#Hey emacs, this is basically -*- perl -*-
#$Id$
## @summary Interface implementation for MemoryStat functions
## @author dl-kernel-qa
## @status private
## @pod ./MemoryStat.pm
## @status 
#   		Change History
##      2011		Initial Checkin
##   12-11-2012         added $verbose_mode value 2, skip logging certain messages
##                        if set the value to 2
## @author jiaz@netapp.com
##


package Os::Memory::MemoryStat;
use strict;
use NACL::APISet;
use Params::Validate
  qw(validate validate_with BOOLEAN HASHREF OBJECT ARRAYREF SCALAR UNDEF);
use Time::Local;
use NATE::Time qw(timestamp);
use Tharn qw($Log);

################################################################################
# Method   : new()
# Objective: Create a new MemoryStat object
# Details  : See POD documentation at the end of this file
################################################################################
sub new {

    #my $class, %args) = @_;
    my $class = shift;
    my ( $self, );

    ## Validate the input
    my %args = validate_with(
        params => \@_,
        spec   => {
            api_set_obj => { isa => 'NACL::APISet' },
            transit_obj => {
                isa      => 'NACL::Transit',
                optional => 1
            },
            verbose_mode => {
                type     => SCALAR,
                optional => 1,
                default  => 0
            },
            load_MemUse => {
                type     => HASHREF,
                optional => 1
            },
            stat_file => {
                type     => SCALAR,
                optional => 1
            },
            timestamp => {
                type     => SCALAR,
                optional => 1
            }
        }
    );

    ### bless instance into class
    $self = bless {}, $class;

    die "Missing required parameter 'api_set_obj'"
      unless ( defined $args{api_set_obj} );

    $self->{api_set_obj} = $args{api_set_obj};
    $self->{transit_obj} = $args{transit_obj} if ( defined $args{transit_obj} );
    $self->{verbose_mode} =
      defined $args{verbose_mode} ? $args{verbose_mode} : 0;

    #Populate stat at creation (from pervious data)
    if ( defined $args{stat_file} ) {
        $self->collect_stat( stat_file => $args{stat_file} );
    }
    else {
        foreach my $column (
            qw (InUse
            MemUse
            HighUse
            Requests
            Sizes
            Pools
            SizeHSEG
            SizePSEG)
          )
        {
            if ( defined $args{"load_$column"} ) {
                my %column_data = %{ $args{"load_$column"} };
                foreach my $mem_type ( keys %column_data ) {
                    $self->{stat}{$mem_type}{$column} = $column_data{$mem_type};
                }
            }
        }
    }

    #Populate s at creation
    if ( defined $args{timestamp} ) {
        $self->{timestamp} = $args{timestamp};
    }

    return $self;
}    #End new()

###############################################################################
# Method:    load_stat
# Objective: load stat information for each memory type shown by vmstat -m
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load_stat {
    my ( $self, %args ) = @_;
    my $state               = $self->_get_transit_state();
    my $api_set_obj         = $self->get_value( name => "api_set_obj" );
    my @vmstat_output       = ();
    my $timestamp           = undef;
    my $verbose_mode        = $self->get_value( name => "verbose_mode" );
    my $d_tag               = "";
    my @dup_mem_type        = ();
    my @dup_mem_type_errors = ();

    $Log->comment("The initial state of the controller is : $state") if ($verbose_mode !=2);

    ## Controller to SYSTEMSHELL if it is not already that state
    if ( defined $state && $state ne 'SYSTEMSHELL' ) {
        $Log->comment("The controller was not in the SYSTEMSHELL state") if ($verbose_mode !=2);
        my @Transit_args = ( to => 'SYSTEMSHELL' );
        unless ( $self->_set_transit_state(@Transit_args) ) {
            $Log->comment( "ERROR : NACL::Transit failed to set controller "
                  . "to systemshell" ) if ($verbose_mode !=2);
        }
    }
    elsif ( defined $state ) {
        $Log->comment("The controller already in the SYSTEMSHELL state") if ($verbose_mode !=2);
    }
    else {
        $Log->comment( "Memory Stat Object has no transit object member, "
              . "no operation needed" ) if ($verbose_mode !=2);
    }

    my $cmd = "vmstat -m";
    $self->{timestamp} = localtime();

    if ( exists $args{saved_vmstat_outfile} ) {

        #get memory stat from a previously saved file
        $cmd = "cat $args{saved_vmstat_outfile}";
        $self->{timestamp} = undef;
    }

    if ( exists $args{new_vmstat_outfile} ) {

        #save memory stat to a file
        my $file = $args{new_vmstat_outfile};
        $cmd .= " > $file; cat $file";
    }

    ## Issue 'vmstat -m' or get vmstat -m output from file
    eval {
        my $output = $api_set_obj->execute_raw_command( command => $cmd );
        @vmstat_output = split( /\n/, $output );
    };
    if ($@) {
        $Log->comment( "$cmd failed with " . $@ );
        return $TCD::FAIL;
    }

    #when using a previous file to load memory stat
    #last line of vmstat output file has TIMESTAMP info
    #get timestamp and delete this line of info
    if ( $vmstat_output[$#vmstat_output] =~ /^TIMESTAMP: (.*)$/ ) {
        $self->{timestamp} = $1;
        delete $vmstat_output[$#vmstat_output];
    }

    #
    #parse info for each memory request type, pop data to $self
    foreach my $line (@vmstat_output) {

        #Originally: /^\s*(\S+[\s*\w*]+)\s+(\d+\s+.*)\s*$/
        #Now:        /^\s*(\S+[\s*\w*]+)\s+([\-]?\d+\s+.*)\s*$/
        #            [\-]? matches zero or 1 '-'
        #To parse if there is a bug causing this column to be a neg number
        if ( $line =~ /^\s*(\S+[\s*\w*]+)\s+([\-]?\d+\s+.*)\s*$/ ) {
            ##each entry contains two parts, the memory type and its info
            ##$1: first column of vmstat output, $1 should start and end with
            ##    non-whitespace character,
            ##    $1 contains letters, numbers, symbols _-.  , spaces;
            ##$2 start with the first non-whitespace string and numbers only
            my $string   = $2;
            my $mem_type = $1;
            $mem_type =~ s/\s+$//;

            # Set $twin_detected to 0
            # if encountered this $mem_type before, set $twin_detected to 1
            my $twin_detected = 0;
            if ( defined $self->{stat}{$mem_type}{string} ) {

                # if 'refresh' is specified, Stat is being loaded again
                # and we expect to see the entries being updated;
                # otherwise we are seeing burt 637293
                if ( !defined $args{refresh} ) {
                    $Log->comment("\nThis is a duplicate name: $mem_type ") if ($verbose_mode != 2);
                    push @dup_mem_type, $mem_type;

                    #InUse  MemUse HighUse Requests Size(s)
                    $twin_detected = 1;
                    $self->{stat}{$mem_type}{string} .= $string;
                }
            }
            else {

                #save the string for debugging
                $self->{stat}{$mem_type}{string} = $string;
            }

            # Go ahead and split the string for more memory statistics
            my @elements = split( /\s+/, $string );

            #Check column number of this line
            #Note: number of elements plus one equal to column number,
            #      the first column is $1 (mem_type)
            #      the rest columns are elements in $2
            if ( @elements != 4 && @elements != 5 && @elements != 7 ) {
                $Log->comment(
                    "Error: wrong format, should have 5, 6 or 8 columns");
                $Log->comment("[$line]");
                return $TCD::FAIL;
            }

            #Populate stats
            #Get the first four columns
            my $msg = "";
            if ($twin_detected) {
                ##update $msg with existing record values
                $msg =
                    "Entry already exists:\n $mem_type InUse "
                  . $self->{stat}{$mem_type}{InUse}
                  . "\n $mem_type MemUse "
                  . $self->{stat}{$mem_type}{MemUse}
                  . "\n $mem_type HighUse "
                  . $self->{stat}{$mem_type}{HighUse}
                  . "\n $mem_type Requests "
                  . $self->{stat}{$mem_type}{Requests};

                #
                ##update $msg with incoming record values
                $msg .=
"\nCombine new values below with existing numbers:\n $mem_type InUse $elements[0], "
                  . "\n $mem_type MemUse $elements[1], "
                  . "\n $mem_type HighUse $elements[2], "
                  . "\n $mem_type Requests $elements[3]";

                #Check values are in good format
                if ( $self->{stat}{$mem_type}{InUse} !~ /\d+/ ) {
                    push @dup_mem_type_errors,
                        "existing InUse value is not in right format"
                      . $self->{stat}{$mem_type}{InUse}
                      . "should be non-negative number";
                }
                if ( $elements[0] !~ /\d+/ ) {
                    push @dup_mem_type_errors,
                        "incoming InUse value is not in right format"
                      . $elements[0]
                      . "should be non-negative number";
                }

                if ( $self->{stat}{$mem_type}{MemUse} !~ /\d+K/ ) {
                    push @dup_mem_type_errors,
                        "existing MemUse value is not in right format"
                      . $self->{stat}{$mem_type}{MemUse}
                      . "should be non-negative number in K";
                }
                if ( $elements[1] !~ /\d+K/ ) {
                    push @dup_mem_type_errors,
                        "Incoming MemUse value is not in right format"
                      . $elements[1]
                      . "should be non-negative number in K";
                }

                if ( $self->{stat}{$mem_type}{Requests} !~ /\d+/ ) {
                    push @dup_mem_type_errors,
                        "existing Requests value is not in right format"
                      . $self->{stat}{$mem_type}{Requests}
                      . "should be non-negative number";
                }
                if ( $elements[3] !~ /\d+/ ) {
                    push @dup_mem_type_errors,
                        "incoming Requests value is not in right format"
                      . $elements[3]
                      . "should be non-negative number";
                }

                ##Add all the numbers
                $self->{stat}{$mem_type}{InUse} += $elements[0];

                #chop the K in MemUse, add the numbers, add K back
                chop( $self->{stat}{$mem_type}{MemUse}, $elements[1] );
                $self->{stat}{$mem_type}{MemUse} += $elements[1];
                $self->{stat}{$mem_type}{MemUse} .= "K";

                if ( $elements[2] eq $self->{stat}{$mem_type}{HighUse} ) {
                    $self->{stat}{$mem_type}{HighUse} = $elements[2];
                }
                else {

                    #This value should always be "-",
                    #If not, we don't give an error here, but concat the value
                    #so we can catch it when calling check_volumn in testcase
                    $self->{stat}{$mem_type}{HighUse} += $elements[2];
                }
                $self->{stat}{$mem_type}{Requests} += $elements[3];

                $msg .=
                    "\nGot new values:\n $mem_type InUse "
                  . $self->{stat}{$mem_type}{InUse}
                  . "\n $mem_type MemUse "
                  . $self->{stat}{$mem_type}{MemUse}
                  . "\n $mem_type HighUse "
                  . $self->{stat}{$mem_type}{HighUse}
                  . "\n $mem_type Requests "
                  . $self->{stat}{$mem_type}{Requests};
            }
            else {
                $self->{stat}{$mem_type}{InUse}    = $elements[0];
                $self->{stat}{$mem_type}{MemUse}   = $elements[1];
                $self->{stat}{$mem_type}{HighUse}  = $elements[2];
                $self->{stat}{$mem_type}{Requests} = $elements[3];
                $msg =
                    "\nObtained these values:\n $mem_type InUse $elements[0] "
                  . "\n $mem_type MemUse $elements[1] "
                  . "\n $mem_type HighUse $elements[2] "
                  . "\n $mem_type Requests $elements[3]";
            }

            if ( @elements == 4 ) {

                # Scenario:  vmstat -m output for nonactive memory use
                # D-smbsetup     0     0K       -      140
                #
                # The first four elements were InUse, MemUse, HighUse, Request
                # already taken in stat
                $Log->comment("$msg") if ($verbose_mode == 1);
            }
            elsif ( @elements == 5 ) {

               # Scenario:  vmstat -m output for active freeBSD memory type
               #        USB    49    16K       -       57  16,32,64,128,512,1024

                if ($twin_detected) {
                    ##Concat the 2 Sizes strings, added comma in between,
                    ##Then update $msg
                    ##Note: as this is a temp workaround 637293
                    ##if a size chunk showed up twice, we still simply concat string,
                    ##will not compare the 2 strings and remove those show up twice
                    ##
                    $msg .= "\nIn addition, existing $mem_type Sizes is "
                      . $self->{stat}{$mem_type}{Sizes};
                    $msg .= "\n Combine new value $mem_type Sizes $elements[4]";
                    $self->{stat}{$mem_type}{Sizes} .= ",$elements[4]";
                    $msg .= "\n Got new $mem_type Sizes "
                      . $self->{stat}{$mem_type}{Sizes};

                }
                else {
                    $self->{stat}{$mem_type}{Sizes} = $elements[4];
                    $msg .=
                      "\n $mem_type Sizes " . $self->{stat}{$mem_type}{Sizes};
                }

                $Log->comment("$msg") if ($verbose_mode == 1);
            }
            elsif ( @elements == 7 ) {
                if ($twin_detected) {
                    ##check Pools
                    ##chop K/, K in SizeHSEG and SizePSEG and add them
                    ##update $msg
                    $msg .=
                      "\nIn addition, more existing values:\n $mem_type Pools "
                      . $self->{stat}{$mem_type}{Pools}
                      . "\n $mem_type SizeHSEG "
                      . $self->{stat}{$mem_type}{SizeHSEG}
                      . "\n $mem_type SizePSEG "
                      . $self->{stat}{$mem_type}{SizePSEG};

                    $msg .=
"\nCombine new values below with existing numbers:\n $mem_type Pools $elements[4], "
                      . "\n $mem_type SizeHSEG $elements[5], "
                      . "\n $mem_type SizePSEG $elements[6] ";

                    if ( $self->{stat}{$mem_type}{SizeHSEG} !~ /\d+K\// ) {
                        push @dup_mem_type_errors,
                            "existing SizeHSEG value is not in right format"
                          . $self->{stat}{$mem_type}{SizeHSEG}
                          . "should be non-negative number in K, trailed by a slash";
                    }
                    if ( $elements[5] !~ /\d+K\// ) {
                        push @dup_mem_type_errors,
                            "Incoming SizeHSEG value is not in right format"
                          . $elements[5]
                          . "should be non-negative number in K, trailed by a slash";
                    }

                    if ( $self->{stat}{$mem_type}{SizePSEG} !~ /\d+K/ ) {
                        push @dup_mem_type_errors,
                            "existing SizePSEG value is not in right format"
                          . $self->{stat}{$mem_type}{SizePSEG}
                          . "should be non-negative number in K";
                    }
                    if ( $elements[6] !~ /\d+K/ ) {
                        push @dup_mem_type_errors,
                            "Incoming SizePSEG value is not in right format"
                          . $elements[6]
                          . "should be non-negative number in K";
                    }

                    if ( $self->{stat}{$mem_type}{Pools} eq $elements[4] ) {
                        $self->{stat}{$mem_type}{Pools} = $elements[4];
                    }
                    else {

                        #This value should always be the same
                        #If not, concat them and catch the error later
                        $self->{stat}{$mem_type}{Pools} .= $elements[4];
                    }

                    chop( $self->{stat}{$mem_type}{SizeHSEG}, $elements[5] );
                    chop( $self->{stat}{$mem_type}{SizeHSEG}, $elements[5] );
                    $self->{stat}{$mem_type}{SizeHSEG} += $elements[5];
                    $self->{stat}{$mem_type}{SizeHSEG} .= "K/";

                    chop( $self->{stat}{$mem_type}{SizePSEG}, $elements[6] );
                    $self->{stat}{$mem_type}{SizePSEG} += $elements[6];
                    $self->{stat}{$mem_type}{SizePSEG} .= "K";

                    $msg .=
                        "\nGot new values:\n $mem_type Pools "
                      . $self->{stat}{$mem_type}{Pools}
                      . "\n $mem_type SizeHSEG "
                      . $self->{stat}{$mem_type}{SizeHSEG}
                      . "\n $mem_type SizePSEG "
                      . $self->{stat}{$mem_type}{SizePSEG};
                }
                else {

         # Scenario:  vmstat -m output for active Dblade memory type
         #   D-libnet    85  1485K       -       96  H/PSEG     1483K/        0K
                    $self->{stat}{$mem_type}{Pools}    = $elements[4];
                    $self->{stat}{$mem_type}{SizeHSEG} = $elements[5];
                    $self->{stat}{$mem_type}{SizePSEG} = $elements[6];
                    $msg .=
                        "\n $mem_type Pools $elements[4]"
                      . "\n $mem_type SizeHSEG $elements[5]"
                      . "\n $mem_type SizePSEG $elements[6]";
                }
                $Log->comment("$msg") if ($verbose_mode == 1);
            }
        }
        elsif ( $line =~ /^\s*D-/ ) {

            #D-tagged mem type listed in wrong format
            $Log->comment("Error: D-tag memory entry listed in wrong format");
            $Log->comment("[$line]");
            return $TCD::FAIL;
        }
        else {
            $Log->comment("Not a memory stat entry: [$line]") if ($verbose_mode == 1);
        }
    }    #End parsing output and populate memory info

    #Add timestamp
    if ( exists $args{new_vmstat_outfile}
        && !exists $args{saved_vmstat_outfile} )
    {
        my $cmd .=
            "echo TIMESTAMP: "
          . $self->{timestamp}
          . " >> $args{new_vmstat_outfile}";
        eval { $api_set_obj->execute_raw_command( command => $cmd ); };
        if ($@) {
            $Log->comment( "$cmd failed with " . $@ );
            return $TCD::FAIL;
        }
    }

    $self->_set_dtag();
    $self->{dup_mem_type}             = [@dup_mem_type];
    $self->{dup_mem_type_check_value} = [@dup_mem_type_errors];
    return $TCD::PASS;
}    #End load_stat

###############################################################################
# Method:    dump_stat
# Objective: dump_stat
# Details:   See POD documentation at the end of this file.
###############################################################################
sub dump_stat {
    my ($self) = @_;
    my %stat = $self->_get_stat();

    foreach my $mem_type ( keys %stat ) {
        $Log->comment("\nMEMORY ALLOCATION FOR [$mem_type]:");
        foreach my $column (
            qw (InUse
            MemUse
            HighUse
            Requests
            Sizes
            Pools
            SizeHSEG
            SizePSEG)
          )
        {
            $Log->comment( "$column : " . $stat{$mem_type}{$column} )
              if ( exists $stat{$mem_type}{$column} );
        }    #End foreach column
    }    #End foreach mem_type
}    #End dump_stat

###############################################################################
# Method:    collect_stat
# Objective: collect_stat
# Details:   parse the output of dump_stat and recreate data
#      See POD documentation at the end of this file.
###############################################################################
sub collect_stat {
    my ( $self, %args ) = @_;
    my $file;
    my $count = 0;
    if ( exists $args{stat_file} ) {

        #save memory stat to a file
        $file = $args{stat_file};
    }
    else {
        $Log->comment(
            "MemoryStat::collect_stat missing required param stat_file");
        return $TCD::FAIL;
    }

    open( MYFILE, $file );
    my $mem_type;
    my $column_name;
    my $column_value;
    foreach my $line (<MYFILE>) {
        chomp;
        if ( $line =~ /MEMORY ALLOCATION FOR \[(.*)\]/ ) {
            $mem_type = $1;
        }
        elsif ( $line =~ /MemUse : (.*)$/ ) {
            $self->{stat}{$mem_type}{MemUse} = $1;
        }
        elsif ( $line =~ /SizePSEG : (.*)$/ ) {
            $self->{stat}{$mem_type}{SizePSEG} = $1;
        }
        elsif ( $line =~ /SizeHSEG : (.*)$/ ) {
            $self->{stat}{$mem_type}{SizeHSEG} = $1;
        }
        else {

            #These lines don't have info of our intrest,
            #Do nothing
        }
    }
    close(MYFILE);
    return $TCD::PASS;
}

###############################################################################
# Method:    compare_stat
# Objective: compare memory stat
# Details:   See POD documentation at the end of this file.
###############################################################################
sub compare_stat {
    my ( $self, %args ) = @_;
    my %base_stat    = $args{base}->_get_stat();
    my %additions    = $args{base}->_get_incoming_pools();
    my %current_stat = $self->_get_stat();
    my @warning      = ();
    my @error        = ();
    my @error_list   = ();
    my @warning_list = ();
    my $verbose_mode = $self->get_value( name => "verbose_mode" );
    my $compare_memory =
      defined $args{Compare_memory} ? $args{Compare_memory} : "all";
    my $percentage = defined $args{percentage} ? $args{percentage} : 10;
    my $threshold  = defined $args{threshold}  ? $args{threshold}  : 2;
    my @filter =
      defined $args{compare_list}
      ? @{ $args{compare_list} }
      : ( keys %current_stat );

    if ( defined $args{kbyte} && !defined $args{threshold} ) {
        $threshold = $args{kbyte};
    }

    if ( defined $args{compare_list} ) {
        $Log->comment(
"PLEASE NOTE: User has specified a list to only monitor the pools below:"
        );
        foreach my $mem_type (@filter) {
            $Log->comment("    $mem_type");
        }
        $Log->comment(
"To monitor pools not listed above, please contact your test developer"
        );
    }

    if ( defined $args{Compare_memory} ) {
        $Log->comment("Comparing only $args{Compare_memory} memory pools");
    }
    else {
        $Log->comment("Comparing both Dblade and Freebsd pools")
          if ($verbose_mode == 1);
    }

    #
    #Compare current output's MemUse of each memtype
    #   with the corresponding baseline MemUse
    #
    foreach my $mem_type (@filter) {

        #Only check memtype that fits $compare_memory
        #  Dblade|freebsd|all
        #    requirement
        if ( $compare_memory =~ /d[-]?blade/i ) {
            next if ( $mem_type !~ /^D-/ );
        }
        elsif ( $compare_memory =~ /freebsd/i ) {
            next if ( $mem_type =~ /^D-/ );
        }

        my $base_mem = undef;
        my $current_mem =
          !defined $args{compare_list}
          ? $current_stat{$mem_type}{MemUse}
          : undef;
        my $current_timestamp = $self->get_value( name => "timestamp" );

        if ( defined $args{compare_list} ) {
            if ( defined $current_stat{$mem_type}{MemUse} ) {
                $current_mem = $current_stat{$mem_type}{MemUse};
            }
            else {
                $Log->comment(
"Please Note: pool \[$mem_type\] in monitor list does not show up in sample,"
                );
                $Log->comment(
"             which could be normal as not all pools are in use at all times."
                );
                $Log->comment(
"             Still, please double check the pool names the list"
                );
                next;
            }
        }

        if (   exists $base_stat{$mem_type}{MemUse}
            || exists $additions{$mem_type}{MemUse} )
        {
            ##Base object has info of $mem_type
            $base_mem =
              ( exists $base_stat{$mem_type}{MemUse} )
              ? $base_stat{$mem_type}{MemUse}
              : $additions{$mem_type}{MemUse};
            chop($base_mem);
            chop($current_mem);
            if ( $base_mem >= $current_mem ) {
                my $msg = "$mem_type did not increase in MemUse";
                $Log->comment("$msg") if ($verbose_mode == 1);
            }
            elsif ($base_mem != 0
                && $base_mem > 0
                && ( $current_mem - $base_mem ) / $base_mem <
                ( $percentage / 100 ) )
            {
                my $msg =
"MemUse by $mem_type did not increase by more than $percentage percent";
                $Log->comment("$msg") if ($verbose_mode == 1);
            }
            elsif ( $current_mem - $base_mem > $threshold ) {

                #default value 2K is to prevents false alarm when system
                #    roundoff a small MemUse
                #    10.45 to 11.5
                #also it does not raise issue for from 1KB to 2KB, that is Okay
                #
                #our main concern is increase from large MemUse, for more than
                # specified $threshold
                my $timestamp_base =
                  $args{base}->get_value( name => "timestamp" );

                $timestamp_base = $additions{$mem_type}{timestamp}
                  if ( exists $additions{$mem_type}{timestamp} );

                my $msg =
                    "Memory Usage Increased: $mem_type MemUse increased "
                  . "from $base_mem KB [at $timestamp_base] to "
                  . "$current_mem KB [at $current_timestamp]";

                ##Check if this is an known issue
                my $double_check = $self->_double_check(
                    baseline    => $base_mem,
                    current     => $current_mem,
                    threshold   => $threshold,
                    memory_type => $mem_type
                );
                next if ( $double_check == $TCD::PASS );
                push( @error,      $msg );
                push( @error_list, $mem_type );
                $Log->comment("$msg") if ($verbose_mode !=2);
            }
            else {

                #MemUse up more than percentage but no more than threshold,
                my $msg =
"OKAY: MemUse by $mem_type increased more than 10 percent, but not exceed threshhold limit ${threshold}K";
                $Log->comment("$msg") if ($verbose_mode == 1);
            }
        }
        else {

            #Base object does not know about $mem_type
            if ( $current_mem ne "0K" ) {

#baseline does not have this memory user, and the user's current_mem is not "0K"
                my $msg =
"Entry for $mem_type was not in previous vmstat -m output, but in a current query "
                  . "with MemUse of $current_mem.";
                push( @warning,      $msg );
                push( @warning_list, $mem_type );
                $Log->comment("$msg") if ($verbose_mode == 1);

                #set value to incoming pools
                #incoming_pool
                #  pool, usage, timestamp
                $Log->comment("Adding $mem_type info to base sample")
                  if ($verbose_mode == 1);
                $args{base}->{incoming_pools}{$mem_type}{MemUse} = $current_mem;
                $args{base}->{incoming_pools}{$mem_type}{timestamp} =
                  $current_timestamp;
            }
            else {

    #baseline does not have this memory user, and the user's current_mem eq "0K"
    #do nothing
            }
        }
    }
    $self->{compare_warning}  = [@warning];
    $self->{compare_error}    = [@error];
    $self->{warning}          = [@warning_list];
    $self->{error}            = [@error_list];
    $self->{compare_basetime} = $args{base}->get_value( name => "timestamp" );
    return ( scalar @error == 0 ) ? $TCD::PASS : $TCD::FAIL;
}    ##End compare_stat()

###############################################################################
# Method:    report_compare_result
# Objective: report compare results
# Details:   See POD documentation at the end of this file.
###############################################################################
sub report_compare_result {
    my ($self) = @_;
    my @error   = @{ $self->get_value( name => "compare_error" ) };
    my @warning = @{ $self->get_value( name => "compare_warning" ) };
    my $banner = "#########################################";
    my @output = ();

    push( @output, $banner );
    my $msg =
      "Compare vmstat -m output at " . $self->get_value( name => "timestamp" );
    push( @output, $msg );
    $msg = "  with vmstat -m output at "
      . $self->get_value( name => "compare_basetime" );
    $msg .= "\nFor memory pools show up after "
      . $self->get_value( name => "compare_basetime" );
    $msg .=
"\n  compare with the pool\'s usage number when it first captured by vmstat -m";
    push( @output, $msg );
    push( @output, $banner );

    if ( scalar @error ) {
        my $err_no = scalar @error;
        push( @output,
"######### Warnings of Excessive Memory Usage Increase: $err_no #########"
        );
        foreach my $err (@error) {
            push( @output, "$err" );
        }
    }
    else {
        my $message =
          "#########No Warnings of Excessive Memory Usage Increase#########";
        push( @output, $message );
    }

    if ( scalar @warning ) {
        my $warn_no = scalar @warning;
        push( @output,
            "######### Warnings of New Memory Usage Entries: $warn_no #########"
        );
        foreach my $warn (@warning) {
            push( @output, "$warn" );
        }
    }

    my $message = "######## End of Warning Messages ########\n";
    push( @output, $message );

    push( @output, $banner );

    if ( scalar @output > 0 ) {
        foreach my $line (@output) {
            $Log->comment("$line");
        }
    }
    return @output;
}    #End report_compare_result

###############################################################################
# Method:    get_memory_type_list
# Objective: get the list of memory type which is in the stat
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_memory_type_list {
    my ($self) = @_;

    if ( !defined $self->{memory_type_list} ) {
        my %stat = $self->_get_stat();
        my @list = keys %stat;
        $self->{memory_type_list} = [@list];
        $self->{num_stat_entry}   = scalar(@list);
    }

    return @{ $self->{memory_type_list} };
}    ## End get_memory_type_list()

###############################################################################
# Method:    get_value
# Objective: get the value from the stat by specifying name, memory_type, etc
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_value {
    my ( $self, %args ) = @_;

    if ( defined $args{name} ) {
        my $name = $args{name};

        if ( $name eq "num_stat_entry" && !defined $self->{$name} ) {
            $self->get_memory_type_list();
        }

        if ( $name =~ /^sum_*/ && !defined $self->{$name} ) {
            $self->_total_dblade_memory_use();
        }

        return $self->{$name};
    }
    elsif ( defined $args{memory_type} && defined $args{column} ) {
        my $mem_type = $args{memory_type};
        my $column   = $args{column};

        if ( $column eq "SizeCACHE" && !defined $self->{cache} ) {
            $self->_calculate_cache_memory();
        }

        if ( $column =~ /sum/ && !defined $self->{sum_dblade} ) {
            $self->_total_dblade_memory_use();
        }

        return $self->{stat}{$mem_type}{$column};
    }
    else {
        return undef;
    }
}    ## End get_value()

###############################################################################
# Method:    check_columns
# Objective: verify memory stat has valid format, not in negative number
# Details:   See POD documentation at the end of this file.
###############################################################################
sub check_columns {
    my ( $self, %args ) = @_;
    my %stat         = $self->_get_stat();
    my @error        = ();
    my $verbose_mode = $self->get_value( name => "verbose_mode" );
    my $check_memory =
      defined $args{Check_memory} ? $args{Check_memory} : "Dblade";

    my @columns_to_check = qw (InUse
      MemUse
      HighUse
      Requests
      Sizes
      Pools
      SizeHSEG
      SizePSEG
    );

    if ( exists $args{column} ) {
        @columns_to_check = @{ $args{column} };
    }

    #
    #   check items in each column are listed
    #    in the right format,
    #    InUse------------positive number or 0
    #    Requests---------positive number or 0
    #    MemUse-----------positive number or 0 in K(Bytes)
    #    SizeHSEG---------positive number or 0 in K(Bytes), trailed by '/'
    #    SizePSEG---------positive number or 0 in K(Bytes)
    #    Pools------------"HSEG/PSEG"  or ""
    #    HighUse----------"-"
    #   If check non-D tagged entries,
    #    Sizes -----------"", or comma seperated positive numbers
    foreach my $mem_type ( keys %stat ) {

        #
        #Only check memtype that fits $check_memory
        #    requirement
        if ( $check_memory =~ /d[-]?blade/i ) {
            next if ( $mem_type !~ /^D-/ );
        }
        elsif ( $check_memory =~ /freebsd/i ) {
            next if ( $mem_type =~ /^D-/ );
        }
        else {
            $Log->comment("No filtering") if ($verbose_mode == 1);
        }

        foreach my $column (@columns_to_check) {
            next unless ( exists $stat{$mem_type}{$column} );
            my $value_to_check = $stat{$mem_type}{$column};

            if ( $column eq "InUse" || $column eq "Requests" )
            {    # these colume expect numbers
                if ( $value_to_check =~ /^\d+$/ ) {
                    my $msg =
"$mem_type $column value $value_to_check is 0 or a positive number";
                    $Log->comment("$msg") if ($verbose_mode == 1);
                }
                elsif ( $value_to_check =~ /^-\d+$/ ) {
                    my $msg =
"Error: $mem_type $column value $value_to_check is a negative number, should be 0 or more";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }
                else {
                    my $msg =
"Error: $mem_type $column value $value_to_check is not a number";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }
            }
            elsif ($column eq "MemUse"
                || $column eq "SizeHSEG"
                || $column eq "SizePSEG" )
            {    #these colume expect numbers in K, SizeHSEG has a trailing /
                my $trail_char = "";
                $trail_char = "\/" if $column eq "SizeHSEG";
                if ( $value_to_check =~ /^\d+K$trail_char$/ ) {
                    my $msg =
"$mem_type $column value $value_to_check is 0 or a positive number";
                    $Log->comment("$msg") if ($verbose_mode == 1);
                }
                elsif ( $value_to_check =~ /^-\d+K$trail_char$/ ) {
                    my $msg =
"Error: $mem_type $column value $value_to_check is a negative number, should be 0 or more";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }
                else {
                    my $msg =
"Error: $mem_type $column value $value_to_check is not a number";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }
            }
            elsif ( $column eq "Pools" ) {
                if ( $value_to_check eq "H\/PSEG" ) {
                    my $msg =
                      "$mem_type $column value $value_to_check is correct";
                    $Log->comment("$msg") if ($verbose_mode == 1);
                }
                else {
                    my $msg =
"Error: $mem_type $column value should be H\/PSEG, not $value_to_check";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }
            }
            elsif ( $column eq "HighUse" ) {
                if ( $value_to_check eq "-" ) {
                    my $msg =
                      "$mem_type $column value $value_to_check is correct";
                    $Log->comment("$msg") if ($verbose_mode == 1);
                }
                else {
                    my $msg =
"Error: $mem_type $column value should be -, not $value_to_check";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }
            }
            elsif ( $column eq "Sizes" )
            { #check freebsd memory allocation has comma seperated numbers in 'Sizes'
                if ( $value_to_check =~ /^\d+[,\d+]?/ ) {
                    my $msg =
                      "$mem_type $column value $value_to_check is correct";
                    $Log->comment("$msg") if ($verbose_mode == 1);
                }
                else {
                    my $msg =
"Error: $mem_type $column value should be positive numbers seperated by comma, not $value_to_check";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }
            }
            else {
                my $msg = "Invalid column show up: $column";
                push( @error, $msg );
                $Log->comment("$msg");
            }
        }
    }
    return (@error);
}    #End check_columns()

###############################################################################
# Method:    check_entries
# Objective: check request numbers and memory usages for D-blade memory types
# Details:   See POD documentation at the end of this file.
###############################################################################
sub check_entries {
    my ( $self, %args ) = @_;
    my %stat = $self->_get_stat();

    my $verbose_mode = $self->get_value( name => "verbose_mode" );

    my @check_rules = qw (Request
      Memory
    );
    my @error = ();

    if ( exists $args{rule} ) {
        @check_rules = @{ $args{rule} };
    }

    #
    #Compare current output's MemUse of each memtype
    #   with the corresponding baseline MemUse
    #
    foreach my $mem_type ( keys %stat ) {

        #Replace "next if ($mem_type!~/^D-/); " with a closest condition to skip
        #   freebsd users which have info in Sizes if they use memory
        next if ( exists $stat{$mem_type}{Sizes} );

        foreach my $rule (@check_rules) {
            if ( $rule eq "Request" ) {

                # InUse <= Requests
                if (   !exists $stat{$mem_type}{InUse}
                    || !exists $stat{$mem_type}{Requests} )
                {
                    my $msg =
"Missing InUse column or Request column info for $mem_type";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }

                if ( $stat{$mem_type}{InUse} <= $stat{$mem_type}{Requests} ) {
                    my $msg =
                        "For $mem_type, Memory InUse $stat{$mem_type}{InUse}"
                      . "is not greater than Memory Requests $stat{$mem_type}{Requests}";
                    $Log->comment("$msg") if ($verbose_mode == 1);
                }
                else {
                    my $msg =
"Error: For $mem_type, Memory InUse $stat{$mem_type}{InUse}"
                      . "is greater than Memory Requests $stat{$mem_type}{Requests}";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }
            }
            elsif ( $rule eq "Memory" ) {

                # make sure SizeHSEG+SizePSEG <= MemUse
                my $hseg = "0K/";
                my $pseg = "0K";
                if ( !exists $stat{$mem_type}{MemUse} ) {
                    my $msg =
"Missing MemUse column or Request column info for $mem_type";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }
                my $used = $stat{$mem_type}{MemUse};
                $hseg = $stat{$mem_type}{SizeHSEG}
                  if ( exists $stat{$mem_type}{SizeHSEG} );
                $pseg = $stat{$mem_type}{SizePSEG}
                  if ( exists $stat{$mem_type}{SizePSEG} );

                chop( $used, $hseg, $hseg, $pseg );
                if ( $hseg + $pseg < $used + 1 )
                {    # add 1 to prevent false error caused by round off
                    my $msg =
"For $mem_type, memory used in PSEG [$pseg K] and HSEG [$hseg K] is not greater "
                      . "all memory in use [$used K]";
                    $Log->comment("$msg") if ($verbose_mode == 1);
                }
                else {
                    my $msg =
"Error: For $mem_type, memory used in PSEG [$pseg K] and HSEG [$hseg K] is greater than "
                      . "MemUse [$used K]";
                    push( @error, $msg );
                    $Log->comment("$msg");
                }
            }
            else {
                my $msg = "Error: Invalid rule selection: $rule";
                push( @error, $msg );
                $Log->comment("$msg");
            }
        }
    }

    return (@error);
}    #End check_entries()

##################END Public Methods############################################

################################################################################
#####################PRIVATE METHODS############################################
#####################private function name starts with underscore###############
#####################to be used within the module###############################
################################################################################
# Method:    _get_stat
# Objective: Retrieve a hash with statistics of each memory type
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_stat {
    my ($self) = @_;
    my %stat = ();

    if ( defined $self->{stat} ) {
        %stat = %{ $self->{stat} };
    }

    return %stat;
}    #End _get_stat()

################################################################################
#####################PRIVATE METHODS############################################
#####################private function name starts with underscore###############
#####################to be used within the module###############################
################################################################################
# Method:    _get_incoming_pools
# Objective: Retrieve a hash with statistics of memory pools that
#             exist in the objects that compare_stat() with $self
#             but not exist in $self
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_incoming_pools {
    my ($self) = @_;
    my %stat = ();

    if ( defined $self->{incoming_pools} ) {
        %stat = %{ $self->{incoming_pools} };
    }

    return %stat;
}    #End _get_incoming_pools()

################################################################################
# Method:    _get_stat
# Objective: Retrieve a hash with statistics of each memory type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _set_dtag {
    my ( $self, %args ) = @_;
    my %memory_stat = $self->_get_stat();

    $self->{Dtag} = "";

    foreach my $mem_type ( keys %memory_stat ) {
        if ( $mem_type =~ /^D-/ ) {
            $self->{Dtag} = "D-";
            last;
        }
    }
}

###############################################################################
# Routine : _calculate_cache_memory
# Objective: For each D-tag memory type, calculate its cache use
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _calculate_cache_memory {
    my ( $self, %args ) = @_;
    my %memory_stat = $self->_get_stat();

    foreach my $mem_type ( keys %memory_stat ) {

        #Replace with a closest condition to skip freebsd users,
        #   freebsd users have info in Sizes if they use memory
        next if ( exists $memory_stat{$mem_type}{Sizes} );

        my $cache = "0K";
        my $hseg  = "0K/";
        my $pseg  = "0K";
        my $used  = $memory_stat{$mem_type}{MemUse};

        #Change the strings into numbers
        #"D-libnet    85  1485K       -       96  H/PSEG     1483K/        0K"
        #$hseg needs two chops: 1483K/ ---> 1483
        #$pseg needs one chop: 0K  ---> 0
        #$used needs one chop: 1485K ---> 1485
        $hseg = $memory_stat{$mem_type}{SizeHSEG}
          if ( exists $memory_stat{$mem_type}{SizeHSEG} );
        $pseg = $memory_stat{$mem_type}{SizePSEG}
          if ( exists $memory_stat{$mem_type}{SizePSEG} );
        chop( $used, $hseg, $hseg, $pseg );

        if ( $used - ( $hseg + $pseg ) > 0 ) {
            $cache = $used - ( $hseg + $pseg );
        }

        $self->{stat}{$mem_type}{SizeCACHE} = "$cache" . "K";
    }
    $self->{cache} = 1;
    return $TCD::PASS;
}    #End _calculate_cache_memory ()

###############################################################################
# Routine : _total_freebsd_memory_use
# Objective: For freebsd memory pools, calculate the sum of MemUse.
#            This is by adding the MemUse of all the memory consumer types
#            (first column in the vmstat -m output) that does not start with "D-"
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _total_freebsd_memory_use {
    my ( $self, %args ) = @_;
    my %memory_stat = $self->_get_stat();

    my $freebsd_memory = 0;
    foreach my $mem_type ( keys %memory_stat ) {

        #skip Dblade users,
        next if ( $mem_type =~ /D-/ );

        my $used = $memory_stat{$mem_type}{MemUse};

        #Change the strings into numbers
        #"D-libnet    85  1485K       -       96  H/PSEG     1483K/        0K"
        #$hseg needs two chops: 1483K/ ---> 1483
        #$pseg needs one chop: 0K  ---> 0
        #$used needs one chop: 1485K ---> 1485
        chop($used);

        $freebsd_memory += $used;
    }

#There are 2 dynamic memory allocators in freebsd.
#One is zone malloc (uma) tracked by 'vmstat -z', one is the malloc tracked by 'vmstat -m'.
#And within 'vmstat -z' it also includes items listed by 'vmstat -m',
#so we have entanglement which is not easy to unravel.
#Thus "sum_freebsd" below is all the memory allocated by malloc within freebsd.
    $self->{sum_freebsd} = $freebsd_memory;
    return $TCD::PASS;
}    #End _total_freebsd_memory_use ()

###############################################################################
# Routine : _total_dblade_memory_use
# Objective: sum all D-tag memory types' use in PSEG, HSEG, CACHE
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _total_dblade_memory_use {
    my ( $self, %args ) = @_;
    my %memory_stat = $self->_get_stat();
    my $verbose_mode = $self->get_value( name => "verbose_mode" );

    my $sum_cache         = 0;
    my $sum_hseg          = 0;
    my $sum_pseg          = 0;
    my $sum_used          = 0;
    my $sum_pre_wafl_pseg = 0;
    my $sum_wsb           = 0;

    if ( !defined $self->{cache} ) {
        $self->_calculate_cache_memory();
    }

    if ( !defined $self->{sum_freebsd} ) {
        $self->_total_freebsd_memory_use();
    }

    foreach my $mem_type ( keys %memory_stat ) {
        next
          if (
            $mem_type !~ /^D-/    #Ontap D-
            || $mem_type =~ /^D-KV/    #skip if D-KV pseg|hseg|cache_alloc
          );

        my $hseg          = "0K/";
        my $pseg          = "0K";
        my $used          = $memory_stat{$mem_type}{MemUse};
        my $cache         = "0K";
        my $pre_wafl_pseg = "0K";

        #This is an example
        #	D-raidv2   2502 7191K     - 16657383  H/PSEG     6672K/        4K
        # $used is "7191K" for "D-raidv2"

        #Get hseg info, "D-raidv2" has $hseg "6672K/"
        #
        if ( exists $memory_stat{$mem_type}{SizeHSEG} ) {
            $hseg = $memory_stat{$mem_type}{SizeHSEG};
            $Log->comment(
                "$mem_type HSEG value $hseg to be added to the sum of HSEG")
              if ($verbose_mode == 1);
        }

  # Get hseg info, "D-raidv2" has $pseg "4K", note that its $pre_wafl_pseg is 0K
  # If the pool name matches /D-sk.*/i, then $pre_wafl_pseg is
        if ( exists $memory_stat{$mem_type}{SizePSEG} ) {

# is this a pre-wafl allocation, all the pool names start with D-Sk or D-sk are pre-wafl allocation
# Note there is an exception, "D-sk" is for post wafl alloc. We use \s+ in the match pattern to make sure there are more letters to it.
            if ( $mem_type =~ /^D-sk \S+/i ) {
                $pre_wafl_pseg = $memory_stat{$mem_type}{SizePSEG};
                $Log->comment(
"$mem_type PSEG value $pre_wafl_pseg to be added to the sum of pre-wafl PSEG"
                ) if ($verbose_mode == 1);
            }
            elsif ( $mem_type =~ /^D-ssd \S+/ )
            {    #D-ssd pvbn hash is also pre-wafl
                $pre_wafl_pseg = $memory_stat{$mem_type}{SizePSEG};
                $Log->comment(
"$mem_type PSEG value $pre_wafl_pseg to be added to the sum of pre-wafl PSEG"
                ) if ($verbose_mode == 1);
            }
	    elsif ( $mem_type =~ /^D-mbuf.*/)
	    ##update per fix of burt659517
	    {
                $pre_wafl_pseg = $memory_stat{$mem_type}{SizePSEG};
                $Log->comment(
"$mem_type PSEG value $pre_wafl_pseg to be added to the sum of pre-wafl PSEG"
                ) if ($verbose_mode == 1);		
	    }
            else {
                $pseg = $memory_stat{$mem_type}{SizePSEG};
                $Log->comment(
                    "$mem_type PSEG value $pseg to be added to the sum of PSEG")
                  if ($verbose_mode == 1);
            }
        }

     #
     #$cache = MemUse - HSEG - PSEG, should be calculated and saved to SizeCACHE
     # already
        $cache = $memory_stat{$mem_type}{SizeCACHE}
          if ( exists $memory_stat{$mem_type}{SizeCACHE} );
        chop( $used, $hseg, $hseg, $pseg, $cache, $pre_wafl_pseg );

        $sum_pseg += $pseg;
        $Log->comment(
            "$mem_type PSEG value $pseg K has been added to the sum of PSEG")
          if ($verbose_mode == 1);
        $sum_hseg += $hseg;
        $Log->comment(
            "$mem_type HSEG value $hseg K has been added to the sum of HSEG")
          if ($verbose_mode == 1);
        $sum_cache         += $cache;
        $sum_used          += $used;
        $sum_wsb           += $used if ( $mem_type =~ /D-wsb/ );
        $sum_pre_wafl_pseg += $pre_wafl_pseg;
    }

#'sum_system' is sum of all the malloc'd memory tracked by 'vmstat -m'
#including freebsd and Dblade
#Some of these numbers are not used in any of our test scripts,
#In the future, if there is need to use all these numbers in testing, we'll
#create a separate function that calculates the total memory(freebsd + D-blade).
    $self->{sum_system} = $sum_used + $self->{sum_freebsd};
    my $msg = "";
    $msg .= " sum_pre_wafl-pseg ${sum_pre_wafl_pseg}K\n"
      if ( $sum_pre_wafl_pseg > 0 );
    $msg .= " sum_wsb ${sum_wsb}K\n" if ( $sum_wsb > 0 );
    $msg .=
        " sum_pseg: ${sum_pseg}K\n"
      . " sum_hseg: ${sum_hseg}K\n"
      . " sum_cache: ${sum_cache}K\n\n"
      . " dynamic allocated Dblade memory: ${sum_used}K\n"
      . " dynamic allocated freebsd memory: "
      . $self->{sum_freebsd} . "K\n"
      . " total dynamic allocated memory: "
      . $self->{sum_system} . "K\n";

    $Log->comment("$msg") if ($verbose_mode != 2);

    my $dtag = $self->{Dtag};

    $self->{stat}{"${dtag}KV cache_alloc"}{sum} = "${sum_cache}K";
    $self->{stat}{"${dtag}KV hseg"}{sum}        = "${sum_hseg}K";
    $self->{stat}{"${dtag}KV pseg"}{sum}        = "${sum_pseg}K";

#NOTE: ideally we should also have sum of pre-wafl pseg, however, there hasn't been an entry of pre-wafl in vmstat -m.
#
    $self->{sum_dblade}        = $sum_used;
    $self->{sum_wsb}           = $sum_wsb;
    $self->{sum_pre_wafl_pseg} = $sum_pre_wafl_pseg;
    my $hseg_string = $memory_stat{"${dtag}KV hseg"}{SizeHSEG};

    #For example, a $hseg_string is '123K/', chop '/'
    chop($hseg_string);

    $msg =
        "${dtag}KV cache_alloc MemUse: "
      . $memory_stat{"${dtag}KV cache_alloc"}{MemUse}
      . "\n${dtag}KV hseg HSEG size: $hseg_string"
      . "\n${dtag}KV pseg PSEG size: "
      . $memory_stat{"${dtag}KV pseg"}{SizePSEG};

    #NOTE: there is no entry for pre-wafl in vmstat -m.

    $Log->comment("$msg");
    return $TCD::PASS;
}    #End _total_dblade_memory_use()

###############################################################################
# Method:    _get_transit_state
# Objective: get filer transit state
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _get_transit_state {
    my ($self) = @_;

    if ( !defined $self->{transit_obj} ) {

        #   $self's api_set_obj member can be initiated with a console connect
        #or with a hostobj thus have ssh|rsh connection, or even more.
        #   For ssh and rsh, _transit_to() and ->change_state won't transit
        #their states as transit_to works through console connection.
        #   So the solution for rsh and ssh is to not pass a transit_obj member
        #when init MemoryStat object, and set ssh|rsh to the right state
        #(in the script) before passing to new() the MemoryStat obj.
        #   As to the return value of _get_transit_state, return undef
        #as we cannot do any transit_state operation here.
        return (undef);
    }
    return ( $self->{transit_obj}->get_state() );
}    #End _get_transit_state()

###############################################################################
# Method:    _set_transit_state
# Objective: Set filer to desired state
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _set_transit_state {
    my ( $self, @Transit_args ) = @_;

    if ( !defined $self->{transit_obj} ) {

        #   $self's api_set_obj member can be initiated with a console connect
        #or with a hostobj thus have ssh|rsh connection, or even more.
        #   For ssh and rsh, _transit_to() and ->change_state won't transit
        #their states as transit_to works through console connection.
        #   So the solution for rsh and ssh is to not pass a transit_obj member
        #when init MemoryStat object, and set ssh|rsh to the right state
        #(in the script) before passing to new() the MemoryStat obj,
        #and always mark _set_transit_state() succeed in this case, a.k.a
        #return $TCD::PASS here.
        return ($TCD::PASS);
    }

    return ( $self->{transit_obj}->change_state(@Transit_args) );
}    #End _set_transit_state()

###############################################################################
# Method:    _double_check
# Objective: Check
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _double_check {
    my ( $self, %args ) = @_;
    my $baseline  = $args{baseline};
    my $current   = $args{current};
    my $dtag      = $self->{Dtag};
    my $mem_type  = $args{memory_type};
    my $threshold = $args{threshold};
    my $verbose_mode        = $self->get_value( name => "verbose_mode" );    

    my %known_issues = (    #burts
        "${dtag}cm"           => "burt 533934",
        "${dtag}kmod exports" => "burt 533928",
        "${dtag}cmds"         => "burt 543807",

        #freebsd components, some has a bigger flutuation than 10%,
        "UMAHash"    => "threshold set to 50K",
        "umtx"       => "threshold set to 50K",
        "md_sectors" => "threshold set to 50K",

        #Ontap components
        #some has a bigger flutuation than 10%,
        #Some takes long time to settle in
        "${dtag}kma cache ovrhd" => "threshold set to 100K",
        "${dtag}tcp"      => "component takes 12:00:00 12 hours to settle",
        "${dtag}adminapi" => "threshold set to 40K",
        "${dtag}wafl"     => "threshold set to 200000K",
        "${dtag}raidv2"   => "threshold set to 4000K",
        "${dtag}KV pseg"  => "threshold set to 20000K",
        "${dtag}fio"      => "check current not exceed 2200K"
    );

    my $solution = undef;
    if ( exists $known_issues{$mem_type} ) {
        $solution = $known_issues{$mem_type};
        if ( $solution =~ /burt \d+/ ) {
            $Log->comment(
                "   $solution has been filed for component $mem_type") if ($verbose_mode !=2);
            return $TCD::PASS;
        }
        elsif ( $solution =~ /component takes.*to settle/ ) {
            $Log->comment("   OKAY $mem_type $solution")  if ($verbose_mode !=2);
            return $TCD::PASS;
        }
        elsif ( $solution =~ /threshold set to (\d+)K/ ) {
            my $new_threshold = $1;
            $new_threshold =
              $threshold > $new_threshold ? $threshold : $new_threshold;
            my $increase = $current - $baseline;
            if ( $increase > $new_threshold ) {
                $Log->comment("   $mem_type $solution")
                  if ( $threshold < $new_threshold ) && ($verbose_mode != 2);
                $Log->comment(
"   Not OKAY $mem_type memory use increases by ${increase}K which exceeds threshold ${new_threshold}K"
                ) if ($verbose_mode !=2);
                return $TCD::FAIL;
            }
            else {
                $Log->comment("   $mem_type: $solution")
                  if ( $threshold < $new_threshold ) && ($verbose_mode != 2);
                $Log->comment(
"   OKAY $mem_type memory use increases by ${increase}K within threshold ${new_threshold}K"
                )  if ($verbose_mode !=2);
                return $TCD::PASS;
            }
        }
        elsif ( $solution =~ /check current use not exceed (\d+)K/ ) {
            my $limit = $1;
            if ( $current > $limit ) {
                $Log->comment(
"   This compoment [$mem_type] takes time to settle, it is okay if memory use not over ${limit}K."
                );
                $Log->comment(
"   Not OKAY $mem_type memory use is ${current}K, exceeds limit ${limit}K"
                );
                return $TCD::FAIL;
            }
            else {
                $Log->comment(
"   This compoment [$mem_type] takes time to settle, it is okay if memory use not over ${limit}K."
                ) if ($verbose_mode !=2);
                $Log->comment(
"   OKAY $mem_type memory use is ${current}K, does not exceeed limit ${limit}K"
                ) if ($verbose_mode !=2);
                return $TCD::PASS;
            }
        }
        else {
            $Log->comment("   $solution");
            return $TCD::FAIL;
        }
    }
    else {
        return $TCD::FAIL;
    }

}    #End _double_check()

1;

############################################################################
#  P O D
############################################################################

=pod

=head1 Package Name

Os::Memory::MemoryStat

=head1 Author

dl-kernel-qa@netapp.com

=head1 Description

Os::Memory::MemoryStat provides a API interface for getting and checking 'vmstat -m'
output.

=head1 See Also

Package is Node agnostic. Supports two connection types

B<Console> console

OR

B<Ssh> ssh

=head1 Public Methods

=head2 B<new( B<api_set_obj>, I<transit_obj, verbose_mode> )>

Creates a new Os::Memory::MemoryStat object.

=over 4

=item Inputs

=over 4

=item B<api_set_obj>

Required : Name of api_set_obj executing command vmstat -m

=item B<transit_obj>

Optional : need this if api_set_obj has console connection

=item B<verbose_mode> (default: 0)

Optional : prints debug message if set to 1

=back

=item Outputs

Returns a Os::Memory::MemoryStat object.

=item Examples

 ### Create a MemoryStat object
 use TCD;
 use Os::Memory::MemoryStat;
 ...
	$MSObj = Os::Memory::MemoryStat->new (
					api_set_obj => $Api_set_obj,
			                transit_obj => $Transit_obj
	);

=back


=head2 B<load_stat( I<saved_vmstat_outfile,new_vmstat_outfile> )>

issue 'vmstat -m', populate object with output info.

=over 4

=item Inputs

=over 4

=item  B<saved_vmstat_outfile> (Optional)

filename of a previously saved vmstat output file

=item  B<new_vmstat_outfile> (Optional)

filename of the file user wants to save current 'vmstat -m' output

=back

=item Outputs

Returns $TCD::PASS|$TCD::FAIL

=item Examples

 $MSObj->load_stat();

=back


=head2 B<dump_stat( I< > )>

print stat info in object

=over 4

=item Inputs

=over 4

None

=back

=item Outputs

None

=item Examples

 $MSObj->dump_stat();

=back


=head2 B<collect_stat( I<stat_file> )>

parse stat info generated by dump_stat

=over 4

=item Inputs

=over 4

=item B<stat_file>

Name of the file that contains dump_stat() result

=back

=item Outputs

Returns $TCD::PASS

=item Examples

 $MSObj->collect_stat(stat_file=>"");

=back


=head2 B<compare_stat( I<base> )>

Compare the object with another object, if memory usage of a memory type
increases more than 10%, generate a error message, returns TCD::PASS|FAIL
at the end

=over 4

=item Inputs

=over 4

=item B<base>  (Required)

MemoryStat Object

=item B<percentage> (Optional)

If the memory usage increase more than this percentage,
and increase more than threshold in Kbyte, raise an error.
default is 10 percent. 

=item B<threshold> (Optional)

If the memory usage increase more than this percentage,
and increase more than threshold in Kbyte, raise an error.
default is 2 (Kbyte). 

=item B<compare_list> (Optional)

reference to an array of pool names. If specify this list, we only compare
those pools that are listed. It works like a filter.

=back

=item Outputs

$TCD::PASS or $TCD::FAIL

=item Examples

 $MSObj->compare_stat(base=>$MSObj2);

=back


=head2 B<report_compare_result()>

Prints compare stat result.

=over 4

=item Inputs

=over 4

=item 

None

=back

=item Outputs

@message : output message

=item Examples

 @report = $MSObj->report_compare_result();

=back


=head2 B<get_memory_type_list(  )>

Retrieve a list of all memory types of the Object

=over 4

=item Inputs

=over 4

None

=back

=item Outputs

Returns a list of all memory types of the Object

=item Examples

 $MSObj->get_memory_type_list();

=back


=head2 B<get_value( I<memory_type>, I<column>  I<|name>)>

Return a value within the object, either by name, or by the memory type with
column.

=over 4

=item Inputs

=over 4

=item B<memory_type> 

A scalar, which is a memory type listed by vmstat -m

=begin html

<pre>
D-regex,
D-kmod cmds,
...
</pre>

=end html

=item B<column> 

An scalar, column names given by vmstat -m, or 

=begin html

<pre>
InUse,
MemUse,
HighUse,
Requests,
Sizes,
Pools,
SizeHSEG,
SizePSEG,
sum
</pre>

=end html


=item B<name> 

object member's name 

=begin html

<pre>
num_stat_entry,
api_set_obj,
verbose_mode
</pre>

=end html

=back

=item Outputs

Returns value of the specified member, or the specified memory info

=item Examples

$api_set_obj = $self->get_value(name=>"api_set_obj");

$hseg_dkv   = $Stat_quiescent -> get_value (memory_type => "D-KV hseg",
						   column      => "SizeHSEG");

=back


=head2 B<check_columns( I<column> )>

Check if data in specified column(s) is valid, return error message

=over 4

=item Inputs

=over 4

=item B<column>
default is to include all columns in the array ref:
[InUse,MemUse,HighUse,Requests,Sizes,Pools,SizeHSEG,SizePSEG]

=begin html

<pre>
reference to array of columns: [column1, column2,...]
InUse,Requests should be numbers
HighUse should be -
MemUse, SizeHSEG, SPSEG should be memory usage in K, can not be negative
Pools should be empty or HSEG/PSEG
Sizes should be comma seperated numbers
</pre>

=end html

=back

=item Outputs

Returns back an array of error messages

=item Examples

    my @check_results = $MSObj ->check_columns(
			        column => ["Pools",
		                          "SizeHSEG",
		                          "SizePSEG"
					  ]);

    my @check_results = $MSObj ->();

=back


=head2 B<check_entries()>

check request numbers and memory usages for D-blade memory types

=over 4

=item Inputs

=over 4

=item B<rule>

reference of arrary of rules

=begin html

<pre>
Request(checks InUse number is less than or equal to requests number),
Memory (MemUse is greater or equal to the sum of HSEG use and PSEG use)
</pre>

=end html

=back

=item Outputs

Returns back an array of error messages 

=item Examples

 $MSObj->check_entries();
 $MSObj->check_entries(rule => [Requests]);

=back


=head2 B<_get_stat()>

Retrieve a hash, the key words are a list of memory types given by vmstat -m,
each memory type has all values given by vmstat -m, such as InUse, Request

=over 4

=item Inputs

=over 4

None

=back

=item Outputs

Returns a hash

=item Examples

 my %hash_stat=$MSObj->_get_stat();

=back


=head2 B<_get_incoming_pools()>

Retrieve a hash, the keys are a list of memory pools that did not show up
at object self's sample time, but show up in objects that compare_stat() with
self. Each pool has MemUse and TimeStamp info which comes from the objects that
call function compare_stat() and use self as base.

=over 4

=item Inputs

=over 4

None

=back

=item Outputs

Returns a hash

=item Examples

 my %hash_stat=$MSObj->_get_incoming_pools();

=back


=head2 B<_set_dtag()>

Set member 'Dtag' value of the object. Either 'D-' or ''.

=over 4

=item Inputs

=over 4

None

=back

=item Outputs

None

=item Examples

 $MSObj->_set_dtag();

=back

=head2 B<_calculate_cache_memory()>

For each D-tag memory type, calculate its cache use
Cache=MemUse-HSEG-PSEG, then save it into 
$self->{stat}{$mem_type}{SizeCACHE}

=over 4

=item Inputs

=over 4

None

=back

=item Outputs

None

=item Examples

 $MSObj->_calculate_cache_memory();

=back

=head2 B<_total_freebsd_memory_use()>
Add up all the MemUse value of those types NOT pre-fixed by D-.
This is the total memory usage used by freebsd kernel modules.
=over 4

=item Inputs

=over 4

None

=back

=item Outputs

None

=item Examples

  $MSObj->_total_freebsd_memory_use();

=back

=head2 B<_total_dblade_memory_use()>

Add up all HSEG usage by D-tag memory types
Add up all PSEG usage by D-tag memory types
Add up all CACHE usage by D-tag memory types
then populate $self->{stat}{"D-KV cache_alloc"|"D-KV hseg"|"D-KV pseg"}{sum}
Then add all three above up, populate $self->{sum_dblade}

=over 4

=item Inputs

=over 4

None

=back

=item Outputs

None

=item Examples

  $MSObj->_total_dblade_memory_use();

=back


=head2 B<_get_transit_state()>

Get filer transit state

=over 4

=item Inputs

=over 4

None

=back

=item Outputs

Returns filer state.

=item Examples

$MSObj->_get_transit_state();

=back


=head2 B<_set_transit_state()>

Set filer to desired state

=over 4

=item Inputs

=over 4

=item * B<@Transit_args>

@Transit_args=(to=>"CLI")

=back

=item Outputs

$TCD::PASS or $TCD::FAIL

=item Examples

$result = $self->_set_transit_state(to=>"SYSTEMSHELL");

=back

=head2 B<_double_check()>

Double check if the memory user is causing a real issue

=over 4

=item Inputs

=over 4

=item B<baseline>

baseline number to compare with

=item B<current> 

current memory use

=item B<threshold> 

a proposed number that memory increase should not exceed

=item B<memory_type> 

memory user (first column in vmstat -m)

=back

=item Outputs

$TCD::PASS or $TCD::FAIL

=item Examples

    $result = $self->_double_check(baseline => $base_mem,
	  			current  => $current_mem,
				threshold=> $threshold,
				memory_type =>$mem_type);

=back

=cut

