# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Vserver Audit module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: PTEFIT::Cifs::Misfit::Features::Audit
# Wrapper for NACL Audit with Misfit messaging

package PTEFIT::Cifs::Misfit::Features::Audit;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::Vserver;
use NACL::CS::Vserver;
use NACL::C::VserverAudit;
use NACL::STask::VserverAudit;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;
use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a new Audit Misfit object
#
#  Returns:
#      A new Audit Misfit object
#
#  Example:
#
#      > my $Audit = PTEFIT::Cifs::Misfit::Features::Audit->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "PTEFIT::Cifs::Misfit::Features::Audit";
    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }
    $self->{node} = $node;


    return $self;
} # End of new

#===============================================================================
#  Function: create
#   This method creates Audit on a vserver or list of vservers depending on the
#   input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to setup audit on a vserver
#      count - The number of vservers to create audit on
#      filter - variable to filter through vservers to create audit on
#      params - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      Audit Attributes:
#        'vserver' - Vserver name (required)
#        'destination-path' - Path to store the audit XML output (required)
#        'rotate-limit' - Number of rotates on the audit XML file (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $Audit->create(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              destination  => '/test_audit',
#      >              'rotate-limit' => '0'
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub create {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create";
        } else {
            $runid = $vserver->{vserver} . "_create";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_create(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Audit create failed");
        return FALSE;
    }
} # End of create

#===============================================================================
#  Function: enable
#   This method enables Audit on a vserver or list of vservers depending on the
#   input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to enable audit on a vserver
#      count - The number of vservers to enable audit on
#      filter - variable to filter through vservers to enable audit on
#      params - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run enable
#
#      Audit Attributes:
#        'vserver' - Vserver name (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $Audit->enable(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >          }
#      >      ]
#      >  );
#===============================================================================

sub enable {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );

    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_enable";
        } else {
            $runid = $vserver->{vserver} . "_enable";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_enable(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "enable",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Audit enable failed");
        return FALSE;
    }
} # End of enable

#===============================================================================
#  Function: disable
#   This method disables Audit on a vserver or list of vservers depending on the
#   input parameters.  This will generate an audit XML file in the destionation.
#
#   Parameters:
#      objects - Array of hashes containing information to disable audit on a vserver
#      count - The number of vservers to disable audit on
#      filter - variable to filter through vservers to disable audit on
#      params - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run disable
#
#      Audit Attributes:
#        'vserver' - Vserver name (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $Audit->disable(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >          }
#      >      ]
#      >  );
#===============================================================================

sub disable {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );

    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_disable";
        } else {
            $runid = $vserver->{vserver} . "_disable";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_disable(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "disable",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Disable audit failed");
        return FALSE;
    }
} # End of disable

#===============================================================================
#  Function: delete
#   This method deletes Audit on a vserver or list of vservers depending on the
#   input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to delete audit on a vserver
#      count - The number of vservers to delete audit on
#      filter - variable to filter through vservers to delete audit on
#      params - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run delete
#
#      Audit Attributes:
#        'vserver' - Vserver name (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $Audit->delete(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >          }
#      >      ]
#      >  );
#===============================================================================

sub delete {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );

    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete";
        } else {
            $runid = $vserver->{vserver} . "_delete";
        }

        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_delete(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Audit delete failed");
        return FALSE;
    }
} # End of delete

#===============================================================================
#  Function: modify
#   This method modify Audit settings on a vserver or list of vservers depending
#   on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to modify audit on a vserver
#      count - The number of vservers to modify audit on
#      filter - variable to filter through vservers to modify audit on
#      params - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run modify
#
#      Audit Attributes:
#        'vserver' - Vserver name (required)
#        'command' - Ontap option for audit to modify (required)
#        'value' - value of the option to modify (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $Audit->modify(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              command  => 'rotate-limit',
#      >              'value'  => '4',
#      >          }
#      >      ]
#      >  );
#===============================================================================

sub modify {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );

    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_modify";
        } else {
            $runid = $vserver->{vserver} . "_modify";
        }

        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_modify(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "modify",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Audit modify failed");
        return FALSE;
    }
} # End of modify

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Create audit on a vserver
#===============================================================================
sub _create {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $destination = $params->get('destination');
    my $rotate_limit = $params->get('rotate-limit');
    my $status = FALSE;

    try {
        NACL::C::VserverAudit->create(
            command_interface => $node,
            'vserver' => $vserver,
            'destination' => $destination,
            'rotate-limit' => $rotate_limit,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already exists/i) {
            $Log->comment('Audit is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to turn create audit on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Audit created');
    } else {
        return NATE::Result::Fail->new('Audit create failed');
    }
} # End of _create

#===============================================================================
# Description: Enable audit on a vserver
#===============================================================================
sub _enable {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $force = $params->get('force');
    my $status = FALSE;

    try {
        $node->apiset_must(interface => 'CLI');
        NACL::C::VserverAudit->enable(
            command_interface => $node,
            'vserver' => $vserver,
            'force' => $force,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already enabled/i) {
            $Log->comment('Audit is already enabled on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to enable audit on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Audit enabled');
    } else {
        return NATE::Result::Fail->new('Audit enable failed');
    }
} # End of _enable

#===============================================================================
# Description: Disable audit on a vserver
#===============================================================================
sub _disable {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $status = FALSE;

    try {
        NACL::STask::VserverAudit->disable(
            command_interface => $node,
            'vserver' => $vserver,
            nacltask_verify   => 1,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already disabled/i) {
            $Log->comment('Audit is already disabled on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to disable audit on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Audit disabled');
    } else {
        return NATE::Result::Fail->new('Audit disable failed');
    }
} # End of _disable

#===============================================================================
# Description: Delete audit on a vserver
#===============================================================================
sub _delete {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $status = FALSE;

    try {
        $node->apiset_must(interface => 'CLI');
        NACL::STask::VserverAudit->purge(
            command_interface => $node,
            'vserver' => $vserver,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('Audit is already deleted on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to delete audit on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::Audit',
            routing_key => '#',
            message => {
                msg => 'Audit.delete.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                }
            }
        );

        return NATE::Result::Pass->new('Audit deleted');
    } else {
        return NATE::Result::Fail->new('Audit delete failed');
    }
} # End of _delete

#===============================================================================
# Description: Delete audit on a vserver
#===============================================================================
sub _modify {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $command = $params->get('command');
    my $value   = $params->get('value');
    my $status = FALSE;

    try {
        $node->apiset_must(interface => 'CLI');
        NACL::C::VserverAudit->modify(
            command_interface => $node,
            'vserver' => $vserver,
            $command => $value,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ / doesn't exist/i) {
            $Log->comment('Audit is already modified on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to modify audit on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Audit modify');
    } else {
        return NATE::Result::Fail->new('Audit modify failed');
    }
} # End of _modify

#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/Audit>

=cut

1;
