################################################################################
# Copyright (c) 2001-2013 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
#
#   Author:
#       dl-ptefit-cifs@netapp.com
#
#   Description:
#       This is the perl module for autolocation.
#
#   Parameters:
#       TBD
#
# @keywords PTEFIT-CIFS
################################################################################
package PTEFIT::Cifs::Misfit::Features::Autolocation;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Vserver;
use NACL::C::VserverCifsOptions;
use NATE::ParamSet;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;
# use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Description: Autolocation constructor
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a new AutoLocation Misfit object
#
# Returns:
#   A new Autolocation Misfit object
#
# Example:
#   my $auto = PTEFIT::Cifs::Misfit::Features::Autolocation->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id}   = "PTEFIT::Cifs::Misfit::Features::Autolocation";
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }
    $self->{node} = $node;
    $self->{node}->apiset_must(interface => 'ZAPI');
    $self->{node}->apiset();
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#===============================================================================
#  Function: enable
#   This method enables autolocation (is-referral-enabled)
#
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      Autolocation Attributes:
#        vserver             - Vserver name
#        is_referral_enabled - Set to true to enable autolocation.
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $auto->enable(
#        objects => [
#                { vserver     => 'vs_base3',
#		   is_referral_enabled => 'true',
#                },
#            ],
#        id => 'enable'
#    );
#
#
#===============================================================================
sub enable {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background1",
        'required' => FALSE,
        'default'  => FALSE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_enable";
        } else {
            $runid = $vserver->{vserver} . "_enable";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => \&_enable,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "enable",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Enable autolocation failed");
        return FALSE;
    }
} # End of enable

#===============================================================================
#  Function: disable
#   This method disables autolocation (is-referral-enabled set to false)
#
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      Autolocation Attributes:
#        vserver             - Vserver name
#        is_referral_enabled - Set to False to disable autolocation.
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $auto->disable(
#        objects => [
#                { vserver     => 'vs_base3',
#                  is_referral_enabled => 'false';
#                },
#            ],
#        id => 'disable'
#    );
#
#
#===============================================================================
sub disable {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background1",
        'required' => FALSE,
        'default'  => FALSE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_disable";
        } else {
            $runid = $vserver->{vserver} . "_disable";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => \&_disable,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "disable",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Disable autolocation failed");
        return FALSE;
    }
} # End of disable


#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Enable autolocation
#===============================================================================
sub _enable {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $is_referral_enabled = &ParamsMgr::get_arg(
        'arg'      => "is_referral_enabled",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverCifsOptions->modify(
            command_interface     => $node,
            'vserver'             => $vserver,
            'is-referral-enabled' => $is_referral_enabled,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
            $Log->error('Failed to enable autolocation');
            NATE::BaseException->throw($exception);
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Autolocation enable failed");
        return FALSE;
    }

} # End of _enable

#===============================================================================
# Description: Disable  autolocation
#===============================================================================
sub _disable {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $is_referral_enabled = &ParamsMgr::get_arg(
        'arg'      => "is_referral_enabled",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverCifsOptions->modify(
            command_interface => $node,
            'vserver' => $vserver,
	    'is-referral-enabled' => $is_referral_enabled,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->error('Failed to disable autolocation');

        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Autolocation disable failed");
        return FALSE;
    }

} # End of _disable


#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/Audit>

=cut

1;
