# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Vserver BranchCache module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: PTEFIT::Cifs::Misfit::Features::BranchCache
# Wrapper for NACL BranchCache with Misfit messaging

package PTEFIT::Cifs::Misfit::Features::BranchCache;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Vserver;
use NATE::ParamSet;
use NACL::C::VserverCifsBranchcache;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;
# use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a new BranchCache Misfit object
#
#  Returns:
#      A new BranchCache Misfit object
#
#  Example:
#
#      > my $BranchCache = PTEFIT::Cifs::Misfit::Features::BranchCache->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id}   = "PTEFIT::Cifs::Misfit::Features::BranchCache";
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }
    $self->{node} = $node;
    $self->{node}->apiset_must(interface => 'ZAPI');
    $self->{node}->apiset();
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->comment("Failed to create a message broker for this module");
        return {};
    }
    return $self;
} # End of new

#===============================================================================
#  Function: create
#   This method creates BranchCache on a vserver or list of vservers depending on the
#   input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to setup branchcache on a vserver
#      count - The number of vservers to create branchcache on
#      filter - variable to filter through vservers to create branchcache on
#      params - hash containing parameters for the branchcache config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      BranchCache Attributes:
#        'vserver' - Vserver name (required)
#        'hash_store_path' - path to where the hash is stored (required)
#        'hash_store_max_size' - Max size of the hash (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $BranchCache->create(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'hash_store_path'  => '/path',
#      >              'hash_store_max_size' => '1GB',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub create {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create";
        } else {
            $runid = $vserver->{vserver} . "_create";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_create,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("BranchCache create failed");
        return FALSE;
    }
} # End of create

#===============================================================================
#  Function: delete
#   This method deletes BranchCache on a vserver or list of vservers depending on the
#   input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to setup branchcache on a vserver
#      count - The number of vservers to delete branchcache on
#      filter - variable to filter through vservers to delete branchcache on
#      params - hash containing parameters for the branchcache config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run delete
#
#      BranchCache Attributes:
#        'vserver' - Vserver name (required)
#        'flush_hashes' - true or false depending if you want to clear
#                         the hash on the vserver (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $BranchCache->delete(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              ''flush_hashes' => 'true',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub delete {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete";
        } else {
            $runid = $vserver->{vserver} . "_delete";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_delete,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("BranchCache delete failed");
        return FALSE;
    }
} # End of delete

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Create branchcache on a vserver
#===============================================================================
sub _create {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $hash_store_path = &ParamsMgr::get_arg(
        'arg'      => "hash_store_path",
        'required' => TRUE,
    );
    my $hash_store_max_size = &ParamsMgr::get_arg(
        'arg'      => "hash_store_max_size",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverCifsBranchcache->create(
            command_interface => $node,
            'vserver' => $vserver,
            'hash-store-path' => $hash_store_path,
            'hash-store-max-size' => $hash_store_max_size,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate/i) {
            $Log->comment('BranchCache is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to turn create branchcache on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('BranchCache created');
    } else {
        return NATE::Result::Fail->new('BranchCache create failed');
    }
} # End of _create

#===============================================================================
# Description: Delete branchcache on a vserver
#===============================================================================
sub _delete {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $flush_hashes = &ParamsMgr::get_arg(
        'arg'      => "flush_hashes",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverCifsBranchcache->delete(
            command_interface => $node,
            'vserver' => $vserver,
            'flush-hashes' => $flush_hashes,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('BranchCache is already deleted on this vserver.');
            $status = TRUE;
        } else {
            $$Log->comment('Failed to turn delete branchcache on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('BranchCache deleted');
    } else {
        return NATE::Result::Fail->new('BranchCache delete failed');
    }
} # End of _delete

#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/BranchCache>

=cut

1;

