################################################################################
# Copyright (c) 2001-2013 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
#
#   Author:
#       dl-ptefit-cifs@netapp.com
#
#   Description:
#       This is the perl module for fpolicy.
#
#   Parameters:
#       TBD
#
# @keywords PTEFIT-CIFS
################################################################################

# Package: PTEFIT::Cifs::Misfit::Features::FPolicy
# Wrapper for NACL FPolicy with Misfit messaging

package PTEFIT::Cifs::Misfit::Features::FPolicy;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Vserver;
use NATE::ParamSet;
use NACL::C::VserverFpolicyPolicyEvent;
use NACL::STask::VserverFpolicy;
use NACL::STask::VserverFpolicyPolicy;
use NACL::C::VserverFpolicyPolicyExternalEngine;
use NACL::C::VserverFpolicyPolicyScope;
use NACL::STask::VserverFpolicyPolicyEvent;
use NACL::STask::VserverFpolicyPolicyExternalEngine;
use NACL::STask::VserverFpolicyPolicyScope;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;
use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a new FPolicy Misfit object
#
#  Returns:
#      A new FPolicy Misfit object
#
#  Example:
#
#      > my $FPolicy = PTEFIT::Cifs::Misfit::Features::FPolicy->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id}   = "PTEFIT::Cifs::Misfit::Features::FPolicy";
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }
    $self->{node} = $node;
    $self->{node}->apiset_must(interface => 'ZAPI');
    $self->{node}->apiset();
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->comment("Failed to create a message broker for this module");
        return {};
    }

    return $self;
} # End of new

#===============================================================================
#  Function: create_fpolicy_event
#   This method creates an FPolicy event on a vserver or list of vservers depending
#    on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to setup an fpolicy event
#      on a vserver
#      count - The number of vservers to create fpolicy on
#      filter - variable to filter through vservers to create fpolicy on
#      params - hash containing parameters for the fpolicy config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'event-name' - name of the fpolicy event you want to create (required)
#        'volume-operation' - whether you want this on the volume level (required)
#        'protocol' - the protocol for the fpolicy event (required)
#        'file-operations' - The file operations for fpolicy (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->create(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'event-name'  => 'event1',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub create_fpolicy_event {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_fpolicy_event";
        } else {
            $runid = $vserver->{vserver} . "_create_fpolicy_event";
        }
        push(@msg_data, {vserver => $vserver->{vserver},
            event_name => $vserver->{'event-name'},
            volume_operation => $vserver->{'volume-operation'},
            protocols => $vserver->{protocols},
            file_operations => $vserver->{'file-operations'}
        });
        push (@vserver_info, {
            codespec => \&_create_fpolicy_event,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_fpolicy_event",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy create failed");
        return FALSE;
    }
} # End of create

#===============================================================================
#  Function: delete_fpolicy_event
#   This method deletes an FPolicy event on a vserver or list of vservers depending
#    on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to delete an fpolicy event
#      on a vserver
#      count - The number of vservers to delete fpolicy event on
#      filter - variable to filter through vservers to delete fpolicy event on
#      params - hash containing parameters for the fpolicy config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run delete
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'event-name' - name of the fpolicy event you want to delete (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->delete_fpolicy_event(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'event-name'  => 'event1',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub delete_fpolicy_event {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete_fpolicy_event";
        } else {
            $runid = $vserver->{vserver} . "_delete_fpolicy_event";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_delete_fpolicy_event,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete_fpolicy_event",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy delete failed");
        return FALSE;
    }
} # End of delete

#===============================================================================
#  Function: create_fpolicy_policy
#   This method creates an FPolicy policy on a vserver or list of vservers depending
#    on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to setup an fpolicy event
#      on a vserver
#      count - The number of vservers to create fpolicy policy on
#      filter - variable to filter through vservers to create fpolicy policy on
#      params - hash containing parameters for the fpolicy policy config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'policy-name' - name of the policy event you want to create (required)
#        'engine' - the engine you want to use for the policy (required)
#        'events' - the events to add to the fpolicy policy (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->create_fpolicy_policy(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'policy-name'  => 'policy_name',
#      >              'engine'  => 'engine_name',
#      >              'events' => ['event1'],
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub create_fpolicy_policy {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_fpolicy_policy";
        } else {
            $runid = $vserver->{vserver} . "_create_fpolicy_policy";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_create_fpolicy_policy,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_fpolicy_policy",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy create policy failed");
        return FALSE;
    }
} # End of create_fpolicy_policy

#===============================================================================
#  Function: create_fpolicy_policy_scope
#   This method creates an FPolicy policy scope on a vserver or list of vservers
#   depending on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to setup an fpolicy scope
#      on a vserver
#      count - The number of vservers to create the scope on on
#      filter - variable to filter through vservers to create policy scope on
#      params - hash containing parameters for the policy scope config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create scope
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'policy-name' - name of the policy event you want to create (required)
#        'shares-to-include' - the shares to include in the fpolicy scope (required)
#        'volumes-to-include' - the volumes to include in the fpolicy scope (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->create_fpolicy_policy_scope(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'policy-name'  => 'policy_name',
#      >              'shares-to-include'  => ['share1'],
#      >              'volumes-to-include' => ['vol1'],
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub create_fpolicy_policy_scope {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_fpolicy_policy_scope";
        } else {
            $runid = $vserver->{vserver} . "_create_fpolicy_policy_scope";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_create_fpolicy_policy_scope,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_fpolicy_policy_scope",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy create policy scope failed");
        return FALSE;
    }
} # End of create_fpolicy_policy_scope

#===============================================================================
#  Function: delete_fpolicy_policy_scope
#   This method deletes an FPolicy policy scope on a vserver or list of vservers
#   depending on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to delete an fpolicy scope
#      on a vserver
#      count - The number of vservers to delete the scope on
#      filter - variable to filter through vservers to delete policy scope on
#      params - hash containing parameters for the policy scope config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run delete scope
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'policy-name' - name of the policy scope you want to delete (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->delete_fpolicy_policy_scope(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'policy-name'  => 'policy_name',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub delete_fpolicy_policy_scope {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete_fpolicy_policy_scope";
        } else {
            $runid = $vserver->{vserver} . "_delete_fpolicy_policy_scope";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_delete_fpolicy_policy_scope,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete_fpolicy_policy_scope",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy delete policy scope failed");
        return FALSE;
    }
} # End of delete_fpolicy_policy_scope

#===============================================================================
#  Function: enable_fpolicy
#   This method enable a FPolicy on a vserver or list of vservers
#   depending on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to enable a fpolicy
#      on a vserver
#      count - The number of vservers to enable fpolicy on
#      filter - variable to filter through vservers to enable fpolicy on
#      params - hash containing parameters for the fpolicy enable
#      background - Option to run in serially or in parallel
#      iterations - number of times to run fpolicy enable
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'policy-name' - name of the policy you want to enable (required)
#        'sequence-number' - the sequence the fpolicy should be in (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->enable_fpolicy(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'policy-name'  => 'policy_name',
#      >              'sequence-number'  => '1'
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub enable_fpolicy {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_enable_fpolicy";
        } else {
            $runid = $vserver->{vserver} . "_enable_fpolicy";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_enable_fpolicy,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "enable_fpolicy",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy enable failed");
        return FALSE;
    }
} # End of enable_fpolicy

#===============================================================================
#  Function: fpolicy_engine_connect
#   This method connects a FPolicy engine to a vserver or list of vservers
#   depending on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to connect an engine
#      to a vserver
#      count - The number of vservers to connect to the engine
#      filter - variable to filter through vservers to connect to the engine
#      params - hash containing parameters for the engine config
#      background - Option to run in serially or in parallel
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'policy-name' - name of the policy to add the engine to (required)
#        'server' - name or ip of the engine (required)
#        'node-name' - name of the node to connect to the engine (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->fpolicy_engine_connect(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'policy-name'  => 'policy_name',
#      >              'server' => '0.0.0.0',
#      >              'node-name' => 'node1',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub fpolicy_engine_connect {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_fpolicy_engine_connect";
        } else {
            $runid = $vserver->{vserver} . "_fpolicy_engine_connect";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_fpolicy_engine_connect,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "fpolicy_engine_connect",
        background => $background,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy engine connect failed");
        return FALSE;
    }
} # End of fpolicy_engine_connect

#===============================================================================
#  Function: fpolicy_engine_disconnect
#   This method disconnects a FPolicy engine to a vserver or list of vservers
#   depending on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to disconnect an engine
#      to a vserver
#      count - The number of vservers to disconnect to the engine
#      filter - variable to filter through vservers to disconnect to the engine
#      params - hash containing parameters for the engine disconnect
#      background - Option to run in serially or in parallel
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'policy-name' - name of the policy to disconnect the engine (required)
#        'server' - name or ip of the engine (required)
#        'node-name' - name of the node to disconnect to the engine (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->fpolicy_engine_disconnect(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'policy-name'  => 'policy_name',
#      >              'server' => '0.0.0.0',
#      >              'node-name' => 'node1',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub fpolicy_engine_disconnect {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_fpolicy_engine_disconnect";
        } else {
            $runid = $vserver->{vserver} . "_fpolicy_engine_disconnect";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_fpolicy_engine_disconnect,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "fpolicy_engine_disconnect",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy engine disconnect failed");
        return FALSE;
    }
} # End of fpolicy_engine_disconnect

#===============================================================================
#  Function: fpolicy_create_external_engine
#   This method connects a FPolicy engine to a vserver or list of vservers
#   depending on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information of external engine
#      count - The number of vservers to connect to the engine
#      filter - variable to filter through vservers to create an ext engine on
#      params - hash containing parameters for the engine config
#      background - Option to run in serially or in parallel
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'engine-name' - name of the engine (required)
#        'port' - port to connect to on server (required)
#        'primary-servers' - ip of the server (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->fpolicy_create_external_engine(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'engine-name'  => 'test_engine',
#      >              'port'  => '8222',
#      >              'primary-servers' => ['0.0.0.0'],
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub fpolicy_create_external_engine {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_fpolicy_create_external_engine";
        } else {
            $runid = $vserver->{vserver} . "_fpolicy_create_external_engine";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_fpolicy_create_external_engine,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "fpolicy_create_external_engine",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy create external engine failed");
        return FALSE;
    }
} # End of fpolicy_create_external_engine

#===============================================================================
#  Function: fpolicy_delete_external_engine
#   This method connects a FPolicy engine to a vserver or list of vservers
#   depending on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information of external engine
#      count - The number of vservers to connect to the engine
#      filter - variable to filter through vservers to delete an ext engine on
#      params - hash containing parameters for the engine config
#      background - Option to run in serially or in parallel
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'engine-name' - name of the engine (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->fpolicy_delete_external_engine(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'engine-name'  => 'test_engine',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub fpolicy_delete_external_engine {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_fpolicy_delete_external_engine";
        } else {
            $runid = $vserver->{vserver} . "_fpolicy_delete_external_engine";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_fpolicy_delete_external_engine,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "fpolicy_delete_external_engine",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy delete external engine failed");
        return FALSE;
    }
} # End of fpolicy_delete_external_engine

#===============================================================================
#  Function: disable_fpolicy
#   This method disable a FPolicy on a vserver or list of vservers
#   depending on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to disable a fpolicy
#      on a vserver
#      count - The number of vservers to disable fpolicy on
#      filter - variable to filter through vservers to disable fpolicy on
#      params - hash containing parameters for the fpolicy config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run fpolicy disable
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'policy-name' - name of the policy you want to disable (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->create_fpolicy_policy_scope(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'policy-name'  => 'policy_name',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub disable_fpolicy {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_disable_fpolicy";
        } else {
            $runid = $vserver->{vserver} . "_disable_fpolicy";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_disable_fpolicy,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "disable_fpolicy",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy disable failed");
        return FALSE;
    }
} # End of disable_fpolicy

#===============================================================================
#  Function: delete_fpolicy_policy
#   This method deletes an FPolicy policy on a vserver or list of vservers depending
#    on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to setup an fpolicy event
#      on a vserver
#      count - The number of vservers to delete fpolicy policy on
#      filter - variable to filter through vservers to delete fpolicy policy on
#      params - hash containing parameters for the fpolicy policy config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run delete
#
#      FPolicy Attributes:
#        'vserver' - Vserver name (required)
#        'policy-name' - name of the policy event you want to delete (required)
#        'engine' - the engine you want to use for the policy (required)
#        'events' - the events to add to the fpolicy policy (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $FPolicy->delete_fpolicy_policy(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'policy-name'  => 'policy_name',
#      >              'engine'  => 'engine_name',
#      >              'events' => ['event1'],
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub delete_fpolicy_policy {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete_fpolicy_policy";
        } else {
            $runid = $vserver->{vserver} . "_delete_fpolicy_policy";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_delete_fpolicy_policy,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete_fpolicy_policy",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FPolicy delete policy failed");
        return FALSE;
    }
} # End of delete_fpolicy_policy

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Create FPolicy on a vserver
#===============================================================================
sub _create_fpolicy_event {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $event_name = &ParamsMgr::get_arg(
        'arg'      => "event-name",
        'required' => TRUE,
    );
    my $volume_operation = &ParamsMgr::get_arg(
        'arg'      => "volume-operation",
        'required' => TRUE,
    );
    my $protocol = &ParamsMgr::get_arg(
        'arg'      => "protocol",
        'required' => TRUE,
    );
    my $file_operations = &ParamsMgr::get_arg(
        'arg'      => "file-operations",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverFpolicyPolicyEvent->create(
            command_interface => $node,
            'vserver' => $vserver,
            'event-name' => $event_name,
            'volume-operation' => $volume_operation,
            'protocol' => @{$protocol},
            'file-operations' => @{$file_operations},
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already exists/i) {
            $Log->comment('FPolicy is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to create FPolicy on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy event created');
    } else {
        return NATE::Result::Fail->new('FPolicy event create failed');
    }
} # End of _create_event

#===============================================================================
# Description: Create FPolicy on a vserver
#===============================================================================
sub _create_fpolicy_policy {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy-name",
        'required' => TRUE,
    );
    my $engine = &ParamsMgr::get_arg(
        'arg'      => "engine",
        'required' => TRUE,
    );
    my $events = &ParamsMgr::get_arg(
        'arg'      => "events",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
        NACL::C::VserverFpolicyPolicy->create(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
            'engine' => $engine,
            'events' => @{$events},
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already exists/i) {
            $Log->comment('FPolicy is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to create FPolicy on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy policy created');
    } else {
        return NATE::Result::Fail->new('FPolicy policy create failed');
    }
} # End of _create_policy

#===============================================================================
# Description: Create FPolicy policy scope on a vserver
#===============================================================================
sub _create_fpolicy_policy_scope {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy-name",
        'required' => TRUE,
    );
    my $shares_to_include = &ParamsMgr::get_arg(
        'arg'      => "shares-to-include",
        'required' => TRUE,
    );
    my $volumes_to_include = &ParamsMgr::get_arg(
        'arg'      => "volumes-to-include",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
        NACL::C::VserverFpolicyPolicyScope->create(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
            'shares-to-include' => @{$shares_to_include},
            'volumes-to-include' => @{$volumes_to_include},
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already exists/i) {
            $Log->comment('FPolicy policy scope is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to create FPolicy policy scope on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy policy scope created');
    } else {
        return NATE::Result::Fail->new('FPolicy policy create scope failed');
    }
} # End of _create_fpolicy_policy_scope

#===============================================================================
# Description: Delete FPolicy policy scope on a vserver
#===============================================================================
sub _delete_fpolicy_policy_scope {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy-name",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
        NACL::C::VserverFpolicyPolicyScope->delete(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('FPolicy policy scope is already deleted on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to delete FPolicy policy scope on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy policy scope deleted');
    } else {
        return NATE::Result::Fail->new('FPolicy policy delete scope failed');
    }
} # End of _delete_fpolicy_policy_scope

#===============================================================================
# Description: Delete FPolicy on a vserver
#===============================================================================
sub _delete_fpolicy_policy {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy-name",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
        NACL::C::VserverFpolicyPolicy->delete(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('FPolicy is already deleted on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to delete FPolicy on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy policy deleted');
    } else {
        return NATE::Result::Fail->new('FPolicy policy delete failed');
    }
} # End of _delete_policy

#===============================================================================
# Description: Delete FPolicy on a vserver
#===============================================================================
sub _delete_fpolicy_event {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $event_name = &ParamsMgr::get_arg(
        'arg'      => "event-name",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
        NACL::C::VserverFpolicyPolicyEvent->delete(
            command_interface => $node,
            'vserver' => $vserver,
            'event-name' => $event_name,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('FPolicy is already deleted on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to delete FPolicy on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy event deleted');
    } else {
        return NATE::Result::Fail->new('FPolicy event delete failed');
    }
} # End of _delete

#===============================================================================
# Description: Enable FPolicy on a vserver
#===============================================================================
sub _enable_fpolicy {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy-name",
        'required' => TRUE,
    );
    my $sequence_number = &ParamsMgr::get_arg(
        'arg'      => "sequence-number",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
        NACL::C::VserverFpolicy->enable(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
            'sequence-number' => $sequence_number
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already enabled/i) {
            $Log->comment('FPolicy is already enabled on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to enable FPolicy on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy policy enabled');
    } else {
        return NATE::Result::Fail->new('FPolicy enable failed');
    }
} # End of _enable_policy

#===============================================================================
# Description: Create FPolicy on a vserver
#===============================================================================
sub _disable_fpolicy {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy-name",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
        NACL::C::VserverFpolicy->disable(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already disabled/i) {
            $Log->comment('FPolicy is already disabled on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to disable FPolicy on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy policy disabled');
    } else {
        return NATE::Result::Fail->new('FPolicy disable failed');
    }
} # End of _disable_policy

#===============================================================================
# Description: Create FPolicy on a vserver
#===============================================================================
sub _fpolicy_engine_connect {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy-name",
        'required' => TRUE,
    );
    my $server = &ParamsMgr::get_arg(
        'arg'      => "server",
        'required' => TRUE,
    );
    my $node_name = &ParamsMgr::get_arg(
        'arg'      => "node-name",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
	   $node->apiset_must(interface => 'CLI');
        NACL::C::VserverFpolicy->engine_connect(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
            'server' => $server,
            'node' => $node_name,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already exists/i) {
            $Log->comment('FPolicy engine is already connected on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to connect to FPolicy engine on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy engine connected');
    } else {
        return NATE::Result::Fail->new('FPolicy engine connect failed');
    }
} # End of _fpolicy_engine_connect

#===============================================================================
# Description: Disconnect FPolicy engine on a vserver
#===============================================================================
sub _fpolicy_engine_disconnect {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy-name",
        'required' => TRUE,
    );
    my $server = &ParamsMgr::get_arg(
        'arg'      => "server",
        'required' => TRUE,
    );
    my $node_name = &ParamsMgr::get_arg(
        'arg'      => "node-name",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
	   $node->apiset_must(interface => 'CLI');
       NACL::C::VserverFpolicy->engine_disconnect(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
            'server' => $server,
            'node' => $node_name,
       );
       $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exists/i) {
            $Log->comment('FPolicy engine is already disconnected on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to disconnect to FPolicy engine on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy engine disconnected');
    } else {
        return NATE::Result::Fail->new('FPolicy engine disconnect failed');
    }
} # End of _fpolicy_engine_disconnect

#===============================================================================
# Description: Create FPolicy on a vserver
#===============================================================================
sub _fpolicy_create_external_engine {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $engine_name = &ParamsMgr::get_arg(
        'arg'      => "engine-name",
        'required' => TRUE,
    );
    my $primary_servers = &ParamsMgr::get_arg(
        'arg'      => "primary-servers",
        'required' => TRUE,
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE,
    );
    my $ssl_option = &ParamsMgr::get_arg(
        'arg'      => "ssl-option",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
        NACL::C::VserverFpolicyPolicyExternalEngine->create(
            command_interface => $node,
            'vserver' => $vserver,
            'engine-name' => $engine_name,
            'primary-servers' => @{$primary_servers},
            'port' => $port,
            'ssl-option' => $ssl_option,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already exists/i) {
            $Log->comment('FPolicy external engine is already on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to create external FPolicy engine on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy external engine created');
    } else {
        return NATE::Result::Fail->new('FPolicy external engine create failed');
    }
} # End of _fpolicy_external_engine

#===============================================================================
# Description: Delete FPolicy on a vserver
#===============================================================================
sub _fpolicy_delete_external_engine {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $engine_name = &ParamsMgr::get_arg(
        'arg'      => "engine-name",
        'required' => TRUE,
    );
    my $node = $self->{node};
    my $status = FALSE;

    try {
        NACL::C::VserverFpolicyPolicyExternalEngine->delete(
            command_interface => $node,
            'vserver' => $vserver,
            'engine-name' => $engine_name,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('FPolicy external engine doesnt exist on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to delete external FPolicy engine on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('FPolicy external engine created');
    } else {
        return NATE::Result::Fail->new('FPolicy external engine delete failed');
    }
} # End of _fpolicy_external_engine_delete

#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/FPolicy>

=cut

1;
