
################################################################################
# Copyright (c) 2001-2013 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
#
#   Author:
#       dl-ptefit-cifs@netapp.com
#
#   Description:
#       This is the thpl that will enable and disable of flexcache volumes
#       
#   Parameters:
#       volume - Name of volume to have action performed
#       vserver - Name of vserver that volume resides on
#
# @keywords PTEFIT-CIFS
################################################################################


package PTEFIT::Cifs::Misfit::Features::FlexCache;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::Volume;
use NACL::C::Cluster;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;
use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

use constant {
    FLEXCACHE_WAIT => 180,
    TIMEOUT        => 360,
};

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a Misfit flexcache object
#
#   Parameters:
#       cluster - name of cluster to perform operation on.
#
#   Returns:
#     New Misfit FlexCache object
#
#   Example:
#     > my $FlexCache = PTEFIT::Cifs::Misfit::Features::FlexCache->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id}   = "PTEFIT::Cifs::Misfit::Features::FlexCache";
    my $interface = NACL::C::Node->find();
    $interface->apiset_must(interface => 'ZAPI');

    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );

    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }

    $node->apiset_must(interface => 'ZAPI');
    $node->apiset();
    $self->{node} = $node;
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#==============================================================================
# Function: create
#   Turns on flexcache. Create will loop through the list of objects provided
#   and enable flexcache with the specified parameters. Provides a wrapper
#   to NACL::C::VolumeFlexcache->create function
#
#   Parameters:
#     objects - Array of hashes containing information for volumes
#     background - Option to specify whether to perform each 'create'
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to turn flexcache on
#     filter - If objects is not passed and a count is specified, use this filter
#         to filter volumes based on attributes
#
#     Flexcache attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $FlexCache->create(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub create {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = {$volume->{id}};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_create(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("FlexCache create failed");
        return FALSE;
    }
} # End of create

#==============================================================================
# Function: delete
#   Turns off flexcache. Delete will loop through the list of objects provided
#   and turn off flexcache with the specified parameters. Provides a wrapper
#   to NACL::C::VolumeFlexcache->off function
#
#   Parameters:
#     objects - Array of hashes containing information for volumes
#     background - Option to specify whether to perform each 'delete'
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to turn flexcache off
#     filter - If objects is not passed and a count is specified, use this filter
#         to filter volumes based on attributes
#
#     Flexcache attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $FlexCache->delete(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub delete {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = {$volume->{id}};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_delete(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "delete",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Flexcache delete failed");
        return FALSE;
    }
} # End of delete


#===============================================================================
#  PRIVATE METHODS
#===============================================================================


#===============================================================================
# Description: Create flexcache volume
#===============================================================================
sub _create {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::FlexCache',
            routing_key => '#',
            message => {
                msg => 'FlexCache.create.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        NACL::C::VolumeFlexcache->create(
            command_interface => $node,
            'origin-volume'   => $volume,
            vserver           => $vserver,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() !~ /Operation has already been enabled/i) {
            $Log->error('Failed to turn on FlexCache');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'FlexCache.create.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume
                    }
                }
            );
            NATE::BaseException->throw($exception);
        } else {
            $status = TRUE;
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
        exchange    => 'PTEFIT::Cifs::Misfit::Features::FlexCache',
        routing_key => '#',
        message => {
            msg => 'FlexCache.create.finished',
            data => {
                node    => $node->{node},
                vserver => $vserver,
                volume  => $volume
                }
            }
        );
        return NATE::Result::Pass->new('FlexCache created');
    } else {
        return NATE::Result::Fail->new('FlexCache create failed');
    }
} # End of _create

#===============================================================================
# Description: Delete: removes flexcache option from volume
#===============================================================================
sub _delete {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::FlexCache',
            routing_key => '#',
            message => {
                msg => 'FlexCache.delete.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        NACL::C::VolumeFlexcache->delete(
            command_interface => $node,
            'origin-volume'   => $volume,
            vserver           => $vserver,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() !~ /Flexcache has already been deleted/i) {
            $Log->error('Failed to turn off flexcache');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'FlexCache.delete.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume
                    }
                }
            );
            NATE::BaseException->throw($exception);
        } else {
            $status = TRUE;
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
        exchange    => 'PTEFIT::Cifs::Misfit::Features::FlexCache',
        routing_key => '#',
        message => {
            msg => 'FlexCache.delete.finished',
            data => {
                node    => $node->{node},
                vserver => $vserver,
                volume  => $volume
                }
            }
        );
        return NATE::Result::Pass->new('FlexCache deleted');
    } else {
        return NATE::Result::Fail->new('FlexCache delete failed');
    }
} # End of delete

#===============================================================================
# MESSAGE CATALOG
#===============================================================================

#===============================================================================
# Message: flexcache.create.started
#   Exchange:
#       PTEFIT::Cifs::Misfit::Features::FlexCache
#
#   Message Structure:
# >     {
# >         "msg": "flexcache.create.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: flexcache.create.finished
#   Exchange:
#       PTEFIT::Cifs::Misfit::Features::FlexCache
#
#   Message Structure:
# >     {
# >         "msg": "flexcache.create.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: flexcache.create.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "flexcache.create.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: flexcache.delete.started
#   Exchange:
#       PTEFIT::Cifs::Misfit::Features::FlexCache
#
#   Message Structure:
# >     {
# >         "msg": "flexcache.delete.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: flexcache.delete.finished
#   Exchange:
#       PTEFIT::Cifs::Misfit::Features::FlexCache
#
#   Message Structure:
# >     {
# >         "msg": "flexcache.delete.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: flexcache.delete.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "flexcache.delete.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
#
#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/FlexCache>

=cut

1;
