# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Vserver LUG module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: PTEFIT::Cifs::Misfit::Features::LUG
# Wrapper for NACL LocalUsers/Groups with Misfit messaging

package PTEFIT::Cifs::Misfit::Features::LUG;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::Vserver;
use San::FIT::Misfit::Core::Workflow;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::MessageBroker;
use NACL::C::VserverCifsUsersAndGroupsLocalUser;
use NACL::C::VserverCifsUsersAndGroupsLocalGroup;
use NACL::C::VserverCifsUsersAndGroupsPrivilege;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a new LUG Misfit object
#
#  Returns:
#      A new LUG Misfit object
#
#  Example:
#
#      > my $LUG = PTEFIT::Cifs::Misfit::Features::LUG->new();
#===============================================================================

sub new {
    my ($class, %args) = @_;

    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "PTEFIT::Cifs::Misfit::Features::LUG";

    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }
    $self->{node} = $node;
    $self->{node}->apiset_must(interface => 'ZAPI');
    $self->{node}->apiset();
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->comment("Failed to create a message broker for this module");
        return {};
    }

    return $self;
} # End of new

#===============================================================================
#  Function: create_local_user
#   This method creates a local user on a vserver or list of vservers depending
#    on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to create local user
#      on a vserver
#      count - The number of vservers to create the local user on
#      filter - variable to filter through vservers to create local user on
#      params - hash containing parameters for the local user config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      LUG Attributes:
#        'vserver' - Vserver name (required)
#        'user-name' - username you want to create (required)
#        'pasword' - the password for the username (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $LUG->create_local_user(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'user-name' => 'user1',
#      >              'password' => 'password1',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub create_local_user {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_local_user";
        } else {
            $runid = $vserver->{vserver} . "_create_local_user";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_create_local_user(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_local_user",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Local User create failed");
        return FALSE;
    }
} # End of create_local_user

#===============================================================================
#  Function: create_local_group
#   This method creates a local group on a vserver or list of vservers depending
#    on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to create local group
#      on a vserver
#      count - The number of vservers to create the local group on
#      filter - variable to filter through vservers to create local group on
#      params - hash containing parameters for the local group config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      LUG Attributes:
#        'vserver' - Vserver name (required)
#        'group-name' - groupname you want to create (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $LUG->create_local_group(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'group-name' => 'group1',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub create_local_group {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_local_group";
        } else {
            $runid = $vserver->{vserver} . "_create_local_group";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_create_local_group(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_local_group",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Local group create failed");
        return FALSE;
    }
} # End of create_local_group

#===============================================================================
#  Function: add_user_to_group
#   This method users to a group on a vserver or list of vservers depending
#    on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to add users to a group
#      on a vserver
#      count - The number of vservers to add users to group
#      filter - variable to filter through vservers to add users to group
#      params - hash containing parameters for the group and users
#      background - Option to run in serially or in parallel
#
#      LUG Attributes:
#        'vserver' - Vserver name (required)
#        'group-name' - the name of the group to add users (required)
#        'member-names' - the users to add to the group (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $LUG->add_user_to_group(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'group-name' => 'group1',
#      >              'member-names' => ['user1'],
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub add_user_to_group {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_add_user_to_group";
        } else {
            $runid = $vserver->{vserver} . "_add_user_to_group";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_add_user_to_group(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "add_user_to_group",
        background => $background,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Add user to group failed");
        return FALSE;
    }
} # End of add_user_to_group

#===============================================================================
#  Function: add_privileges
#   This method adds privileges to a user or groupon a vserver or list of
#   vservers depending on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to add privileges
#      on a vserver
#      count - The number of vservers to add privileges to user/group
#      filter - variable to filter through vservers to add privileges to user/group
#      params - hash containing parameters for the user/group config
#      background - Option to run in serially or in parallel
#
#      LUG Attributes:
#        'vserver' - Vserver name (required)
#        'user-or-group-name' - username or group name (required)
#        'privileges' - the privileges for a user or group (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $LUG->add_privileges(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'user-or-group-name' => 'user1',
#      >              privileges => [''],
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub add_privileges {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_add_privileges";
        } else {
            $runid = $vserver->{vserver} . "_add_privileges";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_add_privileges(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "add_privileges",
        background => $background,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Add privileges failed");
        return FALSE;
    }
} # End of add_privileges

#===============================================================================
#  Function: remove_user_to_group
#   This method removes users from a group on a vserver or list of vservers depending
#    on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to remove users from a group
#      on a vserver
#      count - The number of vservers to remove users from group
#      filter - variable to filter through vservers to remove users from group
#      params - hash containing parameters for the group and users
#      background - Option to run in serially or in parallel
#
#      LUG Attributes:
#        'vserver' - Vserver name (required)
#        'group-name' - the name of the group to remove users (required)
#        'member-names' - the users to remove from the group (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $LUG->remove_user_to_group(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'group-name' => 'group1',
#      >              'member-names' => ['user1'],
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub remove_user_from_group {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_remove_user_from_group";
        } else {
            $runid = $vserver->{vserver} . "_remove_user_from_group";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_remove_user_from_group(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "remove_user_from_group",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Local group create failed");
        return FALSE;
    }
} # End of create_local_group

#===============================================================================
#  Function: delete_local_user
#   This method deletes a local user on a vserver or list of vservers depending
#    on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to delete local user
#      on a vserver
#      count - The number of vservers to delete the local user on
#      filter - variable to filter through vservers to delete local user on
#      params - hash containing parameters for the local user config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run delete
#
#      LUG Attributes:
#        'vserver' - Vserver name (required)
#        'user-name' - username you want to delete (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $LUG->delete_local_user(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'user-name' => 'user1',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub delete_local_user {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete_local_user";
        } else {
            $runid = $vserver->{vserver} . "_delete_local_user";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_delete_local_user(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete_local_user",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Local user delete failed");
        return FALSE;
    }
} # End of delete_local_user

#===============================================================================
#  Function: delete_local_group
#   This method deletes a local group on a vserver or list of vservers depending
#    on the input parameters.
#
#   Parameters:
#      objects - Array of hashes containing information to delete local group
#      on a vserver
#      count - The number of vservers to delete the local group on
#      filter - variable to filter through vservers to delete local group on
#      params - hash containing parameters for the local group config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run delete
#
#      LUG Attributes:
#        'vserver' - Vserver name (required)
#        'group-name' - groupname you want to delete (required)
#
#  Returns:
#      TRUE if successful, FALSE if unsuccessful
#
#  Example:
#
#      > my $status = $LUG->delete_local_group(
#      >     objects => [
#      >          {
#      >              vserver   => 'vs_base1',
#      >              'group-name' => 'group1',
#      >          }
#      >      ]
#      >  );
#===============================================================================
sub delete_local_group {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete_local_group";
        } else {
            $runid = $vserver->{vserver} . "_delete_local_group";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_delete_local_group(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete_local_group",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Local group delete failed");
        return FALSE;
    }
} # End of delete_local_group

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Responsible for generating a list of vservers matching a criteria
#===============================================================================

#===============================================================================
# Description: Create LUG on a vserver
#===============================================================================
sub _create_local_user {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $user_name = $params->get('user-name');
    my $password = $params->get('password');
    my $status = FALSE;

    try {
        NACL::C::VserverCifsUsersAndGroupsLocalUser->create(
            command_interface => $node,
            'vserver' => $vserver,
            'user-name' => $user_name,
            'password' => $password

        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already exist/i) {
            $Log->comment('LUG is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to turn create LUG on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Local User created');
    } else {
        return NATE::Result::Fail->new('Local User create failed');
    }
} # End of _create_local_user

#===============================================================================
# Description: Create LUG on a vserver
#===============================================================================
sub _create_local_group {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $group_name = $params->get('group-name');
    my $status = FALSE;

    try {
        NACL::C::VserverCifsUsersAndGroupsLocalGroup->create(
            command_interface => $node,
            'vserver' => $vserver,
            'group-name' => $group_name,

        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('LUG is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to turn create LUG on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Local group created');
    } else {
        return NATE::Result::Fail->new('Local group create failed');
    }
} # End of _create_local_group

#===============================================================================
# Description: Create LUG on a vserver
#===============================================================================
sub _add_user_to_group {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $group_name = $params->get('group-name');
    my $member_names = $params->get('member-names');
    my $status = FALSE;

    try {
        NACL::C::VserverCifsUsersAndGroupsLocalGroup->add_members(
            command_interface => $node,
            'vserver' => $vserver,
            'group-name' => $group_name,
            'member-names' => $member_names

        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate/i) {
            $Log->comment('LUG is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to turn create LUG on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Local members add to group');
    } else {
        return NATE::Result::Fail->new('Local members added to group failed');
    }
} # End of _add_user_to_group

#===============================================================================
# Description: Create LUG on a vserver
#===============================================================================
sub _add_privileges {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $user_or_group_name = $params->get('user-or-group-name');
    my @privileges = $params->get('privileges');
    my $status = FALSE;

    try {
        NACL::C::VserverCifsUsersAndGroupsPrivilege->add_privilege(
            command_interface => $node,
            'vserver' => $vserver,
            'user-or-group-name' => $user_or_group_name,
            'privileges' => @privileges

        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate/i) {
            $Log->comment('LUG is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to add privileges to user/group');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Added Privilege');
    } else {
        return NATE::Result::Fail->new('Add Privilege failed');
    }
} # End of _add_privileges

#===============================================================================
# Description: Create LUG on a vserver
#===============================================================================
sub _remove_user_from_group {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $group_name = $params->get('group-name');
    my $member_names = $params->get('member-names');
    my $status = FALSE;

    try {
        NACL::C::VserverCifsUsersAndGroupsLocalGroup->remove_members(
            command_interface => $node,
            'vserver' => $vserver,
            'group-name' => $group_name,
            'member-names' => $member_names

        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('LUG is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to turn create LUG on a vserver');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'LUG.created.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Local members remove from group');
    } else {
        return NATE::Result::Fail->new('Local members removed from group failed');
    }
} # End of _remove_user_from_group

#===============================================================================
# Description: Create LUG on a vserver
#===============================================================================
sub _delete_local_user {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $user_name = $params->get('user-name');
    my $status = FALSE;

    try {
        NACL::C::VserverCifsUsersAndGroupsLocalUser->delete(
            command_interface => $node,
            'vserver' => $vserver,
            'user-name' =>  => $user_name,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /found/i) {
            $Log->comment('LUG is already deleted on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to turn delete LUG on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Local User deleted');
    } else {
        return NATE::Result::Fail->new('Local User delete failed');
    }
} # End of _delete_local_user

#===============================================================================
# Description: Create LUG on a vserver
#===============================================================================
sub _delete_local_group {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $group_name = $params->get('group-name');
    my $status = FALSE;

    try {
        NACL::C::VserverCifsUsersAndGroupsLocalGroup->delete(
            command_interface => $node,
            'vserver' => $vserver,
            'group-name' =>  => $group_name,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /found/i) {
            $Log->comment('LUG is already deleted on this vserver.');
            $status = TRUE;
        } else {
            $Log->comment('Failed to turn delete LUG on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Local group deleted');
    } else {
        return NATE::Result::Fail->new('Local group delete failed');
    }
} # End of _delete_local_group

#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/LUG>

=cut

1;

