################################################################################
# Copyright (c) 2001-2013 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
#
#   Author:
#       dl-ptefit-cifs@netapp.com
#
#   Description:
#       This is the perl module for Roaming Profile.
#
#   Parameters:
#       TBD
#
# @keywords PTEFIT-CIFS
################################################################################
package PTEFIT::Cifs::Misfit::Features::RoamingProfile;

# Compiler Directives
use strict;
use warnings;
use TCD;
use Server;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Vserver;
use NATE::ParamSet;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );
use vars qw($Test);

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a new RoamingProfile Misfit object
#
# Returns:
#   A new RoamingProfile Misfit object
#
# Example:
#   my $roaming = PTEFIT::Cifs::Misfit::Features::RoamingProfile->new();
#===============================================================================

sub new {
    my ($class, %args) = @_;
    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id}   = "PTEFIT::Cifs::Misfit::Features::RoamingProfile";
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }
    $self->{node} = $node;
    $self->{node}->apiset_must(interface => 'ZAPI');
    $self->{node}->apiset();

    $Log->exit();
    return $self;
} #End of new

#===============================================================================
#  Function: loginLogoff
#   This method performs login and logoff of roaming users depending on the
#   input parameters.
#
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      RoamingProfile Attributes:
#        volume     - Volume name
#        client     - Windows client from which the login/logoffs are executed
#        share      - CIFS Share in which the roaming profiles should be stored
#        server     - CIFS Server name
#        dc         - Domain controller name
#        domain     - Domain to which the windows clients' belong
#        suser      - Windows clients super user
#        spassword  - Windows clients super user password
#        user       - Base name used to generate list of users
#        password   - Password for above users
#        sleep      - Wait time for each user thread after a write operation
#        deleteUser - Flag to indicate whether to delete users and create new ones
#        users      - Total number of user login/logoffs from each client
#        timeout    - Timeout for main thread to wait on all user threads
#        dcTimer    - Wait time for information to be updated in domain controller
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      > my $status = $roaming->loginLogoff(
#      >  objects => [
#      >          { volume     => 'vol_base222',
#      >            client     => 'ssan-2k8-64-149',
#      >            share      => 's_vol_base222',
#      >            server     => '10.228.131.152',
#      >            dc         => 'ssan-gdl-ad1',
#      >            domain     => 'sj',
#      >            suser      => 'minwin',
#      >            spassword  => 'netapp1!',
#      >            user       => 'ptefit',
#      >            password   => 'netapp1!',
#      >            sleep      => '5',
#      >            deleteUser => '0',
#      >            users      => '100',
#      >            timeout    => '300',
#      >            dcTimer    => '10',
#      >          },
#      >      ],
#      >  id => 'login_logoff'
#      > );
#===============================================================================
sub loginLogoff {
    my ($self, %args) = @_;
    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined ($volumes)) {
        my $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'ZAPI');
        my @vols = NACL::C::Volume->find(
            command_interface => $node,
            filter => $filter
        );
        for (my $i = 0; $i < $count; $i++) {
            my $vserver = $vols[$i]->get_vserver();
            $volumes->[$i] =  {
                vserver  => $vserver->name(),
                volume   => $vols[$i]->volume()
            };
        }
    }

    my @volume_info = ();
    my @msg_data = ();

    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = {$volume->{id}} . "_loginLogoff";
        } else {
            $runid = $volume->{volume} . "_loginLogoff";
        }
        push(@msg_data, {
            volume => $volume->{volume},
        });
        push (@volume_info, {
            codespec => \&_loginLogoff,
            runid => $runid,
            background => $background,
            args => [$self,%{$volume}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "loginLogoff",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Roaming profile login/logoff test failed!");
        return FALSE;
    }
} #End of on

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Execute Login/Logoff
#===============================================================================
sub _loginLogoff {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $volume = &ParamsMgr::get_arg(
        'arg'      => "volume",
        'required' => TRUE,
    );
    my $client = &ParamsMgr::get_arg(
        'arg'      => "client",
        'required' => TRUE,
    );
    my $share = &ParamsMgr::get_arg(
        'arg'      => "share",
        'required' => TRUE,
    );
    my $server = &ParamsMgr::get_arg(
        'arg'      => "server",
        'required' => TRUE,
    );
    my $dc = &ParamsMgr::get_arg(
        'arg'      => "dc",
        'required' => TRUE,
    );
    my $domain = &ParamsMgr::get_arg(
        'arg'      => "domain",
        'required' => TRUE,
    );
    my $suser = &ParamsMgr::get_arg(
        'arg'      => "suser",
        'required' => TRUE,
    );
    my $spassword = &ParamsMgr::get_arg(
        'arg'      => "spassword",
        'required' => TRUE,
    );
    my $user = &ParamsMgr::get_arg(
        'arg'      => "user",
        'required' => TRUE,
    );
    my $password = &ParamsMgr::get_arg(
        'arg'      => "password",
        'required' => TRUE,
    );
    my $sleep = &ParamsMgr::get_arg(
        'arg'      => "sleep",
        'required' => TRUE,
    );
    my $deleteUser = &ParamsMgr::get_arg(
        'arg'      => "deleteUser",
        'required' => TRUE,
    );
    my $users = &ParamsMgr::get_arg(
        'arg'      => "users",
        'required' => TRUE,
    );
    my $timeout = &ParamsMgr::get_arg(
        'arg'      => "timeout",
        'required' => TRUE,
    );
    my $dcTimer = &ParamsMgr::get_arg(
        'arg'      => "dcTimer",
        'required' => TRUE,
    );
    my $status    = TRUE;
    my $node = $self->{node};
    my $result    = "";

    try {
        $node->apiset_must(interface => 'CLI');
        #Invoke upstress.exe which performs login/logoffs.

        if($deleteUser>0){
           $deleteUser="-deleteuser";
        } else {
	       $deleteUser="";
	   }

        $Log->comment("Share: " . $share . " Server ip: " . $server);
        $result = `rsh $client \"X:\\bin\\upstressMT.exe -share $share -server $server -v -dc $dc -domain $domain -suser $suser -spassword $spassword -user $user -password $password -sleep $sleep $deleteUser -users $users -timeout $timeout -dcTimer $dcTimer \"`;

        $Log->comment('Result from upstress execution:' . $result);
        $result =~  s/^\s+|\s+$//g;
        if($result eq ""){
            $Log->comment("No result returned. upstressMT.exe may not have been executed!");
            $status=FALSE;
        }
        if($result =~ /(Error.*)/i) {
	        $Log->comment($1);
            $status=FALSE;
        }
    } otherwise {
        my $exception = shift;
        $Log->error('Failed to execute login/logoff');
    };

    if ($status) {
	   return TRUE;
    } else {
        NATE::BaseException->throw("Roaming profile failed");
        return FALSE;
    }
} #End of _loginLogoff

#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/RoamingProfile>

=cut

1;
