# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Vserver antiVirus module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

package PTEFIT::Cifs::Misfit::Features::antiVirus;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Vserver;
use NACL::C::AntivirusEngine;
use NACL::C::AntivirusOnAccess;
use NACL::C::AntivirusOnAccessPolicy;
use NACL::C::Antivirus;
use NATE::ParamSet;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;
use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );


# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Description: antiVirus constructor
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a new antiVirus Misfit object
#
# Returns:
#   A new antiVirus Misfit object
#
# Example:
#   my $roaming = PTEFIT::Cifs::Misfit::Features::antiVirus->new();
#===============================================================================

sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id}   = "PTEFIT::Cifs::Misfit::Features::antiVirus";
    $self->{node} = NACL::C::Node->find();
    $self->{node}->apiset_must(interface => 'ZAPI');
    $self->{node}->apiset();
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#===============================================================================
#  Function: enable_av_engine
#   This method enabled antivirus engine for a particular vendor
#
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      antiVirus Attributes:
#        vendor   - AntiVirus vendor name (trendmicro/mcafee/sophos)
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $antiVirusy->enable_av_engine(
#        objects => [
#                { vendor     => 'trendmicro',
#                },
#            ],
#        id => 'enable_av_engine'
#    );
#
#
#===============================================================================


#===============================================================================
# Description: Enable antiVirus engine
#===============================================================================
sub enable_av_engine {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($vservers)) {
        $vservers = $self->_get_vserver_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }
    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_enable_av_engine";
        } else {
            $runid = $vserver->{vserver} . "_enable_av_engine";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_enable_av_engine(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "enable_av_engine",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("antiVirus engine enable failed");
        return FALSE;
    }
} # End of enable av engine

#===============================================================================
#  Function: disable_av_engine
#   This method disables the antivirus engine for the current vendor configured
# 
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#       
#      antiVirus Attributes:
#        none
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#       
# Example:
#      my $status = $antiVirusy->enable_av_engine(
#        id => 'disable_av_engine'
#    );
#
#
#===============================================================================
sub disable_av_engine {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($vservers)) {
        $vservers = $self->_get_vserver_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }
    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_disable_av_engine";
        } else {
            $runid = $vserver->{vserver} . "_disable_av_engine";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_disable_av_engine(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "disable_av_engine",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("antiVirus disable engine failed");
        return FALSE;
    }
} # End of disable engine av

#===============================================================================
#  Function: modify_av_engine
#   This method modifies the antivirus engine to the specified vendor
# 
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#       
#      antiVirus Attributes:
#        vendor                - AntiVirus vendor name (trendmicro/mcafee/sophos)
#        trend_http_url        - Installation site url
#        trend_activation_code - activation code
#        timeout               - Wait time for job to complete
#   
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#       
# Example:
#      my $status = $antiVirusy->enable_av_engine(
#        objects => [
#                { vendor                => 'trendmicro',
#                  trend_http_url        => "http://spna-p.activeupdate.trendmicro.com/activeupdate/",
#                  trend_activation_code => PX-5FHE-GY2MC-FQNKZ-8JKRK-9CUQD-B7Z7G",
#                  timeout               => 360,
#                },
#            ],
#        id => 'modify_av_engine'
#    );
#
#
#===============================================================================
sub modify_av_engine {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($vservers)) {
        $vservers = $self->_get_vserver_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }
    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_modify_av_engine";
        } else {
            $runid = $vserver->{vserver} . "_modify_av_engine";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_modify_av_engine(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "modify_av_engine",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("antiVirus modify engine failed");
        return FALSE;
    }
} # End of av engine modify

#===============================================================================
#  Function: create_on_access_policy
#   This method creates on-access policy on a vserver
#
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver                 - Vserver in which the security policy should be created
#        vendor                  - vendor name 
#        policy_name             - policy name
#        scan                    - Turn on/off scanning
#        scan_ro_volume          - Turn on/off scanning read only volumes
#        scan_mandatory          - Turn on/off mandatory scan
#        protocols               - List of protocols,
#        cifs_share              - Include/Exclude cifs share,
#        cifs_share_list         - List of Cifs share or pattern,
#        cifs_scan_execute_acces - Turn on/off scanning files with execute access
#        cifs_file               - Include/Exclude Cifs files
#        cifs_file_list          - List of Cifs files or pattern
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $fSecurity->create_on_access_policy(
#        objects => [
#                { vserver                 => 'vs_base3',
#                  policy_name             => 'policy1',
#                  vendor                  => 'trendmicro',
#                  scan                    => 'on',
#                  scan_ro_volume          => 'off',
#                  scan_mandatory          => 'on',
#                  protocols               => ["cifs"],
#                  cifs_share              => 'include',
#                  cifs_share_list         => ["s_vol_base222"],
#                  cifs_scan_execute_acces => 'off',
#                  cifs_file               => 'include',
#                  cifs_file_list          => "\"^.*\$\"",
#                },
#            ],
#        id => 'create_on_access_policy'
#    );
#
#
#===============================================================================

sub create_on_access_policy {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($vservers)) {
        $vservers = $self->_get_vserver_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }
    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_on_access_policy";
        } else {
            $runid = $vserver->{vserver} . "_create_on_access_policy";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_create_on_access_policy(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_on_access_policy",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("antiVirus create on_access policy failed");
        return FALSE;
    }
} # End of create on-access policy

#===============================================================================
#  Function: delete_on_access_policy
#   This method deletes on-access policy on a vserver
#
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver                 - Vserver in which the security policy should be created
#        policy_name             - policy name
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $fSecurity->delete_on_access_policy(
#        objects => [
#                { vserver                 => 'vs_base3',
#                  policy_name             => 'policy1',
#                },
#            ],
#        id => 'delete_on_access_policy'
#    );
#
#
#===============================================================================


#===============================================================================
# Description: Delete on-access policy
#===============================================================================
sub delete_on_access_policy {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    ); 
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );  
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );  
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );  
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );  
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );  
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }
    
    unless (defined($vservers)) {
        $vservers = $self->_get_vserver_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }
    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete_on_access_policy";
        } else {
            $runid = $vserver->{vserver} . "_delete_on_access_policy";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_delete_on_access_policy(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete_on_access_policy",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("antiVirus delete on_access policy failed");
        return FALSE;
    }

} # End of delete on-access policy


#===============================================================================
#  Function: create_on_access_policy
#   This method modifies on-access policy on a vserver
#
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver                 - Vserver in which the security policy should be created
#        vendor                  - vendor name
#        policy_name             - policy name
#        scan                    - Turn on/off scanning
#        scan_ro_volume          - Turn on/off scanning read only volumes
#        scan_mandatory          - Turn on/off mandatory scan
#        protocols               - List of protocols,
#        cifs_share              - Include/Exclude cifs share,
#        cifs_share_list         - List of Cifs share or pattern,
#        cifs_scan_execute_acces - Turn on/off scanning files with execute access
#        cifs_file               - Include/Exclude Cifs files
#        cifs_file_list          - List of Cifs files or pattern
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $fSecurity->create_policy(
#        objects => [
#                { vserver                 => 'vs_base3',
#                  policy_name             => 'policy1',
#                  vendor                  => 'trendmicro',
#                  scan                    => 'off',
#                  scan_ro_volume          => 'on',
#                  scan_mandatory          => 'off',
#                  protocols               => ["cifs"],
#                  cifs_share              => 'include',
#                  cifs_share_list         => ["s_vol_base222"],
#                  cifs_scan_execute_acces => 'off',
#                  cifs_file               => 'include',
#                  cifs_file_list          => "\"^.*\$\"",
#                },
#            ],
#        id => 'modify_on_access_policy'
#    );
#
#
#===============================================================================
sub modify_on_access_policy {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    ); 
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );  
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );  
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );  
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );  
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );  
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }
    
    unless (defined($vservers)) {
        $vservers = $self->_get_vserver_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }
    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_modify_on_access_policy";
        } else {
            $runid = $vserver->{vserver} . "_modify_on_access_policy";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_modify_on_access_policy(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "modify_on_access_policy",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("antiVirus modify on_access policy failed");
        return FALSE;
    }

} # End of modify on-access policy

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Responsible for generating a list of vservers matching a criteria
#===============================================================================
sub _get_vserver_list {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $operation = &ParamsMgr::get_arg(
        'arg'      => "operation",
        'required' => FALSE,
        'default'  => '',
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vservers = ();
    my @all_vservers = NACL::C::Vserver->find_data_vservers(
        command_interface => $self->{node},
        filter => $filter
    );

    if ($count =~ /^all$/i) {
        $count = scalar(@all_vservers);
    }

    for (my $i = 0; $i < $count; $i++) {
        my $vserver = $all_vservers[$i]->name();
        $vservers[$i] =  {
            vserver  => $vserver,
            %{$params}
        };
    }

    return wantarray ? @vservers : \@vservers;
} # End of _get_vvserver_list

#===============================================================================
# Description: Enable antiVirus engine
#===============================================================================
sub _enable_av_engine {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vendor  = $params->get('vendor');
    my $status = FALSE;

    try {
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.enable_av_engine.started',
                data => {
                    node    => $node->{node},
                    vendor  => $vendor,
                }
            }
        );
        NACL::C::AntivirusEngine->enable(
            command_interface => $node,
            vendor            => $vendor,
        );
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.enable_av_engine.finished',
                data => {
                    node    => $node->{node},
                    vendor  => $vendor,
                }
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already enabled/i) {
            $Log->comment('Antivirus engine is already created for this vendor.');
            $MessageBroker->publish_message(
                exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
                routing_key => '#',
                message => {
                    msg => 'antiVirus.enable_av_engine.finished',
                    data => {
                        node    => $node->{node},
                        vendor  => $vendor,
                    }
                }
            );
            $status = TRUE;
        } else {
            $Log->error('Failed to enable antivirus engine for this vendor');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'antiVIrus.enable_av_engine.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vendor  => $vendor,
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('antiVirus engine enabled');
    } else {
        return NATE::Result::Fail->new('antiVirus engine enable failed');
    }
} # End of _enable_av_engine

#===============================================================================
# Description: Disable antiVirus engine
#===============================================================================
sub _disable_av_engine {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $status = FALSE;

    try {
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.disable_av_engine.started',
                data => {
                    node    => $node->{node},
                }
            }
        );
        NACL::C::AntivirusEngine->disable(
            command_interface => $node,
        );
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.disable_av_engine.finished',
                data => {
                    node    => $node->{node},
                }
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already disabled/i) {
            $Log->comment('Antivirus engine is already disabled for this vendor.');
            $MessageBroker->publish_message(
                exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
                routing_key => '#',
                message => {
                    msg => 'antiVirus.disable_av_engine.finished',
                    data => {
                        node    => $node->{node},
                    }
                }
            );
            $status = TRUE;
        } else {
            $Log->error('Failed to disable antivirus engine for this vendor');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'antiVirus.disable_av_engine.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('antiVirus engine disabled');
    } else {
        return NATE::Result::Fail->new('antiVirus engine disable failed');
    }
} # End of _disable_av_engine


#===============================================================================
# Description: Modify antiVirus engine
#===============================================================================
sub _modify_av_engine {
    my ($class) = @_;
    
    # Global params
    my $params                = NATE::ParamSet->get_global();
    my $node                  = $params->get('node');
    my $vendor                = $params->get('vendor');
    my $trend_http_url        = $params->get('trend_http_url');
    my $trend_activation_code = $params->get('trend_activation_code');
    my $timeout               = $params->get('timeout');
    my $status = FALSE;  
    my $job;
            
    try {   
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.modify_av_engine.started',
                data => {
                    node                    => $node->{node},
                    vendor                  => $vendor,
                    'trend-http-url'        => $trend_http_url,
                    'trend-activation-code' => $trend_activation_code, 
                    timeout                 => $timeout,
                }
            }
        );


        NACL::C::AntivirusEngine->modify(
            command_interface       => $node,
            vendor                  => $vendor,
            'trend-http-url'        => $trend_http_url,
            'trend-activation-code' => $trend_activation_code,
            job_component           => \$job,

        );

        if (defined ($job)) {
            $job->wait_on_job('method-timeout' => $timeout);
        }

        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.modify_av_engine.finished',
                data => {
                    node                    => $node->{node},
                    vendor                  => $vendor,
                    'trend-http-url'        => $trend_http_url,
                    'trend-activation-code' => $trend_activation_code,
                    timeout                 => $timeout,
                }
            }
        );  
        $status = TRUE;
    } otherwise {
        my $exception = shift; 
        $Log->error('Failed to modify antivirus engine for this vendor');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message => {
                msg => 'antiVrrus.modify_av_engine.' . $exception->text(),
                data => {
                          node    => $node->{node},
                          vendor                  => $vendor,
                          'trend-http-url'        => $trend_http_url,
                          'trend-activation-code' => $trend_activation_code,
                          timeout                 => $timeout,
                }
            }
         );
         NATE::BaseException->throw($exception);
    };          
                    
    if ($status) {  
        return NATE::Result::Pass->new('antiVirus engine modified');
    } else {            
        return NATE::Result::Fail->new('antiVirus engine modify failed');
    }           
} # End of _modify_av_engine

#===============================================================================
# Description: Create on_access policy engine
#===============================================================================
sub _create_on_access_policy {
    my ($class) = @_;
                    
    # Global params
    my $params                   = NATE::ParamSet->get_global();
    my $node                     = $params->get('node');
    my $vserver                  = $params->get('vserver');
    my $policy_name              = $params->get('policy_name');
    my $scan                     = $params->get('scan');
    my $scan_ro_volume           = $params->get('scan_ro_volume');
    my $scan_mandatory           = $params->get('scan_mandatory');
    my $protocols                = $params->get('protocols');
    my $cifs_share               = $params->get('cifs_share');
    my $cifs_share_list          = $params->get('cifs_share_list');
    my $cifs_scan_execute_access = $params->get('cifs_scan_execute_access');
    my $cifs_file                = $params->get('cifs_file');
    my $cifs_file_list           = $params->get('cifs_file_list');
    my $status = FALSE;
                
    try {   
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.create_on_access_policy.started',
                data => {
                    node                       => $node->{node},
                    vserver                    => $vserver,
                    name                       => $policy_name,
                    scan                       => $scan,
                    'scan-ro-volume'           => $scan_ro_volume,
                    'scan-mandatory'           => $scan_mandatory,
                    protocols                  => $protocols,
                    'cifs-share'               => $cifs_share,
                    'cifs-share-list'          => $cifs_share_list,
                    'cifs-scan-execute-access' => $cifs_scan_execute_access,
                    'cifs-file'                => $cifs_file,
                    'cifs-file-list'           => $cifs_file_list,
                }
            }
        );
        NACL::C::AntivirusOnAccessPolicy->create(
            command_interface       => $node,
            vserver                    => $vserver,
            name                       => $policy_name,
            scan                       => $scan,
            'scan-ro-volume'           => $scan_ro_volume,
            'scan-mandatory'           => $scan_mandatory,
            protocols                  => $protocols,
            'cifs-share'               => $cifs_share,
            'cifs-share-list'          => $cifs_share_list,
            'cifs-scan-execute-access' => $cifs_scan_execute_access,
            'cifs-file'                => $cifs_file,
            'cifs-file-list'           => $cifs_file_list,

        );
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.create_on_access_policy.finished',
                data => {
                    node                    => $node->{node},
                    vserver                    => $vserver,
                    name                       => $policy_name,
                    scan                       => $scan,
                    'scan-ro-volume'           => $scan_ro_volume,
                    'scan-mandatory'           => $scan_mandatory,
                    protocols                  => $protocols,
                    'cifs-share'               => $cifs_share,
                    'cifs-share-list'          => $cifs_share_list,
                    'cifs-scan-execute-access' => $cifs_scan_execute_access,
                    'cifs-file'                => $cifs_file,
                    'cifs-file-list'           => $cifs_file_list,
                }
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate entry /i) {
            $Log->comment('Antivirus engine is already created on this vserver.');
            $MessageBroker->publish_message(
                exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
                routing_key => '#',
                message => {
                    msg => 'antiVirus.create_on_access_policy.finished',
                    data => {
                        node    => $node->{node},
                        vserver                    => $vserver,
                        name                       => $policy_name,
                        scan                       => $scan,
                        'scan-ro-volume'           => $scan_ro_volume,
                        'scan-mandatory'           => $scan_mandatory,
                        protocols                  => $protocols,
                        'cifs-share'               => $cifs_share,
                        'cifs-share-list'          => $cifs_share_list,
                        'cifs-scan-execute-access' => $cifs_scan_execute_access,
                        'cifs-file'                => $cifs_file,
                        'cifs-file-list'           => $cifs_file_list,
                    }
                }
            );
            $status = TRUE;
        } else {
            $Log->error('Failed to create on-access policy on this vserver');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'antiVirus.create_on_access_policy.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver                    => $vserver,
                        name                       => $policy_name,
                        scan                       => $scan,
                        'scan-ro-volume'           => $scan_ro_volume,
                        'scan-mandatory'           => $scan_mandatory,
                        protocols                  => $protocols,
                        'cifs-share'               => $cifs_share,
                        'cifs-share-list'          => $cifs_share_list,
                        'cifs-scan-execute-access' => $cifs_scan_execute_access,
                        'cifs-file'                => $cifs_file,
                        'cifs-file-list'           => $cifs_file_list,
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };


    if ($status) {
        return NATE::Result::Pass->new('antiVirus on-access policy created');
    } else {
        return NATE::Result::Fail->new('antiVirus on-access policy create failed');
    }
} # End of _create_on_access_policy

#===============================================================================
# Description: Modify on_access policy engine
#===============================================================================
sub _modify_on_access_policy {
    my ($class) = @_;

    # Global params
    my $params                   = NATE::ParamSet->get_global();
    my $node                     = $params->get('node');
    my $vserver                  = $params->get('vserver');
    my $policy_name              = $params->get('policy_name');
    my $scan                     = $params->get('scan');
    my $scan_ro_volume           = $params->get('scan_ro_volume');
    my $scan_mandatory           = $params->get('scan_mandatory');
    my $protocols                = $params->get('protocols');
    my $cifs_share               = $params->get('cifs_share');
    my $cifs_share_list          = $params->get('cifs_share_list');
    my $cifs_scan_execute_access = $params->get('cifs_scan_execute_access');
    my $cifs_file                = $params->get('cifs_file');
    my $cifs_file_list           = $params->get('cifs_file_list');
    my $status = FALSE;

    try {
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.modify_on_access_policy.started',
                data => {
                    node                       => $node->{node},
                    vserver                    => $vserver,
                    name                       => $policy_name,
                    scan                       => $scan,
                    'scan-ro-volume'           => $scan_ro_volume,
                    'scan-mandatory'           => $scan_mandatory,
                    protocols                  => $protocols,
                    'cifs-share'               => $cifs_share,
                    'cifs-share-list'          => $cifs_share_list,
                    'cifs-scan-execute-access' => $cifs_scan_execute_access,
                    'cifs-file'                => $cifs_file,
                    'cifs-file-list'           => $cifs_file_list,
                }
            }
        );
        NACL::C::AntivirusOnAccessPolicy->modify(
            command_interface       => $node,
            vserver                    => $vserver,
            name                       => $policy_name,
            scan                       => $scan,
            'scan-ro-volume'           => $scan_ro_volume,
            'scan-mandatory'           => $scan_mandatory,
            protocols                  => $protocols,
            'cifs-share'               => $cifs_share,
            'cifs-share-list'          => $cifs_share_list,
            'cifs-scan-execute-access' => $cifs_scan_execute_access,
            'cifs-file'                => $cifs_file,
            'cifs-file-list'           => $cifs_file_list,
            
        );      
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',             
            message => {
                msg => 'antiVirus.modify_on_access_policy.finished',
                data => {
                    node                    => $node->{node},
                    vserver                    => $vserver,
                    name                       => $policy_name,
                    scan                       => $scan,
                    'scan-ro-volume'           => $scan_ro_volume,
                    'scan-mandatory'           => $scan_mandatory,
                    protocols                  => $protocols,
                    'cifs-share'               => $cifs_share,
                    'cifs-share-list'          => $cifs_share_list,
                    'cifs-scan-execute-access' => $cifs_scan_execute_access,
                    'cifs-file'                => $cifs_file,
                    'cifs-file-list'           => $cifs_file_list,
                }
            }             
        );                
        $status = TRUE;   
    } otherwise {
        my $exception = shift;
            $Log->error('Failed to modify on-access policy on this vserver');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'antiVirus.modify_on_access_policy.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver                    => $vserver,
                        name                       => $policy_name,
                        scan                       => $scan,
                        'scan-ro-volume'           => $scan_ro_volume,
                        'scan-mandatory'           => $scan_mandatory,
                        protocols                  => $protocols,
                        'cifs-share'               => $cifs_share,
                        'cifs-share-list'          => $cifs_share_list,
                        'cifs-scan-execute-access' => $cifs_scan_execute_access,
                        'cifs-file'                => $cifs_file,
                        'cifs-file-list'           => $cifs_file_list,
                    }
                }
            );
            NATE::BaseException->throw($exception);
        
    };


    if ($status) {
        return NATE::Result::Pass->new('antiVirus on-access policy modified');
    } else {
        return NATE::Result::Fail->new('antiVirus on-access policy modified failed');
    }
} # End of _modify_on_access_policy

#===============================================================================
# Description: Create on_access policy engine
#===============================================================================
sub _delete_on_access_policy {
    my ($class) = @_;

    # Global params
    my $params                   = NATE::ParamSet->get_global();
    my $node                     = $params->get('node');
    my $vserver                  = $params->get('vserver');
    my $policy_name              = $params->get('policy_name');
    my $status = FALSE;

    try {
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.deletee_on_access_policy.started',
                data => {
                    node                       => $node->{node},
                    vserver                    => $vserver,
                    name                       => $policy_name,
                }
            }
        );
        NACL::C::AntivirusOnAccessPolicy->delete(
            command_interface       => $node,
            vserver                    => $vserver,
            name                       => $policy_name,
        );
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
            routing_key => '#',
            message => {
                msg => 'antiVirus.delete_on_access_policy.finished',
                data => {
                    node                    => $node->{node},
                    vserver                    => $vserver,
                    name                       => $policy_name,
                }
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /cannot find/i) {
            $Log->comment('Antivirus engine is already created on this vserver.');
            $MessageBroker->publish_message(
                exchange    => 'PTEFIT::Cifs::Misfit::Features::antiVirus',
                routing_key => '#',
                message => {
                    msg => 'antiVirus.delete_on_access_policy.finished',
                    data => {
                        node    => $node->{node},
                        vserver                    => $vserver,
                        name                       => $policy_name,
                    }
                }
            );
            $status = TRUE;
        } else {
            $Log->error('Failed to delete on-access policy on this vserver');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'antiVirus.delete_on_access_policy.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver                    => $vserver,
                        name                       => $policy_name,
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };


    if ($status) {
        return NATE::Result::Pass->new('antiVirus on-access policy deleted');
    } else {
        return NATE::Result::Fail->new('antiVirus on-access policy delete failed');
    }
} # End of _delete_on_access_policy



#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/Audit>

=cut

1;
