################################################################################
# Copyright (c) 2001-2013 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
#
#   Author:
#       dl-ptefit-cifs@netapp.com
#
#   Description:
#       This perl module for fSecurity
#
#   Parameters:
#       TBD
#
# @keywords PTEFIT-CIFS
################################################################################

package PTEFIT::Cifs::Misfit::Features::fSecurity;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Vserver;
use NACL::C::VserverSecurityFileDirectory;
use NACL::C::VserverSecurityFileDirectoryPolicy;
use NACL::C::VserverSecurityFileDirectoryPolicyTask;
use NACL::C::VserverSecurityFileDirectoryNtfsSacl;
use NACL::C::VserverSecurityFileDirectoryNtfs;
use NACL::C::VserverSecurityFileDirectoryNtfsDacl;
use NACL::C::VserverSecurityFileDirectoryJob;
use NATE::ParamSet;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;
use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================
#===============================================================================
# Constructor: new
#   Create a new fSecurity Misfit object
#
# Returns:
#   A new fSecurity Misfit object
#
# Example:
#   my $fSecurity = PTEFIT::Cifs::Misfit::Features::fSecurity->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id}   = "PTEFIT::Cifs::Misfit::Features::fSecurity";
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }
    $self->{node} = $node;
    $self->{node}->apiset_must(interface => 'ZAPI');
    $self->{node}->apiset();

    $Log->exit();
    return $self;
} # End of new

#===============================================================================
#  Function: create_policy
#   This method creates a security policy on a vserver
#
#  Parameters:
#      objects    - Array of hashes containing information to setup policy  on a vserver
#      count      - The number of vservers to create policy on
#      filter     - variable to filter through vservers to create policy on
#      params     - hash containing parameters for the policy  config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver     - Vserver in which the security policy should be created
#        policy_name - Security policy name
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $fSecurity->create_policy(
#        objects => [
#                { vserver     => 'vs_base3',
#                  policy_name => 'policy1',
#                },
#            ],
#        id => 'create_policy'
#    );
#
#
#===============================================================================
sub create_policy {
    my ($self, %args) = @_;
    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_policy";
        } else {
            $runid = $vserver->{vserver} . "_create_policy";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => \&_create_policy,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_policy",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("fSecurity policy create failed");
        return FALSE;
    }
} # End of create policy

#===============================================================================
#  Function:apply_policy
#   This method applies a security policy on a vserver
#
#  Parameters:
#      objects    - Array of hashes containing information to setup policy on a vserver
#      count      - The number of vservers to create policy on
#      filter     - variable to filter through vservers to create policy on
#      params     - hash containing parameters for the policy config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver     - Vserver in which the security policy should be created
#        policy_name - Security policy name
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $fSecurity->apply_policy(
#        objects => [
#                { vserver     => 'vs_base3',
#                  policy_name => 'policy1',
#                },
#            ],
#        id => 'apply_policy'
#    );
#
#
#===============================================================================
sub apply_policy {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_apply_policy";
        } else {
            $runid = $vserver->{vserver} . "_apply_policy";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => \&_apply_policy,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "apply_policy",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("fSecurity policy apply failed");
        return FALSE;
    }
} # End of apply  policy


#===============================================================================
#  Function:create_ntfs_descriptor
#   This method creates a ntfs security descriptor
#
#  Parameters:
#      objects    - Array of hashes containing information to setup policy on a vserver
#      count      - The number of vservers to create policy on
#      filter     - variable to filter through vservers to create policy on
#      params     - hash containing parameters for the policy config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver       - Vserver in which the security policy should be created
#        ntfs_sd       - security descriptor name
#        owner         - superuser
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $fSecurity->apply_policy(
#        objects => [
#                { vserver       => 'vs_base3',
#                  ntfs_sd       => 'sd1',
#                  owner         => 'sj\\minwin',
#                },
#            ],
#        id => 'create_ntfs_descriptor'
#    );
#
#
#===============================================================================
sub create_ntfs_descriptor {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );

    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_ntfs_descriptor";
        } else {
            $runid = $vserver->{vserver} . "_create_ntfs_descriptor";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_create_ntfs_descriptor,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_ntfs_descriptor",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("fSecurity create NTFS security descriptor failed");
        return FALSE;
    }
} # End of enable


#===============================================================================
#  Function: add_task
#   This method adds a task to the security policy
#
#  Parameters:
#      objects    - Array of hashes containing information to setup policy on a vserver
#      count      - The number of vservers to create policy on
#      filter     - variable to filter through vservers to create policy on
#      params     - hash containing parameters for the policy config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver       - Vserver in which the security policy should be created
#        policy_name   - Security policy name
#        path          - volume path
#        security_type - volume security type (ntfs/unix/mixed)
#        ntfs_mode     - type of security propogation
#        ntfs_sd       - security descriptor name
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $fSecurity->add_task(
#        objects => [
#                { vserver       => 'vs_base3',
#                  policy_name   => 'policy1',
#                  path          => '/vol_base222',
#                  security_type => 'ntfs',
#                  ntfs_mode     => 'propogate',
#                  ntfs_sd       => 'sd1',
#                },
#            ],
#        id => 'add_task'
#    );
#
#
#===============================================================================
sub add_task {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );

    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_add_task";
        } else {
            $runid = $vserver->{vserver} . "_add_task";
        }

        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_add_task,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "add_task",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Add task to policy failed");
        return FALSE;
    }
} #add_task

#===============================================================================
#  Function: add_daccess_control
#   This method adds discretionary access control on NTFS security descriptor
#
#  Parameters:
#      objects    - Array of hashes containing information to setup policy on a vserver
#      count      - The number of vservers to create policy on
#      filter     - variable to filter through vservers to create policy on
#      params     - hash containing parameters for the policy config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver         - Vserver in which the security policy should be created
#        ntfs_sd         - security descriptor name
#        account         - account name to which the access control is applied
#        access_type     - Access type (allow/deny)
#        advanced_rights - Advanced rights (Delete/Write-to/Append, etc)
#        apply_to        - Apply to 'this-folder/sub-folders/files'
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $fSecurity->add_task(
#        objects => [
#                { vserver         => 'vs_base3',
#                  ntfs_sd         => 'sd1',
#                  account         => 'sj\\minwin',
#                  access_type     => 'allow',
#                  advanced_rights => 'Append',
#                  apply_to        => 'files',
#                },
#            ],
#        id => 'add_daccess_control'
#    );
#
#
#===============================================================================
sub add_daccess_control {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );

    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_add_daccess_control";
        } else {
            $runid = $vserver->{vserver} . "_add_daccess_control";
        }

        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_add_daccess_control,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "add_daccess_control",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Add access control to NTFS security descriptor failed");
        return FALSE;
    }
} # End of add_daccess_control

#===============================================================================
#  Function: add_saccess_control
#   This method adds system access control on NTFS security descriptor
# 
#  Parameters:
#      objects    - Array of hashes containing information to setup policy on a vserver
#      count      - The number of vservers to create policy on
#      filter     - variable to filter through vservers to create policy on
#      params     - hash containing parameters for the policy config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver         - Vserver in which the security policy should be created
#        ntfs_sd         - security descriptor name
#        account         - account name to which the access control is applied
#        access_type     - Access type (success/failure)
#        rights          - Read/Write rights
#        advanced_rights - Advanced rights (Delete/Write-to/Append, etc)
#        apply_to        - Apply to 'this-folder/sub-folders/files'
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $fSecurity->add_task(
#        objects => [
#                { vserver         => 'vs_base3',
#                  ntfs_sd         => 'sd1',
#                  account         => 'sj\\minwin',
#                  access_type     => 'success',
#                  rights          => 'read',
#                  advanced_rights => 'Append',
#                  apply_to        => 'this-folder',
#                },
#            ],
#        id => 'add_saccess_control'
#    );
#
#
#===============================================================================
sub add_saccess_control {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );

    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_add_saccess_control";
        } else {
            $runid = $vserver->{vserver} . "_add_saccess_control";
        }

        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => \&_add_saccess_control,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "add_saccess_control",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Add system access control to NTFS security descriptor failed");
        return FALSE;
    }
} # End of add_saccess_control


#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Create file-directory security policy on a vserver
#===============================================================================
sub _create_policy {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy_name",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        $node->apiset_must(interface => 'CLI');
        NACL::C::VserverSecurityFileDirectoryPolicy->create(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate entry/i) {
            $Log->comment('File-directory policy is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to create policy on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("fSecurity policy create failed");
        return FALSE;
    }

} # End of _create

#===============================================================================
# Description: Apply file-directory security policy on a vserver
#===============================================================================
sub _apply_policy {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy_name",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        $node->apiset_must(interface => 'CLI');
        NACL::C::VserverSecurityFileDirectory->apply(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already applied/i) {
            $Log->comment('File-directory policy is already applied on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to apply policy on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("fSecurity policy apply failed");
        return FALSE;
    }

} # End of _apply


#===============================================================================
# Description: Create ntfs security descriptor for a policy
#===============================================================================
sub _create_ntfs_descriptor {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $ntfs_sd = &ParamsMgr::get_arg(
        'arg'      => "ntfs_sd",
        'required' => TRUE,
    );
    my $owner = &ParamsMgr::get_arg(
        'arg'      => "owner",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverSecurityFileDirectoryNtfs->create(
            command_interface => $node,
            'vserver' => $vserver,
            'ntfs-sd' => $ntfs_sd,
            owner    => $owner,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate entry/i) {
            $Log->comment('NTFS security descriptor is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to create NTFS security descriptor on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("fSecurity NTFS security descriptor create failed");
        return FALSE;
    }

} # End of _enable

#===============================================================================
# Description: Add task to a policy
#===============================================================================
sub _add_task {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $policy_name = &ParamsMgr::get_arg(
        'arg'      => "policy_name",
        'required' => TRUE,
    );
    my $path = &ParamsMgr::get_arg(
        'arg'      => "path",
        'required' => TRUE,
    );
    my $security_type = &ParamsMgr::get_arg(
        'arg'      => "security_type",
        'required' => TRUE,
    );
    my $ntfs_mode = &ParamsMgr::get_arg(
        'arg'      => "ntfs_mode",
        'required' => TRUE,
    );
    my $ntfs_sd = &ParamsMgr::get_arg(
        'arg'      => "ntfs_sd",
        'required' => TRUE,
    );
    my $index_num = &ParamsMgr::get_arg(
        'arg'      => "index-num",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        $node->apiset_must(interface => 'CLI');
        NACL::C::VserverSecurityFileDirectoryPolicyTask->add(
            command_interface => $node,
            'vserver' => $vserver,
            'policy-name' => $policy_name,
            path    => $path,
           'security-type' => $security_type,
           'ntfs-mode' => $ntfs_mode,
           'ntfs-sd' => $ntfs_sd,
           'index-num' => $index_num,

        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate entry/i) {
            $Log->comment('Task is already added on this policy.');
            $status = TRUE;
        } else {
            $Log->error('Failed to add task on the policy');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("fSecurity add task to policy failed");
        return FALSE;
    }

} # End of _add_task

#===============================================================================
# Description: Add discretionary access control on NTFS security descriptor
#===============================================================================
sub _add_daccess_control {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $ntfs_sd = &ParamsMgr::get_arg(
        'arg'      => "ntfs_sd",
        'required' => TRUE,
    );
    my $account = &ParamsMgr::get_arg(
        'arg'      => "account",
        'required' => TRUE,
    );
    my $access_type = &ParamsMgr::get_arg(
        'arg'      => "access_type",
        'required' => TRUE,
    );
    my $advanced_rights = &ParamsMgr::get_arg(
        'arg'      => "advanced_rights",
        'required' => TRUE,
    );
    my $apply_to = &ParamsMgr::get_arg(
        'arg'      => "apply_to",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverSecurityFileDirectoryNtfsDacl->add(
            command_interface => $node,
            'vserver' => $vserver,
            'ntfs-sd' => $ntfs_sd,
            account   => $account,
            'access-type' =>  $access_type,
            'advanced-rights' => $advanced_rights,
            'apply-to' => $apply_to,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate entry/i) {
            $Log->comment('Discretionary access control is already added on this NTFS security descriptor.');
            $status = TRUE;
        } else {
            $Log->error('Failed to add access control on this NTFS security descriptor');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("fSecurity add discretionary access control failed");
        return FALSE;
    }


} # End of _add_daccess_control

#===============================================================================
# Description: Add discretionary access control on NTFS security descriptor
#===============================================================================
sub _add_saccess_control {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $ntfs_sd = &ParamsMgr::get_arg(
        'arg'      => "ntfs_sd",
        'required' => TRUE,
    );
    my $account = &ParamsMgr::get_arg(
        'arg'      => "account",
        'required' => TRUE,
    );
    my $access_type = &ParamsMgr::get_arg(
        'arg'      => "access_type",
        'required' => TRUE,
    );
    my $rights = &ParamsMgr::get_arg(
        'arg'      => "rights",
        'required' => TRUE,
    );
    my $advanced_rights = &ParamsMgr::get_arg(
        'arg'      => "advanced_rights",
        'required' => TRUE,
    );
    my $apply_to = &ParamsMgr::get_arg(
        'arg'      => "apply_to",
        'required' => TRUE,
    );
    my $status = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverSecurityFileDirectoryNtfsSacl->add(
            command_interface => $node,
            'vserver' => $vserver,
            'ntfs-sd' => $ntfs_sd,
            account   => $account,
            'access-type' => $access_type,
            rights       => $rights,
            'apply-to' => $apply_to,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate entry/i) {
            $Log->comment('System access control is already added on this NTFS security descriptor.');
            $status = TRUE;
        } else {
            $Log->error('Failed to add system access control on this NTFS security descriptor');
            NATE::BaseException->throw($exception);
        }
    };
    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("fSecurity add system access control failed");
        return FALSE;
    }
} # End of _add_saccess_control

#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/RoamingProfile>

=cut

1;
