################################################################################
# Copyright (c) 2001-2013 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
#
#   Author:
#       dl-ptefit-cifs@netapp.com
#
#   Description:
#       This is the perl module for secTrace
#
#   Parameters:
#       TBD
#
# @keywords PTEFIT-CIFS
################################################################################
package PTEFIT::Cifs::Misfit::Features::secTrace;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Vserver;
use NACL::C::VserverSecurityTraceFilter;
use NACL::C::VserverSecurityTraceTraceResult;
use NATE::ParamSet;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;
# use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================
#===============================================================================
# Constructor: new
#   Create a new secTrace Misfit object
#
# Returns:
#   A new secTrace Misfit object
#
# Example:
#   my $secTrace = PTEFIT::Cifs::Misfit::Features::secTrace->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;
    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id}   = "PTEFIT::Cifs::Misfit::Features::secTrace";
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }
    $self->{node} = $node;
    $self->{node}->apiset_must(interface => 'CLI');
    $self->{node}->apiset();

    $Log->exit();
    return $self;
} # End of new

#===============================================================================
#  Function: create
#   This method creates a security trace filter on a vserver
#
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver     - Vserver in which the security policy should be created
#        win_name    - Windows user name
#        trace-allow - Trace allow events
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $secTrace->create(
#        objects => [
#                { vserver     => 'vs_base3',
#                  win_name    => 'sj\minwin',
#		   trace_allow => 'no',
#                },
#            ],
#        id => 'create'
#    );
#
#
#===============================================================================
sub create {
    my ($self, %args) = @_;
    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create";
        } else {
            $runid = $vserver->{vserver} . "_create";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => \&_create,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("secTrace filter create failed");
        return FALSE;
    }
} # End of create

#===============================================================================
#  Function: delete
#   This method creates a security trace filter on a vserver
#
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver     - Vserver in which the security policy should be created
#        index       - Filter index (integer)
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $secTrace->delete(
#        objects => [
#                { vserver     => 'vs_base3',
#                  index       => 1,
#                },
#            ],
#        id => 'delete'
#    );
#
#
#===============================================================================
sub delete {
    my ($self, %args) = @_;
    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete";
        } else {
            $runid = $vserver->{vserver} . "_delete";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => \&_delete,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("secTrace filter create failed");
        return FALSE;
    }
} # End of delete

#===============================================================================
#  Function: modify
#   This method shows security trace results
#
#  Parameters:
#      objects    - Array of hashes containing information to setup audit on a vserver
#      count      - The number of vservers to create audit on
#      filter     - variable to filter through vservers to create audit on
#      params     - hash containing parameters for the audit config
#      background - Option to run in serially or in parallel
#      iterations - number of times to run create
#
#      fSecurity Attributes:
#        vserver     - Vserver in which the security policy should be created
#        win_name    - user name
#        trace-allow - allow or deny trace
#
# Returns:
#      TRUE if successful, FALSE if unsuccessful
#
# Example:
#      my $status = $secTrace->modify(
#        objects => [
#                { vserver     => 'vs_base3',
#                  win_name    => 'sj\minwin',
#                  trace_allow => 'yes',
#                },
#            ],
#        id => 'modify'
#    );
#
#
#===============================================================================
sub modify {
    my ($self, %args) = @_;
    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_modify";
        } else {
            $runid = $vserver->{vserver} . "_modify";
        }
        push(@msg_data, {
            vserver => $vserver->{vserver},
        });
        push (@vserver_info, {
            codespec => \&_modify,
            runid => $runid,
            background => $background,
            args => [$self,%{$vserver}],
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "modify",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("secTrace filter create failed");
        return FALSE;
    }
} # End of modify


#===============================================================================
#  PRIVATE METHODS
#===============================================================================



#===============================================================================
# Description: Enable sectrace on a vserver
#===============================================================================
sub _create {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $win_name = &ParamsMgr::get_arg(
        'arg'      => "win_name",
        'required' => TRUE,
    );
    my $trace_allow = &ParamsMgr::get_arg(
        'arg'      => "trace_allow",
        'required' => TRUE,
    );
    my $index = &ParamsMgr::get_arg(
        'arg'      => "index",
        'required' => TRUE,
    );
    my $status        = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverSecurityTraceFilter->create(
            command_interface => $node,
            'vserver'     => $vserver,
            'windows-name'    => $win_name,
            'trace-allow' => $trace_allow,
             index          => $index,

        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate entry/i) {
            $Log->comment('NTFS security descriptor is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to create NTFS security descriptor on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("secTrace filter create failed");
        return FALSE;
    }

} # End of _create

#===============================================================================
# Description: Enable audit on a vserver
#===============================================================================
sub _delete {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $index = &ParamsMgr::get_arg(
        'arg'      => "index",
        'required' => TRUE,
    );
    my $status        = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverSecurityTraceFilter->delete(
            command_interface => $node,
            'vserver'     => $vserver,
            index    => $index,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
            $Log->error('Failed to create NTFS security descriptor on a vserver');
            NATE::BaseException->throw($exception);
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("secTrace filter delete failed");
        return FALSE;
    }


} # End of _delete

#===============================================================================
# Description: Enable audit on a vserver
#===============================================================================
sub _modify {
    my ($self, %args) = @_;

    # Global params
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $win_name = &ParamsMgr::get_arg(
        'arg'      => "win_name",
        'required' => TRUE,
    );
    my $trace_allow = &ParamsMgr::get_arg(
        'arg'      => "trace_allow",
        'required' => TRUE,
    );
    my $index = &ParamsMgr::get_arg(
        'arg'      => "index",
        'required' => TRUE,
    );
    my $status        = FALSE;
    my $node = $self->{node};

    try {
        NACL::C::VserverSecurityTraceFilter->modify(
            command_interface => $node,
            'vserver'     => $vserver,
            'windows-name'    => $win_name,
            'trace-allow' => $trace_allow,
            index          => $index,

        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
            $Log->error('Failed to modify NTFS security descriptor on a vserver');
            NATE::BaseException->throw($exception);
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("secTrace filter modify failed");
        return FALSE;
    }

} # End of _modify


#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/Audit>

=cut

1;
