# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Vserver CifsHost module for Misfit
## @author dl-ptefit-cifs@netapp.com
## @status Private

# Package: PTEFIT::Cifs::Misfit::Services::CifsHost
# Module to run Cif Share commands with Misfit messaging

package PTEFIT::Cifs::Misfit::Services::CifsHost;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::Client;
use NACL::C::Vserver;
use NACL::CS::Client::Mount;
use NACL::STask::Client::Mount;
use NACL::STask::VserverCifsShare;
use San::FIT::Misfit::Core::Workflow;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Description: CifsHost constructor
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "PTEFIT::Cifs::Misfit::Services::CifsHost";

    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );

    my $node = undef;
    my $node_cli = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
        $node_cli = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
        $node_cli = NACL::C::Node->find();
    }

    $node->apiset_must(interface => 'ZAPI');
    $node->apiset();
    $self->{node} = $node;
    $self->{node_cli} = $node_cli;
    $self->{node_cli}->apiset_must(interface => 'CLI');
    $self->{node_cli}->apiset();
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#===============================================================================
# Description: Creates a Share
#===============================================================================
sub create_share {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_share";
        } else {
            $runid = $vserver->{vserver} . "_create_share";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_create_share(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_share",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Create failed");
        return FALSE;
    }
} # End of create_share

#===============================================================================
# Description: Maps a Share
#===============================================================================
sub map_share {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_map_share";
        } else {
            $runid = $vserver->{vserver} . "_map_share";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_map_share(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "map_share",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Map failed");
        return FALSE;
    }
} # End of map_share

#===============================================================================
# Description: Unmaps a Share
#===============================================================================
sub unmap_share {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_unmap_share";
        } else {
            $runid = $vserver->{vserver} . "_unmap_share";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_unmap_share(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "unmap_share",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Unmap failed");
        return FALSE;
    }
} # End of unmap_share

#===============================================================================
# Description: Deletes a Share
#===============================================================================
sub delete_share {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete_share";
        } else {
            $runid = $vserver->{vserver} . "_delete_share";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_delete_share(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete_share",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Delete share failed");
        return FALSE;
    }
} # End of delete_share

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Create a share
#===============================================================================
sub _create_share {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $path = $params->get('path');
    my $share_name = $params->get('share_name');
    my $status = FALSE;

    try {
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Services::CifsHost',
            routing_key => '#',
            message => {
                msg => 'CifsHost.create.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    path => $path,
                    share_name => $share_name,
                }
            }
        );

#        NACL::STask::Volume->create(
#            command_interface  => $node,
#            vserver            => $vserver,
#            volume             => $share_name,
#            'junction-path'    => $path,
#            nacltask_if_exists => 'reuse',
#            size               => '5G',
#            aggregate          => 'cifsfit_aggr_node1_1',
#        );

        NACL::STask::VserverCifsShare->create(
        	command_interface =>$node,
            vserver => $vserver,
            path => $path,
			'share-name' =>	$share_name,
			nacltask_if_exists	=>"reuse",
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('CifsHost is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to create Cifs share on a vserver');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'CifsHost.created.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        path => $path,
                        share_name => $share_name,
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Services::CifsHost',
            routing_key => '#',
            message => {
                msg => 'CifsHost.create.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    path => $path,
                    share_name => $share_name,
                }
            }
        );
        return NATE::Result::Pass->new('Share created');
    } else {
        return NATE::Result::Fail->new('Share create failed');
    }
} # End of _create_share

#===============================================================================
# Description: Map a share
#===============================================================================
sub _map_share {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $win_client = $params->get('win_client');
    my $share = $params->get('share');
    my $user = $params->get('user');
    my $password = $params->get('password');
    my $mount_obj;
    my $dir;
    my $status = FALSE;
    my $client_obj;

    try {
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Services::CifsHost',
            routing_key => '#',
            message => {
                msg => 'CifsHost.map.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    share => $share,
                    user => $user,
                    password => $password,
                    win_client => $win_client,
                }
            }
        );
        $client_obj = NACL::C::Client->new(
            name => $win_client,
        );
        my $share_obj = NACL::STask::VserverCifsShare->new(
            command_interface => $node,
            'share-name' =>	$share,
            vserver => $vserver,
        );
        $mount_obj = NACL::STask::Client::Mount->create(
            command_interface => $client_obj,
            share => $share_obj,
            user => $user,
            password => $password,
        );
        $dir = $mount_obj->mount_point();
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /mapped/i) {
            $Log->comment('CifsHost is already mapped on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to map share on a vserver');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'CifsHost.mapped.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        share => $share,
                        user => $user,
                        password => $password,
                        win_client => $win_client,
                        dir => $dir,
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Services::CifsHost',
            routing_key => '#',
            message => {
                msg => 'CifsHost.map.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    share => $share,
                    user => $user,
                    password => $password,
                    win_client => $win_client,
                    dir => $dir,
                }
            }
        );
        return NATE::Result::Pass->new('Share mapped');
    } else {
        return NATE::Result::Fail->new('Share map failed');
    }
} # End of _map_share

#===============================================================================
# Description: Unmap a share
#===============================================================================
sub _unmap_share {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $win_client = $params->get('win_client');
    my $share = $params->get('share');
    my $status = FALSE;

    try {
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Services::CifsHost',
            routing_key => '#',
            message => {
                msg => 'CifsHost.unmap.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    share => $share,
                    win_client => $win_client,
                }
            }
        );
        my $client_obj = NACL::C::Client->new(
            name => $win_client,
        );

        my $mount = NACL::CS::Client::Mount->fetch(
            command_interface => $client_obj,
            filter => {
                share_name => $share,
            }
        );

        $status = NACL::C::Client::Mount->delete(
            command_interface => $client_obj,
            'mount_point' => $mount->{mount_point},
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('CifsHost is already unmapped on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to unmap share on a vserver');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'CifsHost.unmapped.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        share => $share,
                        win_client => $win_client,
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Services::CifsHost',
            routing_key => '#',
            message => {
                msg => 'CifsHost.unmap.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    share => $share,
                    win_client => $win_client,
                }
            }
        );
        return NATE::Result::Pass->new('Share unmapped');
    } else {
        return NATE::Result::Fail->new('Share unmap failed');
    }
} # End of _unmap_share

#===============================================================================
# Description: Delete a share
#===============================================================================
sub _delete_share {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $share_name = $params->get('share_name');
    my $status = FALSE;

    try {
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Services::CifsHost',
            routing_key => '#',
            message => {
                msg => 'CifsHost.delete.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    share_name => $share_name,
                }
            }
        );
        NACL::STask::VserverCifsShare->purge(
        	command_interface =>$node,
            vserver => $vserver,
			'share-name' =>	$share_name,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('CifsHost is already deleted on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to delete Cifs share on a vserver');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'CifsHost.deleted.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        share_name => $share_name,
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'PTEFIT::Cifs::Misfit::Services::CifsHost',
            routing_key => '#',
            message => {
                msg => 'CifsHost.delete.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                }
            }
        );

        return NATE::Result::Pass->new('Share deleted');
    } else {
        return NATE::Result::Fail->new('Share delete failed');
    }
} # End of _delete_share

#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/CifsHost>

=cut

1;

