# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Vserver File module for Misfit
## @author dl-ptefit-cifs@netapp.com
## @status Private

# Package: PTEFIT::Cifs::Misfit::Services::File
# Module to run File commands with Misfit messaging

package PTEFIT::Cifs::Misfit::Services::File;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Vserver;
use NATE::ParamSet;
use NACL::C::Client;
use NACL::C::Client::File;
use NACL::STask::Client::Mount;
use NACL::STask::VserverCifsShare;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;
use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a new File Misfit object
#
#  Returns:
#      A new File Misfit object
#
#  Example:
#
#      > my $File = PTEFIT::Cifs::Misfit::Features::File->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "PTEFIT::Cifs::Misfit::Services::File";

    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );

    my $node = undef;
    my $node_cli = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
        $node_cli = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
        $node_cli = NACL::C::Node->find();
    }

    $node->apiset_must(interface => 'ZAPI');
    $node->apiset();
    $self->{node} = $node;
    $self->{node_cli} = $node_cli;
    $self->{node_cli}->apiset_must(interface => 'CLI');
    $self->{node_cli}->apiset();
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#===============================================================================
# Description: Creates File on a Share
#===============================================================================
sub create_file_on_share {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_file_on_share";
        } else {
            $runid = $vserver->{vserver} . "_create_file_on_share";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_create_file_on_share(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_file_on_share",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    $Log->comment("Create file on share woor: $status");

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Create failed");
        return FALSE;
    }
} # End of create_file_on_share

#===============================================================================
# Description: Creates File on a Share
#===============================================================================
sub create_files_fast {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_create_files_fast";
        } else {
            $runid = $vserver->{vserver} . "_create_files_fast";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_create_files_fast(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create_files_fast",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Create Files Fast failed");
        return FALSE;
    }
} # End of create_files_fast

#===============================================================================
# Description: Creates File on a Share
#===============================================================================
sub set_acl_on_file {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_set_acl_on_file";
        } else {
            $runid = $vserver->{vserver} . "_set_acl_on_file";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_set_acl_on_file(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "set_acl_on_file",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Set ACL failed");
        return FALSE;
    }
} # End of set_acl_on_file

#===============================================================================
# Description: Deletes File on a Share
#===============================================================================
sub delete_file_on_share {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_delete_file_on_share";
        } else {
            $runid = $vserver->{vserver} . "_delete_file_on_share";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_delete_file_on_share(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "delete_file_on_share",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("File delete failed");
        return FALSE;
    }
} # End of delete_file_on_share

#===============================================================================
# Description: Read File on a Share
#===============================================================================
sub read_file_on_share {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_read_file_on_share";
        } else {
            $runid = $vserver->{vserver} . "_read_file_on_share";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_read_file_on_share(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "read_file_on_share",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("File read failed");
        return FALSE;
    }
} # End of read_file_on_share

#===============================================================================
# Description: Write to File on a Share
#===============================================================================
sub write_to_file_on_share {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_write_to_file_on_share";
        } else {
            $runid = $vserver->{vserver} . "_write_to_file_on_share";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_write_to_file_on_share(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "write_to_file_on_share",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("File write failed");
        return FALSE;
    }
} # End of write_to_file_on_share

#===============================================================================
# Description: Copy files
#===============================================================================
sub copy_files {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_copy_files";
        } else {
            $runid = $vserver->{vserver} . "_copy_files";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_copy_files(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "copy_files",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("File copy failed");
        return FALSE;
    }
} # End of copy_files

#===============================================================================
# Description: Rename files
#===============================================================================
sub rename_file {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = {$vserver->{id}} . "_rename_file";
        } else {
            $runid = $vserver->{vserver} . "_rename_file";
        }
        push(@msg_data, {vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_rename_file(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "rename_file",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("File rename failed");
        return FALSE;
    }
} # End of rename_file

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Create file on a vserver
#===============================================================================
sub _create_file_on_share {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $file = $params->get('file');
    my $win_client = $params->get('win_client');
    my $dir = $params->get('dir');
    my $expected_behaviour = $params->get('expected-behaviour', default => "pass");
    my $exception_text;
    my $status = FALSE;

    try {
        my $client = NACL::C::Client->new(
            name => $win_client);
        my $client_obj = $client->apiset(category => "Host",
            interface => "CLI",
            set => "Windows");

        my $output = $client_obj->execute_raw_command(command => "echo $file > $dir\\$file");

        my $output_status = "pass";
        if ($output =~ /denied/i) {
            $output_status = "fail";
        }
        if ($output_status eq $expected_behaviour) {
            $status = TRUE;
        }

        $Log->comment("woor $status = $output_status equals $expected_behaviour");

    } otherwise {
        my $exception = shift;
        $Log->error('Failed to create file');
        NATE::BaseException->throw($exception);
    };

    if ($status) {
       return NATE::Result::Pass->new('File create action passed');
    } else {
      NATE::BaseException->throw("File create action failed");
      return NATE::Result::Fail->new('File create action failed');
    }
} # End of _create



#===============================================================================
# Description: Set ACL on FILE
#===============================================================================
sub _set_acl_on_file {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $file = $params->get('file');
    my $host = $params->get('host');
    my $dir = $params->get('dir');
    my $dacl = $params->get("dacl");
    my $dacl_op = $params->get("dacl_op");
    my $sacl = $params->get("sacl");
    my $sacl_op = $params->get("sacl_op");
    my $Windowsuser = $params->get("user");
    my $Api_Set_Obj;
    my $host_obj;

    my $status = FALSE;

    try {
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        $host_obj = host($host);
        $Api_Set_Obj = NACL::APISet->new(
            hostobj   => $host_obj,
            category  => "Host",
            interface => "CLI",
            set       => "PowerShell",
            connid    => 'exec',
        );
            logcomment("Set ACLs - $dacl_op\[$dacl]/$sacl_op\[$sacl] on [$dir\\$file] for $Windowsuser");
            $Api_Set_Obj->execute_command( command =>  "X:\\agnostic\\lib\\CIFS\\dacls_sacls $dir\\$file $Windowsuser $dacl_op $dacl $sacl_op $sacl");

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate/i) {
            $Log->comment('File is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to turn create file on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('File acls created');
    } else {
        return NATE::Result::Fail->new('File acls failed');
    }
} # End of _set_acl

#===============================================================================
# Description: Delete file on a vserver
#===============================================================================
sub _delete_file_on_share {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $file = $params->get('file');
    my $win_client = $params->get('win_client');
    my $dir = $params->get('dir');
    my $expected_behaviour = $params->get('expected-behaviour', default => "pass");
    my $status = FALSE;
    my $exception_text;

    try {

        my $client = NACL::C::Client->new(
            name => $win_client);
        my $client_obj = $client->apiset(category => "Host",
                        interface => "CLI",
                        set => "Windows");

        my $output = $client_obj->execute_raw_command(command => "erase $dir\\$file /Q", timeout => 7200);
        my $output_status = "pass";
        if ($output =~ /find/i) {
            $exception_text = "Cannot find file";
            $output_status = "fail";
        } elsif ($output =~ /access/i) {
             $exception_text = "Cannot erase file permission denied";
             $output_status = "fail";
        }
        if ($output_status eq $expected_behaviour) {
            $status = TRUE;
        }

    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('File is already deleted on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to turn delete file');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('File delete action passed');
    } else {
        NATE::BaseException->throw("File delete action failed");
        return NATE::Result::Fail->new('File delete action failed');
    }
} # End of _delete

#===============================================================================
# Description: Read file on a vserver
#===============================================================================
sub _read_file_on_share {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $share = $params->get('share');
    my $file = $params->get('file');
    my $dir = $params->get('dir');
    my $win_client = $params->get('win_client');
    my $mount_obj = $params->get('mount');
    my $directory_name = $params->get('directory_name');
    my $expected_behaviour = $params->get('expected-behaviour', default => "pass");
    my $status = FALSE;
    my $exception_text;

    try {

        my $client = NACL::C::Client->new(
                name => $win_client);
        my $client_obj = $client->apiset(category => "Host",
                            interface => "CLI",
                            set => "Windows");

        my $output = $client_obj->execute_raw_command(command => "type $dir\\$file");
        my $output_status = "pass";
        if ($output =~ /denied/i) {
            $exception_text = "Cannot read file";
            $output_status = "fail";
        } elsif ($output =~ /find/i) {
             $exception_text = "Cannot find file";
             $output_status = "fail";
        }
        if ($output_status eq $expected_behaviour) {
            $status = TRUE;
        }

    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('File is already read on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to read file on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('File red action passed');
    } else {
      NATE::BaseException->throw("File read action failed");
      return NATE::Result::Fail->new('File read action failed');
    }
} # End of _read

#===============================================================================
# Description: Write file on a vserver
#===============================================================================
sub _write_to_file_on_share {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $file = $params->get('file');
    my $win_client = $params->get('win_client');
    my $dir = $params->get('dir');
    my $text = $params->get('text');
    my $expected_behaviour = $params->get('expected-behaviour', default => "pass");
    my $exception_text;
    my $status = FALSE;
    my $path = "";

    try {
        my $client = NACL::C::Client->new(
            name => $win_client);
        my $client_obj = $client->apiset(category => "Host",
                        interface => "CLI",
                        set => "Windows");

        my $output = $client_obj->execute_raw_command(command => "echo $text > $dir\\$file");
        my $output_status = "pass";
        if ($output =~ /denied/i) {
             $output_status = "fail";
        }
        if ($output_status eq $expected_behaviour) {
            $status = TRUE;
        }

    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('File is written.');
            $status = TRUE;
        } else {
            $Log->error('Failed to write to file');
            NATE::BaseException->throw($exception);
        }
    };

    $Log->comment("status: $status:$expected_behaviour");

    if ($status) {
        return NATE::Result::Pass->new('File write action passed');
    } else {
        NATE::BaseException->throw("File write action failed");
        return NATE::Result::Fail->new('File write action failed');
    }
} # End of _write_to_file_on_share

#===============================================================================
# Description: Copy files
#===============================================================================
sub _copy_files {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $source = $params->get('source');
    my $destination = $params->get('destination');
    my $file = $params->get('file');
    my $win_client = $params->get('win_client');
    my $dir = $params->get('dir');
    my $status = FALSE;

    try {

        my $client = NACL::C::Client->new(
                name => $win_client);
        my $client_obj = $client->apiset(category => "Host",
                interface => "CLI",
                set => "Windows");

        $client_obj ->execute_raw_command(command => "copy $source $destination", timeout => 3600);

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('File is already read on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to copy files');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('File write');
    } else {
        return NATE::Result::Fail->new('File write failed');
    }
} # End of _copy_files

#===============================================================================
# Description: Rename files
#===============================================================================
sub _rename_file {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $file = $params->get('file');
    my $client_obj = $params->get('client_obj');
    my $new_filename = $params->get('new-filename');
    my $status = FALSE;


    try {
        my $response = $client_obj->rename('old-filename' => $file,
                'new-filename' => $new_filename,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /exist/i) {
            $Log->comment('File is already renamed on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to rename file on a vserver');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('File rename');
    } else {
        return NATE::Result::Fail->new('File rename failed');
    }
} # End of _rename_file

#===============================================================================
# Description: Create file on a vserver
#===============================================================================
sub _create_files_fast {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $file = $params->get('file');
    my $win_client = $params->get('win_client');
    my $dir = $params->get('dir');
    my $num_files = $params->get('num_files');
    my $status = FALSE;

    try {
        $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'CLI');
        my $client = NACL::C::Client->new(
            name => $win_client);
        my $client_obj = $client->apiset(category => "Host",
            interface => "CLI",
            set => "Windows");

        for ($num_files; $num_files > 0;$num_files--){
            my $file_name = $file . "_" . $num_files;
            $client_obj->execute_raw_command(command => "echo $file $num_files > $dir\\$file_name", timeout => 600);
        }
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /duplicate/i) {
            $Log->comment('File is already created on this vserver.');
            $status = TRUE;
        } else {
            $Log->error('Failed to create file');
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        NATE::Result::Pass->new('Files created fast');
        return TRUE;
    } else {
        NATE::Result::Fail->new('File create fast failed');
        return FALSE;
    }
} # End of _create

#===========================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/File>

=cut

1;

