#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/ontap/main/test/lib/Packet_tracer.pm#4 $
#
# Copyright (c) 2011 Network Appliance, Inc.
# All rights reserved.
#

package Packet_tracer;

## @summary
##   Provide an OS independent routine to start, stop and validate the 
##   Packet trace files
##
## @description
##   The library "start_trace_" subroutine attempts to start the packet capture
##   based on the Client and returns the NFS info of the captured packet
##   For Mount options,  gerp the packet with the MOUNT name and 
##   verify for the reply call from the server
##   For NFS options, grep the file name on which the IO is done and
##   verify the reply call from the server
##
##   The library "start_trace_" subroutine starts the packet trace on the clients
##   Params : Client object on which the Packet capture needs to be started
##   reports error is Packet capture is not started
##
##   The library "stop_trace_" subroutine stops the packet trace on the client
##   Params : The client object and the pid info for AIX
##   
##   The library "Check_file_trace_" subroutine Checks for the mount call or 
##   the lookup call validates the reply call from the server
##   Params : The packet trace file to validate, host_type to load the modules
##            for the packet trace, pattern so search MOUNT PATH or the file name
##            return the NFS options for the packet that matches with the pattern 

use strict;
use Params::Validate qw(validate OBJECT SCALAR UNDEF ARRAYREF);
use Exporter qw(import);
use vars qw( @ISA );
use Carp qw(croak);
use Net::TcpDumpLog;
use Net::SnoopLog;

our @EXPORT_OK = qw(&start_trace_
					&Check_file_trace_
                    &stop_trace_);
					
our %protocol = (
			'17' => "udp",
			'6'  => "tcp",
);					

our %packet_info;

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};


################################################################################
## 	Function prototypes
################################################################################

sub start_trace_(%);
sub stop_trace_(%);
sub Check_file_trace_(%);
sub __num_conversion($$$$);

sub start_trace_(%) {

	my %opts = @_;
	
	# Check for correct connection object
    croak "Not a 'Connectrec' object"
      if ! UNIVERSAL::isa($opts{conn}, "Connectrec");

    # Determine the right platform
    # and inherit correct OS dependent class.
    my $os_type = $opts{conn}->hostp()->hosttype();
	
	my $cmd_response;
		
    if ( $os_type =~ /unix-solaris/i ) {
	
		$cmd_response = $opts{conn}->execute(command => "/usr/sbin/snoop $opts{command} &");
        
    } elsif ($os_type =~ /unix-linux/i ) {
	
		$cmd_response = $opts{conn}->execute(command => "/usr/sbin/tcpdump $opts{command} &");
		
	}  elsif ($os_type =~ /unix-aix/i ) {
	
		$cmd_response = $opts{conn}->execute(command => "/usr/sbin/tcpdump $opts{command} &");
		
	}	elsif ($os_type =~ /unix-hpux/i ) {
	
		$cmd_response = $opts{conn}->execute(command => "/usr/local/bin/tcpdump $opts{command} &");
	
	}
	
	if($cmd_response =~ /no suitable device found|No such device exists/i)	{
	
		croak "Failed to start Packet Capture";
	
	}
	
	return $cmd_response;
	
}

# Subroutine : stop_trace_
# Objective  : To stop the packet trace on the client
#
# Input  : Client object
#        : Command to be executed to stop the packet trace
#
# Output : Returns the command response


sub stop_trace_(%)	{

	my %opts = @_;
	
	# Check for correct connection object
    croak "Not a 'Connectrec' object"
      if ! UNIVERSAL::isa($opts{conn}, "Connectrec");

    # Determine the right platform
    # and inherit correct OS dependent class.
    my $os_type = $opts{conn}->hostp()->hosttype();
	
	if($opts{pid})	{  # If the pid of the process to kill is mentioned, kill it without os info
	
		return $opts{conn}->execute(command => "kill -15 $opts{pid}");
	
	}
		
    if ( $os_type =~ /unix-solaris/i ) { 
	
		return $opts{conn}->execute(command => 'pkill -9 /usr/sbin/snoop');
		
    } elsif ($os_type =~ /unix-linux/i ) {
	
		return $opts{conn}->execute(command => 'killall -9 /usr/sbin/tcpdump');
		
	} elsif ($os_type =~ /unix-aix/i ) {
	
		croak "Aix process id to be provided to kill the process";
		
	}	elsif ($os_type =~ /unix-hpux/i ) {
	
		return $opts{conn}->execute(command => 'pkill -9 /usr/local/bin/tcpdump');
		
	}

}

# Subroutine : Check_file_trace_
# Objective  : To Check the pattern in the packet trace file and return the NFS params
#
# Input  : Trace file path
#        : The host type to invoke the modules to read the trace file
#        : Pattern to search in the trace file
#
# Output : Returns NFS params in the captured packet


sub Check_file_trace_(%)	{

	my %opts = @_;
	
	my $file = $opts{path}; # get the path and packet trace file name
	
	my ($log, $reply_xid, $reply_response, $reply_pkt_chk);

	if($opts{host_type} =~ /linux|aix|hpux/)	{ # load the Tcpdump modules if the client is linux

		$log = Net::TcpDumpLog->new();
	
	} elsif($opts{host_type} =~ /solaris/)	{ # load the snoop modules if the client is solaris
	
		$log = Net::SnoopLog->new();
		
	}
	
	$log->read("$file"); # read the packet trace file

	my @Indexes = $log->indexes; # get each packet index
	
	foreach my $index (@Indexes) { # Check in the each packet captured

        my $data = $log->data($index);
		
		my ($rpc_version, $program, $NFS_Version, $procedure_call, $mount_port, $offset);
		
		$packet_info{'Protocol'} = "";	# to load the default params
		$packet_info{'NFS_Version'} = "";
		$packet_info{'program'} = "";
		$packet_info{'procedure_call'} = "";
		$packet_info{'mount_port'} = "";
		$packet_info{'rpc_version'} = "";
		$packet_info{'reply_state'} = "False";
		
		my ($ether_dest,$ether_src,$ether_type,$ether_data) = unpack('H12H12H4a*',$data); # unpack the ethernet info
		
		my $i=1;
		
		my @nfs_info = qw();
		
		my @rpv_version = qw ();
		
		my	@data = split('',$ether_data);  # Convert the data each char into text
		
			foreach(@data)	{
			
				if($i > 80) {
				
					my $temp = chr(ord($_));

					push(@nfs_info,"$temp");

				}

				$i++;

			}
				
		my $nfs_info = join('',@nfs_info);
		
		if($nfs_info =~ /$opts{pattern}/)	{ # sreach for the pattern in the converted text
		
			my $file_name = $&;
		
			my $IPV_VERSION = ord($data[00]); # check if the paket captured is IPV4/IPV6
			
			my $IPV6_OFFSET = 0;
			
			my $IPV4_OPTIONAL_BYTES = ord($data[00]);  # Check for the IP headre size for IPV4
			
			$IPV_VERSION = (($IPV_VERSION & 0xF0) >> 4);
			
			$IPV4_OPTIONAL_BYTES = ((hex($IPV4_OPTIONAL_BYTES & 0x0F)) * 4) - 20;
			
			my $protocol = $data[9];  # Get the protocol info
			
			if($IPV_VERSION == '6')	{
			
				$protocol = $data[6];
				
				$IPV6_OFFSET = 20;
				
				$IPV4_OPTIONAL_BYTES = 0;
			
			}
			
			if($protocol{ord($protocol)} =~ /udp/)	{  # Calculate the offset if the proto is udp
			
				$offset = 36 + $IPV6_OFFSET + $IPV4_OPTIONAL_BYTES;
					
			} elsif ($protocol{ord($protocol)} =~ /tcp/)	{  # calculate the offset if the proto is tcp
				
				my $tcp_optional_bytes = ord($data[32 + $IPV6_OFFSET + $IPV4_OPTIONAL_BYTES]);
				
				$tcp_optional_bytes = ((($tcp_optional_bytes & 0xF0) >> 4) * 4) - 20;
				
				$offset = 52 + $IPV6_OFFSET + $tcp_optional_bytes + $IPV4_OPTIONAL_BYTES;
					
			}
			
			# Check if the is call or reply packet

			my $xid = __num_conversion(ord($data[$offset-8]), ord($data[$offset-7]), ord($data[$offset-6]), ord($data[$offset-5]));
			
			my $packet_type = __num_conversion(ord($data[$offset-4]), ord($data[$offset-3]), ord($data[$offset-2]), ord($data[$offset-1]));
			
			if($packet_type)	{
			
				next;   # skip if the packet type is not a call packet
			
			}
			
			$mount_port = __num_conversion('00', '00', ord($data[22 + $IPV6_OFFSET + $IPV4_OPTIONAL_BYTES]), ord($data[23 + $IPV6_OFFSET + $IPV4_OPTIONAL_BYTES]));
			
			$rpc_version = __num_conversion(ord($data[$offset+0]), ord($data[$offset+1]), ord($data[$offset+2]), ord($data[$offset+3]));
			
			$program = __num_conversion(ord($data[$offset+4]),ord($data[$offset+5]),ord($data[$offset+6]),ord($data[$offset+7]));
			
			$NFS_Version = __num_conversion(ord($data[$offset+8]),ord($data[$offset+9]),ord($data[$offset+10]),ord($data[$offset+11]));
				
			$procedure_call = __num_conversion(ord($data[$offset+12]),ord($data[$offset+13]),ord($data[$offset+14]),ord($data[$offset+15]));
			
			$packet_info{'Protocol'} = $protocol{ord($protocol)};
		
			$packet_info{'rpc_version'} = $rpc_version;
			
			$packet_info{'program'} = $program;
			
			$packet_info{'NFS_Version'} = $NFS_Version;
			
			$packet_info{'procedure'} = $procedure_call;
			
			$packet_info{'mount_port'} = $mount_port;

                        #print " \n      $xid";
			
			# The Next Packet after the call is the reply from the server ========================================================

			do	{
			
			$index++; # Increment the packet count by one to check the reply packet

			$data = $log->data($index);
			
			$i=0;

			($ether_dest,$ether_src,$ether_type,$ether_data) = unpack('H12H12H4a*',$data); # 

			@data = split('',$ether_data);

			foreach (@data)	{
	
				if($i > 80) {

					my $temp = chr(ord($_));

					push(@nfs_info,"$temp");

				}

			$i++;

			}

			$nfs_info = join('',@nfs_info);

			$reply_xid = __num_conversion(ord($data[$offset-8]), ord($data[$offset-7]), ord($data[$offset-6]), ord($data[$offset-5]));

                        #print " \n      $reply_xid";

			$reply_pkt_chk = __num_conversion(ord($data[$offset-4]), ord($data[$offset-3]), ord($data[$offset-2]), ord($data[$offset-1]));

			$reply_response =  __num_conversion(ord($data[$offset+13]), ord($data[$offset+14]), ord($data[$offset+15]), ord($data[$offset+16]));

			} while($xid ne $reply_xid); #Check for the xid reply packet

			#print "	\n 	$reply_response";
			
			if(($reply_response == 0) and ($reply_pkt_chk == 1))	{ # check if the packet is reply packet and NFS response is ok
			
				$packet_info{'reply_state'} = "True"; # Update to True if the reply call is sucessfull
			
			}

			return (%packet_info);
		
		}

	}

}

sub __num_conversion($$$$)	{

			my @nums;
			
			my @hex_values = @_;
			
			foreach	(@hex_values)	{
			
				my $temp = sprintf("%x", $_);
				
				my $dec = hex($temp);
				
				if(($dec >= 0) and ($dec < 16))	{
				
					$temp = "0"."$temp";
				
				}
				
				push(@nums,$temp);
			
			}
			
			my $program = join('',@nums);
			
			return hex($program);

}

1;
