# Perl Module
#
# Purpose:  Provide tools/utils to work with NetApp panics.
#


package Panic_Utils;

use strict;
use File::Basename;
use Data::Dumper;
use Time::Local;

use Data::Dumper;
use Digest::MD5;




BEGIN {
   use Exporter();
   our @ISA           = qw(Exporter);
   our $VERSION       = '0.0.2';
}

our @BackTrace_tags = qw(
   parse_gdb_backtrace_file
   parse_gdb_backtrace_list
   parse_backtrace_from_consol
   parse_consol_backtrace_list
   consol_backtrace_index2epoch
   compare_backtrace
   consol_backtrace_panic_str
   make_bt_signature
);


our @EXPORT     = ( @BackTrace_tags);
our @EXPORT_OK  = qw();
 

our %EXPORT_TAGS = (
   'backtrace'  => [@BackTrace_tags],
);

our $ERROR;
our $DEBUG;
our $CONSOL_RUNAWAY = 48;


our %PRIORITY = (
   'index'        => 10,   # gdb
   'addr'         => 1,    # gdb
   'func'         => 10,   # gdb, console
   'vars'         => 1,    # gdb, console
   'file'         => 5,    # gdb
   'fname'        => 10,   # gdb, console
   'fpath'        => 10,   # gdb
   'linum'        => 2,    # gdb
   'msg'          => 10,   # gdb, console
   'offset'       => 2,    # console
   'ver'          => 5,    # console     from version:
   'fver'         => 5,    # console     from version line with full path
   'vpath'        => 5,    # console     full version path
   'panic_str'    => 10,   # console
   'PANIC_str'    => 10,   # console
   'uptime'       => 0,    # console
   'compile_flag' => 0,    # console
   '_signature'   => 0,    # gdb, console
   '_syntax'      => 0,    # gdb, console
);


# can be any number, treated as a multiplier
our %WEIGHT = (
   'index'        => 100,
   'addr'         => 1,
   'func'         => 100,
   'vars'         => 1,
   'file'         => 80,
   'fname'        => 100,
   'fpath'        => 10,
   'linum'        => 10,
   'msg'          => 100,
   'offset'       => 10,
   'ver'          => 50,
   'fver'         => 5,
   'vpath'        => 5,
   'panic_str'    => 100,
   'PANIC_str'    => 100,
   'uptime'       => 1,
   'compile_flag' => 1,
   '_signature'   => 1,
   '_syntax'      => 1,
);


# day of week translator
our %DoW = (
   Sun => '0',   
   Mon => '1',  
   Tue => '2',
   Wed => '3',
   Thu => '4',
   Fri => '5',
   Sat => '6'
);


our %MONTH_MAP = (
   Jan => '0',    Feb => '1',    Mar => '2',      Apr => '3',
   May => '4',    Jun => '5',    Jul => '6',      Aug => 7,
   Sep => '8',    Oct => '9',    Nov => '10',     Dec => '11'
);

# hex value for bit mask for compare mode
# 0000 0x00 - not used
# 0001 0x01 - line to line
# 0010 0x02 - line to all line
#
our $COMPARE_MODE = 0x1;





#-----------------------------------------------------------------------------
#
# Parse GDB backtrace list
# Args
#    $_[0] = ref to array of bactrace lines
#    $_[1] = aref to populate 
#
# Return
#   wantarray ? (status, error) ; status
# 
# AoH structure 
#   - each Array element represents on parsed line of a backtrace
#   - the Hash in each element represents each field of a panic bactrace line
#
# @AoH = (  {  'index'   => <int>,        the backtrace index at begin of line
#              'addr'    => <hex>,        the hex address right after the index
#              'func'    => <string>,     the function name
#              'vars'    => <string>,     the content of () right after function name
#              'file'    => <string>,     the full filename referenced in backtrace
#              'fname'   => <string>,     the file name, separated from path
#              'fpath'   => <string>,     the path to fill name
#              'linum',  => <int>,        the line number given for given file
#              'msg'     => <string>      string contained within < >, follows #index on backtrace line
#              '_syntax' => <string>      syntax format
#           },
#           { ... },
#           { ... },
#  )
# 
sub parse_gdb_backtrace_list {

   my ($_line,
       $_index, $_addr, $_func, $_vars, $_file, $_linum,
       $_fname, $_fpath, $_fsuffix,
       $_msg,
       $_parsed, $_syntax, $_h,
   );
   $ERROR = '';
   my $_PANIC_str     = '';
   my $_ver           = '';
   my $_vpath         = '';
   my $_fver          = '';
   my $_fsuffix       = '';
   my $_compile_flags = '';
   my @_tokens        = ();
   my $_signature     = '';

   # open file
   foreach $_line  ( @{$_[0]} ) {
      chomp($_line);
      $_line =~ s/\015//g;
      $_parsed = 0;
      $_syntax = '';
      # init the values so nothing carries to next line
      ($_index,$_addr,$_func,$_vars,$_file,$_fname,$_fpath,$_fsuffix,$_linum,$_msg) = ();
      # full backtrace line
      #               idx     0xAddr              func      vars          File       linum   
      if ($_line =~ /#(\d+)\s+([0-9a-fx]*)\s+in\s+(\S+)\s+\((.*)\)\sat\s+([\w\.\/]+):(\d+)$/) {
         $_index = $1;         $_addr  = $2;         $_func  = $3;
         $_vars  = $4;         $_file  = $5;         $_linum = $6;
         # get parts of the name
         ($_fname, $_fpath, $_fsuffix) = &fileparse($_file);   
         $_parsed++;
         $_syntax ='iafvfl';
         # push onto array
         push( @{$_[1]},  { 'index'   => $_index,
                             'addr'    => $_addr,
                             'func'    => $_func,
                             'vars'    => $_vars,
                             'file'    => $_file,
                             'fname'   => $_fname,
                             'fpath'   => $_fpath,
                             'linum'   => $_linum,
                             'msg'     => '',
                             '_syntax' => $_syntax,
                             '_type'   => 'gdb',
                          },
         ); 
      }
      # no addr field in backtrace line
      #                  idx     func      vars          File       linum
      elsif ($_line =~ /#(\d+)\s+(\S+)\s+\((.*)\)\sat\s+([\w\.\/]+):(\d+)$/) {
         $_index = $1;         $_func  = $2;         $_vars  = $3;
         $_file  = $4;         $_linum = $5;
         # get parts of the name
         ($_fname, $_fpath, $_fsuffix) = &fileparse($_file);

         $_parsed++;
         $_syntax ='ifvfl';
         # push onto array
         push( @{$_[1]},  { 'index'   => $_index,
                             'addr'    => '',
                             'func'    => $_func,
                             'vars'    => $_vars,
                             'file'    => $_file,
                             'fname'   => $_fname,
                             'fpath'   => $_fpath,
                             'linum'   => $_linum,
                             'msg'     => '',
                             '_syntax' => $_syntax,
                            '_type'   => 'gdb',
                          },
         );
      }
      # no file info
      #                  idx     0xAddr              func      vars   
      elsif ($_line =~ /#(\d+)\s+([0-9a-fx]*)\s+in\s+(\S+)\s+\((.*)\)$/) {
         $_index = $1;         $_addr  = $2;         $_func  = $3;
         $_vars  = $4;         $_file  = $5;         $_linum = $6;
         $_parsed++;
         $_syntax ='iafv';
         # push onto array
         push( @{$_[1]},  { 'index'   => $_index,
                             'addr'    => $_addr,
                             'func'    => $_func,
                             'vars'    => $_vars,
                             'file'    => '',
                             'fname'   => '',
                             'fpath'   => '',
                             'linum'   => '',
                             'msg'     => '',
                             '_syntax' => $_syntax,
                            '_type'   => 'gdb',
                          },
         );
      }
      # no func name and no file info
      # func name is '??'
      #                  idx     0xAddr              func      vars   
      elsif ($_line =~ /#(\d+)\s+([0-9a-fx]*)\s+in\s+(\?+)\s+\((.*)\)$/) {
         $_index = $1;         $_addr  = $2;         $_func  = $3;
         $_vars  = $4;         $_file  = $5;         $_linum = $6;
         $_parsed++;
         $_syntax ='iaf?v';
         # push onto array
         push( @{$_[1]},  { 'index'   => $_index,
                             'addr'    => $_addr,
                             'func'    => $_func,
                             'vars'    => $_vars,
                             'file'    => '',
                             'fname'   => '',
                             'fpath'   => '',
                             'linum'   => '',
                             'msg'     => '',
                             '_syntax' => $_syntax,
                            '_type'   => 'gdb',
                          },
         );
      }

      # totally different line, i'm calling it message 
      # it has index number and text enclosed in <>
      #                  idx       msg
      elsif ($_line =~ /#(\d+)\s+<(.+)>$/) {
         $_index = $1;
         $_msg   = $2;
         $_parsed++;
         $_syntax ='im';
         # push onto array
         push( @{$_[1]},  { 'index'   => $_index,
                             'addr'    => '',
                             'func'    => '',
                             'vars'    => '',
                             'file'    => '',
                             'fname'   => '',
                             'fpath'   => '',
                             'linum'   => '',
                             'msg'     => $_msg,
                             '_syntax' => $_syntax,
                            '_type'   => 'gdb',
                          },
         );
      } 
      # PANIC string
      elsif ($_line =~ /PANIC:\s+(.+)/) { 
         $_PANIC_str = $1;
         printf("   found PANIC string\n") if $DEBUG;
      }
      # compile flags
      elsif ($_line =~ /cc_flags=(.+)/ ) {
         $_compile_flags = $1;
         printf("   found compile flags [%s]\n", $_compile_flags) if $DEBUG; 
      }
      # version string
      elsif ($_line =~ /version=(.+)/ ) {
         $_line =~ s/://g;
         @_tokens = split(/\s+/, $_line);
         # version given with full path
         if ($_tokens[2] =~ /\x2f/) {
            $_tokens[2] =~ s/\x2f$//;
            ($_fver, $_vpath, $_fsuffix) = &fileparse($_tokens[2]);
            printf("      ver:  [%s]  fullpath: [%s]\n", $_fver, $_vpath) if $DEBUG;
         }
         # simplified version string
         else   {
            $_ver = $_tokens[2];
         }
         printf("   found version [%s]  path [%s]\n", $_fver || $_ver, $_vpath) if $DEBUG;
      }
      # if we did not parse anything on a line that starts with '#'
      else {
         if ($_line =~ /^#/) {
            printf("   !!! unknown gdb pattern: %s\n", $_line);
         }
      }      

      if ($DEBUG) {
         if ($_parsed) {
            printf("parsed %-3u syntax %6s ::>  %-3u %10s  %-25s  %6u %-45s ", 
               $., $_syntax, 
               $_index, $_addr, $_func || $_msg, 
               $_linum, $_file || '-',
            );
            printf("%u ", scalar(@{$_[1]}));
            printf("\n");
         }
      }
   }
   #print "bt:\n" . Dumper($_[1]);

   # push info string on last element, non-parse values start with '_'
   push( @{$_[1]},  {  'fver'         => $_fver,
                       'vpath'        => $_vpath,
                       'ver'          => $_ver,
                       'PANIC_str'    => $_PANIC_str,
                       'compile_flag' => $_compile_flags,
                       '_syntax'      => 'info',
                       '_type'        => 'gdb',
                       '_signature'   => 'none',
                    },
   );
   # get signature and put in last hash
   $_signature = &make_bt_signature($_[1]);
   $_[1]->[-1]{'_signature'} = $_signature;



   $_h = scalar(@{$_[1]});

   return(wantarray ? ($_h, undef) : $_h); 
}
#
#.............................................................................
# 
# Parse console backtrace from list of lines user passed in
# Expecting the start of backtrace to be console line such as:
#    [Ddd Mmm  d hh:mm:ss]panic;
#
# Expect end of backtrace to be console line such as:
#    [Ddd Mmm  d hh:mm:ss]backtrace complete
#    [Ddd Mmm  d hh:mm:ss]DUMPCORE:
#
# Version and panic string info are parsed from lines expected to be as:
#    [Ddd Mmm  d hh:mm:ss]panic:
#    [Ddd Mmm  d hh:mm:ss]version:
#
# Args
#    $_[0] = ref to array of backtrace lines
#    $_[1] = aref to populate 
#
# Return
#   wantarray ? (status, error) ; status
# 
# AoH structure  
#   - each Array element represents on parsed line of a backtrace
#   - the Hash in each element represents each field of a panic bactrace line
#
# @AoH = (  {  'func'    => <string>,     the function name
#              'vars'    => <string>,     the content of () right after function name
#              'fname'   => <string>,     the function name after the 'at'
#              'offset', => <hex>,        the line number given for given file
#              'msg'     => <string>      string contained within < >, follows #index on backtrace line
#              '_syntax' => <string>      syntax format
#           },
#           { ... },
#           { ... },
#  )
 
sub parse_consol_backtrace_list {

   my ($_line,  
       @_tokens, $_fsuffix,
       $_func, $_vars, $_file, $_offset,
       $_syntax, $_P, $_signature,
   );
   my $_flag = 0;
   my $_panic_str     = '';     # panic string on line with panic:
   my $_PANIC_str     = '';     # panic string on line with PANIC:
   my $_ver           = '';     # version
   my $_vpath         = '';     # version path
   my $_fver          = '';     # version from full version path
   my $_uptime        = '';     # uptime line
   my $_compile_flags = '';     

   printf("   parse_consol_backtrace_list element\n") if $DEBUG;
   foreach $_line  ( @{$_[0]} ) {
      ($_func, $_vars, $_file, $_offset) = ();
      # clean up line
      $_line =~ s/^\[\w\w\w.+:\d\d \d\d\d\d\]//;   # timestamp
      $_line =~ s/^\.+//;                          # preceeding '.'
      $_line =~ s/\015//g;                         # ^M
      chomp($_line);
 
      #printf("   line: %s\n", $_line) if $DEBUG;

      next if $_line =~ /^cpuid/;
      next if $_line =~ /^new watchdog thresholds/; 
      next if $_line =~ /^$/;
      next if $_line =~ /^\s+$/;

      # start of panic trace
      if ( $_line =~ /panic:\s(.+)$/ ) {
         $_panic_str = $1;
         $_flag      = 1;
         printf("   found panic_str\n") if $DEBUG;
         next;
      }
      # version info
      if ( $_line =~ /version:/ ) {
         printf("   found version\n") if $DEBUG;
         $_line =~ s/://g;
         @_tokens = split(/\s+/, $_line);
         printf ("      ver tokens:>> %s\n", join('::', @_tokens)) if $DEBUG;
         # version given with full path
         if ($_tokens[3] =~ /\x2f/) {
            $_tokens[3] =~ s/\x2f$//;
            ($_fver, $_vpath, $_fsuffix) = &fileparse($_tokens[3]);
            printf("      ver:  [%s]  fullpath: [%s]\n", $_fver, $_vpath) if $DEBUG; 
         }
         # simplified version string
         else   {
            $_ver = $_tokens[3];
         }
         printf("   found version [%s]  path [%s]\n", $_ver, $_vpath) if $DEBUG;
         next;
      }
      # start of backtrace
      if ( $_line =~ /stack backtrace:/) { 
         $_flag = 1;
         printf("   found stack backtrace\n") if $DEBUG;
         next;
      }
      # end of panic
      if ($_line =~ /backtrace\scomplete/ ||  $_line =~ /DUMPCORE:/ ) {
         $_flag = 0;
         printf("   found backtrace end or DUMPCORE\n") if $DEBUG;
         next;
      }

      # uptime and end of panic
      if ($_line =~ /Uptime:\s+(.+)/) {
         $_uptime = $1;
         $_flag = 0;
         printf("   found uptime\n") if $DEBUG;
      }
      # PANIC string
      if ($_line =~ /PANIC:\s+(.+)/) {
         $_PANIC_str = $1;
         $_flag = 0;
         printf("   found PANIC string\n") if $DEBUG;
      }
      # compile flags
      if ($_line =~ /compile\sflags:\s+(.+)/) {
         $_compile_flags= $1;
         $_flag = 0;
         printf("   found compile flags\n") if $DEBUG;
      }


      if ($_flag ) { 
         # func,(vars),func,offset
         if ( $_line =~ /(.+)\((.*)\)\sat\s(.+?)\+0x([0-9a-f]+)/ ) {
            $_func   = $1;
            $_vars   = $2;
            $_file   = $3;
            $_offset = $4; 
            $_syntax = 'fvfo';
   
            push( @{$_[1]},  { 'func'    => $_func,
                               'msg'     => '',
                               'vars'    => $_vars,
                               'fname'   => $_file,
                               'offset'  => $_offset,
                               '_syntax' => $_syntax,
                               '_type'   => 'consol',
                               
                             },
            );
            printf("   found %-6s %-40s  [%s]  %s\n", $_syntax, $_func, $_vars, $_offset) if $DEBUG; 
         }
         # func,(vars),func
         elsif ( $_line =~ /(.+)\((.*)\)\sat\s(.+)/ ) {
            $_func   = $1;
            $_vars   = $2;
            $_file   = $3;
            $_syntax = 'fvf';
            push( @{$_[1]},  { 'func'    => $_func,
                               'msg'     => '',
                               'vars'    => $_vars,
                               'fname'   => $_file,
                               'offset'  => $_offset,
                               '_syntax' => $_syntax,
                               '_type'   => 'consol',
                             },
            );
            printf("   found %-6s %-40s  [%s] \n", $_syntax, $_func, $_vars, $_offset) if $DEBUG;
         }
         # -- trap vars ---
         elsif ( $_line =~ /\-{3}\strap\s0x([0-9a-f]+),\s(.+)\s\-{3}/ )  {
            $_offset = $1;
            $_vars   = $2;
            $_syntax = 'trap';
            push( @{$_[1]},  { 'func'    => '',
                               'msg'     => 'trap',
                               'vars'    => $_vars,
                               'fname'   => '',
                               'offset'  => $_offset,
                               '_syntax' => $_syntax,
                               '_type'   => 'consol',
                             },
            ); 
            printf("   found %-6s %-40s  [%s]  %s\n", $_syntax, '--- trap ---', $_vars, $_offset) if $DEBUG;
         }
         else {
            printf("   UNKNOWN consol backtrace syntax: [%s]\n", $_line) if $DEBUG;
         }
      }
   }

   # push info string on last element, non-parse values start with '_'
   push( @{$_[1]},  {  'fver'         => $_fver,
                       'vpath'        => $_vpath,
                       'ver'          => $_ver,
                       'panic_str'    => $_panic_str,
                       'PANIC_str'    => $_PANIC_str,
                       'uptime'       => $_uptime,
                       'compile_flag' => $_compile_flags,                       
                       '_syntax'      => 'info',
                       '_type'        => 'consol',
                       '_signature'   => 'none',
                    },
   );
   # get signature and put in last hash
   $_signature = &make_bt_signature($_[1]);
   $_[1]->[-1]{'_signature'} = $_signature;

   printf("   store info: [%s]  %s\n", $_ver, $_signature) if $DEBUG;

   my $_h = scalar(@{$_[1]});
   return(wantarray ? ($_h,  undef) : $_h  );
}
#
#.............................................................................
# 
# Return panic strings
# console back traces have two possible panic strings
# these are denoted by the syntax  
# given a ref to panic token list, return both string
#
# $_[0] = ref to backtrace 
#
# return both strings
#
sub consol_backtrace_panic_str {

    my $_h;

   #printf("con_pstr: %s\n", Dumper($_[0]));   
 
    foreach $_h ( reverse @{$_[0]} ) {
       if ( $_h->{'_syntax'} eq 'info' ) {
          return( $_h->{'PANIC_str'}, $_h->{'panic_str'} );
       }
    } 
    return(undef, undef);
}



#
#.............................................................................
#
# compare backtrace fields
#
# $_[0] - backtrace 1 aref
# $_[1] - backtrace 2 aref to compare to

sub compare_backtrace {

   my $_e   = 0;
   my $_e2  = 0;
   my $_lbl = 0;   # line by line result
   my ($_bt, $_bt2, $_p, $_k, $_chk_this,);
   my $_max_sum     = 0;
   my $_match_sum   = 0;
   my $_max_sum_2   = 0;
   my $_match_sum_2 = 0;
   my $_same        = 0;
   my %_cache       = ();
   my @_priority    = sort keys %PRIORITY;

   # determine backtrace type gdb or consol
   my $_type_1 = $_[0]->[0]->{'_type'};
   my $_type_2 = $_[1]->[0]->{'_type'};

   # set flag to indicate if both are same type
   if ($_type_1 eq $_type_2) { $_same = 1; }

   # backtraces are of same type
   if ($_same) {
      printf("   compare same type bt: %s %s\n", $_type_1, $_type_2) if $DEBUG;
      # $COMPARE_MODE = 0001 0x01  line-to-line
      if ($COMPARE_MODE & 0x01) {
         $_e = 0;
         printf("   compare mode line-to-line bit flag set\n") if $DEBUG;
         # check every line in bt-1 to its respective line in bt-2
         foreach $_bt ( @{$_[0]} ) {
            foreach $_p ( @_priority ) { 
               # check if we have this set to be checked 
               if ($PRIORITY{$_p}) {
                  # check if this field got populated in the parser, no use comparing if empty
                  # plus we have different fields between the different types (gdb|consol)
                  printf("   cmp line: %2u  field: %-12s ", $_e, $_p,) if $DEBUG;
                  # field is defined in bt1
                  if (&_my_defined($_bt->{$_p})) {
                     $_max_sum = $_max_sum + ($PRIORITY{$_p} * $WEIGHT{$_p});
                     printf("incr %4u ", $PRIORITY{$_p} * $WEIGHT{$_p}) if $DEBUG; 
                     # compare respective field for the index
                     if ($_bt->{$_p} eq $_[1]->[$_e]->{$_p}) {
                        $_match_sum = $_match_sum + $PRIORITY{$_p} * $WEIGHT{$_p};
                        printf("same ", ) if $DEBUG;
                     }
                     else {
                        printf("diff ", ) if $DEBUG;
                     }
                  }
                  # field not defined in bt1
                  else {
                     printf("incr %4u ", 0) if $DEBUG;
                     printf("---- ") if $DEBUG;
                  }
                  printf("%6u  %6u ", $_max_sum, $_match_sum) if $DEBUG;
                  printf(" %-35s %-35s \n", $_bt->{$_p}, $_[1]->[$_e]->{$_p}) if $DEBUG;
               }
            }
            $_e++;
         }
         printf("   line-to-line: %8u  %8u  %.4f\n", 
            $_max_sum, $_match_sum,  ($_match_sum/$_max_sum)*100 
         ) if $DEBUG;
      }

      #
      #
      # $COMPARE_MODE = 0001 0x01  line-to-all-line
      if ($COMPARE_MODE & 0x02) {
         $_e = 0;
         %_cache = ();
         printf("   compare mode line-to-all-line bit flag set\n") if $DEBUG;
         foreach $_bt ( @{$_[0]} ) {
            $_chk_this = '';
            # see if we type we have
            if    (&_my_defined($_bt->{'func'})) { $_chk_this = 'func'; }
            elsif (&_my_defined($_bt->{'msg'}))  { $_chk_this = 'msg'; }
            elsif ($_bt->{'_syntax'} eq 'info')  { $_chk_this = '_syntax'; }
            else                                 { $_chk_this = ''; }
   
            printf("   check bt1 entry: %2u %4s %-40s ", $_e, $_chk_this, $_bt->{'func'} || $_bt->{'msg'} || 'info') if $DEBUG;
            # skip when func = ??
            next if $_bt->{'func'} =~ /^\?+$/; 
            # increment max_sum for this line
            foreach $_p (@_priority) {
               if ($PRIORITY{$_p}) {
                  if (&_my_defined($_bt->{$_p})) { 
                     $_max_sum_2 = $_max_sum_2 + ($PRIORITY{$_p} * $WEIGHT{$_p});
                  }
               }
            }       
            printf("max sum %6u\n", $_max_sum_2) if $DEBUG; 
            # check all bt-2 lines, if func or msg is same then compare the rest
            $_e2 = 0;
            foreach $_bt2 (@{$_[1]} ) {
               if ($_cache{$_e2}) {
                  printf("      bt2 entry %u matched previously\n", $_e2) if $DEBUG;
               }
               else { 
                  printf("      bt2 entry %u\n", $_e2) if $DEBUG;
                  if ($_chk_this eq 'func' || $_chk_this eq 'msg' || $_chk_this eq 'info') {
                     if ($_bt->{$_chk_this} eq $_bt2->{$_chk_this}) {
                        printf("         cmp %s %s\n", $_chk_this, $_bt->{$_chk_this}) if $DEBUG;
                        foreach $_p (@_priority) {
                           if ($PRIORITY{$_p}) {
                              if (&_my_defined($_bt->{$_p}) && $_bt->{$_p} eq $_bt2->{$_p}) {
                                 $_match_sum_2  = $_match_sum_2 + ($PRIORITY{$_p} * $WEIGHT{$_p});
                                 $_cache{$_e2}++;
                                 printf("         bt1 %2u  bt2 %2u  incr %6u %6u  matched: %s \n", 
                                    $_e, $_e2,  
                                    $PRIORITY{$_p} * $WEIGHT{$_p}, 
                                    $_match_sum_2, $_p
                                 ) if $DEBUG;
                              }
                           }
                        }
                     }
                  }
               }
               $_e2++;
               printf("   line-to-all-line: %8u  %8u  %.4f\n", 
                  $_max_sum_2, $_match_sum_2,  ($_match_sum_2/$_max_sum_2)*100 
               ) if $DEBUG;
            }
            $_e++;
         }
      }
   }
   # different type
   # for now we are only going to compare the func|msg
   else {
      printf("   compare diff type bt: bt1: %s  bt2: %s\n", $_type_1, $_type_2) if $DEBUG;
      %_cache = ();
      $_e = 0;
      foreach $_bt ( @{$_[0]} ) {
         $_chk_this = '';

         # check what type of line this is
         if    (&_my_defined($_bt->{'func'})) { $_chk_this = 'func';}
         elsif (&_my_defined($_bt->{'msg'}))  { $_chk_this = 'msg';}
         elsif ($_bt->{'_syntax'} eq 'info')  { $_chk_this = 'info'; }
         else                                 { $_chk_this = ''; }

         # pesky string of '?'s
         next if $_bt->{'func'} =~ /^\?+$/;

         # calc our max sum for this bt1 line
         # since bt and bt2 are diff types, only comparing definite common fields
         if ($_chk_this eq 'func') { $_max_sum_2 = $_max_sum_2 + ($PRIORITY{'func'} * $WEIGHT{'func'});}
         if ($_chk_this eq 'msg')  { $_max_sum_2 = $_max_sum_2 + ($PRIORITY{'msg'}  * $WEIGHT{'msg'});}
         if ($_chk_this eq 'info') {
            foreach $_k ( 'fver', 'vpath', 'ver', 'PANIC_str', 'compile_flag' ) {
               if (&_my_defined($_bt->{$_k})) {
                  $_max_sum_2 = $_max_sum_2 + ($PRIORITY{$_k}  * $WEIGHT{$_k});
               }
            }
         }
         printf("   check bt1 %u  %s [%s]  max sum: %u\n", 
            $_e, $_chk_this, 
            $_bt->{'func'} || $_bt->{'msg'} || 'info', 
            $_max_sum_2
         ) if $DEBUG;
         # check all bt-2 lines, only func or msg
         $_e2 = 0;
         foreach $_bt2 (@{$_[1]} ) {
            # skip this bt2 line if already matched
            if ($_cache{$_e2}) {
               printf("      bt2 %2u matched previously ", $_e2) if $DEBUG;
            }
            else {
               printf("      bt2 %2u ", $_e2) if $DEBUG;
               # already checked if bt1 field is defined when setting $_chk_this
               # bt1 line is a func
               if ($_chk_this eq 'func') {
                  if ($_bt->{'func'} eq $_bt2->{'func'}) {
                     $_match_sum_2  = $_match_sum_2 + ($PRIORITY{'func'} * $WEIGHT{'func'});
                     $_cache{$_e2}++;
                     printf("matched func: [%s] match sum: %u ", $_bt->{'func'},  $_match_sum_2) if $DEBUG;
                  } 
               }
               # bt1 line is a msg
               elsif ($_chk_this eq 'msg') {               
                  if ($_bt->{'msg'} eq $_bt2->{'msg'}) {
                     $_match_sum_2  = $_match_sum_2 + ($PRIORITY{'msg'} * $WEIGHT{'msg'});
                     $_cache{$_e2}++;
                     printf("matched msg:  [%s] match sum: %u", $_bt->{'msg'}, $_match_sum_2) if $DEBUG;
                  } 
               }
               # bt1 line is the info line 
               elsif ($_chk_this eq 'info') {
                  foreach $_k ( 'fver', 'vpath', 'ver', 'PANIC_str', 'compile_flag' ) {
                     if (&_my_defined($_bt->{$_k})) {
                        if ($_bt->{$_k} eq $_bt2->{$_k}) { 
                           $_match_sum_2 = $_match_sum_2 + ($PRIORITY{$_k}  * $WEIGHT{$_k});
                           printf("matched info: [%s] match sum: %u", $_k, $_match_sum_2) if $DEBUG;
                        }
                     }
                  }
 
               }
            }
            printf("\n") if $DEBUG;
            $_e2++;
         }
         $_e++;
      }
   }

   printf("   line-by-line: %8u  %8u    all-other-line %8u  %8u\n", 
      $_max_sum, $_match_sum,
      $_max_sum_2, $_match_sum_2,
   ) if $DEBUG;


   return($_match_sum, $_max_sum, $_match_sum_2, $_max_sum_2,);

}

#.............................................................................
# 
# Parse GDB Backtrace file
#
# Args
#    $_[0] - filename, one backtrace per file
#    $_[1] - filename
#    $_[2] - ref to hash to add
#             @{$hash{$_index}} = (line1, line2, line3, ...);
#              index format  <filename>_bt<num>
# Return
#    number of lines parsed
#
# GDB backtrace lines are expected to be '#' directly followed by a number

sub parse_gdb_backtrace_file {

   my $_l;
   my $_fh       = $_[0];
   my $_bt_num   = 0;
   my $_index    = '';
   my $_bt_lines     = 0;
   my $_all_lines    = 0;
   my $_all_bt_lines = 0;
   my $_bt_linum;

   while ( $_l = <$_fh> )  {
      $_all_lines++;
      next if $_l =~ /^$/;
      next if $_l =~ /^\s+$/;
      
      if ($_l =~ /#(\d+)\s/) {
          $_bt_linum = $1;
          $_all_bt_lines++;
          # start of backtrace
          if ($_bt_linum == 0 ) {
             $_bt_lines = 0;
             $_bt_num++;
             $_index = sprintf("%s_bt%u", $_[1], $_bt_num); 
          }
          chomp($_l);
          push(@{$_[2]->{$_index}}, $_l);
          printf("   parsed gdb bt%u line %u: %s\n", $_bt_num, $., $_l) if $DEBUG;
      }
      if ($_l =~ /PANIC:\s+.+/) {
         chomp($_l);
         push(@{$_[2]->{$_index}}, $_l);
         printf("   parsed gdb bt%u line %u: %s\n", $_bt_num, $., $_l) if $DEBUG;
      }
      if ($_l =~ /version=.+/) {
         chomp($_l);
         push(@{$_[2]->{$_index}}, $_l);
         printf("   parsed gdb bt%u line %u: %s\n", $_bt_num, $., $_l) if $DEBUG;
      }
      if ($_l =~ /cc_flags=.+/) {
         chomp($_l);
         push(@{$_[2]->{$_index}}, $_l);
         printf("   parsed gdb bt%u line %u: %s\n", $_bt_num, $., $_l) if $DEBUG;
      }
   }
   return(wantarray ? ($_bt_num, $_all_bt_lines, $_all_lines) :$_bt_num) ;
}
#.............................................................................
# 
# Parse Backtrace from console log
#
# Args
#    $_[0] - filehandle to console log
#    $_[1] - filename
#    $_[2] - ref to hash to add
#            @{$hash{$_index}} = (line1, line2, line3, ...);
#              index format  <filename>_bt<num>_Ddd_Mmm_hh:mm:ss_yyyy
#
# Return
#   wantarry ? ( bt_num, bt_lines, all_lines) : bt_num
# 
sub parse_backtrace_from_consol {

   my $_l;
   my $_keepit       = 0;
   my $_bt_num       = 0;
   my $_index        = '';
   my $_bts          = 0;
   my $_fh           = $_[0];   
   my $_bt_lines     = 0;
   my $_all_lines    = 0;
   my $_all_bt_lines = 0;
   my ($_ts, $_ts_);

   # TODO
   # error check the filehandle

   while ( $_l = <$_fh> )  {

      $_l =~ s/^\[(\w\w\w.+:\d\d \d\d\d\d)]//;
      $_ts = $1;
      $_l =~ s/^\.+//;
      $_all_lines++;
      # start of panic, make an index
      if ( ($_l =~ /^panic:\s/ || $_l =~ /KDB:\sstack\sbacktrace:/) && $_keepit == 0 ) {
         $_keepit   = 1;
         $_bt_lines = 0;
         $_bt_num++;
         $_ts_ = $_ts;
         $_ts_ =~ s/\s+/_/g;
         $_index = sprintf("%s_bt%u_%s", $_[1], $_bt_num, $_ts_);
         printf("Backtrace %u start found: line %u\n", $_bt_num, $.) if $DEBUG ;
      }
      # clean up line and store in array 
      if ($_keepit) {
         $_l =~ s/\015//g;                        # remove ^M
         chomp($_l);                              # remove newline
         $_bt_lines++;
         $_all_bt_lines++;
         push(@{$_[2]->{$_index}}, $_l);
         printf("   add bt line %u:  %s\n", $., $_l) if $DEBUG;
      }


      # end of panic messages, push on list
      if ( $_l =~ /DUMPCORE:\s+END/ && $_keepit) {
         $_keepit = 0;
         printf("Backtrace %u end found: line %u\n", $_bt_num, $.) if $DEBUG;
      }

      # reboot during panic dump
      if ( $_l =~ /CFE\sversion/) {
         $_keepit = 0;
         printf("CFE prompt detected: %u\n", $_bt_num, $.) if $DEBUG;
      }

      # rebooting line 
      if ( $_l =~ /Rebooting\.\.\./) {
         $_keepit = 0;
         printf("Reboot detected: %u\n", $_bt_num, $.) if $DEBUG; 
      }
      if ( $_l =~ /Type\s\'\?'\sfor\sa\slist\sof\scommands/) {
         $_keepit = 0;
         printf("Command list prompt detected: %u\n", $_bt_num, $.) if $DEBUG;
      }
      # catch runaway's
      if ($_bt_lines > $CONSOL_RUNAWAY ) {
         $_keepit = 0;
         printf("Backtrace %u runaway: line %u\n", $_bt_num, $.) if $DEBUG;
      }

   }
   return(wantarray ? ($_bt_num, $_all_bt_lines, $_all_lines) :$_bt_num) ;
}
#
#----------------------------------------------------------------------------
# Make a signtaure string for the backtrace
# Make 1 signature from two parts
#   1 - foreach func/msg string concatenate and md5sum
#   2 - md5 the PANIC string
# combine the strings
#
# Args
#   $_[0] - ref to parsed backtrace AoH
# Return
#   signature
#   
sub make_bt_signature {

   my $_sign = '';
   my $_s1   = '';
   my $_s2   = '';
   my $_p_str = '';
   my $_P_str = '';
   my $_md5   = Digest::MD5->new;
   my ($_line);

   # foreach func/msg concatenate a string and md5 it
   printf("   create panic signature\n") if $DEBUG;
   foreach $_line  ( @{$_[1]} ) {
      if    ($_line->{'func'})       { $_md5->add($_line->{'func'}); }
      elsif ($_line->{'msg'})        { $_md5->add($_line->{'msg'}); }
      elsif ($_line->{'PANIC_str'})  { $_P_str = $_line->{'PANIC_str'};}
      elsif ($_line->{'panic_str'})  { $_P_str = $_line->{'panic_str'};}
      else                           { }
      printf("   md5 this [%s]\n", $_line->{'func'} || $_line->{'msg'} || 'nofunc') if $DEBUG > 1;
   }
   $_s1 = $_md5->hexdigest;
   printf("   md5 hex digest bt lines:  %s\n", $_s1) if $DEBUG;

   # md5 the panic string
   $_md5->reset;
   if    ($_P_str) { $_md5->add($_P_str);}
   elsif ($_p_str) { $_md5->add($_p_str);}
   else                { $_md5->add('no string');}
   $_s2 = $_md5->hexdigest;
   printf("   md5 hex digest panic str: [%s]\n", $_s2) if $DEBUG;
   # combine both digest for a signature
   $_sign = $_s1 . '::' . $_s2;
   printf("   md5 hex signature:        %s\n", $_sign) if $DEBUG;

   $_sign;
}



#
#----------------------------------------------------------------------------
#
# Parse timestamp info form backtrace index made with &parse_backtrace_from_consol
# and make a hash of epoch-to-backtrace
# user can then use to sort based on time since timestamp in not optimal fo that
#
# Args
#  $_[0] = ref to console backtrace hash
#  $_[1] = ref to hash to populate
#
# return 1
#
sub consol_backtrace_index2epoch {
   my ($_mon, $_day, $_year, 
       $_hour, $_minute, $_second,
       $_epoch,
       $_bt_file,
   );
   foreach $_bt_file (sort keys %{$_[0]}) {
      printf("index2epoch: index %s => ", $_bt_file) if $DEBUG;
      # consol backtrace index <file>_bt<num>_Ddd_Mmm_dd_hh:mm:ss_yyyy
      if ($_bt_file =~ /bt\d+_\w+_(\w+)_(\d+)_(\d+):(\d+):(\d+)_(\d+)/) {
         $_mon    = $MONTH_MAP{$1};
         $_day    = $2;
         $_hour   = $3;
         $_minute = $4;
         $_second = $5;
         $_year   = $6;

         $_epoch = timelocal($_second, $_minute, $_hour, $_day, $_mon, $_year);
         printf("epoch: %u ", $_epoch) if $DEBUG;

         $_[1]->{$_epoch} = $_bt_file;

      }
      printf("\n") if $DEBUG;
   }


   1;
}




#-----------------------------------------------------------------------------
#
# Need my own check for something being there.
# Perl's &defined() will fail if something is 0, 0x000000
# Perl's &exist()  will succeed since the parser creates each field but may put
#        an empty string
# So if the hash holds anything, this will give true
# 
# $_[0] - what to check
# Return - 1 if something in the passed var, undef if empty
#
sub _my_defined {
   if ($_[0] =~ /./) {return(1);}
   else              {return(undef);}
}

1;
__END__


=pod

=head1 NAME Panic_Utils

=head1 Synopsis

    # parse console logs
    ($bt, $line, $all) = &parse_backtrace_from_consol(<filehandle>, <tag>, \%consol);

    # tokenize console backtraces
    $ok = &parse_consol_backtrace_list(\@{$consol{$tag}}, \@{$tokens{$tag}}); 

    # parse gdb backtrace file
    ($bt, $line, $all) = &parse_gdb_backtrace_file(<filehandle>, <tag>, \%gdb);

    # tokenize gdb backtrace
    $ok = &parse_gdb_backtrace_list(\@{$gdb{$tag}}, \@{$tokens{$tag}});

    # compare backtrace tokens
    @stats = &compare_backtrace(\@{$tokens{<tag1>}}, \@{$tokens{<tag2>}});

=head1 Description

Provide functions to parse backtrace from console logs and files containing
backtraces from gdb shell bt command output. Given a list of lines parsed as
a backtrace, separate each part of the line into tokens. These tokens
can then be used to statistically compare backtraces to each other.
Comparison is done line to line and token to token. Dual weighted
algorithm is used to derive values for statistical analysis. User
has configuration control to the weighted algorithm which allows the user
to control what tokens are compared and its respective weight to the overall
comparison value


=head1 Global Values

Values contolling operation, access to set or read values should be
done directly from user script.

=over

=item $DEBUG

    $Panic_Utils::DEBUG = <int>;

Debug flag, set to 1 to display messages to STDOUT.


=item $CONSOL_RUNAWAY

    $Panic_Utils::CONSOL_RUNAWAY = 48;

Safegaurd against parsing to many console backtrace lines. In the case of
lines delimiting end of back trace are missing or corrupt, the parse will stop
after this many lines are put into the array being created.

Default is 48.

=item $COMPARE_MODE

Bit mask setting to contol which comparison is done, Line-by-Line,
All-Lines-Compared or both.

=over

=item $Panic_Utils::COMPARE_MODE = 0x01;   # enable Line-by-Line

=item $Panic_Utils::COMPARE_MODE = 0x02;   # enable All-Lines-Compared

=item $Panic_Utils::COMPARE_MODE = 0x03;   # enable both

=back

=item Dual Weighting

=over

=item %PRIORITY

Hash whose keys are the same name as the tokens names used. Thus when
tokens are compared the value of the weight is affected by PRIORITY.
PRIORITY is intended to indicate 'how important' this token is
in the overall comparison. By default it is scaled from 0-10 with
0 nullifying the token.

    $Panic_Utils::PRIORITY::index        = 10    # gdb
    $Panic_Utils::PRIORITY::addr         = 1     # gdb
    $Panic_Utils::PRIORITY::func         = 10    # gdb, console
    $Panic_Utils::PRIORITY::vars         = 1     # gdb, consol
    $Panic_Utils::PRIORITY::file         = 5     # gdb
    $Panic_Utils::PRIORITY::fname        = 10    # gdb, consol
    $Panic_Utils::PRIORITY::fpath        = 10    # gdb
    $Panic_Utils::PRIORITY::linum        = 2     # gdb
    $Panic_Utils::PRIORITY::msg          = 10    # gdb, consol
    $Panic_Utils::PRIORITY::offset       = 2     # consol
    $Panic_Utils::PRIORITY::ver          = 5     # consol     from version:
    $Panic_Utils::PRIORITY::fver         = 5     # consol     from version line with full path
    $Panic_Utils::PRIORITY::vpath        = 5     # consol     full version path
    $Panic_Utils::PRIORITY::panic_str    = 10    # consol
    $Panic_Utils::PRIORITY::PANIC_str    = 10    # consol
    $Panic_Utils::PRIORITY::uptime       = 0     # consol
    $Panic_Utils::PRIORITY::compile_flag = 0     # consol
    $Panic_Utils::PRIORITY::_signature   = 0     # gdb, consol
    $Panic_Utils::PRIORITY::_syntax      = 0     # gdb, consol

=item %WEIGHT

Hash whose keys are the same name as the tokens names used. Thus when
tokens are compared the value of the weight is controlled by WEIGHT.
In short this is a multipler.  The value for this token if matched
is PRIORITY * WEIGHT. This number is added to running total for the
whole backtrace compare.

    $Panic_Utils::WEIGHT::index        = 100    # bt line index
    $Panic_Utils::WEIGHT::addr         = 1      # bt addr value
    $Panic_Utils::WEIGHT::func         = 100    # bt function name
    $Panic_Utils::WEIGHT::vars         = 1      # bt function vars
    $Panic_Utils::WEIGHT::file         = 80     # file name
    $Panic_Utils::WEIGHT::fname        = 100    # full file name
    $Panic_Utils::WEIGHT::fpath        = 10     # full file path
    $Panic_Utils::WEIGHT::linum        = 10     # file line number
    $Panic_Utils::WEIGHT::msg          = 100    # bt message
    $Panic_Utils::WEIGHT::offset       = 10     # offset value
    $Panic_Utils::WEIGHT::ver          = 50     # version string
    $Panic_Utils::WEIGHT::fver         = 5      # full version string
    $Panic_Utils::WEIGHT::vpath        = 5      # version path
    $Panic_Utils::WEIGHT::panic_str    = 100    # panic string in bt
    $Panic_Utils::WEIGHT::PANIC_str    = 100    # panic string following PANIC:
    $Panic_Utils::WEIGHT::uptime       = 1      # sysUptime
    $Panic_Utils::WEIGHT::compile_flag = 1      # compile flags
    $Panic_Utils::WEIGHT::_signature   = 1      # panic signature
    $Panic_Utils::WEIGHT::_syntax      = 1      # bt line syntax

=back

The above PRIORITY and WEIGHT are the heart of the comparison.
When two tokenized list are compared, respective tokens are compared.
Each token in PRIORITY is check if defined in backtrace 1. If defined
then WEIGHT#PRIORITY is added to max_sum. If the tokens match then
WEIGHT#PRIORITY for this token are added to match_sum. These values
are then returned to the user. These values are calculated and kept separate 
for both comparison modes, thus the user is returned 4 numbers. See 
&compare_backtrace description.

=back

=head1 Functions


Parsing backtraces creates Array of Hash (AoH) where each element of the 
array is a hash whose key/values are the tokens parsed from the backtrace
line. For any line parsed, not all key/value pairs will be defined but will 
exist. For example if token foo does not occur on the backtrace line, foo will
exist in hash with an undef value  ('foo' => '',). 

=head2 Data Structure Format of Tokenized Backtrace

The AoH is created and worked upon by the functions listed below.
The first N lines will make the same hash structure.
The last hash entry will give information about the panic.
Data structure format will be described here:

     @AoH = ( {'index'   => <int>,        # the backtrace index at begin of line
               'addr'    => <hex>,        # the hex address right after the index
               'func'    => <string>,     # the function name
               'vars'    => <string>,     # the content of () right after function name
               'file'    => <string>,     # the full filename referenced in backtrace
               'fname'   => <string>,     # the file name, separated from path
               'fpath'   => <string>,     # the path to fill name
               'linum',  => <int>,        # the line number given for given file
               'msg'     => <string>      # string contained within < >, follows #index on backtrace line
               '_syntax' => <string>      # syntax format
              },
              { ... },
              { ... },
         
              {'fver'         => $_fver,          # full version string 
               'vpath'        => $_vpath,         # version path
               'ver'          => $_ver,           # version string
               'panic_str'    => $_panic_str,     # panic string found in backtrace
               'PANIC_str'    => $_PANIC_str,     # panic strinc denoted with PANIC:
               'uptime'       => $_uptime,        # sys uptime
               'compile_flag' => $_compile_flags, # compile flags
               '_syntax'      => 'info',          # syntax descriptor
               '_type'        => 'consol',        # backtrace type
               '_signature'   => 'none',          # backtrace signature
              },


      )

=head2 parse_gdb_backtrace_file

Parse a gdb backtrace from a file containing the output of a core dump
analyzed with gdb and the gdb shell command 'bt' executed. The output 
of the 'bt' command is whats to be parsed. 

    ($bt, $line, $all) = &parse_gdb_backtrace_file(<fh>, <tag>, \%hash);

    Arg 1 - filehandle to parse for gdb shell 'bt' output
    Arg 2 - tag, used to create a key for holding backtrace
    Arg 3 - reference to hash to populate with backtrace lines
            @{$hash{$tag}} = (line1, line2, line3, ...,)

    $tag is made from arg 2 and number of backtraces found, so the first
    backtrace found will have tag_1, then tag_2, ....

   In array context the number of backtraces found, the number of backtrace 
   lines, the number of total lines is returned. In scalar context the 
   number of backtraces found is returned.

The data structure created by this function can be used by to 
create the backtrace list for &parse_gdb_backtrace_list. 


=head2 parse_gdb_backtrace_list

Parse the backtrace obtained from gdb shell bt command output.
GDB backtrace has been parsed from a file and contained in an array
with each backtrace line being a separate element. The array could
of been created with &parse_gdb_backtrace_file or made by the user.


    $ok = &parse_gdb_backtrace_list(\@gdb, \@tokens);
    ($ok, $error) = &parse_gdb_backtrace_list(\@gdb, \@tokens);

     Arg 1 - ref to array of bactrace lines
     Arg 2 - ref to array of hashes to populate with tokens

Return status in scalar context and in array context status and error.

Array reference with second argument is populated with the tokens
parsed from each backtrace line. See Data Structure Format section above. 


=head2 parse_backtrace_from_consol

Parse a gdb backtrace from a console log file.

    ($bt, $line, $all) = &parse_backtrace_from_consol(<fh>, <tag>, \%hash);

    Arg 1 - filehandle to parse for gdb shell 'bt' output
    Arg 2 - tag, used to create a key for holding backtrace
    Arg 3 - reference to hash to populate with backtrace lines
            @{$hash{$tag}} = (line1, line2, line3, ...,)

    $tag is made from arg 2, number of backtraces found and timestamp
    found on the line at start of backtrace. 
    Such as $tag_bt<num>_Ddd_Mmm_d_hh:mm:ss. The timestamp is parsed from the
    console line, has its spaces replaced with '_' and appended to tag string. 

   In array context the number of backtraces found, the number of backtrace 
   lines, the number of total lines is returned. In scalar context the 
   number of backtraces found is returned.

The data structure created by this function can be used by to 
create the backtrace list for &parse_consol_backtrace_list. 


To prevent the parser from getting into a runaway state due to various
reasons, the variable $CONSOL_RUNAWAY can be set so as to stop parsing
after N lines. Default is 48.


Lines that are used by parser to indicate end of backtrace

=over

=item DUMPCORE: END

=item CFE version

=item Rebooting...

=item Type '?' for a list of commands

=back


=head2 parse_consol_backtrace_list

Parse the backtrace lines from given list of console backtrace lines.
Console backtraces have been parsed from a file and contained in an array
with each backtrace line being a separate element. The array could
of been created with &parse_backtrace_from_consol or made by the user.

    $ok = &parse_consol_backtrace_list(\@gdb, \@tokens);
    ($ok, $error) = &parse_gdb_backtrace_list(\@gdb, \@tokens);

     Arg 1 - ref to array of bactrace lines
     Arg 2 - ref to array of hashes to populate with tokens

Return status in scalar context and in array context status and error.

Array reference with second argument is populated with the tokens
parsed from each backtrace line. See Data Structure Format section above. 

=head2 compare_backtrace

Compare two backtrace token list.
Tokens are compred to thier respective tokens for a match.
Token list should be created with &parse_consol_backtrace_list
or parse_gdb_backtrace_list. 

    @stats = &compare_backtrace( $aref1, $aref2);

    Arg 1 - ref to AoH of backtrace line tokens
    Arg 1 - ref to AoH of backtrace line tokens to compare to


There are two comparisons and are controlled by $COMPARE_MODE flag.

Line-by-Line (lbl) - each tokenized backtrace line is compared to its 
respective tokenized entry in the other backtrace. In other words
bt-1_line-1 is compared to bt-2_line1, bt1_line2 <=> bt2_line2, ...

All-Lines-Compared (alc) - each tokenized backtrace line is compared
to all lines in the other backtrace. In other words bt1-line-1 is compared
to all lines in bt2, bt1-line2 is compared to all lines in bt2. Once a line
is matched in bt2 it is not used again in compares.


=head3 Comparison Modes

Comparison is controlled by $COMPARE_MODE bit mask

=over

=item  Line-by-Line set bit 1 true, default = 1

=item  All-Lines-Compared set bit 2 to true, default = 0

=back

See Global Setting section for COMPARE_MODE info.

Comparison numbers are created while checking for match token values.
This number is controlled by the Weighting algorithm, see Global Setting
section for more info. As each token is compared the weights are applied
accordingly. This info is returned to the user 


Values returned represent values for each compison mode (lbl, alc), 
regardless if the comparison was done as per the COMPARE_MODE flag.
Thus 4 values are returned in array context, from the sample line above:

=over

=item $stats[0] - lbl match value (actual)
 
=item $stats[0] - lbl highest possible value (max) 

=item $stats[0] - alc match value (actual)

=item $stats[0] - alc highest possible value (max)

=back

The user can then use these numbers to compare panics.
One such way is to take the ((actual)/max)*100 to normalize as a percentage.

See the PRIORITY and WEIGHT description in Global Setting section for
more comparison info.


=head2 consol_backtrace_panic_str

Retun the panic and PANIC token from a tokenized backtrace.


=head2 make_bt_signature

Used to make a backtrace (panic) signature.

    $signature = &make_bt_signature( $aoh_ref );

Given a reference to a AoH of a tokenized backtrace, this function will create
a MD5 signature of th ebacktrace. Each backtrace line is looked at in list order
and the function or message or panic string token is concatenated and then MD5'd. This MD5 sum
is returned as the signature.

This same function is called internally when creating the MD5 sum but is provided
to the user. If called by the user it will not change the signature possibly 
already in the AoH. User would have to manually do this



=head2 consol_backtrace_index2epoch

Parse timestamp info form backtrace index made with &parse_backtrace_from_consol
and make a hash of epoch-to-backtrace. User can then use to sort based on time 
since timestamp in index string is not optimal for that.

=head1 Example

    #!/usr/software/bin/perl
    #  
    # Parse Panic BackTrace files and compare
    
    use strict;
    use Panic_Utils;
    use Data::Dumper;
    use IO::File;
    
    $Data::Dumper::Indent   = 1;  # default = 2
    
    my %backtraces;
    my %hash;
    
    # gdb backtrace files
    my @bt_files = (
       '/some/path/core.101170979.2008-06-12.19_30_12.nz.bt',
       '/some/path/core.101170979.2008-06-12.19_39_30.nz.bt',
       '/some/path/core.101170979.2008-06-12.19_52_10.nz.bt',
    );
    # 
    # console logs
    my @con_files = ( 
       '/some/path/rtpgx-3050-5.con.log',
       '/some/path/rtpgx-3050-6.con.log',
    ); 
       
    
    my( $bt_file, $con_file, $fh, $ok, $error,
        $bt_file2,
        @results
    );
    
    $Panic_Utils::DEBUG = 0;
    $Panic_Utils::COMPARE_MODE = 0x03;
    
    #
    # Parse the files
    #
    # parse backtrace files and store parsed info
    printf("Parsing Bactrace files ...\n");
    foreach $bt_file (@bt_files) {
       printf("Parse bt file: %s\n", $bt_file);
       # create filehandle
       $fh = IO::File->new($bt_file, 'r');
       if ($fh) {
          # parse the backtrace file
          ($ok, $error) = &parse_gdb_backtrace_file($fh,
                                                   \@{$backtraces{$bt_file}}
          );
          if ($ok) {
             printf("Parsed %u lines\n", $ok);
             # parse the backtrace list
             ($ok, $error) &parse_gdb_backtrace_list(\@{$backtraces{$bt_file}},
                                                     \@{$hash{$bt_file}}
             );
             if ($ok) {printf("Parsed backtrace list\n");}
             else     {printf("ERROR parse_gdb_backtrace_list: %s\n", $error);}
          }
          else     { printf("ERROR: parse_gdb_backtrace_file: %s\n", $error);}
          $fh->close;
       }
       else {
          printf("IO::File error: %s\n", $!);
       }
    }
    #print Dumper(\%backtraces);
    
    # Parse consol log
    printf("Parsing consol log ...\n");
    foreach $con_file (@con_files) {
       printf("Parse con file: %s\n", $con_file);
       # create filehandle
       $fh = IO::File->new($con_file, 'r');
       if ($fh) {
          ($ok, $error) = &parse_backtrace_from_consol($fh, $con_file, \%backtraces);
          if ($ok) {
             printf("Parsed %u panic backtrace lines\n", $ok);
             foreach $bt_file (sort keys %backtraces) {
                if ($bt_file =~ $con_file) {
                   printf("Parse con file bt list: %s\n", $bt_file);
                   ($ok, $error) = &parse_consol_backtrace_list(\@{$backtraces{$bt_file}},
                                                                \@{$hash{$bt_file}}
                   );
                   if ($ok) {printf("Parsed consol backtrace list\n");}
                   else     {printf("ERROR parse_consol_backtrace_list: %s\n", $error);}
                }
             }
          }
          else     {
             printf("ERROR: %s\n", $error);
          }
          $fh->close;
       }
       else {
          printf("IO::File error: %s\n", $!);
       }
    }
    #print Dumper(\%backtraces);
    #print Dumper(\%hash);
    
    printf("%s\n", '='x79);
    
    #
    # Compare the parsed back traces
    #
    
    foreach $bt_file (sort keys %hash) {
       printf("Compare backtrace fields: %s\n", $bt_file);
       foreach $bt_file2 (sort keys %hash) {
          next if $bt_file eq $bt_file2;
          @results = ();
          @results = &compare_backtrace(\@{$hash{$bt_file}}, \@{$hash{$bt_file2}});
          if (&_calc_percent($results[0], $results[1]) > 95.0) {
              printf("   cmp2: %-85s ", $bt_file2);
             printf("   %-6s %6u/%-6u  %6s   %-6s %6u/%-6u  %6s\n",
                'lbl:',  $results[0], $results[1],
                &_calc_percent($results[0], $results[1]),
                'alc:', $results[2], $results[3],
                &_calc_percent($results[2], $results[3]),
             );
          }
       }
       printf("\n");
    }
    
    exit(0);
    
    #=============================================================================
    # Calculate percentage value, pass back value
    # Used to handle when a value of 0 is used. 
    #
    sub _calc_percent {
        if ( $_[0] == 0 )  { return(0); }
        if ( $_[1] == 0 )  { return(0); }
        sprintf("%.2f", ($_[0]/$_[1]) * 100);
    }

=head1 STATUS

   Current 

=head1 AUTHOR

  Scott Parsons (parsons@netapp.com)
  Last revised $Date: 2008/07/14 $

=head1 COPYRIGHT

Copyright 2008, Network Appliance., All Rights Reserved


=cut



   
