# this is basically -*- perl -*-
# vim: set ft=perl tabstop=4 shiftwidth=4 expandtab shiftround: 
#
# $Id:$
#
# Copyright (c) 2007 Network Appliance Inc.
# All rights reserved.
#

=head1 NAME

Params::CIFS - common CIFS definitions for Options.pm parameters

=head1 SYNOPSIS

   use Options;
   use Params::CIFS;

=head1 DESCRIPTION

This package is used to define the common CIFS parameters used by
GX testers. It can be used by anyone using the L<Options> package
as a convienient way to keep your options all consistent.

=head1 METHODS

none

=cut

use Options;
use TracePkg;

my $Trace = new TracePkg;

# figure out where the kerberos binaries live on this client
my $Kerberos_Path = _find_kerberos_default_path();

$Options = new Options;
$Options->parameters( definitions=> [
        {
            name        => "SMBCLIENT",
            description => "smbclient binary",
            default     => "/usr/bin/smbclient",
            file_tests  => "rx",
        },
        {
            name        => "KERBEROS_REALM",
            description => "Name of the Kerberos Realm",
        },
        {
            name        => "KINIT",
            description => "kinit binary",
            default     => "$Kerberos_Path/kinit",
            file_tests  => "rx",
        },
        {
            name        => "KLIST",
            description => "klist binary",
            default     => "$Kerberos_Path/klist",
            file_tests  => "rx",
        },
        {
            name        => "KDESTROY",
            description => "kdestroy binary",
            default     => "$Kerberos_Path/kdestroy",
            file_tests  => "rx",
        },
        {
            name        => "CIFS_DRIVE",
            description => "Drive to map shares to on windows clients",
            default     => "*",
        },
        {
            name        => "CIFS_IGNORE_DRIVES",
            description => "CIFS drive mappings to ignore",
            default     => ["X:"],
            spec        => "s@",
        },
        # we explicity do not set a default here because we want to let the system provide the default
        {
            name        => "CIFS_SHARE_SHARE_PROP",
            description => "Properties that should be active for a particular cifs share",
            spec        => "s@",
            valid_values    => [qw(dynamicshare dynamicpath oplocks browsable showsnapshot)],
        },
        # we explicity do not set a default here because we want to let the system provide the default
        {
            name        => "CIFS_SHARE_SYMLINK_PROP",
            description => "Properties that should be active for a particular cifs share",
            spec        => "s",
            valid_values    => [qw(enable hide read_only)],
        },
        # we explicity do not set a default here because we want to let the system provide the default
        {
            name        => "CIFS_SHARE_FILE_UMASK",
            description => "File Mode Creation Mask for files created on a share",
            spec        => "s",
        },
        # we explicity do not set a default here because we want to let the system provide the default
        {
            name        => "CIFS_SHARE_DIR_UMASK",
            description => "Directory Mode Creation Mask for files created on a share",
            spec        => "s",
        },
        {
            name        => "AUTH_FLAVOR",
            description => "The style of authentication a test will use to log in",
            spec        => "s@",
            default     => [qw(ntlm kerberos)],
            valid_values    => [qw(ntlm kerberos)],
        },
]);

# If there are any options that have changed names here is where we would put them
$Options->compatability( definitions=>
    {
    });

#
# find the kerberos binaries on this client
#
sub _find_kerberos_default_path {
    $Trace->entry(msg => "");
    my $binary_path = '';

    # look for the binary on the system, first we look for it in the most common place, then fall
    # back to searching the user's PATH
    foreach my $path ( "/usr/kerberos/bin", "/usr/bin", split(/:/, $ENV{PATH}) ) {
        if ( -e "$path/kinit" ) {
            $Trace->trace(msg =>
                "Found kinit binary in '$path' using that for the kerberos binary path.");
            $binary_path = "$path";
            last;
        } else {
            $Trace->debug(msg => "Looked in '$path' and have not found kinit yet.");
        }
    }
    
    # if we didn't find a default - we need to set it to something so it fails
    if ( $binary_path eq '' ) {
        $Trace->warn(msg => "Unable to find a sane default path for the kerberos binaries on this host.");
        $binary_path = "no_default_path_for_kerberos_binaries_found";
    }
    $Trace->exit(msg=> "binary_path = '$binary_path'");
    return($binary_path);
}

1;
