## @summary     common parameter definitions for Options.pm users
## @author      snider@netapp.com
## @pod here

=head1 NAME

Params::Common - common definitions for Options.pm parameters

=head1 SYNOPSIS

   use Options;
   use Params::Common;

=head1 DESCRIPTION

This package is used to define the common parameters used by
GX testers. This package is included automatically by L<TestManager>
tests. It can be used by anyone also using the L<Options> package.

=head1 METHODS

none

=cut

package Params::Common;
use Options;
use TracePkg;
use Socket qw(inet_ntoa);
$Trace = new TracePkg;

require Exporter;
our @ISA    = qw(Exporter);
our @EXPORT = qw(
    parse_release_info
    $Release_separator
);
our @EXPORT_OK = qw(
    parse_release_info
    $Release_separator
);

our $Release_separator = ";";

$Options = new Options;

$Options->parameters(
    definitions => [
        {   
            name        => "NODES",
            spec        => "s@",
            subroutine  => [ \&_nodes_to_filer ],
            description => "list of comma-separated nodes against which script will be run",
            sys_prefix  => 1,
        },
        {
            name => "CLUSTER_NAME",
            description => "Name of the cluster",
            default => undef,
            sys_prefix  => 1,
        },
        {   
            name        => "CLIENT_TYPE",
            description => "The type of the client",
            spec        => "s",
        },
        {   
            name        => "CIFS",
            description => "Enables CIFS configuration",
            spec        => "boolean",
            sys_prefix  => 1,
        },
        {   
            name        => "RELEASE",
            description => "the release the FILER is running",
            sys_prefix  => 1,
        },
        {   
            name        => "SEED",
            description => "the seed to pass to srand",
            subroutine  => [ \&_set_random_seed ],
        },
        {   
            name        => "CLIENT",
            spec        => "s@",
            description => "list of comma-separated clients",
        },
        {   
            name        => "CLIENTFILE",
            spec        => "s",
            description => "file containing the list of clients",
            file_tests  => "fr",
            subroutine  => [ \&_read_client_file ],
        },
        {   
            name        => "CLIENTFILE_RANGE",
            spec        => "s@",
            description =>
                "range of the clients read from clientfile. Default is all with 0-0 setting. Ranges can be combined with comma separators",
            default     => "0-0",
        },
        {   
            name        => "WIN_CLIENTS",
            spec        => "s@",
            description => "list of comma-separated windows clients",
        },
        {   
            name        => "FILERFILE",
            spec        => "s",
            description => "file containing the list of filers",
            file_tests  => "fr",
            subroutine  => [ \&_read_filer_file ],
        } ,
        {   
            name        => "vserver",
            description => "the name of a vserver to use",
            sys_prefix  => 1,
        },
        {   
            name        => "DATA_INTERFACES",
            spec        => "s@",
            description =>
                "list of comma-separated data_interfaces, or range[specified as IP/range]",
            delimiter   => ",",
            subroutine  => [ \&_process_interfaces ],
            sys_prefix  => 1,
        },
        {   
            name        => "aggregate",
            spec        => "s@",
            description => "list of comma-separated aggregates",
            sys_prefix  => 1,
        },
        {   
            name        => "SRC_AGGREGATE",
            spec        => "s@",
            description =>
                "list of aggregates to be used as a SOURCE for some operation",
            sys_prefix  => 1,
        },
        {   
            name        => "DEST_AGGREGATE",
            spec        => "s@",
            description =>
                "list of aggregates to be used as a DESTINATION for some operation",
        },
        {   
            name        => "SOURCE",
            description =>
                "the location of any additional source code or needed binaries",
        },
        {   
            name        => "USERID",
            description => "User ID to use during test",
            sys_prefix  => 1,
        },
        {   
            name        => "GROUPID",
            description => "Group ID to use during test",
            sys_prefix  => 1,
        },
        {   
            name        => "share",
            default     => "/",
            description => "FILER export",
            sys_prefix  => 1,
        },
        {   
            name        => "CIFS_DOMAIN",
            spec        => "s",
            description => "The name (NetBios name) of the CIFS domains",
            sys_prefix  => 1,
        },
        {   
            name        => "CIFS_SERVER",
            spec        => "s",
            description => "The name (NetBios name) of the CIFS servers",
            sys_prefix  => 1,
        },
        {   
            name        => "CIFS_SHARE",
            spec        => "s",
            description => "The name (NetBios name) of the CIFS shares",
            sys_prefix  => 1,
        },
        {   
            name        => "CIFS_DRIVE",
            spec        => "s",
            description =>
                "Local drive letter to use on client instead of picking one.",
        },
        {   
            name        => "CIFS_USER",
            spec        => "s",
            description => "CIFS User to map the network drive with.",

        },
        {   
            name        => "CIFS_PASSWD",
            spec        => "s",
            description => "CIFS Password to map the network drive with.",
        },
        {   
            name        => "NODE_RELEASE_INFO",
            spec        => "s@",
            delimiter   => ",",  # Sparse branches include a : in their names 
            description => "node_name;branch;release;build;buildopts[;buildpath;[...]],[...]",
            subroutine  => [ \&_verify_node_release_info ],
            sys_prefix  => 1,
        },
        {   
            name        => "SITE",
            spec        => "s",
            description => "The location of the Nodes and clients",
            valid_values => [
                qw(ptc rtp rtp_cqa rtpeng stresScale ssqa nane svl btc kc smoke_bubble
                    kc-nas kc-oz jeti ctebtc ctegdl ptc_ctl sti gdl beta)
            ],
        },
        {   
            name        => "AGGR_BLOCK_TYPE",
            description =>
                "This specifies the type of aggregate to be filtered, i.e. 32-bit or 64-bit",
            valid_values => [qw(32-bit 64-bit *)],
            sys_prefix  => 1,
        },
        {
            name        => "MAIL_TO",
            description => "E-mail address to send email to",
        },
        {
            name        => "MAIL_FROM",
            description => "E-mail address to send email from",
        },
        {
            name        => "MAIL_CC",
            description => "E-mail address to send email as a cc:",
        },
    ]
);

$Options->compatability(
    definitions => {
        FILER              => "NODES",
        FILER_AGGR         => "aggregate",
        VFILER             => "vserver",
        network_interfaces => "DATA_INTERFACES",
        CLIENTS            => "CLIENT",
        EXPORT             => "share",
        CLIENT_FILE        => "CLIENTFILE",
        CLIENT_FILE_RANGE  => "CLIENTFILE_RANGE",
        WINCLIENT          => "WIN_CLIENTS",
    }
);

my $File_read = undef;

sub parse_release_info {
    my %params = (
        node => undef,
        @_
    );
    $Trace->entry();

    my @release_info = ();
    my $options      = $Options->get_hash();
    foreach my $release_info ( @{ $options->{NODE_RELEASE_INFO} } ) {
        $Trace->debug(msg => "release_info = '$release_info'");
        my ( $node, $branch, $release, $build, $buildopts, $buildpath )
            = split( /$Release_separator/, $release_info );
        if ( ( $params{node} && $params{node} eq $node )
            || !$params{node} )
        {
            @release_info
                = ( $branch, $release, $build, $buildopts, $buildpath );
            last;
        } ## end if ( ( $params{node} &&...
    } ## end foreach my $release_info ( ...
    $Trace->exit( msg => join( ",", @release_info ) );
    return @release_info

} ## end sub parse_release_info

sub _nodes_to_filer {
    my $nodes = shift;
    $Trace->entry();

    # The FILER param used to be the default and was defined in Options.pm.
    # This param is now named NODES and defined in this package. But some
    # tests might still be looking for FILER so define it here for backwards
    # compatability
    $Options->set( FILER => $nodes );
    $Trace->exit();
} ## end sub _nodes_to_filer

sub _set_random_seed {
    my $seed = shift;
    $Trace->entry( msg => "seed=$seed" );
    our $prev_seed;

    if ( !defined($prev_seed) || $seed != $prev_seed ) {
        $Trace->comment( msg => "SEED=$seed\n" );
        srand($seed);
        $prev_seed = $seed;
    }
    $Trace->exit();
} ## end sub _set_random_seed

sub _process_interfaces {
    my $interfaces = shift;
    $Trace->entry( msg => "interfaces=>" . join( ",", @$interfaces ) );

    my @interfaces;
    foreach my $ip ( @{$interfaces} ) {
        if ( $ip =~ /local/ ) {

            #	    $self->{localNbladeIP} = 1;
        } elsif ( $ip =~ /remote/ ) {

            #	    $self->{remoteNbladeIP} = 1;
        } elsif ( $ip =~ /random/ ) {    # this is the default so do nothing
                # this now means to use all of the vifs for the virtual server
            ;
        } elsif ( $ip =~ /(.+)\/(.+)/ ) {
            my ( $ipaddress, $range ) = ( $1, $2 );
            $ipaddress = inet_ntoa( scalar( gethostbyname($ipaddress) ) );
            my ( $a, $b, $c, $d ) = split( /\./, $ipaddress );
            push( @interfaces, "$a.$b.$c.$d" );
            for ( my $i = 1; $i < $range; $i++ ) {
                $d++;
                if ( $d > 255 ) {
                    $d = 0;
                    $c++;
                }
                if ( $c > 255 ) {
                    $c = 0;
                    $b++;
                }
                if ( $b > 255 ) {
                    $b = 0;
                    $a++;
                }
                if ( $a > 255 ) {
                    $Trace->fatal(
                        msg    => "could not parse $ip\n",
                        result => "CONFIG"
                    );
                } ## end if ( $a > 255 )
                push( @interfaces, "$a.$b.$c.$d" );
            } ## end for ( my $i = 1; $i < $range...
        } else {
            push( @interfaces, $ip );
        }
    } ## end foreach my $ip ( @{$interfaces...
    $Options->set( DATA_INTERFACES => \@interfaces );
    $Trace->exit();
} ## end sub _process_interfaces

sub _read_client_file {
    my $file = shift;
    $Trace->entry( msg => "file=>$file" );
    if ($File_read) {
        $Trace->trace( msg => "file already processed" );
        goto done;
    }
    my $clients = $Options->get( option => "CLIENT" );
    my $ranges  = $Options->get( option => "CLIENTFILE_RANGE" );
    $Trace->trace( msg => "Processing CLIENTFILE for range '@$ranges'" );

    my ( @orig_lines, @lines, $total );

    # Save original contents of the file in an array
    if ( open( IF, $file ) ) {

        # remove the comment lines so that the @orig_lines array consists
        # solely of client names
        while (<IF>) {
            next if (/^\s*#/);
            $_=~ s/^\s+//; #remove leading spaces
            $_=~ s/\s+$//; #remove trailing spaces
            chomp;
            push( @orig_lines, $_ );
        }
        $total = scalar(@orig_lines);
        close(IF);
    } ## end if ( open( IF, $file ))

    # Filter out the required range of clients
    foreach my $range (@$ranges) {
        my ( $start, $end );
        if ( $range =~ /%/ ) {

            # If % is specified, if 'start %' has values greater than 1,
            # substract by 1, so you will not miss out any clients. But
            # increment the real value of 'start', so that the same client
            # will not be reused. This is based on the assumption, that user
            # does not repeat the numbers. If he does, then it's most likely
            # that's what he wants. E.g "1-30%,30%-40" makes 1% of the lines
            # in 30% range will be redundant.

            $range =~ s/%//;
            my ( $startp, $endp ) = split( "-", $range );
            if ( $startp <= 1 ) {
                $startp = 0;
            } else {
                $startp = $startp - 1;
            }
            $start = int( $startp * $total / 100 );
            $end   = int( $endp * $total / 100 );
            if ( $endp != 0 && $end == 0 ) {

                # Set atleast one client, if 0 is not intended.
                $end = 1;
            }
            $start = $start + 1;
        } else {

            # No % given, so straight forward to just play with the numbers
            ( $start, $end ) = split( "-", $range );
        }
        my @temp_lines = @orig_lines;
        $start = $start - 1 if ( $start >= 1 );
        $end   = $end - $start;
        $end   = 0 if ( $end < 0 );
        if ( $end == 0 ) {
            push( @lines, splice( @temp_lines, $start ) );
        } else {
            push( @lines, splice( @temp_lines, $start, $end ) );
        }
    } ## end foreach my $range (@$ranges)

    # Save the filtered clients in to CLIENT
    foreach (@lines) {
        chomp;
        next if ( $_ eq '' );
        push( @{$clients}, $_ );
    }
    $File_read = 1;

done:
    $Trace->exit();
} ## end sub _read_client_file

sub _verify_node_release_info {
    my $release_info = shift;

    $Trace->entry( msg => $release_info );
    foreach my $release_info (@$release_info) {
        $Trace->debug(msg => "release_info = '$release_info'");
        # ($buildpath is optional)
        my ( $node, $branch, $release, $build, $buildopts, $buildpath )
            = split( /$Release_separator/, $release_info );
        if ( !( $node && $branch && $release && $build && $buildopts ) ) {
            $Trace->warn(
                msg => "Incorrect format for NODE_RELEASE_INFO:"
                    . "$node:$branch:$release:$build:$buildopts",
                result => "CONFIG"
            );

# Node::NG gave a previous bad config. Warn about it and undo the damage so that we don't
#  have random failures around it
            $Options->set( NODE_RELEASE_INFO => undef );
        } ## end if ( !( $node && $branch...
    } ## end foreach my $release_info (@$release_info)
    $Trace->exit();
} ## end sub _verify_node_release_info

=head1 PARAMS

=over

=item RELEASE

Can be specified to indicate the release a given filer/cluster is running.

=item SEED

This is a seed which can be passed in to srand(). 

=item CLIENT

This is a comma-separated list of clients.

=item CLIENTFILE

This is a file which contains a newline-separated list if clients. The clients are 
internally added to the CLIENTS list. Any line which starts with '#' is considered
a comment and is skipped.

=item CLIENTFILE_RANGE

This is to read the clients selectively from CLIENTFILE. Multiple ranges can be combined with ','. E.g CLIENTFILE_RANGE="0-35,50-70,90-0". "90-0" means that CLIENTFILE is read from line 90 to end of file. "0-35" specifies from the start of the file to line 35 inclusive. By default, this is set to "0-0" which enables to read from start to end of the file.

Values can be given in % as well. E.g CLIENT_FILERANGE=1-20%,90-100%. This 
reads the first and last 20% of the clients (40% total) from the CLIENTFILE.

Any line which is considered to be a comment (see CLIENTFILE above) is skipped and not counted as part of the range.

=item vserver

This param can be used to indicate the name of a virtual server to use.

=item volume

This param can be used to indicate a comma-separated list of volumes to use.

=item DATA_INTERFACES

This param can be used to indicate a comma-separated list of data interfaces to use.
The param can also be specified as IP/range where IP represents the beginning IP address
and the range indicates the number of successive IPs beginning at the IP. For example,
192.168.0.1/10 results in the following:

192.168.0.1
192.168.0.2
192.168.0.3
192.168.0.4
192.168.0.5
192.168.0.6
192.168.0.7
192.168.0.8
192.168.0.9
192.168.0.10

=item aggregate

This param can be used to indicate a comma-separated list of aggregates to use. If the param value
you specify is of the form /aggr.+/ then pick_aggregates() will use perl regexp matching on the 
string in between the '/' characters

=item SOURCE

This param is used to indicate the source path for additional files needed by 
the test.

=item USERID

The user id that should be used for a test.

=item GROUPID

The group id that should be used for a test.

=item share

The name of the export/share to use for the test.

=back

=head1 SEE ALSO

L<Options>
L<TestManager>

=head1 BUGS

none 

=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: $
  Last revised $Date: $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;
