#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Params/Importer.pm#1 $
## @summary     Module to import, and process parameter definitions into Options.pm
## @author      cwilcox@netapp.com
## @pod here

=head1 NAME

Params::Importer - Module to import and process parameter definitions into Options.pm

=head1 SYNOPSIS
This package will be used in the libraries that reside in Params. (Volume.pm, Common.pm, etc.)
Example of how to explicitly specify which parameters to import using an abbreviated testcase and Params::Volume.pm 

B<Testcase>:
#grab only this definition
use Params::Volume qw(VOLUME);

B<Params::Volume.pm>:
use Params::Importer;

sub import {
my $me = shift;
#definitions
my @defs =   (
              {
              name => "volume",
              spec => "s@",
              description => "list of comma-separated volumes",
              },
              {
              name => "VOLUME",
              spec => "s@",
              description => "list of comma-separated volumes",
              },
             );
#compatability check
my %compats = (
                VOLUME => "volume",
                FILER_VOL => "volume",
              );

#grab only these testcase-indicated import params... if specified they will be included in the argument array after the class is shifted off
my @import_params = @_;

#Call the importer subroutine and hand-over parameter information to Params::Importer for Options processing.    
Params::Importer::importer {
           definitions   => \@defs,
           compatability => \%compats,
           import_params => \@import_params,
}

=head1 DESCRIPTION

This package is used to modularize the functionality for Options processing, and enables explicit import of testcase-specified parameters. 
It is used as the instrument to execute calls to Options.pm B<parameters>, B<compatability>, and B<process> methods during parameter import.

=head1 METHODS

B<importer> Receives the B<definitions>, B<compatability> names, and any explicitly named parameters to import (B<import_params>). This module now becomes the instrument to execute the calls to Options methods that occur during import. This will offload the Options processing functionality from the Params::modules; thereby fulfilling their original intended purpose of defining commonly used parameters within their different niches. It also enables a more modular approach to importing parameters, and increases code sharing.

=cut

package Params::Importer;
use Options;
use Data::Dumper;
use TracePkg;

$Trace = new TracePkg;
$Options = new Options;

sub importer {
    $Trace->entry();
    my $params = shift; 
    
    # Create references to definitions array, import_params array, compatibility hash
    my $defs = $params->{definitions};
    my $compat = $params->{compatability};
    my $import_params = $params->{import_params};
    my $num_import_params = scalar(@$import_params);
   
    my @params_array;
    my @defs_array;
    # Create a definitions name array to compare for compatability.
    for my $def_name (@$defs) {
        push (@defs_array, $def_name->{name});
    };
    # %def_name_hash is used to store "used" params, and separates them from "unused" params. 
    my @import_defs;
    my %import_compat;

    # Check to see if we received any explicit import parameters.
    # If we do, use them to check against the compatability hash, and store their definitions to import into Options.pm
    # If no explicit import parameters, use the complete definitions array of hashes.
    if ($num_import_params > 0) {
        @params_array = @$import_params;
        $Trace->debug(msg=>"We are using import params: ".Dumper(\@params_array));
    } else {
        @params_array = @defs_array;
        $Trace->debug(msg=>"We are using all the params: ".Dumper(\@params_array));
    }
    
    # set the "used" compatability values where applicable.
    for my $param_name (@params_array) {
        if (defined $compat->{$param_name}) {
            $import_compat{$param_name} = $compat->{$param_name};
        }
        elsif (!grep(/^$param_name$/, @defs_array)) {
            $Trace->fatal(msg=>"Param $param_name is not a valid parameter in your script's <use> statement.\nPlease check for spelling, and usage errors.", result=>"SCRIPT");
        }     
    }

    # Iterate through the definitions array of hashes, and push the "used" parameter definitions into an @import_defs array to be 
    # passed into parameters.
    my @compat_array = %import_compat;

    for my $definitions (@$defs) {
        my $name = $definitions->{name};
        if ((grep(/^$name$/, @params_array)) || (grep (/^$name$/, @compat_array))) {
            push (@import_defs, $definitions);
        }
    }

    # Import the parameter definitions into Options.pm
    $Trace->debug(msg=>"Here is the import_defs array: ".Dumper(\@import_defs));
    $Options->parameters(definitions =>[@import_defs]);
    
    # Specifies the override values for acceptable alternate param names  
    $Trace->debug(msg=>"Here is the import_compat hash: ".Dumper(\%import_compat));
    $Options->compatability(definitions =>{%import_compat});

    $Trace->exit();
}

1;
