#
# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of libraries to work with parameters
##
## @author latif@netapp.com
## @status public
## @pod here

package ParamsMgr;

# Compiler directives.
use strict;

# Module import.
use Tharn;
use NATE::BaseException qw(:try);
use NATE::ParamSet;
use NATE::Log;

use Error;
use San::SanLog;
use San::SanUtil;
use SubIPC;

use File::Temp qw(tempfile);
use XML::LibXML;
use Tie::IxHash;
use Clone;

# Define local/global variables
# The variable $_IntObject is used to store an internal object that is NOT accessible
# by callers/users. This object is mainly used internal for storing/holding
# a NATE ParamSet, pointers to parameter and any errors that may have occurred.
# When a called wants to check error (or other things) on the package they simply
# need to call the package error function, this will check the internal object
# for any errors and return to the caller.
use vars qw(@ISA $_IntObject $Log);
$Log = NATE::Log::log_global();

@ISA = qw(Error SanLog);

# Exported global variables
###########################
$ParamsMgr::TRUE   = $SanUtil::TRUE;
$ParamsMgr::FALSE  = $SanUtil::FALSE;

# Create internal object to be used for this package
# This call is necessary and should NOT be deleted or moved.
# This will create the internal object and initialize necessary
# variables.
ParamsMgr->_new();

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    load_xml
# Objective: load provided XML file into the NATE Paramset.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub load_xml {
    my (%args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my ($xml_data);

    my $xml_file = delete($args{xml_file});
    $xml_file ||= undef;
    my $xml_string = delete($args{xml_string});
    $xml_string ||= undef;
    my $display_xml = delete($args{display_xml});
    if (not defined($display_xml)) {
        $display_xml = 1;
    }
    my $display_xml_runid = delete($args{display_xml_runid});
    $display_xml_runid ||= "XML-environment";

    my $global_set  = (defined $args{paramset})?delete $args{paramset}:&ParamsMgr::get_global_paramset();

    # Second validation: Check whether an XML file is well formed or not.
    # We don't give a hoot about missing or incorrect elements/attributes here.
    $_IntObject->{xml_parser} = new XML::LibXML;

    # parse the xml file if it exists
    if (defined($xml_file)) {
        unless(-e $xml_file) {
            $_IntObject->_set_err("Could not find XML file: $xml_file");
            return $ParamsMgr::FALSE;
        }

        # Try to parse the file and set the error if the document is invalid.
        try {
            $xml_data = $_IntObject->{xml_parser}->parse_file($xml_file);
        } otherwise {
            my $exception = $@;
            $_IntObject->_set_err("XML file is malformed. Check for missing tags\n" . $exception);
            $xml_data = "";
        };
    } elsif (defined($xml_string)) {
        # otherwise, parse the xml string passed in
        # Try to parse the string and error if the document is invalid
        try {
            $xml_data = $_IntObject->{xml_parser}->parse_string($xml_string);
        } otherwise {
            my $exception = $@;
            $_IntObject->_set_err("XML string is malformed. Check for missing tags\n" . $exception);
            $xml_data = "";
        };
    } else {
        $_IntObject->_set_err("Either -xml_file or -xml_string must be specified");
    }

    # clean up the parser object
    undef $_IntObject->{xml_parser};

    # Check to see if we have any errors to return
    if (&ParamsMgr::error()) {
        return $ParamsMgr::FALSE;
    }

    # display the processed XML in a nicely formatted subtest
    if ($display_xml) {
        &ParamsMgr::_display_xml(runid => $display_xml_runid,
                                 xml_data => $xml_data);
        if (&ParamsMgr::error()) {
            $_IntObject->_set_err("Failed to display XML");
            return $ParamsMgr::FALSE;
        }
    }

    # load the xml document into the nate paramset
    &ParamsMgr::_load_xml_into_nate(xml_data => $xml_data, paramset => $global_set);
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err("Failed to load xml into NATE env");
        return $ParamsMgr::FALSE;
    }

    return $ParamsMgr::TRUE;
}

###############################################################################
# Method:    validate_and_process_xml
# Objective: Validate XML with provided XML schema and generate a Post schema
#            validation XML infoset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_and_process_xml {
    my (%args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my $xml_file = delete($args{xml_file});
    if (!$xml_file) {
        $_IntObject->_set_err("Missing required parameter 'xml_file'");
        return "";
    }
    my $schema_files = delete($args{schema_files});
    if (!$schema_files or ref($schema_files) ne 'ARRAY') {
        $_IntObject->_set_err("Missing or invalid parameter 'schema_files'");
        return "";
    }
    my $root_element = delete($args{root_element});
    $root_element ||= "scenario";

    $_IntObject->sanlog(-msg => "Validating XML\nXML File:\n\t$xml_file\nDTD Files:\n\t" . join("\n\t", @{$schema_files}),
                        -msg_level => 0);

    if (not -e $xml_file) {
        $_IntObject->_set_err("The XML file '$xml_file' does not exist.");
        return "";
    }

    #############################################################
    # (1) Build the temporary DTD schema file from all the schemas
    #############################################################

    my $element_hash;
    my $attr_hash;
    my $insert_elements_hash;
    tie(%{$insert_elements_hash}, "Tie::IxHash");

    # for each schema file, create a datastructure that will maintain
    # element & attribute overrides from overlapping schemas
    foreach my $schema_file (@{$schema_files}) {
        # open this schema for reading
        my $schema_fh;
        if (not open($schema_fh, "<", $schema_file)) {
            $_IntObject->_set_err("Unable to open schema file: '$schema_file' for reading");
            return "";
        }

        my $line_count = 0;
        my $doc_block = 0;

        # read each line of the schema and store in the datastructure
        while (my $line = <$schema_fh>) {
            $line_count++;
            # get rid of preceeding and trailing spaces on the line
            chomp($line);
            $line =~ s/^\s+//;
            $line =~ s/\s+$//;

            # This code block will skip any lines surrounded by =start_doc and =end_doc.
            # These are used to provide documentation for the DTD elements and attributes.
            if ($line =~ /^\=end_doc/i) {
                $doc_block = 0;
                next;
            } elsif ($line =~ /^\=start_doc/i) {
                $doc_block = 1;
                next;
            }

            next if ($doc_block == 1);

            # if it's an element tag, then store it in the element hash
            if ($line =~ /^\s*<\!ELEMENT\s+(\S+)\s+(.+?)>$/) {
                my $element_name = $1;
                my $element_children = $2;
                $element_hash->{$element_name} = $element_children;
            } elsif ($line =~ /^\s*<!ATTLIST\s+\((\S+)\)\s+(\S+)\s+CDATA\s+(\".*?\"|\#\S+|\#FIXED\s+\".*?\")\s+FORMAT\s+(.+?)\s*>\s*$/) {
                # if it's a multi element attribute assignment, store the attribute and value for each element
                # We also have a format tag.
                my @element_names = split(/,/, $1);
                my $attribute_name = $2;
                my $value = "CDATA " . $3;
                my $format = $4;
                my $format_string = $format;

                if ($format =~ /^\s*\/(.+)\/(\S*)/) {
                    # Format was found and it looks to be a regular expression
                    $format = qr/(?$2)$1/;
                } else {
                    $_IntObject->_set_err("Format on attribute $attribute_name of elements @element_names was not parsable");
                    return "";
                }

                foreach my $element_name (@element_names) {
                    $attr_hash->{$element_name}->{$attribute_name} = undef;
                    $attr_hash->{$element_name}->{$attribute_name}->{default} = $value;
                    $attr_hash->{$element_name}->{$attribute_name}->{format} = $format;
                    $attr_hash->{$element_name}->{$attribute_name}->{format_string} = $format_string;
                }
            } elsif ($line =~ /^\s*<!ATTLIST\s+\((\S+)\)\s+(\S+)\s+(.+)>$/) {
                # if it's a multi element attribute assignment, store the attribute and value for each element
                my @element_names = split(/,/, $1);
                my $attribute_name = $2;
                my $value = $3;
                foreach my $element_name (@element_names) {
                    $attr_hash->{$element_name}->{$attribute_name} = undef;
                    $attr_hash->{$element_name}->{$attribute_name}->{default} = $value;
                    $attr_hash->{$element_name}->{$attribute_name}->{format} = undef;
                }
            } elsif ($line =~ /^\s*<!ATTLIST\s+(\S+)\s+(\S+)\s+CDATA\s+(\".*?\"|\#\S+|\#FIXED\s+\".*?\")\s+FORMAT\s+(.+?)\s*>\s*$/) {
                # It's a single attribute assignment tag, store for just the single element
                # We also have a format tag.
                my $element_name = $1;
                my $attribute_name = $2;
                my $value = "CDATA " . $3;
                my $format = $4;
                my $format_string = $format;

                if ($format =~ /^\s*\/(.+)\/(\S*)/) {
                    # Format was found and it looks to be a regular expression
                    $format = qr/(?$2)$1/;
                } else {
                    $_IntObject->_set_err("Format on attribute $attribute_name of element $element_name was not parsable");
                    return "";
                }

                $attr_hash->{$element_name}->{$attribute_name} = undef;
                $attr_hash->{$element_name}->{$attribute_name}->{default} = $value;
                $attr_hash->{$element_name}->{$attribute_name}->{format} = $format;
                $attr_hash->{$element_name}->{$attribute_name}->{format_string} = $format_string;
            } elsif ($line =~ /^\s*<!ATTLIST\s+(\S+)\s+(\S+)\s+(.+)>$/) {
                # otherwise, it's a single attribute assignment tag, store for just the single element
                my $element_name = $1;
                my $attribute_name = $2;
                my $value = $3;
                $attr_hash->{$element_name}->{$attribute_name} = undef;
                $attr_hash->{$element_name}->{$attribute_name}->{default} = $value;
                $attr_hash->{$element_name}->{$attribute_name}->{format} = undef;
            } elsif ($line =~ /^\s*<!INSERT_ELEMENT\s+(\S+)\s+\((.+?)\)>$/) {
                # if this is an insert element, store all the elements that need
                # this element inserted into.
                my $insert_element = $1;
                my @elements_to_insert_into = split(/,/, $2);
                $insert_elements_hash->{$insert_element} = \@elements_to_insert_into;
            } elsif ($line =~ /^\s*<!DELETE_ELEMENT\s+(\S+?)>$/) {
                my $element_to_delete = $1;
                delete($element_hash->{$element_to_delete});
                delete($attr_hash->{$element_to_delete});
            } elsif ($line =~ /^\s*<!CLONE_ELEMENT\s+(\S+)\s+FROM\s+(\S+)>$/) {
                my $destination_element = $1;
                my $source_element = $2;
                # clone the source element's data into the destination element
                # in both the element hash and the attr hash
                $element_hash->{$destination_element} = $element_hash->{$source_element};
                foreach my $attr (keys %{$attr_hash->{$source_element}}) {
                    map { $attr_hash->{$destination_element}->{$attr}->{$_} = $attr_hash->{$source_element}->{$attr}->{$_} } keys %{$attr_hash->{$source_element}->{$attr}};
                }
            } elsif ($line =~ /^\s*<!UNDEF_ATTLIST\s+(\S+)>$/) {
                my $element_to_undef_attlist = $1;

                # set all the attribute values for this element to "#IMPLIED"
                # this will force all the attributes for this element to not
                # be pushed into the final xml string for this element if there is no value
                # defined for it
                foreach my $attr (keys %{$attr_hash->{$element_to_undef_attlist}}) {
                    $attr_hash->{$element_to_undef_attlist}->{$attr}->{default} = "CDATA #IMPLIED";
                }
            } elsif ($line =~ /^\s*#/) {
                # accept "comments" in the dtd files
            } elsif ($line =~ /^\s*$/) {
                # empty lines are acceptable too
            } else {
                # catch all lines that were not parsable
                $_IntObject->_set_err("Invalid syntax in DTD File: '$schema_file:$line_count' : '$line'");
                return "";
            }
        }

        # close out the schema file
        close($schema_fh);
    }

    # generate a temporary file to use as the DTD file we will validate against
    my ($tmp_dtd_fh, $tmp_dtd_file_name) = tempfile();

    # flush out all the elements and attributes into the temporary DTD file
    foreach my $element (keys %{$element_hash}) {
        my $element_children = $element_hash->{$element};
        my $element_str = "<!ELEMENT $element $element_children>";
        print $tmp_dtd_fh $element_str . "\n";
        foreach my $attr (keys %{$attr_hash->{$element}}) {
            my $attr_value = $attr_hash->{$element}->{$attr}->{default};
            my $attr_str = "<!ATTLIST $element $attr $attr_value>";
            # write this attr to the tmp dtd file
            print $tmp_dtd_fh $attr_str . "\n";
        }
    }

    ######################################################
    # (2) clone the XML file passed in and insert the dtd
    # header at the top of the file, create a temp
    # xml_file for this. Process all PUSHELEMENTS by
    # dynamically modifying the xml input
    ######################################################

    # first generate the primary doctype header with the root element
    my @dtd_header;
    push(@dtd_header, "<!DOCTYPE ${root_element} [");

    # insert the entity tag for the temp dtd file
    push(@dtd_header, "<!ENTITY \% ${root_element} SYSTEM \"${tmp_dtd_file_name}\">");
    push(@dtd_header, "\%" . $root_element . ";");
    push(@dtd_header, "]>");

    # open the xml file for reading
    my $old_xml_fh;
    if (not open($old_xml_fh, "<", $xml_file)) {
        $_IntObject->_set_err("Unable to open XML File: '$xml_file' for reading");
        return "";
    }

    # read in the xml file and convert it to a string
    my $xml_string = "";
    while (my $line = <$old_xml_fh>) {
        chomp($line);
        $xml_string .= $line . "\n";
    }

    # remove all commented lines from the xml file
    $xml_string =~ s/<!--(.*?)-->//gs;

    # convert all <include_xml/> tags with the contents of each file they point to
    &ParamsMgr::_process_xml_includes(\$xml_string);
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err("Unable to process XML Includes");
        return "";
    }

    # its possible that include xml files contain comments,
    # so remove any that may be lingering around
    $xml_string =~ s/<!--(.*?)-->//gs;

    # process all dtd insert elements into the xml string
    # this will dynamically update the xml string with all insert elements that
    # are missing
    my @insert_element_keys = keys %{$insert_elements_hash};
    &ParamsMgr::_process_dtd_insert_elements(\$xml_string,
                                             $insert_elements_hash);
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err("Unable to process DTD Insert Elements");
        return "";
    }

    # join the xml string back into an array
    my @processed_xml_file = split("\n", $xml_string);
    undef $xml_string;

    # create a temporary xml file to work with
    my ($tmp_xml_fh, $tmp_xml_file_name) = tempfile();

    # flush out the dtd header into the temp xml file
    foreach my $line (@dtd_header) {
        print $tmp_xml_fh $line . "\n";
    }

    # flush out the processed xml file into the temp xml file
    foreach my $line (@processed_xml_file) {
        # skip the xml tag if it is present, we won't be needing it
        next if ($line =~ /<\?xml|\?>/);

        # write this xml line into the temp file
        chomp($line);
        print $tmp_xml_fh $line . "\n";
    }

    # close out all open files
    close($tmp_xml_fh);
    close($old_xml_fh);
    close($tmp_dtd_fh);

    ############################################################
    # (3) Run xmllint --postvalid --dtdattr $tmp_xml_file_name
    # on the nate control server, and capture the output and
    # exit status. We will use the XML output and feed that into
    # XML::LibXML to parse it.
    ############################################################
    my ($status, @output) = &SanUtil::nate_server_execute(-cmd => "xmllint --postvalid --dtdattr ${tmp_xml_file_name}",
                                                          -show_output => 0);
    if ($status != 0) {
        # capture the relevent errors from the output that were caught from
        # xml lint if the status was bad
        my @errors = grep(/^${tmp_xml_file_name}\:\d+:/, @output);
        if (grep(/\d+: parser error|xmllint: fatal/, @output)) {
            push(@errors, @output);
        }
        my $errstring = join("\n", @errors);
        $_IntObject->_set_err("XML Schema Validation failed:\n$errstring");
        return "";
    }

    # look to see if a dtd child element setting is invalid
    my @bad_dtd_string = grep(/validity error : Content model of client is not determinist:/i, @output);
    if (scalar(@bad_dtd_string)) {
        $_IntObject->_set_err("XML Schema Validation failed, invalid DTD sub-element syntax\n@bad_dtd_string");
        return "";
    }

    # create a giant xml string of the output generated by xmllint
    my $new_xml_string = join("", @output);
    # clean up after ourselves
    undef @output;

    # delete all empty insert elements that may exist
    foreach my $insert_elm (@insert_element_keys) {
        $new_xml_string =~ s/<$insert_elm\/>//g;
    }

    ############################################################
    # (4) At this point, the Schema validation succeeded,
    # Perform all Tharn::param substitutions in the XML and
    # Validate the attribute values with the DTD attr formats
    ############################################################

    # interpolate and substitute in any replacement tharn variables in the
    # xml string
    # users can specify variables as "...${PARAM}..." or "$PARAM"
    &ParamsMgr::_process_tharnparams(xml_string => \$new_xml_string);
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err("Process Tharn Params failed");
        return $ParamsMgr::FALSE;
    }

    # place a newline after every closing tag so that all tags are isolated on their
    # own line.
    $new_xml_string =~ s/>\s*</>\n</g;

    # split the xml string by newlines so every tag is a single element
    my @temp_xml_file = split("\n", $new_xml_string);
    undef $new_xml_string;

    # handle all enumeration elements.
    # using eval here, because this is a recursive function
    # and error handling isnt pretty when using _set_err() recursively
    my @post_enumeration_xml_file;
    eval {
        @post_enumeration_xml_file = &ParamsMgr::_process_enumeration_elements(@temp_xml_file);
    };
    if ($@) {
        $_IntObject->_set_err("Enumeration failed: $@");
        return "";
    }

    undef @temp_xml_file;

    # join the post enumeration xml file back into a string
    my $final_xml_string = join("\n", @post_enumeration_xml_file);

    undef @post_enumeration_xml_file;

    # scan through the xml one more time for &&{ } parameters that have already been interpolated
    &ParamsMgr::_process_tharnparams(xml_string => \$final_xml_string,
                                     already_interpolated => 1);
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err("Process Tharn Params failed");
        return $ParamsMgr::FALSE;
    }

    # put every tag on its own line
    $final_xml_string =~ s/>\s*</>\n</g;

    # Make a call to validate the xml attribute values.  If the DTD file
    # has a FORMAT regex, that will be validated against the actual xml value
    # of the attribute.
    &ParamsMgr::_validate_xml_attr_values(xml_string => \$final_xml_string,
                                          attr_hash  => $attr_hash);
    if (&ParamsMgr::error()) {
        $_IntObject->_set_err("XML Attribute Value Validation failed.  One or more attributes " .
                              "had invalid values");
        return "";
    }

    return $final_xml_string;
}

sub _process_tharnparams {
    my (%args) = @_;

    my $xml_string = delete($args{xml_string});
    my $already_interpolated = delete($args{already_interpolated});
    $already_interpolated ||= 0;
    if (!$xml_string) {
        $_IntObject->_set_err("Missing required parameter 'xml_string'");
        return $ParamsMgr::FALSE;
    }

    # Look for variables in the XML that are either formats:
    # ${PARAM}
    # $PARAM
    # &&{INTERPED_PARAM}
    # &&INTERPED_PARAM
    while ($$xml_string =~ /\$(\{\S+?\})/g or $$xml_string =~ /"\$(\S+)"/g or
          ($already_interpolated and ($$xml_string =~ /\&\&(\{\S+?\})/g or $$xml_string =~ /"\&\&(\S+)"/g))) {
        my $tharn_var = $1;

        # Save a copy of the tharn variable in it's unmodified state for later use
        my $tharn_var_unmod = $tharn_var;

        # get rid of braces from the param name only if starts and ends with a open and close
        if ($tharn_var =~ /^\s*\{/ and $tharn_var =~ /\}\s*$/) {
            $tharn_var =~ s/\{|\}$//g;
        }

        # If the parameter has a default value specified, grab it
        my $default = undef;
        if ($tharn_var =~ /^(\S+)\,(.*)$/) {
            ($tharn_var, $default) = ($1, $2);
        }

        # skip if this tharn var is %\S+ (this is special syntax that allows for
        # dynamic tharnhost file lookups
        if ($tharn_var =~ /^%\S+/) {
            # if the tharn var was in the format ${...}, then replace it in this fashion
            if ($tharn_var_unmod =~ /\{|\}/) {
                $$xml_string =~ s/\$\Q$tharn_var_unmod\E/\&\&$tharn_var_unmod/g;
            } else {
                # otherwise anchor it with quotes to prevent unwanted substitutions
                $$xml_string =~ s/"\$\Q$tharn_var_unmod\E"/"\&\&$tharn_var_unmod"/g;
            }

            # skip this tharn var, it will be dealt with later
            next;
        }

        # Check to see if the parameter is dot notation
        # if it is, search the tharnhost file for the parameter
        if ($tharn_var =~ /^(\S+)\.(\S+)$/) {
            my ($host, $param) = ($1, $2);
            my $values;

            my $name;
            if ($already_interpolated) {
                # then the $host value has already been interpolated,
                $name = $host;
            } else {
                # interpolate the tharn parameter $host
                $name = &Tharn::param("$host");
            }

            # Eval host call so that it does not blow up
            eval {
                $values = &Tharn::host($name);
            };

            # If no error occurred during call of host() process the values
            if (not $@ and exists($values->{$param})) {
                my $value = $values->{$param};
                # Substitute value for dot variable in XML
                $Log->comment("Overriding \'$tharn_var_unmod\' with value \'$value\'");

                # if the tharn var was in the format ${...} or &&{...}, then replace it in this fashion
                if ($tharn_var_unmod =~ /\{|\}/) {
                    $$xml_string =~ s/\$\Q$tharn_var_unmod\E/$value/g;
                    $$xml_string =~ s/\&\&\Q$tharn_var_unmod\E/$value/g if ($already_interpolated);
                } else {
                    # otherwise anchor it with quotes to prevent unwanted substitutions
                    $$xml_string =~ s/"\$\Q$tharn_var_unmod\E"/"$value"/g;
                    $$xml_string =~ s/"\&\&\Q$tharn_var_unmod\E"/"$value"/g if ($already_interpolated);
                }
                next;
            } elsif (defined($default)) {
                my $value = $default;
                # Substitute value for dot variable in XML
                $Log->comment("Overriding \'$tharn_var_unmod\' with default value \'$value\'");

                # if the tharn var was in the format ${...}, then replace it in this fashion
                if ($tharn_var_unmod =~ /\{|\}/) {
                    $$xml_string =~ s/\$\Q$tharn_var_unmod\E/$value/g;
                    $$xml_string =~ s/\&\&\Q$tharn_var_unmod\E/$value/g if ($already_interpolated);
                } else {
                    # otherwise anchor it with quotes to prevent unwanted substitutions
                    $$xml_string =~ s/"\$\Q$tharn_var_unmod\E"/"$value"/g;
                    $$xml_string =~ s/"\&\&\Q$tharn_var_unmod\E"/"$value"/g if ($already_interpolated);
                }

                next;
            }
        }

        my $tharn_value = &Tharn::param("$tharn_var", '-default', $default);
        $Log->comment("Overriding \'$tharn_var_unmod\' with value \'$tharn_value\'");

        # if the tharn var was in the format ${...}, then replace it in this fashion
        if ($tharn_var_unmod =~ /\{|\}/) {
            $$xml_string =~ s/\$\Q$tharn_var_unmod\E/$tharn_value/g;
            $$xml_string =~ s/\&\&\Q$tharn_var_unmod\E/$tharn_value/g if ($already_interpolated);
        } else {
            # otherwise anchor it with quotes to prevent unwanted substitutions
            $$xml_string =~ s/"\$\Q$tharn_var_unmod\E"/"$tharn_value"/g;
            $$xml_string =~ s/"\&\&\Q$tharn_var_unmod\E"/"$tharn_value"/g if ($already_interpolated);
        }
    }

    return $ParamsMgr::TRUE;
}

###############################################################################
# Method:    search
# Objective: Search through available data for requested items using
#            search path and search type provided by caller.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub search {
    my (%args) = @_;

    my ($modified_search_path, $value);

    my $search_path       = exists ($args{search_path})      ? $args{search_path}      :
                            exists ($args{-search_path})     ? $args{-search_path}     :
                                                               [];
    my $search_type       = exists ($args{search_type})      ? $args{search_type}      :
                            exists ($args{-search_type})     ? $args{-search_type}     :
                                                               "";
    my $exact_key         = exists ($args{exact_key})        ? $args{exact_key}        :
                            exists ($args{-exact_key})       ? $args{-exact_key}       :
                                                               0;
    my $key                 = exists ($args{key})              ? $args{key}               :
                            exists ($args{-key})             ? $args{-key}               :
                                                               undef;
    my $key_filter         = exists ($args{key_filter})       ? $args{key_filter}       :
                            exists ($args{-key_filter})      ? $args{-key_filter}      :
                                                               undef;
    my $filter_match         = exists ($args{filter_match})     ? $args{filter_match}     :
                            exists ($args{-filter_match})    ? $args{-filter_match}    :
                                                               1;
    my $lowercase_keys    = exists ($args{lowercase_keys})   ? $args{lowercase_keys}   :
                            exists ($args{-lowercase_keys})  ? $args{-lowercase_keys}  :
                                                               0;
    my $eval_values         = exists ($args{eval_values})         ? $args{eval_values}      :
                            exists ($args{-eval_values})     ? $args{-eval_values}     :
                                                               0;
    my $no_private_keys   = exists ($args{no_private_keys})  ? $args{no_private_keys}  :
                            exists ($args{-no_private_keys}) ? $args{-no_private_keys} :
                                                               0;
    if (not defined($key)) {
        $_IntObject->_set_err("ParamsMgr::search required argument 'key' was not passed in");
        return {};
    }

    my $global_set  = (defined $args{paramset})?delete $args{paramset}:&ParamsMgr::get_global_paramset();

    $modified_search_path = [];

    if ($exact_key == 0) {
        # Search type is not provided, default search path to input plus modified
        if ($search_type eq "") {
            # Setup a search path as defined by ParamSet
            push(@{$modified_search_path}, "");
            foreach my $s_path (@{$search_path}) {
                push(@{$modified_search_path}, $s_path);
            }
        } elsif ($search_type =~ /NODE/i) {
            # Caller wants NODES, setup search for nodes, first the ROOT
            # then CLUSTER, then AGGR, the NODE(*) are under AGGR
            push(@{$modified_search_path}, "");
            push(@{$modified_search_path}, "CLUSTER");
            # Set the key to the parent of the NODE(*), which is AGGR
            $key = "AGGR";
        } elsif ($search_type =~ /CLIENT/i) {
            #multi client group configs can have root client tags in the form
            #of CLIENTS(\d)
            # Caller wants CLIENT(*), setup search for clients, first the ROOT
            # then CLIENTS, the CLIENT(*) are under CLIENTS
            push(@{$modified_search_path}, "");
            # Set the key to the parent of CLIENT(*), which is CLIENTS
            $key = "CLIENTS";
        } elsif ($search_type =~ /SWITCH/i) {
            # Caller wants SWITCH(*), setup search for switches, first the ROOT
            # then SWITCHES, the SWITCH(*) are under SWITCHES
            push(@{$modified_search_path}, "");
            # Set the key to the parent of SWITCH(*), which is SWITCHES
            $key = "SWITCHES";
        } elsif ($search_type =~ /CLUSTER/i) {
            # Caller wants CLUSTER(*), setup search for clusters, first the ROOT
            # then CLUSTERS, the CLUSTER(*) are under CLUSTERS
            push(@{$modified_search_path}, "");
            # Set the key to the parent of CLUSTER(*), which is CLUSTERS
            $key = "CLUSTERS";
        }
    }

    $value = &ParamsMgr::get($key, search => $modified_search_path, merge => 1, expand => 1, paramset => $global_set);
    if (&ParamsMgr::error()) {
        return {};
    }

    # Determine if $value is a hash, if not, this means the value was NOT found in the XML file
    # but in the environment. At this point we need to set an error and return.
    if (ref($value) ne "HASH") {
        return {};
    }

    if ($search_type =~ /NODE/i) {
        # If search_type is NODE(*) delete all else in the $value hash
        foreach my $key (keys(%{$value})) {
            if ($key !~ /NODE\d+/i) {
                delete($value->{$key});
            }
        }
    }

    # Checking for the three cases inside the loop before entering so that if
    # no work would be done in the loop then we skip the loop.
    if(defined($key_filter) or $no_private_keys == 1 or $lowercase_keys == 1) {
        # filter all the keys by the key filter
        # specified by the caller
        foreach my $key (keys %{$value}) {
            if(defined ($key_filter)) {
                # if this key doesnt match the key filter, delete it
                if ($filter_match and $key !~ $key_filter) {
                    delete($value->{$key});
                    next;
                } elsif (not $filter_match and $key =~ $key_filter) {
                    delete($value->{$key});
                    next;
                }
            }

            # if requested to not include private keys, remove them now
            if ($no_private_keys == 1 and $key =~ /^_/) {
                delete($value->{$key});
                next;
            }

            # if requested to lowercase all return keys, do so now.
            if ($lowercase_keys == 1) {
                $value->{lc($key)} = delete($value->{$key});
            }
        }
    }

    # if requested to eval the values of all the keys, do so now
    if ($eval_values == 1) {
        &ParamsMgr::eval_key_values($value);
    }

    return $value;
}

sub lc_keys {
    my ($value) = @_;
    foreach my $key (keys %{$value}) {
        $value->{lc($key)} = delete $value->{$key};
    }
}

sub eval_key_values {
    my ($value) = @_;
    no strict 'subs';
    
    foreach my $key (keys %{$value}) {
        my $old_value = $value->{$key};
        
        # dont eval values that are not strings
        next if (ref($old_value) ne "");
        
        # if the old value is wrapped around eval tags, eval it
        # Eval tags are ARRAYREF="[ ]" or HASHREF="{ }" or CODE="sub{ }" or REGEX="qr//"
        if ($old_value =~ /^\[.*\]$|^\{.*\}$|^\s*sub\s*\{.+\}\s*$|^\s*qr\/.*\/[imsxog]*\s*$/) {
            # store the eval'd version in the key
            $value->{$key} = eval($old_value);
            if ($@) {
                # if the eval died, set an error
                $_IntObject->_set_err($@);
                return {};
            }
        }
    }
    
    use strict;
    
    return $value;
}

###############################################################################
# Method:    walk
# Objective: Walks the paramset object using ParamSet::walk()
# Details:   See POD documentation at the end of this file.
###############################################################################
sub walk {
    my (%args) = @_;

    my @retval = ();

    my $global_set  = (defined $args{paramset})?delete $args{paramset}:&ParamsMgr::get_global_paramset();
    if (&ParamsMgr::error()) {
        return "";
    }

    # Use Exception handling for NATE ParamSet
    try {
        @retval = $global_set->walk(%args);
    } otherwise {
        my $exception = shift;
        $_IntObject->_set_err($exception->text());
        @retval = ();
    };

    return @retval;
}

###############################################################################
# Method:    get_global_paramset
# Objective: gets global paramset name space
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_global_paramset { $_IntObject->{nate_paramset}; }

###############################################################################
# Method:    clear_global_paramset
# Objective: clear global paramset name space
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clear_global_paramset {

    # Undef nate_paramset and construct new ParamSet
    $_IntObject->{nate_paramset} = undef;
    $_IntObject->{nate_paramset} = NATE::ParamSet->new();

    return $ParamsMgr::TRUE;
}

###############################################################################
# Method:    get
# Objective: gets parameter from global parameter name space
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get {
    my ($key, %args) = @_;

    my $retval = "";

    my $global_set  = (defined $args{paramset})?delete $args{paramset}:&ParamsMgr::get_global_paramset();
    if (&ParamsMgr::error()) {
        return "";
    }

    # Use Exception handling for NATE ParamSet
    try {
        $retval = $global_set->get($key, %args);
    } otherwise {
        my $exception = shift;
        $_IntObject->_set_err($exception->text());
        $retval = "";
    };

    return $retval;
}

###############################################################################
# Method:    set
# Objective: sets parameter and value in global name space
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set {
    my (%args) = @_;

    my $retval = $ParamsMgr::FALSE;

    my $global_set  = (defined $args{paramset})?delete $args{paramset}:&ParamsMgr::get_global_paramset();
    if (&ParamsMgr::error()) {
        return $ParamsMgr::FALSE;
    }

    try {
        # allow_replace=>1
        $args{'allow_replace'} = 1;
        $global_set->set(%args);
        $retval = $ParamsMgr::TRUE;
    } otherwise {
        my $exception = shift;
        $_IntObject->_set_err($exception->text());
        $retval = $ParamsMgr::FALSE;
    };

    return $retval;
}

###############################################################################
# Method:    unset
# Objective: unsets the parameter from the global parameter name space
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unset {
    my ($key, %args) = @_;

    my $retval = "";

    my $global_set  = (defined $args{paramset})?delete $args{paramset}:&ParamsMgr::get_global_paramset();
    if (&ParamsMgr::error()) {
        return "";
    }

    # Use Exception handling for NATE ParamSet
    try {
        $retval = $global_set->unset($key, %args);
    } otherwise {
        my $exception = shift;
        $_IntObject->_set_err($exception->text());
        $retval = "";
    };

    return $retval;
}

###############################################################################
# Method:    clone
# Objective: clone current ParamSet and return back clone
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clone {
    my %args = @_;
    my $retval = {};

    my $global_set  = (defined $args{paramset})?delete $args{paramset}:&ParamsMgr::get_global_paramset();
    if (&ParamsMgr::error()) {
        return $retval;
    }

    try {
        $retval = $global_set->clone();
    } otherwise {
        my $exception = shift;
        $_IntObject->_set_err($exception->text());
        $retval = {};
    };

    return $retval;
}

################################################################################
# Method:    process_args
# Objective: Create temporary data structure with args on the given object
# Inputs:    Object to build data structure on
#            Hash of input args to process
# Returns:   $ParamsMgr::TRUE when completed properly
################################################################################
sub process_args {
    my (%args) = @_;

    &ParamsMgr::clear_error();

    %args = &SanUtil::strip_dash(%args);

    # Determine if the caller wants to add arguments under different caller name
    my $caller = $_IntObject->_parent_caller();
    if (defined($args{'CALLER'})) {
        # Set caller to the user requested value and delete the argument from the
        # input arguments hash.
        $caller = $args{'CALLER'};
        delete($args{'CALLER'});
    }

    # If the caller attribute is already defined, delete it so it can be created
    # with the new arguments being passed in.
    if (defined($_IntObject->{_args}->{$caller})) {
        delete($_IntObject->{_args}->{$caller});
    }

    # Go through the arguments and build the data structure on the object provided.
    foreach my $arg (keys(%args)) {
        $_IntObject->{_args}->{$caller}->{$arg} = $args{$arg};
    }

    return $ParamsMgr::TRUE;
}

################################################################################
# Method:    get_arg
# Objective: Return argument value requested
# Inputs:    -arg => argument to retrieve
#            -required => 1 | 0 if the argument is required or not
#            -default => The default value to return if argument is not found
# Returns:   __ERROR__ if argument is not found or hit error attempting to
#            locate argument.
#            Returns argument value if argument is found.
#            Returns default argument value if argument is not found and
#            argument is not -required=>1
################################################################################
sub get_arg {
    my (%args) = @_;

    # Can't call ParamsMgr::get_arg, circular reference
    my $arg            = exists($args{'arg'})             ? $args{'arg'} :
                                                            $args{'-arg'};
    my $required       = exists($args{'required'})        ? $args{'required'} :
                         exists($args{'-required'})       ? $args{'-required'} :
                                                            0;
    my $default        = exists($args{'default'})         ? $args{'default'} :
                                                            $args{'-default'};
    my $ref_type       = exists($args{'ref_type'})        ? $args{'ref_type'} :
                         exists($args{'-ref_type'})       ? $args{'-ref_type'} :
                                                            undef;
    my $default_to_env = exists($args{'default_to_env'})  ? $args{'default_to_env'} :
                         exists($args{'-default_to_env'}) ? $args{'-default_to_env'} :
                                                            0;
    my $dependant_on   = exists($args{'dependant_on'})    ? $args{'dependant_on'} :
                         exists($args{'-dependant_on'})   ? $args{'-dependant_on'} :
                                                            [];
    my $format         = exists($args{'format'})          ? $args{'format'} :
                         exists($args{'-format'})         ? $args{'-format'} :
                                                            qr/.*/;
    my $code_args      = exists($args{'code_args'})       ? $args{'code_args'} :
                         exists($args{'-code_args'})      ? $args{'-code_args'} :
                                                            undef;
    my $eval_code      = exists($args{'eval_code'})       ? $args{'eval_code'} :
                         exists($args{'-eval_code'})      ? $args{'-eval_code'} :
                                                            $ParamsMgr::FALSE;
    my $caller         = defined($args{'CALLER'})         ? $args{'CALLER'} :
                         defined($args{'-CALLER'})        ? $args{'-CALLER'} :
                                                            $_IntObject->_parent_caller();
    my $error_msg      = defined($args{'error_msg'})      ? $args{'error_msg'} :
                         defined($args{'-error_msg'})     ? $args{'-error_msg'} :
                                                            "Unable to locate required argument ($arg) for ($caller)";
    my $error          = "";
    my $retval         = "";

    $arg =~ s/^\-//;

    my $global_set  = (defined $args{paramset})?delete $args{paramset}:&ParamsMgr::get_global_paramset();

    # Check to see if the caller wants to verify that the arg's
    # dependants exist within the processed hash.
    # There are three scenarios that a caller can go into
    #    1) All "dependant on" do not exist
    #        This is not a failure case but an ignore case.
    #        The required flag is shut off (set to 0).
    #    2) One of the "dependant on" is not found
    #        This is a failure, the required flag is enabled (set to 1).
    #    3) All "dependant on" exist
    #        The caller's request will be processed as normal
    if (scalar(@{$dependant_on}) != 0) {
        my $count = 0;
        # Go through the list and check to see if the "dependant on" exists
        foreach my $dep_on (@{$dependant_on}) {
            if (defined($_IntObject->{_args}->{$caller}->{$dep_on})) {
                $count++;
            }
        }

        if ($count == 0) {
            # Found none of the args, set required to 0
            $required = 0;
        }
    }

    # Check to see if required arguments were passed in
    if (not defined($arg)) {
        $_IntObject->_set_err("Required parameter -arg missing");
        return undef;
    }

    if ($required == 0 and not exists($args{'default'}) and not exists($args{'-default'}) and $default_to_env == 0) {
        $_IntObject->_set_err("Optional parameter ($arg) requested but no default provided");
        return undef;
    }

    if (not defined($_IntObject->{_args}->{$caller}->{$arg}) and $default_to_env == 1) {
        # Not defined on the standard stack, look to see if it is defined
        # in the environment by using a recursive call to get_arg
        # Load up the parent parameters in order and see if the argument exists within them
        # Get the set of environment parameters from ParamsMgr
        my $environment = &ParamsMgr::search(search_path => [], key => "ENVIRONMENT", paramset => $global_set);
        if (&ParamsMgr::error()) {
            return undef;
        }

        # Process the environment arguments and then check to see if the arg caller
        # requested exists, if not set $retval to undefined
        &ParamsMgr::process_args(%{$environment}, -CALLER => $caller . "|" . time());
        delete($args{default_to_env});
        delete($args{-default_to_env});

        # recursively call this function with the environment as the args holder
        return &ParamsMgr::get_arg(%args, -CALLER => $caller . "|" . time());
    } elsif (not defined($_IntObject->{_args}->{$caller}->{$arg})) {
        if ($required == 0) {
            # Did not find the argument on the saved data structure
            # return the default specified by caller IF defined
            $retval = $default;
        } else {
            # Add any other errors from previous checks
            if ($error ne "") {
                $error = $error_msg . "\n" . $error;
            }

            $_IntObject->_set_err($error_msg);
            $retval = "";
        }
    } else {
        # Set the return value found
        $retval = $_IntObject->{_args}->{$caller}->{$arg};
    }

    # Check the format of the return, make sure it matches callers request
    # AND if ref type is string (or empty)
    if ($retval !~ $format and $ref_type eq "") {
        $_IntObject->_set_err("Requested format ($format) does NOT match format of the return ($retval)");
    }

    # Check to see if an error has occurred and set it
    if ($error ne "") {
        $_IntObject->_set_err($error_msg);
    } else {
        # Check to see if ref_type is what user requested
        if (defined($ref_type) and defined($retval)) {
            if (ref($retval) ne $ref_type) {
                $_IntObject->_set_err("Requested reference type ($ref_type) does not match requested " .
                                      "argument ($arg) of type (" . ref($retval) . ")");

                # Need to modify the return variable so that it matches the null value
                # for the reference type that was requested by the caller
                if ($ref_type eq "ARRAY") {
                    (wantarray()) ? $retval = () : $retval = [];
                } elsif ($ref_type eq "HASH") {
                    $retval = {};
                } elsif ($ref_type eq "CODE") {
                    $retval = sub {};
                } elsif ($ref_type eq "SCALAR") {
                    $retval = undef;
                } elsif ($ref_type eq "") {
                    $retval = "";
                } else {
                    $retval = undef;
                }
            }
        }
    }

    # if the caller wants an array
    # and the return is an array reference, then dereference the return
    if (ref($retval) eq "ARRAY") {
        if (wantarray()) {
            return @{$retval};
        }
    }

    # if the return value is a CODE reference, then execute the code reference
    if (ref($retval) eq "CODE" and $eval_code eq $ParamsMgr::TRUE) {
        return $retval->($code_args);
    }

    return $retval;
}

###############################################################################
# Method:    debug_level
# Objective: Set the debug level and return it to caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub debug_level {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $debug_level = &ParamsMgr::get_arg(-arg => "debug_level", -required => 0, -default => 0);
    if (&ParamsMgr::error()) {
        return 0;
    }

    $_IntObject->{debug_level} = $debug_level;

    return $debug_level;
}

###############################################################################
# Method:    sanlog
# Objective: Wrapper to SanLog::sanlog() function
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sanlog {
    my ($self, %args) = @_;

    my ($msg_type, $msg, $msg_level, $enter, $exit);

    &ParamsMgr::process_args(%args);
    $enter = &ParamsMgr::get_arg(-arg => "enter", -required => 0, -default => 0);
    $exit = &ParamsMgr::get_arg(-arg => "exit", -required => 0, -default => 0);
    if ($enter == 1 or $exit == 1) {
        # If enter or exit is either specified, msg is not required, if not defined
        # default it to an empty string
        $msg = &ParamsMgr::get_arg(-arg => "msg", -required => 0, -default => "");
    } else {
        $msg = &ParamsMgr::get_arg(-arg => "msg", -required => 1, -default => "");
    }

    $msg_type = &ParamsMgr::get_arg(-arg => "msg_type", -required => 0, -default => "");
    $msg_level = &ParamsMgr::get_arg(-arg => "msg_level", -required => 0, -default => 0);
    if (&ParamsMgr::error()) {
        return $ParamsMgr::FALSE;
    }

    # If the message is empty that could mean that the user specified an enter or exit flag
    # default it to _parent_caller
    if ($msg eq "") {
        $msg = $_IntObject->_parent_caller();
    } else {
        # Prepend the caller name/description to the msg
        $msg = $_IntObject->_parent_caller() . ": " . $msg;
    }

    # Determine if the caller called with an enter or exit request
    # Also set the msg_level to 3 if this is an enter/exit request
    if ($enter == 1) {
        $msg = "[Function Enter] " . $msg;
        $msg_level = 3;
    } elsif ($exit == 1) {
        $msg = "[Function Exit] " . $msg;
        $msg_level = 3;
    }

    # If the message type is warn then call SanUtil::nate_log with
    # warning type and message
    if ($msg_type =~ /warn/i) {
        $Log->warn($msg);
        return $ParamsMgr::TRUE;
    }

    # Call sanlog with the msg and msg_level
    $self->SanLog::sanlog(-msg => $msg, -msg_level => $msg_level);

    return $ParamsMgr::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _new
# Objective: Create new ParamsMgr object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($proto, %args) = @_;

    $_IntObject = {};

    my $class = ref($proto) || $proto;
    bless ($_IntObject, $class);

    # Set attributes on self for Error.pm
    #   error_dont_overwrite_errors - Append errors until cleared.
    #   error_stack_style - 'Normal' means place first error at top of stack
    $_IntObject->{error_dont_overwrite_errors} = 1;
    $_IntObject->{error_stack_style} = "normal";

    # Construct new ParamSet
    $_IntObject->{nate_paramset} = NATE::ParamSet->new();

    # Clear out errors
    &ParamsMgr::clear_error();

    return $_IntObject;
}

###############################################################################
# Method:    _paramset_xml_load
# Objective: Loads the XML data into the NATE paramset
# Details:   See POD documentation at the end of this file.
# Returns:   Returns true when completed
###############################################################################
sub _paramset_xml_load {
    my ($tree, $data, %args) = @_;

    my $global_set  = (defined $args{paramset})?delete $args{paramset}:&ParamsMgr::get_global_paramset();

    # If tree is not empty 'grow' tree by appending a '.' to provide caller with
    # required parameter naming standards (dot notation).
    if ($tree ne "") {
        $tree .= ".";
    }

    my @keys = keys(%{$data});
    foreach my $param (@keys) {
        my $data_param = $data->{$param};
        # If this key has no more children set the value
        if (ref($data_param) eq "HASH") {
            if ($param =~ /^(.*\D)1$/) {
                my $param_only = $1;
                my $multiple_keys_count = 0;
                my $multiple_keys_flag = 0;

                foreach my $key (@keys) {
                    if ($key =~ /^${param_only}\d+$/) {
                        $multiple_keys_count++;
                    }
                    if ($multiple_keys_count > 1) {
                        $multiple_keys_flag = 1;
                        last;
                    }
                }

                if ($multiple_keys_flag) {
                    # leave the count because there are multiples of this type
                    &ParamsMgr::_paramset_xml_load($tree . $param, $data_param, paramset => $global_set);
                } else {
                    # dont add the count because it is the only instance of this type
                    &ParamsMgr::_paramset_xml_load($tree . $param_only, $data_param, paramset => $global_set);
                }
            } else {
                # pass it in as is
                &ParamsMgr::_paramset_xml_load($tree . $param, $data_param,  paramset => $global_set);
            }
        } else {
            # enforce uppercase standards on parameter names
            my ($key, $value) = (uc("$tree$param"), $data_param);

            # Check to see if the parameter is a string or value that needs
            # to be further processed.
            if ($value =~ /^\$/) {
                # The key is a parameter that the value needs to be extracted
                # from the environment.
                # Strip off any leading '$' from the key name
                $value =~ s/^\$//;
                $value = &Tharn::param("$value");
            }

            &ParamsMgr::set(key => $key, value => $value, paramset => $global_set);
        }
    }

    # If the tree has a digit at the end of it, add an _INDEX key to keep track of the digit
    if ($tree =~ /(\d+)\.$/) {
        my ($key, $value) = (uc($tree) . "_INDEX", $1);

        &ParamsMgr::set(key => $key, value => $value, paramset => $global_set);
    }

    return $ParamsMgr::TRUE;
}

###############################################################################
# Method:    _validate_xml_attr_values
# Objective: Process the xml string and check against the FORMAT specification in
#            the dtd attr list.
# Inputs:    attr_hash    Attribute hash
#            xml_string   XML String
# Outputs: Returns $ParamsMgr::TRUE if no errors are found.  If errors are found
#          a report is printed, error is set and $ParamsMgr::FALSE is returned.
###############################################################################
sub _validate_xml_attr_values {
    my (%args) = @_;

    my $xml_string = delete($args{xml_string});
    if (!$xml_string) {
        $_IntObject->_set_err("Missing required parameter 'xml_string'");
        return $ParamsMgr::FALSE;
    }
    my $attr_hash = delete($args{attr_hash});
    if (!$attr_hash) {
        $_IntObject->_set_err("Missing required parameter 'attr_hash'");
        return $ParamsMgr::FALSE;
    }

    # Build Lookup Hash of elements in the configuration XML. The keys of the hash
    # are the elements and the value returned is an array reference to each instance
    # of that element in the XML configuration file.
    # $hash->{element} = array ref of all instances of that element in xml file
    my $xml_hash;
    while ($$xml_string =~ /<\s*(\w+?)\s+(.+?)\/?>/gx) {
        push (@{$xml_hash->{$1}}, $2);
    }

    my @errors = ();
    # Iterate over each element in the xml hash. The value of $element is an array reference
    # to an array containing every instance of the element in the XML configuration file.
    foreach my $element (sort keys %$xml_hash) {

        my @attrs_keys = keys %{$attr_hash->{$element}};
        my @sorted_attrs_keys = sort @attrs_keys;
        my @element_instances = @{$xml_hash->{$element}};

        # Iterate over each instance of the element in the XML configuration.
        foreach my $element_instance (@element_instances) {

            # Now that we have found the element in the xml hash,
            # iterate over each of it's attributes from the attribute
            # hash and see if they can be found in the xml element instance.
            foreach my $attr (@sorted_attrs_keys) {

                if ($element_instance =~ /\s${attr}=\"(.*?)\"/) {
                    my $value = $1;
                    # See if the attribute has a format defined for it and act accordingly
                    if (defined($attr_hash->{$element}->{$attr}->{format})) {
                        unless ($value =~ $attr_hash->{$element}->{$attr}->{format}) {
                            my $msg = "Invalid value '$value' for Element '$element' " .
                                      "attribute '$attr'.  Must satisfy regex of " .
                                      "'$attr_hash->{$element}->{$attr}->{format_string}'\n" .
                                      "xml_element='$element_instance'";
                            push (@errors, $msg);
                        }
                    } ## END if (defined($attr_hash->{$element}->{$attr}->{format})) {
                } ## END if ($element_instance =~ /\b${attr}=\"(.*?)\"\s*/) {
            } ## END foreach my $attr (sort keys %{$attr_hash->{$element}}) {
        } ## END foreach my $element_instance (@element_instances) {
    } ## END foreach my $element (sort keys %$xml_hash) {

    if (scalar(@errors) > 0) {
        my $msg = "\n" . join("\n", @errors);
        $Log->warn($msg);
        $_IntObject->_set_err("XML attribute validation failed");
        return $ParamsMgr::FALSE;
    }

    return $ParamsMgr::TRUE;
}

###############################################################################
# Method:    _process_node
# Objective: process XML node and populate a data structure
# Details:   See POD documentation at the end of this file.
# Returns:   This function will return a hash data structure that is populated
#            by nodes and their attributes.
###############################################################################
sub _process_node {
    my ($parent) = @_;

    # Get the nodes of the XML element ($parent)
    # Example:
    # <parent>
    #    <child> ... </child>
    # ...
    # </parent>
    my $data;
    tie(%{$data}, "Tie::IxHash");
    my @nodes = $parent->childNodes();

    my %seen_nodes;

    foreach my $node (@nodes) {
        my $node_name = $node->getName();
        if (($node_name eq "text") or ($node_name eq "#text")) {
            next;
        }

        $seen_nodes{$node_name}++;

        my $node_attr;
        # Determine if this node has child nodes to process
        if ($node->hasChildNodes()) {
            # Node has child nodes, recursively call _process_node on this node
            $node_attr = &ParamsMgr::_process_node($node);
        }

        my @atts = $node->getAttributes();
        foreach my $attr (@atts) {
            my ($attr_name, $attr_value) = ($attr->getName(), $attr->getValue());
            $node_attr->{$attr_name} = $attr_value;
        }

        # if the node name doesnt end with a digit, then append the count
        if ($node_name !~ /\d$/) {
            $data->{$node_name . $seen_nodes{$node_name}} = $node_attr;
        } else {
            # otherwise dont append the count
            $data->{$node_name} = $node_attr;
        }
    }

    return $data;
}

###############################################################################
# Method:    _load_xml_into_nate
# Objective: Given the XML data document, load it into the Nate ParamSet
###############################################################################
sub _load_xml_into_nate {
    my (%args) = @_;

    my $global_set  = (defined $args{paramset})?delete $args{paramset}:&ParamsMgr::get_global_paramset();

    my $xml_data = delete $args{xml_data};
    if (!$xml_data) {
        $_IntObject->_set_err("Missing required parameter 'xml_data'");
        return $ParamsMgr::FALSE;;
    }

    my $xml_root;
    try {
        # Get the root of the XML Document
        $xml_root = $xml_data->getDocumentElement();
    } otherwise {
        my $exception = shift;
        $_IntObject->_set_err("Error while executing getDocumentElement:\n" . $exception->text());
        $xml_data = "";
    };

    # Check to see if we have any errors to return
    if (&ParamsMgr::error()) {
        return $ParamsMgr::FALSE;
    }

    # Call _process_node starting at the beginning of the tree {}, using the data
    # gathered from getDocumentElement().
    my $xml_data_struct = &ParamsMgr::_process_node($xml_root);
    if (&ParamsMgr::error()) {
        return $ParamsMgr::FALSE;
    }

    # Load up the XML within the NATE paramset, start the "tree" input as an
    # empty string, starting from the beginning of the stack.
    &ParamsMgr::_paramset_xml_load("", $xml_data_struct, paramset => $global_set);
    if (&ParamsMgr::error()) {
        return $ParamsMgr::FALSE;
    }

    return $ParamsMgr::TRUE;
}

###############################################################################
# Method:    _process_dtd_insert_elements
# Objective: process all insert elements specified in DTD inputs
# Returns:   Returns updated XML string containing all insert elements
###############################################################################
sub _process_dtd_insert_elements {
    my ($xml_string, $insert_elements_hash) = @_;

    # bail if there is nothing to do
    return if (not scalar(keys %{$insert_elements_hash}));

    # due to a regex issue with the existence of special characters in the XML string
    # we need to replace them with a temporary sentinel, they will get restored
    # afterwards
    $$xml_string =~ s/\$/##DOLLARSIGN##/g;
    $$xml_string =~ s/\|/##PIPE##/g;
    $$xml_string =~ s/\(/##LPAREN##/g;
    $$xml_string =~ s/\)/##RPAREN##/g;
    $$xml_string =~ s/\^/##CARET##/g;
    $$xml_string =~ s/\*/##SPLAT##/g;
    $$xml_string =~ s/\!/##BANG##/g;
    $$xml_string =~ s/\?/##QUESTION##/g;
    $$xml_string =~ s/\./##DOT##/g;
    $$xml_string =~ s/\[/##LBRACKET##/g;
    $$xml_string =~ s/\]/##RBRACKET##/g;
    $$xml_string =~ s/\{/##LBRACE##/g;
    $$xml_string =~ s/\}/##RBRACE##/g;
    $$xml_string =~ s/\=\>/##EQUALSGREATER##/g;

    # it is possible that an infinite loop may be hit if there
    # exist circularly dependent insert elements,
    # IE: if insert element1 needs to get inserted into insert element2, and vice versa
    # this would cause an infinite loop
    # first, verify that there are no circular depencies on any of the insert elements
    my @insert_into_elements = ();
    foreach my $insert_element (keys %{$insert_elements_hash}) {
        @insert_into_elements = @{$insert_elements_hash->{$insert_element}};

        # verify that none of these push into elements exist in the insert element hash
        # and also containg $insert_element in their dependecy list
        foreach my $insert_into_element (@insert_into_elements) {
            if (exists $insert_elements_hash->{$insert_into_element}) {
                if (grep(/^$insert_element$/, @{$insert_elements_hash->{$insert_into_element}})) {
                    # then a circular dependecy has been found, error out here
                    $_IntObject->_set_err("Discovered a circular <!INSERT_ELEMENT> DTD dependency " .
                                          "between elements '$insert_element' and '$insert_into_element'");
                    return;
                }
            }
        }
    }

    # process all insert elements, this means that all insert elements
    # need to be inserted into each of the parent elements they need to be
    # inserted into.
    @insert_into_elements = ();
    my @insert_elements = ();
    my @dependency_insert_elements = ();
    while (scalar(keys %{$insert_elements_hash})) {
        foreach my $insert_element (reverse keys %{$insert_elements_hash}) {
            # obtain the elements that this insert element will be inserted into
            @insert_into_elements = @{$insert_elements_hash->{$insert_element}};

            # if any of the dependencies for this insert element exist in the insert element hash,
            # then move on to the next insert element, this will allow insert elements
            # to be defined in any order in the DTD files
            @insert_elements = keys %{$insert_elements_hash};
            @dependency_insert_elements = grep {
                my $elm = $_;
                if (grep(/^$elm$/, @insert_elements)) {
                    # then there exist dependencies for this insert element
                    1;
                } else {
                    # no dependencies for this insert element exist of this type
                    0;
                }
            } @insert_into_elements;

            # if there are any dependencies for this insert element, then go on to the next insert element
            # we have to process all dependencies before we can process this one
            if (scalar(@dependency_insert_elements)) {
                next;
            } else {
                # otherwise, we can delete this key from the the insert element hash
                # since we are guaranteed that it will get processed
                delete($insert_elements_hash->{$insert_element});
            }

            # process all the elements that this insert element needs to get inserted into
            foreach my $elm (@insert_into_elements) {

                # convert all empty instances of "<$elm/>" into "<$elm></$elm>";
                $$xml_string =~ s/(<${elm}\b[^<]*?)\/>/${1}><\/${elm}>/gs;

                # scan all instances of <$elm></$elm> and insert the insert element into it
                # if it doesn't already contain one
                while ($$xml_string =~ /(<${elm}\b.*?>)(.*?)(<\/${elm}\b>)/gs) {
                    my $start = $1;
                    my $middle = $2;
                    my $end = $3;

                    # keep track of the regexp position so we can jump back to it
                    my $pos = pos($$xml_string);

                    # count how many times the insert element exists in the middle
                    my $count = 0;
                    while ($middle =~ /<$insert_element\b/g) {
                        $count++;
                    }

                    # if $middle contains no insert element tags, then insert it at the beginning of the $middle
                    if ($count == 0) {
                        $$xml_string =~ s/${start}${middle}${end}/${start}<$insert_element\/>${middle}${end}/;
                        # since the last regexp reset the position, we need to reset it back to where it was.
                        # this will optimize subsequent regex calls against xml_string
                        pos($$xml_string) = $pos;
                    }
                }
            }
        }
    }

    # clean up the subtitutions back to the way they were
    $$xml_string =~ s/##DOLLARSIGN##/\$/g;
    $$xml_string =~ s/##PIPE##/\|/g;
    $$xml_string =~ s/##LPAREN##/\(/g;
    $$xml_string =~ s/##RPAREN##/\)/g;
    $$xml_string =~ s/##CARET##/\^/g;
    $$xml_string =~ s/##SPLAT##/\*/g;
    $$xml_string =~ s/##BANG##/\!/g;
    $$xml_string =~ s/##QUESTION##/\?/g;
    $$xml_string =~ s/##DOT##/\./g;
    $$xml_string =~ s/##LBRACKET##/\[/g;
    $$xml_string =~ s/##RBRACKET##/\]/g;
    $$xml_string =~ s/##LBRACE##/\{/g;
    $$xml_string =~ s/##RBRACE##/\}/g;
    $$xml_string =~ s/##EQUALSGREATER##/\=\>/g;

    return;
}

###############################################################################
# Method:    _process_enumeration_elements
# Objective: process all enumeration elements
# Returns:   Returns updated XML array after all elements have been enumerated
###############################################################################
sub _process_enumeration_elements {
    my (@xml_file) = @_;

    my $elements_hash;
    my @return;

    # iterate through all lines in the xml file
    for (my $i = 0 ; $i < scalar(@xml_file) ; $i++) {
        my $element_tag = $xml_file[$i];

        if ($element_tag =~ /^\s*<\b(\S+)\b.*?(\/)?>/) {
            # parse the name of the element
            my $element_name = $1;

            # determine if this tag contains children
            my $has_children = $2 eq "/" ? 0 : 1;

            # default the element line to be the original line
            my $final_element_line = "$element_tag";

            # if this element has children, then we need to process the children first
            if ($has_children) {
                # create @children, which represents an array of all the children tags for this tag
                # IE:   if @xml_file = (<a>,<b/>,<c/>,</a>)
                # THEN: @children = (<b/>,</c>)
                my @children;
                for (my $j = $i + 1 ; $j < scalar(@xml_file) ; $j++) {
                    if ($xml_file[$j] =~ /^\s*<\/${element_name}>/) {
                        # then we hit the end, generate the children array
                        # and break out
                        @children = @xml_file[$i+1..$j-1];
                        # move the $i pointer up to where we left off in the array $j
                        $i = $j;
                        last;
                    }
                }

                # its possible that the tag had no children, but has an open and close tag.
                # if there are children, then process them.
                if (scalar(@children)) {
                    # recursively handle the children, this should return all the children tags after
                    # they have been processed
                    my @processed_children = &ParamsMgr::_process_enumeration_elements(@children);

                    # now handle the current element for enumeration, and
                    # feed in the processed children so that it they can be inserted in between
                    # this element's open and close tags
                    my $element_post_enumeration = &ParamsMgr::_handle_element_enumeration($element_tag,
                                                                                           $element_name,
                                                                                           @processed_children);
                    # now that we have completed this element, and all it's children, push it onto the return array
                    push(@return, $element_post_enumeration);

                    # move onto the next element in the array
                    next;
                } else {
                    # this is the case where there are no children in this tag,
                    # allow the code to continue on to the next section
                }
            }

            # this element has no children, handle enumeration on this tag, but don't pass in any children elements
            my $element_post_enumeration = &ParamsMgr::_handle_element_enumeration($element_tag, $element_name);

            # now that we have completed this childless element, push it onto the return array
            push(@return, $element_post_enumeration);
        } else {
            # element wasnt parsable, most likely a non xml line IE: DTD data, push it on
            push(@return, $element_tag);
        }
    }

    # we have completed all of the elements passed into this function,
    # return them in processed form.
    return @return;
}

###############################################################################
# Method:    _handle_element_enumeration()
# Objective: Helper function to perform the actual enumeration of an element
# Returns:   Returns the element plus any enumerations that may have been
#            dynamically created.
###############################################################################
sub _handle_element_enumeration {
    my ($element_tag, $element_name, @children) = @_;

    my $attr_hash = {};

    # parse out the enumeration delimiter
    my $enum_syntax;
    if ($element_tag =~ /<${element_name}\s+.+?enum_syntax=\"(.+?)\"/) {
        $enum_syntax = $1;
    } else {
        # otherwise there is no enum_syntax, just return
        # if there are children, then add them in and close the tag
        if (scalar(@children)) {
            $element_tag .= "@children</$element_name>";
        } else {
            # otherwise close out the tag;
            $element_tag =~ s/>\s*$/\/>/;
            $element_tag =~ s/\/\/>\s*$/\/>/;
        }

        return $element_tag;
    }

    # need to convert greater than and less than signs back to normal
    $enum_syntax =~ s/&gt;/>/g;
    $enum_syntax =~ s/&lt;/</g;

    # parse the enum syntax by evaling it
    my $enum_syntax_hash = eval($enum_syntax);
    if ($@) {
        die "Invalid enum_syntax (eval error: $@) : $enum_syntax";
    }

    # make sure it is a hash
    unless (ref($enum_syntax_hash) eq "HASH") {
        die "Invalid enum_syntax (not a hash) : $enum_syntax";
    }

    # scan the enum syntax hash and replace $PARAMS with their values
    &ParamsMgr::_interpolate_enum_syntax_params($enum_syntax_hash);

    # parse the enum syntax hash
    my $primary_key;
    my $primary_key_loop_count = 1;
    my $primary_key_loop_type = "reset";
    foreach my $key (keys %{$enum_syntax_hash}) {
        # verify that all the keys for this key are valid enum_syntax flags
        my @flags = keys (%{$enum_syntax_hash->{$key}});
        my @badflags = grep {
            $_ !~ /(^key$|^delim$|^loop_count$|^increment$|^max_increment$|^prefix$|^suffix$|^permute$|^loop_type$|^search_replace$)/;
        } @flags;
        if (scalar(@badflags)) {
            die "Invalid enum_syntax (Attribute '$key' of element '$element_name' contains " .
                "invalid options '@badflags', you probably misspelled something)";
        }

        # parse out the primary key
        if (defined($enum_syntax_hash->{$key}->{key}) and $enum_syntax_hash->{$key}->{key} == 1) {
            # if there is already a primary key defined, then there is a duplicate primary key specified
            # this is an error
            if (defined($primary_key)) {
                die "Invalid enum_syntax (more than one primary key specified for element $element_name";
            } else {
                $primary_key = $key;
            }

            # store the loop count for the primary key
            if (defined($enum_syntax_hash->{$key}->{loop_count})) {
                $primary_key_loop_count = $enum_syntax_hash->{$key}->{loop_count};
            }

            # store the loop type for the primary key
            if (defined($enum_syntax_hash->{$key}->{loop_type})) {
                $primary_key_loop_type = $enum_syntax_hash->{$key}->{loop_type};
            }
            last;
        }

        # make sure that the user did not specify a loop count on a non-primary key
        elsif ((not defined($enum_syntax_hash->{$key}->{key}) or $enum_syntax_hash->{$key}->{key} == 0) and
                   (defined($enum_syntax_hash->{$key}->{loop_count}))) {
            die "Invalid enum_syntax (Attribute '$key' of element '$element_name' contains " .
                "a loop_count value when it is not explicitly specified as the key)";
        }
    }

    # parse out all the attributes that need to be delimited
    # and build $attr_hash with each attr as a hash key, and
    # all the delimited values of that attribute as the hash value
    my $max_attr = "";
    my $max_values = 0;
    while ($element_tag =~ /\b(\S+?)\s*=\s*\"(.*?)\"/g) {
        my $attr = $1;
        my $value = $2;

        # remove the enumeration syntax attribute
        next if ($attr =~ /^enum_syntax$/);

        my (@delimited_values, @original_values);

        # if this attr is represented in the enum syntax,
        # then its value should be split by the delimiter
        if (exists($enum_syntax_hash->{$attr})) {
            # obtain the delimiter for this attribute
            my $delimiter = $enum_syntax_hash->{$attr}->{delim};
            if (not defined($delimiter)) {
                # default the delimiter to a ':' character
                $delimiter = ":";
            }
            # parse out the delimited values for this attribute
            @delimited_values = split(/$delimiter/, $value);
            @original_values = @delimited_values;
        } else {
            # else its value should be placed in a single element array
            @delimited_values = ($value);
            @original_values = ($value);
        }

        # keep track of which attribute has the max # of values
        if (scalar(@original_values) > $max_values) {
            $max_attr = $attr;
            $max_values = scalar(@original_values);
        }

        # store the attributes and their delimited values
        $attr_hash->{$attr}->{values_temp} = \@delimited_values;
        $attr_hash->{$attr}->{values_original} = \@original_values;
    }

    # verify that a primary key is defined
    if (not defined($primary_key) or $primary_key eq "") {
        # if no primary key is defined, then assume the attr
        # with the max value is the key
        $primary_key = $max_attr;
    }

    # pre-process all attributes that have permute enabled
    my @permute_array_set;
    my @permute_array_elms;
    foreach my $elm (%{$enum_syntax_hash}) {
        if (exists($enum_syntax_hash->{$elm}->{permute}) and
            $enum_syntax_hash->{$elm}->{permute} == 1) {
            # take the values_temp array for this element and add it to the permute array set
            push(@permute_array_set, $attr_hash->{$elm}->{values_temp});
            push(@permute_array_elms, $elm);
        }
    }

    # handle permute elements
    if (scalar(@permute_array_elms)) {
        # create the permutation hash
        my $permute_hash = &ParamsMgr::_permute(@permute_array_set);

        my $i = 0;
        foreach my $elm (@permute_array_elms) {
            my @values_temp;
            my @values_orig;
            # modify the values_temp and values original arrays
            # for each permutated element
            if (defined($permute_hash)) {
                @values_temp = @{$permute_hash->{$i}};
                @values_orig = @{$permute_hash->{$i}};
            } else {
                my $orignal_attributes = '';
                foreach my $orig_attr (keys %{$attr_hash}){
                    my $original_values = defined($attr_hash->{$orig_attr}->{values_original})?
                                            join(",",@{$attr_hash->{$orig_attr}->{values_original}}):
                                            '';
                    $orignal_attributes .= "$orig_attr='$original_values'";
                }
                die "Cannot permute over attribute '$elm' for the element '$element_name', " .
                    "its value is not defined. Did you forget to fill it in?\n" .
                    "xml_element='".$orignal_attributes."'\n (Delimeters may not be the same)";
            }
            $attr_hash->{$elm}->{values_temp} = \@values_temp;
            $attr_hash->{$elm}->{values_original} = \@values_orig;
            $i++;
        }
    }

    # build the xml line for this element based on the attr hash

    # process the primary key values
    my @enumerated_element_tags;
    foreach my $primary_key_value (@{$attr_hash->{$primary_key}->{values_temp}}) {
        my $hash;
        $hash->{$primary_key} = $primary_key_value;
        push(@enumerated_element_tags, $hash);
        delete($attr_hash->{$primary_key});
    }

    # now build the final string
    my $final_string;

    # process the primary key $primary_key_loop_count times
    for (my $i = 0; $i < $primary_key_loop_count; $i++) {
        # process each value of the primary key
        foreach my $orig_elm_hash (@enumerated_element_tags) {
            # make a clone of the elm hash so we don't corrupt the next series of loops that need to process the hash
            my $elm_hash = Clone::clone($orig_elm_hash);

            # add any specified prefixes or suffixes to the primary key value
            if (defined($enum_syntax_hash->{$primary_key}->{prefix})) {
                $elm_hash->{$primary_key} = $enum_syntax_hash->{$primary_key}->{prefix} . $elm_hash->{$primary_key};
            }
            if (defined($enum_syntax_hash->{$primary_key}->{suffix})) {
                $elm_hash->{$primary_key} .= $enum_syntax_hash->{$primary_key}->{suffix};
            }

            # check if this primary key needs to increment
            if (defined($enum_syntax_hash->{$primary_key}->{increment})) {
                if (not defined($enum_syntax_hash->{$primary_key}->{increment_original})) {
                    # store the original starting value of the incremement
                    $enum_syntax_hash->{$primary_key}->{increment_original} = $enum_syntax_hash->{$primary_key}->{increment};
                }
                # if the max incremement has been hit, reset increment back to its original value
                if (defined($enum_syntax_hash->{$primary_key}->{max_increment}) and
                    $enum_syntax_hash->{$primary_key}->{increment} > $enum_syntax_hash->{$primary_key}->{max_increment}) {
                    $enum_syntax_hash->{$primary_key}->{increment} = $enum_syntax_hash->{$primary_key}->{increment_original};
                }

                # increment the count and apply it to the value
                if (length($elm_hash->{$primary_key})) {
                    $elm_hash->{$primary_key} .= "_" . $enum_syntax_hash->{$primary_key}->{increment}++;
                } else {
                    $elm_hash->{$primary_key} = $enum_syntax_hash->{$primary_key}->{increment}++;
                }
            }

            # process all secondary keys
            foreach my $attr (keys %{$attr_hash}) {
                # if there are no more values for this secondary key left,
                # reset from the original list
                if (scalar(@{$attr_hash->{$attr}->{values_temp}}) == 0) {

                    # if the loop type for this secondary key is reverse,
                    # then reverse the original list
                    if ($enum_syntax_hash->{$attr}->{loop_type} =~ /reverse/i) {
                        @{$attr_hash->{$attr}->{values_original}} = reverse(@{$attr_hash->{$attr}->{values_original}});
                    }

                    # restore temporary array from the original
                    my @original = @{$attr_hash->{$attr}->{values_original}};
                    $attr_hash->{$attr}->{values_temp} = \@original;
                }

                # add the next value for this attribute to this hash
                my ($next_value) = shift(@{$attr_hash->{$attr}->{values_temp}});

                # add any specified prefixes or suffixes to the primary key value
                if (defined($enum_syntax_hash->{$attr}->{prefix})) {
                    $next_value = $enum_syntax_hash->{$attr}->{prefix} . $next_value;
                }
                if (defined($enum_syntax_hash->{$attr}->{suffix})) {
                    $next_value .= $enum_syntax_hash->{$attr}->{suffix};
                }

                # check if this attribute needs to get incremented
                if (defined($enum_syntax_hash->{$attr}->{increment})) {
                    if (not defined($enum_syntax_hash->{$attr}->{increment_original})) {
                        # store the original starting value of the incremement
                        $enum_syntax_hash->{$attr}->{increment_original} = $enum_syntax_hash->{$attr}->{increment};
                    }
                    # if the max incremement has been hit, reset increment back to its original value
                    if (defined($enum_syntax_hash->{$attr}->{max_increment}) and
                        $enum_syntax_hash->{$attr}->{increment} > $enum_syntax_hash->{$attr}->{max_increment}) {
                        $enum_syntax_hash->{$attr}->{increment} = $enum_syntax_hash->{$attr}->{increment_original};
                    }
                    # increment the count and apply it to the value
                    if (length($next_value)) {
                        $next_value .= "_" . $enum_syntax_hash->{$attr}->{increment}++;
                    } else {
                        $next_value = $enum_syntax_hash->{$attr}->{increment}++;
                    }
                }

                # update the value for this attr
                $elm_hash->{$attr} = $next_value;
            }

            # interpolate attrs that may be referenced in other attrs
            foreach my $attr (keys %{$elm_hash}) {
                while ($elm_hash->{$attr} =~ /\%\{(\S+?)\}/g) {
                    my $interp_attr = $1;
                    my $unmodified = "%{" . $interp_attr . "}";

                    # if this interp attr exists in the elm_hash, replace it with the value of that attr
                    if (exists($elm_hash->{$interp_attr})) {
                        my $value = $elm_hash->{$interp_attr};

                        # interp the value
                        $elm_hash->{$attr} =~ s/$unmodified/$value/g;
                    } else {
                        # user specified an invalid attr name
                        die "Invalid enum_syntax (The value of attribute '$attr' of element '$element_name' " .
                            "is attempting to interpolate '\%\{$interp_attr\}', " .
                            "but '$interp_attr' does not exist as an attribute to this element, " .
                            "maybe you misspelled '$interp_attr')";
                    }
                }

                # apply the search and replace regex on the final value
                if ($enum_syntax_hash->{$attr}->{search_replace}) {
                    my $regex = $enum_syntax_hash->{$attr}->{search_replace};
                    if ($regex =~ /^\s*s\/(.+?)\/(.+?)\/([ig]*?)\s*$/) {
                        my $find = $1;
                        my $replace = $2;
                        my $options = $3;
                        if ($options) {
                            # only supporting global option and case insensitve here
                            if ($options =~ /g/ and $options !~ /i/) {
                                $elm_hash->{$attr} =~ s/$find/$replace/g;
                            } elsif ($options !~ /g/ and $options =~ /i/) {
                                $elm_hash->{$attr} =~ s/$find/$replace/i;
                            } else {
                                $elm_hash->{$attr} =~ s/$find/$replace/ig;
                            }
                        } else {
                            $elm_hash->{$attr} =~ s/$find/$replace/;
                        }
                    }
                }
            }

            # flush this elm hash into an element (with optional children)
            my $elm_string;
            if ($primary_key eq "enum_iter") {
                # dont add enum_iter to the final element
                $elm_string = "<$element_name";
            } else {
                # make sure the primary key is at the front of the element for readability
                $elm_string = "<$element_name $primary_key=\"" . $elm_hash->{$primary_key} ."\"";
            }
            foreach my $attr (keys %{$elm_hash}) {
                # skip if this is the primary key OR if this is enum_iter
                next if ($attr eq $primary_key or $attr eq "enum_iter");
                $elm_string .= " $attr=\"" . $elm_hash->{$attr} . "\"";
            }

            # if there are children, then add them in and close the tag
            if (scalar(@children)) {
                $elm_string .= ">@children</$element_name>";
            } else {
                # otherwise close out the tag;
                $elm_string .= "/>";
            }

            # add this element string to the final string
            $final_string .= $elm_string;
        }

        # if loop type for the primary key is reverse, then reverse the value list of the primary key
        if ($primary_key_loop_type =~ /reverse/i) {
            @enumerated_element_tags = reverse(@enumerated_element_tags);
        }
    }

    return $final_string;
}

###############################################################################
# Method:    _permute()
# Objective: Create permutations given a set of array references
# Returns:   A hash of all permutation values per element
###############################################################################
sub _permute {
    my @arr = _permute_helper(@_);

    my $hash;

    foreach my $perm (@arr) {
        my $i = 0;
        foreach my $elm (@{$perm}) {
            push(@{$hash->{$i}}, $elm);
            $i++;
        }
    }

    return $hash;
}

###############################################################################
# Method:    _permute_helper()
# Objective: Create all permutations of a set of arrays passed in
# Returns:   An array of array references, each being a permutation
###############################################################################
sub _permute_helper {
    my $last = pop @_;
    unless (@_) {
        return map {[$_]} @$last;
    }
    return map { my $left = $_;
                 map {[@$left, $_]} @$last } _permute_helper(@_);
}

###############################################################################
# Method:    _interpolate_enum_syntax_params()
# Objective: Walk the enum syntax hash and convert dollar variables to their
#            Respective tharn values
###############################################################################
sub _interpolate_enum_syntax_params {
    my ($hash) = @_;

    foreach my $key (keys %{$hash}) {
        if (ref($hash->{$key}) eq "HASH") {
            # recursively handle the child hashes
            &ParamsMgr::_interpolate_enum_syntax_params($hash->{$key});
        } else {
            # otherwise this is a leaf, handle the parameter
            while ($hash->{$key} =~ /\$!([A-Za-z0-9,\.]+)/g) {
                my $tharn_var = $1;
                my $tharn_var_unmod = $tharn_var;
                my $default = "";
                if ($tharn_var =~ /^(\S+)\,(.*)$/) {
                    ($tharn_var, $default) = ($1, $2);
                }

                if ($tharn_var =~ /^(\S+)\.(\S+)$/) {
                    my ($host, $param) = ($1, $2);
                    my $values;

                    # Get the resource name
                    my $name = &Tharn::param("$host");

                    # Eval host call so that it does not blow up
                    eval {
                        $values = host($name);
                    };

                    # If no error occurred during call of host() process the values
                    if (not $@) {
                        my $value = $values->{$param};
                        $hash->{$key} =~ s/\$!\Q$tharn_var_unmod\E/\"$value\"/g;
                    } elsif (defined($default)) {
                        # otherwise use the default value that may be specified
                        my $value = $default;
                        # Substitute value for dot variable in XML
                        $hash->{$key} =~ s/\$!\Q$tharn_var_unmod\E/\"$value\"/g;
                    }
                } else {
                    # pull in the default from the tharn environment
                    my $tharn_value = &Tharn::param("$tharn_var", '-default', $default);
                    $hash->{$key} =~ s/\$!\Q$tharn_var_unmod\E/\"$tharn_value\"/g;
                }
            }

            # now, eval the value of the key if it is wrapped in an eval(...) call
            if ($hash->{$key} =~ /^eval\((.*)\)$/) {
                my $value = $1;
                no strict 'subs';
                my $before = $hash->{$key};
                $hash->{$key} = eval($value);
                if ($@) {
                    die("Error while evaluating enum_syntax code '$before'", $@);
                }
                use strict;
            }
        }
    }

    return;
}

###############################################################################
# Method:    _process_xml_includes()
# Objective: substitute xml includes with their replacement xml files
# Returns:   An updated XML string
###############################################################################
sub _process_xml_includes {
    my ($xml_string) = @_;

    while ($$xml_string =~ /(<include_xml\s+file=\".+?\"\/>)/g) {
        my $include_string = $1;
        my ($include_xml_file_name) = ($include_string =~ /file="(.+?)"/);

        # open the xml include file, read all the contents, and substitute it
        # in the place of $include_string
        my $include_xml_fh;
        if (not open($include_xml_fh, "<", $include_xml_file_name)) {
            $_IntObject->_set_err("Unable to open include XML File: '$include_xml_file_name' for reading");
            return "";
        }
        my @include_xml_file = <$include_xml_fh>;
        chomp(@include_xml_file);
        my $include_xml_string = join("\n", @include_xml_file);

        # replace $include_string, with $include_xml_string
        $$xml_string =~ s/${include_string}/${include_xml_string}/g;

        $Log->comment("Loading XML Include File: $include_xml_file_name");
    }

    return $xml_string;
}
###############################################################################
# Method:    _display_xml
# Objective: Display the XML data in a subtest
###############################################################################
sub _display_xml {
    my (%args) = @_;

    my $runid = delete($args{runid});
    $runid ||= "XML-environment";
    my $xml_data = delete($args{xml_data});
    if (!$xml_data) {
        $_IntObject->_set_err("Missing required parameter 'xml_data'");
        return $ParamsMgr::FALSE;
    }

    my $subtest = &SubIPC::create(-id => "$runid",
                                  -run => sub {
                                      my $str = $xml_data->toString(2);
                                      $Log->comment("---XML Environment---");
                                      print "$str\n";
                                      &SubIPC::pass(-msg => "XML Environment");
                                      return $ParamsMgr::TRUE;
                                  },
                                  -bg => $ParamsMgr::TRUE,
                                  -ipc => $ParamsMgr::TRUE,
                                  -esc => 1);

    &SubIPC::wait_finish(-subtest => [$subtest]);

    if (&SubIPC::error()) {
        $_IntObject->_set_err("Failed to invoke XML Environment subtest", &SubIPC::get_error());
        return $ParamsMgr::FALSE;
    }

    return $ParamsMgr::TRUE;
}

###############################################################################
# Method:    _parent_caller
# Objective: _parent_caller returns the subroutine of caller of depth 2
# Details:   format for the caller:
#      0         1          2      3            4
#      ($package, $filename, $line, $subroutine, $hasargs,
#      5          6          7            8       9         10
#      $wantarray, $evaltext, $is_require, $hints, $bitmask, $hinthash) = caller($i);
#      $i is the caller depth
###############################################################################
sub _parent_caller { (caller(2))[3] }

###############################################################################
# Method:    error
# Objective: Return error flag on ParamsMgr object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub error {

    return $_IntObject->Error::error();
}

###############################################################################
# Method:    errmsg
# Objective: Return the errors string on the ParamsMgr object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub errmsg {

    return $_IntObject->Error::errmsg();
}

###############################################################################
# Method:    get_error
# Objective: Return the errors string on the ParamsMgr object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_error {

    return $_IntObject->errmsg();
}

###############################################################################
# Method:    clear_error
# Objective: Clear the error string(s) set on the ParamsMgr object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clear_error {

    return $_IntObject->_clear_err();
}

1;
