#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Pathdiff.pm#1 $

## @summary     Class to compare two different file paths
## @author      snider@netapp.com
## @pod here

=head1 NAME

Pathdiff - Class to handle file path comparisons (for GX specifically)

=head1 SYNOPSIS

    use Pathdiff;
    my $pathdiff = new Pathdiff;
    $db_file = $pathdiff->snapshot(directory=>"/some/path");
    $pathdiff->compare(source=>"/some/path", target=>"/some/other/path");

=head1 DESCRIPTION

C<Pathdiff> implements a class which is used to compare multiple filesystems.
After creating the object, run the C<snapshot> method to create a snapshot
of a particular filesystem. The object will store this snapshot internally
and will also return a reference to it to the caller. The C<compare> method
can then be called to compare two filesystems. 

=head1 METHODS

=cut

#***********************************************************************
# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDE_PATHDIFF}) {
    my $newModule = $ENV{OVERRIDE_PATHDIFF};
    delete $ENV{OVERRIDE_PATHDIFF};
    print "The main Pathdiff module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package Pathdiff;
use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Data::Dumper;
use tools::qa::pathdiff; # brings in PathDiff
use TracePkg;
use Options;
my $Trace = new TracePkg;
my $Options = new Options;

$Options->parameters(definitions=>
                     [
                      {
                       name=>"LEAVE_PATHDIFF_DATABASE",
                       spec=>"boolean",
                       default=>"false",
                       description=>"should the pathdiff database file be left around?",
                      },
                      {
                       name=>"PATHDIFF_VERBOSITY",
                       default=>0,
                       description=>"the verbosity setting for pathdiff.pm",
                       # this should go away once pathdiff.pm and Pathdiff.pm are merged
                      },
                     ]
			 );

=head2 new

    my $pathdiff = new Pathdiff(pathdiff_config => {});

C<new> is the invocant for the Pathdiff class.

=over

=item pathdiff_config (Optional)

Hash reference to pass to tools::qa::pathdiff->new()

=back


=cut

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {
                @_,
            };
    $Trace->entry();
    bless($self, $class);

    $Options->process() if (!$Options->processed());
    $self->{options} = $Options->get_hash();
    $Trace->exit();
    return $self;
}

=head2 snapshot

    $db_file = $pathdiff->snapshot(directory=>"/some/path");

C<snapshot> will create a snapshot of the provided path. The snapshot will
be stored internally. The call will also return a reference to this 
snapshot to be optionally passed back into C<compare>. The snapshot will
contain metadata and the MD5 checksum for all files/directories in the 
path.

=over

=item directory

Specifies the path to the filesystem to snapshot

=back

=cut

sub snapshot {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'Pathdiff' } ],
                                         {
                                           directory => { type => SCALAR },
                                         }
					 );
    $Trace->entry();

    my $database_file;
    if ($^O !~ /Win/) {
        $database_file = "/tmp/pathdiff.";
    } else {
        $database_file = $ENV{TEMP}."/pathdiff.";
    }

    our $index;
    $index = 0 if (!defined($index));
    
    while ($index < 100) {
        if (! -f "$database_file".$index) {
            $database_file = $database_file.$index;
            last;
        }
        $index++;
    }
    if ($index >= 100) {
        $Trace->fatal(msg=>"Found more than 100 database files");
    }
    $Trace->trace(msg=>"will use database file '$database_file'");
    $self->{directories}{$params{directory}} = $database_file;
    push(@{$self->{db_files}}, $database_file);
    
    my %options = (
                   src => $params{directory},
                   tgt => $database_file,
                   verbosity => $self->{options}{PATHDIFF_VERBOSITY},
               );
    if ($self->{pathdiff_config}) {
        %options = ( (%{$self->{pathdiff_config}}), (%options));
    }
    my $pathdiff = new PathDiff(\%options);
    my $rc = $pathdiff->DoDiff();
    if ($rc) {
        $Trace->fatal(msg=>"Failed to take the snapshot");
    }

    $Trace->exit(msg=>"return $database_file");
    return $database_file;
}

=head2 compare

    $pathdiff->compare(source=>"/some/path", target=>"/some/other/path");
    $pathdiff->compare(source=>$db_file, target=>"/some/other/path");

C<compare> will compare two filesystems looking for differences. It will
die if new files are found in the target, files are missing from the target
or files in the two filesystems are found to have differences in the metadata
or MD5 checksums. Internally, if the source directory has a stored snapshot,
it will use the stored snapshot for comparison. If the source specifies a
snapshot file (returned by C<snapshot>) then that file will be used. 
Otherwise, the source filesystem will be read and compared to the target.

=over

=item source

Specifies either a source filesystem or a snapshot database file returned by
the C<snapshot> method.

item target

Specifies the target filesystem to compare the source against.

=back

=cut

sub compare {
    my ($self, %params) = validate_mixed(
					 @_,
					 [ { type => OBJECT, isa => 'Pathdiff' } ],
                                         {
                                          source => { type => SCALAR },
                                          target => { type => SCALAR },
                                          skip => { type => ARRAYREF, default => [qw(ATime)] },
                                          cleanup => { type => SCALAR | UNDEF, optional => 1 },
                                         }
					 );
    $Trace->entry();

    my $target = $params{target};
    my $source;
    my @delete;
    if ($source=$self->{directories}{$params{source}}) {
        $Trace->trace(msg=>"source set to a database file based on $params{source}: $params{source}");
        push(@delete, $source);
    } elsif (($source) = grep($_ eq $params{source}, @{$self->{db_files}})) {
        $Trace->trace(msg=>"source set to a database file that was passed in: $params{source}");
        push(@delete, $source);
    } else {
        $source=$params{source};
        $Trace->trace(msg=>"source set to $params{source}");
    }
    $Trace->trace(msg=>"target='$target'");
    
    if (! -d $target) {
        $Trace->fatal(msg=>"The target '$target' does not exist!");
    }
    
    my %options = (
                   src => $source,
                   tgt => $target,
                   skip => $params{skip},
                   verbosity => $self->{options}{PATHDIFF_VERBOSITY},
               );
    if ($self->{pathdiff_config}) {
        %options = ( (%{$self->{pathdiff_config}}), (%options));
    }
    my $pathdiff = new PathDiff(\%options);
    my $rc = $pathdiff->DoDiff();
    
    if ($params{cleanup} && scalar @delete) {
        my @errors;
        foreach my $db_file (@delete) {
            unlink($db_file) || push(@errors, $db_file);
        }
        if (@errors) {
            $Trace->warn(msg=>"Could not delete all database files: ".join(" ", @errors));
        }
    }

    if ($rc) {
        $Trace->fatal(msg=>"The source and target differ");
    }

    $Trace->exit();
}

sub DESTROY {
    my $self = shift;
    
    if (!$self->{options}{LEAVE_PATHDIFF_DATABASE}) {
        my @errors;
        foreach my $db_file (@{$self->{db_files}}) {
            $Trace->trace(msg=>"delete '$db_file'");
            unlink($db_file) || push(@errors, $db_file);
        }
        if (@errors) {
            $Trace->warn(msg=>"Could not delete all database files: ".join(" ", @errors));
        }
    } else {
        $Trace->trace(msg=>"Not deleting database files due to LEAVE_PATHDIFF_DATABASE being set");
    }
    
}

=head1 PARAMS

=over

=item LEAVE_PATHDIFF_DATABASE

Specifies whether the internally-kept snapshot files should be left around
when the script exits. The files are deleted by default.

=item PATHDIFF_VERBOSITY

Specifies the verbosity setting to pass to the internally-used pathdiff.pm
script/module. Defaults to 0.

=back

=head1 SEE ALSO


=head1 BUGS


=head1 STATUS

  Shared

=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: //depot/prod/test/Rsierranevada/lib/Pathdiff.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut


1;
