# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  Callout return value constants plus ':Callout' attribute definition
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::Callout

=head1 DESCRIPTION

This package does two things. It defines and implements the ':Callout'
attribute, and exports constants for the values callouts are required
to return.

To define callouts, users should import the ':slsl_results' symbols
from this package, and then tag subroutines that they want to be
callable as callouts with the ':Callout' attribute.

All callouts are passed an initial Perf::AAMP::Test::Resources object
when they are called; other arguments can be passed from the point of
call.

Callouts are required to return one of the result values defined below:

    SLSL_RESULT_CONTINUE  - continue adaptive iteration
    SLSL_RESULT_DONE      - test complete
    SLSL_RESULT_ERROR     - fatal error in callout
    SLSL_RESULT_FAILURE   - test failure in callout
    SLSL_RESULT_STOP      - end adaptive iteration
    SLSL_RESULT_SUCCESS   - successful callout execution


    Example:

        use Perf::AAMP::Callout qw(:slsl_results);

        sub set_volume_policy :Callout {
            my $resources = shift;
            my %opts = validate(
                @_, {
                    volume  => { type => SCALAR },
                    vserver => { type => SCALAR },
                },
            );

            my $policy = $resources->stash->fetch(name => 'policy-name');

            NACL::C::Volume->modify(
                command_interface  => $resources->command_interface(),
                volume             => $opts{volume},
                vserver            => $opts{vserver},
                'qos-policy-group' => $policy,
            );

            return SLSL_RESULT_SUCCESS;
        }


Note that the Perf::AAMP::Callout::Dispatcher package allows callout
subroutines to be called from NATE tests without the caller having to
explicitly pass the $resources argument. See that package's documentation
for details.

=head1 SEE ALSO

    Perf::AAMP::Test
    Perf::AAMP::Test::Resources
    Perf::AAMP::Test::Stash
    Perf::AAMP::Callout::Dispatcher

=cut

package Perf::AAMP::Callout;

use strict;
use warnings;
use base q(Exporter);
use attributes;
use Carp;
use List::MoreUtils qw(any);


# Public constants:
use constant {
    SLSL_RESULT_DONE      => 'DONE',
    SLSL_RESULT_ERROR     => 'ERROR',
    SLSL_RESULT_FAILURE   => 'FAIL',
    SLSL_RESULT_SUCCESS   => 'SUCCESS',
};

# Public list constants:
use constant SLSL_RESULT_CONTINUE => (SLSL_RESULT_SUCCESS, 'continue');
use constant SLSL_RESULT_STOP     => (SLSL_RESULT_SUCCESS, 'stop');

our %EXPORT_TAGS = (
    slsl_results => [
        qw(SLSL_RESULT_CONTINUE
           SLSL_RESULT_DONE
           SLSL_RESULT_ERROR
           SLSL_RESULT_FAILURE
           SLSL_RESULT_STOP
           SLSL_RESULT_SUCCESS
           MODIFY_CODE_ATTRIBUTES),
    ],
);

our @EXPORT_OK = (
    @{$EXPORT_TAGS{slsl_results}},
);


# Below, we define a handler for the ':Callout' attribute.
#
# Note that even though this handler is passed a coderef,
# the subroutine referred to by that coderef has not yet
# been defined, and thus we cannot use the coderef to lookup
# the callout's name.
#
# Perf::AAMP::Test will need the callout's name in order to
# construct its dispatch table, so we'll lookup the names
# the first time _get_callouts() is called. By that time,
# the subroutines will have been defined and the lookup will
# succeed.
#
# As is obvious from the code, we make no attempt to support
# the ':Callout' attribute with anonymous subroutines.
#
# One final point: we use the 'attributes' pragma rather than
# 'Attribute:Handlers' to avoid requiring scripts to inherit
# from this package.

my $callouts = {
    initialized => 0,
    table       => [],
};

sub MODIFY_CODE_ATTRIBUTES {
    my ($pkg, $code, @attrs) = @_;

    confess 'Not a CODE reference!'
        if ref($code) ne 'CODE';

    my $expected = 'Callout';
    my @unknown  = grep { !m/$expected/ } @attrs;

    if (any { m/$expected/ } @attrs) {
        push @{$callouts->{table}}, {
            name => undef,
            pkg  => $pkg,
            code => $code,
        };
    }

    return @unknown;
}


sub _get_callouts {

    # The first time this is called, we'll make a second
    # pass through the callouts table, looking up and
    # storing the callout names.

    if (!$callouts->{initialized}) {
        foreach my $callout (@{$callouts->{table}}) {
            my $code = $callout->{code};
            my $pkg  = $callout->{pkg};
            my $name = undef;

            no strict qw(refs);
            foreach my $sym (values %{$pkg . q{::}}) {
                no warnings qw(uninitialized);
                if (*{$sym}{CODE} == $code) {
                    $name = (split(/::/, $sym))[-1];
                    $callout->{name} = $name;
                }
            }

            confess "Cannot locate name for coderef in package '$pkg'"
                if !defined($name);
        }

        $callouts->{initialized} = 1;
    }

    my @table = @{$callouts->{table}};
    return \@table;
}

1;

