# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  Allow AAMP callouts to be called from NATE-only tests
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::Callout::Dispatcher

=head1 DESCRIPTION

This package allows callouts written for AAMP tests to be called directly
from NATE-only tests. This reduces code duplication by allowing callout
subroutines to be use for both AAMP/SLSL and NATE tests.


    Example:

        use Perf::AAMP::Callout qw(:slsl_results);
        use Perf::AAMP::Callout::Dispatcher;

        my $dispatcher = Perf::AAMP::Callout::Dispatcher->new();
        $dispatcher->stash->add(
            name  => 'policy-name',
            value => 'test_plcy',
        );

        set_volume_policy(volume => 'testvol', vserver => 'vs0');

        sub set_volume_policy :Callout {
            my $resources = shift;
            my %opts = validate(
                @_, {
                    volume  => { type => SCALAR },
                    vserver => { type => SCALAR },
                },
            );

            my $policy = $resources->stash->fetch(name => 'policy-name');

            NACL::C::Volume->modify(
                command_interface  => $resources->command_interface(),
                volume             => $opts{volume},
                vserver            => $opts{vserver},
                'qos-policy-group' => $policy,
            );

            return SLSL_RESULT_SUCCESS;
        }


Note in the example above that the $resources object is automatically
created by the Perf::AAMP::Callout::Dispatcher constructor and passed
implicitly to the callout subroutine. Note that the dispatcher object
and the callout can pass data between each other by means of the stash
object.

=head1 SEE ALSO

    Perf::AAMP::Callout
    Perf::AAMP::Test::Resources
    Perf::AAMP::Test::Stash

=head1 METHODS

=head2 new

Construct and return a Perf::AAMP::Callout::Dispatcher object. This also
has the side effect of making all subroutines tagged with the ':Callout'
attribute directly callable with an initial $resources argument implicitly
passed.

=cut


package Perf::AAMP::Callout::Dispatcher; {
    use strict;
    use warnings;
    use Carp qw(confess);
    use Data::Dumper;
    use Hook::LexWrap;
    use List::MoreUtils qw(any);
    use NATE::ParamSet;
    use Object::InsideOut;
    use Params::Validate qw(:all);
    use Perf::AAMP::Callout qw(:slsl_results !MODIFY_CODE_ATTRIBUTES);
    use Perf::AAMP::Test::Resources;
    use Perf::AAMP::Test::Stash;
    use Tharn qw($Log);


    my @caller_pkg :Field;
    my @resources  :Field;
    my @stash      :Field :Get(stash);


    sub _init :Init {
        my $self = shift;

        # Initialize test stash. User code can access via
        # the stash() method.
        $stash[$$self] = Perf::AAMP::Test::Stash->new();

        # Initialize the test resources object. This can
        # only be accessed from within callouts.
        my $params  = NATE::ParamSet->new(global => 1);
        my @filers  = split(/:/, $params->get('FILERS'));
        my @clients = split(/:/, $params->get('CLIENTS'));

        $resources[$$self] = Perf::AAMP::Test::Resources->new(
            filers  => \@filers,
            clients => \@clients,
            stash   => $stash[$$self],
        );

        # Wrap the user-defined callouts.
        #
        # We only want to include callouts that have been imported into the
        # package that this constructor is being called from. Therefore we'll
        # confirm that the coderef accessible via the callout's name in that
        # package name matches the callout's coderef. Callouts without matching
        # coderefs will be silently skipped.
        $caller_pkg[$$self] = caller(1);       # caller(0) is Object::InsideOut
        foreach my $callout (@{Perf::AAMP::Callout::_get_callouts()}) {
            my $name = $callout->{name};
            my $code = $callout->{code};

            no strict qw(refs);
            no warnings qw(uninitialized);
            if (*{$caller_pkg[$$self] . q{::} . $name}{CODE} == $code) {
                $self->add_callout(name => $callout->{name});
            }
        }
        return;
    }


=head2 add_callout

Register a coderef to be callable as an AAMP callout. Rather than
calling this method explicitly, most callouts are registered
implicitly via the ':Callout' attribute.

=over

=item Arguments

B<name> = string

I<Required> The name of the callout coderef to make directly callable.

=back

=cut

    sub add_callout {
        my $self = shift;
        my %opts = validate(
            @_, {
                name => { type => SCALAR, optional => 0 },
            },
        );

        my $name = $caller_pkg[$$self] . q{::} . $opts{name};
        my @valid_results = (
            SLSL_RESULT_DONE,
            SLSL_RESULT_ERROR,
            SLSL_RESULT_FAILURE,
            SLSL_RESULT_SUCCESS,
        );

        # Tried to implement this using Hook::LexWrap, but evidently
        # that module is broken. For a discussion of the problem, see:
        # http://www.perlmonks.org/?node_id=423259
        $Log->debug("WRAPPING callout '$name'");

        no strict q(refs);
        no warnings q(redefine);

        my $orig = *$name{CODE};
        *$name = sub {
            $Log->debug("CALLING callout '$name' with args: "
                      . Dumper(\@_));

            # Call the original callout subroutine with the resources
            # object prepended to its argument list.
            my ($result, $payload) = $orig->($resources[$$self], @_);
            $Log->debug("RETURNED from callout '$name' with result '$result'"
                      . ', payload: ' . Dumper($payload));

            # Validate the result value returned by the callout.
            if (!any { m/$result/ } @valid_results) {
                confess "Callout '$name' returned invalid result: '$result'";
            }

            # TODO: Should we map the callout return values to
            # something else? What are the use cases for this?
            return wantarray ? ($result, $payload) : $result;
        };
        return;
    }
}

1;

