# -*- perl -*-
# $Id: $
#
## @author   dl-perf-qa@netapp.com
## @summary  Common callouts for QoS NATE/AAMP tests
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::QoS::Callouts

=head1 DESCRIPTION

Callout routines for Cluster QoS AAMP and NATE tests.

=cut

package Perf::AAMP::QoS::Callouts;

use strict;
use warnings;
use Carp;
use Data::Dumper;
use Data::Random qw(rand_words);
use List::MoreUtils qw(any none uniq);
use List::Util qw(max sum);
use NACL::C::Cluster;
use NACL::C::Disk;
use NACL::C::File;
use NACL::C::Lun;
use NACL::C::QosPolicyGroup;
use NACL::CS::QosPolicyGroup;
use NACL::C::QosWorkload;
use NACL::CS::QosWorkload;
use NACL::CS::StorageAggregate;
use NACL::C::Vserver;
use NACL::C::Volume;
use NATE::BaseException qw(:try);
use Params::Validate qw(:all);
use Perf::AAMP::Callout qw(:slsl_results);
use Perf::Stats;
use Readonly;
use Tharn qw($Log sleep);
use base q(Exporter);

our %EXPORT_TAGS = (
    all => [
        qw(
            create_dup_data
            define_policy_group
            disk_util_check
            cpu_util_check
            filersio_create_entry
            filersio_check_entry
            filersio_cleanup
            filersio_set_policy
            flush_cache
            get_sys_proc_names
            get_workload_name
            make_system_wklds_persist
            modify_policy_group
            policy_exists
            reconnect
            remove_system_wkld
            set_filersio_policy
            set_policy
            start_dedup
            start_filersio
            start_sample
            stop_sample
            toggle_nodes
            track_counters
            update_policy_groups
            workload_exists
        ),
    ],
);
our @EXPORT_OK = @{$EXPORT_TAGS{all}};

# Perf::Stats object for current iteration
Readonly::Scalar my $STATS_KEY => 'qos_stats_obj';
# QoS counters for specified aggregates
Readonly::Scalar my $AGGR_COUNTERS_KEY => 'qos_aggr_counters';
# Input parameters to track_counters() subroutine
Readonly::Scalar my $TRACK_COUNTERS_INPUT_KEY => 'track_counters_input';
# Boolean to indicate if the counters list/Perf::Stats object should be recreated
Readonly::Scalar my $REBUILD_COUNTER_LIST_KEY => 'rebuild_counter_list';

# For _debug_data_collection_* subroutines
# The subtests that have been started for sysstat
Readonly::Scalar my $SYSSTAT_SUBTESTS_KEY => 'sysstat_subtests';
# The directory in which to store the debug data logfiles
Readonly::Scalar my $DEBUG_DATA_OUTPUT_DIR_KEY => 'debug_data_output_dir';
# Constants defined for the desired disk utilization based on aggr type
Readonly::Hash my %DISK_UTIL => ('HyA' => 60, 'SSD' => 8, 'HDD' => 80);

=head2 define_policy_group

Create a QoS policy group & assign it to the specified data objects - i.e.
files, luns, volumes or vserver

=over

=item Arguments

B<name> = string

I<Optional> The name of the QoS policy group to be created. If not defined, a
random name is used.

B<containing_volume> = string

I<Optional> The name of the volume in which the file or lun is contained.

B<containing_vserver> = string

I<Required> The name of the vserver in which the QoS policy group is to be
created.

B<slo> = string

I<Optional> The QoS SLO/policy-group parameter to set on the policy group.

B<slo_value> = string

I<Optional> The value of the QoS SLO to set on the policy group.

B<files> = arrayref of string(s)

I<Optional> The names of the files on which to apply this policy group.

B<luns> = arrayref of string(s)

I<Optional> The names of the luns on which to apply this policy group.

B<volumes> = arrayref of string(s)

I<Optional> The names of the volumes on which to apply this policy group.

B<vserver> = arrayref of string(s)

I<Optional> The name of the vserver on which to apply this policy group.

=back

=cut

sub define_policy_group :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            name               => { type => SCALAR,   optional => 1 },
            # containing_volume: The volume in which the file/lun is contained.
            containing_volume  => { type => SCALAR,   optional => 1 },
            # containing_vserver: The vserver in which the policy-group is
            # created.
            containing_vserver => { type => SCALAR,   optional => 0 },
            slo                => { type => SCALAR,   optional => 1 },
            slo_value          => { type => SCALAR,   optional => 1 },
            files              => { type => ARRAYREF, optional => 1 },
            luns               => { type => ARRAYREF, optional => 1 },
            # volumes: The volumes on which to apply this policy group.
            volumes            => { type => ARRAYREF, optional => 1 },
            # vserver: The vserver on which to apply this policy group.
            vserver            => { type => SCALAR,   optional => 1 },
        },
    );

    my $ci = $resources->command_interface();
    my $stash = $resources->stash();

    my ($rand_word) = rand_words();
    # Remove non-word characters, since it would not be a valid name.
    $rand_word =~ s/\W/_/g;
    my $policy_group_name = $opts{name} || 'policy_' . $rand_word;

    # Create the policy group
    my $policy = NACL::C::QosPolicyGroup->create(
        command_interface => $ci,
        'policy-group'    => $policy_group_name,
        vserver           => $opts{containing_vserver},
    );
    if (defined $opts{slo}) {
        $policy->modify($opts{slo} => $opts{slo_value});
    }

    # Assign the created policy group to the specified data objects
    my %cmd_args = (
        command_interface  => $ci,
        vserver            => $opts{containing_vserver},
        'qos-policy-group' => $policy_group_name,
    );

    # Key-value pair of NACL modules for the data object types
    my %data_objects = (
        VolumeFile => $opts{files},
        Lun        => $opts{luns},
        Volume     => $opts{volumes},
        Vserver    => (defined $opts{vserver} ? [$opts{vserver}] : undef),
    );

    if (any { defined($_) } values %data_objects) {
        for my $module (keys %data_objects) {
            my $class = 'NACL::C::' . "$module";

            # Determine command arguments for different data object types
            my $obj_type;
            if ($module eq 'VolumeFile') {
                $obj_type = 'file';
                $cmd_args{volume} = $opts{containing_volume};
            }
            elsif ($module eq 'Lun') {
                $obj_type = lc($module);
                $cmd_args{volume} = $opts{containing_volume};
            }
            else {
                $obj_type = lc($module);
            }

            foreach my $obj (@{$data_objects{$module}}) {
                $cmd_args{$obj_type} = $obj;
                if (!defined $cmd_args{volume}) {
                    delete $cmd_args{volume};
                }
                $class->modify(%cmd_args);
            }
        }
    }

    # Update the Test Summary with the policy group.
    my $slsl_result = update_policy_groups(
        $resources,
        names => [$policy_group_name],
    );

    return ($slsl_result, $policy_group_name);
}

=head2 modify_policy_group

Modify an existing QoS policy-group, i.e set or modify an SLO on it.

=over

=item Arguments

B<name> = string

I<Required> The name of the QoS policy group.

B<slo> = string

I<Required> The QoS SLO/policy-group parameter to set on the policy group.

B<slo_value> = string

I<Required> The value of the QoS SLO to set on the policy group.

=back

=cut

sub modify_policy_group :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            name         => { type => SCALAR,  optional => 0 },
            slo          => { type => SCALAR,  optional => 0 },
            slo_value    => { type => SCALAR,  optional => 0 },
        },
    );

    my $ci = $resources->command_interface();
    my $stash = $resources->stash();

    NACL::C::QosPolicyGroup->modify(
        command_interface => $ci,
        'policy-group'    => $opts{name},
        $opts{slo}        => $opts{slo_value},
    );

    # Update the Test Summary with the policy group.
    return update_policy_groups($resources, names => [$opts{name}]);
}

=head2 update_policy_groups

Update the information about the QoS policy groups, if specified, or all the QoS
policy groups in this iteration, in the Test Summary. See Examples for when you
should call this method.

=over

=item Arguments

B<names> = arrayref of string(s)

I<Optional> The names of the QoS policy groups. Default is all the QoS policy
groups being observed in this iteration.

=item Example

example 1): test creates a new policy group(s) without using the
define_policy_group() method in this module:
    <create QoS Policy group using NACL command>
    <attach QoS Policy group to data object using NACL command>
    update_policy_groups($resources, names => ['policy1']);

example 2): test wants to track a pre-existing policy group in the cluster,
e.g.: test wants to track the preset 'User-Best-Effort' QoS policy group.
    update_policy_groups($resources, names => ['User-Best-Effort']);

example 3) test makes any change to a QoS policy group that was being
observed/tracked in the test.
    <create QoS Policy group 'policy1' & attach it to a data object>
    <do something>
    <add/remove/change the name of a data object from 'policy1'>
    update_policy_groups($resources);

=back

=cut

sub update_policy_groups :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            names         => { type => ARRAYREF,  optional => 1 },
        },
    );

    my $ci = $resources->command_interface();
    my $stash = $resources->stash();
    my @policy_names = ();

    if (defined $opts{names}) {
        @policy_names = @{$opts{names}};
    }
    else {
        my $slo_summary = $stash->slo_summary_fetch(
            iteration => $stash->current_iteration(),
            all_samples => 1,
        );
        @policy_names = keys %$slo_summary;
    }

    # Add each policy group, its workloads & SLOs to the Test Summary.
    for my $name (@policy_names) {
        # Get the workload names in the policy_group
        my $wkld_names = _get_workloads_in_policy_group(
            $resources,
            policy_group => $name,
        );

        # List of configurable QoS SLOs. Add to this list as more SLOs are
        # supported.
        my @slos = qw(max-throughput);

        # Get the SLOs for the policy_group
        my $policy = NACL::CS::QosPolicyGroup->fetch(
            command_interface => $ci,
            filter            => { 'policy-group' => $name },
        );
        my %policy_slos = map {+$_ => $policy->$_} @slos;

        # Test Summary Update:
        $stash->slo_summary_update(
            iteration => $stash->current_iteration(),
            policy_group => $name,
            slo => \%policy_slos,
            workloads => $wkld_names,
        );
    }

    # This method was called because the policy-group(s)/workload(s) in the test
    # changed. We will have to rebuild the counters list & the Perf::Stats
    # object. Set the flag to do this when appropriate.
    $stash->update(
        name  => $REBUILD_COUNTER_LIST_KEY,
        value => 1,
        create_if_nonexistent => 1,
    );

    return SLSL_RESULT_SUCCESS;
}

# Set specified policy via the qos_debug command.
# Deprecated. TODO: Remove.
sub start_filersio :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            volume   => { type => SCALAR,  optional => 0 },
            read_pct => { type => SCALAR,  optional => 0 },
            rand_pct => { type => SCALAR,  optional => 0 },
            mode     => { type => SCALAR,  default  => 'asyncio_pause' },
            pause    => { type => SCALAR,  default  => 1 },
            count    => { type => SCALAR,  default  => 2 },
            slsl     => { type => HASHREF, optional => 0 },
        },
    );

    my $node = $opts{slsl}->{node};
    my $ci = $resources->command_interface();
    my $apiset = $ci->get_7m_or_nodescope_apiset();

    # TODO: Need to execute command on node specified.
    $apiset->execute_command(
        command => "filersio $opts{mode} $opts{read_pct}"
                 . " -r $opts{rand_pct} 4k 0 10g 600 1 2"
                 . " /vol/$opts{volume}/file1 -synchronous -create");

    return SLSL_RESULT_SUCCESS;
}

=head2 filersio_create_entry

Run filersio briefly to get an entry in the QoS tables. We need to have an
entry in the table to track the policy group & workload during the iterations.
Also checks that the filersio entry is created in the QoS table.

=over

=item Arguments

B<node> = string

I<Required> The name of the node on which to create the entry.

B<vol_name> = string

I<Required> The name of the volume in which to create a file using filersio.

=back

=cut

sub filersio_create_entry :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            node       => { type => SCALAR,  optional => 0 },
            vol_name   => { type => SCALAR,  optional => 0 },
        },
    );

    my $ci = $resources->command_interface();
    my $apiset = NACL::APISet->new(
        name => $opts{node},
        category => 'Node',
        interface => 'CLI',
        set => 'Nodescope',
    );

    make_system_wklds_persist($resources, nodes => [$opts{node}]);

    # eval is a temporary workaround. NACL complains about unexpected output.
    # See burt643120.
    eval {
        $apiset->filersio_asyncio_pause(
            'read-percentage' => 0,
            'rand-percentage' => 0,
            'blk-size' => '4k',
            'start-byte' => 0,
            'end-byte' => '10k',
            'run-time' => 10,
            pause => 10,
            count => 1,
            filename => "/vol/$opts{vol_name}/entry_file",
            create => 1,
            synchronous => 1,
        );
    };
    $Log->warn("got NACL error erroneously: $@") if ($@);

    # Wait for the recently created filersio workload/policy-group instances
    # to appear in the QoS & CM statistics table before returning.
    Tharn::sleep(5);

    # Check that the filersio entry is created in QoS workload table.
    my ($slsl_status, $filersio_names) = get_sys_proc_names(
        $resources,
        policy_prefix => '_Filersio_plcy',
        workload_prefix => '_Filersio',
        node => $opts{node},
    );
    filersio_check_entry(
        $resources,
        node => $opts{node},
        workload_name => $filersio_names->{workload_name},
    );

    return SLSL_RESULT_SUCCESS;
}

=head2 filersio_check_entry

Check that an entry for Filersio exists in the QoS workload table. Check that
it's wid matches that from the qos debug output.

=over

=item Arguments

B<node> = string

I<Required> The name of the node on which to check the entry.

B<workload_name> = string

I<Required> The full name of the Filersio workload.

=back

=cut

sub filersio_check_entry :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            node            => { type => SCALAR,  optional => 0  },
            workload_name   => { type => SCALAR,  optional => 0 },
        },
    );

    my $ci = $resources->command_interface();
    my $apiset = NACL::APISet->new(
        name => $opts{node},
        category => 'Node',
        interface => 'CLI',
        set => 'Nodescope',
    );

    my $filersio_wkld =  NACL::CS::QosWorkload->fetch(
        command_interface => $ci,
        filter            => {workload => $opts{workload_name}},
    );
    my $show_wid = $filersio_wkld->wid();

    my $output = $apiset->execute_command(
        command => 'qos_debug filersio lookup',
    );
    $output =~ /Workload id is (\d+)/ or
        confess "Unrecognized output from lookup cmd.\n output = $output";
    my $lookup_wid = $1;

    if ($lookup_wid != $show_wid) {
        confess 'filersio wid from lookup & show commands do not match. '
                . 'This could result in incorrect classification. Ensure '
                . 'proper cleanup before the start of the test.';
    }
    else {
        return SLSL_RESULT_SUCCESS;
    }
}

=head2 filersio_set_policy

Configure the specified policy or SLO parameters for Filersio.

=over

=item Arguments

B<node> = string

I<Required> The name of the node on which to set the policy.

B<preset_policy> = C<besteffort | background>

I<Optional> The preset policy to set for Filersio. Supported policies are:
besteffort, background.

B<reserve> = string

I<Optional> The reserve (target-throughput) SLO value to set.

B<limit> = string

I<Optional> The limit (max-throughput) SLO value to set.

B<latency> = string

I<Optional> The latency (target-latency) SLO value to set.

B<units> = string

I<Optional> The units for the reserve or limit SLO values.

=back

=cut

sub filersio_set_policy :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            node          => { type => SCALAR,  optional => 0 },
            preset_policy => { type => SCALAR,  optional => 1 },
            reserve       => { type => SCALAR,  optional => 1 },
            limit         => { type => SCALAR,  optional => 1 },
            latency       => { type => SCALAR,  optional => 1 },
            units         => { type => SCALAR,  optional => 1 },
        },
    );

    # Delete options that are not 'filersio set' command arguments.
    my $node = delete $opts{node};
    my $preset_policy = delete $opts{preset_policy};

    my $ci = $resources->command_interface();
    my $apiset = NACL::APISet->new(
        name => $node,
        category => 'Node',
        interface => 'CLI',
        set => 'Nodescope',
    );

    # Commands are:
    # qos_debug filersio besteffort
    # qos_debug filersio background
    if (defined $preset_policy) {
        $apiset->execute_command(
            command => "qos_debug filersio $preset_policy"
        );
    }
    # qos_debug filersio set [-reserve <iops> [<burst>]] [-limit <iops> [<burst>]] [-latency <us>] [-units {ops | bytes | kbytes}]
    else {
        my @cmd_args = map { q{-} . $_ . q{ } . $opts{$_} . q{ } } keys %opts;
        $apiset->execute_command(
            command => "qos_debug filersio set @cmd_args",
        );
    }

    $apiset->execute_command(
        command => 'qos_debug filersio get',
    );

    return SLSL_RESULT_SUCCESS;
}

=head2 filersio_cleanup

Cleanup remnants of filersio, if existent.

=over

=item Arguments

B<node> = string

I<Required> The name of the node on which to cleanup the entry.

=back

=cut

sub filersio_cleanup :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            node            => { type => SCALAR,  optional => 0 },
        },
    );

    my $ci = $resources->command_interface();
    my $apiset = NACL::APISet->new(
        name => $opts{node},
        category => 'Node',
        interface => 'CLI',
        set => 'Nodescope',
    );

    try {
        $apiset->filersio_stop();
    }
    catch NACL::APISet::Exceptions::CommandFailedException with {
        my $exception = shift;
        my $err_msg = $exception->text();
        if ($err_msg !~ /no instances currently running/) {
            confess "Error: $err_msg";
        }
    };

    # Delete the filersio workload explicitly to remove cached wid entry
    $apiset->execute_command(
        command => 'qos_debug workload_delete_later clear; '
        .  'qos_debug filersio delete; qos_debug filersio delete',
    );

    # If filersio is now classified as System-Default (wid=1), then the filersio
    # workload has been deleted & cleanup was successful.
    my $output = $apiset->execute_command(
        command => 'qos_debug filersio lookup',
    );
    confess 'filersio workload has its own wid. filersio cleanup failed.'
        if ($output !~ /Workload id is 1/);

    # Check that the Filersio entry is deleted from the QoS workload table.
    my $filersio_wkld;
    try {
        $filersio_wkld = NACL::CS::QosWorkload->fetch(
            command_interface => $ci,
            filter            => {workload => '.*Filersio'},
        );
    }
    catch NACL::Exceptions::NoElementsFound with {
        # NO-OP
    };
    confess 'Filersio workload not deleted from the ngsh QoS workload table'
        if (defined $filersio_wkld);

    return SLSL_RESULT_SUCCESS;
}

# Verify that the specified QoS workload exists in the Workload table
sub workload_exists :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            workload_name => { type => SCALAR, optional => 0 },
        },
    );

    my $ci = $resources->command_interface();

    my $cluster_node = NACL::CS::Cluster->fetch(
        command_interface => $ci,
        filter            => {node => $ci->node()},
    );
    my $uuid = $cluster_node->uuid();
    my $workload_name = join('_', $opts{workload_name}, $uuid);

    my $wkld = NACL::CS::QosWorkload->fetch(
        command_interface => $resources->command_interface(),
        filter            => { workload => $workload_name },
    );

    if (defined $wkld) {
        return SLSL_RESULT_SUCCESS;
    }
    else {
        confess "QoS workload '$workload_name' not found";
    }
}


# Verify that the specified QoS policy exists in the Policy table
sub policy_exists :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            policy_name => { type => SCALAR, optional => 0 },
        },
    );

    my $ci = $resources->command_interface();

    my $cluster_node = NACL::CS::Cluster->fetch(
        command_interface => $ci,
        filter            => {node => $ci->node()},
    );
    my $uuid = $cluster_node->uuid();
    my $policy_name = join('_', $opts{policy_name}, $uuid);

    my $policy = NACL::CS::QosPolicyGroup->fetch(
        command_interface => $ci,
        filter            => { 'policy-group' => $policy_name },
    );
    if (defined $policy) {
        return SLSL_RESULT_SUCCESS;
    }
    else {
        confess "QoS policy '$policy_name' not found";
    }

    return SLSL_RESULT_SUCCESS;
}


=head2 get_workload_name

Return the name of the QoS workload for the specified object. Return undef, if
the QoS workload for the object is not found.
Be as specific as possible when specifying an object to avoid conflicts, i.e.
for a lun, specify the lun name & the names of the volume & the vserver the lun
is in.

=over

=item Arguments

B<file> = string

I<Optional> The name of the file whose workload name you want to retrieve.

B<lun> = string

I<Optional> The name of the lun whose workload name you want to retrieve.

B<volume> = string

I<Optional> The name of the volume whose workload name you want to retrieve or
the volume in which the lun or file is contained.

B<vserver> = string

I<Optional> The name of the vserver whose workload name you want to retrieve or
the vserver in which the lun, file or volume is contained.

=back

=cut

sub get_workload_name :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            file           => { type => SCALAR, optional => 1 },
            lun            => { type => SCALAR, optional => 1 },
            volume         => { type => SCALAR, optional => 1 },
            vserver        => { type => SCALAR, optional => 1 },
        },
    );

    my $ci = $resources->command_interface();

    my $wkld = NACL::CS::QosWorkload->fetch(
        command_interface => $ci,
        filter            => {%opts},
    );
    my $wkld_name = $wkld->workload();

    return (SLSL_RESULT_SUCCESS, $wkld_name);
}


=head2 get_sys_proc_names

The node-specific system processes have a QoS workload & policy on each node in
the cluster whose name is tied to the node's UUID. Return the names of the QoS
workload and Policy group for the specified system process on the specified
node.

=over

=item Arguments

B<policy_prefix> = string

I<Required> The policy name prefix of the system process whose policy name you
want to retrieve.

B<workload_prefix> = string

I<Required> The workload name prefix of the system process whose policy name you
want to retrieve.

B<node> = string

I<Required> The name of the node on which you want to retrieve the system
process workload & policy name.

=item Example

For the BLKR_REALLOC system process, the QoS workload & policy group entries
are:
workload is _WID_SCAN_BLKR_METRICS_<node_uuid>
policy group is _POL_SCAN_BLKR_METRICS_<node_uuid>
To get the full workload & policy group names for the BLKR_REALLOC system
process on NodeA:

    my $sys_proc_names = get_sys_proc_names(
        $resources,
        policy_prefix => '_POL_SCAN_BLKR_METRICS',
        workload_prefix => '_WID_SCAN_BLKR_METRICS',
        node => 'NodeA',
    );
    my $sys_policy_name = $sys_proc_names->{policy_name};
    my $sys_workload_name = $sys_proc_names->{workload_name};

=back

=cut

sub get_sys_proc_names :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            policy_prefix   => { type => SCALAR, optional => 0},
            workload_prefix => { type => SCALAR, optional => 0},
            node            => { type => SCALAR, optional => 0},
        },
    );

    my $ci = $resources->command_interface();
    my %sys_proc = ();

    my $cluster_node = NACL::CS::Cluster->fetch(
        command_interface => $ci,
        filter            => {node => $opts{node}},
    );
    my $uuid = $cluster_node->uuid();
    $sys_proc{policy_name} = join('_', $opts{policy_prefix}, $uuid);
    $sys_proc{workload_name} = join('_', $opts{workload_prefix}, $uuid);

    return (SLSL_RESULT_SUCCESS, \%sys_proc);
}

=head2 flush_cache

Flush the buffer cache on the specified nodes using the nodescope command
'wafl flush cache'.

=over

=item Arguments

B<nodes> = arrayref of string(s)

I<Optional> The names of the nodes on which you want to flush the buffer cache.
Default is all nodes in the cluster.

=back

=cut

sub flush_cache :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            nodes  => { type => ARRAYREF, optional => 1},
        },
    );

    my $ci = $resources->command_interface();
    my @nodes = NACL::C::Cluster->nodes(command_interface => $ci);
    my @nodes_list;

    if (defined $opts{nodes}) {
        # Check that the specified nodes are in the cluster under test.
        foreach my $node (@{$opts{nodes}}) {
            confess "Node '$node' not one of test cluster nodes (@nodes)."
                if (none { $_ eq $node } @nodes);
        }
        @nodes_list = @{$opts{nodes}};
    }
    else {
        # Flush cache on all nodes in the cluster.
        @nodes_list = @nodes;
    }

    foreach my $node_name (@nodes_list) {
        my $apiset = NACL::APISet->new(
            name => $node_name,
            category => 'Node',
            interface => 'CLI',
            set => 'Nodescope',
        );
        $apiset->wafl_flush_cache();
    }

    return SLSL_RESULT_SUCCESS;
}


# Make system workloads persistent, i.e. disable their automatic deletion upon
# completion of the system work.
sub make_system_wklds_persist :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            nodes  => { type => ARRAYREF, optional => 1},
        },
    );

    my $ci = $resources->command_interface();
    my @nodes = NACL::C::Cluster->nodes(command_interface => $ci);
    my @nodes_list;

    if (defined $opts{nodes}) {
        # Check that the specified nodes are in the cluster under test.
        foreach my $node (@{$opts{nodes}}) {
            confess "Node '$node' not one of test cluster nodes (@nodes)."
                if (none { $_ eq $node } @nodes);
        }
        @nodes_list = @{$opts{nodes}};
    }
    else {
        # Do on all nodes in the cluster.
        @nodes_list = @nodes;
    }

    foreach my $node_name (@nodes_list) {
        my $apiset = NACL::APISet->new(
            name => $node_name,
            category => 'Node',
            interface => 'CLI',
            set => 'Nodescope',
        );
        $apiset->execute_command(
            command => 'qos_debug workload_delete_later set;'
        );
    }

    return SLSL_RESULT_SUCCESS;
}


# Enable automatic deletion of system workloads & delete the specified system
# workload
# No longer needed. TODO: Remove.
sub remove_system_wkld :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            workload_name => { type => SCALAR, optional => 0 },
        },
    );

    my $ci = $resources->command_interface();
    my $apiset = $ci->get_7m_or_nodescope_apiset();
    my $workload = $opts{workload_name};

    $apiset->execute_command(
        command => "qos_debug $workload delete;"
    );

    return SLSL_RESULT_SUCCESS;
}

=head2 track_counters

Create a Perf::Stats object with the list of counters to be tracked.
Note: This was split from start_sample because building the instances &
collecting the counters takes a significant amount of time & could interfere
with the measurement cycle.

=over

=item Arguments

B<counters> = arrayref of CM counters

I<Optional> The list of CM counters to be tracked for the test. The format
must comply with that supported by Perf::Stats.pm

B<aggr_names> = arrayref of string(s)

I<Optional> The names of the aggregates whose disk utilization counters are to
be tracked. You must specify this parameter if you want to do the disk
utilization check in your test.

B<slsl> = string

I<Optional> For now, the AAMP tests pass AAMP object information via the slsl
parameter. Specify the aggregates, etc. via this paramter.
This parameter will be removed once the AAMP task for slsl args is fixed.

=back

=cut

sub track_counters :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            counters    => { type => ARRAYREF|UNDEF, optional => 1 },
            aggr_names  => { type => ARRAYREF|UNDEF, optional => 1 },
            # temporary: till the AAMP task for slsl args is fixed.
            slsl => { type => HASHREF, optional => 1 },
        },
    );
    # Note: Do not modify the input parameters (%opts) to this subroutine. They
    # are used later.

    my $ci = $resources->command_interface();
    my $stash = $resources->stash();
    my $filers = $resources->filers();
    my @counters_to_track = defined $opts{counters} ? @{$opts{counters}} : ();
    my @aggr_names = defined $opts{aggr_names} ? @{$opts{aggr_names}} : ();
    my @aggr_counters = ();

    # temporary: AAMP tests pass AAMP object info via the slsl parameter.
    # validate slsl parameter & get the aggr names.
    if (defined $opts{slsl}) {
        my @slsl_params = %{$opts{slsl}};
        my %slsl_opts = validate(
            @slsl_params, {
                aggrs => { type => ARRAYREF, optional => 0 },
            },
        );
        $opts{slsl} = \%slsl_opts;
        my $aggr_list = $opts{slsl}->{aggrs};
        @aggr_names = map {$_->{name}} @$aggr_list;
    }

    # If aggregate names are specified, then add the disk utilization counters
    # to the counters list.

    # Get the instance names for the specified aggregates
    foreach my $aggr (@aggr_names) {
        my $aggr_obj = NACL::CS::StorageAggregate->fetch(
            command_interface => $ci,
            filter => {aggregate => $aggr},
        );

        my $aggr_uuid = $aggr_obj->uuid();
        my $aggr_instance = _get_aggr_instance(
                                $resources,
                                aggr_obj => $aggr_obj
                             );
        my $aggr_node = $aggr_obj->nodes()->[0];

        my $aggr_ctrs = {
            node     => $aggr_node,
            object   => 'resource',
            instance => "$aggr_instance",
            counter  => '(busy_percent|num_instances)',
        };
        push(@counters_to_track, $aggr_ctrs);
        push(@aggr_counters, $aggr_ctrs);
    }

    # Get the names of the QoS workloads & policy-groups in the test.
    my $slo_summary = $stash->slo_summary_fetch(
        iteration => $stash->current_iteration(),
        all_samples => 0,
    );
    my @wkld_names = uniq map { map { @{$_->{workloads}} } @{$_} }
                     values %$slo_summary;
    my @policy_names = keys %$slo_summary;

    # Get the policy-group-id for each policy group.
    my @policy_ids;
    foreach my $pol_name (@policy_names) {
        my $policy_obj = NACL::C::QosPolicyGroup->find(
            command_interface => $ci,
            filter => {'policy-group' => $pol_name},
        );
        my $pgid = $policy_obj->get_one_state_attribute('pgid');
        push(@policy_ids, $pgid);
    }

    # Add the QoS CM workload object counters for the workloads in the test to
    # the counters list.
    my @wkld_ctrs = map {
        {object => 'workload', instance => $_, counter => '.*'}
        } @wkld_names;
    push(@counters_to_track, @wkld_ctrs);

    # Add the QoS CM ratebucket object for the policy-groups in the test to the
    # counters list.
    my @ratebucket_ctrs = map {
        {object => 'ratebucket', instance => '.*\.' . $_, counter => '.*'}
        } @policy_ids;
    push(@counters_to_track, @ratebucket_ctrs);

    # TODO: Add the workload:policy_group combo object to the counters list.

    # Wait for the possibly recently created workload/policy-group instances
    # to appear in the statistics table.
    Tharn::sleep(5);

    $Log->debug("Tracking counters: @counters_to_track");
    my $stats = Perf::Stats->new(
        node     => $filers->[0],
        counters => \@counters_to_track,
        source   => 'statistics',
        all_nodes_in_cluster => 1,
            force_multinode      => 1,
            debug    => 1,
    );

    # Store what will be required later, in the Stash.
    my @store_in_stash = (
        # The Perf::Stats object tracking the QoS counters
        {name => $STATS_KEY, value => $stats},
        # The names of the aggregate's busy_percent counters
        {name => $AGGR_COUNTERS_KEY, value => \@aggr_counters},
        # The input parameters to this subroutine, in the event that we need to
        # call it internally
        {name => $TRACK_COUNTERS_INPUT_KEY, value => \%opts},
        # We've just built the counters list & created the Perf::Stats object,
        # reset the flag
        {name => $REBUILD_COUNTER_LIST_KEY, value => 0},
    );

    foreach (@store_in_stash) {
        $stash->update(
            name  => $_->{name},
            value => $_->{value},
            create_if_nonexistent => 1,
        );
    }

    return SLSL_RESULT_SUCCESS;
}

=head2 start_sample

Start the sampling of the counters specified in track_counters().

=over

=item Arguments

None

=back

=cut

sub start_sample :Callout {
    my $resources = shift;

    my $stash = $resources->stash();

    # Call track_counters() to rebuild the counters in the Perf::Stats object
    # if the rebuild_counter_list flag is set.
    my $rebuild_counter_list = $stash->fetch(name => $REBUILD_COUNTER_LIST_KEY);
    if ($rebuild_counter_list) {
        my $ctrs_input = $stash->fetch(name => $TRACK_COUNTERS_INPUT_KEY);
        track_counters($resources, %$ctrs_input);
    }

    # Note: track_counters creates a new Perf::Stats object, so this fetch must
    # occur after any calls to track_counters().
    my $stats = $stash->fetch(name => $STATS_KEY);

    $stats->start();

    return SLSL_RESULT_SUCCESS;
}

=head2 stop_sample

Stop the sampling of counters started by start_sample() & save the counter
values.

=over

=item Arguments

None

=back

=cut

sub stop_sample :Callout {
    my $resources = shift;
    my $stash = $resources->stash();

    my $stats = $stash->fetch(name => $STATS_KEY);
    my $sample = $stats->stop(exact_format => 0);
    $Log->comment('stats sample: ' . Dumper($sample));

    # Test Summary Update: Store the stats sample in the Test Summary.
    $stash->statistics_summary_update(
        iteration => $stash->current_iteration(),
        sample    => $sample,
    );

    return SLSL_RESULT_SUCCESS;
}

=head2 toggle_nodes

This is a wrapper around Perf::Stats::toggle_nodes(). Inform Perf::Stats about
nodes going up/down in the cluster. This method must be called before the
node(s) goes down.

=over

=item Arguments
Same as those to Perf::Stats::toggle_nodes(). Copied here.
See L<Perf::Stats::toggle_nodes()| Perf::Stats/toggle_nodes> for arguments list.

B<nodes> = arrayref of string(s)

I<Required> Arrayref of node names to toggle.

B<toggle> = boolean

I<Required> 1 to activate, 0 to deactivate.

=back

=cut

sub toggle_nodes :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            nodes =>  {type => ARRAYREF, optional => 0},
            toggle => {type => SCALAR, optional => 0},
        }
    );

    my $stash = $resources->stash();

    my $stats = $stash->fetch(name => $STATS_KEY);

    $stats->toggle_nodes(%opts);

    return SLSL_RESULT_SUCCESS;
}

=head2 disk_util_check

Determine the disk utilization of the aggregates in the test, check if it
is acceptable & return the appropriate result.

=over

=item Arguments

B<iteration_name> = string

I<Optional> The name of the iteration for which you want to do this check.
Default is the current iteration.

B<iteration_type> = C<baseline | test>

I<Optional> The type of the iteration for which you want to do this check.
Supported iteration types are: baseline, test. Default is 'baseline'.

B<util> = number

I<Optional> The desired utilization value to be achieved in the test.
If undefined, desired utilization defaults to 80 for HDDs, 8 for SSDs, 60 for HyAs
(Value is expressed as a number, i.e. for 80%, specify 80)

=back

=cut

sub disk_util_check :Callout {
    my $resources = shift;
    my $stash = $resources->stash();
    my %opts = validate(
        @_, {
            iteration_name => {type => SCALAR, optional => 1, default => $stash->current_iteration()},
            iteration_type => {type => SCALAR, optional => 1, default => 'baseline'},
            util           => {type => SCALAR | UNDEF, optional => 1, default => undef},
        },
    );

    # Get the statistics sample for the iteration from the Test Summary.
    my $sample = $stash->statistics_summary_fetch(
        iteration => $opts{iteration_name},
    )->[0];

    # Get the aggregate counter names to calculate disk utilization.
    my $aggr_counters = $stash->fetch(name => $AGGR_COUNTERS_KEY);

    foreach my $cm_hash (@$aggr_counters) {
       my $bp;
       my $disk_util;
       my $obj_instance;
       my $desired_util;
       if ($cm_hash->{instance} =~ /(DISK_[HDD|SSD].*)\|(DISK_[HDD|SSD].*)/) {
           # if hybrid aggregate used, use util value constant as defined
           $desired_util = $opts{util} || $DISK_UTIL{'HyA'};
           my @aggr_instances = split /\|/ , $cm_hash->{instance};
           my @obj_instances = map {join (':', $cm_hash->{node},
                                    $cm_hash->{object}, $_)}
                                    @aggr_instances;
           my @sum_bps = map {join (':', $_, 'busy_percent')} @obj_instances;
           my @sum_num_instances = map {join (':', $_, 'num_instances')}
                                   @obj_instances;
           # TODO: Research and investigate the correct way to
           # identify the disk utilization percent across multiple disk types
           $disk_util = (sum map {$sample->{$_}} @sum_bps) /
                        (sum map {$sample->{$_}} @sum_num_instances);
        }
        else {
           # If SSD disks used, use util value constant as defined
           if ($cm_hash->{instance} =~ /DISK_SSD/) {
              $desired_util = $opts{util} || $DISK_UTIL{'SSD'};
           }
           else {
              $desired_util = $opts{util} || $DISK_UTIL{'HDD'};
           }
           $obj_instance = join(':', $cm_hash->{node}, $cm_hash->{object},
                                   $cm_hash->{instance});
           $bp = join(':', $obj_instance, 'busy_percent');
           my $num_instances = join(':', $obj_instance, 'num_instances');
           $disk_util = $sample->{$bp} / $sample->{$num_instances};
        }
        if ($disk_util < $desired_util) {
            # Utilization of this aggr instance is < required.
            if ($opts{iteration_type} eq 'baseline') {
                return SLSL_RESULT_CONTINUE;
            }
            else {
                $Log->error('Disk utilization less than expected in iteration. '
                            . "$bp = $disk_util");
                return SLSL_RESULT_FAILURE;
            }
        }
        else {
            $Log->debug("Disk utilization check succeeded for '$obj_instance'. "
                        . "$bp = $disk_util");
            next;
        }
    }

    # Utilization of all aggregates was >= required. Flag this iteration as
    # successful.
    $Log->comment('Disk utilization check succeeded.');

    if ($opts{iteration_type} eq 'baseline') {
        return SLSL_RESULT_STOP;
    }
    else {
        return SLSL_RESULT_SUCCESS;
    }
}

=head2 cpu_util_check

Determine the CPU utilization of the nodes receiving I/O in the test, check if
it is acceptable & return the appropriate result. Complain if the disk
utilization is above a certain value - this is to ensure that the test is
CPUbound only and is not Diskbound.

=over

=item Arguments

B<iteration_name> = string

I<Optional> The name of the iteration for which you want to do this check.
Default is the current iteration.

B<iteration_type> = C<baseline | test>

I<Optional> The type of the iteration for which you want to do this check.
Supported iteration types are: baseline, test. Default is 'baseline'.

B<util> = number

I<Optional> The desired utilization value to be achieved in the test.
(Value is expressed as a number, i.e. for 80%, specify 80)

=back

=cut

sub cpu_util_check :Callout {
    my $resources = shift;
    my $stash = $resources->stash();
    my %opts = validate(
        @_, {
            iteration_name => {type => SCALAR, optional => 1, default => $stash->current_iteration()},
            iteration_type => {type => SCALAR, optional => 1, default => 'baseline'},
            util           => {type => SCALAR, optional => 1, default => 80},
        },
    );

    # Get the statistics sample for the iteration from the Test Summary.
    my $sample = $stash->statistics_summary_fetch(
        iteration => $opts{iteration_name},
    )->[0];

    # Get the aggregate counter names to calculate disk utilization.
    my $aggr_counters = $stash->fetch(name => $AGGR_COUNTERS_KEY);

    foreach my $cm_hash (@$aggr_counters) {
        my $obj_instance = join(':', $cm_hash->{node}, $cm_hash->{object},
                                $cm_hash->{instance});
        my $bp = join(':', $obj_instance, 'busy_percent');
        my $num_instances = join(':', $obj_instance, 'num_instances');

        my $disk_util = $sample->{$bp} / $sample->{$num_instances};

        # Check that the test is not diskbound. At 80% utilization, QoS
        # scheduling/message re-ordering kicks in. So, disk utilization must
        # stay lower than that for a CPUbound test.
        if ($disk_util > 70) {
            $Log->error("Disk utilization is too high for '$obj_instance' "
                        . "$bp = $disk_util");
            return SLSL_RESULT_FAILURE;
        }
    }

    # Get the tracked counters for CPU utilization.
    my @cpu_util_ctrs = grep {$_ =~ /system:system:avg_processor_busy/}
        keys %$sample;

    for my $bp (@cpu_util_ctrs) {
        my $cpu_util = $sample->{$bp};

        if ($cpu_util < $opts{util}) {
            # Utilization of this instance is < required.
            if ($opts{iteration_type} eq 'baseline') {
                return SLSL_RESULT_CONTINUE;
            }
            else {
                $Log->error('CPU utilization less than expected in iteration. '
                            . "$bp = $cpu_util");
                return SLSL_RESULT_FAILURE;
            }
        }
        else {
            $Log->debug('CPU utilization check succeeded. '
                        . "$bp = $cpu_util");
            next;
        }
    }

    # Utilization of all instances was >= required. Flag this iteration as
    # successful.
    $Log->comment('CPU utilization check succeeded.');

    if ($opts{iteration_type} eq 'baseline') {
        return SLSL_RESULT_STOP;
    }
    else {
        return SLSL_RESULT_SUCCESS;
    }
}

=head2 reconnect

Reestablish the NACL connections created by the perf::qos::control::QoSTest
package. It is the responsibility of user code to reestablish any connections
created by it.

(This would typically be called when nodes reboot, etc.)

=cut

sub reconnect :Callout {
    my $resources = shift;
    my $ci = $resources->command_interface();
    my $stash = $resources->stash();

    # Reconnect the main command interface.
    $ci->refresh_command_interface();

    # Reconnect Perf::Stats interfaces.
    if ($stash->exists(name => $STATS_KEY)) {
        my $stats = $stash->fetch(name => $STATS_KEY);
        $stats->reconnect();
        $Log->comment('Reestablished resource command interface'
                    . " and '$STATS_KEY' connections");
    }
    else {
        $Log->comment('Reestablished resource command interface connection');
    }

    return SLSL_RESULT_SUCCESS;
}

### Private functions:


# Return the workload names contained in the specified policy_group.
# Arguments:
# policy_group => Name of the QoS policy group.
sub _get_workloads_in_policy_group {
    my $resources = shift;
    my %opts = validate(
        @_, {
            policy_group    => { type => SCALAR, optional => 0 },
        },
    );

    my $ci = $resources->command_interface();
    my @wklds = ();

    try {
        @wklds = NACL::CS::QosWorkload->fetch(
            command_interface => $ci,
            filter            => {
                'policy-group' => $opts{policy_group},
                class => '*',
            },
        );
    }
    catch NACL::Exceptions::NoElementsFound with {
        $Log->comment("No workloads in policy group $opts{policy_group}");
    };

    my @wkld_names = @wklds ? map {$_->workload()} @wklds : ();

    return \@wkld_names;
}

# Return the instance of the aggregate in use (HDD/SSD/Hybrid).
# Arguments:
# aggr_obj => NACL object of the aggregate being used
sub _get_aggr_instance {
    my $resources = shift;
    my %opts = validate(
        @_, {
            aggr_obj    => { type => OBJECT, optional => 0 },
        },
    );
    my $ci = $resources->command_interface();
    my $aggr_disk_type;
    my $aggr_obj = $opts{aggr_obj};
    my $aggr_name = $aggr_obj->aggregate;
    my $aggr_uuid = $aggr_obj->uuid;
    my @disk_list = $aggr_obj->disklist;
    if ($aggr_obj->hybrid_enabled() eq 'true') {
       $aggr_disk_type = 'hybrid';
    }
    else {
       my @disk_types = map {NACL::C::Disk->find(
                                command_interface => $ci,
                                filter => {disk => $_}
                            )->get_one_state_attribute('type')}
                            @disk_list;

       my $disk_num = scalar uniq @disk_types;
       if ($disk_num == 1) {
          $aggr_disk_type = $disk_types[0];
       }
       else {
          confess "More than one disktype identified on a non-hybrid aggregate";
       }
    }

    my $aggr_instance;
    if ($aggr_disk_type eq 'SSD') {
        $aggr_instance = join('_', 'DISK_SSD', $aggr_name, $aggr_uuid);
    }
    elsif ($aggr_disk_type eq 'hybrid') {
        # If hybrid aggregate, track HDD and SSD instances
        my $aggr_hdd_instance = join('_', 'DISK_HDD', $aggr_name, $aggr_uuid);
        my $aggr_ssd_instance = join('_', 'DISK_SSD', $aggr_name, $aggr_uuid);
        $aggr_instance = join('|', $aggr_hdd_instance, $aggr_ssd_instance);
    }
    else {
        $aggr_instance = join('_', 'DISK_HDD', $aggr_name, $aggr_uuid);
    }

    return $aggr_instance;
}

# Run commands to initialize the data collection required for debugging testcase
# failures. Some of the commands run in this subroutine will zero counters. So
# be careful when calling this subroutine.
# This is a temporary way to provide this data to the Dev team until
# ASUP, CM Archiver, Onboard perfstats are in place.
sub _debug_data_collection_start {
    my $resources = shift;

    my $ci = $resources->command_interface();
    my $stash = $resources->stash();
    my $current_iteration = $stash->current_iteration();

    # Create the directory for the current iteration. This is where all the
    # output/log files will be stored.
    # Note:This creates sub-directories for iterations in the .thpl's directory.
    # This does not create a unique directory for each testcase. Hence this will
    # overwrite the logfiles when multiple testcases are run within a .thpl.
    my $output_dir = "$::LOGDIR/$::RUNID/$current_iteration";
    make_path($output_dir);
    # Store the directory path in the Stash for use later.
    $stash->update(
        name  => $DEBUG_DATA_OUTPUT_DIR_KEY,
        value => $output_dir,
        create_if_nonexistent => 1,
    );

    my @nodes = NACL::C::Cluster->nodes(command_interface => $ci);
    my @sysstat_subtests;

    # Run the dblade commands on each node in the cluster.
    foreach my $node_name (@nodes) {
        my $apiset = NACL::APISet->new(
            name => $node_name,
            category => 'Node',
            interface => 'CLI',
            set => 'Nodescope',
        );

        # Using execute_command() instead of the APISet methods because we
        # just want the raw output.
        my @commands = (
            'ps -z',
            'statit -b',
            # Collecting all objects & counters takes over 2 minutes. Hence
            # only collecting the specified objects.
            'stats start cifs disk edfifo lun nfsv3 nfsv4 nwk_ctx ratebucket '
            . 'resource system volume workload workload_detail',
            'waffinity_stats -z', 'wafl_susp -z',
        );
        foreach my $cmd (@commands) {
            $apiset->execute_command(command => $cmd);
        }

        # Run sysstat commands as subtests.
        my @sysstat_commands = ('sysstat -x 1', 'sysstat -M');
        foreach my $cmd (@sysstat_commands) {
            # The file to store the output in.
            my $filename = $cmd;
            # Remove any spaces & '-' from the filename.
            $filename =~ s/\s+/_/g;
            $filename =~ s/-//g;

            my $sysstat_subtest = subtest(
                sub {
                $apiset->execute_command(
                    command => $cmd,
                    'connectrec-timeout' => -1);
                },
                '-runid', join('_', $current_iteration, $node_name, $filename),
                -bg,
            );
            push (@sysstat_subtests, $sysstat_subtest);
        }
    }

    # Store the sysstat subtests in the Stash for cleanup later.
    $stash->add(
        name  => $SYSSTAT_SUBTESTS_KEY,
        value => [@sysstat_subtests],
    );

    return SLSL_RESULT_SUCCESS;
}

# Run commands to stop the data collection required for debugging testcase
# failures. This is a temporary way to provide this data to the Dev team until
# ASUP, CM Archiver, Onboard perfstats are in place.
sub _debug_data_collection_stop {
    my $resources = shift;

    my $ci = $resources->command_interface();
    my $stash = $resources->stash();
    my $current_iteration = $stash->current_iteration();

    # Fetch the directory in which to store the debug data logfiles.
    my $output_dir = $stash->fetch(name => $DEBUG_DATA_OUTPUT_DIR_KEY);
    confess "Directory to store the logfiles does not exist.($output_dir)"
        if (! -d $output_dir);

    my @nodes = NACL::C::Cluster->nodes(command_interface => $ci);

    # Run the dblade commands on each node in the cluster.
    foreach my $node_name (@nodes) {
        my $apiset = NACL::APISet->new(
            name => $node_name,
            category => 'Node',
            interface => 'CLI',
            set => 'Nodescope',
        );
        # For write tests, some of the commands below dump a lot of output that
        # takes a long time. This increases our iteration time significantly,
        # hence we should remove this data collection ASAP.
        $apiset->set_timeout('timeout' => 300);

        # Using execute_command() instead of the APISet methods because we
        # just want the raw output.
        my @commands = (
            'ps -s',
            'statit -e',
            'stats stop',
            'waffinity_stats',
            'wafl_susp -w',
            'wafl bufstats',
            'wafl cpstats',
        );
        foreach my $cmd (@commands) {
            my $raw_output = $apiset->execute_command(command => $cmd);

            # Store the output of each command in a separate file.
            my $filename = $cmd;
            # Remove any spaces & '-' from the filename.
            $filename =~ s/\s+/_/g;
            $filename =~ s/-//g;
            my $output_file = "$output_dir/$node_name" . '_' . "$filename\.log";
            open(my $fh_out, '>', $output_file)
                or confess "Cannot open '$output_file': $!";

            print {$fh_out} "command: $cmd\n";
            print {$fh_out} "$raw_output\n";

            close($fh_out)
                or $Log->warn("Cannot close '$output_file': $!");
        }
    }

    # Kill the subtests running the sysstat command.
    my $sysstat_subtests = $stash->fetch(name => $SYSSTAT_SUBTESTS_KEY);
    foreach my $subtest (@$sysstat_subtests) {
        $subtest->kill();
    }
    $stash->remove(name => $SYSSTAT_SUBTESTS_KEY);

    return SLSL_RESULT_SUCCESS;
}

### For Demo only. Stubs.

sub create_dup_data :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            volume      => { type => SCALAR,  optional => 0 },
            dedup_args  => { type => SCALAR,  optional => 0 },
        },
    );

    my $ci = $resources->command_interface();
    ###
    # Use your favorite library to create dup data.
    ###
    return SLSL_RESULT_SUCCESS;
}


sub start_dedup :Callout {
    my $resources = shift;
    my %opts = validate(
        @_, {
            volume      => { type => SCALAR,  optional => 0 },
            dedup_args  => { type => SCALAR,  optional => 0 },
        },
    );

    my $ci = $resources->command_interface();
    ###
    # Use your favorite library to configure & start dedup.
    ###
    return SLSL_RESULT_SUCCESS;
}

1;
