version = "1.2.1"

import inspect
from pprint import pprint
import random

from aamp.slsl.macros import *
from aamp.slsl.setlist_parser import (aggregate,callout,cluster,controller,
    datastore,event,io_path,iteration,parameter,path_mapper,test_runner,volume,
    vserver,workload)
from aamp.test_execution.iteration import Iteration
from aamp.test_execution.test_runner import TestRunner

from events import Events

# Test parameters specified via PerfANT or via NATE
test_parameters = {
    ### PerfANT parameters.
    # Default values specified for these parameters are dummy values & will
    # *not* be used in the test.
    'purpose': parameter(type="string", name="purpose", default_value=None,
                         source='perfant'),
    'aggr_list': parameter(type="list", name="aggr_list", default_value="aggr1",
                           source='perfant'),
    'model': parameter(type="string", name="model", default_value=None,
                       source='perfant'),
    'num_nodes': parameter(type="int", name="num_nodes", default_value=2,
                           source='perfant'),
    'num_clients': parameter(type="int", name="num_clients", default_value=2,
                             source='perfant'),
    'data_nics': parameter(type="list", name="data_nics",
                           default_value=['e1b', 'e2b'], source='perfant'),
    'eth_iface': parameter(type="string", name="client_ethernet_nics",
                           default_value="eth1", source='perfant'),
    'fcp_targets': parameter(type="list", name="fcp_targets",
                             default_value=['6a', '6b'], source='perfant'),
    'initiators': parameter(type="list", name="initiators", default_value=['2a'],
                            source='perfant'),

    ### SLSL user parameters.
    # Default values specified for these parameters will be used in the test.
    # QoS test type
    'test_type': parameter(type="string", name="test_type",
                           default_value='Diskbound'),
    # Runtime for user workloads
    'runtime': parameter(type="int", name="runtime", default_value=600),
    'num_vols': parameter(type="int", name="num_vols", default_value=2),
    'vol_size': parameter(type="string", name="vol_size", default_value='150g'),
    'num_files': parameter(type="int", name="num_files", default_value=10),
    'file_size': parameter(type="string", name="file_size", default_value='10g'),
    # IO tool to use for user workloads
    'io_tool': parameter(type="string", name="io_tool", default_value='dwim'),
    # start, end & step values (e.g: for iop rate/thread count) to use in
    # range function to ramp up baseline workload
    'knob_range':parameter(type="list", name="knob_range",
                           default_value=[2000, 10000, 1000]),
    'protocol': parameter(type="string", name="protocol", default_value='nfs',
                          source='perfant'),
    }

# This acts as a Factory. It instantiates an object of the appropriate subclass
# based on the test type.
# Note: We are using the 'test_type' parameter to determine both the module
# name & the class to instantiate. Hence, test_type must be exactly the same as
# the module name & the class name.
def test_setup():
    """Instantiate a particular class of QoS test based on the test type.
    Return an object of that class."""
    test_type = test_parameters['test_type']

    # TODO: Using import is not the ideal way to do this. Find a better way.
    module = __import__(name=test_type,
                        globals=globals(),
                        locals=locals(),
                        level=-1)
    class_obj = getattr(module, test_type)
    return class_obj()


class QoSTest(object):

    """Base class for QoS tests."""


    def __init__(self):
        """Initialize attributes of a QoS test instance."""
        self.test_params = test_parameters
        self.nodes = []
        self.aggregates = []
        self.vservers = []
        self.volumes = []
        self.workloads = []
        self.iterations = []
        self.system_processes = []

# This method loads the default values for method arguments that are not
# specified by the user, from the test_params instance variable. This is
# required because Python does not allow instance variables as defaults in the
# method arguments list.
    def _load_params(self, args):
        """Retrieve parameters specified via PerfANT or via NATE, if not
        specified in the method's argument list."""
        for key in args:
            if args[key] is None:
                if key in self.test_params:
                    args[key] = self.test_params[key]
        return args

    def setup(self):
        """Do setup for a QoS test. Discovers the aggregates & creates
        the specified number of volumes"""
        # Discover nodes, cluster & aggregates
        self.discover_nodes()
        self.discover_cluster(nodes=self.nodes)

        # We can do this after task#1487 is completed. Currently the aggr_list
        # in Params.py must match the one specified to PerfANT.
        ## Get the names of the aggregates created by perfant.
        #aggr_info = self.test_params['aggr_list'].split()
        #aggr_names = []
        #for aggr in aggr_info:
            #aggr_names.append(aggr.split(',')[0])

        #for aggr in aggr_names:
            #self.discover_aggregate(node=self.nodes[0], name=aggr)

        for aggr in self.test_params['aggr_list']:
            self.discover_aggregate(node=self.nodes[0],
                                    name=self.test_params['aggr_list'][0])

        # Create data objects
        self.create_vserver(cluster=self.cluster)

        for num in range(self.test_params['num_vols']):
            self.create_volume(vserver=self.vservers[0],
                               aggregate=self.aggregates[0],
                               vol_size=self.test_params['vol_size'])
        return


### Wrapper methods for each SLSL construct

    def discover_nodes(self, num_nodes=None, model=None, node_group='m1',
                       type='cmode'):
        """Discover & store the nodes in the testbed.
        Arguments:-
        num_nodes: Number of nodes in the cluster
        model: Controller model"""
        # Load test params
        params = {}
        params = self._load_params(
                    inspect.getargvalues(inspect.currentframe())[3])

        for i in range(params['num_nodes']):
            self.nodes.append(controller(model=params['model'],
                                         node_group=params['node_group'],
                                         node_group_index="%s" % (i),
                                         type=params['type']))

    def discover_cluster(self, nodes, name="cluster1"):
        """Discover & store the cluster.
        Arguments:-
        nodes: Array of node values returned by the SLSL controller constructs
        name: Cluster name"""
        # Load test params
        params = {}
        params = self._load_params(
                    inspect.getargvalues(inspect.currentframe())[3])

        self.cluster = cluster(name=params['name'], nodes=params['nodes'])
        return

    def discover_aggregate(self, node, name):
        """Discover & store the aggregates in the cluster.
        Arguments:-
        node: Value returned by a SLSL controller construct
        name: Aggregate name. Must match the aggr name specified in the
              PerfANT aggr_list parameter"""
        self.aggregates.append(aggregate(node=node, name=name))
        return

    def create_vserver(self, cluster, name=None):
        """Create vservers.
        Arguments:-
        cluster: Value returned by a SLSL cluster construct
        name: Name of vserver to be created"""
        # Must do this here instead of in the method definition to get different
        # names.
        if name is None:
            name = 'vs' + str(random.randint(1, 500))

        self.vservers.append(vserver(id=name, cluster=cluster))
        return

    def create_volume(self, vserver, aggregate, vol_size, name=None,
                      type='cmode'):
        """Create volumes.
        Arguments:-
        vserver: Value returned by a SLSL vserver construct
        aggregate: Value returned by a SLSL aggregate construct
        vol_size: Size of volume to be created
        name: Name of volume to be created"""

        # Must do this here instead of in the method definition to get different
        # names.
        if name is None:
            name = 'vol' + str(random.randint(1, 500))

        # Load test params
        params = {}
        params = self._load_params(
                    inspect.getargvalues(inspect.currentframe())[3])

        self.volumes.append(volume(type=params['type'], size=params['vol_size'],
                            name=params['name'], vserver=params['vserver'],
                            aggregate=params['aggregate']))
        return

    def create_datastores(self, volumes, num_files=None, file_size=None,
                          num_luns=None, lun_size=None, protocol=None):
        """Create data on luns or volumes.
        Arguments:-
        volumes: Array of volumes. Each element is a return value from a SLSL
                 volume construct
        num_files: Scalar specifying the number of files to create
        file_size: Scalar specifying the size of each file
        num_luns: Scalar specifying the number of luns to create
        lun_size: Scalar specifying the size of each lun
        protocol: String specifying the protocol"""
        # Load test params
        params = {}
        params = self._load_params(
                    inspect.getargvalues(inspect.currentframe())[3])

        datastores = []
        for vol in params['volumes']:
            if params['protocol'] == 'nfs' or params['protocol'] == 'cifs':
                datastores.append(datastore(type="NAS", volume=vol,
                                            num_files=params['num_files'],
                                            file_size=params['file_size']))
            elif params['protocol'] == 'iscsi' or params['protocol'] == 'fcp':
                datastores.append(datastore(type="SAN", volume=vol,
                                            num_luns=params['num_luns'],
                                            lun_size=params['lun_size']))
            else:
                raise Exception('Unknown protocol: ', params['protocol'])

        return datastores

    def create_paths(self, datastores, vserver, nodes, protocol=None,
                     data_nics=None, eth_iface=None, num_clients=None,
                     fcp_targets=None, initiators=None):
        """Create io paths & path mappers for the datastores specified.
        Arguments:-
        datastores: Array. Each element is a return value from a SLSL datastore
                    construct.
        vserver: Vserver containing the datastores. Must be value from a
                 SLSL vserver construct
        nodes: Array. Nodes to which paths will be created. Each element is a
               return value from a SLSL controller construct
        protocol: Scalar specifying the protocol to use.
        data_nics: Array specifying the data nics on the controller
        eth_iface: Array specifying the ethernet interfaces on the clients
        num_clients: Scalar specifying the number of clients
        fcp_targets: Array specifying the fcp targets on the controller
        initiators: Array specifying the initiators on the clients"""
        # Load test params
        params = {}
        params = self._load_params(
                    inspect.getargvalues(inspect.currentframe())[3])

        io_paths = []
        for ds in params['datastores']:
            if params['protocol'] == 'nfs' or params['protocol'] == 'cifs':
                for node in params['nodes']:
                    io_paths.append(
                        create_default_nas_paths(
                            params['data_nics'], params['eth_iface'],
                            ds, num_clients=params['num_clients'],
                            controller_id=node, vserver=params['vserver']))
            elif params['protocol'] == 'iscsi' or params['protocol'] == 'fcp':
                for node in params['nodes']:
                    io_paths.append(
                        create_default_san_paths(
                            params['fcp_targets'], params['initiators'],
                            ds, num_clients=params['num_clients'],
                            controller_id=node, vserver=params['vserver']))
            else:
                raise Exception('Unknown protocol: ', params['protocol'])

        paths = []
        for ipath in io_paths:
            paths.append(path_mapper(io_paths=ipath))

        return paths

    def create_workload(self, volume=None, vserver=None, node=None,
                        runtime=None,io_tool=None, num_processes=None,
                        **keywords):
        """Create I/O workloads. Create datastore & path mapper, if not
        specified.
        Arguments:-
        volume: value returned from a SLSL volume construct
        vserver: value returned from a SLSL vserver construct
        node: value returned from a SLSL controller construct
        runtime: scalar specifying the runtime for the workload
        io_tool: scalar specifying the io_tool to use for/type of the workload
        num_processes: scalar specifying the number of processes/files for the
                       workload
        **keywords: key-value pairs of arguments for the SLSL workload construct
        """
        # Load test params
        params = self._load_params(
                    inspect.getargvalues(inspect.currentframe())[3])

        # Specify workload parameters
        wkld_params = keywords
        if 'runtime' not in wkld_params:
            wkld_params['runtime'] = params['runtime']
        if 'type' not in wkld_params:
            wkld_params['type'] = params['io_tool']
        if 'num_processes' not in wkld_params:
            wkld_params['num_processes'] = params['num_processes']
        if wkld_params['num_processes'] is None:
            wkld_params['num_processes'] = self.test_params['num_files']

        # For dwim, sio, determine the op rate or threads based on the starting
        # range value, if not specified.
        if wkld_params['type'] == 'dwim':
            knob = 'target_iops'
        if wkld_params['type'] == 'sio':
            knob = 'threads'
        if wkld_params['type'] == 'dwim' or wkld_params['type'] == 'sio':
            if knob not in wkld_params:
                wkld_params[knob] = self.test_params['knob_range'][0]

        # Create data & io access paths, if path_mapper is not specified
        if 'path_mapper' not in wkld_params:
            if volume is None or vserver is None or node is None:
                raise Exception('You must specify either (a):path_mapper or ' +
                                '(b): volume, vserver & nodes')
            datastores = []
            paths = []
            datastores = self.create_datastores(volumes=[volume])
            paths = self.create_paths(datastores=[datastores[0]],
                                      vserver=params['vserver'],
                                      nodes=[params['node']])
            wkld_params['path_mapper'] = paths[0]

        # Store all created workloads & their params, in case we need them later
        # to rebuild a workload.
        wkld = workload(**wkld_params)
        self.workloads.append({'id':wkld, 'params':wkld_params})
        return wkld

    def create_workload_from_baseline(self, baseline_iteration,
                                      baseline_workload):
        """Use the specified baseline workload to build a new workload by
        replacing its knob with the baseline result.
        Arguments:-
        baseline_iteration: baseline iteration dict returned by the
                            create_iteration method
        baseline_workload: workload returned by the create_workload method &
                           used in the baseline iteration"""

        # Complain if specified iteration does not exist in list
        for inst in self.iterations:
            if baseline_iteration.args['id'] == inst.args['id']:
                break
        else:
            raise Exception('Could not find specified iteration in list of ' +
                            'existing iterations')

        # Determine parameters for new workload from baseline workload &
        # iteration
        workload_params = self.get_workload_params(workload=baseline_workload)
        for knob in baseline_iteration.args['knobs'].keys():
            workload_params[knob] = getattr(baseline_iteration.result, knob)

        return workload(**workload_params)

    def create_callout(self, callback_name, type='nate', args=None):
        """Create a SLSL callout.
        Arguments:-
        callback_name: Name of the callout
        type: Type of callout. Default is 'nate'
        args: Arguments to be passed to the callout"""

        if args is None:
            return callout(type=type, callback_name=callback_name)
        else:
            return callout(type=type, callback_name=callback_name, args=args)

    def create_iteration(self, type, workloads, knob=None, knob_range=None,
                         events_obj=None, feedback_callout=None,
                         flush_cache=True, chk_system_wkld=False):
        """Create an iteration. Returns the iteration object.
        Arguments:-
        type: Type of iteration. Supported values are 'baseline' or 'test'
        workloads: array specifying theworkloads to be run in this iteration.
                   elements of the array must be values returned by the SLSL
                   workload construct
        knob: the knob to use to modify the workload. Must be a valid argument
              of the workload construct. Only used in iteration of type
              'baseline'. Has defaults for specific workload types.
        knob_range: The start, end & step values to use for the knob to ramp up
                    the baseline workload
        event_obj: An instance of the Events class specifying the events to
                   occur in the iteration
        feedback_callout: A dict containing the arguments for a feedback callout
                          This is the callout that determines if the baseline
                          iteration has reached its goal.
        flush_cache: Boolean. If true, flush the cache.
        chk_system_wkld: Boolean. If true, add callouts to check entries for the
                         System process in the QoS workload & policy tables"""
        # Load test params
        params = {}
        params = self._load_params(
                    inspect.getargvalues(inspect.currentframe())[3])

        # If an events object is not specified, create one.
        if events_obj is None:
            events_obj = Events()
        # If an events object is specified, verify that it is an instance of
        # the right class.
        else:
            if not isinstance(events_obj, Events):
                raise Exception("Object passed in the 'events_obj' argument " +
                                "is not an instance of the 'Events' class")

        # Create default events for the test
        self.create_default_events(events_obj=events_obj, iteration_type=type)

        # Create events to check for system workload & policy entries
        if chk_system_wkld:
            for sys_proc in self.system_processes:
                events_obj.create(event_name='workload_exists',
                                  event_phase='start_measurement_after',
                                  workload_name=sys_proc['workload_name'])
                events_obj.create(event_name='policy_exists',
                                  event_phase='start_measurement_after',
                                  policy_name=sys_proc['policy_name'])

          # Create a callout to flush the cache at the start of the iteration
        if flush_cache:
            events_obj.create(event_name='flush_cache',
                              event_phase='start_iteration')

        iter_obj = {}
        if type == 'baseline':
            if len(workloads) > 1:
                raise Exception('Baseline iteration can only contain 1 workload')

            if knob is None:
                workload_params = self.get_workload_params(workload=workloads[0])
                if workload_params['type'] == 'dwim':
                    knob = 'target_iops'
                elif workload_params['type'] == 'sio':
                    knob = 'threads'
                else:
                    raise Exception('knobs must be specified for workload of ' +
                                    'type', workload['type'])

            if feedback_callout is None:
                raise Exception('feedback must be specified for baseline ' +
                                'iteration')
            # Create the feedback callout & add it to the events list
            else:
                feedback = self.create_callout(
                                    callback_name=feedback_callout['name'],
                                    args=feedback_callout['args'])
                events_obj.add(event=feedback,
                               event_phase=feedback_callout['phase'])

            iter_obj = iteration(type='adaptive',
                                 workload=workloads[0],
                                 feedback=feedback,
                                 knobs={knob:range(params['knob_range'][0],
                                                   params['knob_range'][1]+
                                                   params['knob_range'][2],
                                                   params['knob_range'][2])},
                                 events=events_obj.get_list())

        elif type == 'test':
            iter_obj = iteration(workloads=workloads,
                                 events=events_obj.get_list())

        else:
            raise Exception('Unknown iteration type:', type)

        self.iterations.append(iter_obj)
        return iter_obj

    def run(self, iterations, events_obj=None, keep_system_wkld=False):
        """Create a testrunner.
        Arguments:-
        iterations: array of iteration dicts returned by the SLSL iteration
                    construct or the create_iteration method
        events_obj: An instance of the Events class specifying the events to
                    occur in the testrunner
        keep_system_wkld: Boolean. If true, set flag to make system workloads
                          persist at the start of the test & explicitly delete
                          them at the end of the test"""
        # Load test params
        params = {}
        params = self._load_params(
                    inspect.getargvalues(inspect.currentframe())[3])

        # If an events object is not specified, create one.
        if events_obj is None:
            events_obj = Events(caller='testrunner')
        # If an events object is specified, verify that it is an instance of
        # the right class.
        else:
            if not isinstance(events_obj, Events):
                raise Exception("Object passed in the 'events_obj' argument " +
                                "is not an instance of the 'Events' class")

        # Create events to make system workloads persist during the test &
        # delete them after the test is complete.
        if keep_system_wkld:
            for sys_proc in self.system_processes:
                events_obj.create(event_name='make_system_wklds_persist',
                                  event_phase='start_test')
                events_obj.create(event_name='remove_system_wkld',
                                  event_phase='end_test',
                                  workload_name=sys_proc['workload_name'])

        test_runner(iterations=iterations, id=self.test_params['purpose'],
                    events=events_obj.get_list())


### Accessor methods
    def get_aggregates(self):
        """Return the list of aggregates (SLSL construct return values)."""
        return self.aggregates

    def get_nodes(self):
        """Return the list of nodes (SLSL construct return values)."""
        return self.nodes

    def get_vservers(self):
        """Return the list of vservers (SLSL construct return values)."""
        return self.vservers

    def get_volumes(self):
        """Return the list of volumes (SLSL construct return values)."""
        return self.volumes

    def get_workloads(self):
        """Return the list of workloads (SLSL construct return values)."""
        return self.workloads

    def get_workload_params(self, workload):
        """Return the parameters of the workload specified.
        Arguments:-
        workload: Workload returned by the create_workload method"""
        for wkld in self.workloads:
            if wkld['id'] == workload:
                return wkld['params']
        else:
            raise Exception('Could not find specified workload in list of ' +
                            'existing workloads')

    def add_system_process(self, workload_name, policy_name):
        """Store the workload name & the policy name of the System Process in
        the test.
        Arguments:-
        workload_name: Workload name of the System Process
        policy_name: Policy name for the System Process"""
        self.system_processes.append({'workload_name':workload_name,
                                      'policy_name': policy_name})
