# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  Library to maintain internal QoS test summary data structure
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::QoS::TestSummary

=head1 DESCRIPTION

Perf::AAMP::QoS::TestSummary provides routines to create and query a QoS
test summary stored in the stash. This test summary may be used by callouts
to share data, as well as by verification functions.

The routines in this library may be called procedurally, or (via
Perf::AAMP::Test::Stash's 'delegation_package' facility) as methods
of the stash object.

For example, as a method:

    use Perf::AAMP::Test::Stash;

    my $stash = Perf::AAMP::Test::Stash->new();
    $stash->delegation_package(pkg => 'Perf::AAMP::QoS::TestSummary');

    $stash->current_iteration(iteration => 'baseline');
    $stash->statistics_summary_update(
        iteration => $stash->current_iteration(),
        sample    => $sample,
    );

Or as a procedure:

    use Perf::AAMP::Test::Stash;
    use Perf::AAMP::QoS::TestSummary qw(current_iteration
                                        statistics_summary_update);

    my $stash = Perf::AAMP::Test::Stash->new();

    current_iteration(iteration => 'baseline', stash => $stash);
    statistics_summary_update(
        iteration => current_iteration(stash => $stash),
        sample    => $sample,
        stash     => $stash,
    );

=head1 SEE ALSO

Perf::AAMP::Test::Stash

=head1 FUNCTIONS

=cut

package Perf::AAMP::QoS::TestSummary;

use strict;
use warnings;
use Carp;
use Data::Dumper;
use Digest::MD5 qw(md5_hex);
use File::Slurp qw(read_file);
use JSON;
use List::MoreUtils qw(none);
use Params::Validate qw(:all);
use POSIX qw(strftime);
use Readonly;
use Tharn qw($Log);

Readonly::Scalar my $SUMMARY_KEY => md5_hex('*QoS*Test*Summary*');

use base q(Exporter);
our %EXPORT_TAGS = (
    all => [
        qw(
            checkpoint
            current_iteration
            config_summary_update
            config_summary_fetch
            io_summary_update
            io_summary_fetch
            iteration_sequence_fetch
            load_summary
            observations
            slo_summary_update
            slo_summary_fetch
            statistics_summary_update
            statistics_summary_fetch
            test_summary_log
            user_summary_update
            user_summary_fetch
        ),
    ],
);
our @EXPORT_OK = @{$EXPORT_TAGS{all}};


### Public functions:


=head2 current_iteration

Return (and optionally set) the current iteration.

=over

=item Arguments

B<iteration> = string

I<Optional> The current iteration.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

B<type> = string

I<Optional> The type of the current iteration.

=back

=cut

sub current_iteration {
    my %opts = validate(
        @_, {
            iteration => {
                type     => SCALAR,
                optional => 1,
            },
            stash     => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
            type     => {
                type     => SCALAR,
                optional => 1,
            },
        },
    );
    $Log->enter(Dumper(\%opts));               # Nicer output after validate()

    my $iteration = $opts{iteration};
    my $stash     = $opts{stash};
    my $summary   = _fetch_summary(stash => $stash);

    # If called without an iteration specified, then simply
    # return the previously-defined current iteration.
    #
    # Note that it's a fatal error if the current iteration
    # has not already been set.
    if (!defined($iteration)) {
        confess 'Current iteration not yet specified'
            if !defined($summary->{current});
        $Log->exit();
        return $summary->{current};
    }

    # If we got here, an iteration was specified.
    #
    # If this is the first time we've seen the specified
    # iteration, create a section for it and append it to
    # the sequence.
    if (none { $_ eq $iteration } @{$summary->{sequence}}) {
        $summary->{iteration}->{$iteration} = _new_section(type => $opts{type});
        push @{$summary->{sequence}}, $iteration;
    }

    # Update the current iteration. Note that we don't require
    # the current iteration to be last in the sequence.
    $summary->{current} = $iteration;

    # Return the current iteration.
    $Log->exit();
    return $summary->{current};
}


=head2 iteration_sequence_fetch

Return the sequence of observed iterations.

=over

=item Arguments

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

B<type> = string

I<Optional> If specified, return only iterations of this type.

=back

=cut

sub iteration_sequence_fetch {
    my %opts = validate(
        @_, {
            stash => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
            type  => {
                type     => SCALAR,
                optional => 1,
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $summary  = _fetch_summary(stash => $opts{stash});
    my @sequence = @{$summary->{sequence}};

    if (defined $opts{type}) {
        @sequence =
            grep { $summary->{iteration}->{$_}->{TYPE} eq $opts{type} }
                 @sequence;
    }

    $Log->exit();
    return \@sequence;
}


=head2 config_summary_update

Update the global test summary configuration information. Either one or
both of the 'QosConfig' and 'SysConfig' parameters must be passed on
each call. Note that configuration information is only stored in the
global section, and that multiple calls overwrite (rather than append to)
existing configuration information.

=over

=item Arguments

B<QosConfig> = hashref

I<Optional> Configuration information for the QoSTest object

B<SysConfig> = hashref

I<Optional> System configuration information

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub config_summary_update {
    my %opts = validate(
        @_, {
            QosConfig => {
                type     => HASHREF,
                optional => 1,
            },
            SysConfig => {
                type     => HASHREF,
                optional => 1,
            },
            stash     => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    confess q{Must specify at least one of: 'QosConfig', 'SysConfig'}
        if none { defined($opts{$_}) } qw(QosConfig SysConfig);

    my $hashref = _fetch_reference(
        stash    => $opts{stash},
        category => 'CONFIG',
    );

    foreach my $section qw(QosConfig SysConfig) {
        if (defined $opts{$section}) {
            $hashref->{$section} = $opts{$section};
        }
    }

    $Log->exit();
    return;
}


=head2 config_summary_fetch

Fetch the global test summary configuration information. Each call
returns both the 'QosConfig' and 'SysConfig' information.

=over

=item Arguments

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub config_summary_fetch {
    my %opts = validate(
        @_, {
            stash     => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $hashref = _fetch_reference(
        stash    => $opts{stash},
        category => 'CONFIG',
    );

    $Log->exit();
    return $hashref;
}


=head2 io_summary_update

Update the test summary I/O information.

=over

=item Arguments

B<id> = string

I<Required> The perf::qos::control::QoSTest::Datastore ID.

B<configuration> = string

I<Optional> The parameters used to create the
perf::qos::control::QoSTest::Datastore object.

B<execution> = string

I<Optional> The execution summary returned by the
perf::qos::control::QoSTest::Datastore object.

B<iteration> = string

I<Optional> The current iteration.

B<lif_owner> = string

I<Optional> The node currently hosting the IO data lif.

B<op_rate> = string

I<Optional> The initial or modified op-rate being used
by the perf::qos::control::QoSTest::Datastore object.

One of 'configuration', 'execution', 'lif_owner', or 'op_rate' must
be passed on each call. The 'iteration' must be specified whenever
'execution', 'lif_owner', or 'op_rate' is passed.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub io_summary_update {
    my @args = @_;

    my $none_of = sub {
        my @exclude = @_;
        return sub {
            my $args = $_[1];
            return none { exists($args->{$_}) } @exclude;
        };
    };

    my %opts = validate(
        @args, {
            configuration => {
                type      => HASHREF,
                optional  => 1,
                callbacks => {
                    mutual_exclusion => $none_of->(qw(execution op_rate)),
                },
            },
            execution     => {
                type      => HASHREF,
                optional  => 1,
                depends   => [ qw(iteration) ],
                callbacks => {
                    mutual_exclusion => $none_of->(
                        qw(configuration lif_owner op_rate),
                    ),
                },
            },
            id            => {
                type      => SCALAR,
                optional  => 0,
            },
            iteration     => {
                type      => SCALAR,
                optional  => 1,
            },
            lif_owner     => {
                type      => SCALAR,
                optional  => 1,
                depends   => [ qw(iteration) ],
                callbacks => {
                    mutual_exclusion => $none_of->(
                        qw(configuration execution op_rate),
                    ),
                },
            },
            op_rate       => {
                type      => SCALAR,
                optional  => 1,
                depends   => [ qw(iteration) ],
                callbacks => {
                    mutual_exclusion => $none_of->(
                        qw(configuration execution lif_owner),
                    ),
                },
            },
            stash         => {
                type      => OBJECT,
                optional  => 0,
                isa       => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    if (defined $opts{configuration}) {
        _register_datastore(%opts);
    }
    elsif (defined $opts{execution}) {
        _append_run_datastore(%opts);
    }
    elsif (defined $opts{lif_owner}) {
        _append_lifowner_datastore(%opts);
    }
    elsif (defined $opts{op_rate}) {
        _append_oprate_datastore(%opts);
    }
    else {
        confess q{Must specify one of '}
              . join(q{', '}, qw(configuration execution lif_owner op_rate))
              . q{'};
    }
    $Log->exit();
    return;
}


=head2 io_summary_fetch

Fetch test summary I/O information.

=over

=item Arguments

B<id> = string

I<Required> The perf::qos::control::QoSTest::Datastore ID.

B<iteration> = string

I<Optional> The current iteration.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub io_summary_fetch {
    my %opts = validate(
        @_, {
            id        => {
                type     => SCALAR,
                optional => 0,
            },
            iteration => {
                type     => SCALAR,
                optional => 1,
            },
            stash     => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $id      = delete $opts{id};
    my $hashref = _fetch_reference(%opts, category => 'IO');

    $Log->exit();
    return $hashref->{$id};
}


=head2 slo_summary_update

Update the test summary Policy-Group/SLO information.

=over

=item Arguments

B<policy_group> = string

I<Required> The policy-group name.

B<slo> = hashref

I<Optional> SLO settings for the policy-group.

B<workloads> = arrayref

I<Optional> QoS workloads in the policy-group

B<iteration> = string

I<Optional> The current iteration.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub slo_summary_update {
    my %opts = validate(
        @_, {
            iteration    => {
                type     => SCALAR,
                optional => 1,
            },
            policy_group => {
                type     => SCALAR,
                optional => 0,
            },
            slo          => {
                type     => HASHREF,
                default  => {},
            },
            stash        => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
            workloads    => {
                type     => ARRAYREF,
                default  => [],
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $policy_group = delete $opts{policy_group};
    my $slo          = delete $opts{slo};
    my $workloads    = delete $opts{workloads};
    my $hashref      = _fetch_reference(%opts, category => 'SLO');

    push @{$hashref->{$policy_group}}, {
        slo       => $slo,
        workloads => $workloads,
    };
    $Log->exit();
    return;
}


=head2 slo_summary_fetch

Fetch test summary I/O information.

=over

=item Arguments

B<all_samples> = string

I<Optional> Flag to return all samples for the iteration. Defaults to false.

B<iteration> = string

I<Optional> The current iteration.

B<policy_group> = string

I<Optional> The policy-group name.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub slo_summary_fetch {
    my %opts = validate(
        @_, {
            all_samples  => {
                type     => SCALAR,
                default  => 0,
            },
            iteration    => {
                type     => SCALAR,
                optional => 1,
            },
            policy_group => {
                type     => SCALAR | UNDEF,
                optional => 1,
            },
            stash        => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $all_samples  = delete $opts{all_samples};
    my $policy_group = delete $opts{policy_group};
    my $hashref      = _fetch_reference(%opts, category => 'SLO');

    if (defined($policy_group)) {
        $Log->exit();
        return $all_samples
             ? $hashref->{$policy_group}           # All obs for plcy-grp
             : [$hashref->{$policy_group}->[-1]];  # Last obs for plcy-grp
    }
    else {
        if ($all_samples) {
            $Log->exit();
            return $hashref;                       # All obs for all plcy-grps
        }

        my $summary;
        foreach my $pg (keys(%{$hashref})) {       # Last obs for each plcy-grp
            $summary->{$pg} = [$hashref->{$pg}->[-1]];
        }
        $Log->exit();
        return $summary;
    }

    confess 'This statement should never be executed';
    # NOTREACHED
}


=head2 statistics_summary_update

Update the test summary Perf::Stats sample information.

=over

=item Arguments

B<iteration> = string

I<Optional> The current iteration.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub statistics_summary_update {
    my %opts = validate(
        @_, {
            iteration => {
                type     => SCALAR,
                optional => 1,
            },
            sample    => {
                type     => HASHREF,
                optional => 0,
            },
            stash     => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $sample   = delete $opts{sample};
    my $arrayref = _fetch_reference(%opts, category => 'STATS');
    push @{$arrayref}, $sample;
    $Log->exit();
    return;
}


=head2 statistics_summary_fetch

Fetch test summary Perf::Stats sample(s).

=over

=item Arguments

B<all_samples> = string

I<Optional> Flag to return all samples for the iteration. Defaults to false.

B<iteration> = string

I<Optional> The current iteration.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub statistics_summary_fetch {
    my %opts = validate(
        @_, {
            all_samples => {
                type     => SCALAR,
                default  => 0,
            },
            iteration   => {
                type     => SCALAR,
                optional => 1,
            },
            stash       => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $all_samples = delete $opts{all_samples};
    my $hashref     = _fetch_reference(%opts, category => 'STATS');

    # Unless the caller requested all samples,
    # only return the last one.
    $Log->exit();
    return $all_samples
         ? $hashref
         : [$hashref->[-1]];
}


=head2 user_summary_update

Update the test summary user-defined data.

=over

=item Arguments

B<name> = string

I<Required> The name to store the data under.

B<value> = string

I<Required> The value to store for the name.

B<iteration> = string

I<Optional> The current iteration.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub user_summary_update {
    my %opts = validate(
        @_, {
            iteration => {
                type     => SCALAR,
                optional => 1,
            },
            name      => {
                type     => SCALAR,
                optional => 0,
            },
            stash     => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
            value     => {
                optional => 0,
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $name    = delete $opts{name};
    my $value   = delete $opts{value};
    my $hashref = _fetch_reference(%opts, category => 'USER');

    $hashref->{$name} = $value;
    $Log->exit();
    return;
}


=head2 user_summary_fetch

Fetch test summary user-defined value.

=over

=item Arguments

B<name> = string

I<Required> The name of the data to fetch.

B<iteration> = string

I<Optional> The current iteration.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub user_summary_fetch {
    my %opts = validate(
        @_, {
            iteration => {
                type     => SCALAR,
                optional => 1,
            },
            name      => {
                type     => SCALAR,
                optional => 0,
            },
            stash     => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $name      = delete $opts{name};
    my $hashref   = _fetch_reference(%opts, category => 'USER');
    my $iteration = $opts{iteration};

    # Fatal error if the specified data doesn't exist.
    confess "No user data '$name'"
          . (defined($opts{iteration}) ? " for iteration '$iteration'" : q{})
        if !exists($hashref->{$name});

    $Log->exit();
    return $hashref->{$name};
}


=head2 checkpoint

Create a verification checkpoint by recording the indices of the latest
observations for the specified iteration.

=over

=item Arguments

B<datastores> = string

I<Optional> The IDs of the Datastore instances that are "interesting"
for this observation.

B<iteration> = string

I<Required> The current iteration.

B<desc> = string

I<Optional> Descriptive label to identify this checkpoint.

B<policy_groups> = string

I<Optional> The names of the policy-groups that are "interesting"
for this observation.

B<should_fail> = hashref | boolean

I<Optional> Specifies that certain verification checks are expected
to fail for certain policy groups for this checkpoint. should_fail
may be either a hashref or a boolean. If it is a hashref, then the
keys should be policy groups and each value should be an arrayref
of the names of the verification functions that are expected to
fail for its policy group. Otherwise, it's treated as a flag to
indicate whether all combinations of policy groups and verification
functions are expected to pass or fail.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub checkpoint {
    my %opts = validate(
        @_, {
            datastores    => {
                type     => ARRAYREF,
                optional => 1,
            },
            desc          => {
                type     => SCALAR | UNDEF,
                optional => 1,
            },
            iteration     => {
                type     => SCALAR,
                optional => 0,
            },
            policy_groups => {
                type     => ARRAYREF | UNDEF,
                optional => 1,
            },
            should_fail   => {
                type     => HASHREF | SCALAR | UNDEF,
                optional => 1,
            },
            stash         => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $datastores    = $opts{datastores};
    my $iteration     = $opts{iteration};
    my $desc          = defined($opts{desc}) ? $opts{desc} : q{};
    my $policy_groups = $opts{policy_groups};
    my $should_fail   = defined($opts{should_fail}) ? $opts{should_fail} : 0;
    my $summary       = _fetch_summary(stash => $opts{stash});

    confess "No such iteration: '$iteration'"
        if !exists($summary->{iteration}->{$iteration});

    # Create checkpoint hashref.
    my $section    = $summary->{iteration}->{$iteration};
    my $checkpoint = {
        DESC        => $desc,                  # Key chosen to sort first
        IO          => {},                     # To be filled in below
        SLO         => {},                     # To be filled in below
        SHOULD_FAIL => $should_fail,
        STATS       => $#{$section->{STATS}},  # Index of last STATS observation
        TIMESTAMP   => strftime('%Y-%m-%d %H:%M:%S %Z', localtime),
    };

    # If the caller didn't specify the "interesting" datastores, save info
    # for all of them. (Otherwise, only save info for the "interesting" ones.)
    my @io_ids = defined($datastores)
               ? @{$datastores}
               : keys(%{$section->{IO}});
    foreach my $id (@io_ids) {
        $checkpoint->{IO}->{$id}->{lif_owner} =
          $#{$section->{IO}->{$id}->{lif_owner}};
        $checkpoint->{IO}->{$id}->{op_rate} =
          $#{$section->{IO}->{$id}->{op_rate}};
        $checkpoint->{IO}->{$id}->{summary} =
          $#{$section->{IO}->{$id}->{summary}};
    }

    # If the caller didn't specify the "interesting" policy groups, save info
    # for all of them. (Otherwise, only save info for the "interesting" ones.)
    my @pgrps = defined($policy_groups)
              ? @{$policy_groups}
              : keys(%{$section->{SLO}});
    foreach my $pg (@pgrps) {
        $checkpoint->{SLO}->{$pg} = $#{$section->{SLO}->{$pg}};
    }

    # Save checkpoint info with iteration data.
    push @{$section->{CHKPT_INDICES}}, $checkpoint;

    $Log->exit();
    return;
}


=head2 observations

Return the checkpointed observations for the specified iteration.

=over

=item Arguments

B<iteration> = string

I<Required> The current iteration.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub observations {
    my %opts = validate(
        @_, {
            iteration => {
                type     => SCALAR,
                optional => 0,
            },
            stash     => {
                type     => OBJECT,
                optional => 0,
                isa      => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $iteration = $opts{iteration};
    my $summary   = _fetch_summary(stash => $opts{stash});

    confess "No such iteration: '$iteration'"
        if !exists($summary->{iteration}->{$iteration});

    my $section = $summary->{iteration}->{$iteration};

    # Create an array of the checkpointed observations for the
    # iteration. Each entry in the array is a triplet composed
    # of IO, SLO, and STATS summaries.
    my @observations;
    foreach my $obs (@{$section->{CHKPT_INDICES}}) {
        my $sample = {
            DESC        => $obs->{DESC},
            IO          => {},
            SLO         => {},
            SHOULD_FAIL => $obs->{SHOULD_FAIL},
            STATS       => {},
        };

        # Collect the IO summaries from the interesting Datastore instances.
        foreach my $id (keys %{$obs->{IO}}) {
            $sample->{IO}->{$id}->{lif_owner} =
              $section->{IO}->{$id}->{lif_owner}->[$obs->{IO}->{$id}->{lif_owner}];
            $sample->{IO}->{$id}->{op_rate} =
              $section->{IO}->{$id}->{op_rate}->[$obs->{IO}->{$id}->{op_rate}];
            $sample->{IO}->{$id}->{summary} =
              $section->{IO}->{$id}->{summary}->[$obs->{IO}->{$id}->{summary}];
        }

        # Collect the SLO summaries from the interesting policy-groups.
        foreach my $pg (keys %{$obs->{SLO}}) {
            $sample->{SLO}->{$pg} =
              $section->{SLO}->{$pg}->[$obs->{SLO}->{$pg}];
        }

        # Include the Perf::Stats summary.
        if ($obs->{STATS} >= 0) {
            $sample->{STATS} =
              $section->{STATS}->[$obs->{STATS}];
        }

        # Add the sample to the ordered list of observations for
        # the iteration.
        push @observations, $sample;
    }

    $Log->exit();
    return \@observations;
}


=head2 test_summary_log

Dump test summary data structure (in JSON format) to logfile.

=over

=item Arguments

B<logdir> = string

I<Optional> The log directory. Defaults to current ntest logdir.

B<logfile> = string

I<Optional> The log filename. Defaults to 'summary.log'.

B<truncate> = boolean

I<Optional> Flag to truncate (rather than append to) the log.
Defaults to false.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub test_summary_log {
    my %opts = validate(
        @_, {
            logdir   => { type => SCALAR, default => $::LOGDIR     },
            logfile  => { type => SCALAR, default => 'summary.log' },
            stash    => { type => OBJECT,
                          isa  => q(Perf::AAMP::Test::Stash)       },
            truncate => { type => SCALAR, default => 0             },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $json      = JSON->new();
    my $logfile   = $opts{logdir} . q{/} . $opts{logfile};
    my $mode      = $opts{truncate} ? q{>} : q{>>};
    my $stash     = $opts{stash};
    my $summary   = $stash->exists(name => $SUMMARY_KEY)
                  ? $stash->fetch(name => $SUMMARY_KEY)
                  : {};

    # JSON doesn't allow comments, so we'll store metadata about the
    # dump in a temporary 'About' section.
    $summary->{About} = {
        Created  => strftime('%Y-%m-%d %H:%M:%S %Z', localtime),
        Filename => $logfile,
        Format   => 'http://www.json.org/',
    };

    # By default, if called more than once, this appends to the
    # test summary log. However, this can be overridden via the
    # 'truncate' parameter.
    open(my $log, $mode, $logfile)
        or confess "Cannot open '$logfile': $!";

    # Output test summary in the JSON format. JSON was chosen
    # because it is both human readable, and easily parsable
    # by numerous programming languages.
    printf {$log} "%s\n",
           $json->canonical->pretty->allow_blessed->encode($summary);

    # Add pointer in the NATE log to the test summary log.
    $Log->comment('Test summary written to log: ' . $logfile);
    close($log)
        or $Log->warn("Cannot close '$logfile': $!");

    # Delete the temporary 'About' section.
    delete $summary->{About};

    $Log->exit();
    return;
}


=head2 load_summary

Load the test summary from a log file previously created by
test_summary_log(). This is intended primarily for creating
Perf::AAMP::QoS::TestSummary unit tests.

=over

=item Arguments

B<force_overwrite> = boolean

I<Optional> Overwrite summary (rather than confess) if one already exists.

B<logfile> = string

I<Required> Pathname of logfile to parse and load.

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash. Implicitly passed when called as a method.

=back

=cut

sub load_summary {
    my %opts = validate(
        @_, {
            force_overwrite => { type => SCALAR, default => 0       },
            logfile         => { type => SCALAR                     },
            stash           => { type => OBJECT,
                                 isa  => q(Perf::AAMP::Test::Stash) },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $overwrite = $opts{force_overwrite},
    my $logfile   = $opts{logfile},
    my $stash     = $opts{stash};

    # Don't overwrite an existing summary, unless the user
    # gives explicit permission.
    confess 'Test summary already exists in stash'
        if $stash->exists(name => $SUMMARY_KEY) && !$overwrite;

    # Attempt to parse the logfile; croak if unsuccessful.
    my $json    = JSON->new();
    my $summary = $json->decode(scalar(read_file $logfile));

    # Avoid storing the 'About' section in the stash.
    delete $summary->{About};

    # Update the summary in the stash, creating the entry
    # if necessary.
    $stash->update(
        name                  => $SUMMARY_KEY,
        value                 => $summary,
        create_if_nonexistent => 1,
    );

    # Record in the NATE log that we loaded the summary.
    $Log->comment('Test summary loaded from: ' . $logfile);
    $Log->exit();
    return;
}


### Private functions:


sub _append_lifowner_datastore {
    my %opts = validate(
        @_, {
            lif_owner => { type => SCALAR },
            id        => { type => SCALAR  },
            iteration => { type => SCALAR  },
            stash     => { type => OBJECT, isa => q(Perf::AAMP::Test::Stash) },
        },
    );

    my $lif_owner = $opts{lif_owner};
    my $id        = $opts{id};
    my $iteration = $opts{iteration};
    my $stash     = $opts{stash};

    # Fetch the iteration's IO hashref...
    my $hashref = _fetch_reference(
        category  => 'IO',
        iteration => $iteration,
        stash     => $stash,
    );

    # ...and append the lif_owner to Datastore's lif_owner arrayref.
    push @{$hashref->{$id}->{lif_owner}}, $lif_owner;
    return;
}


sub _append_oprate_datastore {
    my %opts = validate(
        @_, {
            op_rate   => { type => SCALAR },
            id        => { type => SCALAR  },
            iteration => { type => SCALAR  },
            stash     => { type => OBJECT, isa => q(Perf::AAMP::Test::Stash) },
        },
    );

    my $op_rate   = $opts{op_rate};
    my $id        = $opts{id};
    my $iteration = $opts{iteration};
    my $stash     = $opts{stash};

    # Fetch the iteration's IO hashref...
    my $hashref = _fetch_reference(
        category  => 'IO',
        iteration => $iteration,
        stash     => $stash,
    );

    # ...and append the op_rate to Datastore's op_rate arrayref.
    push @{$hashref->{$id}->{op_rate}}, $op_rate;
    return;
}


sub _append_run_datastore {
    my %opts = validate(
        @_, {
            execution => { type => HASHREF },
            id        => { type => SCALAR  },
            iteration => { type => SCALAR  },
            stash     => { type => OBJECT, isa => q(Perf::AAMP::Test::Stash) },
        },
    );

    my $execution = $opts{execution};
    my $id        = $opts{id};
    my $iteration = $opts{iteration};
    my $stash     = $opts{stash};

    # Fetch the iteration's IO hashref...
    my $hashref = _fetch_reference(
        category  => 'IO',
        iteration => $iteration,
        stash     => $stash,
    );

    # ...and append the execution summary to Datastore's
    # summary arrayref.
    push @{$hashref->{$id}->{summary}}, $execution;
    return;
}


sub _register_datastore {
    my %opts = validate(
        @_, {
            configuration => { type => HASHREF     },
            id            => { type => SCALAR      },
            stash         => {
                type => OBJECT,
                isa  => q(Perf::AAMP::Test::Stash),
            },
        },
    );

    my $configuration = $opts{configuration};
    my $id            = $opts{id};
    my $stash         = $opts{stash};

    # Fetch the global IO hashref...
    my $hashref = _fetch_reference(
        category => 'IO',
        stash    => $stash,
    );

    # ... and store the Datastore's configuration in it.
    $hashref->{$id} = $configuration;
    return;
}


sub _fetch_reference {
    my %opts = validate(
        @_, {
            category    => {
                type     => SCALAR,
                regex    => qr/^(?:CONFIG|IO|SLO|STATS|USER)$/,
                optional => 0,
            },
            iteration    => {
                type     => SCALAR,
                optional => 1,
            },
            stash       => {
                type     => OBJECT,
                isa      => q(Perf::AAMP::Test::Stash),
                optional => 0,
            },
        },
    );

    my $category  = $opts{category};
    my $iteration = $opts{iteration};
    my $stash     = $opts{stash};
    my $summary   = _fetch_summary(stash => $stash);

    # If an iteration was specified, fetch the section for the
    # itersation. Otherwise, fetch the global section.
    if (defined $iteration) {

        # It's a fatal error if the iteration section doesn't
        # already exist.
        confess "No summary data available for iteration: '$iteration'"
            if !exists($summary->{iteration}->{$iteration})
               || none { $_ eq $iteration } @{$summary->{sequence}};

        # Return reference to the category's hashref in
        # the iteration's section.
        return $summary->{iteration}->{$iteration}->{$category};
    }

    # If we're here, an iteration was not specified.
    # Return reference to the category's hashref in
    # the global section.
    return $summary->{global}->{$category};
}


sub _fetch_summary {
    my %opts = validate(
        @_, {
            stash       => {
                type     => OBJECT,
                isa      => q(Perf::AAMP::Test::Stash),
                optional => 0,
            },
        },
    );

    my $stash = $opts{stash};

    # Lazily construct test summary hash the first time we
    # try to fetch it.
    if (!$stash->exists(name => $SUMMARY_KEY)) {
        my $summary = {
            current   => undef,                # Current iteration
            global    => _new_section(),       # Global test summary info
            iteration => {},                   # Iteration-specific info
            sequence  => [],                   # Sequence of iterations
        };
        $summary->{global}->{CONFIG} = {
            QosConfig => {},
            SysConfig => {},
        };
        $stash->add(
            name  => $SUMMARY_KEY,
            value => $summary,
        );
    }

    # Fetch and return summary hash reference from the stash.
    return $stash->fetch(name => $SUMMARY_KEY);
}


sub _new_section {
    my %opts = validate(
        @_, {
            type     => {
                type     => SCALAR | UNDEF,
                optional => 1,
            },
        },
    );
    return {
        IO    => {},                           # Datastore configurations
        SLO   => {},                           # SLO policy info
        STATS => [],                           # Generally N/A for global
        USER  => {},                           # User-defined name/value pairs
        (defined $opts{type} ? (TYPE => $opts{type}) : ()),    # Iteration type
    };
}

1;

