# -*- perl -*-
# $Id: $
#
## @author   dl-perf-qa@netapp.com
## @summary  Common callouts for QoS NATE/AAMP tests
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::QoS::Verify

=head1 DESCRIPTION

Verification functions for Cluster QoS AAMP and NATE tests.

=cut

package Perf::AAMP::QoS::Verify;

use strict;
use warnings;
use Carp;
use Data::Dumper;
use List::MoreUtils qw(any uniq none);
use List::Util qw(reduce first);
use Params::Validate qw(:all);
use Perf::AAMP::QoS::TestSummary qw(
    config_summary_fetch
    io_summary_fetch
    slo_summary_fetch
    iteration_sequence_fetch
    observations
    statistics_summary_fetch
);
use POSIX qw(ceil floor);
use Tharn qw($Log);
use base q(Exporter);

our %EXPORT_TAGS = (
    all => [
        qw(
            compare_with_baseline
            verify_slo
        ),
    ],
);

our @EXPORT_OK = @{$EXPORT_TAGS{all}};


=head1 FUNCTIONS

=head2 compare_with_baseline

Verify that the performance did not degrade for the checkpointed
workloads compared with the baseline iteration.

=over

=item Arguments

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash.

B<error_margin> = number

I<Optional> The allowable error margin.

=back

=cut

sub compare_with_baseline {
    my %opts = validate(
        @_, {
            stash     => {
                type     => OBJECT,
                isa      => q(Perf::AAMP::Test::Stash),
            },
            error_margin => {
                type     => SCALAR,
                default  => 0.02,
            },
        },
    );
    local $Data::Dumper::Indent    = 1;
    local $Data::Dumper::Quotekeys = 0;
    $Log->enter(Dumper(\%opts));               # Nicer output after validate()

    my $stash           = $opts{stash};
    my $error_margin    = $opts{error_margin};
    my $failures        = 0;
    my $baselines_to_compare = iteration_sequence_fetch(stash => $stash,
                                                        type  => 'baseline');

    foreach my $baseline_iter (@{$baselines_to_compare}) {
      my $baseline_sample = statistics_summary_fetch(
           iteration => $baseline_iter,
           stash     => $stash,
      )->[-1];

      my $baseline_slo_summary = slo_summary_fetch(
           iteration => $baseline_iter,
           stash     => $stash,
      );

      # Verify each checkpoint in each test iteration.
      foreach my $iter (@{iteration_sequence_fetch(stash => $stash,
                                                   type  => 'test')}) {
          foreach my $obs (@{observations(iteration => $iter, stash => $stash)}) {

              # Verify that each checkpointed workload meet or exceeded
              # baseline performance.
              foreach my $pg (keys %{$obs->{SLO}}) {
                  foreach my $wkld (@{$obs->{SLO}->{$pg}->{workloads}}) {
                      my $failed_checks = _achieved_baseline_performance(
                          observation   => $obs,
                          baseline_perf => $baseline_sample,
                          baseline_slo  => $baseline_slo_summary,
                          policy_group  => $pg,
                          workload      => $wkld,
                          %opts,
                      );

                      $failures += _unexpected_result(
                          observation   => $obs,
                          policy_group  => $pg,
                          verification  => 'compare_with_baseline',
                          failed_checks => $failed_checks,
                      );
                  }
              }
          }
      }

    }

    $Log->exit();
    return $failures == 0;
}


=head2 verify_slo

Verify that the SLOs were met for the checkpointed workloads.

=over

=item Arguments

B<stash> = Perf::AAMP::Test::Stash object

I<Required> The stash.

B<error_margin> = number

I<Optional> The allowable error margin.

=back

=cut

sub verify_slo {
    my %opts = validate(
        @_, {
            error_margin => {
                type    => SCALAR,
                default => 0.05,
            },
            stash        => {
                type    => OBJECT,
                isa     => q(Perf::AAMP::Test::Stash),
            },
        },
    );
    local $Data::Dumper::Indent    = 1;
    local $Data::Dumper::Quotekeys = 0;
    $Log->enter(Dumper(\%opts));

    my $failures = 0;
    my $stash    = $opts{stash};

    # Verify each checkpoint in each iteration.
    foreach my $iter (@{iteration_sequence_fetch(stash => $stash)}) {
        foreach my $obs (@{observations(iteration => $iter, stash => $stash)}) {

            # Verify that each checkpointed policy-group met its SLO.
            foreach my $pg (keys %{$obs->{SLO}}) {
                my %slo       = %{$obs->{SLO}->{$pg}->{slo}};
                my $workloads = $obs->{SLO}->{$pg}->{workloads};

                # Be careful here. Since we're passing the observation
                # hashref to subroutines, they could potentially rewind
                # the slo hash iterator. Thus we're iterating over a
                # *copy* of the slo hashref. (See %slo assignment, above.)
                while (my ($policy, $value) = each(%slo)) {
                    if ($policy eq 'max-throughput') {
                        my $failed_checks = _verify_max_throughput(
                            observation  => $obs,
                            policy_group => $pg,
                            slo_value    => $value,
                            workloads    => $workloads,
                            %opts,
                        );

                        $failures += _unexpected_result(
                            observation   => $obs,
                            policy_group  => $pg,
                            verification  => 'verify_slo',
                            failed_checks => $failed_checks,
                        );
                    }
                    else {
                        confess "Don't know how to verify policy '$policy'";
                    }
                }
            }
        }
    }

    $Log->exit();
    return $failures == 0;
}


### Private functions:


# If the workload's max throughput was not limited, check
# that it met or exceeded its baseline performance.
sub _achieved_baseline_performance {
    my %opts = validate(
        @_, {
            baseline_perf => { type => HASHREF  },
            baseline_slo =>  { type => HASHREF  },
            error_margin  => { type => SCALAR   },
            observation   => { type => HASHREF  },
            policy_group  => { type => SCALAR   },
            stash         => { type => OBJECT   },
            workload      => { type => SCALAR   },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $baseline_perf        = $opts{baseline_perf};
    my $baseline_slo_summary = $opts{baseline_slo};
    my $error_margin         = $opts{error_margin};
    my $observation          = $opts{observation};
    my $policy_group         = $opts{policy_group};
    my $stash                = $opts{stash};
    my $workload             = $opts{workload};
    my $failures             = 0;

    # Only check workloads without any SLO set (max_throughput or target_latency)
    #
    # This should check all workloads when QoS enforcement is
    # disabled (e.g., through the 'qos settings' command), but
    # at present enforcement settings are not tracked in the
    # stash. This will be addressed by burt661648.

    # The call to config_summary_fetch() will fail with test
    # summary logs created before burts 636926 & 656279 were
    # checked in. The commented-out code below would work with
    # earlier logs.
    #
    # my $nodes = user_summary_fetch(
    #     stash => $stash,
    #     name  => 'Configuration',
    # )->{QosConfig}->{nodes};
    my $nodes = config_summary_fetch(stash => $stash)->{QosConfig}->{nodes};

    # This check will determine if the policy group in the observation
    # should be compared with its baseline,
    if (_pg_valid_for_baseline_comparison(slo_summary => $baseline_slo_summary,
                                          policy_group => $policy_group,
                                          stash        => $stash,
                                          observation  => $observation,
                                         )) {
     my $nodes = config_summary_fetch(stash => $stash)->{QosConfig}->{nodes};
        # Verify that latency did not increase on any node.
        foreach my $node (@{$nodes}) {
            my $baseline_latency =
                $baseline_perf->{"$node:workload:$workload:latency"};
            my $observed_latency =
                $observation->{STATS}->{"$node:workload:$workload:latency"};
            my $max_latency =
                ceil($baseline_latency + ($baseline_latency * $error_margin));

            if ($observed_latency > $max_latency) {
                $Log->comment(
                    "FAIL: Workload '$workload' in policy-group '$policy_group'"
                  . " exceeded baseline latency on node '$node'\n"
                  . "* Max acceptable latency: $max_latency\n"
                  . "* Actual observed latency: $observed_latency"
                );
                $failures++;
            }
            else {
               $Log->comment(
                    "PASS: Workload '$workload' in policy-group '$policy_group'"
                  . " under baseline latency on node '$node'\n"
                  . "* Max acceptable latency: $max_latency\n"
                  . "* Actual observed latency: $observed_latency"
                );
            }
        }

        # Verify that ops did not decrease.
        my $baseline_ops = _sum_across_nodes(
            counter  =>'ops',
            nodes    => $nodes,
            sample   => $baseline_perf,
            workload => $workload,
        );
        my $observed_ops = _sum_across_nodes(
            counter  =>'ops',
            nodes    => $nodes,
            sample   => $observation->{STATS},
            workload => $workload,
        );
        my $min_ops =
            floor($baseline_ops - ($baseline_ops * $error_margin));

        if ($observed_ops < $min_ops) {
            $Log->comment(
                "FAIL: Workload '$workload' in policy-group '$policy_group'"
              . " did not achieve baseline ops\n"
              . "* Min acceptable ops: $min_ops\n"
              . "* Actual observed ops: $observed_ops"
            );
            $failures++;
        }
        else {
            $Log->comment(
                "PASS: Workload '$workload' in policy-group '$policy_group'"
              . " achieved baseline ops\n"
              . "* Min acceptable ops: $min_ops\n"
              . "* Actual observed ops: $observed_ops"
            );
        }

        # Verify that total_data did not decrease.
        my $baseline_total_data = _sum_across_nodes(
            counter  =>'total_data',
            nodes    => $nodes,
            sample   => $baseline_perf,
            workload => $workload,
        );
        my $observed_total_data = _sum_across_nodes(
            counter  =>'total_data',
            nodes    => $nodes,
            sample   => $observation->{STATS},
            workload => $workload,
        );
        my $min_total_data =
            floor($baseline_total_data - ($baseline_total_data * $error_margin));

        if ($observed_total_data < $min_total_data) {
            $Log->comment(
                "FAIL: Workload '$workload' in policy-group '$policy_group'"
              . " did not achieve baseline throughput\n"
              . "* Min acceptable total_data: $min_total_data\n"
              . "* Actual observed total_data: $observed_total_data"
            );
            $failures++;
        }
        else {
            $Log->comment(
                "PASS: Workload '$workload' in policy-group '$policy_group'"
              . " achieved baseline throughput\n"
              . "* Min acceptable total_data: $min_total_data\n"
              . "* Actual observed total_data: $observed_total_data"
            );
        }
    }

    $Log->exit();
    return $failures;
}

## This method processes a policy group present in an observation's SLO hash
## and determines if it is valid for comparison for the current baseline,
# Returns 0 (false) if the policy_group should not be compared
# Returns 1 (true) if the policy_group should be compared
sub _pg_valid_for_baseline_comparison {
    my %opts = validate(
        @_, {
            slo_summary   => { type => HASHREF },
            policy_group  => { type => SCALAR },
            stash         => { type => OBJECT },
            observation   => {type => HASHREF},
        },
    );

    my $slo_summary   = $opts{slo_summary};
    my $policy_group  = $opts{policy_group};
    my $stash         = $opts{stash};
    my $observation   = $opts{observation};

    my $valid = 1;
    my $slo = $observation->{SLO}->{$policy_group}->{slo};

    # If the policy_group has SLO set (max_throughput or target_latency), return 0
    if ((!exists($slo->{'max-throughput'})
        || uc($slo->{'max-throughput'}) eq 'INF')) {
       my @current_baseline_pgs = keys %{$slo_summary};

       # (1) If policy_group not tracked in this baseline, return 0.
       if (none {/^$policy_group$/} @current_baseline_pgs) {
          $valid = 0;
       }

       # (2) If policy_group present in more than one baseline, not possible to
       # determine which baseline to compare against, return 0 and confess.
       my $all_baselines = iteration_sequence_fetch(stash => $stash,
                                                    type  => 'baseline');
       my $policy_in_baseline = 0;
       foreach my $baseline_iter (@{$all_baselines}) {
           my $baseline_slo_summary = slo_summary_fetch(
              iteration => $baseline_iter,
              stash     => $stash,
           );
           my @baseline_pgs = keys %{$baseline_slo_summary};
           if (first {/^$policy_group$/} @baseline_pgs) {
              $policy_in_baseline++;
           }
       }
       if ($policy_in_baseline > 1) {
           $valid = 0;
           confess("Policy group $policy_group tracked in more than one baseline iteration");
       }
    }
    else {
       $valid = 0;
    }
    return $valid;
}

sub _sum_across_nodes {
    my %opts = validate(
        @_, {
            counter  => { type => SCALAR   },
            nodes    => { type => ARRAYREF },
            sample   => { type => HASHREF  },
            workload => { type => SCALAR   },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $counter = $opts{counter};
    my $nodes   = $opts{nodes};
    my $sample  = $opts{sample};
    my $wkld    = $opts{workload};
    my $sum     = 0;

    foreach my $node (@{$nodes}) {
        $sum += $sample->{"$node:workload:$wkld:$counter"};
    }

    $Log->exit();
    return $sum;
}


# Perform validation checks for 'max-throughput' policies,
# log any failures, and return the failure count.
sub _verify_max_throughput {
    my %opts = validate(
        @_, {
            error_margin => { type => SCALAR   },
            observation  => { type => HASHREF  },
            policy_group => { type => SCALAR   },
            slo_value    => { type => SCALAR   },
            stash        => { type => OBJECT   },
            workloads    => { type => ARRAYREF },
        },
    );
    $Log->enter(Dumper(\%opts));

    my $error_margin = $opts{error_margin};
    my $observation  = $opts{observation};
    my $policy_group = $opts{policy_group};
    my $slo_value    = $opts{slo_value};
    my $stash        = $opts{stash};
    my $workloads    = $opts{workloads};

    my $failures = 0;
    my $sample   = $observation->{STATS};

    # Make sure we have CM counter sample.
    confess 'Cannot perform verification without Perf::Stats sample'
        if !scalar(%{$sample});

    # Get the nodes being used for I/O by the Datastore objects.
    my $nodes = _nodes_with_io(
        observation => $observation,
        stash       => $stash,
    );


    # Make sure we've found some nodes.
    confess 'Datastore objects did not provide I/O to any nodes'
        if scalar(@{$nodes}) < 1;

    # Verify SLO limit was enforced, if max-throughput was not "infinite".
    if (uc($slo_value) ne 'INF') {
        # Get the SLO limit and the CM counter implied by the slo_value.
        my ($limit, $counter) = _parse_max_throughput_value($slo_value);
        my $min_expected      = floor($limit - ($limit * $error_margin));
        my $max_expected      = ceil($limit + ($limit * $error_margin));

        # Compare observed counter values with the expected values for
        # each workload.
        my $observed = 0;
        foreach my $wkld (@{$workloads}) {
            foreach my $node (@{$nodes}) {
                $Log->debug("Sampled value: $node:workload:$wkld:$counter: "
                          . $sample->{"$node:workload:$wkld:$counter"});
                $observed += $sample->{"$node:workload:$wkld:$counter"};
            }
        }
        $Log->debug("Total observed: $observed");

        # Verify that no workload exceeded its SLO limit by
        # greater than the error margin.
        if ($observed > $max_expected) {
            $Log->comment(
                "FAIL: Policy-group '$policy_group' exceeded"
              . " its throughput limit ($slo_value):\n"
              . "* Max acceptable for '$counter': $max_expected\n"
              . "* Actual observed for '$counter': $observed"
            );
            $failures++;
        }

        # Verify that no workload fell significantly below its SLO limit.
        # TODO: Confirm sufficient offered load before checking workload.
        if ($observed < $min_expected) {
            $Log->comment(
                "FAIL: Policy-group '$policy_group' failed"
              . " to reach its throughput limit ($slo_value):\n"
              . "* Min acceptable for '$counter': $min_expected\n"
              . "* Actual observed for '$counter': $observed"
            );
            $failures++;
        }
    }

    # Verify that trickle rates are 4/sec for Best-Effort (User & System)
    # 1/sec for Background policies 
    if($policy_group =~ /^(User-Best-Effort|_System-Best-Effort|_System-Background)$/){
        my $nodes = config_summary_fetch(stash => $stash)->{QosConfig}->{nodes};
        my $ops = 0;
	foreach my $node (@$nodes){
            foreach my $wkld (@{$workloads}) {
                    $ops += $sample->{"$node:workload:$wkld:ops"};
            }
        }
        if ((($policy_group =~ /Best-Effort/) &&( $ops < 4))|| 
               (($policy_group =~ /Background/) &&( $ops < 1))){
            $Log->comment(" trickle rate for Best-Effort policy is a min of 4 ops ") 
                   if  ($policy_group =~ /Best-Effort/) ;     
            $Log->comment(" trickle rate for Background policies is a min of 1 ops ")
                   if  ($policy_group =~ /Background/) ; 
            $Log->comment("Actual received ops for $policy_group are $ops ");
            $Log->comment("FAIL: $policy_group did not recive the trickle rate");
            $failures++;
        }
     }
    

    # Verify that all nodes with a Datastore op-rate
    # greater than zero performed some I/O.
    foreach my $id (keys(%{$observation->{IO}})) {
        no warnings qw(uninitialized);
        if ($observation->{IO}->{$id}->{op_rate} > 0 ){
            foreach my $node( @$nodes){
                my $ops = 0;
                foreach my $wkld (@{$workloads}) {
                    $ops += $sample->{"$node:workload:$wkld:ops"};
                }
                if ($ops < 1) { 
                    $Log->comment("FAIL: Node '$node' received no I/O");
                    $failures++;
                }
	    }
        }
    }
    $Log->exit();
    return $failures;
}


# Return the nodes whose nblades are being used for IO.
sub _nodes_with_io {
    my %opts = validate(
        @_, {
            observation => { type => HASHREF  },
            stash       => { type => OBJECT   },
        },
    );
    $Log->enter();                             # Avoid dumping observation again

    my $observation = $opts{observation};
    my $stash       = $opts{stash};
    my @nodes;

    foreach my $id (keys %{$observation->{IO}}) {
        # If a lif_owner is included in the current observation, use it.
        # Otherwise, use the one from the global section.
        if (defined(my $lif_owner = $observation->{IO}->{$id}->{lif_owner})) {
            push @nodes, $lif_owner;
        }
        else {
            push @nodes, io_summary_fetch(
                id    => $id,
                stash => $stash,
            )->{lif_owner};
        }
    }

    my @uniq_nodes = uniq(@nodes);

    $Log->exit(Dumper(\@uniq_nodes));
    return \@uniq_nodes;
}


# Return the throughput limit and the CM counter implied by the
# max-throughput SLO value.
sub _parse_max_throughput_value {
    $Log->enter(@_);
    my $slo_value = shift;

    # ONTAP is very permissive in the values it accepts for max-throughput.
    # See burt609648.
    confess "Invalid max-throughput value: '$slo_value'"
        if $slo_value !~ m/^(\d+(?:[.]\d+)?)([ibkmg]?)/i;

    my $base   = $1;
    my $suffix = uc($2);

    my ($limit, $counter);
    if ($suffix eq q{} || $suffix eq 'I') {
        $limit   = $base;
        $counter = 'ops';
    }
    elsif ($suffix eq 'B') {
        $limit   = $base;
        $counter = 'total_data';
    }
    elsif ($suffix eq 'K') {
        $limit   = $base * 1024;
        $counter = 'total_data';
    }
    elsif ($suffix eq 'M') {
        $limit   = $base * 1024 * 1024;
        $counter = 'total_data';
    }
    elsif ($suffix eq 'G') {
        $limit   = $base * 1024 * 1024 * 1024;
        $counter = 'total_data';
    }
    else {
        confess "Internal error: suffix not matched: '$suffix'";
    }

    $Log->exit();
    return ($limit, $counter);
}


# Compare the number of failed checks against the SHOULD_FAIL
# value for the observation, and return true if an unexpected
# result occurred.
sub _unexpected_result {
    $Log->enter();                             # Avoid dumping observation again
    my %opts = validate(
        @_, {
            failed_checks => { type => SCALAR  },
            observation   => { type => HASHREF },
            policy_group  => { type => SCALAR  },
            verification  => { type => SCALAR  },
        },
    );

    my $failed      = $opts{failed_checks};
    my $pg          = $opts{policy_group};
    my $should_fail = $opts{observation}->{SHOULD_FAIL};
    my $ver_fname   = $opts{verification};

    # If SHOULD_FAIL is a hashref, then only certain combinations
    # of policy groups and verification functions are expected to
    # fail.
    if (ref($should_fail) eq 'HASH') {
        if (exists($should_fail->{$pg})
            && any { $_ eq $ver_fname } @{$should_fail->{$pg}}) {
            $Log->comment(
                "NOTE: Policy-group '$pg' is expected to"
              . " fail '$ver_fname'"
            );
            $Log->exit();
            return $failed == 0;               # Expect some failures.
        }
        else {
            $Log->exit();
            return $failed > 0;                # Expect no failures.
        }
    }

    # Otherwise, interpret SHOULD_FAIL as a boolean, and expect
    # all policy groups and verification functions to fail.
    elsif (!ref($should_fail)) {
        if ($should_fail) {
            $Log->comment(
                'NOTE: All policy groups are expected to'
              . ' fail all verification checks'
            );
            $Log->exit();
            return $failed == 0;               # Expect some failures.
        }
        else {
            $Log->exit();
            return $failed > 0;                # Expect no failures.
        }
    }
    else {
        confess 'Unexpected value for SHOULD_FAIL: ' . Dumper($should_fail);
    }
}

1;

