# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  AAMP/SLSL test object
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::Test

=head1 DESCRIPTION

A Perf::AAMP::Test object represents an instance of an AAMP/SLSL test. AAMP is
an extensible framework for developing and executing performance measurement
tests in a PerfANT lab environment. SLSL is a domain-specific language used to
specify performance measurement tests. The Perf::AAMP::Test class allows NATE
tests to specify parameters and SLSL scripts (called setlists) for AAMP tests,
queue these tests in a PerfANT lab, respond to callouts from AAMP, and query
the results.

Callouts are Perl subroutines in the Perf::AAMP::Test object's address space
that may be called remotely from the AAMP process. They are typically defined
implicitly by being tagged with the ':Callout' attribute, although any coderef
may be registered as a callout via the add_callout() method. In addition to
the arguments passed from AAMP, callout routines receive a test resources
object. See the documentation for the Perf::AAMP::Callout package for
information about writing callout subroutines.

Both callouts and the Perf::AAMP::Test object have access to a common stash
of name/value pairs. They can use the stash to pass data back and forth.


    Example:

        my $test = Perf::AAMP::Test->new(
            test_runner_params => {
                test_type   => 'QOS_DISK_BOUND',
                model       => 'FAS3240_ANY_DOTc',
                client_type => 'IBM3650mLinuxRHEL5U3',
                disk_type   => 'DS2246-S540G107K-24',
                build       => '/x/eng/rlse/DOT/mainN',
            },
        );

        $test->stash->add(
            name  => 'policy-name',
            value => 'test_plcy',
        );

        my $results = $test->execute();


    Example callout:

        sub set_volume_policy :Callout {
            my $resources = shift;
            my %opts = validate(
                @_, {
                    volume  => { type => SCALAR },
                    vserver => { type => SCALAR },
                },
            );

            my $policy = $resources->stash->fetch(name => 'policy-name');

            NACL::C::Volume->modify(
                command_interface  => $resources->command_interface(),
                volume             => $opts{volume},
                vserver            => $opts{vserver},
                'qos-policy-group' => $policy,
            );

            return SLSL_RESULT_SUCCESS;
        }


Perf::AAMP::Test supports pluggable test runners. A test runner implements the
actions requested Perf::AAMP::Test's public methods. For example, the default
PerfANT test runner submits AAMP tests into a PerfANT lab's queue, as described
above. Other test runners may behave differently. The primary purpose of this
feature is to aid in the development of unit tests for the Perf::AAMP::Test
modules.

=head1 REFERENCES

    http://wikid.netapp.com/w/AAMP
    http://wikid.netapp.com/w/AAMP/SLSL
    http://wikid.netapp.com/w/PerfANT

=head1 SEE ALSO

    Perf::AAMP::Callout
    Perf::AAMP::Test::Resources
    Perf::AAMP::Test::Results
    Perf::AAMP::Test::Stash

=head1 METHODS

=head2 new

    Construct and return a Perf::AAMP::Test object.

=over

=item Arguments

B<test_runner_pkg> = string

I<Optional> The name of the test runner package. Defaults to
'Perf::AAMP::Test::Runner::PerfANT'. Test runner packages must
extend the Perf::AAMP::Test::Runner abstract base class.

B<test_runner_params> = hashref

I<Optional> A hashref of parameters passed to the test runner
constructor. The specific entries allowed and/or expected in
the hash are specific to the test runner package.

=back

=cut

package Perf::AAMP::Test; {
    use strict;
    use warnings;
    use Algorithm::Diff qw(sdiff);
    use Carp qw(confess);
    use Data::Dumper;
    use List::MoreUtils qw(uniq);
    use Net::Domain qw(hostfqdn);
    use NATE::Events qw(eventloop_once listener_add);
    use NATE::STAF qw(staf_global);
    use Object::InsideOut;
    use Params::Validate qw(:all);
    use Perf::AAMP::Callout qw(:slsl_results !MODIFY_CODE_ATTRIBUTES);
    use Perf::AAMP::Test::Resources;
    use Perf::AAMP::Test::Stash;
    use Tharn qw($Log);


    # OIO is evidently confused by this module requiring a test_runner_pkg
    # that is dynamically determined at runtime. Setting the following flag
    # lets it avoid erroneously complaining about "unhandled parameters."

    BEGIN {
        no warnings q(once);
        $OIO::Args::Unhandled::WARN_ONLY = 1;
    }


    use constant {
        SLSL_MESSAGE          => 'NATE/SLSL/Callback',
        SLSL_VALIDATE_CALLOUT => '_NATE_VALIDATE_',
        SLSL_INIT_CALLOUT     => '_NATE_INIT_',
        SLSL_TEARDOWN_CALLOUT => '_NATE_TEARDOWN_',
    };


    my @dispatch_table      :Field;
    my @resources           :Field;
    my @results             :Field;
    my @staf                :Field;
    my @stash               :Field;
    my @test_runner         :Field
                            :Type('Perf::AAMP::Test::Runner');
    my @test_runner_params  :Field
                            :Arg(test_runner_params)
                            :Default({});
    my @test_runner_pkg     :Field
                            :Arg(test_runner_pkg)
                            :Default('Perf::AAMP::Test::Runner::PerfANT');


    ### Public methods:

    sub _init :Init {
        my $self = shift;

        # Construct the callout dispatch table.
        #
        # First, insert the internal callouts.

        $self->add_callout(
            name => SLSL_VALIDATE_CALLOUT,
            code => sub { $self->_validate(@_) },
        );
        $self->add_callout(
            name => SLSL_INIT_CALLOUT,
            code => sub { $self->_initialize(@_) },
        );
        $self->add_callout(
            name => SLSL_TEARDOWN_CALLOUT,
            code => sub { $self->_teardown(@_) },
        );

        # Second, insert the user-defined callouts.
        #
        # We only want to include callouts that have been imported into the
        # package that this constructor is being called from. Therefore we'll
        # confirm that the coderef accessible via the callout's name in that
        # package name matches the callout's coderef. Callouts without matching
        # coderefs will be silently skipped.

        my ($caller_pkg) = caller(1);          # caller(0) is Object::InsideOut
        foreach my $callout (@{Perf::AAMP::Callout::_get_callouts()}) {
            my $name = $callout->{name};
            my $code = $callout->{code};

            no strict qw(refs);
            no warnings qw(uninitialized);
            if (*{$caller_pkg . q{::} . $name}{CODE} == $code) {
                $dispatch_table[$$self]->{$callout->{name}} = $callout->{code};
            }
        }

        $staf[$$self]    = NATE::STAF->staf_global();
        $stash[$$self]   = Perf::AAMP::Test::Stash->new();
        $results[$$self] = undef;

        # If no test_runner was specified, default to a PerfANT run.

        eval "require $test_runner_pkg[$$self]"
            or confess "Could not load package $test_runner_pkg[$$self]: $@";

        $test_runner[$$self] = $test_runner_pkg[$$self]->new(
            staf_endpoint      => hostfqdn(),
            staf_handle        => $staf[$$self]->handle(),
            msgtype            => SLSL_MESSAGE,
            internal_callouts  => {
                validate   => SLSL_VALIDATE_CALLOUT,
                initialize => SLSL_INIT_CALLOUT,
                teardown   => SLSL_TEARDOWN_CALLOUT,
            },
            test_runner_params => $test_runner_params[$$self],
        );
        return;
    }


=head2 add_callout

Register a coderef to be callable as an AAMP callout. Rather than
calling this method explicitly, most callouts are registered
implicitly via the ':Callout' attribute.

=over

=item Arguments

B<name> = string

I<Required> The name of the callout (i.e., 'callback_name' in the
SLSL setlist).

B<code> = coderef

I<Required> Coderef to be executed when the callout is called.

=back

=cut

    sub add_callout {
        my $self = shift;
        my %opts = validate(
            @_, {
                name => { type => SCALAR,  optional => 0 },
                code => { type => CODEREF, optional => 0 },
            },
        );
        my $name = $opts{name};
        my $code = $opts{code};

        confess "Callout '$name' already exists"
            if exists $dispatch_table[$$self]->{$name};

        $dispatch_table[$$self]->{$name} = $code;
        return $name;
    }


=head2 remove_callout

Deregister a callout. Once a callout is deregistered, it can no
longer be called.

=over

=item Arguments

B<name> = string

I<Required> The name of the callout to deregister.

=back

=cut

    sub remove_callout {
        my $self = shift;
        my %opts = validate(
            @_, {
                name => { type => SCALAR, optional => 0 },
            },
        );
        my $name = $opts{name};

        confess "No such callout: '$name'"
            if !exists $dispatch_table[$$self]->{$name};

        return delete $dispatch_table[$$self]->{$name};
    }


=head2 execute

Submit an AAMP test into the PerfANT lab queue, listen for and execute
callouts (until the teardown callout is called), then wait for the PerfANT
job to terminate, and return the results object.

=over

=item Arguments

None

=back

=cut

    sub execute {
        my $self = shift;

        # Submit run to PerfANT queue.
        my $run = $test_runner[$$self];
        $run->submit();

        # Continue listening for and performing AAMP/SLSL
        # callouts until we're done.
        $self->_listen();

        # Wait for run to complete and return the results.
        return $results[$$self] = $run->wait();
    }


=head2 results

Return the results object. It is an error to call this before the
PerfANT test completes and the results object is constructed.

Results objects are instances of test-runner-specific classes that
extend the Perf::AAMP::Test::Results abstract base class. See the
documentation for a that class for a description of its methods.

=over

=item Arguments

None

=back

=cut

    sub results {
        my $self    = shift;
        my $results = $results[$$self];

        confess 'Results unavailable'
            if !defined $results || !$results->isa('Perf::AAMP::Test::Results');

        return $results[$$self];
    }


=head2 stash

Return the stash object, which is an instance of the Perf::AAMP::Test::Stash
class. See the documentation for that class for a description of its methods.

=over

=item Arguments

None

=back

=cut

    sub stash {
        my $self = shift;
        return $stash[$$self];
    }


    ### Private methods:

    sub _listen :Private {
        my $self     = shift;
        my $staf     = $staf[$$self];
        my $dispatch = $dispatch_table[$$self];
        my $done     = 0;

        # Definer listener for SLSL_MESSAGE events.
        my $listener = sub {
            my $msg      = $_[0]->{message};
            my $endpoint = $_[0]->{machine};
            my $handle   = $_[0]->{handle};
            my $name     = $msg->{callback_name};
            my $args     = $msg->{args};
            $Log->debug('RECEIVED message from AAMP: ' . Dumper($msg));

            # Receiving an unknown callout is a fatal error.
            if (!exists($dispatch->{$name})) {
                my $error = "Unknown message name: '$name'";
                $staf->queue_post(
                    endpoint => $endpoint,
                    handle   => $handle,
                    type     => SLSL_MESSAGE,
                    message  => {
                        status  => SLSL_RESULT_ERROR,
                        payload => $error,
                    },
                );
                confess $error;
            }

            my $action = $dispatch->{$name};
            my ($result, $payload);

            # Perform the callout action. This assumes the internal _initialize
            # callout has already been invoked, otherwise the resources object
            # will not have been created.
            eval {
                $Log->debug("CALLING callout '$name' with args: "
                          . Dumper($args));

                ($result, $payload) = $action->($resources[$$self], %{$args});

                $Log->debug("RETURNED from callout '$name' with result"
                          . " '$result', payload: " . Dumper($payload));
            };

            # Construct reply. Note that AAMP doesn't understand "DONE",
            # so when we're done, we'll reply with "SUCCESS".
            my $err   = $@;
            my $reply = {
                callback_name => $msg->{callback_name},
                status  => $result,
                payload => defined $payload ? $payload : q{},
            };
            if ($err) {
                $reply->{status}  = SLSL_RESULT_ERROR;
                $reply->{payload} = $err;
            }

            # Send reply to the SLSL process.
            $Log->debug('SENDING reply to AAMP: ' . Dumper($reply));
            $staf->queue_post(
                endpoint => $endpoint,
                handle   => $handle,
                type     => SLSL_MESSAGE,
                message  => $reply,
            );

            # Exit event loop if we're done.
            $done = $name eq SLSL_TEARDOWN_CALLOUT;
        };

        # Register the SLSL_MESSAGE listener.
        listener_add(
            {
                filter   => sub { $_[0]->{type} eq SLSL_MESSAGE },
                callback => $listener,
            },
        );

        # Enter infinite loop waiting for SLSL_MESSAGE events.
        while (!$done) {
            $Log->debug('Waiting for event...');
            eventloop_once();
        }

        $Log->debug('Event loop exited!');
        return;
    }


    ### Internal callouts:

    sub _validate :Private {
        my $self   = shift;
        my $ignore = shift;
        my %opts   = validate(
            @_, {
                validate_list => { type => ARRAYREF, optional => 0 },
            }
        );

        my @nate_callouts = sort(keys %{$dispatch_table[$$self]});
        my @slsl_callouts = uniq(sort(@{$opts{validate_list}}));

        my $diff = sdiff(\@nate_callouts, \@slsl_callouts);
        my $missing_from_nate = $self->_missing_entries(
            diff      => $diff,
            fromfirst => 1,
        );
        my $missing_from_slsl = $self->_missing_entries(
            diff      => $diff,
            fromfirst => 0,
        );

        my $report = '';
        my $errcnt = 0;
        if (scalar(@{$missing_from_nate})) {
            $report .= sprintf(
                "ERROR: SLSL callouts not defined in NATE:\n%s\n",
                join(q{ }, @{$missing_from_nate}),
            );
            $errcnt++;
        }
        if (scalar(@{$missing_from_slsl})) {
            $Log->comment(
                sprintf("NATE callouts not mentioned in SLSL:\n%s\n\n",
                        join(q{ }, @{$missing_from_slsl})),
            );
        }
        confess $report if $errcnt;

        return SLSL_RESULT_SUCCESS;
    }


    sub _missing_entries :Private {
        my $self = shift;
        my %opts = validate(
            @_, {
                diff      => { type => ARRAYREF, optional => 0 },
                fromfirst => { type => SCALAR,   optional => 0 },
            }
        );
        my $diff = $opts{diff};
        my $fromfirst = $opts{fromfirst};

        my @missing;
        my ($regex, $index) = $fromfirst ? (qr/^[c+]$/, 2) : (qr/^[c-]$/, 1);

        foreach my $discrepancy (grep { $_->[0] =~ m/$regex/ } @{$diff}) {
            push @missing, $discrepancy->[$index];
        }

        return \@missing;
    }


    sub _initialize :Private {
        my $self   = shift;
        my $ignore = shift;
        my %opts   = validate(
            @_, {
                filers  => { type => ARRAYREF, optional => 0 },
                clients => { type => ARRAYREF, optional => 0 },
            }
        );

        confess 'SLSL_INIT_CALLOUT called more than once'
            if defined $resources[$$self];

        $resources[$$self] = Perf::AAMP::Test::Resources->new(
            %opts,
            stash => $stash[$$self],
        );

        return SLSL_RESULT_SUCCESS;
    }


    sub _teardown :Private {
        my $self   = shift;
        my $ignore = shift;
        return SLSL_RESULT_SUCCESS;
    }
}

1;

