# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  PerfANT lab resources object
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::Test::Resources

=head1 DESCRIPTION

The Perf::AAMP::Test::Resources object provides access to the PerfANT lab
resources being used by the test.

Note that these resources come into existance only after the PerfANT job
has been dequeued, the test hardware has been allocated, ONTAP installed,
and the cluster configured. For this reason, it is intended that the
resources object be available only through callouts, which are called
only after setup is complete and the AAMP/SLSL test begins executing.
NATE code should not attempt to access PerfANT lab resources outside of
callouts (therefore callouts should not store resource data in the stash),
since the resources will be released when the AAMP job terminates.

=head1 SEE ALSO

    NACL::C::Node
    Perf::AAMP::Test::Stash

=head1 METHODS

=cut

package Perf::AAMP::Test::Resources; {

    use strict;
    use warnings;
    use Carp;
    use NACL::C::Node;
    use Object::InsideOut;
    use Params::Validate qw(:all);

    my @client_list       :Field
                          :Arg(Name => 'clients', Mandatory => 1)
                          :Type(ARRAY_ref);

    my @filer_list        :Field
                          :Arg(Name => 'filers',  Mandatory => 1)
                          :Type(ARRAY_ref);

    my @stash             :Field
                          :Arg(Name => 'stash',   Mandatory => 1)
                          :Type('Perf::AAMP::Test::Stash');

    my @command_interface :Field;


    sub _init :Init {
        my $self = shift;
        my $node = $filer_list[$$self]->[0];

        confess 'Cannot create command_interface: no filers specified'
            if !defined $node;

        $command_interface[$$self] = NACL::C::Node->new(name => $node);
        return;
    }


=head2 clients

Return a list of the client hostnames being used by the test.
The callout can assume that a nate_hosts file exists in the
current sitedir for each client in the list.

=over

=item Arguments

None

=back

=cut

    sub clients {
        my $self    = shift;
        my $clients = $client_list[$$self];

        confess 'client list has not been initialized'
            if !defined $clients;

        return wantarray ? @{$clients} : $clients;
    }


=head2 filers

Return a list of the filer hostnames being used by the test.
The callout can assume that a nate_hosts file exists in the
current sitedir for each filer in the list.

=over

=item Arguments

None

=back

=cut

    sub filers {
        my $self   = shift;
        my $filers = $filer_list[$$self];

        confess 'filer list has not been initialized'
            if !defined $filers;

        return wantarray ? @{$filers} : $filers;
    }


=head2 command_interface

Return a NACL::C::Node object that can be used as a NACL command_interface
in the callout.

=over

=item Arguments

None

=back

=cut

    sub command_interface {
        my $self = shift;
        my $ci   = $command_interface[$$self];

        confess 'command_interface has not been initialized'
            if !defined $ci;

        return $ci;
    }


=head2 stash

Return the Perf::AAMP::Test::Stash object being used by the test.

=over

=item Arguments

None

=back

=cut

    sub stash {
        my $self = shift;

        confess 'stash has not been initialized'
            if !defined $stash[$$self];

        return $stash[$$self];
    }
}

1;

