# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  Base class for AAMP test runners
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::Test::Runner

=head1 DESCRIPTION

Perf::AAMP::Test::Runner is the abstract base class for Perf::AAMP::Test's
pluggable test runners. Each test runner should extend this class and
implements its methods.

Test runner objects are constructed and used by the Perf::AAMP::Test class,
rather than directly by user code.

=head1 METHODS

=head2 new

Construct and return a Perf::AAMP::Test::Runner subclass object.

=over

=item Arguments

B<test_runner_params> = hashref

I<Optional> A hashref of parameters passed to the test runner
constructor. The specific entries allowed and/or expected in
the hash are specific to the test runner package.

=cut

package Perf::AAMP::Test::Runner;

use strict;
use warnings;
use Carp qw(confess);
use File::Basename qw(fileparse);
use Params::Validate qw(:all);
use Tharn qw($Log);


=head2 new

Construct and return a Perf::AAMP::Test::Runner object.

=over

=item Arguments

B<test_runner_params> = hashref

I<Optional> A hashref of test-runner-specific parameters.
Note: this argument may be required by certain test runners.

=back

=cut

sub new {
    confess 'Method must be implemented in subclass: new';
}


=head2 results

Return the test-runner-specific results object. This object is a subclass of
Perf::AAMP::Test::Results.

=over

=item Arguments

None

=back

=cut

sub results {
    confess 'Method must be implemented in subclass: results';
}


=head2 submit

Called by Perf::AAMP::Test's execute() method to "submit" a test for execution.
For AAMP tests, this would involve submitting the test into the PerfANT lab
queue, and waiting for the test be be dequeued and begin execution. For other
test runners (such as ones designed for unit testing Perf::AAMP::Test code),
the actions performed may be quite different.

=over

=item Arguments

None

=back

=cut

sub submit {
    confess 'Method must be implemented in subclass: submit';
}


=head2 wait

Called by Perf::AAMP::Test's execute() method to wait for the "submitted" test
to complete execution.

=over

=item Arguments

None

=back

=cut

sub wait {
    confess 'Method must be implemented in subclass: wait';
}


### Restricted methods:

sub _setlist_pathname {
    my $self = shift;
    my %opts = validate(
        @_, {
            setlist => { type => SCALAR | UNDEF, optional => 1 },
        },
    );

    my $setlist = $opts{setlist};
    my ($test_script, $test_dir) = fileparse($ENV{TESTNAME});

    # If no setlist filename was specified, construct a default one
    # from the test script pathname.
    if (!defined $setlist) {
        my $filename = $test_script;
        $filename =~ s/thpl$/slsl/;
        $setlist = $test_dir . $filename;
    }

    # Otherwise, use the specified setlist filename. (If necessary,
    # prepend the test script directory name so that to make an
    # absolute pathname.)
    elsif ($setlist !~ qr{^/}) {
        $setlist = $test_dir . $setlist;
    }

    # Verify that the setlist file exists.
    confess "Setlist file does not exist: '$setlist'"
        if ! -f $setlist;

    return $setlist;
}


sub _slsl_libs_pathname {
    my $self        = shift;
    my ($nate_libs) = (__FILE__ =~ m{^(.*/Perf/AAMP/)Test/Runner[.]pm$});
    my $slsl_libs   = $nate_libs . 'QoS/SLSL_libs';
    return $slsl_libs;
}


sub _test_script_filename {
    my $self   = shift;
    my $script = fileparse($ENV{TESTNAME});
    return $script;
}


sub _validate_setlist {
    my $self = shift;
    my %opts = validate(
        @_, {
            aamp_install => { type => SCALAR, optional => 0 },
            setlist      => { type => SCALAR, optional => 0 },
        },
    );

    my $aamp_install = $opts{aamp_install};
    my $setlist      = $opts{setlist};
    my $slsl_libs    = $self->_slsl_libs_pathname();
    my $pythonpath   = "$slsl_libs:$aamp_install/"
                     . 'lib/python2.6/site-packages';
    my $sizzle_check = "PYTHONPATH=$pythonpath $aamp_install/"
                     . "bin/sizzle_check.py -c $setlist --no_warnings";

    $Log->debug("VALIDATING setlist with command:\n$sizzle_check");
    my $results = `$sizzle_check`;

    if ($results !~ m/Validation Successful/ms) {
        confess "$setlist is not a valid slsl script:\n$results";
    }
    else {
        $Log->debug("PASSED setlist validation:\n$results");
    }
    return;
}

1;

