# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  Pluggable test runner for Perf::AAMP::Test unit tests
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::Test::Runner::LocalNATE

=head1 DESCRIPTION

Perf::AAMP::Test::Runner::LocalNATE is a test runner for performing unit tests
of the Perf::AAMP::Test NATE libraries. Rather than submitting AAMP tests into
a PerfANT lab queue, Perf::AAMP::Test::Runner::LocalNATE sets up a STAF
listener in a local subtest, sends callout messages to a specified list of
callout subroutines, and logs their execution.

=cut

package Perf::AAMP::Test::Runner::LocalNATE; {
    use strict;
    use warnings;
    use Carp qw(confess);
    use Data::Dumper;
    use List::MoreUtils qw(any);
    use NATE::Events qw(listener_add);
    use NATE::ParamSet;
    use NATE::STAF::Peer;
    use Object::InsideOut qw(Perf::AAMP::Test::Runner);
    use Params::Validate qw(:all);
    use Perf::AAMP::Test::Results::LocalNATE;
    use Tharn qw($Log subtest);

    use constant {
        LOCALNATE_MESSAGE => 'LocalNATE/Callback',
    };


    my @actions             :Field
                            :Type(ARRAY_ref);
    my @activity_log        :Field
                            :Type(ARRAY_ref);
    my @internal_callouts   :Field
                            :Arg(Name => 'internal_callouts', Mandatory => 1)
                            :Type(HASH_ref);
    my @msgtype             :Field
                            :Arg(Name => 'msgtype', Mandatory => 1)
                            :Type(scalar);
    my @peer                :Field
                            :Type('NATE::STAF::Peer');
    my @peer_endpoint       :Field
                            :Arg(Name => 'staf_endpoint', Mandatory => 1)
                            :Type(scalar);
    my @peer_handle         :Field
                            :Arg(Name => 'staf_handle', Mandatory => 1)
                            :Type(numeric);
    my @results             :Field;
    my @runner              :Field;
    my @test_runner_params  :Field
                            :Arg(Name => 'test_runner_params', Mandatory => 1);


    sub _init :Init {
        my $self = shift;

        # Validate callouts parameter.
        my $callouts = $internal_callouts[$$self];
        my @required = qw(initialize teardown validate);
        confess 'LocalNATE tests require these internal callouts: '
              . join(', ', @required)
            if any { !exists($callouts->{$_}) } @required;

        my $actions = $test_runner_params[$$self]->{actions};
        confess 'Missing requred test_runner_params field: actions'
            if !defined($actions) || ref($actions) ne 'ARRAY';
        $actions[$$self] = $actions;

        $peer[$$self] = NATE::STAF::Peer->new(
            endpoint => $peer_endpoint[$$self],
            handle   => $peer_handle[$$self],
        );

        $results[$$self] = undef;
        return;
    }


    sub submit {
        my $self = shift;

        my $listener = sub {
            my $msg = $_[0]->{message};

        };

        listener_add(
            {
                filter   => sub { $_[0]->{type} eq LOCALNATE_MESSAGE },
                callback => sub {
                    $results[$$self] =
                        Perf::AAMP::Test::Results::LocalNATE->new(
                            test_results_params => {
                                activity_log => $_[0]->{message},
                            },
                        );
                },
            },
        );

        $runner[$$self] = subtest(
            sub { $self->_runner(@_) },
            -runid => 'LocalNATE',
            -bg,
            q{--},
            actions => $actions[$$self],
        );
        return;
    }


    sub wait {
        my $self = shift;
        $runner[$$self]->wait_finish();
        return $results[$$self];         # Set via callback in submit()
    }


    sub results {
        my $self = shift;
        return $results[$$self];         # Perf::AAMP::Test confesses if undef
    }


    sub _runner :Private {
        my $self = shift;
        my %opts = validate(
            @_, {
                actions => { type => ARRAYREF, optional => 0 },
            },
        );
        my $actions = $opts{actions};

        # Call internal validate and initialize callouts.

        my $response = $self->_send_message(
            message => $self->_validate_callout(),
        );

        $response = $self->_send_message(
            message => $self->_initialize_callout(),
        );

        # Loop through user-defined callouts until done.

        foreach my $action (@{$actions}) {
            my $name  = $action->{callout_name};
            my $args  = $action->{callout_args};

            $response = $self->_send_message(
                message => {
                    callback_name => $name,
                    args          => $args,
                },
            );

            last if $response->{status} eq 'DONE';
        }

        # Call internal teardown callout.

        $response = $self->_send_message(
            message => $self->_teardown_callout(),
        );

        # Create results & status.
        $peer[$$self]->message_put(
            type    => LOCALNATE_MESSAGE,
            message => $activity_log[$$self],
        );
        return;
    }


    sub _send_message :Private {
        my $self = shift;
        my %opts = validate(
            @_, {
                message => { type => HASHREF, optional => 0 },
            },
        );

        # Send the message.
        $Log->debug('CALLING callout: ' . Dumper($opts{message}));
        $peer[$$self]->message_put(
            type    => $msgtype[$$self],
            message => $opts{message},
        );

        # Return the response.
        my $response = $peer[$$self]->message_get();
        $Log->debug('RECEIVED response: ' . Dumper($response));

        return $self->_log($response);
    }


    sub _log :Private {
        my $self     = shift;
        my $response = shift;

        push @{$activity_log[$$self]}, $response;
        return $response;
    }


    sub _validate_callout :Private {
        my $self     = shift;
        my $actions  = $actions[$$self];
        my $validate = $internal_callouts[$$self]->{validate};

        confess 'Must specify validate internal callout name'
            if !defined $validate;

        my @callout_list = (
            values %{$internal_callouts[$$self]},    # Internal callouts
            map { $_->{callout_name} } @{$actions},  # User-defined callouts
        );

        return {
            callback_name => $validate,
            args => {
                validate_list => [ @callout_list ],
            },
        };
    }


    sub _initialize_callout :Private {
        my $self       = shift;
        my $params     = NATE::ParamSet->new(global => 1);
        my @clients    = split(/:/, $params->get('CLIENTS'));
        my @filers     = split(/:/, $params->get('FILERS'));
        my $initialize = $internal_callouts[$$self]->{initialize};

        return {
            callback_name => $initialize,
            args => {
                clients => [ @clients ],
                filers  => [ @filers  ],
            },
        };
    }


    sub _teardown_callout :Private {
        my $self     = shift;
        my $teardown = $internal_callouts[$$self]->{teardown};

        return {
            callback_name => $teardown,
            args => {},
        };
    }
}

1;

