# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  Test runner for PerfANT tests
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::Test::Runner::PerfANT

=head1 DESCRIPTION

Perf::AAMP::Test::Runner::PerfANT is the default test runner for
Perf::AAMP::Test tests. It provides Perf::AAMP::Test::Runner methods
for submitting AAMP tests into a PerfANT lab queue, waiting for the
AAMP test to complete, and returning a Perf::AAMP::Test::Results::PerfANT
object. It handles the creation of PerfANT and SLSL user parameters,
providing defaults, and allowing the user to override them via its
'test_runner_params' argument.

Objects of this class should be created by the Perf::AAMP::Test constructor,
rather than directly by user code.

=head1 SEE ALSO

    Perf::AAMP::Test
    Perf::AAMP::Test::Runner
    Perf::AAMP::Test::Runner::PerfANT::Installation
    Perf::AAMP::Test::Runner::PerfANT::Parameters

=head1 METHODS

=head2 new

    Construct and return a Perf::AAMP::Test::Runner::PerfANT object.

=over

=item Arguments

B<test_runner_params> = hashref

I<Required> A hashref of PerfANT and SLSL user parameters.

=back

=cut

package Perf::AAMP::Test::Runner::PerfANT; {
    use strict;
    use warnings;
    use Carp qw(confess);
    use Data::Dumper;
    use List::MoreUtils qw(any);
    use Log::Log4perl qw(:easy);
    use LWP::Simple qw(get);
    use Params::Validate qw(:all);
    use Perf::AAMP::Test::Results::PerfANT;
    use Perf::AAMP::Test::Runner::PerfANT::Installation;
    use Perf::AAMP::Test::Runner::PerfANT::Parameters;
    use Tharn qw($Log);
    use URI;
    use XML::LibXML;

    my %inst_params;                           # For loading libraries, etc.
    BEGIN {
        %inst_params =
            Perf::AAMP::Test::Runner::PerfANT::Installation::params();
    }

    use lib "$inst_params{PERFANT_INSTALL}/perl";
    use lib "$inst_params{PERFANT_INSTALL}/perl/PerfCommon/lib";
    use RunSubmissions::Enqueue;
    use RunSubmissions::PerfantParameters;
    no lib "$inst_params{PERFANT_INSTALL}/perl";
    no lib "$inst_params{PERFANT_INSTALL}/perl/PerfCommon/lib";

    use constant {                             # Grover Web services URL
        WS_URL => 'http://grover.rtp.netapp.com/KO/index.php/common/ws/process',
    };

    use namespace::clean;
    use Perf::AAMP::Test::Runner;
    use Object::InsideOut qw(Perf::AAMP::Test::Runner);


    my @installation        :Field
                            :Type(HASH_ref);
    my @internal_callouts   :Field
                            :Arg(Name => 'internal_callouts', Mandatory => 1)
                            :Type(HASH_ref);
    my @msgtype             :Field
                            :Arg(Name => 'msgtype', Mandatory => 1)
                            :Type(scalar);
    my @perfant_lab         :Field
                            :Type('PerfantLab');
    my @perfant_params      :Field
                            :Type('PerfantParameterSet');
    my @results             :Field
                            :Type('Perf::AAMP::Test::Results::PerfANT');
    my @setlist             :Field
                            :Type(scalar);
    my @runid               :Field
                            :Type(scalar);
    my @staf_endpoint       :Field
                            :Arg(Name => 'staf_endpoint', Mandatory => 1)
                            :Type(scalar);
    my @staf_handle         :Field
                            :Arg(Name => 'staf_handle', Mandatory => 1)
                            :Type(numeric);
    my @test_runner_params  :Field
                            :Arg(Name => 'test_runner_params', Mandatory => 1);
    my @test_script         :Field
                            :Type(scalar);


    sub _init :Init {
        my $self = shift;

        # Save the inst_params info.
        $installation[$$self] = {
            aamp    => $inst_params{AAMP_INSTALL},
            base    => $inst_params{INSTALL_BASE},
            lab     => $inst_params{LAB_NAME},
            perfant => $inst_params{PERFANT_INSTALL},
        };
        $Log->debug('INSTALLATION parameters: '
                  . Dumper($installation[$$self]));

        # Validate internal_callouts parameter.
        $Log->debug('VALIDATING internal_callouts: '
                  . Dumper($internal_callouts[$$self]));
        my @required = qw(initialize teardown validate);
        confess 'PerfANT tests require these internal callouts: '
              . join(', ', @required)
            if any { !exists($internal_callouts[$$self]->{$_}) } @required;


        # Initialize logging used by PerfCommon and RunSubmission classes.
        Log::Log4perl->easy_init({
            'log4perl.appender.Main' => 'Log::Log4perl::Appender::Screen',
            'level' => $Log->may_trace() ? 'TRACE'
                     : $Log->may_debug() ? 'DEBUG'
                     : 'INFO',
        });

        # Create and initialize PerfantLab object.
        $perfant_lab[$$self] = PerfantLab->getinstance($inst_params{LAB_NAME});
        $perfant_lab[$$self]->initialize();

        # Create PerfantParameterSet object.
        $perfant_params[$$self] = PerfantParameterSet->new();

        # Set PerfANT options.
        $self->_init_perfant_params(
            perfant_params => $test_runner_params[$$self],
        );
        return;
    }


=head2 submit

Submit AAMP test into the PerfANT lab queue, and poll the queue until the
test has been dequeued.

=over

=item Arguments

None

=back

=cut

    sub submit {
        my $self = shift;
        my $lab  = $perfant_lab[$$self];
        my $pp   = $perfant_params[$$self];

        $Log->comment('Performing sizzle_check on ' . $setlist[$$self]);
        $self->_validate_setlist(
            setlist      => $setlist[$$self],
            aamp_install => $installation[$$self]->{aamp},
        );

        {
            $Data::Dumper::Indent = 1;
            $Log->debug('SUBMITTING PerfANT test with parameters: '
                      . Dumper($perfant_params[$$self]));
        }

        # Submit the test into the lab's PerfANT queue.
        my $submission = QueueSubmission->new($pp, $lab);
        $submission->initialize();
        $submission->add_nate_element($self->_generate_nate_element());

        my $handle = QueueHandle->new($lab);
        my $result = $handle->submit($submission);

        confess 'ERROR in PerfANT test submission: ' . $result->error_text()
            if $result->has_errors();

        # Poll the submission queue until the PerfANT test
        # begins running.
        while (1) {
            my $status = $handle->get_status($submission);
            $Log->comment("PerfANT test status: '$status'");
            last if $status eq 'DEQUEUED';
            Tharn::snooze(60);     # Acquires lock, don't poll more frequently!
        }

        $runid[$$self] = $submission->get_testid();
        $Log->comment("PerfANT test running with testid: $runid[$$self]");
        return;
    }


=head2 wait

Poll the Grover 'runIdSearch' Web service until the PerfANT job has completed.
Then construct and return a Perf::AAMP::Test::Results::PerfANT object.

=over

=item Arguments

None

=back

=cut

    sub wait {
        my $self   = shift;
        my $lab    = $installation[$$self]->{lab};
        my $runid  = $runid[$$self];
        my $user   = $perfant_params[$$self]->get_value('user', 't');
        my $ws     = URI->new(WS_URL);
        my $parser = XML::LibXML->new();

        # Poll Grover until the PerfANT test either completes
        # successfully or exits with an error.
        #
        # For details on the 'runIdSearch' Web service we're using, see the
        # PerformanceEngineering/MeasurementQA/PerformanceMeasurement/Grover/UserGuide
        # page at wikid.netapp.com.
        $ws->query_form({
            action    => 'runIdSearch',
            labs      => qq('$lab'),           # Note: single quotes req'd
            runIdList => $runid,
            userId    => $user,
        });
        my $query = $ws->as_string();

        # TODO: Could this ever get stuck in an infinite loop? Or will
        # we always eventually get a valid 'Result' element (even if the
        # PerfANT test is manually killed)?
        while (1) {

            # Call the runIdSearch Web service and parse its response.
            my $results = get($query)
                or confess "HTTP GET failed on $query";
            my $doc  = $parser->parse_string($results);
            my ($ms) = $doc->findnodes('//message/header/status');

            # Make sure we got a valid reply from the Web service.
            confess "Bad query: $query"
                if $ms->to_literal() ne 'SUCCESS';

            # Try to get the 'Result' element from the parsed XML.
            # If it doesn't exist, we'll assume the PerfANT test is
            # still running, and check again later.
            #
            # TODO: Will this query ever return multiple 'Result' elements
            # in the 'ResultList' (perhaps for multi-node tests)? Will the
            # code below handle that case, or could it become confused by
            # multiple 'RunId' elements?
            my ($reslist) =
                $doc->findnodes('//message/body/MessageBody/ResultList');
            my ($result)  = $reslist->findnodes('./Result');
            if (defined $result) {
                my ($sc)   = $result->findnodes('./StatusCode');
                my $status = $sc->to_literal();

                # TODO: Are there any other 'StatusCode' values we
                # should handle?
                if ($status !~ m/^(?:SUCCESS|ERR_)/) {
                    confess "Unrecognized PerfANT StatusCode: '$status'";
                }

                # Create and return the PerfANT results object.
                $Log->debug("CREATING PerfANT results object with:\n$results");
                $results[$$self] = Perf::AAMP::Test::Results::PerfANT->new(
                    test_results_params => {
                        result_list => $reslist,
                        runid       => $runid,
                        status      => $status,
                    },
                );

                return $results[$$self];
            }

            # Didn't find a 'Result' element this time around.
            # Try again in a couple of minutes.
            Tharn::snooze(120);
        }
        return;                                # Not reached!
    }


=head2 results

Return the Perf::AAMP::Test::Results::PerfANT object already created by wait().

=over

=item Arguments

None

=back

=cut

    sub results {
        my $self = shift;
        return $results[$$self];
    }


=head2 set_params

Set parameter values in the RunSubmissions::PerfantParameters objects.
Assumes every parameter name it recognizes is PerfANT parameter, and
all others are SLSL user parameters. Thus cannot detect typos in the
parameter names!

=over

=item Arguments

Unnamed hash of parameter name/value pairs.

=back

=cut

    sub set_params {
        my $self   = shift;
        my %params = @_;

        my %param_type = (
            aggr_list             => 'm1',
            build                 => 'm1',
            client_type           => 't',
            disk_type             => 'm1',
            ic_type               => 'i',
            initiators            => 't',
            model                 => 'm1',
            num_clients           => 't',
            num_nodes             => 'm1',
            num_shelves           => 'm1',
            ok_options            => 'm1',
            perfant_dev_options   => 't',
            priority              => 't',
            protocol              => 't',
            purpose               => 't',
            setlist_file          => 't',
            sfo_status            => 'm1',
            sound_meter_poststats => 'm1',
            sound_meter_prestats  => 'm1',
            testname              => 't',
            user                  => 't',
        );

        my $pp = $perfant_params[$$self];
        foreach my $name (keys %params) {
            my $type = exists($param_type{$name})
                ? $param_type{$name}
                : 'slsl User Parameters';
            $pp->add_parameter($name, $type, $params{$name});
        }
        return;
    }


    ### Internal methods:

    # Set the initial RunSubmissions::PerfantParameters values. The settings
    # come from defaults, based on 'test_type' and defined in the
    # Perf::AAMP::Test::Runner::PerfANT::Parameters module, and user-supplied
    # values, specified via 'test_runner_params'. These initial settings may
    # be overridden by user calls to the set_params() method.
    sub _init_perfant_params :Private {
        my $self = shift;
        my %opts = validate(
            @_, {
                perfant_params => { type => HASHREF, optional => 0 },
            },
        );

        my $user_params = $opts{perfant_params};
        my $test_type   = delete $user_params->{test_type};

        # Remember setlist pathname and test script filename.
        $setlist[$$self] = $self->_setlist_pathname(
            setlist => $user_params->{setlist_file},
        );
        $test_script[$$self] = $self->_test_script_filename();

        # Initialize PerfANT parameters. First set parameters
        # to their default values.
        my $params = Perf::AAMP::Test::Runner::PerfANT::Parameters::defaults(
            test_type   => $test_type,
            test_script => $test_script[$$self],
        );

        # Make sure we have an absolute pathname for the
        # setlist_file parameter.
        $user_params->{setlist_file} = $setlist[$$self];

        # Merge the user-supplied parameters with the defaults.
        foreach my $key (keys %{$user_params}) {
            $params->{$key} = $user_params->{$key};
        }

        # Verify that values have been specified for all
        # required parameters.
        my @missing;
        foreach my $key (keys %{$params}) {
            if (!defined($params->{$key})) {
                push @missing, $key;
            }
        }
        confess 'Missing required PerfANT parameter(s): ' . join(', ', @missing)
            if scalar(@missing) > 0;

        # Add the user and default parameters to the
        # PerfantParameterSet object.
        $self->set_params(%{$params});

        # Finally, add the hardcoded parameters.
        my $aamp_run_config = $installation[$$self]->{aamp}
                            . '/aamp_run.cfg';
        my $defaults_dir    = $installation[$$self]->{perfant}
                            . '/PerfDev/defaults';  #TODO: should this be lab_name?
        my $pp              = $perfant_params[$$self];

        $pp->add_parameter('aamp_run_config', 't', $aamp_run_config);

        # Evidently, should NOT add the 'defaults_dir' parameter:
        # $pp->add_parameter('defaults_dir', 't', $defaults_dir);
        return;
    }


    # Return a XML::LibXML document for the PerfANT <nate>
    # 'config.xml' element.
    sub _generate_nate_element :Private {
        my $self = shift;
        my %opts = validate(
            @_, {
                endpoint => {
                    type    => SCALAR,
                    default => $staf_endpoint[$$self],
                },
                handle   => {
                    type    => SCALAR,
                    default => $staf_handle[$$self],
                },
                validate => {
                    type    => SCALAR,
                    default => $internal_callouts[$$self]->{validate},
                },
                init     => {
                    type    => SCALAR,
                    default => $internal_callouts[$$self]->{initialize},
                },
                teardown => {
                    type    => SCALAR,
                    default => $internal_callouts[$$self]->{teardown},
                },
                lab      => {
                    type    => SCALAR,
                    default => $::LAB,
                },
                msgtype  => {
                    type    => SCALAR,
                    default => $msgtype[$$self],
                },
                sitedir  => {
                    type    => SCALAR,
                    default => $::SITE_ROOT,
                },
            },
        );

        my $nate_element = <<"EOT";
    <nate>
      <callout>
        <endpoint>$opts{endpoint}</endpoint>
        <handle>$opts{handle}</handle>
        <msgtype>$opts{msgtype}</msgtype>
        <validate_callout>$opts{validate}</validate_callout>
        <init_callout>$opts{init}</init_callout>
        <teardown_callout>$opts{teardown}</teardown_callout>
      </callout>
      <testbed_setup>
        <sitedir>$opts{sitedir}</sitedir>
        <lab>$opts{lab}</lab>
      </testbed_setup>
    </nate>
EOT
        $Log->debug("PerfANT <nate> element:\n$nate_element");

        my $parser  = XML::LibXML->new();
        my $doc     = $parser->parse_string($nate_element);
        my $element = $doc->documentElement();

        $Log->debug('Parsed <nate> element: ' . Dumper($element));
        return $element;
    }
}

1;

