# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  Installation parameters for Perf::AAMP::Test::PerfANT tests.
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::Test::Runner::PerfANT::Installation

=head1 DESCRIPTION

This package defines a params() function that returns a hash of installation
parameters for PerfANT tests.

=cut

package Perf::AAMP::Test::Runner::PerfANT::Installation;

use strict;
use warnings;
use NATE::ParamSet;
use Params::Validate qw(:all);
use base q(Exporter);

our @EXPORT_OK = qw(params);

=head1 FUNCTIONS

=head2 params

Return installation parameters in a hash. These parameters are used by the
PerfANT test runner to submit tests into the specified PerfANT lab's queue.

As Perf::AAMP::Test tests are run in new PerfANT labs, it would be convenient
to add their parameters to this package.

=over

=item Arguments

B<install_base> = string

I<Optional> Defaults to the value of the INSTALL_BASE NATE parameter,
which, in turn, defaults to 'perfqa_stable.' The install base determines
the default values for the AAMP_INSTALL and PERFANT_INSTALL NATE parameters.

B<lab_name> = string

I<Optional> The PerfANT lab. Defaults to the value of the LAB_NAME NATE
parameter, which, in turn, defaults to 'PerfDev.'

=back

=cut

sub params {
    my $paramset = NATE::ParamSet->new(global => 1);
    my %opts     = validate(
        @_, {
            install_base => {
                type    => SCALAR,
                default => $paramset->get(
                    'INSTALL_BASE',
                    default => 'perfqa_stable',
                ),
            },

            lab_name => {
                type    => SCALAR,
                default => $paramset->get(
                    'LAB_NAME',
                    default => 'PerfDev',
                ),
            },
        },
    );

    my $base = $opts{install_base};
    my $lab  = $opts{lab_name};

    if ($base eq 'perfqa_stable') {
        return (
            INSTALL_BASE    => 'perfqa_stable',
            LAB_NAME        => $lab,
            AAMP_INSTALL    => $paramset->get(
                'AAMP_INSTALL',
                default => '/x/eng/rtpperf/perf_qa/stable/aamp_install',
            ),
            PERFANT_INSTALL => $paramset->get(
                'PERFANT_INSTALL',
                default => '/x/eng/rtpperf/perf_qa/stable/perfant',
            ),
        );
    }
    elsif ($base eq 'perfqa_test') {
        return (
            INSTALL_BASE    => 'perfqa_test',
            LAB_NAME        => $lab,
            AAMP_INSTALL    => $paramset->get(
                'AAMP_INSTALL',
                default => '/x/eng/rtpperf/perf_qa/test/aamp_install',
            ),
            PERFANT_INSTALL => $paramset->get(
                'PERFANT_INSTALL',
                default => '/x/eng/rtpperf/perf_qa/test/perfant',
            ),
        );
    }
    elsif ($base eq 'perfdev') {
        return (
            INSTALL_BASE    => 'perfdev',
            LAB_NAME        => $lab,
            AAMP_INSTALL    => $paramset->get(
                'AAMP_INSTALL',
                default => '/x/eng/perfdev/aamp_runtime',
            ),
            PERFANT_INSTALL => $paramset->get(
                'PERFANT_INSTALL',
                default => '/x/eng/perfdev',
            ),
        );
    }
    else {
        return (
            INSTALL_BASE    => $base,
            LAB_NAME        => $lab,
            AAMP_INSTALL    => $paramset->get(
                'AAMP_INSTALL',
                default => $base . '/aamp_runtime',
            ),
            PERFANT_INSTALL => $paramset->get(
                'PERFANT_INSTALL',
                default => $base,
            ),
        );
    }
}

1;

