# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  Test-type-specific PerfANT parameter defaults
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::Test::Runner::PerfANT::Parameters

=head1 DESCRIPTION

This package defines a defaults() function that returns a hashref of
default PerfANT parameter values for the specified test type.

=cut

package Perf::AAMP::Test::Runner::PerfANT::Parameters;

use strict;
use warnings;
use Carp;
use NATE::ParamSet;
use Params::Validate qw(:all);
use base q(Exporter);

our @EXPORT_OK = qw(defaults);


=head1 FUNCTIONS

=head2 defaults

Return default values for PerfANT parameters based on test_type.

=over

=item Arguments

B<test_script> = string

I<Required> The name of the NATE script (i.e., the thpl) running this test.

B<test_type> = string

I<Required> The type (e.g., QOS_DISK_BOUND) of this test. Mapped to the
corresponding test_type understood by the Perf/AAMP/QoS/SLSL_libs, and
used to set test-type-specific defaults.

=back

=cut

sub defaults {
    my %opts = validate(
        @_, {
            test_script => {
                type     => SCALAR,
                optional => 0,
                regex    => qr/[.]thpl$/,
            },

            test_type   => {
                type     => SCALAR,
                optional => 0,
                regex    => qr/^QOS_(?:CPU|CSM|DISK)_BOUND$/,
            },
        },
    );


    # Initialize test_type-independent defaults. Note that
    # not all PerfANT parameters can be assigned a default.
    my $script    = $opts{test_script};
    my $type      = $opts{test_type};
    my ($purpose) = ($script =~ m/^(.*)[.]thpl$/);
    my $paramset  = NATE::ParamSet->new(global => 1);
    my $defs      = {
        aggr_list           => undef,
        build               => undef,
        client_type         => undef,
        disk_type           => undef,
        ic_type             => 'DOT_cluster',
        initiators          => q{},
        model               => undef,
        num_clients         => '2',
        num_nodes           => '2',
        num_shelves         => '1',
        perfant_dev_options => 'ignore_rule_initiators',
        priority            => '3',
        protocol            => undef,
        purpose             => $purpose,
        setlist_file        => undef,
        sfo_status          => 'enabled',
        testname            => 'motley',
        user                => $paramset->get('USER'),
    };

    # Initialize test_type-specific defaults.
    if ($type eq 'QOS_CPU_BOUND') {
        $defs->{test_type} = 'CPUbound';
        $defs->{aggr_list} = 'aggr1,21';  # PqLab per 3/29/2012 heinze email
        $defs->{protocol}  = 'nfs';
    }
    elsif ($type eq 'QOS_DISK_BOUND') {
        $defs->{test_type} = 'Diskbound';
        $defs->{aggr_list} = 'aggr1,4';
        $defs->{protocol}  = 'nfs';
    }
    else {
        # The parameter validation regex should prevent us from
        # hitting this, but we'll check anyway in case the code
        # diverges.
        confess "Unsupported test_type: '$type'";
    }

    return $defs;
}

1;

