# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary  Stash of name/value pairs
## @status   public
##
## @pod here

=head1 NAME

Perf::AAMP::Test::Stash

=head1 DESCRIPTION

The stash is a simple repository of name/value pairs used by callouts to store
and retrieve data. The stash allows callouts to share data with each other, and
with the NATE script that queues the AAMP test, without having to side effect
global variables. This is useful since callout routines are invoked remotely
from AAMP processes.

User code should not create Perf::AAMP::Test::Stash objects directly, rather
it should use the one returned by the stash() method of the Perf::AAMP::Test
or Perf::AAMP::Callout::Dispatcher object. Callouts can access this stash via
the stash() method of their Perf::AAMP::Test::Resources argument.

=head1 SEE ALSO

    Perf::AAMP::Test
    Perf::AAMP::Callout::Dispatcher
    Perf::AAMP::Test::Resources

=head1 METHODS

=cut

package Perf::AAMP::Test::Stash; {
    use strict;
    use warnings;
    use Carp;
    use Data::Dumper;
    use Object::InsideOut;
    use Params::Validate qw(:all);

    my @delegate :Field :Default(undef);
    my @stash    :Field;


=head2 exists

Return true if the named entry exists in the stash, false otherwise.

=over

=item Arguments

B<name> = string

I<Required> The name of the stash entry to check for.

=back

=cut

    sub exists {
        my $self = shift;
        my %opts = validate(
            @_, {
                name => { type => SCALAR, optional => 0 },
            },
        );

        return exists($stash[$$self]->{$opts{name}});
    }


=head2 add

Add a name/value pair to the stash. It is an error if an entry with the
specified name already exists in the stash.

=over

=item Arguments

B<name> = string

I<Required> The name of the entry to add to the stash.

B<value>

I<Required> The initial value to associate with the name in the stash.

=back

=cut

    sub add {
        my $self  = shift;
        my %opts  = validate(
            @_, {
                name  => { type => SCALAR, optional => 0 },
                value => {                 optional => 0 },
            },
        );
        my $name  = $opts{name};
        my $value = $opts{value};

        confess "Already have entry for '$name' in stash"
            if exists($stash[$$self]->{$name});

        $stash[$$self]->{$name} = $value;
        return $name;
    }


=head2 remove

Remove a name/value pair from the stash. It is an error if an entry with
the specified name does not exist in the stash.

=over

=item Arguments

B<name> = string

I<Required> The name of the stash entry to remove.

=back

=cut

    sub remove {
        my $self = shift;
        my %opts = validate(
            @_, {
                name => { type => SCALAR, optional => 0 },
            },
        );
        my $name = $opts{name};

        confess "No entry for '$name' in stash"
            if !exists($stash[$$self]->{$name});

        delete $stash[$$self]->{$name};
        return $name;
    }


=head2 update

Update the value associated with the specified name in the stash. It is an
error if an entry with that name doesn't exist in the stash, unless a true
value is supplied for the 'create_if_nonexistent' parameter.

=over

=item Arguments

B<name> = string

I<Required> The name of the stash entry to update.

B<value>

I<Required> The new value for the entry.

B<create_if_nonexistent> = boolean

I<Optional> Optionally create an entry with the specified name in the stash,
if it doesn't already exist.

=back

=cut

    sub update {
        my $self  = shift;
        my %opts  = validate(
            @_, {
                name  => { type => SCALAR, optional => 0 },
                value => {                 optional => 0 },
                create_if_nonexistent => {
                    type    => SCALAR,
                    default => 0,
                },
            },
        );
        my $create_ok = $opts{create_if_nonexistent};
        my $name      = $opts{name};
        my $value     = $opts{value};

        confess "No entry for '$name' in stash"
            if !exists($stash[$$self]->{$name}) && !$create_ok;

        $stash[$$self]->{$name} = $value;
        return $name;
    }


=head2 fetch

Return the value associated with then name in the stash. It is an error
if an entry with the specified name does not exist in the stash.

=over

=item Arguments

B<name> = string

I<Required> The name of the stash entry to return.

=back

=cut

    sub fetch {
        my $self = shift;
        my %opts = validate(
            @_, {
                name => { type => SCALAR, optional => 0 },
            },
        );
        my $name = $opts{name};

        confess "No entry for '$name' in stash"
            if !exists($stash[$$self]->{$name});

        return $stash[$$self]->{$name};
    }


=head2 dump

Return a string representation of the stash contents.

=cut

    sub dump {
        my $self = shift;

        local $Data::Dumper::Indent    = 1;
        local $Data::Dumper::Quotekeys = 0;
        local $Data::Dumper::Sortkeys  = 1;
        local $Data::Dumper::Terse     = 1;

        chomp(my $str = Dumper($stash[$$self]));
        return $str;
    }


=head2 delegation_package

Specify a package in which to delegate calls to unknown methods.

Subroutines in the specified package must accept a 'stash' keyword
argument, which is the current Perf::AAMP::Test::Stash object (i.e.,
$self). The autoload method will attempt to call a subroutine with
the method name in the specified package. The point of this is to
allow the stash object to be "extended" with test-specific "methods".

    Example:

        my $stash = Perf::AAMP::Test::Stash->new();
        $stash->delegation_package(pkg => 'Perf::AAMP::QoS::TestSummary');

        $stash->statistics_summary_update(
            iteration   => 'baseline',
            observation => $delta,
        );

=over

=item Arguments

B<pkg> = string

I<Required> The name of the package to delegate method calls to.

=back

=cut

    sub delegation_package {
        my $self = shift;
        my %opts = validate(
            @_, {
                pkg => { type => SCALAR | UNDEF, default => undef },
            },
        );

        $delegate[$$self] = $opts{pkg};
        eval "require $delegate[$$self]"
            if defined $delegate[$$self];
        return;
    }


    sub _autoload :Automethod {
        my $self = shift;
        my $name = $_;
        my @args = @_;
        my $pkg  = $delegate[$$self];

        return if !defined($pkg);

        my $subr = $pkg . q{::} . $name;
        no strict q(refs);
        return sub { $subr->(@args, stash => $self) };
    }
}

1;

