# -*- perl -*-
# $Id: //depot/prod/test/main/ontap/boilermaker/lib/Perf/Aggrmetrics.pm#3 $
#
## @author   jroussin@netapp.com
## @summary  
## @status   private
##
## @pod here

package Aggrmetrics;

use strict;
use warnings;

use filer;
use Tharn;
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Module::Common qw(validate_mixed);
use Carp qw(confess);
use Perf::Utils qw(min max get_random_element as_string in_bytes get_file_size);
use Perf::ModeAgnostic qw(say_wrap);

#
# Define some WAFL constants
#
my $TETRIS_SIZE = 32;

sub new {
    my $class = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {
                fc => {type => OBJECT},
                aggr => {type => SCALAR},
                debug => {type => SCALAR | UNDEF, optional => 1, default => 0},
           }
           );
    my $self = \%params;
    
    # TODO: check params
    
    bless $self, $class;
    
    my $segment_size_factor = $self->_get_segment_size_factor();
    my $blocks_per_segment_per_spindle = $TETRIS_SIZE * $segment_size_factor;
    my $num_aggr_disks = $self->_get_num_aggr_disks();
    $self->{blocks_per_segment} = $blocks_per_segment_per_spindle * $num_aggr_disks;
    
    # Keeps track of the ids of the various scanner runs triggered by this object
    $self->{scanner_ids} = [];
    
    $self->{verbose} = "-noverbose";
    if ($self->{debug}){
        $self->{verbose} = "-verbose";
    }
        
    return $self;
}

sub start {
    my $self = shift;

    my $cnt = 5;
    say_wrap($self->{fc}, "priv set -q diag;wafl scan aggrmetrics $self->{aggr};");
    while ($cnt) { 
        my $out = say_wrap($self->{fc}, "priv set -q diag;wafl scan status -A $self->{aggr}");
        my $scanner_id;
        if ($out =~ /\s+(\d+)\s+aggregate metrics/){
            $self->{scanner_id} = $1;
            push @{$self->{scanner_ids}}, $1;
            last;
        } 
        Tharn::snooze(1);
        $cnt--;
        next;
    }

    unless (defined $self->{scanner_id}) {
        confess "aggrmetrics did not start. Unable to parse aggrmetrics scanner id";
    }
}

sub wait {
    my $self = shift;
    
    my $still_running = 1;
    while ($still_running){
        my $out = say_wrap($self->{fc}, "priv set -q diag;wafl scan status -A $self->{aggr}");
        if ($out =~ /\s+(\d+)\s+aggregate metrics/){
            my $id = $1;
            if ($self->{scanner_id} && $self->{scanner_id} != $id){
                logwarn("Last known aggrmetrics scanner id on aggregate '$self->{aggr}' was " .
                        "'$self->{scanner_id}', but current one is $id");
                $self->{scanner_id} = $id;
                push @{$self->{scanner_ids}}, $id;
            }
            sleep 5;
            next;
        }
        else{
            last;
        }
    }
}

sub DESTROY {
    my $self = shift;
    my $old_error = $@;
    eval {
        $self->purge();
    };
    if ($@){
        logcomment("$@");
    }
    $@ = $old_error;
}

sub purge {
    my $self = shift;
    
    my $errors;
    for my $scanner_id (@{$self->{scanner_ids}}){
        my $filename = "WAFL_aggrmetrics." . $self->{aggr} . ".$scanner_id";
        my $out = say_wrap($self->{fc}, "priv set -q diag;rm /vol/vol0/etc/$filename");
        if ($out){
            $errors->{$scanner_id} = $out;
        }
    }
    if ($errors){
        my $str = "Errors while deleting aggrmetrics files:\n";
        for my $sid (keys %$errors){
            $str .= "$sid : " .$errors->{$sid} ." \n";
        }
        confess $errors;
    }
}

sub get_output ($;$){
    my $self = shift;
    my $scanner_id = shift;
    
    if (!defined($scanner_id)){
        if ($self->{scanner_id}){
            $scanner_id = $self->{scanner_id};
        }
        else{
            confess "No preexisting scanner id, you need to specify one explicitly";
        }
    }
    
    my $filename = "WAFL_aggrmetrics." . $self->{aggr} . ".$scanner_id";
    my (undef, undef, $out) = say_wrap($self->{fc}, "priv set -q diag; rdfile /vol/vol0/etc/$filename", $self->{verbose});
    my $found_lines = 0;
    while ($out =~ /(\d+),(\d+),(\d+),(\d+),(\d+)/g){
        $found_lines ++;
        last;
    }
    if (!$found_lines){
        confess "Unable to read WAFL aggrmetrics csv file: $out";
    }
    return \$out;
}

#
# Check how big a segment is going to be
#
sub _get_segment_size_factor {
    my $self = shift;
    my $out = say_wrap($self->{fc},"priv set -q diag;printflag wafl_blkr_segment_size");
    
    if ($out =~ /wafl_blkr_segment_size = (\d+)/){
        logcomment("wafl_blkr_segment_size = $1");
        return $1;
    }
    else{
        return 1;
    }
    
}

sub _get_num_aggr_disks{
    
    my $self = shift;
    
    if (!$self->{num_aggr_disks}){
        my $out = say_wrap($self->{fc},"aggr status -r " . $self->{aggr});
        my $i = 0;
        while ($out =~ /data\s+(\d\w\.\d+)(\.\d+)?\s+/g){
            $i ++;  
        }
        if (!$i){
            confess "Unable to find any data disks for aggregate '" . $self->{aggr} ."'";
        }
        $self->{num_aggr_disks} = $i;
        return $i;
    }
    else{
        return $self->{num_aggr_disks};
    }
}

sub get_realloc_blocks($) {
    my $self = shift;
    $self->start();
    $self->wait();
    return $self->parse_output();
}

sub parse_output($;$){
    my $self = shift;
    my $output_ref = shift;
    
    if (!defined($output_ref)){
        $output_ref = $self->get_output();
    }
    
    my $found_lines = 0;
    my $criteria = 30; # Don't count the full or nearly full segments
    my $realloc_segment_count = 0;
    my $realloc_block_min_count = 0;
    my $max_block_count = 0;
    
    while ($$output_ref =~ /(\d+),(\d+),(\d+),(\d+),(\d+)/g){
        my $position = $1;
        my $utilization = $2;
        my $chain_cost = $3;
        my $stripe_cost = $4;
        my $segment_count = $5;
        
        $found_lines++;
        
        if ($utilization > 0){
            $max_block_count += int($self->{blocks_per_segment} * ($utilization/31)) * $segment_count;
            if ($utilization <= $criteria){
                $realloc_segment_count += $segment_count;
                my $low_block_count_per_segment = max( 0,int($self->{blocks_per_segment}*($utilization - 1)/31) ) + 1;
                my $low_block_count = $segment_count * $low_block_count_per_segment;
                $realloc_block_min_count += $low_block_count;
                logcomment("qualifying segments. Util: $utilization. Count: $segment_count. Blocks: $low_block_count");
            }
            else{
                logcomment("non-qualifying segments. Util: $utilization. Count: $segment_count");
            }
        }
    }
    
    if (!$found_lines){
        confess "Unable to read WAFL aggrmetrics csv file: $$output_ref";
    }
    
    logcomment("Lines read: $found_lines. Qualifying segments: $realloc_segment_count. Min. blocks: $realloc_block_min_count. Max. blocks: $max_block_count");
    if (wantarray){
        return ($realloc_block_min_count,$max_block_count);
    }
    else{
        confess "Subroutine returns array";
    }
}
1;
