# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Perf/Block_Layout.pm#1 $
# Copyright (c) 2008 NetApp, Inc.
# All rights reserved.
# Perf/Block_Layout.pm
#
## @summary  Perl library routines for obtaining and parsing the on-disk block layout.
## @author   farida@netapp.com
## @status   private

package Perf::Block_Layout;

use strict;
use warnings;  
use Tharn;
use Carp;
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Module::Common qw(validate_mixed);
use filer;
use Filer::Version qw(get_ontap_version version_compare);
use Volume qw(vol_root);
use GenericMount qw(mount umount);
use SnapCmds qw(snapshot_status snapshot_exists);

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(scramble_dumpblocks parse_dumpblocks_file count_large_bnos);



# Subroutine to run the scramble dumpblocks command on the filer.
# command: scramble dumpblocks -v -F -S <snapid> -u -z <volume> [fileid]
sub scramble_dumpblocks {
    my %args = validate_mixed(
    @_,
    [],
    {
       fc => {type => OBJECT},
       vol => {type => SCALAR},
       verbose => {type => SCALAR | UNDEF, optional => 1},                    # option -v
       mismatches_only => {type => SCALAR | UNDEF, optional => 1},            # option -F
       snapshot => {type => SCALAR | UNDEF, optional => 1},                   # option -S, snapshot name.
       list_pvbn_unknown => {type => SCALAR | UNDEF, optional => 1},          # option -u
       gzip_compress => {type => SCALAR | UNDEF, optional => 1},              # option -z
       file => {type => SCALAR | UNDEF, optional => 1},                       # can be a fileid or a filename, example /vol/myvol/myfile.
       offline_vol => {type => SCALAR | UNDEF, optional => 1, default => 1},
       flush_cache => {type => SCALAR | UNDEF, optional => 1, default => 1},
       timeout => {type => SCALAR | UNDEF, optional => 1, default => -1},     # default is infinite timeout.
    }
    );

    # Determine the Filer version.
    my $GB = '7.2';              # Version number for RGordonBiersch;
    my $filer_version = get_ontap_version($args{fc}->hostp());
    my $ret = version_compare($filer_version, $GB);
    my $not_GB;
    if ($ret == -1) {
       confess "Scramble dumpblocks is not supported in versions < $GB";
    }
    elsif ($ret == 0) {
       $not_GB = 0;
    }
    elsif ($ret == 1) {
       $not_GB = 1;
    }
    else {
       confess "Could not compare Filer's current version with $GB. Filer's current version is ($filer_version)";
    }

    # Make sure all blocks are on disk
    filer_sync($args{fc});
    filer_sync($args{fc});
    
    # Flush the buffer cache
    if ($args{flush_cache}){
        say($args{fc}, "priv set -q diag; wafl flush cache");
    }

    if ($args{offline_vol}){
        filer_vol_offline($args{fc}, $args{vol});
        filer_vol_online($args{fc}, $args{vol});
    }
    
    my $cmd = "priv set -q test; scramble dumpblocks";
    if ($args{verbose}){
        $cmd .= " -v";
    }
    if ($args{mismatches_only}){
        $cmd .= " -F";
    }
    if ($args{list_pvbn_unknown} && $not_GB){
        $cmd .= " -u";
    } 
    if ($args{gzip_compress} && $not_GB){
        $cmd .= " -z";
    } 

    # -S (snapshot) option (available in 7.3 onwards)
    if ($not_GB) {
       my $snapid;

       # If a snapshot is defined, then find it's snapid and do scramble dumpblocks on that snapshot.
       if (defined $args{snapshot}) {
          my @snaps = snapshot_status($args{fc}, $args{vol});
          foreach my $snap(@snaps) {
             if ($snap->{name} eq $args{snapshot}) {
                $snapid = $snap->{snapid};
             }
          }
       }
       # If a snapshot is not defined, then do scramble dumpblocks on the active filesystem i.e. snapid = 0.
       else {
          $snapid = 0;
       }
       # Make sure the snapshot still exists
       if (($snapid != 0) && (! snapshot_exists($args{fc}, $args{vol}, $args{snapshot}))){
           confess "Snapshot $args{snapshot} (snapid = $snapid) specified for scramble dumpblocks does not exist.";
       }
       $cmd .= " -S $snapid";
    }

    # Specify the volume on which to run the command.
    my $cmdstr = "$cmd $args{vol}";

    # If a file is defined, do scramble dumpblocks on that file.
    my $file_id;
    if (defined $args{file}) {
       if ($args{file} =~ /^0x(\d+)$/) {
          $file_id = $args{file};
       }
       # Find the fileid of the filename specified.
       else {
          my (undef, undef, $out) = say($args{fc}, "priv set -q test; showfh $args{file}");
          if ($out =~ /fileid=(\w+)/) {
             $file_id = $1;
          }
          else {
             confess "Unable to get fileid for file $args{file}: $out";
          }
       }
       $cmdstr .= " $file_id";
    }
  
    my $dumpblock_file;
    my (undef, undef,$out) = say($args{fc}, $cmdstr, '-timeout', $args{timeout}, '-maxidle', 300, -verbose);
    if ($out =~ /scramble dumpblocks: output being saved to .*(WAFL_dumpblocks\S+)/) {
        $dumpblock_file = $1;
    }
    elsif ($out =~ /usage/){
        confess "scramble dumpblocks not supported";
    }
    else {
        confess "Unable to get scramble dumpblocks output: $out";
    }

    # vol_root needs the full path called out to prevent issues from SanFiler.pm using a different Volume.pm
    my $root_vol = Volume::vol_root($args{fc});
    my $filepath = "/vol/$root_vol/etc/$dumpblock_file";
  
    # Rename the file so it won't get overwritten by the next scramble dumpblocks command.
    my $rand_num = int(rand(2**32));
    if ($dumpblock_file =~ /\.gz$/) {
        $dumpblock_file =~ s/\.gz$/_$rand_num\.gz/;
    }
    else {
        $dumpblock_file .=  "_" . $rand_num;
    }
    my $new_filepath = "/vol/$root_vol/etc/$dumpblock_file";
    (undef, undef, $out) = say($args{fc}, "mv $filepath $new_filepath");
    if ($out){
        confess "Unable to move file '$filepath':\n$out";
    }
   return $dumpblock_file;
}


# This subroutine parses the scramble dumpblocks output file.
# Parameters: fc =>Filer Connectrec object, file=> output file from scramble dumpblocks command, 
#             processor => coderef defining how you want to process the individual lines in the dumpblocks file.
sub parse_dumpblocks_file {
   my %args = validate_mixed(
   @_,
   [],
   {
      fc => {type => OBJECT, optional => 0},
      file => {type => SCALAR, optional => 0},        # example WAFL_dumpblocks.myvol
      processor =>{type => CODEREF, optional => 0},
   }
   );
            
   my $fc = $args{fc};
   my $file = $args{file};
   my $processor = $args{processor};

   my $got_header = 0;
   my @headers = ();
   my $do_zip = 0;

   # Mount the root volume.
   # vol_root needs the full path called out to prevent issues from SanFiler.pm using a different Volume.pm
   my $root_vol = Volume::vol_root($fc);
   say($fc, "exportfs -i -o anon=0,rw /vol/$root_vol");
   my $mountpoint = mount($fc->hostp()->hostname(), "/vol/$root_vol");

   # gunzip the file, if needed.
   if ($file =~ /\.gz$/) {
       my $cmd = "sudo gunzip -v $mountpoint/etc/$file";
       system($cmd) == 0 or confess "system $cmd failed. Exit status returned: $?\n";
       $do_zip = 1;
       $file =~ s/\.gz//;
   }
   
   open(IN, "$mountpoint/etc/$file") or confess "Cannot open $file: $!";
   while(<IN>){
      my $line = $_;
      chomp($line);
      if ($line =~ /^#/){
         # Line is a comment, skip
         next;
      }
      elsif ((! $got_header) && ($line =~ /^[a-z]/)){
          # Line is the header line
          @headers = split(/,/,$line);
          $got_header = 1;
      }
      elsif(($got_header) && ($line =~ /^\w/)){
          my @vals = split(/,/,$line);
          my %row;
          @row{@headers} = @vals; 
          
          my $ret_status = $processor->(row=>\%row);
          if (!$ret_status){
              last;
          }
      }
      else{
          confess "No match: $line";
      }
   }
   close(IN);

   # gzip the file, if needed.
   if ($do_zip) {
       my $cmd = "sudo gzip -v $mountpoint/etc/$file";
       system($cmd) == 0 or confess "system $cmd failed. Exit status returned: $?\n";
   }

   umount($mountpoint);
}


# This subroutine returns the number of large (>2^32-1) & small (< 2^32) vvbns & pvbns from the dumpblock file.
# Parameters: fc =>Filer Connectrec object, dumpblock_file=> output file from scramble dumpblocks command, 
sub count_large_bnos {
    my %args = validate_mixed(
    @_,
    [],
    {
       fc               => {type => OBJECT, optional => 0},
       dumpblock_file   => {type => SCALAR, optional => 0},        # example WAFL_dumpblocks.myvol
    }
    );

    my $fc = $args{fc};
    my $dumpblock_file = $args{dumpblock_file};

    my ($large_vvbns, $small_vvbns, $large_pvbns, $small_pvbns);
    $large_vvbns = $small_vvbns = $large_pvbns = $small_pvbns = 0;

    my $processor = sub {
                            my %params = @_;
                            my $row = $params{row};

                                if ($row->{vvbn} > 2**32 - 1) {
                                    $large_vvbns++;
                                }
                                else {
                                    $small_vvbns++;
                                }
                                if ($row->{pvbn} > 2**32 - 1) {
                                    $large_pvbns++;
                                }
                                else {
                                    $small_pvbns++;
                                }
                            return 1;
                         };

    parse_dumpblocks_file(fc=>$fc, file=>$dumpblock_file, processor=>$processor); 

    my $result = {large_vvbns => $large_vvbns,
                  small_vvbns => $small_vvbns,
                  large_pvbns => $large_pvbns,
                  small_pvbns => $small_pvbns,
                  };
    return $result;
}



1;
