# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Perf/Boottimes.pm#1 $
#
## @author   jroussin@netapp.com
## @summary  
## @status   public
##
## @pod here

=head1 NAME

    Perf::Boottimes - Provides raw output and parsing for "boottimes" filer data. 
                      Works in 7-mode and 10-mode.

=head1 SYNOPSIS

    use Perf::Boottimes;
    my $boottimes_out = boottimes_get($FILER);

    my $cp_start;
    my @cp_durations;
    my $processor = sub {   my $row = shift;
                            if (!$cp_start && $row->{event} eq 'nvlog_replay_sync begin'){
                                $cp_start = $row->{microseconds};
                            }
                            elsif ($cp_start && $row->{event} eq 'nvlog_replay_sync end'){
                                push @cp_durations,  $row->{microseconds} - $cp_start;
                                undef $cp_start;
                            }
                            return 1;
                        };
    
    boottimes_parse($boottimes_out, $processor);
    $Log->comment("CP boottimes: @cp_durations");

=head1 DESCRIPTION

    This package provides mode-agnostic retrieving and parsing of boottimes data.

=head1 METHODS

=cut

use strict;
use warnings;


package Perf::Boottimes;{

    use Carp qw(confess);
    use Tharn qw(host connect say $Log);
    use NATE::Log qw(log_global);
    use Data::Dumper;
    use Filer::Version qw(get_wafl_version version_compare WV_BOILERMAKER_MIN);
    use Exporter 'import';
    our @EXPORT = qw( boottimes_get boottimes_parse );

=head2 boottimes_get

    Retrieves boottimes data

        my $out = boottimes_get($FILER);

        # Also takes a Hostrec object
        my $fh = host($FILER);
        $out = boottimes_get($fh);

    Takes a single argument:

=over 8    

=item string or Hostrec

(Required) : Filer name or Hostrec object

=back

=cut

    sub boottimes_get {
        my $host = shift;
        my $host_ref = ref($host);
        if ($host_ref && !$host_ref->isa('Hostrec')){
            confess "Argument must be a string or a Hostrec, and not a '$host_ref'";
        }
        if (!$host_ref){
            $host = host($host);
        }
        if ($host->hosttype =~ /-ng-/){
            return _boottimes_10m_get($host);
        }
        else{
            return _boottimes_7m_get($host);
        }
    }

=head2 boottimes_parse

    Feeds parsed boottimes data one row at a time to a user-provided 'processor' subroutine.
    Will keep looping down the boottimes data until 'processor' returns 0.

        # Gather timings for first CP
        my $first_cp;
        my $processor = sub {   my $row = shift;
                                if ($row->{event} eq 'nvlog_replay_sync begin'){
                                    $first_cp = $row->{microseconds};
                                }
                                elsif ($row->{event} eq 'nvlog_replay_sync end'){
                                    $first_cp = $row->{microseconds} - $first_cp;
                                    return 0;
                                }
                                return 1;
                            };
        my $out = boottimes_get($FILER);
        boottimes_parse($out, $processor);
        print "First CP duration (us): $first_cp\n";

    Takes two arguments:

=over 8

=item string

(Required) : Boottimes data

=item coderef

(Required) : Processor subroutine

=back

=cut

    sub boottimes_parse {
        my ($output,$processor) = @_;
        confess "Invalid first argument" if (!$output || ref($output));
        if (!$processor){
            confess "Missing second argument";
        }
        my $ref_processor = ref($processor);
        if ($ref_processor ne 'CODE'){
            confess "Processor argument must be a coderef and not a $ref_processor";
        }

        my $log = log_global();

        # Remove all blanks lines
        my @lines = split(/(?:\r*\n)+/, $output);
        # Remove first line
        shift @lines;
        # Second line is header
        my $header = shift @lines;
        # trim the header
        $header =~ s/(^\s+)|(\s+$)//g;

        my @fields = map { lc } split(/\s+/,$header);
        for my $line (@lines){
            # Lines are also space delimited, except for the last column
            $line =~ s/(^\s+)|(\s+$)//g;
            my @splits = split(/\s+/, $line);
            my @values = @splits[0..$#fields - 1];
            if ($line =~ /$values[-1]\s+(\S.*\S)\s*$/){
                push @values, $1;
            }
            else{
                confess "Unable to parse last column in '" . Tharn::_string_cookctl_nl($line) ."'";
            }
            my %row;
            @row{@fields} = @values;
            my $ret_status = $processor->(\%row);
            if (!$ret_status){
                last;
            }
        }
    }

=head2 boottimes_get_parsed

    Returns an array of hashes parsed from the boottimes output. Each element of the array
    is a row from output and each key of the hashes is a column.
    Columns are 'microseconds', 'process', 'event'

        my $parsed_out = boottimes_get_parsed($FILER);
        # Find all unique process names
        use List::MoreUtils qw( uniq );
        my @process_names = uniq map { $_->{process} } @$parsed_out;

    Takes a single argument:

=over 8

=item string or Hostrec

(Required) : Filer name or Hostrec object

=back

=cut

    sub boottimes_get_parsed {
        my $out = boottimes_get(shift);
        my $out_parsed = [];
        my $processor = sub { my $row = shift;
                              push @$out_parsed, $row;
                            };
        boottimes_parse($out,$processor);
        return $out_parsed;
    }

    sub _boottimes_10m_get {
        my $host = shift;
        my $server = Server->new();
        $server->add_nodes( nodes=>[$host->hostname()] );
        $server->set_execution_server( -Server=>$host->hostname() );
        my $out = $server->shell( cmd=>"sysctl sysvar.boottimes" );
        if (!$out || $out !~ /sysvar\.boottimes\.boottimes/){
            confess "Unable to get boottimes: $out";
        }
        return $out;
    }

    sub _boottimes_7m_get {
        my $host = shift;
        my $wafl_version = get_wafl_version($host);
        if ($wafl_version < WV_BOILERMAKER_MIN){
            return _boottimes_prebr_get($host);
        }
        host($host->hostname);
        my $fc = connect("#rsh");

        my (undef, undef, $out) = say($fc, "priv set -q test; mhostexec sysctl sysvar.boottimes");
        if (!$out || $out !~ /sysvar\.boottimes\.boottimes/){
            confess "Unable to get boottimes: $out";
        }
        return $out;
    }

    sub _boottimes_prebr_get {
        my $host = shift;
        confess "not implemented yet";
    }

}
1;
