# -*- perl -*-
# $Id: //depot/prod/test/main/ontap/boilermaker/lib/Perf/DSS.pm#4 $
# Copyright (c) 2008-2009 NetApp, Inc.
# All rights reserved.
# Perf/DSS.pm
#
## @summary  Perl library for DSS routines.
## @author   slane@netapp.com
## @status   public
## example
## my $ext = new Perf::DSS(fc=>$fc);
## $ext->enable();
##

package Perf::DSS;

use strict;
use warnings;  
use NOTE;
use Tharn;
use Carp;
use filer;
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Module::Common qw(validate_mixed);
use Filer::Version;
use Perf::Block_Layout;
use Perf::Utils qw(as_string in_blocks);

# list routines from this file	
our @EXPORT = qw(new calc_length get_extent_state set_extent_state);

# If a filer is running a release > MAJOR_RELEASE, then set the max-write-alloc-blocks vol option is available
use constant VOL_OPT_RELEASE => 8.0; # BR.0
#use constant EXTENT_SIZE => 4;

my $init = 0;
BEGIN {
    $DSS::init = $$;
}

sub new {
    my $class = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {
                fc => {type => OBJECT, optional => 0},
                cc => {type => OBJECT, optional => 0},
                test => {type => OBJECT, optional => 0},
                vol => {type => OBJECT, optional => 1},
                aggr => {type => OBJECT | UNDEF, optional => 1},
                log_dir => {type => SCALAR | UNDEF, optional => 1},			 
                mountpoint => {type => SCALAR | UNDEF, optional => 1},			 
                workload_done => {type => SCALAR | UNDEF, optional => 1, default => "FALSE"},			 
                num_files => {type => SCALAR | UNDEF, optional => 1},			 
                snapshot => {type => SCALAR | UNDEF, optional => 1},			 
                ds => {type => OBJECT | UNDEF, optional => 1, default => 'off'},
                delta => {type => OBJECT | UNDEF, optional => 1, default => 'off'},

                # determines whether pre-BR.0 or not
                use_vol_options => {type => SCALAR | UNDEF, optional => 1, default => "FALSE"}, 
                
                wafl_max_write_alloc_blocks => {type => SCALAR | UNDEF, optional => 1},
                wafl_read_ahead_total_hint => {type => SCALAR | UNDEF, optional => 1},
                max_write_alloc_blocks_vol_option => {type => SCALAR | UNDEF, optional => 1},

                san_iscsi => {type => SCALAR | UNDEF, optional => 1},
                datasize => {type => SCALAR | UNDEF, optional => 1},
                
                histo => {type => HASHREF | UNDEF, optional => 1}
           }
           );
    my $self = \%params;     

    $self->{dataset_snap} = undef;
    $self->{file_layouts} = {};

	my $ffc = $self->{fc};

   	# Get the ONTAP version name and release 
    $self->{use_vol_options} = "TRUE";
   	
  	# This method looks for the release version number directly	
    my $filer  = $ffc->hostp();
    $self->{version} = Filer::Version::get_ontap_version($filer);
    logcomment("\$version is $self->{version}\n");   	
  	
  	# If the filer is runing 8.0 (BR.0) or later, then enable vol options 
  	
  	$self->{use_vol_options} = "FALSE";
  	if ($self->{version} >= VOL_OPT_RELEASE) {
  		$self->{use_vol_options} = "TRUE";
  	}  
 	
 	logcomment("use_vol_options is $self->{use_vol_options}\n");		
	
    bless $self, $class;
         
    return $self;
}
	

# calc_length
# This provides a wrapper around the parse_dumpblocks call in Block_Layout.pm
sub calc_length {
	my $dss = shift;
	my $dumpblock_file = shift;
	my $write_blocks = shift;
	
    my $fc = $dss->{fc};
    my $tcd = $dss->{test};
    
    my @max_lengths = ();
    my $current_length = 0;
    my $last_dbn = 1;
    my $last_disk = 0;
    
    my $processor = sub { 
        my %args = @_;
        my $row = $args{row};
        
        if ($row->{level} == 0) {
            if (($row->{disk} == $last_disk) && ($row->{dbn} == ($last_dbn + 1))) {
                $current_length++;
            } else {
                push(@max_lengths, $current_length);
                $current_length = 1;  # An isolated pbn has a length of 1
            }
            $last_disk = $row->{disk};
            $last_dbn = $row->{dbn};
        } else {
            $current_length = 1;  # Found a non-L0 block
        }        
        return 1;
    };

    Perf::Block_Layout::parse_dumpblocks_file(fc=>$fc, file=>$dumpblock_file, processor=>$processor); 
    
    # Create a histogram of the collected data
    logcomment("write_blocks is $write_blocks\n");
    logcomment("layout of disk is a follows (each number represents the number of consecutive blocks on a spindle)\n@max_lengths\n");
    my %max_histo = ();
    $max_histo{$write_blocks} = 0;    # Initialize this data point with 0
    my $num_data_points = 0;
    foreach (@max_lengths) {
        if (exists $max_histo{$_}) {
            $max_histo{$_}++;   
        } else {
            $max_histo{$_} = 1;   
        }
        $num_data_points++;
    }

    my $num_large_writes = 0;
    my $largest_write = 0;
    my $small_chains = 0;
    my $large_chains = 0;
    foreach (sort keys %max_histo) {
#        logcomment("$_ is $max_histo{$_}\n"); 
        if ($_ > $write_blocks) {
            $num_large_writes += $max_histo{$_};
        }

        $largest_write = $_ > $largest_write ? $_ : $largest_write;
        
        if ($_ < 32) {
            $small_chains += $max_histo{$_};
        } elsif (($_ >= 32) && ($_ <= $write_blocks)) {
            $large_chains += $max_histo{$_};
        }
    }
    
    my $num_max_writes = $max_histo{$write_blocks};    
    
    my $ratio = 0;
    if ($num_data_points != 0) {
        $ratio = $num_large_writes/$num_data_points; 
    }
    my $result = {
        num_max_writes => $num_max_writes,
        num_large_writes => $num_large_writes,
        largest_write => $largest_write,
        num_data_points => $num_data_points,
        ratio => $ratio,
        small_chains => $small_chains,
        large_chains => $large_chains,
        histo => \%max_histo,
    };

    return $result;
}


sub DESTROY {
   my $self = shift;

   logcomment("DESTROY sequence initiated on DSS because it fell out of scope.");

   logcomment("Debug: File layout information for each file:", as_string($self->{file_layouts}));

#   umount($self->{mountpoint}, client=>$self->{cc}, force=>1); 
}

1;