# $Id: //depot/prod/test_main/lib/Perf/DWIM.pm#1 $
#
# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.
#
# Perf/DWIM/DWIM_server.pm
#
## @summary Sets up and runs DWIM. It will create the dwim server
## and dwim clients.
##
## @description
##   This library does setup and runs dwim.
##   Summary of all dwim workloads is available upon completion.
## @synopsis
##   use Perf/DWIM.pm
##
#
##
## @keywords   dwim
## @status     public
## @see-also   lib/TCD.pm
## @author     heinze@netapp.com

package Perf::DWIM;


use strict;
use warnings;

use Carp qw(confess);
use Params::Validate qw(validate validate_with SCALAR HASHREF ARRAYREF OBJECT BOOLEAN);
use Perf::DWIM::Server;
use Perf::DWIM::Client;
use POSIX qw(floor ceil);
use Tharn;


## @pod here
=head1 NAME

Perf::DWIM - Run dwim on clients and files specified.

=head1 DESCRIPTION

This package allows you to run the dwim IO tool on all the clients and
files that you specify. It will create the DWIM server and clients to
run the workloads you want and the summary is then given back when stopped.
The summary sums up all dwim clients' data.

Every client will do IO on every file.

See http://wikid.netapp.com/w/Dwim and
http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/Tools/IOTools/Dwim/index.html

=head1 SYNOPSIS

    use Perf::DWIM;

    Example:
        $dwim = Perf::DWIM->new(
            clients => $physical_clients,
            files   => $files,
        );
        $dwim->start(
            op_rate  => 500,
            rand_pct => 80,
            read_pct => 50,
        );
        $dwim->modify(
            op_rate => 200
        );
        $dwim->stop;

        # Collects data from all dwim clients and sums it together
        my $summary_ref = $dwim->summary();



=head1 METHODS

=cut


=head2 new

Returns a dwim object.

=over

=item Arguments

B<clients> = arrayref

I<Required> The physical clients used to run the dwim workload on

B<files> = arrayref

I<Required> The files that dwim will run on, must be the full mounted path

B<id> = scalar

I<Optional> The unique identifier for the DWIM processes.


=back

=cut

sub new {
    my $class = shift;
    my %args = validate(
        @_,
        {
            clients => {type => ARRAYREF, optional => 0},
            files   => {type => ARRAYREF, optional => 0},
            id      => {type => SCALAR, optional => 1},
        }
    );

    my $self = \%args;

    bless $self, $class;

    return $self;
}



=head2 start

Starts the dwim processes. The dwim clients will be started individually.
Call 'stop' to stop the workloads and to get the summary of all the clients.

=over

=item Arguments

B<alignment> = string

I<Optional>  Issue IOs that are misaligned by this many 512byte sectors.
Range: 0-7. Dwim defaults to perfectly aligned.

B<blk_size> = string

I<Optional> The size of IO done. Dwim defaults to '4k'.

B<burst> = string

I<Optional> amplitude[,duration[,period]]: burst amplitude (%),
duration (ms), period (ms). Dwim defaults to non-burty workload.

B<exp_interarrival> = boolean

I<Optional> Dwim will issue IOs at an exponentially distributed rate rather
than fixed intervals, simulating a Poisson process. 0 or 1. Dwim defaults to 0.

B<op_rate> = integer

I<Optional> The number of IOPs. Dwim defaults to 100.

B<port> = number

I<Optional> The port used for distributed dwim, Dwim default to 6336.

B<queue_depth> = number

I<Optional> The number of IOs in flight. Dwim defaults to 256.

B<rand_pct> = number

I<Optional> The percent of random IO. Dwim defaults to 100.

B<read_pct> = number

I<Optional> The read/write percentage. Dwim defaults to 100, all reads.

B<run_time> = number

I<Optional> The length of time dwim will run. Dwim defaults to infinite,
must be stopped manually.

B<verbose> = boolean

I<Optional> Verbose mode. Dwim defaults to 0.

B<wait_time> = number

I<Optional> Dwim will run for this amount of time at a gradually increasing
throughput from 1/3 to full requested throughput before measurement interval.
Dwim defaults to 0.

=back

=cut

sub start {
    my $self = shift;
    my %params = validate_with(
        params => \@_,
        spec => {
            op_rate => {type => SCALAR, regex => qr/^\d+$/,
                optional => 1, default => 100},
        },
        allow_extra => 1,
    );

    my $num_clients = scalar(@{$self->{clients}});
    my $num_files = scalar(@{$self->{files}});

    my $dwim_server_physical_client = NACL::C::Client->new(
        name => $self->{clients}->[0]
    );

    $self->{num_dwim_clients} = $num_clients * $num_files;

    $Log->comment('Start dwim server ' . $self->{id});
    my $dwim_server = Perf::DWIM::Server->new(
        server      => $self->{clients}->[0],
        num_clients => $self->{num_dwim_clients},
        id          => $self->{id},
    );
    $dwim_server->start;

    $self->{dwim_server} = $dwim_server;
    $self->{id} = $dwim_server->get_id();
    # Let DWIM server really come up and be ready before DWIM clients start.
    Tharn::snooze(1);

    my $total_op_rate = delete $params{op_rate};
    $Log->comment("Starting dwim with total op_rate of $total_op_rate");
    my $op_rates = $self->calculate_op_rate(op_rate => $total_op_rate);

    my $queue_depth = ceil(($params{queue_depth} || 256) / $num_files);
    $Log->comment("Queue depth per dwim process set to $queue_depth");

    my @dwim_clients;
    foreach my $dwim_client (@{$self->{clients}}) {
        foreach my $i (0..($num_files - 1)) {
            my $file = $self->{files}->[$i];
            $Log->comment("Start dwim client on $dwim_client for $file");
            my $dwim_client = Perf::DWIM::Client->new(
                client      => $dwim_client,
                server_ip
                    => $dwim_server_physical_client->hostrec->{default_ip},
                port        => $dwim_server->port(),
                id          => $self->{id} . '_' . ($i + 1),
            );
            $dwim_client->start(
                file    => $file,
                op_rate => $op_rates->[$i],
                %params,
            );
            push @dwim_clients, $dwim_client;
        }
    }

    $self->{dwim_clients} = \@dwim_clients;

    return 1;
}


=head2 modify

Adjusts the rates of the running dwim workloads.

=over

=item Arguments

B<op_rate> = number

I<Required> The new value for the total number of IOPs across all clients
and files.

=back

=cut

sub modify {
    my $self = shift;
    my %args = validate(
        @_,
        {
            op_rate    => {type => SCALAR, optional => 0},
        }
    );

    $Log->comment("Modifying dwim clients to total op_rate of $args{op_rate}");
    my $op_rates = $self->calculate_op_rate(op_rate => $args{op_rate});
    my $i = 0;
    foreach my $client (@{$self->{dwim_clients}}) {
        $Log->comment('Modify dwim client on ' . $client->get_id()
            . ' to ' . $op_rates->[$i]);
        $client->modify(
            op_rate => $op_rates->[$i],
        );
        $i++;
    }

    return 1;
}


=head2 stop

Stops the running the dwim workload on all clients.

=over

=item Arguments

B<full_client_output> = boolean

I<Optional> The dwim client logs will only have the output up until the
last start or modify issued. If the full client output per second is wanted
to be logged, setting this will issue a modify to set the op_rate to 0 and
then immediately stop afterwards. Depending on response time, this could
cause the op_rate to taper off before it stops completely. Default is 0 (off).


=back

=cut

sub stop {
    my $self = shift;
    my %args = validate(
        @_,
        {
            full_client_output => {type => BOOLEAN, optional => 1, default => 0},
        }
    );

    if ($args{full_client_output} || $Log->may_debug()) {
        $Log->debug('Modifying all clients to op_rate of 0 to capture all '
            . 'client output before the dwim server stops.');
        foreach my $client (@{$self->{dwim_clients}}) {
            $client->modify(
                op_rate => 0,
            );
        }
    }

    $self->{dwim_server}->stop;

    foreach my $dwim_client (@{$self->{dwim_clients}}) {
        # DWIM clients have probably already stopped so this will error
        # however the STAF processes still need to be cleaned up.
        eval {$dwim_client->stop();};
    }


    return 1;
}


=head2 summary

Returns a hashref with the summary of the work done.
Summary is only available after dwim has stopped.

=over

=item Arguments

None

=back

=cut

sub summary {
    my $self = shift;

    return $self->{dwim_server}->summary();
}


=head2 calculate_op_rate

Returns an arrayref of op_rates for the dwim clients that adds up
to the total op_rate

=over

=item Arguments

B<op_rate> = number

I<Required> The new value for the number of IOPs.

=back

=cut

sub calculate_op_rate {
    my $self = shift;
    my %args = validate(
        @_,
        {
            op_rate    => {type => SCALAR, optional => 0},
        }
    );

    # op_rates will be determined to match the overall op_rate exactly
    # It will be divided evenly among clients and files except the last one
    # which might be slightly more to round up to the total op_rate.

    my $num_files = scalar(@{$self->{files}});
    my $total_op_rate = $args{op_rate};
    my @op_rates;

    my $total_dwim_processes = $self->{num_dwim_clients};
    my $op_rate_1 = floor($total_op_rate / $total_dwim_processes);
    my $op_rate_final =
        $total_op_rate - ($op_rate_1 * ($total_dwim_processes - 1));

    foreach (1..($total_dwim_processes - 1)) {
        push @op_rates, $op_rate_1;
    }
    push @op_rates, $op_rate_final;
    $Log->comment("Op rates will be @op_rates");

    return \@op_rates;
}

1;
