# $Id: //depot/prod/test_main/lib/Perf/DWIM/Client.pm#1 $
#
# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.
#
# Perf/DWIM/Client.pm
#
## @summary Sets up and runs a DWIM client for dwim workloads.
##
## @description
##   This library does setup and runs the dwim client. A dwim server must be
##   setup before the client if used in a server<->client model.
##   This can be used singularly without a DWIM server.
##   Summary of the dwim workload is given to the dwim server or available
##   for this specific client upon completion of the workload.
## @synopsis
##   use Perf/DWIM/Client.pm
##
#
##
## @keywords   wafl qos
## @status     public
## @see-also   lib/TCD.pm
## @author     heinze@netapp.com

package Perf::DWIM::Client;


use strict;
use warnings;

use Data::Dumper qw(Dumper);
use Data::Random qw(rand_words);
use NATE::Events qw{eventloop_drain};
use NATE::Process;
use NATE::STAF::Peer;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT);
use Tharn;


## @pod here
=head1 NAME

Perf::DWIM::Client - DWIM client

=head1 DESCRIPTION

Most users should use Perf::DWIM instead of this library.

This package allows you to create a dwim client. This connects to the dwim
server that handles many clients. Upon completion, the dwim server will
have a summary output from all dwim clients.

See http://wikid.netapp.com/w/Dwim and
http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/Tools/IOTools/Dwim/index.html

=head1 SYNOPSIS

    use Perf::DWIM::Client;

    Example:
        my $dwim_client = Perf::DWIM::Client->new(
            client    => $physical_client,
            server_ip => $ip_addr,
        );
        $dwim_client->start(
            file     => $file,
            op_rate  => '100',
            read_pct => 80,
            rand_pct => 100,
            blk_size => '4k'
        );

        $dwim_client->modify(op_rate => '50');

        # If used with a server, stopping 1 client stops all clients and server
        $dwim_client->stop();

        # Collects summary of this client. If using dwim server, collect
        # the summary from it for summation of all clients.
        my $summary_ref = $dwim_client->summary();



=head1 METHODS

=cut


=head2 new

Returns a dwim client object.

=over

=item Arguments

B<client> = string

I<Required> The client to run the dwim client on

B<port> = string

I<Optional> The port of the dwim server.

B<id> = scalar

I<Optional> The unique ID for the DWIM client process.

B<server_ip> = string

I<Optional> The IP address of the dwim server. The server and client
can be on the same box.


=back

=cut

sub new {
    my $class = shift;
    my %args = validate(
        @_,
        {
            client      => {type => SCALAR, optional => 0},
            id          => {type => SCALAR, optional => 1},
            port        => {type => SCALAR, optional => 1},
                # default is 6336
            server_ip   => {type => SCALAR, optional => 1},
                # default is to run without a server
        }
    );

    my $self = \%args;

    bless $self, $class;

    return $self;
}



=head2 start

Starts the dwim client. The dwim client will connect to the dwim server.

=over

=item Arguments

B<file> = string

I<Required> The file dwim will use.

B<alignment> = string

I<Optional>  Issue IOs that are misaligned by this many 512byte sectors.
Range: 0-7

B<blk_size> = string

I<Optional> The size of IO done. Defaults to '4k'.

B<burst> = string

I<Optional> amplitude[,duration[,period]]: burst amplitude (%),
duration (ms), period (ms)

B<exp_interarrival> = boolean

I<Optional> Dwim will issue IOs at an exponentially distributed rate rather
than fixed intervals, simulating a Poisson process. 0 or 1.

B<op_rate> = integer

I<Optional> The number of IOPs. Defaults to 100.

B<port> = number

I<Optional> The port used for distributed dwim, Default to 6336.

B<queue_depth> = number

I<Optional> The number of IOs in flight. Defaults to 256.

B<rand_pct> = number

I<Optional> The percent of random IO. Defaults to 100.

B<read_pct> = number

I<Optional> The read/write percentage. Defaults to 100, all reads.

B<run_time> = number

I<Optional> The length of time dwim will run. Defaults to infinite,
must be stopped manually.

B<verbose> = boolean

I<Optional> Verbose mode.

B<wait_time> = number

I<Optional> Dwim will run for this amount of time at a gradually increasing
throughput from 1/3 to full requested throughput before measurement interval.

=back

=cut

sub start {
    my $self = shift;

    my %args = validate(
        @_,
        {
            files            => {type => ARRAYREF, optional => 0},
            alignment        => {type => SCALAR, optional => 1},
            blk_size         => {type => SCALAR, optional => 1},
            burst            => {type => SCALAR, optional => 1},
            exp_interarrival => {type => SCALAR, optional => 1},
            op_rate          => {type => SCALAR, regex => qr/^\d+$/,
                optional => 1},
            port             => {type => SCALAR, optional => 1},
            queue_depth      => {type => SCALAR, optional => 1, default => 256},
            rand_pct         => {type => SCALAR, optional => 1},
            read_pct         => {type => SCALAR, optional => 1},
            run_time         => {type => SCALAR, optional => 1},
            verbose          => {type => SCALAR, optional => 1},
            wait_time        => {type => SCALAR, optional => 1},
        }
    );

    # Each dwim client needs a unique ID (random word) so all messages
    # will carry that unique ID. This allows more than 1 dwim client process
    # per physical client. But still only 1 dwim server per physical client.
    # Normally uses the dwim server unique ID + client_#.
    if (!defined $self->{id}) {
        my ($rand_word) = rand_words();
        # Remove non-word characters, since it would not be a valid name.
        $rand_word =~ s/\W/_/g;
        $self->{id} = $rand_word;
    }
    $args{id} = $self->{id};
    $args{server_addr} = $self->{server_ip};
    $args{port} = $self->{port};
    local $Data::Dumper::Terse  = 1;
    my $dwim_client_code = sprintf(<<'EOT', Dumper(\%args));
        {
            my $args = %s;
            my $id = delete $args->{id};
            my $files = delete $args->{files};

            use warnings;
            use strict;

            use Carp qw(confess);
            use Data::Dumper qw(Dumper);
            use File::Spec;
            use IPC::Run qw(run start pump finish timeout);
            use NATE::Events qw{eventloop_drain};
            use NATE::ParamSet;
            use NATE::STAF::Peer qw($Parent);

            my $params = {
                alignment           => ['-a', $args->{alignment}],
                blk_size            => ['-s', $args->{blk_size}],
                burst               => ['-b', $args->{burst}],
                exp_interarrival    => ['-e'],
                op_rate             => ['-i', $args->{op_rate}],
                port                => ['-p', $args->{port}],
                queue_depth         => ['-q', $args->{queue_depth}],
                read_pct            => ['-r', $args->{read_pct}],
                rand_pct            => ['-U', $args->{rand_pct}],
                run_time            => ['-t', $args->{run_time}],
                server_addr         => ['-C', $args->{server_addr}],
                verbose             => ['-v'],
                wait_time           => ['-w', $args->{wait_time}],
            };

            my $nate_params = NATE::ParamSet->new(global => 1);
            my $cmd_path =
                File::Spec->catfile($nate_params->get('TESTBINDIR'), 'dwim');
            my @cmd = ($cmd_path);
            foreach my $key (%%{$args}) {
                if (defined $args->{$key}) {
                    push @cmd, @{$params->{$key}};
                }
            }

            if ($args->{server_addr}) {
                $Log->comment('Running as a client to server '
                    . $args->{server_addr});
            }
            if ($args->{port} && !defined $args->{server_addr}) {
                confess 'Cannot specify port unless a server_addr given';
            }

            push(@cmd, @$files);

            my $in = "\n";
            my $out;
            my $error;

            $Log->comment("Starting dwim with this command: @cmd");
            my $harness = start(\@cmd, \$in, \$out, \$error);
            pump $harness until $out =~ /Time/mgc;

            # Setup to receive messages from STAF
            my $dwim_finished = 0;
            while(!$dwim_finished) {
                $Parent->message_get(
                    type        => "$id" . 'ModifyIO',
                    callback    => sub {
                        $in .= shift;
                        $Log->comment("Received Modify command: $in");
                        pump $harness until $out =~ /Set target_iops|Paused/mgc;
                        $Log->comment("DWIM output at modify: $out\n\n");
                    }
                );
                $Parent->message_get(
                    type        => "$id" . 'StopIO',
                    callback    => sub {
                        $in .= shift;
                        $Log->comment("Received Stop command: $in");
                        pump $harness until
                            $out =~ /Dwim workload complete!/mgc;
                        $dwim_finished = 1;
                    }
                );
                $Parent->message_get(
                    type        => "$id" . 'Finish',
                    callback    => sub {
                        $in .= shift;
                        $Log->comment('Received finish');
                        $dwim_finished = 1;
                    }
                );
                eventloop_drain;
                if ($error) {
                    $Log->error("DWIM Client error: " . Dumper($error));
                }
            }

            finish $harness;

            my %%summary;
            $summary{dwim_client_id} = $id;
            ($summary{run_time}) = $out =~ /Run time \(s\): (\d+.?\d+)/;
            ($summary{avg_throughput_iops}) =
                $out =~ /Avg. TPUT \(iops\): (\d+.?\d+)/;
            ($summary{avg_throughput_kbps}) =
                $out =~ /Avg. TPUT \(kbps\): (\d+.?\d+)/;
            ($summary{avg_rt}) = $out =~ /Avg. RT.*?: (\d+.?\d+)/;
            ($summary{max_rt}) = $out =~ /Max. RT.*?: (\d+.?\d+)/;
            ($summary{avg_low_latency}) =
                $out =~ /Average latency of .*?: (\d+.?\d+)ms/;
            ($summary{avg_qdepth}) = $out =~ /Avg qdepth : (\d+.?\d+)/;
            ($summary{max_qdepth}) = $out =~ /Max qdepth : (\d+)/;
            ($summary{target_achieved}) =
                $out =~ /Percent target throughput achieved: (\d+.?\d+)/;

            $Log->comment("\nComplete dwim output: $out\n");

            $Parent->message_put(
                type    => $id . 'Summary',
                message => \%%summary,
            );
        }
EOT


    $self->{run_id} = 'dwim_client_' . $self->{id};
    my $process = NATE::Process->new(
        codespec        => $dwim_client_code,
        runid           => $self->{run_id},
        inherit_env     => 1,
        inherit_params  => 1,
        client          => $self->{client},
    );
    $process->start();
    $self->{handle} = $process->handle();

    $self->{staf_peer} = NATE::STAF::Peer->new(
        handle      => $self->{handle},
        endpoint    => $self->{client},
    );

    $self->{process} = $process;

    return $process;
}


=head2 modify

Adjusts the rate of the running dwim workload.

=over

=item Arguments

B<op_rate> = number

I<Required> The new value for the number of IOPs.

=back

=cut

sub modify {
    my $self = shift;
    my %args = validate(
        @_,
        {
            op_rate    => {type => SCALAR, optional => 0},
        }
    );

    $Log->comment('Modify dwim client to ' . $args{op_rate});

    $self->{staf_peer}->message_put(
        type    => $self->{id} . 'ModifyIO',
        message => "-i $args{op_rate}\n",
    );

    return 1;
}

=head2 stop

Stops the running the dwim workload on all clients and gives summary.

=over

=item Arguments

None

=back

=cut

sub stop {
    my $self = shift;

    $Log->comment('Stopping dwim client ' . $self->{id});

    $self->{staf_peer}->message_put(
        type    => $self->{id} . 'StopIO',
        message => "-done\n",
    );
    my $summary;
    my $wait_time = time + 30;
    while (!defined $summary && time < $wait_time) {
        $self->{staf_peer}->message_get(
            type        => $self->{id} . 'Summary',
            callback    => sub {
                $summary = shift;
            }
        );
        eventloop_drain;
    }
    $self->{client_summary} = $summary;

    $self->{process}->stop();
    $self->{process}->destroy();

    return 1;
}


=head2 finish

Finishes the dwim workload. Use to collect data from clients when the
server stops and therefore already stops the dwim clients automatically.
This allows you to capture the dwim client summary output.

=over

=item Arguments

None

=back

=cut

sub finish {
    my $self = shift;

    # This method was created to try and capture the output on the client
    # after the server stopped, however in my testing I found that most of
    # the time it did not catch the summary output. Generally looping through
    # all of the clients resulted in only capturing the summary output from
    # the first client's output.
    # This method might never really be used, but keeping in case we need to
    # add this functionality in later.

    $self->{staf_peer}->message_put(
        type    => $self->{id} . 'Finish',
        message => 'finish',
    );
    my $summary;
    my $wait_time = time + 30;
    while (!defined $summary && time < $wait_time) {
        $self->{staf_peer}->message_get(
            type        => $self->{id} . 'Summary',
            callback    => sub {
                $summary .= shift;
            }
        );
        eventloop_drain;
    }
    my @temp_summary = split(m/,/, $summary);
    my %dwim_summary = @temp_summary;
    $self->{client_summary} = \%dwim_summary;

    $self->{process}->stop();
    $self->{process}->destroy();

    return 1;
}


=head2 summary

Returns a hashref with the summary of the work done.
Summary is only available after dwim has stopped.

=over

=item Arguments

None

=back

=cut

sub summary {
    my $self = shift;

    return $self->{client_summary};
}


=head2 get_id

Returns the id of the server process.

=over

=item Arguments

None

=back

=cut

sub get_id {
    my $self = shift;

    return $self->{id};
}



1;
