# $Id: //depot/prod/test_main/lib/Perf/DWIM/Server.pm#1 $
#
# Copyright (c) 2012 Network Appliance, Inc.
# All rights reserved.
#
# Perf/DWIM/Server.pm
#
## @summary Sets up and runs a DWIM server for dwim workloads.
##
## @description
##   This library does setup and runs a dwim server.
##   Summary of all dwim workloads is available upon completion.
## @synopsis
##   use Perf/DWIM/Server.pm
##
#
##
## @keywords   wafl qos
## @status     public
## @see-also   lib/TCD.pm
## @author     heinze@netapp.com

package Perf::DWIM::Server;


use strict;
use warnings;

use Carp qw(confess);
use Data::Dumper qw(Dumper);
use Data::Random qw(rand_words);
use NATE::Events qw{eventloop_drain};
use NATE::Process;
use NATE::STAF::Peer;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF);
use Tharn;


## @pod here
=head1 NAME

Perf::DWIM::Server - DWIM server for using dwim clients

=head1 DESCRIPTION

Most users should use Perf::DWIM instead of this library.

This package allows you to create a dwim server. This allows multiple
dwim clients to run and the summary is then given back to the server.
The summary sums up all dwim clients' data.

See http://wikid.netapp.com/w/Dwim and
http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/Tools/IOTools/Dwim/index.html

=head1 SYNOPSIS

    use Perf::DWIM::Server;

    Example:
        $server = Perf::DWIM::Server->new(
            server      => $physical_client,
            num_clients => $num_clients,
            port        => $server_port,
        );
        $dwim_server->start();

        $client1 = Perf::DWIM::Client->new(
            file      => $file1,
            read_pct  => 50,
            rand_pct  => 50,
            server_ip => $dwim_server_ip_addr,
            port      => $server_port,
        );
        $client2 = Perf::DWIM::Client->new(
            file      => $file2,
            read_pct  => 20,
            rand_pct  => 100,
            server_ip => $dwim_server_ip_addr,
            port      => $server_port,
        );
        $client1->start();
        $client2->start();

        $client1->modify(op_rate => 50);
        $client2->modify(op_rate => 50);

        # Stop the server which also stops all clients
        $dwim_server->stop();

        # Collects data from all clients and sums it together
        my $summary_ref = $dwim_server->summary();



=head1 METHODS

=cut


=head2 new

Returns a dwim server object. There can be only 1 dwim server
running on a physical client

=over

=item Arguments

B<server> = string

I<Required> The hostname to run the dwim server on

B<num_clients> = string

I<Required> The number of clients that dwim will run on

B<id> = scalar

I<Optional> The unique ID for the DWIM server process.

B<port> = number

I<Optional> The port used for distributed dwim, Default = 6336


=back

=cut

sub new {
    my $class = shift;
    my %args = validate(
        @_,
        {
            server      => {type => SCALAR, optional => 0},
            num_clients => {type => SCALAR, optional => 0},
            id          => {type => SCALAR|UNDEF, optional => 1},
            port        => {type => SCALAR, optional => 1},
        }
    );

    my $self = \%args;

    if (!defined $self->{id}) {
        my ($rand_word) = rand_words();
        # Remove non-word characters, since it would not be a valid name.
        $rand_word =~ s/\W/_/g;
        $self->{id} = $rand_word;
    }

    if ($self->{num_clients} < 1) {
        confess 'Must have at least 1 client';
    }

    bless $self, $class;

    return $self;
}



=head2 start

Starts the dwim server. All dwim clients will connect to this server.

The dwim clients will be started individually. Call 'stop' to stop the workload
and to get the summary of all clients.


=over

=item Arguments

=back

=cut

sub start {
    my $self = shift;

    my %args;
    $args{id} = $self->{id};
    $args{num_clients} = $self->{num_clients};

    $args{port} = $self->find_open_port;
    $self->{port} = $args{port};

    local $Data::Dumper::Terse  = 1;
    my $dwim_server_code = sprintf(<<'EOT', Dumper(\%args));
        {
            my $args = %s;
            my $id = $args->{id};
            my $num_clients = $args->{num_clients};
            my $port = $args->{port};

            use Data::Dumper qw(Dumper);
            use File::Spec;
            use IPC::Run qw(run start pump finish timeout);
            use NATE::Events qw{eventloop_drain};
            use NATE::ParamSet;
            use NATE::STAF::Peer qw($Parent);

            $Log->comment('DWIM server starting, '
                . "waiting for $num_clients clients");

            my $nate_params = NATE::ParamSet->new(global => 1);
            my $cmd_path =
                File::Spec->catfile($nate_params->get('TESTBINDIR'), 'dwim');
            my @cmd = ($cmd_path, '-S', $num_clients);
            if ($port) {
                push(@cmd, '-p', $port);
            }

            my $in = "\n";
            my $out;
            my $error;

            $Log->comment("Starting dwim with command: @cmd");
            my $harness = start(\@cmd, \$in, \$out, \$error);
            pump $harness until $out =~ /waiting for/mgc;

            my $dwim_finished;
            # Setup to receive messages from STAF
            while(!$dwim_finished) {
                $Parent->message_get(
                    type        => "$id" . 'ServerStopIO',
                    callback    => sub {
                        $in .= shift;
                        pump $harness until
                            $out =~ /Dwim workload complete!/mgc;
                        $dwim_finished = 1;
                    }
                );
                eventloop_drain;
                if ($error) {
                    $Log->error("DWIM Server error: " . Dumper($error));
                }
            }
            finish $harness;

            my %%summary;
            $summary{dwim_server_id} = $id;
            ($summary{run_time}) =
                $out =~ /TOTAL.*Run time \(s\): (\d+.?\d+)/s;
            ($summary{avg_throughput_iops}) =
                $out =~ /TOTAL.*Avg. TPUT \(iops\): (\d+.?\d+)/s;
            ($summary{avg_throughput_kbps}) =
                $out =~ /TOTAL.*Avg. TPUT \(kbps\): (\d+.?\d+)/s;
            ($summary{avg_rt}) =
                $out =~ /TOTAL.*Avg. RT.*?: (\d+.?\d+)/s;
            ($summary{max_rt}) =
                $out =~ /TOTAL.*Max. RT.*?: (\d+.?\d+)/s;
            ($summary{avg_low_latency}) =
                $out =~ /TOTAL.*Average latency of .*?: (\d+.?\d+)ms/s;
            ($summary{avg_qdepth}) =
                $out =~ /TOTAL.*Avg qdepth : (\d+.?\d+)/s;
            ($summary{max_qdepth}) =
                $out =~ /TOTAL.*Max qdepth : (\d+)/s;
            ($summary{target_achieved}) =
                $out =~ /TOTAL.*throughput achieved: (\d+.?\d+)/s;

            $Parent->message_put(
                type    => $id . 'ServerSummary',
                message => \%%summary,
            );
            $Log->comment("\nComplete dwim output: $out\n");
        }
EOT


    $self->{run_id} = 'dwim_server_' . $self->{id};
    my $process = NATE::Process->new(
        codespec        => $dwim_server_code,
        runid           => $self->{run_id},
        inherit_env     => 1,
        inherit_params  => 1,
        client          => $self->{server},
    );
    $process->start();
    $self->{handle} = $process->handle();

    $self->{staf_peer} = NATE::STAF::Peer->new(
        handle      => $self->{handle},
        endpoint    => $self->{server},
    );

    $self->{process} = $process;

    return $process;
}


=head2 stop

Stops the running the dwim workload on all clients and gives summary.

=over

=item Arguments

None

=back

=cut

sub stop {
    my $self = shift;

    $Log->comment('Stopping dwim server ' . $self->{id});

    $self->{staf_peer}->message_put(
        type    => $self->{id} . 'ServerStopIO',
        message => "-done\n",
    );

    my $summary;
    my $wait_time = time + 30;
    while (!defined $summary && time < $wait_time) {
        $self->{staf_peer}->message_get(
            type        => $self->{id} . 'ServerSummary',
            callback    => sub {
                $summary = shift;
            }
        );
        eventloop_drain;
    }
    $self->{dwim_summary} = $summary;

    $self->{process}->stop();
    $self->{process}->destroy();

    return 1;
}


=head2 summary

Returns a hashref with the summary of the work done.
Summary is only available after dwim has stopped.

=over

=item Arguments

None

=back

=cut

sub summary {
    my $self = shift;

    return $self->{dwim_summary};
}


=head2 port

Returns the port the server is listening on.

=over

=item Arguments

None

=back

=cut

sub port {
    my $self = shift;

    return $self->{port};
}


=head2 get_id

Returns the id of the server process.

=over

=item Arguments

None

=back

=cut

sub get_id {
    my $self = shift;

    return $self->{id};
}


=head2 find_open_port

Returns an open dwim port on the client.

=over

=item Arguments

None

=back

=cut

sub find_open_port {
    my $self = shift;

    # DWIM defaults to port 6336, so start there and go up

    my $client = NACL::C::Client->new(
        name => $self->{server},
    );

    my $cmd;
    my $dwim_tcp_regex;
    if ($client->is_windows()) {
        $cmd = 'netstat -bnp tcp';
        $dwim_tcp_regex = 'TCP\s+\d+\.\d+\.\d+\.\d+:(\d+).*dwim';
    }
    else {
        $cmd = 'sudo netstat -ltp';
        $dwim_tcp_regex = 'tcp.*?:(\d+).*?/dwim';
    }
    my $dwim_ports = $client->execute(command => $cmd);
    my @sorted_ports = sort($dwim_ports =~ m%$dwim_tcp_regex%img);
    my $last_port_used = $sorted_ports[-1];

    $Log->comment("Last port used by dwim: $last_port_used");

    my $next_port;
    if (!defined $last_port_used) {
        $next_port = 6336;
    }
    else {
        $next_port = $last_port_used + 1;
    }

    $Log->comment("DWIM Server will use port $next_port");

    return $next_port;
}


1;
