# -*- perl -*-
#
## @author   dl-perf-qa@netapp.com
## @summary File collection creation, access, data integrity checking and more
## @status   public
##
## @pod here

=head1 NAME

Perf::DataSet - Create, access, track and verify a semi-random collection of mixed size files on a filer over a
          variety of protocols:
            - NFS
            - iSCSI
            - CIFS
            - FCP

=head1 DESCRIPTION

This package allows you to first generate a semi-random collection of files on a filer. See L<Perf::DataSet/add>
for more details.

    Example:
        $dataset = Perf::DataSet->new(filer=>$filer, vol=>$volname, clients=>[$client]);
        $dataset->add(size=>"2g");

Those files can then be used to generate different kinds of workloads. See L<Perf::DataSet/workload> for more details.

    Example:
        $dataset->read(random=>1, amount=>"5g", io_size=>"128k");

By default every read operation will attempt to validate the data obtained against a collection of
checksums kept in memory. This allows for fast stand-alone data integrity checking.

Additionally, this package allows you to track your collection of files through snapshots,
over VSM, QSM, etc. See L<Perf::DataSet/fork> for more details.

    Example:
        $dataset_vsm_dst = $dataset->fork(vol=>vol_dst_name);

Each operation on the collection of files can be performed simultaneously from one or several client hosts, and by
multiple processes on each hosts by default. See L<Perf::DataSet/new> for more details. This allows tests to fully utilize the filer. Note that Additions and Workloads can be throttled to a specific number of MB/s.

=head1 METHODS

=cut


package Perf::DataSet;

use strict;
use warnings;
use Tharn qw(host connect $Log subtest);
BEGIN{
    my @c = caller(2);
    $Log->comment("Loading " . __FILE__ . " from $c[1]:$c[2]");

    # Prevent Object::InsideOut from transforming exceptions. This is
    # necessary because NACL expects all its errors to be
    # NATE::Result::Config objects, but OIO::trap turns them into OIO
    # objects causing true failure reasons to be lost.
    require Object::InsideOut;
    *OIO::trap = sub {die($_[0]);};
}
use Cwd;
use Carp qw(confess longmess);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF validate);
use List::MoreUtils qw(all any none);
use Module::Common qw(validate_mixed);
use Sys::Hostname qw(hostname);
use Perf::Utils qw(min max get_random_element as_string in_bytes deep_copy
                   get_random_elements floor);
use Perf::DataSet::Command;
use Perf::DataSet::Manager;
use Perf::DataSet::Worker;
use Perf::DataSet::Container;
use Perf::DataSet::Element;
use Perf::Sysstat;
use Storable qw();
$Storable::Eval = 1;
use File::Path qw(mkpath);
use Readonly;
use NACL::C::Ndmpcopy;
use NACL::C::Volume;
use NACL::C::VolumeDirectory;
use NACL::STask::Node;
use NACL::STask::VolumeFileCopy;

# Fix for diamond multiple inheritance problem found in several
# sub-classes in Perf::DataSet
use MRO::Compat;
Class::C3::initialize();

Readonly::Scalar our $BLOCK_UPDATED => 4;
Readonly::Scalar our $BLOCK_ADDED   => 8;
# the volume file copy command can only handle 40 files at once
Readonly::Scalar my $VOL_FILE_COPY_LIMIT => 40;


=head2 new

Returns an object representing an empty collection. The data will reside on the filer/vserver and
volume specified, under [.snapshot/<snapshot>/][<qtree>/][<root_dir>]

=over

=item Arguments

B<filer> = string | I<filer connection>

I<Required> The filer on which to create the files

B<vol> = string

I<Required> The name of the volume on which to create the files

B<clients> = arrayref of string(s)

I<Required> The client hostnames from which to create and access the
files. As of now, all clients must be Linux machines unless using
CIFS. If type is 'cifs', only Windows clients are supported.

B<qtree> = string

I<Optional> The qtree name on which to create the files

B<vserver> = string

I<Optional> The vserver on which to create the files. This argument is required if there are more
than one vservers in the cluster the filer belongs to.

B<root_dir> = string

I<Optional> Name of the root folder the files should be put in. Defaults to a random value.

B<nodes> = arrayref of strings or NACL SystemNode components

I<Optional> The nodes in the cluster to be involved in handling
traffic. Defaults to all nodes. If NACL::C::CommandInterfaces are
being passed, they must be NACL::C::SystemNode objects or aliases or
subclasses. This includes NACL::C::Node and NACL::STask::Node.

B<processes> = number

I<Optional> The number of processes to create on each client for the purpose
of handling files. Defaults to 12.

B<type> = C<'nfs' | 'nfsv4' | 'nfsv4.1' | 'cifs' | 'cifs-smb1' | 'cifs-smb2' |
            | 'cifs-smb2.1' | 'cifs-smb3' | 'iscsi' | 'fcp'>

I<Optional> The protocol to be used to connect to the filer. Defaults
to 'nfs'. If 'cifs' is specified, the highest SMB version supported by
the Windows clients will be used. If the client defaults to SMB2 or 3
and one wants to test SMB1, use 'cifs-smb1' to disable SMB2 and 3 on
the filer. Specifying 'cifs-smb2', 'cifs-smb2.1', or 'cifs-smb3' will
ensure the client is capable of supporting the SMB protocol specified.

B<jumbo_frames> = C<0 | 1>

I<Optional> Enables jumbo frames between filer and clients. Defaults to 0.

B<nfs_mntopts> = hashref

I<Optional> Specify nfs mount options such as proto, rsize, wsize,
retrans, timeo, and hard. Defaults for each options are: proto=>tcp,
rsize=>65536, wsize=>65536, retrans=>2, timeo=>undef, hard=>0

I<Note: Do not use> hard I<unless you know what you are doing. If the filer panics with hard
mountpoints, Perf::DataSet will not return until the filer is brought back up, which may not happen if
Perf::DataSet is stuck waiting for the workload to finish.>

I<Note:> If no value is passed for C<timeo>, there are different
defaults set by the NFS version handlers (600 for NFSv3, 6000 for
NFSv4+).

B<no_purge_on_exit> = C<0 | 1>

I<Optional> Do not delete files on the filer when object falls out of scope. Defaults to 1.

B<allow_file_slicing> = C<0 | 1>

I<Optional> Enable file sharing among processes. This allows for more evenly distributed load and
shorter load time. Defaults to 1.

B<single_client_file_slicing> = C<0 | 1>

I<Optional> Enable file sharing among processes across clients. This
allows for more evenly distributed load and shorter load time.
Defaults to 0 for NAS protocols and 1 for SAN.

B<lun_size> = string

I<Deprecated> The LUN size will now be the size of each addition

B<debug> = C<0 | 1>

I<Optional> Enables additional logging. Defaults to 1.

B<seed> = number

I<Optional> Random number generator seed. Setting this number to a known value allows for repeatable
data creation and workloads.

B<load_container_from>

I<Optional> Location of persisted container files. See persist_container for more details.

B<load_container_name>

I<Optional> Name of persisted container to load. Required when load_container_from is specified.
See persist_container for more details.


B<enable_multipathing>

I<Optional> Turn on multipathing for 7-mode SAN Perf::DataSets


B<gather_sysstat>

I<Optional> Collect sysstat output for each filer in the background
and log in the results directory. Each DataSet action will have its
own log file named after the filer, action, and the action's UUID.
Defaults to 1.

B<disable_oplocks> = C<0 | 1>

I<Optional> Disable CIFS oportunistic locking in c-mode. Defaults to 1.

B<lifs>

I<Optional> Provide an arrayref of LIFs to use for testing. By default, DataSet will use all available data LIFs so this parameter allows the restriction of LIF usage.

=item Examples

Specify more than one client:

    $dataset = Perf::DataSet->new(filer=>$FILER, vol=>$FILER_VOL, clients=>[$client1, $client2, $client3]);

Set NFS protocol to TCP:

    $dataset = Perf::DataSet->new(filer=>$FILER, vol=>$FILER_VOL, clients=>[$client], nfs_mntopts => {proto=>tcp});

Create an iSCSI Perf::DataSet

    $dataset = Perf::DataSet->new(filer=>$FILER, vol=>$FILER_VOL, clients=>[$client], type=>'iscsi');

Specify NACL command interface objects for nodes:

    my @nodes = NACL::STask::Node->find();
    $dataset = Perf::DataSet->new(nodes => \@nodes, vol=>$FILER_VOL, clients=>[$client]);

=back

=cut

sub new {
    my $class = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {    #
                # User parameters
                #
                #filer: optional only if 'nodes' was specified
                filer => {type => SCALAR| OBJECT| UNDEF, optional => 1},
                # nodes: optional unless you wish to restrict what nodes get mounted by clients
                nodes => {type => ARRAYREF | UNDEF, optional => 1},
                vol => {type => SCALAR},
                clients => {type => ARRAYREF },
                # vserver: optional only if only one exists in the cluster of filer
                vserver => {type => SCALAR | UNDEF, optional => 1},
                processes => {type => SCALAR | UNDEF, optional => 1, default => 12},
                debug => {type => SCALAR  | UNDEF, optional => 1, default => 1},
                private_subnet_only => {type => SCALAR | UNDEF, optional => 1, default => 0},
                # nfs_mntopts: hashref of options: proto, rsize, wsize, retrans, hard, and timeo
                nfs_mntopts => {type => HASHREF | UNDEF, optional => 1, default => undef},
                # supported types: 'nfs', 'nfsv4', 'nfsv4.1','cifs',
                #                  'cifs-smb1', 'cifs-smb2',
                #                  'cifs-smb2.1', 'cifs-smb3',
                #                  'iscsi', 'fcp'
                type => {type => SCALAR | UNDEF, optional => 1, default => 'nfs'},
                # allow_file_slicing: allow files to be handled in slices by multiple processes
                # This allows for more evenly distributed load and shorter load time
                allow_file_slicing => {type => SCALAR | UNDEF, optional => 1, default => 1},
                # single_client_file_slicing: forces all slices of each file to be handled by processes
                # residing on the same client
                single_client_file_slicing => {type => SCALAR | UNDEF, optional => 1},
                # no_purge: don't purge the files when dataset falls out of scope
                no_purge_on_exit => {type => SCALAR , optional => 1, default => 1},
                # jumbo_frames: enable jumbo frames
                jumbo_frames => {type => SCALAR | UNDEF, optional => 1, default => 0},
                # qtree: name of the qtree
                qtree => {type => SCALAR | UNDEF, optional => 1},
                # lun_prefix: prefix for name of the luns on the filer
                lun_prefix => {type => SCALAR | UNDEF , optional => 1},
                # root_dir: folder to put files in
                root_dir => {type => SCALAR | UNDEF , optional => 1},
                # seed: random number generator seed
                seed => {type => SCALAR | UNDEF, optional => 1},
                # load_from: path to container files to load
                load_container_from => {type => SCALAR | UNDEF, optional => 1},
                # load_from: name of container to load
                load_container_name => {type => SCALAR | UNDEF, optional => 1},
                # enable_multipathing: force multipathing on in 7-mode
                enable_multipathing => {type => SCALAR | UNDEF, optional => 1,
                                        default => 0,},
                gather_sysstat => {type => SCALAR | UNDEF, optional => 1,
                                   default => 1,},
                disable_oplocks => {type => SCALAR | UNDEF, optional => 1,},
                lifs => {type => ARRAYREF | UNDEF, optional => 1,},
            }
           );

    my $self = \%params;

    #
    # Check arguments
    #

    _validate_type($self->{type});

    if ($self->{jumbo_frames} and ! $self->{private_subnet_only}) {
        confess "jumbo_frames only supported on private subnets";
    }

    #
    # type specific checks
    #
    if ($self->{type} eq 'iscsi' || $self->{type} eq 'fcp') {
        if ($self->{allow_file_slicing}) {
            if (defined($self->{single_client_file_slicing})
                && !$self->{single_client_file_slicing}) {
                confess("Option 'single_client_file_slicing' must be turned on"
                        . " for iscsi or fcp Perf::DataSets");
            }
            else {
                $self->{single_client_file_slicing} = 1;
            }
        }

        confess('LIF restriction via "lifs" not supported for SAN DataSets')
            if (defined($self->{lifs}));
    }

    if ($self->{type} =~ /^cifs/) {
        confess(  'CIFS Perf::DataSets do not support file slicing.'
                . ' See BURT489830.')
          if ($self->{allow_file_slicing});

        # disable opportunistic locks by default for CIFS
        $self->{disable_oplocks} //= 1;
    }
    else {
        confess('disable_oplocks only supported for CIFS')
            if ($self->{disable_oplocks});
    }

    confess('Multipathing is only valid for SAN Perf::DataSets')
        if ($self->{type} ne 'iscsi' && $self->{type} ne 'fcp' && $self->{enable_multipathing});

    #
    # Populate $self->{filer} if it's not
    #
    if (!$self->{filer}){
        confess "You must specify a filer or a list of nodes"
            if (!$self->{nodes} || !@{$self->{nodes}});
        $self->{filer} = $self->{nodes}->[0];
    }

    #
    # Check if a NACL object was passed in as the filer argument
    #
    handle_filer_connection($self);

    #
    # Make sure filer is part of nodes
    #
    if ($self->{nodes} && @{$self->{nodes}}){
        my $validated = 0;
        for my $node (@{$self->{nodes}}){
            my $node_ref = ref($node);
            my $node_name;

            if ($node_ref) {
                confess("Expected a NACL::C::SystemNode, got a $node_ref")
                    if (!$node->isa('NACL::C::SystemNode'));

                $node_name = $node->hostrec()->hostname();
                $self->{cis}->{$node_name} = $node;
            }
            else {
                # go ahead and create command interfaces to pass to
                # Node resources
                my $ci = NACL::STask::Node->new(name => $node);
                $self->{cis}->{$node} = $ci;
                $node_name = $node;
            }

            if ($node_name eq $self->{filer}){
                $validated = 1;
            }
        }

        confess(  "Argument 'filer' $self->{filer} is not part of 'nodes' list:"
                . " @{$self->{nodes}}")
          if (!$validated);

        $self->{nodes} = [keys(%{$self->{cis}})];

        # node list is supplied, restrict workloads to just those
        # nodes
        $self->{all_nodes_in_cluster} = 0;
    }
    else {
        $self->{all_nodes_in_cluster} = 1;
    }

    my $filer_host = host($self->{filer});
    my $filer_hosttype = $filer_host->hosttype();

    if ($filer_hosttype =~ /-ng-/) {
        if ('iscsi' eq $self->{type} || 'fcp' eq $self->{type}) {
            # Multipathing must be used in c-mode
            $self->{enable_multipathing} = 1;
        }
    }
    else {
        confess "You may not specify a vserver argument for a 7G filer"
            if ($self->{vserver});

        confess('CIFS SMB3 is not supported in 7-mode')
            if ('cifs-smb3' eq $self->{type});
    }

    # multipathing should always be on for C-mode and FCP, allow the
    # user to force it on for 7-mode iSCSI
    my $mp_and_alua = (($filer_hosttype =~ /-ng-/ || 'fcp' eq $self->{type}) ? 1 : 0);
    $self->{enable_multipathing} ||= $mp_and_alua;
    # ALUA is only supported in C-mode or with FCP
    $self->{enable_alua} = $mp_and_alua;

    # Create a connection if none was passed
    if (!$self->{fc}) {
        $self->{fc} = NACL::STask::Node->new(name => $self->{filer});
    }

    # Translate 'localhost' into hostname
    my $clients = $self->{clients};
    my $count = scalar @$clients;
    my $i = 0;
    while ($i < $count){
        my $client = $clients->[$i];
        if ($client =~ /^localhost$/i){
            $clients->[$i] = hostname();
            $Log->comment("Turned client '$client' into '" . $clients->[$i] ."'" );
        }
        $i ++;
    }


    # Ensure correct host types are used
    foreach my $c (@{$self->{clients}}){
        my $hosttype = host($c)->hosttype();

        confess("Type $self->{type} is not supported on $hosttype host $c")
          if (   ($self->{type} =~ /^(?:nfs|iscsi|fcp)/ && $hosttype !~ /linux/i)
              || ($self->{type} =~ /^cifs/i && $hosttype !~ /windows/i));
    }


    # Validate nfs options.
    my @nfs_args;
    if (defined($self->{nfs_mntopts})) {
        @nfs_args = %{$self->{nfs_mntopts}};
    }
    my %validated = validate(
                   @nfs_args,
                   {
                       # defaults from http://www.netapp.com/us/media/tr-3183.pdf
                       proto   => {type => SCALAR, optional => 1, default => 'tcp'},
                       rsize   => {type => SCALAR, optional => 1, default => 65536},
                       wsize   => {type => SCALAR, optional => 1, default => 65536},
                       retrans => {type => SCALAR, optional => 1, default => 2},
                       # timeo varies by NFS version, will be set by Managers
                       timeo   => {type => SCALAR, optional => 1, default => undef},
                       hard    => {type => SCALAR, optional => 1, default => 0},
                   });
    $self->{nfs_mntopts} = \%validated;

    #
    # Set the seed and display it
    # Rely on Perl's portable seeding mechanism to come up
    # with our own seed
    #
    if (!defined($self->{seed})){
        #my $seed = time ^ $$ ^ unpack("%L*", `ps axww | gzip`);
        my $seed = _get_uid();
        $self->{seed} = $seed;
    }
    $Log->comment("Seed for process $$: $self->{seed}");
    srand $self->{seed};

    $self->{uid} = Perf::Depot->make_uid;
    $self->{pid} = $$;
    $self->{copies} = [];
    $self->{additions} = {};
    $self->{read_only} = 0;

    # SAN datasets reside in a dedicated LUN rather than a dedicated shared folder
    if ($self->{type} =~ /iscsi|fcp/ && ! defined ($self->{lun_prefix})){
        $self->{lun_prefix} = $self->{uid};
    }
    else{
        $self->{root_dir} = $self->{uid} if (! defined $self->{root_dir});
    }

    #
    # Create Manager object common to all Perf::DataSets
    #
    $self->{manager} =
      Perf::DataSet::Manager->new(
                            type                => $self->{type},
                            nfs_mntopts         => $self->{nfs_mntopts},
                            jumbo_frames        => $self->{jumbo_frames},
                            private_subnet_only => $self->{private_subnet_only},
                            );

    #
    # Delete the parameters that were meant for the Manager
    #
    delete $self->{jumbo_frames};
    delete $self->{private_subnet_only};
    delete $self->{nfs_mntopts};

    bless $self, $class;

    #
    # Self register with manager
    #
    my $dataset_resource = $self->{manager}->register_dataset( dataset=>$self );

    # Start all the subtest processes and create the Workers
    $self->{manager}->create_processes($self->{uid});
    $self->{manager}->create_workers($self->{uid});

    #
    # Delete the parameters the Manager is now in charge of
    #
    #delete $self->{clients};
    delete $self->{processes};

    #
    # Perform some initialization steps on the filer
    #
    $self->{manager}->initialize($self->{uid});

    my $load_container_path = delete $self->{load_container_from};
    if (!$load_container_path ){
        $self->{container} = Perf::DataSet::Container->new;
    }
    else{
        my $load_container_name = delete $self->{load_container_name};
        if (!$load_container_name){
            confess "You must specify a load_container_name argument";
        }
        $self->_load_container_from( path => $load_container_path,
                                     name => $load_container_name);
    }

    $Log->comment("Created new Perf::DataSet: $self->{uid}\n" . as_string($self,2));

    return $self;
}

=head2 fork

Returns a new Perf::DataSet object containing the same files as the current object but not necessarily located
in the same path. This allows for data integrity checking with snapshots, QSM, VSM, volcopy, aggrcopy, etc..
In order to be able to run simultaneous workloads on a fork and its parent Perf::DataSet, you must specify
'detach => 1', see below for more details.

=over

=item Arguments

B<filer> = I<filer connection> | string

I<Optional> Filer at which files can be found

B<nodes> = arrayref of strings

I<Optional> Nodes to involve in handling traffic

B<vol> = string

I<Optional> Volume name in which files can be found

B<vserver> = string

I<Optional> Vserver in which files can be found

B<lifs> = arrayref of strings

I<Optional> List of LIFs to fork on. Note: if the parent DataSet had
LIF restrictions, the fork is onto a different vserver or cluster, and
no LIF list is supplied in the fork call, the LIF list will not be
copied to the fork.

B<qtree> = string

I<Optional> Qtree name in which files can be found

B<snapshot> = string

I<Optional> Snapshot name in which files can be found

B<root_dir> = string

I<Optional> Location under /vol/<vol> at which files can be found

B<read_only> = C<0 | 1>

I<Optional> Specifies whether the new Perf::DataSet object is writable. Defaults to 1.

B<copy> = C<0 | 1>

I<Optional> Attempts to copy the files via NDMP to the new filer/volume/path
specified by root_dir. Defaults to 0.

B<detach> = C<0 | 1>

I<Optional> Specifies whether fork object should run in its own processes so that
a fork and its parent Perf::DataSet can run simultaneous workloads. Defaults to 0. Setting
this parameter to 1 increases memory usage by a quantity equal to the memory footprint
of the parent dataset processes on each client.

=item Examples

Handling VSM destination

    # Create data in VSM source
    $dataset->add(size=>'2g');
    # Perform VSM
    ....
    # Verify data in VSM destination
    $dataset_dst = $dataset->fork(vol=>$voldst);
    $dataset_dst->read(random=>0, io_size=>'128k');

Making sure data is valid in snapshot before and after snap restore:

    # Create data in active filesytem
    $dataset->add(size=>'4g');
    $dataset_snap= $dataset->fork(snapshot=>$snap);
    # Overwrite active filesystem
    $dataset->write(random=>1, amount=>'2g', io_size=>'32k')
    # Read snapshot to make sure data is still correct
    $dataset_snap->read(random=>0, io_size=>'128k');
    # Restore snapshot
    ...
    # Now read the active filesystem
    $dataset_restored = $dataset_snap->fork(snapshot=>'');
    $dataset_restored->read(random=>0, io_size=>'128k');

=back

=cut

sub fork {
    my $self = shift;
    my %params;
    my $fork;

    %params = validate_mixed(
           @_,
           [ ],
           {
                filer => {type => SCALAR | OBJECT | UNDEF, optional => 1},
                nodes => {type => ARRAYREF | UNDEF, optional => 1},
                vol => {type => SCALAR | UNDEF, optional => 1},
                qtree => {type => SCALAR | UNDEF, optional => 1},
                lun_prefix => {type => SCALAR | UNDEF, optional => 1},
                snapshot => {type => SCALAR | UNDEF, optional => 1},
                root_dir => {type => SCALAR | UNDEF, optional => 1},
                vserver => {type => SCALAR | UNDEF, optional => 1},
                lifs => {type => ARRAYREF | UNDEF, optional => 1},
                type => {type => SCALAR | UNDEF, optional => 1},
                read_only => {type => SCALAR , optional => 1, default => 1},
                copy => {type => SCALAR | UNDEF, optional => 1, default => 0},
                detach => {type => SCALAR, optional => 1, default => 0},
           }
           );
    $fork = \%params;
    $fork->{uid} = Perf::Depot->make_uid;
    $fork->{parent_uid} = $self->{uid};

    if ($fork->{copy} && $self->is_san()) {
        confess "Option 'copy' not supported on SAN Perf::DataSets";
    }

    if ($self->is_busy){
        my $command = $self->get_active_command;
        $Log->comment("Perf::DataSet '$self->{uid}' is busy. Waiting for '" . $command->description
                    ."' to complete before forking");
        $self->wait;
    }

    $Log->comment("Forking Perf::DataSet '$self->{uid}' to:\n" . as_string($fork,2));

    #
    # Connect to filer
    #
    if ($fork->{filer}){
        handle_filer_connection($fork);

        if ($fork->{filer} ne $self->{filer}){
            if (host($fork->{filer})->hosttype() =~ /-ng/) {
                # Most likely the user meant to use 'nodes' rather than 'filer'
                # The only time it would make sense to get here is if
                # $fork->{filer} and $self->{filer} belong to two different
                # clusters. Make sure that's indeed the case.
                my $node = Perf::Depot->get_resource( class=>'Node', name=>$self->{filer});
                my $cluster = $node->get_cluster;
                my $cluster_nodes = $cluster->get_nodes;
                my $in_same_cluster = any { $_->get_name eq $fork->{filer} } @$cluster_nodes;
                if ($in_same_cluster){
                    confess "Filers '$self->{filer}' and '$fork->{filer}' are in the same cluster. Use the 'nodes' parameter instead of 'filer'";
                }
            }
        }
        else{
            $fork->{fc} = $self->{fc};
            $fork->{cis} = $self->{cis};
        }
        $fork->{all_nodes_in_cluster} = 1;
    }
    elsif ($fork->{nodes} && @{$fork->{nodes}}) {
        # A 'filer' is still required at all times for backwards compatibility reasons
        # TODO: Clean this up once we have cluster interfaces

        foreach my $node (@{$fork->{nodes}}) {
            my $node_ref = ref($node);
            my $node_name;

            if ($node_ref) {
                confess("Expected a NACL::C::SystemNode, got a $node_ref")
                    if (!$node->isa('NACL::C::SystemNode'));

                $node_name = $node->hostrec()->hostname();
                $fork->{cis}->{$node_name} = $node;
            }
            else {
                $fork->{cis}->{$node} = $self->{cis}->{$node}
                    // NACL::STask::Node->new(name => $node);
            }
        }

        $fork->{nodes} = [keys(%{$fork->{cis}})];
        $fork->{filer} = $fork->{nodes}->[0];
        $fork->{fc} = $fork->{cis}->{$fork->{filer}};
        # node fork, restrict node list
        $fork->{all_nodes_in_cluster} = 0;
    }
    else {
        # no filer or nodes provided
        $fork->{filer} = $self->{filer};
        $fork->{fc}    = $self->{fc};
        $fork->{cis}   = $self->{cis};

        $fork->{all_nodes_in_cluster} = $self->{all_nodes_in_cluster};

        # This is needed to check for cluster join/unjoin. It will get
        # recomputed before every workload.
        $fork->{nodes} = $self->{nodes};
    }

    if (!$fork->{vol}){
        $fork->{vol} =  $self->{vol};
    }
    if (!defined($fork->{qtree})){
        $fork->{qtree} =  $self->{qtree};
    }
    if (!defined($fork->{lun_prefix})){
        $fork->{lun_prefix} =  $self->{lun_prefix};
    }
    elsif($self->is_san()) {
        confess "Lun_prefix forks not supported yet";
    }
    if (!defined($fork->{snapshot})){
        $fork->{snapshot} = $self->{snapshot};
    }
    elsif ($self->is_san()) {
        confess "Snapshot forks not supported yet";
    }
    if (!$fork->{root_dir}){
        $fork->{root_dir} = $self->{root_dir};
    }
    if (!$fork->{type}){
        $fork->{type} = $self->{type};
    }
    if (!$fork->{no_purge_on_exit}){
        $fork->{no_purge_on_exit} = $self->{no_purge_on_exit};
    }
    if (!$fork->{vserver}){
        $fork->{vserver} = $self->{vserver};
    }

    # discard lifs when:
    # - vserver is passed
    # - nodes are passed and it's a different cluster
    if (defined($self->{lifs})
        && !defined($fork->{lifs})
        && $self->_same_lif_context($fork)) {
        $fork->{lifs} = $self->{lifs};
    }

    $fork->{pid} = $$;
    # 'Copy' data from parent
    # Mostly shallow copies except for the dataset itself (and additions)
    #
    $fork->{clients} = $self->{clients};
    $fork->{manager} = $self->{manager};
    $fork->{debug} = $self->{debug};
    $fork->{seed} = $self->{seed};
    #$fork->{lun_size} = $self->{lun_size};
    # TODO: Should this parameter be fork-able or kept by the manager?
    $fork->{allow_file_slicing} = $self->{allow_file_slicing};
    $fork->{single_client_file_slicing} = $self->{single_client_file_slicing};
    $fork->{additions} = $self->{additions};
    $fork->{enable_multipathing} = $self->{enable_multipathing};
    $fork->{enable_alua} = $self->{enable_alua};
    $fork->{disable_oplocks} = $self->{disable_oplocks};

    my $do_copy = delete($fork->{copy});

    # We don't want a pointer straight to the parent dataset, or we could end up with
    # collisions between the parent's dataset and our own
    # Note that the container is mostly empty here and only used to keep track
    # of stats
    $fork->{container} = $self->{container}->clone;

    bless $fork, ref($self);

    #
    # Register forked Perf::DataSet with Manager
    #
    $self->{manager}->register_dataset(dataset=>$fork);

    #
    # Fork the Workers
    #
    my $detached = delete $fork->{detach};
    # Detached forks are created via a three step process:
    # 1. New workers get created in new processes
    # 2. Parent workers dump their contents to disk.
    # 3. New workers pick up parent workers contents
    if ($detached){
        $fork->{manager}->create_processes($fork->{uid});
    }
    else{
        # Declare our processes as the fork's also
        my $dataset_resource = Perf::Depot->get_resource( class => 'DataSet', name=>$self->{uid});
        confess "No Perf::DataSet resource found for $self->{uid}" if (!$dataset_resource);
        my $process_group = Perf::Depot->get_resources( class=> 'Process::Group', used_by=>$dataset_resource);
        my $fork_resource = Perf::Depot->get_resource( class => 'DataSet', name=>$fork->{uid});
        $fork_resource->add_dependencies( $process_group );
    }
    $fork->{manager}->create_workers($fork->{uid});

    #
    # Copy the Perf::DataSet data if required
    #
    if ($do_copy) {
        # TODO: Any error checking?
        $self->copy($fork);
    }

    $Log->comment("Created new fork Perf::DataSet:\n" . as_string($fork,2));

    return $fork;
}

sub query {
    my $self = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {
            type => {type => SCALAR | UNDEF, optional => 1},
            count => {type => SCALAR | UNDEF, optional => 1},
            block_size => {type => SCALAR | UNDEF, optional => 1},
            compatible_with_io_size => {type => SCALAR | UNDEF, optional => 1},
            min_size => {type => SCALAR | UNDEF, optional => 1},
            max_size => {type => SCALAR | UNDEF, optional => 1},
            strict_size => {type => SCALAR | UNDEF, optional => 1},
            sort_by => {type => SCALAR | UNDEF, optional => 1,
                        regex => qr/^(?:size|size_on_disk)$/,
                        depends => ['count'] },
            decreasing_order => {type => SCALAR | UNDEF, optional => 1, default=>1},
           }
        );
    my $rset = $self->{container}->filter({
                                type=>$params{type},
                                block_size=>$params{block_size},
                                compatible_with_io_size=>$params{compatible_with_io_size},
                                min_size=>$params{min_size},
                                max_size=>$params{max_size},
                                strict_size=>$params{strict_size},
                                });
    my $result = $rset;

    if ($params{count}){
        #
        # Flatten the dataset container
        #
        my @elements = values %{ $rset->get_elements };

        $result = Perf::DataSet::Container->new;
        my $sort_by = $params{sort_by};
        if ($sort_by){
            #
            # Sort the elements
            #
            my @sorted_elements;
            if ($params{decreasing_order}){
                @sorted_elements = sort { $b->get_field($sort_by) <=> $a->get_field($sort_by) }
                                         @elements;
            }
            else{
                @sorted_elements = sort { $a->get_field($sort_by) <=> $b->get_field($sort_by) }
                                         @elements;
            }

            #
            # Return the first 'count' elements of the sorted list
            #
            my $i = 0;
            while ($i < $params{count}){
                $result->set_element( shift @sorted_elements );
                $i ++;
            }

        }
        else{
            #
            # Pick $count random elements
            #
            my $random_elements = get_random_elements(\@elements,$params{count});
            #
            # Assemble result set
            #
            for my $e (@$random_elements){
                $result->set_element( $e );
            }
        }
    }
    my $relative_path = $self->{manager}->get_relative_path($self->{uid});
    my $path_prefix = "/vol/$self->{vol}/$relative_path";
    $result = $result->get_paths({prefix=>$path_prefix});
    return $result;
}

=head2 get_random_file

Returns a random file in the collection. File paths returned are filer paths starting with /vol

=over

=item Arguments

    None

=item Examples

    $file = $dataset->get_random_file;

=back

=cut

sub get_random_file {
    my $self = shift;
    my $results = $self->query(type=>"files",count=>1);
    return $results->[0];
}

=head2 get_largest_file

Returns the largest file in the collection. File paths returned are filer paths starting with /vol

=over

=item Arguments

    None

=item Examples

    $file = $dataset->get_largest_file;

=back

=cut

sub get_largest_file {
    my $self = shift;
    my $results = $self->query(type=>"files",count=>1,sort_by=>'size');
    return $results->[0];
}

=head2 get_all_files

Returns an arrayref listing all files in the collection. File paths returned are filer paths starting with /vol

=over

=item Arguments

    None

=item Examples

    $files = $dataset->get_all_files;

=back

=cut

sub get_all_files {
    my $self = shift;
    if ($self->{is_client}){
        confess "You cannot invoke this subroutine from a client";
    }
    my %params = @_;
    return $self->query(type=>"files");
}

=head2 workload

Starts an I/O workload on the collection of files and returns a hashref containing
a description of what was done

=over

=item Arguments

B<random> = C<0 | 1>

I<Required> Enables random vs sequential workload

B<read_pct> = number

I<Required> Specifies the percentage of read operations during workload

B<io_size> = string

I<Required> Specifies the size of each IO. You must specify a unit like C<'32k'>. The io_size
must be a multiple of the block_size specified during the 'add' operation

B<amount> = string

I<Optional> Amount of data to read/write during the workload. The default value during a sequential
workload is the size of the dataset. The value C<0> can be specified for asynchronous workloads and
means infinite. You then need to call the method 'abort' to stop the workload.

B<duration> = number

I<Optional> Duration of workload. The value C<0> can be specified for asynchronous workloads and
means infinite. You then need to call the method 'abort' to stop the workload.

B<percentage> = number

I<Optional> Same as 'amount' but expressed in terms of percentage of the total dataset size

B<verify> = C<'checksum' | 'pattern' | 'auto'>

I<Optional> Type of data integrity verification to perform during
read.  Setting to 'auto' will verify using checksums or patterns if
the Perf::DataSet contains a mixture of files created with and without
checksums.  Defaults to C<'checksum'>.

B<verify_count> = C<0 | 1>

I<Optional> When set to 1, a read workload will not stop at the first data integrity check failure
but instead count those failures and finish the workload

B<async> = C<0 | 1>

I<Optional> Return immediately while workload continues in the background. Defaults to 0. Some
limitations apply to what can be done while an asynchronous workload is running:
- If the asynchronous workload is a CIFS workload, you cannot start another read CIFS workload
on the same volume from another Perf::DataSet
- If the asynchronous workload is a SAN workload, you may not be able to start another SAN
workload without first mounting your Perf::DataSet prior to the async workload starting.

    # This may not work
    $dataset->workload( async=>1, ...);
    $dataset2->workload( async=>1, ...);
    # This will work all the time
    $dataset2->mount;
    $dataset->workload( async=>1, ...);
    $dataset2->workload( async=>1, ...);

B<max_tput> = number

I<Optional> Maximum throughput allowed during run in MB/s. 0 means no limit. Defaults to 0

B<files> = arrayref

I<Optional> List of files to exercise the workload on. By default all files participate
in the workload

B<delete_first> = C<0 | 1>

I<Optional> Delete/truncate files before writing. This only works with sequential write workloads.
Defaults to 0.

B<delete_ratios> = arrayref

I<Optional> List of percentages for each 'delete' action: delete, truncate, no-op. Defaults to
C<[33,67,100]>, meaning 33% of each.

B<panic_on_corruption> = C<0 | 1>

I<Optional> Panic all nodes involved in testing when data corruption is detected.

B<skip_flush> = 0|1

I<Optional> Workloads will typically end by forcingly flushing data to disk on the nodes. You may disable
this behavior by setting this parameter to 1. Default to 0.

=item Examples

Most likely you'll be calling the 'workload' method through one of its two wrappers: 'read' and 'write'.

Start a sequential read workload at no more than 15MB/s

    $ret = $dataset->read(random=>0, io_size=>'128k', max_tput=>15);
    print "Workload lasted $ret->{duration}->{total} \n";

Start an asynchronous random write workload

    $dataset->write(random=>1, block_size=>'4k', async=>1, duration=>0);
    ... do something ...
    $ret = $dataset->abort;
    print "Workload wrote $ret->{amount}->{total} Bytes \n";

Start a mixed workload of 50% read and 50%write

    $ret = $dataset->workload(read_pct=>50, random=>1, io_size=>'8k', amount=>'10g');
    print " Read: $ret->{amount}->{read} Bytes \n";
    print " Written: $ret->{amount}->{write} Bytes \n";

Start a workload that only counts the data integrity errors

    $ret = $dataset->read(random=>0, io_size=>'256k', verify_count=>1);
    print "Data integrity errors: $ret->{validation_errors} \n";

=back

=cut

sub workload {
    my $self = shift;
    my $workload = $self->create_command('Workload', @_);
    my $ret = $self->execute($workload);
    return $ret;
}

=head2 modular_workload

Allows users to run arbitrarily complex workloads of their own devising using pre-defined Operation
objects. The user workload needs to be packaged as a subclass of Perf::DataSet::Sequence.
See Perf::DataSet::ModularWorkload for more details.

=over

=item Arguments

B<sequence> = object belonging to a subclass of Perf::DataSet::Sequence

I<Required> See Perf::DataSet::ModularWorkload for more details.

B<async> = C<0 | 1>

I<Optional> Return immediately while workload continues in the background. Defaults to 0. Some
limitations apply to what can be done while an asynchronous workload is running:
- If the asynchronous workload is a CIFS workload, you cannot start another read CIFS workload
on the same volume from another Perf::DataSet
- If the asynchronous workload is a SAN workload, you may not be able to start another SAN
workload without first mounting your Perf::DataSet prior to the async workload starting.

        # This may not work
        $dataset->workload( async=>1, ...);
        $dataset2->workload( async=>1, ...);
        # This will work all the time
        $dataset2->mount;
        $dataset->workload( async=>1, ...);
        $dataset2->workload( async=>1, ...);


B<skip_flush> = 0|1

I<Optional> Workloads will typically end by forcingly flushing data to disk on the nodes. You may disable
this behavior by setting this parameter to 1. Default to 0.

B<skip_container_validate> = 0|1

I<Optional> Once a workload ends, Perf::DataSet will attempt to validate that's seen from the client(s) is
what's expected. You may disable this behavior by setting this parameter to 1. Defaults to 0.

B<on_error_handler> = coderef

I<Optional> Provides users fine-grained control over what should happen to the workload when an error
occurs. The on_error_handler subroutine will be passed the current workload object and the exception
that occured while executing the last Operation. User may then selectively decide to abort the
worload with or without raising an exception or continue.

    my $own_handler = sub {
                            my $workload = shift;
                            my $current_exception = shift;
                            if ($current_exception->isa('Something::Really::Bad'){
                                $Log->comment("Caught exception:\n" . $current_exception->as_string);
                                $workload->abort;
                                $workload->die_on_operation_error;
                            }
                      };
    $dataset->modular_workload( ..., on_error_handler=> $own_handler);

=back

=cut

sub modular_workload {
    my $self = shift;
    my %args = @_;
    my $execute = delete $args{execute};
    $execute = 1 if (!defined($execute));
    my $workload = $self->create_command('ModularWorkload', %args);
    return $workload if (!$execute);
    my $ret = $self->execute($workload);
    return $ret;
}

sub create_command {
    my $self = shift;
    my $command = shift;
    my $obj = Perf::DataSet::Command->new($command,
                                    debug=>$self->{debug},
                                    dataset_uid=>$self->{uid},
                                    @_);
    return $obj;
}

=head2 read

Wrapper around L<Perf::DataSet/workload> for read operations

=over

=item Arguments

See L<Perf::DataSet/workload> for details

=back

=cut

sub read {
    my $self = shift;
    my %params = @_;
    $params{read_pct} = 100;
    return $self->workload(%params);
}

=head2 write

Wrapper around L<Perf::DataSet/workload> for write operations

=over

=item Arguments

See L<Perf::DataSet/workload> for details

=back

=cut

sub write {
    my $self = shift;
    my %params = @_;
    $params{read_pct} = 0;
    return $self->workload(%params);
}


sub copy {
	my $self = shift();
	my $fork = shift();

	if ($self->{fc}->hostrec()->is_cmode()) {
		$self->mount();
		$fork->mount();

		my @src_paths = values(%{$self->get_worker_full_paths()});
		my @dst_paths = values(%{$fork->get_worker_full_paths()});

		my $client = Perf::DataSet::Depot->get_resource(
			class => 'Host',
			name => $self->{clients}->[0],
		);

		$client->run_cmd("rsync -av $src_paths[0]/ $dst_paths[0]/",
		    max_idle => 240,
		    timeout => -1,);

		$fork->unmount();
		$self->unmount();
	}
	else {
		my $src_relative_path = $self->{manager}->get_relative_path($self->{uid});
		my $src =  "/vol/$self->{vol}/$src_relative_path";
		my $dst_relative_path = $self->{manager}->get_relative_path($fork->{uid});
		my $dst = "/vol/$fork->{vol}/$dst_relative_path";

		if ($fork->{filer} ne $self->{filer}) {
			$dst = "$fork->{filer}:$dst";
		}

		NACL::C::Ndmpcopy->ndmpcopy(
			command_interface => $fork->{fc},
			source => $src,
			destination => $dst,);
	}
}


=head2 add_single_file

Wrapper around L<Perf::DataSet/add> to add a single file to the collection.

=over

=item Arguments

B<size> = string

I<Required> Size of the file

See L<Perf::DataSet/add> for complete list

=back

=cut

sub add_single_file {
    my $self = shift;
    my %params = @_;
    my $file_size = $params{size};
    if (!$file_size){
        confess "You must specify a file size";
    }
    # SAN Perf::DataSets do not use the structure and size parameters to Addition
    if ($self->is_nas()) {
        $params{tree_structure} = [[1,0,0,0]];
        $params{size_boundaries} = [$file_size,$file_size];
        $params{size_distribution} = [100];
    }
    else {
        confess("Cannot add a single file to SAN Perf::DataSets with multiple clients")
            if (1 < scalar(@{$self->{clients}}));
    }

    return $self->add(%params);
}


=head2 add_single_size

Wrapper around L<Perf::DataSet/add> to add files of equal size to the
collection to ensure an evenly distributed workload.

=over

=item Arguments

B<size> = string

I<Required> Total size of the addition

See L<Perf::DataSet/add> for complete list

=back

=cut

sub add_single_size {
    my $self = shift;
    my %params = @_;

    my $total_size    = in_bytes($params{size});
    my $process_count = $self->get_processes();
    my $client_count  = scalar(@{$self->get_clients()});
    my $file_count    = $self->get_processes() * $client_count;
    my $buffer_align  = (defined($params{buffer_align}) ? $params{buffer_align} : 512);
    # ensure size is a multiple of $buffer_align
    my $file_size     = $buffer_align * floor(($total_size / $file_count) / $buffer_align);

    $params{tree_structure}     = [[$file_count, 0, 0, 0]];
    $params{size_boundaries}    = [$file_size, $file_size];
    $params{size_distribution}  = [100];
    $params{file_size_priority} = 'size_boundaries';

    return $self->add(%params);
}

=head2 add

Add files to the collection. Returns a hashref containing summary of what exactly got created.

=over

=item Arguments

B<size> = string

I<Required> Amount of data to generate. Must contain a unit such as
'g' or 'm'. Note that the LUNs created during SAN additions will
consume twice as much space as requested since 100% size reservations
are on by default for LUNs. Size volumes accordingly.

B<compute_checksum> = C<0 | 1>

I<Optional> Compute and store a checksum for each "block" of data. Used for data integrity checking.
Defaults to C<1>.

B<block_size> = string

I<Optional> How much data should be used to generate each checksum. The higher the value here, the
smaller the memory footprint. For best performance, this parameter should be set to the smallest IO size
you intend to use in your workloads. Defaults to C<'4k'>.

B<tree_structure> = arrayref

I<Optional> List of arrays each describing number of objects at a specific level
in the folder tree. [number of files, number of hardlinks, number of symlinks, number of directories].
Last item in the list must indicate 0 directories. Defaults to C<[[5, 0, 0, 20],[10, 0, 0, 20],[10, 0, 0, 5],
[5, 0, 0, 0]]>

B<size_boundaries> = arrayref

I<Optional> Define some size intervals from which a random file size
will be chosen.  Each interval will then be assigned a probability
using the next parameter "size_distribution". Defaults to
C<["0B","64B","32k","2036k","2040k","2044k","16m","8g"]>

B<size_distribution> = arrayref

I<Optional> Define probabilities for each size interval defined in size_boundaries. Defaults to
C<[1,3,10,4,4,30,48]>.

Together, size_boundaries and size_distribution are used to generate a
collection of files of various random sizes.  size_boundaries defines
the edges of buckets from which files are selected.  size_distribution
defines the probability that something will be selected from a
bucket. The defaults can be viewed as:

  0B <-1%-> 64B <-3%-> 32k <-10%-> 2036k <-4%-> 2040k <-4%-> 2044k <-30%-> 16m <-48%-> 8g

i.e. there is a 30% chance that a randomly chosen file will be between
2044k and 16m.  Note that those are probabilities, not
guarantees. Smaller files have a higher probabilty to fit within the
"size" limit of the collection than larger files.  size_distribution
probabilities must sum to 100%.  If all probability is allocated so
that there are fewer probabilities than intervals, the larger
intervals will default to 0%.

B<file_size_priority> = 'size' | 'size_boundaries'

I<Optional> This option is only used when the lowest size_boundaries is not 0. 'size' is used to
make a file smaller than the smallest boundary and make the Perf::DataSet size exactly the amount specified.
'size_boundaries' will not make a file smaller than the smallest boundary, but the Perf::DataSet size may
be smaller than the amount specified. Defaults to 'size'.

B<max_tput> = number

I<Optional> Maximum throughput during file creation in MB/s. 0 means no max. Defaults to 0.

B<async> = C<0 | 1>

I<Optional> The call returns immediately while the file creation happens in the background.
Defaults to C<0>.

B<sparse> = float between 0 and 1

I<Optional> Specifies the probability of any given file to be
sparse. Defaults to 0.  For SAN Perf::DataSets, any value other than zero
will result in a sparse raw device file.

B<sparse_ratio> = float between 0 and 1

I<Optional> Specifies what ratio of blocks at level L</sparse_level> should not be written to disk.
Defaults to C<1>. Note that even for a value of C<1> some blocks will have to be written to disk at level
sparse_level in order to populate the metadata blocks at levels above sparse_level

B<sparse_level> = integer between 0 and 7

I<Optional> Specifies at what level the files should be sparse. Defaults to C<0>, meaning that L1 blocks
and above will be fully populated. If a file isn't large enough to contain blocks of level sparse_level or
sparse_level + 1, only the first and last data block in the file will be written to disk.

B<sparse_align> = 'right' | 'left'

I<Optional> Defines which pointer in a sparse metadata block should be populated. 'left' means the first one.
'right' means the last one. Defaults to 'right'.

=item Examples

Add files using defaults:

    $ret = $dataset->add(size=>'4g');
    print "Created $ret->{count}->{files} files \n";

Add 5 files of 20m directly in the root directory

    $dataset->add(size=>'100m', size_boundaries=>['20m','20m'], size_distribution=>[100],
                  tree_structure=>[[5,0,0,0]]);

Add files that are only of sizes between 10m and 20m

    $dataset->add(size=>'100m', size_boundaries=>['10m','20m'], size_distribution=>[100]
                  file_size_priority=>'size_boundaries');

Start an aysnchronous add:

    $dataset->add(size=>'4g', async=>1);
    ...do something...
    $ret = $dataset->wait;

=back

=cut

sub add {
    my $self = shift;
    my %args = @_;
    my $execute = delete $args{execute};
    $execute = 1 if (!defined($execute));

    # Temporary confession to see how users expect SAN Perf::DataSets to behave
    confess 'SAN Perf::DataSets do not support the tree_structure, '
      . 'size_boundaries or size_distribution parameters'
      if ($self->is_san()
          && (any { exists($args{$_}) }
              ('tree_structure', 'size_boundaries', 'size_distribution')));

    $args{type} = $self->get_type();

    my $addition = $self->create_command('Addition', %args);
    # Hack to help with testing
    return $addition if (!$execute);
    my $ret = $self->execute($addition);
    return $ret;
}

=head2 set_max_tput

Changes the maximum throughput setting defined in current running command
such as 'add' or 'read' or 'write'.

=over

=item Arguments

number - Maximum allowed throughput in MB/s

=item Examples

    $dataset->write(random=>1,
                  io_size=>$io_size,
                  duration=>$duration,
                  async=>1);
    sleep 5;
    # Change the maximum throughput from whatever the filer was able to
    #  process to a maximum of 20 MB/s
    $dataset->set_max_tput(20);


=back

=cut

sub set_max_tput {
    my $self = shift;
    my $max_tput = shift;
    my $command = $self->get_active_command;
    if (!$command){
        $Log->debug("Subroutine 'abort' was invoked but no code was executing");
        return;
    }
    $command->set_max_tput($max_tput);
    return;
}

=head2 block_count

Returns the current number of "blocks" in the dataset
"Blocks" here refers to the 'block_size' argument of an Addition
which may be greater than a WAFL block (4k). Only files have blocks.
This number will be very close to ($dataset->size / $block_size) but
not quite due to file tails.

=over

=item Arguments

None

=item Examples

    $blocks = $dataset->block_count;
    print "Number of blocks in current dataset: $blocks \n";

=back

=cut

sub block_count {
    my $self = shift;
    my $section = shift;
    $section = 'total' if (!$section);
    my $ret = $self->{container}->get_stats->get_field('blocks',$section);
    return $ret;
}

=head2 block_on_disk_count

Returns the current number of non-hole "blocks" in the dataset
"Blocks" here refers to the 'block_size' argument of an Addition
which may be greater than a WAFL block (4k). Only files have blocks.
This number will be very close to ($dataset->size / $block_size) but
not quite due to file tails.

=over

=item Arguments

None

=item Examples

    $blocks = $dataset->block_on_disk_count;
    print "Number of non-hole blocks in current dataset: $blocks \n";

=back

=cut

sub block_on_disk_count {
    my $self = shift;
    my $section = shift;
    $section = 'total' if (!$section);
    my $ret = $self->{container}->get_stats->get_field('blocks_on_disk',$section);
    return $ret;
}

=head2 wafl_block_count

Returns the current number of WAFL block (4k) in the dataset
Only files have blocks. This number will be very close to ($dataset->size / 4096) but
not quite due to file tails.

=over

=item Arguments

None

=item Examples

    $blocks = $dataset->wafl_block_count;
    print "Number of WAFL blocks in current dataset: $blocks \n";

=back

=cut

sub wafl_block_count {
    my $self = shift;
    my $section = shift;
    $section = 'total' if (!$section);
    my $ret = $self->{container}->get_stats->get_field('wafl_blocks',$section);
    return $ret;
}

=head2 wafl_block_on_disk_count

Returns the current number of non-hole WAFL block (4k) in the dataset
Only files have blocks. This number will be very close to ($dataset->size / 4096) but
not quite due to file tails.

=over

=item Arguments

None

=item Examples

    $blocks = $dataset->wafl_block_on_disk_count;
    print "Number of non-hole WAFL blocks in current dataset: $blocks \n";

=back

=cut

sub wafl_block_on_disk_count {
    my $self = shift;
    my $section = shift;
    $section = 'total' if (!$section);
    my $ret = $self->{container}->get_stats->get_field('wafl_blocks_on_disk',$section);
    return $ret;
}

=head2 size

Returns the current size of the dataset

=over

=item Arguments

None

=item Examples

    $size = $dataset->size;
    print "Size of current dataset: $size Bytes \n";

=back

=cut

sub size(;$) {
    my $self = shift;
    my $section = shift;
    my $ret = $self->{container}->get_size($section);
    return $ret;
}

=head2 size_on_disk

Returns the current size on disk of the dataset. The size and size_on_disk will only differ
in case sparse files were created

=over

=item Arguments

None

=item Examples

    $size = $dataset->size_on_disk;
    print "Size on disk of current dataset: $size Bytes \n";

=back

=cut

sub size_on_disk(;$) {
    my $self = shift;
    my $section = shift;
    my $ret = $self->{container}->get_size_on_disk($section);
    return $ret;
}

=head2 size_of

Returns the size of a specific file in the dataset

=over

=item Arguments

string - Relative path to the file or full path from /vol

=item Examples

    $file = $dataset->get_random_file;
    $size = $dataset->size_of($file);
    print "File '$file' is of size '$size' \n";

=back

=cut

sub size_of($$) {
    my $self = shift;
    my $path = shift;

    $path =~ s/\/vol\/$self->{vol}\///;
    my $relative_path = $self->get_relative_path;
    $path =~ s/\/?$relative_path\///;

    return $self->{container}->size_of($path);
}

=head2 exists

Returns true if the argument passed in is a valid file or directory

=over

=item Arguments

string - Relative path to the file/dir or full path from /vol

=back

=cut

sub exists($$){
    my $self = shift;
    my $path = shift;

    $path =~ s/\/vol\/$self->{vol}\///;
    my $relative_path = $self->get_relative_path;
    $path =~ s/\/?$relative_path\///;

    my $type = $self->{container}->contains($path);
    if (!$type){
        return 0;
    }
    return 1;
}

=head2 purge

Physically deletes the files from the filer

=over

=item Arguments

None

=item Examples

    $dataset->purge;

=back

=cut

sub purge {
    my $self = shift;
    my %args = @_;
    my $purge = Perf::DataSet::Command->new('Purge',%args, dataset_uid=>$self->{uid});
    my $ret = $self->execute($purge);
    return $ret;
}

sub execute($$) {
    my $self = shift;
    my $command = shift;
    my %args = @_;
    if (!$command || !$command->isa('Perf::DataSet::Command')){
        confess "Argument must be a Perf::DataSet::Command";
    }

    #if ($command && ($command->isa("Perf::DataSet::Workload") || $command->isa("Perf::DataSet::Addition")
    #    || $command->isa("Perf::DataSet::Purge") || $command->isa("Perf::DataSet::Query")) ){
    #    # We're good
    #}
    #else{
    #    confess "Argument must be a Workload, Addition,Purge or Copy";
    #}

    my $async = $command->{async};
    #
    # See if the user has not called wait() since
    # the previous execute statement
    #
    my $active_command = $self->get_active_command;
    if ($active_command){
        $Log->comment("Command '" . $command->description ."' waiting for '"
                    . $active_command->description ."' to complete");
        $self->wait();
    }

    $Log->comment(  "$self->{uid} "
                  . "Starting a '"
                  . $command->description() . "':\n"
                  . as_string($command, $self->{debug} ? 3 : 2));

    if ($self->{gather_sysstat}) {
        $self->start_sysstat($command->description() . '_' . $command->get_uid());
    }

    #
    # Pre Execution
    #
    $command->pre_execute($self);

    #
    # Execution
    #
    my $start_time = time();
    eval{
        $self->{manager}->execute(command=>$command, dataset=>$self->{uid});
    };
    if ($@){
        delete $self->{active};
        confess("Got errors during '" . $command->description() ."':\n$@");
    }
    my $returns;
    if (!$async){
        $returns = $self->wait();
    }
    my $execution_time = time() - $start_time;
    $Log->comment("Command '" . $command->description ."' completed in (${execution_time}s)\n");

    return $returns;
}

=head2 abort

Aborts current asynchronous workload and returns a description of what was done during the workload
Note that this doesn't work on asynchronous additions yet...

=over

=item Arguments

None

=item Examples

    $dataset->write(random=>1, io_size=>'4k', async=>1, duration=>0);
    ... do something ...
    $ret = $dataset->abort;
    Print "Amount written: $ret->{amount}->{write} \n";

=back

=cut

sub abort {
    my $self = shift;
    my $command = $self->get_active_command;
    my %args = @_;
    my $async = $args{async};
    if (!$command){
        $Log->debug("Subroutine 'abort' was invoked but no code was executing");
        return;
    }
    $command->abort;

    if ($async) {
        if ($self->{gather_sysstat}) {
            $self->stop_sysstat();
        }
        return;
    }

    return $self->wait();
}

sub send_event {
    my ($self,$event) = @_;
    confess "You must specify an event" if (!$event);
    confess "Argument must be a Perf::DataSet::Event" if (!$event->isa('Perf::DataSet::Event'));
    $self->{manager}->send_event($event, dataset_uid=>$self->{uid});
    return;
}

sub mount {
    my $self = shift();
    my %args = @_;
    return $self->{manager}->mount($self->{uid}, %args);
}
sub unmount {
    my $self = shift;
    my %args = @_;
    return $self->{manager}->unmount($self->{uid}, %args);
}
sub get_mountpoints {
    my $self = shift;
    return $self->{manager}->get_mountpoints($self->{uid},@_);
}

=head2 get_relative_path

Returns the location of the data in the volume. For a NAS Perf::DataSet for example the path
returned might look like '.snapshot/<snapshot_name>/<root_dir>' or '<qtree>/<root_dir>' or
simply '<root_dir>'.

=over

=item Arguments

None

=back

=cut

sub get_relative_path {
    my $self = shift;
    my $rel_path = $self->{relative_path};
    if (!$rel_path){
        $rel_path = $self->{manager}->get_relative_path($self->{uid});
        $self->{relative_path} = $rel_path;
    }
    return $rel_path;
}

=head2 get_relative_path_info

Returns the same information as get_relative_path but as a hash rather than a string. A NAS Perf::DataSet
whose relative path is '.snapshot/<snapshot_name>/<root_dir>' would return a hash looking like this:
    (
         snapshot => "<snapshot_name>",
         root_dir => "<root_dir>"
    )


=over

=item Arguments

None

=back

=cut

sub get_relative_path_info {
    my $self = shift;
    my %relative_path_info;
    if ($self->get_snapshot){
        $relative_path_info{snapshot} = $self->get_snapshot;
    }
    if ($self->get_qtree){
        $relative_path_info{qtree} = $self->get_qtree;
    }
    if ($self->get_lun_prefix){
        $relative_path_info{lun_prefix} = $self->get_lun_prefix;
    }
    if ($self->get_root_dir){
        $relative_path_info{root_dir} = $self->get_root_dir;
    }
    return %relative_path_info;
}

=head2 get_full_path_info

Returns a hash containing all the information necessary to pinpoint the location of the data.
For example a Perf::DataSet whose data is located in the cluster of filer 'filer0', on vserver 'vs0',
in volume 'vol0', in folder 'root0' would return:

    (
        filer    => 'filer0',
        vserver  => 'vs0',
        vol      => 'vol0',
        root_dir => 'root0',
    )

=over

=item Arguments

None

=back

=cut

sub get_full_path_info {
    my $self = shift;
    my %full_path_info = (  filer   => $self->get_filer,
                            vserver => $self->get_vserver,
                            vol     => $self->get_volume,
                         );
    my %rel_path_info = $self->get_relative_path_info();
    @full_path_info{keys(%rel_path_info)} = values(%rel_path_info);
    return %full_path_info;
}

sub get_worker_full_paths {
    my ($self) = @_;
    return $self->{manager}->get_worker_full_paths($self->{uid});
}


=head2 wait

Blocks until the current asynchronous workload completes

=over

=item Arguments

None

=item Examples

    $dataset->write(random=>1, io_size=>'4k', async=>1, amount=>'4g');
    ... do something ...
    $ret = $dataset->wait;
    Print "Duration: $ret->{duration}->{total} \n";

=back

=cut

sub wait {
    my ($self,%args) = @_;
    my $wait_finish = $args{wait_finish};
    my $command = $self->get_active_command;
    if (!$command){
        $Log->debug("Subroutine 'wait' was invoked but no code was executing");
        return;
    }
    my $returns;
    eval{
        $returns = $self->{manager}->wait($self->{uid}, wait_finish=>$wait_finish);
    };
    my $err = $@;
    delete $self->{active};

    if ($err) {
        if ($self->{gather_sysstat}) {
            $self->stop_sysstat();
        }

        if ($command->isa('Perf::DataSet::Addition')) {
            eval {
                $self->dump_diagnostics();
            };
            if ($@) {
                $Log->warn("Problem dumping diagnostics: $@");
            }
        }

        die $err;
    }

    #if ($self->{debug}){
    #    $Log->comment("Return data after execute:\n" . as_string($returns));
    #}

    #
    # Post execution
    #
    $returns = $command->post_execute($self, $returns);
    #if ($self->{debug}){
    #    $Log->comment("Return data after post-execute:\n" . as_string($returns));
    #}

    if ($self->{gather_sysstat}) {
        $self->stop_sysstat();
    }

    return $returns;
}

sub wait_finish {
    my ($self) = @_;
    $self->{manager}->wait_finish($self->{uid});
    return;
}

=head2 is_busy

Returns '1' if the current asynchronous workload is still working and '0' if it has
stopped. You should then call 'wait' to get a description of what was
done during the workload

=over

=item Arguments

None

=item Examples

    $dataset->write(random=>1, io_size=>'4k', async=>1, amount=>'4g');
    while ($dataset->is_busy){
        ... do something ...
    }
    $ret = $dataset->wait;
    Print "Duration: $ret->{duration}->{total} \n";

=back

=cut

sub is_busy {
    my $self = shift;
    my $command = $self->get_active_command;
    if (!$command){
        $Log->debug("Subroutine 'is_busy' was invoked but no code was executing");
        return 0;
    }
    return $self->{manager}->is_busy($self->{uid});
}

sub _get_actual_sizes {
    my $self = shift;
    my %args = @_;
    my $type        = $args{type};
    my $client = $self->get_clients->[0];
    my $host_obj = Perf::Depot->get_resource( class=>'Host', name=>$client);
    my $dataset_resource = Perf::Depot->get_resource(class => 'DataSet',
                                                        name=>$self->get_uid);
    confess "No Perf::DataSet found?!" if (!$dataset_resource);
    my $client_workers = Perf::Depot->get_resources( class=>'Worker',
                                                        host=>$host_obj,
                                                        used_by=>$dataset_resource);
    my $chosen_worker = $client_workers->[0];
    my $getactualsizes_command = $self->create_command('GetActualSizes', element_type => $type);
    $getactualsizes_command->pre_execute($self);
    return $chosen_worker->execute(command=>$getactualsizes_command, async=>0);
}

sub validate_container {
    my $self = shift;

    # TODO: Validate directories!

    #
    # Verify the size of each file
    #
    my $files = $self->{container}->get_elements('files');
    my $actual_file_sizes = $self->_get_actual_sizes(type        => 'files');
    # Verify file sizes against template sizes
    for my $path (keys %$files){
        my $size = $actual_file_sizes->{$path};
        if (!defined($size)){
            confess "File '$path' missing from return output!\n" .
                    as_string($actual_file_sizes);
        }
        my $container_file = $files->{$path};
        my $container_file_size = $container_file->get_size;
        if ($container_file->is_deleted){
            if ($Perf::DataSet::Element::NOT_FOUND == $size) {
                $Log->comment("File '$path' was properly detected as deleted")
            }
            elsif ( $container_file->is_invalid ){
                $Log->comment("File '$path' not detected as deleted but is invalid");
            }
            else{
                confess "File '$path' should not have been detected on client!";
            }
        }
        else{
            if ($container_file_size != $size){
                my $msg = (
                    $Perf::DataSet::Element::NOT_FOUND == $size
                    ? "File '$path' was not found on disk"
                    : "File '$path' is not of expected size"
                      . " ($size instead of $container_file_size)"
                    );

                if ($container_file->is_invalid){
                    $Log->comment("$msg but is invalid");
                }
                else{
                    confess $msg;
                }
            }
        }
    }
}

sub is_mounted {
    my $self = shift;

    return all { $_->is_latched() } @{$self->get_mountpoints()};
}

# Return filer name
sub get_filer{
    my $self = shift;
    return $self->{filer};
}
# Return node name
sub get_node{
    my $self = shift;
    return $self->{filer};
}
# Return appliance name
sub get_appliance{
    my $self = shift;
    return $self->{appliance};
}
# Return vserver name
sub get_vserver{
    my $self = shift;

    $self->{vserver} = $self->{appliance}
        if (!$self->{vserver});

    return $self->{vserver};
}
sub get_fc{
    my $self = shift;
    return $self->{fc};
}


# Get an arrayref of all the node CommandInterface objects this DataSet
# is using. $self->{cis} is maintained only when DataSet is created with
# a nodes argument instead of a single filer argument and contains
# CommandInterfaces for all the nodes passed. $self->{fc} contains a
# single CommandInterface and is maintained at all times: when the filer
# arg is used and for backward compatibility when the nodes arg is used
sub get_command_interfaces {
    my $self = shift();

    return [
        defined($self->{cis})
        ? values(%{$self->{cis}})
        : $self->{fc}
    ];
}


sub get_volume{
    my $self = shift;
    return $self->{vol};
}
sub get_manager {
    my $self = shift;
    return $self->{manager};
}
sub get_processes{
    my $self = shift;
    my $processes_count = $self->{processes};
    if (!$processes_count){
        #$processes_count = scalar(keys %{$self->{manager}->get_processes}) / scalar(@{$self->get_clients});
        $processes_count = $self->{manager}->get_process_count_per_client;
    }
    return $processes_count;
}
sub get_localhost{
    my $self = shift;
    return $self->{manager}->get_localhost;
}

=head2 get_lun_paths

Returns a list of paths for all the LUNs managed by a Perf::DataSet.

LUN paths will be in the form: /vol/<volume name>/<qtree name>/<LUN name>

=over

=item Examples

 my @lun_paths = $dataset->get_lun_paths();
 print "LUN path is: $lun_paths[0]\n";

 LUN path is: /vol/vl04218_de91_cf6fdfc4/4187206221_739460792_perfqa-linux-5

=back

=cut

sub get_lun_paths {
    my $self = shift();

    confess('Invalid method for Perf::DataSet type: ' . $self->get_type())
        if (!$self->is_san());

    my $dataset_resource =
      Perf::Depot->get_resource(class => 'DataSet', name => $self->{uid});

    return map { $_->get_path() } @{$dataset_resource->get_lun_resources()};
}

sub is_read_only {
    my $self = shift;
    return $self->{read_only};
}
sub get_no_purge_on_exit {
    my $self = shift;
    return $self->{no_purge_on_exit};
}
sub get_snapshot {
    my $self = shift;
    return $self->{snapshot};
}
sub get_qtree {
    my $self = shift;
    return $self->{qtree};
}
sub get_nodes {
    my $self = shift;
    return $self->{nodes};
}
sub get_type{
    my $self = shift;
    return $self->{type};
}

sub get_lifs {
	my $self = shift();

	return $self->{lifs};
}

sub _validate_type {
    my $type = shift();

    if (   $type ne 'nfs'
        && $type !~ 'nfsv4(?:\.1)?'
        && $type !~ /^cifs(?:-smb(?:[13]|2(?:\.1)?))?$/
        && $type ne 'iscsi'
        && $type ne 'fcp') {
        confess "type: $type not supported";
    }
}

sub _set_type {
    my $self     = shift();
    my $new_type = shift();

    _validate_type($new_type);

    $self->{type} = $new_type;
}

sub get_smb_version {
    my $self = shift();
    (my $ver = $self->get_type()) =~ s/^cifs-smb([[:digit:].]+)$/$1/;

    return $ver;
}

sub get_parent_uid{
    my $self = shift;
    return $self->{parent_uid};
}
sub get_lun_prefix{
    my $self = shift;
    return $self->{lun_prefix};
}
sub get_root_dir{
    my $self = shift;
    return $self->{root_dir};
}
sub is_file_slicing_allowed{
    my $self = shift;
    return $self->{allow_file_slicing};
}
sub is_file_slicing_single_client{
    my $self = shift;
    return $self->{single_client_file_slicing};
}

sub get_pwd {
    my $self = shift;
    return $self->{manager}->get_pwd;
}
sub get_clients{
    my $self = shift;
    # Return a copy of the array of clients
    my $c = $self->{clients};
    if (!$c){
        $c = $self->{manager}->get_clients($self->get_uid);
        my @clients = map { $_->get_name } @$c;
        return \@clients;
    }
    else{
        return $c;
    }
}
sub get_uid{
    my $self = shift;
    return $self->{uid};
}
sub get_seed{
    my $self = shift;
    return $self->{seed};
}
sub get_container{
    my $self = shift;
    return $self->{container};
}

sub get_active_command {
    my $self = shift;
    return $self->{manager}->get_active_command($self->{uid});
}

sub is_san {
    my $self = shift();
    return ('iscsi' eq $self->{type} || 'fcp' eq $self->{type});
}

sub is_nas {
    my $self = shift();
    return ($self->{type} =~ m{^(?:nfs|cifs)}msx);
}


sub enable_multipathing {
    my $self = shift();
    return $self->{enable_multipathing};
}


sub enable_alua {
    my $self = shift();
    return $self->{enable_alua};
}


sub use_all_nodes {
    my $self = shift();

    return $self->{all_nodes_in_cluster};
}


sub disable_oplocks {
	my $self = shift();

	return $self->{disable_oplocks};
}


# Add information about an Addition to this Perf::DataSet
# $dataset->insert_addition_info($addition_uid, $addition_size);
sub insert_addition_info {
    my ($self, $addition_uid, $addition_size) = @_;
    $self->{additions}->{$addition_uid} = $addition_size;
}

# Retrieve the hash of Addition UIDs to Addition sizes
sub get_addition_info {
    my ($self) = @_;
    return $self->{additions};
}

# Clear out information about this Perf::DataSet's Additions
sub clear_addition_info {
    my ($self) = @_;
    $self->{additions} = { };
}

# Generate a list of LUNs which should be present in this Perf::DataSet
# ex: $dataset->get_lun_names();
#
# will filter based on Addition UID or hostname if provided:
# ex: $dataset->get_lun_names(addition_uid => ...,
#                             host         => ...,);
sub get_lun_names {
    my $self = shift();
    my %param = @_;
    my $lun_names = [ ];

    my @additions =
        (defined($param{addition_uid}))
      ? ($param{addition_uid})
      : keys(%{$self->{additions}});
    my $hosts =
        (defined($param{host}))
      ? [$param{host}]
      : $self->{clients};

    foreach my $add_uid (@additions) {
        foreach my $host (@$hosts) {
            my $lun_name = $self->make_lun_name(addition_uid => $add_uid,
                                                client_name  => $host,);
            push(@$lun_names, $lun_name);
        }
    }

    return $lun_names;
}

# create a name for a LUN
sub make_lun_name {
    my $self = shift();
    my %params = validate_mixed(
           @_,
           [ ],
           {
               # UID of the addition which is causing this
               # LUN to be created
               addition_uid => {type => SCALAR},

               # Name of the client that will map to this LUN
               client_name  => {type => SCALAR},
            }
       );

    return $self->{lun_prefix}
           . '_'
           . $params{addition_uid}
           . '_'
           . $params{client_name};
}

sub _get_uid {
    return int(rand(2**32));
}

=head2 persist_container

Stores to disk the state of the data (i.e, file paths and checksum data) this Perf::DataSet instance
knows about.

This is useful if creating your Perf::DataSet is expensive and you plan on keeping around the volume
containing it between invocations of "ntest".

Limitation: Loading a container is currently only possible with a Perf::DataSet that contains
the same number of processes as the Perf::DataSet that persisted it.

Note: You should consider taking a snapshot of your volume or aggregate at the same time you call
this method.

=over

=item Arguments

B<to> = string

I<Required> Full path to directory into which data state files will be persisted.
The $to path must be accessible from all clients and the machine executing ntest.

B<name> = string

I<Required> Name of the persisted container. Creates or overwrites symlink $name under $to pointing
to the latest persisted container files the name $name. Old persisted containers called $name are
not deleted.

=item Examples

    $PERSIST_PATH = '/u/username/datasets';
    $PERSIST_NAME = 'my_feature';
    $ROOT_DIR = $PERSIST_NAME;

    # In test 1
    my $dataset = Perf::DataSet->new(..., root_dir => $ROOT_DIR );
    $dataset->persist_container( to => $PERSIST_PATH, name => $PERSIST_NAME );

    # In test 2
    my $dataset = Perf::DataSet->new( ...., root_dir => $ROOT_DIR,
                                      load_container_from => $PERSIST_PATH,
                                      load_container_name => $PERSIST_NAME );

=back

=cut

sub persist_container {
    my $self = shift();
    my %params = validate_mixed(
           @_,
           [ ],
           {
               to    => {type => SCALAR},
               name  => {type => SCALAR},
            }
       );
    my $path = $params{to};
    # Sanitize
    $path =~ s/^\s+//;
    $path =~ s/\s+$//;
    $path =~ s/\/+$//;
    confess "Invalid path!" if (!$path);
    my $root_path = $path;
    my $name = $params{name};
    $path .= "/hist/$name/$self->{uid}";

    $Log->comment("Creating '$path'");
    my $success = eval {mkpath($path);1};
    if (!$success){
        confess  "Unable to create directory '$path':$@ $!";
    }

    # There are containers to persist both in $self and and each Worker
    # associated with this instance of Perf::DataSet.

    # 'clone' does some cleanup on the container (possibly unnecessary here)
    my $freeze = { container=> $self->{container}->clone };
    my $ret;
    eval {$ret = Storable::nstore $freeze, "$path/dataset"};
    if (!$ret || $@){
        confess "Unable to freeze container of Perf::DataSet '$self->{uid}': $@";
    }

    # TODO: If the container was sliced with single_client_file_slicing or if the
    # Perf::DataSet is SAN, then it'll be important to keep track of which containers belong
    # to a same client. Currently all containers will be dumped in the 'workers' folder
    $self->{manager}->freeze_workers(   $self->{uid},
                                        path    => "$path/workers",
                                    );
    # Create symlink
    my $symlink_path = "$root_path/$name";
    if ( -l $symlink_path){
        unlink($symlink_path) or confess "Unable to remove symlink '$symlink_path':$!";
    }
    symlink($path, $symlink_path) or confess "Unable to create symlink '$symlink_path': $!";

    return;
}

sub _load_container_from {
    my ($self,%args) = @_;
    my $path = $args{path};
    my $name = $args{name};
    confess "You must pass a 'path' argument" if (!$path);
    confess "You must pass a 'name' argument" if (!$name);
    $path = join('/',$path,$name);
    confess "Path '$path' is not available" if (! -e $path );
    confess "Path '$path' must be, or point to, a directory" if (! -d _ );

    # A container is currently split in several files, one for the
    # main Perf::DataSet instance and one per Worker instances
    # This means that in order to load a container the number of workers in this Perf::DataSet
    # must match the number of files to load
    # TODO: Allow container loading with a different number of workers

    my $dataset_container_path = "$path/dataset";
    confess "Missing dataset container file" if ( ! -f $dataset_container_path );
    $Log->comment("Thawing container data from '$dataset_container_path'");
    my $thaw;
    eval{$thaw = Storable::retrieve($dataset_container_path)};
    if (!defined($thaw) || $@){
        confess "Unable to load file '$dataset_container_path': $@";
    }
    $self->{container} = $thaw->{container};

    my $worker_containers_path = "$path/workers";
    $self->{manager}->thaw_workers( $self->{uid}, path=>$worker_containers_path, no_unlink=>1 );

    return;
}


sub start_sysstat {
    my $self        = shift();
    my $postfix     = shift();
    my $num_clients = scalar(@{$self->{clients}});
    my @nodes;
    my @sysstats;

    if (defined($self->{nodes})) {
        @nodes = @{$self->{nodes}};
    }
    else {
        @nodes = ($self->{filer});
    }

    foreach my $i (0 .. $#nodes) {
        my $node = $nodes[$i];
        my $sysstat = Perf::Sysstat->new(
                                node   => $node,
                                client => $self->{clients}->[$i % $num_clients],
                                options => '-x',
                                runid   => "${node}_$postfix",);

        $sysstat->start();

        push(@sysstats, $sysstat);

    }

    $self->{sysstats} = \@sysstats;
}


sub stop_sysstat {
    my $self = shift();

    while (0 < scalar(@{$self->{sysstats}})) {
        my $sysstat = pop(@{$self->{sysstats}});

        $Log->comment("Stopping sysstat job");
        $sysstat->get_output();

        $sysstat->stop();
    }
}


# handle a connection object being passed in for filer
sub handle_filer_connection {
    my $dataset = shift();

    my $filer_ref = ref($dataset->{filer});
    if ($filer_ref){
        $dataset->{fc} = $dataset->{filer};
        if ($dataset->{filer}->isa('NACL::C::SystemNode')) {
            $dataset->{filer} = $dataset->{fc}->hostrec()->hostname();
            $dataset->{cis}->{$dataset->{filer}} = $dataset->{fc};
        }
        elsif ($dataset->{fc}->isa('Connectrec')) {
            $Log->warn(
                  "Passing Connectrec objects to DataSet has been deprecated.\n"
                    . "Please use appropriate NACL objects instead.");
            my $hostp = $dataset->{fc}->hostp();
            $dataset->{filer} = $hostp->hostname();
            $dataset->{cis}->{$dataset->{filer}} = NACL::STask::Node->new(hostrec => $hostp);
            $dataset->{fc} = $dataset->{cis}->{$dataset->{filer}};
        }
        else {
            confess("Unsupported type for argument 'filer => $filer_ref'");
        }
    }
}


=head2 reconnect_nodes

Reestablish connections to filers after a reboot, takeover-give back,
etc.

=cut

sub reconnect_nodes {
    my $self             = shift();
    my $dataset_resource = Perf::Depot->get_resource(
        class => 'DataSet',
        name  => $self->{uid});
    my $appliance = $dataset_resource->get_appliance_resource();
    $appliance->reconnect();
}


sub _same_lif_context {
	my $self = shift();
	my $fork = shift();

	confess('LIF forks not supported in 7-mode')
		if ($self->{fc}->hostrec()->hosttype() !~ /-ng-/);

	return 0
		if ($self->{vserver} ne $fork->{vserver});

	my $dataset_rsc = Perf::Depot->get_resource(
		class => 'DataSet',
		name => $self->{uid},);

	my $all_nodes = $dataset_rsc
		->get_appliance_resource()
		->get_cluster()
		->get_node_names();

	foreach my $fork_node (@{$fork->{nodes}}) {
		# fork node not found in current node list, must be a
		# new cluster
		return 0
			if (none { $fork_node eq $_ } @$all_nodes);
	}

	return 1;
}


=head2 set_execution_node

Changes the target node for test commands

=over

=item Arguments

I<Required> Name of the execution node

=back

=cut

sub set_execution_node {
	my $self = shift();
	my $node_name = shift();

	my $node = Perf::Depot->get_resource(
		class => 'Node',
		create_on_miss => 1,
		name => $node_name,);

	$node->get_cluster()->set_execution_node($node);
}


sub _get_dataset_resource {
	my $self = shift();

	return Perf::Depot->get_resource(
		class => 'DataSet',
		name => $self->{uid},
	);
}


=head2 dump_diagnostics

Execute commands in an attempt to diagnose common failures

=cut

sub dump_diagnostics {
	my $self = shift();
	my $cis = $self->get_command_interfaces();

	foreach my $ci (@$cis) {
		$ci->apiset(
			interface => 'CLI',
			set => 'Nodescope',
		)->wafl_cpstats();
	}

	$self->_get_dataset_resource()
		->get_volume_resource()
		->dump_size_usage();
}


sub DESTROY {
    my $self = shift;
    #
    # Skip the destruction for subtests
    #
    return unless ($$ == $self->{pid});

    #
    # It is possible that we error out during destroy itself.
    # This will cause a second call to destroy
    #
    if ($self->{being_destroyed}){
        $Log->comment("Destroy is already taking place, skipping. Current error: $@");
        return;
    }
    else{
        $self->{being_destroyed} = 1;
    }

    #
    # The Manager will be destroyed first usually and will
    # perform all the cleanup
    #
    if (!$self->{manager}){
        return;
    }

    my $current_error = $@;
    if ($current_error){
        $Log->comment("DESTROY sequence initiated on Perf::DataSet '" . $self->{uid}
                . ".' because of following:\n$current_error");

        # Do not shutdown as user may want to continue running with a fork

        #eval{
        #    if ($self->{active}){
        #        $self->abort;
        #    }
        #    $self->{manager}->shutdown;
        #};
    }
    else{
        $Log->comment("DESTROY sequence initiated on Perf::DataSet '" . $self->{uid}
                . ".' because it fell out of scope");
    }
    eval{
        if ($self->{active}){
            $self->abort;
        }
        $self->{manager}->unregister($self->{uid});
    };
    if ($@){
        $Log->comment("Errors during DESTROY of Perf::DataSet '$self->{uid}': $@");
    }
    $@ = $current_error;
    return;
}


# Transitional code
package DataSet;

use Tharn qw($Log);

use base qw(Perf::DataSet);

sub new {
    my $class = shift();

    $Log->warn(  "Use of DataSet->new() has been deprecated.\n"
               . "Please use Perf::DataSet->new() instead.");

    my $self = $class->SUPER::new(@_);

    return $self;
}


1;
