# -*- perl -*-
#
## @author   dl-perf-qa-automation@netapp.com
## @summary
## @status   private
##
## @pod here

package Perf::DataSet::Addition;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log subtest);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Perf::Utils qw(get_random_element as_string in_bytes get_file_size ceil);
use IO::Seekable;
use Time::HiRes qw(time);
use File::Basename qw(basename);
use Perf::DataSet::Container;
use Perf::DataSet::SparserFactory;
use File::Path qw(rmtree mkpath);
use Fcntl;
use List::Util qw(sum min max);

use Storable qw();
use MIME::Base64;
use Perf::DataSet::Element;
use Perf::DataSet::Element::Directory;
use Perf::DataSet::Element::File;

use namespace::clean;
use Perf::DataSet::Command;
# Addition is a subclass of Command
use base qw(Perf::DataSet::Command);
use Perf::DataSet::Addition::NAS;
use Perf::DataSet::Addition::SAN;


#
# Define some WAFL constants
#
my $BLOCKSIZE = 4096;

# Use to generate random strings
my @alphabet = ('a','b','c','d','e','f','g','h','i','j','k','l','m',
        'n','o','p','q','r','s','t','u','v','w','x','y','z',
        'A','B','C','D','E','F','G','H','I','J','K','L','M',
        'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
        '0','1','2','3','4','5','6','7','8','9');
my $alphabet_count = 62;
use constant GENERATE_COMPLETE => -1;

sub validate {
    my $class = shift;
    my %params = Params::Validate::validate(
           @_,
           {
            description => {type => SCALAR | UNDEF, optional => 1, default => "add"},
            uid => {type => SCALAR | UNDEF, optional => 1, default => int(rand(2**32))},
            # root_dir: where to start under the mountpoint
            root_dir => {type => SCALAR | UNDEF, optional => 1, default => '.'},
            # size: overall size of what needs to be added
            size => {type => SCALAR },
            file_name_length => {type => SCALAR | UNDEF, optional => 1, default => 28},
            # compute_checksum: you may choose to compute and store the checksum of each 'block'
            compute_checksum => {type=> SCALAR | UNDEF, optional => 1, default => 1},
            # block_size: Length of pattern. Must be a multiple of BLOCKSIZE
            block_size => {type => SCALAR| UNDEF, optional => 1, default => "4k"},
            # buffer_size: How often do we write to disk
            buffer_size => {type => SCALAR| UNDEF, optional => 1, default => "5m"},
            # buffer_align: Boundaries to align buffer to
            buffer_align => {type => SCALAR| UNDEF, optional => 1, default => 512},
            # buffer_sleep: How much do we sleep between write operation in seconds
            #buffer_sleep => {type => SCALAR | UNDEF, optional => 1, default => 0}, #DEPRECATED
            # max_tput: Maximum throughput in MB/s. 0 means no max.
            max_tput => {type => SCALAR | UNDEF, optional => 1, default => 0},
            # type: supports 'nfs'
            type => {type => SCALAR | UNDEF, optional => 1, default => 'nfs'},
            # async: For asynchronous loads
            async => {type => SCALAR | UNDEF, optional => 1, default => 0},
            # tree_structure: list of arrays each describing number of objects for a specific level
            # in the folder tree
            # [number of files, number of hardlinks, number of symlinks, number of directories]
            # Last line must indicate 0 directories
            tree_structure => {type => ARRAYREF | UNDEF, optional => 1, default => [[5, 0, 0, 20],
                                                [10, 0, 0, 20],
                                                [10, 0, 0, 5],
                                                [5, 0, 0, 0]]},
            # start_offset for large 'empty' files
            #start_offset => {type => SCALAR| UNDEF, optional => 1, default => 0},
            # file or files with many holes
            sparse       => {type => SCALAR, optional => 1, default => 0},
            sparse_ratio => {type => SCALAR, optional => 1, default => 1},
            sparse_level => {type => SCALAR, optional => 1, default => 0},
            sparse_align => {type => SCALAR, optional => 1, default => 'right'},
            # Boundaries and Distribution:
            # In a nutshell this script will attempt to create $DISTRIBUTIONS[$i] percent of files
            #  between $BOUNDARIES[$i] Bytes and $BOUNDARIES[$i + 1] Bytes
            # Distribution must add up to 100%
            size_boundaries => {type => ARRAYREF | UNDEF, optional => 1, default =>
                    ["0B","64B","32k","2036k","2040k","2044k","16m","8g"]},
            size_distribution => {type => ARRAYREF | UNDEF, optional => 1, default =>
                    [1,3,10,4,4,30,48]},
            # If the lowest size boundaries is 0, this will allow dataset to make a small file (size)
            # or it will allow the small file not to be created making a smaller dataset size (size_boundaries)
            file_size_priority => {type => SCALAR, optional => 1, default => 'size'},
            debug => {type => SCALAR | UNDEF, optional => 1, default => 0},
            # skip_flush: don't sync filers before returning
            skip_flush  => {type => SCALAR | UNDEF, optional => 1, default => 0},
           }
           );

    # Make a copy of these arrays so as not to modify the input arrays later on
    my @size_boundaries_in_bytes = in_bytes($params{size_boundaries});
    $params{size_boundaries} = \@size_boundaries_in_bytes;
    $params{original_size_boundaries} = $params{size_boundaries};
    # Make a copy of these arrays so as not to modify the input arrays later on
    $params{original_size_distribution} = $params{size_distribution};
    my @sd_arry = @{$params{size_distribution}};
    my @sb_arry = @{$params{size_boundaries}};
    $params{size_boundaries} = \@sb_arry;
    $params{size_distribution} = \@sd_arry;

    if ($params{buffer_sleep} && $params{max_tput}){
        confess "You cannot specify both buffer_sleep and max_tput";
    }
    # We need at least two items in the boundaries array
    if (@{$params{size_boundaries}} < 2){
         confess "You must specify at least two items in size_boundaries: \n" . as_string($params{size_boundaries});
    }

    if (scalar(@sd_arry) >= scalar(@sb_arry)) {
        confess('Too many elements in size_distribution parameter');
    }

    $params{size} = in_bytes($params{size});
    $params{block_size} = in_bytes($params{block_size});
    $params{buffer_size} = in_bytes($params{buffer_size});

    if ($params{block_size} % $BLOCKSIZE != 0){
        confess "block size ($params{block_size}) must be a factor of $BLOCKSIZE";
    }
    if ($BLOCKSIZE % $params{buffer_align} != 0){
        confess "buffer_align ($params{buffer_align}) must be a divisor of $BLOCKSIZE";
    }
    if ($params{buffer_size} % $params{buffer_align} != 0 ){
        confess "buffer_size must be a multiple of buffer_align ($params{buffer_align})";
    }
    # Check the size distribution adds to to 100%
    my $hundred_percent = 0;
    for my $p (@{$params{size_distribution}}){
        $hundred_percent += $p;
    }
    if ($hundred_percent != 100){
        confess "Your distribution percentages must add up to 100%: \n" . as_string($params{size_distribution});
    }
    if ($params{file_size_priority} !~ /^size$|^size_boundaries$/) {
        confess "file_size_priority must be either 'size' or 'size_boundaries'";
    }

    # Check boundaries go in increasing order
    my $previous=0;
    for my $boundary (@{$params{size_boundaries}}){
        if ($boundary<$previous){
            confess "Your boundaries must go in increasing order";
        }
        else{
            $previous = $boundary;
        }
    }

    # Check sparse is between 0 and 1
    if (exists $params{sparse} && ( $params{sparse} < 0 || $params{sparse} > 1)){
        confess "Parameter 'sparse' must be a value between 0 and 1";
    }

    # Check sparse_level is between 0 and 7
    if (exists $params{sparse_level} &&
            ( $params{sparse_level} != int($params{sparse_level}) ||
              $params{sparse_level} < 0 ||
              $params{sparse_level} > 7)){
        confess "Parameter 'sparse_level' must be an integer between 0 and 6";
    }

    return %params;
}

sub new {
    my $class = shift;
    my %params = @_;

    if ($class eq __PACKAGE__) {
        if ($params{type} =~ /nfs|cifs/) {
            return Perf::DataSet::Addition::NAS->new(%params);
        }
        elsif ('iscsi' eq $params{type} || 'fcp' eq $params{type}) {
            return Perf::DataSet::Addition::SAN->new(%params);
        }
        else {
            confess("Type ($params{type}) not supported");
        }
    }

    %params = $class->validate(@_);

    my $self = \%params;
    bless $self, $class;
    return $self;
}


#
# Input: size of file
# Returns: range index, lower bound, upper bound
#
sub get_range($$){
    my $self = shift;
    my $size = shift;
    my %args = @_;
    my $boundaries = $args{boundaries} || $self->{size_boundaries};

    my $range_format = "%04d";
    my $i = 0;

    my $last_boundary_idx = $#{$boundaries};
    my $last_boundary = in_bytes($boundaries->[$last_boundary_idx]);

    if ($size > $last_boundary) {
        #return (sprintf($range_format,$last_boundary_idx), $boundaries->[$#{$boundaries}], $size);
        return undef;
    }
    while (in_bytes($boundaries->[$i + 1]) < $size){
        $i++;
    }
    if (wantarray){
        return (sprintf($range_format,$i), $boundaries->[$i], $boundaries->[$i + 1]);
    }
    else{
        return $i;
    }
}


sub generate_report {

    my $self = shift;
    my $container = shift;
    my $files = $container->get_elements('files');
    my $size_distribution = $self->{original_size_distribution};

    my %report;

    # Handle subdirectories

    # Handle links

    # Handle files
    foreach my $file (keys(%{$files})){
        my $file_size = $files->{$file}->get_size;
        my $bytes_written = $files->{$file}->get_size_on_disk;
        my ($range, $minsize, $maxsize) = $self->get_range($file_size,
                                                           boundaries=>$self->{original_size_boundaries}
                                                           );
        $report{files}{ranges}{$range}{"Interval"}="[$minsize-$maxsize]";
        $report{files}{ranges}{$range}{"Number of files"} += 1;
        $report{files}{ranges}{$range}{"Size"} += $file_size;
        $report{files}{ranges}{$range}{"Size on disk"} += $bytes_written;
        $report{files}{ranges}{$range}{"Number of sparse files"} += $files->{$file}->get_sparse;
        if ($report{files}{ranges}{$range}{"Max. Size"}){
            if ($file_size > $report{files}{ranges}{$range}{"Max. Size"}){
                $report{files}{ranges}{$range}{"Max. Size"} = $file_size;
            }
        }
        else{
            $report{files}{ranges}{$range}{"Max. Size"} = $file_size;
        }
        if ($report{files}{ranges}{$range}{"Min. Size"}){
            if ($file_size < $report{files}{ranges}{$range}{"Min. Size"}){
                $report{files}{ranges}{$range}{"Min. Size"} = $file_size;
            }
        }
        else{
            $report{files}{ranges}{$range}{"Min. Size"} = $file_size;
        }

        $report{files}{total}{"Number of files"} += 1;
        $report{files}{total}{"Number of sparse files"} += $files->{$file}->get_sparse;
        $report{files}{total}{"Size"} += $file_size;
        $report{files}{total}{"Size on disk"} += $bytes_written;
    }

    foreach my $range (keys(%{$report{files}{ranges}})){
        $report{files}{ranges}{$range}{"Pct Files created"} =
            int($report{files}{ranges}{$range}{"Number of files"} / $report{files}{total}{"Number of files"}
            * 100) . "%";
        $report{files}{ranges}{$range}{"Pct Expected"} = $size_distribution->[$range] . "%";
    }
    return \%report;
}


sub pre_execute($$) {
    my $self = shift;
    my $owner = shift;
    if (!$owner || !ref($owner)){
        confess "You must specify an owner";
    }

    if ($owner->isa('Perf::DataSet::Worker')){
        #
        # Make sure we are going to know how to align the buffers for
        # directio
        #
        my $length_p = length(pack "p", "foo");
        my $length_J = length(pack "J", 0);
        confess "Platform not supported ( p: $length_p, J: $length_J )" if ( $length_p != $length_J );
        my $buffer_align = $self->{buffer_align};
        my $buffer_unit = 'x' x $buffer_align;
        my $standard_buffer = $buffer_unit x ( $self->{buffer_size} / $buffer_align + 1 );
        $self->{standard_buffer} = \$standard_buffer;
        my $align_offset = unpack( "J", pack "p", $standard_buffer ) % $buffer_align;
        $self->{align_offset} = $align_offset? ($buffer_align - $align_offset) : 0;

        #
        # Grab template corresponding to this Worker
        #
        my $process_uid = $owner->get_process_uid;
        $self->{container} = $self->{worker_templates}->{$process_uid};
        if (!$self->{container}){
            confess "Unable to find Worker template for process '$process_uid'";
        }
        else{
            # Reset on_dirty with valid coderef
            $self->{container}->add_on_dirty_to_all;
            $Log->debug("Addition template\n" . $self->{container}->as_string);
        }
        delete $self->{worker_templates};

        my $full_path = $self->{worker_full_paths}->{$process_uid};
        confess "No full path found!" if (!$full_path);
        $owner->{full_path} = $full_path;

        $Log->comment("Changing to directory '$full_path'");
        chdir $full_path or confess "Unable to chdir to '$full_path': $!";

    }
    elsif($owner->isa('Perf::DataSet')){
        # This is a hack to help with test_suite.thpl
        return if ($self->{dataset_pre_executed});

        if ($owner->is_read_only){
            confess "You cannot add into a read-only Perf::DataSet";
        }

        # Tell the owner about this addition so the proceeding call to
        # mount can create raw devices for us.
        $owner->insert_addition_info($self->{uid}, $self->{size});

        # Mount all the workers and store the mountpoints in the
        # Addition.  This needs to take place before generate()
        # because mount() causes SAN Managers to create LUNs and more
        # importantly the raw devices whose names we need.
        $owner->mount();

        my $dataset = Perf::Depot->get_resource(class=>'DataSet', name=>$owner->get_uid);
        if ($self->{sparse}){
            $self->{sparser_factory} = Perf::DataSet::SparserFactory->new(vol => $dataset->get_volume_resource);
        }

        #
        # Generate template for Addition
        #
        my $dataset_template = $self->generate($dataset);
        #$self->{container} = $dataset_template;
        # We don't need to keep all the additions, just the latest template for
        # later verification in post_execute
        $owner->{addition_template} = $dataset_template->clone;
        my $report = $self->generate_report($dataset_template);
        $Log->debug("Template dataset for '$self->{description}':\n" . as_string($report));
        #
        # Divide the work among Workers
        #
        my $processes = $owner->get_manager->get_processes($owner->{uid});
        my $workers_total_count = scalar @$processes;
        $self->{processes} = $workers_total_count;

        my %more_args;
        if ($owner->is_file_slicing_allowed){
            $more_args{file_slicing} = 1;
            if ($owner->is_file_slicing_single_client){
                $more_args{groups} = scalar @{$owner->get_clients};
            }
        }
        else{
            $more_args{file_slicing} = 0;
        }

        my $partitions = $dataset_template->partition($workers_total_count,
                                                      {%more_args},
                                                      );

        my %worker_templates =
          $self->distribute_partitions(partitions => $partitions,
                                       owner      => $owner,
                                       processes  => $processes);

        $self->{worker_templates} = \%worker_templates;

        # Set buffer_sleep to respect max_tput
        $self->adjust_buffer_sleep;

        # Embed the worker paths to the command so that workers can load their own
        $self->{worker_full_paths} = $owner->get_worker_full_paths;

        # This is hack to help with test_suite.thpl
        $self->{dataset_pre_executed} = 1;
    }
    else{
        confess "The owner must be a Perf::DataSet or Perf::DataSet::Worker and not a " . ref($owner);
    }
}


sub adjust_buffer_sleep {
    my $self = shift;
    if ( $self->{max_tput}){
        my $worker_max_tput = $self->{max_tput} / $self->{processes};
        $self->{buffer_sleep} = $self->{buffer_size}/($worker_max_tput * 1024 * 1024);
    }
    else{
        $self->{buffer_sleep} = 0;
    }
    $Log->comment("Set buffer_sleep to $self->{buffer_sleep} seconds");
}


# Create and fill a Container with Elements
sub generate {
    confess('This method must be implemented in a subclass of ' . __PACKAGE__);
}


sub execute($) {
    my $self = shift;
    # Create the files for real
    $self->create();

    my $return = $self->{container}->get_stats->as_hash;

    #
    # We don't want to return the whole container but we still need
    # to pass the size_on_disk of each created file
    #
    my $files = $self->{container}->get_elements('files');
    map { $return->{size_on_disk}->{individual}->{$_->get_path} += $_->get_size_on_disk } values %$files;

    return $return;
}


sub post_execute($$$){
    my $self = shift;
    my $owner = shift;
    if (!$owner || !ref($owner)){
        confess "You must specify an owner";
    }
    my $returns = shift;
    if ($owner->isa('Perf::DataSet::Worker')){
        # Merge the additional files to the current Worker container
        $owner->{container}->merge_with($self->{container});

        return $returns;
    }
    elsif($owner->isa('Perf::DataSet')){
        # Make sure all data is written to disk
        if (!$self->{skip_flush}){
            $Log->comment("Flushing appliance from " . __PACKAGE__ );
            my $appliance = Perf::Depot->get_resource(class=>'Appliance', name=>$owner->get_appliance);
            confess "No appliance found?!" if (!$appliance);
            $appliance->flush;
        }

        # Merge the Addition container with the Perf::DataSet container
        if ($self->{debug}){
            $Log->debug("Merging Perf::DataSet container with Addition container");
        }
        # We stored the Addition container at this location during pre-execute
        my $template = $owner->{addition_template} ;
        $owner->{container}->merge_with($template);

        #if ($self->{debug}){
        #    $Log->comment("Addition Stats per process:\n" . as_string($returns));
        #}
        #
        # Add up all the stats in returns
        #
        my $merged_returns;
        for my $process_uid (keys %$returns){
            my $stats = $returns->{$process_uid};
            for my $s (keys %$stats){
                my $e = $stats->{$s};
                if (ref($e) eq 'HASH'){
                    for my $k (keys %$e){
                        my $v = $e->{$k};
                        if (ref($v) eq 'HASH'){
                            map { $merged_returns->{$s}->{$k}->{$_} += $v->{$_} } keys %$v;
                        }
                        else{
                            $merged_returns->{$s}->{$k} += $v;
                        }
                    }
                }
                else{
                    $merged_returns->{$s} += $e;
                }
            }
        }
        #$Log->comment( "returns:\n" . as_string($returns) );
        #$Log->comment( "merged_returns:\n" . as_string($merged_returns) );

        $Log->debug( "returns:\n" . as_string($returns) );

        #
        # Get the size of each directory
        # So far we assume that all clients can see the entire dataset so running this on
        # any client will work. Multi-client support for iscsi will mean breaking that
        # assumption
        #
        #my $template = $owner->{addition_template} ;
        my $dirs = $template->get_elements('dirs');
        my $actual_dir_sizes = $owner->_get_actual_sizes(type => 'dirs');

        # Merge dir sizes into owner and self container
        for my $path (keys %$dirs){
            my $size = $actual_dir_sizes->{$path};
            if (! defined($size)){
                confess "Directory '$path' missing from return output!\n" .
                        as_string($actual_dir_sizes);
            }

            if ($owner->exists($path)){
                my $size_delta = $size - $owner->size_of($path);
                my $addition_dir = Perf::DataSet::Element::Directory->new(
                                                   key=>$path,
                                                   path=>$path,
                                                   size=>$size_delta,
                                                   );
                $template->set_element($addition_dir);
            }
            my $dataset_dir = Perf::DataSet::Element::Directory->new(
                                        key=>$path,
                                        path=>$path,
                                        size=>$size,
                                        );
            $owner->get_container->set_element($dataset_dir);
        }

        # Verify the size of each file
        $owner->validate_container;

        # TODO: Find a way to validate size_on_disk
        #my $size_on_disk = $container_file->get_size_on_disk;
        #my $actual_size_on_disk = $merged_returns->{size_on_disk}->{individual}->{$path};
        #if ( $actual_size_on_disk  != $size_on_disk ){
        #    confess "File '$path' is not of expected size_on_disk ($actual_size_on_disk "
        #            ."instead of $size_on_disk)";
        #}

        # Clean up data we don't need anymore
        delete $self->{worker_templates};
        delete $owner->{addition_template};

        $owner->unmount(users_ok => 1);

        return $template->get_stats->as_hash;
    }
    else{
        confess "The owner must be a Perf::DataSet or Perf::DataSet::Worker and not a " . ref($owner);
    }
}


#
# Create files, links and directories according to params, boundaries and distributions
#
sub create {
    my $self = shift;
    my $container = $self->{container};
    my $type = $self->{type};
    my $debug = $self->{debug};

    if ($debug){
        $Log->comment("Starting dataset create");
        my $timings = {};
        $timings->{file_open} = 0;
        $timings->{io} = 0;
        $timings->{post_process} = 0;
        $timings->{align_buffer} = 0;
        $timings->{file_seek} = 0;
        $timings->{file_generate} = 0;
        $timings->{dir_make} = 0;
        $self->{timings} = $timings;
        $self->{io_histogram} = {};
    }

    my $files = $container->get_elements('files');
    #my $symlinks = $container->get_symlinks();
    my $dirs = $container->get_elements('dirs');
    if (!$dirs){
        confess "No directories in container!\n" . $container->as_string;
    }

    my $latest_time;
    if ($debug){
        $latest_time = time;
    }

    #
    # First create all the dirs
    #
    for my $dirpath (keys %$dirs){
        eval{mkpath($dirpath,1,0777)};
        if ($@){
            confess "Couldn't create '$dirpath': $@";
        }
        if (! -d $dirpath){
            confess "Creating '$dirpath' failed";
        }
        $dirs->{$dirpath}->mark_created;
    }
    if ($debug){
         my $time_after_dir_makes = time;
         $self->{timings}->{dir_make} += ($time_after_dir_makes - $latest_time);
         $latest_time = $time_after_dir_makes;
    }

    eval {
        #
        # Create all files in descending size order
        #
        my @files_keys = keys %$files;
        for my $filekey (@files_keys){
            my $file = $files->{$filekey};
            my $path = $file->get_path;

            my $flags;
            if ($file->get_size < 64 || $type =~ /^cifs/){
                $flags = O_RDWR|O_CREAT;
            }
            else {
                $flags = O_RDWR|O_CREAT|O_DIRECT;
            }

            if ($debug){
                print "Creating file '$path'\n";
            }
            my $fh = new FileHandle("$path", $flags)
                        or confess "could not create file \"$path\": $!\n";
            $file->mark_created;
            if ($debug){
                my $time_after_file_open = time;
                $self->{timings}->{file_open} += ($time_after_file_open - $latest_time);
                $latest_time = $time_after_file_open;
            }
            $self->{latest_time} = $latest_time;

            eval{
            my $bytes_written = $self->populate_file($fh,file=>$file);
            $latest_time = $self->{latest_time};
            #$file->set_size_on_disk($bytes_written);
            $container->set_element($file);
            };
            if ($@){
                # Make sure we close the file
                $fh->close or $Log->comment("Unable to close file '$path': $!");;
                die $@;
            }
            $fh->close() or confess "error closing file \"$path\": $!\n";

            if ($debug){
                my $time_after_file_close = time;
                $self->{timings}->{file_close} += ($time_after_file_close - $latest_time);
                $latest_time = $time_after_file_close;
            }

        }
    };
    my $error_during_create = $@;
    if ($debug){
        $Log->comment("Timings:\n" . as_string($self->{timings}));
        delete $self->{timings};
        my $io_histogram = $self->{io_histogram};
        my @buckets = sort {$a <=> $b} keys %$io_histogram;
        my $largest_bucket = $buckets[-1];
        my $hist_str;
        my $i = 0;
        while ($i <= $largest_bucket){
            my $range_string = "< " . (($i + 1) * 100) . "ms";
            my $count = $io_histogram->{$i};
            if (! defined($count) ){
                $count = 0;
            }
            $hist_str .= "$range_string : $count \n";
            $i ++;
        }
        $Log->comment("I/O histogram:\n$hist_str");
        delete $self->{io_histogram};
    }
    die $error_during_create if ($error_during_create);
}


sub handle_WorkerComplete {
    my $self = shift;
    my $event = shift;
    $self->{processes} -= $event->{payload};
    $self->adjust_buffer_sleep;
    return;
}


sub handle_MaxTput {
    my $self = shift;
    my $event = shift;
    my $new_max_tput = $event->{payload};
    $self->{max_tput} = $new_max_tput;
    $self->adjust_buffer_sleep;
    return;
}

sub populate_file($$@){
    my $self = shift;
    my $fh = shift;
    my %args = @_;
    my $file = $args{file};
    my $maxtime = $args{maxtime};

    my $filesize = $file->get_size;
    # "file" may actually be a file chunk rather than a whole file
    # "file->{size}" refers to the size of that chunk
    # "file->{size_whole} is the actually size of the file
    my $whole_filesize = $file->get_size_whole;
    if (!$whole_filesize){
        $whole_filesize = $filesize;
    }
    my $is_file_sparse = $file->get_sparse;
    my $filepath = $file->get_path;
    my $offset = $file->get_offset;
    my $block_size = $file->get_block_size;

    my $debug = $self->{debug};
    my $buffer_size = $self->{buffer_size};
    my $buffer = $self->{standard_buffer};
    my $align_offset = $self->{align_offset};

    $Log->comment("Populating file: $filepath");

    # Keeps track of how much data was actually written as opposed to seeked off
    my $bytes_written = 0;

    if (!defined($offset)){
        $offset = 0;
    }
    #my $original_file_size = get_file_size($fh);
    if ($offset){
        $Log->comment("Seeking to offset: $offset");
        my $ret = $fh->sysseek($offset,SEEK_SET);
        confess "Could not seek to '$offset' in file '$filepath': $!" if (!$ret);
        confess "Seek didn't go to the wanted spot ($ret instead of $offset)" if ($ret != $offset);
    }

    my $bytesleft = $filesize;
    my $bufl = 0;
    my $block_num = $offset / $block_size;
    if ($block_num != int($block_num)){
        confess "File offset ($offset) for '$filepath' not aligned on block size ($block_size)";
    }
    my $is_block_hole = 0;
    my $holes_count;
    my $tick_time = time;
    my $last_tick_time;

    my $sparser;
    my $next_hole;
    if ($is_file_sparse){
        my $sparse_ratio = $file->get_sparse_ratio;
        my $sparse_level = $file->get_sparse_level;
        my $sparse_align = $file->get_sparse_align;
        $sparser = $self->{sparser_factory}->sparser(
                        filesize => $whole_filesize,
                        block_size => $block_size,
                        offset => $offset,
                        range => $filesize,
                        sparse_ratio => $sparse_ratio,
                        sparse_level => $sparse_level,
                        sparse_align => $sparse_align,
                        );
        $next_hole = $sparser->get_next_hole;
    }

    while ($bytesleft) {

        #
        # How much do we need to write in this block?
        #
        my $printlen = $block_size;
        if ($printlen > $bytesleft) {
            $printlen=$bytesleft;
        }
        #
        # Is this block the beginning of a chain of hole?
        #
        if ($is_file_sparse && $next_hole && $block_num == $next_hole->[0] ){
            $is_block_hole = 1;
            $holes_count = $next_hole->[1] - $next_hole->[0] + 1;
            if ($holes_count < 1){
                confess "Invalid hole! " . as_string($next_hole);
            }
            $next_hole = $sparser->get_next_hole;
        }

        #
        # Populate block with data
        #
        if (!$is_block_hole){
            my $block_data = $file->generate_block( buffer   =>$buffer,
                                                    offset   =>$bufl + $align_offset,
                                                    size     =>$printlen,
                                                    block_num=>$block_num,
                                                    );
            my $block_data_length = length($block_data);
            if ( $block_data_length != $printlen ){
                confess "data block size different than expected: " .
                        "$block_data_length instead of $printlen";
            }
            $bufl += $printlen;
            $file->record_block($block_num,\$block_data);
        }
        else{
            # Do nothing, we'll do a seek after the write
        }
        #
        # How much space is left in the file?
        #
        if ($is_block_hole) {
            $bytesleft-=($printlen * $holes_count);
        } else {
            $bytesleft-=$printlen;
        }
        if ($bytesleft < 0){
            confess "Something very wrong happened: bytesleft: $bytesleft \n".
                                "printlen: $printlen \n" .
                                "holes_count: $holes_count \n" .
                                "filesize: $filesize \n" .
                                "block_num: $block_num";
        }
        #
        # Is it time to send a write?
        #
        if ( $bufl >= $buffer_size || $bytesleft == 0 || $is_block_hole ){
            # If there is something in the buffer, write it
            # There may not be anything if we are in a chain of block holes
            # at the beginning of the file
            if ($bufl){
                #$Log->comment("We want to write that much: $bufl");

                #
                # Perform write
                #
                my $before_time = time;
                my $error_during_write;
                my $actual_written = $fh->syswrite($$buffer,$bufl,$align_offset) || 0;
                if ($actual_written != $bufl){
                    $error_during_write = "Unable to write full buffer ($actual_written/$bufl) to '$filepath': $!";
                }
                my $io_time = time - $before_time;
                my $io_time_bucket = int($io_time * 1000 / 100);
                $self->{io_histogram}->{$io_time_bucket} ++;

                # Delayed death so can detect IO timeouts
                confess $error_during_write if ($error_during_write);

                #$Log->comment("Written $bufl bytes");
                $bytes_written += $bufl;

                # See if the load should be adjusted
                $self->process_events;

                if ($self->{buffer_sleep}){
                    # Find out how much time was spent not sleeping since the last sleep
                    # or the last io if no sleep happened
                    $last_tick_time = $tick_time;
                    $tick_time = time;
                    my $time_since_last_tick = ($tick_time - $last_tick_time);
                    # Prorate the sleep time based on the buffer length
                    my $sleep_time = ($bufl / $buffer_size) * $self->{buffer_sleep} - $time_since_last_tick;
                    if ($sleep_time > 0 && $sleep_time < 0.01){
                        # Rather than try to sleep for such a small amount of time
                        # Figure out if we should sleep at all
                        $sleep_time = rand(0.01/$sleep_time) <= 1 ? 0.01 : 0;
                    }
                    if ($sleep_time > 0){
                        my $ret = select (undef,undef,undef,$sleep_time);
                        if ($ret == -1){
                            confess "Unable to sleep between writes: $!";
                        }
                        # Include the sleep in the tick period
                        $tick_time = time;
                    }
                }
                $bufl = 0;
            }
            if ($is_block_hole){
                my $seeklen = $printlen * $holes_count;
                my $ret = $fh->sysseek($seeklen,SEEK_CUR);
                confess "Could not seek $seeklen in file '$filepath': $!" if (!$ret);

                # Account for the blocks in the chain other than the first one
                $block_num += $holes_count - 1;
                $is_block_hole = 0;
            }
            if ( defined ($maxtime) && (time() > $maxtime)  ) {
                last;
            }
        }
        $block_num++;
    }
    return $bytes_written;
}


sub copy {
    my $parent = shift;
    my $self = {};

    for my $key ( keys %{$parent} ){
        my $val = $parent->{$key};
        $self->{$key} = $val;
    }

    $self->{container} = $parent->{container}->copy();
    bless $self, ref($parent);
    return $self;
}


# Distribute Containers to Workers
sub distribute_partitions {
    confess('This method must be implemented in a subclass of ' . __PACKAGE__);
}


sub distribute_partitions_validate {
    my $self = shift;
    my %params = Params::Validate::validate(
           @_,
           {
            partitions => {type => ARRAYREF},
            owner      => {isa => 'Perf::DataSet'},
            processes  => {type => ARRAYREF},
            });

    # Verify contents of partitions and processes?

    return %params;
}


1;
