## @author   dl-perf-qa-automation@netapp.com
## @summary
## @status   private
##
## @pod here

package Perf::DataSet::Addition::NAS;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Perf::Utils qw(as_string in_bytes ceil);
use FileHandle;
use IO::Seekable;
use Time::HiRes qw(time);
use Perf::DataSet::Command;
use Perf::DataSet::Container;
use Perf::DataSet::SparserFactory;
use File::Path qw(rmtree mkpath);
use Fcntl;
use List::Util qw(sum min max);

use Storable qw();
use MIME::Base64;
use Perf::DataSet::Element;
use Perf::DataSet::Element::Directory;
use Perf::DataSet::Element::File;

use base qw(Perf::DataSet::Addition);

#
# Define some WAFL constants
#
my $BLOCKSIZE = 4096;

# Use to generate random strings
use constant GENERATE_COMPLETE => -1;


sub pick_file_sparse(){
    my $self = shift;
    my $is_sparse = 0;
    if ($self->{sparse}){
        if ($self->{sparse} != 1){
            $is_sparse = rand(10000) <= $self->{sparse} * 10000?1:0;
        }
        else{
            $is_sparse = 1;
        }
    }
    return {is_sparse=>$is_sparse,
            ratio=>$self->{sparse_ratio},
            level=>$self->{sparse_level},
            align=>$self->{sparse_align}};
}

sub pick_file_size {
    my ($self,$size_left) = @_;
    my $minsize;
    my $maxsize;
    my $distributions = $self->{size_distribution};
    my $boundaries = $self->{size_boundaries};

    #
    # Try to get a file size from each range with a non
    # zero distribution
    #
    if ( ! exists($self->{non_zero_distributions}) ){
        my $non_zero_distributions = 0;
        map { $non_zero_distributions ++ if ($_) } @$distributions;
        $self->{non_zero_distributions} = $non_zero_distributions;
    }

    # The first filesize ranges won't be picked at random using $lot
    # Instead, each range is picked once, starting with the last one
    if (! exists($self->{pick_file_size_range_idx}) ){
        $self->{pick_file_size_range_idx} = @$distributions - 1;
    }
    my $pick_file_size_range_idx = $self->{pick_file_size_range_idx};
    if ($pick_file_size_range_idx >= 0 ){
        while ($distributions->[$pick_file_size_range_idx] == 0){
            $pick_file_size_range_idx --;
            if ($pick_file_size_range_idx < 0){
                last;
            }
        }
        if ($pick_file_size_range_idx >= 0){
            $minsize = $boundaries->[$pick_file_size_range_idx];
            $maxsize = $boundaries->[$pick_file_size_range_idx + 1];
            $pick_file_size_range_idx -- ;
            $Log->debug("Range picked: [$minsize - $maxsize]");
        }
        $self->{pick_file_size_range_idx} = $pick_file_size_range_idx;
    }

    # Now that each range has been picked once we use
    # the distribution numbers to randomly chose each time
    if (! defined($minsize) || ! defined($maxsize)){

        # First remove size intervals that are not attainable given
        # how much size we have left to pick
        $distributions = \@{$self->{size_distribution}};
        $boundaries = \@{$self->{size_boundaries}};
        while (!$distributions->[0]) {
            shift @$boundaries;
            shift @$distributions;
        }
        if (!@$distributions) {
            confess("All distributions are 0, not a valid distribution array.");
        }
        while ($boundaries->[-2] > $size_left){
            last if (@$boundaries == 2);
            pop @$boundaries;
            pop @$distributions;
        }
        if ($boundaries->[0] > $size_left) {
            # Check file size priority and dump leftover or put all of leftover into 1 file
            $Log->debug("Perf::DataSet needs to write the last amount $size_left and is less than " .
                          "the smallest boundary $boundaries->[0]");
            if($self->{file_size_priority} eq 'size_boundaries') {
                $Log->debug("Using size boundaries as priority, Perf::DataSet size will be short $size_left\n" .
                              "If this message comes out more than once, the size will be short for a total of $size_left");
                return GENERATE_COMPLETE;
            }
            else {
                $Log->debug("Using Perf::DataSet size as priority, writing smaller file of $size_left");
                return $size_left;
            }
        }

        # Not sure why this is done. Jerome?
        if ($boundaries->[-1] > $size_left){
            my $new_boundary = ceil($size_left/$self->{buffer_align}) * $self->{buffer_align};
            $new_boundary = min($new_boundary, $boundaries->[-1]);
            $boundaries->[-1] = $new_boundary;
        }

        my $lot = rand( sum @$distributions );
        my $i = 0;
        my $limit = $distributions->[0];
        while ($lot > $limit){
            $i++;
            $limit += $distributions->[$i];
        }
        $minsize=$boundaries->[$i];
        $maxsize=$boundaries->[$i + 1];
    }

    my $chosen_size = $minsize + int(rand($maxsize - $minsize));
    if ($chosen_size < 1){
        $chosen_size = 1;
    }
    # Align the size with align_buffer if possible
    if ($chosen_size > 64){
        $chosen_size = $self->align_file_size( $chosen_size, $minsize, $maxsize);
    }
    return $chosen_size;
}

sub align_file_size {
    my ($self,$chosen_size,$minsize,$maxsize) = @_;
    confess "You must specify a size" if (!defined($chosen_size));

    my $buffer_align = $self->{buffer_align};
    my $modulo = $chosen_size % $buffer_align;
    if ($modulo >= $buffer_align/2){
        # Round up
        $chosen_size += ( $buffer_align - $modulo );
    }
    else{
        # Round down
        $chosen_size -= $modulo;
    }
    # See if size fits in interval
    if ( $chosen_size > $maxsize ){
        $chosen_size -= $buffer_align;
    }
    elsif ( $chosen_size < $minsize ){
        $chosen_size += $buffer_align;
    }
    # One last check before giving up
    if ( $chosen_size > $maxsize || $chosen_size < $minsize ){
        confess "Size range [$minsize - $maxsize] isn't valid. " .
                "It doesn't contain a multiple of $buffer_align bytes";
    }
    if ( $chosen_size % $buffer_align ){
        confess "Chosen size ( $chosen_size) isn't a multiple of $buffer_align";
    }
    if ($chosen_size < 1){
        $chosen_size = 1;
    }
    return $chosen_size;
}


sub distribute_partitions {
    my $self       = shift();
    my %arg        = $self->distribute_partitions_validate(@_);

    my $owner      = $arg{owner};
    my $processes  = $arg{processes};
    my $partitions = $arg{partitions};

    my %groups_map;
    if (   $owner->is_file_slicing_allowed
        && $owner->is_file_slicing_single_client) {
        my $i = 0;
        %groups_map = map { $_ => $i++ } @{$owner->get_clients};
    }
    else {
        %groups_map = map { $_ => 0 } @{$owner->get_clients};
    }
    my %worker_templates;
    for my $p (@$processes) {
        my $p_name     = $p->get_name;
        my $p_hostname = $p->get_hostname;
        my $group      = $partitions->[$groups_map{$p_hostname}];
        my $wt         = shift @$group;
        $worker_templates{$p_name} = $wt;

        # Don't send invalid coderefs across the network
        $wt->clear_on_dirty_for_all;
    }

    return %worker_templates;
}


sub generate($){
    my $self = shift;
    $self->{container} = Perf::DataSet::Container->new;

    $self->{size_so_far} = 0;
    my $loopid = 0;
    my $previous_file_size = 0;
    my $current_file_size = 0;
    #while ( !$loopid || ($self->{container}->get_size() < .99 * $self->{size}
    while ( !$loopid || ($self->{size_so_far} < .99 * $self->{size}
                    && $previous_file_size != $current_file_size) ) {
        $loopid ++;
        my $ret = $self->_generate_level({  pwd    => '',
                                            loopid => $loopid,
                                            level  => 0,
                                            },
                                        @{$self->{tree_structure}},
                                        );
        last if ($ret == GENERATE_COMPLETE);
        $previous_file_size = $current_file_size;
        #$current_file_size = $self->{container}->get_size('files');
        $current_file_size = $self->{size_so_far};
    }
    # At this point we may have tons of empty directories

    if (! $self->{container}->count('files')){
        confess "Unable to generate dataset template with parameters:\n" .
                "size_distribution:\n" . as_string($self->{size_distribution}) ."\n" .
                "size_boundaries:\n" . as_string($self->{size_boundaries});
    }

    return delete $self->{container};
}


#
# Generate list of files, links and directories according to params, boundaries and distributions
#
sub _generate_level {
    my $self = shift;
    my ($preamble,$content, @subcontent) = @_;
    my ($fcount, $hlcount, $slcount, $dcount) = @$content;
    my $pwd = $preamble->{pwd};
    my $level = $preamble->{level};
    my $loopid = $preamble->{loopid};
    my $block_size = $self->{block_size};

    #if($self->{container}->get_size() >= $self->{size}){
    if($self->{size_so_far} >= $self->{size}){
        #$Log->comment("We have reached the size requested ($self->{size}");
        return GENERATE_COMPLETE;
    }

    if ($pwd && $pwd ne '.'){
        # make sure path does not end with a slash
        $pwd =~ s/\/$//;
    }
    else{
        undef $pwd;
    }
    # Loop over files
    my $files_added = 0;
    my $complete;
    FCOUNT: foreach my $f (1..$fcount) {
        # Check that we haven't passed the target
        #my $size_left = $self->{size} - $self->{container}->get_size;
        my $size_left = $self->{size} - $self->{size_so_far};
        if ( $size_left <= 0 ){
            last;
        }
        my $filename = $self->{container}->get_random_filename(name=>"$loopid.$f",
                                                               length=>$self->{file_name_length},
                                                               );
        my $filesize = $self->pick_file_size($size_left);
        if ($filesize == GENERATE_COMPLETE) {
            $complete = 1;
            last FCOUNT;
        }
        my $filesparse = $self->pick_file_sparse();
        my $path = $pwd?"$pwd/$filename":$filename;
        my $type = $self->{type};
        #if ($self->{container}->get_size() + $filesize <= $self->{size}){
        if ($self->{size_so_far} + $filesize <= $self->{size}){
            $self->{file_too_big_counter} = 0;
        }
        else{
            # File is too big, maybe some smaller file will come along
            $self->{file_too_big_counter} ++;
            next if ( $self->{file_too_big_counter} < 10 );

            #
            # Try to finish dataset template with one file
            # so we don't loop forever
            #
            #my $size_remaining = $self->{size} - $self->{container}->get_size();
            my $size_remaining = $self->{size} - $self->{size_so_far};
            my $range = $self->get_range($size_remaining);
            next if ( !defined($range) || !$self->{size_distribution}->[$range]);
            my $aligned_size;
            eval {
                $aligned_size = $self->align_file_size( $size_remaining, @$range );
            };
            # If the eval blows up, skip the file
            next if (!$aligned_size);
            $filesize = $aligned_size;
        }

        my @args = (    key         =>$path,
                        path        =>$path,
                        size        =>$filesize,
                        size_whole  =>$filesize,
                        block_size  =>$block_size,
                        no_checksums=>$self->{compute_checksum}?0:1,
                    );
        my $file_element;
        if (!$filesparse->{is_sparse}){
            $file_element = $self->{container}->create_element('Perf::DataSet::Element::File',{@args});
        }
        else{
            push @args, (   sparser_factory=>$self->{sparser_factory},
                            sparse_ratio=>$filesparse->{ratio},
                            sparse_level=>$filesparse->{level},
                            sparse_align=>$filesparse->{align},
                        );
            $file_element = $self->{container}->create_element('Perf::DataSet::Element::File::Sparse',{@args});
        }
        $self->{container}->set_element( $file_element );
        $files_added ++;
        $self->{size_so_far} += $filesize;
    }
    # Adding symlinks doesn't work
    #foreach my $sl (1..$slcount) {
    #    my $filename = $self->pad_filename("symlink.$loopid.$sl");
    #    # TODO: this doesn't point to a valid file
    #    my $linkto =  $self->{container}->random_file();
    #    my $path = $pwd?"$pwd/$filename":$filename;

    #   if ($self->{container}->get_size() < $self->{size}){
    #       $self->{container}->add_symlink(key=>$path,
    #                                     path=>$path,
    #                                     size=>0,
    #                                     linkto=>$linkto,
    #                                     addition_uid=>$self->uid,
    #                                     size_on_disk=>0);
    #   }
    #   else{
    #       #$Log->comment("We have reached the size requested ($self->{size}");
    #       last;
    #   }
    #}
    if (!$complete) {
        foreach my $d (1..$dcount) {
            #if ($self->{container}->get_size() < $self->{size}){
            if ($self->{size_so_far} < $self->{size}){
                my $filename = $self->{container}->get_random_dirname(name=>"$loopid.$d",
                                                                      length=>$self->{file_name_length},
                                                                      );
                my $path = $pwd?"$pwd/$filename":$filename;
                my $ret = $self->_generate_level({    pwd    => $path,
                                            loopid => $loopid,
                                            level  => $level+1,
                                        },
                                        @subcontent,
                                        );
                last if ($ret == GENERATE_COMPLETE);
                # Since empty folders aren't kept, keep track of how many folders
                # are really added
                $files_added += $ret;
            }
            else{
                #$Log->comment("We have reached the size requested ($self->{size}");
                last;
            }
        }
    }
    # Only create the folder if there is a file somewhere underneath
    if($files_added){
        my $dir_element = $self->{container}->create_element('Perf::DataSet::Element::Directory',{
                                                              key=>$pwd?$pwd:'.',
                                                              path=>$pwd?$pwd:'.',
                                                              size=>0,
                                                              });
        $self->{container}->set_element($dir_element);
        $Log->debug("Generated directory '" . ($pwd?$pwd:'.') ."' with $files_added files underneath")
    }
    return $files_added || GENERATE_COMPLETE;
}


sub copy {
    my $parent = shift;
    my $self = {};

    for my $key ( keys %{$parent} ){
        my $val = $parent->{$key};
        $self->{$key} = $val;
    }

    $self->{container} = $parent->{container}->copy();
    bless $self, ref($parent);
    return $self;
}
1;
