# -*- perl -*-
#
## @author   dl-perf-qa-automation@netapp.com
## @summary
## @status   private
##
## @pod here

package Perf::DataSet::Addition::SAN;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log connect host subtest);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Data::Dumper;
use Perf::DataSet::Addition;
use FileHandle;
use Fcntl;
use Readonly;
use English qw(-no_match_vars);
use Perf::Utils qw(in_bytes);
use base qw(Perf::DataSet::Addition);


# According to the blocks host QA team, the beginning of a raw
# device may have special OS data at the beginning.  They
# recommend skiping at least the first megabyte of a raw
# device file.
# XXX BLS this isn't working as Container assumes an Element with an
# offset is sliced which isn't the case here.  Haven't seen any errors
# though.
Readonly::Scalar my $RAW_DEVICE_DATA_OFFSET => 0; # 1048576;


sub generate {
    my $self = shift();
    my $dataset = shift();

    my $filer_host = host($dataset->get_filer());
    my $filer_hosttype = $filer_host->hosttype();

    $self->{container} = Perf::DataSet::Container->new();

    # the actual LUN sizes could be larger than requested, recalculate
    # size based LUN queries
    $self->{size} = 0;

    my $raw_devices =
      $dataset->get_raw_device_resources(addition_uid => $self->{uid});

    confess("Addition received no raw devices")
        if (!defined($raw_devices) || 0 == scalar(@$raw_devices));

    my %args = (
        offset       => $RAW_DEVICE_DATA_OFFSET,
        block_size   => $self->{block_size},
        no_checksums => $self->{compute_checksum} ? 0 : 1,);

    foreach my $raw_device (@$raw_devices) {
        my $filename     = $raw_device->get_name();
        my $filesize     = $raw_device->get_dev_size();
        my $element_type = 'Perf::DataSet::Element::Raw';

        $self->{size} += $filesize;

        $args{key}        = $filename;
        $args{path}       = $filename;
        $args{size}       = $filesize - $RAW_DEVICE_DATA_OFFSET;
        $args{size_whole} = $filesize;
        $args{raw_size}   = $filesize;
        $args{lun_id}     = $raw_device->get_lun()->get_name();

        if ($self->{sparse}) {
            # Compute the size_on_disk
            my $sparser =
              $self->{sparser_factory}->sparser(
                                          filesize     => $filesize,
                                          block_size   => $self->{block_size},
                                          sparse_ratio => $self->{sparse_ratio},
                                          sparse_level => $self->{sparse_level},
                                          sparse_align => $self->{sparse_align},
                                          );

            $args{sparser_factory} = $self->{sparser_factory};
            $args{sparse_ratio}    = $self->{sparse_ratio};
            $args{sparse_level}    = $self->{sparse_level};
            $args{sparse_align}    = $self->{sparse_align};

            $element_type .= '::Sparse';
        }
        else {
            $element_type .= '::Base';
        }

        my $file_element =
          $self->{container}->create_element($element_type, \%args);

        $self->{container}->set_element($file_element);
    }

    # delete raw device reference to prevent issues passing
    # addition out to workers
    delete $self->{raw_devices};

    return delete $self->{container};
}


sub distribute_partitions {
    my $self       = shift();
    my %args       = $self->distribute_partitions_validate(@_);

    my $owner      = $args{owner};
    my $processes  = $args{processes};
    my $partitions = $args{partitions};

    my %worker_templates;

  PROCESS: foreach my $process (@$processes) {
        my $proc_name = $process->get_name();
        my $proc_host = $process->get_hostname();

      GROUP: foreach my $group (@$partitions) {
            if (0 == scalar(@$group)) {
                next GROUP;
            }
            my $container = $group->[0];
            my $raw_file = $container->random_file();
            # Should raw file offer get_host(), get_addition() and
            # get_prefix() in addition to full get_lun_id()?
            my $lun_id = $raw_file->get_lun_id();
            my @lun_parts = split(/_/, $lun_id);
            my $raw_host = $lun_parts[-1];

            if ($proc_host eq $raw_host) {
                shift(@$group);
                $worker_templates{$proc_name} = $container;
                $container->clear_on_dirty_for_all();
                next PROCESS;
            }
        }
    }

    return %worker_templates;
}


sub create {
    my $self = shift();
    my $container = $self->{container};

    my $files = $container->get_elements('files');

    foreach my $filekey (keys(%$files)) {
        my $file  = $files->{$filekey};
        my $path  = $file->get_path();
        my $flags = O_RDWR | O_DIRECT;

        my $fh = FileHandle->new("$path", $flags)
            or confess("Could not open file \"$path\": $!\n");

        $file->mark_created();

        eval {
            $self->populate_file($fh, file => $file);
            $container->set_element($file);
        };
        if ($@){
            # Make sure we close the file
            $fh->close or $Log->debug("Unable to close file '$path': $!");;
            die $@;
        }

        $fh->close() or confess "error closing file \"$path\": $!\n";
    }

    return;
}

1;
