## @author   dl-perf-qa-automation@netapp.com
## @summary
## @status   private
##
## @pod here


=head1 NAME

Perf::DataSet::CodeRef - Execute small snippets of code on clients and execution host

=head1 DECRIPTION

This package allows you to pass short pieces of code to Workers and
Perf::DataSet to collect information and pass it from clients to the execution host.

  Example:
    my $code = $dataset->create_command(
        'CodeRef',
        pre_execute_code => q!
          sub {
              my $self  = shift();
              my $owner = shift();

              $self->{container} = $owner->{container};

              return;
          };!,
        execute_code => q!
          sub {
              my $self = shift();
              return List::Util::sum(map { $_->get_invalid_block_count() }
                                     values(%{$self->{container}->get_elements('files')})
                                    );
          };!,
        post_execute_code => q!
          sub {
              my $self    = shift();
              my $owner   = shift();
              my $returns = shift();

              if ($owner->isa('Perf::DataSet')) {
                  return List::Util::sum(values(%{$returns}));
              }
          };!,
    );
    my $result = $dataset->execute($code);

=head1 METHODS

=cut

package Perf::DataSet::CodeRef;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Params::Validate qw(CODEREF SCALAR UNDEF);
use Data::Dumper;
use List::Util;
use Perf::DataSet::Command;

# CodeRef is a subclass of Command
use base qw(Perf::DataSet::Command);

=head2 new

Create a new Command that will execute the passed in code

=over

=item Arguments

B<pre_execute_code> = string

I<optional> The code snippet to be executed to prepare Worker and
Perf::DataSet for the execute phase.

B<execute_code> = string

I<optional> The code snippet to be executed during the execute phase.

B<post_execute_code> = string

I<optional> The code snippet to be executed to consolidate all the
results generated by the B<execute> phase.

=back

=cut

sub new {
    my $class = shift;
    my %params = $class->validate(@_);
    my $self = \%params;
    bless $self, $class;
    return $self;
}

sub validate {
    my $self = shift();

    my %params =
      Params::Validate::validate(@_,
                                 {description => {type     => SCALAR | UNDEF,
                                                  optional => 1,
                                                  default  => "code ref"},
                                  type => {type     => SCALAR | UNDEF,
                                           optional => 1,
                                           default  => 'nfs'},
                                  debug => {type     => SCALAR | UNDEF,
                                            optional => 1,
                                            default  => 0},
                                  pre_execute_code => {type => SCALAR,
                                                       optional => 1,},
                                  execute_code => {type     => SCALAR,
                                                   optional => 1,},
                                  post_execute_code => {type     => SCALAR,
                                                        optional => 1,},},);

    return %params;
}


sub execute_coderef {
    my $self      = shift();
    my $code_name = shift();

    return
        if (!defined($self->{$code_name}));

    my $coderef = eval $self->{$code_name};

    confess("Error evaluating code: $self->{$code_name}\n$@\n")
        if ($@);

    return &$coderef($self, @_);
}


sub pre_execute {
    my $self = shift();

    return $self->execute_coderef('pre_execute_code', @_);
}


sub execute {
    my $self = shift();

    return $self->execute_coderef('execute_code', @_);
}

sub post_execute {
    my $self = shift();

    return $self->execute_coderef('post_execute_code', @_);
}

1;
