# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

package Perf::DataSet::Command;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
#use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
#use Module::Common qw(validate_mixed);
use Perf::Utils qw(min max get_random_element as_string in_bytes);
use Perf::Depot;
use Perf::DataSet::Event;
# TODO: Put the command subclasses below in a Command folder
# and load them all
use Perf::DataSet::Addition;
use Perf::DataSet::CodeRef;
use Perf::DataSet::GetActualSizes;
use Perf::DataSet::ModularWorkload;
use Perf::DataSet::Purge;
use Perf::DataSet::Workload;


sub new {
    my $class = shift;
    if ($class ne 'Perf::DataSet::Command'){
        confess "Your Command subclass must implement a 'new' subroutine";
    }
    my $subclass = shift;
    if (!$subclass){
        confess "You must specify a Command subclass";
    }
    my %params = @_;
    my $dataset_uid = delete $params{dataset_uid};
    if (!$dataset_uid){
        confess "You must specify a dataset_uid";
    }

    # Keep only the last part of the package name
    if ($subclass =~ /Perf::DataSet::(.*)\s*$/){
        $subclass = $1;
    }
    my $full_subclass_name = 'Perf::DataSet::' . $subclass;
    my $subclass_package = 'Perf::DataSet::' . $subclass;

    # Try to load the package
    eval "require $full_subclass_name";
    if ($@){
        confess "Unable to load $full_subclass_name: $@";
    }
    # Check to see if the subclass is a descendant of Perf::DataSet::Command
    eval{
        if (!$subclass_package->isa('Perf::DataSet::Command')){
            die "Class '$subclass_package' is not a Perf::DataSet::Command";
        }
    };
    if ($@ && $@ =~ /is not a Perf..DataSet/){
        confess $@;
    }
    elsif ($@){
        confess "Class '$subclass_package' is not a Perf::DataSet::Command: $@";
    }

    #my %args = $subclass_package->validate(%params);
    my $self = $subclass_package->new(%params);
    $self->{dataset_uid} = $dataset_uid;
    $self->{uid} = Perf::Depot->make_uid;

    return $self;
}

sub get_uid {
    my $self = shift;
    return $self->{uid};
}

sub description {
    my $self = shift;
    return $self->{description};
}

sub send_event {
    my ($self,$event,%args) = @_;
    my $dataset_uid = $self->{dataset_uid};
    my $dataset = Perf::Depot->get_resource(class => 'DataSet', name=>$dataset_uid);
    confess "Invalid Dataset '$dataset_uid'" if (!$dataset);
    $dataset->get_manager->send_event($dataset_uid, $event, %args);
}

sub get_events {
    my $self = shift;
    my $events = Perf::DataSet::Event->receive(command_uid=>$self->{uid});
    if ($self->{debug} && $events){
        map {$Log->comment("Received event:\n" . as_string($_))} @$events;
    }
    return $events;
}

sub abort {
    my $self = shift;
    my $event = Perf::DataSet::Event->new(type=>'Abort',
                                    command_uid=>$self->{uid},
                                    );
    $self->send_event($event);
}

sub set_max_tput {
    my ($self,$max_tput) = @_;
    confess "You must pass an argument" if (! defined $max_tput);
    my $event = Perf::DataSet::Event->new(type=>'MaxTput',
                                    command_uid=>$self->{uid},
                                    payload=>$max_tput,
                                    );
    $self->send_event($event);
}

sub notify_worker_completion {
    my ($self,$completions) = @_;
    confess "You must pass an argument" if (! defined $completions);
    confess "Argument cannot be zero" if (!$completions);
    my $event = Perf::DataSet::Event->new(type=>'WorkerComplete',
                                    command_uid=>$self->{uid},
                                    payload=>$completions,
                                    );
    $self->send_event($event);
}

sub process_events {
    my $self = shift;
    my $events = shift;
    if (!$events){
        $events = $self->get_events;
    }
    elsif(ref($events) ne 'ARRAY'){
        confess "You must pass an arrayref";
    }
    else{
        my $all_event_objects = 1;
        map {$all_event_objects = 0 if (!$_->isa('Perf::DataSet::Event'))} @$events;
        if ($all_event_objects){
            confess "You must pass an arrayref of Perf::DataSet::Events";
        }
    }
    if($events){
        for my $event (@$events){
            my $type = $event->{type};
            my $method = "handle_$type";
            if (!$self->can($method)){
                $Log->comment("Command " . ref($self) ."cannot handle events of type '$type'");
                next;
            }
            $self->$method($event);
        }
    }
    return;
}

sub validate {
    confess "Your Command subclass must implement a 'validate' subroutine";
}
sub pre_execute {
    confess "Your Command subclass must implement a 'pre-execute' subroutine";
}
sub execute {
    confess "Your Command subclass must implement a 'execute' subroutine";
}
sub post_execute {
    confess "Your Command subclass must implement a 'post-execute' subroutine";
}
1;
