# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Container - Collection of L<Element> objects.

=head1 DESCRIPTION

This package provides methods to modify and query a collection of Elements. It is used internally
by L<Perf::DataSet> and most L<Command> objects.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Container;
{
    use strict;
    use warnings;
    use Carp qw(confess longmess verbose);
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF SCALARREF);
    use Module::Common qw(validate_mixed);
    use Perf::Utils qw(get_random_element in_bytes deep_copy);
    use List::Util qw( min max );
    use POSIX qw(ceil);
    use Digest::MD5 qw(md5 md5_base64);
    use Time::HiRes qw(time);
    use Data::Dumper;
    use Scalar::Util qw(weaken);

    use Perf::DataSet::SparserFactory;
    use Perf::DataSet::Sparser;
    use Perf::DataSet::ContainerStats;
    use Perf::DataSet::Element;
    # TODO: Load all files under 'Element' folder
    use Perf::DataSet::Element::File;
    use Perf::DataSet::Element::File::Sparse;
    use Perf::DataSet::Element::Raw::Base;
    use Perf::DataSet::Element::Raw::Sparse;
    use Perf::DataSet::Element::Directory;
    use Object::InsideOut qw(Storable);

    # Use to generate random strings
    my @alphabet = ('a'..'z','A'..'Z',0..9);
    my $alphabet_count = scalar @alphabet;

    # Member variables
    my @elements         :Field :Name('elements');
    my @elements_types   :Field :Name('elements_types');
    my @elements_size_sorted :Field :Name('elements_size_sorted');
    my @stats            :Field :Type('Perf::DataSet::ContainerStats') :Name('stats');
    # 'dirty' is set when the number of elements tracked changes
    my @dirty            :Field :Set('set_dirty') :Default(1) :Name('dirty');
    my @recomputing      :Field :Name('recomputing') :Get('is_recomputing') :Set('set_recomputing');
    # Store deleted elements separately
    my @deleted_elements :Field :Name('deleted_elements');
    # Store old elements until container is undirty-ed
    my @dirty_elements   :Field :Name('dirty_elements');
    # Flag set to true when the container 'size' has changed
    my @dirty_size       :Field :Name('dirty_size');

=head2 new

Returns an object representing an empty collection.

=over

=item Arguments

None

=item Examples

    $container = Perf::DataSet::Container->new;

=back

=cut

    sub _init :Init {
        my $self = $_[0];
        $elements[$$self]       = {};
        $dirty_elements[$$self] = {};
        $deleted_elements[$$self]       = {};
        $stats[$$self]          = Perf::DataSet::ContainerStats->new;
        $recomputing[$$self]    = 0;
    }

=head2 clone

Returns a copy of this Container object made of the clones of all the Elements it contains.
See L<Element/clone>.

=over

=item Arguments

None

=item Examples

    $clone = $container->clone;

=back

=cut

    sub _replicate :Replicate {
        my ($self, $clone, $flag) = @_;
        # This will be called in the context of a shallow cloning
        confess "Deep cloning not supported" if ($flag eq 'deep');
        # Don't inherit dirty_elements
        delete $dirty_elements[$$clone];
        my %clone_elements = map { $_->get_key => $_->clone } values %{$self->get_elements};
        $elements[$$clone] = {};
        $clone->set_elements(\%clone_elements);
        my $clone_size = $clone->get_size;
        my $self_size = $self->get_size;
        if ($clone_size != $self_size){
            confess "Self and clone don't have same size ($self_size and $clone_size)!\n".
                    "Clone:\n" . $clone->as_string .
                    "Self:\n" . $self->as_string;
        }
        my $clone_size_on_disk = $clone->get_size_on_disk;
        my $self_size_on_disk = $self->get_size_on_disk;
        if ($clone_size_on_disk != $self_size_on_disk) {
            confess "Self and clone don't have same size_on_disk ( $self_size_on_disk and $clone_size_on_disk)!\n".
                    "Clone:\n" . $clone->as_string .
                    "Self:\n" . $self->as_string;
        }

        # remove callbacks from clone
        $clone->clear_on_dirty_for_all();
    }

    sub as_array {
        my $self = $_[0];
        my $dump = $self->dump;
        my $d = $dump->[1]->{'Perf::DataSet::Container'};
        $d->{stats} = $stats[$$self]->dump;
        my %elements_dump = map { $_->get_key => $_->dump } values %{$self->get_elements};
        $d->{elements} = \%elements_dump;
        return $dump;
    }

=head2 as_string

Returns a string representation of the object. This is required since calling Data::Dumper directly
on the object handle won't show its fields (due to the nature of inside-out objects).

=over

=item Examples

    confess "Something horrible happened! The object:\n" . $container->as_string;

=back

=cut

    sub as_string {
        my $self = $_[0];
        my $string = Dumper( $self->as_array );
        $string =~ s/\$VAR1\s=\s//;
        $string =~ s/;$//;
        return $string;
    }

=head2 get_elements

Returns a hash reference containing all the Elements objects present in the collection.
Format of that hash is as follow:

    { element1_key => element1,
      element2_key => element2,
      ...
    }

See L<Element/new> for description of what a key is.

=over

=item Arguments

string = files | dirs | any other type of Element

I<Optional> If not specified all Elements are returned

include_deleted = 0|1

I<Optional> Include deleted elements or not. Defaults to 1.

=item Examples

    my $files = $container->get_elements('files', include_deleted=>0);
    my $str = join('\n', keys %$files);
    print "Files in this container:\n$str";

=back

=cut

    sub get_elements {
        my $self = shift;
        my ($type, %args);
        if (scalar(@_) % 2){
            $type = shift;
            %args = @_;
        }
        else{
            %args = @_;
        }
        my $include_deleted = $args{include_deleted};
        $include_deleted = 1 if (! defined $args{include_deleted});
        my $no_copy = $args{no_copy};
        my $count_only = $args{count_only};

        my @merging_hashes;
        if ($type){
            push @merging_hashes, $elements[$$self]->{$type};
            push @merging_hashes, $deleted_elements[$$self]->{$type} if ($include_deleted);
        }
        else{
            push @merging_hashes, values %{$elements[$$self]};
            push @merging_hashes, values %{$deleted_elements[$$self]} if ($include_deleted);
        }
        if (scalar @merging_hashes == 1 && $no_copy){
            return $merging_hashes[0];
        }
        my %merge;
        my $total_keys = 0;
        # Merge hashes
        for my $h (@merging_hashes){
            $total_keys += scalar keys %$h;
            if (!$count_only){
                @merge{ keys %$h } = values %$h;
            }
        }
        return $total_keys if ($count_only);
        confess "Collision detected!" if ($total_keys != scalar keys %merge);
        return \%merge;
    }

    sub get_elements_keys {
        my $self = shift;
        my $es = $self->get_elements(@_, no_copy=>1);
        return [keys %$es] if ($es);
    }

    # Prepare elements for RPC network communication
    sub get_rpc_serialized_elements {
        my ($self, $type) = @_;
        my @elements = values %{$self->get_elements($type, no_copy=>1)};
        my @serialized_elements = map { $_->rpc_serialize } @elements;
        return \@serialized_elements;
    }

=head2 create_element

Returns a new Element.

=over

=item Arguments

string = 'Perf::DataSet::Element::File' | 'Perf::DataSet::Element::Directory' | ...

I<Required> Package name of a subclass of Element

hashref

I<Required> Arguments required for the 'new' method of Element

=item Examples

    my $file = $container->create_element('Perf::DataSet::Element::File',
                                          { key => file_key, path => file_path, ...},
                                          );

=back

=cut

    sub create_element($$$) {
        my ($self,$package,$args) = @_;
        #eval "require $package";
        #confess $@ if $@;
        my $e = $package->new(%$args);
        #eval '$e = ' . $package . '->new(%$args)';
        #confess $@ if $@;
        return $e;
    }

=head2 get_element

Returns Element with specific key.

=over

=item Arguments

string

I<Required> key of an Element. See L<Element/new>

=item Examples

    my $file_element = $container->get_element( 'key_of_my_file' );

=back

=cut

    sub get_element {
        my ($self,$key) = @_;
        confess "undefined self!" if (!$self);
        confess "undefined key!" if (!$key);
        my $type = $elements_types[$$self]->{$key};
        return if (!$type);
        my $element = $elements[$$self]->{$type}->{$key};
        if (!$element){
            $element = $deleted_elements[$$self]->{$type}->{$key};
        }
        return $element;
    }

=head2 set_element

Insert Element into the Container

=over

=item Arguments

None

=item Examples

    $container->set_element($element);

=back

=cut

    sub set_element {
        my ($self,$element) = @_;
        if (!$element || !$element->isa('Perf::DataSet::Element')){
            confess "Argument must be a Perf::DataSet::Element";
        }
        my $element_key = $element->get_key;
        my $element_type = $element->get_type;
        my $curr_element = $self->get_element($element_key);
        if (!$curr_element){
            # Marking container dirty forces stats to be recomputed
            $dirty[$$self] = 1;
            # Tell element to tell us when it gets dirty
            $self->add_on_dirty_to($element);
            $elements_types[$$self]->{$element_key} = $element_type;
        }
        elsif( $element->is_dirty ){
            # Keep track of the old element for later comparison
            if ($dirty_elements[$$self]->{$element_key}){
                # There is already an old element for this key. This means 'undirty'
                # hasn't been called since the last time this element changed.
                # Keep the oldest element for accurate comparison during the next 'undirty'
            }
            else{
                $dirty_elements[$$self]->{$element_key} = $curr_element;
            }
            # Record that the size changed between the new and the old element
            if ( $element->get_size != $curr_element->get_size || $element->is_deleted ){
                $dirty_size[$$self] = 1;
            }
        }
        if ($element->is_deleted){
            delete $elements[$$self]->{$element_type}->{$element_key};
            $deleted_elements[$$self]->{$element_type}->{$element_key} = $element;
        }
        else{
            delete $deleted_elements[$$self]->{$element_type}->{$element_key};
            $elements[$$self]->{$element_type}->{$element_key} = $element;
        }
        return;
    }

    # Return true if the container or any of the contained elements are dirty
    sub is_dirty {
        my $self = $_[0];
        return 1 if $dirty[$$self];
        return 1 if ($dirty_elements[$$self] && %{$dirty_elements[$$self]});
        return 0;
    }

    sub is_size_dirty {
        my ($self) = @_;
        return ( $dirty[$$self] || $dirty_size[$$self] );
    }

    sub remove_deleted_elements {
        my ($self) = @_;
        for my $type ( keys %{$deleted_elements[$$self]} ){
            for my $e ( values %{$deleted_elements[$$self]->{$type}} ){
                $Log->comment("Removing deleted element: " . $e->get_key);
                $self->remove_element($e);
            }
        }
    }

    sub undirty {
        my $self = $_[0];
        return if (!$self->is_dirty);

        # Remove deleted elements
        #$self->cull;

        # TODO Computing the stats is currently an expensive operation
        # since it walks the entire list of Elements.
        # Instead, use @dirty_elements to do just what is required to get
        # the new stats.
        my $starttime = time;
        my $s = Perf::DataSet::ContainerStats->new;
        $s->compute($self);
        $stats[$$self] = $s;
        my $runtime = int((time - $starttime)*1000);
        $Log->debug("Container was dirty. Recomputed stats in about $runtime milliseconds");

        # Optimization
        $self->size_sort_elements if ($elements_size_sorted[$$self]);

        $dirty[$$self] = 0;
        $dirty_size[$$self] = 0;

        # Undirty all the elements as well
        map { $_->undirty } values %{$self->get_elements};
        delete $dirty_elements[$$self];

        return;
    }

=head2 get_stats

Returns the ContainerStats object associated with this Container.
See L<ContainerStats> for available methods.

=over

=item Arguments

None.

=item Examples

    $stats = $container->get_stats;
    print "Stats from container:\n" . $stats->as_string;

=back

=cut

    sub get_stats {
        my $self = $_[0];
        $self->undirty;
        return $stats[$$self];
    }

=head2 contains_type

Returns true if the collection contains at least one Element of a specific type

=over

=item Arguments

string = files | dirs | any other type of Element

I<Required>

=item Examples

    if ($container->contains_type('files')){
        print "Container contains files";
    }

=back

=cut

    sub contains_type($$) {
        my ($self,$type)  = @_;
        my $b = $self->get_stats->get_field('type',$type);
        return $b?1:0;
    }

=head2 get_size

Returns aggregated size information in Bytes about the collection of Elements

=over

=item Arguments

string = files | dirs | any other type of Element

I<Optional> Returns total size if type is not specified

=item Examples

    $size = $container->get_size('files');
    print "Size of all files: $size";


=back

=cut

    sub get_size($;$) {
        my ($self,$type) = @_;
        if (!$type){
            $type = 'total';
        }
        elsif(!$self->contains_type($type)){
            return 0;
        }
        return $self->get_stats->get_field('size',$type);
    }

=head2 get_size_on_disk

Returns aggregated "size on disk" (i.e. not counting the holes) information in Bytes about
the collection of Elements.

=over

=item Arguments

string = files | dirs | any other type of Element

I<Optional> Returns total size on disk if type is not specified

=item Examples

    $size = $container->get_size_on_disk('files');
    print "Size on disk of all files: $size";


=back

=cut

    sub get_size_on_disk($;$) {
        my ($self,$type) = @_;
        if (!$type){
            $type = 'total';
        }
        elsif(!$self->contains_type($type)){
            return 0;
        }
        return $self->get_stats->get_field('size_on_disk',$type);
    }

=head2 count

Return the number of Elements of a specific type

=over

=item Arguments

string = files | dirs | any other type of Element

I<Optional> Returns total number of elements if type is not specified

=item Examples

    $count = $container->count('files');
    print "Found $count files";

=back

=cut

    sub count {
        my $self = shift;
        return $self->get_elements(@_, count_only=>1);
    }

=head2 set_elements

Insert multiple Elements into the Container

=over

=item Arguments

hashref or arrayref

I<Required> If a hashref is provided, keys are ignored

=item Examples

    $elements = [$e1,$e2,$e3];
    $container->set_elements($elements);

    $elements = { key1 => $e1, key2 => $e2 };
    $container->set_elements($elements);

=back

=cut

    sub set_elements($$){
        my ($self,$list) = @_;
        if (!$list){
            confess "No elements passed in";
        }
        if (ref($list) eq 'ARRAY'){
            map { $self->set_element($_) } @$list;
        }
        elsif(ref($list) eq 'HASH'){
            map { $self->set_element($_) } values %$list;
        }
        else{
            confess "Unsupported input";
        }
        return;
    }

=head2 merge_with

Inserts all the elements of another Container into this one

=over

=item Arguments

Container

B<Required> The Container whose Elements should be added to this Container

=item Examples

    $container1->merge_with( $container2 );

=back

=cut

    sub merge_with($$){
        my ($self,$merging) = @_;
        if (!$merging || !$merging->isa( __PACKAGE__ )){
            confess "You must pass a " . __PACKAGE__ ;
        }
        $self->set_elements($merging->get_elements);
    }

=head2 remove_element

Removes Element with specific key from Container. Returns the removed Element.

=over

=item Arguments

string | object

<Required> Element key or object

=item Examples

    $container->remove_element( $element->get_key );
    $container->remove_element( $element );

=back

=cut

    sub remove_element {
        my ($self, $key) = @_;
        my $type;
        if (ref($key) && $key->isa('Perf::DataSet::Element')){
            $key = $key->get_key;
            $type = $key->get_type;
        }
        $type = $elements_types[$$self]->{$key} if (!$type);
        confess "No element type found for key $key" if (!$type);
        my $deleted = delete $elements[$$self]->{$type}->{$key};
        if (!$deleted){
            $deleted = delete $deleted_elements[$$self]->{$type}->{$key};
        }
        confess "No element found with key $key" if (! $deleted );
        $dirty[$$self] = 1;
        return $deleted;
    }

    # Undoes partitioning
    sub unpartition {
        my $self = $_[0];
        my $slices = {};
        my $es = $self->get_elements;
        for my $element (values %$es) {
            if (!$element->is_slice){
                #$Log->comment("Element with key " . $element->get_key . " is not a slice");
                next;
            }
            my $path_slices = $slices->{ $element->get_path };
            if (!$path_slices){
                $slices->{ $element->get_path } = [$element];
            }
            else{
                if (!$element->is_coslice($path_slices->[-1]) ){
                    confess "Incompatible elements!\nNew:\n" . $element->as_string
                            . "Last:\n" . $path_slices->[-1]->as_string;
                }
                push @$path_slices,$element;
            }
        }

        for my $path (keys %$slices){
            # At this point we already the elements have compatible field values thanks to
            # the call to 'is_coslice'. All we need to do is make sure we're not missing
            # a slice to reconstruct the file
            my $elms = $slices->{$path};
            my $slice_total_size = 0;
            map { $slice_total_size += $_->get_size } @$elms;
            if ($slice_total_size != $elms->[0]->get_size_whole){
                confess "Slices found don't add up to file size for $path." .
                        "($slice_total_size instead of " .  $elms->[0]->get_size_whole . ")";
            }
            my $file_elem = $elms->[0]->unslice;
            map { $self->remove_element($_->get_key) } @$elms;
            $Log->debug( "Adding element " . $file_elem->get_key . " to self" );
            $self->set_element($file_elem);
        }
    }

=head2 partition

Breaks up the Container into X smaller Containers and returns an arrayref.
Will attempt to generate Containers of equal sizes. If the 'file_slicing' option
is set to 1, it will break up files to make the Containers even closer in size
to each other.

=over

=item Arguments

number

<Required> Number of Container objects to return

B<file_slicing> = 0 | 1

<Optional> Whether or not file can be broken up into smaller slices to achieve better uniformity
in size between partition Container objects.

=item Examples

    my $partitions = $container->partition(4, {file_slicing=>0});

=back

=cut

    sub partition {
        my ($self,$count,$args) = @_;
        my $file_slicing = $args->{file_slicing};
        if (!defined($file_slicing)){
            $file_slicing = 1;
        }
        my $group_count = $args->{groups};
        if (!defined($group_count)){
            $group_count = 1;
        }
        if ($count % $group_count){
            confess "Number of partitions ($count) must be a multiple of the number of groups ($group_count)";
        }

        #
        # Files are distributed among the partitions
        # Directories are present in all partitions
        #
        my $all_files = $self->get_elements('files');
        if (!$all_files){
            confess "No files were found in container before partition";
        }

        # Groups are made of non-sliced files
        if (scalar keys %$all_files < $group_count ){
            confess "Not enough files to create $group_count groups of files";
        }

        my $groups = $self->_create_partitions({elements=>$all_files, count=>$group_count});
        my ($min_size,$max_size, @group_sizes);
        $max_size = 0;
        for my $p (@$groups){
            my $size = $p->get_size('files');
            if ($size > $max_size){
                $max_size = $size;
            }
            if (!defined($min_size) || $size < $min_size){
                $min_size = $size;
            }
            push @group_sizes, $size;
        }
        $Log->debug("Created $group_count groups of partitions:\n" .
                      "- Min. size: $min_size\n" .
                      "- Max. size: $max_size\n" .
                      "- All: [ @group_sizes ]");

        my $partitions;
        my $group_partition_count = $count / $group_count;
        my $i = 0;
        while ($i < $group_count){
            my $group = $groups->[$i];
            my $files = $group->get_elements('files');
            my $created_slices = 0;
            #
            # If we have less files than the number of partitions requested,
            # create file chunks aligned on the block_size
            #
            if (scalar keys %$files < $group_partition_count ){
                if (!$file_slicing){
                    confess "You must allow file slicing or add more files to your dataset";
                }
                $files = $self->_make_slices({elements=>$files,
                                              count=>$group_partition_count,
                                              });
                $created_slices = 1;
            }
            #my $partitions = $self->_create_partitions(elements=>$files, count=>$group_partition_count, type=>'files');
            my $group_partitions = $self->_create_partitions({elements=>$files, count=>$group_partition_count});

            #
            # Find out smallest and biggest partition
            #
            my $max_size = 0;
            my $min_size;
            my @group_partition_sizes;
            for my $p (@$group_partitions){
                my $size = $p->get_size('files');
                if ($size > $max_size){
                    $max_size = $size;
                }
                if (!defined($min_size) || $size < $min_size){
                    $min_size = $size;
                }
                push @group_partition_sizes, $size;
            }

            #
            # Make sure the partitions are of even sizes. If not, create slices
            # and recreate the partitions
            #
            if ($file_slicing && !$created_slices){
                #
                # Don't allow more than 10% difference in size
                # between partitions
                #
                if ($max_size > 1.1*$min_size){
                    $Log->debug("Group partitions are not of even sizes. Re-creating. max_size: $max_size. min_size: $min_size");
                    # Create chunks for files that are bigger than
                    # 10% of the max size
                    my $min_size_file = $max_size * 0.1;
                    $files = $self->_make_slices({elements=>$files,
                                                  count=>$group_partition_count,
                                                  min_size=>$min_size_file,
                                                  });
                    $group_partitions = $self->_create_partitions({elements=>$files, count=>$group_partition_count});

                    #
                    # Find out smallest and biggest partition
                    #
                    $max_size = 0;
                    undef @group_partition_sizes;
                    undef $min_size;
                    for my $p (@$group_partitions){
                        my $size = $p->get_size('files');
                        if ($size > $max_size){
                            $max_size = $size;
                        }
                        if (!defined($min_size) || $size < $min_size){
                            $min_size = $size;
                        }
                        push @group_partition_sizes, $size;
                    }
                }
            }
            $Log->debug("Created $group_partition_count partitions in group $i:\n" .
                          "- Min. size: $min_size\n" .
                          "- Max. size: $max_size\n" .
                          "- All: [ @group_partition_sizes ]");

            # Give the directories to all partitions
            # TODO: This isn't quite right. Instead, try to add only the
            # ones that are needed for the files present in each partition
            my $dirs = $self->get_elements('dirs');
            map { $_->set_elements($dirs) } @$group_partitions;

            $partitions->[$i] = $group_partitions;
            $i ++;
        }
        return $partitions;
    }

    sub _make_slices {
        my ($self,$args) = @_;
        my $input_elements = $args->{elements};
        my $count = $args->{count};
        my $min_size = $args->{min_size};
        if (!defined($min_size)){
            $min_size = 0;
        }
        my $sliced_elements = {};

        # Make sure we deal with a copy
        #$elements = deep_copy($elements);

        my @elements_keys = keys %$input_elements;
        my $initial_elements_count = scalar @elements_keys;
        #
        # Create the file slices
        #
        my $slices_per_file = $count;
        for my $element_key (@elements_keys){
            my $element = $input_elements->{$element_key};
            if (!$element->is_sliceable){
                $sliced_elements->{$element_key} = $element;
                next;
            }
            my $element_size = $element->get_size;
            if ($element_size < $min_size){
                # File is too small
                $sliced_elements->{$element_key} = $element;
                next;
            }
            my $slices = $element->slice($count);
            for my $s (@$slices){
                $sliced_elements->{$s->get_key} = $s;
            }
        }
        #
        # Make sure we created enough slices
        #
        my $new_elements_count = scalar keys %$sliced_elements;
        if ($new_elements_count < $count){
            confess "We didn't create enough file slices ($new_elements_count instead of $count):\n"
                        . as_string($sliced_elements);
        }
        #
        # Make sure we create any slice at all
        #
        if ($new_elements_count == $initial_elements_count){
            confess "We didn't create any file slices! min_size was $min_size. Elements:\n"
                    . as_string($sliced_elements);
        }
        return $sliced_elements;
    }

    sub _make_slices_old {
        my ($self,$args) = @_;
        my $input_elements = $args->{elements};
        my $count = $args->{count};
        my $min_size = $args->{min_size};
        if (!defined($min_size)){
            $min_size = 0;
        }
        # This is required when slicing sparse files
        my $sparser_factory = $args->{sparser_factory};

        my $sliced_elements = {};

        # Make sure we deal with a copy
        #$elements = deep_copy($elements);

        my @elements_keys = keys %$input_elements;
        my $initial_elements_count = scalar @elements_keys;
        #
        # Create the file slices
        #
        my $slices_per_file = $count;
        for my $element_key (@elements_keys){
            my $element = $input_elements->{$element_key};
            if (!$element->is_sliceable){
                confess "Element is not sliceable:\n" . $element->as_string;
            }
            my $element_size = $element->get_size;
            my $element_block_size = $element->get_block_size;
            if ($element_size < 2 * $element_block_size){
                # We can't make slices out of this file. Skip.
                $sliced_elements->{$element_key} = $element;
                next;
            }
            if ($element_size < $min_size){
                # File is too small
                $sliced_elements->{$element_key} = $element;
                next;
            }
            my $size_per_slice;
            if ($element_size >= $count * $element_block_size){
                $size_per_slice = int($element_size / $count / $element_block_size)
                                        * $element_block_size;
            }
            else{
                $size_per_slice = $element_block_size;
            }
            my $slice_count = ceil($element_size / $size_per_slice);
            my $size_of_last_slice = $size_per_slice;
            if ($slice_count * $size_per_slice > $element_size){
                #
                # Just make the last slice a little bit bigger so each slice is at least of
                # size $block_size and there are no more than $count slices
                #
                $size_of_last_slice = $element_size - ($slice_count - 2) * $size_per_slice;
                $slice_count --;
            }
            my $total_slices_size = 0;
            my $i = 0;
            while ($i < $slice_count){
                my $slice_key = $element->get_path() . "/$i";
                my $slice_offset = $i * $size_per_slice;

                my $slice = $element->clone; # Shallow copy
                $slice->set_offset( $slice_offset );
                $slice->set_size( $i < $slice_count - 1 ? $size_per_slice : $size_of_last_slice );
                $slice->set_key( $slice_key );
                $slice->set_size_whole( $element_size );
                if ($slice->get_sparse){
                    confess "You must pass a sparser factory" if (!$sparser_factory);
                    my $sparser = $sparser_factory->sparser(
                            filesize => $slice->get_size_whole,
                            block_size => $slice->get_block_size,
                            offset => $slice->get_offset,
                            range => $slice->get_size,
                            sparse_ratio => $slice->get_sparse_ratio,
                            sparse_level =>$slice->get_sparse_level,
                            sparse_align => $slice->get_sparse_align,
                            );
                    $slice->set_size_on_disk( $sparser->get_size_on_disk);
                }
                else{
                    $slice->set_size_on_disk( $slice->get_size );
                }
                $slice->init;
                $sliced_elements->{$slice_key} = $slice;
                $total_slices_size += $slice->get_size;
                $i ++;
            }
            if ( $total_slices_size != $element_size ){
                confess "Total size of slices for " . $element->get_path ." is not equal to the original file size!" .
                        " $total_slices_size instead of $element_size";
            }
        }
        #
        # Make sure we created enough slices
        #
        my $new_elements_count = scalar keys %$sliced_elements;
        if ($new_elements_count < $count){
            confess "We didn't create enough file slices ($new_elements_count instead of $count):\n"
                        . as_string($sliced_elements);
        }
        #
        # Make sure we create any slice at all
        #
        if ($new_elements_count == $initial_elements_count){
            confess "We didn't create any file slices! min_size was $min_size. Elements:\n"
                    . as_string($sliced_elements);
        }
        return $sliced_elements;
    }

    # TODO: Drop partitions that are above the average
    sub _create_partitions {
        my ($self,$args) = @_;
        my $input_elements = $args->{elements};
        my $count = $args->{count};
        confess "No elements passed in" if (!$input_elements);
        confess "No count passed in" if (!$count);

        #
        # Each partition is a Perf::DataSet::Container
        #
        my $ps;
        for my $idx (1..$count){
            push @$ps, __PACKAGE__->new;
        }
        #
        # Sort elements by decreasing order of size
        #
        my @sorted_keys = sort { $input_elements->{$b}->get_size <=> $input_elements->{$a}->get_size }
                                            keys %$input_elements;
        #
        # Distribute elements among partitions
        #
        # Don't rely on 'get_size' container method to get the size
        # as it's very slow for large number of elements
        my @ps_size = (0) x $count;
        my $p_idx;
        ROUND: while (1){
            # Travel the partitions from top to bottom
            $p_idx = 0;
            while($p_idx < $count){
                my $p = $ps->[$p_idx];
                my $key = shift @sorted_keys;
                if (!$key){
                    # No more elements in this type
                    last ROUND;
                }
                $p->set_element($input_elements->{$key});
                $ps_size[$p_idx] += $input_elements->{$key}->get_size;
                # Keep adding elements until the next partition is smaller than the current one
                if ( $p_idx < $count - 1 &&
                        #$p->get_size('files') <= $ps->[$p_idx + 1]->get_size('files') ){
                        $ps_size[$p_idx]  <= $ps_size[$p_idx + 1]  ){
                    next;
                }
                $p_idx ++;
            }
            # Travel the partitions from bottom to top
            $p_idx = $count - 1;
            while($p_idx >= 0){
                my $p = $ps->[$p_idx];
                my $key = shift @sorted_keys;
                if (!$key){
                    # No more elements in this type
                    last ROUND;
                }
                $p->set_element($input_elements->{$key});
                $ps_size[$p_idx] += $input_elements->{$key}->get_size;
                # Keep adding elements until the next partition is smaller than the current one
                if ( $p_idx > 0 &&
                        #$p->get_size('files') <= $ps->[$p_idx - 1]->get_size('files') ){
                        $ps_size[$p_idx]  <= $ps_size[$p_idx - 1]  ){
                    next;
                }
                $p_idx --;
            }
        }
        return $ps;
    }

=head2 random_file

Returns a non-deleted random file from the collection

=over

=item Arguments

B<weight> = undef | 'size'

I<Optional> Setting this to 'size' increases chances of larger files to get picked

=item Examples

    $file = $container->random_file(weight => 'size');

=back

=cut

    sub random_file {
        my ($self,%args) = @_;
        return $self->random_element(%args, type=>'files');
    }

    sub random_element {
        my ($self,%args) = @_;
        my $type = $args{type};
        my $weight = $args{weight};
        my $include_deleted = $args{include_deleted} || 0;
        if (!$weight){
            my $key = get_random_element($self->get_elements(   $type,
                                                                no_copy=>1,
                                                                include_deleted=>$include_deleted,
                                                            )
                                        );
            return $self->get_element($key);
        }
        elsif($weight eq 'size'){
            my $total_size = $args{total_size};
            if (!$total_size){
                # Potentially expensive call
                $total_size = $self->get_size($type);
            }
            my $_rand = rand($total_size);
            $self->size_sort_elements($type)
                if (!$elements_size_sorted[$$self]->{$type} || $self->is_size_dirty);

            #
            # Binary search of first element with first index greater or equal to $_rand
            #
            my $array = $elements_size_sorted[$$self]->{$type};
            my $posmin = 0;
            my $posmax = scalar(@$array) - 1;
            my $i = 0;
            while($i<50){
                $i ++;
                my $mid = int (($posmin + $posmax) / 2);
                my $result = $array->[$mid]->[0] <=> $_rand;
                if ($result < 0){
                    $posmin = $posmax, next if ($mid == $posmin && $posmax != $posmin);
                    return $array->[$mid + 1]->[1] if $mid == $posmin;
                    $posmin = $mid;
                }
                elsif($result > 0 ){
                    $posmax = $posmin, next if ($mid == $posmax && $posmax != $posmin);
                    return $array->[$mid]->[1] if $mid == $posmax;
                    $posmax = $mid;
                }
                else{
                    return $array->[$mid]->[1];
                }
            }
            confess "Binary search is broken!";
        }
        else{
            confess "Parameter 'weight' value '$weight' not supported";
        }
    }


=head2 get_paths

Returns an arrayref of all the Element paths present in the collection

=over

=item Arguments

B<type> = 'files' | 'dirs' | other Element type

I<Optional> Will return paths for all Elements by default

B<prefix> = string

I<Optional> Prefixes all the path with string

=item Examples

    $file_paths = $container->get_paths({type => 'files'});

=back

=cut

    sub get_paths {
        my ($self,$args) = @_;
        my $type = $args->{type};
        my $prefix = $args->{prefix};
        my @paths;

        my $es = $self->get_elements($type);
        my @element_keys = keys %$es;
        if ($prefix){
            @paths = map {"$prefix/" . $es->{$_}->get_path } @element_keys;
        }
        else{
            @paths = map { $es->{$_}->get_path } @element_keys;
        }
        return \@paths;
    }

=head2 contains

Returns 1 if the collection contains at least one Element corresponding to each path passed as
input parameter.

=over

=item Arguments

arrayref of strings

B<Required> list of paths to evaluate

B<type> = 'files' | 'dirs' | any other Element type

I<Optional> Restrict match to Elements of a specific type

=item Examples

    $paths = [ $path1, $path2 ];
    if ( !$container->contains( $paths, {type=>'files'} )){
        confess "Container does not contain all our filepaths";
    }

=back

=cut

    sub contains {
        my ($self,$paths,$args) = @_;
        my $args_type = $args->{type};
        if (defined($args_type) && ! $self->contains_type($args_type)){
            confess "Not a valid type: $args_type";
        }
        if ( ! ref($paths) ){
            # We were passed a scalar
            $paths = [$paths];
        }
        if (ref($paths) ne 'ARRAY'){
            confess "Invalidate argument of type " . ref($paths) . ":\n" . as_string($paths);
        }
        #
        # This does not handle file slices
        #
        my @types_to_check = ($args_type);
        if (!$args_type){
            @types_to_check = ('files','dirs');
        }
        my @types_found;
        my $types_elements = {};
        map { $types_elements->{$_} = $self->get_elements($_) } @types_to_check;
        for my $path (@$paths){
            my $found = 0;
            for my $t (@types_to_check){
                if (exists $types_elements->{$t}->{$path}){
                    $found = 1;
                    push @types_found, $t;
                    last;
                }
            }
            if (!$found){
                $Log->comment("Element with path '$path' does not exist in Container");
                return 0;
            }
        }
        if (scalar @types_found == 1){
            return $types_found[0];
        }
        else{
            return \@types_found;
        }
    }

    sub size_of($$) {
        my ($self,$key) = @_;
        my $e = $self->get_element($key);
        confess "No element with key '$key' found" if (!$e);
        confess "Element with key '$key' is deleted" if ($e->is_deleted);
        return $e->get_size;
    }

=head2 filter

Returns a Container object containing only the Elements matching the criteria entered as input

=over

=item Arguments

B<type> = 'files' | 'dirs' | any other Element type

I<Optional> Returns only Elements of that type

B<min_size> = number (in Bytes)

I<Optional> Returns only Elements with a size greater than that number

B<max_size> = number (in Bytes)

I<Optional> Returns only Elements with a size lower than that number

B<strict_size> = 0 | 1

I<Optional> Combines with min_size or max_size to mean 'strictly lower than' and
'stricly greater than' respectively. Defaults to 0.

B<level> = number

I<Optional> Returns only Elements that are at a specific level in the folder hierarchy. 0 means root
folder, 1 means subfolders of the root, and so on.

B<block_size> = number in bytes

I<Optional> Returns only Elements with that block size

B<compatible_with_io_size> = number in Bytes

I<Optional> Returns only Elements whose block size is a factor of that number

B<paths> = arrayref of strings

I<Optional> Returns only Elements whose path matches with one of the paths specified

=item Examples

    # Get all the files that can handle 32k reads
    $filtered_container = $container->filter({ type=>'files', compatible_with_io_size=>32768 });

=back

=cut

    #
    # Returns Perf::DataSet::Container containing elements with min<= size <=max
    # If 'strict' is set, returns files with min < size < max
    # Also filters files based on block_size. If "strict_block_size" isn't set
    # it will match all file which block size is a factor of block_size
    #
    sub filter {
        my ($self,$args) = @_;
        my $type = $args->{type};
        my $minsize = $args->{min_size}?$args->{min_size}:0;
        my $maxsize = $args->{max_size};
        my $strict_size = $args->{strict_size};
        my $level = $args->{level};
        my $block_size = $args->{block_size};
        my $io_size = $args->{compatible_with_io_size};
        my $paths = $args->{paths};

        if ($io_size && $block_size ){
            if ($io_size % $block_size != 0){
                confess "The 'block_size' ($block_size) you entered is not compatible with ".
                        "the 'io_size' ($io_size) you entered";
            }
            # io_size is redundant
            undef $io_size;
        }
        if (!$type && $block_size){
            confess "You need to specify a type of 'files' to do filtering on block_size";
        }
        if ($paths && !$type){
            $type = 'files';
        }

        #
        # Filter by type
        #
        my $result = $self;
        if ($type && $self->count){
            $result = __PACKAGE__->new;
            $result->set_elements($self->get_elements($type));
        }
        #
        # Filter by level
        #
        if (defined($level) && $result->count){
            my @level_elements = grep { $_->get_level == $level } values %{$result->get_elements};
            $result = __PACKAGE__->new;
            $result->set_elements(\@level_elements);
        }

        #
        # Filter by path
        #
        if ($paths && $result->count){
            my @path_elements = grep { $_->path_in($paths) } values %{ $result->get_elements };
            $result = __PACKAGE__->new;
            $result->set_elements(\@path_elements);
        }

        #
        # Filter by size
        #
        if ( ($minsize || $maxsize) && $result->count ){
            my @size_elements;
            for my $e (values %{$result->get_elements}){
                my $size = $e->get_size;
                if (!$strict_size){
                    next if ($size < $minsize || ( $maxsize && $size > $maxsize ) );
                    push @size_elements, $e;
                }
                else{
                    next if ($size <= $minsize || ( $maxsize && $size >= $maxsize ) );
                    push @size_elements, $e;
                }
            }
            $result = __PACKAGE__->new;
            $result->set_elements(\@size_elements);
        }
        #
        # Now filter by block sizes if the type is 'files'
        #
        if ($type && ($type eq 'files') && ($block_size || $io_size) && $result->count ){
            my @block_size_elements;
            for my $e ( values %{ $result->get_elements } ) {
                my $file_block_size = $e->get_block_size;
                if ($block_size && $block_size == $file_block_size){
                    push @block_size_elements, $e;
                }
                elsif ($io_size && $io_size % $file_block_size == 0){
                    push @block_size_elements, $e;
                }
            }
            $result = __PACKAGE__->new;
            $result->set_elements(\@block_size_elements);
        }

        return $result;
    }

    sub get_random_filename {
        my ($self,%args) = @_;
        my $name = delete $args{name} || "";
        $name = join('.', 'file', $name);
        return $self->_pad_name(name=>$name, %args);
    }

    sub get_random_dirname {
        my ($self,%args) = @_;
        my $name = delete $args{name};
        $name = join('.', 'dir', $name);
        return $self->_pad_name(name=>$name, %args);
    }

    sub _pad_name {
        my ($self,%args) = @_;
        my $length = $args{length};
        $length = 28 if (!$length);
        my $name = $args{name};
        confess "Argument 'name' is missing" if (!$name);
        my $rng = $args{rng};
        $rng = sub { int(rand($_[0])) } if (!$rng);

        # Pick an acceptable filename size
        my $padlen = max( $rng->($length - length($name)) + 1, 5);
        # Pad the name with random letters and digits
        my @paddata;
        for my $digit (1..$padlen){
            my $r = int(rand($alphabet_count));
            push @paddata,$alphabet[$r];
        }
        my $padstring = join("",@paddata);
        return $name . ".$padstring";
    }

    sub clear_on_dirty_for_all {
        my ($self) = @_;
        my $es = $self->get_elements;
        my $e;
        while ( (undef, $e) = each %$es ){
            $e->clear_on_dirty;
        }
    }

    sub add_on_dirty_to_all {
        my ($self) = @_;
        my $es = $self->get_elements;
        my $e;
        while ( (undef, $e) = each %$es ){
            $self->add_on_dirty_to($e);
        }
    }

    sub add_on_dirty_to {
        my ($self, $element) = @_;
        confess "Missing argument" if (!$element);
        # Storing a strong reference to self inside each Element would
        # cause memory leaks since $self already contains references to
        # those Elements
        weaken($self);

        my $callback = sub { return if (!$self); $self->element_is_dirty($_[0]) };
        $element->add_on_dirty( $callback );
        return;
    }

    sub element_is_dirty {
        my ($self, $element) = @_;
        confess "Missing argument" if (!$element);
        $self->set_element($element);
        return;
    }

    sub size_sort_elements {
        my ($self,$type) = @_;
        # Sort elements in decreasing size order
        # TODO: Is this needed?
        my @size_sorted = sort { $b->get_size  <=> $a->get_size  } values %{$self->get_elements($type, include_deleted=>0, no_copy=>1)};
        # Create a structure easy to traverse from random_element
        my $total_size = 0;
        my @cumulative_size_sorted = map { [$total_size += $_->get_size, $_] } @size_sorted;
        $elements_size_sorted[$$self]->{$type} = \@cumulative_size_sorted;
        return;
    }

=head2 relocate_element

Relocates an element.  This is primarily intended for Elements that
have been cloned to different file paths/names but need to maintain
the same internal data.

=over

=item Arguments

string

I<Required> Element key

string

I<Required> New element path

=back

=cut

    sub relocate_element {
        my ($self,$key,$new_path) = @_;

        # handle slice key notation
        (my $old_path = $key) =~ s{/[[:digit:]]+$}{}msx;
        my $file_elements = $self->get_elements('files', include_deleted => 0);

        my @keys = grep { m{^$old_path}msx } keys(%$file_elements);

        foreach my $key (@keys) {
            my $element = $file_elements->{$key};
            $element->relocate($new_path);
            delete $elements[$$self]->{'files'}->{$key};
            $elements[$$self]->{'files'}->{$element->get_key()} = $element;
        }
    }
}
1;
