# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

package Perf::DataSet::ContainerStats;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Perf::Utils qw();
    use Object::InsideOut qw(Storable);

    my @fields              :Field
                            :Name('fields');
    my @definitions         :Field
                            :Name('definitions');
    my @fields_by_type      :Field
                            :Name('definitions_by_type');

    sub _init :Init {
        my $self = $_[0];
        $fields[$$self]         = {};
        $definitions[$$self]    = {};
        $fields_by_type[$$self] = {};
    }

    sub register($$){
        my ($self,$element) = @_;
        if (!$element || !$element->isa('Perf::DataSet::Element')){
            confess "Argument must be a 'Perf::DataSet::Element'";
        }
        my $element_type = $element->get_type;
        my %element_defs = $element->get_stats_definitions;
        for my $stat (keys %element_defs){
            my $stored_def = $definitions[$$self]->{$stat};
            if ($stored_def && $stored_def ne $element_defs{$stat}){
                confess "Definition for stat '$stat' ($element_defs{$stat}) "
                        ."in Element '" . ref($element) . "' doesn't match "
                        ."existing ($stored_def)";
            }
            elsif (!$stored_def ){
                # Record new field
                $definitions[$$self]->{$stat} = $element_defs{$stat};
            }

            my $type_fields = $fields_by_type[$$self]->{$element_type};
            if (!$type_fields){
                $Log->debug("$$self: First registered field for type '$element_type' is '$stat'");
                $fields_by_type[$$self]->{$element_type} = [$stat];
                if (!$self->field_is_defined($stat)){
                    $self->reset_field($stat);
                }
            }
            else{
                my $field_already_registered = grep(/^$stat$/, @$type_fields);
                if (!$field_already_registered){
                    $Log->debug("$$self: Registering field '$stat' for type '$element_type'");
                    push @$type_fields, $stat;
                    if (!$self->field_is_defined($stat)){
                        $self->reset_field($stat);
                    }
                }
            }
        }
    }

    sub compute($$) {
        my ($self,$container) = @_;
        confess "Stats are already being recomputed!" if ($container->is_recomputing);
        $container->set_recomputing(1);
        my $es = $container->get_elements(include_deleted=>0);
        for my $e (values %$es){
            #next if ($e->is_deleted);
            my $e_type = $e->get_type;
            $Log->debug("$$self: Gathering stats from '$e_type' Element with key " . $e->get_key);
            $self->register($e);
            my $type_fields = $fields_by_type[$$self]->{$e_type};
            if (!$type_fields){
                confess "$$self: Type of elements '$e_type' not registered:\n" .
                        Utils::as_string($fields_by_type[$$self]);
            }
            for my $field (@$type_fields){
                my $field_value = $e->get_field($field);
                $self->set_field( $field, $field_value, $e_type);
            }
        }
        $container->set_recomputing(0);
        return;
    }

    sub reset_field {
        my ($self,$field) = @_;
        if (!$field){
            confess "$$self: You must specify a field";
        }
        $fields[$$self]->{$field} = {};
        return;
    }

    sub as_string {
        my $self = $_[0];
        return Perf::Utils::as_string( $self->as_hash );
    }

    sub as_hash {
        my $self = $_[0];
        return $self->dump->[1]->{'Perf::DataSet::ContainerStats'}->{'fields'};
    }

    sub set_field {
        my ($self,$field,$value,$type) = @_;
        confess "$$self: Field '$field' has not been registered" if (! exists $fields[$$self]->{$field} );
        my $name = $definitions[$$self]->{$field};
        if ($name eq 'SUM_BY_TYPE'){
            if (!defined($value) || !defined($type)){
                confess "$$self: You must pass a value and a type";
            }
            $fields[$$self]->{$field}->{ $type }   += $value;
            $fields[$$self]->{$field}->{ 'total' } += $value;
            $Log->debug("ContainerStats $$self: set field '$field' with value '$value'. $field($type) = $fields[$$self]->{$field}->{ $type }");
        }
        elsif($name eq 'COUNT'){
            if (!defined($value)){
                confess "$$self: You must pass a value";
            }
            $fields[$$self]->{$field}->{$value} ++;
            $Log->debug("ContainerStats $$self: set field '$field' with value '$value'. $field($value) = $fields[$$self]->{$field}->{$value}");
        }
        else{
            confess "$$self: Unsupported '$name'";
        }
    }

    sub get_field {
        my ($self,$field,$arg) = @_;
        return if (! exists $fields[$$self]->{$field} );
        my $name = $definitions[$$self]->{$field};
        if ($name eq 'SUM_BY_TYPE'){
           $arg = 'total' if (!defined($arg));
           return $fields[$$self]->{$field}->{$arg};
        }
        elsif($name eq 'COUNT'){
           return $fields[$$self]->{$field} if (!defined($arg));
           return $fields[$$self]->{$field}->{$arg};
        }
        else{
            confess "$$self: Unsupported '$name'";
        }
    }

    sub get_defined_fields {
        my $self = $_[0];
        my @list = keys %{$fields[$$self]};
        return \@list;
    }
    sub field_is_defined {
        my ($self,$field) = @_;
        return exists $fields[$$self]->{$field};
    }

}
1;
