# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Element - Base class for filesystem objects like a file or a directory.

=head1 DESCRIPTION

This class should not be instantiated directly but rather through a sub-class such as Element::File
or Element::Directory. It provides common methods for all current and future Elements of a Container.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Element;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use MIME::Base64;
    use Storable qw(nfreeze);
    use Readonly;
    use Object::InsideOut qw(Storable);

=head2 get_path

Returns Element's relative path

=head2 get_key

Returns Element's key

=head2 get_size

Returns Element's size

=head2 get_size_on_disk

Returns Element's size on disk

=head2 get_level

Returns Element's level in folder hierarchy (0 means root)

=head2 is_dirty

Returns 1 if Element's size has been modified since inception

=cut

    my @path            :Field
                        :Get('get_path')
                        :Arg(Name=>'path', Mandatory=>1);
    my @key             :Field
                        :Get('get_key')
                        :Set(Name=>'set_key', Perm=>'Restrict(Perf::DataSet::Container)')
                        :Arg(Name=>'key',  Mandatory=>1);
    my @size            :Field
                        :Get('get_size')
                        :Arg(Name=>'size', Mandatory=>1)
                        :Type('numeric');
    my @size_on_disk    :Field
                        :Name('size_on_disk')
                        :Get('get_size_on_disk')
                        :Type('numeric');
    my @on_dirty        :Field
                        :Arg( Name=>'on_dirty', Mandatory=>0)
                        :Type('list(CODE)');
    my @on_dirty_enabled :Field
                        :Name('call_on_dirty')
                        :Default(1);
    my @level           :Field
                        :Get('get_level')
                        :Name('level');
    my @dirty           :Field
                        :Name('dirty')
                        :Get('is_dirty');
    my @deleted         :Field
                        :Name('deleted')
                        :Get('is_deleted');
    my @invalid         :Field
                        :Name('invalid')
                        :Get('is_invalid')
                        :Set('set_invalid');
    my @created         :Field
                        :Name('created')
                        :Get('was_created')
                        :Default(0);

    Readonly::Scalar our $NOT_FOUND => -1;

=head2 new

Returns an object representing an Element. Should never be called by user code. Use Element sub-class
new method instead.

=over

=item Arguments

B<size> = number in Bytes

I<Required> Size of the file or slice

B<size_on_disk> = number if Bytes

I<Required> Amount of non-holes in file or slice

B<path> = string

I<Required> Relative path of this file

B<key> = string

I<Required> Unique identifier for this Element. Usually same as path, except for slices.

=back

=cut

    sub _init :Init {
        my $self = $_[0];

        # The dirty flag is used to indicate an element got changed after creation
        # Elements are therefore instantiated as "undirty"
        $dirty[$$self] = 0;

        # The deleted flag is used to indicate an element got deleted after creation
        # Elements are therefore instantiated as "undeleted"
        $deleted[$$self] = 0;

        # The invalid flag indicates that element may not accurately reflect what's on filer
        $invalid[$$self] = 0;

        # Sparse files will override that default setting
        $size_on_disk[$$self] = $size[$$self];

        #
        # Figure out what directoy level that path belongs.
        # 0 = root
        #
        my $tmp = $path[$$self];
        # remove start and end slash character if any. Paths should not start with a'/'.
        $tmp =~ s/\/$//;
        $tmp =~ s/^\///;
        $_ = $tmp;
        my @levels = /\//g;
        my $path_level = scalar @levels;
        $level[$$self] = $path_level;

    }

    sub _replicate :Replicate {
        my ($self, $clone, $flag) = @_;
        # This will be called in the context of a shallow cloning
        confess "Deep cloning not supported" if ($flag eq 'deep');
        # Don't inherit on_dirty callbacks
        delete $on_dirty[$$clone];
        return;
    }

=head2 set_size_on_disk

Changes Element's size on disk

=over

=item Arguments

number

I<Required> Size in bytes

=back

=cut

    sub set_size_on_disk {
        my ($self,$arg) = @_;
        if ($size_on_disk[$$self] != $arg){
            $size_on_disk[$$self] = $arg;
            $self->dirty;
        }
    }

=head2 set_size

Changes Element's size

=over

=item Arguments

number

I<Required> Size in bytes

=back

=cut

    sub set_size {
        my ($self,$arg) = @_;
        if ($size[$$self] != $arg){
            $size[$$self] = $arg;
            $self->dirty;
        }
    }

    sub disable_on_dirty :Restricted {
        my ($self) = @_;
        $on_dirty_enabled[$$self] = 0;
    }

    sub enable_on_dirty :Restricted {
        my ($self) = @_;
        $on_dirty_enabled[$$self] = 1;
    }

    sub undirty {
        my $self = $_[0];
        $dirty[$$self] = 0;
    }

    sub dirty {
        my $self = $_[0];
        $dirty[$$self] = 1;
        if ($on_dirty_enabled[$$self]){
            confess "No on_dirty callbacks found!" if (!$on_dirty[$$self] || !@{$on_dirty[$$self]});
            for my $callback (@{$on_dirty[$$self]}){
                $callback->($self);
            }
        }
    }

=head2 get_type

Returns Element's type (i.e. 'files', or 'dirs, etc...)

=over

=item Arguments

None

=back

=cut

    sub get_type {
        my $self = $_[0];
        confess "You must implement this subroutine in " . ref($self);
    }

    sub get_stats_definitions :Cumulative {
        return ('type'          => 'COUNT',
                'size'          => 'SUM_BY_TYPE',
                'size_on_disk'  => 'SUM_BY_TYPE',
                );
    }

=head2 path_in

Returns 1 if Element's path is in the list passed as argument

=over

=item Arguments

arrayref of strings

I<Required> Paths to evaluate

=back

=cut

    sub path_in {
        my ($self,$paths) = @_;
        if (!$paths || ref($paths) ne 'ARRAY'){
            confess "Invalid parameters";
        }
        my $match = grep { /^$path[$$self]$/ } @$paths;
        return $match;

    }

=head2 get_field

Returns specific Element field. See subclasses of Element for available fields.

=over

=item Arguments

strings

I<Required> Element field

=back

=cut

    sub get_field {
        my ($self, $field) = @_;
        confess "No field specified" if (!$field);
        my $method = "get_$field";
        if (!$self->can($method)){
            $method = "is_$field";
            confess "Field '$field' not available" if (!$self->can($method));
        }
        my $value;
        $value = $self->$method();
        return $value;
    }

=head2 is_sliceable

Returns 1 if Element can be split up into smaller pieces

=over

=item Arguments

None

=back

=cut

    sub is_sliceable {
        return 0;
    }

=head2 is_slice

Returns 1 if Element is a smaller piece of a larger Element

=over

=item Arguments

None

=back

=cut

    sub is_slice {
        return 0;
    }

    # Prepare element for RPC network communication
    # Don't send binary data across only text
    sub rpc_serialize {
        my $self = $_[0];
        # Remove the callbacks
        my $on_dirty_callbacks = delete $on_dirty[$$self];
        my $frozen = nfreeze($self);
        my $encoded = encode_base64($frozen,'');
        # Put the callbacks back
        $on_dirty[$$self] = $on_dirty_callbacks;
        return $encoded;
    }

    sub as_hash {
        my ($self,$hide_fields) = @_;
        my $dump = $self->dump->[1];
        my $hash = {};
        # Sort class names in inverse alphabetical order
        # this should take care of inheritance hierarchy
        # given the naming convention chosen for Elements
        my @classes = sort { $b cmp $a } keys %$dump;
        for my $class (@classes){
            my $class_fields = $dump->{$class};
            # If key already exists, ignore so that lowest level value is kept
            map { $hash->{$_} = $class_fields->{$_} if (!exists $hash->{$_} && !$hide_fields->{$_}) }
                keys %$class_fields;
        }
        return $hash;
    }

=head2 as_string

Returns a string description of the Element's fields. This is required since using
Data::Dumper on the object itself won't work (see Object::InsideOut documentation)

=over

=item Arguments

None

=back

=cut

    sub as_string {
        my $self = shift;
        my $string = Dumper( $self->as_hash(@_) );
        $string =~ s/\$VAR1\s=\s//;
        $string =~ s/;$//;
        return ref($self) . ":\n$string";
    }

    sub mark_deleted {
        my ($self) = @_;
        if (!$deleted[$$self]){
            $deleted[$$self] = 1;
            $self->dirty;
        }
        return;
    }

    sub clear_deleted {
        my ($self) = @_;
        $deleted[$$self] = 0;
        $self->dirty;
        return;
    }

    sub add_on_dirty {
        my ($self, $callback) = @_;
        confess "Missing arguments" if (!$callback);
        confess "Callback must be a coderef" if ( ref($callback) ne 'CODE' );
        push @{$on_dirty[$$self]}, $callback;
        return;
    }

    sub clear_on_dirty {
        my ($self) = @_;
        return delete $on_dirty[$$self];
    }

    sub mark_created {
        my ($self) = @_;
        $created[$$self] = 1;
        return;
    }

=head2 get_actual_size

Uses system calls on the client to find out how much space the OS
believes a File contains.

=over

=item Arguments

None

=back

=cut

    sub get_actual_size {
        my ($self) = @_;
        my $path = $path[$$self];
        my @stats = stat($path);
        my $size;

        if (!@stats){
            my $err = $!;
            if (! -e $path) {
                $size = $NOT_FOUND;
            }
            else{
                confess "Unable to get size of '$path': $err";
            }
        }
        else{
            $size = $stats[7];
        }

        return $size;
    }

=head2 relocate

Changes the path and key of an element.  This is primarily intended
for Elements that have been cloned and now have different file names.

=over

=item Arguments

string

I<Required> The new name of the element

=back

=cut

    sub relocate {
        my ($self, $new_path) = @_;

        my $old_key = $key[$$self];
        my $old_path = quotemeta($path[$$self]);
        $path[$$self] = $new_path;
        $key[$$self] =~ s/$old_path/$new_path/;

        $Log->comment("Element $old_key relocated to $key[$$self]");
    }
}
1;
