# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Element::File - Representation of a file

=head1 DESCRIPTION

This class represent a file tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Element::File;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw($Log param);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::Utils qw( in_bytes ceil );
    use Data::Dumper qw(Dumper);
    use Digest::MD5 qw(md5 md5_base64);
    use Object::InsideOut qw(Perf::DataSet::Element);
    use Net::Domain qw(hostname);
    use MIME::Base64 qw(encode_base64);
    use File::Path qw(mkpath);
    use Params::Validate;
    use Inline;
    use Time::HiRes qw(time);

    BEGIN{
        # Disable Inline::C optimizations on Windows since there is no a C compiler
        if ($^O =~ /Win32/i || param('NO_INLINE')){
            *generate_block = *generate_block_perl;
        }
        else{
            #
            # Create a different folder per host so compiled stuff matches with host's platform
            # Create a different folder per user too, so permissions don't get in the way
            # Inline::C won't recompile unless C source code changes
            #
            my $inline_prefix = '.';

            if (! -w $inline_prefix) {
                # CTL doesn't allow writing of test related files anywhere
                # except the log directory but doesn't want Inline::C disabled
                # Compile this every time in the LOGDIR for them.
                require NATE::ParamSet;
                my $param = NATE::ParamSet->new(global => 1);
                $inline_prefix = $param->get('LOGDIR_UNIX');
            }

            # Turns out 'getlogin' does not work when run in a remote subtest
            # Use cuserid instead
            require POSIX;
            my $inline_path = "$inline_prefix/_Inline/" . POSIX::cuserid() .'/' . hostname();
            print "Inline path: $inline_path\n";
            mkpath($inline_path, {verbose=>1, mode=>0777});
            Inline->import( C => 'DATA',
                            DIRECTORY   => $inline_path,
                            BUILD_NOISY => 1,
                            BUILD_TIMERS=> 1,
                            CLEAN_BUILD_AREA=>1,
                            CLEAN_AFTER_BUILD=>1,
                            WARNINGS    => 1,
                            PRINT_INFO  => 1,
                            );
            *generate_block = *generate_block_c;
        }
    }
    Inline->init;

=head2 get_block_size

Returns File's block size

=head2 get_size_whole

In case the object is actually a File slice, return the size of the entire File. Otherwise
return the File size.

=head2 get_offset

In case the object is actually a File slice, returns the offset in the file this slice
starts at. Otherwise return 0.

=head2 get_sparse

Returns 1 if file is sparse

=head2 get_sparse_ratio

Returns sparse ratio, i.e. ration of holes versus total number of blocks at level sparse_level
in the file.

=head2 get_sparse_level

Returns level sparsing occured at in the file. 0 means L0, 1 means L1, etc..

=head2 get_first_block

Returns first block number in File. Returns 0 if File is not a slice.

=head2 get_last_block

Returns last block number in File.

=cut

    my @block_size      :Field
                        :Get('get_block_size')
                        :Arg(Name => 'block_size', Mandatory => 1)
                        :Type('numeric');
    my @size_whole      :Field
                        :Get('get_size_whole')
                        :Arg(Name => 'size_whole', Mandatory => 1)
                        :Type('numeric');
    my @offset          :Field
                        :Get('get_offset')
                        :Set(Name=>'set_offset', Perm=>'Restrict(Perf::DataSet::Container)')
                        :Arg(Name => 'offset', Mandatory => 0, Default => 0)
                        :Type('numeric');
    my @no_checksums    :Field
                        :Arg(Name=>'no_checksums', Mandatory=>0, Default=>0)
                        :Get('get_no_checksums');

    my @first_block     :Field :Name('first_block') :Get('get_first_block');
    my @last_block      :Field :Name('last_block')  :Get('get_last_block');
    my @checksums       :Field :Name('checksums');
    my @blockmap        :Field :Name('blockmap') :Deep;
    my @invalid_blocks  :Field :Name('invalid_blocks');
    my @fbns_per_block  :Field :Name('fbns_per_block') :Get('get_fbns_per_block');

    # Use to generate random strings
    # Let @alphabet = all printable characters.  In ASCII, '!' (33d)
    # is the first printable character and '~' (126d) is the last.
    my @alphabet = map { chr } ord('!')..ord('~');

    my $alphabet_count = scalar(@alphabet);


    my $zero_blocks = {};

=head2 new

Returns an object representing a file or a file slice.

=over

=item Arguments inherited from Element

See L<Element/new>.

=item Arguments

B<block_size> = number in Bytes

I<Required> Size of the data pattern written to disk

B<size_whole> = number in Bytes.

I<Required> Size of the entire file if object is a slice

B<offset> = number in Bytes

I<Optional> Must align with block_size. Defaults to 0.

B<sparse> = 0 | 1

I<Optional> Whether the file is sparse. Defaults to 0.

B<sparse_ratio> = number between 0 and 1

I<Optional> Ratio of holes versus non-holes at sparse level. Defaults to 1.

B<sparse_level> = number

I<Optional> Block level at which ratio applies. 0 means L0, etc. Defaults to 0.

=item Examples

    $file = Perf::DataSet::Element::File->new( path=>$path, key=>$key, ... );

=back

=cut

    sub init :Init(Public) {
        my $self = $_[0];
        if ($offset[$$self] < 0){
            confess "offset cannot be negative\n" . $self->as_string;
        }
        if ($size_whole[$$self] < $self->get_size + $offset[$$self] ){
            confess "Size + offset must be less than size_whole\n" . $self->as_string;
        }
        $first_block[$$self] = int( $offset[$$self] / $block_size[$$self] );
        $last_block[$$self] = ceil( ($offset[$$self] + $self->get_size) / $block_size[$$self] ) - 1;
        if ($first_block[$$self] > $last_block[$$self]){
            # Make an exception for empty files where first_block will be set 0
            # but last_block to -1
            if ($self->get_size){
                confess "First block ($first_block[$$self]) larger than last block ($last_block[$$self])!\n"
                        . $self->as_string;
            }
        }
        $fbns_per_block[$$self] = $block_size[$$self]/4096;
        $invalid_blocks[$$self] = { };
    }

    sub create_checksum_store {
        my $self = shift;
        my @checksum_array;
        $#checksum_array = $last_block[$$self] - $first_block[$$self] + 1;
        $checksums[$$self] = \@checksum_array;
    }

    sub compute_size_on_disk {
        my ($self) = @_;
        $self->set_size_on_disk( $self->get_size );
    }

    sub get_sparse {
        return 0;
    }

=head2 clone

Returns a copy of this object. All fields are duplicated except for the checksums values.

=over

=item Arguments

None

=item Examples

    $clone = $file->clone;

=back

=cut

    # During cloning, checksums require specific processing so that values don't get unnecessarily
    # duplicated
    sub _replicate :Replicate
    {
        my ($self, $clone, $flag) = @_;
        # This will be called in the context of a shallow cloning
        confess "Deep cloning not supported" if ($flag eq 'deep');
        return if (!$checksums[$$self]);
        # We want a new array, but not new values (since those are refernces
        # to checksums we want to keep shared as much as possible)
        my @copy = @{$checksums[$$self]};
        $checksums[$$clone] = \@copy;
        return;
    }

    sub as_hash {
        my ($self,$hidden_fields) = @_;
        # Hide the checksums and blockmap by default
        # Passing an empty hashref is different from not passing anything
        # and means "no hidden fields"
        $hidden_fields = { 'checksums' => 1, 'blockmap' => 1} if (!defined($hidden_fields));
        return $self->SUPER::as_hash($hidden_fields);
    }

    # Prepare element for RPC network communication
    # Don't send binary data across only text
    # Don't send checksums or blockmap data
    sub rpc_serialize {
        my $self = $_[0];
        #my $clone = $self->clone;
        #$clone->delete_checksums;
        #$clone->delete_blockmap;
        #return $clone->SUPER::rpc_serialize;

        my $c = $self->delete_checksums;
        my $b = $self->delete_blockmap;
        my $rpc_serialized_self = $self->SUPER::rpc_serialize;
        $self->set_checksums($c) if ($c);
        $self->set_blockmaps($b) if ($b);
        return $rpc_serialized_self;
    }

    sub get_type {
        return 'files';
    }

=head2 get_blocks

Returns number of blocks in this file or file slice. "Blocks" here refer to the internal blocks. See
get_wafl_blocks for WAFL blocks.

=over

=item Arguments

None

=back

=cut

    sub get_blocks {
        my $self = $_[0];
        return ceil($self->get_size / $block_size[$$self]);
    }

=head2 get_blocks_on_disk

Returns number of non-holey blocks in this file or file slice. "Blocks" here refer to the internal blocks. See
get_wafl_blocks_on_disk for WAFL blocks.

=over

=item Arguments

None

=back

=cut

    sub get_blocks_on_disk {
        my $self = $_[0];
        return ceil($self->get_size_on_disk / $block_size[$$self]);
    }

=head2 get_wafl_blocks

Returns number of WAFL L0 blocks in this file or file slice. Returns 1 if the size of the File is
less than 4096.

=over

=item Arguments

None

=back

=cut

    # This does not include L1s
    sub get_wafl_blocks {
        my $self = $_[0];
        return ceil($self->get_size / 4096);
    }

=head2 get_wafl_blocks_on_disk

Returns number of non-holey WAFL L0 blocks in this file or file slice. Returns 1 if the size of
the File is less than 4096.

=over

=item Arguments

None

=back

=cut

    # This does not include L1s
    sub get_wafl_blocks_on_disk {
        my $self = $_[0];
        return ceil($self->get_size_on_disk / 4096);
    }

    # Each key of the definition hash must have a corresponding
    # $self->get_<key> method
    sub get_stats_definitions :Cumulative {
        return ('block_size'            => 'COUNT',
                'sparse'                => 'COUNT',
                #'sparse_level'          => 'COUNT',
                'blocks'                => 'SUM_BY_TYPE',
                'blocks_on_disk'        => 'SUM_BY_TYPE',
                'wafl_blocks'           => 'SUM_BY_TYPE',
                'wafl_blocks_on_disk'   => 'SUM_BY_TYPE',
                );
    }

    sub set_size_whole {
        my ($self, $arg) = @_;
        if ($size_whole[$$self] != $arg){
            $size_whole[$$self] = $arg;
            $self->dirty;
        }
    }

    sub is_sliceable {
        my $self = $_[0];
        return 0 if ($self->is_slice);
        return 0 if ($self->has_block_data);
        return 0 if ($self->get_size < 2 * $self->get_block_size);
        return 1;
    }

    sub is_slice {
        my $self = $_[0];
        return 0 if ($size_whole[$$self] == $self->get_size);
        return 1;
    }

=head2 is_last_slice

Returns 1 if the object represents the tail end of the file

=over

=item Arguments

none

=back

=cut

    sub is_last_slice {
        my $self = $_[0];
        return 1 if ($offset[$$self] + $self->get_size == $size_whole[$$self]);
        return 0;
    }

=head2 is_coslice

Returns 1 if the File object given as argument and this object are two slices of the same file

=over

=item Arguments

Element::File

I<Required> File object to evaluate

=back

=cut

    sub is_coslice($$) {
        my ($self,$other) = @_;
        confess "Argument must be a " . __PACKAGE__ if ( !$other || !$other->isa(__PACKAGE__) );
        # Coslices have same path
        return 0 if ( $self->get_path ne $other->get_path );

        # Some sanity checking
        if ( $self->get_size_whole != $other->get_size_whole ){
            confess "Same path has different size_whole!\n" . $self->as_string . "\n"
                    . $other->as_string;
        }
        if ( ($self->get_offset + $self->get_size > $other->get_offset) &&
             ($other->get_offset + $other->get_size > $self->get_offset) ){
             confess "Elements overlap!\n". $self->as_string . "\n"
                    . $other->as_string;
        }
        if ( $self->get_block_size != $other->get_block_size ){
            confess "Elements report different block sizes\n". $self->as_string . "\n"
                    . $other->as_string;
        }
        if ( $self->get_sparse != $other->get_sparse ){
            confess "Elements report different sparse values\n". $self->as_string . "\n"
                    . $other->as_string;
        }
        if ( $self->is_deleted != $other->is_deleted ){
            confess "Elements report different is_deleted values\n". $self->as_string . "\n"
                    . $other->as_string;
        }

        return 1;
    }

    sub slice {
        my ($self, $count) = @_;
        if (!$self->is_sliceable){
            confess "Element is not sliceable:\n" . $self->as_string;
        }
        my $element_size = $self->get_size;
        my $element_block_size = $self->get_block_size;
        my $element_path = $self->get_path;
        my $size_per_slice;
        if ($element_size >= $count * $element_block_size){
            $size_per_slice = int($element_size / $count / $element_block_size)
                                    * $element_block_size;
        }
        else{
            $size_per_slice = $element_block_size;
        }
        my $slice_count = ceil($element_size / $size_per_slice);
        my $size_of_last_slice = $size_per_slice;
        if ($slice_count * $size_per_slice > $element_size){
            #
            # Just make the last slice a little bit bigger so each slice is at least of
            # size $block_size and there are no more than $count slices
            #
            $size_of_last_slice = $element_size - ($slice_count - 2) * $size_per_slice;
            $slice_count --;
        }
        my @slices;
        my $total_slices_size = 0;
        my $i = 0;
        while ($i < $slice_count){
            my $slice_key = "$element_path/$i";
            my $slice_offset = $i * $size_per_slice;

            my $slice = $self->clone; # Shallow copy
            $slice->disable_on_dirty;
            $slice->set_offset( $slice_offset );
            $slice->set_size( $i < $slice_count - 1 ? $size_per_slice : $size_of_last_slice );
            $slice->set_key( $slice_key );
            $slice->set_size_whole( $element_size );
            $slice->compute_size_on_disk;
            $slice->init;
            $slice->enable_on_dirty;
            push @slices,$slice;
            $total_slices_size += $slice->get_size;
            $i ++;
        }
        if ( $total_slices_size != $element_size ){
            confess "Total size of slices for " . $self->get_path ." is not equal to the original file size!" .
                    " $total_slices_size instead of $element_size";
        }
        return \@slices;
    }

    sub unslice {
        my ($self) = @_;
        confess "Not a slice!" if (!$self->is_slice);
        my $reconstructed = $self->clone;
        $reconstructed->disable_on_dirty;
        $reconstructed->set_key( $self->get_path );
        $reconstructed->set_size( $self->get_size_whole );
        $reconstructed->set_offset( 0 );
        $reconstructed->compute_size_on_disk;
        $reconstructed->init;
        $reconstructed->enable_on_dirty;
        return $reconstructed;
    }

=head2 has_block_data

Returns 1 if this File object contains data about its own blocks i.e. checksums or blockmap info.
See get_checksum and get_blockmap for more details.

=over

=item Arguments

None

=back

=cut

    sub has_block_data {
        my $self = $_[0];
        if (!$no_checksums[$$self]){
            # @checksums[$$self] is pre-grown to the size of the file
            # so a test on scalar @{$checksums[$$self]} isn't the
            # right thing to do. Instead, test if the first element
            # is defined
            return !! $checksums[$$self]->[0];
        }
        else{
            return !! $blockmap[$$self];
        }
    }

=head2 has_checksums

Returns 1 if this File object contains checksums for its blocks.

=over

=item Arguments

None

=back

=cut

    sub has_checksums {
        my $self = $_[0];
        return 0 if ($no_checksums[$$self]);
        return !! $checksums[$$self]->[0];
    }

    sub keep_checksums {
        my ($self,$value) = @_;
        confess "Invalid value $value" if (defined($value) && $value !~ /^(?:0|1)$/);
        if (defined($value)){
            my $not_value = $value?0:1;
            if (!defined($no_checksums[$$self])){
                $Log->comment("Setting no_checkum to $not_value for '" . $self->get_path ."'");
                $no_checksums[$$self] = $not_value;
            }
            elsif ($no_checksums[$$self] != $not_value){
                confess "You are not allowed to change this setting once set";
            }
        }
        return !$no_checksums[$$self];
    }

    sub delete_checksums :Restricted {
        my $self = $_[0];
        return delete $checksums[$$self];
    }

    sub delete_blockmap :Restricted {
        my $self = $_[0];
        return delete $blockmap[$$self];
    }

    sub set_checksums :Restricted {
        my ($self, $c) = @_;
        confess "Missing argument" if (!$c);
        $checksums[$$self] = $c;
        return;
    }

    sub set_blockmaps :Restricted {
        my ($self, $b) = @_;
        confess "Missing argument" if (!$b);
        $blockmap[$$self] = $b;
    }

=head2 before_eof

Returns portion of range contained before the end of file. Range is defined by an offset and a length.

    Example:
        For a $file of size 32k, $file->before_eof('16k', '64k') would return 16384.

=over

=item Arguments

number or string

I<Required> Offset indicating start of range

number of string

I<Required> Length of range

=back

=cut

    sub before_eof {
        my ($self, $offset, $max_length) = @_;
        confess "Invalid arguments" if (! defined $offset || ! defined $max_length );
        $max_length = in_bytes($max_length);
        $offset = in_bytes($offset);
        my $before_eof = $self->get_end_offset - $offset;
        return ($before_eof >= $max_length)? $max_length : $before_eof;
    }

    sub set_size {
        my ($self,$size) = @_;
        $self->SUPER::set_size($size);
        $last_block[$$self] = ceil( ($offset[$$self] + $size) / $block_size[$$self] ) - 1;
        return;
    }

    sub record_block($$$) {
        my ($self,$block_num,$block_data) = @_;
        if ($block_num > $last_block[$$self] || $block_num < $first_block[$$self]){
            confess "Block $block_num is out of range ($first_block[$$self]-$last_block[$$self])";
        }

        # Once an invalid block is overwritten, it should no longer be
        # considered invalid
        delete($invalid_blocks[$$self]->{$block_num});

        if (!$no_checksums[$$self]){
            my $file_blocksize = $block_size[$$self];
            my $block_data_length = length($$block_data);
            if ( $block_data_length < $file_blocksize){
                my $data = pack("a$file_blocksize", $$block_data);
                $block_data = \$data;
            }
            confess "Block data of wrong size" if (length($$block_data) != $file_blocksize);
            my $checksum = md5($$block_data);
            return $self->set_checksum($block_num,\$checksum);
        }
        else{
            return $self->set_blockmap($block_num);
        }
    }

    sub get_checksum {
        my ($self, $block_num) = @_;
        if ($block_num > $self->get_last_block || $block_num < $self->get_first_block){
            confess "block_num ($block_num) is out of range (". $self->get_first_block .
                    "-" . $self->get_last_block . ")";
        }
        my $c = $checksums[$$self]->[$block_num - $first_block[$$self]];
        if (!$c){
            confess "No checksum data stored for '" . $self->get_path . "'";
        }
        return $$c;
    }

    sub set_checksum {
        my ($self,$block_num,$checksum) = @_;
        confess "Bad arguments" if (!defined($block_num) || !$checksum);
        my $index = $block_num - $first_block[$$self];
        my $r;

        if (!$checksums[$$self]){
            $self->create_checksum_store;
        }
        if ($checksums[$$self]->[$index]){
            $r = $Perf::DataSet::BLOCK_UPDATED;
        }
        else{
            $r = $Perf::DataSet::BLOCK_ADDED;
        }
        $checksums[$$self]->[$index] = $checksum;
        return $r;
    }

    sub get_blockmap {
        my ($self, $block_num) = @_;
        return 1 if ($block_num <= $blockmap[$$self]);
        return 0;
    }

    sub set_blockmap {
        my ($self,$block_num) = @_;
        if (!defined $blockmap[$$self]){
            confess "Operation not allowed. Block_num:$block_num. Self:\n" . $self->as_string
                if ($block_num != $first_block[$$self]);
            $blockmap[$$self] = $block_num;
            return $Perf::DataSet::BLOCK_ADDED;
        }
        elsif ($block_num == $blockmap[$$self] + 1){
            $blockmap[$$self] = $block_num;
            return $Perf::DataSet::BLOCK_ADDED;
        }
        elsif($block_num <= $blockmap[$$self]){
            return $Perf::DataSet::BLOCK_UPDATED;
        }
        else{
            confess "Operation not allowed. Block_num:$block_num. Self:\n" . $self->as_string;
        }
    }

    sub validate_block_data {
        my ($self,$args) = @_;
        my $block_data = $args->{block_data};
        my $block_num = $args->{block_num};
        my $corruptions_ok = $args->{corruptions_ok};
        my $corruptions_print = $args->{corruptions_print};
        my $verify = $args->{verify};
        my $no_checksums = $no_checksums[$$self];

        # if this file contains invalid block, check to see if the block
        # currently being validated was flagged
        $corruptions_ok = 1
          if ($self->is_invalid() || $invalid_blocks[$$self]->{$block_num});

        if (!$verify){
            $verify = 'checksum';
        }
        elsif ('auto' eq $verify) {
            # automatic used for a Perf::DataSet containing a combination of
            # checksum and non-checksum files
            $verify = $no_checksums ? 'pattern' : 'checksum';
        }

        confess 'checksum verification requested'
              . ' but file was created without checksums'
                  if ($no_checksums && 'checksum' eq $verify);

        my $validation_criteria =
            $no_checksums ? $self->get_blockmap($block_num)
                          : $self->get_checksum($block_num);

        # Is this block a zero block in a patterned file or a checksum file?
        if ($no_checksums ? ( 0  == $validation_criteria)
                          : ('0' eq $validation_criteria)) {
            $verify = 'zero';
        }

        my $file_path = $self->get_path;

        if ($verify eq 'checksum') {
            my $file_blocksize = $block_size[$$self];
            my $block_data_length = length($$block_data);
            if ( $block_data_length < $file_blocksize){
                my $data = pack("a$file_blocksize", $$block_data);
                $block_data = \$data;
            }
            my $block_data_checksum = md5($$block_data);
            if ($block_data_checksum ne $validation_criteria){
                if (!$corruptions_ok){
                    my $filepath = $self->get_path;
                    my $msg =
                      "Data corruption found in file '$filepath' block '$block_num'. "
                      . "Checksum for block data does not match with original block checksum.\n"
                      . "Element:\n"
                      . $self->as_string . "\n"
                      . 'Stored Checksum: '
                      . encode_base64($validation_criteria) . "\n"
                      . 'Read Checksum: '
                      . encode_base64($block_data_checksum);

                    # STAF doesn't like binary data or maybe just zeros. Don't include the
                    # data read as it may contain \0s
                    $Log->comment( $msg . "\nBlock data read:\n" . $$block_data );
                    confess $msg;
                }
                else{
                    if ($corruptions_print){
                        my $filepath = $self->get_path;
                        my $msg = "Data corruption found in file '$filepath' block '$block_num'. " .
                            "Checksum for block data does not match with original block checksum.\n" .
                            "Element:\n" . $self->as_string ."\n" .
                            'Stored Checksum: ' . encode_base64($validation_criteria) . "\n" .
                            'Read Checksum: ' . encode_base64($block_data_checksum);
                        # STAF doesn't like binary data or maybe just zeros. Don't include the
                        # data read as it may contain \0s
                        $Log->comment( $msg . "\nBlock data read:\n" . $$block_data );
                    }
                    return 0;
                }
            }
        }
        elsif ($verify eq 'zero') {
            # verify a zero block
            my $zero_block = $self->get_zero_block($block_size[$$self]);
            if ($$block_data ne $zero_block){
                if (!$corruptions_ok){
                    confess "Data corruption found in file '$file_path' block '$block_num'." .
                        "Block should have been full of zeros.\n" .
                        "Element:\n" . $self->as_string . "\n" .
                        "Block data read:\n" . $$block_data;
                }
                else{
                    if ($corruptions_print){
                        $Log->comment("Data corruption found in file '$file_path' block '$block_num'." .
                        "Block should have been full of zeros.\n" .
                        "Element:\n" . $self->as_string . "\n" .
                        "Block data read:\n" . $$block_data);
                    }
                    return 0;
                }
            }
        }
        else { # ($verify eq 'pattern')
            # This needs to be done for each wafl block since that's the size of the pattern
            my $printlen = length($$block_data);
            my $params = { header=>"##$file_path##$block_num##$printlen##",
                           fbn=>$self->get_first_fbn($block_num),
                           corruptions_ok => $corruptions_ok,
                           };
            my $left = length($$block_data);
            my $offset = 0;
            while ($left >= 4096){
                $params->{block_data} = \do{substr($$block_data,$offset,4096)};
                if (!$self->validate_wafl_block_data_pattern($params)){
                    if (!$corruptions_ok){
                        confess "Data corruption found in file '$file_path' block '$block_num'";
                    }
                    else{
                        if ($corruptions_print){
                            $Log->comment("Data corruption found in file '$file_path' block '$block_num'");
                        }
                        return 0;
                    }
                }
                $left -= 4096;
                $offset += 4096;
                $params->{fbn} ++;
            }
            if ($left){
                if ($offset){
                    $params->{block_data} = \do{substr($$block_data,$offset)};
                }
                else{
                    $params->{block_data} = $block_data;
                }
                if (!$self->validate_wafl_block_data_pattern($params)){
                    if (!$corruptions_ok){
                        confess "Data corruption found in file '$file_path' block '$block_num'";
                    }
                    else{
                        if ($corruptions_print){
                            my $msg = "Data corruption found in file '$file_path' block '$block_num'";
                            if ($left == 1){
                                my @a = unpack("B*", $$block_data);
                                $msg .= "\nData read: 0b@a\n";
                            }
                            $Log->comment($msg);
                        }
                        return 0;
                    }
                }
            }
        }

        # Block was found to be valid, remove from invalid list
        delete($invalid_blocks[$$self]->{$block_num});

        return 1;
    }

=head2 set_invalid_blocks

Flag blocks in this File for which no guarantee can be made for valid data.

=over

=item Arguments

B<start> = block number

I<Required> The number of the first invalid block

B<num> = count

I<Required> The total number of invalid blocks after B<start>

=back

=cut

    sub set_invalid_blocks {
        my $self = shift();
        my %args = validate(@_,
                            { start => 1,
                              num   => 1, });

        foreach my $block_num ($args{start}..($args{start} + $args{num} - 1)) {
            $invalid_blocks[$$self]->{$block_num} = 1;
        }

        return 1;
    }


=head2 get_invalid_block_count

Return the number of invalid blocks in this file.

=cut

    sub get_invalid_block_count {
        my $self = shift();

        return scalar(keys(%{$invalid_blocks[$$self]}));
    }


    sub validate_wafl_block_data_pattern {
        my ($self,$args) = @_;
        my $str = join('', $args->{header}, $args->{fbn}, '##');
        # a timestamp is 10 characters + 1 decimal + 6 characters for fractional + 2 #s
        my $strl = length($str) + 19;
        my $block_data = $args->{block_data};
        my $printlen = length($$block_data);

        if ($strl >= $printlen){
            if ($printlen > 22){
                # block data should start with a MD5 hash of str
                my $str_md5 = md5_base64($str);
                return 0 if ( index($$block_data, $str_md5) != 0 );
            }
            elsif (1 == $printlen) {
                # Attempting to verify a 1 byte block.  1 byte blocks
                # use the binary pattern 10RRRRRR where R is a random
                # bit.  Mask off the random bits and verify the 2 MSBs
                return 0 if (0x80 != (ord($$block_data) & 0xC0));
            }
            else{
                # The first byte should be representative of the length of the block
                my $first_byte = chr($printlen + 33);
                return 0 if ( index($$block_data, $first_byte) != 0 );
            }
        }
        else{
            # We expect to see $str at the beginning of the block
            return 0 if ( index($$block_data, $str) != 0 );
        }
        return 1;
    }

    # Returns first fbn in block with block_num
    sub get_first_fbn {
        my ($self, $block_num) = @_;
        return $block_num * $fbns_per_block[$$self];
    }

    sub get_zero_block($$) {
        my ($self,$block_size) = @_;
        if (!$block_size){
               confess "You need to specify a block size";
        }
        if (!$zero_blocks->{$block_size}){
            $zero_blocks->{$block_size} = pack("a$block_size",'');
        }
        return $zero_blocks->{$block_size};
    }

    sub get_last_block_size {
        my ($self) = @_;
        my $blocksize = $block_size[$$self];
        my $last_block_size = $self->get_size % $blocksize;
        $last_block_size = $last_block_size ? $last_block_size : $blocksize;
        return $last_block_size;
    }

    sub get_append_offset {
        my ($self) = @_;
        return ($self->get_last_block + 1) * $self->get_block_size;
    }

    sub get_end_offset {
        my ($self) = @_;
        return $self->get_offset + $self->get_size;
    }


    sub generate_block_c {
        my $self = shift;
        my %args = @_;
        my $file_path = $self->get_path;
        my $block_num = $args{block_num};
        my $printlen = $args{size};
        my $buffer = $args{buffer}; # Reference to a string
        my $offset = $args{offset}; # Where to write in buffer
        $offset = 0 if (!defined($offset));

        if (!defined($block_num) || !defined($printlen)){
            confess "You need to specify both inputs ( $block_num | $printlen)";
        }

        if (!$buffer){
            # Pre-grow string
            my $tmp = "x" x $printlen;
            $buffer =  \$tmp;
        }

        if (!ref($buffer) || ref($buffer) ne 'SCALAR'){
            confess "Buffer argument must be a string reference, not a " . ref($buffer);
        }

        my $timestamp = time();
        my $block = _generate_block($$buffer, $offset, $file_path, $block_num,
            $printlen, $fbns_per_block[$$self], $timestamp);
        confess "No block returned for block $block_num in $file_path!" if (!$block);
        confess "Wrong size block $block_num in $file_path" if (length($block) != $printlen);
        return $block;
    }

    sub generate_block_perl {
        my $self = shift;
        my %args = @_;
        my $file_path = $self->get_path;
        my $block_num = $args{block_num};
        my $printlen = $args{size};
        my $buffer = $args{buffer}; # Reference to a string
        my $offset = $args{offset}; # Where to write in buffer
        if ($buffer){
            $offset = $args{offset}; # Where to write in buffer
            $offset = 0 if (!defined($offset));
        }
        if (!defined($block_num) || !defined($printlen)){
            confess "You need to specify all three inputs ($file_path | $block_num | $printlen)";
        }

        my $timestamp = sprintf('%.6f', time());
        my $block_str = "";
        my $header = "##$file_path##$block_num##$printlen##";
        my $left = $printlen;
        my $params = { size   => 4096,
                       header => $header,
                       fbn    => $self->get_first_fbn($block_num),
                       timestamp => $timestamp};
        while ($left >= 4096){
            $block_str .= _generate_wafl_block($params);
            $left -= 4096;
            $params->{fbn} ++;
        }
        if ($left){
            $params->{size} = $left;
            $block_str .= _generate_wafl_block($params);
        }
        if ($buffer){
            substr($$buffer, $offset, length($block_str), $block_str);
        }
        return $block_str;
    }

    sub _generate_wafl_block {
        my ($args) = @_;
        my $str = join('', $args->{header}, $args->{fbn}, '##', $args->{timestamp}, '##');
        my $strl = length($str);
        my $printlen = $args->{size};

        if ($strl >= $printlen) {
            #
            # Replace string with something shorter
            #
            # Start with a MD5 hash of str if possible
            if ($printlen > 22) {
                # only checksum the portion up to the timestamp as
                # that portion won't be reproducible during validation
                # see _generate_wafl_block() for -19 length calculation
                $str = md5_base64(substr($str, 0, -19));
            }
            elsif (1 == $printlen) {
                # lay down a binary pattern 10RRRRRR where R is a
                # random bit, since all other data falls into the
                # ASCII range, the MSB will only be 1 for a 1 byte
                # block.  the 6 bits of randomness will hopefully be
                # enough.
                $str = chr(0x80 | int(rand(0x40)));
            }
            # At least log the size of the block
            else{
                $str = chr($printlen + 33);
            }
            $strl = length($str);
        }

        #
        # Pad block with readable ASCII junk
        #
        my $filllen = $printlen - $strl;
        my $r = $filllen % 4;
        my $c = ($filllen - $r)/4;
        my $i = 0;
        if ($c){
            $str .= join('',
                        $alphabet[int(rand($alphabet_count))] x $c,
                        $alphabet[int(rand($alphabet_count))] x $c,
                        $alphabet[int(rand($alphabet_count))] x $c,
                        $alphabet[int(rand($alphabet_count))] x $c,
                        );
        }
        while ($i < $r){
            $str .= $alphabet[int(rand($alphabet_count))];
            $i ++;
        }

        if (length($str) != $printlen){
            confess "Error creating block. The following is not of size $printlen:\n$str";
        }
        return $str;
    }

    sub _last_bchain_block {
        my $chain = $_[0];
        confess "You must pass a chain" if (!$chain);
        return $chain->[0] + $chain->[1] - 1;
    }

   sub pretty_print_blockmap {
        my $self = $_[0];
        my $filepath = $self->get_path;
        my $blockmap_chains = $blockmap[$$self];
        if (!$blockmap_chains){
            $Log->comment("No used blocks chains for '$filepath'");
        }
        my $s = "Blockmap chains for '$filepath':\n";
        for my $chain (@$blockmap_chains){
            $s .= ("$chain->[0] + $chain->[1] \n");
        }
        $Log->comment($s);
    }
 }
 1;
__DATA__
__C__
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

const unsigned int MAX_BLOCK_SIZE = 524288;
const unsigned int WAFL_BLOCK_SIZE = 4096;
const char alphabet[] = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";
const unsigned int alphabet_len = 94;

unsigned int randint_in_range(unsigned int min, unsigned int max)
{
    return min + (rand() % (max - min));
}

char *_md5_base64( char* string ){
    Inline_Stack_Vars;
    char* args[] = { string, '\0'};
    int count = call_argv("Digest::MD5::md5_base64",G_SCALAR, args);
    char* md5;
    if ( count != 1 ){
        croak("Call to md5_base64 failed");
    }
    md5 = SvPV_nolen(Inline_Stack_Item( Inline_Stack_Items ));
    if ( strlen(md5) != 22 ){
        croak("MD5 string is of length %d instead of 22!\nChecksum: %s\nData: %s", strlen(md5),md5, string);
    }
    return md5;
}

void append_wafl_block( char *wafl_block, char *block_header, IV wafl_block_no, IV size, NV timestamp )
{
    unsigned int header_size_max = 1024;
    char header[header_size_max];
    unsigned int header_len;
    unsigned int fill_len;
    unsigned int r,c;
    char rand_char;
    unsigned int i,j;

    if ( size <= 0 ){
        croak( "Size (%ld) must be strictly greater than zero", size );
    }

    header_len = snprintf( header, header_size_max, "%s%lu##%.6f##",
        block_header, wafl_block_no, timestamp );

    if ( header_len >= header_size_max ){
        croak( "Wafl block header size (%u) larger than %d", header_len, header_size_max);
    }
    if ( header_len >= size ){
       /* Replace string with something shorter
        * Start with a MD5 hash of header if possible
        */
        if (size > 22){
            /* only checksum up to timestamp. see _generate_wafl_block() */
            header[header_len - 19] = '\0';
            strncpy(wafl_block, _md5_base64(header), 22);
            header_len = 22;
        }
        else if ( size == 1 ){
           /* lay down a binary pattern 10RRRRRR where R is a
            * random bit, since all other data falls into the
            * ASCII range, the MSB will only be 1 for a 1 byte
            * block.  the 6 bits of randomness will hopefully be
            * enough.
            */
            *wafl_block = 0x80 | randint_in_range(0, 0x40);
            header_len = 1;
        }
        else{
            /* Log the size of the bock */
            *wafl_block = size + 33;
            header_len = 1;
        }
    }
    else{
        strncpy(wafl_block, header, header_len);
    }
    wafl_block += header_len;

    /*
     * Pad block with readable ASCII junk
     */
    fill_len = size - header_len;
    r = fill_len % 4;
    c = (fill_len - r)/4;

    if (c > 0){
        for ( i=0; i<4; i++ ){
            rand_char = alphabet[randint_in_range( 0, alphabet_len - 1 )];
            for ( j=0; j<c; j++){
                *wafl_block = rand_char;
                wafl_block++;
            }
        }
    }
    for ( i=0; i<r; i++){
        rand_char = alphabet[randint_in_range( 0, alphabet_len - 1 )];
        *wafl_block = rand_char;
        wafl_block++;
    }
    *wafl_block = '\0';
}

SV *_generate_block(SV* buffer_sv, IV offset, char* file_path, IV block_no, IV size, IV fbns_per_block, NV timestamp)
{
    unsigned int header_size_max = 1024;
    char header[header_size_max];
    unsigned int header_len;
    char* block;
    char* wafl_block;
    IV wafl_block_no;
    int left;
    SV* return_block;
    char* buffer;

    if ( size <= 0 ){
        croak( "Size (%lld) must be strictly greater than zero", size );
    }

    if ( fbns_per_block < 1 ){
        croak( "Last argument (%lld) must be greater or equal to one", fbns_per_block );
    }

    if ( SvLEN(buffer_sv) < offset + size ){
        croak( "Buffer passed in is too small. %d instead of %d", SvLEN(buffer_sv), offset + size );
    }

    buffer = SvPVX( buffer_sv );

   /*
    * Generate header common to all wafl blocks in that block that looks like:
    *     ##file_path##block_no##size##
    */

    header_len = snprintf( header, header_size_max, "##%s##%lu##%lu##", file_path, block_no, size );
    if ( header_len >= header_size_max ){
        croak( "Block header size (%u) larger than %d", header_len, header_size_max);
    }

    wafl_block = buffer + offset;
    wafl_block_no = block_no * fbns_per_block;

    left = size;
    while ( left >= WAFL_BLOCK_SIZE ){
        append_wafl_block( wafl_block, header, wafl_block_no, WAFL_BLOCK_SIZE, timestamp );
        wafl_block_no ++;
        left -= WAFL_BLOCK_SIZE;
        wafl_block += WAFL_BLOCK_SIZE;
    }
    if ( left > 0 ){
        append_wafl_block( wafl_block, header, wafl_block_no, left, timestamp );
    }
    return_block = newSVpvn( buffer + offset, size );
    return return_block;
}
