# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Element::File::Sparse - Representation of a sparse file

=head1 DESCRIPTION

This class represent a file tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Element::File::Sparse;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::DataSet::Element::File;
    use Object::InsideOut qw( Perf::DataSet::Element::File);

    # Redefine checkums and blockmap here for convenience and speed
    my @checksums       :Field :Name('checksums');
    my @blockmap        :Field :Name('blockmap') :Deep;
    my @sparser_factory :Field :Arg(Name=>'sparser_factory', Mandatory=>1);
    my @sparser         :Field :Name('sparser');
    my @sparse_ratio    :Field
                        :Get(get_sparse_ratio)
                        :Arg(Name => 'sparse_ratio', Mandatory => 0, Default => 1)
                        :Type('numeric');
    my @sparse_level    :Field
                        :Get(get_sparse_level)
                        :Arg(Name => 'sparse_level', Mandatory => 0, Default => 0)
                        :Type('numeric');
    my @sparse_align    :Field
                        :Get(get_sparse_align)
                        :Arg(Name => 'sparse_align', Mandatory => 0, Default => 'right');

    sub _init :Init {
        my ($self) = @_;
        $self->disable_on_dirty;
        $self->compute_size_on_disk;
        $self->enable_on_dirty;
    }

    sub compute_size_on_disk {
        my ($self) = @_;
        $sparser[$$self] = $sparser_factory[$$self]->sparser(
                                                        filesize     => $self->get_size,
                                                        block_size   => $self->get_block_size,
                                                        sparse_ratio => $sparse_ratio[$$self],
                                                        sparse_level => $sparse_level[$$self],
                                                        sparse_align => $sparse_align[$$self],
                                                        );
        $self->set_size_on_disk($sparser[$$self]->get_size_on_disk);
    }

    sub create_checksum_store {
        my $self = shift;
        my %hash;
        keys %hash = $self->get_blocks_on_disk;
       $checksums[$$self] = \%hash;
    }

    sub create_blockmap {
        my $self = shift;
        my %hash;
        keys %hash = $self->get_blocks_on_disk;
        $blockmap[$$self] = \%hash;
    }

    sub get_sparse {
        return 1;
    }

    sub _replicate :Replicate
    {
        my ($self, $clone, $flag) = @_;
        # This will be called in the context of a shallow cloning
        confess "Deep cloning not supported" if ($flag eq 'deep');
        return if (!$checksums[$$self]);
        # We want a new hash, but not new values (since those are refernces
        # to checksums we want to keep shared as much as possible
        my %copy = %{$checksums[$$self]};
        $checksums[$$clone] = \%copy;
        return;
    }

    sub set_checksum {
        my ($self, $block_num, $checksum) = @_;
        confess "Bad Arguments" if (!defined($block_num) || !$checksum);
        if (!$checksums[$$self]){
            $self->create_checksum_store;
        }
        my $r;
        if (!exists $checksums[$$self]->{$block_num}){
            $r = $Perf::DataSet::BLOCK_ADDED;
        }
        else{
            $r = $Perf::DataSet::BLOCK_UPDATED;
        }
        $checksums[$$self]->{$block_num} = $checksum;
        return $r;
    }

    sub set_checksums :Restricted {
        my ($self, $c) = @_;
        confess "Missing argument" if (!$c);
        $checksums[$$self] = $c;
        return;
    }

    sub set_blockmaps :Restricted {
        my ($self, $b) = @_;
        confess "Missing argument" if (!$b);
        $blockmap[$$self] = $b;
    }

    sub get_checksum {
        my ($self, $block_num) = @_;
        my $c = $checksums[$$self]->{$block_num};
        return defined($c)?$$c:0;
    }

    sub set_blockmap {
        my ($self, $block_num) = @_;
        if ($block_num > $self->get_last_block || $block_num < $self->get_first_block){
            confess "block_num ($block_num) is out of range (". $self->get_first_block .
                    "-" . $self->get_last_block . ")";
        }
        if (!$blockmap[$$self]){
            $self->create_blockmap;
        }
        my $r;
        if (!exists $blockmap[$$self]->{$block_num}){
            $r = $Perf::DataSet::BLOCK_ADDED;
        }
        else{
            $r = $Perf::DataSet::BLOCK_UPDATED;
        }
        $blockmap[$$self]->{$block_num} = undef; #TODO: Is storing undef most efficient?
        return $r;
    }

    sub delete_checksums {
        my $self = $_[0];
        return delete $checksums[$$self];
    }

    sub delete_blockmap {
        my $self = $_[0];
        return delete $blockmap[$$self];
    }

    sub get_blockmap {
        my ($self, $block_num) = @_;
        if ($block_num > $self->get_last_block || $block_num < $self->get_first_block){
            confess "block_num ($block_num) is out of range (". $self->get_first_block .
                    "-" . $self->get_last_block . ")";
        }
        return exists $blockmap[$$self]->{$block_num};
    }

    sub has_block_data {
        my $self = shift;
        return 0 if (!$self->has_checksums && !$self->has_blockmap);
        return 1;
    }
    sub has_checksums {
        my $self = $_[0];
        return 0 if ($self->get_no_checksums);
        return !! keys %{$checksums[$$self]};
    }
    sub has_blockmap {
        my $self = $_[0];
        return 0 if (!$self->get_no_checksums);
        return !! keys %{$blockmap[$$self]};
    }

    sub is_coslice {
        my ($self, $other) = @_;
        my $a = $self->SUPER::is_coslice($other);
        return $a if (!$a);
        if ( $self->get_sparse_level != $other->get_sparse_level ){
            confess "Elements report different sparse_level values\n". $self->as_string . "\n"
                    . $other->as_string;
        }
        if ( $self->get_sparse_align ne $other->get_sparse_align ){
            confess "Elements report different sparse_align values\n". $self->as_string . "\n"
                    . $other->as_string;
        }
        if ( $self->get_sparse_ratio != $other->get_sparse_ratio ){
            confess "Elements report different sparse_ratio values\n". $self->as_string . "\n"
                    . $other->as_string;
        }
        return 1;
    }
}
1;
