# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

package Perf::DataSet::Event;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log $Parent config_result);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use NATE::Events qw(listener_add listener_remove timer_add timer_remove
                    eventloop_once eventloop_drain);
use NATE::Time qw(timeout2time describe_timeout);
use Time::HiRes qw(time);
use Perf::Utils qw(as_string);
use Params::Validate qw(validate SCALAR UNDEF BOOLEAN CODEREF);

sub new {
    my $class = shift;
    my %args = validate(    @_,
                            { type         => {type => SCALAR },
                              command_uid  => {type => SCALAR },
                              payload      => {type => SCALAR | UNDEF, optional=>1},
                              sent_time    => {type => SCALAR | UNDEF, optional=>1},
                              received_time=> {type => SCALAR | UNDEF, optional=>1},
                            },
                        );
    my $self = \%args;
    bless $self, $class;
    return $self;
}

sub get_command_uid {
    my $self = shift;
    return $self->{command_uid};
}

sub set_command_uid {
    my ($self, $uid) = @_;
    confess "You must specify a uid" if (!$uid);
    $self->{command_uid} = $uid;
}

sub get_dataset_uid {
    my $self = shift;
    return $self->{dataset_uid};
}

sub send {
    my $self = shift;
    my $process = shift;
    if(!$process || !$process->isa('Perf::Depot::Resource::Process')){
        confess "You must pass a Perf::Depot::Resource::Process object";
    }
    confess "Events must have a command_uid before being sent" if (!$self->{command_uid});
    my $str = "$self->{type} to $self->{command_uid}";
    if (defined($self->{payload})){
        $str .= " ($self->{payload})";
    }
    eval{
        $process->message_put(message=>'print "Received event ' . "$str [" . time . '] while idle or shutting down\n"',
                              type=>"Perf/DataSet/Event");
    };
    confess "Unable to send event of type $self->{type}: $@" if $@;
}

sub receive {
    my $class = shift;
    my %args = @_;
    my $target_command = $args{command_uid};
    my @events;
    while (my $message = $class->message_get( type=>"Perf/DataSet/Event", timeout=>0, on_timeout=>sub{})){
        my $event = $class->create_event_obj($message);
        if ($target_command && $event->get_command_uid != $target_command){
            $Log->comment("Throwing away event\n" . as_string($event));
        }
        else{
            push @events, $event;
        }
    }
    if (@events){
        return \@events;
    }
}

# TODO: Copy/paste/hack from NATE::Events to remove source filtering.
#       Remove once burt394961 has been fixed
sub message_get {
    my $class = shift;
    my %opts = validate(@_, {
        type           => { type => SCALAR,        optional => 1 },
        include_header => { type => BOOLEAN,       default  => 0 },
        eval_message   => { type => BOOLEAN,       default  => 0 },
        filter         => { type => CODEREF,       default  => sub {return 1;} },
        timeout        => { type => SCALAR|UNDEF,  optional => 1 },
        on_timeout     => { type => CODEREF|UNDEF, optional => 1 },
        callback       => { type => CODEREF|UNDEF, optional => 1 },
    });

    $opts{on_timeout} ||= sub {
        my @error = ("Timed out waiting for queued message.",
                     );
        if ($opts{timeout}) {
            push @error, "  Timeout: ".describe_timeout($opts{timeout})."\n";
        }
        die config_result([@error]);
    };

    my $caller_filter = $opts{filter};
    my $full_filter = sub {
        (!defined($opts{type}) || $_[0]->{type} eq $opts{type}) &&
        $caller_filter->(@_)
    };

    my $entry;
    my %keep_on_exit = (keep_on_exit => timeout2time(defined($opts{timeout}) ? $opts{timeout} : 60));
    my $listener = {filter       => $full_filter,
                    callback     => sub { my $listener;
                                          ($entry, $listener) = @_;
                                          timer_remove($listener->{timer}) if($listener->{timer});
                                          if(($opts{eval_message}) && (!ref($entry->{message}))) {
                                              # frozen message could have '$VAR<n> = ...' stuff
                                              no strict 'vars';
                                              $entry->{message} = eval($entry->{message})
                                                  or die config_result("Error decoding message: $@");
                                              use strict 'vars';
                                          }
                                          $entry = $entry->{message} if(!$opts{include_header});
                                          return $opts{callback}->($entry) if($opts{callback});
                                          return $entry;
                    },
                    once         => 1,
                    keep_on_fork => 0,
                    %keep_on_exit};

    # If there is a callback (for asynchronously dealing with messages),
    #   the only way to deal with the timeout is via a timer event
    my $timed_out = 0;
    my $timer;
    if(defined($opts{timeout})) {
        $timer =    {callback  => sub {
                        listener_remove($listener);
                        $timed_out++;
                        $opts{on_timeout}->(@_);
                    },
                    keep_on_fork => 0,
                    next_time => timeout2time($opts{timeout}),
                    %keep_on_exit};
        $listener->{timer} = $timer;
        timer_add($timer);
    }
    listener_add($listener);

    eventloop_drain(timeout => 0);

    #Async
    return 1 if($opts{callback});

    #Sync
    while (1) {
        return $entry if(defined($entry));
        return if ($timed_out);

        # Don't worry too much about the timeout because the timer event will
        #  trigger if it Really times out
        eval {eventloop_drain(timeout => $opts{timeout});};
        my $error = $@;
        if ($error) {
            timer_remove($timer) if($timer);
            listener_remove($listener);
            die $error;
        }
    }
}

sub create_event_obj {
    my $class = shift;
    my $message = shift;
    if ($message =~ /Received event (\w+) to (\d+) (?:\(\s*([^\)]*)\s*\))?\s*\[([^\]]+)\] /){
        my $type        = $1;
        my $command_uid = $2;
        my $payload     = $3;
        my $sent_time   = $4;
        my $event = $class->new(type         => $type,
                                command_uid  => $command_uid,
                                payload      => $payload,
                                sent_time    => $sent_time,
                                received_time=> time,
                                );
        return $event;
    }
    else{
        confess "Unable to parse message: $message";
    }
}

sub empty {
    my $class = shift;
    $Log->comment("Emptying Event queue");
    while (my $message = $class->message_get( type=>"Perf/DataSet/Event",
                                               timeout=>0,
                                               on_timeout=>sub{})){
        $Log->comment("Throwing away message: $message");
    }
    return;
}

1;
