# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  The Manager class is in charge of all protocol dependent operations
##           It is both a factory ( 'new' will instantiate the right subclass for the given protocol)
##           and a base class for all the various protocol specific subclasses
## @status   private
##
## @pod here

package Perf::DataSet::Manager;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log subtest $Parent);
use NATE::Events qw(eventloop_drain);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Perf::Utils qw(as_string);
use threads qw();
use MIME::Base64 qw(encode_base64);
use Storable qw(nfreeze);
$Storable::Deparse = 1;
use Net::Domain qw();
use Perf::Depot;


# this has to be loaded at compile time, otherwise multiple
# inheritance/MRO breaks
use Perf::DataSet::Manager::SANManager::iSCSIManager;


sub new {
    my $class = shift;
    my %args = @_;
    my $type =  delete $args{type};
    #
    # If 'type' is specified, this method behaves as a Factory
    #
    if ($type){
        if ($type eq 'nfs'){
            require Perf::DataSet::Manager::NFSManager;
            return Perf::DataSet::Manager::NFSManager->new(%args);
        }
        elsif ($type eq 'nfsv4'){
            require Perf::DataSet::Manager::NFSv4Manager;
            return Perf::DataSet::Manager::NFSv4Manager->new(%args);
        }
        elsif ($type eq 'nfsv4.1'){
            require Perf::DataSet::Manager::NFSv41Manager;
            return Perf::DataSet::Manager::NFSv41Manager->new(%args);
        }
        elsif($type eq 'iscsi'){
            # no require here, see above
            return Perf::DataSet::Manager::SANManager::iSCSIManager->new(%args);
        }
        elsif($type eq 'fcp'){
            require Perf::DataSet::Manager::SANManager::FCPManager;
            return Perf::DataSet::Manager::SANManager::FCPManager->new(%args);
        }
        elsif($type =~ /^cifs/) {
            require Perf::DataSet::Manager::CIFSManager;
            return Perf::DataSet::Manager::CIFSManager->new(%args);
        }
        else{
            confess "Type ($type) not supported";
        }
    }
    #
    # If 'type' is not specified, perform common initialization stuff
    #
    my $self = {
                process_count      => undef,
                localhost          => undef,
                has_registered     => 0,
                is_shutting_down   => 0,
               };
    $self->{pid} = $$;
    $self->{pwd} = &Cwd::cwd();
    $self->{tid} = threads->tid;
    bless $self, $class;
    return $self;
}

sub DESTROY {
    my $self = shift;
    #
    # Skip the destruction for subtests and threads
    #
    return unless ($$ == $self->{pid});
    return unless (threads->tid == $self->{tid});
    return unless (!$self->{is_shutting_down});

    $Log->comment("DESTROY of Manager started in process $$");

    my $current_error = $@;
    eval{
        $self->shutdown;
    };
    if ($@){
        $Log->comment("Got errors during DESTROY of Manager:\n$@");
    }
    $@ = $current_error;
    return;
}

sub _get_uid {
    return int(rand(2**32));
}

sub stop_processes {
    my $self = shift;
    my $proc_arrayref = shift;
    my @processes;
    if ($proc_arrayref){
        if (ref($proc_arrayref) eq 'ARRAY'){
            @processes = @$proc_arrayref;
        }
        else{
            confess "Argument must be an arrayref";
        }
    }
    else{
        @processes = keys %{$self->{processes}};
    }

    my $stopped = 0;
    for my $process_uid (@processes){
        eval{
            my $process = $self->{processes}->{$process_uid}->{process};
            if ($process && $process->is_alive()){
                my $ipc = $process->ipc();
                $ipc->quit();
                $process->wait_finish();
                $stopped ++;
            }
            else{
                $Log->comment("Process '$process_uid' was already dead");
            }
        };
        if ($@){
            $Log->comment("Errors during shutdown of process '$process_uid':$@");
        }
    }
    $Log->comment("Shut down $stopped processes");
}

sub shutdown {
    my $self = shift;
    return if ($self->is_shutting_down);
    $self->{is_shutting_down} = 1;
    $Log->comment("Manager is shutting down");

    # Shutting down means finding all the Perf::DataSet talking to this
    # manager and unregistering them
    my $all_datasets = Perf::Depot->get_resources( class=>'DataSet');
    return if (!$all_datasets || !@$all_datasets);
    my @datasets = grep { $_->get_manager == $self } @$all_datasets;

    if (@datasets){
        for my $dataset (@datasets){
            # Since we are shutting down, disregard errors
            # happening during unregister and keep going
            my $dataset_uid = $dataset->get_name;
            eval{
                $self->unregister($dataset_uid);
            };
            if ($@){
                $Log->comment("Error during unregistration of Perf::DataSet '$dataset_uid':$@");
            }
        }
    }
    undef @datasets;
    return;
}

sub register_dataset {
    my $self = shift;
    my %args =
      Params::Validate::validate(@_,
                                 {dataset =>
                                    {type => OBJECT, isa => 'Perf::DataSet'},});

    my $dataset = $args{dataset};
    my $dataset_uid = $dataset->get_uid;

    # First item registered must be the root Perf::DataSet
    if (!$self->{has_registered}){
        if($dataset->get_parent_uid){
            confess "You must start by registering a root Perf::DataSet";
        }
        $self->{process_count_per_client} = $dataset->get_processes;
        my $dataset_clients = $dataset->get_clients;

        # Add localhost
        my $localhost = Net::Domain::hostname(); #Returns short name that host() likes better
        my $fully_qualified_localhost;
        my $short_name_localhost;
        if ($localhost =~/^([^.]+)\./){
            $fully_qualified_localhost = $localhost;
            $short_name_localhost = $1;
        }
        elsif($localhost =~/^([^.]+)$/){
            $short_name_localhost = $1;
        }
        else{
            confess "Can't parse hostname: $localhost";
        }
        # See if the short name matches any of the clients
        my $match = grep {/^$short_name_localhost$/} @$dataset_clients;
        if ($match){
            $localhost = $short_name_localhost;
        }
        # See if fuly qualified name matches
        elsif ($fully_qualified_localhost){
            $match = grep {/^$fully_qualified_localhost$/} @$dataset_clients;
            if ($match){
                $localhost = $fully_qualified_localhost;
            }
        }
        $self->{localhost} = $localhost;
        $Log->debug("Found localhost to be '$localhost'");
        my $localhost_obj = Perf::Depot->get_resource( class=>'Host',
                                                          name=>$self->{localhost});
        if (!$localhost_obj){
            $localhost_obj = Perf::Depot->create_resource(
                class => 'Localhost',
                enable_alua => $dataset->enable_alua(),
                enable_multipathing => $dataset->enable_multipathing(),
                name => $self->{localhost},
            );

            $localhost_obj->register;
        }
        $self->{has_registered} = 1;
    }
    my $dataset_obj = Perf::Depot->get_resource( class=> 'DataSet',  name=>$dataset_uid);
    confess "Perf::DataSet already exists?!!" if ($dataset_obj);
    $dataset_obj = Perf::Depot->create_resource( class=> 'DataSet',  name=>$dataset_uid,
                                                                        dataset=>$dataset);
    $dataset_obj->register;
    return $dataset_obj;
}

sub get_localhost {
    my $self = shift;
    return $self->{localhost};
}

sub get_processes($$) {
    my $self = shift;
    my $dataset_uid = shift;
    confess "You must pass an argument" if (!$dataset_uid);
    my $dataset = Perf::Depot->get_resource( class=>'DataSet', name=>$dataset_uid );
    confess "Invalid dataset_uid $dataset_uid" if (!$dataset);
    my $processes = Perf::Depot->get_resources( class=>'Process', used_by=>$dataset);
    #my @dataset_processes = grep { $_->is_used_by($dataset) } @$processes;
    #return \@dataset_processes;
    my $expected_count = scalar(@{$dataset->get_clients}) * $self->{process_count_per_client};
    my $processes_count = scalar(@$processes);
    if ( $processes_count != $expected_count){
        confess "Unexpected number of processes returned! $processes_count instead of $expected_count";
    }
    return $processes;
}

sub get_process_count_per_client {
    my $self = shift;
    return $self->{process_count_per_client};
}

sub get_pwd {
    my $self = shift;
    return $self->{pwd};
}
sub get_clients {
    my ($self,$dataset_uid) = @_;
    my $dataset = Perf::Depot->get_resource( class=>'DataSet', name=>$dataset_uid);
    confess "No dataset found with uid $dataset_uid" if (!$dataset);
    return $dataset->get_dependencies( class=>'Host');
}

sub get_parent_uid {
    my $self = shift;
    my $dataset_uid = shift;
    my $dataset = Perf::Depot->get_resource( class=>'DataSet', name=>$dataset_uid);
    confess "No dataset found with uid $dataset_uid" if (!$dataset);
    return $dataset->get_parent_uid;
}

sub get_relative_path {
    my $self = shift;
    my $dataset_uid = shift;
    if (!$dataset_uid){
        confess "You must enter a dataset_uid";
    }
    my $dataset = Perf::Depot->get_resource( class=>'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown dataset '$dataset_uid'";
    }
    my %relative_path_info = $dataset->get_relative_path_info;
    my @sorted_info = ('snapshot', 'qtree', 'lun_prefix', 'root_dir');

    my $relative_path = '';
    if ($relative_path_info{snapshot}){
        $relative_path .= '.snapshot/';
    }
    $relative_path .= join('/', grep { defined $_ } map { $relative_path_info{$_} } @sorted_info);
    return $relative_path;
}

sub get_worker_full_paths {
    confess "This method must be implemented in a subclass";
}

#
# Delete data residing on filer for specific Perf::DataSet
#
sub purge {
    confess "This method must be implemented in a subclass";
}

#
# Get rid of all information related to a specific Perf::DataSet
#
sub unregister {
    my ($self,$dataset_uid) = @_;
    confess "You must pass an argument" if (!$dataset_uid);
    my $dataset = Perf::Depot->get_resource( class=>'DataSet', name=>$dataset_uid);
    if (!$dataset ){
        if (!$self->{unregistered}->{$dataset_uid}){
            confess "Unknown Perf::DataSet $dataset_uid";
        }
        $Log->comment("Perf::DataSet '$dataset_uid' was unregistered already");
        return;
    }

    $Log->comment("Unregistering Perf::DataSet '$dataset_uid'");

    #
    # Check that this particular Perf::DataSet isn't currently executing code
    #
    my $active_command = $self->get_active_command($dataset_uid);
    if ($active_command){
        $active_command->abort;
    }

    if (!$dataset->get_no_purge_on_exit && !$dataset->get_snapshot){
        #
        # Figure out if data associated with this Perf::DataSet can be deleted
        # from the appliance
        # Look for datasets pointing to the same location on appliance
        # If none is found and the location is not a snapshot
        # attempt to delete the data. Deletion may fail if data is in
        # a read-only volume
        my $appliance = $dataset->get_appliance;
        my $volume = $dataset->get_volume;
        my $relative_path = $dataset->get_relative_path;
        #my $datasets_colocated = $self->get_datasets(filer        =>$filer,
        #                                             volume       =>$volume,
        #                                             relative_path=>$relative_path,
        #                                             );
        my @datasets_colocated = grep { $_ != $dataset }
          @{Perf::Depot->get_resources(class         => 'DataSet',
                                                appliance     => $appliance,
                                                volume        => $volume,
                                                relative_path => $relative_path,)};
        my $datasets_colocated_count = scalar(@datasets_colocated);
        if (0 < $datasets_colocated_count){
            $Log->comment("No purging since $datasets_colocated_count Perf::DataSet(s) are "
                       ."still pointing to the data");
        }
        else {
            eval {
                $dataset->purge(force => 1, users_ok => 1);
            };
            if ($@) {
                $Log->comment("Unable to purge Perf::DataSet '$dataset_uid': $@");
            }
        }
    }

    # Remove from Depot. That combined with the following undef
    # should trigger a chain reaction where Resources clean up
    # after themselves as they go out
    Perf::Depot->remove_resource($dataset);
    undef $dataset;

    # Perform some house cleaning if the Depot is void of Perf::DataSets
    my $remaining_datasets = Perf::Depot->get_resources( class=>'DataSet');
    if (!$remaining_datasets || !@$remaining_datasets){
        Perf::Depot->remove_unused_and_never_used;
    }

    $self->{unregistered}->{$dataset_uid} = 1;
    return;
}

sub unmount {
    confess "This method must be implemented in subclass";
}

sub mount {
    confess "This method must be implemented in subclass";
}

sub is_mounted {
    my $self = shift;
    my $dataset_uid = shift;
    if (!$dataset_uid){
        confess "You must enter a dataset_uid";
    }
    my $dataset = $self->{datasets}->{$dataset_uid};
    if (!$dataset){
        confess "Unknown dataset '$dataset_uid'";
    }
    return $dataset->is_mounted;
}


sub get_mountpoints {
    my ($self,$dataset_uid) = @_;
    if (!$dataset_uid){
        confess "You must specify a dataset_uid";
    }
    my $dataset = Perf::Depot->get_resource(class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown dataset '$dataset_uid'";
    }
    my %args = @_;
    my $type = $args{type};
    if (!$type){
        $type = 'worker';
    }

    if($type =~ /worker/i){
        my $workers = Perf::Depot->get_resources( class=>'Worker', name=>$dataset_uid);
        my @mountpoints = map {$_->get_dependency( class=>'Mountpoint')} @$workers;
        return \@mountpoints;
    }
    else{
        confess "Unknown type '$type'";
    }
}

sub initialize {
    confess "This method must be implemented in subclass";
}

sub create_workers {
    my ($self, $dataset_uid) = @_;
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    confess "No Perf::DataSet with uid $dataset_uid found!" if (!$dataset);
    my $process_group = Perf::Depot->get_resource( class=>'Process::Group', used_by=>$dataset );
    confess "No process group found for " . $dataset->get_description if (!$process_group);
    my $processes = $process_group->get_processes;
    my $workers = [];
    for my $process_obj (@$processes){
        my $worker_obj = Perf::Depot->create_resource(   class  => 'Worker',
                                                            process=> $process_obj,
                                                            name   => $dataset_uid,
                                                        );
        $dataset->add_dependencies( $worker_obj );

        # The process group will clean up all processes that it contains
        # including the process for this worker. It's therefore imperative that it no get
        # cleaned up before this worker. This means the group is a dependency of the worker.
        # TODO: It would probably feel less wrong if we introduced a Worker::Group. The
        # Worker::Group could then naturally depend on the Process::Group and the cleanup
        # would happen in the right order without having each worker depend on the Process::Group
        $worker_obj->add_dependencies( $process_group );

        push @$workers, $worker_obj;
    }
    for my $worker (@$workers){
        $worker->wait_new;
    }
    return $workers;
}

sub create_processes {
    my ($self,$dataset_uid) = @_;
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    confess "No Perf::DataSet with uid $dataset_uid found!" if (!$dataset);
    my $hosts = $dataset->get_dependencies( class => 'Host');
    confess "No host resources found for " . $dataset->get_description if (!$hosts || !@$hosts);
    my $process_count_per_client = $self->get_process_count_per_client;
    my $process_group = Perf::Depot->create_resource( class => 'Process::Group',
                                                         name  => $dataset_uid );
    $process_group->register;
    for my $host (@$hosts){
        $Log->comment("Starting $process_count_per_client processes on "  . $host->get_description);
        my $factory = Perf::Depot->get_resource( class=> 'Process::Factory',
                                                    host => $host);
        if (!$factory){
            $factory = Perf::Depot->create_resource( class => 'Process::Factory',
                                                        name  => 'Dummy',
                                                        host  => $host);
            $factory->register;
        }
        my $processes = $factory->create_processes( count=>$process_count_per_client );
        $process_group->add_processes( factory      => $factory,
                                        processes   => $processes,
                                     );
    }
    $dataset->add_dependencies( $process_group );
    return $process_group;
}

#
# Execute code in all processes
#
sub execute {
    my $self = shift;
    my %params = Params::Validate::validate(
           @_,
           {
                command => {type => OBJECT, isa => 'Perf::DataSet::Command', optional => 1},
                code => {type => SCALAR , optional => 1},
                args => {type => ARRAYREF , optional => 1},
                dataset => {type => SCALAR },
           }
           );
    my $command = $params{command};
    my $code = $params{code};
    my $args = $params{args};
    if ( (!$command && !$code) || ($command && $code) ){
        confess "You must specify either a command or some code";
    }
    my $dataset_uid = $params{dataset};
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown Perf::DataSet '$dataset_uid'";
    }

    # Make sure all the processes are alive
    $self->_check_processes($dataset_uid);

    if ($command){
        $code = '$::p->{' . $dataset_uid . '}->execute';
        my $frozen_command = nfreeze($command);
        my $encoded_frozen_command = encode_base64($frozen_command, '');
        $args = [ $encoded_frozen_command ];
    }
    $Log->debug("Executing code: $code");

    #
    # Send the code
    #
    my $workers = $dataset->get_dependencies( class=>'Worker');
    confess "No Workers found?!" if (!$workers || !@$workers);

    $self->{active}->{$dataset_uid}->{code} = $code;
    $self->{active}->{$dataset_uid}->{args} = $args;
    $self->{active}->{$dataset_uid}->{command} = $command if ($command);

    my $monitor_id = "monitor_${dataset_uid}_" . time;
    $Log->debug("Starting monitor '$monitor_id' on dataset '$dataset_uid' for:\n"
                . "code: $self->{active}->{$dataset_uid}->{code}\n"
                . "args: $self->{active}->{$dataset_uid}->{args}\n");
    my $subtest;
    eval{
        $subtest = subtest( sub{
                                for my $worker (@$workers){
                                    $worker->execute( code=>$code, args=>$args);
                                }
                                my $panic_on_corruption = $command->{panic_on_corruption};
                                $self->_monitor($dataset_uid, panic_on_corruption=>$panic_on_corruption);
                              }
                            , "-bg",
                            '-runid',$monitor_id,
                            );
        # Calling 'message_get' on a dead subtest doesn't reliably work. Instead use a callback.
        $subtest->message_get(  type    =>'Perf/DataSet/Monitor',
                                callback=> sub { my $result = shift;
                                                $Log->comment("Monitor '$monitor_id' returned a message");
                                                $Log->debug("Monitor '$monitor_id' result:\n" . as_string($result));
                                                if ( ! defined $result ){
                                                    $result = 'UNDEFINED!';
                                                }
                                                $self->{active}->{$dataset_uid}->{result} = $result },
                              );
    };
    my $error = $@;

    # TODO: HACK Get rid of Workers trapped in the monitor closure
    # Remove when issue gets fixed
    if ($subtest){
        delete $subtest->{scriptspec};
        delete $subtest->{full_scriptspec};
    }

    die $error if $error;

    $self->{active}->{$dataset_uid}->{monitor} = $subtest;
    return;
}

sub _monitor {
    my ($self,$dataset_uid,%args) = @_;
    confess "You must pass an argument" if (!$dataset_uid);
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    confess "No Perf::DataSet found with name $dataset_uid" if (!$dataset);
    my $panic_on_corruption = $args{panic_on_corruption};
    # For panics on corruption to be produce interesting data
    # the filer must be panic'ed quickly, hence the shorter sleep time
    #my $sleep_time = $panic_on_corruption?1:5;
    my $sleep_time = 1;

    my $ret = {errors=>undef,
               returns=>undef,
               };

    # This subroutine needs to always report something
    local $@;
    eval{

        # This subroutine is only meant to run in a subtest with IPC
        if ($$ == $self->{pid}){
            confess "Subroutine not meant to be run in main process";
        }

        my $active = $self->{active}->{$dataset_uid};
        my $code = $active->{code};
        my $args = $active->{args};
        my $command = $active->{command};

        # Make sure all processes are still alive
        $self->_check_processes($dataset_uid);

        #
        # Capture results
        #
        my @processes_busy = @{ $dataset->get_used_processes };
        $self->{active}->{$dataset_uid}->{processes} = \@processes_busy;

        my $loop = 0;
        my $errors;
        my $returns;
        while(1){
            my $processes_completed = 0;
            my @processes_idx_done;
            for my $i (0..$#processes_busy){
                my $process = $processes_busy[$i];
                if ($process->has_reply){
                    my $process_uid = $process->get_name;
                    # See if we got a result or an error
                    my $ret;
                    eval { $ret = $process->read_reply };
                    if ($@){
                        $errors->{$process_uid} = $@;
                        $Log->comment( "Error detected in process $process_uid: \n$@" );
                        if ($panic_on_corruption && $@ =~ /Data corruption/i){
                            $Log->comment("Panicing filer after data corruption detection");
                            my $appliance = $dataset->get_appliance_resource;
                            $appliance->panic;
                        }
                        # An error was detected, abort all the remaining workers
                        $command->abort(check_active=>1);
                    }
                    else{
                        $Log->comment("Process $process_uid has completed successfully" );
                        $returns->{$process_uid} = $ret;
                        # Tell the remaining workers this worker has stopped
                        #$command->notify_worker_completion(check_active=>1);
                        #$processes_completed ++;
                    }
                    $Log->comment("Adding index $i");
                    push @processes_idx_done, $i;
                }
            }
            if (@processes_idx_done){
                for my $process_done_ix (sort {$b <=> $a} @processes_idx_done){
                    splice @processes_busy, $process_done_ix, 1;
                }
                last if (!scalar @processes_busy);
                $Log->comment("Remaining busy processes: " . join(" ", map {$_->get_name} @processes_busy) );
                # Tell the remaining workers how many have stopped
                #$command->notify_worker_completion(payload=>$processes_completed, check_active=>1);
                $command->notify_worker_completion(scalar @processes_idx_done);
            }
            # Should block until something is seen
            $Log->comment("Starting eventloop_drain");
            eventloop_drain();
            $Log->comment("Done with eventloop_drain");
        }

        if ($errors){
            my $str = "";
            for my $process_uid (keys %$errors){
                $str .= "Error on '$process_uid':\n$errors->{$process_uid}\n";
            }
            if ($args){
                confess("Got errors during during execution of '$code(@$args)':\n$str");
            }
            else{
                confess("Got errors during during execution of '$code':\n$str");
            }
        }

        $ret->{returns} = $returns;

    };#eval
    if ($@){
        $ret->{errors} = $@;
    }
    $Log->comment("Writing result object back to parent:\n" . as_string($ret));
    $Parent->message_put(message=>$ret, type=>'Perf/DataSet/Monitor');
    $Log->comment("Subroutine _monitor returning");
    return;
}

#
# Wait for processes to finish executing currently active code
# Allow to skip the wait_finish step
#
sub wait {
    my ($self,$dataset_uid,%args) = @_;
    confess "You must pass an argument" if (!$dataset_uid);
    if (!$self->{active}->{$dataset_uid}){
        $Log->debug("No active code for Perf::DataSet '$dataset_uid'");
        return;
    }
    my $monitor = $self->{active}->{$dataset_uid}->{monitor};
    if (!$monitor){
        confess "Unable to find monitoring processes for workload on '$dataset_uid'";
    }
    my $wait_finish = $args{wait_finish};
    $wait_finish = 1 if (!defined($wait_finish));

    my $monitor_id = $monitor->{runid};
    my $waiting_message = "Waiting for monitor '$monitor_id' to return";
    $Log->comment($waiting_message);
    my $ret = delete $self->{active}->{$dataset_uid}->{result};
    while (!defined $ret){
        $Log->comment($waiting_message);
        # The following seems to block for 5s unless a message is seen
        eventloop_drain();
        $ret = delete $self->{active}->{$dataset_uid}->{result};
    }
    if ($ret eq 'UNDEFINED!'){
        $Log->comment("Monitor '$monitor_id' returned an undefined result");
        $self->wait_finish($dataset_uid);
        confess "Monitor '$monitor_id' returned an undefined result";
    }
    elsif ($ret->{errors}){
        # Do we care about the returns?
        $Log->comment("Errors returned by monitor '$monitor_id':\n" . as_string($ret));
        $self->wait_finish($dataset_uid);
        die "$ret->{errors}";
    }
    else{
        $self->wait_finish($dataset_uid) if ($wait_finish);
        return $ret->{returns};
    }
}

sub get_active_command {
    my ($self,$dataset_uid) = @_;
    confess "You must pass a dataset_uid" if (!$dataset_uid);
    my $dataset = Perf::Depot->get_resource(class => 'DataSet', name=>$dataset_uid);
    confess "Invalid Dataset '$dataset_uid'" if (!$dataset);
    if (!$self->{active} || !$self->{active}->{$dataset_uid}){
        $Log->debug("No active command found for dataset '$dataset_uid'");
        return;
    }
    my $command = $self->{active}->{$dataset_uid}->{command};
    confess "No command found even though dataset is active!" if (!$command);
    return $command;
}

sub wait_finish {
    my ($self, $dataset_uid) = @_;
    my $monitor = $self->{active}->{$dataset_uid}->{monitor};
    if (!$monitor){
        confess "Unable to find monitoring processes for workload on '$dataset_uid'";
    }
    $Log->comment("Waiting for subtest of monitor '$monitor->{runid}' to finish");
    $monitor->wait_finish;
    delete $self->{active}->{$dataset_uid};
    return;
}

sub send_event {
    my ($self,$dataset_uid,$event,%params) = @_;
    if (!$event || ! $event->isa('Perf::DataSet::Event')){
        confess "You must specify a Perf::DataSet::Event";
    }
    my $dataset = Perf::Depot->get_resource(class => 'DataSet', name=>$dataset_uid);
    confess "Invalid Dataset '$dataset_uid'" if (!$dataset);
    my $command = $self->get_active_command($dataset_uid);
    confess "No active command found for dataset '$dataset_uid'" if (!$command);
    my $command_uid = $event->get_command_uid;
    confess "Active command uid doesn't match!" if ($command_uid != $command->get_uid);

    my $check_active = $params{check_active};
    my $target_processes;
    if ($check_active){
        $target_processes = $self->{active}->{$dataset_uid}->{processes};
        my $str = join(' ', map { $_->get_name } @$target_processes);
        $Log->comment("Current active processes:\n$str");
    }
    else{
        $target_processes = $dataset->get_used_processes;
    }
    if (!$target_processes || ref($target_processes) ne 'ARRAY' || !@$target_processes){
        confess "No target processes ?!\n" . as_string($self->{active});
    }
    for my $process (@$target_processes){
        my $process_uid = $process->get_name;
        $Log->comment("Sending event $event->{type} to process '$process_uid'");
        $event->send($process);
    }
}

sub is_shutting_down {
    my $self = shift;
    return $self->{is_shutting_down};
}

#
# Check if at least one process on one client is still executing
# code. Depends on clients not writing anything to their RPC
# socket until they're done.
#
sub is_busy {
    my ($self,$dataset_uid) = @_;
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Invalid dataset '$dataset_uid";
    }

    if (!$self->{active}->{$dataset_uid}){
        return 0;
    }
    my $monitor = $self->{active}->{$dataset_uid}->{monitor};
    if (!$monitor){
        confess "No monitoring process found for $dataset_uid!";
    }

    if (!$monitor->is_alive){
        return 0;
    }

    # The monitor process is supposed to send back the results
    # If this hasn't happened yet, it means the command is still going
    my $ret = $self->{active}->{$dataset_uid}->{result};
    return !defined($ret);
 }

sub _check_processes {
    my ($self,$dataset_uid) = @_;
    confess "You must pass an argument" if (!$dataset_uid);
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Invalid dataset '$dataset_uid";
    }
    my $errors = "";
    my $processes = $dataset->get_used_processes;
    for my $process (@$processes){
        if ($process && $process->is_alive()){
            next;
        }
        else{
            my $str = $process->get_description . " is dead\n";
            $errors .= $str;
        }
    }
    if ($errors){
        confess $errors;
    }
    return;
}

sub freeze_workers {
    my ($self, $dataset_uid, %args) = @_;
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    confess "No Perf::DataSet with uid $dataset_uid found!" if (!$dataset);
    my $path = $args{path};
    my $workers = $dataset->get_worker_resources;
    for my $w (@$workers){
        $w->freeze( path => $path );
    }
}

sub thaw_workers {
    my ($self, $dataset_uid, %args) = @_;
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    confess "No Perf::DataSet with uid $dataset_uid found!" if (!$dataset);
    my $path = $args{path};
    confess "Invalid path '$path'" if (!$path || ! -d $path);
    my $no_unlink = $args{no_unlink};

    #
    # Container files are separated in folders, one for each client
    #
    opendir(my $dir, $path) or confess "Unable to open folder '$path': $!";
    my @clients = map { "$path/$_" } grep { ! /^\./ } readdir($dir);
    closedir $dir;
    confess "No folders found under '$path'" if (!@clients);
    $Log->comment("Found folders:\n" . join("\n", @clients));

    my @files;
    my $count = 0;
    for my $client (@clients){
        opendir($dir, $client) or confess "Unable to open folder '$client': $!";
        my @client_files = map { "$client/$_" } grep { ! /^\./ } readdir($dir);
        closedir $dir;
        $Log->comment("Found files:\n" . join("\n", @client_files));
        push @files, \@client_files;
        $count += scalar @client_files;
    }

    my $workers = $dataset->get_worker_resources;
    confess "Wrong number of files ( " . $count . " instead of " . scalar(@$workers) ." )"
        if ($count != scalar @$workers );

    #
    # Assign files from a client to workers residing on a same client
    # This is so as to not split slices that were created with 'single_client_file_slices'
    my $worker_clients;
    for my $w (@$workers){
        my $w_hostname = $w->get_host->get_name;
        push @{$worker_clients->{$w_hostname}}, $w;
    }
    my @worker_client_names = keys %$worker_clients;
    # Assign container files in groups
    for my $files_group (@files){
        my $suitable_group;
        my $i = 0;
        while (!$suitable_group){
            my $worker_group = $worker_clients->{ $worker_client_names[$i] };
            # We need to find a group of workers that is just as big or bigger than the number
            # of container files we're trying to assign
            $suitable_group = $worker_group if ( scalar @$worker_group >= scalar @$files_group );
            # Exit loop if there are no more groups to examine
            last if ($i == $#worker_client_names);
            $i ++;
        }
        confess "Unable to find suitable group!" if (!$suitable_group);
        for my $file (@$files_group){
            my $worker = shift @$suitable_group;
            $worker->thaw( $file, no_unlink=>$no_unlink );
        }
    }

    # TODO: The above isn't enough for SAN Perf::DataSets.
    # SANManager should override this method and complete it
    if ($dataset->is_san){
        confess "Operation not supported yet";
    }

    return;
}

1;
