# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Manager subclass for CIFS on Unix protocol
## @status   private
##
## @pod here

package Perf::DataSet::Manager::CIFSManager;

use strict;
use warnings;
use Carp qw(confess);
use List::Util qw(max);
use Readonly;
use Tharn qw($Log);
use NATE::ParamSet;


BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}

use namespace::clean;
use Perf::DataSet::Manager::NASManager;
use base qw(Perf::DataSet::Manager::NASManager);


Readonly::Array our @SMB_VERSIONS => qw(1 2 2.1 3);


sub volume_mount_setup {
    my ($self,$dataset_uid,%args) = @_;
    $self->SUPER::volume_mount_setup($dataset_uid, %args);
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown Perf::DataSet $dataset_uid";
    }
    my $vol = $dataset->get_volume_resource(class => 'Volume');

    if (!$vol){
        confess "No Volume found for " . $dataset->get_description;
    }
    my $cifs_share = Perf::Depot->get_resource( class=> 'CIFSShare', volume=>$vol );
    if (!$cifs_share) {
        $cifs_share =
          Perf::Depot->create_resource(class  => 'CIFSShare',
                                                volume => $vol,
                                                name   => 'Dummy',
                                                disable_oplocks => $dataset->disable_oplocks());
        $cifs_share->register;
    }
    else {
        confess('oplocks setting mismatch')
            if($dataset->disable_oplocks() != $cifs_share->disabled_oplocks());
    }

    return;
}

sub get_mountpoint_class {
    return 'Mountpoint::CIFS';
}

sub get_mount_opts {
    my ($self,$dataset_uid,%args) = @_;
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown Perf::DataSet $dataset_uid";
    }
    my $ro_volume = $args{read_only};
    my $params = NATE::ParamSet->get_global();

    my %mount_opts;
    my $CIFS_DOMAIN = $params->get('CIFS_DOMAIN');
    my $AD_SERVER_USER = $params->get('AD_SERVER_USER');
    my $AD_SERVER_PASSWORD = $params->get('AD_SERVER_PASSWORD');

    confess "Required param CIFS_DOMAIN not defined" unless defined ($CIFS_DOMAIN);
    confess "Required param AD_SERVER_USER not defined"
        unless defined ($AD_SERVER_USER);
    confess('Required param AD_SERVER_PASSWORD not defined')
        if (!defined ($AD_SERVER_USER));

    $mount_opts{fstype} = 'cifs';
    $mount_opts{user} = "$CIFS_DOMAIN\\$AD_SERVER_USER";
    $mount_opts{password} = $AD_SERVER_PASSWORD;
    $mount_opts{mountopts} = "rw,directio"; # XXX BLS still needed?
    return %mount_opts;
}


sub register_dataset {
    my $self      = shift;
    my $dataset   = $self->SUPER::register_dataset(@_);
    my $type      = $dataset->get_type();
    my $host_objs = $dataset->get_client_resources();
    my $appliance = $dataset->get_appliance_resource();
    my $max_supported_smb_version = max(@SMB_VERSIONS);
    my $requested_smb_version;
    my %host_info;

    foreach my $host (@$host_objs) {
        my $max_smb_version = $host->max_smb_version();

        $host_info{$host->get_name()} = {obj             => $host,
                                         max_smb_version => $max_smb_version,};

        # client doesn't support our current max, degrade
        $max_supported_smb_version = $max_smb_version
            if ($max_supported_smb_version gt $max_smb_version);
    }

    if ('cifs' eq $type) {
        # force a specific SMB version
        $type = "cifs-smb$max_supported_smb_version";
        $requested_smb_version = $max_supported_smb_version;
        $dataset->_set_type($type);
    }
    elsif ($type =~ /^cifs-smb([[:digit:].]+)$/) {
        $requested_smb_version = $1;

        confess(  "Requested SMB$requested_smb_version but maximum version"
                . " supported by clients is: SMB$max_supported_smb_version")
          if ($requested_smb_version > $max_supported_smb_version);

        confess(  'SMB2 requested but client will negotiate to SMB2.1.'
                . ' Need a Windows Vista or non-R2 2008 Server client')
          if (2 == $requested_smb_version && 2.1 <= $max_supported_smb_version);
    }

    my %used_version =
      map { $_->get_smb_version() => 1 }
      grep { $dataset != $_ }
      @{Perf::Depot->get_resources(class              => 'DataSet',
                                            appliance_resource => $appliance,)};

    # disable all versions from one above what was requested up to
    # the maximum possible
    foreach my $version (grep { $_ > $requested_smb_version } @SMB_VERSIONS) {
        confess("SMB$version in use by another DataSet and cannot be disabled")
            if ($used_version{$version});

		# SMB2.1 cannot be independently controlled, see comments about 2.0 clients above
        next
            if (2.1 == $version);

        $appliance->cifs_disable_smb_feature($version);
    }

    my %cur_version = map { $_ => 1 } @{$appliance->cifs_get_smb_features()};

    $requested_smb_version = 2
        if (2.1 == $requested_smb_version);

    $appliance->cifs_enable_smb_feature($requested_smb_version)
        if (!$cur_version{$requested_smb_version});
}


1;
