# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

package Perf::DataSet::Manager::IPManager;

use strict;
use warnings;
use Carp qw(confess longmess);
use Tharn qw($Log host connect say subtest);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use List::MoreUtils qw(all none);
use Perf::Utils qw(min max get_random_element as_string in_bytes deep_copy ndmp_copy);
use Perf::Depot;

use base qw(  Perf::DataSet::Manager );

sub new {
    my $class = shift;
    my %args = @_;
    my $private_subnet_only = delete $args{private_subnet_only};
    my $jumbo_frames        = delete $args{jumbo_frames};
    my $self = $class->SUPER::new(%args);
    $self->{private_subnet_only} = $private_subnet_only;
    $self->{jumbo_frames} = $jumbo_frames;
    bless $self, $class;
    return $self;
}

sub shutdown {
    my $self = shift;
    $self->SUPER::shutdown();
    if($self->{jumbo_frames}){
        eval{
            $self->disable_jumbo_frames;
        };
        if ($@){
            $Log->comment("Error during jumbo frame disabling: $@");
        }
    }
    return;
}

sub register_dataset {
    my $self = shift;
    my %args = @_;
    my $dataset_obj = $self->SUPER::register_dataset(%args);
    #
    # Load the network ports and IP addresses
    # for filer and clients
    my $appliance = $dataset_obj->get_dependency( class=>'Appliance');
    $appliance->detect_ifconfig;
    my $clients = $dataset_obj->get_dependencies( class=>'Host');
    map { $_->detect_ifconfig } @$clients;

    # Detect IP connectivity between filer and clients
    $appliance->detect_ip_connectivity($clients);

    return $dataset_obj;
}

sub create_workers {
    my ($self,$dataset_uid) = @_;
    #
    # Assign a filer IP to this worker. Compute what port is then used on each side
    #
    my $workers = $self->SUPER::create_workers($dataset_uid);
    my $dataset = Perf::Depot->get_resource(class => 'DataSet', name  => $dataset_uid);
    my $appliance = $dataset->get_appliance_resource;
    my %host      = _get_worker_hosts($workers);

    while (my ($hostname, $hostinfo) = each(%host)) {
        $hostinfo->{filer_ips} = _get_accessible_filer_ips(
            appliance => $appliance,
            dataset   => $dataset,
            hostobj   => $hostinfo->{resource},);
    }

    my $restrict_lifs = $dataset->get_lifs();

    if ($restrict_lifs) {
        my @accepted;

        foreach my $restrict_lif (@$restrict_lifs) {
            foreach my $hostinfo (values(%host)) {
                my @accepted;

                foreach my $ip (@{$hostinfo->{filer_ips}}) {
                    my $lif_name = $ip->get_name();

                    push(@accepted, $ip)
                        if ($lif_name eq $restrict_lif);
                }

                $hostinfo->{filer_ips} = \@accepted;
            }
        }
    }

    if ($appliance->isa('Perf::Depot::Resource::Appliance::GX')) {
        if (!_datasets_mix_workloads($appliance)) {
            # ensure all LIFs are the correct protocol
            my $type = ($dataset->is_nas() ? 'nas' : 'san');

            foreach my $hostinfo (values(%host)) {
                my $ip_for = {$type => $hostinfo->{filer_ips}};
                _switch_protocols($ip_for);
            }
        }
        else {
            $Log->debug('Handling mixed NAS/SAN DataSets');

            foreach my $hostinfo (values(%host)) {
                my $ip_for = {nas => [], san => []};

                foreach my $ip (@{$hostinfo->{filer_ips}}) {
                    if (grep(/iscsi/, $ip->get_protocols())) {

                        push(@{$ip_for->{san}}, $ip);
                    }
                    else {
                        $Log->debug('Assigning LIF ' . $ip->get_name() . ' to NAS');
                        push(@{$ip_for->{nas}}, $ip);
                    }
                }

                my $nas_count = scalar(@{$ip_for->{nas}});
                my $san_count = scalar(@{$ip_for->{san}});

                # LIFs have not been properly split yet
                if (0 == $nas_count || 0 == $san_count || 1 < abs($san_count - $nas_count)) {
                    $ip_for = _split_lifs($hostinfo->{filer_ips});
                    my $changed_lifs = _switch_protocols($ip_for);
                    _relocate_workers($changed_lifs);
                }

                if ($dataset->is_nas()) {
                    $hostinfo->{filer_ips} = $ip_for->{nas};
                }
                else {
                    $hostinfo->{filer_ips} = $ip_for->{san};
                }
            }
        }
    }

    $self->_assign_worker_ips(
        hosts => \%host,
        workers => $workers,);

    return $workers;
}


sub _assign_worker_ips {
    my $self = shift();
    my %arg     = @_;
    my $hosts   = $arg{hosts};
    my $workers = $arg{workers};

    foreach my $worker (@$workers) {
        my $host      = $worker->get_host();
        my $hostname  = $host->get_name();
        my $filer_ips = $hosts->{$hostname}->{filer_ips};

        #
        # Pick one filer IP to use for this worker
        #
        my $worker_filer_ips = $self->choose_filer_ips(worker    => $worker,
                                              filer_ips => $filer_ips,);
        foreach my $worker_filer_ip (@$worker_filer_ips) {
            $Log->comment("Chosen " . $worker_filer_ip->get_description . " for " . $worker->get_description);
            #
            # Determine which host port that filer ip is accessible from
            #
            my $host_ports = $worker_filer_ip->get_port->get_accessible_by;
            confess "No accessible client ports found for " . $worker_filer_ip->get_description
                if (!$host_ports || !@$host_ports);
            my @client_ports = grep { $host == $_->get_host } @$host_ports;

            confess "Unable to find host port able to talk to " . $worker_filer_ip->get_description
                if (!@client_ports);

            $worker->add_dependencies($worker_filer_ip);

            foreach my $client_port (@client_ports) {
                my $client_ip = $client_port->get_user(class => 'Interface');
                confess('No client IP found on ' . $client_port->get_description)
                    if (!defined($client_ip));
                $worker->add_dependencies($client_ip);
            }
        }
    } ## end foreach my $worker (@$workers)
} ## end sub _assign_worker_ips


sub _relocate_workers {
    my $changed_lifs = shift();
    my %affected_datasets;

    while (my ($lif_type, $lifs) = each(%$changed_lifs)) {
        foreach my $lif (@$lifs) {
            # if LIF has Workers using a different protocol, redistribute
            my $lif_workers = Perf::Depot->get_resources(
                class      => 'Worker',
                is_user_of => $lif,);

            foreach my $affected_worker (@$lif_workers) {
                my $affected_dataset = Perf::Depot->get_resource(
                    class => 'DataSet',
                    name  => $affected_worker->get_name(),);

                $Log->debug('LIF '
                        . $lif->get_name()
                        . " has type $lif_type and affected DataSet is "
                        . ($affected_dataset->is_san() ? 'san' : 'nas'));

                if (('san' eq $lif_type) ^ $affected_dataset->is_san()) {
                    $lif->remove_user($affected_dataset);
                    $lif->disconnect_user($affected_worker);

                    my $ds_name = $affected_dataset->get_name();

                    if (!defined($affected_datasets{$ds_name})) {
                        $affected_datasets{$ds_name}->{dataset} = $affected_dataset;
                    }

                    push(@{$affected_datasets{$ds_name}->{workers}}, $affected_worker);
                }
            }
        }
    }

    foreach my $ds (values(%affected_datasets)) {
        $ds->{dataset}->_relocate_workers($ds->{workers});
    }
}


sub _split_lifs {
    my $filer_ips = shift();

    _check_lif_balance($filer_ips);

    my %ip_for;
    my $i = 0;

    # NAS gets even numbered LIFs, SAN odd
    foreach my $ip (@$filer_ips) {
        if (0 == $i % 2) {
            push(@{$ip_for{nas}}, $ip);
        }
        else {
            push(@{$ip_for{san}}, $ip);
        }
        $i++;
    }

    return \%ip_for;
}


sub _switch_protocols {
    my $ip_for = shift();

    my %changed_lif;
    while (my ($type, $lifs) = each(%$ip_for)) {
        foreach my $lif (@$lifs) {
            my $changed = 0;
            if ('nas' eq $type && grep(/iscsi/, $lif->get_protocols())) {
                $Log->debug('LIF ' . $lif->get_name() . ' is SAN, we want NAS');

                my $busy_ds = $lif->get_users(
                    class => 'DataSet',
                    san => 1,
                    busy => 1,);
                confess('SAN LIF in use, cannot switch to NAS')
                    if (defined($busy_ds) && 0 < scalar(@$busy_ds));

                $lif->change_protocol([qw(nfs cifs fcache)]);
                $changed = 1;
            }
            elsif ('san' eq $type && !grep(/iscsi/, $lif->get_protocols())) {
                $Log->debug('LIF ' . $lif->get_name() . ' is NAS, we want SAN');

                my $busy_ds = $lif->get_users(
                    class => 'DataSet',
                    nas => 1,
                    busy => 1,);
                confess('NAS LIF in use, cannot switch to SAN')
                    if (defined($busy_ds) && 0 < scalar(@$busy_ds));

                $lif->change_protocol(['iscsi']);
                $changed = 1;
            }

            if ($changed) {
                push(@{$changed_lif{$type}}, $lif);
            }
        }
    }

    return \%changed_lif;
} ## end sub _switch_protocols


sub _get_accessible_filer_ips {
    my %arg       = @_;
    my $appliance = $arg{appliance};
    my $dataset = $arg{dataset};
    my $host      = $arg{hostobj};
    my $host_ips  = Perf::Depot->get_resources(
        class => 'Interface',
        host  => $host);
    confess "No host IPs found!" if (!$host_ips || !@$host_ips);

    #
    # Get all accessible filer IPs from the host
    #
    my $private_filer_ips = [];
    my $public_filer_ips  = [];
    for my $host_ip (@$host_ips) {
        $Log->debug("Finding accessible filer IPs from: "
                . $host_ip->get_description);
        my $accessible_filer_ips = $host_ip->get_accessible;
        next if (!$accessible_filer_ips);
        for my $filer_ip (@$accessible_filer_ips) {
            $Log->debug("Evaluating " . $filer_ip->get_description);
            next if ($filer_ip->get_appliance != $appliance);
            my $filer_ip_host = $filer_ip->get_host->get_name;
            if ($dataset->get_nodes && none { $_ eq $filer_ip_host }
                @{$dataset->get_nodes}) {
                $Log->debug($filer_ip_host
                        . " not found to be on one of [@{$dataset->get_nodes}]"
                        );
                next;
            }
            if ($filer_ip->is_public
                && !$filer_ip->found_in($public_filer_ips)) {
                push @$public_filer_ips, $filer_ip;
                $Log->debug(
                    "Found public filer ip: " . $filer_ip->get_description);
            }
            elsif (!$filer_ip->is_public
                && !$filer_ip->found_in($private_filer_ips)) {
                push @$private_filer_ips, $filer_ip;
                $Log->debug(
                    "Found private filer ip: " . $filer_ip->get_description);
            }
        }
    } ## end for my $host_ip (@$host_ips)

    # Use the filer private IPs in priority
    my $filer_ips = $private_filer_ips;
    if (!$private_filer_ips || !@$private_filer_ips) {
        $filer_ips = $public_filer_ips;
    }

    confess "No accessible filer IPs found" if (!@$filer_ips);

    # sort LIFs based on name so each Node's LIFs can get different
    # protocols in a mixed environment
    $filer_ips = [sort { $a->get_name() cmp $b->get_name() } @$filer_ips];

    return $filer_ips;
} ## end sub _get_accessible_filer_ips


sub _get_worker_hosts {
	my $workers = shift();
	my %host;

	foreach my $worker (@$workers) {
		my $host = $worker->get_host();
		my $hostname = $host->get_name();

		if (!defined($host{$hostname})) {
			$host{$hostname} = {resource => $host};
		}
	}

	return %host;
}


sub _datasets_mix_workloads {
	my $appliance = shift();
	my @is_nas =
		map { ($_->is_nas() ? 1 : 0) }
		@{Perf::Depot->get_resources(
			class     => 'DataSet',
			appliance_resource => $appliance,)};

	return !((all {$_} @is_nas) || (none {$_} @is_nas));
}


# Ensure there are at least 2 LIFs per node so that a protocol family
# won't be isolated to a single IO path
sub _check_lif_balance {
    my $filer_ips = shift();
    my %ips_for;

    foreach my $ip (@$filer_ips) {
        $ips_for{$ip->get_appliance()->get_name()}++;
    }

    while (my ($node, $count) = each(%ips_for)) {
        confess("Not enough LIFs to do mixed workload on node $node")
            if (2 > $count);
    }
}


sub initialize {
    my ($self, $dataset_uid) = @_;
    if ($dataset_uid){
        my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
        confess "Unknown dataset uid $dataset_uid" if (!$dataset);
        if ($self->{jumbo_frames}){
            $self->enable_jumbo_frames($dataset_uid);
        }
    }
}

# Set port MTU on clients and filer to an appropriate value to
# enable jumbo frames.
#
sub enable_jumbo_frames {
    my ($self,$dataset_uid) = @_;
    my $jumbo_mtu = 9000;
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    confess "Unknown dataset uid $dataset_uid" if (!$dataset);
    my $filer = $dataset->get_dependency( class=>'Filer');
    my $filer_ifs = $filer->get_users( class =>'FilerPort');
    my @dataset_filer_ifs = grep { $->is_used_by($dataset) } @$filer_ifs;
    confess "No filer ports found!" if (!@dataset_filer_ifs);

    # Set MTU to enable jumbo frames on all client ports
    # each filer port can be accessed by
    foreach my $filer_if (@dataset_filer_ifs){
        if ($filer_if->is_public){
            # Don't mock with public ports
            next;
        }
        if ($filer_if->is_shared(class => 'DataSet')){
            # More than 1 dataset are using this filer port
            # This is okay as long as the port is already set to mtu 9000
            if ($filer_if->get_mtu != $jumbo_mtu){
                confess $filer_if->get_description . " is being used by another dataset " .
                                                     "that isn't using jumbo frames";
            }
        }
        # Set the MTU on client ifs
        my $accessible_client_ifs = $filer_if->get_accessible_by;
        for my $client_if (@$accessible_client_ifs){
            next if ($client_if->is_public);
            # This port may have already been set by another dataset
            next if ($client_if->get_mtu == $jumbo_mtu);
            # Client port isn't used by anybody, leave alone
            next if (!$client_if->is_used);
            # At this point client port cannot be shared
            if ($client_if->is_shared( class => 'DataSet')){
                confess $client_if->get_description . " is being used by another dataset " .
                                                     "that isn't using jumbo frames";
            }
            $client_if->set_mtu($jumbo_mtu);
        }
        $filer_if->set_mtu($jumbo_mtu);
    }
}

#
# Restore port MTU on all clients and filers to their original values
# or to just one filer if a filer is passed as an argument
#
sub disable_jumbo_frames {
    my ($self,$dataset_uid) = @_;
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    confess "Unknown dataset uid $dataset_uid" if (!$dataset);
    my $filer = $dataset->get_dependency( class=>'Filer');
    my $filer_ifs = $filer->get_users( class =>'FilerPort');
    my @dataset_filer_ifs = grep { $->is_used_by($dataset) } @$filer_ifs;
    confess "No filer ports found!" if (!@dataset_filer_ifs);

    foreach my $filer_if (@dataset_filer_ifs){
        if ($filer_if->is_public){
            # Don't mock with public ports
            next;
        }
        if ($filer_if->is_shared(class => 'DataSet')){
            # More than 1 dataset are using this filer port
            # This is okay as long as the port is already set to right mtu
            if ($filer_if->get_mtu != $filer_if->get_original_mtu){
                confess $filer_if->get_description . " is being used by another dataset " .
                                                     "that may still be using jumbo frames";
            }
        }
        # Set the MTU on client ifs
        my $accessible_client_ifs = $filer_if->get_accessible_by;
        for my $client_if (@$accessible_client_ifs){
            next if ($client_if->is_public);
            # This port may have already been set by another dataset
            next if ($client_if->get_mtu == $client_if->get_original_mtu);
            # Client port isn't used by anybody, leave alone
            next if (!$client_if->is_used);
            # At this point client port cannot be shared
            if ($client_if->is_shared( class => 'DataSet')){
                confess $client_if->get_description . " is being used by another dataset " .
                                                     "that may still be using jumbo frames";
            }
            $client_if->set_mtu($client_if->get_original_mtu);
        }
        $filer_if->set_mtu($filer_if->get_original_mtu);
    }
}


# Select IPs to assign to a worker
# Arguments:
#   worker => A Perf::Depot::Resource::Worker
#   filer_ips => An arrafref of Perf::Depot::Resource::Interface objects
# Returns:
#   The arrayref of IPs to be assigned to a worker
sub choose_filer_ips {
    my $self = shift();
    my %param = Params::Validate::validate(
        @_,
        {worker => {type => OBJECT,
                    isa  => 'Perf::Depot::Resource::Worker',},
         filer_ips => {type => ARRAYREF},
        });

    confess('No appliance IPs provided')
        if (0 == scalar(@{$param{filer_ips}}));

    my $idx = $param{worker}->get_order % scalar(@{$param{filer_ips}});
    return [$param{filer_ips}->[$idx]];
}

1;
