# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Manager subclass specific to NFS
## @status   private
##
## @pod here

package Perf::DataSet::Manager::NASManager;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Perf::Depot;

use namespace::clean;
use Perf::DataSet::Manager::IPManager;
use base qw( Perf::DataSet::Manager::IPManager );


# TODO: This doesn't work well on forked workers, may have to mount the Perf::DataSet first
sub get_worker_full_paths {
    my ($self,$dataset_uid) = @_;
    my $dataset = Perf::Depot->get_resource(class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown Perf::DataSet $dataset_uid";
    }
    my $workers = Perf::Depot->get_resources( class=>'Worker', name=>$dataset_uid);
    my $relative_path = $self->get_relative_path($dataset_uid);
    my %ret = map { $_->get_process_name => join('/',$_->get_dependency( class=>'Mountpoint')->get_name,$relative_path) } @$workers;
    return \%ret;
}


#
# Assign the workers of dataset_uid a mountpoint to the appropriate filer IP and volume
#
sub mount {
    my $self = shift;
    my $dataset_uid = shift;
    my %args = @_;

    confess('Option no longer supported')
        if ($args{local_only});

    my $force_rw = $args{force_rw};
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown Perf::DataSet $dataset_uid";
    }
    my $appliance = $dataset->get_dependency(class =>'Appliance');

    # check to see if any nodes have left the cluster
    $dataset->check_nodes();

    my $vol = $dataset->get_dependency( class=>'Volume');
    if(!$vol){
        confess "No Volume found for " . $dataset->get_description;
    }
    my $mountpoint_class = $self->get_mountpoint_class;


    # Perform some volume level setup
    $self->volume_mount_setup($dataset_uid, force_rw => $force_rw);

    #
    # Compute mount options
    #
    my %mount_opts = $self->get_mount_opts($dataset_uid, force_rw => $force_rw);

    #
    # Worker setup
    #
    my $workers = Perf::Depot->get_resources( class=>'Worker', name=>$dataset_uid);
    confess "No workers found?!" if (!$workers);


    # create and/or assign mountpoints for workers
    for my $worker (@$workers) {
        my $mountpoint =
          $self->get_worker_mount(worker           => $worker,
                                  mountpoint_class => $mountpoint_class,
                                  mount_opts       => \%mount_opts,
                                  volume           => $vol,);
    }

    #
    # Make sure RootDir exists on Mountpoint
    #
    my $root_dir_obj = $dataset->get_rootdir_resource();

    if (!defined($root_dir_obj)) {
        $root_dir_obj =
          Perf::Depot->create_resource(
                                  class    => 'RootDir',
                                  name     => $dataset->get_root_dir(),
                                  qtree    => $dataset->get_qtree_resource(),
                                  volume   => $dataset->get_volume_resource(),
                                  snapshot => $dataset->get_snapshot_resource(),
                                  );

        $root_dir_obj->register();
        $dataset->add_dependencies($root_dir_obj);
        $root_dir_obj->create_root($dataset_uid);
    }

    return;
}

sub unmount {
    my ($self,$dataset_uid, %args) = @_;
    confess "You must provide an argument" if (!$dataset_uid);
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown Perf::DataSet $dataset_uid";
    }
    my $pending_mount = $args{pending_mount};
    my $users_ok = $args{users_ok};

    my $mountpoint_class = $self->get_mountpoint_class();
    my $mountpoints = $dataset->get_mountpoint_resources($mountpoint_class);

    return if (!$mountpoints || !@$mountpoints);

    # Don't stop after the first unmount errors
    # Try to unmount everything before stopping
    my $errors_during_unmount;

    #
    # Unlatch all mountpoints
    #
    for my $mountpoint (@$mountpoints){
        # See if it's being used by somebody
        my $dataset_users = $mountpoint->get_busy_dataset_users() || [];

        if (0 < scalar(@$dataset_users)) {
            my $str = "Unmount not allowed. " . $mountpoint->get_description()
                    . " is being used by:\n"
                    . join('\n', map {$_->get_description()} @$dataset_users);

            if ($users_ok || $pending_mount) {
                $Log->comment($str);
                next;
            }
            else {
                confess($str);
            }
        }

        next if (!$mountpoint->is_latched);
        $mountpoint->unlatch();
    }

    return;
}

sub volume_mount_setup {
    my ($self, $dataset_uid, %args) = @_;
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown Perf::DataSet $dataset_uid";
    }
    my $force_rw = $args{force_rw};
    my $vol = $dataset->get_dependency( class=>'Volume');
    if(!$vol){
        confess "No Volume found for " . $dataset->get_description;
    }

    #
    # Find all the datasets pointing to the volume and
    # export volume as read-only if all datasets are read-only
    #
    my $ro_volume = 1;
    if (!$force_rw){
        my $datasets_to_vol = $vol->get_users(class => 'DataSet');
        map { $ro_volume = 0 if ( ! $_->is_read_only) } @$datasets_to_vol;
    }
    else{
        $ro_volume = 0;
    }
    $vol->export(read_only=>$ro_volume);
    return;
}


sub get_worker_mount {
    my $self = shift();
    my %args = Params::Validate::validate(@_,
                        {worker           => {isa  => 'Perf::Depot::Resource::Worker',},
                         mountpoint_class => 1,
                         mount_opts       => {type => HASHREF},
                         volume           => {isa  => 'Perf::Depot::Resource::Volume',}});

    my $worker          = $args{worker};
    my $worker_filer_ip = $worker->get_appliance_ip();
    my $mountpoint =
      Perf::Depot->get_resource(
                                     class        => $args{mountpoint_class},
                                     appliance_ip => $worker_filer_ip,
                                     host => $worker->get_process()->get_host(),
                                     volume => $args{volume});
    if (!$mountpoint) {
        $mountpoint = Perf::Depot->create_resource(
            class        => $args{mountpoint_class},
            name         => 'Dummy',                   # Will get overwritten
            appliance_ip => $worker_filer_ip,
            host    => $worker->get_process()->get_host(),
            volume  => $args{volume},
            options => $args{mount_opts});
        $mountpoint->register();
    }
    else {
        my $mntopts = $mountpoint->get_mntopts();

        # For now bomb if we detect collision. In the future we may want to come up with a different
        # mountpoint instead
        confess(  'Mountpoint already exists with different mount options'
                . " ('$mntopts' instead of '$args{mount_opts}->{mntopts}'")
          if ($mntopts ne $args{mount_opts}->{mountopts});

        if (!$mountpoint->is_latched()) {
            $mountpoint->latch();
        }
    }

    $worker->add_dependencies($mountpoint);

    return $mountpoint;
} ## end sub mount_worker

1;
