# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Manager subclass specific to NFS
## @status   private
##
## @pod here

package Perf::DataSet::Manager::NFSManager;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use namespace::clean;
use Perf::DataSet::Manager::NASManager;
use base qw(  Perf::DataSet::Manager::NASManager );

sub new {
    my $class = shift;
    my %args = @_;
    my $nfs_mntopts = delete $args{nfs_mntopts};
    my $self = $class->SUPER::new(%args);
    $self->{nfs_mntopts} = $nfs_mntopts;
    bless $self, $class;
    return $self;
}

sub get_mount_opts {
    my ($self,$dataset_uid,%args) = @_;
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown Perf::DataSet $dataset_uid";
    }
    my $force_rw = $args{force_rw};
    my $vol = $dataset->get_dependency( class=>'Volume');
    if(!$vol){
        confess "No Volume found for " . $dataset->get_description;
    }
    my $ro_volume = 1;
    if (!$force_rw){
        my $datasets_to_vol = $vol->get_users(class => 'DataSet');
        map { $ro_volume = 0 if ( ! $_->is_read_only) } @$datasets_to_vol;
    }
    else{
        $ro_volume = 0;
    }
    my %mount_opts;
    if (!$ro_volume){
        $mount_opts{chmod} = '0777';
    }
    # 'soft' mount allow for better recovery after filer panic
    my $mntopts;
    if ($self->{nfs_mntopts}->{hard}) {
        $mntopts = "hard";
    }
    else {
        $mntopts = "soft";
    }

    # 'noac' prevents hosts from seeing different things
    # Don't enable this if there is only one client or if
    # we are not slicing, or if we are slicing only withing
    # clients. In other words, don't enable this unless we
    # really really have to.
    my $clients_count = scalar(@{$dataset->get_clients});
    if ($clients_count > 1){
        if ($dataset->is_file_slicing_allowed && !$dataset->is_file_slicing_single_client){
            $Log->comment("Turning on noac");
            $mntopts .= ",noac";
        }
    }

    # 'udp' mount gave performance boost but isn't well supported
    # under RHEL6
    # 'timeo' and 'retrans' guard against transient
    # network failures
    $self->{nfs_mntopts}->{vers}    ||= 3;
    $self->{nfs_mntopts}->{timeo}   ||= 600;

    if (defined($self->{nfs_mntopts}->{rsize})) {
        $mntopts .= ",rsize=$self->{nfs_mntopts}->{rsize}";
    }
    if (defined($self->{nfs_mntopts}->{wsize})) {
        $mntopts .= ",wsize=$self->{nfs_mntopts}->{wsize}";
    }
    $mntopts .= ",retrans=$self->{nfs_mntopts}->{retrans}";
    $mntopts .= ",timeo=$self->{nfs_mntopts}->{timeo}";

    $mount_opts{proto}              = $self->{nfs_mntopts}->{proto};
    $mount_opts{vers}               = $self->{nfs_mntopts}->{vers};
    $mount_opts{mountopts}          = $mntopts;
    $mount_opts{retries}            = 0;
    $mount_opts{nacltask_if_exists} = 'reuse';

    return %mount_opts;
}

# TODO: Don't do this. Instead modify NFSMountpoint and uCIFSMountpoint so that
# they report a class of 'Mountpoint'
sub get_mountpoint_class {
    #return 'NFSMountpoint';
   return 'Mountpoint::NFS';
}

1;
