## @author dl-perfqa-automation@netapp.com
## @summary Manager subclass for SAN technologies
## @status private
##
## @pod here

=head1 NAME

Perf::DataSet::Manager::SANManager - Manage storage elements in a SAN

=head1 DESCRIPTION

This package manages SAN entities such as LUNs and raw, block devices
on clients and the protocols used to connect the two such as iSCSI or
FCP

=head1 METHODS

=cut

package Perf::DataSet::Manager::SANManager;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log);
use Perf::Utils qw(min max get_random_element as_string in_bytes deep_copy ndmp_copy floor);

BEGIN {
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}

use Data::Dumper;
use Perf::Depot;
use Params::Validate qw(validate SCALAR OBJECT UNDEF);
use List::Util qw(first);
use List::MoreUtils qw(any);

use namespace::clean;
use Perf::DataSet::Manager;
use base qw( Perf::DataSet::Manager );


=head2 C<_init_lun>

Initialize a physical LUN on a storage appliance

=head3 Example

 $self->_init_lun(dataset_uid  => $dataset_uid,
                  host         => $host,
                  lun_size     => $addition_sizes->{$addition_uid},
                  addition_uid => $addition_uid,
                  create_lun   => $create_lun,);


=head3 Arguments

=over

=item dataset_uid => B<string>

The UID of the Perf::DataSet using this LUN

=item host => B<Perf::Depot::Resource::Host>

The client that will be accessing the LUN

=item lun_size => B<string> | B<integer>

The size of the LUN in bytes

=item addition_uid => B<string>

The UID of the Addition that is creating this LUN

=item create_lun => B<boolean>

Flag to create the physical LUN in addition to the Perf::Depot::Resource
representing the LUN

=back

=cut

sub _init_lun {
    my $self = shift();
    my %params = validate(@_,
                          {dataset_uid => {type => SCALAR},
                           host        => {
                                    type => OBJECT,
                                    isa  => 'Perf::Depot::Resource::Host'},
                           lun_size     => {type => SCALAR},
                           addition_uid => {type => SCALAR},
                           create_lun   => {type => SCALAR},});

    my $dataset =
      Perf::Depot->get_resource(class => 'DataSet',
                                   name  => $params{dataset_uid},);
    confess "Unknown Perf::DataSet $params{dataset_uid}" if (!$dataset);

    my $volume = $dataset->get_volume_resource();
    confess "No volume found for " . $dataset->get_description if (!$volume);
    my $qtree = $dataset->get_qtree_resource();

    my $host = $params{host};

    my $lun_size  = int(in_bytes($params{lun_size}) / 1024) . 'k';
    my $host_name = $host->get_name();
    my $lun_name =
      $dataset->make_lun_name(addition_uid => $params{addition_uid},
                              client_name  => $host_name);
    my $lun_type = $host->get_lun_type();
    my $lun_obj =
      Perf::Depot->create_resource(class      => 'Lun',
                                            name       => $lun_name,
                                            volume     => $volume,
                                            qtree      => $qtree,
                                            type       => $lun_type,
                                            size       => $lun_size,
                                            create_lun => $params{create_lun});

    return $lun_obj;
} ## end sub _init_lun


=head2 C<depend_dev_on_filer_resource>

This method must be overridden in a subclass of SANManager.  It should
add a dependency on the filer/node resource that provides access to
a LUN (i.e. network interface for iSCSI, FC adapter for FCP).

=head3 Arguments

=over

=item raw_dev

=item dataset_uid

=back

=head3 Example

 $raw_dev = Perf::Depot->create_resource(class => 'SANDevice'...);

 $san_manager->depend_dev_on_filer_resource(raw_dev => $raw_dev,
                                            dataset_uid => $dataset_uid);

=cut

sub depend_dev_on_filer_resource {
    confess("Must be implemented in a subclass");
}

sub mount {
    my $self        = shift();
    my $dataset_uid = shift();
    my %params      = @_;

    my $dataset =
      Perf::Depot->get_resource(class => 'DataSet',
                                         name  => $dataset_uid,);
    if (!$dataset) {
        confess "Unknown Perf::DataSet $dataset_uid";
    }

    $dataset->check_nodes();
    $self->check_initiators($dataset);

    # LUN_ids = lun_prefix ✕ addition_uids ✕ clients
    my $ds_lun_names   = $dataset->get_lun_names();
    my $vol_obj        = $dataset->get_volume_resource();
    my $qtree_obj      = $dataset->get_qtree_resource();
    my $container      = $dataset->get_container();
    my $addition_sizes = $dataset->get_addition_info();
    my %addition_actuals;
    my $physical_luns  = $vol_obj->get_luns();
    my $client_count   = scalar(@{$dataset->get_clients()});

    # In SAN Perf::DataSets, File Elements are used to access raw device
    # files nodes.  These Raw File Elements store the name of the LUN
    # they are accessing.  Query the Contaniner and build up a LUN
    # name to raw device map.
    my %element_map =
      map { $_->get_lun_id() => $_->get_path() }
      values(%{$container->get_elements('files', include_deleted => 0)});

    # the list of LUN names has been built by Perf::DataSet to include all
    # the present and expected LUNs based on Additions
    foreach my $lun_name (@$ds_lun_names) {
        # parse the LUN name into pieces for later use
        my (undef, $addition_uid, $hostname) = split(/_/, $lun_name, 3);

        my $host =
          Perf::Depot->get_resource(class => 'Host',
                                             name  => $hostname,);
        my $lun_obj =
          Perf::Depot->get_resource(class  => 'Lun',
                                             name   => $lun_name,
                                             volume => $vol_obj,
                                             qtree  => $qtree_obj,);
        my $raw_device;

        if (defined($lun_obj)) {
            # the LUN is in the Depot, confirm the raw device used to
            # access it
            if (exists($element_map{$lun_name})) {
                $raw_device =
                  Perf::Depot->get_resource(
                                                class => 'SANDevice',
                                                name => $element_map{$lun_name},
                                                lun  => $lun_obj,
                                                host => $host,);
            }
            else {
                confess("No Element for LUN $lun_name?");
            }
        }
        else {
            # the LUN is not in the Depot, this LUN is the result of
            # either a fork() or an Addition
            my $lun_phy_present  = any { $lun_name eq $_->get_name() } @$physical_luns;
            my $old_raw_dev_name = undef;

            if ($lun_phy_present) {
                # $lun_name will be in %element_map on a fork
                if (exists($element_map{$lun_name})) {
                    $old_raw_dev_name = $element_map{$lun_name};
                }
                else {
                    # otherwise this must be a LUN fork, we need to find
                    # the old LUN name by the Addition UID and host name
                    my $lun_name_re =
                      join('_', ('^[^_]+', $addition_uid, $hostname,));
                    $old_raw_dev_name = first { $_ =~ m{$lun_name_re}msx }
                                              keys(%element_map);
                }
            }

            my $appliance = $vol_obj->get_appliance();

            # For an Addition, each client will get 1 LUN, so make the
            # LUN size the Addition size ÷ client count
            my $lun_size =
              floor(in_bytes($addition_sizes->{$addition_uid}) / $client_count);

            my $igroup =
                Perf::Depot->get_resource(
                    class     => 'IGroup',
                    appliance => $appliance,
                    host      => $host,
                    protocol  => $self->get_protocol(),);

            confess(  'No IGroup for '
                    . $appliance->get_name() . ' and '
                    . $host->get_name())
              if (!defined($igroup));

            my $new_lun_rsc =
              $self->_init_lun(dataset_uid  => $dataset_uid,
                               host         => $host,
                               lun_size     => $lun_size,
                               addition_uid => $addition_uid,
                               create_lun   => ($lun_phy_present ? 0 : 1),);

            my $lun_map =
              Perf::Depot->create_resource(class     => 'LUNMap',
                                                    name      => 'Dummy',
                                                    igroup    => $igroup,
                                                    lun       => $new_lun_rsc,);

            # wait for LUN map to come online before scanning for new
            # devices
            sleep(15);

            # LUN is now mapped, have client rescan SCSI bus
            $host->discover_san_devices(protocol => $self->get_protocol());

            $raw_device = $self->create_raw_device(lun_map => $lun_map);

            $raw_device->register();
            $self->depend_dev_on_filer_resource(raw_dev     => $raw_device,
                                                dataset_uid => $dataset_uid);

            # LUN does physically exist, must be a volume or
            # LUN fork(), a raw device will need to be relocated
            if ($lun_phy_present) {
                confess("Couldn't determine old device name")
                    if (!defined($old_raw_dev_name));

                my $new_raw_dev_name = $raw_device->get_name();

                # relocate Element in controller Container…
                $container->relocate_element($old_raw_dev_name,
                                             $new_raw_dev_name);

                my $workers =
                  Perf::Depot->get_resources(class   => 'Worker',
                                                      host    => $host,
                                                      used_by => $dataset);

                # …and on all clients
                map {
                    $_->relocate_element($old_raw_dev_name, $new_raw_dev_name)
                } @$workers;
            }
            else {
                # LUN wasn't present.  This means we've created it.
                # The LUN created is sometimes larger than requested
                # so collect actual sizes to update the Perf::DataSet
                # addition info so that later validations won't fail.
                $addition_actuals{$addition_uid} += $new_lun_rsc->get_size();
            }
        } ## end else [ if (defined($lun_obj))

        $self->_confirm_raw_device($raw_device);
    } ## end foreach my $lun_name (@$ds_lun_names)

    # update addition sizes with actual sizes
    while (my ($uid, $size) = each(%addition_actuals)) {
        $dataset->insert_addition_info($uid, $size);
    }

    $self->_depend_workers($dataset);

    return;
} ## end sub mount


sub unmount {
    my ($self,$dataset_uid) = @_;
    confess "You must specify an argument" if (!$dataset_uid);
    my $dataset = Perf::Depot->get_resource( class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown Perf::DataSet $dataset_uid";
    }
    my $dataset_client_objs = $dataset->get_client_resources;
    for my $host (@$dataset_client_objs){
        my $luns = $dataset->get_lun_resources( $host );

        foreach my $lun (@$luns) {
            my $device =
              Perf::Depot->get_resource(class => 'SANDevice',
                                                 lun   => $lun,
                                                 host  => $host);
            if (!$device){
                # The device may not exist if we just did a fork
                $Log->comment("No raw device found on " . $host->get_description .
                                  " for " . $host->get_description );
            }
            else{
                $device->unlatch;
            }
        }
    }

    return;
}


sub get_worker_full_paths {
    my ($self,$dataset_uid) = @_;
    my $dataset = Perf::Depot->get_resource(class => 'DataSet', name=>$dataset_uid);
    if (!$dataset){
        confess "Unknown Perf::DataSet $dataset_uid";
    }
    my $workers = Perf::Depot->get_resources( class=>'Worker', name=>$dataset_uid);

    my %ret = map { $_->get_process_name => $_->get_host()->get_raw_dev_base() } @$workers;
    return \%ret;
}


# Ensure that the raw device used to access a LUN on a client exists
# and is functioning.
sub _confirm_raw_device {
    my $self = shift();
    my $raw_device = shift();

    confess("Expected raw device does not exist in Depot")
        if (!defined($raw_device));

    confess("Argument is not a SANDevice")
        if (!$raw_device->isa('Perf::Depot::Resource::SANDevice'));

    if (!$raw_device->is_latched()) {
        $raw_device->latch();
        $raw_device->check_valid();
    }
    # is_latched() calls check_valid() in the event the device is
    # already latched

    return 1;
}


=head2 C<_depend_workers>

Ensure all Workers depend on the managed raw devices.

=head3 Arguments

=over

=item B<Perf::DataSet>

The controlling Perf::DataSet

=back

=cut

sub _depend_workers {
    my ($self, $dataset) = @_;

    # All raw devices should be present at this point.  Ensure the
    # Workers depends on them.
    foreach my $host (@{$dataset->get_client_resources()}) {
        my $host_devices =
          Perf::Depot->get_resources(class   => 'SANDevice',
                                              host    => $host,
                                              used_by => $dataset,);
        foreach my $device (@$host_devices) {
            my $host_workers =
              Perf::Depot->get_resources(class => 'Worker',
                                                  name  => $dataset->get_uid(),
                                                  host  => $host,);

            confess "No Workers found on "
              . $host->get_description
              . " for dataset '"
              . $dataset->get_uid() . "'"
              if (!$host_workers || !@$host_workers);

            map { $_->add_dependencies($device) } @$host_workers;
        }
    }
}




sub create_raw_device {
    my $self = shift();
    my %param = validate(@_,
                         {lun_map => {type => OBJECT,
                                      isa  => 'Perf::Depot::Resource::LUNMap',
                            },});

    my $host = $param{lun_map}->get_igroup()->get_host();
    my $appliance = $param{lun_map}->get_lun()->get_volume()->get_appliance();


    my $host_san_service =
        Perf::Depot->get_resource(class => $self->get_service_type(),
                                           host  => $host);

    if (!$host_san_service) {
        confess('No SANService found for host ' . $host->get_description());
    }

    my $new_dev =
      Perf::Depot->create_resource(class       => 'SANDevice',
                                            name        => 'Dummy',
                                            lun_map     => $param{lun_map},
                                            host        => $host,
                                            san_service => $host_san_service,);

    return $new_dev;
}


sub get_protocol {
    confess(
          'Subroutine get_protocol() must be implemented in a subclass of '
            . __PACKAGE__);
}


# There are a bunch of clients out there with duplicate initiator names
# due to EngSup disk cloning.  This will lead to failures later on.
sub check_initiators {
    my $self = shift();

    my $dataset = shift();

    confess('Expecting a Perf::DataSet Resource, got a ' . ref($dataset))
      if (!$dataset->isa('Perf::Depot::Resource::DataSet'));

    my $clients = $dataset->get_client_resources();
    my %initiator_client;

    foreach my $client (@$clients) {
        my $client_name = $client->get_name();
        my $initiator_name =
          $client->get_initiator_name(protocol => $self->get_protocol());

        if (!defined($initiator_client{$initiator_name})) {
            $initiator_client{$initiator_name} = $client_name;
        }
        else {
            confess("Duplicate iSCSI initiator name ($initiator_name) detected"
                . " on clients: $initiator_client{$initiator_name} & $client_name"
                . "\nFile an EngSup ticket to have this issue corrected"
                );
        }
    }
}


sub create_services {
    my $self = shift();
    my $dataset_obj = shift();

    my $clients = $dataset_obj->get_dependencies(class => 'Host');
    my $appliance = $dataset_obj->get_appliance_resource();

    foreach my $host (@$clients) {
        my $host_san_service =
            Perf::Depot->get_resource(class => $self->get_service_type(),
                                               host  => $host);

        if (!$host_san_service) {
            $host_san_service = Perf::Depot->create_resource(
                class      => $self->get_service_type(),
                name       => 'Dummy',
                host       => $host,
                appliances => [$appliance],
            );
            $host_san_service->register();
            $host_san_service->configure();
        }

        my $igroup = Perf::Depot->get_resource(
            class       => 'IGroup',
            appliance   => $appliance,
            san_service => $host_san_service,
            protocol    => $self->get_protocol(),);
        if (!defined($igroup)) {
            $igroup = Perf::Depot->create_resource(
                class       => 'IGroup',
                name        => 'Dummy',
                appliance   => $appliance,
                san_service => $host_san_service,
                protocol    => $self->get_protocol(),);
            $igroup->register();
        }

        # ensure igroup exists as long as DataSet does
        $dataset_obj->add_dependencies($igroup);
    }

    return;
}

1;
