# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Manager subclass for iSCSI protocol
## @status   private
##
## @pod here

package Perf::DataSet::Manager::SANManager::iSCSIManager;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use List::MoreUtils qw(uniq);
use Params::Validate qw(validate SCALAR OBJECT UNDEF ARRAYREF);
use Data::Dumper;

use Perf::Depot;

use namespace::clean;
use Perf::DataSet::Manager::IPManager;
use Perf::DataSet::Manager::SANManager;
use base qw(Perf::DataSet::Manager::IPManager Perf::DataSet::Manager::SANManager);


# fix diamond inheritance problem by turning off typical depth first
# search algorithm used for subroutine lookup
use MRO::Compat;
mro::set_mro('Perf::DataSet::Manager::SANManager::iSCSIManager', 'c3');


sub create_workers {
    my $self        = shift();
    my $dataset_uid = shift();

    my $workers = $self->SUPER::create_workers($dataset_uid);

    my $dataset = Perf::Depot->get_resource(class => 'DataSet',
                                                     name  => $dataset_uid,);

    # ensure SAN services are created after iSCSI LIFs are available
    $self->create_services($dataset);

    return $workers;
}


# make iSCSI devices depend on the filer IP they'll be accessing
sub depend_dev_on_filer_resource {
    my $self = shift();
    my %param = validate(
                        @_,
                        {dataset_uid => {type => SCALAR},
                         raw_dev     => {
                             type => OBJECT,
                             isa => 'Perf::Depot::Resource::SANDevice'
                           },});

    my $raw_dev  = $param{raw_dev};
    my $host_obj = $raw_dev->get_host();
    my $workers =
      Perf::Depot->get_resources(class => 'Worker',
                                    name  => $param{dataset_uid},
                                    host  => $host_obj);

    foreach my $worker (@$workers) {
        my $ip_rscs = $worker->get_appliance_ips();

        $raw_dev->add_dependencies($ip_rscs);
    }
}


sub get_protocol {
    my $self = shift();

    return 'iscsi';
}


sub get_service_type {
    my $self = shift();

    return 'SANService::iSCSI';
}


sub choose_filer_ips {
	my $self = shift();
	my %param = validate(@_,
		{worker => {
			type => OBJECT,
			isa  => 'Perf::Depot::Resource::Worker',
		},
		filer_ips => {type => ARRAYREF},
	});

	confess('No appliance IPs provided')
		if (0 == scalar(@{$param{filer_ips}}));

	if ($param{worker}->get_host()->enable_multipathing()) {
		# iSCSI interacts with all filer IPs when multipathing is on
		return $param{filer_ips};
	}
	else {
		# pick a single IP per worker
		return $self->SUPER::choose_filer_ips(%param);
	}
}

1;
