# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Perf::DataSet::ModularWorkload - Runs a user-defined workload packaged as a Sequence of Operations

=head1 DESCRIPTION

Users may define their own workloads by creating a subclass of Perf::DataSet::Sequence that implements
the 'next_operation' method passing an instance of their subclass to $dataset->modular_workload.
This method should return an object belonging to one of the supported Perf::DataSet::Operation subclasses.
Examples: File::Read, File::Overwrite, File::Append, etc... See packages under Perf::DataSet/Operation for
a complete list and documentation.

    Example of a fixed-size random overwrite Sequence:

        # In file .../RandomWriteWorload.pm
        package RandomWriteWorkload;{
            use strict;
            use Object::InsideOut qw( Perf::DataSet::Sequence::File );

            # Define member attribute 'io_size'
            my @io_size :Field :Arg( Name=>'io_size', Mandatory=>1 ) :Get(get_io_size);

            sub next_operation {
                my $self   = shift;
                my $file   = $self->random_file;
                my $offset = $self->random_offset($file);
                my $size= $file->before_eof($offset, $self->get_io_size);
                return Perf::DataSet::Operation::File::Overwrite->new(element => $file,
                                                                offset  => $offset,
                                                                size    => $size);
            }
        }
        1;

        # Back in your test, after having created $dataset and added files into it
        my $random_write = RandomWriteWorkload->new(io_size=>'4k');
        my $ret = $dataset->modular_workload( sequence=>$random_write, duration=>30 );

See Perf::DataSet::Sequence for available methods.

=cut

use strict;
use warnings;

package Perf::DataSet::ModularWorkload;
{
    use Carp qw(confess);
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Params::Validate qw(validate_with SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
    use Time::HiRes qw(time);
    use Perf::Utils qw( in_bytes as_string );
    use Data::Dumper;
    use Storable qw();
    use MIME::Base64;
    use Perf::DataSet::Command;
    use Perf::DataSet::Container;
    use Perf::Depot;
    use Perf::DataSet::Operation;
    use Perf::DataSet::Sequence;

    # Workload is a subclass of Command
    use base qw(Perf::DataSet::Command);

    # Most OSes don't let you open too many files at once
    # Set the maximum of open files to 1000
    our $MAX_OPEN_FILES = 1000;

    # IO latency histogram bucket size (in ms)
    our $BUCKET_MS = 10;

    sub new {
        my $class = shift;
        my %params = $class->validate(@_);
        my $self = \%params;
        $self->{abort} = 0;
        $self->{op_context} = {}; # Cache accessible to Operations
        bless $self, $class;
        # Set some defaults
        $self->die_on_operation_error;
        $self->die_on_file_close_error;
        return $self;
    }

    sub is_aborted {
        my $self = shift;
        return $self->{abort};
    }

    sub set_abort {
        my ($self) = @_;
        $self->{abort} = 1;
    }

=head2 die_on_operation_error

Indicates whether workload should die from the last Operation error should one have occured.

=over

=item Arguments

boolean value

I<Optional> Defaults to 1.

=back

=cut

    sub die_on_operation_error {
        my ($self,$value) = @_;
        $value = 1 if (!defined($value));
        $self->{die_on_operation_error} = $value;
        return;
    }

=head2 will_die_on_operation_error

Reports whether workload will die from the last Operation error should one occur.

=cut

    sub will_die_on_operation_error {
        my ($self) = @_;
        $Log->comment("will_die_on_operation_error: $self->{die_on_operation_error}");
        return $self->{die_on_operation_error};
    }

=head2 die_on_file_close_error

Indicates whether workload should die from the last error that occured while closing files
should there be one.

        # Disable dying from errors on close
        $self->die_on_file_close_error(0);

=over

=item Arguments

boolean value

I<Optional> Defaults to 1.

=back

=cut

    sub die_on_file_close_error {
        my ($self,$value) = @_;
        $value = 1 if (!defined($value));
        $self->{die_on_file_close_error} = $value;
        return;
    }

=head2 will_die_on_file_close_error

Reports hether workload will die from the the last error that occured while closing files
should there be one.

=cut

    sub will_die_on_file_close_error {
        my ($self) = @_;
        $Log->comment("will_die_on_file_close_error: $self->{die_on_file_close_error}");
        return $self->{die_on_file_close_error};
    }

    sub validate {
        my $class = shift;
        my %params =
            validate_with(
                params      => \@_,
                spec        => {
                    sequence => {type => OBJECT, optional => 0, isa => 'Perf::DataSet::Sequence'},
                    #recorder => {type => OBJECT | UNDEF, optional => 1, isa => 'Perf::DataSet::Recorder'},
                    # async: For asynchronous loads
                    async => {type => SCALAR | UNDEF, optional => 1, default => 0},
                    description => {type => SCALAR | UNDEF, optional => 1, default => "workload"},
                    # event_interval: number of seconds between event detection
                    event_interval => {type => SCALAR , optional => 1, default => 1},
                    # event_interval: number of operations between event detection
                    event_enable => {type => SCALAR | UNDEF , optional => 1, default => 1},
                    # skip_flush: don't sync filers before returning
                    skip_flush  => {type => SCALAR | UNDEF, optional => 1, default => 0},
                    # skip_container_validate: don't attempt to validate container against reality after workload
                    skip_container_validate  => {type => SCALAR | UNDEF, optional => 1, default => 0},
                    # start_on_due: workload won't start until it receives a start event
                    #start_on_cue => {type => SCALAR | UNDEF, optional => 1, default => 0},
                    # on_error_handler: subroutine called after operation error
                    on_error_handler => {type => CODEREF | UNDEF, optional => 1},
                    # type: Supports 'nfs'
                    type => {type => SCALAR | UNDEF, optional => 1, default => 'nfs'},
                    # debug
                    debug => {type => SCALAR | UNDEF, optional => 1, default => 0},
                },
                allow_extra => 0,
            );
        return %params;
    }

    sub get_container {
        my ($self) = @_;
        my $sequence = $self->{sequence};
        confess "No sequence found" if (!$sequence);
        my $container = $sequence->get_container;
        confess "No container found" if (!$container);
        return $container;
    }

    sub pre_execute($$) {
        my $self = shift;
        my $owner = shift;
        if (!$owner || !ref($owner)){
            confess "You must specify an owner";
        }

        if ($owner->isa('Perf::DataSet::Worker')){

            eval "require $self->{sequence_package}";
            confess "Unable to load $self->{sequence_package}: $@" if ($@);
            # This is a necessary hack for Object::InsideOut
            {local $@; eval "$self->{sequence_package}->new";}
            $self->{sequence} = Storable::thaw(decode_base64($self->{sequence}));

            $self->{sequence}->pre_execute($owner);

            #
            # This worker may be brand new off a fork
            # Reset the mountpoint and relative path
            #
            my $process_uid = $owner->get_process_uid;

            # Cd to appropriate directory
            my $full_path = $self->{worker_full_paths}->{$process_uid};
            confess "No full path found!" if (!$full_path);
            $owner->{full_path} = $full_path;

            $Log->comment("Changing to directory '$full_path'");
            chdir $full_path or confess "Unable to chdir to '$full_path': $!";
        }
        elsif($owner->isa('Perf::DataSet')){

            # Inherit 'type' (i.e protocol)
            # and pass it to Sequence
            $self->{type} = $owner->get_type;
            $self->{sequence}->set_type($self->{type});

            $self->{sequence}->pre_execute($owner);
            $self->{sequence_package} = ref( $self->{sequence} );
            $self->{sequence} = $self->{sequence}->rpc_serialize;

            # TODO: Move this to the truncation sequence
            #if ($self->{delete_first} && $owner->is_file_slicing_allowed){
            #    confess "Argument delete_first is not supported for datasets that allow file slicing";
            #}

            #
            # Set the current number of active workers
            # This is used to dynamically adjust the load
            #
            my $process_count = scalar(@{$owner->{manager}->get_processes($owner->{uid})});
            $self->{processes} = $process_count;

            #
            # Unmount and remount before reads (to clear client cache)
            # Skip NFS and iSCSI since O_DIRECT works for those protocols
            #
            if ($self->{read_pct} && $self->{type} !~ /nfs|iscsi|fcp/){
                $owner->unmount( pending_mount=>1 );
            }

            # Mount the Perf::DataSet. Paths for each worker have already been communicated during the add
            # No need to do it again.
            $owner->mount;

            # Embed the worker paths to the command so that workers can load their own
            # While this is already done in the Addition, forked datasets
            # may never execute an addition, so do this every time
            $self->{worker_full_paths} = $owner->get_worker_full_paths;

        }
        else{
            confess "The owner must be a Perf::DataSet or Perf::DataSet::Worker and not a " . ref($owner);
        }
    }

    sub post_execute($$){
        my ($self, $owner, $returns) = @_;
        if (!$owner || !ref($owner)){
            confess "You must specify an owner";
        }

        if ($owner->isa('Perf::DataSet::Worker')){
            $self->{sequence}->post_execute($owner,$returns);
        }
        elsif($owner->isa('Perf::DataSet')){
            #
            # Make sure all data is written to disk
            #
            if (!$self->{skip_flush}){
                $Log->comment("Flushing appliance from " . __PACKAGE__ );
                my $appliance = Perf::Depot->get_resource(class=>'Appliance', name=>$owner->get_appliance);
                confess "No appliance found?!" if (!$appliance);
                $appliance->flush;
            }
            #
            # Merge results and return
            #
            my $load_report;
            my $tmp_container = Perf::DataSet::Container->new;
            for my $process_uid (keys %$returns){
                #
                # Handle the 'common' section
                #
                # TODO: Recursive algorithm!
                my $return = $returns->{$process_uid}->{common};
                $load_report->{$process_uid} = $return;
                for my $op (keys %$return){
                    my $op_stats = $return->{$op};
                    for my $op_stat (keys %$op_stats){
                        my $v = $op_stats->{$op_stat};
                        my $v_ref = ref($v);
                        if (!$v_ref){
                            $load_report->{$op}->{$op_stat} += $v;
                        }
                        elsif ($v_ref eq 'HASH'){
                            for my $v_k (keys %$v){
                                my $v_v = $v->{$v_k};
                                my $v_v_ref = ref($v_v);
                                if (!$v_v_ref){
                                    $load_report->{$op}->{$op_stat}->{$v_k} += $v_v;
                                }
                                elsif( $v_ref eq 'HASH' ){
                                    for my $v_v_k (keys %$v_v){
                                        my $v_v_v = $v_v->{$v_v_k};
                                        my $v_v_v_ref = ref($v_v_v);
                                        if (!$v_v_v_ref){
                                            $load_report->{$op}->{$op_stat}->{$v_v_k} += $v_v_v;
                                        }
                                        else{
                                            confess "Ref '$v_v_v_ref' not supported";
                                        }
                                    }
                                }
                                else{
                                    confess "Ref '$v_v_ref' not supported";
                                }
                            }
                        }
                        else{
                            confess "Ref '$v_ref' not supported";
                        }
                    }
                }

                #
                # Handle the 'elements' section
                #
                $return = $returns->{$process_uid}->{elements};
                my @elems = map { Storable::thaw(decode_base64($_)) } @$return;
                $tmp_container->set_elements(\@elems);

            }
            #$Log->comment("Workload:\n" . as_string($load_report));



            $tmp_container->unpartition;
            #$Log->comment("DEBUG: tmp_container:\n" . $tmp_container->as_string);
            $owner->{container}->merge_with($tmp_container);

            if (!$self->{skip_container_validate}){
                $owner->validate_container;
            }

            #$Log->comment("New container stats:\n" . $owner->{container}->get_stats->as_string );

            return $load_report;
        }
        else{
            confess "The owner must be a Perf::DataSet or Perf::DataSet::Worker and not a " . ref($owner);
        }
        return $returns;
    }

    sub execute {
        my ($self) = @_;
        my $sequence = $self->{sequence};
        confess "Invalid sequence" if (!$sequence || !$sequence->isa('Perf::DataSet::Sequence'));
        my $recorder = $self->{sequence}->get_default_recorder;
        confess "Invalid recorder" if (!$recorder || !$recorder->isa('Perf::DataSet::Recorder'));
        $sequence->set_recorder($recorder);

        my $event_interval = $self->{event_interval};
        my $event_enable = $self->{event_enable};

        #if ($self->{start_on_cue}){
        #    $Log->comment("Waiting on start event");
        #    my $event = $self->wait_for_next_event;
        #    confess "Unexpected event:\n" . Dumper($event) if ( $event->{type} ne 'start');
        #    $Log->comment("Received start event");
        #}

        my $op_count = 0;
        my $start_time = time;
        my $last_event_processing = $start_time;
        my $end_time;
        my $last_operation;
        eval{
            while ( !$self->is_aborted ) {
                my @operations = $sequence->next;
                last if (!@operations);
                for my $operation ( @operations ){
                    $last_operation = $operation;
                    eval{
                        $operation->execute($sequence);
                    };
                    # Unrecognized errors abort workload abruptly
                    # others are passed to a user definable handle_error sub
                    if($@){
                        my $err = $@;
                        if (my $e = Perf::DataSet::Operation::Exception->caught){
                            $operation->set_error($e);
                            undef $@;
                        }
                        else{
                            die $err;
                        }
                    }
                    $end_time = time;
                    $recorder->record($operation, $end_time - $start_time);
                    $start_time = $end_time;
                    $op_count ++;
                    $self->handle_error($operation->get_error) if ($operation->has_failed);
                }
                if ($event_enable && ($end_time >= $last_event_processing + $event_interval) ){
                    $self->process_events;
                    $last_event_processing = $end_time;
                }
            }
            $Log->comment("Workload was cleanly aborted") if ($self->is_aborted);
        };
        my $err_during_workload = $@;
        $Log->comment("Workload was abruptly aborted") if ($err_during_workload);
        #eval{$self->{sequence}->close_fhs};
        #my $err_during_close = $@;
        #$self->handle_error($err_during_close) if ($err_during_close);
        my $err_during_last_op = $last_operation?$last_operation->get_error():undef;

        die $err_during_last_op if ($err_during_last_op && $self->will_die_on_operation_error);
        die $err_during_workload if ($err_during_workload);
        #confess $err_during_close if ($err_during_close && $self->will_die_on_file_close_error);

        return $recorder->get_report;
    }

    sub start {
        my ($self) = @_;
        # TODO: Ensure all processes are ready first!!
        my $start_event = Perf::DataSet::Event->new(type=>'start');
        $self->send_event($start_event);
        return;
    }

    # Answer Abort events
    sub handle_Abort {
        my $self = shift;
        $self->set_abort;
    }

    sub handle_error {
        my ($self,$error) = @_;
        confess "No error passed in!" if (!$error);
        if (!ref($error)){
            $Log->comment("Handling error: $error");
        }
        else{
            $Log->comment("Handling error:\n" . Dumper($error));
        }
        my $handler = $self->{on_error_handler};
        # By default, all errors trigger the workload to abort
        # and die
        if (!$handler){
            $self->set_abort;
            $self->die_on_operation_error;
            $self->die_on_file_close_error;
        }
        else{
            $handler->($self,$error);
        }
        return;
    }

    sub get_sequence {
        return $_[0]->{sequence};
    }

}
1;
