# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Operation - Base class for Workload operations

=head1 DESCRIPTION

This class is the base class for all Operation subclasses. It provides all the common functionality.

It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Operation;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    # Declare base classes for catchable errors.
    # All calls to the storage system (i.e. the filer or node) should throw
    # a subclass of StorageException in case of errors
    use Exception::Class ( 'Perf::DataSet::Operation::Exception' =>
                            {
                                description=>"Storage error",
                            }
                         );
    # Display stack trace when printing exceptions. Inherited by all subclasses.
    Perf::DataSet::Operation::Exception->Trace(1);

    use Perf::Utils qw();

    # Member attributes
    use constant ELEMENT    => 0;
    use constant ERROR      => 1;
    use constant LATENCY    => 2; # IO latency
    use constant DURATION   => 3; # Execution duration
    use constant TAG        => 4;

    sub new {
        my ($class,%args) = @_;
        my $self = [];
        bless $self, $class;
        $self->_init(\%args);
        my @remaining_arguments = keys %args;
        if ( @remaining_arguments ){
            confess "Unsupported arguments: " . join(", ", @remaining_arguments);
        }
        return $self;
    }

    # _init should be called in parent->children order and remove entries
    # from the hash reference passed as parameter
    sub _init {
        my $element = delete $_[1]->{element};
        confess "You must supply a Perf::DataSet::Element" if ( !$element );
        $_[0]->[ELEMENT] = $element;
        my $tag = delete $_[1]->{tag};
        $_[0]->[TAG] = $tag;
    }

    sub _as_hash {
        my ($self,$hash) = @_;
        $hash->{element} = $self->[ELEMENT]->dump();
        $hash->{error}   = $self->[ERROR];
        $hash->{latency} = $self->[LATENCY];
        $hash->{duration} = $self->[DURATION];
        $hash->{tag} = $self->[TAG];
    }

    sub execute {
        confess "You must implement this method in a subclass";
    }

    sub get_element {
        return $_[0]->[ELEMENT];
    }

    sub get_latency {
        return $_[0]->[LATENCY];
    }

    sub set_latency {
        $_[0]->[LATENCY] = $_[1];
    }

    sub set_duration {
        $_[0]->[DURATION] = $_[1];
    }

    sub get_duration {
        return $_[0]->[DURATION];
    }

    sub set_tag {
        $_[0]->[TAG] = $_[1];
    }

    sub get_tag {
        return $_[0]->[TAG];
    }


    sub get_categories {
        confess "You must implement this subroutine";
    }

    sub has_failed {
        return defined $_[0]->[ERROR];
    }

    sub set_error {
        my ($self,$error) = @_;
        confess 'You must supply an argument' if (!$error);
        confess "Operation already has an error!\n" . $self->dump(1) if ($self->[ERROR]);
        $self->[ERROR] = $error;

        if ($self->isa('Perf::DataSet::Operation::Modification')) {
            $self->[ELEMENT]->set_invalid(1);
        }

        return;
    }

    sub get_error {
        return $_[0]->[ERROR];
    }

    sub get_size {
        return 0;
    }

    sub dump {
        my ($self,$as_string) = @_;
        my $clone = {};
        $self->_as_hash($clone);
        return $clone if (!$as_string);
        return Perf::Utils::as_string($clone);
    }
}
1;
