# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

FileIO - Base class for Workload File Operations

=head1 DESCRIPTION

This class is the base class for all File Operation subclasses. It provides all the common functionality.

It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Operation::File::Block;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::Utils qw(ceil);
    use Perf::DataSet::Operation::File;
    # Declare exceptions
    # TODO: Add a block number field
    use Exception::Class (  'Perf::DataSet::Operation::FileBlockException' =>
                            {
                                isa => 'Perf::DataSet::Operation::Exception',
                            },
                         );
    use base qw(Perf::DataSet::Operation::File);

    # Member attributes
    use constant PAYLOAD   => 5;
    use constant OFFSET    => 6;
    use constant SIZE      => 7;
    use constant BNO_START => 8;
    use constant BLOCK_SIZE=> 9;

    # Aliases
    use constant FILE      => Perf::DataSet::Operation::ELEMENT;
    use constant ERROR     => Perf::DataSet::Operation::ERROR;

    sub _init {
        my ($self,$args) = @_;
        $self->SUPER::_init($args);
        my $offset = delete $args->{offset};
        confess "You must specify an 'offset' argument" if (! defined $offset);
        my $file = $self->[FILE];
        my $file_blocksize = $file->get_block_size;
        confess "Offset not at beginning of a block" if ($offset % $file_blocksize);
        my $file_offset = $file->get_offset;
        confess "Offset $offset smaller than $file_offset" if ($offset < $file_offset);
        $self->[OFFSET]    = $offset;
        $self->[SIZE]      = delete $args->{size};
        # Cache that data for later
        $self->[BNO_START] = $offset/$file_blocksize;
        $self->[BLOCK_SIZE] = $file_blocksize;
    }

    sub _as_hash {
        my ($self,$hash) = @_;
        $self->SUPER::_as_hash($hash);
        $hash->{offset}    = $self->[OFFSET];
        $hash->{size}      = $self->[SIZE];
        $hash->{payload}   = $self->[PAYLOAD];
        $hash->{bno_start} = $self->[BNO_START];
        $hash->{block_size}= $self->[BLOCK_SIZE];
    }

    sub set_payload {
        $_[0]->[PAYLOAD] = $_[1];
    }

    sub get_payload {
        return $_[0]->[PAYLOAD];
    }

    sub get_offset {
        return $_[0]->[OFFSET];
    }

    sub set_offset {
        $_[0]->[OFFSET] = $_[1];
    }

    sub get_size {
        return $_[0]->[SIZE];
    }

    sub get_bno_start {
        return $_[0]->[BNO_START];
    }

    # Override Operation::set_error to flag invalid blocks in this element
    sub set_error {
        my ($self,$error) = @_;

        confess 'You must supply an argument' if (!$error);
        confess "Operation already has an error!\n" . $self->dump(1) if ($self->[ERROR]);

        $self->[ERROR] = $error;

        if ($self->isa('Perf::DataSet::Operation::File::Modification')) {
            $self->[FILE]->set_invalid_blocks(start => $self->[BNO_START],
                                              num   => ceil($self->[SIZE] / $self->[BLOCK_SIZE]));
        }

        return;
    }
}
1;
