# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

File::Create - Create Operation

=head1 DESCRIPTION

This class is the base class for all File Operation subclasses. It provides all the common functionality.

It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Operation::File::Create;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    use Time::HiRes qw( time );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use FileHandle;
    use IO::Seekable;
    use Perf::DataSet::Operation;
    use Perf::DataSet::Element::File;
    use Perf::DataSet::Operation::File;
    use Perf::DataSet::Operation::File::Modification;
    use base qw(Perf::DataSet::Operation::File Perf::DataSet::Operation::File::Modification);

    # Aliases
    use constant FILE    => Perf::DataSet::Operation::ELEMENT;
    use constant LATENCY => Perf::DataSet::Operation::LATENCY;

    sub _init {
        my ($self,$args) = @_;

        # Remove parameters that are hardcoded below
        # TODO: Support non empty creates!!
        confess "Non-empty creates are currently not supported" if (delete $args->{size});
        delete $args->{size_whole};
        my $file_block_size = delete $args->{block_size};
        confess "You must specify a block_size" if (!$file_block_size);
        delete $args->{key};
        my $filepath = delete $args->{path};
        confess "You must specify a path" if (!$filepath);

        my @file_args = (   key         =>$filepath,
                            path        =>$filepath,
                            size        =>0,
                            size_whole  =>0,
                            block_size  =>$file_block_size,
                        );

        my $sparse = delete $args->{sparse};
        my $file_element;
        if ($sparse){
            # We need to pass %$args to get the sparse_* required fields
            $file_element = Perf::DataSet::Element::File::Sparse->new( @file_args, %$args);
        }
        else{
            $file_element = Perf::DataSet::Element::File->new( @file_args);
        }
        my $tag = delete $args->{tag};
        # Remove everything from $args now that we have create the element
        # and replace with the only argument the Operation supports
        for (keys %$args){ delete $args->{$_} };
        $args->{element} = $file_element;
        $args->{tag} = $tag;
        $self->SUPER::_init($args);
    }

    sub execute {
        my ($self,$sequence) = @_;
        my $file = $self->[FILE];
        confess "File '". $file->get_key  ."' is invalid for this Operation"
            if ($file->was_created && !$file->is_deleted);
        $sequence->get_container->set_element($file);
        my $start_time = time;
        eval {
            # That should create the file
            $sequence->get_fh($file);
        };
        my $err = $@;
        my $e = Perf::DataSet::Operation::Exception->caught();
        $self->[LATENCY] = time - $start_time;
        if ($err){
            $e->rethrow if ($e);
            die $err;
        }
        return;
    }

    sub get_categories {
        return ('create');
    }
 }
1;