# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  
## @status   private
##
## @pod here

=head1 NAME

File::Read - File Read operation

=head1 DESCRIPTION

#TODO

It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Operation::File::Read;
{
    use strict;
    use warnings;
    use Carp qw( confess );    
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use FileHandle;
    use Fcntl qw(:DEFAULT :seek);
    use Time::HiRes qw( time );
    
    use Perf::DataSet::Operation::File::Aligned qw(get_aligned_read_buffer);
    use Perf::DataSet::Operation::File::Block;

    # Declare exceptions
    use Exception::Class (  'Perf::DataSet::Operation::FileReadException' =>  
                            {
                                isa => 'Perf::DataSet::Operation::FileBlockException',
                            },
                         );      
    use base qw(Perf::DataSet::Operation::File::Block);                         
    #
    # Member attributes
    #
    use constant VERIFY         => 10;
    use constant CORRUPTIONS_OK => 11;
    use constant CORRUPTIONS    => 12;
    # TODO: The two fields above should not be properties of the Op
    # but rather of the Workload itself. The Op should simply throw an exception
    # rather than try to handle the corruption itself

    # Aliases
    use constant FILE           => Perf::DataSet::Operation::ELEMENT;
    use constant LATENCY        => Perf::DataSet::Operation::LATENCY;
    use constant SIZE           => Perf::DataSet::Operation::File::Block::SIZE;
    use constant PAYLOAD        => Perf::DataSet::Operation::File::Block::PAYLOAD;
    use constant OFFSET         => Perf::DataSet::Operation::File::Block::OFFSET;
    use constant BNO_START      => Perf::DataSet::Operation::File::Block::BNO_START;    
    use constant BLOCK_SIZE     => Perf::DataSet::Operation::File::Block::BLOCK_SIZE;    
    
    sub _init {
        my ($self,$args) = @_;
        $self->SUPER::_init($args);
        $self->[VERIFY]         = delete $args->{verify} || 'checksum';
        $self->[CORRUPTIONS_OK] = delete $args->{corruptions_ok};
    }
    
    sub _as_hash {
        my ($self,$hash) = @_;
        $self->SUPER::_as_hash($hash);
        $hash->{verify}         = $self->[VERIFY];
        $hash->{corruptions_ok} = $self->[CORRUPTIONS_OK];
        $hash->{corruptions}    = $self->[CORRUPTIONS];
    }
    
    sub get_categories {
        return ('read');
    }
    
    sub execute {
        my ($self,$sequence) = @_;
        my $size   = $self->[SIZE];
        my $offset = $self->[OFFSET];
        my $file   = $self->[FILE];
        my $fh = $sequence->get_fh($file);

        #
        # Seek
        #
        my $res = $fh->sysseek($offset, SEEK_SET);
        if (!defined($res)) {
            my $filesize = $file->get_size;
            my $filepath = $file->get_path;
            confess "Seek to ($offset/$filesize) failed for '$filepath': $!\n";
        }
        #
        # Aligned buffers are required for files opened with O_DIRECT
        #
        my $align_buffer_info = get_aligned_read_buffer($size);     
        #
        # Read
        #
        my $start_time = time;
        $res = $fh->sysread(${$align_buffer_info->[0]},$size,$align_buffer_info->[1]);
        my $err = $!;
        $self->[LATENCY] = time - $start_time;
        if (!defined($res) || $res != $size ) {
            my $filepath = $file->get_path;
            my $error = "Read failed for '$filepath' at offset $offset ($res bytes instead of $size) after " . $self->[LATENCY] ." seconds: $err\n";
            if ($err =~ /Input\/output error/){
                # This may indicate that the filer paniced. Give the caller the means
                # to catch the condition and decide to go ahead or not
                Perf::DataSet::Operation::FileReadException->throw(error=>$error);
            }            
            confess $error;
        }
        #
        # Validate
        #
        my $buffer_content = substr(${$align_buffer_info->[0]}, $align_buffer_info->[1], $size);        
        my $block_size = $self->[BLOCK_SIZE];
        # Split the data into chunks of length block_size
        my @blocks_data = unpack("(a$block_size)*", $buffer_content);        
        my $block_num = $self->[BNO_START];
        for my $block_data (@blocks_data) {
            my $s = $file->validate_block_data({block_num=>$block_num,
                                                block_data=>\$block_data,
                                                verify=>$self->[VERIFY],
                                                corruptions_ok=>$self->[CORRUPTIONS_OK],
                                                });
            if (!$s){            
                $self->[CORRUPTIONS] ++;
            }
            $block_num ++;
        }
        return;
    }
    
}
1;    