# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  
## @status   private
##
## @pod here

=head1 NAME

File::Write - File write operation

=head1 DESCRIPTION

#TODO

It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Operation::File::Write;
{
    use strict;
    use warnings;
    use Carp qw( confess );    
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use FileHandle;
    use Fcntl qw(:DEFAULT :seek);
    use Data::Dumper;
    use List::Util qw(sum);
    use Time::HiRes qw( time );

    use Perf::DataSet::Operation::File::Aligned qw(get_aligned_write_buffer);
    use Perf::DataSet::Operation;
    use Perf::DataSet::Operation::File::Block;
    use Perf::DataSet::Operation::File::Modification;
    
    # Declare exceptions
    use Exception::Class (  'Perf::DataSet::Operation::FileWriteException' =>  
                            {
                                isa => 'Perf::DataSet::Operation::FileBlockException',
                            },
                         );   
    use base qw(Perf::DataSet::Operation::File::Block Perf::DataSet::Operation::File::Modification);    
    
    #
    # Member attributes
    #
    use constant AMOUNT_ADDED   => 10;
    use constant AMOUNT_UPDATED => 11;
    # Aliases
    use constant FILE              => Perf::DataSet::Operation::ELEMENT;
    use constant LATENCY           => Perf::DataSet::Operation::LATENCY;
    use constant SIZE              => Perf::DataSet::Operation::File::Block::SIZE;
    use constant PAYLOAD           => Perf::DataSet::Operation::File::Block::PAYLOAD;
    use constant OFFSET            => Perf::DataSet::Operation::File::Block::OFFSET;
    use constant BNO_START         => Perf::DataSet::Operation::File::Block::BNO_START;
    use constant BLOCK_SIZE        => Perf::DataSet::Operation::File::Block::BLOCK_SIZE;

    use Perf::DataSet::Operation::File::Append;
    use Perf::DataSet::Operation::File::Overwrite;
    
    sub get_amount_added {
        return $_[0]->[AMOUNT_ADDED];
    }
    
    sub get_amount_updated {
        return $_[0]->[AMOUNT_UPDATED];
    }
    
    sub get_categories {
        if ($_[0]->[AMOUNT_UPDATED]){
            return ('write','overwrite');
        }
        return ('write');
    }
    
    sub _as_hash {
        my ($self,$hash) = @_;
        $self->SUPER::_as_hash($hash);
        $hash->{amount_added}   = $self->[AMOUNT_ADDED];
        $hash->{amount_updated} = $self->[AMOUNT_UPDATED];
    }

    sub new {
        my $class = shift;
        return $class->Perf::DataSet::Operation::File::Block::new(@_) if ($class ne __PACKAGE__);
        my %args = @_;
        # Figure out if we should instantiate an overwrite or an append
        my $file = $args{element} || confess "You must provide an 'element' argument";
        my $io_size = $args{size} || confess "You must provide a 'size' argument";
        my $offset = $args{offset};        
        my $file_slice_size = $file->get_size;
        my $file_offset = $file->get_offset;
        if ($offset >= $file_offset){
            if( $offset + $io_size <= $file_offset + $file_slice_size ){
                return Perf::DataSet::Operation::File::Overwrite->new(%args);            
            }
            my $file_whole_size = $file->get_size_whole;        
            if ($offset >= $file_whole_size ){
                return Perf::DataSet::Operation::File::Append->new(%args);
            }
        }
        confess "Operation not supported yet. Operation:\n" . Dumper(\%args) . 
                "\nFile:\n" . Dumper($file->dump);
    }
    
    sub execute {
        my ($self,$sequence) = @_;
        my $size   = $self->[SIZE];
        my $offset = $self->[OFFSET];
        my $file   = $self->[FILE];
        my $block_size = $self->[BLOCK_SIZE];

        #
        # Open the file
        #
        my $fh = $sequence->get_fh($file);

        #
        # Use aligned buffer for O_DIRECT
        #
        my $align_buffer_info = get_aligned_write_buffer($size);          
        
        #
        # Generate data for overwrite
        #
        my $block_num = $self->[BNO_START];
        my $left = $size;
        my @blocks_data;
        while ($left){
            my $curr_block_size = ($block_size <= $left )?$block_size:$left;
            my $bl = $file->generate_block( 
                                            buffer   => $align_buffer_info->[0],
                                            offset   => $size - $left + $align_buffer_info->[1],
                                            block_num=> $block_num,
                                            size     => $curr_block_size,
                                            );
            push @blocks_data, $bl;                                
            $left -= $curr_block_size;
            $block_num++;
        }   
        $self->[PAYLOAD] = \@blocks_data; 
       
        #
        # Seek
        #
        my $res = $fh->sysseek($offset, SEEK_SET);
        if (!defined($res)) {
            my $filesize = $file->get_size;
            my $filepath = $file->get_path;
            confess "Seek to ($offset/$filesize) failed for '$filepath': $!\n";
        }
        #
        # Perform overwrite
        #
        my $start_time = time;
        $res = $fh->syswrite(${$align_buffer_info->[0]},$size,$align_buffer_info->[1]);
        my $err = $!;
        $self->[LATENCY] = time - $start_time;
        if (!defined($res) || $res != $size ) {
            $res = defined($res)?$res:'undef';
            my $filepath = $file->get_path;
            my $error = "Write failed for '$filepath' at offset $offset ($res bytes instead of $size) after " . $self->[LATENCY] ." seconds: $err\n";
            if ($err =~ /Input\/output error/){
                # This may indicate that the filer paniced. Give the caller the means
                # to catch the condition and decide to go ahead or not
                Perf::DataSet::Operation::FileWriteException->throw(error=>$error);
            }
            confess $error;
        }
        $self->process;
        # No need to keep this in memory any longer
        delete $self->[PAYLOAD];        
    }
    
    sub process {
        my ($self) = @_;
        my $file = $self->[FILE];
        my $block_num = $self->[BNO_START];
        for my $block_data (@{$self->[PAYLOAD]}){
            my $ret = $file->record_block($block_num,\$block_data);        
            my $actual_block_size = length($block_data);
            if ($ret == $Perf::DataSet::BLOCK_UPDATED){
                $self->[AMOUNT_UPDATED] += $actual_block_size;
            }
            elsif( $ret == $Perf::DataSet::BLOCK_ADDED ){
                $self->[AMOUNT_ADDED] += $actual_block_size;
            }
            else{
                confess "Invalid return value $ret";
            }            
            $block_num ++;
        }
    }
    
}
1;
    