# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

package Perf::DataSet::Purge;

use strict;
use warnings;
use Carp qw(confess croak carp);
use Tharn qw($Log);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Module::Common qw(validate_mixed);
use Perf::Utils qw(min max get_random_element as_string in_bytes);
use List::MoreUtils qw(any uniq);
use Perf::DataSet::Command;

use base qw(Perf::DataSet::Command);

sub validate {
    my $class = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {
            description => {type => SCALAR | UNDEF, optional => 1, default => "purge"},
            force => {type => SCALAR | UNDEF, optional => 1, default => 0},
            users_ok => {type => SCALAR | UNDEF, optional => 1, default => 0},
           }
        );
}


sub new {
    my $class = shift;
    my %params = $class->validate(@_);
    my $self = \%params;
    bless $self, $class;
    return $self;
}


sub pre_execute($$){
    my $self = shift;
    my $owner = shift;
    if (!$owner || !ref($owner)){
        confess "You must specify an owner";
    }

    if ($owner->isa('Perf::DataSet::Worker')) {
        # Get rid of all the metadata kept by this Worker about the dataset
        # and replace it with a brand new Container object
        $owner->{container} = Perf::DataSet::Container->new;
    }
    elsif ($owner->isa('Perf::DataSet')) {
        if ($owner->is_read_only()) {
            if (!$self->{force}) {
                confess "You cannot purge read-only Perf::DataSet '" . $self->get_uid . "'";
            }
        }
        my $dataset_uid = $owner->get_uid();
        my $dataset =
          Perf::Depot->get_resource(class => 'DataSet',
                                       name  => $dataset_uid);

        #
        # Delete actual data
        #
        $owner->mount(force_rw => 1);

        if ($owner->is_nas()) {
            my $rootdir = $dataset->get_rootdir_resource();

            confess("Couldn't find RootDir object") if (!defined($rootdir));

            my @rootdir_users = grep { $_ != $dataset } @{$rootdir->get_users(class => 'DataSet')};

            if (0 < scalar(@rootdir_users)) {
                my $str =
                    'Purge not allowed. '
                  . $rootdir->get_description()
                  . " is being used by:\n"
                  . join("\n", map { $_->get_description() } @rootdir_users);

                if ($self->{users_ok}) {
                    $Log->comment($str);
                    return;
                }
                else {
                    confess($str);
                }
            }

            $rootdir->delete_root($dataset_uid);
            Perf::Depot->eject_resource($rootdir);
        }
        else { # is SAN
            foreach my $lun (@{$dataset->get_lun_resources()}) {
                my $rscs =
                  Perf::Depot->get_resources(
                        class => 'SANDevice',
                        lun   => $lun,);

                foreach my $rsc (@$rscs, $lun) {
                    Perf::Depot->eject_resource($rsc);

                    # LUNs get destroyed automatically once raw
                    # devices are destroyed
                    if ($rsc != $lun ) {
                        Perf::Depot->remove_resource($rsc);
                    }
                }
            }
        }

        #
        # Delete metadata
        #
        $owner->{container} = Perf::DataSet::Container->new;
    }
    else{
        confess "The owner must be a Perf::DataSet or Perf::DataSet::Worker and not a " . ref($owner);
    }
}

sub execute {
    # Nothing to do
    return;
}

sub post_execute($$$){
    my $self = shift;
    my $owner = shift; # Perf::DataSet object
    my $returns = shift;

    if ($owner->isa('Perf::DataSet::Worker')){
        return $returns;
    }
    elsif ($owner->isa('Perf::DataSet')){
        $owner->clear_addition_info();

        $owner->unmount(users_ok => 1);
    }
}

1;
