# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Sequence - Base class for Workload sequences

=head1 DESCRIPTION

This class is the base class for all Sequence subclasses. It provides all the common functionality.

It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Recorder;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::DataSet::Operation;
    use Object::InsideOut qw(Storable);

    #
    # Member attributes
    #
    my @totals              :Field
                            :Name('totals');
    my @operations_list     :Field
                            :Name('operations_list');
    my @tag_totals          :Field
                            :Name('tag_totals');

    sub _init :Init {
        my ($self) = @_;
        $Log->comment("Instantiating a Recorder of class '" . ref($self) . "' with id '$$self'");
        $totals[$$self]          = {};
        $operations_list[$$self] = [];
        $tag_totals[$$self]     = {};
    }

    sub record {
        my ($self,$operation,$duration) = @_;
        $operation->[ Perf::DataSet::Operation::DURATION  ] = $duration;
        # TODO: Should there be a limit?
        push @{$operations_list[$$self]}, $operation;
        my $aggregate = $totals[$$self];
        $aggregate->{count} ++;
        $aggregate->{duration} += $duration;
        $aggregate->{amount} += $operation->get_size;

        my $current_tag = $operation->get_tag;
        if($current_tag){
            my $tags = $tag_totals[$$self]->{$current_tag};
            if(!defined($tags)){
                $tags ={};
                $tag_totals[$$self]->{$current_tag} = $tags;
            }
            $tags->{count} ++;
            $tags->{duration} += $duration;
            $tags->{amount} += $operation->get_size;
        }
    }

    sub get_amount {
        my ($self) = @_;
        return $totals[$$self]->{amount} || 0;
    }

    sub get_duration {
        my ($self) = @_;
        return $totals[$$self]->{duration} || 0;
    }

    sub get_count {
        my ($self) = @_;
        return $totals[$$self]->{count} || 0;
    }

    # For backwards compatibility
    sub get_report {
        my ($self) = @_;
        my $start_time = time;
        my $report = {common => {}, individual=>{}};
        my $aggregate_records = {};
        my $element_records   = {};
        for my $operation (@{$operations_list[$$self]}){
            my $element_key = $operation->get_element->get_key;
            my $operation_has_failed = $operation->has_failed;
            my $size = $operation->get_size;
            my $duration = $operation->get_duration;
            for my $category ($operation->get_categories){
                my $element_record = $element_records->{$element_key}->{$category};
                if (!defined($element_record)){
                    $element_record = {failed=>{}};
                    $element_records->{$element_key}->{$category} = $element_record;
                }
                my $aggregate_record = $aggregate_records->{$category};
                if (!defined($aggregate_record)){
                    $aggregate_record = {failed=>{}};
                    $aggregate_records->{$category} = $aggregate_record;
                }
                # TODO: distinguish betweens failures!
                if ($operation_has_failed){
                    $element_record = $element_record->{failed};
                    $aggregate_record = $aggregate_record->{failed};
                }
                $element_record->{count} ++;
                $aggregate_record->{count} ++;
                $element_record->{duration} += $duration;
                $aggregate_record->{duration} += $duration;
                if ($size){
                    $element_record->{amount} += $size;
                    $aggregate_record->{amount} += $size;
                    $aggregate_record->{size}->{$size} ++;
                    $element_record->{size}->{$size} ++;
                }
            }
        }
        delete $operations_list[$$self];
        $report->{common}     = $aggregate_records;
        $report->{common}->{all} = $totals[$$self];
        $report->{common}->{tags} = $tag_totals[$$self];
        $report->{individual} = $element_records;
        my $runtime = time - $start_time;
        $Log->comment("Report computed in $runtime seconds");
        return $report;
    }
}
1;
