# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Sequence - Base class for Sequence classes

=head1 DESCRIPTION

This class is the base class for all Sequence subclasses. It provides all the common functionality.

It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Sequence;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Math::Random::MT::Auto ':!auto';
    use Data::Dumper;
    use MIME::Base64;
    use Perf::Utils qw(in_bytes);
    use Storable qw(nfreeze);
    use Perf::DataSet::Container;
    use Perf::DataSet::Recorder;

    use Object::InsideOut qw(Storable);

    # Member attributes
    my @container       :Field
                        #:Arg( Name=>'container', Mandatory => 0 )
                        :Name('container')
                        :Type('Perf::DataSet::Container')
                        :Get('get_container')
                        :Set(Name=>'set_container', Perm=>'Restrict(Perf::DataSet::ModularWorkload)');
    my @recorder        :Field
                        :Arg( Name=>'recorder', Mandatory => 0 )
                        :Type('Perf::DataSet::Recorder')
                        :Set('set_recorder')
                        :Get('get_recorder');
    my @rand_generator  :Field
                        :Arg( Name=>'rand_generator', Mandatory => 0)
                        :Type('Math::Random::MT::Auto');
    my @target_amount   :Field
                        :Arg( Name=>'target_amount', Mandatory => 0 )
                        :Get('get_target_amount')
                        :Set('set_target_amount');
    my @target_duration :Field
                        :Arg( Name=>'target_duration', Mandatory => 0 )
                        :Get('get_target_duration')
                        :Set('set_target_duration');
    my @target_percent  :Field
                        :Arg( Name=>'target_percent', Mandatory => 0 )
                        :Get('get_target_percent')
                        :Set('set_target_percent');
    my @type            :Field
                        :Name('type')
                        :Set('set_type')
                        :Get('get_type');

=head2 new

Instantiate Sequence object.

Do not override this method directly. Instead, define an :Init method in your subclass.

=over

=item Arguments

B<operations> = arrayref of strings

I<Mandatory> List of all the Operations the Sequence may return

B<target_amount> = string or number

I<Optional> Workload will stop when the sum of the sizes of all executed Operations has reached that
number. Examples: '2m', '4g',...

B<target_duration> = number

I<Optional> Workload will stop after that many seconds

B<target_percent> = number

I<Optional> Workload will stop after the sum of the sizes of all executed Operations has reached that
percentage of the total size of the Container

=back

=head2 get_container

Returns list of Elements participating in Sequence as a Perf::DataSet::Container

=head2 get_target_amount

Returns target amount for Sequence. In bytes.

=head2 get_target_duration

Returns target duration for Sequence. In seconds.

=head2 get_target_percent

Returns target percentage for Sequence. A hundred percent is equivalent to a target amount equal to
the total size of the files in the Container.

=cut

    # Prepare element for RPC network communication
    # Don't send binary data across only text
    sub rpc_serialize {
        my $self = $_[0];
        #my $tmp_recorder = $recorder[$$self];
        #$recorder[$$self] = encode_base64(nfreeze($recorder[$$self]),'');
        my $frozen = nfreeze($self);
        my $encoded = encode_base64($frozen,'');
        #$recorder[$$self] = $tmp_recorder;
        return $encoded;
    }

    sub rpc_thaw {
        my ($self) = @_;
        $Log->comment("DEBUG: rpc_thaw self $$self = " . Dumper($self));
        $recorder[$$self] = Storable::thaw(decode_base64($recorder[$$self]));
        my $recorder_package = ref($recorder[$$self]);
        eval "require $recorder_package; $recorder_package->import()";
        confess "Unable to load '$recorder_package': $@" if $@;

        $Log->comment("DEBUG: Recorder: \n" . Dumper($recorder[$$self]->dump) );

        return;
    }

=head2 rand

Returns a random integer between 0 and the input value.
You must use this method whenever you need a random number to ensure repeatibility.

=over

=item Arguments

number

I<Required> Maximum value returned

=over

=cut

    sub rand {
        my ($self,$range) = @_;
        confess "You must pass a value" if (! defined $range);
        return int(rand($range));
        #return int($rand_generator[$$self]->rand($range));
    }

=head2 random_element

Returns a random Element from the list of Elements participating in workload.

=over

=item Arguments

string

I<Optional> Indicates whether how elements should be weighted if at all during the random
selection process. Defaults to empty, only supports 'size'.

=over

=cut

    sub random_element {
        my $self = shift;
        # TODO: Use $self->get_rand to control repeatibility
        return $container[$$self]->random_element(@_);
    }

=head2 filter_elements

Remove all elements but those of a certain type (like 'files' or 'dirs') from the list of elements
to run the sequence on.

=over

=item Arguments

string

I<Required> Type of the elements to keep. Currently supported: files, dirs.

=back

=cut

    sub filter_elements {
        my ($self,$type) = @_;
        confess "You must specify a type" if (!$type);
        my $type_elements = $container[$$self]->get_elements($type);
        confess "No elements of type '$type' found" if (!$type_elements || !%$type_elements);
        my $type_elements_container = Perf::DataSet::Container->new();
        $type_elements_container->set_elements($type_elements);
        $type_elements_container->undirty;
        $container[$$self] = $type_elements_container;
    }

=head2 element_count

Returns the number of elements of a certain type or of all types

=over

=item Arguments

string

I<Optional> Type of the elements provide a count for. Currently supported: files, dirs.

=back

=cut

    sub element_count {
        my ($self, $type) = @_;
        return $container[$$self]->count($type);
    }

=head2 has_next

Returns true if the Sequence hasn't yet completed.
At this level the only criteria used are target_amount, target_percent and target_duration. If you
need a more sophisticated stopping criteria you will need to override this method in your subclass.

=cut

    sub has_next {
        my ($self) = @_;
        my $ta = $target_amount[$$self];
        if ($ta && $recorder[$$self]->get_amount >= $ta){
            return 0;
        }
        my $td = $target_duration[$$self];
        if ( $td && $recorder[$$self]->get_duration >= $td){
            return 0;
        }
        return 1;
    }

    sub next_operation {
        confess "You must implement this subroutine";
    }

    sub pre_execute{
        my ($self,$owner) = @_;
        if ($owner->isa('Perf::DataSet::Worker')){
            $self->set_container( $owner->get_container );
            if ($target_percent[$$self] && $container[$$self]->count('files')){
                # Turn percentage into amount for this Worker
                my $percentage = delete $target_percent[$$self];
                $target_amount[$$self] = int($container[$$self]->get_size('files') * $percentage / 100);
            }
        }
        elsif($owner->isa('Perf::DataSet')){
            my $amount = delete $target_amount[$$self];
            if ($amount){
                $amount = in_bytes($amount);
                my $process_count = scalar(@{$owner->{manager}->get_processes($owner->{uid})});
                $target_amount[$$self] = int($amount / $process_count);
            }
        }
        else{
            confess 0;
        }
    }

    sub post_execute {
        my ($self, $owner, $returns) = @_;
        if ($owner->isa('Perf::DataSet::Worker')){
            $owner->{container}->merge_with($container[$$self]);
            #$Log->comment("DEBUG: Container after post-execute:\n" . $owner->{container}->as_string);

            # While the Worker already contains the correct data about each Element, the Perf::DataSet
            # may not. Therefore we need to send all the elements back to the Perf::DataSet so it can
            # update its own container.
            $returns->{elements} = $container[$$self]->get_rpc_serialized_elements;
        }
        return $returns;
    }

    sub next {
        my ($self) = @_;
        return if (!$self->has_next);
        return $self->next_operation;
    }

    sub get_default_recorder {
        return Perf::DataSet::Recorder->new;
    }

=head2 get_operations

Must be implemented in user subclass. Returns an arrayref of Operation class names potentially
returned by this Sequence.

=cut

    sub get_operations {
        confess "You must implement this subroutine in your subclass";
    }

 }
1;
