# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

DataSet::Sequence::File - Base class for file-based Workload sequences

=head1 DESCRIPTION

This class is the base class for all Sequence subclasses. It provides all the common functionality.

It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::DataSet::Sequence::File;{

    use strict;
    use warnings;
    use Carp qw(confess);
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use FileHandle;
    use Fcntl qw(:DEFAULT :seek);
    use Perf::Utils qw(in_bytes);
    use Tie::Cache::LRU::LinkedList;
    use Object::InsideOut qw( Perf::DataSet::Sequence );

    my @does_modifications :Field :Name('does_modifications') :Get('does_modifications');
    my @max_open_files     :Field :Arg(Name=>'max_open_files', Mandatory=>0) :Default(1000);
    my @filehandles        :Field :Name('filehandles');

    sub pre_execute {
        my ($self,$owner) = @_;
        $self->SUPER::pre_execute($owner);
        if ($owner->isa('Perf::DataSet::Worker')){
            # Get rid of all Elements but files
            $self->filter_elements('files');
            # Don't keep more than max_open_files filehandles
            my $files_count = $self->element_count;
            my %filehandles;
            if ($files_count > $max_open_files[$$self]){
                tie %filehandles, 'Tie::Cache::LRU::LinkedList', $max_open_files[$$self];
            }
            # Create the filehandle cache as an attribute of the worker
            # and the workload. This ensures that filehandles don't get closed
            # during wait(), where it could possibly hang for a while if the
            # filer is down
            $filehandles[$$self] = \%filehandles;
            $owner->{filehandles} = $filehandles[$$self];
        }
        elsif($owner->isa('Perf::DataSet')){
            # Find out if the sequence involves File::Modification Operations
            my $modifications = grep { $_->isa('Perf::DataSet::Operation::File::Modification')}
                                          @{ $self->get_operations };
            $does_modifications[$$self] = !!$modifications;
        }
    }

    sub post_execute {
        my ($self,$owner,$returns) = @_;
        $self->SUPER::post_execute($owner,$returns);
        if ($owner->isa('Perf::DataSet::Worker')){
            delete $owner->{filehandles};
            # This should trigger file closing
            delete $filehandles[$$self];
        }
        return $returns;
    }

    sub random_file {
        my $self = shift;
        return $self->random_element(@_);
    }

    sub random_offset {
        my ($self, $file, $io_size) = @_;
        confess "Invalid file argument" if (!$file || !ref($file) || !$file->isa('Perf::DataSet::Element::File'));
        my $filesize = $file->get_size;
        if (!$io_size) {
            $io_size = $file->get_block_size;
        }
        else{
            $io_size = in_bytes($io_size);
        }
        return $file->get_offset + int( $self->rand($filesize) / $io_size) * $io_size;
    }

    sub get_fh {
        my ($self, $file) = @_;
        my $filepath = $file->get_path;
        my $fh = $filehandles[$$self]->{$filepath};
        return $fh if ($fh);
        my $flags;
        if ($file->is_deleted || !$file->was_created){
            $flags = O_CREAT | O_EXCL;
            $file->mark_created;
            $file->clear_deleted;
        }
        if ($does_modifications[$$self]){
            $flags |= O_RDWR;
        }
        else{
            $flags |= O_RDONLY;
        }
        # Enable O_DIRECT whenever possible to disable client caching
        my $type = $self->get_type;
        confess "No type set on Sequence!" if (!$type);
        if ($type !~ /^cifs/) {
            $flags |= O_DIRECT;
        }
        $fh = FileHandle->new( $filepath, $flags) or confess "Unable to open file '$filepath': $!";
        $filehandles[$$self]->{$filepath} = $fh;
        return $fh;
    }

    sub remove_fh {
        my ($self, $file) = @_;
        my $filepath = $file->get_path;
        my $fh = $filehandles[$$self]->{$filepath};
        if ($fh){
            local $SIG{ALRM} = sub {confess "Timeout during close of '$filepath'\n"};
            alarm 20;
            $fh->close or confess "Unable to close file '$filepath':$!";
            alarm 0;
            # Remove filehandle only if the close has succeeded so that close isn't
            # attempted anyway (during destruction) and possibly hangs
            delete $filehandles[$$self]->{$filepath};
        }
        return $fh;
    }

}
1;
