# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Store a sorted list of numbers in a compact manner by storing ranges when appropriate
## @status   review

package Perf::DataSet::Sparser;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Perf::Utils qw(as_string);
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use POSIX qw(ceil);

my $NONLARGE_AGGR = 0;
my $LARGE_AGGR = 1;
my $FLEX_VOL = 'flex';
my $TRAD_VOL = 'trad';
my $WAFLBLOCK = 4096;
my $MAX_CACHED = 1000;
my $MAX_POINTERS        = {$TRAD_VOL => {$NONLARGE_AGGR => 1024,
                                        },
                           $FLEX_VOL => {$NONLARGE_AGGR => 510,
                                         $LARGE_AGGR  => 255,
                                        },
                            };
my $INODE_MAX_POINTERS  = {$TRAD_VOL => {$NONLARGE_AGGR => 16,
                                        },
                           $FLEX_VOL => {$NONLARGE_AGGR => 8,
                                         $LARGE_AGGR  => 4,
                                        },
                            };
my $INODE_MAX_FILESIZE  ={$TRAD_VOL => {$NONLARGE_AGGR => 64,
                                        },
                           $FLEX_VOL => {$NONLARGE_AGGR => 64,
                                         $LARGE_AGGR  => 64,
                                        },
                            };

sub new {
    my $class = shift;
    my %params = validate(
           @_,
           {
               filesize => {type => SCALAR},
               block_size => {type => SCALAR},
               offset => {type => SCALAR, optional=>1, default=>0}, # In Bytes
               range => {type => SCALAR, optional=>1}, # In Bytes
               sparse_ratio => {type => SCALAR, optional=>1, default=>1},
               sparse_level => {type => SCALAR, optional=>1, default=>0},
               sparse_align => {type => SCALAR, optional=>1, default=>'right', regex=>qr/left|right/},
               vol_type => {type => SCALAR, optional=>1, default=>'flex'},
               large_aggr => {type => SCALAR, optional=>1, default=>0},
               debug => {type => SCALAR, optional=>1, default=>1},
            }
            );

    my $vol_type = $params{vol_type};

    # Sparser can only handle flex, striped, or trad volumes.
    if ($vol_type !~ m/^(?:flex|striped|trad)$/) {
        confess "Don't know what to do with '$vol_type' volumes";
    }

    # As far as Sparser is concerned, striped volumes are just flexvols.
    if ($vol_type eq 'striped') {
        $vol_type = 'flex';
    }

    my $large_aggr = $params{large_aggr};
    if ($large_aggr && $vol_type !~ /flex/){
        confess "Aggregates only contain flexible volumes";
    }
    if (!$params{range}){
        $params{range} = $params{filesize};
    }

    my $self = \%params;
    bless $self, $class;

    $self->{max_pointers} = $MAX_POINTERS->{$vol_type}->{$large_aggr};
    $self->{inode_max_pointers} = $INODE_MAX_POINTERS->{$vol_type}->{$large_aggr};
    $self->{inode_max_filesize} = $INODE_MAX_FILESIZE->{$vol_type}->{$large_aggr};

    if (!defined($self->{max_pointers}) || !defined($self->{inode_max_pointers}) ||
            !defined($self->{inode_max_filesize})){
         confess "Unable to initialize Sparser:\n" . as_string($self);
    }

    # Store these numbers so we don't recompute it too often
    $self->{last_fbn} = ceil($self->{filesize} / $WAFLBLOCK) - 1;
    $self->{last_block} = ceil($self->{filesize} / $self->{block_size}) - 1;
    $self->{range_last_fbn} = ceil( ($self->{range} + $self->{offset}) / $WAFLBLOCK ) - 1;
    $self->{range_last_block} = ceil( ($self->{range} + $self->{offset}) / $self->{block_size} ) - 1;

    # Compute the first $MAX_CACHED elements
    $self->{offset_block} = int($self->{offset}/$self->{block_size});
    $self->{offset_fbn} = int($self->{offset}/$WAFLBLOCK);
    $self->{cache} = $self->_get_populated_blocks();
    $self->{cache_recomputed} = 0;
    $self->{cache_sequential_hit} = 0;
    $self->{cache_sequential_miss} = 0;
    $self->{marker} = undef;
    $self->{no_more_holes} = 0;

    $self->{binary_search} = 0;
    $self->{binary_search_iterations} = 0;

    return $self;
}

sub reset {
    my $self = shift;
    $self->{no_more_holes} = 0;
    $self->{marker} = undef;
    $self->{cache} = $self->_get_populated_blocks();
}

sub get_size_on_disk {
    my $self = shift;
    my $last_block = $self->{last_block};
    my $range_last_block = $self->{range_last_block};
    my $populated_blocks = $self->_get_populated_blocks;
    my $populated_blocks_count = 0;
    my $highest_populated_block_seen = $self->{highest_populated_range_block_seen};
    OUTER: while(1){
        # Compute how many blocks are populated
        for my $e (@$populated_blocks){
            my $ref = ref($e);
            if (!$ref){
                $populated_blocks_count ++;
            }
            elsif($ref eq 'ARRAY'){
                $populated_blocks_count += ($e->[1] - $e->[0] + 1);
            }
            else{
                confess "Yuck";
            }
        }
        my $last_e = $populated_blocks->[-1];
        my $ref = ref($last_e);
        my $offset;
        if (!$ref){
            if ($last_e == $last_block){
                last;
            }
            $Log->debug("End of range not reached yet ($last_e instead of $last_block)");
            $offset = $last_e;
            $populated_blocks_count --;
        }
        elsif($ref eq 'ARRAY'){
            if ($last_e->[1] == $last_block){
                last;
            }
            $Log->debug("End of range not reached yet ($last_e->[1] instead of $last_block)");
            $offset = $last_e->[0];
            $populated_blocks_count -= ($last_e->[1] - $last_e->[0] + 1);
        }
        # We may not at the end of the file yet, loop again
        $populated_blocks = $self->_get_populated_blocks(offset=>$offset);
        if ($self->{highest_populated_range_block_seen} == $highest_populated_block_seen){
            # Even though we asked for a new set of populated blocks we didn't get further
            # We're at the end of our range
            last if ($self->{range_last_block} != $self->{last_block});
            confess "This should not happen";
        }
        $highest_populated_block_seen = $self->{highest_populated_range_block_seen};
    }
    my $size_on_disk = $populated_blocks_count * $self->{block_size};
    if ($highest_populated_block_seen == $last_block){
        # The last block of each file is always populated and may be shorter than block_size
        my $last_block_size = $self->{filesize} - $last_block*$self->{block_size};
        $size_on_disk = $size_on_disk - $self->{block_size}  + $last_block_size;
    }
    return $size_on_disk;
}

sub get_next_hole {
    my $self = shift;

    return undef if ($self->{no_more_holes});
    if ($self->{all_holes}){
        $self->{no_more_holes} = 1;
        $Log->debug("The range is all holes");
        return [$self->{offset_block},$self->{range_last_block}];
    }
    my $marker = $self->{marker};

    #
    # Find first block in hole
    #
    my $block_start;
    my $block_end;
    if (!defined($marker)){
        $block_start = $self->{offset_block};
        my $first_element = $self->{cache}->[0];
        if (_first($first_element) == $self->{offset_block}){
            $block_start = _after($first_element);
            $self->{marker} = 0;
            $marker=0;
        }
        else{
            $block_end = _before($first_element);
        }
    }
    else{
        my $marker_element = $self->{cache}->[$marker];
        if (!defined $marker_element){
            confess "No elements in cache at marker index ($marker)";
        }
        $block_start = _after($marker_element);
    }

    #
    # Find last block in hole
    #
    if (!defined($block_end)){
        my $after_marker_element = $self->{cache}->[$marker +1];
        if (!defined $after_marker_element){
            # Make sure we are not at the end of the file
            # by any chance
            if ( $block_start > $self->{range_last_block} ){
                # End of file, return undef
                if ($self->{debug}){
                    $Log->debug("No more holes in this file or range");
                }
                $self->{no_more_holes} = 1;
                return undef;
            }
            $self->{cache} = $self->_get_populated_blocks(offset=>_first($self->{cache}->[$marker]));
            $self->{cache_recomputed} ++;
            # The first element of the new cache is the last element of the old cache
            $after_marker_element = $self->{cache}->[1];
            $self->{marker} = 0;
            if (! defined($after_marker_element)){
                if ($block_start > $self->{range_last_populated_block} ){
                    $block_end = $self->{range_last_block};
                    $self->{no_more_holes}=1;
                    $Log->debug("No more holes in this file or range");
                }
                else{
                    # That's not right, if we were at the end
                    # of the file or range we would have returned already
                    confess "Unable to get after marker element after block $block_start from cache:\n" . as_string($self);
                }
            }
        }
        if (!defined($block_end)){
            $block_end = _before($after_marker_element);
        }
    }

    if ($block_start >= $self->{last_block}){
        confess "End of hole past next to last block in file!";
    }
    if ($block_start > $self->{range_last_block}){
        if ($self->{debug}){
            $Log->debug("No more holes in this file or range");
        }
        $self->{no_more_holes} = 1;
        return undef;
    }
    if ($block_end >= $self->{last_block}){
        confess "End of hole past next to last block in file!";
    }
    if ($block_end > $self->{range_last_block}){
        $block_end = $self->{range_last_block};
    }


    my $next_hole = [$block_start, $block_end];
    if (!defined($self->{marker})){
        $self->{marker} = 0;
    }
    else{
        $self->{marker} ++;
    }
    return $next_hole;
}

sub _before {
    my $element = shift;
    if (ref($element) eq 'ARRAY'){
        return $element->[0] - 1;
    }
    else{
        return $element - 1;
    }
}

sub _after {
    my $element = shift;
    if (ref($element) eq 'ARRAY'){
        return $element->[1] + 1;
    }
    else{
        return $element + 1;
    }
}

sub _first {
    my $element = shift;
    if (ref($element) eq 'ARRAY'){
        return $element->[0];
    }
    else{
        return $element;
    }
}

sub _last {
    my $element = shift;
    if (ref($element) eq 'ARRAY'){
        return $element->[1];
    }
    else{
        return $element;
    }
}

sub _get_block_num($$) {
    my $self = shift;
    my $fbn = shift;
    return int($fbn/$self->{block_size}*$WAFLBLOCK);
}

sub _get_first_block_fbn($$){
    my $self = shift;
    my $block_num = shift;
    my $fbns_per_block = $self->{block_size}/$WAFLBLOCK;
    return $block_num*$fbns_per_block;
}

sub _get_last_block_fbn($$){
    my $self = shift;
    my $block_num = shift;
    my $fbns_per_block = $self->{block_size}/$WAFLBLOCK;
    my $fbn_end = $block_num*$fbns_per_block + $fbns_per_block - 1;
    my $last_fbn = $self->{last_fbn};
    if (!defined($last_fbn)){
        confess "No last_fbn defined!!\n" .as_string($self);
    }
    if (!defined($fbn_end)){
        confess "No fbn_end defined!!\n" .as_string($self);
    }
    if ($fbn_end > $last_fbn){
        $fbn_end = $last_fbn;
    }
    return $fbn_end;
}

sub _get_fbn_range($$){
    my $self = shift;
    my $block_range = shift;
    my $ref = ref($block_range);
    if ( !$ref ){
        return [$self->_get_first_block_fbn($block_range),
                $self->_get_last_block_fbn($block_range)];
    }
    if ($ref eq 'ARRAY' && scalar @$block_range != 2){
        confess "Input must be a 2 element array";
    }
    return [$self->_get_first_block_fbn($block_range->[0]),
            $self->_get_last_block_fbn($block_range->[1])];
}

sub _get_populated_blocks {
    my $self = shift;
    my %args = @_;
    my $block_offset = $args{offset};
    if (!defined($block_offset)){
        $block_offset = $self->{offset_block};
    }
    if ($block_offset > $self->{range_last_block}){
        confess "Offset ($block_offset) is past range ($self->{range_last_block}) of this Sparser";
    }
    my $fbn_range = $self->_get_fbn_range($block_offset);
    my $fbn_offset = $fbn_range->[0];
    my $populated_fbns = $self->_get_populated_fbns(offset=>$fbn_offset);
    if (!defined($populated_fbns) || !scalar @$populated_fbns){
        if (!$args{offset}){
            if ($self->{range} == $self->{filesize}){
                confess "This should never happen";
            }
            $self->{all_holes} = 1;
        }
        $self->{range_last_populated_block} = $self->{highest_populated_range_block_seen};
        $Log->debug("No populated fbns were found between fbn $fbn_offset and end of range");
        return undef;
    }

    my $populated_blocks = [];
    for my $e_fbn (@$populated_fbns){
        my $e_block;
        #
        # First see if we are dealing with a single block or a chain
        #
        my $ref_e_fbn = ref($e_fbn);
        if($ref_e_fbn eq 'ARRAY'){
            my $begin_block = $self->_get_block_num($e_fbn->[0]);
            my $end_block = $self->_get_block_num($e_fbn->[1]);
            if ($begin_block eq $end_block){
                $e_block = $begin_block;
            }
            else{
                $e_block = [$begin_block,$end_block];
            }
        }
        else{
            $e_block = $self->_get_block_num($e_fbn);
        }

        #
        # Now see how (or if) the new blocks should be added to the list
        #
        my $last_e = $populated_blocks->[-1];
        if (! defined $last_e){
            # The list is currently empty, just add the element to it
            push @$populated_blocks, $e_block;
            next;
        }
        my $ref_last_e = ref($last_e);
        my $ref_e_block = ref($e_block);
        if(!$ref_e_block){
            # Element is a scalar
            # Compare new block num with last element in ret
            if ($ref_last_e eq 'ARRAY'){
                if ($e_block > $last_e->[1] + 1){
                    push @$populated_blocks, $e_block;
                }
                elsif($e_block == $last_e->[1] + 1){
                    $last_e->[1] = $e_block;
                }
                elsif($e_block == $last_e->[1]) {
                    # skip
                }
                else{
                    confess "Last element's end boundary ($last_e->[1]) higher than current block ($e_block)";
                }
            }
            elsif(!$ref_last_e){
                if ($e_block > $last_e + 1){
                    push @$populated_blocks, $e_block;
                }
                elsif ($e_block == $last_e + 1){
                    # Remove the last element
                    pop @$populated_blocks;
                    # Start a chain
                    push @$populated_blocks, [$last_e, $e_block];
                }
                elsif ($e_block == $last_e ){
                    # skip
                }
                else{
                    confess "Last block ($last_e) higher than current block ($e_block)";
                }
            }
            else{
                confess "Last element is neither a scalar nor an arrayref: $ref_last_e";
            }
        }
        elsif ($ref_e_block eq 'ARRAY'){
            if ($ref_last_e eq 'ARRAY'){
                if ($e_block->[0] > $last_e->[1] + 1){
                    push @$populated_blocks, $e_block;
                }
                elsif($e_block->[0] >= $last_e->[1]){
                    # Merge the two chains
                    $last_e->[1] = $e_block->[1];
                }
                else{
                    confess "Last element's end boundary ($last_e->[1]) higher than current".
                            " element start boundary ($e_block->[0])";
                }
            }
            elsif(!$ref_last_e){
                if ($e_block->[0] > $last_e + 1){
                    push @$populated_blocks, $e_block;
                }
                elsif($e_block->[0] >= $last_e){
                    $e_block->[0] = $last_e;
                    pop @$populated_blocks;
                    push @$populated_blocks, $e_block;
                }
                else{
                    confess "Last block ($last_e) higher than current element start boundary ($e_block->[0])";
                }
            }
            else{
                confess "Last element is neither a scalar nor an arrayref: $ref_last_e";
            }
        }
        else{
            confess "Current element is neither a scalar nor an arrayref: $ref_e_block";
        }
    }
    #if ($self->{debug}){
    #    $Log->comment("Populated blocks:\n" . as_string($populated_blocks));
    #}
    my $last_populated_block = $populated_blocks->[-1];
    if (ref($last_populated_block) eq 'ARRAY'){
        $last_populated_block = $last_populated_block->[1];
    }
    if ($last_populated_block > $self->{range_last_block}){
        confess "Last populated block ($last_populated_block) past range last block ($self->{range_last_block})";
    }
    if (!$self->{highest_populated_range_block_seen} ||
            $last_populated_block > $self->{highest_populated_range_block_seen}){
        $self->{highest_populated_range_block_seen} = $last_populated_block;
    }

    return $populated_blocks;
}

sub _get_populated_fbns {
    my $self = shift;
    my %args = @_;
    my $fbn_offset = $args{offset};
    if (!defined($fbn_offset)){
        $fbn_offset = $self->{offset_fbn};
    }
    if ($fbn_offset > $self->{range_last_fbn}){
        confess "fbn offset ($fbn_offset) past range ($self->{range_last_fbn}) for this Sparser";
    }
    my $filesize = $self->{filesize};
    my $last_fbn = $self->{last_fbn};
    my $debug = $self->{debug};
    my $align = $self->{sparse_align};
    my $sparse_level = $self->{sparse_level};
    my $max_pointers = $self->{max_pointers};
    my $sparse_ratio = $self->{sparse_ratio};

    #
    # Find out how many wafl blocks at each level
    #
    my $level_fbns = $self->{level_fbns};
    if (!$level_fbns){
        my $level = 0;
        $level_fbns->[$level] = ceil($filesize / $WAFLBLOCK) + 0;
        while (1){
            if ($filesize <= $self->{inode_max_filesize}){
                $level_fbns->[0] = 1; # That's the inode entry, not really a fbn
                last;
            }
            elsif ($level_fbns->[$level] <= $self->{inode_max_pointers}){
                $level_fbns->[$level + 1] = 1; # That's the inode entry, not really a fbn
                last;
            }
            else{
                $level_fbns->[$level + 1] = ceil($level_fbns->[$level]/$self->{max_pointers}) + 0;
                if($level_fbns->[$level] <= $self->{max_pointers}){
                    $level_fbns->[$level + 2] = 1; # That's the inode
                    last;
                }
            }
            $level ++;
        }
        $self->{level_fbns} = $level_fbns;
        #if ($debug){
        #    $Log->comment("FBNs at each level:\n" . as_string($level_fbns));
        #}
    }

    my $level_chains = [];
    my $populated_level_fbns = {};
    if ( $sparse_level >= @$level_fbns - 1){
        # Either there are no blocks at sparse_level or the highest level
        # blocks in the file are at at sparse_level.
        # In both cases only write the last block in the file
        push @$level_chains, $last_fbn;
        return $level_chains;
    }

    # Offset a each level
    my $level_offsets = {};
    $level_offsets->{0} = $fbn_offset;
    my $i = 0;
    while($i < 8){
        my $level_offset = $level_offsets->{$i};
        if ($level_offset){
            $level_offsets->{$i + 1} = int($level_offset / $max_pointers);
        }
        else{
            $level_offsets->{$i + 1} = 0;
            last;
        }
        $i ++;
    }
    if ($debug){
        $Log->debug("Offset at each level:\n" . as_string($level_offsets) );
    }

    # All fbns after the cutoff should be populated
    # 1 out of every $max_pointers fbns should be populated before that
    my $cutoff = int($level_fbns->[$sparse_level] * $sparse_ratio + .5);
    # Make sure the cutoff isn't past the end of the level
    if ($cutoff > $level_fbns->[$sparse_level] - 1){
        $cutoff = $level_fbns->[$sparse_level] - 1;
    }

    if($debug){
        $Log->debug("cutoff at level $sparse_level: $cutoff");
    }

    # Populate blocks at sparse_level between the offset and the cutoff
    my $marker = int($level_offsets->{$sparse_level} / $max_pointers)*$max_pointers;
    if ($align eq 'left' && $marker ne $level_offsets->{$sparse_level}){
        $marker += $max_pointers;
    }
    if ($align eq 'right'){
        $marker += $max_pointers - 1;
    }
    if ($sparse_level == 0 && $marker > $self->{range_last_fbn}){
        if ($self->{range_last_fbn} < $self->{last_fbn}){
            $Log->debug("First fbn ($marker) is outside range ($self->{range_last_fbn})");
            #$self->{all_holes} = 1;
            return undef;
        }
        else{
            return [$self->{last_fbn}];
        }
    }
    my $element_count = 0;
    if ($sparse_level == 0){
        while ($element_count < $MAX_CACHED
               && $marker < $cutoff
               && $marker <= $self->{range_last_fbn} ){
            push @$level_chains, $marker;
            $marker += $max_pointers;
            $element_count ++;
        }
    }
    else{
        while ($element_count < $MAX_CACHED && $marker < $cutoff ){
            push @$level_chains, $marker;
            $marker += $max_pointers;
            $element_count ++;
        }
    }
    # Add the non-sparse portion
    if ($element_count < $MAX_CACHED){
        my $last_fbn_at_sparse_level = $level_fbns->[$sparse_level] - 1;
        if ($sparse_level == 0){
            if ($last_fbn_at_sparse_level > $self->{range_last_fbn}){
                $last_fbn_at_sparse_level = $self->{range_last_fbn};
            }
        }
        if ($cutoff < $level_offsets->{$sparse_level}){
            $cutoff = $level_offsets->{$sparse_level};
        }
        if ($cutoff < $last_fbn_at_sparse_level){
            push @$level_chains, [$cutoff,$last_fbn_at_sparse_level];
        }
        elsif($cutoff eq $last_fbn_at_sparse_level){
            push @$level_chains,$last_fbn_at_sparse_level;
        }
        else{
            # cutoff is past the end of the level. Skip.
        }
    }
    $populated_level_fbns->{$sparse_level} =  $level_chains;

    my $level = $sparse_level;
    while ($level > 0){
        my $chains = $populated_level_fbns->{$level};
        #if ($debug){
        #    $Log->comment("populated fbns at level $level:\n" . as_string($chains));
        #}
        $level_chains = [];
        my $element_count = 0;
        CHAIN: for my $chain (@$chains){
            my $tmp = $chain;
            if (!ref($chain)){
                # $chain is actually a single fbn
                $tmp = [$chain, $chain];
            }
            my $parent_fbn = $tmp->[0];
            while ($parent_fbn <= $tmp->[1]){
                if($align eq 'right'){
                    my $right_fbn = $parent_fbn*$max_pointers+$max_pointers-1;
                    if ($right_fbn > $level_fbns->[$level - 1] -1){
                        $right_fbn = $level_fbns->[$level - 1] -1;
                    }
                    if ($level == 1){
                        if ($right_fbn > $self->{range_last_fbn}){
                            last CHAIN;
                        }
                        if ($right_fbn < $self->{offset_fbn}){
                            next CHAIN;
                        }
                    }
                    push @$level_chains,$right_fbn;
                }
                elsif($align eq 'left'){
                    my $left_fbn = $parent_fbn*$max_pointers;;
                    if ($level == 1){
                        if( $left_fbn > $self->{range_last_fbn}){
                            last CHAIN;
                        }
                        if ($left_fbn < $self->{offset_fbn}){
                            next CHAIN;
                        }
                    }
                    push @$level_chains,$left_fbn;
                }
                else{
                    confess 0;
                }
                $parent_fbn ++;
                $element_count ++;
                if ($MAX_CACHED && $element_count == $MAX_CACHED){
                    if ($debug){
                        $Log->debug("Reached max element for level " . ($level - 1));
                    }
                    last CHAIN;
                }
            }
        }
        # We don't need to keep the data for the current level
        # and it's using RAM
        delete $populated_level_fbns->{$level};
        $populated_level_fbns->{$level - 1} = $level_chains;
        $level --;
    }

    my $last_populated_fbn = $populated_level_fbns->{0}->[-1];
    if (ref($last_populated_fbn) eq 'ARRAY'){
        $last_populated_fbn = $last_populated_fbn->[1];
    }
    if ($last_populated_fbn > $self->{range_last_fbn}){
        confess "Last populated fbn ($last_populated_fbn) past range end ($self->{range_last_fbn})";
    }
    elsif($last_populated_fbn < $self->{range_last_fbn} && @{$populated_level_fbns->{0}} < $MAX_CACHED){
        $self->{range_last_populated_fbn} = $last_populated_fbn;
        $self->{range_last_populated_block} = int($self->{range_last_populated_fbn}*$WAFLBLOCK/$self->{block_size});
    }

    #if ($debug){
    #    $Log->comment("populated fbns at level 0:\n" . as_string($populated_level_fbns->{0}));
    #}

    return $populated_level_fbns->{0};

}
