# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

package Perf::DataSet::Worker;

use strict;
use warnings;
use Tharn qw($Log param);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Cwd;
use Carp qw(confess longmess);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Module::Common qw(validate_mixed);
use Perf::Utils qw(min max get_random_element as_string in_bytes deep_copy);
use Perf::DataSet::Container;
use Storable qw();
$Storable::Eval = 1;
use File::Path qw(mkpath);
use MIME::Base64;
use Sys::Hostname qw();

sub new {
    my $class = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {
                uid   => {type => SCALAR },
                process_uid => {type => SCALAR },
                seed => {type => SCALAR },
                debug => {type => SCALAR , optional => 1, default => 0},
           }
           );

    my $self = \%params;
    $self->{pwd} = &Cwd::cwd();
    $self->{pid} = $$;
    #$self->{additions} = {};
    $self->{container} = Perf::DataSet::Container->new;

    # Seed the random number generator
    srand $self->{seed};

    bless $self, $class;
    $Log->comment("Created new " . __PACKAGE__ . ":\n" . as_string($self,2));
    return $self;
}

sub fork {
    my $self = shift;
    my %params;
    my $fork;

    %params = validate_mixed(
           @_,
           [ ],
           {
                uid   => {type => SCALAR },
                debug => {type => SCALAR , optional => 1, default => 1},
           }
           );
    $fork = \%params;

    # Loop through all additions and call copy on each.
    # TODO: what a mess, don't store all the Additions
    # in the first place
    #for my $addition ( keys %{$self->{additions}} ){
    #    my $new_addition = $self->{additions}->{$addition}->copy();
    #    $fork->{additions}->{$addition} = $new_addition;
    #}

    $fork->{pid} = $$;

    # 'Copy' data from parent
    # Mostly shallow copies except for the dataset itself (and additions)
    #
    $fork->{pwd} = $self->{pwd};
    $fork->{debug} = $self->{debug};
    $fork->{process_uid} = $self->{process_uid};
    $fork->{event_channel_port} = $self->{event_channel_port};
    $fork->{event_channel} = $self->{event_channel};
    $fork->{seed} = $self->{seed};

    # We don't want a pointer straight to the parent dataset, or we could end up with
    # collisions between the parent's dataset and our own
    $fork->{container} = $self->{container}->clone;

    bless $fork, ref($self);
    $Log->comment("Worker $self->{uid} created fork Worker:\n" . as_string($fork,2));

    return $fork;
}

sub freeze {
    my ($self,%args) = @_;
    my $filedir = $args{path} || "/tmp/Perf/DataSet/Worker/$self->{uid}";
    my $freeze = {};
    #$freeze->{additions} = $self->{additions};
    $freeze->{pwd} = $self->{pwd};
    $freeze->{debug} = $self->{debug};
    $freeze->{container} = $self->{container}->clone();

    $Log->comment("Starting freezing of Worker $self->{uid}");

    my $hostname = Sys::Hostname::hostname;
    $filedir .= "/$hostname";

    $Log->comment("Creating '$filedir'");
    eval {mkpath($filedir)};
    if ($@){
        confess  "Unable to create directory '$filedir':$@";
    }

    my $filepath = "$filedir/$self->{process_uid}";
    my $ret;
    $Log->comment("Freezing worker to '$filepath'");
    eval {$ret = Storable::nstore $freeze, $filepath};
    if (!$ret || $@){
        confess "Unable to freeze Worker in process '$self->{process_uid}' for Perf::DataSet '$self->{uid}': $@";
    }
    return $filepath;
}

sub thaw {
    my ($self, $filepath, %args) = @_;
    if ( ! -f $filepath){
        confess "File '$filepath' does not exist";
    }
    $Log->comment("Thawing worker data from '$filepath'");
    my $thaw;
    eval{$thaw = Storable::retrieve($filepath)};
    if (!defined($thaw) || $@){
        confess "Unable to load file '$filepath': $@";
    }
    unlink $filepath unless ($args{no_unlink});
    map {$self->{$_} = $thaw->{$_}} keys %$thaw;
}

sub execute($$) {
    my $self = shift;
    my $command_frozen_encoded = shift;
    my $returns;
    if (!$command_frozen_encoded){
        confess "You must pass an argument";
    }
    # Decode
    my $command_frozen = decode_base64($command_frozen_encoded);
    # Unfreeze
    my $command = Storable::thaw($command_frozen);

    # Check that the command is a subclass of Perf::DataSet::Command
    if (!$command->isa('Perf::DataSet::Command')) {
        confess("Argument must be a Perf::DataSet::Command, got a (" . ref($command) . ")");
    }

    # DEBUG: Profiling
    my $profile_enabled=0;
    if ($Devel::NYTProf::VERSION && param("DATASET_WORKER_PROFILE")){
        $Log->comment("Enabling NYTProf profiling");
        $profile_enabled ++;
        DB::enable_profile("nytprof.$$." . $command->{uid});
    }

    # Check that the command is really addressed to us
    if ($command->{dataset_uid} != $self->{uid}){
        confess "Command wasn't meant for this worker!\n" .
                "- Command dataset_uid: $command->{dataset_uid}\n" .
                "- Worker uid: $self->{uid}";
    }

    $Log->comment("Starting a '" . $command->description() ."':\n" . as_string($command,3));

    #
    # To enhance repeatibility between Commands, create a new seed here to be applied after
    # the execute has completed
    #
    my $new_seed = int(rand(3000000000));

    # In case we received events after the last command stopped
    # listening for them
    #Perf::DataSet::Event->empty($self->{event_channel});

    eval{
        #
        # Pre Execution
        #
        $command->pre_execute($self);

        #
        # Execution
        #
        my $start_time = time();
        $returns = $command->execute();
        my $end_time = time();
        my $execution_time = $end_time - $start_time;
        $start_time = $end_time;
        $Log->comment("Command '" . $command->description ."' completed in (${execution_time}s)\n");
        #
        # Post execution
        #
        $command->post_execute($self, $returns);
        $end_time = time();
        $execution_time = $end_time - $start_time;
        $Log->comment("Command '" . $command->description ."' post-executed in (${execution_time}s)\n");

        # End of Eval
    };
    my $error = $@;

    # Exit out of mountpoint so unmount can succeed
    $Log->comment("Changing to directory '$self->{pwd}'");
    chdir $self->{pwd} or confess "Unable to chdir to '$self->{pwd}': $!";

    $Log->debug("Returns:\n" . as_string($returns));

    # Apply new seed. This ensures the next Command is repeatable
    srand $new_seed;

    if ($error){
        $Log->comment("Exiting with error:\n" . $error);
        die $error;
    }

    $Log->comment("Exiting Worker::execute");

    if ($profile_enabled){
        $Log->comment("Finishing NYTProf profiling");
        DB::finish_profile();
    }

    return $returns;
}
sub get_container {
    my $self = shift;
    return $self->{container};
}
sub get_process_uid {
    my $self = shift;
    return $self->{process_uid};
}

sub DESTROY {
    my $self = shift;
    #
    # Skip the destruction for subtests
    #
    return unless ($$ == $self->{pid});

    #
    # It is possible that we error out during destroy itself. This will cause
    # a second call to destroy and all kinds of bad things happen then
    #
    if ($self->{being_destroyed}){
        $Log->comment("$self->{uid} " . "Destroy is already taking place, skipping. Current error: $@");
        return;
    }
    else{
        $self->{being_destroyed} = 1;
    }
    #
    # DESTROY may be called outside of any errors. If that's the case
    # don't do the shutdown (unless we are the only Perf::DataSet running)
    #
    my $current_error = $@;
    if ($current_error){
        $self->{on_error} = 1;
    }
    $Log->comment("$self->{uid} " . "DESTROY sequence initiated on client '".$self->{uid}
                ."' because of following:\n$current_error");
    if ($@){
        $Log->comment("$self->{uid} " . "About to override error:\n'$@'\nwith:\n'$current_error'");
    }
    $@ = $current_error;
}

sub relocate_element {
    my $self = shift();
    my %params = validate_mixed(
        @_,
        [ ],
        {
            old => {type => SCALAR },
            new => {type => SCALAR },
        }
    );

    $self->{container}->relocate_element($params{old}, $params{new});
}
1;
