# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

package Perf::DataSet::Workload;

use strict;
use warnings;
use Carp qw(confess);
use Tharn qw($Log);
BEGIN{
    my @c = caller(2);
    $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Module::Common qw(validate_mixed);
use Perf::Utils qw(min max ceil get_random_element as_string in_bytes);
use FileHandle;
use IO::Seekable;
use Data::Dumper qw(Dumper);
use Fcntl ':flock';
use Time::HiRes qw(time);
use Perf::DataSet::Command;
use Perf::DataSet::Container;
use Perf::Depot;
use Storable qw(thaw);
use MIME::Base64;
use Tie::Cache::LRU::LinkedList;

# Workload is a subclass of Command
use base qw(Perf::DataSet::Command);

# Most OSes don't let you open too many files at once
# Set the maximum of open files to 1000
our $MAX_OPEN_FILES = 1000;

sub validate {
    my $class = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {
            #
            # Mandatory parameters
            #
            random => {type => SCALAR },
            read_pct => {type => SCALAR},
            io_size => {type => SCALAR },
            #
            # Mandatory but filled up by each client before execution
            #
            # dataset: Perf::DataSet::Container object
            container => {type => OBJECT | UNDEF, optional => 1},
            #
            # Optional parameters
            #
            description => {type => SCALAR | UNDEF, optional => 1, default => "workload"},
            # type: Supports 'nfs'
            type => {type => SCALAR | UNDEF, optional => 1, default => 'nfs'},
            # file_size_min: excludes smaller files
            file_size_min => {type => SCALAR | UNDEF, optional => 1, default => undef},
            # io_size_min: not supported for type='filersio'
            io_size_min => {type => SCALAR | UNDEF, optional => 1, default => undef},
            # duration: in seconds
            duration => {type => SCALAR | UNDEF, optional => 1, default => undef},
            duration_min_per_file => {type => SCALAR | UNDEF, optional => 1, default => 5},
            # amount: amount to read/write
            amount => {type => SCALAR | UNDEF, optional => 1, default => undef},
            amount_min_per_file => {type => SCALAR | UNDEF, optional => 1, default => undef},
            # percentage: percentage of eligible files to read/write
            percentage => {type => SCALAR | UNDEF, optional => 1, default => undef},
            # verify: Supports 'pattern','checksum','auto','vs_copy'. Only applied for 'nfs' reads.
            verify => {type => SCALAR | UNDEF, optional => 1, default => 'checksum'},
            # verify_count: Instead of stopping the workload at the first error, count errors.
            verify_count => {type => SCALAR | UNDEF, optional => 1, default => 0},
            # debug: Enable debug logging
            debug => {type => SCALAR | UNDEF, optional => 1, default => 1},
            # async: For asynchronous loads
            async => {type => SCALAR | UNDEF, optional => 1, default => 0},
            # start_offset: you can define an offset for all files in the dataset
            start_offset => {type => SCALAR | UNDEF, optional => 1, default => undef},
            # delete_first: Will delete the file before proceeding to workload
            delete_first => {type => SCALAR , optional => 1, default => 0},
            # delete_ratios: proportions of truncate, delete, and noop to perform the
            # deletes. In order to do only truncates set it to [100,0,0]. [50,100,100]
            # does half truncates and half deletes. [33,67,100] truncates a third,
            # deletes a third, and skips a third so that they will be owerwritten later.
            #
            delete_ratios => {type => ARRAYREF, optional => 1, default => [33,67,100]},
            # files: list of files to run the workload on. By default a Workload acts
            # on all the files
            files => {type => ARRAYREF | UNDEF, optional => 1},
            # io_sleep: sleep time between IOs in milliseconds
            #io_sleep => {type => SCALAR , optional => 1, default => 0}, #Deprecated
            # max_tput: maximum throughput allowed during run in MB/s
            max_tput => {type => SCALAR , optional => 1, default => 0},
            # max_tput_interval: number of ios between each sleeps
            max_tput_interval => {type => SCALAR , optional => 1, default => 25},
            # max_tput_min_sleep: minimum amount of time we can reliably sleep for
            max_tput_min_sleep => {type => SCALAR , optional => 1, default => 0.01},
            # event_interval: number of seconds between event detection
            event_interval => {type => SCALAR , optional => 1, default => 1},
            # buffer_align: Boundaries to align buffer to
            buffer_align => {type => SCALAR| UNDEF, optional => 1, default => 512},
            # panic_on_corruption: panic the filer when a data corruption is detected
            # this cannot be set with verify_count
            panic_on_corruption => {type => SCALAR | UNDEF , optional => 1, default => 0},
            # print_on_corruptions: log useful info on corruptions. On by default except if
            # verify_count is enabled
            print_corruptions => {type => SCALAR | UNDEF , optional => 1},
            # skip_flush: don't sync filers before returning
            skip_flush  => {type => SCALAR | UNDEF, optional => 1, default => 0},
           }
        );

    if ($params{random} && !(defined($params{amount}) || defined($params{duration})) ){
        confess "You must specify an amount or a duration when starting a random workload";
    }
    if ($params{delete_first} && $params{random}){
            confess "Argument 'delete_first' is not yet supported on random workloads";
    }
    #if ($params{amount} && $params{duration}){
    #    confess "You cannot specify both an amount and a duration";
    #}
    #if ($params{random} && $params{duration} && $params{type} eq 'nfs'){
    #    confess "Feature not yet supported: random + nfs + duration";
    #}
    if ($params{read_pct} && $params{verify} && $params{type} =~ /nfs/ && $params{verify} !~ /pattern|checksum|auto/){
        confess "feature not yet supported: nfs + verify != pattern|checksum|auto";
    }
    if ($params{read_pct} && $params{verify} && $params{type} =~ /cifs/ && $params{verify} !~ /pattern|checksum|auto/){
        confess "feature not yet supported: cifs + verify != pattern|checksum|auto";
    }
    if ($params{container} && !$params{container}->isa("Perf::DataSet::Container")){
        confess "Argument 'dataset' must be a Perf::DataSet::Container";
    }
    if (defined($params{percentage}) && (defined($params{amount}) || defined($params{duration})) ){
        confess "You can only specify one of (percentage, duration, amount)";
    }
    if (defined($params{amount}) && defined($params{duration})){
        confess "You can only specify one of (percentage, duration, amount)";
    }
    if ($params{panic_on_corruption} && $params{verify_count}){
        confess "You can only specify one of (panic_on_corruption,verify_count)";
    }

    $params{io_size} = in_bytes($params{io_size});
    if ($params{amount}){
        $params{amount} = in_bytes($params{amount});
    }
    if ($params{io_size_min}){
        $params{io_size_min} = in_bytes($params{io_size_min});
    }
    if ($params{amount_min_per_file}){
        $params{amount_min_per_file} = in_bytes($params{amount_min_per_file});
    }

    if ($params{io_size} % $params{buffer_align} != 0){
        confess "buffer_align ($params{buffer_align}) must be a divisor of $params{io_size}";
    }

    if (! defined $params{print_corruptions}){
        $params{print_corruptions} = !$params{verify_count};
    }

    return %params;
}

sub new {
    my $class = shift;
    my %params = $class->validate(@_);
    my $self = \%params;
    $self->{abort} = 0;
    bless $self, $class;
    return $self;
}

sub pre_execute($$) {
    my $self = shift;
    my $owner = shift;
    if (!$owner || !ref($owner)){
        confess "You must specify an owner";
    }

    if ($owner->isa('Perf::DataSet::Worker')){

        #
        # Make sure we are going to know how to align the buffers for
        # directio
        #
        my $length_p = length(pack "p", "foo");
        my $length_J = length(pack "J", 0);
        confess "Platform not supported ( p: $length_p, J: $length_J )" if ( $length_p != $length_J );
        my $buffer_align = $self->{buffer_align};
        my $buffer_unit = 'x' x $buffer_align;

        # Create a read buffer and a write buffer
        # TODO: This really shouldn't be necessary but having a single buffer for both
        # triggers partial writes I don't fully understand. My guess is that sysread
        # "resizes" the buffer SV in a way that is not seen by the kludgy C code
        # in _generate_block. Having two buffers allows me to skip spending much
        # time in perlapi or perlguts to try to figure this out.
        for my $io_type ("read", "write"){
            my $standard_buffer = $buffer_unit x ( $self->{io_size} / $buffer_align + 1 );
            $self->{standard_buffer}->{$io_type} = \$standard_buffer;
            my $align_offset = unpack( "J", pack "p", $standard_buffer ) % $buffer_align;
            $self->{align_offset}->{$io_type} = $align_offset? ($buffer_align - $align_offset) : 0;
        }
        #
        # This worker may be brand new off a fork
        # Reset the mountpoint and relative path
        #
        my $process_uid = $owner->get_process_uid;

        # Cd to appropriate directory
        my $full_path = $self->{worker_full_paths}->{$process_uid};
        confess "No full path found!" if (!$full_path);
        $owner->{full_path} = $full_path;

        $Log->comment("Changing to directory '$full_path'");
        chdir $full_path or confess "Unable to chdir to '$full_path': $!";

        #
        # Get the list of appropriate files
        #
        # Figure out if what is the minimum size we can deal with

        my $min_size = min($self->{file_size_min},$self->{io_size_min});
        my $files_container =
                $owner->get_container->filter({type=>"files",
                                              min_size=>$min_size,
                                              compatible_with_io_size=>$self->{io_size},
                                              paths => $self->{files}});

        $self->{container} = $files_container;

        if ($self->{percentage} && $files_container->count('files')){
            # Turn the percentage into an amount for this Worker
            my $percentage = delete $self->{percentage};
            $self->{amount} = int($files_container->get_size('files') * $percentage / 100);
        }

        # So the Workload can listen for events
        $self->{event_channel} = $owner->{event_channel};

    }
    elsif($owner->isa('Perf::DataSet')){
        #
        # Check that we are not trying to write to a read-only Perf::DataSet
        #
        if ($owner->is_read_only && $self->{read_pct} ne 100 ){
            confess "You cannot write to a read_only Perf::DataSet";
        }

        #
        # Check that we have files in the dataset
        #
        if (!$owner->get_container->get_stats->get_field('size','files')){
            confess "You cannot invoke a workload on an empty dataset. You must first add files.";
        }

        if ($self->{delete_first} && $owner->is_file_slicing_allowed){
            confess "Argument delete_first is not supported for datasets that allow file slicing";
        }

        #
        # Check the files specified do exist in the Perf::DataSet
        #
        if ($self->{files}){
            # Remove from the file paths reference to the volume
            # or the relative_path (i.e. the path between the
            # root of the volume and the root_dir)
            my $vol = $owner->{vol};
            my $relative_path = $owner->get_relative_path;
            my $file_count = scalar @{$self->{files}};
            my $i = 0;
            while ($i < $file_count){
                my $file = shift @{$self->{files}};
                # Strip the path to the volume if present
                $file =~ s/\/vol\/$vol\///;
                # Strip the relative path if present;
                $file =~ s/\/?$relative_path\///;
                push @{$self->{files}}, $file;
                $i ++;
            }
            # Now check that the files exist in the owner's container
            if (! $owner->get_container->contains($self->{files})){
                confess "Not all files passed could be found";
            }
        }

        #
        # Check that we have files in the dataset with an appropriate block_size
        #
        my $valid_block_sizes = 0;
        my @block_sizes;
        if ($self->{files}){
            my %tmp = map {$owner->get_container->get_element($_)->get_block_size => 1} @{$self->{files}};
            @block_sizes = keys %tmp;
        }
        else{
            @block_sizes = keys %{$owner->get_container->get_stats->get_field('block_size')};
        }
        for my $block_size (@block_sizes){
            if ($self->{io_size} % $block_size == 0){
                $valid_block_sizes ++;
            }
        }
        if (!$valid_block_sizes){
            my $block_size_string = "@block_sizes";
            confess "io_size ($self->{io_size}) must be a factor of one of the block sizes " .
                "used to create data ($block_size_string)";
        }

        #
        # Set the current number of active workers
        # This is used to dynamically adjust the load
        #
        my $process_count = scalar(@{$owner->{manager}->get_processes($owner->{uid})});
        $self->{processes} = $process_count;

        if ($self->{amount}){
            # If we were passed files, turn the amount into a percentage
            if ($self->{files}){
                my $amount = in_bytes(delete $self->{amount});
                my $dataset_size = $owner->get_container->get_size('files');
                $self->{percentage} = $amount * 100 / $dataset_size;
            }
            # Split the 'amount' among Workers
            else{
                my $amount_per_worker = int($self->{amount} / $process_count) ;
                $self->{amount} = $amount_per_worker;
            }
        }

        $self->{type} = $owner->get_type;

        # Mount the Perf::DataSet. Paths for each worker have already been communicated during the add
        # No need to do it again.
        $owner->mount;

        # Embed the worker paths to the command so that workers can load their own
        # While this is already done in the Addition, forked datasets
        # may never execute an addition, so do this every time
        $self->{worker_full_paths} = $owner->get_worker_full_paths;

        # Set io_sleep to respect max_tput
        $self->adjust_io_sleep;

    }
    else{
        confess "The owner must be a Perf::DataSet or Perf::DataSet::Worker and not a " . ref($owner);
    }
}

sub post_execute($$){
    my $self = shift;
    my $owner = shift;
    if (!$owner || !ref($owner)){
        confess "You must specify an owner";
    }
    my $returns = shift;

    if ($owner->isa('Perf::DataSet::Worker')){

        # This isn't technically required yet since no workloads remove or add files
        # but in case some do in the future, update the Worker container.
        $owner->{container}->merge_with($self->{container});

        # While the Worker already contains the correct data about each Element, the Perf::DataSet
        # may not. Therefore we need to send all the elements back to the Perf::DataSet so it can
        # update its own container.
        $returns->{elements} = $self->{container}->get_rpc_serialized_elements;

        return $returns;
    }
    elsif($owner->isa('Perf::DataSet')){
        #
        # Make sure all data is written to disk
        #
        if (!$self->{skip_flush}){
            $Log->comment("Flushing appliance from " . __PACKAGE__ );
            my $appliance = Perf::Depot->get_resource(class=>'Appliance', name=>$owner->get_appliance);
            confess "No appliance found?!" if (!$appliance);
            $appliance->flush;
        }

        #
        # Merge results and return
        #
        my $load_report;
        my $tmp_container = Perf::DataSet::Container->new;
        my $max_duration = {read => 0, write => 0, total => 0};
        for my $process_uid (keys %$returns){
            #
            # Handle the 'common' section
            #
            my $return = $returns->{$process_uid}->{common};
            $load_report->{$process_uid} = $return;
            for my $k1 (keys %$return){
                my $v1 = $return->{$k1};
                if ($k1 eq 'duration'){
                    # Keep track of the maximum only
                    for my $k2 (keys %$v1){
                        if ($v1->{$k2} > $max_duration->{$k2}){
                            $load_report->{duration_max}->{$k2} = $v1->{$k2};
                            $max_duration->{$k2} = $v1->{$k2};
                        }
                    }
                }
                else{
                    my $v1_ref = ref($v1);
                    if ($v1_ref eq 'HASH'){
                        for my $k2 (keys %$v1){
                            $load_report->{$k1}->{$k2} += $v1->{$k2};
                        }
                    }
                    elsif(!$v1_ref){
                        $load_report->{$k1} += $v1;
                    }
                    else{
                        confess "Type '$v1_ref' not supported";
                    }
                }
            }

            #
            # Handle the 'elements' section
            #
            $return = $returns->{$process_uid}->{elements};
            my @elems = map { thaw(decode_base64($_)) } @$return;
            $tmp_container->set_elements(\@elems);

            #
            # Handle the 'individual' section.
            # Namely, update the size_on_disk of all files
            # since it may have changed for sparse files after
            # a write workload
            #
            #$return = $returns->{$process_uid}->{individual};
            #my $container = $owner->get_container;
            #for my $filepath (keys %$return){
            #    my $workload_file_data = $return->{$filepath};
            #    my $added_to_file_on_disk = $workload_file_data->{amount}->{write} -
            #                                $workload_file_data->{amount}->{overwrite};
            #    if ($added_to_file_on_disk && $self->{debug}){
            #        $Log->comment("Size_on_disk for '$filepath' grew by $added_to_file_on_disk");
            #    }
            #    my $file = $container->get_elements('files')->{$filepath};
            #    if (!$file){
            #        confess "This should never happen. Really.";
            #    }
            #    my $copy = $file->clone;
            #    my $old_size_on_disk = $file->get_size_on_disk;
            #    $copy->set_size_on_disk($old_size_on_disk + $added_to_file_on_disk);
            #    if ($copy->get_size_on_disk > $copy->get_size){
            #        confess "The size_on_disk for '$filepath' is greater than its size!\n"
            #                . "Clone:\n" . $copy->as_string
            #                . "File:\n"  . $file->as_string;
            #    }
            #    $container->set_element($copy);
            #}
        }
        $Log->comment("Workload completed:\n" . as_string($load_report));

        $tmp_container->unpartition;
        $owner->{container}->merge_with($tmp_container);

        $Log->debug("New container stats:\n" . $owner->{container}->get_stats->as_string );

        $owner->unmount(users_ok => 1);

        return $load_report;
    }
    else{
        confess "The owner must be a Perf::DataSet or Perf::DataSet::Worker and not a " . ref($owner);
    }
}

sub adjust_io_sleep {
    my $self = shift;
    if ($self->{max_tput}){
        my $worker_max_tput = $self->{max_tput} / $self->{processes};
        $self->{io_sleep} = $self->{io_size}/($worker_max_tput * 1024 * 1024);
    }
    else{
        $self->{io_sleep} = 0;
    }
    $Log->comment("Set io_sleep to $self->{io_sleep} seconds");
}

sub handle_WorkerComplete {
    my $self = shift;
    my $event = shift;
    $self->{processes} -= $event->{payload};
    $self->adjust_io_sleep;
    return;
}

sub handle_Abort {
    my $self = shift;
    my $event = shift;
    $self->{abort} = 1;
    return;
}

sub handle_MaxTput {
    my $self = shift;
    my $event = shift;
    $self->{max_tput} = $event->{payload};
    $self->adjust_io_sleep;
    return;
}

# Computes the X axis value for IO histograms, or "bucket"
# Buckets represent an upper bound, the lower bound being represented
# by the preceding bucket value. An exponential scale is used. Examples:
# Latency, bucket
# 1, 1
# 2, 2
# 10, 10
# 12, 20
# 23, 30
# 145, 200
# 398, 400
# etc...
sub _latency_bucket {
    my ($x) = @_; # Time in seconds
    $x *= 1000;
    # Some Perl builds have issues with small values of $x in 'log'
    # Return the bucket <=1ms if $x is less than a millisecond
    # TODO: Try to do something better here
    return 1 if ($x < 1);

    # Computes biggest 10**y number below $x
    my $log10 = 10**int(log($x)/log(10));
    # Return multiple of $log10 immediately above $x
    return ceil($x/$log10)*$log10;
}

sub _histogram {
    my ($io_histogram) = @_;
    confess "Unsupported input" if (!$io_histogram || !ref($io_histogram) || ref($io_histogram ne 'HASH'));
    my @buckets = sort {$a <=> $b} keys %$io_histogram;
    my $largest_bucket = $buckets[-1];
    my $hist_str;
    my $i = 0;
    my $bucket = 0;
    OUTER: while(1){
        for my $b (1..9){
            $bucket = $b*10**$i;
            last OUTER if ($bucket > $largest_bucket);
            my $hist_value = $io_histogram->{$bucket};
            $hist_str .= "<= ${bucket}ms : " . ($hist_value?$hist_value:0) . "\n";
        }
        $i++;
    }
    return $hist_str;
}


sub _get_fh {
    my ($self,$filepath,%args) = @_;
    my $create_file = $args{create_file};
    my $read_pct = $self->{read_pct};
    my $type = $self->{type};

    my $flags;
    if ($create_file){
        $flags = O_CREAT | O_EXCL;
    }
    # Enable writing only if asked to avoid issues on read-only
    # destinations like snapshots and others
    if ($read_pct < 100){
        $flags |= O_RDWR;
    }
    else{
        $flags |= O_RDONLY;
    }
    # Enable O_DIRECT whenever possible to disable client caching
    if ($type !~ /cifs/) {
        $flags |= O_DIRECT;
    }

    my $fh = FileHandle->new($filepath, $flags)
      or confess "Unable to open file '$filepath': $!";

    return $fh;
}


sub _compute_score($$$) {
    my $self = shift;
    my $file = shift;
    my $io_is_read = shift;

    my $filesize = $file->{size};
    my $read_pct = $self->{read_pct};

    my $score = $filesize *
                ($io_is_read?$read_pct/100:(1-$read_pct/100));

    return $score;

}

sub execute {
    my $self = shift;
    my $dataset_container = $self->{container};
    my $max_open_files = $MAX_OPEN_FILES;
    my $debug = $self->{debug};

    # Initialize hash used to log load activity
    my $load_report = {};
    $load_report->{common} = {amount   => {total => 0, read => 0, write => 0},
                              duration => {total => 0, read => 0, write => 0},
                            };
    my $load_report_common_duration = $load_report->{common}->{duration};
    my $load_report_common_amount = $load_report->{common}->{amount};

    #
    # Some debug information
    #
    my $ios = "";
    my $ios_count = 0;
    # Timings are kept for debug only
    my $timings = {file_choose => 0, file_open => 0, io => 0, post_process => 0, file_delete => 0, io_sleep => 0};
    $self->{io_histogram} = {};

    # Maintain a cache of open filehandles
    my %filehandles;
    my $files_count = $dataset_container->get_stats->get_field('type',"files");;
    if ($files_count >= $MAX_OPEN_FILES){
        tie %filehandles, 'Tie::Cache::LRU::LinkedList', $MAX_OPEN_FILES;
    }

    #$filehandles->{stats}->{max_opened} = 0;
    #$filehandles->{stats}->{max_open_files} = $max_open_files;

    my $workload_types;
    if ($self->{read_pct} == 100 || $self->{read_pct} == 0){
        $workload_types = 1;
    }
    else{
        $workload_types = 2;
    }
    my $random_pct = $self->{random}?100:0;
    my $read_pct = $self->{read_pct};
    my $io_size = $self->{io_size};
    my $verify = $self->{verify};
    my $verify_count = $self->{verify_count};
    my $amount = $self->{amount};
    my $duration = $self->{duration};
    my $delete_first = $self->{delete_first};
    my $io_size_min = $self->{io_size_min};
    my $io_sleep = $self->{io_sleep};
    my $max_tput_interval = $self->{max_tput_interval};
    my $max_tput_min_sleep = $self->{max_tput_min_sleep};
    my $event_interval = $self->{event_interval};
    my $async = $self->{async};
    my $buffer_read = $self->{standard_buffer}->{read};
    my $buffer_write = $self->{standard_buffer}->{write};
    my $align_offset_read = $self->{align_offset}->{read};
    my $align_offset_write = $self->{align_offset}->{write};
    my $print_corruptions = $self->{print_corruptions};

    #
    # Make sure we were given files to work on
    #
    my $dataset_size = $dataset_container->get_size("files");
    my $files = $dataset_container->get_elements('files', include_deleted => 0);
    if (! $dataset_container->count("files")){
        if ($self->{files}){
            #
            #   This is an acceptable case
            #
            $Log->comment("Nothing to do. Container for this Workload does not contain any files");
            return $load_report->{common};
        }
        else{
            confess "No files given as parameters";
        }
    }
    else{
        $Log->comment("Starting workload: " . $dataset_container->count("files") .
            " files, ${dataset_size}B, $max_open_files open files max");
        #if ($dataset_container->count("files") == 1){
        #    $Log->comment("Perf::DataSet: \n" . $dataset_container->as_string );
        #}
    }

    # Used for thottling mechanism
    my $tick_time = time;
    my $last_tick_time;

    # Used for event checking
    my $last_event_process = $tick_time;

    #
    # In order to make sure we are able to close all our files at the end of the workload
    # regardless of errors, wrap whole loop in an eval statement
    #
    eval{

    my $done = 0;
    my $file_index = 0;
    my @filekeys = keys %{$files};
    my $socket;
    # This value is not going to change (for now) and is expensive to compute
    # during random workloads
    my $total_size = $dataset_container->get_size('files');

    FILE: while(1){

        my $starttime_in_file = time;

        #
        # Pick a file to do IO on
        #
        my $file;
        if ($random_pct){

            #
            # Every $event_interval IOs process events
            #
            if ($starttime_in_file >= $last_event_process + $event_interval){
                $self->process_events;
                $last_event_process = $starttime_in_file;
                if ($self->{abort}){
                    $Log->comment('Aborting workload');
                    last FILE;
                }
                $io_sleep = $self->{io_sleep};
            }

            #
            # Pick a random file. Each file's chance of getting picked is
            # proportional to its size
            #
            $file = $dataset_container->random_file(weight=>'size', total_size=>$total_size);
        }
        else{

            #
            # Process events
            #
            $self->process_events;
            $last_event_process = $starttime_in_file;
            if ($self->{abort}){
                $Log->comment('Aborting workload');
                last FILE;
            }
            $io_sleep = $self->{io_sleep};

            my $filekey = $filekeys[$file_index];
            if (!$filekey){
                # We've done IOs to all files.  If we have $amount left or duration left,
                # reset index & reloop thru filepaths array
                if ( $self->{duration} && ($load_report_common_duration->{total} < $self->{duration})
                        || ($amount && ($load_report_common_amount->{total} < $amount)) ) {
                    $file_index = 0;
                    $filekey = $filekeys[$file_index];
                }
                #
                # If the user specified a 0 duration or amount
                # just loop forever
                #
                elsif ((defined($self->{duration}) && $self->{duration} == 0)
                       || (defined($self->{amount}) && $self->{amount} == 0)){
                    $file_index = 0;
                    $filekey = $filekeys[$file_index];
                }
                else {
                    #we're done
                    last FILE;
                }
            }
            $file = $files->{$filekey};
            $file_index++;
        }

        my $filesize = $file->get_size;
        my $filepath = $file->get_path;
        my $block_size = $file->get_block_size;
        my $offset = $file->get_offset;
        my $start_byte = $offset;
        if ( $self->{start_offset} ) {
            #check to see if we've reached the offset so this thread can work on this filechunk
            if ( $offset + $filesize <= $self->{start_offset}) {
                next FILE;
            }
            # else, only update $offset if we aren't yet beyond the $start_offset point.
            if ( $offset <= $self->{start_offset}) {
                    $start_byte = $self->{start_offset};
            }
        }
        my $read_time = 0;
        my $end_byte;
        my $amount_read_in_file = 0;
        my $amount_written_in_file = 0;
        my $amount_overwritten_in_file = 0;
        my $duration_read_in_file = 0;
        my $duration_written_in_file = 0;
        my $truncated_file = 0;
        my $deleted_file = 0;
        my $skipped_file = 0;
        my $validation_errors_in_file = 0;
        my $validation_oks_in_file = 0;
        my $fh;
        my $reached_eof = 0;
        my $original_start_byte = $start_byte;
        my $original_end_byte = $end_byte;
        my $latest_time = $starttime_in_file;
        #my $latest_time_debug = $starttime_in_file;

        #
        # Perform truncation if specified
        #
        # TODO: support random truncate workloads
        if ($self->{delete_first}){
            my $delete_rand = rand(100);
            my $delete_ratios = $self->{delete_ratios};
            my $delete_idx = 0;
            while ($delete_idx < 3){
                my $delete_ratio = $delete_ratios->[$delete_idx];
                if ($delete_rand <= $delete_ratio){
                    # Truncate
                    if ($delete_idx == 0){
                        truncate($filepath,0) or confess "Unable to truncate '$filepath': $!";
                        #
                        # For some reason truncates don't always work the first time...
                        #
                        my $tries = 1;
                        while ( ! -z $filepath){
                            truncate($filepath,0) or confess "Unable to truncate '$filepath': $!";
                            $tries ++;
                            if ($tries > 10){
                                confess "Unable to truncate $filepath after $tries attempts";
                            }
                        }
                        $truncated_file = 1;
                        if ($debug){
                            $Log->comment("Truncated $filepath");
                        }
                    }
                    # Unlink
                    elsif ($delete_idx == 1) {
                        #
                        # Remove all filehandles info associated with the file
                        # about to get deleted
                        #
                        $fh = delete $filehandles{$filepath};
                        if ($fh){
                            local $SIG{ALRM} = sub {confess "Timeout during close of '$filepath'\n"};
                            alarm 20;
                            $fh->close or confess "Unable to close file '$filepath':$!";
                            alarm 0;
                        }
                        if ( -f $filepath){
                            unlink($filepath) or confess "Unable to unlink '$filepath': $!";
                            if ( -f $filepath){
                                confess "File $filepath still present after unlink";
                            }
                            $deleted_file = 1;
                            if ($debug){
                                $Log->comment("Deleted $filepath");
                            }
                        }
                        else{
                            confess "Unable to unlink $filepath. File not found";
                        }
                    }
                    # Overwrite
                    elsif ($delete_idx == 2){
                        # Do nothing here; the file will be overwritten below.
                        $skipped_file = 1;
                        if ($debug){
                            $Log->comment("Skipped deletion/truncation of $filepath");
                        }
                    }
                    else{
                        confess "0";
                    }
                    last;
                }
                $delete_idx ++;
            }
        }

        IO_IN_FILE: while (1){

            #
            # Every $event_interval IOs make sure we weren't asked to abort
            # Skip this check for truncation workloads
            #
            my $start_io_in_file = time;
            if (!$random_pct &&  $start_io_in_file >=  $last_event_process + $event_interval ){
                $self->process_events;
                $last_event_process = $start_io_in_file;
                if (!$delete_first && $self->{abort}){
                    $Log->comment('Aborting workload');
                    last FILE;
                }
                $io_sleep = $self->{io_sleep};
            }

            #
            # Pick a location in the file
            #
            if ($random_pct){
                # Random IOs are aligned to the io_size
                # TODO: this doesn't handle start_offset cases correctly.
                $start_byte = $offset + int(rand($filesize) / $io_size) * $io_size;
            }
            else{
                # We reached the end of the file during the last iteration
                if ($start_byte >= $offset + $filesize){
                    last IO_IN_FILE;
                }
            }
            $end_byte = $start_byte + $io_size - 1;
            if ($end_byte > $offset + $filesize - 1){
                if (!$io_size_min ||
                        ($io_size_min && ($filesize - $start_byte > $io_size_min) )){
                    $end_byte = $offset + $filesize - 1;
                }
                else{
                    $start_byte -= $io_size;
                    $end_byte = $start_byte + $io_size - 1;
                    if (!$self->{random}){
                        $reached_eof = 1;
                    }
                }
            }
            my $start_block_num = $start_byte/$block_size;
            my $end_block_num = int($end_byte/$block_size);

            #
            # Figure out if we are going to do a read or a write
            #
            my $io_is_read = 1;
            if ($read_pct != 100){
                if ( int(rand(100)) >= $read_pct ){
                    $io_is_read = 0;
                }
            }
            #
            # Open the file if needed
            # Stash the filehandle for future needs
            $fh = $filehandles{$filepath};
            if (!$fh){
                $fh = $self->_get_fh( $filepath, create_file=>$deleted_file );
                $filehandles{$filepath} = $fh;
            }

            #
            # Seek in the file to the selected position
            #
            my $res = $fh->sysseek($start_byte, SEEK_SET);
            if (!defined($res)) {
                confess "Seek to ($start_byte/$filesize) failed for '$filepath': $!\n";
            }

            my $data;
            my @blocks_data;
            my $actual_io_size = $end_byte - $start_byte + 1;

            if ($debug && $ios_count < 100){
                $ios .= "$file,$start_byte,$actual_io_size\n";
            }
            $ios_count ++;
            my $time_before_io;
            if ($io_is_read){
                $time_before_io = time;
                my $res = $fh->sysread($$buffer_read,$actual_io_size,$align_offset_read);

                if (!defined($res) || $res != $actual_io_size ) {
                    $res ||= 0;
                    my $syserr = $!;
                    # Capture latency of failed IO
                    my $io_time = time - $time_before_io;
                    my $io_time_bucket = _latency_bucket($io_time);
                    $self->{io_histogram}->{read}->{$io_time_bucket} ++;
                    confess "Read failed for '$filepath' at block $start_block_num ".
                            "($res bytes instead of $actual_io_size) after $io_time seconds: $syserr\n";
                }
                $amount_read_in_file += $actual_io_size;

                my $buffer_content = substr($$buffer_read, $align_offset_read, $actual_io_size);
                $data = \$buffer_content;
            }
            else{
                #
                # Generate data for overwrite
                #
                my $left = $actual_io_size;
                for my $block_num ($start_block_num..$end_block_num){
                    my $curr_block_size = ($block_size <= $left )?$block_size:$left;
                    my $bl = $file->generate_block(
                                                    buffer   => $buffer_write,
                                                    offset   => $actual_io_size - $left + $align_offset_write,
                                                    block_num=> $block_num,
                                                    size     => $curr_block_size,
                                                    );
                    push @blocks_data, $bl;
                    $left -= $curr_block_size;
                }

                # Don't include data generation in latency numbers
                $time_before_io = time;
                #
                # Perform overwrite
                #
                my $res = $fh->syswrite($$buffer_write,$actual_io_size,$align_offset_write);
                if (!defined($res) || $res != $actual_io_size){
                    $res ||= 0;
                    my $syserr = $!;
                    # Capture latency of failed IO
                    my $io_time = time - $time_before_io;
                    my $io_time_bucket = _latency_bucket($io_time);
                    $self->{io_histogram}->{write}->{$io_time_bucket} ++;
                    confess "Write failed for '$filepath' at block $start_block_num ".
                            "($res bytes instead of $actual_io_size) after $io_time seconds: $syserr\n";
                }
                $amount_written_in_file += $actual_io_size;
            }
            my $time_after_io = time;
            my $io_time = $time_after_io - $time_before_io;
            my $io_time_bucket = _latency_bucket($io_time);
            if ($io_is_read){
                $self->{io_histogram}->{read}->{$io_time_bucket} ++;
            }
            else{
                $self->{io_histogram}->{write}->{$io_time_bucket} ++;
            }
            #
            # Post process data (validate or checksum)
            #
            if ( ($io_is_read && $verify) || !$io_is_read ){
                if ($io_is_read){
                    # Split the data into blocks
                    #@blocks_data = $$data =~ /(.{1,$block_size})/g;
                    @blocks_data = unpack "(a$block_size)*", $$data;
                }
                confess "No blocks data for blocks $start_block_num to $end_block_num in " . $file->get_path. "!"
                    if (!@blocks_data);
                my $b = 0;
                my $total_blocks_size = 0;
                for my $block_data (@blocks_data){
                    my $actual_block_size = length($block_data);
                    $total_blocks_size += $actual_block_size;
                    if ($io_is_read){
                        my $s = $file->validate_block_data({block_num=>$b + $start_block_num,
                                                            block_data=>\$block_data,
                                                            verify=>$verify,
                                                            corruptions_ok=>$verify_count,
                                                            corruptions_print=>$print_corruptions,
                                                            });
                        if (!$s){
                            $validation_errors_in_file ++;
                        }
                        else{
                            $validation_oks_in_file ++;
                        }
                    }
                    else{
                        my $ret = $file->record_block($b + $start_block_num,\$block_data);
                        if ($ret == $Perf::DataSet::BLOCK_UPDATED){
                            $amount_overwritten_in_file += $actual_block_size;
                        }
                        elsif( $ret == $Perf::DataSet::BLOCK_ADDED ){
                            $file->set_size_on_disk($file->get_size_on_disk + $actual_block_size);
                        }
                        else{
                            confess "Invalid return value $ret";
                        }
                    }
                    $b++;
                }
                # DEBUG;
                if ($total_blocks_size != $actual_io_size){
                    confess "Something very wrong is happening:\n" .
                            "Total \@blocks_data size: $total_blocks_size\n".
                            "Actual IO size: $actual_io_size\n" .
                            "Blocks Data:\n" . join("\n", @blocks_data);
                }
            }
            if ( !$self->{random} && $end_byte == $filesize - 1){
                $reached_eof = 1;
            }

            #
            # Sleep io_sleep seconds if asked to
            #
            my $time_after_process = time;
            my $time_after_sleep = $time_after_process;
            if ($io_sleep && $ios_count % $max_tput_interval == 0){
                # Find out how much time was spent not sleeping since the last sleep
                # or the last time we went through this if no sleep happened
                $last_tick_time = $tick_time;
                $tick_time = $time_after_process;
                my $time_since_last_tick = $tick_time - $last_tick_time;

                my $sleep_time = $io_sleep * $max_tput_interval - $time_since_last_tick;
                if ($sleep_time > 0 && $sleep_time < $max_tput_min_sleep){
                    # Rather than try to sleep for such a small amount of time
                    # Figure out if we should sleep at all
                    $sleep_time = rand($max_tput_min_sleep/$sleep_time) <= 1 ? $max_tput_min_sleep : 0;
                }
                if ($sleep_time > 0){
                    my $ret = select (undef,undef,undef,$sleep_time);
                    if ($ret == -1){
                        confess "Unable to sleep between IOs: $!";
                    }
                    # Include the sleep in the tick period
                    $tick_time = time;
                    $time_after_sleep = $tick_time;
                }
            }

            if ( $read_pct != 0 || $read_pct != 100 ) {
                my $elapsed = $time_after_sleep - $latest_time;
                if ($io_is_read) {
                    $duration_read_in_file += $elapsed;
                } else {
                    $duration_written_in_file += $elapsed;
                }
                $latest_time = $time_after_sleep;
            }

            #
            # Examine exit criteria
            #
            if ($amount && !$delete_first && $load_report_common_amount->{total} + $amount_read_in_file
                            + $amount_written_in_file >= $amount){
                # We are done with this file and all remaning files
                $done = 1;
                last IO_IN_FILE;
            }
            #if (!$amount && $reached_eof ){
            if ( $reached_eof ){
                if ($delete_first && $amount && $load_report_common_amount->{total} + $amount_read_in_file
                            + $amount_written_in_file >= $amount){
                    # Truncation workloads can only be stopped after the end of a file
                    $done = 1;
                }
                if ($delete_first && $duration && $duration_written_in_file + $duration_read_in_file
                            + $load_report_common_duration->{total} >= $duration){
                    # Truncation workloads can only be stopped after the end of a file
                    $done = 1;
                }
                # We are done with this file, go to the next one
                last IO_IN_FILE;
            }
            if ( $duration && !$delete_first && $duration_written_in_file + $duration_read_in_file
                            + $load_report_common_duration->{total} >= $duration ) {
                $done = 1;
                last IO_IN_FILE;
            }
            if ($random_pct){
                # For random loads we pick a different file for each IO
                last IO_IN_FILE;
            }
            #
            # For sequential reads increment the start_byte.
            # Random loads will overwrite this
            #
            $start_byte += $io_size;

        } # End of IO_IN_FILE

        if ( $read_pct == 0 || $read_pct == 100 ) {
            my $elapsed = time - $starttime_in_file;
            if ($read_pct == 0) {
                $duration_written_in_file = $elapsed;
            }
            elsif ($read_pct == 100) {
                $duration_read_in_file = $elapsed;
            }
        }

        if (!$self->{random}){
            $start_byte = $original_start_byte;
            #$end_byte = $original_end_byte;
        }

        #
        # Log the IO(s)
        #

        my $load_report_individual = $load_report->{individual}->{$filepath};
        if (!$load_report_individual){
            $load_report->{individual}->{$filepath} = {};
            $load_report_individual = $load_report->{individual}->{$filepath};
        }
        $load_report_common_duration->{total} += $duration_written_in_file + $duration_read_in_file;
        $load_report_common_duration->{read} += $duration_read_in_file;
        $load_report_common_duration->{write} += $duration_written_in_file;
        $load_report_common_amount->{total} += $amount_read_in_file + $amount_written_in_file;
        $load_report_common_amount->{read} += $amount_read_in_file;
        $load_report_common_amount->{write} += $amount_written_in_file;
        $load_report_common_amount->{overwrite} += $amount_overwritten_in_file;
        $load_report->{common}->{file_deletion}->{truncated} += $truncated_file;
        $load_report->{common}->{file_deletion}->{deleted} += $deleted_file;
        $load_report->{common}->{file_deletion}->{skipped} += $skipped_file;
        $load_report->{common}->{validation_errors} += $validation_errors_in_file;
        $load_report->{common}->{validation_oks} += $validation_oks_in_file;
        $load_report_individual->{duration}->{total} += $duration_written_in_file + $duration_read_in_file;
        $load_report_individual->{duration}->{read} += $duration_read_in_file;
        $load_report_individual->{duration}->{write} += $duration_written_in_file;
        $load_report_individual->{amount}->{total} += $amount_read_in_file + $amount_written_in_file;
        $load_report_individual->{amount}->{read} += $amount_read_in_file;
        $load_report_individual->{amount}->{write} += $amount_written_in_file;
        $load_report_individual->{amount}->{overwrite} += $amount_overwritten_in_file;
        $load_report_individual->{random_pct} = $random_pct;
        #$load_report_individual->{start_byte} = min($start_byte,$load_report_individual->{start_byte});
        #$load_report_individual->{end_byte} = max($end_byte,$load_report_individual->{end_byte});
        $load_report_individual->{io_size} = $io_size;
        $load_report_individual->{verify} = $verify;
        $load_report_individual->{file_deletion}->{truncated} += $truncated_file;
        $load_report_individual->{file_deletion}->{deleted} += $deleted_file;
        $load_report_individual->{file_deletion}->{skipped} += $skipped_file;
        $load_report_individual->{validation_errors} += $validation_errors_in_file;
        $load_report_individual->{validation_oks} += $validation_oks_in_file;

        #
        # Examine exit criteria
        #
        if ($done){
            last FILE;
        }
    } # End of FILE

    }; # EVAL
    my $error_during_workload = $@;
    if ($error_during_workload){
        $Log->comment($error_during_workload);
    }

    my $histograms = delete $self->{io_histogram};
    my @op_types = keys %$histograms;
    for my $op_type (@op_types){
        $Log->comment("$op_type IO histogram:\n" . _histogram($histograms->{$op_type}));
    }
    # Combine histograms
    if (scalar(@op_types) > 1){
        my %combined_histogram;
        for my $op_type (@op_types){
            my $io_histogram = $histograms->{$op_type};
            for my $bucket (keys %$io_histogram){
                $combined_histogram{$bucket} += $io_histogram->{$bucket};
            }
        }
        $Log->comment("Combined IO histogram:\n" . _histogram(\%combined_histogram));
    }
    $Log->comment("Workload whole report:\n" . as_string($load_report));
    $Log->comment("First ios:\n$ios");

    #
    # Close all filehandles
    #
    my $timeouts = 0;
    my $error_during_close = "";
    CLOSE_FILE: while ( my ($file,$fh) = each %filehandles){
        print "Closing filehandle for '$file'\n";
        eval{
            local $SIG{ALRM} = sub {die "Timeout during close\n"};
            alarm 20;
            $fh->close or confess "Unable to close filehandle for file '$file':$!";
            alarm 0;
        };
        if ($@ && $@ ne "Timeout during close\n"){
            $error_during_close .= "$@\n";
            $Log->comment($@);
        }
        if ($@){
            # Timed out
            $timeouts ++;
            $error_during_close .= "Timeout during close of '$file'\n";
            $Log->comment("Timeout during close. Skipping rest of closing sequence");
            last CLOSE_FILE;
        }
    }
    $Log->comment("Deleting filehandles list. May take a while.");
    undef %filehandles;
    $Log->comment("Deleting done");

    #
    # Die from errors during workload first
    #
    if ($error_during_workload){
        die $error_during_workload;
    }
    if ($error_during_close){
        confess $error_during_close;
    }

    return $load_report;
}
1;
