# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Depot - Resource Depot singleton

=head1 DESCRIPTION

This class keeps track of Resources used by Perf libraries
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw($Log);

    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use threads;
    use Perf::Utils qw(as_string);
    use List::Util qw(sum reduce);
    use List::MoreUtils qw(uniq);

    use namespace::clean;
    use Object::InsideOut;
    use Perf::Depot::Resource;


    my @resources       :Field :Name(resources);
    my @empty           :Field :Name(empty) :Get(is_empty);
    my @uids            :Field :Name(uids);
    my @pid             :Field :Name(pid) :Get(get_pid);
    my @tid             :Field :Name(tid) :Get(get_tid);
    my @destroying      :Field :Name(destroying) :Get(is_destroying);

    my $Depot = __PACKAGE__->Object::InsideOut::new;

    # Don't let the depot be cleaned up at global destruction time
    # since it won't happen in the right order. Instead perform
    # an orderly cleanup in the END block below
    END {
        # Not sure when that would happen
        return if (!$Depot);

        # Return if process isn't the root process
        # TODO: Would we ever add resources to the Depot from
        # a different process than the parent one? If so,
        # we'll need to record the pid for all objects
        # and perform the appropriate cleanup based on $$
        return if ($$ != $Depot->get_pid);

        $Log->comment("Cleaning up Depot");

        # Find all the Perf::DataSets still present and call their DESTROY subroutine
        my $dataset_resources = $Depot->get_resources( class => 'DataSet');
        if ($dataset_resources && @$dataset_resources){
            for my $d (@$dataset_resources){
                my $d_dataset = $d->get_dataset;
                if ($d_dataset){
                    # The Perf::DataSet object still exists, call DESTROY
                    $d_dataset->DESTROY;
                }
                else{
                    # Perf::DataSet object is gone already, remove the Perf::DataSet
                    # resource from the Depot and let cleanup happen
                    $Depot->remove_resource($d);
                    undef $d;
                }
            }
            undef $dataset_resources;
        }
        # There should not be anything else, but just in case
        $Depot->remove_unused_and_never_used;

        undef $Depot;
        $Log->comment("Done cleaning up Depot");
    }

    sub new {
        return $Depot;
    }

    sub _init :Init {
        my $self = $_[0];
        $resources[$$self] = {};
        $empty[$$self] = 1;
        $uids[$$self] = {};
        $pid[$$self] = $$;
        $tid[$$self] = threads->tid;
        $destroying[$$self] = 0;
        $Log->comment("Created " . ref($self));
    }

    sub get_resource {
        shift;
        my $self = $Depot;
        my %args = @_;

        my $create_on_miss = delete $args{create_on_miss};
        if (!defined($create_on_miss)){
            $create_on_miss = 0;
        }
        my $rs;
        my $rsa = $self->get_resources(%args);
        if (!$rsa || !@$rsa){
            if ( $create_on_miss ){
                $rs = $self->create_resource(%args);
            }
        }
        else{
            if (scalar(@$rsa) > 1){
                my $str = "More than 1 resource found:\n- ";
                $str .= join("\n- ", map { $_->get_description} @$rsa);
                confess $str ;
            }
            $rs = $rsa->[0];
        }
        return $rs;
    }

    sub create_resource {
        shift;
        my $self = $Depot;
        my %args = @_;
        my $class = delete $args{class};
        confess "You must specify a class" if (!$class);
        $class = 'Perf::Depot::Resource::' . $class;
        eval "use $class";
        if ($@){
            confess "Unable to load $class: $@";
        }
        chomp($class);
        return $class->new(%args);
    }

    sub get_resources {
        shift;
        my $self = $Depot;
        my %args = @_;
        my $class = delete $args{class};
        confess "You must specify a class" if (!$class);
        my $rs = $resources[$$self]->{$class};
        my $matches;
        for my $r (@$rs){
            push(@$matches,$r) if ( $r->match(%args) );
        }
        return $matches;
    }

    sub is_resource_registered {
        shift;
        my $self = $Depot;
        my $resource = shift;
        confess "You must specify a resource" if (!$resource);
        confess "Argument must be a Perf::Depot::Resource" if (!$resource->isa('Perf::Depot::Resource'));
        my $resource_classes = $resource->get_classes;
        my %registered;
        for my $class (@$resource_classes){
            my $rs = $resources[$$self]->{$class};
            if (!$rs || !$resource->found_in($rs)){
                $registered{$class} = 0;
                next;
            }
            $registered{$class} = 1;
        }
        # Check for partial registration
        my $classes_count = scalar(@$resource_classes);
        my $registration_status = sum values %registered;
        if ($registration_status != 0 && $registration_status != $classes_count){
            confess "Resource " . $self->get_description . " is partially registered!\n"
                    . as_string(\%registered);
        }
        else{
            return $registration_status?1:0;
        }
    }

    sub add_resource {
        shift;
        my $self = $Depot;
        my ($resource, %args) = @_;
        confess "You must specify a resource" if (!$resource);
        confess "Argument must be a Perf::Depot::Resource" if (!$resource->isa('Perf::Depot::Resource'));
        confess "Resource " . $self->description . " already exists" if ($self->is_resource_registered($resource));
        my $resource_classes = $resource->get_classes;
        for my $class (@$resource_classes){
            $Log->debug("Adding resource '". $resource->get_description ."' to list of '$class' in depot");
            push @{$resources[$$self]->{$class}}, $resource;
        }
        $empty[$$self] = 0;
    }

    sub remove_resource {
        shift;
        my $self = $Depot;
        my $resource = shift;
        confess "You must specify a resource" if (!$resource);
        confess "Argument must be a Perf::Depot::Resource"
          if (!$resource->isa('Perf::Depot::Resource'));
        confess "Resource " . $resource->get_description . " not in depot!"
          if (!$self->is_resource_registered($resource));

        my $users = $resource->get_users();
        confess(  'Not allowed to remove resource '
                . $resource->get_description()
                . " since it is still being used by:\n"
                . join("\n", map { $_->get_description() } @$users))
          if (defined($users) && 0 < scalar(@$users));

        $Log->debug("Removing resource '". $resource->get_description ."' from depot");
        my $resource_classes = $resource->get_classes;
        for my $class (@$resource_classes){
            my $class_resources = $resources[$$self]->{$class};
            my $class_resources_count = $class_resources?scalar(@$class_resources):0;
            my $i = 0;
            while ($i < $class_resources_count){
                my $r = $class_resources->[$i];
                if ($r == $resource){
                    splice @$class_resources,$i,1;
                    $Log->debug("Resource " . $resource->get_description .
                                  " removed from list of '$class' Resources in depot" );
                    last;
                }
                $i ++;
            }
        }
        # Free the uid if necessary
        my $r_uid = $resource->get_uid;
        $self->free_uid($r_uid) if ($r_uid);
        return;
    }

    sub remove_unused_in {
        shift;
        my $list = shift;
        return if (!$list);
        my $self = $Depot;
        confess "Argument must be a ARRAY_ref" if (ref($list) ne 'ARRAY');
        for my $r (@$list){
            $self->remove_resource($r) if ( !$r->was_never_used && !$r->is_used );
        }
        return;
    }

    sub remove_unused {
        shift;
        my $self = $Depot;
        for my $class (keys %{$resources[$$self]}){
            my $class_resources = $resources[$$self]->{$class};
            $self->remove_unused_in($class_resources);
        }
        return;
    }

    sub remove_unused_and_never_used {
        shift;
        my $self = $Depot;
        $Log->comment("Looking for unused and never used resources");
        while (1){
            my $removed_resources = 0;
            for my $class (keys %{$resources[$$self]}){
                my $class_resources = $resources[$$self]->{$class};
                my $class_resources_scalar = $class_resources?scalar(@$class_resources):0;
                next if (!$class_resources_scalar);
                my $rsrc_str = join(', ', map { $_->get_description } @$class_resources);
                $Log->debug("Looking for unused among $class_resources_scalar $class resource(s): $rsrc_str");
                # We need to make a copy here because remove_resource will splice the
                # $class_resources array if called
                my @class_resources_copy = @$class_resources;
                for my $r (@class_resources_copy){
                    my $r_users = $r->get_users;
                    if (!$r_users || !@$r_users){
                        $Log->debug("Found unused resource: " . $r->get_description);
                        $self->remove_resource($r);
                        $removed_resources ++;
                    }
                    else{
                        my $users_string = '- ' . join("\n- ", map { $_->get_description } @$r_users);
                        $Log->debug("Resource ". $r->get_description ." is being used by:\n$users_string" );
                    }
                    #if ( ! $r->is_used ){
                    #    $Log->comment("Found unused resource: " . $r->get_description);
                    #    $self->remove_resource($r)
                    #}
                }
            }
            last if (!$removed_resources);
            $Log->comment("Searching again for unused and never used resources");
        }
        return;
    }

    sub list_resource_users {
        shift;
        my $self = $Depot;
        for my $class (keys %{$resources[$$self]}){
            my $class_resources = $resources[$$self]->{$class};
            next if (!$class_resources);
            for my $r (@$class_resources){
                my $r_users = $r->get_users;
                next if (!$r_users || !@$r_users);
                my $users_string = '- ' . join("\n- ", map { $_->get_description } @$r_users);
                $Log->comment("Resource ". $r->get_description ." is being used by:\n$users_string" );
            }
        }
        return;
    }

    sub random_number {
        return int(rand(2**32));
    }

    sub make_uid {
        shift;
        my $self = $Depot;
        my $uid = random_number();
        while ($uids[$$self]->{$uid}){
            $Log->comment("Uid collision detected! Creating another uid");
            $uid = random_number();
        }
        $uids[$$self]->{$uid} = 1;
        return $uid;
    }

    sub free_uid {
        my (undef,$uid) = @_;
        confess "You must specify a uid!"  if (!$uid);
        my $self = $Depot;
        delete $uids[$$self]->{$uid};
        return;
    }


    # forcibly remove a resource that is depended upon by other resources
    sub eject_resource {
        shift();
        my $self = $Depot;
        my $resource = shift();

        confess('Must specify a Resource')
            if (!defined($resource) || !$resource->isa('Perf::Depot::Resource'));

        $Log->debug('Ejecting resource: ' . $resource->get_description());

        my $eject_users = $resource->get_users();

        foreach my $user (@$eject_users) {
            $resource->disconnect_user($user);
        }

        # After all strong references to a Resource are gone, it
        # should automatically be GC'd

        return;
    }


    # Eject a given resource from the Depot and all resources that
    # directly or indirectly depend on it up to a given Resource.
    # Return an arrayref of matched "To" resources
    sub eject_from_to {
        shift();
        my %arg  = @_;
        my $from = $arg{from};
        my $tos   = $arg{to};
        my @ejections = ($from);
        my @matches;

        while (0 < scalar(@ejections)) {
            my $eject = shift(@ejections);
            my $users = $eject->get_users();

            USER: foreach my $user (@$users) {
                if ($user->is_dependent_on($eject)) {
                    foreach my $to (@$tos) {
                        if ($user->isa("Perf::Depot::Resource::$to")) {
                            push(@matches, $user);
                            next USER;
                        }
                    }

                    push(@ejections, $user);
                }
            }

            Perf::Depot->eject_resource($eject);
        }

        return [uniq(@matches)];
    } ## end sub eject_from_to
}
1;
