# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource - Base class for resource classes used by Perf::DataSet::Manager

=head1 DESCRIPTION

This class is the base class for all Resource subclasses. It provides all the common functionality.

Resources are organized in a tree like fashion: resources can declare dependencies on other resources
or declare themselves as users of other resources. Resources contain hard references to their
dependencies but only weak links to their users. This allows for cleanup to happen from the top
of the tree down to the lowest resources (usually the clients or filers).

It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use List::MoreUtils qw(any);
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Scalar::Util qw();
    use threads;
    use Perf::Utils qw(as_string);
    use Object::InsideOut;

    my @dependencies    :Field :Name(dependencies);
    my @users           :Field :Name(users);
    my @name            :Field
                        :Arg('Name'=>'name', 'Mandatory' => 1)
                        :Get(get_name);
    my @pid             :Field :Name(pid) :Get(get_pid);
    my @tid             :Field :Name(tid) :Get(get_tid);
    my @destroying      :Field :Name(destroying) :Get(is_destroying);
    my @never_used      :Field :Name(never_used) :Get(was_never_used);
    my @uid             :Field :Arg(Name=>'uid', Mandatory=>0) :Get(get_uid);

    my $dumped = {};

    sub _init :Init {
        my $self = $_[0];
        $Log->comment("Instantiating new resource: " . $self->get_description);
        $users[$$self] = [];
        $never_used[$$self] = 1;
        $dependencies[$$self] = [];
        $pid[$$self] = $$;
        $tid[$$self] = threads->tid;
        $destroying[$$self] = 0;
    }

    sub get_classes {
        my $self = $_[0];
        my @raw_classes = $self->meta->get_classes;
        my @classes;
        for (@raw_classes){
            # We don't keep the base class 'Resource', since it will match
            # all objects and currently wouldn't be used for anything
            push @classes,$1 if (/Resource::(.*)/);
        }
        return \@classes;
    }

    sub set_uid :Restricted {
        my ($self, $id) = @_;
        confess "You must specify an argument" if (!$id);
        $uid[$$self] = $id;
    }

    sub set_name {
        my ($self,$new_name) = @_;
        confess "You must pass an argument" if (!$new_name);
        $Log->comment(  "Resource "
                      . $self->get_description
                      . " is being renamed '$new_name'")
          if ($new_name ne $name[$$self]);
        $name[$$self] = $new_name;
    }

    sub get_resource {
        my $self = shift;
        my %args = @_;
        return  Perf::Depot->get_resource(%args);
    }

    sub get_resources {
        my $self = shift;
        my %args = @_;
        return Perf::Depot->get_resources(%args);
    }

    sub get_description {
        my $self = $_[0];
        my $class = ref($self);
        $class =~ s/Perf::Depot::Resource:://;
        return "$class '$name[$$self]'";
    }

    sub register {
        my $self = shift;
        if (!$self->is_registered){
            Perf::Depot->add_resource($self);
            # Check the dependencies are registered as well
            map { $_->register } @{$dependencies[$$self]};
        }
        else{
            $Log->debug( $self->get_description . " is already registered");
        }
    }

    sub is_registered {
        my $self = shift;
        return Perf::Depot->is_resource_registered($self);
    }

    sub add_dependencies {
        my ($self, $arg) = @_;
        if (!$arg){
            confess "You must specify a resource or array of resources";
        }
        if (ref($arg) ne 'ARRAY'){
            $arg = [$arg];
        }
        for my $resource (@$arg){
            my $ref_resource = ref($resource);
            if (!$ref_resource || $ref_resource eq 'ARRAY' || $ref_resource eq 'HASH'){
                confess "Dependency isn't an object: " . as_string($resource);
            }
            if (!$resource->isa('Perf::Depot::Resource')){
                confess "Argument must be a Resource and not a " . ref($resource);
            }
            # Avoid loops
            if ($resource->found_in($users[$$self])){
                confess "Resource " . $resource->get_description . " is listed as a user of  " .
                    $self->get_description . " already and cannot be added as a dependency of same";
            }
            if (!$self->is_dependent_on($resource)){
                push @{$dependencies[$$self]}, $resource;
                $resource->add_user($self);
            }
            if ($self->is_registered && !$resource->is_registered){
                $resource->register;
            }
            map { $resource->add_user($_) } @{$users[$$self]};
        }
    }

    sub get_dependency {
        my $self = shift;
        my $deps = $self->get_dependencies(@_);
        confess "No dependency found" if (!$deps || !@$deps);
        confess "Too many dependencies found" if ( scalar(@$deps) > 1);
        return $deps->[0];
    }

    sub get_dependencies {
        my $self = shift;
        my %args = @_;
        if (%args){
            my @deps = grep { $_->match(%args) } @{$dependencies[$$self]};
            return \@deps;
        }
        return $dependencies[$$self];
    }

    sub is_shared_by {
        my $self = shift;
        my $sharers = $self->get_users(@_);
        return 0 if (!$sharers);
        return scalar(@$sharers);
    }

    sub found_in {
        my ($self, $list) = @_;
        confess "You must specify an argument" if (!$list);
        confess "You must specify an arrayref" if (ref($list) ne 'ARRAY');
        for my $r (@$list){
            confess "Object is not a Perf::Depot::Resource" if (!$r->isa('Perf::Depot::Resource'));
            return 1 if ($self == $r);
        }
        return 0;
    }

    sub get_users {
        my $self = shift;
        $self->_prune;
        my %args = @_;
        if (%args) {
            my @users = grep { $_->match(%args) } @{$users[$$self]};
            return \@users;
        }
        else {
            return $users[$$self];
        }
    }

    sub get_user {
        my $self = shift;
        my $matching_users = $self->get_users(@_);
        return if (!$matching_users || !@$matching_users);
        confess "More than one user found" if (scalar(@$matching_users) > 1);
        return $matching_users->[0];
    }

    sub is_dependent_on {
        my ($self, $resource) = @_;
        if (!$resource){
            confess "You must specify a resource";
        }
        if (!$resource->isa('Perf::Depot::Resource')){
            confess "Argument must be a Resource and not a " . ref($resource);
        }
        return 1 if ($resource->found_in($dependencies[$$self]));
        return 0;
    }

    sub add_user {
        my ($self,$resource) = @_;
        if (!$resource){
            confess "You must specify a resource";
        }
        if (!$resource->isa('Perf::Depot::Resource')){
            confess "Argument must be a resource and not a " . ref($resource);
        }
        # Avoid loops
        if ($resource->found_in($dependencies[$$self])){
            confess "Resource " . $resource->get_description . " is listed as a dependency of  " .
                    $self->get_description . " already and cannot be added as a user of same";
        }
        my $was_added = 0;
        if (!$self->is_used_by($resource)){
            my $item = $resource;
            $Log->debug("Resource " . $self->get_description . " is now used by " . $resource->get_description);
            push @{$users[$$self]}, $item;
            # To prevent memory leaks and have working cleanup sequence,
            # only store weak references
            Scalar::Util::weaken($users[$$self]->[-1]);
            $never_used[$$self] = 0;
            $was_added = 1;
        }
        map { $_->add_user($resource) } @{$dependencies[$$self]};
        return $was_added;
    }

    sub remove_user {
        my ($self,$resource,%args) = @_;
        my $recursive = $args{recursive};
        if (!$resource){
            confess "You must specify an argument";
        }
        if (!$resource->isa('Perf::Depot::Resource')){
            confess "Argument must be a resource and not a " . ref($resource);
        }
        if (!$self->is_used_by($resource)){
            $Log->debug("Resource " . $self->get_description . " not used by resource " . $resource->get_description);
            return;
        }
        else{
            my $i = 0;
            my $users_count = scalar(@{$users[$$self]});
            while ($i < $users_count){
                my $r = $users[$$self]->[$i];
                if ($r == $resource){
                    splice @{$users[$$self]},$i,1;
                    $Log->debug("Resource " . $resource->get_description . " removed from users of " . $self->get_description );
                    last;
                }
                $i ++;
            }
        }

        # By default we don't tell the dependencies to remove $resource from their
        # users list since we don't know if $self is the only way $resource
        # was added as one of their users
        if ($recursive){
            map { $_->remove_user($resource, recursive=>1)  } @{$dependencies[$$self]};
        }
        # If we are no longer used, remove from depot
        # This should trigger destroy on $self
        my $remaining_users = $self->get_users;
        if (!$remaining_users || !@$remaining_users){
            $Log->debug( "Resource " . $self->get_description . " isn't used anymore" );
            Perf::Depot->remove_resource($self);
        }
        return;
    }

    sub match {
        my ($self, %criteria) = @_;
        # The 'class' is treated differently to respect
        # OO inheritance
        my $class = delete $criteria{class};
        return 0 if ($class && !$self->isa('Perf::Depot::Resource::' . $class));
        for my $field (keys %criteria){
            my $handler_name;
            if ($field =~ /^(get|is)_/){
                $handler_name = $field;
            }
            else{
                $handler_name = "get_$field";
                if (!$self->can($handler_name)){
                    $handler_name = "is_$field";
                }
            }
            return 0 if (!$self->can($handler_name));
            my $value;
            if ($handler_name =~ /_(by|in|on|of)$/){
                $value = $self->${handler_name}($criteria{$field});
                return $value;
            }
            else{
                $value = $self->${handler_name}();
                if (defined($value) && defined($criteria{$field})){
                    if (ref($value) || Scalar::Util::looks_like_number($value)){
                        # We should be able to use '!='
                        return 0 if ( $value != $criteria{$field} );
                    }
                    else{
                        # We should be left with strings, so use 'ne'
                        return 0 if ( $value ne $criteria{$field} );
                    }
                }
                elsif ( defined($value) && !defined($criteria{$field})){
                    return 0;
                }
                elsif ( !defined($value) && defined($criteria{$field})){
                    return 0;
                }
                else{
                    # Both are undef, go to next field
                }
            }
        }
        return 1;
    }

    # TODO: If the Depot/Resource code is ever moved outside of Perf::DataSet (for easier reuse)
    # the two methods below should be put in a subclass of Resource, called Perf::Depot::Resource probably
    # so as to separate Perf::DataSet code from Depot/Resource code
    sub get_busy_dataset_users {
        my $self = $_[0];
        my $dataset_users = $self->get_users( class => 'DataSet');
        return if (!$dataset_users || !@$dataset_users);
        my @busy_datasets = grep { $_->is_busy } @$dataset_users;
        return if (!@busy_datasets);
        return \@busy_datasets;
    }
    sub is_dataset_busy {
        my $self = $_[0];
        my $busy_datasets = $self->get_busy_dataset_users;
        return 0 if (!$busy_datasets || !@$busy_datasets);
        return 1;
    }

    sub is_used {
        my $self = $_[0];
        $self->_prune;
        return $users[$$self]?scalar(@{$users[$$self]}):0;
    }


    sub is_user_of {
        my ($self, $resource) = @_;

        confess('You must specify an argument')
            if (!defined($resource));

        confess('Argument must be a Resource and not a ' . ref($resource))
            if (!$resource->isa('Perf::Depot::Resource'));

        $resource->_prune();

        return $self->found_in($resource->get_users());
    }


    sub is_used_by {
        my ($self,$resource) = @_;
        if (!$resource){
            confess "You must specify an argument";
        }
        if (!$resource->isa('Perf::Depot::Resource')){
            confess "Argument must be a resource and not a " . ref($resource);
        }
        $self->_prune;
        my $r;
        for $r (@{$users[$$self]}){
            return 1 if ($r == $resource);
        }
        return 0;
    }

    sub cleanup :Restricted{
        # NOOP
    }

    sub _destroy :Destroy{
        my $self = $_[0];
        return  if ($$ != $pid[$$self]);
        return  if (threads->tid != $tid[$$self]);
        return  if ($destroying[$$self]);
        $destroying[$$self] = 1;
        $Log->comment("Destroying Resource " . $self->get_description);
        $self->_prune;
        # The call to cleanup below may perform system call. We do not want those
        # to affect the overall exit code as this may affect test results
        local $?;
        eval{ $self->cleanup };
        $Log->comment("Error during cleanup of " . $self->get_description .": $@" ) if ($@);
        map {$_->remove_user($self, recursive=>1)} @{$dependencies[$$self]};
        return;
    }

    # Get rid of stale references
    sub _prune {
        my $self = $_[0];
        my $i = 0;
        my $users_count = scalar(@{$users[$$self]});
        while ($i < $users_count){
            if (!defined($users[$$self]->[$i])){
                $Log->comment("Removing stale 'user' reference from " . $self->get_description);
                splice @{$users[$$self]},$i,1;
                $users_count --;
                next;
            }
            elsif( ref($users[$$self]->[$i]) ){
                confess "Reference is not weak!" if (!Scalar::Util::isweak($users[$$self]->[$i]));
            }
            $i ++;
        }
        return;
    }

    # Remove dependency on a resource that's going away
    sub remove_dependency {
        my ($self, $resource) = @_;

        $Log->debug(  "Removing "
                      . $resource->get_description()
                      . " from "
                      . $self->get_description()
                      . " $name[$$self]s dependencies");

        $dependencies[$$self] =
          [grep { $_ != $resource } @{$dependencies[$$self]}];
    }


    # remove dependency on this resource by a user and remove other
    # resource from list of users
    sub disconnect_user {
        my $self = shift();
        my $user = shift();

        if ($user->is_dependent_on($self)) {
            $user->remove_dependency($self);
        }
        $self->remove_user($user);
    }
}
1;
