# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Appliance - Representation of a NetApp appliance

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Appliance;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log host );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Params::Validate qw(validate);
    use Scalar::Util qw(looks_like_number);
    use Data::Dumper;
    use Perf::Depot;
    use Perf::DataSet::Manager::CIFSManager;
    use List::MoreUtils qw(any none);
    use File::Basename;
    use NACL::C::Network;
    use NACL::C::VserverFcp;
    use NACL::CS::VserverFcpStatus;
    use NACL::Transit;
    use NATE::BaseException qw(:try);

    use namespace::clean;
    use Perf::Depot::Resource;
    use Object::InsideOut qw(Perf::Depot::Resource);

    my @conn :Field :Arg(name=>'conn', mandatory=>0) :Set(set_conn) :Get(get_conn);
    my @san_filer :Field;

    # Track which SMB versions are on/off and which have been
    # purposefully disabled so on destroy, everything we
    # disabled can be reenabled
    # { ver => { is_enabled => 1, was_disabled => 0, original_state => 1}...
    my @smb_feature :Field;


    sub new {
        my $class = shift;
        if ($class ne __PACKAGE__){
            return $class->Object::InsideOut::new(@_);
        }
        my %args = @_;
        #
        # Instantiate the proper Appliance subclass
        # If the filer is a non-ng hosttype, create a Filer
        # if the filer is an ng hosttype, create a Node
        my $name = delete $args{name};
        confess "Argument 'name' is mandatory" if (!$name);
        my $host = host($name);
        my $appliance;

        if ($host->hosttype() =~ /-ng-/) {
            my $hostname = $host->hostname;
            my $node =
              Perf::Depot->get_resource(class => 'Node',
                                                 name  => $hostname);
            if (!$node) {
                $node =
                  Perf::Depot->get_resource(class => 'Node',
                                                     name  => $name);
                if (!$node) {
                    $node =
                      Perf::Depot->create_resource(class => 'Node',
                                                            name  => $hostname,
                                                            %args);
                }
                else {
                    $Log->comment(
                                 "Depot already contains a Node named '$name'");
                }
            }
            else {
                $Log->comment("Depot already contains a Node named '$name'");
            }

            $appliance = $node;

            if ($hostname =~ /^$name(?:\.|$)/){
                # Rename in case the hostname doesn't exactly match
                $appliance->set_name($name);
            }
        }
        else {
            my $filer =
              Perf::Depot->get_resource(class => 'Filer',
                                                 name  => $name);
            if (!$filer) {
                $filer =
                  Perf::Depot->create_resource(class => 'Filer',
                                                        name  => $name,
                                                        %args);
            }
            else {
                $Log->comment("Depot already contains a Filer named '$name'");
            }

            $appliance = $filer;
        }

        return $appliance;
    }

    sub can_ping {
        my ($self, $interface, $pass_thru) = @_;
        confess "You must pass an argument" if (!$interface);
        confess "You must pass a Perf::Depot::Resource::Interface"
            if (! $interface->isa('Perf::Depot::Resource::Interface'));

        my $ci = $self->get_conn();
        my $ret;

        try {
            my $res = NACL::C::Network->ping(command_interface => $ci,
                                             node              => $self->get_name(),
                                             destination       => $interface->get_ip(),
                                             count             => 2,
                                             %$pass_thru,);
            $ret = (('alive' eq $res->state()) ? 1 : 0);
        }
        catch NACL::Exceptions::HostUnreachable with {
            my $ex = shift();
            $Log->comment('Filer to client ping failed. Error is: ' . $ex->text());
            $ret = 0;
        };

        return $ret;
    }

    sub detect_ifconfig {
        confess "You must implement this subroutine in a subclass of " . __PACKAGE__ ;
    }

    sub sync {
        confess "You must implement this subroutine in a subclass of " . __PACKAGE__ ;
    }

    sub flush {
        my ($self) = @_;
        $self->sync;
        sleep 2;
        $self->sync;
        sleep 2;
        $self->sync;
    }

    # TODO: Confirm that no traffic happens on rejected interfaces
    sub detect_ip_connectivity {
        my ($self, $clients) = @_;
        my $appliance_ips = Perf::Depot->get_resources(class=>'Interface', appliance=>$self);
        CLIENT: for my $client (@$clients){
            $Log->comment("Detecting IP connectivity from " . $self->get_description .
                          " and " . $client->get_description);
            my $client_ips = Perf::Depot->get_resources(class=>'Interface', host=>$client);
            CLIENT_IP: for my $client_ip (@$client_ips){
                APPLIANCE_IP: for my $appliance_ip (@$appliance_ips){
                    $Log->debug("Trying " . $client_ip->get_description .
                                  " with " . $appliance_ip->get_description);

                    # Filter out IPs that are not in the same public/private space
                    if ($client_ip->is_public != $appliance_ip->is_public){
                        $Log->debug("Rejecting " . $appliance_ip->get_description . ": not as public or private");
                        next APPLIANCE_IP;
                    }

                    # Public IPs may belong to different subnets, private IPs on the other
                    # hand can't. TODO: how do we detect misconfigured public IPs? Should we also
                    # allow private IPs on different subnets?
                    if (!$appliance_ip->is_public && !$appliance_ip->in_same_subnet($client_ip)){
                        $Log->debug("Rejecting " . $appliance_ip->get_description . ": different subnets");
                        next APPLIANCE_IP;
                    }

                    #
                    # Test the IP both ways
                    #
                    if (!$self->can_ping($client_ip) || !$client->can_ping($appliance_ip)){
                        $Log->debug("Rejecting " . $appliance_ip->get_description .
                                      ": Unable to ping both ways");
                        next APPLIANCE_IP;
                    }

                    #
                    # Check the MTU is the same on both sides
                    #
                    my $appliance_if = $appliance_ip->get_dependency( class=>'Port::Ethernet::Appliance');
                    my $appliance_mtu = $appliance_if->get_mtu;
                    my $client_if = $client_ip->get_dependency( class=>'Port::Ethernet::Host');
                    my $client_mtu = $client_if->get_mtu;
                    if ($appliance_mtu != $client_mtu){
                        $Log->comment("Rejecting " . $appliance_if->get_description .
                                      ": different MTU sizes ($appliance_mtu and $client_mtu respectively)");
                        next APPLIANCE_IP;
                    }
                    #
                    # Record the appliance_ip as a potential connection to this client_ip
                    #
                    $Log->comment("Accepting " . $appliance_ip->get_description);
                    $client_ip->add_accessible($appliance_ip);
                    $appliance_if->add_accessible_by($client_if);
                }#APPLIANCE_IP
            }#CLIENT_IP
        }#CLIENT
    }

    sub get_public_ip {
        my $self = $_[0];
        $self->detect_ifconfig;
        my $ips = Perf::Depot->get_resources(class=>'Interface', appliance=>$self);
        confess "No appliance interface resources found?!" if (!$ips);
        my @public_ips = grep { $_->is_public } @$ips;

        # There might be more than one public IPs if the appliance is a vserver
        # Pick one randomly then
        my $public_ip = $public_ips[int(rand(scalar(@public_ips)))];

        my $public_if = $public_ip->get_port;
        confess $public_if->get_description . " does not belong to an appliance!"
            if (!$public_if->isa('Perf::Depot::Resource::Port::Ethernet::Appliance'));
        return $public_ip;
    }


    sub cifs_init {
        my $self = shift();

        return
          if (defined($smb_feature[$$self]));

        my %smb_enabled = map { $_ => 1 } @{$self->cifs_get_smb_features()};

        foreach my $feature (@Perf::DataSet::Manager::CIFSManager::SMB_VERSIONS,
            'copy_offload') {
            # SMB2.1 can't be independently controlled, it is on if SMB2
            # is on
            my $effective_ver =
              ((looks_like_number($feature) && 2.1 == $feature) ? 2 : $feature);
            my $cur_state = ($smb_enabled{$effective_ver} ? 1 : 0);

            $smb_feature[$$self]->{$feature} = {
                is_enabled     => $cur_state,
                was_disabled   => 0,
                original_state => $cur_state,};
        }
    }


    sub cifs_start {
        confess "You must implement this subroutine in a subclass of " . __PACKAGE__ ;
    }


    sub cifs_share_create {
        my ($self, %args) = @_;

        my $share_name = $args{name};
        confess "You must specify a share name" if (!$share_name);

        my $vol = $args{volume};
        confess('You must specify a volume')
          if (   !defined($vol)
              || !ref($vol)
              || !$vol->isa('Perf::Depot::Resource::Volume'));

        my $cifs_share =
          NACL::STask::VserverCifsShare->create(
                                         command_interface => $self->get_conn(),
                                         vserver           => $self->get_name(),
                                         'share-name'      => $share_name,
                                         volume            => $vol->get_name(),
                                         nacltask_verify   => 1,);

        return;
    }


    sub cifs_share_delete {
        my ($self, $share_name) = @_;

        NACL::STask::VserverCifsShare->purge(
                                         command_interface => $self->get_conn(),
                                         vserver           => $self->get_name(),
                                         'share-name'      => $share_name,
                                         nacltask_verify   => 1,);

        return;
    }


    sub cifs_get_smb_features {
        confess "You must implement this subroutine in a subclass of " . __PACKAGE__ ;
    }


    sub cifs_enable_smb_feature {
        my $self = shift();
        my $feature = shift();

        confess("CIFS SMB$feature has been disabled by another DataSet")
            if ($smb_feature[$$self]->{$feature}->{was_disabled});

        return
            if ($smb_feature[$$self]->{$feature}->{is_enabled});

        $self->_cifs_set_smb_feature(feature => $feature,
                                     enabled => 1,);

        if (looks_like_number($feature) && 2 == $feature) {
            # SMB2.1 can't be independently controlled, it is on if SMB2
            # is on
            $smb_feature[$$self]->{2.1}->{enabled} = 1;
        }
    }


    sub cifs_disable_smb_feature {
        my $self = shift();
        my $feature  = shift();

        $smb_feature[$$self]->{$feature}->{was_disabled} = 1;

        return
            if (!$smb_feature[$$self]->{$feature}->{is_enabled});

        $self->_cifs_set_smb_feature(feature => $feature,
                                     enabled => 0,);

        $smb_feature[$$self]->{$feature}->{is_enabled} = 0;

       # disabling SMB2 also disables SMB2.1 and SMB3/2.2
        if (2 == $feature) {
            foreach my $affected_ver (qw(2.1 3 copy_offload)) {
                $smb_feature[$$self]->{$affected_ver}->{is_enabled} = 0;
                $smb_feature[$$self]->{$affected_ver}->{was_disabled} = 1;
            }
        }
        elsif (3 == $feature) {
            # disabling SMB3 also disables copy_offload, but reenabling SMB3
            # does not automatically reenable copy_offload
            $smb_feature[$$self]->{copy_offload}->{is_enabled} = 0;
            $smb_feature[$$self]->{copy_offload}->{was_disabled} = 1;
        }
    }


    sub _cifs_set_smb_feature {
        confess "You must implement this subroutine in a subclass of " . __PACKAGE__ ;
    }


    sub _cifs_cleanup {
        my $self = shift();

        foreach my $feature (qw(2 3 copy_offload)) {
            my $state = $smb_feature[$$self]->{$feature};

            if ($state->{is_enabled} != $state->{original_state}) {
                if ($state->{original_state}) {
                    $smb_feature[$$self]->{$feature}->{was_disabled} = 0;
                    $self->cifs_enable_smb_feature($feature);
                }
                else {
                    $self->cifs_disable_smb_feature($feature);
                }
            }
        }
    }


    sub get_physical_appliance {
        confess('You must implement this subroutine in a subclass of '
                . __PACKAGE__);
    }


    sub reboot {
        my $self = shift();

        my $transit = NACL::Transit->new(name => $self->get_name());

        my $res = $transit->reboot(inhibit_takeover => 1);

        confess('Reboot failed')
            if ('CLI' ne $res);

        $self->reconnect();
    }


    #
    # San library proxies
    #
    sub _get_san_filer {
        my $self = shift();

        confess(
            'San only interacts with physical appliances (filers or nodes), not '
              . ref($self))
          if (   !$self->isa('Perf::Depot::Resource::Filer')
              && !$self->isa('Perf::Depot::Resource::Node'));

        if (!defined($san_filer[$$self])) {
            require San::SanFiler;
            $san_filer[$$self] = SanFiler->new(filer       => $self->get_name(),
                                               debug_level => 0);

            if ($san_filer[$$self]->error()) {
                my $errmsg = join("\n", $san_filer[$$self]->errmsg());
                confess('Error getting SanFiler for ' .  $self->get_name() . ":\n$errmsg");
            }
        }

        return $san_filer[$$self];
    }

    sub get_san_vserver {
        my $self         = shift();
        my $vserver_name = shift();
        my $san_filer    = $self->_get_san_filer();
        my $san_vserver  = $san_filer->get_vserver_obj(name => $vserver_name,);

        if ($san_filer->error()) {
            my $errmsg = join("\n", $san_filer->errmsg());
            confess(  'error getting san vserver for '
                    . $self->get_description()
                    . ":\n$errmsg");
        }

        return $san_vserver;
    }



    # Retrieve the lib/San representation of a LUN
    # arguments:
    #   1. A Resource::Lun
    sub _get_san_lun {
        my $self = shift();
        my $lun  = shift();

        confess('Argument is not a Lun resource.  Is a: ' . ref($lun))
            if (!$lun->isa('Perf::Depot::Resource::Lun'));

        my $san_volume = $self->_get_san_volume($lun->get_volume());
        my $san_lun = $san_volume->get_lun($lun->get_name());

        return $san_lun;
    }


    sub _get_san_volume {
        confess('You must implement this subroutine in a subclass of '
                . __PACKAGE__);
    }


    #
    # LUN manipulation
    #
    sub lun_create {
        my $self = shift();
        my %arg = @_;

        my $volume = delete($arg{volume});

        # assume volume validated all other params
        confess('Argument is not a Volume resource.  Is a: ' . ref($volume))
            if (!$volume->isa('Perf::Depot::Resource::Volume'));

        my $san_vol = $self->_get_san_volume($volume);

        # San:::Volume wants the LUN name separate from the path
        my $full_path = delete($arg{path});
        $arg{path} = dirname($full_path);
        $arg{name} = basename($full_path);

        my $san_lun = $san_vol->create_lun(%arg);

        confess('Undefined San LUN')
            if (!defined($san_lun));

        if ($san_vol->error()) {
            my $errmsg = join("\n", $san_vol->errmsg());
            confess(  'Error getting San LUN for '
                    . $self->get_description()
                    . ":\n$errmsg");
        }

        return $san_lun;
    }


    sub lun_destroy {
        my $self = shift();
        my $lun  = shift();

        my $san_lun = $self->_get_san_lun($lun);
        $san_lun->destroy();

        confess($san_lun->errmsg())
            if ($san_lun->error());
    }


    sub lun_get_size {
        my $self = shift();
        my $lun  = shift();

        require NACL::CS::Lun;

        my $size = NACL::CS::Lun->fetch(
            command_interface => $self->get_conn(),
            filter => {
                vserver => $self->get_name(),
                path => $lun->get_path(),
            },
            requested_fields => ['size'],
        )->{size};

        return $size;
    }


    sub lun_online {
        my $self = shift();
        my $lun  = shift();

        my $san_lun = $self->_get_san_lun($lun);

        if (!$san_lun->is_online()) {
            confess($san_lun->errmsg())
                if ($san_lun->error());
            $san_lun->online();
        }

        confess($san_lun->errmsg())
            if ($san_lun->error());
    }


    sub lun_map {
        my $self = shift();
        my %args = validate(@_,
                            {lun => {isa => 'Perf::Depot::Resource::Lun'},
                             igroup =>
                               {isa => 'Perf::Depot::Resource::IGroup'},});

        my $san_lun = $self->_get_san_lun($args{lun});

        $san_lun->map($args{igroup}->get_name());

        confess($san_lun->errmsg())
            if ($san_lun->error());
    }


    sub lun_unmap {
        my $self = shift();
        my %args = validate(@_,
                            {lun => {isa => 'Perf::Depot::Resource::Lun'},
                             igroup =>
                               {isa => 'Perf::Depot::Resource::IGroup'},});

        my $san_lun = $self->_get_san_lun($args{lun});

        $san_lun->unmap($args{igroup}->get_name());

        confess($san_lun->errmsg())
          if ($san_lun->error());
    }


    # Retrieve all the LUNs in a volume
    sub get_luns_in_volume {
        my $self = shift();
        my $vol  = shift();

        my $san_vol = $self->_get_san_volume($vol);

        my $luns = $san_vol->get_luns();

        if ($san_vol->error()) {
            confess($san_vol->errmsg());
        }

        return $luns;
    }


    #
    # iGroup
    #
    sub igroup_create {
        confess('You must implement this subroutine in a subclass of '
                . __PACKAGE__);
    }


    # Perform iGroup creation on this Appliance and handle San error checking.
    # Return the name of the iGroup actually created.
    sub _igroup_create {
        my $self      = shift();
        my %args      = @_;
        my $san_filer = $self->_get_san_filer();

        my $san_igroup = $san_filer->create_igroup(%args);

        confess($san_filer->errmsg())
            if ($san_filer->error());

        my $igroup_name = $san_igroup->get_name();

        confess($san_igroup->errmsg())
            if ($san_igroup->error());

        return $igroup_name;
    }


    sub igroup_destroy {
        confess('You must implement this subroutine in a subclass of '
                . __PACKAGE__);
    }

    sub _igroup_destroy {
        my $self      = shift();
        my $igroup    = shift();
        my %args      = @_;
        my $san_filer = $self->_get_san_filer();

        # handle San library API incongruity
        if ($san_filer->isa('San::SanFiler7M')) {
            $san_filer->destroy_igroup($igroup, %args);
        }
        else {
            $san_filer->destroy_igroup(name => $igroup, %args);
        }

        confess($san_filer->errmsg())
          if ($san_filer->error());
    }


    #
    # iSCSI
    #
    sub _get_iscsi_targets {
        my $self      = shift();
        my %pass_thru = @_;
        my $san_filer = $self->_get_san_filer();

        my @targets =
          $san_filer->get_iscsi_targets(%pass_thru);

        if ($san_filer->error()) {
            my $errmsg = join("\n", $san_filer->errmsg());
            confess("Error retrieving iSCSI targets from SanFiler:\n$errmsg");
        }

        confess('No iSCSI targets returned from SanFiler')
          if (0 == scalar(@targets));

        return @targets;
    }

    sub get_iscsi_targets {
        confess('You must implement this subroutine in a subclass of '
                . __PACKAGE__);
    }


    #
    # FCP
    #
    sub fcp_start {
        my ($self, %args) = @_;

        $self->fcp_license();

        try {
            NACL::C::VserverFcp->start(command_interface => $self->get_conn(),
                                       vserver           => $self->get_name(),);
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $ex = shift();

            if ($ex->text() !~ /FCP service is already running./) {
                $ex->throw();
            }
        };

        return;
    }


    sub fcp_license {
        my ($self, %args) = @_;

        try {
            NACL::STask::SystemLicense->add(
                                         command_interface => $self->get_conn(),
                                         feature           => 'FCP',
                                         nacltask_verify   => 1,
                                         nacltask_if_license => 'pass');
        }
        catch NACL::APISet::Exceptions::InvalidParamValueException with {
            my $ex = shift();

            if ($ex->text() !~ /License for package "FCP" is already installed. There is no need to install a Site License./) {
                $ex->throw();
            }
        };
    }

    sub detect_fcp_adapters {
        confess("You must implement this subroutine in a subclass of "
                . __PACKAGE__);
    }


    sub get_fcp_targets {
        confess('You must implement this subroutine in a subclass of '
                . __PACKAGE__);
    }


    sub _get_fcp_targets {
        my $self      = shift();
        my %pass_thru = @_;
        my $san_filer = $self->_get_san_filer();

        my @targets =
          $san_filer->get_targets(%pass_thru);

        if ($san_filer->error()) {
            my $errmsg = join("\n", $san_filer->errmsg());
            confess("Error retrieving FCP targets from SanFiler:\n$errmsg");
        }

        confess('No FCP targets returned from SanFiler')
          if (0 == scalar(@targets));

        return @targets;
    }


    sub _get_target_info {
        my $self                         = shift();
        my $appliance_visible_initiators = $self->get_visible_fcp_initiators();
        my $appliance_targets =
          Perf::Depot->get_resources(class     => 'Port::FCP',
                                              container => $self);
        my %target_info =
          map {
            $_->get_name() => {
                              resource => $_,
                              wwpn     => $_->get_wwpn(),
                              visible_initiators =>
                                $appliance_visible_initiators->{$_->get_name()},
              }
          } @$appliance_targets;

        return \%target_info;
    }


    sub detect_fcp_connectivity {
        my $self        = shift();
        my $clients     = shift();
        my $target_info = $self->_get_target_info();
        my %clients_visible_targets;

        # ensure that each client initiator appears in the
        # list of those logged in to the appliance
      CLIENT: foreach my $client (@$clients) {
            my $host_initiators =
              Perf::Depot->get_resources(class     => 'Port::FCP',
                                                  container => $client);
            my $client_name = $client->get_name();
            my $has_path    = 0;

            foreach my $host_initiator (@$host_initiators) {
                my $host_wwpn = $host_initiator->get_wwpn();

                # loop through targets and record accessibility
              TARGET: foreach my $target (values(%$target_info)) {
                    if (any { $host_wwpn eq $_ }
                        @{$target->{visible_initiators}}) {

                        $target->{resource}->add_accessible_by($host_initiator);

                        # Appliance can see the Host logged in. Do
                        # reverse check
                        $clients_visible_targets{$client_name} =
                          $client->get_visible_fcp_targets()
                          if (!defined($clients_visible_targets{$client_name}));

                        if (none { $target->{wwpn} eq $_ }
                            @{$clients_visible_targets{$client_name}}) {
                            $Log->debug("FCP Target $target->{wwpn} on "
                                . $self->get_description()
                                . " not visible to port $host_wwpn on client $client_name"
                                );
                            next TARGET;
                        }

                        $host_initiator->add_accessible_by($target->{resource});
                        $has_path = 1;
                    }
                    else {
                        $Log->debug(
                                  "$client_name port $host_wwpn not visible to "
                                    . $self->get_description()
                                    . ' port $target->{wwpn} via FCP');
                    }
                }
            }

            confess(  "Client $client_name not visible to "
                    . $self->get_description()
                    . ' via FCP')
              if (!$has_path);
        } ## end foreach my $client (@$clients)
    } ## end sub detect_fcp_connectivity


    # See which initiators are logged into this Appliance's targets
    sub get_visible_fcp_initiators {
        confess('You must implement this subroutine in a subclass of '
                . __PACKAGE__);
    }


    sub _create_fcp_targets {
       confess('You must implement this subroutine in a subclass of '
                . __PACKAGE__);
    }


    sub toggle_fcp_ports {
        confess('You must implement this subroutine in a subclass of '
                . __PACKAGE__);
    }


    sub _toggle_fcp_ports {
        my $self = shift();
        my $set  = shift();

        my $host = host($self->get_name());
        my $apiset = NACL::APISet->new(hostobj   => $host,
                                       category  => 'Node',
                                       interface => 'CLI',
                                       set       => $set,);

        my $response = $apiset->disk_show();
        my $output   = $response->get_parsed_output();

        my @disks = map { $_->{disk} } @$output;
        my %adapter_in_use;
        foreach my $disk (@disks) {
            my ($fc) = $disk =~ /^([[:digit:]][[:alpha:]])\.[[:digit:]]+/;
            confess("Unrecognized disk name format: $disk")
                if (!defined($fc));
            $adapter_in_use{$fc} = 1;
        }

        my $need_reboot = 0;
        my @new_targets;
        $response = $apiset->fcadmin_config();
        $output = $response->get_parsed_output();
        foreach my $port (@$output) {
            my $adapter = $port->{'adapter-name'};

            next
                if ($adapter_in_use{$adapter});

            if ('initiator' eq $port->{type}) {
                if ('online' eq $port->{status}) {
                    try {
                        $apiset->fcadmin_config(d => $adapter);
                    }
                    catch NACL::APISet::Exceptions::UnexpectedOutputException with {
                        $Log->comment(
                                  'API does not fully support taking port down.'
                                    . ' See BURT618047. Ignoring error.');
                    }
                }

                $apiset->fcadmin_config('adapter-name' => $adapter,
                                        type           => 'target');
                $need_reboot = 1;
                push(@new_targets, $adapter);
            }
        }

        if ($need_reboot) {
            $Log->warn('This node must be rebooted to initialize FCP targets');
            $self->reboot();
        }

        return \@new_targets;
    }


    sub panic {
        confess('You must implement this subroutine in a subclass of '
                . __PACKAGE__);
    }


    sub reconnect {
        my $self = shift();
        $self->get_conn()->refresh_command_interface();

        if (defined($san_filer[$$self])) {
            $san_filer[$$self]->reconnect();

            confess($san_filer[$$self]->errmsg())
                if ($san_filer[$$self]->error());
        }
    }


    sub cleanup {
        my $self = shift();

       $self->_cifs_cleanup()
            if (defined($smb_feature[$$self]));
    }
}

1;
