# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Appliance::GX - Base class for GX appliances (Cluster, Vserver, Node)

=head1 DESCRIPTION

This class represent a GX appliance tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Appliance::GX;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::Depot;
    use NACL::STask::Cluster;
    use NACL::STask::NetworkInterface;
    use NACL::STask::NetworkPort;

    use namespace::clean;
    use Perf::Depot::Resource::Appliance;
    use Object::InsideOut qw(Perf::Depot::Resource::Appliance);

    my @cluster :Field :Arg(Name=>'cluster', mandatory=>0)
                       :Type(Perf::Depot::Resource::Cluster)
                       :Set(set_cluster)
                       :Weak;

    sub get_cluster {
        my $self = $_[0];
        my $cl = $cluster[$$self];
        return $cl if ($cl);

        my $ci = $self->get_conn();
        my $cluster_name =
          NACL::STask::Cluster->identity(command_interface => $ci,
                                         requested_fields => ['name'],)
                              ->name();

        $cl =
          Perf::Depot->get_resource(class => 'Cluster',
                                             name  => $cluster_name);
        if (!$cl) {
            $cl =
              Perf::Depot->create_resource(class => 'Cluster',
                                                    name  => $cluster_name,
                                                    execution_node => $self);
            $cl->register;
        }
        $self->set_cluster($cl);
        return $cl;
    }

    sub get_interfaces {
        my $self = $_[0];
        $self->detect_ifconfig;
        return Perf::Depot->get_resources(class=>'Interface::GX', appliance=>$self);
    }

    sub detect_ifconfig {
        my $self  = $_[0];
        my $ci    = $self->get_conn();
        my @ports = NACL::STask::NetworkPort->find(command_interface => $ci);
        my $name  = $self->get_name();

        foreach my $port (@ports) {
            my $port_state = $port->state();
            my $port_name  = $port_state->port();
            my $port_mtu   = $port_state->mtu();
            my $port_node  = $port_state->node();
            my $port_node_obj;

            if ($port_node eq $name) {
                $port_node_obj = $self;
            }
            else {
                $port_node_obj =
                  Perf::Depot->get_resource(class => 'Node',
                                                     name  => $port_node);
            }
            if (!$port_node_obj){
                $Log->comment("No Node resource with name '$port_node'. Skipping port $port_name");
                next;
            }

            my $port_obj =
              Perf::Depot->get_resource(class     => 'Port::Ethernet',
                                                 name      => $port_name,
                                                 container => $port_node_obj);
            if (!$port_obj){
                $port_obj =
                  Perf::Depot->create_resource(
                                                     class => 'Port::Ethernet',
                                                     name  => $port_name,
                                                     mtu   => $port_mtu,
                                                     container => $port_node_obj
                                                     );
                $port_obj->register;
            }
        }

        # NACL ANDs the protocols in the data-protocol arrayref. Since
        # we want only the IP based LIFs, and a LIF can't be iSCSI and
        # one of the NAS protocols at the same time, we have to OR the
        # protocols ourselves.
        my @lifs =
          NACL::STask::NetworkInterface->find(
                                         command_interface => $ci,
                                         filter            => {
                                             vserver         => $name,
                                             'data-protocol' => ['nfs|cifs|iscsi'],
                                           },);

        foreach my $lif (@lifs) {
            my $lif_state     = $lif->state();
            my $lif_name      = $lif_state->lif();
            my $lif_role      = $lif_state->role();
            my $lif_node      = $lif_state->home_node();
            my $lif_port      = $lif_state->home_port();
            my $lif_ipaddress = $lif_state->address();
            my $lif_netmask   = $lif_state->netmask();
            my $lif_is_home   = $lif_state->is_home() eq 'true' ? 1 : 0;

            my $lif_node_obj;
            if ($lif_node eq $name){
                $lif_node_obj = $self;
            }
            else{
                $lif_node_obj = Perf::Depot->get_resource( class=>'Node', name=>$lif_node );
                confess "No Node resource named '$lif_node' found!" if (!$lif_node_obj);
            }

            my $lif_port_obj =
              Perf::Depot->get_resource(class     => 'Port::Ethernet',
                                                 name      => $lif_port,
                                                 container => $lif_node_obj);
            confess "No Port resource called '$lif_port' found for " . $lif_node_obj->get_description
                    if (!$lif_port_obj);

            my $lif_obj = Perf::Depot->get_resource(class=>'Interface',
                                                        name=>$lif_name,
                                                        port=>$lif_port_obj);
            if (!$lif_obj){
                $lif_obj = Perf::Depot->create_resource(class=>'Interface',
                                                            name   =>$lif_name,
                                                            ip     =>$lif_ipaddress,
                                                            mask   =>$lif_netmask,
                                                            role   =>$lif_role,
                                                            is_home=>$lif_is_home,
                                                            port   =>$lif_port_obj,
                                                            appliance=>$self,
                                                            );
                $lif_obj->register;
            }
        }
        confess "Unable to parse interfaces" if (!$self->get_users(class=>'Interface'));
    }



}
1;
