# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Cluster - Representation of a 10-mode cluster

=head1 DESCRIPTION

This class represent a cluster tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Cluster;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use NACL::STask::Cluster;
    # use NACL::STask::Vserver;
    use namespace::clean;
    use Perf::Depot::Resource::Appliance::GX;
    use Object::InsideOut qw(Perf::Depot::Resource::Appliance::GX);

    my @execution_node  :Field
                        :Get(get_execution_node)
                        :Set(set_execution_node)
                        :Arg(Name=>'execution_node', Mandatory=>1)
                        :Type('Perf::Depot::Resource::Node');
    my @nodes           :Field
                        :Name('nodes');

    sub _init :Init {
        my $self = $_[0];
        $nodes[$$self] = [];
        $self->get_nodes();
    }

    sub get_conn {
        my $self = $_[0];
        my $server = $execution_node[$$self];
        if (!$server->has_connection()){
            undef $server;
            confess "Unable to find pingable node" if (!$nodes[$$self]);
            for my $node (@{$nodes[$$self]}){
                if ($node->has_connection()){
                    $Log->comment("Changing execution node of '" . $self->get_description .
                                  "' from '" . $execution_node[$$self]->get_description .
                                  "' to '" . $node->get_description . "'");
                    $server = $node;
                    $execution_node[$$self] = $node;
                    last;
                }
                else{
                    $Log->comment("No connection to '" . $node->get_name . "'");
                }
            }
            confess "Unable to find connected node" if (!$server);
        }
        return $server->get_conn;
    }


    sub get_cluster {
        my $self = $_[0];
        return $self;
    }

    sub get_node_names {
        my ($self) = @_;

        my @nodes =
          NACL::STask::Cluster->nodes(command_interface => $self->get_conn());

        return \@nodes;
    }

    sub get_vserver_names {
        my $self = $_[0];
        my @vservers =
          NACL::STask::Vserver->find_data_vservers(
                                       command_interface => $self->get_conn(),);
        my @names = map { $_->get_one_state_attribute('vserver') } @vservers;

        confess('No vservers found!')
          if (0 == scalar(@names));

        return \@names;
    }

    sub get_vservers {
        #TODO
    }


    sub get_nodes {
        my $self    = shift();
        my %arg = @_;
        my %node_obj;

        if (!$arg{skip_validation}) {
            # Identify all nodes belonging to cluster and register them with Depot
            my $node_names = $self->get_node_names();
            foreach my $node_name (@$node_names) {
                my $node_obj =
                  Perf::Depot->get_resource(class => 'Node',
                                                     name  => $node_name,);

                if (!defined($node_obj)) {
                    $node_obj = Perf::Depot->create_resource(class => 'Node',
                                                                      name  => $node_name);
                    $node_obj->register();
                    $self->add_dependencies($node_obj);
                }

                $node_obj{$node_name} = $node_obj;
            }


            if (scalar(@{$nodes[$$self]}) > scalar(keys(%node_obj))) {
                $self->_process_unjoin(\%node_obj);
            }

            my @node_objs = values(%node_obj);

            $nodes[$$self] = \@node_objs;
        }

        return $nodes[$$self];
    }


    # Process Nodes unjoining from a Cluster
    sub _process_unjoin {
        my $self          = shift();
        my $cur_node_objs = shift();
        my @node_objs     = values(%$cur_node_objs);

        # see which previous Nodes left
        foreach my $prev_node_obj (@{$nodes[$$self]}) {
            my $prev_node_name = $prev_node_obj->get_name();

            if (!defined($cur_node_objs->{$prev_node_name})) {
                my $exec_name = $execution_node[$$self]->get_name();

                # the execution node has left the cluster, use a node
                # from the discovered list
                if ($prev_node_name eq $exec_name) {
                    $execution_node[$$self] = $node_objs[0];
                }

                # remove connections with the missing node
                $prev_node_obj->disconnect_user($self);
            }
        }
    }
}
1;
