# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::DataSet - Representation of a Perf::DataSet

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::DataSet;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use List::MoreUtils qw( any );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use List::MoreUtils qw(all any pairwise);
    use Params::Validate qw(validate SCALAR ARRAYREF);
    use Perf::Depot;
    use Perf::Utils qw(as_string);
    use namespace::clean;
    use Perf::Depot::Resource;
    use Object::InsideOut qw(Perf::Depot::Resource);

    my @dataset :Field
                :Name(dataset)
                :Arg(name =>'dataset', mandatory => 1)
                :Type(Perf::DataSet)
                :Get(get_dataset)
                :Weak;
    #
    # Redirect unknown method calls to the actual Perf::DataSet object
    # Once Perf::DataSet is an OIO the code below won't be necessary
    #
    sub _automethod :Automethod {
        my $self = shift;
        my $d = $dataset[$$self];
        confess 0 if (!$d);
        my @args = @_;
        my $method_name = $_;
        return if (!$d->can($method_name));
        #$Log->comment( __PACKAGE__ . " Automethod called for $method_name" );
        my $handler = sub { return $d->$method_name(@args) };
        #no strict refs;
        #*{__PACKAGE__.'::'.$method_name} = $handler;
        return $handler;
    }

    sub get_used_processes {
        my $self = $_[0];
        my $all_processes = Perf::Depot->get_resources( class=>'Process');
        my @processes = grep { $_->is_used_by($self) } @$all_processes;
        return \@processes if (@processes);
        confess "No processes found" if (!@processes);
    }

    sub get_worker_resources {
        my $self = $_[0];
        my $workers = Perf::Depot->get_resources( class=>'Worker', used_by=>$self);
        confess "No workers found" if (!$workers || !@$workers);
        return $workers;
    }

    sub is_used {
        # Dataset Resources are marked as used for as long as they live
        # even though they are not technically being used by any other
        # resource. This prevents Perf::DataSet resources from being prematurely
        # removed by a call to Perf::Depot->remove_unused
        return 1;
    }

    sub get_lun_resources {
        my ($self, $host) = @_;
        if ($host && !$host->isa('Perf::Depot::Resource::Host')) {
            confess "Argument must be a Perf::Depot::Resource::Host";
        }
        my $lun_objs = [];
        my $host_objs;
        if ($host) {
            $host_objs = [$host];
        }
        else {
            $host_objs = $self->get_client_resources;
        }
        my $vol_obj   = $self->get_volume_resource;
        my $qtree_obj = $self->get_qtree_resource();
        for my $host_obj (@$host_objs) {
            my $lun_names = $self->get_lun_names(host => $host_obj->get_name());

            my @found_lun_objs = map {
                Perf::Depot->get_resource(class  => 'Lun',
                                             name   => $_,
                                             volume => $vol_obj,
                                             qtree  => $qtree_obj,)
            } @$lun_names;

            confess "Expected LUN resources not found in "
              . $vol_obj->get_description
              if (scalar(@found_lun_objs) != scalar(@$lun_names)
                  || (any { !defined($_) } @found_lun_objs));

            push @$lun_objs, @found_lun_objs;
        }
        return $lun_objs;
    } ## end sub get_lun_resources

    sub get_lun_resource {
        my ($self,$host) = @_;
        confess "You must specify an argument" if (!$host);
        my $lun_objs = $self->get_lun_resources($host);
        return $lun_objs->[0];
    }

    sub get_appliance_resource {
        my $self = $_[0];
        my $appliance_name = $self->get_appliance;
        my $appliance_obj = Perf::Depot->get_resource( class=>'Appliance', name=>$appliance_name);
        confess "No Appliance resource found" if (!$appliance_obj);
        return $appliance_obj;
    }

    sub get_filer_resource {
        my $self = $_[0];
        return $self->get_appliance_resource;
    }

    sub get_vserver_resource {
        my $self = $_[0];
        return $self->get_appliance_resource;
    }

    sub get_volume_resource {
        my $self = $_[0];
        my $volname = $self->get_volume;
        my $appliance = $self->get_appliance_resource;
        my $vol_obj = Perf::Depot->get_resource(class=>'Volume', name=>$volname, appliance=>$appliance);
        confess "No volume resource found" if (!$vol_obj);
        return $vol_obj;
    }

    sub get_qtree_resource {
        my $self = $_[0];

        my $volume = $self->get_volume_resource();

        my $qtree_obj = Perf::Depot->get_resource(class  => 'Qtree',
                                                     volume => $volume,);

        return $qtree_obj;
    }

    sub get_client_resources {
        my $self = $_[0];
        my $client_hostnames = $self->get_clients;
        my @resources = map { Perf::Depot->get_resource( class=>'Host', name=>$_ ) } @$client_hostnames;
        confess "No client resources found!" if (!@resources);
        my $resources_found = scalar(@resources);
        my $resources_expected = scalar (@$client_hostnames);
        confess "Not enough client resources found ( $resources_found instead of $resources_expected"
            if ($resources_found != $resources_expected);
        return \@resources;
    }

    sub get_node_resources {
        my $self = $_[0];
        my $nodes = $self->get_nodes;
        my @resources = map { Perf::Depot->get_resource( class=>'Node', name=>$_ ) } @$nodes;
        my $resources_found = scalar(@resources);
        my $resources_expected = scalar (@$nodes);
        confess "Not enough node resources found ( $resources_found instead of $resources_expected"
            if ($resources_found != $resources_expected);
        return \@resources;
    }

    sub _init :Init {
        my $self =$_[0];
        my $dataset_clients = $self->get_clients;

        confess "No clients found for " . $self->get_description ."!"
            if (!$dataset_clients || !@$dataset_clients);

        my $localhost = $self->get_localhost;
        for my $client (@$dataset_clients){
            my $client_obj =
              Perf::Depot->get_resource(class => 'Host',
                                                 name  => $client);
            if (!$client_obj){
                $client_obj = Perf::Depot->create_resource(
                    class => ($client eq $localhost ? 'Localhost' : 'Host'),
                    name => $client,
                    enable_multipathing => $self->enable_multipathing(),
                    enable_alua => $self->enable_alua(),);
                $client_obj->register;
            }
            $self->add_dependencies($client_obj);
        }

        my $appliance_obj =
          Perf::Depot->get_resource(class          => 'Appliance',
                                             name           => $self->get_filer,
                                             conn           => $self->get_fc,
                                             create_on_miss => 1);

        $appliance_obj->register();

        $self->check_nodes($appliance_obj);

        if ($appliance_obj->isa('Perf::Depot::Resource::Node')){
            my $node = $appliance_obj;
            # Create the Vserver and make that the appliance
            my $vserver_name = $self->get_vserver;
            my $cluster = $node->get_cluster;

            my $vserver_names = $cluster->get_vserver_names;
            if (!@$vserver_names){
                confess "No vserver found on " . $cluster->get_description;
            }
            if (!$vserver_name){
                # The user has not specified what vserver to use
                # check if only one exists in the cluster the node
                # belongs to
                if ( scalar(@$vserver_names) == 1 ){
                    $vserver_name = $vserver_names->[0];
                    # TODO: Bad! Find a way to not break encapsulation
                    $dataset[$$self]->{vserver} = $vserver_name;
                }
                else {
                    confess "You must specify a vserver name when several are present in the cluster:\n" .
                            "@$vserver_names";
                }
            }
            else{
                # Check that the vserver_name user entered is valid
                my $match = any {/^$vserver_name$/} @$vserver_names;
                if (!$match){
                    confess "No Vserver called '$vserver_name' found in " . $cluster->get_description;
                }
            }
            $appliance_obj = Perf::Depot->get_resource( class=>'Vserver',
                                                           name=>$vserver_name,
                                                           );
            if (!$appliance_obj){
                $Log->debug("No Vserver named '$vserver_name' found in Depot");
                $appliance_obj = Perf::Depot->create_resource(
                                                           class=>'Vserver',
                                                           name=>$vserver_name,
                                                           cluster=>$cluster);
            }
        }

        # TODO: Bad! Find a way to not break encapsulation
        $dataset[$$self]->{appliance} = $appliance_obj->get_name;

        $self->add_dependencies($appliance_obj);

        my $vol_obj = Perf::Depot->get_resource( class=>'Volume',
                                                    name=>$self->get_volume,
                                                    appliance =>$appliance_obj,
                                                    create_on_miss => 1);
        $vol_obj->register;
        $self->add_dependencies($vol_obj);

        if ($self->get_snapshot){
            my $snap_obj = Perf::Depot->get_resource(class=>'Snapshot',
                                                        name=>$self->get_snapshot,
                                                        volume=>$vol_obj,
                                                        create_on_miss => 1);
            $self->add_dependencies($snap_obj);
        }
        my $qtree_obj;
        if ($self->get_qtree){
            $qtree_obj = Perf::Depot->get_resource( class=>'Qtree',
                                                    name=>$self->get_qtree,
                                                    volume=>$vol_obj,
                                                    create_on_miss => 1);
            $qtree_obj->register;
            $self->add_dependencies($qtree_obj);
        }

        my $rootdir_obj = $self->get_rootdir_resource();
        if ($rootdir_obj){
            $self->add_dependencies($rootdir_obj);
        }
    }


    sub check_nodes {
        my $self          = shift();
        my $appliance_obj = shift();

        if (!defined($appliance_obj)) {
            $appliance_obj =
              Perf::Depot->get_resource(class => 'Appliance',
                                                 name  => $self->get_filer,
                                                 conn  => $self->get_fc,
                                                 create_on_miss => 1);
            $appliance_obj->register();
        }

        if ($appliance_obj->isa('Perf::Depot::Resource::Node')) {
            my $node           = $appliance_obj;
            my $cluster        = $node->get_cluster();
            my $cluster_nodes  = $cluster->get_nodes();
            my %cur_node_objs  = map { $_->get_name() => $_ } @$cluster_nodes;
            my @cur_node_names = sort(keys(%cur_node_objs));
            my $ds_nodes       = $dataset[$$self]->get_nodes();
            my @ds_node_names;

            if (defined($ds_nodes)) {
                @ds_node_names = sort(@$ds_nodes);

                my $node_lists_match =
                  (scalar(@cur_node_names) == scalar(@ds_node_names))
                  && (all { $_ } pairwise { $a eq $b } @cur_node_names,
                      @ds_node_names);

                if (!$node_lists_match) {
                    confess('SAN requires all nodes in cluster to be used')
                        if ($self->is_san());

                    if (scalar(@cur_node_names) < scalar(@ds_node_names)) {
                        $self->_process_unjoin(\%cur_node_objs);
                    }
                    elsif ($dataset[$$self]->use_all_nodes()) {
                        # TODO detect and process a cluster join
                        confess('DataSet does not support adding a node to a cluster.'
                            . ' File a BURT to have support added');
                    }
                }
            }

            if (!$dataset[$$self]->use_all_nodes()) {
                foreach my $node_name (@ds_node_names) {
                    $Log->comment("Validating restrict node: $node_name");
                    my $validated = 0;
                    foreach my $cluster_node (@$cluster_nodes) {
                        $Log->comment("Comparing cluster node: "
                                      . $cluster_node->get_name);
                        if ($cluster_node->get_name() eq $node_name) {
                            $validated = 1;
                            last;
                        }
                    }
                    confess "Node '$node_name' is not a valid node of "
                      . $cluster->get_description()
                      if (!$validated);
                }
            }
            else {
                # TODO: Bad! Find a way to not break encapsulation
                $dataset[$$self]->{nodes} = \@cur_node_names;
            }
        } ## end if ($appliance_obj->isa...
    } ## end sub check_nodes


    sub _process_unjoin {
        my $self          = shift();
        my $cur_node_objs = shift();
        my @relocate_workers;

        foreach my $prev_node_obj (@{$self->get_node_resources()}) {
            my $prev_node_name = $prev_node_obj->get_name();

            if (!defined($cur_node_objs->{$prev_node_name})) {
                # Remove the missing node from the Depot. The Cluster will
                # remain behind. Remove everything up to the Workers that
                # need to be relocated
                my $detached_resources = Perf::Depot->eject_from_to(
                    from => $prev_node_obj,
                    to   => ['Worker', 'Cluster'],);
                push(@relocate_workers,
                    grep { $_->isa('Perf::Depot::Resource::Worker') }
                        @$detached_resources);
            }
        }

        if (0 < scalar(@relocate_workers)) {
            my @node_objs = values(%$cur_node_objs);

            $self->_relocate_workers(\@relocate_workers);
        }
    }


    # Workers need to be ‘relocated.’ They should have had their
    # dependencies on the ejected Mountpoint and Interface removed
    # already. Round-robin through currently existing Node resources
    # assigning new dependencies. This will need to be revisited for
    # more complicated situations with both iSCSI and FCP in
    # Smokejumper. For iSCSI, we'll look at only the Interfaces being
    # used by other workers instead of Mountpoints. For FCP, we'll
    # need to use the FCP Ports instead of ethernet.
    sub _relocate_workers {
        my $self              = shift();
        my $relocate_workers  = shift();
        my $lost_worker_count = scalar(@$relocate_workers);

        return
            if (0 == $lost_worker_count);

        my $appliance = $self->get_appliance_resource();
        my $type = $self->get_type();

        if ($type =~ /^nfs/) {
            $type = 'nfs';
        }
        elsif ($type =~ /^cifs/) {
            $type = 'cifs';
        }
        elsif ($type =~ /^iscsi/) {
            $type = 'iscsi';
        }
        elsif ($type =~ /^fcp/) {
            confess('Submit a BURT to have FCP worker relocation supported');
        }

        my @interfaces;
        my $ifs =
          Perf::Depot->get_resources(class        => 'Interface',
                                              dependent_on => $appliance,);

        foreach my $if (@$ifs) {
            if (grep(/$type/, $if->get_protocols())) {
                push(@interfaces, $if);
            }
        }

        my $lif_count = scalar(@interfaces);

        # ensure Interfaces with least Workers get new Workers first
        @interfaces =
            map  { $_->[0] }
            sort { $a->[1] <=> $b->[1] }
            map  { [$_, $self->_workers_per_lif($_)] } @interfaces;

        # Assign new Interface dependencies to Workers. Mountpoints
        # will be assigned or created later in NASManager->mount()
        foreach my $i (0 .. ($lost_worker_count - 1)) {
            my $interface = $interfaces[$i % $lif_count];

            $relocate_workers->[$i]->add_dependencies([$interface]);
        }
    } ## end sub _relocate_workers


    sub _workers_per_lif {
        my $self = shift();
        my $lif  = shift();

        my $workers = $lif->get_users(class => 'Worker',
                                      name  => $self->get_name());

        confess('No Workers found using Interface: ' . $lif->get_description())
            if (0 == scalar(@$workers));

        return scalar(@$workers);
    }


=head2 get_mountpoint_resources

Retrieve the Mountpoint objects associated with this Perf::DataSet

=cut

    sub get_mountpoint_resources {
        my ($self, $mountpoint_class) = @_;

        my $mountpoints =
          Perf::Depot->get_resources(class   => $mountpoint_class,
                                              used_by => $self,);

        return $mountpoints;

    }


=head2 get_snapshot_resource

Retrieve the Snapshot resource for this Perf::DataSet

=cut

sub get_snapshot_resource {
    my $self = shift();

    return
      Perf::Depot->get_resource(class  => 'Snapshot',
                                   name   => $self->get_snapshot(),
                                   volume => $self->get_volume_resource(),);
}


=head2 get_rootdir_resource

Retrieve the RootDir resource for this Perf::DataSet

=cut

sub get_rootdir_resource {
    my $self = shift();

    return
      Perf::Depot->get_resource(class    => 'RootDir',
                                   name     => $self->get_root_dir(),
                                   volume   => $self->get_volume_resource(),
                                   qtree    => $self->get_qtree_resource(),
                                   snapshot => $self->get_snapshot_resource(),);
}


=head2 get_raw_device_resources

Retrieve the raw device Resource objects created during the mount of a
SAN Perf::DataSet.

=head3 Arguments

=over

=item I<addition_uid> - The UID of the Addition that initiated the
device creation (Optional)

=back

=head3 Example

$dataset->get_raw_device_resources(addition_uid => $add_uid);

=cut

    sub get_raw_device_resources {
        my $self = shift();

        my %params = @_;

        my $vol_obj   = $self->get_volume_resource();
        my $qtree_obj = $self->get_qtree_resource();

        # the created LUNs will be named after the UID of this
        # Addition
        my $lun_names = $self->get_lun_names(%params);

        my @lun_objs = map {
            Perf::Depot->get_resource(class  => 'Lun',
                                               name   => $_,
                                               volume => $vol_obj,
                                               qtree  => $qtree_obj,);
        } @$lun_names;

        # raw devices should map to exactly one LUN
        my @raw_devs = map {
            Perf::Depot->get_resource(class => 'SANDevice',
                                               lun   => $_,)
        } @lun_objs;

        return \@raw_devs;
    } ## end sub get_raw_device_resources
}
1;
