# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Filer - Representation of a 7G filer

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Filer;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw($Log subtest);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::Depot;
    use Net::IP qw(ip_get_mask
                   ip_bintoip
                  );
    use NACL::C::Ifconfig;
    use NACL::C::Options;
    use NACL::C::VserverFcp;
    use NACL::C::VserverNfs;
    use NACL::MTask::CIFS;
    use NACL::STask::SystemLicense;
    use NACL::STask::VserverCifsShare;
    use NATE::BaseException qw(:try);
    use NATE::ParamSet;

    use namespace::clean;
    use Perf::Depot::Resource::Appliance;
    use Object::InsideOut qw(Perf::Depot::Resource::Appliance);


    sub _init :Init {
        my $self = $_[0];
        my $conn = $self->get_conn;
        if (!ref($conn) || !$conn->isa('NACL::C::SystemNode')){
            confess "Argument 'conn' must be a NACL::C::SystemNode and not a " . ref($conn);
        }
        my $name = $conn->hostrec->hostname;
        $self->set_name($name);
    }

    sub get_fc {
        my $self = $_[0];
        return $self->get_conn;
    }

    sub set_fc {
        my ($self,$fc) = $_[0];
        confess "Argument must be a 'NACL::C::SystemNode'"
          if (!$fc || !ref($fc) || !$fc->isa('NACL::C::SystemNode'));
        return $self->set_conn($fc);
    }

    sub detect_ifconfig {
        my $self = $_[0];

        #
        # Get filer IPs from ifconfig command
        #
        my @ifs =
          NACL::C::Ifconfig->find(command_interface => $self->get_fc());
        foreach my $if (@ifs) {
            my $if_name     = $if->interface();
            my $if_loopback = ($if_name =~ /^lo/);
            my $if_mgmt     = ($if_name =~ /^e0M/);

            if ($if_loopback || $if_mgmt) {
                $Log->comment("Skipping Port '$if_name' because it's a "
                                . (  $if_loopback ? 'loopback'
                                   : $if_mgmt     ? 'management'
                                   : q{}) . ' port');
                next;
            }

            my $if_state = $if->state();
            my $if_up    = $if_state->flags()->[0]->{UP};

            if (!$if_up) {
                $Log->comment("Skipping Port '$if_name' because it's not UP");
                next;
            }

            my $current_if =
              Perf::Depot->get_resource(class     => 'Port::Ethernet',
                                                 name      => $if_name,
                                                 container => $self);
            if (!$current_if) {
                my $if_mtu = $if_state->mtu();
                $current_if =
                  Perf::Depot->create_resource(
                                                      class => 'Port::Ethernet',
                                                      name  => $if_name,
                                                      container => $self,
                                                      mtu       => $if_mtu);
                $current_if->register;
            }

            my $ip = $if_state->ipv4()->[0];

            next
                if ($ip =~ /^127\.0/);

            my $mask = $if_state->netmask();

            if ($mask =~ /0x(\w\w)(\w\w)(\w\w)(\w\w)/) {
                $mask = hex($1) . "." . hex($2) . "." . hex($3) . "." . hex($4);
            }
            elsif ($mask =~ /^\d+$/) {
                $mask = ip_bintoip(ip_get_mask($mask, 4), 4);
            }
            else {
                confess "Unable to parse '$mask'";
            }

            my $ip_obj =
              Perf::Depot->get_resource(class => 'Interface',
                                                 name  => $ip,
                                                 port  => $current_if);
            if (!$ip_obj) {
                $ip_obj =
                  Perf::Depot->create_resource(class => 'Interface',
                                                        name  => $ip,
                                                        ip    => $ip,
                                                        port  => $current_if,
                                                        mask  => $mask);
                $ip_obj->register;
            }

            my $partner = $if_state->partner();
            $current_if->add_failover_target($partner)
                if (defined($partner));
        }

        confess 'Unable to parse filer IPs'
          if (!$self->get_users(class => 'Interface'));
    }


    sub sync {
        my ($self) = @_;
        $self->get_fc()->apiset()->sync();
    }

    sub nfs_start {
        my ($self,%args) = @_;
        my $nfs_version = $args{version};

        my $ci = $self->get_fc();
        my $model_name = $ci->apiset()
                            ->sysconfig('model-information' => 1)
                            ->get_parsed_output()
                            ->[0]
                            ->{model_name};

        #
        # License and Start NFS
        #
        if ($model_name =~ /^SA\d/) {
            $Log->comment("Found SA model, using flexcache_nfs");
            NACL::STask::SystemLicense->add(command_interface => $ci,
                                            feature => 'flexcache_nfs',
                                            nacltask_verify => 1,);
        }
        else {
            NACL::STask::SystemLicense->add(command_interface => $ci,
                                            feature           => 'nfs',
                                            nacltask_verify   => 1,);
        }

        NACL::C::VserverNfs->on(command_interface => $ci);

        # Enable nfsv4 on the filer, if nfsv4 is requested.
        if ($nfs_version && $nfs_version eq 4) {
            NACL::C::Options->option(command_interface => $ci,
                                     'option-name' => 'nfs.v4.enable',
                                     'option-value' => 'on',);
        }
    }


    sub add_user {
        my ($self, $resource) = @_;
        my $was_added = $self->SUPER::add_user($resource);
        return if (!$was_added);
        if ($resource->isa('Perf::Depot::Resource::DataSet')){
            my $dataset_type = $resource->get_type;
            if ( $dataset_type eq 'iscsi'){
                # If this is the first Perf::DataSet added, turn off all iSCSI interfaces
                # The IP manager will turn on the ones that are needed later
                my $dataset_users = $self->get_users( class => 'DataSet');
                my $dataset_users_count = scalar(@$dataset_users);
                if ( $dataset_users_count == 1 ){
                    $self->disable_iscsi_interfaces;
                }
                $self->iscsi_start;
            }
            elsif ($dataset_type eq 'nfs'){
                $self->nfs_start;
            }
            elsif( $dataset_type eq 'nfsv4' ){
                $self->nfs_start(version=>4);
            }
            elsif($dataset_type eq 'fcp') {
                $self->fcp_start();
            }
            elsif( $dataset_type =~ /cifs/ ){
                $self->cifs_start;
            }
        }
        return $was_added;
    }


    #
    # CIFS
    #
    sub cifs_start {
        my $self = $_[0];

        # translate environment variables to CIFS setup
        # parameters
        my $params = NATE::ParamSet->get_global();
        my %env_vars = ('FILER_DNS_DOMAINNAME'  => 'dns-domainname',
                        'FILER_DNS_SERVER_LIST' => 'dns-server-list',
                        'CIFS_DOMAIN_LONG'      => 'domain',
                        'AD_SERVER_PASSWORD'    => 'domain-pass',
                        'AD_SERVER_USER'        => 'domain-user',
                        'NTP_SERVER_IP'         => 'timed-server-list');
        my %setup_args = (command_interface => $self->get_conn(),
                          'auth-type'       => 'ad',
                          'create-admin'    => 'no',
                          'dns-enable'      => 1,
                          'cifs-server'     => $self->get_name(),
                          'passwd-group'    => 'yes',
                          'timed-enable'    => 1,
                          if_exists         => 'reuse',);
        while (my ($env_var, $cifs_var) = each(%env_vars)) {
            my $env_val = $params->get($env_var);

            confess(  "Environment variable '$env_var' not set."
                    . " Can't configure CIFS")
              if (!defined($env_val));

            $setup_args{$cifs_var} = $env_val;
        }

        NACL::MTask::CIFS->setup_7Mode(%setup_args);

        $self->cifs_init();

        return;
    }


    # SMB1 is always on, SMB3 is not supported in 7-mode
    sub cifs_get_smb_features {
        my $self = shift();
        my @versions = (1);

        my $smb2 =
            NACL::CS::Options->fetch(command_interface => $self->get_fc(),
                                     filter => {name => 'cifs.smb2.enable'},)
                             ->value();

        push(@versions, 2)
            if ('on' eq $smb2);

        return \@versions;
    }


    sub _cifs_set_smb_feature {
        my $self    = shift();
        my %arg     = @_;

        if (2 == $arg{version}) {
            NACL::C::Options->option(
                                 command_interface => $self->get_conn(),
                                 'option-name'     => 'cifs.smb2.enable',
                                 'option-value' => ($arg{enabled} ? 'on' : 'off'),
                                 );
        }
    }

    sub get_nfsv4_domain_id {
        my ($self) = @_;

        my $domain_id =
          NACL::CS::Options->fetch(command_interface => $self->get_fc(),
                                   filter => {name => 'nfs.v4.id.domain'},);
        return $domain_id->value();
    }

    sub get_physical_appliance {
        my $self = shift();

        return $self;
    }


    #
    # San library interaction
    #
    sub _get_san_volume {
        my $self = shift();
        my $vol  = shift();

        confess(  "Didn't provide a Resource::Volume::SevenMode.\n"
                . 'Got a: '
                . ref($vol))
          if (   !defined($vol)
              || !$vol->isa('Perf::Depot::Resource::Volume::SevenMode'));

        my $san_filer  = $self->_get_san_filer();
        my $san_volume = $san_filer->get_vol($vol->get_name());

        if ($san_filer->error()) {
            my $errmsg = join("\n", $san_filer->errmsg());
            confess(  'Error getting San Volume for '
                    . $self->get_description()
                    . ":\n$errmsg");
        }

        return $san_volume;
    }


    #
    # iGroup
    #
    # Create an iGroup on this Filer.  The name given may manipulated,
    # so return the name of the iGroup actually created.
    sub igroup_create {
        my $self = shift();
        my %args = @_;

        return $self->_igroup_create(%args);
    }

    sub igroup_destroy {
        my $self   = shift();
        my $igroup = shift();

        $self->_igroup_destroy($igroup);
    }


    #
    # iSCSI
    #
    sub iscsi_start {
        my ($self,%args) = @_;

        $self->iscsi_license();

        my $san_filer = $self->_get_san_filer();

        $san_filer->iscsi_start();

        confess($san_filer->errmsg())
            if ($san_filer->error());

        return;
    }


    sub iscsi_interface_status {
        my $self       = shift();
        my $interfaces = $self->get_fc()
                              ->apiset()
                              ->iscsi_interface_show()
                              ->get_parsed_output();
        my %interface_status;

        @interface_status{keys(%$_)} = values(%$_)
            foreach (@$interfaces);

        return \%interface_status;
    }


    sub disable_iscsi_interfaces {
        my $self = $_[0];
        # iscsi command doesn't exist if iscsi isn't licensed
        $self->iscsi_license;

        my $san_filer        = $self->_get_san_filer();
        my $interface_status = $self->iscsi_interface_status();

        while (my ($if, $status) = each(%$interface_status)) {
            if ('enabled' eq $status) {
                $san_filer->iscsi_interface_disable($if);

                confess($san_filer->errmsg())
                    if ($san_filer->error());
            }
        }
    }


    sub iscsi_license {
        my ($self,%args) = @_;

        NACL::STask::SystemLicense->add(command_interface => $self->get_fc(),
                                        feature           => 'iscsi',
                                        nacltask_verify   => 1,);

        return;
    }


    sub get_iscsi_targets {
        my $self = shift();

        return $self->_get_iscsi_targets();
    }


    #
    # FCP
    #
    sub get_fcp_targets {
        my $self = shift();

        return $self->_get_fcp_targets();
    }


    sub get_visible_fcp_initiators {
        my $self = shift();
        my %inits;

        my $out = $self->get_fc()
                       ->apiset()
                       ->fcp_show_initiator()
                       ->get_raw_output();
        my $current_hba;

        foreach my $line (split(/\r?\n/, $out)) {
            if ($line =~ /^Initiators connected on adapter ([[:digit:]]+[[:alpha:]]):/) {
                $current_hba = $1;
            }
            elsif ($line =~ /^([[:xdigit:]:]{23}) .*/) {
                confess('Current HBA not defined')
                    if (!defined($current_hba));
                (my $init = $1) =~ s/://g;
                push(@{$inits{$current_hba}}, $init);
            }
        }

        return \%inits;

        # the below code uses the San libraries but fails when no
        # igroups are set

        # my $sanfiler = $self->_get_san_filer();
        # it makes no sense, but calling get_fcp_initators() without
        # arguments returns a list of filer targets
        # my $targets = $sanfiler->get_fcp_initiators();

        # foreach my $target (@$targets) {
        #     # then you pass a target name as an initiator arg to get the
        #     # list of logged in initiators
        #     my $initiators = $sanfiler->get_fcp_initiators(init => $target);
        #     push(@inits, @$initiators);
        # }
    }


    sub detect_fcp_adapters {
        my $self = shift();

        my $san_filer = $self->_get_san_filer();
        my %adapter_info = $san_filer->fcp_adapters_list_info();

        confess('Error getting FCP targets: ' . $san_filer->errmsg())
            if ($san_filer->error());

        foreach my $adapter (keys(%adapter_info)) {
            if ('ONLINE' ne $adapter_info{$adapter}->{Status}) {
                $Log->debug("Rejecting $adapter: not online");
                next;
            }

            (my $wwpn = $adapter_info{$adapter}->{'FC Portname'})
                =~ s/[[:xdigit:]:]{23} \(([[:xdigit:]]{16})\)/$1/;

            my %fc_id = (class     => 'Port::FCP',
                         name      => $adapter,
                         wwpn      => $wwpn,
                         container => $self,);
            my $fc_hba = Perf::Depot->get_resource(%fc_id);

            if (!$fc_hba) {
                $fc_hba = Perf::Depot->create_resource(%fc_id);
                $fc_hba->register();
            }
        }
    }


    sub _create_fcp_targets {
        my $self = shift();

        my $code = sub {
            $self->toggle_fcp_ports();
        };

        subtest($code, '-runid' => $self->get_name() . '_toggle_fcp');
    }


    sub toggle_fcp_ports {
        my $self = shift();

        return $self->_toggle_fcp_ports('7Mode');
    }
}
1;
