# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Host - Representation of a host

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Host;
{
    use strict;
    use warnings;
    use Carp qw(confess);
    use Tharn qw( $Log host );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Perf::Depot;
    use NACL::C::Client;
    use NACL::C::Client::Ping;
    use NATE::BaseException qw(:try);

    use namespace::clean;
    use Perf::Depot::Resource;
    use Object::InsideOut qw(Perf::Depot::Resource);
    use Perf::Depot::Resource::Host::Unix;
    use Perf::Depot::Resource::Host::Windows;

    my @connection  :Field
                    :Name(connection)
                    :Get(get_connection);
    my @hosttype    :Field
                    :Name(hosttype)
                    :Get(get_hosttype);
    my @enable_multipathing :Field
                            :Arg(enable_multipathing)
                            :Name(enable_multipathing)
                            :Get(enable_multipathing);
    my @enable_alua         :Field
                            :Arg(enable_alua)
                            :Name(enable_alua);
    my @configured_iscsi    :Field
                            :Get(iscsi_was_configured);
    my @started_iscsi       :Field
                            :Get(iscsi_was_started);
    my @configured_fcp      :Field;
    my @sanclients          :Field
                            :Name(sanclients);
    my @sanclient_conn      :Field
                            :Name(sanclient_conn);
    my @san_dev_mgr         :Field
                            :Set(_set_san_dev_mgr);
    my @initiator_name      :Field;

    sub new {
        my $class = shift;
        if ($class ne __PACKAGE__){
            return $class->Object::InsideOut::new(@_);
        }
        my %args = @_;
        my $name = $args{name};
        my $host = host($name);
        my $host_type = $host->hosttype;
        if ($host_type =~ /unix/i){
            return Perf::Depot::Resource::Host::Unix->new(%args);
        }
        elsif($host_type =~ /windows/i){
            return Perf::Depot::Resource::Host::Windows->new(%args);
        }
        else{
            confess "Unsupported hosttype: '$host_type'";
        }
    }

    sub _init :Init {
        my $self = $_[0];
        $sanclients[$$self] = [];
        $configured_iscsi[$$self] = 0;
        $configured_fcp[$$self] = 0;

        $self->_populate;
    }

    sub _populate {
        my $self = $_[0];
        my $ch = host($self->get_name);
        $hosttype[$$self] = $ch->hosttype();
        $connection[$$self] = NACL::C::Client->new(hostrec => $ch);
    }

    sub set_hosttype :Restricted{
        my ($self,$type) = @_;
        $hosttype[$$self] = $type;
    }

    sub run_cmd {
        my $self = shift();
        my $cmd  = shift();
        my %pass_thru = @_;

        return $self->get_connection()->execute(
            command => $cmd,
            %pass_thru,
        );
    }

    sub detect_ifconfig {
        confess "You must implement this method in a subclass";
    }

    sub detect_fcp_adapters {
        my $self      = shift();
        #my $sanclient = $self->_get_sanclient(protocol => 'fcp');
        #my $inits     = $sanclient->get_initiators();

        # the San libraries do not return the operational state of an HBA

        my $sanlun = $self->_get_sanlun(protocol => 'fcp');
        my $adapter_info = $sanlun->display_adapters('fcp');

        my %initiator;
        my $cur_adapter;
        foreach my $line (@$adapter_info) {
            next
                if (q{} eq $line);

            my ($k, $v) = split(/:[[:space:]]+/, $line, 2);

            if ('adapter name' eq $k) {
                $cur_adapter = $v;
            }
            elsif (defined($cur_adapter)) {
                $initiator{$cur_adapter}->{$k} = $v;
            }
            else {
                confess("Unrecognised line: $line in sanlun output:\n"
                        . join("\n", @$adapter_info));
            }
        }

        while (my ($hba, $info) = each(%initiator)) {
            if ('Operational' ne $info->{'port state'}) {
                $Log->debug("Rejecting $hba: not operational");
                next;
            }

            my %fc_id = (class     => 'Port::FCP',
                         name      => $hba,
                         wwpn      => $info->{WWPN},
                         container => $self,);
            my $fc_hba = Perf::Depot->get_resource(%fc_id);

            if (!defined($fc_hba)) {
                $fc_hba = Perf::Depot->create_resource(%fc_id);
                $fc_hba->register();
            }
        }
    }


    sub can_ping {
        my $self = shift();
        my $ipaddr = shift();

        confess('You must pass an argument')
            if (!defined($ipaddr));
        confess('You must pass a Perf::Depot::Resource::Interface')
            if (!$ipaddr->isa('Perf::Depot::Resource::Interface'));

        my $ret;

        try {
            my $res =
              NACL::C::Client::Ping->ping(
                                       command_interface => $connection[$$self],
                                       destination       => $ipaddr->get_ip(),
                                       count             => 1,);
            $ret = ((1 eq $res->received()) ? 1 : 0);
        }
        catch NATE::BaseException with {
            # NACL::Exceptions::HostUnreachable doesn't get thrown for
            # some reason when host doesn't respond
            my $ex = shift();
            $Log->comment('Client to filer ping failed. Error is: ' . $ex->text());
            $ret = 0;
        };

        return $ret;
    }


    #
    # San library interactions
    #
    sub _get_sanclient :Restricted {
        my ($self,%pass_thru) = @_;

        confess('protocol not specified')
            if (!defined($pass_thru{protocol}));

        $pass_thru{host}                     ||= $self->get_name();
        $pass_thru{mp_type}                  ||= $self->get_mp_type();
        $pass_thru{override_vm_manufacturer} ||= 1;
        $pass_thru{alua_support}             ||= $enable_alua[$$self];

        my $sanclient;
        if ($sanclients[$$self]){
            SANCLIENT: for my $e (@{$sanclients[$$self]}){
                # Make a copy;
                my @a = @{$e};
                # Grab contents
                my $s = shift @a;
                my %s_pt = @a;
                # Compare the pass_thru arguments
                for my $k (keys %pass_thru){
                    next SANCLIENT if ( !exists($s_pt{$k}) || $s_pt{$k} ne $pass_thru{$k});
                }
                for my $l (keys %s_pt){
                    next SANCLIENT if ( !exists($pass_thru{$l}) || $pass_thru{$l} ne $s_pt{$l} );
                }
                $Log->debug("Found matching cached SanClient");
                $sanclient = $s;
            }
        }

        if (!$sanclient) {
            my $conn = $sanclient_conn[$$self];

            if ($conn) {
                $pass_thru{connection} = $conn;
            }

            require San::SanClient;

            $sanclient = SanClient->new(%pass_thru);
            push @{$sanclients[$$self]},[$sanclient,%pass_thru];
            $sanclient_conn[$$self] = $sanclient->get_connection;
        }

        if ($sanclient->error) {
            my $errmsg = join("\n", $sanclient->errmsg());
            confess("Error getting a SanClient:\n$errmsg");
        }

        return $sanclient;
    }

    sub _get_san_devmgr {
        my $self = shift();
        my %pass_thru = @_;

        if (!defined($san_dev_mgr[$$self])) {
            my $san_client = $self->_get_sanclient(%pass_thru);

            $san_dev_mgr[$$self] =
                $san_client->create_devmgr(load_devices => 0);

            confess($san_client->errmsg())
                if ($san_client->error());
        }

        return $san_dev_mgr[$$self];
    }

    # sanlun is a NetApp utility for querying the Host OS to discover
    # information about raw devices and the LUNs they're associated
    # with
    sub _get_sanlun {
        my $self      = shift();
        my %pass_thru = @_;

        my $san_client = $self->_get_sanclient(%pass_thru);

        my $sanlun = $san_client->get_sanlun();

        confess($san_client->errmsg())
          if ($san_client->error());

        return $sanlun;
    }

    sub get_lun_type {
        my $self = $_[0];
        confess "Field hosttype not populated for " . $self->get_description if (!$hosttype[$$self]);

        my $lun_type = $hosttype[$$self];
        # Figure out type of LUN that should be created
        # Copy/Paste of code found in San::San
        $lun_type =~ s/unix-solaris/solaris/;    # strip off unix-
        if (grep /(linux|sanblaze)/, $lun_type)
        {
            $lun_type = "linux";
        }
        if (grep /hpux/, $lun_type)
        {
            $lun_type = "hpux";
        }
        if (grep /aix/, $lun_type)
        {
            $lun_type = "aix";
        }
        return $lun_type;
    }

    # This is currently broken.  Running it once works.  Running it a
    # second time fails as it appears to expect MP to be turned on if
    # asked to turn it off.

    # sub disable_multipathing {
    #     my $self = shift();
    #     my %args = @_;
    #     my $sanclient = $self->_get_sanclient(mp_type => 'DMMP', %args);
    #     my $dmmp = $sanclient->{mp};
    #     $dmmp->disable( '-conn' => $sanclient->get_connection );
    #     if ($dmmp->error){
    #         my $errmsg = join("\n", $dmmp->errmsg());
    #         confess $errmsg;
    #     }
    #     return;
    # }

    sub add_user {
        my ($self, $resource) = @_;
        my $was_added = $self->SUPER::add_user($resource);
        return if (!$was_added);

        if (   $resource->isa('Perf::Depot::Resource::DataSet')
            && $resource->is_san()) {
            $enable_multipathing[$$self] = $resource->enable_multipathing();

            if (!$enable_multipathing[$$self]) {
                $self->disable_multipathing();
            }
        }

        return $was_added;
    }

    sub get_initiators {
        my $self = shift();
        my %pass_thru = @_;

        my $san_client = $self->_get_sanclient(%pass_thru);

        my $initiators = $san_client->get_initiators();

        confess($san_client->errmsg())
            if ($san_client->error());

        confess(  'Host '
                . $self->get_name() . ' has '
                . scalar(@$initiators)
                . ' initiators instead of 1')
          if (1 != scalar(@$initiators) && 'iscsi' eq $pass_thru{protocol});

        return $initiators;
    }


    sub get_initiator_name {
        my $self      = shift();
        my %pass_thru = @_;

        if (!defined($initiator_name[$$self])) {
            my $initiators = $self->get_initiators(%pass_thru);

            $initiator_name[$$self] = $initiators->[0]->get_initiator();
        }

        return $initiator_name[$$self];
    }


    # Retrieve the name of the LUN the Host believes a raw device
    # represents
    sub get_lun_path_for_dev {
        my $self = shift();
        my $dev  = shift();

        confess("Expecting a 'SANDevice', got a '" . ref($dev) . "'")
            if (!$dev->isa('Perf::Depot::Resource::SANDevice'));

        my $san_dev_name =
          $self->get_san_dev_name(uniq     => $dev->get_name(),
                                  protocol => $dev->get_protocol(),);
        my $sanlun   = $self->_get_sanlun(protocol => $dev->get_protocol());
        my $lun_path = $sanlun->get_lun_path(device => $san_dev_name);

        confess($sanlun->errmsg())
            if ($sanlun->error());

        return $lun_path;
    }

    # Force Host to rescan for newly created devices
    sub discover_san_devices {
        my $self = shift();
        my %pass_thru = @_;

        my $san_client = $self->_get_sanclient(%pass_thru);

        # SanClient->discover() doesn't do anything for FCP
        # because SanClient->configure() will eventually call for
        # a rescan, but we want to check to ensure that the device
        # has actually shown up before we attempt to configure
        $san_client->_rescan_scsi_bus(option => 'discover_LUNs', %pass_thru);

        confess($san_client->errmsg())
            if ($san_client->error());

        $san_client->discover();

        confess($san_client->errmsg())
            if ($san_client->error());
    }

    # Retrieve a list of unused SAN raw devices
    sub get_free_san_devs {
        my $self = shift();
        my %args = @_;

        my $lun = delete($args{lun});

        confess('Argument is not a Lun resource.  Is a: ' . ref($lun))
            if (!defined($lun) || !$lun->isa('Perf::Depot::Resource::Lun'));

        my $san_devmgr = $self->_get_san_devmgr(protocol => $args{protocol});

        # setup arguments for San:::DevMgr
        $args{lun}    = $lun->get_name();
        $args{volume} = $lun->get_volume()->get_name();

        my @dev_names =
          $san_devmgr->get_free_devices(refresh => 1,
                                        %args,);

        confess($san_devmgr->errmsg())
            if ($san_devmgr->error());

        confess('No SAN devices found')
            if (0 == scalar(@dev_names));

        return @dev_names;
    }

    sub get_unique_dev_name {
        confess('You must implement this method in a subclass of ' . __PACKAGE__);
    }

    sub get_san_dev_name {
        confess('You must implement this method in a subclass of ' . __PACKAGE__);
    }

    sub san_dev_create {
        my $self = shift();
        my %args = @_;

        my $san_devmgr = $self->_get_san_devmgr(protocol => $args{protocol});

        my $san_dev = $san_devmgr->create_device(name => $args{name});

        confess($san_devmgr->errmsg())
            if ($san_devmgr->error());
    }

    # Retrieve the actual size of a SAN raw device
    sub get_san_dev_size {
        confess('You must implement this method in a subclass of ' . __PACKAGE__);
    }


    sub get_mp_type {
        confess('You must implement this method in a subclass of ' . __PACKAGE__);
    }


    #
    # iSCSI
    #
    sub iscsi_start {
        my $self = $_[0];
        # TODO: we ought to record whether iscsi was already started
        # if it wasn't, we ought to stop it once this object goes away

        my $sanclient = $self->_get_sanclient(protocol => 'iscsi');
        my $ret = $sanclient->_iscsi_service(action => 'start');
        if ($sanclient->error){
            my $errmsg = join "\n", $sanclient->errmsg();
            confess "Unable to start iscsi on " . $self->get_description .":\n$errmsg";
        }

        return;
    }


    sub san_cleanup {
        my $self = shift();
        my $proto = shift();
        my $sanclient = $self->_get_sanclient(protocol => $proto);

        if ($enable_multipathing[$$self]) {
            my $mp = $sanclient->{mp};
            $mp->disable(conn => $sanclient->get_connection());

            $Log->warn('Error disabling multipathing: ' . $mp->errmsg())
                if ($mp->error());
        }

        # Log out initiators and remove targets
        $sanclient->deconfigure(clean_level => 2);

        if ($sanclient->error()) {
            my $errmsg = join("\n", $sanclient->errmsg());
            $Log->warn(  "Unable to clean up $proto on "
                       . $self->get_description
                       . ":\n$errmsg");
        }

        $sanclient->_rescan_scsi_bus(protocol => $proto,
                                     option   => 'remove_LUNs');
        $Log->warn('Error rescanning SCSI bus: ' . $sanclient->errmsg())
            if ($sanclient->error());
    }


    sub iscsi_stop {
        my $self = shift();
        my $sanclient = $self->_get_sanclient(protocol => 'iscsi');

        my $ret = $sanclient->_iscsi_service(action => 'stop');
        if ($sanclient->error){
            my $errmsg = join "\n", $sanclient->errmsg();
            confess "Unable to stop iscsi on " . $self->get_description .":\n$errmsg";
        }
        return;
    }

    sub iscsi_configure {
        my ($self,$appliances) = @_;

        return
            if ($configured_iscsi[$$self]);

        # TODO: we ought to record whether iscsi was already
        # configured if it wasn't, we ought to unconfigure it once
        # this object goes away

        # Build list of used IPs on filers
        my $target_objs = [];
        for my $appliance (@$appliances){
            my @targets = $appliance->get_iscsi_targets();
            my $appliance_ips =
              Perf::Depot->get_resources(class     => 'Interface',
                                                  appliance => $appliance);
            confess "No Interface resources found on "
              . $appliance->get_description()
              if (!$appliance_ips || !@$appliance_ips);

            my $found = 0;
            APPLIANCE_IP: for my $appliance_ip (@$appliance_ips){
                my $workers = $appliance_ip->get_users(class => 'Worker');
                if (!$workers || !@$workers){
                    $Log->comment("No workers found using "
                                  . $appliance_ip->get_description);
                    next;
                }
                my $if_ip_addr = $appliance_ip->get_ip();

                # See if the workers belong to an iscsi dataset
                for my $w (@$workers) {
                    my $dataset =
                      Perf::Depot->get_resource(class => 'DataSet',
                                                         name  => $w->get_name);
                    if ($dataset->get_type() eq 'iscsi') {
                        $Log->comment("Will attempt to configure iSCSI between "
                                      . $self->get_description . " and "
                                      . $appliance_ip->get_description);
                        my @valid_targets =
                            grep { $if_ip_addr eq $_->get_ip() } @targets;

                        push @$target_objs, @valid_targets;
                        if (0 < scalar(@valid_targets)) {
                            $found = 1;
                        }
                        next APPLIANCE_IP;
                    }
                }
            }

            confess "No Interface resources found to participate in iSCSI on "
              . $appliance->get_description
              if (!$found);
        }

        my $sanclient =
          $self->_get_sanclient(
                   protocol => 'iscsi',);

        # ensure that the iSCSI daemon is running
        $self->iscsi_start();

        $sanclient->configure(targets => $target_objs);

        if ($sanclient->error){
            my $errmsg = join("\n", $sanclient->errmsg());
            confess $errmsg;
        }

        $configured_iscsi[$$self] = 1;
    }


    sub fcp_configure {
        my $self       = shift();
        my $appliances = shift();

        return
            if ($configured_fcp[$$self]);

        my $target_objs = [];
        foreach my $appliance (@$appliances) {
            my @targets = $appliance->get_fcp_targets();
            push(@$target_objs, @targets);
        }

        my $sanclient = $self->_get_sanclient(protocol => 'fcp');

        $sanclient->configure(targets => $target_objs);

        if ($sanclient->error()) {
            my $errmsg = join("\n", $sanclient->errmsg());
            confess($errmsg);
        }

        $configured_fcp[$$self] = 1;
    }


    sub get_visible_fcp_targets {
        confess('You must implement this method in a subclass of ' . __PACKAGE__);
    }


    sub get_nfsv4_domain_id {
        confess "You must implement this method in a subclass";
    }

    sub cleanup {
        my $self = $_[0];
    }
}
1;
