# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Host::Unix - Representation of a Unix (or linux) host

=head1 DESCRIPTION

This class represent a host tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Host::Unix;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Perf::Depot;
    use Perf::Depot::Resource::Host;
    use Object::InsideOut qw(Perf::Depot::Resource::Host);


    sub detect_ifconfig {
        my $self = $_[0];
        return if (Perf::Depot->get_resources(class=>'Interface', host=>$self));

        # XXX The NACL component doesn't have a find method (BURT649085)
        my $apiset = $self->get_connection()->apiset();
        # XXX NACL doesn't handle /sbin not being in PATH
        my $ifs = $apiset->ifconfig(command_full_path => '/sbin/ifconfig')
                         ->get_parsed_output();

        foreach my $if (@$ifs) {
            my $current_ip = $if->{inet_addr};

            next
                if (!defined($current_ip) || $current_ip =~ /^127\.0/);

            my $if_obj =
              Perf::Depot->get_resource(class     => 'Port::Ethernet',
                                                 name      => $if->{interface},
                                                 container => $self);
            if (!$if_obj) {
                $if_obj =
                  Perf::Depot->create_resource(
                                                      class => 'Port::Ethernet',
                                                      name  => $if->{interface},
                                                      container => $self,
                                                      mtu       => $if->{mtu});
                $if_obj->register();
            }

            my $ip_obj =
              Perf::Depot->get_resource(class          => 'Interface',
                                                 name           => $current_ip,
                                                 ip             => $current_ip,
                                                 mask           => $if->{mask},
                                                 port           => $if_obj,
                                                 create_on_miss => 1);
            $ip_obj->register();
        }
    }


    # TODO BLS remove this once BURT 417854 is implemented by the San
    #          team.
    # NOTE BLS this was fixed but only for devices that've already
    #          been created.  We don't want to create the device
    #          unless necessary so this code will live on for now
    #          until BURT501886 is fixed.
    # Given a linux host, return a hash containing
    # {raw device short name => raw device long name}
    # pairs where the long name is a symlink rooted at /dev and existing
    # in /dev/disk/by-id/.  This is Linux specific; all other *NIX OSs should
    # give devices reliably unique names.
    sub get_dev_map {
        my ($self) = @_;

        confess(ref($self) . '::get_dev_map() only supported on Linux')
          if ($self->get_hosttype() !~ m{^unix-linux}msx);

        my $devs = $self->run_cmd(
            q{/bin/bash -c "for D in /dev/disk/by-id/*; do echo \$(readlink \$D|sed -r 's%\.{2}/%%g') \$(echo \$D|sed 's%/dev/%%'); done"}
        );

        my %dev_map;

        foreach my $line (split(/\r?\n/, $devs)) {
            my @pair = split(/\s/, $line);

            $dev_map{$pair[0]} = $pair[1];
        }

        return \%dev_map;
    }

    sub get_nfsv4_domain_id {
        my ($self) = @_;
        # Look for the line "Domain = ..." in /etc/idmapd.conf
        my $domain_line =
          $self->run_cmd(q{ grep -i ^domain /etc/idmapd.conf  });
        if (!$domain_line){
            confess "Unable to find idmapd domain on client '". $self->get_name ."'";
        }
        if ($domain_line =~ /domain\s+=\s+(.*)$/i){
            my $domain = $1;
            $domain =~ s/^\s+//;
            $domain =~ s/\s+$//;
            return $domain;
        }
        else {
            confess "Unable to parse line '$domain_line'";
        }
    }


    # The device name returned by a San:::DevMgr search is the short
    # name (/dev/sdb) which isn't unique across a SAN or potentially
    # reboots on the same client.  Translate this short name into a
    # long, unique.
    sub get_unique_dev_name {
        my $self = shift();
        my $cn   = shift();

        $cn =~ s{^/dev/}{};

        # multipath device names are already unique
        return $cn
            if ($self->enable_multipathing());

        my $dev_map = $self->get_dev_map();
        my $uniq    = $dev_map->{$cn};

        confess("Couldn't get unique device name for $cn")
            if (!defined($uniq));

        return $uniq;
    }


    # Linux SAN Perf::DataSets use the unique, long raw device name
    # provided by udev, however Sanlun won't accept these for LUN
    # path lookups.  Perform a reverse lookup to get an up-to-date
    # short name.
    sub get_san_dev_name {
        my $self = shift();
        my %args = @_;
        my $ds_uniq_name = delete($args{uniq});

        if ($self->enable_multipathing()) {
            my $devmgr = $self->_get_san_devmgr(%args);
            (my $dev_id = $ds_uniq_name) =~ s{(?:/dev/)?mapper/}{};

            my $san_name = $devmgr->get_dev_name(dev_id => $dev_id);

            confess($devmgr->errmsg())
                if ($devmgr->error());

            return $san_name;
        }
        else {
            my $dev_map = $self->get_dev_map();

            while (my ($san_name, $os_uniq_name) = each(%$dev_map)) {
                return $san_name
                    if ($ds_uniq_name eq $os_uniq_name);
            }
        }

        confess("Couldn't find short name for /dev/$ds_uniq_name");
    }


    sub get_raw_dev_base {
        return '/dev';
    }

    # see comments in Host
    sub disable_multipathing {
        my $self = shift();

        $self->run_cmd('/etc/init.d/multipathd stop');
    }

    sub get_mp_type {
        my $self = shift();

        return ($self->enable_multipathing() ? 'dmmp' : 'none');
    }


    sub get_visible_fcp_targets {
        my $self = shift();

        my $targets =
            $self->run_cmd(
                  'cat /sys/class/fc_remote_ports/*/port_name | sed "s/^0x//"');

        my @targets = split(/\r?\n/, $targets);

        return \@targets;
    }


    sub get_san_dev_size {
        my $self = shift();
        my %args = @_;
        my $dev  = $args{name};
        my $out = $self->run_cmd("sudo /sbin/blockdev --getsize64 $dev");
        $out =~ s/^[[:space:]]+//;
        $out =~ s/[[:space:]]+$//;

        confess("Device size not found for $dev. Got '$out'")
            if ($out !~ /^[[:digit:]]+$/);

        return $out;
    }
}
1;
