# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Host::Windows - Representation of a Windows host

=head1 DESCRIPTION

This class represent a host tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Host::Windows;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use List::MoreUtils qw(firstval);
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::Depot;
    use Readonly;

    use namespace::clean;
    use Perf::Depot::Resource::Host;
    use Object::InsideOut qw(Perf::Depot::Resource::Host);


    # Win  => SMB
    # >=6.2 => 3/2.2
    # >=6.1 => 2.1
    # >=6.0 => 2
    # <6    => 1
    Readonly::Scalar my $WINDOWS_MAJOR_VERSION_SMB2_SUPPORTED  => 6;
    Readonly::Scalar my $WINDOWS_MINOR_VERSION_SMB21_SUPPORTED => 1;
    Readonly::Scalar my $WINDOWS_MINOR_VERSION_SMB3_SUPPORTED  => 2;
    Readonly::Scalar my $DEFAULT_MTU => 1500;

    my @version :Field;
    my @used_drive :Field;


    sub _init : Init {
        my $self = shift();

        $used_drive[$$self] = $self->_get_drive_letters();
    }


    sub _get_drive_letters {
        my $self = shift();
        my $out  = $self->run_cmd('wmic logicaldisk get caption');
        my %drive;

        foreach my $line (split(/\r?\n/, $out)) {
            $line =~ s/[[:space:]]*$//;

            $drive{$line} = 1
              if ($line =~ /^[[:upper:]]:$/);
        }

        foreach my $l ('A' .. 'Z') {
            my $dn = "$l:";

            $drive{$dn} = 0
              if (!defined($drive{$dn}));
        }

        return \%drive;
    }


    sub take_drive_letter {
        my $self = shift();
        my $letter =
          firstval { !$used_drive[$$self]->{$_} } keys(%{$used_drive[$$self]});

        confess('No free drive letters')
          if (!defined($letter));

        $used_drive[$$self]->{$letter} = 1;

        return $letter;
    }


    sub release_drive_letter {
        my $self   = shift();
        my $letter = shift();

        $used_drive[$$self]->{$letter} = 0;

        return $letter;
    }


    sub detect_ifconfig {
        my $self = $_[0];
        return if (Perf::Depot->get_resources(class=>'Interface', host=>$self));

        my $ifs = $self->get_connection()
                       ->apiset()
                       ->ipconfig()
                       ->get_parsed_output();

        foreach my $adapter (@$ifs) {
            my $current_if = $adapter->{adapter};
            my $current_ips = $adapter->{ipv4};

            next
                if (!defined($current_ips) || 0 == scalar(@$current_ips));

            my $current_ip = $current_ips->[0];
            my $current_mask = $adapter->{mask};

            my $if_obj =
              Perf::Depot->get_resource(class     => 'Port::Ethernet',
                                                 name      => $current_if,
                                                 container => $self);
            if (!$if_obj) {
                $if_obj =
                  Perf::Depot->create_resource(
                                                      class => 'Port::Ethernet',
                                                      name  => $current_if,
                                                      container => $self,
                                                      mtu       => $DEFAULT_MTU);
                $if_obj->register();
            }
            my $ip_obj =
              Perf::Depot->get_resource(class => 'Interface',
                                                 name  => $current_ip,
                                                 ip    => $current_ip,
                                                 mask  => $current_mask,
                                                 port  => $if_obj,
                                                 create_on_miss => 1);
            $ip_obj->register();

        }
    }


    sub get_version {
        my $self = shift();

        if (!defined($version[$$self])) {
            my $ver = $self->get_connection()
                           ->apiset()
                           ->ver()
                           ->get_parsed_output()
                           ->[0];

            confess('No version information detected')
                if (!defined($ver));

            my %version;
            @version{'major', 'minor', 'patch'} = split(/\./, $ver->{number});
            $version{os} = $ver->{name};

            $version[$$self] = \%version;
        }

        return $version[$$self];
    }


    sub max_smb_version {
        my $self = shift();

        my $version_info = $self->get_version();

        if ($version_info->{major} >= $WINDOWS_MAJOR_VERSION_SMB2_SUPPORTED) {
            if ($version_info->{minor} >= $WINDOWS_MINOR_VERSION_SMB3_SUPPORTED) {
                return 3;
            }
            elsif ($version_info->{minor} >= $WINDOWS_MINOR_VERSION_SMB21_SUPPORTED) {
                return 2.1;
            }
            else {
                return 2;
            }
        }
        else {
            return 1;
        }
    }
}

1;
