# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Interface - Representation of a network logical interface on a port

=head1 DESCRIPTION

This class represent a network logical interface tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Interface;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Scalar::Util qw();
    use namespace::clean;
    use Perf::Depot::Resource;
    use Object::InsideOut qw(Perf::Depot::Resource);
    use Perf::Depot::Resource::Interface::GX;


    my @port            :Field  :Arg(Name=>'port', Mandatory=>1) :Get(get_port)
                                :Type(Perf::Depot::Resource::Port::Ethernet);
    my @mask            :Field  :Arg(Name=>'mask', Mandatory=>1) :Get(get_mask);
    my @ip              :Field  :Arg(Name=>'ip', Mandatory=>1)   :Get(get_ip)   ;
    my @ip_bits         :Field  :Get(get_bits) :Name(ip_bits);
    my @mask_bits       :Field  :Get(get_mask_bits) :Name(mask_bits);
    my @accessible      :Field  :Name(accessible);

    sub new {
        my $class = shift;
        if ($class ne __PACKAGE__){
            return $class->Object::InsideOut::new(@_);
        }
        my %args = @_;
        my $p = $args{port};
        confess "Argument 'port' is mandatory" if (!$p);
        confess "Argument 'port' must be a Resource" if (!ref($p));
        if (   $p->isa('Perf::Depot::Resource::Port::Ethernet::Filer')
            || $p->isa('Perf::Depot::Resource::Port::Ethernet::Host')) {
            return $class->Object::InsideOut::new(@_);
        }
        else {
            return Perf::Depot::Resource::Interface::GX->new(@_);
        }
    }

    sub _init :Init {
        my $self = $_[0];
        $accessible[$$self] = [];
        # Store the IP address in an array of integers ('name' is only a string representation)
        my @bits = split(/\./,$ip[$$self]);
        my @integer_bits = map {$_ + 0} @bits;
        $ip_bits[$$self] = \@integer_bits;
        $Log->debug("Storing IP Address as [ @integer_bits ] in " . $self->get_description);
        # Store the netmask in an array of integers ('name' is only a string representation)
        @bits = split(/\./,$mask[$$self]);
        my @integer_mask_bits = map {$_ + 0} @bits;
        $mask_bits[$$self] = \@integer_mask_bits;
        $Log->debug("Storing Subnet mask as [ @integer_mask_bits ] in " . $self->get_description);

        $self->add_dependencies($port[$$self]);
    }

    sub get_description {
        my $self = $_[0];
        my $super_description = $self->SUPER::get_description;
        return "$super_description on " . $port[$$self]->get_description;
    }

    sub get_appliance {
        my $self = $_[0];
        return $port[$$self]->get_container;
    }

    sub get_host {
        my $self = $_[0];
        return $port[$$self]->get_container;
    }

    sub is_public {
        my $self = $_[0];
        # RTP
        return 1 if ($ip[$$self] =~ /^10\./);
        # SVL
        return 1 if ($ip[$$self] =~ /^172\./);
        return 0;
    }

    sub get_accessible {
        my $self = $_[0];
        $self->_prune;
        return $accessible[$$self];
    }

    sub _prune {
        my $self = $_[0];
        $self->SUPER::_prune;
        my $accessible_count = scalar(@{$accessible[$$self]});
        my $i = 0;
        while ($i < $accessible_count){
            if (!defined($accessible[$$self]->[$i])){
                $Log->comment("Removing stale 'accessible' reference in " . $self->get_description);
                splice @{$accessible[$$self]},$i,1;
                $accessible_count --;
                next;
            }
            elsif( ref($accessible[$$self]->[$i]) ){
                confess "Reference to resource ". $accessible[$$self]->[$i]->get_description . " is not weak!"
                    if (!Scalar::Util::isweak($accessible[$$self]->[$i]));
            }
            $i ++;
        }
    }

    sub in_same_subnet {
        my ($self, $interface) = @_;
        confess "You must pass a argument" if (!$interface);
        confess "You must pass a Perf::Depot::Resource::Interface"
            if (!$interface->isa('Perf::Depot::Resource::Interface'));

        # The netmask must match
        if ( $self->get_mask ne $interface->get_mask ){
            $Log->comment("Netmasks don't match: " . $self->get_mask . ' vs ' . $interface->get_mask);
            return 0;
        }
        # Make sure IPs are on the same subnet
        my $self_bits = $self->get_bits;
        my $interface_bits = $interface->get_bits;
        my $mask_bits = $self->get_mask_bits;
        my $i = 0;
        while ($i < scalar @$self_bits){
            my $c = $self_bits->[$i] & $mask_bits->[$i];
            my $f = $interface_bits->[$i] & $mask_bits->[$i];
            if ($c != $f){
                return 0;
            }
            $i++;
        }
        return 1;
    }

    sub add_accessible {
        my ($self, $interface) = @_;
        confess "You must pass a argument" if (!$interface);
        confess "You must pass a Perf::Depot::Resource::Interface"
            if (!$interface->isa('Perf::Depot::Resource::Interface'));
        if (!$self->is_accessible($interface)){
            push @{$accessible[$$self]}, $interface;
            Scalar::Util::weaken($accessible[$$self]->[-1]);
        }
    }

    sub is_accessible {
        my ($self, $interface) = @_;
        confess "You must pass a argument" if (!$interface);
        confess "You must pass a Perf::Depot::Resource::Interface"
            if (!$interface->isa('Perf::Depot::Resource::Interface'));
        $self->_prune;
        for my $i (@{$accessible[$$self]}){
            return 1 if ($i == $interface);
        }
        return 0;
    }

    sub set_state {
        my $self  = shift();
        my $state = lc(shift());

        confess("Invalid state: $state")
            if ($state !~ m{(?:up|down)});

        my $container = $self->get_container();

        NACL::C::NetworkPort->modify(
            command_interface => $container->get_conn(),
            node              => $container->get_name(),
            port              => $port[$$self]->get_name(),
            'up-admin'        => (('up' eq $state) ? 'true' : 'false'));
    }
}
1;
