# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Interface - Representation of a network logical interface on a port

=head1 DESCRIPTION

This class represent a network logical interface tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Interface::GX;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Readonly;
    use Scalar::Util qw();
    use NACL::STask::NetworkInterface;

    use namespace::clean;
    use Perf::Depot::Resource::Interface;
    use Object::InsideOut qw(Perf::Depot::Resource::Interface);

    my @is_home :Field :Arg(Name=>'is_home', Mandatory=>1) :Get(is_home);
    my @role    :Field :Arg(Name=>'role', Mandatory=>1) :Get(get_role);
    my @appliance :Field :Arg(Name=>'appliance', Mandatory=>1)
                         :Type('Perf::Depot::Resource::Appliance::GX');
    my @orig_lif_info :Field;
    my @nacl_lif :Field;

    # data-protocol should be in here, but due in inconsistencies in the NACL
    # libraries it needs to be handled separately
    Readonly::Array my @LIF_FIELDS => qw(vserver lif role home-node home-port
                                         address netmask);

    sub _init :Init {
        my $self = $_[0];
        $self->add_dependencies($appliance[$$self]);
    }

    sub get_description {
        my $self = $_[0];
        my $super_description = $self->SUPER::get_description;
        return "$super_description for " . $appliance[$$self]->get_description;
    }

    # Override Interface::get_appliance
    sub get_appliance{
        my $self = $_[0];
        return $appliance[$$self];
    }

    sub get_port_node{
        my $self = $_[0];
        return $self->SUPER::get_appliance;
    }

    sub set_state {
        my $self = shift();
        my $state = lc(shift());

        confess("Invalid state: $state")
            if ($state !~ m{(?:up|down)});

        my $vserver = $self->get_appliance();

        NACL::C::NetworkInterface->modify(
                                         command_interface => $self->get_conn(),
                                         server         => $vserver->get_name(),
                                         lif            => $self->get_name(),
                                         'status-admin' => $state,);
    }


    # returns an array of supported protocols
    sub get_protocols {
        my $self = shift();

        my $vserver = $self->get_appliance();

        return
          NACL::CS::NetworkInterface->fetch(
                                           command_interface => $vserver->get_conn(),
                                           filter            => {
                                                vserver => $vserver->get_name(),
                                                lif     => $self->get_name(),},
                                           requested_fields => ['data-protocol']
                                           )
                                    ->data_protocol();
    }


    sub change_protocol {
        my $self = shift();
        my $new_proto = shift();

        my $vserver      = $self->get_appliance();

        # This works for one LIF and breaks if attempted on a second
        # as it destroys the first LIF to create the second. BURT649531
        # $nacl_lif[$$self] =
        #   NACL::STask::NetworkInterface->set_iscsi_lif(
        #                               command_interface => $vserver->get_conn(),
        #                               vserver           => $vserver->get_name(),
        #                               lif               => $self->get_name(),
        #                               nacltask_if_lif_exists => 'reuse',);

        my $lif_name     = $self->get_name();
        my $vserver_name = $vserver->get_name();
        my $ci           = $vserver->get_conn();
        my $lif =
          NACL::STask::NetworkInterface->new(
                        command_interface => $ci,
                        vserver => $vserver_name,
                        lif => $lif_name,);

        my $lif_state = $lif->state(requested_fields => [@LIF_FIELDS, 'data-protocol']);

        if (!defined($orig_lif_info[$$self])) {
            $orig_lif_info[$$self] =
              {map { $_ => $lif_state->$_() } @LIF_FIELDS};
            $orig_lif_info[$$self]->{'data-protocol'} =
              [$lif_state->data_protocol()];
        }

        $lif->purge(nacltask_verify => 1);

        my %new_info = %{$orig_lif_info[$$self]};

        $new_info{'data-protocol'} = $new_proto;

        $nacl_lif[$$self] =
          NACL::STask::NetworkInterface->create(command_interface => $ci,
                                                nacltask_verify => 1,
                                                %new_info,)
    }


    sub cleanup {
        my $self = shift();

        if (defined($nacl_lif[$$self])) {
            # $nacl_lif[$$self]->delete_iscsi_lif();

            $nacl_lif[$$self]->purge();

            NACL::STask::NetworkInterface->create(
                        command_interface => $self->get_appliance()->get_conn(),
                        nacltask_verify  => 1,
                        %{$orig_lif_info[$$self]},);
        }
    }
}
1;
