# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Lun - Representation of a Lun in a volume or qtree

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Lun;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use NACL::C::UnitNormalization qw(convert_to_bytes_int);
    use Perf::Depot::Resource::Volume;
    use Perf::Depot::Resource::Qtree;
    use Object::InsideOut qw(Perf::Depot::Resource);

    my @volume  :Field :Arg(name=>'volume', mandatory=>0)
                :Type(Perf::Depot::Resource::Volume);
    my @qtree   :Field  :Arg(name=>'qtree', mandatory=>0)
                :Type(Perf::Depot::Resource::Qtree)
                :Get(get_qtree);
    my @type    :Field :Arg(name=>'type', mandatory=>1) :Get(get_type);
    my @size    :Field :Arg(name=>'size', mandatory=>1) :Get(get_size);
    my @create_lun :Field :Arg(name=>'create_lun', mandatory=>0);

    sub _init :Init {
        my $self = shift;

        confess "You must specify a qtree or volume"
          if (!$volume[$$self] && !$qtree[$$self]);

        $self->add_dependencies($volume[$$self]) if ($volume[$$self]);
        $self->add_dependencies($qtree[$$self]) if ($qtree[$$self]);

        if ($create_lun[$$self]){
            $self->create();
        }
        else{
            $self->verify;
            $self->online;
        }
        return;
    }

    sub create {
        my $self =$_[0];

        $self->get_appliance()->lun_create(type   => $type[$$self],
                                           size   => $size[$$self],
                                           path   => $self->get_path(),
                                           volume => $volume[$$self],);

        # LUN create may make the size slightly larger than what was requested.
        # It has been seen to add 2M. Verify within 1% larger, but no smaller.
        my $lun_show_size = convert_to_bytes_int($self->get_actual_size);
        my $lun_size_bytes = convert_to_bytes_int($size[$$self]);
        if ($lun_show_size < $lun_size_bytes || $lun_show_size > $lun_size_bytes * 1.01){
            confess "Size not in acceptable limits for " . $self->get_description .
                    " ($lun_show_size instead of $lun_size_bytes)";
        }
        # Set size to true size created
        $size[$$self] = $lun_show_size;
        $Log->debug("Set Perf::DataSet's LUN size to true size of $size[$$self]");
        return;
    }


    sub get_actual_size{
        my $self = $_[0];

        my $lun_size = $self->get_appliance()->lun_get_size($self);

        return $lun_size;
    }


    sub verify {
        my $self =$_[0];
        my $lun_show_size = convert_to_bytes_int($self->get_actual_size());
        my $lun_size_bytes = convert_to_bytes_int($size[$$self]);
        if ($lun_show_size != $lun_size_bytes){
            confess "Wrong size for " . $self->get_description .
                    " ($lun_show_size instead of $lun_size_bytes or $size[$$self])";
        }
        return;
    }


    sub online {
        my $self = $_[0];

        $self->get_appliance()->lun_online($self);
    }


    sub get_path {
        my $self = $_[0];
        my $str = "/vol/" . $self->get_volume->get_name;
        if ($qtree[$$self]){
            $str .= "/" . $qtree[$$self]->get_name;
        }
        $str .= "/" . $self->get_name;
        return $str;
    }

    sub get_volume {
        my $self = $_[0];
        my $qtree_obj = $qtree[$$self];
        my $vol_obj = $volume[$$self];
        if ($qtree_obj){
            $vol_obj = $qtree_obj->get_volume;
        }
        return $vol_obj;
    }

    sub get_description {
        my $self = $_[0];
        my $str = $self->SUPER::get_description;
        if ($qtree[$$self]){
            $str .= " in " . $qtree[$$self]->get_description;
        }
        else{
            $str .= " in " . $volume[$$self]->get_description;
        }
        return $str;
    }


    sub cleanup {
        my $self = $_[0];

        $self->get_appliance()->lun_destroy($self);
    }


    sub get_appliance {
        my $self = shift();

        return $self->get_volume()->get_appliance();
    }
}
1;
