# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Mountpoint::NAS - Base class for NAS mountpoints

=head1 DESCRIPTION

It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Mountpoint::NAS;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::Depot;
    use List::Util qw(shuffle);
    use NACL::STask::Client::Mount;

    use namespace::clean;
    use Perf::Depot::Resource::Mountpoint;
    use Object::InsideOut qw(Perf::Depot::Resource::Mountpoint);

    my @volume  :Field :Get(get_volume)
                :Arg(Name=>'volume', Mandatory=>1)
                :Type(Perf::Depot::Resource::Volume);
    my @appliance :Field :Get(get_appliance)
                :Arg(Name=>'appliance', Mandatory=>0)
                :Type(Perf::Depot::Resource::Appliance);
    my @appliance_ip :Field :Get(get_appliance_ip)
                 :Arg(Name=>'appliance_ip', Mandatory=>0)
                 :Type(Perf::Depot::Resource::Interface);
    my @options :Field :Get(get_options)
                :Arg(Name=>'options', Mandatory=>1)
                :Type(HASHref);
    my @nacl_obj :Field;


    sub _init :Init {
        my $self = $_[0];
        confess "No host found!" if (!$self->get_host);
        confess "You must specify a appliance or appliance ip address"
            if (!$appliance[$$self] && !$appliance_ip[$$self]);
        $self->add_dependencies($volume[$$self]);
        if (!$appliance[$$self]){
            $appliance[$$self] = $appliance_ip[$$self]->get_appliance();
        }
        if (!$appliance_ip[$$self]){
            $appliance_ip[$$self] = $self->_get_appliance_ip();
        }
        $self->latch;
        $self->add_dependencies($appliance_ip[$$self]);
    }


    sub get_mount_target {
        confess('This method must be implemented in a subclass of ' . __PACKAGE__);
    }


    sub get_mntopts {
        my $self = $_[0];
        return $self->get_options->{mountopts};
    }


    sub get_filer{
        my $self = $_[0];
        return $appliance[$$self];
    }

    sub get_filer_ip{
        my $self = $_[0];
        return $appliance_ip[$$self];
    }

    sub cleanup {
        my $self = $_[0];
        $self->unlatch if ($self->is_latched);
        return;
    }

    sub _get_appliance_ip :Restricted {
        my $self = shift;
        my $appliance_ips = Perf::Depot->get_resources(class     => 'Interface',
                                                          appliance => $appliance[$$self]);
        my @ips = shuffle(@$appliance_ips);
        my $localhost = Perf::Depot->get_resource(class => 'Localhost');

        foreach my $ip (@ips) {
            if ($localhost->can_ping($ip)) {
                return $ip;
            }
        }

        confess 'No filer interfaces accessible from localhost';
    }


    sub do_latch {
        my $self                = shift();
        my %GenericMount_params = %{$self->get_options};

        my $mount =
          NACL::STask::Client::Mount->create(
                       command_interface => $self->get_host()->get_connection(),
                       server            => $self->get_appliance_ip()->get_ip(),
                       nacltask_verify         => 1,
                       %GenericMount_params,
                       $self->get_mount_target(),);
        $nacl_obj[$$self] = $mount;
        my $path          = $mount->get_one_state_attribute('mount_point');

        if (!$self->was_never_latched && $self->get_name ne $path) {
            confess "Latching produced an unexpected path ('$path' instead of "
              . $self->get_name . ")";
        }

        $self->set_name($path);

    }


    sub do_unlatch {
        my $self = shift();

        # is there a way to get a /YES down into the CIFS version of
        # this to ensure there are no hangs unmounting?
        $nacl_obj[$$self]->purge(nacltask_verify => 1,
                                 # BURT625905
                                 # force         => 1,
                             );
    }
}
1;
