# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Port - Representation of a physical communications port

=head1 DESCRIPTION

This class represents a physical port tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Port;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Scalar::Util qw();

    use namespace::clean;
    use Perf::Depot::Resource;
    use Object::InsideOut qw(Perf::Depot::Resource);

    my @accessible_by   :Field
                        :Name(accessible_by)
                        :Get(get_accessible_by);
    my @container       :Field
                        :Arg(Name=>'container', Mandatory=>1)
                        :Get(get_container)
                        :Type('Perf::Depot::Resource');
    my @failover_targets :Field
                         :Name('failover_targets')
                         :Get('get_failover_targets');

    sub _init :Init {
        my $self = $_[0];
        $accessible_by[$$self] = [];
        $failover_targets[$$self] = [];
        $self->add_dependencies($container[$$self]);
    }

    sub get_description {
        my $self = $_[0];
        my $super_description = $self->SUPER::get_description;
        return "$super_description on " . $container[$$self]->get_description;
    }


    sub get_conn {
        confess "This method must be implemented by a subclass";
    }

    sub _prune {
        my $self = $_[0];
        $self->SUPER::_prune;
        my $accessible_by_count = $accessible_by[$$self]?0:scalar(@{$accessible_by[$$self]});
        my $i = 0;
        while ($i < $accessible_by_count){
            if (!defined($accessible_by[$$self]->[$i])){
                $Log->comment("Removing stale 'accessible_by' reference in " . $self->description);
                splice @{$accessible_by[$$self]},$i,1;
                $accessible_by_count --;
                next;
            }
            elsif( ref($accessible_by[$$self]->[$i]) ){
                confess "Reference to resource ". $accessible_by[$$self]->[$i]->get_description . " is not weak!"
                    if (!Scalar::Util::isweak($accessible_by[$$self]->[$i]));
            }
            $i ++;
        }
    }

    sub add_accessible_by {
        my ($self, $port) = @_;
        confess "You must pass a argument" if (!$port);
        confess "You must pass a Perf::Depot::Resource::Port" if (!$port->isa('Perf::Depot::Resource::Port'));
        $self->_prune;
        if (!$self->is_accessible_by($port)){
            push @{$accessible_by[$$self]}, $port;
            Scalar::Util::weaken( $accessible_by[$$self]->[-1] );
        }
    }

    sub is_accessible_by {
        my ($self, $port) = @_;
        confess "You must pass a argument" if (!$port);
        confess "You must pass a Perf::Depot::Resource::Port" if (!$port->isa('Perf::Depot::Resource::Port'));
        $self->_prune;
        return 1 if ($port->found_in($accessible_by[$$self]));
        return 0;
    }

    sub add_failover_target {
        my ($self, $target) = @_;
        confess "You must specify an argument" if (!defined($target));
        confess "Argument must be a scalar" if (ref($target));
        if ( ! grep { $_ eq $target } @{$failover_targets[$$self]}){
            $Log->comment("Adding '$target' to list of failover targets");
            push @{$failover_targets[$$self]}, $target;
        }
        return;
    }
}
1;
