# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Port::Ethernet::Filer - Representation of a network port on a filer

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Port::Ethernet::Filer;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;

    use namespace::clean;
    use Perf::Depot::Resource::Port::Ethernet::Appliance;
    use Object::InsideOut qw(Perf::Depot::Resource::Port::Ethernet::Appliance);

    my @enabled_iscsi :Field :Name(enabled_iscsi) :Set(set_iscsi_enabled)
                                                  :Get(is_iscsi_enabled);


    sub _init :Init {
        my $self = $_[0];
        # Verify appliance is a filer
        confess "Appliance argument must be a Filer"
            if (!$self->get_appliance->isa('Perf::Depot::Resource::Filer'));
    }

    sub get_filer {
        my $self = $_[0];
        return $self->get_appliance;
    }

    sub iscsi_status {
        my $self            = $_[0];
        my $apiset          = $self->get_conn()->apiset();
        my $res             = $apiset->iscsi_interface_show();
        my $complete_status = $self->get_container()->iscsi_interface_status();
        my $status          = $complete_status->{$self->get_name()};
        confess('No iscsi status found for ' . $self->get_description())
          if (!$status);
        return $status;
    }


    sub iscsi_enable {
        my $self = $_[0];
        return if ($self->iscsi_status ne 'disabled');

        my $san_filer = $self->get_container()->_get_san_filer();

        $san_filer->iscsi_interface_enable($self->get_name());

        $self->set_iscsi_enabled(1);
    }


    sub iscsi_disable {
        my $self = $_[0];

        my $san_filer = $self->get_container()->_get_san_filer();

        $san_filer->iscsi_interface_disable($self->get_name());

        $self->set_iscsi_enabled(0);
    }

    # Return MTU of the specified interface on the filer, optionally
    # setting it first if a second argument is supplied.

    sub _mtu {
        my ($self,$mtu) = @_;
        my $if = $self->get_name;
        my $params = {
            match     => "^$if:.*mtu\\s(\\d+)",               # ifconfig regex
            query     => "ifconfig $if",                      # ifconfig query cmd
            update    => "ifconfig $if mtusize ",             # ifconfig mtu update cmd
            wait_match=> "-up\\)",                             # ifconfig regex before returning
        };
        return $self->_ifconfig_mtu($params, $mtu);
    }

    sub add_user {
        my ($self, $resource) = @_;
        my $was_added = $self->SUPER::add_user($resource);
        return if (!$was_added);
        return if (!$resource->isa('Perf::Depot::Resource::DataSet'));

        if ('iscsi' eq $resource->get_type()){
            # Turn on the iscsi interface
            $self->iscsi_enable();
        }

        return $was_added;
    }

    sub cleanup {
        my $self = shift();
        if ($enabled_iscsi[$$self]) {
            $self->iscsi_disable;
        }
    }
}
1;
