# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Process::Base - Representation of a remote process responsible for starting others

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Process::Base;
{
    use strict;
    use warnings;
    use Carp qw( confess cluck );
    use Tharn qw( $Log subtest );
    use NATE::RPC::Freeze qw(freeze);
    use NATE::Events qw( eventloop_drain );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Object::InsideOut qw(Perf::Depot::Resource);

    my @host          :Field :Arg(Name=>'host', Mandatory=>1) :Type(Perf::Depot::Resource::Host) :Get(get_host);
    my @reply         :Field :Name('reply') :Std(Name=>'reply', Restricted=>1);
    my @request       :Field :Name('request') :Std(Name=>'request', Restricted=>1);

    sub _init :Init {
        my $self = $_[0];
        $self->add_dependencies($host[$$self]);
        if (!$self->get_uid){
            $self->set_uid( Perf::Depot->make_uid );
        }
        $self->set_name( $self->get_uid );
        return;
    }

    sub get_description {
        my $self = shift;
        my $d = $self->SUPER::get_description . " on " . $host[$$self]->get_description;
    }

    sub get_staf_peer {
        confess "You must implement this subroutine in a subclass";
    }

    sub get_hostname {
        my $self = shift;
        return $host[$$self]->get_name;
    }

    sub call_request {
        my ($self, $code, $args, $no_log) = @_;
        if ($self->has_reply){
            cluck("Throwing away reply to previous request:\n" . $request[$$self]);
            eval { $self->read_reply };
        }
        elsif ($self->has_request){
            confess("There is a pending request on " . $self->get_description . ":\n" . $request[$$self]);
        }
        if ($args){
            my @frozen_args = map { freeze($_) } @$args;
            $code=$code."(".join(",",@frozen_args).")";
        }
        if (!$no_log){
            $Log->debug($self->get_description . " is executing asynchronous code: $code");
        }
        $request[$$self] = $code;
        my $callback = sub {    my ( $key, $result ) = @_;
                                # Make a copy so we can alter the original below
                                %{$reply[$$self]} = %$result;
                                # Don't leave the error in the result object so
                                # we don't die from remote errors too soon
                                delete $result->{error};
                                $Log->debug($self->get_description . " has returned a result");
                                return;
                            };
        eval { $self->get_staf_peer->perl_eval( code => $code, callback => $callback, async => 1 ) };
        if ($@){
            confess "Got error during call_request on " . $self->get_description . ": $@";
        }
        return;
    }

    sub call {
        my ($self, $code, $args) = @_;
        $Log->debug($self->get_description . " is executing synchronous code: $code");
        $self->call_request($code, $args, 1);
        $self->wait_reply;
        return $self->read_reply;
    }

    sub wait_reply {
        my ($self) = @_;
        confess $self->get_description . " has no request" if (!$self->has_request);
        while (!$self->has_reply){
            $Log->debug("Waiting for '" . $self->get_description ."' to reply");
            eventloop_drain();
        }
        return;
    }

    sub read_reply {
        my ($self) = @_;
        my $ret = delete $reply[$$self];
        if (defined $ret){
            delete $request[$$self];
            if ($ret->{error}){
                die $ret->{error};
            }
            else {
                return $ret->{retval};
            }
        }
        return;
    }

    sub message_put {
        my $self = shift;
        return $self->get_staf_peer->message_put(@_);
    }

    sub message_get {
        my $self = shift;
        return $self->get_staf_peer->message_get(@_);
    }

    sub has_request {
        my ($self) = @_;
        return defined $request[$$self];
    }

    sub has_reply {
        my ($self) = @_;
        return defined $reply[$$self];
    }

    sub cleanup {
        confess "You must implement this method in a subclass";
    }

}
1;
