# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Process::Base - Representation of a remote process responsible for starting others

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Process::Factory;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log subtest );
    use NATE::RPC::Freeze qw(freeze);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Object::InsideOut qw(Perf::Depot::Resource::Process::Base);

    my @process         :Field :Name('process') :Get('get_staf_peer');
    my @response_keys   :Field :Name('response_keys');

    sub _init :Init {
        my $self = $_[0];

        # Make sure factory is unique on client
        my $f = $self->get_resource( class=>'Process::Factory', host=>$self->get_host );
        confess "An existing factory was found on " . $self->get_host->get_description if ($f);

        $response_keys[$$self] = {};


        #
        # Start IPC server on client. Note that this file and Perf::DataSet/Depot.pm
        # must agree on the runid.
        my $host_name = $self->get_host->get_name;
        my $is_windows = $self->get_host->isa('Perf::Depot::Resource::Host::Windows');
        my $code = '
                                    {   use Tharn;
                                        use Carp qw(confess);';
        if (!$is_windows){
            # No need to load Perf::DataSet on Windows since it won't get shared with the
            # processes later on
            $code .= "
                                        use Perf::DataSet;";
        }
        $code .= '
                                        use NATE::Events qw(eventloop_drain);
                                        while(1) { eventloop_drain(); };
                                    }';
        my @args = (    $code,
                       '-bg',
                       '-runid', "Perf_DataSet_log_from_${host_name}_factory_" . $self->get_uid,
                    );

        if (!$self->get_host->isa('Perf::Depot::Resource::Localhost')){
            push @args, ('-client', $host_name);
        }
        $process[$$self] = subtest(@args);
        return;
    }

    sub is_alive {
        my $self = shift;
        return 0 if (!$process[$$self]);
        return $process[$$self]->is_alive;
    }

    sub create_processes {
        my ($self, %args) = @_;
        my $count = $args{count};
        confess "You must pass a count" if (!$count);
        # Generate uids
        my @process_uids = map { Perf::Depot->make_uid } (1..$count);

        my $is_windows = $self->get_host->isa('Perf::Depot::Resource::Host::Windows');
        my $code =  '
                        use Tharn;';
        if (!$is_windows){
            # On Unix, load all necessary libraries in the Factory so they get shared
            # among processes
            $code .='
                        use Carp qw(confess);'."
                        use Perf::DataSet;".'
                        use NATE::Events qw(eventloop_drain);';
        }
        $code .=    '
                        my $i = 0;
                        my @uids = qw/'. "@process_uids" .'/;
                        for my $uid (@uids){
                            $::p->{$uid} = subtest( \'{';
        if ($is_windows){
            # On Windows, each child process has to load its own libraries
            # because 'exec' will be called. Attempts to bypass the 'exec'
            # by calling subtest on a coderef resulted in the factory
            # process dying with a funny exit code
            $code .='
                                                        use Carp qw(confess);
                                                        use NATE::Events qw(eventloop_drain);'."
                                                        use Perf::DataSet;";
        }
        $code .=    '
                                                        while(1){eventloop_drain()}}\',
                                                            \'-bg\',
                                                            \'-runid\',"process_$uid",
                                                            );
                            $i++;
                        }
                        my @handles;
                        for my $uid (@uids){
                            push @handles, $::p->{$uid}->handle;
                        }
                        return \@handles;
                    ';
        my $handles = $self->call($code);
        my @processes;
        my $i = 0;
        my $endpoint = $self->get_staf_peer->endpoint;
        while ($i <= $#process_uids ){
            my $uid = $process_uids[$i];
            my $handle = $handles->[$i];
            push @processes, Perf::Depot->create_resource(  class=> 'Process',
                                                               uid => $uid,
                                                               name => 'Dummy',
                                                               host=>$self->get_host,
                                                               factory=>$self,
                                                               endpoint=>$endpoint,
                                                               handle=>$handle);
            $i++;
        }
        return \@processes;
    }

    sub is_process_alive {
        my ($self, %args) = @_;
        my $uid = $args{uid};
        confess "You must pass a uid" if (!$uid);
        return $self->call( '
                            use Carp qw(confess);
                            my $subtest = $::p->{'.$uid.'};
                            confess "Unknown process '. $uid .'" if(!$subtest);
                            return $subtest->is_alive;'
                          );
    }

    sub cleanup_process {
        my ($self, %args) = @_;
        my $uid = $args{uid};
        confess "You must pass a uid" if (!$uid);
        $self->call('
                                my $subtest = $::p->{'.$uid.'};
                                if (!$subtest){
                                    $Log->comment("No subtest '. $uid .' found");
                                    return;
                                }
                                if ($subtest->is_alive){
                                    $Log->comment("Subtest $uid still alive. Quitting");
                                    $subtest->quit;
                                    if ($subtest->is_alive){
                                        $Log->comment("Subtest $uid still alive. Killing");
                                        $subtest->kill;
                                    }
                                }
                                delete $::p->{'.$uid.'};
                                return;'
                             );
    }

    sub cleanup_processes_start {
        my ( $self, @process_uids ) = @_;
        confess "You must pass at least one uid" if (!@process_uids);
        $self->call_request('
                                for my $uid ( qw/' . "@process_uids" . '/ ){
                                    my $subtest = $::p->{$uid};
                                    if (!$subtest){
                                        $Log->comment("No subtest $uid found");
                                        return;
                                    }
                                    if ($subtest->is_alive){
                                        $Log->comment("Subtest $uid still alive. Quitting");
                                        $subtest->quit;
                                        if ($subtest->is_alive){
                                            $Log->comment("Subtest $uid still alive. Killing");
                                            $subtest->kill;
                                        }
                                    }
                                    delete $::p->{$uid};
                                }
                                return;
                            '
                            );
        return;
    }

    sub cleanup {
        my $self = $_[0];
        if (!$process[$$self]){
            $Log->comment("No subtest found");
            return;
        }
        if ( !$process[$$self]->is_alive){
            $Log->comment("Subtest has already shut down");
            return;
        }
        $self->call( '
                        for my $uid ( keys %::p ){
                            my $subtest = $::p->{$uid};
                            if ($subtest->is_alive){
                                $Log->comment("Subtest $uid still alive. Quitting");
                                $subtest->quit;
                            }
                            $subtest->wait_finish( timeout=>5 );
                            if ($subtest->is_alive){
                                $Log->comment("Subtest $uid still alive. Killing");
                                $subtest->kill;
                            }
                            delete $::p->{$uid};
                        }
                        return;'
                    );
        $process[$$self]->quit;
        $process[$$self]->wait_finish(timeout=>5);
        if ($process[$$self]->is_alive){
            $Log->comment("Process for " . $self->get_description . " still alive. Killing");
            $process[$$self]->kill;
        }
        return;
    }
}
1;
