# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::RootDir - Representation of a Root directory in a volume or qtree

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::RootDir;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN {
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Params::Validate;
    use List::Util qw(first);
    use List::MoreUtils qw(uniq);
    use Object::InsideOut qw(Perf::Depot::Resource);

    my @volume   :Field :Arg(name=>'volume', mandatory=>0) :Type(Perf::Depot::Resource::Volume) :Get(get_volume);
    my @qtree    :Field :Arg(name=>'qtree', mandatory=>0)  :Type(Perf::Depot::Resource::Qtree) :Get(get_qtree);
    my @snapshot :Field :Arg(name=>'snapshot', mandatory=>0) :Type(Perf::Depot::Resource::Snapshot) :Get(get_snapshot);

    sub _init :Init {
        my $self = shift;
        confess "You must specify a qtree or volume" if ( !$volume[$$self] && !$qtree[$$self]);
        $self->add_dependencies($volume[$$self]) if ($volume[$$self]);
        $self->add_dependencies($qtree[$$self]) if ($qtree[$$self]);
        $self->add_dependencies($snapshot[$$self]) if ($snapshot[$$self]);
    }

    sub create_root {
        my $self = shift();
        my $dataset_uid = shift();

        $self->execute_dir_op(
            dataset_uid => $dataset_uid,
            code => q%
              if (!-d "$work_dir") {
                  make_path("$work_dir", {verbose => 1, mode => 0777})
                    or confess("Unable to create directory '$work_dir': $!");
              }
            %);
    }

    # Pass in a code string to be executed by a Worker.  Code can
    # expect the variable $work_dir to be properly set to the Worker's
    # root directory.
    sub execute_dir_op {
        my $self = shift();
        my %arg =
          Params::Validate::validate(@_,
                                     {code        => 1,
                                      dataset_uid => 1,});
        my $dataset =
          Perf::Depot->get_resource(class => 'DataSet',
                                       name  => $arg{dataset_uid},);
        my $workers    = $dataset->get_worker_resources();
        my $full_paths = $dataset->get_worker_full_paths();

        my ($pid, $fullpath) = each(%$full_paths);

        my $target_worker = first { $pid eq $_->get_process_name() } @$workers;

        confess("No Worker found to create working directory: $fullpath")
          if (!defined($target_worker));

        $target_worker->execute(
            async => 0,
            code  => qq%
                          use File::Path qw(make_path remove_tree);
                          my \$work_dir = "$fullpath";
                          $arg{code}
                        %);
    }

    sub delete_root {
        my $self = shift();
        my $dataset_uid = shift();

        $self->execute_dir_op(
            dataset_uid => $dataset_uid,
            code => q%
              remove_tree("$work_dir");
            %);
    }
}
1;
