package Perf::Depot::Resource::SANDevice;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log host );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::Utils qw(in_bytes);
    use List::Util qw(first);
    use Object::InsideOut qw(Perf::Depot::Resource::Mountpoint);


    my @lun_map :Field
                :Arg(Name => 'lun_map', Mandatory => 1)
                :Get(get_lun_map)
                :Type(Perf::Depot::Resource::LUNMap);
    my @set_up :Field;
    my @dev_size :Field
                 :Get(get_dev_size);
    my @san_service
        :Field
        :Arg(Name => 'san_service', Mandatory => 1)
        :Type(Perf::Depot::Resource::SANService);


    sub _init :Init {
        my $self = $_[0];
        $self->add_dependencies($lun_map[$$self]);
        $set_up[$$self] = 0;
        $self->setup;
    }


    sub is_set_up {
        my $self = $_[0];
        return $set_up[$$self] ? 1 : 0;
    }


    sub confirm_latched {
        my $self = $_[0];

        return $self->check_valid();
    }


    sub check_valid {
        my $self          = shift();
        my $host_lun_path = $self->get_host()->get_lun_path_for_dev($self);

        if (  !$host_lun_path
            || $host_lun_path eq ''
            || $host_lun_path =~ /^\s+$/) {
            confess "No LUN path found!";
        }

        my $lun = $self->get_lun();
        my $appliance_lun_path =
            $lun->get_appliance()->get_name() . ':'
          . $lun->get_path();

        confess(  "LUN paths don't match! Host ($host_lun_path) vs. "
                . "Appliance ($appliance_lun_path)")
          if ($host_lun_path ne $appliance_lun_path);

        return 1;
    }


    sub is_latched {
        my $self = $_[0];
        my $is_latched = $self->SUPER::is_latched;
        if ($is_latched){
            $self->confirm_latched;
        }

        return $is_latched;
    }

    # udev automatically sets up raw device for use so this is
    # essentially a no-op
    sub do_latch {
        my $self = shift();
    }

    # no-op
    sub do_unlatch {
        my $self = shift();
    }

    sub get_description {
        my $self = $_[0];
        my $str = $self->SUPER::get_description;
        $str .= " for " . $self->get_lun()->get_description;
        return $str;
    }

    sub get_lun {
        my $self = shift();

        return $lun_map[$$self]->get_lun();
    }

    sub setup {
        my $self = $_[0];
        return if ($set_up[$$self]);

        my $lun              = $self->get_lun();
        my $filer            = $lun->get_volume()->get_physical_appliance();
        my $host_obj         = $self->get_host();
        my $host_name        = $host_obj->get_name;
        my $host_san_service = $san_service[$$self];
        my $protocol         = $self->get_protocol();

        if ($host_san_service->is_used) {
            # The call to San::setup below is going to restart the iscsi service
            # This will cause all kinds of troubles if we already have
            # an iSCSI workload going for example. Make sure none of the
            # Perf::DataSet users are busy.
            my $busy_datasets = $host_san_service->get_busy_dataset_users;
            if ($busy_datasets && @$busy_datasets) {
                my $str = join(', ', map { $_->get_name } @$busy_datasets);
                confess
                  "Operation aborted because of the following busy Perf::DataSet(s): $str";
            }
        }

        my @device_names =
          $host_obj->get_free_san_devs(protocol => $protocol,
                                       lun      => $lun,);

        my $new_dev_name;
        # the unique name we want to use in Perf::DataSet
        my $unique_dev_name;

      FREE_DEVICE: foreach my $device (@device_names) {
            $unique_dev_name = $host_obj->get_unique_dev_name($device);

            my $found_device =
              Perf::Depot->get_resource(class => 'SANDevice',
                                                 host  => $host_obj,
                                                 name  => $unique_dev_name,);

            if (!defined($found_device)) {
                $new_dev_name = $device;
                last FREE_DEVICE;
            }
        }

        confess("No free SAN devices found")
            if (!defined($new_dev_name));

        $host_obj->san_dev_create(name     => $new_dev_name,
                                  protocol => $protocol,);

        $self->set_name($unique_dev_name);

        $dev_size[$$self] =
          in_bytes(lc($host_obj->get_san_dev_size(
                                              name     => $new_dev_name,
                                              protocol => $protocol,
                        )));

        confess("Invalid SAN device size detected!")
            if (!$dev_size[$$self] || 0 == in_bytes($dev_size[$$self]));

        $set_up[$$self] = 1;
    } ## end sub setup


    sub get_protocol {
        my $self = shift();

        return $self->get_lun_map()->get_igroup()->get_protocol();
    }


    sub cleanup {
        my $self = $_[0];

        return if (!$set_up[$$self]);

        $Log->debug("Starting cleanup of " . $self->get_description);

        my $san_service = $san_service[$$self];

        confess "No SAN service found?!" if (!$san_service);
        my $devices = $san_service->get_users(class=>'SANDevice');
        if ($devices){
            for my $d (@$devices){
                my $d_busy_datasets = $d->get_busy_dataset_users;
                my $count = $d_busy_datasets?scalar(@$d_busy_datasets):0;
                confess "Unable to cleanup. Found $count busy Perf::DataSet(s)" if ($count);
                $d->unlatch;
            }
        }

        # $Log->debug( 'There are ' . scalar(@$devices)
        #              . ' other iSCSI devices. Maybe we should leave iSCSI on');
        # if (0 < scalar(@$devices) - 1) {
        #     # San cleanup turns iSCSI off but doesn't turn it back on,
        #     # reenable it for the other devices
        #     $Log->debug('Reenabling iSCSI');
        #     $san_service->start();
        # }

        return;
    }
}

1;
