# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Volume - Representation of a volume on a filer

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Volume;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Object::InsideOut qw(Perf::Depot::Resource);
    use List::MoreUtils qw(any);
    use NACL::CS::Aggregate;
    use NACL::STask::Volume;


    my @appliance :Field
                  :Arg(name => 'appliance', mandatory => 1)
                  :Get(get_appliance)
                  :Type(Perf::Depot::Resource::Appliance);
    my @aggregate_name  :Field :Get(get_aggregate_name) :Set(set_aggregate_name);

    sub new {
        my $class = shift;
        if ($class ne __PACKAGE__){
            return $class->Object::InsideOut::new(@_);
        }
        my %args = @_;
        my $app = $args{appliance};
        confess "Argument 'appliance' is required and must be a Perf::Depot::Resource::Appliance"
            if (!$app || !ref($app) || !$app->isa('Perf::Depot::Resource::Appliance'));
        my $name = $args{name};
        confess "Argument 'name' is required" if (!$name);
        if ($app->isa('Perf::Depot::Resource::Filer')){
            my $vol = Perf::Depot->get_resource( class     =>'Volume::SevenMode',
                                                    name      =>$name,
                                                    appliance =>$app);
            if (!$vol){
                $vol = Perf::Depot->create_resource( class=>'Volume::SevenMode', %args );
            }
            else{
                $Log->comment("Depot already contains a Volume7Mode named '$name' on "
                                . $app->get_description);
            }
            return $vol;
        }
        elsif($app->isa('Perf::Depot::Resource::Vserver')){
            # TODO: Do we need to distinguish between striped and non-striped?
            my $vol = Perf::Depot->get_resource( class     =>'Volume::TenMode',
                                                    name      =>$name,
                                                    appliance =>$app);
            if (!$vol){
                $vol = Perf::Depot->create_resource( class=>'Volume::TenMode', %args );
            }
            else{
                $Log->comment("Depot already contains a Volume10Mode named '$name' on "
                                . $app->get_description ." on " . $app->get_description );
            }
            return $vol;
        }
        else{
            confess "Argument 'appliance' must be Filer or a Vserver";
        }
    }

    sub _init :Init {
        my ($self) = @_;
        $self->add_dependencies( $appliance[$$self] );
    }

    sub get_description {
        my ($self) = @_;
        return $self->SUPER::get_description . ' on ' . $appliance[$$self]->get_description;
    }

    sub export {
        confess "This method must be implemented in a subclass of " . __PACKAGE__;
    }

    sub get_export_path {
        confess "This method must be implemented in a subclass of " . __PACKAGE__;
    }

    sub export_status {
        confess "This method must be implemented in a subclass of " . __PACKAGE__;
    }

    sub is_exported_as_ro {
        confess "This method must be implemented in a subclass of " . __PACKAGE__;
    }

    sub get_space_available {
        confess "This method must be implemented in a subclass of " . __PACKAGE__;
    }

    sub get_type {
        confess "This method must be implemented in a subclass of " . __PACKAGE__;
    }

    sub is_in_large_aggr {
        confess "This method must be implemented in a subclass of " . __PACKAGE__;
    }

    sub get_luns {
        my $self = shift();

        return $appliance[$$self]->get_luns_in_volume($self);
    }

=head2 has_lun

Determine if a given LUN name is physically stored in this volume

=cut

    sub has_lun {
        my ($self, $lun_name) = @_;

        my $luns = $self->get_luns();
        my $matched = any { $_->get_name() eq $lun_name } @$luns;

        return $matched;
    }

    # dump the volume and hosting aggregate usage
    sub dump_size_usage {
        my $self = shift();
        my $appliance = $self->get_appliance();
        my $ci = $appliance->get_conn();
        my $vol = NACL::STask::Volume->new(
            command_interface => $ci,
            volume => $self->get_name(),
            vserver => $appliance->get_name(),
        );
        my $vol_state = $vol->state();

        $Log->comment('Volume usage:');
        foreach my $a (qw(total used available)) {
            $Log->comment("$a = " . $vol_state->$a());
        }

        foreach my $o (qw(show_footprint show_space)) {
            $vol->$o();
        }

        my $aggr_state = NACL::CS::Aggregate->fetch(
            command_interface => $ci,
            filter => {aggregate => $self->get_aggregate_name(),},
        );

        $Log->comment('Aggregate usage:');
        foreach my $a (qw(size usedsize availsize)) {
            $Log->comment("$a = " . $aggr_state->$a());
        }
    }
}
1;
