# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Volume::SevenMode - Representation of a 7G volume

=head1 DESCRIPTION

This class represent a 7-mode volume tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Volume::SevenMode;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Utils::Units qw(:constants);
    use NACL::CS::Volume;

    use namespace::clean;
    use Perf::Depot::Resource::Volume;
    use Object::InsideOut qw(Perf::Depot::Resource::Volume);

    sub get_filer {
        my $self = $_[0];
        return $self->get_appliance;
    }

    sub export {
        my ($self, %args) = @_;
        my $read_only = $args{read_only};
        my $opts_exportfs = ['anon=0','nosuid',$read_only?'ro':'rw'];
        my $opts_exportfs_string = join(',', @$opts_exportfs);
        my $vol = $self->get_name;

        # If we are about to set the 'read_only' option of the volume
        # make sure no other dataset is relying on it being writable
        my $current_exportfs = $self->get_parsed_exportfs;
        if ($current_exportfs && $read_only && $current_exportfs->{'rw'}){
            my $vol_datasets = $self->get_users( class => 'DataSet');
            if ($vol_datasets){
                my @blockers = grep { !$_->is_read_only } @$vol_datasets;
                if (@blockers){
                    my @blockers_description = map { $_->get_description } @blockers;
                    my $blockers_string = join(', ', @blockers_description);
                    confess "Unable to exportfs " . $self->get_description . " as read_only. The ".
                            "following resources require it to be writable: $blockers_string";
                }
            }
        }

        # Setting exportfs option is slow (~1s) and during that time calls to 'exportfs'
        # will fail. Make sure we set those options only when needed.
        my $properly_set = 1;
        for my $opt (@$opts_exportfs){
            if (!$current_exportfs->{$opt}){
                $properly_set = 0;
                last;
            }
        }
        if ($properly_set){
            $Log->comment("Export options on volume '". $self->get_name ."' already properly set");
            return;
        }

        # TODO: Should we make sure no busy Perf::DataSets are using
        # the export?
        $self->get_filer()
             ->get_fc()
             ->apiset()
             ->exportfs(path    => "/vol/$vol",
                        options => $opts_exportfs_string,
                        ignore  => 1,);

        # Check that the command worked (we've seen cases where it failed silently)
        $current_exportfs = $self->get_parsed_exportfs;
        for my $opt (@$opts_exportfs){
            confess "Unable to find exportfs option '$opt'" if (!$current_exportfs->{$opt});
        }
        return;
    }

    sub get_exportfs {
        my $self = $_[0];

        # no high level NACL classes interact with exportfs
        my $apiset = $self->get_filer()->get_fc()->apiset();
        my $res    = $apiset->exportfs();
        my $out    = $res->get_parsed_output();
        my $vol    = 'vol/' . $self->get_name();

        foreach my $export (@$out) {
            return $export->{export_options}
                if ($vol eq $export->{path});
        }

        confess("Volume $vol not found in exports");
    }

    sub get_parsed_exportfs {
        my ($self) = @_;
        my $raw = $self->get_exportfs;
        my %parsed = map {$_ => 1} split(/\s*,\s*/, $raw);
        return \%parsed;
    }

    sub is_exported_as_ro {
        my $self = $_[0];
        my $volume_export_opts = $self->get_exportfs;
        return 1 if ($volume_export_opts =~ /(?:,|^)ro(?:,|$)/);
        return 0;
    }

    sub get_aggregate_name {
        my $self = $_[0];
        my $aggrname = $self->SUPER::get_aggregate_name;
        if (!$aggrname){
            my $volume_state =
              NACL::CS::Volume->fetch(
                              command_interface => $self->get_filer()->get_fc(),
                              filter           => {volume => $self->get_name()},
                              requested_fields => ['aggregate'],);
            my $aggrname = $volume_state->aggregate();

            $self->set_aggregate_name($aggrname);
        }
        return $aggrname;
    }


    sub get_export_path {
        my $self = $_[0];
        return '/vol/' . $self->get_name;
    }

    sub get_type {
        my $self = $_[0];

        my $volume_state =
          NACL::CS::Volume->fetch(
                              command_interface => $self->get_filer()->get_fc(),
                              filter           => {volume => $self->get_name()},
                              requested_fields => ['raidstatus'],);

        # type is stored in raidstatus as a comma separated string
        # along with RAID type and word size
        my $status = $volume_state->raidstatus();
        my $type;

        if ($status =~ /flex/) {
            $type = 'flex';
        }
        elsif ($status =~ /trad/) {
            $type = 'trad';
        }
        elsif ($status =~ /striped/) {
            $type = 'striped';
        }

        return $type;
    }

    sub is_in_large_aggr {
        return 0; # TODO
    }

    sub get_physical_appliance {
        my $self = shift();

        return $self->get_filer();
    }
}
1;
