# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Volume::TenMode - Representation of a 10-mode volume on a filer

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Volume::TenMode;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use NACL::CS::Volume;
    use NACL::STask::Volume;

    use namespace::clean;
    use Perf::Depot::Resource::Volume;
    use Object::InsideOut qw(Perf::Depot::Resource::Volume);

    sub export {
        my ($self, %args) = @_;
        # The read_only argument is probably not necessary in 10-mode
        # TODO: confirm
        my $read_only = $args{read_only};
        $Log->comment("Ignoring read_only argument") if ($read_only);

        my $vserver = $self->get_vserver;
        my $volname = $self->get_name;
        my $junction_info = $self->get_junction_info;
        my $path = $junction_info->{'junction-path'};
        my $active = $junction_info->{'junction-active'};
        if ($path && $path ne '-' && $active eq 'true'){
            $Log->debug("Volume $volname is already exported with active junction-path '$path'");
            return $path;
        }

        $path = "/$volname" if (!$path || $path eq '-');

        # Create the junction or make it active it already exists,
        # associate it to the default export policy. Hopefully nobody mocked with it
        # TODO: validate default export policy or create new policy
        NACL::STask::Volume->mount(command_interface => $vserver->get_conn(),
                                   vserver           => $vserver->get_name(),
                                   volume            => $volname,
                                   'junction-path'   => $path,
                                   active            => 'true',);

        return $path;
    }

    sub get_junction_info {
        my $self    = $_[0];
        my @junction_fields = qw(junction-path junction-active);
        my $vol_state       = $self->get_state(\@junction_fields);
        my %ret             = map { $_ => $vol_state->$_() } @junction_fields;

        return \%ret;
    }

    sub get_export_path {
        my $self = $_[0];
        return $self->get_junction_info->{'junction-path'};
    }

    sub get_vserver {
        my $self = $_[0];
        return $self->get_appliance;
    }


    sub export_status {
        confess "TODO";
    }

    sub is_exported_as_ro {
        confess "TODO";
    }


    sub get_type {
        my $self    = $_[0];
        my $vol_state = $self->get_state(['volume-style']);

        return $vol_state->volume_style();
    }

    sub is_in_large_aggr {
        return 0;                                                      # TODO
    }


    sub get_aggregate_name {
        my $self     = shift();
        my $aggrname = $self->SUPER::get_aggregate_name();

        if (!$aggrname) {
            my $vol_state = $self->get_state(['aggregate']);

            $aggrname = $vol_state->aggregate();

            $self->set_aggregate_name($aggrname);
        }

        return $aggrname;
    }

    sub get_physical_appliance {
        my $self = shift();

        return $self->get_vserver()->get_physical_appliance();
    }


    sub get_state {
        my $self    = shift();
        my $fields  = shift();
        my $vserver = $self->get_vserver();
        my %args;

        if (defined($fields) && 0 < scalar(@$fields)) {
            $args{requested_fields} = $fields;
        }

        my $vol_state = NACL::CS::Volume->fetch(
            command_interface => $vserver->get_conn(),
            filter            => {
                volume  => $self->get_name(),
                vserver => $vserver->get_name(),},
            %args);

        return $vol_state;
    }
}
1;
