# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Vserver - Representation of a Vserver

=head1 DESCRIPTION

This class represent a Vserver tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Vserver;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Subtest qw();
    use Tharn qw( $Log subtest);
    use Scalar::Util qw(looks_like_number);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use NACL::C::SystemServicesNtpConfig;
    use NACL::C::VserverCifsOptions;
    use NACL::C::VserverCifsShareProperties;
    use NACL::CS::VserverCifsOptions;
    use NACL::MTask::CIFS;
    use NACL::STask::VserverCifsShare;
    use NACL::STask::VserverNfs;
    use NATE::ParamSet;
    use namespace::clean;
    use Perf::Depot::Resource::Appliance::GX;
    use Object::InsideOut qw(Perf::Depot::Resource::Appliance::GX);

    sub _init :Init {
        my $self = shift;
        my $cluster = $self->get_cluster;
        confess "You must specify a cluster"
            if (!$cluster || !ref($cluster) || !$cluster->isa('Perf::Depot::Resource::Cluster'));
        $self->add_dependencies($cluster);
    }

    sub get_conn {
        my $self = shift;
        return $self->get_cluster->get_conn;
    }

    sub can_ping {
        my ($self,$target) = @_;
        my $interfaces = $self->get_interfaces;

        # A Vserver is deemed able to ping another interface if any of its interfaces can ping it
        for my $interface (@$interfaces){
            my $interface_node = $interface->get_port_node;
            return 1 if ($interface_node->can_ping($target, {vserver => $self->get_name()}));
        }
        return 0;
    }

    sub sync {
        my ($self) = @_;
        # We want to sync all the nodes
        my $nodes = $self->get_cluster->get_nodes;
        my $synced = 0;
        for my $node (@$nodes){
            # Some nodes may be down
            eval{ $node->sync };
            if ($@){
                $Log->comment("Unable to sync '" . $node->get_description . "': $@");
            }
            else{
                $synced ++;
            }
        }
        if ($synced != @$nodes){
            $Log->comment("Could only sync $synced nodes out of " . @$nodes);
        }
        confess "Unable to sync a single node" if (!$synced);
    }

    sub add_user {
        my ($self, $resource) = @_;
        my $was_added = $self->SUPER::add_user($resource);
        return if (!$was_added);
        if ($resource->isa('Perf::Depot::Resource::DataSet')){
            my $dataset_type = $resource->get_type;
            if ($dataset_type =~ /cifs/) {
                $self->cifs_start;
            }
            elsif ($dataset_type eq 'nfs') {
                $self->nfs_start;
            }
            elsif ($dataset_type eq 'nfsv4') {
                $self->nfs_start(version => 4);
            }
            elsif ($dataset_type eq 'nfsv4.1') {
                # https://library.netapp.com/ecmdocs/ECMP1136576/html/GUID-332D7D40-4A0F-48E3-B1C6-9B5B07045D14.html
                # NFSv4.0 has to be on to use NFSv4.1
                $self->nfs_start(version => 4);
                $self->nfs_start(version => 4.1);
            }
            elsif ($dataset_type eq 'iscsi') {
                $self->iscsi_start();
            }
            elsif ($dataset_type eq 'fcp') {
                $self->fcp_start();
            }
        }
    }


    sub cifs_start {
        my $self    = $_[0];
        my $cluster = $self->get_cluster();
        my $node    = $cluster->get_node_names()->[0];
        my $vserver = $self->get_name();

        # CIFS server names are limited to 15 characters
        # http://support.microsoft.com/kb/909264
        (my $server_name = join('_', ($node, $vserver, 'cifs'))) =~
          s/^(.{15}).*/$1/;

        # most of the AD info is lab specific and stored in NATE params
        my $nate_params = NATE::ParamSet->get_global();
        my @name_servers = split(/,/, $nate_params->get('FILER_DNS_SERVER_LIST'));

        my $ci = $self->get_conn();
        my $cifs =
          NACL::MTask::CIFS->setup(
                        command_interface => $ci,
                        vserver           => $vserver,
                        'cifs-server'     => $server_name,
                        cifs_domain => $nate_params->get('CIFS_DOMAIN_LONG'),
                        username    => $nate_params->get('AD_SERVER_USER'),
                        password    => $nate_params->get('AD_SERVER_PASSWORD'),
                        dns_domain => $nate_params->get('FILER_DNS_DOMAINNAME'),
                        'name-servers' => \@name_servers,
                        if_exists      => 'reuse',
                        # not sure what this is, but it consistently
                        # fails if not disabled
                        nacltask_replay_timeout => 0,);
        sleep(3);
        NACL::C::SystemServicesNtpConfig->modify(command_interface => $self->get_conn(),
                                                 enabled => 'true');

        NACL::C::VserverCifsOptions->modify(command_interface   => $ci,
                                            vserver             => $vserver,
                                            'default-unix-user' => 'nobody',);

        # CIFS information can't be queried until at least one CIFS
        # server has been created
        $self->cifs_init();

        return;
    }


    # return an arrayref of enabled CIFS SMB features
    # [1, 2, 2.1, 3, copy_offload]
    sub cifs_get_smb_features {
        my $self = shift();
        my $state = NACL::CS::VserverCifsOptions->fetch(
            command_interface => $self->get_conn(),
            filter            => {vserver => $self->get_name()},
            requested_fields =>
              [qw(smb2-enabled smb3-enabled copy-offload-enabled)],);

        # SMB1 is always on
        my @features = (1);

        foreach my $feature (qw(smb2 smb3 copy_offload)) {
            my $method = "${feature}_enabled";

            if ($state->can($method)) {
                my $res = $state->$method();
                my $enabled = (defined($res) && 'true' eq $res);
                (my $feature_name = $feature) =~ s/smb//;

                push(@features, $feature_name)
                  if ($enabled);

                push(@features, 2.1)
                  if ( looks_like_number($feature_name)
                    && 2 == $feature_name
                    && $enabled);
            }
        }

        return \@features;
    }


    sub _cifs_set_smb_feature {
        my $self    = shift();
        my %arg     = @_;
        my $feature =
          (looks_like_number($arg{feature})
            ? "smb$arg{feature}"
            : $arg{feature}) . '-enabled';
        $feature =~ s/_/-/g;

        NACL::C::VserverCifsOptions->modify(
            command_interface => $self->get_conn(),
            vserver           => $self->get_name(),
            $feature          => ($arg{enabled} ? 'true' : 'false'),);
    }


    sub cifs_share_create {
        my $self = shift();
        my %args = @_;
        my $disable_oplocks = delete($args{disable_oplocks});

        $self->SUPER::cifs_share_create(%args);

        if ($disable_oplocks) {
            NACL::C::VserverCifsShareProperties->remove(
                command_interface => $self->get_conn(),
                vserver => $self->get_name(),
                'share-name' => $args{name},
                'share-properties' => ['oplocks'],);
        }
    }

    sub nfs_start {
        my ($self, %args) = @_;
        my $nfs_version = $args{version};
        my $conn        = $self->get_conn;
        my $vs_nfs =
          NACL::STask::VserverNfs->new(command_interface => $conn,
                                       vserver           => $self->get_name(),);

        # TODO: Hopefully no licensing is needed. Otherwise do it here

        # Recent ONTAP versions have nfs start/stop/status
        if (!$vs_nfs->status()) {
            $vs_nfs->start();

            confess('Unable to start NFS')
                if (!$vs_nfs->status());
        }

        # Enable nfsv4.X on the filer, if nfsv4.X is requested.
        if ($nfs_version && $nfs_version >= 4) {
            my $state = $vs_nfs->state();
            # The filer wants v4.0 or v4.1.
            # NACL wants v4 for CLI and v4_0 for ZAPI for 4.0 or v4_1 for 4.1
            my $full_version = sprintf('v%0.1f', $nfs_version);
            (my $method = $full_version) =~ s/\./_/;

            if (!$state->can($method)
                && (4 == $nfs_version && !$state->can('v4'))) {
                confess "NFSv$nfs_version not supported on this platform";
            }

            if ($state->$method() ne 'enabled') {
                $vs_nfs->modify($full_version => 'enabled');

                $state = $vs_nfs->get_one_state_attribute($method);

                confess "Unable to enable NFSv$nfs_version"
                  if ($state ne 'enabled');
            }
        }
    }


    sub get_nfsv4_domain_id {
        my ($self) = @_;
        my $conn = $self->get_conn;
        my $vs_nfs = NACL::STask::VserverNfs->new(command_interface => $conn,
                                                  vserver => $self->get_name(),);

        my $state = $vs_nfs->state();

        confess "Unable to find v4-id-domain"
            if (!$state->can('v4_id_domain'));

        return $state->v4_id_domain();
    }


=head2

Retrieve the current node being used to execute commands against this vserver.

=cut

    sub get_physical_appliance {
        my $self = shift();

        return $self->get_cluster()->get_execution_node();
    }


    sub get_iscsi_targets {
        my $self = shift();

        return $self->_get_iscsi_targets(vserver => $self->get_name());
    }


    sub _get_san_volume {
        my $self = shift();
        my $vol  = shift();

        confess(  "Didn't provide a Resource::Volume::TenMode.\n"
                . 'Got a: '
                . ref($vol))
          if (   !defined($vol)
              || !$vol->isa('Perf::Depot::Resource::Volume::TenMode'));

        my $san_vserver = $self->_get_san_vserver();
        my $san_volume = $san_vserver->get_volume_obj(name => $vol->get_name());

        confess($san_vserver->errmsg())
            if ($san_vserver->error());

        return $san_volume;
    }

    sub _get_san_filer {
        my $self  = shift();
        my $node  = $self->get_physical_appliance();

        return $node->_get_san_filer();
    }

    sub _get_san_vserver {
        my $self      = shift();
        my $san_filer = $self->_get_san_filer();
        my $san_vserver =
          $san_filer->get_vserver_obj(name => $self->get_name());

        confess($san_filer->errmsg())
          if ($san_filer->error());

        return $san_vserver;
    }


    # Create an iGroup on this Vserver.  The name given may manipulated,
    # so return the name of the iGroup actually created.
    sub igroup_create {
        my $self = shift();
        my %args = @_;
        my $name = $self->get_name();

        $args{vserver} = $name;
        $args{name}    = "$args{vserver}_$args{name}";

        return $self->_igroup_create(%args);
    }


    sub igroup_destroy {
        my $self   = shift();
        my $igroup = shift();
        my %args   = ( vserver => $self->get_name() );

        $self->_igroup_destroy($igroup, %args);
    }


    sub iscsi_start {
        my $self        = shift();
        my $san_vserver =
          $self->get_san_vserver($self->get_name());
        my $san_iscsi_obj =
          (   $san_vserver->iscsi_exist()
            ? $san_vserver->get_iscsi_obj()
            : $san_vserver->create_iscsi_obj());

        confess("Failed to get San:::iSCSI object\n" . $san_vserver->errmsg())
            if ($san_vserver->error());

        $san_iscsi_obj->start();

        confess("Failed to start iSCSI service\n" . $san_vserver->errmsg())
            if ($san_vserver->error());
    }


    sub panic {
        my $self  = shift();
        my $nodes = $self->get_cluster()->get_node_names();
        my @subtests;

        foreach my $node (@$nodes) {
            my $code = sub { NACL::STask::Node->new(name => $node)->panic(); };
            my $subtest = subtest($code, '-bg', '-runid' => "${node}_panic");

            push(@subtests, $subtest);
        }

        Subtest::wait_finish(subtests => \@subtests);

        return;
    }


    sub reconnect {
        my $self = shift();
        # Cluster->get_nodes() wants to talk to the cluster to check
        # the node list for changes. There may not be any live
        # connections when this is called, so bypass that check.
        my $nodes = $self->get_cluster()->get_nodes(skip_validation => 1);

        foreach my $node (@$nodes) {
            $node->reconnect();
        }
    }


    sub _get_san_fcp_obj {
        my $self        = shift();
        my $san_vserver = $self->get_san_vserver($self->get_name());
        my $san_fcp_obj = (  $san_vserver->fcp_exist()
                           ? $san_vserver->get_fcp_obj()
                           : $san_vserver->create_fcp_obj());

        confess("Failed to get San:::FCP object\n" . $san_vserver->errmsg())
          if ($san_vserver->error());

        return $san_fcp_obj;
    }



    sub fcp_start {
        my $self      = shift();
        my $san_filer = $self->_get_san_filer();

        $san_filer->license_add(service => 'fcp');

        my $san_fcp_obj = $self->_get_san_fcp_obj();

        my $fcp_status = $san_fcp_obj->status();

        confess("Failed to check FCP service\n" . $san_fcp_obj->errmsg())
          if ($san_fcp_obj->error());

        $san_fcp_obj->start()
          if ('up' ne $fcp_status->{'status_admin'});

        confess("Failed to start FCP service\n" . $san_fcp_obj->errmsg())
          if ($san_fcp_obj->error());
    }


    sub detect_fcp_adapters {
        my $self        = shift();
        my $san_fcp_obj = $self->_get_san_fcp_obj();

        my @lifs = $san_fcp_obj->get_lifs();

        confess("Failed to get FCP adapters\n" . $san_fcp_obj->errmsg())
          if ($san_fcp_obj->error());

        if (0 == scalar(@lifs)) {
            $Log->comment('No FCP target LIFs exist, attempting to create');
            $self->_create_fcp_targets();

            # the previous call rebooted the filer, reconnect
            $self->_get_san_filer()->reconnect();

            $san_fcp_obj->_refresh_lifs();

            confess("Failed to refresh FCP adapters\n" . $san_fcp_obj->errmsg())
                if ($san_fcp_obj->error());

            @lifs = $san_fcp_obj->get_lifs();

            confess("Failed to get FCP adapters\n" . $san_fcp_obj->errmsg())
                if ($san_fcp_obj->error());

            confess('No FCP target LIFs were created')
              if (0 == scalar(@lifs));
        }

        foreach my $lif (@lifs) {
            (my $wwpn = $lif->get_wwpn()) =~ s/://g;

            my %fc_id = (class     => 'Port::FCP',
                         name      => $lif->get_name(),
                         wwpn      => $wwpn,
                         container => $self);
            my $fc_lif = Perf::Depot->get_resource(%fc_id);

            if (!$fc_lif) {
                $fc_lif = Perf::Depot->create_resource(%fc_id);
                $fc_lif->register();
            }
        }
    } ## end sub detect_fcp_adapters


    # gx_boot blows away FCP targets. To reinstate them, we have to
    # toggle the adapter firmware from initiator to target, reboot the
    # filer to load the firmware, then create the FCP data LIFs
    sub _create_fcp_targets {
        my $self = shift();
        my $conn = $self->get_conn();
        my $name = $self->get_name();
        my @subtests;

        foreach my $node (@{$self->get_cluster()->get_nodes()}) {
            my $code = sub {
                my $new_ports = $node->toggle_fcp_ports();
                my $node_name = $node->get_name();

                # wait for SAN services to come back online
                sleep(2 * 60);

                foreach my $i (0 .. $#{$new_ports}) {
                    $conn->apiset()->network_interface_create(
                                                vserver => $name,
                                                lif  => $node_name . "_fcp$i",
                                                role => 'data',
                                                'data-protocol' => 'fcp',
                                                'home-node'     => $node_name,
                                                'home-port' => $new_ports->[$i],
                                                );
                }
            };

            my $subtest = subtest($code, '-bg',
                                 '-runid' => $node->get_name() . '_toggle_fcp');

            push(@subtests, $subtest);
        }

        Subtest::wait_finish(subtests => \@subtests);

        return;
    } ## end sub _create_fcp_targets


    sub get_visible_fcp_initiators {
        my $self        = shift();
        my $san_fcp_obj = $self->_get_san_fcp_obj();
        my $inits       = $san_fcp_obj->get_initiators();
        my %inits;

        foreach my $init (@$inits) {
            (my $wwpn = $init->{wwpn}) =~ s/://g;
            push(@{$inits{$init->{lif}}}, $wwpn);
        }

        return \%inits;
    }


    sub get_fcp_targets {
        my $self = shift();

        return $self->_get_fcp_targets(vserver => $self->get_name());
    }
}

1;
