# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary
## @status   private
##
## @pod here

=head1 NAME

Resource::Worker - Representation of a Worker in a process

=head1 DESCRIPTION

This class represent a filer tracked by a Perf::DataSet.
It is written using the "inside-out" object paradigm and the package Object::InsideOut specifically.

=over

=item References

=over

=item More on inside-out objects:      http://www.perlfoundation.org/perl5/index.cgi?inside_out_object

=item Object::InsideOut Documentation: http://search.cpan.org/dist/Object-InsideOut/lib/Object/InsideOut.pod

=back

=back

=head1 METHODS

=cut

package Perf::Depot::Resource::Worker;
{
    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw( $Log );
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Storable qw(nfreeze);
    use MIME::Base64;
    use Socket qw(inet_ntoa);
    use Object::InsideOut qw(Perf::Depot::Resource);

    my @process :Field :Get(get_process) :Arg(Name=>'process', Mandatory=>1) :Type(Perf::Depot::Resource::Process);
    my @order   :Field :Get(get_order);
    my @forked_to :Field :Name('forked_to');
    my @pending_new :Field :Name('pending_new');

    sub _init :Init {
        my $self = $_[0];
        $order[$$self] = 0;
        $forked_to[$$self] = {};
        $pending_new[$$self] = 0;

        # Workers belonging to the same Perf::DataSet share the same name
        my $current_workers = $self->get_resources( class=>'Worker', name=>$self->get_name );
        $order[$$self] = scalar(@$current_workers) if ($current_workers);
        $self->add_dependencies( $process[$$self] );

        #
        # Create Worker in process
        #
        my $dataset_uid = $self->get_name;
        my $parent_uid = $self->get_parent_uid;
        my $parent_worker;
        if ($parent_uid){
            # Find out if parent worker is in the same process or not
            $parent_worker = Perf::Depot->get_resource(  class=>'Worker',
                                                            name=>$parent_uid,
                                                            process=>$process[$$self]);
        }
        if ($parent_worker){
            #
            # In-process fork
            #
            $Log->comment("Starting fork of " . $parent_worker->get_description
                            . "into " . $self->get_description);
            $self->execute(code=>'sub {$::p->{'. $self->get_name .'} = ' .
                                 '$::p->{'. $parent_worker->get_name .'}->fork(@_); 1;}->',
                           args=>['uid', $self->get_name],
                           async=>1,
                           );
            $pending_new[$$self] = 1;
            $parent_worker->add_forked_to( $self->get_name );
        }
        else{
            my $async = 1;

            if ($parent_uid) {
                # The worker has a parent_uid set but no parent worker was found in the same
                # process. This means a detached fork is taking place. The remote new can only
                # be done synchronously unfortunately
                $async = 0;
            }

            # Initial worker for process
            my $seed = Perf::Depot::random_number();
            $self->execute( code=>'sub {$::p->{'.$dataset_uid.'}=Perf::DataSet::Worker->new(@_);1}->',
                            args=>['uid', $dataset_uid,
                                    'seed', $seed,
                                    'process_uid',$process[$$self]->get_name,
                                    ],
                            async=>$async,
                          );
            $pending_new[$$self] = 1 if ($async);
            if($parent_uid){
                #
                # Detached fork
                #
                # Find a suitable parent_worker on the same host
                my $host = $self->get_host;
                my $host_workers = Perf::Depot->get_resources( class=>'Worker',
                                                                  host=>$host,
                                                                  name=>$parent_uid);
                confess "No workers found on " . $host->get_description
                        . " within Perf::DataSet '$parent_uid'" if (!$host_workers);
                my @suitable_workers = grep { ! $_->was_forked_to( $dataset_uid ) } @$host_workers;
                confess "No suitable workers found on " . $host->get_description
                        . " within Perf::DataSet '$parent_uid'" if (!@suitable_workers);
                $parent_worker = $suitable_workers[0];

                # Dump parent_worker content to disk
                my $filepath = $parent_worker->freeze;
                # Load it asynchronously
                $self->thaw($filepath);
                $pending_new[$$self] = 1;
                $parent_worker->add_forked_to( $self->get_name );
            }
        }
    }

    sub wait_new {
        my $self = shift;
        confess "Worker isn't in pending new state" if (!$pending_new[$$self]);
        # read_reply returns immediately if there is no reply
        # first we must wait for the reply to show up
        $process[$$self]->wait_reply;
        my $r = $process[$$self]->read_reply;
        $pending_new[$$self] = 0;
        return $r;
    }

    sub freeze {
        my $self = shift;
        $Log->comment("Starting freeze of " . $self->get_description);
        my $filepath = $self->execute(code=>'$::p->{'. $self->get_name .'}->freeze', args=>[@_], async=>0);
        confess "No filepath returned after freeze of " . $self->get_description if (!$filepath);
        return $filepath;
    }

    sub thaw {
        my ($self, $filepath, %args) = @_;
        confess "You must specify an argument" if (!$filepath);
        $Log->comment("Starting thaw of " . $self->get_description);
        $self->execute(code=>'sub {$::p->{'. $self->get_name .'}->thaw(@_)}->', args=>[$filepath,%args], async=>1);
        return;
    }

    sub was_forked_to {
        my ($self, $dataset_uid) = @_;
        confess "You must specify an argument" if (!$dataset_uid);
        return 1 if ($forked_to[$$self]->{$dataset_uid});
        return 0;
    }

    sub add_forked_to {
        my ($self, $dataset_uid) = @_;
        confess "You must specify an argument" if (!$dataset_uid);
        $forked_to[$$self]->{$dataset_uid} = 1;
        return;
    }

    sub get_description {
        my $self = $_[0];
        return $self->SUPER::get_description . " in " . $process[$$self]->get_description;
    }

    sub get_process_name {
        my $self = $_[0];
        return $process[$$self]->get_name;
    }

    sub get_host {
        my $self = $_[0];
        my $host = $process[$$self]->get_host;
        confess "No host found in " . $process[$$self]->get_description if (!$host);
        return $host;
    }

    sub get_parent_uid {
        my $self = $_[0];
        my $dataset = $self->get_resource(class => 'DataSet', name=>$self->get_name);
        confess "No Perf::DataSet with name '" . $self->get_name . "' found!" if (!$dataset);
        return $dataset->get_parent_uid;
    }


    sub get_appliance_ips {
        my $self = shift();
        my $ips = $self->get_dependencies(class => 'Interface');

        confess('No appliance interface resources found for '
            . $self->get_description())
            if (!defined($ips) || 0 == scalar(@$ips));

        return $ips;
    }


    sub get_appliance_ip {
        my $self = shift();

        my $ips = $self->get_appliance_ips();

        my $filer_ip;
        for my $ip (@$ips){
            if ($ip->get_port->isa('Perf::Depot::Resource::Port::Ethernet::Appliance')){
                confess "Found more than one appliance interfaces!" if ($filer_ip);
                $filer_ip = $ip;
            }
        }
        confess "No appliance interfaces found for " . $self->get_description if (!$filer_ip);
        return $filer_ip;
    }


    sub get_appliance_fcp_ports {
        my $self = shift();
        my $ports = $self->get_dependencies(class => 'Port::FCP');

        confess('No FCP ports found for: ' . $self->get_description())
            if (!defined($ports) || 0 == scalar(@$ports));

        my @filer_ports;
        foreach my $port (@$ports) {
            if ($port->get_container()->isa('Perf::Depot::Resource::Appliance')) {
                push(@filer_ports, $port);
            }
        }

        confess('No appliance FCP ports found for: ' . $self->get_description())
            if (0 == scalar(@filer_ports));

        return \@filer_ports;
    }


    sub execute {
        my ($self, %params) = @_;
        confess "You must call wait_new first" if ($pending_new[$$self]);
        my $code = $params{code};
        my $command = $params{command};
        if (!$code && !$command){
            confess "You must specify code or command";
        }
        if ($code && $command){
            confess "You cannot specify both code and command";
        }
        my $args = $params{args};
        if (defined($args) and ref($args) ne 'ARRAY'){
            confess "Argument 'args' must be an ARRAY reference";
        }
        my $async = $params{async};
        if (!defined($async)){
            $async = 1;
        }
        if ($command){
            $code = '$::p->{' . $self->get_name . '}->execute';
            my $frozen_command = nfreeze($command);
            my $encoded_frozen_command = encode_base64($frozen_command, '');
            $args = [ $encoded_frozen_command ];
        }
        $Log->debug($self->get_description . " executing code: $code");
        if ($async){
            $process[$$self]->call_request($code, $args);
            return;
        }
        else{
            my $ret = $process[$$self]->call($code, $args);
            return $ret;
        }
    }

    sub cleanup {
        my $self = $_[0];
        # Dereference the Worker object from the process
        $self->execute(code => 'delete $::p->{' . $self->get_name .'};1;', async=>0);
        return;
    }

    # Rename/relocate an element in the Worker's container
    sub relocate_element {
        my ($self, $old_location, $new_location) = @_;

        $self->execute(code  => 'sub { $::p->{' . $self->get_name()
                                          . '}->relocate_element(@_); 1;}->',
                       args  => ['old', $old_location, 'new', $new_location],
                       async => 0);
    }
}
1;
