# -*- perl -*-

#----------------------------------------------------------------------
# Disk Performance general development utility libraries
#----------------------------------------------------------------------
# Description:
#   Disk Performance general development utility libraries
#   for use in unit tests and CIT testing.
#----------------------------------------------------------------------
# Initial author(s):
#   Eric Youngquist (on loan from PerfQA)
#   Scott Brasington
#----------------------------------------------------------------------
#

use strict;
use warnings;

package Perf::Dev::DevUtils;

use NATE::ParamSet;

use NACL::APISet::Exceptions::ResponseException qw(:try);

use Tharn;
use TCD;

use Data::Dumper;
$Data::Dumper::Indent=1;

use Perf::Utils qw(in_blocks in_bytes);

#----------------------------------------------------------------------
# Convienence functions

# usage: mkfile [-x stream_name] [-z] size filename [filecount [stepsize]]
#             -z              Detect and deduplicate zeroed data (only valid on
#                             vols with dedup enabled.)
#             filecount       Number of files to create, the names will be
#                             filename.0, filenanem.1 ...,
#             size            Units on the size can be:
#                             b/B     - blocks
#                             k/K     - kilobytes
#                             m/M     - megabytes
#                             g/G     - gigabytes
#             stepsize        number of blocks to skip in between each write

use File::Spec;

sub mkfile_name {

    my $_name = (caller(0))[3];

    $Log->debug("in $_name");

    my ($vol, $size, $prefix) = @_;

    use File::Temp;

    my $template = "XXXXXXXX";

    my $tmp_suffix = mktemp( $template );

    my $file = "${prefix}file_${tmp_suffix}_${size}";

    my $filename = File::Spec->join("/vol", $vol, $file);

    # TODO: try / catch this.
    # See UnitTest: ~/p4/test_main/lib/NACL/UnitTest/APISet/Node/CLI/7Mode/mkfile.thpl

    return $filename;

}

sub _mkfile {

    my $_name = (caller(0))[3];

    $Log->debug("in $_name");

    my ($apiset, $vol, $size, $prefix) = @_;

    my $filename = _mkfile_name($vol, $size, $prefix);

    # TODO: try / catch this.
    # See UnitTest: ~/p4/test_main/lib/NACL/UnitTest/APISet/Node/CLI/7Mode/mkfile.thpl

    my $response = $apiset->mkfile(
        size => $size,
        filename => $filename,
    );

    return $filename;

}

sub _mkfiles {

    my $_name = (caller(0))[3];

    $Log->debug("in $_name");

    my ($vol, $size, $prefix, $number) = @_;

    if ( ! defined $number ) {
        $number = 5;
    }

    my @files;

    foreach ( 1..$number ) {
        push(@files, _mkfile($vol, $size, $prefix));
    }

    return @files;

}

sub scramble_files {

    my ($block_size, @files) = @_;

    my $_name = (caller(0))[3];

    $Log->debug("in $_name");

    if ( $#files < 0) {
        die "ERROR: no files found.";
    }

    my $return_val=0;

    foreach my $file (@files) {
        my $ret = scramble_path($file, $block_size);
        if ( $ret ) {
            # In here, something happened.  If ret = 1, then issue a warning,
            # else return FAIL.
            if ( $ret == 1 ) {
                $return_val = 1;
            } else {
                return 2;
            }
        }
    }

    return $return_val;

}

sub scramble_path {

    my $_name = (caller(0))[3];

    $Log->debug("in $_name");

    my ($apiset, $file, $blocks) = @_;

    # Not in ApiSet, so run "execute_command"

    my $nate_params = NATE::ParamSet->new(global => 1);

    my $scramble_path_pattern_def = "0x112";
    my $scramble_path_pattern = $nate_params->get("SCRAMBLE_PATH_PATTERN", default => $scramble_path_pattern_def);

    my $scramble_path_timeout_def = 60*10;
    my $scramble_path_timeout = $nate_params->get("SCRAMBLE_PATH_TIMEOUT", default => $scramble_path_timeout_def);

    $apiset->set_timeout(timeout => $scramble_path_timeout);

    # Filed NACL Burt 647480 requesting APISET->scramble support for path.
    my $response = $apiset->execute_command(
        command => "scramble path -n $blocks -p $scramble_path_pattern -a $file",
    );

    # Check for errors / warnings

    my $failed_grep_str = ' Invalid path ';

    my $error = grep (/$failed_grep_str/, split("\n", $response));

    my $error_2=1;

    foreach my $line ( split("\n", $response) ) {
        if ( $line =~ /overwrote (\d*)/) {
            my $num = $1;
            if ( $num > 0 ) {
                $error_2=0;
                last;
            }
        }
    }

    if ( $error || $error_2 ) {
        if ( $error ) {
            $Log->error("scramble_path failed:\n$response:\nFailed one of the expressions: \"$failed_grep_str\"");
        }
        if ( $error_2 ) {
            $Log->error("scramble_path failed as 'overwrote 0 blocks' occurred:\n$response:");
        }
        return 2;
    }

    # I was told by dev that "failed to .." string is most likely not
    # an error at this time.  I filed burt burt647412 on this.
    # In the meantime, I am setting this as a warning.

    my $warn = grep (/^failed to/, split("\n", $response));

    if ( $warn ) {
        $Log->comment("scramble_path 'failed to' message:\n$response");
        return 1;
    }

    return 0;

}

sub _wait_on_scanner {

    my $_name = (caller(0))[3];

    $Log->debug("in $_name");

    my ($apiset, $aggr_name, $scanner_id, $sleep_time_max) = @_;

    use List::MoreUtils;

    my $sleep_time_def = 5;

    my $sleep_time = ($sleep_time_max) ? $sleep_time_max : $sleep_time_def;

    my $still_running = 1;

    while ( $still_running ) {

        my $response = $apiset->wafl_scan_status(
            aggregate => $aggr_name,
        );

        my $info = $response->get_parsed_output()->[0]->{info};

        # Extract the scan_id from run all running scanners
        my @scan_id_array = map {$_->{scan_id}} @{$info};

        # See if we have a match with our saved scan_id
        my $what = List::MoreUtils::first_index {/$scanner_id/} @scan_id_array;

        if ( $what != 1 ) {
            $still_running=0;
            last;
        }

        $Log->debug("_wait_on_scanner::wait: sleeping: $sleep_time: -- what: $what : -- still_running: $still_running :");

        sleep($sleep_time);
    }

}

sub _get_scanner_id {

    my ($apiset, $aggr_name, $type_of_scan, $max_cnt) = @_;

    my $_name = (caller(0))[3];

    $Log->debug("in $_name");

    # Defaults

    my $cnt_def = 15;

    # TODO: validate params

    if ( ! $aggr_name ) {
        die "ERROR: could not determine aggr_name:$aggr_name:. It appears to be blank.";
    }

    if ( ! $type_of_scan ) {
        die "ERROR: could not determine type of scan:$type_of_scan:. It appears to be blank.";
    }

    my $cnt = ($max_cnt) ? $max_cnt : $cnt_def;

    my @scanner_ids;

    while ($cnt > 0) { 

        my $response = $apiset->wafl_scan_status(
            aggregate => $aggr_name,
        );

        my $info = $response->get_parsed_output()->[0]->{info};

        my $scanner_id;

        foreach my $ii ( @{$info} ) {
            $Log->debug("ii->{type_of_scan}: " . $ii->{type_of_scan} . " :\n");
            if ( $ii->{type_of_scan} =~ /$type_of_scan/) {
                $scanner_id = $ii->{scan_id};
                push (@scanner_ids, $scanner_id);
                $cnt=0;
                last;
            }
        }

        Tharn::snooze(1);
        $cnt--;
        next;
    }

    if ( $#scanner_ids < 0 ) {
        die "ERROR: scanner id was not found ... ";
    }

    return @scanner_ids;

}

sub _wait_for_filersio {

    my $_name = (caller(0))[3];

    $Log->debug("in $_name");

    my ($apiset, $runtime) = @_;

    if ( ! $runtime ) {
        die "ERROR: $_name: runtime not passed in.\n";
    }

    # Loop over filersio_status until it finishes.  When it finishes, it
    # throws an exception that no instances currently exist.
    # The loop is setup from a range from 1..N, where is
    #   the runtime / sleeptime + 5 for a buffer.

    my $sleep_def = 10;

    my $sleep_time = $sleep_def;

    my $up_range = int ($runtime / $sleep_def) + 5;

    my $done = 0;

    # Initial set the return value for a failure.
    my $return_val = 2;

    $Log->debug("return_val: $return_val : -- up_range: $up_range :");

    foreach my $cntr ( 1..$up_range ) {
        $Log->debug("cntr a): $cntr : -- return_val: $return_val :");
        try {
            my $response = $apiset->filersio_status();
            $Log->debug("Data::Dumper->Dumper(filersio_status - response):" . Data::Dumper->Dumper($response));
            sleep($sleep_time);
        } catch NACL::APISet::Exceptions::CommandFailedException with {
            my $exception_object = shift;
            my $resp_text = $exception_object->text();
            $Log->debug("==> resp: text: " . $resp_text);
            # If we hit an exception that is not the string below, re-throw.
            if ( ! $resp_text =~ /filersio: no instances currently running/ ) {
                my $resp             = $exception_object->get_response_object();
                $Log->comment( "Caught a " . ref($exception_object) . "!" );
                $Log->comment( "Error is: " . $exception_object->text() );
                $Log->comment( "Raw output was:\n" . $resp->get_raw_output() );
                $exception_object->throw();
            }
            $Log->debug("cntr c): $cntr :");
            $done = 1;
        };   # NOTE: must have this semicolon
        if ( $done ) {
            # If we get here, all is well.
            $return_val = 0;
            last;
        }
    }

    $Log->debug("return_val: $return_val :");

    return $return_val;
}

use Params::Validate qw(:all);


#
# TODO: Consider wrapping these file operations into some
#       sort of fileObject that is manipulated as an object.
#

# http://wikid.netapp.com/w/Filersio

sub filersio_asyncio_active {

    my $_name = (caller(0))[3];

    $Log->debug("in $_name");

    my %params = Params::Validate::validate(
        @_,
        {
            'apiset' => {
                type     => OBJECT,
                optional => 0,
            },
            'read-percentage' => {
                type     => SCALAR,
                optional => 0,
            },
            'rand-percentage' => {
                type     => SCALAR,
                optional => 0,
            },
            'blk-size' => {
                type     => SCALAR,
                optional => 0,
            },
            'start-byte' => {
                type     => SCALAR,
                optional => 1,
                default => 0,
            },
            'end-byte' => {
                type     => SCALAR,
                optional => 0,
            },
            'run-time' => {
                type     => SCALAR,
                optional => 0,
            },
            'num-active' => {
                type     => SCALAR,
                optional => 1,
                default => 1,
            },
            'filename' => {
                type     => SCALAR,
                optional => 0,
            },
            'synchronous' => {
                type     => SCALAR,
                optional => 1,
                default => 1,
            },
            'create' => {
                type     => SCALAR,
                optional => 1,
                default => 0,
            },
            'fillonce' => {
                type     => SCALAR,
                optional => 1,
            },
            'fill-pattern' => {
                type     => SCALAR,
                optional => 1,
            },
            'runonce' => {
                type     => SCALAR,
                optional => 1,
            },
            'exp-constant' => {
                type     => SCALAR,
                optional => 1,
            },
            'interval' => {
                type     => SCALAR,
                optional => 1,
            },
            'verifyonce' => {
                type     => SCALAR,
                optional => 1,
            },
            'verify-pattern' => {
                type     => SCALAR,
                optional => 1,
            },
            'maxio' => {
                type     => SCALAR,
                optional => 1,
            },
        }
    );

    # Set apiset and remove it from params

    my $apiset = delete $params{'apiset'};

    my $response = $apiset->filersio_asyncio_active(
        %params,
    );

    return $response;

}

sub filersio_asyncio_pause {

    my $_name = (caller(0))[3];

    $Log->debug("in $_name");

    my %params = Params::Validate::validate(
        @_,
        {
            'apiset' => {
                type     => OBJECT,
                optional => 0,
            },
            'read-percentage' => {
                type     => SCALAR,
                optional => 0,
            },
            'rand-percentage' => {
                type     => SCALAR,
                optional => 0,
            },
            'blk-size' => {
                type     => SCALAR,
                optional => 0,
            },
            'start-byte' => {
                type     => SCALAR,
                optional => 1,
                default => 0,
            },
            'end-byte' => {
                type     => SCALAR,
                optional => 0,
            },
            'run-time' => {
                type     => SCALAR,
                optional => 0,
            },
            'pause' => {
                type     => SCALAR,
                optional => 0,
            },
            'count' => {
                type     => SCALAR,
                optional => 0,
            },
            'filename' => {
                type     => SCALAR,
                optional => 0,
            },
            'synchronous' => {
                type     => SCALAR,
                optional => 1,
                default => 1,
            },
            'create' => {
                type     => SCALAR,
                optional => 1,
                default => 0,
            },
            'fillonce' => {
                type     => SCALAR,
                optional => 1,
            },
            'fill-pattern' => {
                type     => SCALAR,
                optional => 1,
            },
            'runonce' => {
                type     => SCALAR,
                optional => 1,
            },
            'exp-constant' => {
                type     => SCALAR,
                optional => 1,
            },
            'interval' => {
                type     => SCALAR,
                optional => 1,
            },
            'verifyonce' => {
                type     => SCALAR,
                optional => 1,
            },
            'verify-pattern' => {
                type     => SCALAR,
                optional => 1,
            },
            'maxio' => {
                type     => SCALAR,
                optional => 1,
            },
        }
    );

    # Set apiset and remove it from params

    my $apiset = delete $params{'apiset'};

    my $response = $apiset->filersio_asyncio_pause(
        %params,
    );

    return $response;

}

#
# internal helper function to dump exception object info
#
sub generic_print_exception {
    my ($exception_object) = @_;

    my $resp_object = $exception_object->get_response_object();

    $Log->comment("<<::::::::::::BEG::::::::::::::>>\n");
    $Log->comment("Caught:    " . ref($exception_object));
    $Log->comment("Exception: " . $exception_object->text() );
    $Log->comment("Raw Output:\n" . $resp_object->get_raw_output() );
    $Log->comment(">>::::::::::::END::::::::::::::<<\n");
}


#
# Create a file that has a pattern written throughout the file
# This function uses builtin ontap filersio to create and fill the file.
# It can be used in conjunction with verify_file_pattern()
#
#
sub create_file_pattern {

    my $_name = (caller(0))[3];
    $Log->debug("in $_name");

    my %create_params = Params::Validate::validate(
        @_,
        {
# required parameters, these MUST be specified by the caller
            'apiset' => {
                type     => OBJECT,
                optional => 0,
            },
            'filesize' => {
                type     => SCALAR,
                optional => 0,
            },
            'filename' => {
                type     => SCALAR,
                optional => 0,
            },
# optional parameters, these can optionally be specified by the caller
            'pattern' => {
                type     => SCALAR,
                optional => 1,
                default  => -2,  # unique pattern each block
            },
        }
    );

    my $apiset = $create_params{'apiset'};

    $Log->comment("creating file: $create_params{'filename'} size: $create_params{'filesize'}");

# TODO: fixme hard coded timeout
    $apiset->set_timeout(timeout => 600);

# Note: NACL returns a response OBJECT with the statistics
    my $response = $apiset->filersio_asyncio_active(
        'read-percentage'    => 0,   # no reads, 100% writes
        'rand-percentage'    => 0,
        'blk-size'           => '4k',
        'start-byte'         => 0,
        'end-byte'           => $create_params{'filesize'},
        'run-time'           => 0,   # run until completion
        'num-active'         => 1,
        'filename'           => $create_params{'filename'},
        'fillonce'           => $create_params{'pattern'},
        'create'             => 1,
        'synchronous'        => 1,
    );

# force newly created blocks out to disk and out of memory
    $apiset->wafl_flush_cache();

# TODO: anything else that should be verified about the file after created?

    return $response;
}


#
# Verify the contents of a file written with a pattern throughout.
# This function uses builtin ontap filersio to verify the file contents
# It can be used in conjunction with create_file_pattern()
#
#
sub verify_file_pattern {

    my $_name = (caller(0))[3];
    $Log->debug("in $_name");

    my %verify_params = Params::Validate::validate(
        @_,
        {
# required parameters, these MUST be specified by the caller
            'apiset' => {
                type     => OBJECT,
                optional => 0,
            },
            'filesize' => {
                type     => SCALAR,
                optional => 0,
            },
            'filename' => {
                type     => SCALAR,
                optional => 0,
            },
# optional parameters, these can optionally be specified by the caller
            'pattern' => {
                type     => SCALAR,
                optional => 1,
                default  => -2,  # unique pattern each block
            },
        }
    );

    my $apiset = $verify_params{'apiset'};

    $Log->comment("verifying file: $verify_params{'filename'} size: $verify_params{'filesize'}");

# TODO: fixme hard coded timeout
    $apiset->set_timeout(timeout => 600);

# force blocks out of memory so that verify reads must come from disk
    $apiset->wafl_flush_cache();

    my $verify_error = 0;

#
# A try/catch block is used so that any unexpected output or
# exception from the NACL implementation does not cause the
# calling test case to fail.  Addtionally it allows more
# through checking if the verify actually passed or not.
#
    try {

# Verify file contents using builtin ontap filersio

# Note: NACL returns a response OBJECT with the statistics
        my $response = $apiset->filersio_asyncio_active(
            'read-percentage'    => 100, # 100% read, no writes
            'rand-percentage'    => 0,
            'blk-size'           => '4k',
            'start-byte'         => 0,
            'end-byte'           => $verify_params{'filesize'},
            'run-time'           => 0,   # run until completion
            'num-active'         => 1,
            'filename'           => $verify_params{'filename'},
            'verifyonce'         => $verify_params{'pattern'},
            'synchronous'        => 1,
        );

#
# Note: there are some flaws in the filersio implementation that if
# the ontap console messages from filersio arrive in different orders, 
# that the NACL filersio parser will not detect that the verify has failed.  
# The NACL code only will throw the exception if the unexpected message 
# is printed during the statistics portion of the output.  If the
# unexpected message is printed before or after the stats then the
# NACL code may return without throwing an exception which will appear
# as if the verify passed when it actually failed.  We try and deal
# with this limitation here.

# Temporary to aid in debug, dump the response object
        $Log->comment("Data::Dumper->Dumper(response):"
                      . Data::Dumper->Dumper($response));
        $Log->comment("Data::Dumper->Dumper(filersio verify response object):"
                      . Data::Dumper->Dumper($response->get_parsed_output()));

# TODO: check response object and see if $verify_error needs to be set to fail
#
# For now, any non-exception during verify is considered a verify pass.
        $verify_error=0; # temporary if command finishes we indicate pass

    } catch NACL::APISet::Exceptions::UnexpectedOutputException with {
        my $exception_object = shift;
#
# The reason we catch this exception, is we do not want any
# unexpected output from filersio to make this testcase fail.
# Instead leave it to the caller or outer loop to decide if
# the file fails to fragment enough.   This way if filersio has
# some indavertent message or runs out of walf buffers or whatever
# that this function continues and just keeps trying to fragment.
# The outer loop and timeouts will prevent it from running forever.
#
        $Log->comment("$_name: caught filersio exception");
        generic_print_exception($exception_object);
#
# For now, any exception during verify is considered a verify failure.
        $verify_error=1; # Note: cannot return from inside a try/catch block.
    }; # NOTE: must have this semicolon

    $Log->comment("$_name: return value = $verify_error");
    return $verify_error;
}


#
# read a file sequentially
# This function uses builtin ontap filersio to perform the read
#
#
sub read_file_sequentially {

    my $_name = (caller(0))[3];
    $Log->debug("in $_name");

    my %read_params = Params::Validate::validate(
        @_,
        {
# required parameters, these MUST be specified by the caller
            'apiset' => {
                type     => OBJECT,
                optional => 0,
            },
            'filesize' => {
                type     => SCALAR,
                optional => 0,
            },
            'filename' => {
                type     => SCALAR,
                optional => 0,
            },
# optional parameters, these can optionally be specified by the caller
        }
    );

    my $apiset = $read_params{'apiset'};

    $Log->comment("reading file: $read_params{'filename'} size: $read_params{'filesize'}");

# TODO: fixme hard coded timeout
    $apiset->set_timeout(timeout => 600);


# force blocks out of memory so that new reads come from disk
    $apiset->wafl_flush_cache();

# Note: NACL returns a response OBJECT with the statistics
    my $response = $apiset->filersio_asyncio_active(
        'read-percentage'    => 100, # 100% reads, no writes
        'rand-percentage'    => 0,
        'blk-size'           => '4k',
        'start-byte'         => 0,
        'end-byte'           => $read_params{'filesize'},
        'run-time'           => 0,   # run until completion
        'num-active'         => 1,
        'filename'           => $read_params{'filename'},
        'runonce'            => 1,   # read entire file once only
        'synchronous'        => 1,
    );

    return $response;
}


#
# Fragment a file such that when it is read back later it will
# cause WAR and/or if the reallocate scanner is run on the file
# that the file will have blocks that need to be reallocated.
# This function uses internal filersio to fragment the file
# and uses the reallocate scanner to measure the level of
# fragmentation.
# TODO: consider using wafl touch to force a more controlled
# deterministic fragmentation.
# TODO: consider using scramble dumpblocks or showvbns to
# determine if the file is really fragmented versus realloc measure.
#
sub fragment_file_for_war {

    my $_name = (caller(0))[3];
    $Log->debug("in $_name");
    my $_sname = "fragment_file";  # shorter name for log messages
    $Log->comment("$_sname: start");

    my %fragment_params = Params::Validate::validate(
        @_,
        {
# required parameters, these MUST be specified by the caller
            'apiset' => {
                type     => OBJECT,
                optional => 0,
            },
            'filesize' => {
                type     => SCALAR,
                optional => 0,
            },
            'filename' => {
                type     => SCALAR,
                optional => 0,
            },
# TODO: This is needed so reallocate scanner can create a logfile
#       consider if there is a way to remove this requirement and
#       have this subr automatically take care of it.  Root vol??
            'logvolname' => {
                type     => SCALAR,
                optional => 0,
            },
# optional parameters, these can optionally be specified by the caller
            'ratio' => {
                type     => SCALAR,
                optional => 1,
# This default realloate measure level seems to work well as it
# provides a mix of both WAR1 and WAR2 when read back.
                default  => 10,  # level of fragmentation required
            },
        }
    );

    my $apiset = $fragment_params{'apiset'};

# TODO: fixme hard coded timeout
    $apiset->set_timeout(timeout => 600);


# TODO: consider if this subr can automtically find a temporary
#       location for the logfile and clean it up.  root vol??
    my $vol_name = $fragment_params{'logvolname'};
    my $logfile = "/vol/$vol_name/tmp_reallocate_measure.log";

    my $result_ratio_threshold = $fragment_params{'ratio'};

#
# The overall approach is as follows:
# Loop:
#   1. Perform a quick burst of random writes on the file specified
#      then flush these blocks out to disk.  
#   2. Use the reallocate measure scanner to determine the level
#      of file fragmentation.
# Keep looping When the level passes the threshold the file is ready
# or the maximum number of attempts is exhausted.

# TODO: consider using wafl touch to force a more controlled
# deterministic fragmentation.

# TODO: consider using "scramble dumpblocks" or "wafl showvbns" to
# determine if the file is really fragmented versus realloc measure.

# The goal here is to randomly write over a small part of
# the file in a short burst, then flush the blocks out to
# disk and repeat until the file is adequately fragmented
# for purposes of WAR and reallocate scanner.
#
# Lengthly experimentation on a VSIM with 16 block chain lengths
# has showed that this target percentage works well on a VSIM.
# Higher than that and too many adjacent blocks will be grouped
# together making it not as fragmented.  A lower percentage
# will require many more passes of the file.
#
    my $target_percentage = 25;

# First start by taking the size of the file in blocks and
# computing the iteration block count based upon the
# target percentage.   This is the number of blocks wihin
# the file that we want to overwrite in one loop iteration.

    my $filesize_blocks = in_blocks($fragment_params{'filesize'});
    my $iteration_blocks = int($filesize_blocks * $target_percentage/100);

# There are 3 parameters neeed to be computed for that
# are used on each iteration of the fragmentation attempts.
#   1. Total run time for this iteration
#   2. How long to pause between each mini-burst
#   3. How many blocks to write in the mini-burst

# The strategy here is to use the smallest pause interval, as
# this helps even or spread things out as much as possible for WAFL
# as opposed to spikes.  Then just fix the number of IOs,
# for example, 100 bursts (10msec pause) and 4 IOs per burts,
# that is 400 IOs per second.  Then compute how long in seconds
# the run time should be to achieve the total number of
# random writes for that interval.  One thing to keep in mind
# is the 10 second WAFL CP interval.  If the runtime is greater
# than 10 seconds, they a CP will occur during the filersio.
# That is fine, just be aware of that.   The other option that
# may need to be tweaked is the maxios.  This option is use to make
# sure that enough IO buffers are available and that filersio does
# not run out of IOs in flight.
#
# For example:
#   pause = 10 msecs (100 mini-bursts)  FIXED
#   count = 4 IOs  FIXED
#   run-time = 2 seconds (computed)
# This will overwrite
#   total = (2 secs) * (1/10msec) * (4 IOs) = 800 total writes

    my $fio_pause_ms = 10;   # FIXED 10 msecs (100 mini-bursts)
    my $num_bursts_per_sec = int(1/($fio_pause_ms/1000));   # FIXED 100
    my $fio_count = 4;       # FIXED 4 IOs per mini-burts
    my $fio_runtime  = int($iteration_blocks/($fio_count*$num_bursts_per_sec));
    my $fio_maxio = 512;     # FIXED 512 hardcoded pre-allocate this many IOS

#
# Example computation:
#   input file size is 4096 blocks, target percentage is 25%
# NOTE: big rounding errors.
# NOTE: 4096 * 25% target = 1024 (iteration_blocks)
# NOTE: 1024 / 400 = 2.56, rounded down to 2.00 (runtime)
# NOTE: 2 seconds * 400 IOs = 800 total IOS.
# NOTE: target 1024 - actual 800 = missing 224 IOS.
# NOTE: if rounded up to 3 seconds  1024 - 1200, extra 176 IOSs
#

# TODO: consider optional parameter to override this max loop count
    my $cntr = 30;   # hard coded max fragment iterations attempts

    $Log->comment("$_sname: filesize_blocks:   $filesize_blocks blocks.");
    $Log->comment("$_sname: target_percentage: $target_percentage\%");
    $Log->comment("$_sname: iteration_blocks:  $iteration_blocks blocks.");
    $Log->comment("$_sname: fio_runtime:       $fio_runtime secs.");
    $Log->comment("$_sname: fio_pause_ms:      $fio_pause_ms msec.");
    $Log->comment("$_sname: num_bursts:        $num_bursts_per_sec per second.");
    $Log->comment("$_sname: fio_count:         $fio_count");
    $Log->comment("$_sname: fio_maxio:         $fio_maxio");
    $Log->comment("$_sname: logfile:           $logfile");
    $Log->comment("$_sname: Max attempts:      $cntr iterations.");

    my %reallocate_measure_hash;
    my $iterations = 0;

    while ( $cntr ) {
        $iterations++;


#
# A try/catch block is used so that any unexpected output or
# exception from the NACL implementation does not cause the
# calling test case to fail.
#
    try {

# invoke filersio to do the random overwrites

# Note: NACL returns a response OBJECT with the statistics
        my $response = $apiset->filersio_asyncio_pause(
            'read-percentage' => 0,   # no reads, 100% (random) writes
            'rand-percentage' => 100, # 100% random writes
            'blk-size'        => '4k',
            'start-byte'      => 0,
            'end-byte'        => $fragment_params{'filesize'},
            'run-time'        => $fio_runtime,   # total runtime in seconds
            'pause'           => $fio_pause_ms,  # delay between bursts
            'count'           => $fio_count,     # number of IOs per burst
            'maxio'           => $fio_maxio,     # max_ios
            'filename'        => $fragment_params{'filename'},
            'synchronous'     => 1,
        );

    } catch NACL::APISet::Exceptions::UnexpectedOutputException with {
        my $exception_object = shift;
#
# The reason we catch this exception, is we do not want any
# unexpected output from filersio to make this testcase fail.
# Instead leave it to the caller or outer loop to decide if
# the file fails to fragment enough.   This way if filersio has
# some indavertent message or runs out of walf buffers or whatever
# that this function continues and just keeps trying to fragment.
# The outer loop and timeouts will prevent it from running forever.
#
        $Log->comment("$_sname: caught filersio exception");
        generic_print_exception($exception_object);
        $Log->warn("WARN: unexpected filersio exception during file fragmentation");
    }; # NOTE: must have this semicolon


#
# Flush the newly written blocks out to disk that filersio overwrote.
#
        $apiset->wafl_flush_cache();

#
# Run reallocate measure to get the 'Result Ratio' which is
# an indicate of the fragmentation ratio of the file.
#
        my $response = $apiset->reallocate_measure(
            'path' => $fragment_params{'filename'},
            'once' => 1,
            'logfile'  => $logfile,
        );

#
# Flush/eject the blocks out of core as a result of the reallocate measure.
#
        $apiset->wafl_flush_cache();

#
# Read the logfile back via console output
# and store entire file in perl variable
        my $response = $apiset->rdfile(
            'file' => $logfile,
        );

#
# Parse reallocate measure logfile and get important stuff
#
        my $data = $response->get_raw_output();
        my @split = split('\n', $data);
        foreach my $line ( @split ) {
            if ( $line =~/(.*): *(.*)/) {
                $reallocate_measure_hash{$1} = $2;
            }
        }

#
# Get 'Result Ratio' and compare with $result_ratio_threshold
#
        my $result_ratio = $reallocate_measure_hash{'Result Ratio'};
        $Log->comment("$_sname: checking 'result_ratio': $result_ratio > $result_ratio_threshold");
        if ( $reallocate_measure_hash{'Result Ratio'} > $result_ratio_threshold ) {
            $cntr = 0;   # break out of loop (success)
            last;
        }

#
# Decr counter loop and give up attempting to fragment if max exceeded
#
        $cntr--;
        if ( $cntr < 1 ) {
            $Log->warn("$_sname: WARN: result_ratio: $result_ratio never surpassed threshold: $result_ratio_threshold");
# Note: used log warning so that this does not force a TCD failure.
# The caller can decide how to handle this failure instead.
# Note: when fragmentation fails the last measurement logfile is left
# on disk for purposes of debug or triage.
            return 1;   # break out of loop (failure)
        }

    }  # end-while (keep trying more iterations)

    $Log->comment("$_sname: Complete.  Took $iterations iterations.");

# TODO: Anything else that should be verified to insure the file
#       is really fragmented?

# Remove the temporary measurement log file if passed fragmentation test
    my $response = $apiset->rm(paths => "$logfile");

    return 0;  # pass
}



#
# aggr_set_no_delete_log(apiset,aggr_task)
# This function turns on or off delete log on aggr
# NOTE: double negative.  disabled means no_delete_log ON.
# NOTE: must pass in apiset since no aggr_task support for no_delete_log
#
sub aggr_set_no_delete_log {

    my ($apiset, $aggr_task, $on_off) = @_;

    my $_name = (caller(0))[3];
    $Log->debug("in $_name($on_off)");

# There does not appear to be c-mode cli support to set no_delete_log,
# so we are going through nodescope cli.
    if ( 0 ) {
# this code is not hit (enable when NACL support is available)
        $aggr_task->modify('no_delete_log' => $on_off);
    } else {
# Use the nodescope cli
        $apiset->aggr_options(
            'aggregate' => $aggr_task->aggregate,
            'option-name' => 'no_delete_log',
            'option-value' => $on_off,
        );
    }
# TODO: any return codes?  if fails, does above just throw an exception?
    return 0;
}


#
# aggr_set_csc(aggr_task, on_off)
# This function turns on or off aggr space-optimized
#
sub aggr_set_csc {

    my ($aggr_task, $on_off) = @_;

    my $_name = (caller(0))[3];
    $Log->debug("in $_name($on_off)");
    $aggr_task->modify('free-space-realloc' => $on_off);
# TODO: any return codes?  if fails, does above just throw an exception?
    return 0;
}

#
# vol_set_war(vol_task, on_off_fake)
# This function sets the volume war option (read-reallocate)
#
sub vol_set_war {

    my ($vol_task, $on_off_fake) = @_;

    my $_name = (caller(0))[3];
    $Log->debug("in $_name($on_off_fake)");
    $vol_task->modify('read-realloc' => $on_off_fake);
# TODO: any return codes?  if fails, does above just throw an exception?
    return 0;
}


1;


