# -*- perl -*-

#----------------------------------------------------------------------
# WAFL Performance Unit Test library code for External Cache
#----------------------------------------------------------------------
# Description:
#   Library to encapsulate common EC test routines & data to make
#   writing EC testcases as painless as possible
#----------------------------------------------------------------------
# Initial author(s):
#   Brian Naylor
#----------------------------------------------------------------------

package Perf::Dev::ExternalCache;

use strict;
use warnings;

use Data::Dumper;
$Data::Dumper::Indent=1;

use Net::Ping;
use NACL::APISet::Exceptions::ResponseException qw(:try);

# use Perf::Stats;

use Perf::Dev::FilersioProfiles;

use TCD;
use NACL::Realm;
use NACL::C::Node;
use NACL::STask::Aggregate;
use NACL::STask::Volume;

use Perf::Dev::DevUtils;

# Default filer settings for EC tests
# Names are chosen currently to match disk perf testing for synergy..
# Override using NATE parameters (see setup())
my $ec_defaults = {
	aggr_setup	=>	0,
	aggr_disk_count =>	"8",
	vol_size 	=>	"2G",
	aggr_name	=>	"aggr_perf_ut",
	vol_name	=>	"vol_perf_ut",
};

#
# Map some constants to interesting TCD return values
#
use constant PASS  => $TCD::PASS;
use constant WARN  => $TCD::WARN;
use constant FATAL => $TCD::FATAL;
use constant FAIL  => $TCD::FAIL;



#############################################################################
#
# Core class methods
#
#############################################################################

#
# Instantiate a new ExternalCache object, initialize
#
sub new {
	my ($inc) = @_;

	my $class = ref($inc) || $inc;

	my $self = {
		# cached test infrastructure
		tcd			=> undef,
		nate_params		=> undef,
		cached_aggr_task	=> undef,
		node			=> undef,
		apiset			=> undef,
		aggr_setup		=> undef,
		aggr_name		=> undef,
		aggr_disk_count		=> undef,
		vol_name		=> undef,
		vol_size		=> undef,
		files			=> [],

		# cached ec parameters
		stats_start		=> undef,
		type			=> undef,
		blocks			=> undef,
		size			=> undef,

		# settings
		verbose			=> 0,
		notify			=> undef,
		keep_files		=> 0,
		reuse_all		=> 0,
		realm_cleanup		=> undef,
		nacltask_to_cleanup	=> 0,
	};
	bless $self, $class;

	return $self;
}

# Store an instance of TCD for use in logging/etc
# this is required before using anything besides 'init'.
sub set_tcd($)
{
	my ($self, $tcd) = @_;
	$self->{tcd} = $tcd;
}

sub log($)
{
	my ($self, $msg) = @_;
	$self->{tcd}->nlog("ExternalCache: $msg");
}

sub notify($)
{
	my ($self, $test, $expected_cnt, $results) = @_;
	my $cmd = "/bin/mail -s 'Test Results' '". $self->{notify} ."'";

	# hope they don't change the constants!
	my $rc_names = {
		0 => 'FAIL',
		1 => 'PASS',
	};

	if ( defined $self->{notify} ) {
		my $body = "\nYour test is done.\n\n$test\n\nResults:\n";
		foreach my $r (@{$results}) {
			foreach my $t (keys(%{$r})) {
				$body .= "\t- ". $r->{$t} .
					 " : ". $rc_names->{$t} ."\n";
			}
		}
		$body .= "\nLogs: $::LOGDIR\n";

		my $cnt = scalar(@{$results});
		if ( $expected_cnt != $cnt ) {
			$body .= "\nWARNING: Results count did not match ".
				 "expected test count!  Check your logs! ".
				 "($expected_cnt vs $cnt)\n\n";
		}

		if ( ! open(MAIL, "|$cmd") ) {
			$self->log("Couldn't send email notification: $!");
		} else {
			print MAIL $body;
			close(MAIL);
		}
	}
}

#############################################################################
#
# TCD shadow routines
#
#############################################################################

sub init {
	my ($self) = shift;
	my $rc = PASS;

	$self->{nate_params} = NATE::ParamSet->new(global => 1);
	$self->{verbose} = $self->{nate_params}->get('EC_VERBOSE', default => 0);

	# Create a NACL Realm for assisted cleanup
	$self->{realm_cleanup} = NACL::Realm->realm()->cleanup();
	$self->{nacltask_to_cleanup} = 1;

	$rc = $self->reinit;

	if ( $rc == PASS ) {
		my $version = $self->run_nodeshell('version');
		$self->log("Running against ontap build: $version");
		$self->init_cache_stats;
	}

	return $rc;
}

#
# reconnect after a reboot
#
sub reinit
{
	my ($self) = @_;
	my $rc = PASS;

	# Find nodes
	$self->{node} = NACL::C::Node->find();

	if ( ! defined($self->{node}) ) {
		$self->log("Couldn't get node");
		$rc = FAIL;
	} else {
		$self->log("node: " . $self->{node}->name);
		# Get APISet from node
		$self->{apiset} = $self->{node}->get_7m_or_nodescope_apiset();

		if ( $self->enable_cache != PASS ) {
			$rc = FAIL;
		}
	}

	return $rc;
}

sub setup {
	my ($self) = shift;

	$self->{notify} = $self->{nate_params}->get('NOTIFY', default => undef);
	# if notify is set to '1', then get the username from the environment
	if ( defined($self->{notify}) && $self->{notify} eq '1' ) {
		my (@u) = getpwuid($<);
		$self->{notify} = $u[0] . '@netapp.com';
	}

	$self->{aggr_setup} = $self->{nate_params}->get('EC_AGGR_SETUP', default => $ec_defaults->{aggr_setup});
	$self->{aggr_name} = $self->{nate_params}->get('EC_AGGR', default => $ec_defaults->{aggr_name});
	$self->{vol_name} = $self->{nate_params}->get('EC_VOL_NAME', default => $ec_defaults->{vol_name});

	if ( $self->{aggr_setup} ) {
		$self->{aggr_disk_count} = $self->{nate_params}->get('EC_AGGR_DISK_COUNT',
					   default => $ec_defaults->{aggr_disk_count});
		$self->{vol_size} = $self->{nate_params}->get('EC_VOL_SIZE', default => $ec_defaults->{vol_size});

		if ( $self->{verbose} ) {
			print<<EOF;
Test settings:
----------------------------------------------------------------------
      aggr_name_def: $ec_defaults->{aggr_name}
aggr_disk_count_def: $ec_defaults->{aggr_disk_count}
       vol_size_def: $ec_defaults->{vol_size}
       vol_name_def: $ec_defaults->{vol_name}
----------------------------------------------------------------------
          aggr_name: $self->{aggr_name}
    aggr_disk_count: $self->{aggr_disk_count}
----------------------------------------------------------------------
           vol_name: $self->{vol_name}
           vol_size: $self->{vol_size}
EOF
		}

		$self->{cached_aggr_task} = NACL::STask::Aggregate->create(
			command_interface => $self->{node},
			aggregate => $self->{aggr_name},
			diskcount => $self->{aggr_disk_count},
			if_exists => "reuse",
			nacltask_to_cleanup => $self->{nacltask_to_cleanup},
		);

		my $vol_task = NACL::STask::Volume->create(
			command_interface => $self->{node},
			volume => $self->{vol_name},
			aggregate => $self->{cached_aggr_task},
			size => $self->{vol_size},
			nacltask_if_exists => "reuse",
			nacltask_to_cleanup => $self->{nacltask_to_cleanup},
			nacltask_choose_vserver_with_lif => 0,                # XXX bnaylor
		);
	} # aggr setup

	return PASS;
}

sub cleanup {
	my ($self) = shift;

	if ( ! $self->{keep_files} ) {
#		my $file_list = join(' ', @{$self->{files}});
# XXX		my $file_list = $...filersio object -> files
#		my $response = $self->{apiset}->rm(
#			paths => $file_list,
#		);

		# TODO: add $response error checking

		$self->cleanup_files;

	}

# not using this atm
#	if ( ! $self->{reuse_all} ) {
#		if ( $self->{realm_cleanup} ) {
#			$self->{realm_cleanup}->run();
#		} else {
#			$self->{cached_aggr_task}->purge();
#		}
#	}

	return PASS;
}

sub uninit
{
}


#############################################################################
#
# Cache-specific utilities
#
#############################################################################

sub enable_cache
{
	my ($self) = shift;
	my $rc = PASS;

	# no c-mode commands yet, use node scope
	$self->run_nodeshell("options flexscale.enable on");

	if ( ! $self->is_cache_enabled ) {
		$self->log("Failed to enable cache.");
		$rc = FAIL;
	}

	return $rc;
}

sub is_cache_enabled {
	my($self) = shift;

	my $val = $self->get_stat('ext_cache', 'ext_cache', 'state-string');

	return ( $val eq 'Active' );
}

sub get_cache_type {
	my($self) = shift;

	$self->{type} = $self->get_stat('ext_cache_obj', 'ec0', 'type');

	return $self->{type};
}

sub set_option($$) {
	my ($self, $opt, $val) = @_;

	$opt = "flexscale.$opt";
	my $response = $self->run_nodeshell("options $opt $val");

	return $response;
}

# probably don't need to do this so manually..  XXX
sub get_option($) {
	my ($self, $opt) = @_;

	$opt = "flexscale.$opt";

	my $response = $self->run_nodeshell("options $opt");

	my ( @lines ) = split(/\n/, $response);
	for ( @lines ) {
		if ( /^${opt}/ ) {
			my ($o, $v, $j) = $_ =~ /^(^\S+)\s+(\S+)(.*)$/;
			if ( $self->{verbose} ) {
				$self->log("option parse ( opt=$o\nval=$v\njunk=$j\n )");
			}
			$response = $v;
		}
	}

	return $response;
}


#############################################################################
#
# Generic utilities
#   .. these probably don't really even belong here.
#
#############################################################################

# get a raw stat
sub get_stat($$$) {
	my ($self, $obj, $ins, $cnt) = @_;
	my $str = join(':', $obj, $ins, $cnt);
	my $stat_ret;

	my $response = $self->run_nodeshell("stats show -R $str");
	my ( @lines ) = split(/\n/, $response);
	for ( @lines ) {
		if ( /^${str}:(.*)$/ ) {
			$stat_ret = $1;
		}
	}

	return ($stat_ret);
}

# grab all the raw stats for an instance
sub get_all_stats_for_instance($$) {
	my ($self, $obj, $ins) = @_;
	my $stats_ret = {};
	my $str = $obj .':'. $ins;

	my $response = $self->run_nodeshell("stats show -R $str");
	my ( @lines ) = split (/\n/, $response);
	for ( @lines ) {
		if ( /^${str}:([^:]+):(.*)$/ ) {
			$stats_ret->{$str .':'. $1} = $2;
		}
	}

	return $stats_ret;
}

sub init_cache_stats
{
	my ($self) = @_;
	my $strip = 'ext_cache_obj:ec0:';

	$self->{stats_start} =
		$self->get_all_stats_for_instance('ext_cache_obj', 'ec0');

	# run through them all and strip off the object:instance
	foreach my $k (keys(%{$self->{stats_start}})) {
		my $new = $k;
		$new =~ s/^${strip}//;
		$self->{stats_start}->{$new} =
			delete($self->{stats_start}->{$k});
	}

	if ( $self->{verbose} ) {
		$self->log("Starting stats:\n" .
			  Data::Dumper->Dumper($self->{stats_start}));
	}

	$self->{type}   = $self->{stats_start}->{type};
	$self->{blocks} = $self->{stats_start}->{blocks};
	$self->{size}   = $self->{stats_start}->{size};
}

#
# individual cmd interface routines
#
# API-specific stubs map to an underlying wrapper that traps annoying
# exceptions, converts them into either command output (or undef on error),
# logs things, tries to ping the node first, etc.  All of this is to try
# and keep tests from just blowing up and exiting without logging -why-,
# which I find happens all the time with this exception-based model.
#
sub _run_cmd($$)
{
	my($self, $cmd, $api) = @_;

	# ping first to detect panics proactively..
	if ( ! $self->ping ) {
		$self->log("Filer seems dead.  Panic?  Skipping cmd: $cmd");
		return undef;
	}

	if ( $self->{verbose} ) {
		$self->log("Running cmd on node ". $self->{node}->{node} .": $cmd");
	}

	my $text;
	try {
		$text = $api->execute_command(command => $cmd);
	} catch NATE::BaseException with {
		my $error = shift;
		$self->log("Failure to execute command: $cmd.\nError: ".
			   $error->text);
		$text = undef;
	};

	if ( defined($text) && $self->{verbose} ) {
		$self->log("Output: $text");
	}
	return $text;
}

sub run_bsd($)
{
	my($self, $cmd) = @_;
	my $bsd = $self->{node}->apiset(interface => 'CLI', set => 'Systemshell');
	return $self->_run_cmd($cmd, $bsd);
}

sub run_nodeshell($)
{
	my($self, $cmd) = @_;
	my $dblade = $self->{node}->apiset(interface => 'CLI', set => 'Nodescope');
	return $self->_run_cmd($cmd, $dblade);
}

sub run_ngsh($)
{
	my($self, $cmd) = @_;
	my $ngsh = $self->{node}->apiset(category => "Node", interface => "CLI",
					 set => "CMode");
	return $self->_run_cmd($cmd, $ngsh);
}

sub output_chomp($)
{
	my($self, $str) = @_;

	$str =~ s/[\r\n]//gm;

	return $str;
}

sub get_bootarg($) {
	my($self, $bootarg) = @_;

	my $result = $self->run_nodeshell("bootargs get $bootarg");
	$result = $self->output_chomp($result);
	$result =~ s/${bootarg} == //;

	return $result;
}

sub set_bootarg($$) {
	my($self, $bootarg, $value) = @_;
	return $self->run_nodeshell("bootargs set $bootarg $value");
}

#
# Returns free space in a given volume in GB
#
# XXX Better to NACLize this.
#
sub vol_free_space_gb($)
{
	my ($self, $vol) = @_;
	my $ret = -1;

	my $response = $self->run_nodeshell("df -x $vol");

	for( $self->response_lines($response) ) {
		next if /^Filesystem/;
		next if /^$/;

		s/ +/ /g;
		my ($fs, $tot, $used, $avail, $cap, $path) = split(/ /);

		if ( $ret != -1 ) {
			$self->log("vol_free_space_gb confused by multiple results.");
			return 0;
		}
		$ret = $avail; # kb
	}

	$ret = int($ret / 1024 / 1024);

	return $ret;
}


#
# Returns a list of data volumes sorted by free space available
#
sub best_data_vols
{
	my ($self) = @_;
	my $free = {};

	my (@aggrs) =
	NACL::C::Aggregate->find_online_not_root(command_interface =>
						 $self->{node});

	for my $aggr (@aggrs) {
		my @vols = $self->get_vols($aggr);
		for my $vol (@vols) {
			$free->{$vol->{volume}} =
				$self->vol_free_space_gb($vol->{volume});
		}
	}

	return sort { $free->{$a} <= $free->{$b} } keys %{$free};
}

#
# Return a size in gb that represents a reasonable working set
# based on the buffer cache size vs the cache
#
sub generic_workload_size
{
	my($self) = shift;

	my $wafl_bufs = $self->get_stat('wafl', 'wafl', 'wafl_bufs_total');
	my $wb_gb = int(10 * ($wafl_bufs * 4096 / 1024 / 1024 / 1024));
	my $c_gb = int($self->{size} * 0.5);

	return ($wb_gb < $c_gb) ? $wb_gb : $c_gb;
}


sub response_lines($)
{
	my ($self, $raw) = @_;

	$raw =~ s/\r//g;
	my (@lines) = split(/\n/, $raw);

	return @lines;
}

sub get_vservers
{
	my ($self) = @_;

	return
	NACL::C::Vserver->find(command_interface => $self->{node});
}

sub get_aggrs
{
	my ($self) = @_;

	return
	NACL::C::Aggregate->find(command_interface => $self->{node},
				 canned_filters => [ qw(online) ]);
}

sub get_vols($)
{
	my ($self, $aggr) = @_;
	return $aggr->get_volumes();
}

sub set_cache_policy($)
{
	my($self, $policy) = @_;
	my $policy_cmd =
	  'qos settings external-cache modify -external-cache-setting '.
	  $policy . ' -default true';
}


#
# see if a host is alive.  returns 0/1
#
sub ping($)
{
	my ($self) = @_;
	my $ping = new Net::Ping;

	if ( ! $ping->ping($self->{node}->{node}, 5) ) {
		$self->log($self->{node}->{node} ." is not responding to pings! ");
		return 0;
	}

	return 1;
}


#############################################################################
#
# Test helper bundle routines - run i/o, etc.
#
#############################################################################


# kick off some generic cache i/o
#
# If you want to be fancy, use FilersioProfiles directly.
#
# i/o here is started async and can be called more than once.  If you want
# to block until i/o is done, call $ec->wait_for_io
#
sub run_test_io($)
{
	my ($self, $vol) = @_;

	if ( ! defined($vol) ) {
		$self->log("run_test_io requires a volume.");
		return FAIL;
	}

	my $vol_space = $self->vol_free_space_gb($vol);
	if ( $vol_space < 1 ) {
		$self->log("Volume $vol doesn't have much free space ".
			   "($vol_space gb), skipping.");
		return FAIL;
	}

	my $gb = $self->generic_workload_size;
	$self->log("Using workload size $gb gb");

	my $io = new Perf::Dev::FilersioProfiles($self->{apiset});

	my $result =
	$io->start( 'mode'		=> 'asyncio_active',
		    'profile'		=> '4k_mixed_80rand_50read',
		    'volume'		=> $vol,
		    'end-byte'		=> $gb .'g',
		    'run-time'		=> 1800,
		    'synchronous'	=> 0,
		  );
	# should really process result and return pass/fail here XXX
	# - actually you need to try/catch, if it fails it bombs the whole test!

	$self->log(scalar(localtime));

	# save files for later deletion
	for ( $io->files ) {
		push(@{$self->{files}}, $_);
	}

	return PASS;
}

sub cleanup_files
{
	my($self) = shift;

	for( @{$self->{files}} ) {
		$self->log("Deleting $_..");
		$self->run_nodeshell("rm $_");
	}
	$self->{files} = [];
}

#
# When running async filersios, this will wait until they
# all complete and then return.
#
sub wait_for_io
{
	my ($self) = @_;
	my $done = 0;
	my $donestr = 'filersio: no instances currently running';

	$self->log("Waiting for filersios to exit..");

	while ( ! $done ) {

		my $out = $self->run_nodeshell("filersio status");

		if ( ! defined($out) ) {
			$self->log("Wait for i/o failed.  Exiting.");
			return;
		}

		for my $line ( $self->response_lines($out) ) {
			if ( $line =~ /${donestr}/ ) {
				$done = 1;
			}
		}
		sleep(60);
	}
}

#
# Write a filersio -2 pattern file sequentially.
#
sub write_verify_file($)
{
	my ($self, $vol, $write_sz) = @_;

	if ( ! defined($write_sz) ) {
		$write_sz = $self->generic_workload_size;
		$self->log("Using workload size $write_sz gb");
	}

	my $vol_space = $self->vol_free_space_gb($vol);

	if ( $vol_space < $write_sz ) {
		$self->log("Volume $vol only has $vol_space gb, ".
			   "want at least $write_sz.  Aborting.");
		return FAIL;
	}

	$self->set_cache_policy('all');

	my $io = new Perf::Dev::FilersioProfiles($self->{apiset});

	my $result =
	$io->start( 'mode'		=> 'asyncio_active',
		    'profile'		=> '64k_seqwrite_pattern',
		    'volume'		=> $vol,
		    'end-byte'		=> $write_sz .'g',
		    'run-time'		=> -1,
		    'synchronous'	=> 0,
		  );
	# should really process result and eval pass/fail here XXX

	$self->wait_for_io;

	$self->log(scalar(localtime));

	# get the filename (there should be only one)
	push(@{$self->{files}}, $io->files);

	$self->set_cache_policy('random-read');

	return PASS;
}

sub read_verify_file($$)
{
	my ($self, $vol, $file, $read_sz) = @_;

	if ( ! defined($read_sz) ) {
		$read_sz = $self->generic_workload_size;
		$self->log("Using workload size $read_sz gb");
	}

	# if no file, try to pop one from the cache
	if ( ! defined($file) ) {
		$file = pop(@{$self->{files}});
		if ( ! defined($file) ) {
			$self->log("No file for read-verify job.");
			return FAIL;
		}
	}

	my $io = new Perf::Dev::FilersioProfiles($self->{apiset});

	my $result =
	$io->start( 'mode'		=> 'asyncio_active',
		    'profile'		=> '64k_seqread_pattern',
		    'volume'		=> $vol,
		    'end-byte'		=> $read_sz .'g',
		    'synchronous'	=> 0,
		    'run-time'		=> -1,
		    'existing'		=> $file,
		  );
	# XXX same caveat

	$self->wait_for_io;

	$self->log(scalar(localtime));

	$self->log("Deleting $file");
	$self->run_nodeshell("rm $file");

	return PASS;
}

#
# Run enough i/o to wrap the flash.
#
sub wrap_flash
{
	my ($self) = @_;

	# just use whatever the best vol is..
	my ($vol) = ($self->best_data_vols)[0];

	my $cache_blks = $self->{blocks};
	my $write_sz = ($cache_blks * 4096 / 1024 / 1024 / 1024);
	$write_sz = $write_sz + ($write_sz * 0.5);  # add a little overage
	$self->log("wrap_flash: Cache size is $cache_blks blocks, need to ".
		   "write $write_sz gb.");

	# store insert count for later comparison
	my $inserts_start = $self->get_stat('ext_cache_obj', 'ec0', 'inserts');
	$self->log("wrap_flash: Starting wrap at $inserts_start inserts.");

	my $iterations = 1;

	my $vol_space = $self->vol_free_space_gb($vol);

	if ( $vol_space < $write_sz ) {
		$self->log("wrap_flash: Volume $vol only has $vol_space gb free.");
		# bump it down a little to keep from maxing it out
		$vol_space -= 10;
		$iterations = int($write_sz / $vol_space) + 1;
		$write_sz = int($write_sz / $iterations);
		$self->log("wrap_flash: Will do $iterations iterations of $write_sz gb.");
	}

	# turn on all-data
	$self->set_cache_policy('all');

	my $io = new Perf::Dev::FilersioProfiles($self->{apiset});

	for (my $i = 0; $i < $iterations; $i++ ) {
		$self->log(scalar(localtime));
		my $result =
		$io->start( 'mode'		=> 'asyncio_active',
			    'profile'		=> '64k_seqwrite_pattern',
			    'volume'		=> $vol,
			    'end-byte'		=> $write_sz .'g',
			    'synchronous'	=> 0,
			    'run-time'		=> -1,
			  );
# should really process result and eval pass/fail here XXX

		# synchro mode seems to not notice panics/hangs..
		$self->wait_for_io;

		$self->log(scalar(localtime));

		# get the filename (there should be only one)
		my $file;
		foreach my $f ( $io->files ) {
			$file = $f;
		}

		# delete the file (should only be one)
		$self->log("wrap_flash: Deleting iteration $i file: $file");
		$self->run_nodeshell("rm $file");
		$io->reset_files;
	}

	# set the policy back to the default (XXX could get/restore this as well)
	$self->set_cache_policy('random-read');

	my $inserts_end = $self->get_stat('ext_cache_obj', 'ec0', 'inserts');

	$self->log("wrap_flash: iterations complete. ".
		   ($inserts_end - $inserts_start) ." blocks written to flash.");

	if ( ($inserts_end - $inserts_start) < $cache_blks ) {
		$self->log("Not enough inserts occurred to wrap the flash.");
		return FAIL;
	}

	return PASS;
}


#
# Reboot the box and wait for it to come back up.
#
sub reboot
{
	my ($self) = @_;
	my $rc = PASS;
	my $wait = 400;
	my $timeout = 30;
	my $node = $self->{node}->{node};

	$self->log("Rebooting.. ignore the following error:");
	$self->run_ngsh("reboot -node $node");

	$self->log("Waiting $wait seconds..");
	sleep($wait);
	while ( ! $self->ping && $timeout > 0 ) {
		sleep(10);
		$timeout--;
	}

	if ( $timeout == 0 ) {
		$self->log("$node did not come back up in the allotted time.");
		$rc = FAIL;
	} else {
		$self->log("$node is alive.  Waiting one more minute.");
		sleep(60);

		# attempt to reestablish connections
		$rc = $self->reinit;

		# make sure it really rebooted..
		if ( $rc == PASS ) {
			my $uptime = $self->run_nodeshell('uptime');
			if ( $uptime !~ / \d+ mins, / ) {
				$self->log("Box didn't reboot? uptime: $uptime");
				$rc = FAIL;
			}
		}
	}

	return $rc;
}

#
# Check rewarm status after a reboot
#
sub check_rewarm
{
	my($self, $usage) = @_;
	my $rc = PASS;

	my $rw_state = $self->get_stat('ext_cache', 'ext_cache', 'warming_state');
	my $rw_pvols = $self->get_stat('ext_cache', 'ext_cache', 'pvols');
	my $rw_blocks = $self->get_stat('ext_cache_obj', 'ec0', 'rewarm_blocks_reconstructed');
	my $rw_aggrs = $self->get_stat('ext_cache_obj', 'ec0', 'rewarm_aggrs_replayed');
	my $rw_fails = $self->get_stat('ext_cache_obj', 'ec0', 'rewarm_aggrs_failed');
	my $rw_usage = $self->get_stat('ext_cache_obj', 'ec0', 'usage');

	# usage check is optional
	my $usage_ok = 1;
	if ( defined($usage) ) {
		$usage_ok = ( $rw_usage > ($usage - 2) );
	}

	if ( $rw_state ne 'Idle' ||
	     $rw_pvols == 0      ||
	     $rw_blocks == 0     ||
	     $rw_aggrs == 0      ||
	     ! $usage_ok ) {
		$self->log("Rewarm failed for some reason.");
		$self->log("State:    $rw_state\n".
			   "Pvols:    $rw_pvols\n".
			   "Blocks:   $rw_blocks\n".
			   "Replayed: $rw_aggrs\n".
			   "Failed:   $rw_fails\n",
			   "Usage:    $rw_usage vs $usage");
		$self->log("EC_D log follows..");
		$self->run_nodeshell("wafl ec dumplog");
		$rc = FAIL;
	} else {
		$self->log("Rewarm was successful.");
	}

	return $rc;
}



1;

