# -*- perl -*-

#----------------------------------------------------------------------
# WAFL Performance Unit Test library code for specifying run profiles
# for filersio invocations
#----------------------------------------------------------------------
# Description:
#   Create a set of profiles for filersio invocations that can be
#   reused easily
#----------------------------------------------------------------------
# Initial author(s):
#   Brian Naylor
#----------------------------------------------------------------------

package Perf::Dev::FilersioProfiles;

use strict;
use warnings;

use Data::Dumper;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use TCD;

use Perf::Dev::DevUtils;

my $default_file_size = '100m';

use constant PASS => $TCD::PASS;
use constant FAIL => $TCD::FAIL;

my $generic_defaults = {
	'synchronous'		=> 0,
	'create'		=> 1,
	'runonce'		=> 0,
	'start-byte'		=> 0,
	'end-byte'		=> $default_file_size,
	'blk-size'		=> '4k',
	'num-active'		=> 10,
	'read-percentage'	=> 50,
	'rand-percentage'	=> 50,
	'run-time'		=> 600,
};

# only override defaults here..
my $io_types = {
	asyncio_active => {
		'4k_mixed_80rand_50read' => {
			'read-percentage'	=> 50,
			'rand-percentage'	=> 80,
		},
		'64k_seqwrite_pattern' => {
			'read-percentage'	=> 0,
			'rand-percentage'	=> 0,
			'blk-size'		=> '64k',
			'fill-pattern'		=> -2,
			'num-active'		=> 20,
#			'runonce'		=> 1,
		},
		'64k_seqread_pattern' => {
			'read-percentage'	=> 100,
			'rand-percentage'	=> 0,
			'blk-size'		=> '64k',
			'verify-pattern'	=> -2,
			'num-active'		=> 20,
#			'runonce'		=> 1,
		},
	},
	other_types => {
	},
};

sub new($) {
	my ($inc, $apiset) = @_;

	my $class = ref($inc) || $inc;

	if ( ! defined($apiset) ) {
		print "Perf::Dev::FilersioProfiles::new requires an instance of APISet.\n";
		return undef;
	}

	my $self = {
		apiset => $apiset,
		files  => undef,
	};

	bless $self, $class;

	return $self;
}

sub _generate_filename($$$) {
	my ($self, $vol, $size, $base) = @_;

	$base = 'fsio_' . $base;
	my $fname = Perf::Dev::DevUtils::mkfile_name($vol, $size, $base);

	return $fname;
}

sub log($)
{
	my ($self, $str) = @_;
	print "FilersioProfiles: $str\n";
}

#
# Start a given profile
# Override default settings by passing key=>value pairs
#
sub start {
	my ($self, %args) = @_;
	my $mode = $args{mode};
	my $profile = $args{profile};
	my $vol = $args{volume};
	my $existing = $args{existing};
	my $rc = PASS;

	# top-level validation
	if ( ! defined($io_types->{$mode}) ||
	     ! defined($io_types->{$mode}->{$profile}) ) {
		$self->log("Unsupported i/o type specified: $mode/$profile");
		return FAIL;
	}

	# grab defaults for i/o type and then override any passed-in key/value pairs
	my %settings;
	foreach my $key (keys(%{$generic_defaults})) {
		$settings{$key} = $generic_defaults->{$key};
	}

	foreach my $key (keys(%{$io_types->{$mode}->{$profile}})) {
		$settings{$key} = $io_types->{$mode}->{$profile}->{$key};
	}

	# now override anything passed in explicitly that isn't a special
	# argument for this function
	foreach my $key (keys(%args)) {
		if ( $key =~ /^(mode|profile|volume|existing)$/ ) {
			delete($args{$key});
		} else {
			$settings{$key} = $args{$key};
		}
	}

	#
	# Sanity checks!  aka "don't shoot yourself in the foot."  If 'runonce' has
	# been turned on, kill the 'run-time' parameter because it will most likely
	# cause things to happen that you didn't mean..
	#
	if ( $settings{'runonce'} == 1 ) {
		$settings{'run-time'} = 0;
	}
	# add more as needed...

	#
	# ok, the command is set up now..  let's go.
	#
	my $fname;
	if ( defined $existing ) {
		$fname = $existing;
	} else {
		$fname = $self->_generate_filename($vol,
				$settings{'end-byte'}, 'fp_');
	}

	# save file to list of active files
	push(@{$self->{files}}, $fname);

	# disable apiset timeout if synchronous
	my $restore_timeout = undef;
	if ( $settings{synchronous} != 0 ) {
		$restore_timeout = $self->{apiset}->get_timeout;
		$self->{apiset}->set_timeout(timeout => -1);
	}

	$settings{apiset}   = $self->{apiset};
	$settings{filename} = $fname;

	if ( $self->{verbose} ) {
		$self->log("FILERSIO SETTINGS: ". Data::Dumper->Dumper(%settings));
	}

	# this thing blows up on failures as well..
	try {
		my $response = Perf::Dev::DevUtils::filersio_asyncio_active(%settings);
		# check response? XXX
		$self->log("filersio response". Data::Dumper->Dumper($response));
	} catch NATE::BaseException with {
		my $fail = shift;
		$self->log("Filersio failed with: ". $fail->text);
		$rc = FAIL;
	};

	if ( $restore_timeout ) {
		$self->{apiset}->set_timeout(timeout => $restore_timeout);
	}

	return $rc;
}

sub files
{
	my $self = shift;
	return @{$self->{files}};
}

sub reset_files
{
	my $self = shift;
	$self->{files} = undef;
}

1;

