# -*- perl -*-

#----------------------------------------------------------------------
# WAFL Performance Unit Test code templater
#----------------------------------------------------------------------
# Description:
#   Export the TCD stubs required for every test to enable small,
#   concise thpls and avoid cut&pastery
#----------------------------------------------------------------------
# Initial author(s):
#   Brian Naylor
#
# Changes:
# 2013-01-17 bnaylor: Made more generic - support passing in an anonymous
#            object rather than instantiating ExternalCache and pushing
#            it out.  Renamed from ECTest to TestHarness.
#----------------------------------------------------------------------

package Perf::Dev::TestHarness;

#
# Export symbols into user namespace
#
BEGIN {
	require Exporter;
	@ISA = qw(Exporter);
	@EXPORT = qw( main
		      print_header
		      save_result
		      init
		      setup
		      cleanup
		      uninit
		      $tcd
		    );
};

use strict;
use warnings;

use TCD;

#
# Map some constants to interesting TCD return values
#
use constant PASS  => $TCD::PASS;
use constant WARN  => $TCD::WARN;
use constant FATAL => $TCD::FATAL;
use constant FAIL  => $TCD::FAIL;

##
## Globals used in user tests
##
our $tcd;

##
## Internal globals
##
my $test_name;
my $test_count;
my $sidecar = undef;
my @results = ();

#############################################################################
#############################################################################
#
# main
#
# Takes two primary arguments: reference to your @TESTS array, and the name of
# the test suite.
#
# An optional third argument is an object that responds to the same "magical"
# calls as this TCD shim - init, cleanup, etc.  If this argument is defined
# those calls will be executed as well.
#
#############################################################################
#############################################################################

sub main
{
	my ($tests, $name, $obj) = @_;

	# save test name
	$test_name = $name;

	$test_count = scalar(@{$tests}) / 2;  # name + desc tuples
	$test_count += 4;                     # add magical TCD stubs

	$tcd = TCD->new(
		-testcases => $tests,
	);

	if ( defined($obj) ) {
		$sidecar = $obj;
	}

	$tcd->run_test();

	return PASS;
}

#
# hide the glorious semantics of 'caller' from users..
#
sub print_header
{
	$tcd->description($test_name .' '. (caller(1))[3]);
}

#
# Log test results internally
#
sub save_result($)
{
	my ($result) = @_;
	my $test = (caller(1))[3];

	push(@results, {$result => $test});
}

#
# Whatever you do, don't log the word 'panic'...
#
sub did_bad_happen {
	my $rc = PASS;
	my $panic_found = 0;

	$tcd->description("$test_name Did Bad Happen");

	if ( ! opendir(LOGS, $::LOGDIR) ) {
		$tcd->nlog("Couldn't open log directory for did_bad_happen.");
		return WARN;
	}

	my(@files) = readdir(LOGS);
	closedir(LOGS);

	for ( @files ) {
		next unless /\.log$/;

		my $log = $_;
		if ( ! open(IN, $::LOGDIR .'/'. $log) ) {
			$tcd->nlog("Couldn't open $log for did_bad_happen.");
			$rc = WARN;
		}

		while(<IN>) {
			# simplistic until it needs to not be....
			if ( /panic/i ) {
				$tcd->nlog("*** BADNESS at line $. in $log.");
				$panic_found = 1;
			}
		}
		close(IN);
	}

	# don't let file open failures override detected panics..
	if ( $panic_found ) {
		$rc = FAIL;
	}

	return $rc;
}

#############################################################################
#############################################################################
#
# Test harness stubs
#
#############################################################################
#############################################################################

sub init {
	my $rc = PASS;

	$tcd->description("$test_name Init");
	print "LOGDIR:", $::LOGDIR, ":\n";

	if ( defined($sidecar) ) {

		# In case the sidecar object wants a TCD too (logging, etc)
		if ( $sidecar->can('set_tcd') ) {
			$sidecar->set_tcd($tcd);
		}

		if ( $sidecar->init != PASS ) {
			$rc = FAIL;
		}
	}

	save_result($rc);
	return $rc;
}

sub setup {
	my $rc = PASS;

	$tcd->description("$test_name Setup");

	if ( defined($sidecar) ) {
		$rc = $sidecar->setup;
	}

	save_result($rc);
	return $rc;
}

# Teardown test set up. Automatically called after all tests execute.
sub cleanup {
	my ($rc) = PASS;

	$tcd->description("$test_name Cleanup");

	if ( defined($sidecar) ) {
		$rc = $sidecar->cleanup;
	}

	save_result($rc);
	return $rc;
}

# Disconnect from filer and client. Automatically called after all tests execute.
sub uninit {
	my $rc = PASS;

	$tcd->description("$test_name Uninit");

	$rc = did_bad_happen;
	save_result($rc);

	if ( defined($sidecar) ) {
		if ( $sidecar->can('notify') ) {
			$sidecar->notify($test_name, $test_count, \@results);
		}
		$sidecar->uninit;
	}

	return $rc;
}

1;

