# -*- perl -*-
# $Id: //depot/prod/test/main/ontap/boilermaker/lib/Perf/Iomem.pm#2 $
#
## @author   bshelton@netapp.com
## @summary  Convenience functions for Pisces
## @status   public

package Perf::Iomem;

use strict;
use warnings;
use Carp qw(confess);
use Tharn;
use Perf::Utils qw(in_blocks);
use Perf::ModeAgnostic qw(say_wrap);

require Exporter;

our @ISA       = qw(Exporter);
our @EXPORT_OK = qw(iomem_config_disable
                    iomem_config_enable
                    iomem_config_show
                    iomem_ems_messages);


# Disable Pisces card.

sub iomem_config_disable {
    my $fc   = shift;
    my $slot = shift;
    if ($slot !~ /all|\d+/){
        confess "You must specify a valid slot id or 'all'";
    }    
    my %args = @_;
    my $persistent = defined($args{persistent})?$args{persistent}:1;    
    
    my $cmd = "priv set -q diag; iomem config disable";
    if ($slot ne 'all'){
        $cmd .= " -c $slot";
    }
    else{
        $cmd .= " -A";
    }
    if ($persistent){
        $cmd .= " -P";
    }    
    my (undef, undef, $out) = say_wrap($fc, $cmd);

    if ($out !~ m/Card disabled/) {
        confess "Unexpected output from '$cmd': $out";
    }
}


# Enable Pisces card.

sub iomem_config_enable {
    my $fc   = shift;
    my $slot = shift;
    if ($slot !~ /all|\d+/){
        confess "You must specify a valid slot id or 'all'";
    }
    my %args = @_;
    my $persistent = defined($args{persistent})?$args{persistent}:1;

    my $cmd = "priv set -q diag; iomem config enable";
    if ($slot ne 'all'){
        $cmd .= " -c $slot";
    }
    else{
        $cmd .= " -A";
    }
    if ($persistent){
        $cmd .= " -P";
    }
    my (undef, undef, $out) = say_wrap($fc, $cmd);

    if ($out !~ m/Card enabled/) {
        confess "Unexpected output from '$cmd': $out";
    }
}


# Return the config info for Pisces cards in an arrayref.
# Each entry in the array is a hashref of fields in the entry.

sub iomem_config_show {
    my $fc = shift;
    my $id = shift || 'Performance Acceleration Module I';

    # Get filer configuration.

    my @config;
    my $cmd = 'priv set -q diag; iomem config show -A';
    my (undef, undef, $out) = say_wrap($fc, $cmd);

    # Parse Pisces card records.

    foreach my $cfg (split /Slot\s+/, $out) {
        if ($cfg =~ m/^(\d+):\s+(.*?)\s*$/m
            and $2 eq $id) {
            chomp($cfg);

            my $slot       = $1;
            my $name       = $2;
            my $state      = $cfg =~ m/State:\s+(\S+)\s*$/m
                               ? $1
                               : 'Unknown';
            my $memory     = $cfg =~ m/Memory Size:\s+(\d+\s\S+)\s*$/m
                               ? in_blocks($1)
                               : 'Unknown';
            my $serial_num = $cfg =~ m/Serial Number:\s+(\S+)\s*$/m
                               ? $1
                               : 'Unknown';
            my $part_num   = $cfg =~ m/Part Number:\s+(\S+)\s*$/m
                               ? $1
                               : 'Unknown';
            my $board_rev  = $cfg =~ m/Board Revision:\s+(\S+)\s*$/m
                               ? $1
                               : 'Unknown';
            my $fpga_rel   = $cfg =~ m/FPGA Release:\s+(\S+)\s*$/m
                               ? $1
                               : 'Unknown';
            my $fpga_bld   = $cfg =~ m/FPGA Build:\s+(\S+)\s*$/m
                               ? $1
                               : 'Unknown';

            my $dimms;
            while ($cfg =~ m/(\d+):            # DIMM number
                             \s+
                             (\d+\s[A-Z]+)     # Size (typically in GB)
                             \s+
                             (\S+)             # Part number
                             \s+
                             (\S+)             # Serial number
                             \s+
                             (\S+)\s*$         # Status
                            /mgx) {
                push @$dimms, { blks       => in_blocks($2),
                                part_num   => $3,
                                serial_num => $4,
                                status     => $5 }
            }

            # Append hashref to Pisces config info array.

            push @config, { slot     => $slot,
                            id       => $name,
                            state    => $state,
                            mem_blks => $memory,
                            serial   => $serial_num,
                            part     => $part_num,
                            rev      => $board_rev,
                            fpga_rel => $fpga_rel,
                            fpga_bld => $fpga_bld,
                            dimms    => $dimms,
                            text     => "\tslot $cfg" };
        }
        elsif ($cfg){
            print "No card in:\n$cfg\n";
        }
    }

    return \@config;
}


# Return arrayref of current iomem EMS messages.

sub iomem_ems_messages {
    my $fc    = shift;
    my $regex = shift || '<iomem_.*?\/>';
    my @msgs;

    my (undef, undef, $out) =
        say_wrap($fc, 'priv set -q diag; rdfile /etc/log/ems');

    push @msgs, $&
        while ($out =~ m/$regex/sg);

    return \@msgs;
}


## @pod here

=head1 NAME

    Iomem.pm - convenience functions for Iomem

=head1 SYNOPSIS

    use Perf::Iomem qw(iomem_config_show
                       iomem_config_enable);
    use Tharn;

    my $fc = connect($FILER);
    my $cfg = iomem_config_show($fc);

    foreach my $card (@$cfg) {
        iomem_config_enable($fc, $card->{slot});
    }

    ...

=head1 DESCRIPTION

    This module provides functions for interacting with the
    Pisces driver and hardware. In general, these functions
    call iomem commands on the filer. Not all iomem commands
    are currently supported; functions are being added "as
    needed".

=head1 FUNCTIONS

=head2 iomem_config_disable

    Disable the Pisces card in the specified slot.

    Arguments:
        $fc   - filer connection
        $slot - Pisces card slot

=head2 iomem_config_enable

    Enable the Pisces card in the specified slot.

    Arguments:
        $fc   - filer connection
        $slot - Pisces card slot

=head2 iomem_config_show

    Return reference to an array of Pisces card info hashrefs.

    Arguments:
        $fc - filer connection

    Options:
        $id - sysconfig identifier string of Pisces card

=head2 iomem_ems_messages

    Return reference to an array of the iomem messages in
    the current EMS log.

    Arguments:
        $fc    - filer connection

    Options:
        $regex - regular expression to recognize iomem messages

=cut


1;

