# -*- perl -*-
# $Id: //depot/prod/test/main/ontap/boilermaker/lib/Perf/ModeAgnostic.pm#24 $
# Copyright (c) 2008 NetApp, Inc.
# All rights reserved.
# lib/Perf/ModeAgnostic.pm
#
## @summary  Perl library for mode agnostic routines.
## @author   dl-perf-qa@netapp.com
## @status   public
##
## @pod here

=head1 NAME

    Perf::ModeAgnostic

=head1 SYNOPSIS

    use Perf::ModeAgnostic qw(say_wrap
                              cluster_mode);

    my $out  = say_wrap($fc, $cmd);
    my $out  = say_wrap($host, $cmd);
    my $cluster_mode = cluster_mode($fc);
    my $cluster_mode = cluster_mode($host);


=head1 DESCRIPTION

    The functions in this module are used to implement mode agnostic commands
    to enable scripts to run not knowing if the filer is 7-mode or Cluster-mode.

=cut

package Perf::ModeAgnostic;

use strict;
use warnings;
use Carp qw(carp confess);
#use Cwd;
#use File::Basename qw(fileparse);
#use File::Temp qw(tempfile);
#use GenericMount qw(mount umount);
#use IO::File;
use List::MoreUtils qw(none);
use Params::Validate qw(validate HASHREF OBJECT SCALAR UNDEF);
use Tharn;
use filer;
use TracePkg;
use Options;
use Server;
use Data::Dumper;
use Filer::Version qw(are_we_ontapng);

use base qw(Exporter);
our @EXPORT_OK = qw(say_wrap
                    cluster_mode
                    filer_reboot_wrap
                    check_C_or_7_mode);



### Public functions:

=head1 FUNCTIONS

=head2 say_wrap
    Issue command to filer. Use Tharn::say() for 7-mode;
    Server::dblade() for cluster-mode.

    Arguments:

      fc   - filer Connectrec (required, if host not specified)
      host - filer hostname (required, if fc not specified)


    7-mode
    my $out  = say_wrap($fc, $cmd);
        or
    my (undef,undef,$out) = say_wrap($fc, $cmd);

    10-mode
    my $out  = say_wrap($host, $cmd);
        or
    my (undef,undef,$out) = say_wrap($host, $cmd);

=cut

sub say_wrap {
    my $host = shift;    # Filer Connection Object or Host
    my $cmd  = shift;
    my @args = @_;

    my $IS_NG = check_C_or_7_mode($host);
    logcomment("IS_NG value : $IS_NG");

    # If cluster-mode, use Server object.

    if ($IS_NG) {
        my $server = Server->new();
        my @known  = $server->servers();

        if (none { $_ eq $host} @known) {
            confess "Server object cannot connect to '$host'";
        }

        my $node_check = $server->set_execution_server( -Server => $host );
        if (wantarray) {
            return (undef,undef,$server->dblade(node => $host, cmd => $cmd, @args));
        }
        else {
            return $server->dblade(node => $host, cmd => $cmd, @args);
        }
    }

    # Otherwise, open a connection and use say().
    if (wantarray) {
        return say( $host, $cmd, @args );
    } else {
        my ( undef, undef, $out ) = say( $host, $cmd, @args );
        return $out;
    }
}

=head2 cluster_mode

    Return true if host is running cluster-mode,
    false otherwise.

    This is here in order to be available to all
    tests and libraries.

    Arguments:

      fc   - filer Connectrec (required, if host not specified)
      host - filer hostname (required, if fc not specified)

=cut

sub cluster_mode {
    my %params =
        validate(@_,
                 {
                     fc    => { type => HASHREF | UNDEF,          optional => 1 },
                     host  => { type => HASHREF | SCALAR | UNDEF, optional => 1 },
                 });

    my $host = $params{host} || $params{fc};

    # May be passed a Connectrec. In that case,
    # assume 7-mode.

    if (eval { $host->isa('Connectrec') }) {
#        my (undef, undef, $out) = say($host, $cmd, @args);
        $host = $host->hostp()->hostname();
#        return 0;
    }

    # Otherwise, check hostytpe. If it looks like
    # e.g., "filer-ng-cfe", assume cluster-mode.
    #
    # Note: This works for Boilermaker, but will
    # likely need updating for future releases.

    return host($host)->hosttype() =~ m/-ng-/;
}

=head2 filer_reboot_wrap

    Run filer_reboot if 7-mode.
    Run deux_reboot.thpl (copy of script) if 10-mode

    This is here in order to be available to all
    tests and libraries.

    Arguments:

      fc   - filer Connectrec (required, if host not specified)
      host - filer hostname (required, if fc not specified)

=cut

sub filer_reboot_wrap {
    my $fc = shift;    #Filer Connection Object or Hostname
    my @flags = shift;

    if ( cluster_mode( host => $fc ) ) {
        $Log->comment("Rebooting using cluster mode method");

        # copied from deux_reboot.thpl and slightly modified

        ### setup traceing ###
        my $Trace = TracePkg->new();

        my $Options = new Options(RR_FLAG=>1);

        $Trace->entry();

        ### create server object ###
        my $server = new Server(RR_FLAG=>1);

        ### handle multiple nodes with subtest ###
        my @Nodes = $server->servers;
        foreach my $node (@Nodes){
            $Log->comment(msg=>"Starting reboot on $node...");
            # set the execution server
            $Trace->comment(msg=>"Setting execution server to $node...");
            $server->set_execution_server(-Server => $node, RR_FLAG=>1);
            #$server->handle_connections(close => [qw(dblade_connection shell_connection cli_connection)]);

#$server->handle_connections(close => [qw(dblade_connection shell_connection cli_connection)]);

            ### Changes by Anupama
            #$Trace->comment(msg=>'Starting transit_to REBOOT');
            #eval {
            #    $server->transit_to(state => $server->REBOOT(), RR_FLAG=>0);
            #};
            #if ($@) {
            #    my $err = $@;
            #    confess "Cannot transit to reboot: $err";
            #}

            #$Trace->comment(msg=>'Starting transit_to LOGIN');
            #eval {
            #    $server->transit_to(state => $server->LOGIN(), RR_FLAG=>0);
            #};
            #if ($@) {
            #    my $err = $@;
            #    confess "Cannot transit to ngsh: $err";
            #}
            ####$server->establish_connections(dblade_connection => 1);
            #$Trace->exit();
            ### End
            my $host_name =
                eval { $fc->isa('Connectrec') }
                ? $fc->hostp()->hostname()
                : $fc;

            # Workaround to close & re-establish dblade connection, since
            # Node::NG::Transit::transit_to() does not handle this.
            $server->close_connection( type => 'dblade_connection' );

            $Trace->comment( msg => "Rebooting Node $host_name:" );
            $server->transit_to( state            => $server->REBOOT,
                                 inhibit_takeover => 1,
                                 RR_FLAG          => 1);

         # This is needed in case AUTOBOOT is not set. If set, this is a no-op
            $Trace->comment(
                          msg => "Bring Node $host_name up to login prompt" );
            $server->transit_to( state => $server->LOGIN, RR_FLAG=>1 );
            $Trace->comment( msg => "Node $host_name reboot completed" );

            $server->establish_connections( dblade_connection => 1 );
        }
        $Trace->comment( msg => "Reboot completed." );
    } else {
        $Log->comment("Rebooting using 7 mode method");
        filer_reboot($fc);
    }
}
    # Transit to boot menu.


#Function to check if the call is from C-mode Filer or 7-mode Filer
sub check_C_or_7_mode {
    my $host = shift;    #Filer connection Object or Hostname
    my $IS_NG;           # Is this NG (Cluster mode)

    $IS_NG = eval { $host->isa('Connectrec') } ? 0 : 1;
    return $IS_NG;
}

1;
