#!perl
package Perf::PerfANTRun;

=head1 NAME

    Perf::PerfANTRun - Schedule PerfANT measurements

=head1 SYNOPSIS

    use Perf::PerfANTRun;

    my $run = Perf::PerfANTRun->new();
    $run->set_opt('purpose:t="Dynamic Mem Allocation"');
    $run->set_opt('model:m1', 'FAS6080');
    $run->submit();

=cut

use strict;
use warnings;
use Carp qw(confess);
use Clone;
use Cwd qw(abs_path);
use Data::Dumper;
use DateTime;
use File::Basename qw(fileparse dirname);
use File::Copy qw(cp);
use LWP::Simple qw(get);
use NATE::ParamSet;
use Template;
use Template::Stash;
use Tharn qw($Log);
use URI;
use XML::Simple;

# Determine AAMP & PerfANT installations to use
my ($INSTALL_BASE, $PERFANT_INSTALL, $AAMP_INSTALL);
BEGIN {
    my $params = NATE::ParamSet->new(global => 1);
    $INSTALL_BASE = $params->get("INSTALL_BASE", default => "perfqa_stable");
    # we have aliases for the ones we use often
    if ($INSTALL_BASE eq 'perfqa_stable') {
        $PERFANT_INSTALL = '/x/eng/rtpperf/perf_qa/stable/perfant';
        $AAMP_INSTALL = '/x/eng/rtpperf/perf_qa/stable/aamp_install';
    }
    elsif ($INSTALL_BASE eq 'perfqa_test') {
        $PERFANT_INSTALL = '/x/eng/rtpperf/perf_qa/test/perfant';
        $AAMP_INSTALL = '/x/eng/rtpperf/perf_qa/test/aamp_install';
    }
    elsif ($INSTALL_BASE eq 'perfdev') {
        $PERFANT_INSTALL = '/x/eng/perfdev';
        $AAMP_INSTALL = '/x/eng/perfdev/aamp_runtime';
    }
    else {
        $PERFANT_INSTALL = $INSTALL_BASE;
        $AAMP_INSTALL = "$INSTALL_BASE/aamp_runtime";
    }
}

use lib "$PERFANT_INSTALL/perl";
use lib "$PERFANT_INSTALL/perl/PerfCommon/lib";

use PerfCommon::Runid;
use RunSubmissions::AAMP qw(aamp_add_slsl_parameters aamp_add_testbed_parameters);
use RunSubmissions::Rules qw(check_rules);
use RunSubmissions::Locks qw(granular_lock granular_unlock);
use RunSubmissions::QueueFile qw(append_test parse_submit_line);
use RunSubmissions::Scheduler qw(read_resources);
use RunSubmissions::Defaults qw(default_parms load_config store_config);
use RunSubmissions::Config qw($Lab_name set_lab_specific_config $Queue_file $Lock_dir %runid_prefix);
use RunSubmissions::PerfantParameters;

no lib "$PERFANT_INSTALL/perl";
no lib "$PERFANT_INSTALL/perl/PerfCommon/lib";

use Log::Log4perl qw(:easy);

#Log4perl used by PerfCommon and RunSubmissions classes.
Log::Log4perl->easy_init( {'log4perl.appender.Main' => 'Log::Log4perl::Appender::Screen', level => 'DEBUG'});

my $web_service = URI->new('http://grover.rtp.netapp.com/KO/index.php/common/ws/process');


=head1 FUNCTIONS
=head2 new

    usage: Perf::PerfANTRun->new();
           Perf::PerfANTRun->new('123456dab'); #override default runid
=cut

sub new {
    my $class = shift;
    my %opts = @_;
    my %new_options = %{ Clone::clone(\%opts) };
    my $runid = $new_options{RUNID} || $new_options{PARENT} || '111028daa';

    my $uid = _get_uid(); #unique number used to identify test even before a
                #runid is assigned.
    my $lab = _set_lab($runid);
    if ($lab) {
        $Log->debug("Test running in lab $lab");
    } else {
        confess("Failed lab setup");
    }

    my $config = PerfCommon::Runid->new($runid);
    $config->load_exp();

    my $self = {
        runid           =>  $runid,
        uid             =>  $uid,
        config          =>  $config,
        slst            =>  0,  #true if the test has been given a template to generate the .slsl file
        template        =>  Template->new({
            ANYCASE         =>  1, #allow directives to be lowercase
            ABSOLUTE        =>  1, #allow absolute paths
            EVAL_PERL       =>  1, #allow imbedded perl
        }),
        template_vars   =>  {}, #custom template variables
        stash           =>  Template::Stash->new(), #stash for main slsl file
        precommands     =>  [],
        postcommands    =>  [],
        parser          =>  XML::Simple->new(
            NormalizeSpace  =>  2,
            ForceContent    =>  1,
            ForceArray      =>  [
                'group',
                'parameter',
            ],
            KeyAttr         =>  {
                group       =>  'name',
                parameter   =>  'name',
            }
        ),
        options         =>  \%new_options,
        pp              =>  PerfantParameterSet->get_instance(),
    };

    bless $self, $class;

    my $purpose = $config->parameter('purpose:t');
    if ($purpose =~ m/([^#]*) #[\d\-]*$/ms) {
#if $purpose has an old uid attached to it, remove it
        $purpose = $1;
    }
    $purpose = "$purpose #$uid";
    $config->parameter('purpose:t', $purpose);

    if ($self->{config}->{test_variables}->{xml_config}) {
        $self->load_xml($self->{config}->{test_variables}->{xml_config});
    }
    my $params = NATE::ParamSet->new(global => 1);
    $config->user($params->get('USER'));

    return $self;
}

sub clone {
    my $self = shift;

    my $parameters = $self->{config}->parameters();
    my $template_vars = $self->{template_vars};
    my $precommands = $self->{precommands};
    my $postcommands = $self->{postcommands};

    my $new_self = Perf::PerfANTRun->new(%{$self->{options}});
#parameters is organized as:
#{
#   'num_clients' => {
#       't' => '2'
#   }
#}
    foreach my $param (%{$parameters}) {
        foreach my $type (%{$param}) {
            $new_self->set_opt("$param:$type", $parameters->{$param}{$type});
        }
    }
    if (defined $self->{setlist_template}) {#if the setlist was a template, but has been expanded during submission, use the template file instead of the test-specific version
        $new_self->set_opt('setlist_file:t', $self->{setlist_template});
    } else {
        $new_self->set_opt('setlist_file:t', $self->{setlist_file});
    }
    if (defined $self->{base_prestats}) {#if prestats slsl file was already appended to by original test, override the test-specific test with the base test that was used
        $new_self->set_opt('sound_meter_prestats:m1', $self->{base_prestats});
    }
    if (defined $self->{base_poststats}) {#if poststats slsl file was already appended to by original test, override the test-specific test with the base test that was used
        $new_self->set_opt('sound_meter_poststats:m1', $self->{base_poststats});
    }
    foreach my $var (keys %{$template_vars}) {
        $new_self->add_template_var($var, $template_vars->{var});
    }
    foreach my $command (@{$precommands}) {
        $new_self->add_precommand(%{$command});
    }
    foreach my $command (@{$postcommands}) {
        $new_self->add_postcommand(%{$command});
    }
    return $new_self;
}

=head2 set_opt

    usage: $run->set_opt('option:o="value"'); #set option of type 'o' to value
           $run->set_opt('option:o', 'value'); #same as above
    available option types are :
=over
=item * i   interconnect type
=item * t   test variable
=item * m1  controller 1 variable
=item * m2  controller 2 variable
=item * mN  controller N variable (N < 10)
=item * q   queue variable
=back
    a list of available options can be found L<here|http://grover.rtp.netapp.com/KO/index.php/common/nomenu/process?action=getQueueParametersHelp>

=head3 Special Options
    Build can be specified absolutely, or with one of 3 types of shorthand notation
=over
=item * ontap:RironcityN (currently not functional)
=item * ontapng:R10.04xN
=item * DOT:mainN
=back
    ontapng and DOT builds can contain different build types, such as
=over
=item * ontapng:R10.03xN:optimize
=item * DOT:RrollingrockN:debug
=back
=cut

sub set_opt {
    my $self = shift;
    my $opt_string = shift;
    my $opt_value = shift;

    $Log->debug("\$opt_string:\t$opt_string");
    $Log->debug("\$opt_value:\t$opt_value");

    if (!defined $opt_value) {
#capture /opt:t="value"/, /opt:t='value'/, or /opt:t=value/
        chomp $opt_string;
        if ($opt_string =~ m/^(\w+:\w+)\s*=\s*"([^"]*)"$/ms) {
            $opt_string = $1;
            $opt_value = $2;
        } elsif ($opt_string =~ m/^(\w+:\w+)\s*=\s*'([^']*)'$/ms) {
            $opt_string = $1;
            $opt_value = $2;
        } elsif ($opt_string =~ m/^(\w+:\w+)\s*=\s*([^ ]+)$/ms) {
            $opt_string = $1;
            $opt_value = $2;
        } else {
            confess('bad argument to set_opt');
        }
    }

    if ($opt_string eq 'purpose:t') {
#append uid to purpose string
        $opt_value .= ' #' . $self->{uid};
    }

    if ($opt_string eq 'setlist_file:t') {
        if ($opt_value =~ m/\.slst$/) {
            $self->{slst} = 1;
        }
        $self->{setlist_file} = abs_path($opt_value);
    }
    $self->{config}->parameter($opt_string, $opt_value);

    $opt_string =~ s/:/_/g;
    $Log->debug("Setting '$opt_string' to '$opt_value' in stash");
    $self->{stash}->set($opt_string, $opt_value);
    return 1;
}

=head2 set_opts

    usage: my $hash_ref = { 'build:m1' => 'DOT:mainN', 'purpose:t' => 'Example test' };
           $run->set_opts($hash_ref); #set options where key is the name and value is the value

=cut

sub set_opts {
    my $self = shift;
    my $opts = shift;
    $self->set_opt($_, $opts->{$_}) foreach (keys %{$opts});
    return 1;
}

sub set_lab {
#different from _set_lab in that _set_lab actually sets up the configurations to submit a test to a lab queue, whereas set_lab selects which lab will be used for the setup.
    my $self = shift;
    my $lab = shift;
    confess("$lab does not exist") if (!defined $runid_prefix{$lab});
    $self->{lab} = $lab;
    return 1;
}

=head2 submit

    usage: $run->submit();

    submits a test to the queue that the base test (specified with new($base_test)) was run in, default queue being PerfDev

=cut

sub submit {
    my $self = shift;
    if ($self->{submitted}) {#if the test is being submitted for a second time, get a new uid
        confess('Run already submitted! Use clone() to create a new run object');
    }
    $self->{submitted} = 1;

    if (defined $self->{lab}) {
        _set_lab($self->{lab});
    } else {
        _set_lab($self->{runid});
    }

#Command line arguments should always override scripted arguments, so they will only be processed directly before the run is submitted
    my $params = NATE::ParamSet->new(global =>  1);
    my @cli_params = split(qr/;/, $params->get("PERFANT_OPTS"));
    foreach (@cli_params) {
        chomp;
        $self->set_opt($_);
    }
    if ($self->{config}->parameter('testname:t') eq 'motley') { #if AAMP test, set defaults for parameters that may not matter
        $Log->debug('Setting motley defaults');
        my %default_opts = (
            'priority:t'                =>  '3',
            'protocol:t'                =>  'nfs',
            'role:m1'                   =>  'none',
            'sfo_status:m1'             =>  'disabled',
            'target_vols:m1'            =>  '',
            'vvol_list'                 =>  '',
            'ic_type:i'                 =>  'standalone',
        );
        foreach (keys %default_opts) {
            if (!defined $self->{config}->parameter($_)) {
                $self->set_opt($_, $default_opts{$_});
            }
        }
    }
    for (my $i = 1; $i <= 9; $i++) {
#build scan be specified by path to image, or with the short hand [type]:[version](:[build])
#       where [type] is 'DOT', 'ontap', or 'ontapng'
#       [version] is the version of ontap, such as R10.0.4xN, RironcityN, etc.
#       and [build] (optional) can be something such as 'optimize', 'debug', 'optimize.ibm', etc.
#so a build of DOT:mainN would be the same as '/x/eng/rlse/DOT/mainN/final/bedrock/export/x86_64/tarball/image.tgz
        next if (!defined $self->{config}->parameter("build:m$i")); #skip nodes that are not defined
        if ($self->{config}->parameter("build:m$i") =~ m/^(?:DOT)|(?:ontap)|(?:ontapng):[\w:]*/) {
            my @build = split /:/, $self->{config}->parameter("build:m$i");
            if ($build[0] eq 'DOT') {
                my $arch = 'x86_64';
                if (defined $build[2]) { #if build is not default
                    $self->{config}->parameter("build:m$i", "/x/eng/rlse/DOT/$build[1]/final/bedrock/export/$arch.$build[2]/tarball/image.tgz");
                } else {
                    $self->{config}->parameter("build:m$i", "/x/eng/rlse/DOT/$build[1]/final/bedrock/export/$arch/tarball/image.tgz");
                }
            } elsif ($build[0] eq 'ontap') {
                $self->{config}->parameter("build:m$i", "/x/eng/rlse/ontap/$build[1]");
            } elsif ($build[0] eq 'ontapng') {
                my $arch = 'x86_64';
                if (defined $build[2]) {
                    $self->{config}->parameter("build:m$i", "/x/eng/rlse/ontapng/$build[1]/relroot/$arch/$build[2]/image.tgz");
                } else {
                    $self->{config}->parameter("build:m$i", "/x/eng/rlse/ontapng/$build[1]/relroot/$arch/image.tgz");
                }
            } else {
                confess('build ' . $self->{config}->parameter("build:m$i") . ' not recognized');
            }
        }
    }

    $self->set_opt('aamp_run_config:t', "$AAMP_INSTALL/current/aamp_run.cfg");
    $self->set_opt('defaults_dir:t', "$PERFANT_INSTALL/PerfDev/defaults");
    $self->{pp}->add_parameter('aamp_run_config', 't',
                               "$AAMP_INSTALL/current/aamp_run.cfg", 1);
    $self->{pp}->add_parameter('defaults_dir', 't',
                               "$PERFANT_INSTALL/PerfDev/defaults", 1);

    my $test_harness = RunSubmissions::TestHarness->get_instance();
    $test_harness->configure_using_parameter_set($self->{pp});
    $test_harness->log_dir('.');
    $test_harness->show_tests();

    my $submit_string = $self->{config}->resubmit_string();
    $Log->debug("\$submit_string: $submit_string");

    # parse_submit_line initializes the PerfantParameterSet
    my $line_err = parse_submit_line($submit_string);

    if ($line_err) {
        confess("parse_submit_line: $line_err");
    }

    my ($switches_ref, $resources_ref, $backend_sws_ref, $busy_ref,
            $sources_ref, $partner_ref, $protocols_ref, $state_ref) =
            read_resources();

    my $defaults_err = default_parms(1, $switches_ref, $backend_sws_ref,
            $sources_ref, $protocols_ref, $partner_ref, 1);

    if ($defaults_err) {
        confess("default_parms: $defaults_err");
    }

    $self->{pp}->update_parameters_on_resubmit();
    my $doc = $self->{pp}->get_aamp_xml_doc();
    if (defined($doc)) {
        aamp_add_slsl_parameters($doc, $self->{pp});
        aamp_add_testbed_parameters($doc, $self->{pp});
        store_config($doc);
    }

    my ($errors, $warnings) = check_rules(1, $backend_sws_ref, $sources_ref,
            $partner_ref);

    if ($warnings) {
        $Log->warn("check_rules: $warnings");
    }
    if ($errors) {
        confess("check_rules: $errors");
    }

    if (granular_lock('queue', 20, 'PerfANTRun.pm')) {
        confess("Failed to acquire lock: $Lock_dir");
    }
    $Log->debug("Acquired lock $Lock_dir");
    my $submit_err;
    my $success = eval {
        $self->_config_template();
        confess('Failed to expand template')
            if ($self->{setlist_file} =~ m/\.slst$/);
        $self->validate_setlist();
        $self->_config_xml();
        $self->_config_commands();

        $self->{pp}->add_parameter('setlist_file', 't',
                                   $self->{setlist_file}, 1);
        $self->{pp}->add_parameter('sound_meter_prestats', 'm1',
                                   $self->{config}->parameter('sound_meter_prestats:m1'), 1);
        $self->{pp}->add_parameter('sound_meter_poststats', 'm1',
                                   $self->{config}->parameter('sound_meter_poststats:m1'), 1);

        if (!$self->{DRY_RUN}) { #if the object is in debug mode, do not actually submit test.
            $submit_err = append_test($self->{config}->user,
                    $self->{config}->testname);
        }
        1;
    };
    granular_unlock('queue', 'PerfANTRun.pm');
    if (!$success) {
        confess("Failure during submit. $@");
    }
    if ($submit_err) {
        if ($self->{slst} && $self->{setlist_file} =~ m/\.slsl$/) {
#if the file was set to generate a setlist from template, and it had an error during submit, then the template must have been finished
#expanding, so delete the test-specific setlist
            unlink($self->{setlist_file});
        }
        confess("append_test: $submit_err");
    }
    return $self->{config}->resubmit_string() if ($self->{DRY_RUN}); #return the submit string for this test
    return 1;
}

=head2 get_status

    usage: $status = $run->get_status();
           $status = $run->get_status($num_retries); #retry up to $num_retries times if test cannot be found

    return vaues: ('QUEUED', queue_pos) test is in the queue, but has not yet started; test is the queue_pos test in the queue
                  'RUNNING'             test is running
                  'SUCCESS'             test has successfully completed
                  'ERR_*'               test exited with an error, such as ERR_PANIC, ERR_USER-KILL_IN_PARENT, or ERR_OTHER to name a few

    confesses an error if the test cannot be found before $num_retries tries (default 1)

=cut

sub get_status {
    my $self = shift;
    my $max_retries = shift;
    if (!defined $max_retries) {
        $max_retries = 1;
    }
    my $purpose = $self->{config}->parameter('purpose:t');
    $Log->debug("\$purpose = $purpose");
    _set_lab($self->{config}->test_variable('lab'));
    my $retries = 0; #number of times tried

    while ($retries <= $max_retries) {
        if ($retries != 0) {
            sleep(15);
        }
        if (!$self->{has_runid}) { #has not been assigned a runid yet, may still be queued
            $Log->debug("Runid not assigned, may be queued");
            my $queue_pos = 1;
            my $fh;
            if (!open $fh, '<', $Queue_file) {
                $Log->warn("Could not open $Queue_file");
                $retries++;
                next;
            }
            while (<$fh>) {
                next if (m/^\#/);
                if (m/.*?$purpose.*?/) {
                    close $fh;
                    return ('QUEUED', $queue_pos);
                } else {
                    $Log->debug("Queue File: $_");
                    $queue_pos++;
                }
            }
            close $fh;
            $web_service->query_form( {
                action => 'runIdSearch',
                userId => $self->{config}->user(),
                user => $self->{config}->user(),
                purpose => $purpose,
            });
            $Log->debug('URI: ' . $web_service->as_string);
            my $runid_search = get($web_service->as_string)  or confess('Could not access URI ' . $web_service->as_string);
            if ($runid_search =~ m {<Result>.*?<RunId>([\w\d]+)</RunId>.*?</Result>}ms) {
                $self->{runid} = $1;
                $self->{has_runid} = 1;
                $Log->debug("RunId = $self->{runid}");
                return 'RUNNING';
            }
        } else {
            $Log->debug("RunId = $self->{runid}");
            $web_service->query_form( {
                action => 'runIdSearch',
                userId => $self->{config}->user(),
                user => $self->{config}->user(),
                runId => $self->{runid}
            });
            my $runid_search = get($web_service->as_string) or confess('Could not access URI ' . $web_service->as_string);
            if ($runid_search =~ m {<StatusCode>(?:IN)?COMPLETE</StatusCode>}ms) {
#test complete, but completion status (SUCCESS, ERR_*) has not been returned yet; considered 'running'
                return 'RUNNING';
            }

            if ($runid_search =~ m {<StatusCode>([\w\-]+)</StatusCode>}ms) {
                return $1;
            }
        }
        $retries++;
    }
    confess('Test lost');
}

=head2 get_runid

    usage: $run->get_runid();

    Return the runid if available, else return undef.

=cut

sub get_runid {
    my $self = shift;

    return $self->{runid};
}

=head2 add_template_var

    usage: $run->add_template_var('num_clients', 2);

    Add a variable for the setlist template, similar to a #define in C.

=cut

sub add_template_var {
    my $self = shift;
    my $name = shift;
    my $replacement_text = shift; #text that will replace the var name
    $Log->debug("Setting '$name' to '$replacement_text' in stash");

    $self->{template_vars}{$name} = $replacement_text;
    return $self;
}

=head2 load_xml

    usage: $run->load_xml($custom_xml_config);

    load a specific xml config file for this test; currently only handles tags with the attribute name=setlist_file, for motley tests

=cut

sub load_xml {
    my $self = shift;
    my $xml_file = shift;

    my $parser = $self->{parser};
    my $xml = $parser->XMLin($xml_file, KeepRoot => 1)
        or confess("Cannot open $xml_file for reading");

    my $setlist_file = $xml->{test}{group}{slsl}{parameter}{setlist_file}{value}{content};
    $self->{setlist_file} = $setlist_file if (defined $setlist_file);

    return 1;
}

=head2 wait

    usage: $run->wait(); #waits for $run to complete, checking its status on a specific interval
           $run->wait(200); #like wait(), but checks run's status ever 200 seconds

    Waits for the run to complete, or die of an error. Will return the final status (SUCCESS or the error code)

=cut

sub wait {
    my $self = shift;
    my $interval = shift;
    if (!defined $interval) {
        $interval = 300; #default to checking every 5 min
    }
    my $status = $self->get_status(5); #get status, and allow up to 5 retries if the test can't be found
    while ($status !~ m/(?:SUCCESS)|(?:ERR_\w+)/) {
        sleep $interval;
        $status = $self->get_status(5); #get status, and allow up to 5 retries if the test can't be found
    }
    return $status;
}

=head2 add_precommand
=head2 add_postcommand

    usage: $run->add_precommand(...)

    Adds measurement command to be run at the beginning/end of each observation.  Takes hash values to configure command.  Available values are:
=over
=item * id: string, optional, identifies the specs
=over
=item * omitted, an id is auto-generated
=back
=item * command: string, required, command to be run that is valid in the cli context
=item * cli: string, choice: ngsh, bsd, dblade, monitor
=item * priv: string, choice: admin, advanced, diag, test
=item * section: string, the PerfANT section for the output file
=over
=item * an empty section, section => "", will print the banner but no text
=item * omitting the section will print no banner and no text
=back
=item * heading: string, the label that goes next to the section in the output file
=over
=item * an empty heading, heading => "", will no text
=item * omitting the heading keyword will print an auto-generated heading based on the command string
=back
=item * foreground: boolean, if the command should be run as a foreground or background process
=item * output: string, file where stdout will go
=over
=item * omitting, goes the SoundMeter default output file if not specified
=back
=item * error: string, file where stderr will go
=over
=item * omitting, goes the SoundMeter default output file if not specified
=back
=item * pid_file: string, file to store the process pid
=item * labels: string, describe under what circumstances the command will execute compared against the SoundMeter labels at runtime
=over
=item * empty and omitted, the command is always executed
=back
=item * timeout: integer, maximum number of seconds to allow the command to execute
=item * timeout_panic: boolean, if the timeout is specified, and a timeout is triggered, panic the controller
=back

=cut

sub add_precommand {
    my $self = shift;
    my %command = @_;
    my %cloned_command = %{Clone::clone(\%command)};
    $Log->debug('command: ' . Dumper(\%cloned_command));
    push @{$self->{precommands}}, \%cloned_command;
    return 1;
}

sub add_postcommand {
    my $self = shift;
    my %command = @_;
    my %cloned_command = %{Clone::clone(\%command)};
    $Log->debug('command: ' . Dumper(\%cloned_command));
    push @{$self->{postcommands}}, \%cloned_command;
    return 1;
}

sub _config_xml {
    $Log->debug('Creating xml');
    my $self = shift;
    if ($self->{config}->testname() eq 'motley') {
        if (defined $self->{setlist_file}) {
            if (defined $self->{pp}->get_value('xml_config', 'q')) {
                my $setlist_file = $self->{setlist_file};
#save a copy of the setlist to the test directory

                #path the setlist will go to
                my $new_setlist_path = dirname($self->{pp}->get_value('xml_config', 'q'));
                my $setlist_name = fileparse($setlist_file);

                $Log->debug("copying $setlist_name to queue directory");
                cp($setlist_file, "$new_setlist_path/$setlist_name");
                $self->{setlist_file} = "$new_setlist_path/$setlist_name";

#set the setlist of choice in the xml configuration
                my $parser = $self->{parser};
                my $xml_file = $self->{pp}->get_value('xml_config', 'q');
                my $xml_config = $parser->XMLin($xml_file, KeepRoot => 1)
                    or confess("Cannot open $xml_file for reading");
                $xml_config->{test}{group}{slsl}{parameter}{setlist_file}{value}{content} = $self->{setlist_file};

                open my $fh, '>', $xml_file
                    or confess("Cannot open $xml_file for writing: $!");
                print {$fh} $parser->XMLout($xml_config, KeepRoot => 1);
                close $fh or confess("Could not close $xml_file after writing: $!");
            } else {
                $Log->warn('No xml_config:q parameter');
            }
        } else {
            $Log->warn('No setlist');
        }
    } else {
        $Log->comment('Not a motley test, no xml configuration needed.');
    }
    return 1;
}

sub _config_commands {
    my $self = shift;
    my $uid = $self->{uid};
    my (undef, $config_dir) = fileparse($self->{pp}->get_value('xml_config', 'q'));
    my $prestats = ($self->{config}->parameter("sound_meter_prestats:m1") ||
            '/u/spec/aamp/current/aamp/data_collection/prestats.slsl');
    my $poststats = ($self->{config}->parameter("sound_meter_poststats:m1") ||
            '/u/spec/aamp/current/aamp/data_collection/poststats.slsl');
    my ($prestats_name, $prestats_dir) = fileparse($prestats);
    my ($poststats_name, $poststats_dir) = fileparse($poststats);

    #keep the base files for clones
    $self->{base_prestats} = $prestats;
    $self->{base_poststats} = $poststats;

    if (defined $self->{precommands}) {
#create a copy of the prestats slsl file, then append custom commands to it
        my $new_file = "$config_dir/$prestats_name";
        $new_file =~ s/\.slsl$/_$uid\.slsl/;
        cp($prestats, $new_file);
        open my $fh, '>>', $new_file or confess("Could not open $new_file for writing");
        foreach my $command (@{$self->{precommands}}) {
            my @option_strings; #strings of opt=value to put into the command parameter
            foreach my $opt (keys %{$command}) {
                if ($command->{$opt} =~ m/\D/) {#if the option contains anything other than just numbers, stringify it
                    $command->{$opt} = qq/"$command->{$opt}"/;
                }
                push @option_strings, "$opt=$command->{$opt}";
            }
            my $command_string = 'spec(' . join(', ', @option_strings) . ')';
            print {$fh} $command_string;
        }
        close $fh or confess("Cannot close file $fh");
        $self->set_opt("sound_meter_prestats:m1", $new_file);
    }

    if (defined $self->{postcommands}) {
#do the same as above for poststats slsl file
        my $new_file = "$config_dir/$poststats_name";
        $new_file =~ s/\.slsl$/_$uid\.slsl/;
        cp($poststats, $new_file);
        open my $fh, '>>', $new_file or confess("Could not open $new_file for writing");
        foreach my $command (@{$self->{postcommands}}) {
            my @option_strings; #strings of opt=value to put into the command parameter
            foreach my $opt (keys %{$command}) {
                if ($command->{$opt} =~ m/\D/) {#if the option contains anything other than just numbers, stringify it
                    $command->{$opt} = qq/"$command->{$opt}"/;
                }
                push @option_strings, "$opt=$command->{$opt}";
            }
            my $command_string = 'spec(' . join(', ', @option_strings) . ')';
            print {$fh} $command_string;
        }
        close $fh or confess("Cannot close file $fh");
        $self->set_opt("sound_meter_poststats:m1", $new_file);
    }
    return 1;
}

sub _config_template {
    my $self = shift;
    return 1 if ($self->{setlist_file} !~ m/\.slst$/);
    $self->{setlist_template} = $self->{setlist_file};#keep backup of slst file for clones

    my $parameters = $self->{config}->parameters();
    $Log->debug(Dumper($parameters));
    foreach my $namespace (keys %{$parameters->{model}}) { #set filers/defaults for each filer first
        $self->_set_hardware($parameters->{model}{$namespace}, $namespace);
    }
    foreach my $parameter (keys %{$parameters}) {
#        next if ($parameter eq 'model');
        foreach my $namespace (keys %{$parameters->{$parameter}}) {
            $Log->debug("Setting '" . $parameter . '_' . $namespace . " = '$parameters->{$parameter}{$namespace}'");
            if ($parameters->{$parameter}{$namespace} !~ m/\D/) {#if there are no non-digit characters in the parameter,
#                                                                   leave it without quotes
                $self->{stash}->set($parameter . '_' . $namespace, $parameters->{$parameter}{$namespace});
            } else {
                $self->{stash}->set($parameter . '_' . $namespace, qq("$parameters->{$parameter}{$namespace}"));
            }
        }
    }

    my $test_vars = $self->{config}->test_variables();
    foreach (keys %{$test_vars}) {
#there is no .def file for :t variables, but some of those may be specific to a test
#however, there is a part of PerfCommon::Runid that contains the value of these variables
        if ($test_vars->{$_} !~ m/\D/) {
            $self->{stash}->set($_ . '_t', $test_vars->{$_});
        } else {
            $self->{stash}->set($_ . '_t', qq("$test_vars->{$_}"));
        }
    }

    foreach (keys %{$self->{template_vars}}) {
        $self->{stash}->set($_, $self->{template_vars}{$_});
    }

    my $new_setlist_path = $self->{setlist_file}; #setlist is technically not a perfant option, so it needs to be seperate
    $new_setlist_path =~ s/\.slst$/_$self->{uid}\.slsl/;

    $self->{template}->process($self->{setlist_file}, $self->{stash}, $new_setlist_path)
            or confess("Could not expand $self->{setlist_file} into $new_setlist_path: " . $self->{template}->error());
    $self->set_opt('setlist_file:t', $new_setlist_path);
    return 1;
}

=head2 validate_setlist

    usage: $run->validate_setlist();

    Validate setlist file using sizzle_check

=cut

sub validate_setlist {
    my $self = shift;
    my $setlist_file = $self->{setlist_file};

    my $sizzle_check = `PYTHONPATH=$AAMP_INSTALL/current/lib/python2.6/site-packages \
            $AAMP_INSTALL/current/bin/sizzle_check.py -c $setlist_file --no_warnings`;

    if ($sizzle_check =~ m/Validation Successful/ms) {
        $Log->comment("sizzle_check successful");
    }
    else {
        confess("$setlist_file is not a valid slsl script: $sizzle_check");
    }
    return 1;
}

#private method to properly find the base lab directory based on the runid of the original test
sub _set_lab {
    my $runid = shift;
    $Log->debug("Setting lab for run '$runid'");
    my $new_lab_name;
    if (defined $runid_prefix {$runid}) {
#received lab name instead of runid
        $new_lab_name = $runid;
    }
    my $prefix = substr($runid, 6, 1);
    foreach (keys(%runid_prefix)) {
        if ($prefix eq $runid_prefix {$_}) {
            $new_lab_name = $_;
        }
    }
    $Log->debug("\$new_lab_name: $new_lab_name");
    if (!defined $new_lab_name) {
        return 0;
    }
    if (!defined $Lab_name) {
        $Lab_name = $new_lab_name;
        set_lab_specific_config();
    }
    if ($Lab_name ne $new_lab_name) {
        $Lab_name = $new_lab_name;
        set_lab_specific_config();
    }
    return $Lab_name;
}

sub _set_hardware {
    my $self = shift;
    my $name = shift; #name of hardware model (ex. FAS6080 or FAS6070_GX)
    my $namespace = shift; #what to append to names, such as 't', or 'm1' (usually the part after the ':' in PerfANT options)

    $self->{model}{$namespace} = $name;

    my $def_file = $self->{config}->parameter('defaults_dir:t') . "/$name/$name.def";
    open my $fh, '<', $def_file or confess("Failed to get defaults for model $name");

    while (<$fh>) {
        chomp;
        next if (m/^\s*#/);# ignore comments (are comments allowed in the .def file?)
        if (my ($name, $value) = m/^(.*?)\s*=\s*(.*?)$/) {
            if (defined $namespace && $namespace ne '') {
                $Log->debug("Using namespace $namespace for $1");
                $name .= "_$namespace";
            }
            $Log->debug("Setting '$name' = '$value'");
            if ($value =~ m/(?:^\d+$)    |  #if $value is only numbers
                            (?:^"[^"]*"$)|  #if $value is an already quoted string (with double quotes)
                            (?:^'[^']*'$)   #if $value is an already quoted string (with single quotes)
                           /x) {
                $self->{stash}->set($name, $value);
            } else {
                $self->{stash}->set($name, qq{"$value"}); #if the value is not a number, put it in quotes
            }
        }
    }
    close $fh or confess("Failed to close file $name");
    return 1;
}

sub _get_uid {
    my $now = DateTime->now( time_zone => 'local' );
    return sprintf '%02d%02d%02d-%06d-%06d', $now->month, $now->day, $now->year % 100, $now->hms(''), int rand 1_000_000;
}

=head2 append_to_xml

    usage: $run->append_to_xml();

    Append content to the PerfANT XML config file

=cut

sub append_to_xml {
    my $self = shift;
    my $content = shift;

    $Log->debug("Appending to xml");
    $Log->debug("content: \n" . $content);

    my $parser = $self->{parser};
    my $xml_file = $self->{pp}->get_value('xml_config', 'q');
    my $xml_config = $parser->XMLin($xml_file, KeepRoot => 1)
        or confess("Cannot open $xml_file for reading");
    my $xml_content = $parser->XMLin($content, KeepRoot => 1);
    my $xml_content_root = (keys %$xml_content)[0];

    # 'test' is the root of the PerfANT XML config file. Add this additional
    # content under this tag.
    $xml_config->{test}{$xml_content_root} = $xml_content->{$xml_content_root};
    open my $fh, '>', $xml_file
        or confess("Cannot open $xml_file for writing");
    print {$fh} $parser->XMLout($xml_config, KeepRoot => 1);
    close $fh or confess("Could not close $xml_file after writing");

    return 1;
}


1;

