package Perf::PerfANTRun::Results;

use strict;
use warnings;

use Carp qw(confess);

use Tharn qw($Log);

use Perf::PerfANTRun;
use Perf::PerfANTRun::Stats;

use XML::Simple;
use LWP::Simple;
use Data::Dumper;

=head1 NAME

    Perf::PerfANTRun::Results - Class for getting results for a specific Perf::PerfANTRun object

=head1 SYNOPSIS

    my $run = Perf::PerfANTRun->new();
    my $results;
    ... set up run ...
    $run->submit();
    if ($run->get_status() eq 'SUCCESS') {
        $results = Perf::PerfANTRun::Results->new($run);
        my $result = $results->get_result('stats system');
    }

=cut

my $web_service = URI->new('http://grover.rtp.netapp.com/KO/index.php/common/ws/process');


=head1 FUNCTIONS
=head2 new

    usage: my $results = Perf::PerfANTRun::Results->new($my_PerfANTRun);

=cut

sub new {
    my $class = shift;
    my $root_run = shift; #Reference to the PerfANTRun from which to get results

    my $self = {
        root    =>  $root_run,
        parser  =>  XML::Simple->new(
            NormalizeSpace  =>  2,
            ForceContent    =>  1,
            GroupTags       =>  {
                RunList                     =>  'Run',
                IterationObservationList    =>  'IterationObservation',
                Object                      =>  'Instance',
                Instance                    =>  'Counter',
            },
            ForceArray      =>  [
                'Object',
                'Counter',
                'Instance',
                'Run',
                'IterationObservation',
                'Command',
            ],
            KeyAttr         =>  {
                Object      =>  'name',
                Instance    =>  'name',
                Counter     =>  'name',
                Metric      =>  'name',
            },
        ),
    };
    $web_service->query_form({
        action => 'getTestInfo',
        userId => $self->{root}{config}->user(),
        runid => $self->{root}{runid},
    });
    my $xml = get($web_service->as_string) or confess("Could not access URI $web_service->as_string");
    #my $get_test_info_handler = PerfANTRun::Results::TestInfoHandler->new();
    my $xml_tree = $self->{parser}->XMLin($xml);
    $Log->debug(Dumper($xml_tree));

    $self->{metrics} = $xml_tree->{body}{MesssageBody}{MasterTest}{Metrics};

    $Log->trace("foreach run");

    foreach my $run (@{$xml_tree->{body}{MessageBody}{MasterTest}{RunList}}) {
        my $runid = $run->{RunId}{content};
        next if ($runid ne $self->{root}{runid}); #ignore runs with different runid from the master

        $Log->trace("foreach iteration/observation in run $runid");

        foreach my $iter_obs (@{$run->{IterationObservationList}}) {
            my $iterid = $iter_obs->{IterationId}{content};
            my $obsid = $iter_obs->{ObservationId}{content};

            $Log->trace("foreach command in iteration $iterid and observation $obsid");

            foreach my $command (@{$iter_obs->{CommandList}{Command}}) {
                my $command_name = $command->{content};
                next if (!defined $command_name || $command_name eq ''); #ignore empty tags
                $Log->trace("\$self->{commands}{$iterid}{$obsid}{$command_name} = ''");
                $self->{commands}{$iterid}{$obsid}{$command_name} = '';
            }
        }
    }

    bless $self, $class;
    return $self;
}

=head2 get_result

    usage: my $command_result = $results->get_result($iter, $obs, $command);

    returns the output of $command for a given iteration and observation id. If perfant web services has parsed the command to XML, the result will be a hash_ref with the following items:
=over 4

=item * attributes: a hash containing attribute names as its keys and the respective values as its values

=item * tags: an array of other tag hashes that are children of this tag

=item * text: string of text inside of the tag, where beginning and trailing whitespace is removed

=back

    If the command is not XML compatible, returns the raw output

=cut

sub get_result {
    my $self = shift;
    my $iterid = shift;
    my $obsid = shift;
    my $command = shift;
    if (!defined $self->{commands}{$iterid}{$obsid}{$command}) {
        confess("Invalid command $command in iteration $iterid and observation $obsid");
    }
    if ($self->{commands}{$iterid}{$obsid}{$command} ne '') {
        return $self->{commands}{$iterid}{$obsid}{$command}; #command result is cached, return the cached version.
    }
    eval { $self->get_result_xml($iterid, $obsid, $command); };
    if ($@) { #command can't be found as XML, fall back to text.
        warn $@;
        eval { $self->get_result_text($iterid, $obsid, $command); };
        if ($@) { #command can't be found as text.
            confess($@);
        }
    }
    return $self->{commands}{$iterid}{$obsid}{$command};
}

=head2 get_stats

    usage: $system = $results->get_stats($iter, $obs, $command);
           $system = $results->get_stats(1, 0, 'system');

    returns a tree derived from the getCommandXML service for command 'stats $command' in iteration $iter and observation $obs
    (ex: command 'stats system', iteration 1, observation 0)

=cut

sub get_stats {
    my $self = shift;
    my $iter = shift; #iteration id
    my $obs = shift; #observation id (0 if irrelevant)
    my $stats_subcommand = shift; #command after 'stats' ex: for the command 'stats workload', the function should be called with 'workload'
    my $command = "stats $stats_subcommand";

    if (!defined $self->{commands}{$iter}{$obs}{$command}) {
        if (!defined $self->{commands}{$iter}{$obs}{stats}) {
            confess("no such command $command for iteration $iter and observation $obs");
        }
    }

    if ($self->{commands}{$iter}{$obs}{$command} ne '') {
        my $stats = Perf::PerfANTRun::Stats->new($self->{commands}{$iter}{$obs}{$command}); #results are cached. return them.
        return $stats;
    }

    $web_service->query_form(
        action          =>  'getCommandXML',
        userId          =>  $self->{root}{config}->user(),
        runid           =>  $self->{root}{runid},
        iterationid     =>  $iter,
        observationid   =>  $obs,
        command         =>  $command,
    );
    my $xml = get($web_service->as_string) or confess("Could not access URI " . $web_service->as_string);
    my $xml_tree = $self->{parser}->XMLin($xml);
    $xml_tree = $xml_tree->{body}{MessageBody}{command}{Object};
    foreach (keys %{$xml_tree}) {
        $xml_tree->{$_} = $xml_tree->{$_}{Instance};
        my $object = $xml_tree->{$_};
        foreach (keys %{$object}) {
            $object->{$_} = $object->{$_}{Counter};
        }
    }
    my $stats = Perf::PerfANTRun::Stats->new($xml_tree);

    $self->{commands}{$iter}{$obs}{$command} = $xml_tree;

    return $stats;
}

=head2 get_result_xml

    usage: $tree = $results->get_result_xml($iter, $obs, $command);

    returns a hash representing the xml from the getCommandXML service
    (if you did not specify an observation in your test, put 0 for $obs)
    throws an error if there is no xml for the command for a given iteration and observation

=cut

sub get_result_xml {
    my $self = shift;
    my $iter = shift;
    my $obs = shift;
    my $command = shift;

    if (!defined $self->{commands}{$iter}{$obs}{$command}) {
        confess("no such command $command for iteration $iter and observation $obs");
    }
    if (ref $self->{commands}{$iter}{$obs}{$command} eq 'HASH') {
        return $self->{commands}{$iter}{$obs}{$command}; #if result is cached
                #as a hash (indicating that it was from the xml parser),
                #return it
    }

    $web_service->query_form(
        action          =>  'getCommandXML',
        userId          =>  $self->{root}{config}->user(),
        runid           =>  $self->{root}{runid},
        iterationid     =>  $iter,
        observationid   =>  $obs,
        command         =>  $command,
    );

    my $xml = get($web_service->as_string);
    my $xml_tree = $self->{parser}->XMLin($xml);
    $xml_tree = $xml_tree->{body}{MessageBody}{command};

    $self->{commands}{$iter}{$obs}{$command} = $xml_tree;
    return $self->{commands}{$iter}{$obs}{$command};
}

=head2 get_result_text

    usage: $tree = $results->get_result_text($iter, $obs, $command);

    returns the output text for command $command in iteration $iter and observation $obs
    (if you did not specify an observation in your test, put 0 for $obs)
    throws an error if the command was not run for a given iteration and observation

=cut

sub get_result_text {
    my $self = shift;
    my $iter = shift;
    my $obs = shift;
    my $command = shift;

    if (!defined $self->{commands}{$iter}{$obs}{$command}) {
        confess("no such command $command for iteration $iter and observation $obs");
    }
    if (ref $self->{commands}{$iter}{$obs}{$command} eq 'SCALAR' && $self->{commands}{$iter}{$obs}{$command} ne '') {
        return $self->{commands}{$iter}{$obs}{$command}; #if result is cached as text, return it
    }

    $web_service->query_form(
        action          =>  'getCommandText',
        userId          =>  $self->{root}{config}->user(),
        runid           =>  $self->{root}{runid},
        iterationid     =>  $iter,
        observationid   =>  $obs,
        command         =>  $command,
    );

    my $xml = get($web_service->as_string);
    my $xml_tree = $self->{parser}->XMLin($xml);
    $self->{commands}{$iter}{$obs}{$command} = $xml_tree->{body}{MessageBody};

    return $self->{commands}{$iter}{$obs}{$command};
}

1;
