package Perf::PerfANTRun::Stats;

use strict;
use warnings;

use Tharn qw($Log);

use List::Util qw(first);
use Carp qw(confess);
use Data::Dumper;

=head1 NAME

    Perf::PerfANTRun::Stats - Class for handling/formatting results from a stats command

=head1 SYNOPSIS

    my $run = Perf::PerfANTRun->new();
    my $results;

    ... set up run ...
    
    $run->submit();
    if ($run->get_status() eq 'SUCCESS') {
        $results = Perf::PerfANTRun::Results->new($run);
        my $workload_stats = $results->get_stats(1, 0, 'workload');

        my $ops = $workload_stats->get('workload_1', 'ops');
        print "Ops: $ops";

        my $hash_ref = $workload_stats->hash();
        $ops = $hash_ref->{'workload:workload_1:ops'};
        print "Ops: $ops"; #will print the same value as above
    }

=cut

=head1 FUNCTIONS
=head2 new

    usage: my $stats = Perf::PerfANTRun::Stats->new($results->get_result('stats system'));

=cut

sub new {
    my $class = shift;
    my $results_tree = shift; #'raw' tree parsed from xml that will be handled by this class
    my $self = {%{$results_tree}};
    $Log->debug(Dumper($results_tree));
    my @keys = keys %{$results_tree};
    $self->{top_level} = $keys[0]; #there should only be one key in the top level of the tree, and that should be the name of the command that was run, such as 'workload' or 'system'
    bless $self, $class;
    return $self;
}

=head2 get

    usage: my $latency = $workload->get('User-Default', 'read_latency_hist', '<20us'); #gets value of workload:User-Default:read_latency_hist.<20us
           my $ops = $workload->get('User-Default', 'ops'); #gets value of workload:User-Default:ops
           my %latencies = $workload->get('User-Default', 'read_latency_hist'); #gets hash of values of
#                   workload:User-Default:read_latency_hist.<20us, workload:User-Default:read_latency_hist.<1ms,
#                   workload:User-Default:read_latency_hist.<2ms, workload:User-Default:read_latency_hist.<4ms,
#                   workload:User-Default:read_latency_hist.<6ms, etc.

=cut
sub get {
    my $self = shift;
    my $instance = shift; #category for counters, ex. System-Default, vol1_wkld
    my $counter = shift; #specific stat to use, such as ops, read_latency_hist
    my $subcounter = shift; #subcommand, for commands with multiple results (such as latency_hist, where there is latency_hist.<20us, latency_hist.<1ms, latency_hist.<2ms, etc.)

    if (defined $subcounter) {
        $counter .= ".$subcounter";
    }

    if (defined $self->{$self->{top_level}}{$instance}{$counter}) {
        return $self->{$self->{top_level}}{$instance}{$counter}{content};
    } else {
        my @counters = grep {/^$counter/} (keys %{$self->{$self->{top_level}}});
        if (@counters == 0) {
            confess("$counter does not exist");
        }
        my $results = {};
        $results->{$_} = $self->{$self->{top_level}}{$instance}{$_}{content} foreach (@counters);
        return $results;
    }
}

=head2 hash

    usage: my $hashref = $workload_stats->hash();

=cut
sub hash {
    my $self = shift;
    my $hash = {};
    foreach my $instance (keys %{$self->{$self->{top_level}}}) {
        foreach my $counter (keys %{$self->{$self->{top_level}}{$instance}}) {
            $hash->{"$self->{top_level}:$instance:$counter"} = $self->{$self->{top_level}}{$instance}{$counter}{content};
        }
    }
    return $hash;
}

1;
