# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Perf/PerfArchiver.pm#1 $
# Copyright (c) 2007 Network Appliance, Inc.
# All rights reserved.
# Perf/PerfArchiver.pm
#
## @summary  Perl library for Performance Archiver routines.
## @author   farida@netapp.com, slane@netapp.com
## @status   private
## example
## my $pa = new Perf::PerfArchiver(fc=>$fc);
## $pa->enable();
##

package Perf::PerfArchiver;

use strict;
use warnings;
use Tharn;
use Carp;
use filer;
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Module::Common qw(validate_mixed);
use Filer::Version qw(are_we_ontapng);
use Server;
#use perf::performance_archiver::CM_Archiver qw(_say _filer_options_get);

# RESTART option. If this option is passed, then Archiver needs to be disabled before option can be set.
use constant RESTART => 1;

my $init = 0;
BEGIN {
    $PerfArchiver::init = $$;
}

sub new {
    my $class = shift;
    my %params = validate_mixed(
        @_,
        [ ],
        {
                host                    => {type => SCALAR,         optional => 1},
                fc                      => {type => OBJECT,         optional => 1},
                cc                      => {type => OBJECT,         optional => 1},
				client					=> {type => SCALAR,			required => 1},
                IS_NG                   => {type => SCALAR,         optional => 0},
                test                    => {type => OBJECT,         optional => 1},
                enable                  => {type => SCALAR | UNDEF, optional => 1},
                'preset.level'          => {type => SCALAR | UNDEF, optional => 1},
                max_disk_space_percent  => {type => SCALAR | UNDEF, optional => 1},
                mountpoint              => {type => SCALAR | UNDEF, optional => 1},
                arch_dir                => {type => SCALAR | UNDEF, optional => 1},
                preset_dir              => {type => SCALAR | UNDEF, optional => 1},
                # Files generated during testing will be stored in this NATE log dir.(determined in each thpl)
                log_dir                 => {type => SCALAR | UNDEF, optional => 1},
                server                  => {type => SCALAR | UNDEF, optional => 1},
           }
       );

    my $self = \%params;

    my $h = $self->{IS_NG} ? $self->{host} : $self->{fc};
    if (!$h) {
        confess 'Must specify either "fc" or "host" parameter.';
    }

    # Populate the object attributes with corresponding values on the filer.
    my @options_array = qw(enable max_disk_space_percent);
    foreach my $option (@options_array) {
        $self->{$option} = 
			perf::performance_archiver::CM_Archiver::_filer_options_get($h, $self->{IS_NG}, "stats.archive.$option");
    }
    # preset.level with (')s around it is not handled by filer_options_get
    my $out = 
		perf::performance_archiver::CM_Archiver::_say($h, $self->{IS_NG}, "options stats.archive.preset.level");
    $out =~ /stats\.archive\.preset\.level\s+(\S+)/;
    $self->{'preset.level'} = $1;

    # Archiver directories.
    $self->{arch_dir} = 'etc/log/stats/archive';
    $self->{preset_dir} = 'etc/log/stats/archive/.preset';

    bless $self, $class;
    return $self;
}


# Return stats.archive.enable option value, optionally setting
# it first if a second argument is supplied.

sub enable {
    my $self = shift;
    my $val = shift;

    return $self->_perfarch_option('enable', $val);
}


# Return stats.archive.preset.level option value, optionally setting
# it first if a second argument is supplied.

sub preset_level {
    my $self = shift;
    my $val = shift;

    return $self->_perfarch_option('preset.level', $val, RESTART);
}

#Return stats.archive.max_disk_space_percent option value, optionally setting
# it first if a second argument is supplied.

sub max_disk_space_percent {
    my $self = shift;
    my $val = shift;

    return $self->_perfarch_option('max_disk_space_percent', $val);
}


# Private helper function for getting/setting options.

sub _perfarch_option {
    my $self = shift;
    my $opt = shift;
    my $val = shift;
    my $restart = shift;   # If $restart=1, then the archiver needs to be disabled before option can be set.

    my $fc = $self->{fc};
    my $host = $self->{host};
    my $IS_NG = $self->{IS_NG};
    my $con = $IS_NG ? $host : $fc ;

    # If $value is defined, set it, check if the $value was set and return the current value.

    if (defined $val) {

        # Get current value for 'enable' option.
        my $enable_opt = $self->{enable};

        if ($restart) {
            $self->enable("off");
        }

        my $cmd = "options stats.archive.$opt $val";
        my $out = perf::performance_archiver::CM_Archiver::_say($con, $IS_NG, $cmd);

        if ($out ne q{} ) {
            confess "Unexpected output from command $cmd: $out";
        }

        # Reset the original value for 'enable' option.
        if ($restart) {
            $self->enable("$enable_opt");
        }

        # The start/stop of the Archiver is a bit delayed, hence sleep.
        if ($opt eq 'enable') {
            sleep (5);
        }

        # Check if the value was set
        $out = perf::performance_archiver::CM_Archiver::_say($con, $IS_NG, "options stats.archive.$opt");
        if ($out !~ /stats\.archive\.$opt\s+$val/) {
            confess "Command $cmd did not set the expected value.\n Current value is: $out";
        } else {
            $self->{$opt} = $val;                    # Set the value in the PerfArchiver obj
        }

        return $1;

    } else {

        # If $value is not defined, just return the current value.

        my $cmd = "options stats.archive.$opt";

        my $out = perf::performance_archiver::CM_Archiver::_say($con, $IS_NG, $cmd);

        if ($out !~ /stats\.archive\.$opt\s+(\S+)/) {
            confess "Unexpected output from command $cmd: $out";
        }

    return $1;
    }
}

1;
