# -*- perl -*-
# $Id: //depot/prod/test/main/ontap/boilermaker/lib/Perf/Playlist.pm#1 $
#
## @author   bshelton@netapp.com
## @summary  Convenience functions for Adaptive Playlist
## @status   private

package Perf::Playlist;

use strict;
use warnings;
use Carp qw(confess);
use Tharn;

require Exporter;

our @ISA       = qw(Exporter);
our @EXPORT_OK = qw(playlist_create
                    playlist_truncate
                    playlist_delete
                    playlist_dump
                    playlist_status);


# Create playlist metafile in the flex volume's containing aggregate.

sub playlist_create {
    my $fc   = shift;
    my $vol  = shift;
    my %opts = @_;

    # Process optional keyword arguments. Cannot specify "zeros" without
    # also specifying "length".

    my $len   = undef;
    my $zeros = undef;

    foreach my $key (keys %opts) {
        if ($key eq 'length') {
            $len = $opts{length};
        }
        elsif ($key eq 'zeros') {
            $zeros = $opts{zeros};
        }
        else {
            confess "Unrecognized option: $key";
        }
    }

    if (defined $zeros and !defined $len) {
        confess 'Option "zeros" requires option "length".';
    }

    my $cmd = "priv set -q test; playlist $vol create";
    $cmd .= " $len"    if defined $len;
    $cmd .= " $zeros"  if defined $zeros;

    my (undef, undef, $out) = say($fc, $cmd);

    return 1 if $out =~ m/create succeeded/;
    return 0 if $out =~ m/create failed/;
    confess "Unexpected output from '$cmd': $out";
}



# Truncate the flex volume's playlist metafile.

sub playlist_truncate {
    my $fc   = shift;
    my $vol  = shift;

    my $cmd = "priv set -q test; playlist $vol truncate";

    my (undef, undef, $out) = say($fc, $cmd);

    return 1 if $out =~ m/truncate succeeded/;
    return 0 if $out =~ m/truncate failed/;
    confess "Unexpected output from '$cmd': $out";
}


# Delete the flex volume's playlist metafile.

sub playlist_delete {
    my $fc   = shift;
    my $vol  = shift;

    my $cmd = "priv set -q test; playlist $vol delete";

    my (undef, undef, $out) = say($fc, $cmd);

    return 1 if $out =~ m/delete succeeded/;
    return 0 if $out =~ m/remove failed/;
    confess "Unexpected output from '$cmd': $out";
}


# Return information about the flex volume's playlist metafile.

sub playlist_status {
    my $fc   = shift;
    my $vol  = shift;

    my $cmd = "vol status $vol";
    my (undef, undef, $out) = say($fc, $cmd);

    return undef  if $out !~ m/Containing aggregate: '(.*?)'/;

    my $aggr = $1;

    $cmd = "priv set -q test; playlist $vol status";
    (undef, undef, $out) = say($fc, $cmd);

    return undef  if $out =~ m/status failed/;

    confess "Unexpected output from '$cmd': $out"
      if $out !~ m{(playlist:\s+<(.*?)>
                    .*?
                    size=(\d+)\s+
                    \(blocks=(\d+)\) )}sgx;

    my $text = $1;
    my $status = {
        vol    => $vol,
        aggr   => $aggr,
        path   => $2,
        size   => $3,
        blocks => $4
    };

    confess "Unexpected output from '$cmd': $out"
      if $out !~ m{(.*?
                    type=(\d+)\s+\((.*?)\)
                    .*?
                    perm=(\d+)
                    .*?
                    uid=(\d+),\s+gid=(\d+)
                    .*?
                    fileid=(\d+) )}sgx;

    $text .= $1;

    $status->{typeno}   = $2;
    $status->{typename} = $3;
    $status->{perm}     = $4;
    $status->{uid}      = $5;
    $status->{gid}      = $6;
    $status->{fileid}   = $7;

    confess "Unexpected output from '$cmd': $out"
      if $out !~ m{(.*?
                    atime/mtime/ctime/crtime/now=
                    (\d+)/(\d+)/(\d+)/(\d+)/(\d+) )}sgx;

    $text .= $1;

    $status->{atime}  = $2;
    $status->{mtime}  = $3;
    $status->{ctime}  = $4;
    $status->{crtime} = $5;
    $status->{now}    = $6;
    $status->{text}   = $text;

    return $status;
}


# Return the contents of the playlist metafile, along with other
# info about the file.
#
# Return undef if the volume doesn't have a playlist file, or if
# the volume doesn't exist.

sub playlist_dump {
    my $fc   = shift;
    my $vvol = shift;

    # Get containing aggregate, fileid, and number of blocks.

    my $status = playlist_status($fc, $vvol);
    return undef  if !defined $status;

    # Use "dumpblock file -A <aggr> <fileid> <blockno>" command
    # to read the contents of each block of the playlist file.

    my @blocks;

    for (my $blockno = 0; $blockno < $status->{blocks}; $blockno++) {
        my $cmd = "priv set -q test; dumpblock file"
                . " -A $status->{aggr} $status->{fileid} $blockno";
        my (undef, undef, $dump) = say($fc, $cmd);

        push @blocks, $dump;
    }

    return _parse_dump($vvol, $status->{aggr}, $status->{fileid}, \@blocks);
}


# Parse the output of the "playlist <vol> status" command, and return
# hash of interesting info (including blockno 0 contents) about the
# the playlist metafile.

sub _parse_dump {
    my $vvol   = shift;
    my $aggr   = shift;
    my $fileid = shift;
    my $blocks = shift;

    my $info = {
        vol    => $vvol,
        aggr   => $aggr,
        fileid => $fileid,
        blocks => [
        ]
    };

    my $blockno = 0;

    foreach my $dump (@$blocks) {

        # Parse dumpblock's output.

        $dump =~ m/^disk\s+(\d\w\.\d\d),\s+dbn\s+(\d+),\s+vbn\s+(\d+)\n
                   (.*?)
                   cksum_data:\s+
                   disk\s+(\d\w\.\d\d),\s+dbn\s+(\d+)\n
                   (.*)$/sgx;

        my $data_disk   = $1;
        my $data_dbn    = $2;
        my $data_vbn    = $3;
        my $cksum_disk  = $5;
        my $cksum_dbn   = $6;

        my $data_text   = $4;
        chomp $data_text;
        my $data_bytes = _getbytes($data_text);

        my $cksum_text  = $7;
        chomp $cksum_text;
        my $cksum_bytes = _getbytes($cksum_text);

        push @{$info->{blocks}}, {
            blockno => $blockno,

            data    => {
                disk  => $data_disk,
                dbn   => $data_dbn,
                vbn   => $data_vbn,
                text  => $data_text,
                bytes => $data_bytes

            },

            cksum   => {
                disk  => $cksum_disk,
                dbn   => $cksum_dbn,
                text  => $cksum_text,
                bytes => $cksum_bytes
            }
        };

        $blockno++;
    }

    return $info;
}


# Return array of bytes constructed from the hex dump
# of the metafile contents (or its cksum).

sub _getbytes {
    my $src   = shift;
    my $bytes = [];

    # Convert the hex dump into an array of bytes.

    foreach my $line (split /\n/, $src) {
        if ($line !~ m/:\s+
                       0x([0-9a-f]{8})\s+
                       0x([0-9a-f]{8})\s+
                       0x([0-9a-f]{8})\s+
                       0x([0-9a-f]{8})/x) {
            confess 'Pattern matching failed';
        }

        for my $byte (unpack('C*', pack('H*', $1 . $2 . $3 . $4))) {
            push @$bytes, $byte;
        }
    }

    return $bytes;
}


## @pod here

=head1 NAME

    Playlist.pm - convenience functions for Adaptive Playlist

=head1 SYNOPSIS

    playlist_create($fc, $volname);
    playlist_status($fc, $volname);
    playlist_truncate($fc, $volname);
    playlist_delete($fc, $volname);

=head1 DESCRIPTION

    The functions in this module are thin wrappers of the
    priv set test playlist commands.

=head1 FUNCTIONS

=head2 playlist_create

    Create a playlist metafile in the specified flex volume's
    containing aggregate.

    Arguments:

        fc     - filer connection
        vol    - flexvol name

    Options:

        length - size to preallocate for playlist
        zeros  - fill playlist with zeros unless this arg is non-zero

        (The "zeros" option requires the "length" option.)

=head2 playlist_truncate

    Truncate the specified flex volume's playlist metafile,
    if it exists in the containing aggregate.

    Arguments:

        fc  - filer connection
        vol - flexvol name

=head2 playlist_delete

    Delete the specified flex volume's playlist metafile,
    if it exists in the containing aggregate.

    Arguments:

        fc  - filer connection
        vol - flexvol name

=head2 playlist_status

    Return a hashref of information about the flex volume's
    playlist metafile, if it exists.

    Arguments:

        fc  - filer connection
        vol - flexvol name

=head2 playlist_dump

    Return a hashref containing the flex volume's playlist
    metafile contents, along with other info about the file.

    Arguments:

        fc  - filer connection
        vol - flexvol name

=cut


1;

