# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Perf/Priority/CLI.pm#1 $
#
## @author   dl-perf-qa@netapp.com
## @summary  NATE wrapper for ONTAP "priority" CLI commands.
## @status   private
##
## @pod here

=head1 NAME

Perf::Priority::CLI - NATE wrapper for ONTAP "priority" CLI commands

=head1 DESCRIPTION

This uses the ONTAP 7-mode CLI to provide test scripts with an
interface to "priority" commands. This interface is common among
all Perf::Priority subclasses, so that scripts can remain agnostic
of the specific type of filer interaction being used.

Rather than supporting all "priority" commands, this class implements
only those needed by the perf/flexshare/*.thpl tests.

=head1 METHODS

=head2 new

The constructor; returns Perf::Priority::CLI object.

=over

=item Arguments

host - hostname of filer being tested

=back

=cut

package Perf::Priority::CLI; {
    use Object::InsideOut qw(Perf::Priority);
    use Carp;
    use Params::Validate qw(validate SCALAR);
    use Tharn;

    my @host :Field :Std(Name => 'host', Perm => 'restricted');
    my @fc   :Field :Std(Name => 'fc',   Perm => 'restricted');
    my @fh   :Field :Std(Name => 'fh',   Perm => 'restricted');

    my %init_args :InitArgs = (
        host => {
            Mandatory => 1,
            Type      => 'scalar',
        },
    );


    sub _init :Init {
        my $self = shift;
        my $args = shift;

        my $host = $args->{host};
        my $fh   = host($host);

        if ($fh->hosttype()=~ m/-ng-/) {
            confess 'Cluster-mode CLI not supported';
        }

        my $fc = connect($host . '#rsh');

        $self->set_host($host);
        $self->set_fh($fh);
        $self->set_fc($fc);
    }


=head2 on

Wrapper for "priority on" command.

=cut

    sub on {
        my $self = shift;
        $self->_command('priority on', qr/starting|is running/);
    }


=head2 off

Wrapper for "priority off" command.

=cut

    sub off {
        my $self = shift;
        $self->_command('priority off', qr/stopped/);
    }


=head2 set

Wrapper for "priority set <option>=<value>" command.

=over

=item Arguments

option - priority option to set

value  - value to set option to

=back

=cut

    sub set {
        my $self   = shift;
        my %params = validate(@_,
                              {
                                  option => { type => SCALAR, optional => 0, },
                                  value  => { type => SCALAR, optional => 0, },
                              });

        my $option = $params{option};
        my $value  = $params{value};

        $self->_command("priority set $option=$value", qr/^$/);
    }


=head2 set_default

Wrapper for "priority set default <option>=<value>" command.

=over

=item Arguments

option - priority option to set

value  - value to set option to

=back

=cut

    sub set_default {
        my $self   = shift;
        my %params = validate(@_,
                              {
                                  option => { type => SCALAR, optional => 0, },
                                  value  => { type => SCALAR, optional => 0, },
                              });

        my $option = $params{option};
        my $value  = $params{value};

        $self->_command("priority set default $option=$value", qr/^$/);
    }


=head2 set_volume

Wrapper for "priority set volume <vol> <option>=<value>" command.

=over

=item Arguments

vol    - volume name

option - priority option to set

value  - value to set option to

=back

=cut

    sub set_volume {
        my $self   = shift;
        my %params = validate(@_,
                              {
                                  vol    => { type => SCALAR, optional => 0, },
                                  option => { type => SCALAR, optional => 0, },
                                  value  => { type => SCALAR, optional => 0, },
                              });

        my $volname = $params{vol};
        my $option  = $params{option};
        my $value   = $params{value};

        $self->_command("priority set volume $volname $option=$value", qr/^$/);
    }


=head2 delete_volume

Wrapper for "priority delete volume <vol>" command.

=over

=item Arguments

vol - volume name

=back

=cut

    sub delete_volume {
        my $self   = shift;
        my %params = validate(@_,
                              {
                                  vol => { type => SCALAR, optional => 0, },
                              });

        my $volname = $params{vol};

        $self->_command("priority delete volume $volname", qr/^$/);
    }


    # Use Tharn::say() to issue command to host, and check that
    # command output matches regex.

    sub _command :Restricted {
        my $self  = shift;
        my $cmd   = shift;
        my $regex = shift;

        my (undef, undef, $out) = say($self->get_fc(), 'priv set -q diag; ' . $cmd);

        if (defined($regex) && $out !~ m/$regex/) {
            confess "Unexpected output from '$cmd': $out";
        }
    }


    # Cleanup host connection.

    sub _destroy :Destroy {
        my $self = shift;
        disconnect($self->get_fc());
    }
}

1;

