# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Perf/Priority/ZAPI.pm#1 $
#
## @author   dl-perf-qa@netapp.com
## @summary  NATE wrapper for ONTAP "priority" ZAPI commands.
## @status   private
##
## @pod here

=head1 NAME

Perf::Priority::CLI - NATE wrapper for ONTAP "priority" ZAPI commands

=head1 DESCRIPTION

This uses the ONTAP ZAPI to provide test scripts with an
interface to "priority" commands. This interface is common among
all Perf::Priority subclasses, so that scripts can remain agnostic
of the specific type of filer interaction being used.

Rather than supporting all "priority" commands, this class implements
only those needed by the perf/flexshare/*.thpl tests.

=head1 METHODS

=head2 new

The constructor; returns Perf::Priority::ZAPI object.

=over

=item Arguments

host - hostname of filer being tested

=back

=cut

package Perf::Priority::ZAPI; {
    use Object::InsideOut qw(Perf::Priority);
    use Carp;
    use Params::Validate qw(validate SCALAR);
    use Tharn;
    use ZapiTest;

    my @host :Field :Std(Name => 'host', Perm => 'restricted');
    my @zapi :Field :Std(Name => 'zapi', Perm => 'restricted');

    my %init_args :InitArgs = (
        host => {
            Mandatory => 1,
            Type      => 'scalar',
        },
    );


    sub _init :Init {
        my $self = shift;
        my $args = shift;

        my $host = $args->{host};
        my $fc   = connect($host . '#rsh');
        my $zc   = ZapiTest->new($fc->connhost(), $fc->username(), $fc->password());

        $self->set_host($host);
        $self->set_zapi($zc);

        disconnect($fc);
    }


=head2 on

Wrapper for "priority on" command.

=cut

    sub on {
        my $self = shift;
        $self->_command(cmd => 'priority-enable');
    }


=head2 off

Wrapper for "priority off" command.

=cut

    sub off {
        my $self = shift;
        $self->_command(cmd => 'priority-disable');
    }


=head2 set

Wrapper for "priority set <option>=<value>" command.

=over

=item Arguments

option - priority option to set

value  - value to set option to

=back

=cut

    sub set {
        my $self   = shift;
        my %params = validate(@_,
                              {
                                  option => { type => SCALAR, optional => 0, },
                                  value  => { type => SCALAR, optional => 0, },
                              });

        my $option = _map_option($params{option});
        my $value  = _map_value($params{value});

        $self->_command(cmd    => 'priority-set',
                        option => $option,
                        value  => $value);
    }


=head2 set_default

Wrapper for "priority set default <option>=<value>" command.

=over

=item Arguments

option - priority option to set

value  - value to set option to

=back

=cut

    sub set_default {
        my $self   = shift;
        my %params = validate(@_,
                              {
                                  option => { type => SCALAR, optional => 0, },
                                  value  => { type => SCALAR, optional => 0, },
                              });

        my $option = _map_option($params{option});
        my $value  = _map_value($params{value});

        $self->_command(cmd    => 'priority-set-default',
                        option => $option,
                        value  => $value);
    }


=head2 set_volume

Wrapper for "priority set volume <vol> <option>=<value>" command.

=over

=item Arguments

vol    - volume name

option - priority option to set

value  - value to set option to

=back

=cut

    sub set_volume {
        my $self   = shift;
        my %params = validate(@_,
                              {
                                  vol    => { type => SCALAR, optional => 0, },
                                  option => { type => SCALAR, optional => 0, },
                                  value  => { type => SCALAR, optional => 0, },
                              });

        my $volname = $params{vol};
        my $option  = _map_option($params{option});
        my $value   = _map_value($params{value});

        $self->_command(cmd     => 'priority-set-volume',
                        option  => $option,
                        value   => $value,
                        volname => $volname);
    }


=head2 delete_volume

Wrapper for "priority delete volume <vol>" command.

=over

=item Arguments

vol - volume name

=back

=cut

    sub delete_volume {
        my $self   = shift;
        my %params = validate(@_,
                              {
                                  vol => { type => SCALAR, optional => 0, },
                              });

        my $volname = $params{vol};

        # Don't see a "priority-delete-volume" api in
        # ontap/prod/common/adminapi/zapi_priority.c.
        #
        # Is the following an appropriate substitute?

        $self->_command(cmd     => 'priority-set-volume',
                        option  => 'service',
                        value   => 'off',
                        volname => $volname);
    }


    # Convert option names from the ones used
    # by the CLI to ones expected by ZAPI.

    sub _map_option :Restricted {
        my $opt = shift;

        # The CLI uses underscores in option
        # names where ZAPI uses hyphens.

        $opt =~ y/_/-/;

        return $opt;
    }


    # Convert values from the CLI string scale
    # to the 1-10 numeric scale expected by ZAPI.

    sub _map_value :Restricted {
        my $val = shift;

        # If value is already a digit string,
        # return its numeric value.

        if ($val =~ m/^\d+$/) {
            return int($val);
        }

        # If it's a string from the "verylow"
        # to "veryhigh" scale, return its
        # corresponding numeric value.

        my $map = {
            verylow  => 1,
            low      => 3,
            medium   => 5,
            high     => 7,
            veryhigh => 10,
        };

        if (exists($map->{$val})) {
            return $map->{$val};
        }

        # Otherwise, return it unchanged.

        return $val;
    }


    # Issue ZAPI command and check results.

    sub _command :Restricted {
        my $self  = shift;
        my %params = validate(@_,
                              {
                                  cmd     => { type => SCALAR, optional => 0, },
                                  option  => { type => SCALAR, optional => 1, },
                                  value   => { type => SCALAR, optional => 1, },
                                  volname => { type => SCALAR, optional => 1, },
                              });

        my $cmd     = $params{cmd};
        my $option  = $params{option};
        my $value   = $params{value};
        my $volname = $params{volname};
        my %args    = ();

        if (defined($volname)) {
            $args{volume} = $volname;
        }

        if (defined($option)) {
            $args{$option} = $value;
        }

        my $out = $self->get_zapi()->invoke_api($cmd, %args);
        $Log->comment("ZAPI '$cmd' output:\n" . $out->sprint_pretty());

        if (!$out->results_ok()) {
            confess "ZAPI '$cmd' failed (errno = @{[$out->results_errno()]})";
        }
    }
}

1;

