# -*- perl -*-
# $Id: //depot/prod/test/main/ontap/boilermaker/lib/Perf/Reallocation.pm#5 $
#
## @author   jroussin@netapp.com
## @summary
## @status   private

##  example
##  my $realloc = new Reallocation (fc=>$fc,aggr=>$AGGR_NAME,vol=>$FILER_VOL,target_type=>vol);
##  $realloc->reallocate_on_off ("on");
##

print("Loading ".__FILE__."\n");
package Perf::Reallocation;

use strict;
use warnings;
use FS::Util::fragMeister;
use Perf::Stats;
use Perf::Utils qw (as_string setflag printflag wait_for_scan wait_for_reclamation);
use Perf::Delete_Logging;
use Admin::License;
use filer;
use Tharn;
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Module::Common qw(validate_mixed);
use Carp qw(confess);
use Aggregate qw(aggr_status);
use SnapCmds qw(snapshot_list snapshot_status snapshot_exists);
use Fcntl;
use Perf::ModeAgnostic qw(say_wrap);

our (@EXPORT_OK);
use base qw(Exporter);
@EXPORT_OK = qw(IDLE REALLOCATING REDIRECT QUIESCE CHECKING NONE SCRAMBLING DELETING
                DONE RUNNING REDIRECT_PUBLIC_INIT REDIRECT_PUBLIC_ACTIVE_INODES REDIRECT_PUBLIC_ACTIVE_INOFILE
		REDIRECT_PRIVATE_ACTIVE_INODES  REDIRECT_PRIVATE_ACTIVE_INOFILE REDIRECT_PUBLIC_SNAPSHOT_INODES
		REDIRECT_PUBLIC_SNAPSHOT_INOFILE  REDIRECT_PUBLIC_SNAPSHOT_FSINFO REDIRECT_PRIVATE_SNAPSHOT_INODES
		REDIRECT_PRIVATE_SNAPSHOT_INOFILE  REDIRECT_PRIVATE_SNAPSHOT_FSINFO );

#
# Raw status
#
use constant NONE                             => 1 ;
use constant IDLE                             => 2 ;
use constant CHECKING                         => 4 ;
use constant REALLOCATING                     => 8 ;
use constant REDIRECT_PUBLIC_INIT             => 16 ;
use constant REDIRECT_PUBLIC_ACTIVE_INODES    => 32 ;
use constant REDIRECT_PUBLIC_ACTIVE_INOFILE   => 64 ;
use constant REDIRECT_PRIVATE_ACTIVE_INODES   => 128 ;
use constant REDIRECT_PRIVATE_ACTIVE_INOFILE  => 256 ;
use constant REDIRECT_PUBLIC_SNAPSHOT_INODES  => 512 ;
use constant REDIRECT_PUBLIC_SNAPSHOT_INOFILE => 1024 ;
use constant REDIRECT_PUBLIC_SNAPSHOT_FSINFO  => 2048 ;
use constant REDIRECT_PRIVATE_SNAPSHOT_INODES => 4096 ;
use constant REDIRECT_PRIVATE_SNAPSHOT_INOFILE=> 8192 ;
use constant REDIRECT_PRIVATE_SNAPSHOT_FSINFO => 16384 ;
use constant SCRAMBLING                       => 32768 ;
use constant QUIESCE                          => 65536 ;
use constant DELETING                         => 131072 ;

#
# Frequent combination status
#
use constant DONE          => NONE |
                              IDLE;
use constant REDIRECT      => REDIRECT_PUBLIC_INIT |
                              REDIRECT_PUBLIC_ACTIVE_INODES |
                              REDIRECT_PUBLIC_ACTIVE_INOFILE |
                              REDIRECT_PUBLIC_SNAPSHOT_INODES |
                              REDIRECT_PUBLIC_SNAPSHOT_INOFILE |
                              REDIRECT_PUBLIC_SNAPSHOT_FSINFO |
                              REDIRECT_PRIVATE_ACTIVE_INODES |
                              REDIRECT_PRIVATE_ACTIVE_INOFILE |
                              REDIRECT_PRIVATE_SNAPSHOT_INODES |
                              REDIRECT_PRIVATE_SNAPSHOT_INOFILE |
                              REDIRECT_PRIVATE_SNAPSHOT_FSINFO ;
use constant RUNNING       => REALLOCATING |
                              REDIRECT |
                              QUIESCE |
                              CHECKING |
                              SCRAMBLING |
                              DELETING;
#
# Reverse lookup map
#
our %STATUS_STRING =   (
                        1     => "NONE",
                        2     => "IDLE",
                        4     => "CHECKING",
                        8     => "REALLOCATING",
                        16    => "REDIRECT_PUBLIC_INIT",
                        32    => "REDIRECT_PUBLIC_ACTIVE_INODES",
                        64    => "REDIRECT_PUBLIC_ACTIVE_INOFILE",
                        128   => "REDIRECT_PRIVATE_ACTIVE_INODES",
                        256   => "REDIRECT_PRIVATE_ACTIVE_INOFILE",
                        512   => "REDIRECT_PUBLIC_SNAPSHOT_INODES",
                        1024 => "REDIRECT_PUBLIC_SNAPSHOT_INOFILE",
                        2048 => "REDIRECT_PUBLIC_SNAPSHOT_FSINFO",
                        4096 => "REDIRECT_PRIVATE_SNAPSHOT_INODES",
                        8192 => "REDIRECT_PRIVATE_SNAPSHOT_INOFILE",
                        16384 => "REDIRECT_PRIVATE_SNAPSHOT_FSINFO",
                        32768 => "SCRAMBLING",
                        65536 => "QUIESCE",
                        131072 => "DELETING",
                        );

our %COMBO_STATUS_STRING = (32752=> "REDIRECT",
                            3   => "DONE",
                            262140=> "RUNNING",
                            );

our %REDIRECT_PHASES = (
                        'public' => {   0 => REDIRECT_PUBLIC_INIT,
                                        1 => REDIRECT_PUBLIC_ACTIVE_INODES,
                                        2 => REDIRECT_PUBLIC_ACTIVE_INOFILE,
                                        3 => REDIRECT_PUBLIC_SNAPSHOT_INODES,
                                        4 => REDIRECT_PUBLIC_SNAPSHOT_INOFILE,
                                        5 => REDIRECT_PUBLIC_SNAPSHOT_FSINFO},
                        'private' =>{   1 => REDIRECT_PRIVATE_ACTIVE_INODES,
                                        2 => REDIRECT_PRIVATE_ACTIVE_INOFILE,
                                        3 => REDIRECT_PRIVATE_SNAPSHOT_INODES,
                                        4 => REDIRECT_PRIVATE_SNAPSHOT_INOFILE,
                                        5 => REDIRECT_PRIVATE_SNAPSHOT_FSINFO},
                        );

use constant TARGET_AGGR => 1;
use constant TARGET_VOL  => 2;
use constant TARGET_FILE => 3;
use constant RUNNING_SCAN_BLKREALLOC => 10;
use constant RUNNING_SCAN_REDIRECT   => 11;
use constant RUNNING_REALLOCATE      => 12;
use constant RUNNING_SCRAMBLE        => 13;
use constant RUNNING_NONE            => 14;

sub new {
    my $class = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {
                fc => {type => OBJECT},
                debug => {type => SCALAR | UNDEF, optional => 1, default=>0},
                aggr => {type => SCALAR | UNDEF, optional => 0},
                vol => {type => SCALAR | UNDEF, optional => 1},
                #example /vol/myvol/myfile
                files => {type => ARRAYREF | UNDEF, optional => 1},
                # target type of reallocate ui cmd - aggr, vol or file.
                #Default is defined object, or vol if more than one is defined
                target_type => {type => SCALAR | UNDEF, optional => 1}
           }
           );
    my $self = \%params;

    # Automatically detect target_type
    if (!$self->{target_type}){
        if($self->{files}){
            $self->{target_type} = TARGET_FILE;
        }
        elsif($self->{vol}){
            $self->{target_type} = TARGET_VOL;
        }
        elsif($self->{aggr}){
            $self->{target_type} = TARGET_AGGR;
        }
        else{
            confess "You must specify a 'aggr','vol', or 'file' argument";
        }
    }
    else {
        my $type = $self->{target_type};
        if ($type eq 'file'){
            $self->{target_type} = TARGET_FILE;
        }
        elsif($type eq 'vol'){
            $self->{target_type} = TARGET_VOL;
        }
        elsif($type eq 'aggr'){
            $self->{target_type} = TARGET_AGGR;
        }
        else{
            confess "Argument target_type must be one of 'aggr','vol','file' instead of '$self->{target_type}'";
        }
    }

    if (!$self->{fc}){
        confess "You must specify a filer connection";
    }
    else{
        $self->{filer} = $self->{fc}->hostp->hostname;
    }

    # Check arguments
    if ($self->{target_type} eq TARGET_FILE) {
        if(!$self->{files}) {
            confess "if file is target of reallocate command, you must specify files=>[full path to file]";
        }
        for my $file (@{$self->{files}}){
            if ($file !~ /\/vol/){
                confess "File '$file' isn't a full path of the filer";
            }
        }
    }
    elsif ($self->{target_type} eq TARGET_VOL && !$self->{vol}) {
        confess "if vol is target of reallocate command, you must specify vol=> full path to file";
    }
    elsif ($self->{target_type} eq TARGET_AGGR && !$self->{aggr}) {
        confess "You must specify an 'aggr'";
    }

    # Set the target
    if ($self->{target_type} eq TARGET_VOL) {
        $self->{target} = "/vol/$self->{vol}";
    }
    elsif ($self->{target_type} eq TARGET_FILE) {
        if (@{$self->{files}} == 1){
            $self->{target} = $self->{files}->[0];
        }
        else{
            $self->{target} = $self->{files};
        }
    }
    else {
        $self->{target} = $self->{aggr};
    }

    # Keep track of what is running if anything, default is nothing
    $self->{running} = RUNNING_NONE;


    $self->{verbose} = "-noverbose";
    if ($self->{debug}){
        $self->{verbose} = "-verbose";
    }

    # Keep track of various flags
    $self->{initial_wafl_blkr_segment_size} = printflag($self->{fc},'wafl_blkr_segment_size');
    $self->{initial_wafl_blkr_policy}       = printflag($self->{fc},'wafl_blkr_policy');
    $self->{initial_wafl_blkr_min}          = printflag($self->{fc},'wafl_blkr_min');
    $self->{initial_wafl_blkr_max}          = printflag($self->{fc},'wafl_blkr_max');
    $self->{initial_wafl_blkr_fix_pvbn}     = printflag($self->{fc},'wafl_blkr_fix_pvbn');

    $self->{pid} = $$;

    bless $self, $class;
    return $self;
}

sub get_current_stats {
       my $self = shift;
       my $aggr = $self->{aggr};

       if (!$aggr){
            confess "You need an 'aggr'";
        }
        if (!$self->{Stats}){
            $self->{Stats} = new Perf::Stats(fc => $self->{fc},debug=>$self->{debug},
                                         counters =>
                                                ["aggregate:$aggr:blkr_.*",
                                                 "aggregate:$aggr:wv_fsinfo_blkr_cp"]);
            $self->{Stats}->alias(name=>"blocks_redirected",
                                  counters=>["aggregate:$aggr:blkr_blocks_redirected.*"],
                                  hide_counters=>0,
                                  action=>'sum');
        }
        return $self->{Stats}->sample();
}

sub get_aggr {
    my $self = shift;
    return $self->{aggr};
}

sub get_current_layout {
    my $self = shift;
    my %params = @_;
    my $timeout = $params{timeout}; #In secs
    if (!$timeout){
        $timeout = 900;
    }
    my $hostrec = $self->{fc}->hostp();
    my $filer = $hostrec->hostname();

    if (!$self->{aggr} || !$self->{vol}){
        confess "You must specify the 'aggr' and 'vol'";
    }

    my $file = undef;
    if ($self->{target_type} eq TARGET_FILE) {
           $file = $self->{target};
    }

    local $main::CLIENT = 'localhost';         # fragMeister kludge

    my $fm = fragMeister->new($filer, $self->{aggr});
    my %layout = $fm->measure_layout_wait($self->{vol}, timeout=>$timeout, client=>"localhost", file=>$file);
        #$layout{ratio}
        #$layout{optimal}
        #$layout{expected}
        #$layout{message}
    return \%layout;

}

sub data {
    my $self = shift;

    if ($self->{running} eq RUNNING_NONE){
        confess "No data available until you start a reallocation";
    }

    if ($self->{running} eq RUNNING_REALLOCATE && $self->{target_type} ne TARGET_AGGR ){
        # Get the layout ratio
        my $layout = $self->get_current_layout();

        # How do you compute the difference between the old and the new?
        return {old_layout => $self->{data}->{inital_layout},
                new_layout => $layout,
                delta => "TODO"
                };
    }
    elsif($self->{running} eq RUNNING_SCRAMBLE){
        if ($self->{scramble_subtest}->is_alive()){
            $self->wait_for_status( DONE );
        }
        my @subtest_results = $self->{scramble_subtest}->own_results();
        my @subtest_msgs = $subtest_results[0]->msgs();
        my $scrambled_blocks = $subtest_msgs[0];
        delete $self->{scramble_subtest};
        return {"scrambled_blocks" => $scrambled_blocks};
    }
    else{
        # Get new stats sample
        my $s = $self->get_current_stats();

        # Get the stats delta since the last 'start'
        my $d = $self->{Stats}->delta($s, $self->{data}->{initial_stats});
        return {old_stats => $self->{data}->{initial_stats},
                new_stats => $s,
                delta => $d};
    }
}

sub status_string {
    my $self = shift;
    my $status = shift;

    my $status_string = $STATUS_STRING{$status};
    if ($status_string){
        return $status_string;
    }
    $status_string = $COMBO_STATUS_STRING{$status};
    if ($status_string){
        return $status_string;
    }

    for my $d (keys %STATUS_STRING){
        if ($d & $status){
            if ($status_string){
                $status_string .= " | ";
            }
            $status_string .= $STATUS_STRING{$d};
        }
    }
    if (!$status_string){
        confess "Unable to reverse lookup status value $status";
    }
    return $status_string;
}

sub clear_deleted_blocks {
    my $self = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {
                interval => {type => SCALAR | UNDEF, optional => 1},
           }
           );
    my $fc = $self->{fc};
    my $aggr = $self->{aggr};

    if (!$aggr){
        confess "You must specify an aggregate";
    }

    #
    # First make sure the delete logging files are empty
    #
    my $delete_logging_was_on = delete_logging_is_on($fc, $aggr);
    if ($delete_logging_was_on){
        delete_logging_off($fc,$aggr);
        delete_logging_info($fc,$aggr);
    }

    #
    # Run block reclamation on all volumes in aggregate
    #

    # Flag below makes sure blk_reclaim scanner runs quickly
    my $old_flag_value = printflag($fc,'wafl_blk_reclaim_paced');
    setflag($fc,'wafl_blk_reclaim_paced',0);

    my $status_obj = aggr_status($fc, $aggr);
    my $vols_list = $status_obj->volumes();

    for my $vol (@$vols_list){
        my (undef, undef, $out) = say_wrap($fc, "priv set -q diag; wafl scan blk_reclaim $vol -f");
        if ($out){
            confess "Unable to start blk_reclaim scanner: $out";
        }
    }
    for my $vol (@$vols_list){
        wait_for_scan($fc, type => 'vol',
                       target => $vol,
                       no_status => 'container block reclamation',
                       interval => $params{interval},
                 );
    }
    setflag($fc,'wafl_blk_reclaim_paced',$old_flag_value);

    if ($delete_logging_was_on){
        delete_logging_on($fc, $aggr);
        delete_logging_info($fc,$aggr);
    }
}

sub start {
    my $self = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {
                once => {type => SCALAR | UNDEF, optional => 1},
                interval => {type => SCALAR | UNDEF, optional => 1},
                threshold => {type => SCALAR | UNDEF, optional => 1},
                no_threshold_check => {type => SCALAR | UNDEF, optional => 1},
                force => {type => SCALAR | UNDEF, optional => 1},  #one-time full reallocation
                #snapshots => {type => SCALAR | UNDEF, optional => 1}, # DEPRECATED. Replaced by physical_only
                physical_only => {type => SCALAR | UNDEF, optional => 1},
                scanner => {type => SCALAR | UNDEF, optional => 1},
                scramble => {type => SCALAR | UNDEF, optional => 1},
                scramble_invalidate_disk => {type => SCALAR | UNDEF, optional => 1, default=>0},
                scramble_invalidate_ec => {type => SCALAR | UNDEF, optional => 1, default=>0},
                scramble_processes => {type => SCALAR | UNDEF, optional => 1, default=>4},
                timeout => {type => SCALAR | UNDEF, optional => 1, default => 5},
                no_block_reclamation => {type => SCALAR | UNDEF, optional => 1, default => 0},
                redirect_vol => {type => SCALAR | UNDEF, optional => 1},
           }
           );

    # Check arguments
    if ($params{scanner} && $params{scanner} !~ /(blkrealloc|redirect)/){
        confess "Argument 'scanner' must be 'blkrealloc' or 'redirect'";
    }
    if ($params{scanner} && $params{scramble}){
        confess "You must choose between starting scramble or a scanner";
    }

    # Scramble -e does not support multiple processes
    if ($params{scramble} && $params{scramble_invalidate_ec}){
        logcomment("Setting number of scramble process to 1");
        $params{scramble_processes} = 1;
    }

    my $fc = $self->{fc};
    my $aggr = $self->{aggr};
    my $target = $self->{target};

    # Starting a scanner
    if ($params{scanner}){

        $self->{data}->{initial_stats} = $self->get_current_stats();

        if ($params{scanner} =~ /blkrealloc/){
            #
            # In order for segment cleaning to see as many non-full tretrises as possible,
            # we need to free all blocks waiting to get freed. That means deleting the
            # delete log file and running the block reclamation scanner
            #
            if (!$params{no_block_reclamation}){
                $self->clear_deleted_blocks(interval => $params{interval});
            }

            # Start scanner
            my $out = say_wrap($fc,"priv set -q diag;wafl scan blkrealloc $aggr; wafl scan status -A $aggr");
            $self->{running} = RUNNING_SCAN_BLKREALLOC;
            #$self->{target_type} = TARGET_AGGR;
            # Make sure it's started
            if (!$self->is_status(REALLOCATING, out=>$out)){
                $self->wait_for_status(REALLOCATING, timeout=>$params{timeout});
            }
        }
        elsif ($params{scanner} =~ /redirect/){
            # Start scanner
            $self->{running} = RUNNING_SCAN_REDIRECT;
            if ($self->{target_type} eq TARGET_VOL){
                if ($params{redirect_vol}){
                    $self->{vol} = $params{redirect_vol};
                }
                my $out = say_wrap($fc,"priv set -q diag;wafl scan redirect $self->{vol}; wafl scan status $self->{vol}");
                # Make sure it's started
                if (!$self->is_status(REDIRECT, out=>$out)){
                    $self->wait_for_status(REDIRECT, timeout=>$params{timeout}, interval=>0);
                }
            }
            elsif ($self->{target_type} eq TARGET_AGGR){
                if ($params{redirect_vol}){
                    my $vol = $params{redirect_vol};
                    $self->{target_type} = TARGET_VOL;
                    $self->{vol} = $vol;
                    my $out = say_wrap($fc,"priv set -q diag;wafl scan redirect $vol; wafl scan status $vol");
                    # Make sure it's started
                    if (!$self->is_status(REDIRECT, out=>$out)){
                        $self->wait_for_status(REDIRECT, timeout=>$params{timeout},
                                                         interval=>0,
                                                         );
                    }
                }
                else{
                    # Start the scanner on every volume of the aggregate at the same time
                    my $status_obj = aggr_status($fc, $self->{aggr});
                    my $vols_list = $status_obj->volumes();
                    foreach my $vol ( @{$vols_list} ) {
                        my $out = say_wrap($fc,"priv set -q diag;wafl scan redirect $vol; wafl scan status $vol");
                        # Make sure it's started
                        if (!$self->is_status(REDIRECT, out=>$out, vol=>$vol)){
                            $self->wait_for_status(REDIRECT, timeout=>$params{timeout},
                                                             interval=>0,
                                                             vol=>$vol,
                                                             );
                        }
                    }
                }
            }
            else{
                confess "The current target_type does not " .
                        "allow you to run the redirect scanner";
            }
        }
    }
    elsif ($params{scramble}){
        $self->{running} = RUNNING_SCRAMBLE;
        my $filer = $self->{fc}->hostp()->hostname;
        $self->{scramble_subtest} = subtest('{use Perf::Reallocation;Perf::Reallocation::_scramble(@_)}',
                                       '-bg', '-runid',"scramble",'--','filer',$filer,
                                       'fc',$self->{fc},
                                       'files',$self->{files},
                                       'processes',$params{scramble_processes},
                                       'invalidate_disk',$params{scramble_invalidate_disk},
                                       'invalidate_ec',$params{scramble_invalidate_ec});
    }
    else {   #default to reallocate cmd

        # Should you get an initial measure of layout here?
        #$self->{data}->{inital_layout} = $self->get_current_layout();

        # Build string of optional parameters
        my $opts = $params{physical_only}?"-p ":"";
        $opts .= $params{interval}?"-i $params{interval} ":"";
        $opts .= $params{threshold}?"-t $params{threshold} ":"";
        $opts .= $params{force}?"-f ":"";
        $opts .= $params{once}?"-o ":"" ;
        $opts .= $params{no_threshold_check}?"-n ":"";

        my $cmd = "reallocate start $opts $target; reallocate status $target";

        #if we're file with many files, skip to the else
        unless ( $self->{target_type} eq TARGET_FILE  &&  (@{$self->{files}} != 1) ) {
            if ($self->{target_type} eq TARGET_AGGR) {
                #
                # In order for segment cleaning to see as many non-full tretrises as possible,
                # we need to free all blocks waiting to get freed. That means deleting the
                # delete log file and running the block reclamation scanner
                #
                if (!$params{no_block_reclamation}){
                    $self->clear_deleted_blocks(interval => $params{interval});
                }

                $self->{data}->{initial_stats} = $self->get_current_stats();

                $cmd = "reallocate start -A $opts $target; reallocate status $target";

                my $out = say_wrap($fc, $cmd);
                $self->{running} = RUNNING_REALLOCATE;

                if (!$self->is_status(REALLOCATING, out=>$out)){
                    while ($out =~ /$aggr has block reallocation/){
                        logcomment("WARN","blkrealloc still finishing up, can't start via reallocate ui, sleeping");
                        sleep 5;
                        $out = say_wrap($fc, $cmd);
                    }
                    $self->wait_for_status(REALLOCATING, timeout=>$params{timeout});
                }
            }
            else {   #target_type vol or file
                my $out = say_wrap($fc, $cmd);
                $self->{running} = RUNNING_REALLOCATE;
                if (!$self->is_status(REALLOCATING | CHECKING | DELETING, out=>$out)){
                    $self->wait_for_status(REALLOCATING | CHECKING | DELETING,
                                           timeout=>$params{timeout});
                }
            }
        }
        else {
            #foreach file, call scanner
            my $files = $self->{files};
            foreach my $filepath (@$files)   {
                $self->{target} = $filepath;
                $cmd = "reallocate start $opts $filepath; reallocate status $filepath";
                my $out = say_wrap($fc, $cmd);
                $self->{running} = RUNNING_REALLOCATE;
                if (!$self->is_status(REALLOCATING | CHECKING | DELETING, out=>$out)){
                    $self->wait_for_status(REALLOCATING | CHECKING | DELETING,
                                           timeout=>$params{timeout});
                }
            }
        }
    }
}

sub overwrite_old_blocks {
    my $self = shift;
    my %args = @_;
    my $vol_snap_list = $args{vol_snap_list};   # list of snapshots in each volume in the aggr.
    # eg: {
    #     vol1 => ['snap1', 'snap2'];
    #     vol2 => [];         empty array means overwrite all snapshots in this volume.
    #    }
    # If this param is not specified, all the snapshots in all the volumes in the aggr will be done.

    my $aggr = $self->{aggr};
    if (!$aggr){
        confess "You must specify an aggregate";
    }

    $self->clear_deleted_blocks();

    my $fc = $self->{fc};

    my $cmd = "priv set -q test; scramble overwrite";

    # If $vol_snap_list is defined, do only for those volumes, else do for all volumes in the aggr.
    my $vols_list;
    if (defined $vol_snap_list) {
       my @vols = keys (%$vol_snap_list);
       $vols_list = \@vols;
    }
    else {
       my $status_obj = aggr_status($fc, $aggr);
       $vols_list = $status_obj->volumes();
    }

    my $blocks_overwritten = 0;
    for my $vol (@$vols_list){

        #
        # In order to prevent a situation where scramble overwrite never
        # completes, start it on the active filesystem and the snapshots
        # present at the beginning of the run
        #

        # If snapshots are specified for this volume, do only for those snapshots, else do for all snapshots in this volume.
        my %snapids;
        my @snaps = snapshot_status($fc, $vol);

        if (defined $vol_snap_list) {
          my $snap_names = $vol_snap_list->{$vol};
          if (@$snap_names) {
             foreach my $snap(@snaps) {
                if (grep {$_ eq $snap->{name}}@$snap_names) {
                   $snapids{$snap->{snapid}} = $snap->{name};
                }
             }
          }
          else {
             %snapids = map { $_->{snapid} => $_->{name} } @snaps;
          }
        }
        else {
           %snapids = map { $_->{snapid} => $_->{name} } @snaps;
        }
        $snapids{0} = 'activefs';

        for my $snapid ( keys %snapids) {

            # Make sure the snapshot still exists
            if ($snapid ne 0 && !snapshot_exists($fc, $vol, $snapids{$snapid})){
                next;
            }

            #say($fc, "priv set test; statit -b");
            my (undef, undef, $out) = say_wrap($fc, $cmd . " -S $snapid $vol", '-timeout', 50400);
            if ($out =~ /Overwrote (\d+) total blocks/){
                $blocks_overwritten += $1;
            }
            else{
                confess "Unable to run scramble overwrite on $vol: $out";
            }
            #say_wrap($fc, "priv set test; statit -e");
        }
    }
    return $blocks_overwritten;
}

sub stop ($) {
    my $self = shift;
    my %params = validate_mixed(
           @_,
           [ ],
           {
                timeout => {type => SCALAR | UNDEF, optional => 1, default => 20},
           }
           );

    my $target = $self->{target};
    my $fc = $self->{fc};

    if ( $self->{target_type} eq TARGET_FILE  &&  (@{$self->{files}} > 1) ) {
            my $files = $self->{files};
            foreach my $filepath (@$files)   {
                my $cmd = "reallocate stop $filepath; reallocate status $filepath";
                my $out = say_wrap($fc, $cmd);
                $self->{target} = $filepath;
	    	if (!$self->is_status(NONE, out=>$out)) {
	       		$self->wait_for_status(NONE, timeout=>$params{timeout}, interval=>2);
		}
            }
	    $self->{running} = RUNNING_NONE;
    }
    else {
	    my (undef, undef, $out) = say_wrap($fc, " reallocate stop $target; reallocate status $target");
	    if (!$self->is_status(NONE, out=>$out)) {
	       $self->wait_for_status(NONE, timeout=>10, interval=>2);
	    }
    }
}

sub quiesce {
    my $self = shift;
    my $target = $self->{target};
    my $fc = $self->{fc};

    my (undef, undef, $out) = say_wrap($fc, " reallocate quiesce $target; reallocate status $target");
    if (!$self->is_status(QUIESCE, out=>$out)) {
        $self->wait_for_status(QUIESCE, timeout=>5, interval=>1);
    }
}

sub is_status {
    my $self = shift;
    my $status = shift;
    my %params = @_;

    if ($self->status(%params) & $status){
        return 1;
    }
    else{
        return 0;
    }
}

sub status {
    my $self = shift;
    my %params = @_;
    my $out = $params{out};
    my $param_vol = $params{vol}; # Used internally for redirect scanner status
    my $fc = $self->{fc};

    if ($self->{running} eq RUNNING_REALLOCATE){
        if (!$out){
            #say_wrap($fc, "priv set -q diag; wafl scan status -V; wafl scan status -A");
            (undef, undef, $out) = say_wrap($fc, " reallocate status -v $self->{target}");
        }
        return _status_reallocate(\$out);
    }
    elsif ($self->{running} eq RUNNING_SCAN_BLKREALLOC){
        if (!$out){
            (undef, undef, $out) = say_wrap($self->{fc},"priv set -q diag; wafl scan status -A $self->{aggr}");
        }
        return _status_blkrealloc(\$out);
    }
   elsif ($self->{running} eq RUNNING_SCAN_REDIRECT){
        if (!$out){
            (undef, undef, $out) = say_wrap($self->{fc},"priv set -q diag; wafl scan status -V");
        }
        if ($self->{target_type} eq TARGET_VOL){
            return _status_redirect(\$out, $self->{vol});
        }
        elsif ($self->{target_type} eq TARGET_AGGR){
            my $status_obj = aggr_status($fc, $self->{aggr});
            my $vols = $status_obj->volumes();
            if (!$vols || !@$vols){
                return NONE;
            }
            my %vol_status = map { $_ => _status_redirect(\$out, $_) } @$vols;
            # Return the status with the highest value
            my $status = 0;
            map { $status = $vol_status{$_} if ($vol_status{$_} > $status) } @$vols;
            return $status;
        }
    }
    elsif ($self->{running} eq RUNNING_SCRAMBLE){
        if ($self->{scramble_subtest}->is_alive()){
            return SCRAMBLING;
        }
        else{
            return NONE;
        }
    }
    # In case you have never started this reallocation return NONE
    else{
        return NONE;
    }
}

#
# TODO: Can this be fitted inside status?
#
sub multi_files_status {
    my $self = shift;
    my $status = shift;
    my %params = @_;
    my $timeout = $params{timeout}; #In secs
    if (!$timeout){
        $timeout = 300;
    }
    my $retval;
    my $files = $self->{files};
    foreach my $filepath (@$files)   {
        $retval = 0;
        $self->{target} = $filepath;
        if ($self->status() & $status){
            $retval = 1;
        }
        if ( $params{wait} && !$retval ) {
            $self->wait_for_status($status, timeout=>$params{timeout});
            $retval = 1;
        }
        if (!$retval) {  #if the above returns false, exit and return it.
            last;
        }
    }
    return $retval;
}

sub segment_size {
    my $self = shift;
    my $size = shift;

    if (defined($size)){
        setflag($self->{fc},'wafl_blkr_segment_size',$size);
        return;
    }
    else{
        return printflag($self->{fc},'wafl_blkr_segment_size');
    }
}

sub schedule($$){

    my $self = shift;
    my $minute = shift;
    my $target = $self->{target};
    my $fc = $self->{fc};


    # default * * * * run every minute of every hour...

    my $cron;
    if ($minute) {
        $cron = "$minute * * *";
    } else {
        $cron = "* * * *";
    }
    my (undef, undef, $out) = say_wrap($fc, " reallocate schedule -s '$cron' $target; reallocate status $target" );
    if ($out =~ /Schedule: n/) {    #Schedule: n/a
        confess("reallocate sched cmd failed");
    }
}

# TODO: Test this
sub restart {
    my $self = shift;
    my $target = $self->{target};
    my $fc = $self->{fc};

    my (undef, undef, $out) = say_wrap($fc, " reallocate restart $target; reallocate status $target");

    if ($self->{target_type} eq TARGET_AGGR) {
        if (!$self->is_status( REALLOCATING | REDIRECT, out => $out)){
            $self->wait_for_status(REALLOCATING | REDIRECT, timeout=>5);
        }
    } else {
        print "YO!  ensure it restarts\n";
    }
}

sub toggle($$) {
    my $self = shift;
    my $cmd = shift;
    if ($cmd !~ /on|off/){
        confess "Argument must be on|off";
    }
    my $fc = $self->{fc};
    my (undef, undef, $out) = say_wrap($fc, " reallocate $cmd");
    if ($cmd =~ /on/ && $out !~ /(scans are on|scans will be enabled)/){
        confess "Unable to switch reallocate on: $out";
    }
    if ($cmd =~ /off/ && $out !~ /(will be disabled|Reallocation scans are off)/){
        confess "Unable to switch reallocate off: $out";
    }
}

sub on_the_fly {
    my $self = shift;
    my $arg = shift;

    if ($arg){
        if ($arg =~ /^(0|off)$/i){
            setflag($self->{fc},'wafl_blkr_fix_pvbn',0);
        }
        elsif ($arg =~ /^(1|on)$/i){
            setflag($self->{fc},'wafl_blkr_fix_pvbn',1);
        }
        else{
            confess "Argument must be 0, 1, on or off";
        }
        return;
    }
    else{
        return printflag($self->{fc},'wafl_blkr_fix_pvbn');
    }
}

sub wait_for_status($$) {
    my $self = shift;
    my $status = shift;
    my %params = @_;
    my $timeout = delete $params{timeout}; #In secs
    if (!defined($timeout)){
        $timeout = 300;
    }
    my $interval = delete $params{interval}; #In secs
    if (!defined($interval)){
        $interval = 1;
    }
    # Wait to see the status $status until
    # timeout
    # If strictly is set to 0 and we're waiting for
    # one of the REDIRECT statuses, this subroutine
    # returns once we see a status >= to the one asked
    # This is useful when waiting for statuses that are
    # hard to detect
    my $strictly = delete $params{strictly};
    if (!defined($strictly)){
        $strictly = 1;
    }

    my $start_time = time;
    while (1) {
        my $current_status = $self->status(%params);
        if ($current_status & $status){
            last;
        }
        # The redirect scanner goes through phases (1..5)
        # We may miss the phase we're after and see a status of
        # phase + n. Just return if that happens
        if ( !$strictly &&  (REDIRECT & $status) && (REDIRECT & $current_status)
                        && $current_status > $status){
            last;
        }

        if ($timeout && time > $start_time + $timeout) {
            confess "Timeout while waiting for status '" . $self->status_string($status) ."'";
        }
        if ($interval){
            sleep $interval;
        }
    }
}

sub _mount_volume {
    my $self = shift;
    if (!$self->{mountpoint}){
        require GenericMount;

        Admin::License::license( $self->{fc}, "nfs" ) or confess "Unable to licence nfs: $@";
        #my (undef, undef, $reply) = say_wrap( $self->{fc}, "nfs on");
        #if( $reply !~ /NFS server is running/ ) {
        #    confess("Unable to bring NFS server up using 'nfs on' command.\nERROR: $reply");
        #}
        filer_nfs_start($self->{fc});
        ExportFS::export_path( $self->{fc}, "/vol/vol0/etc", OPTIONS => "anon=0,nosuid,rw" );
        my $mountpoint_path = GenericMount::mount($self->{filer},"/vol/vol0/etc",chmod => "0777",
                                                    mountretry=>0,sharing=>"shared");
        confess $@ if $@;
        logcomment("Mounted '$self->{filer}:/vol/vol0' to '$mountpoint_path'");

        $self->{mountpoint} = $mountpoint_path;
    }
    else{
        logcomment("Volume '$self->{filer}:/vol/vol0' is already mounted to '$self->{mountpoint}'");
    }
    return;
}

sub _scramble_dumpblocks {
    my $self = shift;
    my %args = @_;

    my $activefs_only = $args{activefs_only};
    my $verbose = $args{verbose};
    my $mismatches_only = $args{mismatches_only};
    my $filter = $args{filter};
    my $hide_pvbn_unknown = $args{hide_pvbn_unknown};
    my $offline_vol = $args{offline_vol};
    my $flush_cache = $args{flush_cache};
    my $timeout = $args{timeout};
    if (!$timeout){
        $timeout = -1; #infinite timeout
    }

    my $vol = $args{vol};
    if (!$vol){
        if ($self->{vol}){
            $args{vol} = $self->{vol};
        }
        else{
            confess "You need to pass a vol argument";
        }
    }
    if (!defined($mismatches_only)){
        $mismatches_only = 1;
    }
    if (!defined($offline_vol)){
        $offline_vol = 1;
    }

    #
    # Make sure all blocks are on disk
    # Found that this may fail the first time due to timeouts so wrapped
    # in eval and retries
    #
    eval{filer_sync($self->{fc});};
    my $error = $@;
    while($error) {
        logcomment("sync failed with: $error, wait 60 seconds and try again");
        Tharn::snooze(60);
        eval{filer_sync($self->{fc});};
        $error = $@;
    }
    filer_sync($self->{fc});

    #
    # Flush the buffer cache
    #
    if ($flush_cache){
        say_wrap($self->{fc}, "priv set -q diag; wafl flush cache");
    }

    if ($offline_vol){
        filer_vol_offline($self->{fc},$vol);
        filer_vol_online($self->{fc},$vol);
    }

    my $cmd = "priv set -q test; scramble dumpblocks";
    if ($verbose){
        $cmd .= " -v";
    }
    if ($mismatches_only){
        $cmd .= " -F";
    }
    if ($hide_pvbn_unknown){
        $cmd .= " -u";
    }

    # Find all the snapshots currently in the volume
    # Only run on those
    my @snaps = snapshot_status($self->{fc}, $vol);
    my %snapids = map { $_->{snapid} => $_->{name} } @snaps;
    $snapids{0} = 'activefs';

    if ($activefs_only){
        %snapids = (0 => 'activefs');
    }

    my $dumpblock_file = "WAFL_dumpblocks.$vol";
    my $filer_filepath = "/vol/vol0/etc/$dumpblock_file";

    for my $snapid ( keys %snapids){
        # Make sure the snapshot still exists
        if ($snapid ne 0 && !snapshot_exists($self->{fc}, $vol, $snapids{$snapid})){
            next;
        }
        my $cmdstr = "$cmd -S $snapid $vol";
        my (undef, undef,$out) = say_wrap($self->{fc}, $cmdstr, '-timeout', $timeout,
                                        '-maxidle', 300, $self->{verbose});
        if ($out =~ /usage/){
            confess "Scramble dumpblocks not supported";
        }
        elsif ($out !~ /output being saved to/){
            confess "Unable to get mismatched blocks: $out";
        }
        # Rename the file so it won't get overwritten by the next iteration
        (undef, undef, $out) = say_wrap($self->{fc}, "mv $filer_filepath $filer_filepath.$snapid");
        if ($out){
            confess "Unable to move file '$filer_filepath':\n$out";
        }
    }

    #
    # Now merge all the files into one
    #
    if (!$self->{mountpoint}){
        $self->_mount_volume();
    }

    my $mount_filepath = "$self->{mountpoint}/$dumpblock_file";
    if (-f $mount_filepath){
        confess "File $mount_filepath already exists!";
    }
    my $fh = new FileHandle($mount_filepath, O_CREAT | O_WRONLY)
        or confess "Unable to open file $mount_filepath: $!";
    my $need_header = 1;

    for my $snapid ( keys %snapids){
        my $snapid_file = "$mount_filepath.$snapid";
        my $snapid_fh;
        if (-f $snapid_file){
            $snapid_fh = new FileHandle($snapid_file)
                or confess "Unable to open $snapid_file: $!";
        }
        else{
            confess "File '$snapid_file' not found";
        }
        my $is_header = 1;
        while (<$snapid_fh>){
            my $line = $_;
            if ($is_header && $line =~ /^(#|[a-z])/){
                # Line belongs to the header
                if ($need_header){
                    $fh->print($line);
                }
            }
            else{
                $is_header = 0;
                $need_header = 0;
                $fh->print($line);
            }
        }
        $snapid_fh->close;
        unlink  $snapid_file or confess "Unable to delete '$snapid_file: $!";
        $need_header = 0;
    }
    $fh->close;
    return 1;
}

sub get_vol_segments_data {
    my $self = shift;
    my %args = @_;

    my $vol = $args{vol};
    if (!$vol){
        if ($self->{vol}){
            $args{vol} = $self->{vol};
        }
        else{
            confess "You need to pass a vol argument";
        }
    }

    my $run_scramble = $args{run_scramble};
    if (!defined($run_scramble)){
        $run_scramble = 1;
    }
    #
    # Run scramble dumpblocks
    #
    delete $args{verbose};
    delete $args{mismatches_only};
    if ($run_scramble){
        $self->_scramble_dumpblocks(verbose=>1,mismatches_only=>0,%args);
    }

    my $processor = sub {
                        my %args = @_;
                        my $context = $args{context};
                        my $row = $args{row};

                        my $wrg = $row->{wrg};
                        my $sno = $row->{sno};
                        my $segment = $context->{segments}->{$wrg}->{$sno};
                        if ($segment && ($segment->{seg_total} != $row->{seg_total} ||
                            $segment->{seg_free} != $row->{seg_free})){
                                logcomment("Segment $row->{sno} on rg $row->{wrg} used to be "
                                           . "($segment->{seg_total}, $segment->{seg_free}) but "
                                           . "is now ($row->{seg_total}, $row->{seg_free})");
                                $context->{segments_changed}->{$wrg}->{$sno} ++;
                                if (!$segment->{seg_free}){
                                    $context->{segments_full} --;
                                }
                                $context->{used_blocks} -= $segment->{seg_total} - $segment->{seg_free};
                                $context->{segments_count} --;
                                delete $context->{segments}->{$wrg}->{$sno};
                        }
                        if (!$segment){
                            $context->{segments_count} ++;
                            if (!$row->{seg_free}){
                                $context->{segments_full} ++;
                            }
                            $context->{segments}->{$wrg}->{$sno}->{seg_total} = $row->{seg_total};
                            $context->{segments}->{$wrg}->{$sno}->{seg_free} = $row->{seg_free};
                            $context->{used_blocks} += $row->{seg_total} - $row->{seg_free};
                        }

                        return 1;
                    };

    my $context = $self->_parse_dumpblocks_file(processor=>$processor, %args);
    return $context;

}

sub get_mismatched_blocks {
    my $self = shift;
    my %args = @_;

    my $count_only = $args{count_only};
    delete $args{count_only};
    my $summary_only = $args{summary_only};
    delete $args{summary_only};
    if ($count_only && $summary_only){
        confess "You can only specify one of (count_only,summary_only)";
    }
    my $run_scramble = $args{run_scramble};
    delete $args{run_scramble};
    my $filter = $args{filter};
    delete $args{filter};
    my $max_mismatches = $args{max_mismatches};
    delete $args{max_mismatches};
    my $vol = $args{vol};
    if (!$vol){
        if ($self->{vol}){
            $args{vol} = $self->{vol};
        }
        else{
            confess "You need to pass a vol argument";
        }
    }
    if (!defined($max_mismatches)){
        $max_mismatches = 100000;
    }
    if (!defined($count_only) && !$summary_only){
        $count_only = 1;
    }
    if (!defined($run_scramble)){
        $run_scramble = 1;
    }
    if (defined($filter) && ref($filter) ne 'CODE'){
        confess "Argument 'filter' must be a code ref and not a " . ref($filter) ;
    }

    #
    # Run scramble dumpblocks
    #
    if ($run_scramble){
        $self->_scramble_dumpblocks(%args);
    }

    my $processor = sub {
                        my %args = @_;
                        my $context = $args{context};
                        my $mismatch = $args{row};

                        if (!$count_only && !$summary_only && $context->{count} >= $max_mismatches){
                            logcomment("Reached $max_mismatches filtered mismatches. Stopping parsing of dumpblocks file");
                            return 0;
                        }

                        if (!defined($filter) || $filter->($mismatch,$context)){
                            $context->{count} ++;
                            if (!$count_only && !$summary_only){
                                push @{$context->{mismatches}}, $mismatch;
                            }
                            elsif ($summary_only){
                                $context->{summary}->{$mismatch->{snapid}}->{$mismatch->{pub}}->{$mismatch->{fileid}} ++;
                                $context->{summary}->{disks}->{$mismatch->{disk}} ++;
                            }
                        }
                        return 1;
                        };

    #
    # Parse file
    #
    my $context = $self->_parse_dumpblocks_file(processor=>$processor, %args);

    if ($count_only){
        return $context->{count};
    }
    elsif($summary_only){
        return $context->{summary};
    }
    else{
        return $context->{mismatches};
    }
}

sub _parse_dumpblocks_file {
    my $self = shift;
    my %args = @_;
    my $max_lines_logged = $args{max_lines_logged};
    my $processor = $args{processor};
    my $vol = $args{vol};
    if (!$vol){
        if ($self->{vol}){
            $args{vol} = $self->{vol};
        }
        else{
            confess "You need to pass a vol argument";
        }
    }
    if (!defined($processor) || ref($processor) ne 'CODE'){
        confess "Argument 'processor' must be a code ref";
    }
    if (!defined($max_lines_logged)){
        $max_lines_logged = 1000;
    }

    if (!$self->{mountpoint}){
        $self->_mount_volume();
    }
    my $filepath = "$self->{mountpoint}/WAFL_dumpblocks.$vol";
    my $fh;
    if (-f $filepath){
        $fh = new FileHandle($filepath)
                or confess "Unable to open $filepath: $!";
    }
    else{
        confess "File '$filepath' not found";
    }

    my $i = 0;
    my $printed_first_lines = 0;
    my $is_header = 1;
    my @headers;
    my $text;
    my $context = {};
    $context->{count} = 0;

    while (<$fh>){
        my $line = $_;
        chomp($line);
        if ($is_header && $line =~ /^#/){
            # Line is a comment, skip
        }
        elsif ($is_header && $line =~ /^[a-z]/){
            # Line is the header line
            @headers = split(/,/,$line);
            $is_header = 0;
        }
        elsif(!$is_header && $line =~ /^[0-9a-f]/){
            my @vals = split(/,/,$line);
            my %row;
            @row{@headers} = @vals;
            my $ret_status = $processor->(row=>\%row, context=>$context);
            if (!$ret_status){
                last;
            }
        }
        else{
            confess "No match: $line";
        }
        if ($i <= $max_lines_logged){
            $text .= "$line\n";
            if ($i == $max_lines_logged){
                logcomment("First $max_lines_logged lines in scramble dumpblocks file:\n$text");
                $printed_first_lines = 1;
            }
        }
        $i ++;
    }
    if (!$printed_first_lines){
        logcomment("Scramble dumpblocks output:\n$text");
    }
    $fh->close;
    return $context;
}

sub reset {
    my $self = shift;
     $self->{data}->{initial_stats} = $self->get_current_stats();
}

sub attach {
    my $self = shift;
    my %params = @_;
    #blkrealloc, redirect, reallocate
    my $scanner = $params{scanner};
    #force/override running
    if ($scanner) {
        $self->{running} = eval{"RUNNING_SCAN_$scanner"};
    } else {
        $self->{running} = RUNNING_REALLOCATE;
    }
}

#check stats, return either REALLOCATING or REDIRECT or REALLOCATING|REDIRECT
sub detect {
    my $self = shift;
    my %params = @_;
    my $aggr = $self->{aggr};

    my $data = $self->data();
    my $delta = $data->{delta};
    if ($self->{debug}){
        logcomment("stats after attach and detect:\n" . as_string($delta));
    }

    if ($delta->{"aggregate:$aggr:blkr_blocks_reallocated"} &&  $delta->{"aggregate:$aggr:blkr_redirect_blocks_updated"}) {
        return (REALLOCATING|REDIRECT);
    }elsif ($delta->{"aggregate:$aggr:blkr_blocks_reallocated"}) {
        return REALLOCATING;
    }elsif ($delta->{"aggregate:$aggr:blkr_redirect_blocks_updated"}) {
        return REDIRECT;
    }else {
        return 0;
    }

}

sub _scramble {

    my %opts = @_;
    my $filer = $opts{filer};
    my $files = $opts{files};
    my $fc = $opts{fc};
    my $invalidate_disk = $opts{invalidate_disk};
    my $invalidate_ec = $opts{invalidate_ec};
    my $subtest_count = $opts{processes};
    if (!$subtest_count){
        $subtest_count = 4;
    }
    my $filecount = scalar @$files;
    #
    # Declare subroutine to be called as a subtest (ipc)
    #
    my $subcode = <<'EOT';
    {
        use Perf::ModeAgnostic qw(say_wrap);
        use San::SanFiler;

        my ($filer, %opts) = @_;

        #my $fh = host($filer);
        #my $fc = connect ('#rsh', -timeout, 10);
        # Create a SanFiler object
        my $filer_obj = new SanFiler(-filer => $filer);
        confess "Failed to create a SAN Filer object: $filer_obj->errmsg()"
            if ($filer_obj->error());
        my $fc = $filer_obj->connection();
        confess "Failed to retrieve the filer connection: $filer_obj->errmsg()"
            if ($filer_obj->error());

        my $subtest_scrambled_blocks = 0;
        while (1) {
            #
            # Read new message containing either 'stop' or string of
            # scramble commands
            #
            my $cmd = Subtest::ipc()->read_message();
            if ($cmd eq 'stop') {
                logcomment("Scrambled blocks by this subtest : $subtest_scrambled_blocks\n");
                logresult("INFO", $subtest_scrambled_blocks);
                last;
            }
            else {
                $cmd = "priv set -q test; $cmd";
                my $scrambled_blocks = 0;
                my $out;
                eval{
                    (undef, undef, $out) = say_wrap($fc, $cmd, -timeout, 900, -maxidle, 300);
                };
                my $retries = 0;
                my $error = $@;
                while ($error){
                    if ($error =~ /TIMEOUT|IDLE/){
                        if ($retries < 3){
                            logcomment("Caught 'say' timeout. Retrying");
                            sleep 20;
                            eval{
                                (undef, undef, $out) = say_wrap($fc, $cmd, -timeout, 900, -maxidle, 300);
                            };
                            $error = $@;
                            $retries ++;
                        }
                        else{
                            logcomment("Tried 3 times. Giving up");
                            last;
                        }
                    }
                    else{
                        logcomment("Got error while scrambling: $@");
                        last;
                    }
                }

                while ($out =~ /reallocated (\d+) blocks/g){
                    $scrambled_blocks += $1;
                }
                if (!$scrambled_blocks){
                    confess "Unable to scramble blocks using command '$cmd'";
                }
                $subtest_scrambled_blocks += $scrambled_blocks;
            }
        }
    }
EOT

    my $start_time = time;
    #
    # Start all the IPC subtests running the code
    # defined above
    #
    #my $filer = $self->{fc}->hostp()->hostname;
    my @subtests;
    for (1...$subtest_count) {
       my $test = subtest($subcode,'-bg','-ipc', '--', $filer);
       push @subtests, $test;
    }

    #
    # Send files to scramble in batches of 5 files at a time
    # Loop over the available subtests
    #
    my $cmd = "";
    my $files_processed = 0;
    my $subtest_idx = 0;
    my $scramble_call = "scramble path -n 0 -a";
    if ($invalidate_disk){
        $scramble_call .= " -p 0";
    }
    if ($invalidate_ec){
        $scramble_call .= " -e";
    }
    for my $filepath (@$files){
        $cmd .= "$scramble_call $filepath;";
        $files_processed ++;
        if ($files_processed % 5 == 0  || $files_processed == $filecount){
            my $s = $subtests[$subtest_idx];
            $s->ipc()->write_message($cmd);
            $subtest_idx++;
            $subtest_idx = $subtest_idx % $subtest_count;
            $cmd = "";
        }
    }

    #
    # Stop the subtests
    #
    foreach my $s (@subtests) {
       $s->ipc()->write_message("stop");
    }
    foreach my $s (@subtests) {
        $s->wait_finish();
    }
    my $scrambling_time = time - $start_time;

    #
    # Loop through subtests and count all scrambled blocks
    #
    my $total_scrambled_blocks = 0;
    foreach my $s (@subtests) {
        my @results = $s->results();
        my @msgs = $results[0]->messages();
        $total_scrambled_blocks += $msgs[0];
    }
    logcomment("Total blocks scrambled in ${scrambling_time}s: $total_scrambled_blocks");

    logresult("INFO", $total_scrambled_blocks);

    #return $total_scrambled_blocks;
}

sub _status_reallocate {
    my $outref = shift; # Reference to output of reallocate status command
    if ($$outref =~ /No reallocation status was found for/) {
        return NONE;
    }
    elsif ($$outref =~ /State: Idle/) {
        return IDLE;
    }
    elsif ($$outref =~ /State: Redirect:.*phase (\d)/) {
        my $phase = $1;
        my $status = $REDIRECT_PHASES{$phase};
        if (!$status){
            confess "Unknown phase: $phase";
        }
        return $status;
    }
    elsif ($$outref =~ /State: Reallocating/) {
        return REALLOCATING;
    }
    elsif ($$outref =~ /State: Quiesce/) {
        return QUIESCE;
    }
    elsif ($$outref =~ /State: Checking/) {
        return CHECKING;
    }
    elsif ($$outref =~ /State: Deleting/){
        return DELETING;
    }
    #
    # TODO: Remove this elsif. This is a temporary fix for burt260200
    #
    elsif ($$outref =~ /State: Redirect/) {
        return REDIRECT;
    }
    else {
        confess "Unknown status: $$outref";
    }
}

sub _status_blkrealloc {
    my $outref = shift;
    if ($$outref =~ /\s*\d+\s+block reallocation/){
        return REALLOCATING;
    }
    elsif ($$outref =~ /No WAFL scans/){
        return NONE;
    }
    elsif ($$outref =~ /usage:/) {
        return NONE;
    }
    elsif($$outref =~ /Scan id\s+Type of scan\s+progress/){
        # Some scans are running, just not ours
        return NONE;
    }
    else{
        confess "Unknown status : $$outref";
    }
}

sub _status_redirect {
    my $outref = shift;
    my $vol = shift;
    if (!$outref || !$vol){
        confess "You must specify an output and a volume";
    }
    if ($outref && ref($outref) ne 'SCALAR'){
        confess "You must pass a reference to the output";
    }
    if ($$outref =~ /No WAFL scans/){
        return NONE;
    }
    if ($$outref =~ /usage:/) {
        return NONE;
    }

    my @lines = split("\n",$$outref);
    my $current_vol;
    for my $line (@lines){
        if ($line =~ /Volume (.*):/){
            $current_vol = $1;
        }
        elsif($current_vol eq $vol && $line =~ /\s*\d+\s+redirect\s+(\w+)\s+phase (\d)/){
            my $phase = $2;
	    my $phase_state = $1;
            my $status = $REDIRECT_PHASES{$phase_state}{$phase};
            if (!$status){
                confess "Unknown phase: $phase";
            }
            return $status;
        }
    }

    if($$outref =~ /Scan id\s+Type of scan\s+progress/){
        # Some scans are running, just not ours
        return NONE;
    }
    else{
        confess "Unknown status: $$outref";
    }
}

sub DESTROY {
    my $self = shift;
    my $old_error = $@;

    #
    # Skip the destruction for subtests
    #
    return unless ($$ == $self->{pid});

    #
    # Skip the destruction for threads
    #
    my $tid = threads->tid();
    if ($tid != 0){
        return;
    }

    logcomment("\nDESTROY of Reallocation.pm started\n");

    #
    # Re-set the flags to their original values
    #
    setflag($self->{fc},'wafl_blkr_segment_size',$self->{initial_wafl_blkr_segment_size});
    setflag($self->{fc},'wafl_blkr_policy',      $self->{initial_wafl_blkr_policy});
    setflag($self->{fc},'wafl_blkr_min',         $self->{initial_wafl_blkr_min});
    setflag($self->{fc},'wafl_blkr_max',         $self->{initial_wafl_blkr_max});
    setflag($self->{fc},'wafl_blkr_fix_pvbn',    $self->{initial_wafl_blkr_fix_pvbn});

    #
    # TODO: Clean up schedule, on/off,
    #

    say($self->{fc}, "priv set -q diag; rm /vol/vol0/etc/WAFL_dumpblocks.$self->{vol}");

    my $mountpoint_path = $self->{mountpoint};
    if ($mountpoint_path){
        GenericMount::umount($mountpoint_path);
        logcomment($@) if $@;
        delete $self->{mountpoint};
    }
    $@ = $old_error;
}

## @pod here

=head1 NAME

    Reallocation - thin wrapper for filer "reallocate" command

=head1 SYNOPSIS

    use Perf::Reallocation qw(:status);
    use Tharn;

    my $fc = connect($FILER_CONN);
    my $realloc = new Perf::Reallocation(fc => $fc, aggr => "myaggr");
    $realloc->start;
    $realloc->wait_for_status( DONE );
    my $results = $realloc->data();


=head1 DESCRIPTION

    Reallocation is a simple class for interacting with the 'reallocate'
    command from within NATE tests.

=head1 METHODS

=head2 new

    Constructor for Reallocation objects.

    Arguments:
        fc   - NATE fullconnspec
        aggr - name of an aggregate to perform block reallocation on
        vol  - name of a volume to perform file reallocation on
        file - name of a file to perform file reallocation on

    Options:
        target_type    - when specifying multiple arguments you need to
                         explicitly input the type of reallocation
                         you want this object to perform

=head2 start

    Start reallocation on the target

    Options:
        once                - Run reallocate just once
        interval            - Run reallocate every x minutes
        threshold           - Threshold under which file reallocation won't start
        no_threshold_check  - Skip previous CHECK
        force               - force realloction
        scanner             - run a scanner instead of the reallocate command.
                              Value can be 'blkrealloc' or 'redirect'

=head2 restart

    Restart reallocation

=head2 status

    Provide status of current reallocation process

=head2 is_status

    Checks whether the current reallocation status matches with one or several
    possibilities

    Arguments:
        $status   -   Statuses to check against. Example: NONE | IDLE

=head2 wait_for_status

    Block until reallocations status reaches one or several possibilities

    Arguments:
        $status   -   Statuses to check against. Example: NONE | IDLE

=head2 toggle

    Enable or disable reallocation

    Arguments:
        $on_off  - string 'on' | 'off'

=head2 quiesce

    TODO

=head2 wafl_measure_layout

    TODO

=cut

=head2 get_mismatched_blocks

    Returns list of mismatched blocks using scrable dumpblocks

=over

=item Arguments

    B<count_only> = 1 | 0

    I<Optional> Only returns total count of mismatches found instead of list if true. Default: 0.

    B<activefs_only> = 1 | 0

    I<Optional> Only returns mismatches found in active filesystem if true. Default: 0.

    B<mismatches_only> = 1 | 0

    I<Optional> Only returns mismatches instead of description of each block in volume. Default: 1.

    B<run_scramble> = 1 | 0

    I<Optional> Will run scramble dumpblocks if true. Will read existing result file if false. Default: 1.

    B<hide_pvbn_unknown> = 1 | 0

    I<Optional> Will not display mismatches of type PVBN_UNKNOWN. Default: 0.

    B<filter> = coderef

    I<Optional> Filter to run on every mismatch found as the result file is being processed

=item Examples

    # Mismatches in fileid 83 of snapshots are okay. Others are not
    my $mismatch_filter = sub {
                           my $mismatch = shift;
                           if ($mismatch->{fileid} ne 83){
                               return 1;
                           }
                           elsif (!$mismatch->{snapid}){
                               return 1;
                           }
                           return 0;
                          };
    my $mismatches = $reallocation->get_mismatched_blocks(count_only=>1, filter=>$mismatch_filter);

=back

=cut

1;
