# -*- perl -*-
#
## @author   bshelton@netapp.com, jroussin@netapp.com
## @summary  Collect and display filer counters.
## @status   review
## @pod here

=head1 NAME

    Perf::Stats

=head1 SYNOPSIS

    use Perf::Stats;

    my $st = Perf::Stats->new(node => $FILER, counters => [{object   => 'disk',
                                                            instance => '(0b.16|0c.12)',
                                                            counter  => 'user_read_blocks'},
                                                           {object   => 'disk',
                                                            instance => '(0b.16|0c.12)',
                                                            counter  => 'user_write_blocks'},
                                                           {object   => 'ext_cache_obj',
                                                            instance => '.*',
                                                            counter  => 'hit'}]);

    $st->add(['pvbn_alias'], source=>'wafl_stats');

    $st->alias(name=>"rdblks", counters => [{ object   => 'disk',
                                              instance => '.*',
                                              counter  => 'user_read_blocks' }]);

    $st->alias(name=>"wrblks", counters => [{ object   => 'disk'
                                              instance => '.*',
                                              counter  => 'user_write_blocks' }]);

    my $samp1 = $st->sample();

    ...

    my $samp2 = $st->sample();
    my $delta = $st->delta($samp1, $samp2);

    for my $counter (keys %$delta){
        if ($counter !~ /^_/){
            print "$counter = $delta->{$counter} \n";
        }
    }

=head1 DESCRIPTION

    Provides single interface to manipulate "stats", "statistics",
    "cm_client", "wafl stats" (same as "wafl susp"), wafl bufstats,
    and "wafl cpstats" data.

    Works in 7-mode and 10-mode.
    Handles multi-node querying and aggregation.

=head1 LIMITATIONS

    This module may not work very well below GordonBiersch in 7-mode or Boilermaker in C-mode

=head1 METHODS

=cut

package Perf::Stats;

use strict;
use warnings;
use Tharn qw(subtest connect host $Log);
BEGIN{
    my @c = caller(2);
    $Log->comment("Loading " . __FILE__ . " from $c[1]:$c[2]");
}
use Params::Validate qw(validate validate_with SCALAR HASHREF BOOLEAN ARRAYREF OBJECT UNDEF CODEREF);
use Module::Common qw(validate_mixed);
use Carp qw(confess);
use Time::HiRes qw(time);
use Data::Dumper;
use Filer::Version qw(get_ontap_version version_compare);
use Clone;
use Readonly;
#use Server;
use NACL::C::Node;
use List::Util qw(first);
use List::MoreUtils qw(any uniq none);

use Perf::Stats::InstanceInfo;
use Perf::Stats::Handler::CM::Stats;
use Perf::Stats::Handler::WaflStats;
use Perf::Stats::Handler::CPStats;
use Perf::Stats::Handler::BufStats;
use Perf::Stats::Handler::CM::Statistics;
use Perf::Stats::Handler::CM::Cluster::Statistics;
use Perf::Stats::Handler::CM::Cluster::ZAPI;
use Perf::Stats::Handler::CM::CM_Client;
use Perf::Stats::Handler::CM::ZAPI;
use Perf::Stats::AliasFactory;
use Perf::Stats::Entry;

our $GNUPLOT = "/usr/software/bin/gnuplot";
if (! -e $GNUPLOT){
    $GNUPLOT = "/usr/local/bin/gnuplot";
}

my %HANDLERS = ( 'stats'        => 'Perf::Stats::Handler::CM::Stats',
                 'wafl_stats'   => 'Perf::Stats::Handler::WaflStats',
                 'cpstats'      => 'Perf::Stats::Handler::CPStats',
                 'bufstats'     => 'Perf::Stats::Handler::BufStats',
                 'statistics'   => 'Perf::Stats::Handler::CM::Statistics',
                 'cm_client'    => 'Perf::Stats::Handler::CM::CM_Client',
                 'zapi_perf'    => 'Perf::Stats::Handler::CM::ZAPI',
                );

our $MAXBUFSIZE = 1024*1024*1024; # 1g

Readonly::Scalar our $TIMEOUT  => 120;

#
# TODOs:
#   1. Enhance wafl_stats parsing
#   2. Allow walf_stats counters to be type-d by the user
#   3. Automatically set 'format y' based on yrange
#   4. Add support for mem_stats
#   5. Add leak detection subroutines
#

=head2 new

Constructor for Perf::Stats class

=over

=item Arguments

B<node> = string | NACL::APISet

I<Optional> Filer name or NACL::APISet object connected to filer

Note: users passing in NACL::APISet objects are responsible for
setting a reasonable connection timeout value.

B<all_nodes_in_cluster> = 0 | 1

I<Optional> Specifies if counters should be retrieved from all nodes in the cluster.
Defaults to 1 for source 'statistics' unless 'nodes' is set.
Default to 0 for all other sources.

B<force_multinode> = 0 | 1

I<Optional> Ensure Stats accepts counter regexps with a node portion
specified even if operating on a single node.
Defaults to 0.

B<nodes> = arrayref

I<Optional> List of filer names or NACL::APISet objects. Use this argument if
you wish to either:

    - Retrieve counters for more than one filer for sources other than 'statistics'
    - Retrieve counters from a subset of the cluster nodes

With this option set, counters will be prefixed with their node name

B<counters> = arrayref of regular expressions

I<Optional> See 'add' method for more details.

B<source> = 'stats' | 'wafl_stats' | 'cpstats' | 'statistics'
          | 'zapi_perf' | 'cm_client' | 'bufstats'

I<Optional> Defaults to 'stats'. See 'add' method for more details.

=item Examples

    # Retrieve CM counters for $filer
    my $st = Perf::Stats->new(node     => $filer,
                              counters => [{object   => 'disk',
                                            instance => '(0b.16|0c.12)',
                                            counter  => 'user_read_blocks'},
                                           {object   => 'disk',
                                            instance => '(0b.16|0c.12)',
                                            counter  => 'user_write_blocks'},
                                           {object   => 'ext_cache_obj',
                                            instance => '.*',
                                            counter  => 'hit'}],
                              source => 'stats',);

    # Create instance looking at a subset of nodes in cluster
    my $st = Perf::Stats->new(
        nodes    => [$filer1, $filer2],
        counters => [{node     => '.*',
                      object   => 'volume',
                      instance => 'vol0',
                      counter  => '.*'}]

          source => 'statistics',);

    # Query all nodes in cluster
    my $st = Perf::Stats->new( node     => $filer,
                               counters => [".*"],
                               source   => 'wafl_stats',
                               all_nodes_in_cluster => 1,
                             );

    # Handle multinode counters on a single node
    my $st = Perf::Stats->new(node     => $filer,
                              counters => [{node     => '.*',
                                            object   => 'nfs3',
                                            instance => 'all',
                                            counter  => '(?:read|write|total)'}
                                ],
                              source               => 'statistics',
                              all_nodes_in_cluster => 0,
                              force_multinode      => 1,);

=back

=for html <hr>

=cut

sub new {
    my $class    = shift;
    my %params = validate(
                   @_,
                   {
                       node                 => {type => SCALAR   | UNDEF | OBJECT, optional => 1},
                       nodes                => {type => ARRAYREF | UNDEF, optional => 1},
                       all_nodes_in_cluster => {type => SCALAR   | UNDEF, optional => 1},
                       force_multinode      => {type => SCALAR   | UNDEF, optional => 1, default => 0 },
                       counters             => {type => ARRAYREF | UNDEF, optional => 1},
                       source               => {type => SCALAR   | UNDEF, optional => 1, default => "stats"},
                       # For backwards compatibility only
                       fc                   => {type => OBJECT   | UNDEF, optional => 1},
                       debug                => {type => SCALAR   | UNDEF, optional => 1},
                    }
            );

    # Ignored
    delete $params{debug};

    #
    # Set the node/nodes parameter
    #
    my $node = delete $params{node};
    if ($node && !ref($node)) {
        # Cleanup
        $node =~ s/^\s+//;
        $node =~ s/\s+$//;
    }
    my $fc = delete $params{fc};
    if ($fc){
        my $fc_node;
        if ($fc->isa('Connectrec')){
            $fc_node = $fc->hostp->hostname;
        }
        else{
            Carp::confess("Argument fc must be a Connectrec");
        }
        if (!$node){
            $node = $fc_node;
        }
        elsif ($node ne $fc_node){
            Carp::confess("Arguments 'fc' and 'node' conflict ('$fc_node' vs '$node')");
        }
    }

    my $counters = delete $params{counters};
    my $source   = delete $params{source};
    my $self = \%params;

    for my $n (@{$self->{nodes}}) {
        # Sanitize
        $n =~ s/^\s+//;
        $n =~ s/\s+$//;
    }

    # Counter source handlers
    for my $source ( keys %HANDLERS ){
        $self->{handlers}->{$source} = {};
    }

    my %apisets;

    if (!defined($params{nodes}) || 0 == scalar(@{$params{nodes}})) {
        confess "You must specify 'node' or 'nodes'" if (!$node);

        if (ref($node)) {
            if ($node->isa('NACL::APISet')) {
                my $apiset = $node;
                $node = $node->get_name();
                $apisets{$node} = $apiset;
            }
            else {
                confess(ref($node) . ' is not a supported node parameter');
            }
        }

        if ($params{all_nodes_in_cluster}) {
            my $handler_class = $HANDLERS{$source};
            my $handler =
              $handler_class->new(node   => $node,
                                  source => $source,
                                  apiset => delete($apisets{$node}),);
            $params{nodes} = $handler->get_node_names();
            $Log->comment("Found following nodes in cluster:\n"
                          . join("\n", @{$params{nodes}}));
        }
        else {
            $params{nodes} = [$node];
        }
    } ## end if (!$params{nodes})
    elsif ($node) {
        confess("Only specify one of 'node' or 'nodes' ");
    }
    else {
        # make a copy of nodes to prevent interference with the
        # caller's array
        my @apisets = @{$params{nodes}};
        $params{nodes} = \@apisets;

        # handle potential APISet
        foreach my $index (0 .. $#{$params{nodes}}) {
            my $cur_node = $params{nodes}->[$index];

            if (ref($cur_node)) {
                if ($cur_node->isa('NACL::APISet')) {
                    my $node_name = $cur_node->get_name();

                    confess("Duplicate entries provided for node $node_name")
                      if (scalar(grep { $_ eq $node_name } @{$params{nodes}})
                          || defined($apisets{$node_name}));

                    # replace APISet with hostname
                    $params{nodes}->[$index] = $node_name;

                    # and save away the APISet to pass to the node's Handler
                    $apisets{$node_name} = $cur_node;
                }
                else {
                    confess(
                          ref($node) . ' is not a supported connection object');
                }
            }
        }
    } ## end else [ if (!$params{nodes})

    $self->{apisets} = \%apisets;

    # Samples taken by this Stats object will be stamped
    # with this vid. Samples with the same vid
    # can be compared
    $self->{current_vid} = 0;

    $self->{pid} = $$;

    # Tracked Counters are kept in the Handlers but Aliases
    # and AliasFactories are kept in the main object.
    $self->{aliases}        = {};
    $self->{alias_factories}= {};
    bless $self, $class;

    $self->add($counters, source=>$source) if ($counters);
    return $self;

}

=head2 sample

    Takes a sample of the counters and returns a hash

=cut

sub sample {
    my $self = shift;
    my $handlers = shift;
    my $sample = {};
    $self->_stamp($sample);

    if(!$handlers) {
        $handlers = $self->_get_handlers;
    }

    # Send all the sample commands in parallel to all the nodes
    $_->sample_start for (@$handlers);

    for my $handler (@$handlers){
        my $handler_sample = $handler->sample_wait;

        # TODO: This doesn't go through entry_display_name but it's more efficient.
        # Revise as necessary
        if ($self->is_multinode){
            my $node = $handler->get_node;
            my %noderized = map { +"$node:$_" => $handler_sample->{$_} } keys %$handler_sample;
            $handler_sample = \%noderized;
        }

        # The 'add' method is supposed to ensure that no collisions happen below
        @$sample{ keys %$handler_sample } = values %$handler_sample;
    }

    $sample = $self->_handle_aliases($sample);
    $Log->comment("Sampled counters:\n" . _as_string($sample));

    return $sample;
}
sub _handle_aliases {
    my $self = shift;
    my $sample = shift;

    for my $alias (values %{$self->{aliases}}){
        my $name = $alias->get_name;
        my $value = $alias->compute;

        $sample->{$name} = $value;

        # Handle no_zero_values aliases
        if (! exists $sample->{$name} ){
            if ($alias->no_zero_values){
                $sample->{$name} = $alias->get_zero_value;
                next;
            }
            else {
                confess "Alias '$name' missing from sample";
            }
        }
        if (! defined $sample->{$name} || $sample->{$name} eq '' ){
            confess "Alias '$name' has no value in sample";
        }
    }
    return $sample;
}


sub _stamp {
    my ($self,$sample) = @_;
    $sample->{_vid} = $self->{current_vid};
    $sample->{_time} = time;
}

=head2 checkpoint

    Takes a sample of the counter and store it in the object. This subroutine
    is meant to be used with the 'delta' subroutine. Overrides the sample taken
    during 'delta'.

=cut

sub checkpoint {
    my $self = shift;
    $self->{last_sample} = $self->sample();
    return;
}

=head2 monitor

    Poll the counters in the background

    Arguments:
        interval    - polling interval. Optional. Default is 5s.

=cut

sub monitor {
    my $self = shift;
    my %args = @_;
    my $interval = $args{interval};
    if ( !defined($interval) ){
        $interval = 5;
    }
    # TODO: This won't work on Windows. Won't fix until somebody complains.
    my $job = subtest(sub { $self->_collect_samples($interval) },'-bg', '-ipc');
    $self->{monitor}->{job} = $job;
    $self->{monitor}->{logdir} = $job->logdir;
    return $job;
}

sub _collect_samples {
        my ($self, $interval) = @_;
        my $samples = [];
        my $ipc = Subtest::ipc();
        # Check for abort message every 5s
        my $interrupt_interval = 5;
        if ($interrupt_interval > $interval){
            $interrupt_interval = $interval;
        }
        my $next_sample_time = time;
        my $sleep_time;
        while (!$ipc->can_read){
            if (time - $next_sample_time >= 0){
                push @$samples, $self->sample;
                $next_sample_time += $interval;
            }
            $sleep_time = $next_sample_time - time;;
            $sleep_time = $sleep_time > $interrupt_interval?$interrupt_interval:$sleep_time;
            if ($sleep_time > 0){
                select(undef,undef,undef,$sleep_time);
            }
        }
        $ipc->read_message;
        print "Monitoring done. Returning samples\n";
        $ipc->write_obj($samples);
        return;
}

=head2 monitor_stop

    Stop monitoring and return arrayref of collected samples

=cut

sub monitor_stop {
    my $self = shift;
    my $job = $self->{monitor}->{job};
    if (!$job){
        $Log->comment("No monitor job found");
        return;
    }
    if (!$job->isAlive){
        $Log->comment("Monitor job is not alive");
        return;
    }
    my $ipc = $job->ipc;
    $ipc->write_message('stop');
    my $samples = $ipc->read_obj;
    if ($@){
        confess("Unable to get samples from monitor: $@");
    }
    $ipc->quit;
    $job->wait_finish;
    delete $self->{monitor}->{job};

    $self->{monitor}->{samples} = $samples;
    return $samples;
}

sub DESTROY {
    my $self = shift;
    return unless $$ == $self->{pid};

    my $err = $@;
    eval{
        $self->monitor_stop;
    };
    if ($@){
        $Log->comment("Error during destroy of Stats: $@");
    }
    $@ = $err;
}

=head2 graph

    Generates gnuplot input files and graphs for an array of samples. if no arrayref is
    provided, use the array generated during the last 'monitor' operation. Only counters
    whose values are numbers or can be turned into numbers can graphed (such as average,
    percent or rate). Others are ignored.
    Files will be located in the same folder as the logs for the thpl or subtest calling
    this subroutine in a folder whose name starts with 'stats_graph_'.

    Arguments:
        samples     - arrayref of samples. Optional.

=cut

sub graph {
    my $self = shift;
    my %args = @_;
    my $samples = $args{samples};
    if (!defined($samples) && $self->{monitor}->{job}){
        $self->stop;
        $samples = $self->{monitor}->{samples};
    }
    if (ref($samples) ne 'ARRAY'){
        confess("Argument 'samples' must be an arrayref");
    }
    my $vid;
    for my $s (@$samples){
        if ( ref($s) ne 'HASH' || ! exists $s->{_vid}){
            confess("Argument 'samples' must contain Stats samples");
        }
        if (!$vid){
            $vid = $s->{_vid};
        }
        elsif ($s->{_vid} ne $vid){
            confess("Samples must come from the same Stats object");
        }
    }

    # Create a directory to store the graphs
    my $timestamp = time;
    my $dir = "$self->{monitor}->{logdir}/stats_graphs_${vid}_$timestamp";
    $Log->comment("Graphs saved in $dir");
    if (! -e "$dir") {
        mkdir("$dir", 0755) || confess("Can't make $dir: $!\n");
    }
    if (! -d "$dir") {
        confess("$dir isn't a directory\n");
    }

    #
    # For all counters but the ones of type 'raw',
    # we'll graph the deltas rather than the actual values.
    #
    my $sorted_samples = $self->_sort($samples);
    my $deltas = $self->delta(@$sorted_samples, already_sorted=>1, hide_zeros=>0, exact_format=>0);

    #
    # Create one graph per counter
    #

    # Figure out axis scales
    my $start_timestamp = $sorted_samples->[0]->{_time};
    my $end_timestamp = $sorted_samples->[-1]->{_time};
    my $xrange = _choose_range($end_timestamp - $start_timestamp);

    my $errors;
    my @counters = keys %{$samples->[0]};
    for my $counter (@counters){
        # Skip metadata
        if ($counter =~ /^_/){
            next;
        }

        # Find the counter type so we know what to graph
        my $handler = $self->_get_handler($counter);
        my $counter_obj = $handler->tracked_get($counter);
        my $type = $counter_obj->get_type;

        # Find the maximum value for that counter and adjust y-scale
        my $max;
        if ($type eq 'raw'){
            $max = $samples->[0]->{$counter};
            # Skip raw counters that contain string values rather than numbers
            if ($max !~ /^\d+\.?\d*$/){
                $Log->comment("Counter '$counter' does not appear to be a number");
                next;
            }
            map { $max = $_->{$counter} if ($_->{$counter} > $max)} @$samples;
        }
        else{
            $max = $deltas->[0]->{$counter};
            map { $max = $_->{$counter} if ($_->{$counter} > $max)} @$deltas;
        }
        my $yrange = _choose_range($max);

        #
        # Start a pipe to gnuplot with its output redirected to the
        # destination graph file.
        #
        my $parsed_counter = $counter;
        $parsed_counter =~ s/\:/__/g;
        my $graph_file = $dir . "/" . $parsed_counter . '.png';
        my $data_file = $dir . "/" . $parsed_counter . '.data';
        open(GRAPH, "|tee $data_file |$GNUPLOT > $graph_file") || die "Can't open pipe to $GNUPLOT: $!\n";

        print GRAPH "set terminal png\n";
        print GRAPH "set nokey\n";
        print GRAPH "set title \"$counter ($type)\\nVersionid: $vid\\nTimestamp: $timestamp\"\n";
        print GRAPH "set grid\n";

        print GRAPH "set yrange [0:$yrange]\n";
        print GRAPH "set ytics 0,", $yrange/10, ",", $yrange, "\n";

        print GRAPH "set xrange [0:$xrange]\n";
        print GRAPH "set xlabel \"Seconds (s)\"\n";
        print GRAPH "set xtics 0,", $xrange/10, ",", $xrange, "\n";

        #
        # The data points appear on gnuplot's input stream by
        # specifying '-' as the file name.  A line with a single
        # 'e' marks the end of the data.
        #
        if ($type eq 'raw'){
            print GRAPH "plot '-' with linespoints\n";
            for my $sample (@$sorted_samples){
                # Keep relative seconds since start
                my $timestamp = $sample->{_time} - $start_timestamp;
                print GRAPH $timestamp . " " . $sample->{$counter} . "\n";
            }
        }
        else{
            print GRAPH "plot '-' with fsteps\n";
            print GRAPH "0 0\n";
            for my $delta (@$deltas){
                # Keep relative seconds since start
                my $timestamp = $delta->{_time} - $start_timestamp;
                print GRAPH $timestamp . " " . $delta->{$counter} . "\n";
            }
        }
        print GRAPH "e\n";
        close(GRAPH);

        # Make sure a graph was actually created
        if ( ! -f $graph_file || -z $graph_file){
            $errors .= "Unable to create graph for counter '$counter' in '$graph_file'\n";
        }
    }
    if ($errors){
        confess($errors);
    }
}

#
# We want to choose nice values for the x-axis and y-axis ranges.
# Nice values are multiples of 1, 2, or 5.  We choose the smallest
# multiple that spans the passed in maximum value.
#

sub _choose_range
{
    my ($range) = @_;
    my $r1 = 1;
    my $r2 = 2;
    my $r5 = 5;

    while (1) {
        if ($r1 >= $range) {
            return $r1;
    }
        if ($r2 >= $range) {
            return $r2;
    }
        if ($r5 >= $range) {
            return $r5;
    }

    $r1 *= 10;
    $r2 *= 10;
    $r5 *= 10;
    }
}


=head2 zero

    Zeros counters on sources which support it

=over

=item Arguments

B<node> = string

I<Optional> Filer name

B<source> = 'stats' | 'wafl_stats' | 'cpstats' | 'statistics'

I<Optional> By default, all sources will be zeroed.  Note that at this
time, only wafl_stats supports counter zeroing.  All other sources
perform no-ops.

=back

=cut

sub zero {
    my $self      = shift();
    my %pass_thru = @_;

    foreach my $handler (@{$self->_get_handlers(%pass_thru)}) {
        $handler->zero;
    }
}

=head2 add

    Adds counters to the current list of tracked counters

    Arguments:
        $counters   - arrayref of specs counters to collect
        source      - 'stats' or 'wafl_stats' or 'statistics' or 'cpstats'
                      or 'bufstats' or 'cm_client' or 'zapi_perf'
        type        - counter type

    Types:
        The 'type' argument determines how to treat counters during 'delta' method call.
        Types will be automatically detected for source 'stats' and 'statistics'. For other sources,
        the following types are available:
            raw:   Display the most recent value
            delta: Display difference between newer and older value.

        Defaults to 'delta', or 'raw' for bufstats or if counter value
        is not numerical.

    Format:
        For the counter manager based counter sources 'stats' and
        'statistics', each element in the $counters array is a hash of
        regular expressions for the various counter parts.  It is
        specified as:

        { object   => 'object_re',
          instance => 'instance_re',
          counter  => 'counter_re',
          label    => 'label_re' }

        A 'node' field can be specified if operating on multiple
        nodes.  When operating on a single node, the node field can be
        provided as long as the 'force_multinode' flag is set.

        The object part is the only mandatory one.  All unspecified
        counter parts default to '.*'.

        For non-counter manager sources, individual regex strings are passed in:

          $st->add(['pvbn_alias'], source => 'wafl_stats');
          $st->add(['cp_ph.*', 'reason'], source => 'cpstats');

        When working with multiple nodes, each regex will be understood as:

                <node regex>:<regex>

        In order to pass exact counter part names for all counter
        sources, you must take care of quoting special characters.
        This can be achieved manually, or by using \Q and \E or
        'quotemeta'.

    Examples:
        # Add all volume counters from 'stats'
        $stats->add([{object => 'volume'}], source=>'stats');

        # Add all aggregate counters from filer 'murre' from 'statistics'
        $stats->add([{node => 'murre', object => 'aggregate'}], source=>'statistics');

        # Add all volume counters from all nodes from 'statistics'
        $stats->add([{node => '.*', object => 'volume'}], source=>'statistics');

        # Add exact counter names
        $stats->add([{object   => quotemeta($object),
                      instance => quotemeta($instance),
                      counter  => quotemeta($counter_name)}],
                    source=>'stats' );

        # Add various counters from each node
        my $stats = Perf::Stats->new( nodes=>['filera','filerb'], ...);
        $stats->add([{node => 'filera', object => 'volume'},
                     {node => 'filerb', object => 'nfs3'}],
                    source => 'stats');

        # Get all wafl_stats counters from each node
        $stats->add([".*:.*"], source=>'wafl_stats');

=cut

sub add {
    my $self = shift;
    my ( $counters, %args ) = validate_mixed(
           @_,
           [ { type => ARRAYREF } ],
           {
               source     => {type => SCALAR|UNDEF, optional=>1},
               type       => {type => SCALAR|UNDEF, optional=>1},
               # Only for backwards compatibility
               rediscover => {optional=>1},
            },
    );
    if (!@$counters) {
        confess('You must specify counters!');
    }
    $Log->comment("Adding counters:\n" . _as_string($counters));

    my $source = $self->_get_source( counters=>$counters, %args );

    delete $args{source};
    my $handler_class = $HANDLERS{$source};
    my $nodes = $self->get_nodes;

    my $counter_nodes; # Map of node => list of regex indexes



    if($source =~  /(statistics|zapi_perf)/) {
        #setup counter regexs/handlers to handle counters with no node
        my $args = {
                counters => $counters,
                nodes => $nodes,
                source => $source
        };
        $counter_nodes = $self->_check_cluster_counters($args)
    }
    else {
        # In order to warm the Handler caches in parallel, first
        # figure out what nodes we need to talk to
        if ($self->is_multinode){
            foreach my $i (0..$#{$counters}) {
                my $regex = $counters->[$i];
                my ($node_regex, undef) = $self->_parse_counter($regex);
                my @matched_nodes = grep { /^$node_regex$/ } @$nodes;
                confess("Unable to find nodes matching '$node_regex' in (@$nodes)")
                    if (0 == scalar(@matched_nodes));
                push(@{$counter_nodes->{$_}}, $i)
                    for (@matched_nodes);
            }
        }
        else{
            $counter_nodes->{$nodes->[0]} = [0..$#{$counters}];
        }
    }
    #
    # Instantiate handlers and warm the caches in parallel
    #
    my $source_handlers = $self->{handlers}->{$source};
    for my $n (keys %$counter_nodes){
        my $handler = $source_handlers->{$n};
        if (!$handler){
            $handler =
              $handler_class->new(node   => $n,
                                  source => $source,
                                  apiset => delete($self->{apisets}->{$n}),);
            $source_handlers->{$n} = $handler;
        }
        my $n_counters = $counter_nodes->{$n};
        # Strip the node information if necessary
        my @formatted;
        for my $r (@{$counters}[@$n_counters]){
            my (undef, $c) = $self->_parse_counter($r);
            push @formatted, $c;
        }
        $handler->cache_populate_start(\@formatted);
    }
    for my $n (keys %$counter_nodes){
        my $handler = $self->{handlers}->{$source}->{$n};
        $handler->cache_populate_wait;
    }

    # Reverse the node => regex index mapping
    my $node_counters = { };
    while (my ($node,$regex_indexes) = each %$counter_nodes){
        push @{$node_counters->{$_}}, $node for (@$regex_indexes);
    }

    #
    # Detect counters and start tracking
    #
    my @counters_added_total;
    while (my ($regex_index, $cnodes) = each(%$node_counters)) {
        my %add_output;
        my @counters_added_for_regex;
        my (undef, $regex) = $self->_parse_counter($counters->[$regex_index]);

        # If a regex applies to more than one node, it's possible it will only
        # match counters for a subset of the nodes, that's okay
        if (scalar @$cnodes > 1){
            for my $n( @$cnodes){
                my $handler = $self->{handlers}->{$source}->{$n};
                %add_output = $handler->counters_add($regex, %args);
                push @counters_added_for_regex, @{$add_output{added_counters}}
                  if (defined($add_output{added_counters}));
            }
        }
        else{
            my $handler = $self->{handlers}->{$source}->{$cnodes->[0]};
            %add_output = $handler->counters_add($regex, %args);

            push @counters_added_for_regex, @{$add_output{added_counters}}
              if (defined($add_output{added_counters}));
        }

        for my $c (@counters_added_for_regex){
            my $counter_name = $c->get_full_name;
            if ($self->{entries}->{$counter_name}){
                confess("Counter collision detected for '$counter_name'");
            }
            $self->{entries}->{$counter_name} = $c;
        }

        confess("No counters were added for regex:\n" . Dumper($regex))
          if (0 == scalar(@counters_added_for_regex)
              && 0 == $add_output{already_added});

        push @counters_added_total, @counters_added_for_regex;
    }

    my @counters_names;

    if (0 < scalar(@counters_added_total)) {
        # Update dynamic aliases with the new counters if applicable
        for my $factory (values %{$self->{alias_factories}}) {
            # We could instead pass all the tracked counters and that would
            # simplify the merge but that wouldn't scale very well with
            # thousands of counters
            my $aliases = $factory->process(\@counters_added_total);
            $self->_merge_aliases($aliases);
        }

        $self->update_vid;

        @counters_names =
          map { $self->entry_display_name(entry => $_) } @counters_added_total;
    }

    # Don't keep the cache across 'add' calls since the state of the filer may have changed
    $_->cache_purge for ( values %$source_handlers );

    # Don't keep Handlers that aren't tracking any counters
    $self->_clean_handlers();
    # Only return the names of the counters
    return \@counters_names;
}
sub _check_cluster_counters{
    my ($self, $args) = @_;
    my $counters = $args->{counters};
    my $nodes = $args->{nodes};
    my $source = $args->{source};
    my $cluster_name = $self->{cluster_name};

    my $handler;
    if($cluster_name){
        $handler = $self->{handlers}->{$source}->{$cluster_name};
    }
    if(!$handler) {
        my $handler_class = ($source eq 'statistics' ?'Statistics':'ZAPI');
        $handler_class = "Perf::Stats::Handler::CM::Cluster::$handler_class";
        #new cluster class
        $handler = $handler_class->new(source => $source, nodes=>$nodes);

        $cluster_name = $handler->get_cluster_name;
        $self->{cluster_name} = $cluster_name;
        $self->{handlers}->{$source}->{$cluster_name} = $handler;
        $handler->set_node($cluster_name);
    }

    my $i = 0;

    my $counter_nodes = {};
    my $has_nodeless = 0;
    for my $r (@$counters) {
        my($n,$c) = $self->_parse_counter($r);
        #convert to object form if needed
        $r = $handler->_process_regex($c);
        $r->{node} = $n;
        my $objects = $handler->get_node_objects(counters=> [$r]);

        if(scalar (keys %{$objects->{no_node}}) > 0) {
            push (@{$counter_nodes->{$cluster_name}},$i);
        }
        if(scalar (keys %{$objects->{regular}}) > 0 && $n ne $cluster_name) {
            if ($self->is_multinode){
                my @matched_nodes = grep { /^$n/ } @$nodes;
                confess("Unable to find nodes matching '$n' in (@$nodes)")
                if (0 == scalar(@matched_nodes));
                push(@{$counter_nodes->{$_}}, $i)
                    for (@matched_nodes);
            }
            else {
                push @{$counter_nodes->{$nodes->[0]}}, $i;
            }
        }
        $i++;
    }
    return $counter_nodes;
}
=head2 remove

    Removes counters from the current list of tracked counters. This will also update
    all aliases those counters belong to.

    Arguments:
        $counters   - arrayref of specs counters to collect
        source      - 'stats' or 'wafl_stats' or 'statistics' or 'cpstats'

    Example:
        $myvol->delete;
        # Removes all the counters for volume 'myvol'
        $stat->remove( ["volume:myvol:.*"], source=>'statistics');

=cut

sub remove {
    my $self = shift;
    my ($counters, %args) = validate_mixed( @_,
                                            [ { type => ARRAYREF } ],
                                            {
                                                source => { type => SCALAR|UNDEF, optional => 1 },
                                                counters_as_strings => { type => SCALAR|UNDEF, optional => 1 },
                                            },
                                          );
    if (!@$counters){
        confess("You must specify counters!");
    }

    my @counters_removed_total;
    for my $counter_regex (@$counters){
        # Skip undef entries
        next if (!$counter_regex);
        my ($node_r, $counter_r) = $self->_parse_counter($counter_regex);
        my @nodes = grep { $_ =~ /^$node_r$/ } @{$self->get_nodes};
        for my $node (@nodes){
            my $handlers = $self->_get_handlers( node => $node );
            for my $handler (@$handlers){
                my $counters_removed = $handler->counters_remove(
                                                    $counter_r,
                                                    counters_as_strings => $args{counters_as_strings}
                                                );
                delete $self->{entries}->{$_->get_full_name} for (@$counters_removed);
                push @counters_removed_total, @$counters_removed;
            }
        }
    }

    # There might be empty aliases now. Remove them.
    my @updated_aliases = map { @{$_->get_aliases}  } @counters_removed_total;
    for my $alias (@updated_aliases){
        $self->alias_remove( name => $alias->get_name ) if ($alias->is_empty);
    }

    $self->update_vid;

    # Only return the names of the counters
    my @counters_names = map { $self->entry_display_name(entry => $_) } @counters_removed_total;
    return \@counters_names;
}

sub _parse_counter {
    my ($self, $counter_regex) = @_;
    confess "You must pass an argument" if (!$counter_regex);

    if ($self->is_multinode()){
        # support passing in of hashrefs
        if ('HASH' eq ref($counter_regex)) {
            if (defined($counter_regex->{node})) {
                return ($counter_regex->{node}, $counter_regex);
            }
            else {
                return ('.*', $counter_regex);
            }
        }
        else {
            return Perf::Stats::Entry::split_regexp($counter_regex);
        }
    }
    else{
        my $node = quotemeta($self->get_nodes->[0]);
        return ($node, $counter_regex);
    }
}

sub _as_string {
    my $dataref = shift;
    my $depth = shift;
    my $old_depth = $Data::Dumper::Maxdepth;

    # Stringify the reference
    $Data::Dumper::Indent = 1;
    $Data::Dumper::Sortkeys = 1;
    if ($depth){
        $Data::Dumper::Maxdepth = $depth;
    }
    my $string = Data::Dumper->Dump([$dataref],["dataref"]);
    $string =~ s/\$dataref\s=\s//;
    $string =~ s/;$//;
    if ($depth){
        $Data::Dumper::Maxdepth = $old_depth;
    }
    return $string;
}

sub _make_vid {
    return join('_', time, int(rand(2**32)));
}

sub _get_source {
    my $self = shift;
    my %args = validate_with (
        params => \@_,
        spec   => { source  => { type => SCALAR|UNDEF, optional=>1 },
                    counters=> { type => ARRAYREF|UNDEF, optional=>1 },
                    counters_as_strings => { type => SCALAR|UNDEF, optional=>1 },
                },
        allow_extra=>1,
    );
    my $source = $args{source};
    my $regexs = $args{counters};
    my $as_strings = $args{counters_as_strings};

    # If a 'source' argument is passed in validate it
    if ($source){
        return $source if ( grep { $source } keys %HANDLERS );
        confess("Unsupported source '$source'");
    }

    my @sources = uniq map { $_->get_source } @{$self->_get_handlers};
    my $sources_count = scalar @sources;
    # If there aren't any handlers, default to 'stats' for backwards compatibilty
    return 'stats' if (!$sources_count);
    # If there is a single source, return it
    return $sources[0] if ($sources_count == 1);

    # If 'counters' regexs get passed in and they only match counters
    # in a single source
    if ($regexs && @$regexs) {
        my $all_nodes = $self->get_nodes;
        my $source;
        for my $regex (@$regexs){
            my ($node_r, $counter_r) = $self->_parse_counter($regex);
            my @nodes = grep { $_ =~ /^$node_r$/ } @$all_nodes;
            for my $node (@nodes){
                my $handlers = $self->_get_handlers( node => $node );
                next if (!$handlers);
                for my $handler (@$handlers){
                    my $matches = $handler->find(   counters => [$counter_r],
                                                    counters_as_strings=>$as_strings
                                                );
                    if ($matches && @$matches ){
                        my $source_match = $handler->get_source;
                        $Log->comment("Found " . @$matches . " matching counters in source '$source_match'");
                        if ($source && $source_match ne $source ){
                            confess("You must pass a 'source' argument");
                        }
                        $source = $source_match;
                    }
                }
            }
        }
        return $source;
    }

    # At this point, we know there is more than one handler
    # and no regexs were passed. Not much can be done.
    confess("You must pass a 'source' argument");
}

=head2 alias

    Aggregates existing counters together under a single name

    Important: Values will be summed for all counter types except 'average' and 'percent'.
               Counters of type 'average' and 'percent' are already averages and the sum
               of their values would not be an accurate representation of the value of
               the aggregate. Instead, the numerator and denominator of those counters
               are summed.

               Example: $stats->alias(  name => "user_write_latency",
                                        counters => [{ object   => 'disk',
                                                       instance => '.*',
                                                       counter  => 'user_write_latency'}]);

                        The value for each user_write_latency counter is derived and
                        represented as two values by ONTAP: total_latency / io_count.
                        The alias value will be computed as sum(total_latency)/sum(io_count).
                        This computation provides an accurate value for the average
                        user_write_latency per IO across all the disks. The computed value
                        isn't the average user_write_latency per disk however.

    Arguments:
        name        - name of the alias
        counters    - arrayref of counter names or regular expressions

    Options:
        hide_counters       - by default, individual 'aliased' counters are not shown.
                              This flag overrides this behavior.
        counters_as_strings - Treat counters passed in as strings only and not regular
                              expressions. Defaults to 0.
        source              - Currently not support for 'cpstats'
        dynamic             - Specifies whether the alias should be updated when counters get added.
                              Defaults to 0.

=cut

sub alias {
    my ($self, %params) = @_;
    if (!$params{name}){
        confess "You must specify a name";
    }
    my $aliases = $self->_generic_alias( %params, type => 'name' );
    # There should be only one
    return $aliases->[0];
}

sub _generic_alias {
    my $self = shift;
    my %params = validate_with(
           params => \@_,
           spec  => {
                   counters             => {type => ARRAYREF },
                   hide_counters        => {type => SCALAR | UNDEF, optional => 1, default => 1},
                   counters_as_strings  => {type => SCALAR | UNDEF, optional => 1, default => 0},
                   counters_as_objs     => {type => SCALAR | UNDEF, optional => 1, default => 0},
                   dynamic              => {type => SCALAR | UNDEF, optional => 1, default => 0},
                   source               => {type => SCALAR | UNDEF, optional => 1},
                   # For backwards compatibility only
                   action               => { optional => 1 },
                },
            allow_extra => 1,
    );

    # Ignored parameter. For backwards compatibility only.
    delete $params{action};

    my $dynamic = delete $params{dynamic};

    my $source = $self->_get_source( %params );
    $params{source} = $source;

    # Internally counter names are fully qualified. Adjust the 'counters' input.
    my $counters = $params{counters};
    my $counters_objs;
    if ( delete $params{counters_as_objs} ){
        my @counter_names = map { $_->get_full_name } @$counters;
        $params{counters} = \@counter_names;
        $params{counters_as_strings} = 1;
        $counters_objs = $counters;
    }

    my $factory = Perf::Stats::AliasFactory->new( parent => $self, %params );
    if ($dynamic){
        my $name = $factory->get_name;
        confess "There is already a dynamic alias factory called '$name'"
            if ($self->{alias_factories}->{$name});
        $self->{alias_factories}->{$name} = $factory;
    }

    if (!$counters_objs){
        # Act on all the tracked counters
        # TODO: This means users can't create aliases of aliases
        my $handlers = $self->_get_handlers( source => $source );
        my @tracked_counters = map { @{$_->tracked_get()} } @$handlers;
        $counters_objs = \@tracked_counters;
    }

    my $aliases = $factory->process($counters_objs);
    $self->_merge_aliases($aliases);

    $self->update_vid;
    return $aliases;
}

sub _merge_aliases {
    my ($self, $aliases) = @_;
    for my $alias (@$aliases){
        my $name = $alias->get_name;
        if (!$self->{entries}->{$name}){
            confess("An alias named $name already exists")
                if ($self->{aliases}->{$name});
            $self->{aliases}->{$name} = $alias;
            $self->{entries}->{$name} = $alias;
        }
        else{
            confess "Collision detected for alias $name" if (!$self->{aliases}->{$name});
            my $current_alias = $self->{aliases}->{$name};
            my $counters = $alias->get_counters;
            $current_alias->add_counter($_) for (@$counters);
        }
    }
}

sub update_vid {
    my ($self) = @_;
    # Update vid
    $self->{current_vid} = _make_vid();
    # Keep track of all vids produced by this object
    $self->{vids}->{$self->{current_vid}} = 1;
}

# Returns tracked counters matching regexes
sub find {
    my $self = shift;
    my %params = validate(
                           @_,
                           {
                               counters => {type => ARRAYREF},
                               counters_as_strings => { type => SCALAR | UNDEF, optional=>1 },
                               source   => {type=> SCALAR | UNDEF, optional=>1 },
                           }
                        );
    my @counters;
    my $handlers = $self->_get_handlers( source => delete $params{source} );
    confess "Unable to find handlers" if (!$handlers || !@$handlers);

    for my $c (@{$params{counters}}){
        my ($node_r, $counter_r) = $self->_parse_counter($c);
        for my $handler (@$handlers){
            next if ($handler->get_node !~ /^$node_r$/);
            push @counters, @{$handler->find(   counters=>[$counter_r],
                                                counters_as_strings=>$params{counters_as_strings},
                                            )};
        }
    }
    return \@counters;
}

# Sort an arrayref of samples in chronological order (older first)
sub _sort {
    my $self = shift;
    my $samples = shift;
    my @sorted = sort { $a->{_time} <=> $b->{_time} } @$samples;
    return \@sorted;
}

=head2 delta

    Computes the difference between samples passed as parameters or kept
    internally. Exact behavior described below:

    if more than two samples are passed as arguments, returns an arrayref of hashes
    of the difference between chronologically consecutive samples.
    if two samples are passed, returns a hash of the difference
    between the two samples.
    if one sample is passed, a second one is created now and the difference is
    returned as a hash.
    if no samples are passed, a sample is taken now and compared to the last
    sample taken during 'checkpoint' or 'delta'. See examples.

    Arguments:
        $samp1       - first sample. Optional.
        $samp2       - second sample. Optional.
        ...etc

    Options:
        as_string    - Returns a string describing the delta instead of a hash
        hide_zeros   - Counters that have not changed are not included in delta
                       Set to 1 by default.
        exact_format - Returns delta in exact format for average and percent counters,
                       i.e. returns the string '1024/3' rather than the number '341.33'
                       This is the default format for the 'stats' command and therefore
                       the default for this module as well.
        raw_counter  - Counters that are only digits will be included in delta
                       Set to 0 by default.

    Examples:
        1. $before = $stats->sample;
           .. do something ..
           $after = $stats->sample;
           $delta = $stats->delta($before, $after);

        2. $stats->checkpoint;
           ... do something ...
           $delta = $stats->delta;
           ... do somethign else ...
           $delta = $stats->delta;

        3. while (...){
               push @samples, $stats->sample;
               ... do something ...
           }
           $deltas = $stats->delta(@samples);

=cut

sub delta {
    my $self = shift;
    my @samples;
    while ($_[0] && ref($_[0]) eq 'HASH' && $_[0]->{_vid}){
        push @samples, shift;
    }
    my %params = @_;
    my $hide_zeros = $params{hide_zeros};
    $hide_zeros = 1 if (!defined $hide_zeros);
    my $as_string = $params{'as_string'};
    $as_string = 0 if (!defined $as_string);
    my $exact_format = $params{'exact_format'};
    $exact_format = 1 if (!defined($exact_format));
    my $already_sorted = $params{'already_sorted'};
    $already_sorted = 0 if (!defined($already_sorted));
    my $raw_counter = $params{'raw_counter'};
    $raw_counter = 0 if (!defined($raw_counter));

    # If only one sample was passed, get a new one.
    if (scalar @samples == 1) {
        push @samples, $self->sample();
    }

    #
    # If no samples were passed, use the checkpoint one
    # and get a new one
    #
    if (scalar @samples < 2){
        if (!$self->{last_sample}){
            confess("You didn't pass enough samples or forgot to call 'checkpoint' before calling 'delta'");
        }
        else{
            push @samples,$self->{last_sample};
        }
        if (scalar @samples < 2){
            push @samples,$self->sample;
            $self->{last_sample} = $samples[-1];
        }
    }
    my $vid = $samples[0]->{_vid};
    my $message = "The subroutine 'delta' can only compare samples "
                    ."created by the same parent in the same state";
    map { confess($message) if ($_->{_vid} ne $vid )} @samples;
    if (!$self->{vids}->{$vid}){
        confess("Samples were not created by this instance");
    }

    # Sort samples in chronological order (older first)
    my $sorted_samples = \@samples;
    if(!$already_sorted){
        $sorted_samples = $self->_sort(\@samples);
    }
    my $lastidx = scalar(@$sorted_samples) - 1;

    # Compute difference between successive samples
    my @deltas;
    my $i = 0;
    while ($i < $lastidx){

        my $older_samp = $sorted_samples->[$i];
        my $newer_samp = $sorted_samples->[$i + 1];
        my $elapsed = $newer_samp->{_time} - $older_samp->{_time};

        my $delta = {
            _vid => $older_samp->{_vid},
            _time => $newer_samp->{_time},
            _elapsed => $elapsed,
        };

        my $nodes = $self->get_nodes;

        # Get all counters from both samples
        # The following code should not be necessary since samp1 and samp2
        # should have exactly the same keys to begin with
        my @keys_a = uniq(keys %{$older_samp},keys %{$newer_samp});

        foreach my $key (@keys_a) {
            next if ($key =~ /^_/);

            # Find out which entry can compute the delta
            # Remember that internally everything is stored
            # with the node name
            my $entry = $self->_get_entry($key);
            confess "No entry found for '$key'?!" if (!$entry);
            my $entry_delta = $entry->delta(    old        => $older_samp->{$key},
                                                new        => $newer_samp->{$key},
                                                raw_counter=> $raw_counter,
                                                exact_format=>$exact_format,
                                            );
            if ($entry_delta || !$hide_zeros){
                $delta->{$key} = $entry_delta;
            }
        }
        push @deltas, $delta;
        $i ++;
    }

    if ($as_string){
        return join("\n", map { _as_string($_) } @deltas);
    }
    else{
        # Maintain backwards compatibility
        if (scalar @deltas > 1){
            return \@deltas;
        }
        else{
            return $deltas[0];
        }
    }
}


sub _get_entry {
    my ($self, $name) = @_;
    confess "You must pass a name" if (!$name);
    my $entry = $self->{aliases}->{$name};
    return $entry if ($entry);
    if (!$self->is_multinode){
        my $node = $self->get_node;
        $entry = $self->{entries}->{"$node:$name"};
        #its possible that the entry could be a cluster wide counter as well
        #and be stored under $cluster_name:$name
        if(!$entry){
            my $cluster_name = $self->{cluster_name};
            $entry = $self->{entries}->{"$cluster_name:$name"};
        }
    }
    else{
        $entry = $self->{entries}->{$name};
    }
    return $entry;
}

=head2 alias_rename

    Rename an alias

    Arguments:
        current         - current alias name
        new             - new name for the alias

=cut

sub alias_rename {
    my $self = shift;
    my %params = validate(
                            @_,
                            {
                                current => {type => SCALAR },
                                new     => {type => SCALAR },
                            }
                       );
    my $old_name = $params{current};
    my $new_name = $params{new};
    my $alias = $self->{aliases}->{$old_name};
    confess "No alias named '$old_name'" if (!$alias);

    my $source = delete $self->{entries}->{$old_name};
    delete $self->{aliases}->{$old_name};

    $alias->set_name($new_name);

    $self->{entries}->{$new_name} = $source;
    $self->{aliases}->{$new_name} = $alias;

    # Update the named alias factories. This is most likely what the user expects
    my $factory = delete $self->{alias_factories}->{$old_name};
    if ($factory){
        $factory->set_name($new_name);
        $self->{alias_factories}->{$new_name} = $factory;
    }
    return;
}


=head2 alias_remove

    Remove an alias definition

    Arguments:
        name            - name of the alias to remove


=cut

sub alias_remove {
    my ($self,%params) =@_;
    my $alias_to_be_removed = $params{name};
    confess("You must specify a name") if (!$params{name});
    delete $self->{aliases}->{$alias_to_be_removed};
    delete $self->{entries}->{$alias_to_be_removed};

    # Delete the named alias factories. This is most likely what the user expects
    delete $self->{alias_factories}->{$alias_to_be_removed};

    return;
}

=head2 alias_list

    Returns an arrayref of all the aliases present in the system with their definitions

=cut

sub alias_list {
    my $self = shift;
    # Don't hand out our hash, make a shallow copy to avoid side-effects
    return [values %{$self->{aliases}}];
}

sub alias_get {
    my ($self, $alias_name) = @_;
    confess "You must pass a name" if (!$alias_name);
    my $alias = $self->{aliases}->{$alias_name};
    confess "No alias named '$alias_name' found" if (!$alias);
    return $alias;
}

sub alias_exists {
    my ($self, $alias_name) = @_;
    confess "You must pass a name" if (!$alias_name);
    my $alias = $self->{aliases}->{$alias_name};
    return $alias?1:0;
}

sub entry_display_name {
    my $self = shift;
    my %args = validate (   @_,
                            {
                                node  => { type => SCALAR | UNDEF, optional => 1 },
                                name  => { type => SCALAR | UNDEF, optional => 1 },
                                entry => { type => OBJECT | UNDEF, optional => 1, isa=>'Perf::Stats::Entry' },
                            }

                        );
    my $entry = $args{entry};
    my $multinode = $self->is_multinode;
    if ($entry){
        return $entry->get_full_name if ($multinode);
        return $entry->get_name;
    }
    return "$args{node}:$args{name}" if ($multinode);
    return $args{name};
}

=head2 alias_instance

    Automatically creates aliases for counters from a same node whose instances are related in some
    user-defined way. For example you may want to group together disk or volume counters by aggregate.
    This only works for 'stats' or 'statistics' counters. More than one attribute may be specified.
    Returns an arrayref of all aliases created or updated

    Arguments:
        counters        - arrayref of counter names or regular expressions. All counters must belong to
                          the same stats object (like 'disk' or 'volume')
        group_by        - arrayref of default attribute names or coderefs that can be used to get attribute
                          values. Default attributes are defined in Stats::InstanceInfo::<object>.
        source          - Required if you've added counters from multiple sources

    Options:
        hide_counters - by default, individual 'aliased' counters are not shown.
                        This flag overrides this behavior.
        dynamic       - Specifies whether aliases should be updated when counters get added.
                        Defaults to 0.

    Examples:
        Group all currently tracked disk counters by aggregate:
        $stats->alias_instance(counters=>[{object => 'disk'}], group_by=>['aggregate']);

        Group all volume counters by some custom attribute:
        $attribute_getter = sub { $instance = shift; ... ; return $value };
        $stats->alias_instance(counters=>[{object => 'volume'}], group_by=>[$attribute_getter]);

        Group disk counters by aggregate and raid role (parity/data):
        $stats->alias_instance(counters=>[{object => 'disk'}], group_by=>['aggregate','raid_role']);

=cut

sub alias_instance {
    my ($self, %params) = @_;
    my $group_by = $params{group_by};
    confess "You must specify a 'group_by' argument" if (!$group_by);
    confess "Argument 'group_by' must be an array ref" if (ref($group_by) ne 'ARRAY');

    # The group_by parameter must be a list of scalars or coderefs
    for my $group_by_i (@$group_by){
        my $r = ref($group_by_i);
        if ($r && $r ne 'CODE'){
            confess("Argument 'group_by' must be a list containing scalars or coderefs");
        }
    }
    return $self->_generic_alias( %params, type => 'instance' );
}

# For backwards compatibility
sub alias_auto {
    goto &alias_instance;
}

=head2 alias_cluster

    Creates aliases grouping similar counters on every node of a cluster.
    Currenlty only supported for source 'statistics'.

    Arguments:
        counters             - arrayref of counter names or regular expressions

    Options:
        hide_counters        - Individual aliased counters are not shown. Defaults to true.
        skip_string_counters - Skip aliasing on counters of type 'string'
        dynamic              - Specifies whether aliases should be updated when counters get added.
                               Defaults to 0.

    Examples:
        Group all currently tracked system counters under various aliases, one per counter:

            $stats->alias_cluster(counters=>[{node     => '.*',
                                              object   => 'system',
                                              instance => 'system',
                                              counter  => '.*']);

    Dynamic:
        Setting 'dynamic' to 1 will automatically re-call alias_cluster with the current parameters
        on counters being added. Example:

            $stats = Perf::Stats->new;
            $stats->add([{node     => '.*',
                          object   => 'volume',
                          instance => 'vol0',
                          counter  => 'read_latency'}]);

            # Create cluster aliases for all counters matching the regex
            # A single alias will get created called <cluster>:volume:vol0:read_latency
            $stats->alias_cluster( counters=>[{node     => '.*',
                                               object   => 'volume',
                                               instance => 'vol0',
                                               counter  => '*_latency'}], dynamic=>1 );

            # During the add, a new alias called <cluster>:volume:vol0:write_latency
            # will get automatically generated
            $stats->add( [{node     => '.*',
                          object   => 'volume',
                          instance => 'vol0',
                          counter  => 'write_latency'}] );

=cut

sub alias_cluster {
    my ( $self, %params ) = @_;
    return $self->_generic_alias( %params, type => 'cluster' );
}

sub get_nodes {
    my ($self) = @_;
    return $self->{nodes};
}

sub get_node {
    my ($self) = @_;
    confess "Method not valid in multinode" if ($self->is_multinode);
    return $self->{nodes}->[0];
}

sub is_multinode {
    my ($self) = @_;
    if ($self->{force_multinode}) {
        return 1;
    }
    else {
        my $nodes_count = scalar @{$self->{nodes}};
        return $nodes_count>1?1:0;
    }
}

sub get_vid {
    my ($self) = @_;
    return $self->{current_vid};
}

sub _get_handlers {
    my ($self, %args) = @_;
    my $node = $args{node};
    my $source = $args{source};
    my $all = $args{all};
    my @handlers;
    for my $s (keys %{$self->{handlers}}){
        next if ($source and $source ne $s);
        my $source_handlers = $self->{handlers}->{$s};
        if($all) {
            if ($node){
                my $node_handler = $source_handlers->{$node};
                push @handlers, $node_handler if ($node_handler);
            }
            else {
                push @handlers, values %$source_handlers;
            }
        }
        else {
            if ($node){
                my $node_handler = $source_handlers->{$node};
                if($node_handler){
                push @handlers, $node_handler if ($node_handler->is_enabled);
                }
            }
            else{
                for my $h (values %$source_handlers){
                    push @handlers, $h if($h->is_enabled);
                }
            }
        }
    }
    return \@handlers;
}

sub _get_handler {
    my $self = shift;
    my ( $counter, %args ) = validate_mixed(   @_,
                                [ { type => SCALAR } ],
                                {
                                    source              => { type => SCALAR| UNDEF, optional => 1 },
                                    counters_as_strings => { type => SCALAR| UNDEF, optional => 1 },
                                },
                            );

    my $source = $args{source};
    my $as_strings = $args{counters_as_strings};

    # If a 'source' argument is passed in validate it
    confess "Unsupported source '$source'" if ($source && !$HANDLERS{$source});

    my ($node, $counter_r) = $self->_parse_counter($counter);
    my $handlers = $self->_get_handlers(source=>$source, node=>$node);
    if (!$handlers || !@$handlers){
        $Log->comment("No handlers found for source '$source' and node '$node'");
        return;
    }
    my $matched_handler;
    for my $handler (@$handlers){
        my $matches = $handler->find(counters=>[$counter_r],counters_as_strings =>$as_strings);
        if ($matches && @$matches){
            if ($matched_handler){
                confess "More than one handler match. You must pass a 'source' argument";
            }
            return $handler;
        }
    }
    $Log->comment("No handler found for counters '$counter' on node '$node' and source '$source'");
    return;
}

=head2 reconnect

    Reset the connections used by Perf::Stats. Typically
    called after a reboot, etc.

    If a specific node is specified, its connections are
    unconditionally reset. Otherwise, connections for all
    nodes are reset as needed.

=over

=item Arguments:

B<node> = string

I<Optional> node whose connections are to be reset

=back

=cut

sub reconnect {
    my $self   = shift;
    my %params = validate(
        @_,
        {
            node => { type => SCALAR, default => undef },
        },
    );

    my $node     = $params{node};
    my $handlers = $self->_get_handlers;

    for my $handler (@{$handlers}) {
        if (defined $node) {
            if ($handler->get_node eq $node) {
                $handler->reconnect(force => 1);
                last;
            }
        }
        else {
            $handler->reconnect(force => 0);
        }
    }
    return;
}


# Don't keep track of empty Handlers
sub _clean_handlers {
    my $self  = shift();
    my @nodes = @{$self->get_nodes()};
    my $cluster_name = $self->{cluster_name};
    push @nodes,$cluster_name if($cluster_name);

  SOURCE: foreach my $source (keys(%HANDLERS)) {
        my $node_handlers = $self->{handlers}->{$source};

        next SOURCE
            if (0 == keys(%{$node_handlers}));

      NODE: foreach my $node (@nodes) {
            my $handler = $node_handlers->{$node};
            if (defined($handler) && 0 == scalar(@{$handler->tracked_get()})) {
                $Log->trace("Discarding empty Handler for $source on $node");

                delete($self->{handlers}->{$source}->{$node});
            }
        }
    }
    return;
}

=head2 start

    Signals the filer to take the first sample to be used as a beginning reference point.

    Options:
        ready_timeout - Amount of time to wait for the asynchronous calls to statistics start/stop to complete.
                        For objects with lots of data sometimes this can take a very long time to return.
                        Defaults to 20 seconds. Potential risks exist due to timing on multiple nodes being off.
=cut
sub start {
    my $self = shift;
    my %params = validate(
                        @_,
                        {
                            ready_timeout => {type => SCALAR | UNDEF, optional => 1, default => 20},
                        }
                   );
    my $handlers = $self->_get_handlers;
    my @hs;
    $self->{start_time} = time;
    $self->{start} = 1;
    #send start commands in parallel
    for my $handler (@$handlers){
        if ($handler->can('pre_start')){
            $handler->pre_start(ready_timeout => $params{ready_timeout});
        }
        else {
            push @hs,$handler;
        }
    }
    for my $handler (@$handlers){
        my $handler_sample;
        if ($handler->can('post_start')){
            $handler->post_start;
         }
    }
    $self->{first_sample} = $self->sample(\@hs) if (scalar @hs > 0);

    return;
}
=head2 stop

    Takes the second sample and returns a hash containing
    the delta between the two samples.

    Options:
        exact_format - Returns delta in exact format for average and percent counters,
                       i.e. returns the string '1024/3' rather than the number '341.33'
                       This is the default format for the 'stats' command and therefore
                       the default for this module as well. Turning this off will give you the precalculated
                       values from the filer.
=cut
sub stop {
    my $self = shift;
    my %params = validate(
                            @_,
                            {
                                exact_format => {type => SCALAR | UNDEF, optional => 1, default => 1},
                            }
                       );

    my $job = $self->{monitor}->{job};
    my $started = $self->{start};
    my @hs;
    #Ensure backwards compatibility with method name change
    if(!$started && $job) {
        $Log->warn("This method has been depreciated please use 'monitor_stop'");
        return $self->monitor_stop;
    }
    if(!$started) {
        confess("No start command found");
        return;
    }

    my $handlers = $self->_get_handlers;
    for my $handler (@$handlers){
        if($handler->can('pre_stop')) {
            $handler->pre_stop;
        }
        else {
            push @hs,$handler;
        }
    }

    my $stop_time = time;
    my $delta = {};
    $delta->{_vid} = $self->{current_vid};
    $delta->{_time} = $stop_time;
    $delta->{_elapsed} = $stop_time - $self->{start_time};

    for my $handler (@$handlers){
        if ($handler->can('post_stop')) {
            my $handler_sample = $handler->post_stop;

            # TODO: This doesn't go through entry_display_name but it's more efficient.
            # Revise as necessary
            if ($self->is_multinode){
                my $node = $handler->get_node;
                my %noderized = map { +"$node:$_" => $handler_sample->{$_} } keys %$handler_sample;
                $handler_sample = \%noderized;
            }

            # The 'add' method is supposed to ensure that no collisions happen below
            @$delta{ keys %$handler_sample } = values %$handler_sample;
        }
    }
    $delta = $self->_handle_aliases($delta);
    $delta = $self->_get_calculated($delta) if (!$params{exact_format});
    my $second_sample =  $self->sample(\@hs) if (scalar @hs > 0);

    if($self->{first_sample} && $second_sample)
    {
        my $sample = $self->delta($self->{first_sample},$second_sample,
            hide_zeros=>0,exact_format => $params{exact_format});
        @$delta{keys %$sample} = values %$sample;
    }

    undef $self->{start};
    undef $self->{first_sample};
    return $delta;
}
sub _get_calculated{
    my $self = shift;
    my $delta = shift;

    my @keys = uniq(keys %$delta);
    foreach my $key (@keys) {
        next if ($key =~ /^_/);
        my $entry = $self->_get_entry($key);
        confess "No entry found for '$key'?!" if (!$entry);
        if($entry->isa('Perf::Stats::Counter::CM')) {
            $delta->{$key} = $entry->get_calc_value;
        }
        elsif ($entry->isa('Perf::Stats::Alias::CM')) {
            $delta->{$key} = $entry->get_calc_value($delta->{$key});
        }
    }
    return $delta;
}

sub toggle_nodes {
    my $self = shift;
    my %args = validate(
                            @_,
                            {
                                nodes => {type => ARRAYREF},
                                toggle => {type => SCALAR}
                            }
                       );
    my %nodes = map {$_=>1} @{$args{nodes}};
    my $toggle = $args{toggle};
    my $update = 0;
    for my $alias (values %{$self->{aliases}}){
        my $counters = $alias->get_counters;
        for my $c (@$counters){
            my $node = $c->get_node;
            if($nodes{$node} and $toggle == 0){
                confess "Cannot disable node $node while it is in use by an Alias";
            }
        }
    }

    for my $node (keys %nodes) {
        my $handlers = $self->_get_handlers(all=>1,node=>$node);
        for my $handler (@$handlers){
            $Log->trace("Switching $node to position $toggle");
            #ensure node state is actually changing before updating vid.
            if($handler->is_enabled ne $toggle){
                $update = 1;
                $handler->toggle($toggle);
            }
        }
    }
    #if a Cluster handler exists ensure that it's connection
    # is still valid and create a new one if need be.
    my $cluster_name = $self->{cluster_name};
    if($cluster_name){
        for my $source ('statistics','zapi_perf'){
            if($self->{handlers}->{$source}->{$cluster_name}){
                my $handler = $self->{handlers}->{$source}->{$cluster_name};
                $handler->toggle(toggle=>$toggle, nodes=>\%nodes);
            }
        }
    }
    #update vid to invalidate previous samples/checkpoints.
    #delta will not be able to function unless filer is in same state
    #for all samples to be used in the delta call.
    if($update){
        $self->update_vid;
    }

}

=head2 reenable_collection

    Re-enables the collection of statistics for the specified nodes.
    All nodes are enabled by default.

    Important: All previous samples/checkpoints will be invalid for use with
        $stats->delta. Suggestion would be to take a new checkpoint after calling
        this method if a delta is needed to be taken during this filer state.

    Arguments:
        nodes    - arrayref of nodes to toggle.

    Example:
        #enables nodes for statistics collection
        $stat->enable_collection( nodes=> ['vsim1']);

=cut
sub reenable_collection  {
    my $self = shift;
    my %args = validate(
                            @_,
                            {
                                nodes => {type => ARRAYREF},
                            }
                       );
    $self->toggle_nodes(nodes=>$args{nodes}, toggle=>1);

}

=head2 disable_collection

    Disables the collection of statistics for the specified nodes.
    To be used when it is known that a node will be going down.

    Important: All previous samples/checkpoints will be invalid for use with
        $stats->delta. Suggestion would be to take a new checkpoint after calling
        this method if a delta is needed to be taken during this filer state.

    Arguments:
        nodes    - arrayref of nodes to toggle.

    Limitations:
        - Must be called before node goes down.
        - The use of aliases is limited to counters on enabled nodes. This method
          will not work if aliases are present on nodes being disabled.

    Example:
        #disables nodes for statistics collection
        $stat->disable_collection( nodes=> ['vsim1']);

=cut
sub disable_collection {
    my $self = shift;
    my %args = validate(
                            @_,
                            {
                                nodes => {type => ARRAYREF},
                            }
                       );
    $self->toggle_nodes(nodes=>$args{nodes}, toggle=>0);

}

=head2 get_counter_types

    Returns the unit for all counters currently being tracked by
    Perf::Stats as a hash. Returns undef for counters with
    unknown units. The units of aliases will not return.

=cut
sub get_counter_units {
    my $self = shift;
    my $handlers = $self->_get_handlers(all=>1);
    my %counters;
    for my $handler (@$handlers){
        my $counters = $handler->tracked_get;
        for my $c (@$counters){
            $counters{$c->get_full_name} = $c->get_unit;
        }
    }
    return \%counters;
}

1;
