# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Base class for Stats Alias
## @status   review
## @pod here

package Perf::Stats::Alias;{
    use strict;
    use warnings;
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Carp qw( confess );
    use Perf::Stats::Entry;
    use Object::InsideOut qw( Perf::Stats::Entry );
    use Perf::Stats::Alias::CM;

    my @counters    :Field  :Arg(Name=>'counters', Mandatory=>1) :Type('ARRAY_ref(Perf::Stats::Counter)');

    sub new {
        my $class = shift;
        if ($class ne __PACKAGE__){
            return $class->Object::InsideOut::new(@_);
        }
        my %args = @_;
        confess "You must specify counters" if (!$args{counters} || !@{$args{counters}});
        my $is_cm = grep { $_->isa('Perf::Stats::Counter::CM') } @{$args{counters}};
        if ($is_cm){
            return Perf::Stats::Alias::CM->new(%args);
        }
        return $class->Object::InsideOut::new(@_);
    }

    sub _init :Init {
        my ($self) = @_;
        my $alias_types = $counters[$$self]->[0]->get_all_types;

        if ( grep { $_->get_all_types ne $alias_types } @{$counters[$$self]} ){
            my @mismatches =
              map { $_->get_name . ": " . $_->get_all_types() }
              @{$counters[$$self]};
            my $str = join("\n", @mismatches);
            confess("Counters must all be of the same type!\n" . $str);
        }
        $self->set_type($alias_types);

        # Store the list of counters as a hash instead
        my $counters_arr = delete $counters[$$self];

        $self->add_counter($_) foreach (@$counters_arr);
    }

    sub compute {
        my ($self) = @_;
        my $value = 0;
        for my $counter (values %{$counters[$$self]}){
            $value += $counter->get_value;
        }
        return $value;
    }

    sub remove_counter {
        my ($self, $counter) = @_;
        confess "Invalid argument"
            if (!$counter || !ref($counter) || !$counter->isa('Perf::Stats::Counter'));
        my $counter_name = $counter->get_full_name;
        my $removed = delete $counters[$$self]->{$counter_name};
        confess "Unable to find counter '$counter_name' for removal" if (!$removed);
        return $removed;
    }

    sub is_empty {
        my ($self) = @_;
        my $counters_count = scalar @{$self->get_counters};
        return $counters_count?0:1;
    }

    sub get_counters {
        my ($self) = @_;
        return [values %{$counters[$$self]}];
    }

    sub add_counter {
        my ($self,$counter) = @_;
        confess "Invalid argument"
            if (!$counter || !ref($counter) || !$counter->isa('Perf::Stats::Counter'));
        my $counter_name = $counter->get_full_name;
        confess "Counter '$counter_name' already present in alias '" . $self->get_name ."'"
            if ($counters[$$self]->{$counter_name});
        $counters[$$self]->{$counter_name} = $counter;
        $counter->add_alias($self);
        return;
    }

}
1;
