# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Base class of all Alias factories
## @status   review
## @pod here

package Perf::Stats::AliasFactory;{
    use strict;
    use warnings;
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Carp qw( confess );
    use Params::Validate qw( validate_pos ARRAYREF );
    use List::MoreUtils qw( any );
    use Object::InsideOut;
    use Perf::Stats::AliasFactory::Name;
    use Perf::Stats::AliasFactory::Cluster;
    use Perf::Stats::AliasFactory::Instance;

    my %FACTORIES = (   'name'      => 'Perf::Stats::AliasFactory::Name',
                        'cluster'   => 'Perf::Stats::AliasFactory::Cluster',
                        'instance'  => 'Perf::Stats::AliasFactory::Instance',
                    );

    my @name                    :Field :Arg(Name=>'name',           Mandatory=>0 ) :Std('name');
    my @counters                :Field :Arg(Name=>'counters',       Mandatory=>1);
    my @hide_counters           :Field :Arg(Name=>'hide_counters',  Mandatory=>1) :Get('does_hide_counters');
    my @skip_string_counters    :Field :Arg(Name=>'skip_string_counters', Mandatory=>0) :Default(0);
    my @counters_as_strings     :Field :Arg(Name=>'counters_as_strings', Mandatory=>0) :Default(0);
    my @parent                  :Field :Arg(Name=>'parent', Mandatory=>1) :Weak :Get('get_parent') :Type('Perf::Stats');
    my @source                  :Field :Arg(Name=>'source',       Mandatory=>1) :Get('get_source_name');

    sub new {
        my $class = shift;
        if ($class ne __PACKAGE__){
            return $class->Object::InsideOut::new(@_);
        }
        my %args = @_;
        my $type = delete $args{type};
        confess "You must specify a type" if (!$type);
        $class = $FACTORIES{$type};
        confess "No factory class found for type '$type'" if (!$class);
        return $class->new( %args );
    }

    sub _init :Init {
        my ($self) = @_;
        $name[$$self] = time . '_' . int(rand(2**32)) if (!$name[$$self]);
        $Log->trace("Instantiating Alias Factory:\n" . Dumper($self->dump));
    }

    sub get_handlers {
        my ($self, $node) = @_;
        return $parent[$$self]->_get_handlers( source => $source[$$self], node => $node );
    }

    sub get_current_vid {
        my ($self) = @_;
        return $parent[$$self]->get_vid;
    }

    sub get_handler {
        my ($self, $node) = @_;
        my $handlers = $self->get_handlers($node);
        if (@$handlers == 1){
            return $handlers->[0];
        }
        confess "There is more than one handler for source '$source[$$self]'";
    }

    sub match_counters {
        my ($self,$counters_to_match) = @_;
        my @matching_counters;

      COUNTER: foreach my $c (@$counters_to_match){
            confess "Invalid argument"
              if (!$c || !ref($c) || !$c->isa('Perf::Stats::Counter'));

            next COUNTER
                if ($skip_string_counters[$$self] && $c->is_of_type('string'));

            if (!$counters_as_strings[$$self]) {
                my $counter_node = $c->get_node();
                my $handler      = $self->get_handler($counter_node);

                # Handle string vs. hashref counters
              REGEXP: foreach my $re (@{$counters[$$self]}) {
                    my ($node_re, $counter_re) =
                      $parent[$$self]->_parse_counter($re);

                    if ($counter_node =~ /^$node_re$/
                        && $handler->do_match(counter => $c,
                                              regex   => $counter_re)
                      ) {
                        push(@matching_counters, $c);
                        last REGEXP;
                    }
                }
            }
            else {
                my $c_name = $c->get_full_name();
                push(@matching_counters, $c)
                  if (any { $_ eq $c_name } @{$counters[$$self]});
            }
        }
        return \@matching_counters;
    }

    # Returns an array ref of aliases
    sub process_aliases {
        confess "You must implement this method in a subclass";
    }

    sub process {
        my ($self, $counters) = @_;
        if (!$counters || ref($counters) ne 'ARRAY' || !@$counters){
            confess "Invalid first argument";
        }
        my $matching_counters = $self->match_counters($counters);
        return if (!$matching_counters || !@$matching_counters);

        $Log->trace("Found ". @$matching_counters . " matching counters:\n" .
                       join("\n", map {$_->get_full_name} @$matching_counters) . "\n" .
                       "for factory:\n" . Dumper($self->dump));

        if ($hide_counters[$$self]){
            $_->hide for ( @$matching_counters );
        }
        my $aliases = $self->process_aliases($matching_counters);
        confess "Wrong output from process_aliases" if (!$aliases || ref($aliases) ne 'ARRAY');

        return $aliases;
    }

}
1;
