# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Class responsible for generating instance aliases
## @status   review
## @pod here

package Perf::Stats::AliasFactory::Instance;{
    use strict;
    use warnings;
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use Carp qw( confess );
    use Perf::Stats::InstanceInfo;
    use Perf::Stats::AliasFactory;
    use Object::InsideOut qw( Perf::Stats::AliasFactory );

    my @group_by        :Field :Arg(Name=>'group_by',      Mandatory=>1) :Type('ARRAYref');
    my @instance_info   :Field :Name('instance_info');

    sub _init :Init {
        my ($self) = @_;
        my $source = $self->get_source_name;
        if ($source !~ /^stat/ && $source ne 'cm_client' && $source ne 'zapi_perf'){
            confess "Not supported on source '$source'";
        }
    }

    sub alias_groups {
        my ($self, $matching_counters) = @_;
        if (!$matching_counters || ref($matching_counters) ne 'ARRAY'){
            confess "Invalid argument";
        }

        # Do some validation and sort counters by node.
        # Instance aliases only apply to a single node
        my %node_counters;
        my $object;
        for my $counter (@$matching_counters){
            if (!ref($counter) || !$counter->isa("Perf::Stats::Counter")){
                $Log->comment("Invalid input:\n" . Dumper($counter));
                confess "Invalid input";
            }
            my $n = $counter->get_node;
            push @{$node_counters{$n}}, $counter;
            my $counter_object = $counter->get_cm_object;
            if (!$object){
                $object = $counter_object
            }
            elsif($object ne $counter_object){
                confess("Object mismatch: '$counter_object' instead of '$object'." .
                                "The group subroutine only works with counters belonging to " .
                                "the same object (like aggregate or volume or disk)");
            }
        }

        my $groups;
        for my $node ( keys %node_counters ){
            my $counters = $node_counters{$node};
            # Sort counters by instances
            my $instances = {};
            for my $counter (@$counters){
                my $counter_instance = $counter->get_cm_instance;
                push @{$instances->{$counter_instance}}, $counter;
            }

            # Start with a single group
            my $node_groups = {root=>$instances};
            for my $attribute (@{$group_by[$$self]}){
                my $named_attribute = 0;
                my $getter;
                #
                # Create the getter coderef for that attribute
                # Use Perf::Stats::InstanceInfo for scalars
                #
                my $ref = ref($attribute);
                if (!$ref){
                    $named_attribute = 1;
                    my $info = $instance_info[$$self]->{$node}->{$object};
                    if (!$info){
                        $info =
                          Perf::Stats::InstanceInfo->new(
                                           $object,
                                           handler => $self->get_handler($node),
                                           node    => $node,);
                        $instance_info[$$self]->{$node}->{$object} = $info;
                    }
                    if (!$info->can($attribute)){
                        confess("Group attribute '$attribute' is not a method of " . ref($info));
                    }
                    $getter = sub { $info->$attribute(shift) };
                }
                elsif($ref eq 'CODE'){
                    $getter = $attribute;
                }
                else{
                    Carp::confess 0;
                }
                #
                # Turn each group into subgroups based on instance attribute values
                #
                my @group_names = keys %$node_groups;
                for my $group_name (@group_names){
                    my $group_instances = delete $node_groups->{$group_name};
                    for my $instance (keys %$group_instances){
                        my $instance_counters = $group_instances->{$instance};
                        my $value;
                        eval {$value = $getter->($instance)};
                        if ($@ || !defined($value)){
                            if ($named_attribute){
                                confess("Unable to get attribute '$attribute' for $object '$instance': $@");
                            }
                            else{
                                confess("Unable to get unamed attribute for $object '$instance': $@");
                            }
                        }
                        # Create subgroup name by appending the value to the group name
                        my $subgroup_name = $group_name eq 'root'?"$object:$value":"${group_name}_$value";
                        $node_groups->{$subgroup_name}->{$instance} = $instance_counters;
                    }
                }
            }

            # Flatten the last level (instance name)
            for my $group ( keys %$node_groups ){
                push @{$groups->{$node}->{$group}}, map {@$_} values %{$node_groups->{$group}};
            }

            # Subgroup by cm_name
            for my $group ( keys %$node_groups ){
                my $group_counters = delete $groups->{$node}->{$group};
                for my $counter (@$group_counters){
                    my $counter_cm_name = $counter->get_cm_name;
                    push @{$groups->{$node}->{"$group:$counter_cm_name"}}, $counter;
                }
            }
        }
        return $groups;
    }

    sub process_aliases {
        my ($self, $counters) = @_;
        my @aliases;
        my $groups = $self->alias_groups( $counters );

        $Log->trace("TRACE Groups:\n" . Dumper($groups));

        my $hide_counters = $self->does_hide_counters;
        my $source_name = $self->get_source_name;
        my $parent = $self->get_parent;
        for my $node ( keys %$groups ){
            my $node_groups = $groups->{$node};
            for my $group ( keys %$node_groups ){
                my $alias_name = $parent->entry_display_name(   node => $node,
                                                                name => $group,
                                                             );
                my $alias_counters = $node_groups->{$group};
                my $alias = Perf::Stats::Alias->new(  name => $alias_name,
                                                counters => $alias_counters,
                                              );
                push @aliases, $alias;
            }
        }
        # Don't keep the Instance::Info cached as the state on the filer may change
        # between invocations
        delete $instance_info[$$self];

        return \@aliases;
    }

}
1;
