# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Base class for Stats Counter
## @status   review
## @pod here

package Perf::Stats::Counter::CM;{
    use strict;
    use warnings;
    use Tharn qw($Log);
    use Carp qw(confess);
    use Data::Dumper;
    use Params::Validate qw(validate SCALAR ARRAYREF UNDEF);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::Stats::Counter;
    use Perf::Stats::Entry::CM;
    use Object::InsideOut qw(Perf::Stats::Entry::CM Perf::Stats::Counter);

    my @cm_name     :Field :Name('cm_name')      :Get('get_cm_name');
    my @cm_object   :Field :Name('cm_object')    :Get('get_cm_object');
    my @cm_instance :Field :Name('cm_instance')  :Get('get_cm_instance');
    my @cm_label    :Field :Name('cm_label')     :Get('get_cm_label');
    my @sum         :Field :Set(set_sum)    :Get(get_sum);
    my @count       :Field :Set(set_count)  :Get(get_count);
    my @calc_value  :Field :Name('calc_value')  :Std(calc_value);

    my %init_args :InitArgs = (
        spec => {
            Mandatory => 1,
            Type => 'HASH_ref',
        },
    );


    sub _preinit :PreInit {
        my ($self, $args) = @_;

        $args->{name} = q{};
    }


    sub _init :Init {
        my ($self, $args) = @_;

        my @spec = %{$args->{spec}};

        eval {
            validate(@spec,
                     {object   => {type => SCALAR},
                      instance => {type => SCALAR},
                      counter  => {type => SCALAR},
                      label    => {type => SCALAR},
                      prop     => {type => SCALAR | ARRAYREF},
                      unit    => {type => SCALAR | UNDEF},});
        };
        if ($@) {
            confess(  "Error initializing counter:\n"
                    . Dumper($args->{spec})
                    . "Due to:\n$@");
        }

        my $type = $args->{spec}->{prop};

        if (ref($type)) {
            confess("Unsupported reference type: '" . ref($type) . "'")
              if (ref($type) ne 'ARRAY');
            my @known_types =
              grep { /^delta|raw|average|percent|rate|string$/ } @$type;
            if (!@known_types) {
                confess("No known types found in @$type");
            }

            # That's the format understood by Perf::Stats::Counter
            $type = join(',', @$type);
        }
        else {
            # Not sure what to do with counters with no 'prop'. Set
            # them to 'raw' for now.
            $type = 'raw' if (!$type || $type eq '-');
        }

        $cm_object[$$self]   = $args->{spec}->{object};
        $cm_instance[$$self] = $args->{spec}->{instance};
        $cm_name[$$self]     = $args->{spec}->{counter};
        $cm_label[$$self]    = $args->{spec}->{label};

        $self->set_name($self->get_name());
        $self->set_type($type);
        $self->set_unit($args->{spec}->{unit});
    }

    sub get_zero_value {
        my ($self) = @_;
        if ($self->is_of_type('percent') || $self->is_of_type('average')){
            return '0/0';
        }
        return 0;
    }

    sub get_name {
        my $self = shift();

        my $name = "$cm_object[$$self]:$cm_instance[$$self]:$cm_name[$$self]";

        if (defined($cm_label[$$self]) && q{} ne $cm_label[$$self]) {
            $name .= ".$cm_label[$$self]";
        }

        return $name;
    }


    sub get_field {
        my $self   = shift();
        my $field  = shift();
        my $method;

        if ('counter' eq $field) {
            $field = 'name';
        }


        if ('node' eq $field) {
            $method = "get_node";
        }
        else {
            $method = "get_cm_$field";
        }

        confess("No such field: $field")
            if (!$self->can($method));

        return $self->$method();
    }
}
1;
