# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Base class for Stats Counter and Stats Alias
## @status   review
## @pod here

package Perf::Stats::Entry;{
    use strict;
    use warnings;
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Carp qw(confess);
    use Params::Validate qw(validate SCALAR UNDEF);
    use Scalar::Util qw(looks_like_number);
    use Object::InsideOut;

    my @name        :Field  :Arg(Name=>'name', Mandatory=>1) :Get('get_name') :Set('set_name');
    my @type        :Field  :Arg(Name=>'type', Mandatory=>0);
    my @node        :Field  :Arg(Name=>'node', Mandatory=>0) :Std(node);
    my @unit        :Field :Arg(Name=>'unit', Mandatory=>0) :Std(unit);

    sub get_full_name {
        my ($self) = @_;
        return join(':', $node[$$self], $name[$$self]);
    }

    sub override_types {
        my $self = shift();
        my $type = shift();

        $type[$$self] = { $type => 1 };
    }

    sub set_type {
        my ($self, $type_string) = @_;

        return if (!$type_string);

        # Input may be comma separated string
        my @types = split(/\s*,\s*/, $type_string);
        my $type_set = 0;
        for my $t (@types){
            $t =~ s/^\s+//;
            $t =~ s/\s+$//;

            $type[$$self]->{$t} = 1;
        }
    }

    sub get_all_types {
        my ($self) = @_;
        return join(',', keys %{$type[$$self]});
    }

    sub has_type {
        my ($self) = @_;
        return scalar keys %{$type[$$self]};
    }

    sub is_of_type {
        my ($self, $of_type) = @_;
        confess "Missing argument" if (!$of_type);
        return defined $type[$$self]->{$of_type};
    }

    sub no_zero_values {
        my ($self) = @_;
        return $self->is_of_type('no-zero-values');
    }

    sub get_zero_value {
        return 0;
    }

    sub delta {
        my $self = shift;
        my %params = validate ( @_,
                                {
                                    old         => { type => SCALAR },
                                    new         => { type => SCALAR },
                                    raw_counter => { type => SCALAR },
                                    exact_format=> { type => SCALAR },
                                },
                              );
        my $self_name = $self->get_name;
        my $self_type = $self->get_all_types;
        if ($self_type =~ /delta/){
            if (! looks_like_number($params{new}) ) {
                confess("This should not happen.\nCounter: $self_name"
                           ."\nValue: $params{new}"
                           ."\nType: " . $self->get_source);
            }
            if (! looks_like_number($params{old}) ){
                confess("This should not happen.\nCounter: $self_name"
                           ."\nValue: $params{old}"
                           ."\nType: " . $self->get_source);
            }
            return $params{new} - $params{old};
        }
        elsif($self_type =~ /raw/){
            if (!$params{raw_counter}){
                return $params{new};
            }
            else{
                my $newer_samp = $params{new};
                my $older_samp = $params{old};
                if (($newer_samp =~ /^\d+\.?\d*$/) && ($older_samp =~ /^\d+\.?\d*$/)) {
                    return $newer_samp - $older_samp;
                }
                else{
                    return $newer_samp;
                }
            }
        }
        else{
            confess("Unknown type: $self_type");
        }
    }


    # Splits a counter regexp on the counter level separator ':' and
    # returns the head and tail.  This routine is necessary since the
    # counter level separator is also a valid regexp operator character.
    # E.x. passing in:
    #  '(?:nodeA|nodeB):(?:object1|object2):instance[[:digit:]]:counter'
    # and then calling again on the tail would successively yield:
    #  ('(?:nodeA|nodeB)', '(?:object1|object2):instance[[:digit:]]:counter')
    #  ('(?:object1|object2)', 'instance[[:digit:]]:counter')
    #  ('instance[[:digit:]]', 'counter')

    sub split_regexp {
        my $counter = shift();

        # Handle `:' separated disk UUIDs using code found in
        # lib/NACL/APISet/Node/CLI/7Mode/default/stats.pm
        my $re_hex_word    = "[0-9A-F]{8}";
        my $re_uuid        = join( ":", ($re_hex_word) x 10 );

        if ($counter =~ /^($re_uuid):(.*)$/) {
            return ($1, $2);
        }
        else {
            my @pieces = split(/:/, $counter);
            my $pattern = shift(@pieces);

          PIECES: while (0 < scalar(@pieces)) {
                eval { qr/$pattern/; };
                if (!$@) {
                    my $rest = join(':', @pieces);
                    return ($pattern, $rest);
                }
                else {
                    $pattern .= ':' . shift(@pieces);
                }
            }
        }

        confess("Unable to parse '$counter'");
    }
}
1;
