# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Base class for Handlers. Support for sources 'stats', 'wafl_stats', etc.. is provided
##           as subclasses of this class.
## @status   review
## @pod here

package Perf::Stats::Handler;{

    use strict;
    use warnings;
    use Carp qw( confess );
    use Tharn qw($Log host connect);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use List::MoreUtils qw(uniq any);
    use List::Util qw( first );
    use Params::Validate qw(validate SCALAR HASHREF BOOLEAN ARRAYREF OBJECT UNDEF CODEREF);
    use Module::Common qw(validate_mixed);
    use Data::Dumper;
    use Object::InsideOut;
    use Aggregate qw();
    use filer qw();
    use IO::Compress::Gzip qw(gzip $GzipError) ;
    use File::Path qw(make_path);

    my @tracked         :Field :Name('tracked') :Get('_get_tracked');
    my @current_vid     :Field :Name('current_vid');
    my @cache           :Field :Name('cache') :Get('cache_get');
    my @preprocessors   :Field :Name('preprocessors') :Std('preprocessors');
    my @node            :Field :Arg(Name=>'node',       Mandatory=>1) :Std('node');
    my @connection      :Field :Std('connection');
    my @node_is_ng      :Field :Name('node_is_ng')  :Get('node_is_ng');
    my @source          :Field :Arg(Name=>'source', Mandatory=>1) :Get('get_source');
    my @enabled         :Field :Name('enabled') :Get('is_enabled') :Set('set_enabled');
    my @external_conn   :Field :Name('external_conn');
    my @pid             :Field :Name('pid') : Get('get_pid');

    my %init_args :InitArgs = (apiset => {Req     => 0,
                                          Default => undef,
                                          Type    => 'NACL::APISet'},);


    sub _init :Init {
        my ($self, $args) = @_;
        $tracked[$$self]        = {};
        $preprocessors[$$self]  = [];

        my $hostrec  = host($node[$$self]);
        my $hosttype = $hostrec->hosttype;
        $node_is_ng[$$self] = $hosttype =~ /-ng-/ ? 1 : 0;
        $enabled[$$self] = 1;

        my $conn = $args->{apiset};

        if (defined($conn)) {
             confess(ref($conn) . ' is not an acceptable APISet')
                 if (!$self->acceptable_apiset($conn));
            $external_conn[$$self] = 1;
        }
        else {
            $Log->trace("Creating connection for node '$node[$$self]'");
            $conn = $self->create_connection($hostrec);
        }

        $connection[$$self] = $conn;
        local $Data::Dumper::Maxdepth = 4;
        $Log->trace("Instantiated handler:\n" . Dumper($self->dump));
        $pid[$$self] = $$;
    }

    sub is_node_scoped {
        return 1;
    }

    sub get_sample_query {
        confess "You must implement this method in a subclass";
    }

    sub tracked_is {
        my ($self, $counter) = @_;

        confess('You must pass an argument')
            if (!defined($counter));

        return defined($self->tracked_get($counter));
    }


    sub tracked_get {
        my ($self, $counter) = @_;
        if ($counter) {
            return $tracked[$$self]->{$counter};
        }
        else {
            my @tracked = values(%{$tracked[$$self]});
            return \@tracked;
        }
    }


    sub tracked_add {
        my ($self, $counter) = @_;
        confess("You must specify an argument") if (!$counter);
        confess("Argument must be a Perf::Stats::Counter")
            if (!ref($counter) || !$counter->isa('Perf::Stats::Counter'));
        $tracked[$$self]->{$counter->get_name} = $counter;
        $Log->trace("Now tracking '" . $self->get_source . "' counter: " . $counter->get_name);
        return $counter;
    }

    sub tracked_remove {
        my ($self, $counter) = @_;
        confess("You must specify an argument") if (!$counter);
        confess("Argument must be a Perf::Stats::Counter")
            if (!ref($counter) || !$counter->isa('Perf::Stats::Counter'));
        my $removed = delete $tracked[$$self]->{$counter->get_name};
        $Log->trace("Removed '" . $self->get_source . "' counter: " . $removed->get_name)
            if ($removed);
        return $removed;
    }

    sub counters_add {
        my ($self, $regex, %args) = @_;
        my $rediscover   = $args{rediscover};
        my $counter_type = $args{type} || $self->get_default_counter_type;
        my $source       = $self->get_source();
        my $node         = $self->get_node();

        $Log->trace(  "Adding counters matching:\n"
                    . Dumper($regex)
                    . "in source '$source' on node '$node'");

        my @added_counters;
        my @already_added_counters;

        # The cache should have been properly populated already
        my $show_out = $self->cache_get;
        my $counters = $self->parse_out_counters(\$show_out, $regex);

        for my $counter (@$counters){
            if (!$self->tracked_is($counter)){
                my $c = $self->make_counter($counter);

                if (   defined($args{type})
                    || !defined($c->get_all_types())
                    || q{} eq $c->get_all_types()) {

                    # ensure counter has a type and allow users to
                    # override that type
                    $c->override_types($counter_type);
                }

                $self->tracked_add($c);
                push @added_counters,$c;
            }
            else{
                push(@already_added_counters, $counter);
            }
        }

        if (!@added_counters && !@already_added_counters){
            $Log->trace(  "Unable to find counters matching:\n"
                          . Dumper($regex)
                          . "in source: '$source' and"
                          . " node '$node' from counters:\n"
                          . _stringify_show_out($show_out));
        }

        my %ret = (added_counters => \@added_counters,
                   already_added  => scalar(@already_added_counters),);

        return %ret;
    }


    sub make_counter {
        my $self    = shift();
        my $counter = shift();

        return
          Perf::Stats::Counter->new(handler => $self,
                                    node    => $self->get_node(),
                                    name    => $counter,);
    }


    sub _stringify_show_out {
        my $show_out = shift();

        if ('ARRAY' eq ref($show_out)) {
            return
              Dumper(
                 [map { "$_->{object}:$_->{instance}:$_->{counter}" } @$show_out
                 ]);
        }
        else {
            return $show_out;
        }
    }

    sub counters_remove {
        my ($self, $regex, %args) = @_;
        my $counters = $self->find( counters=>[$regex], %args );
        my @removed;
        for my $c (@$counters){
            my $aliases = $c->get_aliases;
            $_->remove_counter($c) for @$aliases;
            # Now remove from the tracked counters list
            push @removed, $self->tracked_remove( $c );
        }
        return \@removed;
    }

    sub sample_build_command {
        confess "You must implement this method in a subclass";
    }

    sub sample_build {
        confess "You must implement this method in a subclass";
    }


    sub has_tracked {
        my $self = shift();

        return (0 < scalar(keys(%{$tracked[$$self]})));
    }


    sub sample_start {
        my ($self) = @_;

        if (!$self->has_tracked()) {
            $Log->comment("No tracked counters, skipping sample on $node[$$self]");
            return;
        }

        my $command = $self->sample_build_command;
        $Log->comment(  'Perf::Stats is sending sample command to '
                      . $self->get_node()
                      . ":\n$command");
        $self->execute_start($command);
        return;
    }

    sub sample_wait {
        my ($self) = @_;

        return {}
            if (!$self->has_tracked());

        my $out =  $self->execute_expect;
        $out = $self->preprocess_output($out);
        my $sample = $self->sample_build($out);

        for my $tracked_counter (@{$self->tracked_get()}){
            my $tracked_counter_name = $tracked_counter->get_name;

            next
                if ($tracked_counter->is_hidden);

            $self->check_sample(sample => $sample, counter => $tracked_counter);

        }

        return $sample;
    }

    sub check_sample {
        my $self = shift();
        my %args = validate(@_,
                            {sample  => {type => HASHREF},
                             counter => {type => OBJECT,
                                         isa  => 'Perf::Stats::Counter'},});
        my $tracked_counter = $args{counter};
        my $tracked_counter_name = $tracked_counter->get_name();
        my $sample          = $args{sample};

        if (!exists $sample->{$tracked_counter_name}) {
            if ($tracked_counter->no_zero_values) {
                $sample->{$tracked_counter_name} =
                  $tracked_counter->get_zero_value;
                return;
            }
            else {
                confess(
                  "Tracked counter '$tracked_counter_name' missing from sample:\n"
                      . Dumper($sample));
            }
        }
        if (!defined $sample->{$tracked_counter_name}
            || $sample->{$tracked_counter_name} eq '') {
            # Allow string counters with no values
            if (!$tracked_counter->is_of_type('string')) {
                confess(  "Tracked counter '$tracked_counter_name'"
                        . " has no value in sample:\n"
                        . Dumper($sample));
            }
        }
    } ## end sub check_sample

    sub zero {
        confess "You must implement this method in a subclass";
    }

    sub find {
        my $self = shift;
        my %params = validate(
               @_,
               {
                   counters => {type => ARRAYREF | UNDEF, optional=>1},
                   counters_as_strings => {type => SCALAR | UNDEF, optional=>1, default=>0},
               }
               );
        # Detect if we called this subroutine from the class rather than an object
        my $class_mode = 0;
        if (!ref($self)){
            $class_mode = 1;
        }
        my @matching_counters;
        # In non-class mode we only compare the regexs with the added counters
        if (!$class_mode){
            if ($params{counters} && @{$params{counters}}){
                if ($params{counters_as_strings}){
                    @matching_counters = map { $self->tracked_get($_) or confess "Unable to find counter '$_'" }
                                             @{$params{counters}};
                }
                else{
                    foreach my $counter (@{$self->tracked_get()}){
                        push(@matching_counters, $counter)
                            if (any { $self->do_match(counter => $counter,
                                                      regex   => $_) }
                              @{$params{counters}});
                    }
                }
            }
            else{
                @matching_counters = values %{$tracked[$$self]};
            }
        }
        else{
            ######
            # TODO
            ######
            confess "Feature not implemented yet";
        }
        return \@matching_counters;
    }

    sub preprocess_command {
        my ($self, $command) = @_;
        if ($self->node_is_ng){
            if ($self->is_node_scoped){
                $command = 'run -node local -command "' . $command . '"';
            }
        }
        else{
            $command = 'priv set -q diag;' . $command;
        }
        return $command;
    }

    sub cache_set {
        my ($self, $value) = @_;
        confess "You must pass a value" if (!$value);
        $Log->trace("Setting cache of '" . $self->get_source .
                      "' handler on node '" . $self->get_node ."' to:\n" .
                      Dumper($value)
                      );
        $cache[$$self] = $value;
    }

    sub cache_build_command {
        confess "You must implement this subroutine";
    }

    sub parse_out_counters {
        confess('This method must be implemented in a subclass of ' . __PACKAGE__);
    }

    sub cache_populate_start {
        my ($self, $regexs) = @_;
        confess "You must pass an argument" if (!$regexs);
        my $cache_command = $self->cache_build_command( $regexs );
        confess "No command returned!" if (!$cache_command);
        $self->execute_start($cache_command);
        return;
    }

    sub get_nate_connection {
        my ($self) = @_;
        my $conn = $connection[$$self];
        if ($conn->isa('NACL::APISet')) {
            # Grab the NATE connection out of the APISet
            $conn = $connection[$$self]->get_connection;
        }

        return $conn;
    }

    # The log_level parameter below sets the level of logging for
    # information going through Connectrec.  We want this to be
    # `debug' instead of the default `comment' as the amount of disk
    # space used for multiple stats explain/list invocations is huge.
    # However, some users want to see the output of these commands,
    # but setting the LOGSPEC to log Connectrec debug for all other
    # invocations also produces tons of output.  As a compromise, set
    # the log level of this specific call to Connectrec to `comment'
    # when `debug' level is set for Perf::Stats.
    sub execute_start {
        my ($self, $command)= @_;
        confess "You must pass a command" if (!$command);
        $command = $self->preprocess_command($command);
        my $conn = $self->get_nate_connection;
        $conn->connect() if $conn->is_disconnected();
        $conn->execute_start(command   => $command,
                             log_level => (
                                   ($Log->can('may_debug') && $Log->may_debug())
                                   ? 'comment'
                                   : 'debug'),);
    }

    sub execute_expect {
        my ($self) = @_;
        my $conn = $self->get_nate_connection;

        # We only want a failure if nothing is received in a given
        # timeframe, not if it takes a long time to receive all the
        # data, hence the infinite timeout. We swap of the timeout for
        # max_idle because users that pass in APISet objects can only
        # manipulate the timeout parameter, not the max_idle. This is
        # an ugly hack that should be removed when/if BURTs 572309
        # & 572313 are implemented.
        my $out = $conn->execute_expect(
            max_idle => $conn->timeout(),
            timeout  => -1,
            # see execute_start()
            log_level => (($Log->can('may_debug') && $Log->may_debug())
                          ? 'comment'
                          : 'debug'),);
        $self->_write_raw_data($out);

        return $out;
    }
    sub _write_raw_data {
        my $self = shift;
        my $out = shift;

        my $p = NATE::ParamSet->new(global=>1);
        my $log_dir = $p->get('LOGDIR').'/raw_data/';

        eval {
            make_path($log_dir);
            my $log_file = $log_dir.$node[$$self]."_$$.gz";

            my $z = new IO::Compress::Gzip $log_file,  'Append'=>1;
            $z->write($out."\n");
            $z->close();
        };
        if($@){
            $Log->warn("Unable to write raw command data to file:\n$@");
        }
    }

    sub cache_populate_wait {
        my ($self) = @_;
        my $out =  $self->execute_expect;
        $out = $self->preprocess_output($out);
        $self->cache_set( $out );
        return;
    }

    sub preprocess_output {
        my ($self, $out) = @_;
        my $preprocessors = $self->get_preprocessors;
        return $out if (!$preprocessors || !@$preprocessors);
        # Pass lines to preprocessors so they don't each have
        # to do it
        my $lines = [split(/\r?\n/, $out)];
        # Add a blank line before additional output
        push @$lines, "\n";
        for my $preproc (@$preprocessors){
            $preproc->($lines);
        }
        # Rebuild output
        $out = join("\n", grep { defined($_) } @$lines);
        $Log->trace("Reconstructed output:\n$out");
        return $out;
    }

    sub preprocessor_add {
        my ($self, $code) = @_;
        confess("Argument must be coderef") if (!$code || !ref($code) || ref($code) ne 'CODE');
        push @{ $preprocessors[$$self]  }, $code;
    }

    sub cache_purge {
        my ($self) = @_;
        delete $cache[$$self];
        return;
    }

    sub get_default_counter_type {
        return 'delta';
    }

    sub reconnect {
        my $self   = shift;
        my %params = validate(
            @_,
            {
                force   => { type => SCALAR, default => 0 },
                timeout => { type => SCALAR, default => 2 },
            },
        );

        my $force   = $params{force};          # Force reset even if connected?
        my $timeout = $params{timeout};        # T/O to verify responsiveness
        my $conn    = $self->get_connection;   # May be APISet or Connectrec

        if ($conn->isa('NACL::APISet')) {
            # NACL::APISet::api_is_session_alive doesn't accept a
            # timeout argument, so we'll use the its Connectrec
            # object to check if the connection is alive.
            my $nate_conn = $self->get_nate_connection;
            if ($force || $nate_conn->is_disconnected(timeout => $timeout)) {
                $conn->api_reconnect_session;
            }
        }
        elsif ($conn->isa('Connectrec')) {
            if ($force || $conn->is_disconnected(timeout => $timeout)) {
                $conn->disconnect;
                $conn->connect;
            }
        }
        else {
            confess 'Unknown connection type';
        }
        return;
    }

    sub disconnect {
        my $self   = shift;
        #Don't mess with passed in ApiSets
        if(!$external_conn[$$self]){
            my $conn = $self->get_nate_connection;
            $conn->disconnect;
        }
        return;
    }

    sub create_connection {
        my $self    = shift();
        my $hostrec = shift();
        my $conn;

        if (!$node_is_ng[$$self]) {
            # We don't necessarily need a brand new connection but
            # one that we know won't be used by multiple handlers
            # at the same time during add or sample suffixing the
            # connid with the source should be unique enough and
            # yet allow reuse across Perf::Stats instances...
            my $connid = "rsh_$source[$$self]";
            $conn = connect("#$connid;conntype=rsh-socket");
            $conn->timeout($Perf::Stats::TIMEOUT);
        }
        else {
            my @old_conns;

            my $node_obj =
                    NACL::C::Node->new(node    => $node[$$self],
                                    hostrec => $hostrec,);

            for my $c ($hostrec->connections){
                #ensure that $c actually exists.
                if($c){
                    if($c->is_disconnected(verify => 0,timeout=>1)){
                        push (@old_conns,$c);
                    }
                }
            }
            my %args = (category  => 'Node',
                            interface => 'CLI',
                            set       => 'CMode',);
            #Hostrecs store all connections until they are destroyed
            #If any connections are available and disconnected re-use
            #that connection to preserve SSH connections.
            if (defined($old_conns[0])){
                $old_conns[0]->connect;
                $args{connection} = $old_conns[0];
            }

            $conn = $node_obj->apiset(%args);
            $conn->set_timeout(timeout => $Perf::Stats::TIMEOUT);
        }

        return $conn;
    }


    sub get_node_names {
        my $self = shift();
        my $conn = $connection[$$self];
        my @node_names;
        my $response;

        if ($conn->isa('NACL::APISet::Node::CLI::CMode')) {
            $response = $conn->system_node_show(fields => 'node');
            @node_names =
              map { $_->{node} } @{$response->get_parsed_output()};
        }
        elsif ($conn->isa('NACL::APISet::Node::CLI::7Mode')) {
            $response = $conn->cf_partner();
            @node_names =
              ($node[$$self], $response->get_parsed_output()->[0]->{partner});
        }
        elsif ($conn->isa('Connectrec')) {
            $response = $conn->execute(command => 'cf partner');
            chomp($response);
            @node_names = ($node[$$self], $response);
        }

        return \@node_names;
    }


    sub acceptable_apiset {
        my $self   = shift();
        my $apiset = shift();

        return (   $apiset->isa('NACL::APISet::Node::CLI::7Mode')
                || $apiset->isa('NACL::APISet::Node::CLI::CMode'));
    }


    sub do_match {
        my $self = shift();
        my %args = validate(@_,
                            {regex   => {type => HASHREF | SCALAR},
                             counter => {type => SCALAR | OBJECT},});

        #HASH based regex's imply a CM counter.
        return if(ref($args{regex}) eq 'HASH');

        my $counter;

        if (ref($args{counter})) {
            confess('counter argument is not a Perf::Stats::Counter object')
                if (!$args{counter}->isa('Perf::Stats::Counter'));

            $counter = $args{counter}->get_name();
        }
        else {
            $counter = $args{counter};
        }

        return ($counter =~ /^$args{regex}$/);
    }


    sub get_cluster_name {
        my $self = shift();
        my $response = $self->get_connection()->cluster_identity_show();
        my $cluster = $response->get_parsed_output->[0]->{name};

        return $cluster;
    }


    sub get_cluster_nodes {
        my $self = shift();

        my $response = $self->get_connection()->cluster_show();
        my $out = $response->get_parsed_output();

        my @nodes = map { $_->{node} } @$out;

        return \@nodes;
    }


    sub get_aggr_list {
        my $self  = shift();
        my $node  = $self->get_node;
        my $hostp = host($node);
        my @aggrs;

        if ($hostp->hosttype =~ /-ng-/) {
            my $response =
              $self->get_connection->storage_aggregate_show(nodes => $node);
            my $lines = $response->get_parsed_output;

            @aggrs = map { $_->{aggregate} } @$lines;
        }
        else {
            @aggrs = filer::filer_aggrs_get($self->get_connection);
        }

        return \@aggrs;
    }


    sub _process_disks {
        my $lines = shift();
        my @ds;

        foreach my $d (@$lines) {
            my $dh = {type => $d->{raidtype}};
            my $name = $d->{disk};
            my $owner = $d->{owner};
            my $pathnames = $d->{diskpathnames};

            if('ARRAY' ne ref($pathnames)) {
                $pathnames = [$pathnames];
            }
           PATHNAME: foreach my $pathname (@$pathnames) {
                 # The name also includes the node, split it.
                 if ($pathname =~ /([^:]+):(.*)/) {
                     if ($1 eq $owner) {
                         $dh->{node} = $owner;
                         $dh->{name} = $2;
                         last PATHNAME;
                     }
                 }
                 else {
                     confess "Unable to parse disk name '$pathname'";
                 }
            }
            confess("No disk name found")
                if (!defined($dh->{name}));
            # Recent builds of RR now use 'uid' instead of 'uuid'
            $dh->{instance} = $d->{uuid} || $d->{uid};
            push @ds, $dh;
        }

        return \@ds;
    }

    sub get_aggr_disks {
        my $self  = shift();
        my $aggr  = shift();
        my $hostp = host($self->get_node);
        my @ds;

        if ($hostp->hosttype =~ /-ng-/){
            my $response = $self->get_connection->storage_disk_show('aggregate'=>$aggr);
            my $lines = $response->get_parsed_output;
            @ds = @{_process_disks($lines)};
        }
        else{
            @ds = Aggregate::aggr_disks( $self->get_connection, $aggr );
        }

        return \@ds;
    }

    sub toggle {
        my ($self,$toggle) = @_;
        #ensure handler is connected to filer.
        #it may not use the same connection as before.
        if($toggle){
            if($self->is_disconnected) {
                my $hostrec  = host($node[$$self]);
                my $conn = $self->create_connection($hostrec);
                $self->set_connection($conn);
            }
        }
        else {
            $self->disconnect();
        }
        $enabled[$$self] = $toggle;
    }
    sub is_disconnected {
        my $self = shift;
        my $connection = $self->get_nate_connection;
        my $out = ($connection->is_disconnected(timeout=>1));

        return $out;
    }
    #since NATE doesn't destroy old connections
    #just disconnect them for re-use.
    sub _destroy :Destroy {
        my $self = shift;
        return unless $$ == $self->get_pid;

        my $err = $@;
        eval{
            if($connection[$$self]){
                $self->disconnect;
            }
        };
        if ($@){
            $Log->comment("Error during destroy of Stats Handler: $@");
        }
        $@ = $err;
    }
}
1;
