package Perf::Stats::Handler::CM; {
    use strict;
    use warnings;
    use Carp qw(confess);
    use Tharn qw($Log);
    use Data::Dumper;
    use List::MoreUtils qw(any);
    use Params::Validate qw(validate SCALAR ARRAYREF HASHREF OBJECT UNDEF);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Perf::Stats::Counter::CM::RPC::CM_Client;
    use Perf::Stats::Counter::CM::RPC;
    use Object::InsideOut qw(Perf::Stats::Handler);

    my $STRING_DEPRECATION_WARNED = 0;

    my @all_objects     :Field;
    my @queries         :Field :Std(queries);


    sub _init :Init {
        my $self = shift();

        $queries[$$self]       = [];
    }


    sub get_all_objects {
        my $self = shift();

        if (!defined($all_objects[$$self])) {
            $all_objects[$$self] = $self->collect_all_objects();
        }

        return $all_objects[$$self];
    }


    sub tracked_add {
        my $self    = shift();
        my $counter = shift();

        confess('You must specify an argument')
            if (!$counter);
        confess("Argument must be a Perf::Stats::Counter::CM")
            if (!ref($counter) || !$counter->isa('Perf::Stats::Counter::CM'));

        my @fields  = $self->get_spec_fields();
        my @values  = map { $counter->get_field($_) } @fields;
        my $tracked = $self->_get_tracked();

        _insert_counter(tree => $tracked, path => \@values, counter => $counter);
    }


    sub flatten {
        my $struct = shift();
        my @ret;

        foreach my $branch (values(%$struct)) {
            if ('HASH' eq ref($branch)) {
                push(@ret, flatten($branch));
            }
            else {
                push(@ret, $branch);
            }
        }

        return @ret;
    }


    sub tracked_get {
        my $self       = shift();
        my $counter    = shift();
        my $tracked    = $self->_get_tracked();

        if (!defined($counter)) {
            my @counters = flatten($tracked);

            return \@counters;
        }

        my $ref    = ref($counter);
        my @fields = $self->get_spec_fields();
        my @field_vals;

        if (('HASH' eq $ref) || (q{} eq $ref)) {
            my $inspect;
            if (q{} eq $ref) {
                $inspect = _string_to_spec(string => $counter, is_re => 0);
            }
            else {
                $inspect = $counter;
            }

            foreach my $field (@fields) {
                push(@field_vals, $inspect->{$field});
            }
        }
        elsif ('Perf::Stats::Counter::CM' eq $ref) {
            foreach my $field (@fields) {
                push(@field_vals, $counter->get_field($field));
            }
        }
        else {
            confess(  'Argument must be a Perf::Stats::Counter::CM or a'
                          . " counter hash, or a spec string, not a: $ref");
        }

        return _get_counter(tree => $tracked, path => \@field_vals);
    }


    # Retrieve a counter from the counter tree
    sub _get_counter {
        #tree   => {type => HASHREF}
        #path   => {type => ARRAYREF}
        my %args = @_;

        return _navigate_tree(%args);
    }

    # Insert a counter into the counter tree
    sub _insert_counter {
        #tree    => {type => HASHREF}
        #path    => {type => ARRAYREF}
        #counter => {type => OBJECT, isa  => 'Perf::Stats::Counter::CM'}
        my %args = @_;

        _navigate_tree(%args);
    }

    # retrieve all the labels and values for a counter
    sub _get_label_values {
        #tree => {type => HASHREF}
        #path => {type => ARRAYREF}
        my %args = @_;

        # copy the path and pop off the label so all labels are retrieved
        my @path = @{$args{path}};
        pop(@path);
        $args{path} = \@path;

        return _navigate_tree(%args);
    }

    # Given a tree of hashrefs (such as tracked counters or a sample),
    # navigate it based on given path of keys and insert or return
    sub _navigate_tree {
        my %args       = @_;
        my $tree       = $args{tree};
        my @nodes      = @{$args{path}};
        my $insert     = $args{counter};
        my $last_node  = pop(@nodes);
        my $cur_branch = $tree;

        foreach my $node (@nodes) {
            if (!defined($cur_branch->{$node})) {
                $cur_branch->{$node} = {};
            }

            $cur_branch = $cur_branch->{$node};
        }

        if (defined($insert)) {
            $cur_branch->{$last_node} = $insert;
        }
        else {
            return $cur_branch->{$last_node};
        }
    }


    sub tracked_remove {
        my $self    = shift();
        my $counter = shift();

        confess('You must specify an argument')
            if (!$counter);
        confess("Argument must be a Perf::Stats::Counter::CM")
            if (!ref($counter) || !$counter->isa('Perf::Stats::Counter::CM'));

        my $tracked = $self->_get_tracked();

        my $removed = delete(
            $tracked->{$counter->get_cm_object()}
                ->{$counter->get_cm_instance()}
                ->{$counter->get_cm_name()}
                ->{$counter->get_cm_label()});

        $Log->trace(  "Removed '"
                    . $self->get_source
                    . "' counter: "
                    . $removed->get_name)
          if ($removed);

        return $removed;
    }


    sub check_sample {
        my $self = shift();
        #sample  => {type => HASHREF}
        #counter => {type => OBJECT, isa  => 'Perf::Stats::Counter::CM'}
        my %args = @_;

        my $counter = $args{counter};
        my $sample  = $args{sample};

        my @fields  = $self->get_spec_fields();
        my @field_vals = map { $counter->get_field($_) } @fields;
        my $label   = $field_vals[-1];
        my $values  = _get_label_values(tree => $sample, path => \@field_vals);

        if (!exists $values->{$label}) {
            if ($counter->no_zero_values()) {
                $values->{$label} = $counter->get_zero_value();
                $sample->{$counter->get_cm_object()}
                       ->{$counter->get_cm_instance()}
                       ->{$counter->get_cm_name()}
                       ->{$label} = $values->{$label};
            }
            else {
                confess(  "Tracked counter:\n"
                        . $counter->dump(1)
                        . "\nmissing from sample:\n"
                        . Dumper($sample));
            }
        }
        if (!defined($values->{$label}) || q{} eq $values->{$label}) {
            # Allow string counters with no values
            if (!$counter->is_of_type('string')) {
                confess(  "Tracked counter:\n"
                        . $counter->get_full_name()
                        . " has no value in sample:\n"
                        . Dumper($sample));
            }
        }
    }


    sub sample_wait {
        my $self = shift();

        my $sample = $self->make_names($self->SUPER::sample_wait());

        return $sample;
    }


    sub make_names {
        my $self = shift();
        my $tree = shift();
        my %sample;

        while (my ($object, $instances) = each(%$tree)) {
            # Internal-only counters and aliases won't have any instances
            if ('HASH' ne ref($instances)) {
                $sample{$object} = $instances;
                next;
            }

            while (my ($instance, $counters) = each(%$instances)) {
                while (my ($counter, $labels) = each(%$counters)) {
                    while (my ($label, $value) = each(%$labels)) {
                        my $fullname = "$object:$instance:$counter";

                        if (q{} ne $label) {
                            $fullname .= ".$label";
                        }

                        $sample{$fullname} = $value;
                    }
                }
            }
        }

        return \%sample;
    }


    sub get_cm_objects {
        my ($self, $counters) = @_;
        if (!$counters || !@$counters) {
            confess("You must specify counters");
        }

        # Query CM for known objects.
        my $list = $self->get_all_objects();;

        # Match user-supplied counters against known objects.
        # Assume everything up to the first ':' in counter
        # is a regex specifying one or more objects.
        my %matched;
        my %regex_seen;
        $Log->debug("DEBUG: counters:\n" . Dumper($counters));

        foreach my $counter_re (@$counters) {
            my $regex     = $self->_process_regex($counter_re);
            my $object_re = $regex->{object};

            next
              if $regex_seen{$object_re};

            $regex_seen{$object_re} = 1;

            my $found = 0;
            foreach my $obj (grep m/^$object_re$/, @$list) {
                $matched{$obj}++;    # Hash since we only want unique matches
                $found++;
                $Log->trace("Found object '$obj' matching '$object_re'");
            }
            if (!$found) {
                confess "No objects found for regex '$object_re' in:\n"
                  . join("\n", @$list);
            }
        }

        # Return arrayref of the matched objects.
        my @objects = keys %matched;
        return \@objects;
    } ## end sub get_cm_objects


    # Validate counter regexp hashrefs and convert traditional `:'
    # separated counter regexps to new data structure
    sub _process_regex {
        my $self  = shift();
        my $regex = shift();

        my $ref_type = ref($regex);

        if ('HASH' eq $ref_type) {
            return $self->_validate_regex_spec($regex);
        }
        elsif (q{} eq $ref_type) {
            return _string_to_spec(string => $regex, is_re => 1);
        }
        else {
            confess(  "Unsupported counter regex specification type ($ref_type)"
                    . ' for '
                    . Dumper($regex));
        }
    }


    sub counters_add {
        my $self            = shift();
        my $regex           = shift();
        my $processed_regex = $self->_process_regex($regex);
        my %ret             = $self->SUPER::counters_add($processed_regex, @_);
        my $ret             = $ret{added_counters};

        if (0 < scalar(@$ret)) {
            # Update list of stats requests
            $self->_compute_queries();
        }

        return %ret;
    }
    sub counters_remove {
        my $self            = shift();
        my $regex           = shift();
        my $processed_regex = $self->_process_regex($regex);
        my $ret             = $self->SUPER::counters_remove($processed_regex, @_);

        if (0 < scalar(@$ret)) {
            # Update list of stats requests
            $self->_compute_queries();
        }

        return $ret;
    }


    sub _string_to_spec {
        # {string => 1,
        # is_re  => 1,});
        my %params = @_;
        my $is_re        = $params{is_re};

        if (!$STRING_DEPRECATION_WARNED && $is_re) {
            $STRING_DEPRECATION_WARNED = 1;
            $Log->warn(  'WARNING: String counter specs are deprecated'
                       . ' for CM based counters'
                       . "\nSee Perf::Stats documentation for "
                       . ' hash based spec description');
        }

        my $obj_inst_cnt = $params{string};

        my ($obj, $inst_cnt) = Perf::Stats::Entry::split_regexp($obj_inst_cnt);
        my ($inst, $cnt_mlabel) = Perf::Stats::Entry::split_regexp($inst_cnt);

        my $label_re = '^([^.]+)' . ($is_re ? '\\' : q{}) . '\.(.+)';

        my ($cnt, $label);
        if ($cnt_mlabel =~ /$label_re/) {
            $cnt   = $1;
            $label = $2;
        }
        else {
            $cnt   = $cnt_mlabel;
            $label = ($is_re ? '.*' : q{});
        }

        my %spec = (object   => $obj,
                    instance => $inst,
                    counter  => $cnt,
                    label    => $label);

        return \%spec;
    }


    sub get_regex_spec {
        my $self = shift();
        my $spec = {node => {type    => SCALAR | UNDEF,
                                      default => '.*',},
                             object   => {type => SCALAR,},
                             instance => {type    => SCALAR | UNDEF,
                                          default => '.*',},
                             counter => {type    => SCALAR | UNDEF,
                                         default => '.*',},
                             label => {type    => SCALAR | UNDEF,
                                       default => '.*',},};

        return $spec;
    }


    sub _validate_regex_spec {
        my $self    = shift();
        my $counter = shift();
        my @params  = %$counter;
        my %spec = validate(@params, $self->get_regex_spec());

        return \%spec;
    }


    sub do_match {
        my $self = shift();
        # regex   => {type => HASHREF | SCALAR}
        # counter => {type => HASHREF | OBJECT}
        my %args = @_;
        my $regex = $self->_process_regex($args{regex});

        # reverse the order of the fields since counter names will
        # vary more than object names, leading to quicker elimination
        my @match_fields = reverse($self->get_spec_fields());

        my %counter;
        my $counter_ref = ref($args{counter});

        if ('HASH' eq $counter_ref) {
            %counter = %{$args{counter}};
        }
        else {
            confess('counter argument is not a Perf::Stats::Counter::CM object')
                if (!$args{counter}->isa('Perf::Stats::Counter::CM'));
            %counter = map { +$_, $args{counter}->get_field($_) } @match_fields;
        }

        foreach my $part (@match_fields) {
            if ($counter{$part} !~ /^$regex->{$part}$/) {
                $Log->debug(  "Rejecting '$part' entry because value"
                              . " ($counter{$part}) does not match"
                              . " regexp ($regex->{$part})");
                return 0;
            }
        }

        return 1;
    }


    # Spec fields are just the fields used to specify a counter
    sub get_spec_fields {
        my $self = shift();

        # order fields from most to least significant
        return ('object', 'instance', 'counter', 'label');
    }


    # Counter fields are the spec fields as well as any other
    # information that should be kept
    sub get_counter_fields {
        my $self = shift();

        my @fields = $self->get_spec_fields();
        push(@fields, ('prop','unit'));

        return @fields;
    }


    sub get_counter_class {
        my $self = shift();

        return 'Perf::Stats::Counter::CM';
    }


    sub make_counter {
        my $self    = shift();
        my $counter = shift();

        my @fields = $self->get_counter_fields();
        my %spec;

        @spec{@fields} = @$counter{@fields};

        my $class = $self->get_counter_class();

        return
          ${class}->new(handler => $self,
                        node    => $self->get_node(),
                        spec    => \%spec,);
    }


    sub collect_all_objects {
        confess('This method must be implemented in a subclass of ' . __PACKAGE__);
    }

    sub _compute_queries {
        confess('This method must be implemented in a subclass of ' . __PACKAGE__);
    }

    sub parse_out_counters {
        my ($self, $show_out, $regex) = @_;

        my $out = $$show_out;
        # CACHE STRUCTURE
        # Hash pairing objects with all there flat counters and instances
        # {
           # object => {
               # counters =>  'counter_name' => {
                           # counter => 'counter_name',
                           # base_counter => 'base_counter_name',
                           # prop => 'properties,
                           # unit => 'unit_type',
                           # object => 'object_name',
                           # #optional
                           # labels => { colums => ['array_ref _columns'], #if columns exist
                                            # rows => ['array_ref _rows'],} #if rows exist
                        # }, 'counter_name2' => {...},...
               # instances => {instance1' => 1, 'instnace2' => 1.},
          # object => {
                # counters => {..},
                # instance => {..},
        # }
        my@counters;
        for my $o (keys %$out){
            #match object
            next if($o !~ /^$regex->{object}$/);

            my $counters = $out->{$o}->{counters};
            for my $line (values %$counters) {
                #match counter
                next if($line->{counter} !~ /^$regex->{counter}$/);

                for my $i (keys %{$out->{$o}->{instances}}){
                    #match instance
                    next if($i !~ /^$regex->{instance}$/);

                    my $line_counters = $self->expand_counters($line);
                    for my $c (@$line_counters){
                        $c->{instance} = $i;

                        #match label
                        if($c->{label} =~ /^$regex->{label}$/) {
                            my $base_counter = $c->{base_counter};
                            if(defined($base_counter) && '-' ne $base_counter)
                            {
                                confess ("No base_counter found for counter: \n".Dumper($c))
                                    if(!defined $counters->{$base_counter});
                                my $base = $self->expand_counters($counters->{$base_counter});

                                #need to match up the base labels to numerator.
                                for my $bc (@$base) {
                                    if(defined $bc->{labels} && $bc->{index} ne $c->{index}) {
                                        next;
                                    }
                                    #by this point base label should match numerator
                                    if(defined $bc->{labels} && $bc->{label} ne $c->{label}){
                                        confess('Mismatching labels for counter and base: \n'.
                                            'counter: '.Dumper($c) .'\n'.
                                            'base: '.Dumper($bc));
                                    }
                                    $bc->{instance} = $i;
                                    push @counters, $bc;

                                    $c->{base_counter} = $bc;
                                }
                            }
                            push @counters, $c;
                            $Log->debug("Accepting counter:\n" . Dumper($c));
                        }
                        else {
                            $Log->debug("Rejecting counter:\n" . Dumper($c));
                        }
                    }
                }
            }
        }
        return \@counters;
    }
    sub expand_counters {
        my ($self, $line_counter) = @_;

        #dereferencing since we do not want to alter original cache
        my $line = {%$line_counter};
        # If labels are present, return one counter per
        my $labels = $line->{labels};

        if (!defined($labels) || $labels eq '-') {
            # empty labels field
            $line->{label} = q{};
            $line->{index} = 0;
            return [$line];
        }
        my @parsed_labels;
        if($labels->{columns}) {
            my $columns = $labels->{columns};
            my $index =0;
            for my $c (@$columns) {
                $c = _label_cleanup($c);
                if($labels->{rows}){
                    my $rows = $labels->{rows};
                    for my $r (@$rows) {
                        $r = _label_cleanup($r);
                        push(@parsed_labels, {label => "$r.$c", index => $index});
                        $index++;
                    }
                }
                else {
                    push(@parsed_labels, {label => $c, index => $index});
                    $index++;
                }
            }
        }

        if (0 == scalar(@parsed_labels)) {
            confess("Unable to parse labels:\n" . Dumper($line));
        }
        my @counters;
        foreach my $label (@parsed_labels) {
            my %cntr_info = %$line;
            $cntr_info{label} = $label->{label};
            $cntr_info{index} = $label->{index};
            push @counters, \%cntr_info;
        }

        return \@counters;
    }
    sub _label_cleanup {
        my $label = shift;
        $label =~ s/^\s*{//;
        $label =~ s/}\s*$//;
        return $label;
    }
}

1;
