package Perf::Stats::Handler::CM::CM_Client; {
    use strict;
    use warnings;
    use Carp qw(confess cluck);
    use Tharn qw($Log);
    use Readonly;
    use Data::Dumper;
    use Params::Validate qw(validate SCALAR HASHREF OBJECT UNDEF);
    use Connectrec;
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Object::InsideOut qw(Perf::Stats::Handler::CM);

    # We have to hardcode the process list because currently the only
    # way to get the list of processes is to query the build tree for
    # the specific ONTAP version installed. This will change if/when
    # BURT558807 is implemented
    #
    # Generated by: $
    # /x/eng/rlse/DOT/RrollingrockN/countergen/src/bin/countergen info \
    #   --list-attr system_processes
    # -or for SN-
    # /x/eng/rlse/DOT/mainN/cm/countergen/src/bin/countergen info \
    #   --list-attr system_processes
    Readonly::Array our @PROCESSES =>
      ('mntsvc', 'kernel', 'mgwd', 'cmd', 'nchm', 'notifyd', 'schm', 'shm');


    sub preprocess_command {
        my $self    = shift();
        my $command = shift();

        return $command;
    }


    sub collect_all_objects {
        my $self = shift();
        my %objects_for_process;

        my $conn = $self->get_nate_connection();

        foreach my $process (@PROCESSES) {
            my $cmd = "cm_client -p $process -a object-list-info";

            my $output = $conn->execute(
                               command   => $cmd,
                               log_level => (
                                   ($Log->can('may_debug') && $Log->may_debug())
                                   ? 'comment'
                                   : 'debug'));
            my @objects = _parse_cm_client_objects($output);
            $objects_for_process{$process} = \@objects;
        }

        return \%objects_for_process;
    }


    sub cache_build_command {
        my $self    = shift();
        my $regexes = shift();

        confess('You must specify an argument')
            if (!$regexes);

        confess('Invalid argument')
          if (('ARRAY' ne ref($regexes)) || (0 == scalar(@$regexes)));

        my %matched;
        my %regex_seen;
        my $objects_for_process = $self->get_all_objects();

        foreach my $counter_re (@$regexes) {
            my $regex   = $self->_process_regex($counter_re);
            my $proc_re = $regex->{process};
            my $obj_re  = $regex->{object};

            next
                if ($regex_seen{$proc_re}->{$obj_re});

            $regex_seen{$proc_re}->{$obj_re} = 1;

            my $found = 0;

            while (my ($proc, $objs) = each(%$objects_for_process)) {
                if ($proc =~ /^$proc_re$/) {
                    foreach my $obj (grep m/^$obj_re$/, @$objs) {
                        $matched{$proc}->{$obj}++;
                        $found++;
                        $Log->debug("Found process '$proc' and object '$obj'"
                                        . " matching:\n" . Dumper($regex));
                    }
                }
            }

            if (!$found) {
                confess("No objects found for regex:\n" . Dumper($regex) . "in:\n"
                  . Dumper($objects_for_process));
            }
        }

        my @cmds;

        while (my ($proc, $objs) = each(%matched)) {
            foreach my $obj (keys(%$objs)) {
                push(@cmds, ("echo Process: $proc",
                             "cm_client -p $proc -a object-instance-list-info -o $obj",
                             "cm_client -p $proc -a object-counter-list-info -o $obj"));
            }
        }

        return join(';', @cmds);
    }


    sub cache_populate_wait {
        my $self = shift();

        $self->SUPER::cache_populate_wait();
        my $out = $self->cache_get();
        my @lines = split(/\r?\n/, $out);

        my $current_process;
        my $current_obj;
        my $counters;
        my $instances;

        foreach my $line (@lines) {
            if ($line =~ /^Process: (.*)$/) {
                $current_process = $1;
            }
            elsif ($line =~ m{^[[:space:]]+
                               object_name:"([^"]+)" [ ]
                              counter_name:"([^"]+)" [ ]
                              base_counter:"([^"]*)" .*
                                properties:"([^"]+)" .*
                                   labelsX:"([^"]*)" [ ]
                                   labelsY:"([^"]*)"
                             }msx) {
                my ($object, $counter, $base, $prop, $x_labels, $y_labels) =
                  ($1, $2, $3, $4, $5, $6);

                my @labels;
                if (defined($x_labels) && q{} ne $x_labels) {
                    my $pos = 0;
                    my @x_labels = split(/,/, $x_labels);

                    foreach my $x_label (@x_labels) {
                        if (defined($y_labels) && q{} ne $y_labels) {
                            my @y_labels = split(/,/, $y_labels);

                            foreach my $y_label (@y_labels) {
                                push(@labels, {label => "$x_label.$y_label",
                                               pos   => $pos});
                                $pos++;
                            }
                        }
                        else {
                            push(@labels, {label => $x_label,
                                           pos   => $pos});
                            $pos++
                        }
                    }
                }
                else {
                    @labels = ({label => q{},
                                pos   => 0});
                }


                foreach my $label (@labels) {
                    $counters->{$current_process}->{$object}->{$counter}
                      ->{$label->{label}} = {prop         => $prop,
                                             query_name   => $counter,
                                             base_counter => $base,
                                             index        => $label->{pos}};
                }
            }
            elsif ($line =~ /object_name:"([^"]+)" instance_name:"([^"]+)"/) {
                push(@{$instances->{$current_process}->{$1}}, $2);
            }
        }

        my %complete_counters;
        my $add_count = 0;

        while (my ($process, $obj_counters) = each(%$counters)) {
            while (my ($object, $full_counter_specs) = each(%$obj_counters)) {
                while (my ($counter, $labels) = each(%$full_counter_specs)) {
                    while (my ($label, $specs) = each(%$labels)) {
                        foreach
                          my $instance (@{$instances->{$process}->{$object}}) {
                            $complete_counters{$process}->{$object}->{$instance}
                              ->{$counter}->{$label} = $specs;
                            $add_count++;
                        }
                    }
                }
            }
        }

        confess("No counters found in:\n$out")
            if (0 == $add_count);

        $self->cache_set(\%complete_counters);

        return;
    }


    sub get_regex_spec {
        my $self = shift();

        my $spec = $self->SUPER::get_regex_spec();

        $spec->{process} = {type => SCALAR};

        return $spec;
    }


    sub get_spec_fields {
        my $self = shift();

        my @fields = $self->SUPER::get_spec_fields();

        unshift(@fields, 'process');

        return @fields;
    }


    sub get_counter_fields {
        my $self = shift();

        my @fields = $self->SUPER::get_counter_fields();

        push(@fields, ('index', 'base_counter'));

        return @fields;
    }


    sub get_counter_class {
        my $self = shift();

        my $base_class    = $self->SUPER::get_counter_class();
        my $counter_class = "${base_class}::RPC::CM_Client";

        return $counter_class;
    }


    sub _parse_cm_client_objects {
        my $objects = shift();
        my @lines = split(/\n/, $objects);
        my @objects;
        foreach my $line (@lines) {
            if ($line =~ /object_name:"([^"]+)"/) {
                push(@objects, $1);
            }
        }

        return @objects;
    }


    sub parse_out_counters {
        my $self  = shift();
        my $cache = shift();
        my $regex = shift();
        my @counters;

        while (my ($process, $objects) = each(%{$$cache})) {
            while (my ($object, $instances) = each(%$objects)) {
                while (my ($instance, $counters) = each(%$instances)) {
                    while (my ($counter_name, $labels) = each(%$counters)) {
                        while (my ($label, $spec) = each(%$labels)) {
                            my $counter = {process  => $process,
                                           object   => $object,
                                           instance => $instance,
                                           counter  => $counter_name,
                                           label    => $label,
                                           %$spec};

                            if ($self->do_match(regex   => $regex,
                                                counter => $counter,)
                              ) {
                                push(@counters, $counter);
                                $Log->trace(
                                      "Accepting counter:\n" . Dumper($counter))
                                  if ($Perf::Stats::DO_TRACE);

                                my $base_counter = $counter->{base_counter};

                                if (defined($base_counter)
                                    && q{} ne $base_counter) {

                                    my %base = (
                                            process  => $process,
                                            object   => $object,
                                            instance => $instance,
                                            counter  => $base_counter,
                                            label    => q{},
                                            %{$$cache->{$process}
                                                     ->{$object}
                                                     ->{$instance}
                                                     ->{$base_counter}
                                                     ->{$label}});

                                    push(@counters, \%base);
                                }
                            }
                            elsif ($Perf::Stats::DO_TRACE) {
                                $Log->trace(
                                     "Rejecting counter:\n" . Dumper($counter));
                            }
                        } ## end while (my ($label, $spec)...
                    } ## end while (my ($counter_name,...
                } ## end while (my ($instance, $counters...
            } ## end while (my ($object, $instances...
        } ## end while (my ($process, $objects...

        return \@counters;
    }


    sub _compute_queries {
        my $self = shift();

        my $tracked = $self->_get_tracked();

        my $queries = {};

        while (my ($process, $objects) = each(%$tracked)) {
            while (my ($object, $instances) = each(%$objects)) {
                while (my ($instance, $counters) = each(%$instances)) {
                    $queries->{$process}->{$object}->{instances}->{$instance} = 1;

                    foreach my $counter (keys(%$counters)) {
                        $queries->{$process}->{$object}->{counters}->{$counter} = 1;
                    }
                }
            }
        }

        $Log->debug("DBG: Queries computed:\n" . Dumper($queries));
        $self->set_queries($queries);
        return;
    }


    sub sample_build_command {
        my $self = shift();
        my $queries = $self->get_queries();
        my @commands;

        my @procs = keys(%$queries);

        while (my ($process, $objects) = each(%$queries)) {
            push(@commands, "echo Process: $process");
            while (my ($object, $insts_cntrs) = each(%$objects)) {
                my $cmd = "cm_client -p $process -a object-get-instances -o $object"
                    . " -i "
                    . join(',', keys(%{$insts_cntrs->{instances}}))
                    . " -c "
                    . join(',', keys(%{$insts_cntrs->{counters}}));
                push(@commands, $cmd);
            }
        }

        my $sample_cmd = join(';', @commands);

        return $sample_cmd;
    }


    sub sample_build {
        my $self    = shift();
        my $out     = shift();
        my $tracked = $self->_get_tracked();

        my @lines = split(/\r?\n/, $out);

        my $cur_proc;
        my @counters;
        my $sample;
        my $values;

        foreach my $line (@lines) {
            if ($line =~ /^Process: (.*)$/) {
                $cur_proc = $1;
            }
            elsif ($line =~ m{object_name:"([^"]+)" [ ]
                              instance_name:"([^"]+)" [ ]
                              counter_name:"([^"]+)" [ ]
                              value:"([^"]*)"
                             }msx) {
                my ($object, $instance, $counter, $value) = ($1, $2, $3, $4);

                # labels aren't included like with statistics so we
                # have to operate directly on the tracked hash
                my $label_hash =
                  Perf::Stats::Handler::CM::_get_label_values(
                      tree => $tracked,
                      path => [$cur_proc, $object, $instance, $counter, 'NO LABEL']
                      );
                my @labels = keys(%$label_hash);

                next
                    if (0 == scalar(@labels));

                if (1 < scalar(@labels) || q{} ne $labels[0]) {
                    # this is an array counter, break up values
                    my @values_arr = split(/,/, $value);

                    foreach my $label (@labels) {
                        my $cm_counter = $label_hash->{$label};
                        $values->{$cur_proc}->{$object}->{$instance}->{$counter}
                            ->{$label} =
                            $values_arr[$cm_counter->get_label_index()];

                        push(@counters,
                             {spec => {process  => $cur_proc,
                                       object   => $object,
                                       instance => $instance,
                                       counter  => $counter,
                                       label    => $label},
                              value => $values_arr[$cm_counter->get_label_index()],
                          });
                    }
                }
                else {
                    $values->{$cur_proc}->{$object}->{$instance}->{$counter}
                      ->{q{}} = $value;

                    push(@counters,
                         {spec => {process  => $cur_proc,
                                   object   => $object,
                                   instance => $instance,
                                   counter  => $counter,
                                   label    => q{}},
                          value => $value,
                      });
                }
            }
        }


        foreach my $counter (@counters) {
            my $c = $counter->{spec};
            my $v = $counter->{value};
            my $tracked = $self->tracked_get($c);

            if (defined($tracked)) {
                $Log->debug(  "Detected tracked counter '"
                            . $tracked->get_name()
                            . "'");

                my $base_counter = $tracked->get_base_counter();

                if (defined($base_counter) && q{} ne $base_counter) {
                    my $base_value = $values->{$c->{process}}
                                            ->{$c->{object}}
                                            ->{$c->{instance}}
                                            ->{$base_counter}
                                            ->{$c->{label}};

                    confess('Invalid value for base counter '
                        . " $c->{process}:$c->{object}:$c->{instance}:$base_counter"
                        ) if (!defined($base_value) || q{} eq $base_value);

                    $tracked->set_sum($v);
                    $tracked->set_count($base_value);

                    $v = $v . '/' . $base_value;
                }

                $tracked->set_value($v);
                if (!$tracked->is_hidden()) {
                    $sample->{$c->{process}}
                           ->{$c->{object}}
                           ->{$c->{instance}}
                           ->{$c->{counter}}
                           ->{$c->{label}} = $v;
                }
            }
        }

        return $sample;
    }

    sub make_names {
        my $self = shift();
        my $tree = shift();
        my %sample;

        while (my ($process, $objects) = each(%$tree)) {
            # Internal-only counters and aliases won't have any instances
            if ('HASH' ne ref($objects)) {
                $sample{$process} = $objects;
                next;
            }

            while (my ($object, $instances) = each(%$objects)) {
                while (my ($instance, $counters) = each(%$instances)) {
                    while (my ($counter, $labels) = each(%$counters)) {
                        while (my ($label, $value) = each(%$labels)) {
                            my $fullname =
                              "$process:$object:$instance:$counter";

                            if (q{} ne $label) {
                                $fullname .= ".$label";
                            }

                            $sample{$fullname} = $value;
                        }
                    }
                }
            }
        }

        return \%sample;
    } ## end sub make_names


    sub tracked_remove {
        my $self    = shift();
        my $counter = shift();

        confess('You must specify an argument')
            if (!$counter);
        confess("Argument must be a Perf::Stats::Counter::CM")
            if (!ref($counter) || !$counter->isa('Perf::Stats::Counter::CM'));

        my $tracked = $self->_get_tracked();

        my $removed = delete(
            $tracked->{$counter->get_cm_process()}
                    ->{$counter->get_cm_object()}
                    ->{$counter->get_cm_instance()}
                    ->{$counter->get_cm_name()}
                    ->{$counter->get_cm_label()});

        $Log->debug(  "Removed '"
                    . $self->get_source
                    . "' counter: "
                    . $removed->get_name)
          if ($removed);

        return $removed;
    }


    sub acceptable_apiset {
        my $self   = shift();
        my $apiset = shift();

        return $apiset->isa('NACL::APISet::Host::CLI::Freebsd');
    }


    sub create_connection {
        my $self    = shift();
        my $hostrec = shift();

        my $node_obj =
          NACL::C::Node->new(node    => $self->get_node(),
                             hostrec => $hostrec,);
        my $conn = $node_obj->apiset(category  => 'Node',
                                     interface => 'CLI',
                                     set       => 'Systemshell',);
        $conn->set_timeout(timeout => $Perf::Stats::TIMEOUT);
        return $conn;
    }


    # execute a command via ngsh and return parsed object
    sub ngsh {
        my $self       = shift();
        my $command    = shift();
        my $args       = shift();
        my $arg_list   = q{};

        while (my ($k, $v) = each(%$args)) {
            $arg_list .= " -$k $v";
        }

        my $connectrec = $self->get_nate_connection();
        my $output =
          $connectrec->execute(
                         command => "ngsh --invoke "
                           . "'set -showallfields true -showseparator \"X;X\" ;"
                           . " $command $arg_list'",
                         log_level => (
                                   ($Log->can('may_debug') && $Log->may_debug())
                                   ? 'comment'
                                   : 'debug'),);

        (my $method_name = $command) =~ s/[[:space:]]+/_/g;
        my $response =
          NACL::APISet::Response::CLI::CMode->new(
                                     output  => $output,
                                     command => $command,
                                     CLI_separator        => 'X;X',
                                     execution_time       => 1,
                                     method_name          => $method_name,
                                     output_compat_mirror => {},);

        return $response->get_parsed_output();
    }


    sub get_cluster_name {
        my $self     = shift();
        my $response = $self->ngsh('cluster identity show');

        return $response->[0]->{name};
    }


    sub get_cluster_nodes {
        my $self     = shift();
        my $response = $self->ngsh('cluster show');

        my @nodes = map { $_->{node} } @$response;

        return \@nodes;
    }


    sub get_node_names {
        my $self     = shift();
        my $response = $self->ngsh('system node show');

        my @node_names =
          map { $_->{node} } @{$response};

        return \@node_names;
    }


    sub get_aggr_list {
        my $self  = shift();
        my $node  = $self->get_node();
        my $lines = $self->ngsh('storage aggregate show', {nodes => $node});

        my @aggrs = map { $_->{aggregate} } @$lines;

        return \@aggrs;
    }


    sub get_aggr_disks {
        my $self  = shift();
        my $aggr  = shift();
        my $lines = $self->ngsh('storage disk show', {aggregate => $aggr});
        my $ds    = Perf::Stats::Handler::_process_disks($lines);

        return $ds;
    }
}

1;
