# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Handler for 'statistics' command output
## @status   review
## @pod here

package Perf::Stats::Handler::CM::Statistics;{

    use strict;
    use warnings;
    use Carp qw(confess);
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Readonly;
    use Data::Dumper;
    use List::MoreUtils qw(uniq);
    use Time::HiRes qw (time);
    use Perf::Utils qw(as_string);
    use Params::Validate qw(validate SCALAR HASHREF UNDEF);
    use NACL::APISet::Response::CLI::CMode;
    use Object::InsideOut qw(Perf::Stats::Handler::CM);
    use NATE::BaseException qw(:try);

    my @cache_command   :Field :Name('cache_command');
    my @sample_command  :Field :Name('sample_command');
    my @counter_flags    :Field :Name('counter_flags');
    my @sample_id   :Field  :Name('sample_id');
    my @sample_ready_timeout :Field :Name('sample_ready_timeout');
    my @start_time :Field  :Name('start_time');
    my @end_time :Field  :Name('end_time');

    # Constant used to limit the length of prepared commands. 30_000
    # was determined through experimentation.
    Readonly::Scalar our $CMD_LENGTH_LIMIT => 30_000;


    sub is_node_scoped {
        return 0;
    }

    sub get_fields {
        my ($self) = @_;
        my $response = $self->get_connection->statistics_show( fields => '?');
        my $out = $response->get_processed_output;
        my @fields;
        my @lines = split(/(?:\r?\n)+/, $out);
        for my $o (@lines) {
            # Each line contains a field name and a field description
            # separated by white space
            $o =~ s/^\s+//;
            $o =~ s/\s+$//;
            my ($field) = split(/\s+/, $o);
            push @fields, $field;
        }
        $Log->trace("Found fields:\n" . join("\n",@fields) );
        return \@fields;
    }

    sub sample_build_command {
        my ($self) = shift;
        my %args = @_;

        my $queries = $self->get_queries();

        my $sample_cmd;

         #Workaround for most cases for burt609182
        #until this is fixed the more counters listed in a query the slower the command returns.
        #This is really only noticeable with a large number of counters listed when people ask for all counters
        #in there regex.

        my @counters = split(/\|/,$queries->{counter});
        delete($queries->{counter}) if (scalar(@counters) > 100);
        #naming samples to avoid collision with multiple nodes
        my $sample_id = $self->_generate_sample_id() if($args{cmd_start});

        foreach my $field ('instance', 'counter', 'object') {
            if($args{cmd_start}) {
                $sample_cmd = 'statistics start -sample-id '.$sample_id.' ' . _concat_args($queries);
            }
            else{
                    $sample_cmd =
                    'statistics show -raw ' . _concat_args($queries);
            }
            # Limit the length of prepared commands.
            last
                if ($CMD_LENGTH_LIMIT > length($sample_cmd));

            # Command is too long, delete some parameters
            if ($field ne 'object') {
                $Log->trace("Command too long. Dropping $field");

                delete($queries->{$field});
            }
            # We have to at least have objects, take a chance and warn user.
            else {
                $Log->warn(
                          "'statistics show' command is exceeds maximum command"
                            . ' length. Errors may occur.');
            }
        }
        if(!$args{no_node}){
            my $node    = $self->get_node();
            #add node filter
            $sample_cmd.=' -node '.$node;
        }
        $sample_command[$$self] = $sample_cmd;
        return $sample_cmd;
    }

    sub _concat_args {
        my $queries = shift();
        my $sample_cmd;

        if (0 < scalar(keys(%$queries))) {
            while (my ($key, $value) = each(%$queries)){
                $sample_cmd .= "-$key $value ";
            }
        }
        else {
            confess('Asked to build a sample command but there are no counters to query');
        }

        return $sample_cmd;
    }

    sub sample_build {
        my $self =shift;
        my $out = shift;
        my %args = validate(@_,
            {cmd_start => {type => SCALAR | UNDEF, optional=>1},});

        my $sample = {};
        my $command = $sample_command[$$self];
        my @counters;
        my $values;
        my $tracked = $self->_get_tracked();
        $out =~ s/\cG//g; #strip bells for long commands  see burt562741
        my $response = NACL::APISet::Response::CLI::CMode->new(
            output         => $out,
            command        => $command,
            CLI_separator  => 'X;X',
            execution_time => 1,
            method_name    => 'statistics_show',
            output_compat_mirror => {},
        );
        my $lines = $response->get_parsed_output;

        my %counter_hash;
        my $elapsed;

        if($args{cmd_start}){
            $elapsed = $end_time[$$self] - $start_time[$$self];
        }
        for my $line (@$lines) {
            my $object = $line->{object};
            my $instance = $line->{instance};
            my $counter = $line->{counter};

            my $label_hash = Perf::Stats::Handler::CM::_get_label_values(
                        tree => $tracked,
                        path => [$object,$instance,$counter, 'NO LABEL']);
            my $value;
            if($line->{values} ne '-'){
                $value  = $line->{values};
            }
            else {
                #old behavior uses valueorlabel in the hash, SN.0 with the display-counters flag uses display-value
                $value = (exists($line->{valueorlabel}) ? $line->{valueorlabel}:$line->{'display-value'});
            }
            my @labels = keys(%$label_hash);
            next
              if (0 == scalar(@labels));
            my $cm_counter;
            my $prop;
            if (1 < scalar(@labels) || q{} ne $labels[0]) {
                foreach my $label (@labels) {
                    $cm_counter = $label_hash->{$label};
                    $prop = $cm_counter->get_all_types;
                    $values->{$object}->{$instance}->{$counter}
                      ->{$label} = $value->[$cm_counter->get_label_index()];
                    my $original= $value->[$cm_counter->get_label_index()];
                    my $val = $value->[$cm_counter->get_label_index()];

                    #strip units and clean up counters values.
                    if($args{cmd_start} && $prop !~ /(string|raw)/ ){
                        $original = $val = $1 if $val =~ /(\d+)/;
                        $original = $val = 0 if $val =~/-/;
                        $val = sprintf('%d',$elapsed * $val) if($prop =~ /rate/);
                    }
                    push(@counters,
                         {spec => {object   => $object,
                                   instance => $instance,
                                   counter  => $counter,
                                   label    => $label},
                          value => $val,
                          original => $original});
                }
            }
            else {
                $cm_counter = $label_hash->{q{}};
                $prop = $cm_counter->get_all_types;
                my $original = $value;

                #strip units and clean up counters values.
                if($args{cmd_start}  && $prop !~ /(string|raw)/ ) {
                    $original = $value = $1 if $value =~ /(\d+)/g;
                    $original = $value = 0 if $value =~/-/;
                    $value = sprintf('%d',$elapsed * $value) if($prop =~ /rate/);
                }
                $values->{$object}->{$instance}->{$counter}->{q{}} =
                  $value;

                push(@counters,
                     {spec => {object   => $object,
                               instance => $instance,
                               counter  => $counter,
                               label    => q{}},
                      value => $value,
                      original => $original});
            }
        }

        foreach my $counter (@counters) {
            my $c       = $counter->{spec};
            my $v       = $counter->{value};
            my $original = $counter->{original};
            my $tracked = $self->tracked_get($c);

            if (defined($tracked)) {
                $Log->debug(
                     "Detected tracked counter '" . $tracked->get_name() . "'");

                my $bc = $tracked->get_base_counter();
                if (defined($bc) && '-' ne $bc) {
                    my $base_value =
                      $values->{$bc->{object}}->{$bc->{instance}}->{$bc->{counter}} ->{$bc->{label}};
                    confess(  'Invalid value for base counter '
                            . " $bc->{object}:$bc->{instance}:$bc->{counter}")
                        if (!defined($base_value) || q{} eq $base_value);

                    # Since start/stop values are not raw we must convert to there raw values
                    # percent types = $value/100 * base_value
                    # average = $value * base_value
                    if($args{cmd_start}) {
                        my $type = $tracked->get_all_types;
                        $v = $v/100 if($type =~/percent/);
                        $v = $v*$base_value;
                    }

                    $tracked->set_sum($v);
                    $tracked->set_count($base_value);
                    $v = $v . '/' . $base_value;
                }
                $tracked->set_calc_value($original);
                $tracked->set_value($v);

                if (!$tracked->is_hidden()) {
                    $sample->{$c->{object}}->{$c->{instance}}->{$c->{counter}}
                      ->{$c->{label}} = $v;
                }
            }
        } ## end foreach my $counter (@counters)
        return $sample;
    }

    sub collect_all_objects {
        my ($self) = @_;
        my @objects;
        my $response = $self->get_connection->statistics_catalog_object_show();
        my $out = $response->get_parsed_output;
        for my $o (@$out){
            push @objects,$o->{object};
        }
        $Log->trace("Found objects:\n" . join("\n", @objects));
        return \@objects;
    }

    sub zero {
        # NOOP
    }

    # This works for now because counters of prop no-zero-values are
    # not actually hidden and so all counters should be visible all
    # the time. Should that change however this subroutine will stop
    # being reliable.
    sub cache_build_command {
        my ($self, $regexs, %params) = @_;

        confess "You must specify an argument" if (!$regexs);
        confess "Invalid argument" if (ref($regexs) ne 'ARRAY' || !@$regexs);

        #use no_node key if needed
        my $key = $params{key};
        $key = 'regular' if (!$key);

        my $objects = $self->get_node_objects(counters=> $regexs,key=> $key);
        my $instance_cmd_part = 'statistics catalog instance show';

        if($key eq 'regular'){
            $instance_cmd_part.= ' -node '. $self->get_node;
        }

        my $counter_cmd_part = 'statistics catalog counter show -describe';

        my $cache_query;
        for my $o (keys %$objects) {
            for my $regex (@{$objects->{$o}}){
                my $instance_cmd = $instance_cmd_part. " -object $o";
                $instance_cmd .= _add_filter('instance',$regex->{instance});

                my $counter_cmd = $counter_cmd_part. " -object $o";
                $cache_query.= $instance_cmd.';'.$counter_cmd.';';
            }

        }

        if ($CMD_LENGTH_LIMIT < length($cache_query)){
            $Log->warn(
            "'statistics catalog' command exceeds maximum command"
            .' length. Errors may occur.');
        }
        $cache_command[$$self] = $cache_query;
        return $cache_query;
    }

    sub _add_filter {
        my $key = shift;
        my $value = shift;

        my $filter = q();
        my $temp = $value;

        #temporarily remove '.*'  '-' '|'  from regex
        #By themselves any one of these chars could be valid for filtering
        #a string of 'workload-1'|'workload-2' is for instance is valid for CLI
        $value =~ s/\.\*//g;
        $value =~ s/\-//g;
        $value =~ s/\|//g;

        #After above removal if any other metachars exist in the regex do not filter.
        #Since quotemeta returns the escaped version of all metachars
        #this will evaluate to false if any metachars are left in the string
        #and an empty string will get returned.
        if ($value eq quotemeta($value)){
            #restore original string.
            $value = $temp;
            #Statistics cli supports '|' and '*" for filtering.
            #convert '.*' in regex to '*'
            $value =~ s/\.\*/*/g;
            $filter = " -$key $value";
        }
        return $filter;
    }

    sub cache_populate_wait {
        my ($self) = @_;
        my $out =  $self->execute_expect;

        my $command = delete $cache_command[$$self];
        confess "No cache query found!" if (!$command);
        $out =~ s/\cG//g; #strip bells for long commands  see burt562741
        my $response;
        my $instances;
        try {
            $response = NACL::APISet::Response::CLI::CMode->new(
                output         => $out,
                command        => $command,
                CLI_separator  => 'X;X',
                execution_time => 1,
                method_name    => 'statistics_catalog_instance_show',
                output_compat_mirror => {},
            );
            $instances = $response->get_parsed_output;
        }
        catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
            #its possible that a node won't have any instances and an exception will be thrown.
            $instances = [];
        };

        $out = $response->{not_parsed};
        my $counters;
        my ($curr_object_name, $curr_counter_name,
               $curr_counter_labels, $curr_counter_type,
               $curr_counter_base, $curr_counter_unit);
        my $label_procs = 0;
        for my $line (@$out) {
            if($line =~ /Object:\s+(.*)/) {
                $curr_object_name = $1;
                $label_procs = 0;
                next;
            }
            if($line =~ /  Counter:\s+(\S+)/) {
                $curr_counter_name = $1;
                undef $curr_counter_labels;
                undef $curr_counter_type;
                $label_procs = 0;
            }
            elsif ($line =~ /Labels:\s+(.*)/) {
                $label_procs = 1;
                $curr_counter_labels = $1;
                #range counters can lose spaces over multiple rows
                if($curr_counter_labels =~ /-$/){
                    $curr_counter_labels.=' ';
                }
                if($curr_counter_labels =~ /^-/){
                    $curr_counter_labels = ' '.$curr_counter_labels;
                }
                $counters->{$curr_object_name}->{$curr_counter_name}->{labels} = $curr_counter_labels;
            }
            elsif ($line =~ /Base Counter:\s+(.*)/) {
                $label_procs = 0;
                $curr_counter_base = $1;
                $counters->{$curr_object_name}->{$curr_counter_name}->{base_counter} = $curr_counter_base;
            }
            elsif ($line =~ /Properties:\s+(.*)/) {
                $label_procs = 0;
                $curr_counter_type = $1;
                $counters->{$curr_object_name}->{$curr_counter_name}->{prop}=$curr_counter_type;
            }
            elsif ($line =~ /Unit:\s+(.*)/) {
                $label_procs = 0;
                $curr_counter_unit = $1;
                $counters->{$curr_object_name}->{$curr_counter_name}->{unit} = $curr_counter_unit;
            }
            elsif($label_procs)
            {
                my $ll = $line;
                $ll =~ s/^\s+//;
                #range counters can lose spaces over multiple rows
                if($ll =~ /-$/){
                    $ll.=' ';
                }
                if($ll =~ /^-/){
                    $ll = ' '.$ll;
                }
                $counters->{$curr_object_name}->{$curr_counter_name}->{labels} .= $ll;
            }
        }

        my $cache = {};
        while (my ($object, $name) = each(%$counters)) {
            while (my ($name, $fields) = each(%$name)) {
                my $c = {object       => $object,
                    counter      => $name,
                    base_counter => $fields->{base_counter},
                    prop         => $fields->{prop},
                    unit         => $fields->{unit},};

                    if(defined $fields->{labels}) {
                        my $labels;
                        my @labels = split(/, /,$fields->{labels});
                        $labels->{columns} = [split(',',$labels[0])];
                        if(scalar(@labels) > 1){
                            $labels->{rows} = [split(',',$labels[1])];
                        }
                        $c->{labels} = $labels;
                    }
                $cache->{$object}->{counters}->{$name} = $c;
            }
        }


        for my $i (@$instances){
            $cache->{$i->{object}}->{instances}->{$i->{instance}} = 1;
        }
        #delete parsed header information
        delete $cache->{object};
        delete $cache->{Object};

        $self->cache_set($cache);
    }

    # There does not appear to be a limit on the length of a command
    # to ngsh; narrow query as much as possible
    sub _compute_queries {
        my $self = shift;
        # Build a tree out of the tracked counters
        my $objects   = {};
        my $instances = {};
        my $names     = {};
        for my $counter (@{$self->tracked_get()}) {
            $objects->{$counter->get_cm_object}     = 1;
            $instances->{$counter->get_cm_instance} = 1;
            $names->{$counter->get_cm_name}         = 1;
        }
        my %query;

        $query{object}   = _make_query(keys(%$objects));
        $query{instance} = _make_query(keys(%$instances));
        $query{counter} = _make_query(keys(%$names));

        $Log->trace("DBG: Queries computed: " . Dumper(\%query));
        $self->set_queries(\%query);
        return;
    }

    # query multiple entities by wrapping the names in "s to escape
    # reserved characters and then join all the names with a '|'
    sub _make_query {
        my @items = @_;

        return join('|', map { q{"} . $_ . q{"} } @items);
    }

    sub acceptable_apiset {
        my $self = shift();
        my $apiset = shift();

        return $apiset->isa('NACL::APISet::Node::CLI::CMode');
    }

    sub get_counter_class {
        my $self = shift();

        my $base_class    = $self->SUPER::get_counter_class();
        my $counter_class = "${base_class}::RPC";
        return $counter_class;
    }

    sub get_counter_fields {
        my $self = shift();

        my @fields = $self->SUPER::get_counter_fields();
        push(@fields, ('index', 'base_counter'));

        return @fields;
    }

    sub execute_start {
        my ($self, $command)= @_;
        #ensure the counter flags have been set before issuing command
        $self->counter_flags;

        $self->SUPER::execute_start($command);
    }
    sub counter_flags {
        my $self = shift();
        if (!defined($counter_flags[$$self])) {
            $counter_flags[$$self] = $self->set_counter_flags;
        }
        return $counter_flags[$$self];
    }

    sub set_counter_flags {
        my $self = shift;
        #need to add the qos_debug zero_suppression for all qos counters to come through
        my $command = 'statistics settings modify -counter-display all;'.
                    'run -node * "priv set test; qos_debug zero_suppression off"';
        my $conn = $self->get_nate_connection;
        my $out = $conn->execute(command => $command);
        if($out =~ /Error: invalid argument "-counter-display"/){
            $Log->warn("Display all counters flag is not available, expect missing tracked counters!");
        }
        if($out !~ /zero_suppression set to OFF/) {
            $Log->warn("QOS zero suppression may not have been turned off");
        }
        return 1;
    }
    sub pre_start {
        my ($self,%args) = @_;
        $sample_ready_timeout[$$self] = $args{ready_timeout};
        
        if (!$self->has_tracked()) {
            $Log->comment("No tracked counters, skipping sample on ".$self->get_node);
            return;
        }
        my $command = $self->sample_build_command(cmd_start => 1);
        $Log->comment(  'Perf::Stats is sending start command to '
                      . $self->get_node()
                      . ":\n$command");
        $self->execute_start($command);
        $start_time[$$self] = time;
        return;
    }

    sub post_start{
        my ($self) = @_;

        my $response = $self->execute_expect;
        $response =~ s/\cG//g; #strip bells at beginning see Burt#562741
        my $sample_id;
        if($response =~ /Sample-id:\s+(.*)/){
            $sample_id = $1;
        }else
        {
            confess("Unable to create sample: ".Dumper($response));
        }
        $sample_id[$$self] = $sample_id;
        $self->wait_for_samples_ready();
        return;
    }

    sub show {
        my ($self) = @_;
        $self->wait_for_samples_ready();
        my $command = 'statistics show -sample-id '.$sample_id[$$self];
        my $out;
        $self->execute_start($command);
        $out = $self->execute_expect;
        $out = $self->preprocess_output($out);

        $sample_command[$$self] = $command;
        my $sample = $self->sample_build($out,cmd_start=>1);

        return $sample;
    }

    sub sample_delete {
        my ($self) = @_;
        my $command = 'statistics samples delete -vserver * -sample-id '.$sample_id[$$self];
        $self->execute_start($command);
        my $out = $self->execute_expect;
    }

    sub pre_stop {
        my ($self) = @_;
        return {}
            if (!$self->has_tracked());
        my $command = 'statistics stop -sample-id '.$sample_id[$$self];
        $Log->comment(  'Perf::Stats is sending stop command to '
                      . $self->get_node());

        $self->execute_start($command);
        $end_time[$$self] = time;
    }

    sub post_stop {
        my ($self) = shift;
        my $out = $self->execute_expect;
        my $sample = $self->show;
        $self->sample_delete();

        for my $tracked_counter (@{$self->tracked_get()}){
            next
                if ($tracked_counter->is_hidden);


            $self->check_sample(sample => $sample, counter => $tracked_counter);

        }
        $sample = $self->make_names($sample);

        return $sample;

    }
    sub toggle {
        my ($self,$toggle) = @_;
        #ensure samples are removed.

        if(!$toggle){
            if(defined $sample_id[$$self]){
                $self->sample_delete;
            }
        }
        #state needs to get reset for counters
        $counter_flags[$$self] = undef;
        $self->SUPER::toggle($toggle);
    }

    sub reconnect {
        my $self = shift;
        #state needs to get reset for counters
        $counter_flags[$$self] = undef;
        $self->SUPER::reconnect();
    }
    sub wait_for_samples_ready {
        my $self=shift;
        my $stats_start = time;
        my $stats_timeout = $sample_ready_timeout[$$self];
        $Log->trace('Wait for statistics samples to be Ready');
        my $status = 'NA';
        while($status ne 'Ready'){
            my $response = $self->get_connection->statistics_samples_show('sample-id'=>$sample_id[$$self]);
            $status = $response->get_parsed_output()->[0]->{'sample-status'};
            if ($stats_start + $stats_timeout < time) {
                     confess ('Timed out waiting for statistics start '
                         . "to be ready. Waited $stats_timeout seconds.");
            }
            Tharn::snooze(1);
        }
        $Log->trace('Statistics samples are now ready.');
    }
    #generate sufficient randomness to account for multiple nodes/samples
    sub _generate_sample_id {
        my $self = shift;
        my $current_samples = $self->get_connection->statistics_samples_show();
        $current_samples = $current_samples->get_parsed_output();

        my $sample_id = 'sample_'.$$.'_'.int(rand(2**64));

        my %ids = map { $_->{'sample-id'} => 1 } @$current_samples;
        #check for collisions
        while($ids{$sample_id}) {
            $sample_id = 'sample_'.$$.'_'.int(rand(2**64));
        }

        return $sample_id;
    }

    sub get_node_objects {
        my ($self,%args) = @_;
        my $counters = $args{counters};
        my $key = $args{key};

        my $objects = $self->get_regex_object_pairs($counters);
        my $node_objects = {regular => {},no_node => {}};

        for my $o (keys %$objects) {
            try{
                my $response = $self->get_connection->statistics_catalog_counter_show(
                    object => $o, counter=>'node_name');
                $node_objects->{regular}->{$o} = $objects->{$o};
            }
            catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
                $node_objects->{no_node}->{$o} = $objects->{$o};
            }
        }

        if($key){
            return $node_objects->{$key};
        }else{
            return $node_objects;
        }
    }

    sub get_regex_object_pairs {
        my ($self, $counters) = @_;
        if (!$counters || !@$counters) {
            confess("You must specify counters");
        }

        # Query CM for known objects.
        my $list = $self->get_all_objects();
        # Match user-supplied counters against known objects.
        # Assume everything up to the first ':' in counter
        # is a regex specifying one or more objects.
        my %matched;
        my %regex_seen;
        $Log->debug("DEBUG: counters:\n" . Dumper($counters));

        foreach my $counter_re (@$counters) {
            my $regex     = $self->_process_regex($counter_re);
            my $object_re = $regex->{object};

            my $found = 0;
            foreach my $obj (grep m/^$object_re$/, @$list) {
                #$matched{$obj}++;    # Hash since we only want unique matches
                push @{$matched{$obj}}, $counter_re;
                $found++;
                $Log->trace("Found object '$obj' matching '$object_re'");
            }
            if (!$found) {
                confess "No objects found for regex '$object_re' in:\n"
                  . join("\n", @$list);
            }
        }
        return \%matched;
    }

    sub _destroy :Destroy {
        my $self = shift;
        return unless $$ == $self->get_pid;

        my $err = $@;
        eval{
            if(defined $sample_id[$$self]){
                $self->sample_delete();
            }
        };
        if ($@){
            $Log->comment("Error during destroy of Stats Handler: $@");
        }
        $@ = $err;
    }

}
1;
