# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Handler for CounterManager 'stats' command
## @status   review
## @pod here

package Perf::Stats::Handler::CM::Stats;{

    use strict;
    use warnings;
    use Carp qw(confess);
    use Tharn qw($Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use List::MoreUtils qw( uniq );
    use Perf::Stats::Handler;
    use Perf::Stats::Counter;
    use Perf::Stats::InstanceInfo;
    use Object::InsideOut qw(Perf::Stats::Handler::CM);

    my @cm_types        :Field;
    my @cm_unit         :Field;
    my @queries         :Field   :Name(queries)         :Std(queries);
    my @instance_info   :Field   :Name(instance_info);
    my @cache_command   :Field;


    our $MAXBUFSIZE = 1024*1024*1024; # 1g

    sub _init : Init {
        my ($self) = @_;
        $cm_types[$$self]      = {};
        $instance_info[$$self] = {};
    }

    sub cache_build_command {
        my ($self, $regexs) = @_;
        confess "You must specify an argument" if (!$regexs);
        confess "Invalid argument" if (ref($regexs) ne 'ARRAY' || !@$regexs);
        my $objects = $self->get_cm_objects($regexs);
        my $cache_query =
          join(';',
               map { "stats explain counters $_;stats list instances $_" }
                 @$objects);
        $cache_command[$$self] = $cache_query;
        return $cache_query;
    }

    sub cache_populate_wait {
        my ($self) = @_;
        $self->SUPER::cache_populate_wait;
        my $out = $self->cache_get;
        my @lines = split(/\r?\n/, $out);
        my ($curr_object_name,     $curr_counter_name,
            @curr_counter_columns, $curr_counter_type, $curr_counter_unit);
        my ($counters,       $instances);
        my ($explain_output, $instance_output);
        my @remove_counters;

        for my $line (@lines) {
            if ($line =~ /Counters for object name: (.*)\s*$/) {
                $curr_object_name = $1;
                $explain_output   = 1;
                $instance_output  = 0;
                next;
            }
            elsif ($line =~ /Instances for object name: (.*)\s*$/) {
                $curr_object_name = $1;
                $explain_output   = 0;
                $instance_output  = 1;
                next;
            }
            if ($explain_output) {
                if ($line =~ /Name:\s+(\S+)/) {
                    $curr_counter_name = $1;
                    $counters->{$curr_object_name}->{counters}->{$curr_counter_name} =
                            {
                                object => $curr_object_name,
                                counter => $curr_counter_name
                            };
                    undef $curr_counter_type;
                    undef $curr_counter_unit;
                }
                elsif ($line =~ /Properties:\s+(.*)/) {
                    $curr_counter_type = $1;
                    $counters->{$curr_object_name}->{counters}->{$curr_counter_name}->
                            {prop} = $curr_counter_type;
                    if($curr_counter_type =~ /no-display/){
                        push @remove_counters, {
                            object => $curr_object_name,
                            counter => $curr_counter_name};
                    }
                }
                elsif ($line =~ /Unit:\s+(.*)/) {
                    $curr_counter_unit = $1;
                    $counters->{$curr_object_name}->{counters}->{$curr_counter_name}->
                            {unit} = $curr_counter_unit;

                }
                elsif ($line =~ /Column names:\s+(.*)/) {
                    my @curr_counter_columns = split(/,\s+/, $1);
                    $counters->{$curr_object_name}->{counters}->{$curr_counter_name}->
                            {labels}->{columns} = \@curr_counter_columns;
                }
                elsif ($line =~ /Row names:\s+(.*)/) {
                    my @curr_counter_rows = split(/,\s+/, $1);
                    $counters->{$curr_object_name}->{counters}->{$curr_counter_name}->
                            {labels}->{rows} = \@curr_counter_rows;
                }
            } ## end if ($explain_output)
            elsif ($instance_output) {
                $line =~ s/^\s+//;
                $line =~ s/\s+$//;
                next if ($line eq '');
                $counters->{$curr_object_name}->{instances}->{$line} = 1;
            }
        } ## end for my $line (@lines)
        for my $c (@remove_counters){
            $Log->trace(
                        "Removing no-display counter $c->{counter} from cache"
                        );
            delete $counters->{$c->{object}}->{counters}->{$c->{counter}};
        }
        $self->cache_set($counters);
        return;
    } ## end sub cache_populate_wait


    sub sample_build_command {
        my ($self) = @_;
        my $query = join(" ",@{$self->get_queries()});
        return "stats show -R $query";
    }


    sub sample_build {
        my ($self, $out) = @_;

        confess("Cannot get stats: $out")
          if (   $out =~ m/Stats for identifier '.*?' was not found/
              || $out =~ /Unknown\/invalid object name/
              || $out =~ /No counter/);

        # Parse output
        my $sample = {};

        for my $line ( split(/\r?\n/, $out) ){
            next
                if ($line =~ /^run -node local/);

            if  ($line =~ /^(.*):([^:]*)\s*$/){
                my $c = $1;
                my $v = $2;
                # When no second match is made, $2 is set to '' instead of undef...

                #hack for ontap_version counter with ':' in value string
                if($line =~ /^(system:system:ontap_version):(.*)/){
                    $c = $1;
                    $v = $2;
                }

                my ($sum,$count);
                if ($v =~ /(\d+)(\/(\d+))?%?/){
                    $sum = $1;
                    $count = $3;
                }
                my $prepared_counter =
                  Perf::Stats::Handler::CM::_string_to_spec(string => $c,
                                                            is_re  => 0);

                my $tracked_counter = $self->tracked_get($prepared_counter);
                if ($tracked_counter){
                    $Log->trace("Detected tracked counter '" . $tracked_counter->get_name ."'");
                    $tracked_counter->set_value($v);
                    #
                    # For counters of type 'average' or 'percent', store
                    # the sum and count separately for easier processing
                    # later on
                    #
                    confess("No type found in counter:\n" . Dumper($tracked_counter->dump))
                        if (!$tracked_counter->has_type);
                    if ($tracked_counter->is_of_type('average') || $tracked_counter->is_of_type('percent')){
                        if (defined($sum) && defined($count)){
                            $tracked_counter->set_sum($sum);
                            $tracked_counter->set_count($count);
                        }
                        else{
                            my $all_types = $tracked_counter->get_all_types;
                            confess("Unable to parse counter '$c' of type '$all_types': $v");
                        }
                    }
                    if (!$tracked_counter->is_hidden){
                        # $sample->{$c} = $v;
                        $sample->{$prepared_counter->{object}}
                            ->{$prepared_counter->{instance}}
                                ->{$prepared_counter->{counter}}
                                    ->{$prepared_counter->{label}} = $v;
                    }
                }
            }
        }
        return $sample;
    }

    sub zero {
        # NOOP
    }

    # Because of rsh limitations we can't just blindly ask for every
    # counter individually Deciding how to group them together is the
    # goal of _compute_queries
    sub _compute_queries {
        my $self = shift;

        # Build a tree out of the tracked counters
        my $tree = {};
        for my $counter (@{$self->tracked_get()}){
            $tree->{$counter->get_cm_object}->{$counter->get_cm_instance}->{$counter->get_cm_name} = 1;
        }

        my $tqueries = [];

        for my $object (keys %$tree){
            # every query will at least mention the object
            my $query = $object;

            my @instances = keys %{$tree->{$object}};
             if ($object ne 'disk' && @instances < 4) {
                for my $instance (@instances){
                    my @counters = keys %{$tree->{$object}->{$instance}};
                    if (@counters > 3){
                        push @$tqueries, $query . ":$instance";
                    }
                    else{
                        for my $counter (@counters){
                            push @$tqueries, $query . ":$instance:$counter";
                        }
                    }
                }
            }
            else{
                my $counters;
                for my $instance (@instances){
                    for my $counter (keys %{$tree->{$object}->{$instance}}){
                        $counters->{$counter}=1;
                    }
                }
                if (keys %$counters > 3){
                       push @$tqueries, $query;
                }
                else{
                    for my $counter (keys %$counters){
                        push @$tqueries, $query . ":*:$counter";
                    }
                }
            }
        }
        $Log->trace("DBG: Queries computed: @$tqueries");
        $self->set_queries($tqueries);
        return;
    }


    sub get_all_objects {
        my ($self) = @_;
        my @list;
        $self->execute_start('stats list objects');
        my $out = $self->execute_expect;
        while ($out =~ m/^\s+(\w+)\s*$/mg) {
            push @list, $1;
        }
        return \@list;
    }


    sub get_default_counter_type {
        # No default
        return;
    }
}
1;
