package Perf::Stats::Handler::CM::ZAPI;
{
    use strict;
    use warnings;
    use Carp qw(confess);
    use Tharn qw($Log);

    BEGIN {
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Data::Dumper;
    use List::MoreUtils qw( uniq );
    use Perf::Utils qw( as_string);
    use Params::Validate qw(validate SCALAR HASHREF OBJECT UNDEF);
    use NACL::APISet::Response::ZAPI::CMode;
    use Object::InsideOut qw(Perf::Stats::Handler::CM);
    use XML::Simple;
    use Tharn;
    use NATE::STAF::Peer qw($Parent);
    use NATE::Process;
    use NATE::ProcManager;

    my @manager : Field : Name('manager');
    my @is_disconnected :Field :Name('is_disconnected');
    my @log_level :Field :Name('log_level');

    sub cache_build_command {
        my ($self, $regexs, %params) = @_;
        confess "You must specify an argument" if (!$regexs);
        confess "Invalid argument" if (ref($regexs) ne 'ARRAY' || !@$regexs);
        #use no_node key if needed
        my $key = $params{key};
        $key = 'regular' if (!$key);

        my $objects = $self->get_node_objects(counters=> $regexs,key=> $key);

        my @queries;
        for my $object (@$objects) {
            my $args = {objectname => $object};
            my $command = {command => 'perf_object_counter_list_info',
                           args    => $args};

            push @queries, $command;
        }

        return \@queries;
    }

    sub cache_populate_wait {
        my ($self) = @_;

        my $out = $self->execute_expect();
        $out = $self->preprocess_output($out);

        my $counters = {};

        #since ZAPIs can take multiple objects have multiple response objects to parse
        for my $o (@$out) {
            my $obj       = $self->_parse_xml($o->{command});
            my $lines_arr = $o->{parsed};
            $lines_arr = $lines_arr->[0]->{counters}[0]->{'counter-info'};
            for my $counter (@$lines_arr){
                my %c;
                $c{counter} = $counter->{name};
                $c{object} = $obj;
                $c{prop} = $counter->{properties};
                $c{unit} = $counter->{unit};

                if(defined $counter->{labels}){
                    my $l =  $counter->{labels}->[0]->{'label-info'};
                    my $labels;
                    if (ref($l) eq 'ARRAY') {
                        $labels->{columns} = [split(',', $l->[0])];
                        $labels->{rows} = [split(',',$l->[1])];
                    }
                    else {
                        $labels->{columns} = [split(',', $l)];
                    }
                    $c{labels} = $labels;
                }

                if(defined $counter->{'base-counter'}){
                    $c{base_counter} = $counter->{'base-counter'};
                }
                $counters->{$obj}->{counters}->{$counter->{name}} = \%c;
            }
            $counters->{$obj}->{instances} = $self->get_all_instances($obj);
        }
        $self->cache_set($counters);
    }

    #get all instances of object
    sub get_all_instances {
        my ($self, $object,%params) = @_;
        confess "You must specify an argument" if (!$object);

        my $args = {objectname => $object};

        my $instance = $self->_get_command_name();
        my $command  = $instance->{command};
        #cluster wide counters can't be filtered
        if(!$params{no_node}){
            $args->{'filter-data'} = "node_name=".$self->get_node();
        }
        #call instances command;
        my $next_tag;
        my $output;
        #Iterate through instances call to get all instances
        do {
            $args->{tag} = $next_tag;
            my $out    = $self->get_connection->$command(%$args);
            $out = $out->get_parsed_output();
            $next_tag = $out->[0]->{'next-tag'};
            push(@$output, $out->[0]->{$instance->{parse}}[0]->{'instance-info'});
        } while($next_tag);

        my $instances;
        for my $out (@$output){
            for my $o (@$out) {
                $instances->{$o->{name}} = 1;
            }
        }
        return $instances;
    }

   sub execute_start {
        my ($self, $command) = @_;
        confess "You must pass a command" if (!$command);
        my @procs;

        for my $com (@$command) {
            my $code = sub { $self->_run_command($com); };
            my $object = $com->{args}->{objectname};
            my $proc =  {codespec => $code};

            push(@procs, $proc);
        }
        $manager[$$self] = NATE::ProcManager->new(
            proc_info => \@procs,
            background => 1,
            parallel =>1);
        $manager[$$self]->start();
    }

    sub _run_command {
        my ($self, $com) = @_;

        my $c = $com->{command};
        my $a = $com->{args};

        $a->{log_level} = $self->log_level;
        $Log->comment("Sending Command:" . $c . "(" .  Dumper($a) . ")");
        my $response = $self->get_connection->$c(%$a);
        my $out;
        $out->{parsed} = $response->get_parsed_output();
        $out->{command} = $response->get_command;

        $Parent->message_put(message=> $out, type=>'CM_DATA');
        sleep;
    }

    sub execute_expect {
        my ($self) = @_;
        my @procs = $manager[$$self]->get_live_children();

        my @response;
        for my $proc (@procs) {
            my $line;

            my $resp = $proc->message_get(type=>'CM_DATA', timeout=> '1200');

            push @response, $resp;
        }
        #stop all children and destroy objects
        $manager[$$self]->stop();
        $manager[$$self]->wait(timeout => '1200');
        $manager[$$self]->destroy();

        return \@response;
    }


    sub collect_all_objects {
        my ($self) = @_;
        my @list;

        my $out = $self->get_connection->perf_object_list_info(log_level=>$self->log_level);
        $out = $out->get_parsed_output();

        #navigate to level needed
        $out = $out->[0]->{objects}[0]->{'object-info'};

        for my $o (@$out) {
            push @list, $o->{name};
        }
        return \@list;
    }

    sub sample_build_command {
        my ($self,%params) = @_;
        my $queries = $self->get_queries();
        my @sample_cmds;

        if (0 < scalar(keys(%$queries))) {
            for my $o (keys %$queries) {
                my $args = {objectname => $o,
                            instances  => $queries->{$o}->{instances},
                            counters   => $queries->{$o}->{counters}};
                if(!$params{no_node}){
                    $args->{'filter-data'} =  "node_name=".$self->get_node();
                }
                my $command = {command => 'perf_object_get_instances',
                               args    => $args};

                push @sample_cmds, $command;
            }

        }
        else {
            confess(
                'Asked to build a sample command but there are no counters to query'
                );
        }
        return \@sample_cmds;
    }

    sub _compute_queries {
        my $self = shift;

        # Build a tree out of the tracked counters
        my $objects = {};

        for my $counter (@{$self->tracked_get()}) {

            my $ob = $counter->get_cm_object;
            $objects->{$ob} = {instances => [],
                               counters  => []}
              unless exists $objects->{$ob};
            my %in = map { $_ => 1 } @{$objects->{$ob}->{instances}};
            my %co = map { $_ => 1 } @{$objects->{$ob}->{counters}};

            push(@{$objects->{$ob}->{instances}}, $counter->get_cm_instance)
              unless exists $in{$counter->get_cm_instance};
            push(@{$objects->{$ob}->{counters}}, $counter->get_cm_name)
              unless exists $co{$counter->get_cm_name};
        }
        $Log->trace("DBG: Queries computed: " . Dumper($objects));
        $self->set_queries($objects);
        return;
    }

    sub sample_build {
        my ($self, $out) = @_;
        my $tracked = $self->_get_tracked();
        my $sample;
        my @counters;
        my $values;

        for my $i (@$out) {
            my $object = $self->_parse_xml($i->{command});
            my $lines = eval{$i->{parsed}->[0]->{instances}[0]->{'instance-data'}};
            if($@)
            {
                confess("Error navigating parsed zapi structure: $@\n"
                    . Dumper($i->{parsed}));
            }
            confess("No Data found in zapi:". Dumper($i->{parsed}))
                if(!defined $lines);

            for my $j (@$lines) {
                my $instance = $j->{name};
                for my $c (@{$j->{counters}->[0]->{'counter-data'}}) {
                    my $counter = $c->{name};
                    confess("Could not read counter name") if (!defined $counter);

                    my $label_hash = Perf::Stats::Handler::CM::_get_label_values(
                      tree => $tracked,
                      path => [$object, $instance, $counter, 'NO LABEL']
                      );
                    my $value  = $c->{value};
                    #null string counters get returned as [{}] in from NACL Parser
                    #see burt654609
                    #converting these to '' to keep delta from breaking
                    if(ref $value ne '') {
                        if(ref $value eq 'ARRAY') {
                            if(ref($value->[0]) eq 'HASH' and !%{$value->[0]}){
                                $value = q{};
                            }
                            else {
                                confess('Invalid format of returned counter value: \n'.Dumper($value));
                            }
                        }
                        else {
                            confess('Invalid format of returned counter value: \n'.Dumper($value));
                        }
                    }
                    my @labels = keys(%$label_hash);

                    next
                      if (0 == scalar(@labels));

                    if (1 < scalar(@labels) || q{} ne $labels[0]) {

                        # this is an array counter, break up values
                        my @values = split(/,/, $value);

                        foreach my $label (@labels) {
                            my $cm_counter = $label_hash->{$label};
                            $values->{$object}->{$instance}->{$counter}
                              ->{$label} = $values[$cm_counter->get_label_index()];

                            push(@counters,
                                 {spec => {object   => $object,
                                           instance => $instance,
                                           counter  => $counter,
                                           label    => $label},
                                  value => $values[$cm_counter->get_label_index()]});
                        }
                    }
                    else {
                        $values->{$object}->{$instance}->{$counter}->{q{}} =
                          $value;

                        push(@counters,
                             {spec => {object   => $object,
                                       instance => $instance,
                                       counter  => $counter,
                                       label    => q{}},
                              value => $value});
                    }
                } ## end for my $c (@{$j->{counters...
            } ## end for my $j (@$lines)
        } ## end for my $i (@$out)

        foreach my $counter (@counters) {
            my $c       = $counter->{spec};
            my $v       = $counter->{value};
            my $tracked = $self->tracked_get($c);

            if (defined($tracked)) {
                $Log->trace(
                     "Detected tracked counter '" . $tracked->get_name() . "'");

                my $bc = $tracked->get_base_counter();

                if (defined($bc) && q{} ne $bc) {
                   my $base_value =
                      $values->{$bc->{object}}->{$bc->{instance}}->{$bc->{counter}} ->{$bc->{label}};
                    confess(  'Invalid value for base counter '
                            . " $bc->{object}:$bc->{instance}:$bc->{counter}")
                        if (!defined($base_value) || q{} eq $base_value);

                    $tracked->set_sum($v);
                    $tracked->set_count($base_value);
                    $v = $v . '/' . $base_value;

                }

                $tracked->set_value($v);
                if (!$tracked->is_hidden()) {
                    $sample->{$c->{object}}->{$c->{instance}}->{$c->{counter}}
                      ->{$c->{label}} = $v;
                }
            }
        } ## end foreach my $counter (@counters)

        return $sample;
    } ## end sub sample_build

    sub _parse_xml {
        my ($self, $command) = @_;

        my $xml = XML::Simple->new();;
        my $parsed_output = eval{$xml->XMLin($command)};
        if($@)
        {
            confess("Could not parse xml command:"
                .Dumper($command)."\nError: $@");
        }

        return $parsed_output->{objectname};
    }

    sub _get_command_name {
        my $self    = shift;
        my $command = 'perf_object_instance_list_info';
        $command .= ($self->node_is_ng ? '_iter' : '');
        my $part = ($self->node_is_ng ? 'attributes-list' : 'instances');
        my $c = {command => $command,
                 parse   => $part};

        return $c;
    }

    sub get_counter_fields {
        my $self = shift();

        my @fields = $self->SUPER::get_counter_fields();

        push(@fields, ('index', 'base_counter'));

        return @fields;
    }

    sub get_counter_class {
        my $self = shift();

        my $base_class    = $self->SUPER::get_counter_class();
        my $counter_class = "${base_class}::RPC";

        return $counter_class;
    }

    sub reconnect {
    }
    sub acceptable_apiset {
        my $self = shift();
        my $apiset = shift();

        return ($apiset->isa('NACL::APISet::Node::ZAPI::7Mode')
             || $apiset->isa('NACL::APISet::Node::ZAPI::CMode'));
    }
    sub create_connection {
        my $self    = shift();
        my $hostrec = shift();

        my $node_obj =
            NACL::C::Node->new(node    => $self->get_node(),
                             hostrec => $hostrec,);

        my $conn = $node_obj->apiset(category  => 'Node',
                                     interface => 'ZAPI',
                                     set       => ($self->node_is_ng ? 'CMode':'7Mode'),);
        $conn->set_timeout(timeout => $Perf::Stats::TIMEOUT);
        #quick command to verify connectivity
        $conn->system_get_ontapi_version(log_level => $self->log_level);


        #ZAPI ApiSets do not track connectivity
        return $conn;
    }
    sub get_node_names {
        my $self       = shift();
        my $conn = $self ->get_connection();

        my @node_names;
        my $response;

        if ($conn->isa('NACL::APISet::Node::ZAPI::CMode')) {
            return $self->get_cluster_nodes();
        }
        else {
            $response = $conn->cf_get_partner ();
            @node_names =
                      ($self->get_node(), $response->get_parsed_output()->[0]->{partner});
        }
        return \@node_names;
    }

    sub get_cluster_name {
        my $self          = shift();
        my $response  = $self->get_connection->cluster_identity_get('desired-attributes' => {'cluster-name' => ''},
                log_level=>$self->log_level);
        my $cluster     = $response->get_parsed_output->[0]->{'attributes'}[0]->{'cluster-identity-info'}[0]->{'cluster-name'};

        return $cluster;
    }
    sub get_cluster_nodes {
        my $self = shift();

        my $response = $self->get_connection->system_node_get_iter(
                            "desired-attributes" => {node => ''},
                            log_level=>$self->log_level);
        my $out = @{$response->get_parsed_output()}[0]->{'attributes-list'}[0]->{'node-details-info'};

        my @node_names = map { $_->{node} } @{$out};

        return \@node_names;
    }
    sub get_aggr_list {
        my $self  = shift();
        my $node  = $self->get_node();

        my @aggrs;
        my $query = {nodes => [$node]};
        my $attributes = {'aggregate-name' => '', nodes => []};

        my $response = $self->get_connection->aggr_get_iter(
            'desired-attributes' => $attributes,query => $query, log_level=>$self->log_level);

        my $lines =$response->get_parsed_output()->[0]->{'attributes-list'}[0]->{'aggr-attributes'};

        @aggrs = map { $_->{'aggregate-name'}} @$lines;

        return @aggrs if (wantarray);
        return \@aggrs;
    }
    sub get_aggr_disks {
        my $self  = shift();
        my $aggr  = shift();
        my $query = { 'disk-raid-info' =>{'disk-aggregate-info' => {'aggregate-name' => $aggr}}};
        my $attributes = {'disk-name' => '','disk-raid-info' => {position => '', 'disk-uid' => ''}};

        my $resp = $self->get_connection->storage_disk_get_iter('desired-attributes' => $attributes, query => $query);
        $resp = $resp->get_parsed_output()->[0]->{'attributes-list'}[0]->{'storage-disk-info'};

        my @ds;
        for my $d (@$resp) {
            my $dh = {type => $d->{'disk-raid-info'}[0]->{'position'}};
            my $name = $d->{'disk-name'};
            if ($name =~ /([^:]+):(.*)/) {
                $dh->{node} = $1;
                $dh->{name} = $2;
            }
            else {
                confess "Unable to parse disk name '$name'";
            }
            $dh->{instance} = $d->{'disk-raid-info'}[0]->{'disk-uid'};
            push @ds, $dh;
        }

        return \@ds;
    }

    sub get_node_objects {
        my ($self,%args) = @_;
        my $counters = $args{counters};
        my $key = $args{key};

        my $objects = $self->get_cm_objects($counters);

        my $node_objects = {regular => {},no_node => {}};
        for my $o (@$objects){
            my $response = $self->get_connection->perf_object_counter_list_info(
                            objectname => $o,log_level=>$self->log_level);
            $response = $response->get_parsed_output();
            my $out = $response->[0]->{counters}[0]->{'counter-info'};
            my $found = 0;
            for my $c (@$out){
                if($c->{name} eq 'node_name'){
                    $node_objects->{regular}->{$o} = 1;
                    $found = 1;
                    last;
                }
            }
            if(!$found){
                $node_objects->{no_node}->{$o} = 1;
            }
        }
        if($key){
            my @objects = keys %{$node_objects->{$key}};
            return \@objects;
        }else{
            return $node_objects;
        }
    }
    #ZAPIs aren't persistant connections
    #tracking "state"
    sub disconnect {
        my $self = shift;
        #ZAPI ApiSets do not track connectivity
        $is_disconnected[$$self] = 1;
    }
    sub is_disconnected {
        my $self = shift;
        return $is_disconnected[$$self];
    }
    sub set_connection {
        my ($self,$conn) = @_;
        $is_disconnected[$$self] = 0;

       $self->SUPER::set_connection($conn);
    }

    #With zapi commands a log_level = comment
    #dumps absolutely everything. Debug reduces the amount of output
    sub log_level {
        my $self = shift();
                if (!defined($log_level[$$self])) {
            $log_level[$$self] = $self->get_log_level;
        }
        return $log_level[$$self];
    }

    sub get_log_level {
        my $self = shift;
        my $level = 'debug';
        if($Log->may_debug()){
            $level = 'comment';
        }
        return $level;
    }
}
1;
