# -*- perl -*-
#
## @author   jroussin@netapp.com
## @summary  Internal class
## @status   review

package Perf::Stats::InstanceInfo::disk;{

    use strict;
    use warnings;
    use Tharn qw(say subtest param host $Log);
    BEGIN{
        my @c = caller(2);
        $Log->debug("Loading " . __FILE__ . " from $c[1]:$c[2]");
    }
    use Params::Validate qw(validate SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
    use Carp qw(confess longmess);
    use Data::Dumper;
    use Object::InsideOut qw( Perf::Stats::InstanceInfo );

    # The package should be loaded already and the line
    # below might load the wrong file so don't do it
    #use Perf::Stats;

    my @instance  :Field :Name('instance');
    my @disks     :Field :Name('disks');
    my @aggregate :Field :Name('aggregate');
    my @raid_role :Field :Name('raid_role');

    sub _init :Init {
        my ($self) = @_;
        $instance[$$self]     = {};
        $disks[$$self]        = {};
        $aggregate[$$self]    = {};
        $raid_role[$$self]    = {};
        my $hostp = host($self->get_node);
        if ($hostp->hosttype !~ /-ng-/){
            $self->_map_instances_to_devices;
        }
    }

    sub _map_instances_to_devices {
        my $self = shift;

        # See if display_name can be found
        eval{
            my $tmp_stats =
              Perf::Stats->new(fc   => $self->get_handler()->get_connection(),
                               node => $self->get_node,);
            $tmp_stats->add([{object => 'disk', counter => 'display_name'}]);
            my $sample = $tmp_stats->sample;
            for my $counter_string (keys %$sample){
                next if ($counter_string =~ /^_/);
                my $counter = $tmp_stats->_get_handler($counter_string)->tracked_get($counter_string);
                my $device = $sample->{$counter_string};
                my $counter_instance = $counter->get_cm_instance;
                $Log->trace("Storing instance '$counter_instance' for device '$device'");
                $instance[$$self]->{$device} = $counter_instance;
            }
        };
        die $@ if ($@);


        my $node_obj =
        NACL::C::Node->new(node    => $self->get_node);
        my $storage_disks = $node_obj->apiset()->storage_show_disk(all => 1)->get_parsed_output();
        for my $disk (@$storage_disks){
            next if(!defined $disk->{disk});

            my $device = $disk->{disk};
            next if ($instance[$$self]->{$device});

            my $struct = $disk;
            my $wwn = $struct->{wwn};
            my $instance_string = $wwn;
            # Remove the semi-colons since they don't match with the format used in the counter names
            $instance_string =~ s/\://g;
            # Add a semi-colon every 8 character
            $instance_string =~ s/(\S{8})(\S)/$1:$2/g;
            # Finish formatting
            $instance_string = uc($instance_string) . (":00000000" x 8);
            #$self->{device}->{$instance} = $device;
            $Log->trace("Storing instance '$instance_string' for device '$device'");
            $instance[$$self]->{$device} = $instance_string;
        }

    }

    #
    # This only queries the list of disks for an aggregate once
    # Should the aggregate get more (or less) disks, stats will be wrong
    #
    sub aggregate {
        my $self = shift;
        my $instance = shift;
        my $node = $self->get_node;

        my $aggr = $aggregate[$$self]->{$instance};
        return $aggr if ($aggr);

        $Log->trace("No match for disk instance '$instance' in:\n" . Dumper($aggregate[$$self]));

        #my $fc = $self->get_connection;
        my @aggrs = $self->_aggrs;
        $Log->trace("Found aggregates: @aggrs ");
        for my $aggr (@aggrs){
            my $aggr_disks = $disks[$$self]->{$aggr};
            if (!$aggr_disks){
                my @ds = $self->_aggr_disks( $aggr );
                $aggr_disks = \@ds;
                $disks[$$self]->{$aggr} = $aggr_disks;
            }
            for my $disk_struct (@$aggr_disks){
                my $disk = $disk_struct->{name};
                my $disk_node = $disk_struct->{node};
                $disk_node = $node if (!$disk_node);
                # 10-mode
                my $disk_instance = $disk_struct->{instance};
                # 7-mode
                $disk_instance = $instance[$$self]->{$disk} if (!$disk_instance);
                if (!$disk_instance){
                    Carp::confess "Unable to find instance name for disk '$disk' on '$disk_node'";
                }
                $aggregate[$$self]->{$disk_instance} = $aggr;
                $aggregate[$$self]->{$disk} = $aggr;
                my $disk_type = $disk_struct->{type};
                $disk_type =~ s/dparity/parity/;
                $raid_role[$$self]->{$disk_instance} = $disk_type;
                $raid_role[$$self]->{$disk} = $disk_type;
                $Log->trace("Disk '$disk' on '$disk_node' (a.k.a '$disk_instance') belongs to " .
                            "aggr '$aggr' with role '$disk_type'");
            }
        }
        $aggr = $aggregate[$$self]->{$instance};
        if (!$aggr){
            # TODO: Handling spare disks this way means that the 'aggregate' method will
            # have to rediscover all aggregate disks every time for each spare disk.
            # Instead, figure out a way to get all spare disks at once
            $aggr = '#spare#';
            $aggregate[$$self]->{$instance} = $aggr;
        }
        return $aggr;
    }

    # raid_role is 'data' or 'parity'
    sub raid_role {
        my $self = shift;
        my $instance = shift;

        my $raid_role = $raid_role[$$self]->{$instance};
        if ($raid_role){
            return $raid_role;
        }

        # Turns out the aggregate subroutine will load all the raid_roles too
        my $disk_aggr = $self->aggregate($instance);
        if ($disk_aggr eq '#spare#'){
            $raid_role = '#none#';
            $raid_role[$$self]->{$instance} = $raid_role;
        }
        else{
            $raid_role = $raid_role[$$self]->{$instance};
            if (!$raid_role){
                Carp::confess "Unable to detect raid role of disk instance '$instance' in aggr '$disk_aggr'";
            }
        }
        return $raid_role;
    }

    # raid_group looks like /aggr0/plex0/rg0
    sub raid_group {
        # TODO
    }

    # Must return an array of hashes like this {name=>, type=>}
    sub _aggr_disks {
        my ($self, $aggr) = @_;
        confess "You must pass an aggregate name" if (!$aggr);

        my $ds = $self->get_handler()->get_aggr_disks($aggr);

        return @$ds if (wantarray);
        return $ds;
    }

    sub _aggrs {
        my ($self) = @_;

        my $aggrs = $self->get_handler()->get_aggr_list();

        return @$aggrs if (wantarray);
        return $aggrs;
    }
}
1;
