# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Perf/Sysstat.pm#1 $
#
## @author   dl-perf-qa@netapp.com
## @summary  Collect, parses, and sends alerts on sysstat output in the background
## @status   review
## @pod here

=head1 NAME

    Perf::Sysstat - Wrapper for filer "sysstat" command. Works in 7-mode and 10-mode.
                    Allows for on-the-fly querying and real-time monitoring of output

=head1 SYNOPSIS

    use Perf::Sysstat;
    my $sysstat = Perf::Sysstat->new(   node => $FILER,
                                        client => $CLIENT,
                                        options => '-x',
                                        interval => 1,
                                    );
    $sysstat->start;
    # Do something
    ...
    my $output = $sysstat->get_output;
    print "Sysstat output so far:\n$output\n";
    # Do some more
    ...
    my $rows = $sysstat->get_rows;
    $sysstat->stop;


=head1 DESCRIPTION

    Perf::Sysstat in its simplest form allows users to gather 'sysstat' output in the background
    while their tests are running. Users may query the object to get the output so far while
    'sysstat' keeps running.
    This class also provides support for real-time monitoring of the output through the use of
    STAF messages and callback subroutines. See L<Perf::Sysstat::CP> for a detailed example.

=head1 METHODS

=cut

package Perf::Sysstat;{

    use strict;
    use warnings;
    use Tharn qw(subtest $Log);
    use Carp qw(confess);
    use Storable qw(nfreeze);
    use MIME::Base64;
    use NATE::Events qw( listener_add listener_remove eventloop_drain );
    use Data::Dumper;
    use Time::HiRes qw( time );
    # The line below isn't technically necessary but help catch errors
    # at compile time in that module
    use Perf::Sysstat::Remote;
    use Perf::Sysstat::Row;
    use Object::InsideOut;

    my @subtest     :Field  :Name('subtest')
                            :Get(Name=>'get_subtest',   Perm=>'Restricted');
    my @node        :Field  :Arg(Name=>'node',          Mandatory=>1);
    my @client      :Field  :Arg(Name=>'client',        Mandatory=>1);
    my @options     :Field  :Arg(Name=>'options',       Mandatory=>0)
                            :Get('get_options')
                            :Set(Name=>'set_options',   Perm=>'Restricted');
    my @interval    :Field  :Arg(Name=>'interval',      Mandatory=>0, Default=>1)
                            :Get('get_interval')
                            :Set(Name=>'set_interval',  Perm=>'Restricted');
    my @count       :Field  :Arg(Name=>'count',         Mandatory=>0);
    my @processor   :Field  :Arg(Name=>'processor',     Mandatory=>0)
                            :Type('CODE')
                            :Set(Name=>'set_callback',  Perm=>'Restricted');
    my @maxbufsize  :Field  :Arg(Name=>'maxbufsize',    Mandatory=>0, Default=>'50m');
    my @debug       :Field  :Arg(Name=>'debug',         Mandatory=>0, Default=>0)
                            :Get('is_debug');
    my @pid         :Field  :Name('pid');
    my @started     :Field  :Name('started') :Get('is_started') :Default(0);


    my %init_args :InitArgs = (runid => {Mandatory => 0,
                                         Type      => 'scalar',},);


=head2 new

Constructor for Perf::Sysstat objects.

    my $sysstat = Perf::Sysstat->new(   node => $filer,
                                        client => $client,
                                        options => '-m',
                                        interval => 1,
                                    );

The following named arguments are accepted,

=over 4

=item node

(Required) : Filer name

=item client

(Required) : Client name. Must be different from local host

=item options

(Optional) : String of valid sysstat options. Defaults to none

Valid options and meanings are listed in 'sysstat' man pages
Currently: [ -s ] [ -u  | -x  | -m  | -f  | -i  | -b ]

=item interval

(Optional) : Number in seconds. Frequency of sysstat output. Default to 1.

=item count

(Optional) : Number of rows sysstat should print out before returning. Defaults to undef

No value means 'sysstat' will run continuously until user issues a call to 'stop' method

=item processor

(Optional) : Subroutine to be called on each row of sysstat output

This subroutine would typically look like this:

    my $processor = sub {
                        my ($row,$context) = @_;
                        # Examine the row of sysstat data
                        my $nfs_ops = $row->get_by_name('NFS');
                        ...
                        # Store some state data
                        $context->{total_nfs_ops} += $nfs_ops;
                        ...
                        # Decide to send an event back to parent to be processed
                        # by the callback subroutine of 'wait_condition'
                        $Tharn::Parent->message_put(message=>$message,
                                                    type=>"Perf::Sysstat",
                                                    );
                    };

=item maxbufsize

(Optional) : Amount of sysstat output kept between calls to get_output or get_rows. Defaults 50m.

With 'interval' set to 1 second and 'options' set to '-x', sysstat produces ~10K every minute

=item debug

(Optional) : Turns on additional logging. Defaults to 0.

=item runid

(Optional) : ID to pass to subtest

=back

=for html <hr>

=cut

    sub _init :Init {
        my $self = shift;
        my $args = shift();

        $pid[$$self] = $$;
        if ($options[$$self] =~ /-c/){
            confess "Count must be specified in 'count' parameter and not 'options'";
        }
        my $code = '{
            use Perf::Sysstat::Remote;
            my $remote = Perf::Sysstat::Remote->new(node      =>\''. $node[$$self] .'\',
                                                    interval  =>\''. $interval[$$self] .'\',
                                                    maxbufsize=>\''. $maxbufsize[$$self] .'\',
                                                    debug     =>\''. $debug[$$self] .'\',
        ';
        if ($options[$$self]){
            $code .='
                                                    options   =>\''. $options[$$self] .'\',';
        }
        if ($count[$$self]){
            $code .='                               count     =>\''. $count[$$self] .'\',';
        }
        if ($processor[$$self]){
            my $serialized_processor = encode_base64(nfreeze($processor[$$self]));
            $code .='                               processor  =>\''. $serialized_processor .'\',';
        }
        $code .= '                                  );
            $remote->run;
            return;
            }';
        my $runid = (defined($args->{runid})
                     ? $args->{runid}
                     : 'sysstat_' . time . '_' . int(rand(2**32)));

        $subtest[$$self] = subtest($code, '-client', $client[$$self],
                                          '-bg',
                                          '-runid', $runid);
        # Wait for the "I'm Up!" message
        $Log->comment("Waiting for Up message");
        my $up = $self->message_get( timeout=>300 );
        $Log->comment("Got it!");
    }

=head2 start

    Start 'sysstat' command on the filer using parameters passed during 'new'

=cut

    sub start {
        my ($self) = @_;
        confess "Already started!" if ($started[$$self]);
        confess "Restarts are not yet supported" if (!$subtest[$$self]);
        my $start_time = time;
        $self->message_put(message=>'start',type=>"Perf::Sysstat");
        my $started = $self->message_get(   type=>"Perf::Sysstat::Remote",
                                            timeout=>30,
                                            on_timeout=> sub{$Log->comment( "[" . __PACKAGE__ ."] "
                                                                      ."Timeout waiting for start confirmation"
                                                                     );
                                                            }
                                        );
        confess "Unexpected message received: $started" if ($started ne 'started');
        my $total_time = time - $start_time;
        $Log->comment("Starting " . __PACKAGE__ . " took $total_time") if ($total_time > 3);
        $started[$$self]=1;
    }

=head2 get_output

    Returns raw 'sysstat' output since 'start' or last 'get_output' or 'get_rows' call

=cut

    sub get_output {
        my ($self, %args) = @_;
        my $timeout = $args{timeout};
        $timeout ||= 30;
        confess "Method unsupported when a 'processor' was specified" if ($processor[$$self]);
        $self->message_put(message=>'get_output',type=>"Perf::Sysstat");
        my $out = $self->message_get(timeout=>$timeout);
        return $out;
    }

=head2 get_rows

    Returns an array of Perf::Sysstat::Row objects, each representing one row of 'sysstat' output
    since 'start' or last 'get_output' or 'get_rows' call. See Perf::Sysstat::Row for more details

        $sysstat = Perf::Sysstat->new(node=>$filer, client=>$client);
        $sysstat->start;
        # Do something
        ...
        $rows = $sysstat->get_rows;
        # Check to see if we were ever in back to back CP
        my @BtB = grep { $_->get_by_name('CP ty') =~ /B/i } @$rows;
        print "Filer experienced " . @BtB . " back-to-back CPs\n";

=cut

    sub get_rows {
        my ($self, %args) = @_;
        my $timeout = $args{timeout};
        $timeout ||= 30;
        confess "Method unsupported when a 'processor' was specified" if ($processor[$$self]);
        $self->message_put(message=>'get_rows',type=>"Perf::Sysstat");
        my $out = $self->message_get(timeout=>$timeout);
        my $headers = $out->{headers};
        my $values = $out->{rows};
        if (!$values || !@$values){
            return [];
        }
        my @rows = map { Perf::Sysstat::Row->new( names=>$headers, columns=>$_  )  } @$values;
        return \@rows;
    }

=head2 stop

    Stops 'sysstat' command on the filer

=cut

    sub stop {
        my ($self) = @_;
        $self->message_put(message=>'stop',type=>"Perf::Sysstat");
        $started[$$self] = 0;
        return;
    }

=head2 wait_condition

    Blocks until a specific condition occurs. Condition is specified through a subroutine
    that listens to events sent by the 'processor' subroutine passed to 'new'.
    See Perf::Sysstat::CP::new and Perf::Sysstat::CP::wait_cp_begin for an example.

=over 4

=item callback

(Required) : subroutine

The callback subroutine will be passed each event and a reference to a structure $listener
it can use to store state data. It would typically look like this:

    my $callback = sub {
        my ($event, $listener) = @_;
        my $message = $event->{message};
        # Inspect $message to see if condition is met
        ...
        # Condition is met!
        $listener->{return} = $ret;
    };

The call to 'wait_condition' is going to block until the callback subroutine sets a value in
$listener->{return}.

=item filter

(Optional) : subroutine

Should the 'processor' subroutine send events of various types, users can specify which ones
to keep if necessary. The 'filter' subroutine will be passed every event received and must
return 1 to mark acceptance, 0 otherwise. Again see Perf::Sysstat::CP::wait_cp_begin for
a simple example

=back

=cut

    sub wait_condition {
        my ($self, %args) = @_;
        my $callback = $args{callback};
        confess "You must pass a 'callback' argument" if (!$callback);
        confess "Argument 'callback' must be a subroutine" if (ref($callback) ne 'CODE');
        my $filter = $args{filter};
        my $listener_filter;
        if (!$filter){
            $listener_filter = sub {
                my $event = shift;
                my $accepted = $self->staf_filter($event);
                if ($accepted) {
                    $Log->comment("Accepting event:\n" . Dumper($event)) if ($debug[$$self]);
                }
                return $accepted;
            };
        }
        elsif(ref($filter) ne 'CODE'){
            confess "Argument 'filter' must be a subroutine";
        }
        else{
            $listener_filter = sub {
                my $event = shift;
                my $accepted = $self->staf_filter($event) && $filter->($event);
                if ($accepted) {
                    $Log->comment("Accepting event:\n" . Dumper($event)) if ($debug[$$self]);
                }
                return $accepted;
            };
        }
        my $listener = {
            once     => 0,
            filter   => $listener_filter,
            callback => $callback,
        };
        listener_add($listener);
        while(! exists $listener->{return}) {
            eventloop_drain();
        }
        listener_remove($listener);
        return $listener->{return};
    }

    sub message_get :Restricted {
        my $self = shift;
        confess "Subtest is not alive" if (!$subtest[$$self]->alive);
        return $subtest[$$self]->message_get(@_);
    }

    sub message_put :Restricted {
        my $self = shift;
        confess "Subtest is not alive" if (!$subtest[$$self]->alive);
        return $subtest[$$self]->message_put(@_);
    }

    sub _destroy :Destroy {
        my $self = shift;
        return if ($$ != $pid[$$self]);
        $Log->comment("Destroying " . __PACKAGE__ );
        if (!$subtest[$$self]->alive){
            $Log->comment("Subtest is not alive, returning");
            return;
        }
        $Log->comment("Sending quit message");
        $subtest[$$self]->message_put(message=>'quit',type=>"Perf::Sysstat");
        # No need to wait for the subtest to finish
        #$st->wait_finish();
        #delete $subtest[$$self];
        return;
    }
}
1;
