# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Perf/Sysstat/CP.pm#2 $
#
## @author   dl-perf-qa@netapp.com
## @summary  Watch sysstat output for CPs
## @status   review
## @pod here

=head1 NAME

    Perf::Sysstat::CP - Subclass of Perf::Sysstat dedicated to watching CPs in real-time

=head1 SYNOPSIS

    use Perf::Sysstat::CP;

    my $sysstat = Perf::Sysstat::CP->new(   node => $FILER, 
                                            client => $CLIENT, 
                                        );

    $sysstat->start;                                
    # Do something
    ...
    # Wait for a CP to start
    $sysstat->wait_cp_begin;
    

=head1 DESCRIPTION

    Perf::Sysstat::CP sends STAF messages at the beginning and end of each CP seen in sysstat output
    These messages can be monitored via Perf::Sysstat::wait_condition. See the method 'wait_cp_begin'
    for a simple example.
    
    The structure sent at the beginning of a CP looks like:
        {
         event    => 'begin',
         type     => 'T', # For a CP from Timer
        }
    The structure sent at the end of a CP looks like:
        {
         event    => 'end',
         type     => 'T', # For a CP from Timer
         duration => 3, # CP lasted about 3s
        }
    
=head1 METHODS

=cut

package Perf::Sysstat::CP;{

    use strict;
    use warnings;
    use Tharn qw(subtest $Log $Parent $Staf);
    use NATE::Events qw(listener_add listener_remove eventloop_drain);
    use Data::Dumper;
    use Carp qw(confess);
    use Perf::Sysstat;
    use List::Util qw();
    use List::MoreUtils qw();
    use Time::HiRes qw(time);
    use Object::InsideOut qw(Perf::Sysstat);
    
    sub _preinit :PreInit {
        my ($self,$args) = @_;
        my $processor = 
            sub {
                my ($row,$context) = @_;
                my $cp_type = $row->get_by_name('CP ty');
                if (!$cp_type){
                    Carp::confess "No CP type found in row!:\n" . Data::Dumper::Dumper($row);
                }
                # Valid CP types (from 'man sysstat'):
                # B,b,F,H,L,S,T,U,Z,V,M,D,N
                # Continuations:
                # :,#
                # None:
                # -
                my @valid_types = qw( B b F H L S T U Z M D N \d);
                my @continuations = map { quotemeta($_) } ( ':','#');
                my $none = '-';
                if ( my $valid_type = List::Util::first {$cp_type =~ /$_/} @valid_types ){
                    my $current_cp = delete $context->{current_cp};
                    if ($current_cp){
                        my %message = %$current_cp;
                        $message{event} = 'end';
                        $Tharn::Parent->message_put(message=>\%message, 
                                                    type=>"Perf::Sysstat::CP/end",
                                                    );
                    }                    
                    $context->{current_cp}->{type} = $valid_type;
                    $context->{current_cp}->{duration} = 1; 
                    $Tharn::Parent->message_put(message=>{
                                                    type=>$valid_type,
                                                    event=>'begin',
                                                    },
                                                type=>"Perf::Sysstat::CP/begin",
                                                );
                }
                elsif( $cp_type eq $none ){
                    my $current_cp = delete $context->{current_cp};
                    if ($current_cp){
                        my %message = %$current_cp;
                        $message{event} = 'end';                    
                        $Tharn::Parent->message_put(message=>\%message, 
                                                    type=>"Perf::Sysstat::CP/end",
                                                    );
                    }
                }
                elsif( List::MoreUtils::any { $cp_type =~ /$_/ } @continuations ){
                    $context->{current_cp}->{duration} ++;
                }
                else{
                    Carp::confess "Unexpected cp type: '$cp_type'";
                }
            };
        $args->{processor} = $processor;
        $args->{options} = '-x';
        $args->{interval} = 1;
        return;    
    }
    
    sub staf_filter {
        my ($self,$event) = @_;
        my $child_handle = $self->get_subtest->handle();
        if ($event->{handle} != $child_handle){
            if ($self->is_debug){
                $Log->comment("Rejecting event because event 'handle' doesn't match " .
                              "child_handle '$child_handle':\n" . Dumper($event));
            }                
            return 0;              
        }
        my $child_staf_endpoint = $self->get_subtest->endpoint();
        if (!$Staf->is_same_endpoint($event->{machine},$child_staf_endpoint)){
            if ($self->is_debug){        
                $Log->comment("Rejecting event because event 'machine' doesn't match " .
                              "child_staf_endpoint '$child_staf_endpoint':\n" . Dumper($event));
            }
            return 0;              
        }                
        if ($event->{type} !~ /Perf::Sysstat::CP/){
            if ($self->is_debug){
                $Log->comment("Rejecting event because type doesn't match 'Perf::Sysstat::CP':\n"
                              . Dumper($event));
            }
            return 0;
        }
        return 1;
    }

=head2 wait_cp_begin

    Blocks until the start of a CP of a certain type or of any type if none is specified
    
=over 4

=item type

(Optional) : CP type to wait for. Default to none.

Valid CP types (from 'man sysstat'): B,b,F,H,L,S,T,U,Z,V,M,D

=back

=cut  
    
    sub wait_cp_begin {
        my ($self, %args) = @_;
        my $cp_type = $args{type};
        my $callback = sub {
            my ($event, $listener)=@_;
            my $got_cp_type = $event->{message}->{type};
            if (($cp_type && $got_cp_type =~ /$cp_type/) || !$cp_type){
                $listener->{return} = $got_cp_type;
            }
        };
        my $filter = sub {
            my $event = shift;
            return 1 if ($event->{type} =~ /begin/);
            return 0;
        };
        return $self->wait_condition(filter=>$filter, callback=>$callback);
    }
    
}
1;