# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Perf/Sysstat/Remote.pm#1 $
#
## @author   dl-perf-qa@netapp.com
## @summary  Collect sysstat output in the background
## @status   review
## @pod here

package Perf::Sysstat::Remote;{            
            
    use strict;
    use warnings;
    use Tharn qw(host connect $Parent $Staf $Log disconnect sleep);
    use NATE::Events qw(listener_add eventloop_drain listener_remove);
    use Carp qw(confess);
    use Data::Dumper;
    use Storable qw( thaw );
    $Storable::Eval = 1;
    use MIME::Base64;
    use Perf::Utils qw(in_bytes);
    use List::Util qw();
    use List::MoreUtils qw();
    use Time::HiRes qw( time );
    use Perf::Sysstat::Row;
    use Object::InsideOut;
    
    my @node            :Field :Arg(Name=>'node',       Mandatory=>1);
    my @interval        :Field :Arg(Name=>'interval',   Mandatory=>1);
    my @maxbufsize      :Field :Arg(Name=>'maxbufsize', Mandatory=>1);
    my @debug           :Field :Arg(Name=>'debug',      Mandatory=>1);
    my @options         :Field :Arg(Name=>'options',    Mandatory=>0);
    my @count           :Field :Arg(Name=>'count',      Mandatory=>0);
    my @processor       :Field :Arg(Name=>'processor',  Mandatory=>0);
    my @running         :Field :Name('running') :Default(0);
    my @fc              :Field :Name('fc');
    my @header_lines    :Field :Name('header_lines');
    my @header_fields   :Field :Name('header_fields');
    my @node_scope      :Field :Name('node_scope') :Get(is_node_scope) :Default(0);
    my @echo_stripped   :Field :Name('echo_stripped');
    
    sub _init :Init {
        my $self = shift;
        $maxbufsize[$$self] = in_bytes($maxbufsize[$$self]);
        my $host = host($node[$$self]);        
        if ($host->hosttype =~ /-ng-/){
            # Rather than trying to connect ourselves, use Server and grab its cli_connection
            # since it corresponds to being connected to ngsh
            require Server;
            my $server = Server->new(nodes=>$node[$$self], nodes_on_demand=>1, no_cli=>1);
            #$server->set_execution_server( -Server=>$node[$$self] );
            $server->establish_connections( cli_connection=>1 );
            my $conn = $server->get_connection( type=>'cli_connection' );
            $conn->maxbufsize($maxbufsize[$$self]);
            $fc[$$self] = $conn;
            $node_scope[$$self] = 1;
        }        
        else{
            $fc[$$self] = connect("#rsh-socket;maxbufsize=$maxbufsize[$$self]");             
        }
        $processor[$$self] = thaw(decode_base64($processor[$$self])) if ($processor[$$self]);        
        $self->load_header;
    }
    
    sub say {
        my ($self, $cmd, @args) = @_;
        if ($node_scope[$$self]){
            # Add the node-scope preambule
            $cmd = "run -node $node[$$self] -command \"$cmd\"";
            my (undef, undef, $out, @more) = Tharn::say($fc[$$self], $cmd, @args);
            if ($out){
                # Remove the command being echoed back to us
                my $quoted_cmd = quotemeta($cmd);
                $out =~ s/.*$quoted_cmd\s*(\r?\n)+//;
            }
            return (undef,undef,$out,@more);
        }
        else{
            return Tharn::say($fc[$$self], $cmd, @args);
        }
    }
    
    sub staf_filter {
        my ($self,$event) = @_;
        my $parent_handle = $Parent->handle();
        if ($event->{handle} != $parent_handle){
            if ($debug[$$self]){
                $Log->comment("Rejecting event because event 'handle' doesn't match " .
                            "parent_handle '$parent_handle':\n" . Dumper($event));
            }                
            return 0;              
        }
        my $parent_staf_endpoint = $Parent->endpoint();
        if (!$Staf->is_same_endpoint($event->{machine},$parent_staf_endpoint)){
            if ($debug[$$self]){
                $Log->comment("Rejecting event because event 'machine' doesn't match " .
                            "parent_staf_endpoint '$parent_staf_endpoint':\n" . Dumper($event));
            }                
            return 0;              
        }                
        if ($event->{type} ne "Perf::Sysstat"){
            if ($debug[$$self]){
                $Log->comment("Rejecting event because type doesn't match 'Perf::Sysstat':\n"
                            . Dumper($event));
            }                
            return 0;
        }
        if ($debug[$$self]){
            $Log->comment("Accepting event:\n" . Dumper($event));                
        }    
        return 1;
    }

    sub run {
        my $self = shift;
        my $quit = 0;
        my $context = {};
        my $listener = {
            once     => 0,
            filter   => sub { my $event = shift; return $self->staf_filter($event) },
            callback => sub {
                my ($event, $listener)=@_;
                my $message = $event->{message};
                $Log->comment("Received message '$message'");
                if ($quit){
                    $Log->comment("Dropping message since we quit");
                    return;
                }
                if ($message eq "start") {
                    # Protect ourselves against seeing the start message twice
                    return if ($running[$$self]);
                    my $sysstat_cmd = 'sysstat ';
                    if ($count[$$self]){
                        $sysstat_cmd .= "-c $count[$$self] ";
                    }
                    $sysstat_cmd .= "$options[$$self] $interval[$$self] ";
                    $self->say( "$sysstat_cmd",
                                '-blind',
                                #'-maxidle',0,
                                );
                    $running[$$self] = quotemeta($sysstat_cmd);
                    $Parent->message_put(message => 'started', type => "Perf::Sysstat::Remote");
                    
                    if ($processor[$$self]){
                        # Sleeping also monitors for incoming events                        
                        my $empties = 0;
                        my $last_output_time = time;
                        sleep $interval[$$self];
                        while (!$quit){
                            my $rows = $self->get_rows;
                            if (!$rows){
                                $empties ++;
                                if ($empties > 2){
                                    my $duration_without_output = time - $last_output_time;
                                    $Log->comment("Didn't get sysstat output in "
                                                  . "$duration_without_output seconds. Exiting processing");
                                    last;              
                                }
                                next;
                            }
                            $empties = 0;
                            $last_output_time = time;
                            for (@$rows){ $processor[$$self]->($_,$context) }; 
                            sleep $interval[$$self];                            
                        }    
                    }
                }
                elsif ($message eq "stop") {
                    return if (!$running[$$self]);
                    $self->stop;
                }
                elsif ($message eq "get_output") {
                    confess "You must call 'start' before you can get any output" 
                        if (!$running[$$self]);
                    my $out = $self->get_output;
                    $Parent->message_put(message => $out, type => "Perf::Sysstat::Remote");
                }
                elsif ($message eq "get_rows") {
                    confess "You must call 'start' before you can get any output" 
                        if (!$running[$$self]);
                    my $r = $self->get_parsed_output;
                    $Parent->message_put(message => $r, type => "Perf::Sysstat::Remote");
                }                
                elsif ($message eq "quit") {
                    $quit = 1;
                    $self->stop;
                }
                else {
                    confess "Unsupported message '$event->{message}'";
                }
            },
        };
        listener_add($listener);
        $Log->comment("Sending UP! message");
        $Parent->message_put(message=>"I am up!", type=>"Perf::Sysstat::Remote");
        eval{
            while(!$quit) {eventloop_drain();}
        };
        my $err = $@;
        $Log->comment("Encountered error: $err");
        listener_remove($listener);
        disconnect($fc[$$self]);
        delete $fc[$$self];
        die $err if ($err);
        $Log->comment("Returning from 'run'");
    }
    
    sub get_output {
        my ($self) = @_;
        my $out="";
        # poll for command output
        my $timed_out=0;
        my $eof = 0;
        while (!$timed_out && !$eof) {
            my $loop_out = $fc[$$self]->expect(
                                            match     =>[qr/.+/s],
                                            timeout   =>0.2,
                                            on_timeout=>sub{$timed_out=1;$Log->comment("Timeout!")},
                                            on_eof    =>sub{$eof = 1; $Log->comment("EOF!")},
                                        );
            if ($loop_out){
                # In node-scope the command is echoed back to the output. Strip it.
                if ($node_scope[$$self] && !$echo_stripped[$$self]){
                    # @running contains the sysstat command being echoed back
                    $echo_stripped[$$self] = ($loop_out =~ s/.*$running[$$self].*//);
                    $loop_out =~ s/^(\r?\n)+//g;                    
                    $Log->comment("Stripped command echo. Result:\n$loop_out") if ($echo_stripped[$$self]);
                }
                $out .= $loop_out;
            }
            elsif( !$timed_out && !$eof ){
                $Log->comment("No output from 'expect', yet we didn't reach eof or timed out!");
            }
            # Remove as many lines from the top as necessary
            if (length($out) > $maxbufsize[$$self]){
                $Log->comment("Truncating output so as to not exceed $maxbufsize[$$self]");
                while (length($out) > $maxbufsize[$$self]){
                    $out =~ s/[^\r^\n]*\r?\n//;
                }
            }
        }
        return $out;
    }
    
    sub get_parsed_output {
        my ($self) = @_;
        my $out = { headers => $header_fields[$$self], rows => [] };
        my $raw_out = $self->get_output;
        return $out if (!$raw_out);        
        my $rows = $out->{rows};        
        my @lines = split(/\r?\n/,$raw_out);
        for (@lines){
            next if (!$self->is_row($_));
            my @values = split(/\s+/,$_);
            # First element may be blank ( that's how split does it ) if the first character on the
            # line is a space (as is most times). The only time it's not a space is when CPU 
            # utilization is at over 100%.
            shift @values if (!$values[0]);
            push @$rows, \@values;
        }
        return $out;
    }
    
    sub get_rows {
        my ($self) = @_;
        my $po = $self->get_parsed_output;
        return if (!$po);
        my $values = $po->{rows};
        if (!$values || !@$values){
            return [];
        }
        my @rows = map { Perf::Sysstat::Row->new(names=>$header_fields[$$self], columns=>$_)}
                       @$values;
        return \@rows;
    }
    
    sub stop {
        my ($self) = @_;
        if ($running[$$self]){
            $fc[$$self]->send_intr();
            $running[$$self] = 0;
            $echo_stripped[$$self] = 0;
        }        
    }
    
    sub load_header {
        my ($self) = @_;
        my (undef, undef, $out) = $self->say("sysstat -c 1 $options[$$self] 1");
        if ($out !~ /CPU/){
            confess "Unable to get sysstat output:\n$out";
        }
        my @lines = split(/\r?\n/, $out);
        # The header is always two lines long
        # The second line is aligned with the end of each word on 
        # the first line. The header 'kB/s' should be replaced
        # with the preceding header's first line, i.e., 'kB/s out' after
        # 'Net in' should become 'Net out'
        # Example:
        # CPU     NFS    CIFS    HTTP     Net   kB/s    Disk   kB/s    Tape   kB/s  Cache
        #                                  in    out    read  write    read  write    age        
        $header_lines[$$self] = [ $lines[0],$lines[1] ];
        my @top_headers;
        my @headers;
        my @first_line_columns = split(/(\s+\S+)/,$lines[0]);
        foreach (@first_line_columns){
            next if (!$_);
            my $column_length = length;
            # Remove the leading and trailing white space
            s/\s+//g;
            push @top_headers, $_;            
            # 'kB/s' header is replaced with preceding top header
            $_ = $top_headers[-2] if ( /kB/ );
            # Figure out if the header spans two rows
            my $column_second_row = substr($lines[1],0,$column_length,'');
            $column_second_row =~ s/\s+//;
            $_ .= " $column_second_row" if ($column_second_row);
            push @headers, $_;
        }        
        $Log->comment("Detected header fields:\n" . join("|", @headers));
        $header_fields[$$self] = \@headers;
    }
    
    sub is_row {
        my $self = shift;
        # Make sure line isn't part of the header
        for (@{$header_lines[$$self]}){
            return 0 if ($_[0] eq $_ );
        }
        return 1;
    }
    
}
1;