# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Perf/Sysstat/Row.pm#1 $
#
## @author   dl-perf-qa@netapp.com
## @summary  Represents one row of sysstat output
## @status   review
## @pod here

=head1 NAME

    Perf::Sysstat::Row - Represents one row of sysstat output
                         
=head1 SYNOPSIS

    use Perf::Sysstat;

    my $sysstat = Perf::Sysstat->new( ... );
    $sysstat->start;
    # Do something
    ...
    my $rows = $sysstat->get_rows;
    # Print observed write throughput 
    print "Throughput:\n";
    for (@$rows){
        print $_->get_by_name('Net in') . "\n";
    }

=head1 DESCRIPTION

    Perf::Sysstat::Row is a wrapper around one data line of sysstat output. Columns can be
    queried by position (starting at zero) or by name (corresponding to the sysstat header)

=head1 METHODS

=cut

package Perf::Sysstat::Row;{  

    use Tharn qw( $Log );
    use Carp qw( confess );
    use Data::Dumper;
    use Object::InsideOut;
    
    my @names     :Field  :Arg(Name=>'names',   Mandatory=>1);
    my @columns   :Field  :Arg(Name=>'columns', Mandatory=>1);
    my @positions :Field  :Name('positions');
    
    sub _init :Init {
        my ($self) = @_;
        my $headers_count = scalar @{$names[$$self]};
        my $columns_count = scalar @{$columns[$$self]};
        if ( $headers_count != $columns_count ){
            $Log->comment("Names:\n" . Dumper($names[$$self]));
            $Log->comment("Columns:\n" . Dumper($columns[$$self]));
            confess "Number of headers ($headers_count) and values ($columns_count) don't match";
        }   
    }
    
    sub load_headers_positions {
        my ($self) = @_;
        my $i = 0;
        $positions[$$self] = {};
        for (@{$names[$$self]}){
            $positions[$$self]->{ $_ } = $i;
            $i ++;
        }
    }

=head2 get_by_index

    Returns value in column at index passed in
    
    Example: if you know the data you're interested in is in the column 0
    
    my $data = $row->get_by_index(0);
    
    
=cut  
    
    sub get_by_index {
        my ($self, $index) = @_;
        confess "You must pass an argument" if (!defined($index));
        return $columns[$$self]->[$index];
    }

=head2 get_by_name

    Returns value in column corresponding to header passed in.
    The column names are detected automatically from the sysstat output. Note that some columns
    get renamed to ensure meaning and uniqueness. 
    
    Example: The following output
    
     CPU     NFS    CIFS    HTTP     Net   kB/s    Disk   kB/s 
                                      in    out    read  write
    
    would result in the following column names:
    
    CPU, NFS, CIFS, HTTP, Net in, Net out, Disk read, Disk write
    
=cut  
    
    sub get_by_name {
        my ($self, $name) = @_;
        confess "You must pass an argument" if (!defined($name));
        $self->load_headers_positions if (!$positions[$$self]);
        my $column_index = $positions[$$self]->{$name};
        if (!defined $column_index){
            $Log->comment("Self:\n" . Dumper($self->dump(1)));
            confess "No column with name '$name' found!"; 
        }
        return $columns[$$self]->[$column_index];
    }
    
=head2 get_names

    Returns all the header names for the row
    
=cut     
    
    sub get_names {
        my ($self) = @_;
        return $names[$$self];
    }

=head2 get_count

    Returns the number of columns in the row
    
=cut     
    
    sub get_count {
        my ($self) = @_;
        return scalar @{$columns[$$self]};
    }
}
1;