/**
 * @file cmapl_common_defs.h
 *
 * CMAPL common definitions file. It contains various data structures
 * used by the CMAPL API.
 */

#ifndef _CMAPL_COMMON_DEFS_H_
#define _CMAPL_COMMON_DEFS_H_

#include <sys/types.h>

#ifdef _KERNEL
#include <machine/_inttypes.h>
#endif /* _KERNEL */

#ifdef _CMAPL_USER_SPACE_
#include <inttypes.h>
#include <stdbool.h>
#endif /* _CMAPL_USER_SPACE_ */

/**
 * This enum is used to specify the value type of a counter. Only 3
 * possibilities are available, 4 byte unsigned integer, 8 byte unsigned
 * integer, or a string.
 */
typedef enum {
	CMAPL_CNT_4BYTES,
	CMAPL_CNT_8BYTES,
	CMAPL_CNT_STRING
} cmapl_cnt_data_type_t;

/**
 * The cmapl_error_t enum lists all the different error values in the CMAPL API.
 * The use of CMAPL_ERROR_FORMAT allows us to generate strings for the errors
 * as the cmapl_error_labels array of strings, so that they can be printed.
 */
 #define CMAPL_ERROR_CODES()							\
	CMAPL_ERROR_FORMAT(CMAPL_COUNTER_NOT_FOUND),		\
	CMAPL_ERROR_FORMAT(CMAPL_COUNTER_DATA_NOT_FOUND),	\
	CMAPL_ERROR_FORMAT(CMAPL_COUNTER_MISMATCH),			\
	CMAPL_ERROR_FORMAT(CMAPL_COUNTER_TYPE_NOT_FOUND),	\
	CMAPL_ERROR_FORMAT(CMAPL_CLOSE_DIRECTORY_FAILED),	\
	CMAPL_ERROR_FORMAT(CMAPL_CLOSE_FILE_FAILED),		\
	CMAPL_ERROR_FORMAT(CMAPL_FILE_NOT_FOUND),			\
	CMAPL_ERROR_FORMAT(CMAPL_FILENAME_FORMAT_ERROR),	\
	CMAPL_ERROR_FORMAT(CMAPL_INSTANCE_NOT_FOUND),		\
	CMAPL_ERROR_FORMAT(CMAPL_INVALID_ARGUMENTS),		\
	CMAPL_ERROR_FORMAT(CMAPL_INVALID_COUNTER_DATA),		\
	CMAPL_ERROR_FORMAT(CMAPL_INVALID_QUERY),			\
	CMAPL_ERROR_FORMAT(CMAPL_INVALID_RECORD),			\
	CMAPL_ERROR_FORMAT(CMAPL_INVALID_SIZE),				\
	CMAPL_ERROR_FORMAT(CMAPL_INVALID_TIMESTAMP),		\
	CMAPL_ERROR_FORMAT(CMAPL_INVALID_VERSION),			\
	CMAPL_ERROR_FORMAT(CMAPL_NULL_POINTER),				\
	CMAPL_ERROR_FORMAT(CMAPL_OBJECT_NOT_FOUND),			\
	CMAPL_ERROR_FORMAT(CMAPL_PARSING_ERROR),			\
	CMAPL_ERROR_FORMAT(CMAPL_PERIOD_NOT_FOUND),			\
	CMAPL_ERROR_FORMAT(CMAPL_READ_FAILED),				\
	CMAPL_ERROR_FORMAT(CMAPL_RECORD_EMPTY),				\
	CMAPL_ERROR_FORMAT(CMAPL_RECORD_NOT_FOUND),			\
	CMAPL_ERROR_FORMAT(CMAPL_SEEK_FAILED),				\
	CMAPL_ERROR_FORMAT(CMAPL_TIMESTAMP_OUT_OF_RANGE),	\
	CMAPL_ERROR_FORMAT(CMAPL_UNLINK_ERROR),				\
	CMAPL_ERROR_FORMAT(CMAPL_VERSION_MISMATCH),			\
	CMAPL_ERROR_FORMAT(CMAPL_WRITE_FAILED)

/* Define CMAPL_ERROR_FORMAT to generate an enum. */
#define CMAPL_ERROR_FORMAT(i) i

/**
 * This enum is used to specify all the possible error types in
 * the CMAPL. All the errors are in the header file as definitions.
 * Please look at the definition of CMAPL_ERROR_CODES for more
 * details.
 */
typedef enum {
	CMAPL_SUCCESS = 0,
	CMAPL_ERROR_CODES(),	/* Inserts codes defined above */
	CMAPL_ERROR_CODES_MAX_INDEX
} cmapl_error_t;
#undef CMAPL_ERROR_FORMAT

/**
 * Define CMAPL_ERROR_FORMAT to generate an array of strings.
 * It is now possible to write:
 *    cmapl_error_t result;
 *    result = cmapl_function(...);
 *    if (CMAPL_SUCCESS != result) {
 *      printf("CMAPL error: cmapl_function returned error %s (%d)\n", cmapl_error_labels[result], (int) result);
 *    }
 */
extern const char * cmapl_error_labels[CMAPL_ERROR_CODES_MAX_INDEX];

/**
 * This enum is used by the cmapl_filter_in structure to specify what value
 * characteristic of a counter to filter by, whether a counter value has to
 * be equal to, less than, or greater than a certain value.
 */
typedef enum {
	CMAPL_EQUAL_TO,
	CMAPL_LESS_THAN,
	CMAPL_GREATER_THAN
} cmapl_op_t;

/*************************************************
 * These are unions used in the parsing library. *
 *************************************************/

/**
 * This union holds the three different values a counter could potentially be.
 * Depending on the counter type, the value will either be a 4 byte integer,
 * 8 byte integer, or a string.
 */
typedef union {
	uint32_t		val_4B;
	uint64_t		val_8B;
	const char*		val_str;
} cmapl_cnt_value_t;

/***********************************************************************
 * These data structures are for the data segment of the archive file. *
 * They will be outputted to the user.                                 *
 ***********************************************************************/

/**
 * These structures contain all the information of a counter data.
 */
typedef struct {
	cmapl_cnt_data_type_t   value_type;             /**< the type of the counter value */
	cmapl_cnt_value_t*      values;                 /**< all the values of the counter */
	uint32_t		num_values;		/**< number of values for this counter */
} cmapl_counter_data_t_v5_0;

typedef cmapl_counter_data_t_v5_0 cmapl_counter_data_t_v6_0;

typedef cmapl_counter_data_t_v6_0 cmapl_counter_data_t;

/**
 * These structures contain the metadata of a single instance.
 */
typedef struct {
	const char*		name;			/**< instance name */
} cmapl_instance_data_t_v5_0;

typedef struct {
	const char**			key_cnt_names;          /**< key counter names */
	cmapl_counter_data_t_v6_0*	key_cnt_values;		/**< key counter values */
	uint32_t			num_key_cnts;           /**< number of key counters */
} cmapl_instance_data_t_v6_0;

typedef cmapl_instance_data_t_v6_0 cmapl_instance_data_t;

/**
 * These structures contain the record header information of a single record.
 */

typedef struct {
	uint32_t                major_v;                /**< major version number */
        uint32_t                minor_v;                /**< minor version number */
        uint64_t                msecs_since_boot;       /**< real sampled timestamp of the record, in milliseconds */
        uint32_t                size;                   /**< size in bytes of the entire record buffer
                                                                including the record header */
        uint32_t                obj_id;                 /**< object id of the object associated with this record */
        uint32_t                num_instances;          /**< number of instances in this record */
        uint32_t                freq_flag;              /**< bitflag of the sampling frequencies of this record */
} cmapl_record_header_data_t_v5_0;

typedef struct {
	uint32_t                major_v;                /**< major version number */
        uint32_t                minor_v;                /**< minor version number */
        uint64_t                msecs_since_base;       /**< real sampled timestamp of the record measured in
								milliseconds since the base time of the file*/
        uint32_t                size;                   /**< size in bytes of the whole record buffer, NOT including
								the record header */
        uint32_t                obj_id;                 /**< object id of the object associated with this record */
        uint64_t                period;                 /**< period of all counters being sampled in this record
								in milliseconds */
} cmapl_record_header_data_t_v6_0;

typedef cmapl_record_header_data_t_v6_0 cmapl_record_header_data_t;

/**
 * These structures hold the information of an attribute definition.
 * An attribute definition describes any attributes and their legal/valid
 * values that might appear in any element definitions later on.
 *
 * In version 5.0, there are 3 different types of attribute definitions possible, each
 * with a different name: "flag", "type", "ctype".
 *
 * "flag" -	This a bitfield that describes the frequency. Each frequency
 *		has a different bitflag value, defined in the <value> element.
 *		Each counter in the following definitions will contain only one
 *		"flag" bitflag. However, each instance header may have several
 *		frequency "flag" bitflags defined.
 * "type" -	This is a simple definition to describe the type of whatever the
 *		attribute is (eg, INTEGER, STRING).
 * "ctype" -	This is an (unfortunately) overloaded bitfield. Valid bitflags
 *		can describe byte sizes, processing type, unit suffices (eg. bit/second).
 *
 * In version 6.0, only the "type" and "ctype" attribute types are available. Frequency
 * is now specified by the counter's period.
 */
typedef struct {
	uint32_t		major_v;		/**< major version number */
	uint32_t		minor_v;		/**< minor version number */
	const char*		name;			/**< Name of the attribute. Either "flag", "type", or
								"ctype" */
	uint32_t		value;			/**< One possible value of the attribute */
	const char*		desc;			/**< Any descriptions used for display */
	const char*		frequency;		/**< Describes the frequency value associated with the
								flag value. Only appears for the attributes with
								name "flag". Valid values are FREQUENCY_1,
								FREQUENCY_2, FREQUENCY_3, FREQUENCY_4, CONFIG. */
} cmapl_attribute_def_t_v5_0;

typedef struct {
	uint32_t		major_v;		/**< major version number */
	uint32_t		minor_v;		/**< minor version number */
	const char*		name;			/**< Name of the attribute. Either "type", or
								"ctype" */
	uint32_t		value;			/**< One possible value of the attribute */
	const char*		desc;			/**< Any descriptions used for display */
} cmapl_attribute_def_t_v6_0;

typedef cmapl_attribute_def_t_v6_0 cmapl_attribute_def_t;

/******************************************************
 * These are data structures used for the XML header. *
 * They are used as input and output.                 *
 ******************************************************/

/**
 * These structures contain all the properties of a single counter.
 */
typedef struct {
	uint32_t                major_v;                /**< major version number */
	uint32_t                minor_v;                /**< minor version number */
	const char*		name;			/**< counter name */
	uint32_t                ctype;                  /**< bitfield indicating the size, type,
								and unit type of the counter */
	uint32_t		flags;			/**< bitfield indicating the frequency
								the counter is sampled at */
	uint32_t		size;			/**< binary size in bytes of this counter,
								per element size if array counter */
	uint32_t		cnt_id;			/**< counter identification number */
	const char*		desc;			/**< counter description */
	uint32_t		base_id;		/**< id of base counter */
	const char**		x_names;		/**< list of first array counter labels */
	uint32_t		num_x_items;		/**< size of first array */
	const char**		y_names;		/**< list of second array counter labels */
	uint32_t		num_y_items;		/**< size of second array */
} cmapl_counter_def_t_v5_0;

typedef struct {
	uint32_t		major_v;		/**< major version number */
	uint32_t		minor_v;		/**< minor version number */
	const char*		name;			/**< counter name */
	const char*		priv;			/**< privilege level of counter */
	uint32_t		ctype;			/**< bitfield indicating the size, type,
								and unit type of the counter */
	uint32_t		size;			/**< binary size in bytes of this counter,
								per element size if array counter */
	bool			is_key;			/**< indicates if key counter true:key false:not key */
	uint64_t		period;			/**< fastest sample period of counter in milliseconds;
								can be ignored if key counter since key
								counters are sampled at every period */
	uint32_t		cnt_id;			/**< counter identification number */
	const char*		desc;			/**< counter description */
	uint32_t		base_id;		/**< id of base counter */
	const char**		x_names;		/**< list of first array counter labels */
	uint32_t		num_x_items;		/**< size of first array */
	const char**		y_names;		/**< list of second array counter labels */
	uint32_t		num_y_items;		/**< size of second array */
} cmapl_counter_def_t_v6_0;

typedef cmapl_counter_def_t_v6_0 cmapl_counter_def_t;

/**
 * This structure describes how different fields in the record header
 * and the rest of the record are divided up and how the record should
 * be parsed.
 */
typedef struct {
	uint32_t		major_v;		/**< major version number */
	uint32_t		minor_v;		/**< minor version number */
	uint32_t		size;			/**< size in bytes that the instance header uses */
	uint32_t		inst_name_size;		/**< size in bytes of each instance name */
	uint32_t		msec_since_boot_size;	/**< size in bytes of the msec_since_boot field that the instance header uses */
	uint32_t		record_size_size;	/**< size in bytes of the record_size field that the instance header uses */
	uint32_t		obj_id_size;		/**< size in bytes of the obj_id field that the instance header uses */
	uint32_t		num_instances_size;	/**< size in bytes of the num_instance field that the instance header uses */
	uint32_t		flag_size;		/**< size in bytes of the flag field that the instance header uses*/
} cmapl_record_header_def_t_v5_0;

/**
 * These structures contain all the metadata of the XML header.
 */
typedef struct {
	uint32_t		major_v;		/**< major version number */
	uint32_t		minor_v;		/**< minor version number */
	uint32_t		generation_id;		/**< indicates if the preset has changed */
	uint64_t		unix_time;		/**< time of the filer, in UNIX epoch time,
								taken at the start of this file in milliseconds */
	uint64_t		msecs_since_boot;	/**< total milliseconds since boot of the filer,
								taken at the start of this file */
} cmapl_header_def_t_v5_0;

typedef struct {
	uint32_t		major_v;		/**< major version number */
	uint32_t		minor_v;		/**< minor version number */
	uint32_t		generation_id;		/**< indicates if the preset has changed */
	const char*		preset_version;		/**< version of the preset in the XML header*/
} cmapl_header_def_t_v6_0;

typedef cmapl_header_def_t_v6_0 cmapl_header_def_t;

/**
 * These structures contain all the metadata of a single object in the XML header.
 */
typedef struct {
	uint32_t		major_v;		/**< major version number */
	uint32_t		minor_v;		/**< minor version number */
	const char*		name;			/**< object name */
	uint32_t		obj_id;			/**< object identification number */
} cmapl_object_def_t_v5_0;

typedef cmapl_object_def_t_v5_0 cmapl_object_def_t_v6_0;

typedef cmapl_object_def_t_v6_0 cmapl_object_def_t;

/**
 * These structures contain all the Ontap metadata related to an archive file.
 */
typedef struct {
	uint32_t		major_v;		/**< major version number */
	uint32_t		minor_v;		/**< minor version number */
	const char*		ontap_version;		/**< version of the Ontap kernel */
	const char*		local_time;		/**< localtime of the filer, taken at the start of this file */
	const char*		gm_time;		/**< UTC time of the filer, taken at the start of this file */
	const char*		hostname;		/**< local name of the filer */
	const char*		model;			/**< model of the filer */
	const char*		serial_number;		/**< serial number of the filer */
	const char*		system_id;		/**< system id of the filer */
} cmapl_system_def_t_v5_0;

typedef struct {
	uint32_t		major_v;		/**< major version number */
	uint32_t		minor_v;		/**< minor version number */
	uint64_t		base_time;		/**< the UNIX epoch time of the file used as the base time for
                                                                all the records written to the file, in milliseconds*/
	const char*		ontap_version;		/**< version of the Ontap kernel */
	const char*		node_uuid;		/**< local UUID of the filer */
	const char*		hostname;		/**< local name of the filer */
	const char*		model;			/**< model of the filer */
	const char*		serial_number;		/**< serial number of the filer */
	const char*		system_id;		/**< system id of the filer */
} cmapl_system_def_t_v6_0;

typedef cmapl_system_def_t_v6_0 cmapl_system_def_t;

/**
 * These structures contain all the metadata related to a single archive file.
 */
typedef struct {
	const char*		file_path;		/**< path to the generic archive file */
	uint64_t		start_time;		/**< start UTC time of the file, in milliseconds */
	uint64_t		end_time;		/**< end UTC time of the file, in milliseconds */
} cmapl_file_def_t_v5_0;

typedef struct {
        const char*             file_path;              /**< path to the generic archive file */
        const char*             meta_path;              /**< path to the .meta meta file */
        const char*             data_path;              /**< path to the .gz data file */
	uint32_t		file_index;		/**< index number of the file */
        uint64_t                start_time;             /**< start UTC time of the file, in milliseconds */
        uint64_t                end_time;               /**< end UTC time of the file, in milliseconds */
} cmapl_file_def_t_v6_0;

typedef cmapl_file_def_t_v6_0 cmapl_file_def_t;

typedef struct {
	uint32_t                major_v;                /**< major version number */
        uint32_t                minor_v;                /**< minor version number */
	char*                   directory_path;         /**< path to the directory */
        uint64_t                start_time;             /**< the start UTC time of the first file to look at, in milliseconds */
        uint64_t                end_time;               /**< the end UTC time of the last file to looks at, in milliseconds */
        uint64_t                period;                 /**< period to round the timestamps by, in milliseconds */
} cmapl_directory_def_t_v5_0;

typedef cmapl_directory_def_t_v5_0 cmapl_directory_def_t_v6_0;

typedef cmapl_directory_def_t_v6_0 cmapl_directory_def_t;

typedef struct {
        uint32_t                major_v;                /**< major version number */
        uint32_t                minor_v;                /**< minor version number */
        char*                   file_path;              /**< path to the corefile */
} cmapl_corefile_def_t_v6_0;

typedef cmapl_corefile_def_t_v6_0 cmapl_corefile_def_t;

/**********************************************************
 * These are data structures used for query input/output. *
 **********************************************************/

/**
 * These structures are used to filter a query by a specific counter value.
 * For example, if you wanted to filter a query to only return records
 * with the counter latency > 120us, cnt_name would be "latency", value_type
 * would be CMAPL_CNT_4BYTES, value would set val_4B to 120, and operator
 * would be CMAPL_GREATER_THAN.
 */
typedef struct {
	uint32_t                major_v;                /**< major version number */
        uint32_t                minor_v;                /**< minor version number */
        const char*		cnt_name;		/**< counter name to filter on */
        cmapl_cnt_data_type_t   value_type;             /**< the type of the counter value */
        cmapl_cnt_value_t       value;                  /**< counter value to filter on */
        cmapl_op_t              op;			/**< the operator used to compare the value */
} cmapl_filter_def_t_v5_0;

typedef cmapl_filter_def_t_v5_0 cmapl_filter_def_t_v6_0;

typedef cmapl_filter_def_t_v6_0 cmapl_filter_def_t;

/**
 * These structures are used to query an archiver file with object name,
 * key counter information, starting and ending timestamp, and the
 * query period. This will be then passed into the cmapl_query()
 * function where it'll be used to search for the desired records.
 */
typedef struct {
	uint32_t                major_v;                /**< major version number */
        uint32_t                minor_v;                /**< minor version number */
        const char*             object_name;		/**< name of the object */
        const char*             instance_name;          /**< name of the instance */
        uint64_t                start_time;             /**< beginning UTC timestamp to start query, in milliseconds */
        uint64_t                end_time;               /**< end UTC timestamp to finish query, in milliseconds */
} cmapl_query_def_t_v5_0;

typedef struct {
	uint32_t                        major_v;        /**< major version number */
        uint32_t                        minor_v;        /**< minor version number */
        const char*                     object_name;    /**< name of the object */
        uint32_t                        num_key_cnts;   /**< number of key counters */
        const char**                    key_cnt_names;  /**< key counter names to search for */
        cmapl_counter_data_t_v6_0*	key_cnt_values;	/**< key counter values to search for */
        uint32_t                        num_cnts;       /**< number of counters to search for */
        const char**                    cnt_names;      /**< counter names to search for */
        uint64_t                        start_time;     /**< beginning UTC timestamp to start query, in milliseconds*/
        uint64_t                        end_time;       /**< end UTC timestamp to finish query, in milliseconds */
        uint64_t                        period;         /**< query period to increment by, in milliseconds */
} cmapl_query_def_t_v6_0;

typedef cmapl_query_def_t_v6_0 cmapl_query_def_t;

/*****************************************
 * CMAPL version structure and functions *
 *****************************************/

/**
 * Different versions of the archive files
 */
#define CMAPL_MAJOR_VER_6		6
#define CMAPL_MAJOR_VER_5		5
#define CMAPL_MINOR_VER_0		0

/**
 * The most recent version of the archive file
 */
#define CMAPL_CURRENT_MAJOR_VERSION	CMAPL_MAJOR_VER_6
#define CMAPL_CURRENT_MINOR_VERSION	CMAPL_MINOR_VER_0

typedef struct {
	uint32_t		major_v;		/**< major version number */
	uint32_t		minor_v;		/**< minor version number */
} cmapl_version_t;

/**
 * Util functions in the CMAPL
 */

/**
 * @name cmapl_get_version
 *
 * @param[in]           data		Data structure that has a version
 * @param[out]		major_v		Major version number
 * @param[out]		minor_v		Minor version number
 * @return              void
 *
 * This method gets the version number of a specific data structure. The major
 * and minor version numbers are passed back through the major_v and minor_v
 * parameters, respectively.
 */
extern void
cmapl_get_version(
                        void* data,
			uint32_t * major_v,
			uint32_t * minor_v);

/**
 * @name cmapl_is_version
 *
 * @param[in]           data            Data structure to check version
 * @param[in]           major_v         Major version number to check
 * @param[in]           minor_v         Minor version number to check
 * @return		bool		Whether the version matches or not
 *
 * This method checks if the specified major and minor version
 * numbers match that of this data structure.
 */
extern bool
cmapl_is_version(
			void * data,
			uint32_t major_v,
			uint32_t minor_v);

/**
 * @name cmapl_set_latest_version
 *
 * @param[in]           data		Data structure that needs version initialized
 * @return              void
 *
 * This method sets the version of a given data structure to be the latest
 * archive file version number.
 */
extern void
cmapl_set_latest_version(
                        void* data);

// type that maps to one of the size definitions below
typedef uint32_t cmapl_cnt_size_t;
// type that maps to one of the type definitions below
typedef uint32_t cmapl_cnt_type_t;
// type that maps to one of the unit definitions below
typedef uint32_t cmapl_cnt_unit_t;
#ifndef _CM_COUNTER_TYPES
#define _CM_COUNTER_TYPES
/*
 * The following is copied from ///cm/cmlib/src/api/cm_common.h. These
 * counter type definitions in these two files should always match.
 * TODO: This should eventually be removed since we shouldn't have
 *       duplicate definitions.
 */

/* these are in absolute values */
#define CM_CNT_4B     0x00000004
#define CM_CNT_8B     0x00000008
#define CM_CNT_16B    0x00000010
#define CM_CNT_32B    0x00000020
#define CM_CNT_64B    0x00000040
#define CM_CNT_96B    0x00000060
#define CM_CNT_128B   0x00000080
#define CM_CNT_160B   0x000000A0
#define CM_CNT_252B   0x000000FC
#define CM_CNT_MAX_BYTESIZE   CM_CNT_252B

/*
 * size of one item in a complex counter, in case of a simple counter it
 * is just the size of the counter
 */
#define CM_CNT_SIZE_MASK 0x000000FF
#define CM_CNT_ITEM_SIZE(counter) (CM_CNT_SIZE_MASK & (counter).type)


/* these 8 bits will indicate the processing to be done on the counter. */
#define CM_CNT_RAW      0x00000100      /* C2 */
#define CM_CNT_RATE     0x00000200      /* (C2 - C1) / timediff */
#define CM_CNT_DELTA    0x00000400      /* C2 - C1 */

/* The CM_CNT_AVERAGE counter type basically does a ratio of
 * counter delta between two points in time divided by the base delta
 * between the same two points in time. Since it is primarliy used
 * for representing average latencies and average chain lengths this
 * counter is called as CM_CNT_AVERAGE.
 */
#define CM_CNT_AVERAGE  0x00000800      /* (C2 - C1) / base diff */
#define CM_CNT_PERCT    0x00001000      /* Scalable (C2 - C1) / base diff */
#define CM_CNT_STRING   0x00002000

#define CM_CNT_TYPE_MASK 0x0000FF00
#define CM_CNT_MATH(counter) (CM_CNT_TYPE_MASK & (counter).type)

/* only one type of math counter type may be set at any time.
 * i.e., masked value for the math type should have only one bit set
 * (power of 2). (x & (x - 1) == 0) if and only if x is a power of 2.
 */
#define CM_CNT_SINGLE_MATH_TYPE(counter)        \
        (((CM_CNT_MATH(counter) - 1) & (CM_CNT_MATH(counter))) == 0)

#define CM_IS_RELA(_counter) (((_counter).type & 0x00001F00) & \
                        (CM_CNT_PERCT|CM_CNT_AVERAGE))

/* Units aka Suffixes */
#define CM_NO_SUFFIX    0x00010000
#define CM_PER_SEC      0x00020000
#define CM_PERCT        0x00030000      /* % */
#define CM_B_PER_SEC    0x00040000
#define CM_KB_PER_SEC   0x00050000
#define CM_MSECS        0x00060000
#define CM_USECS        0x00070000
#define CM_SECS         0x00080000
#define CM_MB_PER_SEC   0x00090000
#define CM_MB           0x000A0000

/* some counters serve as base counter. these are never exported/displayed.
 * however, they are still included in the instance value list for
 * calculations of other counters. For no-display counters, suffix has no meaning.
 * so NODISP is another type of suffix.
 */
#define CM_CNT_NODISP   0x000B0000

#define CM_CNT_UNIT_MASK(_type) (0x000F0000 & (_type))

/* use this macro instead of the old one CM_IS_BASE */
#define CM_CNT_SUPRESS(_counter) ((CM_CNT_UNIT_MASK((_counter).type)) == CM_CNT_NODISP)

#endif /* _CM_COUNTER_TYPES */

// redefine counter type masks for CMAPL use
#define CMAPL_CNT_ITEM_SIZE(ctype) (CM_CNT_SIZE_MASK & ctype)
#define CMAPL_CNT_MATH(ctype) (CM_CNT_TYPE_MASK & ctype)
#define CMAPL_CNT_UNIT_MASK(ctype) (0x000F0000 & ctype)

#endif /* _CMAPL_COMMON_DEFS_H_ */
