/**
 * @file cmapl_corefile.h
 *
 * The CMAPL corefile-level function definitions. These functions provide all the
 * functionality to parse and read archiver core files.
 *
 */

#ifndef _CMAPL_COREFILE_H_
#define _CMAPL_COREFILE_H_

/**
 * @name cmapl_close_corefile
 *
 * @param[in]           corefile                Handle of the corefile to close
 * @return              cmapl_error_t           Error flag
 * @retval		CMAPL_SUCCESS		Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS	The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION	The version of the handle is not known
 *
 * This method closes the specifed corefile handle. All memory held by the handle
 * will be freed and the file handle associated with the corefile will be closed.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_close_corefile(
                        cmapl_corefile_handle_t * corefile);

/**
 * @name cmapl_first_datafile
 *
 * @param[in]           corefile                Handle of the corefile to retrieve datafile
 * @param[out]          datafile                Handle of first datafile in corefile
 * @return              cmapl_error_t           Error flag
 * @retval		CMAPL_SUCCESS		Function finished successfully
 * @retval              CMAPL_CLOSE_FILE_FAILED Closing the extracted file failed
 * @retval              CMAPL_INVALID_ARGUMENTS	The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION	The version of the handle is not known
 * @retval              CMAPL_PARSING_ERROR     Parsing the compressed 7z corefile failed
 * @retval              CMAPL_WRITE_FAILED      Extracting the datafile failed
 *
 * This method returns the handle of the first datafile within a corefile. This datafile
 * is at the root directory of the corefile. The datafile handle returned to the user will
 * be allocated internally, and will be freed if another datafile is retrieved by the user
 * or if the corefile is closed. If there is no datafile in the corefile, then NULL is
 * passed back through the file handle.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_first_datafile(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_file_handle_t ** datafile);

/**
 * @name cmapl_next_datafile
 *
 * @param[in]           corefile                Handle of the corefile to retrieve datafile
 * @param[out]          datafile                Handle of next datafile in the corefile
 * @return              cmapl_error_t           Error flag
 * @retval		CMAPL_SUCCESS		Function finished successfully
 * @retval              CMAPL_CLOSE_FILE_FAILED Closing the extracted file failed
 * @retval              CMAPL_INVALID_ARGUMENTS	The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION	The version of the handle is not known
 * @retval              CMAPL_PARSING_ERROR     Parsing the compressed 7z corefile failed
 * @retval              CMAPL_WRITE_FAILED      Extracting the datafile failed
 *
 * This method returns the handle of the next datafile within a corefile. This datafile
 * is at the root directory of the corefile. The datafile handle returned to the user
 * will be allocated internally, and will be freed if another datafile is retrieved by
 * the user or if the corefile is closed. The corefile handle keeps track of the last
 * datafile that was looked at. If there are no more datafiles remaining, then NULL is
 * returned back to the user through the datafile handle.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_next_datafile(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_file_handle_t ** datafile);

/**
 * @name cmapl_open_corefile
 *
 * @param[in]           file_name               Name of the corefile file to open
 * @param[out]          corefile                Handle of the corefile that's opened
 * @return              cmapl_error_t           Error flag
 * @retval		CMAPL_SUCCESS		Function finished successfully
 * @retval              CMAPL_FILE_NOT_FOUND    Corefile name couldn't be found
 * @retval              CMAPL_INVALID_ARGUMENTS	The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION	The version of the handle is not known
 * @retval              CMAPL_PARSING_ERROR     Parsing the compressed 7z corefile failed
 *
 * This method opens a specific corefile and returns a handle of that corefile
 * back to the user. The user will then be able to interact with and query the
 * corefile through this handle.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_open_corefile(
                        char * file_name,
                        cmapl_corefile_handle_t ** corefile);

#endif /* _CMAPL_COREFILE_H_ */
