/**
 * The CMAPL internal corefile-level function declarations. These functions
 * are version specific declarations. Different version data files will
 * use functions that will have different implementations. The external
 * functions will call one of these functions depending on the version
 * of the data file.
 *
 * The descriptions of the external CMAPL functions related to each of
 * these internal functions are described in cmapl_corefile.h.
 *
 * Syntax of functions:
 *
 *      <function_name>_v<major_version>_<minor_version>
 *
 *      function_name   - the external function name
 *      major_version   - the major version number of the data file
 *      minor_version   - the minor version number of the data file
 *
 *      For Example:    cmapl_open_corefile_v6_0 stands for the
 *                      cmapl_open_corefile function used on
 *                      a version 6.0 data file.
 *
 */

#ifndef _CMAPL_COREFILE_INT_H_
#define _CMAPL_COREFILE_INT_H_

cmapl_error_t
cmapl_close_corefile_v6_0(
                        cmapl_corefile_handle_t * corefile);


cmapl_error_t
cmapl_first_datafile_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_file_handle_t ** datafile);

cmapl_error_t
cmapl_first_directory_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_directory_handle_t ** dir);

cmapl_error_t
cmapl_next_datafile_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_file_handle_t ** datafile);

cmapl_error_t
cmapl_next_directory_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_directory_handle_t ** dir);

cmapl_error_t
cmapl_open_corefile_v6_0(
                        char * file_name,
                        cmapl_corefile_handle_t ** corefile);

///////////////////////////////////////////////////////////////////////////
// The following functions are internal-only fuctions, meaning they have no
// general external version of them.
///////////////////////////////////////////////////////////////////////////

/**
 * @name cmapl_extract_file_v6_0
 *
 * @param[in]           corefile                Handle of the corefile to extract from
 * @param[in]           file_index              Index of the datafile to extract
 * @return              cmapl_error_t           Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This function extracts a specific datafile from a corefile. The file is specified by its
 * index within the corefile. Both the metadata and data files will be extracted temporarily
 * to the working directory. The file should be deleted with cmapl_remove_current_file_v6_0()
 * after it is finished being used.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_extract_file_v6_0(
                        cmapl_corefile_handle_t* corefile,
                        uint32_t file_index);

/**
 * @name cmapl_remove_current_file_v6_0
 *
 * @param[in]           corefile                Handle of the corefile to extract from
 * @return              cmapl_error_t           Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This function deletes the temporarily extracted datafile that is currently opened for
 * use. Both the metadata and data files will be deleted and removed from the current
 * working directory.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_remove_current_file_v6_0(
                        cmapl_corefile_handle_t* corefile);

/**
 * @name cmapl_extract_directory_v6_0
 *
 * @param[in]           corefile                Handle of the corefile to extract from
 * @param[in]           dir_index               Index of the directory to extract
 * @return              cmapl_error_t           Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This function extracts a specific directory from a corefile. The directory is specified by its
 * index within the corefile. The whole directory and all the contents inside of it will be extracted
 * temporarily to the working directory. The file should be deleted with
 * cmapl_remove_current_directory_v6_0() after it is finished being used.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_extract_directory_v6_0(
                        cmapl_corefile_handle_t* corefile,
                        uint32_t dir_index);

/**
 * @name cmapl_remove_current_directory_v6_0
 *
 * @param[in]           corefile                Handle of the corefile to extract from
 * @return              cmapl_error_t           Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This function deletes the temporarily extracted directory that is currently opened for
 * use. The directory and all of its internal contents will be deleted and removed from the
 * current working directory.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_remove_current_directory_v6_0(
                        cmapl_corefile_handle_t* corefile);

/**
 * @name cmapl_add_root_directory_v6_0
 *
 * @param[in]           corefile                Handle of the corefile
 * @param[in]           dir_path                Path of the directory to add
 * @return              cmapl_error_t           Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This function takes a full path to a directory or file, and adds the root directory
 * of that path to the list of directories to open in the corefile handle. For example,
 * if the path passed into dir_path was foo/bar/a.txt, then the directory foo/ would be
 * added to the corefile handle. This list of root directories make it easy to iterate
 * through all the directories in a corefile later and get their directory handles.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_add_root_directory_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        char* dir_path);

#endif /* _CMAPL_COREFILE_INT_H_ */
