/**
 * @file cmapl_directory.h
 *
 * The CMAPL directory-level function declarations. These functions provide all the
 * functionality to look-up and access archive files within directories.
 *
 * TODO: need to handle subdirs within the directory
 */

#ifndef _CMAPL_DIRECTORY_H_
#define _CMAPL_DIRECTORY_H_

/**
 * @name cmapl_close_directory
 *
 * @param[in]           dir             Handle of directory
 * @pre                                 Directory needs to have been opened first
 * @post                                Every file handle that was opened using
 *                                      this directory handle needs to be manually
 *                                      closed when they are done being used. Close
 *                                      them with cmapl_close_data_file()
 * @return              cmapl_error_t   Error flag
 * @retval		CMAPL_SUCCESS		Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS	The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION	The version of the handle is not known
 *
 * This method closes a handle to a directory. It will free any memory that it is using.
 * However, all data files handles that were obtained from this directory handle will
 * not be freed with a call to this function. Those data files need to be closed manually
 * using cmapl_close_data_file().
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_close_directory(
                        cmapl_directory_handle_t * dir);

/**
 * @name cmapl_find_and_open_data_file
 *
 * @param[in]           dir             Handle of directory
 * @param[in]           timestamp       UTC Timestamp of file to search for
 * @param[out]          data_file       Handle of found data file
 * @post                                Memory for the data file will be allocated internally.
 *                                      It will be up to the user to close the data file
 *                                      handle passed back to them.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS			Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS		The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION		The version of the handle is not known
 * @retval		CMAPL_TIMESTAMP_OUT_OF_RANGE	The timestamp is not within the range
 *							of the directory handle
 * @retval		CMAPL_FILE_NOT_FOUND		The was no file available at the
 *							specified timestamp
 *
 * This method takes in a handle of a directory, and searches for a data file that contains
 * data sampled at the specified UTC timestamp. The found data file is passed back through the
 * data_file parameter. If no data file was found, then NULL is passed back.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure
 */
extern cmapl_error_t
cmapl_find_and_open_data_file(
                        cmapl_directory_handle_t * dir,
                        uint64_t timestamp,
                        cmapl_file_handle_t ** data_file);

/**
 * @name cmapl_find_and_open_newest_data_file
 *
 * @param[in]           dir             Handle of directory
 * @param[out]          data_file       Handle of newest data file
 * @post                                Memory for the data file will be allocated internally.
 *                                      It will be up to the user to close the data file
 *                                      handle passed back to them.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes in a handle of a directory, and retrieves the handle of the newest data
 * file within the directory handle's time range. This handle will be passed back through
 * the data_file parameter.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure
 */
extern cmapl_error_t
cmapl_find_and_open_newest_data_file(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file);

/**
 * @name cmapl_find_and_open_oldest_data_file
 *
 * @param[in]           dir             Handle of directory
 * @param[out]          data_file       Handle of oldest data file
 * @post                                Memory for the data file will be allocated internally.
 *                                      It will be up to the user to close the data file
 *                                      handle passed back to them.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes in a handle of a directory, and retrieves the handle of the oldest data
 * file within the directory handle's time range. This handle will be passed back through
 * the data_file parameter.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure
 */
extern cmapl_error_t
cmapl_find_and_open_oldest_data_file(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file);

/**
 * @name cmapl_find_and_open_next_data_file
 *
 * @param[in]           dir             Handle of directory
 * @param[out]          data_file       Handle of next data file
 * @pre                                 A previous call to either cmapl_find_and_open_next_data_file,
 *                                      cmapl_find_and_open_data_file, cmapl_find_and_open_oldest_data_file,
 *                                      or cmapl_find_and_open_newest_data_file must have been called on this
 *					directory.
 * @post                                Memory for the data file will be allocated internally.
 *                                      It will be up to the user to close the data file
 *                                      handle passed back to them.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes in a handle of a directory, and retrieves the handle of the next most
 * recent data file within the directory handle's time range. This handle will be passed back
 * through the data_file parameter. The next data file returned depends on how many previous
 * calls to cmapl_find_and_open_next_data_file, cmapl_find_and_open_data_file,
 * cmapl_find_and_open_oldest_data_file, and cmapl_find_and_open_newest_data_file
 * there were. The internal iterator for the directory will be incremented with each call
 * to this function. If there isn't a next data file, then NULL is passed back.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure
 */
extern cmapl_error_t
cmapl_find_and_open_next_data_file(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file);

/**
 * @name cmapl_get_file_infos
 *
 * @param[in]           dir             Handle of directory
 * @param[out]          num_files       The number of files in the directory handle
 * @param[out]          file_infos      Array of the metadata of all the data files in this directory
 *					handle. This includes file paths, start time, and end time
 *					in UTC milliseconds.
 * @post                                Memory for the array of file infos will be allocated by the
 *                                      function. This memory will be kept track of in the directory
 *                                      handle and will be freed when the directory handle is closed,
 *					or when this function is called again.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval		CMAPL_NULL_POINTER	The program is out of memory
 *
 * This method retrieves the info of every single file in this directory handle. This only
 * includes all the files within the time range specified when opening the file. Each file
 * info includes the full path to the generic file, the path to both the .meta and .gz file,
 * and the start and end timestamp of the file in UTC milliseconds. The number of files and
 * the actual file infos are both passed back to the user through the parameters.
 *
 * The file_infos parameter should be of some type of cmapl_file_def_t, depending on the version.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure
 */
extern cmapl_error_t
cmapl_get_file_infos(
                        cmapl_directory_handle_t * dir,
                        uint32_t * num_files,
                        void ** file_infos);

/**
 * @name cmapl_open_directory
 *
 * @param[in]           dir_def         Input directory definition; directory path and time range
 * @param[out]          dir             Handle of open directory
 * @post                                Memory for the directory will be allocated internally.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS			Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS		The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION		The version of the handle is not known
 * @retval              CMAPL_NULL_POINTER		The program is out of memory or directory
 *							could not be opened
 * @retval		CMAPL_CLOSE_DIRECTORY_FAILED	Closing the directory failed
 * @retval              CMAPL_FILE_NOT_FOUND            No file found in the directory with the
 *                                                      given timestamps
 *
 * This method opens a directory and gets all the archive files that were collected between
 * a certain time range specified inside dir_def in UTC time. The parameter dir_def also specifies the
 * path of the directory to open. A directory handle is then passed back through the dir paramter.
 *
 * The dir_def parameter should be of some type of cmapl_directory_def_t, depending on the version.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_open_directory(
                        void * dir_def,
                        cmapl_directory_handle_t ** dir);

#endif /* _CMAPL_DIRECTORY_H_ */
