/**
 * The CMAPL internal directory-level function declarations. These functions
 * are version specific declarations. Different version data files will
 * use functions that will have different implementations. The external
 * functions will call one of these functions depending on the version
 * off the data file.
 *
 * The descriptions of the external CMAPL functions related to each of
 * these internal functions are described in cmapl_directory.h.
 *
 * Syntax of functions:
 *
 *      <function_name>_v<major_version>_<minor_version>
 *
 *      function_name   - the external function name
 *      major_version   - the major version number of the data file
 *      minor_version   - the minor version number of the data file
 *
 *      For Example:    cmapl_find_counter_v6_0 stands for the
 *                      cmapl_find_counter function used on
 *                      a version 6.0 data file.
 *
 */

#ifndef _CMAPL_DIRECTORY_INT_H_
#define _CMAPL_DIRECTORY_INT_H_

cmapl_error_t
cmapl_find_and_open_data_file_v6_0(
                        cmapl_directory_handle_t * dir,
                        uint64_t timestamp,
                        cmapl_file_handle_t ** data_file);

cmapl_error_t
cmapl_find_and_open_newest_data_file_v6_0(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file);

cmapl_error_t
cmapl_find_and_open_oldest_data_file_v6_0(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file);

cmapl_error_t
cmapl_find_and_open_next_data_file_v6_0(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file);

cmapl_error_t
cmapl_get_file_infos_v6_0(
                        cmapl_directory_handle_t * dir,
                        uint32_t * num_files,
                        cmapl_file_def_t_v6_0 ** file_infos);

cmapl_error_t
cmapl_open_directory_v6_0(
                        cmapl_directory_def_t_v6_0 * dir_def,
                        cmapl_directory_handle_t ** dir);

///////////////////////////////////////////////////////////////////////////
// The following functions are internal-only fuctions, meaning they have no
// general external version of them.
///////////////////////////////////////////////////////////////////////////

/**
 * @name cmapl_get_file_path
 *
 * @param[in]           dir                     Handle of directory
 * @param[in]           file_name               Name of the new file
 * @param[out]		file_path		Whole path of the new file
 * @return              cmapl_error_t           Error flag
 *
 * This method will take a file name and the path of the directory to that
 * file, and output the full path of the file. This path will be passed back
 * through the file_path parameter.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_get_file_path_v6_0(
                        cmapl_directory_handle_t * dir,
                        char * file_name,
                        char ** file_path);

/**
 * @name cmapl_increase_file_array_size
 *
 * @param[in]           dir                     Handle of directory
 * @return              cmapl_error_t           Error flag
 *
 * This method will double the array size of the file array in the
 * directory handle.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_increase_file_array_size_v6_0(
                        cmapl_directory_handle_t * dir);

/**
 * @name cmapl_refresh_directory
 *
 * @param[in]		dir			Handle of directory
 * @return              cmapl_error_t           Error flag
 *
 * This method will update the directory handle's list of archive files that
 * it contains. It's possible that some archive files have been either added
 * or removed since the creation of the directory handle, so this function
 * checks for these changes, and updates its list of archive files given its
 * own time range constraints.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_refresh_directory_v6_0(
			cmapl_directory_handle_t * dir);

/**
 * @name cmapl_sort_files
 *
 * @param[in]           dir                     Handle of directory
 * @return              cmapl_error_t           Error flag
 *
 * This method will sort the directory's internal list of files by the files'
 * start times. This function uses insertion sort since the files will most likely
 * already be in mostly sorted order, and insertion sort is very fast in this
 * scenario.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_sort_files_v6_0(
                        cmapl_directory_handle_t * dir);

/**
 * @name cmapl_update_file_info
 *
 * @param[in]           dir                     Handle of directory
 * @param[in]		file_index		Index of the file info to update
 * @param[in]		file_path		Path of the new file
 * @param[in]		index			Index number of the new file
 * @param[in]		start_time		Start time of the new file in UTC in ms
 * @param[in]		end_time		End time of the new file in UTC in ms
 * @return              cmapl_error_t           Error flag
 *
 * This method updates a specific file info structure with the provided information
 * from the parameters. The specific file info to update is specified by file_index.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_update_file_info_v6_0(
                        cmapl_directory_handle_t * dir,
                        uint32_t file_index,
                        char * file_path,
                        uint32_t index,
                        uint64_t start_time,
                        uint64_t end_time);

#endif /* _CMAPL_DIRECTORY_INT_H_ */
