/**
 * @file cmapl_read.h
 *
 * The CMAPL read-side function declarations. These functions provide all the
 * functionality to read, parse, and query archive files.
 *
 */

#ifndef _CMAPL_READ_H_
#define _CMAPL_READ_H_

/**
 * @name cmapl_find_counter_by_name
 *
 * @param[in]           object          Handle of object
 * @param[in]           counter_name    Name of counter to find
 * @param[out]          counter         Handle of found counter
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval		CMAPL_COUNTER_NOT_FOUND	The counter was not found
 *
 * This method takes in a handle of an object, and searches for a specified counter by name.
 * If the counter is found, a handle of that counter is passed to the counter parameter,
 * and the internal iterator for the object will be set to the found counter. If the counter
 * is not found, then a CMAPL_COUNTER_NOT_FOUND error is returned.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_find_counter_by_name(
                        cmapl_object_handle_t * object,
                        const char * counter_name,
                        cmapl_counter_handle_t ** counter);

/**
 * @name cmapl_find_counter_by_id
 *
 * @param[in]           object          Handle of object
 * @param[in]           counter_id      ID of counter to find
 * @param[out]          counter         Handle of found counter
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_COUNTER_NOT_FOUND The counter was not found
 *
 * This method takes in a handle of an object, and searches for a specified counter by id.
 * If the counter is found, a handle of that counter is passed to the counter parameter,
 * and the internal iterator for the object will be set to the found counter. If the counter
 * is not found, then a CMAPL_COUNTER_NOT_FOUND error is returned.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_find_counter_by_id(
                        cmapl_object_handle_t * object,
                        uint32_t counter_id,
                        cmapl_counter_handle_t ** counter);

/**
 * @name cmapl_find_counter_data
 *
 * @param[in]           instance        Handle of instance
 * @param[in]		counter_name	Name of counter to find data for
 * @param[out]          counter         Handle of counter associated with the found data
 * @param[out]          counter_data    Handle of the counter data found
 * @pre                                 The instance handle must come from one of the
 *                                      following commands: cmapl_first_instance,
 *                                      cmapl_next_instance, cmapl_first_instance_from_record,
 *                                      cmapl_next_instance_from_record
 * @post                                Memory for the counter data will be allocated internally.
 *					Any counter data handle resulting from a previous
 *                                      call to cmapl_next_counter_data, cmapl_find_counter_data,
 *                                      or cmapl_first_counter_data will be freed.
 * @return              cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_COUNTER_NOT_FOUND The counter was not found
 *
 * This method takes in a handle of an instance, and searches for a specified counter
 * data by name. If the counter data is found, a handle containing the counter value is
 * passed to the counter_data parameter, and the internal iterator for the instance will
 * be set to the found counter data. The pointer to the handle of the counter associated
 * with the returned counter data will also be passed back. If the counter data is not
 * found, then a CMAPL_COUNTER_DATA_NOT_FOUND error is returned.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_find_counter_data(
                        cmapl_instance_handle_t * instance,
			const char * counter_name,
                        cmapl_counter_handle_t ** counter,
                        cmapl_counter_data_handle_t ** counter_data);

/**
 * @name cmapl_find_instance
 *
 * @param[in]           timestamp	Handle of timestamp
 * @param[in]           key_cnt_names	List of key counter names
 * @param[in]           key_cnt_values	List of key counter values
 * @param[in]           num_key_cnts	Number of key counters
 * @param[out]          instance	Handle of found instance
 * @post                                Memory for the instance will be allocated internally.
 *					Any instance handle resulting from a previous
 *                                      call to cmapl_next_instance, cmapl_find_instance,
 *                                      or cmapl_first_instance will be freed.
 * @return              cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS			Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS		The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION		The version of the handle is not known
 * @retval		CMAPL_INSTANCE_NOT_FOUND	The instance was not found
 *
 * This method takes in a handle of a timestamp, and searches for a specified whole instance.
 * All the key counter names and values of the instance is required in finding the unique instance.
 * The specific structure used for the key counter values will depend on the version.
 * If the instance is found, a handle of that whole instance is passed to the instance
 * parameter, and the internal iterator for the timestamp will be set to the found instance.
 * If the instance is not found, then a CMAPL_INSTANCE_NOT_FOUND error is returned.
 *
 * The key_cnt_values parameter should be an array of some type of cmapl_counter_data_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_find_instance(
			cmapl_timestamp_handle_t * timestamp,
			const char ** key_cnt_names,
			void * key_cnt_values,
			uint32_t num_key_cnts,
			cmapl_instance_handle_t ** instance);

/**
 * @name cmapl_find_object
 *
 * @param[in]           header          Handle of XML header
 * @param[in]           object_name     Name of object to find
 * @param[out]          object          Handle of found object
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval		CMAPL_OBJECT_NOT_FOUND	The object was not found
 *
 * This method takes in a handle of a XML header, and searches for a specified object.
 * If the object is found, a handle of that object is passed to the object parameter, and
 * the internal iterator for the XML header will be set to the found object. If the object
 * is not found, then a CMAPL_OBJECT_NOT_FOUND error is returned.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_find_object(
                        cmapl_header_handle_t * header,
                        const char * object_name,
                        cmapl_object_handle_t ** object);

/**
 * @name cmapl_first_counter
 *
 * @param[in]           object          Handle of object
 * @param[out]          counter         Handle of first counter
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes in a handle of an object, and retrieves a handle of the first
 * counter defined in that object. This will reset the internal iterator for the object
 * to be the first counter.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_first_counter(
			cmapl_object_handle_t * object,
			cmapl_counter_handle_t ** counter);

/**
 * @name cmapl_first_counter_by_period
 *
 * @param[in]           object          Handle of object
 * @param[in]		period		Period of counter
 * @param[out]          counter         Handle of first counter
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval		CMAPL_PERIOD_NOT_FOUND	The period was not found
 *
 * This method takes in a handle of an object, a period, and retrieves a handle of the first
 * counter defined in that object at the specified period. This will reset the internal iterator
 * for this period of this object to be the first counter of that period.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_first_counter_by_period(
                        cmapl_object_handle_t * object,
                        uint64_t period,
                        cmapl_counter_handle_t ** counter);

/**
 * @name cmapl_first_counter_data
 *
 * @param[in]           instance	Handle of instance
 * @param[out]		counter		Handle of counter associated with data
 * @param[out]          counter_data	Handle of first counter data value
 * @post                                Memory for the counter data will be allocated internally.
 *                                      Any counter data handle resulting from a previous
 *                                      call to cmapl_next_counter_data, cmapl_find_counter_data,
 *                                      or cmapl_first_counter_data will be freed.
 * @return              cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval		CMAPL_NULL_POINTER	The program is out of memory
 *
 * This method takes in a handle of an instance, and retrieves a handle to
 * the first counter data in that instance. This will reset the internal iterator
 * for the instance to be the first counter data. The handle to the counter
 * associated with the counter data will also be passed back to the caller.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure
 */
extern cmapl_error_t
cmapl_first_counter_data(
			cmapl_instance_handle_t * instance,
			cmapl_counter_handle_t ** counter,
			cmapl_counter_data_handle_t ** counter_data);

/**
 * @name cmapl_first_instance
 *
 * @param[in]           timestamp       Handle of timestamp
 * @param[out]          instance        Handle of first instance
 * @post				Memory for the instance will be allocated internally.
 *					Any instance handle resulting from a previous
 *                                      call to cmapl_next_instance, cmapl_find_instance,
 *                                      or cmapl_first_instance will be freed.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_NULL_POINTER      The program is out of memory
 *
 * This method takes in a handle of a timestamp, and retrieves a handle of the first
 * whole instance at that timestamp. This will reset the internal iterator for the
 * timestamp to be the first instance.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_first_instance(
			cmapl_timestamp_handle_t * timestamp,
			cmapl_instance_handle_t ** instance);

/**
 * @name cmapl_first_object
 *
 * @param[in]           header          Handle of XML header
 * @param[out]          object          Handle of first object
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes in a handle of a XML header, and retrieves a handle of the first
 * object defined in that XML header's preset. This will reset the internal iterator
 * for the header to be the first object.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_first_object(
			cmapl_header_handle_t * header,
			cmapl_object_handle_t ** object);

/**
 * @name cmapl_first_timestamp
 *
 * @param[in]           query           Handle of query result
 * @param[out]          timestamp       Handle of the first timestamp
 * @post                                Memory for the timestamp will be allocated internally.
 *					Any timestamp handle resulting from a previous
 *                                      call to cmapl_next_timestamp or
 *                                      cmapl_first_timestamp will be freed.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval		CMAPL_RECORD_EMPTY	The record at this timestamp was empty
 *
 * This method takes in a handle to the result of a query, and returns back the
 * first valid timestamp handle. The first timestamp handle returned represents the
 * first timestamp from the query that is both a multiple of all the periods of all the
 * counters being requested and that also contains records sampled at that timestamp
 * for each of those counters' periods. The timestamp handle will contain all the
 * whole instance handles at that timestamp. If there are no more instances to return,
 * the next instance handle passed back will be NULL.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_first_timestamp(
			cmapl_query_handle_t * query,
			cmapl_timestamp_handle_t ** timestamp);

/**
 * @name cmapl_get_counter_data_value
 *
 * @param[in]           counter_data		Handle of counter data
 * @param[out]          counter_data_value	Value of counter found
 * @post					Memory allocated by this function will be
 *						freed when the data file is closed
 * @return              cmapl_error_t		Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes the handle of a counter data and gets its value. That data is passed
 * back to the counter_data_value parameter. The specific data structure that is returned
 * will depend on the version of the data file.
 *
 * The counter_data_value parameter should be of some type of cmapl_counter_data_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_counter_data_value(
                        cmapl_counter_data_handle_t * counter_data,
                        void * counter_data_value);

/**
 * @name cmapl_get_counter_info
 *
 * @param[in]           counter         Handle of counter
 * @param[out]          counter_info    Metadata of counter
 * @post				Memory allocated by this function will be
 *					freed when the data file is closed
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes the handle of a counter and gets its metadata. That data is passed
 * back through the counter_info parameter. The specific data structure that is returned
 * will depend on the version of the data file.
 *
 * The counter_info parameter should be of some type of cmapl_counter_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_counter_info(
			cmapl_counter_handle_t * counter,
			void * counter_info);

/**
 * @name cmapl_get_counter_type
 *
 * @param[in]           counter         Handle of counter
 * @param[out]          size            Size of the counter in bytes
 * @param[out]          type            Type of the counter, such as raw or average
 * @param[out]          unit            Suffix units of the counter, such as secs or msecs
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes the handle of a counter and retrieves the counter's type information.
 * This includes the counter's size, type, and unit. Possible values for each type is defined
 * in cmapl_common_defs.h .
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_counter_type(
                        cmapl_counter_handle_t * counter,
                        cmapl_cnt_size_t * size,
                        cmapl_cnt_type_t * type,
                        cmapl_cnt_unit_t * unit);

/**
 * @name cmapl_get_file_info
 *
 * @param[in]           data_file       Handle of data file
 * @param[out]          file_info	Metadata of file
 * @post                                Memory allocated by this function will be
 *                                      freed when the data file is closed
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes the handle of a data file and gets its metadata. That data is passed
 * back through the file_info parameter. The specific data structure that is returned
 * will depend on the version of the data file.
 *
 * The file_info parameter should be of some type of cmapl_file_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_file_info(
                        cmapl_file_handle_t * data_file,
                        void * file_info);

/**
 * @name cmapl_get_header
 *
 * @param[in]           data_file	Handle of data file
 * @param[out]          header		Handle of header
 * @post                                The function will allocate the memory that's
 *                                      passed back to the header parameter. It
 *                                      will be freed when the data file is closed
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method gets the header handle associated with a specific data file. That data
 * passed back through the header parameter.
 *
 *The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_header(
                        cmapl_file_handle_t * data_file,
                        cmapl_header_handle_t ** header);

/**
 * @name cmapl_get_header_info
 *
 * @param[in]           header          Handle of header
 * @param[out]          header_info     XML header information
 * @post                                The function will allocate the memory that's
 *					passed back to the header_info parameter. It
 *					will be freed when the data file is closed
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method looks in a header handle for all the XML header information and returns
 * a header data structure. This information is passed back by reference through the
 * header_info parameter. The specific data structure that is returned will depend on
 * the version of the data file.
 *
 * The header_info parameter should be of some type of cmapl_header_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_header_info(
			cmapl_header_handle_t * header,
			void * header_info);

/**
 * @name cmapl_get_instance_data
 *
 * @param[in]           instance        Handle of instance
 * @param[out]          instance_data   Metadata of instance
 * @post                                Memory allocated by this function will be
 *                                      freed when the data file is closed
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes the handle of an instance and gets its metadata. That data is
 * passed back to the instance_data parameter.
 *
 * The instance_data parameter should be of some type of cmapl_instance_data_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_instance_data(
			cmapl_instance_handle_t * instance,
			void * instance_data);

/**
 * @name cmapl_get_object_info
 *
 * @param[in]           object          Handle of object
 * @param[out]          object_info     Metadata of object
 * @post                                Memory allocated by this function will be
 *                                      freed when the data file is closed
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes the handle of a object and gets its metadata. That data is passed
 * back to the object_info parameter. The specific data structure that is returned
 * will depend on the version of the data file.
 *
 * The object_info parameter should be of some type of cmapl_object_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_object_info(
			cmapl_object_handle_t * object,
			void * object_info);

/**
 * @name cmapl_get_object_periods
 *
 * @param[in]           object          Handle of object
 * @param[out]          periods         List of all this object's counters' periods in milliseconds
 * @param[out]          num_periods     Number of periods in list
 * @post                                The function will allocate the memory that's passed back to
 *					the periods parameter. It will be the responsibility of the
 *					caller to free the array of periods.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes a handle of a certain object and finds all the different periods
 * its counters are sampled at. A pointer to the list of periods and the number of
 * periods are passed back by reference. The periods are in order from smallest to
 * largest.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_object_periods(
			cmapl_object_handle_t * object,
			uint64_t ** periods,
			uint32_t * num_periods);

/**
 * @name cmapl_get_system_info
 *
 * @param[in]           header          Handle of header
 * @param[out]          system_info     System/Ontap metadata of the archive file
 * @post                                The function will allocate the memory that's
 *					passed back to the system_info parameter. This
 *					will be freed when the data file is closed.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method gets all the system metadata associated with the archive file and its XML
 * header. This data will be stored in a cmapl_hdr_ontap_t data structure and passed back
 * through the system_info parameter. The specific data structure that is returned
 * will depend on the version of the data file.
 *
 * The system_info parameter should be of some type of cmapl_system_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_system_info(
			cmapl_header_handle_t * header,
			void * system_info);

/**
 * @name cmapl_get_timestamp
 *
 * @param[in]           timestamp       Handle of timestamp
 * @param[out]          time		Actual UTC time of this timestamp handle in milliseconds
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method gets the UTC time of a timestamp handle in milliseconds.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_timestamp(
                        cmapl_timestamp_handle_t * timestamp,
                        uint64_t * time);

/**
 * @name cmapl_next_counter
 *
 * @param[in]           object          Handle of object
 * @param[out]          counter         Handle of next counter
 * @pre                                 A previous call to either cmapl_next_counter,
 *					cmapl_find_counter, or cmapl_first_counter must have
 *                                      been called on this object.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes in a handle of an object, and retrieves a handle of the next counter
 * defined in that object. The next counter returned depends on how many previous calls
 * to cmapl_next_counter, cmapl_find_counter, and cmapl_first_counter there were. The
 * internal iterator for the object will be incremented with each call to this function.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_next_counter(
			cmapl_object_handle_t * object,
			cmapl_counter_handle_t ** counter);

/**
 * @name cmapl_next_counter_by_period
 *
 * @param[in]           object          Handle of object
 * @param[in]		period		Period of counter in milliseconds
 * @param[out]          counter         Handle of next counter
 * @pre                                 A previous call to either cmapl_next_counter_by_period,
 *                                      cmapl_find_counter, or cmapl_first_counter_by_period
 *                                      must have been called on this object.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval		CMAPL_PERIOD_NOT_FOUND	The period was not found
 *
 * This method takes in a handle of an object, and retrieves a handle of the next counter
 * sampled at the specified period defined in that object. The next counter returned depends
 * on how many previous calls to cmapl_next_counter_by_period, cmapl_find_counter, and
 * cmapl_first_counter_by_period there were. The internal iterator for the object will be
 * incremented with each call to this function.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_next_counter_by_period(
                        cmapl_object_handle_t * object,
			uint64_t period,
                        cmapl_counter_handle_t ** counter);

/**
 * @name cmapl_next_counter_data
 *
 * @param[in]           instance	Handle of instance
 * @param[out]		counter		Handle of counter associated with data
 * @param[out]          counter_data	Handle of next counter data value
 * @pre					A previous call to either cmapl_next_counter_data,
 *					cmapl_find_counter_data, or cmapl_first_counter_data
 *					must have been called on this instance.
 *                                      The instance handle must come from one of the
 *                                      following commands: cmapl_first_instance,
 *                                      cmapl_next_instance, cmapl_first_instance_from_record,
 *                                      cmapl_next_instance_from_record
 * @post                                Memory for the counter data will be allocated internally.
 *                                      Any counter data handle resulting from a previous
 *                                      call to cmapl_next_counter_data, cmapl_find_counter_data,
 *                                      or cmapl_first_counter_data will be freed.
 * @return              cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_NULL_POINTER      The program is out of memory
 *
 * This method takes in a handle of an instance, and retrieves the handle of the
 * next counter data. The next counter data returned depends on how many previous calls
 * to cmapl_next_counter_data, cmapl_find_counter_data, and cmapl_first_counter_data
 * there were. The internal iterator for the instance will be incremented with each call
 * to this function. A handle to the counter associated with the found counter data
 * will also be passed back to the caller. The counter data are returned in the order
 * they were requested in the query. If no counter was requested, then all the counters
 * are returned in order of id.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_next_counter_data(
			cmapl_instance_handle_t * instance,
			cmapl_counter_handle_t ** counter,
			cmapl_counter_data_handle_t ** counter_data);

/**
 * @name cmapl_next_instance
 *
 * @param[in]           timestamp       Handle of timestamp
 * @param[out]          instance        Handle of next instance
 * @pre                                 A previous call to either cmapl_next_instance,
 *					cmapl_find_instance, or cmapl_first_instance must
 *					have been called on this timestamp.
 * @post                                Memory for the instance will be allocated internally.
 *					Any instance handle resulting from a previous
 *                                      call to cmapl_next_instance, cmapl_find_instance, or
 *                                      cmapl_first_instance will be freed.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_NULL_POINTER      The program is out of memory
 *
 * This method takes in a handle of an timestamp, and retrieves a handle of its next
 * whole instance. The next instance returned depends on how many previous calls to
 * cmapl_next_instance, cmapl_find_instance, and cmapl_first_instance there were.
 * The internal iterator for the timestamp will be incremented with each call to this
 * function. If there are no more instances to return, the next instance handle passed
 * back will be NULL.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_next_instance(
			cmapl_timestamp_handle_t * timestamp,
			cmapl_instance_handle_t ** instance);

/**
 * @name cmapl_next_object
 *
 * @param[in]           header          Handle of XML header
 * @param[out]          object          Handle of object
 * @pre                                 A previous call to either cmapl_next_object,
 *					cmapl_find_object, or cmapl_first_object must
 *					have been called on this object.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method takes in a handle of a XML header, and retrieves a handle of the next object
 * defined in that XML header's preset. The next object returned depends on how many previous
 * calls to cmapl_next_object, cmapl_find_object, and cmapl_first_object there were. The
 * internal iterator for the XML header will be incremented with each call to this function.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_next_object(
			cmapl_header_handle_t * header,
			cmapl_object_handle_t ** object);

/**
 * @name cmapl_next_timestamp
 *
 * @param[in]           query           Handle of query result
 * @param[out]          timestamp	Handle of next timestamp
 * @pre                                 A previous call to either cmapl_next_timestamp
 *                                      or cmapl_first_timestamp must have been called
 *					on this query.
 * @post				Memory for the timesatmp will be allocated internally.
 *					Any timestamp handle resulting from a previous
 *					call to cmapl_next_timestamp or
 *					cmapl_first_timestamp will be freed.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_NULL_POINTER      The program is out of memory
 * @retval		CMAPL_RECORD_EMPTY	The record at the next timestamp is empty
 *
 * This method takes in a handle of the result of a query and retrieves the next timestamp
 * handle that's a result of this query. The next timestamp returned depends on how many
 * previous calls to cmapl_next_timestamp and cmapl_first_timestamp there were.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_next_timestamp(
                        cmapl_query_handle_t * query,
                        cmapl_timestamp_handle_t ** timestamp);

/**
 * @name cmapl_open_data_file
 *
 * @param[in]           file_path       Path to data file to be opened
 * @param[in]           access_type     Access type for the opened data file; Can be
 *                                      either "r" or "rw"/"wr"
 * @param[out]          data_file       Handle of opened data file
 * @param[out]          meta_path       Path to the metadata file
 * @param[out]          data_path       Path to the record data file
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_NULL_POINTER      The program is out of memory
 * @retval              CMAPL_PARSING_ERROR	Parsing the XML header failed
 * @retval		CMAPL_CLOSE_FILE_FAILED	Closing the file failed
 * @retval		CMAPL_READ_FAILED	Reading the file failed
 *
 * This method opens a certain archive file given the path to the file and an access type.
 * The access type can be either "r" or "rw"/"wr". "r" means the file will be read-only.
 * "rw" or "wr" means the file can both be read and rewritten. The handle of the opened
 * data file will be passed back through the data_file parameter.
 *
 * The archive file is split into two separate files. One is uncompressed and contains all the
 * metadata and indexing information. The other is compressed and contains all the record data.
 * The paths to both of these files will be passed back to the user. The file_path used to open
 * the file will be the same path used when creating the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_open_data_file(
                        const char * file_path,
                        const char * access_type,
                        cmapl_file_handle_t ** data_file,
			const char ** meta_path,
			const char ** data_path);

/**
 * @name cmapl_query
 *
 * @param[in]           data_file       Handle of data file
 * @param[in]           query_def       Input query definition; filter on object, instances,
 *					and counters
 * @param[out]          query           Query handle containing a list of timestamps, each
 *					of which also contains a list of matching handles
 *					of instances
 * @post                                The function will allocate the memory that's passed
 *					back to the query parameter. Calling this function
 *					again will free the memory used by the output of the
 *					previous call to this function. Closing the file will
 *					also free the memory of the output of this function.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_NULL_POINTER      The program is out of memory
 * @retval		CMAPL_INVALID_TIMESTAMP The start and/or end timestamps are out of range

 *
 * This method queries a data file with certain parameters to filter on, and the
 * result of the query is then stored inside a query handle data structure.
 * The result will be passed by reference to the query parameter. The resulting query
 * handle will contain timestamp handles that match the the query period. The result
 * of the query can be filtered by a query data structure. The query data structure
 * depends on the version of the file specifies parameters to filter on
 * such as object name, key counter name/value pairs, regular counters, start and end
 * timestamps, and the period.
 *
 * The query_def parameter should be of some type of cmapl_query_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure
 */
extern cmapl_error_t
cmapl_query(
			cmapl_file_handle_t * data_file,
			void * query_def,
			cmapl_query_handle_t ** query);

#ifdef _CMAPL_FIJI_
/**
 * @name cmapl_disable_fiji
 *
 * @param[in]           error           Error flag that is to be untriggered; only errors
 *                                      allowed at CMAPL_READ_FAILED, CMAPL_SEEK_FAILED,
 *                                      and CMAPL_WRITE_FAILED.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 *
 * This method disables the fault injections for the functions fread, fseek, fwrite, and their
 * gz* equivalents. Depending on which error code is passed into this function, one of those
 * functions will no longer be forced to trigger an error. Passing in CMAPL_READ_FAILED will
 * cause fread and gzread to behave normally. Passing in CMAPL_SEEK_FAILED will cause fseek
 * and gzseek to behave normally. Passing in CMAPL_WRITE_FAILED will cause fwrite and gzwrite
 * to behave normally.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure
 *
 * NOTE: This function is for QA use only! No one other than QA should ever need to call
 * this function!
 */
extern cmapl_error_t
cmapl_disable_fiji(
                        cmapl_error_t error);

/**
 * @name cmapl_enable_fiji
 *
 * @param[in]           error		Error flag that is to be triggered; only errors
 *					allowed at CMAPL_READ_FAILED, CMAPL_SEEK_FAILED,
 *					and CMAPL_WRITE_FAILED.
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 *
 * This method enables the fault injections for the functions fread, fseek, fwrite, and their
 * gz* equivalents. Depending on which error code is passed into this function, one of those
 * functions will always return an error. Passing in CMAPL_READ_FAILED will cause fread and
 * gzread to fail. Passing in CMAPL_SEEK_FAILED will cause fseek and gzseek to fail. Passing
 * in CMAPL_WRITE_FAILED will cause fwrite and gzwrite to fail.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure
 *
 * NOTE: This function is for QA use only! No one other than QA should ever need to call
 * this function!
 */
extern cmapl_error_t
cmapl_enable_fiji(
                        cmapl_error_t error);
#endif /* _CMAPL_FIJI_ */

#endif /* _CMAPL_READ_H_ */
