/**
 * @file cmapl_write.h
 *
 * The CMAPL write-side function declarations. These functions provide all the
 * functionality to create, update, and modify archive files.
 *
 */

#ifndef _CMAPL_WRITE_H_
#define _CMAPL_WRITE_H_

/**
 * @name cmapl_add_attr_def
 *
 * @param[in]		header		Handle of XML header that will add this new attribute definition
 * @param[in]		attr_def	Definition of the attribute to be added
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval		CMAPL_VERSION_MISMATCH	Version of arguments don't match
 * @retval		CMAPL_NULL_POINTER	The progam is out of memory
 *
 * This method adds a new attribute definition to the given XML header.
 * An attribute definition describes any attributes and their legal/valid
 * values that might appear in any element definitions later on.
 *
 * The attr_def parameter should be of some type of cmapl_attribute_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_add_attr_def(
			cmapl_header_handle_t * header,
			void * attr_def);

/**
 * @name cmapl_add_counter
 *
 * @param[in]		object		Handle of object that will add the new counter
 * @param[in]		counter_def	Data structure containing new counter's metadata
 * @param[out]		counter		Handle of new counter
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_VERSION_MISMATCH  Version of arguments don't match
 * @retval		CMAPL_COUNTER_MISMATCH	The counter has already been added before
 *						but their properties don't match
 * @retval		CMAPL_NULL_POINTER      The progam is out of memory
 *
 *
 * This method adds a new counter to an object in the preset of a XML header.
 * The definition of this new counter must be passed in as a cmapl_counter_def_t data
 * structure through the counter_def parameter. A handle of this counter is passed back through
 * the counter parameter.
 *
 * The counter_def parameter should be of some type of cmapl_counter_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_add_counter(
			cmapl_object_handle_t * object,
			void * counter_def,
			cmapl_counter_handle_t ** counter);

/**
 * @name cmapl_add_object
 *
 * @param[in]		header		Handle of XML header that will add this new object
 * @param[in]		object_data	Data structure containing new object's metadata
 * @param[out]		object		Handle of new object
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_VERSION_MISMATCH  Version of arguments don't match
 * @retval              CMAPL_NULL_POINTER      The progam is out of memory
 *
 * This method adds a new object to the preset of the XML header.
 * The definition of this new object will be passed in by the user through the
 * object_def parameter. A handle of this object is passed back to the object parameter.
 *
 * The object_def parameter should be of some type of cmapl_object_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_add_object(
			cmapl_header_handle_t * header,
			void * object_def,
			cmapl_object_handle_t ** object);

/**
 * @name cmapl_close_data_file
 *
 * @param[in]		data_file	Handle of data file
 * @pre					Data file needs to have either been opened or created first
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval		CMAPL_SEEK_FAILED	Seeking into the file failed
 * @retval		CMAPL_CLOSE_FILE_FAILED	Closing the file handle failed
 *
 * This method closes a data file after it is done being written to. All information
 * about the data file still in memory will be freed as well.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_close_data_file(
			cmapl_file_handle_t * data_file);

/**
 * @name cmapl_copy_counter
 *
 * @param[in]		object		Handle of object that is adding the new copied counter
 * @param[in]		counter		Handle of counter to be copied
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 *
 * This method takes in a handle to a counter and copies the counter into a given
 * object. The data of both counters will be identical.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_copy_counter(
			cmapl_object_handle_t * object,
			cmapl_counter_handle_t * counter);

/**
 * @name cmapl_copy_header
 *
 * @param[in]		data_file	Handle of archive file that is using the new copied header
 * @param[in]		header		Handle of XML header to be copied
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 *
 * This method takes in a handle to a XML header and copies the XML header into a
 * given data file. The data of both headers will be identical, including all the objects
 * and counters in its preset information.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_copy_header(
			cmapl_file_handle_t * data_file,
			cmapl_header_handle_t * header);

/**
 * @name cmapl_copy_object
 *
 * @param[in]		header		Handle of XML header that will add the new copied object
 * @param[in]		object		Handle of object to be copied
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 *
 * This method takes in a handle to a object and copies the object into a given
 * XML header. The data of both objects will be identical, including all the
 * counters of the object.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_copy_object(
			cmapl_header_handle_t * header,
			cmapl_object_handle_t * object);

/**
 * @name cmapl_create_data_file
 *
 * @param[in]		file_path	Path to this data file
 * @param[in]		major_v		Major version of this data file
 * @param[in]		minor_v		Minor version of this data file
 * @param[in]		duration	Time in milliseconds this file will sample for
 * @param[out]		data_file	Handle of new data file
 * @param[out]		meta_path	Path to the metadata file
 * @param[out]		data_path	Path to the record data file
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_NULL_POINTER      The progam is out of memory
 *
 * This method creates a new archive file that the user can start writing records to. The version of
 * this file will be passed in, indicating how this file will be parsed. Both a major and a minor
 * version number needs to be specified. A duration for the file will also be specified indicating
 * how long this file is to be sampled for in milliseconds.
 *
 * The archive file will be split into two separate files. One is uncompressed and contains all the
 * metadata and indexing information. The other is compressed and contains all the record data. The
 * paths to both of these files will be passed back to the user.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_create_data_file(
			const char * file_path,
			uint32_t major_v,
			uint32_t minor_v,
			uint64_t duration,
			cmapl_file_handle_t ** data_file,
			const char ** meta_path,
			const char ** data_path);

/**
 * @name cmapl_create_header
 *
 * @param[in]		data_file	Handle of archive file that will use this new header
 * @param[in]		header_def	Metadata of XML header
 * @param[in]		system_def	System information for XML header
 * @param[out]		header		Handle of new XML header
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_VERSION_MISMATCH  Version of arguments don't match
 * @retval              CMAPL_NULL_POINTER      The progam is out of memory
 *
 * This method creates a new handle of a XML header to be written to. The XML header
 * The header_data parameter specifies all the metadata of this XML header. The system
 * information related to this archive file is passed in through the system_def
 * parameter. A handle of this XML header is passed back through the header parameter.
 *
 * The header_def parameter should be of some type of cmapl_header_def_t struct,
 * depending on the version of the file.
 *
 * The system_def parameter should be of some type of cmapl_system_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_create_header(
			cmapl_file_handle_t * data_file,
			void * header_def,
			void * system_def,
			cmapl_header_handle_t ** header);

/**
 * @name cmapl_delete_counter
 *
 * @param[in]		object		Handle of object where counter is to be deleted from
 * @param[in]		counter		Handle of counter to be deleted
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method deletes and removes a counter from a specified object in the preset
 * of a XML header. The object will result in one less counter and the pointer to
 * the counter handle will be freed.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_delete_counter(
			cmapl_object_handle_t * object,
			cmapl_counter_handle_t * counter);

/**
 * @name cmapl_delete_object
 *
 * @param[in]		header		Handle of XML header where object is to be deleted from
 * @param[in]		object		Handle of object to be deleted
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method deletes and removes an object from the preset of a XML header. As
 * a result, all the counters of this object will be deleted and removed as well.
 * The XML header will result in one less object and the pointer to the object
 * handle will be freed, along with all of its counters.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_delete_object(
			cmapl_header_handle_t * header,
			cmapl_object_handle_t * object);

/**
 * @name cmapl_finalize_header
 *
 * @param[in]		data_file	Handle of archive file that will have it's header written
 * @pre					The XML header of the given archive file must be finished
 *					being created before this function is called.
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_NULL_POINTER      The progam is out of memory
 *
 * This method will write the XML header associated with a specific archive file
 * to the actual file. This method should only be called once the XML header for
 * an archive file is finished being constructed. The XML header will be the first
 * thing written to the archive file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_finalize_header(
			cmapl_file_handle_t * data_file);

/**
 * @name cmapl_ftell
 *
 * @param[in]           data_file       Handle of archive file
 * @param[out]          count		Number of uncompressed bytes written so far
 * @return		cmapl_error_t	Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method returns the number of uncompressed bytes written to the data file so far. This
 * value will be passed back through the count parameter.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure..
 */
extern cmapl_error_t
cmapl_ftell(
			cmapl_file_handle_t * data_file,
			uint64_t * count);

/**
 * @name cmapl_get_newest_record_time
 *
 * @param[in]           data_file       Handle of data file
 * @param[out]          newest_time     UTC time since epoch of the last record in the file
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method retrieves the time stamp of the last record in a given data file, in terms
 * of UTC time since epoch.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_newest_record_time(
                        cmapl_file_handle_t * data_file,
                        uint64_t * newest_time);

/**
 * @name cmapl_get_oldest_record_time
 *
 * @param[in]           data_file       Handle of data file
 * @param[out]          oldest_time     UTC time since epoch of the first record in the file
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method retrieves the time stamp of the first record in a given data file, in terms
 * of UTC time since epoch.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_oldest_record_time(
                        cmapl_file_handle_t * data_file,
                        uint64_t * oldest_time);

/**
 * @name cmapl_unlink
 *
 * @param[in]           file_path	Path of the data file
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_NULL_POINTER      The progam is out of memory
 * @retval		CMAPL_UNLINK_ERROR	Deleting the file failed
 *
 * This method deletes the compressed data file and meta file associated with the file_path
 * that is passed in.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure..
 */
extern cmapl_error_t
cmapl_unlink(
                        const char * file_path);

// TODO: remove function
/**
 * @name cmapl_write_records
 *
 * @param[in]		data_file	Handle of data file
 * @param[in]		records_buffer	Buffer of raw binary records
 * @param[in]		buffer_size	Size in bytes of the buffer
 * @return		cmapl_error_t	Error flag
 *
 * This method writes some number of records to the given data file. The ordering
 * of the records written much match the order that was specified during the creation
 * of the data file. It is up to the caller to regulate this. The raw binary of the
 * records is to be passed in to be written to the archive file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
/*
extern cmapl_error_t
cmapl_write_records(
			cmapl_file_handle_t * data_file,
			char * records_buffer,
			size_t buffer_size);
*/

/**
 * @name cmapl_write_record
 *
 * @param[in]           data_file       Handle of data file
 * @param[in]           record_info     Metadata of the record
 * @param[in]           record_buffer   Buffer of raw binary record
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 * @retval              CMAPL_VERSION_MISMATCH  Version of arguments don't match
 * @retval              CMAPL_NULL_POINTER      The progam is out of memory
 * @retval		CMAPL_WRITE_FAILED	Writing to the file failed
 *
 * This method writes a single record to the given data file. The metadata
 * is passed in through the record_info parameter. The actual data structure
 * passed in will depend on the version of the data file. The raw binary of
 * the record is to be passed in to be written to the archive file.
 *
 * The record_info parameter should be of some type of cmapl_record_header_data_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_write_record(
                        cmapl_file_handle_t * data_file,
                        void * record_info,
                        char * record_buffer);

/**
 * @name cmapl_set_start_time
 *
 * @param[in]           data_file       Handle of data file
 * @param[in]           start_time      Timestamp of the first sample in the file,
                                        in msecs since base time of the file
 * @return              cmapl_error_t   Error flag
 * @retval              CMAPL_SUCCESS           Function finished successfully
 * @retval              CMAPL_INVALID_ARGUMENTS The argument was NULL or invalid
 * @retval              CMAPL_INVALID_VERSION   The version of the handle is not known
 *
 * This method sets the start time of the data file to the value start_time specified
 * by the user. This start time represents the timestamp of the earliest sample in the
 * file represented in msecs since boot.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 *
 * NOTE: This function should only be used by the ccma when writing archive files.
 */
extern cmapl_error_t
cmapl_set_start_time(
                        cmapl_file_handle_t * data_file,
                        uint64_t start_time);

#endif /* _CMAPL_WRITE_H_ */
