/**
 * The CMAPL internal write-side function declarations. These functions
 * are version specific declarations. Different version data files will
 * use functions that will have different implementations. The external
 * functions will call one of these functions depending on the version
 * off the data file.
 *
 * The descriptions of the external CMAPL functions related to each of
 * these internal functions are described in cmapl_write.h. The rest are
 * internal-only functions and are described in this file.
 *
 * Syntax of functions:
 *
 *	<function_name>_v<major_version>_<minor_version>
 *
 *	function_name	- the external function name
 *	major_version	- the major version number of the data file
 *	minor_version	- the minor version number of the data file
 *
 *	For Example:	cmapl_create_counter_1 stands for the
 *			cmapl_create_counter function used on
 *			a version 1 data file.
 *
 */

#ifndef _CMAPL_WRITE_INT_H_
#define _CMAPL_WRITE_INT_H_

cmapl_error_t
cmapl_add_attr_def_v6_0(
			cmapl_header_handle_t * header,
			cmapl_attribute_def_t_v6_0 * attr_def);

cmapl_error_t
cmapl_add_counter_v6_0(
                        cmapl_object_handle_t * object,
                        cmapl_counter_def_t_v6_0 * counter_def,
                        cmapl_counter_handle_t ** counter);

cmapl_error_t
cmapl_add_object_v6_0(
                        cmapl_header_handle_t * header,
                        cmapl_object_def_t_v6_0 * object_def,
                        cmapl_object_handle_t ** object);

cmapl_error_t
cmapl_close_data_file_v6_0(
			cmapl_file_handle_t * data_file);

cmapl_error_t
cmapl_create_data_file_v6_0(
			const char * file_path,
			uint32_t major_v,
			uint32_t minor_v,
			uint64_t duration,
			cmapl_file_handle_t ** data_file,
			const char ** meta_path,
                        const char ** data_path);

cmapl_error_t
cmapl_create_header_v6_0(
			cmapl_file_handle_t * data_file,
			cmapl_header_def_t_v6_0 * header_def,
			cmapl_system_def_t_v6_0 * system_def,
			cmapl_header_handle_t ** header);

cmapl_error_t
cmapl_delete_counter_v6_0(
			cmapl_object_handle_t * object,
			cmapl_counter_handle_t * counter);

cmapl_error_t
cmapl_delete_object_v6_0(
			cmapl_header_handle_t * header,
			cmapl_object_handle_t * object);

cmapl_error_t
cmapl_finalize_header_v6_0(
			cmapl_file_handle_t * data_file);

cmapl_error_t
cmapl_get_newest_record_time_v6_0(
                        cmapl_file_handle_t * data_file,
                        uint64_t * newest_time);

cmapl_error_t
cmapl_get_oldest_record_time_v6_0(
                        cmapl_file_handle_t * data_file,
                        uint64_t * oldest_time);

/*
// TODO: remove function
cmapl_error_t
cmapl_write_records_v6_0(
			cmapl_file_handle_t * data_file,
			char * records_buffer,
			size_t buffer_size);
*/

cmapl_error_t
cmapl_write_record_v6_0(
                        cmapl_file_handle_t * data_file,
                        cmapl_record_header_data_t_v6_0 * record_info,
                        char * record_buffer);

cmapl_error_t
cmapl_set_start_time_v6_0(
                        cmapl_file_handle_t * data_file,
                        uint64_t start_time);


///////////////////////////////////////////////////////////////////////////
// The following functions are internal-only fuctions, meaning they have no
// general external version of them.
///////////////////////////////////////////////////////////////////////////

/**
 * @name cmapl_get_num_record_offsets_v6_0
 *
 * @param[in]		start_time		The start time of the records being sampled
 * @param[in]		end_time		The end time of the records being sampled
 * @param[in]           period			The period of the record
 * @param[out]          num_offsets		The number of offsets for this period
 * @return              cmapl_error_t           Error flag
 *
 * This method calculates the number of record offsets for a specific period. Each offset
 * is used to point to some record in the file that is sampled at this particular period.
 * The number of offsets should be the duration of the file divided by the period. For
 * config periods, there should only be one record offset since there will only be one
 * config record per object per file.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_get_num_record_offsets_v6_0(
			uint64_t start_time,
			uint64_t end_time,
			uint64_t period,
			uint32_t * num_offsets);

/**
 * @name cmapl_update_object_index_v6_0
 *
 * @param[in]           data_file               Handle of data file
 * @param[in]           r_period_timestamp      The programmed timestamp of the record in ms
 * @param[in]           r_period                The period of the record in ms
 * @param[in]           r_oid                   The object ID of the record
 * @return              cmapl_error_t           Error flag
 *
 * This method updates the object index with information from a single record. This includes
 * updating the object and period table, the correct period offsets table, and the iterative
 * records table if necessary.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_update_object_index_v6_0(
                        cmapl_file_handle_t * data_file,
                        uint64_t r_period_timestamp,
                        uint64_t r_period,
                        uint32_t r_oid);

/**
 * @name cmapl_write_iterative_record_offsets_v6_0
 *
 * @param[in]           data_file       Handle of data file
 * @param[in]           period          Period for these records
 * @param[in]           offset_index    Index into the array of the iterative records
 *                                      we care about
 * @param[out]          written_count   Number of bytes written to the file so far
 * @param[in]           write_flag      Flag that indicates whether to actually write
 *                                      the offsets (true), or to simulate the write
 *                                      and update the object index instead (false)
 * @return              cmapl_error_t   Error flag
 *
 * This method writes and updates the iterative record offsets for a certain object,
 * period, and timestamp. The correct period is specified by the period parameter, and
 * the iterative records that need to be written or updated are indicated by the
 * offset_index parameter. The write_flag parameter specifies whether the offsets are
 * to be written to the file (true) or if they are suppose to be updated in the object
 * index (false).
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_write_iterative_record_offsets_v6_0(
                        cmapl_file_handle_t * data_file,
                        cmapl_period_t * period,
                        uint32_t offset_index,
                        uint64_t * written_count,
                        bool write_flag);

/**
 * @name cmapl_write_object_index_v6_0
 *
 * @param[in]           data_file       Handle of data file
 * @return              cmapl_error_t   Error flag
 *
 * This method writes the whole object index to the specified data file. The object index
 * will located records based on object ID, then period, and then the programmed timestamp.
 * Iterative records for a given object, period, and timstamp will also be accounted for.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_write_object_index_v6_0(
                        cmapl_file_handle_t * data_file);

/**
 * @name cmapl_write_period_offsets_v6_0
 *
 * @param[in]           data_file       Handle of data file
 * @param[in]           period          Period for these records
 * @param[in]           obj_count       The specific object the period offsets belong to
 * @param[out]          written_count   Number of bytes written to the file so far
 * @param[in]           write_flag      Flag that indicates whether to actually write
 *                                      the offsets (true), or to simulate the write
 *                                      and update the object index instead (false)
 * @return              cmapl_error_t   Error flag
 *
 * This method writes and updates the period offsets for a certain object. The object
 * is specified by the obj_count parameter which stands for the position of the object
 * when all the objects in the file are sorted by object ID. The write_flag parameter
 * specifies whether the offsets are to be written to the file (true) or if they are
 * suppose to be updated in the object index (false).
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
cmapl_error_t
cmapl_write_period_offsets_v6_0(
                        cmapl_file_handle_t * data_file,
                        cmapl_period_t * period,
                        uint32_t obj_count,
                        uint64_t * written_count,
                        bool write_flag);

#endif /* _CMAPL_WRITE_INT_H_ */
