###############################################################################
# $Id: //depot/prod/test/Rsierranevada/lib/PerfLib/PerfLib.pm#1 $
#
# Copyright (c) 2007-2009 NetApp, Inc.
# All rights reserved.
###############################################################################
package PerfLib;

use strict;
use DBI;
use Data::Dumper;
use Time::HiRes;

#
# --- Table structures ---
# 

my $test_table = {name => "test",
    columns => {
        id => "INTEGER PRIMARY KEY AUTO_INCREMENT",
        name => "VARCHAR(255)",
        short_desc => "VARCHAR(255)",
        long_desc => "VARCHAR(255)"
    }
};
my $run_table = {name => "run",
    columns => {
        id => "INTEGER PRIMARY KEY AUTO_INCREMENT",
        test_id => "INTEGER",
        testcase => "VARCHAR(255)",
        active => "INTEGER" # whether the current run is still running
        # other stuff to go in here:
        #   username -> the user who ran the test
        #   start_time -> when the run was started
        #   end_time -> when the run ended
    }
};
my $result_table = {name => "result",
    columns => {
        id => "INTEGER PRIMARY KEY AUTO_INCREMENT",
        run_id => "INTEGER",
        sub_test => "VARCHAR(255)",
        samples => "INTEGER"
        # other stuff to go in here:
        #   timestamp -> when the result was inserted 
        #   cluster -> a list of nodes (which is a unique identifier for a cluster)
        #              This and timestamp can be used to uniquely relate to Aneesh's "Cluster Health Monitor"-database 
    }
};
my $result_type_table = {name => "result_type",
    columns => {
        id => "INTEGER PRIMARY KEY AUTO_INCREMENT",
        test_id => "INTEGER",
        name => "VARCHAR(255)",
        result_desc => "VARCHAR(255)", # 'desc' is a keyword in mysql, so this column is now called 'result_desc'
        unit => "VARCHAR(255)"
    }
};
my $sample_table = {name => "sample",
    columns => {
        id => "INTEGER PRIMARY KEY AUTO_INCREMENT",
        result_id => "INTEGER",
        result_type => "INTEGER",
        data => "VARCHAR(255)"
    }
};

#
# --- Internal Routines ---
#

# _complain is the routine used to print/log error messages. By default, it prints to stdout. 
# Point $_log_print_code_ref to a code reference, and it'll use that to substitute print().
# _log_print_code_ref = sub
my $_log_print_code_ref = undef;
sub _complain {
    my $self = {@_};
    my $level = (keys %$self)[0];
    my $msg = $self->{$level};

 
    if (defined $_log_print_code_ref) {
        $_log_print_code_ref->(level=>$level, msg=>$msg);
    }
    else {
        print $msg . "\n";
    }
}

# Setup DBI, connect to mysql, and return a database handle
sub _perflib_setup_dbi {
    my $self = {@_};
    my $dbh;
    
    my $host = $self->{host};
    my $user = $self->{user};
    my $pass = $self->{pass};
    unless ((defined $host) and (defined $user)) {
        _complain(msg=>"_perflib_setup_dbi: error: I need more args");
        return undef;
    }
    unless (defined $pass) {
        $pass = "";
    }

    eval {
        $dbh = DBI->connect("DBI:mysql:database=perflib_db:host=$host", $user, $pass); 
    }; if ($@) {
        _complain(msg=>"DBI connect failed (mysql:'$host':'$user':'$pass')");
        return undef;
    }

    unless (defined $dbh) {
        _complain(msg=>"DBH connect failed, returned undef");
        return undef;
    }

    $dbh->{PrintError} = 0;
    

    return $dbh;
}

# To insert single-quotes into sql, you need to use special formating. To insert a single-quote, you must use two single-quotes.
sub _sanitize_sql_string {
    my $input = shift;
    my $output = "";
    foreach (split(//, $input)) {
        if ($_ eq "'") { $output .= "''"; }
        else { $output .= $_; }
    }
    return $output;
}

# CREATE one of the tables defined at the top of this file (e.g. $sample_table, $test_table, ...)
sub _perflib_create_one_table {
    my $self = {@_};
    my $dbh = $self->{dbh};
    my $name = $self->{table}->{name};
    my $columns = $self->{table}->{columns};

    # Be sure to sort @column_names, since the order of keys in a hash is unpredictable/unstable
    my @column_names = sort {uc($a) cmp uc($b)} (keys %$columns);
    
    # Bulid a CREATE TABLE statement, with the columns in alphabetical order
    my $column_string = "";
    foreach (@column_names) {
        $column_string .= "$_ " . $columns->{$_} . ", ";
    }

    # Crop off the last two characters ', ' 
    $column_string = substr($column_string, 0, -2);
    
    my $do_result = $dbh->do("CREATE TABLE $name ($column_string);");
    unless (defined $do_result) {
        _complain(msg=>"_perflib_create_one_table: error: do() failed, err = ".$dbh->errstr);
    }
}

# CREATE all the tables used by PerfLib, just in case they don't already exist
sub _perflib_create_all_tables {
    my $self = {@_};
    my $dbh = $self->{dbh};
    eval {
        foreach ($sample_table, $test_table, $result_table, $run_table, $result_type_table) {
            _perflib_create_one_table(dbh=>$dbh, table=>$_);
        }
    }; if ($@) {
        _complain(msg=>"Create tables failed, $@");
    }
}

# do a simple insert on a table, given a {column_name=>'value'} hash
sub _perflib_sql_add_row {
    my $self = {@_};
    my $table = $self->{table};
    my $dbh = $self->{dbh};
    my $row = $self->{row};
    my @keys = keys %$row;

    my $columns = "";
    my $values = "";
    foreach (@keys) {
        $columns .= $_.', ';
        $values .= "'" . _sanitize_sql_string($row->{$_}) . "', ";
    }
    $values = substr($values, 0, -2);
    $columns = substr($columns, 0, -2);

    my $id = -1;
    eval {
        my $insert = "INSERT INTO " . $table->{name} . " (" . $columns . ") VALUES (" . $values . ");";
        my $do_result = $dbh->do($insert);

        if ($do_result == 1) {
            #$id = $dbh->func('last_insert_rowid'); # FIXME: in mysql, this is 'last_insert_id'
            #$id = $dbh->func('last_insert_id');
            #$id = $dbh->last_insert_id(undef, undef, ($table->{name}, "id"));
            $id = $dbh->selectrow_array("SELECT LAST_INSERT_ID()");
            #_complain(msg=>"_perflib_sql_add_row: debug: verify that this is sane: id = $id");
        }
        else {
            _complain(msg=>"_perflib_sql_add_row: error: do() returned $do_result, err = ".$dbh->errstr);
        }
    }; if ($@) {
        _complain(msg=>"insert failed, $@"); 
    }
    
    return $id;
}

# open a statement handle in DBI for a query
sub _perflib_simple_query_raw {
    my $self = {@_};
    my $dbh = $self->{dbh};
    my $sql = $self->{sql};

    my $sth;
    eval {
        $sth = $dbh->prepare($sql);
        if (defined $sth) {
            my $r = $sth->execute();
        }
    }; if ($@) {
        _complain(msg=>"couldn't execute query \n-- Query -- $sql\n-- Error -- $@");
        return undef;
    }
    return $sth;
}

# perform a simple sql query, return results as an array of rows, which each row is a hash, with columns for keys
sub _perflib_simple_query_hash {
    my $self = {@_};
    my $table_name = $self->{table}->{name};
    my $where_clause = $self->{where_clause}; # e.g. "WHERE id=5 and foo='blah'"
    my $columns = $self->{columns}; # e.g. ['id', 'some_col', ...]
    my $dbh = $self->{dbh};

    # build a string of the given columns, "id, some_col, another_col"
    my $columns_str = "";
    foreach (@$columns) {
        $columns_str .= $_ . ', ';
    }
    $columns_str = substr($columns_str, 0, -2);

    # send the query 
    my $sth = _perflib_simple_query_raw(dbh=>$dbh, sql => "SELECT $columns_str FROM $table_name $where_clause;");
    unless (defined $sth) {
        _complain(msg=>"_perflib_simple_query_hash: couldn't send raw query");
        return undef;
    }

    # parse the output
    my @output;
    while (my @row = $sth->fetchrow()) {
        my $row_hash = {};
        for (my $i=0; $i < (scalar @row); $i++) {
            $row_hash->{$columns->[$i]} = $row[$i];
        }
        push @output, $row_hash;
    }

    return \@output;
}

sub _perflib_simple_update {
    my $self = {@_};
    my $table_name = $self->{table}->{name};
    my $where_clause = $self->{where_clause};
    my $updates = $self->{updates};
    my $dbh = $self->{dbh};

    # build a string of updates, "foo='bar', blah='banana'" 
    my $updates_str = "";
    foreach (keys %$updates) {
        $updates_str .= $_ . '=' . _sanitize_sql_string($updates->{$_}) . ', ';
        # printf("updates_str = %s\n", $updates_str);
    }
    $updates_str = substr($updates_str, 0, -2);

    # send the update
    my $do_result = undef;
    eval {
        my $update = "UPDATE $table_name SET $updates_str $where_clause;";
        # printf("_perflib_simple_update: debug: update string == %s\n", $update);
        $do_result = $dbh->do($update)
    }; if ($@) {
        _complain(msg=>"Update failed, $@");
        return 0;
    }

    unless ((defined $do_result) and ($do_result > 0)) {
        _complain(msg=>"Update failed, do() returned $do_result");
        return 0;
    }
    return 1;
}

# This just makes sure that $self contains the arguments it's supposed to contain
sub _perflib_test_for_argument_existence {
    my $self = shift;
    foreach (@_) {
        unless (exists $self->{$_}) {
            return 0;
        }
    }
    return 1;
}

sub _perflib_lookup_test_id {
    my $self = {@_};
    my $dbh = $self->{dbh};
    my $name = $self->{name};
    my $rows = _perflib_simple_query_hash(
        dbh => $dbh, 
        table => $test_table, 
        columns => ['name', 'id'], 
        where_clause => "WHERE name='$name'");

    if ( (defined $rows) and ((scalar @$rows)==1) ) {
        return $rows->[0]->{id};
    }
    return -1;
}

sub _perflib_lookup_result_types {
    my $self = {@_};
    my $dbh = $self->{dbh};
    my $test_id = $self->{test_id};
    my $rows = _perflib_simple_query_hash(dbh=>$dbh, table => $result_type_table, 
        columns => ['id', 'name', 'result_desc', 'unit'], where_clause => "WHERE test_id=$test_id");
    
    my %hash;
    foreach (@$rows) {
        $hash{$_->{name}} = {id=>$_->{id}, result_desc=>$_->{result_desc}, unit=>$_->{unit}};
    }

    return \%hash;
}
    
sub _perflib_create_base_object {
    my $self = {@_};

    my $dbh = _perflib_setup_dbi(host=>$self->{dbi_host}, user=>$self->{dbi_user}, pass=>$self->{dbi_pass});
    unless (defined $dbh) {
        _complain(msg=>"_perflib_create_base_object: error: _perflib_setup_dbi failed");
        return undef;
    }
     
    my $obj = {
        name => $self->{name},
        test_id => undef,
        short_desc => undef,
        long_desc => undef,
        result_types => undef,
        run_id => undef,    
        pid => $$,
        dbh => $dbh,
        dbi_host => $self->{dbi_host},
        dbi_user => $self->{dbi_user},
        dbi_pass => $self->{dbi_pass}
    };
    bless($obj);
    return $obj;
}

sub _perflib_check_dbh {
    my $self = {@_};
    my $obj = $self->{obj};

    if ($obj->{pid} != $$) {
        _complain(msg=>"_perflib_check_dbh: debug: my PID changed (maybe I was subtested). Restarting DBI...");
        $obj->{dbh} = _perflib_setup_dbi(host=>$obj->{dbi_host}, user=>$obj->{dbi_user}, pass=>$obj->{dbi_pass});
        unless (defined $obj->{dbh}) {
            _complain(msg=>"_perflib_check_dbh: error: _perflib_setup_dbi failed");
            return 0;
        }
    }

    return 1;
}

#
# --- External API ---
# 
    
#
# This is a constructor for PerfLib. It attempts to lookup a *type of test* that already exists in the database. The test is uniquely identified by its name.
# If the test can't be found, this will return undef. 
#
# $test_obj = LookupTest(name=>"my_test_name")
#
sub LookupTest {
    my $self = {@_};
        $_log_print_code_ref = $self->{_log_print_code_ref}; # configure the path for stderr/log output
    my $name = $self->{name};

    my $obj = _perflib_create_base_object(name=>$name, dbi_host=>"rtpgx-tools1.rtp.netapp.com", dbi_user=>"sqluser"); # no password
    unless (defined $obj) {
        _complain(msg=>"LookupTest: error: _perflib_create_base_object failed");
        return undef;
    }

    my $test_id = _perflib_lookup_test_id(dbh=>$obj->{dbh}, name=>$name);
    if ($test_id == -1) {
        return undef;
    }

    my $rows = _perflib_simple_query_hash(dbh=>$obj->{dbh}, table => $test_table, columns => ['short_desc', 'long_desc'], where_clause => "WHERE test_id=$test_id");
    unless (defined $rows) {
        _complain(msg=>"LookupTest: error: query test_table failed. Probably a database problem!");
        return undef;
    }
    
    my $result_types = _perflib_lookup_result_types(dbh=>$obj->{dbh}, test_id=>$test_id); 
    
    $obj->{test_id} = $test_id;
    $obj->{short_desc} = $rows->[0]->{short_desc};
    $obj->{long_desc} = $rows->[0]->{long_desc};
    $obj->{result_types} = $result_types;
    $obj->{_log_print_code_ref} = $self->{_log_print_code_ref};
    return $obj;
}

#
# This is a constructor for PerfLib. It attempts to create a new type of *Test* in the database. The name must be unique. If a test with that name already exists,
# then we'll return undef. If you want to instantiate PerfLib for a test that already exists, use PerfLib::LookupTest().
#
# $test_obj = PerfLib::CreateTest(
#                      name=>"worst_case_foobar_test", 
#                      short_desc=>"Measure the worst-case foobar performance.",
#                      long_desc=>"This test will setup a cluster with FOOBAR_AFFINITY set to 'high', and will iterate over nodes using the round-foobar scheduling algortihm...
#
sub CreateTest {
    my $self = {@_};
        $_log_print_code_ref = $self->{_log_print_code_ref}; # configure the path for stderr/log output
    unless (_perflib_test_for_argument_existence($self, 'name', 'long_desc', 'short_desc')) {
        _complain(msg=>"CreateTest needs more arguments");
        return undef;
    }

    my $name = $self->{name};
    my $long_desc = $self->{long_desc};
    my $short_desc = $self->{short_desc};

    # create/bless an object with basic data filled in
    my $obj = _perflib_create_base_object(name=>$name, dbi_host=>"rtpgx-tools1.rtp.netapp.com", dbi_user=>"sqluser"); # no password
    unless (defined $obj) {
        _complain(msg=>"CreateTest: error: _perflib_create_base_object failed");
        return undef;
    }

    # in case we're working on a new database, try to create all the tables
    _perflib_create_all_tables(dbh=>$obj->{dbh});

    # if a test with this name already exists, fail
    unless (_perflib_lookup_test_id(dbh=>$obj->{dbh}, name=>$name) == -1) {
        _complain(msg=>"Could not create test '$name' because it already exists");
        return undef;
    }

    my $test_id = _perflib_sql_add_row(table=>$test_table, dbh=>$obj->{dbh}, row=>{name=>$name, long_desc=>$long_desc, short_desc=>$short_desc});             
    if ($test_id == -1) {
        _complain(msg=>"CreateTest: error: _perflib_sql_add_row failed");
        return undef;
    }

    # my $obj = {
    #     name=>$name,             # The name of the test (which is itself a unique identifier)
    #     test_id=>$test_id,       # the primary key of this test in the database
    #     short_desc=>$short_desc, # a brief (maybe single-sentence) description of what this test does, (what it measures).
    #     long_desc=>$long_desc,   # a longer, more detailed description of what this test does (how it measures it, why it measures it).
    #     result_types=>{},        # a hash-of-hashes describing the *types* of results this test measures (kb/s, meters/fortnight, midichlorians)
    #     dbh=>$dbh,               # (private) our dbi connection / database handle
    #     run_id=>undef            # (private) the primary key of the current run
    # };

    $obj->{result_types} = {}; # this is a new test, so there are no result_types
    $obj->{test_id} = $test_id;
    $obj->{name} = $name;
    $obj->{long_desc} = $long_desc;
    $obj->{short_desc} = $short_desc;
    $obj->{_log_print_code_ref} = $self->{_log_print_code_ref};
    return $obj;
}    
           
#
# Add a new *type* of result to the given test. An example of a *type* of result is "write_latency". This might measure, for example,
# the length of time to complete a single write operation. It's a unit of time, measured in microseconds. 
#
# $test_obj->CreateResultType(  name => "write_latency",
#                               desc => "This is the latency of a single write operation, as measured by the filer.",
#                               unit => "microseconds");
#
sub CreateResultType {
    my $obj = shift;
        $_log_print_code_ref = $obj->{_log_print_code_ref}; # configure the path for stderr/log output
    my $self = {@_};
    unless (_perflib_test_for_argument_existence($self, 'name', 'desc', 'unit')) {
        _complain(msg => "CreateResultType needs more arguments");
        return 0;
    }

    # check whether the DBI handle is still valid (and then fix it)
    unless (_perflib_check_dbh(obj=>$obj)) {
        _complain(msg => "CreateResultType: error: _perflib_check_dbh failed");
        return 0;
    }

    my $name = $self->{name};
    my $desc = $self->{desc};
    my $unit = $self->{unit};

    # check whether a result_type with this name already exists
    if (defined $obj->{result_types}->{$name}) {
        _complain(msg=>"CreateResultType: type '$name' already exists for this test");
        return 0;
    }

    # insert a new result_type into the db
    my $result_type_id = _perflib_sql_add_row(dbh=>$obj->{dbh}, table=>$result_type_table, 
        row=>{name=>$name, result_desc=>$desc, unit=>$unit, test_id=>$obj->{test_id}});

    if ($result_type_id == -1) {
        _complain(msg => "CreateResultType failed");
        return 0;
    }  
  
    # push the new result type into our object
    $obj->{result_types}->{$name} = {result_desc=>$desc, unit=>$unit, id=>$result_type_id};
    return 1;
}     
    
#
# This is a method that begins a new *run*. It creates a new run-entry in the database, and sets the state of test_obj to 'running'. 
# After calling this, your test is presumed to be measuring data. 
# When the run has completed, call EndRun()
#
# $test_obj->StartRun()
#
sub StartRun {
    my $obj = shift;
        $_log_print_code_ref = $obj->{_log_print_code_ref}; # configure the path for stderr/log output
    my $self = {@_};

    if (defined $obj->{run_id}) {
        _complain(msg=>"StartRun: error: this test is already running");
        return 0;
    }

    # check whether the DBI handle is still valid (and then fix it)
    unless (_perflib_check_dbh(obj=>$obj)) {
        _complain(msg => "StartRun: error: _perflib_check_dbh failed");
        return 0;
    }

    my $test_id = $obj->{test_id};
    my $run_id = $obj->{run_id};
    my $dbh = $obj->{dbh};
    my $testcase = (defined $self->{testcase}) ? $self->{testcase} : " "; # optional
    
    my $run_id = _perflib_sql_add_row(table=>$run_table, dbh=>$dbh, row=>{test_id=>$test_id, active=>1, testcase=>$testcase});
    if ($run_id == -1) {
        _complain(msg=>"StartRun: error: couldn't insert into run table");
        return 0;
    }
    $obj->{run_id} = $run_id;

    return 1;
}

#
# Call this when your run is completed, and you are no longer measuring data. 
# After calling this, you may not add new results.
# After calling this, you may call StartRun() to begin a new run.
#
# test_obj->EndRun();
#
sub EndRun {
    my $obj = shift;
        $_log_print_code_ref = $obj->{_log_print_code_ref}; # configure the path for stderr/log output
    my $self = {@_};
    my $test_id = $obj->{test_id};
    my $run_id = $obj->{run_id};

    $obj->{run_id} = undef; # if !defined $test->{run_id}, then the test isn't running
    
    # check whether the DBI handle is still valid (and then fix it)
    unless (_perflib_check_dbh(obj=>$obj)) {
        _complain(msg => "EndRun: error: _perflib_check_dbh failed");
        return 0;
    }

    unless (defined $run_id) {
        _complain(msg=>"EndRun: error: this test isn't currently running");
        return 0;
    }    

    unless (_perflib_simple_update(dbh => $obj->{dbh}, table => $run_table, 
        where_clause => "WHERE id=$run_id",
        updates => {active=>0})) {
            _complain(msg=>"EndRun: error: couldn't update our run in the database");
            return 0;
    }

    return 1;
}

#
# This will add a result into the database. Provide a list of key=>value pairs, in the form result_type=>"result value".
# The test must be running. 
# If you want to insert this result as part of a subtest (from TCD for example), use subtest=>"subtest name".
# 
# test_obj->AddResult(subtest=>"my_subtest_01", 
#   samples=>{throughput=>"10", most_active_node=>"rtp-3070-50", least_active_node=>"rtp-3070-12", other_result=>"other value"})
#
sub AddResult {
    my $obj = shift;
        $_log_print_code_ref = $obj->{_log_print_code_ref}; # configure the path for stderr/log output
    my $self = {@_};
    my $samples = $self->{samples};
    
    # check whether the DBI handle is still valid (and then fix it)
    unless (_perflib_check_dbh(obj=>$obj)) {
        _complain(msg => "AddResult: error: _perflib_check_dbh failed");
        return 0;
    }

    my $run_id = $obj->{run_id};
    unless (defined $run_id) {
        _complain(msg=>"AddResult: error: This test is not running.");
        return 0;
    }

    # confirm that the specified result_types exist
    foreach (keys %$samples) {
        unless (exists $obj->{result_types}->{$_}) {
            _complain(msg=>"AddResult: error: result type $_ doesn't exist");
            return 0;
        }
    }

    # add the result in to the db
    my $result_id = _perflib_sql_add_row(table => $result_table, dbh => $obj->{dbh}, row => {
        sub_test => ((exists $self->{subtest})?$self->{subtest}:""), 
        samples => scalar (keys %$samples),
        run_id => $run_id
    });
    if ($result_id == -1) {
        _complain(msg=>"AddResult: error: couldn't insert row into result table");
        return 0;
    }

    # add the specified samples into the db
    foreach my $key (keys %$samples) {
        my $value = $samples->{$key};
        
        my $sample_id = _perflib_sql_add_row(table => $sample_table, dbh => $obj->{dbh}, row => {
            result_type => $obj->{result_types}->{$key}->{id}, 
            result_id => $result_id,
            data => $value
        });
        if ($sample_id == -1) {
            _complain(msg=>"AddResult: error: couldn't insert row into sample table");
            return 0;
        }
    }

    return 1;
}


#
# RunQuery: This is a temporary (or permanent?) method to let you directly query the database.
# We should probably have better ways of querying data out of the database in the future.
#
# e.g.
# $sth = $perf_obj->RunQuery(query=>"SELECT name FROM test WHERE MATCH(long_desc) AGAINST ('CIFS');");
# ... to find the name of a test whose long-description contains the word CIFS
#
# we'll just return the statement handle for now
#
sub RunQuery {
    my $self = shift;
    my $args = {@_};
        $_log_print_code_ref = $self->{_log_print_code_ref}; # configure the path for stderr/log output
    
    # check whether the DBI handle is still valid (and then fix it)
    unless (_perflib_check_dbh(obj=>$self)) {
        _complain(msg => "AddResult: error: _perflib_check_dbh failed");
        return 0;
    }

    my $query = $self->{query};
    unless (defined $query) {
        _complain(error=>"PerfLib::RunQuery: error: need more arguments");
        return undef;
    } 
    
    my $sth = _perflib_simple_query_raw(dbh=>$self->dbh, sql=>$query);
    return $sth;
}

1;
