###############################################################################
# $Id: //depot/prod/test/Rsierranevada/lib/PerfLib/TCD.pm#1 $
#
# Copyright (c) 2007-2009 NetApp, Inc.
# All rights reserved.
###############################################################################
package PerfLib::TCD;

use strict;
use lib ".";
use PerfLib::PerfLib;
use TCD;
use Data::Dumper;
use TracePkg;
use Time::HiRes;

our @ISA = ("TCD");

# 
# - Overridden Method -
#
# Override TCD::new(). Instantiate a PerfLib object and either create or lookup an existing test using the name
# of the script (e.g. 'foobar_test' for foobar_test.thpl) as the PerfLib-test name.
#
# PerfLib::TCD::new() supports new optional arguments,
#   -short_desc => a brief (maybe one sentence) description of what this test does
#   -long_desc  => a longer, more detailed description of what this test does (how/why it does it)
#   -perflib_test_name => the name PerfLib will use to uniquely identify this test. If not provided, it'll use the name of the script
#
sub new {
    my $proto = shift;
    my @raw_args = @_;
    my $class = ref($proto) || $proto;
    my $self  = $class->SUPER::new(@raw_args);

    # parse our custom, optional arguments
    my $args = {@raw_args};
    my $long_desc = (exists $args->{'-perflib_long_desc'}) ? $args->{'-perflib_long_desc'} : " " ;
    my $short_desc = (exists $args->{'-perflib_short_desc'}) ? $args->{'-perflib_short_desc'} : " " ;
    my $testname = (exists $args->{'-perflib_test_name'}) ? $args->{'-perflib_test_name'} : "TCD_" . $self->{testscript};

    # create a code reference to pass PerfLib errors through $self->nlog()
    # FIXME: I think storing this coderef in $self is self-referential, and it may need its reference count decremented
    my $logging_ref = sub {my $args = {@_}; $self->nlog("PerfLib:".$args->{level}.": ".$args->{msg});};

    # $self->nlog("Attention! The currently running script is " . $self->{testscript});
    # $self->nlog(Dumper($self));
    my $perf_obj = PerfLib::LookupTest(name=>$testname, _log_print_code_ref=>$logging_ref);
    unless (defined $perf_obj) {
        # If it doesn't already exist, then create it
        $perf_obj = PerfLib::CreateTest(name => $testname, 
                                        short_desc => $short_desc,
                                        long_desc => $long_desc,
                                        _log_print_code_ref => $logging_ref);
        unless (defined $perf_obj) {
            die "PerfLib::TCD::new(): error: PerfLib::CreateTest failed";
        }
    }

    $self->nlog(Dumper($perf_obj));

    $self->{perf_obj} = $perf_obj;

    bless ($self, $class); 
    return $self;
}

#
# A TCD script calls run_test() to let TCD take over control of execution. 
# ... uh, don't do anything here yet
#
# sub run_test {
#     my $self = shift;
#     my @raw_args = @_;
# 
#     $self->nlog("run_test: DEBUG: I'm called");
#     my $result = $self->SUPER::run_test(@raw_args);
#     # my $self = $class->SUPER::new(@raw_args);
#     $self->nlog("run_test: DEBUG: I'm done");
# 
#     return $result;
# }

sub _exists_in_array {
    foreach (@{$_[1]}) {if ($_[0] eq $_) {return 1;} }
    return 0;
}

#
# - Overridden Method -
# TCD calls _run_method() to begin a testcase (and also init(), setup(), cleanup(), and uninit())
# We'll use this section to demarcate the beginning and end of a run in PerfLib
#
sub _run_method {
    my $self = shift;
    my @raw_args = @_;
    my $method_name = $raw_args[0];
    
    # If this isn't a *real* testcase, and it's just one of the init/release functions,
    # then do nothing. Just pass the call to SUPER.
    if (_exists_in_array($method_name, ['init', 'setup', 'cleanup', 'uninit'])) {
        return $self->SUPER::_run_method(@raw_args);
    }

    # First find the corresponding 'testcase' to this method name in $self->{testcases}
    # (TCD allows testcases and their corresponding methods to have differing case, for some reason)
    my $testcase = undef;
    foreach (keys %{$self->{testcases}}) {
        if ( lc($_) eq lc($method_name) ) {
            $testcase = $_;
            last;
        }
    }
    unless (defined $testcase) {
        die "PerfLib::TCD::_run_method: error: no testcase matches my method's name ($method_name)";
    }
    $self->nlog("PerfLib::TCD::_run_method: debug: starting a run for $testcase (".$self->{testcases}->{$testcase}.")");


    # Start a run for this testcase
    unless ($self->{perf_obj}->StartRun(testcase=>$testcase)) {
        die "PerfLib::TCD::_run_method: error: PerfLib::StartRun failed";
    }

    # Let SUPER begin the testcase
    my $result = $self->SUPER::_run_method(@raw_args);

    # The testcase returns, now call PerfLib::EndRun()
    unless ($self->{perf_obj}->EndRun()) {
        die "PerfLib::TCD::_run_method: error: PerfLib:EndRun failed";
    }
    
    return $result;
}

#
# This will add result *types* to the running test. These types are stored permanently in the database.
# Provide a list of hashes with the form:
#   add_result_types( {name=>"throughput", unit=>"KB/s", desc=>"The average throughput on the foobar interface."}, ...)
# The number of new inserted results will be returned. Or if there's an error, we'll throw an exception. 
#
sub create_result_types {
    my $self = shift;
    my $perf_obj = $self->{perf_obj};

    my $inserted = 0;
    foreach my $hash (@_) {
        # sanity check input
        unless ( (exists $hash->{unit}) and (exists $hash->{name}) and (exists $hash->{desc}) ) {
            die "PerfLib::TCD::create_result_types: error: bad formating for arguments - each type's hash needs 'unit', 'name', and 'desc' defined";
        }

        # lookup name - if the result type DOES exist, verify that the details match
        if (exists $perf_obj->{result_types}->{$hash->{name}}) {
            my $existing = $perf_obj->{result_types}->{$hash->{name}};
            
            # if the details *don't* match, then log a warning
            if ($existing->{unit} ne $hash->{unit}) {
                $self->nlog("PerfLib::TCD::create_result_types: WARNING: the result type '" . $hash->{name} ."' already exists and has a different unit ('".
                    $existing->{unit} . "' vs '" . $hash->{unit} . "'");
            }
            if ($existing->{result_desc} ne $hash->{desc}) {
                $self->nlog("PerfLib::TCD::create_result_types: WARNING: the result type '" . $hash->{name} ."' already exists and has a different description ('".
                    $existing->{result_desc} . "' vs '" . $hash->{desc} . "'");
            }
            # $self->nlog("TCD_PerfLib::create_result_types: debug: ".Dumper($existing));
        }
        # else, if the result type doesn't exist, then create it 
        else {
            unless ($perf_obj->CreateResultType(name=>$hash->{name}, desc=>$hash->{desc}, unit=>$hash->{unit})) {
                die "PerfLib::TCD::create_result_types: error: CreateResultType failed!";
            }
            $inserted++;
        }
    }

    return $inserted;
}
        
#
# This will add a result (a set of samples that are described by a result_type) into the database.
# Provide a hash of samples, e.g.
#   add_result( samples => {throughput=>"40", most_active_node=>"rtpgx-3050-19", all_nodes_are_healthy=>"true"} )
# Where 'throughput', 'most_active_node', and 'all_nodes_are_healthy' are the names of result_types
#
# FIXME: fix subtest thing
#
sub add_result {
    my $self = shift;
    my $args = {@_};
    my $perf_obj = $self->{perf_obj};
    my $samples = $args->{samples};

    # sanity check arguments
    unless (defined $samples) {
        die "PerfLib::TCD::add_result: error: missing samples=>{} argument"; 
    }

    # sanity check test state (we must be "running")
    unless (defined $perf_obj->{run_id}) {
        $self->nlog("PerfLib::TCD::add_result: error: add_result() can only be called when the test is running. (It can only be called within a testcase.)");
        return 0;
    }

    # ship the samples off to PerfLib
    return $perf_obj->AddResult(subtest=>"FIXME", samples=>$samples);
} 

#
# FIXME: add documentation
# FIXME: implement me!
# I believe I will need either (node and if) or (ip)
#
sub start_nfs_trace {
    my ($self, %args) = @_;

    my $data_if = $args{data_if};
    my $data_ip = $args{data_ip};
    my $node = $args{node};

    if ((defined $data_if) and (defined $node) and !(defined $data_ip)) {
        # do some logic
    }
    elsif ((defined $data_ip) and !(defined $node) and !(defined $data_if)) {
        # do some logic
    }
    else {
        #$Trace->fatal(msg=>"provide start_nfs_trace *either* (data_ip) or (node and data_if)");
    }
    
    my $nfs_trace_obj = PerfLib::NfsTrace::new(Server => $self->get_server(), node => $node);

}
    
#
# This starts a a "HiRes" timer, and returns a timer object. Provide a result_type so that, when you stop the timer,
# it can automatically insert the measurement into the database.
# FIXME: the result_type must have a "unit" of seconds (but we don't sanity check that)
#        make this more robust.
#
# $timer_obj = $Test->start_timer(result_type=>"some_macro_time_measurement");
# ...
# $timer_obj->end();
#    
sub start_timer {
    my ($self, %args) = @_;

    $self->nlog("start_timer: debug: result_type = " . $args{result_type});

    return perflib_timer::_new(test=>$self, result_type=>$args{result_type});
}


#
# define a new object type for our built-in timer, so that start_timer() can return an object that you can call end() on
#
package perflib_timer;

#
# private constructor for perflib_timer
#
sub _new {
    my $self = {@_};
    my $obj = {};

    $obj->{start} = [ Time::HiRes::gettimeofday() ];
    $obj->{test} = $self->{test};
    $obj->{result_type} = $self->{result_type};
    bless($obj);
    return $obj;
}
    
#
# call $timer_obj->end() to stop the timer and insert the measured length of time into the database, 
# under the result_type given when you called PerfLib::TCD::start_timer()
# returns the elapsed time as a real in seconds
sub end {
    my $self = shift;
    my $delta_t = Time::HiRes::tv_interval($self->{start});
    my $test = $self->{test};

    unless ($test->add_result(samples=>{$self->{result_type} => $delta_t})) {
        die "PerfLib::TCD: perflib_timer::end(): error: add_result() failed";
    }
    
    #FIXME: don't let users call ->end() twice. Deallocate this object, or something. I dunno. Whatever.

    return $delta_t;
}
