# Copyright (c) 2006 Network Appliance, Inc.
# All rights reserved.

# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Platform/BmcUtils.pm#1 $

## @summary BMC command library
## 
## @description
##    This package defines routines which execute commands on a BMC
##    using the "say" command in Tharn.  These routines were developed
##    on a Shuar filer running GB.1.
##
##    All the subroutines here expect a filer "#console" connection as their
##    first argument.  
##
##    The routines are:
##      
##       phash       - print out contents of hash
##       parray      - print out contents of array
##       login2ontap - supplies the user name and password in an attempt
##                     to login to ontap.  
##
##       bmc_head_power_cycle - Power cycles the filer head using BMC's
##                              "system power cycle" command.
##
##       bmc_head_reset - Resets the filer head using BMC's "system reset" command.
##
##       bmc_command - Handles processing for common BMC commands.
##
##       get_bmc_event_info - Gets number of events, free bytes, etc
##       initialize_sel     - Makes sure there are at least N free entries in
##                            the SEL, if not clears SEL.  Returns current number of 
##                            entries.
##       find_sel_entries   - Returns array of found entries, optionally errors if
##                            entry is found or not found, depending on arguments
##
##       reboot_bmc         - reboot the BMC, via Ontap
##       ssh2bmc            - SSH into the BMC from $CLIENT
##
##       change_ontap_password - Change current Ontap password
##
##       get_bmc_ip_from_bmc   - Gets IP info from BMC
##       get_bmc_ip_from_ontap - Gets IP info from ontap
##       get_system_info_from_bmc - Gets system info from BMC
##       checkBmcConLogForPanic - Dumps the console logs ("system log" command from BMC) from the BMC and checks for the presence of panics. 
##       checkBmcConLogForCECC  - Dumps the console logs ("system log" command from BMC) from the BMC and checks for the presence of CECC.
##      
## @dependencies 
## The functions depend on exact output format to parse and extract results.
## Even a minor change in the UI can result in these functions not working properly.
##      
## @bugs
##      

package Platform::BmcUtils;
require Exporter;

our @ISA        = qw(Exporter);
our @EXPORT     = qw( phash parray login2ontap bmc_head_power_cycle bmc_head_reset bmc_command get_bmc_ip get_bmc_event_info initialize_sel find_sel_entries reboot_bmc ssh2bmc change_ontap_password get_bmc_ip_from_bmc get_bmc_ip_from_ontap get_system_info_from_bmc checkBmcConLogForPanic checkBmcConLogForCECC);
our @EXPORT_OK  = qw();
our $VERSION    = 1.00;

use strict;
use Tharn;
use NATE::Result;
use boot;
# Needed for $EVAL_ERROR
use English;

my $MAX_SEL_ENTRIES = 512;

# remove prior to submittal
sub phash {
  my %hash = @_;
  my @keys = keys(%hash);
  foreach my $i (@keys) { 
    print "  hash element $i = |$hash{$i}|\n"; 
  }
}

sub parray {
  my @params = @_;
  foreach my $i (@params) { print "  array element |$i|\n"; }
}

## Options are: -timeout in seconds
##
sub login2ontap {
  my ($fc, %options) = @_;

  my $fh = $fc->hostp();
  my $runlength = ($options{-timeout}) ? $options{-timeout} : 600;
  my $bootOntap = ($options{-bootOntap}) ? $options{-bootOntap} : "yes";

  # get username and password from $fc
  my $username = ($fh->{username}) ? $fh->{username} : "root";
  my $dpassword = ($fh->{password}) ? $fh->{password} : $fh->{default_password};

  my $password = ($options{-password}) ? $options{-password} : $dpassword;

  my $begtime = time;

  my $finished=0;
  my $firmware=0;
  while ( ((time-$begtime) < $runlength) && ($finished == 0) ) {

    my $state = boot_state( $fc, transient=>1 );

#print "\nSTATE = $state \n\n";

    if ( $state =~ /USERNAME/ ) {
      say( $fc, "$username", -blind, -swallow, -timeout, 300 );
      say( $fc, "$password", -blind, -swallow, -timeout, 300 );
    } elsif ( $state =~ /PASSWORD/ ) {
      say( $fc, "$password", -blind, -swallow, -timeout, 300 );
    } elsif ( $state =~ /FIRMWARE/ ) {

      if ( $bootOntap =~ m/yes/i ) {
        if ( $firmware > 2 ) {
          say( $fc, "boot_ontap", -blind, -swallow, -timeout, 300 );
        }
      }
      $firmware++;

    } elsif ( $state =~ /BOOT_MENU/ ) {
      if ( $bootOntap =~ m/yes/i ) {
        # this should be for normal bootup
        say( $fc, "1", -blind, -swallow, -timeout, 300 );
      }
    } elsif ( $state =~ /UP/ ) {
      $finished = 1;
    }

    sleep 10;
  }

  if ( $finished == 0 ) {
     tstlog(-fatal, "ERROR: Could not log into Ontap"); 
  }
}

## Options are: -login <yes/no> - login to ontap after power cycling the head?
##
sub bmc_head_power_cycle {
  my ($fc, %options) = @_;
  my $login = ($options{-login}) ? $options{-login} : "yes";
  
  bmc_command($fc, "system power cycle", -login, $login);
}

## Options are: -login <yes/no> - login to ontap after power cycling the head?
##
sub bmc_head_reset {
  my ($fc, %options) = @_;
  my $login = ($options{-login}) ? $options{-login} : "yes";

  bmc_command($fc, "system reset", -login, $login);
}

## Options are: -login <yes/no> - login to ontap after power cycling the head?
## The BMC oem shell is entered at the start of this procedure and
## is exited at the conclusion of this call.
##
sub bmc_command {
  my ($fc, $cmd, %options) = @_;
  my $BMC_CTRL_SEQ = "\cg";
  my @BMC_PRMT = ( -noprompt, -prompt, '^bmc.*-> ' );
  my $login = ($options{-login}) ? $options{-login} : "yes";

  say( $fc, "exit\r\r", -blind, -swallow );

  my $res="";
  sleep 1;
  say($fc, $BMC_CTRL_SEQ, -nocr, @BMC_PRMT, -swallow );
  say($fc, "oem", @BMC_PRMT, -swallow );
  my ($one, $two, $three, $four) = say($fc, "\r", @BMC_PRMT, -swallow );

  if ( $cmd =~ /system power cycle/ ) {
    # the ?i: is for ignore case
    say($fc, $cmd, "Continue\\? (?i:\\[y\\/N\\])", -noprompt, -swallow );
    say($fc, "y", @BMC_PRMT, -swallow );
    say($fc, "exit", -blind, -swallow );
    if ( $login =~ m/yes/i ) {
      login2ontap( $fc );
    }
  } elsif ( $cmd =~ /system reset/ ) {
    # the ?i: is for ignore case
    say($fc, "priv set advanced", @BMC_PRMT, "-timeout", 120, -swallow );
    say($fc, $cmd, "Continue\\? (?i:\\[y\\/N\\])", -noprompt, -swallow );
    say($fc, "y", @BMC_PRMT, "-timeout", 300, -swallow );
    say($fc, "exit", -blind, -swallow );
    if ( $login =~ m/yes/i ) {
      login2ontap( $fc );
    }
  } elsif ( $cmd =~ /system core/ ) {
    say($fc, "system core", "Continue\\? (?i:\\[y\\/N\\])", -noprompt, -swallow );
    say($fc, "y", @BMC_PRMT, "-timeout", 300, -swallow );
    say($fc, "exit", -blind, -swallow );
    if ( $login =~ m/yes/i ) {
      login2ontap( $fc );
    }
  } elsif ( $cmd =~ /system power off/ ) {
    say($fc, "system power off", "Continue\\? (?i:\\[y\\/N\\])", -noprompt, -swallow );
    say($fc, "y", @BMC_PRMT, "-timeout", 300, -swallow );
  } elsif ( $cmd =~ /system power on/ ) {
    say($fc, "system power on",  @BMC_PRMT, -swallow );
    say($fc, "exit", -blind, -swallow );
    if ( $login =~ m/yes/i ) {
      login2ontap( $fc );
    }
  } elsif ( $cmd =~ /panic/ ) {
   say($fc, "priv set diag; ngs smash; set panic=1", "-wait", "\(Soft reset by BMC self-repair|Hard reset by main CPU\)", -noprompt, -swallow, "-timeout", 300 );
   # Even after all of the messages, console still doesn't come back for a few seconds
   sleep 10;
  } elsif ( $cmd =~ /wd_reset/ ) {
   say($fc, "priv set test; watchdog wd_reset", "-wait", "\(Soft reset by BMC self-repair|Hard reset by main CPU\)", -noprompt, -swallow, "-timeout", 300 );
   # Even after all of the messages, console still doesn't come back for a few seconds
   sleep 10;
  } elsif ( $cmd =~ /hang a/ ) {
   say($fc, "priv set test; watchdog hang a", "-wait", "\(Soft reset by BMC self-repair|Hard reset by main CPU\)", -noprompt, -swallow, "-timeout", 300 );
   # Even after all of the messages, console still doesn't come back for a few seconds
   sleep 10;
  } elsif ( $cmd =~ /hang b/ ) {
   say($fc, "priv set test; watchdog hang b", "-wait", "\(Soft reset by BMC self-repair|Hard reset by main CPU\)", -noprompt, -swallow, "-timeout", 300 );
   # Even after all of the messages, console still doesn't come back for a few seconds
   sleep 10;
  }  else {
    (undef, undef, $res, undef) = say($fc, $cmd, @BMC_PRMT, "-timeout", 300, -swallow );
    say( $fc, "exit\r", -blind, -swallow );
  }
  say( $fc, "exit\r", -blind, -swallow );
  return $res;
}


sub get_bmc_event_info {
  my ($fc, %options) = @_;

  # anonymous hash
  my $info_hash = {};

  my $out = bmc_command($fc, "events info");

  $out =~ s/^(\s\n)*//;
  my @lines = split ( /\n/, $out );
  chomp(@lines);
  my $buff;

  foreach my $line (@lines) {
    (undef, $buff) = split ( /:/, $line, 2 );
    $buff =~ s/^\s*//;
    $buff =~ s/\s*$//;
    if ( $line =~ /version/ ) {
      $info_hash->{version} = $buff;
    } elsif ( $line =~ /Total SEL entries/ ) {
      $info_hash->{numEntries} = $buff;
    } elsif ( $line =~ /Free Bytes/ ) {
      $info_hash->{freeBytes} = $buff;
    } elsif ( $line =~ /Recent Addition Time/ ) {
      $info_hash->{lastAdded} = $buff;
    }
  }

#  phash ( %$info_hash );
  return $info_hash;
}

sub clear_sel_if_necessary { 
  my ($fc, %options) = @_;
  my $needed = ($options{-numFree}) ? $options{-numFree} : 100;

  my $hash = get_bmc_event_info( $fc );
  my $free = $MAX_SEL_ENTRIES - $hash->{numEntries};

  if ( $free <  $needed ) {
    bmc_command( $fc, "priv set advanced; events clear" );
  }
}

sub initialize_sel { 
  my ($fc, %options) = @_;
  my $needed = ($options{-numFree}) ? $options{-numFree} : 100;

  clear_sel_if_necessary( @_ );
  my $numEntries = (get_bmc_event_info( $fc ))->{numEntries};

  return $numEntries;
}

sub numerically { $a <=> $b }

sub find_sel_entries {
  my ($fc, %options) = @_;
  my $start = ($options{-start}) ? $options{-start} : 0;
  my $id = ($options{-id}) ? $options{-id} : "";
  my $index = ($options{-index}) ? $options{-index} : "";
  my $sensor = ($options{-sensor}) ? $options{-sensor} : "";
  my $descr = ($options{-description}) ? $options{-description} : "";
  my @permissible = ($options{-permissible}) ? $options{-permissible} : "";
  my $fail_if_absent = ($options{-fail_if_absent}) ? $options{-fail_if_absent} : "no";
  my $fail_if_present = ($options{-fail_if_present}) ? $options{-fail_if_present} : "no";

  my $numEntries = (get_bmc_event_info( $fc ))->{numEntries};
  my $end = ($options{-end}) ? $options{-end} : (get_bmc_event_info( $fc ))->{numEntries};

  my $num = $numEntries - $start + 1;
  my $out;

  if ( $id ne "" ) {
    $out = bmc_command($fc, "events search Id=$id");
  } else {
    $out = bmc_command($fc, "events latest $num");
  }

  $out =~ s/^(\s\n)*//;
  my @lines = split ( /\n/, $out );
  chomp(@lines);
  @lines = (sort numerically @lines);

  @lines = grep( /^\d/, @lines );
 
  my @matches;
  my $ind;

  foreach my $line (@lines) {
    if ( $line =~ /$sensor/ ) {
      if ( $line =~ /$descr/ ) {
        my @lines = split ( /\n/, $out );
        ($ind, undef) = split ( / /, $line, 2 );

        if ( ($ind > $start) && ( $ind <= $end) ) {
          push(@matches, $line);        
        }
      }
    }
  }

  if ( $fail_if_absent =~ m/yes/i ) {
    if ( !@matches ) {
      tstlog(-fail, "ERROR: Did not find $id $sensor $descr"); 
    }
  }

  if ( $fail_if_present =~ m/yes/i ) {
    if ( @matches ) {
      tstlog(-fail, "ERROR: Should not have found $id $sensor $descr"); 
    }
  }
  return @matches;
}

sub reboot_bmc {
  my ($fc, %options) = @_;
  my $timeout = ($options{-timeout}) ? $options{-timeout} : 90;
  my $method = ($options{-method}) ? $options{-method} : "ontap";

  # eventually add support for rebooting via bootloader
  say($fc, "bmc reboot", -noprompt, -prompt, "The BMC rebooted successfully", -timeout, $timeout, -swallow );
}

sub ssh2bmc {
  my ($bc, %options) = @_;
  my $password = ($options{-password}) ? $options{-password} : "\n";
  my $connhost = ($options{-connhost}) ? $options{-connhost} : "";

  my $retry=3;
  my $status=0;

  $bc->{password} = $password;
  if ( $connhost ne "" ) {
    $bc->{connhost} = $connhost;
  }

  while ( $retry > 0 ) {
    eval{ $bc = connect("#bmc") };
    #phash ( %$bc );
    if ( $EVAL_ERROR ) {

      logcomment("Retrying after timeout");
      $retry--;

      if ( keys(%$bc) ) {
        disconnect($bc);
      }

      # for some reason we can't specify the password BEFORE we create the connection, so
      # we have to fail one connect and then set the password on the connectrec object
      # then we can try again.
      $bc->{password} = $password;
      sleep 5;
    } else {
      $retry=0;
      $status=1;
    }
  }

  $bc->{status}=$status;
  return $bc;
}

sub change_ontap_password {
  my ($fc, $passwd, undef) = @_;
  say( $fc, "passwd", -noprompt, -prompt, "New password:", -swallow );
  say( $fc, "$passwd", -noprompt, -prompt, "Retype new password:", -swallow );
  say( $fc, "$passwd", -swallow );
}

sub get_bmc_ip_from_ontap {
  my ($fc, %options) = @_;

  # anonymous hash
  my $ip_hash = {};

  say( $fc, "exit", -blind );

  my (undef, undef, $out, undef) = say($fc, "bmc status", -swallow );

  $out =~ s/^(\s\n)*//;
  my @lines = split ( /\n/, $out );
  chomp(@lines);
  my @matches = grep( /IP|MAC|DHCP/, @lines );
  my $buff;

  foreach my $match (@matches) {
    #chop($match);
    #chop($match);
    (undef, $buff) = split ( /:/, $match, 2 );
    $buff =~ s/^\s*//;
    $buff =~ s/\s*$//;
    if ( $match =~ /Gateway/ ) {
      $ip_hash->{Gateway} = $buff;
    } elsif ( $match =~ /IP address:/ ) {
      $ip_hash->{IP} = $buff;
    } elsif ( $match =~ /IP mask:/ ) {
      $ip_hash->{Mask} = $buff;
    } elsif ( $match =~ /DHCP:/ ) {
      $ip_hash->{Dhcp} = $buff;
    } else {
      $ip_hash->{MAC} = $buff;
    }
  }
  return $ip_hash;
}

sub get_bmc_ip_from_bmc {
  my ($fc, %options) = @_;

  # anonymous hash
  my $ip_hash = {};

  my $out = bmc_command($fc, "bmc config");

  $out =~ s/^(\s\n)*//;
  my @lines = split ( /\n/, $out );
  chomp(@lines);
  my @matches = grep( /ipaddr|netmask|gateway|mac|dhcp/, @lines );

  my $buff;

  foreach my $match (@matches) {
    (undef, $buff) = split ( /:/, $match, 2 );
    $buff =~ s/^\s*//;
    $buff =~ s/\s*$//;
    if ( $match =~ /gateway/ ) {
      $ip_hash->{Gateway} = $buff;
    } elsif ( $match =~ /ipaddr/ ) {
      $ip_hash->{IP} = $buff;
    } elsif ( $match =~ /netmask/ ) {
      $ip_hash->{Mask} = $buff;
    } elsif ( $match =~ /dhcp/ ) {
      $ip_hash->{Dhcp} = $buff;
    } else {
      $ip_hash->{MAC} = $buff;
    }
  }

  return $ip_hash;
}

sub get_system_info_from_bmc {

  my ($fc, %options) = @_;

  # anonymous hash
  my $system_hash = {};

  my $out = bmc_command($fc, "system show");

  $out =~ s/^(\s\n)*//;
  my @lines = split ( /\n/, $out );
  chomp(@lines);
  my @matches = grep( /power|fault|reset|firmware/, @lines );

  my $buff;

  foreach my $match (@matches) {
    (undef, $buff) = split ( /:/, $match, 2 );
    $buff =~ s/^\s*//;
    $buff =~ s/\s*$//;
    if ( $match =~ /power/ ) {
      $system_hash->{power} = $buff;
    } elsif ( $match =~ /fault/ ) {
      $system_hash->{fault} = $buff;
    } elsif ( $match =~ /reset/ ) {
      $system_hash->{reset} = $buff;
    } elsif ( $match =~ /firmware/ ) {
      $system_hash->{firmware} = $buff;
    }
  }

  return $system_hash;
}

# Issues the "system log" command from the BMC and then searches the log for panics. 
# The function also has the option to clear the BMC log by rebooting the BMC from ontap. 
# Can be used in cases where there is no access within a script to check ONTAP console log output. 
sub checkBmcConLogForPanic {
 my ($fc, %options) = @_;
 my $BMC_CTRL_SEQ = "\cg";
 my @BMC_PRMT = ( -noprompt, -prompt, '^bmc.*-> ' );
 my $clearLog = ($options{-clearLog}) ? $options{-clearLog} : "no";
 my $outputinfo;

 say($fc, $BMC_CTRL_SEQ, -nocr, @BMC_PRMT, -swallow );
 say($fc, "oem", @BMC_PRMT, -swallow );
 (undef, undef, $outputinfo) = say ($fc, "system log", @BMC_PRMT, -swallow,-timeout, 300);
 if ($outputinfo =~ /PANIC/) {
            logcomment ("Filer hit an unexpected panic.\n");
            die "FATAL","Filer hit an unexpected panic";
 }
 sleep 15;

 # This functionality is provided so that upon the next invocation of the
 # function, the system log will only contain most recent syslog entries
 if ( $clearLog =~ m/yes/i ) {
    logcomment ("Rebooting the BMC to clear its system log.\n");
    say($fc, "exit", -blind, -swallow );
    login2ontap($fc);
    reboot_bmc ($fc);
 } else {
    say($fc, "exit", -blind, -swallow );
    login2ontap($fc);
  }
}

# Issues the "system log" command from the BMC and then searches the log for CECC. 
# The function also has the option to clear the BMC log by rebooting the BMC from ontap. 
# Can be used in cases where there is no access within a script to check ONTAP console log output. 
sub checkBmcConLogForCECC {
 my ($fc, %options) = @_;
 my $BMC_CTRL_SEQ = "\cg";
 my @BMC_PRMT = ( -noprompt, -prompt, '^bmc.*-> ' );
 my $clearLog = ($options{-clearLog}) ? $options{-clearLog} : "no";
 my $outputinfo;

 say($fc, $BMC_CTRL_SEQ, -nocr, @BMC_PRMT, -swallow );
 say($fc, "oem", @BMC_PRMT, -swallow );
 (undef, undef, $outputinfo) = say ($fc, "system log", @BMC_PRMT, -swallow,-timeout, 300);
 # Assume that an EMS message similar to the following is generated
 # [cecc_log.entry:warning]: 1: Non Fatal Correctable DRAM ECC Channel A
 if ($outputinfo =~ /DRAM ECC/) {
            logcomment ("Filer hit CECC Errors.\n");
            die "FATAL","Filer hit unexpected CECC Errors";
 }
 sleep 15;

 # This functionality is provided so that upon the next invocation of the
 # function, the system log will only contain most recent syslog entries
 if ( $clearLog =~ m/yes/i ) {
    logcomment ("Rebooting the BMC to clear its system log.\n");
    say($fc, "exit", -blind, -swallow );
    login2ontap($fc);
    reboot_bmc ($fc);
 } else {
    say($fc, "exit", -blind, -swallow );
    login2ontap($fc);
  }
}


1;
