# this is basically -*-perl-*-
# $Id: //depot/prod/test/Rsierranevada/lib/Platform/Environment/Base.pm#1 $
#
## @summary	Base.pm - Provide a base class for the Environmental Subsystem
## @author	dl-platform-qa@netapp.com
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod		here

=head1 NAME

Base - Provide a base class for the Environmental Subsystem

=head1 SYNOPSIS

  use Platform::Environment;

  # Base constructor when using a filer
  $obj = Platform::Environment::Base->new(hostspec => $hostspec);

  # (List Sensors)
  # (List of jointly monitored chassis sensors (classes) information)
  # Displays all non-shelf environmental information for the filer by class:
  # environment chassis | environment status chassis
  $out = $obj->chassis_show;

  # (List all sensors in a monitor-wise hierarchical fashion)
  # Lists all sensors in the system, their current readings, states and their
  # thresholds:
  # environment chassis list-sensors | environment status chassis list-sensors
  $out = $obj->chassis_list;

  # (List all sensors regardless of whether they are monitored)
  # Displays all the available environmental information:
  # environment status
  $out = $obj->status_show;

  # Display the environmental status for the specified class:
  # environment chassis $class
  $out = $obj->chassis_show(class=>$class);

  # Test states and Policies of a sensor:
  # environment chassis "SENSOR_NAME" test-all
  $out = $obj->chassis_test(sensor=>$sensor);

  # Test states and Policies of all sensors:
  # environment chassis test-all
  $out = $obj->chassis_test(sensor=>'all');

  # Test states and Policies of sensor(s) forever. Use 'runtime' option to
  # determine when to stop the command using control C:
  # environment chassis test-all repeat
  $out = $obj->chassis_test(sensor=>$sensor, repeat=>1, runtime=>10);

  # Set state of a sensor:
  # environment chassis $sensor set-state $state
  $out = $obj->chassis_set(state=>$state, sensor=>$sensor);

  # Read Current Value of a sensor
  # environment chassis "SENSOR_NAME" read-curr
  $out = $obj->chassis_read_curr(sensor=>$sensor);

  # Print Attributes of a sensor
  # environment chassis "SENSOR_NAME" print
  $out = $obj->chassis_print(sensor=>$sensor);

  # Quick Print Attributes of a sensor
  # environment chassis "SENSOR_NAME" quick-print
  $out = $obj->chassis_quick_print(sensor=>$sensor);

  # Read Thresholds of a sensor
  # environment chassis "SENSOR_NAME" read-thresh
  $out = $obj->chassis_read_thresh(sensor=>$sensor);

  # Environmental Shutdown
  # Supported actions = enable|disable
  # environment chassis ok-halt
  # environment chassis no-halt
  $out = $obj->shutdown_enable;
  $out = $obj->shutdown_disable;

  # Environmental Policies
  # Supported actions = enable|disable|query
  # environment chassis ok-policy
  # environment chassis no-policy
  # environment chassis is-policy
  $out = $obj->policies_enable;
  $out = $obj->policies_disable;
  $out = $obj->policies_query;

  # Initiate 'Top-of-the-Hour' behavior by updating all the sensors:
  # environment chassis hourly
  $out = $obj->top_of_the_hour_sensors_update

  # I2C Stress Tests
  # i2c_stress start
  # i2c_stress status
  # i2c_stress fail-verbose-off
  # i2c_stress fail-verbose-on
  # i2c_stress stop
  $out = $obj->stress_start;
  $out = $obj->stress_status;
  $out = $obj->stress_fail_verbose_off;
  $out = $obj->stress_fail_verbose_on;
  $out = $obj->stress_stop;

  # Get list of all sensors that should be in the system:
  # Lists all expected sensors for a given platform.
  $out = $obj->sensor_list([type => $type]);

  # Get sysconfig items that are common across all filers:
  $out = get_common_cfg();

  # Get the sensor status for testing verifications:
  # Returns test_status, class_status, state_status, and binary_reading
  $out = get_sensor_status;

  # Example usage (Model Name: FAS6080)
  # Usage: environment status |
  #          [status] [shelf [<adapter>]] |
  #          [status] [shelf_log] |
  #          [status] [shelf_stats] |
  #          [status] [shelf_power_status] |
  #          [status] [chassis [all | list-sensors | Voltage | PSU 1 | PSU 2 |
  #                             Temperature | System_Fan | CF_status |
  #                             NVRAM6-temperature-1 | NVRAM6-battery-1]]

  # Example usage (Model Name: FAS960)
  # Usage: environment status |
  #          [status] [shelf [<adapter>]] |
  #          [status] [shelf_log] |
  #          [status] [shelf_stats] |
  #          [status] [shelf_power_status] |
  #          [status] [chassis [all | list-sensors | Fan | Power | Temp |
  #                             Power Supply | RTC Battery |
  #                             NVRAM4-temperature-7 | NVRAM4-battery-7]]

  # Example usage (Model Name: FAS270)
  # Usage: environment status |
  #          [status] [shelf [<adapter>]] |
  #          [status] [shelf_log] |
  #          [status] [shelf_stats] |
  #          [status] [shelf_power_status] |
  #          [status] [chassis [all | list-sensors | Voltage | Temperature |
  #                             Power Supplies | Battery]]

  # Example usage (Model Name: FAS3040)
  # Usage: environment status |
  #          [status] [shelf [<adapter>]] |
  #          [status] [shelf_log] |
  #          [status] [shelf_stats] |
  #          [status] [shelf_power_status] |
  #          [status] [chassis [all | list-sensors | Temperature | PSU 1 |
  #                             PSU 2 |Voltage | SYS FAN |
  #                             NVRAM6-temperature-1 | NVRAM6-battery-1]]

  # Example usage (Model Name: FAS2020)
  # Usage: environment status |
  #          [status] [shelf [<adapter>]] |
  #          [status] [shelf_log] |
  #          [status] [shelf_stats] |
  #          [status] [shelf_power_status] |
  #          [status] [chassis [all | list-sensors | Voltage | Temperature |
  #                             Power Supplies OutVoltage | Battery |
  #                             Power Supply 1 | Power Supply 2]]

=head1 DESCRIPTION

Provides routines for diagnostic command line interface to the filer's
physical environment.

Note: The column widths for 'environment chassis list-sensors' output are coded
      in the @width array. This assumption is based on fixed column widths.
      Parsing dynamically at this time is not an option since sensor names vary
      in length and there is nothing common to each column such as double
      spaces between each one.

WARNING: These commands are dangerous and can cause data corruption or
         system failure if used incorrectly or unwisely. Do NOT use the
         commands unless you know EXACTLY what you are doing.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-platform-qa@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/Platform/Environment/Base.pm#1 $
  Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

  Copyright 2008 Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package Platform::Environment::Base;

use strict;
use warnings;

use Carp qw(croak);
use SayWrap qw(sayEOF);
use Params::Validate qw(validate SCALAR);
use Server;

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

#---------------------------------------------------------------------
# Global variables
#---------------------------------------------------------------------
our $AUTOLOAD;

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#---------------------------------------------------------------------
# Local variables
#---------------------------------------------------------------------
# Various sensor status for testing verifications:
my %status = (
    test_status    => 'passed',
    class_status   => 'ok|charging',
    state_status   => 'normal',
    binary_reading => 'good',
);
# Sysconfig items that are common across all filers:
my %table = (
    common_cfg => [
        'System Serial Number',
        'Model Name',
        'System ID',
        'Kernel version',
        'System Rev',
        'Firmware release',
        'Mother Board',
    ]
);

#---------------------------------------------------------------------

=head2 new

  $base_obj = Platform::Environment::Base->new(hostspec => $hostspec
                                               [, connspec => $connspec ]);

Creates a base object. This is an abstract class, so do not use
this constructor directly, use Platform::Environment->new() instead.

=for html <hr />

=cut

#---------------------------------------------------------------------

sub new {
    my ($class, %opts) = @_;
    return bless \%opts, $class;
}

#---------------------------------------------------------------------
# AUTOLOAD to exposed all keys in the blessed object as read-only methods
# Do not expose private keys (the ones which start with a '_')
#---------------------------------------------------------------------
sub AUTOLOAD {
    my $self = shift;
    my $method = $AUTOLOAD;
    $method =~ s/.*:://;
    if ( $method =~ /^_/ || !exists $self->{$method} ) {
        croak qq(Can't locate object method "$method" via package ") .
              __PACKAGE__ . '"';
    }
    return $self->{$method};
}

#---------------------------------------------------------------------

=head2 chassis_show

  $out = $base_obj->chassis_show([ class => $class ]);

The following named arguments are accepted:

=over

=item Arguments

=over

=item class

(Optional) : Set of jointly monitored chassis sensors or 'all' for all classes.

=back

=item Returns

An array reference of the environment chassis information.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub chassis_show {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            class => { type => SCALAR, optional => 1 },
        }
    );

    return $self->_chassis(action => 'show', %opts);
}

#---------------------------------------------------------------------

=head2 chassis_test

  $out = $base_obj->chassis_test(sensor => $sensor);

Test states and Policies of a sensor or all sensors.

When testing sensor(s) forever using the 'repeat' argument, the 'runtime'
option will determine when to stop the command by issuing a control C.

The following named arguments are accepted:

=over

=item Arguments

=over

=item sensor

(Required) : Chassis sensor name or 'all' for all sensors.

=item repeat

(Required) : Repeats the test-all commands forever. Default value is 0.

=item runtime

(Optional) : Runtime used with the repeat argument to determine when to stop
the command by issuing a control C. Default value is 5 (minutes).

=back

=item Returns

A hash reference of the environment chassis information loaded in an array.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub chassis_test {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            sensor  => { type => SCALAR},
            repeat  => { type => SCALAR, default => 0 },
            runtime => { type => SCALAR, default => 5 },
        }
    );

    return $self->_chassis(action => 'test', %opts);
}

#---------------------------------------------------------------------

=head2 chassis_set

  $out = $base_obj->chassis_set(sensor => $sensor, state => $state);

Set state of a sensor.

The following named arguments are accepted:

=over

=item Arguments

=over

=item sensor

(Required) : Chassis sensor name.

=item state

(Required) : Warning state. Supported options for linear sensors are 'normal'
for normal sensor operation, 'warn-low' to force low temp warning state,
'warn-high' to force high temp warning state, 'crit-low' to force critical low
temp state, and 'crit-high' to force critical high temp state. A binary sensor
might be in either 'normal' or 'bad' state.

=back

=item Returns

An array reference of the environment chassis information.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub chassis_set {
    my $self = shift;
    my $regex = qr/^(normal|warn-low|warn-high|crit-low|crit-high|bad)$/;
    my (%opts) = validate (
        @_,
        {
            sensor => { type => SCALAR },
            state  => { type => SCALAR, regex => $regex },
        }
    );

    return $self->_chassis(action => 'set', %opts);
}

#---------------------------------------------------------------------

=head2 chassis_list

  $out = $base_obj->chassis_list();

List all sensors in a monitor-wise hierarchical fashion.
The list will include all sensors in the system, their current readings, states
and their thresholds.

=over

=item Returns

A hash reference of the environment chassis information loaded in an array.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub chassis_list {
    my $self = shift;

    return $self->_chassis(action => 'list');
}

#---------------------------------------------------------------------

=head2 chassis_read_curr

  $out = $base_obj->chassis_read_curr(sensor => $sensor);

Read current value of a sensor.

The following named arguments are accepted:

=over

=item Arguments

=over

=item sensor

(Required) : Chassis sensor name.

=back

=item Returns

A hash reference of the environment chassis information.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub chassis_read_curr {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            sensor => { type => SCALAR },
        }
    );

    return $self->_chassis(action => 'read-curr', %opts);
}

#---------------------------------------------------------------------

=head2 chassis_print

  $out = $base_obj->chassis_print(sensor => $sensor);

Print attributes of a sensor.

The following named arguments are accepted:

=over

=item Arguments

=over

=item sensor

(Required) : Chassis sensor name.

=back

=item Returns

A hash reference of the environment chassis information.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub chassis_print {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            sensor => { type => SCALAR },
        }
    );

    return $self->_chassis(action => 'print', %opts);
}

#---------------------------------------------------------------------

=head2 chassis_quick_print

  $out = $base_obj->chassis_quick_print(sensor => $sensor);

Quick print attributes of a sensor.

The following named arguments are accepted:

=over

=item Arguments

=over

=item sensor

(Required) : Chassis sensor name.

=back

=item Returns

A hash reference of the environment chassis information.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub chassis_quick_print {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            sensor => { type => SCALAR },
        }
    );

    return $self->_chassis(action => 'quick-print', %opts);
}

#---------------------------------------------------------------------

=head2 chassis_read_thresh

  $out = $base_obj->chassis_read_thresh(sensor => $sensor);

Read the thresholds of a sensor.

The following named arguments are accepted:

=over

=item Arguments

=over

=item sensor

(Required) : Chassis sensor name.

=back

=item Returns

A hash reference of the environment chassis information.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub chassis_read_thresh {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            sensor => { type => SCALAR },
        }
    );

    return $self->_chassis(action => 'read-thresh', %opts);
}

#---------------------------------------------------------------------

=head2 sensor_list

  $out = $base_obj->sensor_list([type => $type]);

Lists all expected sensors for a given platform.

The following named arguments are accepted:

=over

=item Arguments

=over

=item type

(Optional) : Controller location. The types supported are 'volt', 'psu',
'psu_temp', 'psu_fan', 'temp', 'fan', 'batt', 'nvram_temp', 'nvram_batt', and
'cluster'. All types are returned if type is not defined.

=back

=item Returns

A array reference of the expected environment sensors.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub sensor_list {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            type => { type => SCALAR, optional => 1 },
        }
    );
    my (@list, $key);

    foreach $key (keys %{$self->{_loc_map}}) {
        if (!defined $opts{type} || $key =~ /$opts{type}/) {
            push @list, @{$self->{_loc_map}{$key}};
        }
    }

    return \@list;
}

#---------------------------------------------------------------------

=head2 status_show

 $out = status_show();

Displays all the available environmental information:

=over

=item Returns

An array reference of the environment chassis information or croak on failure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub status_show {
    my $self = shift;
    my ($cmd, $out, @lines, $line, @list);

    # Build the environment command
    $cmd = "environment status ";

    # Issue the command
    $out = sayEOF($self->{conn}, $cmd);

    # The command must be supported by the filer to continue and the
    # Environmental Subsystem must be available
    if ($out =~ /not found/i || $out =~ /Usage/i) {
        croak ("Filer does not support command: $cmd");
    } elsif ($out =~/Environmental initialization/i) {
        croak ("Error: $out");
    }

    # Put the messages per line into a list (array)
    @lines = split /\n/, $out;
    foreach $line (@lines) {
        next if !$line;

        # Take out leading spaces
        $line =~ s/^\s+//;

        # Load up the array
        push @list, $line;
    }

    return \@list;
}

#---------------------------------------------------------------------

=head2 policies_enable

 $out = $base_obj->policies_enable();

Enable Environmental Policies

=over

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub policies_enable {
    my $self = shift;

    return $self->_policies(action => 'enable');
}

#---------------------------------------------------------------------

=head2 policies_disable

 $out = $base_obj->policies_disable();

Disable Environmental Policies

=over

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub policies_disable {
    my $self = shift;

    return $self->_policies(action => 'disable');
}

#---------------------------------------------------------------------

=head2 policies_query

 $out = $base_obj->policies_query();

Query Environmental Policies

=over

=item Returns

The current policy state. (YES|NO)

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub policies_query {
    my $self = shift;

    return $self->_policies(action => 'query');
}

#---------------------------------------------------------------------

=head2 shutdown_enable

 $out = $base_obj->shutdown_enable();

Enable Environmental Shutdown

=over

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub shutdown_enable {
    my $self = shift;

    return $self->_shutdown(action => 'enable');
}

#---------------------------------------------------------------------

=head2 shutdown_disable

 $out = $base_obj->shutdown_disable();

Disable Environmental Shutdown

=over

=item Returns

1 on success.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub shutdown_disable {
    my $self = shift;

    return $self->_shutdown(action => 'disable');
}

#---------------------------------------------------------------------

=head2 stress_start

  $out = $base_obj->stress_start;

Start the I2C stress tests.

=over

=item Returns

A hash reference of the results of the I2C stress test command.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub stress_start {
    my $self = shift;

    return $self->_stress(action => 'start');
}

#---------------------------------------------------------------------

=head2 stress_status

  $out = $base_obj->stress_status

Start the I2C stress tests.

=over

=item Returns

A hash reference of the I2C stress test status.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub stress_status {
    my $self = shift;

    return $self->_stress(action => 'status');
}

#---------------------------------------------------------------------

=head2 stress_fail_verbose_off

  $out = $base_obj->stress_fail_verbose_off;

Switch off details of failed sensors - I2C stress tests.

=over

=item Returns

A hash reference of the results of the I2C stress test command.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub stress_fail_verbose_off {
    my $self = shift;

    return $self->_stress(action => 'fail-verbose-off');
}

#---------------------------------------------------------------------

=head2 stress_fail_verbose_on

  $out = $base_obj->stress_fail_verbose_on;

Switch on details of failed sensors - I2C stress tests.

=over

=item Returns

A hash reference of the results of the I2C stress test command.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub stress_fail_verbose_on {
    my $self = shift;

    return $self->_stress(action => 'fail-verbose-on');
}

#---------------------------------------------------------------------

=head2 stress_stop

  $out = $base_obj->stress_stop;

Stop the I2C stress tests.

=over

=item Returns

A hash reference of the results of the I2C stress test command.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub stress_stop {
    my $self = shift;

    return $self->_stress(action => 'stop');
}

#---------------------------------------------------------------------

=head2 top_of_the_hour_sensors_update

 $out = $obj->top_of_the_hour_sensors_update

Initiate 'Top-of-the-Hour' behavior by updating all the sensors:

=over

=item Returns

1 on success or croak on failure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub top_of_the_hour_sensors_update {
    my $self = shift;
    my ($cmd, $out);

    # Build the environment command
    $cmd = "environment chassis hourly ";

    # Issue the command
    $out = sayEOF($self->{conn}, $cmd);

    # The command must be supported by the filer to continue and the
    # Environmental Subsystem must be available
    if ($out =~ /not found/i || $out =~ /Usage/i) {
        croak ("Filer does not support command: $cmd");
    } elsif ($out =~/Environmental initialization/i) {
        croak ("Error: $out");
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 get_sensor_status

 $out = get_sensor_status;

Get the sensor status for testing verifications:

=over

=item Returns

A hash reference of the various environmental sensor status.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_sensor_status {
    my $self = shift;

    return bless \%status;
}

#---------------------------------------------------------------------

=head2 get_common_cfg

 $out = get_common_cfg();

The 'sysconfig -M' contents that are common across all filers:

=over

=item Returns

A hash reference of the common 'sysconfig -M' contents.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_common_cfg {
    my $self = shift;

    return bless \%table;
}

#---------------------------------------------------------------------

=head2 sysconfig_out

  $out = $base_obj->sysconfig_out();

Get 'sysconfig -M' FRU information for a given platform.

=over

=item Returns

A hash reference of the system FRUs or croak on failure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub sysconfig_out {
    my $self = shift;
    my ($out, $cmd, $key, @list, $line, $hash, %ret);

    # Build the command
    $cmd = "sysconfig -M ";

    # Issue the command
    $out = sayEOF($self->{conn}, $cmd);

    # The command must be supported by the filer to continue and the
    # Environmental Subsystem must be available
    if ($out =~ /not found/i || $out =~ /Usage/i) {
        croak ("Filer does not support command: $cmd");
    } elsif ($out =~/Environmental initialization/i) {
        croak ("Error: $out");
    }

    # Parse the information

    # Take out leading blanks
    $out =~ s/^\s+//mg;

    # Split the output
    foreach $line (split /\n/, $out) {
        next if !($line =~ s/^!//);

        # Split the line based on exclamation marks
        @list = ($line =~ /([^!]*)!/g);

        # The first element of the array contains the system FRU name or an
        # informational name with additional information in the rest of the
        # elements (ei: model|serial|version|part|etc)
        if (@list == 2) {
            # The split will be a key and its value
            $ret{$list[0]} = $list[1];
        } elsif (@list > 2) {
            # The split will be a key and and a hash
            $key = $list[0];
            $hash = {
                serial => $list[1],
                partno => $list[2],
            };
            if (@list > 3) {
                # Remove elements from the end of @list so that {other} will
                # contain an array of the remaining fields
                # (some are not yet defined)
                $hash->{other} = [splice @list, 3];
            }
            if (!exists $ret{$key}) {
                # Key does not exist yet, value is a hash
                $ret{$key} = $hash;
            } else {
                # Key does exist, aggregate all values into an array
                if (ref $ret{$key} eq 'HASH') {
                    # Change hash into an array
                    $ret{$key} = [ $ret{$key} ];
                }
                push @{$ret{$key}}, $hash;
            }
        } else {
            croak "Data can not be parsed";
        }
    }

    return \%ret;
}

#---------------------------------------------------------------------
=head2 Gxsysconfig_out

  $out = $base_obj->Gxsysconfig_out();

Get 'sysconfig -M' FRU information for a given platform.

=over

=item Returns

A hash reference of the system FRUs or croak on failure.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub Gxsysconfig_out {
    my $self = shift;
    my ($out, $cmd, $key, @list, $line, $hash, %ret, $Server);

    # Build the command
    $cmd = "run local sysconfig -M ";

    # Issue the command
    #$out = sayEOF($self->{conn}, $cmd);
    $Server = $self->{conn};
    $out = $Server->cli(cmd =>$cmd);

    # The command must be supported by the filer to continue and the
    # Environmental Subsystem must be available
    if ($out =~ /not found/i || $out =~ /Usage/i) {
        croak ("Filer does not support command: $cmd");
    } elsif ($out =~/Environmental initialization/i) {
        croak ("Error: $out");
    }

    # Parse the information

    # Take out leading blanks
    $out =~ s/^\s+//mg;

    # Split the output
    foreach $line (split /\n/, $out) {
        next if !($line =~ s/^!//);

        # Split the line based on exclamation marks
        @list = ($line =~ /([^!]*)!/g);

        # The first element of the array contains the system FRU name or an
        # informational name with additional information in the rest of the
        # elements (ei: model|serial|version|part|etc)
        if (@list == 2) {
            # The split will be a key and its value
            $ret{$list[0]} = $list[1];
        } elsif (@list > 2) {
            # The split will be a key and and a hash
            $key = $list[0];
            $hash = {
                serial => $list[1],
                partno => $list[2],
            };
            if (@list > 3) {
                # Remove elements from the end of @list so that {other} will
                # contain an array of the remaining fields
                # (some are not yet defined)
                $hash->{other} = [splice @list, 3];
            }
            if (!exists $ret{$key}) {
                # Key does not exist yet, value is a hash
                $ret{$key} = $hash;
            } else {
                # Key does exist, aggregate all values into an array
                if (ref $ret{$key} eq 'HASH') {
                    # Change hash into an array
                    $ret{$key} = [ $ret{$key} ];
                }
                push @{$ret{$key}}, $hash;
            }
        } else {
            croak "Data can not be parsed";
        }
    }

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# Parse output 'environment chassis' command
#---------------------------------------------------------------------
sub _chassis {
    my $self = shift;
    my $rx = qr/^(show|test|set|list|read-curr|print|quick-print|read-thresh)$/;
    my $rx2 = qr/^(normal|warn-low|warn-high|crit-low|crit-high|bad)$/;
    my (%opts) = validate (
        @_,
        {
            action  => { type => SCALAR, regex => $rx },
            class   => { type => SCALAR, optional => 1 },
            sensor  => { type => SCALAR, optional => 1 },
            state   => { type => SCALAR, optional => 1, regex => $rx2 },
            repeat  => { type => SCALAR, default => 0 },
            runtime => { type => SCALAR, default => 5 },
        }
    );

    my ($out, $cmd, %hash, $i, $line, @lines, @list, %ret, $offset, $chead,
        $fhead, @ckeys, @fkeys, @cols, @vals, $width, $conn, $end_time,
        $connected, $err);

    # These are the column widths for 'environment chassis list-sensors'
    # output. This assumption is based on fixed column widths. Parsing
    # dynamically at this time is not an option since sensor names vary in
    # length and there is nothing common to each column such as double spaces
    # between each one.
    my @width = qw(25 15 11 13 11 11 10);

    # Example usage (Model Name: FAS6080)
    #  Usage: environment status |
    #           [status] [shelf [<adapter>]] |
    #           [status] [shelf_log] |
    #           [status] [shelf_stats] |
    #           [status] [shelf_power_status] |
    #           [status] [chassis [all | list-sensors | Voltage | PSU 1 |
    #                              PSU 2 | Temperature | System_Fan |
    #                              CF_status |
    #                              NVRAM6-temperature-1 | NVRAM6-battery-1]]

    # Build the environment command
    $cmd = "environment chassis ";

    # Display the environmental status for the specified class
    if (defined($opts{class})) {
        croak "Invalid class $opts{class}"
          if !exists $self->{_class_map}{$opts{class}};
        $cmd .= "$self->{_class_map}{$opts{class}} ";
    }

    # Set action
    if ($opts{action} =~ /read-curr|print|quick-print|read-thresh/) {
        $cmd .= "$opts{sensor} $opts{action} ";
    } elsif ($opts{action} eq 'set') {
        # Set state of a sensor
        $cmd .= "$opts{sensor} set-state $opts{state}";
    } elsif ($opts{action} eq 'list') {
        # Lists all sensors in the system with their current readings, states
        # and thresholds
        $cmd .= "list-sensors";
    } elsif ($opts{action} =~ /test/) {
        # Test all states and Policies of a sensor
        if ($opts{sensor} =~ /all/) {
            $cmd .= "test-all ";
        } else {
            $cmd .= "$opts{sensor} test-all ";
        }
    }

    # Set the repeat argument
    if ($opts{repeat}) {
        $cmd .= "$opts{repeat} ";
    }

    # Get the messages
    if ($opts{repeat}) {
        $conn = $self->{conn};

        # A telnet connection is required to get messages
        if ($conn->conntype ne 'telnet') {
            Tharn::host($conn->hostp->id);
            $conn = Tharn::connect('#telnet');
            $connected = 1;
        }

        # Use the say command with the '-blind' option so that the say command
        # does not timeout
        Tharn::say($conn, $cmd, '-blind');

        # Eval the say command so that if there is an error before runtime has
        # completed, the repeats forever command can be stopped.
        eval {
            # Gather messages
            $out = '';
            $end_time = time + $opts{runtime}*60;
            while ($end_time > time) {
                # Get all the messages line by line
                $line = Tharn::say($conn, '', '.*\r', '-noprompt', '-nocr');
                # Get only the summary messages otherwise all console
                # messages will be collected
                if ($line =~ /Sensor/) {
                    $out .= "$line\n";
                }
            }
        };
        $err = $@;

        # Stop the repeat forever command with a '^C' (send interrupt)
        eval { $conn->send_intr() };

        # Terminating test of all sensors early because of a '^C' may
        # take a few seconds to complete
        sleep(10);

        # Disconnect the 'telnet' connection if it was set here
        eval { Tharn::disconnect('#telnet') } if $connected;
        croak $err if $err;

        # Remove the carriage return
        $out =~ s/\r//g;
    } else {
        $out = sayEOF($self->{conn}, $cmd, timeout=>1800);
    }

    # The command must be supported by the filer to continue and the
    # Environmental Subsystem must be available
    if ($out =~ /not found/i || $out =~ /Usage/i) {
        croak ("Filer does not support command: $cmd");
    } elsif ($out =~/Environmental initialization/i) {
        croak ("Error: $out");
    }

    # Parse list messages
    if ($opts{action} =~ /list/) {
        # Split the headings from the data
        ($fhead, $out) = split /^-+.*\n/m, $out;

        # Split the 2 rows of headings
        ($chead, $fhead) = split /\n/, $fhead;

        # Split each row of column headings
        @fkeys = ($fhead =~ /(\S+)/g);
        @ckeys = split /\s{2,}/, $chead;

        # Join the column headings into a single field per column
        for($i=2; $i< scalar(@ckeys); $i++) {
            $ckeys[$i] .= " $fkeys[$i-2]";
        }

        # Split each of the messages
        @lines = split /\n/, $out;

        # Setup the column widths (size and offset)
        @cols = ();
        $offset = 0;
        foreach $width (@width) {
            push @cols, { size => $width, offset => $offset };
            $offset += $width;
        }

        # Do not truncate data on the last column
        $cols[-1]{size} = 99999;

        foreach $line (@lines) {
            # Take out leading spaces
            $line =~ s/^\s+//;

            next if !$line;

            foreach $i ( 0..@cols-1 ) {
                if ( $cols[$i]{offset} > length($line) ) {
                     # This line has no more data for this column
                     $vals[$i] = '';
                } else {
                    # Get data given by the current offset and size
                    $vals[$i] = substr(
                    $line,
                    $cols[$i]{offset},
                    $cols[$i]{size},
                    );
                    # Remove any unnecessary spaces
                    $vals[$i] =~ s/^\s+//;
                    $vals[$i] =~ s/\s+$//;
                }
            }

            # Join up the information using the headings as keys
            @hash{@ckeys} = @vals;

            # Load up all the sensor information into an array
            push @list, _change_keys(\%hash);
        }
    } elsif ($opts{action} eq 'print' || $opts{action} eq 'read-thresh') {
        # Parse 'print|read-thresh' messages
        $out =~ s/,/\n/;
        %ret = ($out =~ /\s*([^:|^=]+)[:|=]\s+(.+)/g);

        return \%ret;
    } else {
        # Parse show|set|test|read-curr|quick-print messages

        # Take out leading spaces
        $out =~ s/^\s+//;

        # Take out ending period '.'
        $out =~ s/\.$//;

        # Put the messages per line into a list (array)
        @list = split /\n/, $out;
    }

    return \@list;
}

#---------------------------------------------------------------------
# Internal method
# enable|disable|query Environmental Policies
#---------------------------------------------------------------------
sub _policies {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            action => { type => SCALAR, regex => qr/^(enable|disable|query)$/ },
        }
    );
    my ($cmd, $out);

    # Build the environment command
    $cmd = "environment chassis ";

    if ($opts{action} eq 'enable') {
        $cmd .= "ok-policy";
    } elsif ($opts{action} eq 'disable') {
        $cmd .= "no-policy";
    } else {
        $cmd .= "is-policy";
    }

    # Issue the command
    $out = sayEOF($self->{conn}, $cmd);

    # The command must be supported by the filer to continue and the
    # Environmental Subsystem must be available
    if ($out =~ /not found/i || $out =~ /Usage/i) {
        croak ("Filer does not support command: $cmd");
    } elsif ($out =~/Environmental initialization/i) {
        croak ("Error: $out");
    }

    $out = 1 if !$out;

    return $out;
}

#---------------------------------------------------------------------
# Internal method
# start|status|fail-verbose-off|fail-verbose-on|stop I2C Stress Tests
#---------------------------------------------------------------------
sub _stress {
    my $self = shift;
    my $regex = qr/^(start|status|fail-verbose-off|fail-verbose-on|stop)$/;
    my (%opts) = validate (
        @_,
        {
            action => { type => SCALAR, regex => $regex },
        }
    );
    my ($cmd, $chead, $line, @lines, $val, $label, $out, %ret);

    # Build the environment command
    $cmd = "priv set -q test; i2c_stress $opts{action} ";

    # Issue the command
    $out = sayEOF($self->{conn}, $cmd);

    # The command must be supported by the filer to continue and the
    # Environmental Subsystem must be available
    if ($out =~ /not found/i || $out =~ /Usage/i) {
        croak ("Filer does not support command: $cmd");
    } elsif ($out =~/Environmental initialization/i) {
        croak ("Error: $out");
    }

    # Parse the information

    # Split each line
    @lines = split /\n/, $out;

    # Parse each line of status message(s)
    foreach $line (@lines) {
        next if($line eq '');

        if ($line =~ /\=/) {
            # Split the heading and value
            ($chead, $val) = ($line =~ /\s*(.+?)\=\s*(\d+)/);

            # Change spaces to underscore
            $chead =~ s/\s/_/g;

            # Load return hash
            $ret{$chead} = $val;
        } else {
            # Load all Messages into 1 hash for general information
            $ret{msg} .= "$line\n";

            # Break out the test status for easy access
            if ($line =~ /test/) {
                $label = 1;
                if ($line =~ /not/) {
                    $label = 0;
                }

                # 0 = Test is not Running or 1 = Test is Running
                $ret{"is_test_running"} = $label;
            } elsif ($line =~ /error reporting/) {
                # Break out the error reporting status for easy access
                # This will be terse|verbose
                ($label) = ($line =~ /is\s+(\w+)/);
                $ret{"error_reporting"} = $label;
            }
        }
    }

    return \%ret;
}

#---------------------------------------------------------------------
# Internal function to change spaces on the keys of the specified hash
#---------------------------------------------------------------------
sub _change_keys($;) {
    my $hash = shift;
    my (%ret, $key);

    # Process all keys in the hash
    foreach ( keys %$hash ) {
        $key = $_;
        # Remove trailing spaces
        $key =~ s/\s$//g;
        # Change spaces to underscore
        $key =~ s/\s/_/g;
        $ret{$key} = $hash->{$_};
    }

    return \%ret;
}

#---------------------------------------------------------------------
# Internal method
# enable|disable Environmental Shutdown
#---------------------------------------------------------------------
sub _shutdown {
    my $self = shift;
    my (%opts) = validate (
        @_,
        {
            action => { type => SCALAR, regex => qr/^(enable|disable)$/ },
        }
    );
    my ($cmd, $out);

    # Build the environment command
    $cmd = "environment chassis ";

    if ($opts{action} eq 'enable') {
        $cmd .= "ok-halt";
    } else {
        $cmd .= "no-halt";
    }

    # Issue the command
    $out = sayEOF($self->{conn}, $cmd);

    # The command must be supported by the filer to continue and the
    # Environmental Subsystem must be available
    if ($out =~ /not found/i || $out =~ /Usage/i) {
        croak ("Filer does not support command: $cmd");
    } elsif ($out =~/Environmental initialization/i) {
        croak ("Error: $out");
    }

    return 1;
}
#---------------------------------------------------------------------
sub Gx_envsensors {
    my $self = shift;
    my $rx1 = qr/^(PSU1|PSU2|Fan-FRU1|Fan-FRU2|Fan-FRU3|Fan-FRU4|Fan-FRU5)$/;
    my $rx2 = qr/^(temperature|fan|voltage|agent)$/;
    my $rx3 = qr/^(normal|not present|warning low|warning high|critical low|critical high|bad|failed|retry|unknown)$/;
    my (%opts) = validate (
        @_,
        {
            filer         => { type => SCALAR},
            sensorindex   => { type => SCALAR},
            sensorname    => { type => SCALAR},
            fru           => { type => SCALAR, regex => $rx1},
            sensortype    => { type => SCALAR, regex => $rx2},
            units         => { type => SCALAR},
            state         => { type => SCALAR, regex => $rx3},
            critlow       => { type => SCALAR},
            warnlow       => { type => SCALAR},
            warnhi        => { type => SCALAR},
            crithi        => { type => SCALAR},
        }
    );

    my ($out, $cmd, %hash, );

    if(defined($opts{filer})) {
        $cmd = "environment sensors show -node $opts{filer} ";
    } else {
        croak ("Filer options is not passed to the subroutine");
    }

    if(defined($opts{sensorindex})) {
        $cmd .= "-index $opts{sensorindex} ";
    }

    if(defined($opts{sensorname})) {
        $cmd .= "-name $opts{sensorname} ";
    }

    if(defined($opts{fru})) {
        $cmd .= "-fru $opts{fru} ";
    }
    
    if(defined($opts{sensortype})) {
        $cmd .= "-type $opts{sensortype} ";
    }

    if(defined($opts{units})) {
        $cmd .= "-units $opts{units} ";
    }

    if(defined($opts{state})) {
        $cmd .= "-state $opts{state} ";
    }

    if(defined($opts{critlow})) {
        $cmd .= "-crit-low $opts{critlow} ";
    }

    if(defined($opts{warnlow})) {
        $cmd .= "-warn-low $opts{warnlow} ";
    }

    if(defined($opts{warnhi})) {
        $cmd .= "-warn-hi $opts{warnhi} ";
    }

    if(defined($opts{crithi})) {
        $cmd .= "-crit-hi $opts{crithi} ";
    }

    my $Server = Server->new(nodes => $opts{filer});
    $out = $Server->cli(cmd => $cmd);

    print "Executing command $cmd on filer $opts{filer} \n";
    foreach my $line (split(/\n/, $out))
    {
      print "The sensors value line : $line \n";
    }


} #End of Gx_envsensors

1;
