# -*- perl -*-
# $Id: $
#
#	Copyright (c) 2011 NetApp.
#	All Rights reserved
##
## @summary	This module provides common functions for firmware version checks 
##         	on the system.
##
## @description
## FWversionsUtils.pm is a common module for FW version check utilities 
## 
## Sub-Routines:
## 1. Sysconfig_v 
##	-	Returns the 'sysconfig -v' output from ONTAP CLI.
## 2. ONTAP_Ver 
##	-	Returns the ONTAP version from 'version' command output on ONTAP CLI.
## 3. BIOS_Ver 
##	-	Returns the BIOS version from 'sysconfig -v' output on ONTAP CLI.
## 4. Service_Processor_Ver 
##	-	Returns the Service Processor version from the output of 
##	 	'sysconfig -v' output on ONTAP CLI..
## 5. Sasadmin_Shelf 
##	-	Returns the shelf adapters from the output of "sasadmin shelf" 
##	 	command on ONTAP CLI.
## 6. Canister_VPD_structure 
##	-	Returns the "Canister VPD structure" version from the output of 
##	 	'sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
##	.	<command>
## 7. Canister_CPLD 
##	-	Returns the "Canister CPLD" version from the output of 
##	 	'sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
##	.	<command>
## 8. Midplane_VPD_Structure 
##	-	Returns the "Midplane VPD Structure" version from the output of 
##	 	'sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
##	.	<command>
## 9. Midplane_CPLD 
##	-	Returns the "Midplane CPLD" version from the output of 
##	 	'sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
##	.	<command>
## 10. Power_Supply_Firmware 
##	-	Returns the "Power Supply Firmware" version from the output of 
##	 	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
##	 	<command>"
## 11. SP_Info 
##	-	Returns SP information from teh output of SP status command on ONTAP CLI.
## 12. Trim 
##	-	Function to remove whitespace from the start and end of the string.
## 13. Ltrim 
##	-	Left trim function to remove leading whitespace.
## 14. Rtrim 
##	-	Right trim function to remove trailing whitespace.
##
## @status private
##
## @author sandeepk@netapp.com, dl-platform-qa@netapp.com
##
###############################################################################

package Platform::FWversionsUtils;

use strict;
use warnings;
use POSIX;
use Shell;
use Tharn;
use BTS_Config;
use Params::Validate qw(validate OBJECT SCALAR HASHREF ARRAYREF UNDEF);
use SPARSE::Util::Logger qw(logStep logWarn);
use Storage::Sasadmin qw (sasadmin_expander_cli);
use SayWrap qw(sayEOF);
use Server;
use LogReader;

require Exporter;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

our %EXPORT_TAGS = ( 'all' => [ qw(

) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our $gSP_prompt = "SP.*>";
my $conn;

our @EXPORT = qw(
	$gSP_prompt 
	Sysconfig_v 
	ONTAP_Ver 
	BIOS_Ver 
	Service_Processor_Ver 
	Sasadmin_Shelf 
	Canister_VPD_structure 
	Canister_CPLD 
	Midplane_VPD_Structure 
	Midplane_CPLD 
	Power_Supply_Firmware 
	SP_Info 
	Trim 
	Ltrim 
	Rtrim 
);

#---------------------------------------------------------------------
# Package Version
#---------------------------------------------------------------------
our $VERSION = '1.0';
BEGIN {
    # Die with a script error when an input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#############################     Sub-routines     #############################

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the 'sysconfig -v' output from ONTAP CLI.
# INPUT    :	$_[0] = Filer Connection handle, $_[1] = Filer Mode.
# RETURN   :	output of 'sysconfig -v' command
#-------------------------------------------------------------------------------
sub Sysconfig_v(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};

	my $response;
	my $cmd = "sysconfig -v";

	if($mode eq '7mode') {
		(undef,undef,$response) = say ($conn, $cmd, -swallow, -timeout, 100);
	} elsif ($mode eq '10mode') {
		$cmd = "run local " .$cmd;
		$response = $conn->cli(cmd => $cmd);
# 		$response = $conn->cli(cmd => $cmd, additional_prompts_timeout => "60",
# 		              additional_prompts => {"::.?>" => "FINAL_PROMPT",});
	}
	return $response;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the ONTAP version from 'version' command output on ONTAP CLI.
# INPUT    :	$_[0] = Filer Connection handle, $_[1] = Filer Mode.
# RETURN   :	ONTAP version from "version" command output
#-------------------------------------------------------------------------------
sub ONTAP_Ver(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
		}
	);
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my ($response,@tmp_out_ver_arr,$ret_ontap_ver);
	my $cmd = "version";
	
	if($mode eq '7mode') {
		(undef,undef,$response) = say ($conn, $cmd, -swallow, -timeout, 100);
	} elsif ($mode eq '10mode') {
		$cmd = "run local " .$cmd;
		$response = $conn->cli(cmd => $cmd);
# 		$response = $conn->cli(cmd => $cmd, additional_prompts_timeout => "60",
# 		              additional_prompts => {"::.?>" => "FINAL_PROMPT",});
	}
	logStep ("ONTAP_Version response - *$response*");
	$response =~ /Release\s{1,}/i;
	@tmp_out_ver_arr = split (/\s{1,}/,$');
	$ret_ontap_ver = $tmp_out_ver_arr[0];
	
	return $ret_ontap_ver;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the BIOS version from 'sysconfig -v' output on ONTAP CLI.
# INPUT    :	$_[0] = Filer Connection handle, $_[1] = Filer Mode.
# RETURN   :	BIOS version from "sysconfig -v" command output
#-------------------------------------------------------------------------------
sub BIOS_Ver(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => {type=>OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my ($sysconfig_output,@bios_out_arr,$bios_ver);

	$sysconfig_output = Sysconfig_v(conn       => $conn,
	                                mode       => $mode);

	$sysconfig_output =~ /BIOS version:\s{1,}/;
	@bios_out_arr = split (/\n/,$');
	$bios_ver = $bios_out_arr[0];
	chomp($bios_ver);
	return $bios_ver;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the Service Processor version from the output of 
#           	'sysconfig -v' command on ONTAP CLI.
# INPUT    :	$_[0] = Filer Connection handle, $_[1] = Filer Mode.
# RETURN   :	BIOS version from "sysconfig -v" command output
#-------------------------------------------------------------------------------
sub Service_Processor_Ver(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my ($sysconfig_output,@sys_out_arr,$sp_info,@sp_split_op,$sp_ver);

	$sysconfig_output = Sysconfig_v(conn       => $conn,
	                                mode       => $mode);

	if ($sysconfig_output =~ (/Service Processor\s{1,}/) || 
		$sysconfig_output =~ (/Remote LAN Module\s{1,}/) || 
		$sysconfig_output =~ (/Baseboard Management Controller/)) {
		@sys_out_arr = split (/\n/,$');
	}
	foreach (@sys_out_arr) {
		if ($_ =~ /Firmware Version:\s{1,}/) {
			@sp_split_op = split (/\n/,$');
			$sp_ver = $sp_split_op[0];
		}
	}
	return $sp_ver;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the shelf adapters from the output of "sasadmin shelf" 
#           	command on ONTAP CLI.
# INPUT    :	$_[0] = Filer Connection handle, $_[1] = Filer Mode.
# RETURN   :	Shelf adapter names from "sasadmin shelf" command output
#-------------------------------------------------------------------------------
sub Sasadmin_Shelf(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $cmd = "sasadmin shelf";
	my (@shelf_adapters,@out_lines,$response,$adapter);
	
	if($mode eq '7mode') {
# 		$response = sayEOF($conn, $cmd);
# 		$response = $conn->console(cmd => $cmd);
	(undef,undef,$response) = say ($conn, $cmd, -swallow, -timeout, 100);
	} elsif ($mode eq '10mode') {
		$cmd = "run local " .$cmd;
		$response = $conn->cli(cmd => $cmd, additional_prompts_timeout => "60",
		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
	}
	@out_lines = split(/\n/,$response);
	for (my $i=0;$i<$#out_lines;$i++) {
		if (($out_lines[$i] =~ /Expanders on channel\s/i) 
				&& ($out_lines[$i+1] !~ /Cannot Complete operation/i)) {
			my @split_line = split(/Expanders on channel\s/i, $out_lines[$i]);
			$adapter = $split_line[1];
			chop($adapter);
			push (@shelf_adapters,$adapter);
		}
	}
	return @shelf_adapters;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Canister VPD structure" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer host object(for 7mode), Filer Connection handle
#           	(for Cmode), $_[1] = Filer Mode,$_[2] = Adapter name,
#           	$_[3] = Shelf ID.
# RETURN   :	Canister VPD structure version.
#-------------------------------------------------------------------------------
sub Canister_VPD_structure(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my (@sas_out_arr,$output);
	
	if ($mode eq '7mode') {
		my $cmd_7mode = "priv set -q diag; sasadmin expander_cli $adapter.$shelf_id ver";
		(undef,undef,$output) = say ($conn, $cmd_7mode, -swallow, -timeout, 100);
# 	$output = sasadmin_expander_cli(host=>$conn,
# 	                               adapter_name=>$adapter,
# 	                               shelf_id=>$shelf_id,
# 	                               command=>"ver",
# 	                               return_data=>1);
	} elsif ($mode eq '10mode') {
		my $cmd_Cmode = "run local \"priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver\"";
		#logStep ("\nCanister_VPD_structure command:\n$cmd_Cmode");
		$output = $conn->cli(cmd => "$cmd_Cmode", additional_prompts_timeout => "60",
		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
	}
	#logStep("LIB: Canister_VPD_structure - output - $output");
	$output =~ /Canister VPD structure\s{1,}:\s{1,}/i;
	@sas_out_arr = split (/\n/,$');
	my $can_vpd_struct = $sas_out_arr[0];
	return $can_vpd_struct;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Canister CPLD" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer host object(for 7mode), Filer Connection handle
#           	(for Cmode), $_[1] = Filer Mode,$_[2] = Adapter name,
#           	$_[3] = Shelf ID.
# RETURN   :	Canister CPLD version.
#-------------------------------------------------------------------------------
sub Canister_CPLD(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my (@can_cpld_out_arr,$output);
	
	if ($mode eq '7mode') {
		my $cmd_7mode = "priv set -q diag; sasadmin expander_cli $adapter.$shelf_id ver";
		(undef,undef,$output) = say ($conn, $cmd_7mode, -swallow, -prompt, -timeout, 100);
	} elsif ($mode eq '10mode') {
		my $cmd_Cmode = "run local \"priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver\"";
# 		logStep ("\Canister_CPLD command:\n$cmd_Cmode");
		$output = $conn->cli(cmd => "$cmd_Cmode", additional_prompts_timeout => "60",
		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
	}
	#logStep ("LIB: Canister_CPLD output - $output");

	$output =~ /Canister CPLD\s{1,}:\s{1,}/i;
	@can_cpld_out_arr = split (/\n/,$');
	my $can_cpld = $can_cpld_out_arr[0];
	return $can_cpld;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Midplane VPD Structure" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer host object(for 7mode), Filer Connection handle
#           	(for Cmode), $_[1] = Filer Mode,$_[2] = Adapter name,
#           	$_[3] = Shelf ID.
# RETURN   :	Midplane VPD Structure version.
#-------------------------------------------------------------------------------
sub Midplane_VPD_Structure(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my ($output,@mid_out_arr);
	
	if ($mode eq '7mode') {
		my $cmd_7mode = "priv set -q diag; sasadmin expander_cli $adapter.$shelf_id ver";
		(undef,$output) = say ($conn, $cmd_7mode,"PCM 1 VPD CRC", -swallow, -timeout, 100);
	} elsif ($mode eq '10mode') {
		my $cmd_Cmode = "run local \"priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver\"";
		$output = $conn->cli(cmd => "$cmd_Cmode", additional_prompts_timeout => "60",
		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
	}
	#logStep ("LIB: Midplane_VPD_Structure output - $output");

	$output =~ /Midplane VPD Structure\s{1,}:\s{1,}/i;
	@mid_out_arr = split (/\n/,$');
	my $midplane_vpd_struct = $mid_out_arr[0];
	return $midplane_vpd_struct;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Midplane CPLD" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer host object(for 7mode), Filer Connection handle
#           	(for Cmode), $_[1] = Filer Mode,$_[2] = Adapter name,
#           	$_[3] = Shelf ID.
# RETURN   :	Midplane CPLD version
#-------------------------------------------------------------------------------
sub Midplane_CPLD(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my ($output,@m_cpld_out_arr);
	
	if ($mode eq '7mode') {
		my $cmd_7mode = "priv set -q diag; sasadmin expander_cli $adapter.$shelf_id ver";
		(undef,$output) = say ($conn, $cmd_7mode,"PCM 1 VPD CRC", -swallow, -timeout, 100);
	} elsif ($mode eq '10mode') {
		my $cmd_Cmode = "run local \"priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver\"";
		$output = $conn->cli(cmd => "$cmd_Cmode", additional_prompts_timeout => "60",
		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
	}
	#logStep ("LIB: Midplane_CPLD output - $output");

	$output =~ /Midplane CPLD\s{1,}:\s{1,}/i;
	@m_cpld_out_arr = split (/\n/,$');
	my $midplane_cpld = $m_cpld_out_arr[0];
	return $midplane_cpld;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Power Supply Firmware" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer host object(for 7mode), Filer Connection handle
#           	(for Cmode), $_[1] = Filer Mode,$_[2] = Adapter name,
#           	$_[3] = Shelf ID.
# RETURN   :	PCM 1 and PCM 2 FWs as an array
#-------------------------------------------------------------------------------
sub Power_Supply_Firmware(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my ($output,@psu_out_arr);
	
	if ($mode eq '7mode') {
		my $cmd_7mode = "priv set -q diag; sasadmin expander_cli $adapter.$shelf_id ver";
		(undef,undef,$output) = say ($conn, $cmd_7mode, -swallow, -prompt, -timeout, 100);
# 		$output = sasadmin_expander_cli(host=>$conn,
# 		                           adapter_name=>$adapter,
# 		                           shelf_id=>$shelf_id,
# 		                           command=>"ver",
# 		                           return_data=>1);
	} elsif ($mode eq '10mode') {
		my $cmd_Cmode = "run local \"priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver\"";
		$output = $conn->cli(cmd => "$cmd_Cmode", additional_prompts_timeout => "60",
		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
	}
# 	logStep ("Power_Supply_Firmware output - $output");

	my @pcm_fw;
	my ($pcm1,$pcm2,$pcm3,$pcm4);
	my (@out_arr1,@out_arr2,@out_arr3,@out_arr4);
	if ($output =~ /PCM 1 firmware\s{1,}:\s{1,}/i) {
		$pcm1 = $';
		@out_arr1 = split (/\n/,$pcm1);
		push (@pcm_fw,$out_arr1[0]);
	}
	if ($output =~ /PCM 2 firmware\s{1,}:\s{1,}/i) {
		$pcm2 = $';
		@out_arr2 = split (/\n/,$pcm2);
		push (@pcm_fw,$out_arr2[0]);
	}
	if ($output =~ /PCM 3 firmware\s{1,}:\s{1,}/i) {
		$pcm3 = $';
		@out_arr3 = split (/\n/,$pcm3);
		push (@pcm_fw,$out_arr3[0]);
	}
	if ($output =~ /PCM 4 firmware\s{1,}:\s{1,}/i) {
		$pcm4 = $';
		@out_arr4 = split (/\n/,$pcm4);
		push (@pcm_fw,$out_arr4[0]);
	}
	return @pcm_fw;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns SP information from the output of SP status command on 
#           	ONTAP CLI.
# INPUT    :	$_[0] = Filer Connection handle, $_[1] = Filer Mode.
#           	$_[2] = Service processor type ('sp','rlm','bmc').
# RETURN   :	Informations on SP in the form of hash
#-------------------------------------------------------------------------------
sub SP_Info(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			processor_type  => { type => SCALAR, default => 'sp'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $proc_type = $opts{processor_type};
	my $cmd;
	if ($proc_type =~ /sp/i) {
		$cmd = "sp status";
	} elsif ($proc_type =~ /rlm/i) {
		$cmd = "rlm status";
	} if ($proc_type =~ /bmc/i) {
		$cmd = "bmc status";
	}
	my ($response,%sp_info);
	if($mode eq '7mode') {
# 		$response = sayEOF($conn, $cmd);
# 		$response = $conn->console(cmd => $cmd);
	(undef,undef,$response) = say ($conn, $cmd, -swallow, -timeout, 100);
	} elsif ($mode eq '10mode') {
		$cmd = "run local " .$cmd;
		$response = $conn->cli(cmd => $cmd, additional_prompts_timeout => "60",
		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
	}
	
	if ($proc_type =~ /bmc/i) {
		$response =~ /IP address:\s{1,}/i;
		my @tmp_out_arr_4 = split (/\n/,$');
		$sp_info{'ipv4_addr'} = $tmp_out_arr_4[0];
	} else {
		$response =~ /IPv4 configuration:\s{1,}/i;
		my @tmp_out_arr_1 = split (/\n/,$');
		my @tmp_out_arr_2 = split (/:\s{1,}/,$tmp_out_arr_1[0]);
		my @tmp_out_arr_3 = split (/\n/,$tmp_out_arr_2[1]);
		$sp_info{'ipv4_addr'} = $tmp_out_arr_3[0];
	}
	chomp ($sp_info{'ipv4_addr'});
	return %sp_info;
}

# Function to remove whitespace from the start and end of the string
sub Trim($)
{
	my $string = shift;
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}

# Left trim function to remove leading whitespace
sub Ltrim($)
{
	my $string = shift;
	$string =~ s/^\s+//;
	return $string;
}

# Right trim function to remove trailing whitespace
sub Rtrim($)
{
	my $string = shift;
	$string =~ s/\s+$//;
	return $string;
}

1;
