# -*- perl -*-
# $Id: $
#
#	Copyright (c) 2011 NetApp.
#	All Rights reserved
##
## @summary	This module provides common functions for firmware version checks 
##         	on the system.
##
## @description
## FWversionsUtils.pm is a common module for FW version check utilities 
## 
## Sub-Routines:
## 1. Sasadmin_Shelf 
##	-	Returns the shelf adapters from the output of "sasadmin shelf" 
##	 	command on ONTAP CLI.
## 2. Canister_VPD_structure 
##	-	Returns the "Canister VPD structure" version from the output of 
##	 	'sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
##	.	<command>
## 3. Canister_CPLD 
##	-	Returns the "Canister CPLD" version from the output of 
##	 	'sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
##	.	<command>
## 4. Midplane_VPD_Structure 
##	-	Returns the "Midplane VPD Structure" version from the output of 
##	 	'sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
##	.	<command>
## 5. Midplane_CPLD 
##	-	Returns the "Midplane CPLD" version from the output of 
##	 	'sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
##	.	<command>
## 6. Power_Supply_Firmware 
##	-	Returns the "Power Supply Firmware" version from the output of 
##	 	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
##	 	<command>"
##
## @status private
##
## @author sandeepk@netapp.com, dl-platform-qa@netapp.com
##
###############################################################################

package Platform::FWversionsUtils;

use strict;
use warnings;
use POSIX;
use Shell;
use Tharn;
use BTS_Config;
use Params::Validate qw(validate OBJECT SCALAR HASHREF ARRAYREF UNDEF);
use SPARSE::Util::Logger qw(logStep logWarn);
use Storage::Sasadmin qw (sasadmin_expander_cli);
use SayWrap qw(sayEOF);
use Server;
use LogReader;

require Exporter;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

our %EXPORT_TAGS = ( 'all' => [ qw(

) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(
	Sasadmin_Shelf 
	Canister_VPD_structure 
	Canister_CPLD 
	Midplane_VPD_Structure 
	Midplane_CPLD 
	Power_Supply_Firmware 
);

#---------------------------------------------------------------------
# Package Version
#---------------------------------------------------------------------
our $VERSION = '1.0';
BEGIN {
    # Die with a script error when an input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

#############################     Sub-routines     #############################

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the shelf adapters from the output of "sasadmin shelf" 
#           	command on ONTAP CLI.
# INPUT    :	$_[0] = Filer Connection handle, $_[1] = Filer Mode.
# RETURN   :	Shelf adapter names from "sasadmin shelf" command output
#-------------------------------------------------------------------------------
sub Sasadmin_Shelf(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $cmd = "sasadmin shelf";
	my (@shelf_adapters,@out_lines,$response,$adapter);
	
	$response = $conn->execute_command(command => $cmd);
	@out_lines = split(/\n/,$response);
	for (my $i=0;$i<$#out_lines;$i++) {
		if (($out_lines[$i] =~ /Expanders on channel\s/i) 
				&& ($out_lines[$i+1] !~ /Cannot Complete operation/i)) {
			my @split_line = split(/Expanders on channel\s/i, $out_lines[$i]);
			$adapter = $split_line[1];
			chop($adapter);
			push (@shelf_adapters,$adapter);
		}
	}
	return @shelf_adapters;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Canister firmware" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer Connection object $_[1] = Filer Mode, 
#           	$_[2] = Adapter name, $_[3] = Shelf ID.
# RETURN   :	Canister firmware version.
#-------------------------------------------------------------------------------
sub Canister_firmware(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my (@sas_out_arr,$op);
	
	my $cmd = "priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver";
	
	$op = $conn->execute_command(command => $cmd);
	$op =~ /Canister firmware\s{1,}:\s{1,}/i;
	@sas_out_arr = split (/\n/,$');
	my $can_fw = $sas_out_arr[0];
	return $can_fw;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Canister VPD structure" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer Connection object $_[1] = Filer Mode, 
#           	$_[2] = Adapter name, $_[3] = Shelf ID.
# RETURN   :	Canister VPD structure version.
#-------------------------------------------------------------------------------
sub Canister_VPD_structure(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my (@sas_out_arr,$op);
# 	my $cmd = "run local \"priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver\"";
# 	
# 	if ($mode eq '7mode') {
# 	$op = sasadmin_expander_cli(host=>$conn,
# 	                               adapter_name=>$adapter,
# 	                               shelf_id=>$shelf_id,
# 	                               command=>"ver",
# 	                               return_data=>1);
# 	} elsif ($mode eq '10mode') {
# 		#logStep ("\nCanister_VPD_structure command:\n$cmd");
# 		$op = $conn->cli(cmd => "$cmd", additional_prompts_timeout => "60",
# 		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
# 	}
# 	#logStep("Canister_VPD_structure - output - $op");
	
	my $cmd = "priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver";
	
	$op = $conn->execute_command(command => $cmd);
	$op =~ /Canister VPD structure\s{1,}:\s{1,}/i;
	@sas_out_arr = split (/\n/,$');
	my $can_vpd_struct = $sas_out_arr[0];
	return $can_vpd_struct;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Canister CPLD" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer Connection object $_[1] = Filer Mode, 
#           	$_[2] = Adapter name, $_[3] = Shelf ID.
# RETURN   :	Canister CPLD version.
#-------------------------------------------------------------------------------
sub Canister_CPLD(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my (@can_cpld_out_arr,$output);
# 	my $cmd = "run local \"priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver\"";
# 	
# 	if ($mode eq '7mode') {
# 		$output = sasadmin_expander_cli(host=>$conn,
# 		                           adapter_name=>$adapter,
# 		                           shelf_id=>$shelf_id,
# 		                           command=>"ver",
# 		                           return_data=>1);
# 	} elsif ($mode eq '10mode') {
# # 		logStep ("\Canister_CPLD command:\n$cmd");
# 		$output = $conn->cli(cmd => "$cmd", additional_prompts_timeout => "60",
# 		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
# 	}
# # 	logStep ("Canister_CPLD output - $output");

	my $cmd = "priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver";
	
	$output = $conn->execute_command(command => $cmd);
	$output =~ /Canister CPLD\s{1,}:\s{1,}/i;
	@can_cpld_out_arr = split (/\n/,$');
	my $can_cpld = $can_cpld_out_arr[0];
	return $can_cpld;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Midplane VPD Structure" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer Connection object $_[1] = Filer Mode, 
#           	$_[2] = Adapter name, $_[3] = Shelf ID.
# RETURN   :	Midplane VPD Structure version.
#-------------------------------------------------------------------------------
sub Midplane_VPD_Structure(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my ($output,@mid_out_arr);
# 	my $cmd = "run local \"priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver\"";
# 	
# 	if ($mode eq '7mode') {
# 		$output = sasadmin_expander_cli(host=>$conn,
# 		                           adapter_name=>$adapter,
# 		                           shelf_id=>$shelf_id,
# 		                           command=>"ver",
# 		                           return_data=>1);
# 	} elsif ($mode eq '10mode') {
# 		$output = $conn->cli(cmd => "$cmd", additional_prompts_timeout => "60",
# 		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
# 	}
# # 	logStep ("Midplane_VPD_Structure output - $output");

	my $cmd = "priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver";
	
	$output = $conn->execute_command(command => $cmd);
	$output =~ /Midplane VPD Structure\s{1,}:\s{1,}/i;
	@mid_out_arr = split (/\n/,$');
	my $midplane_vpd_struct = $mid_out_arr[0];
	return $midplane_vpd_struct;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Midplane CPLD" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer Connection object $_[1] = Filer Mode, 
#           	$_[2] = Adapter name, $_[3] = Shelf ID.
# RETURN   :	Midplane CPLD version
#-------------------------------------------------------------------------------
sub Midplane_CPLD(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my ($output,@m_cpld_out_arr);
# 	my $cmd = "run local \"priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver\"";
# 	
# 	if ($mode eq '7mode') {
# 		$output = sasadmin_expander_cli(host=>$conn,
# 		                           adapter_name=>$adapter,
# 		                           shelf_id=>$shelf_id,
# 		                           command=>"ver",
# 		                           return_data=>1);
# 	} elsif ($mode eq '10mode') {
# 		$output = $conn->cli(cmd => "$cmd", additional_prompts_timeout => "60",
# 		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
# 	}
# # 	logStep ("Midplane_CPLD output - $output");

	my $cmd = "priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver";
	
	$output = $conn->execute_command(command => $cmd);
	$output =~ /Midplane CPLD\s{1,}:\s{1,}/i;
	@m_cpld_out_arr = split (/\n/,$');
	my $midplane_cpld = $m_cpld_out_arr[0];
	return $midplane_cpld;
}

#-------------------------------------------------------------------------------
# FUNCTION :	Returns the "Power Supply Firmware" version from the output of 
#           	"sasadmin expander_cli <adapter_name>.<shelf_id>[.<module id>] 
#           	<command>"
# INPUT    :	$_[0] = Filer Connection object $_[1] = Filer Mode, 
#           	$_[2] = Adapter name, $_[3] = Shelf ID.
# RETURN   :	PCM 1 and PCM 2 FWs as an array
#-------------------------------------------------------------------------------
sub Power_Supply_Firmware(%;)
{
	my (%opts) = validate (
		@_,
		{
			conn       => { type => OBJECT },
			mode       => { type => SCALAR, default => '7mode'},
			adapter    => { type => SCALAR, default => '0a'},
			shelf_id   => { type => SCALAR, default => '0'},
		}
	);
	
	my $conn = $opts{conn};
	my $mode = $opts{mode};
	my $adapter = $opts{adapter};
	my $shelf_id = $opts{shelf_id};
	my ($output,@psu_out_arr);
# 	my $cmd = "run local \"priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver\"";
# 	
# 	if ($mode eq '7mode') {
# 		$output = sasadmin_expander_cli(host=>$conn,
# 		                           adapter_name=>$adapter,
# 		                           shelf_id=>$shelf_id,
# 		                           command=>"ver",
# 		                           return_data=>1);
# 	} elsif ($mode eq '10mode') {
# 		$output = $conn->cli(cmd => "$cmd", additional_prompts_timeout => "60",
# 		              additional_prompts => {"\\*?>" => "FINAL_PROMPT",});
# 	}
# # 	logStep ("Power_Supply_Firmware output - $output");

	my $cmd = "priv set -q diag;sasadmin expander_cli $adapter.$shelf_id ver";
	
	$output = $conn->execute_command(command => $cmd);
	my @pcm_fw;
	my ($pcm1,$pcm2,$pcm3,$pcm4);
	my (@out_arr1,@out_arr2,@out_arr3,@out_arr4);
	if ($output =~ /PCM 1 firmware\s{1,}:\s{1,}/i) {
		$pcm1 = $';
		@out_arr1 = split (/\n/,$pcm1);
		push (@pcm_fw,$out_arr1[0]);
	}
	if ($output =~ /PCM 2 firmware\s{1,}:\s{1,}/i) {
		$pcm2 = $';
		@out_arr2 = split (/\n/,$pcm2);
		push (@pcm_fw,$out_arr2[0]);
	}
	if ($output =~ /PCM 3 firmware\s{1,}:\s{1,}/i) {
		$pcm3 = $';
		@out_arr3 = split (/\n/,$pcm3);
		push (@pcm_fw,$out_arr3[0]);
	}
	if ($output =~ /PCM 4 firmware\s{1,}:\s{1,}/i) {
		$pcm4 = $';
		@out_arr4 = split (/\n/,$pcm4);
		push (@pcm_fw,$out_arr4[0]);
	}
	return @pcm_fw;
}

1;
