# Copyright (c) 2006 Network Appliance, Inc.
# All rights reserved.
#
# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/Platform/PlatUtils.pm#1 $
## @summary Platform Utilities
##
## @description
##    This package defines some common routines used by some of the platform
##    automation.
##
##    All the subroutines here expect a filer "#console" connection as their
##    first argument.
##
##    The routines are:
##
##       nvram_jettison - 		Jettision NVMEM contents using platform_error
##       watchdog_hang - 		Perform a watchdog hard-hang from Ontap
##       reboot_ontap - 		Reboot from Ontap
##       chassis_power_cycle - 		Power cycle the chassis
##       panic - 			Panic FILER
##       random - 			Return a random integer between start and end
##       timecheck - 			Compare FILER date/time to that of CLIENT
##       date_time_to_epoch - 		Convert a time in Ontap "date" format (i.e. "Mon Dec  4 15:32:01 EST 2006") to
##                     			  the number of seconds since "Jan 1, 1970 00:00:00", makes time comparisons 
## 					  easier.
##       start_internal_hammer - 
##
##         Usage:  'start_internal_hammer <spec1>' or 'start_internal_hammer <spec2>'
##           Where spec1 allows volume and blocksize specification for each thread:
##             -vol<N>=>"volname"       Names of up to 5 volumes to start hammer threads on (i.e. -vol1 thru -vol5)
##             -blk<N>=>blocksize       Blocksize for each thread
##   
##             Example (start 3 threads; one on tvol4, 2 on tvol5 with different blocksizes):
##               start_internal_hammer -vol1=>tvol4, -blk1=>1000, -vol2=>tvol5, -blk2=>10000, -vol3=>tvol5, -blk1=>100
##
##           Or where spec2 allows up to 5 threads created on a single volume and blocksize
##             -vol=>"volname"          Names of the volume on which you want to start up to 5 hammer threads.
##             -blk=>blocksize          Blocksize for all threads to be started.
##             -threads=>threads        Number of threads to create
##  
##             Example (start 5 threads on tvol4 with a blocksize of 10,000):
##               start_internal_hammer -vol=>tvol4, -blk=>10000, -threads=>5
##
##      stop_internal_hammer -          Abort all internal hammer threads
##      internal_hammer_running -       Is internal hammer running?
##      dump_sys_config_info    -       Function to dump system configuration information.
##
##
## @dependencies
## The functions depend on exact output format to parse and extract results.
## Even a minor change in the UI can result in these functions not working properly.
##
## @bugs
##

package Platform::PlatUtils;
require Exporter;

our @ISA 	= qw(Exporter);
our @EXPORT 	= qw(nvram_jettison watchdog_hang wait_for_ontap_up reboot_ontap chassis_power_cycle panic random timecheck date_time_to_epoch start_internal_hammer start_fill_internal_hammer stop_internal_hammer internal_hammer_running dump_sys_config_info);
our @EXPORT_OK 	= qw();
our $VERSION 	= 1.00;

use strict;
use Tharn;
use NATE::Result;
use boot;
use Platform::BmcUtils;

##       nvram_jettison - 		Jettision NVMEM contents using platform_error
sub nvram_jettison {
  my ($fc, %options) = @_;
  my $login = ($options{-login}) ? $options{-login} : "yes";

  say($fc, "priv set test; platform_error nvram jettison", "-noprompt", "-timeout", 120 );
  sleep 2;
  say($fc, "y\r", "-nocr", "-noprompt", "-timeout", 600, "-wait", "previous boot of system marked nvlog to be DISCARDED" );

  if ( $login =~ m/yes/i ) {
    login2ontap($fc, -timeout, 600);
  }
}

sub wait_for_ontap_up {
  my ($fc, %options) = @_;
  my $timeout = ($options{-timeout}) ? $options{-timeout} : "600";

  my $begtime = time;

  my $finished=0;
  while ( ((time-$begtime) < $timeout) && ($finished == 0) ) {

    sleep 10;

    my $state = boot_state( $fc, transient=>1 );

#print "\nSTATE = |$state|\n";

    if ( $state =~ /USERNAME|PASSWORD|UP/ ) {
      $finished = 1;
    }
  }

  return $finished;
}

##       watchdog_hang - 		Perform a watchdog hard-hang from Ontap
sub watchdog_hang {
  my ($fc, %options) = @_;
  my $login = ($options{-login}) ? $options{-login} : "yes";

  say($fc, "priv set test; watchdog hard_hang", "-noprompt" );

  my $finished = wait_for_ontap_up( $fc, -timeout=>600 );

  if ( $finished == 0 ) {
     tstlog(-fail, "ERROR: Filer did not come up to Ontap after a watchdog hard hang");
  }

  if ( $login =~ m/yes/i ) {
    login2ontap($fc, -timeout, 600);
  }
}

##       reboot_ontap - 		Reboot from Ontap
sub reboot_ontap {
  my ($fc, %options) = @_;
  my $login = ($options{-login}) ? $options{-login} : "yes";

  say($fc, "reboot -t 0", -noprompt );

  my $finished = wait_for_ontap_up( $fc, -timeout=>600 );

  if ( $finished == 0 ) {
     tstlog(-fail, "ERROR: Filer did not come up to Ontap after a reboot");
  }

  if ( $login =~ m/yes/i ) {
    login2ontap($fc, -timeout, 600);
  }
}

##       chassis_power_cycle - 		Power cycle the chassis
sub chassis_power_cycle {
  my ($fc, $cc, %options) = @_;
  my $login = ($options{-login}) ? $options{-login} : "yes";
  my $fh = $fc->{hostp};

  my $cmd = "/usr/local/bin/power_cycle $fh->{hostname} reboot";
  if ( $fh->{powercmd} ne "" ) {
    $cmd = $fh->{powercmd};
  }

  say($cc, "$cmd", -timeout, 120);

  if ( $login =~ m/yes/i ) {
    login2ontap($fc, -timeout, 600);
  }
}

##       panic - 			Panic FILER
sub panic {
  my ($fc, %options) = @_;
  my $login = ($options{-login}) ? $options{-login} : "yes";

  say($fc, "priv set test; panic", "-noprompt" );

  my $finished = wait_for_ontap_up( $fc, -timeout=>600 );

  if ( $finished == 0 ) {
     tstlog(-fail, "ERROR: Filer did not come up to Ontap after a panic");
  }

  if ( $login =~ m/yes/i ) {
    login2ontap($fc, -timeout, 600);
  }
}

##       timecheck - 			Compare FILER date/time to that of CLIENT
sub timecheck {
  my ($fc, $cc, %options) = @_;
  logcomment("timecheck start .....");
  my $filerTimeAdjust;
  my $hostTimeAdjust;
  my ($t1, $t2, $t3, $elapsedFiler, $elapsedHost);
  $t1 = time;
  my (undef, undef, $ftime, undef) = say($fc, "date", -timeout, 300, -swallow);
  $t2 = time;
  my (undef, undef, $ctime, undef) = say($cc, "date", -timeout, 300, -swallow);
  $t3 = time;

  $ftime = (split(/\n/, $ftime))[1];
  chop($ftime); chop($ftime);
  chop($ctime); chop($ctime);

  $elapsedFiler = $t2 - $t1;
  $elapsedHost  = $t3 - $t2;

  my $nftime = date_time_to_epoch( $ftime );
  my $nctime = date_time_to_epoch( $ctime );

  my $diff = abs( $nftime - $nctime );

  # Assume that time returned by command is at the very close to the point where "say" returns.
  # then $nctime should subtract off $elapsedHost

  $nctime = $nctime - $elapsedHost;

  $diff = abs( $nftime - $nctime );

  if ( $diff > 60 ) {
    logcomment("FAIL: Filer time |$ftime| is $diff (> 60) seconds out of sync with client time |$ctime|");
    return 0;
  } else {
    logcomment("PASS: Filer time |$ftime| is $diff (> 60) seconds out of sync with client time |$ctime|");
    return 1;
  }
}

# options vol, blocks
# command would look like "start_internal_hammer $fc, -vol1=>vol0, -blk1=>1000, -vol2=>tvol4, -vol3=>vol0, -blk3=>10000, -vol4=>tvol4, -vol5=>tvol5, -blk5=>1000"
# command would look like "start_internal_hammer $fc, -vol=>tvol4, -threads=>5, -blk=>10000"

sub start_internal_hammer {
  my ($fc, %options) = @_;
  my $threads = 5;
  my $blocksize = 100000;
  my $numthreads;
  my $filename;
  my $vol;
  my $blk;

  logcomment("Starting Internal Hammer");

  if ( $options{-vol} ) {
    $vol = $options{-vol};
    $blk = ($options{-blk}) ? $options{-blk} : $blocksize;
    $numthreads = ($options{-threads}) ? $options{-threads} : $threads;

    for ( my $i = 0; $i < $numthreads; $i++) {
      $filename = "hammer.$i";
      $vol = $options{-vol};
      $blk = ($options{-blk}) ? $options{-blk} : $blocksize;
      say( $fc, "priv set -q diag; hammer -f -1 /vol/$vol/$filename $blk", "-timeout", 120);
    }

  } elsif ( $options{-vol1}  ) {
    for ( my $i = 1; $i <= $threads; $i++) {
      if ( $options{"-vol${i}"}  ) {
        $vol = $options{"-vol${i}"};
        $blk = ($options{"-blk${i}"}) ? $options{"-blk${i}"} : $blocksize;
        $filename = "hammer.$i";
        say( $fc, "priv set -q diag; hammer -f -1 /vol/$vol/$filename $blk", "-timeout", 120);
      }
    }
  } else {
    tstlog( '-script', "start_internal_hammer must have a -vol or -vol1 option");
  }
}


sub start_fill_internal_hammer {
  my ($fc, %options) = @_;
  my $blocksize = 100000;
  my $blk = ($options{-blk}) ? $options{-blk} : $blocksize;

  logcomment("Starting Internal Hammer Fill $blk");
  say( $fc, "priv set -q diag; hammer fill $blk", "-timeout", 120);
}

sub internal_hammer_running {
  my ($fc, %options) = @_;

  boot($fc);
  my ( undef, undef, $status, undef ) = say($fc, "priv set -q diag; hammer status", "-timeout", 120, -swallow);

  if ($status =~ /blacksmith/) {
    return 1;
  } else {
    return 0;
  }

}

sub stop_internal_hammer {
  my ($fc, %options) = @_;
  my $hammer_status;
  my $wait_interval = 3;
  my $time_waited = 0;
  my $max_wait = 1800; # wait 30 minutes for hammer to abort

  boot($fc);
  my $o = say($fc, "priv set -q diag; hammer abort", "-timeout", 120);

  logcomment("Stopping Internal Hammer");

  while ( $hammer_status && $time_waited < $max_wait) {
    $hammer_status = say($fc, "priv set -q diag; hammer status", "-timeout", 120);
    $time_waited += $wait_interval;
    sleep $wait_interval;
  }

  if ( $hammer_status ) {
    tstlog( '-comment', "stop_internal_hammer failed to stop hammer threads after $time_waited seconds");
    return 0;
  } else {
    tstlog( '-comment', "stop_internal_hammer stopped hammer threads after $time_waited seconds");
    return 1;
  }
}

##       date_time_to_epoch - 		Convert a time in Ontap "date" format (i.e. "Mon Dec  4 15:32:01 EST 2006") to
##
sub date_time_to_epoch {
  my ($td, undef) = @_;

  use Date::Manip qw(ParseDate UnixDate);
  my $date = ParseDate($td);
  if (!$date) {
      return 0;
  } else {
      my $s = UnixDate($date, "%s");
      return $s;
  }
}

##       random - 			Return a random integer between start and end
#                                       "random(5)" will give a random number between 0 and 5, 
#                                       "random(5,10)" will give you random numbers between 5 and 10 inclusive.
sub random {
  my ($start, $end) = @_;

  if ( $end ) {
    my $range = $end - $start;
    my $num = (int rand( $range )) + $start;
    return $num;
  } else {
    my $num = int rand( $start );
    return $num;
  }
}

sub dump_sys_config_info {
  my ($fc, %options) = @_;
	logcomment("\t**** System Config. Information Dump Start BEGIN ****");
	logcomment("\t**** Date ****");
	say($fc, "date",       '-ok');
	logcomment("\t**** Version ****");
	say($fc, "version",       '-ok');
	logcomment("\t**** Full version info ****");
	say($fc, "version -b",       '-ok');
	logcomment("\t**** Sysconfig -av ****");
	say($fc, "sysconfig -av",  '-ok');
	logcomment("\t**** Sysconfig -M ****");
	say($fc, "sysconfig -M",  '-ok');
	logcomment("\t**** Volume status -r ****");
	say($fc, "vol status -r", '-ok');
	logcomment("\t**** Volume status -f ****");
	say($fc, "vol status -f", '-ok');
	logcomment("\t**** sasadmin expander_map ****");
	say($fc, "sasadmin expander_map",    '-ok');
	logcomment("\t**** sasadmin expander_phy_state ****");
	say($fc, "sasadmin expander_phy_state 0c",    '-ok');
	logcomment("\t**** sensor status ****");
	say($fc, "environment status chassis list-sensors",    '-ok');
	logcomment("\t**** bootfs dir ****");
	say($fc, "priv set -q diag; bootfs dir -r 0e.0:1:", '-ok');
	logcomment("\t**** bootfs chkdsk ****");
	say($fc, "priv set -q diag; bootfs chkdsk 0e.0", '-ok');
	sleep 10;
	logcomment("\t**** System Config. Information Dump Start end ****");
}

1;
