# -*- perl -*-
#/* $Id: //depot/prod/test/main/lib/Platform/bios/BIOSUtils.pm
#
#
# Copyright (c) 2013 NetApp, Inc.
# All rights reserved.
#
## @summary
## BIOSUtils.pm is a common module for all the  scripts under
##              system_firmware/bios/efi directory.
## @description
## This module privodes common functions for use in testing
##        scripts under system_firmware/bios/efi directory.
##
## Sub-routines:
##    Testharness_Status
##    Get_Model_Memory
##    Get_SRAM_Pelog_Status
##    DIMM_Num_And_Error_Address
##    Check_Ontap_Version
##    Check_BIOS_Version
##    Check_AUTOBOOT_true
##    SP_Status_LOADER
##    Get_Sys_Model
##    Get_Sys_Memory
##    Get_BIOS_Boot_Time_From_Console
##
## @status private
##
## @author dhana@netapp.com, dl-platform-qa@netapp.com
##
###############################################################################

package Platform::bios::BIOSUtils;

use strict;
use warnings;

#
# Import exceptions
#

require Exporter;
our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
    Testharness_Status
    Get_Model_Memory
    Get_SRAM_Pelog_Status
    DIMM_Num_And_Error_Address
    Check_Ontap_Version
    Check_BIOS_Version
    Check_AUTOBOOT_true
    SP_Status_LOADER
    Get_Sys_Model
    Get_Sys_Memory
    Get_BIOS_Boot_Time_From_Console
);

our $VERSION = '1.0';

#Memory size are in GB
my %memory = (
    'FAS8020' => "24",
    'FAS8040' => "32",
    'FAS8060' => "64",
    'FAS8080' => "128",
);

#Pelog SRAM fields expected hash table per model string
#For BBA, there are 4 channels per node
my %pelog = (
    'FAS8020' => {
        'dimm-1' => 'node(0) chan(0) dimm(0)',
        'dimm-2' => 'node(0) chan(1) dimm(0)',
        'dimm-3' => 'node(0) chan(2) dimm(0)',
        'dimm-4' => 'node(0) chan(3) dimm(0)',
    },
    'FAS8040' => {
        'dimm-1' => 'node(0) chan(0) dimm(0)',
        'dimm-2' => 'node(0) chan(1) dimm(0)',
        'dimm-3' => 'node(0) chan(2) dimm(0)',
        'dimm-4' => 'node(0) chan(3) dimm(0)',
    },
    'FAS8060' => {
        'dimm-1' => 'node(0) chan(0) dimm(0)',
        'dimm-2' => 'node(0) chan(1) dimm(0)',
        'dimm-3' => 'node(0) chan(2) dimm(0)',
        'dimm-4' => 'node(0) chan(3) dimm(0)',
        'dimm-5' => 'node(1) chan(0) dimm(0)',
        'dimm-4' => 'node(1) chan(1) dimm(0)',
        'dimm-4' => 'node(1) chan(2) dimm(0)',
        'dimm-4' => 'node(1) chan(3) dimm(0)',
    },
    'FAS8080' => {
        'dimm-1' => 'node(0) chan(0) dimm(0)',
        'dimm-2' => 'node(0) chan(1) dimm(0)',
        'dimm-3' => 'node(0) chan(2) dimm(0)',
        'dimm-4' => 'node(0) chan(3) dimm(0)',
        'dimm-5' => 'node(1) chan(0) dimm(0)',
        'dimm-4' => 'node(1) chan(1) dimm(0)',
        'dimm-4' => 'node(1) chan(2) dimm(0)',
        'dimm-4' => 'node(1) chan(3) dimm(0)',
    },
);

#---------------------------------------------------------------------------
# FUNCTION: This function check if the testharness -status displays enabled
#                                status for $test_point
# INPUT: Filer connection object to firmware($Firm_Obj), ($test_point) where
#                                error is injected.
# RETURN: Returns enabled or disabled status
#---------------------------------------------------------------------------
sub Testharness_Status($$) {

    my ($test_status,);
    my $status = 'unknown'; ;
    my $Firm_Obj  = $_[0];
    my $testpoint = $_[1];

    $test_status = $Firm_Obj->execute_raw_command(
        command => "priv set test; testharness -status");
    foreach my $test_status_line (split(/\n/, $test_status)) {
        if ($test_status_line !~ /^\-/) {
            my @element = split(/\|/, $test_status_line);
            if (defined($element[1])) {
                $element[1]
                    =~ s/^\s+|\s+$//g;    #remove leading & trailing spaces
                if ($element[1] eq "$testpoint") {
                    if ($element[4] !~ /Enabled/i) {
                        $status = 'disabled';
                    } else {
                        $status    = 'enabled';
                    }
                }
            }
        }    #End of if loop
    }    #End of foreach loop

    return $status;
}    ##end of sub Testharness_Status

#---------------------------------------------------------------------------
# FUNCTION: This function returns the expected memory for the model as per CCG
#
# INPUT: system model, example FAS8040, etc
#
# RETURN: Expected memory from stored %memory hash
#---------------------------------------------------------------------------
sub Get_Model_Memory($) {

    my $model = $_[0];
    my $memory = 'unknown';

    if(defined( $memory{$model} )) {
    $memory = $memory{$model};
    }

    return $memory;
}    ##end of sub Get_Model_Memory

#---------------------------------------------------------------------------
# FUNCTION: Verify SRAM fields are updated for dimm errors by validating
#               pelog output for given model and dimm
#
# INPUT: 1. Firm_Obj to connect to system, 2. System model string &
#               3. Dimm number as displayed in sysconfig -M output
# RETURN: 'enabled' when SRAM fields are enabled in pelog output
#         'disabled' when SRAM fields are not enabled in pelog output
#---------------------------------------------------------------------------
sub Get_SRAM_Pelog_Status($$$) {

    my $Firm_Obj = $_[0];
    my $model    = $_[1];
    my $dimm     = $_[2];
    my $result = 'unknown';

    my $exp_ssram = $pelog{$model}{$dimm};
    if(defined( $exp_ssram )) {

    my $sys_pelog = $Firm_Obj->execute_raw_command(
        command => "priv set test; pelog -a -g=1");

    #Append backslash for regular expression parsing
    $exp_ssram =~ s/\(/\\\(/g;
    $exp_ssram =~ s/\)/\\\)/g;

    if ($sys_pelog !~ /$exp_ssram/i) {
        $result = 'disabled';
    } elsif ($sys_pelog =~ /$exp_ssram/i) {
        $result = 'enabled';
    }
    } #End of if defined loop

    return $result;
}    ##end of sub Get_SRAM_Pelog_Status($$$)

#---------------------------------------------------------------------------
# FUNCTION: Takes DIMM string from sysconfig -M and returns DIMM number and 
#           error injection address
#   Example: dimm-2 string returns DIMM number 2 & error injection address 0x100
# INPUT: DIMM string from sysconfig -M output 
#
# RETURN: DIMM number and error injection address
#
#---------------------------------------------------------------------------
sub DIMM_Num_And_Error_Address($) {

    my $dimm = $_[0];
    my ($dimm_num, $slot_inj, $error_addr);

    if ($dimm =~ /dimm\-(\d+)/i) {
    $dimm_num = $1;
    }

    $slot_inj = $dimm_num - 1;
    $error_addr = "0x" . $slot_inj . "00";

    return ($dimm_num, $error_addr);
} ##end of sub DIMM_Num_And_Error_Address($)
#---------------------------------------------------------------------------
# FUNCTION: Parses version output and returns ontap version
#
# INPUT: Filer connection API set object
#
# RETURN: Ontap version
#
#---------------------------------------------------------------------------
sub Check_Ontap_Version($) {

    my $Api_Set_Obj = $_[0];
    my ($ver_raw, @ver_parse, $version);

    $ver_raw   = $Api_Set_Obj->version();
    @ver_parse = @{$ver_raw->get_parsed_output()};
    $version   = $ver_parse[0]->{'ontap_version'};

    return ($version);
} ## end sub Check_Ontap_Version($)
#---------------------------------------------------------------------------
# FUNCTION: Parses sysconfig -a output and returns bios version
#
# INPUT: Filer connection API set object
#
# RETURN: BIOS version
#
#---------------------------------------------------------------------------
sub Check_BIOS_Version($) {

    my $Api_Set_Obj = $_[0];
    my $biosver = 'unknown';
    my ($sys, @sys_parse, $num);

    $sys       = $Api_Set_Obj->sysconfig("detailed-information" => "-a");
    @sys_parse = @{$sys->get_parsed_output()};
    $num       = scalar @{$sys_parse[0]->{'slot_information'}};

    for (my $j = 0; $j < $num; $j++) {
        if (defined(
                $sys_parse[0]->{'slot_information'}[$j]->{'bios_version'}
            )
            ) {
            $biosver
                = $sys_parse[0]->{'slot_information'}[$j]->{'bios_version'};
        }
    }

    return ($biosver);
} ## end sub Check_BIOS_Version($)
#---------------------------------------------------------------------------
# FUNCTION: Check if AUTOBOOT is set to true, if not set it to true
#
# INPUT: Firmware connection API set object
#
# RETURN: None
#
#---------------------------------------------------------------------------
sub Check_AUTOBOOT_true($) {

    my $Firm_Obj = $_[0];
    my ($env_val);

    $env_val = $Firm_Obj->execute_raw_command(command => 'printenv AUTOBOOT');

    if ($env_val =~ /true/i) {
    } elsif (($env_val =~ /false/i) || ($env_val =~ /Undefined/i)) {
        $Firm_Obj->execute_raw_command(command => 'setenv AUTOBOOT true');
    }
} ## end sub Check_AUTOBOOT_true($)
#---------------------------------------------------------------------------
# FUNCTION: Verifies SP status at the LOADER
#
# INPUT: Firmware connection API set object
#
# RETURN: Returns "down" if link is down, 
#                 "up" if link is up & 
#                 "config" if ip address, netmask and gateway is not configured
#
#---------------------------------------------------------------------------
sub SP_Status_LOADER($) {

    my $Firm_Obj = $_[0];
    my ($sp_loader_raw, @sp_loader_parse, $count,
        $sp_ip_loader,  $sp_net_loader,   $sp_gw_loader,
    );
    my $status = 'up';

    $sp_loader_raw   = $Firm_Obj->sp_status();
    @sp_loader_parse = @{$sp_loader_raw->get_parsed_output()};

    #Verify SP's Ethernet link status is up
    if ($sp_loader_parse[0]->{'ethernet_link'} =~ /down/i) {
        $status = 'down';
    } else {

        $count = scalar @{$sp_loader_parse[0]->{'ipv4_settings'}};
        for (my $k = 0; $k < $count; $k++) {
            if (defined(
                    $sp_loader_parse[0]->{'ipv4_settings'}[$k]->{'ip_address'}
                )
                ) {
                $sp_ip_loader = $sp_loader_parse[0]->{'ipv4_settings'}[$k]
                    ->{'ip_address'};
            }
            if (defined(
                    $sp_loader_parse[0]->{'ipv4_settings'}[$k]->{'netmask'}
                )
                ) {
                $sp_net_loader
                    = $sp_loader_parse[0]->{'ipv4_settings'}[$k]->{'netmask'};
            }
            if (defined(
                    $sp_loader_parse[0]->{'ipv4_settings'}[$k]->{'gateway'}
                )
                ) {
                $sp_gw_loader
                    = $sp_loader_parse[0]->{'ipv4_settings'}[$k]->{'gateway'};
            }

        }
    }    #End of else

    if (   (!defined($sp_ip_loader))
        || (!defined($sp_net_loader))
        || (!defined($sp_gw_loader))) {

        $status = 'config';
    }

    return $status;
} ## end sub SP_Status_LOADER($)
#---------------------------------------------------------------------------
# FUNCTION: Get system Model name
#
# INPUT: Filer connection API set object
#
# RETURN: Returns  system model string
#
#---------------------------------------------------------------------------
sub Get_Sys_Model($) {

    my $Api_Set_Obj = $_[0];
    my $model = 'unknown';
    my ($sys, @sys_parse, $num, $sys_model);

    $sys       = $Api_Set_Obj->sysconfig("detailed-information" => "-a");
    @sys_parse = @{$sys->get_parsed_output()};
    $num       = scalar @{$sys_parse[0]->{'slot_information'}};

    for (my $j = 0; $j < $num; $j++) {
        if (defined(
                $sys_parse[0]->{'slot_information'}[$j]->{'model_name'}
            )
            ) {
            $sys_model
                = $sys_parse[0]->{'slot_information'}[$j]->{'model_name'};
        }
    }

    if($sys_model =~ /(FAS\d+)/) {
    $model = $1;
    }

    return $model;
} ## end sub Get_Sys_Model($)
#---------------------------------------------------------------------------
# FUNCTION: Get system Memory
#
# INPUT: Filer connection API set object
#
# RETURN: Returns  system memory from sysconfig -a output
#
#---------------------------------------------------------------------------
sub Get_Sys_Memory($) {

    my $Api_Set_Obj = $_[0];
    my $sysmem = 'unknown';
    my ($sys, @sys_parse, $num );

    $sys       = $Api_Set_Obj->sysconfig("detailed-information" => "-a");
    @sys_parse = @{$sys->get_parsed_output()};
    $num       = scalar @{$sys_parse[0]->{'slot_information'}};

    for (my $j = 0; $j < $num; $j++) {
        if (defined(
                $sys_parse[0]->{'slot_information'}[$j]->{'memory_size'}
            )
            ) {
            $sysmem
                = $sys_parse[0]->{'slot_information'}[$j]->{'memory_size'};
            last;
        }
    }

    return $sysmem;
} ## end sub Get_Sys_Memory($)
#---------------------------------------------------------------------------
# FUNCTION: Get_BIOS_Boot_Time_From_Console
#
# INPUT: Bios initialization console output string
#
# RETURN: Returns BIOS boot time in seconds 
#
#---------------------------------------------------------------------------
sub Get_BIOS_Boot_Time_From_Console($) {

    my $console_output = $_[0];
    my $boot_time_in_secs = 'unknown';
    my $bios_boot_time;

    if($console_output =~ /BIOS\s+boot\s+time\:\s+(\d+)\s+ms/i) {
    $bios_boot_time = $1;
    }
    if(defined($bios_boot_time)) {
       $boot_time_in_secs = ($bios_boot_time/1000);
    }

    return $boot_time_in_secs;
} ## end sub Get_BIOS_Boot_Time_From_Console($)
#---------------------------------------------------------------------------

1;
