# -*- perl -*-
# /* $Id: //depot/prod/test/main/platqa/non_volatile/nvram9/NVR9_Utils.pm $ */
#
# Copyright (c) 2009 NetApp, Inc.
# All rights reserved.
##
## @summary
## NVR9Utils.pm is a common module for all the NVRAM9 related test cases.
##
## @description
## This module provides common functions for use in testing NVRAM9 related
## functionalities.
##
## Sub-routines:
##
## 1. Find_System_Model_And_NVRAM_Device_Name
##	- To find system model of the filer and type of NVRAM installed.
##
## 2. Get_Bootargs
##	- To find the current vlaue of the Environmental variable
##
## 3. Filersio_Start
##	- To start FilerSIO stress.
##
## 4. Filersio_Stop
##	- To stop FilerSIO stress.
##
## 5. Filersio_Status
##	- To check FilerSIO stress.
##
## 6. verify_console_message
##	- To verify console messages on boot up.
##
## 7. verify_sysctl
##	- To verify sysctl values from system shell.
##
## 8. remove-filersio_instance
##	- To remove previou filersio instances files from systemshell.
##
## 9. show_sysstat
##	- To check the system statistics after i/0 run.
##
## 9. Set_Or_Reset_Bootargs
##	- To set/reset the vlaue of the Environmental variable
##
## @status private
##
## @author gajjela@netapp.com,dl-platform-qa@netapp.com
###############################################################################

package Platform::non_volatile::NVR9_Utils;

use TCD;
use NACL::APISet;
use Data::Dumper;
use Dumpvalue;

# Import exceptions
use NACL::APISet::Exceptions::ResponseException qw(:try);

use vars qw(
  @ISA @EXPORT
);

require Exporter;
@ISA = ('Exporter');

#List of functions supported by module
@EXPORT = qw(Find_System_Model_And_NVRAM_Device_Name
  Filersio_Start
  Filersio_Stop
  Filersio_Status
  verify_console_message
  verify_sysctl
  remove_filersio_instance
  show_sysstat
  Get_Bootargs
  Set_Or_Reset_Bootargs
);

our ( %dimm_size, %device_name, );

#List of Filers with corresponding NVRAM Dimm sizes
%dimm_size = (
    FAS6210 => '4096 MB',
    FAS8020 => '4096 MB',
    FAS6240 => '8192 MB',
    FAS8040 => '8192 MB',
    FAS8060 => '8192 MB',
    FAS6280 => '16384 MB',
    FAS8080 => '16384 MB',
    FAS6220 => '4096 MB',
    FAS6250 => '4096 MB',
    FAS6290 => '4096 MB',
);

#List of Filers with supported NVRAM cards
%device_name = (
    FAS6210 => 'nvram8',
    FAS8020 => 'nvram9',
    FAS6240 => 'nvram8',
    FAS8040 => 'nvram9',
    FAS8060 => 'nvram9',
    FAS6280 => 'nvram8',
    FAS8080 => 'nvram9',
    FAS6220 => 'nvram8',
    FAS6250 => 'nvram8',
    FAS6290 => 'nvram8',
);

#---------------------------------------------------------------------
# FUNCTION: To Find the FAS model of the system and type of NVRAM used
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Log object
# RETURN: 	FAS model and NVRAM used if success else 0 on FAIL
#---------------------------------------------------------------------

sub Find_System_Model_And_NVRAM_Device_Name {
    my $conn = $_[0];
    my $Test = $_[1];
    my ( $Fas_Model, $Device_Name ) = 0;

    try {
        my $response = $conn->sysconfig( "model-information" => 1 );
        my $parsed_output = $response->get_parsed_output();
        foreach my $t ( @{$parsed_output} ) {
            foreach my $tt ( keys %{$t} ) {
                if ( $tt =~ /model_name/ ) {
                    $Fas_Model = $t->{$tt};
                }    # end if
            }    # end inner foreach
        }    # end outer foreach
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $Test->nlog( "Error in setup: " . $exception_object->text() );
        my $resp = $exception_object->get_response_object();
        $Test->nlog( "Raw output:\n" . $resp->get_raw_output() );
        $Test->nlog("***** FAS model is not found from Sysconfig output *****");
        return ( $Fas_Model, $Device_Name );
    };

    #Finding the nvram type installed in the filer
    if ( exists( $device_name{"$Fas_Model"} ) ) {
        $Device_Name = $device_name{$Fas_Model};
    }
    else {
        $Test->nlog(
"***** FAS model $fasmodel is not listed in Device List.Please update this info 
					with corresoinding script owner*****"
        );
    }

    return ( $Fas_Model, $Device_Name );

} ## end sub Device_Name

#---------------------------------------------------------------------
# FUNCTION: To Start Filersio in the system
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Log object
# RETURN: 	1 if Filersio run successfully else 0.
#---------------------------------------------------------------------

sub Filersio_Start {
    my ( $count, $Original_State, $cmd, $response );
    my $Api_Set_Obj = $_[0];
    my $Test        = $_[1];
    my $result      = 1;

    $count = 3;

    #Execute fielrsio of 4 instances
    for ( $count = 0 ; $count <= 3 ; $count++ ) {
        try {
            my $response = $Api_Set_Obj->filersio_asyncio_active(
                'read-percentage' => 0,
                'rand-percentage' => 4,
                'blk-size'        => '4k',
                'start-byte'      => 0,
                'end-byte'        => '10g',
                'run-time'        => 3000,
                'num-active'      => 2,
                'filename'        => "/vol/vol0/testfile$count",
                'create'          => 1,
            );
        } ## end try
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            my $resp             = $exception_object->get_response_object();
            $Test->nlog( "Caught a " . ref($exception_object) . "!" );
            $Test->nlog( "Error is: " . $exception_object->text() );
            $Test->nlog( "Raw output was:\n" . $resp->get_raw_output() );
            $result = 0;
            last;
        };    #end catch
    }    # end for loop

    return $result;
} ## end sub Filersio_Start

#---------------------------------------------------------------------
# FUNCTION: To show Sysstat in the system
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Log object
# RETURN: 	1 if sysstat run successfully else 0.
#---------------------------------------------------------------------

sub show_sysstat {
    my $Api_Set_Obj = $_[0];
    my $Test        = $_[1];
    my $response;
    my $result = 1;

    try {
        my $response = $Api_Set_Obj->sysstat(
            'count'     => 6,
            'interval'  => 1,
            "print-all" => 1,
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        my $resp             = $exception_object->get_response_object();
        $Test->nlog( "Caught a " . ref($exception_object) . "!" );
        $Test->nlog( "Error is: " . $exception_object->text() );
        $Test->nlog( "Raw output was:\n" . $resp->get_raw_output() );
        $result = 0;
    };

    return $result;
} ## end sub show_sysstat

#---------------------------------------------------------------------
# FUNCTION: To Stop Filersio in the system
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Log object
# RETURN: 	1 if Filersio stops successfully else 0.
#---------------------------------------------------------------------

sub Filersio_Stop {
    my $Api_Set_Obj = $_[0];
    my $Test        = $_[1];
    my $result      = 1;

    try {

        my $response = $Api_Set_Obj->filersio_stop();
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        my $resp             = $exception_object->get_response_object();
        $Test->nlog( "Caught a " . ref($exception_object) . "!" );
        $Test->nlog( "Error is: " . $exception_object->text() );
        $Test->nlog( "Raw output was:\n" . $resp->get_raw_output() );
        $result = 0;
    };
    return $result;
} ## end sub Filersio_Stop

#---------------------------------------------------------------------
# FUNCTION: To Check Filersio status in the system
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Log object
# RETURN: 	1 if Filersio status run successfully else 0.
#---------------------------------------------------------------------

sub Filersio_Status {
    my $Api_Set_Obj = $_[0];
    my $Test        = $_[1];
    my $result      = 1;

    try {
        my $response = $Api_Set_Obj->filersio_status();
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        my $resp             = $exception_object->get_response_object();
        $Test->nlog( "Caught a " . ref($exception_object) . "!" );
        $Test->nlog( "Error is: " . $exception_object->text() );
        $Test->nlog( "Raw output was:\n" . $resp->get_raw_output() );
        $result = 0;
    };
    return $result;
} ## end sub Filersio_Status

#---------------------------------------------------------------------
# FUNCTION: To verify system console logs
# INPUT: 	$_[0]=console messages shud found array,
#			$_[1]= console messages shud not found array
#			$_[2]=Log oject, $_[3]= console log
# RETURN: 	1 if there no unexpected messages found else 0
#---------------------------------------------------------------------

sub verify_console_message {
    my ( $str_must_present, $str_must_absent, $Test, $Console_Log_Filer_A ) =
      @_;
    my $result = 1;
    my $line;
    my $found;
    my $msg;

    if (@$str_must_present) {
        foreach $msg (@$str_must_present) {
            $Test->nlog("msg = $msg\n");
        }    # end for loop

        my $Print_Once = 0;
        foreach $msg (@$str_must_present) {
            $found = 1;

            #$Test->nlog("msg = $msg\n");
            foreach $line ( split /\n/, $Console_Log_Filer_A ) {
                if ( $line =~ /.*$msg.*/ ) {
                    $Test->nlog("INFO: Found Message: $line");
                    $found = 0;
                    last;
                }    # end if
            }    #end inner for loop
            if ( $found != 0 ) {
                if ( $Print_Once == 0 ) {
                    $Test->nlog("Following messages not found in console log.");
                    $Test->nlog(" \'$msg\' \n");
                    $result     = 0;
                    $Print_Once = 1;
                    return $result;
                }    # end inner if

            }    # end outer if
        }    #end outer for loop
    }    # end outer if

    if (@$str_must_absent) {
        $Print_Once = 0;
        $Test->nlog("Searching for Unexpected console messages");
        foreach $msg (@$str_must_absent) {
            $found = 0;
            foreach $line ( split /\n/, $Console_Log_Filer_A ) {
                if ( $line =~ /.*$msg.*/ ) {
                    $Test->nlog("INFO: Found Message: $line");
                    $found = 1;
                    last;
                }    # end if
            }    #end inner for loop
            if ( $found == 1 ) {
                if ( $Print_Once == 0 ) {
                    $Test->nlog(
                        "Following unexpected message found in console log.");
                    $Test->nlog(
                        "==================================================");
                    $Test->nlog(" \'$msg\' \n");
                    $Print_Once = 1;
                    $result     = 0;

                    #return $result;
                }    # end inner if
            }    # end outer if
        }    #end outer for loop

    }    # end outer if
    return $result;
} ## end sub verify_console_message

#---------------------------------------------------------------------
# FUNCTION: To verify sysctl output
# INPUT: 	$_[0]=console messages shud found array,$_[1]=Log oject,
#			$_[2]= dev_nvram9_0_restage_done expected value
#			$_[3]= dev_nvram9_0_restage_fail expected value
#			$_[3]= Nvram_Device_Name
# RETURN: 	0 if sysctl values found expectedly else 0
#---------------------------------------------------------------------

sub verify_sysctl {
    my $Api_Set_Obj_Systemshell   = $_[0];
    my $Test                      = $_[1];
    my $dev_nvram9_0_restage_done = $_[2];
    my $dev_nvram9_0_restage_fail = $_[3];
    my $Nvram_Device_Name         = $_[4];
    my $result                    = 1;
    my $result_1                  = 1;
    my $result_2                  = 1;

    $Api_Set_Obj_Systemshell->api_reconnect_session()
      if !$Api_Set_Obj_Systemshell->api_is_session_alive();
    $Test->nlog("restage done expected value:$dev_nvram9_0_restage_done");
    my $response = $Api_Set_Obj_Systemshell->sysctl(
        'name' => "dev.$Nvram_Device_Name.0.restage_done" );
    my @bootarg_set = @{ $response->get_parsed_output() };
    my $restage_done_param =
      $bootarg_set[0]->{"dev.$Nvram_Device_Name.0.restage_done"};
    if ( $restage_done_param != $dev_nvram9_0_restage_done ) {
        $Test->nlog(
"dev.$Nvram_Device_Name.0.restage_done is => $restage_done_param it should be :$dev_nvram9_0_restage_done\n"
        );
        $result_1 = 0;
    }
    else {
        $Test->nlog(
            "dev.$Nvram_Device_Name.0.restage_done is => $restage_done_param \n"
        );
    }

    $response = $Api_Set_Obj_Systemshell->sysctl(
        'name' => "dev.$Nvram_Device_Name.0.restage_fail" );
    @bootarg_set = @{ $response->get_parsed_output() };
    my $restage_fail_param =
      $bootarg_set[0]->{"dev.$Nvram_Device_Name.0.restage_fail"};
    if ( $restage_fail_param != $dev_nvram9_0_restage_fail ) {
        $Test->nlog(
"dev.$Nvram_Device_Name.0.restage_fail is => $restage_fail_param it should be :$dev_nvram9_0_restage_fail \n"
        );
        $result_2 = 0;
    }
    else {
        $Test->nlog(
            "dev.$Nvram_Device_Name.0.restage_fail is => $restage_fail_param \n"
        );
    }

    if ( ( $result_1 != 1 ) || ( $result_2 != 1 ) ) {
        $result = 0;
    }
    else {
        $result = 1;
    }    #ends else
    return $result;
} ## end sub verify_sysctl

#---------------------------------------------------------------------
# FUNCTION: To remove previously run Filersio instances in the system
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Log object
# RETURN: 	1 if filersio instances removed successfully else 0
#---------------------------------------------------------------------

sub remove_filersio_instance {
    my $Api_Set_Obj = $_[0];
    my $Test        = $_[1];
    my $count;
    my $result = 1;

    for ( $count = 0 ; $count <= 3 ; $count++ ) {
        try {
            my $response =
              $Api_Set_Obj->rm( 'paths' => "/vol/vol0/testfile$count" );
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $exception_object = shift;
            $Test->nlog( "Error in setup: " . $exception_object->text() );
            my $resp = $exception_object->get_response_object();
            $Test->nlog( "Raw output:\n" . $resp->get_raw_output() );
            $result = 0;
        };
    }

    return $result;
} ## end sub remove_filersio_instance

#---------------------------------------------------------------------
# FUNCTION: To find current bootarg vlue
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Log object
#			$_[2]= bootarg name
# RETURN: 	Bootarg VALUE if success else 0
#---------------------------------------------------------------------

sub Get_Bootargs {
    my $Api_Set_Obj = $_[0];
    my $Test        = $_[1];
    my $bootarg     = $_[2];
    my $Orginal_bootarg_value;

    try {
        $response = $Api_Set_Obj->bootargs_get( arg => "$bootarg" );
        my $parsed_output = $response->get_parsed_output();
        foreach my $t ( @{$parsed_output} ) {
            foreach my $tt ( keys %{$t} ) {
                if ( $tt =~ /$bootarg/ ) {
                    $Orginal_bootarg_value = $t->{$tt};
                }    # end if
            }    # end inner foreach
        }    #
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $Test->nlog( "Error is: " . $exception_object->text() );
        my $resp = $exception_object->get_response_object();
        $Test->nlog( "Raw output was:\n" . $resp->get_raw_output() );
        $Orginal_bootarg_value = 0;

        #return $Orginal_bootarg_value;
    };

    return $Orginal_bootarg_value;
} ## end sub Get_Bootargs

#------------------------------------------------------------------------------
# FUNCTION: To find current bootarg vlue
# INPUT: 	$_[0]=Filer connection handle, $_[1]= Log object
#			$_[2]= bootarg name, $_[3]= bootarg value
# RETURN: 	Bootarg VALUE if success else 0
#------------------------------------------------------------------------------

sub Set_Or_Reset_Bootargs {
    my $Api_Set_Obj   = $_[0];
    my $Test          = $_[1];
    my $bootarg       = $_[2];
    my $bootarg_value = $_[3];
    my $modified_bootarg_value;
    try {
        $response = $Api_Set_Obj->bootargs_set(
            arg   => "$bootarg",
            value => "$bootarg_value"
        );
        my $parsed_output = $response->get_parsed_output();
        foreach my $t ( @{$parsed_output} ) {
            foreach my $tt ( keys %{$t} ) {
                if ( $tt =~ /$bootarg/ ) {
                    $modified_bootarg_value = $t->{$tt};
                }    # end if
            }    # end inner foreach
        }    #
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        $Test->nlog( "Error is: " . $exception_object->text() );
        my $resp = $exception_object->get_response_object();
        $Test->nlog( "Raw output was:\n" . $resp->get_raw_output() );
        $modified_bootarg_value = 0;

        #return $modified_bootarg_value;
    };
    return $modified_bootarg_value;
} ## end sub Set_Or_Reset_Bootargs

#------------------------------------------------------------------------------

1;

#------------------------------------------------------------------------------
