# this is basically -*- perl -*-
#
# $Id: //depot/prod/test/Rsierranevada/lib/Populate.pm#1 $
#
# Copyright (c) 2006 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##   Create files and directories using available clients.
##
## @description
##   This library provides routines to create any number of files and directories,
##   and various types of files.
##
## @dependencies
##
## @synopsis
##
## @duration
##
## @example
##
## @keywords
##
## @attribute
##   listtestcases=no
##
## @status
##   review
##
## @author
##   snider@netapp.com,okeefe@netapp.com,bryans@netapp.com
##
########################################################################

=head1 NAME

Populate.pm - Create files and directories using available clients.

=head1 DESCRIPTION

C<Populate> provides routines to populate/create many objects
using all available clients to create the dataset as quickly
as possible.

=head1 SYNOPSIS

    use Populate qw(create_set delete_set);

    create_set(clients => [qw(client1 client2)],
	       base_name => "foo",
	       size => "5k",
	       start => 0,
	       count => 100,
	       testdir => "/t/test",
               streams => 1);

    delete_set(clients => [qw(client1 client2)],
	       base_name => "foo",
	       start => 0,
	       count => 100,
	       testdir => "/t/test");

=head1 ROUTINES

=cut

my $LibraryLoaded;
if ($ENV{OVERRIDEPOPULATE}) {
    my $newModule = $ENV{OVERRIDEPOPULATE};
    delete $ENV{OVERRIDEPOPULATE};
    print "The main Populate module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package Populate;
use Module::Common qw(validate_mixed);
use Params::Validate qw(CODEREF ARRAYREF SCALAR UNDEF);
use Params::Common;
use RemoteExecution;
use Fcntl qw(:flock);
use Data::Dumper;
use FileManip;
use RandomData;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(
		 create_set
		 delete_set
		 create_files
		 delete_files
		 );
our @EXPORT_OK = qw(
		    create_set
		    delete_set
		    _handle_set
		    create_files
		    delete_files
		    );

use FindBin qw($RealBin);           # full directory pathname for this script
my $Nateeval = "$RealBin/nateeval";

use Options;
use TracePkg;
use Tharn;
use strict;

my $Remote = new RemoteExecution;
my $Trace = new TracePkg();
my $Options;

$Options = new Options();
$Options->parameters(definitions=>
            [
            ]);


#***********************************************************************

=head2 create_set

    create_set(clients => [qw(client1 client2)],
	       base_name => "foo",
	       size => "5k",
	       start => 0,
	       count => 100,
	       testdir => "/t/test",
               streams => 1);

C<create_set> is used to create the objects.

=over

=item clients

(Required) The anonymous list of clients that are used for the data creation.

=item timeout

(Optional) How long to wait for the subprocesses to finish.

=item testdir

(Required) The name of the directory to create the objects in, i.e. the directory path
on the client machines that is likely mounted from a file server.

=item base_name

(Optional) The string name to use for the objects. A period and an index for each
object will be directly appended to this string. The default value is set to "f.",
i.e. objects will be named "f.0", "f.1", "f.2", ...

=item name_size

(Optional) The size of the object name to use. For example, if C<name_size> is
16, then the twelfth object will be named "ttttttttttttt.11".

=item type

(Optional) A string (or the first character of the string) indicating the type of
objects to create. The strings are "text", "sparse", "dir", "symlink", "link", 
"pipe" (named pipe), "block" (block special), "char" (character special).
The default is to randomly choose one of the types for each object created.

=item start

(Optional) This is the beginning index for the objects to be created.
The default value is 0.

=item count

(Optional) This is the number of objects to create. The default is 100.

=item size

(Required) This is the size of the object to be created. If not specified, a random
size will be chosen between 0 and C<max_size>.

=item max_size

(Optional) The maximum size of object to create when C<size> is not specified.
The size of each object will be randomly chosen between 0 and C<max_size>. If
C<max_size> is not specified, it will default to 10M.

=item skip

(Optional) Distance to seek into a sparse file before the first write. If not specified,
a random value will be chosen between 0 and C<size>.

=item max_skip

(Optional) The maximum skip at the beginning of a sparse file when C<skip> is not
specified. The initial skip will be randomly chosen from between 0 and C<max_skip>.

=item gap

(Optional) Inter-write gap for the writes within a sparse file. If not specified,
a random value will be chosen between 1 and C<size>.

=item max_gap

(Optional) The maximum gap when C<gap> is not specified. Each gap will be randomly
chosen from between 0 and C<max_gap>.

=item verify

(Optional) A flag indicating to verify the object immediately after creation.
The default is 1.

=item streams

(Optional) A flag indicating whether or not to create streams/named attributes on text
files and directories.  Note that creating streams/named attributes will increase the 
file count.  Right now it is only valid to use this with Windows or Solaris clients
over CIFS/NFSv4 on Flex or Acro volumes; Coral is not supported.
The default is 0.

=back

=cut

#***********************************************************************

#
# Convenience routine for now, will remove later
#
sub create_files {
    $Trace->entry();
    $Trace->warn(msg=>"create_files is deprecated, please use create_set instead");
    my $rc = create_set(@_);
    $Trace->exit();
    return $rc;
}

#***********************************************************************

# Call the remote run code with create as the desired op
sub create_set {
    $Trace->entry();
    my $rc = _run_remotely(op=>"create", @_);
    $Trace->exit();
    return $rc;
}

#***********************************************************************

=head2 delete_set

    delete_set(clients => [qw(client1 client2)],
	       base_name => "foo",
	       start => 0,
	       count => 100,
	       testdir => "/t/test");

C<delete_set> is used to delete the objects.

=over

=item clients

(Required) The anonymous list of clients that are used for the object deletion.

=item timeout

(Optional) How long to wait for the subprocesses to finish.

=item testdir

(Required) The name of the directory to delete the objects from.

=item base_name

(Optional) The string name to use for the objects. A period and an index for each
object will be directly appended to this string. The default value is set to 'f.',
i.e. objects will be named 'f.0', 'f.1', 'f.2', ...

=item name_size

(Optional) The size of the object name to use. For example, if C<name_size> is
16, then the twelfth object will be named 'ttttttttttttt.11'.

=item start

(Optional) This is the beginning index for the objects to be deleted.
The default value is 0.

=item count

(Optional) This is the number of objects to delete. The default is 100.

=back

=cut

#***********************************************************************

#
# Convenience routine for now, will remove later
#
sub delete_files {
    $Trace->entry();
    $Trace->warn(msg=>"delete_files is deprecated, please use delete_set instead");
    my $rc = delete_set(@_);
    $Trace->exit();
    return $rc;
}

#***********************************************************************

# Call the remote run code with delete as the desired op
sub delete_set {
    $Trace->entry();
    my $rc = _run_remotely(op=>"delete", @_);
    $Trace->exit();
    return $rc;
}

#***********************************************************************

#
# Fix up the params and run the code on each of the clients
#
sub _run_remotely {
    $Trace->entry();
    $Options->process() if (!$Options->processed());
    my $options = $Options->get_hash();

    # All of the params for the object handling routines need to be specified here.
    # We could put the params in the individual ops and do a combination to get the list
    # of params to include here, but that probably will wait for another change.
    my %params = validate_mixed(
        @_,
        [],
        {
	    op => { type => SCALAR },
	    clients => { type => ARRAYREF | UNDEF },
	    timeout => { type => SCALAR, default => 600 },
	    testdir => { type => SCALAR },
	    base_name => { type => SCALAR | UNDEF, optional => 1, default => 'f' },
	    name_size => { type => SCALAR | UNDEF, optional => 1 },
	    type => { type => SCALAR | UNDEF, optional => 1 },
	    start => { type => SCALAR | UNDEF, default => 0 },
	    count => { type => SCALAR | UNDEF, optional => 1, default => 100 },
	    size => { type => SCALAR | UNDEF, optional => 1 },
	    max_size => { type => SCALAR | UNDEF, optional => 1, default => 10 * 1024 * 1024 },
	    skip => { type => SCALAR | UNDEF, optional => 1 },
	    gap => { type => SCALAR | UNDEF, optional => 1 },
	    blocksize => { type => SCALAR, default => 32 * 1024 },
	    verify => { type => SCALAR | UNDEF, optional => 1, default => 1 },
	    link_src_count => { type => SCALAR | UNDEF, optional => 1, default => 1 },
            symlink_src_count => { type => SCALAR | UNDEF, optional => 1, default => 1 },	    
            streams => { type => SCALAR | UNDEF, default => 0 },
        }
    );

    $Options->unit_to_base($params{start}, 
			   $params{count}, 
			   $params{size}, 
			   $params{max_size},
			   $params{skip},
			   $params{gap},
			   );
    $Trace->trace(msg=>"base_name=$params{base_name}, start=$params{start}, count=$params{count}, size=$params{size}, max_size=$params{max_size}, skip=$params{skip}, gap=$params{gap}");
    $Trace->debug(msg=>"Params:".Dumper(\%params));
    if (defined($params{clients})){
	$Trace->trace(msg => "taking non-subset route");
	my $jobs = scalar(@{$params{clients}});
	if ( !$jobs) {
	    $Trace->fatal(msg=>"not enough client instances", result=>"CONFIG");
	}

	#
	# %counts tracks the names of the variables that we'll be adjusting when we 
	# pass them down to each subprocess on the clients. The hash key is the start
	# value and the key is the count. These variables are used to adjust the values
	# in %params for the call to each client for each object type.
	#
	my %counts = (
		      start => "count",
		      );
	#
	# %remainders keeps track of how many of each object type are left after we
	# split the job evenly across the clients. If there is a remainder, we tack
	# an additional item onto each job until we've eliminated the remainder then
	# we reduce the number of items per job by 1 which should leave us with the
	# right total number.
	#
	my %remainders;

	#
	# Go through each of the counters that we're using and calculate the number 
	# of each type that we'll pass off to each client job.
	#
	while (my ($key, $value) = each %counts){
	    $remainders{$value} = $params{$value} % $jobs;
	    $params{$value} = int($params{$value} / $jobs);
	    # There are leftovers after evenly splitting the jobs, so adjust the count
	    # up by one to handle creating these extras. This number gets adjusted down
	    # later once the leftovers have been accounted for.
	    $params{$value}++ if $remainders{$value};
	}

	my $count = 0;
	foreach my $client (@{$params{clients}}) {
	    if ($count){
		# Check to see if we've started enough jobs to fulfill the extra items
		# we had left over from splitting the jobs. If so, adjust the number
		# of items per job by one.
		while (my ($key, $value) = each %counts){
		    $params{$value}-- if (($remainders{$value} - $count) == 0);
		}
	    }
	    $count++;

	    my $params = {
		LOGDIR => "$options->{LOGDIR}/test_setup", 
	    };
	    my $command = "use Populate qw(_handle_set); ";
	    $command .= "_handle_set(";
	    foreach my $key (sort keys %params){
		$command .= "$key=>" . ($key =~ /clients/ ? "undef" : "\"$params{$key}\"") . ", ";
	    }
	    $command .= ")";

	    $Trace->trace(msg => "Command is $command");
	    # Start the remote job.  Note that all jobs run in parallel.
	    $Remote->execute(host=>$client, 
			     cmd=>$command, 
			     nateeval=>1,
			     runid=>"$params{op}_set_on_$client",
			     nate_params=>$params, 
			     handler=>sub{});
	    #
	    # Adjust the start values for each item type so that the next client
	    # begins at the proper offest.
	    #
	    while (my ($key, $value) = each %counts){
		$params{$key} += $params{$value};
	    }
	}
	$Trace->trace(msg => "Waiting $params{timeout} for subtests to exit");
	my $result = $Remote->wait(timeout=>$params{timeout});
	die $result if ($result);
    } else {
	# Could just run the code on this client, but we won't for now.
	$Trace->fatal(msg => "No clients specified");
    }
    $Trace->exit();
}

#***********************************************************************

#
# Handle running the op on each object type.
#
sub _handle_set {
    my %params = (@_);
    my $partial_name_size;
    my $name;
    $Trace->entry(msg=>"start=$params{start}, count=$params{count}");
    $Trace->trace(msg=>"changing into dir $params{testdir}");
    if ($^O !~ /Win/i) {
	if ( !chdir($params{testdir}) ) {
	    $Trace->fatal(msg=>"chdir error for $params{testdir}: $!");
        }
    } else {
	#On Windows, we currently only create text files or directories
	if (defined($params{type}) && $params{type} ne "" && $params{type} !~ /text/i && $params{type} !~ /dir/i) {
	    $Trace->fatal(msg => "$params{type} is not supported on this platform. Current supported types for this platform: 'text', 'dir'. ");
	}
    }
    #
    # Go through each known object type and perform the proper operation the right
    # number of times.
    #
    $Trace->trace(msg=>"$params{op} $params{count} objects, numbered starting from $params{start}");
    for (my $i = 0; $i < $params{count}; $i++){
	# set an ordinal value for the sub to use in deciding when 
	# to create hardlink and symlink source files
	$params{ord} = $i;
	if ( defined($params{name_size}) && $params{name_size} ne "" ) {
	    # The name must take into account the size of the start value string, and the period.
	    $partial_name_size = $params{name_size} - length(($params{start}+$i)) - 1;
	    if ( $partial_name_size <= 0 ) {
		$partial_name_size = 1;
	    }
	    $name = ("t" x $partial_name_size) . "." . ($params{start}+$i);
	} else {
	    $name = $params{base_name} . "." . ($params{start}+$i);
	}
	if ($^O =~ /Win/i) {
	    my $newname =  $params{testdir}.'/'.$name;
	    $name = $newname;
	}
	_callsub("object", $name, %params);
    }
    $Trace->exit();
}

#***********************************************************************

#
# Handle calling the proper subroutine for op & type
#
sub _callsub {
    my ($type, $path, %params) = @_;

    # This routine dynamically creates a subroutine name and then calls it.
    # The caller of this routine can make a single call with a parameter ("type")
    # that indicates which subroutine to call.

    # As of now, this routine only calls _create_object and _delete_object,
    # and all the different objects are handled within those two subroutines.
    # But when code is added for reading/writing, this subroutine can be used
    # to call routines such as _read_object and _write_object, etc. and then
    # creating higher level wrappers like read_set and write_set.

    my $routine = "_$params{op}_$type";
    my $rc;
    {
	no strict 'refs';
	# Look up the coderef for the built-up routine
	my $coderef = *{"$routine"}{CODE};
	if ($coderef and defined(&$coderef)){
	    # Cool, it exists, so call it.
	    $rc = $coderef->($path, %params);
	} else {
	    # Bummer it doesn't exist, so bomb.
	    $Trace->fatal(msg => "No subroutine $routine found");
	}
    }
    return $rc;
}

#***********************************************************************

# Below are all of the object handling routines.

#***********************************************************************

#
# Handle creating a single object.  Parameters have already been validated
#
sub _create_object {
    my ($name, %params) = @_;
    my $size_bytes = $params{size};
    my $final_write = 0;
    my $orig_contents;
    my $lcnt;
    my $link_num;
    my @types = qw(text sparse dir symlink link pipe block char);
    my $stream_content;
    my $stream_name;
    my $cifsutf8 = 0;

    $Trace->entry();
    $Trace->debug(msg=>"params = ".Dumper(\%params));

    if ($^O =~ /Win/i) {
        @types = qw(text dir);
        $cifsutf8 = 1;
    } elsif ($^O =~ /Linux/i) {
        # Disallow streams/named attributes if not Windows or SunOS
        if ($params{streams}) {
            $Trace->warn(msg=>"Disabling streams/named attribute population because they are not supported from Linux clients.");
            $params{streams} = 0;
        }
    }

    # Before creating an object, remove the old one with the same name, if it exists.
    _delete_object($name, %params);
    # For parameters that weren't specified and aren't defaulted, choose random values.
    if ( !defined($params{type}) || $params{type} eq "" ) {
	$params{type} = $types[int(rand(scalar(@types)))];
	$Trace->trace(msg=>"randomly chose object type '$params{type}' for $name");
    }
    if ( !defined($params{size}) || $params{size} eq "") {
	$size_bytes = int(rand($params{max_size}));
	$Trace->trace(msg=>"randomly chose size $size_bytes for $name");
    }

    # take the modulus of ordinal, and desired link count
    if ($params{type} =~ /^link$|^symlink$/i) {
        $lcnt = $params{type}."_src_count";
	$link_num = $params{ord} % $params{$lcnt};    
    }
    $Trace->comment(msg=>"creating $name, type '$params{type}'");
    if ( $params{type} =~ /text|sparse/i ) {
	$Trace->trace(msg=>"creating $name of size $size_bytes");
	if ( !open(FILE, ">$name") ) {
	    $Trace->fatal(msg=>"open() error for $name: $!");
	}
	if ( $params{type} =~ /sparse/i && $size_bytes ) {
	    if ( !defined($params{skip}) || $params{skip} eq "" ) {
		$params{skip} = int(rand($size_bytes));
		$Trace->trace(msg=>"randomly chose skip $params{skip} for $name");
	    }
	    my $current_offset = $params{skip};
	    do {
		if ( ($current_offset >= $size_bytes) && (!$final_write) ) {
		    $current_offset = $size_bytes - 1;
		    $final_write = 1;
		}
		$Trace->comment(msg=>"    seeking to offset $current_offset in $name");
		if ( !seek(FILE, $current_offset, 0) ) {
		    $Trace->fatal(msg=>"seek() error within $name: $!");
		}
#		$Trace->trace(msg=>"seeked to $current_offset in $name");
		$orig_contents .= "\c@" x ($current_offset - length($orig_contents));
		$Trace->comment(msg=>"    writing to $name");
		if ( !print(FILE "f") ) { # hex code 66
		    $Trace->fatal(msg=>"print() error within $name: $!");
		}
#		$Trace->trace(msg=>"wrote to $current_offset in $name");
		$orig_contents .= "f";
		if ( !defined($params{gap}) || $params{gap} eq "") {
		    $params{gap} = int(rand($size_bytes));
		    # we don't want a gap of 0, lest we have an infinite loop
		    if ( $params{gap} eq 0 ) {
			$params{gap} = 1;
		    }
#		    $Trace->debug(msg=>"randomly chose gap $params{gap} for $name");
		}
		$current_offset += $params{gap};
	    } while ( $current_offset < $size_bytes );
	} else {
	    while ( $size_bytes ) {
		my $write_size = ($size_bytes > $params{blocksize}) ? $params{blocksize} : $size_bytes;
		$orig_contents .= "w" x $write_size; # hex code 77
		$Trace->comment(msg=>"writing $write_size bytes to $name");
		if ( !print(FILE "w" x $write_size) ) {
		    $Trace->fatal(msg=>"print() error within $name: $!");
		}
		$Trace->trace(msg=>"$write_size bytes written to $name");
		$size_bytes -= $write_size;
	    }
	}
#	$Trace->trace(msg=>"orig_contents = $orig_contents");
	if ( !close(FILE) ) {
	    $Trace->fatal(msg=>"close() error for $name: $!");
	}
        if ( $params{streams} ) {
            my $size = $params{size} || 1025;
            my $name_size = $params{name_size} || 16;
            my $rnd = randomString->new(size => $size, mixcase => 1, randomsize => 1, cifsutf8 => $cifsutf8);
            $stream_content = $rnd->String();
            $rnd = randomString->new(size => $name_size, mixcase => 1, randomsize => 1, cifsutf8 => 0);
            $stream_name = '';
            while (length($stream_name) < 1) {
                $stream_name = $rnd->String();
            }

            $Trace->comment(msg=>"Writing stream to $name");
            if (! FileManip::write_stream(stream => $stream_name, file => $name, contents => $stream_content) ) {
                $Trace->fatal(msg=>"write_stream() error for file $name: $!");
            }
        }
	if ( $params{verify} ) {
	    $Trace->comment(msg=>"verify: opening $name");
	    if ( !open(FILE, "$name") ) {
		$Trace->fatal(msg=>"open() error for $name: $!");
	    }
	    $Trace->comment(msg=>"verify: locking $name");
	    if ( !flock(FILE, LOCK_SH) ) {
		$Trace->fatal(msg=>"flock() error for $name: $!");
	    }
	    my $current_contents;
	    my $i;
	    my $buf_size = 4096;
	    my $contents_offset = 0;
	    $Trace->comment(msg=>"verify: reading $name");
	    do {
		$i = sysread(FILE, $current_contents, $buf_size, $contents_offset);
		if ($i > 0) {
		    $contents_offset += $i;
		}
	    } while ($i > 0);
#	    $Trace->trace(msg=>"current_contents = $current_contents");
	    if ( !close(FILE) ) {
		$Trace->fatal(msg=>"close() error for $name: $!");
	    }
	    if ( $current_contents ne $orig_contents ) {
		$Trace->fatal(msg=>"$name has incorrect contents.\nOriginal = $orig_contents\nCurrent = $current_contents\n");
	    }
            if ( $params{streams} ) {
                my $stream_read = FileManip::read_stream(stream => $stream_name, file => $name);
                if ($stream_read ne $stream_content) {
                    $Trace->fatal(msg=>"Stream on $name has incorrect contents.\nOriginal = $stream_content\n\nCurrent = $stream_read\n");
                }
            }
	}
    } elsif ( $params{type} =~ /dir/i ) {
        if ( !mkdir($name) ) {
            $Trace->fatal(msg=>"failed to create directory $name: $!");
        }
        if ( $params{streams} ) {
            my $size = $params{size} || 1025;
            my $name_size = $params{name_size} || 16;
            my $rnd = randomString->new(size => $size, mixcase => 1, randomsize => 1, cifsutf8 => $cifsutf8);
            $stream_content = $rnd->String();
            $rnd = randomString->new(size => $name_size, mixcase => 1, randomsize => 1, cifsutf8 => 0);
            $stream_name = '';
            while (length($stream_name) < 1) {
                $stream_name = $rnd->String();
            }

            $Trace->comment(msg=>"Writing stream to $name");
            if (! FileManip::write_stream(stream => $stream_name, file => $name, contents => $stream_content) ) {
                $Trace->fatal(msg=>"write_stream() error for directory $name: $!");
            }
        }
    } elsif ( $params{type} =~ /symlink/i ) {
	system("touch symlink_target_$link_num");
	if ( !symlink("symlink_target_$link_num", $name) ) {
	    $Trace->fatal(msg=>"failed to create symlink for $name: $!");
	}
	if ( $params{verify} && ! -l $name ) {
		$Trace->fatal(msg=>"$name failed verification: $!");
	}
    } elsif ( $params{type} =~ /link/i ) {
	system("touch link_target_$link_num");
	if ( !link("link_target_$link_num", $name) ) {
	    $Trace->fatal(msg=>"failed to create hard link for $name: $!");
	}
	if ( $params{verify} && ! -f $name ) {
		$Trace->fatal(msg=>"$name failed verification: $!");
	}
    } elsif ( $params{type} =~ /pipe/i ) {
	if ( system("mknod $name p") ) {
	    $Trace->fatal(msg=>"failed to create named pipe $name: $!");
	}
	if ( $params{verify} && ! -p $name ) {
		$Trace->fatal(msg=>"$name failed verification: $!");
	}
    } elsif ( $params{type} =~ /block/i ) {
	if ( system("sudo mknod $name b 8 0") ) {
	    $Trace->fatal(msg=>"failed to create block special $name: $!");
	}
	if ( $params{verify} && ! -b $name ) {
		$Trace->fatal(msg=>"$name failed verification: $!");
	}
    } elsif ( $params{type} =~ /char/i ) {
	if ( system("sudo mknod $name c 8 0") ) {
	    $Trace->fatal(msg=>"failed to create char special $name: $!");
	}
	if ( $params{verify} && ! -c $name ) {
		$Trace->fatal(msg=>"$name failed verification: $!");
	}
    } else {
	$Trace->fatal(msg=>"Invalid type '$params{type}' for $name");
    }
    $Trace->exit(msg => "done creating $name");
}

#***********************************************************************

#
# Handle deleting single a object
#
sub _delete_object {
    my ($name, %params) = @_;
    $Trace->entry();
    if ( -e $name ) {
	$Trace->trace(msg=>"deleting $name");
	if ( -d $name ) {
	    if ( !rmdir($name) ) {
		$Trace->fatal(msg=>"rmdir() error for $name: $!");
	    }
	} else {
	    if ( !unlink($name) && $! !~ /No such file or directory/ ) {
		$Trace->fatal(msg=>"unlink() error for $name: $!");
	    }
	}
    } else {
	$Trace->trace(msg=>"$name does not exist");
    }
    $Trace->exit();
}

=head1 PARAMS

=head1 SEE ALSO

=head1 BUGS

We need to provide a generic mount mechanism so that all of the clients 
have the same directories available

=head1 STATUS

Shared

=head1 ATTRIBUTES

=head1 AUTHOR

Bill Snider (snider@netapp.com)
Brian O'Keefe (okeefe@netapp.com)
Bryan Schmersal (bryans@netapp.com)

$Id: //depot/prod/test/Rsierranevada/lib/Populate.pm#1 $
Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;
