package PortableQC::Client;

use strict;
use IO::Socket;
use URI;
use SOAP::Lite;
use SOAP::Transport::TCP;
use Data::Dumper;
use Carp qw(croak cluck carp);

#
# Our SOAP transport is essentially TCP, so it makes sense to reuse
# SOAP::Transport::TCP. We reimplement send_receive here to support
# statefulness by keeping a TCP connection alive.
#
sub SOAP::Transport::TCP::Client::send_receive {

    my($self, %parameters) = @_;
    my($envelope, $endpoint, $action) =
        @parameters{qw(envelope endpoint action)};

    $endpoint ||= $self->endpoint;
    warn "URLs with 'tcp:' scheme are deprecated. Use 'tcp://'. Still continue\n"
        if $endpoint =~ s!^tcp:(//)?!tcp://!i && !$1;
    my $uri = URI->new($endpoint);

    local($^W, $@, $!); # <- what?

    # if the socket isn't already connected, then connect it
    unless (exists $self->{_qcif_sock}) {
        my $sock = new IO::Socket::INET->new(
            PeerAddr => $uri->host,
            PeerPort => $uri->port,
            Proto => 'tcp'
        );
        unless ($sock) {
            die("SOAP::Transport::TCP::Client::send_receive() couldn't connect".
            " QCif server");
        };
        $self->{_qcif_sock} = $sock;
    }

    SOAP::Trace::debug($envelope);

    # bytelength hack. See SOAP::Transport::HTTP.pm for details.
    my $bytelength = SOAP::Utils::bytelength($envelope);
    $envelope = pack('C0A*', $envelope)
        if !$SOAP::Constants::DO_NOT_USE_LWP_LENGTH_HACK && length($envelope) != $bytelength;

    my $msg = "";
    if ($self->{_qcif_sock}) {
        my $sock = $self->{_qcif_sock};
        print $sock sprintf("QCif%08x", $bytelength) . $envelope;
        
        SOAP::Trace::debug("Reading magic bytes...");
        my ($magic, $lenhex);
        if ((read($sock, $magic, 4) != 4) and ($magic ne 'QCif')) { # read magic bytes
            close($sock);
            die "bogus magic bytes";
        }
        SOAP::Trace::debug("Reading length...");
        if ((read($sock, $lenhex, 8) != 8)) { # read msg length
            close($sock);
            die "couldn't read msg length";
        }

        my $msglen = hex($lenhex);
        SOAP::Trace::debug("Length = $msglen");
        # fixme: sanity check $msglen

        my $buffer;
        my $bytesread=0;
        my $curlen = 0;
        my $remaining = ($msglen - $curlen);
        while ($bytesread=read($sock, $buffer, $remaining)) {
            SOAP::Trace::debug("Read a chunk ($bytesread bytes)");
            $msg .= $buffer;
            $curlen += $bytesread;
            $remaining = ($msglen - $curlen);
        };
        SOAP::Trace::debug("returned from while loop");
        unless (defined $bytesread) {
            close($sock);
            die "Connection to server closed mid-message";
        }
    }
    my $code = $@ || $!;

    $self->code($code);
    $self->message($code);
    $self->is_success(!defined $code || $code eq '');
    $self->status($code);

    SOAP::Trace::debug($msg);

    SOAP::Trace::debug("I just debugged a message");
    return $msg;
}

# Constructor for a PortableQC::Client object.
# e.g.
# my $qc_conn = PortableQC::Client::connect(
#   host => "qcserv.rtp.netapp.com", # required
#   port => 29027 # optional
#   user => "QCUser", # required (QC login)
#   pass => "mypass", # required (QC login password)
#   domain => "ONTAP", # required (QC domain)
#   project => "ONTAP", # required (QC project)
#   die_on_fatal => 1|0, # optional (default = 0)
#
sub PortableQC::Client::connect {
    my $args = {@_};
    my $obj = {};

    # check for required args
    foreach my $arg qw(host user pass domain project) {
        unless (exists $args->{$arg}) {
            carp("I require the $arg argument");
            return undef;
        }
        $obj->{$arg} = $args->{$arg};
    }

    # check for port
    $obj->{port} = (exists $args->{port})?$args->{port}:29027;
    # die on fatal?
    $obj->{die_on_fatal} = (exists $args->{die_on_fatal})?$args->{die_on_fatal}:0;
    
    # create a SOAP::Transport::TCP::Client instance
    # (using our overided send_receive() method)
    # Be careful to only keep one reference to this object!
    # The connection can only be closed when it's DESTROYed
    $obj->{conn} = SOAP::Lite->new();
    $obj->{conn}->uri('urn:ServerFunctions');
    $obj->{conn}->proxy(
        sprintf('tcp://%s:%u', $obj->{host}, $obj->{port})
    );

    # Send a "login" message. The server will internally
    #   * Create a QCConnection instance
    #   * Log in with user/pass
    #   * ConnectProject to domain/project
    #   * return "ok" if everything worked
    my $result;
    eval {
        my $som = $obj->{conn}->SOAPHandler({
            command => "login",
            user => $obj->{user},
            pass => $obj->{pass},
            domain => $obj->{domain},
            project => $obj->{project}
        });
        $result = $som->result;
    };
    unless ((defined $result) and ($result->{result} eq "ok")) {
        # This failure could be caused many ways, but it's
        # not easy to tell which specifically. The connection
        # could have been refused, or it could have connected,
        # but QC rejected login. If this causes problems, we can add
        # more useful debugging code.
        if ($obj->{die_on_fatal}) {
            unless (defined $result) {
                $result = {result=>"notok", fatal=>1, error=>"not_connected", error_msg=>"Couldn't connect to PortableQC server"};
            }
            die $result;    
        }
        return undef;
    }

    # Otherwise, everything worked.
    $obj->{disconnecting} = 0;
    bless $obj;
    return $obj;
}

# Return true if (we think) the connection is still valid.
# 
# (The best way to see if the connection still works is
#  to call a simple QC command.)
#
sub PortableQC::Client::isConnected {
    my $self = shift;
    return (exists $self->{conn})?1:0; 
}

# Disconnect from the server. 
# This object cannot be reused. Instantiate a new one.
#
sub PortableQC::Client::disconnect {
    my $self = shift;
    
    $self->{die_on_fatal} = 0; # call(disconnect) will definitely return fatal

    if ((exists $self->{conn}) and (not $self->{disconnecting})) {
        $self->{disconnecting} = 1; 
        $self->call(command=>"disconnect");
        delete $self->{conn} if (exists $self->{conn});
        $self->{disconnecting} = 0;
    }
}


# Call a server-side function (with the prefix _qc_).
# e.g. $qc->call(
#   command=>"testplan_query", 
#   limit => 500,
#   "TS_USER_20" => "dp/functional_move/*.thpl"
# );
#
# The "command=>" functions are implemented in PortableQC::ServerFunctions
#
# Returns undef or dies (depending on die_on_fatal) on fatal (disconnection)
# returns {result=>"notok",error=>"..."} on mild failure,
# and returns {result=>"ok", ...} on success.
#
sub PortableQC::Client::call {
    my $self = shift;
    my $args = {@_};
    my $result;

    # make sure we're connected
    unless (exists $self->{conn}) {
        $result = {
            result=>"notok", 
            fatal=>1, 
            error => "not_connected",
            error_msg=>"This PortableQC::Client object is no longer connected",
        };

        delete $self->{conn};

        if ($self->{die_on_fatal}) { 
            die $result;
        } else {
            return $result;
        }
    }

    # make sure we have the correct arguments
    unless (exists $args->{command}) {
        return {
            result=>"notok", 
            error => "bogus_command",
            error_msg=>"Missing 'command' argument",
        };
    }

    # send the request over the wire
    eval {
        my $som = $self->{conn}->SOAPHandler($args);
        $result = $som->result;
    };

    # If undef, then the server connection is dead
    unless (defined $result) {
        $self->disconnect();
        $result = {
            result=>"notok", 
            fatal=>1, 
            error => "not_connected",
            error_msg=>"I received undef from server (disconnection)",
        };
        if ($self->{die_on_fatal}) {
            die $result;
        } else {
            return $result;
        }
    }

    # If fatal=>1, then the server disconnected after sending the message
    if ($result->{fatal}) {
        Dumper($self);
        delete $self->{conn};
        if ($self->{die_on_fatal}) {
            die $result;
        } else {
            return $result;
        }
    }

    return $result;
}

# A setter for die_on_fatal
sub set_die_on_fatal {
    my $self = shift;
    my $value = shift;

    $self->{die_on_fatal} = ($value)?1:0;
}


1;
