package PortableQC::JAMSTestsetTools;

use strict;
use PortableQC::Client;
# use Exporter qw(import);
# use vars qw(@EXPORT_OK);
use Params::Validate qw(validate :types);
use Carp;
use Data::Dumper;

# @EXPORT_OK = qw(TSTestExists CreateTSTest EnumerateTSTests);

sub _parse_testset_path {
    my $testset = shift;

    # Split apart $testset into $testset_path and $testset_name
    # First remove all double slashes, foobar\\blah -> foobar\blah
    my @testset_components = split /\\/, $testset;
    my @sanitized_testset_components;
    for (my $i=0; $i<(scalar @testset_components); $i++) {
        my $component = $testset_components[$i];
        unless (length($component)==0) {
            push(@sanitized_testset_components, $component);
        }
    }

    # Do we have at least two path components?
    unless ((scalar @sanitized_testset_components) > 1) {
        die sprintf("_parse_testset_path: Implausible testset path (only has one component). Testset: [%s]", $testset);
    }

    # Take the last component to be the testset_name, and other components the testset_path
    my $testset_name = pop(@sanitized_testset_components);
    my $testset_path = "Root\\" . join("\\", @sanitized_testset_components);

    return [$testset_name, $testset_path];
}

sub _end_connection {
    my $obj = shift;
    eval {
        if (defined $obj->{qc}) {
            $obj->{qc}->disconnect();
        }
    };
    $obj->{qc} = undef;
}

sub _start_connection {
    my $obj = shift; # not necessarially a blessed object

    # kill the old connection, if it exists
    _end_connection($obj);

    # Setup a PortableQC connection
    my $qc;
    eval {
        $qc = PortableQC::Client::connect(
            host => $obj->{host},
            port => $obj->{port},
            user => $obj->{user},
            pass => $obj->{pass},
            domain => $obj->{domain},
            project => $obj->{project},
            die_on_fatal => 1,
        );
    }; if ($@) {
        croak(sprintf("JAMSTestsetTools::_start_connection: Unable to connect to PortableQC server.\nArgs: %s\nResponse: %s",
            Dumper($obj), Dumper($@))
        );
    }

    $obj->{qc} = $qc;
}

sub new {
    my $class = shift;
    my %args = validate(@_, {
        host => { # The host of the PortableQC server
            type => SCALAR,
            default => "qcserver.rtp.netapp.com",
        },
        port => { # The port of the PortableQC server
            type => SCALAR,
            callbacks => { is_16_bits => sub {my $p=shift; ($p > 0) and ($p < 0x10000);} },
            default => 29027,
        },
        user => { # The username for connecting to QC (required)
            type => SCALAR,
        },
        pass => { # The password for connecting to QC (required)
            type => SCALAR,
        },
        domain => { # The domain for connecting to QC
            type => SCALAR,
            default => "ONTAP",
        },
        project => { # The project for connecting to QC
            type => SCALAR,
            default => "ONTAP",
        },
    });

    my $obj = {
        host => $args{host},
        port => $args{port},
        user => $args{user},
        pass => $args{pass},
        domain => $args{domain},
        project => $args{project},
        qc => undef,
    };
    
    # Setup a PortableQC connection
    _start_connection($obj); # this might croak 

    bless $obj;
    return $obj;
}

sub DESTROY {
    my $obj = shift;

    eval {
        if (defined $obj->{qc}) {
            _end_connection($obj);
        }
    };
}

sub EnumerateTSTests {
    my $obj = shift;
    my %args = validate(@_, {
        testset => {
            type => SCALAR,
        },
    });

    # If something killed the connection, try to restart it (or croak trying)
    unless (defined $obj->{qc}) {
        _start_connection($obj);
    } 

    # Perform some basic sanity/cleanup on the input
        # Trim leading/trailing whitespace
        $args{testset} =~ s/^\s+|\s+$//g;

        # assert non-zero length
        if (length($args{testset})==0) {
            die sprintf("EnumerateTSTests: Testset has a length==0: testset:[%s]", $args{testset});
        }

        # Embedded wildchars (*) shouldn't be allowed. They'll cause
        # false positives, and they can cause QC to behave badly
        # (e.g. running gigantic SQL queries internally)
        if ($args{testset} =~ /\*/) {
            die sprintf("EnumerateTSTests: Testset has an asterisk in its path: testset:[%s]", $args{testset});
        }
    
    # Parse testset path
    my ($testset_name, $testset_path);
    eval {($testset_name, $testset_path) = @{_parse_testset_path($args{testset})};};
    if ($@) {
        die sprintf("EnumerateTSTests: %s", $@);
    }

    # Query for the test, given the testset.
    my $result;
    eval {
        $result = $obj->{qc}->call(
            command => "testlab_query_testset_test",
            testset_path => $testset_path,
            testset_name => $testset_name,
            limit => 10000,
        );
    }; if ($@) {
        _end_connection($obj); # try to cleanly tear down the connection
        die sprintf("EnumerateTSTests: PortableQC connection died during testlab_query_testset_test " .
                    "Arguments = %s\nResult = %s",
                    Dumper(\%args), Dumper($result));
    } 
    elsif ($result->{result} ne "ok") {
        # Otherwise a different kind of error occured (not a fatal error)
        die sprintf("EnumerateTSTests: testlab_query_testset_test returned an error. Arguments: %s\nResult: %s",
            Dumper(\%args), Dumper($result));
    }
    elsif ($result->{more_results_left}) {
        _end_connection($obj);
        die sprintf("EnumerateTSTests: testlab_query_testset_test returned >10,000 results. This is probably a bug, ".
            "but if not, you may need to increase the 10,000 result limit.\nArgs: %s", Dumper($obj));
    }

    # The returned tests ($result->{matches}) are in a slightly weird format. QC prepends these test names with an instance
    # number (e.g. "[1]my_test_001") because a test can exist multiple times in a testset. The instance number distinguishes
    # them.
    # So, we'll build a new structure to hold the results.
    # ( e.g. $new_structure = [{name=>"my_test_001", instance=>"1", qc_path=>'Subject\ONTAP\blah'},...] )
    #
    my $new_structure = [];
    foreach (@{$result->{matches}}) {
        my $test = {qc_path => $_->{path}};
        unless ((($_->{test} =~ /\[(\d*)\](.*)/) != 2) or (length($1)==0) or (length($2)==0)) {
            die sprintf("EnumerateTSTests: testlab_query_testset_test returned a match in an unexpected format: %s\n",
                Dumper($_));
        }
        $test->{instance} = $1;
        $test->{name} = $2;
        push @$new_structure, $test; 
    }

    return $new_structure;
}

sub TSTestExists {
    my $obj = shift;
    my %args = validate(@_, {
        testset => {
            type => SCALAR,
        },
        test => {
            type => SCALAR,
        },
    }); 

    # If something killed the connection, try to restart it (or croak trying)
    unless (defined $obj->{qc}) {
        _start_connection($obj);
    }

    # Perform some basic sanity/cleanup on the input
        # Trim leading/trailing whitespace
        $args{testset} =~ s/^\s+|\s+$//g;
        $args{test} =~ s/^\s+|\s+$//g;

        # assert non-zero length
        if (length($args{test})==0 or length($args{testset})==0) {
            die sprintf("TSTestExists: Test or testset has a length==0: test:[%s] testset:[%s]",
                $args{test}, $args{testset});
        }

        # Embedded wildchars (*) shouldn't be allowed. They'll cause
        # false positives, and they can cause QC to behave badly
        # (e.g. running gigantic SQL queries internally)
        if (($args{test} =~ /\*/) or ($args{testset} =~ /\*/)) {
            die sprintf("TSTestExists: Test or testset has an asterisk in its path: test:[%s] testset:[%s]",
                $args{test}, $args{testset});
        }

    # Parse testset path
    my ($testset_name, $testset_path);
    eval {($testset_name, $testset_path) = @{_parse_testset_path($args{testset})};};
    if ($@) {
        die sprintf("TSTestExists: %s", $@);
    }

    # Query for the test, given the testset.
    my $result;
    eval {
        $result = $obj->{qc}->call(
            command => "testlab_query_testset_test",
            testset_path => $testset_path,
            testset_name => $testset_name,
            limit => 2,
            "TS_NAME" => $args{test},
        );
    }; if ($@) {
        _end_connection($obj); 
        die sprintf("TSTestExists: PortableQC connection died during testlab_query_testset_test " .
                    "Arguments = %s\nResult = %s", 
                    Dumper(\%args), Dumper($result));
    }

    # PortableQC returns this error if it can't find the testset 
    if (($result->{result} ne "ok") and ($result->{error} eq "missing_testset")) {
        # No, the testset/test doesn't exist
        return 0;
    }
    elsif ($result->{result} ne "ok") {
        # Otherwise a different kind of error occured (not a fatal error)
        die sprintf("TSTestExists: testlab_query_testset_test returned an error. Arguments: %s\nResult: %s", 
            Dumper(\%args), Dumper($result));
    }

    # If PortableQC returned "ok", that means that it found the testset.
    # Now check that it found a matching test
    
    if ((scalar @{$result->{matches}}) == 0) {
        # It didn't find any matching tests
        return 0;
    }
    # NOTE: It's possible to return more than 1 match.
    #       We might want to consider that an error in the future.

    # Success! We found the testset AND the test
    return 1;
}

# CreateTSTest: Copy a test into a testset. 
#               If the testset doesn't exist, it will be created.
#               If the path to the testset doesn't exist, it will be created.
#
# Arguments:
#   host => PortableQC server host (required)
#   port => PortableQC server port (default 29027)
#   user => QC username (required)
#   pass => QC pasword (required)
#   domain => QC domain (default ONTAP)
#   project => QC project (default ONTAP)
#   testset => Fully qualified path to a QC testset (e.g. ONTAP/Boilermaker/boilermker.1/Rboilermaker/01_PreFFT/my_testset_001) (required)
#   test_script_path => The relative NATE script path for a (Test Plan) test (required)
#   test_name => The name of a (Test Plan) test (required)
#   product_models => the '000 Product Models' field for the testset (required)
#   executing_subteam => the '004 Executing Subteam' field for the testset (required)
#   preexisting_feature => the '003 Pre-existing Feature Test Set' field for the testset (required)
#   ontap_mode => the '005 ONTAP Mode' field for the testset (required)
#   description => the description for the testset (optional)
#
# Note: IF the testset already exists, THEN none of its fields will be changed.
#
sub CreateTSTest {
    my $obj = shift;
    my %args = validate(@_, {
        testset => {type => SCALAR},
        test_script_path => {type => SCALAR},
        test_name => {type => SCALAR},
        product_models => {type => SCALAR},
        executing_subteam => {type => SCALAR},
        preexisting_feature => {type => SCALAR},
        ontap_mode => {type => SCALAR},
        description => {
            type => SCALAR,
            default => '<html><body>Auto-created by JAMS/PortableQC</body></html>',
        },
    });

    # If something killed the connection, try to restart it (or croak trying)
    unless (defined $obj->{qc}) {
        _start_connection($obj);
    }

    # Perform some basic sanity/cleanup on the input
    foreach my $key (qw(testset test_name test_script_path)) {
        # Trim leading/trailing whitespace
        $args{$key} =~ s/^\s+|\s+$//g;

        # assert non-zero length
        if (length($args{$key})==0) {
            die sprintf("CreateTSTest: The %s argument has a length==0", $key);
        }

        # Embedded wildchars (*) shouldn't be allowed. They'll cause
        # false positives, and they can cause QC to behave badly
        # (e.g. running gigantic SQL queries internally)
        if ($args{$key} =~ /\*/) {
            die sprintf("CreateTSTest: The %s=>'%s' argument has an asteriks (wildchar) - which will mess up a QC query", 
                $key, $args{$key});
        }
    }

    # Parse testset path
    my ($testset_name, $testset_path);
    eval {($testset_name, $testset_path) = @{_parse_testset_path($args{testset})};};
    if ($@) {
        die sprintf("CreateTSTest: %s", $@);
    }

    # Search for the test (in Test Plan), given its name and NATE script path
    my $result;
    eval {
        $result = $obj->{qc}->call(
            command => "testplan_query",
            limit => "1",
            "TS_NAME" => $args{test_name},
            "TS_USER_20" => $args{test_script_path},
        );
    }; if ($@) {
        _end_connection($obj);
        die sprintf("CreateTSTest: connection died during testplan_query.\nArgs = %s\nResult = %s",
            Dumper(\%args), Dumper($result));
    }

    # Test whether testplan_query worked
    if ($result->{result} ne "ok") {
        die sprintf("CreateTSTest: testplan_query returned notok: error=%s, error_msg=%s",
            $result->{error}, $result->{error_msg});
    }

    # Test whether testplan_query returned 1 (one) single matching test
    # FIXME: For the moment, we consider matching 2 or more tests to be a failure.
    #        Someone might want to change this logic, someday.
    if ((scalar @{$result->{matches}}) != 1) {
        die sprintf("CreateTSTest: testplan_query returned %u matches for the test. Test name:[%s] Script path:[%s]",
            $args{test_name}, $args{test_script_path});
    }
    
    # Otherwise, we successfully found the test in Test Plan
    my $test_hash = $result->{matches}->[0];
    my $test_path = $result->{matches}->[0]->{path};
    if ($args{test_name} ne $result->{matches}->[0]->{test}) {
        # Weird behavior - better kill QC
        _end_connection($obj);
        die sprintf("CreateTSTest: Sanity check failed! I found the test in Test Plan, but it returned a different ".
                    "test_name than I expected. expected:[%s] got:[%s]",
                    $args{test_name}, $result->{matches}->[0]->{test});
    }

    # Now try to create the testset, or at least, confirm that it exists
    eval {
        $result = $obj->{qc}->call(
            command => "testlab_create_testset",
            path => $testset_path,
            testset => $testset_name,
            force_create_path => 1,
            update_if_exists => 0,
            fields => {
                "CY_COMMENT" => $args{description},
                "CY_USER_04" => $args{ontap_mode},
                "CY_USER_02" => $args{preexisting_feature},
                "CY_USER_03" => $args{executing_subteam},
                "CY_USER_30" => $args{product_models},
            },
        );
    }; if ($@) {
        _end_connection($obj);
        die sprintf("CreateTSTest: connection died during testlab_create_testset.\nArgs = %s\nResult = %s",
             Dumper(\%args), Dumper($result));
    }

    # Test whether testlab_create_testset worked
    unless ($result->{result} eq "ok") {
        die sprintf("CreateTSTest: testlab_create_testset failed.\nArgs = %s\nResult = %s",
            Dumper(\%args), Dumper($result));
    }

    # The testset now exists.  The final step is to "copy" the test into the testset.
    eval {
        $result = $obj->{qc}->call(
            command => "testlab_add_test",
            testset_path => $testset_path,
            testset_name => $testset_name,
            test_path => $test_path,
            test_name => $args{test_name},
            fields => {
                "TC_ACTUAL_TESTER" => $obj->{user},
            },
        );
    }; if ($@) {
        _end_connection($obj);
        die sprintf("CreateTSTest: connection died during testlab_add_test.\nArgs = %s\nResult = %s",
             Dumper(\%args), Dumper($result));
    }

    # Test whether it worked
    unless ($result->{result} eq "ok") {
        die sprintf("CreateTSTest: testlab_add_test failed. The testset exists, but adding the test failed. \nArgs = %s\nResult = %s",
             Dumper(\%args), Dumper($result));
    }

    # Sanity check: do a query to make sure the test was actually added
    eval {
        $result = $obj->{qc}->call(
            command => "testlab_query_testset_test",
            testset_path => $testset_path,
            testset_name => $testset_name,
            limit => 2,
            "TS_NAME" => $args{test_name},
        );
    }; if ($@) {
        _end_connection($obj);
        die sprintf("CreateTSTest: connection died during testlab_query_testset_test.\nArgs = %s\nResult = %s",
             Dumper(\%args), Dumper($result));
    }

    # Confirm that we actually see the test in the testset
    # So long as we got >0 results back, I'll consider that a PASS.
    unless (($result->{result} eq "ok") and ((scalar @{$result->{matches}}) > 0) ) {
        # QC might be going insane - better kill the connection
        _end_connection($obj);
        die sprintf("CreateTSTest: I couldn't verify that we actually added the test.\nArgs = %s\nResult = %s",
             Dumper(\%args), Dumper($result));
    }

    # Everything worked
    return 1;
}

1;
