package PortableQC::Server;

use strict;
require Win32::OLE; use Win32::OLE;
use SOAP::Lite;
use POE qw(Wheel::SocketFactory
  Wheel::ReadWrite
  Driver::SysRW
  Filter::Stream
);
use Log::Log4perl qw(:easy);
use PortableQC::ServerFunctions;
use Data::Dumper;
use Exporter qw(import);
use vars qw(@ISA @EXPORT @EXPORT_OK);

@EXPORT = qw(StartPortableQCServer);

# Entry point to PortableQC::Server. Begin serving and block indefinitely. 
#
# StartPortableQCServer(arg=>val, ...); 
sub StartPortableQCServer {
    my $args= {@_};
    my $args_defaults = {
        port => 29027, # the port to run this server
        forks => 5, # how many forked servers should we run
        log_path => "/var/log/portableqc/", # where to store the logs
        log_level => "trace", # what level of logging to write
    };
    my $server_state = {};
    foreach my $arg (keys %$args_defaults) {
        $server_state->{$arg} = (exists $args->{$arg})?$args->{$arg}:$args_defaults->{$arg};
    }
    
    _setup_logging($server_state);
    my $log = $server_state->{log};

    my $ServerFunctionsObj = PortableQC::ServerFunctions->new($server_state);
    $server_state->{ServerFunctionsObj} = $ServerFunctionsObj;

    # Start a socket
    POE::Session->create(
        inline_states => {
            _start         => \&socket_start,
            _stop          => \&socket_stop,
            do_fork        => \&socket_do_fork,
            got_error      => \&socket_got_error,
            got_sig_int    => \&socket_got_sig_int,
            got_sig_child  => \&socket_got_sig_child,
            got_connection => \&socket_got_connection,
            _child         => sub { undef },
        },
        heap => {
            server_state => $server_state, 
        },
    );

    POE::Kernel->run();

    # if we're connected to QC, try to cleanup gracefully
    if (exists $server_state->{com}) {
        com_cleanup($server_state);
        sleep(3); # wait a few seconds to let VB clean up its resources
    }

    $log->info("StartPortableQCServer is exiting...");
}

sub _setup_logging {
    my $server_state = shift;

    my $conf = q(
        log4perl.category = DEBUG, Screen 

        log4perl.appender.Screen = Log::Log4perl::Appender::Screen
        log4perl.appender.Screen.layout = Log::Log4perl::Layout::PatternLayout::Multiline
        log4perl.appender.Screen.layout.ConversionPattern = [%P %p %d] %m{chomp}%n
    );
    Log::Log4perl->init(\$conf);
    $server_state->{log} = get_logger();
}

# This callback opens a socket and listens for sigint
sub socket_start {
    my ($kernel, $heap) = @_[KERNEL, HEAP];

    $heap->{server} = POE::Wheel::SocketFactory->new(
        BindPort     => $heap->{server_state}->{port},
        SuccessEvent => "got_connection",
        FailureEvent => "got_error",
        Reuse        => "yes",

    );

    $kernel->sig(INT => "got_sig_int");
    
    $heap->{server}->pause_accept(); # only the children may accept connections
    $heap->{children}   = {};
    $heap->{is_a_child} = 0;

    "Server $$ has begun listening on port ", $heap->{server_state}->{port}, "\n";

    $kernel->yield("do_fork"); 
}

# This callback is called when our socket closes
sub socket_stop {
    my ($kernel, $heap) = @_[KERNEL, HEAP];
    $heap->{server_state}->{log}->info("Server stopped.");
}

# Shut down the server if it encounters an error.
sub socket_got_error {
    my ($heap, $syscall, $errno, $error) = @_[HEAP, ARG0 .. ARG2];
    my $log = $heap->{server_state}->{log};
    
    $log->warn("Server got $syscall error $errno: $error. Shutting down...");
    delete $heap->{server};
}

# The parent may fork() more processes as needed.  Because events are
# copied during fork(), we must make sure this is only done in the
# parent.  Otherwise, hilarity in the form of forkbombing may ensue.

sub socket_do_fork {
    my ($kernel, $heap) = @_[KERNEL, HEAP];
    my $log = $heap->{server_state}->{log};

    if ($heap->{is_a_child}) {
        return ;
    };

    while (scalar(keys %{$heap->{children}}) < $heap->{server_state}->{forks}) {
        my $pid = fork();

        # Fork failed.  Try again shortly.
        # FIXME: this can be an infinite loop
        unless (defined($pid)) {
            $log->warn("Server fork failed: $!. Will retry fork shortly");
            $kernel->delay(do_fork => 1);
            return;
        }

        # Parent.  Add the child process to its list.
        if ($pid) {
            $log->info("Server forked child $pid");
            $heap->{children}->{$pid} = 1;
            $kernel->sig_child($pid, "got_sig_child");
            next;
        }

        # Child.  Clear the child process list.
        $kernel->has_forked();
        $heap->{is_a_child} = 1;
        $heap->{children}   = {};

        $heap->{server}->resume_accept(); # resume accepting connections 
                                          # but pause again after accepting 1
        return;
    }
}

# Gracefully stop a server on SIGINT.  sig_handled() prevents the
# signal from gracelessly forcing the server to stop.

sub socket_got_sig_int {
    my $log = $_[HEAP]->{server_state}->{log};
    $log->warn("Server received SIGINT");
    delete $_[HEAP]->{server};
    $_[KERNEL]->sig_handled();
}

# Reap child processes, and fork new ones if the server is still
# listening for connections.  Ignore child processes that we've
# already stopped tracking.

sub socket_got_sig_child {
    my ($kernel, $heap, $child_pid) = @_[KERNEL, HEAP, ARG1];
    my $log = $heap->{server_state}->{log};

    if (delete $heap->{children}->{$child_pid}) {
        $log->info("Server reaped child $child_pid");
        $kernel->yield("do_fork") if exists $_[HEAP]->{server}; # fork later
    }
}

# A client connected to us! Setup a connection.
#
sub socket_got_connection {
    my ($heap, $socket, $peer_addr, $peer_port) = @_[HEAP, ARG0, ARG1, ARG2];
    my $log = $heap->{server_state}->{log};

    # fixme: log client info here
    $log->info("Server received a connection");
    $heap->{server}->pause_accept(); # this child will never accept a connection again

    POE::Session->create(
        inline_states => {
            _start      => \&connection_start,
            _stop       => \&connection_stop,
            got_data    => \&connection_got_data,
            got_error   => \&connection_got_error,
            _parent     => sub { 0 },
        },
        heap => {
            socket    => $socket,
            peer_addr => $peer_addr,
            peer_port => $peer_port,
            server_state => $heap->{server_state},
        },
    );

    # Kill the server obj, we don't allow any more connections after the first one. 
    if ($heap->{is_a_child}) {
        delete $heap->{server};
    }
}

### Interactor callbacks.

# The interactor has started for a client socket.  Begin reading
# and writing on the socket.

sub connection_start {
    my ($heap, $kernel, $connected_socket, $address, $port) =
        @_[HEAP, KERNEL, ARG0, ARG1, ARG2];
    my $server_state = $heap->{server_state};
    my $log = $server_state->{log};
    
    my $ServerFunctionsObj = $server_state->{ServerFunctionsObj};
    $heap->{soap} = SOAP::Server->new();
    $heap->{soap}->dispatch_with({"urn:ServerFunctions" => $ServerFunctionsObj});

    # Initialize state for raw TCP reading/writing
    $heap->{msg_len_curlen} = 0; # the number of bytes OF the message length we currently have
    $heap->{msg_len_raw} = ""; # string-representation of the message length
    $heap->{msg_len} = -1; # the message length
    $heap->{msg_curlen} = 0; # the number of bytes OF the message we currently have
    $heap->{msg} = ""; # the incomming SOAP request

    $heap->{client} = POE::Wheel::ReadWrite->new(
        Handle => $heap->{socket},
        Driver => POE::Driver::SysRW->new(),
        Filter => POE::Filter::Stream->new(),

        InputEvent => 'got_data',
        ErrorEvent => 'got_error',

        AutoFlush => 1,
    );

    $log->info("Client handler started");
}

# Log the fact that the interactor has stopped.
#
sub connection_stop {
    $_[HEAP]->{server_state}->{log}->info("Client handler stopped");
}

# We got some data from the client. Once we've collected a full SOAP request, 
# ship it off to the SOAP handler.
#
sub connection_got_data {
    my ($heap, $input) = @_[HEAP, ARG0];
    my $server_state = $heap->{server_state};
    my $log = $server_state->{log};

    # The message format is [4 magic bytes][8 bytes - msg length in bigendian/hexadecimal][n msg bytes...]

    while ((length $input) > 0) {
        # if we don't have the header yet
        if ($heap->{msg_len_curlen} < 12) {
            my $needed = 12-$heap->{msg_len_curlen}; # how many bytes do we need?
            my $available = length $input; # how many are available?
            my $copy = ($available>$needed)? $needed : $available; # fetch as many as possible
            $heap->{msg_len_raw} .= substr($input, 0, $copy); # append msg_len_raw with bytes
            $input = substr($input, $copy); # crop $input
            $heap->{msg_len_curlen} += $copy; # msg_len_curlen is now longer
        }  

        # if we have 12 bytes, check first 4 magic bytes 
        if ($heap->{msg_len_curlen} >= 12) {
            my $magic = substr($heap->{msg_len_raw}, 0, 4);
            my $msg_len_hex = substr($heap->{msg_len_raw}, 4, 8);
            if ($magic ne "QCif") { 
                my $complaint = "connection_got_data: got bogus magic: $magic";
                $log->error($complaint);
                die $complaint;
            }
            my $msg_len = hex($msg_len_hex);
            $log->debug("msg_len = $msg_len ($msg_len_hex)"); 
            $heap->{msg_len} = $msg_len; 
            # fixme: verify that $heap->{msg_len} is realistic
          }

        # if we've started reading the message,
        if ($heap->{msg_len} > -1) { 
            my $needed = $heap->{msg_len}-$heap->{msg_curlen};
            my $available = length $input;
            my $copy = ($available>$needed)? $needed : $available;
            $heap->{msg} .= substr($input, 0, $copy);
            $heap->{msg_curlen} += $copy;
            $input = substr($input, $copy);

            if ($heap->{msg_curlen} >= $heap->{msg_len}) {
                $log->debug("Got a whole message!");
                # process message
                my $response = $heap->{soap}->handle($heap->{msg}); 

                # send response
                $heap->{client}->put(sprintf("QCif%08x",(length $response)));             

                my $chunklen = 10000;
                for (my $i=0; $i<length($response); $i+=$chunklen) {
                    $heap->{client}->put(substr($response, $i, $chunklen));
                    $heap->{client}->flush() while $heap->{client}->get_driver_out_octets();
                }

                # if we need to terminate now, do so
                if ($server_state->{terminate_after_write}) {
                    $log->debug("terminate_after_write is true, so I'm terminating");
                    delete $heap->{client}; 
                    return ;
                }

                # reinit message-handling state
                $heap->{msg_len_curlen} = 0; # the number of bytes OF the message length we currently have
                $heap->{msg_len_raw} = ""; # string-representation of the message length
                $heap->{msg_len} = -1; # the message length
                $heap->{msg_curlen} = 0; # the number of bytes OF the message we currently have
                $heap->{msg} = ""; # the incomming SOAP request 
            }
        }
    }
    $log->debug("connection_got_data called");
}

# Stop interacting if a socket error has occurred.
#
sub connection_got_error {
    my ($heap, $operation, $errnum, $errstr) = @_[HEAP, ARG0, ARG1, ARG2];
    my $log = $heap->{server_state}->{log};
    $log->warn("Client handler got $operation error from $errnum: $errstr. Shutting down...");
    delete $heap->{client};
}

1;
