package PortableQC::ServerFunctions;
use strict;
require Win32::OLE; use Win32::OLE;
use Data::Dumper;
use Exporter qw(import);
use vars qw(@ISA @EXPORT @EXPORT_OK); 

@EXPORT = qw(
    com_cleanup
);

sub new {
    my $class = shift;
    my $server_state = shift;

    $server_state->{terminate_after_write} = 0; # close socket and exit() after writing last message

    my $obj = {server_state => $server_state};
    bless $obj, $class;
    return $obj;
}

# This is the only method called directly by SOAP
# SOAPHandler() handles the incomming message, and dispatches
# the request to a _qc_* method in PortableQC::ServerFunctions.
sub SOAPHandler {
    my $self= shift;
    my $args = shift;
    my $result;

    my $server_state = $self->{server_state};
    my $log = $server_state->{log};
    $log->debug("SOAPHandler called, command=".$args->{command}); 
    
    # sanitize $args->{command}, so a user can't inject code
    # into this eval()
    if ((not exists $args->{command}) or ($args->{command} =~ m/[^_a-zA-Z0-9]/)){
        return {
            result => "notok",
            error => "bogus_command",
            error_msg => "bogus command name",
        };
    }

    # run the command
    my $command = delete $args->{command};
    # FIXME: you can do $result = eval(...);
    eval sprintf('$result = _qc_%s($server_state, $args);', $command);

    # an uncaught die isn't fatal (but maybe it should be)
    if ($@) {
        return {
            result => "notok",
            error => "uncaught_die",
            error_msg => "Eval failed ($@)"
        };
    }

    if ((exists $result->{fatal}) and ($result->{fatal})) {
        # if result->fatal, then we should disconnect after
        # sending this final message
        # (An unrecoverable error or something happened.)
        $server_state->{terminate_after_write} = 1;
        $log->debug("SOAPHandler: '$command' fataled");
    }

    return $result;
}


# Try to destroy the COM object in such a way that it won't crash when
# the program terminates. 
# Visual Basic leaves threads running and accessing memory long after it's
# supposed to, and when perl terminates and free()s its memory, those
# threads will crash gracelessly.
sub com_cleanup {
    my $server_state = shift;
    my $log = $server_state->{log};
    my $com = delete $server_state->{com};

    warn "com_cleanup: defined com => " . ((defined $com)?'true':'false');
    if (defined $com) {
        $log->debug("com_cleanup: trying to disconnect...");
        $com->Disconnect();
        $log->debug("com_cleanup: did disconnect, releasing connection...");
        $com->ReleaseConnection();
        $log->debug("com_cleanup: did release conn, quitting...");
        $com->Quit();
        $log->debug("com_cleanup: did quit, DESTROYing com object...");
        $com->DESTROY();
        $com = undef;
        $log->debug("com_cleanup: did DESTROY.");
    }
}


# handle "login" message
# Connect to QC and log in
sub _qc_login {
    my $server_state = shift;
    my $args = shift;

    # Sanity check - you can only login once
    if (exists $server_state->{com}) {
        return { result => "notok", error=>"insanity", error_msg => "you already called login"};
    }

    # Dynamically load QC COM library, and instantiate a TDConnection
    $server_state->{com} = new Win32::OLE("TDApiOle80.TDConnection");

    # Enable the ability to enumerate "collection" objects
    Win32::OLE->Option(_NewEnum => 1);

    # Connect to QC server
    $server_state->{com}->InitConnectionEx("http://rtpqcas04.rtp.netapp.com:8080/qcbin");

    # Login
    $server_state->{com}->ConnectProjectEx(
        $args->{domain},
        $args->{project},
        $args->{user},
        $args->{pass}
    );

    # Verify everything worked
    my $didconnect = $server_state->{com}->Connected();
    my $didlogin = $server_state->{com}->LoggedIn();

    # Everything didn't work - this is a fatal error
    unless ($didconnect and $didlogin) {
       return {
            result => "notok",
            didconnect => $didconnect,
            didlogin => $didlogin,
            error_msg => "Couldn't connect/login",
            error => "failed_login",
            fatal => 1, # disconnect and terminate after sending
       };
    }

    

    # Otherwise, everything did work
    return {
        result => "ok",
    };
}


# Internally, the Test Plan columns don't have human-friendly names.
# "Nate Script Path" is actually "TS_USER_20".
# If you want to query for tests, you need to use the internal
# field names.
# This returns a hash map, e.g. {"TS_USER_20"=>"NATE Script Path", ...}
sub _qc_testplan_field_map {
    my $server_state = shift;
    my $self = shift;

    # No arguments necessary.

    my $TestFactory = $server_state->{com}->TestFactory;
    my $Fields = $TestFactory->Fields;

    my $field_map = {};
    foreach my $val (in $Fields) {
        my $field_name = $val->name;
        my $Property = $val->Property;
        my $field_userlabel = $Property->UserLabel;
        $field_map->{$field_name} = $field_userlabel;
    }

    return {
        result => "ok", field_map=>$field_map
    };
}

# This is like _qc_testplan_field_map, except it finds
# internal field names for testsets
sub _qc_testlab_field_map {
    my $server_state = shift;
    my $self = shift;

    # No arguments necessary.

    my $TestSetFactory = $server_state->{com}->TestSetFactory;
    my $Fields = $TestSetFactory->Fields;

    my $field_map = {};
    foreach my $val (in $Fields) {
        my $field_name = $val->name;
        my $Property = $val->Property;
        my $field_userlabel = $Property->UserLabel;
        $field_map->{$field_name} = $field_userlabel;
    }

    return {
        result => "ok", field_map=>$field_map
    };
}

# This is like _qc_testlab_field_map, except it finds internal
# field names for the tests that are INSIDE testsets.
sub _qc_testlab_test_field_map {
    my $server_state = shift;
    my $self = shift;

    # no arguments

    my $TSTestFactory = $server_state->{com}->TSTestFactory;
    my $Fields = $TSTestFactory->Fields;

    my $field_map = {};
    foreach my $val (in $Fields) {
        my $field_name = $val->name;
        my $Property = $val->Property;
        my $field_userlabel = $Property->UserLabel;
        $field_map->{$field_name} = $field_userlabel;
    }

    return {result=>"ok", field_map=>$field_map};
}

sub _qc_testplan_list_folder {
    my $server_state = shift;
    my $self = shift;


    # Check arguments
    my $raw_path = $self->{path};
    unless (defined $raw_path) {
        return {result=>"notok", error=>"bad_args", error_msg=>"I'm expecting a 'path' argument"};
    }

    # Digest the path
    my @path_components;
    foreach my $folder (split /\\/, $raw_path) {
        if (length($folder) > 0) {
            push @path_components, $folder;
        }
    }

    # '\' (root) = "Subject\"
    if ((scalar(@path_components)>0) and ($path_components[0] eq 'Subject')) {
        pop @path_components; # if path starts with 'Subject\', remove it. We implicitly start with Subject\
    }

    # Traverse the path components, until we find the folder's SubjectNode
    my $TreeManager = $server_state->{com}->TreeManager;
    my $node = $TreeManager->TreeRoot('Subject');

    my $path_so_far = '\\';
    foreach my $folder (@path_components) {
        $path_so_far .= $folder.'\\';
        $node = $node->FindChildNode($folder);
        unless (defined $node) {
            return {
                result=>"notok",
                error=>"missing_test",
                error_msg=>"I can't find this path in Test Plan: ".$path_so_far,
            };
        }
    }
    
    my @folder_list;

    # first enumerate the subfolders
    my $num_children = $node->Count();
    for (my $i=0; $i < $num_children; $i++) {
        my $child = $node->Child($i+1);
        my $name = $child->Name();
        push @folder_list, {name=>"$name", type=>"folder"};
    }

    # enumerate the tests
    my $node_id = $node->NodeID();
    my $TestFactory = $node->TestFactory;
    my $Filter = $TestFactory->Filter;
    $Filter->SetProperty("Filter", "TS_SUBJECT", $node_id);
    my $tests = $TestFactory->NewList($Filter->text);
    foreach my $test (in $tests) {
        my $name = $test->Field("TS_NAME");
        push @folder_list, {name=>"$name", type=>"test"};
    }

    # done
    return {result=>"ok", folder_list=>\@folder_list};
}

# Search for tests (in Test Plan) by specifying filters for fields.
# This is just like the filter at the top of the screen in the QC GUI
# testplan_query(
#   limit => 10, # optional, maximum number of matches to return (default=50)
#   "TS_USER_20" => "dp/functional_move/*.thpl", # example filter
#   "TS_NAME" => "my_testcase_055", # example filter
#   ... => ...
# )
# Returns matches=>[{path='Subject\qc\path\', name='testcase_name'}, ...]
sub _qc_testplan_query {
    my $server_state = shift;
    my $self = shift;

    my $TestFactory = $server_state->{com}->TestFactory;
    my $Filter = $TestFactory->Filter;

    # The limit (default=50) is the maximum number of results to return
    my $limit = (exists $self->{limit})?$self->{limit}:50;
    delete $self->{limit};

    # All other keys are field names (to filter on)
    foreach my $key (keys %$self) {
        # Populate a "Filter" object with our filters
        $Filter->SetProperty("Filter", $key, $self->{$key});
    }

    # Search for matching tests
    my $Matches = $TestFactory->NewList($Filter->text);

    my $matches = [];
    my $i=0;
    my $more_results_left = 0;
    # Iterate over the $Matches collection
    foreach my $test (in $Matches) {
        # if we've reached our $limit, break
        if ($i >= $limit) {
            $more_results_left = 1;
            last;
        }
        # Use the TS_SUBJECT field to retrieve a Subject object,
        # which will give us the Test Plan path.
        my $TestName = $test->Field("TS_NAME");
        my $Subject = $test->Field("TS_SUBJECT");
        my $Path = $Subject->Path;
        if (defined $Path) {
            push @$matches, {path=>$Path,test=>$TestName};
        }
        $i++;
    }

    return {
        result => "ok",
        more_results_left => $more_results_left,
        matches => $matches,
    };
}

sub _qc_testplan_retrieve_fields_by_id {
    my $server_state = shift;
    my $self = shift;
    my $log = $server_state->{log};

    # Check arguments
    foreach my $arg (qw(id fields)) {
        unless ((exists $self->{$arg}) and (defined $self->{$arg})) {
            return {result=>"notok", error=>"bad_args", error_msg=>"I'm expecting a '$arg' argument"};
        }
    }

    my $TestFactory = $server_state->{com}->TestFactory();
    my $Test = $TestFactory->Item($self->{id} + 0);
    unless (defined $Test) {
        return {result=>"notok", error=>"missing_test", error_msg=>"I can't find test #".$self->{id}};
    }    

    # Iterate over all the fields we're supposed to lookup, and store the data
    my $fields = {};
    foreach my $f (@{$self->{fields}}) {
        my $raw_field_data = $Test->Field($f);
        # Some fields return complex objects (instead of strings)
        # Sometimes serializing Win32::COM objects freezes perl,
        # so explicitly stringify the data before shipping it back to the
        # client.
        # (Except for these special cases)
        if (not (defined $raw_field_data)) { # The result is undef, don't try to dereference it
            $fields->{$f} = "$raw_field_data";
        }
        elsif (lc($f) eq "ts_subject") { # The result is a SubjectNode object, return its path
            $fields->{$f} = $raw_field_data->Path();
        }
        elsif ((lc($f) eq 'ts_vts') or (lc($f) eq 'ts_creation_date')) { # The result is a VT_DATE variant, stringify it
            $fields->{$f} = $raw_field_data->Date('MMM dd yyyy') . ' ' .
                $raw_field_data->Time('HH:mm:ss');
        }
        else { # Return the raw field as a string
            $fields->{$f} = "$raw_field_data";
        }
    }

    return {
        result => "ok",
        data => $fields,
    };
}

# Given a Test Plan path (e.g. Subject\ONTAP\Core\CSI\BAT)
# and a Test Plan test-name (e.g. vol_copy)
# Return the test's fields (as specified by $self->fields{})
sub _qc_testplan_retrieve_fields {
    my $server_state = shift;
    my $self = shift;
    my $log = $server_state->{log};

    # Check arguments
    foreach my $arg (qw(path test fields)) {
        unless ((exists $self->{$arg}) and (defined $self->{$arg})) {
            return {result=>"notok", error=>"bad_args", error_msg=>"I'm expecting a '$arg' argument"};
        }
    }

    # Sanitize $self->{path}
    my @path_comp;
    foreach my $folder (split /\\/, $self->{path}) {
        push(@path_comp, $folder) if (length($folder)>0);
    }
    if (((scalar @path_comp)>0) and ($path_comp[0] ne 'Subject')) {
        unshift @path_comp, 'Subject';
    }
    $self->{path} = join('\\', @path_comp);

    my $TreeManager = $server_state->{com}->TreeManager;
    my $TestPathTree = $TreeManager->NodeByPath($self->{path});

    unless (defined $TestPathTree) {
        return {
            result=>"notok", 
            error=>"missing_test", 
            error_msg=>"I can't find that path in Test Plan"
        };
    }

    my $SubjectID = $TestPathTree->NodeID();

    # TestFactory::Filter is broken and awful. We'll manually build our own filter string
    my $FilterText =
    sprintf("[Filter]{\r\nTableName:TEST,\r\nColumnName:TS_NAME,\r\nLogicalFilter:%s,\r\nVisualFilter:%s,\r\nNO_CASE:\r\n}\r\n{\r\nTableName:TEST,\r\nColumnName:TS_SUBJECT,LogicalFilter:%u,VisualFilter:\\%08x\\%s,\r\nNO_CASE:\r\n}",
        $self->{test}, $self->{test}, $SubjectID, length($self->{path}), $self->{path});
    warn "FilterText = " . $FilterText;
    
    # _find_test_by_name_path(test=>$self->{test}, path=>$self->{path},
    #     TestFactory=>$server_state->{com}->TestFactory);

    # [Filter]{
    # TableName:TEST,
    # ColumnName:TS_NAME,
    # LogicalFilter:EMS_event_log,
    # VisualFilter:EMS_event_log,
    # NO_CASE:
    # }
    # {
    # TableName:TEST,
    # ColumnName:TS_SUBJECT,
    # LogicalFilter:31559,
    # VisualFilter:\00000029\Subject\ONTAP\Core\CSI\Supportability\EMS,
    # NO_CASE:
    # }

    my $List = $TestPathTree->FindTests(
        "",
        1,
        $FilterText,
    );

    unless (defined $List) {
        return {
            result=>"notok",
            error => "missing_test",
            error_msg => "I found the path, but not the test.",
        };
    }

    # There should be at least one result
    # I'm not sure whether it's possible to have two identically-named
    # tests in the same folder - we'll just use the first result.
    my $i=0;
    my $first_match;
    foreach my $test (in $List) {
        $i++;
        if ($i >= 2) {
            last;
        }
        $first_match = $test;
    }

    if ($i < 1) {
        return {
            result=>"notok", 
            error => "missing_test",
            error_msg => "I found the path, but not the test.",
        };
    }
    if ($i > 1) { # This condition may not be possible
        $log->warn(sprintf("Weird: found multiple Test Plan tests matching path(%s) and test(%s)", 
            $self->{path}, $self->{test})
        );
    }

    # Iterate over all the fields we're supposed to lookup, and store the data
    my $fields = {};
    foreach my $f (@{$self->{fields}}) {
        my $raw_field_data = $first_match->Field($f);
        # Some fields return complex objects (instead of strings)
        # Sometimes serializing Win32::COM objects freezes perl,
        # so explicitly stringify the data before shipping it back to the
        # client.
        # (Except for these special cases)
        if (not (defined $raw_field_data)) { # The result is undef, don't try to dereference it
            $fields->{$f} = "$raw_field_data";
        }
        elsif (lc($f) eq "ts_subject") { # The result is a SubjectNode object, return its path
            $fields->{$f} = $raw_field_data->Path();
        }
        elsif ((lc($f) eq 'ts_vts') or (lc($f) eq 'ts_creation_date')) { # The result is a VT_DATE variant, stringify it
            $fields->{$f} = $raw_field_data->Date('MMM dd yyyy') . ' ' . 
                $raw_field_data->Time('HH:mm:ss');
        }
        else { # Return the raw field as a string
            $fields->{$f} = "$raw_field_data";
        }
    }

    return {
        result => "ok",
        data => $fields,
    };
}

sub _find_test_by_name_path {
    my $args = {@_};
    my $TestFactory = $args->{TestFactory};
    my $path = $args->{path};
    my $test = $args->{test};

    warn sprintf("Finding test from name='%s' path='%s'\n", $test, $path);

    my $Filter = $TestFactory->Filter;
    $Filter->SetProperty("Filter", "TS_NAME", $test);
    $Filter->SetProperty("Filter", "TS_SUBJECT", $path);
    my $filter_text = $Filter->text;

    warn "Filter: ";
    warn $filter_text;

    # [Filter]{
    # TableName:TEST,
    # ColumnName:TS_NAME,
    # LogicalFilter:EMS_event_log,
    # VisualFilter:EMS_event_log,
    # NO_CASE:
    # }
    # {
    # TableName:TEST,
    # ColumnName:TS_SUBJECT,
    # LogicalFilter:31559,
    # VisualFilter:\00000029\Subject\ONTAP\Core\CSI\Supportability\EMS,
    # NO_CASE:
    # }



}

sub _qc_testlab_retrieve_fields {
    my $server_state = shift;
    my $self = shift;

    # check arguments
    foreach my $arg (qw(path testset fields)) {
        unless (exists $self->{$arg}) {
            return {
                result=>"notok", 
                error => "bad_args",
                error_msg=>"I'm expecting a '$arg' argument",
            };
        }
    }

    # Lookup the folder path
    my $tstm = $server_state->{com}->TestSetTreeManager;
    my $TestSetPath = $tstm->NodeByPath($self->{path});
    unless (defined $TestSetPath) {
        return {
            result=>"notok", 
            error=>"missing_testset", 
            error_msg=>"I can't find that path in Test Lab",
        };
    }

    # Lookup the testset. Do a search for testsets in $self->{path} with
    # the name "$self->{testset}"
    my $TestSetFactory = $TestSetPath->TestSetFactory;
    my $Filter = $TestSetFactory->Filter;
    $Filter->SetProperty("Filter", "CY_CYCLE", $self->{testset});
    my $TestSetList = $TestSetPath->FindTestSets("", 1, $Filter->text);

    # It is not possible for two testsets to exist in the same directory with
    # the same name - so we don't need to handle that case.
    my $test;
    foreach my $_test (in $TestSetList) {
        $test = $_test;
        last;
    }
    unless (defined $test) {
        return {
            result => "notok", 
            error => "missing_testset",
            error_msg => "The path exists, but the testset does not",
        };
    }

    # Iterate over $self->{fields}, retrieving the data from $test
    my $field_data = {};
    foreach my $field (@{$self->{fields}}) {
        my $dat = $test->Field($field);
        $dat = (defined $dat)?"$dat":"NULL"; # string-ify the result
        $field_data->{$field} = $dat;
    }

    return {result=>"ok", field_data=>$field_data};
}

# This is a helper function for _qc_testlab_create_testset
# Create a given testset folder path, however deep it goes.
sub _create_testset_path {
    my $self = {@_};
    my $path = $self->{path};
    my $com = $self->{com};

    my @path_components;
    foreach my $folder (split /\\/, $path) {
        if (length($folder) > 0) {
            push @path_components, $folder;
        }
    }

    # might as well sanity check the path here
    # it should begin with 'Root'
    if ( (scalar(@path_components)<1) or ($path_components[0] ne "Root") ) {
        die "bogus path (doesn't begin with 'Root')";
    }

    shift @path_components; # we'll start iterating over the path after 'Root'
    my $current_path = "Root";
    my $tstm = $com->TestSetTreeManager;

    # iterate over the already-existing path components
    while ((scalar @path_components) > 0) {
        my $node = $tstm->NodeByPath("$current_path\\".$path_components[0]);
        unless (defined $node) {
            last;
        }
        $current_path .= "\\" . (shift @path_components);
    }

    # @path_components now contains the set of folders that don't exist
    # starting at $current_path.
    # Create these new folders.
    foreach my $folder_name (@path_components) {
        my $node = $tstm->NodeByPath($current_path);
        unless (defined $node) {
            die "couldn't open this path: $current_path";
        }

        # try to add $folder_name
        my $childnode = $node->AddNode($folder_name);
        unless (defined $childnode) {
            die "couldn't create this path: $current_path\\$folder_name";
        }
        $childnode->Post;
        $current_path .= "\\$folder_name";
    }
}

# $qc->call(
#   command => "testlab_create_testset",
#
#   path => 'Root\RollingRock\rollingrock.0', # (required)
#
#   testset => "my_testset_001", # (required)
#
#   fields => {"CY_COMMENT"=>"This testset contains blah blah...",
#              "CY_USER_03"=>"WAFL",
#              ... => ...,
#   },                      # (required) fields is a hashref of testset fields
#                           # to set. NOTE: CY_CYCLE and CY_FOLDER_ID are ignored.
#
#   force_create_path => 0, # (optional) if the path doesn't exist, create it.
#                           # if the path doesn't exist, and force_create_path=0,
#                           # the call will fail.
#                           # (default = 0 (don't create path))
# )
#   update_if_exists => 0,  # (optional) if the testset already exists, should
#                           # we update its fields?
#                           # (default = 0 (don't update fields if testset exists))
#
# Returns: {
#   result => "ok"|"notok", # ok -> success, notok -> failure
#   created_new_path => 1|0, # did (or did not) create a new path
#                            # (this can only be true if force_create_path was set true)
#   created_new_testset => 1|0, # This is false if the testset already existed, and
#   updated_fields => 1|0 # This is true if we modified the testset's fields
#
sub _qc_testlab_create_testset {
    my $server_state = shift;
    my $self = shift;

    # Should we create the path if it doesn't exist?
    my $force_create_path = (exists $self->{force_create_path})?$self->{force_create_path}:0;

    # If the testset already exists, should we update its fields?
    my $update_if_exists = (exists $self->{update_if_exists})?$self->{update_if_exists}:0;

    # Did we create the path?
    my $created_new_path = 0;
    # Did we create a new testset?
    my $created_new_testset = 0;


    # Check arguments
    foreach my $arg (qw(path fields testset)) {
        unless (exists $self->{$arg}) {
            return {result => "notok",
                    error_msg => "Argument '$arg' is required",
                    error => "bad_args",
                    created_new_path => 0,
                    created_new_testset => 0,
                    updated_fields => 0,
            };
        }
    }

    # Check whether the path already exists
    my $tstm = $server_state->{com}->TestSetTreeManager;
    my $TestSetPath = $tstm->NodeByPath($self->{path});

    unless (defined $TestSetPath) {
        # If if path DNE, and $force_create_path is false, return error
        unless ($force_create_path) {
            return {result => "notok",
                    error_msg => "path does not exist",
                    error => "missing_testset", # close enough
                    created_new_path => 0,
                    created_new_testset=>0,
                    updated_fields => 0,
            };
        }

        # Otherwise, create the path
        eval {_create_testset_path(com=>$server_state->{com}, path=>$self->{path});};
        if ($@) {
            return {result => "notok",
                    error_msg => "create_testset_path died: $@",
                    error => "failed_to_write",
                    created_new_path => 0,
                    created_new_testset=>0,
                    updated_fields => 0,
            };
        }

        # Lookup $TestSetPath again
        $TestSetPath = $tstm->NodeByPath($self->{path});
        unless (defined $TestSetPath) {
            return {result => "notok",
                    error_msg => "failed to create path",
                    error => "failed_to_write",
                    created_new_path=>0,
                    created_new_testset=>0,
                    updated_fields => 0,
            };
        }

        $created_new_path = 1; # we DID create a new path
    }

    # Lookup the old testset, to see if it exists
    my $TestSetFactory = $server_state->{com}->TestSetFactory;
    my $Filter = $TestSetFactory->Filter;
    $Filter->SetProperty("Filter", "CY_CYCLE", $self->{testset});
    my $testsets = $TestSetPath->FindTestSets("", 1, $Filter->text);

    my $testset;

    # Is this a valid, already-existing testset?
    # There can be at-most one result
    if (defined $testsets) {
        foreach my $ts (in $testsets) {
            $testset = $ts;
            last;
        }
    }
    if (defined $testset) {
        # we found an existing testset
        # unless we should update existing testsets, return
        unless ($update_if_exists) {
            return {
                result=>"ok",
                created_new_path=>$created_new_path,
                created_new_testset=>0, # did NOT create new testset
                updated_fields => 0,
            };
        }
        $created_new_testset = 0;
    }
    else {
        # The testset doesn't exist yet - we have to create it

        $testset = $TestSetFactory->AddItem([
            $self->{testset},
            $TestSetPath->NodeID()
        ]);

        unless (defined $testset) {
            return {result=>"notok" ,
                error_msg=>"couldn't create new testset",
                error => "failed_to_write",
                created_new_path=>$created_new_path,
                created_new_testset=>0, # did NOT create new testset
                updated_fields => 0,
            };
        }
        $created_new_testset = 1;
    }

    # Now populate its fields and return
    my $fields = $self->{fields};
    # These two fields are overridden by $self->{testset} and {path} respectively
    delete $fields->{"CY_CYCLE"} if (exists $fields->{"CY_CYCLE"});
    delete $fields->{"CY_FOLDER_ID"} if (exists $fields->{"CY_FOLDER_ID"});

    foreach my $f (keys %$fields) {
        $testset->SetProperty("Field", $f, $fields->{$f});
    }
    $testset->Post();

    return {
        result => "ok",
        created_new_path => $created_new_path,
        created_new_testset => $created_new_testset,
        updated_fields => 1,
    };
}

sub _qc_testlab_query_testset_test {
    my $server_state = shift;
    my $self = shift;
    
    # Check for required arguments
    my $testset_path = delete $self->{testset_path};
    my $testset_name = delete $self->{testset_name};
    unless ((defined $testset_path) and (defined $testset_name)) {
        return {result => "notok", error => "bad_args", error_msg=>"testset_path and testset_name need to be provided"};
    }

    # Optional argument(s)
    my $limit = delete $self->{limit};
    unless (defined $limit) {$limit = 10;}

    # First find the testset
    my $TestSetTreeManager = $server_state->{com}->TestSetTreeManager;
    my $TestSetFolder = $TestSetTreeManager->NodeByPath($testset_path);
    unless (defined $TestSetFolder) {
        return {result=>"notok", error=>"missing_testset", error_msg=>"Couldn't find that testset_path"};
    }
    my $List = $TestSetFolder->FindTestSets("$testset_name");
    warn "HEY! TESTSET_NAME = $testset_name";

    # Pull out the first testset
    my $ts;
    foreach my $_ts (in $List) {
        $ts = $_ts;
        last;
    }
    unless (defined $ts) {
        return {result=>"notok", error=>"missing_testset", error_msg=>"Couldn't find that testset"};
    }

    # Do a query for tests inside the testset
    my $TSTestFactory = $ts->TSTestFactory;
    my $Filter = $TSTestFactory->Filter;
    foreach my $f (keys %$self) {
        $Filter->SetProperty("Filter", $f, $self->{$f});
    }

    my $TSTestList = $TSTestFactory->NewList($Filter->text);
    my $matches = [];
    my $i=0;
    my $more_results_left = 0;
    foreach my $tst (in $TSTestList) {
        $i++;
        if ($i > $limit) {
            $more_results_left = 1;
            last;
        }

        my $test_name = $tst->Field("TS_NAME");
        my $Subject = $tst->Field("TS_SUBJECT");
        my $test_path = $Subject->Path;
        push @$matches, {path=>$test_path, test=>$test_name};
    }

    return {
        result => "ok",
        more_results_left => $more_results_left,
        matches => $matches,
    };
}

        



# This function adds a test to a testset, given:
#   The qc path and name of a test
#   The qc path and name of a testset
#   A set of fields to set (most importantly TC_ACTUAL_TESTER a.k.a. "001 Assigned Engineer")
# On success, returns {result=>"ok", already_existed=>1|0}
#   If already_existed is true, then the test already existed
#   in the testset AND WE DID NOT UPDATE ITS FIELDS. This is due
#   to a bug (apparently) in QC, where the API cannot update an
#   existing TSTest's fields.
#
# $qc->call(
#   command => "testlab_add_test",
#
#   testset_path => 'Root\RollingRock\rollingrock.0', # (required)
#
#   testset_name => "my_testset_001", # (required)
#
#   test_path => 'Subject\some\testplan\path', # (required)
#
#   test_name => 'wafl_exploder', # (required)
#
#   fields => {"TC_ACTUAL_TESTER" => "rutenbar",
#              "TC_STATUS" => "No Run",
#              ... => ...,
#   },                           # (required)
sub _qc_testlab_add_test {
    my $server_state = shift;
    my $self = shift;

    # Check for required arguments
    foreach my $arg (qw(test_path test_name testset_path testset_name fields)) {
        unless ((exists $self->{$arg}) and (defined $self->{$arg})) {
            return {result=>"notok", error=>"bad_args", error_msg=>"I need this argument: $arg"};
        }
    }
    # $server_state->{log}->debug(Dumper($self));

    # First find the testset
    my $TestSetTreeManager = $server_state->{com}->TestSetTreeManager;
    my $TestSetFactory = $server_state->{com}->TestSetFactory;
    my $TestSetPath = $TestSetTreeManager->NodeByPath($self->{testset_path});
    unless (defined $TestSetPath) {
        return {
            result=>"notok", 
            error=>"missing_testset", 
            error_msg=>"I couldn't find that testset path",
        };
    }
    my $Filter = $TestSetFactory->Filter;
    $Filter->SetProperty("Filter", "CY_CYCLE", $self->{testset_name});
    my $TestSetList = $TestSetPath->FindTestSets("", 1, $Filter->text);
    my $ts;
    foreach my $_ts (in $TestSetList) {
        $ts = $_ts;
        last;
    }
    unless (defined $ts) {
        return {
            result=>"notok", 
            error => "missing_testset",
            error_msg=>"Couldn't find that testset",
        };
    }

    # Find the testcase
    my $TestFactory = $server_state->{com}->TestFactory;
    my $TreeManager = $server_state->{com}->TreeManager;
    my $TestPathTree = $TreeManager->NodeByPath($self->{test_path});
    unless (defined $TestPathTree) {
        return {
            result=>"notok", 
            error => "missing_test",
            error_msg=>"I couldn't find that test path",
        };
    }
    my $SubTestFactory = $TestPathTree->TestFactory;
    $Filter = $TestFactory->Filter;
    $Filter->SetProperty("Filter", "TS_NAME", $self->{test_name});
    my $TestList = $SubTestFactory->NewList($Filter->text);
    my $t;
    foreach my $_t (in $TestList) {
        $t = $_t;
        last;
    }
    unless (defined $t) {
        return {
            result=>"notok", 
            error => "missing_test",
            error_msg=>"I couldn't find that test",
        };
    }

    # See if the testcase already exists in the testset
    my $TSTestFactory = $ts->TSTestFactory;
    $Filter = $TSTestFactory->Filter;
    $Filter->SetProperty("Filter", "TC_TEST_ID", $t->ID());
    my $TSTestList = $TSTestFactory->NewList($Filter->text);
    my $tst;
    my $tst_instance; # if the TSTest already exists, remember its instance ID
    foreach my $_tst (in $TSTestList) {
        $tst = $_tst;
        last;
    }
    if (defined $tst) {
        return {result=>"ok", already_existed=>1};
    }

    my $newtst = $TSTestFactory->AddItem($t->ID());
    my $fields = $self->{fields};
    # $server_state->{log}->debug("I'm about to iterate over fields...");
    foreach my $f (keys %$fields) {
        # $server_state->{log}->debug("Setting field [%s]=[%s]", "$f", "$fields->{$f}");
        $newtst->SetProperty("Field", "$f", "$fields->{$f}");
    }
    $newtst->Post();

    return {result=>"ok", already_existed=>0};
}

# For unittesting and debugging,
# call ServerTime() to make sure TDConnection is still active
sub _qc_debug_ping {
    my $server_state = shift;
    my $self = shift;
    my $server_time = $server_state->{com}->ServerTime();
    unless (defined $server_time) {
        return {result => "notok",
                error_msg => 'ServerTime() returned undef',
                fatal => 1,
        };
    }
    return {result => "ok"};
}

# For unittesting and debugging,
# sleep for $sleep_time before returning
sub _qc_debug_wait {
    my $server_state = shift;
    my $self = shift;
    
    my $sleep_time = $self->{sleep_time};
    sleep($sleep_time);

    return {result=>"ok"};
}

# For unittesting and debugging,
# return fatal=>1 (this should cause the server/client to disconnect)
sub _qc_debug_fatal {
    my $server_state = shift;
    my $self = shift;

    return {
        result => "notok",
        fatal => 1,
    }
}

# Cleanly disconnect and terminate
sub _qc_disconnect {
    my $server_state = shift;
    return {
        result => "ok",
        fatal => 1,
    };
}

1;
