#!/usr/software/bin/perl

use lib "/usr/software/lib/perl5/site_perl/5.8.8/";
use lib "/x/eng/kc-users/rutenbar/p4/test-dev/lib/";

use Test::More tests=>8;
use Getopt::Long;
use PortableQC::Client;
use Data::Dumper;
use strict;

main();
sub main {
    my $state = {
        host => "127.0.0.1",
        port => 29027,
        user => undef,
        pass => undef,
        domain => "ONTAP",
        project => "ONTAP",
        conn => undef,
    };
    parse_arguments($state);
    do_tests($state);
}


sub parse_arguments {
    my $state = shift;

    my $help;
    GetOptions(
        "host=s" => \$state->{host},
        "port=i" => \$state->{port},
        "user=s" => \$state->{user},
        "pass=s" => \$state->{pass},
        "domain=s" => \$state->{domain},
        "project=s" => \$state->{project},
        "help" => \$help,
        "h" => \$help,
    );
    my $helpmsg = 
        sprintf("usage: $0 \n") .
        sprintf("\t[--host qcserver.rtp.netapp.com] (required)\n") .
        sprintf("\t[--port 29027] (default: 29027)\n") .
        sprintf("\t[--user qcusername] (required)\n") .
        sprintf("\t[--pass qcpassword] (required)\n") .
        sprintf("\t[--domain QC_Domain] (default: ONTAP)\n") .
        sprintf("\t[--project QC_Project] (default: ONTAP)\n");

    if ($help) {
        printf("%s", $helpmsg);
        exit(0);
    }

    foreach my $arg (qw(host user pass)) {
        unless (defined $state->{$arg}) {
            printf("The %s argument is required\n\n%s", $arg, $helpmsg);
            exit(0);
        }
    }
}

sub search_field_map {
    my ($field_map, $value) = @_;
    my $match;
    foreach my $key (keys %$field_map) {
        if ($field_map->{$key} eq $value) {
            if (defined $match) {
                return undef;
            }
            $match = $key;
        }
    }
    return $match;
}

sub do_tests {
    my $state = shift;
    my ($conn1, $conn2);
    my $result;
    
    # First create two identical, simultaneous connections. 
    
    eval {
        $conn1 = PortableQC::Client::connect(
            host => $state->{host},
            port => $state->{port},
            user => $state->{user},
            pass => $state->{pass},
            domain => $state->{domain},
            project => $state->{project},
            die_on_fatal => 1,
        );
    }; 
    unless (ok((not $@), 'Connect to PortableQC::Server')) {
        diag("Could not connect to server");
        goto CLEANUP;
    }

    eval {
        $conn2 = PortableQC::Client::connect(
            host => $state->{host},
            port => $state->{port},
            user => $state->{user},
            pass => $state->{pass},
            domain => $state->{domain},
            project => $state->{project},
            die_on_fatal => 1,
        );
    }; 
    unless (ok((not $@), 'Connect to PortableQC::Server again')) {
        diag("I could connect to the server once, but the second time failed.  This might be a serious problem!");
        goto CLEANUP;
    }

    # Fetch field maps
    my ($testplan_field_map, $testlab_field_map, $testlab_test_field_map);
    subtest('Fetch field maps' => sub {
        plan tests => 3;

        $result = $conn1->call(command => "testplan_field_map");
        ok(($result->{result} eq "ok"), "Test Plan field map");
        $testplan_field_map = $result->{field_map};

        $result = $conn2->call(command => "testlab_field_map");
        ok(($result->{result} eq "ok"), "Test Lab field map");
        $testlab_field_map = $result->{field_map};

        $result = $conn1->call(command => "testlab_test_field_map");
        ok(($result->{result} eq "ok"), "Test Lab / test field map");
        $testlab_test_field_map = $result->{field_map};

        # print Dumper($testlab_test_field_map);
    });

    

    # Find Test Plan tests, and retrieve their information    
    my ($test_path, $test_name);
    subtest('Find a test in Test Plan' => sub {
        plan tests => 3; 

        # Find the internal field name for '040 NATE Script Path'
        my $script_path_field= search_field_map($testplan_field_map, '040 NATE Script Path');
        unless (ok((defined $script_path_field), 'Finding "NATE Script Path" internal field name')) {
            return ;
        }

        # Find a test
        $result = $conn1->call(
            command => "testplan_query",
            limit => 1,
             # it's reasonable to expect there'll be >=1 match for this query
            $script_path_field => 'dp/*.thpl',
        );
        ok( 
            (($result->{result} eq "ok") and (scalar(@{$result->{matches}}) == 1)), 
            "Finding functional_move.thpl"
        );
        $test_path = $result->{matches}->[0]->{path};
        $test_name = $result->{matches}->[0]->{test};
         
        # Lookup various fields for the test
        $result = $conn2->call(
            command => "testplan_retrieve_fields",
            path => $test_path,
            test => $test_name,
            fields => ['TS_Description', 'TS_RESPONSIBLE', 'TS_USER_15'],
        );
        ok(($result->{result} eq "ok"), 'Lookup various test fields') or 
            diag(Dumper($result));
    });

    # Create a testset, and then verify that it was written
    my ($testset_path, $testset_name);
    subtest('Create and verify testset' => sub {
        plan tests => 3;

        $testset_path = 'Root\Sandbox\PortableQC';
        $testset_name = 'qc_unittest_testset_001';
        my $fields = {
            "CY_USER_30" => 'N/A', 
            "CY_USER_03" => 'Unknown',
            "CY_USER_02" => 'N',
            "CY_USER_04" => 'NA',
            "CY_COMMENT" => '<html><body>This is testset generated by the unittest for PortableQC.  Random number: ' . int(rand(10000000)) . '</body></html>',
        };

        $result = $conn1->call(
            command => "testlab_create_testset",
            path => $testset_path,
            testset => $testset_name,
            force_create_path => 1,
            update_if_exists => 1,
            fields => $fields,
        );
        ok(($result->{result} eq "ok"), "Create a testset with conn1");

        my @field_names = keys %$fields;
        $result = $conn2->call(
            command => "testlab_retrieve_fields",
            path => $testset_path,
            testset => $testset_name,
            fields => \@field_names,
        );
        ok(($result->{result} eq "ok"), "Find testset with conn2");

        foreach my $f (keys %$fields) {
            if ($result->{field_data}->{$f} eq $fields->{$f}) {
                delete $fields->{$f};
            }
        }
        ok (((scalar(keys %$fields)) == 0), 'Verify fields match');
    });

    # Add a test to the testset
    $result = $conn1->call(
        command => "testlab_add_test",
        testset_path => $testset_path,
        testset_name => $testset_name,
        test_path => $test_path,
        'test_name' => $test_name,
        # FIXME: Don't always use my name here
        'fields' => {"TC_ACTUAL_TESTER" => "rutenbar"},
    );
    ok(($result->{result} eq "ok"), "Add a test to the testset");
    
    # Verify that we can still see that test
    $result = $conn2->call(
        command => "testlab_query_testset_test",
        testset_path => $testset_path,
        testset_name => $testset_name,
        limit => 50,
        "TS_NAME" => $test_name,
        "TS_SUBJECT" => $test_path,
    );
    ok(($result->{result} eq "ok"), "Lookup test-in-testset info");

    # print Dumper($result);


    # Verify that fatal-ing conn2 dies correctly
    eval {
        $result = $conn2->call(command=>"debug_fatal");
    };
    ok ((($@) and (not $conn2->isConnected())), "Fataled connection makes call() die");



CLEANUP:
    # Do cleanup here:
    if (defined $conn1) {$conn1->disconnect();}
    if (defined $conn2) {$conn2->disconnect();}
};


