#
# $Id: //depot/prod/test/Rsierranevada/lib/PostProc/LogMessage.pm#1 $
#

use strict;
use warnings;

package PostProc::LogMessage;

use TracePkg;
use Carp qw/confess/;
use Data::Dumper;

#
# class data & methods
#
my $Trace = new TracePkg();
$Trace->set_options(separateNewline => 1);

our %msgs_by_tokens;
our @nondescript_msgs;

sub msgs_by_freq
{
    my $obj = shift();
    if (ref($obj) && $obj->isa(qw/PostProc::LogMessage/)) {
        $Trace->warn(msg => "Static method called with context, ignoring.");
    }

    return [sort( 
        { scalar(@{$msgs_by_tokens{$b}}) <=> scalar(@{$msgs_by_tokens{$a}}) } 
        keys(%msgs_by_tokens))];
}

sub msg_total_size
{
    my $obj = shift();
    if (ref($obj) && $obj->isa(qw/PostProc::LogMessage/)) {
        $Trace->warn(msg => "Static method called with context, ignoring.");
        $obj = shift() || $Trace->Fatal(msg => "Need a set of tokens.");
    }

    if (!exists($msgs_by_tokens{$obj})) {
        confess("Tokens \"$obj\" aren't a hash key.\n");
    }

    return (scalar(@{$msgs_by_tokens{$obj}}) * 
        length($msgs_by_tokens{$obj}->[0]->{line}));
}

sub msgs_by_size
{
    my $obj = shift();
    if (ref($obj) && $obj->isa(qw/PostProc::LogMessage/)) {
        $Trace->warn(msg => "Static method called with context, ignoring.");
        $obj = shift() || $Trace->Fatal(msg => "Need a set of tokens.");
    }
    return [sort( 
        { 
            length($msgs_by_tokens{$b}->[0]->{line}) * scalar(@{$msgs_by_tokens{$b}}) >
            length($msgs_by_tokens{$a}->[0]->{line}) * scalar(@{$msgs_by_tokens{$a}})
        }
        keys(%msgs_by_tokens)
    )];
}

sub nondescript_msgs
{
    my $obj = shift();
    if (ref($obj) && $obj->isa(qw/PostProc::LogMessage/)) {
        $Trace->warn(msg => "Static method called with context, ignoring.");
    }

    return [sort( { $a->{line} cmp $b->{line} } @nondescript_msgs)];
}

#
# Methods
#

sub new ()
{
    $Trace->entry();
    my $class = shift();

    my %p = @_;

    my $self = ();

    $self->{num} = $p{num};
    $self->{toss_re} = $p{'toss-regexes'};
    $self->{remove_re} = $p{'remove-regexes'};

    $self->{line} = $p{line};
    chomp($self->{line});

    #
    # Tokenize the log line
    #
    $self->{tokens} = [split('\s+', $self->{line})];
    

    # For some reason there's an extra undefined array element
    $Trace->debug(msg => "Created line with ".scalar(@{$self->{tokens}})." tokens ($self->{line}).");

    #
    # Scan for undefined tokens
    #
    for (my $i = 0; $i < scalar(@{$self->{tokens}}); $i ++) {
        if (!defined($self->{tokens}->[$i])) {
            $Trace->comment(msg => "Token at $i isn't defined.");
            $Trace->comment(msg => "line is $self->{line}");
            $Trace->comment(msg => "Previous token is ".$self->{tokens}->[$i-1]);
            confess("Token at $i isn't defined\n");
        }
    }

    bless($self, $class);

    $Trace->exit();
    return $self;
}


#
# toss_tokens looks at the tokenized line and removes any common and
# meaningless tokens.  There are significant space/speed tradeoffs made here.
#
sub toss_tokens
{
    my $self = shift();
    $Trace->entry();
    $Trace->trace(msg => "Tossing tokens from line: ".join(", ", @{$self->{tokens}}));

    TOKEN: for (my $i = 0; 
        exists($self->{tokens}->[$i]);
        $i++) {

        if (!defined($self->{tokens}->[$i])) {
            # this should never happen
            $Trace->fatal(msg => "Index $i not defined: ".join(", ", @{$self->{tokens}}));
        }

        foreach my $re (@{$self->{toss_re}}) {
            # check each pattern to see if this token is needed.
            if ($self->{tokens}->[$i] =~ m/$re/) {
                splice(@{$self->{tokens}}, $i, 1);
                $i--;
                next TOKEN;
            }
        }
    }

    $self->{nondescript} = (scalar(@{$self->{tokens}}) ? 0 : 1);

    $Trace->trace(msg => "Resulting line: ".join(", ", @{$self->{tokens}}));
    $Trace->exit();
}

#
# remove_tokens looks at the tokenized line and removes any common and
# meaningless tokens.  There are significant space/speed tradeoffs made here.
#
sub remove_tokens
{
    my $self = shift();
    $Trace->entry();
    $Trace->trace(msg => "Removing tokens from line: ".join(", ", @{$self->{tokens}}));

    TOKEN: for (my $i = 0; 
        exists($self->{tokens}->[$i]);
        $i++) {

        if (!defined($self->{tokens}->[$i])) {
            # this should never happen
            $Trace->fatal(msg => "Index $i not defined: ".join(", ", @{$self->{tokens}}));
        }

        foreach my $re (@{$self->{remove_re}}) {
            # check each pattern to see if this token is needed.
            if ($self->{tokens}->[$i] =~ m/$re/) {
                splice(@{$self->{tokens}}, $i, 1);
                $i--;
                next TOKEN;
            }
        }
    }

    $self->{nondescript} = (scalar(@{$self->{tokens}}) ? 0 : 1);

    $Trace->trace(msg => "Resulting line: ".join(", ", @{$self->{tokens}}));
    $Trace->exit();
}

sub is_nondescript
{
    my $self = shift();
    return $self->{nondescript};
}

sub token_string
{
    my $self = shift();
    $Trace->entry();
    $Trace->exit();
    return join(" ", @{$self->{tokens}});
}

sub hash_entry()
{
    my $self = shift();
    $Trace->entry();

    if (!scalar(@{$self->{tokens}})) {
        push(@nondescript_msgs, $self);
        $Trace->exit();
        return;
    }

    if (!exists($self->{hashed}) || ($self->{hashed} != 1)) {

        my $tokenstring = join(" ", @{$self->{tokens}});

        # If there's already a copy of the tokens, delete them since they're
        # not needed.
        if (exists($msgs_by_tokens{ $tokenstring })) {
            $self->{tokens} = [];
        }
        push(@{$msgs_by_tokens{ $tokenstring }}, $self);

        if (scalar(@{$msgs_by_tokens{ $tokenstring }}) > 10) {
            delete($self->{line});
        }

        $self->{hashed} = 1;
    }
    $Trace->exit();
}

sub msg_by_tokens
{
    return \%msgs_by_tokens;
}

sub nondescript_msg
{
    return \@nondescript_msgs;
}

sub msg_info
{
}

sub DESTROY
{
    my $self = shift();
}

1;
