# PERL Module to assist with Remote Power management of servers and other equipment.
#
# Handle Remote Power operations of the Cyclades and Black Box power servers in use.
#
# Public:
#   Power::new	    creates a new object
#   Power::off	    turns off a machine/port
#   Power::on	    turns on a machine/port
#   Power::cycle    bounces the power to a machine/port
#   Power::status   checks the machine/port status
#   Power::stats    checks power/environmental stats on power server
#
# Private:
#   Power::connect  connects to a power server
#   Power::validateRconsPortsInfo
#		    validates and fills in object information from the RconsPorts file
#
# Standard Usage:
#
# # preferred and safest usage (all validation checks in place)
# my $power = new Power ( Machine => "$hostname" );
# if ( defined $power ) {
#   $power->cycle || die "Unable to cycle power for \"$hostname\"\n";
# } else {
#   print "Cannot operate on host \"$hostname\" Is it defined in RconsPorts?\n";
# }
# 
# # less safe usage (can't validate that you aren't resetting someone else's device)
# my $power = new Power ( Server => "pm1" );
# if ( defined $power ) {
#   my $status = $power->status( Port => 5 );
#   if ( $status =~ /off/i ) {
#	$power->on( Port => 5 );
#   } elsif ( $status =~ /on/i ) {
#	$power->off( Port => 5 );
#   }

# standard library overload function
my $LibraryLoaded;
if ($ENV{OVERRIDEPOWER}) {
    my $newModule = $ENV{OVERRIDEPOWER};
    delete $ENV{OVERRIDEPOWER};
    print "The Power module will be replaced with $newModule\n-------- Ignore the following redefined subroutine warnings ----------\n";
    require($newModule);
    print "----------------------------------------------------------------------\n";
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package Power;
use strict;
use warnings;
use Expect;

use lib "/x/eng/systest/perl";
use RconsPorts;

# create the object.  fill in and validate what we can from the RconsPorts file
# options:
#   Server	name of remote power server (defined in RconsPorts::rpServer) to connect to
#   Port	port number of remote power server to operate on
#   Machine	machine name (defined in RconsPort::machPort) to operate on
#   debug	turn on debugging information
#   relaxnamecheck	the cyclades power strip / cyclades acs is really buggy.  more often
#			than not the stupid PMD on the cyclades "forgets" what is attached to it.
#			so with this flag we can relax the name check in on/off/cycle to warn
#			if the name is empty, but still fail if the name does not match
#			
#
#   If a server is passed in, it will be validated against the RconsPorts file to make sure it is defined in there.  If
#   it isn't defined in the file the object will be returned as undef.
#
#   If a machine is passed in, it will be validated against the RconsPorts file to make sure it is defined there.  Also
#   the Server and Port for this object will be filled in, unless they were also passed in.  If they were, then we validate 
#   what was passed in against what the file has and return undef if they don't match (why would you pass in a machine name
#   and the server/port combo for it anyway).
sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my $self = {
	Server	    => undef,
	Port	    => undef,
	Machine	    => undef,
	debug	    => undef,
	relaxnamecheck	=> 1,
	@_};

    #
    # Cheesy detection of a network port argument.
    #
    if ( $self->{Machine} =~ m/((\w|-)+)\:(\w+)/ ) {
        use Power::NetPort;
        return new Power::NetPort(NetPort => $self->{Machine});
    }

    my $name = "Power::new:";
    bless $self, $class;

    # call the _validateRconsPortsInfo routine if we were given a machine or server to use to make sure we have 
    # good data in from the caller.
    if ( (defined $self->{Server} || defined $self->{Machine} ) && $self->validateRconsPortsInfo ) {
	return $self;
    } else {
	return undef; 
    }
}

# connect to a given remote power management server. noone outside of this module should need to call this function
# Only connects to the default server that is in the object.  No support for passing in an alternate server.
# XXX the cyclades telnet server seems to suck ass - so we use ssh on the cyclades instead of telnet
sub connect {
    my $self = shift;
    my $name = "Power::connect:";
    my $username = "power";
    my $password = "rewop";
    my $timeout = 10;
    
    # make sure that everything still looks good
    $self->validateRconsPortsInfo || return undef;

    # open a new connection.  
    # we loop through several times before giving up because the blackbox power modules only allow 1 person on at once.
    my $connected = 0;
    for ( my $i = 0 ; $i < 5 ; $i++ ) {
	$self->{expect} = new Expect;
	$self->{expect}->log_user(0);
	( $self->{debug} ) ? $self->{expect}->debug(1) : $self->{expect}->debug(0);
	$self->{expect}->raw_pty(1);
	if ( $RconsPorts::rpServer{$self->{Server}}{"Type"} =~ /cyclades/i ) { # cyclades login
	    print "$name calling \"/usr/bin/ssh $username\@$RconsPorts::rpServer{$self->{Server}}{IP}\"\n" if $self->{debug};
	    $self->{expect}->spawn("/usr/bin/ssh $username\@$RconsPorts::rpServer{$self->{Server}}{IP}");
	    if ( $self->{expect}->expect($timeout,
	    [qr/Are you sure you want to continue connecting \(yes\/no\)\?/i, sub { my $exp = shift; $exp->send("yes\n"); exp_continue; }],
	    [qr/REMOTE HOST IDENTIFICATION HAS CHANGED/i, sub { my $exp = shift; print $exp->exp_before() . $exp->exp_match() . $exp->exp_after(); $i = 10; }].
	    [qr/login: /i, sub { my $exp = shift; $exp->send("$username\n"); exp_continue; }],
	    [qr/password: /i, sub { my $exp = shift; $exp->send("$password\n"); exp_continue; }], 
	    [qr/power\]\$ /i]) ) {
		print "$name connected to $self->{Server} in " . ($i+1) . " attempts\n" if $self->{debug};
		$connected = 1;
		last;
	    } else {
		print "$name unable to connet to $self->{Server} ($i)\n" if $self->{debug};
	    }
	} elsif ( $RconsPorts::rpServer{$self->{Server}}{"Type"} =~ /blackbox/i ) { # blackbox login
	    $self->{expect}->spawn("/usr/bin/telnet $RconsPorts::rpServer{$self->{Server}}{IP}");
	    if ( $self->{expect}->expect($timeout, 'NPS> ') ) {
		print "$name connected to $self->{Server} in " . ($i+1) . " attempts\n" if $self->{debug};
		$connected = 1;
		last;
	    } else { # we probably got a connection refused
		sleep 2;
	    }
	} else {
	    print "$name Unsupported type \"$RconsPorts::rpServer{$self->{Server}}{Type}\"\n";
	    return undef;
	}
    }
    if ( !$connected ) {
	print "$name Unable to connect to power server $self->{Server} after 5 tries.\n";
	$self->{expect} = undef;
	return undef;
    } else {
	return 1;
    }
}

# turn a plug(s) off
# takes an optional port number to cycle and a force parameter to force an OFF call to the server
# returns undef on failure and 1 on success
sub off {
    my $self = shift;
    my %params = (
	Port	=>  $self->{Port},
	Force	=>  undef,
	@_);
    my $name = "Power::off:";
    my $timeout = 10;
    my @status;

    # if a Port was passed in and there is no name assigned - use the passed in port instead of the port in object
    if ( defined $self->{Machine} ) {
	if ( $params{Port} ne $self->{Port} ) {
	    print "$name Port \"$params{Port}\" passed in, but Machine is defined as \"$self->{Machine}\".  Using port \"$self->{Port}\" from RconsPorts instead of user specified port.\n";
	    $params{Port} = $self->{Port};
	}
    }
    
    # make sure we have a port
    if ( !defined $params{Port} ) {
	print "$name No port to operate on.  Cannot execute command.\n";
	return undef;
    }

    # check the port status before the operation
    @status = $self->status(Port => $params{Port});
    if ( defined $status[$params{Port}] ) {
	if ( !defined $params{Force} && ( $status[$params{Port}] =~ /,off/i ) ) {
	    print "$name Port \"$params{Port}\" is already OFF.  No need to turn it OFF.\n" if $self->{debug};
	    return 1;
	}
    } else {
	print "$name Unable to get status of port \"$params{port}\" Cannot continue.\n";
	return undef;
    }
    
    # if we have a machine name we need to make sure the port name matches it
    if ( defined $self->{Machine} ) {
	my ( $pmPort, $machineName, $powerStatus, $lockStatus ) = split(/,/, $status[$params{Port}]);
	if ( ( $self->{relaxnamecheck} == 1 ) &&  ($machineName =~ /\(undefined\)/i) ) {
	    warn "$name The name on the port is undefined, you could possibly be powering off the wrong machine,\nor the Cyclades is hosed up again.  Continuing anyway.\n";
	} elsif ( $machineName !~ /$self->{Machine}/i ) {
	    print "$name The machine name \"$self->{Machine}\" does not match the name on the port \"$machineName\".\nRefusing to power off.  Please correct the configuration of this port.\n";
	    return undef;
	}
    }

    # if we aren't connected, connect (blackbox needs to disconnect after every command)
    if ( !defined $self->{expect} ) {
	$self->connect || return undef;
    }

    # execute the command on the different power server types
    if ( $RconsPorts::rpServer{$self->{Server}}{"Type"} =~ /cyclades/i ) { # cyclades
	my $prompt = "power\]\$ ";
	$self->{expect}->send("pmCommand $RconsPorts::rpServer{$self->{Server}}{Port} off $params{Port}\n");
	if ( $self->{expect}->expect($timeout, $prompt) ) {
	    my $output = $self->{expect}->exp_before();
	    if ( $output !~ /$params{Port}: Outlet turned off/i ) {
		print "$name Power Server \"$self->{Server}\" encountered an error turning off port \"$params{Port}\"\n\"$output\"\n";
		return undef;
	    }
	}
    } elsif ( $RconsPorts::rpServer{$self->{Server}}{"Type"} =~ /blackbox/i ) { # blackbox
	my $prompt = 'NPS> ';
	$self->{expect}->send("/OFF $params{Port}\n\r");
	if ( $self->{expect}->expect($timeout, $prompt) ) {
	    my $output = $self->{expect}->exp_before();
	    if ( $output =~ /Invalid Plug Specified/i ) {
		print "$name Power Server \"$self->{Server}\" did not like port \"$params{Port}\"\n\"$output\"\n";
		return undef;
	    }
	}
	$self->{expect} = undef; # we need to disconnect from a blackbox unit
    } else {
	print "$name Unsupported type \"$RconsPorts::rpServer{$self->{Server}}{Type}\"\n";
	return undef;
    }

    # now check the status again.  verify that the port is off
    @status = $self->status(Port => $params{Port});
    if ( $status[$params{Port}] !~ /,off/i ) {
	print "$name OFF call must have failed.  Port does not appear to be OFF. \"$status[$params{Port}]\"\n";
	return undef;
    }
    return 1;
}

# cycle the power for a given machine or server/port combo
# if there is no machine identified, a port may be passed in to override the default port to operate on
# returns 1 on success undef on failure
sub cycle {
    my $self = shift;
    my %params = (
	Port	=>  $self->{Port},
	@_);
    my $name = "Power::cycle:";
    my $timeout = 10;
    my @status;

    # if a Port was passed in and there is no name assigned - use the passed in port instead of the port in object
    if ( defined $self->{Machine} ) {
	if ( $params{Port} ne $self->{Port} ) {
	    print "$name Port \"$params{Port}\" passed in, but Machine is defined as \"$self->{Machine}\".  Using port \"$self->{Port}\" from RconsPorts instead of user specified port.\n";
	    $params{Port} = $self->{Port};
	}
    }
    
    # make sure we have a port
    if ( !defined $params{Port} ) {
	print "$name No port to operate on.  Cannot execute command.\n";
	return undef;
    }

    # check the port status before the operation
    @status = $self->status(Port => $params{Port});
    if ( defined $status[$params{Port}] ) {
	if ( $status[$params{Port}] =~ /,off/i ) {
	    print "$name WARNING: Port \"$params{Port}\" is OFF.  Cycling it will turn it on.\n";
	}
    } else {
	print "$name Unable to check status on port \"$params{Port}\"  Cannot continue.\n";
	return undef;
    }
    
    # if we have a machine name we need to make sure the port name matches it
    if ( defined $self->{Machine} ) {
	my ( $pmPort, $machineName, $powerStatus, $lockStatus ) = split(/,/, $status[$params{Port}]);
	if ( ( $self->{relaxnamecheck} == 1 ) &&  ($machineName =~ /\(undefined\)/i) ) {
	    warn "$name The name on the port is undefined, you could possibly be powering cycling the wrong machine,\nor the Cyclades is hosed up again.  Continuing anyway.\n";
	} elsif ( $machineName !~ /$self->{Machine}/i ) {
	    print "$name The machine name \"$self->{Machine}\" does not match the name on the port \"$machineName\".\nRefusing to power cycle.  Please correct the configuration of this port.\n";
	    return undef;
	}
    }
    
    # if we aren't connected, connect (blackbox needs to disconnect after every command)
    if ( !defined $self->{expect} ) {
	$self->connect || return undef;
    }

    # execute the command on the different power server types
    if ( $RconsPorts::rpServer{$self->{Server}}{"Type"} =~ /cyclades/i ) { # cyclades
	my $prompt = "power\]\$ ";
	$self->{expect}->send("pmCommand $RconsPorts::rpServer{$self->{Server}}{Port} cycle $params{Port}\n");
	if ( $self->{expect}->expect($timeout, $prompt) ) {
	    my $output = $self->{expect}->exp_before();
	    if ( $output !~ /$params{Port}: Outlet turned off/i ) {
		print "$name Power Server \"$self->{Server}\" encountered an error turning off port \"$params{Port}\"\n\"$output\"\n";
		return undef;
	    }
	    if ( $output !~ /$params{Port}: Outlet turned on/i ) {
		print "$name Power Server \"$self->{Server}\" encountered an error turning on port \"$params{Port}\"\n\"$output\"\n";
		return undef;
	    }

	}
    } elsif ( $RconsPorts::rpServer{$self->{Server}}{"Type"} =~ /blackbox/i ) { # blackbox
	my $prompt = 'NPS> ';
	$self->{expect}->send("/BOOT $params{Port}\n\r");
	if ( $self->{expect}->expect($timeout, $prompt) ) {
	    my $output = $self->{expect}->exp_before();
	    if ( $output =~ /Invalid Plug Specified/i ) {
		print "$name Power Server \"$self->{Server}\" did not like port \"$params{Port}\"\n\"$output\"\n";
		return undef;
	    }
	}
	$self->{expect} = undef; # we need to disconnect from a blackbox unit
    } else {
	print "$name Unsupported type \"$RconsPorts::rpServer{$self->{Server}}{Type}\"\n";
	return undef;
    }

    # now check the status again.  verify that the port is off
    @status = $self->status(Port => $params{Port});
    if ( $status[$params{Port}] !~ /,on/i ) {
	print "$name CYCLE call must have failed.  Port does not appear to be ON. \"$status[$params{Port}]\"\n";
	return undef;
    }
    
    return 1;
}

# power on a specified machine or server/port combo
# if a machine isn't specified, a port may be passed in to override the default port to operate on
# returns 1 on success, undef on failure
sub on {
    my $self = shift;
    my %params = (
	Port	=>  $self->{Port},
	Force	=>  undef,
	@_);
    my $name = "Power::on:";
    my $timeout = 10;
    my @status;

    # if a Port was passed in and there is no name assigned - use the passed in port instead of the port in object
    if ( defined $self->{Machine} ) {
	if ( $params{Port} ne $self->{Port} ) {
	    print "$name Port \"$params{Port}\" passed in, but Machine is defined as \"$self->{Machine}\".  Using port \"$self->{Port}\" from RconsPorts instead of user specified port.\n";
	    $params{Port} = $self->{Port};
	}
    }
    
    # make sure we have a port
    if ( !defined $params{Port} ) {
	print "$name No port to operate on.  Cannot execute command.\n";
	return undef;
    }

    # check the port status before the operation
    @status = $self->status(Port => $params{Port});
    if ( defined $status[$params{Port}] ) {
	if ( !defined $params{Force} && ( $status[$params{Port}] =~ /,on/i ) ) {
	    print "$name Port \"$params{Port}\" is already ON.  No need to turn it ON.\n" if $self->{debug};
	    return 1;
	}
    } else {
	print "$name Unable to get status for port \"$params{Port}\" Cannot continue\n", @status;
	return undef;
    }
    
    # if we have a machine name we need to make sure the port name matches it
    if ( defined $self->{Machine} ) {
	my ( $pmPort, $machineName, $powerStatus, $lockStatus ) = split(/,/, $status[$params{Port}]);
	if ( ( $self->{relaxnamecheck} == 1 ) &&  ($machineName =~ /\(undefined\)/i) ) {
	    warn "$name The name on the port is undefined, you could possibly be powering on the wrong machine,\nor the Cyclades is hosed up again.  Continuing anyway.\n";
	} elsif ( $machineName !~ /$self->{Machine}/i ) {
	    print "$name The machine name \"$self->{Machine}\" does not match the name on the port \"$machineName\".\nRefusing to power on.  Please correct the configuration of this port.\n";
	    return undef;
	}
    }

    # if we aren't connected, connect (blackbox needs to disconnect after every command)
    if ( !defined $self->{expect} ) {
	$self->connect || return undef;
    }

    # execute the command on the different power server types
    if ( $RconsPorts::rpServer{$self->{Server}}{"Type"} =~ /cyclades/i ) { # cyclades
	my $prompt = "power\]\$ ";
	$self->{expect}->send("pmCommand $RconsPorts::rpServer{$self->{Server}}{Port} on $params{Port}\n");
	if ( $self->{expect}->expect($timeout, $prompt) ) {
	    my $output = $self->{expect}->exp_before();
	    if ( $output !~ /$params{Port}: Outlet turned on/i ) {
		print "$name Power Server \"$self->{Server}\" encountered an error turning on port \"$params{Port}\"\n\"$output\"\n";
		return undef;
	    }
	}
    } elsif ( $RconsPorts::rpServer{$self->{Server}}{"Type"} =~ /blackbox/i ) { # blackbox
	my $prompt = 'NPS> ';
	$self->{expect}->send("/ON $params{Port}\n\r");
	if ( $self->{expect}->expect($timeout, $prompt) ) {
	    my $output = $self->{expect}->exp_before();
	    if ( $output =~ /Invalid Plug Specified/i ) {
		print "$name Power Server \"$self->{Server}\" did not like port \"$params{Port}\"\n\"$output\"\n";
		return undef;
	    }
	}
	$self->{expect} = undef; # we need to disconnect from a blackbox unit
    } else {
	print "$name Unsupported type \"$RconsPorts::rpServer{$self->{Server}}{Type}\"\n";
	return undef;
    }

    # now check the status again.  verify that the port is off
    @status = $self->status(Port => $params{Port});
    if ( $status[$params{Port}] !~ /,on/i ) {
	print "$name ON call must have failed.  Port does not appear to be ON. \"$status[$params{Port}]\"\n";
	return undef;
    }
    
    return 1;
}

# get the status of a plug(s)
# if port is passed in as "undef" we will get the status of all ports
# returns status string or array of status strings
sub status {
    my $self = shift;
    my %params = (
	Port	=>  $self->{Port},
	@_);
    my $name = "Power::status:";
    my $timeout = 10;
    my @status;

    # if we have a port, it must be numeric
    if ( defined $params{Port} && ( $params{Port} !~ /^\d+$/ ) ) {
	print "$name Port passed in \"$params{Port}\" is not numeric.\n";
	return undef;
    }

    # if we aren't connected, connect (blackbox needs to disconnect after every command)
    if ( !defined $self->{expect} ) {
	$self->connect() || return undef;
    }

    # get the server/port status 
    if ( $RconsPorts::rpServer{$self->{Server}}{"Type"} =~ /cyclades/i ) { # cyclades
	my $prompt = "power\]\$ ";
	$status[0] = "pmPort,machineName,powerStatus,lockStatus";
	if ( defined $params{Port} ) {
	    $self->{expect}->send("pmCommand $RconsPorts::rpServer{$self->{Server}}{Port} status $params{Port}\n");
	} else {
	    $self->{expect}->send("pmCommand $RconsPorts::rpServer{$self->{Server}}{Port} status all\n");
	}
	if ( $self->{expect}->expect($timeout, $prompt) ) {
	    foreach my $line ( split /\n/, $self->{expect}->exp_before ) {
		if ( $line =~ /^\s+(\d+)\s+([\w+|\d+]+)\s+(\w+)\s+(\w+)\s+(\w+)/ ) {
		    $status[$1] = "$1,$2,$4,$3";
		} elsif ( $line =~ /^\s+(\d+)\s+(\w+)\s+(\w+)\s+(\w+)/ ) {
		    $status[$1] = "$1,(undefined),$3,$2";
		}
	    }
	} else {
	    print "$name error trying to get status. " . $self->{expect}->exp_error() . "\n";
	    return undef;
	}
    } elsif ( $RconsPorts::rpServer{$self->{Server}}{"Type"} =~ /blackbox/i ) { # blackbox
	my $prompt = 'NPS> ';
	$status[0] = "pmPort,machineName,powerStatus";
	$self->{expect}->send("/S\n\r");
	if ( $self->{expect}->expect($timeout, $prompt) ) {
	    foreach my $line ( split /\n/, $self->{expect}->exp_before ) {
		if ( $line =~ /^\s+(\d+)\s+.\s+(\w+|.\w+.|.\w+.\w+.)\s+.\s+(\w+)/ ){
		    $status[$1] = "$1,$2,$3";
		}
	    }
	} else {
	    print "ERROR\n" . $self->{expect}->exp_before;
	}
	$self->{expect} = undef; # we need to disconnect from a blackbox unit
    } else {
	print "$name Unsupported type \"$RconsPorts::rpServer{$self->{Server}}{Type}\"\n";
	return undef;
    }
    return wantarray ? @status : $status[$params{Port}];
}

# get any available stats
sub stats {
}

# query the RconsPorts file and validate what we have against what is here
sub validateRconsPortsInfo {
    my $self = shift;
    my $name = "Power::validateRconsPortsInfo:";

    # If a machine name was passed in, check the RconsPorts file for all the relevant information.  Validate
    # the port and server that were passed in as well in case something on the caller.  If they weren't passed in at
    # all, fill them in.
    if ( defined $self->{Machine} ) {
	if ( defined $RconsPorts::machPort{$self->{Machine}} ) {
	    print "$name Found Machine $self->{Machine} in RconsPorts file.\n" if $self->{debug};
	    my $mach = \%{$RconsPorts::machPort{$self->{Machine}}};
	    if ( defined $self->{Server} ) {  # Server passed in
		if ( $self->{Server} ne $$mach{Power}[0] ) {
		    print "$name Server passed in was \"$self->{Server}\" but the RconsPorts file says machine \"$self->{Machine}\" is on Power Server \"$$mach{Power}[0]\"\n";
		    return undef;
		} else {
		    print "$name Server passed in matches Server in RconsPorts file.\n" if $self->{debug};
		}
	    } else { 
		$self->{Server} = $$mach{Power}[0];
	    }
	    if ( defined $self->{Port} ) { # Port passed in
		if ( $self->{Port} ne $$mach{Power}[1] ) {
		    print "$name Port passed in was \"$self->{Port}\" but the RconsPorts file says machine \"$self->{Machine}\" is on Port \"$$mach{Power}[1]\" of Power Server \"$self->{Server}\".\n";
		    return undef;
		} else {
		    print "$name Port passed in matches port in RconsPorts file.\n" if $self->{debug};
		}
	    } else {
		$self->{Port} = $$mach{Power}[1];
	    }
	} else {
	    print "$name Could not find Machine $self->{Machine} in RconsPorts file.\n";
	    return undef;
	}
    }
    
    # if were were given a server, but not a name (IE we didn't fill in stuff from above)
    if ( defined $self->{Server} && !defined $self->{Machine}) {
	if ( defined $RconsPorts::rpServer{$self->{Server}} ) {
	    print "$name Found Server $self->{Server} in ports file.\n" if $self->{debug};
	} else {
	    print "$name Could not find Server $self->{Server} in RconsPorts file.\n";
	    return undef;
	}
    }

    return 1;
}

1;
