#Hey emacs, this is basically -*- perl -*-
# $Id$
#Copyright (c) 2005 Network Appliance, Inc.
#All rights reserved.
#
## @summary Library of functions to standardize printing
##
## @description
## This package implements a Print object. The object has 4
## main methods: Print, SetVerbosity, SetLogVerbosity and SetLogFile
##
## @author      snider@netapp.com
##
## @status public
##
## @pod here

use strict;
use warnings;
#***********************************************************************
# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDEPRINT}) {
    my $newModule = $ENV{OVERRIDEPRINT};
    delete $ENV{OVERRIDEPRINT};
    print "The main Print module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package Print;
use Term::ANSIColor;
use Options;
use Carp qw(confess);

=head1 NAME

Print - Basic Print Functions

=head1 SYNOPSIS

  use Print;
  my $print = $print->new();
  $print->SetVerbosity(verbosity=>2);
  $print->Print(min=>1, message=>"This message shows up at verbosity level\
          1 or higher\n");

=cut

my $Options;
my $Verbosity;
my $Log_verbosity;
my $Header = 1;
my $Log_handle;
my $Self;
my $Program;

BEGIN {
    $Verbosity = 3;
    $Log_verbosity = 3;
    $Options = new Options;
    $Options->parameters(definitions=>
			 [
			  {
			      name=>"verbosity",
			      spec=>"i",
			      subroutine=>[\&SetVerbosity, undef],
			      description=>"change the level of output",
			      default=>$Verbosity,
			  },
			  {
			      name=>"logverbosity",
			      spec=>"i",
			      subroutine=>[\&SetLogVerbosity, undef],
			      description=>"change the level of log output",
			      default=>$Log_verbosity,
			  },
			  {
			      name=>"color",
			      spec=>"!",
			      default=>1,
			      description=>"enable color printing or not",
			      subroutine=>[\&_disable_color],
			  },
			  {
			      name=>"program",
			      description=>"the name of the calling script",
			  },
			  {
			      name=>"header",
			      spec=>"!",
			      default=>1,
			      description=>"enable the date and script header",
			      subroutine=>[\&_disable_header],
			  },
			  ]
			 );
}

# set auto reset to true
$Term::ANSIColor::AUTORESET = 1;

my %colors = (
    good	=> "green",
    bad		=> "red",
    warn	=> "yellow", 
    normal	=> "clear",
    info	=> "cyan",
    debug	=> "white on_blue",
    threadinfo	=> "blue on_cyan",
    label	=> "magenta",
    bench	=> "blue on_white",
);


#***********************************************************************

=head1 C<new()>

=over 4

=item Description

This is the constructor for the object

=item Arguments

=over 4

=item verbosity

This setting allows the user to set the verbosity at invocation

=item log_verbosity

This setting allows the user to set the log_verbosity at invocation

=item output_handler

This setting allows the user to pass a subroutine reference which
is called for each Print statement

=item convert_output

This setting allows the user to cause the output to be converted to HTML

=item log_file

This setting allows the user to cause the output to be logged to a file

=back

=item Returns

Returns the object

=back

=cut

#***********************************************************************
sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {verbosity => undef, # optional
	        log_verbosity => undef, # optional
                output_handler => undef, # optional
		convert_output => undef, # ?
		log_file => undef,
		options => undef,
		@_, # get any parameters that were explicitly passed
	        };
    bless($self, $class); 

    # check to see if we are running inside of NATE
    if (exists($main::{'Tharn::'})) {
	$self->{NATE} = 1;
	# If this file is running inside of NATE and logging to a file
	#  then set the default verbosity level to 3
	$Verbosity = 3 if ($ENV{LOGDIR} ne '');
    }
    # when running inside of nate, $0 is set to 'full script options' so
    #  we need to first split this by spaces and split the first element
    #  (which is the actual script) by the '/' character and take the
    #  last element which will leave us with the real script name
    
    my $full_script = (split(/\s+/, $0))[0];
    my @program = split("/", $full_script);
    $Program = pop @program;

    if ($Self) {
	if (defined($self->{verbosity}) && $self->{verbosity} != $Verbosity) {
	    confess "Print->new() was called with verbosity set to $self->{verbosity} but the class already exists with verbosity $Verbosity";
	}
	return $Self;
    }

    $Verbosity = $self->{Verbosity} if (defined($self->{verbosity}));
    $Log_verbosity = $self->{log_verbosity} if (defined($self->{log_verbosity}));
    $self->SetLogFile(log_file=>$self->{log_file}) if ($self->{log_file});

    $Self = $self;
    $| = 1;
    return $self;
} # end new

#***********************************************************************

=head1 C<Print()>

=over 4

=item Description

This method will print the message only when the current verbosity setting
is at the minimum specified (or higher). It will prepend the message with
a timestamp and the name of the program.

=item Arguments

=over 4

=item min

This defines the minimum verbosity level for which the message should be 
printed. A negative value will not print the date and program.

=item message

This is the message that should be printed.

=item status

This field will cause the message to be printed in different colors.

=item die

This field, if set, will cause the routine to print the message in
a die statement

=back

=item Returns

1 unless it dies

=back

=cut

#***********************************************************************
sub Print {
    my $self = shift;
    my %params = (
		  min => 1,
		  message => undef,
		  status => undef,
		  die => undef,
		  @_);

    my $status = $params{status};
    my $verbosityMin = $params{min};
    my $message = $params{message};
    my %output = ("stdout" => ""); 
    my $color = undef;
    my $header = $Header;

    if ($self->{NATE}) {
	$Program = Tharn::param("program", -default, $Program);
    }
    # if there was no status passed in we assume normal.  if there
    # is no entry in the %colors for the status passed in, we assume the
    # caller is passing a color explicitly and we assign whatever they passed
    # for the call as the color (without checking it to be sure its valid).
    # the exception is the "html" status which means we are going to be doing
    # an ANSI -> HTML conversion and we want the color to be normal, but not
    # touch the status
    $status = "normal" if (!$status);
    if ( defined $colors{$status} ) {
	$color = $colors{$status};
    } else {
	$color = $status;
    }
    # set justPrint so we skip the pretty headers
    if ( $verbosityMin < 0 ) {
	$verbosityMin = 0 - $verbosityMin;
	$header = 0;
    }
    
    # Under CYGWIN, the localtime call in Perl always returns UTC and not the
    # regional time.  So this is a little hack to get around that.  This will only
    # muck with Cygwin on W2K, because under XP the problem doesn't seem to exist.
    # (technically I believe I saw this on W2K under Active Perl too, but I can't reproduce
    # that now so I'm just limiting the change to Cygwin)
    my $time = scalar localtime();
    if (($^O =~ /cygwin/) && (`uname -s` =~ /CYGWIN_NT-5.0/i)) {
	$time = scalar localtime(time()-18000);
    }
    # If you'd like to process your output, this is where it gets handed off to the handler of 
    # your choosing.  Note that this is pre-coloring, so you won't get bright red text passed 
    # to your handler. Your handler will get only one argument.
    if ( defined( $self->{output_handler} ) ) { 
	if ( $Verbosity >= $verbosityMin ) {
	    $self->{output_handler}($time." $Program " . $message);
	}
	return 1;
    }
    # clean up the message - we'll push EOL back on later
    my $eoln = undef;
    if ($message && substr($message, -1) eq "\n") {
	$eoln = 1;
	chomp($message);
    }

    # pop on the header
    if ( $header ) {
	$output{stdout} .= $time." $Program: ";
	$output{log} .= $time." $Program: " if ( defined $Log_handle );
    }

    # should we be printing ANSI, or HTML to stdout?
    if ($self->{convert_output}) {
	# we want the log to be unconverted all the time
	$output{log} .= $message if ( defined $Log_handle );
	# strip the ANSI sequenses and replace with HTML coloring
	$output{stdout} .= $self->ANSItoHTML($message);
    } else { # no conversion needed
	$output{stdout} .= colored($message, $color);
	$output{log} .= colored($message, $color) if ( defined $Log_handle );
    }

    if ($eoln) {
	$output{stdout} .= "\n";
	$output{log} .= "\n" if ( defined $Log_handle );
    }

    confess $output{stdout} if ($params{die});
    print $output{stdout} if ( $Verbosity >= $verbosityMin);

    if ($self->{NATE}) {
#	logcomment($output{log}) if ($output{log} && $Log_verbosity >= $verbosityMin);
    } else {
	# if there is a file handle open, we print to it
	print $Log_handle $output{log} if ($Log_handle && ($Log_verbosity >= $verbosityMin));
    }
    return 1;
} # end Print

#***********************************************************************

=head1 C<SetVerbosity()>

=over 4

=item Description

function to set the Verbosity in the object

=item Arguments

=over 4

=item verbosity

This specifies the new verbosity setting for the program

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub SetVerbosity {
    my ($self, $verbosity) = @_;
    $self = $Self if (!$self);

    $Verbosity = $verbosity;
    $self->Print(min=>4, message=>"Print::SetVerbosity verbosity level now set to $Verbosity\n");
    return 1;
} # end SetVerbosity

#***********************************************************************

=head1 C<SetLogVerbosity()>

=over 4

=item Description

function to set the log verbosity

=item Arguments

=over 4

=item verbosity

This specifies the new log verbosity setting for the program

=back

=item Returns

List the return values

=back

=cut

#***********************************************************************
sub SetLogVerbosity {
    my ($self, $verbosity) = @_;
    $self = $Self if (!$self);

    $Log_verbosity = $verbosity;
    $self->Print(min=>4, message=>"Print::SetLogVerbosity verbosity level now set to $Log_verbosity\n");
    return 1;
} # end SetLogVerbosity

#***********************************************************************

=head1 C<SetLogFile()>

=over 4

=item Description

function to setup the log file

=item Arguments

=over 4

=item log_file

This value will cause a logfile to be created and all output at 
the log_verbosity setting will be appended to this logfile. If set
to 0 or "", the file will be set to /tmp/Program.$$

=back

=item Returns

nothing

=back

=cut

#***********************************************************************
sub SetLogFile {
    my $self = shift;
    my %params = (log_file => undef,
		  @_, # get any parameters that were explicitly passed
		  );

    my $file = $params{log_file};
    if ( !$file || $file eq "" ) { # if told to save output without a file
	$file = "/tmp/$Program.$$";
    }
    $self->Print(min=>2,message=>"Saving output to $file\n");
    if ( -e $file ) {
	$self->Print(min=>1, message=>"Output file $file already exists and will be overwritten.\n", status=>"warn");
    }
    if (open($Log_handle, ">$file")) {
	my $oldfh = select $Log_handle;
	$| = 1;
	select $oldfh;
    } else {
	$self->Print(min=>1,message=>"Could not open $file\n", status=>"bad");
    }
}

sub _disable_color {
    my $color = shift;
    # turn off colors if we need to.
    if ( !$color ) {
	$Self->Print(min=>3, message=>"Print::_disable_color: Disabling color output via the ANSI_COLORS_DISABLED environment variable.\n");
	$ENV{ANSI_COLORS_DISABLED} = "true";
    }
}

sub _disable_header {
    my $header = shift;

    $Header = $header;
}

1;
