# $Id: //depot/prod/test/main/lib/Process.pm#1 $ 

## @summary Provides a mechanism for generating specific background
#           functionality from Clients
##
## @owner GX Automation Framework
##
## @author Michael Edmonds (michael.edmonds@netapp.com)
## @author GX Automation Framework (dl-afd@netapp.com)
##
#
# Copyright (c) 2007-2008 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

=pod

=head1 NAME

Process.pm

=head1 DESCRIPTION

C<Process> Provides a mechanism for generating performing Client Specific
    actions in the background.

=head1 See Also

Package inherits from the following modules. See these modules for further 
interface detail.

C<Process::Load> provides such Methods as new(), setup(), cleanup(), valid_io_types()
C<Class.pm>
C<Err.pm>

=head1 SYNOPSIS

    use Process::Load; #Process::Load ISA Process
    my $Hmr_Load = Process::Load->new(type => "hammer", clients => @Clients);
    my $hammer_bin_params = {bin       => "/usr/local/test/bin/hammer",
                            filer_core => 0,
                            options    => ""};
    $Hmr_Load->bin_conf($hammer_bin_params);
    $Hmr_Load->setup(map=>\@maps);
    $Hmr_Load->start();
    
    ...
    # do stuff
    
    my %status = $Hmr_Load->status();
    # check status
    
    # do stuff
    
    ...
    
    $Hmr_Load->stop();
    $Hmr_Load->cleanup();


=head1 Public Methods

=cut

package Process;
use strict;
use Data::Dumper;
use Params::Validate qw(validate SCALAR ARRAYREF BOOLEAN);
use NATE 3.0;

use base qw(Class);

use TracePkg;
our $Trace;

BEGIN
{
    $Trace = TracePkg->new();
}
#end BEGIN

=head2 pause

=over 4

C<pause()>

=item Description

Tries to pause the process.
Note:
    pause *may* not be available for all processes.
    it will return a Err::NA() if called for a Process that can't pause

=item Input

None

=item Output

integer. <ok (no error) | NA (not applicable)>
Method will return fatal or die if other error occurs

=item Examples

    $cthon_load->start();
    
    if($cthon_load->pause() == $cthon_load->NA())
    {
        $Trace->comment(cmd => "pause doesn't exist for this load type.");
    }
    if($cthon_load->resume() == $cthon_load->NA())
    {
        $Trace->comment(cmd => "resume doesn't exist for this load type.");
    }
    
    ---------------------------------------
    
    $hmr_load->start();
    
    $hmr_load->pause();
    ...
    # do something
    ...
    $hmr_load->resume();

=back

=cut

sub pause
{
    $Trace->entry();
    my $self = shift;
    $self->_pause();
    if ($self->is_error())
    {
        if($self->error_code() == $self->NA())
        {
            $Trace->exit(msg=>"return = " . $self->NA());
            return $self->NA();
        }
        else
        {# We Probably already died at this point
            return $self->FATAL();
        }
    }

    $self->_result();
    $Trace->exit(msg=>"return = " . $self->ok());
    return $self->ok();
}
#end pause()

=head2 resume

=over 4

C<resume()>

=item Description

Tries to resume the process.
Note:
    resume *may* not be available for all processes.
    it will return a Err::NA() if called for a Process that can't pause/resume

=item Input

None

=item Output

integer. <ok (no error) | NA (not applicable)>
Method will return fatal or die if other error occurs

=item Examples

    $cthon_load->start();
    
    if($cthon_load->pause() == $cthon_load->NA())
    {
        $Trace->comment(cmd => "pause doesn't exist for this load type.");
    }
    if($cthon_load->resume() == $cthon_load->NA())
    {
        $Trace->comment(cmd => "resume doesn't exist for this load type.");
    }
    
    ---------------------------------------
    
    $hmr_load->start();
    
    $hmr_load->pause();
    ...
    # do something
    ...
    $hmr_load->resume();

=back

=cut

sub resume
{
    $Trace->entry();
    my $self = shift;
    $self->_resume();
    if ($self->is_error())
    {
        if($self->error_code() == $self->NA())
        {
            $Trace->exit(msg=>"return = " . $self->NA());
            return $self->NA();
        }
        else
        {# We Probably already died at this point
            return $self->FATAL();
        }
    }

    $self->_result();
    $Trace->exit(msg=>"return = " . $self->ok());
    return $self->ok();
}
#end resume()


=head2 start

=over 4

C<start(%params)>

=item Description

Some pre-start verification is done if needed
Generates the command string
Some post-start stuff is done if needed

=item Input

=item * clients => <\@Clients | $client>

(Optional)

Default = $self->{client_list} (All Clients)
This Specifies which Clients to start

=item Output

None

=item Example

    $hmr_load->start(); # Start on ALL Clients
    
    ------------------------------------------
    
    $hmr_load->start(clients => $client); # Start on a single client
    
    ------------------------------------------
    
    $hmr_load->start(clients => \@Clients); # Start on a set of clients

=back

=cut

sub start
{
    $Trace->entry();
    my $self = shift;
    
    my $hash = {clients => {type => SCALAR|ARRAYREF, default => $self->{client_list}},
                };
    my %params = validate(@_, $hash);

    $self->_pre_start(clients=>$params{clients});
    if ($self->is_error())
    {
        $Trace->comment(msg=>"ERROR " . Dumper($self->result()));
        return;
    }

    $self->_start(clients=>$params{clients});
    if ($self->is_error())
    {
        $Trace->comment(msg=>"ERROR " . Dumper($self->result()));
        return;
    }
    
    $self->_post_start(clients=>$params{clients});
    if ($self->is_error())
    {
        $Trace->comment(msg=>"ERROR " . Dumper($self->result()));
        return;
    }

    $self->_result();
    $Trace->exit();
}
# end start()

=head2 type

=over 4 

C<type()>

=item Description

returns a string representing the type.

=item Input

None

=item Output

String

=item Example

    use Process::Load;
    my $Hmr_Load = Process::Load->new(type => "hammer", clients => @Clients);
    $Hmr_Load->setup(...);
    $Hmr_Load->start();
    $Trace->comment(msg => "started " . $Hmr_Load->type() . " in the bg.");
    ...
    
    ... if error ...
    $Trace->fatal(msg => "Error occured when doing something while " .
                         $Hmr_Load->type() . " was running.");

=back

=cut

sub type
{
    my $self = shift;
    return $self->{type};
}
# end type()

=head2 results

=over 4 

C<results(%params)>

=item Description

returns an array of Tharn Result Objects depending on the level param.

=item Input

=item * level => <0 | 1 | 2>

(Optional)
Default = 2

0 returns an Array of Array of Result Objects
    This is ALL Result Objects from every Client (if it has any).
    
1 returns an Array of WORST Result Objects
    This is the Worst result from each Client

2 returns an Array of bad Result Objects
    This if the worst result is Non-<PASS | INFO | PARAM>
    you kinda want this to be empty all of the time

**An empty array can be because it's not at a point of reporting a result.
    this *may* occur if this is called while it's running (for some things).

=item * clients => <\@Clients | $client>

(Optional)

Default = $self->{client_list} (All Clients)
This Specifies which Clients to get the results from

=item Output

Array of some sort

=item Example

    use Process::Load;
    my $Hmr_Load = Process::Load->new(type    => Process::Load::HAMMER(),
                                      clients => @Clients);
    ...
    $Hmr_Load->start();
    ...
    my @bad = $Hmr_Load->results();
    if(scalar(@bad) > 0)
    {
        $Trace->fatal(msg=>Dumper(@bad));
    }

    ... OR ...
    my @worst = $Hmr_Load->results(level => 1);
    $Trace->debug(msg=>Dumper(@worst));

    ... OR ...
    my @all = $Hmr_Load->results(level => 0);
    $Trace->debug(msg=>Dumper(@all));


=back

=cut

sub results
{
    $Trace->entry();
    my $self = shift;
    my $hash = {level   => {type => SCALAR|ARRAYREF, default => 2,
                            callbacks=>{test => sub {(int($_[0]) > -1)
                                                                &&
                                                      (int($_[0]) < 3)},}
                           },
                clients => {type => SCALAR|ARRAYREF, default => [keys(%{$self->{clients}})]},
               };
    my %params = validate(@_, $hash);

    $Trace->debug(msg=>"Clients = " . join(", ", @{$params{clients}}));
    my @results;
    foreach my $client (@{$params{clients}})
    {
        my @client_results = $self->{clients}->{$client}->{cmd_obj}->results();
        if(@client_results) # It's ok if there are NO RESULTS
        {
            if($params{level} == 0)# All RESULTS
            {
                push(@results, @client_results);
            }
            elsif($params{level} > 0)
            {
                my $worst_result = NATE::Result::worst(@client_results);
                $Trace->debug(msg=>"worst_result = " . Dumper($worst_result));
                if($worst_result)
                {
                    if($params{level} == 1) # All WORST RESULTS
                    {
                        push(@results, $worst_result);
                    }
                    else # ONLY BAD RESULTS
                    {
                        my $type = $worst_result->type();
                        if(!($type =~ /pass|info|param/i))
                        {
                            $Trace->debug(msg=>"BAD_result = " . Dumper($worst_result));
                            push(@results, $worst_result);
                        }
                    }
                }
            }
        }
    }

    $self->_result();
    $Trace->debug(msg => "results = " . Dumper(@results));
    $Trace->exit();
    return @results;
}
# end results()

=head2 stop

=over 4

C<stop(%params)>

=item Input

=item * hard => <0 | 1>

(Optional)
Default = 0
indicates whether it should take the time to try to stop the load gracefully
a value of 1 just kills the running subtests

=item * status_check => <0 | 1>

(Optional)
Default = 1
1 indicates that you want it to check the status of the clients before stopping
if it checks that status and the load is NOT Running on a Client, it will FATAL

0 does not check the status, it just does the stop

=item * clients => <\@Clients | $client>

(Optional)

Default = $self->{client_list} (All Clients)
This Specifies which Clients stop the load on

=item Output

None

=item Example

    $hmr_load->stop(); # Stop on ALL Clients
    
    ------------------------------------------
    
    $hmr_load->stop(clients => $client); # Stop on a single client
    
    ------------------------------------------
    
    $hmr_load->stop(clients => \@Clients); # Stop on a set of clients

=back

=cut

sub stop
{
    $Trace->entry();
    my $self = shift;
    my $hash = {status_check => {type => BOOLEAN, default => 1,},
                hard         => {type => BOOLEAN, default => 0,},
                clients      => {type => SCALAR|ARRAYREF,
                                 default => [keys(%{$self->{clients}})]},
                };
    my %params = validate(@_, $hash);

    if($params{status_check})
    {
        my $status = $self->status(clients => $params{clients});
        if ($self->is_error())
        {
            $Trace->comment(msg=>"ERROR " . Dumper($self->result()));
            return;
        }

        if(int($status->{stopped}) > 0)
        {
            foreach my $client (@{$status->{stopped}})
            {
                $Trace->warn(msg=>$self->type()." on ".$client." was NOT running".
                             " when it was asked to stop\n");
            }
        }
    }

    if($params{hard})
    {
        my $msg = ["comment", "HARD STOP"];
        foreach my $client (@{$params{clients}})
        {
            $self->{clients}->{$client}->{cmd_obj}->kill(message       => $msg,
                                                         child_message => $msg);
            $self->{clients}->{$client}->{cmd_obj} = undef;
        }
    }
    else
    {
        $self->_stop(clients => $params{clients});
        if($self->is_error() && ($self->error_code() ne $self->NA()))
        {
            $Trace->warn(msg=>"ERROR " . Dumper($self->result()));
            return;
        }
    }
    #$self->{cmd_objs} = undef;
    #delete($self->{cmd_objs});
    $self->_result();
    $Trace->exit();
}
# end stop()

=head2 status

=over 4

C<status(%params)>

=item Description

returns a hash containing all of the nodes that are running and stopped

=item Input

=item * clients => <\@Clients | $client>

(Optional)

Default = $self->{client_list} (All Clients)
This Specifies which Clients get the status of

=item Output

Hash Reference
$status = {clients => integer <quanitity of Clients>,
           running => \@running_Clients,
           stopped => \@stopped_Clients}

=item Examples

    my %status = $hmr_load->status();
    if ($hmr_load->is_error())
    {
       $Trace->fatal(msg=>"ERROR " . Dumper($self->result()));
       return;
    }

    if(int($status{stopped}) > 0)
    {
        foreach(@{$status{stopped}})
        {
            $Trace->fatal(msg => $hmr_load->type() . " on " . $_ .
                          " was NOT running");
        }
    }

=back

=cut

sub status
{
    $Trace->entry();
    my $self = shift;
    my $hash = {clients => {type => SCALAR|ARRAYREF, default => [keys(%{$self->{clients}})]},
                };
    my %params = validate(@_, $hash);

    my @running;
    my @stopped;

    foreach my $client (@{$params{clients}})
    {
        if($self->{clients}->{$client}->{cmd_obj}->is_alive())
        {
            push(@running, $client);
        }
        else
        {
            push(@stopped, $client);
        }
    }

    if(int(@{$params{clients}}) != (int(@running) + int(@stopped)))
    {
        $self->_result(rc  => $self->FAIL(),
                       msg => "something is wrong with the status\n" .
                       "Client  count = " . int(@{$params{clients}}) . "\n" .
                       "Running count = " . int(@running) . "\n" .
                       "Stopped count = " . int(@stopped));
    }

    my %result = (clients => int(@{$params{clients}}),
                  running => \@running,
                  stopped => \@stopped);

    $self->_result();
    $Trace->debug(msg => "status = " . Dumper(\%result));
    $Trace->exit();
    return \%result;
}
# end status()

sub _normalize_for_filenames
{
    $Trace->entry();
    my ($self,$input) = @_;
    my $output;
    # replace nasty chars with '_'
    $output = join("_", split(' ', $input));
    $output = join("_", split('/', $input));
    $output = join("_", split('\\\\', $input));
    $output = join("_", split(':', $input));
    $output = join("_", split('\*', $input));
    $output = join("_", split('\?', $input));
    $output = join("_", split('\"', $input));
    $output = join("_", split('\<', $input));
    $output = join("_", split('\>', $input));
    $output = join("_", split('\|', $input));
    $Trace->exit(msg => "return = " . $output);
    return $output;
}
# end _normalize_for_filenames()

sub __destroy
{
    $Trace->entry();
    my $self = shift;

    my $msg = ["comment", "HARD STOP"];
    foreach my $subtest_obj (@{$self->{cmd_objs}})
    {
        $subtest_obj->kill(message => $msg, child_message => $msg);
    }

    $Trace->exit();
}
# end __destroy()

1;
