# $Id: //depot/prod/test/main/lib/Process/CodeRef.pm#1 $

## @summary Provides a mechanism for executing specific code functions in the background
##
## @owner GX Automation Framework
##
#
# Copyright (c) 2007,2008 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

=pod

=head1 NAME

Process::CodeRef.pm

=head1 DESCRIPTION

C<Process::CodeRef> Provides a mechanism for executing specific functions in the
    background from a set of Clients.

=head1 See Also

Process.pm
Process provides such Public Methods as:
    start()
    stop()
    status()
    pause()
    resume()
    ...

=head1 SYNOPSIS

    use Process::CodeRef;
    my $Check_Epoch = Process::CodeRef->new(code    => \&func,
                                            args    => \@func_args,
                                            verify  => \@stuff_to_verify,
                                            clients => \@Clients);
    $Check_Epoch->start();
    
    ...
    # do stuff
    
    my %status = $Check_Epoch->status();
    # check status
    
    # do stuff
    ...
    
    $Check_Epoch->stop();


=head1 Public Methods

=cut

package Process::CodeRef;
use strict;
use Data::Dumper;
use Params::Validate qw(validate SCALAR CODEREF ARRAYREF);
use B::Deparse;

use TracePkg;
use base qw(Process::Setup);

use Tharn qw ($Staf);
our $Trace;
BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

=head2 new

C<new(%params)>

=over

=item Input

=item * code => \&func

Required)
This is the Code reference to execute on the remote clients

=item * clients => <\@Clients | $Clients[0]>

(Required)
Takes a SCALAR or ARRAY REFERENCE
Clients that you want to execute this code on.

=item * args => <\@args | $args[0]>

(Optional)
Takes a SCALAR or ARRAY REFERENCE
These are the Arguments that need to be passed so that we can execute func(@args).

=item * verify => <\@stuff_to_verify | $thing_to_verify>

(Optional)
Takes a SCALAR or ARRAY REFERENCE
These are the things that will be verified when func(@args) returns.

=item * id => <some string/integer id>

(Optional)
Setting this to something will tweak the runids for the subtests that is spawns.
This is so you can have multiple objects of the same description that do
different things on the same Clients.
e.g.:
new(code => \&func, args => [1,2,3], ..., desc => "func_name", id => "1-3")
new(code => \&func, args => [4,5,6], ..., desc => "func_name", id => "4-6")

=item * desc => <description or function name>

(Optional)
default = "code"
Setting this to something will tweak the runids for the subtests that is spawns.
This is so you can have multiple objects of the same code.


=item Output

Class Reference

=item Examples

    my $Check_Epoch = Process::CodeRef->new(code    => \&epoch_checker,
                                            args    => [$file1, file2],
                                            verify  => 1,
                                            desc    => "epoch_checker",
                                            clients => \@Clients);

=back

=cut

sub new
{
    $Trace->entry();

    my $hash = {clients    => {type => SCALAR|ARRAYREF},
                code       => {type => CODEREF},
                args       => {type => SCALAR|ARRAYREF, default => []},
                iterations => {type => SCALAR,          default => -1},
                verify     => {type => SCALAR|ARRAYREF, default => []},
                id         => {type => SCALAR,          default => ""},
                desc       => {type => SCALAR,          default => "code"},
                error_type => {type => SCALAR,          default => Err::DIE()},
                };
    my $class = shift;
    my %params = validate(@_, $hash);

    my $self = {};

    bless($self, $class);

    #$self->{need_setup} = 1;
    $self->_configure_error(type => $params{error_type});
    
    my $deparse = B::Deparse->new("-p", "-sC");
    my %args = (code => $deparse->coderef2text($params{code}), iterations => $params{iterations});

    # normalize the Clients var
    if(ref($params{clients}) ne "ARRAY")
    {
        $params{clients} = [$params{clients}];
    }

    # normalize the args var
    if(ref($params{args}) ne "ARRAY")
    {
        $params{args} = [$params{args}];
    }
    $args{args} = $params{args};

    # normalize the Verify var
    if(ref($params{verify}) ne "ARRAY")
    {
        $params{verify} = [$params{verify}];
    }
    $args{verify} = $params{verify};

    # normalize the id, if it exists
    if(defined($params{id}))
    {
        # replace nasty chars with '_'
        $params{id} = $self->_normalize_for_filenames($params{id});
    }

    # normalize the description, if it exists
    if(defined($params{desc}))
    {
        # replace nasty chars with '_'
        $params{desc} = $self->_normalize_for_filenames($params{desc});
    }

    $self->{client_list} = $params{clients};
    $self->{type} = "code";
    $self->{cmd} = \%args;
    $self->{runid_appender} = $params{desc} . $params{id} . "_";

    $self->_result();
    $Trace->exit();
    return $self;
}
# end new()

sub _start
{
    $Trace->entry();
    my $self = shift;

    # 7 day timeout
    my $timeout = ((3600*24)*7);
    
    my %args = %{$self->{cmd}};

    foreach my $client (@{$self->{client_list}})
    {
        if((!(defined($self->{clients}->{$client}->{cmd_obj})))
                                ||
            (!($self->{clients}->{$client}->{cmd_obj}->is_alive())))
        {
            my $runid = $self->{runid_appender} . $client;

            if(defined($self->{clients}->{$client}->{cmd_obj}))
            {
                # Let's not overwrite our old files
                $runid = $runid . "_" . rand(time());
            }
            my $cmd_obj;

            $cmd_obj = Tharn::subtest("{" . $self->__run_code() . "}", -bg, 
                                        -client  => $client,
                                        -timeout => $timeout,
                                        -runid   => $runid,
                                        -onexit  => sub {},
                                        "--"     => %args,
                                       );
            $self->{clients}->{$client}->{cmd_obj} = $cmd_obj;
            $Trace->debug(msg=>"Spawned " . $runid . " on " . $client);
            push(@{$self->{cmd_objs}}, $cmd_obj);
        }
        else
        {
            $Trace->comment(msg=>"\n\nWhat's going on??\n" .
                            $client . " has already been started from this " .
                            "object!!\n\n");
        }
    }
    $Staf->sync(flag=>"Process/CodeRef/Start",  peers=> $self->{cmd_objs});
    $Staf->wait_sync(flag=>"Process/CodRef/Start");

    $self->_result();
    $Trace->exit();
}
# end _start()

sub __run_code
{
    my $self = shift;
    # When you get an error here move your cursor to the line below and type NNNj in vim 
    # to go to the NNN'th relative line number
    return << 'SCRIPT';

    use strict;
    use Params::Validate qw(validate SCALAR ARRAYREF);
    use Data::Dumper;
    use TracePkg;
    use Tharn qw($Log $Parent);
    use NATE::Events qw(listener_add eventloop_once);
    use NATE::STAF qw(kOk);

    my $hash = {code       => {type => SCALAR},
                args       => {type => ARRAYREF},
                verify     => {type => ARRAYREF},
                iterations => {type => SCALAR}};
    
    my %params = validate(@_, $hash);
    my @args;
    my @control_vals;
    my $stopped = 0;
    my $pause = 0;
    my $iterations = 1;

    $Log->debug("Eval checking code");
    eval "sub test $params{code}";
    if ($@) {
        my $errstr = "Eval of code failed: $@";
        $Log->error($errstr);
        die $errstr;
    }

    @args = @{$params{args}};
    $iterations = $params{iterations};
   
    sub pause {
        $Log->comment("sync_request_setup for Pause");
        $Parent->sync_request_setup(flag     => "Process/CodeRef/Pause",
                                    callback => sub {
                                        $Log->comment("Got a Pause request");
                                        my ($result) = @_;
                                        NATE::STAF::default_error(%{$result})
                                            if ($result->{rc} != kOk);
                                        $pause = 1;
                                        pause();
                                    });
    }

    sub resume {
        $Log->comment("sync_request_setup for Resume");
        $Parent->sync_request_setup(flag     => "Process/CodeRef/Resume",
                                    callback => sub {
                                        $Log->comment("Got a Resume request");
                                        my ($result) = @_;
                                        NATE::STAF::default_error(%{$result})
                                            if ($result->{rc} != kOk);
                                        $pause = 0;
                                        resume();
                                    });
    }
    
    $Parent->sync_request_setup(flag     => "Process/CodeRef/Stop",
                                callback => sub {
                                    $Log->comment("Got a Stop request");
                                    my ($result) = @_;
                                    NATE::STAF::default_error(%{$result})
                                        if ($result->{rc} != kOk);
                                    $stopped = 1;
                                });


    sub handle_failure
    {
        my $msg = shift;
        $Log->error($msg);                                       
        Tharn::addresult(type => "FAIL", abort => 0, $msg);
        $stopped = 1;
    }

    pause();
    resume();

    $Log->comment("Waiting on sync_request for Process/CodeRef/Start");
    $Parent->sync_request(flag => "Process/CodeRef/Start");
    while(!$stopped && ($iterations != 0))
    {
        my @outputs = test(@args);
        $Log->comment("Output = " . Dumper(@outputs));             
        $iterations--;

        if (scalar(@{$params{verify}}) == 0) {
            $Log->comment("No validation requested");
        } elsif (scalar(@outputs) == scalar(@{$params{verify}})) {
            my @control_vals = @{$params{verify}};
            my $i = 0;
            $Log->comment("Validate: ".Dumper({control => @control_vals, output => @outputs}));
            foreach my $output (@outputs) {
                if (!($output =~ qr/$control_vals[$i]/)) {
                    my $msg = "Mismatch:\n Verify = " . $control_vals[$i] .
                            "\nOutput = " . $output . "\n";
                    handle_failure($msg);
                    exit(0);                   
                }
                $i++;
            }
        } else {
            my $msg = "Output Array Size Mismatch: Verify Array = " .
                        @control_vals . "Output Array = " . @outputs;
            handle_failure($msg);
            exit(0);
        }
           
        # If we're still around, log a PASS
        Tharn::addresult(type => "PASS", msgs => "No Error");
        $Log->comment("Reported a Pass result.\n Looking for Pause/Stop/Resume now");

        do {
            $Log->debug("eventloop_once");
            eventloop_once(timeout=>.25);
        } while($pause);                                                                   #110
    }

    # I don't think this will ever hit
    handle_failure("we stopped early") if(!$stopped);

    exit(0);

SCRIPT
}
# end __run_code()

=head2 wait

=over 4 

C<wait()>

=item Description

Calls Subtest::wait_finish on the client processes. Should only be used if
the iterations parameter is set. 

=back

=cut

sub wait
{
    $Trace->entry();
    my $self = shift;
    Subtest::wait_finish(subtests=>$self->{cmd_objs});
    $self->_result();
    $self->_clear();
    $Trace->exit();    
}

sub _stop
{
    $Trace->entry();
    my $self = shift;
    $Staf->sync(flag => "Process/CodeRef/Stop",peers=> $self->{cmd_objs});
    $Staf->wait_sync(flag => "Process/CodeRef/Stop");

    foreach my $client (keys(%{$self->{clients}}))
    {
        my $subtest_obj = $self->{clients}->{$client}->{cmd_obj};
        $subtest_obj->message_put(type    => "Process/CodeRef",
                                  message => "Leave");
    }
    Subtest::wait_finish(subtests=>$self->{cmd_objs});
    $self->_result();
    $self->_clear();
    $Trace->exit();
}
# end _stop

sub _pause
{
    $Trace->entry();
    my $self = shift;
    my $flag = "Process/CodeRef/Pause";
    $Staf->sync(flag      => $flag, peers => $self->{cmd_objs}); 
    $Staf->wait_sync(flag => $flag);

    $self->_result();
    $Trace->exit();
}
# end _pause

sub _resume
{
    $Trace->entry();
    my $self = shift;
    my $flag = "Process/CodeRef/Resume";
    $Staf->sync(flag      => $flag, peers => $self->{cmd_objs}); 
    $Staf->wait_sync(flag => $flag);
    $self->_result();
    $Trace->exit();
}
# end _resume

sub _pre_start
{
    return;
}
# end _pre_start()

sub _post_start
{
    return;
}
# end _pre_start()

#Clear out any subtest command objects from previous runs
sub _clear
{
    $Trace->entry();
    my $self = shift;

    $Trace->comment(msg=>"clearing client cmd_obj entries");
    #delete($self->{cmd_objs});
    foreach my $client (@{$self->{client_list}}) {
        if (defined($self->{clients}->{$client}->{cmd_obj})) {
            delete ($self->{clients}->{$client}->{cmd_obj});
        } else {
            $Trace->comment(msg=>"$client is not defined! skipping");
        }
    }
    return;
}

DESTROY
{
    my $self = shift;
    $self->__destroy();
}
# end DESTROY

1;



