# $Id: //depot/prod/test/main/lib/Process/Load.pm#1 $

## @summary Provides a mechanism for generating specific background load on Nodes
##
## @owner GX Automation Framework
##
## @author Michael Edmonds (michael.edmonds@netapp.com)
## @author GX Automation Framework (dl-afd@netapp.com)
##
#
# Copyright (c) 2007-2009 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

=pod

=head1 Author

Michael Edmonds

=head1 Owner

GX Automation Framework (dl-afd@netapp.com)

=head1 NAME

Process::Load.pm

=head1 DESCRIPTION

C<Process::Load> Provides a mechanism for generating specific background load on
    Nodes from a set of Clients.

=head1 See Also

Process.pm
Process provides such Public Methods as:
    start()
    stop()
    status()
    ...

Process::Load::IO::<IO TOOL>.pm (e.g. Process::Load::IO::Hammer.pm)
Process::Load::IO::<IO TOOL> provides such Public Methods as:
    bin_conf()

=head1 SYNOPSIS

    use Process::Load;
    my $Hmr_Load = Process::Load->new(type    => Process::Load::HAMMER(),
                                      clients => \@Clients);
    $Hmr_Load->setup(map=>\@maps);
    $Hmr_Load->start();

    ...
    # do stuff

    my %status = $Hmr_Load->status();
    # check status

    # do stuff

    ...

    $Hmr_Load->stop();
    $Hmr_Load->cleanup();


=head1 Public Methods

=cut

package Process::Load;
use strict;
use Data::Dumper;
use Params::Validate qw(validate SCALAR BOOLEAN ARRAYREF);
use Time::HiRes;

use Tharn qw($Log subtest host $Staf);

use TracePkg;
use base qw(Process::Setup);

our $Trace;
BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

=head2 new

=over 4

C<new(%params)>

=item Input

=item * type => <"hammer" | "hammerlocking" | "sio" | "dt" | "cthon" | ...>

(Optional)
Default = "hammer"
this tells Process::Load what type of io to subclass

=item * clients => \@Clients

(Optional)
you can set the list of Clients here or in the mount method.
BE CAREFUL NOT TO have duplicate CLIENTS. (No check for that is done)

=item * id => <some SCALAR id>

(Optional)
Setting this to something will tweak the runids for the subtests that is spawns.
This is so you can have multiple objects of the same IO type.

=item * error_type => <Err::DIE() | Err::RETURN()>

(Optional)
Default = Err::DIE()
this tells Error Handler to either do a die on an error or return no matter what
    and let the user handle the error

=item Output

Class Reference

=item Example

    my $hmr_load = Process::Load->new(type => "hammer", id => "nfsv2");

=back

=cut

sub new
{
    $Trace->entry();

    my $hash = {type       => {type => SCALAR, default => "hammer",
                               callbacks => {
                                          test => sub
                                            {
                                                my $load = shift(@_);
                                                (grep ($_ eq $load, Process::Load::valid_io_types()))
                                            }
                                          }
                            },
                clients    => {type => SCALAR|ARRAYREF, optional => 1},
                id         => {type => SCALAR,  default => rand(Time::HiRes::time())},
                error_type => {type => SCALAR,  default => Err::DIE()},
                interactive=> {type => BOOLEAN, default => 0},
                };

    my $class = shift;
    my %params = validate(@_, $hash);

    my $self = {};

    bless($self, $class);

    $self->_configure_error(type => $params{error_type});
    $self->{deny_hosttype} = [];

    my $type = $params{type};
    if($type eq $self->HAMMER())
    {
        require Process::Load::IO::Hammer;
        Process::Load::IO::Hammer->_new($self);
    }
    elsif($type eq $self->HAMMERLOCKING())
    {
        require Process::Load::IO::HammerLocking;
        Process::Load::IO::HammerLocking->_new($self);
    }
    elsif($type eq $self->SIO())
    {
        require Process::Load::IO::SIO;
        Process::Load::IO::SIO->_new($self);
    }
    elsif($type eq $self->CTHON())
    {
        require Process::Load::IO::CThon;
        Process::Load::IO::CThon->_new($self);
    }
    elsif($type eq $self->RANDOM())
    {
        require Process::Load::IO::Random;
        Process::Load::IO::Random->_new($self);
    }
    elsif($type eq $self->CRUD())
    {
        require Process::Load::IO::Crud;
        Process::Load::IO::Crud->_new($self);
    }
    elsif($type eq $self->CUSTOM())
    {
        require Process::Load::IO::Custom;
        Process::Load::IO::Custom->_new($self);
    }
    elsif($type eq $self->DT())
    {
        require Process::Load::IO::Dt;
        Process::Load::IO::Dt->_new($self);
    }
    elsif($type eq $self->READONLYDT())
    {
        require Process::Load::IO::ReadOnlyDt;
        Process::Load::IO::ReadOnlyDt->_new($self);
    }
    # Normalize Clients
    if((defined($params{clients})) && (ref($params{clients}) ne "ARRAY"))
    {
        $params{clients} = [$params{clients}];
    }

    # Normalize id, if it exists
    if(defined($params{id}))
    {
        # replace nasty chars with '_'
        $params{id} = $self->_normalize_for_filenames($params{id});
    }

    $self->{client_list} = $params{clients};
    $self->{interactive} = $params{interactive};
    $self->{type} = $type;
    $self->{runid_appender} = $self->{type} . $params{id} . "_";
    $self->{id} = $params{id};
    $self->bin_conf();

    $self->_result();
    $Trace->exit();
    return $self;
}
# end new()

=head2 is_running

=over 4

C<is_running(%params)>

=item Input

=item * clients => \@Clients

(Optional)

=item Output

Count of clients where load is running.

=item Example

    my $is_running = $load->is_running();

=back

=cut

sub is_running
{
    $Trace->entry();
    my $self = shift;
    my %params = validate(
        @_,
        {
            clients => {
                type => SCALAR|ARRAYREF,
                default => $self->{client_list},
            },
        }
    );
    my $clients = $params{clients};

    # Normalize Clients
    if (!ref($clients)) {
        $clients = [$clients];
    }

    my $alive_count = 0;
    foreach my $client (@$clients) {
        if (defined($self->{clients}->{$client}->{cmd_obj})) {
            eval {
                if ($self->{clients}->{$client}->{cmd_obj}->is_alive()) {
                    $alive_count++;
                }
            };
        }
    }

    $Trace->exit();
    return $alive_count;
}
# end is_running()

sub _pre_start
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => [keys(%{$self->{clients}})]},});

    $params{clients} = [$params{clients}] if(ref($params{clients}) ne "ARRAY");

    if($self->_need_setup())
    {
        $self->_check_connected(clients => $params{clients});
        return if ($self->is_error());
    }

    $self->_gen_cmd(clients => $params{clients});
    return if ($self->is_error());

    $self->__pre_start(clients => $params{clients});
    return if ($self->is_error());

    $self->_result();
    $Trace->exit();
}
# end _pre_start()

sub _check_connected
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => [keys(%{$self->{clients}})]},});

    foreach my $client (@{$params{clients}})
    {
        if(!($self->{clients}->{$client}->{info}->{is_connected}))
        {
            $self->_result(rc => $self->FATAL(), msg => $client .
                           " isn't mounted/mapped/connected yet!");
            return;
        }
    }
    $self->_result();
    $Trace->exit();
}
# end _check_connected

sub _start
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => [keys(%{$self->{clients}})]},});

    my $timeout = 3600 * 24 * 7; # 1 week
    my %username;
    # CTHON needs to be root
    %username = (-username => "root") if($self->type() eq $self->CTHON());

    my @new_subtest;
    foreach my $client (@{$params{clients}})
    {
        my $cmd = $self->{clients}->{$client}->{cmd};
        $cmd->{id} = $self->{id};
        $Trace->debug(msg=>"Command = " . Dumper($cmd));
        $Trace->debug(msg=>"Client = "  . $client);
        if (!$self->is_running(clients => $client)) {
            my $runid = $self->{runid_appender} . "execute_" . $client . "_" .
                        $self->{clients}->{$client}->{random};

            if(defined($self->{clients}->{$client}->{cmd_obj}))
            {
                # Let's not overwrite our old files
                if(!(exists($self->{clients}->{$client}->{index})))
                {
                    $self->{clients}->{$client}->{index} = 1;
                }
                my $index = $self->{clients}->{$client}->{index};
                $runid = $runid . "_" . $index;
                $self->{clients}->{$client}->{index} = $index++;
            }
            if(!(defined($self->{clients}->{$client}->{host_obj})))
            {
                $self->{clients}->{$client}->{host_obj} = host($client);
            }
            my $cmd_obj = subtest("{".$self->_execute()."}", -bg,
                                  -client  => $self->{clients}->{$client}->{host_obj},
                                  -timeout => $timeout,
                                  -runid   => $runid,
                                  -onexit  => sub {},
                                  %username,
                                  "--"     => %{$cmd}
            );

            $Trace->debug(msg=>"subtested start for " . $runid);
            $self->{clients}->{$client}->{cmd_obj} = $cmd_obj;
            push(@new_subtest, $cmd_obj);
        }
        else
        {
            $Trace->comment(msg => "\n\nWhat's going on??\n" .
                                    $client . " has already been started from this " .
                                    "object!!\n\n");
        }
    }
    push(@{$self->{cmd_objs}}, @new_subtest);
    $Staf->sync(flag  => "Process/Load/IO/".$self->{type}."/Start/".$self->{id},
                peers => \@new_subtest, timeout => 600);

    $self->_result();
    $Trace->exit();
}
# end _start()

sub _post_start
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => $self->{client_list}},});

    $self->__post_start(clients => $params{clients});
    return if ($self->is_error());

    $self->_result();
    $Trace->exit();
}
# end _post_start()

sub _stop
{
    $Trace->entry();
    my $self = shift;
    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => [keys(%{$self->{clients}})]},});

    my @wait_clients;
    $params{clients} = [$params{clients}] if(ref($params{clients}) ne "ARRAY");

    my @subtest_objs = map {$self->{clients}->{$_}->{cmd_obj}} @{$params{clients}};
    $Staf->sync(flag  => "Process/Load/IO/".$self->{type}."/Stop/".$self->{id},
                peers => \@subtest_objs);
    Subtest::wait_finish(subtests => \@subtest_objs);
    $self->_result();
    $Trace->exit();
}
# end _stop

sub __pre_start
{
    return;
}
# end __pre_start

sub __post_start
{
    return;
}
# end __post_start

sub _pause
{
    $Trace->entry();
    my $self = shift;
    $self->_result(rc => $self->NA(), msg => "No pause in " . $self->type());
    $Trace->exit();
}
# end _pause

sub _resume
{
    $Trace->entry();
    my $self = shift;
    $self->_result(rc => $self->NA(), msg => "No resume in " . $self->type());
    $Trace->exit();
}
# end _resume


=head2 HAMMER

=over 4

C<HAMMER()>

=item Description

Used as a const (but easier to use w/ an IDE).
Just returns "hammer"

=item Input

None

=item Output

String "hammer"

=item Example

    Process::Load->new(type => Process::Load::HAMMER());

=back

=cut

sub HAMMER
{
    return "hammer";
}
#end HAMMER()

=head2 HAMMERLOCKING

=over 4

C<HAMMERLOCKING()>

=item Description

Used as a const (but easier to use w/ an IDE).
Just returns "hammerlocking"

=item Input

None

=item Output

String "hammerlocking"

=item Example

    Process::Load->new(type => Process::Load::HAMMERLOCKING());

=back

=cut

sub HAMMERLOCKING
{
    return "hammerlocking";
}
#end HAMMERLOCKING()

=head2 SIO

=over 4

C<SIO()>

=item Description

Used as a const (but easier to use w/ an IDE).
Just returns "sio"

=item Input

None

=item Output

String "sio"

=item Example

    Process::Load->new(type => Process::Load::SIO());

=back

=cut

sub SIO
{
    return "sio";
}
#end SIO()

=head2 CTHON

=over 4

C<CTHON()>

=item Description

Used as a const (but easier to use w/ an IDE).
Just returns "cthon"

=item Input

None

=item Output

String "cthon"

=item Example

    Process::Load->new(type => Process::Load::CTHON());

=back

=cut

sub CTHON
{
    return "cthon";
}
#end CTHON()

=head2 RANDOM

=over 4

C<RANDOM()>

=item Description

Used as a const (but easier to use w/ an IDE).
Just returns "random"

=item Input

None

=item Output

String "random"

=item Example

    Process::Load->new(type => Process::Load::RANDOM());

=back

=cut

sub RANDOM
{
    return "random";
}
#end RANDOM()

=head2 CRUD

=over 4

C<CRUD()>

=item Description

Used as a const (but easier to use w/ an IDE).
Just returns "crud"

=item Input

None

=item Output

String "crud"

=item Example

    Process::Load->new(type => Process::Load::CRUD());

=back

=cut

sub CRUD
{
    return "crud";
}
#end CRUD()

=head2 DT

=over 4

C<DT()>

=item Description

Used as a const (but easier to use w/ an IDE).
Just returns "dt"

=item Input

None

=item Output

String "dt"

=item Example

    Process::Load->new(type => Process::Load::DT());

=back

=cut

sub DT
{
    return "dt";
}
#end DT()

=head2 READONLYDT

=over 4

C<READONLYDT()>

=item Description

Used as a const (but easier to use w/ an IDE).
Just returns "readonlydt"

=item Input

None

=item Output

String "readonlydt"

=item Example

    Process::Load->new(type => Process::Load::READONLYDT());

=back

=cut

sub READONLYDT
{
    return "readonlydt";
}
#end DT()

=head2 CUSTOM
C<CUSTOM()>

=item Description

Used as a const (but easier to use w/ an IDE).
Just returns "custom"

=item Input

None

=item Output

String "custom"

=item Example

    Process::Load->new(type => Process::Load::CUSTOM());

=back

=cut

sub CUSTOM
{
    return "custom";
}
#end CUSTOM()

=head2 valid_io_types

C<valid_io_types()>

lets you know what loads are availible incase you need to iterate through all of them

=over 4

=item Input

None

=item Output

Array of Avialible Load Types

=item Example

    my @availible_loads = Process::Load::valid_io_types();
    my @Loads;
    foreach(@availible_loads)
    {
        my $load_obj = Process::Load->new(type => $_);
        push(@Loads, $load_obj);
    }
    ...

=back

=cut

sub valid_io_types
{
    return qw(hammer hammerlocking cthon random sio crud readonlydt dt custom);
}
# end valid_io_types()

DESTROY
{
    my $self = shift;
    $self->__destroy();
}
# end DESTROY

1;
