# $Id: //depot/prod/test/main/lib/Process/Load/IO/CThon.pm#1 $ 

## @summary Allows Process::Load to perform Load from CThon
##
## @owner GX Automation Framework
##
## @author Michael Edmonds (michael.edmonds@netapp.com)
## @author GX Automation Framework (dl-afd@netapp.com)
##
#
# Copyright (c) 2007-2008 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

=pod

=head1 Author

Michael Edmonds

=head1 Owner

GX Automation Framework (dl-afd@netapp.com)

=head1 NAME

Process::Load::IO::CThon.pm

=head1 DESCRIPTION

C<Process::Load::IO::CThon> takes care of cthon specific commands.

=head1 Public Methods

=cut

package Process::Load::IO::CThon;
use strict;
use Data::Dumper;
use Params::Validate qw(validate SCALAR ARRAYREF);

use Tharn;
use TracePkg;
use base qw(Process::Load);

our $Trace;

BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

sub _new
{
    $Trace->entry();

    my $class = shift;

    my $self = shift;

    bless($self, $class);

    $self->_result();
    $self->{deny_hosttype} = ["win"];
    $Trace->debug(msg=>"self = " . Dumper($self));
    $Trace->exit();
    return $self;
}
# end _new()

=head2 B<bin_conf(%params)>

=over 4

=item Description

This Configures CThon parameters.
This Method is Optional.
Only needs to be called if you want to modify the defaults.

=item Inputs

=item * bin_base => <Path to Base of the cthon executable>

(Optional)
Default = "/usr/software/test/bin/cthon04"
This is the NOT the Binary to use, but the base.

=item * group => <a | b | g | s | l>

(Optional)
Default = a

a == "basic general special lock"
b == "basic"
g == "general"
s == "special"
l == "lock"

=item * type => <t | f>

(Optional)
Default = t

t == "timing"
f == "functional"

=item Output

None

=item Example

    use Process::Load;
    my $Cthon_Load = Process::Load->new(type => "cthon", clients => @Clients);
    my %cthon_bin = (bin_base => "/usr/sotware/test/bin/cthon04",
                     group    => 'a',
                     type     => 't');
    $Cthon_Load->bin_conf(%cthon_bin);
    ...

=back

=cut

sub bin_conf
{
    $Trace->entry();
    my $self = shift;
    my $hash = {bin_base=> {type    => SCALAR,
                            default => "/usr/software/test/bin/cthon04"},
                group   => {type      => SCALAR,
                            default   => 'a',
                            callbacks => {test => sub {($_[0] eq 'a')
                                                              ||
                                                        ($_[0] eq 'b')
                                                              ||
                                                        ($_[0] eq 'g')
                                                              ||
                                                        ($_[0] eq 's')
                                                              ||
                                                        ($_[0] eq 'l')}}
                            },
                type    => {type      => SCALAR,
                            default   => 't',
                            callbacks => {test => sub {($_[0] eq 't')
                                                              ||
                                                        ($_[0] eq 'f')}}
                            },
                };

    my %params = validate(@_, $hash);

    $self->{bin_conf} = \%params;
    $self->_special_setup_options();
    $self->_result();
    $Trace->exit();
}
# end bin_conf

sub _special_setup_options
{
    my $self = shift;
    $self->{_special_mount_options} = "";
    $self->{need_setup} = 0;
}
# end _special_setup_options()

sub _gen_cmd
{
    $Trace->entry();
    my $self = shift;

    my @client_list = keys(%{$self->{clients}});
    
    my $hash = {clients => {type => SCALAR|ARRAYREF, default => \@client_list},};
    my %params = validate(@_, $hash);
    
    my $no_core;
    my $logdir;

    foreach my $client (@{$params{clients}})
    {
        my $cmd = {bin_base => $self->{bin_conf}->{bin_base},
                   cmd      => "./server -" . $self->{bin_conf}->{group} .
                               " -" . $self->{bin_conf}->{type} .
                               " -p " . $self->{clients}->{$client}->{nfs_info}->{junction}.
                               " -m " . $self->{clients}->{$client}->{info}->{path},
                   options  => " -o " . $self->{clients}->{$client}->{nfs_info}->{options},
                   version  => $self->{clients}->{$client}->{nfs_info}->{version},
                   data_ip  => $self->{clients}->{$client}->{info}->{data_ip},
                   dir      => $self->{clients}->{$client}->{info}->{path},};
        $self->{clients}->{$client}->{cmd} = $cmd;
    }
    $self->_result();
    $Trace->exit();
}
# end _gen_cmd()

sub _execute
{
    my $self = shift;
    return << 'SCRIPT';

    use strict;
    use Params::Validate qw(validate SCALAR);
    use File::Spec::Functions qw(catdir);

    use TracePkg;
    use Tharn qw($Parent);
    use NATE::Events qw(eventloop_once);

    my $Trace = TracePkg->new();                                            #10
    $Trace->entry();
    my $hash = {bin_base => {type => SCALAR},
                id       => {type => SCALAR},
                cmd      => {type => SCALAR},
                dir      => {type => SCALAR},
                options  => {type => SCALAR},
                version  => {type => SCALAR},
                data_ip  => {type => SCALAR},};
    my %params = validate(@_, $hash);

    my $os = $^O;              #21
    my $arch;
    my $export = "";
    my $done = 0;
    if($os =~ qr/linux/i)
    {
        chomp($arch = `uname -i`);
        if($arch =~ qr/_64/i)
        {
            $os  .= "_" . $arch;                                            #30
        }
    }
    elsif($os =~ qr/sunos/i)
    {
        chomp($arch = `uname -p`);
        if($arch ne "sparc")
        {
            $os  .= "-" . $arch;
        }
    }
    if(!(($os =~ qr/linux/i) && ($params{version} == 4)))
    {
        $params{options} .= ",vers=" . $params{version};
    }
    else
    {
        # This is NOT so good.
        $export = " export MNTFILESYSTEM='-t nfs4' ;";
        $params{options} =~ s/,vers=4//;
    }                                                                       #50

    if (!chdir($params{bin_base}))
    {
        my $msg = "Could not cd to " .$params{bin_base} . " : ".$!;
        $Trace->warn(msg => $msg);
        Tharn::fail_result($msg)->store();
        exit(0);
    }
    $Trace->comment(msg=>"PWD = " . $params{bin_base});

    if (! -x catdir($params{bin_base},"server"))
    {
        my $msg = catdir($params{bin_base},"server")." is not executable: ".$!;
        $Trace->comment(msg => $msg);
        Tharn::fail_result($msg)->store();
        exit(0);
    }
    my $cmd = "/bin/sh -c \"$export echo y | " . $params{cmd} . $params{options} .
              " " . $params{data_ip} . "\"";

    my $stopped = 0;                                                        #71
    my $finished = 0;
    my $error = 0;

    $Parent->sync_request_setup(flag     => "Process/Load/IO/cthon/Stop/".$params{id},
                                callback => sub {
                                    my ($result) = @_;
                                    print "\n\nI was told to STOP!!\n\n";
                                    NATE::STAF::default_error(%{$result})
                                        if ($result->{rc} != NATE::STAF::kOk);
                                    $stopped = 1;
                                    $Trace->comment(msg => "\nThis will exit as soon as " .
                                                           "it's complete\n");
                                });

    $Parent->sync_request(flag => "Process/Load/IO/cthon/Start/".$params{id}, timeout => 600);
    AGAIN:
    {
        open(WTR, $cmd . " 2>&1 \n |");
        $Trace->comment(msg=>"Sending " . $cmd);

        while (<WTR>)                                                       #92
        {
            my $line = $_;
            print $line;
            if($line =~ qr/all tests completed/i)
            {
                $finished = 1;
                close(WTR);
                chomp($line);
                Tharn::pass_result($line)->store();                         #101
                $finished = 1;
            }
            elsif($line =~ qr/tests failed/i)
            {
                close(WTR);
                chomp($line);
                Tharn::fail_result($line)->store();
                $error = 1;
                $finished = 1;
            }

            if($finished)                                                   #113
            {
                if(!$error && !$stopped)
                {
                    $finished = 0;
                    redo AGAIN;
                }
                else
                {
                    close(WTR);
                }
            }
            else
            {
                eventloop_once(timeout=>1);
            }
        }
    }
    if((!$error) && ($finished))
    {
        Tharn::pass_result("Everything is all good")->store();
    }
    else                                                                    #135
    {
        Tharn::fail_result("Something bad happened")->store();
    }
    $Trace->exit();
    exit(0);
# This Can NOT be indented
SCRIPT
}
# end _execute()

1;
