# $Id: //depot/prod/test/main/lib/Process/Load/IO/Crud.pm#1 $ 

## @summary Allows Process::Load to perform Load from Crud
##
## @owner GX Automation Framework
##
## @author Michael Edmonds (michael.edmonds@netapp.com)
## @author GX Automation Framework (dl-afd@netapp.com)
##
#
# Copyright (c) 2007-2008 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

=pod

=head1 Author

Michael Edmonds

=head1 Owner

GX Automation Framework (dl-afd@netapp.com)

=head1 NAME

Process::Load::IO::Crud.pm

=head1 DESCRIPTION

C<Process::Load::IO::Crud> takes care of Crud specific commands.


=head1 Public Methods

=cut

package Process::Load::IO::Crud;
use strict;
use Data::Dumper;
use Params::Validate qw(validate SCALAR BOOLEAN ARRAYREF);
use File::Spec::Functions qw(catdir);

use Tharn;
use TracePkg;
use base qw(Process::Load);

our $Trace;

BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

sub _new
{
    $Trace->entry();

    my $class = shift;

    my $self = shift;

    bless($self, $class);

    $self->_result();
    $Trace->debug(msg=>"self = " . Dumper($self));
    $Trace->exit();
    return $self;
}
# end _new()

=head2 C<bin_conf(%params)>

=over 4

=item Description

This Configures crud parameters.
This Method is Optional.
Only needs to be called if you want to modify the defaults.

=item Inputs

=item * bin => <Path to Crud Binary>

(Optional)

Default is different depending on the OS

=item * options => <String containing additional options in CLI form>

(Optional)

Default = ""

=item Output

None

=item Example

    use Process::Load;
    my $Crud_Load = Process::Load->new(type => Process::Load::CRUD(), clients => @Clients);
    my %crud_bin_params = (bin       => "/x/eng/localtest/noarch/agnostic/cluster/fs/crud/bin/linux/crud",
                            options    => "");
    $Crud_Load->bin_conf(%crud_bin_params);
    $Crud_Load->setup();
    ...

=back

=cut

sub bin_conf
{
    $Trace->entry();
    my $self = shift;
    my $hash = {bin     => {type => SCALAR, optional => 1},
                options => {type => SCALAR, default  => ""},
                timeout => {type => SCALAR, default  => ""},
                };
    my %params = validate(@_, $hash);

    $self->{bin_conf} = \%params;
    $self->_special_setup_options();
    $self->_result();
    $Trace->exit();
}
# end bin_conf

sub _gen_cmd
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => [keys(%{$self->{clients}})]},});

    my %bin_params;
    $bin_params{timeout} = $self->{bin_conf}->{timeout};
    $bin_params{bin}     = $self->{bin_conf}->{bin} if(exists($self->{bin_conf}->{bin}));
    my $opts = $self->{bin_conf}->{options};
    foreach my $client (@{$params{clients}})
    {
        my $dir = catdir($self->{clients}->{$client}->{info}->{path},
                $self->{clients}->{$client}->{random});
        my $cmd = {%bin_params, opts=> "$opts -f $dir", dir => "$dir"};
        $self->{clients}->{$client}->{cmd} = $cmd;
    }
    $self->_result();
    $Trace->exit();
}
# end _gen_cmd()

sub _special_setup_options
{
    my $self = shift;
    $self->{_special_mount_options} = "";
    $self->{need_setup} = 1;
}
# end _special_setup_options()

sub _execute
{
    my $self = shift;
    return << 'SCRIPT';

    use strict;
    use Data::Dumper;
    use Params::Validate qw(validate SCALAR);

    use TracePkg;
    use Tharn qw($Parent param subtest);
    use NATE::OS qw(is_windows);

    my $Trace = TracePkg->new();
    $Trace->entry();                                                        #10
    my $is_windows = is_windows();

    my $hash = {bin     => {type => SCALAR, optional => 1},
                id      => {type => SCALAR},
                dir     => {type => SCALAR},
                timeout => {type => SCALAR, default  => ""},
                opts    => {type => SCALAR},};
    my %params = validate(@_, $hash);

    my $stopped = 0;
    my $error = 0;

    sub handle_failure
    {
        my $msg = shift;
        $Trace->warn(msg => $msg);
        Tharn::fail_result($msg)->store();
        $error = 1;
    }
    # end handle_failure()

    if(!(exists($params{bin})))
    {
        if(!$is_windows)
        {
            $params{bin} = "/x/eng/localtest/noarch/agnostic/cluster/fs/crud/bin/linux/crud";
        }
        else
        {
            $params{bin} = "X:\\agnostic\\cluster\\fs\\crud\\bin\\win32\\crud.exe";
            $params{timeout} = "-t 120" if(!$params{timeout});
        }
    }
    if (! -x $params{bin})
    {
        handle_failure($params{bin} . " is not executable: $!");            #45
        exit(0);
    }

    mkdir($params{dir}); # write the files in a unique dir inside the mount


    $Parent->sync_request_setup(flag     => "Process/Load/IO/crud/Stop/".$params{id},
                                callback => sub {
                                    my ($result) = @_;
                                    print "\n\nI was told to STOP!!\n\n";
                                    NATE::STAF::default_error(%{$result})
                                        if ($result->{rc} != NATE::STAF::kOk);
                                    $stopped = 1;
                                    foreach my $subtest (Subtest::get_live())
                                    {
                                        $subtest->kill();
                                    }
                                    $Trace->comment(msg => "\nThis will exit as soon as " .
                                                           "it's complete\n");
                                });

    my $crud_instance;
    $Parent->sync_request(flag => "Process/Load/IO/crud/Start/".$params{id}, timeout => 600);
    my @args = split(" ",join(" ",($params{opts}, $params{dir}, $params{timeout})));
    my $full_cmd = join(" ", ($params{bin}, @args));

    AGAIN:
    {
        # Binaries always use STAF by default                               #74
        print "\n\nExecuting: $full_cmd\n\n";
        $crud_instance = subtest($params{bin}, -bg, -onexit => sub{}, '--', @args);
        $crud_instance->wait_finish();
        if(!$stopped)
        {
            $error = $crud_instance->exit_status();
            redo AGAIN if(!$error);
        }
    }

    # if we got here and we were NOT finished or we're corrupt, then fail
    if($stopped && !$error)
    {
        Tharn::pass_result("Everything is all good")->store();
    }
    else
    {
        Tharn::fail_result("Something bad happened")->store();              #92
    }

    $Trace->exit();
    exit(0);
# This Can NOT be indented
SCRIPT
}
# end _execute()

1;
