# $Id: //depot/prod/test/Rsierranevada/lib/Process/Load/IO/Custom.pm#1 $ 

## @summary Allows Process::Load to perform Load from a custom application
##
## @owner GX Automation Framework
##
## @author Michael Edmonds (michael.edmonds@netapp.com)
## @author Collin McHale (collinm@netapp.com)
## @author GX Automation Framework (dl-afd@netapp.com)
##
#
# Copyright (c) 2007-2009 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

=pod

=head1 Author

Michael Edmonds,
Collin McHale

=head1 Owner

GX Automation Framework (dl-afd@netapp.com)

=head1 NAME

Process::Load::IO::Custom.pm

=head1 DESCRIPTION

C<Process::Load::IO::Custom> takes care of custom commands.


=head1 Public Methods

=cut

package Process::Load::IO::Custom;
use strict;
use Data::Dumper;
use Params::Validate qw(validate SCALAR BOOLEAN ARRAYREF);
use File::Spec::Functions qw(catdir);

use Tharn;
use TracePkg;
use base qw(Process::Load);

our $Trace;

BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

sub _new
{
    $Trace->entry();

    my $class = shift;

    my $self = shift;

    bless($self, $class);

    $self->_result();
    $Trace->debug(msg=>"self = " . Dumper($self));
    $Trace->exit();
    return $self;
}
# end _new()

=head2 C<bin_conf(%params)>

=over 4

=item Description

This Configures custom parameters.
This Method is Optional.
Only needs to be called if you want to modify the defaults.

=item Inputs

=item * linux_bin => <Path to Linux Binary>

Location of OS specific binary

=item * win_bin => <Path to Windows Binary>

Location of OS specific binary

=item * options => <String containing additional options in CLI form>

(Optional)

Default = ""

=item Output

None

=item Example

    use Process::Load;
    my $Custom_Load = Process::Load->new(type => Process::Load::CUSTOM(), clients => @Clients);
    my %cusotm_bin_params = (linux_bin       => "/usr/local/test/bin/app1",
                            win_bin          => "X:\bin\custom\app1",
                            options    => "param1=value1 param2=value2");
    $Custom_Load->bin_conf(%custom_bin_params);
    $Custom_Load->setup();
    ...

=back

=cut

sub bin_conf
{
    $Trace->entry();
    my $self = shift;
    my $hash = {linux_bin        => {type => SCALAR,  optional => 1,},
                win_bin        => {type => SCALAR,  optional => 1,},
                options    => {type => SCALAR,  default=> "",},
                };

    my %params = validate(@_, $hash);

    $self->{bin_conf} = \%params;
    $self->_special_setup_options();
    $self->_result();
    $Trace->exit();
}
# end bin_conf

sub _gen_cmd
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => [keys(%{$self->{clients}})]},});
    
    my $no_core;
    my $logdir;
    my %bin_params;

    if(param('LOGDIR'))
    {
        $logdir = param('LOGDIR') . param('JOBID');
    }
    else
    {
        $logdir = $ENV{PWD};
    }

    if($self->{bin_conf}->{linux_bin})
    {
        $bin_params{linux_bin} = $self->{bin_conf}->{linux_bin};
    }
    if($self->{bin_conf}->{win_bin})
    {
        $bin_params{win_bin} = $self->{bin_conf}->{win_bin};
    }
    if(defined($self->{bin_conf}->{options}))
    {
        $bin_params{opts} = $self->{bin_conf}->{options};
    }

    foreach my $client (@{$params{clients}})
    {
        my $dir = catdir($self->{clients}->{$client}->{info}->{path},
                         $self->{clients}->{$client}->{random});

        my $cmd = {%bin_params, dir => $dir};
        $self->{clients}->{$client}->{cmd} = $cmd;
    }
    $self->_result();
    $Trace->exit();
}
# end _gen_cmd()

sub _special_setup_options
{
    my $self = shift;
    $self->{_special_mount_options} = ",noac,rsize=32768,wsize=32768";
    $self->{need_setup} = 1;
}
# end _special_setup_options()

sub _execute
{
    my $self = shift;
    return << 'SCRIPT';

    use strict;
    use Data::Dumper;
    use Params::Validate qw(validate SCALAR);

    use TracePkg;
    use Tharn qw($Parent param subtest);
    use NATE::OS qw(is_windows);

    my $Trace = TracePkg->new();
    $Trace->entry();                                                        #10
    my $is_windows = is_windows();

    my $hash = {linux_bin  => {type => SCALAR, optional => 1},
                win_bin  => {type => SCALAR, optional => 1},
                dir  => {type => SCALAR},
                id   => {type => SCALAR},
                opts => {type => SCALAR, default  => ""},};
    my %params = validate(@_, $hash);

    my $stopped = 0;
    my $error = 0;

    sub handle_failure
    {
        my $msg = shift;
        $Trace->warn(msg => $msg);
        Tharn::fail_result($msg)->store();
        $error = 1;
    }
    # end handle_failure()

    if(!(exists($params{bin})))                                             #31
    {
        if(!$is_windows)
        {
            $params{bin} = $params{linux_bin};
        }
        else
        {
            $params{bin} = $params{win_bin};
        }
    }
    if (! -x $params{bin})
    {
        handle_failure($params{bin} . " is not executable: $!");
        exit(0);
    }

    mkdir($params{dir}); # write the files in a unique dir inside the mount
    if($is_windows && ($params{opts} !~ /iterations/))                    #49
    {
        $params{opts} .= " -iterations 100";
    }
    my $has_iterations = ($params{opts} =~ /iterations/);

    $Parent->sync_request_setup(flag     => "Process/Load/IO/custom/Stop/".$params{id},
                                callback => sub {
                                    my ($result) = @_;
                                    print "\n\nI was told to STOP!!\n\n";
                                    NATE::STAF::default_error(%{$result})
                                        if ($result->{rc} != NATE::STAF::kOk);
                                    $stopped = 1;

                                    if(!$has_iterations)
                                    {
                                        foreach my $subtest (Subtest::get_live())
                                        {
                                            $subtest->kill();
                                        }
                                    }
                                    $Trace->comment(msg => "\nThis will exit as soon as " .
                                                           "it's complete\n");
                                });

    my $custom_instance;                                                    #74
    $Parent->sync_request(flag => "Process/Load/IO/custom/Start/".$params{id}, timeout => 600);
    # Replaces the directory key with the mount
    $params{opts} =~ s/####DIR####/$params{dir}/g;
    my @args = split(" ",join(" ", $params{opts}));

    AGAIN:
    {
        # Binaries always use STAF by default
        $custom_instance = subtest($params{bin}, -bg, -onexit => sub{}, '--', @args);
        Subtest::wait_finish(subtests=>[$custom_instance]);
        $error = $custom_instance->exit_status();
        print "\n\nERROR = $error\n\n";
        if($has_iterations && (!$error && !$stopped))
        {
            redo AGAIN;
        }
    }

    # if we got here and we were NOT finished or we're corrupt, then fail
    if($stopped && !$error)
    {                                                                       #95
        Tharn::pass_result("Everything is all good")->store();
    }
    else
    {
        Tharn::fail_result("Something bad happened")->store();
    }

    $Trace->exit();
    exit(0);
# This Can NOT be indented
SCRIPT
}
# end _execute()

1;
