# $Id: //depot/prod/test/Rsierranevada/lib/Process/Load/IO/Dt.pm#2 $ 

## @summary Allows Process::Load to perform Load from Dt 
##
## @owner GX Automation Framework
##
## @author dl-volops-qa 
##
#
# Copyright (c) 2010 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

=pod

=head1 Author

dl-volops-qa

=head1 Owner

VolOps QA (dl-volops-qa@netapp.com)

=head1 NAME

Process::Load::IO::Dt.pm

=head1 DESCRIPTION

C<Process::Load::IO::Dt> takes care of dt specific commands.


=head1 Public Methods

=cut

package Process::Load::IO::Dt;
use strict;
use Data::Dumper;
use Params::Validate qw(validate SCALAR BOOLEAN ARRAYREF);
use File::Spec::Functions qw(catdir);

use Tharn;
use TracePkg;
use base qw(Process::Load);

our $Trace;

BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

sub _new
{
    $Trace->entry();

    my $class = shift;

    my $self = shift;

    bless($self, $class);

    $self->_result();
    $Trace->debug(msg=>"self = " . Dumper($self));
    $Trace->exit();
    return $self;
}
# end _new()

=head2 C<bin_conf(%params)>

=over 4

=item Description

This configures dt parameters.
This Method is Optional.
Only needs to be called if you want to modify the defaults.

=item Inputs

=item * bin => <Path to Dt Binary>

(Optional)

Default is different depending on the OS

=item * version => <stable or latest> 

(Optional)

Default = latest

=item * options => <String containing additional options in CLI form>

(Optional)

Default = ""

=item Output

None

=item Example

    use Process::Load;
    my $Dt_Load = Process::Load->new(type => Process::Load::DT(), clients => @Clients);
    my %dt_bin_params = (bin       => "/x/eng/localtest/arch/i386-redhat-linux8/bin/dt.stable",
                         options    => "");
    $Dt_Load->bin_conf(%dt_bin_params);
    $Dt_Load->setup();
    ...

=back

=cut

sub bin_conf
{
    $Trace->entry();
    my $self = shift;
    my $hash = {bin        => {type => SCALAR,  optional => 1,},
                version    => {type => SCALAR,  optional => 1,
                               default => 'latest',
                               callbacks => {testver => sub {
                                ($_[0] =~ qr/stable/i)
                                       ||
                                ($_[0] =~ qr/latest/i)}},
                              },
                options    => {type => SCALAR,  default=> "",
                              },
                };

    my %params = validate(@_, $hash);

    $self->{bin_conf} = \%params;
    $self->_special_setup_options();
    $self->_result();
    $Trace->exit();
}
# end bin_conf

sub _gen_cmd
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => [keys(%{$self->{clients}})]},});
    
    my $logdir;
    my %bin_params;

    if($ENV{LOGDIR})
    {
        $logdir = $ENV{LOGDIR} . $ENV{JOBID};
    }
    else
    {
        $logdir = $ENV{PWD};
    }

    if($self->{bin_conf}->{bin})
    {
        $bin_params{bin} = $self->{bin_conf}->{bin};
    }
    if(defined($self->{bin_conf}->{options}))
    {
        $bin_params{opts} = $self->{bin_conf}->{options};
    }

    if(defined($self->{bin_conf}->{version}))
    {
        $bin_params{version} = $self->{bin_conf}->{version};
    }

    foreach my $client (@{$params{clients}})
    {
        $Log->debug("Client data: " . Dumper $self->{clients}->{$client});
        my $dir = catdir($self->{clients}->{$client}->{info}->{path},
                         $self->{clients}->{$client}->{random});

        my $cmd = {%bin_params, dir => $dir};
        $self->{clients}->{$client}->{cmd} = $cmd;

        $Log->debug("Client command: " . Dumper $cmd);
    }
    $self->_result();
    $Trace->exit();
}
# end _gen_cmd()

sub _special_setup_options
{
    my $self = shift;
#    $self->{_special_mount_options} = ",noac,rsize=32768,wsize=32768";
    $self->{need_setup} = 1;
}
# end _special_setup_options()

sub _execute
{
    my $self = shift;
    return << 'SCRIPT';

    use strict;
    use Data::Dumper;
    use Params::Validate qw(validate SCALAR);

    use Tharn;
    use TracePkg;
    use Tharn qw($Parent param subtest);
    use NATE::OS qw(is_windows);

    my $Trace = TracePkg->new();
    $Trace->entry();                                                        #10
    my $is_windows = is_windows();

    my $hash = {bin     => {type => SCALAR, optional => 1},
                dir     => {type => SCALAR},
                id      => {type => SCALAR},
                version => {type => SCALAR, default => "stable"},
                opts    => {type => SCALAR, default  => ""},};
    my %params = validate(@_, $hash);

    my $stopped = 0;
    my $error   = 0;

    sub handle_failure
    {
        my $msg = shift;
        $Trace->warn(msg => $msg);
        Tharn::fail_result($msg)->store();
        $error = 1;
    }
    # end handle_failure()

    if(!(exists($params{bin})))                                             #31
    {
        if(!$is_windows)
        {
            if ($^O =~ /linux/i) 
            {
                $params{bin} = "/x/eng/localtest/arch/i386-redhat-linux8/bin/dt." . $params{version};
                $params{bin} = "/x/eng/local/test/arch/i386-redhat-linux8/bin/dt" if(!(-e $params{bin}));
            } 
            elsif ($^O =~ /solaris|sunos/i)
            {
                $params{bin} = "/x/eng/localtest/arch/sparc-sun-solaris8/bin/dt." .  $params{version};
                $params{bin} = "/x/eng/localtest/arch/sparc-sun-solaris8/bin/dt" if (!(-e $params{bin}));
            } 
            else 
            {
                $params{bin} = "/usr/local/test/bin/dt";
            }
        }
        else #windows
        {
             $params{bin} = "X:\\\\SAN\\\\bin\\\\common\\\\dt." .  $params{version} .  "\\\\dt.exe";
             $params{bin} = "X:\\\\SAN\\\\bin\\\\common\\\\dt.exe" if (!(-e $params{bin}));
        }
    }
    if (! -x $params{bin})
    {
        handle_failure($params{bin} . " is not executable: $!");
        exit(0);
    }

    # Set the trigger based on OS if not defined and noprog is specified
    if ($params{opts} =~ /enable=noprog/) 
    {
        if ($params{opts} !~ /trigger/) 
        {
            if ($is_windows) 
            {
                $params{opts} .= " trigger=cmd:'//jade.rtp.netapp.com/users/rtmiller/Tools/dt.d-WIP/Scripts/dt_noprog_script.bat'";
#                trigger=cmd:\'\\jade.rtp.netapp.com\users\rtmiller\Tools\dt.d-WIP\Scripts\dt_noprog_script.bat\'';
            } 
            else
            {
                $params{opts} .= ' trigger=cmd:"/u/rtmiller/Tools/dt.d-WIP/Scripts/dt_noprog_script.ksh"';
            }
        }
    }

    mkdir($params{dir}); # write the files in a unique dir inside the mount

    if ($params{opts} !~ ' of=') {
        $params{opts} .= ' of=' . $params{dir} . '/dt.out';
    }

    # Selected a default runtime of 120 seconds.  This can be overridden
    # by passing in a value for the runtime option
    # cluster/fs/dt/dt.thpl uses
    if ($params{opts} !~ ' runtime=') {
        $params{opts} .= ' runtime=120';
    }

    $Parent->sync_request_setup(flag     => "Process/Load/IO/dt/Stop/".$params{id},
                                callback => sub {
                                    my ($result) = @_;
                                    print "\n\nI was told to STOP!!\n\n";
                                    NATE::STAF::default_error(%{$result})
                                        if ($result->{rc} != NATE::STAF::kOk);
                                    $stopped = 1;

                                    $Trace->comment(msg => "\nThis will exit as soon as " .
                                                           "it's complete\n");
                                });

    my $dt_instance;
    $Parent->sync_request(flag => "Process/Load/IO/dt/Start/".$params{id}, timeout => 600);

    my @args = split(" ",$params{opts});
$Trace->comment(msg => "Args: " . Dumper \@args);


    AGAIN:
    {

        $Trace->debug(msg => "Calling dt with the following arguments: " . 
                    Dumper \@args);        

        # Binaries always use STAF by default
        $dt_instance = subtest($params{bin}, -bg, -onexit => sub{}, '--', @args);
        Subtest::wait_finish(subtests=>[$dt_instance]);

        $error = $dt_instance->exit_status();
        print "\n\nERROR = $error\n\n";
        if (!$stopped) {
            redo AGAIN if (!$error);
        }    
    }

    # if we got here and we were NOT finished or we're corrupt, then fail
    if($stopped && !$error)
    {                                                                       #90
        Tharn::pass_result("Everything is all good")->store();
    }
    else
    {
        Tharn::fail_result("Something bad happened")->store();
    }

    $Trace->exit();
    exit(0);
# This Can NOT be indented
SCRIPT
}
# end _execute()

1;
