# $Id: $ 

## @summary Allows Process::Load to perform Load from Hammer with locking
##
## @owner GX Automation Framework
##
## @author Michael Edmonds (michael.edmonds@netapp.com)
## @author Collin McHale (collin.mchale@netapp.com)
## @author GX Automation Framework (dl-afd@netapp.com)
##
#
# Copyright (c) 2007-2008 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

=pod

=head1 Author

Michael Edmonds
Collin McHale

=head1 Owner

GX Automation Framework (dl-afd@netapp.com)

=head1 NAME

Process::Load::IO::HammerLocking.pm

=head1 DESCRIPTION

C<Process::Load::IO::Hammer> takes care of hammer specific commands with locking.


=head1 Public Methods

=cut

package Process::Load::IO::HammerLocking;
use strict;
use Data::Dumper;
use Params::Validate qw(validate SCALAR BOOLEAN ARRAYREF);
use File::Spec::Functions qw(catdir);

use Tharn;
use TracePkg;
use base qw(Process::Load);

our $Trace;

BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

sub _new
{
    $Trace->entry();

    my $class = shift;

    my $self = shift;

    bless($self, $class);

    $self->_result();
    $Trace->debug(msg=>"self = " . Dumper($self));
    $Trace->exit();
    return $self;
}
# end _new()

=head2 C<bin_conf(%params)>

=over 4

=item Description

This Configures hammer parameters.
This Method is Optional.
Only needs to be called if you want to modify the defaults.

=item Inputs

=item * bin => <Path to Hammer Binary>

(Optional)

Default is different depending on the OS

=item * filer_core => <0 | 1>

(Optional)

Default = 0

I Don't think Hammer can make a GX node core, but if it gets fixed, this option
is present

=item * options => <String containing additional options in CLI form>

(Optional)

Default = "-lockfiles"
The method should blow up if you pass "interactive" or "[no]filercore"

=item Output

None

=item Example

    use Process::Load;
    my $Hmr_Load = Process::Load->new(type => Process::Load::HAMMERLOCKING(), clients => @Clients);
    my %hammer_bin_params = (bin       => "/usr/software/test/bin/hammer.current",
                            filer_core => 0,
                            options    => "-lockfiles");
    $Hmr_Load->bin_conf(%hammer_bin_params);
    $Hmr_Load->setup();
    ...

=back

=cut

sub bin_conf
{
    $Trace->entry();
    my $self = shift;
    my $hash = {bin        => {type => SCALAR,  optional => 1,},
                filer_core => {type => BOOLEAN, default => 0,},
                options    => {type => SCALAR,  default=> "-lockfiles",
                               callbacks => {testoptions => sub {
                                ($_[0] !~ qr/interactive/i)
                                       &&
                                ($_[0] !~ qr/filercore/i)}},
                               },
                };

    my %params = validate(@_, $hash);

    $self->{bin_conf} = \%params;
    $self->_special_setup_options();
    $self->_result();
    $Trace->exit();
}
# end bin_conf

sub _gen_cmd
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => [keys(%{$self->{clients}})]},});
    
    my $no_core;
    my $logdir;
    my %bin_params;

    if(param('LOGDIR'))
    {
        $logdir = param('LOGDIR') . param('JOBID');
    }
    else
    {
        $logdir = $ENV{PWD};
    }

    if($self->{bin_conf}->{bin})
    {
        $bin_params{bin} = $self->{bin_conf}->{bin};
    }
    if(defined($self->{bin_conf}->{options}))
    {
        $bin_params{opts} = $self->{bin_conf}->{options};
    }

    foreach my $client (@{$params{clients}})
    {
        if($self->{bin_conf}->{filer_core})
        {
            $bin_params{core} = " -filercore " .
            $self->{clients}->{$client}->{info}->{node};
        }
        my $dir = catdir($self->{clients}->{$client}->{info}->{path},
                         $self->{clients}->{$client}->{random});

        my $cmd = {%bin_params, dir => $dir};
        $self->{clients}->{$client}->{cmd} = $cmd;
    }
    $self->_result();
    $Trace->exit();
}
# end _gen_cmd()

sub _special_setup_options
{
    my $self = shift;
    $self->{_special_mount_options} = ",noac,rsize=32768,wsize=32768";
    $self->{need_setup} = 1;
}
# end _special_setup_options()

sub _execute
{
    my $self = shift;
    return << 'SCRIPT';

    use strict;
    use Data::Dumper;
    use Params::Validate qw(validate SCALAR);

    use TracePkg;
    use Tharn qw($Parent param subtest);
    use NATE::OS qw(is_windows);

    my $Trace = TracePkg->new();
    $Trace->entry();                                                        #10
    my $is_windows = is_windows();

    my $hash = {bin  => {type => SCALAR, optional => 1},
                core => {type => SCALAR, default  => '-nofilercore'},
                dir  => {type => SCALAR},
                id   => {type => SCALAR},
                opts => {type => SCALAR, default  => "-lockfiles"},};
    my %params = validate(@_, $hash);

    my $stopped = 0;
    my $error = 0;

    sub handle_failure
    {
        my $msg = shift;
        $Trace->warn(msg => $msg);
        Tharn::fail_result($msg)->store();
        $error = 1;
    }
    # end handle_failure()

    if(!(exists($params{bin})))                                             #32
    {
        if(!$is_windows)
        {
            $params{bin} = "/usr/software/test/bin/hammer.current";
            $params{bin} = "/usr/local/test/bin/hammer.current" if(!(-e $params{bin}));
        }
        else
        {
            $params{bin} = "X:\\bin\\hammer.current.exe";
        }
    }
    if (! -x $params{bin})
    {
        handle_failure($params{bin} . " is not executable: $!");
        exit(0);
    }

    mkdir($params{dir}); # write the files in a unique dir inside the mount
    if($is_windows && ($params{opts} !~ /iterations/) && ($params{opts} !~ /runtime/))  #51
    {
        $params{opts} .= " -iterations 100";
    }
    my $has_iterations = ($params{opts} =~ /iterations/);

    $Parent->sync_request_setup(flag     => "Process/Load/IO/hammerlocking/Stop/".$params{id},
                                callback => sub {
                                    my ($result) = @_;
                                    print "\n\nI was told to STOP!!\n\n";
                                    NATE::STAF::default_error(%{$result})
                                        if ($result->{rc} != NATE::STAF::kOk);
                                    $stopped = 1;

                                    if(!$has_iterations)
                                    {
                                        foreach my $subtest (Subtest::get_live())
                                        {
                                            $subtest->kill();
                                        }
                                    }
                                    $Trace->comment(msg => "\nThis will exit as soon as " .
                                                           "it's complete\n");
                                });

    my $hammer_instance;                                                    #76
    $Parent->sync_request(flag => "Process/Load/IO/hammerlocking/Start/".$params{id}, timeout => 600);
    my @args = split(" ",join(" ",$params{core}, $params{dir}, $params{opts}));

    AGAIN:
    {
        # Binaries always use STAF by default
        $hammer_instance = subtest($params{bin}, -bg, -onexit => sub{}, '--', @args);
        Subtest::wait_finish(subtests=>[$hammer_instance]);
        $error = $hammer_instance->exit_status();
        print "\n\nERROR = $error\n\n";
        if($has_iterations && (!$error && !$stopped))
        {
            redo AGAIN;
        }
    }

    # if we got here and we were NOT finished or we're corrupt, then fail
    if(!$error)
    {                                                                       #95
        Tharn::pass_result("Everything is all good")->store();
    }
    else
    {
        Tharn::fail_result("Something bad happened")->store();
    }

    $Trace->exit();
    exit(0);
# This Can NOT be indented
SCRIPT
}
# end _execute()

1;
