# $Id: //depot/prod/test/main/lib/Process/Load/IO/Random.pm#1 $ 

## @summary Allows Process::Load to perform Load from Random
##
## @owner GX Automation Framework
##
## @author Michael Edmonds (michael.edmonds@netapp.com)
## @author GX Automation Framework (dl-afd@netapp.com)
##
#
# Copyright (c) 2007-2008 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

=pod

=head1 Author

Michael Edmonds

=head1 Owner

GX Automation Framework (dl-afd@netapp.com)

=head1 NAME

Process::Load::IO::Random.pm

=head1 DESCRIPTION

C<Process::Load::IO::Random> takes care of Random specific commands.

=head1 Public Methods

=cut

package Process::Load::IO::Random;
use strict;
use Data::Dumper;
use Params::Validate qw(validate SCALAR BOOLEAN ARRAYREF);

use Tharn;
use TracePkg;
use base qw(Process::Load);

our $Trace;

BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

sub _new
{
    $Trace->entry();

    my $class = shift;

    my $self = shift;

    bless($self, $class);

    $self->_result();
    $self->{deny_hosttype} = ["win"];
    $Trace->debug(msg=>"self = " . Dumper($self));
    $Trace->exit();
    return $self;
}
# end _new()

=head2 B<bin_conf(%params)>

=over 4

=item Description

This Configures the Random parameters.
This Method is Optional.
Only needs to be called if you want to modify the defaults.

=item Inputs

=item * bin => <Path to a specific Random executable>

(Optional)

Default is different depending on the OS

=item * seed

(Optional)

=item * maxoffset => <Positive Number>

(Optional)

=item * maxfilelen => <Positive Number>

(Optional)

=item * uid

(Optional)

=item * gid

(Optional)

=item * random => <0 | 1>

(Optional)

default = 0

if random => 1, it sets the '-R' flag for the random binary

=item * data => <0 - 255>

(Optional)

=item * maxdifflength => <Positive Number>

(Optional)

=item Output

None

=item Example

    use Process::Load;
    my $Random_Load = Process::Load->new(type => Process::Load::RANDOM(), clients => @Clients);
    my %random_bin = (random => 1, seed => time());
    $Random_Load->bin_conf(%random_bin);
    ...

=back

=cut

sub bin_conf
{
    $Trace->entry();
    my $self = shift;
    my $hash = {bin        => {type => SCALAR, optional => 1,},
                maxoffset  => {type => SCALAR, optional => 1,
                               callbacks => {testcore => sub {(int($_[0]) > 0)}},
                               },
                maxfilelen => {type => SCALAR, optional => 1,
                                callbacks => {testcore => sub {(int($_[0]) > 0)}},
                               },
                maxdifflen => {type => SCALAR, optional => 1,
                                callbacks => {testcore => sub {(int($_[0]) > 0)}},
                               },
                seed       => {type => SCALAR, optional => 1},
                uid        => {type => SCALAR, optional => 1},
                gid        => {type => SCALAR, optional => 1},
                data       => {type => SCALAR, optional => 1,
                               callbacks => {testcore => sub {(int($_[0]) > -1)
                                                                      &&
                                                              (int($_[0]) < 256)}},
                               },
                random     => {type => BOOLEAN, optional => 1,},
    };

    my %params = validate(@_, $hash);

    $self->{bin_conf} = \%params;
    $self->_special_setup_options();
    $self->_result();
    $Trace->exit();
}
# end bin_conf

sub _special_setup_options
{
    my $self = shift;
    $self->{_special_mount_options} = "";
    $self->{need_setup} = 1;
}
# end _special_setup_options()

#usage: random [-s seed] [-o maxoffset] [-l maxfilelen] [-v] [-u uid] [-g gid] [-r] [-R] [-e max-diff-length] [-b data] [-t time-to-run] scratch_dir

sub _gen_cmd
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type => SCALAR | ARRAYREF,
                               default => [keys(%{$self->{clients}})]},});

    my $cmd = " -v ";
    
    $params{bin} = $self->{bin_conf}->{bin}          if(exists($self->{bin_conf}->{bin}));
    $cmd .= "-o " . $self->{bin_conf}->{maxoffset}  if(exists($self->{bin_conf}->{maxoffset}));
    $cmd .= "-l " . $self->{bin_conf}->{maxfilelen} if(exists($self->{bin_conf}->{maxfilelen}));
    $cmd .= "-e " . $self->{bin_conf}->{maxdifflen} if(exists($self->{bin_conf}->{maxdifflen}));
    $cmd .= "-s " . $self->{bin_conf}->{seed}       if(exists($self->{bin_conf}->{seed}));
    $cmd .= "-u " . $self->{bin_conf}->{uid}        if(exists($self->{bin_conf}->{uid}));
    $cmd .= "-g " . $self->{bin_conf}->{gid}        if(exists($self->{bin_conf}->{gid}));
    $cmd .= "-b " . $self->{bin_conf}->{data}       if(exists($self->{bin_conf}->{data}));
    $cmd .= "-R "                                   if($self->{bin_conf}->{random});

    foreach my $client (@{$params{clients}})
    {
        my $dir = $self->{clients}->{$client}->{info}->{path} ."/".
                        $self->{clients}->{$client}->{random};
        my $command = {cmd => $cmd, dir => $dir};
        $self->{clients}->{$client}->{cmd} = $command;
    }
    $self->_result();
    $Trace->exit();
}
# end _gen_cmd()

sub _execute
{
    my $self = shift;
    # Dude, a nested script.
    return << 'SCRIPT';

    use strict;
    use Data::Dumper;
    use Params::Validate qw(validate SCALAR);

    use Tharn qw(fail_result $Parent);
    use TracePkg;

    my $Trace = TracePkg->new();
    $Trace->entry();
    my $hash = {bin => {type => SCALAR, optional => 1},                     #10
                cmd => {type => SCALAR},
                dir => {type => SCALAR},
                id =>  {type => SCALAR},};
    my %params = validate(@_, $hash);
    my $dir = $params{dir};

    if(!(mkdir($params{dir})))
    {
        my $msg = "Could Not Create " . $params{dir};
        $Trace->warn(msg => $msg);
        Tharn::fail_result($msg)->store();
        exit(0);
    }

    if(!(exists($params{bin})))
    {
        $params{bin} = "/usr/software/test/bin/random";
        $params{bin} = "/usr/local/bin/random" if(!(-e $params{bin}));
    }

    if (! -x $params{bin})                                                  #30
    {
        my $msg = $params{bin} . " is not executable: ".$!;
        $Trace->warn(msg => $msg);
        Tharn::fail_result($msg)->store();
        exit(0);
    }

    my $cmd = $params{cmd} . " " . $params{dir};

    my $stopped = 0;

    $Parent->sync_request_setup(flag     => "Process/Load/IO/random/Stop/".$params{id},
                                callback => sub {
                                    my ($result) = @_;
                                    print "\n\nI was told to STOP!!\n\n";
                                    NATE::STAF::default_error(%{$result})
                                        if ($result->{rc} != NATE::STAF::kOk);
                                    $stopped = 1;
                                    foreach my $subtest (Subtest::get_live())
                                    {                                       #50
                                        $subtest->kill();
                                    }
                                    $Trace->comment(msg => "\nThis will exit as soon as " .
                                                           "it's complete\n");
                                });

    $Parent->sync_request(flag => "Process/Load/IO/random/Start/".$params{id}, timeout => 600);
    my $random_instance = Tharn::subtest($params{bin},
                                         -bg,
                                         -onexit => sub{},
                                         "--",split(" ",$cmd));

    $random_instance->wait_finish();
    # if we're here and $stopped is Not 1, then something bad happened
    #   and we should be Done
    if(!$stopped)
    {
        fail_result("Random Exited without asking it to." .
                    "\nExitCode = ".$random_instance->exit_code())->store();
    }                                                                       #70
    else
    {
        Tharn::pass_result("Everything is all good")->store();
    }

    $Trace->exit();
    exit(0);
# This Can NOT be indented
SCRIPT
}
# end _execute()

1;
