# $Id: //depot/prod/test/main/lib/Process/Load/IO/SIO.pm#1 $ 

## @summary Allows Process::Load to perform Load from SIO
##
## @owner GX Automation Framework
##
## @author Michael Edmonds (michael.edmonds@netapp.com)
## @author GX Automation Framework (dl-afd@netapp.com)
##
#
# Copyright (c) 2007-2008 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#
##
## @status public
## @pod here

=pod

=head1 Author

Michael Edmonds

=head1 Owner

GX Automation Framework (dl-afd@netapp.com)

=head1 NAME

Process::Load::IO::SIO.pm

=head1 DESCRIPTION

C<Process::Load::IO::SIO> takes care of SIO specific commands.


=head1 Public Methods

=cut

package Process::Load::IO::SIO;
use strict;
use Data::Dumper;
use Params::Validate qw(validate SCALAR ARRAYREF);
use File::Spec::Functions qw(catdir);

use Tharn;
use TracePkg;
use base qw(Process::Load);

our $Trace;

BEGIN
{
    $Trace = TracePkg->new();
}
# end BEGIN

sub _new
{
    $Trace->entry();

    my $class = shift;

    my $self = shift;

    bless($self, $class);

    $self->_result();
    $Trace->debug(msg=>"self = " . Dumper($self));
    $Trace->exit();
    return $self;
}
# end _new()

=head2 C<bin_conf(%params)>

=over 4

=item Description

This Configures SIO parameters.
This Method is Optional.
Only needs to be called if you want to modify the defaults.

=item Inputs

=item * bin => <Path to SIO Binary>

(Optional)

Default is different depending on the OS

=item * reads => <0 - 100 | 'random'>

(Optional)

Default = random

percentage of accesses that are reads

=item * rand => <0 - 100>

(Optional)

Default = 100

percentage of acceses that are random

=item * block_size => <512 - 256k | random>

(Optional)

Default = random

size of I/O's. (e.g. 1024, 2k, 4k)

=item * start => <positive integer>

(Optional)

Default = 0

lower bound for access location in each file

=item * file_size => <positive integer>

(Optional)

Default = 1m

total bytes accessed in each file (e.g. 1024, 100m, 2g, 1000k)

=item * threads => <positive integer>

(Optional)

Default = 2

Concurrent I/O generators

=item * adv_args => <advanced command line arguments>

(Optional)

Default = ""

=item * file_qty => <integer greater than 0>

(Optional)

Default = 1

quantity of files that this lib should generate for the test

=item * file_names => <$file_name | \@file_names>

(Optional)

this is if you want to hit the same files from different locations

=item Output

None

=item Example

    use Process::Load;
    my $SIO_Load = Process::Load->new(type => Process::Load::SIO(), clients => @Clients);
    my %sio_bin_params = (file_qty  => 100,
                          file_size => "1g",
                          adv_args => "-break_on_dc -fixedfill -256");
    $SIO_Load->bin_conf(%sio_bin_params);
    $SIO_Load->mount();
    ...

=back

=cut

sub bin_conf
{
    $Trace->entry();
    my $self = shift;
    my $hash = {bin        => {type => SCALAR, optional => 1,},
                reads      => {type => SCALAR, default => "random",
                               callbacks => {test => sub {((int($_[0]) > -1) && (int($_[0]) < 101)) || ($_[0] eq "random")}}},
                rand       => {type => SCALAR, default => 100,
                               callbacks => {test => sub {(int($_[0]) > -1) && (int($_[0]) < 101)}}},
                block_size => {type => SCALAR, default => "random"},
                start      => {type => SCALAR, default => 0,
                               callbacks => {test => sub {int($_[0]) > -1}}},
                file_size  => {type => SCALAR, default => "1m"},
                threads    => {type => SCALAR, default => 2,
                               callbacks => {test => sub {int($_[0]) > 0}}},
                file_qty   => {type => SCALAR, default => 1,
                               callbacks => {test => sub {int($_[0]) > 0}}},
                adv_args   => {type => SCALAR, default => ""},
                file_names => {type => SCALAR|ARRAYREF, optional => 1},
                };

    my %params = validate(@_, $hash);

    if(ref($params{file_names}) ne "ARRAY")
    {
        $params{file_names} = [$params{file_names}] if(exists($params{file_names}));
    }

    $self->{bin_conf} = \%params;
    $self->_special_setup_options();
    $self->_result();
    $Trace->exit();
}
# end bin_conf


sub _gen_cmd
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => [keys(%{$self->{clients}})]},
                               });

    my %bin_params;
    my $cmd = join(" ",($self->{bin_conf}->{reads},
                        $self->{bin_conf}->{rand},
                        $self->{bin_conf}->{block_size},
                        $self->{bin_conf}->{start},
                        $self->{bin_conf}->{file_size},
                        "0",
                        $self->{bin_conf}->{threads}));

    if($self->{bin_conf}->{bin})
    {
        $bin_params{bin} = $self->{bin_conf}->{bin};
    }

    foreach my $client (@{$params{clients}})
    {
        my @files;
        my $dir = catdir($self->{clients}->{$client}->{info}->{path},$self->{clients}->{$client}->{random});
        if(exists($self->{bin_conf}->{file_names}))
        {
            foreach my $file ($self->{bin_conf}->{file_names})
            {
                push(@files, catdir($dir,$file));
            }
        }
        else
        {
            for(my $i = 0 ; $i < $self->{bin_conf}->{file_qty} ; $i++)
            {
                push(@files, catdir($dir,$client . "_" . $i . ".file"));
            }
        }
        $bin_params{dir} = $dir;
        $bin_params{cmd} = $cmd . " " . join(" ", @files) . " " .
                            $self->{bin_conf}->{adv_args};
        $bin_params{key} = $client . $self->{clients}->{$client}->{random};

        $self->{clients}->{$client}->{cmd} = {%bin_params};
    }
    $self->_result();
    $Trace->exit();
}
# end _gen_cmd()

sub _special_setup_options
{
    my $self = shift;
    $self->{_special_mount_options} = ",noac";
    $self->{need_setup} = 1;
}
# end _special_setup_options()

sub __pre_start
{
    $Trace->entry();
    my $self = shift;

    my %params = validate(@_, {clients => {type    => SCALAR|ARRAYREF,
                                           default => [keys(%{$self->{clients}})]},
                               });

    my %bin_params;
    my $blocksz =  $self->{bin_conf}->{block_size};

    $bin_params{cmd} = " 0 0 " . $blocksz . " 0 " . $self->{bin_conf}->{file_size} .
                        " 0 " . $self->{bin_conf}->{threads};
    $bin_params{adv_args} = " -fillonce -debug " . $self->{bin_conf}->{adv_args};

    if($self->{bin_conf}->{bin})
    {
        $bin_params{bin} = $self->{bin_conf}->{bin};
    }

    my @subtest_objs;

    foreach my $client (@{$params{clients}})
    {
        my @files;
        my $dir = catdir($self->{clients}->{$client}->{info}->{path}, $self->{clients}->{$client}->{random});
        if(exists($self->{bin_conf}->{file_names}))
        {
            foreach my $file ($self->{bin_conf}->{file_names})
            {
                push(@files, catdir($dir, $file));
            }
        }
        else
        {
            for(my $i = 0 ; $i < $self->{bin_conf}->{file_qty} ; $i++)
            {
                push(@files, catdir($dir, $client . "_" . $i . ".file"));
            }
        }
        my $cmd = {%bin_params, files => \@files, dir => $dir};

        if(!(defined($self->{clients}->{$client}->{host_obj})))
        {
            my $host_obj = Tharn::host($client);
            $self->{clients}->{$client}->{host_obj} = $host_obj;
        }

        my $runid = $self->{runid_appender} . "mkfiles_" . $client . "_" .
                            $self->{clients}->{$client}->{random};

        my $cmd_obj = Tharn::subtest("{".$self->_mkfiles()."}", -bg, 
                                     -client  => $self->{clients}->{$client}->{host_obj},
                                     -onexit  => sub {},
                                     -runid   => $runid,
                                     "--"     => %{$cmd}
                                    );

        $self->{clients}->{$client}->{mkfile_thread} = $cmd_obj;
        $Trace->debug(msg=>"subtested _mkdir for " . $runid);
        push(@subtest_objs, $cmd_obj);
    }

    Subtest::wait_finish(subtests=>\@subtest_objs);
    $Trace->debug(msg=>"All mkfiles finished");

    # Wait for the subtests to return for EACH Client
    foreach my $client (keys(%{$self->{clients}}))
    {
        my $worst_result = NATE::Result::worst($self->{clients}->{$client}->{mkfile_thread}->results());
        if($worst_result->type() ne "PASS")
        {
            $Trace->warn(msg => "Something Bad happened while making SIO " .
                         "files on " . $client);
            $self->_result(rc => $self->FATAL(),
                           msg => join(",", $worst_result->msgs()));
        }
        delete($self->{clients}->{$client}->{mkfile_thread});
    }
    $self->_result();
    $Trace->exit();
}
# end __pre_start

sub _mkfiles
{
    my $self = shift;
    return << 'SCRIPT';

    use strict;
    use Data::Dumper;
    use Params::Validate qw(validate SCALAR ARRAYREF);

    use TracePkg;
    use Tharn;
    use NATE::OS qw(is_windows);

    my $Trace = TracePkg->new();
    $Trace->entry();                                                       #10
    my $hash = {bin      => {type => SCALAR, optional => 1},
                dir      => {type => SCALAR},
                cmd      => {type => SCALAR},
                files    => {type => ARRAYREF},
                adv_args => {type => SCALAR}};
    my %params = validate(@_, $hash);

    my $is_windows = is_windows();

    sub handle_failure                                                      #20
    {
        my $msg = shift;
        $Trace->warn(msg => $msg);
        Tharn::fail_result($msg)->store();
    }
    # end handle_failure()

    if(!(mkdir($params{dir})))
    {
        handle_failure("Could Not Create " . $params{dir});                 #30
        exit(0);
    }

    if(!(exists($params{bin})))
    {
        $params{bin} = (!$is_windows) ? "/usr/software/test/bin/sio" :
                                         "X:\\bin\\sio.exe";
    }

    if (! -x $params{bin})
    {
        my $msg = $params{bin} . " is not executable: $!";
        handle_failure($msg);
        exit(0);
    }
    foreach my $file (@{$params{files}})
    {
        if(open(SIOFILE, ">" . $file))
        {
            $Trace->debug(msg => "Created " . $file);                       #50
            close(SIOFILE);
        }
        else
        {
            handle_failure("Cannot open $file: $!");
            exit(0);
        }
    }

    my $cmd = join(" ",($params{bin},$params{cmd},@{$params{files}},$params{adv_args}));
    $Trace->comment(msg=>"executing " . $cmd);
    if(system($cmd . " 2>&1") != 0)                                         #70
    {
        handle_failure("Problem occured w/ -fillonce");
        exit(0);
    }
    else
    {
        Tharn::pass_result("Everything is all good")->store();
    }

    $Trace->exit();                                                         #80
    exit(0);
# This Can NOT be indented
SCRIPT
}
# end _mkfiles()

sub _execute
{
    my $self = shift;
    return << 'SCRIPT';

    use strict;
    use Data::Dumper;
    use Params::Validate qw(validate SCALAR);
    use File::Spec::Functions qw(catdir);

    use TracePkg;
    use Tharn qw(param $Parent);
    use NATE::OS qw(is_windows);

    my $Trace = TracePkg->new();                                            #10
    $Trace->entry();
    my $hash = {bin => {type => SCALAR, optional => 1},
                id  => {type => SCALAR},
                dir => {type => SCALAR},
                cmd => {type => SCALAR},
                key => {type => SCALAR}};
    my %params = validate(@_, $hash);

    my $is_windows = is_windows();

    sub handle_failure
    {
        my $msg = shift;
        $Trace->warn(msg => $msg);
        Tharn::fail_result($msg)->store();
    }
    # end handle_failure()

    mkdir($params{dir}); # don't catch error in case dir exists already;

    my $key_prepender = param('LOGDIR')||"";                                #30
    $params{key} = catdir($key_prepender . "/". $params{key});

    if(!(exists($params{bin})))
    {
        $params{bin} = (!$is_windows)? "/usr/software/test/bin/sio" :
                                        "X:/bin/sio.exe";
    }

    if (! -x $params{bin})
    {
        my $msg = $params{bin} . " is not executable: $!";
        handle_failure($msg);
        exit(0);
    }

    my $stopped = 0;
    my $cmd = join(" ",($params{cmd},"-niceoutput","-debug","-stop","-".$params{key}));

    $Parent->sync_request(flag => "Process/Load/IO/sio/Start/".$params{id}, timeout => 600);
    my $sio_instance = subtest($params{bin},                                #50
                               -bg,
                               -onexit => sub{},
                               "--", split(" ",$cmd));

    $Parent->sync_request_setup(flag     => "Process/Load/IO/sio/Stop/".$params{id},
                                callback => sub {
                                    my ($result) = @_;
                                    print "\n\nI was told to STOP!!\n\n";
                                    NATE::STAF::default_error(%{$result})
                                        if ($result->{rc} != NATE::STAF::kOk);
                                    $stopped = 1;
                                    if(open(SIOFILE, ">" . $params{key}))
                                    {
                                        close(SIOFILE);
                                    }
                                    else
                                    {
                                        handle_failure("Cannot open " . $params{key} . ": $!");
                                        foreach my $subtest (Subtest::get_live())
                                        {                                   #70
                                            $subtest->kill();
                                        }
                                    }
                                    $Trace->comment(msg => "\nThis will exit as soon as " .
                                                           "it's complete\n");
                                });

    $sio_instance->wait_finish();

    # if we're here and $finished is Not 1, then something bad happened
    #   and we should be Done
    if(!$stopped)
    {
        handle_failure("SIO Exited without asking it to.");
    }
    else
    {
        Tharn::pass_result("Everything is all good")->store();
    }

    $Trace->exit();                                                         #91
    exit(0);
# This Can NOT be indented
SCRIPT
}
# end _execute()

1;
