# $Id: //depot/prod/test/main/lib/Process/Setup.pm#1 $

#
# Copyright (c) 2008 NetApp, Inc., All Rights Reserved
# Permission is granted to use this script on a NetApp Product.
#   Any other use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @pod here

package Process::Setup;
use strict;
use Data::Dumper;
use Params::Validate qw(validate ARRAYREF HASHREF SCALAR);

use Tharn qw($Log subtest host $Staf param);

#for CIFS
use MapShare qw(:all);

#for NFS
use Mount::MountClass;

#for IPv6
use Net::IP qw(ip_get_version);

use TracePkg;
use base qw(Process);

our $Trace;
BEGIN
{
    $Trace = TracePkg->new();
}

sub _need_setup
{
    my $self = shift;
    return $self->{need_setup};
}
# end _need_setup()

=head2 setup

=over 4

C<setup(%params)>

=item Description

Setup the internals of the Process:Load object
and if necessary,
setup the connection between the clients and the nodes.

=item Input

=item * map => \%map

(required)

This should be passed in from Knopp's Config::Cluster.pm,
    but it can be created by hand too

=item * clients => \@Clients

(Optional if you gave the Client ArrayRef to the new() method)

Default = $self->{client_list}
This Specifies the which Clients to perform the setup on

=item * nfs_options => <\%HashRef>

(Optional)

$nfs_options->{path} = <Mount Point Path>
    (Default = the method generates the mount point path)
    Instead of generating a path, it will use this path to mount to.
        If this is Not an ABSOLUTE PATH, it will automatically prepend the path with:
        "/t/<thpl name>/"

$nfs_options->{options} = <Mount options>
    (Default = "hard,intr" + <possibly other options depeding on the Load Gen>)

$nfs_options->{transport} = <udp|tcp>
    (Default = "tcp")

$nfs_options->{version} = <1 .. 4.1>
    (Default = 3)

=item * cifs_options => <\%HashRef>

(Optional)

$cifs_options->{workgroup} = <workgroup of which the node would be part of>
    (Default = "WORKGROUP")

$cifs_options->{domain}    = <domain where the username exists>

$cifs_options->{username}  = <username on the domain/workgoup which has access to the share>

$cifs_options->{password}  = <password of the "username">

$cifs_options->{drive}     = <A-Z|a-z>
    (Default = "*" (mapping the share to next available drive letter))

=item * cleanup => <0 | 1>

(Optional)

Default = 0

Setting this to true would cause it to try to delete everything in the volume
    that the client is going to mount to

=item Output

None

=item Example

    $hmr_load->setup(clients => \@Clients, map => \%map);

=back

=cut

sub setup
{
    $Trace->entry();
    my $self = shift;

    my $hash = {clients      => {type => ARRAYREF|SCALAR, default => $self->{client_list}},
                map          => {type => ARRAYREF|HASHREF},
                nfs_options  => {type => HASHREF, optional => 1},
                cifs_options => {type => HASHREF, optional => 1},
                cleanup      => {type => SCALAR,  default  => 0},
    };
    my %params = validate(@_, $hash);

    # Verify that we have Clients and a Map
    if(!(defined($params{clients}) && defined($params{map})))
    {
        $self->_result(rc => $self->CONFIG(), msg => "'clients' AND 'map' must be defined");
        return;
    }

    # Normalize Clients
    if(ref($params{clients}) ne "ARRAY")
    {
        $params{client_list} = [$params{clients}];
    }
    else
    {
        $params{client_list} = $params{clients};
    }

    # Add the client_list to $self if it doesn't exist
    $self->{client_list} = $params{client_list} if(!(defined($self->{client_list})));

    my $do_setup = $self->_need_setup();

    $params{cifs_options}->{drive} = "*" if(!(exists($params{cifs_options}->{drive})));

    my $test_file = param('RUNID');
    $test_file =~ s|/|_|;

    my @clients_to_delete;

    # Build the connection info and start the connection process for EACH Client
    foreach my $client (@{$params{client_list}})
    {
        my $mnt_pt;
        my $map_hash = $params{map}->{$client};
        $map_hash = @{$map_hash}[0] if(ref($map_hash) eq "ARRAY");
        
        # Override default NFS options if defined in the client map
        if(exists($map_hash->{nfs_options}))
        {
            $params{nfs_options}->{options} = $map_hash->{nfs_options}->{options} if (exists($map_hash->{nfs_options}->{options}));
            $params{nfs_options}->{version} = $map_hash->{nfs_options}->{version} if (exists($map_hash->{nfs_options}->{version}));
            $params{nfs_options}->{transport} = $map_hash->{nfs_options}->{transport} if (exists($map_hash->{nfs_options}->{transport}));
            $params{nfs_options}->{path} = $map_hash->{nfs_options}->{path} if (exists($map_hash->{nfs_options}->{path}));
        }
        
        # Validate NFS options
        if(!(exists($params{nfs_options}->{options})))
        {
            $params{nfs_options}->{options} = "hard,intr";
        }
        
        if(!(exists($params{nfs_options}->{transport})))
        {
            $params{nfs_options}->{transport} = "tcp";
        }
        elsif($params{nfs_options}->{transport} !~ /^udp$|^tcp$|udp6$|^tcp6$/)
        {
            $self->_result(rc  => $self->CONFIG(),
                           msg => "NFS transport [" . $params{nfs_options}->{transport} .
                                  "] is not supported.");
            return;
        }
        
        if(!(exists($params{nfs_options}->{version})))
        {
            $params{nfs_options}->{version} = 3;
        }
        elsif($params{nfs_options}->{version} !~ /^[1234]/)
        {
            $self->_result(rc  => $self->CONFIG(),
                           msg => "NFS Version [" . $params{nfs_options}->{version} .
                                  "] is not supported.");
            return;
        }
    
        if(($params{nfs_options}->{transport} eq "udp") && ($params{nfs_options}->{version} =~ /^4/))
        {
            $self->_result(rc => $self->CONFIG(), msg => "NFSv".$params{nfs_options}->{version}." over UDP is not supported");
            return;
        }
        
        $self->{clients}->{$client}->{host_obj} = host($client) if(!(defined($self->{clients}->{$client}->{host_obj})));
        # Modify mount options for IPv6
        if (ip_get_version($map_hash->{data_ip}) == 6) {
            if($self->{clients}->{$client}->{host_obj}->{hosttype} !~ /win/i) {
                $params{nfs_options}->{transport} = $params{nfs_options}->{transport}.'6' if ($params{nfs_options}->{transport} !~ /6$/);
                $map_hash->{data_ip} = '['. $map_hash->{data_ip}.']';
            } else {
                $map_hash->{data_ip} =~ s/:/-/g;
                $map_hash->{data_ip} = $map_hash->{data_ip}.'.ipv6-literal.net';
            }
        }
        
        # Append special mount options
        $params{nfs_options}->{options} .= $self->{_special_mount_options};
        
        my $options;
        if ($params{nfs_options}->{version} =~ /^4\.*(\d*)/) {
            my $minor = $1;
            $options = $params{nfs_options}->{options} .
                       ",vers=4" .
                       ",proto=" . $params{nfs_options}->{transport};
            $options .= ",minorversion=$minor" if (defined($minor) and length($minor) > 0);
        } else {
            $options = $params{nfs_options}->{options} .
                       ",vers=" . $params{nfs_options}->{version} .
                       ",proto=" . $params{nfs_options}->{transport};
        }
        
        if(defined($map_hash->{path}))
        {
            # We were given the Specific Directory to work from in the map
            $mnt_pt = $map_hash->{path};
        }
        elsif(defined($params{nfs_info}->{path}))
        {
            # We were given the Specific Directory to Mount to
            if($params{nfs_info}->{path} !~ /^\//)
            {
                $mnt_pt = "/t/" . $test_file . "/" . $params{nfs_info}->{path};
            }
            else
            {
                $mnt_pt = $params{nfs_info}->{path};
            }
        }
        else
        {
            $mnt_pt = "/t/".$self->{runid_appender}.$map_hash->{volume}->{volume};
        }

        if(!(exists($self->{clients}->{$client}->{random})))
        {
            $self->{clients}->{$client}->{seed} = time();
            $self->{clients}->{$client}->{random} = rand($self->{clients}->{$client}->{seed});
        }

        $map_hash->{node} = $map_hash->{data_ip} if(!(defined($map_hash->{node})));

        my $info = {data_ip      => $map_hash->{data_ip},
                    node         => $map_hash->{node},
                    path         => $mnt_pt,# this will be overwritten if $self->{need_setup} = 1
                    is_connected => 0};
        my $nfs_info_path = $info->{path} ;
	$info->{path} .= $self->{clients}->{$client}->{random};
        $self->{clients}->{$client}->{cleanup_dir} = $info->{path};# . "/" . $self->{clients}->{$client}->{random};

        $self->{clients}->{$client}->{info} = $info;
        $self->{clients}->{$client}->{host_obj} = host($client)
            if(!(defined($self->{clients}->{$client}->{host_obj})));

        if(int($self->{deny_hosttype}) > 0)
        {
            foreach my $outcast (@{$self->{deny_hosttype}})
            {
                if($self->{clients}->{$client}->{host_obj}->{hosttype} =~ /$outcast/i)
                {
                    $self->{clients}->{$client} = undef;
                    eval{delete($self->{clients}->{$client});};
                    push(@clients_to_delete, $client);
                    $Trace->warn(msg => "ALL " . $outcast .
                                " clients are forbidden to execute " .
                                $self->type() . "\n" . $client . " has " .
                                "been deleted from the list of Clients");
                }
            }
        }

        if($self->{clients}->{$client}->{host_obj}->{hosttype} !~ qr/win/i)
        {
            $self->{clients}->{$client}->{nfs_info} = {mount_pt => $nfs_info_path,
                                                       junction => $map_hash->{volume}->{"junction-path"},
                                                       options  => $options,
                                                       version  => $params{nfs_options}->{version},
                                                       };
        }
        else
        {
            $params{cifs_options}->{share} = "\\\\$info->{data_ip}\\" . $map_hash->{share};
            $self->{clients}->{$client}->{cifs_info} = \%{$params{cifs_options}};
        }

        if($do_setup && (exists($self->{clients}->{$client})))
        {
            my $runid = $self->{runid_appender} . "setup_" . $client .
                            "_" . $self->{clients}->{$client}->{random};

            my $client_path;
            if($self->{clients}->{$client}->{host_obj}->{hosttype} !~ qr/win/i)
            {
                my %os_mntopts;
                my %opts = (server => $info->{data_ip}, path => $self->{clients}->{$client}->{nfs_info}->{junction},
                            chmod  => "0777", client => $client, mntopts=> $options,
                            mountpoint => $mnt_pt, mountretry => 5,
                            sharing => "shared", timeout => 30);
                my @tmp = split /,/, $options;
                foreach my $opt (@tmp)
                {
                    # Disable chmod of read only mounts
                    if ($opt =~ /^ro$/) {
                        $opts{chmod} = -1;
                    }
                    if ($opt =~ /=/)
                    {
                        my @t = split /=/, $opt;
                        $os_mntopts{$t[0]} = $t[1];
                    }
                    else
                    {
                        $os_mntopts{$opt} = $MountClass::novalue;
                    }
                }

                if (!$os_mntopts{soft})
                {
                    $os_mntopts{hard} = $MountClass::novalue;
                }
                if (!$os_mntopts{nointr})
                {
                    $os_mntopts{intr} = $MountClass::novalue;
                }
                if (!$os_mntopts{quota})
                {
                    $os_mntopts{noquota} = $MountClass::novalue;
                }

                eval
                {
                    # Get the hostrec entry for the client and create a connection to it
                    my $h = Tharn::host($opts{client});
                    $opts{cc} = Tharn::connect($opts{client}."#staf-persay");
                    # Create the MountClass object
                    my $mountobj = Mount::MountClass->new($opts{cc});
                    $client_path = $mountobj->mount(\%opts, \%os_mntopts);
                };
            }
            else
            {
                my @shares = list_maps(client => $client);
                foreach my $share_hash (@shares)
                {
                    $share_hash->{path} =~ s/^(\S+)/$1/;
                    my $path = $1;
                    $Trace->comment(msg=>$path);
                    $Trace->comment(msg=>$self->{clients}->{$client}->{cifs_info}->{share});
                    if($path eq $self->{clients}->{$client}->{cifs_info}->{share})
                    {
                        $client_path = $share_hash->{drive};
                        last;
                    }
                }
                 eval{$client_path = map_cifs_share(%{$self->{clients}->{$client}->{cifs_info}}, client => $client);}
                    if(!(defined($client_path)));
            }
            if(!(defined($client_path)))
            {
                $self->_result(rc  => $self->FATAL(),
                               msg => "Error occured while mounting/mapping/connecting\n" . $@);
            }
            $self->{clients}->{$client}->{info}->{path} = $client_path;
            $self->{clients}->{$client}->{info}->{is_connected} = 1;
            $self->{clients}->{$client}->{cleanup_dir} = $client_path . "/" .
                $self->{clients}->{$client}->{random};
            $Trace->trace(msg=> "Successfully connected to Node from " . $client.
                                " on " . $client_path);
        }
    }

    foreach my $client (@clients_to_delete)
    {
        my $z;
        for(my $i = 0 ; $i < int(@{$self->{client_list}}) ; $i++)
        {
            if(@{$self->{client_list}}[$i] eq $client)
            {
                $z = $i;
                $i = int(@{$self->{client_list}});
            }
        }
        splice(@{$self->{client_list}}, $z, 1);
    }

    if((int(@{$self->{client_list}}) < 1) || (int(keys(%{$self->{clients}})) < 1))
    {
        $Trace->warn(msg => "There are No Clients to execute on.");
        $Trace->warn(msg => "client_list = " . Dumper($self->{client_list}));
        $Trace->warn(msg => "clients = " . Dumper($self->{clients}));

        $self->_result(rc => $self->CONFIG(), msg => "All of the " .
                       join(", ", @{$self->{deny_hosttype}}) .
                       " clients have been deleted resulting in a lack of" .
                       " clients.");
    }

    if($params{cleanup})
    {
        foreach my $client (keys(%{$self->{clients}}))
        {
            my $dir = $self->{clients}->{$client}->{cleanup_dir};
            print "Here is my cleanup dir: $dir\n";
            my $endpoint = $self->{clients}->{$client}->{host_obj}->net_best_address();
            $Staf->fs_delete(
                endpoint => $endpoint,
                path     => $dir,
                children => 1,
                on_error => sub {},
            );
        }
    }

    $self->_result();
    $Trace->exit();
}
# end setup()

=head2 cleanup

=over 4

C<cleanup(%params)>

=item Description

Cleanup the internals of the Process:Load object
and if necessary,
Cleanup the connection between the clients and the nodes.

=item Input

=item * clients => \@Clients

(Optional)

Default = all clients that it knows about

Specifies the Client[s] to disconnect from the node

=item * map => \%map

(Optional)

Specifies some client/node maping

This should be passed in from Knopp's Config::Cluster.pm,
    but it can be created by hand too

=item * nfs_path => <full path of mount point>

(Optional)

This is if you want to umount a specific dir

=item * cifs_share => <\\\\server\\share>

(Optional)

This is if you want to disconnect a specific map

=item * cleanup => <0 | 1>

(Optional)

Default = 0

Setting this to true would cause it to try to delete everything that this
    object wrote to the volume

=item Output

None

=item Example

    $hmr_load->cleanup();

=back

=cut

sub cleanup
{
    $Trace->entry();
    my $self = shift;
    my $hash = {clients   => {type => ARRAYREF|SCALAR, optional => 1},
                map       => {type => ARRAYREF|HASHREF,optional => 1},
                nfs_path  => {type => SCALAR, optional => 1},
                cifs_share=> {type => SCALAR, optional => 1},
                cleanup   => {type => SCALAR, default  => 0},
                };
    my %params = validate(@_, $hash);

    if(!(exists($params{clients})))
    {
        if(exists($self->{client_list}))
        {
            $params{clients} = $self->{client_list};
        }
        elsif(exists($self->{clients}))
        {
            my @client_list = keys(%{$self->{clients}});
            $params{clients} = \@client_list;
        }
        else
        {
            $self->_result(rc  => $self->FATAL(),
                           msg => "Can't Find Clients to cleanup from");
            return;
        }
    }

    $params{clients} = [$params{clients}] if(ref($params{clients}) ne "ARRAY");

    if($self->_need_setup())
    {
        foreach my $client (@{$params{clients}})
        {
            if($params{cleanup})
            {
                my $dir = $self->{clients}->{$client}->{cleanup_dir};
                my $endpoint = $self->{clients}->{$client}->{host_obj}->net_best_address();
                $Staf->fs_delete(endpoint => $endpoint, path => $dir, children => 1);
            }

            my $cleanup = 0;
            if((exists($self->{clients}->{$client}->{info})) &&
               ($self->{clients}->{$client}->{info}->{is_connected}))
            {
                $cleanup = 1;
            }
            else
            {
                $cleanup = 1;
            }
            if($cleanup)
            {
                if(!(defined($self->{clients}->{$client}->{host_obj})))
                {
                    $self->{clients}->{$client}->{host_obj} = host($client);
                }

                if(!(exists($self->{clients}->{$client}->{random})))
                {
                    $self->{clients}->{$client}->{seed} = time();
                    $self->{clients}->{$client}->{random} = rand($self->{clients}->{$client}->{seed});
                }
                my $runid = $self->{runid_appender} . "cleanup_".$client.
                                 "_" . $self->{clients}->{$client}->{random};
                my $path;
                my $share;
                if(defined($self->{clients}->{$client}->{info}))
                {
                    $path = $self->{clients}->{$client}->{info}->{path};
                }
                else
                {
                    $Trace->debug(msg=>"Client map = " . Dumper($params{map}->{$client}));
                    if($self->{clients}->{$client}->{host_obj}->{hosttype} !~ qr/win/i)
                    {
                        if(exists($params{nfs_path}))
                        {
                            # We were given the Specific Directory to Mount to
                            if($params{nfs_path} !~ /^\//)
                            {
                                $path = "/t/" . $1 . "/" . $params{nfs_path};
                            }
                            else
                            {
                                $path = $params{nfs_path};
                            }
                        }
                        elsif(exists($params{map}->{$client}))
                        {
                            my $map_hash = $params{map}->{$client};
                            if(ref($map_hash) eq "ARRAY")
                            {
                                $map_hash = @{$map_hash}[0];
                            }
                            $path = "/t/" . $self->{runid_appender} .
                                        $map_hash->{volume}->{volume};
                        }
                        else
                        {
                            $self->_result(rc  => $self->FATAL(),
                                           msg => "I don't know what to umount");
                            return;
                        }
                    }
                    else
                    {
                        if(exists($params{cifs_share}))
                        {
                            $share = $params{cifs_share};
                        }
                        else
                        {
                            $share = "\\\\".$params{map}->{$client}->{data_ip}.
                                     "\\" . $params{map}->{share};
                            if(!(defined($share)))
                            {
                                $self->_result(rc => $self->CONFIG(),
                                               msg => "I don't know what to unmap.");
                                return;
                            }
                        }
                    }
                }
                $Trace->debug(msg=>"Path = " . $path);
                if($self->{clients}->{$client}->{host_obj}->{hosttype} !~ qr/win/i)
                {
                    my $umount;
                    my $h = Tharn::host($client);

                    # Create the MountClass object
                    my %opts_ref;
                    $opts_ref{force} = 1;
                    $opts_ref{mountpoint} = $path;
                    $opts_ref{cc} = Tharn::connect($client."#staf-persay");
                    $opts_ref{timeout} = 60;
                    my $mountobj = Mount::MountClass->new($opts_ref{cc});

                    eval{$umount = $mountobj->umount(\%opts_ref);};
                    if(!(defined($umount)))
                    {
                        $Trace->comment(msg=>"Umount = " . Dumper($@) . "\n" . Dumper($umount));
                        my $err = $@;
                        if($err !~ qr/no such file or directory/i)
                        {
                            $self->_result(rc  => $self->FATAL(),
                                           msg => "Error occured while umounting: ".
                                                  $err);
                            return;
                        }
                    }
                }
                else
                {
                    my %options = (force => 1, client => $client);
                    if(defined($path))
                    {
                        $options{drive} = $path;
                    }
                    else
                    {
                        $options{mappath} = $share;
                    }
                    eval{unmap_cifs_share(%options);};
                    if($@)
                    {
                        $self->_result(rc  => $self->FATAL(),
                                       msg => "Error occured while unmapping.\n" . $@);
                    }
                }
            }
        }
    }

    # Cleaning up so setup, start, stop, cleanup, ... can be run in a loop.
    foreach my $client (@{$params{clients}})
    {
        eval{delete($self->{clients}->{$client}->{info});};
        eval{delete($self->{clients}->{$client}->{nfs_info});};
        eval{delete($self->{clients}->{$client}->{cifs_info});};
        eval{delete($self->{clients}->{$client}->{cmd});};
    }
    $Trace->debug(msg=>"cleanup finished for " . join(",", @{$params{clients}}));
    $self->_result();
    $Trace->exit();
}
# end cleanup()


1;
