# $Id$
#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Protocol Client File Module for any operating system
## @author dl-wafl-proto-qa@netapp.com
## @status shared
## @pod here

=pod

=head1 NAME

Protocol::AgnosticClientFile

=head1 DESCRIPTION

AgnosticClientFile provides API's to manipulate files on any client.
This class SHOULD NOT be called directly. Its methods are called by the
Protocol::ClientFile and Protocol::ClientFilehandle packages. To use these
methods instantiate a Protocol::ClientFile and use its public methods.

=head1 METHODS (not public)

=cut

############################################################################
#  AgnosticFileOperations
############################################################################

package Protocol::AgnosticClientFile;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR GLOB OBJECT ARRAYREF);
use Digest::SHA;
use NATE::BaseException;
use Tharn qw($Log);

=head2 checksum()

Checksum a file

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be opened

=item Returns

  SCALAR: The hash of the file

=item Example

 my $file_hash = Protocol::AgnosticClientFile::checksum(file_name   => $file_name);

=back

=back

=cut

sub checksum {
    my $dummy_arg = shift; # Skip the package name
    my %args = Params::Validate::validate(
        @_,
        {
            file_name => { type => SCALAR },
        }
    );
    my $file_name = $args{file_name};
    my $fh;
    open $fh, "<", $file_name or NATE::BaseException->throw(
        "File could not be opened for checksum."
    );;
    my $digest = Digest::SHA->new(256);
    $digest->addfile($fh);
    close $fh;
    return $digest->hexdigest;
}

=head2 exists()

Check if a file exists

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be checked

=item Returns

  1 if the file is found
  0 if the file is not found

=item Example

 if (Protocol::AgnosticClientFile::exists(file_name => $file_name)) {
    # file found
    $Test->nlog("File $file_name found!");
 } else {
    # file not found
    $Test->nlog("File $file_name not found!");
 }

=back

=back

=cut

sub exists {
    my $dummy_arg = shift; # Skip the package name
    my %args = Params::Validate::validate(
        @_,
        {
            file_name => { type => SCALAR },
        }
    );
    my $file_name = $args{file_name};

    # Check if file is a symlink
    my $is_symlink = ( -l $file_name );

    # Check if file exists
    my $does_file_exist = ( -e $file_name );

    if ( ( defined($is_symlink) && $is_symlink == 1) || 
        ( defined($does_file_exist) && $does_file_exist == 1) ) {
        return 1;
    }
    else {
        return 0;
    }
}

=head2 size()

Return a file's size in bytes

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file whose size is requested

=item Returns

  SCALAR value of the file's size in bytes

=item Example

 my $file_size = Protocol::AgnosticClientFile::size(file_name => $file_name);
 $Test->nlog("File $file_name size is $file_size bytes");

=back

=back

=cut

sub size {
    my $dummy_arg = shift; # Skip the package name
    my %args = Params::Validate::validate(
        @_,
        {
            file_name => { type => SCALAR },
        }
    );

    my $file_size = (stat($args{file_name}))[7]
        or NATE::BaseException->throw(
                    "failed to get size for file " .
                    "$args{file_name}!\n$@\n"
                );

    return $file_size;
}

=for html list_dir 

=head2 list_dir

Return the contents of a directory.
All files/pipes/directories contained in the directory are listed by name.
No other information is provided.

=over

=item Options

=over

=item dir_path => $dir_path

 (Required) The full path of the directory to be listed

=back

=item Returns

  ARRAYREF (reference to a list of files listed in the directory.)

=item Example

 my $list_ref = Protocol::AgnosticClientFile::list_dir( dir_path => $dir_path );
 my @list = @{$list_ref};
 foreach my $f (@list) {
   $Test->nlog ("\$f = $f\n");
   }

=back

=cut

sub list_dir {
    my $dummy_arg = shift; # Skip the package name
    my %args = Params::Validate::validate(
        @_,
        {
            dir_path => { type => SCALAR },
        }
    );

    my $dir_path = $args{dir_path};
    opendir( D, $dir_path ) || NATE::BaseException->throw(
        "failed to open directory with given path: " . $dir_path . "\n$!\n" );

    ( my @list = readdir(D) ) || NATE::BaseException->throw(
        "failed to list directory with given path: " . $dir_path . "\n$!\n" );

    closedir(D) || NATE::BaseException->throw(
        "failed to close directory with given path: " . $dir_path . "\n$!\n" );

    return \@list;
}

=head2 chdir

Change the working directory to the provided path

=over

=item Options

=over

=item dir_path => $dir_path

 (Required) The full path of the directory to be made as working directory

=back

=item Returns

Nothing, if successful.
Throws exception, if error occurs.

=item Example

 $fileops_obj->chdir( dir_path => $dir_path );

=back

=cut

sub chdir {
    my $dummy_arg = shift; # Skip the package name
    my %args = Params::Validate::validate(
        @_,
        {
            dir_path => { type => SCALAR },
        }
    );

    my $dir_path = $args{dir_path};
    if ( !chdir( $dir_path ) ) {
        NATE::BaseException->throw(
            "failed to change to directory with given path: " . $dir_path . "\n$!\n" );
    }
}

=head2 stat()

Return a file's statistics

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file whose statistics are requested

=item Returns

ARRAYREF An array ref to an array of statistics about the file whose path is provided.

Statistics in order:
0   Device number of file system
1   Inode number
2   File mode (type and permissions)
3   Number of (hard) links to the file
4   Numeric user ID of file.s owner
5   Numeric group ID of file.s owner
6   The device identifier (special files only)
7   File size, in bytes
8   Last access time since the epoch
9   Last modify time since the epoch
10  Inode change time (not creation time!) since the epoch
11  Preferred block size for file system I/O
12  Actual number of blocks allocated

=item Example

my $stat_ref = Protocol::AgnosticClientFile::stat($filename);
my @file_stats = @{$stat_ref};
for my $stats (@file_stats){
...
}

=back

=back

=cut

sub stat {

    my $dummy_arg = shift; # Skip the package name
    my %args = Params::Validate::validate(
        @_,
        {
            file_name => { type => SCALAR },
        }
    );

    my @file_stat = ( stat( $args{file_name} ) ) or NATE::BaseException->throw(
        "failed to get size for file " . "$args{file_name}!\n$!\n" );

    return \@file_stat;
}

=head2 utime 

Update the access and modification timestamps on a file

=over

=item Options

=over

=item file_name => $file_name

 (Required) The full path of the file to modify the timestamps on

=item atime => $atime

 (Options) The new access time (defaults to undef = current time)

=item mtime => $mtime

 (Optional) The new modification time (defaults to undef = current time)

=back

=item Returns

Nothing, if successful.
Throws exception, if error occurs.

=item Example

 $fileops_obj->utime(file_name => $file_name);

=back

=cut

sub utime {

    my $dummy_arg = shift; # Skip the package name

    my %args = Params::Validate::validate(
        @_,
        {
            file_name => { type => SCALAR },
	    atime     => {
	    	type     => SCALAR,
		optional => 1,
		default  => undef,
	    },
	    mtime     => {
	    	type     => SCALAR,
		optional => 1,
		default  => undef,
	    },
        }
    );

    my $file_name = $args{file_name};
    my $atime     = $args{atime};
    my $mtime     = $args{mtime};

    if (!utime($atime, $mtime, $file_name)) {
        NATE::BaseException->throw("Failed to update the timestamps ".
	    "on file: $file_name");
    }

    return;

}

=head1 SEE ALSO

=over

=item Protocol::ClientFileOperations

=item Protocol::ClientFilehandle

=item Protocol::LinuxClientFile

=item Protocol::WindowsClientFile

=back

=head1 STATUS

    Shared

=head1 AUTHOR

    WAFL Protocol QA (dl-wafl-proto-qa@netapp.com)

=head1 COPYRIGHT

    Copyright 2011, Network Appliance Inc., All Rights Reserved.

=cut

1;
