# $Id$

## @summary Library for CIFS change notification
## @pod here

=head1 NAME

=head1 SYNOPSIS

    use Protocol::CIFS::ChangeNotify qw(:all);

=head1 ROUTINES

=cut

package Protocol::CIFS::ChangeNotify;
use strict;

use Protocol::CIFS::ReadDirectoryChanges;

BEGIN {
  use Module::Common;
  Module::Common::setup_export_tags(
    ops => [ qw(cifsNotifyStart
                cifsNotifyDetails
                cifsNotifyActionToStr
                cifsNotifyFilterToStr
                cifsNotifyStop) ],
    constants => [ qw(FILE_NOTIFY_CHANGE_FILE_NAME
                      FILE_NOTIFY_CHANGE_DIR_NAME
		      FILE_NOTIFY_CHANGE_NAME
		      FILE_NOTIFY_CHANGE_ATTRIBUTES
		      FILE_NOTIFY_CHANGE_SIZE
                      FILE_NOTIFY_CHANGE_LAST_WRITE
		      FILE_NOTIFY_CHANGE_LAST_ACCESS
		      FILE_NOTIFY_CHANGE_CREATION
		      FILE_NOTIFY_CHANGE_EA
		      FILE_NOTIFY_CHANGE_SECURITY
		      FILE_NOTIFY_CHANGE_STREAM_NAME
		      FILE_NOTIFY_CHANGE_STREAM_SIZE
		      FILE_NOTIFY_CHANGE_STREAM_WRITE

		      FILE_ACTION_ADDED
		      FILE_ACTION_REMOVED
		      FILE_ACTION_MODIFIED
		      FILE_ACTION_RENAMED_OLD_NAME
		      FILE_ACTION_RENAMED_NEW_NAME
		      FILE_ACTION_ADDED_STREAM
		      FILE_ACTION_REMOVED_STREAM
		      FILE_ACTION_MODIFIED_STREAM) ]
  );
}

use Params::Validate v0.70 qw(validate SCALAR SCALARREF HASHREF UNDEF);
use Carp;
use base qw(Exporter);

sub cifsNotifyStart {
    my $dummy_arg = shift; # Skip the package name

    my (%args) = validate (@_, {
	sPath		=> { type => SCALAR },
	bWatchSubtree	=> { type => SCALAR, optional => 1, default => 0 },
	uFilter		=> { type => SCALAR, optional => 1, default => 0xfff }
    });

    return new Protocol::CIFS::ReadDirectoryChanges(path => $args{sPath},
                                           subtree => $args{bWatchSubtree},
					   filter => $args{uFilter});
}

sub cifsNotifyStop {
    my $dummy_arg = shift; # Skip the package name

    my (%args) = validate (@_, {
        hNotify		=> { type => HASHREF }
    });

    return $args{hNotify}->destroy();
}

sub cifsNotifyDetails {
    my $dummy_arg = shift; # Skip the package name

    my (%args) = validate (@_, {
	hNotify		=> { type => HASHREF }
    });

    return join("\0",$args{hNotify}->details());
}

sub cifsNotifyActionToStr {
    my (%args) = validate (@_, {
	nAction		=> { type => SCALAR }
    });

    if ($args{nAction} == FILE_ACTION_UNSPECIFIED) {
	return "Unspecified";
    } elsif ($args{nAction} == FILE_ACTION_ADDED) {
	return "Added";
    } elsif ($args{nAction} == FILE_ACTION_REMOVED) {
	return "Removed";
    } elsif ($args{nAction} == FILE_ACTION_MODIFIED) {
	return "Modified";
    } elsif ($args{nAction} == FILE_ACTION_RENAMED_OLD_NAME) {
	return "Rename (Old)";
    } elsif ($args{nAction} == FILE_ACTION_RENAMED_NEW_NAME) {
	return "Rename (New)";
    } elsif ($args{nAction} == FILE_ACTION_ADDED_STREAM) {
	return "Added (Stream)";
    } elsif ($args{nAction} == FILE_ACTION_REMOVED_STREAM) {
	return "Removed (Stream)";
    } elsif ($args{nAction} == FILE_ACTION_MODIFIED_STREAM) {
	return "Modified (Stream)";
    }

    return "Unknown";
}

sub cifsNotifyFilterToStr {
    my (%args) = validate (@_, {
	nFilter		=> { type => SCALAR }
    });

    my $filter;

    if ($args{nFilter} & FILE_NOTIFY_CHANGE_FILE_NAME) {
	_addFilterStr(\$filter, "File name");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_DIR_NAME) {
	_addFilterStr(\$filter, "Dir name");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_ATTRIBUTES) {
	_addFilterStr(\$filter, "Attributes");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_SIZE) {
	_addFilterStr(\$filter, "Size");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_LAST_WRITE) {
	_addFilterStr(\$filter, "Last write");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_LAST_ACCESS) {
	_addFilterStr(\$filter, "Last access");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_CREATION) {
	_addFilterStr(\$filter, "Creation");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_EA) {
	_addFilterStr(\$filter, "Extended attributes");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_SECURITY) {
	_addFilterStr(\$filter, "Security");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_STREAM_NAME) {
	_addFilterStr(\$filter, "Stream name");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_STREAM_SIZE) {
	_addFilterStr(\$filter, "Stream size");
    }
    if ($args{nFilter} & FILE_NOTIFY_CHANGE_STREAM_WRITE) {
	_addFilterStr(\$filter, "Stream write");
    }

    return $filter;
}

sub _addFilterStr {
    my $filter = shift;
    my $text = shift;

    ($$filter) ? $$filter .= ", " : 1;
    $$filter .= $text;
}

1;
