# Hey emacs, this is basically -*- perl -*-
# vim: set filetype=perl tabstop=4 shiftwidth=4 shiftround expandtab:
# $Id: //depot/prod/test/main/lib/Protocol/ClientFileHandle.pm#1 $

#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Protocol Client File Handle Module
## @author dl-wafl-proto-qa@netapp.com
## @status shared
## @pod here

=pod

=head1 NAME

Protocol::ClientFileHandle

=for html <hr>

=head1 DESCRIPTION

This library is used to peform basic file-handle level operations on a file such
as read, write, lock and unlock.  A Protocol::ClientFileHandle object instance
is returned when a file is opened via a Protocol::ClientFileOperations object
instance.

=for html <hr>

=head1 SYNOPSIS

# The Protocol::ClientFileHandle instance was returned from a previous

# open call for a file using the Protocol::ClientFileOperations instance.

# Write to the file

my $num_bytes = $fh_obj->write(
    offset => 0,
    num_bytes => 50
);

# Read from the file

my $contents = $fih_obj->read(
    offset=> 0,
    num_bytes => 10
);

# Acquire a byte-range lock on the file

$fh_obj->lock(
    lock_type => write-lock,
    lock_wait => non-waiting,
    offset => 0,
    num_bytes => 20
);

# Release the byte-range lock on the file

$fh_obj->unlock(
    offset => 0,
    num_bytes => 20
);

=cut

package Protocol::ClientFileHandle;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;

use Data::Dumper;

use Protocol::Executor;

use NATE::BaseException;
use Params::Validate qw(:all);
use Tharn qw($Log);

=for html <hr>

=head1 METHODS

=head2 new

Instantiates a Protocol::ClientFileHandle object instance.

=over

=item Options

=over

=item client => $client

 (Required) The NACL::C::Client object instance.

=item file_name => $file_name

 (Required) The fully specified name of the file being accessed

=item file_handle => $file_handle

 (Required) The system file handle for the file

=item executor => $executor

 (Required) An object of Protocol::Executor,
            usually got from Protocol::ClientFileOperations

=back

=item Returns

 Returns a new Protcol::ClientFileHandle object instance if successful,
 throws an error on failure.

=item Example

 my $file_obj = Protocol::ClientFileHandle->new(
     client => $client_obj,
     file_name => $file_name,
     file_handle => $file_handle,
     executor => $executor
 );

=back

=cut

sub new() {

    my $self = {};
    my $class = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            client => {isa => 'NACL::C::Client'},
            file_name => {type => SCALAR},
            file_handle => {type => SCALAR},
            executor => {isa => 'Protocol::Executor'}
        }
    );
    bless ($self, $class) unless ref($class);

    # Internal object bookkeeping

    $self->{client_obj} = $args{client};
    $self->{file_name} = $args{file_name};
    $self->{file_handle} = $args{file_handle};
    $self->{executor} = $args{executor};
    $self->{locking_requests} = ();

    # Determine the client type and set the appropriate
    # implementor library package

    my $client_type = $args{client}->hostrec()->{hosttype};
    if ($client_type =~ qr/nix/i) {
        $self->{client_type} = 'linux';
        $self->{platform_pkg} = 'Protocol::LinuxClientFile';
    } elsif ($client_type =~ qr/indow/i) {
        $self->{client_type} = 'windows';
        $self->{platform_pkg} = 'Protocol::WindowsClientFile';
    } else {
        NATE::BaseException->throw(
           "Unsupported client type ($client_type)."
        );
    }

    return $self;

}


=for html <hr>

=head2 read

Reads the number of bytes starting at the specified offset.

=over

=item Options

=over

=item offset => $offset

 (Required) The offset into the file to begin reading from

=item num_bytes => $num_bytes

 (Required) The number of bytes to read from the file

=back

=item Returns

 Returns the contents of the read operation.

=item Example

 my $contents = $fh_obj->read(
     offset => $offset,
     num_bytes => $num_bytes
 );

=back

=cut

sub read() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            offset => {type => SCALAR},
            num_bytes => {type => SCALAR},
        }
    );

    # Remotely execute the read command

    my $contents = $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'read',
        args => {
            file_handle => $self->{file_handle},
            offset => $args{offset},
            num_bytes => $args{num_bytes},
        },
    );

    return Protocol::Executor::decode($contents);
}

=for html <hr>

=head2 write

Writes a given number of bytes or a specified content string (either
the num_bytes or contents parameter but not both must be specified)
to a given file starting at the specified offset.

=over

=item Options

=over

=item offset => $offset

 (Required) The offset into the file to begin writing at

=item num_bytes => $num_bytes

 (Optional) The number of bytes to write to the file
 Either the num_bytes or contents parameter must be specified

=item contents => $contents

 (Optional) The contents string to write to the file
 Either the num_bytes or contents parameter must be specified

=back

=item Returns

 The number of bytes written if successful, throws an error on failure.

=item Example

 my $bytes_written = $fh_obj->write(
     offset => $offset,
     contents => $contents,
 );

=back

=cut

sub write() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            offset => {
                type => SCALAR
            },
            num_bytes => {
                type => SCALAR,
                optional => 1,
            },
            contents => {
                type => SCALAR,
                optional => 1,
            },
        }
    );

    if (!exists $args{num_bytes} xor exists $args{contents}) {
        NATE::BaseException->throw(
            "Either the num_bytes or contents parameter but ".
            "not both must be specified."
        );
    }

    # If the num_bytes parameter was specified, form a string
    # of a fixed format of the requested size to write

    if (exists $args{num_bytes}) {
        my $line = '123456789\n';
        my $len = length($line);
        $args{contents} = ($line x int($args{num_bytes} / $len)) .
            substr($line, 0, $args{num_bytes} % $len);
    }

    # Remotely execute the write command

    my $bytes_written = $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'write',
        args => {
            file_handle => $self->{file_handle},
            offset => $args{offset},
            contents => Protocol::Executor::encode($args{contents}),
        }
    );

    return $bytes_written;

}

=for html <hr>

=head2 set_allocation_size

Sets the file's allocation size using this filehandle to the
specified number of bytes (past-EOF, at EOF, or less than EOF)

=over

=item Options

=over

=item allocation_size => $allocation_size

 (Required) The total file allocation size to set (in bytes)

=back

=item Returns

Nothing if successful, throws an error on failure.

=item Example

 $fh_obj->set_allocation_size(
     allocation_size => ($file_size + 2048),
 );

=back

=cut

sub set_allocation_size() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            allocation_size => {
                type => SCALAR,
            },
        }
    );

    if ($self->{client_type} ne 'windows') {
        NATE::BaseException->throw(
           "Unsupported Error: setting allocation sizes is only supported on Windows platforms!"
        );
    }
    # Remotely execute the write command

    return $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'set_allocation_size',
        args => {
            file_handle => $self->{file_handle},
            allocation_size => $args{allocation_size},
        }
    );

}

=for html <hr>

=head2 lock

Acquire a byte-range lock on a file.

=over

=item Options

=over

=item lock_type => $lock_type

 (Optional) The type of lock to acquire
 Valid values are:  exclusive, shared
 Defaults to exclusive

=item lock_wait => $lock_wait

 (Optional) The waiting symantics for the lock request
 Valid values are:  waiting, non-waiting
 Defaults to non-waiting

=item offset => $offset

 (Required) The offset into the file to acquire the lock at

=item num_bytes => $num_bytes

 (Required) The number of bytes to acquire the lock on

=item waiting_id => $waiting_id

 (Optional) When requesting a waiting operation, optionally
 pass a SCALAR to be populated with the ID with which to
 refer to the waiting request (used to check the status)

=back

=item Returns

 Returns nothing if successful, throws an error on failure.
 In the event of a waiting lock request, if the waiting_id
 parameter is specified, it will be populated with an ID with
 which to obtain the current status of the waiting request.

=item Example

 # non-waiting shared lock request
 $fh_obj->lock(
     lock_type => 'shared',
     lock_wait => 'non-waiting',
     offset => $offset,
     num_bytes => $num_bytes
 );

 # waiting exclusive lock request
 my $waiting_id;
 $fh_obj->lock(
     lock_type => 'exclusive',
     lock_wait => 'waiting',
     offset => $offset,
     num_bytes => $num_bytes,
     waiting_id => $waiting_id
 );

 # get waiting lock status using the waiting_id
 my $wait_lock_status = $fh_obj->get_waiting_lock_status(
                             waiting_id => $waiting_id
                         );

=back

=cut

sub lock() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            lock_type => {
                type => SCALAR,
                optional => 1,
                default => 'exclusive-lock',
            },
            lock_wait => {
                type => SCALAR,
                optional => 1,
                default => 'non-waiting',
            },
            offset => {type => SCALAR},
            num_bytes => {type => SCALAR},
            waiting_id => {
                type => SCALARREF,
                optional => 1,
            },
        }
    );

    # Translate the arguments and build up the platform-specific lock
    # command parameters

    my $lock_type = _translate_lock_type(
        $self->{client_type},
        $args{lock_type}
    );
    my $lock_wait = _translate_lock_wait(
        $self->{client_type},
        $args{lock_wait}
    );

    # Remotely execute the lock command

    my $lock_id = $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'lock',
        args => {
            file_handle => $self->{file_handle},
            lock_type => $lock_type,
            lock_wait => $lock_wait,
            offset => $args{offset},
            num_bytes => $args{num_bytes},
        },
    );

    # Update the waiting ID parameter that was optionally passed in
    # in the case of a waiting request

    if ($args{lock_wait} eq 'waiting' && exists $args{waiting_id}) {
        ${$args{waiting_id}} = $lock_id;
    }

    # Bookkeeping to keep track of the client lock operation requests

    push @{$self->{locking_requests}},
        {
            request_type => 'lock',
            lock_type => $args{lock_type},
            lock_wait => $args{lock_wait},
            offset => $args{offset},
            num_bytes => $args{num_bytes},
        };

    return;

}

=for html <hr>

=head2 unlock

Releases a byte-range lock on a file.

=over

=item Options

=over

=item offset => $offset

 (Required) The starting offset at which to release the lock

=item num_bytes => $num_bytes

 (Required) The number of bytes to release from the lock

=item unlock_wait => $unlock_wait

 (Optional) The waiting symantics for the unlock
 Valid values are:  waiting, non-waiting
 Defaults to non-waiting
 Only currently supported for Unix

=item waiting_id => $waiting_id

 (Optional) When requesting a waiting operation, optionally
 pass a SCALAR to be populated with the ID with which to
 refer to the waiting request (used to check the status)

=back

=item Returns

 Returns nothing if successful, throws an error on failure.
 In the event of a waiting request, if the waiting_id
 parameter is specified, it will be populated with an ID with
 which to obtain the current status of the waiting request.

=item Example

 $fh_obj->unlock(
     offset => $offset,
     num_bytes = $num_bytes,
 );

=back

=cut

sub unlock() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            offset => {type => SCALAR},
            num_bytes => {type => SCALAR},
            unlock_wait => {
                type => SCALAR,
                optional => 1,
                default => 'non-waiting',
            },
            waiting_id => {
                type => SCALAR,
                optional => 1,
            },
        }
    );

    # Translate the arguments and build up the platform-specific
    # unlock command parameters

    my $unlock_wait = _translate_lock_wait(
        $self->{client_type},
        $args{unlock_wait}
    );

    my %exec_args = (
        file_handle => $self->{file_handle},
        offset => $args{offset},
        num_bytes => $args{num_bytes},
    );
    if ($self->{client_type} eq 'linux') {
        $exec_args{unlock_wait} = $unlock_wait;
    }

    # Remotely execute the unlock command

    my $lock_id = $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'unlock',
        args => \%exec_args,
    );

    # Update the waiting ID parameter that was optionally passed in
    # in the case of a waiting request

    if ($args{unlock_wait} eq 'waiting' && exists $args{waiting_id} &&
        $self->{client_type} eq 'linux') {
        $args{waiting_id} = $lock_id;
    }

    # Bookkeeping to keep track of the client lock operation requests

    push @{$self->{locking_requests}},
        {
            request_type => 'unlock',
            offset => $args{offset},
            num_bytes => $args{num_bytes},
            lock_wait => $args{unlock_wait},
        };

    return;

}

=for html <hr>

=head2 get_lock_requests

Gets the current client view of the requested lock operations
for a given file handle.

=over

=item Options

=over

=item None

=back

=item Returns

 Returns an array of successful lock operation requests from the
 client point of view where each element in the array is a hash
 storing information about the request.  For each lock operation
 request, the hash contains the following information:
     request_type => 'lock' or 'unlock'
     lock_type => 'shared' or 'exclusive'
     lock_wait => 'waiting' or 'non-waiting'
     offset => <offset>
     num_bytes => <num_bytes>

=item Example

 my @lock_array = $fh_obj->get_lock_requests()

=back

=cut

sub get_lock_requests() {

    my $self = shift;

    # Enable for debugging
    #$self->_dump_lock_requests();

    return \@{$self->{locking_requests}};

}

=for html <hr>

=head2 get_waiting_lock_status

Gets the current status for a given waiting lock request.

=over

=item Options

=over

=item waiting_id => $waiting_id

 (Required) The lock ID of the waiting lock to return status on.

=back

=item Returns

 Returns the current status for a waiting lock request if successful,
 throws an error on failure.  Possible status values are:  waiting, granted.

=item Example

 my $wait_lock_status = $fh_obj->get_waiting_lock_status(
     waiting_id => $waiting_id
 );

=back

=cut

sub get_waiting_lock_status() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            waiting_id => {type => SCALAR},
        }
    );

    # Remotely execute the get_wait_status command

    my $thread_status = $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'get_wait_status',
        args => {
            file_handle => $self->{file_handle},
            wait_request_id => $args{waiting_id},
        },
    );

    # Translate the thread status to a lock status

    my $wait_status = $thread_status;
    if ($thread_status eq 'waiting') {
        $wait_status = $thread_status;
    } elsif ($thread_status eq 'exited') {
        $wait_status = 'granted';
    } else {
        # 'invalid' thread status
        NATE::BaseException->throw(
            "Unable to get current lock status for thread ".
            "ID ($args{waiting_id})."
        );
    }

    return $wait_status;

}

=for html <hr>

=head2 kill_waiting_lock_request

Kill a currently active waiting lock request.

=over

=item Options

=over

=item lock_id => $lock_id

 (Required) The lock ID of the waiting lock request to kill

=back

=item Returns

 Returns nothing if successful, throws an error on failure.

=item Example

 $fh_obj->kill_waiting_lock_request(
     lock_id => $lock_id
 );

=back

=cut

sub kill_waiting_lock_request() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            lock_id => {type => SCALAR},
        }
    );

    # Remotely execute the get_wait_status command

    $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'kill_waiting_request',
        args => {
            file_handle => $self->{file_handle},
            wait_request_id => $args{lock_id}
        },
    );

    return;

}

# Private functions

sub _dump_lock_requests () {

    my $self = shift;

    $Log->comment("Total Number of Lock Operation Requests = ".
        @{$self->{locking_requests}});
    foreach my $req_hash (@{$self->{locking_requests}}) {
        $Log->comment("Lock Operation Request Definition:");
        foreach my $key (keys %$req_hash) {
            $Log->comment("->  $key = $req_hash->{$key}");
        }
    }

    return;

}

sub _translate_lock_type () {

    my ($client_type, $lock_type) = @_;
    my %lock_type_values;

    if ($client_type eq 'windows') {
        %lock_type_values = (
            exclusive => 'LOCKFILE_EXCLUSIVE_LOCK',
            shared => '',
        );
    } elsif ($client_type eq 'linux') {
        %lock_type_values = (
            exclusive => 'F_WRLCK',
            shared => 'F_RDLCK',
        );
    } else {
        NATE::BaseException->throw(
           "Unsupported client type ($client_type)."
        );
    }

    if (exists $lock_type_values{$lock_type}) {
        return $lock_type_values{$lock_type};
    } else {
        return $lock_type;
    }

}

sub _translate_lock_wait () {

    my ($client_type, $lock_wait) = @_;
    my %lock_wait_values;

    if ($client_type eq 'windows') {
        %lock_wait_values = (
            waiting => '',
            'non-waiting' => 'LOCKFILE_FAIL_IMMEDIATELY',
        );
    } elsif ($client_type eq 'linux') {
        %lock_wait_values = (
            waiting => 'F_SETLKW',
            'non-waiting' => 'F_SETLK',
        );
    } else {
        NATE::BaseException->throw(
           "Unsupported client type ($client_type)."
        );
    }

    if (exists $lock_wait_values{$lock_wait}) {
        return $lock_wait_values{$lock_wait};
    } else {
        return $lock_wait;
    }

}

=head1 SEE ALSO

=over

=item Protocol::ClientFileOperations

=item Protocol::WindowsClientFile

=item Protocol::LinuxClientFile

=item Protocol::Executor

=back

=head1 STATUS

Shared

=head1 AUTHOR

WAFL Protocol QA (dl-wafl-proto-qa@netapp.com)

=head1 COPYRIGHT

Copyright 2011, Network Appliance Inc., All Rights Reserved.

=cut

1;
