# Hey emacs, this is basically -*- perl -*-
# vim: set filetype=perl tabstop=4 shiftwidth=4 shiftround expandtab:
# $Id: //depot/prod/test/main/lib/Protocol/ClientFileOperations.pm#1 $

#
# Copyright (c) 2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

# TODO: Consider adding readlink

## @summary Protocol Client File Operations Module
## @author dl-wafl-proto-qa@netapp.com
## @status shared
## @pod here

=pod

=head1 NAME

Protocol::ClientFileOperations

=for html <hr>

=head1 DESCRIPTION

This library is used to perform basic file system level operations
on a client such as open, close, rename, delete, link and unlink.

=for html <hr>

=head1 SYNOPSIS

# A NACL::C::Client object instance was previously created

# Create a Protocol::ClientFileOperations object instance to operate on

$fileops_obj = Protocol::ClientFileOperations->new(client => $client);

# Open a file (returns a Protocol::ClientFileHandle instance

$fh_obj = $fileops_obj->open(
    file_name => $file_name,
    access_mode => $access_mode,
    open_mode => $open_mode
);

# Close the open file handles for a file

$fileops_obj->close(file_name => $file_name);

# Rename a file

$fileops_obj->rename(
    file_name => $file_name,
    new_file_name => $new_file_name
);

# Delete a file

$fileops_obj->delete(file_name => $new_file_name);

# Destroy the Protocol::ClientFileOperations instance

$fileops_obj->destroy();

=cut

package Protocol::ClientFileOperations;

use Exporter qw(import);
@EXPORT_OK = qw();

use strict;
use warnings;

use Protocol::Executor;
use Protocol::ClientFileHandle;

use NATE::BaseException;
use Params::Validate qw(:all);
use Tharn qw($Log);

=for html <hr>

=head1 METHODS

=head2 new

=over

Instantiates a new Protocol::ClientFileOperations object for a given client.

=item Options

=over

=item client => $client

 (Required) The NACL::C::Client object instance.

=back

=item Returns

 Returns a new Protocol::ClientFileOperations object instance if successful,
 throws an error on failure.

=item Example

 my $fileops_obj = Protocol::ClientFileOperations->new(client => $client);

=back

=cut

sub new() {

    my $self = {};
    my $class = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            client => {isa => 'NACL::C::Client'}
        }
    );
    bless ($self, $class) unless ref($class);

    # Internal object bookkeeping

    $self->{client_obj} = $args{client};
    $self->{executor} = Protocol::Executor->new();

    # Determine the client type and set the appropriate
    # implementor library package

    my $client_type = $args{client}->hostrec()->{hosttype};
    if ($client_type =~ qr/nix/i) {
        $self->{client_type} = 'linux';
        $self->{platform_pkg} = 'Protocol::LinuxClientFile';
    } elsif ($client_type =~ qr/indow/i) {
        $self->{client_type} = 'windows';
        $self->{platform_pkg} = 'Protocol::WindowsClientFile';
    } else {
        NATE::BaseException->throw(
            "Unsupported client type ($client_type)."
        );
    }

    return $self;

}

=for html <hr>

=head2 destroy

=over

Destroys the Protocol::ClientFileOperations object instance.

=item Options

=over

=item None

=back

=item Returns

 Returns nothing if successful, throws an error on failure.

=item Example

 $fileops_obj->destroy();

=back

=cut

sub destroy() {

    my $self = shift;

    # Destroy the Executor object to trigger cleanup of the
    # underlying ClientRPC connections

    $self->{executor}->_destroy();

    return;

}

=for html <hr>

=head2 open

Opens a file on a client.

=over

=item Options

=over

=item file_name => $file_name

 (Required) The fully specified name of the file to be opened

=item access_mode => $access_mode

 (Required) The desired access mode for the file being opened
  Valid values are:  read, write, read-write

=item open_mode => $open_mode

 (Required) The desired open mode for the file being opened
 Valid values are:  open-existing, open-always, create-new,
 create-always, truncate-existing, open-append

=item share_mode => $share_mode

 (Optional) The desired share mode for the file being opened
 Valid values are:  shared-delete, shared-read, shared-write,
 shared-read-delete, shared-write-delete, shared-read-write,
 shared-read-write-delete, none
 This is ignored for Unix and defaults to shared-read-write for Windows

=item permissions => $permissions

 (Optional) The permissions (4 bits) for the file being opened
 This is ignored for Windows.

=item attributes => $attributes

 (Optional) The desired flags and attributes for the file being opened
 This can be any supported flags and attributes (or combinations combined
 with "|") specified in the MSDN documentation for CreateFile().
 This is ignored for Unix.

=back

=item Returns

 Returns a Protocol::ClientFileHandle object instance from which
 to access the file if successful, throws an error on failure.

=item Example

 my $fh_obj = $fileops_obj->open(
     file_name => $file_name,
     access_mode => $access_mode,
     open_mode => $open_mode
 );

=back

=cut

sub open() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
            access_mode => {type => SCALAR},
            open_mode => {type => SCALAR},
            share_mode => {
                optional => 1,
                type => SCALAR,
                default => 'shared-read-write',
            },
            attributes => {
                optional => 1,
                type => SCALAR,
            },
            permissions => {
                optional => 1,
                type => SCALAR,

            },
        }
    );

    # Translate the arguments and build up the platform-specific open
    # command parameters

    my $access_mode = _translate_access_mode(
        $self->{client_type},
        $args{access_mode}
    );
    my $open_mode = _translate_open_mode(
        $self->{client_type},
        $args{open_mode}
    );
    my $share_mode = _translate_share_mode(
        $self->{client_type},
        $args{share_mode}
    );

    my %exec_args = (
        file_name => $args{file_name},
        access_mode => $access_mode,
        open_flags => $open_mode,
    );

    if ($self->{client_type} eq 'linux') {
        if (exists $args{permissions}) {
            $exec_args{permissions} = $args{permissions};
        }
    }
    if ($self->{client_type} eq 'windows') {
        $exec_args{share_mode} = $share_mode;
        if (exists $args{attributes}) {
            $exec_args{attributes} = $args{attributes};
        }
    }

    # Issue the file open command on the remote client

    my $file_handle = $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'open',
        args => \%exec_args,
    );

    # Get a new ClientFileHandle object instance

    my $fh_obj = Protocol::ClientFileHandle->new(
        client => $self->{client_obj},
        file_name => $args{file_name},
        file_handle => $file_handle,
        executor => $self->{executor}
    );

    # Bookkeeping for the file handle information

    if (!exists $self->{file_mapping}{$args{file_name}}) {
        $self->{file_mapping}{$args{file_name}} = ();
    }
    push @{$self->{file_mapping}{$args{file_name}}}, $fh_obj;

    return $fh_obj;

}

=for html <hr>

=head2 close

Closes one or all open file handles to a file.

=over

=item Options

=over

=item file_name => $file_name

 (Optional) The fully specified name of the file to close
 Either the file_name or file_handle parameter must be specified.

=item file_handle => $file_handle_obj

 (Optional) The Protocol::ClientFileHandle object instance for
 the file handle to close.
 Either the file_name or file_handle parameter must be specified.

=back

=item Returns

 Nothing if successful, throws an error on failure.

=item Examples

 # Close all open file handles on a file

 $fileops_obj->close(file_name => $file_name)

 # Close a single open file handle on a file

 $file_ops_obj->close(file_handle => $fh_obj)

=back

=cut

sub close() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {
                type => SCALAR,
                optional => 1,
            },
            file_handle => {
                isa => 'Protocol::ClientFileHandle',
                optional => 1,
            },
            use_syscall => {
                type => SCALAR,
                optional => 1,
                default => 0,
            },
        }
    );

    # Check that either the file_name or file_handle parameter
    # has been specfified

    if (!(exists $args{file_name} xor exists $args{file_handle})) {
        NATE::BaseException->throw(
            "Either the file_name or file_handle parameter must ".
            "be specified."
        );
    }

    # Build the list of file handle object(s) to operate on

    my @fh_obj_list;
    my $file_name;
    if (exists $args{file_name}) {
        $file_name = $args{file_name};
        @fh_obj_list = @{$self->{file_mapping}{$file_name}};
    } else {
        push @fh_obj_list, $args{file_handle};
        $file_name = $args{file_handle}->{file_name};
    }

    # Remotely execute the close command(s)

    foreach my $fh_obj (@fh_obj_list) {
        my %close_args = (file_handle => $fh_obj->{file_handle});
        if ($args{use_syscall}) {
            $close_args{use_syscall} = 1;
        }
        $self->{executor}->execute(
            client => $self->{client_obj},
            package => $self->{platform_pkg},
            method => 'close',
            args => \%close_args,
        );

        # Cleanup the bookkeeping by removing the closed
        # file handle from the list

        @{$self->{file_mapping}{$file_name}} = grep {
            ! ($_->{file_handle} eq $fh_obj->{file_handle})
        } @{$self->{file_mapping}{$file_name}};

    }

    # If all open file handles are now closed, delete the
    # file handle mapping entry for the file

    if (!@{$self->{file_mapping}{$file_name}}) {
        delete $self->{file_mapping}{$file_name};
    }

    return;

}

=for html <hr>

=head2 rename

Renames a file.

=over

=item Options

=over

=item file_name => $file_name

 (Required) The fully specified name of the file being renamed

=item new_file_name => $new_file_name

 (Required) The new file name

=back

=item Returns

 Nothing if successful, throws an error on failure.

=item Example

 $fileops_obj->rename(
     file_name => $file_name,
     new_file_name => $new_file_name
 );

=back

=cut

sub rename() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
            new_file_name => {type => SCALAR},
        }
    );

    # Remotely execute the rename command

    $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'rename',
        args => {
            file_name => $args{file_name},
            new_file_name => $args{new_file_name},
        },
    );

    # Update the bookkeeping?

    return;

}

=for html <hr>

=head2 move

Moves a file.

=over

=item Options

=over

=item file_name => $file_name

 (Required) The fully specified name of the file being moved

=item new_path => $new_path

 (Required) The new path name for the file

=back

=item Returns

 Nothing if successful, throws an error on failure.

=item Example

 $fileops_obj->move(
     file_name => $file_name,
     new_path => $new_path
 );

=back

=cut

sub move() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
            new_path => {type => SCALAR},
        }
    );

    # Remotely execute the move command

    $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'move',
        args => {
            file_name => $args{file_name},
            new_path => $args{new_path},
        },
    );

    # Update the bookkeeping?

    return;

}

=for html <hr>

=head2 delete

Deletes a file.

=over

=item Options

=over

=item file_name => $file_name

 (Required) The fully specified name of the file being deleted

=back

=item Returns

 Nothing if successful, throws an error on failure.

=item Example

 $fileops_obj->delete(
     file_name => $file_name
 );

=back

=cut

sub delete() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
        }
    );

    # Remotely execute the delete command

    $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'delete',
        args => {
            file_name => $args{file_name},
        },
    );

    # Update the bookkeeping?

    return;

}

=for html <hr>

=head2 link

Creates a link to a file.

=over

=item Options

=over

=item file_name => $file_name

 (Required) The fully specified name of the file to create the link to

=item link_name => $link_name

 (Required) The name of the link to create

=back

=item Returns

 Nothing if successful, throws an error on failure.

=item Example

 $fileops_obj->link(
     file_name => $file_name,
     link_name => $link_name
 );

=back

=cut

sub link() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
            link_name => {type => SCALAR},
        }
    );

    # Remotely execute the link command

    $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'link',
        args => {
            file_name => $args{file_name},
            link_name => $args{link_name},
        },
    );

    return;

}

=for html <hr>

=head2 symlink

Creates a symbolic link to a file.

=over

=item Options

=over

=item file_name => $file_name

 (Required) The fully specified name of the file to symbolically link to

=item link_name => $link_name

 (Required) The name of the symblic link to create

=back

=item Returns

 Nothing if successful, throws an error on failure.

=item Example

 $fileops_obj->symlink(
     file_name => $file_name,
     link_name => $link_name
 );

=back

=cut

sub symlink() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
            link_name => {type => SCALAR},
        }
    );

    # Remotely execute the symlink command

    $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'symlink',
        args => {
            file_name => $args{file_name},
            link_name => $args{link_name},
        }
    );

    return;

}

=for html <hr>

=head2 unlink

Removes a file link.

=over

=item Options

=over

=item link_name => $link_name

 (Required) The name of the link to remove

=back

=item Returns

 Nothing if successful, throws an error on failure.

=item Example

 $fileops_obj->unlink(
     link_name => $link_name
 );

=back

=cut

sub unlink() {

    my $self = shift;

    # Validate the input parameters

    my %args = Params::Validate::validate(
        @_,
        {
            link_name => {type => SCALAR},
        }
    );

    # Remotely execute the unlink command

    $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'unlink',
        args => {
            link_name => $args{link_name},
        },
    );

    return;

}

=for html <hr>

=head2 checksum

Compute and return a checksum for the file

=over

=item Options

=over

=item file_name => $file_name

 (Required) The full path of the file to compute the checksum for

=back

=item Returns

 Checksum if successful, throws an error on failure.

=item Example

 $fileops_obj->checksum(
    file_name => $file_name,
 );

=back

=cut

sub checksum() {
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
        }
    );

    return $self->{executor}->execute(
        client => $self->{client_obj},
        package => 'Protocol::AgnosticClientFile',
        method => 'checksum',
        args => {
            file_name => $args{file_name},
        },
    );
}

=for html <hr>

=head2 exists

Check if a file exists

=over

=item Options

=over

=item file_name => $file_name

 (Required) The full path of the file to be checked

=back

=item Returns

  1 if the file is found
  0 if the file is not found

=item Example

 if ($fileops_obj->exists(file_name => $file_name)) {
    # file found
    $Test->nlog("File $file_name found!");
 } else {
    # file not found
    $Test->nlog("File $file_name not found!");
 }

=back

=cut

sub exists() {
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
        }
    );

    return $self->{executor}->execute(
        client => $self->{client_obj},
        package => 'Protocol::AgnosticClientFile',
        method => 'exists',
        args => {
            file_name => $args{file_name},
        },
    );
}

=for html <hr>

=head2 size

Return a file's size

=over

=item Options

=over

=item file_name => $file_name

 (Required) The full path of the file to be checked

=back

=item Returns

  SCALAR (size of the file in bytes)

=item Example

 my $file_size = $fileops_obj->size(file_name => $file_name);
 $Test->nlog("The size of the file $file_name is $file_size bytes");

=back

=cut

sub size() {
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
        }
    );

    return $self->{executor}->execute(
        client => $self->{client_obj},
        package => 'Protocol::AgnosticClientFile',
        method => 'size',
        args => {
            file_name => $args{file_name},
        },
    );
}

=for html stat

=head2 size

Return a file's stats

=over

=item Options

=over

=item file_name => $file_name

 (Required) The full path of the file to be checked

=item Returns

ARRAYREF An array ref to an array of statistics about the file whose path is provided.

Statistics in order:
0   Device number of file system
1   Inode number
2   File mode (type and permissions)
3   Number of (hard) links to the file
4   Numeric user ID of file.s owner
5   Numeric group ID of file.s owner
6   The device identifier (special files only)
7   File size, in bytes
8   Last access time since the epoch
9   Last modify time since the epoch
10  Inode change time (not creation time!) since the epoch
11  Preferred block size for file system I/O
12  Actual number of blocks allocated

=item Example

my $stat_ref = $fileops_obj->stat(file_name => $filename);
my @file_stats = @{$stat_ref};
for my $stats (@file_stats){
...
}

=back

=back

=cut

sub stat() {
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
        }
    );

    return $self->{executor}->execute(
        client => $self->{client_obj},
        package => 'Protocol::AgnosticClientFile',
        method => 'stat',
        args => {
            file_name => $args{file_name},
        },
    );
}

=for html list_dir

=head2 list_dir

Return a the contents of a directory

=over

=item Options

=over

=item dir_path => $dir_path

 (Required) The full path of the directory to be listed

=back

=item Returns

  ARRAYREF (reference to a list of files listed in the directory.)

=item Example

 my $list_ref = $fileops_obj->list_dir( dir_path => $dir_path );
 my @list = @{$list_ref};
 foreach my $f (@list) {
   $Test->nlog ("\$f = $f\n");
   }

=back

=cut

sub list_dir() {
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            dir_path => {type => SCALAR},
        }
    );

    return $self->{executor}->execute(
        client => $self->{client_obj},
        package => 'Protocol::AgnosticClientFile',
        method => 'list_dir',
        args => {
            dir_path => $args{dir_path},
        },
    );
}

=for html chdir

=head2 chdir

Change the working directory to the provided path

=over

=item Options

=over

=item dir_path => $dir_path

 (Required) The full path of the directory to be made as working directory

=back

=item Returns

Nothing, if successful.
Throws exception, if error occurs.

=item Example

 $fileops_obj->chdir( dir_path => $dir_path );

=back

=cut

sub chdir() {
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            dir_path => {type => SCALAR},
        }
    );

    return $self->{executor}->execute(
        client => $self->{client_obj},
        package => 'Protocol::AgnosticClientFile',
        method => 'chdir',
        args => {
            dir_path => $args{dir_path},
        },
    );
}

=head2 mkdir

Creates a directory on a client.

=over

=item Options

=over

=item directory_name => $dir_name

 (Required) The fully specified name of the directory to be created

=item permissions => $permissions

 (Optional) The permissions (4 bits) for the directory being created
 This is ignored for Windows.

=back

=item Returns

 Nothing.  Throws an error on failure.

=item Example

 $fileops_obj->mkdir(directory_name => $dir_name);

=back

=cut

sub mkdir() {
    my $self = shift;

    # Validate the input parameters
    my %args = Params::Validate::validate(
        @_,
        {
            directory_name => { type => SCALAR },
            permissions => {
                optional => 1,
                type => SCALAR,

            },
        }
    );

    my %exec_args = (
        directory_name => $args{directory_name},
    );

    if ($self->{client_type} eq 'linux') {
        if (exists $args{permissions}) {
            $exec_args{permissions} = $args{permissions};
        }
    }

    # Issue the mkdir command on the remote client
    $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'mkdir',
        args => \%exec_args,
    );

    return;
}

=for html utime 

=head2 utime

Modify the last access and modification timestamps on a file

=over

=item Options

=over

=item file_name => $file_name

 (Required) The full path of the filename to modify the timestamps on

=item mtime => $mtime

 (Optional) The new modification timestamp for the file (defaults to the
            current time)

=item atime => $atime

 (Optional) The new access timestamp for the file (defaults to the
            current time)

=back

=item Returns

Nothing, if successful.
Throws exception, if error occurs.

=item Example

 $fileops_obj->utime(file_name => $file_name);

=back

=cut

sub utime() {
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
            mtime     => {
                type     => SCALAR,
                optional => 1,
                default  => undef,
            },
            atime     => {
                type     => SCALAR,
                optional => 1,
                default  => undef,
            },
        }
    );

    my %exec_args = (
        file_name => $args{file_name},
    );
    if (defined $args{mtime}) {
        $exec_args{mtime} = $args{mtime};
    }
    if (defined $args{atime}) {
        $exec_args{atime} = $args{atime};
    }

    return $self->{executor}->execute(
        client  => $self->{client_obj},
        package => 'Protocol::AgnosticClientFile',
        method  => 'utime',
        args    => \%exec_args,
    );
}

=for html <hr>

=head2 rmdir

Removes a directory on a client.

=over

=item Options

=over

=item directory_name => $dir_name

 (Required) The fully specified name of the directory to be removed

=back

=item Returns

 Nothing.  Throws an error on failure.

=item Example

 $fileops_obj->rmdir(directory_name => $dir_name);

=back

=cut

sub rmdir() {
    my $self = shift;

    # Validate the input parameters
    my %args = Params::Validate::validate(
        @_,
        {
            directory_name => { type => SCALAR },
        }
    );

    my %exec_args = (
        directory_name => $args{directory_name},
    );

    # Issue the rmdir command on the remote client
    $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'rmdir',
        args => \%exec_args,
    );

    return;
}

=for html <hr>

=head2 list_streams

Get list of the streams associated with a file.  This is not supported for Unix.

=over

=item Options

=over

=item file_name => $file_name

 (Required) The fully specified name of the file for which the stream names
 are to be returned.

=back

=item Returns

 A list of the stream names.  Throws an error on failure.

=item Example

 my @streams = $fileops_obj->list_streams(file_name => $file_name);

=back

=cut

sub list_streams() {
    my $self = shift;

    # Validate the input parameters
    my %args = Params::Validate::validate(
        @_,
        {
            file_name => {type => SCALAR},
        }
    );

    # Remotely execute the command
    my $result = $self->{executor}->execute(
        client => $self->{client_obj},
        package => $self->{platform_pkg},
        method => 'list_streams',
        args => {
            file_name => $args{file_name},
        }
    );

    return @$result;
}

=for html <hr>

=head2 client

Returns the client object associated with this instance.

=over

=item Returns

 The NACL::C::Client object associated with this instance.

=item Example

 my $client = $fileops_obj->client();

=back

=cut

sub client() {
    my $self = shift;
    my %args = Params::Validate::validate(@_, {});

    return $self->{client_obj};
}

=for html <hr>

=head2 cifs_notify_start

Start a CIFS watch on a particular directory.

=over

=item Options

=over

=item path_name => $path_name

 (Required) The fully specified name of the directory.

=back

=item Returns

 A handle to a watch object that is used in subsequent CIFS watch calls.

=item Example

 my $watch_handle = $fileops_obj->cifs_notify_start(path_name => $path_name);

=back

=cut

sub cifs_notify_start() {
    my $self = shift;

    # Validate the input parameters
    my %args = Params::Validate::validate(
        @_,
        {
            path_name => {type => SCALAR},
            watch_subtree => {type => SCALAR},
            filter => {type => SCALAR},
        }
    );

    # Remotely execute the command
    my $result = $self->{executor}->execute(
        client => $self->{client_obj},
        package => 'Protocol::CIFS::ChangeNotify',
        method => 'cifsNotifyStart',
        args => {
            sPath => $args{path_name},
            bWatchSubtree => $args{watch_subtree},
            uFilter => $args{filter},
        }
    );

    return $result;
}

=for html <hr>

=head2 cifs_notify_stop

Stop a CIFS watch on a particular directory.

=over

=item Options

=over

=item notify_handle => $notify_handle

 (Required) Handle to the watch object.

=back

=item Returns

 1 on success, 0 on failure.

=item Example

 my $result = $fileops_obj->cifs_notify_stop(notify_handle => $notify_handle);

=back

=cut

sub cifs_notify_stop() {
    my $self = shift;

    # Validate the input parameters
    my %args = Params::Validate::validate(
        @_,
        {
            notify_handle => {type => HASHREF},
        }
    );

    # Remotely execute the command
    my $result = $self->{executor}->execute(
        client => $self->{client_obj},
        package => 'Protocol::CIFS::ChangeNotify',
        method => 'cifsNotifyStop',
        args => {
            hNotify => $args{notify_handle},
        }
    );

    return $result;
}

=for html <hr>

=head2 cifs_notify_details

Retrieve CIFS watch updates on a particular directory.

=over

=item Options

=over

=item notify_handle => $notify_handle

 (Required) Handle to the watch object.

=back

=item Returns

 Array of CIFS watch updates.

=item Example

 my @results = $fileops_obj->cifs_notify_details(notify_handle => $notify_handle);

=back

=cut

sub cifs_notify_details() {
    my $self = shift;

    # Validate the input parameters
    my %args = Params::Validate::validate(
        @_,
        {
            notify_handle => {type => HASHREF},
        }
    );

    # Remotely execute the command
    my $result = $self->{executor}->execute(
        client => $self->{client_obj},
        package => 'Protocol::CIFS::ChangeNotify',
        method => 'cifsNotifyDetails',
        args => {
            hNotify => $args{notify_handle},
        }
    );

    my @results = split /\0/,$result;
    return @results;
}

# Private Functions

sub _translate_access_mode() {

    my ($client_type, $access_mode) = @_;
    my %access_mode_values;

    if ($client_type eq 'windows') {
        %access_mode_values = (
            none => '0',
            read => 'GENERIC_READ',
            write => 'GENERIC_WRITE',
            delete => 'DELETE',
            'read-delete' => 'GENERIC_READ|DELETE',
            'read-write' => 'GENERIC_READ|GENERIC_WRITE',
            'write-delete' => 'GENERIC_WRITE|DELETE',
            'read-write-delete' => 'GENERIC_READ|GENERIC_WRITE|DELETE',
        );
    } elsif ($client_type eq 'linux') {
        %access_mode_values = (
            read => 'O_RDONLY',
            write => 'O_WRONLY',
            'read-write' => 'O_RDWR',
        );
    } else {
        NATE::BaseException->throw(
           "Unsupported client type ($client_type)."
        );
    }

    if (exists $access_mode_values{$access_mode}) {
        return $access_mode_values{$access_mode};
    } else {
        return $access_mode;
    }

}

sub _translate_open_mode() {

    my ($client_type, $open_mode) = @_;
    my %open_mode_values;

    if ($client_type eq 'windows') {
        %open_mode_values = (
            'open-existing' => 'OPEN_EXISTING',
            'open-always' => 'OPEN_ALWAYS',
            'create-always' => 'CREATE_ALWAYS',
            'create-new' => 'CREATE_NEW',
            'open-append' => '',
            'truncate-existing' => 'TRUNCATE_EXISTING'
        );
    } elsif ($client_type eq 'linux') {
        %open_mode_values = (
            'open-existing' => '',
            'open-always' => 'O_CREAT',
            'create-always' => 'O_CREAT',
            'create-new' => 'O_CREAT|O_EXCL',
            'open-append' => 'O_APPEND',
            'truncate-existing' => 'O_TRUNC'
        );
    } else {
        NATE::BaseException->throw(
           "Unsupported client type ($client_type)."
        );
    }

    if (exists $open_mode_values{$open_mode}) {
        return $open_mode_values{$open_mode};
    } else {
        return $open_mode;
    }

}

sub _translate_share_mode() {

    my ($client_type, $share_mode) = @_;
    my %share_mode_values;

    if ($client_type eq 'windows') {
        %share_mode_values = (
            'none' => 0,
            'shared-delete' => 'FILE_SHARE_DELETE',
            'shared-read' => 'FILE_SHARE_READ',
            'shared-write' => 'FILE_SHARE_WRITE',
            'shared-read-delete' => 'FILE_SHARE_READ|FILE_SHARE_DELETE',
            'shared-write-delete' => 'FILE_SHARE_WRITE|FILE_SHARE_DELETE',
            'shared-read-write' => 'FILE_SHARE_READ|FILE_SHARE_WRITE',
            'shared-read-write-delete' =>
                'FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE',
        );
    } elsif ($client_type eq 'linux') {
        %share_mode_values = ();
    } else {
        NATE::BaseException->throw(
           "Unsupported client type ($client_type)."
        );
    }

    if (exists $share_mode_values{$share_mode}) {
        return $share_mode_values{$share_mode};
    } else {
        return $share_mode;
    }

}

=head1 SEE ALSO

=over

=item Protocol::ClientFileHandle

=item Protocol::WindowsClientFile

=item Protocol::LinuxClientFile

=item Protocol::Executor

=back

=head1 STATUS

Shared

=head1 AUTHOR

WAFL Protocol QA (dl-wafl-proto-qa@netapp.com)

=head1 COPYRIGHT

Copyright 2011, Network Appliance Inc., All Rights Reserved.

=cut

1;
