# $Id: //depot/prod/test/main/lib/Protocol/Executor.pm#1 $

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Execution handler for file ops on remote clients
## @author dl-wafl-proto-qa@netapp.com
## @status shared

=pod

=head1 NAME

Protocol::Executor

=head1 DESCRIPTION

This library is entirely an internal library, meant for
a very limited purpose of serving the Protocol file ops
libraries. Hence there is no POD exposed.
This module uses NACL::STask::Client::RPC::Worker under the covers.

=cut

############################################################################
#  Executor
############################################################################

package Protocol::Executor;

use strict;
use warnings;

use NACL::STask::Client::RPC::Worker;
use Params::Validate qw(validate SCALAR BOOLEAN OBJECT ARRAYREF HASHREF);

# methods

#------------------------------------------------------------------------------
#
# new()
#
# Get a new object of the Executor:
#
# Inputs  : None
#
# Returns : A new object of this class
#
# Example :
#       my $executor = $executor->new();
#
#------------------------------------------------------------------------------

sub new
{
    my $class = shift;
    my $self = {};
    %{$self->{clients}} = ();

    bless($self, $class);
    return $self;
}

#------------------------------------------------------------------------------
#
# execute()
#
# Executes a method on a client using NACL::STask::Client::RPC::Worker.
# Creates a NACL::STask::Client::RPC::Worker object per client,
# and caches it for multiple calls.
#
# Inputs  :
#        client  => A client object (NACL::C::Client)
#
#        package => The package containing the method to execute
#
#        method  => The method in the above package to exeute
#
#        args    => Reference to a hash of args to pass to the above method
#
# Returns : SCALAR, The output of the executed method
#
# Example :
#       my %file_args = (
#                           file_handle => $fh1,
#                           offset      => $offset2,
#                           num_bytes   => length($contents));
#                       );
#
#       my $file_data = $executor->execute(client    => $client1,
#                                          package   => 'Protocol::LinuxClientFile',
#                                          method    => "read",
#                                          args      => \%file_args
#                                         );
#
#------------------------------------------------------------------------------

sub execute
{
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            client  => { isa  => 'NACL::C::Client' },
            package => { type => SCALAR },
            method  => { type => SCALAR },
            args    => { type => HASHREF, optional => 1 },
        }
    );

    # grab args and clean up the %args hash
    my $client  = $args{client};
    delete $args{client};
    my $package = $args{package};
    delete $args{package};
    my $method  = $args{method};
    delete $args{method};
    my @method_args = (defined $args{args}) ? %{$args{args}} : ();

    # get a NACL::STask::Client::RPC::Worker object for this client
    my $client_rpc = $self->_get_client_rpc(client => $client, package => $package);

    # execute the call
    my $output = $client_rpc->call(
                                package     => $package,
                                method      => $method,
                                args        => \@method_args,
                            );


    return $output;

}

=for html <hr>

=head2 encode

Converts NUL characters in a string into something printable.

=over

=item Options

=over

=item data (positional)

 (Required) The string to encode

=back

=item Returns

 The input string with all NUL characters encoded as printable.

=item Example

 $encoded_data = Executor::encode(data => $data);

=back

=cut

sub encode {
    Params::Validate::validate_pos(@_, { type => SCALAR });
    my $data = shift;

    # convert each NUL into escaped #0 (and # into ##)
    $data =~ s/[#\x00]/$& eq '#' ? '##' : '#0'/eg;

    return $data;
}

=for html <hr>

=head2 decode

Converts encoded characters in a string from a call to encode back to their
original form.

=over

=item Options

=over

=item data (positional)

 (Required) The string to decode

=back

=item Returns

 The input string with all encoded characters restored.

=item Example

 $data = Executor::decode(data => $encoded_data);

=back

=cut

sub decode {
    Params::Validate::validate_pos(@_, { type => SCALAR });
    my $data = shift;

    # convert each escaped pair back to its proper character
    $data =~ s/#([0#])/$1 eq '#' ? '#' : chr(0)/eg;

    return $data;
}

#------------------------------------------------------------------------------
#
# _get_client_rpc()
#
# INTERNAL ONLY, gets an object of NACL::STask::Client::RPC::Worker for a client.
# Creates a new one if the object does not eist in the cache. Also loads packages
# as part of creating the NACL::STask::Client::RPC::Worker object.
#
# Inputs  : client : A client object (NACL::C::Client)
#
# Returns : An object of type NACL::STask::Client::RPC::Worker
#
#------------------------------------------------------------------------------

sub _get_client_rpc
{
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            client  => { isa  => 'NACL::C::Client' },
            package => { type => SCALAR , optional => 1 },
        }
    );

    # pull details from the NACL::C::Client object
    my $client_obj  = $args{client};
    my $client_name = $client_obj->{client};

    # get a NACL::STask::Client::RPC::Worker object
    if (!$self->_client_exists(client => $client_name)) {
        # cache RPC objects per client
        $self->{clients}{$client_name}{rpc} =
                    NACL::STask::Client::RPC::Worker->new(client => $client_obj);
        $self->{clients}{$client_name}{object} = $client_obj;
        $self->{clients}{$client_name}{rpc}->start();
    }

    # load the required libs
    if (defined $args{package}) {
        my $package = $args{package};

        # first check if the lib was already loaded
        if (!$self->_is_package_loaded(client => $client_obj, package => $package)) {

            $self->{clients}{$client_name}{rpc}->call( eval => "use Tharn qw(param \$Log);" );
            $self->{clients}{$client_name}{rpc}->call( eval => "use $package" );

            # package is loaded, cache it!
            $self->{clients}{$client_name}{packages}{$package} = 1;
        }
    }

    # return from cache
    return $self->{clients}{$client_name}{rpc};
}

#------------------------------------------------------------------------------
#
# _is_package_loaded()
#
# INTERNAL ONLY, checks if the requested package was loaded on the client.
#
# Inputs  : client : A client object (NACL::C::Client)
#
# Returns : BOOLEAN (1 for true, or 0 for false)
#
#------------------------------------------------------------------------------

sub _is_package_loaded
{
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            client  => { isa  => 'NACL::C::Client' },
            package => { type => SCALAR},
        }
    );

    my $client_name = $args{client}->{client};
    my $package     = $args{package};

    return 0    if (!defined $self->{clients}{$client_name}{packages}{$package});
    return 1;
}

#------------------------------------------------------------------------------
#
# _client_exists()
#
# INTERNAL ONLY, checks if the requested client exists in our cache.
#
# Inputs  : client : the client's name (SCALAR)
#
# Returns : BOOLEAN (1 for true, or 0 for false)
#
#------------------------------------------------------------------------------

sub _client_exists
{
    my $self = shift;
    my %args = Params::Validate::validate(
        @_,
        {
            client  => { type => SCALAR },
        }
    );

    return 0    if (!defined $args{client});

    # first check if we have any clients in our cache at all
    my %clients = %{$self->{clients}};
    return 0    if (scalar keys %clients < 1);

    # check if the client exists in our cache
    my $client_name = $args{client};
    return 1    if (defined $clients{$client_name});
    return 0; # false
}

#------------------------------------------------------------------------------
#
# _destroy()
#
# Cleans up NACL::STask::Client::RPC::Worker objects for all clients.
# Wipes out the local cache of client data such as, client names,
# NACL::C::Client and NACL::STask::Client::RPC::Worker objects
#
# Inputs  : None
#
# Returns : None
#
#------------------------------------------------------------------------------

sub _destroy
{
    my $self = shift;

    # cleanup NACL::STask::Client::RPC::Worker objects per client
    my %clients = %{$self->{clients}};
    foreach my $client_name (keys %clients) {
        if (defined $clients{$client_name}{rpc}) {
            # destroy the NACL::STask::Client::RPC::Worker objects
            $clients{$client_name}{rpc}->kill()
                    if (defined $clients{$client_name}{rpc});
            # delete the stale NACL::STask::Client::RPC::Worker object from the cache
            delete $self->{clients}{$client_name}{rpc};
        }
    }

    # clean up the cache (empty the cache rather than deleting the hash key)
    %{$self->{clients}} = ();

    return;
}



=head1 SEE ALSO

=over

=item Protocol::ClientFileOperations

=item Protocol::ClientFilehandle

=item Protocol::LinuxClientFile

=item Protocol::WindowsClientFile

=back

=head1 STATUS

    Shared

=head1 AUTHOR

    WAFL Protocol QA (dl-wafl-proto-qa@netapp.com)

=head1 COPYRIGHT

    Copyright 2011, Network Appliance Inc., All Rights Reserved.

=cut


1;
