# $Id: //depot/prod/test/main/lib/Protocol/LinuxClientFile.pm#1 $

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Protocol Client File Module for Linux
## @author dl-wafl-proto-qa@netapp.com
## @status shared
## @pod here

=pod

=head1 NAME

Protocol::LinuxClientFile

=head1 DESCRIPTION

LinuxClientFile provides API's to manipulate files on a Linux client.
This class SHOULD NOT be called directly. It's methods are called by the
Protocol::ClientFile and Protocol::ClientFilehandle packages.

=head1 METHODS (not public)

=cut

############################################################################
#  LinuxFileOperations
############################################################################

package Protocol::LinuxClientFile;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR GLOB OBJECT ARRAYREF);
use Fcntl;
use File::FcntlLock;
use threads 'exit' => 'threads_only';
use NATE::BaseException;
use Tharn qw($Log);

use Protocol::Executor;
use constant DUMMY_ARG => 'dummy_arg';

#---------------------------------------------------------------------------
# internal bookkeeping for file_handles because they cannot be passed back
# and forth between different processes and clients (e.g. - subtest).
# also saving thread objects for waiting lock requests
#---------------------------------------------------------------------------
# file_handle map
my %Files;
my $Thread_Count = 0;

=head2 open()

Open (or create) a file

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be opened

=item access_mode

 (Required) The desired access mode for the file being opened
 Sample values are:
    O_RDONLY  => open in Read mode
    O_WRONLY  => open in Write mode
    O_RDWR    => open in Read-Write mode

=item open_flags

 (Required) The desired open mode for the file being opened
 Sample values are:
    <no_flag>       => open-existing
    O_CREAT         => open-always / create-always
    O_CREAT|O_EXCL  => create-new
    O_APPEND        => open-append
    O_TRUNC         => truncate-existing

=item permissions

 (Optional) The Unix permissions (4-bits) for the file being opened (Default: 0777)

=item Returns

 SCALAR: file handle returned by the OS, or throw an eror

=item Example

 my $file_handle = open(file_name   => $file_name,
                        access_mode => O_RDWR,
                        open_flags  => O_CREAT,
                        permissions => 0777);
 if ($@) {
     # print "open file failed: $@\n";
 }

=back

 Notes : Calls Perl's sysopen()

=back

=cut

sub open
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($legal,
        $file_name,
        $file_handle,
        $flags,
        $permissions,
        @inputs
       );

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
            access_mode => {
                type => SCALAR,
            },
            open_flags => {
                type => SCALAR,
            },
            permissions => {
                type     => SCALAR,
                optional => 1,
                default  => '0777',
                regex    => qr/^\d+$/,
            },
        }
    );

    $file_name = $args{file_name};

    # translate flags to integer equivalent and OR together
    $flags |= eval "$args{access_mode}";

    # append to open flags mask
    $flags |= eval "$args{open_flags}" if $args{open_flags};

    # set the umask to '0' to allow user permissions
    umask(0);

    # convert the permissions string to OCTAL (if required)
    if (defined $args{permissions} and $args{permissions} ne "") {
        if ($args{permissions} =~ /^0/) {
            # convert perms to octal
            $permissions = oct $args{permissions};
        } else {
            # assume perms are already octal
            $permissions = $args{permissions};
        }
    }

    # open the file
    sysopen($file_handle, $file_name, $flags, $permissions)
            or NATE::BaseException->throw("failed to open file: $file_name: $!");

    # save the OS-returned file handle and return the
    # latest index into our FH DB, with a "LINUX" prefix
    # to use as a user reference to the OS file handle
    return _add_file_handle(file_handle => $file_handle,
                            file_name   => $file_name);
}

=for html <hr>

=head2 close()

Close a file handle

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle of the file to be closed (returned by open())

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 close(file_handle => $file_handle);  # $file_handle returned by open()
 if ($@) {
     $Test->nlog("close file failed: $@\n");
 }

=back

 Notes : Calls Perl's close()

=back

=cut

sub close
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my $file_handle;

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
	    use_syscall => {
                type => SCALAR,
		default => 0,
		optional => 1,
            },
        }
    );

    # get the file_handle from the global hash
    $file_handle = _get_file_handle(file_handle => $args{file_handle});

    # close the file
    if ($args{use_syscall}) {
	require("sys/syscall.ph");
	$! = 0;
	my $file_no = fileno($file_handle);
	if (syscall(&SYS_close,$file_no) == -1) {
	    if ($!) {
		my $err_msg = "failed to close file handle: $file_no" .
			      " with error number: $!";
		NATE::BaseException->throw($err_msg);
	    }
	}
    } else {
	close($file_handle) or
		    NATE::BaseException->throw("failed to close file: $!");

    }

    # delete the file_handle from our records
    _delete_file_handle(file_handle => $args{file_handle});

    return 0;
}

=for html <hr>

=head2 rename()

Rename a file to a new name

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be renamed

=item new_file_name

 (Required) The new file name

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 rename(file_name     => $file_name,
        new_file_name => $new_file_name);
 if ($@) {
     $Test->nlog("rename file failed: $@\n");
 }

=back

 Notes : Calls Perl's rename()

=back

=cut

sub rename
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_name, $new_file_name);

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
            new_file_name => {
                type => SCALAR,
            },
        }
    );

    # Since this is a call to a local method, and since the NACL Worker STask
    # likes to call methods as ->method_name(), we need to pass this DUMMY_ARG
    # and let the target method shift it off the list
    return &move(DUMMY_ARG,
                 file_name => $args{file_name},
                 new_path  => $args{new_file_name});
}

=for html <hr>

=head2 move()

Move a file to a new location

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be moved

=item new_path

 (Required) The new file path

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 move(file_name => $file_name,
      new_path  => $new_file_path);
 if ($@) {
     $Test->nlog("move file failed: $@\n");
 }

=back

 Notes : Calls Perl's rename()

=back

=cut

sub move
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_name, $new_file_path);

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
            new_path => {
                type => SCALAR,
            },
        }
    );

    $file_name     = $args{file_name};
    $new_file_path = $args{new_path};

    CORE::rename($file_name, $new_file_path)
            or NATE::BaseException->throw(
                "failed to move/rename file $file_name to $new_file_path: $!");

    return 0;
}
=for html <hr>

=head2 delete()

Delete a file

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be deleted

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 delete(file_name => $file_name);
 if ($@) {
     $Test->nlog("delete file failed: $@\n");
 }

=back

 Notes : Calls Perl's unlink()

=back

=cut

sub delete
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my $file_name;

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
        }
    );

    $file_name = $args{file_name};

    unlink $file_name
            or NATE::BaseException->throw(
                "failed to delete file $file_name: $!");

    return 0;
}

=for html <hr>

=head2 link()

Create a link to a file

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be linked

=item link_name

 (Required) The link name to be created

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 link(file_name => $file_name,
      link_name => $link_name);
 if ($@) {
     $Test->nlog("link file failed: $@\n");
 }

=back

 Notes : Calls Perl's link()

=back

=cut

sub link
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_name, $link_name);

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
            link_name => {
                type => SCALAR,
            },
        }
    );

    $file_name = $args{file_name};
    $link_name = $args{link_name};

    link($file_name, $link_name)
            or NATE::BaseException->throw(
                "failed to link file $file_name to link $link_name: $!");

    return 0;
}

=for html <hr>

=head2 symlink()

Create a symlink to a file

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be linked

=item link_name

 (Required) The link name to be created

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 symlink(file_name => $file_name,
         link_name => $link_name);
 if ($@) {
     $Test->nlog("symlink file failed: $@\n");
 }

=back

 Notes : Calls Perl's symlink()

=back

=cut

sub symlink
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_name, $link_name);

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
            link_name => {
                type => SCALAR,
            },
        }
    );

    $file_name = $args{file_name};
    $link_name = $args{link_name};

    symlink($file_name, $link_name)
            or NATE::BaseException->throw(
                "failed to symlink file $file_name to link $link_name: $!");

    return 0;
}

=for html <hr>

=head2 unlink()

Remove a file link

=over 4

=item Options

=over

=item link_name

 (Required) The link to be deleted

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 unlink(link_name => $link_name);
 if ($@) {
     $Test->nlog("unlink failed: $@\n");
 }

=back

 Notes : Calls Perl's unlink()

=back

=cut

sub unlink
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my $link_name;

    # validate inputs
    %args = validate (
        @_,
        {
            link_name => {
                type => SCALAR,
            },
        }
    );

    $link_name = $args{link_name};

    unlink $link_name
            or NATE::BaseException->throw(
                "failed to unlink file link $link_name: $!");

    return 0;
}

=for html <hr>

=head2 mkdir()

Create a directory

=over 4

=item Options

=over

=item directory_name

 (Required) The directory to be created

=item permissions

 (Optional) The Unix permissions (4-bits) for the directory being created
 (Default: 0777)

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 mkdir(directory_name => $dir_name);
 if ($@) {
     $Test->nlog("mkdir failed: $@\n");
 }

=back

 Notes : Calls Perl's mkdir()

=back

=cut

sub mkdir
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;

    # validate inputs
    %args = validate(
        @_,
        {
            directory_name => { type => SCALAR },
            permissions => {
                type => SCALAR,
                default  => 0777,
                regex => qr/^\d+$/,
            },
        }
    );

    my $dir_name = $args{directory_name};

    mkdir($dir_name, $args{permissions})
        or NATE::BaseException->throw(
            "failed to create directory $dir_name: $!");

    return 0;
}

=for html <hr>

=head2 rmdir()

Remove a directory

=over 4

=item Options

=over

=item directory_name

 (Required) The directory to be removed

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 rmdir(directory_name => $dir_name);
 if ($@) {
     $Test->nlog("rmdir failed: $@\n");
 }

=back

 Notes : Calls Perl's rmdir()

=back

=cut

sub rmdir
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;

    # validate inputs
    %args = validate(
        @_,
        {
            directory_name => { type => SCALAR },
        }
    );

    my $dir_name = $args{directory_name};

    rmdir($dir_name)
        or NATE::BaseException->throw(
            "failed to remove directory $dir_name: $!");

    return 0;
}

=for html <hr>

=head2 read()

Read from a file handle

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle to read from

=item offset

 (Required) The offset into the file to begin reading from

=item num_bytes

 (Required) The number of bytes to read from the file

=item Returns

 SCALAR data if successfully read, or throws an error for FAILURE

=item Example

 my $contents = read(file_handle => $file_handle,
                     offset      => 10,
                     num_bytes   => 50);
 if ($@) {
     $Test->nlog("read file failed: $@\n");
 } else {
    $Test->nlog("contents read from the file: $contents\n");
 }

=back

=back

=cut

sub read
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_handle,
        $offset,
        $num_bytes,
        $buffer,
        $num_bytes_read);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            offset => {
                type => SCALAR,
            },
            num_bytes => {
                type => SCALAR,
            },
        }
    );

    # get the file_handle from the global hash
    $file_handle = _get_file_handle(file_handle => $args{file_handle});
    $offset      = $args{offset};
    $num_bytes   = $args{num_bytes};

    # seek to the offset
    sysseek($file_handle, $offset, SEEK_SET)
                        or NATE::BaseException->throw("failed to seek to offset $offset: $!");

    # read
    $num_bytes_read = sysread($file_handle, $buffer, $num_bytes);
    if (!defined $num_bytes_read) {
        NATE::BaseException->throw("failed to read $num_bytes bytes from file: $!");
    }

    # print the actual number of bytes read
    $Log->comment("read: $num_bytes_read bytes read from file\nread: Data:\n$buffer\n");

    return Protocol::Executor::encode($buffer);
}

=for html <hr>

=head2 write()

Write to a file handle

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle to write to

=item offset

 (Required) The offset into the file to begin writing to

=item contents

 (Required) The contents to write to the file

=item Returns

 SCALAR number of bytes that were successfully written,
 or throws an error for FAILURE

=item Example

 my $num_bytes_wrote = write(file_handle => $file_handle,
                             offset      => 10,
                             contents    => $contents);
 if ($@) {
     $Test->nlog("write file failed: $@\n");
 }
 # else, print $num_bytes_wrote

=back

=back

=cut

sub write
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_handle,
        $offset,
        $contents,
        $num_bytes_wrote);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            offset => {
                type => SCALAR,
            },
            contents => {
                type => SCALAR,
            },
        }
    );

    # get the file_handle from the global hash
    $file_handle = _get_file_handle(file_handle => $args{file_handle});
    $offset      = $args{offset};
    $contents    = Protocol::Executor::decode($args{contents});

    # seek to the offset
    sysseek($file_handle, $offset, SEEK_SET)
                        or NATE::BaseException->throw("failed to seek to offset $offset: $!");

    # write
    $num_bytes_wrote = syswrite($file_handle, $contents);
    if (!defined $num_bytes_wrote) {
        NATE::BaseException->throw("failed to write to file: $!");
    }

    return $num_bytes_wrote;
}

=for html <hr>

=head2 lock()

Lock a byte range in a file

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle to lock

=item offset

 (Required) The offset into the file to begin locking from

=item num_bytes

 (Required) The number of bytes to lock

=item lock_type

 (Required) The type of lock to acquire
 Valid values are:
    F_RDLCK       => read (or shared) lock
    F_WRLCK       => write (or exclusive) lock

=item lock_wait

 (Required) The waiting symantics for the lock
 Valid values are:
    F_SETLK       => non-waiting lock
    F_SETLKW      => waiting lock (puts the waiting request in the
                     background and returns control)

=item Returns

 SCALAR or throws an error for FAILURE
     For non-waiting locks, 0 is returned.
     For non-waiting locks, a thread ID is returned.

=item Example

 $thread_id = lock(file_handle  => $file_handle,
                   offset       => 10,
                   num_bytes    => 50,
                   lock_type    => F_WRLCK,
                   lock_wait    => F_SETLKW);
 if ($@) {
     $Test->nlog("lock file failed: $@\n");
 } else {
     # use the returned $thread_id for get_wait_status()
     # and kill_waiting_request()
 }

=back

 Notes : Calls Fcntl::FileLock->lock()

=back

=cut

sub lock
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_handle,
        $offset,
        $num_bytes,
        $lock_type,
        $lock_wait,
        $locker,
        $lock_op);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            offset => {
                type => SCALAR,
            },
            num_bytes => {
                type => SCALAR,
            },
            lock_type => {
                type => SCALAR,
            },
            lock_wait => {
                type => SCALAR,
            },
        }
    );

    # get the file_handle from the global hash
    $file_handle = _get_file_handle(file_handle => $args{file_handle});

    # get offset, and number of bytes to lock
    $offset      = $args{offset};
    $num_bytes   = $args{num_bytes};

    # translate flags to integer equivalent
    $lock_wait  |= eval "$args{lock_wait}";
    $lock_type  |= eval "$args{lock_type}";
    $lock_op     = ($args{lock_type} eq "F_UNLCK") ? "unlock" : "lock";

    # set up the lock structure
    $locker = new File::FcntlLock;
    $locker->l_type($lock_type);        # type = F_WRLCK or F_RDLCK
    $locker->l_whence(SEEK_SET);
    $locker->l_start($offset);
    $locker->l_len($num_bytes);
    $locker->l_pid($$);

    # lock - check if this is a waiting lock request
    if ($args{lock_wait} eq "F_SETLKW") {
        # spawn a thread for this waiting lock request
        my $lock_thread = threads->new(\&_lock_thread, $locker, $file_handle, $lock_wait);

        # save the thread object
        _add_lock_thread(file_handle => $args{file_handle},
                         lock_thread => $lock_thread);

        # return the latest index into our Thread DB to
        # use as a user reference to the lock request
        return $Thread_Count;
    }

    # else, this is a non-waiting lock request
    $locker->lock($file_handle, $lock_wait);     # lock_type = F_SETLK or F_SETLKW

    # check for lock error
    if ($locker->error()) {
        NATE::BaseException->throw("failed to $lock_op $num_bytes bytes in the file:\n"
            . "(" . $locker->lock_errno() . ") " . $locker->error());
    }

    return 0;
}

=for html <hr>

=head2 unlock()

Unlock a byte range lock on a file (uses F_UNLCK)

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle to unlock

=item lock_id

 (Required*) The byte range specified by the lock ID
             * Not required if offset and length are provided

=item offset

 (Required*) The offset into the file to begin unlocking from
             * Not required if lock_id is provided

=item num_bytes

 (Required*) The number of bytes to unlock
             * Not required if lock_id is provided

=item unlock_wait

 (Optional) The waiting symantics for the lock

Valid values are:
    F_SETLK       => (Default) non-waiting unlock request
    F_SETLKW      => waiting unlock request (puts the waiting request
                     in the background and returns control)

=item Returns

 SCALAR or throws an error for FAILURE
     For non-waiting locks, 0 is returned.
     For non-waiting locks, a thread ID is returned.

=item Examples

 unlock(file_handle => $file_handle,
        offset      => 10,
        num_bytes   => 50);
 if ($@) {
     $Test->nlog("unlock file failed: $@\n");
 }

 $thread_id = unlock(file_handle => $file_handle,
                     lock_id     => $lock_id,    # returned by lock()
                     unlock_wait => F_SETLKW);
 if ($@) {
     $Test->nlog("unlock file failed: $@\n");
 } else {
     # use the returned $thread_id for get_wait_status()
     # and kill_waiting_request()
 }

=back

 Notes : Calls Fcntl::FileLock->lock() with F_UNLCK

=back

=cut

sub unlock
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;

    # set the lock_type in args to F_UNLCK
    $args{lock_type} = 'F_UNLCK';

    # convert unlock_wait to lock_wait
    if (defined $args{unlock_wait}) {
        $args{lock_wait} = $args{unlock_wait};
        delete $args{unlock_wait};
    }

    # since all code (except F_UNLCK) is common to
    # lock() and unlock(), call lock() with F_UNLCK
    # Since this is a call to a local method, and since the NACL Worker STask
    # likes to call methods as ->method_name(), we need to pass this DUMMY_ARG
    # and let the target method shift it off the list

    return &lock(DUMMY_ARG, %args);
}

=for html <hr>

=head2 get_wait_status()

Return the status for a waiting lock operation request

=over 4

=item Options

=over

=item wait_request_id

 (Required) The waiting lock request's thread ID, returned by lock()

=item Returns

 SCALAR ("waiting", "exited", "invalid")
     waiting = if the thread is still alive
     exited  = if the thread has exited
     invalid = if the thread object is undef (thread does not exist)

=item Example

 my $wait_lock_status = get_wait_status(
                            file_handle     => $file_handle,
                            wait_request_id => $thread_id);
 $Test->nlog("lock wait status is: $wait_lock_status\n");

=back

=back

=cut

sub get_wait_status
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;

    my ($file_handle,
        $wait_request_id);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            wait_request_id => {
                type     => SCALAR
            },
        }
    );

    # get the file_handle index
    $file_handle = $args{file_handle};

    # get the waiting request's thread index
    $wait_request_id = $args{wait_request_id};

    # get the thread object from our global hash
    my $thread_object = _get_lock_thread(file_handle => $file_handle,
                                         lock_thread => $wait_request_id);

    # check if the thread object exists
    return "invalid"    if (!defined $thread_object);

    # check if the thread is still alive
    # if yes, return TRUE
    return "waiting"    if ($thread_object->is_running());

    # else... the thread seems to have exited
    # so, return FALSE
    return "exited";
}

=for html <hr>

=head2 kill_waiting_request()

Return the status for a waiting lock/unlock operation request

=over 4

=item Options

=over

=item wait_request_id

 (Required) The waiting lock request's thread ID, returned by lock()

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 kill_waiting_request(file_handle     => $file_handle,
                      wait_request_id => $wait_request_id);
 if ($@) {
     $Test->nlog("failed to kill waiting request: $@\n");
 }

=back

=back

=cut

sub kill_waiting_request
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;

    my ($file_handle,
        $wait_request_id);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            wait_request_id => {
                type     => SCALAR
            },
        }
    );

    # get the file_handle index
    $file_handle = $args{file_handle};

    # get the waiting request's thread index
    $wait_request_id = $args{wait_request_id};

    # get the thread object from our global hash
    my $thread_object = _get_lock_thread(file_handle => $file_handle,
                                         lock_thread => $wait_request_id);

    return 0    if (!defined $thread_object);

    # stop the thread
    $thread_object->kill('KILL')
            or NATE::BaseException->throw(
                "failed to kill waiting lock request $wait_request_id: $!");

    #
    if ($thread_object->is_running()) {
        NATE::BaseException->throw("\nwaiting lock thread $wait_request_id did not die!");
    }
    if ($thread_object->is_joinable()) {
        #$thread_object->join();
    }

    return 0;
}

=for html <hr>

=head2 list_streams()

Get list of the streams associated with a file.

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file for which the stream names
 are to be returned.

=item Returns

 Throws an exception as this method is not supported on Linux.

=item Example

 my @streams = list_streams(file_name => $file_name);

=back

=back

=cut

sub list_streams {
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;

    # validate inputs
    %args = validate(
        @_,
        {
            file_name => {
                type => SCALAR,
            },
        }
    );

    throw NATE::BaseException('list_streams is not supported on Linux');
}


#=========================================================================#
#                            INTERNAL METHODS                             #
#=========================================================================#

sub _add_file_handle
{
    my %args   = @_;

    # generate a random number for the FH key
    my $time   = time();
    my $rand   = rand();
    my $rand2  = rand();
    my $fh_key = "LINUX$time$rand$rand2";

    # add the file name to our global hash
    $Files{$fh_key}{file_name}   = $args{file_name};
    # add the OS file handle to our global hash
    $Files{$fh_key}{file_handle} = $args{file_handle};

    return $fh_key;
}

sub _get_file_handle
{
    my %args     = @_;
    my $fh_index = $args{file_handle};

    # return the hash entry for the requested
    # file handle index from our global hash
    return $Files{$fh_index}{file_handle};      # 'undef' if nonexistent
}

sub _delete_file_handle
{
    my %args     = @_;
    my $fh_index = $args{file_handle};

    # delete the hash entry for the requested
    # file handle index from our global hash
    delete $Files{$fh_index};

    return;
}

#--------------------------------------------------------------------
# INTERNAL ONLY:
# this method is for creating a thread for waiting lock requests.
# this method should NOT be called directly by any external caller.
#--------------------------------------------------------------------
sub _lock_thread
{
    # Thread 'cancellation' signal handler
    $SIG{'KILL'} = sub {
                        return;
                   };

    my ($locker, $file_handle, $lock_wait) = @_;

    $locker->lock($file_handle, $lock_wait);     # lock_type = F_SETLK or F_SETLKW
    # check for lock error
    if ($locker->error()) {
        NATE::BaseException->throw("failed to lock the file:\n"
            . "(" . $locker->lock_errno() . ") " . $locker->error());
    }

    return;
}

sub _add_lock_thread
{
    my %args     = @_;
    my $fh_index = $args{file_handle};

    # increment the global counter
    $Thread_Count++;

    # add the lock thread to our global hash, per FH
    $Files{$fh_index}{waiting_requests}{$Thread_Count} = $args{lock_thread};

    return;
}

sub _get_lock_thread
{
    my %args         = @_;
    my $fh_index     = $args{file_handle};
    my $thread_index = $args{lock_thread};

    # return the hash entry for the requested
    # file handle index from our global hash
    return $Files{$fh_index}{waiting_requests}{$thread_index};      # 'undef' if nonexistent
}

sub _delete_lock_thread
{
    my %args         = @_;
    my $fh_index     = $args{file_handle};
    my $thread_index = $args{lock_thread};

    # delete the hash entry for the requested
    # file handle's thread index from our global hash
    delete $Files{$fh_index}{waiting_requests}{$thread_index};

    return;
}



=head1 SEE ALSO

=over

=item Protocol::ClientFileOperations

=item Protocol::ClientFilehandle

=item Protocol::WindowsClientFile

=back

=head1 STATUS

    Shared

=head1 AUTHOR

    WAFL Protocol QA (dl-wafl-proto-qa@netapp.com)

=head1 COPYRIGHT

    Copyright 2011, Network Appliance Inc., All Rights Reserved.

=cut


1;
