# $Id: //depot/prod/test/main/lib/Protocol/WindowsClientFile.pm#1 $
# vim: set filetype=perl tabstop=4 shiftwidth=4 shiftround expandtab:

#
# Copyright (c) 2001-2011 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Protocol Client File Module for Windows
## @author dl-wafl-proto-qa@netapp.com
## @status shared
## @pod here

=pod

=head1 NAME

Protocol::WindowsClientFile

=head1 DESCRIPTION

WindowsClientFile provides API's to manipulate files on a Windows client.
This class SHOULD NOT be called directly. It's methods are called by the
Protocol::ClientFile and Protocol::ClientFilehandle packages.

=head1 METHODS (not public)

=cut

############################################################################
#  WindowsFileOperations
############################################################################

package Protocol::WindowsClientFile;

use strict;
use warnings;

use Params::Validate qw(validate SCALAR GLOB OBJECT ARRAYREF);
use Win32::API;
use Win32API::File qw(
    :Func :Misc :FILE_ :FILE_FLAG_ :FILE_SHARE_ :FILE_ATTRIBUTE_ :GENERIC_ );
use constant KERNEL32 => 'C:\WINDOWS\system32\kernel32.dll';
use threads 'exit' => 'threads_only';
use NATE::BaseException;
use Tharn qw($Log);

use Protocol::AgnosticClientFile;
use Protocol::Executor;
use constant DUMMY_ARG => 'dummy_arg';
use constant DELETE => '65536';

#---------------------------------------------------------------------------
# internal bookkeeping for file_handles because they cannot be passed back
# and forth between different processes and clients (e.g. - subtest).
# also saving thread objects for waiting lock requests
#---------------------------------------------------------------------------
# file_handle map
my %Files;
my $Thread_Count = 0;

# single instances of our Win32::API prototypes
my ($LockObject, $UnlockObject, $CreateFile, $CloseHandle, $BackupRead,
    $BackupSeek);


=head1 Methods (not public)

=head2 open()

Open (or create) a file

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be opened

=item access_mode

 (Required) The desired access mode for the file being opened
 Sample values are:
    GENERIC_READ               => open in Read mode
    GENERIC_WRITE              => open in Write mode
    GENERIC_READ|GENERIC_WRITE => open in Read-Write mode

=item open_flags

 (Required) The desired open mode for the file being opened
            These flags can be combined with "|"

 Sample values are:
    CREATE_ALWAYS     => create-always (overwrites if exists)
    OPEN_EXISTING     => open-existing
    OPEN_ALWAYS       => open-always   (creates if does not exist)
    CREATE_NEW        => create-new
    TRUNCATE_EXISTING => truncate-existing

=item share_mode

(Optional) The desired share mode for the file being opened
           These flags can be combined with "|"

 Sample values are:
    0                 => no sharing (deny delete, read, write)
    FILE_SHARE_DELETE => allow shared delete (and rename)
    FILE_SHARE_READ   => allow shared reading
    FILE_SHARE_WRITE  => allow shared writing

=item attributes

(Optional) The desired flags and attributes for the file being opened
           These flags can be combined with "|"

 Sample values are:
    undef             => [Default] no attributes or additional flags
    FILE_ATTRIBUTE_*  => Examples: FILE_ATTRIBUTE_NORMAL|FILE_ATTRIBUTE_TEMPORARY
    FILE_FLAG_*       => Examples: FILE_FLAG_DELETE_ON_CLOSE|FILE_FLAG_WRITE_THROUGH

=item Returns

 SCALAR: file handle returned by the OS, or throws an eror

=item Example

 my $file_handle = open(file_name   => $file_name,
                        access_mode => GENERIC_READ|GENERIC_WRITE,
                        open_flags  => TRUNCATE_EXISTING,
                        share_mode  => FILE_SHARE_READ|FILE_SHARE_WRITE,
                        attributes  => FILE_FLAG_DELETE_ON_CLOSE);
 if ($@) {
     # print "open file failed: $@\n";
 }

=back

 Notes: For full details of supported flags and attributes,
        refer to MSDN's documentation for CreateFile() at:
        http://msdn.microsoft.com/en-us/library/aa363858(v=vs.85).aspx

=back

=cut

sub open
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($legal,
        $file_name,
        $file_handle,
        $access_mode,
        $open_flag,
        $share_mode,
        $attributes,
        @inputs,
       );

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
            access_mode => {
                type => SCALAR,
            },
            open_flags => {
                type => SCALAR,
            },
            share_mode => {
                type     => SCALAR,
                optional => 1,
                default  => 0,
            },
            attributes => {
                type     => SCALAR,
                optional => 1,
                default  => undef,
            },
        }
    );

    $file_name = $args{file_name};

    # translate access mode
    $access_mode |= eval "$args{access_mode}";

    # translate open flags
    $open_flag |= eval "$args{open_flags}";

    # translate shared mode flags
    $share_mode |= eval "$args{share_mode}"
            if (defined $args{share_mode});

    # translate flags & attributes
    $attributes |= eval "$args{attributes}"
            if (defined $args{attributes});

    # open the file
    $file_handle = Win32API::File::CreateFile($file_name, $access_mode,
                                              $share_mode, [], $open_flag,
                                              $attributes, [])
            or NATE::BaseException->throw("failed to open file: $file_name: $^E");

    # save the OS-returned file handle and return the
    # latest index into our FH DB, with a "WINDOWS" prefix
    # to use as a user reference to the OS file handle
    return _add_file_handle(file_handle => $file_handle,
                            file_name   => $file_name);
}

=for html <hr>

=head2 close()

Close a file handle

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle of the file to be closed (returned by open())

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 close(file_handle => $file_handle);  # $file_handle returned by open()
 if ($@) {
     $Test->nlog("close file failed: $@\n");
 }

=back

 Notes : Calls Win32API::File::CloseHandle()

=back

=cut

sub close
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my $file_handle;

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
        }
    );

    # get the file_handle from the global hash
    $file_handle = _get_file_handle(file_handle => $args{file_handle});

    # close the file
    Win32API::File::CloseHandle($file_handle) or
            NATE::BaseException->throw("failed to close file: $^E");

    # delete the file_handle from our records
    _delete_file_handle(file_handle => $args{file_handle});

    return 0;
}

=for html <hr>

=head2 rename()

Rename a file to a new name

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be renamed

=item new_file_name

 (Required) The new file name

=item flags

 (Optional)

 Sample values are (can be a combination of these):
    MOVEFILE_COPY_ALLOWED       => Allows files to be moved between file systems
    MOVEFILE_REPLACE_EXISTING   => Replace an existing file if it exists
    MOVEFILE_WRITE_THROUGH      => Won't return until the operation has finished
                                   and been flushed to disk
    MOVEFILE_DELAY_UNTIL_REBOOT => Move file after OS reboot. If new_file_name is
                                   not set, the file is deleted after OS reboot

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 rename(file_name     => $file_name,
        new_file_name => $new_file_name);
 if ($@) {
     $Test->nlog("rename file failed: $@\n");
 }

=back

 Notes : Calls Win32API::File::MoveFileEx()

=back

=cut

sub rename
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_name, $new_file_name);

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
            new_file_name => {
                type => SCALAR,
            },
            flags => {
                type => SCALAR,
                optional => 1,
            },

        }
    );

    my %flags;
    $flags{flags} = $args{flags}        if (defined $args{flags});

    # Since this is a call to a local method, and since the NACL Worker STask
    # likes to call methods as ->method_name(), we need to pass this DUMMY_ARG
    # and let the target method shift it off the list
    return &move(DUMMY_ARG,
                 file_name => $args{file_name},
                 new_path  => $args{new_file_name},
                 %flags);
}

=for html <hr>

=head2 move()

Move a file to a new location

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be moved

=item new_path

 (Required) The new file path

=item flags

 (Optional)

 Sample values are (can be a combination of these):
    MOVEFILE_COPY_ALLOWED       => Allows files to be moved between file systems
    MOVEFILE_REPLACE_EXISTING   => Replace an existing file if it exists
    MOVEFILE_WRITE_THROUGH      => Won't return until the operation has finished
                                   and been flushed to disk
    MOVEFILE_DELAY_UNTIL_REBOOT => Move file after OS reboot. If new_file_name is
                                   not set, the file is deleted after OS reboot

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 move(file_name => $file_name,
      new_path  => $new_file_path);
 if ($@) {
     $Test->nlog("move file failed: $@\n");
 }

=back

 Notes : Calls Win32API::File::MoveFileEx()

=back

=cut

sub move
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_name, $new_file_name, $flags);

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
            new_path => {
                type => SCALAR,
                optional => 1,      # allow 'undef' for negative testing
            },
            flags => {
                type => SCALAR,
                optional => 1,
            },
        }
    );

    $file_name     = $args{file_name};
    $new_file_name = $args{new_path};
    $flags        |= (defined $args{flags}) ? eval "$args{flags}" : 0;

    Win32API::File::MoveFileEx($file_name, $new_file_name, $flags)
            or NATE::BaseException->throw(
                    "failed to move/rename file $file_name to $new_file_name: $^E");

    return 0;
}

=for html <hr>

=head2 delete()

Delete a file

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be deleted

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 delete(file_name => $file_name);
 if ($@) {
     $Test->nlog("delete file failed: $@\n");
 }

=back

 Notes : Calls Win32API::File::DeleteFile()

=back

=cut

sub delete
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my $file_name;

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
        }
    );

    $file_name = $args{file_name};

    Win32API::File::DeleteFile($file_name)
            or NATE::BaseException->throw("failed to delete file $file_name: $^E");

    return 0;
}

=for html <hr>

=head2 link()

Create a link to a file

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be linked

=item link_name

 (Required) The link name to be created

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 link(file_name => $file_name,
      link_name => $link_name);
 if ($@) {
     $Test->nlog("link file failed: $@\n");
 }

=back

 Notes : Calls Perl's link()

=back

=cut

sub link
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_name, $link_name);

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
            link_name => {
                type => SCALAR,
            },
        }
    );

    $file_name = $args{file_name};
    $link_name = $args{link_name};

    link($file_name, $link_name)
            or NATE::BaseException->throw(
                "failed to link file $file_name to link $link_name: $^E");

    return 0;
}

=for html <hr>

=head2 symlink()

Create a symlink to a file

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file to be linked

=item link_name

 (Required) The link name to be created

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example
 symlink(file_name => $file_name,
         link_name => $link_name);
 if ($@) {
     $Test->nlog("symlink file failed: $@\n");
 }

=back

 Notes:
 - Perl's symlink() should be supported for SMB2.0+ (Windows Vista/2008 onwards)
 - If not, we need the CPAN module, Win32::Symlink installed
   [http://search.cpan.org/~autrijus/Win32-Symlink-0.04/Symlink.pm]

=back

=cut

sub symlink
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_name, $link_name);

    # validate inputs
    %args = validate (
        @_,
        {
            file_name => {
                type => SCALAR,
            },
            link_name => {
                type => SCALAR,
            },
        }
    );

    $file_name = $args{file_name};
    $link_name = $args{link_name};

    symlink($file_name, $link_name)
            or NATE::BaseException->throw(
                "failed to symlink file $file_name to link $link_name: $^E");

    return 0;
}

=for html <hr>

=head2 unlink()

Remove a file link

=over 4

=item Options

=over

=item link_name

 (Required) The link to be deleted

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 unlink(link_name => $link_name);
 if ($@) {
     $Test->nlog("unlink file failed: $@\n");
 }

=back

 Notes : Calls Perl's unlink()

=back

=cut

sub unlink
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my $link_name;

    # validate inputs
    %args = validate (
        @_,
        {
            link_name => {
                type => SCALAR,
            },
        }
    );

    $link_name = $args{link_name};

    unlink $link_name
            or NATE::BaseException->throw("failed to unlink file link $link_name: $^E");

    return 0;
}

=for html <hr>

=head2 mkdir()

Create a directory

=over 4

=item Options

=over

=item directory_name

 (Required) The directory to be created

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 mkdir(directory_name => $dir_name);
 if ($@) {
     $Test->nlog("mkdir failed: $@\n");
 }

=back

 Notes : Calls Perl's mkdir()

=back

=cut

sub mkdir
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;

    # validate inputs
    %args = validate(
        @_,
        {
            directory_name => { type => SCALAR },
        }
    );

    my $dir_name = $args{directory_name};

    mkdir($dir_name)
        or NATE::BaseException->throw(
            "failed to create directory $dir_name: $!");

    return 0;
}

=for html <hr>

=head2 rmdir()

Remove a directory

=over 4

=item Options

=over

=item directory_name

 (Required) The directory to be removed

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 rmdir(directory_name => $dir_name);
 if ($@) {
     $Test->nlog("rmdir failed: $@\n");
 }

=back

 Notes : Calls Perl's rmdir()

=back

=cut

sub rmdir
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;

    # validate inputs
    %args = validate(
        @_,
        {
            directory_name => { type => SCALAR },
        }
    );

    my $dir_name = $args{directory_name};

    rmdir($dir_name)
        or NATE::BaseException->throw(
            "failed to remove directory $dir_name: $!");

    return 0;
}

=for html <hr>

=head2 read()

Read from a file handle

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle to read from

=item offset

 (Required) The offset into the file to begin reading from

=item num_bytes

 (Required) The number of bytes to read from the file

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 my $contents = read(file_handle => $file_handle,
                     offset      => 10,
                     num_bytes   => 50);
 if ($@) {
     $Test->nlog("read file failed: $@\n");
 } else {
    $Test->nlog("contents read from the file: $contents\n");
 }

=back

 Notes : Calls Win32API::File::setFilePointer and Win32API::File::ReadFile()

=back

=cut

sub read
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_handle,
        $offset,
        $num_bytes,
        $buffer,
        $num_bytes_read);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            offset => {
                type => SCALAR,
            },
            num_bytes => {
                type => SCALAR,
            },
        }
    );

    # get the file_handle from the global hash
    $file_handle = _get_file_handle(file_handle => $args{file_handle});
    $offset      = $args{offset};
    $num_bytes   = $args{num_bytes};

    # set the current pointer to $offset
    Win32API::File::setFilePointer($file_handle, $args{offset}, FILE_BEGIN)
            or NATE::BaseException->throw("failed to set pointer to $offset before reading: $^E");

    # read
    Win32API::File::ReadFile($file_handle, $buffer, $num_bytes, $num_bytes_read, [])
            or NATE::BaseException->throw("failed to read $num_bytes bytes from file: $^E");

    # print the actual number of bytes read
    $Log->comment("read: $num_bytes_read bytes read from file\nread: Data:\n$buffer\n");

    return Protocol::Executor::encode($buffer);
}

=for html <hr>

=head2 write()

Write to a file handle

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle to write to

=item offset

 (Required) The offset into the file to begin writing to

=item contents

 (Required) The contents to write to the file

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 write(file_handle => $file_handle,
       offset      => 10,
       contents    => $contents);
 if ($@) {
     $Test->nlog("write file failed: $@\n");
 }

=back

 Notes : Calls Win32API::File::setFilePointer and Win32API::File::WriteFile()

=back

=cut

sub write
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_handle,
        $offset,
        $contents,
        $num_bytes_wrote);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            offset => {
                type => SCALAR,
            },
            contents => {
                type => SCALAR,
            },
        }
    );

    # get the file_handle from the global hash
    $file_handle = _get_file_handle(file_handle => $args{file_handle});
    $offset      = $args{offset};
    $contents    = Protocol::Executor::decode($args{contents});

    # set the current pointer to $offset
    Win32API::File::setFilePointer($file_handle, $args{offset}, FILE_BEGIN)
            or NATE::BaseException->throw("failed to set pointer to $offset before writing: $^E");

    # write
    Win32API::File::WriteFile($file_handle, $contents, 0, $num_bytes_wrote, [])
            or NATE::BaseException->throw("failed to write to file: $^E");

    return $num_bytes_wrote;
}

=for html <hr>

=head2 set_allocation_size()

Set file allocation size for an opened file handle

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle to use

=item allocation_size

 (Required) The new allocation size of the file to set

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 set_allocation_size(file_handle => $file_handle,
                     allocation_size => 10);
 if ($@) {
     $Test->nlog("failed to set the allocation size on the file: $@\n");
 }

=back

 Notes :
 if allocation_size <= EOF
  - truncate to EOF
 if allocation_size > EOF (past-EOF)
 - seek to allocation_size offset - 1
 - write 1-byte at allocation_size offset - 1

=back

=cut

sub set_allocation_size
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_handle,
        $allocation_size,
        $eof_offset,
        $file_name);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            allocation_size => {
                type => SCALAR,
            },
        }
    );

    # get the file_handle from the global hash
    $file_handle     = _get_file_handle(file_handle => $args{file_handle});
    $file_name       = _get_file_name(file_handle => $args{file_handle});
    $eof_offset      = Protocol::AgnosticClientFile::size($dummy_arg, file_name => $file_name);
    $allocation_size = $args{allocation_size};

    if ($allocation_size <= $eof_offset) {
        # set the current pointer to new allocation size offset
        $Log->comment("Performing SEEK to offset $allocation_size");
        Win32API::File::setFilePointer($file_handle, $allocation_size, FILE_BEGIN)
                or NATE::BaseException->throw("failed to set pointer to $allocation_size before truncating (set EOF): $^E");

        # import the 'SetEndOfFile' API
        my $SetEndOfFile = new Win32::API('kernel32', 'SetEndOfFile', 'N', 'N')
                or NATE::BaseException->throw("failed to import a prototype object for SetEndOfFile: $^E");

        # truncate to the allocation size offset
        $Log->comment("Setting EOF at offset $allocation_size to truncate the file");
        $SetEndOfFile->Call($file_handle)
                or NATE::BaseException->throw("failed to truncate file to length $allocation_size: $^E");

    } else {
        my $write_offset = ($allocation_size - 1);

        # set the current pointer to new allocation size offset - 1
        $Log->comment("Performing SEEK to offset $write_offset");
        Win32API::File::setFilePointer($file_handle, $write_offset, FILE_BEGIN)
                or NATE::BaseException->throw("failed to set pointer to $write_offset before writing: $^E");

        # write
        my $num_bytes_wrote = 0;
        $Log->comment("Writing 1 byte at offset $write_offset to commit space reservation");
        Win32API::File::WriteFile($file_handle, 'x', 0, $num_bytes_wrote, [])
                or NATE::BaseException->throw("failed to write 1 byte to file: $^E");

    }

    return;
}

=for html <hr>

=head2 lock()

Lock a byte range in a file

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle to lock

=item offset

 (Required) The offset into the file to begin locking from

=item num_bytes

 (Required) The number of bytes to lock

=item lock_type

 (Required) The type of lock to acquire
 Valid values are:
    <no_flag>               => read  (or shared) lock
    LOCKFILE_EXCLUSIVE_LOCK => write (or exclusive) lock

=item lock_wait

 (Required) The waiting symantics for the lock
Valid values are:
    LOCKFILE_FAIL_IMMEDIATELY => non-waiting lock
    <no_flag>                 => waiting lock (puts the waiting request in the
                               background and returns control)

=item wait_request_id => \$wait_request_id

 (Required*) A reference to a SCALAR, to save the thread ID (TID)
             of a waiting lock request.
            (This is required only if lock_type F_SETLKW (waiting) is used)

=item Returns

 SCALAR lock ID to reference the lock, or undef

=item Example

 $lock_id = lock(file_handle => $file_handle,
                 offset      => 10,
                 num_bytes   => 50,
                 lock_type   => LOCKFILE_EXCLUSIVE_LOCK,
                 lock_wait   => LOCKFILE_FAIL_IMMEDIATELY);
 if ($@) {
     $Test->nlog("lock file failed: $@\n");
 }

=back

 Notes : imports and calls Win32::API(KERNEL32,'LockFileEx','NNNNP','N')

=back

=cut

sub lock
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_handle,
        $offset,
        $num_bytes,
        $lock_flags);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            offset => {
                type => SCALAR,
            },
            num_bytes => {
                type => SCALAR,
            },
            lock_type => {
                type     => SCALAR,
                optional => 1,
                default  => undef,
            },
            lock_wait => {
                type     => SCALAR,
                optional => 1,
                default  => undef,
            },
        }
    );

    # constants to map to Windows' LockFileEx() values
    use constant LOCKFILE_FAIL_IMMEDIATELY => 1;
    use constant LOCKFILE_EXCLUSIVE_LOCK   => 2;
    use constant LOCKFILE_WAITING_LOCK     => 3; # internal constant

    # get the file_handle from the global hash
    $file_handle = _get_file_handle(file_handle => $args{file_handle});

    # get offset, and number of bytes to lock
    $offset      = $args{offset};
    $num_bytes   = $args{num_bytes};

    # translate flags to integer equivalent
    # - lock type
    $lock_flags |= eval "$args{lock_type}"      if (defined $args{lock_type} and
                                                    $args{lock_type} ne "");

    # - waiting lock flag
    if (defined $args{lock_wait}) {
        $lock_flags |= eval "$args{lock_wait}";
    } else {
        $args{lock_wait} = LOCKFILE_WAITING_LOCK;
    }

    # import LockFileEx()'s prototype for requesting locks
    if (!defined $LockObject){
        $LockObject = new Win32::API(KERNEL32, 'LockFileEx','NNNNNP','N')
                or NATE::BaseException->throw(
                    "failed to import a prototype object for LockFileEx: $^E");
    }

    # lock - check if this is a waiting lock request
    if ($args{lock_wait} eq LOCKFILE_WAITING_LOCK()) {
        # spawn a thread for this waiting lock request
        my $lock_thread = threads->new(\&_lock_thread,
                                       $file_handle,
                                       $lock_flags,
                                       $num_bytes,
                                       $offset
                                      );

        # save the thread object
        _add_lock_thread(file_handle => $args{file_handle},
                         lock_thread => $lock_thread);

        # return the latest index into our Thread DB to
        # use as a user reference to the lock request
        return $Thread_Count;
    }

    # else, this is a non-waiting lock request
    my $status = 1;    # default: SUCCESS
    eval {
        ## Set file pointer to correct offset
        # $Log->comment("Performing SEEK to offset ".$offset);
        # Win32API::File::setFilePointer($file_handle, $offset, FILE_BEGIN)
            # or NATE::BaseException->throw("Could not seek on file $file: $^E");

        $status = $LockObject->Call($file_handle,      # from open()
                                    $lock_flags,       # LOCKFILE_EXCLUSIVE_LOCK, LOCKFILE_FAIL_IMMEDIATELY
                                    0,                 # RESERVED
                                    $num_bytes,        # offset (LOW)
                                    0,                 # offset (HIGH)
                                    pack('PPLLL', 0, 0, $offset, 0, 0) # LPOVERLAPPED
                                   );
        $Log->comment("lock command status = $status");
    };
    if ($status != 1) {
        NATE::BaseException->throw("".(caller(0))[3].": $^E");
    }

    return 0;
}

=for html <hr>

=head2 unlock()

Unlock a byte range lock on a file (uses F_UNLCK)

=over 4

=item Options

=over

=item file_handle

 (Required) The file handle to unlock

=item lock_id

 (Required*) The byte range specified by the lock ID
             * Not required if offset and length are provided

=item offset

 (Required*) The offset into the file to begin unlocking from
             * Not required if lock_id is provided

=item num_bytes

(Required*) The number of bytes to unlock
             * Not required if lock_id is provided

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 unlock(file_handle => $file_handle,
        offset      => 10,
        num_bytes   => 50);
 if ($@) {
     $Test->nlog("unlock file failed: $@\n");
 }

=back

 Notes : imports and calls Win32::API(KERNEL32,'UnlockFileEx','NNNNP','N')

=back

=cut

sub unlock
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my ($file_handle,
        $offset,
        $num_bytes);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            offset => {
                type => SCALAR,
            },
            num_bytes => {
                type => SCALAR,
            },
        }
    );

    # get the file_handle from the global hash
    $file_handle = _get_file_handle(file_handle => $args{file_handle});

    # get offset, and number of bytes to lock
    $offset      = $args{offset};
    $num_bytes   = $args{num_bytes};

    # import UnlockFileEx()'s prototype for requesting unlock ops
    if (!defined $UnlockObject){
        $UnlockObject = new Win32::API(KERNEL32, 'UnlockFileEx','NNNNP','N')
                or NATE::BaseException->throw(
                    "failed to import a prototype object for UnlockFileEx: $^E");
    }

    # unlock
    my $status = 1;    # default: SUCCESS
    eval {
        $status = $UnlockObject->Call($file_handle,      # from open()
                                      0,                 # RESERVED
                                      $num_bytes,        # offset (LOW)
                                      0,                 # offset (HIGH)
                                      pack('PPLLL', 0, 0, $offset, 0, 0) # LPOVERLAPPED
                           );
    };
    if ($status != 1) {
        NATE::BaseException->throw("".(caller(0))[3].": $^E");
    }

    return 0;
}

=for html <hr>

=head2 get_wait_status()

Return the status for a waiting lock operation request

=over 4

=item Options

=over

=item wait_request_id

 (Required) The waiting lock request's thread ID, returned by lock()

=item Returns

 SCALAR ("waiting", "exited", "invalid")
     waiting = if the thread is still alive
     exited  = if the thread has exited
     invalid = if the thread object is undef (thread does not exist)

=item Example

 my $wait_lock_status = get_wait_status(
                            file_handle     => $file_handle,
                            wait_request_id => $thread_id);
 $Test->nlog("lock wait status is: $wait_lock_status\n");

=back

=back

=cut

sub get_wait_status
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;

    my ($file_handle,
        $wait_request_id);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            wait_request_id => {
                type     => SCALAR
            },
        }
    );

    # get the file_handle index
    $file_handle = $args{file_handle};

    # get the waiting request's thread index
    $wait_request_id = $args{wait_request_id};

    # get the thread object from our global hash
    my $thread_object = _get_lock_thread(file_handle => $file_handle,
                                         lock_thread => $wait_request_id);

    # check if the thread object exists
    return "invalid"    if (!defined $thread_object);

    # check if the thread is still alive
    # if yes, return "waiting"
    return "waiting"    if ($thread_object->is_running());

    # else... the thread seems to have exited
    # so, return "exited"
    return "exited";
}

=for html <hr>

=head2 kill_waiting_request()

Return the status for a waiting lock/unlock operation request

=over 4

=item Options

=over

=item wait_request_id

 (Required) The waiting lock request's thread ID, returned by lock()

=item Returns

 0 for SUCCESS, or throws an error for FAILURE

=item Example

 kill_waiting_request(file_handle     => $file_handle,
                      wait_request_id => $wait_request_id);
 if ($@) {
     $Test->nlog("failed to kill waiting request: $@\n");
 }

=back

=back

=cut

sub kill_waiting_request
{
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;

    my ($file_handle,
        $wait_request_id);

    # validate inputs
    %args = validate (
        @_,
        {
            file_handle => {
                type => SCALAR,
            },
            wait_request_id => {
                type     => SCALAR
            },
        }
    );

    # get the file_handle index
    $file_handle = $args{file_handle};

    # get the waiting request's thread index
    $wait_request_id = $args{wait_request_id};

    # get the thread object from our global hash
    my $thread_object = _get_lock_thread(file_handle => $file_handle,
                                         lock_thread => $wait_request_id);

    return 0    if (!defined $thread_object);

    # stop the thread
    $thread_object->kill('KILL')
            or NATE::BaseException->throw(
                    "failed to kill waiting lock request $wait_request_id: $!");

    #
    if ($thread_object->is_running()) {
        NATE::BaseException->throw("\nwaiting lock thread $wait_request_id did not die!");
    }
    if ($thread_object->is_joinable()) {
        #$thread_object->join();
    }

    return 0;
}

=for html <hr>

=head2 list_streams()

Get list of the streams associated with a file.

=over 4

=item Options

=over

=item file_name

 (Required) The fully specified name of the file for which the stream names
 are to be returned.

=item Returns

 A reference to an array of the stream names.  Throws an error on failure.

=item Example

 my $streams = list_streams(file_name => $file_name);

=back

 Notes: imports and calls Win32::API CreateFile, CloseHandle, BackupRead and
 BackupSeek.

=back

=cut

my $GENERIC_READ = 0x80000000;
my $OPEN_EXISTING = 3;
my $INVALID_HANDLE_VALUE = -1;
my $BACKUP_ALTERNATE_DATA = 0x00000004;
my @STREAM_ID_KEYS = qw(id attrs size_low size_high name_size);

sub list_streams {
    my $dummy_arg = shift; # Skip the package name
    my %args = @_;
    my $file_name;

    # validate inputs
    %args = validate(
        @_,
        {
            file_name => {
                type => SCALAR,
            },
        }
    );

    # get file name
    $file_name = $args{file_name};

    # import our Win32::API prototypes
    if (!defined $CreateFile) {
        $CreateFile = new Win32::API(KERNEL32, 'CreateFile', 'PNNPNNN', 'N') ||
            throw NATE::BaseException("Cannot load CreateFile() function: $^E");
    }
    if (!defined $CloseHandle) {
        $CloseHandle = new Win32::API(KERNEL32, 'CloseHandle', 'N', 'I') ||
            throw NATE::BaseException("Cannot load CloseHandle() function: $^E");
    }
    if (!defined $BackupRead) {
        $BackupRead = new Win32::API(KERNEL32, 'BackupRead', 'NPNPIIP', 'I') ||
            throw NATE::BaseException("Cannot load BackupRead() function: $^E");
    }
    if (!defined $BackupSeek) {
        $BackupSeek = new Win32::API(KERNEL32, 'BackupSeek', 'NNNPPP', 'I') ||
            throw NATE::BaseException("Cannot load BackupSeek() function: $^E");
    }

    my $handle = $CreateFile->Call(
        $file_name . "\0", $GENERIC_READ, 0, 0, $OPEN_EXISTING, 0, 0);
    if ($handle == $INVALID_HANDLE_VALUE) {
        throw NATE::BaseException(
            "CreateFile failed opening file $file_name: $^E");
    }

    my @stream_names = ();

    my $failed;
    my $p_context = pack('L', 0);
    for (;;) {
        my $p_bytes_read = pack('L', 0);
        my $p_stream_id = pack('L5', 0, 0, 0, 0, 0);
        if (!$BackupRead->Call($handle, $p_stream_id, length($p_stream_id),
                               $p_bytes_read, 0, 0, $p_context)) {
            last;
        }

        my $bytes_read = unpack('L', $p_bytes_read);
        last if ($bytes_read == 0);

        my %stream;
        @stream{@STREAM_ID_KEYS} = unpack('L5', $p_stream_id);

        if ($stream{id} == $BACKUP_ALTERNATE_DATA) {
            my $size = $stream{name_size};
            my $name = "\0" x (Win32::API::IsUnicode() ? $size : 2 * $size);
            if (!$BackupRead->Call($handle, $name, $size, $p_bytes_read, 0, 0,
                                   $p_context)) {
                $failed = "BackupRead failed unexpectedly: $^E";
                last;
            }
            # translate from Unicode (sort of) and strip trailing NULs
            $name =~ tr/\0//d;
            # extract name (strip :: and attribute name)
            $name =~ s/^:(.*):.*$/$1/;
            push(@stream_names, $name);
        }

        # move to next stream
        my $pSeekedBytesLow = pack('L', 0);
        my $pSeekedBytesHigh = pack('L', 0);
        if (!$BackupSeek->Call($handle, $stream{size_low}, $stream{size_high},
                               $pSeekedBytesLow, $pSeekedBytesHigh,
                               $p_context)) {
            last;
        }
    }

    # Win32 API insists we do a BackupRead Abort when we're done
    $BackupRead->Call($handle, undef, 0, 0, 1, 0, $p_context);
    $CloseHandle->Call($handle);

    throw NATE::BaseException($failed) if $failed;

    return \@stream_names;
}


#=========================================================================#
#                            INTERNAL METHODS                             #
#=========================================================================#

sub _add_file_handle
{
    my %args = @_;

    # generate a random number for the FH key
    my $time   = time();
    my $rand   = rand();
    my $rand2  = rand();
    my $fh_key = "WINDOWS$time$rand$rand2";

    # add the file name to our global hash
    $Files{$fh_key}{file_name}   = $args{file_name};
    # add the OS file handle to our global hash
    $Files{$fh_key}{file_handle} = $args{file_handle};

    return $fh_key;
}

sub _get_file_handle
{
    my %args     = @_;
    my $fh_index = $args{file_handle};

    # return the hash entry for the requested
    # file handle index from our global hash
    return $Files{$fh_index}{file_handle};      # 'undef' if nonexistent
}

sub _get_file_name
{
    my %args     = @_;
    my $fh_index = $args{file_handle};

    # return the hash entry for the requested
    # file handle index from our global hash
    return $Files{$fh_index}{file_name};      # 'undef' if nonexistent
}

sub _delete_file_handle
{
    my %args     = @_;
    my $fh_index = $args{file_handle};

    # delete the hash entry for the requested
    # file handle index from our global hash
    delete $Files{$fh_index};

    return;
}

#--------------------------------------------------------------------
# INTERNAL ONLY:
# this method is for creating a thread for waiting lock requests.
# this method should NOT be called directly by any external caller.
#--------------------------------------------------------------------
sub _lock_thread
{
    # Thread 'cancellation' signal handler
    $SIG{'KILL'} = sub {
                        return;
                   };

    my ($file_handle, $lock_flags, $num_bytes, $offset) = @_;

    eval {
        ## Set file pointer to correct offset
        # $Log->comment("Performing SEEK to offset ".$offset);
        # Win32API::File::setFilePointer($file_handle, $offset, FILE_BEGIN)
            # or NATE::BaseException->throw("Could not seek on file $file: $^E");

        my $status = $LockObject->Call($file_handle,      # from open()
                                       $lock_flags,       # LOCKFILE_EXCLUSIVE_LOCK, LOCKFILE_FAIL_IMMEDIATELY
                                       0,                 # RESERVED
                                       $num_bytes,        # offset (LOW)
                                       0,                 # offset (HIGH)
                                       pack('PPLLL', 0, 0, $offset, 0, 0) # LPOVERLAPPED
                                      );
        $Log->comment("lock command status = $status");
    };
    if ($@) {
        chomp $@;
        NATE::BaseException->throw("".(caller(0))[3].": $^E (".$@.")");
    }

    return;
}

sub _add_lock_thread
{
    my %args     = @_;
    my $fh_index = $args{file_handle};

    # increment the global counter
    $Thread_Count++;

    # add the lock thread to our global hash, per FH
    $Files{$fh_index}{waiting_requests}{$Thread_Count} = $args{lock_thread};

    return;
}

sub _get_lock_thread
{
    my %args         = @_;
    my $fh_index     = $args{file_handle};
    my $thread_index = $args{lock_thread};

    # return the hash entry for the requested
    # file handle index from our global hash
    return $Files{$fh_index}{waiting_requests}{$thread_index};      # 'undef' if nonexistent
}

sub _delete_lock_thread
{
    my %args         = @_;
    my $fh_index     = $args{file_handle};
    my $thread_index = $args{lock_thread};

    # delete the hash entry for the requested
    # file handle's thread index from our global hash
    delete $Files{$fh_index}{waiting_requests}{$thread_index};

    return;
}



=head1 SEE ALSO

=over

=item Protocol::ClientFileOperations

=item Protocol::ClientFilehandle

=item Protocol::LinuxClientFile

=back

=head1 STATUS

    Shared

=head1 AUTHOR

    WAFL Protocol QA (dl-wafl-proto-qa@netapp.com)

=head1 COPYRIGHT

    Copyright 2011, Network Appliance Inc., All Rights Reserved.

=cut


1;
