# This library adds functionality to the standard NATE TestInfo package.
#
# Why not include this in the standard TestInfo package?  Good question.
# I'm not sure that most folks would want it and I need the functionality
# now without going through a review process.  The functionality presented
# here is fairly straightforward but is useful.

package QATestInfo;

use base qw(NATE::TestInfo);
use Data::Dumper;

# Our version of "new" for this class.  Calls the standard TestInfo root
# and then blesses the returned object into this class
sub root {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = NATE::TestInfo->root(@_);

    bless($self, $class);
}

# TestsRunningAt
#
# Goes through all of the test items and determines if that test was running
# at $time (expressed in seconds since 1/1/70).  Returns the list of tests
# that were determined to be running
sub TestsRunningAt {
    my $self = shift;
    my $time = shift;
    my @retval;

    foreach my $result ($self->descendants()){
	if (($result->start_time() < $time) && ($result->end_time() > $time)){
	    # This is a test that was running when we saw the problem
	    push(@retval, $result);
	}
    }

    @retval;
}

# TestChain
#
# Determines the "path" from a test to the root and returns the testids
# of the tests and parent nodes in an array (or array ref if in scalar
# context).
sub TestChain {
    my ($self, $test) = @_;

    my @retval;

    return unless $test;
    unshift(@retval, $test->test_id);
    while (my $pid = $test->parent_id()){
	unshift(@retval, $pid);
	$test = $test->parent();
    }
    if (wantarray){
	return @retval;
    } else {
	return \@retval;
    }
}

# IsDescendant
#
# Determines if $t2 is a child of $t1.  Returns 0 if $t2 is not a child
# and non-zero otherwise.
sub IsDescendant {
    my ($self, $t1, $t2) = @_;

    my @descentants = $t1->descentants();
    my $testid = $t2->test_id();

    my @result = grep {$_->test_id() == $testid} @descentants;
    return scalar @result;
}

# IsParent
#
# Determines if $t2 is a parent of $t1.  Returns 0 if $t2 is not a
# parent of $t1 and non-zero otherwise.
sub IsParent {
    my ($self, $t1, $t2) = @_;
    my $testid = $t2->test_id();

    my @result = grep(/^$testid$/, $self->TestChain($t1));
    return scalar @result;
}

# ReduceTests
#
# Reduces a list of tests down to the lowest level tests in the list.
# For example, if this sub is passed the following chain of tests:
#   cthon
#   cthon/cthon
#   cthon/cthon/udpv2_local
# The cthon and cthon/cthon tests will be removed from the list since
# they often add noise and confusion to a list of results since they
# likely failed due to the subtest failure.
#
# Returns the pruned down list of tests.
sub ReduceTests {
    my ($self, @tests) = @_;
    my @chains;

    foreach my $test (@tests){
	push(@chains,join('::', $self->TestChain($test)));
    }

    my @retval;
    # Reduce the chains
    foreach my $chain (sort {length($a) <=> length($b)} @chains){
	if (grep(/^$chain/, @chains) == 1){ # Only save unique chains
	    if (my $test = $self->GetTestByTestId((split('::', $chain))[-1])){
		push(@retval, $test);
	    }
	}
    }
    return @retval;
}

# GetTesetByTestId
#
# Finds a test in the group by the testid.  Returns the test if one is found
# and undef otherwise.
sub GetTestByTestId {
    my ($self, $testid) = @_;

    if (!defined($testid) || ($testid =~ /^\s*$/)){
	return undef;
    }
    my @list = grep {$_->test_id() == $testid} $self->descendants();

    warn "Too many matches for test $testid" if @list > 1;
    if (@list < 1){
	warn "Too few matches for test $testid";
	return undef;
    }

    return $list[0];
}

1;
