#!/usr/bin/perl -w

# $Id$
# Copyright (c) 2008 Network Appliance, Inc.
# All rights reserved.

#************************************************************************
#
# NAME
#   QC_Client.pm - module to communicate remotely with Quality Center.
#
# SYNOPSIS
#   Provides remote access to Quality Center to filter on tests and create
#   test sets.
#
# DESCRIPTION
#   This Perl module is meant to be a programmatic way of interfacing with
#   Quality Center using Perl.
#
# AUTHOR
#   Julie Glenn <julieg@netapp.com>
#
#************************************************************************

## @summary     Routines to pass Quality Center API requests to a proxy 
##              a Windows machine.  Actual API is performed using Visual Basic
##              calls.
## @description
##              These routines send request messages and receive responses 
##              from a socket server on a Windows machine which converts the
##              requests to Visual Basic API calls to Quality Center.
## @author      julieg@netapp.com
## @notes       The routines die if a socket connection cannot be established. 

package QC_Client;

use strict; 
use Data::Dumper;
use IO::Socket;
use Params::Validate qw(validate SCALAR HASHREF ARRAYREF GLOBREF GLOB UNDEF);
use TracePkg;
use Socket;
use vars qw(@ISA @EXPORT_OK $VERSION);

#------------
# Prototypes
#------------
sub _validate_testset_name($);
sub _process_test_cases($$$$);

=pod

=head1 NAME

QC_Client.pm = Programmatic interface to Quality Center

=head1 AUTHOR

Julie Glenn <julieg@netapp.com>

=head1 DESCRIPTION

The routines contained in this module serve as an interface between Perl
scripts and Quality Center.  They communicate via sockets to a Quality Center
proxy that is written in Visual Basic and resides on a Windows machine.  The
proxy processes the requests from these routines and formulates them into
Quality Center API calls which retrieve data from or write data to the Quality
Center tables.  Once table access is complete, responses are then sent back 
from the proxy and processed by these routines.  For actions that require
read-only access to QC, the guest account is used by default.  This is a
read-only account.  For actions that require write access to QC, the "autotest"
account is used by default.

=head1 SYNOPSIS

    use QC_Client;

    # If you want to get a filtered list of autotest tests
    $href = tw_get_tests(domain  => "ONTAP",
                         project => "boilermaker_0");

    # If you want to get a filtered list of tests based on a custom
    # filter
    my %filter_set = (
                      "TS_USER_07" => "WAFL",
                     );

    $href = tw_get_tests_filtered(filters => \%filter_set,
                                  domain  => "ONTAP",
                                  project => "boilermaker_0");

    # If you want to get a list of field names and their corresponding
    # types
    my $href = tw_get_value_lists(domain  => "ONTAP",
                                  project => "boilermaker_0");

    # If you want to get a list of tests in a specific test set
    my $href = tw_get_testset_tests(domain       => "ONTAP",
                                    project      => "Boilermaker_0",
                                    testset_name => qw(\Root\regression\Core\WAFL\7mode\WAFL\automated_regression\MS6\BAT\WAFL001_Space_Management));

    # If you want to get a list of test set names in a specific folder
    my $aref = tw_get_testset_names(domain       => "ONTAP",
                                    project      => "Boilermaker_0",
                                    testset_path => qw(\Root\regression\Core\WAFL\7mode\WAFL\automated_regression\MS6\BAT));

    # If you want to get a list of test data for a specified list of test
    # instances 
    my $aref = tw_get_by_instance_id(domain            => "ONTAP",
                                     project           => "Boilermaker_0",
                                     test_inst_id_aref => \@test_inst_ids);

    # If you want to get a list of test set names in all folders 
    my $aref = tw_get_all_testsets(domain       => "ONTAP",
                                   project      => "Boilermaker_0");

    # If you want to create test runs for a list of tests in a test set
    my $href = tw_create_test_run(domain       => "ONTAP",
                                  project      => "Boilermaker_0",
                                  testset_name => qw(\Root\FFT\Core\mhost),
                                  run_prefix   => "boilerms6_smokeHouse_",
                                  test_names   => $href);



=head1 PUBLIC SUBROUTINES

=cut

#------------------------
# Define constant values
#------------------------

# Standard message types
use constant GET_FILT_TEST_REQ      => "001";
use constant GET_TESTSET_NAME_REQ   => "002";
use constant GET_VAL_LISTS_REQ      => "003";
use constant GET_TESTSET_REQ        => "004";
use constant CREATE_TEST_RUN_REQ    => "005";
use constant GET_TEST_INST_DATA_REQ => "006";

# User-defined field names
use constant TS_KEYWORDS      => "TS_USER_15";
use constant TS_FUNC_SUBAREA  => "TS_USER_08";

use constant DEF_SOCK_TIMEOUT => 90;    
use constant MAX_SOCK_TIMEOUT => 1200;  
use constant PEER_PORT        => 5001;
use constant QC_SERV_IP       => 'qcserver.rtp.netapp.com';
use constant QC_SOCK_TIMEOUT  => 30;
use constant ROOT_NODE        => 'Root\\';
use constant TERM_STRING      => "END_OF_DATA";
use constant QC_DEFAULT_USER  => "Qc_Guest";
use constant QC_DEFAULT_PASS  => "QC123";
 
my $qc_field_names = "TS_CREATION_DATE|TS_DESCRIPTION|"
                     . "TS_DEV_COMMENTS|TS_EXEC_STATUS|TS_NAME|"
                     . "TS_RESPONSIBLE|TS_USER_[0-3][0-9]|"
                     . "TS_SUBJECT|TS_STEPS|TS_TYPE|TS_ESTIMATE_DEVTIME|"
                     . "TS_TEMPLATE|TS_STATUS|TS_VTS|TS_PATH|TS_TEST_ID";

my $Trace = new TracePkg();

$VERSION    = 0.01;
our @ISA    = qw(Exporter);
our @EXPORT = qw(tw_get_tests tw_get_tests_filtered tw_get_testset_names tw_get_value_lists tw_get_testset_tests tw_create_test_run tw_get_all_testsets tw_get_by_instance_id);

#------------------------------------------------------------------------------
#
# _create_qc_socket
#
#     This subroutine creates a TCP socket through which to send client requests
#     to Quality Center.  The QC Server IP address is used for the connection.
#
# Input:
#     None 
#
# Output:
#     None
#
# Return value:
#     $sock - Socket filehandle
#
#------------------------------------------------------------------------------
sub _create_qc_socket{

    $Trace->debug(msg=>"Creating socket.  Peer IP Address: " . QC_SERV_IP);

    #---------------------
    # Create a new socket
    #---------------------
    my $sock = new IO::Socket::INET (
                                     PeerAddr => QC_SERV_IP,
                                     PeerPort => PEER_PORT,
                                     Proto => 'tcp',
                                     Timeout => QC_SOCK_TIMEOUT,
                                    );
    $Trace->fatal(msg=>"Could not create socket: $!", result=>"FAIL") 
        unless $sock;

    $sock->autoflush(1);

    $Trace->exit(msg=>"QC socket successfully created");

    return $sock;
}

=pod

=head2 B<tw_get_tests>

=over 4 

=item Description

This routine queries Quality Center for list of filtered tests that are to be
executed for Smokehouse.  A predefined filter of "TS_USER_15 = "*autotest*" is
used.  TS_USER_15 is the Keywords field.  A list of filtered tests is returned.

=item Arguments 

 domain  => Name of the QC domain to use when obtaining the
            list of test sets (Required)

 project => Name of the QC project to use (Required)

 qc_user => Name of the QC user (Optional)

 qc_pwd  => Password for qc_user (Optional)

=item Returns 

Returns a hash reference containing a list of filtered tests in the form:

    <test_case_id> => {
                       'project'       => <project_name>,
                       'test_name'     => <test_name>, 
                       'nate_script'   => <nate_script_path>, 
                       'nate_params'   => '<nate_params>',
                       'test_owner'    => <test_owner>,
                       'ontap_support' => <ontap_support>,
                       'func_area'     => <functional_area>,
                       'func_subarea'  => <functional_subarea>
                      }  

    <project_name> is the fully-qualified path name for the folder 
    containing the test (e.g., "Subject/folder1/folder2"),

    <test_name> is the name of the test (e.g., "mhost_no_key"),

    <nate_script_path> is the relative p4 path containing the NATE 
    .thpl or .stest file (e.g., "nblade/cifs/auth/sanity.thpl"),

    <nate_params> are the NATE params that are defined for the test 
    case in the Test Plan module (e.g., "TESTCASE=<testcase_name>").  
    Note that Test Lab NATE params are not included,

    <test_owner> is the name of the person responsible for 
    implementing the test,

    <ontap_support> identifies the ONTAP mode(s) for which this test
    case applies,

    <functional_area> is the Functional Area to which the test belongs,

    <functional_subarea> is the Functional Subarea to which the test belongs.

=item Example

    # Get a list of Smokehouse tests from Quality Center
    $href = tw_get_tests(domain  => "ONTAP",
                         project => "boilermaker_0");

=back

=cut

#------------------------------------------------------------------------------
#
# tw_get_tests
#     This routine sends a request to the QC proxy on a Windows machine.
#
# Input:
#     domain   - QC domain
#     project  - QC project
#     qc_user  - QC user name 
#     qc_pwd   - QC password 
#
#     The values for domain, project, qc_user, and qc_pwd are used for
#     logging in to QC.  Domain and project are required.
#
# Output:
#     None
#
# Return value:
#     test_href - Hash reference containing test ID and a hash of test
#                 data as follows:
#
#     <test_case_id> => {
#                        'project'       => <project_name>,
#                        'test_name'     => <test_name>, 
#                        'nate_script'   => <nate_script_path>, 
#                        'nate_params'   => '<nate_params>',
#                        'test_owner'    => <test_owner>,
#                        'ontap_support' => <ontap_support>
#                        'func_area'     => <functional_area>,
#                        'func_subarea'  => <functional_subarea>
#                       }  
#
#------------------------------------------------------------------------------
sub tw_get_tests {

    my %params = validate (
                              @_,
                              {   
                                  domain  => {type    => SCALAR},
                                  project => {type    => SCALAR},
                                  qc_user => {type    => SCALAR|UNDEF,
                                              default => QC_DEFAULT_USER},
                                  qc_pwd  => {type    => SCALAR|UNDEF,
                                              default => QC_DEFAULT_PASS},
                              }
                          );

    my $domain  = $params{domain};    # QC Domain to log in to
    my $project = $params{project};   # QC Project to log in to 
    my $qc_user = $params{qc_user};   # QC User ID
    my $qc_pwd  = $params{qc_pwd};    # QC Password   

    my $href;         # Hash containing the test ID (key) and test name (value)
    my %filter;       # Filter to use for filtering on Smokehouse tests

    #----------------------------------------------
    # Set up the filters to be used for smokehouse 
    #----------------------------------------------
    $filter{+TS_KEYWORDS} = "*autotest*";

    $href = tw_get_tests_filtered(filters => \%filter,
                                  domain  => $domain,
                                  project => $project,
                                  qc_user => $qc_user,
                                  qc_pwd  => $qc_pwd);

    $Trace->exit(msg => "Filtered test retrieval from QC complete");

    return $href;

}

=pod

=head2 B<tw_get_tests_filtered>

=over 4 

=item Description

This routine queries the Quality Center TEST table for list of filtered tests.  
Multiple filters can be specified.  Values can be and'ed or or'ed together for 
a single field by enclosing each value in single or double quotes.  A list 
of filtered tests is returned.

=item Arguments 

 filters => A hash ref containing the field names and values on which 
            to filter.  The hash key is the field name and the value 
            is the value(s) on which to filter.  The filter text can 
            include "and" and "or" but individual values must be 
            enclosed be enclosed in quotes.  Filter text that 
            contains spaces must also be enclosed in quotes. Filter
            text may contain wildcard (*) characters.

 domain  => Name of the QC domain to use when obtaining test list 
            (Required)

 project => Name of the QC project to use (Required)

 qc_user => Name of the QC user (Optional)

 qc_pwd  => Password for qc_user (Optional)

=item Returns 

Returns a hash reference containing a list of filtered tests in the form:

    <test_case_id> => {
                       'project'       => <project_name>,
                       'test_name'     => <test_name>, 
                       'nate_script'   => <nate_script_path>, 
                       'nate_params'   => '<nate_params>'
                       'test_owner'    => <test_owner>,
                       'ontap_support' => <ontap_support>
                       'func_area'     => <functional_area>,
                       'func_subarea'  => <functional_subarea>
                      }  

    <project_name> is the fully-qualified path name for the folder 
    containing the test (e.g., "Subject/folder1/folder2"),

    <test_name> is the name of the test (e.g., "mhost_no_key"),

    <nate_script_path> is the relative p4 path containing the NATE 
    .thpl or .stest file (e.g., "nblade/cifs/auth/sanity.thpl"),

    <nate_params> are the NATE params that are defined for the test 
    case in the Test Plan module (e.g., "TESTCASE=<testcase_name>").  
    Note that Test Lab NATE params are not included,

    <test_owner> is the name of the person responsible for 
    implementing the test,

    <ontap_support> identifies the ONTAP mode(s) for which this test
    case applies,

    <functional_area> is the Functional Area to which the test belongs,

    <functional_subarea> is the Functional Subarea to which the test belongs.

=item Example

    my %filter_set = (
                     "TS_RESPONSIBLE" => "myuserid"
                     "TS_NAME"        => "'mhost*' or '*dblade*'"
                     "TS_USER_10"     => "'Data Integrity'"
                     );

    # Get a list of filtered tests from Quality Center
    $href = tw_get_tests_filtered(filters => \%filter_set,
                                  domain  => "ONTAP",
                                  project => "boilermaker_0");

=back

=cut

#------------------------------------------------------------------------------
#
# tw_get_tests_filtered
#     This routine sends a request to the QC proxy on a Windows machine.
#
# Input:
#     filters - Hash reference containing the name of the QC field to filter on
#               and the value on which to filter.
#     domain  - QC domain
#     project - QC project
#     qc_user - QC user name 
#     qc_pwd  - QC password 
#
#     The values for domain, project, qc_user, and qc_pwd are used for
#     logging in to QC.  Domain and project are required.
#
# Output:
#     None
#
# Return value:
#     test_href - Hash reference containing test ID/test name.  Test ID is
#                 the key and the test name is the value.
#
#------------------------------------------------------------------------------
sub tw_get_tests_filtered {

    my %hash;        # Hash containing the test ID (key) and test name (value)
    my @line_data;   # Array of tab-delimited lines containing test ID/test name
    my $tst_count;   # Number of tests returned from QC

    $tst_count = 0;
   
    my %params = validate (
                              @_,
                              {   
                                  filters => {type => HASHREF},
                                  domain  => {type => SCALAR},
                                  project => {type => SCALAR},
                                  qc_user => {type    => SCALAR|UNDEF,
                                              default => QC_DEFAULT_USER},
                                  qc_pwd  => {type    => SCALAR|UNDEF,
                                              default => QC_DEFAULT_PASS},
                              }
                          );

    my @lines = ();
    my ($key, $value);

    my $filter_href = $params{filters};
    my $domain      = $params{domain};
    my $project     = $params{project};
    my $qc_user     = $params{qc_user};
    my $qc_pwd      = $params{qc_pwd};

    #--------------------------------------------------
    # Put something in so password doesn't get ignored
    #--------------------------------------------------
    if (!$qc_pwd) {
         $qc_pwd = " ";
    }

    $Trace->entry(msg => "Getting a filtered list of tests using the " .
                         "following filter data: " .  Dumper $filter_href);

    $Trace->debug(msg => "QC Domain:  " . $domain . "\n" .
                         "QC Project: " . $project);

    #----------------------------------------------------------
    # Include the QC domain and project in the message request
    #----------------------------------------------------------
    push(@lines, "$domain\n");
    push(@lines, "$project\n");
    push(@lines, "$qc_user\n");
    push(@lines, "$qc_pwd\n");

    my $num_keys = keys %$filter_href;

    if ($num_keys < 1) {
        $Trace->fatal(msg      => "No filters were passed to " .
                                  "tw_get_tests_filtered.  Filter set is empty",
                      result   => "CONFIG");
    }

    #--------------------------------
    # Add each filter to the request
    #--------------------------------
    foreach my $key (sort (keys %$filter_href)) {

        #--------------------------------------
        # Make sure the QC field name is valid 
        #--------------------------------------
        if (!($key =~ /$qc_field_names/)) {
            $Trace->fatal(msg    => "QC Field Name $key is invalid",
                          result => "CONFIG"); 
            return -1;
        } else {
            push(@lines, "$key\t$filter_href->{$key}\n");
        }
    }  

    #----------------------------------------------
    # Add the termination string to the line array
    #----------------------------------------------
    push(@lines, TERM_STRING);

    #-----------------------------------------
    # Create a socket to talk to the QC proxy
    #-----------------------------------------
    my $sock = _create_qc_socket();

    #---------------------------------------------
    # Send the request to get tests to the server
    #---------------------------------------------
    print $sock GET_FILT_TEST_REQ;

    #---------------------------------------- 
    # Send the list of filters to the server
    #---------------------------------------- 
    print $sock @lines;

    #----------------------------------------
    # Read the list of tests from the server
    #----------------------------------------
    $Trace->trace(msg => "Reading data from QC server");
    my $ret_val = _read_sock(sock => $sock);

    if ($ret_val == -1) {
        $Trace->fatal(msg => "Error occurred while trying to read data from " .
                             "socket during get filtered tests request");
    }

    @lines = @$ret_val;

    my ($err_code, $err_msg, $href);

    #-----------------------------------------
    # Add all returned test cases to the hash
    #-----------------------------------------
    $href = _process_test_cases(\@lines, 0, \$err_code, \$err_msg);

    if ($href == -1) {
        $Trace->fatal(msg => "An error occurred trying to filter "
                      . "on Smokehouse tests\nError code: " .
                      $err_code . "\nError message: " . $err_msg,
                      result => "FAIL");
    }

    close($sock);

    $Trace->exit(msg => "Filtered test retrieval from QC completed");

    return $href;

}

=pod

=head2 B<tw_get_value_lists>

=over 4

=item Description

This routine gets a list of filterable QC field names and label names along 
with their corresponding data types.  If a field is a selection list, the list 
name and list of possible values is also returned.  For multiple-selection 
lists, the type is "multi".  For single selection lists, the type is "list".

=item Arguments 

 domain       => Name of the QC domain to use when obtaining test
                 list (Required)

 project      => Name of the QC project to use (Required)

 qc_user      => Name of the QC user (Optional)

 qc_pwd       => Password for qc_user (Optional)

 qc_table     => QC table name to return field names for
                 (Optional; default = "TEST")

=item Returns 

Returns a hash reference containing a list of fields in the form:

    <QC_field_name> => {
                          'Type'      => <data_type>,
                          'Values'    => [<list_values>],
                          'Label Name'=> <label_name>,
                       }  

    <data_type> is the data type for the QC field.  Possible values 
    are:
        char   - free text field
        DATE   - date field
        list   - single selection list
        memo   - memo field
        multi  - multiple selection list
        number - integer

    <list_values> is an array of selection values that are available 
    for the list.  This is only returned if the type is "list" or 
    "multi".

    <label_name> is the label name associated with the QC field.  It 
    is the label that is displayed on the QC GUI.

    The Function Subarea field is a special case.  It is a list of 
    lists.  Each main sublist corresponds to a Functional Area.  The 
    format of the data for Functional Subareas is slightly different 
    than that for other lists:

    <QC_field_name> => {
                          'Type'      => <data_type>,
                          'Values'    => {
                                          'Functional Area' => [<list_values>],
                                         }
                          'Label Name'=> <label_name>,
                       }  

=item Example

   my $href = tw_get_value_lists(domain  => "ONTAP",
                                 project => "Boilermaker_0"); 

=back

=cut

#------------------------------------------------------------------------------
#
# tw_get_value_lists
#
#     This routine sends a request to the QC proxy on a Windows machine for
#     a list of all field names in the TEST table.  All fields, along with
#     their associated types and QC user interface field names, are returned 
#     in a hash.  If a field is a list, its value list and list name are also
#     included in the hash.  Possible field types include the following:
#     
#     char   - free text field
#     DATE   - date field 
#     list   - single selection list
#     memo   - memo field
#     multi  - multiple selection list
#     number - integer
#
# Input:
#     domain   - QC domain
#     project  - QC project
#     qc_user  - QC user name 
#     qc_pwd   - QC password 
#     qc_table - QC table to obtain the value lists from (default = TEST)
#
#     The values for domain, project, qc_user, and qc_pwd are used for
#     logging in to QC.  Domain and project are required.
#
# Output:
#     None
#
# Return value:
#     test_href - Hash reference containing test ID/test name.  Test ID is
#                 the key and the test name is the value.
#
#------------------------------------------------------------------------------
sub tw_get_value_lists {

    my %params = validate (
                              @_,
                              {   
                                  domain   => {type    => SCALAR},
                                  project  => {type    => SCALAR},
                                  qc_user  => {type    => SCALAR|UNDEF,
                                               default => QC_DEFAULT_USER},
                                  qc_pwd   => {type    => SCALAR|UNDEF,
                                               default => QC_DEFAULT_PASS},
                                  qc_table => {type    => SCALAR|UNDEF,
                                               default => "TEST"},
                              }
                          );

    my $domain   = $params{domain};   # QC Domain to log in to
    my $project  = $params{project};  # QC Project to log in to 
    my $qc_user  = $params{qc_user};  # QC User ID
    my $qc_pwd   = $params{qc_pwd};   # QC Password   
    my $qc_table = $params{qc_table}; # QC table name from which to obtain lists

    #--------------------------------------------------
    # Put something in so password doesn't get ignored
    #--------------------------------------------------
    if (!$qc_pwd) {
         $qc_pwd = " ";
    }

    $Trace->entry(msg => "tw_get_value_lists called with the following " .
                         "parameters: \n" .
                         "QC domain:   $domain\n" .
                         "QC project:  $project\n" .
                         "QC user:     $qc_user\n" .
                         "QC table:    $qc_table");

    my %hash;        # Hash containing the test ID (key) and test name (value)
    my @line_data;   # Array of tab-delimited lines containing test ID/test name
    my $tst_count;   # Number of tests returned from QC
    my @qc_data;     # Array of QC login data to pass to the socket server

    $tst_count = 0;
   
    #-----------------------------------------
    # Create a socket to talk to the QC proxy
    #-----------------------------------------
    my $sock = _create_qc_socket();

    #---------------------------------------------
    # Send the request to get tests to the server
    #---------------------------------------------
    print $sock GET_VAL_LISTS_REQ;

    push(@qc_data, "$domain\n");
    push(@qc_data, "$project\n");
    push(@qc_data, "$qc_user\n");
    push(@qc_data, "$qc_pwd\n");
    push(@qc_data, "$qc_table\n");
    push(@qc_data, TERM_STRING);

    print $sock @qc_data;
    $Trace->trace(msg => "QC login data sent to server");

    #------------------------------------------
    # Read the lists of values from the server
    #------------------------------------------
    my $ret_val = _read_sock(sock => $sock);

    if ($ret_val == -1) {
        $Trace->fatal(msg => "Error occurred while trying to read data from " .
                             "socket during get value lists request");
    }

    my @lines = @$ret_val;

    my $done     = 0;

    my ($key, $label_name, $line, $field_count, $list_name, $field_type, 
        $num_values);

    $field_count = 0;
    $done        = 0;

    #---------------------------
    # Add each list to the hash
    #---------------------------
    while (not $done) {
        $line = shift @lines;
        chomp($line);

        if ($line ne TERM_STRING) {
            @line_data  = split("\t", $line);
            if (substr($line_data[0], 0, 6) ne "\tError") {

                my @value_list;

                #--------------------------------------------
                # Extract the list name and number of values 
                #--------------------------------------------
                $key        = shift @line_data;
                $label_name = shift @line_data;
                $field_type = shift @line_data;

                $hash{$key}{"Label Name"} = $label_name;
                $hash{$key}{"Type"}       = $field_type;

                #----------------------------------------------
                # If this field is a list or multi-select list
                #----------------------------------------------
                if (($field_type eq "list") or ($field_type eq "multi")) {
                    $num_values = shift @line_data;

                    #--------------------------------------------
                    # If this is not the Functional Subarea data
                    #--------------------------------------------
                    if ($key ne TS_FUNC_SUBAREA) {

                        #--------------------------------------
                        # Read the values for the current list
                        #--------------------------------------
                        for (my $cnt = 0; $cnt < $num_values; $cnt++) {
                            $line = shift @lines;
                            chomp($line);
                            push (@value_list, $line);
                        } 

                        $hash{$key}{"Values"} = \@value_list; 

                    } else {

                        my $num_lists = $num_values;

                        #-------------------------------------------------
                        # Process each of the Functional Subarea sublists
                        #-------------------------------------------------
                        for (my $cnt = 0; $cnt < $num_lists; $cnt++) {
                            $line = shift @lines;
                            chomp($line);
                            @line_data  = split("\t", $line);
                            $list_name  = shift @line_data;
                            $num_values = shift @line_data;

                            #--------------------------------------
                            # Read the values for the current list
                            #--------------------------------------
                            my @value_sublist;
                            for (my $item_cnt = 0; $item_cnt < $num_values;
                                 $item_cnt++) {
                                $line = shift @lines;
                                chomp($line);
                                push (@value_sublist, $line);
                            }
                            if ($num_values > 0) {
                                $hash{$key}{"Values"}{$list_name} = 
                                    \@value_sublist;
                            } else {
                                $hash{$key}{"Values"}{$list_name} = "";
                            }
                        }
                    }
                }
                $field_count++;
                
            #--------------------------------------------------------------
            # An error occurred - extract the error code and error message
            #--------------------------------------------------------------
            } else {
                shift @line_data;
                my $err_code = shift @line_data;
                shift @line_data;
                my $err_msg  = shift @line_data;
                $Trace->warn(msg => "An error occurred trying to read "
                             . "value lists from QC\nError code: " .
                             $err_code . "\nError message: " . $err_msg);
            }
        } else {
            $done = 1;
        } 
    }

    $Trace->debug(msg => "There were " . $field_count . " fields returned from "
                  . "QC");

    $Trace->exit(msg => "Value lists returned from QC: " . Dumper \%hash);
    close($sock);

    return \%hash;

}

=pod

=head2 B<tw_get_testset_names>

=over 4

=item Description

This routine gets a list of all test sets in the specified path.

=item Arguments 

 testset_path => Fully-qualified path of the folder containing the
                 test sets to retrieve (Required)

 domain       => Name of the QC domain to use when obtaining the
                 list of test sets (Required)

 project      => Name of the QC project to use (Required)

 qc_user      => Name of the QC user (Optional)

 qc_pwd       => Password for qc_user (Optional)

=item Returns 

Returns an array reference containing a list of of all test sets retrieved.  -1 is returned if an error occurred. 

=item Example

   my $aref = tw_get_testset_names(domain       => "ONTAP",
                                   project      => "Boilermaker_0",
                                   testset_path => qw(Root\regression\Core\WAFL\7mode\WAFL\automated_regression\BAT)
                                   ); 

=back

=cut

#------------------------------------------------------------------------------
#
# tw_get_testset_names
#
# Input:
#    testset_path -  The fully-qualified path of the folder containing
#                    the test set names to retrieve.  Individual folder names 
#                    should be delimited by a backslash (\) character.  
#                    If "\Root" is excluded from the path, it is prepended.  At 
#                    least one subfolder name (after \Root) is required.
#                    (Required)
#
#     domain        - QC domain (Required)
#
#     project       - QC project (Required)
#
#     qc_user       - QC user name 
#
#     qc_pwd        - QC password
#
#     The values for domain, project, qc_user, and qc_pwd are used for
#     logging in to QC.  Domain and project are required.
#
# Output:
#     None
#
# Return value:
#     aref - Reference to an array containing test set names (only)
#            -1 is returned if errors occurred
#
#------------------------------------------------------------------------------
sub tw_get_testset_names {

    my %params = validate (
                              @_,
                              {   
                                  testset_path => {type    => SCALAR},
                                  domain       => {type    => SCALAR},
                                  project      => {type    => SCALAR},
                                  qc_user      => {type    => SCALAR|UNDEF,
                                                   default => QC_DEFAULT_USER},
                                  qc_pwd       => {type    => SCALAR|UNDEF,
                                                   default => QC_DEFAULT_PASS},
                              }
                          );

    #--------------------
    # Process parameters
    #--------------------
    my $testset_path = $params{testset_path};
    my $domain       = $params{domain};
    my $project      = $params{project};
    my $qc_user      = $params{qc_user};
    my $qc_pwd       = $params{qc_pwd};

    #--------------------------------------------------
    # Put something in so password doesn't get ignored
    #--------------------------------------------------
    if (!$qc_pwd) {
         $qc_pwd = " ";
    }

    $Trace->entry(msg => "Generating test set name list with test set path" .
                         Dumper $testset_path);

    my @lines = ();

    #------------------------------------------------------------------------
    # Replace any forward slashes with back slashes in the testset path name
    #------------------------------------------------------------------------
    $testset_path =~ s/\//\\/g;

    #--------------------------------------------
    # If testset_path starts with "\", remove it
    #--------------------------------------------
    if (substr($testset_path, 0, 1) eq "\\") {
        $testset_path = substr($testset_path, 1);
    }

    #---------------------------------------------------------------
    # Split the test set into an array of folders and test set name 
    #---------------------------------------------------------------
    my @folders = split(/\\/, $testset_path);
    my $num_folders = @folders;

    #-------------------------------------------------
    # There must be at least one subfolder after Root
    #-------------------------------------------------
    if (($num_folders <= 1) && ($folders[0] != '\Root')) {
        $Trace->fatal(msg => "There must be at least one subfolder included "
                             . "in the test set name", result => "CONFIG");
        return -1;
    }

    #-----------------------------------------------
    # Check the folder names for invalid characters
    # Invalid characters include: \^*
    #-----------------------------------------------
    foreach my $folder (@folders) {
        if ($folder =~ /[\\\^\*]+/) {
            $Trace->fatal(msg => 'Invalid character \^* contained in test ' .
                                 'set folder name ' . $folder,
                          result => "FAIL"); 
            return -1;
        } 
    }

    #----------------------------------------------------
    # If "Root\" is not included in the path, prepend it
    #----------------------------------------------------
    if (substr($testset_path, 0, 5) ne ROOT_NODE) {
        $testset_path = ROOT_NODE . $testset_path;
    }   

    #--------------------------------------------------------------------
    # Add a newline character to the end of the testset name if there is 
    # not one already there
    #--------------------------------------------------------------------
    my $str_len = length($testset_path);

    if (substr($testset_path, $str_len - 1, 1) ne "\n") { 
        $testset_path .= "\n";
    }

    #--------------------------------------
    # Add the QC login data to the request
    #--------------------------------------
    push (@lines, "$domain\n");
    push (@lines, "$project\n");
    push (@lines, "$qc_user\n");
    push (@lines, "$qc_pwd\n");

    #-------------------------------------
    # Add the testset path to the request
    #-------------------------------------
    push (@lines, $testset_path);

    #----------------------------------------------
    # Add the termination string to the line array
    #----------------------------------------------
    push(@lines, TERM_STRING);

    #-----------------------------------------
    # Create a socket to talk to the QC proxy
    #-----------------------------------------
    my $sock = _create_qc_socket();

    #----------------------------------------------------
    # Send the request to get the list of test set names 
    #----------------------------------------------------
    print $sock GET_TESTSET_NAME_REQ;

    $Trace->debug(msg => "Smokehouse get test set name: " .  Dumper \@lines);

    #-------------------------------- 
    # Send the request to the server
    #-------------------------------- 
    print $sock @lines;

    #-------------------------------------------------
    # Read the list of test set names from the server
    #-------------------------------------------------
    my $done     = 0;
    my @ts_names = ();
    my $ts_count = 0;

    $Trace->trace(msg => "Reading data from QC server");
    my $ret_val = _read_sock(sock => $sock, timeout => 600);

    if ($ret_val == -1) {
        $Trace->fatal(msg => "Error occurred while trying to read data from " .
                             "socket during get test set names request".
                             " $!");
    }

    @lines = @$ret_val;

    #--------------------------------------
    # Process the response from the server 
    #--------------------------------------
    foreach my $line (@lines) {
        chomp($line);
        if ($line ne TERM_STRING) {
            if (substr($line, 0, 5) ne "Error") {

                push (@ts_names, $line);
                $ts_count++;
            
            #----------------------------------------------------------
            # An error occurred - extract the error code and error 
            # message
            #----------------------------------------------------------
            } else {
                my @line_data  = split("\t", $line);
                shift @line_data;
                my $err_code = shift @line_data;
                shift @line_data;
                my $err_msg  = shift @line_data;
                $Trace->fatal(msg => "An error occurred trying to " .
                                     "retrieve test set names\n" .
                                     "Error code: " .  $err_code . 
                                     "\nError message: " . $err_msg,
                              result => "FAIL");
            }
        } else {   

            #---------------------------------------
            # Finished reading data from the server
            #---------------------------------------
            $done = 1;
        }
    } 

    close($sock);

    $Trace->exit(msg => "Test set test retrieval complete");

    return \@ts_names; 

}

=pod

=head2 B<tw_get_all_testsets>

=over 4

=item Description

This routine gets a list of all test sets in the Test Lab module for a
project. 

=item Arguments 

 domain       => Name of the QC domain to use when obtaining the
                 list of test sets (Required)

 project      => Name of the QC project to use (Required)

 qc_user      => Name of the QC user (Optional)

 qc_pwd       => Password for qc_user (Optional)

=item Returns 

Returns an array reference containing a list of of all test sets retrieved.  -1 is returned if an error occurred. 

=item Example

   my $aref = tw_all_testsets(domain       => "ONTAP",
                              project      => "Boilermaker_0",
                             ); 

=back

=cut

#------------------------------------------------------------------------------
#
# tw_get_all_testsets
#
# Input:
#     domain        - QC domain (Required)
#
#     project       - QC project (Required)
#
#     qc_user       - QC user name 
#
#     qc_pwd        - QC password 
#
#     The values for domain, project, qc_user, and qc_pwd are used for
#     logging in to QC.  Domain and project are required.
#
# Output:
#     None
#
# Return value:
#     aref - Reference to an array containing test set paths/names
#            -1 is returned if errors occurred
#
#------------------------------------------------------------------------------
sub tw_get_all_testsets {

    my %params = validate (
                              @_,
                              {   
                                  domain       => {type    => SCALAR},
                                  project      => {type    => SCALAR},
                                  qc_user      => {type    => SCALAR|UNDEF,
                                                   default => QC_DEFAULT_USER},
                                  qc_pwd       => {type    => SCALAR|UNDEF,
                                                   default => QC_DEFAULT_PASS},
                              }
                          );

    #--------------------
    # Process parameters
    #--------------------
    my $domain  = $params{domain};
    my $project = $params{project};
    my $qc_user = $params{qc_user};
    my $qc_pwd  = $params{qc_pwd};

    $Trace->entry(msg => "Retrieving list of all test sets");

    my $aref = tw_get_testset_names(testset_path => "",
                                    domain       => $domain,
                                    project      => $project,
                                    qc_user      => $qc_user,
                                    qc_pwd       => $qc_pwd);

    $Trace->exit(msg => "Test set name retrieval complete");

    return $aref; 

}

=pod

=head2 B<tw_get_testset_tests>

=over 4

=item Description

This routine queries Quality Center for the list of tests that are contained in 
the specified test set.  The fully-qualified path for the test set name is 
passed in.  A list of all tests contained in the test set is returned.

=item Arguments 

 testset_name => <path_to_test_set> (Required)

 domain       => Name of the QC domain to use when obtaining test 
                 list (Required)

 project      => Name of the QC project to use (Required)

 qc_user      => Name of the QC user (Optional)

 qc_pwd       => Password for qc_user (Optional)

=item Returns 

Returns a hash reference containing a list of filtered tests in the form:

    <test_case_id> => {
                       'project'       => <project_name>,
                       'test_name'     => <test_name>, 
                       'nate_script'   => <nate_script_path>, 
                       'nate_params'   => '<nate_params>',
                       'test_owner'    => <test_owner>,
                       'ontap_support' => <ontap_support>,
                       'func_area'     => <functional_area>,
                       'func_subarea'  => <functional_subarea>
                       'lab_params'    => <lab_params>
                      }  

    <project_name> is the fully-qualified path name for the folder 
    containing the test (e.g., "Subject/folder1/folder2"),

    <test_name> is the name of the test (e.g., "mhost_no_key"),

    <nate_script_path> is the relative p4 path containing the NATE 
    .thpl or .stest file (e.g., "nblade/cifs/auth/sanity.thpl"),

    <nate_params> are the NATE params that are defined for the test 
    case in the Test Plan module (e.g., "TESTCASE=<testcase_name>").  
    Note that Test Lab NATE params are not included,

    <test_owner> is the name of the person responsible for 
    implementing the test,

    <ontap_support> identifies the ONTAP mode(s) for which this test
    case applies,

    <functional_area> is the Functional Area to which the test belongs,

    <functional_subarea> is the Functional Subarea to which the test belongs,

    <lab_params> are the NATE params that are defined for the test
    instance in the Test Lab module.  

=item Example

    # Get a list of tests in the specified testset from Quality 
    # Center
    $href = tw_get_testset_tests(testset_name => qw(\Root\regression\Core\WAFL\7mode\WAFL\automated_regression\MS6\BAT\WAFL001_Space_Management),
                                 domain       => "ONTAP",
                                 project      => "boilermaker_0");

=back

=cut

#------------------------------------------------------------------------------
#
# tw_get_testset_tests
#
# Input:
#    testset_name -  The name of the test set, including the fully-qualified
#                    path for the folder to contain the test set.  Individual
#                    folder names should be delimited by a backslash (\)
#                    character.  If "\Root" is excluded from the path, it is
#                    prepended.  At least one subfolder name (after \Root) is
#                    required.
#
#     domain        - QC domain
#
#     project       - QC project
#
#     qc_user       - QC user name 
#
#     qc_pwd        - QC password 
#
#     The values for domain, project, qc_user, and qc_pwd are used for
#     logging in to QC.  Domain and project are required.
#
# Output:
#     None
#
# Return value:
#     test_href - Hash reference containing test ID and a hash of test
#                 data as follows:
#
#     <test_case_id> => {
#                        'project'       => <project_name>,
#                        'test_name'     => <test_name>, 
#                        'nate_script'   => <nate_script_path>, 
#                        'nate_params'   => '<nate_params>',
#                        'test_owner'    => <test_owner>,
#                        'ontap_support' => <ontap_support>,
#                        'lab_params'    => '<lab_params>'
#                       }  
#
#                 -1 is returned if test set name is invalid
#
#------------------------------------------------------------------------------
sub tw_get_testset_tests {

    my %params = validate (
                              @_,
                              {   
                                  testset_name => {type    => SCALAR},
                                  domain       => {type    => SCALAR},
                                  project      => {type    => SCALAR},
                                  qc_user      => {type    => SCALAR|UNDEF,
                                                   default => QC_DEFAULT_USER},
                                  qc_pwd       => {type    => SCALAR|UNDEF,
                                                   default => QC_DEFAULT_PASS},
                              }
                          );

    #--------------------
    # Process parameters
    #--------------------
    my $testset_name = $params{testset_name};
    my $domain       = $params{domain};
    my $project      = $params{project};
    my $qc_user      = $params{qc_user};
    my $qc_pwd       = $params{qc_pwd};

    my %hash;        # Hash containing the test ID (key) and test name (value)
    my @line_data;   # Array of tab-delimited lines containing test ID/test name
    my $tst_count;   # Number of tests returned from QC

    $tst_count = 0;

    #--------------------------------------------------
    # Put something in so password doesn't get ignored
    #--------------------------------------------------
    if (!$qc_pwd) {
         $qc_pwd = " ";
    }

    $Trace->entry(msg => "Getting testset with test set name " .
                         Dumper $testset_name);

    my @lines = ();
    my ($key, $value);

    #------------------------------------------------------------------------
    # Replace any forward slashes with back slashes in the testset path name
    #------------------------------------------------------------------------
    $testset_name =~ s/\//\\/g;

    #--------------------------------------------
    # If testset_name starts with "\", remove it
    #--------------------------------------------
    if (substr($testset_name, 0, 1) eq "\\") {
        $testset_name = substr($testset_name, 1);
    }

    #---------------------------------------------------------------
    # Split the test set into an array of folders and test set name 
    #---------------------------------------------------------------
    my @folders = split(/\\/, $testset_name);
    my $ts_name = pop(@folders);

    my $num_folders = @folders;

    #-------------------------------------------------
    # There must be at least one subfolder after Root
    #-------------------------------------------------
    if ($num_folders <= 1) {
        $Trace->fatal(msg => "There must be at least one subfolder included "
                             . "in the test set name", result => "CONFIG");
        return -1;
    }

    #-----------------------------------------------
    # Check the folder names for invalid characters
    # Invalid characters include: \^*
    #-----------------------------------------------
    foreach my $folder (@folders) {
        if ($folder =~ /[\\\^\*]+/) {
            $Trace->fatal(msg => 'Invalid character \^* contained in test ' .
                                 'set folder name ' . $folder,
                          result => "FAIL"); 
            return -1;
        } 
    }

    #------------------------------------------------
    # Check the test set name for invalid characters
    # Invalid characters include: \^,"*
    #------------------------------------------------
    if ($ts_name =~ /[\\\^\,\"\*]+/) {
        $Trace->fatal(msg => 'Invalid character \^,"* contained in test set ' .
                             'name ' .  $ts_name,
                      result => "FAIL"); 
        return -1;
    } else {

        #----------------------------------------------------
        # If "Root\" is not included in the path, prepend it
        #----------------------------------------------------
        if (substr($testset_name, 0, 5) ne ROOT_NODE) {
            $testset_name = ROOT_NODE . $testset_name;
        }   

        #--------------------------------------------------------------------
        # Add a newline character to the end of the testset name if there is 
        # not one already there
        #--------------------------------------------------------------------
        my $str_len = length($testset_name);

        if (substr($testset_name, $str_len - 1, 1) ne "\n") { 
            $testset_name .= "\n";
        }

        #--------------------------------------
        # Add the QC login data to the request
        #--------------------------------------
        push (@lines, "$domain\n");
        push (@lines, "$project\n");
        push (@lines, "$qc_user\n");
        push (@lines, "$qc_pwd\n");

        #--------------------------------
        # Add the testset to the request
        #--------------------------------
        push (@lines, $testset_name);

        #----------------------------------
        # Add the termination string array
        #----------------------------------
        push(@lines, TERM_STRING);

        #-----------------------------------------
        # Create a socket to talk to the QC proxy
        #-----------------------------------------
        my $sock = _create_qc_socket();

        #------------------------------------------------
        # Send the request to filter on Smokehouse tests
        #------------------------------------------------
        print $sock GET_TESTSET_REQ;

        $Trace->debug(msg => "Test set retrieval message: " . 
                      Dumper \@lines);

        #-------------------------------- 
        # Send the request to the server
        #-------------------------------- 
        print $sock @lines;

        #----------------------------------------
        # Read the list of tests from the server
        #----------------------------------------
        $Trace->trace(msg => "Reading data from QC server");

        my $ret_val = _read_sock(sock => $sock);

        if ($ret_val == -1) {
            $Trace->fatal(msg => "Error occurred while trying to read data " .
                                 "from socket during get test set tests " .
                                 "request");
        }

        my @lines = @$ret_val;

        my ($err_code, $err_msg, $href);

        $href = _process_test_cases(\@lines, 1, \$err_code, \$err_msg);

        if ($href == -1) {
            $Trace->fatal(msg => "An error occurred trying to " .
                                 "retrieve test set tests\n" .
                                 "Error code: " .  $err_code . 
                                 "\nError message: " . $err_msg,
                                 result => "FAIL");
        }
        close($sock);
        $Trace->exit(msg => "Test set test retrieval from QC complete");

        return $href;
    }

}

=pod

=head2 B<tw_get_by_instance_id>

=over 4

=item Description

This routine queries Quality Center for the list of test instance IDs that are
passed in.  A reference to a hash containing the data for all specified test 
instances is returned.

=item Arguments 

 test_inst_id_aref => Array reference of test instance IDs (Required)

 domain            => Name of the QC domain to use when obtaining test 
                      list (Required)

 project           => Name of the QC project to use (Required)

 qc_user           => Name of the QC user (Optional)

 qc_pwd            => Password for qc_user (Optional)

=item Returns 

Returns a hash reference containing a list of test instance data in the form:

    <test_case_id> => {
                       'project'       => <project_name>,
                       'test_name'     => <test_name>, 
                       'nate_script'   => <nate_script_path>, 
                       'nate_params'   => '<nate_params>',
                       'test_owner'    => <test_owner>,
                       'ontap_support' => <ontap_support>,
                       'func_area'     => <functional_area>,
                       'func_subarea'  => <functional_subarea>
                       'lab_params'    => <lab_params>
                      }  

    <project_name> is the fully-qualified path name for the folder 
    containing the test (e.g., "Subject/folder1/folder2"),

    <test_name> is the name of the test (e.g., "mhost_no_key"),

    <nate_script_path> is the relative p4 path containing the NATE 
    .thpl or .stest file (e.g., "nblade/cifs/auth/sanity.thpl"),

    <nate_params> are the NATE params that are defined for the test 
    case in the Test Plan module (e.g., "TESTCASE=<testcase_name>").  
    Note that Test Lab NATE params are not included,

    <test_owner> is the name of the person responsible for 
    implementing the test,

    <ontap_support> identifies the ONTAP mode(s) for which this test
    case applies,

    <functional_area> is the Functional Area to which the test belongs,

    <functional_subarea> is the Functional Subarea to which the test belongs,

    <lab_params> are the NATE params that are defined for the test
    instance in the Test Lab module.  

=item Example

    # Get the test data for the specified test instances 
    $href = tw_get_by_instance_id(test_inst_id_aref => \@test_inst_ids),
                                  domain            => "ONTAP",
                                  project           => "boilermaker_0");

=back

=cut

#------------------------------------------------------------------------------
#
# tw_get_by_instance_id
#
# Input:
#    test_inst_id_aref - Array reference containing a list of test IDs for 
#                        which to return test data. 
#
#    domain            - QC domain
#
#    project           - QC project
#
#    qc_user           - QC user name 
#
#    qc_pwd            - QC password 
#
#    The values for domain, project, qc_user, and qc_pwd are used for
#    logging in to QC.  Domain and project are required.
#
# Output:
#    None
#
# Return value:
#    test_href - Hash reference containing test ID and a hash of test
#                data as follows:
#
#     <test_case_id> => {
#                        'project'       => <project_name>,
#                        'test_name'     => <test_name>, 
#                        'nate_script'   => <nate_script_path>, 
#                        'nate_params'   => '<nate_params>',
#                        'test_owner'    => <test_owner>,
#                        'ontap_support' => <ontap_support>,
#                        'lab_params'    => '<lab_params>'
#                       }  
#
#                 -1 is returned if test set name is invalid
#
#------------------------------------------------------------------------------
sub tw_get_by_instance_id {

    $Trace->entry();

    my %params = validate (
                              @_,
                              {   
                                  test_inst_id_aref => {type    => ARRAYREF},
                                  domain            => {type    => SCALAR},
                                  project           => {type    => SCALAR},
                                  qc_user           => {type    => SCALAR|UNDEF,
                                                        default => QC_DEFAULT_USER},
                                  qc_pwd            => {type    => SCALAR|UNDEF,
                                                        default => QC_DEFAULT_PASS},
                              }
                          );

    #--------------------
    # Process parameters
    #--------------------
    my $test_inst_id_aref = $params{test_inst_id_aref};
    my $domain            = $params{domain};
    my $project           = $params{project};
    my $qc_user           = $params{qc_user};
    my $qc_pwd            = $params{qc_pwd};

    my @line_data;   # Array of tab-delimited data returned from the QC proxy 
    my $tst_count;   # Number of tests returned from QC

    $tst_count = 0;

    #--------------------------------------------------
    # Put something in so password doesn't get ignored
    #--------------------------------------------------
    if (!$qc_pwd) {
         $qc_pwd = " ";
    }

    $Trace->debug(msg => "Getting test data for specified list of test " .
                         "instances " . Dumper $test_inst_id_aref);

    my @lines = ();

    #--------------------------------------
    # Add the QC login data to the request
    #--------------------------------------
    push (@lines, "$domain\n");
    push (@lines, "$project\n");
    push (@lines, "$qc_user\n");
    push (@lines, "$qc_pwd\n");

    my $test_cnt = @{$test_inst_id_aref};

    if ($test_cnt < 1) {
        $Trace->fatal(msg => "There were no test instance IDs passed to "
                             . "tw_get_by_instance_id", result => "CONFIG");
    }

    #------------------------------------------
    # Add the test instance IDs to the request
    #------------------------------------------
    foreach my $test_inst_id (@{$test_inst_id_aref}) {
       push (@lines, $test_inst_id . "\n");
    }

    #----------------------------------
    # Add the termination string array
    #----------------------------------
    push(@lines, TERM_STRING);

    #-----------------------------------------
    # Create a socket to talk to the QC proxy
    #-----------------------------------------
    my $sock = _create_qc_socket();

    #--------------------------------------------
    # Send the request to get test instance data 
    #--------------------------------------------
    print $sock GET_TEST_INST_DATA_REQ;

    $Trace->debug(msg => "Test instance data retrieval message: " . 
                  Dumper \@lines);

    #-------------------------------- 
    # Send the request to the server
    #-------------------------------- 
    print $sock @lines;

    #----------------------------------------
    # Read the list of tests from the server
    #----------------------------------------
    $Trace->trace(msg => "Reading data from QC server");

    my $ret_val = _read_sock(sock => $sock);

    if ($ret_val == -1) {
        $Trace->fatal(msg => "Error occurred while trying to read data " .
                             "from socket during get test instance by ID " .
                             "request");
    }

    my @ret_lines = @$ret_val;

    my ($err_code, $err_msg, $href);

    $href = _process_test_cases(\@ret_lines, 1, \$err_code, \$err_msg);

    if ($href == -1) {
        $Trace->fatal(msg => "An error occurred trying to " .
                             "retrieve test instance test data\n" .
                             "Error code: " .  $err_code . 
                             "\nError message: " . $err_msg,
                             result => "FAIL");
    }
    close($sock);
    $Trace->exit(msg => "Test instance test data retrieval from QC complete");

    return $href;

}

=pod

=head2 B<tw_create_test_run>

=over 4

=item Description

This routine creates a test run for every specified test case in the specified
test set.  If a test does not exist in the test set, it is added to the test
set and a run is created for it.

=item Arguments 

 testset_name => <$path_to_test_set> (Required)

 run_prefix   => Prefix to use for naming individual runs.  The 
                 current date, time, and test number are appended 
                 to the prefix 
                 (e.g., <prefix>_yyyy-mm-dd_hhmmss_Test_x).  
                 (Optional; default = "Autotest_Run_")

 retries      => Number of times to retry when a record is locked.  
                 The wait time between retries is 1/2 second.
                 (Optional; default = 4);

 domain       => Name of the QC domain to use when obtaining test 
                 list (Required)

 project      => Name of the QC project to use (Required)

 qc_user      => Name of the QC user (Optional; default = "autotest")

 qc_pwd       => Password for qc_user (Optional; default = <password 
                 for autotest>)

=item Returns 

Returns a hash reference containing a list of created runs and their
corresponding test information in the form 

    <test_case_id> => {
                       'test_name' => <test_name>, 
                       'run_id'    => <run_id>, 
                       'run_name'  => <run_name>,
                      }  

    <test_name> is the name of the test (e.g., "mhost_no_key"),

    <run_id> is the ID assigned to the run by Quality Center.  
    This value can be used to post results to QC when executing 
    ntest from the command line.

    <run_name> is the name assigned to the run 
    (e.g., <prefix>_yyyy-dd-mm_hhmmss_Test_x)

=item Example

    # Create a set of runs for the specified tests
    $href = tw_get_testset_tests(testset_name => qw(\Root\regression\Core\WAFL\7mode\WAFL\automated_regression\MS6\BAT\WAFL001_Space_Management),
                                 run_prefix   => "boilerms6_smokeHouse_"
                                 domain       => "ONTAP",
                                 project      => "boilermaker_0");

=back

=cut

#------------------------------------------------------------------------------
#
# tw_create_test_run
#
# Input:
#     testset_name - The name of the test set, including the fully-qualified
#                    path for the folder to contain the test set.  Individual
#                    folder names should be delimited by a backslash (\)
#                    character.  If "\Root" is excluded from the path, it is
#                    prepended.  At least one subfolder name (after \Root) is
#                    required.
#
#     run_prefix   - Prefix to use for the test runs being created.
#                    "yyyy-mm-dd_hhmmss_Test_x" will be appended to the prefix for
#                    each run created.  (default - "Autotest_Run_")
#
#     test_names   - Reference to a hash containing a list of test names and their
#                    corresponding test IDs.  Hash key is the test ID and hash
#                    value is the test name.
#
#     domain       - QC domain
#
#     project      - QC project
#
#     qc_user      - QC user name (default = "autotest")
#
#     qc_pwd       - QC password (default = "")
#
#     The values for domain, project, qc_user, and qc_pwd are used for
#     logging in to QC.  Domain and project are required.
#
# Output:
#     None
=pod

=head2 B<tw_create_test_run>

=over 4

=item Description

This routine creates a test run for every specified test case in the specified
test set.  If a test does not exist in the test set, it is added to the test
set and a run is created for it.

=item Arguments 

 testset_name => <$path_to_test_set> (Required)

 run_prefix   => Prefix to use for naming individual runs.  The 
                 current date, time, and test number are appended 
                 to the prefix 
                 (e.g., <prefix>_yyyy-mm-dd_hhmmss_Test_x).  
                 (Optional; default = "Autotest_Run_")

 retries      => Number of times to retry when a record is locked.  
                 The wait time between retries is 1/2 second.
                 (Optional; default = 4);

 domain       => Name of the QC domain to use when obtaining test 
                 list (Required)

 project      => Name of the QC project to use (Required)

 qc_user      => Name of the QC user (Optional; default = "autotest")

 qc_pwd       => Password for qc_user (Optional; default = <password 
                 for autotest>)

=item Returns 

Returns a hash reference containing a list of created runs and their
corresponding test information in the form 

    <test_case_id> => {
                       'test_name' => <test_name>, 
                       'run_id'    => <run_id>, 
                       'run_name'  => <run_name>,
                      }  

    <test_name> is the name of the test (e.g., "mhost_no_key"),

    <run_id> is the ID assigned to the run by Quality Center.  
    This value can be used to post results to QC when executing 
    ntest from the command line.

    <run_name> is the name assigned to the run 
    (e.g., <prefix>_yyyy-dd-mm_hhmmss_Test_x)

=item Example

    # Create a set of runs for the specified tests
    $href = tw_get_testset_tests(testset_name => qw(\Root\regression\Core\WAFL\7mode\WAFL\automated_regression\MS6\BAT\WAFL001_Space_Management),
                                 run_prefix   => "boilerms6_smokeHouse_"
                                 domain       => "ONTAP",
                                 project      => "boilermaker_0");

=back

=cut

#------------------------------------------------------------------------------
#
# tw_create_test_run
#
# Input:
#     testset_name - The name of the test set, including the fully-qualified
#                    path for the folder to contain the test set.  Individual
#                    folder names should be delimited by a backslash (\)
#                    character.  If "\Root" is excluded from the path, it is
#                    prepended.  At least one subfolder name (after \Root) is
#                    required.
#
#     run_prefix   - Prefix to use for the test runs being created.
#                    "yyyy-mm-dd_hhmmss_Test_x" will be appended to the prefix for
#                    each run created.  (default - "Autotest_Run_")
#
#     test_names   - Reference to a hash containing a list of test names and their
#                    corresponding test IDs.  Hash key is the test ID and hash
#                    value is the test name.
#
#     domain       - QC domain
#
#     project      - QC project
#
#     qc_user      - QC user name (default = "autotest")
#
#     qc_pwd       - QC password (default = "")
#
#     The values for domain, project, qc_user, and qc_pwd are used for
#     logging in to QC.  Domain and project are required.
#
# Output:
#     None
#
# Return value:
#     test_href - Hash reference containing test ID, test name, run ID, and
#                 run name; -1 is returned if test set name is invalid
#
#------------------------------------------------------------------------------
sub tw_create_test_run {

    my %params = validate (
                              @_,
                              {   
                                  testset_name => {type    => SCALAR},
                                  run_prefix   => {type    => SCALAR,
                                                   default => "Autotest_Run_"},
                                  retries      => {type    => SCALAR,
                                                   default => 4},
                                  test_names   => {type    => HASHREF},
                                  domain       => {type    => SCALAR},
                                  project      => {type    => SCALAR},
                                  qc_user      => {type    => SCALAR|UNDEF,
                                                   default => "autotest"},
                                  qc_pwd       => {type    => SCALAR|UNDEF,
                                                   default => "ptcapi"},
                              }
                          );

    #--------------------
    # Process parameters
    #--------------------
    my $testset_name = $params{testset_name};
    my $run_prefix   = $params{run_prefix};
    my $retries      = $params{retries};
    my $hashref      = $params{test_names};
    my $domain       = $params{domain};
    my $project      = $params{project};
    my $qc_user      = $params{qc_user};
    my $qc_pwd       = $params{qc_pwd};

    my %hash;        # Hash containing the test ID (key) and test name (value)
    my @line_data;   # Array of tab-delimited lines containing test ID/test name
    my $tst_count;   # Number of tests returned from QC

    $tst_count = 0;

    #--------------------------------------------------
    # Put something in so password doesn't get ignored
    #--------------------------------------------------
    if (!$qc_pwd) {
         $qc_pwd = " ";
    }

    $Trace->entry(msg => "Getting testset with test set name " .
                         Dumper $testset_name);

    my @lines = ();
    my ($key, $value);

    if (_validate_testset_name(\$testset_name) == 0) {

        #--------------------------------------
        # Add the QC login data to the request
        #--------------------------------------
        push (@lines, "$domain\n");
        push (@lines, "$project\n");
        push (@lines, "$qc_user\n");
        push (@lines, "$qc_pwd\n");

        #--------------------------------
        # Add the testset to the request
        #--------------------------------
        push (@lines, $testset_name);

        #---------------------------------------------
        # Add the test run name prefix to the request
        #---------------------------------------------
        push (@lines, "$run_prefix\n");

        #------------------------------------------
        # Add the number of retries to the request
        #------------------------------------------
        push (@lines, "$retries\n");

        #-----------------------------------
        # Add each test name to the request
        #-----------------------------------
        foreach my $key (sort (keys %$hashref)) {

            #-----------------------------------------------------------------
            # Make sure the test name does not contain any invalid characters
            # Invalid characters include:  \/:"?<>|*%'
            #-----------------------------------------------------------------
            if ($hashref->{$key} =~ /[\\\/\:\"\?\<\>\|\*\%\']+/) {
                $Trace->fatal(msg => "Invalid character " . qw(\/:"?<>|*%') .
                                     "contained in test name " .
                                     $hashref->{$key}, result => "FAIL"); 
                return -1;
            } else {
                push(@lines, "$key\t$hashref->{$key}\n");
            }
        }  

        #----------------------------------
        # Add the termination string array
        #----------------------------------
        push(@lines, TERM_STRING);

        #-----------------------------------------
        # Create a socket to talk to the QC proxy
        #-----------------------------------------
        my $sock = _create_qc_socket();

        #------------------------------------------------
        # Send the request to filter on Smokehouse tests
        #------------------------------------------------
        print $sock CREATE_TEST_RUN_REQ;

        $Trace->debug(msg => "Test run creation message: " . 
                      Dumper \@lines);

        #-------------------------------- 
        # Send the request to the server
        #-------------------------------- 
        print $sock @lines;

        #----------------------------------------
        # Read the list of tests from the server
        #----------------------------------------
        my $done = 0;
        my $timeout;


        #-------------------------------------------------------
        # Determine the socket read timeout value based on
        # number of retries.  Maximum timeout is 20 minutes;
        # default timeout is 90 seconds.
        #-------------------------------------------------------

        # If specified to wait forever 
        if ($retries == 0) {    
            $timeout = MAX_SOCK_TIMEOUT;  # Timeout after max retry
        # If number of retries * 1/2 sec is greater than default timeout    
        } elsif (($retries / 2) > DEF_SOCK_TIMEOUT) {
            $timeout = $retries / 2;
        # No retries specified - use default timeout
        } else {
            $timeout = DEF_SOCK_TIMEOUT;
        }
            
        while (not $done) {
            $Trace->trace(msg => "Reading data from QC server");

            my $ret_val = _read_sock(sock    => $sock,
                                     timeout => $timeout);

            if ($ret_val == -1) {
                $Trace->fatal(msg => "Error occurred while trying to read " .
                                     "data from socket during test run " .
                                     "creation request");
            }   

            my @lines = @$ret_val;

            my ($test_id, $test_name, $run_id, $run_name);

            #---------------------------
            # Add each test to the hash
            #---------------------------
            foreach my $line (@lines) {
                chomp($line);
                if ($line ne TERM_STRING) {
                    @line_data  = split("\t", $line);
                    if (substr($line_data[0], 0, 5) ne "Error") {

                        $test_id   = shift @line_data;
                        $test_name = shift @line_data;
                        $run_id    = shift @line_data;
                        $run_name  = shift @line_data;

                        $hash{$test_id}{"test_name"} = $test_name;
                        $hash{$test_id}{"run_id"}    = $run_id;
                        $hash{$test_id}{"run_name"}  = $run_name;
                        $tst_count++;
                
                    #----------------------------------------------------------
                    # An error occurred - extract the error code and error 
                    # message
                    #----------------------------------------------------------
                    } else {
                        shift @line_data;
                        my $err_code = shift @line_data;
                        shift @line_data;
                        my $err_msg  = shift @line_data;
                        $Trace->fatal(msg => "An error occurred trying to " .
                                             "retrieve test set tests\n" .
                                             "Error code: " .  $err_code . 
                                             "\nError message: " . $err_msg,
                                      result => "FAIL");
                    }
                } else {   

                    #---------------------------------------
                    # Finished reading data from the server
                    #---------------------------------------
                    $done = 1;
                }
            } 
        }

        close($sock);

    } else {   # Invalid testset name
        return -1;
    }

    $Trace->exit(msg => "Test run creation complete");

    return \%hash; 

}

#------------------------------------------------------------------------------
#
# _validate_testset_name
#
#    This function prepends "\Root" to the test set name if it is not 
#    already included, ensures that at least one subfolder name is
#    included, converts any "/" characters to "\" in the testset path, adds a
#    newline character to the end of the testset name and ensures that there 
#    are no invalid characters in any of the folder names or the test set name.
#
# Input:
#    testset_name -  The name of the test set, including the fully-qualified
#                    path for the folder to contain the test set.  Individual
#                    folder names should be delimited by a backslash (\)
#                    character.  If "\Root" is excluded from the path, it is
#                    prepended.  At least one subfolder name (after \Root) is
#                    required.
#
# Return value:
#    0 - Valid testset name
#   -1 - Invalid testset name
#
#------------------------------------------------------------------------------
sub _validate_testset_name($) 
{

    my $ts_name_ref = shift;

    my $ret_val = 0;

    #------------------------------------------------------------------------
    # Replace any forward slashes with back slashes in the testset path name
    #------------------------------------------------------------------------
    ${$ts_name_ref} =~ s/\//\\/g;

    #--------------------------------------------
    # If testset_name starts with "\", remove it
    #--------------------------------------------
    if (substr(${$ts_name_ref}, 0, 1) eq "\\") {
        ${$ts_name_ref} = substr(${$ts_name_ref}, 1);
    }

    #---------------------------------------------------------------
    # Split the test set into an array of folders and test set name 
    #---------------------------------------------------------------
    my @folders = split(/\\/, ${$ts_name_ref});
    my $ts_name = pop(@folders);

    my $num_folders = @folders;

    #-------------------------------------------------
    # There must be at least one subfolder after Root
    #-------------------------------------------------
    if ($num_folders <= 1) {
        $Trace->fatal(msg => "There must be at least one subfolder included "
                             . "in the test set name", result => "CONFIG");
        $ret_val = -1;
    }

    #-----------------------------------------------
    # Check the folder names for invalid characters
    # Invalid characters include: \^*
    #-----------------------------------------------
    foreach my $folder (@folders) {
        if ($folder =~ /[\\\^\*]+/) {
            $Trace->fatal(msg => 'Invalid character \^* contained in test ' .
                                 'set folder name ' . $folder,
                          result => "FAIL"); 
            return -1;
        } 
    }

    #------------------------------------------------
    # Check the test set name for invalid characters
    # Invalid characters include: \^,"*
    #------------------------------------------------
    if ($ts_name =~ /[\\\^\,\"\*]+/) {
        $Trace->fatal(msg => 'Invalid character \^,"* contained in test set ' .
                             'name ' .  $ts_name,
                      result => "FAIL"); 
        return -1;
    } else {

        #----------------------------------------------------
        # If "Root\" is not included in the path, prepend it
        #----------------------------------------------------
        if (substr(${$ts_name_ref}, 0, 5) ne ROOT_NODE) {
            ${$ts_name_ref} = ROOT_NODE . ${$ts_name_ref};
        }   

        #--------------------------------------------------------------------
        # Add a newline character to the end of the testset name if there is 
        # not one already there
        #--------------------------------------------------------------------
        my $str_len = length(${$ts_name_ref});

        if (substr(${$ts_name_ref}, $str_len - 1, 1) ne "\n") { 
            ${$ts_name_ref} .= "\n";
        }
    } 

    return $ret_val;
}

#------------------------------------------------------------------------------
#
# _read_sock
#
#    This function reads data from the socket until the termination string is
#    encountered and generates a timeout if the read is not complete within
#    90 seconds.
#
# Input:
#    sock    - File handle for the socket from which to read. (Required)
#    timeout - Number of seconds to wait before timing out when reading from
#              the socket. Default = 90 seconds (Optional)
#
# Return value:
#    $buf - Buffer containing data read. 
#      -1 - Error occurred 
#
#------------------------------------------------------------------------------
sub _read_sock() 
{

    my %params = validate (
                              @_,
                              {   
                                  sock    => {type    => GLOBREF},
                                  timeout => {type    => SCALAR,
                                              default => 90},
                              }
                          );

    #--------------------
    # Process parameters
    #--------------------
    my $sock    = $params{sock};
    my $timeout = $params{timeout};

    #------------------------------------------
    # Read the entire response from the server
    #------------------------------------------
    my $done = 0;
    my $ret_val;

    eval {

        local $SIG{ALRM} = sub { die "timeout" };

        alarm $timeout;
        while (not $done) {
            $Trace->trace(msg => "Reading data from QC server");
            my @lines = <$sock>;

            #---------------------------------------------------------
            # Check to see if the termination string was encountered.
            # It should be on a line by itself
            #---------------------------------------------------------
            foreach my $line (@lines) {
                chomp ($line);
                if ($line eq TERM_STRING) {
                    $done = 1;
                }
            }
            push(@$ret_val, @lines);
        }
        my $time_rem = alarm 0;
        $Trace->debug(msg => "Time remaining on timer was " . $time_rem);
    };

    alarm 0;

    if ($@) {

        if ($@ =~ /timeout/) {
            $Trace->debug(msg => "Timeout occurred while trying to read " .
                                 "socket");
        } else {
            $Trace->debug(msg => "Non-timeout failure occurred while trying " .
                                 "to read socket");
        }
        $ret_val = -1;
    }

    return $ret_val;
}

#------------------------------------------------------------------------------
#
# _process_test_cases
#
#    This function processes the test cases that are returned from the QC
#    proxy and puts the results in a hash.  "<newline>" is replaced with "\n",
#    "<tab>" is replaced with "\t", and "<NULL>" is replaced with "" in the 
#    NATE Params field as well.
#
# Input:
#
# Return value:
#    href - Reference to the hash containing the processed test cases. 
#   -1 - Error was returned from the QC proxy 
#
#------------------------------------------------------------------------------
sub _process_test_cases($$$$) 
{

    my $lines         = shift;
    my $test_set      = shift;
    my $err_code_ref  = shift;
    my $err_msg_ref   = shift;
    my $tst_count = 0;

    my (%hash, $func_area, $func_subarea, $lab_params, @line_data, $nate_params,
        $nate_script, $ontap_support, $project, $ret_val, $test_id, $test_name, 
        $test_owner);

    #---------------------------
    # Add each test to the hash
    #---------------------------
    foreach my $line (@$lines) {
        chomp($line);
        if ($line ne TERM_STRING) {
            @line_data  = split("\t", $line);
            if (substr($line_data[0], 0, 5) ne "Error") {

                $test_id       = shift @line_data;
                $test_name     = shift @line_data;
                $nate_script   = shift @line_data;
                $nate_params   = shift @line_data;
                $project       = shift @line_data;
                $test_owner    = shift @line_data;
                $ontap_support = shift @line_data;
                $func_area     = shift @line_data;
                $func_subarea  = shift @line_data;

                #----------------------------------------
                # If test lab NATE params were sent back
                #----------------------------------------
                if ($test_set) {
                    $lab_params = shift @line_data;
                }

                #------------------------------------------------------
                # Convert double backslashes to single forward slashes
                #------------------------------------------------------
                $project     =~ s/\\/\//g;  

                #----------------------------------------------------
                # Separate each param with spaces
                #----------------------------------------------------
                $nate_params   =~ s/\<NULL\>//g;
                $ontap_support =~ s/\<NULL\>//g;
                $func_area     =~ s/\<NULL\>//g;
                $func_subarea  =~ s/\<NULL\>//g;
                $hash{$test_id}{"test_name"}     = $test_name;
                $hash{$test_id}{"nate_script"}   = $nate_script; 
                $hash{$test_id}{"nate_params"}   = $nate_params;
                $hash{$test_id}{"project"}       = $project;
                $hash{$test_id}{"test_owner"}    = $test_owner;
                $hash{$test_id}{"ontap_support"} = $ontap_support;
                $hash{$test_id}{"func_area"}     = $func_area;
                $hash{$test_id}{"func_subarea"}  = $func_subarea;

                #----------------------------------------
                # If test lab NATE params were sent back
                #----------------------------------------
                if ($test_set) {
                    $lab_params =~ s/<NULL>//g;
                    $hash{$test_id}{"lab_params"} = $lab_params;
                }
                $tst_count++;
                
            #--------------------------------------------------------------
            # An error occurred - extract the error code and error message
            #--------------------------------------------------------------
            } else {
                shift @line_data;
                ${$err_code_ref} = shift @line_data;
                shift @line_data;
                ${$err_msg_ref}  = shift @line_data;
                $ret_val = -1;
            }
        }
    } 

    #----------------------------------------------------
    # If no errors occurred, return the test information
    #----------------------------------------------------
    if ($ret_val != -1) {
         $ret_val         = \%hash;
         ${$err_code_ref} = 0;
         ${$err_msg_ref}  = "";
    }

    $Trace->debug(msg => "There were " . $tst_count . " tests returned from "
                  . "QC");

    $Trace->debug(msg => "Tests returned from QC " . Dumper \%hash);

    return $ret_val;

}

1;
