# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Qtree/Setup.pm#1 $

## @summary Library of qtree setup commands
## @author archana@netapp.com, dl-test-infra@netapp.com     
## @status Shared
## @pod here


=head1 NAME

Setup - Set up multiple qtrees

=head1 SYNOPSIS

   use Qtree::Setup qw(:all);

   my $h = host($FILER);
   my $fc = connect("#rsh");

   setup_qtrees_from_string($conn, $FILER_VOL_LIST);

   setup_qtrees($conn,
                 {vol=>"aggr1",type=>"aggr",...},
                 {vol=>"vol1",type=>"flex",aggr=>"aggr1",...},
                 {vol=>"clone1",type=>"clone",parent=>"vol1",...},
                );


=head1 DESCRIPTION

Use this lib to seup different type of qtrees.

To test this lib, use Filer/setup_qtree_list.thpl


=head1 ROUTINES

=cut


package Qtree::Setup;
use strict;
use warnings;
use base qw(Exporter);
BEGIN
{
  our %EXPORT_TAGS = (
    all => [ qw(
                setup_qtrees 
                setup_qtrees_from_string
                )
           ] );
  Exporter::export_ok_tags('all');
}

use Tharn qw(tstlog logcomment script_result);
use Carp qw(croak confess);
use Params::Validate v0.70 qw(validate SCALAR OBJECT HASHREF);
use QtreeCmds qw(qtreePresent qtreeCreate qtreeSecurity qtreeOplocks qtreeStatus);
use Volume qw(vol_exists vol_root);
BEGIN
{
    Params::Validate::validation_options(
        on_fail => sub {my @msgs = @_; chomp(@msgs);
                confess(script_result(@msgs))}
        );
}


=head2 setup_qtrees_from_string

As C<setup_qtrees>, but takes a string description instead of an
array of hashes.  This is intended to be used when the description of
qtrees to (re)create comes from an external source, such as as a
parameter to a test.

=over

=item Arguments

=over

=item conn

(Required) : connection to filer

=item string

(Required) : A space separated list of qtree descriptions. Each
qtree description is a a comma separated list of 
I<attribute>=I<value> pairs.  For example:

   qtree=foo aggr=foo2,type=aggr vol=foo3,type=flex,aggr=foo2
    ???

Refer to C<setup_qtrees> for possible attributes and more details

=back

=back

=cut

sub setup_qtrees_from_string ( $$ ) {

    my ($conn, $qtree_list) = @_;

    my @list;
    # extract attributes and create hash
    my @qtrees = split(/\s+/, $qtree_list);
    foreach my $qtree (@qtrees) {
        my $args = {};
        my @attributes = split(/,/, $qtree);
        foreach my $attr (@attributes) {
            my ($key, $value) = ($attr =~ /(.*?)=(.*)/);
            $args->{$key} = $value;
        }
        push (@list, $args);
    }
    setup_qtrees($conn, @list);

    return 1;
}


=head2  setup_qtrees

Create several qtrees.

=over

=item Returns

1 on success

dies otherwise.


=item Arguments

=over

=item conn

(Required) : connection to filer

=item qtree descriptions

(Required) : Each argument after the connection is a hash reference
describing a particular qtree to be created.

Possible elements are:

=over

=item qtree

Name of qtree to be created.

=item path

Path of qtree

=back

   setup_qtrees($conn,
     {vol=>'foo',type=>'flex',aggr=>'xxx'},
     {vol=>'boo',type=>'clone',parent=>'foo'}
    );

=back

=back


=cut

sub setup_qtrees ( $@ ) {

    my ($conn, @list) = @_;

    foreach my $item (@list) {
        $item->{conn} = $conn;
        _setup(%$item);
        croak("$@\n") if ($@);
    }

}

sub _setup ( % ) {
    my (%args) = validate (@_, { 
                 conn => { type=>OBJECT|HASHREF, isa=>'Connectrec' },
                 qtree => { type=>SCALAR, regex=>qr/^(\/vol\/)?\w+(\/\w+)?/, optional=>0 },
                 security => { type=>SCALAR, regex=>qr/^unix|ntfs|mixed$/, optional=>1 },
                 oplocks => { type=>SCALAR, regex=>qr/^enable|disable$/, optional=>1 },
                 #purge_existing => { type=>SCALAR, regex=>qr/^0|1$/, optional=>1, default=>0 },
                }
            );

    # check if volume in qtree path exists on filer
    my ($vol, $path);
    if ($args{qtree} =~ /^\/vol\/(\w+)\/\w+$/) {
        ($vol) = $1;
        if (!vol_exists($args{conn}, $vol)) {
            croak("Unable to create qtree $args{qtree} since volume '$vol' ". 
                    "does not exist\n");
        }
        $path = $args{qtree};
    } elsif ($args{qtree} =~ /^\w+$/) {
        $vol = vol_root($args{conn});
        tstlog('-commline', "Qtree $args{qtree} must live in root volume '$vol'");
        $path = "/vol/$vol/$args{qtree}";
    }


    # check if qtree already exists on filer
    if (qtreePresent($args{conn}, $path)) {
        tstlog('-commline', "Qtree '$path' already exists");
    } else {
        logcomment("Creating qtree $path on volume $vol");
        qtreeCreate($args{conn}, $path);
        croak("$@\n") if ($@);
    }

    # set security style if asked to
    if (defined($args{security})) {
        tstlog('-commline', "Setting security style for qtree '$path' to '$args{security}'");
        qtreeSecurity($args{conn}, $path, $args{security});
        croak("$@\n") if ($@);
    }

    # set oplocks if asked to
    if (defined($args{oplocks})) {
        tstlog('-commline', "Setting oplocks for qtree '$path' to '$args{oplocks}'");
        my $on = ($args{oplocks} eq 'enable') ? 1 : 0;
        qtreeOplocks($args{conn}, $path, $on);
        croak("$@\n") if ($@);
    }

    # verify the qtree that we created
    my $qStat = qtreeStatus($args{conn}, $path);
    croak("$@\n") if ($@);

    #croak("Qtree $path was not created\n") if (!$qstat{$path}); 

    croak("Unexpected security style '$qStat->{security}' for qtree $path. Expected '$args{security}'\n") 
        if ($args{security} && ($qStat->{security} ne $args{security})); 
    
    croak("Unexpected oplocks '$qStat->{oplocks}' for qtree $path. Expected '$args{oplocks}d'\n") 
        if ($args{oplocks} && ($qStat->{oplocks} ne $args{oplocks}.'d')); 

    croak("Unexpected status '$qStat->{status}' for qtree $path. Expected it to be 'normal'\n") 
        if ($qStat->{status} ne 'normal'); 

    logcomment("Created qtree '$path' successfully");

    return (1);

}



1;

=head1 STATUS

    Shared

=head1 ATTRIBUTES

    ONTAP_COMMANDS = none

    ONTAP_ZAPI_APIS = none

=head1 AUTHOR

    archana@netapp.com, dl-test-infra@netapp.com
    $Id: //depot/prod/test/Rsierranevada/lib/Qtree/Setup.pm#1 $
    Last revised $Date: 2013/01/22 $

=head1 COPYRIGHT

    Copyright 2005, Network Appliance Inc., All Rights Reserved.

=cut

