# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/QtreeCmds.pm#1 $

## @summary Library of qtree commands and related functions
## @author dl-johny@netapp.com
##         (FSS WAFL Quality)
## @status public
## @pod here


=head1 NAME

QtreeCmds - Library of qtree commands and related functions

=head1 DESCRIPTION

This package defines routines which use the filer's "qtree" command.
They gather and return information about qtree, attempt to create,
change the qtree, and return success or failure.

=head1 FUNCTIONS

=cut

package QtreeCmds;

BEGIN {
    use Exporter qw(import);
    our @EXPORT = qw(qtreeCreate qtreeRename qtreeOplocks qtreeSecurity
		qtreeStatus qtreeStatusVol qtreeDelete
		qtreeList qtreePresent qtreeComponents
		);
}

use strict;
use Tharn;
use SayWrap qw(sayEOF);
use Carp qw(croak);
use Params::Validate v0.70 qw(validate validate_pos SCALAR HASHREF UNDEF);
use Module::Common qw(validate_mixed);
use Admin::License qw(license is_licensed);
use FS::Util::Utilities qw(deltree uSystem);
use Utils qw(make_conn);
use Volume qw(vol_exists);
use ExportFS qw (export_path unexport_path list_exports);
use mount qw(mount umount);
use SPARSE::Util::Logger;
use filer qw(filer_vfiler_command);
use Data::Dumper;

my $sudo_cmd = defined($main::SUDO_PATH) ? $main::SUDO_PATH : "/usr/local/bin/sudo";

my $VOL_REGEX = qr{[a-zA-Z_]\w*};
my $QTREE_REGEX = qr{[^/]*[^/\s]};  # no space at end - even though ontap allows
                                    # but it messes up getting the name of
                                    # qtree from qtree status
my $VFILER_REGEX = qr{[\w_-]+};
my $QTREE_PATH_REGEX = qr{^/vol/($VOL_REGEX)(?:/($QTREE_REGEX))?$};

# An internal function, just to avoid lots of cut & copy within this file.
# The 'qtree status' command was just 'qtree' on older releases, so we
# need to try both in case we're trying to run scripts across releases.
sub qtree_status($;$$) {
  my ($conn, $volname, $timeout) = @_;

  my $cmd = "qtree status -i -v";
  $cmd .= " $volname" if ($volname);
  my $text;
  if (defined $timeout) {
    $text = sayEOF($conn, $cmd, timeout => $timeout);
  } else {
    $text = sayEOF($conn, $cmd);
  }

  if ($text =~ /Invalid argument "status"/) {
    $cmd = "qtree";
    $cmd .= " $volname" if ($volname);
    $text = sayEOF($conn, $cmd);
  }

  # Remove header of the 'qtree status' output
  $text =~ 
    s/^Volume\s+Tree\s+Style\s+Oplocks\s+Status\s+ID\s*(Owning vfiler)?\s*\-+\s+\-+\s+\-+\s+\-+\s\-+\s+\-+\s+(\-+)?\s*//;

  return $text;
}

=head2 qtreeCreate($connobj,$qtreepath)

Create the qtree specified in $qtreepath

Arguments

=over

=item connobj

(Required) : Connection object to filer

=item qtreepath

(Required) : Path of qtree to be created.
Normally must be provided in the form "/vol/<volname>/<qtreename>".
If just the qtreename is provided then the qtree will be created in 
the root volume.

=back

Returns

success,  (if qtree created with no incident)

on failure (if qtree could not be created) returns undef with an error message in $@.

=for html <hr />

=cut

sub qtreeCreate($$) {

  my ($conn, $qtreepath) = validate_pos(
                    @_,
                    { type=>SCALAR|HASHREF, optional=>0 }, 
                    { type=>SCALAR, optional=>0 } # can be /vol/<vol>/<qtree> or <qtree>
                    );

  return eval {
# my $text = sayEOF($conn, "qtree create \'$qtreepath\'");
my $cmd = "qtree create \'$qtreepath\'";

# Convert command if simulating a vfiler context
$cmd = filer_vfiler_command($conn, $cmd);
my $text = sayEOF($conn, $cmd);
croak("$text\n") if ($text);
    return 1;
  };
}

=head2 qtreeRename($conn, $qt_src, $qt_dst)

Renames an existing qtree from $qt_src to $qt_dst

Arguments

=over

=item conn

(Required) : Connection object to filer

=item qt_src

(Required) : Path of the source qtree to be rename
Must be of the form "/vol/<volname>/<qtreename>".

Please note that this routine at present does not provide the ability to remount the qtree
if already in use. If functionality can be added with the addition of an optional parameter
when needed (see qtreeDelete for example).

=item qt_dst

(Required) : Path to rename the qtree to
Must be of the form :/vol/<volname>/<qtreename>"

=back

Returns

success, (if qtree is renamed)

on failure (if qtree could not be rename) returns undef with an error message in $@.

=for html <hr />

=cut

sub qtreeRename(%) {
  my (%args) = validate(
		@_,
		{
			conn => { type=>SCALAR|HASHREF, optional=>0 },
			qt_src => { type=>SCALAR, optional=>0 },
			qt_dst => { type=>SCALAR, optional=>0 }
		}
	);

  my $text = sayEOF($args{conn}, "priv set test; qtree rename "
		. "$args{qt_src} $args{qt_dst}");
  croak("$text\n") if ($text);
  croak("Unable to locate new qtree $args{qt_dst}")
	unless qtreePresent($args{conn}, $args{qt_dst});
  return 1;
}

=head2 qtreeDelete($connobj,$qtreepath)

Delete the qtree specified in $qtreepath.

Arguments

=over

=item fspec

(Required) : Filerspec of filer.

=item qtree_path

(Required) : Path of qtree to be created.
Format is: "/vol/<volname>/<qtreename>".

=item mountopts

(Optional): mount options to use.

=back

Returns

success,  (if qtree does not exist, or if it was successfully deleted)

on failure (if qtree could not be deleted) returns undef with an error message in $@.

=for html <hr />

=cut

sub qtreeDelete($$;%) {
  my ($fspec, $qtree_path, %opts) = validate_mixed( @_,
	      [
                    { type=>SCALAR|HASHREF, optional=>0 }, 
      		    { type=>SCALAR, optional=>0, regex=>$QTREE_PATH_REGEX },
	      ],
	      {
		    #  This option is deprecated, but is still allowed for backwards compatibility
		    #  The option is ignored.
        	    FORCE => {type => SCALAR|UNDEF, optional => 1},
        	    VOL_MOUNTPATH => {type => SCALAR|UNDEF, optional => 1},
        	    MOUNTOPTS => {type => SCALAR|UNDEF, optional => 1},
	      },
    );

    my $conn = make_conn($fspec);
    my $default_ip = $conn->hostp()->default_ip();
    my $vol_mountpath = $opts{VOL_MOUNTPATH};

    if ($qtree_path !~ m%(/vol/([^\/]+))/([^\/]+)%) {
	croak("qtreeDelete: Invalid qtree_path: $qtree_path");
    }
    my ($vol_path, $vol, $qtree) = ($1, $2, $3);
    if (!qtreePresent($conn, $qtree_path)) {
	return(1);
    }
    logresult('INFO', "Deleting qtree $qtree_path");
    my $vol_security = 
        qtreeSecurity($conn, $vol_path)
            or croak("qtreeDelete: Could not get security of volume: $vol");
    my $qtree_security = 
        qtreeSecurity($conn, $qtree_path)
            or croak("qtreeDelete: Could not get security of volume: $vol");
    if ($vol_security eq 'ntfs') {
    	qtreeSecurity($conn, $vol_path, 'mixed')
   	     or croak("qtreeDelete: Could not set security to 'mixed' of volume: $vol");
    }
    if ($qtree_security eq 'ntfs') {
    	qtreeSecurity($conn, $qtree_path, 'mixed')
    	    or croak("qtreeDelete: Could not set security to 'mixed' of qtree: $qtree_path");
    }

    my ($mount_func, $umount_func);
    $mount_func = \&mount::mount;
    $umount_func = \&mount::umount;

    if (!is_licensed($conn, "nfs")) {
        license($conn, "nfs")
            or croak("qtreeDelete: could not license nfs on filer: $@");
    }
     my $local_qtree_path;
     my $local_mnt_pt;
     my $is_exported;

     if ($vol_mountpath) {
         $local_qtree_path = "$vol_mountpath/$qtree";
     } else {
        my $mountopts = $opts{MOUNTOPTS} || $main::MOUNTOPTS || 'actimeo=0,noac,hard,intr,noquota';
        my $curr_exports = list_exports($conn);
        $is_exported = ($curr_exports->{$vol_path}) ? 1 : 0;
        export_path($conn, $vol_path, OPTIONS=>"rw,anon=0,sec=sys");

        my $mntdir = "/t/$default_ip/qtree_delete_" . $$;
        while (-e $mntdir) {
	     $mntdir .= "_";
         }
         uSystem("$sudo_cmd mkdir $mntdir");
         uSystem("$sudo_cmd chmod 0777 $mntdir");
         $local_mnt_pt = &$mount_func($default_ip, $vol_path,
    					mountpoint=>$mntdir,
    					mountopts=>$mountopts)
    			   or croak("qtreeDelete: Mount failed: $@");
         $local_qtree_path = "$local_mnt_pt/$qtree";
     }
     eval {
        deltree($local_qtree_path)
            or logresult('WARN', "qtreeDelete: Error returned from deltree $local_qtree_path: $@");
     };
     if (qtreePresent($conn, $qtree_path)) {
        eval {
             sayEOF($conn, "priv set test; qtree delete $qtree_path");
        };
     }

     if ($local_mnt_pt) {
         &$umount_func($local_mnt_pt, force=>1)
            or logresult('WARN', "qtreeDelete: umount error $local_mnt_pt: $@");
        uSystem( "$sudo_cmd rmdir $local_mnt_pt");
        if (-d $local_mnt_pt) {
             uSystem("$sudo_cmd umount -f $local_mnt_pt");
             uSystem("$sudo_cmd chmod 0777 $local_mnt_pt");
             uSystem("$sudo_cmd rmdir $local_mnt_pt");
        }
        if (!$is_exported) {
            unexport_path($conn, $vol_path);
        }
        croak("qtreeDelete: Could not umount $local_mnt_pt")
            if (uSystem("mount | grep $local_mnt_pt") =~ /$local_mnt_pt/);
        croak("qtreeDelete: Could not delete mountpoint: $local_mnt_pt")
            if (-d $local_mnt_pt);
    }

    if (qtreePresent($conn, $qtree_path)) {
        croak("qtreeDelete: Could not delete qtree: $qtree_path")
    } else {
	logresult('INFO', "Deleted qtree $qtree_path");
    }

    qtreeSecurity($conn, $vol_path, $vol_security)
        or croak("qtreeDelete: Could not set security to 'unix' of volume: $vol");

    return 1;
}

=head2 qtreeOplocks($connobj,$qtreepath,$on)

Gets or sets the current oplocks setting for a qtree.

Arguments

=over

=item connobj

(Required) : Connection object to filer

=item qtreepath

(Required) : Path to qtree. Should be of the form "/vol/<volname>[/<qtreename>]"

=item on

(Optional) : If specified, sets the current oplocks setting as specified,
If not specified, gets and returns the current oplocks setting.

=back

Returns

=over 

=item Set : (if $on is specified)

returns success or failure with an error message in $@ on failure.

=item Get : ($on is not specified)

on success returns the current oplocks setting,

on failure returns undef with an error message in $@.

=back

=for html <hr />

=cut

sub qtreeOplocks($$;$) {

  my ($conn, $qtreepath, $on) = validate_pos(
                    @_,
                    { type=>SCALAR|HASHREF, optional=>0 }, 
                    { type=>SCALAR, optional=>0, 
                         regex=>$QTREE_PATH_REGEX },
                    { type=>SCALAR, optional=>1,  },
                    );

  return eval {

    if (defined $on) {
      # Set the default oplocks setting.
      my $action = ($on ? "enable" : "disable");
      my $cmd = "qtree oplocks $qtreepath $action";
      my $text = sayEOF($conn, $cmd);
      if ($text) {
	$text =~ s/^qtree\s*:\s*//;
	die "no output\n" if ($text eq "");
	die "$text\n";
      }
      return 1;

    } else {
      # Retrieve and return the current oplocks setting.
      my $qt_exp = quotemeta($qtreepath);
      my $cmd = "qtree oplocks $qtreepath";
      my $text = sayEOF($conn, $cmd);
      unless ($text =~ /^$qt_exp\/? has (?:unix|ntfs|mixed) security style and oplocks are (enabled|disabled)\.$/) {
	$text =~ s/^qtree\s*:\s*//;
	die "no output\n" if ($text eq "");
	die "$text\n";
      }
      return ($1 eq "enabled");
    }

  };
}

=head2 qtreeSecurity($connobj,$qtreepath,$type)

Gets or sets the current security setting for a qtree.

Arguments

=over

=item connobj

(Required) : Connection object to filer

=item qtreepath

(Required) : Path to qtree. Should be of the form "/vol/<volname>[/<qtreename>]"

=item type

(Optional) : If specified, sets the current security style.
If not specified returns the current security setting.

Possible values : 'ntfs' / 'unix' / 'mixed'

=back

Returns

=over

=item Set : (if type is specified)

returns success or failure with an error message in $@ on failure.

=item Get : (type is not specified)

returns the current security setting,

or on failure returns undef with an error message in $@.

=back

=for html <hr />

=cut

sub qtreeSecurity($$;$) {

  my ($conn, $qtreepath, $type) = validate_pos(
                    @_,
                    { type=>SCALAR|HASHREF, optional=>0 }, 
                    { type=>SCALAR, optional=>0, 
                         regex=>$QTREE_PATH_REGEX },
                    { type=>SCALAR, optional=>1, 
                         regex=>qr/^(unix|mixed|ntfs)$/ },
                    );

  return eval {
    my $qt_exp = quotemeta($qtreepath);

    if (defined $type) {
      # Set the default security setting.
      my $cmd = "qtree security $qtreepath $type";
      my $text = sayEOF($conn, $cmd);

      # Certain warnings can get issued if you change the security style to
      # enable a protocol which is not licensed or active, or if quotas are
      # turned on.  Strip out that cruft here.
      $text =~ s/^qtree: Warning (NFS|CIFS) not licensed - proceeding\.\s*//;
      $text =~ s/^qtree: Since CIFS is not active, files in this qtree that have ACLs\s+will not be accessible from NFS\. Also, it will not be possible to change\s+the permissions or ownership of files in this qtree from NFS\.\s+Change its security style to unix if access from NFS is required\s+before CIFS is started\.\s*//;
      $text =~ s/^qtree: Changing the security style of qtree $qt_exp\/? will\s+change the visibility of existing Windows security descriptors \(a\.k\.a\. ACLs\)\.\s+This may affect the disk space usage values in the quota data base\.\s+Turn quotas off and then on to recompute disk space usage\.\s*//;

      if ($text) {
	$text =~ s/^qtree\s*:\s*//;
	die "no output\n" if ($text eq "");
	die "$text\n";
      }
      return 1;

    } else {
      # Retrieve and return the current security setting.
      my $cmd = "qtree security $qtreepath";
      my $text = sayEOF($conn, $cmd);
      unless ($text =~ /^$qt_exp\/? has (unix|ntfs|mixed) security style and oplocks are (?:enabled|disabled)\.$/) {
	$text =~ s/^qtree\s*:\s*//;
	die "no output\n" if ($text eq "");
	die "$text\n";
      }
      return $1;
    }

  };

}

=head2 qtreeStatus($connobj,$qtreepath)

Gets and returns the status of the specified qtree.

Arguments

=over

=item connobj

(Required) : Connection object to filer

=item qtreepath

(Required) : Path to qtree. Should be of the form "/vol/<volname>[/<qtreename>]"
Routine will die if volume does not exist.

=back

Returns

On success, returns a hash.

On failure, returns undef with an error message in $@.

=for html <hr />

=cut

sub qtreeStatus($$) {

  my ($conn, $qtreepath) = validate_pos(
                    @_,
                    { type=>SCALAR|HASHREF, optional=>0 }, 
                    { type=>SCALAR, optional=>0, 
                         regex=>$QTREE_PATH_REGEX },
                    );

  return eval {

    my ($volname, $qtreename) = qtreeComponents($qtreepath);
    croak("Invalid qtree path '$qtreepath'\n") unless (defined $volname);

    my %HoH = qtreeStatusVol($conn, $volname);
    croak("$@\n") if ($@);
    
    if ($HoH{$qtreepath}) {
        return($HoH{$qtreepath});
    } else {
        croak("Could not get status of Qtree '$qtreepath'. Probably does not exist\n");
    }

  };
}


=head2 qtreeStatusVol($connobj,$volname)

Gets and returns the status all qtrees associated with the volume $volname
If $volname is not specified then it will return the status of all qtrees on the filer.

Arguments

=over

=item connobj

(Required) : Connection object to filer

=item volname

(Optional) : Volume whose qtree info is to be obtained.
If not specified then will return the status of all qtrees on the filer.
Routine will die if volume does not exist.

=back

Returns

On success returns a hash of hashes.

On failure, returns undef with an error message in $@.

Bugs

Will return incorrect qtree name if the qtree name starts/ends with a space.

=for html <hr />

=cut


sub qtreeStatusVol($;$) {

  my ($conn, $volname) = validate_pos(
                    @_,
                    { type=>SCALAR|HASHREF, optional=>0 }, 
                    { type=>SCALAR, optional=>1, 
                         regex=>$VOL_REGEX },
                    );

  return eval {

    my %HoH;

    my ($text, $volexp);
    if ($volname) {
        my $volname_exp = quotemeta($volname);
        $text = qtree_status($conn, $volname);
        $volexp = $volname_exp;
    } else {
        $text = qtree_status($conn);
        $volexp = $VOL_REGEX;
    }
    my $expected = 
        qr/^($volexp)\s+($QTREE_REGEX)?\s*(unix|mixed|ntfs)\s+(enabled|disabled)\s+(\w+)\s+(\d+)\s*($VFILER_REGEX)?\s*$/;

    my @lines = split(/\n/,$text);

    foreach my $line (@lines) {
        croak("Error parsing qtree status output |$line|\n")
          if ($line !~ $expected);

        my $path = "/vol/".$1;
        $path .= "/$2" if ($2);

        $HoH{$path}{security} = $3;
        $HoH{$path}{oplocks}  = $4;
        $HoH{$path}{status}   = $5;
        $HoH{$path}{id}       = $6;
        $HoH{$path}{vfiler}   = $7;
    }

    return(%HoH);

  };
}

=head2 qtreeList($connobj,$volname)

Gets and returns a list of qtrees. 

Arguments

=over

=item connobj

(Required) : Connection object to filer

=item volname

(Optional) : Name of volume whose qtree's are required.

=back

Returns

If $volname is provided, returns only the qtree paths of qtrees in that volume.

If no qtrees are present, return an empty list.

All returned paths are of the form "/vol/volname/qtreename".

Bugs

Will return incorrect qtree name if the qtree name starts/ends with a space.

=for html <hr />

=cut

sub qtreeList($;$) {

  my ($conn, $volname) = validate_pos(
                    @_,
                    { type=>SCALAR|HASHREF, optional=>0 }, 
                    { type=>SCALAR, optional=>1, 
                         regex=>$VOL_REGEX },
                    );

  return eval {

    my $text = qtree_status($conn, $volname);
    # if volume is specified then $text will contain qtree info for that vol only

    # use volume name in regex if provided
    my $vol_regex = ($volname) ? $volname : $VOL_REGEX;

    # Break into lines
    my @lines = split(/\n/, $text);

    # Strip out all lines for volumes (we are interested only in qtrees)
    # we dont have to worry about qtrees for other volumes because
    # if volume is specified then qtree_status will give details for qtrees
    # present on specified volume only.
    @lines = grep { ! /^$vol_regex\s+(unix|mixed|ntfs)\s+(enabled|disabled)\s+\w+\s+\d+\s*($VFILER_REGEX)*\s*$/ } @lines;

    # Parse qtree output into a list of qtree paths
    # Fail fatally if something doesn't parse
    my @qtrees = map {
        ( /^($vol_regex)\s+
            ($QTREE_REGEX)\s+
            (?:unix|mixed|ntfs)\s+
            (?:enabled|disabled)\s+
            \w+\s+
            \d+\s*
            \S*\s*
        $/x ? "/vol/$1/$2" : die "Error parsing qtree line: $_\n")
    } @lines;

    return @qtrees;
  };
}

=head2 qtreePresent($connobj,$qtreepath, TIMEOUT => $TIMEOUT)

Determines if a qtree is present or not

Arguments

=over

=item connobj

(Required) : Connection object to filer

=item qtreepath

(Required) : Path to qtree. Should be of the form "/vol/<volname>[/<qtreename>]"

=item timeout

(Optional) : timeout period (in seconds) waiting for the command to complete

=back

Returns

1 if qtree is present

0 if qtree is absent

undef with an error message in $@ in case of failure

=for html <hr />

=cut

sub qtreePresent($$;%) {

  my ($conn, $qtreepath, %opts) = validate_mixed(
    @_,
    [ { type=>SCALAR|HASHREF, optional=>0 },
      { type=>SCALAR, optional=>0,
	  regex=>$QTREE_PATH_REGEX } ],
      {
        TIMEOUT => {type => SCALAR|UNDEF, optional => 1}
      });
 
  return eval {
    my ($volname, $qtreename) = qtreeComponents($qtreepath);
    die $@ unless (defined $volname);
    return(0) unless (vol_exists($conn, $volname));
    my $volname_exp = quotemeta($volname);
    my $qtree_exp = quotemeta($qtreename);

    my $text = qtree_status($conn, $volname, $opts{TIMEOUT});

    # Return whether the command output has the qtree
    return($text =~ /^$volname_exp\s+$qtree_exp\s+(unix|mixed|ntfs)\s+(enabled|disabled)\s+\w+\s+\d+\s*($VFILER_REGEX)?\s*$/m);
  };
}

=head2 qtreeComponents(qtreepath)

Breaks up a qtree path and returns a list
of the qtree components ($volume, $qtreename).

Arguments

=over

=item connobj

(Required) : Connection object to filer

=item qtreepath

(Required) : Path to qtree. Should be of the form "/vol/<volname>[/<qtreename>]"

=back

Returns

On success a list of the qtree components ($volume, $qtreename).

On error (if the path is invalid) returns failure with an error message in $@.

=for html <hr />

=cut

sub qtreeComponents($) {
  my ($qtreepath) = @_;
  return eval {
    die "Invalid qtree path: $qtreepath\n"
      unless ($qtreepath =~ $QTREE_PATH_REGEX);
    return ($1, $2);
  }
}

=head1 BUGS

OnTap allows qtree names to start/end in a 'space'. This library currently does not handle it.
All qtree names ending in a space will be considered as an invalid name however qtree names 
starting with a space will not be invalid.
OnTap accepts the path in case of volumes to end with a slash. eg. /vol/vol0/, but this lib 
does not accept that regex.

=cut

1;
