#!/bin/sh /usr/software/bin/perl

package Queue;

BEGIN {
  die "Cannot run as root" if $< == 0;
}

use Data::Dumper;
use Time::HiRes qw(gettimeofday tv_interval);
use Getopt::Long qw(:config pass_through);
use Errno qw(EEXIST);
use Carp qw(carp cluck confess croak);
use Net::SMTP;
use Pg;
use strict;
use warnings;

# Callback is the function that allows the calling program to have a chance to reject the job before it is pulled off the queue
sub Callback {
  my $self = shift;
  my $item = $_[0];
  my $retval = 0;

  if (exists($self->{CallBack})){
    # Call the callback if it exists
    $retval = $self->{CallBack}(@_);
  } elsif (exists($item->{QueueId})){
    # Verify that the queue matches if it has been specified
    if ($item->{QueueId} eq $self->{QueueId}){
      $retval = 1
    }
  } else {
    # Otherwise, we're good to go
    $retval = 1;
  }
  $retval;
}

sub ShowResult {
  my ($self, $DBResultHandle, $verbose) = @_;

  my $DBResultStatus = $DBResultHandle->resultStatus();
  my $retval = 0;
  my $str = "";

  if ($DBResultStatus == PGRES_EMPTY_QUERY){
    $str = "PGRES_EMPTY_QUERY";
  } elsif ($DBResultStatus == PGRES_COMMAND_OK){
    #    print "PGRES_COMMAND_OK\n";
    $retval = 1;
  } elsif ($DBResultStatus == PGRES_TUPLES_OK){
    #    print "PGRES_TUPLES_OK\n";
    $retval = 1;
  } elsif ($DBResultStatus == PGRES_COPY_OUT){
    $str =  "PGRES_COPY_OUT";
  } elsif ($DBResultStatus == PGRES_COPY_IN){
    $str = "PGRES_COPY_IN";
  } elsif ($DBResultStatus == PGRES_BAD_RESPONSE){
    $str = "PGRES_BAD_RESPONSE";
  } elsif ($DBResultStatus == PGRES_NONFATAL_ERROR){
    $str = "PGRES_NONFATAL_ERROR";
  } elsif ($DBResultStatus == PGRES_FATAL_ERROR){
    $str = "PGRES_FATAL_ERROR\n";
  } elsif ($DBResultStatus) {
    cluck "Unknown error $DBResultStatus\n";
    print Dumper($DBResultStatus);
  }
  print "Return value is $str\n" if $verbose;

  return $retval;
}

sub SqlExec {
  my ($self, $cmd, $verbose) = @_;

  cluck "Executing command: $cmd" if $verbose;
  if ($self->{Reconnect} || !defined($self->{DBConn})){
    $self->DBConn();
    $self->{Reconnect} = 0;
  }
#  $self->{DBConn}->reset;
  my $dbstatus = $self->{DBConn}->status;
  if ($dbstatus == PGRES_CONNECTION_OK){
#    print "DBConn is OK\n";
  } elsif ($dbstatus == PGRES_CONNECTION_BAD){
    print "DBConn is BAD\n";
  } else {
    print "DBConn is in an unknown state\n";
  }
  print Dumper($self) if $verbose;
  return undef unless $self->{DBConn};
  my $DBResultHandle;
  my $DBResultStatus;
  if ($DBResultHandle = $self->{DBConn}->exec($cmd)){
    $DBResultStatus = $self->ShowResult($DBResultHandle, $verbose);
  }

  if (wantarray()){
    print "Returning: res = $DBResultHandle, DBResultStatus = $DBResultStatus\n" if ($verbose);
    return ($DBResultHandle, $DBResultStatus);
  } else {
    print "Returning: res = $DBResultHandle\n" if ($verbose);
    return $DBResultHandle;
  }
}

sub DBConn {
  my $self = shift;

  my $conn = Pg::connectdb("dbname=queue host=zoo.hq.netapp.com user=bryans password=bryans");
  die "Could not establish connection to database" if $conn->status;

  $self->{DBConn} = $conn;
}

sub new {
  my $invocant = shift;
  my $class = ref($invocant) || $invocant;

  my $host = `/bin/hostname`;
  chomp $host;

  my $self = {
	      Program => $0,
	      Host => $host,
	      Run => 0,
	      SpinState => 0,
	      Pid => $$,
#	      QueueId => $0 . "_" . $host . "_" . $$,
	      Priority => 100,
	      Mail => 1,
	      Reconnect => 0,
	      @_
	     };

  $| = 1;
  my $retval = bless($self, $class);
  my $RunCode = $retval->ParseOpts(%{$self});
  $retval->DBConn();
  my @ARGVC = @ARGV;
  $retval->{Args} = \@ARGVC;
  if ($retval->{Run}){
    die "Must have a qid specified when running!" if !$self->{QueueId};
    print "Running\n";
    my $DBResultHandle = $self->SqlExec("begin;");
    $DBResultHandle = $self->SqlExec("select oid,* from queues where name = '$self->{QueueId}' for update;");
    my $rows = $DBResultHandle->ntuples();
    if ($rows == 0){
      my ($DBResultHandle, $DBStatus) = $self->SqlExec("insert into queues (name, state, host, pid) values ('$self->{QueueId}',1,'$self->{Host}','$self->{Pid}');");
      if (!$DBStatus){
	die "Could not insert run information into queues: $DBStatus";
      }
    } else {
      $DBResultHandle = $self->SqlExec("update queues set host = '$self->{Host}', pid = '$self->{Pid}' where name = '$self->{QueueId}';");
    }
    $DBResultHandle = $self->SqlExec("commit;");
  }

  ($retval, $RunCode);
}

sub Set {
  my $self = shift;
  my %params = (@_);

  foreach my $key (sort keys %params){
    $self->{$key} = $params{$key};
    print "$key set to $self->{$key}\n" if $self->{Debug};
    if (($key eq "Cluster")){
      $self->{QueueId} = $params{$key};
    }
  }
}

sub Get {
  my ($self, $key) = @_;

  if (exists($self->{$key})){
    return $self->{$key};
  }
  return undef;
}

sub CompleteJob {
  my $self = shift;
  my $verb = 0;


  print Dumper($self) if $self->{Debug};
  if ($self->{Oid}){
    print "Finishing job\n";
    my $DBResultHandle = $self->SqlExec("begin;", $verb);
    $DBResultHandle = $self->SqlExec("update jobs set \"state\" = 2, finishtime = now() where oid = '$self->{Oid}';", $verb);
    $DBResultHandle = $self->SqlExec("update queues set job = null where name = '$self->{QueueId}';", $verb);
    $DBResultHandle = $self->SqlExec("commit;", $verb);
    if ($self->{Mail}){
      if (my $smtp = Net::SMTP->new("sepia.hq.netapp.com")){
#	cluck "Sending finish mail: $$";
	$DBResultHandle = $self->SqlExec("select user, id from jobs where oid = '$self->{Oid}';");
	my @row = $DBResultHandle->fetchrow();
	$smtp->mail("queue\@netapp.com");
	$smtp->to($row[0] . "\@netapp.com");
	$smtp->to("bryans\@netapp.com");
	$smtp->data();
	$smtp->datasend("To: $row[0]\@netapp.com\n");
	$smtp->datasend("Subject: Job $row[1] completed\n");
	$smtp->datasend("\n\n");
	$smtp->datasend("Your job has finished\n");
	$smtp->dataend();
	$smtp->quit();
      }
    }
    delete $self->{Oid};
  }
}
sub bt {
  my $self = shift;
  my $print = shift;
  if (!defined($print)){
    $print = 0;
  }
  my $i = 0;
  my $retval = 0;
  while (my ($pack, $file, $line, $subname, $hasargs, $wantarray, $evaltext, $is_require) = caller($i++)){
    if ($file =~ /RemoteProcess/){
      $retval++;
    }
    if ($print){
      print " p = $pack" if defined($pack);
      print " f = $file" if defined($file);
      print " l = $line" if defined($line);
      print " s = $subname" if defined($subname);
      print " a = $hasargs" if defined($hasargs);
      print " w = $wantarray" if defined($wantarray);
      print " e = $evaltext" if defined($evaltext);
      print " w = $wantarray" if defined($wantarray);
      print " e = $evaltext" if defined($evaltext);
      print " r = $is_require" if defined($is_require);
      print "\n";
    }
  }
  $retval;
}

sub DESTROY {
  my $self = shift;

  cluck "Destroy called" if $self->{Debug};
  $self->{Reconnect}++;
  return if $self->bt();
  print Dumper($self) if ($self->{Debug} > 1);
  if (1 == 0){
    my $i = 0;
    while (my ($pack, $file, $line, $subname, $hasargs, $wantarray, $evaltext, $is_require) = caller($i++)){
      print "$pack, $file, $line, $subname, $hasargs, $wantarray, $evaltext, $is_require\n";
    }
  }
  print "Queue::Destroy called\n" if $self->{Debug};
  if (!defined($self->{DBConn})){
    print "Re-establishing database connection\n" if ($self->{Debug});
    $self->{DBConn} = Pg::connectdb("dbname=queue host=zoo.hq.netapp.com user=bryans password=bryans");
  }
  if ($self->{Run}){
    $self->CompleteJob();
    my ($DBResultHandle, $DBStatus) = $self->SqlExec("update queues set host = null, pid = null where name = '$self->{QueueId}';",0);
  }
}

sub ParseOpts {
  my $self = shift;
  my %options = (@_);

  my $QueueIdArg = 'qid=s';
  if ($self->{QueueIdArg}){
    $QueueIdArg = "qid|$self->{QueueIdArg}=s";
  }

  print "$QueueIdArg\n" if $self->{Debug};
  my $retval = 1;
  my %QueueOptions;

  my $queueres = GetOptions(\%QueueOptions,
			    'pri=i',
			    'low' => sub {$QueueOptions{pri} = 100},
			    'high' => sub {$QueueOptions{pri} = 0},
			    'run',
			    $QueueIdArg,
			   );

  if (exists($QueueOptions{pri})){
    $self->{Priority} = $QueueOptions{pri};
  }
  if ($QueueOptions{qid}){
    $self->{QueueId} = $QueueOptions{qid};
  }
  if ($QueueOptions{run}){
      $retval = 2;
      $self->{Run} = 1;
  }

  if ($self->{QueueIdArg}){
    push @ARGV, "-" . $self->{QueueIdArg} => $self->{QueueId} if $self->{QueueId};
  }

  $retval;
}

sub Enqueue {
  my $self = shift;
  my $retval = 0;

  my %InOptions = (
	       @_,
	      );

  my @CommandArguments;
  if (exists($InOptions{Args})){
    # This had better be a ref to an array.....
    # We want to make our own copy since the caller could still trash the args....
    @CommandArguments = @{$InOptions{Args}};
    delete $InOptions{Args};
  } else {
    # No argument were explicitly passed, so use the arguments that we saved from the command line
    @CommandArguments = @{$self->{Args}};
  }
  if (@CommandArguments){
    # Only enqueue if there are any arguments.
    foreach my $key (keys %InOptions){
      # Can't handle undef values in the database.....
      if (!defined($InOptions{$key})){
	delete $InOptions{$key};
      }
    }
    if ($InOptions{Cluster}){
      # Substitute QueueId for Cluster
      $self->{QueueId} = $InOptions{Cluster};
    }
    delete $InOptions{Cluster};

    # Set up the data for the insert
    my %TableData = (
		Args => join(" ", @CommandArguments),
		User => $ENV{USER},
		Host => $self->{Host},
		Program => $self->{Program},
		Priority => $self->{Priority},
		State => 0,
		Options => join('#fSeP#', %InOptions),
		Queue => $self->{QueueId},
	       );

    # Now build up the insert statement
    my $Fields = "";
    my $Values = "";
    foreach my $key (sort keys %TableData){
      if (defined($TableData{$key})){
	if (length($Fields)){
	  $Fields .= ", ";
	}
	$Fields .= "\"" . lc $key . "\"";
	if (length($Values)){
	  $Values .= ", ";
	}
	$Values .= "'$TableData{$key}'";
      }
    }
    # The following call could fail, as could all DB ops, but this would really tick off the user.....
    my ($DBResultHandle, $DBStatus) = $self->SqlExec("insert into jobs ($Fields) values ($Values);", 0);
    if (!$DBStatus){
      $self->ShowResult($DBResultHandle, 1);
      warn "Insert failed";
      return 0;
    }
    my $oid = $DBResultHandle->oidStatus;
    $DBResultHandle = $self->SqlExec("select id from jobs where oid = '$oid';");
    my @row = $DBResultHandle->fetchrow();
    my $job;
    if (@row){
      print "Job $row[0] enqueued\n";
      $job = $row[0];
      $retval = 1;
    }

    if ($job){
      if (defined($self->{QueueId})){
	$DBResultHandle = $self->SqlExec("select id from jobs where \"state\" = 0 order by priority, oid where queue = '$self->{QueueId}';");
      } else {
	$DBResultHandle = $self->SqlExec("select id from jobs where \"state\" = 0 order by priority, oid;");
      }
      my $num = 0;
      my $loc = -1;
      while (@row = $DBResultHandle->fetchrow()){
	if ($row[0] == $job){
	  $loc = $num;
	}
	$num++;
      }
      if ($loc){
	$loc++;
      }
      if ($num == 1){
	if ($loc == 0){
	  print "Your job is the only one in the queue.\n";
	} else {
	  print "There is $num job in the queue, but it looks like yours has been pulled out already.\n";
	}
      } else {
	print "There are $num jobs in the queue and your job is at position $loc.\n";
      }

      print "You can check job status at http://zoo.hq.netapp.com/~bryans/show_queue.php\n";
    }
  }
  return $retval;
}

sub NextItem {
  my $self = shift;
  my %params = (@_);

  my @row;
  my %data;
  my $DBResultHandle;
  my $DBResultStatus;

  # Make sure that we marked our last job finished, otherwise we shouldn't pick up a new one.
  $self->CompleteJob();
  while (1){
    # Make sure the queue is runnable
    $DBResultHandle = $self->SqlExec("select state from queues where name = '$self->{QueueId}';");
    my $rows = $DBResultHandle->ntuples();
    if ($rows){
      @row = $DBResultHandle->fetchrow();
      if (!$row[0]){
	$self->Spin(Dir => -1);
	sleep 1;
	return unless $self->{Run} == 1;
	next;
      }
    }

    $DBResultHandle = $self->SqlExec("begin;");
    $DBResultHandle = $self->SqlExec("select oid,* from jobs where \"state\" = 0 and (queue = '$self->{QueueId}' or queue is null) order by priority, oid for update;");
    $rows = $DBResultHandle->ntuples();
    if (!$rows){
      $self->Spin();
      sleep 1;
      $DBResultHandle = $self->SqlExec("commit;");
      return unless $self->{Run} == 1;
      next;
    }

    # At this point I have a job that needs to be executed.  How do I know if this job is for me?
    # This should be in a transaction so that nobody else picks up the job while we're checking it out
    # and we can update/commit or rollback as needed.
    # The challenge is avoiding picking up the job again.  The old method went through the jobs until
    # it found one that was for this executor, locking the whole queue in the process.  We can improve
    # on this by only selecting runnable jobs (keeping historical and currently executing jobs available)
    # and the select criteria can specifically exclude jobs that are targeted for other clusters.
    while (@row = $DBResultHandle->fetchrow()){
      for (my $i = 0; $i < $DBResultHandle->nfields; $i++){
	my $name = $DBResultHandle->fname($i);
	if (defined($row[$i]) && ($row[$i] =~ /#fSeP#/)){
	  my @narr = split(/#fSeP#/, $row[$i]);
	  my %ndata = @narr;
	  foreach my $key (keys %ndata){
	    $data{$key} = $ndata{$key};
	  }
	} else {
	  $data{ucfirst $name} = $row[$i];
	}
      }
      # Do the verification here
      if (!$self->Callback(\%data, %params)){
	# This is not my job if callback isn't true
	undef @row;
      }

      last if @row;
    }
    # Break if we have a row that we can actually execute.
    last if @row;
  }
  print Dumper(\@row) if $self->{Debug};
  $DBResultHandle = $self->SqlExec("update jobs set \"state\" = 1, executetime = now(), equeue = '$self->{QueueId}' where oid = '$data{Oid}';");
  $DBResultHandle = $self->SqlExec("update queues set job = $data{Oid} where name = '$self->{QueueId}';");
  $DBResultHandle = $self->SqlExec("commit;");
  if ($self->{Mail}){
    if (my $smtp = Net::SMTP->new("sepia.hq.netapp.com")){
#      cluck "Sending start mail: $$";
      $DBResultHandle = $self->SqlExec("select user, id from jobs where oid = '$data{Oid}';");
      my @row = $DBResultHandle->fetchrow();
      print Dumper(\@row) if $self->{Debug};
      $smtp->mail("queue\@netapp.com");
      $smtp->to($row[0] . "\@netapp.com");
      $smtp->to("bryans\@netapp.com");
      $smtp->data();
      $smtp->datasend("To: $row[0]\@netapp.com\n");
#      $smtp->datasend("To: bryans\@netapp.com\n");
      $smtp->datasend("Subject: Job $row[1] started\n");
      $smtp->datasend("\n\n");
      $smtp->datasend("Your job has started\n");
      $smtp->dataend();
      $smtp->quit();
    }
  }
  $self->{Oid} = $data{Oid};

  # I'm going to be returning control to the caller and they may fork, so I'm going to close the connection
  $self->{DBConn} = undef;

  \%data;
}

sub Shutdown {
  my $self = shift;

  $self->{Run} = 2;
}

sub Spin {
  my $self = shift;
  my %params = (Dir => 1,
		@_);

  if ($self->{SpinState} == 0){
    print " |\r";
  } elsif ($self->{SpinState} == 1){
    print " /\r";
  } elsif ($self->{SpinState} == 2){
    print " -\r";
  } elsif ($self->{SpinState} == 3){
    print " \\\r";
  }
  $self->{SpinState} += $params{Dir};
  if ($self->{SpinState} > 3){
    $self->{SpinState} = 0;
  } elsif ($self->{SpinState} < 0){
    $self->{SpinState} = 3;
  }
}

package main;

my %options = (
	       horizontal => 0,
	       showhex => 0,
	       displaysep => 10,
	      );

my @keys;
my %max;
my %data;
my %data2;

sub handledata {
  my ($key, $format, $format2, @values) = @_;
  if ($options{horizontal}){
    push(@keys, $key) if (!grep(/^$key$/, @keys));
    my $value = sprintf($format, $values[0]);
    my $value2 = sprintf($format2, $values[1]);
    if (!defined($max{$key})){
      $max{$key} = length($key);
    }    my $vallen = length($value);
    if ($vallen > $max{$key}){
      $max{$key} = $vallen;
    }
    my $vallen2 = length($value2);
    if ($vallen2 > $max{$key}){
      $max{$key} = $vallen2;
    }
    push(@{$data{$key}}, $value);
    push(@{$data2{$key}}, $value2);
  } else {
    printf("$key: " . $format, $values[0]);
    if (length($format2)){
      printf(" (" . $format2 . ")", $values[1]);
    }
    print "\n";
  }
}

use Getopt::Long;
#use Time::HiRes qw(gettimeofday tv_interval);

sub HandleFile {
  my ($aref) = @_;
  my %options = (
		 horizontal => 0,
		 showhex => 0,
		 displaysep => 10,
               );
  @ARGV = split(/\s+/, $aref->{Args});
  my $file;
  my $error;

  warn "Error parsing command line" unless GetOptions(\%options, 'horizontal', 'showhex', 'displaysep=i', 'time');

  while ($file = shift @ARGV){
    $error = 0;
    #    my $starttime = [gettimeofday];
    $! = 0;
    my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = lstat($file);
    #    my $elapsed = tv_interval ( $starttime );
    #    printf("Stat took %.4f seconds\n", $elapsed) if ($options{time});
    #    print "$dev:$ino:$mode:$nlink:$uid:$gid:$rdev:$size:$atime:$mtime:$ctime:$blksize:$blocks\n";
    if ($!){
      warn "Could not stat $file";
      print "  Cause: ", $!, "\n";
      next;
    }

    handledata("File", "%s", "", $file);
    #    handledata("Inode", "%x", "%s", $ino, ($ino >> 16) . "." . ($ino & 0xFF));
    handledata("Inode", "%x", "", $ino);
    handledata("Mode", "%o", "", $mode);
    handledata("Lnk", "%d", "", $nlink);
    handledata("Uid", "%d", "%x", $uid, $uid);
    handledata("Gid", "%d", "%x", $gid, $uid);
    handledata("Rdv", "%d", "", $rdev);
    handledata("Size", "%d", "%x", $size, $size);
    handledata("Access", "%s", "%x", scalar(localtime($atime)), $atime);
    handledata("Modify", "%s", "%x", scalar(localtime($mtime)), $mtime);
    handledata("Change", "%s", "%x", scalar(localtime($ctime)), $ctime);
    handledata("Prf BS", "%d", "%x", $blksize, $blksize);
    handledata(".5K B", "%d", "%x", $blocks, $blocks);
    if (!$options{horizontal}){
      print "\n";
    }
  }

  if ($options{horizontal}){
    my $totlen = 0;
    foreach my $key (@keys){
      printf("%-" . ($max{$key} + 2) . "s", $key);
      $totlen += $max{$key} + 2;
    }
    print "\n";
    my $size = scalar @{$data{$keys[0]}};
    for (my $i = 0; $i < $size; $i++){
      foreach my $key (@keys){
	printf("%-" . ($max{$key} + 2) . "s", $data{$key}[$i]);
      }
      print "\n";
      if ($options{showhex}){
	foreach my $key (@keys){
	  printf("%-" . ($max{$key} + 2) . "s", $data2{$key}[$i]);
	}
	print "\n";
      }
      if ($options{displaysep}){
	#	    print "Here $options{displaysep}\n";
	if ((($i % $options{displaysep}) == 0) && ($i > 0)){
	  print "-"x$totlen, "\n";
	}
      }
    }
  }
}

use Data::Dumper;

sub main {
  my $run = 1;
  my ($queue, $rv) = Queue->new(Debug => 0);

  $SIG{INT} = sub {
    print "clearing run\n";
    $queue->Shutdown();
    $run = 0;
  };

  if ($rv == 2){
    # I used this code along with the use lib from above to debug the DESTROY problem with SpinDex
    #    eval "use SpinDexClient";
    #    my $sd = new SpinDexClient();
    #    sleep 2;
    #    $sd->DESTROY();
    while ($run){
      my $ref = $queue->NextItem();
#      print Dumper($ref);
      if (!defined($ref)){
	select undef, undef, undef, .5;
	$queue->Spin();
      } else {
	print $ref->{Args} . "\n";
	sleep 10;
	HandleFile($ref);
      }
    }
  } else {
    my $ret = $queue->Enqueue(Release => "2.5.5");
    print "Enqueue returned $ret\n";
  }
  $queue = undef;
}

if (!defined caller){
  # Someone called us straight...they must want to test
  main();
}

1;
