### //depot/prod/test/main/lib/Quota.pm#39 - edit change 919876 (ktext)
# $Id: //depot/prod/test/Rsierranevada/lib/Quota.pm#1 $

## @summary Library for Quota related functions
##
## @status public
## 
## @author dl-johny@netapp.com
##         (FSS WAFL Quality)
##
## @keywords Quota
## @pod here

package Quota;


#------------------------------------------------------
#   Imports
#------------------------------------------------------
BEGIN { require "FS/commonConstants.pl"; };
use Carp qw(croak);
use base qw(Exporter);
use Utils qw(make_conn);
use Tharn;
use Module::Common qw(validate_mixed);
use Params::Validate qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF UNDEF);
use filer qw( filer_file_exists );

use SPARSE::Util::Logger;
use Users qw(createUsers createGroups setPrimaryGroup listCIFSUsers
		removeCIFSUser addPasswdEntry addCIFSUser addPasswdList);
use QtreeCmds qw(qtreeCreate qtreePresent);
use SayWrap qw(sayEOF);
use CLI::FileIO qw(rdfile wrfile);
use Filer::Version qw(:releases get_wafl_version UNSUPPORTED_FEATURE_EXCEPTION);
use Data::Dumper;
use TestEnv::MinWin qw(execute_cmd);
use Utils::Units qw(:constants guess_unit strip_unit_spec convert_units convert_to_bytes);
#use setup qw(is_cifs_setup);
use CIFS::Setup qw(is_cifs_setup);
use FS::Util::Setup qw(mount_root_vol_if_needed);
use Volume qw(vol_container);
use Aggregate qw(aggr_status);

################################################################################
#use FS::Util::Utilities qw(util_chmod util_mkfile util_unlink util_stat
#			    util_filesize util_exists util_rmdir delfile deltree);
#use FileManip qw(read_file write_file setattr_file getattr_file stat_file remove_file);
################################################################################


#------------------------------------------------------
#   Module Vars
#------------------------------------------------------
our ($Validation, $VERSION);
my $MOUNTPOINT = "";

#acceptable difference 2 DISK BLOCKS i.e 8 BLOCKS
use constant QUOTA_REPORT_DELTA => 8 * BLOCK_SIZE_KB; 

no warnings("uninitialized");
use strict;


BEGIN
{   

    #--------------------------------------------
    #   Exports
    #--------------------------------------------

    $Validation = 1; # unused as of now
    $VERSION    = '0.01';

    my @constants = qw( QUOTA_REPORT_DELTA );
    my @quota_subs = qw(

	readUsermapFile
	writeUsermapFile
	removeUsermap
	addUsermap
	writeUsermapFileHashList
	writeQuotaFile
	setDefaultQuota
	appendQuotaFile
	appendUsermapFile

	setTreeQuotas
	setGroupQuotas
	setUserQuotas

	setQuota
	setSingleQuota
	setMappedQuota
	setQuotaList
	setTreeQuota
	setGroupQuota
	setUserQuota
	setCIFSUserQuota
	setQuotasForUsers
	setQuotasForCIFSUsers
	setQuotasForGroups
	setQuotasForTrees

	getQuotaTargetAndType

	enableQuota
	enableQuotaLogging

	disableQuota
	removeQuotaEntry
	modifyQuotaEntry

	findQuota
	deleteQuota
	mapQuota
	unmapQuota
	readQuotaFile
	stripQuotaSpec

	resizeQuota
	statusQuota
	statusQuotaAll
	
	verifyQuotaDiskLimit

	queryQuotaReport
	verifyQuotaReport

	convertQuotaValue
	adjustedConvertQuotaValue
	getQuotaReport

	chownCIFSFile
	rshCIFSClient

        quotaUserMapping
	mapUsers
	unmapUser
	mapToCIFSUser
	mapToCIFSUserList
	unmapToCIFSUser
	getCIFSServer
	terminate_cifs

	mapCIFSUser
	mapCIFSUserList
	unmapCIFSUser

	_markLogTail
	_getLogTail
	map_cifs_share

	createQtrees

	vm_stat
	log_quota_error

	_rdfile
	_wrfile

	openEMSLog
	markEMSLog
	closeEMSLog
	pollEMSLogLines
	pollEMSLogRecords
    );

    Module::Common::setup_export_tags(
	quota => [ @quota_subs, @constants ]
    );
}


#------------------------------------------------------
#     Prototypes
#------------------------------------------------------
sub _rdfile($$;$);
sub _wrfile($$$@);



=head1 NAME

Quota - Basic Quota Functions

=head1 SYNOPSIS

	use Quota qw(:quota);

=head1 DESCRIPTION

	This module provides quota related operations

=head1 FUNCTIONS

	All functions are aimed to be quota specific.



=head2 setQuota($filerspec, %quotaopt);

	sets the quota for the specified quota target in /etc/quotas

	returns success or failure

=over 4

=item filerspec

	Required: filer specification

=item quotaopt

	quotaopt	(Required)	A hash describing the quota 
					specification with following
			target	=> ..
			type	=> ..
			disk	=> ..
			files	=> ..
			thold 	=> ..
			sdisk	=> ..
			sfiles	=> ..

=back

=cut 

sub setQuota  
{
	my ($filerspec, %quotaopt) = validate_mixed(@_,
			[{ type => SCALAR|HASHREF }],
			{ 
				target 	=> { type => SCALAR },
				type 	=> { type => SCALAR },
				disk 	=> {type => SCALAR, optional => 1,
					default => "-"},
				files 	=> {type => SCALAR, optional => 1,
					default => "-"},
				thold	=> {type => SCALAR, optional => 1,
					default => "-"},
				sdisk 	=> {type => SCALAR, optional => 1,
					default => "-"},
				sfiles 	=> {type => SCALAR, optional => 1,
					default => "-"},
			}
			);

	return eval {
    		my ($target, $type);
		my ($addcommentline) = 1;
		my ($line);

		my $filerconn = make_conn($filerspec);

		my $newquotafile = "";

		# don't croak here since split may report here if /etc/quotas
		# is empty
		my @oldquotafile = split(/\n/, rdfile($filerconn, 
						"/etc/quotas"));
	
		foreach my $line (@oldquotafile) {
			$addcommentline = 0
				if ($line =~ /\s+target\s+type/);
			
			if (($line =~ /^\s*#/) || ($line =~ /^\s*$/)) {
				$newquotafile .= $line . "\n";
				next;
			}

			$line =~ s/\s*,\s*/,/g;
			($target, $type) = ($line =~ /^\s*(\S+)\s+(\S+)/);

			if (defined($target) && defined($type)) {
				next if ($target eq $quotaopt{target} && 
					 $type eq $quotaopt{type})
			}
			
			$newquotafile .= $line . "\n";
		}

		my $lineout;

		my $commentline = sprintf "# %15s %15s %6s %6s %6s %6s %6s \n", 
				"target", "type", "disk", "files", "thold", 
				"sdisk", "sfiles";
	
		# if old quota file has no comments on column names, add it
		$lineout .= $commentline
					if ($addcommentline == 1);

		$lineout .= sprintf "%15s %15s %6s %6s %6s %6s %6s \n",
			$quotaopt{target}, $quotaopt{type}, $quotaopt{disk},
			$quotaopt{files}, $quotaopt{thold}, $quotaopt{sdisk},
			$quotaopt{sfiles};

    		$lineout =~ s/\s+$//;
		$newquotafile .= $lineout . "\n";

		logTrc("updating quotas file with the rules .. $newquotafile");
		wrfile($filerconn, "/etc/quotas", $newquotafile)
			or croak("can't write to /etc/quotas : $@");
		
		return 1;
	};
}


=head2 _validateQuotaOpt(%quotaopt);

	Validates quota options hash 'quotaopt'

=over 4

=item quota opts

	quotaopt	(Required)	A hash describing the quota 
					specification with following
			target	=> ..
			type	=> ..
			disk	=> ..
			files	=> ..
			thold 	=> ..
			sdisk	=> ..
			sfiles	=> ..
			user => ..
			group => ..
			tree => ..
			vol => ..
			default => ..
			usermap => ..

=back

=cut 

sub _validateQuotaOpt(%)
{
	my ( %quotaopt) = eval {validate_mixed(@_,
			[],
			{ 
				type 	=> { type => SCALAR },
				target 	=> { type => SCALAR|UNDEF|UNDEF, optional => 1},
				default	=> {type => SCALAR|UNDEF, optional => 1,
					default => 0},
				qtree	=> {type => SCALAR|UNDEF, optional => 1,
					default => ""},
				tree	=> {type => SCALAR|UNDEF, optional => 1,
					default => ""},
				id	=> {type => SCALAR|UNDEF, optional => 1,
					default => ""},
				user	=> {type => SCALAR|UNDEF, optional => 1,
					default => ""},
				group	=> {type => SCALAR|UNDEF, optional => 1,
					default => ""},
				vol	=> {type => SCALAR|UNDEF, optional => 1,
					default => ""},
				volume	=> {type => SCALAR|UNDEF, optional => 1,
					default => ""},
				disk 	=> {type => SCALAR|UNDEF, optional => 1,
					default => "-"},
				files 	=> {type => SCALAR|UNDEF, optional => 1,
					default => "-"},
				thold	=> {type => SCALAR|UNDEF, optional => 1,
					default => "-"},
				sdisk 	=> {type => SCALAR|UNDEF, optional => 1,
					default => "-"},
				sfiles 	=> {type => SCALAR|UNDEF, optional => 1,
					default => "-"},
				usermap	=> {type => SCALAR|UNDEF, optional => 1 },
			}
		) };;
	if ($@) {
		logError("_validateQuotaOpt: $@");
	}
	return(1);
}


=head2 stripQuotaSpec($quota_in)

	Returns 'subtable' of the HASHREF quota_in which
	contains only those fields which are valid for a quota spec.

=over 4

=item quota_in

	Required: HASHREF which represents a quota spec. May contain
	additional fields besides those valid for  a quota spec.

	Returns a 'subtable' of quota_in which only has 'valid'
	fields for a quota spec.

=back

=cut

sub stripQuotaSpec($)
{
	my ($in) = validate_mixed(@_,
			[{ type => HASHREF }],
			{}
		);
 	my @valid_fields = qw( type target default qtree tree id user group vol volume
 				disk files thold sdisk sfiles usermap );
 	my %valid_hash = ();
 	foreach (@valid_fields) {
 		$valid_hash{$_} = 1;
 	}
 	my $quota_out = { %$in };
 	foreach my $key (keys(%$in)) {
		if (!$valid_hash{$key}) {
	 		delete($quota_out->{$key});
		}
 	}
 	return( $quota_out );
}


=head2 setSingleQuota($filerspec, %quotaopt);

	sets the quota for the specified quota target in /etc/quotas

	returns success or failure

=over 4

=item filerspec

	Required: filer specification

=item quota opts

	quotaopt	(Required)	A hash describing the quota 
					specification with following
			target	=> ..
			type	=> ..
			disk	=> ..
			files	=> ..
			thold 	=> ..
			sdisk	=> ..
			sfiles	=> ..
			user => ..
			group => ..
			tree => ..
			vol => ..
			default => ..
			usermap => ..

=back

=cut 

sub setSingleQuota($%)
{
	my ($filerspec, %quotaopt) = @_;
	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);
	my $filerconn = make_conn($filerspec);
	$quotaopt{tree} = $quotaopt{tree} || $quotaopt{qtree};
	my $qtree;
        if (exists($quotaopt{tree})) {
		$qtree = $quotaopt{tree};
	} elsif (exists($quotaopt{qtree})) {
		$qtree = $quotaopt{tree};
		$quotaopt{tree} = $qtree;
	}
	delete($quotaopt{qtree});

	_validateQuotaOpt(%quotaopt);
	my ($quota_out) = { %quotaopt };
	($quota_out, undef) = getQuotaTargetAndType($quota_out);
	return( _mergeQuotaEntries($filerspec, [ $quota_out ]) );
}


=head2 setMappedQuota($filerspec, %quotaopt);

	sets the quota for the specified quota target in /etc/quotas,
	with QUOTA_PERFORM_USER_MAPPING enabled for that quota entry

	returns success or failure

=over 4

=item filerspec

	Required: filer specification

=item quota opts

	quotaopt	(Required)	A hash describing the quota 
					specification with following
			target	=> ..
			type	=> ..
			disk	=> ..
			files	=> ..
			thold 	=> ..
			sdisk	=> ..
			sfiles	=> ..
			user => ..
			group => ..
			tree => ..
			vol => ..
			default => ..

=back

=cut 

sub setMappedQuota($%)
{
	my ($filerspec, %quotaopt) = @_;
	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);
	$quotaopt{usermap} = 1;
	return ( setSingleQuota($filerspec, %quotaopt) );
}


=head2 setQuotaList($filerspec, $quotas, %quotaopt);

	Adds a list of quotas to the /etc/quotas file.

	returns success or failure

=over 4

=item filerspec

	Required: filer specification

=item quotas

	Required.  ARRAYREF whose elemets are each a HASHREF which represents a new
			quota entry. The fields of these entries are the same as
			the 'quotaopt' fields for setQuota.

			target	=> ..
			type	=> ..
			disk	=> ..
			files	=> ..
			thold 	=> ..
			sdisk	=> ..
			sfiles	=> ..
			user => ..
			group => ..
			tree => ..
			vol => ..
			default => ..
			usermap => ..


=item quota opts

	quotaopt	Optional
			'Default' values. These values are used to 'fill in'
			any undefined values in the list of quotas.

=back

=cut 

sub setQuotaList($$%)
{
	my ($filerspec, $quotas, %quotaopt) = @_;
	my @validate = ($filerspec, $quotas);
	validate_mixed(@validate,
			[
			    { type => SCALAR|HASHREF },
			    { type => ARRAYREF },
			],
			{}
	);

	my $filerconn = make_conn($filerspec);
	my @quota_out = ();
	foreach my $q (@$quotas) {
		_setQuotaDefaults($q, %quotaopt);
		($q, undef) = getQuotaTargetAndType($q);
		push(@quota_out, $q);
		_validateQuotaOpt(%$q);
	}
	return( _mergeQuotaEntries($filerspec, \@quota_out) );
}


=head2 setQuotasForUsers($filerspec, $users, %quotaopt)

	Given a set of users, adds a quota entry for each of them.
	The user's 'login' is used for the quota entry.

	returns success or failure

=over 4

=item filerspec 

	Required	Filer Specification

=item users

	Required	HASHREF. Each hashtable key/value pair represents an existing user.
			hashtable has same format as that returned by readPasswdFile

=item quota opts

	quotaopt	(Required)	A hash describing the quota 
					specification with following
			tree => ..
			vol => ..
			disk	=> ..
			files	=> ..
			thold 	=> ..
			sdisk	=> ..
			sfiles	=> ..

			users_per_quota => ..   ## for single quota line with multiple users

=back

=cut 

sub setQuotasForUsers($$%)
{
	my ($filerspec, $users, %quotaopt) = @_;

	my @validate = ($filerspec, $users);
	validate_mixed(@validate,
			[
				{ type => SCALAR|HASHREF },
				{ type => HASHREF|ARRAYREF },
			],
			{}
		);

	return eval {
		if (!defined($quotaopt{type})) {
			$quotaopt{type} = 'user';
		}
		my $users_per_quota = 1;
		if (defined($quotaopt{users_per_quota})) {
			$users_per_quota = $quotaopt{users_per_quota};
		}
		delete($quotaopt{users_per_quota});
		my @user_quotas = ();
		my $user_list = [];
		if (ref($users) =~ /ARRAY/) {
			foreach (@$users) {
				if (!ref($_)) {
					push(@$user_list, $_);
				} elsif (ref($_) =~ /HASH/) {
					push(@$user_list, $_->{login});
				}
			}
		} elsif (ref($users) =~ /HASH/) {
			while (my ($uname, $user_ref) = each(%$users)) {
				push(@$user_list, $uname);
			}
			@$user_list = sort(@$user_list);
		}
		for (my $uidx=0; $uidx<=$#$user_list; $uidx+=$users_per_quota) {
			my @ulist = @$user_list;
			splice(@ulist, 0, $uidx) if ($uidx > 0);
			splice(@ulist,  $users_per_quota, $#ulist+1);
			my $unames = join(",", @ulist);
			my $q = { %quotaopt };
			$q->{user} = $unames;
			_validateQuotaOpt(%$q);
			push(@user_quotas, $q);
		}
		return( setQuotaList($filerspec, \@user_quotas, %quotaopt) );
	};
}


=head2 setQuotasForCIFSUsers($filerspec, $users, %quotaopt)

	Given a set of users, adds a quota entry for each of them.
	The user's 'winlogin' is used for the quota entry.
	CIFS must be running on the filer.

	returns success or failure

=over 4

=item filerspec 

	Required	Filer Specification

=item users

	Required	HASHREF. Each hashtable key/value pair represents an existing user.
			hashtable has same format as that returned by readPasswdFile

=item quota opts

	quotaopt	(Required)	A hash describing the quota 
					specification with following
			tree => ..
			vol => ..
			disk	=> ..
			files	=> ..
			thold 	=> ..
			sdisk	=> ..
			sfiles	=> ..

			users_per_quota => ..   ## for single quota line with multiple users

=back

=cut 

sub setQuotasForCIFSUsers($$%)
{
	my ($filerspec, $users, %quotaopt) = @_;

	my @validate = ($filerspec, $users);
	validate_mixed(@validate,
			[
				{ type => SCALAR|HASHREF },
				{ type => HASHREF|ARRAYREF },
			],
			{}
		);

	return eval {
		my $filerconn = make_conn($filerspec);
		my $cifs_server = getCIFSServer($filerconn);
		if ($@ || !$cifs_server) {
			croak("CIFS server is undefined");
		}
		if (!defined($quotaopt{type})) {
			$quotaopt{type} = 'user';
		}
		my $users_per_quota = 1;
		if (defined($quotaopt{users_per_quota})) {
			$users_per_quota = $quotaopt{users_per_quota};
		}
		delete($quotaopt{users_per_quota});
		my @user_quotas = ();
		my $user_list = [];
		if (ref($users) =~ /ARRAY/) {
			foreach (@$users) {
				if (!ref($_)) {
					push(@$user_list, $_);
				} elsif (ref($_) =~ /HASH/) {
					push(@$user_list, $_->{login});
				}
			}
		} elsif (ref($users) =~ /HASH/) {
			while (my ($uname, $user_ref) = each(%$users)) {
				push(@$user_list, $uname);
			}
			@$user_list = sort(@$user_list);
		}
		for (my $uidx=0; $uidx<=$#$user_list; $uidx++) {
			$$user_list[$uidx] = $cifs_server . "\\" . $$user_list[$uidx];
		}
		for (my $uidx=0; $uidx<=$#$user_list; $uidx+=$users_per_quota) {
			my @ulist = @$user_list;
			splice(@ulist, 0, $uidx) if ($uidx > 0);
			splice(@ulist,  $users_per_quota, $#ulist+1);
			my $unames = join(",", @ulist);
			my $q = { %quotaopt };
			$q->{user} = $unames;
			_validateQuotaOpt(%$q);
			push(@user_quotas, $q);
		}
		return( setQuotaList($filerspec, \@user_quotas, %quotaopt) );
	};
}


=head2 setQuotasForGroups($filerspec, $groups, %quotaopt)

	Given a set of groups, adds a quota entry for each of them.

	returns success or failure

=over 4

=item filerspec 

	Required	Filer Specification

=item groups

	Required	HASHREF. Each hashtable key/value pair represents an existing group.
			hashtable has same format as that returned by readGroupFile

=item quota opts

	quotaopt	(Required)	A hash describing the quota 
					specification with following
			tree => ..
			vol => ..
			disk	=> ..
			files	=> ..
			thold 	=> ..
			sdisk	=> ..
			sfiles	=> ..

=back

=cut 

sub setQuotasForGroups($$%)
{
	my ($filerspec, $groups, %quotaopt) = @_;

	my @validate = ($filerspec, $groups);
	validate_mixed(@validate,
			[
				{ type => SCALAR|HASHREF },
				{ type => HASHREF|ARRAYREF },
			],
			{}
		);

	return eval {
		if (!defined($quotaopt{type})) {
			$quotaopt{type} = 'group';
		}
		my @group_quotas = ();
		my $group_list = [];

		if (ref($groups) =~ /ARRAY/) {
			foreach (@$groups) {
				if (!ref($_)) {
					push(@$group_list, $_);
				} elsif (ref($_) =~ /HASH/) {
					push(@$group_list, $_->{group});
				}
			}
		} elsif (ref($groups) =~ /HASH/) {
			while (my ($gname, $group_ref) = each(%$groups)) {
				push(@$group_list, $gname);
			}
		}
		foreach my $gname (@$group_list) {
			my $q = { %quotaopt };
			$q->{group} = $gname;
			_validateQuotaOpt(%$q);
			push(@group_quotas, $q);
		}
		return( setQuotaList($filerspec, \@group_quotas, %quotaopt) );
	};
}


=head2 setQuotasForTrees($filerspec, $trees, %quotaopt)

	Given a set of qtrees, adds a quota entry for each of them.

	returns success or failure

=over 4

=item filerspec 

	Required	Filer Specification

=item qtrees

	Required	ARRAYREF, which gives a list of qtree names.

=item quota opts

	quotaopt	(Required)	A hash describing the quota 
					specification with following
			vol => ..
			disk	=> ..
			files	=> ..
			thold 	=> ..
			sdisk	=> ..
			sfiles	=> ..

=back

=cut 

sub setQuotasForTrees($$%)
{
	my ($filerspec, $trees, %quotaopt) = @_;

	my @validate = ($filerspec, $trees);
	validate_mixed(@validate,
			[
				{ type => SCALAR|HASHREF },
				{ type => ARRAYREF },
			],
			{}
		);

	return eval {
		if (!defined($quotaopt{type})) {
			$quotaopt{type} = 'tree';
		}
		my @tree_quotas = ();
		my $tree_list = [ @$trees ];
		foreach my $tree (@$tree_list) {
			my $q = { %quotaopt };
			$q->{tree} = $tree;
			_validateQuotaOpt(%$q);
			push(@tree_quotas, $q);
		}
		return( setQuotaList($filerspec, \@tree_quotas, %quotaopt) );
	};
}


=head2 setUserQuota($filerspec, %quotaopt)

	Sets a new 'user' quota entry from 'quotaopt'

=over 4

=item filerspec 

	Required Filer Specification

=item options   

	Optional
		default => ..	If '1', then set a 'default' type quota entry
		user	=> ..	Name of user to set quota for.
		vol	=> ..	Volume
		tree	=> ..	Qtree
		disk   => ..	quota size limit
		files  => ..	file limit
		thold  => ..	threshold
		sdisk  => ..	soft disk limit
		sfiles => ..	soft files limit

=back

=cut

sub setUserQuota($%)
{
	my ($filerspec, %quotaopt) = @_;

	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);

	if (!$quotaopt{user}) {
		croak("'user' option is required");
	}
	$quotaopt{type} = 'user';
	$quotaopt{tree} = $quotaopt{tree} || $quotaopt{qtree};
	delete($quotaopt{qtree});
	_validateQuotaOpt(%quotaopt);
	return( setSingleQuota($filerspec, %quotaopt) );
}


=head2 setCIFSUserQuota($filerspec, %quotaopt)

	Sets a new 'user' quota entry from 'quotaopt'
	Fails unless CIFS is running on filer.

=over 4

=item filerspec 

	Required Filer Specification

=item options   

	Optional
		default => ..	If '1', then set a 'default' type quota entry
		user	=> ..	Name of user to set quota for.
		vol	=> ..	Volume
		tree	=> ..	Qtree
		disk   => ..	quota size limit
		files  => ..	file limit
		thold  => ..	threshold
		sdisk  => ..	soft disk limit
		sfiles => ..	soft files limit

=back

=cut

sub setCIFSUserQuota($%)
{
	my ($filerspec, %quotaopt) = @_;

	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);

	my $filerconn = make_conn($filerspec);
	my $cifs_server = getCIFSServer($filerconn);
	if ($@ || !$cifs_server) {
		croak("CIFS server is undefined");
	}
	if (!$quotaopt{user}) {
		croak("'user' option is required");
	}
	$quotaopt{user} = $cifs_server . "\\" . $quotaopt{user};
	$quotaopt{type} = 'user';
	$quotaopt{tree} = $quotaopt{tree} || $quotaopt{qtree};
	delete($quotaopt{qtree});
	_validateQuotaOpt(%quotaopt);
	return( setSingleQuota($filerspec, %quotaopt) );
}


=head2 setGroupQuota($filerspec, %quotaopt)

	Sets a new 'group' quota entry from 'quotaopt'

=over 4

=item filerspec 

	Required Filer Specification

=item options   

	Optional
		default => ..	If '1', then set a 'default' type quota entry
		group	=> ..	Name of group io set quota for.
		vol	=> ..	Volume
		tree	=> ..	Qtree
		disk   => ..	quota size limit
		files  => ..	file limit
		thold  => ..	threshold
		sdisk  => ..	soft disk limit
		sfiles => ..	soft files limit

=back

=cut

sub setGroupQuota($%)
{
	my ($filerspec, %quotaopt) = @_;
	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);

	if (!$quotaopt{group}) {
		croak("'group' option is required");
	}
	$quotaopt{type} = 'group';
	$quotaopt{tree} = $quotaopt{tree} || $quotaopt{qtree};
	delete($quotaopt{qtree});
	_validateQuotaOpt(%quotaopt);
	return( setSingleQuota($filerspec, %quotaopt) );
}


=head2 setTreeQuota($filerspec, %quotaopt)

	Sets a new 'tree' quota entry from 'quotaopt'

=over 4

=item filerspec 

	Required Filer Specification

=item options   

	Optional
		default => ..	If '1', then set a 'default' type quota entry
		vol	=> ..	Volume
		tree	=> ..	Qtree
		disk   => ..	quota size limit
		files  => ..	file limit
		thold  => ..	threshold
		sdisk  => ..	soft disk limit
		sfiles => ..	soft files limit

=back

=cut

sub setTreeQuota($%)
{
	my ($filerspec, %quotaopt) = @_;
	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);
#	$quotaopt{tree} = $quotaopt{tree} || $quotaopt{qtree};
	if ($quotaopt{tree}) {
	} elsif ($quotaopt{qtree}) {
		$quotaopt{tree} = $quotaopt{qtree};
	} else {
		croak("'tree' or 'qtree' option is required");
	}
	$quotaopt{type} = 'tree';
	delete($quotaopt{qtree});
	_validateQuotaOpt(%quotaopt);
	return( setSingleQuota($filerspec, %quotaopt) );
}


=head2 setUserQuota($filerspec, %quotaopt)

	Sets new 'default' quota entry(s) from 'quotaopt'

=over 4

=item filerspec 

	Required Filer Specification

=item options   

	Optional
		vol	=> ..	Volume
		tree	=> ..	Qtree
		disk   => ..	quota size limit
		files  => ..	file limit
		thold  => ..	threshold
		sdisk  => ..	soft disk limit
		sfiles => ..	soft files limit

	For this routine, 'type' may contain one or more of 'user', 'group', 'tree',
	separated by spaces. A Default entry will be set for each of these types.
		type => "user"
		type => "user tree"
		type => "tree group"
		type => "user group tree"

=back

=cut

sub setDefaultQuota($%)
{
	my ($filerspec, %quotaopt) = @_;
	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);

	$quotaopt{default} = 1;
	my @types = split(/\s+/, $quotaopt{type});
	foreach my $t (@types) {
		$quotaopt{tree} = $quotaopt{tree} || $quotaopt{qtree};
		delete($quotaopt{qtree});
		$quotaopt{type} = $t;
		_validateQuotaOpt(%quotaopt);
		setSingleQuota($filerspec, %quotaopt);
	}
}


=head2 appendQuotaFile($filerspec, @textlist)

	Appends '$text' to /etc/quotas file.

	returns success or failure

=cut 

sub appendQuotaFile($@)
{
        my ($filerspec, @text_list) = @_;
        my $text = join("\n", @text_list);
        $text .= "\n" unless ($text =~ /\n$/);

	my $filerconn = make_conn($filerspec);
	my $old_text = rdfile($filerconn, "/etc/quotas");
	croak($@) if ($@);
	$old_text .= "\n" unless ($old_text =~ /\n$/);
	wrfile($filerconn, "/etc/quotas", $old_text . $text)
		or croak("can't write to /etc/quotas : $@");
	return(1);
}


=head2 appendUsermapFile($filerspec, @textlist)

	Appends '$text' to /etc/usermap.cfg file.

	returns success or failure

=cut 

sub appendUsermapFile($@)
{
        my ($filerspec, @text_list) = @_;
        my $text = join("\n", @text_list);
	my $filerconn = make_conn($filerspec);
	my $old_text = rdfile($filerconn, "/etc/usermap.cfg");
	croak($@) if ($@);
	$old_text .= "\n" unless ($old_text =~ /\n$/);
	wrfile($filerconn, "/etc/usermap.cfg", $old_text . $text)
		or croak("can't write to /etc/usermap.cfg : $@");
	sleep(4);
	return(1);
}


=head2 writeQuotaFile($filerspec, @textlines)

	Appends '$text' to /etc/quotas file.

	returns success or failure

=cut 

sub writeQuotaFile($@)
{
        my ($filerspec, @text_list) = @_;
        my $text = join("\n", @text_list);
        $text .= "\n" unless ($text =~ /\n$/);

	my $filerconn = make_conn($filerspec);
	wrfile($filerconn, "/etc/quotas", $text)
			or croak("can't write to /etc/quotas : $@");
	return(1);
}


=head2 writeUsermapFile($filerspec, @textlines)

	Writes '$text' to /etc/usermap.cfg file.
	Existing /etc/usermap.cfg is overwritten (not appended).

	returns success or failure

=cut 

sub writeUsermapFile($@)
{
        my ($filerspec, @text_list) = @_;
        my $text = join("\n", @text_list);
	my $filerconn = make_conn($filerspec);
	wrfile($filerconn, "/etc/usermap.cfg", $text)
			or croak("can't write to /etc/usermap.cfg : $@");
	sleep(4);
	return(1);
}


=head2 writeUsermapFileHashList($filerspec, @hashlist)

	Creates new /etc/usermap.cfg file based on contents of @hashlist.
	Each member of @hashlist is a HASHREF with same format as that
	returned by readUsermapFile
	returns success or failure

=over 4

=item filerspec

	Required: filer specification

=item	hashlist

	Required: List of HASHREF representing entries to be written to /etc/usermap.cfg.

=back

=cut 

sub writeUsermapFileHashList($@)
{
        my ($filerspec, @hashlist) = @_;
	my $filerconn = make_conn($filerspec);
	my @lines = ();
	foreach my $entry (@hashlist) {
		my $line = join(" ",
			        ($entry->{cifs_user}, $entry->{direction}, $entry->{nfs_user}));
		push(@lines, $line);
	}
	writeUsermapFile($filerconn, @lines);
}

=head2 addUsermap($filerspec, %opts)

	Adds specified user mapping to /etc/usermap.cfg

=over 4

=item filerspec

	Required: filer specification

=item opts

	cifs_user => ..  optional.

	nfs_user =>  ..  optional.

	direction => ..  optional.

=back

=cut 

sub addUsermap($%)
{
    my ($filerspec, %opts) = validate_mixed(@_, 
			[
			{ type => SCALAR|HASHREF },
			],
			{
				cifs_user  => { type => SCALAR, optional => 1,
					default => "*\\*" },
				nfs_user  => { type => SCALAR, optional => 1,
					default => "*" },
				direction  => { type => SCALAR, optional => 1,
					default => "==" },
				add_workgroup  => { type => SCALAR, optional => 1,
					default => 1 },
				add_cifs_user_quotes  => { type => SCALAR, optional => 1,
					default => 0 },
				match_workgroup  => { type => SCALAR, optional => 1,
					default => 0 },
			}
		);
	my $fconn = make_conn($filerspec);
	my $cifs_user = $opts{cifs_user};
	my $nfs_user = $opts{nfs_user};
	my $direction = $opts{direction};
	my $workgroup = getCIFSServer($fconn) || "*";
	my $match_cifs_user = sub {
					my ($test_user) = @_;
					my $cifs_user_part = $cifs_user;
					my $test_user_part = $test_user;
					$cifs_user_part =~ s/\"//g;
					$test_user_part =~ s/\"//g;
					if ($opts{add_workgroup} && $cifs_user_part !~ m%\\%) {
						$cifs_user_part = "$workgroup\\" . $cifs_user_part;
					}
					if ($opts{add_workgroup} && $test_user_part !~ m%\\%) {
						$test_user_part = "$workgroup\\" . $test_user_part;
					}
					if ($opts{match_workgroup}) {
						return($cifs_user_part eq $test_user_part);
					} else {
						$cifs_user_part =~ s/^(.*\\)//;
						$test_user_part =~ s/^(.*\\)//;
						return($cifs_user_part eq $test_user_part);
					}
				  };
	if (!$cifs_user) {
		croak("cifs_user cannot be NULL");
	}
	$cifs_user =~ s/\"//g;
	if ($opts{add_workgroup} && $cifs_user !~ m%\\%) {
		$cifs_user = "$workgroup\\" . $cifs_user;
	}
	if ($opts{add_cifs_user_quotes}) {
		$cifs_user = "\"$cifs_user\"";
	}
	if ($direction !~ /(==|=>|<=)/) {
		croak("direction must be one of '<=', '==', '=>'");
	}
	my $new_entry = { cifs_user => $cifs_user,
			  nfs_user => $nfs_user,
			  direction => $direction };
	my @usermap_list = readUsermapFile($fconn);
	my @filtered_usermap_list = ();
	foreach my $entry (@usermap_list) {
		if ($direction eq "==") {
			if (&$match_cifs_user($entry->{cifs_user}) ||
			    $nfs_user eq $entry->{nfs_user}) {
				next;
			}
		} elsif ($direction eq "=>") {
			if (&$match_cifs_user($entry->{cifs_user})) {
				next;
			}
		} elsif ($direction eq "<=") {
			if ($nfs_user eq $entry->{nfs_user}) {
				next;
			}
		}
		push(@filtered_usermap_list, $entry);
	}
	### push(@filtered_usermap_list, $new_entry);
	### writeUsermapFileHashList($fconn, @filtered_usermap_list);
	push(@usermap_list, $new_entry);
	writeUsermapFileHashList($fconn, @usermap_list);
}


=head2 removeUsermap($filerspec, %opts)

	Removes specified user mapping(s) from /etc/usermap.cfg

=over 4

=item filerspec

	Required: filer specification

=item opts

	cifs_user => ..  optional.

	nfs_user =>  ..  optional.

	direction => ..  optional.

=back

=cut 

sub removeUsermap($%)
{
    my ($filerspec, %opts) = validate_mixed(@_, 
			[
			{ type => SCALAR|HASHREF },
			],
			{
				cifs_user  => { type => SCALAR, optional => 1,
					default => undef },
				nfs_user  => { type => SCALAR, optional => 1,
					default => undef },
				direction  => { type => SCALAR, optional => 1,
					default => undef },
				add_workgroup  => { type => SCALAR, optional => 1,
					default => 1 },
				match_workgroup  => { type => SCALAR, optional => 1,
					default => 0 },
			}
		);
	my $fconn = make_conn($filerspec);
	my $cifs_user = $opts{cifs_user};
	my $nfs_user = $opts{nfs_user};
	my $direction = $opts{direction};
	my $workgroup = getCIFSServer($fconn) || "*";
	my $match_cifs_user = sub {
					my ($test_user) = @_;
					my $cifs_user_part = $cifs_user;
					my $test_user_part = $test_user;
					$cifs_user_part =~ s/\"//g;
					$test_user_part =~ s/\"//g;
					if ($opts{add_workgroup} && $cifs_user_part !~ m%\\%) {
						$cifs_user_part = "$workgroup\\" . $cifs_user_part;
					}
					if ($opts{add_workgroup} && $test_user_part !~ m%\\%) {
						$test_user_part = "$workgroup\\" . $test_user_part;
					}
					if ($opts{match_workgroup}) {
						return($cifs_user_part eq $test_user_part);
					} else {
						$cifs_user_part =~ s/^(.*\\)//;
						$test_user_part =~ s/^(.*\\)//;
						return($cifs_user_part eq $test_user_part);
					}
				  };
	if (defined($cifs_user) && !$cifs_user) {
		croak("cifs_user cannot be NULL");
	}
	$cifs_user =~ s/\"//g;
	if ($opts{add_workgroup} && $cifs_user !~ m%\\%) {
		$cifs_user = "$workgroup\\" . $cifs_user;
	}
	$cifs_user = "\"$cifs_user\"";
	if (defined($direction) && $direction !~ /(==|=>|<=)/) {
		croak("direction must be one of '<=', '==', '=>'");
	}
	my @usermap_list = readUsermapFile($fconn);
	my @filtered_usermap_list = ();
	foreach my $entry (@usermap_list) {
		if ( (!defined($direction) || $direction eq $entry->{direction}) &&
		     (!defined($cifs_user) || &$match_cifs_user($entry->{cifs_user})) &&
		     (!defined($nfs_user) || $nfs_user eq $entry->{nfs_user}) ) {
			next;
		}
		push(@filtered_usermap_list, $entry);
	}
	writeUsermapFileHashList($fconn, @filtered_usermap_list);
}



=head2 readUsermapFile($filerspec)

	Reads '/etc/usermap.cfg', and returns a list of HASHREF which represnts
	the entries of this file. Each HASHREF has the format:
		{
			cifs_user => xxxx,
			nfs_user => xxxx,
			direction => xxxx
		}
=over 4

=item filerspec

	Required: filer specification

=back

=cut

sub readUsermapFile($)
{
	my ($filerspec) = @_;
	my $fconn = make_conn($filerspec);
	my $file = "/etc/usermap.cfg";
	my @usermap_list = ();
	my $file_txt;
	if (filer_file_exists($fconn, $file)) {
		$file_txt = rdfile($fconn, "$file");
		die($@) if ($@);
	} else {
		return( @usermap_list );
	}
	my @old_lines = split(/\n/, $file_txt);
	foreach (@old_lines) {
		next if (/^\s*(#.*)?$/);
		$_ =~ s/^\s*//;
		$_ =~ s/\s*$//;
		my $entry = {};
		($entry->{cifs_user}, $entry->{direction}, $entry->{nfs_user}) =
		    split(/\s+/);
		push(@usermap_list, $entry);
	}
	return( @usermap_list );
}



=head2 enableQuota($filerspec, $vol, %opts)
	
	Enable quota for specified volume on the filer

=over 4

=item filerspec

	Required: filer specification

=item vol

	Required: volume name

=item opts

	timeout => ..		Optional

	reenable => ..		Optional

	nowait => ..    	Optional.
			            If '1', then omit '-w' flag in 'quota on' command.

	dontcheck => ..	 	Optional.
				    If '1', then don't do status check after turning
				    quotas on.

	poll_interval => ..	Optional. Default is 10 sec.
 
=back

=cut

sub enableQuota($$%)
{
    my ($spec, $vol, %options) = validate_mixed(@_, 
			[
			{ type => SCALAR|HASHREF },
			{ type => SCALAR }
			],
			{
				timeout  => { type => SCALAR, optional => 1,
					default => $main::CMDTIMEOUT},

				poll_interval  => { type => SCALAR, optional => 1,
					default => 10},

				reenable => { type => SCALAR, optional => 1,
					default => 0},

				nowait   => { type => SCALAR, optional => 1,
					default => undef},

				dontcheck   => { type => BOOLEAN, optional => 1,
					default => undef},
			}
			);

    logDbg("Enter enableQuota: $vol  Options:\n\n".Dumper(\%options));
    return eval {
	my $conn = make_conn($spec);
	my $poll_interval = $options{poll_interval};
	my $timeout = $options{timeout} || $main::CMDTIMEOUT || 300;
	my $ver = get_wafl_version($conn->hostp());
	if (!defined($options{nowait})) {
		if ($ver < WV_ANCHORSTEAM_MIN) {
			$options{nowait} = 1;
		} elsif (!defined($options{nowait})) {
			$options{nowait} = 0;
		}
	}

	if ($options{reenable} && statusQuota($conn, $vol) ne "off") {
		say($conn, "quota off $vol", -timeout, $timeout);
		my (undef, undef, $disable_msg) = say($conn, "quota off $vol", -timeout, $timeout);
		if ($disable_msg !~ /(^$|quotas( are)? already off)/) {
			croak("enableQuota: could not disable quotas for $vol: $disable_msg");
			# logWarn("enableQuota: could not disable quotas for $vol: $disable_msg");
			# return(0);
		}
		for (my $tm = 0; $tm < $timeout; $tm += $poll_interval) {
	    		last if (statusQuota($conn, $vol) eq "off");
			sleep($poll_interval);
		}
	}

	my $flags = "";
	if (!$options{nowait}) {
		$flags = "-w";
	}

	$@ = "";
	my (undef, undef, $enable_msg) = say($conn, "quota on $flags $vol", -timeout, $timeout);
	if ($enable_msg !~ /(Starting quota initialization|quotas( are)? already on)/) {
		croak("enableQuota: could not enable quotas: $enable_msg");
		# logWarn("enableQuota: could not enable quotas: $enable_msg");
		# return(0);
	}
        if ($options{dontcheck}) {
    	    logDbg("Exit enableQuota: $vol");
            return (1);
	}

	for (my $tm = 0; $tm < $timeout; $tm += $poll_interval) {
	    sleep($poll_interval);
	    if (statusQuota($conn, $vol) eq "on") {
    	    	logDbg("Exit enableQuota: $vol");
	    	return (1);
	    }
	}
	croak("could not enable quotas for $vol, check message logs");
    };
}


=head2 statusQuota($filerspec, $vol, %opts)
	
	Check quota status for specified volume on the filer

=over 4

=item filerspec

	Required: filer specification

=item vol

	Required: volume name

=item opts hash

	wait => ..		Optional. If '1', then function will 'poll' quota status
				    until it reaches a 'final' state -- 'on' or 'off',
				    or until timeout is reached.

	poll_interval => ..	Optional. Default is 10 sec.

        percentage => ..        Optional. Return the percentage complete
                                    when the state is initializing if set to 1.
                                    Default is 0.

	timeout => ..		Optional. 

=back

=cut

sub statusQuota($$%)
{
	my ($spec, $vol, %opts) = validate_mixed(@_, 
			[
			{ type => SCALAR | HASHREF},
			{ type => SCALAR }
			], 
			{
				wait  => { type => SCALAR, optional => 1,
					      default => 0},

				poll_interval  => { type => SCALAR, optional => 1,
					      		default => 10},

                                percentage  => { type => BOOLEAN, default => 0},

				timeout  => { type => SCALAR, optional => 1, default => $main::CMDTIMEOUT},
			}
			);

    return eval {
	my $conn = make_conn($spec);
	my $ver = get_wafl_version($conn->hostp());
	my $timeout = defined($opts{timeout}) ? $opts{timeout} : 300;
	my $poll_interval = $opts{poll_interval};
	my $status_cmmd;
	if ($ver < WV_GORDONBIERSCH_MIN) {
		$status_cmmd = "quota";
	} else{ 
		$status_cmmd = "quota status $vol";
	}

	my $status;
	my $elapsed = 0;
	while ( 1 ) {
		my $status = sayEOF($conn, $status_cmmd, timeout=>$timeout);
		my @lines = split(/\n/, $status);
		my $line;
		my $msg;
		foreach (@lines) {
			if (/^$vol: (.*)/) {
				$line = $_;
				$msg = $1;
				last;
			}
		}
		if (!$line) {
			croak("No quota status for $vol");
		}
		if ($msg =~ /quota resize loading/ || $msg =~ /resize in progress/) {
			$status = "resize";
		} elsif ($msg =~ /quotas are initializing/) {
                        if ($opts{percentage}) {
                            $msg =~ /\((\d+)\% done\)/;
                            $status = "initialize: $1 percent done";
                        } else {
			$status = "initialize";
                        }
		} elsif ($msg =~ /quotas are (on|off)/) {
			$status = $1;
		} else {
			croak("can't parse quota status for $vol:\n\n$line");
		}
		my $final = ($status =~ /(on|off)/);
		if ( !$opts{wait} || $final || ($elapsed > $timeout) ) {
			return($status);
		}
		sleep($poll_interval);
		$elapsed += $poll_interval;
	}
	return($status);
    };
}


=head2 statusQuotaAll($filerspec %opts)
	
	Check quota status for all volumes on the filer
	Returns ARRAY representing 'quota status' output

=over 4


=item filerspec

	Required: filer specification

=item opts hash

	wait => ..		Optional. If '1', then function will 'poll' quota status
				    until it reaches a 'final' state -- 'on' or 'off',
				    or until timeout is reached.

	poll_interval => ..	Optional. Default is 10 sec.

	timeout => ..		Optional. 

=back

=cut

sub statusQuotaAll($%)
{
	my ($spec, %opts) = validate_mixed(@_, 
			[
			{ type => SCALAR | HASHREF},
			], 
			{
				wait  => { type => SCALAR, optional => 1,
					      default => 0},

				poll_interval  => { type => SCALAR, optional => 1,
					    		  default => 10},

				timeout  => { type => SCALAR, optional => 1,
					      default => $main::CMDTIMEOUT},
			}
			);

    return eval {
	my $conn = make_conn($spec);
	my $ver = get_wafl_version($conn->hostp());
	my $timeout = defined($opts{timeout}) ? $opts{timeout} : 300;
	my $poll_interval = $opts{poll_interval};
	my $status_cmmd;
	if ($ver < WV_GORDONBIERSCH_MIN) {
		$status_cmmd = "quota";
	} else{ 
		$status_cmmd = "quota status";
	}

	my @status_list = ();
	my $elapsed = 0;
	while ( 1 ) {
		my $status = sayEOF($conn, $status_cmmd, timeout=>$timeout);
		my @lines = split(/\n/, $status);
		my $count = 0;
		my $final = 1;
		foreach my $line (@lines) {
			if ($line =~ /^([\w\_\-\.]+): (.*)/) {
				my $vol = $1;
				my $msg = $2;
				my $status;
				if ($msg =~ /quota resize loading/ || $msg =~ /resize in progress/) {
					$status = "resize";
				} elsif ($msg =~ /quotas are initializing/) {
					$status = "initialize";
				} elsif ($msg =~ /quotas are (on|off)/) {
					$status = $1;
				} else {
					croak("can't parse quota status\n\n$line");
				}
				$final &= ($status =~ /(on|off)/);
				push( @status_list, { volume => $vol, status => $status, volnum => $count++ } );
			}
		}
		$elapsed += $poll_interval;
		if ( !$opts{wait} || $final || ($elapsed > $timeout) ) {
			logDbg("statusQuotaAll  Returning:\n\n".Dumper(\@status_list));
			return( @status_list );
		}
		sleep($poll_interval);
	}
	return( @status_list );
    };
}


=head2 disableQuota($filerspec, $vol, %opts)
	
	Disable quota for specified volume on the filer

=over 4

=item filerspec

	Required: filer specification

=item vol

	Required: volume name

=item opts hash

	timeout => ..		Optional

	poll_interval => ..	Optional. Default is 10 sec.

=back

=cut

sub disableQuota($$%)
{
	my ($spec, $vol, %options) = validate_mixed(@_, 
			[
			{ type => SCALAR | HASHREF},
			{ type => SCALAR }
			],
			{
			      poll_interval  => { type => SCALAR, optional => 1,
				              default => 10},

			      timeout  => { type => SCALAR, optional => 1,
				              default => $main::CMDTIMEOUT},
			}
			);

    logDbg("Enter disableQuota: $vol  Options:\n\n".Dumper(\%options));
    return eval {
	my $conn = make_conn($spec);
	my $timeout = defined($options{timeout}) ? $options{timeout} : 300;
	my $poll_interval = $options{poll_interval};
	my $done = 0;
	my $disable_msg;
	for (my $tm = 0; $tm < $timeout; $tm += $poll_interval) {
		(undef, undef, $disable_msg) = say($conn, "quota off $vol", -timeout, $timeout);
		$disable_msg =~ s/^\s*//;
		$disable_msg =~ s/\s*$//;
	     	if ($disable_msg =~ /quota initialization is finishing/i) {
			$done = 0;
	     	} elsif ($disable_msg =~ /try .* again/i) {
			$done = 0;
		} elsif ($disable_msg !~ /(^$|quotas( are)? already off)/) {
			croak("disableQuota: could not disable quotas: $disable_msg");
			# logWarn("disableQuota: could not disable quotas: $disable_msg");
			# return(0);
		} else {
			$done = 1;
			last;
		}
		sleep($poll_interval);
	}
	unless ($done) {
		croak("disableQuota: could not disable quotas: $disable_msg");
	}
	for (my $tm = 0; $tm < $timeout; $tm += $poll_interval) {
    		if (statusQuota($conn, $vol) eq "off") {
    		    logDbg("Exit disableQuota: $vol");
    		    return (1);
		}
		sleep($poll_interval);
	}
	logcomment("Exit (FAIL) disableQuota  Line: ".__LINE__);
	croak("could not disable quotas for $vol, check message logs");
    };
}


=head2 removeQuotaEntry($filerspec, $pattern)

	Remove specified entry from /etc/quotas file

=over 4

=item filerspec

	Required: filer specification

=item pattern

	Required: entry to remove from /etc/quotas
	May be SCALAR for single pattern or ARRAYREF for multiple patterns
	If multiple patterns are given, then all patterns must match
	quota entry for this entry to be removed.

=back

=cut

sub removeQuotaEntry($$)
{
	my ($filerspec, $pattern) = validate_mixed(@_,
			[
			{ type => SCALAR | HASHREF},
			{ type => SCALAR | ARRAYREF}
			], {}
			);

	my @patterns;
	if (ref($pattern) eq 'ARRAY') {
		@patterns = @$pattern;
	} else {
		@patterns = ($pattern);
	}
	my $filerconn = make_conn($filerspec);

	my $newlines = "";

	my $linebuff = rdfile($filerconn, "/etc/quotas");
	croak($@) if ($@);
	my @lines = split(/\n/, $linebuff);
	return(1) unless (@lines);

	foreach my $line (@lines) {
		my $match = 1;
		foreach my $patt (@patterns) {
			if ($line !~ /$patt/) {
				$match = 0;
				last;
			}
		}
		if (!$match) {
			$newlines .= $line . "\n";
		}
	}

	wrfile($filerconn, "/etc/quotas", $newlines)
		or croak("can't write to /etc/quotas : $@");
	return(1);
}


=head2 modifyQuotaEntry($filerspec, $pattern)

	Modify specified (existing) entry from /etc/quotas file
	with new values

=over 4

=item filerspec

	Required: filer specification

=item pattern

	Required: entry to remove from /etc/quotas
	May be SCALAR for single pattern or ARRAYREF for multiple patterns
	If multiple patterns are given, then all patterns must match
	quota entry for this entry to be removed.

=item opts

	Optional args for modifying quota entry:
		disk => ...
		files => ...

=back

=cut

sub modifyQuotaEntry($$%)
{
	my ($filerspec, $pattern, %opts) = validate_mixed(@_,
			[
				{ type => SCALAR | HASHREF},
				{ type => SCALAR | ARRAYREF}
			],
			{
				disk 	=> { type => SCALAR, optional => 1 },
				files 	=> { type => SCALAR, optional => 1 },
			}
			);

	my @patterns;
	if (ref($pattern) eq 'ARRAY') {
		@patterns = @$pattern;
	} else {
		@patterns = ($pattern);
	}
	my $filerconn = make_conn($filerspec);

	my $newlines = "";
	my $found = 0;
	my $linebuff = rdfile($filerconn, "/etc/quotas");
	croak($@) if ($@);
	my @lines = split(/\n/, $linebuff);
	return(0) unless (@lines);

	foreach my $line (@lines) {
		$line =~ s/^\s*//;
		my $match = 1;
		foreach my $patt (@patterns) {
			if ($line !~ /$patt/) {
				$match = 0;
				last;
			}
		}
		if ($match) {
			$found = 1;
			my @fields = split(/\s+/, $line);
			$fields[2] = $opts{disk} if (defined($opts{disk}));
			$fields[3] = $opts{files} if (defined($opts{files}));
			$newlines .= join("\t", @fields) . "\n";
		} else {
			$newlines .= $line . "\n";
		}
	}

	wrfile($filerconn, "/etc/quotas", $newlines)
		or croak("can't write to /etc/quotas : $@");
	return($found);
}


#
# _findInList
#
#	Searches a given ARRAYREF for matching members, and returns
#	any matches.
#
# Args:
#
#	list:		Required. ARRAYREF from which matching members
#			are to be chosen
#
#	args:		Key, value pairs which are used to select
#			matching members

sub _findInList($%)
{
	my ($arraylist, %args) = @_;
	my (@keys) = keys(%args);
	my @matches = ();
	for my $key (@keys) {
		delete($args{$key}) unless (defined($args{$key}));
	}
	@keys = keys(%args);
	for my $elt (@$arraylist) {
		my $match = 1;
		my $args_ref = { %args };
		while ( my ($key, $value) = each(%args)) {
			if ($elt->{$key} ne $value) {
				$match = 0;
				last;
			}
		}
		if ($match) {
			push(@matches, $elt);
		}
	}
	return(@matches);
}


=head2 readQuotaFile($filerspec)

	Reads /etc/quotas file, parses entries, and returns
	list (ARRAYREF) of quota entries

=over 4

=item filerspec

	Required: filer specification

=item opts

	rootvolmnt => ...	If given, specifies mount directory
				for root volume. NFS will be used to
				access /etc/quotas file.

=back

=cut

sub readQuotaFile
{
	my ($filerspec, %opts) = validate_mixed(@_,
			[
				{ type => SCALAR|HASHREF},
			],
			{
				rootvolmnt => { type => SCALAR, optional => 1 },
			},
	);
	my $filerconn = make_conn($filerspec);

	my @lines = _rdfile($filerspec, "/etc/quotas", $opts{rootvolmnt});
	my $quota_list = [];
	my $usermap = "OFF";
	my $line_num = 0;
	foreach my $line (@lines) {
		$line =~ s/^\s*//;
		$line =~ s/\s*$//;
		$line_num++;
		next if ($line =~ /^\s*(#.*)?$/);
		if ($line =~ /^\s*QUOTA_PERFORM_USER_MAPPING\s+(ON|OFF)/i) {
			$usermap = uc($1);
			next;
		}
		my $qinfo = _parse_quotas_file_line($line);
		$qinfo->{line_number} = $line_num;
		$qinfo->{usermap} = ($usermap eq "ON");
		push(@$quota_list, $qinfo);
	}
	return ( $quota_list );
}


=head2 findQuota($filerspec, %opts)

	Returns list of quota entries that match the criteria
	given by %opts

=over 4

=item filerspec

	Required: filer specification

=item opts

	Optional args for specifying "match" criteria
		target => ...
		type => ...
		user => ...
		qtree => ...
		volume => ...

=back

=cut

sub findQuota
{
	my ($filerspec, %opts) = @_;
	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);
		
	my $opts_copy = { %opts };
	(undef, $opts_copy) = getQuotaTargetAndType($opts_copy);
	my $filerconn = make_conn($filerspec);
	my $quota_entries = readQuotaFile($filerconn);
	return( _findInList($quota_entries, %$opts_copy) );
}


=head2 deleteQuota($filerspec, %opts)

	Deletes list of quota entries that match the criteria
	given by %opts

=over 4

=item filerspec

	Required: filer specification

=item opts

	Optional args for specifying "match" criteria
		target => ...
		type => ...
		user => ...
		qtree => ...
		volume => ...

=back

=cut

sub deleteQuota
{
	my ($filerspec, %opts) = @_;
	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);
		
###	_validateQuotaOpt(%opts);

	my $filerconn = make_conn($filerspec);
	my @matched_entries = findQuota($filerconn, %opts);
	my @to_delete = ();
	foreach (@matched_entries) {
		my $bare_quota = stripQuotaSpec($_);
		push(@to_delete, $bare_quota);
	}
	return( _mergeQuotaEntries($filerspec, undef, \@to_delete) );
}


=head2 mapQuota($filerspec, %opts)

	"Maps" the list of quota entries that match the criteria
	given by %opts by preceding them with a
	"QUOTA_PERFORM_USER_MAPPING ON" directive.

=over 4

=item filerspec

	Required: filer specification

=item opts

	Optional args for specifying "match" criteria
		target => ...
		type => ...
		user => ...
		qtree => ...
		volume => ...

=back

=cut

sub mapQuota
{
	my ($filerspec, %opts) = @_;
	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);
		
	my $filerconn = make_conn($filerspec);
	my @to_map = findQuota($filerconn, %opts);
	foreach my $q (@to_map) {
		$q->{usermap} = 1;
	}
	return( setQuotaList($filerconn, \@to_map) );
}


=head2 unmapQuota($filerspec, %opts)

	"Unmaps" the list of quota entries that match the criteria
	given by %opts by preceding them with a
	"QUOTA_PERFORM_USER_MAPPING OFF" directive.

=over 4

=item filerspec

	Required: filer specification

=item opts

	Optional args for specifying "match" criteria
		target => ...
		type => ...
		user => ...
		qtree => ...
		volume => ...

=back

=cut

sub unmapQuota
{
	my ($filerspec, %opts) = @_;
	my @validate = ($filerspec);
	validate_mixed(@validate,
			[{ type => SCALAR|HASHREF }],
			{}
		);
		
	my $filerconn = make_conn($filerspec);
	my @to_unmap = findQuota($filerconn, %opts);
	foreach my $q (@to_unmap) {
		$q->{usermap} = 0;
	}
	return( setQuotaList($filerconn, \@to_unmap) );
}


=head2 	enableQuotaLogmsg()

=cut

sub enableQuotaLogmsg()
{
	croak "TBD";
}

=head2 resizeQuota($filerspec, $vol)

	Execute 'quota resize' command on the given volume.

	Returns:  "0" if failed. Otherwise returns "1", and status message (if any).

=over 4

=item filerspec 

	Required Filer Specification

=item vol	  

	Required Volume name

=item opts

	timeout => ..		Optional

	poll_interval => ..	Optional. Default is 10 sec.

	dontcheck => ..	 	Optional
				    If '1', then don't do status check after turning
				    quotas on.

=back 

=cut

sub resizeQuota($$%)
{
	my ($spec, $vol, %options) = validate_mixed(@_, 
			[
			{ type => SCALAR | HASHREF},
			{ type => SCALAR }
			],
			   {
			        poll_interval  => { type => SCALAR, optional => 1,
				              default => 10},

                                timeout  => { type => SCALAR, optional => 1,
                                                 default => $main::CMDTIMEOUT},

				dontcheck   => { type => BOOLEAN, optional => 1,
					default => undef},
			   }
			);

    logDbg("Enter resizeQuota: $vol  Options:\n\n".Dumper(\%options));
    return eval {
	my $conn = make_conn($spec);
	my $timeout = defined($options{timeout}) ? $options{timeout} : 300;
	my $poll_interval = $options{poll_interval};
	my (undef, undef, $resize_msg) = say($conn, "quota resize $vol", -timeout, $timeout);
	# my $status = statusQuota($spec, $vol);
	# return( ($status eq 'on') );
	# return wantarray ? (1, $resize_msg) : 1;
	my $check_passed = 0;
        if ($options{dontcheck}) {
		$check_passed = 1;
	}
	for (my $tm = 0; $tm < $timeout; $tm += $poll_interval) {
    		sleep($poll_interval);
    		if (statusQuota($conn, $vol) eq "on") {
			$check_passed = 1;
			last;
		}
	}
	if (!$check_passed) {
		croak("could not resize quotas for $vol, check message logs");
	}
        logDbg("Exit resizeQuota: $vol");
	if (wantarray()) {
		return( (1, $resize_msg) );
	} else {
		return( 1 );
	}
    };
}


=head2 	setTreeQuotas($filerspec, $vol, $qtree, %opts)

	Sets Tree Type Quota rules

=over 4

=item filerspec 

	Required Filer Specification

=item vol	  

	Required Volume name

=item qtree

	Required qtree name to create and set rule for

=item options hash

	Optional
		disk   => ..	quota size limit
		files  => ..	file limit
		thold  => ..	threshold
		sdisk  => ..	soft disk limit
		sfiles => ..	soft files limit

=back 

=cut

sub setTreeQuotas($$$%)
{
	my ($filerspec, $vol, $qtree, %opts)
			= validate_mixed(@_,
				[
				{ type => SCALAR|HASHREF },
				{ type => SCALAR },
				{ type => BOOLEAN }
				  # qtree is defined as BOOLEAN so that it can 
				  # be undef. If undef, we create a quota rule
				  # with target = '*'
				],
				{ 
				disk 	=> { type => SCALAR, optional => 1 },
				files 	=> { type => SCALAR, optional => 1 },
				thold 	=> { type => SCALAR, optional => 1 },
				sdisk 	=> { type => SCALAR, optional => 1 },
				sfiles	=> { type => SCALAR, optional => 1 },
				},
			);
	return eval {
		my $qtreepath = "/vol/$vol";
		my ($target, $type);
	
		if ($qtree) {
			# create qtree only if we are supposed to
			logTrc("create qtree and extract qtreepath");
			$qtreepath = _create_qtree($filerspec, $vol, $qtree)
				or croak $@;
			$target = $qtreepath;
			$type = "tree";
		} else {
			# if no qtree is defined, target would be '*'
			$target = "*";
			$type = "tree\@$qtreepath";
		}
	
		logTrc("set Tree quota rule on $filerspec ". Dumper(\%opts));
		setQuota($filerspec, target => $target, 
			type => $type, %opts)
				or croak $@;

		return 1;
	};
}


=head2 setGroupQuotas($filerspec, $vol, $qtree, $nusers, %opts)

	Sets Group Type Quota rules

=over 4

=item filerspec 

	Required Filer Specification

=item vol

	Required Volume name

=item qtree

	Required qtree name to create and set rule for

=item nusers

	Required number of users to set quota rules for

=item options hash

	Optional
		disk   => ..	quota size limit
		files  => ..	file limit
		thold  => ..	threshold
		sdisk  => ..	soft disk limit
		sfiles => ..	soft files limit

=back

=cut

sub setGroupQuotas($$$%)
{
	my ($filerspec, $vol, $qtree, $nusers, %opts)
			= validate_mixed(@_,
				[
				{ type => SCALAR|HASHREF },
				{ type => SCALAR },
				{ type => BOOLEAN },
				  # qtree is defined as BOOLEAN so that it can 
				  # be undef. If undef, we create a quota rule
				  # with target=$group, type='group@/vol/$vol'
				{ type => BOOLEAN },
				  # type is BOOLEAN so it can be undef
				  # If undef, target = '*' 
				],
				{ 
				disk 	=> { type => SCALAR, optional => 1 },
				files 	=> { type => SCALAR, optional => 1 },
				thold 	=> { type => SCALAR, optional => 1 },
				sdisk 	=> { type => SCALAR, optional => 1 },
				sfiles	=> { type => SCALAR, optional => 1 },
				},
			);
	return eval {
		my ($type, $i);

		my $qtreepath = "/vol/$vol";

		if ($qtree) {
			# create qtree only if we are supposed to
			logTrc("create qtree and extract qtreepath");
			$qtreepath = _create_qtree($filerspec, $vol, $qtree)
				or croak $@;
		}
		$type = "group\@$qtreepath";

		my $filerconn = make_conn($filerspec);

		if ($nusers) {
			logTrc("create group");
			my %groups_hash	= createGroups($filerconn, 1)
				or croak("createGroups failed: .$@");
			my @groups = keys %groups_hash;
			my $group = $groups[0];
			my %users_hash 	= createUsers($filerconn, $nusers, group=>$group)
				or croak("createUsers failed: .$@");
			my @users = keys %users_hash;

			logTrc("set Groups quota rule on $filerspec");
			setQuota($filerspec, target => $groups[0], 
				type => "group\@$qtreepath", %opts)
						or croak $@;
	
			return (\%users_hash, \%groups_hash);

		} else {
			setQuota($filerspec, target => "*", type => $type, 
				 %opts)
					or croak $@;
			return \{target => "*", type => $type};
		}
	};
}


=head2 setUserQuotas($filerspec, $vol, $qtree, $nusers, %opts)

	Sets User Type Quotas 

=over 4

=item filerspec 

	Required Filer Specification

=item vol

	Required Volume name

=item qtree

	Required qtree name to create and set rule for

=item nusers

	Required number of users to set quota rules for

=item options   

	Optional
		disk   => ..	quota size limit
		files  => ..	file limit
		thold  => ..	threshold
		sdisk  => ..	soft disk limit
		sfiles => ..	soft files limit

		users_per_quota => ..   ## for single quota line with multiple users

=back

=cut

sub setUserQuotas($$$$%)
{
	my ($filerspec, $vol, $qtree, $nusers, %opts)
			= validate_mixed(@_,
				[
				{ type => SCALAR|HASHREF },
				{ type => SCALAR },
				{ type => BOOLEAN },
				  # qtree is defined as BOOLEAN so that it can 
				  # be undef. If undef, we create a quota rule
				  # with target = $user, type ='user@/vol/$vol'
				{ type => BOOLEAN },
				  # type is BOOLEAN so it can be undef
				  # If undef, target = '*' 
				],
				{ 
				disk 	=> { type => SCALAR, optional => 1 },
				files 	=> { type => SCALAR, optional => 1 },
				thold 	=> { type => SCALAR, optional => 1 },
				sdisk 	=> { type => SCALAR, optional => 1 },
				sfiles	=> { type => SCALAR, optional => 1 },
				users_per_quota => { type => SCALAR, optional => 1, default => 1 },
				},
			);
	return eval {
		my ($type, $i);

		my $qtreepath = "/vol/$vol";
	
		if ($qtree) {
			# create qtree only if we are supposed to
			logTrc("create qtree and extract qtreepath");
			$qtreepath = _create_qtree($filerspec, $vol, $qtree)
				or croak $@;
		}
		$type = "user\@$qtreepath";
		$opts{type} = $type;
		my $filerconn = make_conn($filerspec);

		if ($nusers) {
			logTrc("create a group");
			my %groups_hash = createGroups($filerconn, 1)
				or croak("createGroups failed: .$@");
			my @groups = keys %groups_hash;
			my $group = $groups[0];

			logTrc("create $nusers users");
			my %users_hash = createUsers($filerconn, $nusers, group=>$group)
				or croak("createUsers failed: .$@");
			my @users = keys %users_hash;
			setQuotasForUsers($filerconn, \%users_hash, %opts);
			logTrc("set user quota rule on $filerspec @users");

			return (\%users_hash , \%groups_hash);

		} else {
			delete($opts{users_per_quota});
			setQuota($filerspec, target => "*", type => $type, 
				 %opts)
					or croak $@;

			return \{target => "*", type => $type};
		}
	};
}


=head2 VerifyQuotaDiskLimit($uid, $gid, $quota, $file)

	Verifies quota disk limits by creating file of size upto disk limit
	(ported from waft/quotas/quota_sanity.thpl)

	Returns:

		1 	if attempt to exceed quota fails

		0 	otherwise

=over 4

=item uid 

	Required	uid of the user

=item gid 

	Required	gid of the user

=item quota

	Required	quota disk limit

=item file

	Required	file to create as user 

=back 

=cut

sub verifyQuotaDiskLimit($$$$)
{
    	my ($uid, $gid, $quota, $test_file) = @_;
	use IO::Handle;
        my ($saved_euid) = $>; $> = 0;
        my ($saved_egid) = $);
	my ($quota_1k) = int($quota/1024);
        if ($> != 0) {
            die("changing to specified user/group is not "
                . "possible as current euid != 0, you need to be root!");
        }

        logTrc("change to specified user:group $uid:$gid");
        $) = $gid; $> = $uid;

        if (!open(FH, ">$test_file")) {
                $> = $saved_euid; $) = $saved_egid;
                die("could not create $test_file as user $>");
        }
        my ($low_pct_lim) = 98;
	my ($high_pct_lim) = 102;
        my ($filler_block) = "a" x 1024;
        my ($error) = 0;

        my ($num_1k_low) = int($low_pct_lim * $quota_1k/ 1024);
        my ($num_1k_written);
        my ($num_1k_expected) = $num_1k_low - QUOTA_REPORT_DELTA;
        my ($fragment) = $low_pct_lim % 1024;

        logTrc("verify that we can almost reach the specified quota ".
	       "of the user:group $uid:$gid");
	logTrc("Attempting to write $num_1k_low 1k blocks to quota limit of $quota_1k kB ....");
        for($num_1k_written = 0;
	   ($num_1k_written < $num_1k_low) && !$error;
	   $num_1k_written++)
	{
                print FH $filler_block or $error = 1;
        }
	logTrc(".... wrote $num_1k_written 1k blocks " .
		"-- expected at least $num_1k_expected of quota $quota_1k kB");
	if (!$error) {
		logTrc("flushing filehandle");
	        FH->flush() or $error = 1;
	}
	if ($error) {
		$> = $saved_euid; $) = $saved_egid;
		if ($num_1k_written == 0) {
			die("Could not write any 1k blocks");
		} elsif ($num_1k_written < $num_1k_expected) {
			die("Expected to be able to write at least $num_1k_expected, " .
			     "but was only able to write $num_1k_written for quota: $quota");
		} else {
			return(1);
		}
	}

        if ($error == 0 && $fragment > 0) {
	    logTrc("writing $fragment byte fragment");
            print FH "b" x $fragment or $error = 1;
	    logTrc("flushing filehandle");
	    if (!$error) {
            	FH->flush() or $error = 1;
	    }
        }
	if ($error) {
            $> = $saved_euid; $) = $saved_egid;
            logTrc("could not write $fragment byte fragment to file $test_file");
	    return(1);
        }

        logTrc("verify that we cannot go beyond the specified limit ".
	       "of the user $uid");
	my ($pct_over) = $high_pct_lim - 100;
        my ($num_1k_over) = int($quota_1k * ($pct_over/100));
	my ($num_1k_high) = $quota_1k + $num_1k_over;

	logTrc("Attempting to write $num_1k_over 1k blocks over quota $quota_1k kB");
        for($num_1k_written = 0;
	    ($num_1k_written < $num_1k_over) && !$error;
	    $num_1k_written++)
	{
                print FH $filler_block or $error = 1;
                FH->flush() or $error = 1;
                close FH;
        }
	logTrc("Wrote $num_1k_written (of $num_1k_over) 1k blocks over quota $quota_1k kB");
	$> = $saved_euid; $) = $saved_egid;
	if ($error) {
		logTrc("PASS: Attempt to exceed quota failed");
		return(1);
	} else {
		logTrc("FAIL: Attempt to exceed quota succeeded");
		logresult("FAIL", "Attempt to exceed quota succeeded for user [$> : $>]   " .
			"quota: $quota_1k kB   written past quota: $num_1k_written");
	}
}


=head2 convertQuotaValue($value)

	Convert quota limit values to be filer consistent
		- Quota Report always returns kbyte limits
		- Filer always rounds off values to upper 4k blocks 

=over 4

=item value

	Required	value in kbytes to convert

=item options   

	Optional
	    Output value conversion:
		output_units =>
			UNIT_B    ..   Output value will be in bytes
			UNIT_KB   ..   Output value will be in Kbytes
			UNIT_MB   ..   Output value will be in Mbytes
			UNIT_GB   ..   Output value will be in Gbytes
			UNIT_TB   ..   Output value will be in Tbytes

=back 

    Data blocks are 4Kbytes.

=cut

sub convertQuotaValue($%)
{
	my ($value_in, %opt) = validate_mixed(@_,
			[{ type => SCALAR }],
			{
				input_units    => { type => SCALAR, optional => 1 },
				output_units   => { type => SCALAR, optional => 1 },
			});
	# By default, convert expected quota limit values to k-bytes
	# since quota report always returns k-bytes limits
	my $input_units = $opt{input_units} || UNIT_KB;
	my $output_units = $opt{output_units} || UNIT_KB;
 	logDbg("convertQuotaValue  Line: ".__LINE__."\nValue in: \"$value_in\"\n".
		"Input units:   " . ($opt{input_units}  ? $opt{input_units}  : "No 'input_units specified, assuming 'Kbyte' (k)")."\n".
		"Output units:  " . ($opt{output_units} ? $opt{output_units} : "No 'output_units specified, assuming 'Kbyte' (k)"));

	$value_in = lc($value_in);
	if ($value_in =~ /^\d+$/) {
		if ($input_units eq UNIT_KB) {
			$value_in .= "k";
		} elsif ($input_units eq UNIT_MB) {
			$value_in .= "m";
		} elsif ($input_units eq UNIT_GB) {
			$value_in .= "g";
		} elsif ($input_units eq UNIT_TB) {
			$value_in .= "t";
		}
	}
	my $bytes_in = convert_to_bytes($value_in);
	my $fourkb_in = int($bytes_in / 4096);
	my $rem = ($bytes_in % 4096);
	$fourkb_in++ if ($rem);
	my $kb_out = 4 * $fourkb_in;
	my $value_out = convert_units("$kb_out", UNIT_KB, $output_units);
	logDbg("convertQuotaValue  Line: ".__LINE__.
		"\nConverted quota 'disk' value: \"$value_in\",  with units: \"$input_units\"\n".
		"to value: \"$value_out\",  units: \"$output_units\"");
	return( $value_out );
}



=head2 adjustedConvertQuotaValue($value)

	Convert quota limit values for a single file, or for a group of files
	all of the same size to be consistent with
	values given by quota report. This function computes the change in
	the 'kbytes_used' field of quota report when a single file of a
	given size is created or destroyed.  The approximation should
	be within 1% of the actual quota report value. This approximation is
	closer than that given by convertQuotaValue.

	Besides returning the expected value, a "margin" is also returned.
	The difference between the approximate value and the expected value
	should be less than the "margin".

	The expected scenario for using this function is:

		# Get initial (actual) quota report value
		my $qr_kb_used_1 = ....
		# create a file of size $filesize 
		# Get second quota report value which will reflect new file
		my $qr_kb_used_2 = ....
		my $actual_diff = abs($qr_kb_used_2 - $qr_kb_used_1);
		my ($expected_diff, $margin) = convertQuotaValue($filesize);
		if ( abs($actual_diff - $expected_diff) > $margin ) {
			# Return FAIL or FATAL error
		}

	... or, for multiple files ($nfiles) of the same size:
		# Get initial (actual) quota report value
		my $qr_kb_used_1 = ....
		# create a file of size $filesize 
		# Get second quota report value which will reflect new file
		my $qr_kb_used_2 = ....
		my $actual_diff = abs($qr_kb_used_2 - $qr_kb_used_1);
		my ($expected_diff, $margin) = convert$nfiles * QuotaValue($filesize);
		if ( abs($actual_diff - $expected_diff) > $margin ) {
			# Return FAIL or FATAL error
		}

		Note that quota calculations also include directories both for
		'nfiles', and for 'kbytes_used' values.

	Returns:

		converted quota value
		"maximum difference" between expected and actual quota report value

=over 4

=item value

	Required	value in kbytes to convert

=item options   

	Optional
	    Output value conversion:
		output_units =>
			UNIT_B    ..   Output value will be in bytes
			UNIT_KB   ..   Output value will be in Kbytes
			UNIT_MB   ..   Output value will be in Mbytes
			UNIT_GB   ..   Output value will be in Gbytes
			UNIT_TB   ..   Output value will be in Tbytes


	    blksize_kb     Optional.   filesystem block size in Kb.
			      Default: 8.

=back 

    Data blocks are 4Kbytes.

=cut

sub adjustedConvertQuotaValue($%)
{
	my ($value_in, %opt) = validate_mixed(@_,
			[{ type => SCALAR }],
			{
				input_units    => { type => SCALAR, optional => 1 },
				output_units   => { type => SCALAR, optional => 1 },
				blksize_kb     => { type => SCALAR, optional => 1 },
				debug          => { type => BOOLEAN, optional => 1, default => 0 },
			});

	my $blksize_kb = $opt{blksize_kb} || 4;
	# By default, convert expected quota limit values to k-bytes
	# since quota report always returns k-bytes limits
	my $input_units = $opt{input_units} || UNIT_KB;
	my $output_units = $opt{output_units} || UNIT_KB;
	$value_in = lc($value_in);
        if ($opt{debug}) {
                logDbg("adjustedConvertQuotaValue  Line: ".__LINE__."\nValue_in:\t\"$value_in\"\n\nOpt:\n\n".Dumper(\%opt));
	}
	if ($value_in =~ /^\d+$/) {
		if ($input_units eq UNIT_KB) {
			$value_in .= "k";
		} elsif ($input_units eq UNIT_MB) {
			$value_in .= "m";
		} elsif ($input_units eq UNIT_GB) {
			$value_in .= "g";
		} elsif ($input_units eq UNIT_TB) {
			$value_in .= "t";
		}
	}
	my $bytes_in = convert_to_bytes($value_in);
        if ($opt{debug}) {
                logDbg("adjustedConvertQuotaValue  Line: ".__LINE__."\nValue_in:\t\"$value_in\"\nInput_units: \"$input_units\"\nBytes_in: $bytes_in");
	}

	# Files less than 64 bytes are stored in inode
	if ($bytes_in < 64) {
	        if ($opt{debug}) {
	                logDbg("adjustedConvertQuotaValue  Line: ".__LINE__."\nReturning: \"(0, 0)\"");
		}
		return(0, 0);
	}

	my $blks_in = int($bytes_in / ($blksize_kb * 1024));
	my $blks_out = $blks_in;
	my $rem = ($bytes_in % ($blksize_kb * 1024));
	$blks_out++ if ($rem);
	my $bytes_out = $blksize_kb * $blks_out * 1024;
	my $kb_out = convert_units($bytes_out, UNIT_B, UNIT_KB);
        if ($opt{debug}) {
                logDbg("adjustedConvertQuotaValue  Line: ".__LINE__."\nBlksize_kb:\t\"$blksize_kb\"\nBlks_in:\t\"$blks_in\"\n".
		         "bytes_in: $bytes_in\nBlks_out: $blks_out\nBytes_out: $bytes_out\nKb_out: $kb_out");
	}

	my $max_diff_kb = 0;
	if ($kb_out > (32)) {
		if ($blksize_kb == 4) {
			$max_diff_kb += 4;
		} elsif ($blksize_kb == 8) {
			$max_diff_kb += 8;
		}
	        if ($opt{debug}) {
	                logDbg("adjustedConvertQuotaValue  Line: ".__LINE__."\nMax_diff_kb:   \"$max_diff_kb\"");
		}
	}
	if ($kb_out > (2040)) {
		if ($blksize_kb == 4) {
			$max_diff_kb += 8;
		} elsif ($blksize_kb == 8) {
			$max_diff_kb += 16;
		}
	        if ($opt{debug}) {
	                logDbg("adjustedConvertQuotaValue  Line: ".__LINE__."\nKb_out:   \"$max_diff_kb\"");
		}
	}
	if ($kb_out > (6132)) {
		if ($blksize_kb == 4) {
			$max_diff_kb += 4;
		} elsif ($blksize_kb == 8) {
			$max_diff_kb += 8;
		}
	        if ($opt{debug}) {
	                logDbg("adjustedConvertQuotaValue  Line: ".__LINE__."\nKb_out:   \"$max_diff_kb\"");
		}
	}
	if ($kb_out > (8160)) {
		if ($blksize_kb == 4) {
			$max_diff_kb += 4;
		} elsif ($blksize_kb == 8) {
			$max_diff_kb += 8;
		}
	        if ($opt{debug}) {
	                logDbg("adjustedConvertQuotaValue  Line: ".__LINE__."\nKb_out:   \"$max_diff_kb\"");
		}
	}
	if ($kb_out > (10220)) {
		if ($blksize_kb == 4) {
			$max_diff_kb += 4;
		} elsif ($blksize_kb == 8) {
			$max_diff_kb += 8;
		}
	        if ($opt{debug}) {
	                logDbg("adjustedConvertQuotaValue  Line: ".__LINE__."\nKb_out:   \"$max_diff_kb\"");
		}
	}
	my $adj_kb_out = $kb_out + $max_diff_kb;
	if ($kb_out > 0) {
		 $max_diff_kb += int($kb_out / 100);
	}
	my $max_diff = convert_units($max_diff_kb, UNIT_KB, $output_units);
	my $value_out = convert_units($adj_kb_out, UNIT_KB, $output_units);
	logTrc("adjustedConvertQuotaValue: converted \"$value_in\" to \"$value_out\" be quota consistent");
        if ($opt{debug}) {
                logDbg("adjustedConvertQuotaValue  Line: ".__LINE__.
		            "\nKb_out:\t\"$kb_out\"\nMax_diff:\t\"$max_diff\"\n".
			    "Value_in:\t$value_in\"\nKb_out: \"$kb_out\"\n\n".
			    "Value_out: $value_out\nMax_diff: $max_diff");
	}

	if (wantarray) {
		return( $value_out, $max_diff );
	} else {
		return( $value_out );
	}
}


=head2 verifyQuotaReport($filerspec, type=>$type, volume=>$volume, 
			kbytes_limit=>$value ... )

	Reads Quota Report and matches expected values

=over 4

=item filerspec 

	Required Filer Specification

=item options   

	Optional	values to match with in quota report
		type   => ..	type of rule (tree|user|group)
		volume => ..	volume name to match
		id     => ..	id for the quota rule
		disk   => ..	quota size limit
		files  => ..	file limit
		thold  => ..	threshold
		sdisk  => ..	soft disk limit
		sfiles => ..	soft files limit

	Other optional args:
		flag	=>	Flag to pass to 'quota report' command, if any

		See other options under 'queryQuotaReport'

=back

=cut

sub verifyQuotaReport($%)
{
    my ($filerspec, %opt) = @_;
    my @matched = queryQuotaReport($filerspec, %opt);
    croak("can't match for given pattern! ".
		"Expected values =\n\n". Dumper(\%opt).
		"\n\nGot values =\n\n". Dumper(\@matched))
    	unless (@matched);
    return($matched[0]);
}


=head2 queryQuotaReport($filerspec, type=>$type, volume=>$volume, 
			kbytes_limit=>$value ... )

	Reads Quota Report and matches expected values

=over 4

=item filerspec 

	Required Filer Specification

=item options   

	Optional	values to match with in quota report -- 'query' options:
		type   => ..	type of rule (tree|user|group)
		volume => ..	volume name to match
		id     => ..	id for the quota rule
		disk   => ..	quota size limit
		files  => ..	file limit
		thold  => ..	threshold
		sdisk  => ..	soft disk limit
		sfiles => ..	soft files limit

	Other optional args -- 'non-query' options:
		flag	                   => ..   Flag to pass to 'quota report' command, if any

		match_debug                => ..   Default is zero. If 1, then show additional
					                debug messages for the matching process.
							Default is '0'.

		split_multiple_entries     => ..   If '1', then match 'target' and 'id' fields of queries
						        against any of a comma-separated list of quota entry values.
						        (Used for 'multiple-user entry' quotas)
						        Default is '1'.

		match_qtree_name           => ..   If '1', then match qtree name in query to qtree path in
						        quota entry.
							Default is '1'.

		match_disk_limit_margin  => ..   If defined, then match 'disk', 'thold', and 'sdisk' limit attributes if
						          'entry' and 'query' values are within the specified margin
						          where the margin is given in Kbytes -- e.g. "4" ==> "4 Kb"
							  Default is 4 (4 Kbyte block size).

		strip_cifs_user_prefix     => ..   If '1', then for 'user' entry attrs of 'target' and 'id', strip
							the 'server-name' prefix from CIFS user name before matching.
							For example:

							           WQA-F11\test_user_1   ==>  test_user_1

							Default is '1'.

=back

=cut

sub queryQuotaReport($%)
{
	my ($filerspec, %opt) = validate_mixed(@_,
			[{ type => SCALAR|HASHREF }],
			{
				#   'Non-query' options:
				flag	                  => { type => SCALAR,    optional => 1},
				match_debug               => { type => BOOLEAN,   optional => 1,   default => 0 },
				split_multiple_entries    => { type => BOOLEAN,   optional => 1,   default => 1 },
				match_qtree_name          => { type => BOOLEAN,   optional => 1,   default => 1 },
				strip_cifs_user_prefix    => { type => BOOLEAN,   optional => 1,   default => 1 },
				adjust_quota_values       => { type => BOOLEAN,   optional => 1,   default => 1 },
				blksize_kb                => { type => SCALAR,    optional => 1,   default => 0},
				match_disk_limit_margin   => { type => SCALAR,    optional => 1},

				#  Quota 'query' options
				type 			=> { type => SCALAR },
				volume 			=> { type => SCALAR },
				id			=> { type => SCALAR, optional => 1},
				target  		=> { type => SCALAR, optional => 1},
				tree			=> { type => SCALAR, optional => 1},
				quotaspec		=> { type => SCALAR, optional => 1},
				files_used		=> { type => SCALAR, optional => 1},
				files_limit		=> { type => SCALAR, optional => 1},
				kbytes_used 		=> { type => SCALAR, optional => 1},
				kbytes_limit		=> { type => SCALAR, optional => 1},
				kbytes_soft_limit	=> { type => SCALAR, optional => 1},
				disk_used       	=> { type => SCALAR, optional => 1},
				disk_limit      	=> { type => SCALAR, optional => 1},
				disk_soft_limit      	=> { type => SCALAR, optional => 1},
			}
		);

   my $match_debug = $opt{match_debug};

$match_debug = 1; ################################

   my $filerconn = eval { make_conn($filerspec) };
   my $volume = $opt{volume};
   my $blksize_kb = $opt{blksize_kb};
   unless ($blksize_kb) {
        my $aggr = vol_container($filerconn, $volume);
	my $aggr_status = aggr_status($filerconn, $aggr);
	if ($match_debug) {
	    logDbg("Aggr Status:\n\n".Dumper($aggr_status));
	    my $status_str = $aggr_status->{status};
	    if ($status_str =~ /64/) {
		$blksize_kb = 8;
            } else {
                $blksize_kb = 4;
            }
        }
   }
   logDbg("Blksize_kb: \"$blksize_kb\"");

   my $match_disk_limit_margin;
   if (defined($opt{match_disk_limit_margin})) {
	$match_disk_limit_margin = $opt{match_disk_limit_margin};
   } else {
	$match_disk_limit_margin = $opt{blksize_kb} || 8;
   }
   $match_disk_limit_margin ||= 0;

   logDbg("Enter queryQuotaReport:\n\n".Dumper(\%opt));

   my @misc_query_options = qw(		type
					volume
					id
					target
					tree
					quotaspec
				);
  my @file_query_options = qw(
					files_used
					files_limit
				  );
   my @disk_query_options = qw(		thold
					kbytes_used
					kbytes_limit
					kbytes_soft_limit
					disk_used
					disk_limit
					disk_soft_limit
				  );

   my @query_options = ( @misc_query_options, @file_query_options, @disk_query_options );

   my @non_query_options = ();
   foreach my $k ( keys(%opt) ) {
	unless (grep( {$_ eq $k} @query_options )) {
		push(@non_query_options, $k);
	}
   }
   logDbg("queryQuotaReport  Line: ".__LINE__."\n\nNon-Query Options:\n\n".Dumper(\@non_query_options));

   $opt{id} = $opt{target}
	if (!defined($opt{id}) && defined($opt{target}));
   $opt{kbytes_used} = $opt{disk_used}
	if (!defined($opt{kbytes_used}) && defined($opt{disk_used}));
   $opt{kbytes_limit} = $opt{disk_limit}
	if (!defined($opt{kbytes_limit}) && defined($opt{disk_limit}));

    for my $key ( @file_query_options ) {
	next if (!defined($opt{$key}) || ($opt{$key} eq ""));
	my $old_val = $opt{$key};
	$opt{$key} = convert_to_bytes($old_val);
	if ($match_debug) {
		logDbg("queryQuotaReport  Line: ".__LINE__.
		         "\nConverting:  \"$key\" = \"$old_val\"  to  \"$opt{$key}\"");
	}
   }

    for my $key ( @disk_query_options ) {
	next if (!defined($opt{$key}) || ($opt{$key} eq ""));
	my $old_val = $opt{$key};
	next if (!defined($old_val) || ($old_val eq "") || ($old_val eq "-"));
	next if ($old_val =~ /^\d+$/);
	$opt{$key} = convert_units(strip_unit_spec($old_val), guess_unit($old_val), UNIT_KB);
	if ($match_debug) {
		logDbg("queryQuotaReport  Line: ".__LINE__.
		         "\nConverting:  \"$key\" = \"$old_val\"  to  \"$opt{$key}\"");
	}
    }

    return eval {

	my $report_flag = ($opt{flag} || "");
	my $query = { %opt };
	my %gqr_opts = ();
	$gqr_opts{flag} = $report_flag
	    if (defined($report_flag));

	# Delete 'non-query' options from $query
	foreach ( @non_query_options ) {
		delete($query->{$_});
	}
	for my $attr (keys(%$query)) {
		delete($query->{$_}) unless (defined($query->{$_}));
	}
	if (defined($query->{tree}) && ($query->{tree} eq '' || $query->{tree} eq '-')) {
		$query->{tree} = '-';
	}

	logDbg("queryQuotaReport  Line: ".__LINE__."\n  -- Expecting match:\n" .  Dumper($query));

	my @report = getQuotaReport($filerconn, %gqr_opts);
	my $entry = {};
	my @matched = ();
	foreach $entry (@report) {
		my $match;
		if ($match_debug) {
			logDbg("queryQuotaReport  Line: ".__LINE__."\ncomparing Entry\n\n"
				. Dumper($entry) . "\n\nwith query\n\n".  Dumper($query));
		}
		for my $attr (keys(%$query)) {
			if ($match_debug) {
				logDbg("queryQuotaReport  Line: ".__LINE__."\nNext Attr: \"$attr\"");
			}
			if (!defined($entry->{$attr})) {
				$match = 0;
				last;
			}
			my $ent_val = $entry->{$attr};
			my @qry_alt_values = split(m%\|%, $query->{$attr});
			$match = 0;
			if ($match_debug) {
				logDbg("queryQuotaReport  Line: ".__LINE__.
				         "\nAttr: \"$attr\"\nOrig query value: \"$query->{$attr}\n".
					"Query Alternative Values: join("|", @qry_alt_values");
			}
			foreach my $qry_val (@qry_alt_values) {
				if ($match_debug) {
					logDbg("queryQuotaReport  Line: ".__LINE__.
					         "\nMatching \"$attr\" attribute\n".
					         "Query: \"$qry_val\"\nEntry: \"$ent_val\"");
				}
				if ( ($attr eq 'tree') &&
				            (($qry_val eq '') || ($qry_val eq '-')) &&
				           (($ent_val eq '') || ($ent_val eq '-')) )
			        {
					$match = 1;
					if ($match_debug) {
						logDbg("queryQuotaReport  Line: ".__LINE__.
						         "\nMatching 'tree' \"$attr\" attribute\n".
						         "Query: \"$qry_val\"\nEntry: \"$ent_val\"\n".
							 "Match: \"$match\"");
					}
				} elsif ( grep( {$attr eq $_}  qw(target id quotaspec) ) ) {
					if (($opt{split_multiple_entries} && ($entry->{type} eq 'user')) && 
					  ($ent_val =~ /,/) )
					{
						my @entry_list = split(/\s*,\s*/, $ent_val);
						$qry_val =~ m%\\?([^\\]+)$%;
						my $query_base_name = $1;
						my @stripped_entry_list =  map( { $_ =~ m%\\?([^\\]+)$%; $1 }  @entry_list );
						my $entry_base_name = $1;
						$match = grep( {$_ eq $qry_val}  @entry_list ) ||
						           grep( { lc($_) eq lc($qry_val) }  @entry_list ) ||
						           grep( { $_ eq $query_base_name }  @stripped_entry_list ) ||
						           grep( { lc($_) eq lc($query_base_name) }  @stripped_entry_list );
						$match = ($match) ? 1 : 0;
						if ($match_debug) {
							logDbg("queryQuotaReport  Line: ".__LINE__.
							        "\nMatching 'multiple-entry' attribute \"$attr\"\n".
								"Entry: \"$ent_val\"\nEntry list: \"@entry_list\"\n".
								"Stripped entry list: \"@stripped_entry_list\"\n".
								"Query: \"$qry_val\"  Stripped query: \"$query_base_name\"\n".
								"Match: \"$match\"");
						}
					} elsif ( $opt{match_qtree_name} && ($entry->{type} eq 'tree') ) {
						$ent_val =~ m%\/?([^\/]+)$%;
						my $entry_qt_name = $1;
						$qry_val =~ m%\/?([^\/]+)$%;
						my $query_qt_name = $1;
						$match = ($ent_val eq $qry_val) || ($entry_qt_name eq $query_qt_name);
						$match = ($match) ? 1 : 0;
						if ($match_debug) {
							logDbg("queryQuotaReport  Line: ".__LINE__.
							        "\n'qtree' attribute \"$attr\"\n".
								"Entry: \"$ent_val\"  --  \"$entry_qt_name\"\n".
								"Query: \"$qry_val\"  --  \"$query_qt_name\"\n".
								"Match: \"$match\"");
						}
					} elsif ( $opt{strip_cifs_user_prefix} && ($entry->{type} eq 'user') ) {
						$ent_val =~ m%\\?([^\\]+)$%;
						my $entry_base_name = $1;
						$qry_val =~ m%\\?([^\\]+)$%;
						my $query_base_name = $1;
						$match = ($ent_val eq $qry_val) || ($entry_base_name eq $query_base_name);
						$match = ($match) ? 1 : 0;
						if ($match_debug) {
							logDbg("queryQuotaReport  Line: ".__LINE__.
							        "\n'user' attribute \"$attr\"\n".
								"Entry: \"$ent_val\"  --  \"$entry_base_name\"\n".
								"Query: \"$qry_val\"  --  \"$query_base_name\"\n".
								"Match: \"$match\"");
						}
					} else {
						$match = ($ent_val eq $qry_val);
						if ($match_debug) {
							logDbg("queryQuotaReport  Line: ".__LINE__.
							        "\nAttribute \"$attr\"\n".
							        "Entry Type: \"$entry->{type}\"\n".
								"Entry: \"$ent_val\"\n".
								"Query: \"$qry_val\"\n".
								"Match: \"$match\"");
						}
					}
				} elsif (($attr =~ /^(disk_limit|disk_soft_limit|kbytes_limit|kbytes_soft_limit|thold)$/) &&
					  ($qry_val =~ /^\d+$/) && ($ent_val =~ /^\d+$/))
				{
					my $margin = $match_disk_limit_margin;
					$match = (abs($qry_val - $ent_val) < $margin);
					$match = ($match) ? 1 : 0;
					if ($match_debug) {
						logDbg("queryQuotaReport  Line: ".__LINE__.
						        "\n'disk' attribute \"$attr\"\n".
							"Entry: \"$ent_val\"\n".
							"Query: \"$qry_val\"\n".
							"Margin:\"$margin\"\n".
							"Match: \"$match\"");
					}
				} elsif (($attr eq 'kbytes_used') &&
					($qry_val =~ /^\d+$/) && ($ent_val =~ /^\d+$/))
				{
					my $adj_ent_val = $ent_val;
					my $margin = $match_disk_limit_margin;
					my $adj_margin;
					$match = 0;
					if ($opt{adjust_quota_values}) {
						($adj_ent_val, $adj_margin) =
							adjustedConvertQuotaValue($ent_val,
								                  blksize_kb => $blksize_kb,
										  input_units=>UNIT_KB,
										  output_units=>UNIT_KB,
                                                                                  blksize_kb=>$blksize_kb);
						$margin = $adj_margin if ($adj_margin > $margin);
						$match  ||=  (($qry_val >= $ent_val)  &&  ($qry_val <= $adj_ent_val));
						$match  ||=  (abs($qry_val - $ent_val)      <= $margin);
						$match  ||=  (abs($qry_val - $adj_ent_val)  <= $margin);
						$match = ($match) ? 1 : 0;

						if ($match_debug) {
							logDbg("queryQuotaReport  Line: ".__LINE__.
							        "\nAttribute:       'kbytes_used'\n".
								"adjust_quota_values: \"True\"\n".
								"Entry:             \"$ent_val\"\n".
								"Adjusted Entry:    \"$adj_ent_val\"\n".
								"Query:             \"$qry_val\"\n".
								"Margin:            \"$margin\"\n".
								"Match: \"$match\"");
						}
					} else {
						$margin = $match_disk_limit_margin;
						$match = (abs($qry_val - $ent_val) <= $margin);
						$match = ($match) ? 1 : 0;
						if ($match_debug) {
							logDbg("queryQuotaReport  Line: ".__LINE__.
								"match_disk_limit_margin: \"$opt{match_disk_limit_margin}\"\n".
							        "\nAttribute:       'kbytes_used'\n".
								"Entry:             \"$ent_val\"\n".
								"Query:             \"$qry_val\"\n".
								"Margin:            \"$margin\"\n".
								"Match: \"$match\"");
						}
					}
				} else {
					$match = ($ent_val eq $qry_val);
					if ($match_debug) {
						logDbg("queryQuotaReport  Line: ".__LINE__.
						        "\nDefault Match:\nAttribute \"$attr\"\n".
							"Entry: \"$ent_val\"\n".
							"Query: \"$qry_val\"\n".
							"Match: \"$match\"");
					}
				}
				$match = ($match) ? 1 : 0;
				if ($match_debug) {
					logDbg("queryQuotaReport  Line: ".__LINE__.
						"\nEntry: \"$ent_val\"\n".
						"Query: \"$qry_val\"\n".
						"Match: \"$match\"");
				}
				last if ($match);
			}
			if ($match_debug && !$match) {
				logDbg("queryQuotaReport  Line: ".__LINE__.
				        "\nMatch Failed ".
					"on attribute: \"$attr\"\n".
				        "Entry: \"$ent_val\"\n".
				        "vs. query: \"$query->{$attr}\"");
			}
			if (!$match) {
				last;
			}
		}
		if ($match) {
			if ($match_debug) {
				logDbg("queryQuotaReport  Line: ".__LINE__.
				          "\nMatch Succeeded for entry:\n\n".Dumper($entry));
			}
			push(@matched, $entry);
		}
	}
	logDbg("queryQuotaReport Returning:\n\n".Dumper(\@matched));
	return (@matched);
    };
}


=head2 getQuotaReport ($filerspec, %options)

	Reads Quota Report and returns an array of hash values of the report

=over 4

=item filerspec 

	Required	Filer Specification

=item timeout

	Optional

=item flag

	Optional	"flag" to pass to "quota report" command, if any

=back

=cut

sub getQuotaReport($%)
{
	my ($filerspec, %options) = validate_mixed(@_, 
			[
			{ type => SCALAR | HASHREF},
			],
			{
				flag => { type => SCALAR, optional => 1 },
				path => { type => SCALAR, optional => 1 },
                        	timeout  => { type => SCALAR, optional => 1,
                                                 default => $main::CMDTIMEOUT},
			}
			);

	logDbg("Enter getQuotaReport  Line: ".__LINE__."\n\nOptions:\n\n".Dumper(\%options));
	my $filerconn = make_conn($filerspec);
	my $timeout = defined($options{timeout}) ? $options{timeout} : 300;

	my @report;
  	my $cmd = "quota report";
	if ($options{flag}) {
		my $flag;
		$flag = $options{flag};
		$flag = "-" . $flag
			unless ($flag =~ /^-/);
		$cmd .= "  $flag"
	}
	if ($options{path}) {
		my $path = $options{path};
		$cmd .= " $path";
	}
	my @lines = split(/\n/, sayEOF($filerconn, $cmd, timeout=>$timeout));

	croak "We are trying to get a quota report while quotas are still ".
	      "initializing on some volumes.. quotas should be enabled ".
	      "before quota report can be fetched, please check your filer!"
		if ($lines[0] =~ /quotas are still initializing/);

	foreach my $line (@lines) {
		# skip over the report column names 
		next if $line =~ /^Type|K-Bytes|^-----/;
		push @report , _parse_quota_report_line($line, $options{flag});
	}

	logDbg("getQuotaReport   --  Returning:\n\n" . Dumper(@report));

	return wantarray ? @report : \@report;
}


=head2 quotaUserMapping ($filerspec, $value)

	Sets QUOTA_PERFORM_USER_MAPPING directive in /etc/quotas to 
	specified value - on or off

=over 4

=item filerspec 

	Required	Filer Specification

=item value 

	Required	'on' or 'off'

=back

=cut

sub quotaUserMapping($$%)
{
    my ($filerspec, $value, %options) = validate_mixed(@_,
                 [
                   { type => SCALAR|HASHREF },
                   { type => SCALAR }
                 ],
                  {
                           timeout  => { type => SCALAR, optional => 1,
                                               default => $main::CMDTIMEOUT},
		  }
	);

	my $filerconn = make_conn($filerspec);
	my $timeout = defined($options{timeout}) ? $options{timeout} : 300;
	my $quota_file = '/etc/quotas';

	return eval {
		if ($value !~ /on/i && $value !~ /off/i) {
			croak ("incorrect value passed to");
		}
		$value = uc($value);
		my $line = "QUOTA_PERFORM_USER_MAPPING $value\n";

		my $oldquotatext = rdfile($filerconn, $quota_file);
		my @oldquotalines;
		if (!$oldquotatext) {
			@oldquotalines = ();
		} else {
			@oldquotalines = split(/\n/, $oldquotatext);
		}
                if (@oldquotalines == 0) {
			my $commentline = sprintf "# %15s %15s %6s %6s %6s %6s %6s", 
				"target", "type", "disk", "files", "thold", 
				"sdisk", "sfiles";
                        push(@oldquotalines, $commentline);
                        push(@oldquotalines, $line);
		} else {
			my $lastline = $oldquotalines[$#oldquotalines];
			if ($lastline =~ /^\s*QUOTA_PERFORM_USER_MAPPING/) {
				$oldquotalines[$#oldquotalines] = $line;
			} else {
				push(@oldquotalines, $line);
			}
		}
		my $newquotatext = join("\n", @oldquotalines);
		logTrc("will write the following $quota_file: $newquotatext");
		#XXX temporary work around. wrfile() is broken for empty files
		if ((!defined($oldquotatext))) {
		      	say($filerconn,
			    "wrfile $quota_file $line", '-blind',
			    -timeout, $timeout);
		} else {
			# general error checking. not very smart.
			wrfile($filerconn, $quota_file, $newquotatext)
				or croak $@;
		}
		
		# general error checking. not very smart.
		my $text = rdfile ($filerconn, $quota_file)
					or croak($@);
		if ($text !~ "$line") {
			croak ("did not find \'$line\' in output: $text");
		}
		return 1;
	};
}


=head2 mapUsers ($filerspec, $cifsuser, $nfsuser)
	
	Sets the user mapping of a cifs user to a nfs user by adding
	an entry into the /etc/usermaps.cfg file

=over 4

=item filerspec 

	Required	Filer Specification

=item cifsuser 

	Required	username of cifs user

=item nfsuser 

	Required	username of nfs user

=back

=cut

sub mapUsers($$$%)
{
    my ($filerspec, $cifsuser, $nfsuser, %options) =
          validate_mixed(@_,
                 [
                   { type => SCALAR|HASHREF },
                   { type => SCALAR },
                   { type => SCALAR }
                 ],
                  {
                           timeout  => { type => SCALAR, optional => 1,
                                               default => $main::CMDTIMEOUT},
		  }
	     );

	# logDbg("mapUsers  BEFORE:\n\n".Dumper(readUsermapFile($filerspec)));
	addUsermap( $filerspec,
		     cifs_user => $cifsuser,
		     nfs_user => $nfsuser,
		     direction => "==");
	# logDbg("mapUsers  AFTER:\n\n".Dumper(readUsermapFile($filerspec)));

}


=head2 unmapUser ($filerspec, $cifsuser)
	
	Removes the user mapping of a cifs user to a nfs user by removing
	the entry into the /etc/usermaps.cfg file for CIFS user $cifsuser

=over 4

=item filerspec 

	Required	Filer Specification

=item cifsuser 

	Required	username of cifs user

=back

=cut

sub unmapUser($$%)
{
    my ($filerspec, $cifsuser, %options) =
          validate_mixed(@_,
                 [
                   { type => SCALAR|HASHREF },
                   { type => SCALAR }
                 ],
                  {
                           timeout  => { type => SCALAR, optional => 1,
                                               default => $main::CMDTIMEOUT},
		  }
	     );

	removeUsermap($filerspec, cifs_user => $cifsuser);
}



=head2 chownCIFSFile ($hostspec, $userid, $winpath)

	Set owner of file using CIFS

=over 4

=item hostspec 

	Required	Host Specification for CIFS client

=item userid

	Required	CIFS id of new owner

=item winpath

	Required	Windows path of file to have new owner

=back

=cut

sub chownCIFSFile {
    my ($hostspec, $userid, $winpath, %opts) =
          validate_mixed(@_,
                 [
                   { type => SCALAR|HASHREF },
                   { type => SCALAR },
                   { type => SCALAR },
                 ],
		 {
				recursive => { type => SCALAR, optional => 1, default => 0 },
		 }
           );
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $cifs_cmmd;
	if ($opts{recursive}) {
		$cifs_cmmd = "x:\\bin\\subinacl.exe /file $winpath /setowner\=$userid";
	} else {
		$cifs_cmmd = "x:\\bin\\subinacl.exe /subdirectories $winpath /setowner\=$userid";
	}
###	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	my $output = execute_cmd($cifs_hostname, $cifs_cmmd);
	chomp($output);
	$output =~ s/\s+/ /g;
	$output =~ s%[^\w\ \t=\\\/\-\\._]%%g;
	if ($output !~ /is the new owner/) {
		 croak("chownCIFSFile:  Could not change owner of\n".
			 "file \"$winpath\" to \"$userid\":\n\n$output");
	}
	return(1);
}

sub str_to_charray {
	my ($str) = (@_);
	my $rest = $str;
	my @carry = ();
	while ($rest) {
		$rest =~ s%^(.)(.*)%$2%;
		push(@carry, $1);
	}
	return(@carry);
}


=head2 rshCIFSClient ($hostspec, $cifs_cmmd)

	Issue command to CIFS client via rsh

=over 4

=item hostspec 

	Required	Host Specification for CIFS client

=item cifs_cmmd

	Required	RSH Command string

=back

=cut

sub rshCIFSClient {
    my ($hostspec, $cifs_cmmd) =
          validate_mixed(@_,
                 [
                   { type => SCALAR|HASHREF },
                   { type => SCALAR },
                 ],
		 {});
	my $cifsconn = make_conn($hostspec, "#rsh");
	my $cifs_hostname = $cifsconn->hostp()->hostname();
	my $output = `rsh $cifs_hostname \"$cifs_cmmd\"`;
	chomp($output);
	return($output);
}


=head2 mapToCIFSUser ($filerspec, $user_ref, %opts)

	Given an existing NFS user, creates a new CIFS user, and
	maps the CIFS user to the NFS user

=over 4

=item filerspec 

	Required	Filer Specification

=item user_ref

	Required	HASHREF. Represents user to be mapped,

=item timeout

	Optional

=item winlogin

	Optional	login name for new CIFS user. Default is to append
			'_win' to login of NFS user

=item winpasswd

	Optional	passwd to use for CIFS user.

=item maxusers

	Optional	If defined, then this value is used as an upper limit
			on the number of local CIFS users to create.

=item use_domain

	Optional	If '0' or undef, then no domain prefix is used for the CIFS user.
			If '1', then the CIFS server name is used as the domain.
			Otherwise, the value of use_domain is interpreted as the domain
			name to use for the CIFS user.

=item direction

	Optional	If specified, then one of: '==', '=>', '<='. If not specified, then
			'==' is the default.

=back

=cut

sub mapToCIFSUser($$%)
{
    my ($filerspec, $user_ref, %opts) =
          validate_mixed(@_,
                 [
                   { type => SCALAR|HASHREF },
                   { type => HASHREF },
                 ],
                  {
                           timeout  => { type => SCALAR, optional => 1,
                                               default => $main::CMDTIMEOUT},
			   winlogin => { type => SCALAR, optional => 1 },
			   winpasswd => { type => SCALAR, optional => 1,
						default => $main::CIFS_PASSWD || "cifspass123" },
			   maxusers => { type => SCALAR, optional => 1 },
			   use_domain => { type => SCALAR|UNDEF, optional => 1,
						default => 1 },
			   direction => { type => SCALAR|UNDEF, optional => 1,
						default => '==' },
		  }
	    );

	my $filerconn = make_conn($filerspec);
	my $dir = $opts{direction};
	my ($uid) =  $user_ref->{uid};
	my ($gid) =  $user_ref->{gid};
	my ($login) = $user_ref->{login};
	my ($winpasswd) = $opts{winpasswd};
	my ($winlogin) = $opts{winlogin} || $login . "_win";
	my $domain_prefix = "";
	my $domain = undef;
	if ($opts{use_domain}) {
		if ($opts{use_domain} eq "1") {
			my $server = getCIFSServer($filerconn);
			$domain = $server;
		} else {
			$domain = $opts{use_domain};
		}
		$domain_prefix = "$domain\\";
		$user_ref->{windomain} = $domain;
	}
	my $num_new_users = undef;
	if ($opts{maxusers}) {
		my $num_current = scalar(listCIFSUsers($filerconn));
		$num_new_users = $opts{maxusers} - $num_current;
		return($user_ref)
		    if ($num_new_users < 0);
	}
	$user_ref->{winlogin} = $winlogin;
	addPasswdEntry($filerconn, $winlogin, $uid, $gid, passwd=>$winpasswd);
        addCIFSUser($filerconn, $winlogin, $winpasswd, "Users")
	    || die("addCIFSUser failed: $@");

	my $qualified_user_name = $domain_prefix . $winlogin;
	mapUsers($filerconn, $qualified_user_name, $login);

	return($user_ref);
}


=head2 unmapToCIFSUser ($filerspec, $user_ref, %opts)
	
	Removes the user mapping of a cifs user to a nfs user by removing
	the entry into the /etc/usermaps.cfg file for CIFS user $cifsuser

=over 4

=item filerspec 

	Required	Filer Specification

=item user_ref

	Required	HASHREF. Represents user to be un-mapped,

=back

=cut

sub unmapToCIFSUser($$%)
{
    my ($filerspec, $user_ref, %opts) =
          validate_mixed(@_,
                 [
                   { type => SCALAR|HASHREF },
                   { type => HASHREF },
                 ],
                  {} );
	my $filerconn = make_conn($filerspec);
	my $cifs_user = $user_ref->{winlogin};
	unmapUser($filerconn, $cifs_user);
	die($@) if ($@);
	removeCIFSUser($filerconn, $cifs_user);
	die($@) if ($@);
	return(1);
}


=head2 mapToCIFSUserList ($filerspec, $users, %opts)

	Given a list of existing NFS users, create new 'CIFS' users
	based on the existing NFS users. For each existing user,
	adds a passwd entry for the new CIFS user, and maps the CIFS
	user to the NFS user.

=over 4

=item filerspec 

	Required	Filer Specification

=item users

	Required	HASHREF. Each hashtable key/value pair represents an existing user.
			hashtable has same format as that returned by readPasswdFile

=item timeout

	Optional

=item winpasswd

	Optional	passwd to use for CIFS user.

=item maxusers

	Optional	If defined, then this value is used as an upper limit
			on the number of local CIFS users to create.

=item use_domain

	Optional	If '0' or undef, then no domain prefix is used for the CIFS user.
			If '1', then the CIFS server name is used as the domain.
			Otherwise, the value of use_domain is interpreted as the domain
			name to use for the CIFS user.

=item direction

	Optional	If specified, then one of: '==', '=>', '<='. If not specified, then
			'==' is the default.

=back

=cut

sub mapToCIFSUserList($$%)
{
    my ($filerspec, $users, %opts) =
          validate_mixed(@_,
                 [
                   { type => SCALAR|HASHREF },
                   { type => HASHREF },
                 ],
                  {
                           timeout  => { type => SCALAR, optional => 1,
                                               default => $main::CMDTIMEOUT},
			   winpasswd => { type => SCALAR, optional => 1,
						default => $main::CIFS_PASSWD || "cifspass123" },
			   maxusers => { type => SCALAR, optional => 1 },
			   use_domain => { type => SCALAR|UNDEF, optional => 1,
						default => 1 },
			   direction => { type => SCALAR|UNDEF, optional => 1,
						default => '==' },
		  }
	   );

	my $filerconn = make_conn($filerspec);
	my $domain_prefix = "";
	my $domain = undef;
	if ($opts{use_domain}) {
		if ($opts{use_domain} eq "1") {
			my $server = getCIFSServer($filerconn);
			$domain = $server;
		} else {
			$domain = "$opts{use_domain}";
		}
		if ($domain) {
			$domain_prefix = "$domain\\";
		}
	}
	my $num_new_users = undef;
	if ($opts{maxusers}) {
		my $num_current = scalar(listCIFSUsers($filerconn));
		$num_new_users = $opts{maxusers} - $num_current;
		$num_new_users = 0
		    if ($num_new_users < 0);
	}
	my @newlines = ();
	my $count = 0;
	my $dir = $opts{direction};
	my @user_list;

	if (ref($users) =~ /ARRAY/) {
		foreach (@$users) {
			if (!ref($_)) {
				push(@user_list, $_);
			} elsif (ref($_) =~ /HASH/) {
				push(@user_list, $_->{login});
			}
		}
	} elsif (ref($users) =~ /HASH/) {
		while (my ($uname, $user_ref) = each(%$users)) {
			push(@user_list, $uname);
		}
		@user_list = sort(@user_list);
	}

	foreach my $user_ref (@user_list) {
		my $login = $user_ref->{login};
		if (defined($num_new_users && ($count++ > $num_new_users))) {
			last;
		}
		my ($uid) =  $user_ref->{uid};
		my ($gid) =  $user_ref->{gid};
		my ($winpasswd) = $opts{winpasswd};
		my ($winlogin) = $login . "_win";
		$user_ref->{windomain} = $domain
		    if ($domain);
		$user_ref->{winlogin} = $winlogin;
		### addPasswdEntry($filerconn, $winlogin, $uid, $gid, passwd=>$winpasswd);
	        addCIFSUser($filerconn, $winlogin, $winpasswd, "Users")
		    || die("addCIFSUser failed: $@");
		my $qualified_user_name = $domain_prefix . $winlogin;
		mapUsers($filerconn, $qualified_user_name, $login);
	}
	addPasswdList($filerconn, $users, cifs_user=>1);

	return(1);
}


#######################################################
#
# Deprecated -- for backwards compatibility only.....
# 

sub mapCIFSUser {
    my ($fspec, $uref, %opts) = @_;
    return( mapToCIFSUser($fspec,$uref,%opts) );
}
sub mapCIFSUserList {
    my ($fspec, $users, %opts) = @_;
   return( mapToCIFSUserList($fspec,$users,%opts) );
}
sub unmapCIFSUser {
    my ($fspec, $uref, %opts) = @_;
    return( unmapToCIFSUser($fspec,$uref,%opts) );
}
#
#######################################################




#-----------------------------------------------------------------------
# Internal routines ..
#-----------------------------------------------------------------------


# _setQuotaDefaults
#
#	Given a HASHREF which specifies a quota entry, supply
#	default values for undefined fields
#
sub _setQuotaDefaults($%)
{
	my ($quota, %quotaopt) = @_;

	my %default_values = (
		default => 0,
		### usermap => 0,
		disk => "-",
		files => "-",
		thold => "-",
		sdisk => "-",
		sfiles => "-",
	);
	while (my ($key, $val) = each(%quotaopt)) {
		if (!defined($quota->{$key})) {
			$quota->{$key} = $val;
		}
	}
	while (my ($key, $val) = each(%default_values)) {
		if (!defined($quota->{$key})) {
			$quota->{$key} = $val;
		}
	}
	return($quota);
}


# getQuotaTargetAndType
#
#	Given a quotaspec (HASHREF), determines complete 'target' and 'type'
#	for this quotaspec, along with serveral other fields. Returns two
#	HASHREFs which which represent the original quotaspec plus these
#	additional fields.
#
#	The first 'bare' HASHREF returned only has those fields defined which
#	are valid for passing to _mergeQuotaEntries:
#		target
#		type
#		disk
#		files
#		thold
#		sdisk
#		sfiles
#
#	The second HASHREF these fields, plus some additional ones:
#		basetype
#		target
#		target
#		volpath
#		volname
#		treepath
#		treename
#		username
#		groupname
#		default_quota
#
sub getQuotaTargetAndType($)
{
	my ($quota_in) = validate_mixed(@_,
			[
			    { type => HASHREF }
			],
			{}
		    );

	logDbg("Enter getQuotaTargetAndType\n\n".Dumper($quota_in));
	### my $quota_out = { %$quota_in };
	my $quota_copy = { %$quota_in };
	while ( my ($k,$v) = each(%$quota_copy) ) {
		if (!defined($v)) {
			delete($quota_copy->{$k});
		}
	}
	$quota_copy->{tree} = $quota_copy->{tree} || $quota_copy->{qtree};
	delete($quota_copy->{qtree});
	my $type = $quota_copy->{type};
	my $target = $quota_copy->{target} || $quota_copy->{id};
	my $default_quota;

	my $volopt = $quota_copy->{vol} || $quota_copy->{volume};
	my $treeopt = $quota_copy->{tree};
	my $useropt = $quota_copy->{user};
	my $groupopt = $quota_copy->{group};

	my ($groupname, $username) = (undef, undef);
    	my ($basetype, $resttype) = (undef, undef);
	my ($volpath, $treepath, $volname, $treename) =
		(undef,   undef,     undef,    undef);

    	($basetype, undef, $resttype) = 
		($type =~ m%([^\@]+)(\@(.*))?%);
	if (!$basetype) {
		die("getQuotaTargetAndType  Line: ".__LINE__."\n".
		      "Could not determine quota type -- quota opts:\n\n".
		       Dumper($quota_copy));
	}
	if ($target) {
		$default_quota = ($target eq '*') ? 1 : 0;
	} else {
		$default_quota = $quota_copy->{default};
		if ($default_quota) {
			$target = '*';
		}
	}
	if ($resttype) {
		$volopt = undef;
		$treeopt = $resttype;
	}
	if ($target) {
		if ($basetype eq 'user') {
			$useropt = $target;
		} elsif ($basetype eq 'group') {
			$groupopt = $target;
		} elsif ($basetype eq 'tree') {
			$treeopt = $target;
		}
	}
	if (!defined($treename) && $treeopt) {
		if ($treeopt =~ (m%/vol/%)) {
			($treepath, undef, $treename) =
				($treeopt =~ m%(/vol/[^/]+(/(.*))?)%);
			if (!defined($volname)) {
				($volpath, $volname) =
					($treeopt =~ m%(/vol/([^/]+))%);
				$volopt = undef;
			}
		} else {
			$treename = $treeopt;
		}
	}
	if ( !defined($volname) && defined($volopt) ) {
		if ($volopt =~ (m%/vol/%)) {
			($volpath, $volname) = ($volopt =~ m%(/vol/([^/]+))%);
		} else {
			$volname = $volopt;
		}
	}

	# volname, treename s/b defined by now....
	if (!defined($volname)) {
		die("getQuotaTargetAndType  Line: ".__LINE__."\n".
		      "Could not determine quota volume -- quota opts:\n\n".
		      Dumper($quota_copy));
	}
	if (!defined($volpath)) {
		$volpath = "/vol/$volname";
	}
	if (!defined($treepath)) {
		if (!defined($treename)) {
			$treename = "";
		}
		if ($treename && !$default_quota) {
			$treepath = "${volpath}/$treename";
		} else {
			$treepath = $volpath;
		}
	}
	if ($treename eq '*' && $basetype ne 'tree') {
		die("getQuotaTargetAndType  Line: ".__LINE__."\n".
		      "'wildcard' tree name invalid except for 'tree' quota type\n\n".
		      Dumper($quota_copy));
	}

	if ($basetype eq 'user') {
		if ($default_quota) {
			$username = '*';
			$target = '*';
		} elsif ($useropt) {
			$username = $useropt;
			$target = $username;
		}
		if (!$username) {
			die("getQuotaTargetAndType  Line: ".__LINE__."\n".
			      "Could not determine 'user' target -- quota opts:\n\n".
			       Dumper($quota_copy));
		}
		$default_quota = ($target eq '*') ? 1 : 0;
		$type = "user\@$treepath";
	} elsif ($basetype eq 'group') {
		if ($default_quota) {
			$groupname = '*';
			$target = '*';
		} elsif ($groupopt) {
			$groupname = $groupopt;
			$target = $groupname;
		}
		if (!$groupname) {
			die("getQuotaTargetAndType  Line: ".__LINE__."\n".
			      "Could not determine 'group' target -- quota opts:\n\n".
			       Dumper($quota_copy));
		}
		$default_quota = ($target eq '*') ? 1 : 0;
		$type = "group\@$treepath";
	} elsif ($basetype eq 'tree') {
		if (!$treename) {
			# Could either 'croak', or else assume 'default' quota in this case.
#			$default_quota = 1;
			die("getQuotaTargetAndType  Line: ".__LINE__."\n".
			      "Could not determine 'tree' target -- quota opts:\n\n".
			       Dumper($quota_copy));
		} elsif (!defined($target) && !defined($default_quota)) {
			$default_quota = ($treename eq '*') ? 1 : 0;
		}
		if ($default_quota) {
			$target = '*';
			$default_quota = 1;
			$type = "tree\@$volpath";
		} else {
			$target = $treepath;
			$type = "tree";
		}
	}

	$default_quota = ($default_quota) ? 1 : 0;
	### $quota_copy = {};
	for my $key ( qw( type basetype target id volpath volname vol treepath
				treename tree username groupname default ) ) {
		delete($quota_copy->{$key});
	}

	my $quota_out = { %$quota_copy };
	$quota_out->{type}         = $type         if ($type);
	$quota_out->{basetype}     = $basetype     if ($basetype);
	$quota_out->{target}       = $target       if ($target);
	$quota_out->{id}           = $target       if ($target);
	$quota_out->{volpath}      = $volpath      if ($volpath);
	$quota_out->{volname}      = $volname      if ($volname);
	$quota_out->{vol}          = $volname      if ($volname);
	$quota_out->{treepath}     = $treepath     if ($treepath);
	$quota_out->{treename}     = $treename     if ($treename);
	$quota_out->{tree}         = $treename     if ($treename);
	$quota_out->{qtree}        = $treename     if ($treename);
	$quota_out->{user}         = $username     if ($username) && !$quota_out->{user};
	$quota_out->{username}     = $username     if ($username);
	$quota_out->{group}        = $groupname    if ($groupname) && !$quota_out->{group};
	$quota_out->{groupname}    = $groupname    if ($groupname);
	$quota_out->{default}      = $default_quota;

	my $bare_quota;
	if (!$basetype || !$target) {
		$bare_quota = undef;
	} else {
		$bare_quota = {
				target => $quota_out->{target},
				type => $quota_out->{type},
				usermap => $quota_out->{usermap},
				disk => $quota_out->{disk},
				files => $quota_out->{files},
				thold => $quota_out->{thold},
				sdisk => $quota_out->{sdisk},
				sfiles => $quota_out->{sfiles}
			};
	}

	logDbg("Exit getQuotaTargetAndType\n\n".Dumper($bare_quota)."\n\n".Dumper($quota_out));
	return( $bare_quota, $quota_out );
}


=head2 _mergeQuotaEntries($filerspec, $newquotas);

	sets the quota for the specified quota target in /etc/quotas

	returns success or failure

=over 4

=item filerspec

	Required: filer specification

	newquotas:	(Required)
			ARRAYREF, whose elements are each a HASHREF which describes a
			single quota entry. The HASHREF fields are:

			target	=> ..
			type	=> ..
			disk	=> ..
			files	=> ..
			thold 	=> ..
			sdisk	=> ..
			sfiles	=> ..
			usermap => ..

	deleted:	(Optional)
			ARRAYREF, whose elements are each a HASHREF which describes
			a single (existing) quota entry. These entries will be deleted
			from the /etc/quotas file.

=back

=cut 

sub _mergeQuotaEntries($$;$)
{
	my ($filerspec, $newquotas, $deleted) = @_;
    		my ($target, $type);
	my ($addcommentline) = 1;

	# '%newquota_hash' will be used to find
	# conflicts between existing and new quota entries
	$newquotas = [] unless ($newquotas);
	my %newquota_hash = ();
	foreach my $q (@$newquotas) {
		_validateQuotaOpt(%$q);
		if (!$q->{target}) {
			croak("quota target is undefined");
		}
		if ($q->{type} !~ /^(user|group|tree)(@.*)?$/) {
			croak("invalid quota type");
		}
		if ( ($q->{type} !~ /^tree/) && ($q->{type} !~ m%@/vol/[^/]+%) ) {
			croak("invalid quota type");
		}
		if ( $q->{default} && ($q->{type} !~ m%@/vol/[^/]+%) ) {
			croak("invalid quota type");
		}
		my $key = "$q->{target}::$q->{type}";
		$newquota_hash{$key} = $q;
	}

	$deleted = [] unless ($deleted);
	my %deleted_hash = ();
	foreach my $q (@$deleted) {
		_validateQuotaOpt(%$q);
		my $key = "$q->{target}::$q->{type}";
		$deleted_hash{$key} = $q;
	}

	my $filerconn = make_conn($filerspec);

	# don't croak here since split may report here if /etc/quotas
	# is empty
	my $oldquotatext = rdfile($filerconn, "/etc/quotas");
	my @oldquotalines = split(/\n/, $oldquotatext);

	my @old_mapped = ();
	my @old_unmapped = ();
	my $current_list = \@old_unmapped;
	my $last_map_line = "OFF";

	# Sort existing quotas into 'mapped' and 'unmapped' lists
	foreach my $line (@oldquotalines) {
		if ($line =~ /^\s*$/) {
			next;
		} elsif ($line =~ /^\s*#/) {
			push(@$current_list, $line);
			next;
		}
		$line =~ s/\s+/ /;
		$line =~ s/\s*,\s*/,/g;
		$line =~ s/\s*$//;
		$line =~ s/^\s*//;
		if ($line =~ /^\s*QUOTA_PERFORM_USER_MAPPING\s*ON/i) {
			$current_list = \@old_mapped;
			$last_map_line = "ON";
		} elsif ($line =~ /^\s*QUOTA_PERFORM_USER_MAPPING\s*OFF/i) {
			$current_list = \@old_unmapped;
			$last_map_line = "OFF";
		} else {
			push(@$current_list, $line);
		}
	}

	my @unmapped = ();
	my @mapped = ();

	# delete any conflicting existing quota entries
	# from the 'unmapped' list
	foreach my $line (@old_unmapped) {
		$addcommentline = 0
		    if ($line =~ /\s+target\s+type/);
		if ($line =~ /^\s*$/) {
			next;
		}
		if ($line =~ /^\s*#/) {
			push(@unmapped, $line);
			next;
		}
		($target, $type) = ($line =~ /^\s*(\S+)\s+(\S+)/);
		my $key = "${target}::${type}";
		if (!$newquota_hash{$key} && !$deleted_hash{$key}) {
			push(@unmapped, $line)
		}
	}

	# delete any conflicting existing quota enties
	# from the 'mapped' list
	foreach my $line (@old_mapped) {
	        if ($line =~ /\s+target\s+type/) {
			next;
		}
		if ($line =~ /^\s*$/) {
			next;
		}
		if ($line =~ /^\s*#/) {
			push(@mapped, $line);
			next;
		}
		($target, $type) = ($line =~ /^\s*(\S+)\s+(\S+)/);
		my $key = "${target}::${type}";
		if (!$newquota_hash{$key} && !$deleted_hash{$key}) {
			push(@mapped, $line)
		}
	}

	# if old quota file has no comments on column names, add it
	if ($addcommentline == 1) {
		my $commentline = sprintf "# %15s %15s %6s %6s %6s %6s %6s", 
			"target", "type", "disk", "files", "thold", 
			"sdisk", "sfiles";
		unshift(@unmapped, $commentline);
	}

	# Give default values to 'disk', 'files', 'sdisk', 'sfiles', and 'thold'
	foreach my $q (@$newquotas) {
		$q->{usermap} = 0 unless ($q->{usermap} eq "1");
		foreach my $key (qw(disk files sdisk sfiles thold)) {
			if (!defined($q->{$key}) || $q->{$key} eq "") {
				$q->{$key} = "-";
			}
		}
	}

	# For each new quota entry, add it to the 'mapped' list if
	# its 'usermap' value is 1, or to the 'unmapped' list if its
	# 'usermap' value is 0. If its 'usermap' value is undef,
	# then use the 'last_map_line' value to determine 'mapped' or
	# 'unmapped'.
	#
	# We preserve the final 'usermap' state of the existing file,
	# so that any quota entries appended to the new file will have
	# the same 'usermap' state as if they were appended to the
	# existing file.
	foreach my $q (@$newquotas) {
		my $newline = sprintf "%15s %15s %6s %6s %6s %6s %6s",
			$q->{target}, $q->{type}, $q->{disk},
			$q->{files}, $q->{thold}, $q->{sdisk},
			$q->{sfiles};
		$newline =~ s/^\s+//;
		if (!defined($q->{usermap})) {
			if ($last_map_line eq "ON") {
				push(@mapped, $newline);
			} else {
				push(@unmapped, $newline);
			}
		} elsif ($q->{usermap} == 1) {
			push(@mapped, $newline);
		} else {
			push(@unmapped, $newline);
		}
	}

	# Construct the new file from the 'unmapped' and 'mapped' lists
	# of quotas, and preserve the final 'usermap' state of the
	# existing file (ON or OFF).
	my @output_list = ();
	if (@unmapped) {
		push(@output_list, @unmapped);
	}
	if (@mapped || $last_map_line eq "ON") {
		push(@output_list, "QUOTA_PERFORM_USER_MAPPING ON");
	}
	if (@mapped) {
		push(@output_list, @mapped);
		if ($last_map_line eq "OFF") {
			push(@output_list, "QUOTA_PERFORM_USER_MAPPING OFF");
		}
	}

	# write out the new file
	for (my $i=0; $i<=$#output_list; $i++) {
		$output_list[$i] =~ s/^\s+//;
	}
	my $newquotatext = join("\n", @output_list);
	$newquotatext .= "\n";
	logTrc("updating quotas file with the rules .. $newquotatext");
	wrfile($filerconn, "/etc/quotas", $newquotatext)
		or croak("can't write to /etc/quotas : $@");
	
	return 1;
}


# _getFilerCIFSSession
#
#	Returns the current CIFS server name for a filer
#
sub getCIFSServer($%)
{
    
	my ($filerspec, %options) = @_;
	my $filerconn = make_conn($filerspec);
	my $timeout = $options{timeout} || $main::CMDTIMEOUT || 300;
	
	if (!is_cifs_setup(host=> $filerconn)) {
		logWarn("CIFS not running, server name is undefined");
		return(undef);
	}
	# extract Server name from filer
	my $string = sayEOF($filerconn, "cifs sessions", timeout=>$timeout)
			or croak $@;
	my $server = "";
	if ($string =~ /^Server Registers as '([\w\d\-]*)' in workgroup/) {
		$server = $1;
	} else {
		logWarn("Can't parse server name from cifs sessions, ".
			"is CIFS setup?");
	}
	return($server);
}

# terminate_cifs
#
#	Stops CIFS on filer
#	Companion to SPARSE::Util::Filer_Utils::setup_cifs
#
#	SPARSE::Util::Filer_Utils::terminate_cifs takes a SPARSE
#	'filer' object as arg, not a NATE connectrec or filerspec.
#
sub terminate_cifs($)
{
	my ($filerspec) = @_;
	my $filerconn = make_conn($filerspec);
	sayEOF($filerconn, "cifs terminate");
}


# _parse_quota_report_line
#
#	Parses line from quota report output, and returns
#	HASHREF which represents this line
#
#	Fields:
#		type
#		id
#		target
#		volume
#		tree
#		kbytes_used
#		kbytes_limit
#		files_used
#		files_limit
#		quotaspec
#
sub _parse_quota_report_line($;$)
{
	my ($line, $flag) = @_;
	my $fields = {};
	my $msg;
	if ($line =~ /quotas are off/) {
		$msg = "_parse_quota_report_line  Line: ".__LINE__.
				"\nquotas are off! error=$line";
		logcomment($msg);
		croak($msg);
	}

	# logDbg("Enter _parse_quota_report_line  Line: ".__LINE__."\n\nLine: \"$line\"\nflag: \"$flag\"");

    $flag =~ s/[^a-z]//g;
    # '-q' and '-u' flags do not affect line structure, they can be ignored.
    # '-x' and '-t' options give same line structure ('multiple IDs' must be handled externally to this fcn).
    # '-v' option is not handled yet.
    # -- This leaves only 's' and 'x' options to be handled, and the combinations 'sx' and 'xs'.
    $flag =~ s/q//;
    $flag =~ s/u//;
    $flag =~ s/-//g;
    if ($flag =~ /x/ || $flag =~ /t/) {
	  $flag =~ s/[tx]//g;
	  $flag .= 'x';
    }

    if ($flag eq '') {
	# note that $9 is not always there
	#                     $1                  $2                        $3                  $4             $5       $6       $7       $8                $9
	if ($line =~ /^(user|group|tree)\s+([\w\d\*_\-,\.\\\/]+)\s+([\w\d\*_\-,\.\\\/]+)\s+([\w\d\.\_\-\\,]+)\s+(\d+|-)\s+(\d+|-)\s+(\d+)\s+(\d+|-)\s*([\w\d\.\_\*\-\/\\]*)\s*$/) {

		$fields->{type} 	= $1;
		$fields->{id}		= $2;
		$fields->{target}	= $2;
		$fields->{volume} 	= $3;
		$fields->{tree} 	= $4;
		$fields->{kbytes_used} 	= $5;
		$fields->{kbytes_limit}	= $6;
		$fields->{files_used} 	= $7;
		$fields->{files_limit} 	= $8;
		$fields->{quotaspec} 	= $9 . $10;

		$fields->{target} = $fields->{id};
		$fields->{disk_used} = $fields->{kbytes_used};
		$fields->{disk_limit} = $fields->{kbytes_limit};

	} else {
		$msg = "_parse_quota_report_line  Line: ".__LINE__.
		               	"\nCan't parse quota report format!\n".
		                "flag=\"$flag\"\nQuota Entry: \"$line\"";
		logcomment($msg);
		croak($msg);
	}
	return $fields;


    } elsif ($flag eq 's') {

	#                     $1                  $2                      $3                   $4              $5        $6       $7        $8       $9        $10        $11
	if ($line =~ /^(user|group|tree)\s+([\w\d\*_\-,\.\\\/]+)\s+([\w\d\*_\-,\.\\\/]+)\s+([\w\d\.\_\-]+)\s+(\d+|-)\s+(\d+|-)\s+(\d+|-)\s+(\d+|-)\s+(\d+|-)\s+(\d+|-)\s*([\w\d\.\_\*\-\/\\]*)\s*$/) {

		$fields->{type} 		= $1;
		$fields->{id}			= $2;
		$fields->{target}		= $2;
		$fields->{volume} 		= $3;
		$fields->{tree} 		= $4;
		$fields->{kbytes_used} 		= $5;
		$fields->{kbytes_limit}		= $6;
		$fields->{kbytes_soft_limit}	= $7;
		$fields->{files_used} 		= $8;
		$fields->{files_limit} 		= $9;
		$fields->{files_soft_limit} 	= $10;
		$fields->{quotaspec}   		= $11 . $12;

		$fields->{target} = $fields->{id};
		$fields->{disk_used} = $fields->{kbytes_used};
		$fields->{disk_limit} = $fields->{kbytes_limit};
		$fields->{disk_soft_limit} = $fields->{kbytes_soft_limit};

	} else {
		$msg = "_parse_quota_report_line  Line: ".__LINE__.
		               	"\nCan't parse quota report format!\n".
		                "flag=\"$flag\"\nQuota Entry: \"$line\"";
		logcomment($msg);
		croak($msg);
	}
	return $fields;

    } elsif ($flag eq 'x') {

	#                     $1                  $2                        $3              $4            $5        $6       $7          $8       $9        $10        $11
	if ($line =~ /^(user|group|tree)\s+([\w\d\*_\-,\.\\\/]+)\s+([\w\d\*_\-,\.\\\/]+)\s+([\w\d\_\-]+)\s+(\d+|-)\s+(\d+|-)\s+(\d+|-)\s+(\d+)\s+(\d+|-)\s*([\w\d\*\_\-\\\,\/]*)\s*$/) {	

		$fields->{type} 	= $1;
		$fields->{id}		= $2;
		$fields->{target}	= $2;
		$fields->{volume} 	= $3;
		$fields->{tree} 	= $4;
		$fields->{kbytes_used} 	= $5;
		$fields->{kbytes_limit}	= $6;
		$fields->{thold}	= $7;
		$fields->{files_used} 	= $8;
		$fields->{files_limit} 	= $9;
		$fields->{quotaspec} 	= $10 . $11;

		$fields->{target} = $fields->{id};
		$fields->{disk_used} = $fields->{kbytes_used};
		$fields->{disk_limit} = $fields->{kbytes_limit};

	} else {
		$msg = "_parse_quota_report_line  Line: ".__LINE__.
		               	"\nCan't parse quota report format!\n".
		                "flag=\"$flag\"\nQuota Entry: \"$line\"";
		logcomment($msg);
		croak($msg);
	}
	return $fields;

    } elsif ($flag eq 'sx' || $flag eq 'xs') {

	#                     $1                  $2                      $3                   $4              $5        $6       $7        $8       $9        $10        $11
	if ($line =~ /^(user|group|tree)\s+([\w\d\*_\-,\.\\\/]+)\s+([\w\d\*_\-,\.\\\/]+)\s+([\w\d\.\_\-]+)\s+(\d+|-)\s+(\d+|-)\s+(\d+|-)\s+(\d+|-)\s+(\d+|-)\s+(\d+|-)\s+(\d+|-)\s*([\w\d\.\_\*\-\/\\]*)\s*$/) {

		$fields->{type} 		= $1;
		$fields->{id}			= $2;
		$fields->{target}		= $2;
		$fields->{volume} 		= $3;
		$fields->{tree} 		= $4;
		$fields->{kbytes_used} 		= $5;
		$fields->{kbytes_limit}		= $6;
		$fields->{kbytes_soft_limit}	= $7;
		$fields->{thold}		= $8;
		$fields->{files_used} 		= $9;
		$fields->{files_limit} 		= $10;
		$fields->{files_soft_limit} 	= $11;
		$fields->{quotaspec}   		= $12 . $13;

		$fields->{target} = $fields->{id};
		$fields->{disk_used} = $fields->{kbytes_used};
		$fields->{disk_limit} = $fields->{kbytes_limit};
		$fields->{disk_soft_limit} = $fields->{kbytes_soft_limit};

	} else {
		$msg = "_parse_quota_report_line  Line: ".__LINE__.
		               	"\nCan't parse quota report format!\n".
		                "flag=\"$flag\"\nQuota Entry: \"$line\"";
		logcomment($msg);
		croak($msg);
	}
	return $fields;

    } else {
		$msg = "_parse_quota_report_line  Line: ".__LINE__.
                	     "\nI can't read the report with flag $flag yet ... ";
		logcomment($msg);
		croak($msg);
    }
}



# _parse_quotas_file_line
#
#	Given a line from an /etc/quotas file,
#	return a HASHREF with entries for:
#		target
#		type
#		disk
#		files
#		thold
#		sdisk
#		sfiles
#
sub _parse_quotas_file_line($)
{
	my ($line) = @_;
	my ($target, $type, $disk, $files, $thold, $sdisk, $sfiles) =
		split(/\s+/, $line);
	my ($quota) = { target=>$target,
			type=>$type,
			disk=>$disk,
			files=>$files,
			thold=>$thold,
			sdisk=>$sdisk,
			sfiles=>$sfiles };
	(undef, $quota) = getQuotaTargetAndType($quota);
	return($quota);
}



sub _mount_root_vol($;$)
{
	my ($filerspec, $rootvol) = @_;
	
	$rootvol = "vol0" unless defined($rootvol);
	my $host = host($filerspec);

	my $rootmntdir = mount::mount($host->{default_ip}, "/vol/$rootvol")
		or croak("unable to mount root volume: $@");

	$MOUNTPOINT = $rootmntdir;
	return $rootmntdir;
}


sub _create_qtree($$$)
{
	my ($filerspec, $vol, $qtree) = @_;
	
	my $filerconn = make_conn($filerspec);

	my $qtreepath = "/vol/$vol/$qtree";
	if (!qtreePresent($filerconn, $qtreepath)) {
		qtreeCreate($filerconn, $qtreepath)
			or croak $@;
	}
		
	return $qtreepath;
}


=head2 createQtrees($filerspec, $volume, $ntrees)

	Create a number of qtrees in a volume.

=over 4

=item filerspec 

	Required Filer Specification

=item volume

	Required volume to contain qtrees

=item ntrees

	Required. Number of qtrees to create

=back

=cut

sub createQtrees($$$)
{
	my ($filerspec, $volume, $ntrees) =
               validate_mixed(@_,
			[{ type => SCALAR|HASHREF },
			 { type => SCALAR },
			 { type => SCALAR }],
			{}
		);
	my $filerconn = make_conn($filerspec);
	my $qtidx;
	my $basename = "qt_";
	my @qtrees = ();
	for ($qtidx=1000; ; $qtidx+=1000) {
		last if (!qtreePresent($filerconn, "/vol/$volume/$basename${qtidx}"));
	}

	for (my $ncreated=0; $ncreated<$ntrees; $ncreated++) {
		while (qtreePresent($filerconn, "/vol/$volume/$basename${qtidx}")) {
			$qtidx -= ($qtidx % 10);
			$qtidx += 10;
		}
		my $qtreepath = "/vol/$volume/$basename${qtidx}";
		qtreeCreate($filerconn, $qtreepath)
			or croak $@;
		push(@qtrees, $qtreepath);
		$qtidx++;
	}
	return (@qtrees);
}


my $_logTailStart = 0;

sub _markLogTail
{
        my ($filerspec) = @_;
        my $filerconn = make_conn($filerspec);
        my @oldlog = split(/\n/, rdfile($filerconn, "/etc/messages"));
        $_logTailStart = @oldlog;
}

sub _resetLogTail
{
        $_logTailStart = 0;
}

sub _getLogTail
{
        my ($filerspec) = @_;
        my $filerconn = make_conn($filerspec);
        my @log = split(/\n/, rdfile($filerconn, "/etc/messages"));
        my @newlog = @log;
        splice(@newlog, 0, $_logTailStart);
        $_logTailStart = 0;
        logresult('INFO', "New Log:\n\n".Dumper(\@newlog)."\n\n");
#        return( join("\n", @newlog) );
        return( @newlog );
}

my $debug_cifs_errors = 1;



=head2 map_cifs_share (%opts)

	Wrapper around MapShare::map_cifs_share for diagnostics, etc.

=over 4

=item Optional Args -- see MapShare::map_cifs_share

=over

=item share

=item client

=item server

=item workgroup

=item domain

=item username

=item password

=item timeout

=back

=back

=cut

sub map_cifs_share(%)
{
	my (%opts) = @_;
	if (!$debug_cifs_errors) {
		return( MapShare::map_cifs_share(%opts) );
	}
	my $host_ip = $opts{server};
	my ($host_line) = grep(/lab.netapp.com/, split(/\n/, `/usr/local/bin/host $host_ip`));
	if ($host_line !~ m/([^\s\.]+)\S*\.netapp\.com/) {
	}
	my $host = $1;
	if (!$host) {
	}
        my $filerconn = make_conn($host);
	logresult('INFO', "Enter Quota::map_cifs_share -- calling MapShare::map_cifs_share");
	_markLogTail($filerconn);
	my ($err, $retval);
	eval { $retval = MapShare::map_cifs_share(%opts); $err = $@; };
	if ($err || $@) {
		$err = $@ unless ($err);
		sleep(3);
		my @loglines = _getLogTail($filerconn);
		if (@loglines) {
			my $loglines = join("\n", @loglines);
			logError("map_cifs_share: $err\nFrom syslog:\n\n$loglines");
		} else {
			logError("map_cifs_share: $err");
		}
	} else {
		_resetLogTail();
		return( $retval );
	}
}

=head2 log_quota_error($err)

	Checks error string, if string matches 'CONFIG' error test, logs
	a config error, otherwise logs a 'FATAL' error.

	The idea is to use this function to log errors returned by the 'main'
	routine of a test, rather than logging all these errors as 'FATAL' errors,
	as was done previously. We want to log certain errors from the 'main' routine
	as 'CONFIG' errors. For example, the CIFS "System error NNNNNN has occured" type
	errors, which indicate a problem with the CIFS client.

=over 4

=item $err

=back

=cut

sub log_quota_error($)
{
	my ($err) = @_;
	if ($err =~ /system error\D+(\d+)/i) {
		logresult('CONFIG', "CIFS error -- check CIFS client and filer:\n".$err);
	} else {
		logresult('FATAL', $err);
	}
}


=head2 vmstat($filerspec)

	Returns a hash table which contains the output of a 'vm_stat' command.

=over 4

=item filerspec 

	Required Filer Specification

=back

=cut

sub vm_stat($)
{
        my ($filerspec) = @_;
        my $filerconn = make_conn($filerspec);
	my $vms_output = eval { sayEOF($filerconn, "priv set test; vm_stat") };
	if ($vms_output =~ /unrecognized privilege/) {
		logWarn("vm_stat: unrecognized privilege");
		return(undef);
	} elsif ($vms_output =~ /vm_stat not found/) {
		logWarn("vm_stat: command not found");
		return(undef);
	} elsif ($@) {
		logWarn("vm_stat: $@");
		return(undef);
	} elsif ($vms_output !~ /^System/m || $vms_output !~ /^WAFL/m || $vms_output !~ /Buffers/m) {
		logWarn("vm_stat: Can't parse vm_stat output:\n\n$vms_output");
		return(undef);
	}
	# croak($@) if ($@);
	my %system_format = (
		"Total Pages in System:"	=>	"system_total_pages",
		"Total Pages Allocated:"	=>	"system_alloc_pages",
		"Total Free Pages:"		=>	"system_total_free_pages",
		"Non-WAFL Free Pages:"		=>	"system_non_wafl_free_pages",
		"WAFL Free Pages:"		=>	"system_wafl_free_pages",
		"no VM Free Pages:"		=>	"system_no_vm_free_pages",
	);
	my %wafl_format = (
		"Pages From WAFL:"		=>	"wafl_pages_from_wafl",
		"Pages Returned To WAFL:"	=>	"wafl_pages_returned",
		"Failures while stealing from WAFL:" => "wafl_failures_while_stealing",
		"Times Pages stolen immediately:" =>	"wafl_times_pages_stolen_immed",
		"Free Pages in WAFL:"		=>	"wafl_free_pages",
		"Free buffers in WAFL:"		=>	"wafl_free_buffers",
		"WAFL recycled bufs:"		=>	"wafl_recycled_buffers",
	);

	my $Line;
	my $vmstat_info = {};
	my @vms_lines = split(/\n/, $vms_output);
	while (@vms_lines) {
		$Line = shift(@vms_lines);
		last if ($Line =~ /^System/);
	}

	croak("Could not parse vm_stat output") unless (@vms_lines);
	while (@vms_lines) {
		$Line = shift(@vms_lines);
		last if ($Line =~ /^WAFL$/);
		if ($Line =~ /^\s*(\S.+:)\s*(\d+)/) {
			my $field = $system_format{$1};
			if ($field) {
				$vmstat_info->{$field} = $2;
			}
		}
	}

	croak("Could not parse vm_stat output") unless (@vms_lines);
	while (@vms_lines) {
		$Line = shift(@vms_lines);
		last if ($Line =~ /^Sleep\/Wakes$/);
		if ($Line =~ /^\s*(\S.+:)\s*(\d+)/) {
			my $field = $wafl_format{$1};
			if ($field) {
				$vmstat_info->{$field} = $2;
			}
		}
	}
	croak("Could not parse vm_stat output") unless (@vms_lines);

	while (@vms_lines) {
		$Line = shift(@vms_lines);
		last if ($Line =~ /^Hseg$/);
	}
	croak("Could not parse vm_stat output") unless (@vms_lines);
	$Line = shift(@vms_lines);
	croak("Could not parse vm_stat output") unless (@vms_lines);
###	while ( $Line = shift(@vms_lines) && ($Line =~ /^(\w+)(\s+\d+)+\s+(\d+)/) ) {
	while (@vms_lines) {
		$Line = shift(@vms_lines);
		last unless ($Line);
		if ($Line =~ /^(\w+)(\s+\d+)+\s+(\d+)/) {
			$vmstat_info->{"hseg_" . $1} = $3;
		}
	}

	croak("Could not parse vm_stat output") unless (@vms_lines);
	while (@vms_lines) {
		$Line = shift(@vms_lines);
		last if ($Line =~ /^Buffers MemoryPool/);
	}

	croak("Could not parse vm_stat output") unless (@vms_lines);
	while (@vms_lines) {
		$Line = shift(@vms_lines);
		last unless ($Line);
		if ($Line =~ /^\s*(\d+)\s+(\S+)$/) {
			$vmstat_info->{$2} = $1;
		}
	}
	return($vmstat_info);
}



#-----------------------------------------------------------------------
#  _rdfile, _wrfile
#-----------------------------------------------------------------------

# _rdfile($filerspec, $file, $rootvolmnt)
#
#	Internal function to read a file.
#	Will use NFS if "rootvolmount" is defined and
#	not null. Otherwise will use 'rdfile'
#
#  sub _rdfile($$;$);
#
sub _rdfile {
	my ($filerspec, $file, $rootvolmnt) = validate_mixed(@_,
			[
				{ type => SCALAR|HASHREF},
				{ type => SCALAR},
				{ type => SCALAR|UNDEF},
			],
			{},
	);
	my $filerconn = make_conn($filerspec);
	my @lines;
	if ($rootvolmnt && -d $rootvolmnt) {
		my $path = "$rootvolmnt/$file";
		open(_RDFILE, "$path")
		    or croak("Cannot open file for reading $path: $!");
		@lines = <_RDFILE>;
		map { chomp() }  @lines;
		close(_RDFILE) 
		    or die("close: $!");
	} else {
		my $linebuff = rdfile($filerconn, $file);
		croak($@) if ($@);
		@lines = split(/\n/, $linebuff);
	}
	return( @lines );
}


# _wrfile($filerspec, $file, $rootvolmnt, @lines)
#
#	Internal function to write a file.
#	Will use NFS if "rootvolmount" is defined and
#	not null. Otherwise will use 'wrfile'
#
#  sub _wrfile($$$@);
#
sub _wrfile {
	my ($filerspec, $file, $rootvolmnt, @lines) = @_;
	my @validate = ($filerspec,  $file, $rootvolmnt);
	($filerspec, $file, $rootvolmnt)
	     = validate_mixed(@validate,
			[
				{ type => SCALAR|HASHREF},
				{ type => SCALAR},
				{ type => SCALAR|UNDEF},
			],
			{},
	);
	my $filerconn = make_conn($filerspec);
	map { chomp() }  @lines;
	my $linebuff = join("\n", @lines);
	$linebuff .= "\n";
	if ($rootvolmnt && -d $rootvolmnt) {
		my $path = "$rootvolmnt/$file";
		open(_WRFILE, ">$path")
		    or croak("Cannot open file for writing $path: $!");
		print _WRFILE $linebuff;
		close(_WRFILE) 
		    or die("close: $!");
	} else {
		wrfile($filerconn, $file, $linebuff)
		    or croak $@;
	}
	return( @lines );
}



#-----------------------------------------------------------------------
#    EMS log fcns.
#-----------------------------------------------------------------------

sub openEMSLog($);
sub markEMSLog($);
sub closeEMSLog();
sub pollEMSLogLines();
sub pollEMSLogRecords();

my $_ems_log_mark = undef;

sub openEMSLog($) {
	my ($filer) = @_;
	my $root_mnt = mount_root_vol_if_needed(filer=>$filer, localhost=>1);
	die("mount_root_vol_if_needed: $@") unless ($root_mnt);
	my $ems_path = $root_mnt . "/" . "etc/log/ems";
	if (!defined($_ems_log_mark)) {
		open(EMS, "<", $ems_path) or die("Could not open EMS log for reading: $!");
	}
	seek(EMS, 0, 0);
	$_ems_log_mark = tell(EMS);
	return(1);
}

sub markEMSLog($) {
	my ($filer) = @_;
	openEMSLog($filer)
	    or die("openEMSLog: $@");
	seek(EMS, 0, 2);
	$_ems_log_mark = tell(EMS);
	sleep(4);
	return(1);
}

sub closeEMSLog() {
	if (defined($_ems_log_mark)) {
		close(EMS) or die("close: $!");
		$_ems_log_mark = undef;
	}
	return(1);
}

sub pollEMSLogLines() {
	if (!defined($_ems_log_mark)) {
		die("must run 'markEMSlog' before running this routine");
	}
	my @lines = ();
	my $last_found = 0;
	my $last_line;
	while (<EMS>) {
		chomp;
		next if (/^\s*$/);
		push(@lines, $_);
		$last_line = $_;
	}
	$last_found = ($last_line =~ m%<\/LR>%);

	return(@lines) if ($last_found);

#################################################
#	sleep(10);
#	while (<EMS> && !$last_found) {
#		chomp;
#		next if (/^\s*$/);
#		push(@lines, $_);
#		$last_line = $_;
#		$last_found = ($_ =~ m%<\/LR>%);
#	}
#	if (!$last_found) {
#		logWarn("Quota  Line: ".__LINE__."   pollEMSLog   No 'record end' (</LR>) line found -- ".
#			  "Last Line:\n\"$last_line\""); ############
#	}
#################################################

	return( @lines );
}


sub pollEMSLogRecords() {
	my @lines = pollEMSLogLines();
	die("pollEMSLogLines: $@") if ($@);
	my @records = ();
	my ($rec, $event);
	my $flag_to_field = { 'd'=>'date', 'n'=>'filer', 'pn'=>'partner', 't'=>'timestamp', 'id'=>'id',
				  'p'=>'priority', 's'=>'status', 'o'=>'ontap_thread', 'vf'=>'vfiler' };
	my @priority_list = qw( emergency alert critical error warning notice info debug );

	foreach my $curr_line ( @lines ) {
		chomp($curr_line);
		my $rem_line;
		next if ($curr_line =~ /^\s*$/);
		if ($rec && defined($rec->{lines})) {
			push(@{$rec->{lines}}, $curr_line);
		}
		if ($curr_line =~ m%<LR (.*)>%) {
			$rec = {};
			$rec->{lines} = [$curr_line];
#			my $hdr = {};
#			$rec->{header} = $hdr;
			$rec->{events} = [];
			$rem_line = $1;
			while ($rem_line) {
				if ($rem_line =~ m%(\w+)="([^"]*)"(.*)%) {
					my $flg = $1;
					my $val = $2;
					$rem_line = $3;
					my $fld = $flag_to_field->{$flg} || $flg;
#					$hdr->{$fld} = $val;
					$rec->{$fld} = $val;
				} else {
					logWarn("pollEMSLogRecords  Line: ".__LINE__."  Can't parse line:\n".
						  "\"$curr_line\"\n\n\"$rem_line\"");
					return();
				}
			}
		} elsif ($curr_line =~ m%<([^/]\S*)(.*)%) {
			$event = { event_type=>$1 };
			$rem_line = $2;	
			while ($rem_line) {
				if ($rem_line =~ m%(\w+)="([^"]*)"(.*)%) {
					my $fld = $1;
					my $val = $2;
					$rem_line = $3;
					$event->{$fld} = $val;
				} elsif ($rem_line =~ m%/>%) {
					$rem_line = "";
					push(@{$rec->{events}}, $event);
					$event = undef;
				} else {
					logWarn("pollEMSLogRecords  Line: ".__LINE__."  Can't parse line:\n".
						  "\"$curr_line\"\n\n\"$curr_line\"");
					return();
				}
			}
		} elsif ($curr_line =~ m%</LR>%) {
			push(@records, $rec);
			$rec = undef;
			$event = undef;
		} else {
			if (!$event) {
#				die("pollEMSLogRecords:  No active 'event' while reading line:\n\"$curr_line\"");
				logWarn("pollEMSLogRecords:   No active 'event' while reading line:\n\"$curr_line\"");
				return();
			}
			$rem_line = $curr_line;
			while ($rem_line) {
				if ($rem_line =~ m%(\w+)="([^"]*)"(.*)%) {
					my $fld = $1;
					my $val = $2;
					$rem_line = $3;
					$event->{$fld} = $val;
				} elsif ($rem_line =~ m%/>%) {
					$rem_line = "";
					push(@{$rec->{events}}, $event);
					if (!defined($rec->{event})) { 
						$rec->{event} = $event;
					}
					$event = undef;
				} else {
					logWarn("pollEMSLogRecords  Line: ".__LINE__."  Can't parse line:\n".
						  "\"$curr_line\"\n\n\"$curr_line\"");
					return();
				}
			}
		}
	}
	return( @records );
}



#-----------------------------------------------------------------------
# END block to cleanup ..
#-----------------------------------------------------------------------

END {
	if ($MOUNTPOINT ne "") {
		logTrc("unmounting mountpnt $MOUNTPOINT");
		mount::umount($MOUNTPOINT)
			or logWarn("umount failed: $@");
		$MOUNTPOINT = "";
	}
};



1;


