#$Id: //depot/prod/test/main/lib/Raid/Label.pm#1 $
##
## @summary Label - Library module for Label command
## @author kbala@netapp.com
## @status private 
## @pod here

=head1 NAME

Label

=head1 DESCRIPTION

This library contains all raid label commands and utililties.

=head1 SYNOPSIS

use Label;

my $host = host($FILER);

=cut

package Label;

use strict;
use warnings;

use Carp qw(confess croak);
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF UNDEF);
use Tharn qw(say script_result host logcomment logresult param);
use Module::Common qw(validate_mixed);

BEGIN {
       Params::Validate::validation_options(
				            on_fail => sub { my @msgs = @_; chomp(@msgs);
        			            Carp::confess( Tharn::script_result(@msgs) ) }
  			 	           );
}

use vars ('@ISA', '@EXPORT', '@EXPORT_OK');
require Exporter;
use Utils qw(make_conn);
use Data::Dumper;
use RaidInfo;
use SayWrap;

@ISA = qw(Exporter);
@EXPORT = qw(
             label_list
             label_edit
             label_editv1
             label_examine
             label_examinev1
             label_restore
             label_keep
             label_save
             label_compare
             label_makespare
             label_copy
             label_makeraid
             label_buildtree
             label_summary
             label_edit_interact
             label_edit_interactv1
             label_wipe
             label_wipev1
            );

my $RaidSqa_TIMEOUT = 300;

=head2 label_list 

label_list(host=>$host, conntype=>$conntype, volname=>$volname, fsid=>$fsid)

-or-

label_list(host=>$host, conntype=>$conntype, volname=>$volname)

-or-

label_list(host=>$host, conntype=>$conntype)

Shows how disks will be assimilated (or not) into aggregates. 

This is useful for identifying the structure of an entire aggregate, or seeing if a disk doesn't get assimilated properly.

It returns label list output 

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item volname

(Optional) : Name of aggregate  or traditional volume to be created.

=item fsid

(Optional) : fsid of the disk

=back

=cut

#------------------------------------------------------------------------------

sub label_list {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                volname => { type=>SCALAR, optional=>1 },
                                fsid => { type=>SCALAR, optional=>1 },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $fsid = $opts{fsid};
    my $volname = $opts{volname};
    my ($cmd, $fc, $output);

    if ( defined($volname) ) {
        $cmd = "-v $volname";
    } elsif (defined($fsid) ) {
        $cmd = "-f fsid $fsid";
    } else {
        $cmd = "";
    }

    $fc = make_conn($host, $conntype);
        
    (undef, undef, $output) = say($fc, "label list $cmd", '-prompt', "\\*> ");

    die "label list $cmd failed" if ($output !~ /Raidtree/);

    return $output;
}

#------------------------------------------------------------------------------

=head2 label_summary

label_summary(host=>$host)

Shows a quick snapshot of the labels of all the disks in the system. 

Good for identifying which disks belong to an aggregate
 
The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in
maintenance mode)

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=back

=cut

#------------------------------------------------------------------------------

sub label_summary {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
				filermode => {type=>SCALAR, default=>"normal", optional=>1 },
                                conntype => {type=>SCALAR, default=>"#rsh"},
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my ($fc, $output);

    if ($opts{filermode} ne "normal") {
        $opts{conntype} = "#console";
    }

    $fc = make_conn($host, $conntype);

    (undef, undef, $output) = say($fc, "label summary", '-prompt', "\\*> ");

    return $output;
}

#------------------------------------------------------------------------------

=head2 label_examine

label_examine(host=>$host, conntype=>$conntype, disk=>$disk)

-or-

label_examine(host=>$host, disk=>$disk)

Dumps the contents of one disk's labels in excruciating detail.

You can identify problems with the label contents and glean some information about the state of an aggregate with this command. 

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in
maintenance mode)

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Required) : Name of disk to be examined

=back

=cut

#------------------------------------------------------------------------------

sub label_examine {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
				filermode => {type=>SCALAR, default=>"normal", optional=>1 },
                                conntype => {type=>SCALAR, default=>"#rsh"},
                                disk => { type=>SCALAR },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};
    my ($fc, $output);

    if ($opts{filermode} ne "normal") {
        $opts{conntype} = "#console";
    }

    $fc = make_conn($host, $conntype);

    (undef, undef, $output) = say($fc, "label examine $disk", '-prompt', "\\*> ");

    return $output;
}

#------------------------------------------------------------------------------

=head2 label_examinev1

label_examine(host=>$host, conntype=>$conntype, disk=>$disk)

-or-

label_examinev1(host=>$host, disk=>$disk)

Like label examine, but for the old RAID (RAIDv1) labels. This lets you look at useful information like the disk layout table (disk TOC)

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in
maintenance mode)

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Required) : Name of disk to be examined

=back

=cut

#------------------------------------------------------------------------------

sub label_examinev1 {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                                conntype => {type=>SCALAR, default=>"#rsh"},
                                disk => { type=>SCALAR },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};
    my ($fc, $output);

    if ($opts{filermode} ne "normal") {
        $opts{conntype} = "#console";
    }

    $fc = make_conn($host, $conntype);

    (undef, undef, $output) = say($fc, "label examinev1 $disk", '-prompt', "\\*> ");

    return $output;
}

#------------------------------------------------------------------------------

=head2 label_compare

label_compare(host=>$host, conntype=>$conntype, disk1=>$disk, disk2=>$disk)

Compare the labels of two disks. 

This used to be useful when doing label copying; with the advent of label makeraid and label buildtree

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item filermode

(Optional) : The mode filer is up at. It can be either normal or maint (if it is running in
maintenance mode)

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk1

(Required) : Name of disk1  

=item disk2

(Required) : Name of disk2

=back

=cut

#------------------------------------------------------------------------------

sub label_compare {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
			        filermode => {type=>SCALAR,default=>"normal",optional=>1 },
                                conntype => {type=>SCALAR, default=>"#rsh"},
                                disk1 => { type=>SCALAR },
                                disk2 => { type=>SCALAR },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk1 = $opts{disk1};
    my $disk2 = $opts{disk2};

    my $fc = make_conn($host, $conntype);
    my $output;

    if ($opts{filermode} ne "normal") {
        $opts{conntype} = "#console";
    }
	
    (undef, undef, $output) = say($fc, "label compare $disk1 $disk2",'-prompt', "\\*> ");

    die "Disk $disk1 does not exist" if ($output =~ /$disk1 does not exist/);
    die "Disk $disk2 does not exist" if ($output =~ /$disk2 does not exist/);

    return $output;
}


#------------------------------------------------------------------------------

=head2 label_edit_interact

label_edit_interact(host=>$host, conntype=>$conntype, disk=>$disk, param=>[ [$label, $value], [$label1, $value1] ] )

It lets you change most (but not all) values of the new RAID (RAIDv2) labels

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Required) : Name of disk to edit

=item param

(Required) : It is a array reference with label to edit and value for it. 
             eg [ [ magic, 100 ], ["software_version", "10"] ]

=back

=cut

#------------------------------------------------------------------------------

sub label_edit_interact {
    my (%opts) = validate( @_, {
                                 host => { type=>OBJECT, isa=>'Hostrec' },
                                 conntype => {type=>SCALAR, default=>"#console"},
                                 disk => { type=>SCALAR },
                                 param => { type=>ARRAYREF },
                                }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};
    my $label_to_change = $opts{label_to_change};
    my $param = $opts{param};
    my $param_count = scalar(@$param) - 1;
    my ($fc, $match, $before, $output, $prompt, @param, $flag);

    $fc = make_conn($host, $conntype);
    $flag = 0;

    foreach my $key_index (0..$param_count) {

        if ($param->[$key_index]->[0] =~ / /) {
	    @param = split (/ /, $param->[$key_index]->[0]);
            $flag = 1;
	}
		
	if ($flag == 0) {
	    say($fc, "label edit", -prompt, "Enter name of disk whose labels are to be edited:");
                        
  	    ($before, $output, $prompt) = say($fc, "$disk", -prompt, "Enter name of *");
            if ( $prompt =~ /Error/) {
                say($fc, "\n", -prompt, ">");
                return ($prompt);
            }

	    say($fc, "$param->[$key_index]->[0]", -prompt, "Field $param->[$key_index]->[0] is");	
	    say($fc, "$param->[$key_index]->[1]", -prompt, "Enter name of top-level field to edit:");
	    say($fc, "\n", -prompt, "Enter name of disk whose labels are to be edited:");
	    say($fc, "\n", -prompt, ">");
	} else {
            say($fc, "label edit", -prompt, "Enter name of disk whose labels are to be edited:");

            ($before, $output, $prompt) = say($fc, "$disk", -prompt, "Enter name of *");
            if ( $prompt =~ /Error/) {
                say($fc, "\n", -prompt, ">");
                return ($prompt);
            }

            say($fc, "$param[0]", -prompt, "Enter name of $param[0]object field to edit:");
	    say($fc, "$param[1]", -prompt, "Field $param[1] is");
	    say($fc, "$param->[$key_index]->[1]", -prompt, "Enter name of $param[0]object field to edit:");
	    say($fc, "\n", -prompt, "Enter name of top-level field to edit:");	
	    say($fc, "\n", -prompt, "Enter name of disk whose labels are to be edited:");
	    say($fc, "\n", -prompt, ">");
	    $flag = 0;
	}
		
    }
    return 1;
}

#------------------------------------------------------------------------------

=head2 label_editv1_interact

label_edit_interactv1(host=>$host, conntype=>$conntype, disk=>$disk, param=>[ [$label, $value], [$label1, $value1] ] )

Like label_edit_interact, but lets you edit the old RAID (RAIDv1) labels. 

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Required) : Name of disk to edit

=item param

(Required) : It is a array reference with label to edit and value for it.
             eg [ [ magic, 100 ], ["software_version", "10"] ]

=back

=cut

#------------------------------------------------------------------------------

sub label_editv1_interact {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                disk => { type=>SCALAR },
                                param => { type=>ARRAYREF },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};
    my $label_to_change = $opts{label_to_change};
    my $param = $opts{param};
    my $param_count = scalar(@$param) - 1;
    my ($fc, $before, $output, $prompt, @param, $flag);

    $fc = make_conn($host, $conntype);
    $flag = 0;

    foreach my $key_index (0..$param_count) {
      #  ($before, $output, $prompt) = say($fc, "label editv1 $disk", -prompt, "Enter name of *");
     say($fc, "label editv1", -prompt, "Enter name of disk whose labels are to be edited:");
     ($before, $output, $prompt) = say($fc, "$disk", -prompt, "Enter name of *"); 
       if ( $prompt =~ /Error/) {
            say($fc, "\n", -prompt, ">");
            return ($prompt);
        }

        say($fc, "$param->[$key_index]->[0]", -prompt, "Field $param->[$key_index]->[0] is");
       	say($fc, "$param->[$key_index]->[1]", -prompt, "Enter name of v1 label field to edit:");
        say($fc, "\n", -prompt, "Enter name of disk whose labels are to be edited:");
	say($fc, "\n", -prompt, ">");
    }
    return 1;
}

#------------------------------------------------------------------------------

=head2 label_edit

label_edit(host=>$host, conntype=>$conntype, disk=>$disk, param=>[ [$label, $value], [$label1, $value1] ] )

It lets you change most (but not all) values of the new RAID (RAIDv2) labels

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Required) : Name of disk to edit 

=item param

(Required) : It is a array reference with label to edit and value for it.
             eg [ [ magic, 100 ], ["software_version", "10"] ]

=back

=cut

#------------------------------------------------------------------------------

sub label_edit { 
    my (%opts) = validate( @_, {
                                 host => { type=>OBJECT, isa=>'Hostrec' },
                                 conntype => {type=>SCALAR, default=>"#console"},
                                 disk => { type=>SCALAR },
			         label_to_change => { type=>SCALAR, default=>"all" },
                                 param => { type=>ARRAYREF },
                                }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};
    my $label_to_change = $opts{label_to_change};
    my $param = $opts{param};
    my $param_count = scalar(@$param) - 1;
    my ($fc, $output, $prompt);
	
    $fc = make_conn($host, $conntype);

    foreach my $key_index (0..$param_count) {
        if ( ($label_to_change eq 1) || ($label_to_change eq 2) ) {	
	    ( undef , undef, $output ) = say($fc, "label edit $disk!$label_to_change $param->[$key_index]->[0] $param->[$key_index]->[1]", '-prompt', "\\*>",'-wait','\n');
	     return ($output);
		
	} else {
	    (undef, undef, $output, $prompt) = say($fc, "label edit $disk $param->[$key_index]->[0] $param->[$key_index]->[1]", '-prompt', "\\*>", '-wait', '\n');

            if ($output =~ /enter label editing mode/) {
                $output = label_edit_interact(
                                              host => $host,
                                              disk => $disk,
                                              param => $param,
                                             );
                return ($output) if ($output);
                return 1;
            }
	    return ($output);
        }
    }	
}

#------------------------------------------------------------------------------

=head2 label_editv1

label_editv1(host=>$host, conntype=>$conntype, disk=>$disk, param=>[ [$label, $value], [$label1, $value1] ] )

Like label edit, but lets you edit the old RAID (RAIDv1) labels. 

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Required) : Name of disk to edit

=item param

(Required) : It is a array reference with label to edit and value for it.
             eg [ [ magic, 100 ], ["software_version", "10"] ]

=back

=cut

#------------------------------------------------------------------------------

sub label_editv1 { 
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                disk => { type=>SCALAR },
                                label_to_change => { type=>SCALAR, default=>"all" },
                                param => { type=>ARRAYREF },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};
    my $label_to_change = $opts{label_to_change};
    my $param = $opts{param};
    my $param_count = scalar(@$param) - 1;
    my ($fc, $output, $prompt);

    $fc = make_conn($host, $conntype);

    foreach my $key_index (0..$param_count) {
        if ( ($label_to_change eq 1) || ($label_to_change eq 2) ) {
            ( undef, undef ,$output )= say($fc, "label editv1 $disk!$label_to_change $param->[$key_index]->[0] $param->[$key_index]->[1]", '-prompt', "\\*>", '-wait', '\n');
            return ($output);
        } else {
                $output = label_editv1_interact(
                                                host => $host,
                                                disk => $disk,
                                                param => $param,
                                               );

                 return ($output) if ($output);
                 return 1;
        }
    }
}

#------------------------------------------------------------------------------

=head2 label_keep

label_keep(host=>$host, conntype=>$conntype)

-or-

label_keep(host=>$host, conntype=>$conntype, disk=>$disk,)

-or-

label_keep(host=>$host, conntype=>$conntype, volume=>$volume)

-or-

label_keep(host=>$host, conntype=>$conntype, fsid=>$fsid)

Erase the saved copies of the labels. 

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Optional) : Name of disk whose label gets erased

=item volume

(Optional) : Name of aggregate  or traditional volume to be created.

=item fsid

(Optional) : fsid of the disk

=back

=cut

#------------------------------------------------------------------------------

sub label_keep {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                disk => { type=>SCALAR, optional=>1 },
                                volname => { type=>SCALAR, optional=>1 },
                                fsid => { type=>SCALAR, optional=>1 },
                                }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};
    my $fsid = $opts{fsid};
    my $volname = $opts{volname};
    my ($cmd, $fc, $output);

    if ( defined($disk) ) {
        $cmd = "-d $disk";
    } elsif (defined($volname) ) {
        $cmd = "-v $volname";
    } elsif (defined($fsid) ) {
        $cmd = "-f $fsid";
    } else {
            $cmd = "all";
    }

    $fc = make_conn($host, $conntype);

    (undef, undef, $output) = say($fc, "label keep $cmd", '-prompt', "\\*> ");
    die "label save $cmd failed : $output" if ($output);

    return 1;
}

#------------------------------------------------------------------------------

=head2 label_restore

label_restore(host=>$host, conntype=>$conntype)

-or-

label_restore(host=>$host, conntype=>$conntype, disk=>$disk,)

-or-

label_restore(host=>$host, conntype=>$conntype, volume=>$volume)

-or-

label_restore(host=>$host, conntype=>$conntype, fsid=>$fsid)

Restores the saved copy of the labels, overwriting the current value of the labels

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Optional) : Name of disk whose label gets restored

=item volume

(Optional) : Name of aggregate  or traditional volume to be created.

=item fsid

(Optional) : fsid of the disk

=back

=cut

#------------------------------------------------------------------------------

sub label_restore {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                disk => { type=>SCALAR, optional=>1 },
                                volname => { type=>SCALAR, optional=>1 },
                                fsid => { type=>SCALAR, optional=>1 },
                                }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};
    my $fsid = $opts{fsid};
    my $volname = $opts{volname};
    my ($cmd, $fc, $output);

    if ( defined($disk) ) {
        $cmd = "-d $disk";
    } elsif (defined($volname) ) {
        $cmd = "-v $volname";
    } elsif (defined($fsid) ) {
        $cmd = "-f $fsid";
    } else {
        $cmd = "all";
    }

    $fc = make_conn($host, $conntype);

    (undef, undef, $output) = say($fc, "label restore $cmd", '-prompt', "\\*> ");

    die "label restore $cmd failed : $output" if ($output);

    return 1;
}

#------------------------------------------------------------------------------

=head2 label_save

label_save(host=>$host, conntype=>$conntype)

-or-

label_save(host=>$host, conntype=>$conntype, disk=>$disk,)

-or-

label_save(host=>$host, conntype=>$conntype, volume=>$volume)

-or-

label_save(host=>$host, conntype=>$conntype, fsid=>$fsid)

Stores the current value of the labels in a special location

You should always do this before running any command which modifies the labels. 

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Optional) : Name of disk whose label gets saved

=item volume

(Optional) : Name of aggregate  or traditional volume to be created.

=item fsid

(Optional) : fsid of the disk

=back

=cut

#------------------------------------------------------------------------------

sub label_save {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                disk => { type=>SCALAR, optional=>1 },
                                volname => { type=>SCALAR, optional=>1 },
                                fsid => { type=>SCALAR, optional=>1 },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};
    my $fsid = $opts{fsid};
    my $volname = $opts{volname};
    my ($cmd, $fc, $output);

    if ( defined($disk) ) {
        $cmd = "-d $disk";
    } elsif (defined($volname) ) {
        $cmd = "-v $volname";
    } elsif (defined($fsid) ) {
        $cmd = "-f $fsid";
    } else {
        $cmd = "all";
    }

    $fc = make_conn($host, $conntype);

    (undef, undef, $output) = say($fc, "label save $cmd", '-prompt', "\\*> ");

    die "label save $cmd failed : $output" if ($output);

    return 1;
}

#------------------------------------------------------------------------------

=head2 label_makeraid 

label_save(host=>$host, conntype=>$conntype, disk=>$disk,)

A 'gee-whiz' editing tool that patches older disks into an aggregate.

If you have an out-of-date disk from a volume, and it meets the requirements (it's not a spare, many fields haven't changed),

this can automatically bring back an old disk. 

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Required) : Name of disk 

=back

=cut

#------------------------------------------------------------------------------

sub label_makeraid {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                disk => { type=>SCALAR },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};

    die "Need to provide the disk for makeraid" if ( !defined($disk) );
	
    my $fc = make_conn($host, $conntype);
    my $output;

    (undef, undef, $output) = say($fc, "label makeraid $disk",'-prompt', "\\*> ");

    return ($output);
}

#------------------------------------------------------------------------------

=head2 label_makespare

label_makespare(host=>$host, conntype=>$conntype, disk=>$disk,)

Writes a SPARE disk label to a disk, turning it into a spare disk regardless of its former state. 

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Required) : Name of disk to be make spare

=back

=cut

#------------------------------------------------------------------------------

sub label_makespare {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                disk => { type=>SCALAR },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};

    die "Need to provide the disk for makespare" if ( !defined($disk) );

    my $fc = make_conn($host, $conntype);
    my $output;

    (undef, undef, $output) = say($fc, "label makespare $disk",'-prompt', "\\*> ");

    return ($output);
}

#------------------------------------------------------------------------------

=head2 label_buildtree

label_buildtree(host=>$host, conntype=>$conntype, name=>$name, checksum_type=>$checksum_type, raid_disks=>$raid_disks)

-or-

label_buildtree(host=>$host, conntype=>$conntype, name=>$name, fsid=>$fsid, checksum_type=>$checksum_type, raid_disks=>$raid_disks)

-or- 

label_buildtree(host=>$host, conntype=>$conntype, name=>$name, checksum_type=>$checksum_type, raid_type=>$raid_type, raid_disks=>$raid_disks)

-or-

label_buildtree(host=>$host, conntype=>$conntype, name=>$name, checksum_type=>$checksum_type, vol_type=>$vol_type)

-or-

label_buildtree(host=>$host, conntype=>$conntype, name=>$name, checksum_type=>$checksum_type, new_raidgroup_disks=>$new_raidgroup_disks)

-or-

label_buildtree(host=>$host, conntype=>$conntype, name=>$name, checksum_type=>$checksum_type, new_plex_disks=>$new_plex_disks)

-or-

label_buildtree(host=>$host, conntype=>$conntype, name=>$name, checksum_type=>$checksum_type, disk_size=>$disk_size)  

This command lets you create an aggregate out of whole cloth, but you have to specify everything

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item name

(Required) : Specifying the new volume

=item fsid

(Optional) : Specifying the new volumes FSID 
  
=item checksum_type 

(Required) : Specifying the checksum type with b|z|i|d|n 
	     where the letter is b for block checksums 
	     z for zoned checksums
	     i for invalid/initializing checksum data
	     d for disabled zone checksums
	     n for none 

=item raid_type 

(Optional) : Specify RAID type [ raid_dp|raid4|raid0 ]. Default is 'raid4' 
 
=item vol_type

(Optional) : Specify that the volume is an aggregate. Default is to make a trad volume  

=item new_raidgroup_disks 

(Optional) : Start new RAID group; disks after this flag are in a different RAID group than those before it

=item new_plex_disks

(Optional) : Start new plex; disks after this flag are in a different plex than those before it
  
=item disk_size

(Optional) : Set the size of the previous disk to <n> blocks
 
=back

=cut

#------------------------------------------------------------------------------

sub label_buildtree {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
		 	        name => { type=>SCALAR},
			        fsid => { type=>SCALAR, optional=>1 },
			        checksum_type => { type=>SCALAR },
			        raid_type => { type=>SCALAR, default=>"raid4", optional=>1 },		
                                raid_disks => { type=>SCALAR },
                                vol_type => { type=>SCALAR, default=>"trad", optional=>1 },
                                new_raidgroup_disks => { type=>SCALAR, optional=>1 },
                                new_plex_disks => { type=>SCALAR,  optional=>1 },
                                disk_size => { type=>SCALAR, optional=>1 },
                               }  
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $name = $opts{name};
    my $fsid = $opts{fsid};
    my $checksum_type = $opts{checksum_type};
    my $raid_type = $opts{raid_type};
    my $raid_disks = $opts{raid_disks};
    my $vol_type = $opts{vol_type};
    my $new_raidgroup_disks = $opts{new_raidgroup_disks};
    my $new_plex_disks = $opts{new_plex_disks};
    my $disk_size = $opts{disk_size};	
    my ($cmd, $output);

    my $fc = make_conn($host, $conntype);

    $cmd = "";
    $cmd = $cmd . " -n $name" if ( defined($name) ); 
    $cmd = $cmd . " -f $fsid" if ( defined($fsid) );
    $cmd = $cmd . " -c $checksum_type" if ( defined($checksum_type) );
    $cmd = $cmd . " -t $raid_type" if ( defined($raid_type) );
    $cmd = $cmd . " $raid_disks" if ( defined($raid_disks) );

    if ( $vol_type =~ /flex/ ) { 
        $cmd = $cmd . " -a";
    } elsif ( $vol_type =~ /trad/ ) { 
        $cmd = $cmd;
    } else {
        die "vol_type should be either flex or trad";
    }

    $cmd = $cmd . " -r $new_raidgroup_disks" if ( defined($new_raidgroup_disks) );
    $cmd = $cmd . " -p $new_plex_disks" if ( defined($new_plex_disks) );
    $cmd = $cmd . " -s $disk_size" if ( defined($disk_size) );
 
    (undef, undef, $output) = say($fc, "label buildtree $cmd",'-prompt', "\\*> ");

    #die "label buildtree $cmd failed : $output" if ($output);

    return $output;

}

#------------------------------------------------------------------------------

=head2 label_wipe

label_wipe(host=>$host, conntype=>$conntype, disk=>$disk,)

Erases the RAID Area of a disk.

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Required) : Name of disk to be wiped 

=back

=cut

#------------------------------------------------------------------------------

sub label_wipe {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                disk => { type=>SCALAR },
                               }
                          );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};

    die "Need to provide the disk for wipe" if ( !defined($disk) );

    my $fc = make_conn($host, $conntype);
    my $output;

    (undef, undef, $output) = say($fc, "label wipe $disk",'-prompt', "\\*> ");

  #  die "label wipe $disk failed : $output" if ($output);

    return $output;
}

#------------------------------------------------------------------------------

=head2 label_wipev1

label_wipev1(host=>$host, conntype=>$conntype, disk=>$disk,)

Erases the old RAID (RAIDv1) labels of a disk

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk

(Required) : Name of disk to be wiped

=back

=cut

#------------------------------------------------------------------------------

sub label_wipev1 {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                disk => { type=>SCALAR },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk = $opts{disk};

    die "Need to provide the disk for wipev1" if ( !defined($disk) );

    my $fc = make_conn($host, $conntype);
    my $output;

    (undef, undef, $output) = say($fc, "label wipev1 $disk",'-prompt', "\\*> ");

    return ($output);
}

#------------------------------------------------------------------------------

=head2 label_copy

label_copy(host=>$host, conntype=>$conntype, disk1=>$disk, disk2=>$disk)

Copy a disk label from one disk to another.

An optional argument will change the RGDN of the label so that the destination disk 

doesn't appear to be in the same spot as the source disk during assimilation 

The following arguments are accepted:

=over

=item host

(Required): Hostspec to filer

=item conntype

(Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.

=item disk1

(Required) : Name of disk1

=item disk2

(Required) : Name of disk2

=back

=cut

#------------------------------------------------------------------------------

sub label_copy {
    my (%opts) = validate( @_, {
                                host => { type=>OBJECT, isa=>'Hostrec' },
                                conntype => {type=>SCALAR, default=>"#console"},
                                disk1 => { type=>SCALAR },
                                disk2 => { type=>SCALAR },
                                id => { type=>SCALAR, optional=>1 },
                               }
                         );

    my $host = $opts{host};
    my $conntype = $opts{conntype};
    my $disk1 = $opts{disk1};
    my $disk2 = $opts{disk2};
    my $id = $opts{id};

    my $fc = make_conn($host, $conntype);
    my $output;
	
    if (!$id) {
        (undef, undef, $output) = say($fc, "label copy $disk1 $disk2", '-prompt', "\\*> ");
    } else {
        (undef, undef, $output) = say($fc, "label copy $disk1 $disk2 -i $id", '-prompt', "\\*> ");
    }

    return $output;
}

1;
