# -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/Raid/RaidCmds.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @summary This library provides RAID utility functions
## that could be used in RAID subsystem testing such as:
## hammer, mount, unmount, and others. 
## 
## @description 
##  The library defines common routines to be use for RAID
##  testing.
##  
## @usage
##  Here is the list of the defined modules and functions which
##  can be used for both filers and simulators.
##
## @author tomdoan@netapp.com, dl-step-internal@netapp.com
##
## @status testing
 
=head1 NAME

RaidCmds - Raid common commands library module

=head1 SYNOPSIS

    use RaidCmds qw(RaidCmds_volaggr_remove_all)

    my $host = host($FILER); 
    my $conn = connect("#rsh"); 

    my $ret = RaidCmds_volaggr_remove_all($conn); 

=head1 DESCRIPTION 
 
The library provides routines to manipulate RAID subsystem via
the RAID CLI commands such as vol/aggr, plex, disk, scrub, and
others. 

=head1 ROUTINES

 Function prototypes: 

 RaidCmds_volaggr_create - create a vol/aggr with size, n disks
 RaidCmds_volaggr_exists - check if the vol/aggr exists
 RaidCmds_volaggr_remove_all - removes all vols except root
 RaidCmds_volaggr_verify_disks_rgdn - verify disks rgdn are on order
  
 RaidCmds_aggrvol_create - create a tradvol or aggr with given options, if aggr created will also create a flexvol that will use most of the space. -Htu

=over

=cut 

package Raid::RaidCmds; 

use strict;
use warnings;

use Carp qw(confess);
use Params::Validate qw(validate validate_pos SCALAR ARRAYREF HASHREF UNDEF OBJECT);
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}
use RaidInfo; 
use Raid::RaidUtils qw(RaidUtils_hammer_abort
		       RaidUtils_debug
		       RaidUtils_convert_to_mb); 
use Tharn qw(say tstlog);
use Utils qw(make_conn_and_complain make_conn);  
use Module::Common qw(validate_mixed); 

use Volume qw(vol_create vol_exists vol_create_flex vol_destroy vol_offline vol_status);  
use Aggregate qw(aggr_create aggr_exists aggr_destroy aggr_offline); 

use vars qw (@ISA @EXPORT_OK); 

require Exporter; 
@ISA = qw(Exporter); 
@EXPORT_OK = qw(RaidCmds_test
		RaidCmds_volaggr_remove_all
		RaidCmds_volaggr_create
		RaidCmds_volaggr_exists
		RaidCmds_volaggr_verify_disks_rgdn

		RaidCmds_aggrvol_create
		);

# ----------------
# define constants
# ----------------
use constant SAY_TIMEOUT => 240; 
use constant PASS => 1; 
use constant FAIL => 0; 
use constant NOOP => 2; 

=head1 RaidCmds_volaggr_exists(conn=>$conn,name=>$name,objtype=>'vol/aggr') 

 The routine verify if the specify vol/aggr exists.
 Return 1 if exists or 0 if it is not. 

=over

=item conn

  Required: filer connection 

=cut  

=item name

  name: name of the vol/aggr
=cut 

=item objtype
 
  objtype =>'vol'or  objtype =>'aggr'

=cut  

sub RaidCmds_volaggr_exists($$%) {
    RaidUtils_debug("RaidCmds_volaggr_exists: ..."); 

    my $conn = make_conn_and_complain(shift); 
    my ($objname, %opts) = 
	validate_mixed(
		       @_,
		       [ {type => SCALAR} ],
		       { 
			   objtype => { type => SCALAR, optional => 1, 
					regex => qr/vol|aggr/ , default => 'vol'},
		       }
		       ); 
    
    if (aggr_exists($conn, $objname)) {
	RaidUtils_debug("aggr $objname exists"); 
	return PASS; 
    }

    if (vol_exists($conn, $objname)) {
	RaidUtils_debug("vol $objname exists"); 
	return PASS; 
    }

    RaidUtils_debug("RaidCmds_volaggr_exists: object <$objname> does not exist"); 
    return FAIL; 
} #RaidCmds_volaggr_exists

=head1 RaidCmds_volaggr_remove($conn, $objname) 

 The routine removes the specified object regardless
 if the specified object is vol or aggr. Since the 
 object could be of type 'vol' then in the next run
 it could be of type 'aggr'. Therefore, the routine
 will not attempt to double check if objectname is
 vol or aggr. 

 It tries to stop all hammer activities before 

=over

=item conn

  Required: filer connection 

=cut  

sub RaidCmds_volaggr_remove($$) {
    RaidUtils_debug("RaidCmds_volaggr_remove: ..."); 

    my $conn = make_conn_and_complain(shift); 
    my $objname = $_[0]; 

    RaidUtils_hammer_abort($conn); 

    # ------------------------------
    # remove the specified aggregate
    # ------------------------------
    if (aggr_offline($conn, $objname)) {
	if (!aggr_destroy($conn, $objname)) {
	    tstlog(-comment, "Unable to remove aggr <$objname>"); 
	    return FAIL;
	}
	tstlog(-comment, "Successfully remove aggregate <$objname>"); 
	return PASS; 
    }

    # ------------------------------
    # remove the specified volume
    # ------------------------------
    if (!vol_offline($conn, $objname)) {
	if (!vol_destroy($conn, $objname)) {
	    tstlog(-comment, "Unable to remove aggr <$objname>"); 
	    return FAIL;
	}
	tstlog(-comment, "Successfully remove volume <$objname>"); 
	return PASS; 
    }
    tstlog(-fail, "Unable to remove the specified vol/aggr <$objname>"); 
} #RaidCmds_volaggr_remove

=head1 RaidCmds_volaggr_remove_all($conn) 

 The routine removes all aggr (but not the root volume). 
 It tries to stop all hammer activities. 

=over

=item conn

  Required: filer connection 

=cut  

sub RaidCmds_volaggr_remove_all($) {
    my $conn = make_conn_and_complain(shift); 

    my $ro = new RaidInfo($conn); 

    # ----------------
    # stop all hammers
    # ----------------
    RaidUtils_hammer_abort($conn); 

    # -----------------------
    # remove all test volumes
    # -----------------------
    my %aggrs = $ro->all_aggregates();
    my $vol0 = $ro->root_volume(); 

    foreach my $aggr (keys %aggrs) {
	if ($vol0 eq $aggr) {
	    next; 
	}
	RaidUtils_debug("Removing aggr <$aggr>"); 
	if (!aggr_offline($conn, $aggr)) {
	    tstlog(-comment, "Unable to offline aggr <$aggr> or the volume already offline"); 
	    #return FAIL;
	}
	aggr_destroy($conn, $aggr); 
    } #each vol

    return PASS; 
} #RaidCmds_volaggr_remove_all

=head1 RaidCmds_volaggr_create(conn=>$conn, name=>$name, %opts) 

 The routine creates the specified vol/aggr with the specified
 name based on the specified options of object type, size, 
 raidtype, and number of disks. 

 Return FAIL(0) if the specified name could not be created. 
 
=over

=item conn

  Required: filer connection 

=cut 

=item name

  Required: name of the obj (vol or aggr)

=cut 

=item opts

    objtype  => 'vol|aggr' - vol or aggregate (default is vol)
    size      => 'nn[k|m|g|t]' - storage must have a unit [kmgt]
    raidtype => 'raid4|raid_dp' - raidtype (default is raid4)
    num_disks => 'n' - specify the number of disks

=cut 
 
sub RaidCmds_volaggr_create($$%) {
  RaidUtils_debug("Cmds_volaggr_create: ..."); 

  my $conn = make_conn_and_complain(shift); 
  my ($objname, %opts) = 
      validate_mixed(
		   @_,
		   [ {type => SCALAR} ],
		   { 
		    objtype => { type => SCALAR, optional => 1, 
				  regex => qr/vol|aggr/ , default => 'vol'},
		    size => { type => SCALAR, optional => 0, 
			      regex => qr/\d+[kmgt ]/ }, 
		    raidtype => { optional => 0, regex => qr/raid4|raid_dp/, default => 'raid4' },
		    num_disks => { optional => 1, regex => qr/\d+/ }, 
		   }
		  ); 
  
  my $ro = new RaidInfo($conn); 
  my %spare_disks = $ro->spare_disks(); 
  my $total_size = 0; 
  my @spares = ();
 
  # -----------------------------------------
  # remove if the vol/aggr is already existed
  # whether ot not it is aggr or volume
  # -----------------------------------------
  if (RaidCmds_volaggr_exists($conn, $objname)) {
      RaidUtils_hammer_abort($conn); 
      RaidCmds_volaggr_remove($conn, $objname); 
  } 
  
  # -------------------------------------------------------
  # get a list of spares of total size > the specified size
  # and we make sure to have more than 1 disk (this is for
  # testing under mini disks). 
  # -------------------------------------------------------
  if (defined($opts{size})) {
      my $size = $opts{size}; 
      
      # convert to the underline MB unit
      $size = RaidUtils_convert_to_mb($size); 
      print("size converted <$size>"); 
      
      foreach my $disk (keys %spare_disks) {
	  my $do = $ro->disk($disk); 
	  my $dsz = $do->parameter("avl_mb"); 
	  
	  $total_size = $total_size + $dsz;
	  RaidUtils_debug("$disk total_size = <$total_size>");       
	  push (@spares, $disk); 
	  if ($total_size > $size && $#spares > 1) {
	      RaidUtils_debug("calling last with $disk total_size = <$total_size>");       
	      last; 
	  }
      } #each spare 
      
      if ($total_size < $size) {
	  tstlog(-comment, "RaidCmds_volaggr_create: insufficient storage to create a filesystem <$objname> with size <$size> from a total spare of <$total_size>"); 
	  return FAIL; 
    }
      
  } #define size

  tstlog(-comment, "RaidCmds_volaggr_create: disk storage to create a vol <$objname> with size <$total_size>"); 
  
  RaidUtils_debug("RaidCmds_volaggr_create: spares= " . scalar(@spares) . "\n"); 
  
  # ----------------------------------------
  # creating a test vol with a specific type
  # ----------------------------------------
  my $timeout = 4 * 60 * 60; # 4 hours
  if (($opts{objtype} eq "vol")) {
      if (!vol_create($conn, $objname, \@spares, timeout=>$timeout, raidtype=>$opts{raidtype})) {
	  tstlog(-comment, "Unable to create the test volume <$objname> in <$timeout>"); 
	  return FAIL; 
      }
      tstlog(-comment, "Successfully created volume <$objname>"); 
      return PASS; 
  } # vol create
  
  # --------------------
  # creating a test aggr 
  # --------------------
  if (!aggr_create($conn, $objname, \@spares, timeout=>$timeout, raidtype=>$opts{raidtype})) {
      tstlog(-comment, "Unable to create the test volume <$objname> in <$timeout>"); 
      return FAIL; 
  } # aggr create
  
  tstlog(-comment, "Successfully created aggr <$objname>");
  return PASS; 
} #RaidCmds_volaggr_create

=head1 RaidCmds_volaggr_verify_disks_rgdn(conn=>$conn, objname=>$objname, options=>$opts) 

 The routine invokes "vol status -r" to retrieve the list
 of raigroups and disks. The routine checks to make
 sure the list of disks from the UI should match the
 order of disk_rgdn from raid_config. 
 
 Return 1 if all raigroup disks are in the right order with
 the disk_rgdn number. 

 Return 0 otherwise. 

=over

=item conn

  Required: filer connection 

=cut 

=item vol 

  Required: vol/aggr name

=cut 
 
sub RaidCmds_volaggr_verify_disks_rgdn($$;%) {
    RaidUtils_debug("RaidCmds_volaggr_verify_disks_rgdn:..."); 
    my $conn = make_conn_and_complain(shift);   
    my ($objname, %opts) = 
	validate_mixed
	    (@_,
	     [ {type => SCALAR} ],
	     { 
		 objtype => { type => SCALAR, optional => 1, 
			      regex => qr/vol|aggr/ , default => 'vol'}
	     }
	     ); 
    
    my $ro = new RaidInfo($conn);
    my $cmd = $opts{objtype}; 

    # ------------------------------------------------- 
    # The test requires to retrieve UI output. 
    # From the output the test tries to match the order
    # of disk against the disk_rgdn number. 
    # -------------------------------------------------
    my $output = _saycmd($conn, "$cmd status -r $objname"); 
    my $num_disks = 0; 
    my $disk_count = 0; # count the number of disk in each raidgroup
 
    foreach my $line (split(/\n/, $output)) {
	my @rg_list = (); 
	# ----------------------------------------
	# searching through the output buffer and 
	# look for the dash lines (--------). This
	# signals a new raidgroup is started. 
	# After the dash lines, the list of disk should
	# be followed:
	#
	#   RAID Disk Device  HA  SHELF BAY CHAN 
	# --------- ------  ------------- ---- ----
	# dparity   3a.75   3a    4   11  FC:B   - 
	# parity    3b.121  3b    7   9   FC:B   - 
	# data      3b.123  3b    7   11  FC:B   - 
	#
	# ----------------------------------------
	if ($line =~ /-------/) {
	    $disk_count = 0; #reset the disk_count 
	    print("\n"); 
	    next; 
	}
	
	# skip blank lines and others
	if ($line =~ /RAID/ || $line =~ /^(\s)*$/ || $line =~ /Volume/ ||
	    $line =~ /Aggregate/ || $line =~ /Plex/) {
	    next; 
	}

	# ---------------------------------------------
	# get each disk and verify whether or not each 
	# one is on the right order list
	# ---------------------------------------------
	if ($line =~ /parity/ || $line =~ /data/) {
	    my @disks = split(/\s+/, $line); 
	    my $do = $ro->disk($disks[2]);
	    my $nth = $do->parameter("disk_rgdn"); 

	    if ($nth == $disk_count) {
		print("<$disks[2]> is at <$disk_count> position which matches the rgdn: <$nth>\n"); 
	    } else {
		tstlog(-fatal, "<$disks[2]> is at <$disk_count> position which does not match with teh rgdn: <$nth>\n"); 
	    }

	    $disk_count++; 
	} #each input line
    }
} #RaidCmds_volaggr_match_disks_rgdn

###########Htu

=head2 RaidCmds_aggrvol_create

    RaidCmds_aggrvol_create($conn,$option,$filer_vol,$aggr,$timeout)

The Routine create a aggr or tradvol depend on $aggr value (1 for aggr, 0 for 
tradvol) with give $option.
=over

=item Arguments

=over

=item conn
(Required) : the filer connection object

=item filer_vol
(Required) : name of the aggr or tradvol

=item option
(Required) : options for vol/aggr create

=item timeout
(Optional) : timeout value for say cmd

=back
=for html <hr>

=cut

sub RaidCmds_aggrvol_create {
        my (%opts) = validate(@_, {
			conn => {type=>HASHREF},
                        filer_vol => {type=>SCALAR},
                        option => {type=>SCALAR},
                        aggr => {type=>SCALAR},
			timeout => {type=>SCALAR, default=>"300"},
                        });

	my $conn = $opts{conn};
        my $filer_vol = $opts{filer_vol};
        my $option = $opts{option};
        my $cmd = ($opts{aggr})? "aggr" : "vol";
	my $timeout = $opts{timeout};

        tstlog ("-comment", "CREATING $cmd $filer_vol ....\n");
        my $output=say($conn, "priv set -q diag; $cmd create $filer_vol $option", -timeout, $timeout);
        sleep 30;

        my $raid_obj=new RaidInfo($conn);
        my $aggr_obj =$raid_obj->aggregate("$filer_vol");

        if (!$aggr_obj) {
               tstlog ("-comment", "CREATION $cmd $filer_vol FAILED.\n");
               return 0;
        }
        sleep 30;

        RaidCmds_wait_on_aggr_disks_zeroing(conn=>$conn, filer_vol=>$filer_vol);

        # Create flex volume having same name of aggregate with 3/4th of its size.

        if($cmd eq "aggr") {
                my $size = 3072 * $aggr_obj->parameter("raidsize");
                my $o = say($conn, "vol create $filer_vol $filer_vol $size",-timeout, $timeout);
        }
        return 1
}   ###RaidCmds_aggrvol_create


sub RaidCmds_wait_on_aggr_disks_zeroing {
        my (%opts) = validate(@_, {
                                conn => {type=>HASHREF},
                                filer_vol => { type=>SCALAR },
                                }
                        );
        my $conn = $opts{conn};
        my $filer_vol = $opts{filer_vol};

        my $raid_obj=new RaidInfo($conn);
        my $aggr_obj =$raid_obj->aggregate("$filer_vol");
        my %disklist=$aggr_obj->all_disks();
        my @disklist=keys(%disklist);
        my $count = 1;
        my $disk;
        while ($count > 0) {
                $raid_obj->refresh();
                $aggr_obj =$raid_obj->aggregate("$filer_vol");
                $count = 0;
                tstlog("-comment", "disklist in test volume/aggr is: @disklist ");
                foreach $disk (@disklist) {
                        my $disk_obj=$aggr_obj->disk("$disk");
                        if ($disk_obj->parameter("fsm_state") eq "ZEROING" || $disk_obj->parameter("fsm_state") eq "NOTZEROED") {
                                tstlog( "-comment","disk $disk in test volume/aggr is not zeroed yet ");
                                $count++;
                        }
                }
                tstlog("-comment", "waiting for $count disks to zero");
                sleep (4*$count);
        }
}   ###RaidCmds_wait_on_aggr_disks_zeroing

    
sub RaidCmds_test() {
    print ("Raidcmds_test\n"); 
}

# -------------------------------------------------
# define the say routine to be
# used only by all defined functions in the library
# -------------------------------------------------
sub _saycmd($$) {
    my ($conn, $msg) = @_; 

    my (undef, undef, $output) = say($conn, "priv set -q diag; $msg ", -timeout, SAY_TIMEOUT); 
    return $output; 
} #_saycmd

1; 

