# -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/Raid/RaidUtils.pm#1 $
#
# Copyright (c) 2005 Network Appliance, Inc.
# All rights reserved.
#
## @summary This library provides RAID utility functions
## that could be used in RAID subsystem testing such as:
## hammer, mount, unmount, and others. 
## 
## @description 
##  The library defines common routines to be use for RAID
##  testing.
## @status testing
## @author tomdoan@netapp.com, dl-step-internal@netapp.com
##

=head1 RaidUtils

The library provides routines to manipulate Raid subsystem
testing. 

=head1 SYSNOPSIS

use Raid::RaidUtils qw( RaidUtils_func_names ...); 

=head1 DESCRIPTION

This library provides RAID utility functions
that could be used in RAID subsystem testing such as:
hammer, mount, unmount, and others. 

=cut

package Raid::RaidUtils; 

use strict;
use warnings; 

use Carp qw(confess);
use Params::Validate qw(validate validate_pos SCALAR ARRAYREF HASHREF UNDEF OBJECT);
BEGIN {
Params::Validate::validation_options(
    on_fail => sub {my @msgs = @_; chomp(@msgs);
        Carp::confess(Tharn::script_result(@msgs))}
   );
}

use RaidInfo; 
use Tharn;
use vars qw (@ISA @EXPORT_OK $VERBOSE); 

require Exporter; 
@ISA = qw(Exporter); 
@EXPORT_OK = qw(
		RaidUtils_cleanup
		RaidUtils_convert_to_mb
		RaidUtils_debug
		RaidUtils_fiji_ata_inject_hw
		RaidUtils_fiji_ata_inject_sim
		RaidUtils_get_total_disks_size
		RaidUtils_get_total_spares_size
		RaidUtils_hammer_abort
		RaidUtils_hammer_pause
		RaidUtils_hammer_restart
		RaidUtils_hammer_start
		RaidUtils_hammer_start_multiple
		RaidUtils_init
		RaidUtils_is_sim
		RaidUtils_root_mount
		RaidUtils_root_umount
		RaidUtils_sim_pull
		RaidUtils_sim_pull_all
		RaidUtils_sim_push
		RaidUtils_sim_push_all
		RaidUtils_testme_driver
		
		RaidUtils_get_group_list   
		RaidUtils_get_disklist_by_group_type
	     );

# ----------------
# define constants
# ----------------
use constant SAY_TIMEOUT => 240; 
use constant TRUE => 1; 
use constant FALSE => 0; 

=head2 Raidutils_Cleanup

    Raidutils_Cleanup($Conn,$Disk);

The Routine Cleans Up The Test Environment Such As Stop
All Hammers, Remove test volume, and unfail all failed
disks.

=over

=item Arguments

=over

=item conn

(Required) : the filer connection object

=item disk

(Required) : name of the specified disk to be injected.

=back

=for html <hr>

=cut

sub RaidUtils_cleanup($) {
    my ($conn) = @_;
    
    RaidUtils_hammer_restart($conn); 
    RaidUtils_hammer_abort($conn); 

} #RaidUtils_cleanup


=head2 RaidUtils_convert_to_mb

    RaidUtils_convert_to_mb($size);

The routine converts the specified storage unit from 
(t - terrabytes, g - gigabytes, m - megabytes, and k -kbytes)
to Mbytes. 
The routine returns a converted value in Mbytes. 

=over

=item Arguments

=over

=item size

(Required) : The parameter should have one of the following format: 
    - nt, ng, nm, or nk

=back

=cut

sub RaidUtils_convert_to_mb($) {
   my ($size) = $_[0]; 
   
   RaidUtils_debug(" RaidUtils_convert_to_mb:..."); 
    
    if ($size =~ /t/) {
	my $num = substr($size, 0, length($size) -1); 
	$num = $num * 1000000; 
	return $num; 
    }

    if ($size =~ /g/) {
	my $num = substr($size, 0, length($size) -1);
	$num = $num * 100; 
	return $num; 
    }

    if ($size =~ /m/) {
	my $num = substr($size, 0, length($size) - 1); 
	return $num; 
    }
    
    if ($size =~ /k/) {
	my $num = substr($size, 0, length($size) -1 ); 
	return ($num / 1000); 
    }

    RaidUtils_debug("Unable to convert <$size> "); 
    return 0; 
} #RaidUtils_convert_to_mb


=head2 RaidUtils_fiji_ata_inject_hw

    RaidUtils_fiji_ata_inject_hw($conn, $disk);

The routine injects fiji points to the specified disk. 
This is not for simulator. 

=over

=item Arguments

=over

=item conn

(Required) : the filer connection object

=item disk

(Required) :  name of the specified disk to be injected. 

=back

=for html <hr>

=cut

sub RaidUtils_fiji_ata_inject_hw($$$) {
    my ($conn, $disk, %opts) = @_;
    
    my $fiji_types; 
    my $fiji_type_hw = $opts{fiji_hw}; 

    RaidUtils_fiji_clear_all();
    for my $ec (split(' ', $fiji_type_hw)) {
	_debug("Injecting fiji error code <$ec>", 2); 
	my $output = RaidUtils_say($conn, "priv set -q test; scsi set_error -e $ec -l 1 $disk"); 
	$output = RaidUtils_say($conn, "priv set -q test; scsi force_error -e $disk"); 
    } #each fiji type

} #RaidUtils_fiji_ata_inject_hw


=head2 RaidUtils_fiji_ata_inject_sim

    RaidUtils_fiji_ata_inject_sim($conn, $disk);

Injects fiji points to the specified disk under simulator.

=over

=item Arguments

=over 

=item conn

(Required) :  filer connection object

=item disk

(Required) :  name of the specified disk to be injected. 

=back

=cut

sub RaidUtils_fiji_ata_inject_sim($$) {
    my ($conn, $disk, %opts) = @_;

    my $fiji_sim = $opts{fiji_sim}; 

    #RaidUtils_fiji_clear_all();
    foreach my $fiji (split(' ', $fiji_sim)) {

	RaidUtils_debug(-comment, "ata_fiji_inject_sim: inject point <$fiji>"); 
	
        if (uc($fiji) eq "DISK.IO.AGRSV.TIMEOUT.FAILURE") {
            RaidUtils_fiji_add_filter("(disk.io.agrsv.timeout.failure (eq \@disk $disk) return 1)");
            next;
        }
	
        if (uc($fiji) eq "RAID.DISK.READ.DONE") {
            RaidUtils_fiji_add_filter("(raid.disk.read.done (eq \@disk $disk)
 return buf_Errno E_DISK_AGRSV_TIMEOUT)");
            next;
        }

        if (uc($fiji) eq "RAID.LABEL.WRITE.DONE") {
            RaidUtils_fiji_add_filter("(raid.label.write.done (eq \@disk $disk) return buf_Errno E_DISK_AGRSV_TIMEOUT))");
            next;
        }

    }#foreach

} #RaidUtils_fiji_ata_inject_sim


=head2 RaidUtils_init

    RaidUtils_init($conn);

 Initialize the test environment covers multiple steps:
    1. start zeroing all spares
    2. pull all simulator disks
    3. stop all hammer activies
    4. 

=over

=item Arguments

=over

=item conn

(Required) :  specified the filer connection object

=back

=cut

sub RaidUtils_init($) {
    my $conn = $_[0]; 

    # 1. start zeroing all spares
    my $output = RaidUtils_say($conn, "spare_zero"); 

    # 2. pulling all simulator disks
    RaidUtils_disk_simpush_all(); 

    # 3. abort all hammers 
    RaidUtils_hammer_abort();
 
    # 4. disk online all
    RaidCmds_disk_online_all(); 

    # 5. clear all fiji points
    RaidCmds_fiji_clear_all(); 

}#RaidUtils_init


=head2 RaidUtils_is_sim

    RaidUtils_is_sim($conn); 

Returns 1 if the specified system is a simulator

=over

=item Arguments

=over

=item conn 

(Required) :  specified the filer connection object

=back

=cut

sub RaidUtils_is_sim($) {
    my $conn = $_[0]; 
    my $output = RaidUtils_say($conn, "version -v"); 

    if ($output=~ "<td>" || $output=~ "<Ld>" || $output=~ "<od>" || $output=~ "<Rd>" || $output=~ "<Pd>" || $output=~ "<tdc>"){
	return TRUE;
    }
    return FALSE;
} #RaidUtils_is_sim


=head2 RaidUtils_get_current_index

    RaidUtils_get_current_index($conn, $filername, $filename);

 Gets the current index of the specified file
 that is part of the root volume of the specified filer. 

=over

=item Arguments

=over

=item $conn

(Required) :  specifies the filer connection object

=item filername

(Required) :  specifies the name of the filer

=item filename

(Required) :  one of the filenames that resides on the root volume
  of the specified filer. 

=item Returns

  The current index of the specified filename

=back

=cut

sub RaidUtils_get_current_index($$$) {
    my ($conn, $filername, $filename) = @_; 
    
    RaidUtils_root_mount($conn);
    
    open(FH, $filername) || die "Unable to open the EMS message file <$filername>"; 
    
    # --------------------------------
    # get the current index of the EMS 
    # --------------------------------
    seek(FH, 0, 2);
    my $index = tell(FH); 
    close(FH);
    RaidUtils_root_umount($conn); 

    RaidUtils_debug("The current index of EMS = <$index>\n"); 
    return $index; 
} #RaidUtils_get_current_index


=head2 RaidUtils_get_total_disks_size

    RaidUtils_get_total_disks_size($conn);

The routine computes the total disk space available on the 
specified filer.

=over

=item Arguments

=over

=item conn

(Required) :  filer connection object

=item Returns

  Total disk space of the specified filer

=back

=cut

sub RaidUtils_get_total_disks_size($) {
    my $conn = $_[0]; 

    RaidUtils_debug("Staring RaidUtils_get_total_disks_size"); 
    my $total_disks_size = 0; 
    my $ro = new RaidInfo($conn); 
    my %disks = $ro->all_disks(); 

    # traverse through each disk and add them together
    # 
    foreach my $disk (keys %disks) {
	my $do = $ro->disk($disk); 
	my $dsz = $do->parameter("avl_mb"); 
	RaidUtils_debug("calculating disk <$disk> $dsz $total_disks_size \n"); 

	$total_disks_size = $total_disks_size + $dsz; 
    }
    return ($total_disks_size);
} #RaidUtils_get_total_disks_size


=head2 RaidUtils_get_total_spares_size

    RaidUtils_get_total_spares_size($conn); 
The routine computes the total disk storage of all spares. 

=over

=item Arguments

=over

=item conn

(Required) :  filer connection object

=back

=item Returns

The total disk storage of all spares in kbytes. 

=cut

sub RaidUtils_get_total_spares_size($) {
    my $conn = $_[0]; 

    RaidUtils_debug("Starting RaidUtils_get_total_spares_size"); 
    my $total_spares_size = 0; 
    my $ro = new RaidInfo($conn); 
    my %disks = $ro->spare_disks(); 

    # --------------------------------------------------
    # compute the total spare disks size 
    # --------------------------------------------------
    foreach my $disk (keys %disks) {
	my $do = $ro->disk($disk); 
	my $dsz = $do->parameter("used_sz"); 
	$total_spares_size = $total_spares_size + $dsz; 
    }
    return ($total_spares_size);
} #RaidUtils_get_total_spares_size


=head2 RaidUtils_hammer_abort

    RaidUtils_hammer_abort($conn);

The routine aborts all hammer activities. It waits until
the hammer is done. 

=over

=item Arguments

=over

=item fc

(Required) :  filer connection object

=back

=cut

sub RaidUtils_hammer_abort($) {
    my $conn = $_[0]; 
    my %opts; 
    $opts{blind} = "yes"; 

    # restart all hammers
    RaidUtils_hammer_restart($conn); 

    # abort hammers
    RaidUtils_say($conn, "hammer abort", \%opts); 

    my $output = RaidUtils_say($conn, "hammer status", \%opts); 
    while ($output ne "") {
	sleep(5); 
	$output = RaidUtils_say($conn, "hammer status", \%opts); 
	print "output = <$output>\n"; 
    }
    return TRUE; 
} #RaidUtils_hammer_abort


=head2 RaidUtils_hammer_pause

    RaidUtils_hammer_pause($conn); 

 The routine pauses all hammer activities and waits
 until there is no more activity. 

=over

=item Arguments

=over

=item fc

(Required) :  filer connection object

=back

=cut

sub RaidUtils_hammer_pause($) {
    my $conn = $_[0]; 

    my $output = RaidUtils_say($conn, "hammer pause"); 

    $output = RaidUtils_say($conn, "hammer status"); 
    while ($output !~ "") {
	sleep(5); 
	$output = RaidUtils_say($conn, "hammer status"); 
    }

} #RaidUtils_hammer_pause


=head2 RaidUtils_hammer_restart

    RaidUtils_hammer_restart($conn);

The routine resumes all hammer activities. 
It wait untils hammer is online. 

=over

=item Arguments

=over

=item fc

(Required) :  filer connection object

=back

=cut

sub RaidUtils_hammer_restart($) {
    my $conn = $_[0]; 
    my %opts = qw/blind yes/; 

    RaidUtils_say($conn, "hammer restart", \%opts); 
    return TRUE; 
} #RaidUtils_hammer_restart


=head2 RaidUtils_root_mount

    RaidUtils_root_mount($conn, $filername);

Mounts the root volume of the specified filer 

=over

=item Arguments

=over

=item fc 

(Required) :  filer connection object

=item filername

(Required) :  specifies the filername 

=back

=cut

sub RaidUtils_root_mount($$)
{
    my ($conn, $filername) = @_;
    
} #RaidUtils_mount_root_vol


=head2 RaidUtils_root_umount

    RaidUtils_root_umount($filername);

Unmount the root volume on the CLIENT

=over

=item Arguments

=over

=item filername

(Required) :  filer name

=back

=cut

sub RaidUtils_root_umount($)
{
    my ($filer_name) = $_[0];

    # -----------------------    
    # umount the root volume
    # -----------------------
} #RaidUtils_umount_root_vol


=head2 RaidUtils_sim_pull

    RaidUtils_sim_pull(conn=>$conn, disk=>$disk);

pulls a specified disk. 

=over

=item Arguments

=over

=item conn 

(Required) :  specified the connection object

=item disk

(Required) :  name of a disk

=back

=cut

sub RaidUtils_sim_pull($$) {
    my ($conn, $disk) = @_; 

    my $output = RaidUtils_say($conn, "disk simpull $disk"); 
    if ($output eq "") {
	return TRUE; 
    }
    return FALSE;
}#RaidUtils_sim_pull

=head2 RaidUtils_sim_push

    RaidUtils_sim_push(conn=>$conn, disk=>$disk);

Routine pushs a specified disk. For simulator only. 

=over

=item Arguments 

=over

=item conn 

(Required) :  specified the filer connection object

=item disk 

(Required) :  name of a disk

=back

=cut

sub RaidUtils_sim_push($$) {
    my ($conn, $disk) = @_; 
    
    my $ro = new RaidInfo($conn);
    my $do = $ro->disk($disk); 
    my $dsz = $do->parameter("avl_blks"); 
    
    my @sel_disk = (); 
    my $disk_alb = ":" . $dsz; 
    my @buf = split(/\./, $disk);  
    my $disk_name = $buf[1] . ":"; 

    my(@list) = glob(",disks/,pulled/*:*");
    my (@name, @names);
    foreach my $e (@list) {
        if ($e =~ /$disk_name/ && $e =~ /$disk_alb/) {
            my @sel_disk = split("/", $e); 
            RaidUtils_debug("found a match $sel_disk[2]", 2);
            last; 
        }
    } # each disk filename

    my ($full_disk_name) = $sel_disk[2]; 
    
    my $output = RaidUtils_say($conn, "disk simpush $full_disk_name"); 
    
    if ($output eq "") {
	return TRUE; 
    }

    RaidUtils_debug("RaidUtils_sim_push: unable to simpush disk <$full_disk_name>", 1); 
    return FALSE;
} #RaidUtils_sim_push

# -------------------------------------------
# define the local routines to be used by all 
# defined functions within the library
# -------------------------------------------
sub RaidUtils_say($$$) {
    RaidUtils_debug("RaidUtils_say:... "); 
    my ($conn, $cmds, $opts) = @_; 

    while ( my ( $key, $value) = each(%$opts)) {
	print "key=$key" . " " . "value=$value" . "\n"; 
    }

    my $priv = "diag"; 

    if (defined $opts->{blind}) {
	my (undef, undef, $output) = say($conn, "priv set -q $priv; $cmds ", -blind, '-timeout', SAY_TIMEOUT); 
	return $output; 
    }
    
    my (undef, undef, $output) = say($conn, "priv set -q $priv; $cmds ", '-timeout', SAY_TIMEOUT); 
    return $output; 
} #RaidUtils_say

=head2 RaidUtils_debug

    RaidUtils_debug($msg); 

The function only displays the message if 
the VERBOSE parameter is set to 1. 

=over

=item Arguments

=over

=item msg

The msg to be displayed when VERBOSE is set.

=back 

=item Return

None

=cut

sub RaidUtils_debug($) {
    if ($VERBOSE) {
	print("\nDEBUG: $_[0]\n"); 
    }

} #RaidUtils_debug

=head2 RaidUtils_testme_driver

This routine is written to load and execute the testme modules defined
within a library. Upon invocation, the routine automatically
loads and executes the testme_<fucntions>. Each test module testme_<func>
should be written to test/verify a specific function defined in the lib.
The test driver should be named accordingly testme_<funcname>.

This module will simplify the testing effort of any library function
as well as adding and removing functions are much simpler. 

Invocation - refer to the implementation of the testme module in Disk.pm for 
invoking the testme_driver using ntest. 

=cut

=item Arguments

=over

=item filer 

(Required) : the filer hostspec. 

=item testme_functions

(Required) : a list of function to be tested. This should come
from the parameter TESTME_FUNCTIONS

=item lib_name

(Required) : specify the lib full qualifier. Ex: "Storage::Disk"

=item export_list

(Required) : specify the global var EXPORT_OK

=back

=item Return

None

=cut

sub RaidUtils_testme_driver($$$@) {
    my ($filer, $testme_functions, $lib_name, @export_list) = @_; 

    param('VERBOSE', '-default', '0'); 

    my $host = host($filer);
    
    # define a callback hash table
    my %hcb = (); 
    
    # --------------------------
    # load the supported routines
    # --------------------------
    foreach my $fn (@export_list) {
	my $tfcn = $lib_name .  "::testme_" . $fn; 
	$hcb{$fn} = \&$tfcn; 
    } 
    
    foreach my $fn (split(/ /, $testme_functions)) {
	if (!defined($hcb{$fn})) {
	    tstlog('-fail', "Undefined function name <$fn>"); 
	    next; 
	}
	my $fcb = $hcb{$fn}; 
	&$fcb($host); 
    } 
} #RaidUtils_testme_driver

################ Htu
=head2 Raidutils_get_group_list

    Raidutils__get_group_list($conn,$filer_vol);

The Routine returns the array of raidgroup list of a given $filer_vol.
if no $filer_vol given, it returns all the raidgroups on the host.
=over

=item Arguments

=over

=item conn

(Required) : the filer connection object

=item filer_vol 
(Required) : name of the aggr or tradvol

=back

=for html <hr>

=cut


sub RaidUtils_get_group_list {
        my (%opts) = validate(@_, {
                                conn => { type=>HASHREF },
                                filer_vol => { type=>SCALAR|UNDEF,default=>undef
},
                                }
                        );

        my $conn = $opts{conn};
        my $filer_vol = $opts{filer_vol};

        my $raid_obj = new RaidInfo($conn);
        my @groups;

        if (!$filer_vol) {
                my %g = $raid_obj->all_raid_groups();
                @groups = keys(%g);
        }
        else {
                my $aggr = $raid_obj->aggregate($filer_vol);
                my %g = $aggr->all_raid_groups();
                @groups = keys(%g);
        }
        return @groups;
}    ###RaidUtils_get_group_list 


=head2 Raidutils_get_disklist_by_group_type

    RaidUtils_get_disklist_by_group_type($conn,$group,$type);

The Routine returns the array of disk list of a given $group that disk type is $type.
=over

=item Arguments

=over

=item conn

(Required) : the filer connection object

=item group
(Required) : name of the raidgroup

=item type
(Required) : the disktype (DATA|PARITY|DPARITY)
=back

=for html <hr>

=cut

sub RaidUtils_get_disklist_by_group_type {
        my (%opts) = validate(@_, {
                                conn => { type=>HASHREF },
                                group => { type=>SCALAR},
				type=> { type=>SCALAR},
				}
                        );

        my $conn = $opts{conn};
        my $group = $opts{group};
        my $type = $opts{type};

        my $raid_obj = new RaidInfo($conn);
        my $group_obj = $raid_obj->raid_group("$group");
	my %group_disk_list_hash = $group_obj->all_disks();
        my @group_disk_list = keys(%group_disk_list_hash);

	my @disk_list;
	foreach my $disk(@group_disk_list) {
		#tstlog ("-comment", "STEPING- $disk \n");
		my $disk_obj =$raid_obj->disk("$disk");
		my $disk_type = $disk_obj->parameter("parent_state");
		#tstlog ("-comment", "$disk TYPE IS $disk_type \n");
		if ($disk_type eq $type ) {
		#	tstlog ("-comment", "FOUND ONE DISK - $disk \n");
			push (@disk_list, $disk);
		}
	}
		
	return @disk_list;
}
1; 
