# $Id: $
##
## @package RaidLW
## @summary RaidLW.pm - RAID lost write library for manipluation of checksum values
##
## @status TESTME
## @change 2010-01-08 pravin.salgond@netapp.com: Burt
##                            1. Added new functions update_checksum, update_data_checksum, update_embedded_checksum
##                            2. Splitted get_checksum into get_data_checksum and get_embedded_checksum
##                            3. Splitted set_checksum into set_data_checksum and set_embedded_checksum
##                            4. Removed subroutines read_block and write_block
##                            5. Updated Checksum anatomy
##                            6. Renamed parameters, volume_raid_type to raid_type and disk_raid_type to disk_type
##                            7. Removed compute_checksum argument from inject_lost_write and all setter functions from checksum class
##                            8. Added functions add_to_FBN and add_to_VBN
## @change 2010-09-07 pravin.salgond@netapp.com: Burt433294 added cpcount support in inject_lost_write_error
## @author pravin.salgond@netapp.com, melan.singh@netapp.com
## @change 2010-10-04 pravin.salgond@netapp.com:
##                            1. Added new functions add cpcount
##                            2. added new function get_gen_count
##
##--------------------------------------------------------
## Package : RaidLW
##--------------------------------------------------------
##   @package           RaidLW
##   @synopsis          new RaidLW(disk_name => $disk_name, disk_type => disk_type,
##                        dbn => $dbn,raid_type => $raid_type,filer_vol => $filer_vol)
##
##   @description       This package defines subroutine to inject, verify and compare gen_id for disks
##
##   @routines
##          compare_random_number
##          compare_gen_count
##          compare_gen_id
##          read_file_to_dd_block
##          write_dd_block_to_file
##          inject_lost_write_error
##--------------------------------------------------------

package RaidLW;

use strict;
use warnings;
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF UNDEF);
use Tharn qw(say script_result host $Log logresult param);
use Utils qw(make_conn);
use Shell qw(chmod sudo cat);
use filer;
use mount;
use RaidSqav3 qw(AUTOLOAD);
use Data::Dumper;
use NACL::APISet;
use NACL::APISet::Exceptions::ResponseException qw(:try);
use NACL::APISet::Exceptions::NoMatchingEntriesException qw(:try);
use Storage::Scsi qw(scsi_set_error scsi_force_error_e scsi_force_error_d);

require Exporter;

BEGIN {

    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess(Tharn::script_result(@msgs));
        });
}
use vars ('$CLIENT', '$OBJ_COUNTER');
my $CMDTIMEOUT = 120;
##--------------------------------------------------------
## Subroutine : new
##--------------------------------------------------------
##   @name              new
##   @synopsis          new RaidLW(disk_name => $disk_name, disk_type => disk_type,
##                                  dbn => $dbn,raid_type => $raid_type,filer_vol => $filer_vol)
##
##   @description       This subroutine creates an object for Raid_lw class
##
##   @arg host              (Required) Hostspec to filer.
##   @arg conntype          (Optional) Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.
##   @arg filermode         (Optional) Specifies to run tests in normal or maintenance mode. Default is normal.
##                                     For maintenance mode the variable will be "MAINT"
##   @arg filer_vol         (optional) This represents the volume name to which disk belongs.
##   @arg disk_name         (Required) This represents name of the disk to which block data belongs
##   @arg dbn               (Required) This represents the dbn number from/to which data is to be read/written.
##   @arg disk_type         (optional) It is raid_type for a disk mentioned by disk_name
##   @arg disk_block_type   (optional) Type of the block, possible: data, checksum
##   @arg raid_type         (optional) This should be raid_type of a volume to which disk represented by disk_name belongs to.
##   @arg debug             (optional) If set to 1, additional log messages will be printed, default 0
##   @arg cksum_type        (optional) This represents checksum type info of a particular volume/aggregate
##   @arg vserver           (optional) The name of virtual server, default is $main::VSERVER (its generaly vs0).
##   @arg dd_option         (optional) This is option required for dd command, default "-c", for checksum type block it should be "".
##   @return                This returns a object for Raid_lw class
##--------------------------------------------------------

sub new {
    if (scalar(@_) % 2) {
        shift;
    }
    my (%args) =
      validate(@_,
               {api_set_obj     => {type => OBJECT,         isa     => 'NACL::APISet', optional => 1},
                host            => {type => OBJECT,         isa     => 'Hostrec',      optional => 1},
                filermode       => {type => SCALAR,         default => "normal",       optional => 1},
                conntype        => {type => SCALAR | UNDEF, default => undef,          optional => 1},
                disk_name       => {type => SCALAR | UNDEF, default => undef,          optional => 1},
                raid_type       => {type => SCALAR | UNDEF, default => undef,          optional => 1},
                filer_vol       => {type => SCALAR | UNDEF, default => undef,          optional => 1},
                dbn             => {type => SCALAR | UNDEF, default => undef,          optional => 1},                
                disk_type       => {type => SCALAR | UNDEF, default => undef,          optional => 1},
                cksum_type      => {type => SCALAR | UNDEF, default => undef,          optional => 1},
                vserver         => {type => SCALAR,         default => $main::VSERVER, optional => 1},
                dd_option       => {type => SCALAR,         default => "-c",           optional => 1},
                disk_block_type => {type => SCALAR,         default => "data",         optional => 1},
                mount           => {type => SCALAR,         default => 1,              optional => 1},
                unmount         => {type => SCALAR,         default => 1 ,             optional => 1},
                mountpoint      => {type => SCALAR | UNDEF, default => undef ,         optional => 1},
                mountvol        => {type => SCALAR | UNDEF, default => undef ,         optional => 1},                
                debug           => {type => SCALAR,         default => 0,              optional => 1},
                fiji_repeat_count => {type => SCALAR,         default => 1,              optional => 1},
                }
              );
    my %self = ();
    $self{'api_set_obj'}     = $args{api_set_obj};
    $self{'host'}            = $args{host};                #Eventually this parameter will be removed once all scripts are ported to APISet
    $self{'filermode'}       = $args{filermode};
    $self{'conntype'}        = $args{conntype};
    $self{'disk_name'}       = $args{disk_name};
    $self{'filer_vol'}       = $args{filer_vol};
    $self{'disk_type'}       = lc $args{disk_type};
    $self{'raid_type'}       = lc $args{raid_type};
    $self{'dbn'}             = $args{dbn};
    $self{'cksum_type'}      = lc $args{cksum_type};
    $self{'debug'}           = $args{debug};
    $self{'vserver'}         = $args{vserver};
    $self{'dd_option'}       = ($args{disk_block_type} eq "checksum") ? "" : $args{dd_option};
    $self{'disk_block_type'} = $args{disk_block_type};
    $self{'mount'}           = $args{mount};
    $self{'unmount'}         = $args{unmount};
    $self{'mountpoint'}      = $args{mountpoint};
    $self{'mountvol'}        = $args{mountvol};
    $self{'fiji_repeat_count'} = $args{fiji_repeat_count};

    $self{'error'}         = undef;            #It stores last error occured (Not implemented yet)
    $self{'timeout'}       = 120;
    my %conn_args = ();

    if (!(defined($args{host}) or defined($args{api_set_obj}))) {
        die("RaidLW: None of host or NACL objects are passed as arguments");
    } elsif (defined($args{api_set_obj})) {
        $conn_args{api_set_obj} = $args{api_set_obj};
    } elsif (defined($args{host})) {
        $conn_args{host} = $args{host};
    }
    $self{conn_args} = \%conn_args;

    #Set host parameter for other subroutines
    if (!defined($args{'host'})) {
        if (defined($args{api_set_obj}->{hostobj})) {
            $self{'host'} = $args{api_set_obj}->{hostobj};
        } elsif (defined($args{node})) {
            $self{'host'} = host($args{node});
        } elsif (defined($args{api_set_obj}->{connection})) {
            $self{'host'} = $args{api_set_obj}->{connection}->{connhost};
        } elsif (defined($main::FILER)) {
            my $host = host($main::FILER);
            $self{'host'} = $host;
        }
    } ## end if (!defined($args{'host'...

    if (!defined($self{'disk_name'}) && !defined($self{'filer_vol'})) {
        die("RaidLW: Arguments missing - disk_name, filer_vol");
    }
    #define conntype
    if (!defined($self{'conntype'})) {
        if ($self{api_set_obj}->{set} =~ /CMode/) {
            $self{'conntype'} = "#ssh";
            $self{'cmode'}    = 1;
        } else {
            $self{'conntype'} = "#rsh";
            $self{'cmode'}    = 0;
        }
    }

    my $raidconfig_obj = new RaidConfig(
        host        => $self{host},
        conntype    => $self{conntype},
        api_set_obj => $self{api_set_obj}
    );
    $self{'raidconfig_obj'} = $raidconfig_obj;
    $self{timeout} = 120;

    if ($self{filermode} ne "normal") {
        $self{conntype} = "#console";
    }

    $self{'conn_obj'} = make_conn($self{host}, $self{conntype});

    if (defined($self{'disk_name'})) {
        if (!defined($self{'disk_type'})) {
            $self{'disk_type'} = $raidconfig_obj->get_disk_type(disk_name => $self{disk_name});
        }
    }

    if ($self{'disk_type'} =~ /data/) {
        $self{'disk_rgdn'} = $raidconfig_obj->get_disk_rgdn(disk_name => $self{disk_name});
    }

    # Gets volume name if not provided.
    if (!defined($self{'filer_vol'})) {
        $self{filer_vol} =
          RaidSqa_get_disk_aggr_name(%{$self{conn_args}},
                                     filermode => $self{filermode},
                                     conntype  => $self{conntype},
                                     disk_name => $self{disk_name});
    }

    # Gets volume raid type as raid_dp/raid4 if not provided.
#    if (!defined($self{'raid_type'})) {
#        $self{raid_type} =
#          RaidSqa_get_aggr_options(%{$self{conn_args}},
#                                   filermode => $self{filermode},
#                                   conntype  => $self{conntype},
#                                   filer_vol => $self{filer_vol},
#                                   options   => 'raidtype');
#    }
#
#    if (!defined($self{'cksum_type'})) {
#        my $cksum_type =
#          RaidSqa_get_aggr_cksum_type(%{$self{conn_args}},
#                                      conntype  => $self{conntype},
#                                      filer_vol => $self{filer_vol});
#        $self{cksum_type} = $cksum_type;
#    }

    $RaidLW::OBJ_COUNTER++;
    $Log->comment("RaidLW:: ObjectCounter: $RaidLW::OBJ_COUNTER");
    my $self = \%self;
    bless $self;

    if ($self{mount} == 0){
        if(!defined($self->{mountpoint}) or !defined($self->{mountvol})){
            # removed die as always mounting is not required to inject errors.
            $Log->comment("RaidLW: Arguments missing - mountpoint, mountvol");
        }
    }else{
        if(!defined($self->{mountvol})){
            if ($self->{cmode}) {
                $self->_handle_mountvol(action => "create");
            }else{
                $self->{mountvol} = RaidSqa_get_root_vol(%{$self->{conn_args}});
            
            }
        }
        
        if (!defined($CLIENT)) {
            $CLIENT = "localhost";
        }

        my $mountpoint = RaidSqa_aggr_mount(%{$self->{conn_args}}, filer_vol=>$self->{mountvol}, client=>$CLIENT, vserver => $self->{vserver});
        if(!$mountpoint) {
            die("RaidLW.pm: Error: aggr ".$self->{mountvol}." could not be mounted");
        }
        $self->{mountpoint} = $mountpoint;
        sleep(5);
    }

    return $self;
} ## end sub new

##--------------------------------------------------------
## Subroutine : get
##--------------------------------------------------------
##   @name             get
##   @synopsis         $Raid_lw_obj->get(key => "dd_block");
##
##   @description      This function returns value of a given key
##
##   @arg key     (Required) One of the field of RaidLW
##
##   @return           value of a given key
##
##--------------------------------------------------------
sub get {
    my $self        = shift;
    my (%opts) = validate(@_, {key => {type => SCALAR}});

    return $self->{$opts{key}};
}

##--------------------------------------------------------
## Subroutine : set
##--------------------------------------------------------
##   @name             set
##   @synopsis         $Raid_lw_obj->set(key => "dd_block", value=> "-c");
##
##   @description      This function sets the key value pairs of RaidLW fields.
##
##   @arg key          (Required) One of the field of RaidLW
##   @arg value        (Required) value of RaidLW field
##
##   @return           none
##
##--------------------------------------------------------
sub set {
    my $self        = shift;
    my (%opts) = validate(@_, {key => {type => SCALAR}, value => {type => SCALAR}});

    $self->{$opts{key}} = $opts{value};
}

##--------------------------------------------------------
## Subroutine : get_raid_type
##--------------------------------------------------------
##   @name             get_raid_type
##   @synopsis         $Raid_lw_obj->get_raid_type();
##
##   @description      This function sets the key value pairs of RaidLW fields.
##
##   @arg key          (Required) One of the field of RaidLW
##   @arg value        (Required) value of RaidLW field
##
##   @return           none
##
##--------------------------------------------------------
sub get_raid_type {
    my $self = shift;

    if($self->{raid_type} eq ""){
        $self->{raid_type} = RaidSqa_get_aggr_options(%{$self->{conn_args}},
                               filermode => $self->{filermode},
                               conntype  => $self->{conntype},
                               filer_vol => $self->{filer_vol},
                               options   => 'raidtype');
    }

    $Log->comment("RaidLW.pm: raid_type of vol ".$self->{filer_vol} .": ".$self->{raid_type});
    return($self->{raid_type});
}

##--------------------------------------------------------
## Subroutine : get_checksum_type
##--------------------------------------------------------
##   @name             get_checksum_type
##   @synopsis         $Raid_lw_obj->get_checksum_type();
##
##   @description      This function sets the key value pairs of RaidLW fields.
##
##   @arg key          (Required) One of the field of RaidLW
##   @arg value        (Required) value of RaidLW field
##
##   @return           none
##
##--------------------------------------------------------
sub get_checksum_type {
    my $self = shift;
    if(!defined($self->{cksum_type})){
        $self->{cksum_type} =  RaidSqa_get_aggr_cksum_type(%{$self->{conn_args}},
                                          conntype  => $self->{conntype},
                                          filer_vol => $self->{filer_vol});
    }
    $Log->comment("RaidLW.pm: checksum_type of vol ".$self->{filer_vol} .": ".$self->{cksum_type});
    return ($self->{cksum_type});
}

##--------------------------------------------------------
## Subroutine : get_disk_raid_type
##--------------------------------------------------------
##   @name             get_disk_raid_type
##   @synopsis         $Raid_lw_obj->get_disk_raid_type();
##
##   @description      This function sets the key value pairs of RaidLW fields.
##
##   @arg key          (Required) One of the field of RaidLW
##   @arg value        (Required) value of RaidLW field
##
##   @return           none
##
##--------------------------------------------------------
sub get_disk_raid_type {
    my $self = shift;
    my $raidconfig_obj = $self->{'raidconfig_obj'};
    my $disk_type = $raidconfig_obj->get_disk_type(disk_name => $self->{disk_name});
    return ($disk_type);
}

##--------------------------------------------------------
## Subroutine : print_stripes
##--------------------------------------------------------
##   @name             print_stripes
##   @synopsis         $RaidLW_obj->print_stripes();
##   @description      This function dumps the checksum for all disk in a raidgroup.
##   @arg              disk_list  - List of the disk for which stripes will be dumped.
##   @arg              dbn        - stripe number to be dumped.
##   @arg              group_name - Group name to get disk list and dump the stripes.
##   @return           none
##--------------------------------------------------------
sub print_stripes {
	my $self = shift;

    my (%opts) = validate(@_, {
		disk_list    => {type => ARRAYREF|UNDEF,default=>undef,optional => 1},
		dbn => {type => SCALAR,default => $self->{dbn},optional => 1},
		group_name => {type => SCALAR|UNDEF,default=>undef,optional => 1},
		});

	my $disk_list = $opts{'disk_list'};
	my $dbn       = $opts{'dbn'};
	my $group_name = $opts{'group_name'};
	my @disks_list=();

	if (!defined($group_name) && !defined($disk_list)) {
		die("RaidLW.pm::get_stripes: Error: Atleast disk_list or group_name should be provided");
	}

	if (defined($disk_list)) { 
		@disks_list = @$disk_list;
	} else {
		@disks_list = RaidSqa_get_disk_list(%{$self->{conn_args}}, group_name => $group_name);
	}
	$Log->comment("RaidLW.pm: disk list to get stripe info : @disks_list");

	foreach my $disk_name (@disks_list) {
		$self->set(key => "disk_name", value=> "$disk_name");
		my $disk_raid_type = $self->get_disk_raid_type();
		logcomment("RaidLW.pm: checksum for disk - $disk_name, disk_raid_type - $disk_raid_type, dbn - $dbn");
		my $cmd = "priv set test; dumpblock dbn -c $disk_name $dbn";
		RaidSqa_filer_cmd(%{$self->{conn_args}}, cmd => $cmd);
	}
}

##--------------------------------------------------------
## Subroutine : get_stripes
##--------------------------------------------------------
##   @name             get_stripes
##   @synopsis         $RaidLW_obj->get_stripes();
##   @description      This function dumps the checksum for all disk in a raidgroup.
##   @arg              disk_list  - List of the disk for which checksum object is to be created.
##   @arg              dbn  - stripe number to be dumped
##   @arg              group_name - Group name to get disk list and create checksum object.
##   @return           none
##--------------------------------------------------------
sub get_stripes {
    my $self = shift;
# if disk_list is provided get checksum object of all the disk for provided dbn; ignore raid_group.
# if disk_list is not provided and raid_group is provided get all disk from that rg and get checksum object of all the disk for provided dbn.
# return hashref in form of $disk_dbn->{disk_name} = $checksum_obj;
# unmount if unmount is 1.

    my (%opts) = validate(@_, {
        disk_list        => {type => ARRAYREF|UNDEF,default => undef,               optional => 1},
        dbn              => {type => SCALAR,        default => $self->{dbn},        optional => 1},
        dd_option        => {type => SCALAR,        default => $self->{dd_option},  optional => 1},
        disk_block_type  => {type => SCALAR,        default => $self->{disk_block_type}, optional => 1},
        group_name       => {type => SCALAR|UNDEF,  default => undef,               optional => 1},
        });

    my $disk_list       = $opts{'disk_list'};
    my $dbn             = $opts{'dbn'};
    my $dd_option       = $opts{'dd_option'};
    my $disk_block_type = $opts{'disk_block_type'};
    my $group_name      = $opts{'group_name'};
    my @disks_list      = ();

    my (%disk_dbn_hash);

    # Check if disk_list or group_name is provided else die out.

    if (!defined($group_name) && !defined($disk_list)) {
        die("RaidLW.pm::get_stripes: Error: Atleast disk_list or group_name should be provided");
    }

    if (defined($disk_list)) {
        @disks_list = @$disk_list;
    } else {
        @disks_list = RaidSqa_get_disk_list(%{$self->{conn_args}}, group_name => $group_name);
    }
    $Log->comment("RaidLW.pm: disk list to get stripe info : @disks_list");

    foreach my $disk_name (@disks_list) {
        $self->set(key => "disk_name",  value=> "$disk_name");
        $self->set(key => "dbn",        value=> "$dbn");
        $self->set(key => "dd_option",  value=> "$dd_option");
        $self->set(key=>  "disk_type",  value=> $self->get_disk_raid_type());
        
        $Log->comment("RaidLW.pm: Get checksum object for disk $disk_name");

        my $dd_block_file = "/dd_block_file";
        my $mountvol   = $self->get_mountvol();

        if($self->{cmode}) {
            $dd_block_file = "/vol/$mountvol/".$dd_block_file;
        }
        RaidSqa_dd(%{$self->{conn_args}},       of      => $dd_block_file,
                   din   => $self->{disk_name}, bin     => $self->{dbn}, 
                   count => 1,                  options => $dd_option);

        my $dd_block = $self->read_file_to_dd_block(filename => $dd_block_file);
        my $ckse = Checksum->new(disk_name       =>   $disk_name,
                                 raid_type       =>   $self->get_raid_type(),
                                 dd_block        =>   $dd_block,
                                 disk_type       =>   $self->{disk_type},                                 
                                 disk_block_type =>   $disk_block_type,
                                 debug           =>   $self->{debug});
        $disk_dbn_hash{$disk_name} = $ckse;
    }
    return(\%disk_dbn_hash);
}

##--------------------------------------------------------
## Subroutine : get_mountpoint
##--------------------------------------------------------
##   @name             get
##   @synopsis         $Raid_lw_obj->get_mountpoint()
##
##   @description      This function returns value of mount point
##
##   @arg key     None
##
##   @return           value of mount point
##
##--------------------------------------------------------
sub get_mountpoint{
    my $self = shift;
    return $self->{mountpoint};
}

##--------------------------------------------------------
## Subroutine : get_mountvol
##--------------------------------------------------------
##   @name             get_mountvol
##   @synopsis         $Raid_lw_obj->get_mountvol()
##
##   @description      This function returns value of mount vol
##
##   @arg key     None
##
##   @return           value of mount vol
##
##--------------------------------------------------------
sub get_mountvol{
    my $self = shift;
    return $self->{mountvol};
}

##--------------------------------------------------------
## Subroutine : get_disk_block_type
##--------------------------------------------------------
##   @name             get_disk_block_type
##   @synopsis         $Raid_lw_obj->get_disk_block_type()
##
##   @description      This function returns block type for given disk and dbn
##
##   @arg disk_name    Name of disk
##   @arg dbn          block number whose block type is to be returned
##
##   @return           Type of block (data or checksum)
##
##--------------------------------------------------------
sub get_disk_block_type {
    my $self = shift;
    my (%opts) = validate(@_, {
                                    disk_name  => {type => SCALAR},
                                    dbn   => {type => SCALAR},
                              });
    my $disk = $opts{disk_name};
    my $dbn  = $opts{dbn};
    my $cksum_dbn = undef;

    my $cmd = "priv set -q diag: dumpblock dbn -c $disk $dbn";
    my $output = RaidSqa_filer_cmd(%{$self->{conn_args}}, cmd=>$cmd);

    if($output =~ /dbn $dbn/im) {
        if($output =~ /cksum_dbn (\d+)/im) {
            $cksum_dbn = $1;
        }
    } else {
        die("RaidLW.pm: Dumpblock command failed");
    }

    if(!defined($cksum_dbn)) {
        return "data";          #we will return default values in case of cksum_dbn not present
    } else {
        if($self->get_checksum_type() ne "block" && $dbn eq $cksum_dbn) {
            return "checksum";
        } else {
            return "data";
        }
    }
}

##--------------------------------------------------------
## Subroutine : compare_random_number
##--------------------------------------------------------
##   @name             compare_random_number
##   @synopsis         $Raid_lw_obj->compare_random_number();
##
##   @description      This function compares random number among data disks, among parity/dparity disks and also each
##                     data disk with parity and dparity disk from a raid group.
##   @arg              None
##   @return           Returns the hash reference for the mismatch found during comparing random number.
##   Exception         None
##--------------------------------------------------------
## Data structure for random number
##
#   %compare_random_number_hash={
#        #foreach each data disk with parity and dparity disk
#        data_parity_mismatch => {
#                    parity => {
#                        # Actual random no for data disk(for no mismatch should be same as expected)
#                        actual => {
#                            data_disk_id => "random no for data disk"
#                            }
#                        # Expected random no to be matched
#                        expected => {
#                            parity_disk_id => "random no for parity disk"
#                            }
#                        }
#                    dparity => {
#                        # Actual random no for data disk(for no mismatch should be same as expected)
#                        actual => {
#                            data_disk_id => "random no for data disk"
#                            }
#                        # Expected random no to be matched
#                        expected => {
#                            dparity_disk_id => "random no for dparity disk"
#                            }
#                        }
#                    }
#        #comparision between parity and dparity disk
#        parity_mismatch => {
#                    parity => {
#                        parity_disk_id => "random no for parity disk"
#                        }
#                    dparity => {
#                        dparity_disk_id => "random no for dparity disk"
#                        }
#                    }
#        #comparision between all data disk random no with reference random no.
#        data_mismatch => {
#                    # Actual random no for data disk(for no mismatch should be same as expected)
#                    actual => {
#                        data_disk_id1 => "random no for data disk"
#                        data_disk_id2 => "random no for data disk"
#                        data_disk_id3 => "random no for data disk"
#                        }
#                    # Expected random no to be matched
#                    expected => {
#                        data_disk_id => "reference random no for data disk"
#                        }
#                    }
#        }
##--------------------------------------------------------

sub compare_random_number {
    my $self = shift;
    my $api_set_obj      = $self->{api_set_obj};
    my $filermode        = $self->{filermode};
    my $conntype         = $self->{conntype};
    my $disk_name        = $self->{disk_name};
    my $disk_rgdn        = $self->{disk_rgdn};
    my $filer_vol        = $self->{filer_vol};
    my $raid_type        = $self->get_raid_type();
    my $disk_type        = $self->{disk_type};
    my $disk_block_type  = $self->{disk_block_type};
    my $dbn              = $self->{dbn};
    my $conn             = $self->{conn_obj};
    my $dd_option        = $self->get(key=>"dd_option");

    my $dparity_cksum_object = '';
    my $random_no_dparity    = '';
    my @dparity_disk_list    = ();
    my %data_disk_object;
    my %mismatch;
    my $mountvol  = $self->{mountvol};


    my $dump_file = "/".join("_", "fileblock",$filer_vol,$disk_type,$disk_name, $dbn);
    if ($self->{cmode}) {
        $dump_file = "/vol/$mountvol".$dump_file;
    }

    my $group_name = RaidSqa_get_disk_group_name(%{$self->{conn_args}},
                                                 filermode => $filermode,
                                                 conntype  => $conntype,
                                                 disk_name => $disk_name);

    if (!$group_name) {
        die("RaidLW: UndefinedValue: group_name is not defined: $group_name");
    }

    # -------------------------------------------------------------------------------------------
    # Get disk list for data and parity disk and if raid_dp then dparity list also.
    # -------------------------------------------------------------------------------------------
    my @data_disk_list =
      RaidSqa_get_disk_list_by_raid_type(%{$self->{conn_args}},
                                         filermode  => $filermode,
                                         conntype   => $conntype,
                                         raid_type  => 'data',
                                         group_name => $group_name);
    $Log->comment(
         "RaidLW::compare_random_number: List of data disk from group $group_name : @data_disk_list"
    );

    my @parity_disk_list =
      RaidSqa_get_disk_list_by_raid_type(%{$self->{conn_args}},
                                         filermode  => $filermode,
                                         conntype   => $conntype,
                                         raid_type  => 'parity',
                                         group_name => $group_name);
    $Log->comment(
        "RaidLW::compare_random_number: List of parity disk from group $group_name : @parity_disk_list"
    );

    @dparity_disk_list =
      RaidSqa_get_disk_list_by_raid_type(%{$self->{conn_args}},
                                         filermode  => $filermode,
                                         conntype   => $conntype,
                                         raid_type  => 'dparity',
                                         group_name => $group_name);

    if ($raid_type eq 'raid_dp' && scalar(@dparity_disk_list)) {
        $Log->comment(
            "RaidLW::compare_random_number: List of dparity disk from group $group_name : @dparity_disk_list"
        );

        # -------------------------------------------------------------------------------------------
        # Create object for dparity disk and get random number.
        # -------------------------------------------------------------------------------------------
        RaidSqa_dd(%{$self->{conn_args}},
                   filermode => $filermode,
                   conntype  => $conntype,
                   of        => $dump_file,
                   din       => $dparity_disk_list[0],
                   bin       => $self->{dbn},
                   count     => 1,
                   options   => $dd_option);

        my $dd_block_dparity = $self->read_file_to_dd_block(filename => $dump_file);

        $dparity_cksum_object = Checksum->new(dd_block        => $dd_block_dparity,
                                              raid_type       => $raid_type,
                                              disk_type       => 'dparity',
                                              disk_block_type => $disk_block_type,
                                              disk_name       => $dparity_disk_list[0],
                                              debug           => $self->{debug});
        $random_no_dparity = $dparity_cksum_object->get_random_number();
    } ## end if ($raid_type eq 'raid_dp'...

    my ($dd_block_parity, $parity_cksum_object, $random_no_parity);
    if (scalar(@parity_disk_list)) {

        # -------------------------------------------------------------------------------------------
        # Create object for parity disk and get random number..
        # -------------------------------------------------------------------------------------------
        RaidSqa_dd(%{$self->{conn_args}},
                   filermode => $filermode,
                   conntype  => $conntype,
                   of        => $dump_file,
                   din       => $parity_disk_list[0],
                   bin       => $self->{dbn},
                   count     => 1,
                   options   => $dd_option);

        $dd_block_parity = $self->read_file_to_dd_block(filename => $dump_file);

        $parity_cksum_object = Checksum->new(dd_block        => $dd_block_parity,
                                             raid_type       => $raid_type,
                                             disk_type       => 'parity',
                                             disk_block_type => $disk_block_type,
                                             disk_name       => $parity_disk_list[0],
                                             debug           => $self->{debug});
        $random_no_parity = $parity_cksum_object->get_random_number();
    } ## end if (scalar(@parity_disk_list...

    # -------------------------------------------------------------------------------------------
    # Create object for all data disks and store it in hash.
    # -------------------------------------------------------------------------------------------
    foreach my $disk (@data_disk_list) {
        RaidSqa_dd(%{$self->{conn_args}},
                   filermode => $filermode,
                   conntype  => $conntype,
                   of        => $dump_file,
                   din       => $disk,
                   bin       => $self->{dbn},
                   count     => 1,
                   options   => $dd_option);

        my $dd_block_data = $self->read_file_to_dd_block(filename => $dump_file);
        my $data_cksum_object = Checksum->new(dd_block        => $dd_block_data,
                                              raid_type       => $raid_type,
                                              disk_type       => 'data',
                                              disk_block_type => $disk_block_type,
                                              disk_name       => $disk,
                                              debug           => $self->{debug});
        $data_disk_object{$disk} = $data_cksum_object;
    } ## end foreach my $disk (@data_disk_list)

    #COMPARISION::
    # -------------------------------------------------------------------------------------------
    # Compare random number for parity and dparity disk and update the hash if mismatch.
    # -------------------------------------------------------------------------------------------
    if (   $raid_type eq 'raid_dp'
        && (scalar(@parity_disk_list))
        && (scalar(@dparity_disk_list))) {
        if ($random_no_parity != $random_no_dparity) {
            $mismatch{'parity_mismatch'}{'parity'}{$parity_disk_list[0]}   = $random_no_parity;
            $mismatch{'parity_mismatch'}{'dparity'}{$dparity_disk_list[0]} = $random_no_dparity;
        }
    }

    # -------------------------------------------------------------------------------------------
    # Compare random number for all data disks with first data disk in @data_disk_list and update the hash if mismatch.
    # -------------------------------------------------------------------------------------------
    my $data_rand_no_reference = $data_disk_object{$data_disk_list[0]}->get_random_number();
    foreach my $disk (@data_disk_list) {
        my $data_random_number = $data_disk_object{$disk}->get_random_number();
        if ($data_random_number != $data_rand_no_reference) {

            # Check if only 1Byte i.e. MSB matches
            my $data_rand_no_reference_1B = ($data_rand_no_reference & 0xFF000000) >> 24;
            my $data_random_number_1B     = ($data_random_number & 0xFF000000) >> 24;

            # Check if other 3Byte of any data disk is zero(0)
            my $data_rand_no_reference_3B = ($data_rand_no_reference & 0x00FFFFFF);
            my $data_random_number_3B     = ($data_random_number & 0x00FFFFFF);
            if ($data_rand_no_reference_1B == $data_random_number_1B) {
                if ((!$data_rand_no_reference_3B) || (!$data_random_number_3B)) {
                    $Log->comment(
                        "RaidLW::compare_random_number: $data_random_number vs $data_rand_no_reference"
                    );
                } else {
                    $mismatch{'data_mismatch'}{'expected'}{$data_disk_list[0]} =
                      $data_rand_no_reference;
                    $mismatch{'data_mismatch'}{'actual'}{$disk} = $data_random_number;
                }
            } else {
                $mismatch{'data_mismatch'}{'expected'}{$data_disk_list[0]} =
                  $data_rand_no_reference;
                $mismatch{'data_mismatch'}{'actual'}{$disk} = $data_random_number;
            }
        } ## end if ($data_random_number...
    } ## end foreach my $disk (@data_disk_list)

    # -------------------------------------------------------------------------------------------
    # Compare random number for all data disks with parity/dparity disk and update the hash if mismatch.
    # -------------------------------------------------------------------------------------------
    foreach my $disk (@data_disk_list) {
        if (scalar(@dparity_disk_list) || scalar(@parity_disk_list)) {

            # Convert 4 byte data random no to 1 byte as parity and dparity random no is of 1 byte only.
            my $data_random_number = $data_disk_object{$disk}->get_random_number();
            my $data_rand_no_1B    = ($data_random_number & 0xFF000000) >> 24;
            if (scalar(@parity_disk_list)) {
                if ($data_rand_no_1B != $random_no_parity) {
                    $mismatch{'data_parity_mismatch'}{'parity'}{'expected'}{$parity_disk_list[0]} =
                      $random_no_parity;
                    $mismatch{'data_parity_mismatch'}{'parity'}{'actual'}{$disk} = $data_rand_no_1B;
                }
            }
            if (scalar(@dparity_disk_list)) {
                if ($data_rand_no_1B != $random_no_dparity) {
                    $mismatch{'data_parity_mismatch'}{'dparity'}{'expected'}
                      {$dparity_disk_list[0]} = $random_no_dparity;
                    $mismatch{'data_parity_mismatch'}{'dparity'}{'actual'}{$disk} =
                      $data_rand_no_1B;
                }
            }
        } ## end if (scalar(@dparity_disk_list...
    } ## end foreach my $disk (@data_disk_list)

    return (\%mismatch);
} ## end sub compare_random_number

##--------------------------------------------------------
## Subroutine : inject_zeroblock_error
##--------------------------------------------------------
##   @name             inject_zeroblock_error
##   @synopsis         $RaidErrInject_obj->inject_zeroblock_error();
##   @description      This function injects zeroblock error on provided disk and dbn.
##   @arg              disk_name - Name of the disk.
##   @arg              dbn - Disk block no where zero block error to be injected.
##   @return           1 on success or error string on failure
##--------------------------------------------------------

sub inject_zeroblock_error {
	my $self = shift;

    my (%opts) = validate(@_, {
        disk_name => {type => SCALAR,default => $self->{disk_name},optional=>1},
        dbn => {type => SCALAR,default => $self->{dbn},optional=>1},
        });

    my $disk_name = $opts{'disk_name'};
    my $dbn       = $opts{'dbn'};

	my $dumpblock_dbn= "priv -level -q diag; dumpblock dbn -c $disk_name $dbn";
    logcomment("RaidLW.pm: Checksum contents before zeroblock err injection");
    RaidSqa_filer_cmd(%{$self->{conn_args}}, cmd => $dumpblock_dbn);

    logcomment("RaidLW.pm: Injecting zeroblock error on disk - $disk_name dbn - $dbn");
	my $cmd_zeroblock = "priv -level -q test;zeroblock dbn $disk_name $dbn";
    my $output = RaidSqa_filer_cmd(%{$self->{conn_args}}, cmd => $cmd_zeroblock);
    logcomment("RaidLW.pm: Output of zeroblock cmd - $output");
    if ($output =~ /(zeroblock dbn: expected <disk_name>|zeroblock dbn: expected <disk_name>)/i) {
        die("RaidLW: Error: zeroblock command - $output");
    }
#    my @lines = split(/\r\n/, $output);
#    logcomment("RaidLW.pm: all line - @lines");
#    foreach my $line (@lines) {
#        logcomment("RaidLW.pm: inside loop line - $line");
#        if ($line =~ /(zeroblock dbn: expected <disk_name>|zeroblock dbn: expected <disk_name>)/i) {
#            die("RaidLW: Error: zeroblock command - $output");
#        }
#    }
#
##    if ($output =~ /\w+\r*\n*/) {
##        die("RaidLW: Error: zeroblock command - $output");
##    }

    logcomment("RaidLW.pm: Checksum contents after zeroblock err injection");
    RaidSqa_filer_cmd(%{$self->{conn_args}}, cmd => $dumpblock_dbn);
    return 1;
}

##--------------------------------------------------------
## Subroutine : inject_scsi_error
##--------------------------------------------------------
##   @name             inject_scsi_error
##   @synopsis         $RaidErrInject_obj->inject_scsi_error();
##   @description      This function injects scsi error on provided disk and dbn.
##   @arg              disk - Name of the disk where media error to be injected
##   @arg              errordesc  - Name of the error sequence to set
##   @arg              loopcount - Number of times to loop error sequence; -1 forever. Default: undef.
##   @return           1 on success or error string on failure.
##--------------------------------------------------------

sub inject_scsi_error {
    my $self = shift;

    my (%opts) = validate(@_, {
        disk_name    => {type => SCALAR,default => $self->{disk_name},optional => 1},
        errordesc => {type => SCALAR},
        loopcount => {type=>SCALAR | UNDEF,default=>undef,optional=>1},    
        });

    # my $host      = $self->{host};
    # my $conntype  = $self->{conntype};
    my $disk_name = $opts{'disk_name'};
    my $error       = $opts{'errordesc'};
    my $loopcount  = $opts{'loopcount'};
    logcomment("RaidErrInject.pm: Setting scsi error - $error; disk - $disk_name");
    eval {
        scsi_set_error(host => $self->{'host'},
                       device    => $disk_name,
                       errordesc => $error,
                       loopcount => $loopcount);
        scsi_force_error_e(host => $self->{'host'}, device => $disk_name);
        # scsi_set_error(%{$self->{conn_args}},
                       # device    => $disk_name,
                       # errordesc => $error,
                       # loopcount => $loopcount);
        # scsi_force_error_e(%{$self->{conn_args}}, device => $disk_name);
    };
    if ($@) {
        die("RaidErrInject.pm: scsi_set_error and scsi_force _error_e error - $error disk - $disk_name failed: $@");
    } else {
        logcomment("RaidErrInject.pm: scsi_set_error and scsi_force _error_e set on disk $disk_name");
    }
}

##--------------------------------------------------------
## Subroutine : inject_media_error
##--------------------------------------------------------
##   @name             inject_media_error
##   @synopsis         $RaidLW_obj->inject_media_error();
##   @description      This function injects media error on provided disk and dbn.
##   @arg              disk - Name of the disk where media error to be injected
##   @arg              dbn  - Block number where media error to be injected.
##   @arg               storage_type - Storage type of the disk.
##   @return           1 on success or error string on failure.
##--------------------------------------------------------

sub inject_media_error {
    my $self = shift;

    my (%opts) = validate(@_, {
            disk_name    => {type => SCALAR,default => $self->{disk_name},optional => 1},
            dbn => {type => SCALAR,default => $self->{dbn},optional => 1},
            storage_type => {type => SCALAR | UNDEF,default => undef,optional => 1},
            fiji_repeat_count => {type => SCALAR ,default => $self->{fiji_repeat_count},optional => 1},
            inject_method => {type => SCALAR, default => 'fiji', optional => 1},
        });

    my $disk_name           = $opts{'disk_name'};
    my $dbn                 = $opts{'dbn'};
    my $storage_type        = $opts{'storage_type'};
    my $inject_method       = $opts{'inject_method'};
    my $fiji_repeat_count   = $opts{'fiji_repeat_count'};

    #if ($storage_type =~ /ata|bsas|sata/i) {
    # 26-May-2011: fiji should work on FCAL and SAS also.
    
    $Log->comment("RaidLW.pm: Inject method - $inject_method");
    die("RaidLW.pm: Invalid value for inject_method - $inject_method, possible values - fiji, disk_mung")
        if($inject_method !~ /fiji|disk_mung/i);

    if ($inject_method =~ /fiji/i) {

        my $media_error_1 = '(repeat 100000 (raid.disk.read.done.check (eq @disk' . " $disk_name)(return 1)))";
        my $media_error_2 = '(repeat '."$fiji_repeat_count".'(raid.disk.read.done.buf (&&(eq @disk'
                             . " $disk_name) (eq " . '@bno'
                             . " $dbn))(return (buf_Errno E_DISK_BAD_SECTOR_ERROR))))";

        #inject 1st fiji
        if (!RaidSqa_fiji_filter(%{$self->{conn_args}},
                                 fiji_point => $media_error_1)) {
            die("Error injecting fiji - $media_error_1 on disk - $disk_name");
        }
        
        #inject 2nd fiji
        if (!RaidSqa_fiji_filter(%{$self->{conn_args}},
                                    fiji_point => $media_error_2)) {
            die("Error injecting fiji - $media_error_2 on disk - $disk_name");
        } 
        
        $Log->comment("RaidLW.pm: Media error injected successfully using fiji");
        return 1;
    }
    
    if($inject_method =~ /disk_mung/i) {
        if(!defined($storage_type)) {
            $Log->comment("RaidLW.pm: Info: Storage type of disk is required for using disk_mung");
            $storage_type = RaidSqa_get_disk_storage_type(%{$self->{conn_args}}, disk_name => $disk_name);
            $Log->comment("RaidLW.pm: Info: Storage type of disk $disk_name - $storage_type");
            die("RaidLW.pm: Error: Get storage type failed, disk - $disk_name, storage_type - $storage_type") if(!$storage_type);
        }
        
        my $sectors_per_dbn = 0;
        $sectors_per_dbn = 8 if($storage_type =~ /FCAL|SAS/i);
        
        if($storage_type =~ /ATA|BSAS|SATA/i) {             ##TODO: Does disk_mung work for SSD too?
            $Log->comment("RaidLW.pm: Info: Checksum type of disk is required to determine sectors per dbn");
            my $checksum_type = RaidSqa_get_disk_checksum_type(%{$self->{conn_args}}, disk_name => $disk_name);
            $Log->comment("RaidLW.pm: Info: Checksum type of disk $disk_name - $checksum_type");
            die("RaidLW.pm: Error: Get checksum type failed, disk - $disk_name, cksum_type - $checksum_type") if(!$checksum_type);
            $sectors_per_dbn = ($checksum_type =~ /block/) ? 9 : 8;
        }
        
        die("RaidLW.pm: Error: Compute sectors per dbn failed - $sectors_per_dbn") if(!$sectors_per_dbn);
        if(!RaidSqa_disk_mung(%{$self->{conn_args}}, disk_name => $disk_name, sector => $dbn * $sectors_per_dbn)) {
            die("RaidLW.pm: Error: Inject media error using disk_mung failed");
        }

        $Log->comment("RaidLW.pm: Media error injected successfully using disk_mung.");
        return 1;        
    }

    $Log->warn("RaidLW.pm: Warning: Media error may not be injected.");
    return 0;

} ## end sub inject_media_error

##--------------------------------------------------------
## Subroutine : inject_parity_inconsistency_error
##--------------------------------------------------------
##   @name             inject_parity_inconsistency_error
##   @synopsis         $RaidLW_obj->inject_parity_inconsistency_error();
##   @description      This function compares gen count for all data disks provided with parity/dparity disk.
##   @arg              None
##   @return           Returns the hash reference for the mismatch found during comparing gen count.
##   Exception         None
#
##  Inject all error and if data_4k is modified skip data_checksum update and update embedded checksum
##
##--------------------------------------------------------
sub inject_parity_inconsistency_error {
    my $self        = shift;
    my (%opts) = validate(@_, {
        mount            => {type => SCALAR,         default  => 0,                   optional => 1},
        unmount          => {type => SCALAR,         default  => 0,                   optional => 1},
        mountpoint       => {type => SCALAR | UNDEF, default  => undef,               optional => 1},
        filer_vol        => {type => SCALAR,         optional => 1},
        disk             => {type => SCALAR,         default  => $self->{disk_name},  optional => 1},
        dbn              => {type => SCALAR,         default  => $self->{dbn},        optional => 1},
        dd_option        => {type => SCALAR,         default  => $self->{dd_option},  optional => 1},
        disk_block_type  => {type => SCALAR,         default  => $self->{disk_block_type}, optional => 1},
#        vserver          => {type => SCALAR,         default =>  $self->{vserver}, optional => 1},
        });
    my $mount           = $opts{'mount'};
    my $unmount         = $opts{'unmount'};
    my $mountpoint      = $opts{'mountpoint'};
    my $filer_vol       = $opts{'filer_vol'};
    my $disk            = $opts{'disk'};
    my $dbn             = $opts{'dbn'};
    my $dd_option       = $opts{'dd_option'};
    my $disk_block_type = $opts{'disk_block_type'};
    my $mountvol        = $self->{mountvol};
    my $conntype        = $self->{conntype};
 #   my $vserver         = $opts{vserver};
    my $vserver         = $self->{vserver};
    $self->{mountpoint} = $mountpoint if (defined($mountpoint));

    my ($dump_file,$disk_type,$original_val,$modified_val);
    my $dd_block_file = "/dd_block_file";

    eval {
        #Algo required
        #1. if mount is required mount the volume
        #2. dd block data to a file
        #3. create checksum object of the dd block
        #4. inject all required fields on checksum
        #5. save the checksum entries to a file
        #6. dd file to disk block
        #7. if unmount is required umount the volume

        # If requested to mount, mount the mentioned filer_vol
        if ($mount  == 1) {
            logcomment("RaidLW.pm: mounting the aggr $filer_vol");
            $mountpoint = RaidSqa_aggr_mount(%{$self->{conn_args}}, conntype=>$conntype, filer_vol=>$filer_vol, vserver => $vserver);
            if(!$mountpoint) {
                die("RaidLW.pm: Error: aggr $filer_vol could not be mounted");
            }
            $self->{mountpoint} = $mountpoint;
        } 

        if($self->{cmode}) {
            $dd_block_file = "/vol/$mountvol/".$dd_block_file;
        }
        
        $self->set(key=>"disk_type", value=>$self->get_disk_raid_type());
        #2. dd block data to a file
        RaidSqa_dd(%{$self->{conn_args}},  conntype  => $conntype, of => $dd_block_file,
                   din     => $disk,       bin       => $dbn,      count => 1,
                   options => $dd_option);

        #3. create checksum object of the dd block
        my $dd_block = $self->read_file_to_dd_block(filename => $dd_block_file);   # read file to dd_block
        my $ckse = Checksum->new(disk_name       => $disk,
                                 raid_type       => $self->get_raid_type(),
                                 dd_block        => $dd_block,
                                 disk_type       => $self->{disk_type},
                                 disk_block_type => $disk_block_type,
                                 debug           => $self->{debug});
                
        #4. inject pi error. corrupt data_4k and update both checksum without modifying any other field
        my $data_4k_str = 'the quick brown fox jumps over the lazy dog.1234567890 'x10;
        $ckse->set_data_4k(data_4k => $data_4k_str);
        $ckse->update_checksum();
        if ($self->{debug}) {
            $Log->comment("RaidLW::DEBUG:  Data Checksum ",     $ckse->get_data_checksum());
            $Log->comment("RaidLW::DEBUG:  Embedded Checksum ", $ckse->get_embedded_checksum());
        }

        #5. save the checksum entries to a file
        my $new_dd_block = $ckse->get_dd_block();
        $self->write_dd_block_to_file(filename => $dd_block_file, dd_block => $new_dd_block);  # write dd_block to file

        #6. dd file to disk block
        RaidSqa_dd(%{$self->{conn_args}},
                   conntype  => $conntype,
                   if        => $dd_block_file,
                   dout      => $disk,
                   bout      => $dbn,
                   count     => 1,
                   options   => $dd_option);

        #7. if unmount is required umount the volume
        if ($unmount == 1) {
            if (defined($self->{mountpoint})) {
                umount($self->{mountpoint});
            }
        }
    };
    if ($@) {
        die "RaidLW.pm: Error: Checksum error injection failed: $@";
    }
}

##--------------------------------------------------------
## Subroutine : inject_checksum_error
##--------------------------------------------------------
##   @name             inject_checksum_error
##   @synopsis         $RaidLW_obj->inject_checksum_error();
##   @description      This function compares gen count for all data disks provided with parity/dparity disk.
##   @arg              mount -> 1 if mount if required 0 if not.
##   @arg              unmount -> 1 if unmount if required 0 if not.
##   @arg              mountpoint -> If defined then the provided mountpoint will be used.
##   @arg              filer_vol -> Name of the volume.
##   @arg              inject_hash -> Hash populated with data to modify checksum fields.
##   @arg              disk_block_type  -> Type of the block, possible: data, checksum
##   @return           Returns error string in case of failure
##   Exception         None
##
##   inject_hash  => {
#                  disk_name             => "$disk_name",
#                  disk_type             => "DATA",
#                  dbn                   => "$dbn",
#                  disk_rgdn             => "$disk_rgdn",
#        
#                  data_4k               =>"$data_4k",
#                  checksum_version      => "checksum_version",
#                  checksum_flags        => "checksum_flags",
#                  data_checksum         => "$data_checksum",
#                  embedded_checksum     => "$embedded_checksum",
#
#                  gen_count             =>"$gen_count",
#                  random_number         =>"$random_number",
#                  DBN                   =>"$DBN",
#                  FBN                   => "$FBN",
#                  VBN                   => "$VBN",
#                  cpcount               => "cpcount",
#                  buftreeid             => "buftreeid"
#                }
##
##--------------------------------------------------------
sub inject_checksum_error {

    my $self        = shift;
    my (%opts) = validate(@_, {
        mount           => {type => SCALAR,       default => 0,optional => 1},
        unmount         => {type => SCALAR,       default => 0 ,optional => 1},
        mountpoint      => {type => SCALAR|UNDEF, default => undef ,optional => 1},
        filer_vol       => {type => SCALAR,       optional => 1},
        inject_hash     => {type => HASHREF},
        disk_name       => {type => SCALAR,       default => $self->{disk_name} ,optional => 1},
        dbn             => {type => SCALAR,       default => $self->{dbn} ,optional => 1},
        dd_option       => {type => SCALAR,       default => $self->{dd_option} ,optional => 1},
        disk_block_type => {type => SCALAR,       default => $self->{disk_block_type} ,optional => 1},
        disk_type       => {type => SCALAR,       default => $self->get_disk_raid_type() ,optional => 1},
        });
    my $mount            = $opts{'mount'};
    my $unmount          = $opts{'unmount'};
    my $inject_hash      = $opts{'inject_hash'};
    my $filer_vol        = $opts{'filer_vol'};
    my $mountpoint       = $opts{'mountpoint'};
    my $disk_name        = $opts{'disk_name'};
    my $dbn              = $opts{'dbn'};
    my $dd_option        = $opts{'dd_option'};
    my $disk_block_type  = $opts{'disk_block_type'};
    my $disk_type        = $opts{'disk_type'};
    my $disk_rgdn        = $inject_hash->{'disk_rgdn'}; # comma separated values    
    my @rgdn_no          = split(/,/, $disk_rgdn) if defined($disk_rgdn);
my $vserver         = $self->{vserver};

    my $mountvol      = $self->{mountvol};
    my $conntype      = $self->{conntype};

    $self->{mountpoint} = $mountpoint if (defined($mountpoint));

    my ($original_val,$modified_val, @error_types);
    my $dd_block_file = "/dd_block_file";

    if($disk_block_type eq "data") {
        @error_types = qw(data_4k checksum_version checksum_flags data_checksum embedded_checksum gen_count random_number DBN FBN VBN cpcount buftreeid);
    } else {
        #For checksum block, only below fields are applicable (#TODO: Check VBN)
        @error_types = qw(checksum_version checksum_flags embedded_checksum DBN VBN);
    }

    $Log->comment("RaidLW.pm:: inject_checksum_error: User requested checksum errors:\n".Dumper($inject_hash));
    $Log->comment("RaidLW.pm:: inject_checksum_error: disk_block_type=> $disk_block_type, dd_option=> $dd_option");
    eval {
        #Algo required
        #1. if mount is required mount the volume
        #2. dd block data to a file
        #3. create checksum object of the dd block
        #4. inject all required fields on checksum
        #5. save the checksum entries to a file
        #6. dd file to disk block
        #7. if unmount is required umount the volume

        # If requested to mount, mount the mentioned mountvol
        if ($mount  == 1) {
            $Log->comment("RaidLW.pm: mounting the aggr $mountvol");
            $mountpoint = RaidSqa_aggr_mount(%{$self->{conn_args}}, conntype=>$conntype, filer_vol=>$mountvol, vserver => $vserver);
            if(!$mountpoint) {
                die("RaidLW.pm: Error: aggr $mountvol could not be mounted");
            }
            $self->{mountpoint} = $mountpoint;
        } 

        if($self->{cmode}) {
            $dd_block_file = "/vol/$mountvol/".$dd_block_file;
        }

        #2. dd block data to a file
        RaidSqa_dd(%{$self->{conn_args}},  
                    conntype  => $conntype, 
                    of        => $dd_block_file,
                    din       => $disk_name, 
                    bin       => $dbn, 
                    options   => $dd_option,
                    count     => 1);

        #3. create checksum object of the dd block
        my $dd_block = $self->read_file_to_dd_block(filename => $dd_block_file);
        my $ckse = Checksum->new(disk_name       => $disk_name,
                                 raid_type       => $self->get_raid_type(),
                                 dd_block        => $dd_block,
                                 disk_type       => $disk_type,
                                 disk_block_type => $disk_block_type,
                                 debug           => $self->{debug});
                
        #4. inject all required fields on checksum
        foreach my $error_key (@error_types) {
            if (!defined($inject_hash->{$error_key})) {
                next;
            }

            logcomment("RaidLW.pm: Modify checksum field - $error_key");
            no strict 'refs';
            if ($error_key eq "gen_count") {
                if ($disk_type =~ /parity/i) {
                    if (!@rgdn_no) {
                        die("For modifying $error_key of $disk_type disk disk_rgdn value of data is required");
                    }
                    foreach my $disk_rgdn_value (@rgdn_no) {
                        logcomment("RaidLW.pm: Modify checksum field - $error_key at RGDN = $disk_rgdn_value");
                        $ckse->set_parity_gen_count($error_key => $inject_hash->{$error_key}, disk_rgdn => $disk_rgdn_value);
                    }
                } else {
                    $ckse->set_data_gen_count($error_key => $inject_hash->{$error_key});
                }
            } else {
                my $function_get = "get_"."$error_key";
                my $function_set = "set_"."$error_key";

                $original_val = $ckse->$function_get();
                $ckse->$function_set(lc($error_key) => $inject_hash->{$error_key});
                $modified_val = $ckse->$function_get;

                logcomment("RaidLW.pm: Modifying field - $error_key, original_val -  $original_val, new_value = $modified_val");
            }
            use strict 'refs';
        }

        # If data or embedded checksum has to be corrupted ignore updating with correct values
        if (!exists($inject_hash->{data_4k})){
            if($disk_block_type eq "data" && (!exists($inject_hash->{data_checksum}))){
                $ckse->update_data_checksum();
            }
            if(!exists($inject_hash->{embedded_checksum})){
                $ckse->update_embedded_checksum();
            }
        }    
        
        if ($self->{debug}) {
            $Log->comment("RaidLW::DEBUG: Data Checksum ".$ckse->get_data_checksum());
            $Log->comment("RaidLW::DEBUG: Embedded Checksum ".$ckse->get_embedded_checksum());
        }

        #5. save the checksum entries to a file
        my $new_dd_block = $ckse->get_dd_block();
        $self->write_dd_block_to_file(filename => $dd_block_file, dd_block => $new_dd_block);

        #6. dd file to disk block
        RaidSqa_dd(%{$self->{conn_args}},
                   conntype  => $conntype,
                   if        => $dd_block_file,
                   dout      => $disk_name,
                   bout      => $dbn,
                   count     => 1,
                   options   => $dd_option);

        #7. if unmount is required umount the volume
        if ($unmount == 1) {
            if (defined($self->{mountpoint})) {
                umount($self->{mountpoint});
            }
        }
    };
    if ($@) {
        die "RaidLW.pm: Error: Checksum error injection failed: $@";
    }
}

##--------------------------------------------------------
## Subroutine : set_raw_checksum_word
##--------------------------------------------------------
##   @name             set_raw_checksum_word
##   @synopsis         $RaidLW_obj->set_raw_checksum_word();
##   @description      This function sets new value corresponding to word number provided.
##   @arg              word   -> Which part of checksum field to modify
##   @arg              value  -> New value to set.
##   @return           Returns error string in case of failure
##
##   Structure of 64 byte checksum
##              +------+-------+------+-----+
##              |  1   |   2   |  3   |  4  |
##              +------+-------+------+-----+
##              |  5   |   6   |  7   |  8  |
##              +------+-------+------+-----+
##              |  9   |   10  |  11  |  12 |
##              +------+-------+------+-----+
##              |  13  |   14  |  15  |  16 |
##              +------+-------+------+-----+
##      word is one of the above value of 64Bytes split as 4B word each.
##
##--------------------------------------------------------
sub set_raw_checksum_word {

    my $self = shift;
    my (%opts) = validate(@_, {
        word    => {type => SCALAR},
        value   => {type => SCALAR},
        });

    my $wordNo        = $opts{'word'};
    my $valString     = $opts{'value'};
    my $dd_option     = $self->{dd_option};
    my $mountvol      = $self->{mountvol};
    my $conntype      = $self->{conntype};

    my ($mountpoint, $fullFileName);
    my $dd_block_file = "/dd_block_file";

    #1. Validate the hex value provided.
    if ($valString !~ /^0x[0-9A-Fa-f]{2,8}$/) {
        die("RaidLW.pm: Value should include '0x' and 2 - 8 characters in hex format for eg : 0x1234abcf");
    }
    $valString = hex($valString);

    $mountpoint = $self->{mountpoint};
    #$fullFileName =  "$mountpoint/$dd_block_file";

    if($self->{cmode}) {
        $dd_block_file = "/vol/$mountvol/".$dd_block_file;
    }

    $self->set(key=>"disk_type", value=>$self->get_disk_raid_type());
    #2. dd block data to a file
    RaidSqa_dd(%{$self->{conn_args}},  
                conntype  => $conntype, 
                of        => $dd_block_file,
                din       => $self->{disk_name}, 
                bin       => $self->{dbn}, 
                options   => $self->{dd_option},
                count     => 1);

    #3. create checksum object of the dd block
    my $dd_block = $self->read_file_to_dd_block(filename => $dd_block_file);
    my $ckse = Checksum->new(disk_name       => $self->{disk_name},
                             raid_type       => $self->get_raid_type(),
                             dd_block        => $dd_block,
                             disk_type       => $self->{disk_type},
                             disk_block_type => $self->{disk_block_type},
                             debug           => $self->{debug});

    #4. Set the value corresponding to word provided.
    $ckse->set_raw_checksum_word(word=>$wordNo,value=>$valString);

    #5. Get new dd block to write it back to disk
    my $new_dd_block = $ckse->get_dd_block();
    $self->write_dd_block_to_file(filename => $dd_block_file, dd_block => $new_dd_block);

    #6. dd file to disk block
    RaidSqa_dd(%{$self->{conn_args}},
               conntype  => $conntype,
               if        => $dd_block_file,
               dout      => $self->{disk_name},
               bout      => $self->{dbn},
               options   => $self->{dd_option},
               count     => 1);
}

##--------------------------------------------------------
## Subroutine : compare_gen_count
##--------------------------------------------------------
##   @name             compare_gen_count
##   @synopsis         $Raid_lw_obj->compare_gen_count();
##   @description      This function compares gen count for all data disks provided with parity/dparity disk.
##   @arg              None
##   @return           Returns the hash reference for the mismatch found during comparing gen count.
##   Exception         None
##--------------------------------------------------------
## Data structure for gen count
##
#   %compare_gen_count_hash={
#            #for each disk provide the differences in gen_count
#            disk_id1 => {
#                    #compare data disk, parity disk and dparity disk
#                    data => "gen_count value of data disk for disk_id1",
#                    parity => "gen_count value in parity disk for disk_id1",
#                    dparity => "gen_count value in dparity disk for disk_id1"
#                }
#            disk_id2 => {
#                    #compare data disk, parity disk and dparity disk
#                    data => "gen_count value of data disk for disk_id2",
#                    parity => "gen_count value in parity disk for disk_id2",
#                    dparity => "gen_count value in dparity disk for disk_id2"
#            }
#        }
##
##--------------------------------------------------------
sub compare_gen_count {
    my $self = shift;

    # my $host        = $self->{host};
    my $api_set_obj = $self->{api_set_obj};
    my $filermode   = $self->{filermode};
    my $conntype    = $self->{conntype};
    my $disk_name   = $self->{disk_name};
    my $disk_rgdn   = $self->{disk_rgdn};
    my $filer_vol   = $self->{filer_vol};
    my $disk_type   = $self->{disk_type};
    my $dbn         = $self->{dbn};
    my $conn        = $self->{conn_obj};
    my $raidconfig_obj    = $self->{raidconfig_obj};
    my $raid_type   = $self->get_raid_type();

    my $dd_option         = $self->get(key=>"dd_option");
    my $disk_block_type   = $self->get(key=>"disk_block_type");


    my $group_name = '';
    my (@parity_disk_list, @dparity_disk_list, @data_disk_list);
    my %mismatch             = ();
    my $dparity_cksum_object = '';
    my $mountvol             = $self->{mountvol};

    my $dump_file = "/".join("_", "fileblock",$filer_vol,$disk_type,$disk_name, $dbn);
    if ($self->{cmode}) {
        $dump_file = "/vol/$mountvol".$dump_file;
    }

    $group_name = RaidSqa_get_disk_group_name(%{$self->{conn_args}},
                                              filermode => $filermode,
                                              conntype  => $conntype,
                                              disk_name => $disk_name);

    # -------------------------------------------------------------------------------------------
    # Get disk list for data and parity disk and if raid_dp then dparity list also.
    # -------------------------------------------------------------------------------------------
    @data_disk_list =
      RaidSqa_get_disk_list_by_raid_type(%{$self->{conn_args}},
                                         filermode  => $filermode,
                                         conntype   => $conntype,
                                         raid_type  => 'data',
                                         group_name => $group_name);
    $Log->comment(
           "RaidLW::compare_gen_count: List of data disk from group $group_name : @data_disk_list");

    @parity_disk_list =
      RaidSqa_get_disk_list_by_raid_type(%{$self->{conn_args}},
                                         filermode  => $filermode,
                                         conntype   => $conntype,
                                         raid_type  => 'parity',
                                         group_name => $group_name);
    $Log->comment(
         "RaidLW::compare_gen_count: List of parity disk from group $group_name : @parity_disk_list"
    );

    @dparity_disk_list =
      RaidSqa_get_disk_list_by_raid_type(%{$self->{conn_args}},
                                         filermode  => $filermode,
                                         conntype   => $conntype,
                                         raid_type  => 'dparity',
                                         group_name => $group_name);

    $Log->comment(
        "RaidLW::compare_gen_count: List of parity disk from group $group_name : @dparity_disk_list"
    );
    if (!(scalar(@dparity_disk_list) || scalar(@parity_disk_list))) {
        return \%mismatch;
    }

    if (scalar(@dparity_disk_list)) {
        $Log->comment(
            "RaidLW::compare_gen_count: List of dparity disk from group $group_name : @dparity_disk_list"
        );

        # -------------------------------------------------------------------------------------------
        # Create object for dparity disk.
        # -------------------------------------------------------------------------------------------
        RaidSqa_dd(%{$self->{conn_args}},
                   filermode => $filermode,
                   conntype  => $conntype,
                   of        => $dump_file,
                   din       => $dparity_disk_list[0],
                   bin       => $self->{dbn},
                   count     => 1,
                   options   => $dd_option);

        my $dd_dparity = $self->read_file_to_dd_block(filename => $dump_file);
        $dparity_cksum_object = Checksum->new(dd_block        => $dd_dparity,
                                              raid_type       => $raid_type,
                                              disk_type       => 'dparity',
                                              disk_block_type => $disk_block_type,
                                              disk_name       => $dparity_disk_list[0],
                                              debug           => $self->{debug});

    } ## end if (scalar(@dparity_disk_list...

    # -------------------------------------------------------------------------------------------
    # Create object for parity disk.
    # -------------------------------------------------------------------------------------------
    my $parity_cksum_object;
    if (scalar(@parity_disk_list)) {
        RaidSqa_dd(%{$self->{conn_args}},
                   filermode => $filermode,
                   conntype  => $conntype,
                   of        => $dump_file,
                   din       => $parity_disk_list[0],
                   bin       => $self->{dbn},
                   count     => 1,
                   options   => $dd_option);

        my $dd_parity = $self->read_file_to_dd_block(filename => $dump_file);
        $parity_cksum_object = Checksum->new(dd_block        => $dd_parity,
                                             raid_type       => $raid_type,
                                             disk_type       => 'parity',
                                             disk_block_type => $disk_block_type,
                                             disk_name       => $parity_disk_list[0],
                                             debug           => $self->{debug});
    } ## end if (scalar(@parity_disk_list...

    # -------------------------------------------------------------------------------------------
    # Create object for each data disk and get gen count and compare with parity/dparity and update hash if mismatch.
    # -------------------------------------------------------------------------------------------
    $raidconfig_obj->refresh();
    foreach my $disk (@data_disk_list) {
        my $disk_rgdn = $raidconfig_obj->get_disk_rgdn(disk_name => $disk);
        RaidSqa_dd(%{$self->{conn_args}},
                   filermode => $filermode,
                   conntype  => $conntype,
                   of        => $dump_file,
                   din       => $disk,
                   bin       => $self->{dbn},
                   count     => 1,
                   options   => $dd_option);

        my $dd_data = $self->read_file_to_dd_block(filename => $dump_file);

        # -------------------------------------------------------------------------------------------
        # Creating object for each data disk.
        # -------------------------------------------------------------------------------------------
        my $data_cksum_object = Checksum->new(dd_block        => $dd_data,
                                              raid_type       => $raid_type,
                                              disk_type       => 'data',
                                              disk_name       => $disk,
                                              disk_block_type => $disk_block_type,
                                              debug           => $self->{debug});

        # -------------------------------------------------------------------------------------------
        # Get gen count for data disk and parity disk.
        # -------------------------------------------------------------------------------------------
        my $data_gen_count = $data_cksum_object->get_data_gen_count();

        if (scalar(@parity_disk_list)) {
            my $parity_gen_count =
              $parity_cksum_object->get_parity_gen_count(disk_rgdn => $disk_rgdn);

            # -------------------------------------------------------------------------------------------
            # Compare gen count for data disk with parity disk.
            # -------------------------------------------------------------------------------------------
            if ($data_gen_count != $parity_gen_count) {
                $mismatch{$disk}{'data'}   = $data_gen_count;
                $mismatch{$disk}{'parity'} = $parity_gen_count;
            }
        } ## end if (scalar(@parity_disk_list...

        if (scalar(@dparity_disk_list)) {

            # -------------------------------------------------------------------------------------------
            # Get gen count for dparity disk and compare with data disk.
            # -------------------------------------------------------------------------------------------
            my $dparity_gen_count =
              $dparity_cksum_object->get_parity_gen_count(disk_rgdn => $disk_rgdn);
            if ($data_gen_count != $dparity_gen_count) {
                $mismatch{$disk}{'data'}    = $data_gen_count;
                $mismatch{$disk}{'dparity'} = $dparity_gen_count;
            }
        } ## end if (scalar(@dparity_disk_list...
    } ## end foreach my $disk (@data_disk_list)
    return \%mismatch;
} ## end sub compare_gen_count

##--------------------------------------------------------
## Subroutine : compare_gen_id
##--------------------------------------------------------
##   @name             compare_gen_id
##   @synopsis         $Raid_lw_obj->compare_gen_id();
##   @description      This function compares gen count and random number for all disks provided with parity/dparity disk.
##   @arg              None
##   @return           Returns the hash reference for the mismatch found during comparing gen count and random number.
##--------------------------------------------------------
sub compare_gen_id {
    my $self = shift;

    # my $host        = $self->{host};
    my $api_set_obj = $self->{api_set_obj};
    my $filermode   = $self->{filermode};
    my $conntype    = $self->{conntype};
    my $disk_name   = $self->{disk_name};

    my %gen_id_mismatch = ();

    # -------------------------------------------------------------------------------------------
    # Compare random number and gen count and update the hash.
    # -------------------------------------------------------------------------------------------

    my $mismatch_rand_no = $self->compare_random_number();
    if ((keys %$mismatch_rand_no) >= 1) {
        $gen_id_mismatch{'random_number'} = $mismatch_rand_no;
    }

    my $mismatch_gen_count = $self->compare_gen_count();
    if ((keys %$mismatch_gen_count) >= 1) {
        $gen_id_mismatch{'gen_count'} = $mismatch_gen_count;
    }
    return (\%gen_id_mismatch);

} ## end sub compare_gen_id

##--------------------------------------------------------
## Subroutine : read_file_to_dd_block
##--------------------------------------------------------
##   @name             read_file_to_dd_block
##   @synopsis         $Raid_lw_obj->read_file_to_dd_block(filename => $filename);
##
##   @description      This function writes to an dd block array from a file.
##
##   @arg filename     (Required) Name of the file from where block data is to be read.
##
##   @return           returns reference to array.
##
##   Exception MountFailedException       This exception is thrown when mount fails.
##   Exception FileInvalidException       This exception is thrown is exception if mentioned file does not exist
##--------------------------------------------------------

sub read_file_to_dd_block {
    my $self        = shift;
    # my $host        = $self->{host};
    my $api_set_obj = $self->{api_set_obj};
    my (%opts) = validate(@_, {
                                    filename  => {type => SCALAR},
                                    dd_option => {type => SCALAR, default=>$self->{dd_option}, optional=>1},
                              });
                              
    my $file      = $opts{filename};
    my $dd_option = $opts{dd_option};
    my $file_length = ($dd_option eq "-c") ? 4160 : 4096;
    my $filer_vol = '/';
    my ($string, $main_string);

    my $file_name = '';
    if ($file =~ m(^(/(?:vol)?/[^/]*)?/(.*))) {
        $filer_vol = $1 if ($1);
        $file_name = $2;
    }

    my $nfs_client_file = "$self->{mountpoint}/$file_name";

    sudo("chmod 777 $nfs_client_file");
    sleep(10);

    my $FILE_FH    = '';
    my @block_data = ();
    open(FILE_FH, $nfs_client_file)
      or die "FileInvalidException: RaidLW:: read_file_to_dd_block: failed to open file $nfs_client_file";
    $main_string = "";
    for (my $count = 0 ; $count < $file_length ; $count++) {
        my $str = '';
        read(FILE_FH, $str, 1);
        push(@block_data, ord($str));
    }
    close(FILE_FH);
    return (\@block_data);

} ## end sub read_file_to_dd_block

##--------------------------------------------------------
## Subroutine : write_dd_block_to_file
##--------------------------------------------------------
##   @name             write_dd_block_to_file
##   @synopsis         $Raid_lw_obj->write_dd_block_to_file(dd_block => $dd_block,filename => $filename);
##
##   @description      This function writes the content of dd_block to a file.
##
##   @arg filename     (Required) Name of the file to write data to.
##   @arg dd_block     (Required) Content of dd_block.
##
##   @return           None
##
##   Exception MountFailedException       This exception is thrown when mount fails.
##   Exception FileInvalidException       This exception is thrown is exception if mentioned file does not exist
##--------------------------------------------------------
sub write_dd_block_to_file {
    my $self        = shift;
    # my $host        = $self->{host};
    my $api_set_obj = $self->{api_set_obj};
    my (%opts) = validate(@_, {
                                    filename => {type => SCALAR}, 
                                    dd_block => {type => ARRAYREF},                                  
                               });

    my $file     = $opts{filename};
    my $dd_block = $opts{dd_block};
    my %args     = (action => "mount");
    my ($string, $main_string);
    my $filer_vol = '/';

    my $file_name = '';
    if ($file =~ m(^(/(?:vol)?/[^/]*)?/(.*))) {
        $filer_vol = $1 if ($1);
        $file_name = $2;
    }

    my $nfs_client_file = "$self->{mountpoint}/$file_name";
    sudo("chmod 777 $nfs_client_file");
    $Log->comment("RaidLW::write_dd_block_to_file: nfs_client_file: $nfs_client_file");
    sleep(10);

    my $FILE_FH    = '';
    my @block_data = @$dd_block;

    open(FILE_FH, ">$nfs_client_file")
      or die "FileInvalidException: RaidLW::write_dd_block_to_file: failed to open file $nfs_client_file";
    for (my $count = 0 ; $count < scalar(@block_data) ; $count++) {
        my $char = chr(int($block_data[$count]));
        print FILE_FH $char or die "RaidLW::write_dd_block_to_file: Could not write to file: $nfs_client_file";
    }
    close(FILE_FH);
} ## end sub write_dd_block_to_file

##--------------------------------------------------------
## Subroutine : inject_lost_write_error
##--------------------------------------------------------
##   @name                inject_lost_write
##   @synopsis            $Raid_lw_obj->inject_lost_write(gen_count => $gen_count,random_number => $random_number,fbn => $fbn, lost_write_hash => $lost_write_hash);
##
##   @description         This function modifies three different values, namely 'gen_count', 'random_number' and 'fbn'
##
##   @arg gen_count       Optional This defines generation count to be set on disk
##   @arg random_number   Optional This defines random number to be set on disk
##   @arg fbn             Optional This defines fbn number to be set on disk
##   @arg cpcount         Optional This defines cpcount to be set on disk
##   @arg inject_on       Optional This defines where to inject lost write error, if pariy/dparity, then it will edit values on parity disks
##                            inject_on is not applicable for fbn
##   @arg data_disk_rgdn  Optional This defines rgdn number for data disk on/for which error has to be injected
##   @arg lost_write_hash Optional This defines lost write values for disks, this should be a reference to a hash
##
##   @return              None
##
##   Exception
##
## Lostwrite hash format
## ----------------------------------------------------------------------------------------------------------------
##------ALL THE DISKS IN THIS HASH SHOULD BELONG TO VOLUME TO WHICH DISK OF THIS OBJECT BELONGS--------
## %lw_hash = {
##      'disk_id'   => {
##              type   => 'data'                    #could be parity or dparity
##              dbn    =>  '$dbn_number'            #Stripe on which error is to be injected
##              gen_count =>   '$gen_count'         #possible values 0,1,2,3,"+1","+2","-1","-2"
##              random_number =>  '$random_number'
##              fbn  => '$fbn'
##              cpcount  => '$cpcount'                #cpcount
##           }
##      'disk_id'  =>  {
##              type =>   'parity'                  #could be parity or dparity
##              dbn   => '$dbn_number'
##              random_number  => '$random_number'
##              fbn  => '$fbn'
##              gen_count    {
##                  '$disk_rgdn'  =>  $gen_count    #disk_rgdn number    0to28 #gen_count    0,1,2,3,"+1","+2","-1","-2"
##                  '$disk_rgdn'  =>  $gen_count    #disk_rgdn number    0to28 #gen_count    0,1,2,3,"+1","+2","-1","-2"
##               }
##              cpcount  => '$cpcount'
##          }
##      }
##
### EXAMPLE HASH
## %lw_hash = {
##      'v3.54'  =>  {
##              type  =>  'data'
##              dbn   =>  '13256'
##              gen_count =>   "+2"
##              random_number =>  0x12345678
##              fbn =>  2
##              cpcount =>  4
##           }
##      'v3.56'  =>  {
##              type  =>  'parity'
##              dbn   =>  '13256'
##              random_number =>  0x12
##              gen_count =>   {
##                  2  =>  "+1"
##                  3  =>  "-1"
##               }
##              cpcount =>  4
##          }
##      }
##
## Definitions for terminologies used in above hash
## Above hash enables us to inject lost write errors on multiple disks in a given raid group simultaneously through a single function call.
## Assumptions: All disks in this hash should belong to single raid group
##
## Definitions:
## 'disk_id'    these are keys of first lw_hash and represents disk names on which error is to be injected, value of this would a ref to hash again
## 'type'       This is a key of a hash for each disk_id. Value of this key represents the type of a disk 'disk_id' (possible values data, parity and dparity)
## 'dbn'        This is key of a hash for each disk_id. Value of this key represents stripe number on which error is to be injected.
## 'fbn'        This is key of a hash for each disk_id. Value of this key represents new fbn number to be written on block represeted by 'disk_id' and 'dbn'
##              This value can be of following format "0xABCDEFABCDEFABCDEF", "+20", "-10" or "1234945"
## 'random_number'    This is key of a hash for each disk_id. Value of this key represents new random_number to be written on block represeted by 'disk_id' and 'dbn'
##              This value can be in hex/decimal/string format
## 'gen_count'    This is key of a hash for each disk_id. Value of this key will differ based on the value of key 'type'
##             if type is 'data', then value of 'gen_count' represents generation count that is to be written on the disk. possible values are 0,1,2,3,"+1","+2","-1","-2"
##             if type is 'parity/dparity', then value of this will be a reference to a hash.
##             The key for new hash represents the rdgn number of a disk and value represents generation count to be written for disk having that rgdn
##             disk rgdn (rgdn  raid group disk number) is nothing but a position/rank of a disk in a raid group. This can be retrieved by raid_config info listdisk command
##------------------------------------------------------------------------------
sub inject_lost_write_error {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {gen_count       => {type => SCALAR | UNDEF,  default => undef,  optional => 1},
                random_number   => {type => SCALAR | UNDEF,  default => undef,  optional => 1},
                fbn             => {type => SCALAR | UNDEF,  default => undef,  optional => 1},
                cpcount         => {type => SCALAR | UNDEF,  default => undef,  optional => 1},
                inject_on       => {type => SCALAR,          default => 'data', optional => 1},
                data_disk_rgdn  => {type => SCALAR | UNDEF,  default => undef,  optional => 1},
                lost_write_hash => {type => HASHREF | UNDEF, default => undef,  optional => 1},});

    my $gen_count     = $opts{gen_count};
    my $random_number = $opts{random_number};
    my $fbn           = $opts{fbn};
    my $cpcount       = $opts{cpcount};
    my $inject_on     = lc $opts{inject_on};
    my $lw_hash       = $opts{lost_write_hash};

    my $disk_rgdn = (defined($self->{disk_rgdn})) ? $self->{disk_rgdn} : $opts{data_disk_rgdn};
    my $filer_vol  = $self->{filer_vol};
    my $raid_type = $self->get_raid_type();
    my $conn      = $self->{conn_obj};

    my $dd_option       = $self->get(key=>"dd_option");
    my $disk_block_type = $self->get(key=>"disk_block_type");

#    if ($self->get_checksum_type() =~ /(zoned_v2)|(advanced_zoned)|(zoned v2)/) {
#        $dd_option = "-c";
#    }

    #raidsqa parameters
    # my $host        = $self->{host};
    my $api_set_obj = $self->{api_set_obj};
    my $filermode   = $self->{filermode};
    my $conntype    = $self->{conntype};
    my $mountvol    = $self->{mountvol};

    my @disk_list = ();
    my ($disk_name, $disk_type);

    if (   (!defined($random_number))
        && (!defined($fbn))
        && (!defined($gen_count))
        && (!defined($cpcount))
        && (!defined($lw_hash))) {
        die "ArgumentMissingException: Arguments are not defined properly";
    }

    #If inject_on is parity or dparity, set disk_name to to parity/dparity disk, else set to disk_name of which object is created
    if ($inject_on =~ /parity/) {

        if (!defined($disk_rgdn)) {
            die("RaidLW: data_disk_rgdn parameter value required");
        }

        #get parity and dparity disks
        my $group_name =
          RaidSqa_get_disk_group_name(%{$self->{conn_args}},
                                      filermode => $self->{filermode},
                                      conntype  => $self->{conntype},
                                      disk_name => $self->{disk_name});
        my @parity_disk_list =
          RaidSqa_get_disk_list_by_raid_type(%{$self->{conn_args}},
                                             filermode  => $self->{filermode},
                                             conntype   => $self->{conntype},
                                             raid_type  => $inject_on,
                                             group_name => $group_name);
        $disk_name = $parity_disk_list[0];
        if (!defined($disk_name)) {
            die("RaidLW: Could not find a disk of type: $inject_on");
        }
        $disk_type = $inject_on;
    } else {
        $disk_name = $self->{disk_name};
        $disk_type = $self->{disk_type};
    }

    #---------------------------------------
    #get the list of the disk on which error is to be injected
    #---------------------------------------
    if (defined($opts{lost_write_hash})) {
        @disk_list = keys %$lw_hash;
    } else {

        #If lw hash is not defined, create one using the available object values
        push(@disk_list, $disk_name);
        my %temp_hash = ();
        $temp_hash{$disk_name}->{type}          = $disk_type;
        $temp_hash{$disk_name}->{dbn}           = $self->{dbn};
        $temp_hash{$disk_name}->{random_number} = $opts{random_number};
        $temp_hash{$disk_name}->{fbn}           = $opts{fbn};
        $temp_hash{$disk_name}->{cpcount}       = $opts{cpcount};
        if ($inject_on =~ /data/) {
            $temp_hash{$disk_name}->{gen_count} = $opts{gen_count};
        } elsif ($inject_on =~ /parity/) {
            $temp_hash{$disk_name}->{gen_count}->{$disk_rgdn} = $opts{gen_count};
        }
        $lw_hash = \%temp_hash;
    } ## end else [ if (defined($opts{lost_write_hash...

    if ($self->{debug}) {
        $Log->comment("RaidLW: Inject_lost_write: LW_HASH:: " . Dumper($lw_hash));
    }

    #---------------------------------------------------------
    # Inject lost write error for each disk in the disk_list
    #---------------------------------------------------------
    foreach my $disk (@disk_list) {
        $Log->comment("RaidLW::inject_lost_write_error: Dump block's checksum before injecting error");
        my $cmd = "priv set -q diag; dumpblock dbn -c $disk $lw_hash->{$disk}->{dbn}";
        $Log->comment("RaidLW:: inject_lost_write_error: Error on $disk $lw_hash->{$disk}->{dbn}");
        if($disk_block_type eq "checksum") {
            die("RaidLW:: inject_lost_write_error: lost write can not be injected on checksum block");
        }
        my $dd_option  = ($disk_block_type eq "data") ? "-c" : "";
        my $filename = "/fileblock_" . $self->{disk_name} . "_" . $lw_hash->{$disk}->{dbn};
        my ($client_file, $dump_file);
        $client_file = $self->{mountpoint}.$filename;
        if($self->{cmode}) {
            $dump_file = "/vol/$mountvol/".$filename;
        } else {
            $dump_file = $filename;
        }

        eval {

            #---------------------------------------
            #Create Checksum object for the disk
            #---------------------------------------
            RaidSqa_dd(%{$self->{conn_args}},
                       filermode => $filermode,
                       conntype  => $conntype,
                       of        => $dump_file,
                       din       => $disk,
                       bin       => $lw_hash->{$disk}->{dbn},
                       count     => 1,
                       options   => $dd_option);

            my $dd_block = $self->read_file_to_dd_block(filename => $dump_file);
            my $ckse = Checksum->new(disk_name        => $disk,
                                     disk_type        => $lw_hash->{$disk}->{type},
                                     raid_type        => $raid_type,
                                     dd_block         => $dd_block,
                                     disk_block_type  => $disk_block_type,
                                     debug            => $self->{debug});

            #---------------------------------------
             #Modify generation count for the disk
            #---------------------------------------
            if (defined($lw_hash->{$disk}->{gen_count})) {

                $Log->comment(
                    "RaidLW::inject_lost_write_error: Inject gen_count: $lw_hash->{$disk}->{gen_count}"
                );

                #For data disks
                if ($lw_hash->{$disk}->{type} =~ /data/i) {
                    $ckse->set_data_gen_count(gen_count => $lw_hash->{$disk}->{gen_count});

                    #For parity disks
                } elsif ($lw_hash->{$disk}->{type} =~ /parity/i) {
                    foreach my $rgdn (keys %{$lw_hash->{$disk}->{gen_count}}) {
                        my $gencount = $lw_hash->{$disk}->{gen_count}->{$rgdn};
                        $ckse->set_parity_gen_count(gen_count => $gencount, disk_rgdn => $rgdn);
                    }
                }
            } ## end if (defined($lw_hash->...

            #---------------------------------------
            # Modify random number
            #---------------------------------------
            if (defined($lw_hash->{$disk}->{random_number})) {
                $Log->comment(
                    "RaidLW::inject_lost_write_error: Inject random_number: $lw_hash->{$disk}->{random_number}"
                );
                $ckse->set_random_number(random_number => $lw_hash->{$disk}->{random_number});
            }

            #-----------------------------------------------------
            # Modify fbn number
            #-----------------------------------------------------
            if (defined($lw_hash->{$disk}->{fbn})) {
                $Log->comment(
                     "RaidLW::inject_lost_write_error: Inject FBN: $lw_hash->{$disk}->{fbn}");
                $ckse->set_FBN(fbn => $lw_hash->{$disk}->{fbn});
            }

            #-----------------------------------------------------
            # Modify cpcount number
            #-----------------------------------------------------
            if (defined($lw_hash->{$disk}->{cpcount})) {
                $Log->comment(
                    "RaidLW::inject_lost_write_error: Inject cpcount: $lw_hash->{$disk}->{cpcount}"
                );
                $ckse->set_cpcount(cpcount => $lw_hash->{$disk}->{cpcount});
            }
            
#            #---------------------------------------------------------
#            # TODO: corrupting Data_4k.  This has to be made as configurable
#            #    param in future. For now it is hardcoded.
#            #---------------------------------------------------------
#            # Commented as this will result into PI error
#            my $data_4k_str = '';
#            for (my $i = 0 ; $i < 10 ; $i++) {
#                $data_4k_str = $data_4k_str . "hello $i\n";
#            }
#            $ckse->set_data_4k(data_4k => $data_4k_str);

            #---------------------------------------------------------
            # Dump checksum object on the block
            #---------------------------------------------------------
            $ckse->update_checksum();
            if ($self->{debug}) {
                $Log->comment("RaidLW::DEBUG:  Data Checksum ".     $ckse->get_data_checksum());
                $Log->comment("RaidLW::DEBUG:  Embedded Checksum ". $ckse->get_embedded_checksum());
            }

            my $new_dd_block = $ckse->get_dd_block();
            $self->write_dd_block_to_file(filename => $dump_file, dd_block => $new_dd_block);

            RaidSqa_dd(%{$self->{conn_args}},
                       filermode => $filermode,
                       conntype  => $conntype,
                       if        => $dump_file,
                       dout      => $disk,
                       bout      => $lw_hash->{$disk}->{dbn},
                       count     => 1,
                       options   => $dd_option);

        };
        if ($@) {
            die "InjectLostWriteFailed: Inject lostwrite failed with error: $@";
        }
        $Log->comment("RaidLW::inject_lost_write_error: Dump block's checksum after injecting error");
        RaidSqa_filer_cmd(%{$self->{conn_args}}, cmd=>$cmd);
    } ## end foreach my $disk (@disk_list)
} ## end sub inject_lost_write_error

sub DESTROY {
    my $self = shift;
    my $error;
    if ($@) {
        $error = $@;
        $Log->comment("RaidLW.pm: DESTROY: Error had occured: $error");
    }

    if($self->{unmount}){
        if(!defined($self->{mountpoint})){
            die "RaidLW: Argument missing for unmount - mountpoint ";
        }
        umount($self->{mountpoint});
    }

    $RaidLW::OBJ_COUNTER--;
    $Log->comment("RaidLW: Decrementing obj_counter to $RaidLW::OBJ_COUNTER");
    if (!$RaidLW::OBJ_COUNTER) {
        if ($self->{cmode} && $self->{unmount}) {
            $Log->comment("RaidLW: Reached objects last Counter, destroying RaidLW_mountvol");
            $self->_handle_mountvol(action => "destroy");
        }
    }
    die($error) if(defined($error));
} ## end sub DESTROY

sub _handle_mountvol {
    my $self        = shift;
    my %args        = @_;
    my $action      = $args{action};
    my $volume      = $args{filer_vol};
    my $api_set_obj = $self->{api_set_obj};

    $Log->comment("Checking for metrocluster enabled or not to create mirror or non-mirror dummyvol");
    my $mirror = 0;
    my $disks_count = 2;
    try {
        my $response = $api_set_obj->metrocluster_show('fields' => 'state');
        my @parsed_output = @{$response->get_parsed_output()};
        foreach my $node (@parsed_output) {
        	if (defined(${$node}{state})) {
            	if (grep (/Enabled/, ${$node}{state})) {
			$disks_count=4;
                    	$mirror=1;
            	    }
           	}
        }
    } catch NACL::APISet::Exceptions::NoMatchingEntriesException with {
          $Log->comment("Test is being executed in a Pre-FS build");
    };
    $Log->comment("Mirror is $mirror to create dummy mirror volume (1 for MCC configuration and 0 for non-MCC");

    $Log->comment("Performing $action action for mountvol");
    if ($action =~ /create/) {
        my $mountvol = "cmodedummyvol";
        if (RaidSqa_check_aggr_exist_anytype(%{$self->{conn_args}}, filer_vol => $mountvol, node => '*')) {
            $Log->comment("RaidLW: Reusing $mountvol for this object");
            $self->{mountvol} = $mountvol;
        }else{
            my @pool0_disks = RaidSqa_get_disk_list_by_pool(%{$self->{conn_args}}, pool_name  => "Pool0", spare_only => "yes");

            if (scalar(@pool0_disks) < 3) {
                $Log->comment("RaidLW: Pool0: " . scalar(@pool0_disks));
                die("RaidLW: _create_dummy_vol: Minimum 3 spare disks are required to create a dummy vol - $mountvol");
            }

            $Log->comment("RaidLW: Creating Dummy volume for this object");
            my $result = RaidSqa_aggr_create(%{$self->{conn_args}},
                                             filer_vol  => $mountvol,
                                             raid_type  => "raid4",
                                             aggr_type  => 1,
                                             disk_count => $disks_count,
                                             raid_size  => 2,
                                             mirror     => $mirror,
                                             fake_la    => 0,
                                             aggr_bits  => 64,
                                             vserver    => $self->{vserver});
            if (!$result) {
                die("RaidLW: Could not create a dummy volume - \"$mountvol\" for this object");
            } else {
                $self->{mountvol} = $mountvol;
            }
        } ## end if ($new_vol)
    } ## end if ($action =~ /create/)
    if ($action =~ /destroy/) {
        $Log->comment("RaidLW: destroyed object");
	#if (defined($self->{dummyvol})) {
        if (defined($self->{mountvol})) {
            $self->_remove_dummyvol();
        }
    }
} ## end sub _handle_mountvol

sub _remove_dummyvol {
    my $self        = shift;
    my $api_set_obj = $self->{api_set_obj};

    if (defined($self->{mountvol})) {
        my $result = RaidSqa_aggr_remove(%{$self->{conn_args}},
                                         filer_vol => $self->{mountvol},
                                         aggr_type => 1);
        if (!$result) {
            die("RaidLW: could not remove $self->{mountvol}");
        } else {
            $Log->comment("RaidLW: $self->{'mountvol'} is removed");
        }
    } 
} 

##--------------------------------------------------------
## Package : Checksum
##--------------------------------------------------------
##   @name              Checksum
##   @synopsis          Checksum->new(disk_name => $disk_name, disk_type => disk_type,
##                                  dbn => $dbn,raid_type => $raid_type,filer_vol => $filer_vol)
##
##   @description       This package defines subroutine to get, set Different checksum fields
##
##   @routines
##          get_data_checksum
##          set_data_checksum
##          get_embedded_checksum
##          set_embedded_checksum
##          get_data_gen_count
##          get_parity_gen_count
##          get_parity_gen_array
##          set_data_gen_count
##          set_parity_gen_count
##          get_random_number
##          set_random_number
##          validate_random_number
##          get_dd_block
##          set_dd_block
##          validate_gen_count
##          validate_parity_gen_count
##          update_checksum
##          update_data_checksum
##          update_embedded_checksum
##          get_next_gen_count
##          print_gen_id
##          get_DBN
##          get_FBN
##          get_VBN
##          get_checksum_version
##          get_checksum_flags
##          get_cpcount
##          get_buftreeid
##          set_DBN
##          set_FBN
##          set_VBN
##          set_checksum_version
##          set_checksum_flags
##          set_cpcount
##          set_buftreeid
##          get_data_4k
##          set_data_4k
##          #Internal Functions
##          _calculate_gen_count
##          _get_8bytes
##
##--------------------------------------------------------

package Checksum;

use Data::Dumper;
use strict;
use warnings;
require Exporter;
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF UNDEF);
use Tharn qw(say script_result host $Log logresult param);
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF UNDEF);

BEGIN {
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess(Tharn::script_result(@msgs));
        });
}

##--------------------------------------------------------
## Subroutine : new
##--------------------------------------------------------
##   @name             new
##   @synopsis         Checksum::new(disk_name => $disk_name, disk_type => disk_type, raid_type => $raid_type, dd_block => $dd_block, debug => $debug)
##
##   @description      This subroutine creates an object for Checksum class
##
##   @arg dd_block     (Required) This represents complete block(4096+64 bytes) in string format (concatenations of 4bytes)
##   @arg disk_name    (Required) This represents name of the disk to which block data belongs
##   @arg disk_type    (Required) It is raid_type for a disk mentioned by disk_name
##   @arg vlock_type   (Optional) Type of the block, possible: data, checksum, default data
##   @arg raid_type    (Required) This should be raid_type of a volume to which disk represented by disk_name belongs to.
##
##   @return           This returns a object for Checksum class
##--------------------------------------------------------
sub new {

    #If self is sent during creation of object, remove it from @_
    if (scalar(@_) % 2) {
        shift;
    }
    my (%opts) =
      validate(@_,
               {dd_block        => {type => ARRAYREF},
                disk_type       => {type => SCALAR},
                disk_name       => {type => SCALAR | UNDEF, optional => 1},
                raid_type       => {type => SCALAR},
                disk_block_type => {type => SCALAR, default => "data", optional => 1},
                debug           => {type => SCALAR, default => 0, optional => 1}});

    my %self = ();
    $self{'disk_name'}        = $opts{disk_name};
    $self{'disk_type'}        = $opts{disk_type};
    $self{'raid_type'}        = $opts{raid_type};
    $self{'dd_block'}         = $opts{dd_block};    #This will be a ref to array which will contain 4096 or 4096+64 bytes of data
    $self{'debug'}            = $opts{debug};
    $self{'disk_block_type'}  = $opts{disk_block_type};
    $self{'checksum_offset'}  = ($self{'disk_block_type'} eq "data") ? 4096 : 2048;

    my $ref_self = \%self;
    $ref_self->{'gen_offset'} = ($self{raid_type} eq 'raid4') ? 1 : 2;
    bless $ref_self;
    $ref_self->{'checksum'} = $ref_self->set_dd_block(dd_block => $ref_self->{dd_block});
    $ref_self->dump_ckse_without_data() if($ref_self->{debug});
    return $ref_self;
} ## end sub new

##--------------------------------------------------------
## Subroutine : get_data_checksum
##--------------------------------------------------------
##   @name            get_data_checksum
##   @synopsis        $checksum_obj->get_data_checksum()
##
##   @description     This returns data checksum stored in object value
##
##   @return          Returns the checksum for data for the block of which object is created
##--------------------------------------------------------
sub get_data_checksum {
    my $self = shift;
    return $self->{checksum}->{ckse_checksum};
}

##--------------------------------------------------------
## Subroutine : get_embedded_checksum
##--------------------------------------------------------
##   @name             get_embedded_checksum
##   @synopsis         $checksum_obj->get_embedded_checksum()
##
##   @description      This returns the checksum value for checksum data (embedded checksum) stored in object value
##
##   @arg   type       Checksum type that should be returned (Possible values   embedded or data, default  embedded)
##
##   @return           Returns the checksum value of specified type
##--------------------------------------------------------
sub get_embedded_checksum {
    my $self = shift;
    return $self->{checksum}->{ckse_embedded_cksum};
}

##--------------------------------------------------------
## Subroutine : get_data_gen_count
##--------------------------------------------------------
##   @name             get_data_gen_count
##   @synopsis         $checksum_obj->get_data_gen_count()
##
##   @description      Returns the generation count of a disk and dbn for which object was created
##
##   @return           Generation count of a disk
##
##   Exception DiskRaidTypeInvalid    This exception is thrown if raid_type is not set to data
##--------------------------------------------------------
sub get_data_gen_count {
    my $self = shift;
    if ($self->{disk_type} !~ /data/i) {
        if ($self->{debug}) {
            $Log->comment("Checksum::get_data_gen_count: Raid_type is " . $self->{disk_type});
        }
        die "DiskRaidTypeInvalid: This function is defined for data disks only: $self->{disk_type}";
    }
    return $self->{checksum}->{ckse_lw_id}->{gen_count};
} ## end sub get_data_gen_count

##--------------------------------------------------------
## Subroutine : get_parity_gen_count
##--------------------------------------------------------
##   @name              get_parity_gen_count
##   @synopsis          $checksum_obj->get_parity_gen_count(disk_rgdn => $disk_rgdn);
##
##   @description       This functions returns the generation count from parity disk for a data disk specified by its rgdn number.
##
##   @arg disk_rgdn     rgdn number for a disk whose gen_count is required from parity disk for which object is created
##
##   @return            return the gen_count for data disk mentioned by disk_rgdn
##
##   Exception DiskRaidTypeInvalid    This exception is thrown if raid_type is not set to parity
##--------------------------------------------------------
sub get_parity_gen_count {
    my $self = shift;

    my (%opts) = validate(@_, {disk_rgdn => {type => SCALAR},});

    my $disk_rgdn = $opts{disk_rgdn};

    if ($self->{disk_type} !~ /parity/i) {
        if ($self->{debug}) {
            $Log->comment("Checksum::get_parity_gen_count: Raid_type is " . $self->{disk_type});
        }
        die
          "DiskRaidTypeInvalid: This function is defined for parity or dparity disks only: $self->{disk_type}";
    }
    return $self->{checksum}->{ckse_lw_id}->{gen_bitmap}->[$disk_rgdn];
} ## end sub get_parity_gen_count

##--------------------------------------------------------
## Subroutine : get_parity_gen_array
##--------------------------------------------------------
##   @name              get_parity_gen_array
##   @synopsis          $checksum_obj->get_parity_gen_array();
##
##   @description       Returns the array for gen_count from parity disk
##
##   @return            Array of gen_count for all data disks from a group to which parity disk belongs
##
##   Exception DiskRaidTypeInvalid    This exception is thrown if raid_type is not set to parity
##--------------------------------------------------------
sub get_parity_gen_array {
    my $self = shift;

    if ($self->{disk_type} !~ /parity/i) {
        if ($self->{debug}) {
            $Log->comment("Checksum::get_parity_gen_array: Raid_type is" . $self->{disk_type});
        }
        die
          "DiskRaidTypeInvalid: This function is defined for parity or dparity disks only: $self->{disk_type}";
    }
    my $ref_gen_bitmap = $self->{checksum}->{ckse_lw_id}->{gen_bitmap};
    my @get_bitmap     = @$ref_gen_bitmap;
    return @get_bitmap;
} ## end sub get_parity_gen_array

##--------------------------------------------------------
## Subroutine : get_random_number
##--------------------------------------------------------
##   @name              get_random_number
##   @synopsis          $checksum_obj->get_random_number();
##
##   @description       This function gets the random number from the object data.
##
##   @arg               None
##
##   @return            Returns the random number of disk for which object is created
##                      For parity disk, return 1 byte MSB value
##
##   Exception     None
##--------------------------------------------------------
sub get_random_number {
    my $self = shift @_;

    # returns the random number.
    return ($self->{'checksum'}->{'ckse_lw_id'}->{'random_number'});
}

##--------------------------------------------------------
## Subroutine : set_raw_checksum_word
##--------------------------------------------------------
##   @name             set_raw_checksum_word
##   @synopsis         $checksum_obj->set_raw_checksum_word();
##   @description      This function sets new value corresponding to word number provided.
##   @arg              None
##   @return           None
##
##   Structure of 64 byte checksum
##              +------+-------+------+-----+
##              |  1   |   2   |  3   |  4  |
##              +------+-------+------+-----+
##              |  5   |   6   |  7   |  8  |
##              +------+-------+------+-----+
##              |  9   |   10  |  11  |  12 |
##              +------+-------+------+-----+
##              |  13  |   14  |  15  |  16 |
##              +------+-------+------+-----+
##      word is one of the above value of 64Bytes split as 4B word each.
##
##--------------------------------------------------------
sub set_raw_checksum_word {
    my $self = shift;

    my (%opts) = validate(@_, {
        word    => {type => SCALAR},
        value => {type => SCALAR},
        });

    my @_4bytes    = ();
    my $word       = $opts{'word'};
    my $value      = $opts{'value'};
    my $dd_block   = $self->get_dd_block();
    my @dd_block   = @$dd_block;

    #--------------------------------------------------------
    # Convert one byte array to 4 bytes words
    #--------------------------------------------------------
    for (my $index = 4096 ; $index < 4160 ; $index += 4) {
        my $word =
          ($dd_block[$index + 3] << 24) | ($dd_block[$index + 2] << 16) | ($dd_block[$index + 1] << 8) |
          ($dd_block[$index]);
        push(@_4bytes, $word);
    }

    $_4bytes[$word - 1] = $value;

    my @new_dd_block = @dd_block[0 .. 4095];

    my @temp_array = ();
    #--------------------------------------------------------
    # Convert one byte array to 4 bytes words
    #--------------------------------------------------------
    for (my $index = 4096 ; $index < 4160 ; $index += 4) {
        my $word = $_4bytes[($index - 4096) / 4];
        push(@temp_array,($word & 0x000000FF));
        push(@temp_array,(($word & 0x0000FF00) >> 8));
        push(@temp_array,(($word & 0x00FF0000) >> 16));
        push(@temp_array,(($word & 0xFF000000) >>24));
    }

    push(@new_dd_block,@temp_array);

    $self->set_dd_block(dd_block=>\@new_dd_block);
}

##--------------------------------------------------------
## Subroutine : compare
##--------------------------------------------------------
##   @name              compare
##   @synopsis          $checksum_obj->compare();
##   @description       This function compares the current checksum object with mentioned
##                      checksum object or dd_block.
##   @arg               checksum_object -> Checksum object to compare with existing object.
##   @arg               disk_type -> raid_type of disk . Possible values data/parity/dparity.
##   @arg               raid_type -> raid_type of the volume to which disk belongs.Possible values raid4/raid_dp.
##   @arg               disk_rgdn -> disk_rgdn value of the data disk.
##   @arg               dd_block -> dd block to compare with existing object.
##   @return            Returns hash reference with mismatch output.

##       mismatch output returned will be in below format

#   $VAR1 = {
#       'FBN' => {
#                  'source' => '0x0000000000000000',
#                  'destination' => '0x00000000000000a2'
#                },
#       'DBN' => {
#                  'source' => 0,
#                  'destination' => 23234
#                },
#       'embedded_checksum' => {
#                                'source' => 0,
#                                'destination' => 1112005733
#                              },
#       'gen_count' => {
#                        'disk_rgdn-3' => {
#                                           'source' => 0,
#                                           'destination' => 1
#                                         },
#                        'disk_rgdn-5' => {
#                                           'source' => 0,
#                                           'destination' => 1
#                                         },
#                        'disk_rgdn-4' => {
#                                           'source' => 0,
#                                           'destination' => 1
#                                         },
#                        'disk_rgdn-6' => {
#                                           'source' => 0,
#                                           'destination' => 1
#                                         },
#                        'disk_rgdn-2' => {
#                                           'source' => 0,
#                                           'destination' => 1
#                                         }
#                      },
#       'cpcount' => {
#                      'source' => 0,
#                      'destination' => 15
#                    },
#	}
##--------------------------------------------------------
sub compare{
    my $self = shift;

    my (%opts) = validate(@_, {
        checksum_object => {type => OBJECT, isa => 'Checksum', optional => 1},
        disk_type    => {type =>  SCALAR, default => ""},
        disk_rgdn => {type => SCALAR|UNDEF,default=>undef,optional=>1},
        dd_block => {type =>  ARRAYREF, default => undef, optional => 1},
        raid_type => {type =>SCALAR, default => "raid4", optional => 1},
        ignore_genid_mismatch => {type =>SCALAR, default => 0, optional => 1},
    });

    my $checksum_object       = $opts{checksum_object};
    my $dd_block              = $opts{dd_block};
    my $raid_type             = $opts{raid_type};
    my $disk_type             = $opts{disk_type};
    my $ignore_genid_mismatch =  $opts{ignore_genid_mismatch};
    my $disk_block_type       = $checksum_object->{disk_block_type};
    my $disk_rgdn             = $opts{disk_rgdn};
    my @rgdn                  = ($raid_type eq "raid4") ? (1..16) : (2..28);
    my @disk_rgdn_value       = defined($disk_rgdn) ? $disk_rgdn : @rgdn;
    my $ignore_embed_cksum    = 0;
    my @checksum_fields       = qw(data_4k checksum_version checksum_flags data_checksum gen_count random_number DBN FBN VBN cpcount buftreeid embedded_checksum );

    if (!defined($checksum_object) and !defined($dd_block)) {
        die "Checksum: Error: Either checksum_object or dd_block should be passed as param";
    }

    my $compareResult = {};

    if (defined($checksum_object)) {
        foreach my $field (@checksum_fields) {
            no strict 'refs';

            my $function_name = "get_".$field;
            my $src_val = '';
            my $dst_val = '';
            if ($field eq "gen_count") {
                if ($disk_type =~ /parity/i) {
                    foreach my $disk_rgdn (@disk_rgdn_value) {
                        $src_val = $self->get_parity_gen_count(disk_rgdn=>$disk_rgdn);
                        $dst_val = $checksum_object->get_parity_gen_count(disk_rgdn=>$disk_rgdn);
                        if("$src_val" ne "$dst_val"){
                            $compareResult->{$field}->{"disk_rgdn_".$disk_rgdn} = {source=>$src_val, destination=>$dst_val};
                        }
                    }
                } else {
                    $src_val = $self->get_data_gen_count();
                    $dst_val = $checksum_object->get_data_gen_count();
                }
            }else{
                $src_val = $self->$function_name();
                $dst_val = $checksum_object->$function_name();
            }
            if("$src_val" ne "$dst_val"){
                # Burt 523300 - Randon number reconstruction when only parity drive is present
                # Burt 575101 - If sprintf returns a string less that 6, the check with string compare would fail
                #               Instead comparing based on actual values of MSB byte
                if($field eq "random_number" && $ignore_genid_mismatch == 1) {
                    my $msb_src_val = $src_val & 0xFF000000;
                    my $msb_dst_val = $dst_val & 0xFF000000;
                    $Log->comment("compare: Info: msb source - $msb_src_val, msb destination - $msb_dst_val");
                    if($msb_src_val != $msb_dst_val || ($dst_val & 0x00FFFFFF)) {
                        $compareResult->{$field} = {source=>$src_val, destination=>$dst_val};
                    } else {
                        $Log->comment("compare: Info: As only parity drive is available, the first byte of random number is successfully reconstructed.");
                        $ignore_embed_cksum=1;
                    }
                }
                elsif($field eq "embedded_checksum" && $ignore_embed_cksum == 1) {
                    $Log->comment("compare: Info: Ignoring comparision for embedded_checksum since random number is partially reconstructed.");
                }
                else {
                    $compareResult->{$field} = {source=>$src_val, destination=>$dst_val};
                }
            }

            use strict 'refs';
        }
    } elsif(defined($dd_block)){
        my $checksum_block = Checksum->new(
                                 raid_type       => $raid_type,
                                 dd_block        => $dd_block,
                                 disk_block_type => $disk_block_type, 
                                 disk_type       => $disk_type,
                                 debug           => $self->{debug});

        foreach my $field (@checksum_fields) {
            no strict 'refs';

            my $function_name = "get_".$field;
            my $src_val = '';
            my $dst_val = '';
            if ($field eq "gen_count") {
				if ($disk_type =~ /parity/i) {
					foreach my $disk_rgdn (@disk_rgdn_value) {
						$src_val = $self->get_parity_gen_count(disk_rgdn=>$disk_rgdn);
						$dst_val = $checksum_object->get_parity_gen_count(disk_rgdn=>$disk_rgdn);
						if("$src_val" ne "$dst_val"){
							$compareResult->{$field}->{"disk_rgdn-".$disk_rgdn} = {source=>$src_val, destination=>$dst_val};
						}
					}
                } else {
                    $src_val = $self->get_data_gen_count();
                    $dst_val = $checksum_block->get_data_gen_count();
                }
            }else{
                $src_val = $self->$function_name();
                $dst_val = $checksum_block->$function_name();
            }
            if("$src_val" ne "$dst_val"){
                $compareResult->{$field} = {source=>$src_val, destination=>$dst_val};
            }

            use strict 'refs';
        }
    }
    return $compareResult;
}

##--------------------------------------------------------
## Subroutine : set_random_number
##--------------------------------------------------------
##   @name              set_random_number
##   @synopsis          $checksum_obj->set_random_number(random_number => $random_number);
##
##   @description       This function updates the random number in object data.
##
##   @arg random_number Scalar (could be hex/decimal/string/empty_string)
##                      Represents the new random number to be updated in object data
##                      if set to empty string (e.g ''), random_number will be taken randomly from 0 to 0xFFFFFFFF
##
##   @return            None
##
##   Exception          None
##--------------------------------------------------------
sub set_random_number {
    my $self = shift @_;
    my %opts = validate(@_,
                        {
                            random_number => {type => SCALAR | UNDEF, default => undef, optional => 1},
                        });

    my $random_number = $opts{'random_number'};

    #-------------------- Check if random number is valid or not---------------
    if (defined($random_number)) {
        if ($random_number =~ /^0x[0-9A-F]{1,8}$/i) {
            $random_number = hex($random_number);
        } elsif ($random_number =~ /[A-F]{1,8}/i) {
            $random_number = hex($random_number);
        } elsif ($random_number eq '') {
            my $range = ($self->{disk_type} =~ /data/i) ? 0xFFFFFFFF : 0xFF;
            $random_number = int rand($range);
        } else {
            $random_number = int($random_number);
        }

        #----------------- For invalid Random number throw exception ------------------
    } else {
        die "ArgumentInvalidException: Parameter random_number does not have any value defined";
    }

    #-------------------------------------------------------------------------------

    #------------------------------------------------------------------------
    # update object with new value for random_number.
    #------------------------------------------------------------------------
    $self->{'checksum'}->{'ckse_lw_id'}->{'random_number'} = $random_number;

} ## end sub set_random_number

##--------------------------------------------------------
## Subroutine : validate_random_number
##--------------------------------------------------------
##   @name                validate_random_number
##   @synopsis            $checksum_obj->validate_random_number(random_number => $random_number);
##
##   @description         This function will compare the given random_number with random_number present in the object data. .
##
##   @arg random_number   Scalar
##                        Expected random_number which should ideally match with random_number present in object data
##
##   @return              1 on success or exception(RandomNumberMismatchException) on failure
##
##   Exception RandomNumberMismatchException    This exception is thrown if random number is not matched.
##--------------------------------------------------------
sub validate_random_number {
    my $self = shift @_;
    my %opts = validate(@_, {random_number => {type => SCALAR},});

    my $random_number = $opts{'random_number'};

    #--------------------------------------------------------------------------
    # Compare stored random number with user sepcified random number
    #--------------------------------------------------------------------------
    if (($self->{'checksum'}->{'ckse_lw_id'}->{'random_number'} != $random_number)) {
        die(  "RandomNumberMismatchException : Random number for disk $self->{disk} is"
            . "$self->{'checksum'}->{'ckse_lw_id'}->{'random_number'} expected : $random_number");
    } else {
        return (1);
    }

    #-------------------------------------------------------------
} ## end sub validate_random_number

##--------------------------------------------------------
## Subroutine : get_dd_block
##--------------------------------------------------------
##   @name             get_dd_block
##   @synopsis         $checksum_obj->get_dd_block()
##
##   @description      This function converts internal checksum data structure to an array
##
##   @arg              None
##   @return           reference to array which contains complete block data byte by byte
##                     array is in Little Endian format.
##--------------------------------------------------------
sub get_dd_block {

    my $self    = shift;
    my @_4bytes = (1 .. 16);

    my $ref_checksum = $self->{checksum};
    $self->dump_ckse_without_data() if($self->{debug});

    #----------------initialize all to zero------------------------
    foreach my $byte (@_4bytes) { $_4bytes[$byte-1] = 0; }

    #-----------------------------------------------------
    # Update the values for other checksum fields
    #-----------------------------------------------------
    $_4bytes[0] =
      ($ref_checksum->{ckse_flags} << 8) |
    $ref_checksum->{ckse_version};    #checksum flags and checksum version    
    $_4bytes[1] = $ref_checksum->{ckse_checksum};          #checksum of 4KB data
    $_4bytes[2] = $ref_checksum->{ckse_DBN};               #DBN value from checksum area
    $_4bytes[3] = $ref_checksum->{ckse_embedded_cksum};    #Checksum of 64 byte raid checksum
    $_4bytes[4] = $ref_checksum->{ckse_VBN_LSB};           #Least significant four bytes of VBN
    $_4bytes[5] = $ref_checksum->{ckse_VBN_MSB};           #Most significant four bytes of VBN

    #--------------------------------------------------------------
    # Disect the gen_id bytes into gen_count and random_number
    #--------------------------------------------------------------
    if ($self->{disk_type} =~ /parity/i) {
        my $start_rgdn = $self->{gen_offset};
        $Log->comment("Checksum::get_dd_block: START_RGDN is $start_rgdn");
        if ($self->{debug}) {
            $Log->comment("Checksum::get_dd_block: Gen_offset: $start_rgdn");
        }
        my @gen_bitmap = @{$ref_checksum->{'ckse_lw_id'}->{gen_bitmap}};
        my $gen_byte1  = $gen_bitmap[$start_rgdn];
        my $gen_byte2  = $ref_checksum->{'ckse_lw_id'}->{random_number} << 24;

        foreach my $rgdn ($start_rgdn + 1 .. $start_rgdn + 15) {
            $gen_byte1 = $gen_byte1 | ($gen_bitmap[$rgdn] << ((($rgdn - $start_rgdn) * 2)));
        }

        foreach my $rgdn ($start_rgdn + 16 .. 27) {
            $gen_byte2 = $gen_byte2 | ($gen_bitmap[$rgdn] << (($rgdn - $start_rgdn - 16) * 2));
        }
        $_4bytes[6] = $gen_byte1;    #Represents part of gen_id bitmap on parity disk
        $_4bytes[7] = $gen_byte2;    #Represents second part of gen_id bitmap on parity disk + random_number
    } elsif ($self->{disk_type} =~ /data/i) {
        $_4bytes[6] = $ref_checksum->{ckse_lw_id}->{gen_count};      #two bit gen count of data disk
        $_4bytes[7] = $ref_checksum->{ckse_lw_id}->{random_number};  #Four byte random number
    }

    #------------------------------------------------------

    #-----------------------------------------------------
    # Update the values for remaining checksum fields
    #-----------------------------------------------------
    $_4bytes[8]  = $ref_checksum->{ckse_FBN_LSB};      #least significant four bytes of FBN
    $_4bytes[9]  = $ref_checksum->{ckse_buftreeid};    #four byte buftreeID
    $_4bytes[10] = $ref_checksum->{ckse_cpcount};      #four byte cpcount

    #----------------------------------------------------------
    # Get the pseudo flag values
    #----------------------------------------------------------
    my $temp = $ref_checksum->{ckse_pseudo_flags};     # pseudo bad flag
    $_4bytes[11] = $temp | (($ref_checksum->{ckse_notflexvol_flag} & 0x01) << 2);    #1=>flexvol 0=>tradvol
    $_4bytes[12] = $ref_checksum->{ckse_FBN_MSB};    #most significant four bytes of FBN

    ##ZeroPadding
    my $zero_val = 0;
    for (my $count = 13 ; $count < 16 ; $count++) {
        $_4bytes[$count] = $zero_val & 0xFFFFFFFF;
    }
    for (my $count = 0 ; $count < 16 ; $count++) {
        if ($self->{debug}) {
            $Log->comment("Checksum::get_dd_block: Index is $count $_4bytes[$count]");
        }
    }

    #----------------------------------------------------------------
    # Copy data_4k block
    #----------------------------------------------------------------
    for (my $index = 0 ; $index < 4096 ; $index++) {
        $self->{dd_block}->[$index] = $ref_checksum->{data_4k}->[$index];
    }

    #------------------------------------------------------------------------------
    # Break four byte words into one byte array elements (little_endian)
    #------------------------------------------------------------------------------
    my $cksum_offset = $self->{checksum_offset};

    for (my $index = $cksum_offset ; $index <= $cksum_offset+63 ; $index += 4) {
        my $word = $_4bytes[($index - $cksum_offset) / 4];
        $self->{dd_block}->[$index]     = $word & 0x000000FF;
        $self->{dd_block}->[$index + 1] = ($word & 0x0000FF00) >> 8;
        $self->{dd_block}->[$index + 2] = ($word & 0x00FF0000) >> 16;
        $self->{dd_block}->[$index + 3] = ($word & 0xFF000000) >> 24;
    }

    #----------------------------------------------------------------------------
    return $self->{dd_block};
} ## end sub get_dd_block

##--------------------------------------------------------
## Subroutine : set_dd_block
##--------------------------------------------------------
##   @name             set_dd_block
##   @synopsis         $checksum_obj->set_dd_block()
##
##   @description      This accepts a data block in form of an array and convert it into a data structure.
##                     This function assumes that the block sent in form of array belongs to same disk with
##                     which object was created
##
##   @arg dd_block     reference to array which contains complete block data byte by byte
##                     Should be in Little Endian.
##   @return           None
##--------------------------------------------------------
sub set_dd_block {
    my $self = shift;

    my (%opts) = validate(@_, {dd_block => {type => ARRAYREF},});

    my @_bytes     = @{$opts{dd_block}};
    my @_4bytes    = ();
    my %checksum   = ();
    my %lw_id      = ();
    my @gen_bitmap = (1 .. 28);            ##define an array of 28 elements
    foreach my $index (0 .. 27) { $gen_bitmap[$index] = 0; }

    #--------------------------------------------------------
    # Convert one byte array to 4 bytes words
    #--------------------------------------------------------
    my $cksum_offset = $self->{checksum_offset};
    for (my $index = $cksum_offset ; $index <= $cksum_offset+63 ; $index += 4) {
        my $word =
          ($_bytes[$index + 3] << 24) | ($_bytes[$index + 2] << 16) | ($_bytes[$index + 1] << 8) |
          ($_bytes[$index]);
        push(@_4bytes, $word);
    }

    #---------------------------------------------------------

    #---------------------------------------------------------
    # Assign array values to checksum object
    #---------------------------------------------------------
    $checksum{data_4k}             = [@_bytes[0..4095]];
    $checksum{ckse_version}        = $_4bytes[0] & 0x000000FF;
    $checksum{ckse_flags}          = ($_4bytes[0] & 0x0000FF00) >> 8;
    $checksum{ckse_checksum}       = $_4bytes[1];
    $checksum{ckse_DBN}            = $_4bytes[2];
    $checksum{ckse_embedded_cksum} = $_4bytes[3];
    $checksum{ckse_VBN_LSB}        = $_4bytes[4];
    $checksum{ckse_VBN_MSB}        = $_4bytes[5];

    #---------------------------------------------------------
    # split the 4byte words to get different checksum values
    #---------------------------------------------------------
    if ($self->{disk_type} =~ /parity/) {
        my ($lw_id_byte, $start_rgdn);
        $start_rgdn = $self->{gen_offset};

        foreach my $rgdn ($start_rgdn .. 27) {
            my $offset = 0;
            if ($rgdn < (16 + $start_rgdn)) {
                $lw_id_byte = $_4bytes[6];
            } else {
                $lw_id_byte = $_4bytes[7];
                $offset     = 16;
            }
            ##Create a mask value for two bits corrosponding to disk_rgdn
            my $mask = (0x03 << ($rgdn - $start_rgdn - $offset) * 2);
            my $temp = $lw_id_byte & $mask;
            $gen_bitmap[$rgdn] = $temp >> (($rgdn - $start_rgdn - $offset) * 2);
        } ## end foreach my $rgdn ($start_rgdn...
        $lw_id{'gen_bitmap'}    = \@gen_bitmap;
        $lw_id{'random_number'} = ($_4bytes[7] & 0xFF000000) >> 24;
    } else {
        $lw_id{'gen_count'}     = $_4bytes[6] & 0x00000003;
        $lw_id{'random_number'} = $_4bytes[7];
    }
    $checksum{ckse_lw_id} = \%lw_id;

    #---------------------------------------------------------
    # Assign remaining array values to checksum object
    #---------------------------------------------------------
    $checksum{ckse_FBN_LSB}         = $_4bytes[8];
    $checksum{ckse_buftreeid}       = $_4bytes[9];
    $checksum{ckse_cpcount}         = $_4bytes[10];
    $checksum{ckse_pseudo_flags}    = $_4bytes[11] & 0x3;
    $checksum{ckse_notflexvol_flag} = ($_4bytes[11] & 0x4) >> 2;
    $checksum{ckse_FBN_MSB}         = $_4bytes[12];

    #-----------------------------------------------------------
    $self->{checksum} = \%checksum;
} ## end sub set_dd_block


##--------------------------------------------------------
## Subroutine : dump_ckse_without_data
##--------------------------------------------------------
##   @name             dump_ckse_without_data
##   @synopsis         $checksum_obj->dump_ckse_without_data()
##
##   @description      Dumpe the checksum object hash with data_4k = null
##
##   @arg              none
##   @return           none
##--------------------------------------------------------
sub dump_ckse_without_data {
    my $self = shift;
    my %checksum = %{$self->{checksum}} ;
    $checksum{data_4k} = undef;
    $Log->comment("Checksum:: Dumper: " . Dumper(\%checksum));
    #$Log->comment("Checksum w/o data:: Dumper: " . Dumper($self->{checksum}));
}

##--------------------------------------------------------
## Subroutine : validate_gen_count
##--------------------------------------------------------
##   @name             validate_gen_count
##   @synopsis         $checksum_obj->validate_gen_count()
##
##   @description      This function matches gen_count passsed as argument with the object value
##
##   @arg  gen_count   Gen count which is to be matched with the object value
##   @return           Returns true if gen counts are matchig
##--------------------------------------------------------
sub validate_gen_count {
    my $self = shift;

    my (%opts) = validate(@_, {gen_count => {type => SCALAR},});

    my $expected_gen_count = $opts{gen_count};

    my $current_gen_count = $self->{checksum}->{ckse_lw_id}->{gen_count};
    if (hex($expected_gen_count) == hex($current_gen_count)) {
        $Log->comment("Checksum::validate_gen_count: gen_count is valid");
        return 1;
    } else {
        $self->dump_ckse_without_data() if($self->{debug});
        die("GenCountMismatch: gen_count: Actual: $current_gen_count, Expected: $expected_gen_count");
    }
} ## end sub validate_gen_count

##--------------------------------------------------------
## Subroutine : validate_parity_gen_count (all the values should be passed in decimal or hex format)
##--------------------------------------------------------
##   @name             validate_parity_gen_count
##   @synopsis         $checksum_obj->validate_parity_gen_count()
##
##   @description      This function matches gen_count  passsed as argument with the object value
##
##   @arg  gen_count   Gen count which is to be matched with the object value
##   @arg  disk_rgdn   rgdn number of a data disk for which value is to be matched
##   @return           Returns true if gen counts are matchig
##--------------------------------------------------------
sub validate_parity_gen_count {
    my $self = shift;

    my (%opts) = validate(@_, {gen_count => {type => SCALAR}, disk_rgdn => {type => SCALAR},});
    my $expected_gen_count = $opts{gen_count};
    my $disk_rgdn          = $opts{disk_rgdn};

    my $current_gen_count = $self->{checksum}->{ckse_lw_id}->{gen_bitmap}->[$disk_rgdn];

    if ($expected_gen_count == $current_gen_count) {
        $Log->comment(
            "Checksum::validate_parity_gen_count: Current gen count is same as of expected gen count"
        );
        return 1;
    } else {
        $Log->comment(
            "Checksum::validate_parity_gen_count: disk_rgdn => $disk_rgdn, gen_count=>  $expected_gen_count, current_gen_count => $current_gen_count"
        );
        die
          "GenCountMismatch: Current gen count: $current_gen_count and Expected gen count: $expected_gen_count";
    }
} ## end sub validate_parity_gen_count

##--------------------------------------------------------
## Subroutine : get_next_gen_count
##--------------------------------------------------------
##   @name              get_next_gen_count
##   @synopsis          $checksum_obj->get_next_gen_count()
##
##   @description       It increments the current object value for gen count by 1 and returns it
##
##   @arg               None
##
##   @return            Returns the next gen count for object value
##--------------------------------------------------------
sub get_next_gen_count {
    my $self              = shift;
    my $current_gen_count = $self->get_data_gen_count();
    my $next_gen_count =
      $self->_calculate_gen_count(gen_count => "+1", current_gen_count => $current_gen_count);
    return $next_gen_count;
}

##--------------------------------------------------------
## Subroutine : get_gen_id
##--------------------------------------------------------
##   @name             get_gen_id
##   @synopsis         $checksum_obj->get_gen_id()
##
##   @description      It returns a hashref, with gen_count and random number
##   @arg              None
##
##   @return           Reference to a hash having below format
##
##   FOR DATA DISK OBJECTS
##   %lw_id_hash = {
##                     gen_count  =>  <gen_count>         <gen_count> is 2 bit generation count on data disk
##                     ramdom number =>   <random_no>     <random_no> is 4 byte random_number on data disk
##                  }
##
##   FOR PARITY DISK OBJECTS
##   %lw_id_hash = {
##                     gen_bitmap   => <ref_bitmap_array>  <ref_bitmap_array> a reference to array which contains gen_count bitmap on parity disks
##                     ramdom number =>   <random_no>      <random_no> is 4 byte random_number on data disk
##                 }
##
##
##--------------------------------------------------------
sub print_gen_id {
    my $self      = shift;
    my $ref_lw_id = $self->{ckse_lw_id};
    $Log->comment("Checksum::print_gen_id: LW GEN ID: " . Dumper($ref_lw_id));
    return $ref_lw_id;
}

##--------------------------------------------------------
## Subroutine : get_DBN
##--------------------------------------------------------
##   @name             get_DBN
##   @synopsis         $checksum_obj->get_DBN()
##
##   @description      This function gets the value for DBN field from checksum.
##
##   @arg              None
##
##   @return           Returns the value for parameter DBN
##--------------------------------------------------------
sub get_DBN {
    my $self = shift;
    return $self->{checksum}->{ckse_DBN};

}

##--------------------------------------------------------
## Subroutine : get_FBN
##--------------------------------------------------------
##   @name            get_FBN
##   @synopsis        $checksum_obj->get_FBN()
##
##   @description     This function gets the value for FBN field from checksum (hex equivalent of 64 bit value).
##
##   @arg             None
##
##   @return          Returns the value for parameter FBN
##--------------------------------------------------------
sub get_FBN {
    my $self = shift;

    #covert decimal values to hex string
    my $fbn_lsb;
    my $fbn_msb;

    if ($self->{checksum}->{ckse_FBN_LSB} == 0) {
        $fbn_lsb = "00000000";
    } else {
        $fbn_lsb = sprintf("%08x", $self->{checksum}->{ckse_FBN_LSB});
    }

    if ($self->{checksum}->{ckse_FBN_MSB} == 0) {
        $fbn_msb = "0x00000000";
    } else {
        $fbn_msb = sprintf("%#010x", $self->{checksum}->{ckse_FBN_MSB});
    }

    #return a string of hex characters representing 64 bit value
    # NOTE: We can not store 64 bit value in decimal in 32 bit perl.
    return ("$fbn_msb" . "$fbn_lsb");

} ## end sub get_FBN

##--------------------------------------------------------
## Subroutine : get_VBN
##--------------------------------------------------------
##   @name             get_VBN
##  @synopsis          $checksum_obj->get_VBN()
##
##   @description      This function gets the value for VBN field from checksum.
##
##   @arg              None
##
##   @return           Returns the value for parameter VBN in hex string
##--------------------------------------------------------
sub get_VBN {
    my $self = shift;

    #covert decimal values to hex string
    my $vbn_lsb;
    my $vbn_msb;

    if ($self->{checksum}->{ckse_VBN_LSB} == 0) {
        $vbn_lsb = "00000000";
    } else {
        $vbn_lsb = sprintf("%08x", $self->{checksum}->{ckse_VBN_LSB});
    }

    if ($self->{checksum}->{ckse_VBN_MSB} == 0) {
        $vbn_msb = "0x00000000";
    } else {
        $vbn_msb = sprintf("%#010x", $self->{checksum}->{ckse_VBN_MSB});
    }

    #return a string of hex characters representing 64 bit value
    # NOTE: We can not store 64 bit value in decimal in 32 bit perl.
    return ("$vbn_msb" . "$vbn_lsb");
} ## end sub get_VBN

##--------------------------------------------------------
## Subroutine : get_checksum_version
##--------------------------------------------------------
##   @name             get_checksum_version
##  @synopsis          $checksum_obj->get_checksum_version()
##
##   @description      This function gets the value for checksum_version field from checksum.
##
##   @arg              None
##
##   @return           Returns the value for parameter checksum_version
##--------------------------------------------------------
sub get_checksum_version {
    my $self = shift;
    return $self->{checksum}->{ckse_version};
}

##--------------------------------------------------------
## Subroutine : get_checksum_flags
##--------------------------------------------------------
##   @name             get_checksum_flags
##  @synopsis          $checksum_obj->get_checksum_flags()
##
##   @description      This function gets the value for checksum_flags field from checksum.
##
##   @arg              None
##
##   @return           Returns the value for parameter checksum_flags
##--------------------------------------------------------
sub get_checksum_flags {
    my $self = shift;
    return $self->{checksum}->{ckse_flags};
}

##--------------------------------------------------------
## Subroutine : get_cpcount
##--------------------------------------------------------
##   @name             get_cpcount
##  @synopsis          $checksum_obj->get_cpcount()
##
##   @description      This function gets the value for cpcount field from checksum.
##
##   @arg              None
##
##   @return           Returns the value for parameter cpcount
##--------------------------------------------------------
sub get_cpcount {
    my $self = shift;
    return $self->{checksum}->{ckse_cpcount};
}

##--------------------------------------------------------
## Subroutine : get_buftreeid
##--------------------------------------------------------
##   @name             get_buftreeid
##   @synopsis         $checksum_obj->get_buftreeid()
##
##   @description      This function gets the value for buftreeid field from checksum.
##
##   @arg              None
##
##   @return           Returns the value for parameter buftreeid
##--------------------------------------------------------
sub get_buftreeid {
    my $self = shift;
    return $self->{checksum}->{ckse_buftreeid};
}

##--------------------------------------------------------
## Subroutine : set_data_checksum
##--------------------------------------------------------
##   @name              set_data_checksum
##   @synopsis          $checksum_obj->set_data_checksum(checksum => $checksum);
##
##   @description       This sets the object value for data checksum field with parameter value passed to function
##
##   @arg  checksum     (Required) checksum value to be set
##
##   @return            None
##--------------------------------------------------------
sub set_data_checksum {
    my $self = shift;

    my (%opts) = validate(@_, {data_checksum => {type => SCALAR},});
    my $checksum_value = $opts{'data_checksum'};

    $self->{checksum}->{ckse_checksum} = $checksum_value;
}

##--------------------------------------------------------
## Subroutine : set_embedded_checksum
##--------------------------------------------------------
##   @name             set_embedded_checksum
##   @synopsis         $checksum_obj->set_embedded_checksum(checksum => $checksum);
##
##   @description      This sets the object value for embedded checksum field with parameter value passed to function
##
##   @arg checksum     (Required) checksum value to be set
##
##   @return           None
##--------------------------------------------------------
sub set_embedded_checksum {
    my $self = shift;

    my (%opts) = validate(@_, {embedded_checksum => {type => SCALAR},});
    my $checksum_value = $opts{'embedded_checksum'};
    $self->{checksum}->{ckse_embedded_cksum} = $checksum_value;
}

##--------------------------------------------------------
## Subroutine : update_data_checksum
##--------------------------------------------------------
##   @name              update_data_checksum
##   @synopsis          $checksum_obj->update_data_checksum();
##
##   @description       This calculates the new data checksum and updates the object value
##
##   @arg               None
##
##
##   @return            None
##--------------------------------------------------------
sub update_data_checksum {
    my $self = shift;

    my $checksum_value = $self->_compute_checksum(type => 'data');
    $self->{checksum}->{ckse_checksum} = $checksum_value;
}

##--------------------------------------------------------
## Subroutine : update_embedded_checksum
##--------------------------------------------------------
##   @name              update_embedded_checksum
##   @synopsis          $checksum_obj->update_embedded_checksum();
##
##   @description       This calculates the new embedded checksum and updates the object value
##
##   @arg  checksum     (Required) checksum value to be set
##
##   @return            None
##--------------------------------------------------------
sub update_embedded_checksum {
    my $self = shift;
    my $checksum_value = $self->_compute_checksum(type => 'embedded');
    $self->{checksum}->{ckse_embedded_cksum} = $checksum_value;
}

##--------------------------------------------------------
## Subroutine : update_checksum
##--------------------------------------------------------
##   @name              update_checksum
##   @synopsis          $checksum_obj->update_checksum();
##
##   @description       This calculates the new data & embedded checksum value and updates the object
##
##   @arg               None
##
##
##   @return            None
##--------------------------------------------------------
sub update_checksum {

    my $self = shift;
    $self->update_data_checksum();
    $self->update_embedded_checksum();
}

##--------------------------------------------------------
## Subroutine : set_data_gen_count
##--------------------------------------------------------
##   @name              set_data_gen_count
##   @synopsis          $checksum_obj->set_data_gen_count(gen_count=>$gen_count);
##
##   @description       set the generation count for data disk of which the object is created
##
##   @arg  gen_count    Value for generation count which is to be set
##
##   @return            None
##--------------------------------------------------------
sub set_data_gen_count {
    my $self = shift;

    my (%opts) = validate(@_, {gen_count => {type => SCALAR},});

    my $gen_count = $opts{gen_count};

    my $current_gen_count = $self->{checksum}->{ckse_lw_id}->{gen_count};

    if ($self->{disk_type} !~ /data/i) {
        if ($self->{debug}) {
            $Log->comment("Checksum::set_data_gen_count: disk_type is $self->{disk_type}");
        }
        die 'DiskRaidTypeInvalid: This function is defined for data disks only';
    }

    my $new_gen_count =
      $self->_calculate_gen_count(gen_count         => $gen_count,
                                  current_gen_count => $current_gen_count);

    if ($self->{debug}) {
        $Log->comment("Checksum::set_data_gen_count: New_gen_count is $new_gen_count");
    }
    $self->{checksum}->{ckse_lw_id}->{gen_count} = $new_gen_count;
} ## end sub set_data_gen_count

##--------------------------------------------------------
## Subroutine : set_parity_gen_count
##--------------------------------------------------------
##   @name             set_parity_gen_count
##   @synopsis         $checksum_obj->set_parity_gen_count(disk_rgdn => $disk_rgdn, gen_count=>$gen_count);
##
##   @description      This function will set gen_count value on parity disk for a data disk specified by disk_rgdn
##
##   @arg  disk_rgdn   The rgdn number of data disk for which value is to be set on parity disk
##   @arg  gen_count   Value for generation count which is to be set
##
##   @return           None
##--------------------------------------------------------
sub set_parity_gen_count {
    my $self = shift;

    my (%opts) = validate(@_, {disk_rgdn => {type => SCALAR}, gen_count => {type => SCALAR},});

    my $disk_rgdn = $opts{disk_rgdn};
    my $gen_count = $opts{gen_count};

    if ($disk_rgdn > 28) {
        die "DiskRgdnInvalidException:: DiskRgdn can not be greater than 28";
    }

    if ($self->{disk_type} !~ /parity/i) {
        if ($self->{debug}) {
            $Log->comment("Checksum::set_parity_gen_count: Disk_raid_type is $self->{disk_type}");
        }
        die 'DiskRaidTypeInvalid: This function is defined for parity or dparity disks only';
    }

    my $current_gen_count = $self->{checksum}->{ckse_lw_id}->{gen_bitmap}->[$disk_rgdn];
    my $new_gen_count =
      $self->_calculate_gen_count(gen_count         => $gen_count,
                                  current_gen_count => $current_gen_count);
    $self->{checksum}->{ckse_lw_id}->{gen_bitmap}->[$disk_rgdn] = $new_gen_count;
} ## end sub set_parity_gen_count

##--------------------------------------------------------
## Subroutine : set_DBN
##--------------------------------------------------------
##   @name            set_DBN
##   @synopsis        $checksum_obj->set_DBN(ckecksum_flags => $value)
##
##   @description     Updates the checksum object's DBN field.
##
##   @arg  DBN        (Required) The value of DBN to be set
##
##   @return          None
##--------------------------------------------------------
sub set_DBN {
    my $self = shift;

    my (%opts) = validate(@_, {dbn => {type => SCALAR},});

    my $dbn = $opts{dbn};
    $self->{checksum}->{ckse_DBN} = $dbn;

}

##--------------------------------------------------------
## Subroutine : set_FBN
##--------------------------------------------------------
##   @name             set_FBN
##   @synopsis         $checksum_obj->set_FBN( fbn =>  $value)
##
##   @description      Updates the checksum object's FBN fields.
##
##   @arg  FBN         (Required) The value of FBN to be set (This should be 8byte value in hex format with zero padding)
##                     Example : 0x00000000000F00A1;
##
##   @return           None
##--------------------------------------------------------
sub set_FBN {

    my $self = shift;
    my (%opts) = validate(@_, {fbn => {type => SCALAR},});

    my $fbn = $opts{fbn};

    if ($self->{debug}) {
        $Log->comment("Checksum::set_FBN: FBN:: $fbn");
    }

    if ($fbn =~ /(\+)|(-)/) {

        #Value contains + or - sign, which means we have to add given value to existing fbn value
        $self->add_to_FBN(value => $fbn);
        return;
    }

    if ($fbn !~ /^0x/) {
        $fbn = uc sprintf("%016x", $fbn);
        $fbn = "0x" . $fbn;
    }

    my @chars = split(//, $fbn);
    if (scalar(@chars) != 18) {
        die
          "ArgumentInvalidException: Input FBN string should have length of 18 characters including '0x'";
    }

    my $fbn_msb_str = substr($fbn, 2,  8);
    my $fbn_lsb_str = substr($fbn, 10, 8);

    if ($self->{debug}) {
        $Log->comment("Checksum: set_FBN:: FBN values: $fbn_msb_str:: $fbn_lsb_str");
    }

    $self->{checksum}->{ckse_FBN_LSB} = hex($fbn_lsb_str);
    $self->{checksum}->{ckse_FBN_MSB} = hex($fbn_msb_str);
} ## end sub set_FBN

##--------------------------------------------------------
## Subroutine : set_VBN
##--------------------------------------------------------
##   @name             set_VBN
##   @synopsis         $checksum_obj->set_VBN(vbn => $value)
##
##   @description      Updates the checksum object's VBN field.
##
##   @arg  VBN         (Required) The value of VBN to be set
##                     it Should be a string of hex character of 64 bit. For e.g 0xABCD1123EFDD4498
##                     String length should be of 18 characters
##
##   @return           None
##--------------------------------------------------------
sub set_VBN {

    my $self = shift;
    my (%opts) = validate(@_, {vbn => {type => SCALAR},});

    my $vbn = $opts{vbn};

    if ($vbn =~ /(\+)|(-)/) {

        #Value contains + or - sign, which means we have to add given value to existing vbn value
        $self->add_to_VBN(value => $vbn);
        return;
    }

    if ($vbn !~ /^0x/) {
        $vbn = uc sprintf("%016x", $vbn);
        $vbn = "0x" . $vbn;
        if ($self->{debug}) {
            $Log->comment("Checksum::set_VBN: VBN  $vbn");
        }
    }

    my @chars = split(//, $vbn);
    if (scalar(@chars) != 18) {
        die
          "ArgumentInvalidException: Input VBN string should have length of 18 characters including '0x'";
    }

    my $vbn_msb_str = substr($vbn, 2,  8);
    my $vbn_lsb_str = substr($vbn, 10, 8);

    $self->{checksum}->{ckse_VBN_LSB} = hex($vbn_lsb_str);
    $self->{checksum}->{ckse_VBN_MSB} = hex($vbn_msb_str);
} ## end sub set_VBN

##--------------------------------------------------------
## Subroutine : add_to_FBN
##--------------------------------------------------------
##   @name             add_to_FBN
##   @synopsis         $checksum_obj->add_to_FBN(value =>$value)
##
##   @description      Performs addition or substraction to object's FBN value.
##
##   @arg  value       (Required) The value which is to be added to FBN (could be +ve or -ve)
##
##   @return            None
##--------------------------------------------------------
sub add_to_FBN {
    my $self = shift;
    my (%opts) = validate(@_, {value => {type => SCALAR},});

    my $val = $opts{value};

    my $big_FBN = Math::BigInt->new($self->get_FBN());
    if (!defined($big_FBN)) {
        die("Checksum: add_to_FBN:: could not create BigInt object");
    }

    $big_FBN->badd($val);
    my $hex_val = uc sprintf("%016x", $big_FBN);
    if ($self->{debug}) {
        $Log->comment("Checksum: add_to_FBN:: after add, fbn value: $big_FBN");
    }
    $self->set_FBN(fbn => "0x$hex_val");
} ## end sub add_to_FBN

##--------------------------------------------------------
## Subroutine : add_to_VBN
##--------------------------------------------------------
##   @name             add_to_VBN
##   @synopsis         $checksum_obj->add_to_VBN(value => $value)
##
##   @description      Performs addition or substraction to object's VBN value.
##
##   @arg  value       (Required) The value which is to be added to VBN (could be +ve or -ve)
##
##   @return           None
##--------------------------------------------------------
sub add_to_VBN {
    my $self = shift;
    my (%opts) = validate(@_, {value => {type => SCALAR},});

    my $val = $opts{value};

    my $big_VBN = Math::BigInt->new($self->get_VBN());

    $big_VBN->badd($val);
    my $hex_val = uc sprintf("%016x", $big_VBN);
    $self->set_VBN(vbn => "0x$hex_val");
} ## end sub add_to_VBN

##--------------------------------------------------------
## Subroutine : add_to_cpcount
##--------------------------------------------------------
##   @name             add_to_cpcount
##   @synopsis         $checksum_obj->add_to_cpcount(value =>$value)
##
##   @description      Performs addition or substraction to object's cpcount value.
##
##   @arg  value       (Required) The value which is to be added to cpcount (could be +ve or -ve)
##
##   @return            None
##--------------------------------------------------------
sub add_to_cpcount {
    my $self = shift;
    my (%opts) = validate(@_, {value => {type => SCALAR},});

    my $val = $opts{value};

#    my $big_cpcount = Math::BigInt->new($self->get_cpcount());
#    if (!defined($big_FBN)) {
#        die("Checksum: add_to_FBN:: could not create BigInt object");
#    }
#
#    $big_cpcount->badd($val);
#    my $hex_val = uc sprintf("%016x", $big_cpcount);
#    if ($self->{debug}) {
#        $Log->comment("Checksum: add_to_cpcount:: after add, fbn value: $big_cpcount");
#    }
    $self->set_cpcount(cpcount => "$val");
} ## end sub add_to_cpcount

##--------------------------------------------------------
## Subroutine : set_checksum_version
##--------------------------------------------------------
##   @name             set_checksum_version
##   @synopsis         $checksum_obj->set_checksum_version(ckecksum_flags => $value)
##
##   @description      Updates the checksum object's checksum_version field.
##
##   @arg  checksum_version      (Required) The value of checksum_version to be set
##
##   @return           None
##--------------------------------------------------------
sub set_checksum_version {
    my $self = shift;
    my (%opts) = validate(@_, {version => {type => SCALAR},});

    my $version = $opts{version};
    $self->{checksum}->{ckse_version} = $version;
}

##--------------------------------------------------------
## Subroutine : set_checksum_flags
##--------------------------------------------------------
##   @name              set_checksum_flags
##   @synopsis          $checksum_obj->set_checksum_flags(ckecksum_flags => $value)
##
##   @description       Updates the checksum object's ckecksum_flags field.
##
##   @arg  ckecksum_flags      (Required) The value of ckecksum_flags to be set
##
##   @return            None
##--------------------------------------------------------
sub set_checksum_flags {
    my $self = shift;
    my (%opts) = validate(@_, {ckecksum_flags => {type => SCALAR},});

    my $flags = $opts{ckecksum_flags};
    $self->{checksum}->{ckse_flags} = $flags;

}

##--------------------------------------------------------
## Subroutine : set_cpcount
##--------------------------------------------------------
##   @name             set_cpcount
##   @synopsis         $checksum_obj->set_cpcount(cpcount => $value)
##
##   @description      Updates the checksum object's cpcount field.
##
##   @arg  cpcount     (Required) The value of cpcount to be set
##
##   @return           None
##--------------------------------------------------------
sub set_cpcount {
    my $self = shift;
    my (%opts) = validate(@_, {cpcount => {type => SCALAR},});
    my $cpcount = $opts{cpcount};
    my $current_cpcount = $self->{checksum}->{ckse_cpcount};
    my $new_cpcount;

    if($cpcount =~ /0x/) {
        if($cpcount =~ /\+|-/) {
            my $neg = ($cpcount =~ /^-/) ? -1 : 1 ;
            $cpcount =~ s/\+|-//;
            $cpcount = hex($cpcount) * $neg;
            $new_cpcount = $current_cpcount + $cpcount;
        } else {
            $cpcount = hex($cpcount);
        }
    } elsif($cpcount =~ /\+|-/) {
        $new_cpcount = $current_cpcount + $cpcount;
    } else {
        $new_cpcount = $cpcount;
    }

    $self->{checksum}->{ckse_cpcount} = $new_cpcount;
}

##--------------------------------------------------------
## Subroutine : set_buftreeid
##--------------------------------------------------------
##   @name              set_buftreeid
##   @synopsis          $checksum_obj->set_buftreeid(buftreeid => $value)
##
##   @description       Updates the checksum object's buftreeid field.
##
##   @arg  buftreeid    (Required) The value of buftreeid to be set.
##
##   @return            None
##--------------------------------------------------------
sub set_buftreeid {
    my $self = shift;
    my (%opts) = validate(@_, {buftreeid => {type => SCALAR},});

    my $buftreeid = $opts{buftreeid};

    $self->{checksum}->{ckse_buftreeid} = $buftreeid;
}

##--------------------------------------------------------
## Subroutine : set_data_4k
##--------------------------------------------------------
##   @name             set_data_4k
##   @synopsis         $checksum_obj->set_data(data_4k => string value)
##
##   @description      Updates the string value of 4k data.
##
##   @arg  data_4k (Required) The value of 4k data block as string value to be set
##
##   @return           None
##--------------------------------------------------------
sub set_data_4k {
    my $self = shift;
    my (%opts) = validate(@_, {data_4k => {type => SCALAR},});

    my $data_str    = $opts{data_4k};
    my $index_end = ($self->{disk_block_type} eq "data") ? 4095 : 4031;
    my @data_as_hex = ();

    my @arr = split(//, $data_str);
    my $index = 0;
    for (; $index < scalar(@arr) ; $index++) {
        push(@data_as_hex, ord($arr[$index]));
    }
    @data_as_hex = @data_as_hex[0 .. $index_end];
    for (my $i = $index ; $i <= $index_end ; $i++) {
        $data_as_hex[$i] = 0;
    }

    $self->{checksum}->{data_4k} = [@data_as_hex];
} ## end sub set_data_4k

##--------------------------------------------------------
## Subroutine : get_data_4k
##--------------------------------------------------------
##   @name             get_data_4k
##   @synopsis         $checksum_obj->get_data()
##
##   @description      return the string value of 4k data.
##
##   @return           String format of 4k data
##--------------------------------------------------------
sub get_data_4k {
    my $self = shift;

    my @data_4k = @{$self->{checksum}->{data_4k}};

    my $data_4k_str = '';
    for (my $i = 0 ; $i < scalar(@data_4k) ; $i++) {
        $data_4k_str = $data_4k_str . chr($data_4k[$i]);
    }

    return $data_4k_str;
} ## end sub get_data_4k

######################################################## INTERNAL FUNCTIONS ######################################################

##--------------------------------------------------------
## Subroutine : _compute_checksum
##--------------------------------------------------------
##   @name              _compute_checksum
##   @synopsis          $checksum_obj->_compute_checksum(dd_block=>$dd_block, type=>$type)
##
##   @description       This function calculates the checksum of specified type and returns the value
##
##   @arg  dd_block     A reference to array which contains the block data byte by byte
##                      in little endian format, default is object data block
##   @arg  type         The type of checksum to be calculated, default is embedded
##
##   @return            The checksum value in decimal format
##--------------------------------------------------------
sub _compute_checksum {
    my $self = shift;

    my (%opts) =
      validate(@_,
               {dd_block => {type => ARRAYREF, default => $self->get_dd_block(), optional => 1},
                type     => {type => SCALAR,   default => 'embedded',            optional => 1},});

    my $dd_block = $opts{dd_block};
    my $type     = $opts{type};

    my $disk_block_type = $self->{disk_block_type};
    if($disk_block_type eq "checksum" and $type eq "data") {
        $Log->warn("Checksum::_compute_checksum: AZCS checksum blocks do not have data checksum");
        return;
    }

    my $cksum_offset = $self->{checksum_offset};
    my ($_ref_8bytes, @_8bytes);

    #------------------------- INIT VARIABLES -----------------
    my $base  = Math::BigInt->new('0xfff1');
    my $sum1l = Math::BigInt->new('1');
    my $sum2l = Math::BigInt->new('0');
    my $sum1  = Math::BigInt->new('0');
    my $sum2  = Math::BigInt->new('0');

    #----------------------------------------------------------

    #-----------------------------------------------------------
    if ($type =~ /data/i) {
        my @orig_block = @$dd_block;
        my @data = splice(@orig_block, 0, $cksum_offset);   #This should only for data blocks and not checksum
        $_ref_8bytes = $self->_get_8bytes(dd_block => \@data);
        @_8bytes = @$_ref_8bytes;
    } elsif ($type =~ /embedded/i) {
        #set embedded checksum fields to zero;
        my @orig_block = @$dd_block;
        my @checksum = splice(@orig_block, $cksum_offset, 64);
        foreach my $index (12 .. 15) {
            $checksum[$index] = 0;
        }
        $_ref_8bytes = $self->_get_8bytes(dd_block => \@checksum);
        @_8bytes = @$_ref_8bytes;
    } ## end elsif ($type =~ /embedded/i)

    #--------------------------------------------------------------

    #--------------------------------------------------------------
    # DO1
    #--------------------------------------------------------------
    my $len = scalar(@_8bytes);
    for (my $index = 0 ; $index < $len ; $index++) {
        my $bval = Math::BigInt->new($_8bytes[$index]);
        $sum1l->badd($bval);
        $sum2l->badd($sum1l);
        if ($sum1l->is_nan() or $sum2l->is_nan()) {
            die "ArgumentInvalidException: $_8bytes[$index] is not valid hex 8 byte string";
        }
    }
    #--------------------------------------------------------------

    #--------------------------------------------------------------
    # SLOW MOD
    #--------------------------------------------------------------
    ($sum1l, $sum1) = ($sum1l->bdiv($base));
    ($sum2l, $sum2) = ($sum2l->bdiv($base));

    if ($self->{debug}) {
        $Log->comment("Checksum::_compute_checksum: sum1: $sum1\t sum2 = $sum2");
    }

    #--------------------------------------------------------------

    #---------------------------------------------------------------
    # Return the computed checksum into decimal format
    #---------------------------------------------------------------
    my $shift = 16;
    $sum2->blsft($shift);
    $sum2->bior($sum1);
    my $cksum     = $sum2;
    my $cksum_val = int($cksum);

    #convert $cksum_val to hex string
    my $cksum_str = sprintf("%#010x", $cksum_val);
    if($self->{debug}) {
        $Log->comment("Checksum::_compute_checksum: Computed checksum is $cksum_str");
    }
    return hex($cksum_str);

    #---------------------------------------------------------------
} ## end sub _compute_checksum

##--------------------------------------------------------
## Subroutine : _calculate_gen_count
##--------------------------------------------------------
##   @name              _calculate_gen_count
##   @synopsis          INTERNAL_FUNCTION : should not be called
##
##   @description       parses the gen_count like +1, -2 etc and gives the new gen_count to be set
##
##   @arg gen_count     (Required) gen_count to be parsed (possible values: +/-{1,2} or (0,1,2.3))
##   @arg current_gen_count    (Required)
##
##   @return new_gen_count    This is the parsed value of the gen count which was passed as argument to this function
##--------------------------------------------------------
sub _calculate_gen_count {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {gen_count         => {type => SCALAR},
                current_gen_count => {type => SCALAR | UNDEF, default => undef, optional => 1}});
    my $gen_count         = $opts{gen_count};
    my $current_gen_count = $opts{current_gen_count};
    my $new_gen_count     = 0;

    #-----------------------------------------------------------------
    # For gen_count +1 or +2
    #-----------------------------------------------------------------
    my $mod_value = 4;
    if ($gen_count =~ /^\+{1}([1|2])$/) {
        if (!defined($current_gen_count)) {
            if ($self->{debug}) {
                $Log->comment("Checksum::_calculate_gen_count: $gen_count $current_gen_count");
            }
            die "ArgumentInvalidException: Current gen count is not passed to a function";
        } else {
            if (($current_gen_count + $1) >= 4) {
                $new_gen_count = ($current_gen_count + $1) % $mod_value;
                $new_gen_count++;
            } else {
                $new_gen_count = ($current_gen_count + $1) % $mod_value;
            }
        }

        #-----------------------------------------------------------------
        # For gen_count -1 or -2
        #-----------------------------------------------------------------
    } elsif ($gen_count =~ /^-{1}([1|2])$/) {
        if (!defined($current_gen_count)) {
            if ($self->{debug}) {
                $Log->comment("Checksum::_calculate_gen_count: $gen_count $current_gen_count");
            }
            die "ArgumentInvalidException: Current gen count is not passed to a function";
        } else {
            if (($current_gen_count - $1) <= 0) {
                $new_gen_count = ($current_gen_count - $1 + $mod_value - 1) % $mod_value;
            } else {
                $new_gen_count = ($current_gen_count - $1) % $mod_value;
            }
        }

        #-----------------------------------------------------------------
        # For gen_count 0,1,2,3
        #-----------------------------------------------------------------
    } elsif ($gen_count =~ /^[0|1|2|3]$/) {
        $new_gen_count = $gen_count;
    } else {
        if ($self->{debug}) {
            $Log->comment("Checksum::_calculate_gen_count: $gen_count $current_gen_count");
        }
        die "ArgumentInvalidException: Invalid gen_count is sent to a function $gen_count";
    }

    #-----------------------------------------------------------------
    return $new_gen_count;
} ## end sub _calculate_gen_count

##--------------------------------------------------------
## Subroutine : _get_8bytes
##--------------------------------------------------------
##   @name              _get_8bytes
##   @synopsis          INTERNAL_FUNCTION : should not be called by external env.
##
##   @description       This function convert little endian bytes to big endian 8 byte quad word
##
##   @arg dd_block      reference to array containing bytes in little endian format
##
##   @return _8bytes    a reference to array containing 8byte strings in hex format (big endian)
##--------------------------------------------------------
sub _get_8bytes {
    my $self = shift;
    my (%opts) = validate(@_, {dd_block => {type => ARRAYREF},});
    my $dd_block  = $opts{dd_block};
    my @_dd_block = @$dd_block;
    my $len       = scalar(@_dd_block);

    #---------------------------------------------------------------
    # Zero padding if the len of input arr is not in multiple of 8
    #---------------------------------------------------------------
    if ($len % 8) {
        $Log->comment(
            "Checksum::_get_8bytes: Zero padding if the len of input dd_block is not in multiple of 8"
        );
        my $index = $len % 8;
        foreach (1 .. $index) {
            push(@_dd_block, 0);
        }
        $len += $index;
    } ## end if ($len % 8)

    #--------------------------------------------------------------------------
    #Convert 0x01 23 45 67 89 ab cd ef to 0xefcdab8967452301
    #$_bytes[$index+3] <<24 ) | ($_bytes[$index+2] << 16) | ($_bytes[$index+1] <<8) | ($_bytes[$index])
    #--------------------------------------------------------------------------
    my @_8bytes = ();
    for (my $index = 0 ; $index < $len ; $index += 8) {
        my $_4byte1 =
          ($_dd_block[$index + 7] << 24) | ($_dd_block[$index + 6] << 16) |
          ($_dd_block[$index + 5] << 8) | $_dd_block[$index + 4];
        my $_4byte2 =
          ($_dd_block[$index + 3] << 24) | ($_dd_block[$index + 2] << 16) |
          ($_dd_block[$index + 1] << 8) | $_dd_block[$index];
        my $_8byte;

        #convert VBN 4 bytes LSB and MSB words and create a hex string of 64bit
        if ($_4byte1 == 0) {
            $_8byte = "0x00000000" . sprintf("%08x", $_4byte2);
        } elsif ($_4byte1) {
            $_8byte = sprintf("%#010x", $_4byte1) . sprintf("%08x", $_4byte2);
        }
        push(@_8bytes, $_8byte);
    } ## end for (my $index = 0 ; $index...

    return \@_8bytes;
} ## end sub _get_8bytes

##--------------------------------------------------------
## Package : RaidConfig
##--------------------------------------------------------
##   @name              RaidConfig
##   @synopsis          new RaidConfig(host => $host, conntype => $conntype)
##
##   @description       This package defines subroutine to get irad_config parameter values for disk, group and aggr
##
##   @routines
##
##          refresh
##          get_disk_rgdn
##          get_disk_type
##          get_group_scrub_progress
##          get_group_scrub_suspended_block
##          get_group_raid_lw_scrub
##          get_aggr_raid_lw_enabled
##          get_aggr_raid_lw_allowed
##
##--------------------------------------------------------
package RaidConfig;
use strict;
use warnings;
use RaidInfo;
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF UNDEF);
use Tharn qw(say script_result host $Log logresult param);
use Params::Validate qw(validate OBJECT SCALAR ARRAYREF HASHREF UNDEF);

BEGIN {
    Params::Validate::validation_options(
        on_fail => sub {
            my @msgs = @_;
            chomp(@msgs);
            Carp::confess(Tharn::script_result(@msgs));
        });
}

use Utils qw(make_conn);

##--------------------------------------------------------
## Subroutine : new
##--------------------------------------------------------
##   @name             new
##   @synopsis         new RaidConfig(host => $host, conntype => $conntype)
##
##   @description      This subroutine creates an object for RaidConfig class
##
##   @arg host         (Required): Hostspec to filer.
##   @arg conntype     (Optional): Connection type to use. Default is "#rsh". For filermode "MAINT" it is "#console" by default.
##
##   @return           This returns a object for RaidConfig class
##--------------------------------------------------------
sub new {
    if (scalar(@_) % 2) {
        shift;
    }
    my (%opts) =
      validate(@_, {
        api_set_obj     => {type => OBJECT, isa => 'NACL::APISet'},
        host            => {type => OBJECT, isa => 'Hostrec'},
        conntype        => {type => SCALAR},
    });

    my %raid_config = ();
    $raid_config{host}     = $opts{host};
    $raid_config{conntype} = $opts{conntype};

    my $conn = make_conn($raid_config{host}, $raid_config{conntype});
    my $raid_obj = new RaidInfo($conn, api_set_obj => $opts{api_set_obj});
    $raid_config{raid_obj} = $raid_obj;
    my $ref_raid_config = \%raid_config;

    #bless
    bless $ref_raid_config;
    return $ref_raid_config;
} ## end sub new

##--------------------------------------------------------
## Subroutine : refresh
##--------------------------------------------------------
##   @name            refresh
##   @synopsis        $raid_config->refresh()
##
##   @description     This function updates the raid_config object.
##
##   @arg             None
##
##   @return          None
##--------------------------------------------------------
sub refresh {
    my $self = shift;
    $self->{raid_obj}->refresh();
}

##--------------------------------------------------------
## Subroutine : get_disk_rgdn
##--------------------------------------------------------
##   @name             get_disk_rgdn
##   @synopsis         $raid_config->get_disk_rgdn(disk_name=>  $disk_name,refresh => 1);
##
##   @description      This function gets the value for parameter disk_rgdn.
##
##   @arg  disk_name   (required) Name of disk for which parameter value to be extracted.
##   @arg  refresh     (optional) if set to 1 ;updates the raid_config object.
##
##   @return           Returns the value for disk_rgdn parameter
##--------------------------------------------------------
sub get_disk_rgdn {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {disk_name => {type => SCALAR},
                refresh => {type => SCALAR, default => 1, optional => 1}});
    my $disk_name = $opts{disk_name};
    my $refresh   = $opts{refresh};

    if ($refresh) {
        $self->{raid_obj}->refresh();
    }

    my $disk_obj = $self->{raid_obj}->disk($disk_name);
    if (!defined($disk_obj)) {
        my $disk_path_name2 = $self->get_disk_active_path(disk_name => $disk_name);
        my $disk_obj_path2 = $self->{raid_obj}->disk($disk_path_name2);
        if (!defined($disk_obj_path2)) {
            die "DiskInvalidException: Disk $disk_name and $disk_path_name2 does not exist";
        } else {
            return ($disk_obj_path2->parameter('disk_rgdn'));
        }
    } else {
        return ($disk_obj->parameter('disk_rgdn'));
    }
} ## end sub get_disk_rgdn

##--------------------------------------------------------
## Subroutine : get_disk_type
##--------------------------------------------------------
##   @name              get_disk_type
##   @synopsis          $raid_config->get_disk_type(disk_name =>  $disk_name,refresh => 1);
##
##   @description       This function gets the value for parameter raid_type of a disk.
##
##   @arg  disk_name    (required) Name of disk for which parameter value to be extracted.
##   @arg  refresh      (optional) if set to 1 ;updates the raid_config object.
##
##   @return            Returns the value for disk_type parameter
##--------------------------------------------------------
sub get_disk_type {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {disk_name => {type => SCALAR},
                refresh => {type => SCALAR, default => 1, optional => 1}});
    my $disk_name = $opts{disk_name};
    my $refresh   = $opts{refresh};

    if ($refresh) {
        $self->{raid_obj}->refresh();
    }

    my $disk_obj = $self->{raid_obj}->disk($disk_name);
    
    
    
    if (!defined($disk_obj)) {
        my $disk_path_name2 = $self->get_disk_active_path(disk_name => $disk_name);
        my $disk_obj_path2 = $self->{raid_obj}->disk($disk_path_name2);
        if (!defined($disk_obj_path2)) {
            die "DiskInvalidException: Disk $disk_name and $disk_path_name2 does not exist";
        } else {
            return ($disk_obj_path2->parameter('raid_type'));
        }
    } else {
        return ($disk_obj->parameter('raid_type'));
    }
} ## end sub get_disk_type


##--------------------------------------------------------
## Subroutine : get_disk_active_path
##--------------------------------------------------------
##   @name             get_disk_active_path
##   @synopsis         $raid_config->get_disk_rgdn(disk_name=>  $disk_name,refresh => 1);
##
##   @description      This function gets the value for parameter path_name2.
##
##   @arg  disk_name   (required) Name of disk for which parameter value to be extracted.
##   @arg  refresh     (optional) if set to 1 ;updates the raid_config object.
##
##   @return           Returns the value for path_name2 parameter
##--------------------------------------------------------
sub get_disk_active_path {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {disk_name => {type => SCALAR},
                refresh => {type => SCALAR, default => 1, optional => 1}});
                
    my $disk_name = $opts{disk_name};
    my $refresh   = $opts{refresh};

    if ($refresh) {
        $self->{raid_obj}->refresh();
    }

    my %disks_hash = $self->{raid_obj}->all_disks();
    
    foreach my $key (keys %disks_hash) {
        if ($disk_name eq $disks_hash{$key}{"path_name2"}) {
            my $disk_name_path2 = $disks_hash{$key}{"name"};
            $Log->comment("get_disk_active_path: The disk name for $disk_name seems to have changed to $disk_name_path2 after reboot.");
            return $disk_name_path2;
        }
    }

} ## end sub get_disk_active_path



##--------------------------------------------------------
## Subroutine : get_group_scrub_progress
##--------------------------------------------------------
##   @name             get_group_scrub_progress
##   @synopsis         $raid_config->get_group_scrub_progress(group_name=>  $group_name,refresh=>  1);
##
##   @description      This function gets the value for parameter scrub_progress.
##
##   @arg  group_name  (required) Name of group from which parameter value to be extracted.
##   @arg  refresh     (optional) if set to 1 ;updates the raid_config object.
##
##   @return           Returns the value for scrub_progress parameter
##--------------------------------------------------------
sub get_group_scrub_progress {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {group_name => {type => SCALAR},
                refresh => {type => SCALAR, default => 1, optional => 1},});

    my $group_name = $opts{group_name};
    my $refresh    = $opts{refresh};

    if ($refresh) {
        $self->{raid_obj}->refresh();
    }
    my $group_obj = $self->{raid_obj}->raid_group($group_name);

    if (!defined($group_obj)) {
        die "GroupInvalidException: Group name $group_name does not exist";
    } else {
        return ($group_obj->parameter('scrub_progress'));
    }
} ## end sub get_group_scrub_progress

##--------------------------------------------------------
## Subroutine : get_group_scrub_suspended_block
##--------------------------------------------------------
##   @name             get_group_scrub_suspended_block
##   @synopsis         $raid_config->get_group_scrub_suspended_block(group_name =>  $group_name,refresh => 1);
##
##   @description      This function gets the value for parameter scrub_suspended_block.
##
##   @arg  group_name  (required) Name of group from which parameter value to be extracted.
##   @arg  refresh     (optional) if set to 1 ;updates the raid_config object.
##
##   @return           Returns the value for scrub_suspended_block parameter
##--------------------------------------------------------
sub get_group_scrub_suspended_block {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {group_name => {type => SCALAR},
                refresh => {type => SCALAR, default => 1, optional => 1},});

    my $group_name = $opts{group_name};
    my $refresh    = $opts{refresh};

    if ($refresh) {
        $self->{raid_obj}->refresh();
    }
    my $group_obj = $self->{raid_obj}->raid_group($group_name);

    if (!defined($group_obj)) {
        die "GroupInvalidException: Group name $group_name does not exist";
    } else {
        return ($group_obj->parameter('scrub_suspended_block'));
    }
} ## end sub get_group_scrub_suspended_block

##--------------------------------------------------------
## Subroutine : get_group_raid_lw_scrub
##--------------------------------------------------------
##   @name             get_group_raid_lw_scrub
##   @synopsis         $raid_config->get_group_raid_lw_scrub(group_name => $group_name,refresh => 1);
##
##   @description      This function gets the value for parameter group_raid_lw_scrub.
##
##   @arg  group_name  (required) Name of group from which parameter value to be extracted.
##   @arg  refresh     (optional) if set to 1 ;updates the raid_config object.
##
##   @return           Returns the value for group_raid_lw_scrub parameter
##--------------------------------------------------------
sub get_group_raid_lw_scrub {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {group_name => {type => SCALAR},
                refresh => {type => SCALAR, default => 1, optional => 1},});

    my $group_name = $opts{group_name};
    my $refresh    = $opts{refresh};

    if ($refresh) {
        $self->{raid_obj}->refresh();
    }
    my $group_obj = $self->{raid_obj}->raid_group($group_name);

    if (!defined($group_obj)) {
        die "GroupInvalidException: Group name $group_name does not exist";
    } else {
        return ($group_obj->parameter('group_raid_lw_scrub'));
    }
} ## end sub get_group_raid_lw_scrub

##--------------------------------------------------------
## Subroutine : get_aggr_raid_lw_enabled
##--------------------------------------------------------
##   @name             get_aggr_raid_lw_enabled
##   @synopsis         $raid_config->get_aggr_raid_lw_enabled(filer_vol => $filer_vol,refresh => 1);
##
##   @description      This function gets the value for parameter raid_lw_enabled.
##
##   @arg  filer_vol    (required) Name of volume from which parameter value to be extracted.
##   @arg  refresh     (optional) if set to 1 ;updates the raid_config object.
##
##   @return           Returns the value for raid_lw_enabled parameter
##--------------------------------------------------------
sub get_aggr_raid_lw_enabled {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {aggr_name => {type => SCALAR},
                refresh => {type => SCALAR, default => 1, optional => 1},});

    my $filer_vol = $opts{aggr_name};
    my $refresh  = $opts{refresh};

    if ($refresh) {
        $self->{raid_obj}->refresh();
    }
    my $group_obj = $self->{raid_obj}->aggregate($filer_vol);

    if (!defined($group_obj)) {
        die "AggregateInvalidException: Aggr of name $filer_vol does not exist";
    } else {
        return ($group_obj->parameter('raid_lw_enabled'));
    }
} ## end sub get_aggr_raid_lw_enabled

##--------------------------------------------------------
## Subroutine : get_aggr_raid_lw_allowed
##--------------------------------------------------------
##   @name             get_aggr_raid_lw_allowed
##   @synopsis         $raid_config->get_aggr_raid_lw_allowed(filer_vol => $filer_vol,refresh => 1);
##
##   @description      This function gets the value for parameter raid_lw_allowed.
##
##   @arg  filer_vol    (required) Name of volume from which parameter value to be extracted.
##   @arg  refresh     (optional) if set to 1 ;updates the raid_config object.
##
##   @return           Returns the value for raid_lw_allowed parameter
##--------------------------------------------------------
sub get_aggr_raid_lw_allowed {
    my $self = shift;
    my (%opts) =
      validate(@_,
               {aggr_name => {type => SCALAR},
                refresh => {type => SCALAR, default => 1, optional => 1},});

    my $filer_vol = $opts{aggr_name};
    my $refresh  = $opts{refresh};

    if ($refresh) {
        $self->{raid_obj}->refresh();
    }
    my $group_obj = $self->{raid_obj}->aggregate($filer_vol);

    if (!defined($group_obj)) {
        die "AggregateInvalidException: Aggr of name $filer_vol does not exist";
    } else {
        return ($group_obj->parameter('raid_lw_allowed'));
    }
} ## end sub get_aggr_raid_lw_allowed

##--------------------------------------------------------
## Subroutine : get_gen_count
##--------------------------------------------------------
##   @name             get_gen_count
##   @synopsis         $checksum_obj->get_gen_count()
##
##   @description      Returns the generation count/disk rgdn of a disk and dbn for which object was created.
##   @arg disk_rgdn    string representing rgdn number for a disk whose gen_count is required from parity disk for 
##                     which object is created.
##   @arg isparity     Boolean.  
##                     0 if gen count of data disk is returned.
##                     1 if gen count for data disk mentioned by disk_rgdn in parity disk is returned.
##
##   @return           gen_count of a disk or the gen_count for data disk mentioned by disk_rgdn 
##--------------------------------------------------------

sub get_gen_count {
    my $self = shift; 
    my (%opts) = validate(@_, {disk_rgdn => {type => SCALAR, default => undef, optional => 1} ,
                               isparity => {type => SCALAR, default => 0},});
    my $isparity = $opts{isparity};
    my $rgdn     = $opts{disk_rgdn};
    my $disk_type = $self->get(key => "disk_type");
    if (!$isparity) {
        if ($disk_type!~ /data/i ) {
            if ($self->{debug}) {
                $Log->comment("Checksum::get_gen_count: Raid_type is " . $disk_type);
            }
            die "DiskRaidTypeInvalid: This function is defined for data disks only: $disk_type";
        }
        return $self->{checksum}->{ckse_lw_id}->{gen_count};
    } else {
        if (!defined($rgdn)) {
            die "UndefinedDiskRgdn: RaidLW::get_gen_count : rgdn does not exist";
        } else {
            if ($disk_type!~ /parity/i) {
                if ($self->{debug}) {
                    $Log->comment("Checksum::get_parity_gen_count: Raid_type is " . $disk_type);
                }
                die
                    "DiskRaidTypeInvalid: This function is defined for parity or dparity disks only: $self->{disk_type}";
            }
            return $self->{checksum}->{ckse_lw_id}->{gen_bitmap}->[$rgdn];
        }
    }
} ## end get_gen_count

1;

