# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/RegistryCmds.pm#1 $

## @summary Library of registry and options commands
##
## @description
##    This package defines routines which use the filer's "registry" command
##    and "options" command.  The get, parse, return, and set registry
##    information and filer options.
##
## @author dl-dpgqa-automation@netapp.com
## @cf     false
## @status Shared

package RegistryCmds;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(registryGet registrySet registryWalk
	     registryDelete registryDeltree
	     getOption setOption walkOptions);

use SayWrap;
use Tharn;
use strict;

## @name ::registryGet
## @usage registryGet($filer, $key)
## @synopsis
##    Get the registry value for the given key and return it.
##    Returns undef if the key did not exist or had no value.
##    On error, returns undef with error message in $@.
sub registryGet ($$) {
  my ($filer, $key) = @_;
  my $keyexp = quotemeta($key);

  my $text = sayEOF($filer, "priv set -q advanced ; registry get $key");

  return eval {
    return undef if ($text =~ /$keyexp does not exist\./);
    die "$text\n" unless ($text =~ /^$keyexp=(.*)$/s);
    return ($1 ? $1 : undef);
  };
}

## @name ::registrySet
## @usage registrySet($filer, $key, $value)
## @synopsis
##    Set the registry value for the given key.
##    Returns success or failure, with error message in $@ on failure.
sub registrySet ($$$) {
  my ($filer, $key, $value) = @_;

  my $text = sayEOF($filer, "priv set -q advanced ; registry set $key \"$value\"");

  return eval {
    die "$text\n" if ($text =~ /An error occurred while trying to commit the registry/);

    # make sure it was set in the registry. ONTAP does nothing if key was
    # not set.
    my $val = registryGet($filer, $key);
    if (!$@ && defined($val) && ($val=~/$value/)) {
	return 1;
    } else {
	die "key '$key' was not set in the registry.\n";
    }
  };
}

## @name ::registryDelete
## @usage registryDelete($filer, $key)
## @synopsis
##    Delete the given key and its value from the registry.
##    Returns success or failure, with error message in $@ on failure.
sub registryDelete ($$) {
  my ($filer, $key) = @_;
  my $keyexp = quotemeta($key);

  my $text = sayEOF($filer, "priv set -q advanced ; registry delete $key");

  return eval {
    $text =~ s/^$keyexp deleted\.\s*$//;
    die "$text\n" if ($text);
    return 1;
  };
}

## @name ::registryDeltree
## @usage registryDeltree($filer, $prefix)
## @synopsis
##    Delete all keys beginning with the specified prefix from the registry.
##    Returns success or failure, with error message in $@ on failure.
sub registryDeltree ($$) {
  my ($filer, $prefix) = @_;
  my $prefexp = quotemeta($prefix);

  my $text = sayEOF($filer, "priv set -q advanced ; registry deltree $prefix");

  return eval {
    $text =~ s/^Everything under $prefexp is deleted\.\s*$//;
    die "$text\n" if ($text);
    return 1;
  };
}

## @name ::registryWalk
## @usage registryWalk($filer, $prefix)
## @synopsis
##    Walks the registry looking for keys that start with the given prefix.
##    Returns a reference to a hash of encountered keys to their values.
##    Returns undef if no keys with the given prefix exist.
##    On error, returns undef with error message in $@.
sub registryWalk ($$) {
  my ($filer, $prefix) = @_;
  my $prefexp = quotemeta($prefix);
  my $text = sayEOF($filer, "priv set -q advanced ; registry walk $prefix");

  return eval {
    my %ret = ();
    my ($curkey, $curval);
    chomp $text;
    for my $line ( split /\n/, $text  ) {
	if ( $line =~ /^(\S*)\=(.*)$/s ) {
	    ($curkey,$curval) = ($1,$2);
	    $ret{$curkey} = $curval;
	} else {
	    $ret{$curkey} .= "\n$line" if ( $curkey );
	}
    }
    return ( (scalar(keys %ret) > 0) ? \%ret : undef );
  };
}

## @name ::getOption
## @usage getOption($filer, $opt)
## @synopsis
##    Get the value for the given option and returns it.
##    Returns undef if the option did not exist or had no value.
##    On error, returns undef with error message in $@.
sub getOption ($$) {
  my ($filer, $opt) = @_;
  my $optexp = quotemeta($opt);

  my $text = sayEOF($filer, "options $opt");

  return eval {
    die "$text\n"
      unless ($text =~ /^$optexp\s+(.*?)\s*$/);
    return ($1 ? $1 : undef);
  };
}

## @name ::setOption
## @usage setOption($filer, $opt, $value)
## @synopsis
##    Set the value for the given option.
##    Returns success or failure, with error message in $@ on failure.
sub setOption ($$$) {
  my ($filer, $opt, $value) = @_;

  my $text = sayEOF($filer, "options $opt \"$value\"");

  return eval {
    $text =~ s/^\s*$//;
    die "$text\n" if ($text);
    return 1;
  };
}

## @name ::walkOptions
## @usage walkOptions($filer, $prefix)
## @synopsis
##    Gets all options that start with the given prefix.
##    Returns a reference to a hash of the options to their values.
##    Returns undef if no options beginning with the prefix exist.
##    On error, returns undef with error message in $@.
sub walkOptions ($$) {
  my ($filer, $prefix) = @_;
  my $prefexp = quotemeta($prefix);

  my $text = sayEOF($filer, "options $prefix");

  return eval {
    return undef if ($text =~ /No such option $prefexp/);
    my @lines = split(/\n/, $text);
    @lines = grep !/^\s*$/, @lines;
    my %ret = map { die "Error: $_\n" unless ($_ =~ /^\s*(\S+)\s+(\S*)\s*$/);
		    $1 => $2 } @lines;
    return ( (scalar(keys %ret) > 0) ? \%ret : undef );
  };
}
