#Hey emacs, this is basically -*- perl -*-
# $Id$

## @summary     Handles remote execution of a job/test via spindex or subtest
## @author      snider@netapp.com
## @pod here

=head1 NAME

RemoteExecution - Library to handle all remote execution on clients

=head1 SYNOPSIS

    use RemoteExecution;
    my $Remote = new RemoteExecution;
    @hosts = $Remote->check_clients(@hosts);
    foreach my $host (@hosts) {
      $Remote->execute(host=>host,
                       cmd=>"command to execute",
                       runid=>runid,
                       nate_params=>{},
                       handler=>\&routine);
    }
    my $result = $Remote->wait(timeout=>600);

=head1 DESCRIPTION

C<RemoteExecution> implements an object which is used to manage processes
on remote clients. The object will either use SpinDex or NATE's subtest in
order to execute the commands on the clients. The default is to use subtest
but can be changed by setting the param C<SPINDEX=1>.

The object implements routines to check the health of the clients, execute
commands on the clients (without blocking) and to wait on the jobs to finish.

=head1 METHODS

=cut

# the following chunk of code allows this module to be overridded with another
# one if the variable is set. If it is set then print a warning, delete the ENV
# variable so that it is not seen in the new module, load the new one, and
# set the flag that indicates that we have already loaded the module (so that
# we don't load the new one
my $LibraryLoaded;
if ($ENV{OVERRIDEREMOTEEXECUTION}) {
    my $newModule = $ENV{OVERRIDEREMOTEEXECUTION};
    delete $ENV{OVERLOADREMOTEEXECUTION};
    delete $ENV{OVERRIDEREMOTEEXECUTION};
    print "The main RemoteExecution module will be replaced with $newModule\n";
    require($newModule);
    $LibraryLoaded = 1;
}
return 1 if ($LibraryLoaded);

package RemoteExecution;

use strict;
use Module::Common qw(validate_mixed);
use Params::Validate qw(SCALAR HASHREF ARRAYREF OBJECT UNDEF CODEREF);
use Time::HiRes qw( gettimeofday tv_interval );
use Tharn;
use SayWrap qw(sayEOF);
use Utils qw(make_conn);
use Cwd;
use Data::Dumper;
use TracePkg;
use Text::ParseWords;
my $Trace = new TracePkg;
use Options;
my $Options;

$Options = new Options;
$Options->parameters(definitions=>
                     [
                      {
                       name=>"SPINDEX",
                       spec=>"s",
                       description=>"spindex should be used instead of subtest",
                      },
                      {
                       name=>"NATE_HOSTS",
                       spec=>"s",
                       description=>"the location of the nate_hosts files (needed for SpinDex)",
                      },
                      {
                       name=>"LOGDIR",
                       description=>"The location of the logs",
                      },
                      {
                       name=>"NATE_LIB",
                       description=>"The location of the scripts/libraries",
                      },
                      {
                       name=>"LOGNAME",
                       description=>"the name of the caller",
                      },
                      {
                       name=>"MAX_UPTIME",
                       description=>"The maximum uptime allowed for a client",
                      },
                      {
                       name=>"CHECK_CLIENT_HEALTH",
                       description=>"Flag indicating that the health of the clients should be checked prior to the test running",
                       default=>0,
                      },
                     ],
                    );
my $Spindex;

# Set a variable for which ntest to call. In the case of spindex, the same
# ntest that was used to invoke the current script should be used to start
# the remote tests
my $Ntest = "/usr/local/bin/ntest";
my $i = 0;
while (my @caller = caller($i++)) {
    if ($caller[0] eq "ntest") {
	$Ntest = $caller[1];
	last;
    }
}

=head2 new

    my $Remote = new RemoteExecution;

C<new> is the invocant for the RemoteExecution object. This method will
create the object and check the value of the C<SPINDEX> param to determine
whether spindex or subtest will be used. The method will call C<process> on
the Options object to insure that C<SPINDEX> has been processed.

=cut

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
    my $self = {};
    $Trace->entry();
    bless($self, $class);

    $Options->process(test_cases=>undef) if (!$Options->processed());
    if ($Options->get(option=>"SPINDEX")) {
        # When running with Spindex, TestManager has to pass all of the params that it knows
        # about to the remote test on the client. This is starting to overwhelm some clients,
        # particularly the windows clients. They seem to have a problem with a limited amount
        # of %ENV space. So this is the end of SpinDex.
        # 
        $Trace->fatal(msg=>"SPINDEX is no longer supported");
	$SIG{INT} = \&cleanup; # catch Ctrl-C to try and kill the remote processes.  Only do
	                       # this in SpinDex case as ignoring INT in subtest will lead to
	                       # problems like burt 199841
	eval "use SpinDexClient";
	if  ($@) {
	    print $@, "\n\n\n";
	    exit(1);
	}
	$Spindex = SpinDexClient->new();
    }
    $Options->process() if (!$Options->processed());
    # %parameters is used to track the processes that were attempted. This
    # will make a retry easier to attempt
    $self->{parameters} = {};
    $Trace->exit(msg=>"return $self");
    $self->{pid_list} = [];
    return $self;
}

my %Clients;

# This routine gets called for each client so scripts with many clients can overwhelm a log file.
#  For this reason, this call is no longer using tracing.
sub _check_client {
    my ($output, $client) = @_;
    my $max_uptime = $Options->get(option=>"MAX_UPTIME");
    if (defined($max_uptime) && $output =~ /load average: (\d+\.*\d+), (\d+\.*\d+),/) {
	my $uptime = $2;
	if ($uptime > $max_uptime) {
	    $Trace->warn(msg=>"Skipping client '$client' because the load ($uptime) is higher than the defined max ($max_uptime)\n");
	    delete $Clients{$client};
	}
    }
}


=head2 check_clients

    @hosts = $Remote->check_clients(@hosts);

C<check_clients> will attempt to run the C<uptime> command on all of the
clients. The method drops the clients that did not respond and returns a list 
of those clients that did respond.

You can unset the nate param C<CHECK_CLIENT_HEALTH> to have any script
that normally calls C<check_clients> to skip this step and return the full
list of clients without checking them.

Since C<check_clients> uses C<uptime> to check the client health, you can
also specify the nate param C<MAX_UPTIME> to drop any clients from the list
which have a current uptime higher than the value specified.

WARNING: this method does not support subtest yet.

=cut
  
sub check_clients {
    my ($self, @clients) = @_;
    $Trace->entry();
    map $Clients{$_}++, @clients;
    if ($Options->get(option=>"CHECK_CLIENT_HEALTH") && $Spindex) {
	$Spindex->{PrintFormat} = undef;
	foreach my $client (@clients) {
	    if (($Spindex->Rexec(Host => $client, 
				 PrintFormat => undef,
				 Wait => 0,
				 Command => "uptime", 
				 EonOut => \&_check_client)) < 0){
		$Trace->warn(msg=>"Skipping host '$client' because it cannot be reached\n");
		delete $Clients{$client};
	    }
	}
	$Spindex->WaitGroup(Timeout => 60);
	delete $Spindex->{PrintFormat};
    }
    
    # now that all of the processReset the pid and client lists since this
    # code is not using handle_results(). Not doing this will cause the 
    # totals to be wrong in the final result
    $self->{pid_list} = [];
    $self->{clients} = [];
    my @return_clients = ();
    while (my ($client, $count) = each %Clients) {
	while ($count--) {
	    push(@return_clients, $client);
	}
    }
    $Trace->exit(msg=>"return ".join(" ", @return_clients));
    return @return_clients;
}

=head2 execute

    $Remote->execute(host=>host,
                     cmd=>"command to execute",
                     runid=>runid,
                     nate_params=>{},
                     handler=>\&routine);

C<execute> is the method used to actually remotely execute a given command
on a host. The method uses the C<SPINDEX> nate param to decide whether or not
to use spindex or subtest to actually execute the command on the remote host.
The command is executed in the background and the method will return unless
the C<wait> param is set to 1. The caller can then use the C<wait> method to 
wait for the job to finish.

In the case of using spindex, C<execute> will build up the full command and
params to pass to ntest on the remote host. It will determine the full path
to the current ntest and use the same one on the remote host. It will also
perform a cd to the current directory on the remote host prior to executing
the ntest. It will also set C<LOGNAME> to the current user on the remote host
prior to running ntest. A warn result will be generated if the host could not
be reached. C<execute> will also build up a list of common NATE params to pass
to the remote ntest. These include: C<-noconsole> and C<LOGDIR>.

In the case of subtest, C<execute> will pass the command as is to subtest
and will use the C<-bg> option unless C<wait> is set to 1. If C<handler> is 
specified, then C<-onexit> will be set to the routine. If C<runid> is set, 
the it is passed as the runid to subtest.

=over

=item host

(Required) This is the name of the remote host on which to execute the command.

=item cmd

(Required) This is the command to execute on the host. This command will be
run via ntest on the remote host.

=item runid

(Required) This is the name of the job for use in the LOGDIR. In the case
of spindex, the remote LOGDIR will be set to C<$LOGDIR/runid>; the resulting
logs for this job will be found in a directory called C<runid>/latest in the
current C<LOGDIR>. In the case of subtest, C<-runid> will be set to runid; the
resulting logs will be found in C<runid.log> and the directory C<runid> in the
current C<LOGDIR>.

=item jobid

(Optional) This param is obsolete.

=item handler

(Optional) This param is not used in the case of spindex. In the case of 
subtest, the C<onexit> param to subtest will be set to this routine.

=item nate_params

(Optional) This is a reference to a hash of params that should be passed to the
C<cmd>.

=item wait

(Optional) This param indicates that the method should wait on the C<cmd> to
finish rather than returning immediately. The default is set to undef.

=item output

(Optional) This param is obsolete.

=back

=cut

sub execute {
    my $self = shift;
    my %params = 
    validate_mixed(
		   @_,
		   [ ],
		   {
		       host => {type=>SCALAR},
		       cmd => {type=>SCALAR | CODEREF},
		       nateeval => {type=>SCALAR | UNDEF, optional => 1, default => 0 },
		       runid => {type => SCALAR | UNDEF},
		       jobid => {type => SCALAR | UNDEF, optional => 1},
		       handler => {type => CODEREF | UNDEF, optional => 1},
		       nate_params => {type => HASHREF, optional => 1, default => {}},
		       wait => {type => SCALAR | UNDEF, optional => 1, default => 0},
		       output => {type => SCALAR | UNDEF, optional => 1, default => 1 },
		       username => {type => SCALAR | UNDEF, optional => 1 },
                       retry_cifs_errors => {type => SCALAR | UNDEF, optional => 1 },
		   }
		   );
    $Trace->entry();
    my $host = $params{host};
    my $eonout = 0 if ($params{output}); # the normal case for SpinDex
    my $pid;
    my %nate_params = %{$params{nate_params}};
    my $comment = "kicking off $params{runid} on client $host";
    $Trace->comment(msg=> $comment);

    # store the parameter listing for this client in case it is needed for a retry
    $self->{parameters}{$host} = \%params;
    # save this host name in the master client list. This list tracks which
    #  hosts were successfully started
    push(@{$self->{clients}}, $host);
    my $user = $params{username};
    if ($Spindex) {
        $user = getpwuid($<) if (!$user);
	if ($params{nateeval}){
	    use FindBin qw($RealBin);           # full directory pathname for this script
	    my $Nateeval = "$RealBin/nateeval";

	    $params{cmd} = "$Nateeval -t -e '$params{cmd}'";
	}
	# this should allow for perl code to be passed in rather than as a cmd
	my $nate_command;
	if ($nate_params{LOGDIR}) {
	    $nate_params{LOGDIR} .= "/$params{runid}";
	    $eonout = undef;
	} else {
	    $nate_params{LOGDIR} = "-";
	}
	my $nate_params = $self->convert_params_to_options(%nate_params);
	my $cwd = getcwd();

	if ($params{cmd} =~ /nateeval/) {
	    my ($filer,$nate_lib) = $Options->get(option=>[qw(FILER NATE_LIB)]);
	    $filer = $filer->[0];
	    $nate_command = "export NATE_LIB='$nate_lib' && export FILER=$filer && $params{cmd}";
	} else {	    
	    if ($params{cmd} !~ ".thpl") {
		$params{cmd} = "'$params{cmd}'";
	    }
	    my $nate_flags = "-noconsole -plain -username $user";
	    
	    # Start the remote tests in the same dir as we are in. This will work
	    # with testbeds.
	    $nate_command = "$Ntest $nate_flags $nate_params $params{cmd}";
	}
	my $spindex_command = "cd $cwd && export LOGNAME=$user && $nate_command";
	$Trace->comment(msg=>"Running Rexec($host, $spindex_command) as $user\n");
        my @st_begin = gettimeofday();
	$pid = $Spindex->Rexec(Host => $host,
			       Command => $spindex_command,
			       Wait => 0,
			       EonOut => $eonout,
			       User => $user,
			      );
        push(@{$self->{start_times}}, tv_interval(\@st_begin));
	if ( $pid < 0 ) {
	    $Trace->warn(msg=>"ERROR kicking off \"$spindex_command\" on host $host as $user, rc = $pid\n");
            pop(@{$self->{start_times}});
	} else {
            $Trace->debug(msg => "Spindex start on $host returned in ".${$self->{start_times}}[$#{$self->{start_times}}]." seconds.");
	    $Trace->debug(msg=>"$host as $user (internal pid = $pid)\n");
	    push(@{$self->{pid_list}}, $pid);
            $self->{clients_started}++;
	    $self->{clients_pids}{$pid} = $host;
            $self->{runid}{$pid} = $params{runid};
	}
    } else {
	my @params = $self->convert_hash_to_array(%{$params{nate_params}});
        my @cmd_parts;
        my $cmd;
        #
        # Do a simple check for non-thpl executables.  This isn't foolproof.
        # TODO: make this work in all cases
        #
        if ($params{cmd} !~ m/\.thpl/ && !$params{nateeval}) {
            @cmd_parts = quotewords(' ', 1, $params{cmd});
            # save the command
            $params{cmd} = shift @cmd_parts;

            my $h = host($host);
            if ($h->hosttype =~ /windows/ && $params{cmd} =~ m!^/{1,}\S+.+\.pl!) {
                # find the full path to the script
                $Trace->fatal(msg=>"CIFS support for non-thpls is not ready yet",
                              result=>"SCRIPT");
#                $params{cmd} = "//stealth/user/snider/p4/test/$params{cmd}";
                unshift(@cmd_parts, $params{cmd});
                $params{cmd} = "X:/perl581/bin/perl.exe";
            }
            # Put the separator before the arguments to the command
	    unshift @cmd_parts, '--' if @cmd_parts;

	    # Strip off quoting on the individual arguments are they were already split by quotewords
	    # and leaving them in place will cause them to be backslashed and quoted again which likely
	    # isn't the desired effect.
	    foreach (@cmd_parts){
		s/^(\'|\")(.*)(\1)$/$2/;
	    }
            $Trace->debug(msg => "Non-thpl executable detected, breaking command apart from arguments.\n".
                "Command: $params{cmd}\n".
                "Arguments (including separator): " . join(" ", @cmd_parts), 
            );
        } elsif ($params{cmd} =~ /\s+/ && !$params{nateeval}) {
            foreach my $part (split(/\s+/, $params{cmd})) {
                if ($part =~ /=/) {
                    push(@cmd_parts, $part);
                } else {
                    # this does not work if multiple scripts are specified: last one wins
                    $params{cmd} = $part;
                }
            }
        }
	if ($params{nateeval}){
	    $params{cmd} = "{".$params{cmd}."}";
	}
	my @subtest = ($params{cmd}, @params, -client, $host);
        push(@subtest, -username, $user) if (defined($user));
	push(@subtest, -runid, $params{runid}) if (defined($params{runid}));
	push(@subtest, -bg) if (!$params{wait});
	push(@subtest, -onexit, $params{handler}) if ($params{handler});
        # Append non-thpl command parts (if any).
        push(@subtest, @cmd_parts) if @cmd_parts;
	$Trace->debug(msg=>"Running subtest(".join(",", @subtest).")");
        eval {
            my @st_begin = gettimeofday();
            $pid = subtest(@subtest);
            push(@{$self->{start_times}}, tv_interval(\@st_begin));
            $Trace->debug(msg => "Subtest start on $host returned in ".${$self->{start_times}}[$#{$self->{start_times}}]." seconds.");
            $self->{clients_started}++;
	    $self->{clients_pids}{$pid} = $host;
            $self->{runid}{$pid}= $params{runid};
	    push(@{$self->{pid_list}},$pid);
        };

        if ($@) {
            my $msg = "Failed to start subtest on $host\n";

            if (UNIVERSAL::isa($@, 'NATE::Result')) {
                $msg .= $@->type()." Error.  Message(s): ".join(", ", $@->messages());
            } else {
                $msg .= "Error: $@";
            }
            
            pop(@{$self->{start_times}});
            $self->{clients_started}--;
            
            $Trace->warn(msg => $msg);
            $pid = 0;
        }
        if ($params{retry_cifs_errors}) {
            # windows clients sometimes run into problems trying to redirect STDOUT so
            # try to detect the failures and respawn them
            my @restarted = ();
            # make a local copy so that the loop can mofify the real one
            my @pid_list = @{$self->{pid_list}};
            foreach my $pid (@pid_list) {
                if (grep(/Unable to redirect STDOUT/, $pid->results())) {
                    $Trace->comment(msg=>"Found a LOGDIR problem on '$host'");
                    if ($self->{parameters}{$host}) {
                        @{$self->{pid_list}} = grep($_->pid() != $pid->pid(), @{$self->{pid_list}});
                        undef $pid;
                        $Trace->comment(msg=>"Retry running command on $host.");
                        $self->execute(%{$self->{parameters}{$host}}, retry_cifs_errors => 0);
                    }
                }
            }
        }
    }
    $Trace->exit(msg=>"return $pid");
    return $pid;
}

sub get_test_stats
{
    my $self = shift();
    my %stats;
    
    $stats{clients_started} = $self->{clients_started};
    $stats{total_start_time} = eval(join(" + ", @{$self->{start_times}}));
    $stats{average_start_time} = $stats{total_start_time} / $stats{clients_started};

    return \%stats;
}

my @Output;

# This method is obsolete so I am not making doc for it right now
sub send {
    my $self = shift;
    my %params = (
		  wait => 1,
		  @_);
    $Trace->entry();
    my $output;
    my $host = $params{host};
    if ($Spindex) {
	$Trace->comment(msg=>"Kicking off command on host $host\n");
	if ($params{cmd} ne ".thpl") {
	    $params{cmd} = "'$params{cmd}'";
	}
	my $pid = $Spindex->Rexec(Host => $host,
				  Command => $params{cmd},
				  PrintFormat => undef,
				  EonOut => sub { @Output = @_; },
				  );
	$Spindex->WaitPid(Pid => $pid, Block => 1);
	if ( $pid < 0 ) {
	    $Trace->fatal(msg=>"ERROR kicking off \"$params{cmd}\" on host $host, rc = $pid\n");
	} else {
	    $Trace->debug(msg=>"Running \'$params{cmd}\' on host $host (internal pid = $pid)\n");
	    $output = join("\n", @Output);
	    $Trace->debug(msg=>"Got '$output'");
	}
    } else {
	my $hconn = make_conn($host);
	$Trace->debug(msg=>"Running sayEOF($params{cmd})");
	eval {
	    $output = sayEOF($hconn, $params{cmd});
	};
    }
    $Trace->exit(msg=>"return '$output'");
    return $output;
}

=head2 wait

    my $result = $Remote->wait(timeout=>600);

C<wait> should be called after C<execute> has been called to run commands
in the background. This allows for many jobs to be run in parallel. This
method will wait for the specified amount of time to pass for all commands
to finish. If C<timeout> passes and some jobs are still running, then the
method will return and the final result will show which jobs timed out.
The results for the jobs that did finish will be returned as normal.

If no error is encountered, the C<wait> will return a PASS result. If an 
error does occurs, C<wait> will return a single result in the form of:

=over

fails / total client failures
client1 = result
client2 = result

=back

See C<handle_results> for more info on the results that are returned.

=over

=item timeout

(Optional) This param specifies the number of seconds that C<wait> should
wait before timing out the commands and returning an error.

=back

=cut

sub wait {
    my ($self, %params) = @_;
    $Trace->entry();
    my $timeout = $params{timeout};
    my $result_msg;
    my $result_obj;
    my $success_msg;
    my $fails = 0;
    my $total = scalar @{$self->{pid_list}};
    my $timeout_failure;

    if ($Spindex) {
	my $rc;
	$Trace->comment(msg=>"Waiting for all hosts to finish\n");
	$rc = $Spindex->WaitGroup(#WaitGroupName => "mywaitgroup",
				  Timeout => $timeout,
				  );
	if ( $rc < 0 ) {
	    $result_msg .= "ERROR waiting for all hosts to finish after $timeout seconds, rc = $rc\n";
	} else {
	    $Trace->comment(msg=>"All hosts have finished\n");
	}
	# Whether or not the timeout occurred, call handle_results to process the results we do have. 
	$fails = $self->handle_results(\$result_msg, \$success_msg, @{$self->{pid_list}});
	
	my $logdir = $Options->get(option=>"LOGDIR");
	unlink("$logdir/latest") if ($logdir);
	
        # If any of the remote processes finished with non-zero return codes, log it.
	if ($Spindex->SeenError() != 0 ) {
	    $Trace->warn(msg=>"spindex saw an error");
	}
    } else {
        #
        # TODO: Change this so it checks each subtest individually 
        # to see if it's alive and if it's reported and results.
        #

        eval {
            Subtest::wait_finish(subtests => $self->{pid_list},
                                 timeout  => $timeout,
                                );
        };
        $timeout_failure = $@;
        if ($@) {
            $Trace->comment(msg=>"Error from Subtest::wait_finish(): $@");
        }
        
        $fails = $self->handle_results(\$result_msg, \$success_msg, @{$self->{pid_list}});
    }
    
    # make sure that at least one job started successfully. Successfully
    #  started clients show up in $self->{clients}. If it is empty, then no
    #  clients started and an error needs to be returned. If it is not
    #  empty, then at least one started and the others issued warnings.
    if (scalar @{$self->{clients}} && ! $total) {
	$result_msg = "No client jobs were started\n" . $result_msg;
	$result_obj = NATE::Result::Fail->new($result_msg);
    } elsif ($timeout_failure) {
        $result_msg = "$timeout_failure\n$result_msg";
	$result_obj = NATE::Result::Fail->new($result_msg);
    } elsif ($result_msg) {
	$result_msg = "$fails / $total client failures\n" . $result_msg;
	$result_obj = NATE::Result::Fail->new($result_msg);
    } elsif ($success_msg) {
	$result_msg = $success_msg;
	$result_obj = NATE::Result::Pass->new($result_msg);
    }
    
    # clean up the master list for the next execute/wait call
    $self->{pid_list} = [];
    $self->{clients} = [];

    # reset the stats since the execute/wait routines can be
    #  called multiple times by the same script.
    delete $self->{clients_started};
    delete $self->{start_times};

    $Trace->exit(msg=>"return '$result_msg'");
    return $result_obj;
}

=head2 wait_pid

  my @pids = $Remote->wait_pids();

C<wait_pid> will return a list of pids that have already finished.

=cut

sub wait_pid {
    my ($self, %params) = @_;
    $Trace->entry();
    my @pids;
    if ($Spindex) {
#	$Trace->comment(msg=>"Waiting for all hosts to finish\n");
	@pids = $Spindex->WaitPid();
    } else {
	# now find the ones that are finished
	# the call to get_all returns all known jobs so only return the 
	#  new ones that finished
	foreach my $t (@{$self->{pid_list}}) {
            if (!$t->is_alive() && !$self->{pids_fin}{$t}) {
                push(@pids, $t);
                $self->{pids_fin}{$t} = 1;
            }
	}
    }
    $Trace->exit(msg=>"");
    return @pids;
}

=head2 handle_results

  my $result = $Remote->handle_results(\$result, \$success, @pids);

C<handle_results> will examine the list of C<pids> and check the results
for each one. For each pid that has encountered an error, a failure counter
is incremented and the result is added to the C<$result> reference that is 
passed in. The final result will include the total number of failures and
the total number of jobs that were run. If the pid is not finished running,
then the result is set to C<TIMEOUT>.

This method is normally only called by C<RemoteExecution> or its associated
libraries. It should not be necessary to call this from outside.

=cut

sub handle_results {
    my ($self, $result_msg, $success_msg, @pids) = @_;
    $Trace->entry();
    my $rc;
    my $fails;
    if ($Spindex) {
        # Check the return codes (exit values) from the actual remote tests.
	foreach my $pid (@pids) {	
	    my $client = $self->{clients_pids}{$pid};
	    $rc = $Spindex->ExitValue(Pid => $pid);
	    if (!defined($rc)) {
		$fails++;
		$$result_msg .= "$client = TIMEOUT\n";
		next;
	    }
	    $rc = $rc >> 8;
	    $Trace->comment(msg=> "$self->{runid}{$pid} on client $client returned $rc\n");
	    if ( $rc ) {
		if ($rc == 1) {
		    $fails++;
		    $$result_msg .= "$client = syntax error\n";
		} elsif ($rc == 250) {
		    $$result_msg .= "$client = WARN\n";
		} elsif ($rc == 251) {
		    $$result_msg .= "$client = NA\n";
		} elsif ($rc == 252) {
		    $fails++;
		    $$result_msg .= "$client = INSPECT\n";
		} elsif ($rc == 253) {
		    $fails++;
		    $$result_msg .= "$client = CONFIG\n";
		} elsif ($rc == 254) {
		    $fails++;
		    $$result_msg .= "$client = FAIL/FATAL\n";
		} elsif ($rc == 255) {
		    $fails++;
		    $$result_msg .= "$client = SCRIPT\n";
		} else {
		    $fails++;
		    $$result_msg .= "$client = UNKNOWN ($rc)\n";
		}
	    }
	}
    } else {
	# now get the results
	foreach my $subtest (@pids) {
	    my $client = $self->{clients_pids}{$subtest};
	    my $client_fails = 0;
	    $Trace->comment(msg=> "$self->{runid}{$subtest} on client $client returned\n");
	    foreach my $result ($subtest->results()) {
		if ($result->severity() >= NATE::Result::Fail->severity()) {
		    $client_fails++;
		    $$result_msg .= "$client: " . join("\n", $result->messages()) . "\n";
		}
		if ($result->type() eq NATE::Result::Pass->type()) {
		    $$success_msg = join("\n", $result->messages());;
		}
	    }
	    $fails++ if ($client_fails);
	}
    }
    $Trace->exit(msg=>"return $fails");
    return $fails;
}
	
=head2 exit_value

  $rc = $Remote->exit_value(pid=>$pid);

C<exit_value> will query the given pid for its return code and return it.

This method is normally only called by C<RemoteExecution> or its associated
libraries. It should not be necessary to call this from outside.

=cut

sub exit_value {
    my $self = shift;
    my %params = 
    validate_mixed(
		   @_,
		   [ ],
		   {
                    pid => {type=>SCALAR | OBJECT},
		   }
                  );
    $Trace->entry();
    my $rc;
    if ($Spindex) {
        $rc = $Spindex->ExitValue(Pid =>$params{pid});
        $Spindex->ClearProclist(Pids => $params{pid});
    } else {
        if (!UNIVERSAL::isa($params{pid}, "Subtest")) {
            $Trace->exit(result=>"SCRIPT",
                         msg=>"pid '$params{pid}' is not a Subtest object");
        }
	$rc = $params{pid}->exit_status();
    }
    $Trace->exit(msg=>"return $rc");
    return $rc;
}

=head2 kill

    $Remote->kill();

C<kill> should be called after C<execute> has been called to run commands
in the background. This routine kills all subtests created by this object.

=over

=item pid

(Optional) The pid to be killed. By default, it kills all processes.

=back

=cut

sub kill {
    my ($self, %params) = @_;
    $Trace->entry();

    if ($Spindex) {
        # To Do
    } else {
        if($params{pid}) {
            $params{pid}->kill();
        } else {
            # Kill all PIDs associated with this object
            foreach my $subtest (@{$self->{pid_list}}) {
                $subtest->kill();
            }
            # clean up the master list for the next execute/wait call
            $self->{pid_list} = [];
            $self->{clients} = [];
            
            # reset the stats since the execute/wait routines can be
            #  called multiple times by the same script.
            delete $self->{clients_started};
            delete $self->{start_times};
        }
    }
}

# This is really an internal method so I will not add pod for it
sub convert_params_to_options {
    my ($self, %params) = @_;
    $Trace->entry();
    
    my $string = "";
    while (my ($param, $value) = each %params) {
	if (ref($value) eq "ARRAY") {
	    $value = join(",", @$value);
	}
	$string .= " $param=\"$value\"";
    }
    $Trace->exit(msg=>"return $string");
    return $string;
}

# This is really an internal method so I will not add pod for it
sub convert_hash_to_array {
    my ($self, %params) = @_;
    $Trace->entry();
    
    my @array = ();
    while (my ($param, $value) = each %params) {
	if (ref($value) eq "ARRAY") {
	    $value = join(",", @$value);
	}
	push(@array, "$param=$value");
    }
    $Trace->exit(msg=>"return @array");
    return @array;
}

# This is really an internal method so I will not add pod for it
sub cleanup {
    $Trace->comment(msg=>"cleaning up\n");
    $SIG{INT} = undef;
    if ($Spindex) {
	$Spindex->Rkill(Pid => "all");
	$Spindex->DESTROY();
	exit;
    }
}

=head1 PARAMS

=over

=item SPINDEX

Setting this param to anything (1, true, etc) causes this package to use
spindex rather than subtest for remote execution. It defaults to undef.

=item MAX_UPTIME

Setting this param will cause C<check_clients> to drop any client from the
list if the uptime on the box is higher than this value.

=item CHECK_CLIENT_HEALTH

This param controls whether or not C<check_clients> actually runs the
health checking on the clients. It defaults to 1. Setting the value to
C<''> turns it off.

=item NATE_HOSTS

This option definition is needed when running with SPINDEX=1. If it is 
defined then it will be passed to the remote ntest invocation.

=item LOGDIR

This option definition is needed when running with SPINDEX=1. If it is 
defined then it will be passed to the remote ntest invocation.

=item NATE_LIB

This option definition is needed when running with SPINDEX=1. If it is 
defined then it will be passed to the remote ntest invocation.

=item NATE_LIB_PREPEND

This option definition is needed when running with SPINDEX=1. If it is 
defined then it will be passed to the remote ntest invocation.

=item LOGNAME

This option definition is needed when running with SPINDEX=1. If it is 
defined then it will be passed to the remote ntest invocation.

=back

=head1 SEE ALSO

L<SpinDexClient>
L<Tharn>

=head1 BUGS

More time has been spent testing with SPINDEX=1.

=head1 STATUS

Shared

=head1 ATTRIBUTES


=head1 AUTHOR

  Bill Snider (snider@netapp.com)
  $Id: $
  Last revised $Date: $

=head1 COPYRIGHT

Copyright 2006, Network Appliance., All Rights Reserved

=cut

1;
