# -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/Revert/Revert.pm#1 $

## @description  Revert library (OBSOLETE)
##
## @summary Revert.pm provides mechanisms to control the reversion process for
##          a filer.
##
## @example my $revert = Upgrade->new($conn, to_version => '6.1')
##
## @status obsolete -- use Filer::Revert
##
## @author shenry@netapp.com (orginal amandag@netapp.com)

package Revert;

use strict;
use vars qw($VERSION $RELEASES @ISA @EXPORT @EXPORT_OK);

use Sysconfig;
use SayWrap qw(sayEOF sayWAIT);
use Carp;
use filer qw(filer_version_get);
use SPARSE::Util::Logger qw(logDbg logTrc);
use FS::Util::Utilities qw(revert_to);
use FS::Util::Setup qw(mount_root_vol umount_vol);
use Data::Dumper;
use Volume qw(vol_root vol_status vol_list vol_online);
use Tharn qw(subtest connect host say param);
use Utils qw(make_host make_conn);
use install;
use Upgrade::Upgrade qw(my_create_vol_and_mount);
use SnapCmds qw(snapshot_list snapshot_delete);

@ISA    = qw();
@EXPORT = qw(allowed new v_compare do download install_release);
$VERSION = '0.2';
$RELEASES = [ {
fosters => 5.2,
becks => 5.3,
guinness => 6.0,
vb => 6.1,
redhook => 6.2,
krakus => 6.3,
fullsail => 6.4,
scrimshaw => 6.5,
anchorsteam => 7.0},
("5.2","5.3","6.0","6.1","6.2","6.3","6.4", "6.5","7.0") ];

use vars qw($DELETE_SNAPSHOTS $ONLINE_VOLS);
param("DELETE_SNAPSHOTS", '-default', 0);
param("ONLINE_VOLS", '-default', 0);

# ---------------------
# sub new()
# Constructor requires a filer connection record as the first argument
# All other parameters are supplied in a named fashion. If not
# supplied, the "to_version" argument will default to undef
# Some of the records for the constructor include:
#  the filer connection record
#  the filer IP address
#  the to_version that the test runner has chosen
#  the from_version, retrieved from the filer
#  the previous versions, also retrieved from the filer -- "previous"
#    refers to the acceptable versions for revert, not the revert history
#  the filer hostspec
#  the Sysconfig object
# ---------------------
sub new {
  my $class = shift;
  my $conn = shift;
  my %parms = @_;

  $ENV{TIMEOUT} = 1200 if $ENV{TIMEOUT} < 1200;
  $ENV{CMDTIMEOUT} = 2400 if $ENV{CMDTIMEOUT} < 2400;
  _zero_spares($conn);

  my @prev_versions = _prev_versions($conn);

  bless {
    _conn          => $conn,
    _filer_ip      => $conn->hostp->default_ip,
    _to_version    => $parms{to_version} || undef,
    _from_version  => _current_version($conn),
    _prev_versions => [ @prev_versions ],
    _f             => $parms{FILER},
    _sysconfig     => undef,
    _mount         => undef,
  }, $class;
}

# --------------------
# Read-only Accessors
# These convenience methods do not prevent the caller from writing code like
#  $self->{_conn}
# but they do protect against potential mistakes like
#  $self{_conn}
# or a situation that assigns undef to $self->{_conn} accidentally, which
# would be disastrous. $self->conn would simply ignore the undef and all
# would be well.
#
# usage example:
#  my $from_version = $self->from_version();
#
# The read-only accessors are for each member of the object that were
# blessed in the constructor that shouldn't be changed.
#
# It would be appropriate to upgrade these hand-coded methods with a
# caching AUTOLOAD approach, but there is some disagreement on what is
# an acceptable level of Perl education required to for library code.
# --------------------
sub conn          { $_[0]->{_conn}          }
sub fhostspec     { $_[0]->{_f}             }
sub lconn         { $_[0]->{_lconn}         }
sub filer_ip      { $_[0]->{_filer_ip}      }
sub from_version  { $_[0]->{_from_version}  }
sub prev_versions { $_[0]->{_prev_versions} }
sub sysconfig {
  my $self = shift;
  if (! $self->{_sysconfig}) {
    my $cfg = Sysconfig->new($self->fhostspec) || croak $@;
    $self->{_sysconfig} = $cfg;
  }
  return $self->{_sysconfig};
}

# --------------------
# Read-Write Accessors
# Again, these are convenience methods that are called in object fashion.
# If an argument is supplied, the corresponding object member is updated
# to reflect this new value, e.g.
#
#  $revert->to_version("6.5");
#
# Thus the user can either set the value as above, or retrieve the value
# as below
#  my $to_version = $revert->to_version();
#
# 0 is not a valid version.
#
# --------------------
sub to_version {
  my $self = shift;
  if (my $version = shift) {
    $self->{_to_version} = $version;
  }
  return $self->{_to_version};
}

sub mount {
  my $self = shift;
  if (my $mount = shift) {
    $self->{_mount} = $mount;
  }
  return $self->{_mount};
}


# --------------------
# sub allowed
# The revert is only allowed if four specific "tests" pass:
# (1) version
# (2) hardware
# (3) config
# (4) run status
# --------------------
sub allowed {
  my $self = shift;
  if ($self->_version_allows()   &&
      $self->_hardware_allows()  &&
      $self->_config_allows()    &&
      $self->_running_status_allows() ) {
    return 1
  } else {
    return 0
  }
}

# --------------------
# sub _version_allows
# _version_allows checks whether the "to_version" that's been
# requested is compatible with the acceptable previous versions
# --------------------
sub _version_allows {
  my $self = shift;
  return undef unless $self->to_version;

  logDbg("_version_allows: to_version=" . $self->to_version());
  logDbg("_version_allows: pre_versions=" . join " ", @{$self->prev_versions()});
  my $short_vers = join(".",(split(/\./,$self->to_version()))[0,1]);
  logDbg("_version_allows: short_vers=$short_vers");
  if (grep { _integer_version($short_vers) }
       map { _integer_version($_) } @{$self->prev_versions()} ) {
    return 1
  } else {
    return 0
  }
}

# --------------------
# sub _hardware_allows
# N/A
# --------------------
sub _hardware_allows {
  my $self = shift;

# It would certainly be more logical, for the moment, if we let a
# Sysconfig object be its own object, not roll it into the revert object,
# so why don't the following two lines of code work?
# (Note: whether or not Sysconfig is in @ISA doesn't affect the error)
#

  return 1;
}

# --------------------
# sub _config_allows
# N/A
# --------------------
sub _config_allows {
  my $self = shift if ref $_[0];
  return 1;
}

# --------------------
# sub _running_status_allows
# N/A
# --------------------
sub _running_status_allows {
  my $self = shift if ref $_[0];
  return 1;
}

# --------------------
# sub v_compare
# v_compare is a class method that takes two netapp version strings
# and converts them to a representative integer. If the first version is
# higher than the second version, v_compare returns 1. If they are equal, it
# returns 0. If the first version is lower than the second version, it returns
# -1
# See the documentation for _integer_version for more information about
# acceptable string formats
# Currently, it croaks if the format is wrong. This needs to be fixed.
# --------------------
sub v_compare {
  my $self = shift if ref $_[0];
  my ($v1, $v2) = @_;
  my $result = $self->_integer_version($v1) - $self->_integer_version($v2);
  if ($result > 0 ) {
    return 1
  } elsif ($result < 0 ) {
    return -1
  } elsif ($result == 0) {
    return 0
  } else {
    return undef
  }
}

# --------------------
# sub _integer_version
# Versioning follows very strict rules here
# If a word is supplied, then we attempt to look up the version
# (1) Versions are split on a single period
# (2) No letters or other characters are allowed
# (3) Only the major and minor digits are significant
# (4) A partial version of .11 is greater than a partial version of .2
# but a partial version of .20 is greater than a partial version of .11
# (i.e. right-padded zeroes are significant)
# --------------------
sub _integer_version {
  my $self = shift if ref $_[0];
  my $version = shift;

  $version =~ s/(X\d+|RC\d+)/.0/;
  my @partials = (split /\./, $version)[0..2];
  if (grep /\D/, @partials) {
    # attempt to detect a current version by looking at prev versions
    my $current_version = undef;
    my $prev_version = $self->prev_versions->[1];
    # get the version after the prev version from the RELEASES lookup array
    for (my $i = 1; $i < @$RELEASES; $i++) {
      $current_version = $RELEASES->[$i+1] if $prev_version eq $RELEASES->[$i];
    }
    if (! defined $current_version ) {
      croak "$version was not found in lookup table. lookup table contains: ".
            join " ", values %{$RELEASES->[0]};
    }
    @partials = (split /\./, $current_version)[0..2] or
      croak "Version '$version' contains an illegal character"
  }
  my $intvers = join "", map { sprintf("%06d", $_) } @partials;
  logDbg("_integer_version($version) => $intvers (partials=@partials)");
  return ( $intvers ) ||
    croak "Could not split version '$version' into partials";
}

# --------------------
# sub _current_version
# This retrieves and returns what the filer thinks its current version is
# --------------------
sub _current_version {
  my $self = shift;
  my $conn = $self =~ 'Connectrec' ? $self : $self->conn;

  my $version = filer_version_get($conn) or croak $@;
  logDbg("The version on the filer is: $version");
  return $version;
}

# --------------------
# sub _prev_versions
#
# Here's a slightly more experimental method for you, called _prev_versions;
# the title refers to the two versions that a filer can revert to from its
# current version, e.g. 6.5 and 6.4 if the filer's at 7.0.
#
# May be called either as a class method or as an object method.
# If called as a class method, it must be supplied with a filer connection
# record as the argument. If called as an object method, the method already
# has a copy of the filer connection record, so no arguments are needed.
#
# This method can be called in list or scalar context as well;
# In list context, this returns the previous two versions that the filer
# could potentially revert to. In scalar context, this returns the version
# immediately previous to the filer's current version.
#
# usage examples:
# my $last_version = Revert::_prev_versions($conn);   # scalar
# my $last_version = $self->_prev_versions();         # scalar
# my @last_2_vrsns = _prev_versions($conn);           # list
# my @last_2_vrsns = $self->_prev_versions();         # list
# my @last_1_vrsn  = scalar $self->_prev_versions();  # force scalar context
#
# ASSUMPTION: The filer command `revert_to` always lists acceptable versions in
# ascending order, e.g. 6.4 followed by 6.5
# --------------------
sub _prev_versions {
  my $self = shift;
  my $conn = $self =~ 'Connectrec' ? $self : $self->conn;

  my $out = (say($conn, 'priv set -q diag; revert_to;',-eval_error))[2];
  my @lines = split /\n/, $out;
  my @versions;

  logDbg("\n\nThe raw result of _prev_versions was: ". join "\t", @lines);
  foreach (@lines) {
    if ($_ =~ /(\d+\.\d+)/) {
      push @versions, $1;
    }
  }
  if (wantarray) {
    return @versions
  } else {
    return $versions[1] || undef
  }
}


# --------------------
# sub make_mount_point
# Issue a call to NFS/Mount.thpl to create and mount a mountpoint
# for the root volume on the filer
# --------------------
sub make_mount_point {
  my $self = shift;

  my $filer_ip  = $self->filer_ip;
  my $root_vol   = vol_root($self->conn) || croak $@;
  my $mountpoint = "/t/$filer_ip/$root_vol";
  $mountpoint = mount_root_vol($self->fhostspec, undef, $mountpoint,
                               (filer_conn => $self->conn) ) ||
    croak $@;

  $self->{_mount} = $mountpoint;

  return $mountpoint;
}

# --------------------
# sub determine_archive_path
# Based on the hardware config and to_version, determine which
# archive to untar
# on a per-test basis, you can send $revert->install(kernel => 'debug')
# or you can override on the command line with $RELEASESTYPE.
# --------------------
sub determine_archive_path {
  my $self = shift;
  my %parms = @_;

  my ($processor, $dist) = ('','');
  # Get the kernel from the kernel => ... param passed in through install()
  # or simply set to _pbo (shipping) by default
  # kernel path postfix options are "releases", "releases_debug", or
  # "releases_pbo";
  my $kernel = defined $ENV{RELEASESTYPE} ? $ENV{RELEASESTYPE} :
                   defined $parms{kernel} ? $parms{kernel} : "releases_pbo";

  # Get the processor so we know which dist
  my $version_string = (say($self->conn, "version -v"),-eval_error)[2];
  if ($version_string =~ /\<(\d).?\>/) {
    $processor = $1;
  } elsif ($version_string =~ /\<(\D).?\>/) {
    $processor = '3';
  } else {
    croak "Could not determine hardware from version string: $version_string";
  }
  if ($processor == 1) {
    $dist = "alpha";
  } elsif ($processor == 3) {
    $dist = "pc";
  } elsif ($processor == 6) {
    $dist = "mips";
  } else {
    croak "This version of Revert only supports alpha (1), pc (3), and mips ".
          "(6), and the processor version detected was ($processor)";
  }

  my $archive = "/rlse/".
                ($self->to_version() || croak "No revert version").
                "/$kernel/".
                "$dist/dist/tar_floppy_$dist";
  croak "Path $archive was not found" unless -e $archive;
  return $archive;
}

# --------------------
# sub install_release
# This method sets up the mount on the client, downloads the data files,
# and runs the install program
# --------------------
sub install_release {
  my $self = shift;
  my %parms = @_;

  my $mount = $self->mount($self->make_mount_point);

  my $archive = $self->determine_archive_path(%parms);
  my $output;
  print "PROCEEDING TO MOUNTPOINT $mount: ";
  $output = `cd $mount && pwd`;
  print "Current working directory is: $output \n";
  #my $key = getc();
  print "PROCEEDING TO UNTAR $archive: ";
  $output = `cd $mount && sudo tar -xvf $archive`;
  print "$output \n";
  print "PROCEEDING TO INSTALL: ";
  $output = `cd $mount && sudo ./install_netapp $mount`;
  my @lines = split /\n/, $output;
  if ($output !~ /download/) {
    croak "The expected prompt to download the kernel was not seen: \n";
  }
  print join "\n", @lines[( scalar(@lines) - 15 )..( scalar @lines )];

  #print "\n\nwaiting for input: ";
  #my $key = getc();
  return 1;
}


# --------------------
# sub do
# Execute the steps necessary to revert the filer
# after all other steps have been taken to initialize
# and information about the versions and paths
# have been gathered
# You can send $revert->do(blind => 1) in order to return immediately
# You can also use interactive => 1 if for whatever reason you want to use the
# -i option
# When you do interactive, then the filer does a query/response thing.
# In this case, you have to tell it the wait string or strings for the query,
# and what to say to the query for example
# waitstring      => [ 'offline attribute' ]
# revert_response => "y"
# Return the response from the filer
# --------------------
sub do {
  my $self = shift;
  my %parms = @_;

  my $interactive = $parms{interactive} || 0;
  my $blind = $interactive ? 1 : ($parms{blind} || 0);
  my $waitstrings = $parms{waitstring} || [ ];
  my $rvt_response = $parms{revert_response} || "n";

  my $response = '';
  my $abbrev_version = join ".", (split /\./, $self->to_version)[0,1];
  $waitstrings = [ map { quotemeta($_) } @$waitstrings ];

  print "abbrv version is $abbrev_version";


  say($self->conn, "nfs off", -eval_error);# or croak "Could not turn off NFS";
  say($self->conn, "cifs terminate", -eval_error);# or croak "Could not turn off CIFS";

  # Get the kernel
  $self->download();

  $self->online_vols();

  $self->delete_all_snaps();

  my $cmd = "revert_to -i " . $abbrev_version;

  print "PROCEEDING TO REVERT TO $abbrev_version\n";
  $response = (say($self->conn, $cmd, ( $blind ? '-blind' : ( ) ),-eval_error))[2];

  if ($interactive) {
    print "waitstrings are ", @$waitstrings;
    $self->conn->expect(match => $waitstrings, before => $response) or die $@;
    print "\n\n\nREPONSE to revert_to command was: $response, answering '$rvt_response'\n\n\n";
    $self->conn->send($rvt_response);
    my $revert_waitstrings = [ map { quotemeta($_) } ("revert the system",
                                                      "above conditions") ];
    print "\n\n\nwaitstrings are ", @$revert_waitstrings , "\n\n\n";
    $self->conn->expect(match => $revert_waitstrings, before => $response,
                        max_idle => 60);
    print "\n\n\nREPONSE to response was: $response\n\n\n";
    return;

  } else {
    return $response;
  }
}

# --------------------
# sub download
# Issue the download command
# Return the response from the filer
# --------------------
sub download {
  my $self = shift;
  my ($match, $after) = ('','');

  return if $self->conn->hostp->hosttype eq 'filer-simulator';

  print "PROCEEDING TO DOWNLOAD THE KERNEL: ";
  ($match, undef, $after) = say($self->conn, "download",
                                "boot blocks successfully copied to disk",
                                -eval_error) or croak "Could not download new kernel successfully";
  print "after is $after";
  return $after;
}

# --------------------
# sub _zero_spares
# Check environment for $ZERO_SPARES (enabled by default)
# Can disable from command line or in a test with
# $ZERO_SPARES=0 or $ZERO_SPARES=[n|no|N|No]
# Sends a background message to the filer to zero the spare disks
# Which can reduce the amount of time it takes to run a test
# --------------------
sub _zero_spares {
  if (! defined $ENV{ZERO_SPARES} || $ENV{ZERO_SPARES} =~ /y|1/i) {
    print "Zeroing spares now\n";
    sayEOF(shift, "disk zero spares;result;");
    sleep(5);
  }
}


# --------------------
# sub online_vols
# Check environment for $ONLINE_VOLS (enabled by default)
# Can disable from command line or in a test with
# $ONLINE_VOLS=0 or $ONLINE_VOLS=[n|N|no|NO]
# Bring all vols online
# croak if they can't be brought online
# --------------------
sub online_vols {
  my $self = shift;
  my @vols = vol_list($self->conn);

  if ((! defined $ENV{ONLINE_VOLS} or $ENV{ONLINE_VOLS} =~ /y|1/i)
	  or ($ONLINE_VOLS =~ /y|1/i)) {
    print "Onlining the volumes now\n";
    map { vol_online($self->conn, $_) or croak $@ } @vols;
  }
}

# --------------------
# sub filer_snaps
# --------------------
sub filer_snaps {
  my $self = shift;
  my @vols = vol_list($self->conn);
  my @snaps = map { snapshot_list($self->conn, $_) } @vols;
  return @snaps;
}

# --------------------
# sub delete_all_snaps
# Check environment for $DELETE_SNAPSHOTS (disabled by default)
# Can enable from command line or in a test with
# $DELETE_SNAPSHOTS=1 or $DELETE_SNAPSHOTS=[y|yes|Y|Yes]
# --------------------
sub delete_all_snaps {
  my $self = shift;
  return unless ($ENV{DELETE_SNAPSHOTS} =~ /y|1/i or
		 $DELETE_SNAPSHOTS =~ /y|1/i);

  my @snaps;
  print "WARNING: All snapshots on this filer are about to be deleted. \n",
        "The test will pause for 30 seconds to allow you to change your mind\n",
        "starting now.\n";
  sleep(30);
  my @vols = vol_list($self->conn);
  foreach my $vol (vol_list($self->conn)) {
    foreach my $snap (snapshot_list($self->conn, $vol)) {
      snapshot_delete($self->conn, $vol, $snap) && push @snaps, $snap;
    }
  }
  return @snaps;
}

1;
