# this is basically -*- perl -*-
# $Id: //depot/prod/test/Rsierranevada/lib/RichData.pm#2 $
#
## @summary RichData - Rich data set creation and manipulation
## @author	dl-dpgqa-automation@netapp.com
## @status	Public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

RichData - Rich data set creation and manipulation for unix and windows

=head1 SYNOPSIS

  use RichData;

  populate(path => 'Z:\temp', client => $client);

  modify(path => '/tmp/data');

  read_tree(path => '/tmp/data');

  delete_tree(path => '/tmp/data');

  compare(path => '/tmp/data', cmppath => '/t/data');

  $counters_h = get_counters();

  $template_h = get_template(mode => 'modify', name => 'default');

  # Create a tree with 1000 files
  %options = total_files_pop_opts(files => 1000);
  populate(path => '/tmp/data', client => $client, %options);

=head1 DESCRIPTION

Provides routines to create and modify a rich data tree.
It creates directories, files, streams (named attributes), links, symbolic
links and sparse files. It can create the file system objects with ACL's and
attributes like archive, hidden, compress and encrypt.

Modifies the data tree by removing, renaming, creating more objects,
modifying ACL's, modifying the contents of files and streams, truncating
and modifying (set/clear) attributes.

Refer to tools/dp/cifs_tools_design.txt for more detailed information
on the parameters used by the population and manipulation subroutines.

=head1 STATUS

  Public

=head1 AUTHOR

  dl-dpgqa-automation@netapp.com
  $Id: //depot/prod/test/Rsierranevada/lib/RichData.pm#2 $
  Last revised $Date: 2013/04/24 $

=head1 COPYRIGHT

  Copyright 2006, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package RichData;

use strict;
use warnings;
use base qw(Exporter);

use POSIX qw();
use Carp qw(croak);
use Cwd qw(getcwd);
use Storable qw(dclone);
use FileManip qw(remove_file vol_info);
use dp::snapmirror::qsm::TargetedData;
use Params::Validate qw(validate SCALAR HASHREF);

our @EXPORT_OK = qw(
    populate
    modify
    read_tree
    delete_tree
    compare
    get_counters
    get_template
    total_files_pop_opts
);

#---------------------------------------------------------------------
# Version for package
#---------------------------------------------------------------------
our $VERSION = '1.0';

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub populate(%);
sub modify(%);
sub read_tree(%);
sub delete_tree(%);
sub compare(%);
sub get_counters();
sub get_template(%);
sub total_files_pop_opts(%);
sub _streams_supported($$;);
sub _check_options($\%;);
sub _init_create_opts(\%;);
sub _get_seed();
sub _get_cont_opts(\%$;);
sub _get_args(\%$$;);
sub _create_none(\%;);

#---------------------------------------------------------------------
# Package variables
#---------------------------------------------------------------------
my $_seed;
# Local copy of the data counters.  counters are the number of a
# specific object type which has been modified or created. See
# subroutine get_counters() for more information.
my $counters = {};
# A hash that provides the default parameters for population
my %pop_template;
# A hash that provides the default parameters for modify
my %mod_template;
# A hash that provides the default parameters for read tree
my %read_template;
# A hash that provides the default parameters for compare
my %compare_template;
my @cont_opts = qw(
    dirs
    files
    streams
    sparse
    compress
    encrypt
    archive
    hidden
    acl
);
my %args_map = (
    file_size_avg   => 'size_avg',
    file_size_dev   => 'size_dev',
    stream_size_avg => 'size_avg',
    stream_size_dev => 'size_dev',
);

#---------------------------------------------------------------------

=head2 ::populate

Populate a rich data set.

=over

=item Arguments

B<path> => I<string>

I<Required.>
Top level directory of tree to populate. If it does not exists it
will be created.

B<client> => I<hostspec>

I<Optional.>
Client where populate will run. Default is localhost.

B<template> => I<string>

I<Optional.>
Template for population which sets the default values of all the following
options. Default is the 'default' template.

B<name_avg> => I<number>

I<Optional.>
Average number of characters for the name of objects.

B<name_dev> => I<number>

I<Optional.>
Standard deviation of the number of characters for the name of objects.

B<dir_avg> => I<number>

I<Optional.>
Average number of directories to create under each directory.

B<dir_dev> => I<number>

I<Optional.>
Standard deviation of the number of directories to create under each directory.

B<file_avg> => I<number>

I<Optional.>
Average number of files to create under each directory.

B<file_dev> => I<number>

I<Optional.>
Standard deviation of the number of files to create under each directory.

B<file_size_avg> => I<number>

I<Optional.>
Average size of files.

B<file_size_dev> => I<number>

I<Optional.>
Standard deviation of the size of files.

B<depth_avg> => I<number>

I<Optional.>
Average directory depth.

B<depth_dev> => I<number>

I<Optional.>
Standard deviation of directory depth.

B<stream_avg> => I<number>

I<Optional.>
Average number of streams to create on file/directory.

B<stream_dev> => I<number>

I<Optional.>
Standard deviation of the number of streams to create on file/directory.

B<stream_size_avg> => I<number>

I<Optional.>
Average size of streams.

B<stream_size_dev> => I<number>

I<Optional.>
Standard deviation of the size of streams.

B<dir_acl_num> => I<number>

I<Optional.>
Number of directories to add an ACL.

B<dir_acl_pct> => I<number>

I<Optional.>
Percentage of directories to add an ACL.

B<file_acl_num> => I<number>

I<Optional.>
Number of files to add an ACL.

B<file_acl_pct> => I<number>

I<Optional.>
Percentage of files to add an ACL.

B<dir_stream_num> => I<number>

I<Optional.>
Number of directories to add streams.

B<dir_stream_pct> => I<number>

I<Optional.>
Percentage of directories to add streams.

B<file_stream_num> => I<number>

I<Optional.>
Number of files to add streams.

B<file_stream_pct> => I<number>

I<Optional.>
Percentage of files to add streams.

B<sparse_num> => I<number>

I<Optional.>
Number of files to be sparse files.

B<sparse_pct> => I<number>

I<Optional.>
Percentage of files to be sparse files.

B<link_num> => I<number>

I<Optional.>
Number of files to be links.

B<link_pct> => I<number>

I<Optional.>
Percentage of files to be links.

B<Slink_num> => I<number>

I<Optional.>
Number of files to be symbolic links.

B<slink_pct> => I<number>

I<Optional.>
Percentage of files to be symbolic links.

b<compress_num> => I<number>

I<Optional.>
Number of objects to set the compress attribute.

B<compress_pct> => I<number>

I<Optional.>
Percentage of objects to set the compress attribute.

B<encrypt_num> => I<number>

I<Optional.>
Number of objects to set the encrypt attribute.

B<encrypt_pct> => I<number>

I<Optional.>
Percentage of objects to set the encrypt attribute.

B<archive_num> => I<number>

I<Optional.>
Number of objects to set the archive attribute.

B<archive_pct> => I<number>

I<Optional.>
Percentage of objects to set the archive attribute.

B<hidden_num> => I<number>

I<Optional.>
Number of objects to set the hidden attribute.

B<hidden_pct> => I<number>

I<Optional.>
Percentage of objects to set the hidden attribute.

B<acl_opts> => I<hash_options>

I<Optional.>
Hash reference of options for how to create each ACL.

      acl_masks     => Use a maximum of this number of access mask flags (0-18)
                       (default is 8; 0 means use maximum available).
      acl_flags     => Use a maximum of this number of flags (0-7)
                       (default is 4; 0 means use maximum available).
      acl_types     => Use a maximum of this number of types (0-3)
                       (default is 1; 0 means use maximum available).
      acl_users     => Use a maximum of this number of users (0-7)
                       (default is 2; 0 means use maximum available).
      users         => List of users/groups to use instead of the defaults.
      acl           => Create file with an ACL (1 or 2).
                       1: the number of ACEs varies between 1 and the product
                          of the acl_* options
                       2: the number of ACEs is the product of the
                          acl_* options (fixed).

B<sparse_opts> => I<hash_options>

I<Optional.>
Hash reference of options for how to create sparse files.

      ndata      => Number of data regions to create.
      ndata_dev  => Standard deviation of the number of data regions.
      offset     => Offset between data regions.
      offset_dev => Standard deviation of the offset between data regions.
      size       => Size of each data region.
      size_dev   => Standard deviation of the size of each data region.

B<fill_vol> => I<number>

I<Optional.>
The percentage of space usage that will not be exceeded.
Only supported on Windows. Default is 100.

B<timeout> => I<number>

I<Optional.>
Timeout value for function execution.
Default is 1800.

B<dbfile> => I<string>

I<Optional.>
Unique database file name to allow multiple clients to work in the same 
directory.

=item Example

=back

     populate(path => '/tmp/data');

=for html <hr />

=cut

#---------------------------------------------------------------------
sub populate(%) {
    my (%opts) = validate (
        @_,
        {
            path            => { type => SCALAR },
            client          => { type => SCALAR,   default => 'localhost' },
            template        => { type => SCALAR,   default => 'default' },
            name_avg        => { type => SCALAR,  optional => 1 },
            name_dev        => { type => SCALAR,  optional => 1 },
            dir_avg         => { type => SCALAR,  optional => 1 },
            dir_dev         => { type => SCALAR,  optional => 1 },
            file_avg        => { type => SCALAR,  optional => 1 },
            file_dev        => { type => SCALAR,  optional => 1 },
            file_size_avg   => { type => SCALAR,  optional => 1 },
            file_size_dev   => { type => SCALAR,  optional => 1 },
            depth_avg       => { type => SCALAR,  optional => 1 },
            depth_dev       => { type => SCALAR,  optional => 1 },
            stream_avg      => { type => SCALAR,  optional => 1 },
            stream_dev      => { type => SCALAR,  optional => 1 },
            stream_size_avg => { type => SCALAR,  optional => 1 },
            stream_size_dev => { type => SCALAR,  optional => 1 },
            dir_acl_num     => { type => SCALAR,  optional => 1 },
            dir_acl_pct     => { type => SCALAR,  optional => 1 },
            file_acl_num    => { type => SCALAR,  optional => 1 },
            file_acl_pct    => { type => SCALAR,  optional => 1 },
            dir_stream_num  => { type => SCALAR,  optional => 1 },
            dir_stream_pct  => { type => SCALAR,  optional => 1 },
            file_stream_num => { type => SCALAR,  optional => 1 },
            file_stream_pct => { type => SCALAR,  optional => 1 },
            sparse_num      => { type => SCALAR,  optional => 1 },
            sparse_pct      => { type => SCALAR,  optional => 1 },
            link_num        => { type => SCALAR,  optional => 1 },
            link_pct        => { type => SCALAR,  optional => 1 },
            slink_num       => { type => SCALAR,  optional => 1 },
            slink_pct       => { type => SCALAR,  optional => 1 },
            compress_num    => { type => SCALAR,  optional => 1 },
            compress_pct    => { type => SCALAR,  optional => 1 },
            encrypt_num     => { type => SCALAR,  optional => 1 },
            encrypt_pct     => { type => SCALAR,  optional => 1 },
            archive_num     => { type => SCALAR,  optional => 1 },
            archive_pct     => { type => SCALAR,  optional => 1 },
            hidden_num      => { type => SCALAR,  optional => 1 },
            hidden_pct      => { type => SCALAR,  optional => 1 },
            acl_opts        => { type => HASHREF, optional => 1 },
            sparse_opts     => { type => HASHREF, optional => 1 },
            fill_vol        => { type => SCALAR,   default => 100 },
            timeout         => { type => SCALAR,   default => 1800 },
            dbfile          => { type => SCALAR,  default => "" },
        }
    );
    my ($data, $sparse_opts, $acl_opts);

    # Set options according to template
    # Options in template are overwritten by explicitly setting the options
    $acl_opts    = $pop_template{$opts{template}}{acl_opts};
    $sparse_opts = $pop_template{$opts{template}}{sparse_opts};
    %opts = ( %{$pop_template{$opts{template}}}, %opts, _mode => 'populate' );

    # Sparse file options
    $opts{sparse_opts} = {
        # Values for sparse file options from template
        %$sparse_opts,
        # Overwrite defaults with user values
        %{$opts{sparse_opts}},
    };

    # ACL options
    $opts{acl_opts} = {
        # Values for ACL options from template
        %$acl_opts,
        # Overwrite defaults with user values
        %{$opts{acl_opts}},
    };

    # Create the data manipulation object
    $data = dp::snapmirror::qsm::TargetedData->new(
        path    => $opts{path},
        client  => $opts{client},
        fillpct => $opts{fill_vol},
        save    => 1,
        seed    => _get_seed,
        timeout => $opts{timeout},
        dbfile  => $opts{dbfile},
    );

    # Ignore unsupported options -- log a warning
    _check_options($data, %opts);

    # Initialize create options
    _init_create_opts(%opts);

    # Catch Control-C to save current state of data tree
    local $SIG{INT} = sub { exit 1 };

    # Create files on the root directory of the tree
    $data->create_files(%{$opts{_file_args}});

    # Create the rest of the tree
    if ( $opts{depth_avg} > 0 ) {
        $data->create_dirs(%{$opts{_dir_args}});
    }

    # Save the current counters so the next call to get_counters()
    # will return the latest information.
    $counters = $data->counters;

    return;
}

#---------------------------------------------------------------------

=head2 ::modify

Modify a rich data set.
The *_opt options is used to overwrite some of the main options, especially
the conditional options (dirs, files, streams, sparse, compress, encrypt,
archive, hidden and acl). See examples for more details.

=over

=item Arguments

B<path> => I<string>

I<Required.>
Top level directory of tree to modify.

B<client> => I<hostspec>

I<Optional.>
Client where modify will run. Default is localhost.

B<template> => I<string>

I<Optional.>
Template for modify which sets the default values of all the following
options. Default is the 'default' template.

B<name_avg> => I<number>

I<Optional.>
Average number of characters for the name of objects.

B<name_dev> => I<number>

I<Optional.>
Standard deviation of the number of characters for the name of objects.

B<dir_avg> => I<number>

I<Optional.>
Average number of directories to create under each directory.

B<dir_dev> => I<number>

I<Optional.>
Standard deviation of the number of directories to create under each directory.

B<file_avg> => I<number>

I<Optional.>
Average number of files to create under each directory.

B<file_dev> => I<number>

I<Optional.>
Standard deviation of the number of files to create under each directory.

B<file_size_avg> => I<number>

I<Optional.>
Average size of files.

B<file_size_dev> => I<number>

I<Optional.>
Standard deviation of the size of files.

B<stream_avg> => I<number>

I<Optional.>
Average number of streams to create on file/directory.

B<stream_dev> => I<number>

I<Optional.>
Standard deviation of the number of streams to create on file/directory.

B<stream_size_avg> => I<number>

I<Optional.>
Average size of streams.

B<stream_size_dev> => I<number>

I<Optional.>
Standard deviation of the size of streams.

B<dir_acl_num> => I<number>

I<Optional.>
Number of directories to add an ACL.

B<dir_acl_pct> => I<number>

I<Optional.>
Percentage of directories to add an ACL.

B<file_acl_num> => I<number>

I<Optional.>
Number of files to add an ACL.

B<file_acl_pct> => I<number>

I<Optional.>
Percentage of files to add an ACL.

B<dir_stream_num> => I<number>

I<Optional.>
Number of directories to add streams.

B<dir_stream_pct> => I<number>

I<Optional.>
Percentage of directories to add streams.

B<file_stream_num> => I<number>

I<Optional.>
Number of files to add streams.

B<file_stream_pct> => I<number>

I<Optional.>
Percentage of files to add streams.

B<sparse_num => I<number>

I<Optional.>
Number of files to be sparse files.

B<sparse_pct> => I<number>

I<Optional.>
Percentage of files to be sparse files.

B<link_num> => I<number>

I<Optional.>
Number of files to be links.

B<link_pct> => I<number>

I<Optional.>
Percentage of files to be links.

B<slink_num> => I<number>

I<Optional.>
Number of files to be symbolic links.

B<slink_pct> => I<number>

I<Optional.>
Percentage of files to be symbolic links.

B<compress_num> => I<number>

I<Optional.>
Number of objects to set the compress attribute.

B<compress_pct> => I<number>

I<Optional.>
Percentage of objects to set the compress attribute.

B<encrypt_num=> I<number>

I<Optional.>
Number of objects to set the encrypt attribute.

B<encrypt_pct> => I<number>

I<Optional.>
Percentage of objects to set the encrypt attribute.

B<archive_num> => I<number>

I<Optional.>
Number of objects to set the archive attribute.

B<archive_pct> => I<number>

I<Optional.>
Percentage of objects to set the archive attribute.

B<hidden_num> => I<number>

I<Optional.>
Number of objects to set the hidden attribute.

B<hidden_pct> => I<number>

I<Optional.>
Percentage of objects to set the hidden attribute.

B<create_on_dir_num> => I<number>

I<Optional.>
Number of directories to add more objects.

B<create_on_dir_pct> => I<number>

I<Optional.>
Percentage of directories to add more objects.

B<create_on_file_num> => I<number>

I<Optional.>
Number of files to add more objects.

B<create_on_file_pct> => I<number>

I<Optional.>
Percentage of files to add more objects.

B<modify_file_num> => I<number>

I<Optional.>
Number of files to modify.

B<modify_file_pct> => I<number>

I<Optional.>
Percentage of files to modify.

B<modify_stream_num> => I<number>

I<Optional.>
Number of streams to modify.

B<modify_stream_pct> => I<number>

I<Optional.>
Percentage of streams to modify.

B<modify_sparse_num> => I<number>

I<Optional.>
Number of sparse files to modify.

B<modify_sparse_pct> => I<number>

I<Optional.>
Percentage of sparse files to modify.

B<remove_dir_num> => I<number>

I<Optional.>
Number of directories to remove. It only removes the leaf directories
(directories which do not contain other directories)

B<remove_dir_pct> => I<number>

I<Optional.>
Percentage of directories to remove. It only removes the leaf directories
(directories which do not contain other directories)

B<remove_file_num> => I<number>

I<Optional.>
Number of files to remove.

B<remove_file_pct> => I<number>

I<Optional.>
Percentage of files to remove.

B<remove_stream_num> => I<number>

I<Optional.>
Number of streams to remove.

B<remove_stream_pct> => I<number>

I<Optional.>
Percentage of streams to remove.

B<rename_dir_num> => I<number>

I<Optional.>
Number of directories to rename.

B<rename_dir_pct> => I<number>

I<Optional.>
Percentage of directories to rename.

B<rename_file_num> => I<number>

I<Optional.>
Number of files to rename.

B<rename_file_pct> => I<number>

I<Optional.>
Percentage of files to rename.

B<truncate_file_num> => I<number>

I<Optional.>
Number of files to truncate.

B<truncate_file_pct> => I<number>

I<Optional.>
Percentage of files to truncate.

B<truncate_stream_num> => I<number>

I<Optional.>
Number of streams to truncate.

B<truncate_stream_pct> => I<number>

I<Optional.>
Percentage of streams to truncate.

B<acl_mod_dir_num> => I<number>

I<Optional.>
Maximum number of directories to modify ACL's.

B<acl_mod_dir_pct> => I<number>
Percentage of directories to modify ACL's.

I<Optional.>

B<acl_mod_file_num> => I<number>

I<Optional.>
Maximum number of files to modify ACL's.

B<acl_mod_file_pct> => I<number>

I<Optional.>
Percentage of files to modify ACL's.

B<acl_del_dir_num> => I<number>

I<Optional.>
Maximum number of directories to delete ACL's.

B<acl_del_dir_pct> => I<number>

I<Optional.>
Percentage of directories to delete ACL's.

B<acl_del_file_num> => I<number>

I<Optional.>
Maximum number of files to delete ACL's.

B<acl_del_file_pct> => I<number>

I<Optional.>
Percentage of files to delete ACL's.

B<sattrs_dir_num> => I<number>

I<Optional.>
Maximum number of directories to set attributes.

B<sattrs_dir_pct> => I<number>

I<Optional.>
Percentage of directories to set attributes.

B<sattrs_file_num> => I<number>

I<Optional.>
Maximum number of files to set attributes.

B<sattrs_file_pct> => I<number>

I<Optional.>
Percentage of files to set attributes.

B<dirs> => C<yes|no|any>

I<Optional.>
Modify directories if they contain directories.

B<files> => C<yes|no|any>

I<Optional.>
Modify directories if they contain files.

B<streams> => C<yes|no|any>

I<Optional.>
Modify directories/files if they contain streams.

B<sparse> => C<yes|no|any>

I<Optional.>
Modify files if they are sparse files.

B<compress> => C<yes|no|any>

I<Optional.>
Modify files if they are compressed.

B<encrypt> => C<yes|no|any>

I<Optional.>
Modify files if they are encrypted.

B<archive> => C<yes|no|any>

I<Optional.>
Modify files if they are archived.

B<hidden> => C<yes|no|any>

I<Optional.>
Modify files if they are hidden.

B<acl> => C<yes|no|any>

I<Optional.>
Modify directories/files with ACL.

B<modify_file_opt> => I<hash_options>

I<Optional.>
Modify options for files. Half of the files to be modified use a full
modify which replaces the entire contents of the file. The other half,
use a block modify which only changes the contents of the given number
of blocks.

      file_size_avg => New average size of file (full modify).
                       Default is the value of main file_size_avg option.
      file_size_dev => Standard deviation of new size of file (full modify).
                       Default is the value of main file_size_dev option.
      blocks        => Number of blocks to modify. Default is 4.
      block_size    => Size of each block to modify. Default is 1024.

B<modify_stream_opt> => I<hash_options>

I<Optional.>
Modify options for streams. Same as modify_file_opt.

B<modify_sparse_opt> => I<hash_options>

I<Optional.>
Modify options for sparse files. Overwrite the sparse options given by
the main option sparse_opts.

B<remove_dir_opt> => I<hash_options>

I<Optional.>
Remove options for directories.

B<remove_file_opt> => I<hash_options>

I<Optional.>
Remove options for files.

B<rename_dir_opt> => I<hash_options>

I<Optional.>
Rename options for directories.

B<rename_file_opt> => I<hash_options>

I<Optional.>
Rename options for files.

B<truncate_file_opt> => I<hash_options>

I<Optional.>
Truncate options for files.

      file_size_avg => New average size of file.
                       Default is the value of main file_size_avg option.
      file_size_dev => Standard deviation of new size of file.
                       Default is the value of main file_size_dev option.

B<truncate_stream_opt> => I<hash_options>

I<Optional.>
Truncate options for files. Same as truncate_file_opt.

B<acl_mod_dir_opt> => I<hash_options>

I<Optional.>
Modify ACL options for directories.

B<acl_mod_file_opt> => I<hash_options>

I<Optional.>
Modify ACL options for files.

B<acl_del_dir_opt> => I<hash_options>

I<Optional.>
Remove ACL options for directories.

B<acl_del_file_opt> => I<hash_options>

I<Optional.>
Remove ACL options for files.

B<sattrs_dir_opt> => I<hash_options>

I<Optional.>
Set attributes options for directories.

B<sattrs_file_opt> => I<hash_options>

I<Optional.>
Set attributes options for files.

B<acl_opts> => I<hash_options>

I<Optional.>
Hash reference of options for how to create each ACL.

      acl_masks => Use a maximum of this number of access mask flags (0-18)
                   (default is 8; 0 means use maximum available).
      acl_flags => Use a maximum of this number of flags (0-7)
                   (default is 4; 0 means use maximum available).
      acl_types => Use a maximum of this number of types (0-3)
                   (default is 1; 0 means use maximum available).
      acl_users => Use a maximum of this number of users (0-7)
                   (default is 2; 0 means use maximum available).
      users     => List of users/groups to use instead of the defaults.
      acl       => Create file with an ACL (1 or 2).
                   1: the number of ACEs varies between 1 and the product
                      of the acl_* options
                   2: the number of ACEs is the product of the
                      acl_* options (fixed).

B<sparse_opts> => I<hash_options>

I<Optional.>
Hash reference of options for how to create sparse files.

      ndata      => Number of data regions to create.
      ndata_dev  => Standard deviation of the number of data regions.
      offset     => Offset between data regions.
      offset_dev => Standard deviation of the offset between data regions.
      size       => Size of each data region.
      size_dev   => Standard deviation of the size of each data region.

B<sattrs_opts> => I<hash_options>

I<Optional.>
Hash reference of options for attributes.

      compress => Set or clear the compress attribute.
      encrypt  => Set or clear the encrypt attribute.
      archive  => Set or clear the archive attribute.
      hidden   => Set or clear the hidden attribute.

B<fill_vol> => I<number>

I<Optional.>
The percentage of space usage that will not be exceeded.
Only supported on Windows. Default is 100.

B<timeout> => I<number>

I<Optional.>
Timeout value for function execution.
Default is 1800.

B<dbfile> => I<string>

I<Optional.>
Unique database file name to allow multiple clients to work in the same
directory.


=item Example

=back

     # Modify tree using the defaults in the 'default' template
     modify(path => '/tmp/data');

     # Create new objects and modify existing objects only on files/directories
     # which contain ACL's, but only modify contents of files without ACL's
     modify(
         path => '/tmp/data'
         acl => 'yes',
         modify_file_opt => {
             # Overwrite main option
             acl => 'no',
         },
     );

=for html <hr />

=cut

#---------------------------------------------------------------------
sub modify(%) {
    my (%opts) = validate (
        @_,
        {
            path                => { type => SCALAR },
            client              => { type => SCALAR,   default => 'localhost' },
            template            => { type => SCALAR,   default => 'default' },
            name_avg            => { type => SCALAR,  optional => 1 },
            name_dev            => { type => SCALAR,  optional => 1 },
            dir_avg             => { type => SCALAR,  optional => 1 },
            dir_dev             => { type => SCALAR,  optional => 1 },
            file_avg            => { type => SCALAR,  optional => 1 },
            file_dev            => { type => SCALAR,  optional => 1 },
            file_size_avg       => { type => SCALAR,  optional => 1 },
            file_size_dev       => { type => SCALAR,  optional => 1 },
            stream_avg          => { type => SCALAR,  optional => 1 },
            stream_dev          => { type => SCALAR,  optional => 1 },
            stream_size_avg     => { type => SCALAR,  optional => 1 },
            stream_size_dev     => { type => SCALAR,  optional => 1 },
            dir_acl_num         => { type => SCALAR,  optional => 1 },
            dir_acl_pct         => { type => SCALAR,  optional => 1 },
            file_acl_num        => { type => SCALAR,  optional => 1 },
            file_acl_pct        => { type => SCALAR,  optional => 1 },
            dir_stream_num      => { type => SCALAR,  optional => 1 },
            dir_stream_pct      => { type => SCALAR,  optional => 1 },
            file_stream_num     => { type => SCALAR,  optional => 1 },
            file_stream_pct     => { type => SCALAR,  optional => 1 },
            sparse_num          => { type => SCALAR,  optional => 1 },
            sparse_pct          => { type => SCALAR,  optional => 1 },
            link_num            => { type => SCALAR,  optional => 1 },
            link_pct            => { type => SCALAR,  optional => 1 },
            slink_num           => { type => SCALAR,  optional => 1 },
            slink_pct           => { type => SCALAR,  optional => 1 },
            compress_num        => { type => SCALAR,  optional => 1 },
            compress_pct        => { type => SCALAR,  optional => 1 },
            encrypt_num         => { type => SCALAR,  optional => 1 },
            encrypt_pct         => { type => SCALAR,  optional => 1 },
            archive_num         => { type => SCALAR,  optional => 1 },
            archive_pct         => { type => SCALAR,  optional => 1 },
            hidden_num          => { type => SCALAR,  optional => 1 },
            hidden_pct          => { type => SCALAR,  optional => 1 },
            create_on_dir_num   => { type => SCALAR,  optional => 1 },
            create_on_dir_pct   => { type => SCALAR,  optional => 1 },
            create_on_file_num  => { type => SCALAR,  optional => 1 },
            create_on_file_pct  => { type => SCALAR,  optional => 1 },
            modify_file_num     => { type => SCALAR,  optional => 1 },
            modify_file_pct     => { type => SCALAR,  optional => 1 },
            modify_stream_num   => { type => SCALAR,  optional => 1 },
            modify_stream_pct   => { type => SCALAR,  optional => 1 },
            modify_sparse_num   => { type => SCALAR,  optional => 1 },
            modify_sparse_pct   => { type => SCALAR,  optional => 1 },
            remove_dir_num      => { type => SCALAR,  optional => 1 },
            remove_dir_pct      => { type => SCALAR,  optional => 1 },
            remove_file_num     => { type => SCALAR,  optional => 1 },
            remove_file_pct     => { type => SCALAR,  optional => 1 },
            remove_stream_num   => { type => SCALAR,  optional => 1 },
            remove_stream_pct   => { type => SCALAR,  optional => 1 },
            rename_dir_num      => { type => SCALAR,  optional => 1 },
            rename_dir_pct      => { type => SCALAR,  optional => 1 },
            rename_file_num     => { type => SCALAR,  optional => 1 },
            rename_file_pct     => { type => SCALAR,  optional => 1 },
            truncate_file_num   => { type => SCALAR,  optional => 1 },
            truncate_file_pct   => { type => SCALAR,  optional => 1 },
            truncate_stream_num => { type => SCALAR,  optional => 1 },
            truncate_stream_pct => { type => SCALAR,  optional => 1 },
            acl_mod_dir_num     => { type => SCALAR,  optional => 1 },
            acl_mod_dir_pct     => { type => SCALAR,  optional => 1 },
            acl_mod_file_num    => { type => SCALAR,  optional => 1 },
            acl_mod_file_pct    => { type => SCALAR,  optional => 1 },
            acl_del_dir_num     => { type => SCALAR,  optional => 1 },
            acl_del_dir_pct     => { type => SCALAR,  optional => 1 },
            acl_del_file_num    => { type => SCALAR,  optional => 1 },
            acl_del_file_pct    => { type => SCALAR,  optional => 1 },
            sattrs_dir_num      => { type => SCALAR,  optional => 1 },
            sattrs_dir_pct      => { type => SCALAR,  optional => 1 },
            sattrs_file_num     => { type => SCALAR,  optional => 1 },
            sattrs_file_pct     => { type => SCALAR,  optional => 1 },
            dirs                => { type => SCALAR,  optional => 1 },
            files               => { type => SCALAR,  optional => 1 },
            streams             => { type => SCALAR,  optional => 1 },
            sparse              => { type => SCALAR,  optional => 1 },
            compress            => { type => SCALAR,  optional => 1 },
            encrypt             => { type => SCALAR,  optional => 1 },
            archive             => { type => SCALAR,  optional => 1 },
            hidden              => { type => SCALAR,  optional => 1 },
            acl                 => { type => SCALAR,  optional => 1 },
            modify_file_opt     => { type => HASHREF, optional => 1 },
            modify_stream_opt   => { type => HASHREF, optional => 1 },
            modify_sparse_opt   => { type => HASHREF, optional => 1 },
            remove_dir_opt      => { type => HASHREF, optional => 1 },
            remove_file_opt     => { type => HASHREF, optional => 1 },
            rename_dir_opt      => { type => HASHREF, optional => 1 },
            rename_file_opt     => { type => HASHREF, optional => 1 },
            truncate_file_opt   => { type => HASHREF, optional => 1 },
            truncate_stream_opt => { type => HASHREF, optional => 1 },
            acl_mod_dir_opt     => { type => HASHREF, optional => 1 },
            acl_mod_file_opt    => { type => HASHREF, optional => 1 },
            acl_del_dir_opt     => { type => HASHREF, optional => 1 },
            acl_del_file_opt    => { type => HASHREF, optional => 1 },
            sattrs_dir_opt      => { type => HASHREF, optional => 1 },
            sattrs_file_opt     => { type => HASHREF, optional => 1 },
            acl_opts            => { type => HASHREF, optional => 1 },
            sparse_opts         => { type => HASHREF, optional => 1 },
            sattrs_opts         => { type => HASHREF, optional => 1 },
            fill_vol            => { type => SCALAR,   default => 100 },
            timeout             => { type => SCALAR,   default => 1800 },
            dbfile              => { type => SCALAR,  default => "" },
        }
    );
    my ($data, %dir_opts, %file_opts, %stream_opts, %args);
    my (%fcont_opts, %dcont_opts, $sparse_opts, $acl_opts, $sattrs_opts);
    my (%cont_opts, $size, $dev);
    my $dir_pct    = 100;
    my $file_pct   = 100;
    my $stream_pct = 100;

    # Set options according to template
    # Options in template are overwritten by explicitly setting the options
    $acl_opts    = $mod_template{$opts{template}}{acl_opts};
    $sparse_opts = $mod_template{$opts{template}}{sparse_opts};
    $sattrs_opts = $mod_template{$opts{template}}{sattrs_opts};
    %opts = ( %{$mod_template{$opts{template}}}, %opts, _mode => 'modify' );

    # ACL options
    $opts{acl_opts} = {
        # Values for ACL options from template
        %$acl_opts,
        # Overwrite defaults with user values
        %{$opts{acl_opts}},
    };

    # Sparse file options
    $opts{sparse_opts} = {
        # Values for sparse file options from template
        %$sparse_opts,
        # Overwrite defaults with user values
        %{$opts{sparse_opts}},
    };

    # Set attribute options
    $opts{sattrs_opts} = {
        # Values for set attribute options from template
        %$sattrs_opts,
        # Overwrite defaults with user values
        %{$opts{sattrs_opts}},
    };

    # Create the data manipulation object
    $data = dp::snapmirror::qsm::TargetedData->new(
        path    => $opts{path},
        client  => $opts{client},
        fillpct => $opts{fill_vol},
        save    => 1,
        create  => 0,
        seed    => _get_seed,
        timeout => $opts{timeout},
        dbfile  => $opts{dbfile},
    );

    # Set all elements to original
    $data->set_original;

    # Ignore unsupported options -- log a warning
    _check_options($data, %opts);

    # Initialize create options
    _init_create_opts(%opts);

    # Catch Control-C to save current state of data tree
    local $SIG{INT} = sub { exit 1 };

    # Get main 'contain' attributes from options
    %cont_opts = map { $_ => $opts{$_} } grep { exists $opts{$_} } @cont_opts;

    # Setup both files and dirs 'contain' options
    %dcont_opts = %cont_opts,
    %fcont_opts = %cont_opts,
    delete $fcont_opts{dirs};
    delete $fcont_opts{files};
    delete $dcont_opts{sparse};

    # Options for removing directories which have no directories (dir leaves)
    %dir_opts = (
        _get_cont_opts(%dcont_opts, $opts{remove_dir_opt}),
        dirs     => 'no',
        number   => $opts{remove_dir_num},
        percent  => 100 * $opts{remove_dir_pct} / $dir_pct,
        original => 1,
        mode     => 'remove',
    );

    # Options for removing files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{remove_file_opt}),
        number   => $opts{remove_file_num},
        percent  => 100 * $opts{remove_file_pct} / $file_pct,
        original => 1,
        mode     => 'remove',
    );

    # Options for removing streams
    %stream_opts = (
        number   => $opts{remove_stream_num},
        percent  => 100 * $opts{remove_stream_pct} / $stream_pct,
        original => 1,
        mode     => 'remove',
    );

    # Remove dirs/files/streams
    $data->modify(
        dirs    => \%dir_opts,
        files   => \%file_opts,
        streams => \%stream_opts,
    );

    # Calculate remaining percentage for next modify
    $dir_pct    -= $opts{remove_dir_pct};
    $file_pct   -= $opts{remove_file_pct};
    $stream_pct -= $opts{remove_stream_pct};
    $dir_pct     = 1 if $dir_pct    < 1;
    $file_pct    = 1 if $file_pct   < 1;
    $stream_pct  = 1 if $stream_pct < 1;

    # Options for modifying files (full file modify)
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{modify_file_opt}),
        sparse   => 'no',
        number   => $opts{modify_file_num} / 2,
        percent  => 50 * $opts{modify_file_pct} / $file_pct,
        original => 1,
        mode     => 'modify',
        args     => {
            size     => _get_args(%opts, 'file_size_avg', 'modify_file_opt'),
            size_dev => _get_args(%opts, 'file_size_dev', 'modify_file_opt'),
        },
    );

    # Options for modifying streams (some blocks modify)
    # XXX FileManip does not support block modify for solaris
    if ( $data->hosttype =~ /solaris/ ) {
        %args = (
            size     => _get_args(%opts, 'stream_size_avg','modify_stream_opt'),
            size_dev => _get_args(%opts, 'stream_size_dev','modify_stream_opt'),
        );
    } else {
        $size = _get_args(%opts, 'stream_size_avg', 'modify_stream_opt');
        %args = (
            blocks   => exists $opts{modify_stream_opt}{blocks} ?
                               $opts{modify_stream_opt}{blocks} : 4,
            size     => exists $opts{modify_stream_opt}{block_size} ?
                               $opts{modify_stream_opt}{block_size} : 1024,
            max_size => $size,
        );
    }
    %stream_opts = (
        number   => $opts{modify_stream_num} / 2,
        percent  => 50 * $opts{modify_stream_pct} / $stream_pct,
        original => 1,
        mode     => 'modify',
        args     => \%args,
    );

    # Modify files/streams
    $data->modify(
        files   => \%file_opts,
        streams => \%stream_opts,
    );

    # Swap the modification args so
    # full stream modify and some blocks modify for files
    $file_opts{args} = {
        blocks   => exists $opts{modify_file_opt}{blocks} ?
                           $opts{modify_file_opt}{blocks} : 4,
        size     => exists $opts{modify_file_opt}{block_size} ?
                           $opts{modify_file_opt}{block_size} : 1024,
        max_size => _get_args(%opts, 'file_size_avg', 'modify_file_opt'),
    };
    $stream_opts{args} = {
        size     => _get_args(%opts, 'stream_size_avg','modify_stream_opt'),
        size_dev => _get_args(%opts, 'stream_size_dev','modify_stream_opt'),
    };

    # Make sure it modifies the correct number
    $file_opts{number}   += $opts{modify_file_num} % 2;
    $stream_opts{number} += $opts{modify_stream_num} % 2;

    # Modify files/streams
    $data->modify(
        files   => \%file_opts,
        streams => \%stream_opts,
    );

    # Calculate remaining percentage for next modify
    $file_pct   -= $opts{modify_file_pct};
    $stream_pct -= $opts{modify_stream_pct};
    $file_pct    = 1 if $file_pct   < 1;
    $stream_pct  = 1 if $stream_pct < 1;

    # Options for modifying sparse files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{modify_sparse_opt}),
        sparse   => 'yes',
        number   => $opts{modify_sparse_num},
        percent  => $opts{modify_sparse_pct},
        original => 1,
        mode     => 'modify',
        args     => {
            sparse_opts => { %{$opts{sparse_opts}},
                             (exists $opts{modify_sparse_opt}{sparse_opts} ?
                              %{$opts{modify_sparse_opt}{sparse_opts}} : ()) }
        },
    );

    # Modify sparse files
    $data->modify(
        files => \%file_opts,
    );

    # Options for renaming directories
    %dir_opts = (
        _get_cont_opts(%dcont_opts, $opts{rename_dir_opt}),
        number   => $opts{rename_dir_num},
        percent  => 100 * $opts{rename_dir_pct} / $dir_pct,
        original => 1,
        mode     => 'rename',
    );

    # Options for renaming files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{rename_file_opt}),
        number   => $opts{rename_file_num},
        percent  => 100 * $opts{rename_file_pct} / $file_pct,
        original => 1,
        mode     => 'rename',
    );

    # Rename dirs/files
    $data->modify(
        dirs  => \%dir_opts,
        files => \%file_opts,
    );

    # Calculate remaining percentage for next modify
    $dir_pct  -= $opts{rename_dir_pct};
    $file_pct -= $opts{rename_file_pct};
    $dir_pct   = 1 if $dir_pct  < 1;
    $file_pct  = 1 if $file_pct < 1;

    # Options for truncating files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{truncate_file_opt}),
        number   => $opts{truncate_file_num},
        percent  => 100 * $opts{truncate_file_pct} / $file_pct,
        original => 1,
        mode     => 'trunc',
        args     => {
            size     => _get_args(%opts, 'file_size_avg', 'truncate_file_opt'),
            size_dev => _get_args(%opts, 'file_size_dev', 'truncate_file_opt'),
        },
    );

    # Options for truncating streams
    $size = _get_args(%opts, 'stream_size_avg', 'truncate_stream_opt');
    $dev  = _get_args(%opts, 'stream_size_dev', 'truncate_stream_opt');
    %stream_opts = (
        number   => $opts{truncate_stream_num},
        percent  => 100 * $opts{truncate_stream_pct} / $stream_pct,
        original => 1,
        mode     => 'trunc',
        args     => {
            size     => $size,
            size_dev => $dev,
        },
    );

    # Truncate files/streams
    $data->modify(
        files   => \%file_opts,
        streams => \%stream_opts,
    );

    # Calculate remaining percentage for next modify
    $file_pct   -= $opts{truncate_file_pct};
    $stream_pct -= $opts{truncate_stream_pct};
    $file_pct    = 1 if $file_pct   < 1;
    $stream_pct  = 1 if $stream_pct < 1;

    # Options for modifying ACL's from directories
    %dir_opts = (
        _get_cont_opts(%dcont_opts, $opts{acl_mod_dir_opt}),
        number   => $opts{acl_mod_dir_num},
        percent  => 100 * $opts{acl_mod_dir_pct} / $dir_pct,
        original => 1,
        mode     => 'acl_mod',
    );

    # Options for modifying ACL's from files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{acl_mod_file_opt}),
        number   => $opts{acl_mod_file_num},
        percent  => 100 * $opts{acl_mod_file_pct} / $file_pct,
        original => 1,
        mode     => 'acl_mod',
    );

    # Modify ACL's
    $data->modify(
        dirs  => \%dir_opts,
        files => \%file_opts,
    );

    # Calculate remaining percentage for next modify
    $dir_pct  -= $opts{acl_mod_dir_pct};
    $file_pct -= $opts{acl_mod_file_pct};
    $dir_pct   = 1 if $dir_pct  < 1;
    $file_pct  = 1 if $file_pct < 1;

    # Options for removing ACL's from directories
    %dir_opts = (
        _get_cont_opts(%dcont_opts, $opts{acl_del_dir_opt}),
        number   => $opts{acl_del_dir_num},
        percent  => 100 * $opts{acl_del_dir_pct} / $dir_pct,
        original => 1,
        mode     => 'acl_del',
    );

    # Options for removing ACL's from files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{acl_del_file_opt}),
        number   => $opts{acl_del_file_num},
        percent  => 100 * $opts{acl_del_file_pct} / $file_pct,
        original => 1,
        mode     => 'acl_del',
    );

    # Remove ACL's
    $data->modify(
        dirs  => \%dir_opts,
        files => \%file_opts,
    );

    # Calculate remaining percentage for next modify
    $dir_pct  -= $opts{acl_del_dir_pct};
    $file_pct -= $opts{acl_del_file_pct};
    $dir_pct   = 1 if $dir_pct  < 1;
    $file_pct  = 1 if $file_pct < 1;

    # Options for setting attributes on directories
    %dir_opts = (
        _get_cont_opts(%dcont_opts, $opts{sattrs_dir_opt}),
        number   => $opts{sattrs_dir_num},
        percent  => 100 * $opts{sattrs_dir_pct} / $dir_pct,
        original => 1,
        mode     => 'sattrs',
        args     => $opts{sattrs_opts},
    );

    # Options for setting attributes on files
    %file_opts = (
        _get_cont_opts(%fcont_opts, $opts{sattrs_file_opt}),
        number   => $opts{sattrs_file_num},
        percent  => 100 * $opts{sattrs_file_pct} / $file_pct,
        original => 1,
        mode     => 'sattrs',
        args     => $opts{sattrs_opts},
    );

    # Modify attributes
    $data->modify(
        dirs  => \%dir_opts,
        files => \%file_opts,
    );

    # Calculate remaining percentage for next modify
    $dir_pct  -= $opts{acl_del_dir_pct};
    $file_pct -= $opts{acl_del_file_pct};
    $dir_pct   = 1 if $dir_pct  < 1;
    $file_pct  = 1 if $file_pct < 1;

    # Options for adding new objects to directories
    %dir_opts = (
        %dcont_opts,
        number   => $opts{create_on_dir_num},
        percent  => 100 * $opts{create_on_dir_pct} / $dir_pct,
        original => 1,
        depth    => 0, # Maintain same depth
        mode     => 'create',
        args     => $opts{_dir_args},
    );

    # Options for adding new objects to files
    %file_opts = (
        %fcont_opts,
        number   => $opts{create_on_file_num},
        percent  => 100 * $opts{create_on_file_pct} / $file_pct,
        original => 1,
        mode     => 'create',
        args     => $opts{_file_args},
    );

    # Create more file system objects
    $data->modify(
        dirs  => \%dir_opts,
        files => \%file_opts,
    );

    # Save the current counters so the next call to get_counters()
    # will return the latest information.
    $counters = $data->counters;

    return;
}

#---------------------------------------------------------------------

=head2 ::read_tree

Read contents of files and streams.

=over

=item Arguments

B<path> => I<string>

I<Required.>
Top level directory of tree to read.

B<client> => I<hostspec>

I<Optional.>
Client where read will run. Default is localhost.

B<template> => I<string>

I<Optional.>
Template for reading a tree which sets the default values of all the following
options. Default is the 'default' template.

B<file_num> => I<number>

I<Optional.>
Number of files to read.

B<file_pct> => I<number>

I<Optional.>
Percentage of files to read.

B<sparse_num> => I<number>

I<Optional.>
Number of files to be sparse files.

B<sparse_pct> => I<number>

I<Optional.>
Percentage of files to be sparse files.

B<stream_num> => I<number>

I<Optional.>
Number of streams to read.

B<stream_pct> => I<number>

I<Optional.>
Percentage of streams to read.

B<dir_acl_num> => I<number>

I<Optional.>
Number of directory ACL's to read.

B<dir_acl_pct> => I<number>

I<Optional.>
Percentage of directory ACL's to read.

B<file_acl_num> => I<number>

I<Optional.>
Number of file ACL's to read.

B<file_acl_pct> => I<number>

I<Optional.>
Percentage of file ACL's to read.

B<dir_attr_num> => I<number>

I<Optional.>
Number of directory attributes to read.

B<dir_attr_pct> => I<number>

I<Optional.>
Percentage of directory attributes to read.

B<file_attr_num> => I<number>

I<Optional.>
Number of file attributes to read.

B<file_attr_pct> => I<number>

I<Optional.>
Percentage of file attributes to read.

B<dirs> => C<yes|no|any>

I<Optional.>
Read directories if they contain directories.

B<files> => C<yes|no|any>

I<Optional.>
Read directories if they contain files.

B<sparse> => C<yes|no|any>

I<Optional.>
Read files if they are sparse files.

B<streams> => C<yes|no|any>

I<Optional.>
Read directories/files if they contain streams.

B<compress> => C<yes|no|any>

I<Optional.>
Read files if they are compressed.

B<encrypt> => C<yes|no|any>

I<Optional.>
Read files if they are encrypted.

B<archive> => C<yes|no|any>

I<Optional.>
Read files if they are archived.

B<hidden> => C<yes|no|any>

I<Optional.>
Read files if they are hidden.

B<acl> => C<yes|no|any>

I<Optional.>
Modify directories/files with ACL.

B<timeout> => I<number>

I<Optional.>
Timeout value for function execution.
Default is 1800.

B<dbfile> => I<string>

I<Optional.>
Unique database file name to allow multiple clients to work in the same
directory.

=item Example

=back

     read_tree(path => '/tmp/data');

=for html <hr />

=cut

#---------------------------------------------------------------------
sub read_tree(%) {
    my (%opts) = validate (
        @_,
        {
            path          => { type => SCALAR },
            client        => { type => SCALAR,  default => 'localhost' },
            template      => { type => SCALAR,  default => 'default' },
            file_num      => { type => SCALAR, optional => 1 },
            file_pct      => { type => SCALAR, optional => 1 },
            sparse_num    => { type => SCALAR, optional => 1 },
            sparse_pct    => { type => SCALAR, optional => 1 },
            stream_num    => { type => SCALAR, optional => 1 },
            stream_pct    => { type => SCALAR, optional => 1 },
            dir_acl_num   => { type => SCALAR, optional => 1 },
            dir_acl_pct   => { type => SCALAR, optional => 1 },
            file_acl_num  => { type => SCALAR, optional => 1 },
            file_acl_pct  => { type => SCALAR, optional => 1 },
            dir_attr_num  => { type => SCALAR, optional => 1 },
            dir_attr_pct  => { type => SCALAR, optional => 1 },
            file_attr_num => { type => SCALAR, optional => 1 },
            file_attr_pct => { type => SCALAR, optional => 1 },
            dirs          => { type => SCALAR, optional => 1 },
            files         => { type => SCALAR, optional => 1 },
            sparse        => { type => SCALAR, optional => 1 },
            streams       => { type => SCALAR, optional => 1 },
            compress      => { type => SCALAR, optional => 1 },
            encrypt       => { type => SCALAR, optional => 1 },
            archive       => { type => SCALAR, optional => 1 },
            hidden        => { type => SCALAR, optional => 1 },
            acl           => { type => SCALAR, optional => 1 },
            timeout       => { type => SCALAR,  default => 1800 },
            dbfile        => { type => SCALAR, default => "" },
        },
    );
    my ($data, %cont_opts, %dcont_opts, %fcont_opts);

    # Set options according to template
    # Options in template are overwritten by explicitly setting the options
    # The mode is set to read because we are in the read_tree subroutine.
    %opts = ( %{$read_template{$opts{template}}}, %opts, _mode => 'read' );

    # Create the data manipulation object
    $data = dp::snapmirror::qsm::TargetedData->new(
        path    => $opts{path},
        client  => $opts{client},
        save    => 1,
        create  => 0,
        seed    => _get_seed,
        timeout => $opts{timeout},
        dbfile  => $opts{dbfile},
    );

    # Set all elements to original
    $data->set_original;

    # Ignore unsupported options -- log a warning
    _check_options($data, %opts);

    # Get main 'contain' attributes from options
    %cont_opts = map { $_ => $opts{$_} } grep { exists $opts{$_} } @cont_opts;

    # Setup both files and dirs options for create
    %dcont_opts = %cont_opts,
    %fcont_opts = %cont_opts,
    delete $fcont_opts{dirs};
    delete $fcont_opts{files};
    delete $dcont_opts{sparse};

    $data->read_tree(
        dirs    => {
            %dcont_opts,
            number  => 99999999999,
            percent => 100,
            args    => {
                acl_num  => $opts{dir_acl_num},
                acl_pct  => $opts{dir_acl_pct},
                attr_num => $opts{dir_attr_num},
                attr_pct => $opts{dir_attr_pct},
            },
        },
        files   => {
            %fcont_opts,
            number  => $opts{file_num},
            percent => $opts{file_pct},
            args    => {
                acl_num    => $opts{file_acl_num},
                acl_pct    => $opts{file_acl_pct},
                attr_num   => $opts{file_attr_num},
                attr_pct   => $opts{file_attr_pct},
                sparse_num => $opts{sparse_num},
                sparse_pct => $opts{sparse_pct},
                links_num  => $opts{links_num},
                links_pct  => $opts{links_pct},
                slinks_num => $opts{slinks_num},
                slinks_pct => $opts{slinks_pct},
            },
        },
        streams => {
            number  => $opts{stream_num},
            percent => $opts{stream_pct},
        },
    );

    # Save the current counters
    $counters = $data->counters;

    return;
}

#---------------------------------------------------------------------

=head2 ::delete_tree

Remove the directory tree.

The following named arguments are accepted:

=over

=item Arguments

B<path> => I<string>

I<Required.>
Top level directory of tree to removed.

B<client> => I<hostspec>

I<Optional.>
Client where read will run. Default is localhost.

B<contents_only> => C<0|1>

I<Optional.>
If set, do not delete tree itself, just everything within it.
Defaults to 1.

B<timeout> => I<number>

I<Optional.>
Timeout value for function execution.
Default is 1800.

B<dbfile> => I<string>

I<Optional.>
Unique database file name to allow multiple clients to work in the same
directory.

=item Returns

A hash reference of the template.

=item Example

=back

     delete_tree(path => '/tmp/data');

=for html <hr />

=cut

#---------------------------------------------------------------------
sub delete_tree(%) {
    my (%opts) = validate (
        @_,
        {
            path          => { type => SCALAR },
            client        => { type => SCALAR, default => 'localhost' },
            contents_only => { type => SCALAR, default => 1 },
            timeout       => { type => SCALAR, default => 1800 },
            dbfile        => { type => SCALAR, default => "" },
        },
    );
    my ($data, %common_opts);

    # Create the data manipulation object
    $data = dp::snapmirror::qsm::TargetedData->new(
        path    => $opts{path},
        client  => $opts{client},
        create  => 0,
        save    => 1,
        timeout => $opts{timeout},
        dbfile  => $opts{dbfile},
    );

    # Set common options for dirs, files and streams options
    %common_opts = (
        percent  => 100,
        mode     => 'remove',
        compress => 'any',
        encrypt  => 'any',
        archive  => 'any',
        hidden   => 'any',
        acl      => 'any',
    );

    # Remove all objects in the tree
    $data->modify(
        dirs => {
            %common_opts,
            dirs    => 'any',
            files   => 'any',
            streams => 'any',
        },
        files => {
            %common_opts,
            streams => 'any',
            sparse  => 'any',
        },
        streams => {
            %common_opts,
        },
    );

    # Destroy object
    undef $data;

    if ( ! $opts{contents_only} ) {
        # Remove top level directory
        remove_file(
            file => $opts{path},
            client => $opts{client},
            recursive => 1,
        );
    }

    return 1;
}

#---------------------------------------------------------------------

=head2 ::compare

Compare two directory trees.

The following named arguments are accepted:

=over

=item Arguments

B<path> => I<string>

I<Required.>
Path of primary tree for comparison.

B<cmppath> => I<string>

I<Required.>
Path of secondary tree for comparison.

B<client> => I<hostspec>

I<Optional.>
Client where compare will run. Default is localhost.

B<template> => I<string>

I<Optional.>
Template for comparison which sets the default values of all the following
options. Default is the 'default' template.

B<contents> => C<0|1>

I<Optional.>
If true, compares the contents of the files otherwise
only checks whether files/directories are present.

B<streams> => I<number>

I<Optional.>
If true, compares contents of file and directory streams.

B<attributes> => I<number>

I<Optional.>
If true, compares file and directory attributes including
ACL, compression, encryption, archive and hidden.

B<amctime> => C<0|1>

I<Optional.>
If true, compares file and directory atime, mtime and ctime.

B<dirs> => C<yes|no|any>

I<Optional.>
Compare directories if they contain directories.

B<files> => C<yes|no|any>

I<Optional.>
Compare directories if they contain files.

B<sparse> => C<yes|no|any>

I<Optional.>
Compare files if they are sparse files.

B<streams> => C<yes|no|any>

I<Optional.>
Compare directories/files if they contain streams.

B<compress> => C<yes|no|any>

I<Optional.>
Compare files if they are compressed.

B<encrypt> => C<yes|no|any>

I<Optional.>
Compare files if they are encrypted.

B<archive> => C<yes|no|any>

I<Optional.>
Compare files if they are archived.

B<hidden> => C<yes|no|any>

I<Optional.>
Compare files if they are hidden.

B<acl> => C<yes|no|any>

I<Optional.>
Compare directories/files with ACL.

B<timeout> => I<number>

I<Optional.>
Timeout value for function execution.
Default is 1800.

B<dbfile> => I<string>

I<Optional.>
Unique database file name to allow multiple clients to work in the same
directory.

=item Returns

The number of differences found.

=item Example

=back

     compare(path => '/tmp/data', cmppath => '/t/data');

=for html <hr />

=cut

#---------------------------------------------------------------------
sub compare(%) {
    my (%opts) = validate (
        @_,
        {
            path        => { type => SCALAR },
            cmppath     => { type => SCALAR },
            client      => { type => SCALAR,  default => 'localhost' },
            template    => { type => SCALAR,  default => 'default' },
            contents    => { type => SCALAR, optional => 1 },
            streams_cmp => { type => SCALAR, optional => 1 },
            attributes  => { type => SCALAR, optional => 1 },
            amctime     => { type => SCALAR, optional => 1 },
            dirs        => { type => SCALAR, optional => 1 },
            files       => { type => SCALAR, optional => 1 },
            sparse      => { type => SCALAR, optional => 1 },
            streams     => { type => SCALAR, optional => 1 },
            compress    => { type => SCALAR, optional => 1 },
            encrypt     => { type => SCALAR, optional => 1 },
            archive     => { type => SCALAR, optional => 1 },
            hidden      => { type => SCALAR, optional => 1 },
            acl         => { type => SCALAR, optional => 1 },
            timeout     => { type => SCALAR,  default => 1800 },
            dbfile      => { type => SCALAR, default => ""},
        },
    );
    my ($data, $ret);

    # Set options according to template
    # Options in template are overwritten by explicitly setting the options
    %opts = ( %{$compare_template{$opts{template}}}, %opts, _mode=>'compare' );

    # Create the data manipulation object
    $data = dp::snapmirror::qsm::TargetedData->new(
        path    => $opts{path},
        client  => $opts{client},
        save    => 1,
        create  => 0,
        seed    => _get_seed,
        timeout => $opts{timeout},
        dbfile  => $opts{dbfile},
    );

    # Set all elements to original
    $data->set_original;

    # Compare both trees
    $ret = $data->compare(
        path       => $opts{cmppath},
        contents   => $opts{contents},
        streams    => $opts{streams_cmp},
        attributes => $opts{attributes},
        amctime    => $opts{amctime},
    );

    # Save the current counters
    $counters = $data->counters;

    return $ret;
}

#---------------------------------------------------------------------

=head2 ::get_counters

Get the data counters. The counters are the number of a
specific object type which has been modified or created. The following
is a list of the different counters:

    total   => Total number of objects in the tree
    create  => Create counters
    remove  => Remove counters
    rename  => Rename counters
    trunc   => Truncate counters
    copy    => Copy counters
    modify  => Modify counters
    sattrs  => Attributes set counters
    read    => Read counters

Each of the counters have the following specific object counters:

    dirs    => Number of directories
    files   => Number of files
    streams => Number of streams
    sparse  => Number of sparse files
    links   => Number of links
    slinks  => Number of symbolic links
    acls    => Number of ACL's

=over

=item Returns

A hash reference of all the counters.

=item Example

=back

     $counters_h = get_counters();

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_counters() {
    return $counters;
}

#---------------------------------------------------------------------

=head2 ::get_template

Get a template for either populate, modify, read tree, or compare.

The following named arguments are accepted:

=over

=item Arguments

B<mode> => C<populate|modify|read_tree|compare>

I<Optional.>
Mode of template to get. Defaults to 'populate'.

B<name> => I<string>

I<Optional.>
Name of template to get. Defaults to 'default'.

=item Returns

A hash reference of the template.

=item Example

=back

     $template_h = get_template(mode => 'modify', name => 'default');

=for html <hr />

=cut

#---------------------------------------------------------------------
sub get_template(%) {
    my (%opts) = validate (
        @_,
        {
            mode => { type => SCALAR, default => 'populate',
                     regex => qr/^(populate|modify|read_tree|compare)$/ },
            name => { type => SCALAR, default => 'default' },
        },
    );
    my $template;

    # Select templates for the specified mode
    $template = $opts{mode} eq 'populate' ? \%pop_template :
                $opts{mode} eq 'modify'   ? \%mod_template :
                $opts{mode} eq 'compare'  ? \%compare_template :
                \%read_template;

    if ( ! exists $template->{$opts{name}} ) {
        croak "Template '$opts{name}' does not exist for $opts{mode}";
    }

    return $template->{$opts{name}};
}

#---------------------------------------------------------------------

=head2 ::total_files_pop_opts

Get the populate options to create a tree with the given total number
of files. It calculates the best directory depth to accommodate the total
number of files given. The actual number of files created on the tree
will be about +/-4% of the value given when 'files' is less than 2000
and about +/-1% otherwise.

The following named arguments are accepted:

=over

=item Arguments

B<files> => I<number>

I<Required.>
Total number of files to create in the tree.

B<file_avg> => I<number>

I<Optional.>
Number of files to create per directory. This is just a hint to calculate
the correct values to create the total number of files given, the actual
value for this parameter will be close to this value when the options are
returned. Defaults to 100.

B<dir_avg> => I<number>

I<Optional.>
Number of directories to create per directory. This is just a hint to calculate
the correct values to create the total number of files given, the actual
value for this parameter will be close to this value when the options are
returned. Defaults to 5.

=item Returns

A hash of the options for populate.

=item Example

=back

     %options = total_files_pop_opts(files => 1000);

=for html <hr />

=cut

#---------------------------------------------------------------------
sub total_files_pop_opts(%) {
    my (%opts) = validate (
        @_,
        {
            files    => { type => SCALAR },
            file_avg => { type => SCALAR, default => 100 },
            dir_avg  => { type => SCALAR, default => 5 },
        },
    );
    my %options = %{dclone($pop_template{none})};
    my $F = $opts{files};     # Total number of files
    my $n = $opts{file_avg};  # Number of files per directory
    my $m = $opts{dir_avg};   # Number of directories per directory
    my $M = 0;
    my $d = 0;
    my $fdev = 0;

    # Check for valid parameters
    croak "Parameter file_avg must be greater than or equal to one" if $n < 1;
    croak "Parameter dir_avg must be greater than or equal to zero" if $m < 0;

    if ($F <= $n) {
        # All files will be at the root of the tree
        $n = $F - 1;  # Compensate for the db file
        $m = 0;
    } elsif ($m > 1) {
        # Geometric series solution to calculate the directory depth
        # Round the directory depth to the nearest integer
        $d = (log(($F*($m-1)/$n) + 1) / log($m)) - 0.5;
        $d = $d > 1 ? int($d) : 1;

        # Re-calculate the number of directories per directory (m)
        # Round it to the nearest integer
        $M = $F/$n;  # Total number of directories
        $m = int($M**(1/$d) + 0.5);
        # Re-calculate the number of files per directory (n)
        if ($m > 1) {
            # Re-calculate the total number of directories given
            # the new number of directories per directory
            $M = ($m**($d+1) - 1)/($m - 1);
            # Re-calculate the number of files per directory given
            # the new total number of directories
            $n = $F/$M;
            # Add a little deviation on the number of files per directory
            # if number of files per directory is a fractional number
            $fdev = $n/20 if ($n != int($n));
        } else {
            # One directory per directory, so calculate depth needed
            $m = 1;
            $d = $F/$n - 1;
            $d = 0 if $d < 0;
            $n = $F/2;
        }
    } elsif ($m == 1) {
        # One directory per directory, so calculate depth needed
        $d = $F/$n - 1;
        $d = 0 if $d < 0;
        $d = 1 if $d > 1;
        $n = $F/2;
    } else {
        # All files will be at the root of the tree
        $n = $F - 1;  # Compensate for the db file
        $m = 0;
    }

    # Set the options to return
    $options{template}  = 'none';
    $options{file_avg}  = $n;
    $options{file_dev}  = $fdev;
    $options{dir_avg}   = $m;
    $options{depth_avg} = $d;

    return %options;
}

#---------------------------------------------------------------------
# Internal function
# Check if path supports streams
#---------------------------------------------------------------------
sub _streams_supported($$;) {
    my ($data, $opts) = @_;
    my ($opt_pct, $opt_num, $ver, $rel, $path, $dobj, $err);
    my $streams = 0;

    if ( $data->hosttype =~ /solaris/ && ($opts->{_mode} eq 'read'
      || $opts->{_mode} eq 'populate' || $opts->{_mode} eq 'modify') ) {
        # Check if streams need to be created/modified
        foreach $opt_pct (grep { /stream_pct$/ } keys %$opts) {
            ($opt_num = $opt_pct) =~ s/pct$/num/;
            if ( $opts->{$opt_pct} > 0 && $opts->{$opt_num} > 0 ) {
                # Options are set to create streams
                $streams = 1;
                last;
            }
        }
        if ( $streams ) {
            # Check if streams are supported
            ($ver, $rel) = ((POSIX::uname)[2] =~ /(\d+)\.(\d+)/);
            $ver .= sprintf("%03d", $rel);
            $streams = 0;
            if ($ver >= 5010) {
                # This is Solaris 5.8 or later which supports streams,
                # now check if path actually supports streams
                $path = "$opts->{path}/.test_streams.$$";

                dp::snapmirror::qsm::TargetedData::logVerbose(
                    dp::snapmirror::qsm::TargetedData::VERBOSE_INFO,
                    'Trying to create a stream to check whether path ' .
                    'supports streams'
                );

                # Create the data manipulation object
                $dobj = dp::snapmirror::qsm::TargetedData->new(
                    path    => $path,
                    client  => $opts->{client},
                    timeout => 120,
                    dbfile  => $opts->{dbfile},
                );

                # Create a single file with a stream
                eval {
                    $dobj->create_files(
                        number => 1,
                        dev    => 0,
                        streams => {
                            number => 1,
                            dev    => 0,
                        },
                    );
                };
                $err = $@;

                # Clean up test directory
                eval {
                    dp::snapmirror::qsm::TargetedData::logVerbose(
                        dp::snapmirror::qsm::TargetedData::VERBOSE_INFO,
                        'Remove test stream'
                    );

                    # Destroy the data object
                    $dobj = undef;

                    # Remove test stream
                    remove_file(
                        file   => $path,
                        client => $opts->{client},
                        recursive => 1,
                    );
                };

                # Check for errors
                if ($err =~ /Could not create stream/) {
                    # Streams are not supported on path
                    $streams = 0;
                } elsif ($err) {
                    # Unknown error
                    croak $err;
                } else {
                    # Streams are supported on path
                    $streams = 1;
                }
            }
        }
    }

    return $streams;
}

#---------------------------------------------------------------------
# Internal function
# Ignore unsupported options and log a warning
#---------------------------------------------------------------------
sub _check_options($\%;) {
    my ($data, $opts) = @_;
    my (%vol_info, $vol);
    my $streams = 0;
    my %flags = (
        acls     => 0,
        sparse   => 1,
        encrypt  => 0,
        compress => 0,
    );

    if ( $data->hosttype =~ /windows/ ) {
        # Streams are supported
        $streams = 1;
        # Get volume information
        $vol = $opts->{path};
        $vol = getcwd() if $vol !~ /^.:/;
        $vol =~ s/:.*/:/;
        %vol_info = vol_info(client => $opts->{client}, vol => $vol);
        %flags = %{$vol_info{sysflags}};
    } else {
        # Check if path supports streams
        $streams = _streams_supported($data, $opts);
    }

    if ( $data->hosttype !~ /unix/ ) {
        # Ignore symbolic link option if not unix
        $opts->{slink_pct} = 0;
        dp::snapmirror::qsm::TargetedData::logVerbose(
            dp::snapmirror::qsm::TargetedData::VERBOSE_INFO,
            'Symbolic link option is ignored: Symbolic links ' .
            'are not supported on ' . $data->hosttype
        );
    }

    if ( ! $streams ) {
        # Ignore stream options if OS does not support streams
        ( $opts->{$_} = 0 ) foreach grep { /stream_pct$/ } keys %$opts;
        dp::snapmirror::qsm::TargetedData::logVerbose(
            dp::snapmirror::qsm::TargetedData::VERBOSE_INFO,
            'Stream options are ignored: streams are not supported on path'
        );
    }

    if ( ! $flags{acls} ) {
        # Ignore ACL options, it is only supported on windows
        ( $opts->{$_} = 0 ) foreach grep { /acl_pct$/ } keys %$opts;
        ( $opts->{$_} = 0 ) foreach grep { /^acl_.*_pct$/ } keys %$opts;
        dp::snapmirror::qsm::TargetedData::logVerbose(
            dp::snapmirror::qsm::TargetedData::VERBOSE_INFO,
            'ACL options are ignored: ACLs are not supported on ' .
            ($data->hosttype !~ /windows/ ? $data->hosttype : 'volume')
        );
    }

    if ( ! $flags{sparse} ) {
        $opts->{sparse_pct} = 0;
        dp::snapmirror::qsm::TargetedData::logVerbose(
            dp::snapmirror::qsm::TargetedData::VERBOSE_INFO,
            'Sparse options are ignored: ' .
            'Sparse files are not supported on volume'
        );
    }

    if ( $data->hosttype !~ /windows/ ) {
        # Ignore Windows attribute options, it is only supported on windows
        $opts->{_attr_args} = {};
        ( $opts->{$_} = 0 ) foreach grep { /^sattrs_.*_pct$/ } keys %$opts;
        ( $opts->{$_} = 0 ) foreach grep { /^.*_attr_pct$/ } keys %$opts;
        dp::snapmirror::qsm::TargetedData::logVerbose(
            dp::snapmirror::qsm::TargetedData::VERBOSE_INFO,
            'Windows attribute options are ignored: Windows ' .
            'attributes are not supported on ' . $data->hosttype
        );
    } else {
        if ( ! $flags{encrypt} ) {
            $opts->{encrypt_pct} = 0;
            delete $opts->{sattrs_opts}{encrypt};
            dp::snapmirror::qsm::TargetedData::logVerbose(
                dp::snapmirror::qsm::TargetedData::VERBOSE_INFO,
                'Encrypt options are ignored: ' .
                'Encrypt is not supported on volume'
            );
        }

        if ( ! $flags{compress} ) {
            $opts->{compress_pct} = 0;
            delete $opts->{sattrs_opts}{compress};
            dp::snapmirror::qsm::TargetedData::logVerbose(
                dp::snapmirror::qsm::TargetedData::VERBOSE_INFO,
                'Compress options are ignored: ' .
                'Compress is not supported on volume'
            );
        }

        # Windows attribute options
        $opts->{_attr_args} = {
            archive_num  => $opts->{archive_num},
            archive_pct  => $opts->{archive_pct},
            hidden_num   => $opts->{hidden_num},
            hidden_pct   => $opts->{hidden_pct},
            encrypt_num  => $opts->{encrypt_num},
            encrypt_pct  => $opts->{encrypt_pct},
            compress_num => $opts->{compress_num},
            compress_pct => $opts->{compress_pct},
        };
    }

    return;
}

#---------------------------------------------------------------------
# Common starter options
# Initialize all hashes for the create options for both
# the populate and modify
#---------------------------------------------------------------------
sub _init_create_opts(\%;) {
    my $opts = shift;

    # Name options
    $opts->{_name_args} = {
        name_size     => $opts->{name_avg},
        name_size_dev => $opts->{name_dev},
    };

    # Creating streams option hash
    $opts->{_stream_args} = {
        number => $opts->{stream_avg},
        dev    => $opts->{stream_dev},
        args   => {
            %{$opts->{_name_args}},
            size     => $opts->{stream_size_avg},
            size_dev => $opts->{stream_size_dev},
        }
    };

    # Creating files option hash
    $opts->{_file_args} = {
        number  => $opts->{file_avg},
        dev     => $opts->{file_dev},
        streams => $opts->{_stream_args},
        args    => {
            %{$opts->{_name_args}},
            %{$opts->{_attr_args}},
            %{$opts->{acl_opts}},
            size        => $opts->{file_size_avg},
            size_dev    => $opts->{file_size_dev},
            acl_num     => $opts->{file_acl_num},
            acl_pct     => $opts->{file_acl_pct},
            stream_num  => $opts->{file_stream_num},
            stream_pct  => $opts->{file_stream_pct},
            sparse_num  => $opts->{sparse_num},
            sparse_pct  => $opts->{sparse_pct},
            sparse_opts => $opts->{sparse_opts},
            link_num    => $opts->{link_num},
            link_pct    => $opts->{link_pct},
            slink_num   => $opts->{slink_num},
            slink_pct   => $opts->{slink_pct},
        }
    };

    # Creating directories option hash
    $opts->{_dir_args} = {
        number  => $opts->{dir_avg},
        dev     => $opts->{dir_dev},
        files   => $opts->{_file_args},
        streams => $opts->{_stream_args},
        args    => {
            %{$opts->{_name_args}},
            %{$opts->{_attr_args}},
            %{$opts->{acl_opts}},
            acl_num    => $opts->{dir_acl_num},
            acl_pct    => $opts->{dir_acl_pct},
            stream_num => $opts->{dir_stream_num},
            stream_pct => $opts->{dir_stream_pct},
        }
    };

    if ( $opts->{_mode} eq 'populate' ) {
        # Add the depth options for populate
        $opts->{_dir_args}{depth_avg} = $opts->{depth_avg},
        $opts->{_dir_args}{depth_dev} = $opts->{depth_dev},
    }

    return;
}

#---------------------------------------------------------------------
# Internal function
#---------------------------------------------------------------------
sub _get_seed() {
    my $seed;

    # Get seed to use
    if ( defined($_seed) ) {
        # One from the package already been seeded
        $seed = $_seed + 1;
    } elsif ( defined($main::DATA_SEED) ) {
        # From the environment
        $seed = $main::DATA_SEED;
    } else {
        # Create a new seed
        do {
            $seed = time() ^ ($$ + ($$ << 15));
        } while (!$seed);
    }
    $_seed = $seed;
    return $seed;
}

#---------------------------------------------------------------------
# Internal function
#---------------------------------------------------------------------
sub _get_cont_opts(\%$;) {
    my ($cont_opts, $opts) = @_;

    return (
        %$cont_opts,
        map { $_ => $opts->{$_} } grep { exists $opts->{$_} } @cont_opts,
    );
}

#---------------------------------------------------------------------
# Internal function
#---------------------------------------------------------------------
sub _get_args(\%$$;) {
    my ($opts, $key, $ind_opt) = @_;

    return exists $opts->{$ind_opt}{$args_map{$key}} ?
                  $opts->{$ind_opt}{$args_map{$key}} : $opts->{$key};
}

#---------------------------------------------------------------------
# Internal function
#---------------------------------------------------------------------
sub _create_none(\%;) {
    my $template = shift;
    my $item;
    my @list = qw(dir file depth);

    # Copy the default options to the 'none' template
    $template->{none} = { %{$template->{default}} };

    foreach $item ( keys %{$template->{none}} ) {
        $template->{none}{$item} = 0 if $item =~ /_(num|pct)$/;
    }

    foreach $item ( @list ) {
        if ( exists $template->{none}{"${item}_avg"} ) {
            $template->{none}{"${item}_avg"} = 0;
            $template->{none}{"${item}_dev"} = 0;
        }
    }
}

#---------------------------------------------------------------------
# Common options for the default templates for populate and modify
# Refer to tools/dp/cifs_tools_design.txt for more detailed information
# on the parameters in the templates below.
#---------------------------------------------------------------------
my %_pop_mod_opts = (
    dir_avg         => 3.2,
    dir_dev         => 0.2,
    file_size_avg   => 32768,
    file_size_dev   => 16384,
    name_avg        => 8,
    name_dev        => 3,
    stream_size_avg => 10000,
    stream_size_dev => 5000,

    dir_stream_num  => 99999999999,
    dir_stream_pct  => 3,
    file_stream_num => 99999999999,
    file_stream_pct => 3,

    link_num        => 99999999999,
    link_pct        => 1,
    slink_num       => 99999999999,
    slink_pct       => 1,
    compress_num    => 99999999999,
    compress_pct    => 5,
    encrypt_num     => 99999999999,
    encrypt_pct     => 5,
    archive_num     => 0,
    archive_pct     => 0,
    hidden_num      => 99999999999,
    hidden_pct      => 5,
    sparse_num      => 99999999999,
    sparse_pct      => 1,
    dir_acl_num     => 99999999999,
    dir_acl_pct     => 4,
    file_acl_num    => 99999999999,
    file_acl_pct    => 4,

    acl_opts        => { acl=>1 },
    sparse_opts => {
        ndata      => 3,
        ndata_dev  => 2,
        offset     => 1000000,
        offset_dev => 200000,
        size       => 4096,
        size_dev   => 1024,
    },
);
my %_common_cont_opts = (
    dirs     => 'any',
    files    => 'any',
    sparse   => 'any',
    streams  => 'any',
    compress => 'any',
    encrypt  => 'any',
    archive  => 'any',
    hidden   => 'any',
    acl      => 'any',
);

#---------------------------------------------------------------------
# Compare templates
#---------------------------------------------------------------------
%compare_template = (
    default => {
        %_common_cont_opts,
        contents    => 1,
        streams_cmp => 1,
        attributes  => 1,
        amctime     => 0,
    },
    none => {
        %_common_cont_opts,
        contents    => 0,
        streams_cmp => 0,
        attributes  => 0,
        amctime     => 0,
    },
);

#---------------------------------------------------------------------
# Read templates
#---------------------------------------------------------------------
%read_template = (
    default => {
        %_common_cont_opts,
        file_num      => 99999999999,
        file_pct      => 100,
        sparse_num    => 99999999999,
        sparse_pct    => 100,
        links_num     => 99999999999,
        links_pct     => 100,
        slinks_num    => 99999999999,
        slinks_pct    => 100,
        stream_num    => 99999999999,
        stream_pct    => 100,
        dir_attr_num  => 99999999999,
        dir_attr_pct  => 100,
        file_attr_num => 99999999999,
        file_attr_pct => 100,
        dir_acl_num   => 99999999999,
        dir_acl_pct   => 100,
        file_acl_num  => 99999999999,
        file_acl_pct  => 100,
    },
);
_create_none(%read_template);

#---------------------------------------------------------------------
# Populate templates
#---------------------------------------------------------------------
%pop_template = (
    default => {
        %_pop_mod_opts,
        file_avg   => 40,
        file_dev   => 2,
        stream_avg => 5,
        stream_dev => 0.5,
        depth_avg  => 4.3,
        depth_dev  => 0.2,
    },
);
_create_none(%pop_template);
$pop_template{small} = {
    %{$pop_template{default}},
    depth_avg     => 3,
    file_avg      => 20,
    file_size_avg => 8192,
    file_size_dev => 2048,
};

#---------------------------------------------------------------------
# Modify templates
#---------------------------------------------------------------------
%mod_template = (
    default => {
        %_pop_mod_opts,
        %_common_cont_opts,
        file_avg   => 8,
        file_dev   => 2,
        stream_avg => 0.5,
        stream_dev => 0.5,

        create_on_dir_num   => 99999999999,
        create_on_dir_pct   => 20,
        create_on_file_num  => 99999999999,
        create_on_file_pct  => 20,
        modify_file_num     => 99999999999,
        modify_file_pct     => 20,
        modify_stream_num   => 99999999999,
        modify_stream_pct   => 20,
        modify_sparse_num   => 99999999999,
        modify_sparse_pct   => 20,
        remove_dir_num      => 99999999999,
        remove_dir_pct      => 4,
        remove_file_num     => 99999999999,
        remove_file_pct     => 4,
        remove_stream_num   => 99999999999,
        remove_stream_pct   => 4,
        rename_dir_num      => 99999999999,
        rename_dir_pct      => 5,
        rename_file_num     => 99999999999,
        rename_file_pct     => 5,
        truncate_file_num   => 99999999999,
        truncate_file_pct   => 4,
        truncate_stream_num => 99999999999,
        truncate_stream_pct => 4,
        acl_mod_dir_num     => 99999999999,
        acl_mod_dir_pct     => 5,
        acl_mod_file_num    => 99999999999,
        acl_mod_file_pct    => 5,
        acl_del_dir_num     => 99999999999,
        acl_del_dir_pct     => 5,
        acl_del_file_num    => 99999999999,
        acl_del_file_pct    => 5,
        sattrs_dir_num      => 99999999999,
        sattrs_dir_pct      => 5,
        sattrs_file_num     => 99999999999,
        sattrs_file_pct     => 5,

        sattrs_opts => {
            compress => 1,
            encrypt  => 1,
            hidden   => 1,
        },
    },
);
_create_none(%mod_template);
$mod_template{small} = {
    %{$mod_template{default}},
    file_avg           => 4,
    file_dev           => 1,
    create_on_dir_pct  => 10,
    create_on_file_pct => 10,
    remove_dir_pct     => 15,
    remove_file_pct    => 15,
};

1;

