# $Id: //depot/prod/test/Rsierranevada/lib/Rquotad.pm#1 $
#
## @summary Rquotad - Rquotad package for making Rquotad requests to a filer
## @author	jsh
## @notes	Nothing to it.
## @pod 	here

package Rquotad;

=head1 NAME

Rquotad - Rquotad class for making Rquotad requests to a filer

=head1 SYNOPSIS

use Rquotad;

# Making a request

%{success, blocks_curr, blocks_soft, blocks_hard, blocks_timelimit,
 files_curr, files_soft, files_hard, files_timelimit} =
rquotaQuery($host, $path [,$uid]);

=head1 DESCRIPTION

The B<Rquotad> module provides access to file system quotas. It is 
dependent upon the netapp B<nfs::Xdr> and B<nfs::Rpc> packages for most of its 
functionality. It doesn't provide much, but it provides rquotad access to NATE 
tests.

=head1 STATUS

Works

=head1 AUTHOR

jsh

=head1 COPYRIGHT

Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

use strict;
use warnings;

use Tharn qw(logcomment);
use nfs::Rpc;
use nfs::Xdr;
use Carp qw(croak);
use Params::Validate qw(validate SCALAR OBJECT ARRAYREF HASHREF UNDEF);

use vars qw(@ISA @EXPORT);

require Exporter;
@ISA = ('Exporter');

@EXPORT = qw(
	rquotaQuery
);


BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};


# some constants that define Rquotad
my $RQUOTAPROG = 100011;
my $RQUOTAVERS = 1;
my $RQUOTAPROC_GETQUOTA = 1;
my $RQUOTAPROC_GETACTIVEQUOTA = 2;
my $RQUOTA_OK = 1;
my $RQUOTA_NOQUOTA = 2;
my $RQUOTA_EPERM = 3;
sub getquota_unpack(%);

#------------------------------------------------------------------

=head2 rquotaQuery

%{success, blocks_curr, blocks_soft, blocks_hard, blocks_timelimit,
 files_curr, files_soft, files_hard, files_timelimit} =
rquotaQuery($hostname, $path [,$uid]);

Make a query to $host for path $path. 

Parameters:

=over 4

=item hostname

(Required) : Hostname of filer to query.

=item path

(Required) : Path to query about.

=item uid


(Optional) : UID of the user to make the query as.  Defaults to $<.

=back

Return:

=over 4

Returns a hash with the results.

%{success} declares if the query was a success:

=over 4

1 == OK; 2 == NOQUOTA; 3 == EPERM;

=back

=back

=cut

#------------------------------------------------------------------

sub rquotaQuery($$;$) {
	my $host = shift;
	my $path = shift;
	my $uid;
	if (defined $_[0]) {
		$uid = shift;
	} else {
		$uid = $<;
	}
	
	my $rpc_obj = nfs::Rpc->clnt_create(hostname => $host, 
		program => $RQUOTAPROG, version => $RQUOTAVERS, proto => 'udp');

	my $args = nfs::Xdr->new();
	$args->opaque(encode => $path, variable => (length $path));
	$args->uint(encode => $uid);

	my $out = $rpc_obj->clnt_call(proc => $RQUOTAPROC_GETQUOTA,
		arg => $args->{message}, out => \&getquota_unpack);
	
	if ($out->{success} eq $RQUOTA_OK) {
		logcomment("query success: uid $uid");
	} elsif ($out->{success} eq $RQUOTA_NOQUOTA) {
		logcomment("query failed: no quotas for uid $uid");
	} elsif ($out->{success} eq $RQUOTA_EPERM) {
		logcomment("query failed: permission denied");
	}

	return %{$out};
}

#------------------------------------------------------------------

=head2 getquota_unpack (not exported)

\%{success, blocks_curr, blocks_soft, blocks_hard, blocks_timelimit,
 files_curr, files_soft, files_hard, files_timelimit} =
Rquotad::getquota_unpack(%{xdr});

Unpacks the results of a RQUOTAPROC_GETQUOTA call.  Returns a hash ref to
the results.

Parameters:

=over 4

=item %{xdr}

(Required) : Hash containing xdr object.

=back

Returns:

=over 4

Hash reference to the results.

\%{success, blocks_curr, blocks_soft, blocks_hard, blocks_timelimit,
 files_curr, files_soft, files_hard, files_timelimit};

%->{success} declares if the query was a success:

=over 4

1 == OK; 2 == NOQUOTA; 3 == EPERM;

=back

=back

=cut

#------------------------------------------------------------------

sub getquota_unpack(%) {
	my (%opts) = validate (
		@_,
		{
		    xdr => { type => OBJECT },
		}
	    );

	my ($size, $active, $bhardlimit, $bsoftlimit, $bcurblocks, $fhardlimit,
	    $fsoftlimit, $curfiles, $btimeleft, $ftimeleft) = 
		(0,0,0,0,0,0,0,0,0,0);

	my $xdr = $opts{xdr};
	my $success = $xdr->uint();

	if ($success eq $RQUOTA_OK) {
		$size = $xdr->uint();
		$active = $xdr->uint();
		$bhardlimit = $xdr->uint();
		$bsoftlimit = $xdr->uint();
		$bcurblocks = $xdr->uint();
		$fhardlimit = $xdr->uint();
		$fsoftlimit = $xdr->uint();
		$curfiles = $xdr->uint();
		$btimeleft = $xdr->uint();
		$ftimeleft = $xdr->uint();
	}

	my %return = (
		success => $success,
		blocks_curr => $bcurblocks,
		blocks_soft => $bsoftlimit,
		blocks_hard => $bhardlimit,
		blocks_timelimit => $btimeleft,
		files_curr => $curfiles,
		files_soft => $fsoftlimit,
		files_hard => $fhardlimit,
		files_timelimit => $ftimeleft
		);

	return \%return;
}
