# Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/SAN.pm#3 $

## @summary SAN Routines
## @author archana@netapp.com
## @status public
## @pod here


=head1 NAME

SAN - SAN routines.

=head1 ROUTINES

=cut

package SAN;

use strict;


use Tharn;
use Module::Common qw(validate_mixed);
use Params::Validate v0.70 qw(validate validate_pos SCALAR BOOLEAN HASHREF);
use Admin::License qw(license);
use Utils qw(make_host make_conn);
use Carp qw(croak carp);
use SayWrap qw(sayEOF);
use Volume qw(vol_root vol_exists vol_status);
use mount qw(mount);
use setup qw(is_fcp_setup is_iscsi_setup);
use Utils::Units qw(convert_to_bytes convert_to_units UNIT_MB);


BEGIN
{
    my @igroup_subs = qw(
        igroup_exist
        igroup_create
        igroup_info
        igroup_destroy
    );
    my @iscsi_subs = qw(
        iscsi_show_initiator
        iscsi_nodename
        iscsi_target_info
        get_iscsi_sym_link
    );
    my @host_subs = qw(
        initiator_node_info
    );
    my @lun_subs = qw(
        lun_setup
        lun_exist
        lun_map
        lun_destroy
        lun_status
        lun_disks
        create_lun
    );
    my @fs_subs = qw(
        format_lun
        mkfs_mount_fcp
        mkfs_mount_iscsi
    );
    my @veritas_subs = qw(
        create_veritas_vol
        vxdisk_list
        vx_dmp_status
        vxdiskadd
        vx_get_vol_info
        );

    Module::Common::setup_export_tags(
        igroup   => [ @igroup_subs ],
        veritas  => [ @veritas_subs ],
        lun      => [ @lun_subs ],
        fs       => [ @fs_subs ],
        host     => [ @host_subs ],
        iscsi    => [ @iscsi_subs ],
    );
}

sub iscsi_nodename ( $% ) ;

=head2 lun_setup($filer,$lun_path,$lun_size,$igrp_name,%args)

create a lun via lun setup

=over

=item Arguments

=over

=item filer

(Required) : hostspec of filer

=item lun_path

(Required) : Path of lun

=item lun_size

(Required) : Size of lun

=item igrp_name

(Required) : name of initiator group

=item args

(Optional) : additional arguments

=item lun_os_type

(Optional) : OS Type. Possible values : default/solaris/windows.
Default:default

=item lun_id

(Optional) : ID of lun

=item lun_comment

(Optional) : Lun Comment. Default : No comment

=item igrp_type

(Optional) : Type of Initiator Group. Possible values : FCP/iSCSI.
Default:FCP

=item igrp_os_type

(Optional) : Initiator OS Type. Possible values : default/solaris/windows.
Default:default

=item spc_resv

(Optional) : Space Reserverd : y/n Default: yes

=item port_names

(Optional) : Comma Seperated list of port names

=back


EXAMPLES

    lun_setup($filer, $lun_path, $lun_size, $igrp_name,
            lun_os_type=>"solaris", lun_id=>"2", 
            lun_comment=>"test lun",
            igrp_type="fcp", igrp_os_type=>"solaris",
            spc_resv=>"yes", 
            port_names=>"10:00:00:00:c9:2d:56:5c");

=item Notes

Will create only one lun at a time. Implemented for FCP only

=back

=for html <hr />

=cut

sub lun_setup ( $$$$;% ) {

    my ($filer,$lun_path,$lun_size,$igrp_name,%args) = @_;

    my $h = host($filer);
    my $c = connect("#telnet");

    # check for existance of lun in filer
    if (lun_exist($filer,$lun_path)) {
        logresult('config', "LUN SETUP FAILED. LUN $lun_path already exists.");
    }

    use constant RE_LUN_CREATE => "Do you want to create a LUN\?";
    use constant RE_LUN_OS_TYPE => "OS type of LUN \\(default/solaris/windows\\) \\[default]: ";
    use constant RE_LUN_PATH => "Enter LUN path: ";
    use constant RE_SPC_RESV => "Do you want the LUN to be space reserved\?";
    use constant RE_LUN_SIZE => "Enter LUN size: ";
    use constant RE_LUN_COMMENT => "Enter comment string: ";
    use constant RE_IGRP_NAME => "Name of initiator group";
    use constant RE_USE_EXIST_IGRP => "Do you want to use this existing group\?";
    use constant RE_IGRP_TYPE => "Type of initiator group .* \\(FCP/iSCSI\\)";
    use constant RE_PORT_NAMES => "Enter comma separated portnames: ";
    use constant RE_NODE_NAMES => "Enter comma separated nodenames:";
    use constant RE_IGRP_OS_TYPE => "OS type of initiator group \".*\" \\(default/solaris/windows\\) \\[default]: ";
    use constant RE_LUN_ID => "LUN ID at which initiator group \".*\" sees \".*\"";
    use constant RE_ACCEPT_CONFIG => "Do you want to accept this configuration\?";
    use constant RE_CREATE_NEXT_LUN => "Do you want to create another LUN\?";

    say($c,"lun setup",'-blind');

    # wait for following "lun setup" prompts, and act on them, until the
    # prompt coming back is one of the builting say prompts (i.e. the
    # filer's normal prompt, showing that lun setup is done)

    my $first=1;
    while (1) {
        my $prompt;
        (undef, undef, undef, $prompt) = 
            say($c,"",-prompt,
                [RE_LUN_CREATE, RE_LUN_OS_TYPE, RE_LUN_PATH, RE_SPC_RESV, 
                RE_LUN_SIZE, RE_LUN_COMMENT, RE_IGRP_NAME, RE_USE_EXIST_IGRP,
                RE_IGRP_TYPE, RE_PORT_NAMES, RE_NODE_NAMES, RE_IGRP_OS_TYPE, 
                RE_LUN_ID, RE_ACCEPT_CONFIG, RE_CREATE_NEXT_LUN]);

        if ($prompt =~ RE_LUN_CREATE) {
            say($c,"\r",'-blind');
        } elsif ($prompt =~ RE_LUN_OS_TYPE) {
            if ($args{lun_os_type}) {
                say($c,$args{lun_os_type},'-blind');
            } else {
                say($c,"\r",'-nocr','-blind');
            }
        } elsif ($prompt =~ RE_LUN_PATH) {
            say($c,$lun_path,'-blind');
        } elsif ($prompt =~ RE_SPC_RESV) {
            if ($args{spc_resv}) {
                say($c,$args{spc_resv},'-blind');
            } else {
                say($c,"\r",'-nocr','-blind');
            }
        } elsif ($prompt =~ RE_LUN_SIZE) {
            say($c,$lun_size,'-blind');
        } elsif ($prompt =~ RE_LUN_COMMENT) {
            if ($args{lun_comment}) {
                say($c,$args{lun_comment},'-blind');
            } else {
                say($c,"\r",'-nocr','-blind');
            }
        } elsif ($prompt =~ RE_NODE_NAMES) {
            tstlog('-script', "SORRY. The iSCSI part is not yet implemented in this script");
        } elsif ($prompt =~ RE_PORT_NAMES) {
            if (!$args{port_names}) {
                tstlog('-script', "Insufficient Arguments. Specify port names");
            } else {
                if ($first) {
                    say($c,$args{port_names},'-blind');
                    $first--;
                } else {
                    say($c,"\r",'-blind');
                    $first++;
                }
            }
        } elsif ($prompt =~ RE_IGRP_NAME) {
            say($c,$igrp_name,'-blind');
        } elsif ($prompt =~ RE_USE_EXIST_IGRP) {
            say($c,"\r",'-blind');
        } elsif ($prompt =~ RE_IGRP_TYPE) {
            if ($args{igrp_type}) {
                say($c,$args{igrp_type},'-blind');
            } else {
                say($c,"\r",'-nocr','-blind');
            }
        } elsif ($prompt =~ RE_IGRP_OS_TYPE) {
            if ($args{igrp_os_type}) {
                say($c,$args{igrp_os_type},'-blind');
            } else {
                say($c,"\r",'-nocr','-blind');
            }
        } elsif ($prompt =~ RE_LUN_ID) {
           if ($args{lun_id}) {
               say($c,$args{lun_id},'-blind');
           } else {
               say($c,"\r",'-nocr','-blind');
           }
       } elsif ($prompt =~ RE_ACCEPT_CONFIG) {
           say($c,"\r",'-blind');
       } elsif ($prompt =~ RE_CREATE_NEXT_LUN) {
           say($c,"\r",'-blind');
        } else {
            # builtin prompt, i.e. filer's normal prompt
            last;
        }
    }

    if (lun_exist($filer,$lun_path)) {
        logcomment("LUN SETUP of $lun_path SUCCEEDED");
    } else {
        logresult('fatal', 'LUN SETUP FAILED');
    }
}



=head2 lun_exist($filer,$lun_path)

Checks if specified lun is present on filer.

=over

=item Returns

1 (if lun exists)

0 (if absent)

=back

=for html <hr />

=cut

sub lun_exist($$) {

    my ($filer, $lun_path) = validate_pos(
                    @_,
                    { type=>SCALAR|HASHREF, optional=>0 }, 
                    { type=>SCALAR, optional=>0, 
                         regex=>qr/^\/vol\/\w+\/\w+$/ },
                    );

    my $conn = make_conn($filer);

    my $out = sayEOF($conn, "lun show $lun_path");
    return(($out =~ /lun show: No such LUN exists/) ? 0 : 1);

}

=head2 lun_map($filer,$lun_path,$igroup,%args)

Maps a lun to igroup.

=over

=item Arguments

=over

=item filer

(Required) : Filer where lun lives

=item lun_path

(Required) : path of the lun. Eg: /vol/vol0/lun0

=item igroup

(Required) : name of igroup to map to

=item lun_id

(Optional) : ID to map to. If not specified then OnTap will decide the lun id.

=back

=item Notes

Ignores if lun was already mapped to specified igroup.

Does not handle if we run out of id's.

Assumes lun and igroup already exist on filer.

=item Returns

On success - returns 1

On failure - dies

=back

=for html <hr />

=cut

sub lun_map( $$$% ) {

    my ($filer, $lun_path, $igroup, %args) = validate_mixed(
                    @_,
                    [ { type=>SCALAR|HASHREF, optional=>0 }, 
                      { type=>SCALAR, optional=>0, 
                         regex=>qr/^\/vol\/\w+\/\w+$/ },
                      { type=>SCALAR, optional=>0 }, 
                    ],
                    {
                     lun_id => { type=>SCALAR, optional=>1,
                                regex=>qr/^\d+$/ },
                    } );

    my $conn = make_conn($filer);

    my $cmd = "lun map $lun_path $igroup";
    $cmd .= " $args{lun_id}" if (defined $args{lun_id});

    my $expected = ($args{lun_id} ? qr/(already mapped to this group|auto-assigned $igroup=($args{lun_id}))/
                                  : qr/(already mapped to this group|auto-assigned $igroup=(\d+))/);

    my $out = sayEOF($conn, $cmd);
    croak("Unexpected output : $out\n") if ($out !~ $expected);

    return(1);
    
}

=head2 lun_destroy($filer,$lun_path,%opts)

destroys a lun on the filer

=over

=item Arguments

=over

=item filer

(Required) : filer where lun exists

=item lun_path

(Required): path to lun which is to be destroyed

=item force

(Optional): This will issue the lun destroy command with the '-f' option.
If set will delete a lun even if it is online and / or mapped.

=item verify

(Optional): Will verify if the lun was *really* destroyed.

=item die_non_existing

(Optional): If set the routine will die if the lun does not exist.
If not set will ignore absence of the lun to be destroyed and will
return 1.

=back

=item Returns

1 on success

dies on failure

=back

=for html <hr />

=cut

sub lun_destroy ( $$% ) {

    my ($filer, $lun_path, %args) = validate_mixed(
                    @_,
                    [ { type=>SCALAR|HASHREF, optional=>0 }, 
                      { type=>SCALAR, optional=>0, 
                         regex=>qr/^\/vol\/\w+\/\w+$/ },
                    ],
                    {
                     force => { type=>SCALAR, optional=>1,
                                regex=>qr/^0|1$/, default=>1 },
                     verify => { type=>SCALAR, optional=>1,
                                regex=>qr/^0|1$/, default=>1 },
                     die_non_existing => { type=>SCALAR, optional=>1,
                                default=>0 },
                    }
                    );

    my $conn = make_conn($filer);

    my $cmd = "lun destroy";
    $cmd .= " -f" if $args{force};
    $cmd .= " $lun_path";
    my $out = sayEOF($conn, $cmd);

    if ($out =~ /No such LUN exists/i) {
        if ($args{die_non_existing}) {
            croak("$out\n");
        } else {
            return(1);
        } 
    } 

    # no output to verify ; die if we get something
    croak("Unexpected output : $out\n") if ($out);
    # if lun is online then this wont work. use -f (force) option
    
    # verify if lun got deleted, if asked to
    croak("Lun $lun_path did not get destroyed\n") 
        if ($args{verify} && lun_exist($filer, $lun_path));

    return (1);

}


=head2 lun_disks($filer,$client,$lun_path,%opts)

Gets the disks associated with the lun.
Currently works for solaris fcp lun only.

=over

=item Arguments

=over

=item filer

(Required) : Hostspec of filer

=item client

(Required) : Hostspec of client (solaris only)

=item lun_path

(Required) : Path to lun

=item sanlun_path

(Optional) : Location of sanlun binary. Default: /usr/sbin

=back

=item Returns

list of disks associated with the lun

=item Notes

Currently works for solaris FCP luns only.

Assumes lun can be seen on the solairs client via 'sanlun'.

=back

=for html <hr />

=cut

sub lun_disks( $$$% ) {

    my ($filer, $client, $lun_path, %args) = validate_mixed(
            @_,
            [ { type=>SCALAR|HASHREF, optional=>0 }, 
              { type=>SCALAR|HASHREF, optional=>0 },
              { type=>SCALAR, optional=>0, regex=>qr/^\/vol\/\w+\/\w+$/ },
            ],
            {
             sanlun_path => { type=>SCALAR, optional=>1, default=>'/usr/sbin', 
                            regex=>qr/^(\/\w+)+$/ },
            } );

    my $sudo = "/usr/local/bin/sudo";

    my $fh = make_host($filer);
    my $filer_id = $fh->{id};
    my $ch = host($client);
    my $client_id = $ch->{id};
    my $cc = make_conn($client);

    # do multiple times since sanlun fails on first attempt and then passes on linux iscsi
    sayEOF($cc, "$sudo $args{sanlun_path}/sanlun lun show $filer_id:$lun_path");
    sleep(5);
    sayEOF($cc, "$sudo $args{sanlun_path}/sanlun lun show $filer_id:$lun_path");
    sleep(5);
    my $out = sayEOF($cc, "$sudo $args{sanlun_path}/sanlun lun show $filer_id:$lun_path");
    croak("Host attach kit 'sanlun' missing on $client_id: $out\n") 
        if ($out =~ /command not found/i);

    my @lun_dev;
    while ($out =~ /$filer_id:\s+$lun_path\s+(\/dev\/\w+)\s+.*GOOD/gm) {
        push(@lun_dev,$1) if ($1);
    }

    return(@lun_dev);
}



=head2 lun_status($filer,$client,$adapter,$lun_path,%args)

Get status of lun on client adapter.

=over

=item Returns

(%status)

=back

=for html <hr />

=cut

sub lun_status( $$$$% ) {

    my ($filer, $client, $adapter, $lun_path, %args) = validate_mixed(
            @_,
            [ { type=>SCALAR|HASHREF, optional=>0 }, 
              { type=>SCALAR|HASHREF, optional=>0 },
              { type=>SCALAR, optional=>0 },
              { type=>SCALAR, optional=>0, regex=>qr/^\/vol\/\w+\/\w+$/ },
            ],
            {
             sanlun_path => { type=>SCALAR, optional=>1, default=>'/usr/sbin', 
                            regex=>qr/^(\/\w+)+$/ },
            } );

    my $sudo ="/usr/local/bin/sudo";

    my $fh = make_host($filer);
    my $filer_id = $fh->id();
    my $cc = make_conn($client);

    # do twice since sanlun fails on first attempt and then passes on linux iscsi
    sayEOF($cc, "$sudo $args{sanlun_path}/sanlun lun show $filer_id:$lun_path");
    sleep(5);
    sayEOF($cc, "$sudo $args{sanlun_path}/sanlun lun show $filer_id:$lun_path");
    sleep(5);
    my $out = sayEOF($cc, "$sudo $args{sanlun_path}/sanlun lun show $filer_id:$lun_path");

    my %status;
    ($status{dev_file},$status{size},$status{state}) = 
        ($out =~ /\s+$filer_id:\s+$lun_path\s+(\S+)\s+$adapter\s+\S+\s+\((\d+)\)\s+(\S+)/);
    return(%status);
}


#sub mkfs_hp()
#{
#"sudo newfs -F hfs -o largefiles /dev/vg01/rtada-u23-lun0 ";
#"mkdir ?p /t/tada-u23/lunHP";
#"sudo mount -F hfs /dev/vg01/tada-u23-lun0 /t/tada-u23/lunHP/";
#mount();
#}

=head2 mkfs_mount_fcp($filer,$client,$disk,%opts)

Make a file system and mount if specified
Works for fcp disks only and on solaris clients.

=over

=item Arguments

=over

=item filer

(Required) : hostspec of filer

=item client

(Required) : client hostspec. Currently works from solaris only.

=item disk

(Required) : disk to be made a file system and mounted

=item mount

(Optional) : flag denoting if disk is to be mounted. Default: 1 (yes, mount)

=item fstype

(Optional) : type of file system to create. Default : 'ufs'

=item timeout

(Optional) : time(sec) required to making filesystem. Default: 200

=back

=item Notes

TBD write the vxfs part also

TBD try to merge this routine and mkfs_mount_iscsi

=back

=for html <hr />

=cut

sub mkfs_mount_fcp( $$$% ) {

    my ($filer, $client, $disk, %args) = validate_mixed(
            @_,
            [ { type=>SCALAR|HASHREF, optional=>0 }, 
              { type=>SCALAR|HASHREF, optional=>0 },
              { type=>SCALAR, optional=>0 },
              #{ type=>SCALAR, optional=>0, regex=>qr/^\/vol\/\w+\/\w+$/ },
            ],
            {
             mount => { type=>SCALAR, optional=>1, default=>1, 
                            regex=>qr/^0|1$/ },
             fstype => { type=>SCALAR, optional=>1, default=>'ufs', 
                            regex=>qr/^\w+$/ },
             timeout => { type=>SCALAR, optional=>1, default=>400, 
                            regex=>qr/^\d+$/ },
            } );

    my $h = make_host($filer);
    my $ch = make_host($client);
    my $cc = make_conn($client,"#rlogin");

    my $sudo = '/usr/local/bin/sudo';

    $disk =~ s/rdsk/dsk/g; # need a block device
    $disk =~ s/s2/s6/g;    # use slice 6

    if ($args{fstype} eq 'ufs') {
        say($cc,"$sudo newfs $disk",
            '-prompt', "newfs: construct a new file system");
        my (undef, undef, $out) = say($cc, "y", '-timeout', $args{timeout});
        croak("$out\n") if ($out =~ /ERROR/i);
    } elsif ($args{fstype} eq 'vxfs') {
        croak("$args{fstype} Not yet supported\n");
    } else {
        croak("Unknown file system type $args{fstype}\n");
    }

    if ($args{mount}) {
        my $mount_point = mount($h->id(), $disk, type=>"device", client=>$ch->id())
            or croak("Could not mount $disk on $client: $@\n");
        return($mount_point);
    } else {
        print "Skipping mount";
        return(1);
    }
}

=head2 mkfs_mount_iscsi($filer,$client,$disk,%opts)

Make a file system and mount if specified.
Works for iscsi disks only and on linux clients.

=over

=item Arguments

=over

=item filer

(Required) : hostspec of filer

=item client

(Required) : client hostspec. Currently works from linux only.

=item disk

(Required) : disk to be made a file system and mounted

=item mount

(Optional) : flag denoting if disk is to be mounted. Default: 1 (yes, mount)

=item fstype

(Optional) : filesystem type ext3, ext2 etc. Default: ext3

=item timeout

(Optional) : time(sec) required to making filesystem. Default: 400

=back

=item Notes

TBD try to merge this routine and mkfs_mount_fcp

=back

=for html <hr />

=cut

sub mkfs_mount_iscsi ( $$$% ) {

    my ($filer, $client, $disk, %args) = validate_mixed(
            @_,
            [ { type=>SCALAR|HASHREF, optional=>0 }, 
              { type=>SCALAR|HASHREF, optional=>0 },
              { type=>SCALAR, optional=>0, regex=>qr/^\/dev\/sd\w$|^\/dev\/iscsi\/bus\d+\/target\d+\/lun\d+\/disk$/ },
            ],
            {
             mount => { type=>SCALAR, optional=>1, default=>1, 
                            regex=>qr/^0|1$/ },
             fstype => { type=>SCALAR, optional=>1, default=>"ext3" },
             timeout => { type=>SCALAR, optional=>1, default=>400, 
                            regex=>qr/^\d+$/ },
            } );

    my $h = make_host($filer);
    my $filer_id = $h->{id};
    my $fc = make_conn($filer);
    my $ch = host($client);
    my $client_id = $ch->{id};
    my $cc = make_conn($client,"#rlogin");

    my $sudo = '/usr/local/bin/sudo';

    # Make the file system
    say($cc, "$sudo /sbin/mkfs -V -t $args{fstype} $disk", '-prompt', "Proceed anyway\?");
    my (undef, undef, $out) = say($cc, "y", '-timeout', $args{timeout});
    croak("$out\n") if ($out =~ /ERROR/i);

    # Mount the filesystem
    if ($args{mount}) {
        my $mount_point = mount($filer_id, $disk, type=>'device', client=>$client_id)
            or croak("Could not mount $disk on $client_id: $@");
        return($mount_point);
    } else {
        print "Skipping mount";
        return(1);
    }

}


=head2 get_iscsi_sym_link($filer,$client,$lun_path)

Get iSCSI symbolic link for a lun.

=over

=item Arguments

=over

=item filer

(Required) : filer who is hosting the lun.

=item client

(Required) : iSCSI client

=item lun_path

(Required) : lun path

=back

=item Returns

on success returns the symbolic link to the lun path

on failure dies with error in $@

=back

=for html <hr />

=cut

sub get_iscsi_sym_link ( $$$ ) {

    my ($filer, $client, $lun_path) = validate_pos(
            @_,
            { type=>SCALAR|HASHREF, optional=>0 }, 
            { type=>SCALAR|HASHREF, optional=>0 },
            { type=>SCALAR, optional=>0, regex=>qr/^\/vol\/\w+\/\w+$/ },
            );

	# we should not need the /dev/iscsi path, good thing to.  It does not exist on newer versions
	# of the iscsi driver and never worked on 64 bit systems.
   my $cc = make_conn($client);
   my $out = sayEOF($cc,"/usr/local/bin/sudo /usr/sbin/sanlun lun show");
   my @lines = split(/\n/,$out);
	foreach my $line (@lines) {
		if ($line =~ /rtpqa-fas980-5:\s+$lun_path\s+(\S+)\s+/) {
			return ($1);
		}
	}
   croak("Failed to find devicename for lun.  $client $lun_path\n")
}


=head2 format_lun($lun,$client)

Formats a lun on solaris

Wrapper for 'format' command on solaris.
Will format a lun to a fixed type of partition table.


=over

=item Arguments

=over

=item lun

(Required) : Lun to be formatted

=item client

(Required) : hostspec of solaris client

=back


=back

=for html <hr />

=cut

sub format_lun($$) {

    my ($lun, $client) = @_;
    my ($raw_dsk,$format_dsk) = ($lun =~ m#(/dev/rdsk/(c\d+t\d+d\d+)s\d+)#);

    my $sudo = '/usr/local/bin/sudo';

    # First format my disks.
    # This is an anonymous array with first element as the one of 
    # the format command and the second as the prompt it would wait for.
    my $c = connect("$client#rlogin");
    foreach my $aref( 
        ["$sudo format -d $format_dsk", "format>"],
        ["partition", "partition>"],
        ["modify", "Choose base \\(enter number\\) \\[0\\]\\?"],
        ["1",   "table based on above table\\[yes\\]\\?"],
        ["yes", "Free Hog partition\\[6\\]\\?"],
        ["6",   "Enter size of partition '0'"],
        ["1c",  "Enter size of partition '1'"],
        [" ",   "Enter size of partition '3'"],
        [" ",   "Enter size of partition '4'"],
        [" ",   "Enter size of partition '5'"],
        [" ",   "Enter size of partition '7'"],
        [" ",   "Okay to make this the current partition table\\[yes\\]\\?"],
        ["yes", "Enter table name \\(remember quotes\\)"],
        ["\"CAT\"", "Ready to label disk, continue\\?"],
        ["yes", "partition>"],
        ["quit", "format>"],
        ["quit"," "],
    ) {
        my ($cmd,$waitprompt) = @$aref;
        my (undef,undef,undef,$prompt) = say ($c,"$cmd",
                                '-prompt',"$waitprompt",
                                '-timeout',"300");
    }
    # Let us send an extra carriage return so that things 
    # are said on the prompt later
    say($c, "");
}


=head2 create_lun($filer,$client,%opts)

Creates a lun for solaris / linux on filer and will map it to 
solaris fcp or linux iSCSI client respectively.


=over

=item Arguments

=over

=item filer

(Required) : hostspec of the filer

=item client

(Required): hostspec of the solaris client

=item filer_vol

(Optional) : volume on filer where lun is to be created

=item igroup

(Optional) : name of igroup to map lun to.  Will create the igroup if it does
not exist.

=item lun_name

(Optional) : name of lun to create. Default: 'cat_lun'

=item lun_size

(Optional) : size of lun to create. Default: 4194304 (4m)

=item lun_path

(Optional) : path of lun (including lun name) this will overide lun_name

=item noreserve

(Optional): If defined then lun will be created with no reservations

=item format

(Optional) : Flag to specify if lun is to be formatted or not.
Possible values : 0 / 1.  Default: 1

=item vg_name

(Optional) : Volume group name. Used for lun setup on HP-UX hosts.
Default: 'vg_ntap'

=item lv_name

(Optional) : Logical volume name. Used for lun setup on HP-UX hosts.
Default: 'lvol'

=item sanlun_path

(Optional) : location of binary 'sanlun'.  Default: /usr/sbin

=item devfsadm_path

(Optional) : location of binary 'devfsadm'. Default: /usr/sbin

=back


=item Notes

Works for Solaris FCP, Hpux FCP and Linux iSCSI clients only


=item Returns

On success - a list of lun devices attached to the created lun

On failure - croaks with error in $@

=back

=for html <hr />

=cut

sub create_lun ( $$;% ) {

    my ($filer, $client, %args) = validate_mixed(
            @_,
            [ { type=>SCALAR|HASHREF }, 
              { type=>SCALAR|HASHREF }
            ],
            {
             protocol => { type=>SCALAR, optional=>1, 
                            regex=>qr/^(iscsi|fcp)$/, default=>'fcp' },
             filer_vol => { type=>BOOLEAN, optional=>1,  default=>undef, 
                            regex=>qr/^\w+$/ },
             lun_name => { type=>SCALAR, optional=>1, default=>'cat_lun',
                            regex=>qr/^\w+$/ },
             lun_path => { type=>BOOLEAN, optional=>1, default=>undef,
                            regex=>qr/^\/vol\/\w+\/\w+$/ },
             lun_size => { type=>SCALAR, optional=>1, 
                            regex=>qr/^\d+(k|m|g|t)$/i, 
                            default=>4194304 }, # 4m
             igroup => { type=>SCALAR, optional=>1 },
             lun_id => { type=>SCALAR, optional=>1, regex=>qr/^\d+$/ },
             purge_existing => { type=>SCALAR, optional=>1, default=>0, 
                            regex=>qr/^0|1$/ },
             noreserve => { type=>SCALAR, optional=>1, default=>'' },
             ostype => { type=>SCALAR, optional=>1, 
                        regex=>qr/^(image|solaris|windows|hpux|aix|linux)$/ },
             format => { type=>SCALAR, optional=>1, regex=>qr/^(0|1)$/, 
                            default=>1 },
             vg_name => { type=>SCALAR, optional=>1, regex=>qr/^\w+$/, 
                            default=>'vg_ntap' },
             lv_name => { type=>SCALAR, optional=>1, regex=>qr/^\w+$/,
                            default=>'lvol' },
             sanlun_path => { type=>SCALAR, optional=>1, 
                            default=>'/usr/sbin', regex=>qr/^(\/\w+\/?)+$/ },
             devfsadm_path => { type=>SCALAR, optional=>1, 
                            default=>'/usr/sbin', regex=>qr/^(\/\w+\/?)+$/ },
            } );

    eval {

    my $ch = make_host($client);
    my $client_hosttype = $ch->{hosttype};
    my $client_id = $ch->{id};
    my $cc = make_conn($client, "#rlogin");

    my $sudo = '/usr/local/bin/sudo';
    my $path;
    if ($client_hosttype =~ /^unix-(hpux|solaris)/) {
        $path = '/usr/bin';
        croak("This routine does not support iscsi on solaris / hp-ux")
            if ($args{protocol} eq 'iscsi');
    } elsif ($client_hosttype =~ /^unix-linux/) {
        $path = '/bin';
        croak("This routine does not support fcp on linux")
            if ($args{protocol} eq 'fcp');
    } else {
        croak("$client_hosttype client $client_id not supported\n");
    }


    my $h = make_host($filer);
    my $filer_id = $h->{id};
    my $fc = make_conn($filer);

    # die if iSCSI/FCP is not setup on filer
    if ($args{protocol} eq 'fcp') {
        croak("FCP is not setup on filer")
            if (!is_fcp_setup($fc));
    } else {
        croak("iSCSI is not setup on filer")
            if (!is_iscsi_setup($fc));
    }


    # Create the lun_path as /vol/$filer_vol/$lun_name if not provided
    # If $filer_vol not defined use root volume
    # If defined then check if it exists and is online
    if (!defined($args{lun_path})) {
        my $filer_vol;
        if (!defined($args{filer_vol})) {
            $filer_vol = vol_root($fc);
        } else {
            if (vol_exists($fc, $args{filer_vol})) {
                my $obj = vol_status($fc, $args{filer_vol})
                                or croak($@);
                my $state = $obj->state();
                if (defined($state) && $state eq 'online') {
                    $filer_vol = $args{filer_vol};
                } else {
                    croak("$args{filer_vol} is not online\n");
                }
            } else {
                croak("$args{filer_vol} does not exist\n");
            }
        }
        $args{lun_path} = "/vol/$args{filer_vol}/$args{lun_name}";
    }



    # If igroup is defined use it, else create one of same name as client.
    # If the user defined igroup does not exist then create it.
    $args{igroup} = $client_id if (!$args{igroup});
    my @opt_args = (protocol=>$args{protocol}, verify=>1, purge_existing=>0,
                sanlun_path=>$args{sanlun_path});
    push(@opt_args,ostype=>$args{ostype}) if ($args{ostype});
    igroup_create($filer,$client,$args{igroup},
                    @opt_args);

    # Verify if $igroup really belongs to $client
    # If so how many adapters its connected to 
    # so that the lun must have that many or more entries in 
    # sanlun lun show
    my (undef, undef, @nodes) = igroup_info($filer,$args{igroup});
    my (%initiator) = initiator_node_info($client, protocol=>$args{protocol}, 
                                    sanlun_path=>$args{sanlun_path});

    my @client_adapter;
    foreach my $nodename (@nodes) {
        $nodename =~ s/://g if ($args{protocol} eq 'fcp'); # remove the : for FCP nodes
        while (my ($adapter, $node) = each (%initiator)) {
            if ($nodename eq $node) {
                logcomment("Igroup:$args{igroup} is connected to Host $client_id via $adapter Node $node");
                push(@client_adapter,$adapter);
            }
        }
    }
    croak("Host:$client_id is not associated with FILER:$filer_id IGROUP:$args{igroup}")
        if (!@client_adapter);


    my $out;
    if (lun_exist($filer, $args{lun_path})) {
        if ($args{purge_existing}) {
            lun_destroy($filer, $args{lun_path}, force=>1, verify=>1);
        } else {
            logcomment("Using existing lun $args{lun_path}");
        }
    } else {
        # Create lun of size $lun_size
        my $cmd = "lun create -s $args{lun_size}";
        $cmd .= " -t $args{ostype}" if ($args{ostype});
        $cmd .= " -o $args{noreserve}" if ($args{noreserve});
        $cmd .= " $args{lun_path}";
        $out = sayEOF($fc, $cmd);
        #croak("Unable to create LUN $args{lun_path} of size $args{lun_size} on $filer_id : $out")
        #    if ($out !~ /created a LUN of size/i);
        # output of lun create is not consistent so verify if lun exists and fail if required
        croak("Unable to create LUN $args{lun_path} of size $args{lun_size} on $filer_id : $out")
            if (!lun_exist($filer, $args{lun_path}));
        logcomment("Created lun $args{lun_path}") 
    }
   

    # Map to the igroup that we just created with
    # edit file /etc/sd.conf on CLIENT so that lun can be recognized
    # we dont care about the lun id... let ontap autoassign it
    # will we reach max # of lun id on the igroup ??? shld we check for it
    # what are the other failures ????
    # if the lun exisited and was already mapped to this igroup this will be ok
    # it was mapped to another igroup shld be fine too since we can map
    # a lun to multiple igroups
    #@opt_args = (lun_id=>$args{lun_id}) if $args{lun_id};
    if ($args{lun_id}) {
        lun_map($filer, $args{lun_path}, $args{igroup}, lun_id=>$args{lun_id});
    } else { 
        lun_map($filer, $args{lun_path}, $args{igroup});
    }
    logcomment("Mapped lun $args{lun_path} to $args{igroup}");


    # Check if CLIENT can see the newly created lun
    if ($args{protocol} eq 'fcp') {
        if ($client_hosttype =~ /^unix-solaris/) {
            # XXX - do we need to run devfsadm - sanlun should suffice ???
            # First Run devfsadm so that solaris will detect the newly created lun
            # XXX - make sure lputil is ok TBD: make entry for lun in /etc/sd.conf
            (undef, undef, $out) = say($cc, "$sudo $args{devfsadm_path}/devfsadm", '-timeout', 600);
            #croak("Incorrect FCP client $client_id: $out") # will we hit this ever ?
            #    if ($out =~ /command not found/);

        } elsif ($client_hosttype =~ /^unix-hpux/) {

            # discover the lun's
            (undef, undef, $out) = say($cc, "$sudo ioscan -fn -C disk && echo '_ioscan_ok_'",
                                    '-timeout', 2000);
            croak("Failed to scan for IO devices: $out\n")
                if ($out !~ '_ioscan_ok_');
            
            # create the device files
            (undef, undef, $out) = say($cc, "$sudo insf -C disk && echo '_insf_ok_'",
                                    '-timeout', 1000);
            croak("Failed to create the device files : $out\n")
                if ($out !~ '_insf_ok_');

            # list lun -> device file mapping
            #(undef, undef, $out) = say($cc, "$sudo ioscan -funC disk && echo '_ioscan_ok_'",
            #                        '-timeout', 800);
            #croak("Failed to scan for IO devices: $out\n")
            #    if ($out !~ '_ioscan_ok_');

            # obtain the disks for the lun
            my @lun_dev = lun_disks($filer, $cc, $args{lun_path}, 
                                sanlun_path=>'/opt/NetApp/santools/bin/sanlun');
            my $device = $lun_dev[0]; # pick first disk
            shift(@lun_dev); # remove the first element

            # create the vol grp
            create_volgrp($cc, $args{vg_name}, $device, more_devices=>@lun_dev);

            # Init LUN for use with HP LVM
            logcomment("Initialize LUN for use with HP LVM");
            (undef, undef, $out) = say($cc, 
                "$sudo pvcreate /dev/rdsk/$device && echo '_pvcreate_ok_'");
            croak("Failed to initialize '/dev/rdsk/$device' for use with HP LVM : $out\n")
                if ($out !~ '_pvcreate_ok_');

            # Fully populate the alternate paths (might take a long time)
            logcomment("Fully populate the alternate paths");
            (undef, undef, $out) = say($cc, 
                "$sudo /opt/NetApp/santools/bin/ntap_config_path $args{vg_name} && echo '_ntap_config_ok_'",
                                '-timeout', 1000);
            croak("Failed to fully populate the alternate lun paths. $out\n")
                if ($out !~ '_ntap_config_ok_');

            # create a logical volume
            my $lun_size_mb = convert_to_units(convert_to_bytes($args{lun_size}), 
                                            units=>UNIT_MB); # convert lun size to MB
            (undef, undef, $out) = say($cc, 
                "$sudo lvcreate -L $lun_size_mb -n $args{lv_name} $args{vg_name} && echo '_lvcreate_ok_'");
            croak("Failed to create the logical volume. $out\n")
                if ($out !~ '_lvcreate_ok_');

            (undef, undef, $out) = say($cc, 
                "$sudo lvdisplay -v /dev/$args{vg_name}/$args{lv_name} && echo '_lvdisplay_ok_'");
            croak("Failed to get execute lvdisplay. $out\n")
                if ($out !~ '_lvdisplay_ok_');

            # make a hfs file system and mount
            #mkfs_hp();

        } else {
            croak("create_lun: Should not have reached here\n");
        }
    } else {
        # iSCSI part
        (undef, undef, $out) = say($cc, "$sudo /etc/init.d/iscsi restart && echo '_iscsi_restart_ok_'",
                                '-timeout', 200);
        croak("Failed to restart iSCSI : $out\n") if ($out !~ '_iscsi_restart_ok_');
        croak("Incorrect iSCSI client $client_id: $out") # will we hit this ever ?
            if ($out =~ /command not found/i);
        logcomment("iscsi restarted");
    }

    sleep(5);
    # do twice since sanlun fails on first attempt and then passes on linux iscsi
    say($cc, "$sudo $args{sanlun_path}/sanlun lun show $filer_id:$args{lun_path}", 
                '-timeout', 600);
    sleep(5);
    say($cc, "$sudo $args{sanlun_path}/sanlun lun show $filer_id:$args{lun_path}", 
                '-timeout', 600);
    sleep(5);
    (undef, undef, $out) = say($cc, 
                "$sudo $args{sanlun_path}/sanlun lun show $filer_id:$args{lun_path}", 
                '-timeout', 600);
    croak("CONFIG: FCP software missing $client: $out")  # will we hit this ever ?
            if ($out =~ /command not found/);
    croak("Cannot see the lun $args{lun_path} on $client_id : $out")
            if ($out =~ /not available/);

    my @lun_dev;
    if ($args{protocol} eq 'fcp') {
        # FCP
        # Verify that the output of sanlun is correcto
        # extract disks from sanlun output
        # make sure we see correct number of paths to lun depending on cfmode
        # randomly pick one disk
        # format disk

        # expected output
        #sanlun lun show tada-u23:/vol/vol0/lun0
        #    filer:         lun-pathname         device filename         adapter         lun size        lun state
        #  tada-u23:       /vol/vol0/lun0        /dev/rdsk/c3t3d0s2      lpfc0     10m (10485760)        GOOD
        #  tada-u23:       /vol/vol0/lun0        /dev/rdsk/c4t3d0s2      lpfc1     10m (10485760)        GOOD
 
        # pick the device filename
        while ($out =~ /$filer_id:\s+$args{lun_path}\s+(\/dev\/rdsk\/c\dt\dd\ds\d)\s+(lpfc|td)\d+\s+.*GOOD/gm) {
            push(@lun_dev,$1) if ($1);
        }

        # make sure we see multiple entries in sanlun if igroup is associated with multiple adapters
        # number of paths is dependant on the cfmode of the fcp (standby=2paths, partner=4paths)
        # mixed not sure
        my $lun_paths_expected = scalar(@client_adapter);
        if ($args{protocol} eq 'fcp') {
            $out = sayEOF($fc, "fcp show cfmode");
            if ($out =~ /partner/) {
                $lun_paths_expected *= 2;
            } elsif ($out =~ /mixed/) {
                $lun_paths_expected *= 4;
            }
        }

        carp("Unable to see $lun_paths_expected entries in sanlun for lun $filer_id:$args{lun_path}\n$out\n")
            if (scalar(@lun_dev) != $lun_paths_expected);

        logcomment("LUN devices are @lun_dev");

        # Now format the disk
        if ($args{format}) {
            my $disk = $lun_dev[int(rand(@lun_dev))];
            format_lun($disk, $client);
        }

    } else {
        # iSCSI
        # make sure sanlun output is good
        # extract device name from output

        # expected output
        #sanlun lun show tada-u23:/vol/vol0/luns
        #    filer:         lun-pathname         device filename          adapter        protocol        lun size        lun state
        #  tada-u23:       /vol/vol0/luns        /dev/sdb                iscsictl1       iSCSI      10m (10485760)       GOOD

        # pick the device filename
        #while ($out =~ /$filer_id:\s+$args{lun_path}\s+(\/dev\/sd\w+)\s+iscsictl\d+\s+iSCSI\s+.*GOOD/gm) {
        while ($out =~ /$filer_id:\s+$args{lun_path}\s+(\/dev\/sd\w+)\s+<unknown>|iscsictl\d+\s+iSCSI\s+.*GOOD/gm) {
            push(@lun_dev,$1) if ($1);
        }
        croak("Unable to see device entries in sanlun for lun $filer_id:$args{lun_path}\n$out\n")
            if (!scalar(@lun_dev));

        logcomment("LUN devices are @lun_dev");
    }

    # now fcp lun/disk is created, opened, formatted(if specified) and ready for use
    return(@lun_dev);

    };
}

=head2 create_volgrp($client,$vg_name,$device,%opts)

If using a high availibility configuration on HP-UX then we need
to create logical volumes from a pool of available LUN storage.
This routine will create a volume group on HP-UX clients.

more_devices is a space separated list of devices. eg. "c10t0d0 c11t1d0"

=cut

sub create_volgrp($$$%) {
    my ($client, $vg_name, $device, %args) = validate_mixed(
                    @_,
                    [ { type=>SCALAR|HASHREF, optional=>0 }, 
                      { type=>SCALAR, optional=>0, regex=>qr{^\w+$} },
                      { type=>SCALAR, optional=>0, regex=>qr{^c\d+t\d+d\+$} },
                    ],
                    {
                     die_existing => { type=>SCALAR, optional=>1,
                                regex=>qr/^0|1$/, default=>0 },
                     more_devices => { type=>SCALAR, optional=>1,
                                regex=>qr{^(c\d+t\d+d\d+\s?)+$}, default=>0 },
                    }
                    );

    my $ch = make_host($client);
    my $client_hosttype = $ch->{hosttype};
    my $client_id = $ch->{id};
    my $cc = make_conn($client);
    my $sudo = '/usr/local/bin/sudo';

    die("Unsupported client $client_id. Must be of type HP-UX")
        if ($client_hosttype !~ /^unix-hpux/);

    # check if it exists already
    my $exist = volgrp_exist($client,$vg_name);
    croak("$@\n") if ($@);
    if ($exist) {
        croak("Vol group '$vg_name' exists already\n")
            if ($args{die_existing});
        logcomment("Using existing vol group '$vg_name'");
    }
    logcomment("Creating vol group '$vg_name'");
    # TBD if volgrp is in path form then use accordingly
    # create the vol group
    my (undef, undef, $out) = say($cc, "$sudo mkdir /dev/$vg_name && echo '_mkdir_ok_'");
    croak("Failed to make directory '/dev/$vg_name' : $out\n")
        if ($out !~ '_mkdir_ok_');

    (undef, undef, $out) = say($cc, "$sudo mknod /dev/$vg_name/group c 64 0x10000 && echo '_mknod_ok_'");
    croak("Failed to make node '/dev/$vg_name/group' : $out\n")
        if ($out !~ '_mknod_ok_');

    my $cmd = "$sudo vgcreate $vg_name $device";
    $cmd .= " $args{more_devices}" if ($args{more_devices});
    (undef, undef, $out) = say($cc, "$cmd && echo '_vgcreate_ok_'");
    croak("Failed to create the vol group : $out\n")
        if ($out !~ '_vgcreate_ok_');

    return(1);

}


=head2 volgrp_exist($client,$vg_name)

=over

=item Arguments

=over

=item $client

(Required): client (only HP-UX supported)

=item $vg_name

(Required): volume group or path. Eg. '/dev/vg01' or 'vg01'

=back

=item Notes

Supported on HP-UX clients only

=item Returns

1 on volume group exists

0 if volume group does not exist

on failure dies with error in $@

=back

=cut

sub volgrp_exist($$) {
    my ($client, $vg_name) = validate_pos( @_,
                    { type => SCALAR|HASHREF }, 
                    { type => SCALAR },
                    );

    my $ch = make_host($client);
    my $client_hosttype = $ch->{hosttype};
    my $client_id = $ch->{id};
    my $cc = make_conn($client);

    die("Unsupported client $client_id. Must be of type HP-UX")
        if ($client_hosttype !~ /^unix-hpux/);

    my $out = sayEOF($cc, "vgdisplay $vg_name");
    return(($out =~ /does not exist/i) ? 0 : 1);
}


=head2 igroup_exist($filer,$igroup)

Checks if igroup exists or not

=over

=item Arguments

=over

=item filer

(Required) : Hostspec of filer

=item igroup

(Required) : name of igroup

=back


=back

=for html <hr />

=cut

sub igroup_exist ( $$ ) {

    my ($filer, $igroup) = validate_pos(
                    @_,
                    { type=>SCALAR|HASHREF, optional=>0 }, 
                    { type=>SCALAR, optional=>0 },
                    );

    my $conn = make_conn($filer);
    my $out = sayEOF($conn, "igroup show $igroup");
    return (($out =~ /No such initiator group exists/i) ? 0 : 1);

}


=head2 igroup_destroy($filer,$igroup)

=over

=item Notes

if lun maps for the igroup exist then routine will die.

=back

=cut

sub igroup_destroy ( $$% ) {

    my ($filer, $igroup, %args) = validate_mixed(
                    @_,
                    [ { type=>SCALAR|HASHREF, optional=>0 }, 
                      { type=>SCALAR, optional=>0 },
                    ],
                    {
                     force => { type=>SCALAR, optional=>1,
                                regex=>qr/^0|1$/, default=>0 },
                     verify => { type=>SCALAR, optional=>1,
                                regex=>qr/^0|1$/, default=>1 },
                    }
                    );

    my $conn = make_conn($filer);
    my $cmd = "igroup destroy";
    $cmd .= " -f" if $args{force};
    $cmd .= " $igroup";
    my $out = sayEOF($conn, $cmd);
    croak("Unexpected output : $out\n") if ($out); # die if we get any output
    # if lun maps for the igroup exist then routine will die.
    
    croak("igroup $igroup did not get destroyed\n") 
        if ($args{verify} && igroup_exist($filer, $igroup));

    return (1);

}


=head2 igroup_info($filer,$igroup)

Get igroup information

=over

=item Returns

($protocol, $ostype, @nodes)

=item Notes

does only -v option

=back

=cut

sub igroup_info($$) {

    my ($filer, $igroup) = validate_pos(
                    @_,
                    { type => SCALAR|HASHREF }, 
                    { type => SCALAR },
                    );

    my $conn = make_conn($filer);

    my $out = sayEOF($conn, "igroup show -v $igroup");
    croak("$out\n") if ($out =~ /not found|No such.*exists/i);

    # split into lines
    my @lines = split(/\n/,$out);
    croak("Received insufficient output : $out\n")
        unless (scalar(@lines) >= 3);

    # do first line
    my $line = shift(@lines);
    croak("Unknown format for first line of igroup show -v : '$line'") unless
        ($line =~ /^\s*$igroup \((FCP|iSCSI)\):$/);
    my $protocol = $1;

    # do second line
    $line = shift(@lines);
    croak("Unknown format for second line of igroup show -v : '$line'") unless
        ($line =~ /^\s*OS Type: (default|solaris|windows|hpux|aix|linux)$/);
    my $ostype = $1;

    # do third line
    # this line occurs iff filer(console) is in advanced/diag/test mode
    if ($protocol eq 'FCP') {
        $line = shift(@lines);
        if ($line !~ /^\s*Host Multipathing Software: (Optional|Required)$/) {
            if ($line =~ /^\s*Member:/) {
                unshift(@lines,$line); # put the line back
            } else {
                croak("Unknown format for third line of igroup show -v : '$line'");
            }
        }
    }

    # do rest of lines
    my @nodes;
    while (@lines) {
        $line = shift(@lines);
        my $node;
        if ($protocol eq 'FCP') {
            croak("Unknown format for igroup show -v : '$line'") unless
                # TBD check to see if this works for different cfmodes
                ($line =~ /^\s*Member: (\w\w:\w\w:\w\w\:\w\w:\w\w:\w\w:\w\w:\w\w)\s*(?:\((?:not logged in|logged in on): (?:\w+,?\s*)+\))?$/);
                # XXX - from TS(7.1) onwards will show if adapter has logged in
            $node = $1;
        } else {
            croak("Unknown format for igroup show -v : '$line'") unless
                ($line =~ /^\s*Member: (iqn.\S+)/);  # Changed from what is shown below for BURT 242536 - mporterf
#                ($line =~ /^\s*Member: (iqn.\S+.com.cisco.\d+.\w+)\s*(?:\((?:not logged in|logged in on): \w+\))?$/);
                # XXX - from TS(7.1) onwards will show if adapter has logged in
            $node = $1;
        }
        push(@nodes, $node);
    }

    return($protocol, $ostype, @nodes);

}



=head2 igroup_create($filer,$client,$igroup,%opts)

Create an igroup.
Works only for solaris fcp or linux iscsi client.

=over

=item Arguments

=over

=item filer 

(Required) : Hostspec of filer

=item client

(Required) : Hostspec of solaris fcp or linux iscsi client

=item igroup

(Required) : Name of igroup to create

=item sanlun_path

(Optional) : Path where sanlun binary lives. Default: /usr/sbin

=back

=item Returns 

on success returns 1
dies on failure

=back

=for html <hr />

=cut

sub igroup_create( $$$% ) {

    my ($filer, $client, $igroup, %args) = validate_mixed(
                    @_,
                    [ { type => SCALAR|HASHREF }, 
                      { type => SCALAR|HASHREF },
                      { type => SCALAR } ],
                    {
                     ostype => { type=>SCALAR, optional=>1, 
                                    regex=>qr/^(default|solaris|windows|hpux|aix|linux)$/ },
                     protocol => { type=>SCALAR, optional=>1, 
                                    regex=>qr/^(iscsi|fcp)$/, default=>'fcp' },
                     sanlun_path => { type=>SCALAR, optional=>1, 
                                    default=>'/usr/sbin', regex=>qr/^(\/\w+\/?)+$/ },
                     verify => { type=>SCALAR, optional=>1, default=>1,
                                    regex=>qr/^0|1$/ },
                     purge_existing => { type=>SCALAR, optional=>1, default=>0,
                                    regex=>qr/^0|1$/ },
                    } );

    my $ch = make_host($client);
    my $client_id = $ch->{id};
    my ($client_os) = ($ch->{hosttype} =~ /^unix-(linux|solaris)/);

    if ($args{ostype}) {
        croak("igroup_create: ostype is specified as $args{ostype}, ". 
            "but $client_id has OS $client_os") if ($args{ostype} ne $client_os);
    }
            
    my $h = make_host($filer);
    my $filer_id = $h->{id};
    my $conn = make_conn($filer);

    if (igroup_exist($filer,$igroup)) {
        if ($args{purge_existing}) {
            logcomment("Will destroy existing igroup $igroup");
            igroup_destroy($filer,$igroup);
        } else {
            logcomment("igroup $igroup already exists");
            return(1);
            # XXX - no verification is done to see if existing igroup
            # is user wants
        }
    }

    # Create igroup on filer associated with CLIENT
    # If igroup does not exist the creation will succeed else 
    # creation will fail and we just use the existing igroup ... 
    # hopefully its the igroup associated with the client 
    # and adapter that we want
    my (%initiator) = initiator_node_info($client, protocol=>$args{protocol}, 
                        sanlun_path=>$args{sanlun_path});
    my @client_adapter = keys(%initiator);
    my @client_node = values(%initiator);


    logcomment("Creating igroup $igroup on $filer_id associated with".
                 " $client_id @client_adapter @client_node");
    my $cmd = "igroup create";
    $cmd .= ($args{protocol} eq 'fcp' ? ' -f' : ' -i');
    $cmd .= " -t $args{ostype}" if ($args{ostype});
    $cmd .= " $igroup @client_node";

    my $out = sayEOF($conn, $cmd);
    croak("Unexpected output : $out\n") if ($out); # die if we get any output

    if ($args{verify}) {
        # verify if igroup creation succeeded
        # in case of FCP verify if client adapter logged in
        $out = sayEOF($conn, "igroup show $igroup");
        croak("Creation of igroup $igroup failed : $out\n") if ($out =~ /no.*exists/i);
        if ($args{protocol} eq 'fcp') {
            if ($args{ostype} =~ /^unix-solaris/) {
                croak("Adapter(s) @client_adapter from $client did not log into ".
                    " filer $filer_id igroup $igroup: $out\n")
                    if ($out =~ /not logged in/i);
            } else {
                logcomment("Skipping verification of adapters logging into igroup");
            }
        }
    }

    return(1);

}



=head2 initiator_node_info($spec,%opts)

This routine will get the node information from the client.
Its basically a wrapper around the command
    sanlun (iscsi/fcp) show adapter.

FCP nodes are specified as world wide port names (WWPN), written
as 16 hexadecimal characters with optional (:) characters.

iSCSI nodes are written in the dotted-domain fashion.


=over

=item Arguments

=over

=item spec

(Required) : hostspec of FCP / iSCSI client
Currently only solaris FCP or linux iSCSI is supported

=item protocol

(Optional) : name of protocol to use. Possible values iscsi, fcp.

=item adapter

(Optional) : name of adapter on client

=item sanlun_path

(Optional) : place where sanlun binary lives. Default: /usr/sbin

=back

=item Notes

This routine depends heavily on output of 'sanlun'

-v and -c option not implemented as yet

does not support multiple adapters for iscsi

=item Returns


=back

=for html <hr />

=cut

sub initiator_node_info( $% ) {

    my ($spec, %args) = validate_mixed(
                    @_,
                    [ { type => SCALAR|HASHREF }, ],
                    {
                     protocol => { type=>SCALAR, optional=>1, 
                                    regex=>qr/^(iscsi|fcp)$/, default=>'fcp' },
                     adapter => { type=>SCALAR, optional=>1, 
                                    regex=>qr/^(lpfc|td|iscsictl|all)/ },
                     sanlun_path => { type=>SCALAR, optional=>1, 
                                    default=>'/usr/sbin', regex=>qr/^(\/\w+\/?)+$/ },
                    } );

    my $ch = make_host($spec);
    my $hosttype = $ch->{hosttype};
    my $hostname = $ch->{id};
    my $cc = make_conn($spec);

    my $sudo = '/usr/local/bin/sudo';
    my $path;
    if ($hosttype =~ /^unix-(hpux|solaris)/) {
        $path = '/usr/bin';
        croak("initiator_node_info: Routine does not support iSCSI on solaris / hp-ux")
            if ($args{protocol} eq 'iscsi');
    } elsif ($hosttype =~ /^unix-linux/) {
        $path = '/bin';
        croak("initiator_node_info: Routine does not support FCP on linux")
            if ($args{protocol} eq 'fcp');
    } else {
        croak("initiator_node_info: $hosttype client $hostname not supported\n");
    }


	# ok, sanlun does not show the initiatorname for iscsi.  I can't find a command line that
	# will return it.  So lets change it to a grep on the initiatorname.iscsi file.
   # Leaving the code the same for the FCP case.  RFC 3720 covers the iscsi spec and section
	# 3.2.6.3 covers naming conventions. RFC 3980 covers proposed extensions.
	my $cmd;
	if ($args{protocol} eq 'fcp') {
    	$cmd = "$sudo $args{sanlun_path}/sanlun $args{protocol} show adapter";
    	$cmd .= " $args{adapter}" if ($args{adapter});
	}
	else { # iscsi
		# find the uncommented InitiatorName line from /etc/initiatorname.iscsi and 
		# trim off the 'InitiatorName=' part.
		# May be able to do without the full paths, but lets not take chances.
		$cmd = "/usr/local/bin/sudo /usr/bin/grep '^InitiatorName'  /etc/initiatorname.iscsi | /usr/bin/cut -c 15-";
	}

    my $out = sayEOF($cc, $cmd);
    croak("Host attach kit 'sanlun' missing on $hostname: $out\n") 
        if ($out =~ /command not found/);
    croak("$out\n") if ($out =~ /No supported adapters are present/);
    croak("$out\n") if ($out =~ /Unable to get adapter info/);

    my $expected;
    if ($args{protocol} eq 'fcp') {
        # fcp part
        $expected = ($args{adapter} ? qr/^($args{adapter})\s+WWPN:(\w{16})$/ 
                                    : qr/^((?:lpfc|td)\d+)\s+WWPN:(\w{16})$/);
    } else {
        # iscsi part
        $expected = qr/^((iqn\.\d\d\d\d-\d\d\.com\.cisco:.*))$/;
#iscsictl1               initiator name: iqn.1987-05.com.cisco:01.66b2e5cc627f
    }

    my %adapters;
    my @lines = split(/\n/, $out);
    while (@lines) {
        my $line = shift(@lines);
        next if (!$line); # next if blank line
        if ($line =~ $expected) {
            $adapters{$1}=$2;
        } else {
            croak("Unexpected output.\nExpected : '$expected'.\nGot : '$line'\n");
        }
    }

    return(%adapters);

}



=head2 iscsi_show_initiator($spec,%opts)

=over

=item Arguments

=over

=item spec

(Required): filer host id / connection to filer / hostspec of filer.

=item adapter

(Optional): name of adapter to get initiator for

=back

=item Returns

On success returns a list containing 
($adapter, $target_pg, $initiator_name, $isid);
dies on failure

=item Notes

adapter option not implemented as yet

does not work for multiple initiators

    tada-u22(takeover)> iscsi show initiator
    Initiators connected on adapter iswta:
      Tgt_PG  iSCSI Initiator Name / ISID
         2    iqn.1987-05.com.cisco:01.65d299af6340 / 00:02:3d:00:00:01
         2    iqn.1987-05.com.cisco:01.66b2e5cc627f / 00:02:3d:00:00:01

    Adapter iswtb operates on behalf of the partner host.
    Use the partner command to view this adapter when in takeover mode.

=back

=cut

sub iscsi_show_initiator ( $% ) {

    my ($spec, %args) = validate_mixed(
                    @_,
                    [ { type => SCALAR|HASHREF }, ],
                    {
                     adapter => { type=>SCALAR, optional=>1, regex=>qr/^iswt(a|b)$/ },
                    # XXX - Not yet implemented
                    } );

    my $conn = make_conn($spec);
    my $txt = sayEOF($conn, "iscsi show initiator");
    croak("$txt\n") if ($txt =~ /service is not licensed/);

    if ($txt =~ /Initiators connected on adapter (iswt(a|b)):.*\s+Tgt_PG\s+iSCSI Initiator Name\s+\/\s+ISID.*\s+(\d+)\s+(iqn.\S+.com.cisco.\d+.\w+)\s+\/\s+(\w\w:\w\w:\w\w:\w\w:\w\w:\w\w)/) {
        return($1, $3, $4, $5);
    } else {
        croak("Unexpected output : $txt\n");
    }
}

=head2 iscsi_nodename($spec,%opts)

Gets or sets the iscsi nodename on the filer.
If option new_nodename=>'' is specified then 
a set operation is performed else the current nodename
of the filer is returned.

=over

=item Arguments

=over

=item spec

(Required): filer host id / connection to filer / hostspec of filer.

=item new_nodename

(Optional): new node name to be set (must start with 'iqn.1992-08.com.netapp:')

=item verify

(Optional): Flag to verify the setting of the new nodename. 
Possible Values: 1/0. Default: 1(do verification)

=back

=item Returns

Set operation, 
On success returns 1
dies on failure

Get operation,
On success returns nodename
dies on failure

=back

=cut

sub iscsi_nodename ( $% ) {

    my ($spec, %args) = validate_mixed(
                    @_,
                    [ { type => SCALAR|HASHREF }, ],
                    {
                     new_nodename => { type=>SCALAR, optional=>1, 
                                regex=>qr/^iqn.1992-08.com.netapp:\w+.\w+$/ },
                     verify => { type=>SCALAR, optional=>1, default=>1, 
                                regex=>qr/0|1/, depends=>['new_nodename'] },
                    } );

    my $conn = make_conn($spec);

    if ($args{new_nodename}) {
        # set the nodename to new value
        my $txt = sayEOF($conn, "iscsi nodename $args{new_nodename}");
        croak("$txt\n") if ($txt =~ /service is not licensed/);

        # expect no ouput
        croak("Unexpected output : $txt\n") if ($txt);

        if ($args{verify}) {
            my $new_nodename = iscsi_nodename($conn);
            croak("Verification failed. iSCSI nodename has not changed". 
                "to '$args{new_nodename}' as expected. Its '$new_nodename'")
            if ($new_nodename !~ $args{new_nodename});
        }

        return(1);

    } else {

        # get the nodename
        my $txt = sayEOF($conn, "iscsi nodename");
        croak("$txt\n") if ($txt =~ /service is not licensed/);

        if ($txt =~ /iSCSI target nodename: (iqn.1992-08.com.netapp:\w+.\w+)/) {
            return($1);
        } else {
            croak("Unexpected output : $txt\n");
        }
    }
}

=head2 iscsi_target_info($client,%opts)

=over

=item Arguments

=over

=item client

(Required) : iSCSI client.

=item target_name

(Optional) : iSCSI target name.
If provided will get the target information for just this target.

=back

=item Returns

on success returns a Hash of Hashes with target information
indexed by target name

on failure dies with error in $@

=back

=cut

sub iscsi_target_info ( $% ) {

    my ($client, %args) = validate_mixed(
                @_,
                [ { type=>SCALAR|HASHREF, optional=>0 } ],
                {
                 target_name => { type=>SCALAR, optional=>1,
                            regex=>qr/^iqn.\S+.com.netapp:sn.\d+$/ },
                } );

    my $ch = make_host($client);
    my $cc = make_conn($client,"#rlogin");

    my $sudo = '/usr/local/bin/sudo';

    my $out = sayEOF($cc, "$sudo /sbin/iscsi-ls"); 
    my @blocks = split(/\*{79}/,$out);
    shift(@blocks); shift(@blocks); # throw out version line
    my %HoH;
    foreach my $block (@blocks) {

        my $expected = qr/TARGET NAME\s+:\s+(iqn.\S+.com.netapp:\S+).*HOST NO\s+:\s+(\d+).*BUS NO\s+:\s+(\d+).*TARGET ID\s+:\s+(\d+).*TARGET ADDRESS\s+:\s+(\d+.\d+.\d+.\d+:\d+)/s;
        if ($block =~ $expected) {
            my $target_name = $1;
            next if ($args{target_name} && ($target_name ne $args{target_name}));
            $HoH{$target_name}{host_no} = $2;
            $HoH{$target_name}{bus_no} = $3;
            $HoH{$target_name}{target_id} = $4;
            $HoH{$target_name}{target_address} = $5;
        } else {
            croak("Got unexpected result.\n$block");
        }
    }

    return(%HoH);
}

=head2 create_veritas_vol($filer,$client,%opts)

create a veritas volume


=over

=item Arguments

=over

=item filer

(Required) : hostspec of filer


=item client

(Required) : hostspec of client


=item igroup

(Optional) : name of igroup to use


=item filer_vol

(Optional) : volume on filer to use


=item lun_name 

(Optional) : name of lun to create


=item lun_size

(Optional) : size of lun


=item veritas_disk_grp

(Optional) : disk grop to use for veritas volume.


=item veritas_vol

(Optional) : name of veritas volume


=item veritas_vol_size

(Optional) : size of veritas volume


=back

=back

=for html <hr />

=cut

sub create_veritas_vol($$%) {


    my ($filer,$client,%args) = @_;
    my $sudo ="/usr/local/bin/sudo";

    my $lun_size = defined($args{lun_size}) ? $args{lun_size} : '4g';
    my $lun_name = defined($args{lun_name}) ? $args{lun_name} : 'multi_lun';
    my $igroup = defined($args{igroup}) ? $args{igroup} : 'multi_igrp';
    my $veritas_disk_grp = defined($args{veritas_disk_grp}) ?  $args{veritas_disk_grp} : 'catdg';
    my $veritas_vol = defined($args{veritas_vol}) ?  $args{veritas_vol} : 'cat_vol';
    my $veritas_vol_size = defined($args{veritas_vol_size}) ?  $args{veritas_vol_size} : '3g';
    #-----------------------------------------------------
    # If filer_vol not defined use root volume
    # if defined then check if it exists and is online
    #-----------------------------------------------------
    my $fc = make_conn($filer);

    my $filer_vol;
    if (!defined($args{filer_vol})) {
        $filer_vol = vol_root($fc);
    } else {
        if (vol_exists($fc, $args{filer_vol})) {
            my $obj = vol_status($fc, $args{filer_vol})
                            or die($@);
            my $state = $obj->state();
            if (defined($state) && $state eq 'online') {
                $filer_vol = $args{filer_vol};
            } else {
                die("$args{filer_vol} is not online\n");
            }
        } else {
            die("$args{filer_vol} does not exist\n");
        }
    }


    # create lun (make sure it maps both adapters)
    my @all_disks = create_lun($filer,$client,
                    filer_vol=>$filer_vol,
                    lun_name=>$lun_name,
                    lun_size=>$lun_size,
                    igroup=>$igroup) or
        logresult('FATAL', "Unable to create lun $lun_name on $filer :$@");
    logcomment("Lun device created is.... @all_disks");

    # add disk to veritas
    my $lun_dev = $all_disks[int(rand(@all_disks))]; # random pick one
    my ($disk) = ($lun_dev =~ /\/dev\/rdsk\/(c\dt\dd\d)s\d/);
    vxdiskadd($client,$disk,disk_group=>$veritas_disk_grp) or
        logresult('FATAL', "create_veritas_vol: Unable to add disk $disk to veritas grp $veritas_disk_grp on $client");


    my %device = vxdisk_list($client,$disk);
    logcomment("DEVICE:$disk TYPE:$device{type} DISK:$device{disk} GROUP:$device{group} STATUS:$device{status}");
    logresult('FATAL', "create_veritas_vol: Something wrong with veritas disk $disk ". 
                        "status: $device{status} on client $client")
        if ($device{status} !~ /online/);

    # verify if its multipath and both paths are enabled
    my (%dev_multi) = vx_dmp_status($client,$disk);
    logresult('FATAL', "create_veritas_vol: Not multi-path. Number of paths ". 
                        "to $disk on $client is $dev_multi{num_path}")
        if ($dev_multi{num_path} < 2);

    logcomment("PATH 1 is  ... $dev_multi{path1_dev} $dev_multi{path1_state}");
    logcomment("PATH 2 is  ... $dev_multi{path2_dev} $dev_multi{path2_state}");
    logresult('FATAL', "create_veritas_vol: Both paths to $disk on $client are not enabled")
        if ($dev_multi{path1_state} !~ /enabled/) || ($dev_multi{path2_state} !~ /enabled/);

    # create veritas volume
    my $cc = make_conn($client);
    my $out = sayEOF($cc, 
        "$sudo /usr/sbin/vxassist -g $veritas_disk_grp make $veritas_vol $veritas_vol_size $device{disk}");
    logresult('FATAL', "Problem with veritas volume creation: $out")
        if ($out =~ /(ERROR|Cannot allocate space)/i);

    # how to check for errors
    # there must be a block device file in /dev/vx/dsk/<disk_group>/<vol_name>
    return("/dev/vx/dsk/$veritas_disk_grp/$veritas_vol");

}




=head2 vxdisk_list($client,$disk,%opts)

get the output of vxdisk list

=over

=item Arguments

=over

=item client

(Required) : Hostspec of client running veritas

=item disk  

(Required) : disk name

=item veritas_path

(Optional) : path where veritas is installed.  Default: /usr/sbin

=back

=item Returns

(%device)

=back

=for html <hr />

=cut

# TBD fix this for following scenario
#r vxtada-c63.lab.netapp.com% disk list
#DEVICE       TYPE      DISK         GROUP        STATUS
#c1t0d0s2     sliced    -            -            error
#c3t0d0s2     sliced    catdg01      catdg        online
#c3t0d1s2     sliced    disk01       rootdg       online
#c3t0d2s2     sliced    -            -            online invalid
sub vxdisk_list($$%) {

    my ($client,$disk,%args) = @_;
    my %device;

    return eval {
    my $veritas_path = defined($args{veritas_path}) ? $args{veritas_path} : '/usr/sbin';
    my $sudo ="/usr/local/bin/sudo";

    my $cc = make_conn($client);
    my $out = sayEOF($cc,"$sudo $veritas_path/vxdisk list");
    die("Veritas not installed on $client") if ($out =~ /command not found/i);
    return(0) if ($out =~ /Disk not in the configuration/i);

    $disk =~ s/s\d//g; # throw off the slice number
    ($device{type},$device{disk},$device{group},$device{status}) = 
            ($out =~ /$disk\w\d\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/);

    return(%device);

    };
}


=head2 vx_dmp_status($client,$disk,%opts)

get multi-path status

=over

=item Arguments

=over

=item client

(Required) : Hostspec of client running veritas

=item disk  

(Required) : disk name

=item veritas_path

(Optional) : path where veritas is installed. Default: /usr/sbin

=back

=item Returns

(device)

=back

=for html <hr />

=cut

sub vx_dmp_status($$%) {

    my ($client,$disk,%args) = @_;
    my %device;

    return eval {
    my $veritas_path = defined($args{veritas_path}) ? $args{veritas_path} : '/usr/sbin';
    my $sudo ="/usr/local/bin/sudo";

    my $cc = make_conn($client);

    $disk =~ s/s\d//g; # throw off the slice number
    my $out = sayEOF($cc,"$sudo $veritas_path/vxdisk list $disk");
    die("Veritas not installed on $client") if ($out =~ /command not found/i);
    return(0) if ($out =~ /Disk not in the configuration/i);

    ($device{num_path}) = ($out =~ /numpaths:\s+(\d+)/gm);

    if ($device{num_path} == 2) {
        # will do for 2 paths only... # there shld be a better way to do this
        ($device{path1_dev},$device{path1_state},$device{path2_dev},$device{path2_state}) = 
            ($out =~ /numpaths:\s+\d+.*(c\dt\dd\ds\d)\s+state=(\S+).*(c\dt\dd\ds\d)\s+state=(\S+)/s);
    } elsif ($device{num_path} == 1) {
        ($device{path1_dev},$device{path1_state}) = 
            ($out =~ /numpaths:\s+\d+.*(c\dt\dd\ds\d)\s+state=(\S+)/s);
    } else {
        die("vx_dmp_status: script unable to handle more than 2 paths");
    }

    return(%device);

    };
}



=head2 vxdiskadd($client,$disk,%opts)

Add a disk to veritas


=over

=item Arguments

=over

=item client

(Required) : Hostspec of client running veritas

=item disk  

(Required) : disk to be added

=item disk_group

(Optional) : name of group to add the disk to.  Default: undef

=item veritas_path

(Optional) : path where veritas is installed.  Default: /usr/sbin

=back

=back

=for html <hr />

=cut

sub vxdiskadd($$%) {

    my ($client,$disk,%args) = @_;
    my $sudo ="/usr/local/bin/sudo";

    return eval {

    my $disk_group = defined($args{disk_group}) ? $args{disk_group} : undef; # default disk grp
    my $veritas_path = defined($args{veritas_path}) ? $args{veritas_path} : '/usr/sbin';

    my $cc = connect("$client#rlogin");
    my $out = sayEOF($cc,"$sudo $veritas_path/vxdctl enable");
    die("Veritas not installed on $client") if ($out =~ /command not found/i);
    sayEOF($cc,"$sudo $veritas_path/vxddladm addjbod vid=NETAPP");
    sayEOF($cc,"$sudo $veritas_path/vxddladm listjbod");

    logcomment("Adding disk $disk to veritas disk group $disk_group");

    my %replytable = (
        "Continue operation\?.*"                                => "y\r",
        "Which disk group.*"                                    => "$disk_group\r",
        "Use a default disk name for the disk\?.*"              => "y\r",
        "Add disk as a spare disk for.*"                        => "n\r",
        "Exclude disk from hot-relocation use\?"                => "n\r",
        "Continue with operation\?.*"                           => "y\r", 
        "Encapsulate this device\?.*"                           => "n\r",
        "Instead of encapsulating, initialize\?.*"              => "y\r",
        "Use a default private region length for this disk\?.*" => "y\r",
        "Create a new group named.*?.*"                         => "y\r",
        "Hit RETURN to continue.*"                              => "\r",
    );

    sayEOF($cc,"$sudo $veritas_path/vxdiskadd $disk",'-blind');
    my $tries;
    for ($tries=100;$tries;$tries--) {
        my (undef, undef, undef, $prompt) = say($cc,"",-prompt,[keys %replytable]);

        my $reply;
        foreach my $regex (keys %replytable) {
            if ($prompt =~ $regex) {
                $reply = $replytable{$regex};
            }
        }
        if ($reply) {
            say($cc,$reply,'-nocr','-blind');
        } else {
            # normal prompt, so vxdiskadd must be complete
            last;
        }
    }

    #disconnect($cc);

    if (!$tries) {
        die "vxdiskadd: too many prompts during \"vxdiskadd $disk\"\n";
    }

    my %device = vxdisk_list($client,$disk,veritas_path=>$veritas_path);
    return(($device{status} !~ /online/) ? 0 : 1);

    };
}


=head2 vx_get_vol_info($client,$vol_path,%opts)

get veritas volume info

=over

=item Arguments

=over

=item client

(Required) : Hostspec of client

=item vol_path

(Required) : Path of veritas volume. Specify vol_path like 
/dev/vx/dsk/catdg/cat_vol. For my limited use of veritas this routine will 
work but it will not work if u make volumes using multiple disks etc

=item veritas_path

(Optional) : path where vxdisk can be found.  Default: /usr/sbin

=back

=item Returns

List containing ($disk_grp, $vol, $disk, $device);

=back

=for html <hr />

=cut

sub vx_get_vol_info($$%) {

    my ($client,$vol_path,%args) = @_;
    my $sudo ="/usr/local/bin/sudo";
    
    eval {
    my $veritas_path = defined($args{veritas_path}) ? $args{veritas_path} : '/usr/sbin';

    my $cc = make_conn($client);
    my $out = sayEOF($cc,"/usr/bin/ls $vol_path");
    die("Could not find $vol_path on $client") if ($out =~ /not found/i);

    my ($disk_grp, $vol) = ($vol_path =~ /\/dev\/vx\/dsk\/(\S+)\/(\S+)/);

    $out = sayEOF($cc,"$sudo $veritas_path/vxdisk scandisks");
    die("Veritas not installed on $client") if ($out =~ /command not found/i);
    die("Unable to veritas scandisks on $client $out") if ($out =~ /ERROR/i);

    $out = sayEOF($cc,"$sudo $veritas_path/vxprint -g $disk_grp");
    die($out) if ($out =~ /ERROR/i);
    my ($disk) = ($out =~ /sd\s+(\S+)-01\s+$vol-01\s+/gm);
    my ($device) = ($out =~ /dm\s+$disk\s+(c\dt\dd\ds\d)\s+/gm);
    logcomment("VERITAS DISK GROUP:$disk_grp VOLUME:$vol DISK:$disk DEVICE:$device");

    return($disk_grp,$vol,$disk,$device); 
    };
}

1;
