# -*-perl-*x
#
# $Id: //depot/prod/test/Rsierranevada/lib/SFOTest.pm#2 $
#
# Copyright (c) 2008 Network Appliance, Inc.
# All rights reserved.
#
## @summary
##     Library functions for vserver setup and sfo scenarios
##
## @description
##     The sfo_test library consists of the following functions:
##     setup_vserver - Creates the vserver with all the necessary setup
##     cleanup_vserver - Deletes the vserver along with all setup
##     transit_node_to_cfe - Transits a node to CFE prompt
##     transit_node_to_cli - To get a node booted and running ngsh
##     reboot_node - Reboots a node
##     panic_node - Panics a node
##     create_volume - Creates a volume on a paricular vserver
##     delete_volume - Deletes a volume from a paricular vserver
##     get_root_object - Gets the root aggregate/volume name
##     create_aggregate - Creates Aggregate
##     delete_aggregate - Deletes Aggregate
##     create_vserver - Creates just the Vserver
##     create_vif - Creates netwok interface
##     create_vserver_nfs - Creates Vserver NFS
##     create_vserver_export_policy - Creates Export Policy
##     create_route - Creates  Network Route
##     create_failover_rules - Creates Failover Policies
##     delete_route - Deletes network route
##     delete_vif - Deletes network interface
##     delete_routing_group - Deletes network routing group
##     delete_export_ruleset - Deletes export ruleset
##     delete_vserver_nfs - Deletes vserver NFS
##     delete_vserver - Deletes just the vserver
##     start_traffic - Mounts and starts traffic on a volume
##     verify_traffic - Verifies whether traffic is running
##     verify_storage_failover_status - Verifies if storage failove show is showing connected for both nodes
##     enable_storage_failover_status - Sets the enabled flag in storage faiover to true
##     get_storage_failover_field - Gets the corresponding field name from the storage failover status
##     giveback_from_node - Issues a give back from a given node.
##     up_both_nodes - Brings up both nodes - this will be usefull when a test fails in between.
##     set_aggr_type - Sets the aggr_type to either sfo or cfo
##     enable_ha - Enables HA on both nodes
##     diable_gdb - Disables gdb on both nodes
##     get_object_status - Gets status field after running vol/aggr status
##     get_object_fsid - Gets fsid of aggr/vol
##     set_object_fsid - Sets fsid on aggr/vol
##     run_vreport_tool - Fixes any stale volumes/aggrs/vservers

##
## @usage
##     use SFOTest;
##
## @author prasank@netapp.com
##
## @status private
package SFOTest;

use strict;
use warnings;
use VifMgr::VifConfig;
use Tharn qw (subtest host $Log);
use Params::Validate qw (validate validate_pos SCALAR SCALARREF OBJECT ARRAYREF HASHREF);
use Client;
use Populate qw(create_set delete_set);
use File::Basename qw (basename);
use File::Temp qw(:mktemp);
use Data::Dumper;
use NACL::APISet;
use Readonly;
Readonly my $IP => "0.0.0.0";
Readonly my $IP_MASK => "0.0.0.0/0";

#Global Variables
my $Prefix = "sfo_test";
our %ApiSetObjs;
use vars ('@ISA', '@EXPORT');
@ISA = qw(Exporter);
@EXPORT =
  qw(setup_vserver cleanup_vserver transit_node_to_cfe transit_node_to_cli reboot_node 
  panic_node create_volume delete_volume get_root_object create_aggregate delete_aggregate 
  create_vserver create_vif create_vserver_nfs create_vserver_export_policy 
  create_route create_failover_rules delete_route delete_vif delete_routing_group 
  delete_export_ruleset delete_vserver_nfs delete_vserver start_traffic verify_traffic 
  verify_storage_failover_status enable_storage_failover_status get_storage_failover_field 
  giveback_from_node up_both_nodes set_aggr_type enable_ha disable_gdb get_object_status
  get_aggr_name get_object_fsid set_object_fsid run_vreport_tool wait_till_file_exists
  verify_ems_messages get_ems_sequence_id get_ems_message_ids verify_nvsave_file
  cleanup_nvsave_files change_object_state verify_nvsave_delete clear_fiji get_node_scope_option
  wait_for_storage_failover_state wait_till_state transit_node_to_maint execute_console_cmd
  get_volume_option set_volume_option get_member_vol get_member_aggr set_coral_heartbeat
  get_ems_details verify_ems_messages_generated mount_volume run_nodescope_command
  get_latest_ems_seq_id set_storage_failover_field disable_ha takeover giveback wait_on_console
  set_node_scope_option enable_automatic_options disable_automatic_options get_object_field
  wait_on_cf_status get_ha_state get_state get_partner_node verify_nvsave_files get_console_filename
  reestablish_connection wait_on_aggr_state setup_failover_rules execute_cmd_on_ngsh
  power_cycle_node stop_traffic get_aggr_disk_list remove_disk_ownership assign_disk_ownership);

########################################################################
# Function:    setup_vserver
#
# Description: Create all the necessary vserver setup including network interfaces export policy etc.
#
# Parameters:  Server,filer_name,vserver,port_index,vserver_index.aggr_name.aggr_size
# Returns:     dies with error message on failure
#
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub setup_vserver {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            aggr_name     => {type => SCALAR, optional => 1},
                            aggr_size     => {type => SCALAR, optional => 1},
                            vserver_name  => {type => SCALAR, optional => 1},
                            vserver_index => {type => SCALAR, optional => 1},
                            port_index    => {type => SCALAR, optional => 1},
                            volume        => {type => SCALAR, optional => 1},
                            volume_count  => {type => SCALAR, optional => 1},
                            vserver_root_aggr  => {type => SCALAR, optional => 1},
                            volume_size   => {type => SCALAR, optional => 1},});
    my ($server,      $hostobj,
        $root_volume, $vserver_name,
        $aggr_name,   $aggr_type,
        $volume_size, $aggr_size,
        $raid_type,   $loopcount,
        $reply,       $filername_with_underscores,
        $volume_name, $vserver_index,
        $port_index,  $volume_count,
        $filername,   $vserver_root_aggr);

    $filername = $options{filername};
    $server    = $options{serverobj};

    #Set the current server
    #$server->set_execution_server(-Server => $filername);
    $filername_with_underscores = $filername;
    $filername_with_underscores =~ s/-/_/g;
    defined($options{aggr_size})
      ? ($aggr_size = $options{aggr_size})
      : ($aggr_size = 3);
    defined($options{raid_type})
      ? ($raid_type = $options{aggr_type})
      : ($raid_type = "raid_dp");
    defined($options{volume_size})
      ? ($volume_size = $options{volume_size})
      : ($volume_size = "2G");
    defined($options{vserver_index})
      ? ($vserver_index = $options{vserver_index})
      : ($vserver_index = 1);
    defined($options{port_index})
      ? ($port_index = $options{port_index})
      : ($port_index = 1);
    defined($options{aggr_name})
      ? ($aggr_name = $options{aggr_name})
      : ($aggr_name = get_aggr_name(filername=>$filername));
    defined($options{vserver_name})
      ? ($vserver_name = $options{vserver_name})
      : ($vserver_name =
         $Prefix . "_vs" . $vserver_index . "_" . $filername_with_underscores);
    defined($options{volume})
      ? ($volume_name = $options{volume})
      : ($volume_name = $Prefix . "_vol");
    defined($options{volume_count})
      ? ($volume_count = $options{volume_count})
      : ($volume_count = 1);
    defined($options{vserver_root_aggr})
      ? ($vserver_root_aggr = $options{vserver_root_aggr})
      : ($vserver_root_aggr = $filername_with_underscores."_aggr1");

    $root_volume = $vserver_name . "_root";
    $hostobj     = host($filername);
    $Log->trace("Entering create_vserver");
    #############################################
    # Create Aggregate
    ############################################
    $Log->trace("Create Aggregate");
    die "Create Aggregate FAILED"
      if (create_aggregate(serverobj => $server,
                           filername => $filername,
                           aggr_name => $aggr_name,
                           aggr_size => $aggr_size,
                           raid_type => $raid_type,));
    #############################################
    # Create VServer
    ############################################
    $Log->trace("Create VServer");
    die "Create Vserver FAILED"
      if (create_vserver(serverobj    => $server,
                         filername    => $filername,
                         vserver_name => $vserver_name,
                         aggr_name    => $vserver_root_aggr,
                         root_volume  => $root_volume));
    #############################################
    # Change the root vol permission
    ############################################
    $server->volume_modify(vserver            => $vserver_name,
                           volume             => $root_volume,
                           "unix-permissions" => "0777");
    #############################################
    # Change the root vol files param
    ############################################
    $server->volume_modify(vserver            => $vserver_name,
                           volume             => $root_volume,
                           "files" => "1024");
    #############################################
    # Create VIFs
    ############################################
    $Log->trace("Create VIFS");
    die "Vif createion FAILED"
      if (create_vif(serverobj    => $server,
                     filername    => $filername,
                     vserver_name => $vserver_name,
                     port_index   => $port_index,
                     hostobj      => $hostobj));
    #############################################
    # Create Volume
    ############################################
    for (my $k = 1 ; $k <= $volume_count ; $k++) {
        $Log->trace("Create test Volume $volume_name" . "" . $k);
        die "Volume creation FAILED"
          if (create_volume(serverobj    => $server,
                            filername    => $filername,
                            volume       => $volume_name . "" . $k,
                            vserver_name => $vserver_name,
                            volume_size  => $volume_size,
                            aggr_name    => $aggr_name));
    } ## end for (my $k = 1 ; $k <= ...
    #############################################
    # Create VServer NFS
    ############################################
    $Log->trace("Create VServer NFS");
    die "Create Vserver NFS FAILED"
      if (create_vserver_nfs(serverobj    => $server,
                             filername    => $filername,
                             vserver_name => $vserver_name));
    #############################################
    # Create VServer Export Policy
    ############################################
    $Log->trace("Create VServer Export Policy");
    die "Create Vserver Export Policy FAILED"
      if (create_vserver_export_policy(serverobj    => $server,
                                       filername    => $filername,
                                       vserver_name => $vserver_name));
    #############################################
    # Create network routing-group route
    ############################################
    $Log->trace("Create network routing-group route");
    die "Create route FAILED"
      if (create_route(serverobj    => $server,
                       filername    => $filername,
                       vserver_name => $vserver_name,
                       port_index   => $port_index));

#commenting temporaily due to burt436037 
if(0)
{
    #############################################
    # Create VServer Failover(s)
    ############################################
    my @nodes = $server->servers;
    foreach my $node (@nodes) {
        if ($node ne $filername) {
            die "Create VServer Failover for $filername FAILED"
              if (create_failover_rules(serverobj     => $server,
                                        filer1        => $filername,
                                        filer2        => $node,
                                        port_index    => $port_index,
                                        vserver_index => 1));
            last;
        }
    } ## end foreach my $node (@nodes)
}
}    #end_setup_vserver

########################################################################
# Function:    cleanup_vserver
#
# Description: Delete all the vserver related setup including network interfaces export policy etc.
#
# Parameters:  Server,filer_name,vserver,port_index,vserver_index.aggr_name
# Returns:     dies with error message on failure
#
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub cleanup_vserver {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            aggr_name     => {type => SCALAR, optional => 1},
                            vserver_name  => {type => SCALAR, optional => 1},
                            vserver_index => {type => SCALAR, optional => 1},
                            port_index    => {type => SCALAR, optional => 1},
                            volume        => {type => SCALAR, optional => 1},
                            volume_count  => {type => SCALAR, optional => 1},
                            volume_size   => {type => SCALAR, optional => 1},});

    my ($server,                     $hostobj,
        $root_volume,                $vserver_name,
        $aggr_name,                  $aggr_type,
        $volume_size,                $vifcfg,
        $aggr_size,                  $raid_type,
        $loopcount,                  $reply,
        $filername_with_underscores, $volume_name,
        $voulume_count,              $filername,
        $vserver_index,              $port_index,
        $volume_count);

    $filername = $options{filername};
    $server    = $options{serverobj};

    #Set the current server
    #$server->set_execution_server(-Server => $filername);
    $filername_with_underscores = $filername;
    $filername_with_underscores =~ s/-/_/g;
    defined($options{vserver_index})
      ? ($vserver_index = $options{vserver_index})
      : ($vserver_index = 1);
    defined($options{port_index})
      ? ($port_index = $options{port_index})
      : ($port_index = 1);
    defined($options{aggr_name})
      ? ($aggr_name = $options{aggr_name})
      : ($aggr_name = get_aggr_name(filername=>$filername));
    defined($options{vserver_name})
      ? ($vserver_name = $options{vserver_name})
      : ($vserver_name =
         $Prefix . "_vs" . $vserver_index . "_" . $filername_with_underscores);
    defined($options{volume})
      ? ($volume_name = $options{volume})
      : ($volume_name = $Prefix . "_vol");
    defined($options{volume_count})
      ? ($volume_count = $options{volume_count})
      : ($volume_count = 1);
    $root_volume = $vserver_name . "_root";
    $hostobj     = host($filername);
    $vifcfg      = new VifMgr::VifConfig();

    #$server->set_execution_server(-Server => $filername);
    ###########################################
    # Fix any stale objects
    ###########################################
    run_vreport_tool(serverobj=>$server);
    ###########################################
    #Delete extra cluster interface
    ###########################################
    die "Deletion of extra cluster interface FAILED"
	if(delete_extra_cluster_interface(serverobj=>$server,
				filername=>$filername));
    ###########################################
    #Delete the failover rules
    ###########################################
    $Log->trace("Delete the failover rules - $filername");
    eval {
        $server->network_interfaces_failover_delete(
                                                server   => $vserver_name,
                                                vif      => 'data1',
                                                priority => "1,2,3,4,5,6,7,8,9",
        );
    };

    #dont worry about the failure since its a blind cleanup
    ###########################################
    #Delete test volume
    ###########################################
    for (my $k = 1 ; $k <= $volume_count ; $k++) {
        $Log->trace("Delete test volume $volume_name" . "_" . $k." on $filername");
        die "Volume deleteion - $volume_name" . "" 
          . $k
          . " FAILED"
          if (delete_volume(serverobj    => $server,
                            filername    => $filername,
                            volume       => $volume_name . "" . $k,
                            vserver_name => $vserver_name));
    } ## end for (my $k = 1 ; $k <= ...
    ###########################################
    #Delete route
    ###########################################
    $Log->trace("Delete route");
    $Log->trace("Delete route FAILED")
      if (delete_route(serverobj    => $server,
                       filername    => $filername,
                       vserver_name => $vserver_name,
                       port_index   => 1));
    ###########################################
    #Delete network interface
    ###########################################
    $Log->trace("Delete Vif");
    $Log->trace("Delete Vif FAILED")
      if (delete_vif(serverobj    => $server,
                     filername    => $filername,
                     vserver_name => $vserver_name,
                     port_index   => 1));
    ###########################################
    #Delete routing group
    ###########################################
    $Log->trace("Delete routing group");
    $Log->trace("Delete routing group FAILED")
      if (delete_routing_group(serverobj    => $server,
                               filername    => $filername,
                               vserver_name => $vserver_name));
    ###########################################
    #Delete export ruleset
    ###########################################
    $Log->trace("Delete export ruleset");
    $Log->trace("Delete export ruleset FAILED")
      if (delete_export_ruleset(serverobj    => $server,
                                filername    => $filername,
                                vserver_name => $vserver_name));
    ###########################################
    #Delete Vserver NFS
    ###########################################
    $Log->trace("Delete Vserver NFS");
    die "Delete Vserver NFS FAILED"
      if (delete_vserver_nfs(serverobj    => $server,
                             filername    => $filername,
                             vserver_name => $vserver_name));
    ###########################################
    #Delete all Volumes
    ###########################################
    $Log->trace("Delete root volume");
    die "Delete root volume FAILED"
      if (delete_all_volumes(serverobj    => $server,
                        filername    => $filername,
                        vserver_name => $vserver_name));
    ###########################################
    #Delete Vserver
    ###########################################
    $Log->trace("Delete Vserver");
    die "Delete Vserver FAILED"
      if (delete_vserver(serverobj    => $server,
                         filername    => $filername,
                         vserver_name => $vserver_name));
    ###########################################
    #Delete Aggregate
    ###########################################
    $Log->trace("Delete Aggregate");
    die "Delete Aggregate FAILED"
      if (delete_aggregate(serverobj => $server,
                           filername => $filername,
                           aggr_name => $aggr_name));
} ## end sub cleanup_vserver

########################################################################
# Function:   transit_node_to_cfe
#
# Description: Transit the node to CFE
#
# Parameters:  $server, $filername
#
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub transit_node_to_cfe {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
			    timeout       => {type => SCALAR, optional=>1, default=>300},
                            filername     => {type => SCALAR},});
    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $timeout = $options{timeout};
    my $error_msg = "";
    $server->set_execution_server(-Server => $filername);
    $Log->trace("Invoking power cycle on $filername");
    
    my $host = host($filername);
    if(power_cycle_node(serverobj=>$server,
		     filername=>$filername,
		     error_msg=>\$error_msg)) {
  	$Log->trace("Some problem happened during power_cycle - ".$error_msg);
  	return 1;
    } else {
  	$Log->trace("Power Cycle succeeded");	
    }
    if(wait_till_state( serverobj     => $server,
                     filername        => $filername,
                     timeout          => $timeout,
                     state        => "CFE",
                   )) {
       $Log->trace("Did not reach CFE stte");
       return 1;
    }
    return 0;

} ## end sub transit_node_to_cfe

########################################################################
# Function:    transit_node_to_cli
#
# Description: transit_node_to_cli
#
# Parameters:  None
#
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub transit_node_to_cli {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
			    timeout       => {type => SCALAR, optional=>1, default=>300},
                            filername     => {type => SCALAR},});
    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $state="";
    my $timeout = $options{timeout};
    my $current=20;
    $Log->trace("Entered transit_node_to_cli on $filername");
    $server->set_execution_server(-Server => $filername);
    # For some reason, the transit_to_cli is
    # failing because it thinks its already in CLI
    # so calling $server->state once to refresh the state
    eval {
        $state = $server->state();
    };



    eval {
    	$state = $server->state();
    };
    if($@) {
    	$Log->trace("We have reached here probably if the $filername 's current state is PANIC - state - $state - error - ",$@);
    }
    $Log->trace("CURRENT STATE - $state");
    ################################################
    # Now Lets attempt a transit to CLI
    ################################################
    $Log->trace("Lets attempt a TRANSIT TO CLI with current state being ---  $state");
    eval {
      $server->transit_to(
        state => $server->CLI(),
	timeout=>$timeout,
      );
    };
    if($@) {
      $Log->trace("Transit1 to CLI on $filername FAILED " . $@);
      return 1;
    } 
    return 0;
} ## end sub transit_node_to_cli

########################################################################
# Function:    reboot_node
#
# Description: Reboot Node
#
# Parameters:  $server, $filername
#
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################

sub reboot_node {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
			    timeout       => {type => SCALAR, optional=>1, default=>300},
			    no_powercycle => {type => SCALAR, optional=>1}});
    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $timeout = $options{timeout};
    my $fail_count=0;
    my $no_powercycle;
    defined($options{no_powercycle})
      ? ($no_powercycle = 1)
      : ($no_powercycle= 0);

    $Log->trace("Reboot the node");
    $server->set_execution_server(-Server => $filername);
    if($no_powercycle) {
      eval {
       $server->system_reboot("inhibit-takeover" => "true",
                                   node => $filername);
      };
      if(transit_node_to_cli(serverobj=>$server,
			filername=>$filername)) {
    	$Log->trace("As part of reboot_node Transit $filername to CLI FALIED");
	return 1;
      }
      return 0;
    }

    $Log->trace("As part of reboot_node first call Transit to CFE on $filername");
    if(transit_node_to_cfe(serverobj=>$server,
			timeout=>$timeout,
			filername=>$filername)) {
    	$Log->trace("As part of reboot_node Transit $filername to CFE FALIED");
	$fail_count++;
    }
    $Log->trace("As part of reboot_node Lets wait for sometime before transiting $filername to CLI");
    sleep(10);
    if(transit_node_to_cli(serverobj=>$server,
			timeout=>$timeout,
			filername=>$filername)) {
    	$Log->trace("As part of reboot_node Transit $filername to CLI FALIED");
	$fail_count++;
    }
    if($fail_count > 0) {
        $Log->trace("Rebooting $filername FAILED.");
	return 1;
    } else {
        $Log->trace("Rebooting $filername succeeded...");
	return 0;
    }
} ## end sub reboot_node

########################################################################
# Function:    panic_node
#
# Description: Panic Node
#
# Parameters:  $server,filername
#
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub panic_node {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},});
    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    $server->set_execution_server(-Server => $filername);
    eval {
        $server->system_node_run(node       => $filername,
                                 '-Timeout' => 300,
                                 command    => "\"priv set test;panic -D\"");
    };
    if ($@) {
        $Log->trace("$filername successfully paniced");
        return 0;
    } else {
        $Log->trace("$filername not paniced ");
        return 1;
    }
} ## end sub panic_node

########################################################################
# Function:    create_volume
#
# Description: Create volume
#
# Parameters:  $server,filername
#
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub create_volume {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            aggr_name     => {type => SCALAR, optional => 1},
                            vserver_name  => {type => SCALAR, optional => 1},
                            vserver_index => {type => SCALAR, optional => 1},
                            port_index    => {type => SCALAR, optional => 1},
                            volume        => {type => SCALAR, optional => 1},
                            junction_path => {type => SCALAR, optional => 1},
                            space_guarantee => {type => SCALAR, optional => 1},
                            state => {type => SCALAR, optional => 1},
                            volume_type => {type => SCALAR, optional => 1},
                            type => {type => SCALAR, optional => 1},
                            snapshot_policy => {type => SCALAR, optional => 1},
                            volume_size   => {type => SCALAR, optional => 1},});

    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $volume_name = $options{volume};
    my ($volume_size, $filername_with_underscores, $aggr_name, $vserver_name,
        $vserver_index, $junction_path, $space_guarantee, $state, $volume_type, 
	$stripe_format, $type, $snapshot_policy, $nvfail);
    $filername_with_underscores = $filername;
    $filername_with_underscores =~ s/-/_/g;
    defined($options{volume_size})
      ? ($volume_size = $options{volume_size})
      : ($volume_size = "30G");
    defined($options{aggr_name})
      ? ($aggr_name = $options{aggr_name})
      : ($aggr_name = get_aggr_name(filername=>$filername));
    defined($options{vserver_index})
      ? ($vserver_index = $options{vserver_index})
      : ($vserver_index = 1);
    defined($options{vserver_name})
      ? ($vserver_name = $options{vserver_name})
      : ($vserver_name =
         $Prefix . "_vs" . $vserver_index . "_" . $filername_with_underscores);
    defined($options{junction_path})
      ? ($junction_path = $options{junction_path})
      : ($junction_path = "/$volume_name".$filername_with_underscores);
    defined($options{space_guarantee})
      ? ($space_guarantee= $options{space_guarantee})
      : ($space_guarantee= "none");
    defined($options{state})
      ? ($state= $options{state})
      : ($state= "online");
    defined($options{volume_type})
      ? ($volume_type= $options{volume_type})
      : ($volume_type= "flex");
    defined($options{type})
      ? ($type= $options{type})
      : ($type= "RW");
    defined($options{snapshot_policy})
      ? ($snapshot_policy= $options{snapshot_policy})
      : ($snapshot_policy= "default");

    if(($volume_type eq "coral") || ($volume_type eq "acro")) {
      defined($options{aggr_name})
        ? ($aggr_name = $options{aggr_name})
        : ($aggr_name = "stripedaggr_0");
    }
    $stripe_format="full";

    if($volume_type eq "coral") {
	$stripe_format="data";
    }
    
    $server->set_execution_server(-Server => $filername);
    eval {
        $server->volume_show(vserver => $vserver_name,
                             volume  => $volume_name);
    };
    if ($@) {
        eval {
          if(($volume_type eq "coral") || ($volume_type eq "acro")) {
            $server->volume_new(vserver         => $vserver_name,
                                volume          => $volume_name,
                                aggregate       => $aggr_name,
                                "junction-path" => $junction_path,
                                state           => $state,
                                size               => $volume_size,
                                "security-style"   => "mixed",
        #                       foreground => "true",
                                "unix-permissions" => "0777",
                                "space-guarantee"  => $space_guarantee,
                                "maxdir-size"      => "4090M",
                                "stripe-format"    => $stripe_format,
                                "snapshot-policy"  => $snapshot_policy,
                                "type"             => $type);
         } else {
            $server->volume_new(vserver         => $vserver_name,
                                volume          => $volume_name,
                                aggregate       => $aggr_name,
                                "junction-path" => $junction_path,
                                state           => $state,
                                size               => $volume_size,
                                "security-style"   => "mixed",
        #                       foreground => "true",
                                "unix-permissions" => "0777",
                                "space-guarantee"  => $space_guarantee,
                                "maxdir-size"      => "4090M",
                                "snapshot-policy"  => $snapshot_policy,
                                "type"             => $type);
         }
        };
        if ($@) {
            $Log->trace("Creation of volume $volume_name for $filername FAILED",
                        $@);
            return 1;
        } else {
            $Log->trace(
                   "Creation of volume $volume_name for $filername succeeded.");
        }
    } else {
        $Log->trace(
            "VOLUME $volume_name on $filername already exists.......skipping creation..."
        );
    }
    return 0;
} ## end sub create_volume

########################################################################
# Function:    delete_all_volumes
#
# Description: Deletes all  volumes from a vserver/filer
#
# Parameters:  $server,filername
#
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub delete_all_volumes {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            vserver_name  => {type => SCALAR, optional => 1},
                            vserver_index => {type => SCALAR, optional => 1},});

    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my ($volume_name,$filername_with_underscores, $vserver_name, $vserver_index, @vols);
    $filername_with_underscores = $filername;
    $filername_with_underscores =~ s/-/_/g;
    defined($options{vserver_index})
      ? ($vserver_index = $options{vserver_index})
      : ($vserver_index = 1);
    defined($options{vserver_name})
      ? ($vserver_name = $options{vserver_name})
      : ($vserver_name =
         $Prefix . "_vs" . $vserver_index . "_" . $filername_with_underscores);
    my $root_volume = $vserver_name . "_root";

    $Log->trace("Delete all volumes");
    eval {
        @vols = $server->volume_show(vserver => $vserver_name);
    };
    if(!$@) {
       eval {
	  foreach my $vol_of (@vols){
               $volume_name = $vol_of->{volume};
               next if($volume_name eq $root_volume);
               die "Delete volume $volume_name FAILED" 
 		if(delete_volume(serverobj    => $server,
                        filername    => $filername,
                        volume       => $volume_name, 
                        vserver_name => $vserver_name));
	  }
          die "Delete root volume $root_volume FAILED"
                if(delete_volume(serverobj    => $server,
                        filername    => $filername,
                        volume       => $root_volume,
                        vserver_name => $vserver_name));

       };
       if ($@) {
           $Log->trace(
                "Deletion of volume $volume_name of vserver  $vserver_name FAILED:".
                $@);
           return 1;
       } else {
           $Log->trace(
               "Deletion of volume $volume_name of vserver  $vserver_name succeded"
           );
       }
    }
    return 0;
} ## end sub delete_all_volumes

########################################################################
# Function:    change_object_state
#
# Description: Change agrr/vol state to online/offline/resticed
#
# Parameters:  $server,filername,object_name,object_type,state
#
# Returns:     0 on Success
#	       1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub change_object_state {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            object_name        => {type => SCALAR},
                            object_type        => {type => SCALAR},
                            state         => {type => SCALAR},
                            vserver_name  => {type => SCALAR, optional => 1},
                            error_msg      => {type => SCALARREF, optional =>1},
                            vserver_index => {type => SCALAR, optional => 1},});

    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $object_name = $options{object_name};
    my $object_type = $options{object_type};
    my $state = $options{state};
    my $error_string = "";
    my $error_msg = $options{error_msg};

    my ($filername_with_underscores, $vserver_name, 
	$vserver_index);
    $filername_with_underscores = $filername;
    $filername_with_underscores =~ s/-/_/g;
    defined($options{vserver_index})
      ? ($vserver_index = $options{vserver_index})
      : ($vserver_index = 1);
    defined($options{vserver_name})
      ? ($vserver_name = $options{vserver_name})
      : ($vserver_name =
         $Prefix . "_vs" . $vserver_index . "_" . $filername_with_underscores);

    $Log->trace("Changing object state to $state");
    my @objs;
    eval {
	if($object_type eq "aggr") {
       	     @objs = $server->storage_aggregate_show(aggregate => $object_name);
        } else {
             @objs = $server->volume_show(vserver => $vserver_name,
                             volume  => $object_name);
        }
    };
    if(!$@) {
       if($objs[0]->{state} eq $state) {
	  return 0;
       }
       eval {
         if(($state eq "offline") && ($object_type eq "vol")) {
           #Unmount
           $server->volume_unmount(vserver => $vserver_name,
   			       foreground => "true",
                                "-Timeout" => "600",
                                  volume  => $object_name);
         }
         #change state
         if($object_type eq "vol") {
           $server->volume_modify(vserver => $vserver_name,
   		       foreground => "true",
               		state=>$state,
                        "-Timeout" => "600",
                        volume  => $object_name);
         } else {
           $server->storage_aggregate_modify(aggregate => $object_name,
   		       foreground => "true",
               		state=>$state,
                        "-Timeout" => "600");
         }
       };
       if ($@) {
           $Log->trace(
                "State change of object $object_name  $state FAILED:".
                $@);
           $error_string .= "$@";
           if(defined($error_msg)) {
             $$error_msg .= $error_string;
           }
           return 1;
       } else {
           $Log->trace(
                "State change of object $object_name  $state succeeded"
           );
       }
    }
    eval {
        if($object_type eq "aggr") {
             @objs = $server->storage_aggregate_show(aggregate => $object_name);
        } else {
             @objs = $server->volume_show(vserver => $vserver_name,
                             volume  => $object_name);
        }
    };

    return 0;
} ## end sub change_object_state

########################################################################
# Function:    delete_volume
#
# Description: Delete volume
#
# Parameters:  $server,filername
#
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub delete_volume {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            volume        => {type => SCALAR},
                            vserver_name  => {type => SCALAR, optional => 1},
                            vserver_index => {type => SCALAR, optional => 1},});

    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $volume_name = $options{volume};
    my ($filername_with_underscores, $vserver_name, $vserver_index);
    $filername_with_underscores = $filername;
    $filername_with_underscores =~ s/-/_/g;
    defined($options{vserver_index})
      ? ($vserver_index = $options{vserver_index})
      : ($vserver_index = 1);
    defined($options{vserver_name})
      ? ($vserver_name = $options{vserver_name})
      : ($vserver_name =
         $Prefix . "_vs" . $vserver_index . "_" . $filername_with_underscores);
    my @vols;

    $Log->trace("Delete volume");
    $server->set_execution_server(-Server => $filername);
    eval {
        @vols = $server->volume_show(vserver => $vserver_name,
                             volume  => $volume_name);
    };
    if(!$@) {
       eval {
          if(($vols[0]->{'state'} eq "online") || ($vols[0]->{'state'} eq "restricted")) {
              #Unmount, but dont unmount if its a vserver root
			  if( $vols[0]->{'vsroot'} ne "true" ) {
                       $server->volume_unmount(vserver => $vserver_name,
      			                   foreground => "true",
                                            "-Timeout" => "600",
                                             volume  => $volume_name);
	          }
              #Offline
              $server->volume_modify(vserver => $vserver_name,
      			       foreground => "true",
   				state=>"offline",
                                   "-Timeout" => "600",
                                     volume  => $volume_name);
           }
	   #Delete
           $server->volume_delete(vserver => $vserver_name,
   			       foreground => "true",
   			       'force'=>"true",
                                "-Timeout" => "600",
                                  volume  => $volume_name);
       };
       if ($@) {
           $Log->trace(
                "Deletion of volume $volume_name of vserver  $vserver_name FAILED:".
                $@);
           return 1;
       } else {
           $Log->trace(
               "Deletion of volume $volume_name of vserver  $vserver_name succeded"
           );
       }
    }
    return 0;
} ## end sub delete_volume

########################################################################
# Function:    get_root_object
#
# Description: Get the name of root Aggregate/Volume
#
# Parameters:  Server
#                   filer_name
#                   object_type - can be aggr or vol
#                   local_or_partner(optional)
#
# Returns:     Root Aggr/Vol on Success
#	               NULL on failure
#
# Notes:
#   - Global parameters referenced: $server
########################################################################
sub get_root_object {
    my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},
                            object_type      => {type => SCALAR},
                            local_or_partner => {type => SCALAR, optional => 1},
                           });

    my $server      = $options{serverobj};
    my $filer_name  = $options{filername};
    my $object_type = $options{object_type};
    my $local_or_partner;
    my $cmd_output,

      defined($local_or_partner)
      ? ($local_or_partner = $options{local_or_partner})
      : ($local_or_partner = "");

    #$server->set_execution_server(-Server => $filer_name);
    $cmd_output =
      $server->system_node_run(
                          node       => $filer_name,
                          '-Timeout' => 60,
                          command => "\"$local_or_partner $object_type status\""
      );

    my @lines = split('\n', $cmd_output);    # split into lines
    foreach my $line (@lines) {
        $line =~ s/^\s*//;
        $line =~ s/,\s*/,/g;
        my ($vol, $state, $status, $opts) = split(/\s+/, $line);

        if (defined($opts) && ($opts =~ /^root/)) {
            $Log->trace("Returning $object_type - " . $vol);
            return ($vol);
        }
    } ## end foreach my $line (@lines)
    $Log->trace("Returning $object_type - nothing");
    return "";
} ## end sub get_root_object

########################################################################
# Function:    create_aggregate
#
# Description: Create Aggregate
#
# Parameters:  Server,filer_name,aggr_name,aggr_size,aggr_type
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub create_aggregate {
    my %options = validate(@_,
                           {serverobj => {type => OBJECT},
                            filername => {type => SCALAR | ARRAYREF},
                            aggr_name => {type => SCALAR},
                            aggr_size => {type => SCALAR},
                            block_type => {type => SCALAR, optional=>1, default=>undef},
                            raid_type => {type => SCALAR},});
    my $server    = $options{serverobj};
    my $filernameparam = $options{filername};
    my $aggr_name = $options{aggr_name};
    my $aggr_size = $options{aggr_size};
    my $raid_type = $options{raid_type};
    my $block_type ;
    if (defined($options{block_type})) {
        $block_type = $options{block_type};
    }else{
        $block_type = undef;
    }
    
    my ($loopcount, $reply);
    my @filernames;
    if(ref($filernameparam) eq "ARRAY") {
      push(@filernames,@$filernameparam);
    } else {
      push(@filernames,$filernameparam);
    }
   
    #$server->set_execution_server(-Server => $filernames[0]);
    my @aggrs;
    eval { @aggrs = $server->storage_aggregate_show(aggregate => $aggr_name); };
    if ($@) {
        eval {
            if(defined ($block_type)){
                $server->storage_aggregate_new(aggregate => $aggr_name,
                                           diskcount => $aggr_size,
					   nodes     => join(',',@filernames),
					   'block-type' => $block_type,
                                           raidtype  => $raid_type);
            }else{
                $server->storage_aggregate_new(aggregate => $aggr_name,
                                           diskcount => $aggr_size,
					   nodes     => join(',',@filernames),
                                           raidtype  => $raid_type);    
            }
        };
        if ($@) {
            $loopcount = 5;
            while ($loopcount--) {
                $reply =
                  $server->storage_aggregate_show(aggregate => $aggr_name);
                if ($reply->[0]->{state} =~ /creating/i) {
                    $Log->trace("Verifying aggregate creation");
                    sleep 5;
                    next;
                } else {
                    $Log->trace("AGGREGATE $aggr_name successfully created...");
                    last;
                }
            } ## end while ($loopcount--)

            if ($loopcount <= 0) {
                $Log->trace("AGGREGATE $aggr_name creation FAILED");
                return 1;
            }
        } ## end if ($@)
    } else {
        $Log->trace(
              "AGGREGATE $aggr_name already exists.......skipping creation...");
       set_aggr_type( serverobj => $server,
                            aggr_name => $aggr_name,
                            aggr_type => "sfo");
	my  $diskcount = $aggrs[0]->{'diskcount'};
	if($diskcount < $aggr_size)
	{
		$diskcount = ($aggr_size - $diskcount);
		die "Aggr Add Disks FAILED"
			if (aggr_add_disks(serverobj => $server,
					  aggr_name => $aggr_name,
	                        	  disk_count => $diskcount));
	}


    }
    return 0;
} ## end sub create_aggregate

########################################################################
# Function:    delete_aggregate
#
# Description: Delete Aggregate
#
# Parameters:  Server,filer_name,aggr_name,aggr_size,aggr_type
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub delete_aggregate {
    my %options = validate(@_,
                           {serverobj => {type => OBJECT},
                            filername => {type => SCALAR},
                            aggr_name => {type => SCALAR},});
    my $server    = $options{serverobj};
    my $filername = $options{filername};
    my $aggr_name = $options{aggr_name};
    #$server->set_execution_server(-Server => $filername);
    eval { $server->storage_aggregate_show(aggregate => $aggr_name); };
    if(!$@) {
       if (change_object_state(serverobj=>$server,
          filername=>$filername,
          object_name=>$aggr_name,
          object_type=>"aggr",
          state=>"restricted")) {
         $Log->trace("Cannot offline the Aggregate");
         return 1;
       }

       eval { $server->storage_aggregate_delete(aggregate => $aggr_name); };
       if ($@) {
           $Log->trace("Deletion of AGGREGATE $aggr_name FAILED", $@);
           return 1;
       } else {
           $Log->trace("Deletion of AGGREGATE $aggr_name succeded");
       }
    }
    return 0;
} ## end sub delete_aggregate

########################################################################
# Function:    create_vserver
#
# Description: Create Vserver
#
# Parameters:  Server,filer_name,aggr_name,aggr_size,aggr_type
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub create_vserver {
    my %options = validate(@_,
                           {serverobj    => {type => OBJECT},
                            filername    => {type => SCALAR},
                            aggr_name    => {type => SCALAR},
                            vserver_name => {type => SCALAR},
                            root_volume  => {type => SCALAR},});
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $aggr_name    = $options{aggr_name};
    my $vserver_name = $options{vserver_name};
    my $root_volume  = $options{root_volume};

    #$server->set_execution_server(-Server => $filername);
    eval { $server->vserver_show(vserver => $vserver_name); };
    if ($@) {
        eval {
            $server->vserver_new(vserver                     => $vserver_name,
                                 rootvolume                  => $root_volume,
                                 aggregate                   => $aggr_name,
                                 'ns-switch'                 => "file",
                                 language                    => "en",
                                 'rootvolume-security-style' => "unix",
                                 -Timeout                    => 300,);
        };
        if ($@) {
            $Log->trace("VSERVER $vserver_name creation FAILED", $@);
            return 1;
        } else {
            $Log->trace("VSERVER $vserver_name successfully created.");
        }
    } else {
        $Log->trace(
             "VSERVER $vserver_name already exists.......skipping creation...");
    }
    return 0;
} ## end sub create_vserver

########################################################################
# Function:    create_vserver
#
# Description: Create Vserver
#
# Parameters:  Server,filer_name,vserver,port_index
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################

sub create_vif {
    my %options = validate(@_,
                           {serverobj    => {type => OBJECT},
                            filername    => {type => SCALAR},
                            vserver_name => {type => SCALAR},
                            port_index   => {type => SCALAR},
                            hostobj      => {type => OBJECT},});
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $vserver_name = $options{vserver_name};
    my $port_index   = $options{port_index};
    my $hostobj;
    defined($options{hostobj})
      ? ($hostobj = $options{hostobj})
      : ($hostobj = host($filername));

    #$server->set_execution_server(-Server => $filername);
    eval {
        $server->network_interface_show(server => $vserver_name,
                                        vif    => 'data' . $port_index);
    };
    if ($@) {
        eval {
            $server->network_interface_new(
                           server      => $vserver_name,
                           vif         => 'data' . $port_index,
                           'home-node' => $filername,
                           'home-port' => $hostobj->{"DATA_PORT" . $port_index},
                           address     => $hostobj->{"DATA_IP" . $port_index},
                           netmask     => $hostobj->{"DATA_MASK" . $port_index},
                           role        => 'data',
                           'failover-policy' => 'nextavail',
                           'status-admin'    => 'up',
                           'auto-revert'     => 'true',);
        };
        if ($@) {
            $Log->trace("Creation of VIF data1 for $filername FAILED", $@);
            return 1;
        } else {
            $Log->trace("Creation of VIF data1 for $filername succeeded...");
        }
    } else {
        $Log->trace(
            "VIF data1 for $filername already exists.......skipping creation...."
        );
    }
    return 0;
} ## end sub create_vif

########################################################################
# Function:    create_vserver_nfs
#
# Description: Create Vserver NFS
#
# Parameters:  Server,filer_name,vserver
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################

sub create_vserver_nfs {
    my %options = validate(@_,
                           {serverobj    => {type => OBJECT},
                            filername    => {type => SCALAR},
                            vserver_name => {type => SCALAR},});
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $vserver_name = $options{vserver_name};

    #$server->set_execution_server(-Server => $filername);
    eval { $server->protocol_nfs_server_show(vserver => $vserver_name); };
    if ($@) {
        eval {
            $server->protocols_nfs_server_new(vserver => $vserver_name,
                                              access  => "true",
                                              "rpcsec-ctx-high" => 30,
                                              "rpcsec-ctx-idle" => 30,
                                              "v3-enable"       => "true",
                                              "udp-enable"      => "true",
                                              "tcp-enable"      => "true",
                                              spinauth          => "true",);
        };
        if ($@) {
            $Log->trace("VServer NFS creation FAILED..." . $@);
            return 1;
        } else {
            $Log->trace("VServer NFS creation succeeded...");
        }
    } ## end if ($@)
    return 0;
} ## end sub create_vserver_nfs

########################################################################
# Function:    create_vserver_export_policy
#
# Description: Create Vserver Export Policy
#
# Parameters:  Server,filer_name,vserver
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub create_vserver_export_policy {
    my %options = validate(@_,
                           {serverobj    => {type => OBJECT},
                            filername    => {type => SCALAR},
                            vserver_name => {type => SCALAR},});
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $vserver_name = $options{vserver_name};
    my $clientmatch  = $IP_MASK;

    #$server->set_execution_server(-Server => $filername);
    eval {
        $server->protocol_exportruleset_rule_show(vserver    => $vserver_name,
                                                  policyname => "default",
                                                  ruleindex  => "1",);
    };
    if ($@) {
        eval {
            $server->protocol_exportruleset_rule_new(vserver => $vserver_name,
                                                     policyname  => "default",
                                                     ruleindex   => "1",
                                                     protocol    => "any",
                                                     clientmatch => $clientmatch,
                                                     rorule      => "any",
                                                     rwrule      => "any",
                                                     anon        => "65534",
                                                     superuser   => "any");
        };
        if ($@) {
            $Log->trace("VServer export policy creation FAILED...", $@);
            return 1;
        } else {
            $Log->trace("VServer export policy creation succeeded...");
        }
    } ## end if ($@)
    return 0;
} ## end sub create_vserver_export_policy

########################################################################
# Function:    create_route
#
# Description: Create Route
#
# Parameters:  Server,filer_name,vserver,port_index
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub create_route {
    my %options = validate(@_,
                           {serverobj    => {type => OBJECT},
                            filername    => {type => SCALAR},
                            vserver_name => {type => SCALAR},
                            port_index   => {type => SCALAR, optional => 1},});
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $vserver_name = $options{vserver_name};
    my $port_index;
    my $destination=$IP;
    my $route_mask = $IP;
    my $vifcfg = new VifMgr::VifConfig();
    defined($options{port_index})
      ? ($port_index = $options{port_index})
      : ($port_index = 1);
    my $hostobj = host($filername);

    #$server->set_execution_server(-Server => $filername);
    eval {
        $vifcfg->create_route(
                            server  => $vserver_name,
                            vif     => 'data' . $port_index,
                            gateway => $hostobj->{"DATA_GATEWAY" . $port_index},
                            destination => $destination,
                            route_mask  => $route_mask,);
    };

    if ($@) {
        $Log->trace("VServer route creation FAILED...", $@);
        return 1;
    } else {
        $Log->trace("VServer route creation succeded...");
    }
    return 0;
} ## end sub create_route

########################################################################
# Function:    create_failover_rules
#
# Description: Create Failover Rules
#
# Parameters:  Server,filer1,filer2,vserver,port_index
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub create_failover_rules {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filer1        => {type => SCALAR},
                            filer2        => {type => SCALAR},
                            vserver_name  => {type => SCALAR, optional => 1},
                            port_index    => {type => SCALAR, optional => 1},
                            vserver_index => {type => SCALAR, optional => 1},});
    my $server                     = $options{serverobj};
    my $filer1                     = $options{filer1};
    my $filer2                     = $options{filer2};
    my $filername_with_underscores = $filer1;
    $filername_with_underscores =~ s/-/_/g;
    my $hostobj      = host($filer2);
    my $vserver_name = $options{vserver_name};
    my ($port_index, $vserver_index);
    defined($options{port_index})
      ? ($port_index = $options{port_index})
      : ($port_index = 1);
    defined($options{vserver_index})
      ? ($vserver_index = $options{vserver_index})
      : ($vserver_index = 1);
    defined($options{vserver_name})
      ? ($vserver_name = $options{vserver_name})
      : ($vserver_name =
         $Prefix . "_vs" . $vserver_index . "_" . $filername_with_underscores);

    my @failover_rules =  $server->network_interfaces_failover_show(
                                        server => $vserver_name,
                                        vif =>'data1');
    foreach my $failover_rule(@failover_rules) {
      if($failover_rule->{priority} == 0) {
        next;
      }
      $server->network_interfaces_failover_delete(
                                server   => $vserver_name,
                                vif      => 'data1',
                                priority => $failover_rule->{priority});
    }
    eval {
        $server->network_interfaces_failover_new(
                                   server   => $vserver_name,
                                   vif      => 'data' . $port_index,
                                   priority => "1",
                                   filer    => $filer2,
                                   port => $hostobj->{"DATA_PORT" . $port_index}
        );
    };
    if ($@) {
        $Log->trace("Network Interface failover creation for $filer1 FAILED...",
                    $@);
        return 1;
    } else {
        $Log->trace(
                 "Network Interface failover creation for $filer1 succeded...");
    }
    return 0;
} ## end sub create_failover_rules

########################################################################
# Function:    delete_route
#
# Description: Delete Route
#
# Parameters:  Server,filername,vserver,port_index
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub delete_route {
    my %options = validate(
        @_,
        {serverobj    => {type => OBJECT},
         filername    => {type => SCALAR},
         vserver_name => {type => SCALAR},
         port_index   => {type => SCALAR, optional => 1},

        });
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $vserver_name = $options{vserver_name};
    my $port_index;
    my $destination=$IP;
    my $group_mask = $IP;

    defined($options{port_index})
      ? ($port_index = $options{port_index})
      : ($port_index = 1);
    my $vifcfg = new VifMgr::VifConfig();

    #$server->set_execution_server(-Server => $filername);
    my @show;
    eval {
     @show = $server->network_interface_show(
            server => $vserver_name,
            vif         => 'data' . $port_index
    );
    };
    if(!$@) {
       my $group = $show[0]->{'routing-group'};
       # Delete only if it already exist
       eval {
               $server->network_routing_groups_route_show(
                   server          => $vserver_name,
                   'routing-group' => $group,
                   destination     => "$destination/$group_mask"
               )
       };
       if(!$@) {
          eval {
              $vifcfg->delete_route(server      => $vserver_name,
                                    vif         => 'data' . $port_index,
                                    destination => $destination,
                                    route_mask  => $group_mask,);
          };
          if ($@) {
              $Log->trace(
                  "Deletion of route for data$port_index of vserver  $vserver_name FAILED",
                  $@);
              return 1;
          } else {
              $Log->trace(
                  "Deletion of route for data$port_index of vserver  $vserver_name succeded"
              );
          }
       }
    }
    return 0;
} ## end sub delete_route

########################################################################
# Function:    delete_vif
#
# Description: Delete Vif
#
# Parameters:  Server,filername,vserver,port_index
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub delete_vif {
    my %options = validate(
        @_,
        {serverobj    => {type => OBJECT},
         filername    => {type => SCALAR},
         vserver_name => {type => SCALAR},
         port_index   => {type => SCALAR, optional => 1},

        });
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $vserver_name = $options{vserver_name};
    my $port_index;
    defined($options{port_index})
      ? ($port_index = $options{port_index})
      : ($port_index = 1);

    #$server->set_execution_server(-Server => $filername);
    eval {
        $server->network_interface_show(server => $vserver_name,
                                          vif    => 'data' . $port_index);
    };
    if(!$@) {
       eval {
           $server->network_interface_delete(server => $vserver_name,
                                             vif    => 'data' . $port_index);
       };
       if ($@) {
           $Log->trace(
               "Deletion of network interface for data$port_index of vserver  $vserver_name FAILED",
               $@);
           return 1;
       } else {
           $Log->trace(
               "Deletion of network interface for data$port_index of vserver  $vserver_name succeded"
           );
       }
    }
    return 0;
} ## end sub delete_vif

########################################################################
# Function:    delete_routing_group
#
# Description: Delete Routing Group
#
# Parameters:  Server,filername,vserver
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub delete_routing_group {
    my %options = validate(@_,
                           {serverobj    => {type => OBJECT},
                            filername    => {type => SCALAR},
                            vserver_name => {type => SCALAR},});
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $vserver_name = $options{vserver_name};
    my $vifcfg       = new VifMgr::VifConfig();

    #$server->set_execution_server(-Server => $filername);
    eval {
	$server->network_routing_groups_show(server => $vserver_name);
    };
    if(!$@) {
       eval { $vifcfg->delete_routing_group(server => $vserver_name); };
       if ($@) {
           $Log->trace(
                      "Deletion of routing group of vserver  $vserver_name FAILED",
                      $@);
           return 1;
       } else {
           $Log->trace(
                   "Deletion of routing group of vserver  $vserver_name succeded");
       }
    }
    return 0;
} ## end sub delete_routing_group

########################################################################
# Function:    delete_export_ruleset
#
# Description: Delete Export Ruleset
#
# Parameters:  Server,filername,vserver
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub delete_export_ruleset {
    my %options = validate(@_,
                           {serverobj    => {type => OBJECT},
                            filername    => {type => SCALAR},
                            vserver_name => {type => SCALAR},});
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $vserver_name = $options{vserver_name};

    #$server->set_execution_server(-Server => $filername);
    eval {
        $server->protocol_exportruleset_rule_show(vserver    => $vserver_name,
                                                  policyname => "default",
                                                  ruleindex  => "1",);
    };
    if(!$@) {
       eval {
           $server->protocol_exportruleset_rule_delete(vserver    => $vserver_name,
                                                       policyname => "default",
                                                       ruleindex  => "1");
       };
       if ($@) {
           $Log->trace(
                     "Deletion of export ruleset of vserver  $vserver_name FAILED",
                     $@);
           return 1;
       } else {
           $Log->trace(
                  "Deletion of export ruleset of vserver  $vserver_name succeded");
       }
    }
    return 0;
} ## end sub delete_export_ruleset

########################################################################
# Function:    delete_vserver_nfs
#
# Description: Delete Vserver NFS
#
# Parameters:  Server,filername,vserver
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub delete_vserver_nfs {
    my %options = validate(@_,
                           {serverobj    => {type => OBJECT},
                            filername    => {type => SCALAR},
                            vserver_name => {type => SCALAR},});
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $vserver_name = $options{vserver_name};

    #$server->set_execution_server(-Server => $filername);
    eval { $server->protocols_nfs_server_show(vserver => $vserver_name); };
    if(!$@) {
       eval { $server->protocols_nfs_server_delete(vserver => $vserver_name); };
       if ($@) {
           $Log->trace("Deletion of NFS of vserver  $vserver_name FAILED", $@);
           return 1;
       } else {
           $Log->trace("Deletion of NFS of vserver  $vserver_name succeded");
       }
    }
    return 0;
} ## end sub delete_vserver_nfs

########################################################################
# Function:    delete_vserver
#
# Description: Delete Vserver
#
# Parameters:  Server,filername,vserver
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub delete_vserver {
    my %options = validate(@_,
                           {serverobj    => {type => OBJECT},
                            filername    => {type => SCALAR},
                            vserver_name => {type => SCALAR},});
    my $server       = $options{serverobj};
    my $filername    = $options{filername};
    my $vserver_name = $options{vserver_name};

    #$server->set_execution_server(-Server => $filername);
    eval { $server->vserver_show('vserver' => $vserver_name); };
    if(!$@) {
       eval { $server->vserver_delete('vserver' => $vserver_name); };
       if ($@) {
           $Log->trace("Deletion of VSERVER  $vserver_name FAILED", $@);
           return 1;
       } 
    }
    $Log->trace("Deletion of VSERVER  $vserver_name succeded");
    return 0;
} ## end sub delete_vserver

########################################################################
# Function:    start_traffic
#
# Description: Starts the Traffic on the specified junction path
#
# Parameters:  Junction Path (should be of the form dataip1:/junctionpath)
#	       Mount Path
#	       Client Name
#
# Returns:     Handler to the hammer subtest
#	               Handler to the mountObject(which can be used in the cleanup function)
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub start_traffic {
    my %options = validate(@_,
                           {junction_path => {type => SCALAR},
                            mount_path    => {type => SCALAR},
                            extra_path    => {type => SCALAR,optional=>1},
                            client_name   => {type => SCALAR},
                            dataset_size   => {type => SCALAR,optional=>1},
                            mount_only => {type => SCALAR,optional=>1},
                            traffic_only => {type => SCALAR,optional=>1, default=>0},
                            hammer_dir => {type => SCALAR,optional=>1, default=>"hammer"},
                            mount_options => {type => SCALAR,optional=>1},
                            log_file      => {type => SCALAR,optional=>1},
                            sudo_path     => {type => SCALAR}});
    my $junction_path = $options{junction_path};
    my $mount_path    = $options{mount_path};
    my $client_name   = $options{client_name};
    my $sudo_path     = $options{sudo_path};
    my $traffic_only = $options{traffic_only};
    my $hammer_dir = $options{hammer_dir};
    my $dataset=0;
    my $mount_only=0;
    my ($extra_path, $traffic_dir, $mount_options, $log_file,$mntobj);
    defined($options{dataset_size})
      ? ($dataset = $options{dataset_size})
      : ($dataset= 0);

    defined($options{extra_path})
      ? ($extra_path= $options{extra_path})
      : ($extra_path = "");

    defined($options{mount_only})
      ? ($mount_only= $options{mount_only})
      : ($mount_only= 0);

    defined($options{mount_options})
      ? ($mount_options= $options{mount_options})
      : ($mount_options= "rw");

    defined($options{log_file})
      ? ($log_file = $options{log_file})
      : ($log_file = "hamlog.txt");

    my $client = new Client(-host     => $client_name,
                            -timeout  => 100,
			    -recreate_stale=>"false",
                            -protocol => "nfs");
    if ($client->error) {
        $Log->trace($client->errmsg);
	die "Problem in connecting to $client_name";
    }
    $Log->trace("Junctions Path: $junction_path");
    $Log->trace("Mount Path : $mount_path");
if(!$traffic_only)
{
    # Do a blind unmount
    eval {
        $client->execute("$sudo_path umount -f  $mount_path");
    };
    #dont worry about the error message
    #just a measure to unmount if it it already mounted
    # NFS share to mount
    $mntobj = $client->create_mnt(-share => $junction_path,
                                     -path  => $mount_path,
				     "-options" => $mount_options);
    if ($mntobj->error) {
        $Log->trace("MOUNT ERROR :". $mntobj->errmsg());
	die "Problem in mounting $junction_path to $mount_path";
    }
}
    $traffic_dir = "$mount_path/$extra_path/$hammer_dir";
    $client->execute("$sudo_path mkdir -p $traffic_dir");
    $client->execute("$sudo_path chmod 0777 $traffic_dir");
    my $handle;
    if(!$mount_only) {
       eval {
           $handle = subtest($sudo_path,     '-client',
                             $client_name,   '-bg',
   			  '-onexit', sub {
   				   	   my ($sec,$min,$hour,$mday,$mon,$year,$wday,
   					   $yday,$isdst)=localtime(time);
   					   my $str = sprintf "%4d-%02d-%02d %02d:%02d:%02d\n", $year+1900,$mon+1,$mday,$hour,$min,$sec;
   					   $Log->trace("PRAS SUBTEST ONEXIT CALLED");
   					   $Log->trace("TIME WAS $str");
   
   			  },
   			  '-ifscript', '-ok',
   			  '-iffatal', '-ok',
                             '--',           "/usr/software/test/bin/hammer",
                             "$traffic_dir", 
                             "-nofilercore", 
		#		"-blocksize",
                       #      "random",
   		#	  "-mode", "creates", 
   			  "-logfile", "../"."$log_file", 
   	#		  "-maxfsize", "6000",
           #			  "-fill"
	);
       };
       if ($@) {
           $Log->trace("Failure in starting HAMMER : " . $@);
           die "Failure in starting HAMMER : " . $@;
       } else {
           $Log->trace("Starting HAMMER SUCCESSFULLY...");
       }
    }
    if($dataset) {
      eval {
            create_set(clients => [$client_name],
                  base_name => "foo",
                  size => $dataset,
                  start => 0,
                  count => 100,
                  testdir => $mount_path);
      };
      if($@) {
        $Log->trace("Some problem in creating dataset : ".$@);
      }
    }
    
    return ($handle, $mntobj);
} ## end sub start_traffic

########################################################################
# Function:   verify_traffic
#
# Description: Verifies whether the traffic is running
#
# Parameters:  Single or Array of Mount Objects
#
# Returns:     0 on success
#	       1 on failure
#
# Notes:
#
########################################################################
sub verify_traffic {
    my %options = validate(@_,
                           { mountobjs => {type => SCALAR | ARRAYREF},
                             hamlog_file => {type => SCALAR, optional=>1, default=>"hamlog.txt"},
			                 error_msg => {type => SCALARREF, optional =>1},
                             time_out=> {type => SCALAR, optional=>1, default=>60},
                             time_interval => {type => SCALAR, optional=>1, default=>5},

			   });
    my $mount_objs_param = $options{mountobjs};
    my $hamlog_file = $options{hamlog_file};
    my $error_string = "";
    my $error_msg = $options{error_msg};
    my $time_out = $options{time_out};
    my $time_interval = $options{time_interval};

    my ($status, @output,$cmd, $prev_count,
	$cur_count, $end_time, @mount_objs) ;

    if(ref($mount_objs_param) eq "ARRAY") {
      push(@mount_objs,@$mount_objs_param);
    } else {
      push(@mount_objs,$mount_objs_param);
    }
    foreach my $mnt(@mount_objs) {
      $end_time = time()+ $time_out;
      $cmd="wc -l ".$mnt->{path}."/hammer/$hamlog_file | awk '{print \$1}'";
      do {
	if($mnt->is_mounted) {
            $cur_count=$prev_count=0;
            ($status, @output) = &ClientUtil::execute($mnt->{conn},$cmd);
            $prev_count=$output[0];
	    sleep($time_interval);
            ($status, @output) = &ClientUtil::execute($mnt->{conn},$cmd);
            $cur_count=$output[0];
        $Log->comment("PREV COUNT = $prev_count ---- CUR COUNT = $cur_count");
	}
      } while ((time() < $end_time) && ($cur_count <= $prev_count));
      if($cur_count <= $prev_count) {
        $error_string .= "Verify traffic failed for : ".$mnt->{path}."\n";
        if(defined($error_msg)) {
           $$error_msg .= $error_string;
        }
	$Log->trace("Verify traffic failed for : ".$mnt->{path});
	return 1;
      } 
    }
    return 0;

}
########################################################################
# Function:   verify_storage_failover_status
#
# Description: Verify storage failover status
#
# Parameters:  server, filername
#
# Returns:     0 on success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub verify_storage_failover_status {
    my %options = validate(@_,
                           {serverobj => {type => OBJECT},
                            filername => {type => SCALAR},});
    my $server    = $options{serverobj};
    my $filername = $options{filername};

    my $found          = 0;
    my $end_time = time()+300;
    my $failovershow;

    my $response;
    my @parsed_output;
    my $state_value;
    my $enabled_value;
    my $field_name="state,enabled";
    
  
    my $api_set_obj;
    if(!defined($ApiSetObjs{$filername})
	      || !$ApiSetObjs{$filername}->api_is_session_alive()) {
        my $hostobj=host($filername);
	$ApiSetObjs{$filername} = NACL::APISet->new(hostobj=>$hostobj,
                                     category  => "Node",
                                     interface => "CLI",
				     set =>  "CMode",
                                    );
    }

    $api_set_obj = $ApiSetObjs{$filername};

    $server->set_execution_server(-Server => $filername);
    while(time() < $end_time) {
        $response = $api_set_obj->storage_failover_show('fields' => $field_name);
        @parsed_output = @{$response->get_parsed_output()};
        $found=0; 
        for (my $i = 0 ; $i < scalar(@parsed_output) ; $i++) {
          $state_value = $parsed_output[$i]->{state};
          $enabled_value = $parsed_output[$i]->{enabled};
          $Log->comment("State = $state_value, Enabled = $enabled_value");
          $found++
             if($state_value eq "connected");
        }
        if ($found == scalar(@parsed_output) ) {
            return 0;
        }
        sleep(3);
    }
    return 1;
} ## end sub verify_storage_failover_status

########################################################################
# Function:    enable_storage_failover_status
#
# Description: If by chance the test has failed and the storage failover stauts is
#                   giveback_partial then we need to issue a give back so that the other tests
#                   can run without any problem , also set the enabled flag to true
#
# Parameters:  server,filername
#
# Returns:     0 on success
#                 1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub enable_storage_failover_status {
    my %options    = @_;
    my $server     = $options{serverobj};
    my $filername  = $options{filername};
    my $fail_count = 0;
    if (get_storage_failover_field(serverobj  => $server,
                                   filername  => $filername,
				   wait_before_check=>2,
                                   field_name => "enabled") eq "false"
      ) {
        $Log->trace("Seting takeover Enabled flag to true");
        eval {
            $server->storage_failover_modify(node    => $filername,
                                             enabled => "true");
        };
        if ($@) {
            $Log->trace("FAILED TO modify the enabled flag " . $@);
            $fail_count++;
        }
    } ## end if (get_storage_failover_field...
    ###########################################
    # If by chance the test has failed and the storage failover stauts is
    # giveback_partial then we need to issue a give back so that the other tests
    # can run without any problem
    ###########################################
    if (get_storage_failover_field(serverobj  => $server,
                                   filername  => $filername,
				   wait_before_check=>2,
                                   field_name => "state"
        ) =~ /giveback_partial/
      ) {
        $Log->trace("Issuing a giveback");
        if (giveback_from_node(serverobj => $server,
                               filername => $filername)
          ) {
            $Log->trace("FAILED to giveback ");
            $fail_count++;
        }
    } ## end if (get_storage_failover_field...
    ($fail_count > 0) ? (return 1) : (return 0);
} ## end sub enable_storage_failover_status

########################################################################
# Function:    get_storage_failover_field
#
# Description: Return the corresponding field-value from the storage failover status command
#
# Parameters:  server,filer_name, field
#
# Returns:     field-value  - eg: connected, giveback_partial etc...
#
# Notes:
#   - Global parameters referenced: None
########################################################################

sub get_storage_failover_field {
    my %options = validate(@_,
                           {serverobj  => {type => OBJECT},
                            filername  => {type => SCALAR},
                            field_name => {type => SCALAR},
                            cf_status => {type => SCALAR, optional => 1},
			    wait_before_check => {type => SCALAR, optional => 1}});

    my $server     = $options{serverobj};
    my $filername  = $options{filername};
    my $field_name = $options{field_name};
    my ($wait_before_check, $cf_status); 
    defined($options{wait_before_check})
     ? ($wait_before_check = $options{wait_before_check})
     : ($wait_before_check = 120);
    defined($options{cf_status})
     ? ($cf_status= $options{cf_status})
     : ($cf_status= 1);

    ##################################################
    #Just wait for some time before checking the values
    ##################################################
    sleep($wait_before_check);
    $server->set_execution_server(-Server => $filername);
    # The following eval with cf status helps in debugging 
    if($cf_status) {
      eval {
         $server->system_node_run(
                      node       => $filername,
                      '-Timeout' => 60,
                      command => "\"priv set test;cf status\"");
      };
    }
    my $api_set_obj;
    if(!defined($ApiSetObjs{$filername})
	      || !$ApiSetObjs{$filername}->api_is_session_alive()) {
        my $hostobj=host($filername);
	$ApiSetObjs{$filername} = NACL::APISet->new(hostobj=>$hostobj,
                                     category  => "Node",
                                     interface => "CLI",
				     set =>  "CMode",
                                    );
    }

    $api_set_obj = $ApiSetObjs{$filername};
    
    my $response = $api_set_obj->storage_failover_show('fields' => $field_name, 'node' => $filername);
    my @parsed_output = @{$response->get_parsed_output()};
    my $field_value = $parsed_output[0]->{$field_name};
    $Log->comment("$field_name = $field_value");
    return $field_value;
} ## end sub get_storage_failover_field

########################################################################
# Function:    giveback_from_node
#
# Description: Issue a storage failover giveback
#
# Parameters:  server,filername
#
# Returns:     0 on success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################

sub giveback_from_node {
    my %options = validate(@_,
                           {serverobj => {type => OBJECT},
                            filername => {type => SCALAR},});
    my $server    = $options{serverobj};
    my $filername = $options{filername};
    my $iteration=0;

    $server->set_execution_server(-Server => $filername);
    eval {
        $server->storage_failover_giveback(fromnode   => $filername,
					   "override-vetoes" => "true",
                                           foreground => "true");
    };
    if ($@) {
	if($@ =~ /send-home is already in progress/) {
       	   $Log->trace("Looks like sendhome is already in progress..lets wait...");
  	   while(1) {
             if (get_storage_failover_field(serverobj  => $server,
                                   filername  => $filername,
                                   field_name => "state"
             ) =~  "giveback_partial"
             ) {
		 sleep 10;
		 if($iteration++ > 10) {
        	   $Log->trace("Timed out waiting for giveback partial to become connected");
		   last;
		 }
	       } else {
		 return 0;
	       }
	   }
	}
        $Log->trace("Some problem with storage failover giveback " . $@);
        return 1;
    } else {
        $Log->trace("Successfully issued storage faiover giveback");
        return 0;
    }
} ## end sub giveback_from_node

########################################################################
# Function:    up_both_nodes
#
# Description: Bring up both Nodes
#
# Parameters:  server,filer1,filer2
#
# Returns:     0 on Success
#	               1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################

sub up_both_nodes {
    my %options = validate(@_,
                           {serverobj => {type => OBJECT},
                            filer1    => {type => SCALAR},
                            filer2    => {type => SCALAR,optional => 1},});
    my $server     = $options{serverobj};
    my $filer1     = $options{filer1};
    my $filer2     = $options{filer2};
    my $fail_count = 0;

    $Log->trace("Bring up node1");
    $fail_count++ 
	if (transit_node_to_cli(serverobj => $server,
		filername => $filer1));
    if(defined($filer2)) {
       $Log->trace("Bring up node2");
       $fail_count++ 
	 if (transit_node_to_cli(serverobj=>$server, 
		filername=>$filer2));
    }
    ###########################################
    # If by chance the test has failed and the storage failover stauts is
    #  giveback_partial then we need to issue a give back so that the other tests
    #  can run without any problem , also set the enabled flag to true
    ###########################################
    if(defined($filer2)) {
      $fail_count++
        if (enable_storage_failover_status(serverobj => $server,
                                         filername => $filer1));
    }
    ($fail_count > 0) ? (return 1) : (return 0);
} ## end sub up_both_nodes

########################################################################
# Function:    set_aggr_type
#
# Description: Set Aggr type to sfo/cfo
#
# Parameters:  server,filername,aggr_name,aggr_type
#
# Returns:     0 on Success
#	       1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################

sub set_aggr_type {
    my %options = validate(@_,
                           {serverobj => {type => OBJECT},
                            filername => {type => SCALAR, optional => 1},
                            aggr_name => {type => SCALAR},
                            foreground => {type => SCALAR, optional => 1, default=> "true"},
                            aggr_type => {type => SCALAR},
                            error_msg => {type => SCALARREF, optional =>1},
			   });
    my $server    = $options{serverobj};
    my $filername = $options{filername};
    my $aggr_name = $options{aggr_name};
    my $aggr_type = $options{aggr_type};
    my $foreground = $options{foreground};
    my $error_string = "";
    my $error_msg = $options{error_msg};
    eval {
        $server->set_execution_server(-Server => $filername)
	  if(defined($filername));
        $server->storage_aggregate_modify(aggregate   => $aggr_name,
                                          "ha-policy" => $aggr_type,
					  foreground=>$foreground);
    };
    if ($@) {
	print "PRAS: Aggre modify to $aggr_type FAILED " . $@. "\n";
        $Log->trace("Aggre modify to $aggr_type FAILED " . $@);
        $error_string .= "Aggre modify to $aggr_type FAILED : $@";
        if(defined($error_msg)) {
           $$error_msg .= $error_string;
        }
        return 1;
    }
    return 0;
} ## end sub set_aggr_type

########################################################################
# Function:    get_ha_state 
#
# Description: Return the ha state from cluster ha show
#
# Parameters:  server,filer
#
# Returns:     Returns ha state
#
########################################################################

sub get_ha_state
{
  my %options = validate(@_,
 		    {serverobj     => {type => OBJECT},
		    filer     => {type => SCALAR},});
  my $hash;
  my $server= $options{serverobj};
  my $filer = $options{filer};

  $server->set_execution_server(-Server => $filer);
  $hash = $server->cluster_ha_show();
  return $hash->[0]{configured};
}

########################################################################
# Function:    get_partner_node
#
# Description: Return the name of the partner node
#
# Parameters:  server,filer
#
# Returns:     the name of the partner node
#
########################################################################

sub get_partner_node
{
  my %options = validate(@_,
 		    {serverobj     => {type => OBJECT},
		    filer     => {type => SCALAR},});
  my ($show, $node, $enabled);
  my $server= $options{serverobj};
  my $filer = $options{filer};

  $server->set_execution_server(-Server => $filer);
  $show = $server->storage_failover_show(-Timeout => 100, 
			-SkipINFO => 1);
  for (my $i = 0 ; $i <= (scalar(@{$show}) - 1) ; $i++) {
    $node=$show->[$i]{node};
    $enabled = $show->[$i]{enabled};
    return $node
     if($node ne $filer);
     #if(($node ne $filer) && ($enabled eq "true"));
  }
  return undef;
}

########################################################################
# Function:    enable_ha
#
# Description: Enable HA
#
# Parameters:  server,filer1,filer2
#
# Returns:     0 on Success
#	       1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################

sub enable_ha
{
  my %options = validate(@_,
 		    {serverobj     => {type => OBJECT},
		    filer1     => {type => SCALAR},
		    filer2     => {type => SCALAR, optional => 1}});
  my $hash;
  my $server    = $options{serverobj};
  my $filer1 = $options{filer1};
  my $filer2 = $options{filer2};
  my $fail_count=0;
  my @filers;
  push(@filers,$filer1);
  push(@filers,$filer2)
    if(defined($filer2));
  foreach my $filer(@filers) {
    $server->set_execution_server(-Server => $filer);
    $hash = $server->cluster_ha_show();
    if ($hash->[0]{configured} eq "false") {
      $Log->trace("Enabling HA");
      eval {
        $server->cluster_ha_modify(configured => "true",
                                foreground => "true");
      };
      if($@) {
        $Log->trace("Enabling HA FAILED for $filer ".$@);
        $fail_count++;
      }
    }
    eval {
      enable_storage_failover_status(serverobj => $server,
                                         filername => $filer);
    };
  }
  if ($fail_count) {
      return 1;
  } else {
      return 0;
  }
}

########################################################################
# Function:    disable_ha
#
# Description: Disable HA
#
# Parameters:  server,filer1,filer2
#
# Returns:     0 on Success
#	       1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################

sub disable_ha
{
  my %options = validate(@_,
 		    {serverobj     => {type => OBJECT},
		    filer1     => {type => SCALAR},
		    filer2     => {type => SCALAR, optional => 1}});
  my $hash;
  my $server    = $options{serverobj};
  my $filer1 = $options{filer1};
  my $filer2 = $options{filer2};
  my $fail_count=0;
  my @filers;
 
  push(@filers,$filer1);
  push(@filers,$filer2)
    if(defined($filer2));
  foreach my $filer(@filers) {
    $server->set_execution_server(-Server => $filer);
    $hash = $server->cluster_ha_show();
    if ($hash->[0]{configured} eq "true") {
      $Log->trace("Disabling HA");
      eval {
        $server->cluster_ha_modify(configured => "false",
                                foreground => "true");
      };
      if($@) {
        $Log->trace("Disabling HA FAILED for $filer ".$@);
        $fail_count++;
      }
    }
  }
  if ($fail_count) {
      return 1;
  } else {
      return 0;
  }
}
########################################################################
# Function:    disable_gdb
#
# Description: Disable GDB from the Loader prompt
#
# Parameters:  None
#
# Returns:     0 on Success
#	               1 on failure
#
########################################################################

sub disable_gdb {
  my %options = validate(@_,
 		    {serverobj     => {type => OBJECT},
		    filer1     => {type => SCALAR},
		    filer2     => {type => SCALAR, optional => 1},
		    autoboot   => {type => SCALAR, optional => 1},
		    sfo        => {type => SCALAR, optional => 1},
			});
  my $hash;
  my $server    = $options{serverobj};
  my $filer1 = $options{filer1};
  my $filer2 = $options{filer2};
  my $fail_count=0;
  my ($host,$gdb_disable,$cur_auto_boot);
  my $reboot_required=0;
  my (@filers, $auto_boot);
  
  push(@filers,$filer1);
  push(@filers,$filer2)
    if(defined($filer2));

  defined($options{autoboot})
   ? ($auto_boot = $options{autoboot})
   : ($auto_boot = "false");

  my $sfo = $options{sfo};

  my @setup_subtests;
  my $st;
  foreach my $filer(@filers) {
    $server->set_execution_server(-Server => $filer);
    $server->handle_connections( close => [qw(cli_connection)] );
  }
  foreach my $filer(@filers) {
    $st = main::subtest(
      sub {
        $server->handle_connections( open => [qw(cli_connection)] );
        $server->set_execution_server(-Server => $filer);
        $host =
          $server->shell(cmd => "/bin/hostname", noprompt => 0, -Errors => 0);
        $gdb_disable = $server->shell(cmd      => "/bin/kenv 'gdb-on-panic?'",
                                         noprompt => 0,
                                         -Errors  => 0);
        if ($gdb_disable ne "false") {
            $Log->trace("GDB on $host is true, set to false");
            $server->shell(cmd      => "sudo sysctl 'kern.bootargs=gdb-on-panic?=false'",
                           noprompt => 0,
                           -Errors  => 0);
            sleep 5;
            $gdb_disable = $server->shell(cmd      => "/bin/kenv 'gdb-on-panic?'",
                                          noprompt => 0,
                                          -Errors  => 0);
            if ($gdb_disable ne "false") {
                $Log->trace("Failed to set GDB to false");
	        $fail_count++;
            } else {
                $Log->trace("Succesfully set GDB to false");
                $Log->comment("Setting reboot flag since we alter gdb flag");
	        $reboot_required=1;
            }
            $server->shell(cmd      => "sudo sysctl 'kern.bootargs=kcs.clam.dont_to_if_partner_ooq?=true'",
                           noprompt => 0,
                           -Errors  => 0);
            sleep 5;
            $server->shell(cmd      => "/bin/kenv 'kcs.clam.dont_to_if_partner_ooq?'",
                                          noprompt => 0,
                                          -Errors  => 0);

        } else {
            $Log->trace("GDB on $host is '$gdb_disable'");
        }
        #Also set AUTOBOOT to false
        $cur_auto_boot = $server->shell(cmd      => "/bin/kenv 'AUTOBOOT'",
                                         noprompt => 0,
                                         -Errors  => 0);
        if ($cur_auto_boot ne $auto_boot) {
            $Log->trace("AUTOBOOT on $host is $cur_auto_boot , set to $auto_boot");
        	$server->shell(cmd      => "sudo sysctl 'kern.bootargs=AUTOBOOT=$auto_boot'",
                   noprompt => 0,
                   -Errors  => 0);
	   $reboot_required=1;
           $Log->comment("Setting reboot flag since we altered AUTOBOOT flag");
        }
        #set auto-giveback to true as a safetey measure
        eval {
          $server->storage_failover_modify(node=>$filer,'auto-giveback'=>'true');
        };
        if(defined($sfo)) {
          eval {
            $server->storage_failover_modify(node=>$filer,'auto-giveback'=>'true');
          };
        }

        #Also disable autosupport since it may veto giveback
        eval {
          $server->system_node_autosupport_modify(node=>$filer,
				state=>"disable");
        };
        #set takeover on panic and on reboot to true
        eval {
          $server->storage_failover_modify(node=>$filer,'onpanic'=>'true');
          $server->storage_failover_modify(node=>$filer,'onreboot'=>'true');
        };
        eval {
          $server->system_node_run(
                              node       => $filer,
                              '-Timeout' => 60,
                              command => "\"priv set test;setflag wafl_metadata_visible 1\""
          );
        };

        eval {
          #$server->network_interface_modify(server=>"vs0",'lif'=>'*', 'auto-revert'=>'false');
          $server->network_interface_revert(server=>"*",'lif'=>'*' );
          delete_extra_cluster_interface(serverobj=>$server,
                                    filername=>$filer);
        };
        eval {
         $server->network_interface_revert(server=>"*",'lif'=>'*' );
        };
        if($reboot_required) {
          $Log->trace("Trying to reboot both nodes since the boot flags are set");
          eval {
            $server->system_reboot("inhibit-takeover" => "true",
                                           node => $filer);
          };
          sleep(10);
          if(transit_node_to_cli(serverobj=>$server,
                      filername=>$filer)) {
                   $Log->trace("Transit $filer to CLI FALIED");
                   $Log->comment("Transit node to CLI FAILED");
                   $fail_count++;
          }
          sleep(10);
          $Log->comment("ATTEMPTING RECONNECT");
          #$server->transit_to(state => $server->CLI(), force_all_reconnect => 1);
          $Log->comment("RECONNCTION SUCCESSFULL");
        }
        
        },
             -bg,
             -runid, "initial_setup_$filer",
    );
    push(@setup_subtests,$st);

  }
  $Log->trace("Just seeing if the setup subtests for $filer1 and $filer2 finished otherwise wait");
  Subtest::wait_finish(subtests=>\@setup_subtests);
  $Log->trace("setup subtests seem to have finished");
  foreach my $filer(@filers) {
    $server->set_execution_server(-Server => $filer);
    $server->handle_connections( open => [qw(cli_connection)] );
  }
  ###########################################
  # Fix any stale objects
  ###########################################
  eval {
    run_vreport_tool(serverobj=>$server);
  };
  if ($fail_count) {return 1;} else { return 0};
}
## end sub disable_gdb

########################################################################
# Function:    get_object_status
#
# Description: Get the status of  Aggregate/Volume
#
# Parameters:  Server
#                   filer_name
#                   object_name - can be aggr or vol name
#                   object_type - can be aggr or vol
#                   local_or_partner(optional)
#
# Returns:     Aggr/Vol Status on Success
#	               NULL on failure
#
# Notes:
#   - Global parameters referenced: $server
########################################################################
sub get_object_status {
    my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},
                            object_name      => {type => SCALAR},                            
                            object_type      => {type => SCALAR},
                            local_or_partner => {type => SCALAR, optional => 1},
                           });

    my $server      = $options{serverobj};
    my $filer_name  = $options{filername};
    my $object_name = $options{object_name};    
    my $object_type = $options{object_type};
    my $local_or_partner;
    my $cmd_output,

      defined($options{local_or_partner})
      ? ($local_or_partner = $options{local_or_partner})
      : ($local_or_partner = "");

    $server->set_execution_server(-Server => $filer_name);
    $cmd_output =
      $server->system_node_run(
                          node       => $filer_name,
                          '-Timeout' => 60,
                          command => "\"$local_or_partner $object_type status $object_name\""
      );

    my @lines = split('\n', $cmd_output);    # split into lines
    shift(@lines);
    my $line = $lines[0];
    $line =~ s/^\s*//;
    $line =~ s/,\s*/,/g;
    my @fields = split(/\s+/,$line);
    return $fields[1];
} ## end sub get_object_status

sub wait_on_aggr_state {
    my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},
                            aggr_name      => {type => SCALAR},
                           });

    my $server     = $options{serverobj};
    my $filername  = $options{filername};
    my $aggr_name  = $options{aggr_name};
    my @objs;
    my $end_time = time()+120;

    $server->set_execution_server(-Server => $filername);
    while(time() < $end_time) {
      eval {
          @objs = $server->storage_aggregate_show(aggregate => $aggr_name);
      };
      if(!$@) {
         return 0
          if($objs[0]->{state} =~ /online|offline|restricted|partial/);
      }
      sleep(5);
    }
    return 1;
}

########################################################################
# Function:    get_object_fsid
#
# Description: Get FSID of  Aggregate/Volume
#
# Parameters:  Server
#                   filer_name
#                   object_name - can be aggr or vol name
#                   object_type - can be aggr or vol
# Returns:     Aggr/Vol FSID on Success
#	               NULL on failure
#
# Notes:
#   - Global parameters referenced: $server
########################################################################
sub get_object_fsid {
    my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},
                            object_name      => {type => SCALAR},                            
                            object_type      => {type => SCALAR},
                           });

    my $server      = $options{serverobj};
    my $filer_name  = $options{filername};
    my $object_name = $options{object_name};    
    my $object_type = $options{object_type};
    my $cmd_output,

    $server->set_execution_server(-Server => $filer_name);
    $cmd_output =
      $server->system_node_run(
                          node       => $filer_name,
                          '-Timeout' => 60,
                          command => "\"priv set -q test;$object_type read_fsid $object_name\""
      );

     #Sample : Aggregate stripedaggr_0_000 has an FSID of 0x4a4f4711.
     if($cmd_output =~ /has an FSID of (.*)\./) {
            $Log->trace("Returning FSID as $1");
            return ($1);
     } else {
    	$Log->trace("Returning $object_name - nothing");
    	return "";
     }
} ## end sub get_object_fsid

########################################################################
# Function:    run_vreport_tool
#
# Description: Will fix any stale volumes/aggrs/vservers
#
# Parameters:  Server
# Returns:     0 on success
#	       1 on failure
#
# Notes:
#   - Global parameters referenced: $server
########################################################################
sub run_vreport_tool {
    my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                           });
   my $server = $options{serverobj};
   my (@vreports, $nodename, $dsid, $command, $output);
   eval {
       @vreports = $server->debug_vreport_show();
   };
   foreach my $vreport(@vreports) {
       if ( $vreport->{object} =~ /unknown/i ) {
           $vreport->{attributes} =~ /"Node Name: (\S+)".*DSID:(\d+) MSID/;
           ($nodename, $dsid) = ($1, $2);
           $nodename =~ s/\"//g;
           $command = "volume lost-found delete -node $nodename -dsid $dsid";
           eval {
               $server->cli(cmd => $command, output => \$output);
           };
       }
       eval {
           $server->debug_vreport_fix(
                    type=>$vreport->{type},
                    object=>$vreport->{object});
       };
   }

}


########################################################################
# Function:    set_object_fsid
#
# Description: Set FSID of  Aggregate/Volume
#
# Parameters:  Server
#                   filer_name
#                   object_name - can be aggr or vol name
#                   object_type - can be aggr or vol
#                   fsid - FSID to set
# Returns:     0 on success
#	       1 on failure
#
# Notes:
#   - Global parameters referenced: $server
########################################################################
sub set_object_fsid {
    my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},
                            object_name      => {type => SCALAR},                            
                            object_type      => {type => SCALAR},
                            fsid             => {type => SCALAR},
                            vserver_name     => {type => SCALAR, optional =>1},
                            error_msg        => {type => SCALARREF, optional =>1},
                           });

    my $server      = $options{serverobj};
    my $filer_name  = $options{filername};
    my $object_name = $options{object_name};    
    my $object_type = $options{object_type};
    my $fsid = $options{fsid};
    my $error_string = "";
    my $error_msg = $options{error_msg};
    my ($cmd_output, $cmd, $fail_count);
    $fail_count=0;
    my $vserver_name;
   

    $server->set_execution_server(-Server => $filer_name);

    eval {
       if($object_type =~/aggr/) {
          $cmd = "priv set -q test;$object_type restrict $object_name;$object_type rewrite_fsid $object_name $fsid;$object_type online $object_name";
          $cmd_output = $server->system_node_run(
                             node       => $filer_name,
                             '-Timeout' => 60,
            command => "\"$cmd\""
         );
         $fail_count++
           if($cmd_output !~ /has had its FSID changed from/);
         
       } else {
        my $filername_with_underscores = $filer_name;
        $filername_with_underscores =~ s/-/_/g;
        defined($options{vserver_name})
          ? ($vserver_name = $options{vserver_name})
          : ($vserver_name =
         $Prefix . "_vs1" . "_" . $filername_with_underscores);
        ###  Get the junction path
        my @volumes = $server->volume_show(vserver => $vserver_name,
                                volume  => $object_name);
        my $junction_path = $volumes[0]->{'junction-path'};
   
        ###  Vol unmount
        $server->volume_unmount(vserver=>$vserver_name,
                                "foreground"=>"false",
                                "-Timeout"               => "600",
                                volume=>$object_name);
   
         ###  Vol restict
         $server->volume_modify(vserver     => $vserver_name,
                              volume             => $object_name,
                              "state" => "restricted");
   
         ###  set FSID
         $cmd = "priv set -q test;$object_type rewrite_fsid $object_name $fsid;";
   
         $server->system_node_run(
                             node       => $filer_name,
                             '-Timeout' => 60,
            command => "\"$cmd\""
         );
   
         ###  Vol online
         $server->volume_modify(vserver     => $vserver_name,
                              volume             => $object_name,
                              "state" => "online");
   
         ###  Vol mount
         $server->volume_mount(vserver=>$vserver_name,
                                volume=>$object_name,
                                "-Timeout"               => "600",
                                "junction-path"=>$junction_path);
       }
    };
    if($@ || $fail_count) {
        $Log->trace("Some problem with setting fsid : ".$@);
        $error_string .= "Some problem with setting fsid : $@";
        if(defined($error_msg)) {
           $$error_msg .= $error_string;
        }
        return 1;
    } else {
        $Log->trace("Successfully set FSID $fsid on $object_name");
        return 0;
    }
} ## end sub get_object_fsid
sub delete_extra_cluster_interface {
    my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR}});

    my ($filername,$server,$group,@show);

    $filername = $options{filername};
    $server    = $options{serverobj};
    my $cluster_name = "Cluster";
    my $hostobj = host($filername);
    if(defined($hostobj->{"CLUSTER_NAME"})) {
	$cluster_name = $hostobj->{"CLUSTER_NAME"};
    }
        eval {
          @show = $server->network_interface_show(
            server => $cluster_name,
            vif=>"cluster_mgmt"
        );
    };
    if($@) {
            $Log->trace("Network interface show for Server Cluster FAILED:".$@);
            return 0;
    }
    if(scalar(@show) == 0) {
            return 0;
    }
    $group = $show[0]->{'routing-group'};
    if ($server->network_routing_groups_route_exists(
            server          => $cluster_name,
            'routing-group' => $group,
            destination => "*",
            )) {
       eval {
        $server->network_routing_groups_route_delete(
            server          => $cluster_name,
            'routing-group' => $group,
            destination     => "*",
        );
       };
    }
    eval {
	  $server->network_interface_delete(
	         server => $cluster_name,
                 vif=>"cluster_mgmt"
	  );
    };
    if ($server->network_routing_groups_exists(
           server          => $cluster_name,
            'routing-group' => $group,
            )) {
           eval {
            $server->network_routing_groups_delete(
            server          => $cluster_name,
            'routing-group' => $group);
           };
    }
    return 0;
}

##Standby Utility function
sub display_message
{
        my $msg = shift;
        print "********************************************\n";
        print $msg."\n";
        print "********************************************\n";
}

sub get_aggr_name {
    my %options = validate(@_,
                           {filername        => {type => SCALAR}});
    my $filername_with_underscores = $options{filername};
    $filername_with_underscores =~ s/-/_/g;
   return( $Prefix . "_aggr_" . $filername_with_underscores."_a" );
}


########################################################################
# Function:    verify_nvsave_file()
#
# Description: Verifies the presence of nvsave files on local/partner node
#
# Parameters:  Server
#                   filer_name
#                   fsid - fsid of aggr/vol
# Returns:     0 on success
#	       1 on failure
#
########################################################################
sub verify_nvsave_file {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR|ARRAYREF},
                            partner_or_local => {type => SCALAR, optional => 1},
                            fsid             => {type => SCALAR|ARRAYREF},
                            error_msg      => {type => SCALARREF, optional =>1},
                            number_of_files      => {type => SCALAR, optional =>1},
                            single_node      => {type => SCALAR, optional =>1, default=>0},
                           });

  my $server = $options{serverobj};
  my $error_string = "";
  my $error_msg = $options{error_msg};
  my $filer = $options{filername};
  my $fsidparam = $options{fsid};
  my $single_node = $options{single_node};
  my $number_of_files = $options{number_of_files};
  my ($cmd, $diag, $setflag, $ls, $fsid, %output, $state,
	$partner, $partner_or_local, $in_takeover, @filers, 
	%no_nvsave, $partner_node, $nvsave_pat);

  defined($options{partner_or_local})
  ? ($partner_or_local = $options{partner_or_local})
  : ($partner_or_local = "");
  
  my (@fsids, @matches);
  if(ref($fsidparam) eq "ARRAY") {
    push(@fsids,@$fsidparam);
  } else {
    push(@fsids,$fsidparam);
  }

  $server->set_execution_server(-Server => $filer);
  # Form the path as specified in the example below
  #Eg. Path /vol/vol0/metadir/nvsave/vol4a9c3a31.log0
  $diag = "priv set diag";
  $ls = "ls /vol/vol0/metadir/nvsave/";
  $setflag = "setflag wafl_metadata_visible 1";
  $partner = "partner";

  # hash specifies the other node to verify
  $no_nvsave{local} = "partner";
  $no_nvsave{partner} = "local";

  #Get the node state (to see if its in takeover mode)
  $state = get_storage_failover_field(
			    serverobj  => $server,
                            filername  => $filer,
                            wait_before_check=>2,
                            field_name => "state");

  $in_takeover=1;
  if($state ne "takeover") {
    $in_takeover=0;
    if(!$single_node) {
      $partner_node = get_partner_node(
				serverobj=> $server,
			        filer    => $filer);
    } else {
      $partner_node = undef;
    }
    $partner="";
  } else {
    $partner_node = $filer;
  }

  foreach my $fsid(@fsids) {
    $fsid =~ s/0x//g;
    $nvsave_pat = "v?vol".$fsid.".log[0-1]";
    $cmd= "$diag\;$setflag\;${ls}";
    $output{$fsid}{local}   = run_nodescope_command(serverobj     => $server,
                          filername     => $filer,
                          command       => $cmd);

    if(defined($partner_node)) {
       $cmd= "$partner $diag\;$partner $setflag\;$partner ${ls}";
       $output{$fsid}{partner} = run_nodescope_command(serverobj     => $server,
                          filername     => $partner_node,
                          command       => $cmd);
    }
    
    if(defined($output{$fsid}{$partner_or_local})) {
      #It has to be either partner or local
      @matches = ($output{$fsid}{$partner_or_local} =~ /$nvsave_pat/g);
      if(!scalar(@matches)) {
	$error_string .= "Expected nvsave file for fsid $fsid on $partner_or_local root is absent\n";
        $Log->trace("Expected nvsave file on $partner_or_local root is absent");

        if(defined($number_of_files) && scalar(@matches) != $number_of_files) {
          $error_string .= "Expected $number_of_files, but found ".scalar(@matches)." files";
          $Log->trace("Expected $number_of_files, but found ".scalar(@matches)." files");
	}
	if(defined($error_msg)) {
	   $$error_msg .= $error_string;
	}
        return 1;
      }
      if(defined($partner_node) && ($output{$fsid}{$no_nvsave{$partner_or_local}}
			=~ /$nvsave_pat/g) ) {
        $Log->trace($no_nvsave{$partner_or_local}. " root also contains nvsave files".
				" expected not be be present");
        $error_string .= "Did not expect nvsave file for fsid $fsid on $no_nvsave{$partner_or_local} root\n";
        if(defined($error_msg)) {
           $$error_msg .= $error_string;
        }

	return 1;
      }
    } else {
      # Verify on both
      @matches = ($output{$fsid}{local} =~ /$nvsave_pat/g);
      if(!scalar(@matches)) {
        $error_string .= "Expected nvsave file for fsid $fsid on local root is absent\n";
        if(defined($number_of_files) && scalar(@matches) != $number_of_files) {
          $error_string .= "Expected $number_of_files, but found ".scalar(@matches)." files";
          $Log->trace("Expected $number_of_files, but found ".scalar(@matches)." files");
        }
        if(defined($error_msg)) {
           $$error_msg .= $error_string;
        }
	return 1;
      } 
      @matches = ($output{$fsid}{partner} =~ /$nvsave_pat/g);
      if(defined($partner_node) && !scalar(@matches )) {
        $error_string .= "Expected nvsave file for fsid $fsid on partner root is absent\n";
        if(defined($number_of_files) && scalar(@matches) != $number_of_files) {
          $error_string .= "Expected $number_of_files, but found ".scalar(@matches)." files";
          $Log->trace("Expected $number_of_files, but found ".scalar(@matches)." files");
        }
        if(defined($error_msg)) {
           $$error_msg .= $error_string;
        }
        return 1;
      }
    }
  }
  return 0;  
}

########################################################################
# Function:    verify_nvsave_delete()
#
# Description: Verifies that nvsave files are deleted from both nodes
#
# Parameters:  Server
#                   filer1,filer2
#                   fsid - fsid of aggr/vol
# Returns:     0 on success
#	       1 on failure
#
########################################################################
sub verify_nvsave_delete {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR|ARRAYREF},
                            partner_or_local => {type => SCALAR, optional => 1},
                            fsid             => {type => SCALAR|ARRAYREF},
                            error_msg      => {type => SCALARREF, optional =>1},
                           });

  my $server = $options{serverobj};
  my $filer = $options{filername};
  my $fsidparam = $options{fsid};
  my ($partner_or_local, @filers);
  my $error_string = "";
  my $error_msg = $options{error_msg};
  defined($options{partner_or_local})
  ? ($partner_or_local = $options{partner_or_local})
  : ($partner_or_local = "");
  
  my @fsids;
  if(ref($fsidparam) eq "ARRAY") {
    push(@fsids,@$fsidparam);
  } else {
    push(@fsids,$fsidparam);
  }

  $server->set_execution_server(-Server => $filer);
  my $partner_node = get_partner_node(
				serverobj=> $server,
			        filer    => $filer);
  push(@filers,$filer);
  push(@filers,$partner_node)
    if(defined($partner_node));
  foreach my $filer(@filers) {
    if(!verify_nvsave_file(serverobj=>$server,
                                 filername=>$filer,
                                 fsid=>$fsidparam,
                                )) {
      $error_string .= "Nvsave files not deleted on $filer";
      if(defined($error_msg)) {
         $$error_msg .= $error_string;
      }
      $Log->trace("Nvsave files not deleted on $filer");
      return 1;
    }
  }
  return 0;
}


########################################################################
# Function:    clear_fiji()
#
# Description: Clears all the fijis from filer1,filer2
#
# Parameters:  Server
#              filers
# Returns:     0 on success
#	       1 on failure
#
########################################################################
sub clear_fiji{
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filer        => {type => SCALAR|ARRAYREF},
                            error_msg      => {type => SCALARREF, optional =>1},
                           });
  my $server = $options{serverobj};
  my $filerparam = $options{filer};
  my @filers;
  my $error_string = "";
  my $error_msg = $options{error_msg};

  if(ref($filerparam) eq "ARRAY") {
    push(@filers,@$filerparam);
  } else {
    push(@filers,$filerparam);
  }
  my $fail_cnt=0;
  my $output;
  my $cmd= "\"priv set test;fiji clear all;fiji nvclear all;tp clear;fiji show;tp list;\"";
  foreach my $filer(@filers) {
    eval {
      $output=$server->system_node_run(node => $filer,
                       command => $cmd);
    };
    if($@) {
      $fail_cnt++;
      $error_string .= "$@\n";
    }

  }
  if(defined($error_msg)) {
    $$error_msg .= $error_string;
  }
  return $fail_cnt;
}
########################################################################
# Function:    get_ems_sequence_id()
#
# Description: Gets the sequence id of the latest ems message
#
# Parameters:  messsage - ems message
#
# Returns:     ems_seq_id  on success
#              -1 on failure
########################################################################
sub get_ems_sequence_id {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},
                            message             => {type => SCALAR},
                           });

  my $server = $options{serverobj};
  my $filer = $options{filername};
  my $ems_message_to_look = $options{message};
  my @ems_messages;

  $server->set_execution_server(-Server => $filer);
  eval {
     @ems_messages = $server->event_log_show(node=>$filer,
			messagename=> "$ems_message_to_look");
  };
  if($@) {
	$Log->trace("Some problem while looking for ems message :".$@);
	return -1;
  } 
  if(defined($ems_messages[0])) {
	return $ems_messages[0]->{seqnum};
  }
  return -1;
}

########################################################################
# Function:    get_ems_message_ids()
#
# Description: Gets the sequence id of the latest ems message for a
#	       list of messages
#
# Parameters:  ems_messsages - hash ref of ems messages to look for
#
# Returns:     array of hashes with {ems_message => ems_id} pairs
########################################################################
sub get_ems_message_ids {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            ems_messages     => {type => HASHREF},
                           });
  my $server = $options{serverobj};
  my $ems_messages_hash_ref = $options{ems_messages};
  my (@ems_details, $ems_messages);

  foreach my $filer(%$ems_messages_hash_ref) {
    $ems_messages = $ems_messages_hash_ref->{$filer};
    foreach my $ems_message(@$ems_messages) {
       push @ems_details, {ems_message=>$ems_message,
			filername=>$filer,
		ems_id=>get_ems_sequence_id(serverobj=>$server,
				filername=>$filer,
				message=>$ems_message)};
    }
  }
  return @ems_details;
}


########################################################################
# Function:    verify_ems_messages()
#
# Description: Verifies the presence ems_messages
#
# Parameters:  ems_details - array ref of {ems_message=>ems_id} pairs 
#
# Returns:     1 on Failure
#              0 on Success
########################################################################
sub verify_ems_messages {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            ems_details      => {type => ARRAYREF},
                           });
  my $server = $options{serverobj};
  my $filer = $options{filername};
  my $ems_details = $options{ems_details};
  foreach my $ems_detail(@$ems_details) {
    return 1
      if($ems_detail->{ems_id} >= get_ems_sequence_id(serverobj=>$server,
			 filername=>$ems_detail->{filername},
                         message=>$ems_detail->{ems_message}));
   }
   return 0;
}

########################################################################
# Function:    cleanup_nvsave_files()
#
# Description: Cleanup any stray nvsave files
#
# Parameters:  Server
#                   filer_name
#                   fsid - fsid of aggr/vol
# Returns:     0 on success
#              1 on failure
#
########################################################################
sub cleanup_nvsave_files {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filer1        => {type => SCALAR},
                            filer2        => {type => SCALAR,optional => 1},
                            error_msg      => {type => SCALARREF, optional =>1},
                           });

  my $server = $options{serverobj};
  my $filer1 = $options{filer1};
  my $filer2 = $options{filer2};
  my $error_string = "";
  my $error_msg = $options{error_msg};
  my $output;
  my @nvsave_files;
  my @filers;
  push(@filers,$filer1);
  push(@filers,$filer2)
    if(defined($filer2));

  #Eg. Path /vol/vol0/metadir/nvsave/vvol4a9c3a31.log0
  eval {
       foreach my $filer(@filers) {
         $output=$server->system_node_run(node => $filer,
                          command => "\"priv set test;setflag wafl_metadata_visible 1;ls /vol/vol0/metadir/nvsave/\"");
         @nvsave_files = ($output =~ /(v?vol.*?log[0-1](\.stale)?)/g);
         foreach my $nvsave_file(@nvsave_files) {
           $server->system_node_run(node => $filer,
                          command => "\"priv set test;setflag wafl_metadata_visible 1;rm /vol/vol0/metadir/nvsave/$nvsave_file\"");
         }
     }
  };
  if($@) {
     $Log->trace("Problem with nvsave cleanup:$@");
     $error_string .= "Problem with nvsave cleanup:$@";
     if(defined($error_msg)) {
        $$error_msg .= $error_string;
     }
     return 1;
  }
  return 0;
}

sub wait_till_file_exists {
 my %options = @_;
 my $filename = $options{filename};
 do {
	sleep 2;
 } while(-e $filename);
}

########################################################################
# Function:    wait_for_storage_failover_state
#
# Description: Waits till storage failover show, shows a specified state
#
# Parameters:  state - can be connected/giveback_partial etc..
#
# Returns:     0  on Success
#              1  on failure
#
# Notes:
#      -
########################################################################
sub wait_for_storage_failover_state {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},
                            state        => {type => SCALAR},
                            timeout	 => {type => SCALAR, optional=>1, default=>300},
                           });

  my $server = $options{serverobj};
  my $filer = $options{filername};
  my $state = $options{state};
  my $timeout = $options{timeout};
  $Log->step("Waiting for the storage failover show, to show $state");
  $server->set_execution_server(-Server => $filer);
  my $end_time = time()+$timeout;
  do {
        if(get_storage_failover_field(serverobj  => $server,
                                   filername  => $filer,
                                   wait_before_check=>2,
                                   field_name => "state") eq $state) {
          return 0;
        }
        sleep(5);
  } while(time() < $end_time);
  $Log->trace("Did not find the state $state");
  return 1;
}

sub get_state {
    my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},});
    my $server= $options{serverobj};
    my $filername = $options{filername};
    my $state="";

    $server->set_execution_server(-Server => $filername);
    eval {
        $state = $server->state();
    };
    return $state;
}
########################################################################
# Function:    wait_till_state
#
# Description: Waits till the node reaches specified state
#
# Parameters:  filername, state
#
# Returns:     0 on success
#              1 on failure
#
# Notes:
#   - Global parameters referenced: $Server
########################################################################
sub wait_till_state {
    my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},
                            state        => {type => SCALAR},
                            timeout      => {type => SCALAR, optional=>1},
                           });
    my $server= $options{serverobj};
    my $filername = $options{filername};
    my $expected_state= $options{state};
    my $state="";
    my $timeout;
    defined($options{timeout})
      ? ($timeout = $options{timeout})
      : ($timeout = 120);
    my $end_time = time()+$timeout;
    ################################################
    # Lets wait until we see the specified state
    ################################################
    $server->set_execution_server(-Server => $filername);
    do {
        sleep(5);
        eval {
                $state = $server->state();
        };
        if($state eq $expected_state) {
                return 0;
        }
    } while (time() < $end_time);
    $Log->trace("Timed out waiting for CFE prompt on $filername");
    return 1;
}

########################################################################
# Function:    transit_node_to_maint
#
# Description:  Transit the node to maintainence mode
#
# Parameters:  $server, $filername
#
# Returns:     0 on success
#                 1 on failure
#
# Notes:
#   
########################################################################
sub transit_node_to_maint {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},});
    my $server      = $options{serverobj};
    my $filername   = $options{filername};

    $Log->trace("Transitioning $filername to MAINT mode");  
    $server->set_execution_server(-Server => $filername);
    eval {    
      $server->transit_to(state => "IBM");  
      $server->console(cmd=>"maint",
                  additional_prompts_timeout => 1800,
                  additional_prompts => {
                  "^\\*>"      => "FINAL_PROMPT"});

    };
    if ($@) { 
      $Log->trace("transit_to maint mode Failed: $@");
      return 1;
    }
    return 0;
}

########################################################################
# Function:    execute_console_cmd
#
# Description: Execute a command on the console
#
# Parameters:  $server, $filername, $command
#
# Returns:     0 on success
#                 1 on failure
#
# Notes:
#   
########################################################################
sub execute_console_cmd{
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            output        => {type => SCALARREF, optional => 1},
                            error_msg      => {type => SCALARREF, optional =>1},
                            command     => {type=>SCALAR},});
    my $server       = $options{serverobj};
    my $filername   = $options{filername};
    my $command  = $options{command};
    my $output = $options{output};
    my $error_string = "";
    my $error_msg = $options{error_msg};

    $Log->trace("Execute the command $command on the Filer : $filername");  
    $server->set_execution_server(-Server => $filername);
    eval {    
      $$output = $server->console(cmd => $command, noprompt => 1);
    };
    if ($@) { 
      $Log->trace("Command Execution Failed : $@");
      $error_string .= "Command Execution Failed : $@";
      if(defined($error_msg)) {
         $$error_msg .= $error_string;
      }
      return 1;
    }
    return 0;
}

########################################################################
# Function:    get_volume_option 
#
# Description: To get the value of the vol option specified
#
# Parameters:  $server, $filername, $volume, $vserver, $option
#
# Returns:     value on success
#              die on failure
########################################################################
sub get_volume_option{
   my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
			    volume	  => {type => SCALAR},
			    vserver       => {type => SCALAR},
			    option	  => {type => SCALAR},});

   my $server        = $options{serverobj};
   my $filername     = $options{filername};
   my $volume        = $options{volume};
   my $vserver_name  = $options{vserver};
   my $option        = $options{option};
   my @output;

   $server->set_execution_server(-Server => $filername); 
   eval {
      @output = $server->volume_show(vserver => $vserver_name,
                                     volume  => $volume);
   };
   if ($@) {
      $Log->trace("Command Execution Failed : $@");
      die "Error in get_volume_option :$@";
   }
   if(defined($output[0])){
      return $output[0]->{$option}; 
   }
   die "Error in get_volume_option :$@";
}
########################################################################
# Function:    set_volume_option 
#
# Description: To set the value of the vol option specified
#
# Parameters:  $server, $filername, $volume, $option, $vserver, $value
#
# Returns:     0 on success
#              1 on failure
########################################################################
sub set_volume_option{
   my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
			    volume	  => {type => SCALAR},
			    vserver       => {type => SCALAR},
			    option	  => {type => SCALAR},
			    value	  => {type => SCALAR},});

   my $server       = $options{serverobj};
   my $filername    = $options{filername};
   my $volume       = $options{volume};
   my $vserver_name = $options{vserver};
   my $option       = $options{option};
   my $value        = $options{value};

   $server->set_execution_server(-Server => $filername); 
   eval {
     $server->volume_show(vserver => $vserver_name,
                          volume  => $volume);
   };
   if ($@) {
     $Log->trace("Command Execution Failed : $@");
     return 1;
   }
   eval {
      $server->volume_modify(vserver => $vserver_name,
                             volume  => $volume,
			     $option => $value);
   };
   if ($@) {
      $Log->trace("Command Execution to modify volume options failed: $@");
      return 1;
   }
   return 0;
}


########################################################################
# Function:    get_member_aggr
#
# Description: Gets the member aggregate name from the specified node
#
# Parameters:  $server, $filername, $aggregate
#
# Returns:     0 on success
#                 1 on failure
#
# Notes:
#   
########################################################################
sub get_member_aggr
{
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            aggregate     => {type => SCALAR},
                            filername     => {type => SCALAR},
                           });
    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $aggregate   = $options{aggregate};
    my @aggrs=$server->storage_aggregate_member_show(
			aggregate=>"$aggregate*"); 
    foreach my $aggr_of (@aggrs){	
       if($aggr_of->{node} =~ /$filername/) {
	 return $aggr_of->{aggregate};
       }
    }
    return "";
}


########################################################################
# Function:    get_member_vol
#
# Description: Gets the member volume name from the specified node
#
# Parameters:  $server, $filername, $aggregate
#
# Returns:     0 on success
#                 1 on failure
#
# Notes:
#   
########################################################################
sub get_member_vol
{
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            volume        => {type => SCALAR},
                            aggregate     => {type => SCALAR},
                            filername     => {type => SCALAR},
                           });
    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $volume      = $options{volume};
    my $aggregate   = $options{aggregate};
    my $member_aggr = get_member_aggr(serverobj=>$server,
				aggregate=>$aggregate,
				filername=>$filername);


    my @vols =$server->volume_member_show(
			volume=>"$volume*"); 

    foreach my $vol_of (@vols){	
       if($vol_of->{aggregate} =~ /$member_aggr/) {
	 return $vol_of->{volume};
       }
    }
    return "";
}

########################################################################
# Function:    set_coral_heartbeat
#
# Description: Sets coral heartbet to on/off
#
# Parameters:  $server, $filername, $value
#
# Returns:     0 on success
#              1 on failure
#
# Notes:
#   
########################################################################
sub set_coral_heartbeat {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR|ARRAYREF},
                            value         => {type => SCALAR},
                           });
    my $server          = $options{serverobj};
    my $filernameparam  = $options{filername};
    my $value           = $options{value};
    my $cmd;
    my @filernames;
    if(ref($filernameparam) eq "ARRAY") {
      push(@filernames,@$filernameparam);
    } else {
      push(@filernames,$filernameparam);
    }
    foreach my $filer(@filernames) {
      $cmd = "coraltest heartbeat -f $filer $value";
      $server->set_execution_server(-Server => $filer);
      eval {
        $server->shell(cmd => $cmd, noprompt => 0, -Errors => 0);
      };
      if($@) {
        $Log->trace("Error in set_coral_heartbeat: $@");
        return 1;
      } 
    }
    return 0;
}

########################################################################
# Function:    get_ems_details()
#
# Description: Gets the ems details for a single or list of ems messages
#
# Parameters:  a single message or a array refrence of messages
#
# Returns:     {ems_name=>message, ems_seqnum=> -1, ems_detail=>""} on failure
#              {ems_name=>message, ems_seqnum=> seqnum, ems_detail=>"details .." } on success 
########################################################################
sub get_ems_details{
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},
                            messages         => {type => SCALAR|ARRAYREF},
                           });

  my $server = $options{serverobj};
  my $filer = $options{filername};
  my $ems_messages_param = $options{messages};
  my (@ems_info, @ems_messages, @messages);
  if(ref($ems_messages_param) eq "ARRAY") {
    push(@messages,@$ems_messages_param);
  } else {
    push(@messages,$ems_messages_param);
  }

  $server->set_execution_server(-Server => $filer);
  foreach my $message(@messages) {
     eval {
        @ems_messages = $server->event_log_show(node=>$filer,
   			messagename=> "$message");
     };
     if($@) {
   	$Log->trace("Some problem while looking for ems message :".$@);
           push (@ems_info,{emsname=>$message, 
   		seqnum=>-1,
   		detail=>""});
     } elsif (defined($ems_messages[0])) {
           push (@ems_info,{emsname=>$message, 
   		seqnum=>$ems_messages[0]->{seqnum},
   		detail=>$ems_messages[0]->{event}});
     } else {
           push (@ems_info,{emsname=>$message, 
   		seqnum=>-1,
   		detail=>""});
     }
  }
  return @ems_info;
}


########################################################################
# Function:    verify_ems_messages_generated()
#
# Description: Verifies the presence ems_messages
#
# Parameters:  ems_details - array of ems_details
#
# Returns:     1 on Failure
#              0 on Success
########################################################################
sub verify_ems_messages_generated {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            ems_details      => {type => ARRAYREF},
                            filername        => {type => SCALAR},
                            fsid             => {type => SCALAR},
                           });
  my $server = $options{serverobj};
  my $filer = $options{filername};
  my $ems_details = $options{ems_details};
  my $fsid= $options{fsid};
  $fsid = hex($fsid);
  my @latest_ems_detail;
  foreach my $ems_detail(@$ems_details) {
    @latest_ems_detail = get_ems_details(serverobj=>$server,
			     filername=>$filer,
                             messages=>$ems_detail->{emsname});
    return 1
      if($ems_detail->{seqnum} >= $latest_ems_detail[0]->{seqnum});
    return 1
      if($latest_ems_detail[0]->{detail} !~ /$fsid/);
			 
   }
   return 0;
}

########################################################################
# Function:    mount_volume()
#
# Description: Mount a volume
#
# Parameters:  
# 	data_ip 	: DATA_IP1 entry in the tharnhost     
#       junction_path   : eg: /myvolume
#       mount_path      : eg: /t/mnt/
#       client_name     : by default its assumed to be localhost 
#
# Returns:     dies      on  Failure
#              mountobj  on Success
# Usage:
#           my $mntobj = mount_volume(.....);
#           my $cmd="ls ".$mntobj->{path}"
#           my ($status, @output) = &ClientUtil::execute($mntobj->{conn},$cmd);
########################################################################
sub mount_volume {
    my %options = validate(@_,
                           {
                            data_ip       => {type => SCALAR},
			    junction_path => {type => SCALAR},
                            mount_path    => {type => SCALAR},
                            client_name   => {type => SCALAR, optional=>1},
                           });
    my $junction_path = $options{junction_path};
    my $mount_path    = $options{mount_path};
    my $data_ip = $options{data_ip};
    my $client_name;
    defined($options{client_name})
      ? ($client_name= $options{client_name})
      : ($client_name= "localhost");

    my $share = "$data_ip:/$junction_path";
    my $client = new Client(-host     => $client_name,
                            -timeout  => 100,
                            -protocol => "nfs");
    # NFS share to mount
    my $mntobj = $client->create_mnt(-share => $junction_path,
                                     -path  => $mount_path);
    if ($mntobj->error) {
        $Log->trace("MOUNT ERROR :". $mntobj->errmsg());
        die "Problem in mounting $junction_path to $mount_path";
    }
    return $mntobj;
}

########################################################################
# Function:    run_nodescope_command()
#
# Description: Runs given command from node scope
#
# Parameters:  
# 	filername,command
#
# Returns:     dies      on  Failure
#              command output on Success
########################################################################

sub run_nodescope_command
{
  my %options = validate(@_,
                         {
                          serverobj     => {type => OBJECT},
                          filername     => {type => SCALAR | ARRAYREF},
                          command       => {type => SCALAR},
			  noprompt      => {type => SCALAR, default=>0, optional=>1},
                          timeout       => {type => SCALAR, optional=>1},
                         });
  my $server = $options{serverobj};
  my $cmd = $options{command};
  my $filernameparam= $options{filername};
  my $noprompt = $options{noprompt};
  my @filernames;
  if(ref($filernameparam) eq "ARRAY") {
    push(@filernames,@$filernameparam);
  } else {
    push(@filernames,$filernameparam);
  }
  my ($output, $timeout);
  
  defined($options{timeout})
      ? ($timeout = $options{timeout})
      : ($timeout = 600);
  foreach my $filer(@filernames) {
    $server->set_execution_server(-Server => $filer); 
    eval {
      $server->system_node_run(node    => $filer,
                             command   => "\"$cmd\"",
			     noprompt => $noprompt,
                             '-Timeout' => $timeout,
                             output     => \$output);
    };
    if($@) {
      $Log->comment("Error running command : $@");
      die $@;
    }
  }
  return $output;
}

########################################################################
# Function:    get_latest_ems_seq_id()
#
# Description: Gets the sequence id of the latest specified ems message
#              for the specied fsid
#
# Parameters:  ems_name, filername, fsid
#
# Returns:     -1 on Failure
#              seq_id  on Success
########################################################################
sub get_latest_ems_seq_id {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            ems_name         => {type => SCALAR},
                            filername        => {type => SCALAR},
                            fsid             => {type => SCALAR},
                            error_msg      => {type => SCALARREF, optional =>1},
                           });
  my $server = $options{serverobj};
  my $filer = $options{filername};
  my $ems_name = $options{ems_name};
  my $fsid= $options{fsid};
  my $error_string = "";
  my $error_msg = $options{error_msg};

  $fsid = hex($fsid);
  my @ems_messages;
  $server->set_execution_server(-Server => $filer); 
  eval {
     @ems_messages = $server->event_log_show(node=>$filer,
  	messagename=> "$ems_name",
	event=>"*$fsid*");
  };
  if($@) {
    $Log->trace("Some problem while looking for ems message :".$@);
    $error_string .= "Some problem while looking for ems message :$@";
    if(defined($error_msg)) {
      $$error_msg .= $error_string;
    }
    return -1;
  } elsif (defined($ems_messages[0])) {
    return $ems_messages[0]->{seqnum};
  } else {
    if(defined($error_msg)) {
      $$error_msg .= "Unknown error while looking for EMS $ems_name";
    }
    return -1;
  }
}

########################################################################
# Function:    set_storage_failover_field()
#
# Description: Sets storage failover field
#
# Parameters:  filername, field, value
#
# Returns:     1 on Failure
#              0 on Success
########################################################################
sub set_storage_failover_field {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            filername        => {type => SCALAR},
                            field            => {type => SCALAR},
                            value            => {type => SCALAR},
                            error_msg      => {type => SCALARREF, optional =>1},
                           });
  my $server = $options{serverobj};
  my $filer  = $options{filername};
  my $field  = $options{field};
  my $value  = $options{value};
  my $error_string = "";
  my $error_msg = $options{error_msg};

  $server->set_execution_server(-Server => $filer); 
  return 0
    if (get_storage_failover_field(serverobj  => $server,
                                   filername  => $filer,
                                   wait_before_check=>2,
                                   field_name => $field 
        ) eq $value);
  eval {
    $server->storage_failover_modify(node=>$filer,"$field"=>"$value");
  };
  if($@) {
    $Log->trace("Setting storage failover $field to $value FAILED", $@);
    $error_string .= "Setting storage failover $field to $value FAILED : $@";
    if(defined($error_msg)) {
       $$error_msg .= $error_string;
    }
    return 1;
  }
  return 0;
}


########################################################################
# Function:    takeover()
#
# Description: Initiates a takeover
#
# Parameters:  bynode, takeover_option (normal|immediate|force|allow-version-mismatch)
#
# Returns:     1 on Failure
#              0 on Success
########################################################################
sub takeover {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            bynode           => {type => SCALAR},
                            takeover_option  => {type => SCALAR},
                            error_msg      => {type => SCALARREF, optional =>1},
                           });
  my $server = $options{serverobj};
  my $bynode = $options{bynode};
  my $option = $options{takeover_option};
  my $error_string = "";
  my $error_msg = $options{error_msg};

  $server->set_execution_server(-Server => $bynode); 
  eval {
     $server->storage_failover_takeover(
                          bynode   => $bynode,
			   option   => $option,
                          -Timeout => 60);
  };
  if($@) {
    print "PRAS: Takeover FAILED".$@."\n";
    $Log->trace("Takeover FAILED", $@);
    $error_string .= "Takeover FAILED $@\n";
    if(defined($error_msg)) {
      $$error_msg .= $error_string;
    }
    return 1;
  }
  return 0;
}

########################################################################
# Function:    giveback()
#
# Description: Initiates a giveback
#
# Parameters:  fromnode, partner_waiting(optional), override_vetoes(optional)
#
# Returns:     1 on Failure
#              0 on Success
########################################################################
sub giveback {
  my %options = validate(@_,
                           {serverobj        => {type => OBJECT},
                            fromnode         => {type => SCALAR},
                            partner_waiting  => {type => SCALAR, optional=>1, default=>"true"},
                            override_vetoes  => {type => SCALAR, optional=>1, default=>"true"},
                            error_msg      => {type => SCALARREF, optional =>1},
                           });
  my $server = $options{serverobj};
  my $fromnode = $options{fromnode};
  my ($partner_waiting, $override_vetoes);
  my $error_string = "";
  my $error_msg = $options{error_msg}; 
  my $return=0;

  $partner_waiting = $options{partner_waiting};
  $override_vetoes = $options{override_vetoes};

  $server->set_execution_server(-Server => $fromnode); 
  eval {
     $server->storage_failover_giveback (
                          fromnode=> $fromnode,
			  "require-partner-waiting" => $partner_waiting,
 		          "override-vetoes" => $override_vetoes,
                          -Timeout => 60);
  };
  if($@) {
    $Log->trace("Giveback FAILED", $@);
    $error_string .= "Giveback FAILED: $@";
    if(defined($error_msg)) {
       $$error_msg .= $error_string;
    }
    
    $return=1;
  }
  eval {
     $server->storage_failover_show_giveback();

  };
  return $return;
}

########################################################################
# Function:    set_node_scope_option()
#
# Description: Sets an option from node scope
#
# Parameters:  filername, option, value
#
# Returns:     1 on Failure
#              0 on Success
########################################################################
sub set_node_scope_option {
  my %options = validate(@_,
                           {serverobj  => {type => OBJECT},
                            filername  => {type => SCALAR},
                            option     => {type => SCALAR},
                            value      => {type => SCALAR},
                           });
  my $server = $options{serverobj};
  my $filer  = $options{filername};
  my $option = $options{option};
  my $value  = $options{value};

  $server->set_execution_server(-Server => $filer); 
  eval {
     run_nodescope_command(
               serverobj     => $server,
               filername     => $filer,
               command       => "priv set test;options $option $value"
         );
  };
  return 1 if($@);
  return 0;
}

########################################################################
# Function:    get_node_scope_option()
#
# Description: Gets an option from node scope
#
# Parameters:  filername, option
#
# Returns:     null on Failure
#              value on Success
########################################################################
sub get_node_scope_option {
  my %options = validate(@_,
                           {serverobj  => {type => OBJECT},
                            filername  => {type => SCALAR},
                            option     => {type => SCALAR},
                           });
  my $server = $options{serverobj};
  my $filer  = $options{filername};
  my $option = $options{option};
  my $value  = "";

  $server->set_execution_server(-Server => $filer); 
  eval {
     $value = run_nodescope_command(
               serverobj     => $server,
               filername     => $filer,
               command       => "priv set test;options $option"
         );
  };
  return $value;
}

########################################################################
# Function:    disable_automatic_options()
#
# Description: Disable all the automatic options 
#
# Parameters:  filername
#
# Returns:     1 on Failure
#              0 on Success
########################################################################
sub disable_automatic_options {
  my %options = validate(@_,
                           {serverobj  => {type => OBJECT},
                            filername => {type => SCALAR | ARRAYREF},
                           });
  my $server = $options{serverobj};
  my $filernameparam= $options{filername};
  my @filernames;
  if(ref($filernameparam) eq "ARRAY") {
    push(@filernames,@$filernameparam);
  } else {
    push(@filernames,$filernameparam);
  }
  my @option_names = ("cf.giveback.auto.enable",
		       "cf.takeover.on_failure",
		       "cf.takeover.on_network_interface_failure",
		       "cf.takeover.on_panic",
		       "cf.takeover.on_reboot",
		       "cf.takeover.on_short_uptime",
		       "cf.hw_assist.enable");		
  foreach my $filer(@filernames) {
     $server->set_execution_server(-Server => $filer); 
     foreach my $option(@option_names) {
        return 1
   	if(set_node_scope_option(serverobj=>$server,
   			      filername=>$filer,
   			      option=>$option,
   			      value=>"off"));
     }
  }
}

########################################################################
# Function:    enable_automatic_options()
#
# Description: Enable all the automatic options 
#
# Parameters:  filername
#
# Returns:     1 on Failure
#              0 on Success
########################################################################
sub enable_automatic_options {
  my %options = validate(@_,
                           {serverobj  => {type => OBJECT},
                            filername => {type => SCALAR | ARRAYREF},
                           });
  my $server = $options{serverobj};
  my $filernameparam= $options{filername};
  my @filernames;
  if(ref($filernameparam) eq "ARRAY") {
    push(@filernames,@$filernameparam);
  } else {
    push(@filernames,$filernameparam);
  }
  my @option_names = ("cf.giveback.auto.enable",
		       "cf.takeover.on_failure",
		       "cf.takeover.on_network_interface_failure",
		       "cf.takeover.on_panic",
		       "cf.takeover.on_reboot",
		       "cf.takeover.on_short_uptime",
		       "cf.hw_assist.enable");		
  foreach my $filer(@filernames) {
     $server->set_execution_server(-Server => $filer); 
     foreach my $option(@option_names) {
        return 1
   	if(set_node_scope_option(serverobj=>$server,
   			      filername=>$filer,
   			      option=>$option,
   			      value=>"on"));
     }
  }
}

########################################################################
# Function:    wait_on_cf_status
#
# Description: Waits till a particular message is found
#              by  cf status command
#
# Parameters:  filername,message
#
# Returns:     0  on Success
#              1  on failure
#
########################################################################
sub wait_on_cf_status {
  my %options = validate(@_,
                           {serverobj  => {type => OBJECT},
                            filername => {type => SCALAR},
			    message => {type => SCALAR},
                           });
   my $server = $options{serverobj};
   my $message = $options{message};
   my $filername = $options{filername};
   my $end_time = time()+300;
   my $found=0;
   my $output="";

   while(time() < $end_time) {
       eval {
        $output = $server->system_node_run(
                         node       => $filername,
                         '-Timeout' => 60,
                         command => "\"priv set test;cf status\"");
       };
       if($output =~ /$message/) {
           $Log->trace("Finally cf status command gave :  $output");
	   return 0; 
       }
       sleep 5;
   }
   $Log->trace("Did not see $message from cf status command");
   return 1;
}

########################################################################
# Function:    execute_cmd_on_ngsh
#
# Description: Execute a command on the ngsh
#
# Parameters:  $server, $filername, $command
#
# Returns:     Output string on success
#              null  on failure
#
# Notes:
#   
########################################################################
sub execute_cmd_on_ngsh {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            command     => {type=>SCALAR},
                            error_msg      => {type => SCALARREF, optional =>1},
			   });
    my $server       = $options{serverobj};
    my $filername   = $options{filername};
    my $command  = $options{command};
    my $error_string = "";
    my $error_msg = $options{error_msg};

    my $output = "";

    $Log->trace("Execute the command $command on the Filer : $filername");  
    $server->set_execution_server(-Server => $filername);
    eval {    
      $server->cli(cmd => $command, output => \$output);
    };
    if ($@) { 
      $Log->trace("Command Execution Failed : $@");
      $error_string .= "Command Execution Failed : $@";
      if(defined($error_msg)) {
        $$error_msg .= $error_string;
      }
      return "";
    }
    return $output;
}

########################################################################
# Function:    reestablish_connection
#
# Description: Execute a command on the ngsh
#
# Parameters:  $server, $filername
#
# Returns:     error string on failure
#              1 on success
#
# Notes:
#   
########################################################################
sub reestablish_connection {
 
    # There are those times when you need to re-establish the connection to the filer
    # If using the console I don't know, but you do.
 
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},});
			   
    my $server  = $options{serverobj};
    my $filer   = $options{filername};
 
    eval {
         $server->handle_connections(close => [qw(shell_connection cli_connection)]);
    };
    sleep 10;
    eval {
        $server->handle_connections(open => [qw(shell_connection cli_connection)]);
    };
    eval {
        $server->set_execution_server(-Server => $filer);
    };
    return ($@);
}


sub setup_failover_rule {
   my %options = validate(@_,
                          {serverobj     => {type => OBJECT},
                           filername     => {type => SCALAR},
                           vserver_name  => {type => SCALAR},});

   my $server  = $options{serverobj};
   my $filername   = $options{filername};
   my $vserver_name= $options{vserver_name};
   my $partner_node = get_partner_node(
                                serverobj=> $server,
                                filer    => $filername);
   my $filername_with_underscores = $filername;
   $filername_with_underscores =~ s/-/_/g;
   my $vif = $filername_with_underscores."_data1";
   my  $hostobj   = host($filername);

   eval {
        $server->network_interfaces_failover_delete(
                                                server => $vserver_name,
                                                vif    => $vif,
                                                priority => "1,2,3,4,5,6,7,8,9",
        );
    };
    if ($@) {

        #This failure is ok since this is a blind cleanup
    }

    eval {
        $server->network_interfaces_failover_new(
                                   server   => $vserver_name,
                                   vif      => $vif,
                                   priority => "1",
                                   filer    => $partner_node,
                                   port => $hostobj->{"DATA_PORT1"}
        );
    };
    if ($@) {
        $Log->trace("Network Interface failover creation for $filername FAILED...",
                    $@);
        return 1;
    } 
    return 0;
}

sub setup_failover_rules {
    my %options = validate(@_,
                           {serverobj => {type => OBJECT},
                            filername => {type => SCALAR | ARRAYREF},
                            vserver_name  => {type => SCALAR},});
    my $server    = $options{serverobj};
    my $vserver_name    = $options{vserver_name};
    my $filernameparam = $options{filername};
    my @filernames;
    if(ref($filernameparam) eq "ARRAY") {
      push(@filernames,@$filernameparam);
    } else {
      push(@filernames,$filernameparam);
    }
    foreach my $filer(@filernames) {
      $server->set_execution_server(-Server => $filer);
      return 1
        if(setup_failover_rule(serverobj=>$server,
			    filername=>$filer,
			    vserver_name=>$vserver_name));
    }
    return 0;
}


########################################################################
# Function:   power_cycle_node
#
# Description: Power cycle a node
#
# Parameters:  $server, $filername
#
# Returns:     0 on Success
#	       1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub power_cycle_node {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            error_msg      => {type => SCALARREF, optional =>1},
			   });
    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $error_string = "";
    my $error_msg = $options{error_msg};

    my $host = host($filername);

    $server->set_execution_server(-Server => $filername);
    $Log->trace("Invoking power cycle on $filername");
    eval {
        $server->rlm(cmd => "system power cycle", -Timeout => 920,
                                  additional_prompts_timeout => 1800,
                  additional_prompts => {
                  "Continue.*"      => "y",
                  "RLM.*|bmc.*|SP.*"    => 'FINAL_PROMPT'});

    };
    if($@) {
	$Log->trace("Some problem happened during power_cycle - ".$@);
        $error_string .= "Some problem happened during power_cycle - $@";
        if(defined($error_msg)) {
           $$error_msg .= $error_string;
        }
	return 1;
    } else {
	$Log->trace("Power Cycle succeeded");	
	return 0;
    }
}

########################################################################
# Function:   stop_traffic
#
# Description: Stops the traffic
#
# Parameters:  Refrence to subtests array
#	       Reference to mountobjects array
#
# Returns:     0 on Success
#	       1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub stop_traffic {
    my %options = validate(@_,
                           {subtests=> {type => ARRAYREF},
                            error_on_corruption=> {type => SCALAR, optional=>1},
                            hamlog_file=> {type => SCALAR, optional=>1, default=>"hamlog.txt"},
                            error_string=> {type => SCALARREF, optional=>1},
                            mountobjs=> {type => ARRAYREF},});
    my $subtests  = $options{subtests};
    my $mountobjs = $options{mountobjs};
    my $error_on_corruption = $options{error_on_corruption};
    my $hamlog_file = $options{hamlog_file};
    my $error_string = $options{error_string};
    my ($err_msg,$cmd);
    my $fail_count = 0;
    ###########################################
    # Stop all the Hammer instances
    ###########################################
    $Log->trace("Stop all the Hammer instances");
    while (my $st = pop @$subtests) {
        if (!$st->isAlive()) {
           $Log->trace("Looks like traffic is not running");
           $fail_count++
	     if($error_on_corruption);
           $err_msg .= "Traffic is not running : $fail_count\n";
	} else {
            $Log->trace("Traffic is still on........");
            eval { $st->killme(); };
            if ($@) {
                $Log->trace("Traffic Stop FAILED........");
                $fail_count++;
                $err_msg .= "PRAS2: Stop Traffic FAILED: $fail_count\n";
            } else {
                $Log->trace("Traffic Stop succeded........");
            }
        } ## end if (defined($k) && $k->isAlive...
    } ## end while (my $k = pop @Subtests)
    ###########################################
    # Give some time to make sure the hammer has stopped.
    ###########################################
    $Log->trace("Give some time to make sure the hammer has stopped.");
    sleep(10);
    ###########################################
    # Remove all the mount points.
    ###########################################
    $Log->trace("Remove all the mount points.");
    while (my $mnt = pop @$mountobjs) {
        if ($mnt->is_mounted) {
            $cmd="grep failed ".$mnt->{path}."/hammer/$hamlog_file";
            my ($status, @output) = &ClientUtil::execute($mnt->{conn},$cmd);
            if(@output) {
	       $Log->trace("*****************CORRUPTION FOUND***********");
               $err_msg .= "Corruption Found: @output\n";
	       $fail_count++
		 if($error_on_corruption);
            }
            $mnt->destroy();
            if ($mnt->error) {
                $Log->trace($mnt->errmsg());
		$err_msg .= "Unable to unmount : ".$mnt->errmsg()."\n";
                $fail_count++;
            }
        } else {
                $Log->trace("Volume not mounted\n");
		$err_msg .= "Volume not mounted\n";
                $fail_count++;
	}
    }
    if ($fail_count) {
	$$error_string = $err_msg
	  if(defined($error_string));
        return 1;
    } else {
        return 0;
    }
} 

########################################################################
# Function:   get_aggr_disk_list
#
# Description: Gets the list of disks of a given aggregate
#
# Parameters:  $server, $filername, $aggregate
#
# Returns:     array of disks on success
#	       dies on failure on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub get_aggr_disk_list {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            aggregate     => {type => SCALAR},
			   });
    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $aggregate   = $options{aggregate};
    my (@objs,@disks);
    $server->set_execution_server(-Server => $filername);
    eval {
      @objs = $server->storage_aggregate_show(aggregate => $aggregate);
    };
    if(!$@) {
       print Dumper($objs[0]->{disklist});
       my $row = $objs[0]->{disklist};
       my @disks;
       foreach my $dsk(@$row) {
          $dsk =~ s/.*://g;
          push @disks,$dsk;
       }
       return @disks;
    }
    die "Get aggr disks FAILED:$@";
}

########################################################################
# Function:   remove_disk_ownership
#
# Description: Removes disk ownership for a given set of disks
#
# Parameters:  $server, $filername, $disk/@disks
#
# Returns:     0 on success
#              1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub remove_disk_ownership {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            disks         => {type => SCALAR|ARRAYREF},
                            error_msg      => {type => SCALARREF, optional =>1},
			   });
    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $diskparam   = $options{disks};
    my $error_string = "";
    my $error_msg = $options{error_msg};

    my @disks;
    if(ref($diskparam) eq "ARRAY") {
      push(@disks,@$diskparam);
    } else {
      push(@disks,$diskparam);
    }
    $server->set_execution_server(-Server => $filername);
    
    my $cmd = "disk remove_ownership -f ".join(" ",@disks);
    eval {
      run_nodescope_command(serverobj     => $server,
                          filername     => $filername,
                          command       => $cmd);
    };
    if($@) {
      $Log->trace("Some problem with removing disk ownership : $@");
      $error_string .= "Some problem with removing disk ownership : $@";
      if(defined($error_msg)) {
         $$error_msg .= $error_string;
      }
      return 1;
    }
    return 0;
}

########################################################################
# Function:   assign_disk_ownership
#
# Description: Assigns disk ownership for a given set of disks
#
# Parameters:  $server, $owner, $disk/@disks
#
# Returns:     0 on success
#              1 on failure
#
# Notes:
#   - Global parameters referenced: None
########################################################################
sub assign_disk_ownership {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            owner         => {type => SCALAR},
                            filername     => {type => SCALAR},
                            disks         => {type => SCALAR|ARRAYREF},
                            error_msg      => {type => SCALARREF, optional =>1},
			   });
    my $server      = $options{serverobj};
    my $owner       = $options{owner};
    my $filername   = $options{filername};
    my $diskparam   = $options{disks};
    my $error_string = "";
    my $error_msg = $options{error_msg};
    my @disks;
    if(ref($diskparam) eq "ARRAY") {
      push(@disks,@$diskparam);
    } else {
      push(@disks,$diskparam);
    }
    
    $server->set_execution_server(-Server => $filername);
    my $cmd = "disk assign  ".join(" ",@disks)." -o $owner";
    eval {
      run_nodescope_command(serverobj     => $server,
                          filername     => $filername,
                          command       => $cmd);
    };
    if($@) {
      $Log->trace("Some problem with assigning disk ownership : $@");
      $error_string .= "Some problem with assigning disk ownership : $@";
      if(defined($error_msg)) {
         $$error_msg .= $error_string;
      }
      return 1;
    }
    return 0;
}

sub get_object_field {
    my %options = validate(@_,
                           {serverobj     => {type => OBJECT},
                            filername     => {type => SCALAR},
                            object_name        => {type => SCALAR},
			    object_type	  => {type => SCALAR},
			    field_name    => {type => SCALAR},
			    vserver       => {type => SCALAR, optional=>1},
			   });
    my $server      = $options{serverobj};
    my $filername   = $options{filername};
    my $object_name = $options{object_name};
    my $object_type = $options{object_type};
    my $field_name = $options{field_name};
    my $vserver= $options{vserver};

    my @objs;
    $server->set_execution_server(-Server => $filername);
    eval {
        if($object_type eq "aggr") {
             @objs = $server->storage_aggregate_show(aggregate => $object_name);
        } else {
             @objs = $server->volume_show(vserver => $vserver,
                             volume  => $object_name);
        }
    };
    if(!$@) {
       return $objs[0]{$field_name}
         if(defined($objs[0]{$field_name}));
    }
    die "Problem in getting $field_name of $object_name : $@";
}

sub filegrep {
    my %options = validate(@_,
                           {
                            filename     => {type => SCALAR},
			    message	 => {type => SCALAR},
			    seek         => {type => SCALAR, optional=>1, default=>0},
			   });
    my $filename   = $options{filename};
    my $message = $options{message};
    my $seek = $options{seek};
    my $found=0;
    open FILE, $filename or die $!;
    seek(FILE, $seek, 0); 
    my @lines = <FILE>;
    if(grep(/$message/, @lines)) {
     $found=1;
    }
    close(FILE);
    return 0
      if($found);
    return 1; 
}

sub get_console_filename {
    my %options = validate(@_,
                           {
                            filername     => {type => SCALAR},
			   });
    my $filername   = $options{filername};
    my $logdir = $ENV{LOGDIR};
    my $logfile=$logdir."/console\-$filername"."\.txt";

    return $logfile;
}

sub wait_on_console {
    my %options = validate(@_,
                           {
                            filername     => {type => SCALAR},
			    message	  => {type => SCALAR},
			    seek	  => {type => SCALAR, optional=>1, default=>undef},
			    timeout       => {type => SCALAR, optional=>1, default=>300},
			   });
    my $filername   = $options{filername};
    my $message = $options{message};
    my $timeout = $options{timeout};
    my $logdir = $ENV{LOGDIR};
    my $logfile=$logdir."/console\-$filername"."\.txt";
    my $end_time = time()+$options{timeout}; 
    my $seek = $options{seek};

    if(!defined($seek)) {
      $seek = -s $logfile;
    }
    
    do {
      sleep(2);
      if ( !filegrep(filename=>$logfile,
		message=>$message,
	        seek=>$seek)) { 
	return 0;
      }
    } while(time() < $end_time);
    return 1;
}


########################################################################
## Function:   aggr_add_disks
##
## Description: Aggregate Add
##
## Parameters:  server,filername,aggr_name,disk_count
##
## Returns:     0 on Success
##              1 on failure
##
## Notes:
##   - Global parameters referenced: None
#########################################################################
#
sub aggr_add_disks{
    my %options = validate(@_,
                           {serverobj => {type => OBJECT},
                            filername => {type => SCALAR, optional => 1},
                            aggr_name => {type => SCALAR},
                     	    disk_count => {type => SCALAR},
                            foreground => {type => SCALAR, optional => 1, default=> "true"},
                            error_msg => {type => SCALARREF, optional =>1},
                           });
    my $server    = $options{serverobj};
    my $filername = $options{filername};
    my $aggr_name = $options{aggr_name};
    my $disk_count = $options{disk_count};
    my $foreground = $options{foreground};
    my $error_string = "";
    my $error_msg = $options{error_msg};
    eval {
        $server->set_execution_server(-Server => $filername)
          if(defined($filername));
        $server->storage_aggregate_add_disks(aggregate   => $aggr_name,
                                          "diskcount" => $disk_count,
                                          foreground=>$foreground);
    };
    if ($@) {
        $Log->trace("Aggr Add disk to Aggregate $aggr_name FAILED " . $@);
        $error_string .= "Aggr Add disk to Aggregate $aggr_name FAILED  : $@";
        if(defined($error_msg)) {
           $$error_msg .= $error_string;
        }
        return 1;
    }
    return 0;
} ## end sub aggr_add_disks

##end_of_file
1;
