#!/bin/sh /usr/software/bin/perl 

#
#  Class that models SMF tables
#
package Table;

use strict;
no strict "refs";

use Data::Dumper;
use FindBin;
use lib "$FindBin::Bin/../../lib";
use TracePkg;

my $t = TracePkg->new();

sub new {

    # an instance of "this" is always the first arg
    # the second arg is a set of named arguments used to init
    #  a hash.
    my ($self, %arg) = @_;

    # Fill the member vars for this instance
    my $object = {
	_name     => $arg{name},
	_desc     => $arg{desc},
	_attrs    => $arg{attrs},
	_fields   => undef,
	_fieldsOrder => undef,
	_commands => undef,
	_cmdDir   => undef,
	_methods  => undef,
	_priv     => "???",
	_type     => $arg{type},
	};

    # For table types of create, replicated, persistent, 
    # or automatic, create the new and delete commands
    if ( $object->{_attrs}=~ 
	 /(create|replicated|persistent|automatic)/ ) {
	$object->{_commands}{new} = 
	    Command->new( name => "new" );
	$object->{_commands}{delete} = 
	    Command->new( name => "delete" );
    }

    # This tags the hash pointed to by $object as a 
    #  type of Table
    bless $object, $self;

    # And give the caller the referece to the bless object.
    return $object;
}


sub getName {
    my ($self, %arg) = @_;
    return $self->{_name};
}


sub addField {
    my ($self, %arg) = @_;

    # Add the name of the field object to a list of names
    # I need to do this to maintain the order that I inserted
    # the elements into the %_fields hash.
    push @{ $self->{_fieldsOrder} }, ( $arg{name} );

    # Now add the new Field object to the hash.  
    $self->{_fields}{$arg{name}} = $arg{field};

    #print "just added $arg{field} to $self->{_fields}\n";
    #print "The fields are now: ", keys %{$self->{_fields}}, "\n";
    # Now set the default priv for the field.  It (initially
    # matches that of the table
    $arg{field}->setPriv( priv => $self->{_priv} );

    # Now create the modify command:
    # if this is a write/modify field, 
    #  and the command doesn't already exist 
    #  and this isn't an action table
    if ( ($arg{field}->getQualifier() =~ /(write|modify)/) &&
	 !($self->{_type} =~ /ACTION/) ) {
	if ( !$self->hasCommand("modify") ) {
	    #print "We need to create the modify command\n";	
	    $self->{_commands}{modify} = 
		Command->new( name => "modify");
	    $t->debug(msg=>"Added modify command: for table:" .
		      $self->{_name} );
	}
    } 
}


sub setFields {
    my ($self, %arg) = @_;
    $self->{_fields} = $arg{fields};
}


sub setFieldsOrder {
    my ($self, %arg) = @_;
    $self->{_fieldsOrder} = $arg{fieldsOrder};
}


sub isActionTable {
    my ($self) = @_;
    if ( exists $self->{_type} && 
	 $self->{_type} =~ /ACTION/ ) {
	    return 1;
    } else {
	return undef;
    }
}


sub getField { 
    my ($self, $name) = @_;
    if ( exists $self->{_fields}{$name} ) {
	return $self->{_fields}{$name};
    } else {
	return undef;
    }
}


sub getFields { 
    my ($self) = @_;

    return $self->{_fields};
}


sub getFieldsOrder { 
    my ($self) = @_;

    return $self->{_fieldsOrder};
}


sub getKeys {
    my ($self) = @_;
    my $key;
    my @keyList = ();

    for $key (  @{$self->{_fieldsOrder}} ) {
	if ( $self->{_fields}{$key}->isKey() ) {
	    push @keyList, $self->{_fields}{$key};
	}
    }

    if ( scalar(@keyList) > 0 ) {
	$t->debug(msg=>"Returning keys");
	return @keyList;
    }
    return undef;
}


sub hasCommand {
    my ($self, $name) = @_;

    if ( exists $self->{_commands}{$name} ) {
	return 1;
    } else {
	return 0;
    }
}


sub getCommand {
    my ($self, $name) = @_;
    if ( exists $self->{_commands}{$name} ) {
	return $self->{_commands}{$name};
    }
    return ;
}


sub getCmds {
    my ($self) = @_;
    if ( exists $self->{_commands} ) {
	return $self->{_commands};
    }
    return ;
}


# Return a hash of all locally stored commands AND
# all commands stored within locally stored Methods
sub getAllCmds {
    my ($self) = @_;
    my %allCommands = undef;
    my ($cmd, $method);

    # If there are no commands, return immediately
    if ( !defined $self->{_cmdDir} ) {
	$t->debug(msg=>"No cmd dir, so no commands: $self->{_name}");
	return undef;
    }

    #$t->debug(msg=>"Table is:\n" . Dumper($self));
    # Walk the Methods grabbing commands
    for $method (keys %{ $self->{_methods} }) {
	$t->debug(msg=>"Working on method: $method");
	#$t->debug(msg=> Dumper($self->{_methods}->{$method}));
	$allCommands{$self->{_methods}->{$method}->getCommand()->getName()} =
	    $self->{_methods}->{$method}->getCommand();
    }
    for $cmd (keys %{ $self->{_commands} }) {
	$allCommands{$self->{_cmdDir}. " " .
			 $self->{_commands}->{$cmd}->getName()} =
	    $self->{_commands}->{$cmd};
    }
    #print Dumper(%allCommands);
    return \%allCommands;
}


sub setCmdDir {
    my ($self, $cmdDir) = @_;

    $self->{_cmdDir} = $cmdDir;
}


sub getCmdDir {
    my ($self) = @_;
    if ( exists $self->{_cmdDir} ) {
	return $self->{_cmdDir};
    }
    return ;
}


sub delCommand {
    my ($self, $name) = @_;
    if ( exists $self->{_commands}{$name} ) {
	delete $self->{_commands}{$name};
	return;
    }
    #print "no such command: $name\n";
    return;
}


sub addCommand {
    
    my ($self, %arg) = @_;

    # Add the new command the list hash of commands.
    $self->{_commands}{$arg{name}} = $arg{command};
    #my $commandName = $arg{command}->getCmdString();
    #print "Added $arg{name} - $commandName\n";
}


sub addMethod {
    my ($self, %arg) = @_;
    $self->{_methods}{$arg{name}} = $arg{method};
    #@{ $self->{_methods} } = @methods;
}


sub getMethods {
    my ($self) = @_;
    if ( exists $self->{_methods} ) {
	return $self->{_methods};
    }
    return ;
}


# Set the privilege for this table.  We can either get
# the priv from the table's attributes section, or 
# default it to passed named ("default") arg.
sub setPriv {
    
    my ($self, %arg) = @_;

    # Set the priv level if coded in the attr section
    if ( $self->{_attrs} =~ /admin/ ) {
	$self->{_priv} = "admin";
    } elsif ( $self->{_attrs} =~ /advanced/ ) {
	$self->{_priv} = "advanced";
    } elsif ( $self->{_attrs} =~ /diag/ ) {
	$self->{_priv} = "diag";
    }

    # Set passed default if not already set
    if ( $self->{_priv} =~ /\?\?\?/ ) {
	$self->{_priv} = $arg{default};
    }	
}


sub getPriv {
    my ($self) = @_;
    return $self->{_priv};
}

sub getDesc {
    my ($self) = @_;
    return $self->{_desc};
}

sub getAttrs {
    my ($self) = @_;
    return $self->{_attrs};
}

sub printMe {

    my ($self) = @_;
    my $key;

    print "Table name: \t$self->{_name}\tType: $self->{_type}\n";
    print "\tDescription: \t$self->{_desc}\n";
    print "\tAttributes: \t$self->{_attrs}\n";
    print "\tPrivilege: \t$self->{_priv}\n";
    #print "\tCommand: \n";
    for $key ( keys %{ $self->{_commands} }  ) {
	#print "\t    ";
	$self->{_commands}{$key}->printMe();
    } 
    
    print "\tFields: \n";
    #print "\t\tFields hash: ",  %{$self->{_fields}}, "\n";
    # Print the fields in the order they were discovered/created
    #for $key ( keys %{ $self->{_fields} }  ) {
    for $key (  @{$self->{_fieldsOrder}} ) {
	$self->{_fields}{$key}->printMe();
    } 
    print "\tMethods: \n";
    for $key ( keys %{ $self->{_methods} } ) {
	$self->{_methods}{$key}->printMe();
    } 
    print "\n";
}


1;
