# $Id: //depot/prod/test/Rsierranevada/lib/SMIS.pm#1 $
#
# Copyright (c) 2012 NetApp, Inc.
# All rights reserved.
#
## @summary Library of functions to configure and test SMI-S functionality 
##
## @description The functions in this module load, configure and execute
## SMI-S commands used to test the reliability and functionality of the 
## Data ONTAP SMI-S Agent.
##
##
## @param   ASSOCIATION_LIST            Required:       Text file with association entries
##
## @param   AGENT_BIN_PATH              Required:       Location on server where agent release is installed
##
## @param   AGENT_BIN_START             Required:       Name of agent start script 
##
## @param   AGENT_BIN_STOP              Required:       Name of agent stop script 
##
## @param   AGENT_LOG                   Required:       Name of agent logfile
##
## @param   AGENT_PROC			        Required:       Agent process name
##
## @param   AGENT_TOOLS_PATH            Required:       Location of agent binary tools
##
## @param   CIM_PROPERTY                Required:       CIM Property Name, e.g. ONTAP_StorageSystem
##
## @param   CLIENT                      Optional:       Unix client where the SMIS commands are executed, can be agent or separate system.
##
## @param   CLIENT_BIN_PATH		        Required:       Location of SMI-S client tool
##
## @param   CLIENT_BIN			        Required:       Name of SMI-S client tool
##
## @param   CLIENT_RETRY			    Required:       # of times command should be retried.
##
## @param   DOWNLOADER			        Required:       Path to downloading tool
##
## @param   FILER                       Required:       Hostspec of filer
##
## @param   INS                         Required:       interop name space
##
## @param   JAVA                        Required:       Path to java binary accessible to the ndisp server
##
## @param   PKG_INSTALLER               Required:       Release installer name
##
## @param   FORCE_INSTALL               Required:       Destination for release on test host
##
## @param   PKG_REMOVER                 Required:       Release uninstaller package name, e.g. uninstall_smisproxy
##
## @param   PKG_XTENSION                Required:       Extension used for the release package, e.g. 'tar'
##
## @param   PRODUCT_NAME                Required:       Name of product under test, e.g. 'DOT SMI-S Agent'
##
## @param   QUICK_TEST                  Optional:       Perform quick testing without loading lastest rlse
##
## @param   RLSE_URL                    Required:       Path to SMIS release
##
## @param   SLP_PROC                    Required:       slp process name
##
## @param   SLP_START                   Required:       Name of slp start script
##
## @param   SLP_STOP                    Required:       Name of slp stop script
##
## @param   SXPORT                      Required:       SMI-S secure transport name, e.g. https
##
## @param   TESTDEBUG                   Optional:       If this parameter is boolean TRUE then debug messages will be printed to the log.
##
## @param   VER_FILE                    Required:       SMIS release version
##
## @param   VNS                         Required:       SMI-S VNS - Vendor Name Space, e.g. root/ontap
##
## @param   XPORT                       Required:       SMI-S transport name, e.g. http
##
## @status  public
##
## @usage
## See example script test/SMIS/associator_names.thpl
##
## @author galan@netapp.com
## @change 08-10-2008 galan:      Added Caymus profile framework
## @change 08-23-2011 galan:      Added support for Pegasus CIMOM and cimclient
## @change 01-12-2012 galan:      Added support for Windows package download and uninstall/install 
## @status public
## @pod here


package SMIS;
use strict;
use warnings;
use misc;
use Connectrec qw();
use Hostrec qw();
use boot qw(reboot);
require "tharnit_lib.pl";
use Tharn;
use Tharn qw(snooze);
use Data::Dumper;
use File::Spec;
use File::Path;
use File::Copy;
use TestEnv::MinWin qw(:all);
use MapShare qw(:all);
use Cwd;
use NATE::ParamSet;
use NATE::Log;

use vars qw(@ISA @EXPORT $AUTOLOAD $FILER $CLIENT 
    $AGENT_BIN $QUICK_TEST $RLSE_URL $PRODUCT_NAME $FORCE_INSTALL
    $PKG_XTENSION $PKG_INSTALLER $PKG_REMOVER $CIM_PROPERTY 
    $AGENT_BIN_PATH $AGENT_TOOLS_PATH $AGENT_BIN_START $AGENT_LOG
    $AGENT_PROC $SLP_PROC $PROFILES $CLIENT_BIN_PATH $CLIENT_BIN $SXPORT $XPORT $ns $INS $JAVA);

use base qw(Exporter);
@ISA = ('Exporter');

BEGIN {
our @EXPORT_OK = qw( 
    get_config
    stop_srvcs
    proc_state
    slpd_start
    slpd_stop
    stop_cwbem
    start_srvcs
    providers
    exec_providers
    path_to_cifs
    start_cwbem
    filer_version
    filer_id
    filer_cmd
    pkg_info
    pkg_xtension
    cimom_ver
    client_os_details
    object_path
    associator_names
    enumerate
    populate_cache
    pkg
);
}

=head1 NAME

SMIS - Functions for testing SMIS Provider

=head1 SYNOPSIS

Note: C<$CLIENT> B<MUST> be a Linux, Windows 2008 or Windows 8 machine.

This library handles the following basic tasks:
1. Installing any SMI-S release
2. Starting and stopping the SMI-S Agent
3. Provides a common framework for SMIS tests to establish and manipulate
configurations for tests


=head1 DESCRIPTION

Note: C<$CLIENT> B<MUST> be a Linux, Windows 2008 or Windows 8 machine.

This library handles the basic tasks of installing any SMI-S release,
starting and stopping the SMI-S Agent and providing a common framework
for SMIS tests to establish and manipulate configurations for tests.
Configuration is persisted per-filer, so that the user may easily direct
a specific filer(or group of filers) to be serviced by a specific SMIS
provider, etc.


The filer_obj() function is used to create a config object which will be used
for testing; at least one config object must be created for each filer or sim
to be tested. Each object will initially contain the defaults for each
parameter passed to the script that uses this module; these values may be
overridden for individual filers or for groups of filers using looping or
iteration. 

The recommended method for creating an arbitrary number of filer objects is
illustrated in the init() function of the example script linked below,
associator_names.thpl. By creating a hash of config objects with integer keys,
it becomes very easy to do things like e.g. "the first 10 filers should 
talk to provider A, the next 10 filers should talk to provider B, etc."

=head1 FUNCTIONS

All functions, unless otherwise stated, return a true value on success, and
false on failure.


=cut

=head2 hosts_obj($FILER, $CLIENT, $debug)

Creates a configuration object in memory.

This method takes a regular Tharn C<$FILER> object, and returns a
configuration object containing details of how C<$FILER> is configured.
Without an object created in this way, most of the other methods in this
module will not work properly or at all. 

Passing in anything that evaluates to BOOLEAN TRUE for the $debug argument
will cause this module to log detailed debugging information whenever one of
its subroutines are called. Below is an example of a typical configuration
object as printed by Data::Dumper:

 $VAR1 = bless( {
                 '_agent_tools_path' => '/usr/ws/bin',
                 '_xport' => 'http',
                 '_debug_module_verbose' => undef,
                 '_filer_hostname' => 'mgt-u2',
                 '_filer_ip      ' => '172.16.180.101',
                 '_provider_class' => undef,
                 '_client_connect' => bless( {
                                               'connhost' => '172.16.180.16',
                                               'conncmd' => undef,
                                               '_comm' => bless( {
                                                                   'pty' => undef,
                                                                   'logbuf' => '',
                                                                   'connp' => $VAR1->{'_client_connect'},
                                                                   'accum' => ''
                                                                 }, 'Tharn_Comm::Pty' ),
                                               'maxbufsize' => 1048576,
                                               'connport' => undef,
                                               'password' => 'sundance',
                                               'connid' => undef,
                                               'timeout' => '400',
                                               'reserved_ports' => undef,
                                               'debug' => undef,
                                               'id' => 'mgt-ln16#rsh',
                                               'first_connect' => 1,
                                               'shell_type' => 'bash',
                                               'logfile' => undef,
                                               'password_in_username' => undef,
                                               'conntype' => 'rsh-persay',
                                               'max_reconnect' => undef,
                                               'prompts' => [
                                                              'mgt-ln16__prompt__> '
                                                            ],
                                               'username' => 'root',
                                               'hostp' => bless( {
                                                                   'prompt_take' => undef,
                                                                   'system_id' => undef,
                                                                   'hostname' => 'mgt-ln16',
                                                                   'prompt_fw' => [
                                                                                    '^\\{[0-9]*\\} ok '
                                                                                  ],
                                                                   'prompt_norm' => [
                                                                                      'mgt-ln16__prompt__> '
                                                                                    ],
                                                                   'mac_address' => undef,
                                                                   'makedisks' => undef,
                                                                   'CONFIGS' => 'fs-ext,fs-ext2,fs-ext3,fs-hfs-a,fs-reiser,h-atm-no,h-nfs,h-nfsv4,hcpu-i386,hcpu-i486,hcpu-i586,hcpu-i686,hrsh-root,linux,linux-rh,rawio,rhel,rhel4.0,rhes,rhes4.0,rpc-none,unix',
                                                                   'OS' => 'Linux',
                                                                   'default_ip_7g' => undef,
                                                                   'id' => 'mgt-ln16',
                                                                   'prompt_virt' => undef,
                                                                   'OS_VER' => '2.6.9-5.EL',
                                                                   'simhost' => undef,
                                                                   'part_host' => undef,
                                                                   'default_ip_gx' => undef,
                                                                   'simdir' => undef,
                                                                   'default_ip' => '172.16.180.16',
                                                                   'irptchar' => undef,
                                                                   'connection_fields' => {
                                                                                            'rsh' => {
                                                                                                       'password' => 'sundance',
                                                                                                       'timeout' => '400',
                                                                                                       'username' => 'root'
                                                                                                     }
                                                                                          },
                                                                   'default_password' => 'sundance',
                                                                   'bootimage' => undef,
                                                                   'powercmd' => 'power_cycle mgt-ln16',
                                                                   'hosttype' => 'unix-linux-redhat'
                                                                 }, 'Hostrec::Unix' ),
                                               'inherit' => 0,
                                               'max_reconnect_timewait' => undef
                                             }, 'Connectrec' ),
                 '_pkg_remover' => 'uninstall_smisproxy',
                 '_agent' => '172.16.180.16',
                 '_logdir_unix' => '/x/eng/testbeds/mgt/svl_SMIS/logs/SMIS_1186615778_0--20070808_162939',
                 '_force_install' => '/tmp/smis/releases',
                 '_pkg_xtension' => 'tar',
                 '_agent_log' => 'cwbemserver.out',
                 '_prompt' => [
                                'mgt-u2\\*?> '
                              ],
                 '_object_path' => 'instance of ONTAP_FilerData {
                                        ElementName = "mgt-u1";'
                                    };
                                    ,                 
                 '_provider_association' => undef,
                 '_client_bin_path' => '/usr/wsi/sdk/bin',
                 '_filer_password' => 'sundance',
                 '_vns' => 'root/ontap',
                 '_agent_bin_path' => '/usr/ws/server/agent/bin',
                 '_agent_bin_start' => 'start_server',
                 '_rlse_url' => '/x/eng/rlse/smis/*/releases/dist',
                 '_sxport' => 'https',
                 '_client' => 'mgt-ln16',
                 '_filer' => 'mgt-u2',
        #         '_pkg_installer' => 'install_smisproxy',
               }, 'SMIS' );

Note: This example is a dump of a typical config object at runtime, immediately
after instantiation. Sort order is not guaranteed.

B<Usage:> C<$obj = filer_obj($FILER);>

B<Returns:> C<$obj> configuration object

=cut

sub hosts_obj () {
	
    my $_class = shift;
    my $_filer = shift;
    my $_client = shift;
    my $_debug = shift;
    
    my $message;
    
    $message = "In hosts_obj(), \$_client is" . " $_client and \$_filer is $_filer";
    $Log->comment($message);

    my $_client_obj  =  host($_client);
    my $_client_connect  =  connect("$main::CLIENT#rsh-persay");
    my $_client_hostname  =  $_client_obj->hostname();
    my $_client_host_type  =   $_client_obj->hosttype();
    my $_client_ip  =  $_client_obj->default_ip();   
    my $_client_os  =  $_client_obj->OS();      
    my $_client_os_ver  =  $_client_obj->OS_VER();      
    my $_client_username  =  $_client_obj->DEFAULT_USERNAME();      
    my $_client_password  =  $_client_obj->default_password();      
    my $_client_power = $_client_obj->powercmd();
    
    # Verify client username and password is set
    if ($_client_username eq "") {
         my $message = "\n\n$_client_hostname 'DEFAULT_USERNAME' is not set in the tharnhost file.\n\n\n\n";
         if ($_client_os eq "Linux") {
             $_client_username = "root";
         } else {
             $_client_username = "Administrator";
         }
         
         $Log->comment($message);
    }
    if (!$_client_password) {
         my $message = "\n\n$_client_hostname 'default_password' is not set in the tharnhost file.\n\n\n\n";     
         $Log->error($message);
    }
    
    my $_filer_obj   =   host($_filer);
    my $_filer_connect  =  connect("$main::FILER#rsh-persay");
    my $_filer_hostname   =   $_filer_obj->hostname();
    my $_filer_ip   =   $_filer_obj->default_ip();   
    my $_filer_password   =   $_filer_obj->default_password();
    my $_prompt     =   $_filer_obj->prompt_norm();

    $message = "My client is $_client_hostname\nMy client OS ver is $_client_os_ver\nMy default filer ip is $_filer_ip\n";
    $Log->comment($message);

    my $self = {
        _client           =>  $_client,
        _client_connect   =>  $_client_connect,
        _client_hostname  =>  $_client_hostname,        
        _client_host_type =>  $_client_host_type, 
        _client_ip        =>  $_client_ip,
        _client_os        =>  $_client_os,
        _client_os_ver    =>  $_client_os_ver,
        _client_username  =>  $_client_username, 
        _client_password  =>  $_client_password, 
        _client_power     =>  $_client_power,
        _filer            =>  $_filer,
        _filer_connect	  =>  $_filer_connect,
        _filer_hostname   =>  $_filer_hostname,
        _filer_ip         =>  $_filer_ip,        
        _filer_password   =>  $_filer_password,
        _prompt           =>  $_prompt,
    };

    my %_cfg = (
        _project                => $main::PROJECT,
        _association_list       => $main::ASSOCIATION_LIST,        
        _agent_bin              => $main::AGENT_BIN,
        _agent_bin_path         => $main::AGENT_BIN_PATH,        
        _agent_bin_start        => $main::AGENT_BIN_START,
        _agent_bin_stop         => $main::AGENT_BIN_STOP,
        _agent_log              => $main::AGENT_LOG,
        _agent_proc             => $main::AGENT_PROC,
        _agent_tools_path       => $main::AGENT_TOOLS_PATH,
        _debug_module_verbose   => $_debug,
        _client_retry           => $main::CLIENT_RETRY,                
        _class_path             => $main::CLASS_PATH,                
        _client_bin             => $main::CLIENT_BIN,                
        _client_bin_path        => $main::CLIENT_BIN_PATH,
        _create_instance        => $main::CREATE_INSTANCE,
        _downloader_list        => $main::DOWNLOADER_LIST,
        _downloader_get         => $main::DOWNLOADER_GET,
        _ins                    => $main::INS,
        _java                   => $main::JAVA,
        _logdir_unix            => $main::LOGDIR_UNIX,
        _pkg_installer          => $main::PKG_INSTALLER,
        _force_install          => $main::FORCE_INSTALL,
        _pkg_remover            => $main::PKG_REMOVER,
        _pkg_xtension           => $main::PKG_XTENSION,
        _port                   => $main::PORT,
        _profiles               => $main::PROFILES,        
        _product_name           => $main::PRODUCT_NAME,
        _quick_test             => $main::QUICK_TEST,
        _rlse_url               => $main::RLSE_URL,
        _slp_proc               => $main::SLP_PROC,
        _slp_start              => $main::SLP_START,
        _slp_stop               => $main::SLP_STOP,
        _sport                  => $main::SPORT,
        _sxport                 => $main::SXPORT,
        _ver_file               => $main::VER_FILE,
        _vns                    => $main::VNS,
        _xport                  => $main::XPORT,

    ); # Temporary hash to store test config details

    # Load info from temporary hash into $self

    foreach my $key (keys %_cfg) {
    
        $self->{$key} = $_cfg{$key};
    }
    
    bless $self, ref($_class) || $_class;


    $message = "\"Configuration Retrieved:\n\" . Dumper($self)";
    $Log->debug($message);

    return $self;

}

=head2 smi_association()

This subroutine will get the AssociatorNames and AssociatorInstances of the specified object.

B<Usage:> C<$obj-E<gt>smi_association($smis_class, $namespace, $args, $query_type)>  

B<Returns:> A List of the AssociatorNames (Traverses an association) and instances of the results class. 

=cut

sub smi_association() {

    my $self = shift;
    my $smis_class = shift;
    my $namespace = shift;
    my $args = shift;
    my $query_type = shift;

    my $dbg = $self->get_debug_module_verbose();

    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $class_path = $self->{_class_path};
    my $java = $self->{_java};
    my $port = $self->{_port};
    my $xport = $self->{_xport};
    my $retry = $self->{_client_retry};
    my $query;
    my $message;

    my $results;

    for (my $i = 0; $i < $retry; $i++) {
          			
		if ($smis_class && $args) {
			$query = "$java -classpath $class_path $args smicli $client_username $client_password $xport://$client_ip:$port $namespace $smis_class $query_type";
        
			$message = "smi_association(): " . "query is\n" . "$query\n";
			$Log->comment("$message");  
			$results = `$query`; # get query
		} else {

            if (!$smis_class && !$args) {
                $message = "\$smis_class and \$args" . "\n" . "have empty values\n";
            }

            elsif (!$smis_class) {
                $message = "\$smis_class" . "\n" . "has empty value\n";
            }

            elsif (!$args) {
                $message = "\$args" . "\n" . "has empty value\n";
            } else {
                $message = "Incorrect values have been specified in \$smis_class and/or \$args" . "\n";
                $Log->comment($message);
            }

            my $msg = "smi_association(): " . $message . "\n";
            $Log->comment($msg);

        }

        if ($results !~ m/Connection failed/) {

            last;

        } else {

            if ($i == $retry) {
                my $message = "smi_association(): " . "Agent $client_hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n" . "\n";
                $Log->comment($message);
               
            }
       }

    }

    $results =~ s/\s+$//;
	$Log->comment($results);
    return $results;

}

=head2 smi_references()

This subroutine will get the ReferenceNames and ReferenceInstances of the specified object.

B<Usage:> C<$obj-E<gt>smi_references($smis_class, $namespace, $args, $query_type)> 

B<Returns:> A List of the ReferenceNames (traversed association) and instances of the results class. 

=cut

sub smi_references() {

    my $self = shift;
    my $smis_class = shift;
    my $namespace = shift;
    my $args = shift;
    my $query_type = shift;

    my $dbg = $self->get_debug_module_verbose();

    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $class_path = $self->{_class_path};
    my $java = $self->{_java};
    my $port = $self->{_port};
    my $xport = $self->{_xport};
    my $retry = $self->{_client_retry};
    my $query;
	my $message;

    my $results;

    for (my $i = 0; $i < $retry; $i++) {
		
		if ($smis_class && $args) {
			$query = "$java -classpath $class_path $args smicli $client_username $client_password $xport://$client_ip:$port $namespace $smis_class $query_type";
        
			$message = "smi_references(): " . "query is\n" . "$query\n";
			$Log->comment("$message");  
			$results = `$query`; # get query
		} else {

            if (!$smis_class && !$args) {
                $message = "\$smis_class and \$args" . "\n" . "have empty values\n";
            }
            elsif (!$smis_class) {
                $message = "\$smis_class" . "\n" . "has empty value\n";
            }
            elsif (!$args) {
                $message = "\$args" . "\n" . "has empty value\n";
            } else {
                $message = "Incorrect values have been specified in \$smis_class and/or \$args" . "\n";
                $Log->comment($message);
                
            }

            my $msg = "smi_references(): " . $message . "\n";
            $Log->comment($msg);

        }

        if ($results !~ m/Connection failed/) {

            last;

        } else {

            if ($i == $retry) {
				 my $message = "smi_references(): " . "Agent $client_hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n" . "\n";
                $Log->error($message);

            }
       }

    }

    $results =~ s/\s+$//;
	$Log->comment($results);
    return $results;

}

=head2 smi_methods()

This subroutine will execute the SBLIM invokeMethod object.

B<Usage:> C<$obj-E<gt>smi_methods($smis_class, $namespace, $args, $query_type)> 

B<Returns:> The value of the specified method either 0 or 4096 (if successful).

=cut

sub smi_methods() {

    my $self = shift;
    my $smis_class = shift;
    my $namespace = shift;
    my $args = shift;
    my $query_type = shift;
	
    my $dbg = $self->get_debug_module_verbose();

    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $class_path = $self->{_class_path};
    my $java = $self->{_java};
    my $port = $self->{_port};
    my $xport = $self->{_xport};
    my $retry = $self->{_client_retry};
    my $query;
	my $message;

    my $results;

    for (my $i = 0; $i < $retry; $i++) {
	
		if ($smis_class && $args) {
			$query = "$java -classpath $class_path $args smicli $client_username $client_password $xport://$client_ip:$port $namespace $smis_class $query_type";
        
			$message = "smi_methods(): " . "query is\n" . "$query\n";
			$Log->comment("$message");  
			$results = `$query`; # get query
		} else {

            if (!$smis_class && !$args) {
                $message = "\$smis_class and \$args" . "\n" . "have empty values\n";
            }

            elsif (!$smis_class) {
                $message = "\$smis_class" . "\n" . "has empty value\n";
            }

            elsif (!$args) {
                $message = "\$args" . "\n" . "has empty value\n";
            } else {
                $message = "Incorrect values have been specified in \$smis_class and/or \$args" . "\n";
                $Log->comment($message);
                
            }
            my $msg = "smi_methods(): " . $message . "\n";
            $Log->comment($msg);

        }

        if ($results !~ m/Connection failed/) {

            last;

        } else {

            if ($i == $retry) {
				 my $message = "smi_methods(): " . "Agent $client_hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n" . "\n";
                $Log->error($message);
            }
       }

    }
    $results =~ s/\s+$//;
	$Log->comment($results);
    return $results;
}

=head2 smi_getmethods()

This subroutine will get the method names and parameters of given a class.

B<Usage:> C<$obj-E<gt>smi_getmethods($smis_class, $namespace, $args, $query_type)>  

B<Returns:> All Method Names and Parameters of the class. 

=cut

sub smi_getmethods() {

    my $self = shift;
    my $smis_class = shift;
    my $namespace = shift;
    my $args = shift;
    my $query_type = shift;

    my $dbg = $self->get_debug_module_verbose();

    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $class_path = $self->{_class_path};
    my $java = $self->{_java};
    my $port = $self->{_port};
    my $xport = $self->{_xport};
    my $retry = $self->{_client_retry};
    my $query;
	my $message;

    my $results;

    for (my $i = 0; $i < $retry; $i++) {
		
		if ($smis_class && $args) {
			$query = "$java -classpath $class_path $args smicli $client_username $client_password $xport://$client_ip:$port $namespace $smis_class $query_type";
        
			$message = "smi_getmethods(): " . "query is\n" . "$query\n";
			$Log->comment("$message");  
			$results = `$query`; # get query
		} else {

            if (!$smis_class && !$args) {
                $message = "\$smis_class and \$args" . "\n" . "have empty values\n";
            }
            elsif (!$smis_class) {
                $message = "\$smis_class" . "\n" . "has empty value\n";
            }

            elsif (!$args) {
                $message = "\$args" . "\n" . "has empty value\n";
            } else {
                $message = "Incorrect values have been specified in \$smis_class and/or \$args" . "\n";
                $Log->comment($message);
                
            }

            my $msg = "smi_getmethods(): " . $message . "\n";
            $Log->comment($msg);

        }

        if ($results !~ m/Connection failed/) {

            last;

        } else {

            if ($i == $retry) {
			 my $message = "smi_getmethods(): " . "Agent $client_hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n" . "\n";
                $Log->error($message);
            }
       }

    }

    $results =~ s/\s+$//;
	$Log->comment($results);
    return $results;

}

=head2 smi_indications()

This subroutine will get the create, modify and delete instances.

B<Usage:> C<$obj-E<gt>smi_indications($namespace, $args, $query_type)>  

B<Returns:> indication instances 

=cut

sub smi_indications() {

    my $self = shift;
    my $namespace = shift;
    my $args = shift;
    my $query_type = shift;

    my $dbg = $self->get_debug_module_verbose();

    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $class_path = $self->{_class_path};
    my $java = $self->{_java};
    my $retry = $self->{_client_retry};
    my $query;
    my $message;

    my $results;

    for (my $i = 0; $i < $retry; $i++) {
          			
		if ($args) {
			$query = "$java -classpath $class_path $args smicli $client_username $client_password $client_ip $namespace $query_type";
        
			$message = "smi_indications(): " . "query is indication testing1223----->\n" . "$query\n";
			$Log->comment($message);  
			$results = `$query`; # get query
		} else {

            if (!$args) {
                $message = "\$args" . "\n" . "has empty values\n";
            } else {
                $message = "Incorrect values have been specified in \$args" . "\n";
                $Log->comment($message);
            }

            my $msg = "smi_indications(): " . $message . "\n";
            $Log->comment($msg);

        }

        if ($results !~ m/Connection failed/) {

            last;

        } else {

            if ($i == $retry) {
                my $message = "smi_indications(): " . "Agent $client_hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n" . "\n";
                $Log->comment($message);
               
            }
       }

    }

    $results =~ s/\s+$//;
	$Log->comment($results);
    return $results;

}


=head2 smi_slp_getservices()

This subroutine will get SLP advertised WBEM services. 

B<Usage:> C<$obj-E<gt>smi_slp_getservices($query_type)>  

B<Returns:> A List of all SLP advertised WBEM services.

=cut

sub smi_slp_getservices() {
    my $self = shift;
	my $query_type = shift;
    my $client_hostname = $self->{_client_hostname};
    my $class_path = $self->{_class_path};
    my $java = $self->{_java};
    my $retry = $self->{_client_retry};     
    my $query;
    my $message;
    my $results;
 	
    for (my $i = 0; $i < $retry; $i++) {
		$query = "$java -classpath $class_path smicli $query_type";
		$message = "smi_slp_getservices(): " . "query is\n" . "$query\n";
		$Log->comment("$message"); 
		$results = `$query`; # get query
		
		if ($results !~ m/Connection failed/) {

            last;

        } else {
            if ($i == $retry) {
				 my $message = "smi_slp_getservices(): " . "Agent $client_hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n" . "\n";
                $Log->error($message);
            }
       }
    }
	$results =~ s/\s+$//;
	$Log->comment($results);
    return $results;
}



=head2 smi_slp_gethostaddress()

This subroutine will get SLP advertised WBEM services. 

B<Usage:> C<$obj-E<gt>smi_slp_getservices($query_type)>  

B<Returns:> Get the SLP advertised WBEM services.

=cut

sub smi_slp_gethostaddress() {
    my $self = shift;
	my $query_type = shift;
    my $client_hostname = $self->{_client_hostname};
    my $class_path = $self->{_class_path};
    my $java = $self->{_java};
    my $retry = $self->{_client_retry};     
    my $query;
    my $message;
    my $results;
	my $args = shift;
 	
    for (my $i = 0; $i < $retry; $i++) {
		if ($args) {
			$query = "$java -classpath $class_path $args smicli $query_type";
			$message = "smi_slp_getservices(): " . "query is\n" . "$query\n";
			$Log->comment("$message"); 
			$results = `$query`; # get query
		} else {

            if (!$args) {
                $message = "\$args" . "\n" . "has empty values\n";
            } else {
                $message = "Incorrect values have been specified in \$args" . "\n";
                $Log->comment($message);
            }

            my $msg = "smi_slp_gethostaddress(): " . $message . "\n";
            $Log->comment($msg);
        }
		
		
		if ($results !~ m/Connection failed/) {

            last;

        } else {
            if ($i == $retry) {
				 my $message = "smi_slp_gethostaddress(): " . "Agent $client_hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n" . "\n";
                $Log->error($message);
            }
       }
    }
	$results =~ s/\s+$//;
	$Log->comment($results);
    return $results;
}

=head2 add_stor_sys()

This subroutine will add the storage system to the agent
mentioned in the config object and install it.

B<Usage:> C<$obj-E<gt>add_stor_sys()>  

B<Returns:> Storage system string if successful, FALSE (undef) otherwise. 

=cut

sub add_stor_sys() {
    my $self = shift;
    my $class = shift;
    my $dbg = $self->get_debug_module_verbose();
    my $hostname = $self->{_client_hostname};    
    my $hostip = $self->{_hostip};    
    my $client_username = $self->{_username};
    my $client_password = $self->{_password};
    my $client_ip = $self->{_client_ip};
    my $filer_ip = $self->{_filer_ip};
    my $filer_password = $self->{_filer_password};
    my $filer_hostname = $self->{_filer_hostname};
    my $filer_username = $self->{_filer_username};    
    my $ns = $self->{_vns};
    my $class_path = $self->{_class_path};
    my $retry = $self->{_client_retry};
    my $java = $self->{_java};
    my $xport = $self->{_xport};
    my $port = $self->{_port};
    my $results;
    my $query;
    my $message;

    $query = "$java -classpath $class_path -Dproperty=\"hostName\" smicli $client_username $client_password $xport://$client_ip:$port $ns ONTAP_FilerData ei";

    $message = "User is $client_username and Password is $client_password";
    $Log->debug($message);

    my $counter=0;
    until ($results =~ m/$filer_ip/) {
        $counter++;
        if ($counter <= $retry) {
            $results = `$query`;    # get filerid
            if ($results eq "") {
                $query = "$java -classpath $class_path -DuserName=\"$filer_username\" -Dpassword=\"$filer_password\" -DcomMechanism=\"HTTP\" smicli $client_username $client_password $xport://$client_ip:$port $ns create ONTAP_FilerData";
                $results = system($query);    # add filer
                snooze (5);
            } else {
                $results =~ s/.+hostName\s=\s(.+)/$1/;
                $results =~ s/\s+//g ;
            }
        } else {

            $message = "\"filer_id(): \" . \"Agent $hostname is not responding to:\\n" . $query . "\n" . \"After $retry retries\n\" .$results";
            $Log->debug($message);


            return 1;
        }
    }

    $message = "\"add_stor_sys(): \" .  Dumper($results)";
    $Log->comment($message);

    return $results;
}

=head2 agent_start()

This subroutine will start the agent process described in the config
object.

B<Usage:> C<$obj-E<gt>agent_start()>  

B<Returns:> Boolean, TRUE ('1') if the processes are successfully started, FALSE ('0') otherwise. 

=cut

sub agent_start(){
    my $self = shift;
    my $dbg = $self->get_debug_module_verbose();
    my $product = $self->{_product_name};
    my $agent_proc = $self->{_agent_proc};
    my $agent_bin_path = $self->{_agent_bin_path};
    my $agent_bin_start = $self->{_agent_bin_start};
    my $client_host_type = $self->{_client_host_type};
    my $pid;
    my $message;

    $self->agent_stop();    # Make sure we don't already have a server running

    $message = "Will start $product";
    $Log->comment($message);
    
    if ($client_host_type eq "windows") { 
        # Get the value of $ENV{ProgramFiles} if $hostname is a minwin client?
        my $progfiles = $self->minwin_client("Program Files");       	
        $self->client_exec("$progfiles\\$agent_bin_path\\$agent_bin_start","3600");
        snooze (5);
    } else {    
    	$self->client_exec("cd $agent_bin_path;$agent_bin_path/$agent_bin_start >& /dev/null","3600");
    	snooze (5);
    }
    
    $pid = $self->get_pid($agent_proc);    
    
    if ($pid ne "" && $pid =~ m/\d+/) {
        $message = "Successfully started $product with PID: \"$pid\"";
        $Log->comment($message);
        snooze (10);
    } else {
        $message = "Unable to start $product.";
        $Log->comment("FATAL", $message);
        return 0;
    }

return 1;

}

=head2 agent_stop()

This subroutine will stop the agent process described in the config
object.

B<Usage:> C<$obj-E<gt>agent_stop()>  

B<Returns:> Boolean, TRUE ('1') if the process is successfully killed(or if no processes found), FALSE ('0') otherwise. 

=cut

sub agent_stop(){
    my $self = shift;
    my $dbg = $self->get_debug_module_verbose();
    my $product = $self->{_product_name};
    my $agent_proc = $self->{_agent_proc};
    my $agent_bin_stop = $self->{_agent_bin_stop};
    my $agent_bin_path = $self->{_agent_bin_path};
    my $client_host_type = $self->{_client_host_type};
    my $pid;
    my $message;

    $pid = $self->get_pid($agent_proc);
    
    if ($pid) {
        if ($pid =~ m/[0-9]/) {
            $message = "Will stop $product PID $pid";

            if ($client_host_type eq "windows") { 
			    # Get the value of $ENV{ProgramFiles} if $hostname is a minwin client?
        	    my $progfiles = $self->minwin_client("Program Files");
                $self->client_exec("$progfiles\\$agent_bin_path\\$agent_bin_stop",3600);
    	    } else {
        	    $self->client_exec("kill $pid","3600");
    	    }
    	    
    	    snooze (10);
    	    
            $pid = $self->get_pid($agent_proc);
                           	
    	    if ($pid && $pid =~ m/[0-9]/) {
        	    $message = "Failed to kill $product process with PID: \"$pid\"";
                $Log->comment($message);
                return 0;                            
            } else {
        	    $message = "Killed $product process";
                $Log->comment($message);             
                return 1;
            }
        }
    } else {
        $message = "No $product process to stop.";
        $Log->comment($message);             
        return 1;
    }
}

=head2 cimom_ver()

This subroutine will get the software version of filer mentioned in the config
object.

B<Usage:> C<$obj-E<gt>cimom_ver($namespace, $class)>

B<Returns:> CIMOM version if successful, FALSE (undef) otherwise.

=cut

sub cimom_ver() {
    my $self = shift;
    my $namespace = shift;
    my $class = shift;
    my $dbg = $self->get_debug_module_verbose();
    my $client_ip = $self->{_client_ip};
    my $hostname = $self->{_client_hostname};
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $class_path = $self->{_class_path};
	if (!$namespace || $namespace eq ""){ $namespace = $self->{_ins} };
	if (!$class || $class eq ""){ $class = "ONTAP_SMIAgentSoftware" };
    my $retry = $self->{_client_retry};
    my $xport = $self->{_xport};
    my $port = $self->{_port};
    my $java = $self->{_java};
    my $query;

    my $results = 0;
    my $counter = 0;
    until ($results =~ m/VersionString/) {
        $counter++;
        if ($counter <= $retry) {
            $query = "$java -classpath $class_path -Dproperty=\"VersionString\" smicli $client_username $client_password $xport://$client_ip:$port $namespace $class ei";
            $results = `$query`;    # get ONTAP_SMIAgentSoftware
            snooze (5);
        } else {
            logcomment("cimom_ver(): " . "Agent $hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n" . Dumper($results));
            return 1;
        }
    }

    $results =~ s/VersionString\s=\s(.+)/$1/;
    $results =~ s/\s+//g;

    logcomment("cimom_ver(): " .  Dumper($results));    

    return $results;
}

=head2 client_exec()
Execute commands on a client host
B<Usage:> C<$obj-E<gt>client_exec($cmd, $timeout)>  

B<Returns:> Result of command.

=cut

sub client_exec() {
    my $self = shift;
    my $cmd = shift;
    my $timeout = shift;    
    my $params = NATE::ParamSet->new(global=>1);
    my $client = $params->get('CLIENT') || die('Please set CLIENT=<HostID of client>');

    my $host_obj = Hostrec->new(id=>$client);
    my $connobj = Connectrec->new(host => $host_obj, template => "default");
    my $TIMEOUT = $params->get("TIMEOUT") || "$timeout";

    my $cmd_results = $connobj->execute(command => "$cmd",
                                        timeout => $TIMEOUT,
                                        check_status => 1);

    $cmd_results =~ s/__status__=.+//;

    my $dconn = $connobj->disconnect();
    my $conntype = $connobj->conntype();
    my @prompts = $connobj->prompts();
    $Log->debug($conntype,@prompts);

    my $message = "client_exec(): Executed command $cmd.\n\n";
    $Log->comment($message);

    return $cmd_results;
}

=head2 client_os_details()

This subroutine will get the version of the O/S on the client
mentioned in the config object.

B<Usage:> C<$obj-E<gt>client_os_details()>  

B<Returns:> Client O/S version string if successful, FALSE (undef) otherwise. 

=cut

sub client_os_details() {
    my $self = shift;
    my $cinfo = shift;
    my @retval;
    my $dbg = $self->get_debug_module_verbose();
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $client_host_type = $self->{_client_host_type};
    my $hostname = $self->{_client_hostname};     
    my $client;
    my $logcomment;
    my @query;
    my $query;

    if ($client_host_type eq "windows") {
        $client = "windows";
    }
    elsif ($client_host_type eq 'unix-linux-redhat' || 'unix-linux-sles10' || 'unix-solaris' || 'unix-sunos') {
        $client = "unix";
    } else {
        logresult('FATAL',"$hostname '$client_host_type' is not a supported O/S");
    }
    
    if ( $cinfo eq 'os' ) {
        if ($client_host_type eq "windows") {
            @retval = "windows";
        }
        elsif ($client_host_type eq "unix-linux-redhat") {
            @retval = "linux";
        }
        elsif ($client_host_type eq "unix-linux-sles10") {
            @retval = "linux";
        }
        elsif ($client_host_type eq "unix-solaris") {
            @retval = "solaris";
        }
        elsif ($client_host_type eq "unix-sunos") {
            @retval = "solaris";
        }
        $logcomment = "nate hosttype: ";
    }

    if ( $cinfo eq 'arch' ) {
        my $arch;
        if ( $client eq "unix"  ) {
            $arch = "/bin/uname -p";
            @query = $self->client_exec($arch,3600);    # get client architecture
            $query[0] =~ s/\s//; 
            @retval = $query[0];
        }    
        elsif ( $client eq "windows" ) {
            $arch = $self->wmic("Win32_OperatingSystem","OSArchitecture");
            @query = split('\n',$arch);
            
            foreach $query(@query) {
                next if ($query =~ m/Win32_OperatingSystem/);
                next if ($query =~ m/OSArchitecture/);
                push(@retval,$query); 
            }
        }
        
        $logcomment = "Architecture: ";
    }

    if ( $cinfo eq 'version' ) {
        my $version;
        if ( $client eq "unix"  ) {
            my $version = "/bin/cat /etc/*-release";
            #$version = File::Spec->catfile('/etc/*-release');
            my @version = $self->client_exec($version,3600);    # get client distribution
            $version[0] =~ s/\s//; 
            @retval = $version[0];
        }    
        elsif ( $client eq "windows" ) {
            $version = $self->wmic("Win32_OperatingSystem","Version");
            @query = split('\n',$version);
         
            foreach $query(@query) {
                next if ($query =~ m/Win32_OperatingSystem/);
                next if ($query =~ m/Version/);
                push(@retval,$query); 
            }
        }
        
        $logcomment = "Version: ";
    }
    
    if ( $cinfo eq 'platform' ) {
        my $platform;
        if ( $client eq "unix"  ) {
            my $platform = "/bin/cat /etc/*-release";
            my @platform = $self->client_exec($platform,3600);    # get client distribution
            $platform[0] =~ s/\s//; 
            @retval = $platform[0];
        }
        elsif ( $client eq "windows" ) {
            $platform = $self->wmic("Win32_OperatingSystem","Version");
            @query = split('\n',$platform);
            
            foreach $query(@query) {
                next if ($query =~ m/Win32_OperatingSystem/);
                next if ($query =~ m/Version/);
                push(@retval,$query); 
            }
        } 
        $logcomment = "Platform: ";
    }

    if ( $cinfo eq 'distro' ) {
        my $distro;
        if ( $client eq "unix"  ) {
            my $distro = "/bin/cat /etc/*-release";
            my @distro = $self->client_exec($distro,3600);    # get client distribution
            $distro[0] =~ s/\s//; 
            @retval = $distro[0];
        }
        elsif ( $client eq "windows" ) {
            $distro = $self->wmic("Win32_OperatingSystem","Caption");
            @query = split('\n',$distro);
            
            foreach $query(@query) {
                next if ($query =~ m/Win32_OperatingSystem/);
                next if ($query =~ m/Caption/);
                push(@retval,$query); 
            }
        } 
        $logcomment = "Distribution: ";
    }

    our $str = join(' ', @retval);

    $logcomment = "\n\nclient_os_details(): $logcomment $str\n\n";
    $Log->comment($logcomment);

    chomp($str);

    return
    $str;
}

=head2 smi_enumerate()

This subroutine will get Enumerates all instances and names of a given CIM class. 

B<Usage:> C<$obj-E<gt>smi_enumerate(<$smis_class>, <$namespace>, <$query_type>, [args], [connection])>  

B<Returns:> A List of the intance paths and names returned by the enumeration.

=cut

sub smi_enumerate() {
    my $self = shift;
    my $smis_class = shift;
    my $namespace = shift;
    my $query_type = shift;
    my $args = shift;
    my $connection = shift;

    my $dbg = $self->get_debug_module_verbose();

    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $class_path = $self->{_class_path};
    my $java = $self->{_java};
    my $retry = $self->{_client_retry};     
    my $query;
    my $message;
    my $results;
    my $xport;
    my $port;    
    
   
    if ($connection && $connection eq "secure") {
        $xport = $self->{_sxport};
        $port = $self->{_sport};     
    } else {
        $xport = $self->{_xport};
        $port = $self->{_port};     
    }

    for (my $i = 0; $i < $retry; $i++) {

            if ($args) {
                $query = "$java -classpath $class_path $args smicli $client_username $client_password $xport://$client_ip:$port $namespace $smis_class $query_type";
            } else {
                $query = "$java -classpath $class_path smicli $client_username $client_password $xport://$client_ip:$port $namespace $smis_class $query_type";
            }
            print "\n##$query##\n";
			my $message = "smi_enumerate(): " . "query is\n" . "$query\n";
            #$Log->comment($message);

            $results = `$query`; # get query
            $message = "\nenumerate(): " . "query is\n" . $query . "\n". "results is\n" . $results . "\n";

        if ($results !~ m/[Connection failed|Unable to connect]/) {

            last;

        } else { 
			
			if ($i == $retry) {
                $message = "smi_enumerate(): " . "Agent $client_hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n" . $results . "\n";
                last;
            }
       }

    }
    
    $Log->comment($message);
    $results =~ s/\s+$//;
	$Log->comment($results);
    return $results;

}

=head2 filer()
Add, verify and delete storage systems on a SMI-S provider
B<Usage:> C<$obj-E<gt>filer($request, $client_hostname, $filer_ip, $filer_hostname, $retry, $java)>  

B<Returns:> Boolean, TRUE ('1') if the storage system exists or is added, FALSE ('0') otherwise.


=cut

sub filer() {
    my $self = shift;
    my $request = shift;
    my $client_hostname = $self->{_client_hostname};
    my $filer_ip = $self->{_filer_ip};
    my $filer_hostname = $self->{_filer_hostname};
    my $retry = $self->{_client_retry};
    my $java = $self->{_java};
    my $results;
    my $message;

    # Add the storage system defined in testbed.info

    for (my $i = 0; $i < $retry; $i++) {
        if ($request eq "add") {

            # Verify if the filer exists before adding
            my $verify = &verify_ss($self);
            snooze(20);

            if (!$verify || $verify !~ m/$filer_ip/) {

                # Filer does not exist so add
                $results = &add($self);

                my $verify_results = &verify_ss($self);
                
                if ($verify_results !~ m/$filer_ip/) {
                    $message = "filer(): " . "The $request request failed for $filer_ip.\n";
                    $Log->error($message); 
                }

            } else {
                # Filer already exists
                    $message = "filer(): " . "Storage system $filer_ip exists in the repository.\n";
                    $Log->comment($message);
                    last;
            }

        }elsif  ($request eq "verify") {
         
            $results = &verify_ss($self);
            return $results;
            
        }elsif  ($request eq "delete") {
         
            $results = &delete($self);
            return $results;
            
        } else {

            $message = "filer(): " . "\"$request\" is not a valid option" . "\n";
            $Log->comment($message);
            die Dumper($request);

        }

        if ($results) {

            last;

        } else {

            if ($i == $retry) {
                $message = "filer(): " . "Agent $client_hostname is not responding to:\n" . $request . "request\n" . "After $retry retries" . "\n";
                $Log->comment($message);
                die Dumper($results);
            }
        }
    }

    sub add {
        my $self = shift;
        my $class = "ONTAP_FilerData";
        my $option = "create";
        my $ns = $self->{_vns};
        my $filer_ip = $self->{_filer_ip};
        my $f = $self->{_filer_connect};
        my $filer_username = $f->username();
        my $filer_password = $f->password();
        my $args = "-DhostName=\"$filer_ip\" -Dport=\"80\" -DuserName=\"$filer_username\" -Dpassword=\"$filer_password\" -DcomMechanism=\"HTTP\"";
        my $results = $self->smi_instance($class,$option,$ns,$args);
        return $results;
    }

    sub verify_ss {
        my $self = shift;
        my $filer_ip = $self->{_filer_ip};
        my $args = "-Dss=\"$filer_ip\" -Dproperty=\"hostName\"";
        my $class = "ONTAP_FilerData";
        my $ns = $self->{_vns};
        my $query = "ei";
        my $results = $self->smi_enumerate($class,$ns,$query,$args);
        $results =~ s/hostName\s=\s//;
        return $results;
    }

    sub delete {
        my $self = shift;
        my $class = "ONTAP_FilerData";
        my $option = "delete";
        my $ns = $self->{_vns};
        my $filer_ip = $self->{_filer_ip};
        my $args = "-DONTAP_FilerData.hostName=\"$filer_ip\" -Dport=\"80\"";
        my $results = $self->smi_instance($class,$option,$ns,$args);
        return $results;
    }
        return 0;
}

=head2 filer_cmd()

This subroutine will return the output of any ONTAP command

B<Usage:> C<$obj-E<gt>filer_cmd()>  

B<Returns:> ONTAP command output if successful, FALSE (undef) otherwise. 

=cut

sub filer_cmd() {
    my $self = shift;
    my $cmd = shift;
    my $dbg = $self->get_debug_module_verbose();
    my $c = connect("$self->{_filer_hostname}#rsh-persay");
    my @r = say($c, "$cmd");
    
    my $logcomment = "'Connect Object in filer_cmd() is:\n' . Dumper($c) .  '\n\n$cmd command returned\n' . Dumper(@r))";
    $Log->debug($logcomment);

    my $str = $r[2];
    chomp($str);
    return $str;
}

=head2 filer_id()

This subroutine will get the id of filer via SMI-S mentioned in the config object.

B<Usage:> C<$obj-E<gt>filer_id()>  

B<Returns:> Filer id if successful, FALSE (undef) otherwise. 

=cut

sub filer_id() {
    my $self = shift;
    my $namespace = shift;
    my $class = shift;
    my $query = shift;
    my $args = shift;
    my $hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $username = $self->{_client_username};
    my $password = $self->{_client_password};
    my $class_path = $self->{_class_path};
    if ($namespace eq "") { $namespace = $self->{_vns} };
    if ($class eq "") { $class = "ONTAP_StorageSystem" };    
    my $retry = $self->{_client_retry};
    my $java = $self->{_java};
    my $logcomment;
    my $message;

    my $results = 0;
    my $counter = 0;
    until ($results =~ m/ONTAP:/) {
        $counter++;
        if ($counter <= $retry) {
            $query = $self->smi_enumerate($query,$ns,$class,$args);
            $results = `$query`;    # get filerid
            snooze (5);
        } else {
            $message = "filer_id(): " . "Agent $hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n";
            $Log->debug($message);
            
            Dumper($results);
             
            return 1;
        }
    }

    $results =~ s/.+ONTAP:(.+)/$1/;
    $results =~ s/\s+//g ;

    $message = "filer_id(): $results";
    $Log->comment($message);
     
    return $results;
}
    
=head2 filer_version()

This subroutine will get the version of ONTAP running on the filer
mentioned in the config object.

B<Usage:> C<$obj-E<gt>filer_version()>  

B<Returns:> Filer version string if successful, FALSE (undef) otherwise. 

=cut

sub filer_version() {
    my $self = shift;
    my $dbg = $self->get_debug_module_verbose();
    my $c = connect("$self->{_filer_hostname}#rsh-persay");
    my @r = say($c, "version");
    
    my $message = "'\n\nVersion command returned\n\n' . @r'";
    $Log->comment($message);
    
    my $str = $r[2];
    chomp($str);
    return $str;
}

=head2 get_pid()

This subroutine will stop the agent process described in the config
object.

B<Usage:> C<$obj-E<gt>get_pid("process_name")>  

B<Returns:> Process id of supplied process name. 

=cut

sub get_pid() {
    my $self = shift;
    my $proc_name = shift;
    my $client_host_type = $self->{_client_host_type};
    my $pid;
    my @retval;
    my $results;
    my @query;
    my $query;

    if ($client_host_type eq "windows") {
        $results = $self->wmic("Win32_Process","Processid","Caption","like","\%$proc_name\%"); # Get PID
        
        if ($results ne "") {
            @query = split('\n',$results);
            
            foreach $query(@query) {
                next if ($query =~ m/Win32_Process/i);
                next if ($query =~ m/Processid/i);
                $query =~ s/.+\W+(.+)/$1/g;
                push(@retval,$query); 
            }
            
            $pid = join(' ', @retval);
        }
    } else {
        $results = $self->client_exec("pgrep $proc_name","3600"); # Get PID

    	if ($results ne "") {
    	    @query = split('\n',$results);
    	    
    	    foreach $query(@query) {
    	        next if ($query !~ m/\d+/);
    	        $query =~ s/[\r|\n]//g;  # Strip newlines
    	        $query =~ s/\s+//g;  # Strip extra spaces
    	        push(@retval,$query);
    	    }
    	}
    	
    	$pid = join(' ', @retval);
    }
    return $pid;
}

=head2 install_pkg()

B<Usage:> C<$obj-E<gt>install_pkg($package, $pkg_xtension, $client_hostname, $client_os)>  

B<Returns:> Install package. 

=cut

sub install_pkg() {
    my $self = shift;
    my $package = shift;
    my $pkg_xtension = $self->{_pkg_xtension};
    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_os = $self->{_client_host_type};
    my $pkg_installer = $self->{_pkg_installer};
    my $pkg;
    my $pkg_path;
    my $install;

    if ($client_os eq "windows") {

        $package =~ s/(.+\\)(.+\.$pkg_xtension)/$1$2/;

        $pkg_path = $1;
        $pkg = $2;

        $install = $self->client_exec("wmic.exe /node:\'$client_ip\' product call install true,\'\' ,\'$package\'","3600");
        return($install);
    } else {
        $package =~ s/\\/\//;
        $package =~ s/(.+)\/(.+\.$pkg_xtension)//;

        $pkg_path = $1;
        $pkg = $2;

        $install = $self->client_exec("cd $pkg_path;rm tar_image*; /bin/tar xvf $pkg;$pkg_path/install_smisproxy -k","3600");
        return($install);
    }

    my $message = "install_pkg(): " . "Package $pkg has been installed on host $client_hostname" . "\n";
    $Log->comment($message);    
}

=head2 smi_instance()
Create, modify & delete instances on a SMI-S provider
B<Usage:> C<$obj-E<gt>smi_instance($class, <create|modify|delete>, $name_space, $args, $client_ip, $client_hostname, $client_password)>  

B<args example:>-DhostName=\"$self->{filer_ip}\" -Dport=\"80\" -DuserName=\"$self->{filer_username}\" -Dpassword=\"$self->{filer_password}\" -DcomMechanism=\"HTTP\"
B<Returns:> The path of the create | modify | delete instance.

=cut

sub smi_instance() {
    my $self = shift;
    my $class = shift;
    my $option = shift;
    my $ns = shift;
    my $args = shift;
    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $retry = $self->{_client_retry};
    my $java = $self->{_java};
    my $class_path = $self->{_class_path};
    my $xport = $self->{_xport};
    my $port = $self->{_port};
    my $message;
    my $instance;
    my $results;
    my $operator;
    my $dumper_msg = "";

    for (my $i = 0; $i < $retry; $i++) {
        if ($class && $args) {

            if ($option eq "create") {
                $operator = "ci";
            }
            elsif ($option eq "modify") {
                $operator = "mi";
            }
            elsif ($option eq "delete") {
                $operator = "di";
            }

            $instance = "$java -classpath $class_path $args smicli $client_username $client_password $xport://$client_ip:$port $ns $class $operator 2>&1";
            $results = system($instance); # exec query
            snooze(20);

        } else {

            if (!$class && !$args) {
                $message = "\$class and \$args" . "\n" . "have empty values\n";
            }

            elsif (!$class) {
                $message = "\$class" . "\n" . "has empty value\n";
            }

            elsif (!$args) {
                $message = "\$args" . "\n" . "has empty value\n";
            } else {
                $message = "Incorrect values have been specified in \$class and/or \$args" . "\n" . "$class\n$args" . "\n";
                $Log->error($message);
            }

            my $msg = "smi_instance(): " . $message . "\n";
            $Log->comment($msg);            

        }

        if ($results ne "") {

            last;

        } else {

            if ($i == $retry) {                
                $message = "smi_instance(): " . "Agent $client_hostname is not responding to:\n" . $instance . "\n" . "After $retry retries" . "\n";
                $Log->comment($message);
            }

        }

    }

    $message = "smi_instance(): $option instance of $class on agent $client_hostname \n";
    $Log->comment("$message\n$instance\n$results\n");    
    
    return $results;
}

=head2 smi_instances()

This subroutine will get the create, modify and delete instances.

B<Usage:> C<$obj-E<gt>smi_instances($smis_class, $namespace, $args, $query_type)>  

B<Returns:> instances(Traverses an association) and instances of the results class. 

=cut

sub smi_instances() {

    my $self = shift;
    my $smis_class = shift;
    my $namespace = shift;
    my $args = shift;
    my $query_type = shift;

    my $dbg = $self->get_debug_module_verbose();

    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $class_path = $self->{_class_path};
    my $java = $self->{_java};
    my $port = $self->{_port};
    my $xport = $self->{_xport};
    my $retry = $self->{_client_retry};
    my $query;
    my $message;

    my $results;

    for (my $i = 0; $i < $retry; $i++) {
          			
		if ($smis_class && $args) {
			$query = "$java -classpath $class_path $args smicli $client_username $client_password $xport://$client_ip:$port $namespace $smis_class $query_type";
        
			$message = "smi_instances(): " . "query is venkat testing----->\n" . "$query\n";
			$Log->comment("$message");  
			$results = `$query`; # get query
		} else {

            if (!$smis_class && !$args) {
                $message = "\$smis_class and \$args" . "\n" . "have empty values\n";
            }

            elsif (!$smis_class) {
                $message = "\$smis_class" . "\n" . "has empty value\n";
            }

            elsif (!$args) {
                $message = "\$args" . "\n" . "has empty value\n";
            } else {
                $message = "Incorrect values have been specified in \$smis_class and/or \$args" . "\n";
                $Log->comment($message);
            }

            my $msg = "smi_instances(): " . $message . "\n";
            $Log->comment($msg);

        }

        if ($results !~ m/Connection failed/) {

            last;

        } else {

            if ($i == $retry) {
                my $message = "smi_instances(): " . "Agent $client_hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n" . "\n";
                $Log->comment($message);
               
            }
       }

    }

    $results =~ s/\s+$//;
	$Log->comment($results);
    return $results;

}

=head2 minwin_client()
Verify if Windows host is MinWin compatible.
B<Usage:> C<$obj-E<gt>minwin_client([$request])>  

B<Returns:> 1 if MinWin client.  Location of Program Files if requested.

=cut

sub minwin_client() {
    my $self = shift;
    my $request = shift;
    my $dbg = $self->get_debug_module_verbose();
    my $hostname = $self->{_client_hostname};    
    my $arch = $self->client_os_details("arch");
    my $client_host_type = $self->{_client_host_type};
    my $progfiles;
    my $message;
	my $minwin_check = "";
            
    if ($client_host_type eq "windows") {
	    # Is this a minwin client?
	    $minwin_check = minwin_compatible($hostname);
        
        if ( $minwin_check eq "1" ) {
            $message = "Client $hostname is minwin compatible";
            $Log->debug($message);

    	    if ($request) {
    	        if ($request eq "Program Files"){
    		        # Get the location of $ENV{ProgramFiles}
                    if ($arch =~ m/32-bit/) {
                        $progfiles = $self->client_exec("set ProgramFiles(x86)","60"); 
                    }
                    elsif ($arch =~ m/64-bit/) {
                        $progfiles = $self->client_exec("set ProgramFiles(x86)","60");                    
                    } else {
                        $message = "\"minwin_client(): \" . \"The architecture '$arch' does not match supported architecture types.\" . \"\n\"";
                        $Log->error($message);
                    }          
                    if ($progfiles eq "") {
                        $message = "\"minwin_client(): \" . \"Cannot find the Program Files directory\" . \"\n\"";
                        $Log->error($message);
                    }
                
                    $progfiles =~ s/.+=(.+)/$1/;
                    $progfiles =~ s/\n+//g;
                    $progfiles =~ s/\r+//g;                    
                    $progfiles =~ s/(\w{1}\:\\)(.+)/$1"$2"/; 
                    return $progfiles;
    	        }
            }          
        } else {
    	    my $minwinurl = "http://wikid.netapp.com/w/HHG/Software-4.0/".
    		    "Tools-7.0/NATE-9.1/User_Guides-2.1/MinWin_Windows_Clients";
    	    $message = "Cannot connect to client: $hostname.\nPlease refer to $minwinurl.";	
    	    $Log->error($message);
	    }
	    
	    return(1);
	
    } else {
        # Client is a UNIX host... return 1
    }   return(1);
}

=head2 object_path()

Gets an array of object paths from the provider

B<Usage:> C<$obj-E<gt>object_path($class)>  

B<Returns:> Object path if successful, FALSE ('0') otherwise. 

=cut

sub object_path() {
    my $self = shift;
    my $class = shift;
    my $dbg = $self->get_debug_module_verbose();
    my $hostname = $self->{_client_hostname};    
    my $hostip = $self->{_hostip};    
    my $username = $self->{_username};
    my $password = $self->{_password};
    my $ns = $self->{_vns};
    my $class_path = $self->{_class_path};
    my $retry = $self->{_client_retry};
    my $java = $self->{_java};
    my $objpath;
    my $results;
    my $query;
    my $message;

    $message = "User is $username and Password is $password";
    $Log->debug($message);

    my $filer_ip = $self->{_filer_ip};    # get filer_ip
    my $filer_id = $self->filer_id("root/ontap","ONTAP_StorageSystem","ei","-Dss=$filer_ip");    # get filer_id

    my $counter=0;
    until ($results =~ m/ONTAP:/) {
        $counter++;
        if ($counter <= $retry) {  
            $query = "$java -classpath $class_path smicli $hostip $username $password http 5988 $ns ONTAP_StorageSystem ni";
            $results = `$query`;    # get ONTAP_SMIAgentSoftware
            snooze (5);
        } else {
            $message = "\"object_path(): \" . \"Agent $hostname is not responding to:\n\" . $query . \"\n\" . \"After $retry retries\n\"";
            $Log->debug($message);
            die Dumper($results);
        }
    }

    $message = "object_path(): " . Dumper($results);
    $Log->debug($message);
		
    return $objpath;
}

=head2 pkg()
B<Usage:> C<$obj-E<gt>pkg($client_hostname)>  

B<Returns:> Downloaded package location.

=cut

sub pkg() {
    my $self = shift;
    my $client_hostname = $self->{_client_hostname};
    my $force_install = $self->{_force_install};
    my $new_rlse_installer = "";
    my $current_rlse = "";
    my $prodcode;
    my $download_path;
    my $message;
    my $verify_pkg = "";
    my $chk_rm_pkg = "";
    my $chk_inst_pkg = "";
    my $installed_rlse = "";
    my $downloaded_rlse;

    # Get the name of the release installer defined in RLSE_URL
    $new_rlse_installer = $self->rlse("Package");
    
    # Verify if a rlse was previously downloaded
    $downloaded_rlse = &download_chk($self);
        
    if (!$downloaded_rlse || $downloaded_rlse ne $new_rlse_installer) {
        # Download release defined in RLSE_URL
        $download_path = &download($self);
    } else {
        $download_path = $downloaded_rlse;
    }

    # Verify if a rlse is installed
    $installed_rlse = &verify_pkg($self);

    # The requested release is currently installed 
    if ($installed_rlse && $installed_rlse eq $new_rlse_installer) {
     
        if ($force_install eq "true") {
            # FORCE_INSTALL option is set to true so remove currently installed release
            $self->install_sequence("$client_hostname","$download_path","$new_rlse_installer","old","$installed_rlse");
        } else {
            # Do not download or install package if it is already installed
            $message = "$new_rlse_installer is already installed on $client_hostname and will not be re-installed.";
        }
    }
    elsif (!$installed_rlse || $installed_rlse eq "") {
        # A release is not installed
        $self->install_sequence("$client_hostname","$download_path","$new_rlse_installer","new");
    } else {
        # A release not matching $new_rlse_installer is installed
        $self->install_sequence("$client_hostname","$download_path","$new_rlse_installer","old","$installed_rlse");
    }

    # Local methods

    sub download() {
        # Download release defined in RLSE_URL
        my $self = shift;
        my $download_path = $self->rlse("Download");
        return $download_path;
    }
    
    sub download_chk() {
        # Verify if a package was previously downloaded
        my $self = shift;
        my $download_chk = $self->rlse("Check");
        $download_chk =~ s/\s+//;
        return $download_chk;
    }

    sub install() {
        # Install release defined in RLSE_URL
        my $self = shift;
        my $download_path = shift;
        $self->install_pkg($download_path);
    }
    
    sub install_sequence() {
        my $self = shift;
        my $client_hostname = shift;
        my $download_path = shift;
        my $new_rlse_installer = shift;
        my $install_option = shift;
        my $installed_rlse = shift;
        my $chk_inst_pkg;
        my $chk_rm_pkg;
        my $message;

        if ($install_option eq "new") {
         
            &install($self,$download_path);
            
        }
        elsif ($install_option eq "old") {
            $self->services_stop();
        
            $message = "pkg(): " . "$installed_rlse will be uninstalled on $client_hostname." . "\n";
            $Log->comment($message);
            &remove($self);
        
            # Verify if package was removed  
            $chk_rm_pkg = &verify_pkg($self);

            # Package removed
            if (!$chk_rm_pkg || $chk_rm_pkg eq "" || $chk_rm_pkg =~ m/"No such file or directory"/) {
                # Install release defined in RLSE_URL
                &install($self,$download_path);
            }
        }
         
        # Verify if a rlse is installed
        $chk_inst_pkg = &verify_pkg($self);

        if ($chk_inst_pkg && $new_rlse_installer =~ m/$chk_inst_pkg/i) {
            $message = "pkg(): " . "$new_rlse_installer was installed on $client_hostname.";
            $Log->comment($message);
        } else {
            $message = "pkg(): " . "$new_rlse_installer was not installed on $client_hostname." . "\n\n";
            $Log->comment($message);
            die Dumper($chk_inst_pkg);
        } 
    }
    
    sub remove() {
        # Obtain the msi product code for the currently installed release
        my $self = shift;
        my $prodcode = $self->pkg_info("ProductCode");

        # Remove currently installed release
        $self->remove_pkg($prodcode);
    }
    
    sub verify_pkg() {
        # Verify if a package is installed
        my $self = shift;
        my $verify_pkg = $self->pkg_info("Package");
        return $verify_pkg;
    }
    
    #my $msg = "pkg(): " . $message . "\n";
    #$Log->comment($msg); 
    
    return 1;
}

=head2 pkg_info()

B<Usage:> C<$obj-E<gt>pkg_info($property)>  
$property = Package || ProductCode

B<Returns:> Requested property value from installed package. 

=cut

sub pkg_info() {
    my $self = shift;
    my $property = shift;
    my $pkg_xtension = $self->{_pkg_xtension};
    my $client_hostname = $self->{_client_hostname};
    my $product_name = $self->{_product_name};
    my $client_os = $self->{_client_host_type};
    my $ver_file = $self->{ _ver_file};
    my $prod;
    my $prod_query;
    my @prod_query_results;
    my @prod;
    my $property_res;
    my $message;
    my $logcomment = "";
    my $line;

    if ($client_os eq "windows") {

        # WMI query to obtain package details
        $prod_query = $self->wmic("Win32_Product","PackageName","name","=","$product_name");
        
        # Proceed if the WMI query returns a value
        if ($prod_query) {

            # Split the WMI query results
            @prod_query_results = split(/\n/,$prod_query); 

            foreach $line(@prod_query_results) {
                next if ($product_name !~ m/$line/);

                # Obtain the line that matches $product_name
                if ($property eq "Package") {
                    # Obtain the installed package name
                    $line =~ s/.+\|(.+$pkg_xtension)\|.+/$1/g;
                    if ($line eq "") {
                        $logcomment = "\n$product_name is not installed on $client_hostname\n";
                    } else {
                        $logcomment = "\n$product_name $line is installed on $client_hostname\n";
                    }
                }
                if ($property eq "ProductCode") {
                    # Obtain the package Product Code
                    $line =~ s/({.+}).+\|(.+$pkg_xtension)\|.+/$1/g;
                    $line = $1;
                    my $installed_product = $2;
                    $logcomment = "\nProduct code is $line for $installed_product\n";
                }
                
                push(@prod,$line);
            }
        }
            
    } else {
        # Add Linux code here.
        if ($property eq "Package") {
            $ver_file = "/bin/cat /usr/ontap/smis/pegasus/version.txt 2>&1";
            $prod_query = $self->client_exec($ver_file,3600);    # get client distribution
            # Proceed if the query returns a value
            if ($prod_query) {
                # Split the query results
                @prod_query_results = split(/\n/,$prod_query);                             
                
                foreach $line(@prod_query_results) {
                    next if ($line !~ m/Release/);
                    $line =~ s/Release version:\s(.+)/$1/g;
                    $line =~ s/\s+$//;
                    # Replacing "." with "-" to match the installer naming convention
                    $line =~ s/\./\-/g;
                    $line = $line.".".$pkg_xtension;
                                        
                    push(@prod,$line);
                }
            }
            if (!@prod || @prod eq "") {
                $logcomment = "\n$product_name is not installed on $client_hostname\n";
            } else {
                $logcomment = "\n$product_name @prod is installed on $client_hostname\n";
            }    
        }
    }
    
    $property_res = $prod[0];
    $message = "pkg_info(): " . $logcomment . "\n\n";
    $Log->comment($message);

    return $property_res;
}

=head2 remove_pkg()

B<Usage:> C<$obj-E<gt>remove_pkg($prodcode)>  

B<Returns:> Removes package using the MSI Product Code number.

=cut

sub remove_pkg() {
    my $self = shift;
    my $prodcode = shift;
    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_os = $self->{_client_host_type};
    my $pkg_remover = $self->{_pkg_remover};
    my $product = $self->{_product_name};
    my $current_rlse = $self->pkg_info("Package");
    my $logcomment = "";
    my $uninstall;

    if ($client_os eq "windows") {
        if ($prodcode) {
            # Remove package
            $uninstall = $self->client_exec("wmic.exe /node:\"$client_ip\" product where name='$product' call uninstall /nointeractive","3600");
            my @uninstall = split('\n',$uninstall);

            foreach my $line(@uninstall) {
                next if (!($line =~ m/ReturnValue/i));
                $line =~ s/.+(\d+).+/$1/;

                if ($line =~ m/0/i) {
                    $logcomment = "Package $current_rlse removed on host $client_hostname";
                } else {
                    $logcomment = "Package $current_rlse was not removed on host $client_hostname\nReturnValue = $line";
                }
            };

        } else {

            $logcomment = "Package \$prodcode required to uninstall $product";
            die;
        }

    } else {
        # Remove package
        $uninstall = $self->client_exec("/usr/ontap/smis/pegasus/bin/uninstall_smisproxy -s","3600");
    }

    my $message = "remove_pkg(): " . $logcomment  . "\n";
    $Log->comment($message);
    
    return 1;
}

=head2 rlse()

B<Usage:> C<$obj-E<gt>rlse($reguest)>  
B$request = Package || Download

B<Returns:> Downloaded package location. 

=cut

sub rlse() {
    my $self = shift;
    my $request = shift;
    my $client_hostname = $self->{_client_hostname};
    my $project = $self->{_project};
    my $rlse_url = $self->{_rlse_url};
    my $pkg_xtension = $self->{_pkg_xtension};
    my $client_os = $self->{_client_host_type};
    my $downloader_list = $self->{_downloader_list};
    my $downloader_get = $self->{_downloader_get};

    my $pkg;
    my @pkg;
    my $map;
    my $download_location;
    my $downloader_query;
    my $pkg_chk;
    my $get_pkg;
    my $package;
    my $logcomment = "";
    my $message;
    my $temp;

    if ($client_os eq "windows") {
        $downloader_query = $self->client_exec("$downloader_list $rlse_url","3600");
    } else {
        $downloader_query = `$downloader_list $rlse_url 2>&1`;
    }

    my @downloader_out = split('\n',$downloader_query);

    foreach $pkg(@downloader_out) {
        next if (!($pkg =~ /.$pkg_xtension/));
        $pkg =~ s/.+\<a href=.+\>(.+)\<\/a\>.+/$1/;
        push(@pkg,$pkg);
    }

    if ($pkg[0] !~ m/$pkg_xtension/) {
        # Invalid url path was defined in RLSE_URL
        $message = "rlse(): " . "Package with extension \"$pkg_xtension\" was not found in \"$rlse_url\"" . "\n";
        $Log->comment($message);
        die Dumper($pkg[0]);
    }

    if ($request eq "Package") {
        $package = $pkg[0];
    }
    
    if ($request eq "Check") {
        if ($client_os eq "windows") {
            $temp = $self->client_exec("set TEMP","60");
            $temp =~ s/.+=(.+)/$1/g;
            $temp =~ s/\s+//g;         
            $download_location = "$temp\\$project";
            $package = $self->client_exec("dir /B $download_location\\$pkg[0]","3600");         

            if ($package =~ m/"File Not Found"/i) {
                $package = "";
            } else {
                $package = "$download_location\\$package"; 
            }
        } else {
            $download_location = "/tmp/$project";
            $package = $self->client_exec("ls $download_location/$pkg[0]","3600");
            $package =~ s/\s+$//;        

            if ($package =~ m/"No such file or directory"/i) {
                $package = "";
            } else {
                $package = "$download_location/$package"; 
            }
        }
    }
     
    if ($request eq "Download") {
        
        if ($client_os eq "windows") {
         
            $temp = $self->client_exec("set TEMP","60");
            $temp =~ s/.+=(.+)/$1/g;
            $temp =~ s/\s+//g;         
            $download_location = "$temp\\$project";
            $pkg_chk = $self->client_exec("dir /B $download_location\\$pkg[0]","3600");

            if ($pkg_chk !~ m/$pkg[0]/) {
                $self->client_exec("mkdir $download_location","60");
            }
            
            if ($pkg_chk =~ m/$pkg[0]/i) {
                $message = "rlse(): " . "Package \"$pkg[0]\" was previoulsy downloaded to $download_location" . "\n\n";
                $Log->comment($message);
            } else {
                $get_pkg = $self->client_exec("$downloader_get $download_location\\$pkg[0] $rlse_url$pkg[0]","3600");
                $pkg_chk = $self->client_exec("dir /B $download_location\\$pkg[0]","60");
            }
        } else {
            # Add Linux code here.
            $download_location = "/tmp/$project";
            
            # Verify if download directory exits
            my $dir_chk = $self->client_exec("ls $download_location","3600");
            $dir_chk =~ s/\s+$//;

            # Create download directory
            if ($dir_chk =~ m/No such file or directory/i) {
                $self->client_exec("mkdir $download_location","60");
            }
            
            # Verify target installer exits
            $pkg_chk = $self->client_exec("ls $download_location/$pkg[0]","3600");
            $pkg_chk =~ s/\s+$//;

            # Download installer if it does not exist in the download location
            if ($pkg_chk =~ m/No such file or directory/i) {
                $get_pkg = $self->client_exec("$downloader_get $download_location/$pkg[0] $rlse_url$pkg[0]","3600");
                $pkg_chk = $self->client_exec("ls $download_location/$pkg[0]","60");
            }
            # Installer exists in the download location
            elsif ($pkg_chk =~ m/$pkg[0]/i) {                                    
                $message = "rlse(): " . "Package \"$pkg[0]\" was previoulsy downloaded to $download_location" . "\n\n";
                $Log->comment($message);
            }
        }
        
        # Verify package was downloaded
        if ($pkg_chk !~ m/$pkg[0]/) {
            $message = "rlse(): " . "Package \"$pkg[0]\" was not found in $download_location" . "\n\n";
            $Log->error($message);
                    
        } else {
            $package = "$download_location\\$pkg[0]"; 
        }

        $logcomment = "Package located in $package\n\n";
    }

    $message = "rlse(): " . $logcomment . "\n";
    $Log->comment($message);

    return $package;
}

=head2 services_start()

This subroutine will start the agent and slpd described in the config
object, stopping existing processes if running. This is essentially just
a wrapper for slpd_start() and agent_start().

B<Usage:> C<$obj-E<gt>services_start()>  

B<Returns:> Boolean, TRUE ('1') if successful, FALSE ('0') otherwise. 

=cut

sub services_start(){
    my $self = shift;

    #$self->slpd_start() && $self->agent_start() ? return 1 : return 0;
    $self->agent_start() ? return 1 : return 0;

}

=head2 services_stop()

This subroutine will stop the agent and slpd described in the config
object. This is essentially just a wrapper for slpd_stop() and agent_stop().

B<Usage:> C<$obj-E<gt>services_stop(@proc_names)>  

B<Returns:> Boolean, TRUE ('1') if successful, FALSE ('0') otherwise. 

=cut

sub services_stop(){
    my $self = shift;

    #$self->slpd_stop() && $self->agent_stop() ? return 1 : return 0;
    $self->agent_stop() ? return 1 : return 0;
}

=head2 slpd_start()

This subroutine starts slpd server. slpd_stop() is called from within this
subroutine before attempting to start slpd, to ensure that there are no other
slpd processes running.

B<Usage:> C<$obj-E<gt>slpd_start()>  

B<Returns:> Boolean, TRUE ('1') if the process is successfully started, FALSE ('0') otherwise. 

=cut

sub slpd_start() {
    my $self = shift;
    my $dbg = $self->get_debug_module_verbose();
    my $slp_proc = $self->{_slp_proc};
    my $agent_tools_path = $self->{_agent_tools_path};
    my $slp_start = $self->{_slp_start};
    my $client_host_type = $self->{_client_host_type};
    my $hostname = $self->{_client_hostname};
    my $pid;
    my $message;

    my $proc_chk = $self->slpd_stop(); # First, make sure there are no slpd processes running

    if ($proc_chk eq "0") {
        $message = "Unable to stop existing $self->{_slp_proc} processes";
        $Log->log(FAIL($message));
    }

    if ($client_host_type eq "windows") { 
        # Get the value of $ENV{ProgramFiles} if $hostname is a minwin client?

        my $progfiles = $self->minwin_client("Program Files");
		$self->client_exec("\'$progfiles\\$agent_tools_path\\$slp_start\'","3600"); # start slpd                   
		snooze (5);
	} else {    
    	$self->client_exec($agent_tools_path . "/$slp_start","3600"); # start slpd     	
	}
	
	$pid = $self->get_pid($slp_proc);
	
    if ($pid =~ m/[0-9]/) {
        $message = "$slp_proc started, PID $pid";
        $Log->comment($message);
        return 1;
    } else {
        $message = "Unable to start $slp_proc process.";
        $Log->comment($message);
        return 0;
    }
}

=head2 slpd_stop()

This subroutine stops all slpd server processes found on the client.

B<Usage:> C<$obj-E<gt>slpd_stop()>  

B<Returns:> Boolean, TRUE ('1') if the processes are successfully stopped(or if there are none to be stopped, FALSE ('0') otherwise. 

=cut

sub slpd_stop() {
    my $self = shift;
    my $dbg = $self->get_debug_module_verbose();
    my $slp_proc = $self->{_slp_proc};
    my $agent_tools_path = $self->{_agent_tools_path};
    my $slp_stop = $self->{_slp_stop};
    my $client_host_type = $self->{_client_host_type};
    my $message;

    my $pid;
    my $slppid;

    $slppid = $self->get_pid($slp_proc);
    
	if ($pid =~ m/[0-9]/) {
		$message = "Will stop $slp_proc PID $pid";
        $Log->comment($message);

    	if ($client_host_type eq "windows") { 
        	# Get the value of $ENV{ProgramFiles} if $hostname is a minwin client?
        	my $progfiles = $self->minwin_client("Program Files");		     		
			$self->client_exec("$progfiles\\$agent_tools_path\\$slp_stop","3600"); # stop slpd                   
			snooze (10);
    	} else {
    	    $self->client_exec($agent_tools_path."/".$slp_stop,"3600"); # stop slpd
    	}
        
        $slppid = $self->get_pid($slp_proc);
                	
		if ($slppid =~ m/[0-9]/) {
			$message = "\"Failed to stop $slp_proc process with PID:\n\".$pid";
            $Log->comment($message);
			return 0;
		} else {
			$message = "\"Killed $slp_proc process with PID:\n\".$pid";
            $Log->comment($message);
			return 1;
		}
    } else {
        $message = "No $slp_proc process to stop.";
        $Log->comment($message);
        return 1;
    }
}

=head2 wmic()

This subroutine performa a WMI query from the WMIC linux WMI client to a Windows host.

B<Usage:> C<$obj-E<gt>wmic($class, $property, [$filter_name, $operator, $filter])>  

B<Returns:> WMI query results. 

=cut

sub wmic() {
    my $self = shift;
    my $class = shift;
    my $property = shift;
    my $filter_name = shift;
    my $operator = shift;    
    my $filter = shift;
    my $client_hostname = $self->{_client_hostname};
    my $client_ip = $self->{_client_ip};
    my $client_username = $self->{_client_username};
    my $client_password = $self->{_client_password};
    my $retry = $self->{_client_retry};
    my $query;
    my $results;
    my $message;

    for (my $i = 0; $i < $retry; $i++) {

        # WMI query
        if ($filter_name && $filter) {
            $query = "wmic -U $client_hostname/$client_username%$client_password  //$client_ip \"select $property from $class where $filter_name $operator '$filter'\"";
        } else {
            $query = "wmic -U $client_hostname/$client_username%$client_password  //$client_ip \"select $property from $class\"";
        }
        
        $results = `$query 2>&1`;
        
        if ($results =~ m/command not found/i) {
            $message = "wmic(): " . "NDISP host does not have WMIC installed" . "\n";
            $Log->comment($message);
            die Dumper($results,$query);
        }
        if ($results ne "") {
            if ($results =~ m/NT code 0x80041010/i) {
                $message = "wmic(): " . "Agent $client_hostname does not have Windows Management Infrastructure installed" . "\n" .
                "Please install 'WMI Windows Installer Provider' along with 'Hotfix KB 972189'" . "\n";
                $Log->comment($message);
                die Dumper($results,$query);
            }
                last;

        } else {

            if ($i == $retry) {
                $message = "wmic(): " . "Agent $client_hostname is not responding to:\n" . $query . "\n" . "After $retry retries\n";
                $Log->comment($message);
                die Dumper($results);
            }

        }

    }

    #$message = "wmic(): " . "Instance of $class $property on $client_hostname\n";
    #$Log->comment($message);
    #Dumper($results);    
    return $results;
}            

###############################################################################
# SET Accessor methods

=head2 set_debug_module_verbose($setting)

Updates module debug parameter in configuration object. Accepts Boolean. May
be used to print out debug info to be logged any time a particular configuration
object is used, depending on where in your script the parameter is set.

B<Usage:> C<$obj-E<gt>set_debug_module_verbose($setting)>

=cut


=head2 set_client($setting)

Updates _client parameter in configuration object. This is generally going to
be $CLIENT in the calling script, and may not need to be set manually.
Returns parameter value.

B<Usage:> C<$obj-E<gt>set_client($setting)>

=cut


=head2 set_add_filer_interval($setting)

Updates _add_filer_interval parameter in configuration object. This is the
minimum interval between adding filers. Returns parameter value.

B<Usage:> C<$obj-E<gt>set_add_filer_interval($setting)>

=cut


=head2 set_agent($setting)

Updates _agent parameter in configuration object. Accepts hostname or IP
of machine that will host release. Returns parameter value.

B<Usage:> C<$obj-E<gt>set_agent($setting)>

=cut


=head2 set_agent_bin_start($setting)

Updates _agent_bin_start parameter in configuration object. Accepts name of start
script for agent, e.g. "start_server". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_agent_bin_start($setting)>

=cut


=head2 set_agent_bin_stop($setting)

Updates _agent_bin_stop parameter in configuration object. Accepts name of stop 
script for agent, e.g. "stop_server". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_agent_bin_stop($setting)>

=cut


=head2 set_agent_bin_path($setting)

Updates _agent_bin_path parameter in configuration object. Accepts full
path to directory location on server where release is installed, 
e.g. "/usr/ws/server/agent/bin". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_agent_bin_path($setting)>

=cut


=head2 set_agent_log($setting)

Updates _agent_log parameter in configuration object. Accepts name of
logfile for agent, e.g. "cwbemserver.out". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_agent_log($setting)>

=cut


=head2 set_agent_proc($setting)

Updates _agent_proc parameter in configuration object. Accepts name of
agent process, e.g. "cwbemserver". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_agent_proc($setting)>


=head2 set_slp_proc($setting)

Updates _slp_proc parameter in configuration object. Accepts name of
slpd process, e.g. "slpd". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_slp_proc($setting)>

=cut


=head2 set_agent_tools_path($setting)

Updates _agent_tools_path parameter in configuration object. Accepts
path to binary tools on server, e.g. "/usr/ws/bin". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_agent_tools_path($setting)>

=cut


=head2 set_ins($setting)

Updates _ins parameter in configuration object. Accepts string for 
INS(Interop Name Space) setting, e.g. "interop". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_ins($setting)>

=cut


=head2 set_sxport($setting)

Updates _sxport parameter in configuration object. Accepts a string
indicating secure protocol, e.g. "https". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_sxport($setting)>

=cut


=head2 set_vns($setting)

Updates _vns parameter in configuration object. Accepts string for Vendor
Name Space, e.g. "root/ontap". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_vns($setting)>

=cut


=head2 set_xport($setting)

Updates _xport parameter in configuration object. Accepts a string
indicating non-secure non-secure transport name, e.g. "http". Returns
parameter value.

B<Usage:> C<$obj-E<gt>set_xport($setting)>

=cut


=head2 set_pkg_installer($setting)

Updates _pkg_installer parameter in configuration object. This is the
filename of the release installer script/binary, e.g. "install_smisproxy".
Returns parameter value.

B<Usage:> C<$obj-E<gt>set_pkg_installer($setting)>

=cut


=head2 set_force_install($setting)

Updates _force_install parameter in configuration object. This is the target
dir on the test host to which the release will be copied. Returns parameter
value.

B<Usage:> C<$obj-E<gt>set_force_install($setting)>

=cut


=head2 set_pkg_remover($setting)

Updates _pkg_remover parameter in configuration object. This is the
filename of the release uninstaller script/binary, e.g. "uninstall_smisproxy".
Returns parameter value.

B<Usage:> C<$obj-E<gt>set_pkg_remover($setting)>

=cut


=head2 set_pkg_xtension($setting)

Updates _pkg_xtension parameter in configuration object. This is the extension
used for the release package, e.g. "tar". Returns parameter value.

B<Usage:> C<$obj-E<gt>set_pkg_xtension($setting)>

=cut


=head2 set_rlse_url($setting)

Updates _rlse_url parameter in configuration object. This is the path to the 
SMIS release directory. Returns parameter value.

B<Usage:> C<$obj-E<gt>set_rlse_url($setting)>

=cut


=head2 set_client_bin_path($setting)

Updates _client_bin_path parameter in configuration object. Accepts path to 
location of WSI SDK tools, e.g. /usr/wsi/sdk/bin Returns parameter value.

B<Usage:> C<$obj-E<gt>set_client_bin_path($setting)>

=cut


=head2 set_object_path($setting)

Updates _object_path parameter in configuration object. Accepts filer id
, e.g. 0033609176 Returns parameter value.

B<Usage:> C<$obj-E<gt>set_object_path($setting)>

=cut


=head2 set_ver_file($setting)

Updates _ver_file parameter in configuration object. This is the target
version on the test host. Returns parameter value.

B<Usage:> C<$obj-E<gt>set_ver_file($setting)>

=cut


###############################################################################
# GET Accessor methods


=head2 get_debug_module_verbose($setting)

Retrieves module debug parameter from configuration object. Contains Boolean. May
be used to print out debug info to be logged any time a particular configuration
object is used, depending on where in your script the parameter is set.

B<Usage:> C<$obj-E<gt>get_debug_module_verbose($setting)>

=cut


=head2 get_client($setting)

Retrieves _client parameter from configuration object. This is generally going to
be $CLIENT in the calling script, and may not need to be set manually.
Returns parameter value.

B<Usage:> C<$obj-E<gt>get_client($setting)>

=cut


=head2 get_add_filer_interval($setting)

Retrieves _add_filer_interval parameter from configuration object. This is the
minimum interval between adding filers. Returns parameter value.

B<Usage:> C<$obj-E<gt>get_add_filer_interval($setting)>

=cut


=head2 get_agent($setting)

Retrieves _agent parameter from configuration object. Contains hostname or IP
of machine that will host release. Returns parameter value.

B<Usage:> C<$obj-E<gt>get_agent($setting)>

=cut


=head2 get_agent_bin_start($setting)

Retrieves _agent_bin_start parameter from configuration object. Contains name of start
script for agent, e.g. "start_server". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_agent_bin_start($setting)>

=cut


=head2 get_agent_bin_stop($setting)

Retrieves _agent_bin_stop parameter from configuration object. Contains name of start
script for agent, e.g. "stop_server". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_agent_bin_stop($setting)>

=cut


=head2 get_agent_bin_path($setting)

Retrieves _agent_bin_path parameter from configuration object. Contains full
path to directory location on server where release is installed, 
e.g. "/usr/ws/server/agent/bin". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_agent_bin_path($setting)>

=cut


=head2 get_agent_log($setting)

Retrieves _agent_log parameter from configuration object. Contains name of
logfile for agent, e.g. "cwbemserver.out". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_agent_log($setting)>

=cut


=head2 get_agent_proc($setting)

Retrieves _agent_proc parameter from configuration object. Contains name of
agent process, e.g. "cwbemserver". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_agent_proc($setting)>

=cut


=head2 get_slp_proc($setting)

Retrieves _slp_proc parameter from configuration object. Contains name of
slp process, e.g. "slpd". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_slp_proc($setting)>

=cut


=head2 get_agent_tools_path($setting)

Retrieves _agent_tools_path parameter from configuration object. Accepts
path to binary tools on server, e.g. "/usr/ws/bin". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_agent_tools_path($setting)>

=cut


=head2 get_ins($setting)

Retrieves _ins parameter from configuration object. Contains string for 
INS(Interop Name Space) setting, e.g. "interop". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_ins($setting)>

=cut


=head2 get_java($setting)

Retrieves _java parameter from configuration object.
Returns java path.

B<Usage:> C<$obj-E<gt>get_java($setting)>

=cut


=head2 get_sxport($setting)

Retrieves _sxport parameter from configuration object. Contains a string
indicating secure protocol, e.g. "https". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_sxport($setting)>

=cut


=head2 get_vns($setting)

Retrieves _vns parameter from configuration object. Contains string for Vendor
Name Space, e.g. "root/ontap". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_vns($setting)>

=cut


=head2 get_xport($setting)

Retrieves _xport parameter from configuration object. Contains a string
indicating non-secure non-secure transport name, e.g. "http". Returns
parameter value.

B<Usage:> C<$obj-E<gt>get_xport($setting)>

=cut


=head2 get_pkg_installer($setting)

Retrieves _pkg_installer parameter from configuration object. This is the
filename of the release installer script/binary, e.g. "install_smisproxy".
Returns parameter value.

B<Usage:> C<$obj-E<gt>get_pkg_installer($setting)>

=cut


=head2 get_force_install($setting)

Retrieves _force_install parameter from configuration object. This is the target
dir on the test host to which the release will be copied. Returns parameter
value.

B<Usage:> C<$obj-E<gt>get_force_install($setting)>

=cut


=head2 get_pkg_remover($setting)

Retrieves _pkg_remover parameter from configuration object. This is the
filename of the release uninstaller script/binary, e.g. "uninstall_smisproxy".
Returns parameter value.

B<Usage:> C<$obj-E<gt>get_pkg_remover($setting)>

=cut


=head2 get_pkg_xtension($setting)

Retrieves _pkg_xtension parameter from configuration object. This is the extension
used for the release package, e.g. "tar". Returns parameter value.

B<Usage:> C<$obj-E<gt>get_pkg_xtension($setting)>

=cut


=head2 get_rlse_url($setting)

Retrieves _rlse_url parameter from configuration object. This is the path to the 
SMIS release directory. Returns parameter value.

B<Usage:> C<$obj-E<gt>get_rlse_url($setting)>

=cut


=head2 get_client_bin_path($setting)

Retrieves _client_bin_path parameter from configuration object. Contains path to 
location of WSI SDK tools, e.g. /usr/wsi/sdk/bin Returns parameter value.

B<Usage:> C<$obj-E<gt>get_client_bin_path($setting)>

=cut


=head2 get_ver_file($setting)

Retrieves _ver_file parameter from configuration object. This is the target
version on the test host. Returns parameter value.

B<Usage:> C<$obj-E<gt>get_ver_file($setting)>

=cut


###############################################################################
###############################################################################
###############################################################################

# Use AUTOLOAD for similar functions which get/set attributes.
# First, define attributes for which AUTOLOAD is valid

{
    my %_attrs = (
        _association_list         =>  'read/write',  
        _agent_bin_path           =>  'read/write',
        _agent_bin_start          =>  'read/write',
        _agent_bin_stop           =>  'read/write',
        _agent_log                =>  'read/write',
        _agent_proc               =>  'read/write',
        _agent_tools_path         =>  'read/write',
        _class_path               =>  'read/write',
        _client_bin		          =>  'read/write',        
        _client_bin_path          =>  'read/write',
        _client_retry             =>  'read/write',
        _create_instance          =>  'read/write',
        _downloader_list          =>  'read/write',
        _downloader_get           =>  'read/write',
        _force_install            =>  'read/write',
        _ins                      =>  'read/write',
        _java                     =>  'read/write',
        _pkg_installer            =>  'read/write',
        _pkg_remover              =>  'read/write',
        _pkg_xtension             =>  'read/write',
        _port                     =>  'read/write',        
        _product_name             =>  'read/write',        
        _profiles                 =>  'read/write',
        _quick_test               =>  'read/write',
        _rlse_url                 =>  'read/write',
        _slp_proc                 =>  'read/write',
        _slp_start                =>  'read/write',
        _slp_stop                 =>  'read/write',
        _sport                    =>  'read/write',        
        _sxport                   =>  'read/write',
        _ver_file                 =>  'read/write',
        _vns                      =>  'read/write',
        _xport                    =>  'read/write',
        _client_hostname          =>  'read/write',
        _client_host_type         =>  'read/write',
        _client_ip                =>  'read/write',
        _client_os                =>  'read/write',
        _client_os_ver            =>  'read/write',
        _client_password          =>  'read/write',        
        _client_power             =>  'read/write',         
        _client_username          =>  'read/write',        
        _debug_module_verbose     =>  'debug',
        _filer_hostname           =>  'read/write',
        _filer_ip                 =>  'read/write',
        _filer_password           =>  'read/write',
        _last_message             =>  'read/write',
        _object_path              =>  'read/write',
        _project                  =>  'read/write',
        _prompt                   =>  'read',
    );
        
    sub _accessible { 
        my ($self, $attr, $mode) = @_;
        $_attrs{$attr} =~ /$mode/;
    }
   
    # Extra accessor function to make this hash available elsewhere. 
    # Additional logic added so that certain elements of the %_attr hash
    # which are not attributes of the filer configuration(such as
    # debug and connect directives) are not returned.
    
    sub _get_attributes {

        my %_h;

        foreach my $_key ( keys %_attrs ) {

            my $_value = $_attrs{$_key};

            if ($_key !~ m/_debug|_connect|hostname|prompt/ ) {

                $_h{$_key} = $_value;
            }
        }

        return %_h;
    }
}

sub AUTOLOAD {

    no strict "refs";
    my ($self, $setting) = @_;

    my $dbg = $self->{'_debug_module_verbose'}; # Set Debug flag
    
    if ($AUTOLOAD =~ /.*::set(_\w+)/ 
                    && $self->_accessible($1, 'write' ) ) {
        my $_attr_name = $1;

        # Anonymous subroutine to update object
        my $anon = sub {
            $self->{$_attr_name} = $setting;
        };

    return $setting;
    
    }
    
    elsif ($AUTOLOAD =~ /.*::get(_\w+)/ && $self->_accessible($1, 'read' ) ) {
    
        my $_attr_name = $1;
        return $self->{$_attr_name};
        
    }
    
    elsif ($AUTOLOAD =~ /.*::get(_\w+)/ && $self->_accessible($1, 'debug' ) ) {
    
        my $_attr_name = $1;
        return $self->{$_attr_name};
        
    }
    
    elsif ($AUTOLOAD =~ /.*::set(_\w+)/ && $self->_accessible($1, 'debug' ) ) {
    
        my $_attr_name = $1;
        
        if ( $setting =~ /on|1|true|TRUE|True/ ) {
        
            $setting = "1";
        }
        
        else {
        
            $setting = undef;
        }
        
        $self->{$_attr_name} = $setting;
        return $self->{$_attr_name};
    }
    elsif ($AUTOLOAD =~ /DESTROY/ ) {
    
        return 1;
    }
    else {
    
        logresult("SCRIPT", "No such attribute : $1");
    }
    
return 0;
}


=head1 SEE ALSO

=for html
<b>
<a href="./SMIS-associator-names.thpl.html">../SMIS/associator_names.thpl</a><br>
<a href="./SMIS-template_using_internal_subroutine-thpl.html">../SMIS/template_using_internal_subroutine.thpl</a><br>
For an example of using TCD to generate one testcase per file in a given directory see SSH/SSHredder.thpl.

</b>

=head1 HISTORY

=head1 TODO

=over

=item * TBD after code review

=back

=cut

    
=head1 COPYRIGHT

Copyright (c) 2012 NetApp, Inc.

=cut

1;

