#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/SNMP/TrapReceiver.pm#1

#***********************************************************************
#
# NAME
#   SNMP::TrapReceiver.pm - 
#
# SYNOPSIS
#   An SNMP Trap Receiver implemented via PERL.
#
# DESCRIPTION
#   This library provides a basic SNMP trap listener/receiver.
#
# AUTHOR
#   dl-cluster-qa
#
# HISTORY
# 9/13/2009 Initial revision
#***********************************************************************

## @summary     SNMP::TrapReceiver.pm

package SNMP::TrapReceiver;

use strict;

use Options;
use Tharn qw(host snooze subtest $Parent $Log fatal_result);

use NATE::Result;
use Data::Dumper;

use SNMP::TrapMgr qw(snmp_trap_mgr oid_lex_sort TRAP SNMP_TRAP_PORT);
use Getopt::Std;

sub new {
    my $this = shift;
    my $class = ref($this) || $this;

    my %options = (
        client=>'',
        ipv6=>'',
        port=>'162',
        @_
    );

    my @subArgs;
    if (!$options{client}) {
        $Log->error("SNMP Trap Receiver requires a client parameter \n");
        return 0;
    }

    my ($addr,$domain);
    if ( $options{ipv6} ) {
        $addr   = $options{ipv6};
    } else {
        my $clientobj = host($options{client});
        $addr = $clientobj->default_ip();
    }

    # I suppose in the future default_ip() might be IPv6?
    if ( $addr =~ /:/ ) {
        $domain = "udp/ipv6";
    } else {
        $domain = "udp/ipv4";
    }

    @subArgs = ('{use SNMP::TrapReceiver; SNMP::TrapReceiver::run(addr=>"'.$addr.'"'.
                                                                ',domain=>"'.$domain.'"'.
                                                                ',port=>"'.$options{port}.'"'.
                                                                '); }', '-bg', 
                                                                '-username', 'root', 
                                                                '-client', "$options{client}");
    my $subtest = subtest @subArgs;
    my $self = {
        "subtest" => $subtest,
        @_,
    };
    bless($self, $class);
    sleep(5); 
    return $self;
}

sub ready {
    my $self = shift;
    my $message;
    my $i = 0;
    while ($i++ < 10) {
          eval { $message=$self->{"subtest"}->message_get(type => "VERIFY",
          timeout => 3); 
          };
          if ($message =~ /SNMPREADY/) {
              $Log->comment( "SNMPTRAPD started\n");
              return 1; 
          }
          sleep(20);
    }
   $Log->comment("Failed to start SNMPTRAPD\n");
   return 0;
}

sub stop {
    my $self = shift;
    $Log->trace("Stopping the server...\n");
    $self->{"subtest"}->kill();
    sleep(10); #wait a few seconds to ensure the process dies and the port is free
    return;
}

sub get_entry {
    my $self = shift;
    my $message=undef;

    #get the single message from the queue
    $message=$self->{"subtest"}->message_get(type => "SNMP_Trap", timeout => 3);
    $Log->debug("Found raw message:  ".Dumper($message)."\n");
    return $message;
}

sub get_all {
    my @msgConcat = ();
    my $msgTmp = undef;

    while (1) {
        $msgTmp=undef;
        #get an entry that is on the queue, a message reported from the server
        eval {
            $msgTmp = get_entry(@_);
        };
        #if no message available, then exit
        if (!defined $msgTmp) {
            last;
        }
        #else add the message to the array
        else {
            $Log->debug("Adding message to array: ".Dumper($msgTmp));
            push @msgConcat, $msgTmp;
        }
    }
    #return the array of messages
    return \@msgConcat;
}

sub run {
    my %params = (
            addr=>undef,
            port=>"162",
            domain=>"udp/ipv4",
            @_,
    );
    #our %OPTS;

    # Do not buffer output
    $| = 1;

    $Log->comment( "Listening for traps on addr($params{addr}), port($params{port}), domain($params{domain})" );

    if ( $params{domain} =~ /ipv6/ ) {
        $Log->comment( "    Note: with domain($params{domain}, SNMPv1 traps received will have 0.0.0.0 as agent-addr" );
    }

    $Parent->message_put(type => "VERIFY", message => "SNMPREADY");
    # Start the trap manager
    snmp_trap_mgr(
       -localaddr => $params{addr},
       -domain    => $params{domain},
       -port      => $params{port},
       -callback  => [\&_trap_received_cb],
    );
}

sub _trap_received_cb
{
   my ($this) = @_;

   my $ip = $this->hostname;
   my $received_time;
   if ($this->error ne '') {

      printf("%s -ERROR- %s: %s.\n\n", _timestamp(), $ip, $this->error);

   } else {
      $received_time=_timestamp();
      if ($this->type == TRAP) {
         $Log->comment("Received Trap-PDU from ".$this->agent_addr." at $received_time :\n");
         printf("\tcommunity:     %s\n", $this->community);
         printf("\tenterprise:    %s\n", $this->enterprise);     
         printf("\tagent-addr:    %s\n", $this->agent_addr);
         printf("\tgeneric-trap:  %s\n", $this->generic_trap);
         printf("\tspecific-trap: %s\n", $this->specific_trap);
         printf("\tup-time:    %s\n", $this->time_stamp);
      } else {
         $Log->comment("Received SNMPv2-Trap-PDU from ".$this->agent_addr." at $received_time :\n");
         printf("\tcommunity:     %s\n", $this->community);
         printf("\terror-status:  %s\n", $this->error_status);
         printf("\terror-index:   %s\n", $this->error_index);
      }

      printf("\tvariable-bindings:\n");
      foreach ($this->var_bind_names) {
         printf("\t\t%s => %s\n", $_, $this->var_bind_list->{$_});
      }

      print "\n";
      my $trap={
        received_time=>$received_time,
        community=>$this->community,
        enterprise=>$this->enterprise,
        agent_address=>$this->agent_addr,
        generic_trap=>$this->generic_trap,
        specific_trap=>$this->specific_trap,
        up_time=>$this->time_stamp,
        var_binds=>$this->var_bind_list,
        error_status=>$this->error_status,
        error_index=>$this->error_index,
        };
      $Parent->message_put(type=>"SNMP_Trap", message=>$trap);

   }
}

sub _timestamp
{
   my @t = localtime;
   sprintf("%02d-%02d-%04d %02d:%02d:%02d",
      $t[4]+1, $t[3], $t[5]+1900, $t[2], $t[1], $t[0]
   );
}


1;


=pod

=head1 NAME

SNMP::TrapReceiver.pm - Basic SNMP Trap Receiver.


=head1 SYNOPSIS

    use SNMP::TrapReceiver;
    my $trap_receiver = new SNMP::TrapReceiver(client => $CLIENT);
    my $single_trap = $trap_receiver->get_entry();
    my $traps = $trap_receiver->get_all();
    $trap_receiver->stop();

=head1 DESCRIPTION

This library provides a basic SNMP Trap Receiver in PERL.
  
    
=head1 METHODS

new(client=>CLIENT);
        Description:
            Creates a new instance of the SNMP::TrapReceiver object.  It will
                bind to port 162 (UDP) on the specified client.
        Inputs:
            client => the client to run this SNMP::TrapReceiver on.
        Outputs:
            SNMP::TrapReceiver object instance
        Examples:
            <see synopsis above>
            
stop();
        Description:
            Stops an SNMP::TrapReceiver instance
        Inputs:
            <none>
        Outputs:
            <none>
        Examples:
            <see synopsis above>
            
get_entry();
        Description:
            Gets a single trap entry that was sent to the SNMP::TrapReceiver
        Inputs:
            <none>
        Outputs:
            %trap => A single trap entry that was sent to the server.
        Examples:
            <see synopsis above>

get_all();
        Description:
            Get all entries that were sent to the SNMP::TrapReceiver
        Inputs:
            <none>
        Outputs:
            $posts => All entries that were sent to the SNMP::TrapReceiver.  A
                ref to an array of hashes (traps).
        Examples:
            <see synopsis above>
            
run();
        Description:
            The main SNMP::TrapReceiver method.  This method will bind to
                port 162.  It will make a callback to the cb method anytime a
                trap is received.  This will cause the trap to be collected and
                posted to the queue to be read.
        Inputs:
            <none>
        Outputs:
            <none>
        Examples:
            <see synopsis above>
=cut
