#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/SPARSE/Resource/Setup.pm#1 $
#
## @summary	Setup - Common setup and cleanup for FlexCache tests
## @author	dl-nb-flexcache-qa (PFS QA NB)
## @status	public
## @notes	
## @pod here

=head1 NAME

Setup - Common setup and cleanup for NFS & PFS tests

=head1 SYNOPSIS

  use SPARSE::Resource::Setup;

  my %setup_h;
  $out = nfs_setup(setup_h => \%setup_h,
                   nfs_version => 3,
                   create_files => 5,
                   mount_vol => 1,
                   save_exports => 1,
                   mt_root_only => 'off');
  nfs_cleanup(setup_h => \%setup_h);

  # Setup as a flexcache origin filer
  $out = nfs_setup(setup_h => \%setup_h,
                   flexcache_access => 'host=flexcache_host');

  # Setup as a flexcache filer
  $out = nfs_setup(setup_h => \%setup_h,
                   flexcache => 1);

=head1 DESCRIPTION

Provides a common setup and cleanup functionality for testing NFS.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/main/lib/nfs/Setup.pm#10 $
  Last revised $Date: 2011/03/04 $

=head1 COPYRIGHT

  Copyright 2006, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package SPARSE::Resource::Setup;

use strict;
use warnings;
use base qw(Exporter);

use Carp qw(croak);
use POSIX qw(uname);
use Tharn qw(host connect);
use Fcntl qw(O_CREAT O_TRUNC);
use GenericMount qw(mount umount);
use Network::Utils qw(run_getXXbyYY);
use nfs::Utils qw(validate_nfs_version);
use SPARSE::Util::Logger qw(logStep logWarn);
use FS::Util::Setup qw(create_aggr create_vol common_cleanup);
use filer qw(filer_options_get filer_options_set filer_qtree_get
             filer_qtree_security_set filer_model_get);
use FileManip qw(read_file write_file remove_file);
use Aggregate qw(aggr_exists aggr_create aggr_purge);
use Admin::License qw(license is_licensed unlicense);
use Params::Validate qw(validate SCALAR ARRAYREF HASHREF);
use nfs::GenericExport qw(export_path list_exports export_reset_persistent);
use Volume qw(vol_create vol_create_flex vol_exists vol_destroy vol_offline);
use Filer::Version qw(:releases get_wafl_version);

my $f = host($main::FILER);

if ($f->hosttype =~ /filer-ng/m){

    require Server;
    require nfs_setup_gx;
}


BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our $VERSION = '1.0';

our @EXPORT_OK = qw(
  nfs_setup
  nfs_cleanup
);

#---------------------------------------------------------------------
# Subroutine prototypes
#---------------------------------------------------------------------
sub nfs_setup(;%);
sub nfs_cleanup(;%);
sub testme();

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my $root_volume = '/vol/vol0';
my $exp_file    = '/etc/exports';
my $ng_file     = '/etc/netgroup';
my $rc_file     = '/etc/rc';
my $gen_access  = 'sec=sys,rw,nosuid';
my $test_file   = "NFSfile.$$";
my $nfs_license = 'nfs';
my $flexcache_license;

#---------------------------------------------------------------------

=head2 nfs_setup

  $out = nfs_setup(setup_h => \%setup_h,
                   [, nfs_allowed => \@allowed ]
                   [, nfs_version => 2|3|4|'all'|'not2'|'not3'|'not4' ]
                   [, filer => $filer ]
                   [, setup_done => $setup_done ]
                   [, vol_name => $vol_name ]
                   [, vol_type => $vol_type ]
                   [, vol_size => $vol_size ]
                   [, num_extra_vol => 20 ]
                   [, aggr_name => $aggr_name ]
                   [, aggr_size => $aggr_size ]
                   [, create_files => $num_files ]
                   [, mount_vol => 0|1 ]
                   [, mount_root => 0|1 ]
                   [, root_access => 0|1 ]
                   [, save_exports => 0|1 ]
                   [, save_netgroup => 0|1 ]
                   [, save_rc => 0|1 ]
                   [, local_ipaddr => 0|1 ]
                   [, read_deleg => 'on'|'off' ]
                   [, write_deleg => 'on'|'off' ]
                   [, acl_enable => 'on'|'off' ]
                   [, mt_root_only => 'on'|'off' ]
                   [, options => \%options ]
                   [, flexcache => 0|1 ]
                   [, flexcache_access => $access ]);

Set up environment for NFS tests. It creates a volume and a number
of test files on that volume. It enables NFS version 4 options if specified.
All setup state is passed back to the test on the 'setup_h' named option.
This hash is populated with the following:

  { fc            => Filer connection object,
    files         => List of files created by setup,
    volume        => Full path of created volume,
    filer         => Hostspec of filer,
    vol_name      => Name of created volume,
    vol_type      => Type of created volume,
    vol_list      => List of extra volumes created,
    aggr_name     => Name of created aggregate,
    local_name    => Local host name as seen by the filer,
    local_ipaddr  => Local ip address as seen by the filer,
    nfs_versions  => List of valid NFS versions,
    vol_mtpoint   => Test volume mountpoint,
    root_mtpoint  => Root volume mountpoint,
    exports_file  => Path of the exports file,
    netgroup_file => Path of the netgroup file,
    rc_file       => Path of the rc file }

See nfs_cleanup for a more in-depth description
of some of the items in the above hash.
The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Hash reference where setup state is passed back to test.

=item nfs_allowed

(Optional) : An array reference indicating what NFS versions are
allowed for this test. An error occurs if the nfs_version parameter
does not match one of these values. Default is to allow all versions.

=item nfs_version

(Optional) : NFS version(s) that will be tested.
Default is 'all' for testing versions 2, 3 and 4.

=item filer

(Optional) : Hostspec of filer to setup. Default is $main::FILER.

=item setup_done

(Optional) : If set to 1 the volume must already exists and if set to 0
the volume is created. Default is $main::SETUP_DONE.

=item vol_name

(Optional) : Name of volume to use/create.
Default is $main::FILER_VOL.

=item vol_type

(Optional) : Type of volume to use/create.
Default is $main::VOL_TYPE.

=item vol_size

(Optional) : Size of volume to use/create.
Default is $main::VOL_SIZE.

=item num_extra_vol

(Optional) : The number of volumes to create in addition to the volume
specified by vol_name. The volume names will be placed in the returned
hash under 'vol_list'. The volumes will be exported from the filer but
not mounted by the client. Default is 0.

=item aggr_name

(Optional) : Name of aggregate to use/create.
Default is $main::FILER_AGGR.

=item aggr_size

(Optional) : Size of aggregate to use/create.
Default is $main::AGGR_SIZE.

=item create_files

(Optional) : Create this number of files on test volume.
Default is 0 (no files are created).

=item mount_vol

(Optional) : Mount the test volume using the highest NFS version
provided by the nfs_version parameter.
Default is 0 (test volume is not mounted).

=item mount_root

(Optional) : Mount the root volume using the highest NFS version
provided by the nfs_version parameter.
Default is 0 (root volume is not mounted).

=item root_access

(Optional) : Export test volume with root access.
Default is 0 (the export options are left intact).

=item save_exports

(Optional) : Save the exports file.
Default is 0 (exports file is not saved).

=item save_netgroup

(Optional) : Save the netgroup file.
Default is 0 (netgroup file is not saved).

=item save_rc

(Optional) : Save the rc file.
Default is 0 (rc file is not saved).

=item local_ipaddr

(Optional) : Resolve local ip address as seen by the filer.
Default is 0 (do not resolve local ip address).

=item read_deleg

(Optional) : Enable/disable read delegations.

=item write_deleg

(Optional) : Enable/disable write delegations.

=item acl_enable

(Optional) : Enable/disable NFS version 4 ACL's.

=item mt_root_only

(Optional) : Allow only privileged NFS mounts, i.e. only the root
user can mount the volume.

=item options

(Optional) : Reference to a hash of options to set, the key being the
option name and value to set.

=item flexcache

(Optional) : Setup as a flexcache filer. If true, license flexcache and nfs
then create an aggregate if it does not already exist. Default is 0.

=item flexcache_access

(Optional) : Setup as an origin filer. If defined, enable flexcache and
set the flexcache access to this value.

=item export_options

(Optional) : Options to pass to exportfs when exporting volumes.
Default is 'rw,anon=0'.

=back

=item Returns

1 if succeeds else croaks with error.

=back

=for html <hr />

=cut

#---------------------------------------------------------------------
sub nfs_setup(;%) {
    my (%opts) = validate (
        @_,
        {
            setup_h          => { type => HASHREF },
            nfs_allowed      => { type => ARRAYREF, default => [2, 3, 4] },
            nfs_version      => { type => SCALAR,   default => 'all' },
            filer            => { type => SCALAR,
                               default => $main::FILER },
            setup_done       => { type => SCALAR,
                               default => $main::SETUP_DONE },
            vol_name         => { type => SCALAR,
                               default => $main::FILER_VOL },
            vol_type         => { type => SCALAR,
                               default => $main::VOL_TYPE },
            vol_size         => { type => SCALAR,
                               default => $main::VOL_SIZE },
            num_extra_vol    => { type => SCALAR,   default => 0 },
            aggr_name        => { type => SCALAR,
                               default => $main::FILER_AGGR },
            aggr_size        => { type => SCALAR,
                               default => $main::AGGR_SIZE },
            create_files     => { type => SCALAR,   default => 0,
                                 regex => qr/^\d+$/ },
            mount_vol        => { type => SCALAR,   default => 0,
                                 regex => qr/^[01]$/ },
            mount_root       => { type => SCALAR,   default => 0,
                                 regex => qr/^[01]$/ },
            root_access      => { type => SCALAR,   default => 0,
                                 regex => qr/^[01]$/ },
            mt_root_only     => { type => SCALAR,  optional => 1,
                                 regex => qr/^(on|off)$/ },
            options          => { type => HASHREF, optional => 1 },
            save_exports     => { type => SCALAR,   default => 0,
                                 regex => qr/^[01]$/ },
            save_netgroup    => { type => SCALAR,   default => 0,
                                 regex => qr/^[01]$/ },
            save_rc          => { type => SCALAR,   default => 0,
                                 regex => qr/^[01]$/ },
            local_ipaddr     => { type => SCALAR,   default => 0,
                                 regex => qr/^[01]$/ },
            read_deleg       => { type => SCALAR,  optional => 1,
                                 regex => qr/^(on|off)$/ },
            write_deleg      => { type => SCALAR,  optional => 1,
                                 regex => qr/^(on|off)$/ },
            acl_enable       => { type => SCALAR,  optional => 1,
                                 regex => qr/^(on|off)$/ },
            flexcache        => { type => SCALAR,   default => 0 },
            flexcache_access => { type => SCALAR,  optional => 1 },
            export_options   => { type => SCALAR,   default => 'rw,anon=0' },
        }
    );
    my ($fc, $fh, $mountpoint, $root_mtpoint, @nfsver, %nfsver_h, $op,
        %qtree_out, $vol, $option);
    my $set_h = $opts{setup_h};
    my $ng_check;
    my $f = host($main::FILER);

    if ($f->hosttype =~ /filer-ng-cfe/m){
                $ng_check = 1;
    }
    $set_h->{rm_netgroup_file}    = 0;
    $set_h->{restore_root_export} = 0;

    logStep('BEGIN: nfs_setup: Setup environment for test');

    if (!$ng_check) {
        logStep ("set_env for 7G/7M system");
        # Connect to filer
        $set_h->{filer} = $opts{filer};
        $fh = host($set_h->{filer});
        $fc = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
        $fc->timeout($main::CMDTIMEOUT);
        $set_h->{fc} = $fc;

        logStep('Validate NFS version');
        @nfsver = validate_nfs_version(version => $opts{nfs_version},
                                       allowed => $opts{nfs_allowed});
        $set_h->{nfs_versions} = \@nfsver;

        # Create hash to simplify check for specific versions
        %nfsver_h = map { $_ => 1 } @nfsver;

        # File may have been left in NTFS-only mode which means all volumes
        # and qtrees created will have the security style set to NTFS which
        # will cause NFS client access to fail. Make sure the default security
        # style is set to UNIX.
        if ( filer_options_get($fc, 'wafl.default_security_style') eq 'ntfs' ) {
            filer_options_set($fc, 'wafl.default_security_style', 'unix');
        }

        # Set filer options
        if ( defined $opts{options} ) {
            foreach $option ( keys %{$opts{options}} ) {
                # Save current value of option
                $set_h->{_options}{$option} = filer_options_get($fc, $option);

                if ( $opts{options}{$option} ne $set_h->{_options}{$option} ) {
                    # Set option to value specified
                    filer_options_set($fc, $option, $opts{options}{$option});
                } else {
                    # Option was not set, do not clean it up
                    delete $set_h->{_options}{$option};
                }
            }
        }

        if ( $nfsver_h{4} ) {
            # Save current NFSv4 state
            logStep('Save current NFSv4 state');
            my $rdeleg    = $opts{read_deleg};
            my $wdeleg    = $opts{write_deleg};
            my $aclenable = $opts{acl_enable};
            # Get current state of NFSv4
            $set_h->{nfs_v4_enable} = filer_options_get($fc, 'nfs.v4.enable');
            if ( defined($rdeleg) ) {
                # Get current state of read delegations
                $set_h->{nfs_v4_rdel} =
                  filer_options_get($fc, 'nfs.v4.read_delegation')
            }
            if ( defined($wdeleg) ) {
                # Get current state of write delegations
                $set_h->{nfs_v4_wdel} =
                  filer_options_get($fc, 'nfs.v4.write_delegation')
            }
            if ( defined($aclenable) ) {
                # Get current state of ACL's
                $set_h->{nfs_v4_acl} =
                  filer_options_get($fc, 'nfs.v4.acl.enable')
            }

            logStep('Enable NFSv4 on server');
            if ( $set_h->{nfs_v4_enable} eq 'off' ) {
                # Enable NFSv4
                filer_options_set($fc, 'nfs.v4.enable', 'on');
            }
            if ( defined($rdeleg) && $rdeleg ne $set_h->{nfs_v4_rdel} ) {
                # Change state of read delegations
                filer_options_set($fc, 'nfs.v4.read_delegation', $rdeleg);
            }
            if ( defined($wdeleg) && $wdeleg ne $set_h->{nfs_v4_wdel} ) {
                # Change state of write delegations
                filer_options_set($fc, 'nfs.v4.write_delegation', $wdeleg);
            }
            if ( defined($aclenable) && $aclenable ne $set_h->{nfs_v4_acl} ) {
                # Change state of ACL's
                filer_options_set($fc, 'nfs.v4.acl.enable', $aclenable);
            }
        }

        if ( defined($opts{mt_root_only}) ) {
            # Save current state of mount_rootonly option
            $set_h->{mt_root_only} = filer_options_get($fc, 'nfs.mount_rootonly');
            if ( $opts{mt_root_only} ne $set_h->{mt_root_only} ) {
                # Change state of option
                filer_options_set($fc, 'nfs.mount_rootonly', $opts{mt_root_only});
            }
        }

        # Save current export options to be restored at nfs_cleanup
        logStep('Save current export options');
        $set_h->{export_list} = list_exports($fc);

        if ( $opts{local_ipaddr} ) {
            # Get localhost ip address and name as seen by the filer
            my ($local_name) = (uname())[1];
            $op = run_getXXbyYY($fc, 'gethostbyname_r', $local_name);
            ($set_h->{local_ipaddr}) = ($op =~ /addresses:\s+(\d+\.\d+\.\d+\.\d+)/);
            ($set_h->{local_name})   = ($op =~ /name:\s+(.+)/);
        }

        if ( $opts{flexcache} ) {
            logStep('Create aggregate for flexcache');
            # Create aggregate to host the flexcache volume
            $op = aggr_exists($fc, $opts{aggr_name});
            if ( $@ =~ /No .* named '$opts{aggr_name}' exists/ ) {
                $set_h->{aggr_created} = 1;
                #aggr_create($fc, $opts{aggr_name}, $opts{aggr_size})
                create_aggr(filer_conn=>$fc, aggr_size=>$opts{aggr_size})
                  or croak "nfs_setup: $@";
            } elsif ( $@ ) {
                croak "nfs_setup: $@";
            }

            # License flexcache
            logStep('License flexcache');

            # Set the flexcache license name based on the wafl
            # version.
            if (get_wafl_version($fh) >= WV_IRONCITY_MIN) {
                # IC.0 or later flexcache license
                $flexcache_license = 'flexcache_nfs';
            } else {
                # Pre-IC.0 flexcache license
                $flexcache_license = 'flex_cache';
            }
            if ( ! is_licensed($fc, $flexcache_license) ) {
                $set_h->{flexcache_licensed} = 1;
                license($fc, $flexcache_license) or croak "nfs_setup: $@";
            }
	
            ## there may be an option setting that is necessary 
            ## if we're using the flexcache_nfs license
	    if ($flexcache_license =~ /_nfs/) {
	        ## if the licensed_feature option exists
	        ## we also need to set that for flexcache_nfs to work
	        ## appropriately
	        my $return=filer_options_get ($fc, 'licensed_feature.flexcache_nfs.enable');
	        ## null return means it didn't exist so if it exists but isn't on
	        ## we need to turn it on in addition to adding the license.
	        if (($return ne '') && 
		    ($return ne 'on')){
		    filer_options_set($fc, 'licensed_feature.flexcache_nfs.enable', 'on');
	        }
	    }

            # License nfs
            logStep('License nfs');
            if ( (filer_model_get($fc) !~ /^SA/i)
                && (! is_licensed($fc, $nfs_license)) ) {
                $set_h->{nfs_licensed} = 1;
                license($fc, $nfs_license) or croak "nfs_setup: $@";
            }
        } else {
            # Common setup -- creates a volume
            logStep('Create test volume');
            create_vol(filer        => $set_h->{filer},
                       filer_vol    => $opts{vol_name},
                       vol_type     => $opts{vol_type},
                       vol_size     => $opts{vol_size},
                       filer_aggr   => $opts{aggr_name},
                       aggr_size    => $opts{aggr_size},
                       extern_setup => $opts{setup_done}) or croak "nfs_setup: $@";

            # Volume name and type
            $set_h->{vol_name} = $opts{vol_name};
            $set_h->{vol_type} = $opts{vol_type};

            # Full path for volume
            $set_h->{volume} = "/vol/$opts{vol_name}";

            %qtree_out = filer_qtree_get($set_h->{fc}, vols => 1);
            if ($qtree_out{$opts{vol_name}}{''}{Style} !~ /(mixed|unix)/) {
                $set_h->{qtree_style_org} =
                    $qtree_out{$opts{vol_name}}{''}{'Style'};

                # Set the qtree security mode on the default volume
                logStep("Setting volume qtree security mode setting to unix");
                if (filer_qtree_security_set($set_h->{fc}, $set_h->{volume},
                                             'unix') == 0) {
                    croak("qtree set failed on $set_h->{volume}");
                }
            }
        }

        # Aggregate name
        $set_h->{aggr_name} = $opts{aggr_name};

        if ( defined($opts{flexcache_access}) ) {
            if ( filer_options_get($fc, 'flexcache.enable') eq 'off' ) {
                # Enable flexcache
                $set_h->{flexcache_enabled} = 1;
                filer_options_set($fc, 'flexcache.enable', 'on');
            }

            # Save current access to restore it at cleanup
            $set_h->{flexcache_access} = filer_options_get($fc, 'flexcache.access');
            filer_options_set($fc, 'flexcache.access', $opts{flexcache_access});
        }

        if ( $opts{mount_root} || $opts{save_exports}
          || $opts{save_netgroup} || $opts{save_rc} ) {
            logStep('Export root volume');
            $set_h->{restore_root_export} = 1;
            export_path($fc, $root_volume, OPTIONS=>"$opts{export_options}")
              or croak "nfs_setup: $@";

            logStep('Mount root volume');
            $root_mtpoint = mount($fc->connhost(),
                                  $root_volume,
                                  mntopts => "vers=$nfsver[-1]")
              or croak "nfs_setup: $@";

            # Save mount point so if setup should fail
            # it will get properly unmounted
            $set_h->{root_mtpoint} = $root_mtpoint;

            if ( $opts{save_exports} ) {
                logStep('Save the exports file');
                $set_h->{exp_contents} = read_file(file => "$root_mtpoint$exp_file",
                                                   client => 'localhost',
                                                   user   => 'root');
                $set_h->{exports_file} = $exp_file;
            }

            if ( $opts{save_netgroup} ) {
                logStep('Save the netgroup file');
                $set_h->{ng_contents} = eval {
                    read_file(file   => "$root_mtpoint$ng_file",
                              client => 'localhost',
                              user   => 'root');
                };
                if ( $@ =~ /No such file/ ) {
                    # Netgroup file does not exists,
                    # so remove it at nfs_cleanup
                    $set_h->{rm_netgroup_file} = 1;
                } elsif ( $@ ) {
                    croak $@;
                }
                $set_h->{netgroup_file} = $ng_file;
            }

            if ( $opts{save_rc} ) {
                logStep('Save the rc file');
                $set_h->{rc_contents} = read_file(file => "$root_mtpoint$rc_file",
                                                  client => 'localhost',
                                                  user   => 'root');
                $set_h->{rc_file} = $rc_file;
            }

            if ( ! $opts{mount_root} ) {
                logStep('Unmount root volume');
                umount($root_mtpoint) or croak "nfs_setup: $@";

                # Already unmounted, no need to unmount it
                # anymore at nfs_cleanup if setup should fail
                delete $set_h->{root_mtpoint};
            }
        }

        # Mount the test volume only if it is specified
        # or need to create test files
        if ( $set_h->{volume}
          && ($opts{root_access} || $opts{mount_vol} || $opts{create_files}) ) {
            logStep('Export test volume');
            export_path($fc, $set_h->{volume}, OPTIONS=>"$opts{export_options}")
              or croak "nfs_setup: $@";

            if ( $opts{mount_vol} || $opts{create_files} ) {
                logStep('Mount test volume');
                $mountpoint = mount($fc->connhost(),
                                    $set_h->{volume},
                                    mntopts => "vers=$nfsver[-1]",
                                    chmod  => '0777') or croak "nfs_setup: $@";

                # Save mount point so if setup should fail
                # it will get properly unmounted
                $set_h->{vol_mtpoint} = $mountpoint;

                logStep('Create test files') if $opts{create_files};
                my $fcontents = join('', ('A'..'Z', "\n"))x10;
                foreach my $index ( 1..$opts{create_files} ) {
                    # Create each file with the same contents and
                    # save the file name and inode for future use
                    my $tfile = "$test_file.$index";
                    my $inode = ( write_file(client   => 'localhost',
                                             file     => "$mountpoint/$tfile",
                                             fstat    => 1,
                                             contents => $fcontents) )[1];
                    push @{$set_h->{files}}, { file => $tfile, inode => $inode };
                }

                if ( ! $opts{mount_vol} ) {
                    logStep('Unmount test volume');
                    umount($mountpoint) or croak "nfs_setup: $@";

                    # Already unmounted, no need to unmount it
                    # anymore at nfs_cleanup if setup should fail
                    delete $set_h->{vol_mtpoint};
                }
            }
        }

        if ( $set_h->{volume} && ! $opts{root_access}
          && $set_h->{export_list}{$set_h->{volume}} ) {
            # Re-export the volume with pre-create_vol options
            export_path($fc, $set_h->{volume},
                        OPTIONS=>$set_h->{export_list}{$set_h->{volume}})
              or croak "nfs_setup: $@";
        }

        if ( $opts{num_extra_vol} > 0 ) {
            logStep("Create and export the $opts{num_extra_vol} volumes");
            my $pid = $$;
            foreach my $index ( 1..$opts{num_extra_vol} ) {
                # Generate volume name that doesn't exist on filer
                my $gen = 0;
                do {
                    $vol = 'nfs_vol_' . $pid . '_' . $index . '_' . $gen;
                    $gen++;
                } while ( vol_exists($fc, $vol) );
                # Create volume
                if ( $opts{vol_type} eq 'flex' ) {
                    vol_create_flex($fc, $vol, $opts{aggr_name},
                                    $opts{vol_size}, guarantee => 'none')
                        or croak "nfs_setup: $@";
                } else {
                    vol_create($fc, $vol, $opts{vol_size})
                        or croak "nfs_setup: $@";
                }
                push @{$set_h->{vol_list}}, $vol;

                # Export volume
                export_path($fc, "/vol/$vol", OPTIONS=>"$opts{export_options}")
                    or croak "nfs_setup: $@";
            }
        }
    }

    else {
        logStep ("nfs_setup for 10Mode System");
        my %set_h;
        my $ref1 = nfs_setup_gx::set_env_gx( setup_h  => \%set_h);

        my $hostname = $ref1->{hostname};
        $set_h->{volume} = $ref1->{volume};
   
        logStep ("Volume is $set_h->{volume}");
        host($hostname);

        $set_h->{fc} = connect($main::FILER_CONN);
        my $fc = $set_h->{fc};
        # Full path for volume
        #$set_h->{volume} = $ref1->{volume};

        if ( $opts{mount_vol} || $opts{create_files} ) {
            logStep('Mount test volume');
            $mountpoint = mount($fc->connhost(),
                                $set_h->{volume},
                                mntopts => "vers=4",
                                chmod  => '0777') or croak "nfs_setup: $@";

            # Save mount point so if setup should fail
            # it will get properly unmounted
            $set_h->{vol_mtpoint} = $mountpoint;

            logStep('Create test files') if $opts{create_files};
            my $fcontents = join('', ('A'..'Z', "\n"))x10;
            foreach my $index ( 1..$opts{create_files} ) {
                # Create each file with the same contents and
                # save the file name and inode for future use
                my $tfile = "$test_file.$index";
                my $inode = ( write_file(client   => 'localhost',
                                         file     => "$mountpoint/$tfile",
                                         fstat    => 1,
                                         contents => $fcontents) )[1];
                push @{$set_h->{files}}, { file => $tfile, inode => $inode };
            }
        }
        if ( ! $opts{mount_vol} ) {
            logStep('Unmount test volume');
            umount($mountpoint) or croak "nfs_setup: $@";
            # Already unmounted, no need to unmount it
            # anymore at nfs_cleanup if setup should fail
            delete $set_h->{vol_mtpoint};
        }
    }

    logStep('END: nfs_setup');
    return 1;
}

#---------------------------------------------------------------------

=head2 nfs_cleanup

  nfs_cleanup(setup_h => \%setup_h);

Clean up environment for NFS locks tests. It destroys the volume and
removes any files given in list. It restores the NFSv4 state and any
other options and files modified by nfs_setup.

The following named arguments are accepted:

=over

=item Arguments

=over

=item setup_h

(Required) : Hash reference where setup state is passed.
The following is a list of items used to cleanup the environment:

files => Array reference of files to remove.
Each entry can either be a file name or a hash reference having
the file name on the 'file' key. If file name has a wild card (? or *)
all files that match are removed recursively.
Also if entry is a hash which contains the 'recursive' key set to 1,
the directory is removed recursively.
File names are relative to the 'volume' option.

volume => Absolute path of volume.

vol_name => Name of volume.

vol_mtpoint => The mount point of test volume.

root_mtpoint => The mount point of root volume.

=back

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub nfs_cleanup(;%) {
    my (%opts) = validate (
        @_,
        {
            setup_h => { type => HASHREF },
        }
    );
    my ($fc, $access, $option);
    my ($save_vol_type, $save_filer_vol, $save_filer_aggr, $save_cleanup);
    my $exports_file_restored = 0;
    my $set_h = $opts{setup_h};
    my $mountpoint   = $set_h->{vol_mtpoint};
    my $root_mtpoint = $set_h->{root_mtpoint};

    if ( defined($set_h->{cleanup}) ) {
        $save_cleanup = $main::CLEANUP;
        $main::CLEANUP = $set_h->{cleanup};
    }

    logStep("CLEANUP == $main::CLEANUP");

    eval {
        eval {
            logStep('Re-connect to filer');
            host($set_h->{filer});
            $fc = connect($main::FILER_CONN, '-timeout', $main::TIMEOUT);
            $fc->timeout($main::CMDTIMEOUT);
        };
        if ( $@ ) {
            logWarn("nfs_cleanup: $@");
            undef $fc;
        }

        if ( $main::CLEANUP ne 'none' ) {
            eval {
                if ( $set_h->{volume} ) {
                    my $remove_files = ( defined($set_h->{files}) and
                                         @{$set_h->{files}} > 0 );
                    if ( defined($fc) && $remove_files && ! $mountpoint ) {
                        logStep('Export test volume');
                        export_path($fc, $set_h->{volume},
                                    OPTIONS=>'rw,anon=0') or croak $@;

                        logStep('Mount test volume');
                        $mountpoint = mount($fc->connhost(),
                                            $set_h->{volume},
                                            mountretry => 2) or croak $@;
                    }

                    if ( $mountpoint && $remove_files ) {
                        my $file;
                        logStep('Remove created files');
                        foreach ( @{$set_h->{files}} ) {
                            my $ishash = (ref($_) eq 'HASH');
                            $file = $ishash ? $_->{file} : $_;
                            my $recursive = ($file =~ /[*?]/ ? 1 :
                                             $ishash ? exists($_->{recursive}) ?
                                             $_->{recursive} : 0 : 0);
                            eval { remove_file(client    => 'localhost',
                                               file      => "$mountpoint/$file",
                                               recursive => $recursive) };
                            logWarn("nfs_cleanup: $@") if $@;
                        }
                    }

                    if ( $mountpoint ) {
                        logStep('Unmount test volume');
                        umount($mountpoint) or croak "nfs_cleanup: $@";
                    }
                }

                # Set flag if it is needed to restore root files
                my $need_root_mt = ( $set_h->{exp_contents}
                  || $set_h->{ng_contents} || $set_h->{rc_contents} );

                if ( $root_mtpoint || $need_root_mt ) {
                    if ( defined($fc) && ! $root_mtpoint && $need_root_mt ) {
                        logStep('Export root volume');
                        export_path($fc, $root_volume,
                                    OPTIONS=>'rw,anon=0') or croak $@;

                        logStep('Mount root volume');
                        $root_mtpoint = mount($fc->connhost(),
                                              $root_volume,
                                              mountretry => 2) or croak $@;
                    }

                    if ( $root_mtpoint && $set_h->{exp_contents} ) {
                        logStep('Restore the exports file');
                        eval { write_file(file     => $root_mtpoint . $exp_file,
                                          client   => 'localhost',
                                          user     => 'root',
                                          oflags   => O_CREAT|O_TRUNC,
                                          contents => $set_h->{exp_contents}) };
                        logWarn("nfs_cleanup: $@") if $@;
                        $exports_file_restored = 1;
                    }

                    if ( $root_mtpoint && $set_h->{ng_contents} ) {
                        logStep('Restore the netgroup file');
                        eval { write_file(file     => $root_mtpoint . $ng_file,
                                          client   => 'localhost',
                                          user     => 'root',
                                          oflags   => O_CREAT|O_TRUNC,
                                          contents => $set_h->{ng_contents}) };
                        logWarn("nfs_cleanup: $@") if $@;
                    } elsif ( $root_mtpoint && $set_h->{rm_netgroup_file} ) {
                        logStep('Remove the netgroup file');
                        eval { remove_file(file   => $root_mtpoint . $ng_file,
                                           client => 'localhost',
                                           user   => 'root') };
                        logWarn("nfs_cleanup: $@") if $@;
                    }

                    if ( $root_mtpoint && $set_h->{rc_contents} ) {
                        logStep('Restore the rc file');
                        eval { write_file(file     => $root_mtpoint . $rc_file,
                                          client   => 'localhost',
                                          user     => 'root',
                                          oflags   => O_CREAT|O_TRUNC,
                                          contents => $set_h->{rc_contents}) };
                        logWarn("nfs_cleanup: $@") if $@;
                    }

                    if ( $root_mtpoint ) {
                        logStep('Unmount root volume');
                        umount($root_mtpoint) or croak "nfs_cleanup: $@";
                    }
                }

                if ( defined($fc) && defined($set_h->{vol_list}) &&
                     @{$set_h->{vol_list}} > 0 ) {
                    # Destroy any extra volumes created which has the
                    # side effect of unexporting them
                    foreach my $vol (@{$set_h->{vol_list}}) {
                        eval {
                            if (vol_exists($fc, $vol)) {
                                vol_offline($fc, $vol) or
                                    logWarn("nfs_cleanup: Cannot offline $vol: $@");
                                vol_destroy($fc, $vol) or
                                    logWarn("nfs_cleanup: Cannot destroy $vol: $@");
                            }
                        };
                        logWarn("nfs_cleanup: $@") if $@;
                    }
                }
            };

            logWarn("nfs_cleanup: $@") if $@;

            if ( defined($fc) ) {
                # Restore filer options
                if ( defined($set_h->{nfs_v4_enable}) ) {
                    logStep('Restore NFSv4 state');
                    eval { filer_options_set($fc, 'nfs.v4.enable',
                                             $set_h->{nfs_v4_enable}) };
                    logWarn("nfs_cleanup: $@") if $@;
                }
                if ( defined($set_h->{nfs_v4_rdel}) ) {
                    eval { filer_options_set($fc, 'nfs.v4.read_delegation',
                                             $set_h->{nfs_v4_rdel}) };
                    logWarn("nfs_cleanup: $@") if $@;
                }
                if ( defined($set_h->{nfs_v4_wdel}) ) {
                    eval { filer_options_set($fc, 'nfs.v4.write_delegation',
                                             $set_h->{nfs_v4_wdel}) };
                    logWarn("nfs_cleanup: $@") if $@;
                }
                if ( defined($set_h->{nfs_v4_acl}) ) {
                    eval { filer_options_set($fc, 'nfs.v4.acl.enable',
                                             $set_h->{nfs_v4_acl}) };
                    logWarn("nfs_cleanup: $@") if $@;
                }
                if ( defined($set_h->{mt_root_only}) ) {
                    eval { filer_options_set($fc, 'nfs.mount_rootonly',
                                             $set_h->{mt_root_only}) };
                    logWarn("nfs_cleanup: $@") if $@;
                }

                foreach $option ( keys %{$set_h->{_options}} ) {
                    eval { filer_options_set($fc, $option,
                                             $set_h->{_options}{$option}) };
                    logWarn("nfs_cleanup: $@") if $@;
                }

                # Restore flexcache status
                if ( defined($set_h->{flexcache_access}) ) {
                    eval { filer_options_set($fc, 'flexcache.access',
                                             $set_h->{flexcache_access}) };
                    logWarn("nfs_cleanup: $@") if $@;
                }
                if ( defined($set_h->{flexcache_enabled}) ) {
                    eval { filer_options_set($fc, 'flexcache.enable', 'off') };
                    logWarn("nfs_cleanup: $@") if $@;
                }
                if ( defined($set_h->{flexcache_licensed}) ) {
                    unlicense($fc, $flexcache_license);
                    logWarn("nfs_cleanup: $@") if $@;
		    ## there may be an option setting that is necessary 
		    ## if we're using the flexcache_nfs license
		    if ($flexcache_license =~ /_nfs/) {
			## if the licensed_feature option exists
			## we also need to set that for flexcache_nfs to work
			## appropriately
			my $return=filer_options_get ($fc, 'licensed_feature.flexcache_nfs.enable');
			## null return means it didn't exist so if it exists but isn't on
			## we need to turn it on in addition to adding the license.
			if (($return ne '') && 
			    ($return ne 'on')){
			    filer_options_set($fc, 'licensed_feature.flexcache_nfs.enable', 'off');
			    logWarn("nfs_cleanup: $@") if $@;
			}
		    }
                }
            }
        }

        if ( $main::CLEANUP ne 'none' && $exports_file_restored ) {
            logStep('Restore all persistent exports');
            eval { export_reset_persistent($fc) };
            logWarn("nfs_cleanup: $@") if $@;
        }

        if ( $main::CLEANUP eq 'test' && $set_h->{volume} ) {
            if ( defined($set_h->{export_list})
              && $set_h->{export_list}{$set_h->{volume}} ) {
                $access = $set_h->{export_list}{$set_h->{volume}};
            } else {
                $access = $gen_access;
            }
            logStep('Restore export options for test volume');
            export_path($fc, $set_h->{volume}, OPTIONS=>$access)
              or logWarn("nfs_cleanup: $@");
        }

        if ( $main::CLEANUP eq 'test' && $set_h->{restore_root_export} ) {
            if ( defined($set_h->{export_list}) and
              $set_h->{export_list}{$root_volume} ) {
                $access = $set_h->{export_list}{$root_volume};
            } else {
                $access = $gen_access;
            }
            logStep('Restore export options for root volume');
            export_path($fc, $root_volume, OPTIONS=>$access)
              or logWarn("nfs_cleanup: $@");
        }

        if ( defined($set_h->{qtree_style_org}) ) {
            logStep("Restore volume qtree security mode setting");
            eval { filer_qtree_security_set($set_h->{fc}, $set_h->{volume},
                                            $set_h->{qtree_style_org}) };
        }

        if ( defined($fc) ) {
            if ( defined($set_h->{aggr_created}) ) {
                # Remove created aggregate
                if ( $main::CLEANUP eq 'all' ) {
                    aggr_purge($fc, $set_h->{aggr_name});
                    logWarn("nfs_cleanup: $@") if $@;
                }
                # Unlicense NFS
                if ( defined($set_h->{nfs_licensed}) ) {
                    unlicense($fc, $nfs_license);
                    logWarn("nfs_cleanup: $@") if $@;
                }
            } else {
                # Run common cleanup and log a warning in order
                # to continue and restore the export options for volume
                # Since common_cleanup deals with $main::FILER_VOL and
                # $main::FILER_AGGR, save the current contents of these
                # variables and set them to the correct values
                $save_vol_type    = $main::VOL_TYPE;
                $save_filer_vol   = $main::FILER_VOL;
                $save_filer_aggr  = $main::FILER_AGGR;
                $main::VOL_TYPE   = $set_h->{vol_type};
                $main::FILER_VOL  = $set_h->{vol_name};
                $main::FILER_AGGR = $set_h->{aggr_name};
                logStep('Run common cleanup');
                common_cleanup($fc) or logWarn("nfs_cleanup: $@");
            }
        }
    };

    if ( $save_cleanup ) {
        # Restore original contents
        $main::CLEANUP = $save_cleanup;
    }

    if ( $save_filer_vol ) {
        # Restore original contents of global variables
        $main::VOL_TYPE   = $save_vol_type;
        $main::FILER_VOL  = $save_filer_vol;
        $main::FILER_AGGR = $save_filer_aggr;
    }
    logWarn("nfs_cleanup: $@") if $@;
    logStep('END: CLEANUP');
    return;
}

#---------------------------------------------------------------------

=head1 testme

Self test routine for this library

Use test_lib.thpl to invoke this routine.

Example
    ntest -hw PKGNAME=nfs::Setup SUBNAME=testme FILER=myfiler \
    util/test_pkg.thpl

=over

=item Arguments

=over

=item FILER

(Required) : Filer to test.

=back

=item Returns

Nothing

=back

=for html <fr />

=cut

#---------------------------------------------------------------------
sub testme() {
    require nfs::CommonParams;
    require Tharn;
    import Tharn qw(param logcomment logresult);
    use vars qw($FILER);
    logcomment("Performing NFS Setup self test");

    param('FILER', '-mesg', 'Filer to test');

    my %setup_h;

    # Setup test
    eval { nfs_setup(setup_h => \%setup_h) };
    logresult('FAIL', "nfs_setup failed: $@") if $@;

    # Cleanup test
    eval { nfs_cleanup(setup_h => \%setup_h) };
    logresult('FAIL', "nfs_cleanup failed: $@") if $@;

    # Do not log a PASS result if there has been a problem
    foreach my $result ( Tharn::own_results() ) {
        return if ( $result->severity() > 2 );
    }

    logresult('PASS', "SELF TEST of nfs/Setup.pm");
}

1;

