#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/Rsierranevada/lib/SPARSE/Util/Filer_Utils.pm#1 $
#
# Copyright (c) 2002 Network Appliance, Inc.
# All rights reserved.

## @summary	Library that defines convenient filer procedures.
## @description
##	This package provides procedural API for quickly setting
##	up sparse volume environments.
##
## @usage
##	use Filer_Utils;
##
## @author dl-sparseeng

package SPARSE::Util::Filer_Utils;

use strict;
use Tharn;

use Exporter qw(import);
BEGIN {
    our @EXPORT	= qw(
		&create_std_arcvol
		&create_std_srcvol

		&sparse_quick_filer_setup
		&sparse_quick_pfs_setup
		&sparse_quick_rod_setup
		&sparse_quick_sv_qtree_setup
		&sparse_quick_sv_vol_setup
		&sparse_quick_sv_setup
		&sparse_quick_common_setup
		&sparse_complete_relation
		&check_supported_release

		&iron_and_check_output
		&wack_and_check_output
		&corrupt_filesystem

		&create_and_restrict_volume

		&setup_cifs
		&terminate_cifs
		
		&setup_http
		&setup_ftp
		&halt_and_reboot
		&simBoot
	        &simHalt
	    );
}

use wack;
use boot;
use cat qw(do_shell);
use Params::Validate qw(HASHREF SCALAR validate);
use SPARSE::Constants;
use SPARSE::Resource::Filer;
use SPARSE::Resource::Volume;
use SPARSE::Util::Logger;
use SPARSE::Util::FSTest0;
use SPARSE::Util::Utilities;
use SPARSE::Util::DataSets qw(create_small_dataset create_large_dataset);
use Volume qw(vol_purge vol_exists vol_destroy vol_root);
use mount qw(mount umount);
use config_base qw(host_is_alive);
use Users qw(removeCIFSUser);
use Data::Dumper;
use Filer::Version qw(:releases get_wafl_version UNSUPPORTED_FEATURE_EXCEPTION);
use CLI::FileIO qw(rdfile wrfile);

# CIFS utils
# use setup;
use CIFS::Setup qw(cifs_setup);
use TestEnv::MinWin qw(:all);

no warnings("uninitialized");

# ------------------------------------------------------------------------

# ------------------------------------------------------------------------
# Internal code
# ------------------------------------------------------------------------

# Internal routine to kill off a volume by conn and volname
sub _purge_vol($$)
{
	my ($hspec, $vname) = @_;
	my $h = _hname($hspec);
	my $filer = sparse_filer($hspec);

	# connect to the vfiler0 console
	my $fc	  = defined($filer->{vfiler0}) ? 
					$filer->{vfiler0}{conn} :
					$filer->{conn};

	if (vol_exists($fc, $vname)) {
		logTrc("killing old dest vol $h $vname");
		vol_purge($fc, $vname)
				or logError("early purge vol failed: $@");
	}
}

# Internal routine to extract a name from a hostspec
sub _hname($)
{
	($_ = $_[0]) =~ s/[\.;].*//;
	return $_;
}

# ------------------------------------------------------------------------
# Sparse dataset creation utilities
# ------------------------------------------------------------------------

## @name	::create_std_arcvol
## @summary	Creates a standard archive volume for RoD testing
## @description
##	Returns the archive volume on the filer if exists
##	Otherwise create the specific dataset and return it
##
## @arg	hspec hostspec where archive volume is located optional,
##					default $main::FILER_DST
## @arg	arc_volname archive volume name optional
## @arg	dataset_type one of: small, large or test optional, default 'small'

sub create_std_arcvol(%)
{
	my (%opt) = validate(@_,
	{
		arc_hspec    => { TYPE=>SCALAR, optional=>1,
					default=>$main::FILER_DST },
		arc_volname  => { TYPE=>SCALAR, optional=>1 },
		dataset_type => { TYPE=>SCALAR, optional=>1,
					default=>$main::DATASET_TYPE }
	}
	);

	my $filer = sparse_filer($opt{arc_hspec}, \%opt);

	my $arc_type = $opt{dataset_type};
	croak("Invalid archive type $arc_type, "
		. " must be small or large")
		unless $arc_type eq "small" || $arc_type eq "large";

	my $arc_volname = $opt{arc_volname} || $arc_type . DFLT_ARC_VOL;
	my $create_dataset = !vol_exists($filer->{conn}, $arc_volname);
	my $arc_vol = sparse_volume($filer, $arc_volname,
					{volSize=>DFLT_STDSRC_SIZE});
	if ($create_dataset) {
		#c66 hardcoded now
		my %ds_opt = (
			vol=>$arc_vol,
			cifs_client=>"c66"
		);
		if ($arc_type eq "small") {
			create_small_dataset(%ds_opt);
		} else {
			create_large_dataset(%ds_opt);
		}
	}

	return $arc_vol;
}

## @name	::create_std_srcvol
## @summary	Creates a standard source volume for RoD testing
## @description
##	Returns the standard source volume on the source filer if exists
##	Otherwise it checks the destination filer for an archive volume
##	and create the standard volume by sv backup to the source filer
##
## @arg	arc_hspec host spec of archive filer optional, default $main::FILER_DST
## @arg	arc_volname specifies the archive volume name optional,
##						default $main::FILER_SRC
## @arg	src_hspec host spec of destination source filer optional,
##						default $main::FILER_DST 
## @arg	src_vol_name specifies the destination standard src volume name optional
## @arg	dataset_type specify dataset_type in the event it is created, optional

sub create_std_srcvol(%)
{
	my (%opt) = validate(@_,
	{
		arc_hspec => { TYPE=>SCALAR, optional=>1,
					default=>$main::FILER_DST },
		arc_volname => { TYPE=>SCALAR, optional=>1 },
		src_hspec => { TYPE=>SCALAR, optional=>1,
					default=>$main::FILER_SRC },
		src_volname => { TYPE=>SCALAR, optional=>1,
					default=>DFLT_STDSRC_VOL },
		dataset_type => { TYPE=>SCALAR, optional=>1,
					default=>$main::DATASET_TYPE }
	}
	);

	my $src_filer = sparse_filer($opt{arc_hspec}, \%opt);
	my $dst_filer = sparse_filer($opt{src_hspec}, \%opt);

	my ($std_volname, $ovol);
	if ($opt{dataset_type} eq "test") {
		# override src volume name, test src volume
		$std_volname = DFLT_SVOL;
		$ovol = sparse_volume($src_filer, "vvolRodTest", {purge=>1});
		
		$src_filer->filer_say("qtree create $ovol->{testDirVolName}");
		FSBuildFileSet($ovol, \%opt);
	} else {
		$std_volname = $opt{src_volname};
		
		if (vol_exists($dst_filer->{conn}, $std_volname)) {
			logInfo("DFLT_STDSRC_VOL chould be a copy of the "
				. "Archive vol, destroy and re-create "
				. "unless it is the std name");
			if ($std_volname eq DFLT_STDSRC_VOL) {
				return sparse_volume($dst_filer, $std_volname);
			}
			vol_destroy($dst_filer->{conn}, $std_volname);
		}

		my %arc_opt;
		$arc_opt{arc_hspec} = $opt{arc_hspec};
		$arc_opt{arc_volname} = $opt{arc_volname};
		$arc_opt{dataset_type} = $opt{dataset_type};
		$ovol = create_std_arcvol(%arc_opt);
		$opt{timeout} = 900 unless $opt{timeout} >= 900;
		$opt{volSize} = "10g" unless exists $opt{volSize};

	}
	return $ovol->snapvault_replicate($dst_filer, $std_volname, \%opt);
}

# ------------------------------------------------------------------------
# Sparse quick setup utilities
# ------------------------------------------------------------------------

## @name	::sparse_quick_filer_setup
## @summary	create a filer and volume
## @description
##	Instantiate sparse Filer and Volume objects.
##	The volume is destroyed first and then mounted.
## @arg		hostspec optional hostspec of the filer.
## @arg		volname optional
## @arg		opts optional {aggr=>.. , volSize=>..}
## @return	volume object
## @example
##	$svol = sparse_quick_filer_setup($FILER, "src");
sub sparse_quick_filer_setup(;$$$)
{
	my ($hspec, $vname, $o) = @_;

	$hspec = $main::FILER_SRC	unless $hspec;
	$vname = DFLT_SVOL		unless $vname;

	my $h = _hname($hspec);
	logTrc("quick setup for $h : $vname");
	my $filer = sparse_filer($hspec, $o);

	my %opts = %$o if defined($o);
	$opts{purge} = 1 unless $o->{nopurge};

	return sparse_volume($filer, $vname, \%opts);
}

## @name	::sparse_quick_pfs_setup
## @summary	create PFS mapping between filer+volume pairs
## @description
##	Instantiate sparse Filer and Volume objects for PFS
##	relationship between source (origin) and destination (proxy).
##	Both volumes are destroyed first and then mounted.
## @arg		src-hostspec optional hostspec of source filer
## @arg		src-volname optional
## @arg		dest-hostspec optional hostspec of destination filer
## @arg		dest-volname optional
## @arg		opts optional {aggr=>}
## @return	source and destination volume objects
## @example
##	($svol, $dvol) =
##		sparse_quick_pfs_setup($FILER_SRC, "src", $FILER_DST, "dst");
sub sparse_quick_pfs_setup(;$$$$$)
{
	my ($shspec, $svname, $dhspec, $dvname, $o) = @_;

	$dhspec = $main::FILER_DST	unless $dhspec;
	$dvname = DFLT_DVOL		unless $dvname;
	$shspec = $main::FILER_SRC	unless $shspec;
	$svname = DFLT_SVOL		unless $svname;

	my $s = _hname($shspec);
	my $d = _hname($dhspec);
	my $buildSrc = $o->{buildSrc} || \&FSBuildFileSet;

	logTrc("quick setup for pfs $s+$svname : $d+$dvname");
	my %srcopts = %$o;
	foreach my $srckey (grep(/^src(Aggr|Vol)/, keys %srcopts)) {
		# When setting up source volume, allow overrides of
		# "aggr|vol" parameters with "srcAggr|srcVol" parameters
		my $key = $srckey;
		$key =~ s/^srcAggr/aggr/;
		$key =~ s/^srcVol/vol/;
		$srcopts{$key} = $srcopts{$srckey};
	}
	if ($srcopts{vol_type} eq 'trad' && !defined($srcopts{tradVol})) {
		$srcopts{tradVol} = 1;
	}
	if (DFLT_SVOL_TYPE eq 'trad' && !defined($srcopts{tradVol})) {
		$srcopts{tradVol} = 1;
	}
	my $svol   = sparse_quick_filer_setup($shspec, $svname, \%srcopts);

	do {	logStep("populate src");
		&$buildSrc($svol, $o); } unless $o->{nofileset};

	my %dstopts = %$o;
	foreach my $dstkey (grep(/^dstAggr/, keys %dstopts)) {
		# When setting up destination volume, allow overrides of
		# "aggr|vol" parameters with "dstAggr|dstVol" parameters
		my $key = $dstkey;
		$key =~ s/^dstAggr/aggr/;
		$key =~ s/^dstVol/vol/;
		$dstopts{$key} = $dstopts{$dstkey};
	}
	my $dfiler = sparse_filer($dhspec, \%dstopts);
	my $dvol   = $svol->setup_sparse_relation("pfs", $dfiler, $dvname,
			\%dstopts);

	return ($svol, $dvol);
}

## @name	::sparse_quick_rod_setup
## @summary	set up and run RoD between filer+volume pairs
## @description
##	Instantiate sparse Filer and Volume objects for RoD
##	relationship between source and destination.
##	Both volumes are destroyed first and then mounted.
##	Populate the source, initiate the RoD relationship.
##	Caller must start the restore.
## @arg		src-hostspec optional hostspec of source filer
## @arg		src-volname optional
## @arg		dest-hostspec optional hostspec of destination filer
## @arg		dest-volname optional
## @arg		opts optional {aggr=>}
## @return	source and destination volume objects
## @notes
##	The need to populate the source before "mapping" the dest
##	means we cannot simply use a version of sparse_quick_pfs_setup()
## @example
##	($svol, $dvol) =
##		sparse_quick_rod_setup($FILER_SRC, "src", $FILER_DST, "dst");
#--------------------------------------------------------------------
sub sparse_quick_rod_setup(;$$$$$)
{
	my ($shspec, $svname, $dhspec, $dvname, $o) = @_;

	$dhspec = $main::FILER_DST	unless $dhspec;
	$dvname = DFLT_DVOL		unless $dvname;
	$shspec = $main::FILER_SRC	unless $shspec;
	$svname = DFLT_SVOL		unless $svname;

	my $s = _hname($shspec);
	my $d = _hname($dhspec);
	my $buildSrc = $o->{buildSrc} || \&FSBuildFileSet;

	logStep("quick setup for rod_remote $s+$svname : $d+$dvname");
	my $svol   = sparse_quick_filer_setup($shspec, $svname, $o);

	do {	logStep("populate src");
		&$buildSrc($svol, $o); } unless $o->{nofileset};

	my $dfiler = sparse_filer($dhspec, $o);
	$dfiler->rod_license($svol->{filer});
	$dfiler->setup_aggr($o->{aggr})	if $o->{aggr};
	my $dvol = $svol->setup_sparse_relation("rod_remote", $dfiler,
		$dvname, $o);

	return ($svol, $dvol);
}

sub sparse_quick_sv_qtree_setup(@)
{ sparse_quick_sv_setup("rod_sv_qtree", @_); }
sub sparse_quick_sv_vol_setup(@)
{ sparse_quick_sv_setup("rod_sv_vol", @_); }

## @name	::sparse_quick_sv_setup
## @summary	set up and run SV RoD between filer+volume pairs
## @description
##	Instantiate sparse Filer and Volume objects for RoD
##	Create relationship between source and destination.
##	Populate the source, initiate the RoD relationship.
##	Volumes are destroyed first and then mounted.
## @arg		mode optional rod_sv_qtree, rod_sv_vol
## @arg		src-hostspec optional hostspec of source filer
## @arg		src-volname optional
## @arg		dest-hostspec optional hostspec of destination filer
## @arg		dest-volname optional
## @arg		opts optional {aggr=>}, {rate=>}
## @return	source and destination volume objects
##
## @example
##	($svol, $dvol) =
## 	     sparse_quick_sv_setup("rod_sv_qtree",
##				   $FILER_SRC, "src", $FILER_DST, "dst");
#--------------------------------------------------------------------
sub sparse_quick_sv_setup(;$$$$$$)
{
	my ($mode, $shspec, $svname, $dhspec, $dvname, $o) = @_;

	$mode   = $mode || $main::SPARSE_TYPE || "rod_snapvault";
	$dhspec = $main::FILER_DST	unless $dhspec;
	$dvname = DFLT_DVOL		unless $dvname;
	$shspec = $main::FILER_SRC	unless $shspec;
	$svname = DFLT_SVOL		unless $svname;

	my $s = _hname($shspec);
	my $d = _hname($dhspec);
	my $buildSrc = $o->{buildSrc} || \&FSBuildFileSet;

	logStep("quick setup for $mode $s+$svname : $d+$dvname");
	my $ovol   = sparse_quick_filer_setup($dhspec, "vvolRodOrig", $o);
	my $dfiler = $ovol->{filer};

	if (!$o->{nofileset}) {
		# create qtree containing fileset
		# XXX qtree only if rod_sv_qtree?
		logTrc("create qtree $ovol->{testDirVolName}");
		$dfiler->filer_say("qtree create $ovol->{testDirVolName}",
				   "^\$");
		&$buildSrc($ovol, $o);
	}

	my $sfiler = sparse_filer($shspec, $o);
	my $svol   = $ovol->snapvault_replicate($sfiler, $svname);

	$dfiler->setup_aggr($o->{aggr})	if $o->{aggr};

	my $dvol = $svol->setup_sparse_relation($mode, $dfiler, $dvname, $o);

	return ($svol, $dvol);
}

## @name	::sparse_quick_common_setup
## @summary	set up a sparse volume relationship between filer+volume pairs
## @description
##	Instantiate sparse Filer and Volume objects for a RoD or PFS
##	relationship between source and destination.
##	Both volumes are destroyed first and then mounted.
##	Populate the source, initiate the svol relationship.
##	Caller must start the restore (if applicable).
## @arg		sparse-type optional $SPARSE_TYPE
## @arg		src-hostspec optional hostspec of source filer
## @arg		src-volname optional
## @arg		dest-hostspec optional hostspec of destination filer
## @arg		dest-volname optional
## @arg		opts optional {aggr=>}
## @return	source and destination volume objects
## @notes
##	The need to populate the source before "mapping" the dest
##	means we cannot simply use a version of sparse_quick_pfs_setup()
## @example
##	($svol, $dvol) =
## 	     sparse_quick_common_setup($SPARSE_TYPE, $FILER_SRC, "src",
##					$FILER_DST, "dst");
#--------------------------------------------------------------------
sub sparse_quick_common_setup(;$$$$$$)
{
	my $mode = shift;
	my ($dhspec, $dvname) = ($_[2], $_[3]);

	# Clean up dst vol first, as it might hold snapshot refs on src
	$dhspec = $main::FILER_DST	unless $dhspec;
	$dvname = DFLT_DVOL		unless $dvname;
	logTrc("common setup: $dhspec $dvname ");
	_purge_vol($dhspec, $dvname);

	$mode = $main::SPARSE_TYPE	unless $mode;
	return &sparse_quick_rod_setup(@_)	if $mode eq "rod_remote";
	return &sparse_quick_sv_qtree_setup(@_)	if $mode eq "rod_sv_qtree";
	return &sparse_quick_sv_qtree_setup(@_)	if $mode eq "rod_snapvault";
	return &sparse_quick_sv_vol_setup(@_)	if $mode eq "rod_sv_vol";
	return &sparse_quick_pfs_setup(@_)	if $mode eq "pfs";
	logError("unknown sparse relationship type: $mode");
}

## @name	::sparse_complete_relation
## @summary	complete and check in-progress sparse relation
## @description
##	Used by frameworks to finish things up.  The following
##	caveats apply:
##		the $vol is a sparse volume
##		and we're the proc that created the vol
##		and we haven't seen an exception or a Tharn abort
## @arg		sparse-vol required
sub sparse_complete_relation($)
{
	my ($dvol) = @_;

	# we want to avoid doing this if there was an error
	# or we might be a child process (subtest)
	return if 
		!defined($dvol)
		|| !defined($dvol->{mode})
		|| $dvol->{volPid} != $$
		|| $@ ne ""
		|| (defined($Tharn::Worst_rslt)
		    && $Tharn::Worst_rslt->is_abortive());

	logTrc("running cleanup");
	$dvol->complete_sparse_relation();
	FSCheckFileSet($dvol);
}

## @name	::check_supported_release
## @summary	check that Src and Dst ONTAP releases support the desired
##		SPARSE relation
## @description
##	Used by scripts to report "NA" result and exit if ONTAP release
##	does not support the SPARSE feature
## @arg		mode optional SPARSE mode (defaults to SPARSE_TYPE)
## @arg		shspec optional Src filer spec (defaults to FILER_SRC)
## @arg		dhspec optional Dst filer spec (defaults to FILER_DST)

sub check_supported_release(;$$$)
{
	my ($mode, $shspec, $dhspec) = @_;
	$mode = $main::SPARSE_TYPE unless ($mode);
	$shspec = $main::FILER_SRC unless ($dhspec);
	$dhspec = $main::FILER_DST unless ($dhspec);
	my $shost = host($shspec);
	my $dhost = host($dhspec);
	my $sver = get_wafl_version($shost);
	my $dver = get_wafl_version($dhost);

	if (defined($mode)) {
		if ($mode =~ /pfs/) {
			if ($sver < WV_ANCHORSTEAM_MIN) {
				die na_result("$mode Server: ".UNSUPPORTED_FEATURE_EXCEPTION. 
						"\n\"$mode\" is unavailable in this version\n");
			}
			if ($dver < WV_ANCHORSTEAM_MIN) {
				die na_result("$mode Cache: ".UNSUPPORTED_FEATURE_EXCEPTION. 
						"\n\"$mode\" is unavailable in this version\n");
			}
		} elsif ($mode =~ /rod/) {
			if ($sver < WV_GORDONBIERSCH_MIN) {
				die na_result("$mode Source: ".UNSUPPORTED_FEATURE_EXCEPTION.
						"\n RoD is unavailable in this version\n");
			}
			if ($sver != $dver) {
				die na_result("$mode Source and Destination: ".
						UNSUPPORTED_FEATURE_EXCEPTION ."\n".
						"RoD Source/Destination wafl versions do not match\n");
			}
		} else {
			logWarn("A check for version compatability of SPARSE_TYPE == $mode"
					." can't be performed!");
		}
	} else {
		logWarn("No SPARSE_TYPE mode defined, a check for version ".
			"compatability of the feature you are testing can't ".
			"be performed!");
	}
}


# ------------------------------------------------------------------------
# WACK/Iron Utilities
# ------------------------------------------------------------------------

## @name	::iron_and_check_output
## @summary	run wafl iron and check the results
## @description
##	Remove /etc/messages after appending it to /etc/messages.0.
##	Start wafliron, wait until it is finished and then check the
##	message file for corruption reports.
## @arg		filer required
## @arg		aggrname required
## @arg         expres optional expected iron result
## @example
##	iron_and_check_output($dfiler, "vol1");
sub iron_and_check_output($$;$)
{
	my ($filer, $aname, $expres) = @_;
	my $istatus;

	logStep("iron $filer->{hostName} : $aname");

	logTrc("check aggregate $aname is not root or already ironing");
	my $response = $filer->filer_say("aggr status $aname",
					 undef, "root|ironing");

	logDbg("mount root vol to flush /etc/messages");
	my $mount_root = mount($filer->{hostName}, "/");
	
	my $file_name = $mount_root . "/etc/messages";
	my $file0_name = $mount_root . "/etc/messages\.0";

	do_shell("sudo touch $file0_name");
	do_shell("sudo chmod 0777 $file0_name");
	do_shell("sudo cat $file_name >> $file0_name");
	do_shell("sudo rm  $file_name");

	logTrc("starting wafliron on aggregate $aname");
	$filer->filer_say("priv set -q advanced; aggr wafliron start $aname");

	do {
		sleep(10);
		$istatus = $filer->filer_say("aggr status $aname");
	} until ($istatus !~ /ironing/);

	logTrc("examine $file_name for any corruption.");
	open(IRONOUT, "<$file_name") or logError("Could not open $file_name!");
	while ($_ = <IRONOUT>) {
		if (/\[\S+\]\: WAFLIRON/) {
			if (!defined($expres)) {
				umount($mount_root);
				logError("WAFLIRON found corruption.\n$_");
			}
		}
	}
	close(IRONOUT);
	umount($mount_root);
}

## @name	::wack_and_check_output
## @summary	run wack and check the results
## @description
##	Use wack()
##	Start wafliron, wait until it is finished and then check the
##	message file for corruption reports.
## @arg		filer required
## @arg		aggrname required
## @arg		timeout required
## @arg		bootimage required
## @arg		result optional expected wack result
## @arg		commit optional whether or not to commit changes
## @example
##	wack_and_check_output($dfiler, "vol1", 3600, "maytag.td");
sub wack_and_check_output($$$$;$$)
{
	my ($filer, $aname, $timeout, $bootimage, $expres, $commit) = @_;
	$expres = WACK_CLEAN if !defined($expres);
	$commit = 0 if !defined($commit);

	logStep("wack $filer->{hostName} : $aname expecting $expres");

	my $result = wack(host($filer->{hostSpec}),
				args=>"$aname",
				timeout=>"$timeout",
				bootimage=>"$bootimage",
				commit=>"$commit");

	logWarn("wack of $filer->{hostName}:$aname "
		 . "result=$result [expecting $expres]"
		 . "$@") if (!($result eq $expres))
}

## @name	::corrupt_filesystem
## @summary	Introduce an inconsistency to the given volume
## @description
##		It will put the given volume in restrict mode,
##		set raid_allow_remote_access flag and introduce	
##		the given inconsistency using hitlist.pl.
##		The hack should have been compiled before using this script.
##		Also the hack and simulators should be running on the same
##		type of machines.
##		If optional arg is set for sparse config metafile,
##		given inconsistency will be introduced to
## 		sparse config metafile in aggregate's hidden metadir.
##		Sparse metafile exists in /vol/<aggr>/vvol/<fsid>/sparse. 
## @arg	 	vol required
## @arg	 	inconsistency required
## @arg		path optional
## @arg		if sparse config metafile optional
## @example
##	corrupt_filesystem($dvol, "inode / nlink \$currInode_nlink+100; \n");
sub corrupt_filesystem($$;$$)
{
	use vars qw($ONTAPDIR);

	my ($vol, $inconsistency, $treetop, $metafile) = @_;
	$treetop = $ONTAPDIR unless defined($treetop);

	logTrc("Check if hack exists");
	logError("hack doesn't exist in $treetop/tools/metadata_tools/hack/")
		unless -e "$treetop/tools/metadata_tools/hack/hack";

	logTrc("Save inconsistency in a temporary file");
	my @hostname = split(/[.;\s]/, $vol->{filer}{hostName});
	my $tmpdir = defined($ENV{TMPDIR}) ? $ENV{TMPDIR} : "/tmp";
	my $tmpfile = "$tmpdir/hackscript.$$";
	my $fh;
	open($fh, ">$tmpfile")
		or logError("Can't create a hack script file $tmpfile");
	if ($metafile) {
		printf $fh 
		       "hack select inode 85.inode \"vvols/%08x/sparse\";\n",
		        $vol->read_fsid();
		print $fh "hack select indirect 0;\n";
	}
	print $fh $inconsistency;
	close($fh);

	logTrc("Set volume in restrict mode and flag for remote raid access");
	$vol->restrict();
	$vol->{filer}->filer_say("priv set test; "
		. "setflag raid_allow_remote_access 1");
	my $volname = $vol->{name};

	if ($metafile) {
		$vol->{filer}->filer_say("aggr restrict $vol->{aggrName}");
	}

	logTrc("Pass the inconsistency to hitlist and hack");
	my $flag = "-" . $hostname[1] if ($hostname[1] eq "sim");
	$volname = "aggr:$vol->{aggrName}" if $metafile;
	system("cd $treetop/tools/metadata_tools/hack; "
		. "hitlist.pl corrupt -f $hostname[0] $flag -v $volname "
		. "-b / -i $tmpfile -noask");

	unlink $tmpfile;
}

# ------------------------------------------------------------------------
# Sparse volume utilities
# ------------------------------------------------------------------------
## @name        ::create_and_restrict
## @summary     creates and restricts a volume
## @arg         dst filer required
## @arg         dst volname required
## @arg         opts optional
sub create_and_restrict_volume($$;$)
{
	my ($filer, $volname, $opts) = @_;
	logTrc("create and restrict volume $volname ");

	my %o = %$opts if defined($opts);
	$o{purge} = 1;

	my $vol = sparse_volume($filer, $volname, \%o);
	$vol->restrict();

	return $vol;
}

# ------------------------------------------------------------------------
# CIFS utilities
# ------------------------------------------------------------------------

## @name	::setup_cifs
## @summary	sets up CIFS on the filer
## @arg		filer spec required
## @arg		CIFSpassword optional
## @arg		AuthenticationType optional
## @arg		WorkGroup optional
## @arg		CIFSDomain optional
## @arg		DomainUser optional
## @arg		DomainUserPassword optional
sub setup_cifs($;$$$$$$)
{
	my ($filerspec, $cifs_passwd, $auth_type, $workgroup, $domain, 
	    $domain_user, $domain_passwd) 
		= @_;

	logTrc("set up CIFS on $filerspec");
	$cifs_passwd = "sundance" unless defined($cifs_passwd);

	# default auth type = 2 i.e. WorkGroup Authentication
	$auth_type   = 2 unless defined($auth_type);
	$workgroup = "workgroup" unless defined($workgroup);

	$domain      = "lab.netapp.com" unless defined($domain);
	$domain_user = "labadmin" unless defined($domain_user);
	$domain_passwd = "sundance" unless defined($domain_passwd);

	# WINDOWS CLIENT SETUP
	# make sure to install and start rshd on win client
	# 'rshd run'
	
	my $host = host($filerspec);
	my $fc = connect("#rsh");
	my $fh = $fc->{'hostp'};
	my $filer_ip = $fh->default_ip();

	my $hosts = rdfile($fc, "/etc/hosts");
	$hosts .= "172.19.2.20     lab.netapp.com\n"
			unless ($hosts =~ /172.19.2.20/);
	wrfile($fc, "/etc/hosts", $hosts);

	logTrc("remove administrator user (if exists) ".
	       "since we are re-creating it with passwd $cifs_passwd");
	my $ret = removeCIFSUser($fc, "administrator")
			or logError($@);
	
	CIFS::Setup::cifs_setup(
			    host        => $filerspec, 
			    cifs_pass   => $cifs_passwd,
			    workgroup   => $workgroup, 
			    #domain      => $domain, 
			    domain_user => $domain_user, 
			    domain_pass => $domain_passwd,
                            passwd_group => 'yes'
			);
	logError($@) if ($@);
	return 1;
}

## @name	::terminate_cifs
## @summary	terminates CIFS on the filer
sub terminate_cifs($)
{
	my ($filer) = @_;

	logTrc("terminate CIFS on $filer->{hostName}");
	$filer->filer_say("cifs terminate",
			  "CIFS local server has shut down");
}

# ------------------------------------------------------------------------
# Miscellaneous utilities
# ------------------------------------------------------------------------

## @name	::setup_http
## @summary	setup HTTP on the filer
## @arg		filer required SPARSE filer object
## @arg		opts optional optional set of arguments http rootdir and timeout
## @example setup_http($dvol->{filer}, { rootdir => "/vol/".$dvol->{name}, 
##									     timeout => 600});   
sub setup_http($;$)
{
	my ($filer, $opts) = @_;

	my $rootdir = $opts->{rootdir} || "/vol/".vol_root($filer->{conn});
	my $timeout = $opts->{timeout} || CMDTIMEOUT;
	
	use Admin::License qw(license);
	license($filer->{conn}, "http")
		|| logError("license http failed : $@");

	$filer->filer_say("options httpd.enable on", "^\$");
	$filer->filer_say("options httpd.rootdir $rootdir", "^\$");
	$filer->filer_say("options httpd.timeout $timeout", "^\$");
}


## @name	::halt_and_reboot
## @summary	halts and reboots filer
##		uses SPARSE/Util/Utilities.pm and boot.pm
## @arg		filer spec required
## @arg		boot image required
sub halt_and_reboot($$)
{
	my ($filerspec, $bootimage) = @_;
	simHalt($filerspec) 
		if $filerspec =~ /simulator/;
	logTrc("booting $filerspec");
	boot(host($filerspec), bootimage=>$bootimage)
		|| logError("bootup of $filerspec failed! $@");
}

## @name        ::simHalt
## @summary     brings down a simulator
## @arg         sim  hostpec or name of filer. required
## @arg         timeout how long to wait for filer 
##                 to go down. optional
sub simHalt($;$)
{
	my ($sim, $timeout) = @_;
	$timeout = CMDTIMEOUT unless defined($timeout);
	
	if ($sim =~ ';') {
		$sim = (split(';', $sim))[0];
	}

	logTrc("bring_down_sim with argument $sim");

	# return if filer is already down
	if (!(host_is_alive("$sim"))) {
		logDbg("$sim is already down");
		return;
	}
	
	logTrc("bringing down $sim");
	subtest("rsh $sim -l root 'halt'","-iffatal", "-ok",-bg,-search,[]);
	logTrc("waiting $timeout seconds for filer to go down");
	sleep($timeout);
	logError("could not bring down the filer $sim")
		if (host_is_alive($sim));
}

## @name        ::simBoot
## @summary     brings up a simulator-requires a
##                 corresponding tharnhostfile in
##                 /usr/local/share/tharn/tharnhosts
## @arg         sim  hostpec or name of filer. required
## @arg         ontap-controls which maytag is envoked. optional.
## @arg         opts -optional
sub simBoot($;$$)
{
	my ($sim, $ontap, $opts) = @_;
	
	$ontap = $ENV{ONTAPDIR} unless defined($ontap);
	my $tharnhost_dir  = "/usr/local/share/tharn/tharnhosts";
	my $sim_name  = (split(';', $sim))[0] if ($sim =~ ';'); #hosttype
	my $tharnhost = (split(/\./, $sim))[0];
	
	logTrc("in sim_boot with arguments: sim: $sim");
	if (!-e "$tharnhost_dir/$tharnhost") {
		logError("sim_boot can only be called for a simulator "
			 ."with a corresponding tharnhost file\n "
			 ."no file $tharnhost is found in $tharnhost_dir")
	}
	
	#(needs to have corresponding tharnhost file).
	my $newfs = $ontap."/test/Filer/newfs.thpl";
	logTrc("launching subtest $newfs for $tharnhost");
	my $subtest = subtest($newfs, "FILER=$tharnhost",
			      "ONTAPDIR=$ontap", 
			      "--", '-console');
	
	logError("could not boot up $sim")
		if ($subtest->exit_status);

	logError("can not ping $sim")
		if (!host_is_alive($sim));
	
	if (defined($opts->{say})) {
		my $conn =  Tharn::connect($sim_name
					   .";hosttype=filer-simulator"
					   . DFLT_FILER_CONN);
		logError("could not connect to filer $sim_name")
			if (!defined($conn));
		say($conn, $opts->{say}, -blind);
	}
	logTrc("Booted $sim");
}

# ------------------------------------------------------------------------
1;
# ------------------------------------------------------------------------
