#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/ontap/main/test/lib/SPARSE/Util/Logger.pm# $
#
# Copyright (c) 2002 Network Appliance, Inc.
# All rights reserved.

## @summary Library that supports Logging operations.
## @description
##	Conditional logging.
##	Log messages are assigned a level and a global log level
##	determines whether a given message will be logged.
##	Logging is layered on NATE logcomment(...)
##	Levels are ordered by verbosity:
##		describe
##		step
##		trc
##		dbg
##		info
##	Enabling a log level enables all lower levels.
##	To conform to NATE we add an error logging message (fatal).
##
##	For example:	logTrc("blah...");
##
## @usage
##	use Logger;
## @author dl-sparseeng

package SPARSE::Util::Logger;

use strict;

use Tharn;

# ------------------------------------------------------------------------

use vars ('@ISA', '@EXPORT');
require Exporter;
@ISA	= qw(Exporter);
@EXPORT	= qw(
		&logXXX

		&logDescribe
		&logStep
		&logTrc
		&logDbg
		&logInfo

		&logPass
		&logWarn
		&logError

		&logParam

		&logLevel
		&logCond
		);

# ------------------------------------------------------------------------

my %logLevels = (
	XXX		=> 0,
	describe	=> 100,
	step		=> 200,
	trc		=> 300,
	dbg		=> 400,
	info		=> 500,
);

# set default log level; log msgs at or below this llevel
my $logLevel = $logLevels{$ENV{'LOGLEVEL'} || "step"}; 


# ------------------------------------------------------------------------

## @name	::logIt
## @summary	conditionally writes a log message
## @description
##		Evaluate log level and determine whether to log.
## @arg		level-name required
## @arg		message required
sub logIt($$)
{
	my ($level_name, $msg) = @_;
	my $lvl = $logLevels{$level_name};
	return if ($lvl > $logLevel);

	chomp($msg) 
		if ($msg =~ /\\n$/);
	
	logcomment("$lvl " . join("\n$lvl ", split(/\n/, $msg)));
}

# ------------------------------------------------------------------------

## @name	::logXXX
## @summary	Hacks and tmp logging
## @arg		message required
sub logXXX($)	{ logIt("XXX", $_[0]); }

## @name	::logDescribe
## @summary	Log the purpose of the test
## @arg		message required
sub logDescribe($)	{ logIt("describe", $_[0]); }

## @name	::logStep
## @summary	Log a step in the test purpose
## @arg		message required
sub logStep($)		{ logIt("step", $_[0]); }

## @name	::logTrc
## @summary	Log a trace messge
## @arg		message required
sub logTrc($)		{ logIt("trc", $_[0]); }
#
## @name	::logDbg
## @summary	Log debugging information
## @arg		message required
sub logDbg($)		{ logIt("dbg", $_[0]); }

## @name	::logInfo
## @summary	Log details
## @arg		message required
sub logInfo($)		{ logIt("info", $_[0]); }

# ------------------------------------------------------------------------

## @name	::logPass
## @summary	Log a pass result
## @arg		message required
sub logPass($)		{ logresult('PASS', $_[0]); }

## @name	::logWarn
## @summary	Log a warning message and continue execution
## @arg		message required
sub logWarn($)		{ logresult('WARN', $_[0]); }

## @name	::logError
## @summary	Log fatal error and exit
## @arg		message required
sub logError($)		{ logresult('FATAL', $_[0]); }

# ------------------------------------------------------------------------

## @name	::logParam
## @summary	Log a paramter to be handed to the parent test
## @arg		variable required
## @arg		value required
sub logParam($$)	{ logresult('PARAM', "$_[0]=$_[1]"); }

# ------------------------------------------------------------------------

## @name	::logLevel
## @summary	Set/query logging level
## @arg		level|value Optional:
## @return	old level
sub logLevel(;$)
{
	$_ = $_[0];
	defined($_) or return $logLevel;
	my $oldLevel = $logLevel;
	if (/\d+/) {
		$logLevel = $_;
	} else {
		defined($logLevels{$_})
			or logError("unknown log level \"$_\", use ["
				. join(' ', keys(%logLevels)) . "]\n");
		$logLevel = $logLevels{$_};
	}
	return $oldLevel;
}

## @name	::logCond
## @summary	Determine whether a log level is in use.
## @example
##		if (logCond("debug")) { ... }
## @arg		level required
## @return	true if logging currently enabled for this level
sub logCond($)
{
	my $lvl = $_[0];
	defined($logLevels{$lvl}) or logError("bad log level $lvl");
	return $logLevels{$lvl} <= $logLevel;
}

# ------------------------------------------------------------------------
1;
