#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/nfs/Locks.pm#7 $
#
## @summary Locks - Common functionality for testing NFS locks
## @author	dl-maryv (FSS Protocols QA)
## @status	public
## @notes	Created by EP Network Storage Performance Lab, Inc.
## @pod here

=head1 NAME

Locks - Common functionality for testing NFS locks

=head1 SYNOPSIS

  use nfs::Locks;

  my %setup_h;
  $out = nfslocks_setup(setup_h => \%setup_h,
                        clients => @clients,
                        mount => 1);
  $out = lock_check(fc => $fc, owner => { pid=>$$, uid=>1234 },
                    ipaddr => '10.0.0.25', type => 'read',
                    file   => '/vol/testvol/lock_file');
  $out = nsm_unreachable_make(fc => $fc,
                              clients => [{ ipaddr=>'10.0.0.4',
                                            status => 'noresp' },
                                          { ipaddr=>'10.0.0.6',
                                            status => 'unreach' }]);
  $out = name_services_unreachable_make(fc => $fc);
  $out = nfslocks_filer_state_restore(fc => $fc);
  nfslocks_cleanup(%setup_h);

=head1 DESCRIPTION

Provides a common functionality for testing NFS locks.
It includes functions for common setup and cleanup.

=head1 STATUS

  Shared

=head1 AUTHOR

  dl-maryv (FSS Protocols QA)
  $Id: //depot/prod/test/main/lib/nfs/Locks.pm#7 $
  Last revised $Date: 2008/07/03 $

=head1 COPYRIGHT

  Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ROUTINES

=cut

package SPARSE::Frameworks::Locks;

use strict;
use warnings;
use base qw(Exporter);

use nfs::Rpc;
use Carp qw(croak);
use POSIX qw(uname);
use Tharn qw(host connect);
use CLI::FileIO qw(rdfile wrfile);
use GenericMount qw(mount umount);
use Network::Nis qw(get_nis_info);
use Network::Utils qw(run_getXXbyYY);
use nfs::Utils qw(validate_nfs_version);
use FileManip qw(write_file remove_file);
use TestEnv::MinWin qw(minwin_compatible);
use SPARSE::Util::Logger qw(logStep logWarn);
use MapShare qw(map_cifs_share unmap_cifs_share);
use nfs::GenericExport qw(export_path list_exports);
use Params::Validate qw(validate UNDEF SCALAR OBJECT ARRAYREF HASHREF);
use FS::Util::Setup qw(create_vol common_cleanup);
use SPARSE::Frameworks::filer qw(filer_options_get filer_options_set filer_qtree_get
             filer_qtree_security_set filer_lock_status_file_get);
use ExportFS qw(share_add share_delete);
use Filer::Version qw(get_wafl_version supported_feature UNSUPPORTED_FEATURE_EXCEPTION are_we_ontapng);
use Data::Dumper;

BEGIN {
    # Die with a script error when an
    # input parameter fails validation
    Params::Validate::validation_options(
        on_fail => sub { my @msgs = @_;
                         chomp(@msgs);
                         Carp::confess(Tharn::script_result(@msgs)) }
    );
};

our $VERSION = '1.0';

our @EXPORT_OK = qw(
  lock_check
);

#---------------------------------------------------------------------
# Function prototypes
#---------------------------------------------------------------------
sub lock_check(%;);

#---------------------------------------------------------------------
# Package local variables
#---------------------------------------------------------------------
my @file_locks;
my $export_list;
my $nfs_v4_id_domain;
my $nfs_v4_enable;
my $nfs_v4_rdel;
my $nfs_v4_wdel;
my $mt_root_only;
my %host_ipaddr_h;
my %nis_servers_h;
my %rc_contents_h;
my %hosts_contents_h;
my $hosts_file  = '/etc/hosts';
my $rc_file     = '/etc/rc';
my $dns_file    = '/etc/resolv.conf';
my $nsswitch    = '/etc/nsswitch.conf';
my $test_file   = "NFSlocks.$$";
my $nsm_program = 100024;
my $usermap = '/etc/usermap.cfg';
my $filer_usermap_text;
my $cifs_share_completed = 0;
my $qtree_sec_saved;

#---------------------------------------------------------------------


#---------------------------------------------------------------------
sub lock_check(%;) {
    my (%opts) = validate (
        @_,
        {
            fc      => { type => OBJECT },
            owner   => { type => HASHREF },
            file    => { type => SCALAR, default => '' },
            ipaddr  => { type => SCALAR, regex => qr/^\d+\.\d+\.\d+\.\d+$/ },
            type    => { type => SCALAR, regex => qr/^(read|write)$/ },
            inode   => { type => SCALAR, regex => qr/^\d+$/, optional => 1 },
            sbyte   => { type => SCALAR, regex => qr/^\d+$/,  default => 0 },
            blength => { type => SCALAR, regex => qr/^\d+$/,  default => 0 },
            buffer  => { type => SCALAR, regex => qr/^[01]$/, default => 0 },
            partner => { type => SCALAR, regex => qr/^[01]$/, default => 0 }
        }
    );
    my ($lock_h, $inode, $owner, $out);

    # Hash to map lock type on filer to input types
    my %lock_type = ( 0=>'read', 1=>'write' );
    my %cifs_type = ( no=>'read', yes=>'write' );

	#mica 10 Mode locks information  
	my $VSERVER_NAME = $main::VSERVER_NAME;
 
	#my %lock_info = (fc => $fc_origin,
	#                         type => 'read',
	#                         ipaddr => $ipaddr,
	#                         owner => get_lock_owner(),
	#                         sbyte => 0,
	#                         blength => 0,
	#                         file => "/vol/$FILER_VOL/$file");
	#        eval { lock_check(%lock_info) };
	#     vserver_locks_show => {
	#              command => "vserver locks show",
	#              options => \%Options,
	#              required_options => undef,
	#              optional_options => [qw(path volume bytelock-mandatory bytelock-length vserver oplock-level vif 
	#				  protocol bytelock-exclusive lockid bytelock-offset sharelock-mode type node )],

	if (my $server = are_we_ontapng($opts{fc})) {
		logStep("##mica Checking the 10M lock");
        sleep(5);
		my (@outputArray,$convPath);
		eval {
        	@outputArray = $server->vserver_locks_show(vserver=>$VSERVER_NAME);
		    foreach(@outputArray)
		    {
					logStep("volume = ". $_->{'volume'} . ", path = ". $_->{'path'});
					$convPath = "/vol/". $_->{'volume'} ."/". $_->{'path'} ;
					#mica check that locks exist as per request from script 
					if ( ($opts{file} =~ /$convPath/ ) ) {
						logStep("Lock exists for this file $opts{file} in 10M filer as $convPath");
						return 1;
					} 
			}
			return 0;
		};
		if ($@) {
			return 0;
		} else {
			return 1;
		}
	}


    # Get all locks on filer -- grouped by file inode
    if ( ! $opts{buffer} ) {
        @file_locks = filer_lock_status_file_get($opts{fc},
                                                 file    => $opts{file},
                                                 partner => $opts{partner});
    }

    # Set the correct inode if only the file parameter was passed
    if ( @file_locks and $opts{file} ne '' and !defined($opts{inode}) ) {
        $opts{inode} = $file_locks[0]{inode};
    }

    foreach $lock_h ( @file_locks ) {
        # Only process file locks
        next unless exists($lock_h->{sbyte});

        # Convert hostname to ipaddr for NLM locks
        # from the point of view of the filer
        if ( !exists($lock_h->{ipaddr}) ) {
            # Get the ipaddr of host from save copy
            if ( !($lock_h->{ipaddr} = $host_ipaddr_h{$lock_h->{host}}) ) {
                # This is a new host, no ip address is saved,
                # get the ipaddress from the filer
                $out = run_getXXbyYY($opts{fc}, 'gethostbyname_r',
                                     $lock_h->{host});
                ($out) = ($out =~ /addresses:\s+(\d+\.\d+\.\d+\.\d+)/);
                $lock_h->{ipaddr} = $out;
                # Save host and ipaddr for faster subsequent lookups
                $host_ipaddr_h{$lock_h->{host}} = $out;
            }
        }

        # Get correct lock owner according to protocol
        $owner = $opts{owner}{($lock_h->{proto} eq 'NLM' ? 'pid' : 'uid')};
        $owner = '' unless defined($owner);

        # On CIFS ignore blength (whole file is locked with flock)
        my $ltype_test = $lock_h->{proto} eq 'CIFS' ?
          ( $cifs_type{$lock_h->{excl}} eq $opts{type} ) :
          ( $lock_type{$lock_h->{ltype}} eq $opts{type}
           and $lock_h->{blength} == $opts{blength} );

        # Check if this is the lock
        if ( $ltype_test
          and $lock_h->{state} eq 'GRANTED'
          and $lock_h->{inode} == $opts{inode}
          and $lock_h->{owner} eq $owner
          and $lock_h->{sbyte} == $opts{sbyte}
          and $lock_h->{ipaddr} eq $opts{ipaddr} ) { 
            return 1;
		}
    }

    # Lock not found
     $inode = defined($opts{inode}) ? sprintf(' %d <0x%x>', $opts{inode}, $opts{inode}) :croak "Could not find $opts{type} lock for " ."inode$inode on client $opts{ipaddr}";
     #croak "Could not find $opts{type} lock for " ."inode$inode on client $opts{ipaddr}";
}

1;

