#$Id: //depot/prod/test/main/lib/SPARSE/Resource/Vol.pm#1 $
#
#Hey emacs, this is basically -*- perl -*-
#vim:filetype=perl
#
# Copyright (c) 2004 Network Appliance, Inc.
# All rights reserved.

## @summary	Library which defines a Volume object.
## @description
##	Abstract common volume tasks and data.
##	Reuse applicable NATE lib functionality.
##
##	This is the base "volume" class; the intent is for
##	other kinds of application volumes (e.g., PFS, RoD,
##	dense) to inherit from this class. The first example
##	of inheritance is SPARSE::Resource::Volume.pm
##
## @usage
##	use SPARSE::Resource::Vol;
## @author dl-sparseeng

package SPARSE::Resource::Vol;

use strict;
use Tharn;
use SPARSE::Frameworks::filer;
use mount();
use SPARSE::Frameworks::Volume qw(vol_exists vol_purge vol_status
		vol_create_flex vol_create vol_destroy
		vol_online vol_offline vol_restrict
		vol_container
		);
	# WARNING: see burt105917.
	# vol_create_flex is subject to being renamed as "vol_create".
	# vol_create is subject to being renamed as "vol_create_paired".
use Data::Dumper;
use SPARSE::Constants;
use SPARSE::Util::Logger;
use SPARSE::Util::Utilities;

# ------------------------------------------------------------------------

use vars ('@ISA', '@EXPORT');
require Exporter;
@ISA	= qw(Exporter);
@EXPORT	= qw(
		&vol
		);

# ------------------------------------------------------------------------

## @name	::volume
## @summary	create a volume object
## @description
##	Create or assimilate volume on filer.
## @arg		filer required
## @arg		volname required
## @arg		opts optional
## @return	volume object
sub vol($$;$)
{
	my ($vfiler, $name, $opts) = @_;
	my $vol = {};

	$vol->{name}	 = $name;
	$vol->{filer}	 = $vfiler;

	logTrc("volume name=$name filer=$vfiler->{hostName}");
	_vol_create($vol, $opts);

	logInfo("volume " . Dumper($vol));
	bless($vol);

	if (!$opts->{subTest}) {
	    $vol->exportfs($opts);
	}
	$vol->mount($opts) unless $opts->{nomount};	# user will do it later
	return $vol;
}

sub DESTROY($)
{
	my ($vol) = @_;

	my $exc_state = $@;

	# if this process did not mount, it will not unmount
	$vol->umount() if ($vol->{mountPid} == $$);
	logDbg("destroying $vol->{name} object");

	$@ = $exc_state;
}

## @name	::exportfs
## @summary	export the volume
## @arg		opts optional
## @example	$vol->exportfs()
sub exportfs($;$)
{
	my ($vol, $opts) = @_;

	my $o =  ($opts->{noPersistExport}) ? "-io" : "-p";
	$vol->{filer}->filer_say("exportfs $o anon=0 /vol/$vol->{name}", "^\$");
}

## @name	::mount
## @summary	mount the volume
## @arg		opts optional
## @example	
##		$vol->mount({mountopts=>"vers=2"});
##		For nfsv4 ACLS,
##			$vol->mount({nfsAcls=1})
sub mount($;$)
{
	my ($vol, $opts) = @_;

	my $host   = $vol->{filer}{hostName};
		my $fname   = "/vol/$vol->{name}";
		my $mnt     = "/t/$host/$vol->{name}";
		my $mntopts = $opts->{mountopts} ? $opts->{mountopts}
						 : "actimeo=0,noac,hard,intr";

		logTrc("mount $host:$fname $mnt");
		$vol->{mountPoint} = mount::mount($host, $fname,
				mountpoint=>$mnt,
				mountopts=>$mntopts,
				vers=> $opts->{nfsAcls} ? "4" : "",
				proto=> $opts->{nfsAcls} ? "tcp" : "",
				timeout=> CMDTIMEOUT,
				)
			or logError("mount failed: $@");

	$vol->set_testdir();
	$vol->{mountPid} = $$;
}

## @name	::umount
## @summary	unmount the volume
sub umount($)
{
	my ($vol) = @_;

	return unless $vol->{mountPoint};

	logTrc("unmount $vol->{mountPoint}");
		# NATE workaround
		# cannot run mount::umount($vol->{mountPoint}) in a destructor
		uSystem(mount::_umount_command(
				host("localhost")->hosttype(),
				$vol->{mountPoint}))
			and logError("unmount failed: $!");

	$vol->{mountPoint} = undef;
	$vol->{mountPid} = 0;
}

## @name	::set_testdir
## @summary	set test dir on volume
## @arg		dir optional default is DFLT_TEST_DIR
sub set_testdir($;$)
{
	my ($vol, $dir) = @_;
	$dir = DFLT_TEST_DIR	unless $dir;

	$vol->{testDirBaseName} = $dir;
	$vol->{testDir} = "$vol->{mountPoint}/$dir";
	$vol->{testDirVolName} = "/vol/$vol->{name}/$dir";
	logTrc("set test dir name=$vol->{testDir}");
}

## @name	::make_testdir
## @summary	create test dir on volume
## @arg		dir optional
sub make_testdir($;$)
{
	my ($vol, $dir) = @_;

	$vol->set_testdir($dir);

	# set up dir if needed
	if (!-e $vol->{testDir}) {
		logTrc("create test dir name=$vol->{testDir}");

		# XXX need root priv
		system("r mkdir -p $vol->{testDir}")
			and logError("mkdir $vol->{testDir} failed");
	}
	# XXX need root priv
	system("r chmod 0777 $vol->{testDir}")
		and logError("chmod $vol->{testDir} failed");
}

## @name	::mkfile
## @summary	run "mkfile" on a filer
## @arg		size required
## @arg		basename required
## @example	$vol->mkfile("32m", "$vol->{testDirBaseName}/bigfile");
sub mkfile($$$)
{
	my ($vol, $size, $basename) = @_;
	my $fc	  = defined($vol->{filer}{vfiler0}) ? 
					$vol->{filer}{vfiler0}{conn} :
					$vol->{filer}{conn};

	return filer_file_make($fc, "/vol/$vol->{name}/$basename", $size);
}

## @name	::rawstat
## @summary	Lookup a raw CM stat
## @arg		counter required
## @arg		val optional expected match
## @return	if above val specified confirm a match, else value of counter
## @example	for getting a value
##			my $stat = $dvol->rawstat(
##				"wv_fsinfo_blks_used_by_plane0");
## 		for checking a value
##			$dvol->rawstat("wv_fsinfo_blks_used_by_plane0", 1000);
sub rawstat($$;$)
{
	my ($vol, $counter, $val) = @_;
	my $object = "volume";
	my $instance = $vol->{name};
	my $key = "$object:$instance:$counter";

	my $ret = $vol->{filer}->stats({raw => 1, key => $key});
	my $value = $ret->{$object}{$instance}{$counter};

	if (!defined($val)) {
		return $value;
	}

	if ($val != $value) {
		logError("Value of $key was $value ".
			 "and not the expected value $val");
	}
}

# ------------------------------------------------------------------------
#			internal routines
# ------------------------------------------------------------------------

## @name	::_vol_create
## @summary	create a (flexible) volume on a filer (internal)
## @arg		size required
## @arg		purge required
sub _vol_create($$$)
{
	my ($vol, $opts) = @_;

	# get the vfiler0 console
	my $vfiler0 = defined($vol->{filer}{vfiler0})
				? $vol->{filer}{vfiler0}
				: $vol->{filer};
	my $fc	    = $vfiler0->{conn};
	my $vname   = $vol->{name};

	my $exists = vol_exists($fc, $vname);
	logTrc("vol $vname ". ($exists ? "already exists" : "does not exist"));

	if ($exists && $opts->{purge}) {	# exterminate
		logTrc("purging vvol $vname");
		vol_purge($fc, $vname)
			or logError("purge vol $vname failed: $@");
		$exists = 0;
	}
	logError("volume \"$vname\" does not exist")
		if (!$exists && $opts->{nocreate});

	# create and online vol if necessary
	if (!$exists) {
		if ($opts->{tradVol}) {
			my $disklist = $opts->{aggrSize} || DFLT_AGGR_SIZE;
			logTrc("creating pvol $vname");
			vol_create($fc, $vname, $disklist, timeout=>CMDTIMEOUT)
				or logError("create aggr $vname failed: $@");
		} else {
			# pick aggr and set it up
			$vol->{aggrName} = ($opts->{aggr} || DFLT_AGGR);
			$vfiler0->setup_aggr($vol->{aggrName}, $opts);

			my $size = $opts->{volSize} || DFLT_VOL_SIZE;
			logTrc("creating vvol $vname size=$size");
			vol_create_flex(
				$fc, $vname, $vol->{aggrName}, $size,
				timeout	=> CMDTIMEOUT,
				$opts->{volOpts} ? @{$opts->{volOpts}} : ()
				)
				or logError("create volume $vname failed: $@");
		}
	} elsif (!$opts->{tradVol}) {
		# exists and is a vvol -- fill in the aggr
		$vol->{aggrName} = vol_container($fc, $vname);
	}

	# finally move the newly created volume to the vfiler
	if (defined($vol->{filer}{vfiler0})) {
		my $vfiler = $vol->{filer};
		$vfiler0->filer_say("vfiler move vfiler0 "
				   ."$vfiler->{vfilerName} /vol/$vname");
	}

	$vol->{volPid} = $$;	# creating process
}
