# -*- perl -*-
# $Id: //depot/prod/test/main/lib/SPARSE/Util/DataSets.pm#1 $

## @summary DataSets creation and manipulation utilities
## 
## @author	dl-dpgqa-automation@netapp.com
## @status  Shared
## @pod		here

=head1 NAME

DataSets - Routines to create data for testing

=head1 SYNOPSIS

	use SPARSE::Util::DataSets qw (
						create_small_dataset
						create_large_dataset
						create_dataset
						read_dataset
						);

=head1 AUTHOR

	dl-dpgqa-automation@netapp.com

=head1 STATUS

	Shared

=head1 COPYRIGHT
	
	Copyright 2005, Network Appliance Inc., All Rights Reserved.

=head1 ATTRIBUTES
	
	ONTAP_COMMANDS  = NONE (directly)
	ONTAP_ZAPI_APIS = NONE (directly)

=head1 DESCRIPTION

Module containing routines to create test data

Routines to create test data given a SPARSE::Resource::Volume object
The data created includes:
	- FSBuildFileSet
	- compat data
	- Directories of a certain depth, and data files per directory
	- NFS special files: sockets, named pipes, fifo,s device files and 
	  hardlinks
	- Level 0 - 4 files with various % of fill from 0% - 100%

The directory structure of the volume is as follows:

	TEST - contains FSBuildFileSet data
	TEST/level_dir - which contains level 0 - 5 files, at different fill 
			 level

	compat_data - containing compat_data

	qtree_0/dir_width - wide directory
	qtree_0/dir_depth - level <n> deep directory
	qtree_0/special_files - Unix special files, sockets, device file and 
				named pipe
	qtree_0
	.
	.
	qtree_<n> - minimum number of qtrees

	qtree_1/cifs_dir - contain cifs data


=head1 ROUTINES

There are 2 pre-defined data sets that this module contains routines to create,
a small and a large data set. In addition to FSBuildFileSet data and compat
data, the data sets created contains a number of qtrees, default  10, named
nfs_qtree<n>, where <n> starts from 0. Under nfs_qtree0, the following are
created:

	- a wide directory, default 100 sub directories
	- a deep directory, default 10

Additional interesting files are files from level 0 - 5, these are located
at TEST/nfs_files, and includes the following:

	- files per directory depth, default 45
	- level of files to create, default 2, files at different file perc at 
	  each level

For the large data set creation, files of level 5 are also created.

=cut

package SPARSE::Util::DataSets;

use strict;
use warnings;

use Carp;

use Params::Validate qw(HASHREF SCALAR validate);

use waflobjCreator qw (	create_socket
						create_dev
						create_fifo
						create_symlink
						create_hardlink
		      );

use QtreeCmds qw(qtreeCreate qtreeSecurity qtreePresent);
use blocks_filer qw(filer_lun_create filer_lun_exists);
use ExportFS qw(export_path unexport_path share_add share_delete share_exists);
use SPARSE::Resource::Volume qw (sparse_volume);
use SPARSE::Util::FSTest0 qw(FSBuildFileSet FSCheckFileSet FSModifyFileSet);
use FSR::FSRutils qw(load_compat_data);
use DataManip;
use Tharn;

use SPARSE::Util::FileIO;
use SPARSE::Util::Logger;

require Exporter;
our @ISA = qw (Exporter);
our @EXPORT_OK = qw (
					create_small_dataset
					create_large_dataset
					create_dataset
					read_dataset
					modify_dataset
					);

=head2 create_small_dataset

Routine to create small data set given a Client and a SPARSE::Resource::Volume
object Pass throught routine, which calls create_dataset using pre-defined
%small_opt

=over

=item params %opt
	vol - sparse volume object where dataset is to be created
	cifs_client - cifs client where cifs data will be created
	nfs_client - nfs client where nfs client will be created
				 default to localhost

=back

=cut

sub create_small_dataset(%)
{
	my (%opt) = validate(@_,
		{
		    vol			=> { TYPE=>HASHREF, optional=>0 },
			cifs_client => { TYPE=>SCALAR, optional=>0 },
			nfs_client 	=> { TYPE=>SCALAR, optional=>1, default=>"localhost" }
		}
	);
	
	my %small_opt = (
			min_qtree => 10
		,	min_dir_width => 100
		,	max_dir_depth => 10
		,	max_files_per_depth => 45
		,	max_level_file => 2
		,	file_level_fill_perc => 10
		,	lun_size => "4m"
	);
	$small_opt{cifs_client} = $opt{cifs_client};
	$small_opt{nfs_client} = $opt{nfs_client};
	$small_opt{vol} = $opt{vol};
	
	create_dataset(%small_opt);
}

=head2 create_large_dataset

Routine to create large test data set given a Client and a
SPARSE::Resource::Filer object calls create_data_set with a pre-defined
%large_opt to specify large data sets

=over

=item params %opt
	vol - sparse volume object where dataset is to be created
	cifs_client - cifs client where cifs data will be created
	nfs_client - nfs client where nfs client will be created
				 default to localhost

=back

=cut

sub create_large_dataset ($%)
{
	my (%opt) = validate(@_,
		{
			vol => { type => HASHREF },	
			cifs_client => { TYPE => SCALAR, optional => 0 },
			nfs_client => { TYPE=>SCALAR, optional=>1, default=>"localhost" }
		}
	);

	logXXX("create_large_dataset: "
		. "This routine have not been tested yet, you're on your own");
	my %large_opt = (
			min_qtree => 200
		,	min_dir_width => 500
		,	max_dir_depth => 1000
		,	max_files_per_depth => 45
		,	max_level_file => 5
		,	file_level_fill_perc => 80
		,	lun_size => "1t"
	);
	$large_opt{cifs_client} = $opt{cifs_client};
	$large_opt{nfs_client} = $opt{nfs_client};
	$large_opt{vol} = $opt{vol};

	create_dataset(%large_opt);
}


=head2 create_dataset

	Generic routine to create a dataset which is determined by the entries
	in parameter %opt

=over

=item param %opt
	vol	=> SPARSE::Resource::Volume object
	cifs_client => hostspec of CIFS client. Required
	nfs_client => hostrec of NFS client. Default localhost
	min_qtree => Minimum number of required qtrees.
	min_dir_width => Max number of directories at root level.
	max_dir_depth => Max directory depth.
	max_files_per_depth => max number of files at each dir level.
	max_level_file => Max file level to be created.
	file_level_fill_perc => Perc. of file size of each level to fill.
	lun_size => Required Lun size.

=back

=cut

sub create_dataset(%)
{
	my (%opt) = @_;

	$opt{op_type} = "create";
	_do_dataset(%opt);
}

=head2 read_dataset

	Routine to read the given datasets

=over

=item param %opt
	vol	=> SPARSE::Resource::Volume object
	cifs_client => hostspec of CIFS client. Required
	nfs_client => hostrec of NFS client. Default localhost
	min_qtree => Minimum number of required qtrees.
	min_dir_width => Max number of directories at root level.
	max_dir_depth => Max directory depth.
	max_files_per_depth => max number of files at each dir level.
	max_level_file => Max file level to be created.
	file_level_fill_perc => Perc. of file size of each level to fill.
	lun_size => Required Lun size.

=back

=cut 

sub read_dataset(%)
{
	my (%opt) = @_;

	# XXX assume small dataset for now.. 
	$opt{nfs_client} = "localhost";
	$opt{min_qtree} = 10;
	$opt{min_dir_width} = 100;
	$opt{max_dir_depth} = 10;
	$opt{max_files_per_depth} = 45;
	$opt{max_level_file} = 2;
	$opt{file_level_fill_perc} = 10;
	$opt{lun_size} = "4m";

	# we always read 
	$opt{op_type} = "read";
	_do_dataset(%opt);
}

=head2 modify_dataset

	Routine to modify the datasets in a pre-defined way

=over

=item param %opt
	Same as those supplied to read_datasets

=back

=cut

sub modify_dataset(%)
{
	my (%opt) = validate(@_,
		{
			vol => { type => HASHREF },	
			cifs_client => { TYPE => SCALAR, optional => 0 },
			nfs_client => { TYPE=>SCALAR, optional=>1, default=>"localhost" },
		 	data_subtype => {type=>SCALAR, optional=>1, default=>"level_files",
				regex => qr/^level_files|FSBuild|compat_data|cifs_data|special$/}
		}
	);

	# XXX assume small dataset for now.. 
	$opt{nfs_client} = "localhost";
	$opt{min_qtree} = 10;
	$opt{min_dir_width} = 100;
	$opt{max_dir_depth} = 10;
	$opt{max_files_per_depth} = 45;
	$opt{max_level_file} = 2;
	$opt{file_level_fill_perc} = 10;
	$opt{lun_size} = "4m";

	# we always modify
	$opt{op_type} = "modify";
	_do_dataset(%opt);
}

# -------------------------------------------------------------------------
# private workhorse methods
# -------------------------------------------------------------------------

sub _do_dataset(%)
{
	my (%opt) = validate(@_,
		{	vol	=> { type => HASHREF }
		,	cifs_client => {type=>SCALAR, optional=>0}
		,	nfs_client => {type=>SCALAR, optional=>0}
		,	min_qtree => {type=>SCALAR, optional=>0}
		,	min_dir_width => {type=>SCALAR, optional=>0}
		,	max_dir_depth => {type=>SCALAR, optional=>0}
		,	max_files_per_depth => {type=>SCALAR, optional=>0}
		,	max_level_file => {type=>SCALAR, optional=>0}
		,	file_level_fill_perc => {type=>SCALAR, optional=>0}
		,	lun_size => {type=>SCALAR, optional=>0}
		, 	op_type	 => {type=>SCALAR, optional=>0, 
						regex => qr/^create|read|modify$/}
		,	data_subtype => {type=>SCALAR, optional=>1, default=>"level_files"}
		}
	);

	my $vol = $opt{vol};

	# 1. compat data
	# XXX load_compat_data uses 'restore' to load the data files, and 
	# 'restore' at present causes filer to panic
# XXX compat_data cannot be restored due to filer bug
#	if ($opt{op_type} eq 'create') {
#		load_compat_data($vol->{filer}->{hostName}, "/vol/$vol->{name}")
#			or croak("load_compat_data Failed, @_");
#	} elsif ($opt{op_type} eq 'read') {
#		# XXX TBD
#		logTrc("TBD");
#	} elsif ($opt{op_type} eq 'modify') {
#		# XXX TBD
#		logTrc("TBD");
#	}

	# 2. FSBuildFileSet
	if ($opt{op_type} eq 'create') {
		# make test_dir a qtree
		$vol->{filer}->filer_say("qtree create $vol->{testDirVolName}");
		FSBuildFileSet ($vol, \%opt);  #XXX does not propagate errors
	} elsif ($opt{op_type} eq 'read') {
		logTrc("checking and read verifying data");
		FSCheckFileSet($opt{vol}); # XXX does not propagate errors
	} elsif ($opt{op_type} eq 'modify') {
		# XXX does not propagate errors
		FSModifyFileSet($opt{vol}) if $opt{data_subtype} eq "FSBuild";
	}

	# 3. Level 0 - 5 files
	my $level_dir = "level_dir";
	if ($opt{op_type} eq 'create') {
		fioMkDir($vol, $level_dir); # XXX does not propagate errors
	} elsif ($opt{op_type} eq 'read') {
		# verify if dir exists
		logTrc("TBD");
	} elsif ($opt{op_type} eq 'modify') {
		# modify dir ?
		logTrc("TBD");
	}
	_do_level_files($vol, $level_dir, %opt)
		or croak("Failed to create various level files, @_");

	# 4. qtrees of various types
	my $qtree_prefix = "qtree_";
	my $qtree_cnt = 0;
	my $qtree_name = undef;
	logTrc($opt{op_type} . " " . $opt{min_qtree} ." qtree(s)");
	do {
		$qtree_name = "/vol/".$vol->{name}."/".
				 $qtree_prefix.$qtree_cnt;

		if ($opt{op_type} eq 'create') {
			logDbg("Creating qtree $qtree_name");
			qtreeCreate($vol->{filer}->{conn}, $qtree_name)
				or croak("Failed to create $qtree_name, @_");
		} elsif ($opt{op_type} eq 'read') {
			qtreePresent($vol->{filer}->{conn}, $qtree_name)
				or croak("qtree $qtree_name does not exist!");
		} elsif ($opt{op_type} eq 'modify') {
			qtreeSecurity($vol->{filer}->{conn}, $qtree_name, "mixed")
				or croak($@);
		}

		$qtree_cnt++;
	} while $qtree_cnt < $opt{min_qtree};

	# 5. lun
	my $lun_path = $vol->{testDir}."/lun";
# XXX lun restore filer bug
#	if ($opt{op_type} eq 'create') {
#		filer_lun_create($vol->{filer}->{conn}, $lun_path, $opt{lun_size});
#		croak("Failed to create lun $lun_path, lun size $opt{lun_size}, @_")
#			unless filer_lun_exists($vol->{filer}->{conn}, $lun_path);
#	} elsif ($opt{op_type} eq 'read') {
#		filer_lun_exists($vol->{filer}->{conn}, $lun_path)
#			or croak ("lun $lun_path does not exist!");
#	} elsif ($opt{op_type} eq 'modify') {
#		# modify a lun?
#		logTrc("TBD");
#	}

	# mount the volume if required
	my $need_umount = ! $vol->{mountPoint};
	if ( $need_umount ) {
		$vol->exportfs();
		$vol->mount();
	}
	croak("Failed to chmod on $vol->{mountPoint}")
		unless r_chmod(777, $vol->{mountPoint});

	# 6. deep and wide dirs
	my $nfs_path = $vol->{mountPoint}."/".$qtree_prefix."0";
	if ($opt{op_type} eq 'create') {
		_do_nfs_dirs($opt{nfs_client}, $nfs_path, %opt)
			or croak("Failed to create nfs data");
	} elsif ($opt{op_type} eq 'read') {
		# XXX TBD
		logTrc("TBD");
	} elsif ($opt{op_type} eq 'modify') {
		# XXX TBD
		logTrc("TBD");
	}

	# 7. special files
	eval {
		logTrc($opt{op_type} . " NFS special files in batches, with suffix");
		my $spec_dir = $nfs_path."/special_files";

		if ($opt{op_type} eq 'create') {
			croak("Failed to create dir $spec_dir") 
				if !mkdir($spec_dir);
		} elsif ($opt{op_type} eq 'read') {
			croak("directory $spec_dir does not exist!")
				unless -d $spec_dir;
		} elsif ($opt{op_type} eq 'modify') {
			croak("directory $spec_dir does not exist!")
				unless -d $spec_dir;
			# XXX TBD
			logTrc("TBD");
		}

		_do_spec_files ($spec_dir, 1, $opt{op_type});
		_do_spec_files ($spec_dir, 2, $opt{op_type});
	} or croak("Failed to create special files, @_");
		
	# 8. cifs data creation
	_do_cifs_data ($vol, $qtree_name, $opt{cifs_client}, $opt{op_type}) 
		or croak("Failed to create cifs data, @_");

	# XXX create some space resvd sparse files from NFS
	# XXX create some sparse files from CIFS

	if ( $need_umount ) { # unmount temporary mount
		logTrc("Unmount temporary mount");
		$vol->umount()
			or croak("umount of $vol->{mountPoint} failed, @_");
	}
}

sub _do_nfs_dirs ($$%) 
{
	my ($client, $dir_path, %opt) = @_;

	logTrc("Create directory width of $opt{min_dir_width}, "
		. "each containing one data file");
	populate_tree ($client, $dir_path."/dir_width", PRECREATE=>1
		,	DIR_AVG=>$opt{min_dir_width}, DIR_DEV=>0
		,	DEPTH_AVG=>1, DEPTH_DEV=>0
		,	FILE_AVG=>1, FILE_DEV=>0
		,	BREADTH_AVG=>128, BREADTH_DEV=>0
	);

	logTrc("Create " . $opt{max_dir_depth} . " directory depth");
	populate_tree ($client, $dir_path."/dir_depth", PRECREATE=>1
		,	DIR_AVG=>1, DIR_DEV=>0
		,	DEPTH_AVG=>$opt{max_dir_depth} - 1, DEPTH_DEV=>0
		,	FILE_AVG=>$opt{max_files_per_depth}, FILE_DEV=>0
		,	BREADTH_AVG=>64, BREADTH_DEV=>0
	);

	return 1;
}

sub _do_level_files
{
	my ($vol, $dir_path, %opt) = @_;

	croak "what op do you want? create, read or modify ?"
		unless ($opt{op_type} =~ /create|read|modify/);

	return 1 if $opt{op_type} eq "modify" && $opt{data_subtype} ne "level_files";

	logTrc($opt{op_type}." level files at $dir_path for $opt{max_level_file} "
		. "level files, with $opt{file_level_fill_perc}\% fill");

	# array of max file size at each file level
	my @file_level_max = (	
				64
			,	32768
			,	16711680
			,	8522956800
			,	4346707968000
			,	17557623775232
	);
	my @file_level_size;
	$file_level_size[0] = 28;

	# create array of file size to create for each file level,
	# using the file_level_fill_perc value
	my $level = 1;

	# Calculate file size to create for each file level
	do {
		if ( $opt{file_level_fill_perc} ) {
			$file_level_size[$level] = $file_level_max[$level-1]
			+	int ((($file_level_max[$level]-
					$file_level_max[$level-1])
						* $opt{file_level_fill_perc}
						)/100);
		} else {
			$file_level_size[$level] = $file_level_max[$level-1] 
						   + 1;
		}
		$level++;
	} while $level <= $#file_level_max;

	$level = 0;
	# Create/Read/Modify files of different fill, at each file level
	do {
		logTrc($opt{op_type} . " level $level files at "
			. "$opt{file_level_fill_perc}\% fill");
		my $wrt_buff_sz = 5001; # odd number to span block boundary
		for my $perc_interval ( 0, 1, 33, 66, 100 ) {
			# skip all intervals, other then 0 and 100, 
			# when $wrt_buff_sz is greater then the file size, 
			# as they would be the same
			next if $wrt_buff_sz > $file_level_size[$level]
				&& !($perc_interval==0 || $perc_interval==100);

			_do_lvl_file(
					$vol
				,	$dir_path."/file_level_".$level.
					"_interval_".$perc_interval."_perc"	
				,   $level
				,	$wrt_buff_sz
				,	$file_level_size[$level]
				,	$perc_interval
				,	$opt{op_type}
				);
		}
		$level++;
	} while $level <= $#file_level_size && $level <= $opt{max_level_file};

	return 1;
}

sub _do_lvl_file ($$$$$$$)
{
	my ($vol, $fname, $lvl, $wrt_buff_sz, $fsize, $perc_interval, $op_type) 
			= @_;

	croak("wrong size perc_interval ($perc_interval)") 
		if $perc_interval > 100;
	
	# set interval size as a function of the file size and percentage
	# interval specified, only exception - 0%
	my $interval_size = ($perc_interval == 0 ? 
				$wrt_buff_sz : int ($fsize*$perc_interval/100));
	$interval_size = ($wrt_buff_sz < $fsize ? $wrt_buff_sz : $fsize)
			unless $interval_size;

	# for small files, the interval size is smaller than the write buffer
	$wrt_buff_sz = $interval_size 
			if ($interval_size < $wrt_buff_sz);

	# write buffer becomes 1% of the interval_size if the file level
	# is greater than 3, i.e, level 4 and 5 files
	$wrt_buff_sz = int ($interval_size/100) 
			if ($lvl > 3);

	my $offset = ($perc_interval == 0 ? 0 : $interval_size);
	logTrc("$op_type file $fname, size $fsize, at $interval_size byte gaps, "
		. "starting at offset $offset");

	my $mode = ($op_type eq "create") ? ">" : "+<";
	my $f = fioOpen($vol, $fname, $mode);
	my $msg;
	# write/read at intervals until $fsize is reached
	do {
		$wrt_buff_sz = $fsize - $offset 
			if ($offset + $wrt_buff_sz > $fsize);

		$msg = "file $fname, $op_type $wrt_buff_sz bytes at ".
		       "offset $offset";

		$f->Seek($offset);
		$f->WriteSeq($wrt_buff_sz)
			if ($op_type eq 'create');
		$f->ReadSeq($wrt_buff_sz)
			if ( $op_type eq 'read');
		
		$offset += $interval_size;
	} while $offset < $fsize;

	logTrc("Last $op_type : $msg");

	if ($op_type eq 'create') {
		$f->Seek($fsize-1); 
		$f->WriteSeq(1); # To set the file size
	} elsif ($op_type eq 'read') {
		$f->Stat();
		croak "$op_type on " . $f->{name} . ", file size verify failed! " .
		      $f->{Stat}{size} ." != $fsize "
				if ($f->{Stat}{size} != $fsize);
	} elsif ($op_type eq 'modify') {
		# XXX revisit for other types of modify
		if ( rand 100 < 50 ) {	
			$f->Stat();
			my $old_fsize = $f->{Stat}{size};
	
			my $new_fsize = int ($fsize / 2);
			$f->Truncate($new_fsize);
			$f->Stat();
			croak("file not truncated " . $f->{name} . ", "
				. "Size: old $old_fsize, "
				. "expected $new_fsize, "
				. "actual " . $f->{Stat}{size})
				unless $f->{Stat}{size} == $new_fsize;
			logTrc("file $f->{name} truncated in half");
		} else {
			$f->Unlink (); # remove the file
			if ( -e $fname ) {
				croak("Failed to remove file $fname");
			} else {
				logTrc("File $fname removed");
			}
		}
	}

	return 1;
}

sub _do_spec_files ($$$)
{
	my ($spec_dir, $idx, $op_type) = @_;

	logTrc("$op_type special files");
	if ($op_type eq 'create') {
		logDbg("Create socket $idx");
		create_socket($spec_dir."/socket$idx");

		logDbg("Create device file $idx");
		create_dev ($spec_dir."/dev_file$idx", $idx, $idx);

		logDbg("Create named pipe $idx");
		create_fifo($spec_dir."/named_pipe$idx");
	}

	return 1;
}

sub _do_cifs_data($$$$)
{
	my ($vol, $qtreepath, $client, $op_type) = @_;

	# XXX yeah, hardcoded .. fixme!
	my $cifs_user = "administrator"; # CIFS_USER
	my $cifs_pass = "sundance"; # CIFS_PASSWORD
	# DIRECTORY base name of dir
	# DEPTH levels of subdirs
	# FILENAME base name of file
	# MAXFILES max no. of files in each dir
	# MAXSTREAMS max no. of streams in each file and dir
	# MAXSIZE max size of each file
	# MAXACES max no. of aces per file
	my $random = "NO"; # RANDOM should be NO

	logTrc("$op_type cifs data at $qtreepath via $client");

	if ($op_type eq 'create') {
		logDbg("Change qtree security of $qtreepath to mixed");
		qtreeSecurity($vol->{filer}->{conn}, $qtreepath, "mixed")
			or croak($@);
	}
	
	logDbg("$op_type cifs share to mount (delete share if it already exists)");
	my $share = "test_data";

	if ($op_type eq 'create') {
		share_delete($vol->{filer}->{conn}, $share)
			or croak($@);
		share_add($vol->{filer}->{conn}, $share, $qtreepath, verify=>1)
			or croak($@);
	} elsif ($op_type eq 'read') {
		share_exists($vol->{filer}->{conn}, $share)
			or croak($@);
	}
	
	if ($op_type eq 'create') {
		logDbg("subtest cifs_data_set.thpl");

		# setting NATE_LIB=http://web.netapp.com/engineering/rlse/mainN with
		# ntest invocation should work... 
		my $onlychild = subtest(
				"http://web.netapp.com/engineering/rlse/mainN/test/NDMP/CIFSDATA/cifs_data_set.thpl",
				-client, $client,
				"FILER=".$vol->{filer}->{hostName},
				"CIFS_SHARE=$share",
				"CIFS_USER=$cifs_user",
			   	"CIFS_PASSWORD=$cifs_pass",
				"RANDOM=$random"
				);

		$onlychild->wait_finish();
		croak("creation of cifs data failed, check subtest logs")
			if ($onlychild->exit_status() != 0);
=pod
	# This is the way its supposed to work, NTEST FIX COMING UP..
	my $onlychild = subtest(
			"NDMP/CIFSDATA/cifs_data_set.thpl",
			-client, $client,
			"FILER=".$vol->{filer}->{hostName},
			"CIFS_SHARE=$share",
			"NATE_LIB=http://web.netapp.com/engineering/rlse/mainN/test;http://web.netapp.com/engineering/rlse/mainN/test/lib",
			"ONTAPDIR=http://web.netapp.com/engineering/rlse/mainN",
			"NATE_TESTDIR=http://web.netapp.com/engineering/rlse/mainN/test;http://www.netapp.com/engineering/rlse/mainN/test/lib",
			"CIFS_USER=$cifs_user",
		   	"CIFS_PASSWORD=$cifs_pass",
			"RANDOM=$random"
			);
=cut
	} elsif ($op_type eq 'read') {
			logTrc("TBD");
	} elsif ($op_type eq 'modify') {
			logTrc("TBD");
	}
	
	return 1;
}

1;
