#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/ontap/main/test/lib/SPARSE/Util/FSTest0.pm# $
#
# Copyright (c) 2002 Network Appliance, Inc.
# All rights reserved.

## @summary	Library that provides file testing APIs
## @description
## 	Functions to build, modify, check and apply operations on
##	FileSystem Objects
##
## @author dl-sparseeng

package SPARSE::Util::FSTest0;
use File::Path;

use strict;
use Tharn;
use fileSystem qw(create_win_data);
use SPARSE::Resource::Volume;
use SPARSE::Resource::Volume10M; # mica - needed for 10M support
use SPARSE::Util::FileIO;
use SPARSE::Util::Logger;
use SPARSE::Util::Utilities;
no warnings("uninitialized");

# ------------------------------------------------------------------------

use vars ('@ISA', '@EXPORT');
require Exporter;
@ISA	= qw(Exporter);
@EXPORT	= qw(
		&FSBuildFileSet
		&FSCheckFileSet
	        &FSModifyFileSet
		&FSFillFileSet
		&FSRunFileSet
		&FSFiles
		&FSCheckSnapFileSetData
		);

# ------------------------------------------------------------------------

## @name	::FSBuildFileSet
## @description
##	Create a set of interesting filesystem objects
## @arg		vol required
## @arg		options optional
sub FSBuildFileSet($;$)
{
	my ($vol, $opts) = @_;

	logTrc("FS build file set: dir=$vol->{testDir}\n");

	if (!-e $vol->{testDir}) {
		$vol->make_testdir();
	}

	my $doFile = sub ($$) {
		my ($vol, $file) = @_;
		return (fioOpen($vol, $file, ">"));
	};

	my $doData = sub ($$$$;$) {
		my ($vol, $f, $off, $size, $seed) = @_;

		$f->Seek($off);
		$f->WriteSeq($size, $seed);
	};

	my $doSize = sub ($$$) {
		my ($vol, $f, $size) = @_;
		$f->Truncate($size);
	};

	my $doMkDir = sub ($$) {
		my ($vol, $name) = @_;
		fioMkDir($vol, $name);
	};

	my $doLink = sub ($$$) {
		my ($vol, $f, $name) = @_;
		$f->Link($name);
	};

	my $doSymLink = sub ($$$) {
		my ($vol, $f, $name) = @_;
		$f->SymLink($name);
	};

	if (defined($opts->{nfsAcls})) {
		logDbg("create nfsv4 acl file");
		$opts->{doAcl} = sub ($$$$$) {
			my ($vol, $file, $entrytype, $uid, $perms) = @_;
			fioOpen($vol, $file, ">");
			uCommand("setfacl -m $entrytype:$uid:$perms "
				 . "$vol->{testDir}/$file",
				 undef, "Invalid|error|not supported", 0);
		}
	}
	
	if (defined($opts->{ntAclsStreams})) {
		logDbg("create nt acl and stream files");
		$opts->{doAcl} = sub ($) {
			my ($vol) = @_;

			create_win_data($vol->{shareTestPath},
					client=>$vol->{shareClient}) 
			|| logError("fileSystem::create_win_data failed! $@");
		}
	}

	FSRunFileSet($vol, $doFile, $doSize, $doData, $doLink,
		$doSymLink, $doMkDir, $opts);

	if (defined($opts->{ntAclsStreams})) {
		$vol->unmap_share();
	}
}

## @name	::FSModifyFileSet
## @description
##	Create a set of interesting filesystem objects
## @arg		vol required
sub FSModifyFileSet($)
{
	my ($vol) = @_;

	logTrc("FS modify file set: dir=$vol->{testDir}\n");

	my $doFile = sub ($$) {
		my ($vol, $file) = @_;
		return (fioOpen($vol, $file, ">"));
	};

	# Modify files
	my $doData = sub ($$$$;$) {
		my ($vol, $f, $off, $size, $seed) = @_;
		$seed = 8;
		$f->Seek($off);
		$f->WriteSeq($size, $seed);
	};
	
	# Grow files
	my $doSize = sub ($$$) {
		my ($vol, $f, $size) = @_;
		$f->Truncate($size * 2);
	};

        # Remove dir/files and add new dir
	my $doMkDir = sub ($$) {
		my ($vol, $name) = @_;
		rmtree("$vol->{testDir}/$name");
		fioMkDir($vol, "$name");
	};

	# Add new links
	my $doLink = sub ($$$) {
		my ($vol, $f, $name) = @_;
		$f->Link($name);
	};

	# Change sym link
	my $doSymLink = sub ($$$) {
		my ($vol, $f, $name) = @_;
		$f->Chmod(0222);
		rename("$vol->{testDir}/$name","$vol->{testDir}/$name"."new");
	};

	FSRunFileSet($vol, $doFile, $doSize, $doData, $doLink,
		$doSymLink, $doMkDir);
}

sub _filestr($$)
{
	my ($vol, $f) = @_;
	return "vol $vol->{filer}->{hostName}:$vol->{name} ino $f->{Stat}{ino}";
}

## @name	::FSCheckFileSet
## @description
##	Check a set of interesting filesystem objects
## @arg		vol required
## @arg		options optional
sub FSCheckFileSet($;$)
{
	my ($vol, $opts) = @_;

	logTrc("FS check file set: dir=$vol->{testDir}\n");

	my $doFile = undef;

	my $doData = sub ($$$$;$) {
		my ($vol, $f, $off, $size, $seed) = @_;
		$f->Seek($off);
		$f->ReadSeq($size, $seed);
	};

	my $doSize = sub ($$$) {
		my ($vol, $f, $size) = @_;
		$f->Stat();
		logError("wrong size "._filestr($vol,$f)
			." (got $f->{Stat}{size}, should be $size)")
			unless $f->{Stat}{size} == $size;
	};

	my $doMkDir = sub ($$) {
		my ($vol, $name) = @_;
		logError("missing dir $vol->{testDir}/$name")
			unless -d "$vol->{testDir}/$name";
	};

	my $doLink = sub ($$$) {
		my ($vol, $f, $name) = @_;
		$f->Stat();
		logError("low link count "._filestr($vol,$f)
			." (nlink is $f->{Stat}{nlink})")
			unless $f->{Stat}{nlink} > 1;
	};

	my $doSymLink = sub ($$$) {
		my ($vol, $f, $name) = @_;
		logError("missing symlink $vol->{testDir}/$name")
			unless -l "$vol->{testDir}/$name";
	};

	if (defined($opts->{nfsAcls})) {
		logDbg("check nfsv4 acl file");
		$opts->{doAcl} = sub ($$) {
			my ($vol, $file) = @_;
			return uCommand("getfacl $vol->{testDir}/$file",
					undef, "No such file or directory");
		}
	}

	if (defined($opts->{ntAclsStreams})) {
		# XXX fileSystem::check_win_data doesn't exist 
		# see burt115595 
		logDbg("TODO: check NT ACL and STREAM files");
		#$opts->{doAcl} = sub ($$$) {
		#	my ($vol, $testdir, $client) = @_;
		#	fileSystem::check_win_data("$vol->{mountPoint}:".
		#		"\\$testdir", $client) 
		#	|| logError("fileSystem::create_win_data failed! $@");
		#}
	}

	FSRunFileSet($vol, $doFile, $doSize, $doData, $doLink,
		$doSymLink, $doMkDir, $opts);
}

## @name	::FSFillFileSet
## @description
##	Read-fills (completely) a set of filesystem objects. Ignores holes.
## @arg		vol required
sub FSFillFileSet($)
{
	my ($vol) = @_;

	logTrc("FS fill file set: dir=$vol->{testDir}\n");

	my $doData = sub ($$$$;$) {
		my ($vol, $f, $off, $size, $seed) = @_;
		$f->Stat();
		$f->Seek(0);
		$f->Read($f->{Stat}{size}, sub {});
	};

	FSRunFileSet($vol, undef, undef, $doData, undef, undef, undef);
}

## @name	::FSRunFileSet
## @description
## 	Apply ops to a standard set of interesting filesystem objects
## @arg	vol required
## @arg	doFile required	regular file handler (should return a FileIO object)
## @arg	doSize required	size handler (called to grow the file - passed size)
## @arg	doData required	data handler (called for each write - passed len@offset)
## @arg	doLink required	hard link handler (called for each hard link)
## @arg	doSymLink required symlink handler (called for each symlink)
## @arg	doMkDir required dir handler (called for each directory)
## @arg	options
sub FSRunFileSet($$$$$$$;$)
{
	my ($vol, $doFile, $doSize, $doData, $doLink, $doSymLink,
		$doMkDir, $opts) = @_;

	# defaults - can pass undef for any handler
	if (!defined($doFile)) {
		$doFile = sub ($$) {
			my ($vol, $file) = @_;
			return (fioOpen($vol, $file, "+<"));
		};
	}

	if (!defined($doSize)) { $doSize = sub {}; }
	if (!defined($doData)) { $doData = sub {}; }
	if (!defined($doLink)) { $doLink = sub {}; }
	if (!defined($doSymLink)) { $doSymLink = sub {}; }
	if (!defined($doMkDir)) { $doMkDir = sub {}; }

	sub reg_no_holes($$$$$$) {
		my ($vol, $file, $size, $doFile, $doSize, $doData) = @_;
		my ($f);
		$f = $doFile->($vol, $file);
		$doSize->($vol, $f, $size);
		$doData->($vol, $f, 0, $size);
	}

	logTrc("FS run file set: dir=$vol->{testDir}\n");

	logDbg("level 0 file, full");
	reg_no_holes($vol, "lvl0_full", 64, $doFile, $doSize, $doData);

	logDbg("level 0 file, partial");
	reg_no_holes($vol, "lvl0_partial", 32, $doFile, $doSize, $doData);

	logDbg("level 1 file, no holes");
	reg_no_holes($vol, "lvl1_noholes", 32*1024, $doFile, $doSize, $doData);

	logDbg("level 1 file, no holes, partial last block");
	reg_no_holes($vol, "lvl1_noholes_partiallastblock", (32*1024) - 512,
		$doFile, $doSize, $doData);

	{
	logDbg("level 1 file, holy middle");
	my $f = $doFile->($vol, "lvl1_holymiddle");
	$doSize->($vol, $f, 32*1024);
	$doData->($vol, $f, 0, 4*1024);
	$doData->($vol, $f, 16*1024, 16*1024);
	}

	{
	logDbg("level 1 file, holy ends");
	my $f = $doFile->($vol, "lvl1_holyends");
	$doSize->($vol, $f, 32*1024);
	$doData->($vol, $f, 0, 16*1024);
	}

	{
	logDbg("level 2 file, holy begin");
	my $f = $doFile->($vol, "lvl2_holybegin");
	$doSize->($vol, $f, 64*1024);
	$doData->($vol, $f, 16*1024, 16*1024);
	}

	{
	logDbg("level 3 file, holy");
	my $f = $doFile->($vol, "lvl3_holy");
	$doSize->($vol, $f, 32*1024*1024);
	$doData->($vol, $f, 3*1024*1024, 1*4096);
	$doData->($vol, $f, 11*1024*1024, 256*4096);
	$doData->($vol, $f, 25*1024*1024, 1*4096);
	}

	logDbg("level 1 dir, 100 entries");
	my $n = "lvl1_dir_100entries";
	$doMkDir->($vol, $n);
	my $f = fioOpen($vol, "lvl1_noholes");
	for my $ii (1 .. 100) {
		$doLink->($vol, $f, "$n/$ii");
	}

	logDbg("symlink");
	$doSymLink->($vol, $f, "sym");
		
	if (defined($opts->{nfsAcls})) {
		$opts->{doAcl}->($vol, "acl_file", "u", "ram", "rwx");
	}

	if (defined($opts->{ntAclsStreams})) {
		$vol->create_mapped_share();
		$opts->{doAcl}->($vol);
	}
}

## @name	::FSFiles
## @description
## 	Returns a list of fio objects to the regular files in the fileset
## @arg
##	vol required
sub FSFiles($)
{
	my ($vol) = @_;
	our @files = ();

	my $doFile = sub ($$) {
		my ($vol, $f) = @_;
		push(@files, fioOpen($vol, $f));
	};

	logDbg("looking for files");

	FSRunFileSet($vol, $doFile, undef, undef, undef, undef, undef);

	logDbg("returning " . scalar(@files) . " fio objects");
	return @files;
}

# ------------------------------------------------------------------------

1;
