#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/ontap/main/test/lib/SPARSE/Util/FileIO.pm# $
#
# Copyright (c) 2002 Network Appliance, Inc.
# All rights reserved.

## @summary	Library that defines a FileIO object
## @description
##	Abstract basic file IO operations.
##
## @author dl-sparseeng

package SPARSE::Util::FileIO;

use strict;
use Tharn;
use SPARSE::Constants;
use SPARSE::Util::Logger;
use SPARSE::Util::Utilities;
use FS::Util::Utilities qw(deltree);
use POSIX;
use Fcntl;
use Params::Validate
	qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF);
use Module::Common qw(validate_mixed);

no warnings("uninitialized");

# ------------------------------------------------------------------------

use vars ('@ISA', '@EXPORT');
require Exporter;
@ISA	= qw(Exporter);
@EXPORT	= qw(
		&fioOpen
		&fioMkDir
		&fioRmDir
	        &VfyDirAbsent
	        &VfyDirPresent
	        &fioMkFifo

		&fioRCmpTestDirSnaps
		&fioRCmpTestDirs
		fioRCmpDir
		&fioDiffFileContents

	        &StatCheck
	        &VfyFileAbsent
	        &VfyFilePresent

	        &Chown
	        &Chgrp
		&FileUnlink
		&fioLock

		&r_chmod
		&r_rmrf
		);

# ------------------------------------------------------------------------

## @name	::fioOpen
## @description
## 	Construct file handle based on vol and test dir
## @arg		vol required
## @arg		file-name required
## @arg		mode optional "+>" trunc+write+read
## @arg         mask optional which umask to use when opening file
## @arg         perms optional. file perms.
## @arg         err optional. expected error
## @return 	file object
sub fioOpen($$;$$$$)
{
	my ($vol, $name, $mode, $mask, $perms, $err)	= @_;
	my $fio			= {};
	$err = 0      unless defined($err);
	$mode = "+<"  unless defined($mode);
	$perms = 0666 unless defined($perms);

	$name =~ s@^/*@@;
	$fio->{name}    = $name;			# name of file
	$fio->{path}    = "$vol->{testDir}/$name";	# full path

	$fio->{vol_path} = $fio->{path};
	$fio->{vol_path} =~ s/\Q$vol->{mountPoint}\E//;

	$fio->{vol}	= $vol;			# vol object reference
	bless($fio);

	logTrc("open: $fio->{path} mode=$mode mask=$mask "
		. "perms=$perms err=$err");

	# if user wants specific umask to execute for open, do that
	my $umask = umask($mask) if (defined($mask));

	# Perl file handle
	$fio->{fh} = new IO::File($fio->{path}, $mode, $perms);

	if (!defined($fio->{fh}) || $err) {
		_process_error($!, $err, "open file $fio->{path}:");
		return undef;
	}
	$fio->{offset} = $fio->{fh}->tell();	# offset (take care of appends)

	# if we tweaked with mask, set it back
	umask($umask) if (defined($mask));

	$fio->Stat();

	return ($fio);
}


# ------------------------------------------------------------------------

sub DESTROY($)
{
	my ($fio) = @_;

	my $exc_state = $@;

	logTrc("destroy handle $fio->{name}");
	$fio->Close();

	$@ = $exc_state;
}

# ------------------------------------------------------------------------

## @name	::Error
## @summary	Report error on a file
## @arg		op-name required
## @arg		msg required
## @arg         err  -optional
sub Error($$$)
{
	my ($fio, $op, $err)	= @_;

	_process_error($err, 0,
		       "$op failed, name=$fio->{path}"
		       . " off=$fio->{offset} size=$fio->{Stat}{size}\n");
}

## @name	::Open
## @summary	(Re)Open existing file handle
## @arg		mode optional "+>" trunc+write+read
sub Open($;$)
{
	my ($fio, $mode) = @_;
	$mode = "+<" unless defined($mode);

	$fio->{fh}->open($fio->{path}, $mode)	or $fio->Error("open", $!);
}

## @name	::Close
## @summary	Close file handle
sub Close($)
{
	my ($fio)	= @_;

	return unless $fio->{fh};
	logTrc("close: $fio->{path}");
	$fio->{fh}->close();
}

## @name	::Unlink
## @summary	Unlink file
## @arg         err optional error code
##
## N.B. this may not quite do what you expect - since the file is open
## over NFS this will likely do a rename (see FileUnlink)
##
sub Unlink($;$)
{
	my ($fio, $err)	= @_;
	$err = 0 unless $err;

	logTrc("unlink: $fio->{path}");
	my $rv = unlink($fio->{path});
	if ($err || ($rv == 0)) {
		logWarn("Error code from Unlink!!! err = $err , rv =$rv");
		_process_error($!, $err, "unlink $fio->{path}");
	}
}

## @name	::Rename
## @summary	Renames file
## @arg         new_name required
## @arg         err optional error code
sub Rename($$;$)
{
	my ($fio, $new_name, $err) = @_;

	$err = 0 unless $err;
	my $new_path = $fio->{path};
	my $vol = $fio->{vol};
	$new_path =~ s/\Q$fio->{name}\E/$new_name/;

	logTrc("rename $fio->{path} to $new_path");

	my $rv = rename("$fio->{path}", $new_path);
	if ($err || ($rv == 0)) {
		logWarn("Error code from Rename!!! err = $err , rv =$rv");
		_process_error($!, $err, "rename $fio->{path}");
		return;
	}

	$fio->{name} = $new_name;
	$fio->{path} = $new_path;

	#XXX TODO assume we are in TEST DIR or its subdirs for now
	$fio->{vol_path} = $fio->{path};
	$fio->{vol_path} =~ s/\Q$vol->{mountPoint}\E//;
}

## @name	::Stat
## @summary	Get stats for file
## @notes
##	Set values for $f->{Stat}{dev ino mode nlink uid gid ...}
sub Stat($)
{
	my ($fio)	= @_;

	my @stat = $fio->{fh}->stat()		or $fio->Error("stat", $!);

	my @yuk = qw( dev ino mode nlink uid gid rdev
			size atime mtime ctime blksize blocks );
	for (0..$#yuk) {
		$fio->{Stat}{$yuk[$_]} = $stat[$_];
	}

	logTrc("stat: name=$fio->{path}"
		. " ino=$fio->{Stat}{ino}"
		. " size=$fio->{Stat}{size}"
		. " nlink=$fio->{Stat}{nlink}");
}

## @name     ::StatCheck
## @summary  Check stat data structure for specific fields
## @arg      opts-stat elelments that need to be verified. 
##                required
sub StatCheck($$)
{
	my ($fio, $opts) = @_;

	# stat it, just in case
	$fio->Stat();

	foreach my $key (keys(%$opts)) {
		if (!defined($fio->{Stat}{$key})) {
			logError("unknown Stat key $key");
		}
		if ($fio->{Stat}{$key} ne $opts->{$key}) {
			logError("$fio->{path} $key expected $opts->{$key}"
				. " but got $fio->{Stat}{$key}");
		}
	}
}

## @name	::Info
## @summary	Create info about file
## @return	String with info about file
sub Info($)
{
	my ($fio)	= @_;
	my $vol		= $fio->{vol};

	$fio->Stat();
	my $info = "--- File Info"
		. " name=$fio->{path}"
		. " ino=$fio->{Stat}{ino}"
		. " size=$fio->{Stat}{size}"
		. " nlink=$fio->{Stat}{nlink}\n"
		. "VBNs:"
		. $vol->{filer}->filer_say(PRIVSETTEST .
			"wafl showvbns $vol->{testDirVolName}/$fio->{name}");
	return ($info);
}

## @name	::Seek
## @summary	Seek to offset in file using Perl "sysseek" (lseek) for unbuffered I/O
## @arg		offset required
## @arg		whence optional SEEK_SET
sub Seek($$;$)
{
	my ($fio, $off, $whence)	= @_;

	$whence = SEEK_SET unless defined($whence);
	logTrc("seek: name=$fio->{path} off=$off whence=$whence");
	logDbg("fd=" .fileno($fio->{fh}));

	$fio->{offset} = $fio->{fh}->sysseek($off, $whence);
	$fio->Error("seek", $!) if !defined($fio->{offset});
}

my $BUFSIZ = 4096;

# Verify content of block with default sequence
sub vfyDefault($$$)
{
	my ($off, $buf, $len) = @_;

	my $rseq = substr($buf, 0, 1);	# first byte
	my $xseq = &uSeqStr($off, 1);	# seq based on offset
	($rseq eq $xseq)
		or logError("seq mismatch \"$rseq\" != \"$xseq\"");
}

## @name	::Read
## @summary	Read and verify general file data
## @arg		length required
## @arg		verify optional sub callback(off, buf, len)
## @example
##	$f->Read(1024, \&mycallback);
sub Read($;$)
{
	my ($fio, $len, $verify)	= @_;

	logTrc("read: name=$fio->{path} len=$len off=$fio->{offset}");
	logDbg("fd=" .fileno($fio->{fh}));

	$verify = \&vfyDefault unless defined($verify);

	while ($len) {
		my $buf;
		$fio->{fh}->sysread($buf, ($len > $BUFSIZ) ? $BUFSIZ : $len);
		my $rlen = length($buf);

		$fio->Error("read", ($! ? $! : "EOF")) if ($rlen == 0);

		&$verify($fio->{offset}, $buf, $rlen);

		$len -= $rlen;
		$fio->{offset} += $rlen;
	}
}

## @name	::ReadSeq
## @summary	Read and verify sequence file data
## @arg		length required
## @arg		sequence optional offset into sequence
sub ReadSeq($$;$)
{
	my ($fio, $len, $seq)	= @_;

	logTrc("read seq: name=$fio->{path} len=$len off=$fio->{offset}");

	$seq = 0 unless (defined($seq));
	uReadSeq($fio->{fh}, $seq, $len);
	$fio->{offset} += $len;
}

## @name	::Write
## @summary	Write general file data
## @arg		data required
## @arg		length required
sub Write($$$)
{
	my ($fio, $buf, $len)	= @_;

	logTrc("write: name=$fio->{path} len=$len off=$fio->{offset}");
	logDbg("fd=" .fileno($fio->{fh}));
	logDbg("\"" . substr($buf, 0, $len) . "\"");

	$fio->{fh}->syswrite($buf, $len)	or $fio->Error("write", $!);
	$fio->{offset} += $len;
}

## @name	::WriteSeq
## @summary	Write sequence file data
## @arg		length required
## @arg		sequence optional offset into sequence
sub WriteSeq($$;$)
{
	my ($fio, $len, $seq)	= @_;

	logTrc("write seq: name=$fio->{path} len=$len off=$fio->{offset}");

	$seq = 0 unless (defined($seq));
	uSendSeq($seq, $len, $fio->{fh});

	$fio->{offset} += $len;
}

## @name	::WriteBlk
## @summary	Write a 4KB block of sequence string data
## @arg		seed optional, offset into sequence
## @arg		seed2 optional, repeat count of the byte in the sequence 
##		string at the offset privided as seed. (see uSendBlk for detail)
sub WriteBlk($;$$)
{
        my ($fio, $seed, $seed2)   = @_;

        logTrc("write block:seed=$seed name=$fio->{path}"
		. " off=$fio->{offset} seed2=$seed2");

        $seed = 0 unless (defined($seed));
        $seed2 = 0 unless (defined($seed2));
        my $wlen = uSendBlk($seed, $fio->{fh}, $seed2);

        $fio->{offset} += $wlen;
}

## @name	::Truncate
## @summary	Truncate file to size
## @arg		size optional
sub Truncate($;$)
{
	my ($fio, $len)	= @_;

	$len = 0 unless (defined($len));
	logTrc("truncate name=$fio->{path} len=$len");

	$fio->{fh}->truncate($len)	or $fio->Error("truncate", $!);
}

## @name	::Chown
## @summary	Change file's owner
## @arg		owner required
sub Chown($$)
{
	my ($fio, $owner) = @_;
	uSystem("r chown $owner $fio->{path}");
}


## @name	::Chgrp
## @summary	Change file's group
## @arg		gid required
sub Chgrp($$)
{
	my ($fio, $gid) = @_;
	uSystem("r chgrp $gid $fio->{path}");
}

## @name	::Link
## @summary	Create link to a file
## @arg		dest-name required
sub Link($$;$)
{
	my ($fio, $name, $err)	= @_;

	logTrc("link name=$fio->{path} link=$fio->{vol}{testDir}/$name");
	my $rv = link($fio->{path}, "$fio->{vol}{testDir}/$name");
	if ($err || ($rv == 0) )  {
		logWarn("Actual Error code =$! Error code from Link!!! err = $err , rv =$rv");
		_process_error($!, $err,
			"link $fio->{vol}{testDir}/$name to $fio->{path}");
	}
}

## @name	::SymLink
## @summary	Create symbolic link to a file
## @arg		dest-name required
## @arg         err-optional
sub SymLink($$;$)
{
	my ($fio, $name, $err)	= @_;
	$err = 0 unless $err;

	logTrc("symlink name=$fio->{path} link=$fio->{vol}{testDir}/$name");
	my $rv = symlink($fio->{path}, "$fio->{vol}{testDir}/$name");
	if ($err || ($rv == 0) )  {
		logWarn("Actual Error code =$! Error code from Link!!! err = $err , rv =$rv");
		_process_error($!, $err, "symlink $fio->{vol}{testDir}/$name "
			       ." to $fio->{path}");
	}
}

## @name        ::ReadLink
## @summary     Read symbolic link information
## @arg         err-optional
sub ReadLink($;$)
{
	my ($fio, $err)  = @_;

	logTrc("symlink name=$fio->{path}");
	my $rv = readlink($fio->{path});
	if ($err || !defined($rv)) {
		logWarn("Error code from ReadLink!!! err = $err , rv =$rv");
		_process_error($!, $err, "readlink $fio->{path}");
	}
	return $rv;
}


## @name	::Chmod
## @summary	Change mode of file
## @arg		mode required
sub Chmod($$;$)
{
	my ($fio, $mode, $err)	= @_;
	$err = 0 unless $err; # OS error code, not perl

	logTrc("chmod name=$fio->{path} mode=$mode");
	my $rv = chmod($mode, $fio->{path});
	# if expected error or failed, process err
	if ($err || $rv == 0) {
		logWarn("Error code from Chmod!!! err = $err , rv =$rv");
		_process_error($!, $err, "chmod $fio->{path}");
	}
}

## @name	::_process_error
## @summary     internal error processing function. useful when
##              not operating on fio Objects
## @arg		got required
## @arg         expected required
## @arg         op optional.
sub _process_error($$;$)
{
	my ($got, $expected, $op) = @_;
	$op = "" unless $op;
	$expected = 0 unless $expected;

	if ($got && ($got != $expected)) {
		# got unexpected error
		logError("$op got ".sprintf("%d", $got)
			 ." ($got) instead of expected $expected ");
	}

	if ($expected && !$got) {
		# expected error, but did not get it
		logError("$op succeeded when expected error $expected but got ! of $got");
	}
}

# ------------------------------------------------------------------------

## @name	::SameAs
## @summary	Verify similarity of files
## @arg		other-file required
## @arg		optional parts to ignore
sub SameAs($$;$)
{
	my ($f1, $f2, $ignore)	= @_;
	$f1->Stat();
	$f2->Stat();

	for (keys(%{$f1->{Stat}})) {
		next if $_ eq "dev";	# may not match
		next if $_ =~ $ignore;
		$f1->{Stat}{$_} == $f2->{Stat}{$_}
			or logError("files $f1->{path} $f2->{path} differ: $_"
				. " $f1->{Stat}{$_} != $f2->{Stat}{$_}");
	}
}

## @name     ::VfyFileAbsent
## @summary  Verify that the file does not exist
## @arg      path-path to a file. required
## @arg      opts-optional
sub VfyFileAbsent($;$)
{
	my ($path, $opts) = @_;

	if (! -e $path) {
		return;
	}
	my $extra_info = $opts->{err_msg} || "";
	logError("file $path exists. $extra_info");
}

## @name     ::VfyFilePresent
## @summary  Verify that the file exists
## @arg      path-path to a file. required
## @arg      opts-optional
sub VfyFilePresent($;$)
{
	my ($path, $opts) = @_;

	if (-e $path) {
		return;
	}
	my $extra_info = $opts->{err_msg} || "";
	logError("file $path does not exist. $extra_info");
}


## @name	::fioDiffFileContents
## @summary	compare two files
## @description
##	Read two files into memory and
##	compare them line by line.
## @arg		file1 required
## @arg		file2 required
sub fioDiffFileContents($$)
{
	my ($file1, $file2) = @_;

	my $line = 0;
	my ($f1_contents, $f2_contents);

	# it would be simpler to read files into memory, but they could
	# be quite large. also, lets not depend on local diff utility.
	# we skip checking file size, since even if the sizes are different
	# we want to see first found difference
	open(S_FH, $file1)
		or logError("could not open $file1");
	open(P_FH, $file2)
		or logError("could not open $file2");
	while ((!eof(P_PH)) || (!eof(S_FH))) {
		$line++;
		$f1_contents = <P_FH>;
		$f2_contents = <S_FH>;
		if ($f1_contents ne $f2_contents) {
			close(S_FH);
			close(P_FH);
			logError("fioDiffFiles:files are different at "
				 . " line $line:\n"
				 . "$file1:$line - $f1_contents \n"
				 . "$file2:$line - $f2_contents \n");
		}
	}
	if (eof(P_PH) != eof(S_FH)) {
		logError("fioDiffFiles:files are different at "
			 . " line $line:\n"
			 . "$file1:$line - $f1_contents \n"
			 . "$file2:$line - $f2_contents \n");
	}
	close(S_FH) or logError("could not close $file1: $!");
	close(P_FH) or logError("could not close $file2: $!");
}

## @name	::fioRCmpSnapTestDirs
## @summary 	Recursively compare test directories of two volumes
## @arg 	vol1 required
## @arg         vol2 required
sub fioRCmpTestDirSnaps($$)
{
	my ($vol1, $vol2) = @_;
	return fioRCmpDir("$vol1->{testDir}/.snapshot",
			  "$vol2->{testDir}/.snapshot");
}

## @name	::fioRCmpTestDirs
## @summary 	Recursively compare test directories of two volumes
## @arg 	vol1 required
## @arg         vol2 required
sub fioRCmpTestDirs($$;$)
{
	my ($vol1, $vol2) = @_;

	return fioRCmpDir($vol1->{testDir},
			  $vol2->{testDir});
}

## @name	::fioRCmpDir
## @summary	Recursively compare two directories
## @description execute 'ls -lAR' on two directories
##	writing results to outfiles. compare outfiles.
## @arg		dir_src required
## @arg		dir_dst required
## @arg		tmp_dir optional
sub fioRCmpDir($$;$;$)
{
	my ($dir_src, $dir_dst, $tmp_dir) = @_;

	if (!defined $tmp_dir) {
		$tmp_dir = defined $ENV{TMPDIR} ? $ENV{TMPDIR} : "/tmp";
	}

	my $out_src = "$tmp_dir/out_src.$$";
	my $out_dst = "$tmp_dir/out_dst.$$";

	# LINUX ls hack
	# clock skew can result in a different format being reported
	# in the date field
	sleep(1);
	system("cd $dir_src && ls -lAR > $out_src") == 0
		or logError("could not list $dir_src "
			    . "and redirect to $out_src");

	system("cd $dir_dst && ls -lAR > $out_dst") == 0
		or logError("could not list $dir_dst "
			    . "and redirect to $out_dst");

	fioDiffFileContents($out_src, $out_dst);

	unlink($out_src, $out_dst);
}

# ------------------------------------------------------------------------

## @name	::fioMkDir
## @summary 	Create a directory
## @arg 	dir required
## @arg 	mode optional eg 0777
## @arg         err optional error code. (only when mkdir fails)
# XXX iffy - new object? overload file?
sub fioMkDir($$;$$)
{
	my ($vol, $name, $mode, $err)	= @_;
	$mode = 0777	unless $mode;
	$err = 0        unless $err;
	my $full_name =  "$vol->{testDir}/$name";

	logTrc("mkdir name=$full_name mode=".sprintf("%o", $mode));
	# this is not as clean as it should be
	# (retval of mkdir() vs $!)
	my $rv = mkdir($full_name);
	if ($err || $rv == 0) {
		logWarn("Error code from fioMkDir!!! err = $err , rv =$rv");
		_process_error($!, $err, "mkdir $full_name");
		logWarn("err = $err , rv =$rv ...");
		return; # nothing to chmod, so bail
	}

	# mkdir() takes a $mode parameter that gets 'and'ed
	# with umask, so instead of mucking with umask, just chmod it
	chmod($mode, $full_name)
		or logError("could not change permissions of "
			    ."$full_name to ".sprintf("%o", $mode));
}

## @name	::fioRmDir
## @summary 	Remove a subdirectory of volume's test dir
## @arg 	vol required
## @arg 	name required
## @arg         err optional. expected error (only when rmdir fails)
sub fioRmDir($$;$)
{
	my ($vol, $name, $err) = @_;
	my $full_name = "$vol->{testDir}/$name";

	logTrc("rmdir $full_name");

	my $rv = rmdir("$full_name");
	if ($err || $rv == 0) {
		logWarn("Error code from fioRmDir!!! err = $err , rv =$rv");
		_process_error($!, $err, "rmdir $full_name");
	}
}

## @name	::VfyDirAbsent
## @summary 	Verify that a directory does not exist 
##              (assumes subdir of test dir)
## @arg 	vol required
## @arg 	name required
sub VfyDirAbsent($$)
{
	my ($vol, $name) = @_;
	if (-d "$vol->{testDir}/$name") {
		logError("directory $vol->{testDir}/$name is presnet");
	}
}

## @name	::VfyDirPresent
## @summary 	Verify that a directory exists (assumes subdir of test dir)
## @arg 	vol required
## @arg 	name required
sub VfyDirPresent($$)
{
	my ($vol, $name) = @_;
	if (! -d "$vol->{testDir}/$name") {
		logError("directory $vol->{testDir}/$name is absent");
	}		

}

## @name	::fioMkFifo
## @summary 	Create a pipe
## @arg 	name required
## @arg 	err optional
## @arg         options optional 
sub fioMkFifo($;$%)
{	
	# XXX REMOVE THIS UNREADABLE GIBBERISH
	# XXX TODO need to change Daisuke's validate_mixed() 
	# so that later we can substitte $optsions HASHREF instead
	# of the %otpions HASH to make things consistent
	my ($name, $err, %options) =
		validate_mixed(@_,
		[
			{ type=>SCALAR },
			{ type=>SCALAR, optional=>1, default=>0 }
		],
		{
			mask=> {type=>SCALAR, optional=>1, default=>undef},
			mode=> {type=>SCALAR, optional=>1, default=>0666}
		}
		);
	my $opts = \%options;
	my $mode = $opts->{mode};
	my $mask;
	my $old_umask;

	# tweak mask if need to
	if (defined ($mask = $opts->{mask}))  {
		$old_umask = umask();
		umask($mask) 
			or logError("could not change umask to $mask $!");
	}

	# create fifo, process error if need be
	my $rv = POSIX::mkfifo($name, $mode);
	if (!defined($rv) || $err) {
		logWarn("Error code from fioMkFifo!!!, Actual error code = $!, expected err = $err, rv =$rv");
		_process_error($!, $err, "fioMkFifo $name");
	}

	# if we tweaked with mask, set it back
	if (defined($mask)) {
		umask($old_umask) 
			or logError("could not change umask to $old_umask $!");
	}
}

## @name	::FileUnlink
## @summary	Unlink file
## @arg         err optional error code
sub FileUnlink($;$)
{
	my ($filename, $err)	= @_;
	$err = 0 unless $err;

	logTrc("unlink: $filename");
	my $rv = unlink($filename);
	if ($err || $rv == 0) {
		logWarn("Error code from FileUnlink!!! err = $err , rv =$rv");
		_process_error($!, $err, "unlink $filename");
	}
}

## @name	::fioLock
## @summary	create a file region lock
## @notes
## Mimics system call, fcntl(), via Perl's Fcntl::fcntl() to
## create a file region lock. Cannot use Perl's Fcntl::flock() as
## that will lock the entire file, rather than a region in the file.
##
## The flock struct must be observed when this method is used on different
## OS. If the information is not packed correctly in $lock, file may
## not be locked properly.
##
## Retry failure for 3 times before giving up.
##
## Lock usage: 
##  General: $fio->fioLock($type, $whence, $start, $len);
##           $type   - F_RDLCK  | F_WRLCK  | F_UNLCK
##           $whence - SEEK_SET | SEEK_CUR | SEEK_END
##           $start  - offset in bytes, relative to $whence
##           $len    - length in bytes
##
##  Write exclusive lock: 
##    $fio->fioLock(F_WRLCK, SEEK_SET, 0, 4096);
##  Read shared lock:
##    $fio->fioLock(F_RDLCK, SEEK_CUR, 0, 1024);
##  Unlock:
##    $fio->fioLock(F_UNLCK, SEEK_END, 0, 2096);
sub fioLock($$;$$$$)
{
	my ($fio, $type, $whence, $start, $len, $err) = @_;
	$whence = SEEK_SET	unless $whence;
	$start = 0		unless $start;
	$len = 0	        unless $len;	# entire file
	$err = 0		unless $err;

	# typical flock struct found on Solaris, Linux, and OSF1
	# do a "grep 'struct flock' /usr/include/*/*" to check
	#
	# struct  flock   {
	#    short   l_type;
	#    short   l_whence;
	#    off_t   l_start;
	#    off_t   l_len;          /* len = 0 means until end of file */
	#    pid_t   l_pid;
	# };
	#
	# only using first 4 fields (no pid)
	my $flock_struct = "ssll";
	my @flock = ($type, $whence, $start, $len);
	my $lock = pack($flock_struct, @flock);

	my $retry = 0;
	my $rv;

	logTrc("fcntl name=$fio->{path} type=$type whence=$whence "
		. "start=$start len=$len err=$err");

	while ($retry < 3) {
		# perl's fcntl returns an undef value if fails
		$rv = fcntl($fio->{fh}, F_SETLKW, $lock) or $rv = -1;
		last if ($rv != -1);
		logWarn("$fio->{path}: Encountered file region locking error. "
			. "The rest of the test may fail due to this error.");
		++$retry;
	}
	if ($err || $rv != 0) {
		logWarn("Error code from fioLock!!! err = $err , rv =$rv");
		_process_error($!, $err, "flock $fio->{path}");
	}
}

# ------------------------------------------------------------------------

# FIXME These need to be evicted
#some wrapper functions for better .thpl cleanup and consistent
#retval handling

## @name	::r_chmod
## @summary	chmod as root
## @arg		mode required
## @arg		filename required
## @return	result
sub r_chmod($$)
{
	my ($mode, $file) = @_;
	logTrc("r chmod name=$file mode=".sprintf("%o", $mode));
	return (! system("r chmod $mode $file"));	# XXX 'r' command
}


## @name	::r_rmrf
## @summary	WARNING: recursive rm as root
## @arg		dir required
sub r_rmrf($)
{
	my ($dir) = @_;
	logError("r_rmrf of suspect dir $dir") if ($dir =~ m{^/+[^/]*/*$});
	logTrc("r rm -rf dir=$dir");
 	return (! system("r rm -rf $dir"));     # XXX 'r' command
}

1;
