#Hey emacs, drop dead
# $Id: //depot/prod/test/main/lib/SPARSE/Util/FilerPerfWatch.pm#2 $

## @author dl-sparseeng
## @summary  A set of routines for taking filer performance measurements
#
# Typical usage:
#	my $filer = sparse_filer($FILER_DST);
#	my $wat = perf_watch($filer, "prof,statit");
#	$wat->watchStart();
#	sleep(10);
#	$wat->watchStop();
#	$wat->watchReport();       
#
# Output is a set of result files

# XXX
#	ponder what to do about "pause" op
#	nice to be able to configure (pass params) to ops
#	ponder an interface to CM aka "stats" cmd
#	ponder a single output file a la perfAnt so we can share tools
#	(and eventually usurp that steaming pile of Expect code...)

package SPARSE::Util::FilerPerfWatch;

use strict;

use SPARSE::Resource::Volume;
use SPARSE::Util::Logger;
use SPARSE::Constants;
use POSIX qw(strftime);
use File::Basename;
use Data::Dumper;

no warnings("uninitialized");

# ------------------------------------------------------------------------

use vars ('@ISA', '@EXPORT');
require Exporter;
@ISA	= qw(Exporter);
@EXPORT	= qw(
		&perf_watch
		);

# ------------------------------------------------------------------------

# callback table for "watchers"
# this encapsulates the set of watcher choices and the entry points
# to manipulate them
#
my %watchers = (
#	Setup		Start		Stop		Report
"prof" =>
[	\&profSetup,	\&profStart,	\&profStop,	\&profReport],
"statit" =>
[	\&statitSetup,\&statitStart,	\&statitStop,	\&statitReport],
"waflsusp" =>
[	\&wSuspSetup,	\&wSuspStart,	\&wSuspStop,	\&wSuspReport],
"waffinitystats" =>
[	\&wfStatSetup,	\&wfStatStart,	\&wfStatStop,	\&wfStatReport],
"pktt" =>
[	\&pkttSetup,	\&pkttStart,	\&pkttStop,	\&pkttReport],
);

sub perf_watch($$)
{
	my ($filer, $w) = @_;
	my $watch = {};

	logStep("setting up watch keepers \"$w\"");
	for (split(/,/, $w)) {
		defined($watchers{$_}) or logError("unknown watcher \"$_\"");
		push(@{$watch->{watchers}}, $_);
	}
	$watch->{filer} = $filer;
	$watch->{ts}    = strftime("%d%b%Y-%H%M", localtime());

	$_ = $filer->filer_say("version -v", "^NetApp");
	my ($build, $flags) = /NetApp Release ([^:]+):[^<]+<(\w+)>/;
	defined($build) or logError("bad filer version $_");
	$watch->{build}  = basename($build);
	$watch->{flags}  = $flags;
	$watch->{kernel} = "$build/prod/build/maytag.$flags";

	bless($watch);
	logInfo("watcher " . Dumper($watch));

	watchIt($watch, 0);
	return $watch;
}

sub watchIt($$)
{
	my ($watch, $op) = @_;
	for (@{$watch->{watchers}}) {
		logTrc(${["setup", "start", "stop", "report"]}[$op] . " $_");
		$watchers{$_}->[$op]($watch);
	}
}

sub watchStart($)  { watchIt($_[0], 1); }
sub watchStop($)   { watchIt($_[0], 2); }
sub watchReport($) { watchIt($_[0], 3); }

# ------------------------------------------------------------------------

sub profSetup()
{
	my ($watch) = @_;
	$watch->{filer}->filer_say(PRIVSETDIAG . "prof reset");
}

sub profStart()
{
	my ($watch) = @_;
	$watch->{filer}->filer_say(PRIVSETDIAG . "prof on; prof reset", "^\$");
}

sub profStop()
{
	my ($watch) = @_;
	$watch->{filer}->filer_say(PRIVSETDIAG . "prof off", "^\$");
	$watch->{filer}->filer_say(PRIVSETDIAG . "prof dump", "^\$");
	$watch->{filer}->filer_say(PRIVSETDIAG . "prof free", "^\$");
}

sub profReport()
{
	my ($watch) = @_;
	my $filer = $watch->{filer};

	my $kern = $watch->{kernel};
	if (!-r $kern) {
		$kern = "/rlse/$kern";
		logWarn("profile - no kernel \"$kern\"") unless -r $kern;
	}
	my $profCmd;
	my $flags = $watch->{flags};
	if ($flags =~ /3/) {
		$profCmd = "i386-netapp-gprof";
	} else {
		logWarn("profile - unknown kernel type \"$flags\"");
	}

	my $vol = sparse_volume($filer, "vol0");
	my $bld = $watch->{build};
	my $profile = "$vol->{mountPoint}/gmon-$bld-$flags.out";
	my $outfile = "gprof.$bld.$watch->{ts}";
	logWarn("profile - no profile found \"$profile\"") unless -r $profile;

	$profCmd .= " $kern ${profile}* > $outfile";
	logTrc($profCmd);
	system($profCmd);
	if ((stat($outfile))[7] == 0) {
		logWarn("failed to create profile: $profCmd");
		system("rm $outfile");
	}
}

# ------------------------------------------------------------------------

sub statitSetup() { }

sub statitStart()
{
	my ($watch) = @_;
	$watch->{filer}->filer_say(PRIVSETDIAG. "statit -b", "^\$");
}

sub statitStop()
{
	my ($watch) = @_;
	$watch->{statit} =
		$watch->{filer}->filer_say(PRIVSETDIAG. "statit -e");
}

sub statitReport()
{
	my ($watch) = @_;

	logTrc("write out statit report");
	open(RPT, "> statit.$watch->{build}.$watch->{ts}");
	print RPT $watch->{statit};
	close(RPT);
}

# ------------------------------------------------------------------------

sub wSuspSetup() { }

sub wSuspStart()
{
	my ($watch) = @_;
	$watch->{filer}->filer_say(PRIVSETDIAG. "wafl_susp -z", "^\$");
}

sub wSuspStop()
{
	my ($watch) = @_;
	$watch->{wsusp} =
		$watch->{filer}->filer_say(PRIVSETDIAG. "wafl_susp -w");
}

sub wSuspReport()
{
	my ($watch) = @_;

	logTrc("write out wafl susp report");
	open(RPT, "> wsusp.$watch->{build}.$watch->{ts}");
	print RPT $watch->{wsusp};
	close(RPT);
}

# ------------------------------------------------------------------------

sub wfStatSetup() { }

sub wfStatStart()
{
	my ($watch) = @_;
	$watch->{filer}->filer_say(PRIVSETDIAG. "waffinity_stats zero", "^\$");
}

sub wfStatStop()
{
	my ($watch) = @_;
	$watch->{wfstat} =
		$watch->{filer}->filer_say(PRIVSETDIAG. "waffinity_stats");
}

sub wfStatReport()
{
	my ($watch) = @_;

	logTrc("write out waffinity stats report");
	open(RPT, "> wfstat.$watch->{build}.$watch->{ts}");
	print RPT $watch->{wfstat};
	close(RPT);
}

# ------------------------------------------------------------------------

sub pkttSetup()
{
	my ($watch) = @_;
	$watch->{filer}->filer_say("pktt stop all");
}

sub pkttStart()
{
	my ($watch) = @_;
	$watch->{filer}->filer_say("pktt start all -d /",
				   "started packet trace");
}

sub pkttStop()
{
	my ($watch) = @_;
	$watch->{filer}->filer_say("pktt pause all", "Tracing suspended");
}

sub pkttReport()
{
	my ($watch) = @_;
	my $filer = $watch->{filer};
	my $fname = "pktt.$$.trc";
	$filer->filer_say("pktt dump all");
	$filer->filer_say("pktt stop all", "Tracing stopped");

	my $list = $filer->filer_say("pktt list");
	if ($list =~ /\.trc$/) {
		logDescribe("pktt trace file $list");
	} else {
		logWarn("pktt - no trace files found $list");
	}
}

# ------------------------------------------------------------------------

1;

__END__
