#Hey emacs, this is basically -*- perl -*-
#$Id: //depot/prod/test/main/lib/SPARSE/Util/Perf.pm#1 $
#
# Copyright (c) 2003 Network Appliance, Inc.
# All rights reserved.

## @summary	Library that provides performance testing APIs
## @description
## 	Functions to encapsulate performance testing and statistics collection
##
## @author dl-sparseeng

package SPARSE::Util::Perf;

use strict;
use Tharn;
use SPARSE::Util::Logger;
use Time::HiRes qw(gettimeofday tv_interval);

# ------------------------------------------------------------------------

require Exporter;
our @ISA	= qw(Exporter);
our @EXPORT	= qw(
		&PerfCreate
		);

# ------------------------------------------------------------------------

## @name	::PerfCreate
## @summary     Create a Perf object
## @description
## 	Create a new Perf object to measure a performance run.
##
##      Caller is expected to write code of the form:
##
##             $p = PerfCreate()
##             $p->monitor_filer($filer); # SPARSE::Resource::Filer obj
##             $p->monitor_filer($filer); # ...
##
##             $p->begin('title')
##             while ... {
##                 $p->op_begin()
##                      [do something]
##                 $p->op_end()
##             }
##             $p->end()
## @return     Perf object

sub PerfCreate()
{
	my $self = {};
	@{$self->{filers}} = ();
	$self->{title} = "uninitialized";
	return bless($self);
}

sub _statit_begin($)
{
	my ($self) = @_;

	for my $filer (@{$self->{filers}}) {
		logStep("starting statit for $self->{title} ".
			"run on $filer->{hostName}");
		$filer->filer_say("priv -level advanced; statit -b",
			'(^$)|the unused stats snapshot is being discarded');
	}
}

## @name	::begin
## @summary     Start a performance run
## @description
## 	Start the performance run.  Begin collecting results.
## @arg         title required name of performance run

sub begin($$)
{
	my ($self, $title) = @_;

	$self->{title} = $title;

	logTrc("Perf begin $self->{title} run");

	$self->{title} = $title;
	$self->{max_op_time} = 0;
	$self->{min_op_time} = 1000000; # MAXINT... humbug
	$self->{last_begin} = [gettimeofday];

	$self->_statit_begin();
}

sub _statit_end($)
{
	my ($self) = @_;

	for my $filer (@{$self->{filers}}) {
		my $statit_output = $filer->filer_say(
			"priv -level advanced; statit -e");
		my $logfile;
		my $log;
		if (defined($::LOGDIR) && $::LOGDIR ne "-" && $::LOGDIR ne "") {
			$logfile = $::LOGDIR . "/statit_" .
				$self->{title} . "_" . $filer->{hostName};
			$logfile =~ s/ /_/g;
			logStep("statit output for $self->{title} ".
				"run on $filer->{hostName} ".
				"saved in $logfile");
		} else {
			# STDOUT
			$logfile = "-";
			logStep("statit output for $self->{title} ".
				"run on $filer->{hostName} ".
				"follows");
		}
		open($log, ">".$logfile) ||
			logError("Perf: unable to open statit log: $logfile");
		print $log $statit_output;
		close($log) ||
			logError("Perf: unable to close statit log: $logfile");
	}
}

## @name	::end
## @summary     End a performance run
## @description
## 	End the performance run.  Save and/or display results.

sub end($)
{
	my ($self) = @_;

	logTrc("Perf end $self->{title} run");

	my $now = [gettimeofday];

	$self->_statit_end();

	my $elapsed = tv_interval($self->{last_begin}, $now);
	my $avg_op_time = $elapsed / $self->{ops};

	# XXX anything more appropriate than logStep for this?
	logStep("title:         " . $self->{title});
	logStep("num ops:       " . $self->{ops});
	logStep("time elapsed:  " . $elapsed);
	logStep("avg op time:   " . $avg_op_time);
	logStep("min op time:   " . $self->{min_op_time});
	logStep("max op time:   " . $self->{max_op_time});
	logStep("(all times in seconds)");

	$self->{elapsed_time} = $elapsed;
}

sub op_begin($)
{
	# This is per-op.  Be as efficient as possible.
	my ($self) = @_;
	$self->{last_op_begin} = [gettimeofday];
}

sub op_end($)
{
	# This is per-op.  Be as efficient as possible.
	my ($self) = @_;
	my $now = [gettimeofday];
	my $elapsed = tv_interval($self->{last_op_begin}, $now);
	$self->{max_op_time} = $elapsed if $elapsed > $self->{max_op_time};
	$self->{min_op_time} = $elapsed if $elapsed < $self->{max_op_time};
	++$self->{ops};
}

sub elapsed_time($)
{
	my ($self) = @_;
	return $self->{elapsed_time};
}

## @name	::monitor_filer
## @summary	Monitor a filer during performance runs
## @description
## 	Add the given SPARSE::Resource::Filer object to a list of
##      filers to monitor during the performance run.
## @arg
##      filer required filer to monitor (SPARSE::Resource::Filer object)

sub monitor_filer($$)
{
	my ($self, $filer) = @_;

	logTrc("Perf monitoring enabled for filer \"$filer->{hostName}\"");
	push(@{$self->{filers}}, $filer);
}

# ------------------------------------------------------------------------

1;
