# -*- perl -*-
#  $Id: //depot/prod/test/main/lib/Volume.pm#32 $

## @summary Library of volume commands
## @author dl-waflqa@netapp.com
##         (FSS WAFL Quality)
## @cf     false
## @status public
## @pod here

# Submitted into Perforce via burt103435 - Reorganize Test Libraries

package Volume::StatusObject;
use Class::MethodMaker
    new_hash_init => 'new',
    get_set       => [ qw(status type state aggr) ],
    hash          => [ qw(options) ];

package Volume;
use strict;
use Tharn;
use Data::Dumper;
use Module::Common qw(validate_mixed);
use Aggregate::ImplTemplate;
use Filer::Version qw(get_wafl_version supported_feature UNSUPPORTED_FEATURE_EXCEPTION are_we_ontapng);
use SPARSE::Util::Logger;
use Carp qw(croak);
use Sysconfig;
use Utils::Units qw (:constants convert_units);
use Zapi;
use ZapiErrno;
use NaElement;


# use with disk_free() command.
use constant DF_SPACE   => 'space';
use constant DF_INODES  => 'inodes';
use constant DF_PVT_INODES  => 'private inodes';
use constant DF_RESERVE => 'reserve';
our($Impl, $Validation, $VERSION);

BEGIN
{
	$Validation = 1;
	$VERSION    = '0.01';

	my @df_space_types  = qw( DF_SPACE DF_PVT_INODES DF_INODES DF_RESERVE );
	my @vol_subs        = qw(
		is_disk_hot
		make_disk_hot
		disk_blocks_written
		vol_add
		vol_clone
		vol_clone_split
		vol_clone_split_status
		vol_copy
		vol_container
		vol_count_inherit
		vol_create
		vol_create_flex
		vol_destroy
		vol_disk_free
		vol_disks
		vol_exists
		vol_join
		vol_lang
		vol_list
		vol_max_files
		vol_meta_i2p
		vol_mirror
		vol_offline
		vol_online
		vol_options
		vol_playlist_copy_in
		vol_playlist_copy_out
		vol_playlist_dump
		vol_playlist_status
		vol_purge
		vol_raidgroup_status
		vol_read_fsid
		vol_rename
		vol_restrict
		vol_rewrite_fsid
		vol_root
		vol_scrub
		vol_snap_reserve
		vol_size
		vol_split
		vol_spare_status
		vol_status
		vol_match_plex_raid_info
		vol_wait_change
                vol_blocks 				
		vol_unsparse
		vol_vvlabel
                tricky_vol_create_zapi
		vol_determine_flex_root_size
		volume_options_list_info
		volume_special_options_hook
		volume_set_option
	);

	Module::Common::setup_export_tags(
		# note: vol_disk_free is preferred over disk_free
		df  => [ 'vol_disk_free', 'disk_free', @df_space_types ],
		vol => [ @vol_subs ]
	);

	# $Impl object is global in this package
	$Impl = Aggregate::ImplTemplate->new(
          cmd        => 'vol',
          validation => $Validation
        );
}

use Params::Validate
	qw(validate_pos validate BOOLEAN SCALAR ARRAYREF HASHREF OBJECT);
# use Admin::License; # enable this when we enable warning prompt for snaplock
use SayWrap qw(sayEOF);
use Utils qw(make_conn_and_complain);
	# XXX - upon popular request, make_conn() shall be used to allow
	#       hostspec/Hostspec/Connectrec as the "filer" argument. But
	#       we want to make sure to complain if it's passed anything
	#       other than a Connectrec, so we will use
	#       make_conn_and_complain(), which will "complain loudly"
	#       if anything other than a Connectrec is passed.
use Utils::Units qw(:convert :constants);
use Aggregate qw(aggr_wait_change);
use Filer::Version qw(get_wafl_version WV_SCRIMSHAW_MAX UNSUPPORTED_FEATURE_EXCEPTION);
use Data::Dumper;

# XXX - these are mappings used for the disk_free() function
my %DF_REPORT_TYPES = (
	UNIT_B()  => '-b', # is this supported?
	UNIT_KB() => '-k',
	UNIT_MB() => '-m',
	UNIT_GB() => '-g',
	UNIT_TB() => '-t', # is this supported?
);

my %DF_SPACE_TYPE_OPTS = (
	DF_SPACE()   => '',
	DF_INODES()  => '-i',
	DF_PVT_INODES()  => '-P -i',
	DF_RESERVE() => '-r'
);

=head1 NAME

Volume - Basic Volume Functions

=head1 SYNOPSIS

  use Volume qw(:all);
  # $conn = connect(...);

  $rv = vol_create($conn, $vol, [ $disk1, $disk2 ... ]);
  $rv = vol_create_flex($conn, $vol, 'myaggr', '20m'[, ...]);
  $rv = vol_copy($conn, $vol_src, $vol_dest, [ options ... ]);
  $rv = vol_copy_status($conn, $opno, [ options ... ]);
  $rv = vol_copy_abort($conn, $opno, [ options ... ]);
  $rv = vol_clone($conn, $clone_vol, $parent_vol, [ options ]);
  $rv = vol_clone_split($conn, $clone_vol, [ options ]);
  $rv = vol_clone_split_status($conn, $clone_vol);
  $rv = vol_destroy($conn, $vol, { destroy_corrupt_volumes => ... });
  $rv = vol_purge($conn, $vol, { destroy_corrupt_volumes => ... });
  $rv = vol_meta_i2p($conn, $aggr);
  $rv = vol_online($conn, $vol);
  $rv = vol_offline($conn, $vol);
  $rv = vol_restrict($conn, $vol);
  $rv = vol_rename($conn, $oldvol, $newvol);
  $rv = vol_exists($conn, $vol);
  $rv = vol_wait_change($conn, $vol, want => { status => qr(online) });
  $rv = vol_scrub($conn, $vol, $mode);
  $rv = vol_unsparse($conn, $path, $name, $size);
  $rv = is_disk_hot ($conn, $vol, $chosen_disk);
  $rv = make_disk_hot ($conn, $vol, $chosen_disk, $interval) ;
  $rv = disk_blocks_written ($conn, $vol_name, $chosen_disk);
  # ($completed, $lastscrub) = vol_scrub($conn, $vol, 'status');

  # old style of getting info from vol_status()
  ($state, $status, $aggr, %options) = vol_status($conn, $vol);

  # new style of getting info from vol_status()
  $status_obj = vol_status($conn, $vol);
  $type   = $status_obj->type();
  $state  = $status_obj->state();
  $status = $status_obj->status();
  $aggr   = $status_obj->parent();
  %options= $status_obj->options();

  %options = vol_options($conn, $vol);

  # the ZAPI interfaces to 'vol options':
  volume_set_option(host=>$host,volume=>$volname,option=>"option",value=>"value");
  %options = volume_options_list_info(host=>$host,volume=>$volname);

  @blocks= vol_blocks($conn, $vol);
  $size  = vol_size($conn, $vol);
  @disks = vol_disks($conn, $vol);
  $vname = vol_root($conn);
  $num   = vol_max_files($conn, $vol);

  $rv = disk_free($conn, $path, $space_type, $report_type);

  vol_playlist_copy_in($conn, $vol, $path);
  vol_playlist_copy_out($conn, $vol, $path);
  $str     = vol_playlist_dump($conn, $vol);
  $hashref = vol_playlist_status($conn, $vol);

  $hash = vol_spare_status($conn);
  $hash = vol_raidgroup_status($conn, $vol);

  # load specific parts only (recommended)
  use Volume qw(:df);  # load disk_free() and DF_* constants
  use Volume qw(:vol);  # load commonly used vol_* functions

  # turn validation off
  {
    local $Volume::Validation = 0;
    vol_online(...);
  }

=head1 DESCRIPTION

This module defines routines related to OnTap's volumes.

This software requires at least Anchorsteam, because it uses many
"Virtual Volume" features that are not available in previous releases.

This library is still in beta status, testers welcome, but please be
warned that the behavior and interface may change.

=head2 $Volume::Validation

Certain functions have the ability to issue extra command(s) and double
check if the function actually had the intended effect. On very rare
occasions when you don't want this validation to be done, you can turn it
on/off via a package variable $Volume::Validation

Validation is turned on by default. To turn off validation, set the value
C<$Volume::Validation> to a false value. If you combine this with the
C<local()> operator, you can limit the effect to a block:

  use Volume qw(vol_online);
  vol_online(...); # validation effective
  {
    local $Volume::Validation = 0;
    vol_online(...); # no validation
  }
  vol_online(...); # validation effective

Currently C<vol_create()>, C<vol_create_flex>, C<vol_add()>, C<vol_online()>,
C<vol_offline()>, C<vol_restrict()>, C<vol_size()>, and C<vol_rename()>
support this feature.

=head1 FUNCTIONS

All functions, unless otherwise stated, return a true value on success,
and false on failure. On failure C<$@> is set to the appropriate error
value via C<die()>.

Function declarations below all take C<$conn>, which is a Connection object
(Connectrec) that is created by calling C<Tharn::connect()>, and a volume
name (C<$vol>)

=head2 vol_create($conn,$vol,$disklist[,%opts])

Creates a new paired volume.

  WARNING: This function currently only creates paired volumes, as it
  was ported over from a library predating virtual volumes. In the
  next release this function will be renamed to vol_create_paired(),
  and vol_create() will be for virtual volume creation only.
  See burt 105917

The C<$disklist> argument can be either a reference to an array, or a scalar.
If it's scalar, C<vol_create()> will try to use it as the
E<lt>ndisksE<gt>[@E<lt>disk-sizeE<gt>] notation, while if it's an array ref,
the contents of the array will be treated as disk names and they will be
prepended with the '-d' option.

You may specify the various difference options in C<%opts>
(if any of the following is unspecified, the system default value will be
used):

=over 4

=item mirror

Enable/disable volume mirroring

=item force

Force volume creation

=item language

Specify the language on the volume

=item rgsize

Specify the raidgroup size.

=item raidtype

Specify the raid type. The value must be either 'raid4' or 'raid_dp'

=item randomize_args

Randomizes the command line arguments (except for disk-list)

=item snaplock

If specified, attempts to create a snaplock volume (via the -L option
to "vol create"). Default is false.

=item snaplock_enterprise

If specified, attempts to create a snaplock_enterprise volume (via the -L
option to "vol create"). Default is false.

=item timeout

The time in seconds to wait until vol_create() confirms that the created
volume exists and is online. Default is 3600.

=back

If $Validation is turned on, then vol_create() will try to make sure that
the volume that was just created actually comes online.

Examples:

  # create a volume with 2 disks
  vol_create($conn, 'vol1', 2);

  # create a volume using disks '2a.16' and '2a.17'
  vol_create($conn, 'vol1', [ qw( 2a.16 2a.17 ) ]);

  # create a snaplock compliant volume
  vol_create($conn, 'vol1', 2, snaplock => 1);

=cut

sub vol_create
{
	my ($conn, $vname, @disks) = @_;

	my $VSERVER_NAME = $main::VSERVER_NAME;
	my $VSERVER_ROOTVOL = $main::VSERVER_ROOTVOL;
	my $VSERVER_ROOTVOLUME_SECURITY_STYLE = $main::VSERVER_ROOTVOLUME_SECURITY_STYLE; 
	my $FILER10M_AGGR = $main::FILER10M_AGGR;
    my $VSERVER_DLIF = $main::VSERVER_DLIF;

	if (my $server = are_we_ontapng($_[0])) {
		#return $server->volume_online(volume=>$_[1]);
		
			# Find if tricky or not  
			my $isTricky=0;
			eval {
        		#Finding the version in a tricky filer
				my $result = $server->system_image_show('iscurrent'=>'true');
        		my $entry;
        		my $osString; 
					foreach $entry (@$result) {
						my %hashVal = %$entry;
            			$osString = $hashVal{'version'};
						logStep(Dumper($entry));
            			logStep("Oh my version = $osString");

					}
				logStep("OSversion : $osString");

				if (($osString =~ /ricky/) || ($osString =~ /10.0.4/)) {
					$isTricky = 1;
				}
			};
    		if ($@) {
        		#Finding the version in a BR filer
				my $result = $server->system_node_image_show('iscurrent'=>'true');
        		my $entry;
       			my $osString; 
				foreach $entry (@$result) {
					my %hashVal = %$entry;
            		$osString = $hashVal{'version'};
					logStep(Dumper($entry));
            		logStep("Oh my version = $osString");
				}
				logStep("OSversion : $osString");
				if (($osString =~ /ricky/) || ($osString =~ /10.0.4/)) {
					$isTricky = 1;
				}

		    }

		# mica create a volume if not present 
		logStep("##mica Checking 10Mode volume $vname.");
		eval {
        	$server->volume_show(vserver=>$VSERVER_NAME,
                			volume=>$vname);
		};
		if ($@) {
			eval {
  				    logStep("##mica vol create a 10M/Tricky filer vol ");

					if ($isTricky == 1) {
						if ($main::VSERVER_AGGR_TYPE =~ /striped/) {
						logStep("Creating Tricky (striped) volume: $vname on aggr: $FILER10M_AGGR");
						} else {
						logStep("Creating Tricky (flex) volume: $vname on aggr: $FILER10M_AGGR");
						}
        				$server->volume_new(vserver=>$VSERVER_NAME,
                			volume=>$vname,
                            aggregate=>$FILER10M_AGGR,
                            size=>'4g',
							"unix-permissions"=>"---rwxrwxrwx",
                            "junction-path"=>"/$vname");
					
					} else {
						if ($main::VSERVER_AGGR_TYPE =~ /striped/) {
							logStep("Creating BR striped (coral) volume: $vname on aggr: $FILER10M_AGGR");
        					$server->volume_new(vserver=>$VSERVER_NAME,
                			volume=>$vname,
                            aggregate=>$FILER10M_AGGR,
                            size=>'4g',
							'stripe-format'=>'data',
							"unix-permissions"=>"---rwxrwxrwx",
                            "junction-path"=>"/$vname");					
						} else { 
							logStep("Creating BR Flex volume: $vname on aggr: $FILER10M_AGGR");
        					$server->volume_new(vserver=>$VSERVER_NAME,
                			volume=>$vname,
                            aggregate=>$FILER10M_AGGR,
                            size=>'4g',
							"unix-permissions"=>"---rwxrwxrwx",
                            "junction-path"=>"/$vname");
                    	}
					
					}

			};
			if ($@) {
				exit -1;
			} else {		
				return 1;
			}
		} else {
			logStep("##mica the volume $vname already exists .");
		}

	}

	$Impl->validation($Validation);
	return ($Impl->cmd_create(@_));
}


=head2 vol_create_flex($conn, $vol, $aggr, $size[, %args])

Creates a new flexible volume.

Besides connection C<$conn> and volume name C<$vol>
You may specify the following options in C<%args>:

=over 4

=item aggr

Required. Name of the containing aggregate.

=item size

Required. Specified as a number (defaults to bytes
and is rounded up to the nearest 4kb).
or #k to specify size in kilobytes. Ex: 10000k
or #m to specify size in megabytes. Ex:   200m
or #g to specify size in gigabytes. Ex:     3g
or #t to specify size in terabytes. Ex.     1t

=item skip_volume_special_options_hook
Optional. Default is zero. If '1', then no call
is made to 'volume_special_options_hook'. In this case,
all 'volume_special_options_hook' related options such
as compression, encription, A_SIS, etc. will be ignored.

=item guarantee
Optional. Preallocates space. Defaults to 'volume' which
means that the size of the volume will be preallocated

=item compression
Optional. Can be used to create a volume without compression
even if COMPRESSION_ENABLE is turned on

=item encryption
Optional. Can be used to create a volume without encryption
even if ENCRYPTION_ENABLE is turned on

=item language

Optional. Specify the language on the volume.

=item source

Optional. Source volume for creating a flexcache volume.
Format is 'remotehost:remotevolume'.

=item timeout

Optional. The time in seconds to wait until vol_create() confirms that the
created volume exists and is online. Default is 3600.

=back

If $Validation is turned on, then vol_create_flex() will try to make sure that
the volume that was just created actually comes online.

Examples:

  # create a flexible volume 'new_vol' 30 mb big on aggr 'MY_aggr'
  vol_create_flex($conn, 'new_vol', 'MY_aggr', '30m');

  # create a flexible volume 'another_vol' 41938944  bytes big (40956kb)
  # on aggregate 'TEST'. Preallocate space for space-reserved files
  # and LUNs on the volume

  vol_create_flex($conn, 'another_vol', 'TEST', '41943000',
		  guarantee=>'file'
		 );
 # Note that the size of 41943000 without the units gets interpreted as
 # 41943000 bytes and rounded to the nearest 4kb - 41938944 bytes (40956kb)

  # create a flexcache volume 'flexcache_vol' 30 mb big on aggr 'MY_aggr'
  # with a source volume 'src_vol' on filer 'origin'
  vol_create_flex($conn, 'flexcache_vol', 'MY_aggr', '30m',
                  source => 'origin:src_vol');

=cut

sub vol_create_flex
{
	logcomment("Entering vol_create_flex  Line: ".__LINE__);

	my $conn = make_conn_and_complain(shift);
	my($vol, $aggr, $size, %args) = validate_mixed( @_,
		[
			{ type => SCALAR },
			{ type => SCALAR },
			{ type => SCALAR },
		],
		{
			language 	=> { type => SCALAR, optional => 1 },
			guarantee  	=> { type => SCALAR, optional => 1 },
			compression  => { type => SCALAR, optional => 1 },
			encryption  	=> { type => SCALAR, optional => 1 },
                        asis                 => { type => SCALAR, optional => 1 },
			source   	=> { type => SCALAR, optional => 1 },
			skip_volume_special_options_hook
			    => { type => SCALAR, optional => 1, default => 0 },
			timeout 	=> { type => SCALAR, optional => 1,
				     	default => 3600, regex => qr/^\d+$/ }
		}
	);
        
	#mica - if the filer is 10M filer use 10M call      
	if (my $server = are_we_ontapng($conn)) {
		#	return eval {
		#my @vslist = $server->vserver_show();
		#$server->volume_new(volume=>$vol, vserver=>$vslist[0]->{vserver}, parent=>$vslist[0]->{rootvolume}, "junction-path"=>$vol, aggregate=>$aggr, size=>$size);
		#return 1;
		#};

		my $VSERVER_NAME = $main::VSERVER_NAME;
		my $FILER10M_AGGR = $main::FILER10M_AGGR;


			# Find if tricky or not  
			my $isTricky=0;
			eval {
        		#Finding the version in a tricky filer
				my $result = $server->system_image_show('iscurrent'=>'true');
        		my $entry;
        		my $osString; 
					foreach $entry (@$result) {
						my %hashVal = %$entry;
            			$osString = $hashVal{'version'};
						logStep(Dumper($entry));
            			logStep("Oh my version = $osString");

					}
				logStep("OSversion : $osString");

				if (($osString =~ /ricky/) || ($osString =~ /10.0.4/)) {
					$isTricky = 1;
				}
			};
    		if ($@) {
        		#Finding the version in a BR filer
				my $result = $server->system_node_image_show('iscurrent'=>'true');
        		my $entry;
       			my $osString; 
				foreach $entry (@$result) {
					my %hashVal = %$entry;
            		$osString = $hashVal{'version'};
					logStep(Dumper($entry));
            		logStep("Oh my version = $osString");
				}
				logStep("OSversion : $osString");
				if (($osString =~ /ricky/) || ($osString =~ /10.0.4/)) {
					$isTricky = 1;
				}

		    }

		# mica create a volume if not present 
		logStep("##mica Checking 10Mode volume $vol");
		eval {
        	$server->volume_show(vserver=>$VSERVER_NAME,
                			volume=>$vol);
		};
		if ($@) {
			eval {
			logStep("##mica vol create a 10M/Tricky filer vol ");

					if ( $isTricky == 1 ) {
						if ($main::VSERVER_AGGR_TYPE =~ /striped/) {
						logStep("Creating Tricky (striped) volume: $vol on aggr: $FILER10M_AGGR");
						} else {
						logStep("Creating Tricky (flex) volume: $vol on aggr: $FILER10M_AGGR");
						}
						$server->volume_new(vserver=>$VSERVER_NAME,
                			volume=>$vol,
                            aggregate=>$FILER10M_AGGR,
                            size=>$size,
							"unix-permissions"=>"---rwxrwxrwx",
                            "junction-path"=>"/$vol");
					
					} else {
						if ($main::VSERVER_AGGR_TYPE =~ /striped/) {
							logStep("Creating BR striped (coral) volume: $vol on aggr: $FILER10M_AGGR");
        					$server->volume_new(vserver=>$VSERVER_NAME,
                			volume=>$vol,
                            aggregate=>$FILER10M_AGGR,
                            size=>$size,
							'stripe-format'=>'data',
							"unix-permissions"=>"---rwxrwxrwx",
                            "junction-path"=>"/$vol");					
						} else { 
							logStep("Creating BR flex volume: $vol on aggr: $FILER10M_AGGR");
        					$server->volume_new(vserver=>$VSERVER_NAME,
                			volume=>$vol,
                            aggregate=>$FILER10M_AGGR,
                            size=>$size,
							"unix-permissions"=>"---rwxrwxrwx",
                            "junction-path"=>"/$vol");
                    	}
					
			}


			#make volume   
			#logStep("##mica creating 10M/Tricky volume .");
			#$server->volume_new(vserver=>$VSERVER_NAME,
			#    			volume=>$vol,
			#                aggregate=>$FILER10M_AGGR,
			#                size=>$size,
			#				"unix-permissions"=>"---rwxrwxrwx",
			#                "junction-path"=>"/$vol");
			};
			if ($@) {
				exit -1;
			} else {		
				return 1;
			}
		} else {
			logStep("##mica the volume $vol already exists.");
		}
		
		return 1;
    } #end 10M filer call check


    # check for attempts to create flex volumes on pre-anchorsteam systems
	my $msgref;
	if (!supported_feature($conn->hostp(), 'FLEXVOL', \$msgref)) {
            logTrc("vol_create_flex: can't continue - version is pre-Anchorsteam: '$aggr'");
            Carp::croak na_result(UNSUPPORTED_FEATURE_EXCEPTION .". ". $msgref);
	}

	return eval {
                   
		# now we create command to send to the filer
		my $cmd = "vol create $vol ";
		my %say_args = %args;
		delete($say_args{skip_volume_special_options_hook});
		if ($args{language}) {
			$cmd .= "-l $args{language} ";
			delete $say_args{language};
		}
		if ($args{guarantee}) {
			$cmd .= "-s $args{guarantee} ";
			delete $say_args{guarantee};
		}
		my @special_volume_options = qw(compression encryption asis a_sis);
		foreach (@special_volume_options) {
                        delete $say_args{$_};
		}
		
		$cmd .= "$aggr ";
		$cmd .= "$size ";

        #mica - got to do some stuff here 
		#if the origin filer is a 10m/Tricky filer then 
		#use the dlif ip to mount stuff

		my $ngorigin_check=0;
		my $dlifIp;

        logTrc ("#mica Origin Filer is ($main::FILER)");
		if ($main::SRC_10MODE == 1) {
			#my $f = host($args{source});
			#if ($f->hosttype =~ /filer-ng-cfe/m)
			$ngorigin_check = 1;
			my $f = host($main::FILER);
			$dlifIp = $f->{DATA_IP1};
		}

		my $VSERVER_ROOTVOL = $main::VSERVER_ROOTVOL;
		if ($args{source}) {
			if($ngorigin_check == 1) {
				my @volAll = split(':',$args{source});
                my $filerO = $main::FILER;
 				my $originSrc = $volAll[0];
				my $volOnly = $volAll[1];
				logTrc("#mica == 10Mode Volume using dlif ip in -S switch");
				if ($volOnly =~ /vol0/) { #some tests use an existing vol for 7g 
					$volOnly = $VSERVER_ROOTVOL;
				} 
				if ($originSrc =~ /$filerO/) {
				} else {
					$dlifIp = $originSrc
				}
				$cmd .= "-S $dlifIp".":"."$volOnly";
				delete $say_args{source};
			} else {
				$cmd .= "-S $args{source}";
				delete $say_args{source};
			}
		}

		# Allow 'skip_volume_special_options_hook' to override individual
		# 'volume special option' values for compression, encription,
		# A_SIS, etc. If this param is set, then 'volume_special_options_hook'
		# should not be called in any case.
		# (currently, only implement 'compression' volume special option.)
		if ($args{skip_volume_special_options_hook}) {
                	$args{compression} = "no";
		}


        logTrc ("#mica : The constructed command ===> $cmd");
		# check that we have succeeded
		my $text = sayEOF($conn, $cmd, %say_args);
		unless ($text =~ "Creation of volume \'$vol\'"
			&& $text =~ "has completed") {
			die "$text\n";
		}
    
		if ($args{compression} =~ /no/  || $args{asis} =~ /no/ || $args{encryption} =~ /no/) {
                logDbg("Volume::volume_special_options_hook not called becuase  ....".
                               " compression=$args{compression} or asis=$args{asis} or encryption=$args{encryption}");
        } else {
                 logDbg("Volume::vol_create_flex  Calling ".
                 "Volume::volume_special_options_hook  ....");
                  Volume::volume_special_options_hook(filer=>$conn, volume=>$vol);
        }
        # validate that we have succeeded
        if ($Validation) {
                        vol_exists($conn, $vol) or die $@;
			vol_wait_change($conn, $vol,
					timeout => $args{timeout},
					want    => {
					   state => qr(^online$),
					   status => qr(normal|mirrored|raid)
						   }
        ) or die $@;

		# XXX/tapl This check is not required. Completion of
		# flex vol creation does not rely ont the readiness of
		# the underlying aggregate.
		aggr_wait_change($conn, $aggr,
					 timeout => $args{timeout},
					 state   => qr(^online$),
					 status  => qr(normal|mirrored|raid)
					) or die $@;
		}
		
		return 1;
	
	}; # eval block
	logcomment("Exiting vol_create_flex  Line: ".__LINE__);
}


=head2 vol_clone($conn,$clone,$parent, %args)

Clone a volume.

=cut

sub vol_clone($$$;%)
{
	my $conn = make_conn_and_complain(shift);

	my ($clone, $parent, %args) = validate_mixed( @_,
		[
			{ type => SCALAR },
			{ type => SCALAR },
		],
		{
			snap_name	=> { type => SCALAR, optional => 1 },
			guarantee  	=> { type => SCALAR, optional => 1 },
		}
	);

	return eval {
		my $cmd = "vol clone create ".$clone;
		$cmd .= " -s ".$args{guarantee} if $args{guarantee};
		$cmd .= " -b ".$parent;
		$cmd .= " ".$args{snap_name} if $args{snap_name};


		# create the clone
		my $text = sayEOF($conn, $cmd);
		die $@ if $@ =~ /not found|create a new|already exists/;

		# validate that we have succeeded
		if ($Validation) {
			vol_exists($conn, $clone) or die $@;
			vol_wait_change($conn, $clone,
				timeout => $args{timeout} ? $args{timeout} : $main::TIMEOUT,
				want    => {
				   state => qr(^online$),
				   status => qr(normal|mirrored|raid)
					   }
				) or die $@;
		}

		return 1;
 	};	# eval block
}


=head2 vol_clone_split($conn,$clone, %args)

Split a cloned a volume.

=cut

sub vol_clone_split($$;%)
{
	my $conn = make_conn_and_complain(shift);

	my ($clone, %args) = validate_mixed( @_,
		[
			{ type => SCALAR },
		],
		{
			wait	=> { type => SCALAR, optional => 1 },
		}
	);

	return eval {
		my $cmd = "vol clone split start ".$clone;

		# split the clone
		my $text = sayEOF($conn, $cmd);
		die $@ if $@ =~ /not found|create a new|already exists/;

		# validate that we have succeeded
		if ($args{wait}) {
			$cmd = "vol clone split status ".$clone;
			do {
				$text = sayEOF($conn, $cmd);
				sleep(2);
				die $text if $text =~ /not found|not being split/i;
				last if $text =~ /clone split complete/i;
			} while ($text !~ /The volume is not a clone/i);
		}
		return 1;
 	};	# eval block
}


=head2 vol_clone_split_status($conn,$clone)

Status of splitting a cloned volume.

=cut

sub vol_clone_split_status($$)
{
	my $conn = make_conn_and_complain(shift);
	my ($clone) = @_;
	return eval {
		my $text;
		my $cmd = "vol clone split status ".$clone;
		do {
			$text = sayEOF($conn, $cmd);
			sleep(2);
			die $text if $text =~ /not found|not being split/i;
			last if $text =~ /clone split complete/i;
		} while ($text !~ /The volume is not a clone/i);

		return 1;
 	};	# eval block
}


=head2 Volume::vol_copy($conn, $srcvol, $destvol, %opts);

Volume copy. Optional snap_name; optional wait for copy to complete.
Use vol_copy_status with return value if not waiting.

Only tested on small volumes which wait...

=cut

sub vol_copy($$$;%)
{
	my $conn = make_conn_and_complain(shift);

	my ($srcvol, $destvol, %opts) = validate_mixed( @_,
		[
			{ type => SCALAR },
			{ type => SCALAR },
		],
		{
			wait		=> { type => SCALAR, optional => 1 },
			snap_name  	=> { type => SCALAR, optional => 1 },
			restrict	=> { type => SCALAR, optional => 1 },
			all_snaps	=> { type => SCALAR, optional => 1 },
		}
	);

	#mica - if the filer is 10M filer use 10M call      
	if (my $server = are_we_ontapng($conn)) {

		my $VSERVER_NAME = $main::VSERVER_NAME;
		my $FILER10M_AGGR = $main::FILER10M_AGGR;
		my $foreground = "false";

		if ($opts{wait} == 1 ) {
			$foreground ="true";
		}

		if ($opts{restrict} == 1) {
			eval {
				logStep("##mica vol restrict a 10M/Tricky filer vol before volcopy ");

				#remove junctions first  
				logStep("##mica unmounting junctions before volcopy task.");
				$server->volume_unmount(vserver=>$VSERVER_NAME,
                			volume=>$srcvol);

				#make volume restricted - this does not work in 10M  
				#logStep("##mica restricting 10M/Tricky volume before volcopy task .");
				#$server->volume_modify(vserver=>$VSERVER_NAME,
				#			volume=>$srcvol, state=>'restricted');
			};
			if ($@) {
				logStep("##mica vol restrict before volcopy failed : $@");
				exit -1;
			} else {		
			}
		}

		eval {
			logStep("##mica vol copy a 10M/Tricky filer vol src and dest");
        	$server->volume_copy(vserver=>$VSERVER_NAME,
                			volume=>$srcvol,
                            'destination-aggregate'=>$FILER10M_AGGR,
                            'destination-volume'=>$destvol,
                            "foreground"=>'true');

			if ($opts{restrict} == 1) {
				#make volume restricted  
				#logStep("##mica restricting 10M/Tricky volume before volcopy task .");
				#$server->volume_modify(vserver=>$VSERVER_NAME,
				#			volume=>$srcvol, state=>'online');
			}
			
			#restore junctions first  
			logStep("##mica mounting junctions after volcopy task.");
			$server->volume_mount(vserver=>$VSERVER_NAME,
                			volume=>$srcvol,'junction-path'=>"/$srcvol");
		};
		if ($@) {
			logStep("##mica copying 10M/Tricky volume to another volume Failed : $@");
			exit -1;
		} else {		
			return 1;
		}

	} #end 10M filer call check
                                                                                                       
    # comes here if 7M 
	return eval {
		my $cmd = "vol copy start ";
		$cmd .= "-S " if $opts{all_snaps};
		$cmd .= "-s ".$opts{snap_name} if $opts{snap_name};
		$cmd .= " $srcvol $destvol";
		my $opid;

		# optionally restrict the volume first
		if ($opts{restrict}) {
			vol_restrict($conn, $destvol)
			    or die $@;
		}
		# start the copy
		my $text = sayEOF($conn, $cmd);
		die $@ if $@ =~ /not found|create a new|already exists/;
		die $text if $text =~ /Volume doesn't exist|Missing volume name|volume is online|restrict it before|is iron restricted|cannot open|Cannot init|cannot use both/i; #'
		if ($text =~ /100% done/) {
			return "done";
		} elsif ($text =~ /operation id.*(\d+)/) {
			$opid = $1;
		}
		if ($opts{wait}) {
			$cmd = "vol copy status $opid";
			do {
				sleep(2);
				$text = sayEOF($conn, $cmd);
			} while ($text !~ /100% done|No operations in progress./);
		}
		return 1;
 	};	# eval block
}

=head2 Volume::vol_copy_status($conn, $opid, %opts);

get the status of a prevously started vol copy operation

=cut

sub vol_copy_status($$;%)
{
	my $conn = make_conn_and_complain(shift);
	my ($opid, %opts) = validate_mixed( @_,
		[
			{ type => SCALAR },
		],
		{
			wait	=> { type => SCALAR, optional => 1 },
		}
	);

	return eval {
		my $text;
		my $cmd = "vol copy status $opid";
		do {
			$text = sayEOF($conn, $cmd);
			sleep(2);
		} while ($opts{wait} and $text !~ /100% done|No operations in progress./);
		return $text;
	};	# eval block
}

=head2 Volume::vol_copy_abort($conn, $opid, %opts);

abort a volume copy operation

=cut

sub vol_copy_abort($$;%)
{
	my $conn = make_conn_and_complain(shift);
	my ($opid, %opts) = @_;

	return eval {
		my $cmd = "vol copy abort $opid";
		my $text = sayEOF($conn, $cmd);
		return $text;
	};	# eval block
}

=head2 vol_destroy($conn,$vol,%args)

Destroy a volume.

=over 4

=item destroy_corrupt_volumes

If '1', then volume is destroyed even if it is corrupt. If '0'
then corrupt volumes are preserved

=back

=cut

sub vol_destroy
{
	my ($conn, $vname, %opts) = @_;

	my $VSERVER_NAME = $main::VSERVER_NAME;
	my $VSERVER_ROOTVOL = $main::VSERVER_ROOTVOL;
	my $VSERVER_ROOTVOLUME_SECURITY_STYLE = $main::VSERVER_ROOTVOLUME_SECURITY_STYLE; 
	my $FILER10M_AGGR = $main::FILER10M_AGGR;
    my $VSERVER_DLIF = $main::VSERVER_DLIF;

	if (my $server = are_we_ontapng($_[0])) {
		#return $server->volume_online(volume=>$_[1]);
		eval {
			#remove junctions 
			logStep("##mica unmounting junctions .");
			$server->volume_unmount(vserver=>$VSERVER_NAME,
                			volume=>$vname);

			#delete volume   
			logStep("##mica deleting volume .");
			$server->volume_delete(vserver=>$VSERVER_NAME,
                			volume=>$vname);
		};
		if ($@) {
			exit -1;
		} else {		
			return 1;
		}
	}

	if (!defined($opts{destroy_corrupt_volumes})) {
		$opts{destroy_corrupt_volumes} = $::DESTROY_CORRUPT_VOLUMES;
	}
	my @args = ($conn, $vname, %opts);
        $Impl->validation($Validation);
	return ($Impl->cmd_destroy(@args));
}

=head2 vol_add($conn,$vol,$disklist[,%opts])

Adds disks to a volume.
For parameter $disklist, if the type is scalar, specify the number of disks
to be added. If the type is an array, vol_add() will use the '-d' option
to add each disk via its disk name taken from each element of the array.

vol_add() accepts the following option in %opt :

=over 4

=item force

Force disk addition

=item rg

Specify which raidgroup to add this disk to

=item timeout

Time to wait for validation that the disk has been added

=back

=cut

sub vol_add
{

	$Impl->validation($Validation);
	return ($Impl->cmd_add(@_));
}

=head2 vol_purge($conn,$vol,%args)

A simple convenience function to set a volume offline, and then destroy it.

=over 4

=item destroy_corrupt_volumes

If '1', then volume is destroyed even if it is corrupt. If '0'
then corrupt volumes are preserved

=back

=cut

sub vol_purge
{
	#my ($conn, $vol, %opts) = @_;
	#    if (my $server = are_we_ontapng($_[0])) {
	#    return $server->volume_purge(volume=>$_[1]);
	#}

	my ($conn, $vname, %opts) = @_;

	my $VSERVER_NAME = $main::VSERVER_NAME;
	my $VSERVER_ROOTVOL = $main::VSERVER_ROOTVOL;
	my $VSERVER_ROOTVOLUME_SECURITY_STYLE = $main::VSERVER_ROOTVOLUME_SECURITY_STYLE; 
	my $FILER10M_AGGR = $main::FILER10M_AGGR;
    my $VSERVER_DLIF = $main::VSERVER_DLIF;

	if (my $server = are_we_ontapng($_[0])) {
		#return $server->volume_online(volume=>$_[1]);
		eval {
			#remove junctions 
			logStep("##mica unmounting junctions .");
			$server->volume_unmount(vserver=>$VSERVER_NAME,
                			volume=>$vname);

			#delete volume   
			logStep("##mica deleting volume .");
			$server->volume_delete(vserver=>$VSERVER_NAME,
                			volume=>$vname);
		};
		if ($@) {
			exit -1;
		} else {		
			return 1;
		}
	}

	if (!defined($opts{destroy_corrupt_volumes})) {
		$opts{destroy_corrupt_volumes} = $::DESTROY_CORRUPT_VOLUMES;
	}

	my @args = ($conn, $vname, %opts);
	$Impl->validation($Validation);
	return ($Impl->cmd_purge(@args));
}

=head2 vol_online($conn,$vol[,%opts])

Change the state of the volume to "online".
Returns success (if the volume was changed or is already in the desired state).
or failure otherwise (if the volume couldn't change or didn't exist).
vol_online() accepts the following option in %opts :

=over 4

=item force

Force onlining of volume.

=back

=cut

sub vol_online
{
 
	my ($conn, $vname, %opts) = @_;

	my $VSERVER_NAME = $main::VSERVER_NAME;
	my $VSERVER_ROOTVOL = $main::VSERVER_ROOTVOL;
	my $VSERVER_ROOTVOLUME_SECURITY_STYLE = $main::VSERVER_ROOTVOLUME_SECURITY_STYLE; 
	my $FILER10M_AGGR = $main::FILER10M_AGGR;
    my $VSERVER_DLIF = $main::VSERVER_DLIF;

	if (my $server = are_we_ontapng($_[0])) {
		#return $server->volume_online(volume=>$_[1]);
		eval {
			logStep("##mica vol online a 10M/Tricky filer vol ");

			#remove junctions first  
			#logStep("##mica unmounting junctions .");
			#$server->volume_unmount(vserver=>$VSERVER_NAME,
			#    			volume=>$vname);

			#make volume online  
			logStep("##mica onlining 10M/Tricky volume .");
			$server->volume_modify(vserver=>$VSERVER_NAME,
                			volume=>$vname, state=>'online');

			#remove junctions first  
			logStep("##mica mounting junctions .");
			$server->volume_mount(vserver=>$VSERVER_NAME,
                			volume=>$vname,'-junction-path'=>"/$vname");

		};
		if ($@) {
			exit -1;
		} else {		
			return 1;
		}
	}

	$Impl->validation($Validation);
	return ($Impl->cmd_online(@_));
}

=head2 vol_offline($conn,$vol[,%opts])

Change the state of the volume to "offline".
Returns success (if the volume was changed or is already in the desired state).
or failure otherwise (if the volume couldn't change or didn't exist).
vol_offline() accepts the following option in %opts :

=over 4

=item force

Force offlining of volume. Will run with 'priv set -q diag' since it
will only work in that mode.

=item cifsdelaytime

The number of seconds to delay before taking the volume offline, during
which time CIFS uses are warned of the pending loss of service.

=back

=cut

sub vol_offline
{

#    if (my $server = are_we_ontapng($_[0])) {
#	    return $server->volume_offline(volume=>$_[1]);
#	}
	my ($conn, $vname, %opts) = @_;

	my $VSERVER_NAME = $main::VSERVER_NAME;

	if (my $server = are_we_ontapng($_[0])) {
		#return $server->volume_online(volume=>$_[1]);
		eval {
			logStep("##mica vol offlining a 10M/Tricky filer vol ");

			#remove junctions first  
			logStep("##mica unmounting junctions .");
			$server->volume_unmount(vserver=>$VSERVER_NAME,
                			volume=>$vname);

			#make volume online  
			logStep("##mica onlining 10M/Tricky volume .");
			$server->volume_modify(vserver=>$VSERVER_NAME,
                			volume=>$vname, state=>'offline');
		};
		if ($@) {
			exit -1;
		} else {		
			return 1;
		}
	}

	$Impl->validation($Validation);
	return ($Impl->cmd_offline(@_));
}

=head2 vol_container($conn,$vol)

Get the aggregate that the volume is associated with

=cut

sub vol_container
{

	my $conn = make_conn_and_complain(shift);
	my($vol)  = validate_pos(@_, {type => SCALAR});

	return eval {
		my $volexp = quotemeta($vol);
		my $text = sayEOF($conn, "vol container $vol");
		if($text !~
			/Volume '$volexp' is contained in aggregate '(\w+)'/) {
			die "$text\n";
		}
		return $1;
	};
}

=head2 vol_count_inherit($conn,$vol)

Get the inherited blocks of the associated flex volume
This function uses 'priv set test' on the filer.

=cut

sub vol_count_inherit($$;)
{

	my $conn = make_conn_and_complain(shift);
	my($vol) = validate_pos(@_, { type => SCALAR });
	return eval {
		my $cmd = "priv set -q test;vol count_inherited $vol";

		# get inherited blocks
		my $text = sayEOF($conn, $cmd);
		die $@ if ($@ =~ /No volume named|Count the number/i);

		my $cnt;
		if ($text =~ /Inherited count\s+:\s+(\d+)/i) {
			$cnt = $1;
		} else {
			die "Cannot get Inherited Block size: $text\n";
		}
		return $cnt;
	};
}

=head2 vol_restrict($conn,$vol[,%opts])

Change the state of the volume to "restricted".
Returns success (if the volume was changed or is already in the desired state).
or failure otherwise (if the volume couldn't change or didn't exist).
vol_restrict() accepts the following option in %opts :

=over 4

=item force

Force restricting of volume. Will run with 'priv set -q diag' since it
will only work in that mode.

=item cifsdelaytime

The number of seconds to delay before making the volume restricted, during
which time CIFS uses are warned of the pending loss of service.

=back

=cut

sub vol_restrict
{
	my ($conn, $vname, %opts) = @_;

	my $VSERVER_NAME = $main::VSERVER_NAME;
	my $VSERVER_ROOTVOL = $main::VSERVER_ROOTVOL;
	my $VSERVER_ROOTVOLUME_SECURITY_STYLE = $main::VSERVER_ROOTVOLUME_SECURITY_STYLE; 
	my $FILER10M_AGGR = $main::FILER10M_AGGR;
    my $VSERVER_DLIF = $main::VSERVER_DLIF;

	if (my $server = are_we_ontapng($_[0])) {
		#return $server->volume_online(volume=>$_[1]);
		eval {
			logStep("##mica vol restrict a 10M/Tricky filer vol ");

			#remove junctions first  
			logStep("##mica unmounting junctions .");
			$server->volume_unmount(vserver=>$VSERVER_NAME,
                			volume=>$vname);

			#make volume restricted  
			logStep("##mica restricting 10M/Tricky volume .");
			$server->volume_modify(vserver=>$VSERVER_NAME,
                			volume=>$vname, state=>'restricted');
		};
		if ($@) {
			exit -1;
		} else {		
			return 1;
		}
	}

	#if (my $server = are_we_ontapng($_[0])) {
	#    return $server->volume_restrict(volume=>$_[1]);
	#}
	
	$Impl->validation($Validation);
	return ($Impl->cmd_restrict(@_));
}

=head2 vol_rename($conn,$oldvol,$newvol)

Renames the name of volume $oldvol to $newvol. Returns success or failure.

=cut

sub vol_rename
{
	my ($conn, $vname, $newname) = @_;

	my $VSERVER_NAME = $main::VSERVER_NAME;
	my $VSERVER_ROOTVOL = $main::VSERVER_ROOTVOL;
	my $VSERVER_ROOTVOLUME_SECURITY_STYLE = $main::VSERVER_ROOTVOLUME_SECURITY_STYLE; 
	my $FILER10M_AGGR = $main::FILER10M_AGGR;
    my $VSERVER_DLIF = $main::VSERVER_DLIF;

	if (my $server = are_we_ontapng($_[0])) {
		#return $server->volume_online(volume=>$_[1]);
		eval {
			logStep("##mica vol renaming a 10M/Tricky filer vol ");

			#make volume online  
			logStep("##mica renaming 10M/Tricky volume .");
			$server->volume_rename(vserver=>$VSERVER_NAME,
                			volume=>$vname, newname=>$newname);
		};
		if ($@) {
			exit -1;
		} else {		
			return 1;
		}
	}

	$Impl->validation($Validation);
	return ($Impl->cmd_rename(@_));
}

=head2 vol_read_fsid($conn,$vol)

   my $fsid = vol_read_fsid($conn, $vol);

Returns the FSID (in hexadecimal) of a volume.

=cut

sub vol_read_fsid
{

	$Impl->validation($Validation);
	return ($Impl->cmd_read_fsid(@_));
}


=head2 vol_rewrite_fsid($conn,$vol,%args)

  vol_rewrite_fsid($conn, $vol, fsid => $vol_fsid, [force => (0|1)]);

Rewrites the FSID of the desired volume. Returns the *original* FSID of
the volume prior to the FSID rewrite.

Do not force a rewrite of FSID of a volume if you have no idea what
you''re doing. Note: You need to restrict the volume first prior to
calling this API.

vol_rewrite_fsid() accepts the following options in %opts:

=over 4

=item fsid

The desired FSID to be written to the aggregate.

=item force

Optional. Force rewriting of the FSID even though an existing of the
same FSID exists on the system.

=back

=cut

sub vol_rewrite_fsid
{
	$Impl->validation($Validation);
	return ($Impl->cmd_rewrite_fsid(@_));
}


sub vol_join
{
	die "TBD";
}

sub vol_lang
{

	my $conn = make_conn_and_complain(shift);
	my($vol, %args) = validate_mixed(
		@_,
		[ {type => SCALAR, optional => 1, default => undef } ],
		{
			language => { type => SCALAR, optional => 1 }
		}
	);

	if ( $vol eq undef ) {
		# return hash of supported languages on the filer
		my $text = sayEOF($conn, "vol lang");
		my %list;
		foreach my $line (split "\n", $text) {
			if ( $line =~ /\(.+\)/ ) {
				$line =~ s/(\s+)(\S+)\s+/$2=/;
				my ($key, $val) = split "=", $line;
				$list{$key} = $val;
			}
		}
		return \%list;
	}

	return eval {
		vol_exists($conn, $vol) or die $@;

		if($args{language}) {
                        # Some language change requests result in a series of prompts
                        # to verify the changes before proceeding. If $conn is a rsh
                        # connection, the command fails with an error message. If $conn
                        # is a telnet/console connection, the prompts are received and
                        # sayEOF fails. The workaround for either $conn type is to
                        # undefine the language on the volume, then set it to the
                        # desired language. When done this way, there are no prompts.
                        logcomment("Undefining language before setting new language.");
                        my $text = sayEOF($conn, "vol lang $vol 0");
                        die "Could not undefine language settings before making change: $text"
                          unless $text =~ /language information on disk reset/;
			my $text = sayEOF($conn,
					"vol lang $vol $args{language}");
			die "$text\n" unless
				$text =~ /The new language mappings will be available after reboot/;

			if($Validation) {
				my $code = vol_lang($conn, $vol) or die $@;
				die "Unable to set language of volume ".
				    "$vol to $args{language}"
					if($code ne $args{language});
			}
			return 1;
		} else {
			my $text = sayEOF($conn, "vol lang $vol");
			die "Could not parse vol lang $vol: $text\n" unless
			        $text =~ /^\s*Volume\s+language\s+is\s+(?:undefined)?(\S+)\s+\((.*)\)/;
			my($code, $charset) = ($1 , $2);
			return wantarray ? ($code, $charset) : $code;
		}
	}
}


=head2 vol_list($conn)

Returns the list of known volume names, listed from the 'vol status' command

  my @vols = vol_list($conn);

=cut

sub vol_list
{

    if (my $server = are_we_ontapng($_[0])) {
	    return $server->volume_show();
	}
	$Impl->validation($Validation);
	return ($Impl->cmd_list(@_));
}

=head2 vol_match_plex_raid_info($volexp, $string)

Takes a string variable that is the output of vol status command
with all the lines before plex raid info stripped out.
Matches that string with expected strings of plex and
RAID groups for the given volume.  All plexes and their RAID
group information is matched.  Will return an empty string on success.
Otherwise will return the string that did not match.

=cut

sub vol_match_plex_raid_info
{

	$Impl->validation($Validation);
	return ($Impl->cmd_match_plex_raid_info(@_));
}

=head2 vol_status($conn,$vol[,$timeout])

Returns the state, status, host aggregate name, and a hash of volume options
of the given volume.

  my($state, $status, $aggr, %options) = vol_status($conn, $vol);

Note that the volume options that "vol status" command returns is not the
same as the extensive one that is returned by vol_options.

When called in a scalar context, returns an object. Object's accessor methods
have to be called to access volume's status info from the object.

  my $status_obj = vol_status($conn, $vol);
  my $type   = $status_obj->type();
  my $state  = $status_obj->state();
  my $status = $status_obj->status();
  my $aggr   = $status_obj->parent();
  my %options= $status_obj->options();

IMPORTANT: The options presented by vol_status() are grabbed from calling
'vol options', rather than parsing the options column from 'vol status'.
This is no different from calling vol_options() directly.

=cut

sub vol_status
{

        if (my $server = are_we_ontapng($_[0])) {
	    my %vol;
	    if ($server->volume_show(volume=>$_[1], want_hash=>\%vol)) {
		my $status = Aggregate::ImplTemplate::Status->new(cmd=>$Impl->cmd(),
								  validation=>$Impl->validation());
		$status->{type} = "flex";
		$status->{state} = $vol{state};
		if (wantarray()) {
		    die "vol_status against Tricky does not support returning an array yet\n";
		}
		return $status;
	    }
	    return undef;
	}
	$Impl->validation($Validation);
	return $Impl->cmd_status(@_);
}

# XXX - wait for state/status change
=head2 vol_wait_change($conn,$vol[,%opts])

C<vol_wait_change()> will wait for asynchronous processing on a volume that
may change its state/status/type. For example, to wait for a volume to
come up online, you would do:

  use Volume qw(vol_wait_change);
  ...
  my $conn = connect(...);
  my $rv = vol_wait_change($conn, 'vol0', want => { state => qr(^online$) });

The C<want> parameter specifies which field(s) to match.
Its value is a hash reference where each key is a field name returned by
C<vol_status()> (state, status, or aggr) and the corresponding hash value
is a regular expression to match against.

So in order to match the state as either online or offline, you can do:

  vol_wait_change($conn, 'vol0', want => { state => qr{^(?:on|off)line$} });

If you specify multiple keys in the C<want> hash, all of those criteria
must be met before C<vol_wait_change()> can successfully return.

You can also specify a C<timeout> and a C<delay> parameter, which specify
the total time to try getting the desired status, and the time to sleep
in between match attempts, respectively.

=cut

sub vol_wait_change
{

	my $conn = make_conn_and_complain(shift);
	my($vol, %args) = validate_mixed(
		@_,
		[ { type => SCALAR } ],
		{
			timeout => { type => SCALAR, optional => 1,
				default => 900 },
			delay   => { type => SCALAR, optional => 1,
				default => 60 },
			want    => { type => HASHREF, optional => 1,
				default => { state => qr/online/ } },
		}
	);
	
    return eval {
		# for efficiency's sake, create a closure which does
		# the matching.
		my $want = $args{want};
		my @requirements;
		foreach my $field qw(state status raidtype) {
			if($want->{$field}) {
				if(ref($want->{$field}) ne 'Regexp') {
					$want->{$field} = qr($want->{$field});
				}
				push @requirements, sprintf(
				       '$%s =~ /$want->{%s}/g', $field, $field);
			}
		}
	
		my $checker = eval sprintf( <<'EOM', join('&&', @requirements));
			sub {
				my($state, $status, $type) = @_;
				return %s;
			}
EOM
		die if $@;

		my $sleepamt = $args{delay};
		my $endtime  = time() + $args{timeout};
		while(1) {
			last if time() > $endtime;

			my @status = vol_status($conn, $vol)
				or die $@;

			if($checker->(@status)) {
				return 1;
			}
			select(undef,undef,undef,$sleepamt);
		}
		die "Timed out waiting for change in volume '$vol': " .
			join( ",", map{ "$_ => $want->{$_}" } keys %{$want} );
	};
}

sub vol_wait_zero
{

	my($conn, $vol) = validate_pos( @_, { isa => 'Connectrec' }, 1 );
}

=head2 vol_options($conn,$vol[,%opts])

Sets or gets the options of a volume. If the function is called with two
arguments, returns a hash of option keys/value pairs for the specified volume.
In scalar context, returns a reference to the hash.

When called with three arguments, will set the optional parameters for
the volume depending on what the $opts{key} parameter says.
Returns a hash of the new options set on the volume; in scalar context,
returns a reference to the hash.

The syntax of the optional parameters is as follows:

  $volopts = vol_options($conn, $vol, minra=>1);
  $volopts = vol_options($conn, $vol, raidsize=>"10");

  $volopts = vol_options($conn, $vol, root=>'');

=cut

sub vol_options
{

	my $conn = make_conn_and_complain(shift);
	my($vol, %opts) = validate_mixed(@_,
		[{ type => SCALAR}],
		{
			root		=> { type => BOOLEAN, optional => 1,
				regex => qr/^\s*$/ },

			svo_enable 	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			nosnapdir 	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			svo_allow_rman 	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			svo_checksum 	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			minra 		=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			fs_size_fixed 	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			snapmirrored 	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			convert_ucode 	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			nvfail 		=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			create_ucode	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			nosnap		=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			compression	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			encryption	=> { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },                        
			no_atime_update => { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			svo_reject_errors => {type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			ignore_inconsistent => {type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },

			resyncsnaptime	=> { type => SCALAR, optional => 1,
				regex => qr/\d+/ }, # > 0 - valid value
			fractional_reserve => {type => SCALAR, optional => 1,
				regex => qr/\d+/ }, # 0-100 - valid value
			maxdirsize 	=> { type => SCALAR, optional => 1,
				regex => qr/\d+/ }, # 4-??? valid value
			schedsnapname 	=> { type => SCALAR, optional => 1,
				regex => qr/ordinal|create_time/i },
			raidsize 	=> { type => SCALAR, optional => 1,
				regex => qr/\d+/ },  # 2-14 valid value
			# Update cmd_options when a new option which has a valid value of 1/0

			raidtype 	=> { type => SCALAR, optional => 1,
				regex => qr/raid_dp|raid4|raid0/i },
			guarantee 	=> { type => SCALAR, optional => 1 ,
				regex => qr/volume|file|none/i },
			snaplock_default_period	=> {
			    type => SCALAR,
			    optional => 1,
			    regex => qr/\d+[sdmy]|min|max|infinite/
			},
			snaplock_minimum_period	=> {
			    type => SCALAR,
			    optional => 1,
			    regex => qr/\d+[sdmy]|infinite/
			},
			snaplock_maximum_period	=> {
			    type => SCALAR,
			    optional => 1,
			    regex => qr/\d+[sdmy]|infinite/
			},
			no_i2p => { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			acregmax => { type => SCALAR, optional => 1,
				      regex => qr/\d+/ },
			acdirmax => { type => SCALAR, optional => 1,
				      regex => qr/\d+/ },
			acsymmax => { type => SCALAR, optional => 1,
				      regex => qr/\d+/ },
			actimeo => { type => SCALAR, optional => 1,
			             regex => qr/\d+/ },
			flexcache_min_reserve => {type => SCALAR, optional => 1}, # 0-100 - valid value
			snapshot_clone_dependency => {type => SCALAR,
						      optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
			flexcache_autogrow => { type => SCALAR, optional => 1,
				regex => qr/on|off/i },
			nbu_archival_snap => { type => SCALAR, optional => 1,
				regex => qr/0|1|true|false|yes|no|on|off/i },
                        extent => {type => SCALAR, optional => 1,
                                   regex => qr/on|off|space_optimized/i},
                        try_first => {type => SCALAR, optional => 1,
                                      regex => qr/volume_grow|snap_delete/i},
                        read_realloc => {type => SCALAR, optional => 1,
                                         regex => qr/on|off|space_optimized/i},

		}
	);

	$Impl->validation($Validation);
	return $Impl->cmd_options($conn, $vol, %opts);
}

=head2 vol_exists($conn,$vol)

Returns true if a given volume name exists.
(CAVEAT: This function is currently implemented to use vol_options())

=cut

sub vol_exists
{

    if (my $server = are_we_ontapng($_[0])) {
		return $server->volume_exists(volume=>$_[1]);
	}
	$Impl->validation($Validation);
	return ($Impl->cmd_exists(@_));
}


=head2 vol_size($conn,$vol[,$size])

Sets or gets the size of volume. If the function is called with two
arguments, returns the current size of the specified volume expressed
in kilobytes.

  $size_in_kilobytes = vol_size($conn, $vol);

(Note that this is different from previous incarnations of this function
where you could specify the units that the returned size was expressed
in. If you want it in some other units, see Utils::Units::convert_units())

When called with three arguments C<vol_size()> will either set or resize
the volume depending on what the $size parameter says, and will return
the new size of the volume.  The syntax of the size parameter is as follows:

  $size_in_kb = vol_size($conn, $vol, '+10M');  # increase by 10MB
  $size_in_kb = vol_size($conn, $vol, '-100k'); # decrease by 100KB
  $size_in_kb = vol_size($conn, $vol, '1g');   # set to 1GB

If $Validation flag is on when you change the size of volume, C<vol_size()>
will attempt to check if the new reported size is equal to the pre-computed
expected size.

=cut

# The size may be a number optionally followed by a unit specification.
# if no unit spec is given, the number is in bytes.
my $size_re = qr{\d+${Utils::Units::ALLOWED_UNITS_RE}?};

# XXX - private function -- just gets the size in bytes
sub _get_vol_size
{

	my $conn = make_conn_and_complain(shift);
	my($vol) = validate_pos(@_, { type => SCALAR } );
	return eval {
		my $volexp = quotemeta($vol);
		my $status = vol_status($conn, $vol)
			or die $@;
		if ($status->type() eq 'flex') {
			
			my $text = sayEOF($conn, "vol size $vol");
			if($text !~ /vol\s+size:\s+Flexible\s+volume\s+'$volexp'\s+has\s+size\s+($size_re)/) {
				die "$text\n";
			}
			return convert_to_bytes($1);
		} elsif ($status->type() eq 'trad') {
			my $blocks_status = vol_blocks($conn, $vol)
				or die $@;
			return $blocks_status->{block_size} * $blocks_status->{blocks}; # in bytes already
		} else {
			die "unknown volume type ".$status->type();
		}
		
	};
}


sub vol_size
{

	my $conn = make_conn_and_complain(shift);
	my($vol, $sizespec) = validate_pos(@_,
		{ type => SCALAR }, # volume name
		{   # optional size "spec" to change size
			type     => SCALAR,
			optional => 1,
			regex    => qr/^[+-]?$size_re/,
		},
	);
	return eval {
		my $volexp = quotemeta($vol);

		if(!$sizespec) {
			# no size spec given. just get size
			my $bytes = _get_vol_size($conn, $vol) or die $@;
			return convert_units($bytes, UNIT_B, UNIT_KB) or die $@;
		} else {
			my $new_size = 0;

# XXX - In order to get validation working, we
# need to be careful with ontap trying to be
# eh, user-friendly. The vol size command
# actually prints out the new size after a
# disk resize in the units that the user
# specified in. So for example,
#
#   +10m   =>  volX size set to XXXm
#   -100k  =>  volX size set to XXXXXXk
#
# etc. The the numbers are trunctaed so that
# the size is never show in franctional numbers.
# so we must make the comparison in the
# right units!

			my $spec_unit;
			if ($Validation) {			
				# first, get the current size of the volume
				# so that we can pre-compute the expected size,
				# once the re-sizing is done

				my $original_size = _get_vol_size($conn, $vol) or die $@;

				# now use this regexp to figure out how much
				# we're adding or subtracting.
				if( $sizespec !~ /^([+-])?($size_re)$/ ) {
					die "Couldn't parse size argument: ".
						"$sizespec\n";
				}

				my $prefix = $1;
				my $number = $2;

				# any calculations must be done using this unit.
				# otherwise the truncation fails miserably

				$spec_unit = guess_unit($number) or die $@;
				$original_size = convert_units(
					$original_size,
					UNIT_B,
					$spec_unit
				);
				my $size_change_in_unit =
						strip_unit_spec($number);

				if ($prefix eq '+') {
					$new_size = $original_size +
						    $size_change_in_unit;
				} elsif ($prefix eq '-') {
					$new_size = $original_size -
						    $size_change_in_unit;
				} else {
					$new_size = $size_change_in_unit;
				}
				# truncate here, so that comparison works later.
				$new_size = int($new_size);
			}

			# set the size
			my $cmd  = "vol size $vol $sizespec";
			my $text = sayEOF($conn, $cmd);
			my $reported_size = 0;
			if ($text =~ /Flexible\s+volume\s+'$volexp'\s+size\s+(limit\s+|)set\s+to\s+($size_re)/s) {
				$reported_size = $2;
			} else {
				die "$text\n";
			}

			my $unit = guess_unit($reported_size);
			if ($Validation) {
				# verify that we actually got the desired change

				# $reported_size should already be in the same
				# unit as $sizespec, which should be the unit
				# of $new_size.
				# also, don't forget to truncate before
				# comparison...
				my $tmp = int(convert_units(
					strip_unit_spec($reported_size),
					$unit,
					$spec_unit
				));
				if ($tmp != $new_size) {
					die "vol size reported $tmp after ".
					    "resize operation but expected ".
					    "${new_size} (units in $spec_unit)";
				}
			}
			# return in kb
			return convert_units(
				strip_unit_spec($reported_size),
				$unit,
				UNIT_KB
			);
		}
	
	};
}

=head2 vol_blocks($conn,$vol)

Returns a hash of volume block size information (block_size blocks fs_size)

=cut

sub vol_blocks {

	$Impl->validation($Validation);
	return ($Impl->cmd_blocks(@_));
}

=head2 vol_disks($conn,$vol)

Returns an array of disk descriptors for all disks on the volume.  Each
disk descriptor is a hash of the form (type name size)

=cut

sub vol_disks
{

	$Impl->validation($Validation);
	return $Impl->cmd_disks(@_);
}

=head2 vol_root($conn)

Returns the name of the root volume on the given filer.

=cut

sub vol_root
{

	$Impl->validation($Validation);
	return ($Impl->cmd_root(@_));
}

=head2 vol_max_files($conn,$vol[,$count])

Gets or sets the maximum file count for the specified volume.

The third argument is optional. If given, it specifies the
maximum file count to set on the volume, and the function returns
success or failure with an error message in $@ on failure.

If the third argument is omitted, the function returns a tuple of
($max, $used), representing the maximum number and currently used
number of files in the given volume,

=cut

use constant MAXFILES_SET => 'set';
use constant MAXFILES_GET => 'get';

sub vol_max_files
{

	my $conn = make_conn_and_complain(shift);
	my($vol, $count) = validate_pos(@_,
		{ type => SCALAR },
		{ type => SCALAR, optional => 1 }
	);

	return eval {
		my $volexp = quotemeta($vol);
		if (defined $count) {
			# Set maxfiles
			my $text = sayEOF($conn, "maxfiles -f $volexp $count");
			return _vol_max_files_parser($conn, $volexp,
						     $text, MAXFILES_SET);

		} else {
			# Get maxfiles
			my $text = sayEOF($conn, "maxfiles $volexp");
			return _vol_max_files_parser($conn, $volexp,
						     $text, MAXFILES_GET);
		}
	};
}

sub _vol_max_files_parser
{

	  my ($conn, $vol, $text, $mode) = @_;

	  my %maxfiles_report = (
              '7.0' => {
                  MAXFILES_SET => "The new maximum number of files will be rounded to \\d+",
                  MAXFILES_GET => "Volume $vol\: maximum number of files is currently (\\d+) \\((\\d+) used\\)",
              },
	      '6.5' => {
		  MAXFILES_SET => '', # unused
                  MAXFILES_GET => "Volume $vol\: maximum number of files is currently (\\d+) \\((\\d+) used\\)",
              },
          );

	  my $ver = get_wafl_version($conn->hostp());
	  my $version = ($ver > WV_SCRIMSHAW_MAX) ? '7.0' : '6.5';

	  if ($mode eq MAXFILES_SET) {
		  if ($ver > WV_SCRIMSHAW_MAX) {
			  # anchorsteam onwards
			  die "$text\n"
			    if ($text !~ /$maxfiles_report{$version}{MAXFILES_SET}/);
		  } else {
			  # pre-anchorsteam
			  # on success, no output. failure if there's an output
			  die "$text\n" if ($text);
		  }
		  return 1;

	  } elsif ($mode eq MAXFILES_GET) {
		  # same output for all releases
		  if ($text =~ /$maxfiles_report{$version}{MAXFILES_GET}/) {

			  return (wantarray ? ($1, $2) : $1);
		  } else {
			  die "$text\n";
		  }
	  }
}


=head2 vol_scrub($conn,$vol,$mode)

Performs the vol scrub operation, specified by C<$mode>, which can
be one of the following:

  start
  stop
  resume
  suspend
  status

Returns 1 if the operation is successful, *except* for C<status>, which
returns the percentage completed and the last scrub time.

   my($completed, $lastscrub) = vol_scrub($conn, $vol, 'status');

C<completed> is set to 0E0 ("0 but true") for both when the percentage
is at 0 or there is no scrubbing being performed on the volume.
This is to allow the function to return a "0" while allowing a
construct like:

   my($completed, $lastscrub) = vol_scrub($conn, $vol, 'status') or die;

=cut

sub vol_scrub
{

	$Impl->validation($Validation);
	return $Impl->cmd_scrub(@_);
}

=head2 vol_snap_reserve($conn,$vol[,%opt])

Get/Set snap reserve on the volume specified by C<$vol>.

This method accepts the following option in %opt :

    reserve - snap reserve in percentage

This method returns the value of snap reserve of the volume if the operation is successful.

Get the value of snap reserve of a volume:

   my($reserve) = vol_snap_reserve($conn, $vol);

Set the value of snap reserve of a volume to 30 percent:

   my($new_reserve) = vol_snap_reserve($conn, $vol, reserve => 30);

=cut

sub vol_snap_reserve
{
	$Impl->validation($Validation);
	return $Impl->cmd_snap_reserve(@_);
}

=head2 vol_split($conn,$vol,$newvol)

Split a volume

=cut

sub vol_split
{

	$Impl->validation($Validation);
	return ($Impl->cmd_split(@_));

}

sub vol_verify
{
	die "TBD";
}

=head2 vol_unsparse($conn,$path,$name,$size[,timeout=>$timeout])

Creates a file on the specified filer of the specified size
without actually writing any data to the disk.

Example:

    vol_unsparse( $conn, "/vol/vol0/home", "filler", "10g" );

Uses the B<unsparse> command. Currently, there are a few bugs
in unsparse -- the file it creates isn't readable (wafl checksum
errors, and performance is dismal), and if it determines that there
isn't enough room in the volume for the file, it aborts before writing
the last unsparse chunk, leaving up to 4GB still free!

=cut

sub vol_unsparse {

    my $conn = make_conn_and_complain(shift);
    my($path, $name, $size, %args) = validate_mixed( @_,
	[
	    { type => SCALAR },
	    { type => SCALAR },
	    { type => SCALAR },
	],
	{
	    timeout => { type => SCALAR, optional => 1,
			 default => 3600, regex => qr/^\d+$/ }
	},
    );
    # Using a rsh connection for unsparse command
    # makes detecting success much more reliable
    # If $conn is not rsh, then create a new one
    if ($conn->conntype() !~ /rsh/) {
	host $conn->hostp()->hostname();
	$conn = connect("#rsh");
    }
    return eval {
	my $r = sayEOF( $conn, "priv set test; unsparse -v -l $size -d $path $name", TIMEOUT => $args{timeout} );
        #
        # The expected output is a single line that starts with 'unsparse:'
        #
        my @lines = split("\n", $r);
        if (@lines > 1){
            croak $r;
        }
        my $line = shift @lines;
        if ($line !~ /unsparse:/){
             croak $r;
        }
    };
}

=head2 disk_free($conn,$path[,$space_type,$report_type])

Runs the C<df> command on the specified filer.
In scalar context returns the I<total> amount of free disk space
In array context returns the I<total>, I<used>, and I<avail> values.

$report_type can be one of the constants from Utilts::Units
(UNIT_KB UNIT_MB UNIT_GB UNIT_TB)

$space_type can be one of the following constants, which are exported
with the :df export tag:

  DF_SPACE
  DF_INODES
  DF_PVT_INODES
  DF_RESERVE

C<$report_type> and C<$space_type> are both optional. If unspecified,
C<DF_SPACE> and C<DF_KB> will be used, respectively.

These constants can be exported explicitly or by using the :df tag:

  use Volume qw(:all);
  use Volume qw(:df);
  use Volume qw(DF_KB DF_SPACE)

  disk_free($conn, $path, DF_SPACE, DF_KB);

=cut

# Note: obsoleted by vol_disk_free()
#
sub disk_free
{

	my ($conn, $path, $space_type, $report_type) = validate_pos(@_,
		{ isa => 'Connectrec' },
		{ type => SCALAR },
		{ type => SCALAR, optional => 1, default => DF_SPACE },
		{ type => SCALAR, optional => 1, default => UNIT_KB }
	);
	

	$Impl->validation($Validation);
	return ($Impl->cmd_disk_free($conn,
				     vol=>$path,
				     space_type=>$space_type,
				     report_type=>$report_type));
}

=head2 vol_spare_status($conn)

Given a connection to a filer, fetches the output of 'vol status -s' and
returns a hash containing lists of Volume::DiskStatus objects.

The list of disks are keyed by the "pool" that they belong in. If there
are no pools defined, everything will go into slot "0"

  my $conn   = connect(...);
  my $hash = vol_spare_status($conn);
  foreach my $disk (@{$hash->{0}}) {
    print $disk->name, "\n";
  }


In scalar context, the reference to the resulting hash is returned.
In array context, the hash values are returned directly.

  my $hash = vol_spare_status($conn);
  my %hash = vol_spare_status($conn);

=cut

sub vol_spare_status
{

	$Impl->validation($Validation);
	return $Impl->cmd_spare_status(@_);
}

=head2 vol_raidgroup_status($conn,$vol)

Given a connection to a filer and a target volume, parses the output of
C<vol status -r $vol> and returns a hash.

  NOTE: If this function is going to be used in more than one script,
  the returned structure needs to be changed a bit more...

This has is a 4-level nested hash. The top level contains some information
about the volume. The second level can be accessed via $hash->{plex},
which is an array of plexes. The third level has the raidgroup information,
and can be accessed via $plex->{raidgroups}. The last level a list of
Volume::DiskStatus objects, and can be accessed via $raidgroup->{disks}

  need to work on this doc...

=cut

sub vol_raidgroup_status
{
	$Impl->validation($Validation);
	return $Impl->cmd_raidgroup_status(@_);
}

=head2 vol_disk_free($conn, %options)

Runs df command on the specified volume/path. Below is the description of the
parameters of this subroutine.

=over 4

=item $conn

This is the connection to filer host on which C<df> command is to be executed.

=item %options

This is a hash with following optional keys.

path
	This is the name of the volume or path.

space_type
	The types of statistics to be reported. There are three types:
	DF_SPACE, DF_INODES, DF_PVT_INODES and DF_RESERVE. Default is DF_SPACE.

report_type
	Units in which the above type of statistics are reported.
	Currently, there are following units: UNIT_KB, UNIT_MB, UNIT_GB and UNIT_TB.
	The default is UNIT_KB.

snapshot 
	If the value of this field is non-zero, the statistics are
	calculated for the snapshot of the requested volume/path.  The default is 0.

sync
sync_delay
	'sync' option is either 0 or 1. If 0, no delay or sync is performed
	If 1, then sync and delay are done, and either 'sync_delay' option or
	$::DF_SYNC_DELAY global param (see "commonParams.pl") is used to
	determine the length of the delay.

	A 'df' performed immediately after a file operation (e.g. 'truncate')
	might not reflect the file operation in its results. The delay is
	intended to allow all previous file operations to "register" before
	performing the df.

	Default value for 'sync' is now 1. See BURT 306218

	Note that there is no way to "guarantee" a delay sufficient to capture
	all previous changes. A greater delay is just more likely to suffice than
	a smaller delay (i.e. delay values are "arbitrary")

no_poll
	Optional. Default is '0'. If '1', then no polling of WAFL counters or other
	values is done before the 'df' operation. However, the delay specified by
	'sync'/'sync_delay' is still done before the 'df' operation.

=back

This subroutine returns a hash reference in scalar context and an array in
array context.

  use Volume qw(:df)
  use Utils::Units qw(:constants)

  my ($vol_df) = vol_disk_free($conn, path => "vol0",
			space_type => DF_SPACE, report_type => UNIT_KB)

  my ($total, $used, $avail, $capacity) = vol_disk_free($conn,
		path => "/vol/vol0/dir", space_type => DF_SPACE,
		report_type => UNIT_KB)

  my ($snap_df) = vol_disk_free($conn, path => "/vol/vol1",
		snapshot => 1)

Elements of arrary and values of hash depends on C<space_type> parameter.

=over 4

=item DF_SPACE

Scalar context - The returned hash contains C<total space>, C<used space>,
C<available space> and C<capacity (percent)> keyed by C<total>, C<used>,
C<avail> and C<capacity> respectively.

Array context - The returned array contains C<total space>, C<used space>,
C<available space> and C<capacity (percent)> in the specified order.

=item DF_RESERVE

Scalar context - The returned hash contains C<total space>, C<used space>,
C<available space> and C<reserved space> keyed by C<total>, C<used>,
C<avail> and C<reserved> respectively.

Array context - The returned array contains C<total space>, C<used space>,
C<available space> and C<reserved space> in the specified order.

=item DF_INODES

Scalar context - The returned hash contains C<used nodes>, C<free inodes>
and C<used percentage> keyed by C<iused>, C<ifree> and C<percent_iused>
respectively.

Array context - The returned array contains C<used nodes>, C<free inodes>
and C<used percentage> in the specified order.

With 'dual inode' feature, returns values for 'public' inodes, while
'DF_PVT_INODES' returns values for 'private' inodes.

=item DF_PVT_INODES

Only valid with 'dual inode' feature. Similar to 'DF_INODES', but returns
values for 'private' inodes, whereas 'DF_INODES' returns values for
'public' inodes.

=back

=cut

sub vol_disk_free
{
	$Impl->validation($Validation);
	my ($conn, %args) = @_;
	return $Impl->cmd_disk_free($conn, %args, type=>'vol');
}


=head2 make_disk_hot($conn, $vol, $chosen_disk, $interval)

This fucntion makes the given disk hot from the given volume for the
specified amount of time

=over 4

=back

=cut

sub make_disk_hot($$$$)
{

    my ($conn, $vol, $chosen_disk, $interval) = @_;
    my $output;

    $output = sayEOF($conn, "priv set -q test; ".
                            "vol disk hot $vol $chosen_disk $interval;");
    if ($output !~ "vol disk hot succeeded for $vol/$chosen_disk")
    {
        croak ("vol disk hot failed: $output");
    }
    return 1;
}

=head2 disk_blocks_written($conn, $vol_name, $chosen_disk)

This function returns a hash array filled with disk information, information
about number of block written and queued on the disk. Also whether the disk
is hot or not and how many other disks are hot.

=over 4

=back

=cut

sub disk_blocks_written($$$)
{

    my ($conn, $vol_name, $chosen_disk) = @_;
    my $output;
    my @word;
    my ($disk, $vol, $queued, $writing, $hot, $otherhot);


    $output = sayEOF($conn, "priv set -q test; ".
                             "vol disk writing $vol_name $chosen_disk");
    @word = split(" ", $output);
    if ($output =~ "was not found")
    {
        croak( "expected 0 blocks, got: $output");
    }

    $disk = $word[1];
    $vol = $word[4];
    $queued = $word[6];
    $writing = $word[9];
    if ($word[18] eq "hot,")
    {
        $hot = 1;
        $otherhot = $word[19];
    }
   else
    {
        $hot = 0;
        $otherhot = $word[20];
    }
    return { disk => $disk,
             vol => $vol,
             queued => $queued,
             written => $writing,
             hot => $hot,
             otherhot => $otherhot
           };
}

=head2 is_disk_hot($conn, $vol, $chosen_disk)

Checks for disk being hot or not
Returns boolean, 0 if not hot, 1 if hot

=over 4

=back

=cut

sub is_disk_hot($$$)
{

    my ($conn, $vol, $chosen_disk) = @_;
    my $output;

    $output = sayEOF($conn, "priv set -q test; ".
                            "vol disk hot $vol $chosen_disk 0;");
    return 0 if ($output =~ "is not hot") ;
    return 1 if ($output =~ "is hot") ;
    croak( "vol disk hot failed: $output");
}


=head2 vol_determine_flex_root_size($conn)

Calculate/determine the minimum allowed (or recommended)
flex root volume size.
Returns size of flex root vol. (E.g 25m or 15g)

    # ONTAP Code
    # /* return recommend flexible root volume size in bytes */
    # uint8
    # vol_reco_flex_root_size(void)
    # {
    # #ifdef simulator
    #     return 25 << 20; /* 25 MB */
    # #else
    #	if (sk_kmod_simulator)
    #		return 25 << 20; /* 25 MB just like regular sim. */
    #	/*
    #	 * recommended root size = 1.2(6GB + no.of.cores * coresize)
    #	 * size of core is approximated to 0.5 (RAM + NVRAM)
    #	 */
    #	uint8 gig = 1024 * 1024 * 1024;
    #	uint8 sz = 6 * gig + 3 * (get_memsize() + nv_size) / 2;
    #	sz = sz + sz/5;
    #	return roundup(sz, gig);
    # #endif
    # }

    # If System Board is not found by all slots,
    # try explicitely with 0. When multiple devices exist in
    # slot 0, System Board string is over written by later strings

    # slot 0: System Board 733 MHz (NetApp System Board VII A1)
    #        Model Name:         F825
    #        Part Number:        110-00013
    #        Revision:           A1
    #        Serial Number:      273317
    #        Firmware release:   4.3.1_i1
    #        Processors:         1
    #        Memory Size:        1024 MB
    #        CIOB Revision ID:   5
    # slot 0: FastEnet-10/100 Ethernet Controller
    #        Device Type:        82559
        #        e0 MAC Address:     00:a0:98:00:e9:53 (auto-100tx-fd-up)
    # slot 0: NetApp ATA/IDE Adapter 0a (0x000001f0)
    #        0a.0                 SanDisk  SDCFB-256        1.24 245MB 512B/sec
    #        t (103314C2503F0653)

=over 4

=back

=cut

sub vol_determine_flex_root_size
{

    my ($conn) = @_;
    my $fh = $conn->{hostp};
    my $size; # Return

    my ($nvram_size, $memory_size);

    if($fh->{hosttype} =~ /filer-simulator/) {
	$size = "25m";
    } else {
	my $obj = Sysconfig->new($fh->{id});
	my @all_slots = $obj->all_slots();

	foreach my $slot (@all_slots) {
	    my $type = $obj->slot($slot, "type");
	    my $info = $obj->slot($slot, "info");
	    print "Slot=$slot, Type = $type, Info = $info\n";
	    if($type =~ /System Board/ || $info =~ /System Board/) {
		$memory_size = $obj->slot($slot, "memory_size");
	    } elsif($type =~ /NVRAM/ || $info =~ /NVRAM/) {
		$nvram_size = $obj->slot($slot, "memory_size");
	    }
	}
	
	$memory_size = $obj->slot(0, "memory_size")
	    if(!$memory_size);

	logDbg("System Board Memory Size = $memory_size, NVRAM = $nvram_size");

	# Strip MB from the string
	$memory_size =~ s/ MB//g;
	$nvram_size =~ s/ MB//g;
	
	# If the size is in GB, convert to MB.
	if($memory_size =~ / GB/) {
	    $memory_size =~ s/ GB//;
	    $memory_size = convert_units($memory_size, UNIT_GB, UNIT_MB);
	}
	
	if($nvram_size =~ / GB/) {
	    $nvram_size =~ s/ GB//;
	    $nvram_size = convert_units($nvram_size, UNIT_GB, UNIT_MB);
	}
	
	my $core_size = ($memory_size + $nvram_size) / 2;
	$size = (6 * 1024) + (3 * $core_size);
	
	logDbg("Rec Flex root vol size in MB : $size");
	
	$size = convert_units($size, UNIT_MB, UNIT_GB);
	# Round up the size
	$size = (abs($size) > int($size)) ? int($size) + 1 : int($size);
	$size = $size."g";
    }
    return $size;
}


=head2 Volume::vol_vvlabel(fc=><filerconn>,volume=><volname>)

Volume::vol_vvlabel() will return the raid label in a hash when only
the required parameters are specified. If a fieldname is specified
then that field will be modified in the raid label. The caller
should restrict the volume if required before modifying the raid
label field. This routine only works with flex volumes.
This function uses 'priv set test' on the filer.

vol_vvlabel() accepts the following options in %opts:

=over

=item Arguments

=over

=item fc

Required. Connection to filer.

=item volume

Required. Flex volume name.

=item field

Optional. Name of the raid label field to modify.

=item value

Optional. New value for the raid label field.

=back

=item Returns

=over

The function returns a hash of the volume options. Croaks on error.

=back

=back

=for html <hr />

=cut

sub vol_vvlabel(%)
{

    my %opts = validate(
        @_,
        {
            fc     => { type => OBJECT },
            field  => { type => SCALAR, optional => 1, depends => [ 'value' ] },
            value  => { type => SCALAR, optional => 1, depends => [ 'field' ] },
            volume => { type => SCALAR },
        }
    );
    my ($cmd, @info, $line, $out, %retval);
    my $cmd_prefix = 'priv set -q test; vol vvlabel';

    # If field is set then set the field in the raid label
    if (defined($opts{field})) {
        # Setting a field requires a value
        if (!defined($opts{value})) {
            die("'field' options requires 'value' option");
        }
        $cmd = "$cmd_prefix $opts{volume} $opts{field} $opts{value}";
        $out = sayEOF($opts{fc}, $cmd);
        if ($out =~ /(Volume vvlabel:|No such command)/i) {
            die("Modifying raid label failed: $out");
        }
    }

    # Get the raid label info
    $cmd = "$cmd_prefix $opts{volume}";
    $out = sayEOF($opts{fc}, $cmd);
    if ($out =~ /(Volume vvlabel:|No such command)/i) {
        die("Reading raid label failed: $out");
    }

    # Put the raid label info into a hash
    @info = split(/\r*\n/, $out);
    foreach $line (@info) {
        # Remove white space at the beginning and end of the line
        $line =~ s/^\s*//;
        $line =~ s/\s*$//;
        # If there is still characters on the line then process it
        if ($line) {
            $line =~ /(.*) (.*)/;
            $retval{"$1"} = $2;
        }
    }

    return %retval;
}


=head2 vol_meta_i2p($conn, $aggr)

Returns the output of the volume's meta_i2p command as a hashref,
with fileids as keys, and filenames as values.

  my $i2p = vol_meta_i2p($conn, $aggr);

=cut

sub vol_meta_i2p {

	$Impl->validation($Validation);
	return ($Impl->cmd_meta_i2p(@_));
}


use Volume::Playlist;

=head2 vol_playlist_copy_in($conn, $vol, $path)

Overwrites the volume's playlist file with the one stored at the specified
path.

The following parameters are all required:

=over 4

=item $conn

The connection to filer host on which the volume resides.

=item $vol

The name of the volume whose playlist file is to be overwritten.

=item $path

The pathname of the playlist file to copy in.

=back

=cut

sub vol_playlist_copy_in {

	my $conn = shift;
	my $vol  = shift;
	my $path = shift;
	Volume::Playlist::playlist_copy_in(conn => $conn, vol => $vol, path => $path);
}


=head2 vol_playlist_copy_out($conn, $vol, $path)

Copies the volume's playlist file to the specified path.

The following parameters are all required:

=over 4

=item $conn

The connection to filer host on which the volume resides.

=item $vol

The name of the volume whose playlist file is to be copied out

=item $path

The pathname to copy the playlist file to.

=back

=cut

sub vol_playlist_copy_out {

	my $conn = shift;
	my $vol  = shift;
	my $path = shift;
	Volume::Playlist::playlist_copy_out(conn => $conn, vol => $vol, path => $path);
}


=head2 vol_playlist_dump($conn, $vol)

Returns a dump of the volume's playlist file contents as a string.

The following parameters are all required:

=over 4

=item $conn

The connection to filer host on which the volume resides.

=item $vol

The name of the volume whose playlist file is to be dumped.

=back

=cut

sub vol_playlist_dump {

	my $conn = shift;
	my $vol  = shift;
	return Volume::Playlist::playlist_dump(conn => $conn, vol => $vol);
}


=head2 vol_playlist_status($conn, $vol)

Returns the volume's playlist file status as a hashref.

The following parameters are all required:

=over 4

=item $conn

The connection to filer host on which the volume resides.

=item $vol

The name of the volume whose playlist file is to be dumped.

=back

=cut

sub vol_playlist_status {

	my $conn = shift;
	my $vol  = shift;
	return Volume::Playlist::playlist_status(conn => $conn, vol => $vol);
}


=head2 Volume::volume_set_option(host=><hostspec>,volume=><volname>,option=><option>,value=>,value>)

Volume::volume_set_option() is the ZAPI set interface to the CLI 'vol
options' as implemented by Volume::vol_options().

=cut

sub volume_set_option(%)
{

    my (%args) = @_;

    my $host = $args{host};
    # prepare the ZAPI
    my $zcmd = new NaElement("volume-set-option");
    $zcmd->child_add(new NaElement("volume","$args{volume}"));
    $zcmd->child_add(new NaElement("option-name","$args{option}"));
    $zcmd->child_add(new NaElement("option-value","$args{value}"));

    my $zapi = Zapi->new($host->default_ip(),
			 "root",
			 $host->default_password());
    $zapi->{fail} = \&_fail_default;

    my $xo = $zapi->invoke_elem($zcmd, "any");
    #logDbg("vol-set-option ZAPI returns: ". $xo->toString());
    if ($Validation) {
        my %opts = volume_options_list_info(host => $host,
                                            volume => $args{volume});
        if ($opts{$args{option}} ne $args{value}) {
            die("ZAPI volume_set_option($args{option}) returned $opts{$args{option}} instead of $args{value}");
        }
    }
    return 1;
}

=head2 Volume::volume_options_list_info(host=><hostspec>,volume=><volname>)

Volume::volume_options_list_info() is the ZAPI get interface to the
CLI 'vol options' as implemented by Volume::vol_options().

=over 4

The function returns a hash of the volume options.

=back

=cut

sub volume_options_list_info(%)
{

    my (%args) = @_;

    my $host = $args{host};
    # prepare the ZAPI
    my $zcmd = new NaElement("volume-options-list-info");
    $zcmd->child_add(new NaElement("volume","$args{volume}"));

    my $zapi = Zapi->new($host->default_ip(),
			 "root",
			 $host->default_password());
    $zapi->{fail} = \&_fail_default;

    my $xo = $zapi->invoke_elem($zcmd, "any");
    #logDbg("volume-options-list-info ZAPI returns: ". $xo->toString());

    my %retval;
    if ($xo->results_ok()) {
        my $vo = $xo->child_get("options");
        if ($vo) {
            my $num_elems = $vo->child_max_number();
            #logDbg("child_get(options) returned $num_elems elements");
            for (my $i=0; $i < $num_elems; $i++) {
                my $vl = $vo->child_get_nth($i);
                my $vln = $vl->child_get("name")->{content};
                my $vlv = $vl->child_get("value")->{content};
                $retval{$vln} = $vlv;
            }
        }
    } else {
        die("volume-options-list-info ZAPI returns ". $xo->toString());
    }
    return %retval;
}



# !!!!!!!!!!!!!!!!!!!   Start Change   !!!!!!!!!!!!!!!!!!!!!!!!!

=head2 volume_special_options_hook(filer=><filerspec>, volume=><volname> );
=head2 volume_special_options_hook(filer=><filerspec>, volume=><volname>, license=>[0|1], actions=>HASHREF );
=head2 volume_special_options_hook(filer=><filerspec>, volume=><volname>, license=>[0|1], disable_all_except=>HASHREF );

=over 4

volume_special_options_hook() uses global parameters and function arguments
volume options/modes such as compression and A_SIS. If either the "actions" or 
"disable_all_except" argument is given, then the global parameters are ignored,
and the supplied hash determines any actions performed.

This function is intended to be called from the various volume and volume clone
creation/setup library routines currently used by WAFL QA scripts. The intention is
to not modify existing scripts at all except to add the following line to the script's
BEGIN block if not already there:

      require commonParams.pl;

... and to only add a single line to the affected library functions:

	volume_special_options_hook(filer=><filerspec>, volume=><volname>, license=>[0|1], actions=>HASHREF );
	- OR -
	volume_special_options_hook(filer=><filerspec>, volume=><volname>, license=>[0|1], disable_all_except=>HASHREF );

The "filer" and "volume" arguments are required. All other arguments ("actions", "license") are optional.

C<$HOSTSPEC> can be any one of: HOSTSPEC string, HOSTREC object, or CONNECTREC object.

=item do_license
The "do_license" argument determines whether or not licensing is setup by this function:
	1 => licensing is performed, 0 => licensing is not performed.
The default value of do_license is "1".

=item actions
=item disable_all_except

The "actions" and "disable_all_except" arguments are mutually exclusive. For the "actions", argument,
an action will be performed on a given feature/option ONLY if there is an entry in the "actions" hashtable
associated with that feature/option. A value of "0" for the key means the feature is disabled, a "1" value
means the feature is enabled.

For "disable_all_except" argument, every feature handled by this function will be affected. A given feature
will be disabled UNLESS the hashtable value for the key associated with the feature is "1".

For example (assume valid features are "compression" and "a_sis"

	"actions" arg.			"disable_all_except" arg.	Actions performed
	---------------------------------------------------------------------------------------------------------
	{} (empty hashtable)						No actions performed

					{} (empty hashtable)		compression disabled, A_SIS disabled.

	{ a_sis=>1 }							A_SIS is enabled

					{ a_sis=>1 }			A_SIS is enabled, compression disabled.

	{ a_sis=>1, compression=>0 } 					A_SIS is enabled, compression disabled

					{ a_sis=>1, compression=>0 }	A_SIS is enabled, compression disabled.


If either the "actions" or "disable_all_except" argument is given, then all
global parameter values are ignored.

If new features are to be handled by this function, then both this function and
the file lib/FS/Util/commonParams.pl must be modified to handle the new feature.
will need to be updated to handle these new options. Currently the options handled,
and the associated commonParams.pl parameters are:

	Feature		Global Parameter	"actions"/"disable_all_except"		Comments
	-----------------------------------------------------------------------------------------
	Compression	$COMPRESSION_ENABLE	compression
	A_SIS		$A_SIS_ENABLE		a_sis


Global parameters are used only if no "actions" or "disable_all_except" argument is specified.
For all parameters, "0" => disable feature, and "1" => enable feature.

The "actions" and "do_license" arguments are primarily to accomodate a particular existing use
of this function. For most situations, this form of calling the function is appropriate:

	volume_special_options_hook(filer=>FILER, volume=>VOLUME);

.... with no "actions", "disable_all_except", or "do_license" arguments givien.
To disable all special features on a volume (e.g. for a "cleanup" routine), the
following form may be used:

	volume_special_options_hook(filer=>FILER, volume=>VOLUME, disable_all_except=>{});

The volume must exist and be online before calling this function -- it is called
after the volume is created.

The function "croaks" if an error occurs.

=back

=cut

use Dense qw(dense_toggle dense_config dense_start dense_wait_for_status);
use Admin::License qw(license is_licensed);
use compression qw(enable_compression_option disable_compression_option);
use Params::Validate qw (UNDEF);
use Utils qw(make_conn);

sub volume_special_options_hook
{

my %raw_opts = (@_);
$raw_opts{filer} = ref($raw_opts{filer}) if (ref($raw_opts{filer}));
logDbg("Enter volume_special_options_hook  Line: ".__LINE__."   Opts:\n\n".Dumper(\%raw_opts));
	my(%args) = validate_mixed( @_,
		[],
		{
			filer			=> { type => SCALAR|OBJECT },
			volume			=> { type => SCALAR },
			do_license		=> { type => SCALAR|UNDEF, optional=>1, default=>1 },
			actions			=> { type => HASHREF|UNDEF, optional=>1, default=>undef },
			disable_all_except	=> { type => HASHREF|UNDEF, optional=>1, default=>undef },
		}
	);
	die($@) if ($@);

	my $do_license;
	if (defined($args{do_license})) {
		$do_license = $args{do_license};
	} else {
		$do_license = 1;
	}

	my $filer = $args{filer};
	my $fc = make_conn($filer);
	my $fh = $fc->hostp();
	my $filer_name = $fh->hostname();
	my $vol = $args{volume};
	my $vol_path = "/vol/$vol";
	my $actions = $args{actions};
	my $disable_all_except = $args{disable_all_except};
	logDbg("volume_special_options_hook\n\nFiler:	\"$filer_name\"\nVolume:	".
		   "\"$vol\"\ndo_license: \"$do_license\"");

	if (defined($actions)) {
		logDbg("volume_special_options_hook  'actions' argument given:\n\n".
			 Dumper($actions));
	} elsif (defined($disable_all_except)) {
		logDbg("volume_special_options_hook  'disable_all_except' ".
			 "argument given:\n\n".Dumper($disable_all_except));
		while ( my ($key,$val) = each(%$disable_all_except) ) {
			if (defined($val)) {
				$actions->{$key} = $val;
			} else {
				$actions->{$key} = 0;
			}
		}
	} else {
		logDbg("volume_special_options_hook  No 'actions' or 'disable_all_except'\n".
			"arguments given, will set 'actions' hash from parameters:\n\n".
			"COMPRESSION_ENABLE: \"".(defined($::COMPRESSION_ENABLE)?$::COMPRESSION_ENABLE:"UNDEF")."\"\n".
			"ENCRYPTION_ENABLE: \"".(defined($::ENCRYPTION_ENABLE)?$::ENCRYPTION_ENABLE:"UNDEF")."\"\n".		   
			"ASIS_ENABLE:	\"".(defined($::ASIS_ENABLE)?$::ASIS_ENABLE:"UNDEF")."\"\n"
					);
		$actions = {
				'compression'	=> $::COMPRESSION_ENABLE,
				'encryption'	=> $::ENCRYPTION_ENABLE,						
				'a_sis'		=> $::ASIS_ENABLE,
				'asis'		=> $::ASIS_ENABLE,
			};
		
		while ( my ($key,$val) = each(%$actions) ) {
				if (!defined($val)) {
					delete($actions->{$key});
				} elsif ($val =~ /no/i) {
					delete($actions->{$key});
				} elsif ($val =~ /1|on|true/i) {
					$actions->{$key} = 1;
				} elsif ($val =~ /0|off|false/i) {
					$actions->{$key} = 0;
				} else {
					delete($actions->{$key});
				}
		}
	}

	my $asis_action = undef;
	if (defined($actions->{a_sis})) {
		$asis_action = $actions->{a_sis};
	} elsif (defined($actions->{asis})) {
		$asis_action = $actions->{asis};
	}
	if (defined($asis_action)) {
		$actions->{a_sis} = $asis_action;
		$actions->{asis} = $asis_action;
	}

	if (!vol_exists($fc, $vol)) {
		croak("Volume $vol does not exist");
	}
	if ($actions->{skip_special_options_hook}) {
		logDbg("volume_special_options_hook  --  'skip_special_options_hook' is set, exiting");
		return;
	}
	logDbg("volume_special_options_hook   Will do 'actions':\n\n".Dumper($actions));
	
	###############################################################################
	#
	#	Start 'Compression' specific section
	#
	# turn compression 'on' / 'off' if asked to
	if (!defined($actions->{compression})) {
		logTrc("volume_special_options_hook  No action given for 'compression' feature ");
	} else {
		logTrc("volume_special_options_hook  Calling  _vol_set_compression");
		_vol_set_compression(filer=>$fc, volume=>$vol, toggle=>$actions->{compression}, do_license=>$do_license);
		die($@) if ($@);
		logTrc("volume_special_options_hook  Returned from  _vol_set_compression");
	}
	#
	#	End 'Compression' specific section
	#
	###############################################################################

	###############################################################################
	#
	#	Start 'Encryption' specific section
	#
	# turn encryption 'on' / 'off' if asked to
	if (!defined($actions->{encryption})) {
		logTrc("volume_special_options_hook  No action given for 'encryption' feature ");
	} else {
		logTrc("volume_special_options_hook  Calling  _vol_set_encryption");
		_vol_set_encryption(filer=>$fc, volume=>$vol, toggle=>$actions->{encryption}, do_license=>$do_license);
		die($@) if ($@);
		logTrc("volume_special_options_hook  Returned from  _vol_set_encryption");
	}
	#
	#	End 'Encryption' specific section
	#
	###############################################################################

	###############################################################################
	#
	#	Start 'A_SIS' specific section
	#
	# turn a_sis 'on' / 'off' if asked to
	my $new_asis_state = undef;
	# will only enable a_sis - setting up schedules, starting it etc is not done in this routine
	if (!defined($actions->{asis})) {
		logTrc("volume_special_options_hook  No action given for 'a_sis' feature ");
	} else {
		logTrc("volume_special_options_hook  Calling  _vol_set_asis");
		_vol_set_asis(filer=>$fc, volume=>$vol, toggle=>$actions->{asis}, do_license=>$do_license);
		die($@) if ($@);
		logTrc("volume_special_options_hook  Returned from  _vol_set_asis");
	}
	#
	#	End 'ASIS' specific section
	#
	###############################################################################

	logTrc("Exit  volume_special_options_hook");
}


=head2 set_volume_special_options(filer=><filerspec>, volume=><volname>, %args );

Wrapper function for 'volume_special_options_hook', instead of 'actions' hash, options
'compression', 'encryption', 'asis', etc. are used as keyword options directly

Examples:

    set_volume_special_options(filer=>$filer, volume=>$vol, compression=>1);
    set_volume_special_options(filer=>$filer, volume=>$vol, compression=>1, asis=>0, do_license=>1);

=over 4

=item compression
Optional. Can be used to create a volume with/without compression
if set, overrides COMPRESSION_ENABLE global var.

=item encryption
Optional. Can be used to create a volume with/without encryption
if set, overrides ENCRYPTION_ENABLE global var.

=item asis (or a_sis)
Optional. Can be used to create a volume with/without A_SIS
if set, overrides ASIS_ENABLE global var.

=item do_license
Optional. If '1', then licenses are added for specified features.
Licenses are not added if '0'. Default is '1'.

=item skip_volume_special_options_hook
Optional. Default is zero. If '1', then no call
is made to 'volume_special_options_hook'. In this case,
all 'volume_special_options_hook' related options such
as compression, encription, A_SIS, etc. will be ignored.

sub set_volume_special_options
{
my %raw_opts = (@_);
$raw_opts{filer} = ref($raw_opts{filer}) if (ref($raw_opts{filer}));
logDbg("Enter set_volume_special_options  Line: ".__LINE__."   Opts:\n\n".Dumper(\%raw_opts));

	my(%args) = validate_mixed( @_,
		[],
		{
			filer		=> { type => SCALAR|OBJECT },
			volume		=> { type => SCALAR },
			skip_special_options_hook => { type => SCALAR|UNDEF, optional=>1, default=>0 },
			do_license	=> { type => SCALAR|UNDEF, optional=>1, default=>1 },
			compression	=> { type => SCALAR|UNDEF, optional=>1, default=>undef },
			encryption	=> { type => SCALAR|UNDEF, optional=>1, default=>undef },
			asis		=> { type => SCALAR|UNDEF, optional=>1, default=>undef },
			a_sis		=> { type => SCALAR|UNDEF, optional=>1, default=>undef },
		}
	);
	die($@) if ($@);

	if ($args{skip_special_options_hook}) {
		return;
	}

	my $asis_action = undef;
	if (defined($args{a_sis})) {
		$asis_action = $args{a_sis};
	} elsif (defined($args{asis})) {
		$asis_action = $args{asis};
	}
	if (defined($asis_action)) {
		$args{a_sis} = $asis_action;
		$args{asis} = $asis_action;
	}
	my %actions_hash = ();
	my @special_volume_options = qw(compression encryption asis a_sis);
	foreach my $opt (@special_volume_options) {
		if (undefined($args{$opt})) {
			next;
		} elsif ($args{$opt} =~ /on|enable|1|start|true/i) {
			$actions_hash{$opt} = 1;
		} elsif ($args{toggle} =~ /off|disable|0||stop|false|^$/i) {
			$actions_hash{$opt} = 0;
		} else {
			die("set_volume_special_options:  Unexpected value for \"$opt\" option: \"$args{$opt}\"");
		}
	}

	return( volume_special_options_hook(filer=>$conn,
						volume=>$vol,
						actions => \%actions_hash) );
}



=head2 _vol_set_asis($conn,$vol[,%opts])

Turn ASIS on/off for volume.

=cut

sub _vol_set_asis
{


my %raw_opts = (@_);
$raw_opts{filer} = ref($raw_opts{filer}) if (ref($raw_opts{filer}));
logDbg("Enter _vol_set_asis  Line: ".__LINE__."   Opts:\n\n".Dumper(\%raw_opts));
my (%args);
	my(%args) = validate_mixed( @_,
		[],
		{
			filer		=> { type => SCALAR|OBJECT },
			volume		=> { type => SCALAR },
			do_license	=> { type => SCALAR|UNDEF, optional=>1, default=>1 },
			check_license	=> { type => SCALAR|UNDEF, optional=>1, default=>1 },
			toggle		=> { type => SCALAR},
			dense_wait_timeout => { type => SCALAR|UNDEF, optional=>1, default=>1800 },
		}
	);

	die($@) if ($@);

	my $check_license = $args{check_license};
	my $do_license;
	if (defined($args{do_license})) {
		$do_license = $args{do_license};
	} else {
		$do_license = 1;
	}

	my $filer = $args{filer};
	my $fc = make_conn($filer);
	my $fh = $fc->hostp();
	my $filer_name = $fh->hostname();
	my $vol = $args{volume};
	my $vol_path = "/vol/$vol";
	my $actions = $args{actions};
	my $dense_wait_timeout = $args{dense_wait_timeout};

	my $toggle;
	if (!defined($args{toggle})) {
		  logWarn("_vol_set_asis  'toggle' option not specified");
		  die("_vol_set_asis  'toggle' option not specified");
	} elsif ($args{toggle} =~ /on|enable|1|start|true/i) {
		$toggle = 1;
	} elsif ($args{toggle} =~ /off|disable|0||stop|false|^$/i) {
		$toggle = 0;
	 } else {
		  die("_vol_set_asis  Unexpected value for 'toggle' option: \"$args{toggle}\"");
	}

	if ($toggle) {
		# will only enable a_sis - setting up schedules, starting it etc is not done in this routine
		if ($do_license) {
			logTrc("_vol_set_asis  Licensing A_SIS and nearstore_option");
			logDbg("_vol_set_asis  Licensing nearstore_option");
			license($fc, 'nearstore_option');
			if ($@) {
				logWarn("_vol_set_asis Error: could not license 'nearstore_option': $@");
				die("_vol_set_asis  Error: could not license 'nearstore_option': $@");
			}
			logDbg("_vol_set_asis  Licensing 'A_SIS'");
			license($fc, 'a_sis');
			if ($@) {
				logWarn("_vol_set_asis Error: could not license 'a_sis: $@");
				die("_vol_set_asis  Error: could not license 'a_sis': $@");
			}
		}
		if ( !$check_license || (is_licensed($fc,'nearstore_option') && is_licensed($fc,'a_sis')) ) {
			logTrc("_vol_set_asis  Enabling/starting A_SIS");
			eval { dense_toggle(host=>$fh, path_name=>$vol_path, state=>"on") };
			if ($@) {
				logWarn("_vol_set_asis  Error: dense_toggle(\"on\") returned: $@");
				die("_vol_set_asis  Error: dense_toggle(\"on\") returned: $@");
			}
			logDbg("_vol_set_asis  Returned from dense_toggle(\"on\")");
			my $sched_string;
			my $test_mode;
			my $filer_output = sayEOF($fc, "priv set test");
			if ($filer_output =~ /unrecognized privilege/i) {
				$test_mode = 0;
				$sched_string = 'sun-sat@1-23';
				logDbg("_vol_set_asis:  'priv test' not available on filer");
			} else {
				$test_mode = 1;
				$sched_string = 'sun-sat@1-23@~0-59';
				logDbg("_vol_set_asis:  'priv test' is available on filer");
			}
			logDbg("_vol_set_asis  Calling dense_config with arguments:\n\n".
					"host:	\"${fh}:<$filer_name>\"\n".
					"path_name:	\"$vol_path\"\ntest_mode:	$test_mode\n".
					"schedule_string:	\"$sched_string\"");
			eval { dense_config(host=>$fh, path_name=>$vol_path, test_mode=>1, schedule_string=>$sched_string) };
			if ($@) {
				    logresult('INFO', "dense_config returned: $@");
			}
		} else {
			logresult('INFO', "_vol_set_asis  Error: 'a_sis' not properly licensed, cannot enable 'a_sis'");
		}
		# Need to make sure A_SIS is 'idle' before exiting....
		logTrc("_vol_set_asis  Calling 'dense_wait_for_status' -- want A_SIS to be in 'IDLE' state before returning");
		logDbg("_vol_set_asis  Calling 'dense_wait_for_status(IDLE)'  time_out: $dense_wait_timeout");
		my $start_time = time();
		eval { dense_wait_for_status(host=>$filer_name, path_name=>$vol_path, status=>"IDLE", time_out=>$dense_wait_timeout,) };
		if ($@) {
			logWarn("_vol_set_asis  Error: 'dense_wait_for_state(IDLE)' returned: $@");
		}
		my $wait_time = time() - $start_time;
		logDbg("_vol_set_asis  Returned from 'dense_wait_for_status(IDLE)' -- Waited $wait_time sec.");
	} else {
		logTrc("_vol_set_asis  Disabling/stopping A_SIS");
		eval { dense_toggle(host=>$fh, path_name=>$vol_path, state=>"off") };
		if ($@) {
			logresult('INFO', "dense_toggle(\"off\") returned: $@");
		}
		logDbg("_vol_set_asis  Returned from dense_toggle(\"off\")");
	}
	return(1);
}



=head2 _vol_set_compression($conn,$vol[,%opts])

Turn compression on/off for volume.

=cut

sub _vol_set_compression
{

my %raw_opts = (@_);
$raw_opts{filer} = ref($raw_opts{filer}) if (ref($raw_opts{filer}));
logDbg("Enter _vol_set_compression  Line: ".__LINE__."   Opts:\n\n".Dumper(\%raw_opts));
	my(%args) = validate_mixed( @_,
		[],
		{
			filer		=> { type => SCALAR|OBJECT },
			volume		=> { type => SCALAR },
			do_license	=> { type => SCALAR|UNDEF, optional=>1, default=>1 },
			check_license	=> { type => SCALAR|UNDEF, optional=>1, default=>1 },
			toggle		=> { type => SCALAR},
		}
	);
	die($@) if ($@);

	my $do_license;
	if (defined($args{do_license})) {
		$do_license = $args{do_license};
	} else {
		$do_license = 1;
	}

	my $check_license = $args{check_license};
	my $filer = $args{filer};
	my $fc = make_conn($filer);
	my $fh = $fc->hostp();
	my $filer_name = $fh->hostname();
	my $vol = $args{volume};
	my $vol_path = "/vol/$vol";
	my $actions = $args{actions};

	my $toggle;
	if (!defined($args{toggle})) {
		  die("_vol_set_compression  'toggle' option not specified");
	} elsif ($args{toggle} =~ /on|enable|1|start|true/i) {
		$toggle = 1;
	} elsif ($args{toggle} =~ /off|disable|0||stop|false|^$/i) {
		$toggle = 0;
	} else {
		die("_vol_set_compression  Unexpected value for 'toggle' option: \"$args{toggle}\"");
	}

	if ($toggle) {
		if ($do_license) {
			logTrc("_vol_set_compression  Licensing compression and nearstore_option");
			logDbg("_vol_set_compression  'nearstore_option'");
			license($fc, 'nearstore_option');
			if ($@) {
				logWarn("_vol_set_compression  Error: could not license 'nearstore_option': $@");
				die("_vol_set_compression  Error: could not license 'nearstore_option': $@");
			}
			logDbg("_vol_set_compression  'compression'");
			license($fc, 'compression') or die($@);
			if ($@) {
				logWarn("_vol_set_compression  Error: could not license 'compression': $@");
				die("_vol_set_compression  Error: could not license 'compression': $@");
			}
		}
		if ( !$check_license || (is_licensed($fc,'nearstore_option') && is_licensed($fc,'compression')) ) {
			logTrc("_vol_set_compression  Enabling Compression");
			enable_compression_option(filer=>$fh->id(), volname=>$vol);
			if ($@) {
				logresult('INFO', "_vol_set_compression Error returned from  enable_compression_opton:\n$@");
			}
			logDbg("_vol_set_compression  Returned from enable_compression");
		} else {
			logWarn("vol_set_compression:  'compression' not properly licensed, cannot enable compression");
			die("vol_set_compression:  'compression' not properly licensed, cannot enable compression");
		}
	} else {
		logTrc("_vol_set_compression  Disabling Compression");
		disable_compression_option(filer=>$fh->id(), volname=>$vol);
		if ($@) {
			logresult('INFO', "_vol_set_compression Error returned from _disable_compression_opton:\n$@");
		}
		logDbg("_vol_set_compression  Returned from disable_compression");
	}
	return(1);
}


=head2 _vol_set_encryption($conn,$vol[,%opts])

Turn encryption on/off for volume.

=cut

sub _vol_set_encryption
{

my %raw_opts = (@_);
$raw_opts{filer} = ref($raw_opts{iler}) if (ref($raw_opts{filer}));
logDbg("Enter _vol_set_encryption  Line: ".__LINE__."   Opts:\n\n".Dumper(\%raw_opts));
	my(%args) = validate_mixed( @_,
		[],
		{
			filer		=> { type => SCALAR|OBJECT },
			volume		=> { type => SCALAR },
			do_license	=> { type => SCALAR|UNDEF, optional=>1, default=>1 },
			check_license	=> { type => SCALAR|UNDEF, optional=>1, default=>1 },
			toggle		=> { type => SCALAR},
		}
	);
	die($@) if ($@);

	my $check_license = $args{check_license};
	my $do_license;
	if (defined($args{do_license})) {
		$do_license = $args{do_license};
	} else {
		$do_license = 1;
	}

	my $filer = $args{filer};
	my $fc = make_conn($filer);
	my $fh = $fc->hostp();
	my $filer_name = $fh->hostname();
	my $vol = $args{volume};
	my $vol_path = "/vol/$vol";
	my $actions = $args{actions};

	my $toggle;
	if (!defined($args{toggle})) {
		  die("_vol_set_encryption  'toggle' option not specified");
	} elsif ($args{toggle} =~ /on|enable|1|start|true/i) {
		$toggle = 1;
	} elsif ($args{toggle} =~ /off|disable|0||stop|false|^$/i) {
		$toggle = 0;
	} else {
		die("_vol_set_encryption  Unexpected value for 'toggle' option: \"$args{toggle}\"");
	}

logWarn("_vol_set_encryption not yet implemented");
#die("_vol_set_encryption not yet implemented");
return(0);

#	if ($toggle) {
#		if ($do_license) {
#		}
#	} else {
#	}

	return(1);
}


# !!!!!!!!!!!!!!!!!!!   End Change   !!!!!!!!!!!!!!!!!!!!!!!!!



=head1 Volume::DiskStatus

Volume::DiskStatus is a data-storage only object. It has no methods,
and contains the following fields:

=head2 state

=head2 name

=head2 ha

=head2 shelf

=head2 bay

=head2 chan

=head2 used_mb

=head2 used_blks

=head2 phys_mb

=head2 phys_blks

=cut


package SPARSE::Volume::DiskStatus;
use strict;
use Class::MethodMaker
	new_with_init => 'new',
	get_set => [ qw(name ha shelf bay chan used_mb used_blks phys_mb phys_blks state) ]
;

sub init
{
	my($self, %args) = @_;
	foreach my $field qw(name ha shelf bay chan used_mb used_blks phys_mb phys_blks state) {
		if(exists $args{$field}) {
			$self->$field($args{$field});
		}
	}
}




1;

=head1 SEE ALSO

L<Module::Common>
L<Params::Validate>
L<Utils::Units>

=head1 HISTORY

This library used be about three different modules, which got consolidated
by a person who had no idea how volumes work. If there are discrepancies
in style, that's probably why ;)

=head1 TODO

vol_mirror()

vol_verify()

disk_free -  maybe in diff package

vol_raidgroup_status, vol_spare_status, vol_disks  -> move to aggr?

=cut

