#Hey emacs, this is basically -*- perl -*-
# $Id: //depot/prod/test/main/lib/SPI.pm#1

#***********************************************************************
#
# NAME
#   SPI.pm - 
#
# SYNOPSIS
#   Library for queries against the SPI web service interface
#   
#
# DESCRIPTION
#   This library provides functions used to query the SPI (Service Processor
#   Interface) web service used in the RSA Partner File Collection feature
#   being introduced in RR.1
#
# AUTHOR
#   dl-cluster-qa
#
# HISTORY
#   06/15/2011 - Initial version
#
#***********************************************************************

## @summary 

package SPI;

#use statements
use strict;
use Data::Dumper;
use Tharn qw($Log snooze);
use Options;
use Params::Common;
use Params::Validate qw(:all);

use NACL::C::Node;
use NACL::C::Cluster;
use NACL::C::Interface;
use NACL::C::Vserver;
use NACL::C::VserverServicesWeb;
use NACL::C::VserverServicesWebAccess;
use NACL::Exceptions::NoElementsFound;
use LWP::UserAgent;
use HTTP::Request;
use HTTP::Response;
use HTML::TokeParser;
use Data::Compare;

#Global variables

#Methods
###############################################################################
#Sub: configure_spi_websvc
#Input Params: hash
#Return values: none
#Summary: Enables or disables the spi web service on the specified node
###############################################################################
sub configure_spi_websvc {
    my %params = (
            enabled => 'true',
            @_,
    );
    
    my $node = NACL::C::Node->find();
    my $enabled = $params{enabled};
    
    my $cserver = NACL::C::Vserver->find(command_interface => $node,
                                         filter => { type => "admin" });
     
    # Enable or disable the SPI web service                                     
    eval {
        NACL::C::VserverServicesWeb->modify(command_interface => $node,
                                        vserver => "*",
                                        name    => "spi",
                                        enabled => $enabled);
    };

    # Create or delete access to the SPI web service
    my %args = (vserver => $cserver, name => "spi", role => "admin");
    if ($enabled eq "true") {

        eval {
            NACL::C::VserverServicesWebAccess->create(command_interface => $node, %args);

        };
     }
    if ($enabled eq "false") {
        eval {
            NACL::C::VserverServicesWebAccess->delete(command_interface => $node, %args);
        };
    }      
      
}


###############################################################################
#Sub: get_node_identifiers
#Input Params: none
#Return values: hash containing retrieved XML
#Summary: Queries the SPI interface on /spi and retrives node-identifiers for the cluster
###############################################################################
sub get_node_identifiers {

    use NACL::C::Interface;
    use XML::LibXML;
    use DataUtils;
    
    # Obtain the cserver LIF IP
    my $node = NACL::C::Node->find();
    my $Interface = NACL::C::Interface->find(command_interface => $node,
                                             filter => {role => "cluster-mgmt"});
    my $Interface_state = $Interface->state();
    my $IP = $Interface_state->address();
    
    # Hit the /spi URL and obtain the response
    my %hash;
    my $hp = \%hash;
    my $parser = XML::LibXML->new();
    my $uri = "http://$IP/spi/";
    my $ua = LWP::UserAgent->new(agent=>"RSA");
    my $req = HTTP::Request->new(GET => $uri );
    my $password = $node->hostrec()->{default_password};
    $req->authorization_basic('admin',$password);
    $Log->comment("Performing GET on $uri");
    my $result = $ua->request($req);
    
    my $http_status = $result->status_line();
    $Log->comment("HTTP STATUS = $http_status");

    # BURT 655103 - If /spi does not return 200 OK the first time, retry for a
    # short while in case Apache hasn't reacted to the WSDB update yet
    my $retry_count = 0;
    while (($http_status !~ /200/) && ($retry_count < 6)) {
        snooze(3);
        $retry_count++;
        $Log->comment("Retrying GET on $uri...");
        $result = $ua->request($req);
        $http_status = $result->status_line();
        $Log->comment("HTTP STATUS = $http_status");        
    }
    
    if ($http_status !~ /200/) {
        $Log->warn("GET $uri returned HTTP STATUS $http_status");
        return 0;
    }
        
    my $content = $result->content();
    $Log->comment("\nContent: ".Dumper($content));
    
    # Parse the XML into a hash
    my $xmlDoc = $parser->parse_string($result->content());
    my @xmlnodes = $xmlDoc->getElementsByTagName('spi-cluster-index');
  
    DataUtils::build_hash_from_xml( xmlnode => $xmlnodes[0], hash => $hp );    
    return $hp;
    
}

###############################################################################
#Sub: get_directory_listing
#Input Params: hash
#Return values: array containing directory listing
#Summary: Retrieves a directory listing via SPI for specified remote node and directory
###############################################################################

sub get_directory_listing {
    my %params = (
            remote_identifier => undef,
            directory => undef,
            @_,
    );
    
    if (!$params{remote_identifier}) {
        $Log->error("Must supply a node-name, system-id, or UUID for remote node \n");
        return 0;
    }
    if (!$params{directory}) {
        $Log->error("Must supply /etc/log/ or /etc/crash/ \n");
        return 0;
    }    
    my @listing;
    
    # Obtain the cserver LIF IP
    my $node = NACL::C::Node->find();
    my $Interface = NACL::C::Interface->find(command_interface => $node,
                                             filter => {role => "cluster-mgmt"});
                                             
    my $Interface_state = $Interface->state();
    my $IP = $Interface_state->address();

    # Hit /spi URL for specified directory and obtain the response
    my $uri = "http://$IP/spi/".$params{remote_identifier}.$params{directory};
    my $ua = LWP::UserAgent->new(agent=>"RSA");
    $Log->comment("Performing GET on $uri...");
    my $req = HTTP::Request->new(GET => $uri );
    my $password = $node->hostrec()->{default_password};
    $req->authorization_basic('admin',$password);
    my $result = $ua->request($req);
    #$Log->comment(Dumper($result));
    
    my $http_status = $result->status_line();
    $Log->comment("HTTP STATUS = $http_status");
    
    while ($http_status =~ /202/) {
        snooze(3);
        $Log->comment("Retrying GET on $uri...");
        $result = $ua->request($req);
        $http_status = $result->status_line();
        $Log->comment("HTTP STATUS = $http_status");        
    }
    
    if ($http_status !~ /200/) {
        $Log->warn("GET $uri returned HTTP STATUS $http_status");
        return 0;
    }
    
    # Parse the response, and push entries to the array
    my $content = $result->content();
    my $tokens = HTML::TokeParser->new(\$content);

    while (my $token = $tokens->get_tag("a")) {
        my $item = $token->[1]{href} || "-";
        my $description = $tokens->get_trimmed_text("/a");
        next if ($description =~ /Parent Directory/); # Ignore the link to Parent Directory
        if ($item =~ /\%3A/) {
            $item =~ s/%3A/:/g;
        }
        #$Log->comment("Pushing $item");
        push (@listing,$item);
     
    }
   
    return @listing;
      
}


###############################################################################
#Sub: get_and_verify_directory_listing
#Input Params: hash
#Return values: 1 on success, 0 on failure
#Summary: Retrieves a directory listing via SPI for specified remote node and directory,
#         and again via systemshell, then compares the two to verify they are the same
###############################################################################

sub get_and_verify_directory_listing {
    my %params = (
            remote_identifier => undef,
            directory => undef,
            remote_node_name => undef,
            @_,
    );
    
    if (!$params{remote_identifier}) {
        $Log->error("Must supply a node-name, system-id, or UUID for remote node \n");
        return 0;
    }
    if (!$params{directory}) {
        $Log->error("Must supply /etc/log/ or /etc/crash/ \n");
        return 0;
    }
    if (!$params{remote_node_name}) {
        $Log->error("Must supply name of remote node \n");
        return 0;
    }
    
    my $remote_identifier = $params{remote_identifier};
    my $dir = $params{directory};
    
    my @websvc_listing = get_directory_listing(remote_identifier=>$remote_identifier, directory=>$dir);
    if ($websvc_listing[0] eq '0') {
        $Log->comment("Failed to retrieve directory listing for $params{directory}");
        return 0;
    }
    if (!@websvc_listing) {
        $Log->comment("No contents found");
    }
    my @websvc_listing_sorted = sort(@websvc_listing);
    
    my $node = NACL::C::Node->find();
    
    # Get directory listing via systemshell so we can compare
    my $Api_Set_Obj = NACL::APISet->new(name   => $params{remote_node_name}, # Prototype does not do remote mounts yet,
                                                                              # so we have to compare to local filesystem 
                                        category  => "Node",
                                        interface => "CLI",
                                        set       => "Systemshell"
                                        );
    my $shell_response = $Api_Set_Obj->execute_raw_command(command=>"ls -1pAf /mroot".$dir);  # To mimic the output from spi...
                                                                                              # -1 = each file on separate line
                                                                                              # -p = include / indicator after dirs
                                                                                              # -A = include hidden . files but not . & ..
                                                                                              # -f = do not sort
    my @shell_listing = split(/\n/,$shell_response);
    my @shell_listing_sorted = sort(@shell_listing);

    if (!Compare(\@shell_listing_sorted, \@websvc_listing_sorted)) {
       $Log->comment("Listing returned by SPI webservice does not match listing from Systemshell");
       $Log->comment("Systemshell listing: ".Dumper(@shell_listing_sorted));
       $Log->comment("Web service listing: ".Dumper(@websvc_listing_sorted));
       return 0;
    }
    
    return 1;
      
}


###############################################################################
#Sub: get_file
#Input Params: hash
#Return values: 1 on success, 0 on failure
#Summary: Retrieve a file through SPI web service and save to savedir
###############################################################################
sub get_file {
    my %params = (
            remote_identifier => undef,
            directory => undef,
            filename => undef,
            savedir => undef,
            @_,
    );
    
    if (!$params{remote_identifier}) {
        $Log->error("Must supply a node-name, system-id, or UUID for remote node \n");
        return 0;
    }
    if (!$params{directory}) {
        $Log->error("Must supply /etc/log/ or /etc/crash/ \n");
        return 0;
    }
    if (!$params{filename}) {
        $Log->error("Must supply a file name to transfer \n");
        return 0;
    }
    if (!$params{savedir}) {
        $Log->error("Must supply a directory on client to which $params{filename} can be written \n");
        return 0;
    }

    my $node = NACL::C::Node->find();
    my $Interface = NACL::C::Interface->find(command_interface => $node,
                                             filter => {role => "cluster-mgmt"});
                                             
    my $Interface_state = $Interface->state();
    my $IP = $Interface_state->address();

    my $uri = "http://$IP/spi/".$params{remote_identifier}.$params{directory}."/".$params{filename};
    my $ua = LWP::UserAgent->new(agent=>"RSA");
    $Log->comment("Performing GET on $uri...");
    my $req = HTTP::Request->new(GET => $uri );
    my $password = $node->hostrec()->{default_password};
    $req->authorization_basic('admin',$password);
    my $result = $ua->request($req);
    #$Log->comment(Dumper($result));
    
    my $http_status = $result->status_line();
    $Log->comment("HTTP STATUS = $http_status");
        
    while ($http_status =~ /202/) {
        snooze(3);
        $Log->comment("Retrying GET on $uri...");
        $result = $ua->request($req);
        $http_status = $result->status_line();
        $Log->comment("HTTP STATUS = $http_status");        
    }
        
    if ($http_status !~ /200/) {
        $Log->warn("GET $uri returned HTTP STATUS $http_status");
        return 0;
    }
    
    my $content = $result->content();
    
    system ("mkdir -p $params{savedir}");
    my $write_path = $params{savedir}."/".$params{filename};
    $Log->comment("Writing file to $write_path");
    open (FH,">$write_path");
    print FH $content;
    close (FH);
    
    return 1;
}


###############################################################################
#Sub: get_http_status_code
#Input Params: hash
#Return values: scalar containing HTTP status code string
#Summary: 
###############################################################################

sub get_http_status_code {
    my %params = (
            remote_identifier => undef,
            directory => undef,
            filename => undef,
            @_,
    );
    
    if (!$params{remote_identifier}) {
        $Log->error("Must supply a node-name, system-id, or UUID for remote node \n");
        return 0;
    }
    if (!$params{directory}) {
        $Log->error("Must supply a directory to query \n");
        return 0;
    }    
    
    # Obtain the cserver LIF IP
    my $node = NACL::C::Node->find();
    my $Interface = NACL::C::Interface->find(command_interface => $node,
                                             filter => {role => "cluster-mgmt"});
                                             
    my $Interface_state = $Interface->state();
    my $IP = $Interface_state->address();

    # Hit /spi URL for specified directory/filename and obtain the response
    my $uri = "http://$IP/spi/".$params{remote_identifier}.$params{directory}."/".$params{filename};
    my $ua = LWP::UserAgent->new(agent=>"RSA");
    $Log->comment("Performing GET on $uri...");
    my $req = HTTP::Request->new(GET => $uri );
    my $password = $node->hostrec()->{default_password};
    $req->authorization_basic('admin',$password);
    my $result = $ua->request($req);
    $Log->comment(Dumper($result));
    
    my $http_status = $result->status_line();
    $Log->comment("HTTP STATUS = $http_status");
    return $http_status;
    
}


###############################################################################
#Sub: get_partial_content
#Input Params: hash
#Return values: 0 on fail, 1 on success
#Summary: 
###############################################################################

sub get_partial_content {
    my %params = (
            remote_identifier => undef,
            @_,
    );
    
    if (!$params{remote_identifier}) {
        $Log->error("Must supply a node-name, system-id, or UUID for remote node \n");
        return 0;
    }
    
    # Obtain the cserver LIF IP
    my $node = NACL::C::Node->find();
    my $Interface = NACL::C::Interface->find(command_interface => $node,
                                             filter => {role => "cluster-mgmt"});
                                             
    my $Interface_state = $Interface->state();
    my $IP = $Interface_state->address();

    # Hit /spi URL for specified directory/filename and obtain the response
    my $uri = "http://$IP/spi/".$params{remote_identifier}."/etc/log/mlog/mgwd.log";
    my $ua = LWP::UserAgent->new(agent=>"RSA");
    $Log->comment("Performing HEAD on $uri...");
    my $req = HTTP::Request->new(HEAD => $uri );
    my $password = $node->hostrec()->{default_password};
    $req->authorization_basic('admin',$password);
    my $result = $ua->request($req);
    $Log->comment(Dumper($result));
    my $content_len = $result->header('Content-Length');
    $Log->comment("CONTENT LENGTH IS: $content_len");

    # GET for only the first 500 bytes of mgwd.log should return a status code of 206 (Partial Success)
    $Log->comment("\nPerforming partial GET for bytes 0-499...");
    $req = HTTP::Request->new(GET => $uri );
    $req->authorization_basic('admin',$password);
    $req->header(Range=>'bytes=0-499');
    $result = $ua->request($req);
    $Log->comment(Dumper($result));
    
    my $http_status = $result->status_line();
    my $partial_content_len = $result->header('Content-Length');
    
    if (($http_status !~ /206/) || ($partial_content_len ne '500')) {
        $Log->comment("Failed to verify partial GET of 500 bytes (HTTP status: $http_status and Content-Length: $partial_content_len");
        return 0;
    }
    
    # GET for a range 1000 bytes beyond the end should return a status code of 416 (Out of Range)
    my $out_of_range_len = $content_len+1000;
    $Log->comment("\nPerforming partial GET for out of range content (501-".$out_of_range_len.")...");
    $req = HTTP::Request->new(GET => $uri );
    $req->authorization_basic('admin',$password);
    $req->header(Range=>"bytes=0-$out_of_range_len");
    $result = $ua->request($req);
    $Log->comment(Dumper($result));
    
    $http_status = $result->status_line();
    $partial_content_len = $result->header('Content-Length');
    
    if ($http_status !~ /416/) {
        $Log->comment("HTTP STATUS = $http_status");
        $Log->comment("Failed to verify HTTP status for out of range GET (expected 416, got $http_status)");
        return 0;
    }
    
    return 1;
    
}

return 1;

=pod

=head1 NAME

SPI.pm - Provides methods for obtaining directory listings and downloading files via the SPI (Service Processor Interface) web service.


=head1 SYNOPSIS

    use SPI;
    my $node_ids = SPI::get_node_identifiers();
    
    my @listing = SPI::get_directory_listing(remote_identifier=>"csiqa-6070-13", directory=>"/etc/log");
    
    if (!SPI::get_file(remote_identifier=>"csiqa-6070-13", directory=>"/etc/log",
                              filename=>$listing[1], savedir=>$LOGDIR)) {
        $Log->comment("Error retrieving file");
        return $TCD::FAIL;
    }  


=head1 DESCRIPTION

  
    
=head1 METHODS

    configure_spi_websvc();      
        Description:
            This method will allow you to enable or disable the SPI web service
                on the cluster-mgmt interface (c-server LIF).
        Inputs:
            enabled => <true|false> Enabled setting for spi webservice
        Outputs:
            None
        Examples:
            see full example below
            
            
    get_node_identifiers();
        Description:
            This method will query the SPI webservice on http://<cserver-lif>/spi and
                obtain the node-identifiers for nodes in the cluster.
                Node-identifiers for SPI can be either node-name, system-id, or UUID.
        Inputs:
            None
        Outputs:
            Hash containing node-identifiers, or 0 if GET was deemed unsuccessful
        Examples:
            see full example below
            
                     
    get_directory_listing();
        Description:
            This method will query the SPI webservice on http://<cserver-lif>/spi and obtain a directory
                listing for the specified /mroot subdirectory.
        Inputs:
            remote-identifier => <text> The node-name, system-id, or UUID for the node to be queried
                using the SPI webservice.  Obtained using get_node_identifiers().
            directory => <text> The /mroot sub-directory for which a listing should be obtained;
                i.e. /etc/log or /etc/crash 
        Outputs:
            Array containing directory contents
        Examples:
            see full example below
            
    get_and_verify_directory_listing();
        Description:
            This method will query the SPI webservice on http://<cserver-lif>/spi and obtain a directory
                listing for the specified /mroot subdirectory, then obtain the same directory listing via
                systemshell, and verify the two match.
        Inputs:
            remote-identifier => <text> The node-name, system-id, or UUID for the node to be queried
                using the SPI webservice.  Obtained using get_node_identifiers().
            directory => <text> The /mroot sub-directory for which a listing should be obtained;
                i.e. /etc/log or /etc/crash
            remote_node_name => <text> The node-name of the remote node node,
                to be accessed via systemshell.
        Outputs:
            1 if successful, 0 if unsuccessful
        Examples:
            see full example below
            
    get_file();
        Description:
            This method will query the SPI webservice on http://<cserver-lif>/spi and download the
                filename specified from the /mroot subdirectory specified.
        Inputs:
            remote_identifier => <text> The node-name, system-id, or UUID for the node to be queried
                using the SPI webservice.  Obtained using get_node_identifiers().
            directory => <text> The /mroot sub-directory for which a listing should be obtained; i.e.
                /etc/log or /etc/crash
            filename => <text> The filename to download
            savedir => <text> A directory on the test client to which the downloaded file should be
                saved.
        Outputs:
            1 if successful, 0 if unsuccessful
        Examples:
            see full example below
            
            
    get_http_status_code();
        Description:
            This method will query the SPI webservice on http://<cserver-lif>/spi plus the specified
                directory and filename.  Returns the HTTP status code of the GET.
        Inputs:
            remote_identifier => <text> The node-name, system-id, or UUID for the node to be queried
                using the SPI webservice.  Obtained using get_node_identifiers().
            directory => <text> The /mroot sub-directory for which a listing should be obtained; i.e.
                /etc/log or /etc/crash
            filename => <text> The filename to download
        Outputs:
            HTTP status code returned by the GET
        Examples:
            see full example below            
            
            
=head1 EXAMPLES

Ex1: Enable SPI web service and query for node-identifiers

    use SPI;
    
    SPI::configure_spi_websvc(enabled => "true");
    my $node_ids = SPI::get_node_identifiers();
    
     
=cut
