#!/usr/local/bin/perl
# $Id: //depot/prod/test/Rsierranevada/lib/SPM.pm#1 $ 
## @summary     Provides the api to monitor, kill, and gather information on SPM managed processes.
## @author      cwilcox@netapp.com,dl-cluster-qa@netapp.com
## @pod here

=head1 NAME

SPM.pm

=head1 SYNOPSIS
 SYNOPSIS:
    use Server;
    use SPM;

    $obj = new SPM(server => $Server);

    $obj->initialize_all(server => $Server);

    $obj->find_avs(server => $Server);
    # finds the node where the cluster instance of avs resides
    # returns the node, or warning that avs doesn't exist

    $obj->kill_proc(
        procs=>"all",
        server=>$Server,
        nodes=>"all",
        iterations=>10,
        interval=>10,
       );
    # Given the parameters, instructs which node(s), proc(s),
    # how many iterations, and the duration at which to wait
    # between iterations of killing procs.

=head1 DESCRIPTION

C<SPM> provides the api to monitor, kill, and gather information on SPM managed processes.

=head1 FUNCTIONS

=cut

## @summary     SPM library

package SPM;

#use statements
use strict;
use Data::Dumper;
use Server;
use Tharn qw($Log);
use Options;
use Params::Common;
use Params::Validate qw(:all);
use Carp;

my @FIELDS =  qw(node name handle pid restarts status 
                 cmd-line max-restarts current-rss 
                 page-reclaims page-faults 
                 voluntary-context-swaps 
                 involuntary-context-swaps 
                 num-swaps num-fs-input 
                 num-fs-output state
                );

#my @PROCS = qw(ucoreman time_state named vifmgr notifyd vldb secd avs mgwd);
my @PROCS = qw(ucoreman time_state named vifmgr notifyd vldb secd);

=pod

=head2 new

C<new(server=>$Server_obj)>


B<Description>
Creates the hash data structure '$hash{node}{proc}{field}' for each node in Server obj
for each proc in global var @PROCS for each field in @FIELDS

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Blessed hash structure as an SPM object 

=cut

sub new {
    my %proc_struct;
    my $class = shift;
    my %params = validate(
    @_,
    {
       server => { type  => HASHREF },
    }
    );
    
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    
    foreach my $node (@Nodes) {
        foreach my $proc (@PROCS) {
            foreach my $field (@FIELDS) {
              $proc_struct{$node}{$proc}{$field} = undef ;
            }
        }
    }
 
    my $self = {%proc_struct};
    bless $self,$class;

    $self->initialize_all(server=>$Server);
    
    return $self;
}

=pod

=head2 initialize_all

C<initialize_all(server=>$Server_obj)>

B<Description>
Instantiates the hash data structure '$hash{node}{proc}{field}' for each node in Server obj
for each proc in global var @PROCS for each field in global var @FIELDS.

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Updated hash structure as an SPM object

=cut

sub initialize_all {
    my $self = shift;
    my %params = validate(
    @_,
    {
       server => { type  => HASHREF },
    }
    );
    my $output;
    my $Server = $params{server}; 
    my @Nodes = $Server->servers();
    
    foreach my $node (@Nodes) {
            my $proc;
            eval{
                $output = $Server->spm_show(node=>$node);
            };
            if ($@) {
                $Log->error("spm show command failed on node $node");
                return $@;
            } else {
                foreach my $index (@$output) {
                    foreach my $field (@FIELDS) {
                        $self->{$node}{$index->{name}}{$field} = $index->{$field};
                    }
                }
            }
    }
    return $self;    
}

=pod

=head2 find_avs

C<initialize_all(server=>$Server_obj)>

B<Description>
Finds the node where the cluster instance of avs resides. 

B<Input>

=over 4

=item * B<server> => $Server_obj

=back

B<Output>
Returns the node if an instance of antivirus is found, 
or the warning: "There are no instances of antivirus running in the cluster."

=cut

sub find_avs {
    my $self = shift;
    my %params = validate(
    @_,
    {
       server => { type  => HASHREF },
    }
    );
    my $Server = $params{server};
    my @Nodes = $Server->servers();
    my $not_found = "There are no instances of antivirus running in the cluster.";

    foreach my $node (@Nodes) {
        my $output;
        eval{
             $output = $Server->spm_show(node=>$node,name=>"avs");
            };

        if ($@ =~ /no entries matching your query/) {
            $Log->comment("There are no instances of antivirus on node $node");
        } elsif ($@) {
            $Log->comment("The spm show command failed due to an unknown error: ".Dumper($@));
        } else { 
            return $node; 
        }
    }
    $Log->warn($not_found);
    return $not_found;
}

=pod

=head2 kill_proc

C<kill_proc(
    procs=>"all",
    server=>$Server,
    nodes=>"all",
    iterations=>1,
    interval=>10,
   );>

B<Description>
Given the parameters, instructs which node(s), proc(s), how many iterations, 
and the duration at which to wait between iterations of killing procs.

B<Input>

=over 4

=item * B<server> 
The Server object.

=item * B<procs>
(Optional)
The procs to be killed. The process name(s) can be passed as an array reference to
list desired procs, a string of a single proc name, or the string, "all", which 
will loop through the global class variable: 
@PROCS:
    ucoreman 
    time_state 
    named 
    vifmgr 
    notifyd 
    vldb 
    secd

The default is "all".

=item * B<server>
(Required)
The Server object.

=item * B<nodes>
(Optional)
The nodes upon which resides the processes to be killed. The nodes name(s) can be passed 
as an array reference to list desired nodes, a string of a single node name, or the string, 
"all", which will loop through the nodes instantiated in the Server object.

The default is "all".

=item * B<iterations>
(Optional)
The number of iterations which the process will be killed.

The default is 1 iteration.

=item * B<interval>
(Optional)
The duration in seconds between killing processes.

The default is 5 seconds.

=back

B<Output>
Returns a list of processes per node that are in a "slow_restart_waiting" state which
would mean that the process has entered into the back-off algorithm loop.

node:process:state 

=cut

sub kill_proc {
    my $self = shift;
    my @node_array;
    my @proc_array;
    my @waiting_list;
    my $output;

    my %params = validate(
    @_,
    {
        procs       => { type  => SCALAR|ARRAYREF,
                         default => "all",
                       },
        server      => { type  => HASHREF },
        nodes       => { type  => SCALAR|ARRAYREF,
                         default => "all",
                       },
        iterations  => { type => SCALAR,
                         default => 1,
                       },
        interval    => { type => SCALAR, 
                         default => 5,
                       },
        signal      => { type => SCALAR,
                         default => "KILL",
                       },
    }
    );

    my $Nodes = $params{nodes};
    my $Server = $params{server};
    my $iterations = $params{iterations};
    my $interval = $params{interval};
    my $procs = $params{procs};
    my $signal = $params{signal};

    if (($Nodes ne "all") && (ref($Nodes) ne "ARRAYREF")) {
        push @node_array, $Nodes;
    } elsif ($Nodes eq "all") {
        @node_array = $Server->servers();
    } else {
        @node_array = @$Nodes;
    }

    if (($procs ne "all") && (ref($procs) ne "ARRAYREF")) {    
        push @proc_array, $procs;
    } elsif ($procs eq "all") {
        @proc_array = @PROCS;
    } else {
        @proc_array = @$procs;
    }

    foreach my $node (@node_array) { 
        $Server->set_execution_server(-Server =>$node);
        foreach my $proc (@proc_array) {
            # Grab the values to update the object
            eval{
                $output = $Server->spm_show(node=>$node,name=>$proc);
                };
            if ($@) {
                $Log->error(die "spm show command failed for $proc process on node $node");
            } else {
                $self->{$node}{$proc}{pid} = $output->[0]{pid};
                $self->{$node}{$proc}{state} = $output->[0]{state};
            }
            # If process is in a "slow_restart_waiting" state, skip the kill, and push node->proc->state to waiting_list.
            if ($self->{$node}{$proc}{state} eq "slow_restart_waiting") {
                $Log->comment("$proc process on $node is in a slow_restart_waiting state. Skipping kill.");
                push @waiting_list, "$node:$proc:$self->{$node}{$proc}{state}";
            } else {
                # For each iteration, do a process kill
                for (my $i = 0; $i < $iterations; $i++) {
                    my $pid = $self->{$node}{$proc}{pid};
                    eval{
                        $Server->shell(cmd=>"sudo kill -s $signal $pid");
                        };
                    if ($@) {
                        $Log->error(die "Attempt to kill $proc managed by spm on $node unexpectantly failed: $@.");
                    } else {
                        $Log->comment("Killed $proc process on $node.");
                        # SPM can be slow to grab the new status after the kill.
                        # While the process pid in the object matches the output of the spm show command,
                        # perform another show to allow it to catch up.
                        my $max_wait_time = 900; # Wait 15 minutes for the new PID
                        my $start_wait_time = time;
                        my $current_wait_time = $start_wait_time;
                        while ($self->{$node}{$proc}{pid} eq $pid) {
                            eval{
                                $output = $Server->spm_show(node=>$node,name=>$proc);
                                };
                            if ($@) {
                                $Log->error(die "spm show command failed for $proc process on node $node");
                                last;
                            } elsif ($output->[0]{state} eq "slow_restart_waiting") {
                                $Log->comment("$proc process on $node is in a slow_restart_waiting state. Breaking out of state-watch loop.");
                                last;
                            } else {
                                $self->{$node}{$proc}{pid} = $output->[0]{pid};
                                $self->{$node}{$proc}{state} = $output->[0]{state};
                                
                                $current_wait_time = time;
                                if (($current_wait_time - $start_wait_time) > $max_wait_time) {
                                    $Log->error(die "process [$proc] on node [$node] did not die when we told it to.  The PID is the same after waiting [$max_wait_time]s");
                                    last;
                                }
                            }
                        }
                    }
                    # This is the interval between iterations for same proc
                    sleep($interval);
                }
            }
        }    
    }
    return \@waiting_list;
}

=pod

=head1 LIMITATIONS    
The mgwd proc is omitted from being killed due to the current inability of Server.pm to reconnect after shell 
session is lost.

=cut

1    
