## Routines for the manipulation of NFS export policies 
##
## @author      dascenj@netapp.com
##
## @pod here

=head1 NAME

    SSQA::Cmode::ExportPolicy - manipulates export policies 

=head1 SYNOPSIS

    Several routines for manipulation of NFS export policies. 

=head1 DESCRIPTION

    This library is based on NACL, and requires a NACL node object to be
    supplied. (See NACL docs for more information). This is a central 
    place to store routines that interact with NFS export policies.

=head1 TODO

    I don't know, it is already pretty awesome.  

=head1 BUGS

    Some probably exist but have not been listed yet.

=head1 STATUS

  Shared

=head1 AUTHOR

  Jd (dascenj@netapp.com)

=head1 COPYRIGHT

Copyright 2010, Network Appliance., All Rights Reserved


=cut

package SSQA::CMode::ExportPolicy;

use warnings;
use strict; 

use NATE::BaseException qw(:try);
use NACL::C::ExportPolicyRule;
use NACL::C::ExportPolicy;
use SSQA::CMode::ExportPolicyRule;
use Params::Validate qw(validate validate_with OBJECT SCALAR ARRAYREF HASHREF BOOLEAN UNDEF);
use Data::Dumper;
use Tharn qw ( $Log config_result);

=head2 choose_random 
 
    Returns a random policy name from the CMode cluster. Logs
    a comment if a policy is not found. Will not return the 
    default export policy.

=cut

sub choose_random {
    my (%params) = @_;

    my $defaultPolicyName = "default";
    my $randomPolicyName = $defaultPolicyName;
    
    # Get a random policy name from the array. 
    # lets not mess with the default policy 
    my $policyObj = {};
    my $counter = 0;
    my $maxRetries = 10;
    while ($randomPolicyName =~ /$defaultPolicyName/i) {
        # pick one policy and make sure to satisfy the while condition.
        $policyObj = @{$params{policyArray}}[rand(0 - scalar(@{$params{policyArray}}))];
        $randomPolicyName = $policyObj->policyname();
      
        $counter++;
          
        if ($counter == $maxRetries) {
           $Log->comment("Could not find a random policy to deliver. Perhaps you didn't do setup?");
           return $policyObj = {};
        }
    }

    $Log->comment("Randomly selected policy " . $policyObj->policyname());

    return $policyObj;
}

=head2 generate_name 
 
    Creates a new policy name. The name_base_text option
    can be supplied to identify to identify the policy. The default
    is SSQA_TestPolicy. All new policy names will utilize the base 
    value and some randomized charecters.

=cut

sub generate_name {
    my (%params) =
        validate(@_, {
            node        => {
               type     => HASHREF,
               optional => 0,
            },
            name_base_text => { type => SCALAR,
                            optional => 1,
                             default => "SSQA_TestPolicy",
            }
        }
    );

    my $newPolicyName;
    my $stringIsUnique = 0;   

    require String::Random;
    my $randomData = new String::Random;

    while ($stringIsUnique == 0) {   
        # Setup a unique policy name
        $newPolicyName = $params{name_base_text} . "_" . $randomData->randregex('\w\w\w') . "_" . $randomData->randregex('\d\d\d');

        $Log->comment("Chose a random new policy name of " . $newPolicyName . ". Verify that the name does not already exist.");

        # make sure that the name is really unique 
        $stringIsUnique = &_check_unique_name(node => $params{node},
                                           newName => $newPolicyName);
    };
 
    return $newPolicyName;
}

# This private routine is used by "generate_name" to 
# verify that the newly chosen name is truly unique.

sub _check_unique_name {

    my (%params) =
        validate(@_, {
            node        => {
               type     => HASHREF,
               optional => 0,
            },
            newName => { type => SCALAR,
                     optional => 0,
            }
        }
    );

    my $noPolicyText = "There are no entries matching your query.";
    my $isUnique = 0;

    try {
        NACL::C::ExportPolicy->find(filter => {policyname => $params{newName}},
                         command_interface => $params{node});
    } catch NATE::BaseException with {
            my $ex = shift;
            $Log->comment("EXCEPTION: " . $ex->text());

            # If the policy does not exist, an exception is thrown. The text includes 
            # a message we can watch for. 
            if ($ex->text() =~ /$noPolicyText/i) {
                $Log->comment("The preceeding exception is expected when the policy name is unique.");
                $isUnique = 1;
            }
    };
 
    if ($isUnique) {
        return 1;
    }

    return 0;
}

=head2 find_volumes_by_policy 
 
    Given a policy name, returns an array of 
    volumes which are assigned the given policy.
 
=cut

sub find_volumes_by_policy {
    my (%params) =
        validate(@_, {
            node => {
               type     => HASHREF,
               optional => 0,
            },
            policy => {
               type     => HASHREF,
               optional => 0,
            }
        }
    );

    require NACL::C::Volume;

    my @vols = NACL::C::Volume->find(command_interface => $params{node},
                                     filter            => {policy => $params{policy}->policyname()});

    if (not $vols[0]->volume() =~ /./) { 
        $Log->comment("The policy named " . $params{policy}->policyname() . " is not currently applied to any volume.");
    } else { 
        foreach my $vol (@vols) {
            $Log->comment("The policy named " . $params{policy}->policyname() . " is currently assigned to " . 
                           $vol->volume().  " at virtual server " . $vol->vserver());
        }
    }
    return \@vols;
}

=head2 policy_is_applied_to_volume 

    Checks to see if a policy is applied to any volumes. 
    returns true or false. 

=cut

sub policy_is_applied_to_volume {
    my (%params) = @_;

    my @vols = &find_volumes_by_policy(node => $params{node},
                                     policy => $params{policy});
   
   # Look to see if there is a volume...if so... 
   if (@vols) {
       return 1;
   }

   return 0;
}

=head2 default_export_allow_all

    This routine modifies the default export policy. The routine
    can be used enable and disable the "allow all" function of the 
    default export policy. The default behavior is to create a rule 
    to allow all clients access through NFS. Setting the enable option 
    to false when calling this method will cause the rule created by enable 
    to be removed. 

=cut
sub default_export_allow_all { 
    $Log->enter();
    my (%params) = validate_with (
                params => \@_, 
                spec => {
                    enable => { type => SCALAR,
                            optional => 1,
                             default => 1,
                    },
                    vserver => { type => SCALAR,
                             optional => 0,
                    },
                    command_interface => {
                    type     => HASHREF,
                    optional => 0,
                    },
                },
                allow_extra => 1,
             );
 
        # Use a rediculous index, so that when delete, we know which rule to aim for.
        my $ruleIndex = "234498234";
        try {
            if ($params{enable}) { 
                my ($rule) = SSQA::CMode::ExportPolicyRule::generate_random();

                $Log->trace("Configuring the default rule to allow all access");
                NACL::C::ExportPolicyRule->create(vserver => $params{vserver},
                                        command_interface => $params{command_interface},
                                               policyname => "default",
                                                 protocol => $rule->{protocol},
                                                ruleindex => $ruleIndex,
                                              clientmatch => "0.0.0.0/0",
                                                   rorule => $rule->{rorule},
                                                   rwrule => $rule->{rwrule},
                                                superuser => $rule->{superuser},);

            } else { 
                $Log->trace("Removing the allow all access from the default rule");
                NACL::C::ExportPolicyRule->delete(vserver => $params{vserver},
                                        command_interface => $params{command_interface},
                                               policyname => "default",
                                                ruleindex => $ruleIndex,);
            }

        } catch NATE::BaseException with {
               my $ex = shift;
               die config_result("Modification of the default export rule failed.");
        }

       $Log->exit(); 
       return;
}

1;
  
