# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/San/Application/IgroupMgr.pm#1 $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Utility for creating Igroup test suite
## It provides following functionality
##  * setup test suite
##  * cleanup test suite
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::Application::IgroupMgr;

# compiler directives
use strict;
use Params::Validate qw( :all );
use San::Application::Igroupconstant;
use San::SanMgr;

# inherit base class
use base qw( San::Task::IgroupHelper );

###############################################################################
# Method:    new
# Objective: Class Constructor.
# Parameter: See below
# Returns:   this class blessed object reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {'sanmgr_obj'     => '',
                'sanmgr_created' => $SanUtil::FALSE};
    @{$self->{'wwpn_alias_info'}} = ();

    # bless instance into class
    bless $self, $class;
    return $self;
}

###############################################################################
# Method:    setup
# Objective: To create SanMgr Object and create the environment using SanMgr.
# Returns:  If succeeds returns TRUE else FALSE with error message set
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup {
    my ($self, %args) = @_;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {'file' => {type     => SCALAR,
                               optional => 1},
                    'tcd_test' => {type     => OBJECT,
                                   required => 1,
                                   isa      => 'TCD'},
                    'sanmgr_obj' => {type     => OBJECT,
                                     optional => 1,
                                     isa      => 'San::SanMgr'},
                    'helper_debug_level' => {type     => SCALAR,
                                             optional => 1,
                                             default  => 0,
                                             regex    => qr/^\d+$/}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => $self->dumper_get(\%args), -msg_level => 3);

    # set debug level for IgroupMgr.pm and IgroupMgr.pm
    $self->set_debug_level($args{'helper_debug_level'});

    # if user pass the SanMgr Object then use it, else create
    # SanMgr Object and do setup
    if (defined $args{'sanmgr_obj'}) {
        $self->{'sanmgr_obj'}     = $args{'sanmgr_obj'};
        $self->{'sanmgr_created'} = $SanUtil::FALSE;
    } else {
        my $sanmgr_obj = new San::SanMgr('-file' => $args{'file'});
        if ($sanmgr_obj->error()) {
            $self->_set_err($sanmgr_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # do SanMgr::setup
        $sanmgr_obj->setup('-tcd_test' => $args{'tcd_test'});
        if ($sanmgr_obj->error()) {
            $self->_set_err($sanmgr_obj->errmsg());
            return $SanUtil::FALSE;
        }
        $self->{'sanmgr_obj'}     = $sanmgr_obj;
        $self->{'sanmgr_created'} = $SanUtil::TRUE;
    }
    return $SanUtil::TRUE;
}    # end of sub setup

###############################################################################
# Method:    cleanup
# Objective: To cleanup what setup created
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    my ($self) = @_;

    $self->_clear_err();
    my $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return $SanUtil::FALSE;
    }
    my $errstr = "";

    # cleanup what IgroupMgr::setup created...
    # destroy wwpn_alias names if created
    $self->cleanup_wwpn_alias();
    if ($self->error()) {
        $errstr .= "FCP wwpn alias cleanup failed: " . $self->errmsg() . "\n";
    }

    # if SanMgr is created by IgroupMgr::setup then call
    # SanMgr::cleanup
    if ($self->{'sanmgr_created'} == $SanUtil::TRUE) {

        # cleanup what SanMgr created
        $sanmgr_obj->cleanup();
        if ($sanmgr_obj->error()) {
            $errstr .= "SanMgr::Cleanup failed: " . $sanmgr_obj->errmsg();
        }
    }
    if ($errstr ne "") {
        $self->_set_err($errstr);
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}    # end of sub cleanup

###############################################################################
# Method:    get_sanmgr_obj
# Objective: To get the SanMgr Object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanmgr_obj {
    my ($self) = shift;

    $self->_clear_err();

    # verify the sanmgr object initialized
    unless (ref($self->{'sanmgr_obj'}) =~ /San::SanMgr/) {
        $self->_set_err("IgroupMgr::setup need to be called prior to this function");
        return {};
    }
    return $self->{'sanmgr_obj'};
}    # end of sub get_sanmgr_obj

###############################################################################
# Method:    get_vserver_objs
# Objective: To get the vserver objects that SanMgr::setup create/reuse
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vserver_objs {
    my ($self) = shift;

    my (@vserver_objs, $sanmgr_obj);

    $self->_clear_err();

    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return (wantarray()) ? () : [];
    }

    # retrieve the vserver objects created/reused by SanMgr
    @vserver_objs = $sanmgr_obj->get_objects('-type' => "vserver");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # additional check to make sure we have some objects
    # since get_objects function returns empty array without any error set
    unless (scalar(@vserver_objs)) {
        $self->_set_err("No vserver object found");
        return (wantarray()) ? () : [];
    }
    return (wantarray()) ? @vserver_objs : \@vserver_objs;
}    # end of sub get_vserver_objs

###############################################################################
# Method:    get_supported_igroup_ostype
# Objective: To get the supported igroup ostype
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_igroup_ostype {
    my ($self) = shift;

    my (@filers_obj, $sanmgr_obj, $san_filer);

    $self->_clear_err();

    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return (wantarray()) ? () : [];
    }

    @filers_obj = $sanmgr_obj->get_objects('-type' => "node");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # additional check to make sure we have some objects
    # since get_objects function returns empty array without any error set
    unless (scalar(@filers_obj)) {
        $self->_set_err("No Node object found");
        return (wantarray()) ? () : [];
    }
    $san_filer = $filers_obj[0];
    unless (ref($san_filer) =~ /San::ControllerCluster::Filer10M/) {
        $self->_set_err("Not a vaild SanFiler object");
        return (wantarray()) ? () : [];
    }
    my $cmd = "lun igroup create -ostype *";
    my @ret_values = $san_filer->execute(cmd => $cmd, -raw_data => 1);
    if ($san_filer->error()) {
        my $err_msg = $san_filer->errmsg();

        # expected output
        # ERROR: "*" is an invalid value for field "-ostype <solaris|windows|hpux|aix|linux|netware|vmware|openvms|default>"
        # extract between <>
        if ($err_msg =~ qr/.+ostype <(.+?)>.+/) {
            my $ostype = $1;

            # remove the defualt ostype from the list, if exist
            $ostype =~ s/default//;
            my @ostypes = split /\|/, $ostype;
            $self->sanlog(-msg       => "Igroup Supported OS types are: $ostype",
                          -msg_level => 3);
            return (wantarray()) ? @ostypes : \@ostypes;
        } else {
            $self->_set_err("Not expected output: $err_msg");
            return (wantarray()) ? () : [];
        }
    }
    $self->_set_err("Able to execute the command: $cmd");
    return (wantarray()) ? () : [];
}    # end of sub get_vserver_objs

###############################################################################
# Method:    create_wwpn_alias_on_vserver
# Objective: to create wwpn alias on given vserver.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_wwpn_alias_on_vserver {
    my ($self, %args) = @_;

    my ($vs_name, $vs_obj, $sanmgr_obj, $tpgroup_name);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(IGRP_VSERVER_OBJ) => {type     => OBJECT,
                                           required => 1,
                                           isa      => 'San::ControllerCluster::Filer10M::Vserver'},
                    'wwpn_array_ref' => {type => ARRAYREF, required => 1},
                    'no_of_alias'    => {
                                      type     => SCALAR,
                                      optional => 1,
                                      default  => 1,
                                      regex    => qr/^\d+$/}});
    };
    if ($@) {
        $self->_set_err($@);
        return {};
    }
    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return {};
    }

    $vs_obj  = $args{(IGRP_VSERVER_OBJ)};
    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Unable to get vserver name\nReason:\n" . $vs_obj->errmsg());
        return {};
    }

    # get the fcp object to create alias
    my $fcp_obj = $vs_obj->get_fcp_obj();
    if ($vs_obj->error()) {
        $self->_set_err("Unable to get fcp object of " . "vserver($vs_name)\nReason:\n" . $vs_obj->errmsg());
        return {};
    }
    my $wwpn_alias_info;
    $wwpn_alias_info->{'fcp_obj'} = $fcp_obj;
    my @igrp_initiators = @{$args{'wwpn_array_ref'}};
    foreach my $fcp_init (@igrp_initiators) {
        my @wwpn_alias = ();
        my $inti_name  = $fcp_init;

        # remove all the ":"
        $inti_name =~ s/://g;
        for (my $index = 1; $index <= $args{'no_of_alias'}; $index++) {
            my $wwpn_alias_name = "a_" . $index . "_" . $inti_name;
            $fcp_obj->wwpnalias_set('-wwpn'  => $fcp_init,
                                    '-alias' => $wwpn_alias_name);
            if ($fcp_obj->error()) {
                $self->_set_err("Unable to set wwpn alias on " . "vserver($vs_name)\nReason:\n" . $fcp_obj->errmsg());
                return {};
            }
            push(@wwpn_alias, $wwpn_alias_name);
        }
        $wwpn_alias_info->{'alias'}->{$fcp_init} = join ",", @wwpn_alias;
    }
    push(@{$self->{'wwpn_alias_info'}}, $wwpn_alias_info);

    if ($self->get_debug_level() >= 3) {
        my %just_alias_attr = %{$wwpn_alias_info};
        delete $just_alias_attr{'fcp_obj'};
        $self->sanlog(-msg => "create_wwpn_alias_on_vserver::Alais Info:\n" . $self->dumper_get(\%just_alias_attr, 5));
    }
    return $wwpn_alias_info;
}    # end sub create_wwpn_alias_on_vserver

###############################################################################
# Method:    cleanup_wwpn_alias
# Objective: to cleanup wwpn alias on given vserver created using
#            function create_wwpn_alias_on_vserver.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup_wwpn_alias {
    my ($self) = @_;

    $self->_clear_err();

    my $errstr                = "";
    my @wwpn_alias_info_array = @{$self->{'wwpn_alias_info'}};
    foreach my $wwpn_alias_info (@wwpn_alias_info_array) {
        my $fcp_obj = $wwpn_alias_info->{'fcp_obj'};

        # delete all the alias associated with initiator
        # use -wwpn key to remove all the alias, instead of -alias
        foreach my $init (keys %{$wwpn_alias_info->{'alias'}}) {
            $fcp_obj->wwpnalias_remove('-wwpn' => $init);
            if ($fcp_obj->error()) {
                $errstr .= $fcp_obj->errmsg() . "\n";
            }
        }
    }
    if ($errstr ne "") {
        $self->_set_err($errstr);
        return $SanUtil::FALSE;
    }
    @{$self->{'wwpn_alias_info'}} = ();
    return $SanUtil::TRUE;
}    # end sub cleanup_wwpn_alias

###############################################################################
# Method:    get_wwpn_alias_on_vserver
# Objective: to get wwpn/alias on given value on particular vserver.
#            NOTE: caller need to check for the any record found...
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_wwpn_alias_on_vserver {
    my ($self, %args) = @_;

    my ($vs_name, $vs_obj, $sanmgr_obj, $tpgroup_name);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(IGRP_VSERVER_OBJ) => {type     => OBJECT,
                                           required => 1,
                                           isa      => 'San::ControllerCluster::Filer10M::Vserver'},
                    'what'  => {type => SCALAR, required => 1, regex => qr/^(alias|wwpn)/},
                    'value' => {type => SCALAR, required => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return (wantarray()) ? () : [];
    }
    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return (wantarray()) ? () : [];
    }

    $vs_obj  = $args{(IGRP_VSERVER_OBJ)};
    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Unable to get vserver name\nReason:\n" . $vs_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # get the fcp object to create alias
    my $fcp_obj = $vs_obj->get_fcp_obj();
    if ($vs_obj->error()) {
        $self->_set_err("Unable to get fcp object of " . "vserver($vs_name)\nReason:\n" . $vs_obj->errmsg());
        return (wantarray()) ? () : [];
    }
    my %cmd_args;
    if ($args{'what'} =~ /alias/) {
        $cmd_args{'alias'} = $args{'value'};
    } else {
        $cmd_args{'wwpn'} = $args{'value'};
    }

    my @wwpn_alias = $fcp_obj->wwpnalias_get(%cmd_args);
    if ($fcp_obj->error()) {
        $self->_set_err("Unable to get wwpn alias on " . "vserver($vs_name)\nReason:\n" . $fcp_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # caller need to check for the any record found...
    return (wantarray()) ? @wwpn_alias : \@wwpn_alias;
}    # end sub get_wwpn_alias_on_vserver

###############################################################################
# Method:    get_protocol_lifs_name_in_vserver
# Objective: to get all the specified 'protocol' lifs name for the given vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_protocol_lifs_name_in_vserver {
    my ($self, %args) = @_;

    my ($vs_obj, $vs_name, $protocol_type, @lif_objs, @lif_names);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {'vserver_obj' => {type     => OBJECT,
                               required => 1,
                               isa      => 'San::ControllerCluster::Filer10M::Vserver'},
             'protocol_type' => {
                 type      => SCALAR,
                 required  => 1,
                 callbacks => {'One of the value: fcp|iscsi|mixed' => sub { $_[0] =~ /fcp|iscsi/i }
                              }}});
    };
    if ($@) {
        $self->_set_err($@);
        return undef;
    }
    $protocol_type = $args{'protocol_type'};

    # get the vserver object and the name of the vserver
    $vs_obj  = $args{'vserver_obj'};
    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get vserver name.\nReason:\n" . $vs_obj->errmsg());
        return undef;
    }

    # get all the specified protocol lif objects, this function returns
    # the fcp lifs in a given vserver, not only the lif created by setup,
    # including already exist on the given vserver
    @lif_objs = $vs_obj->get_lifs(-type => $protocol_type);
    if ($vs_obj->error()) {
        $self->_set_err(
                 "Failed to get all '$protocol_type' lifs from " . "vserver($vs_name).\nReason:\n" . $vs_obj->errmsg());
        return undef;
    }
    unless (scalar(@lif_objs)) {
        $self->_set_err("No '$protocol_type' lif found in the vserver($vs_name)");
        return undef;
    }
    foreach my $lif_obj (@lif_objs) {

        # get name of the lifs
        my $lifname = $lif_obj->get_name();
        if ($lif_obj->error()) {
            $self->_set_err(
                "Failed to get '$protocol_type' lif name from " . "vserver($vs_name).\nReason:\n" . $lif_obj->errmsg());
            return undef;
        }
        push(@lif_names, $lifname);
    }
    unless (scalar(@lif_names)) {
        $self->_set_err("No '$protocol_type' lif found in the vserver($vs_name)");
        return undef;
    }

    # cmode library can not send space separated lif name
    # so join with comma. refer burt:383323 for more details - 10/21/09
    my $lif_names = join ',', @lif_names;
    return $lif_names;
}    # end sub get_protocol_lifs_name_in_vserver

###############################################################################
# Method:    generate_iscsi_initiators
# Objective: Creates one or more random iSCSI initiators
## @arg $numNodeNames Optional: The number of initiators requested, default is 1
## @arg $prefix Optional: The iSCSI nodename prefix, default is 'iqn.1992-08.com.netapp'
# Details:   See POD documentation at the end of this file.
###############################################################################
sub generate_iscsi_initiators() {
    my ($self, $numNodeNames, $prefix) = @_;
    my @nodeNames = ();
    if (!defined($prefix)) {
        $prefix = 'iqn.1992-08.com.netapp';
    }
    $numNodeNames = 1 unless defined $numNodeNames;

    for (my $n = 0; $n < $numNodeNames; $n++) {
        my @parts = ();
        my ($random, $number);
        push(@parts, $n);
        for (my $x = 0; $x < 2; $x++) {
            $random = int(rand(999)) + 1;
            push(@parts, $random);
        }
        my $nodename = "$prefix." . join(".", @parts);
        push(@nodeNames, $nodename);
    }
    return (wantarray) ? @nodeNames : $nodeNames[0];
}    # end of sub generate_iscsi_initiators

###############################################################################
# Method:    generate_fcp_initiators
# Objective: Creates one or more random FCP initiators
## @arg $numNodeNames Optional: The number of initiators requested, default is 1
# Details:   See POD documentation at the end of this file.
###############################################################################
sub generate_fcp_initiators() {
    my ($self, $numNodeNames) = @_;
    my @nodeNames = ();
    $numNodeNames = 1 unless defined $numNodeNames;

    # nodename must be conformance
    # 50:0a:09:80:8X:XX:XX:XX
    for (my $n = 0; $n < $numNodeNames; $n++) {
        my @hexnumbers = ();
        @hexnumbers = split(/:/, "50:0a:09:80");    # fcp conformance
        push(@hexnumbers, int(rand(10)) + 80);      # random between 80-89
        my ($random, $number);
        for (my $x = 5; $x < 8; $x++) {
            $random = int(rand(255)) + 1;
            $number = sprintf("%02x", $random);
            push(@hexnumbers, $number);
        }
        my $nodename = join(":", @hexnumbers);
        push(@nodeNames, $nodename);
    }
    return (wantarray) ? @nodeNames : $nodeNames[0];
}    # end of sub generate_fcp_initiators

###############################################################################
#                        Private Functions
###############################################################################
###############################################################################

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

IgroupMgr.pm

=head1 Author

Muthukumaran Govindarajan (muthug@netapp.com)

=head1 Description

IgroupMgr.pm provides an common API interface to test igroup functionality with the main intent to avoid repeated code from thpl level.
IgroupMgr is inherited from IgroupHelper.pm, which has common igroup utility function. IgroupMgr.pm uses SanMgr and San Library. Apart
from SanMgr::setup, it will do additional setup and cleanup needed for igroup testing.

Since this package base class is San::Task::IgroupHelper, From IgroupMgr object, we can access the San::Task::IgroupHelper function directly.


=head1 See Also

Package inherits from the San::Task::IgroupHelper module. See these San::Task::IgroupHelper POD
for further interface detail.

B<San::Task::IgroupHelper>

B<San::SanMgr>

=head1 Attributes

B<sanmgr_obj> 

Holds the SanMgr reference. IgroupMgr creates SanMgr object or user can pass the reference after setup.

=head1 Revision History

Jan-18-2010 muthug@netapp.com: Original template created
Jul-13-2010 muthug@netapp.com: Modified to use SanMgr

=head1 Public Methods

=head2 B<new()>

Class Constructor method

=over 4

=item Inputs

None

=item Outputs

Returns a IgroupMgr blessed object reference.

=item Examples

   #
   # Create a IgroupMgr object
   #
   my $igrpmgr_obj = new IgroupMgr();
   if ($igrpmgr_obj->error()) {
      my $err_msg = $igrpmgr_obj->errmsg();
      $Test->nlog("Failed to instantiate IgroupMgr object:\nReason: $err_msg");
      return $TCD::FAIL;
   }

=back

=head2 B<setup(%args)>

Create necessary resources on filers and clients using SanMgr library.  It takes same parameters as SanMgr::setup.
In addition it takes following parameters. Refer SanMgr Pod for SanMgr::setup parameters.

=over 4

=item Inputs

=item * I<all inputs that SanMgr::setup takes>

=item * I<-file> <optional> 

Name of the (param definition) xml file with full path. if xml name not specified, then it will default to script name(with script path) with .xml

Ex: If script name ~/san/feature/igroup/igroup_basic_func_cm.thpl and if you do not pass the xml file, it will search the file from
~/san/feature/igroup/igroup_basic_func_cm.xml.

=item * I<-debug_level> <optional>

Level of debug to set it for San Library

=item * I<-sanmgr_obj> <optional> 

Reference to SanMgr object. Object has to be created and setup would have been called on SanMgr.
If this param is not passed, then IgroupMgr::setup will create SanMgr and do the setup.

=item Outputs

Returns TRUE resource are created successful. If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

      #
      # Instantiate IgroupMgr Object 
            
      $igrpmgr_obj = new IgroupMgr();
      if ($igrpmgr_obj->error()) {
         my $err_msg = $igrpmgr_obj->errmsg();
         $Test->nlog("Failed to instantiate IgroupMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $igrpmgr_obj->setup("-file" => full_path/thplname.xml);
      if ($igrpmgr_obj->error()) {
         $Test->nlog($igrpmgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<cleanup()>

To Cleanup what IgroupMgr::setup created. If SanMgr created outside of IgroupMgr, then it wont cleanup up the SanMgr.

=over 4

=item Inputs

None

=item Outputs

Returns TRUE resource are created successful. If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

      #
      # Instantiate IgroupMgr Object 
            
      $igrpmgr_obj = new IgroupMgr();
      if ($igrpmgr_obj->error) {
         my $err_msg = $igrpmgr_obj->errmsg();
         $Test->nlog("Failed to instantiate IgroupMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $igrpmgr_obj->setup("-file" => full_path/thplname.xml);
      if ($igrpmgr_obj->error()) {
         $Test->nlog($igrpmgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      ...
      #
      # Call cleanup 
      #
      $igrpmgr_obj->cleanup();
      if ($igrpmgr_obj->error()) {
         $Test->nlog($igrpmgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_sanmgr_obj()>

To get the SanMgr objects. It will return what IgroupMgr::setup created or user passed.

NOTE: This is called by IgroupMgr::Setup, mostly no need to call it.

=over 4

=item Inputs

None

=item Outputs

Returns a reference to SanMgr object if successful. If an error occurred the error flag is set, errmsg is set to the failure reason,
and {} (empty hash reference) is passed back as the return value.

=item Examples

      #
      # Instantiate IgroupMgr Object (see above examples) 
            
      $igrpmgr_obj = new IgroupMgr();
      ...
      ...
      ...
      #
      # get the SanMgr Object. 
      #
      my $SanMgr_obj = $igrpmgr_obj->get_sanmgr_obj();
      if ($igrpmgr_obj->error()) {
         $Test->nlog($igrpmgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_vserver_objs()>

To get the vserver objects that SanMgr::setup created/reused

NOTE;

This functions returns the vserver object which is/are created or reused by San::SanMgr:setup.
You can get all the vservers in the controller by using SanFiler::get_vservers().

=over 4

=item Inputs

None

=item Outputs

Returns a reference to array or array of vserver objects if successful. If an error occurred the error flag is set, errmsg is set to the failure reason,
and empty array ([] or ()) is passed back as the return value.

=item Examples

      #
      # Instantiate IgroupMgr Object (see above examples) 
            
      $igrpmgr_obj = new IgroupMgr();
      ...
      ...
      ...
      #
      # get the vserver objects created by setup 
      #
      my @Vserver_objects = $igrpmgr_obj->get_vserver_objs();
      if ($igrpmgr_obj->error()) {
         $Test->nlog($igrpmgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<create_wwpn_alias_on_vserver(%args)>

To create wwpn alias on given vserver of specified wwpns.

=over 4

=item Inputs

=item * I<-vserver_obj> <required> 

Reference to vserver object

=item * I<-wwpn_array_ref> <required>

Array reference to wwpn names. 

=item Outputs

Returns a reference to hash of wwpn alias name if successful. If an error occurred the error flag is set, errmsg is set to the failure reason,
and empty hash reference ({}) is passed back as the return value.

if sucess hash looks like

$alias_info->{'fcp_obj'}
$alias_info->{'alias'}->{'fcp_initiator_name_1'} = "a1,a2,a3";
$alias_info->{'alias'}->{'<fcp_initiator_name_2'} = "b1,b2";
$alias_info->{'alias'}->{'<fcp_initiator_name_3'} = "c1";

=item Examples

    #
    # Instantiate IgroupMgr Object (see above examples) 
        
    $igrpmgr_obj = new IgroupMgr();
    ...
    ...
    ...
    # get the vserver objects created by setup 
    my $vs_obj = ...;
    
    my @igrp_initiators_alias = $igrpmgr_obj->generate_fcp_initiators(4);
    my $alias_info =
    $igrpmgr_obj->create_wwpn_alias_on_vserver('-vserever_obj' => $vs_obj,
                                               'wwpn_array_ref'   => \@igrp_fcp_initiators);
    if ($igrpmgr_obj->error()) {
        $Test->nlog($igrpmgr_obj->errmsg());
        return $TCD::FAIL;
    }
    foreach my $init (keys %{$alias_info->{'alias'}}) {
        # comma sperated alias list
        my $alias = $alias_info->{'alias'}->{$init};
    }
    ...
    ...

=back

=head2 B<generate_iscsi_initiators(...)>

Creates one or more random iSCSI initiators.

=over 4

=item Inputs

=item * I<1st param, Number of Node Names> <Optional, default: 1> 

Integer value to indicates number of iSCSI initiators to create. 

=item * I<2nd param, prefix for the iSCSI name> <Optional, default: 'iqn.1992-08.com.netapp'>

Prefix name of iSCSI node name, should comply with iqn standards.

=item Outputs

Returns a reference to array or array of iSCSI initiators name.

=item Examples

    #
    # Instantiate IgroupMgr Object (see above examples) 
        
    $igrpmgr_obj = new IgroupMgr();
    ...
    ...
    my @igrp_iscsi_initiators = $igrpmgr_obj->generate_iscsi_initiators(4);
    ...
    ...

=back


=head2 B<generate_fcp_initiators(...)>

Creates one or more random FCP initiators. Initiators will be conformance 50:0a:09:80:8X:XX:XX:XX

=over 4

=item Inputs

=item * I<1st param, Number of initiators> <Optional, default: 1> 

Integer value to indicates number of FCP initiators to generate. 

=item Outputs

Returns a reference to array or array of FCP initiators name.

=item Examples

    #
    # Instantiate IgroupMgr Object (see above examples) 
        
    $igrpmgr_obj = new IgroupMgr();
    ...
    ...
    my @igrp_iscsi_initiators = $igrpmgr_obj->generate_fcp_initiators(4);
    ...
    ...

=back

=cut
