# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/San/Application/LunMgr.pm#1 $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Utility for creating Lun test suite
## It provides following functionality
##  * setup test suite
##  * get() methods to return objects of various types (vserver, volume, etc.)
##  * cleanup test suite
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::Application::LunMgr;

# compiler directives
use strict;

use File::Basename;    # fileparse - for splitting dir,file,ext
use San::SanMgr;
use constant LUN_TYPES_SUPPORTED => qw(vmware hyper_v windows_2008 windows_gpt windows linux xen solaris solaris_efi hpux aix netware openvms);

# inherit base class
use base qw( San::Task::LunHelper );

###############################################################################
# Method:    new
# Objective: Class Constructor.
# Parameter: See below
# Returns:   this class blessed object reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {'sanmgr_obj'     => '',
                'sanmgr_created' => $SanUtil::FALSE,
                'cserver_obj'    => {},
                'node_objs'    => [],
                'qtree_count'    => 0,
                'LunAttrsArray'  => []};

    # bless instance into class
    bless $self, $class;
    return $self;
}

###############################################################################
# Method:    setup
# Objective: To create SanMgr Object and create the environment using SanMgr.
# Returns:  If succeeds returns TRUE else FALSE with error message set
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup {
    my ($self, %args) = @_;

    my $debug_level = 0;
    my ($san_setup_xml_file,
        $sanmgr_obj,
        @vserver_objs,
        @igroup_objs,
        $vol_state,
        $aggr_state,
        $tcd_test);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /debug_level/) {
            if (!length($debug_level = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
        } elsif ($key =~ /file/) {
            if (!length($san_setup_xml_file = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return $SanUtil::FALSE;
            }
        } elsif ($key =~ /sanmgr_obj/) {
            $sanmgr_obj = $value;
            if (ref($sanmgr_obj) !~ /San::SanMgr/) {
                $self->_set_err("invalid value for \"-$key\": $value. Need reference to San::SanMgr");
                return $SanUtil::FALSE;
            }
        } elsif ($key =~ /tcd_test/) {
            $tcd_test = $value;
            if (ref($tcd_test) !~ /TCD/) {
                $self->_set_err("invalid value for \"-$key\": $value. Need reference to TCD");
                return $SanUtil::FALSE;
            }
        } elsif ($key =~ /qtree_count/) {
            $self->{'qtree_count'} = $value;
            if ($self->{'qtree_count'} !~ /[0-9]/ ) {
                $self->_set_err("invalid value for \"-$key\": $value. qtree_count should be set to value between 0 and 9");
                return $SanUtil::FALSE;
            }
        } elsif ($key =~ /vol_state/) {
            $vol_state = $value;
            if ($vol_state !~ /ONLINE/) {
                $self->_set_err("invalid value for \"-$key\": $value. vol_state must be one of <...>");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err("invalid \"-$key\": passed");
            return $SanUtil::FALSE;
        }
    }

    # set debug level for LunMgr.pm
    $self->set_debug_level($debug_level);

    # display argument hash if debug_level is high
    $self->sanlog(-msg => $self->dumper_get(\%args), -msg_level => 3);

    # if user pass the SanMgr Object then use it, else create
    # SanMgr Object and do setup
    unless (ref($sanmgr_obj) =~ /San::SanMgr/) {
        $sanmgr_obj = new San::SanMgr('-file' => _get_xml_filename($san_setup_xml_file));
        if ($sanmgr_obj->error()) {
            $self->_set_err($sanmgr_obj->errmsg());
            return $SanUtil::FALSE;
        }
        $self->{'sanmgr_created'} = $SanUtil::TRUE;
        $self->{'sanmgr_obj'} = $sanmgr_obj;
    }
    # do SanMgr::setup, this will create vservers/volumes/luns/etc based on XML file
    $sanmgr_obj->setup(-debug_level   => $debug_level,
                       -cmd_interface => "cli",
                       -tcd_test      => $tcd_test);
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # get cserver object
    my @cluster_objs = $sanmgr_obj->get_objects(-type => "cluster");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $self->{'cserver_obj'} = $cluster_objs[0]->get_cserver_obj();
    if ($cluster_objs[0]->error()) {
        $self->_set_err($cluster_objs[0]->errmsg());
        return $SanUtil::FALSE;
    }

    # get node objects
    $self->{'node_objs'} = $sanmgr_obj->get_objects(-type => "node");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # FIXME: unnecessary once burt 437466 is implemented
    if ($self->{'qtree_count'} >= 1) {
        $self->_create_qtrees();
        if ($self->error()) {  # error msg is already set in _create_qtrees()
            return $SanUtil::FALSE;
        }
    }

    # create the LUNs to be used for test, if specified
    #$self->create_tc_luns();
    #if ($self->error()) {
    #    return $SanUtil::FALSE;
    #}

    # validate the setup -
    # TODO - following code is commented out for later implementation.
    # just a place holder for now....

    # $self->validate();
    # if ($self->error()) {

    #    # error set, just return it
    #    return $SanUtil::FALSE;
    # }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    create_tc_luns
# Objective: To create the luns which are needed for the test cases
# Details:
###############################################################################
sub create_tc_luns {
    my ($self, %args) = @_;

    $self->_clear_err();

    my (@volumes, $lun_name_prefix, $lun_os_type, $space_resv, @lun_os_types, $in_qtree,
        $volume_name, $from_file, $unmapped_luns, $mapped_luns, @container_objs, $volume_obj);
    my $luns_per_ostype = 1;
    my $lun_size = 40 * 1024 * 1024;
    my $verify_serial = 'true';
    $from_file = 0;
    $unmapped_luns = "true";
    $mapped_luns = "true";
    # Need to update this method to create LUNs of various "types"
    # create by size, create from file, clone, restore

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /-volumes/) {
            if (ref($value) !~ /ARRAY/) {
                $self->sanlog(-msg => "invalid value for \"$key\": "
                            . $value
                            . "Need ARRAY reference of volume information");
                return ();
            }
            @volumes = @$value;
        } elsif ($key =~ /-luns_per_ostype/) {
            if (!length($luns_per_ostype = $value)) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } elsif ($key =~ /-lun_size/) {
            if (!length($lun_size = $value)) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } elsif ($key =~ /-lun_name_prefix/) {
            if (!length($lun_name_prefix = $value)) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } elsif ($key =~ /-lun_os_type/) {
            if (!length($lun_os_type = $value)) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } elsif ($key =~ /-option/) {
            if (!length($space_resv = $value)) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } elsif ($key =~ /-from_file/) {
            if (!length($from_file = $value)) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } elsif ($key =~ /-unmapped_luns/) {
            if (!length($unmapped_luns = $value)) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } elsif ($key =~ /-mapped_luns/) {
            if (!length($mapped_luns = $value)) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } elsif ($key =~ /-volume_name/) {
            if (!length($volume_name = $value)) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } elsif ( $key =~ /in_qtree/ ) {
            if ( !length( $in_qtree = $value ) ) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } elsif ( $key =~ /verify_serial/ ) {
            if ( !length( $verify_serial = $value ) ) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return ();
            }
        } else {
            $self->sanlog(-msg => "invalid \"$key\": passed. Valid parameter: volumes|luns_per_ostype|lun_size|lun_name_prefix|lun_os_type|option|volume_name");
            return ();
        }
    }

    unless (defined($volumes[0])) {
        $self->_set_err("mandatory parameter \"-volumes\" is needed");
        return $SanUtil::FALSE;
    }

    unless (defined($lun_name_prefix)) {
        $self->_set_err("mandatory parameter \"-lun_name_prefix\" is needed");
        return $SanUtil::FALSE;
    }

    unless ($unmapped_luns =~ /true/ or $mapped_luns =~ /true/) {
        $self->_set_err("at least one of \"-unmapped\" or  \"-mapped\" must be set to true");
        return $SanUtil::FALSE;
    }

    my $lun_count = 0;
    
    # if an os type was specified, only create luns of that os type
    # otherwise, create luns of all OS types supported
    if ($lun_os_type) {
        push @lun_os_types, $lun_os_type;
    } else {
        foreach my $lun_os (LUN_TYPES_SUPPORTED) {
            push @lun_os_types, $lun_os;
        }
    }

    # combine the requested lun containers (i.e., volumes and qtrees)
    @container_objs = @volumes;
    unless ( $in_qtree eq 'false' ) {
        my $qtree_objs = $self->get_qtree_objs();
        if (scalar(@{$qtree_objs})) {
            push @container_objs, @{$qtree_objs};
        }
    }

    # Create an unmapped lun of each OS type in each volume/qtree and each vserver
    # and a mapped lun of each OS type in each volume/qtree and each vserver
    # TBD: should we be creating offline LUNs also??
    foreach my $container_obj (@container_objs) {
        if (ref($container_obj) =~
               /San::ControllerCluster::Filer10M::Vserver::Volume::Qtree/) {
            $volume_obj = $container_obj->get_parent_obj();
        } else {
            $volume_obj = $container_obj;
        }
        if ($volume_name) {
            next unless ($volume_obj->get_name() eq $volume_name);
        }
        foreach my $os_type (@lun_os_types) {

            for (my $i = 1; $i <= $luns_per_ostype; $i++) {
                # create the unmapped lun if we were asked to do so
                if ($unmapped_luns =~ /true/) {
                    $lun_count++;
                    # create the unmapped lun
                    my $lun_name = $lun_name_prefix . "_unmapped_" . $os_type . "_" . $lun_count;
                    my %cmd_args;
                    $cmd_args{'-container_obj'} = $container_obj;
                    $cmd_args{'-lun_name'}   = $lun_name;
                    $cmd_args{'-size'}       = $lun_size;
                    $cmd_args{'-ostype'}     = $os_type;
                    $cmd_args{'-verify_serial'} = $verify_serial;
                    if ($space_resv) {
                        $cmd_args{"-option"} = $space_resv;
                    }
                   $cmd_args{"-file-path"} = "file"."_"."$lun_name" if($from_file == 1);
                    
                    my $lun_attr_expected =
                      $self->create_lun(%cmd_args);
                    if ($self->error()) {
                        # function will destroy the lun(s) and log the failure
                        #return _cleanup_tc_luns('-err_msg'       => $self->errmsg());
                        return ();
                    }
                    # verify the lun created
                    $self->verify_lun_attributes('-lun_attr_info' => $lun_attr_expected);
                    if ($self->error()) {

                        # this function will return TCD::FAIL
                        #return _cleanup_tc_luns('-err_msg'       => $self->errmsg());
                        return ();
                    }
                    push(@{$self->{LunAttrsArray}}, $lun_attr_expected);
                }
                # create the mapped lun if we were asked to do so
                if ($mapped_luns =~ /true/) {
                    $lun_count++;
                    my $lun_name = $lun_name_prefix . "_mapped_" . $os_type . $lun_count;
                    my %cmd_args;
                    $cmd_args{'-container_obj'} = $container_obj;
                    $cmd_args{'-lun_name'}   = $lun_name;
                    $cmd_args{'-size'}       = $lun_size;
                    $cmd_args{'-ostype'}     = $os_type;
                    $cmd_args{'-verify_serial'} = $verify_serial;
                    if ($space_resv) {
                        $cmd_args{"-option"} = $space_resv;
                    }
                   $cmd_args{"-file-path"} = "file"."_"."$lun_name" if($from_file == 1);
                    my $lun_attr_expected =
                      $self->create_lun(%cmd_args);
                    if ($self->error()) {
                        # function will destroy the lun(s) and log the failure
                        #return _cleanup_tc_luns('-err_msg'      => $self->errmsg());
                        return ();
                    }
                    my $igroup_os_type;
                    # there are a couple of lun os types that would map to an igroup
                    # type of windows (windows, windows_gpt, etc)
                    if ($os_type =~ /windows/) {
                        $igroup_os_type = "windows";
                    } elsif ($os_type =~ /hyper_v/) {
                        $igroup_os_type = "vmware";
                    } elsif ($os_type =~ /xen/) {
                        $igroup_os_type = "vmware";
                    } elsif ($os_type =~ /solaris/) {
                        $igroup_os_type = "solaris";
                    } else {
                        $igroup_os_type = $os_type;
                    }
                    
                    $igroup_os_type = "windows";
                    
                    my $vs_obj = $volume_obj->get_parent_obj();
                    my $vs_name = $vs_obj->get_name();
                    my @igroup_objs = $self->get_igroup_objs('-vserver_name'        => $vs_name,
                                                             '-os_type'             => $igroup_os_type);
                    if ($self->error()) {
                         my $err = "Unable to get igroup(s) on vserver: $vs_name\nReason:\n"
                                   . $self->errmsg();
                        $self->sanlog(-msg => $err);
                        #return _cleanup_tc_luns('-err_msg'      => $LUNMgr_obj->errmsg());
                        return ();
                    }
                    $self->map_lun_to_igroup(-lun_attr_info  => $lun_attr_expected,
                                             -igroup_obj     => $igroup_objs[0]);
                    if ($self->error()) {
                         my $err = "Unable to get igroup(s) on vserver: $vs_name\nReason:\n"
                                   . $self->errmsg();
                        $self->sanlog(-msg => $err);
                        #return _cleanup_tc_luns('-err_msg'      => $LUNMgr_obj->errmsg());
                        return ();
                    }

                    # verify the lun created
                    $self->verify_lun_attributes('-lun_attr_info' => $lun_attr_expected);
                    if ($self->error()) {

                        # this function will return TCD::FAIL
                        #return _cleanup_tc_luns('-err_msg'      => $LUNMgr_obj->errmsg());
                        return ();
                    }
                    # map_lun_to_igroup will add this lun to the array...
                    # push(@{$self->{LunAttrsArray}}, $lun_attr_expected);
                } # end of if ($mapped_luns =~ /true/)
            }    # end of for (my $i=1; $i <= $LunsPerOsType; $i++)
        }   # end of for (my $lun_os_type)
    }    # end of foreach my $container (@container_objs)

    return (wantarray()) ? @{$self->{LunAttrsArray}} : \@{$self->{LunAttrsArray}};
}

###############################################################################
# Method:    cleanup_tc_luns
# Objective: To destroy the luns which are created by the test cases
# Details:   this can be called with error message.
###############################################################################
sub cleanup_tc_luns {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $err_str = "";
    my @lun_objs_array;
    my $force;

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /-err_msg/) {
            if (!length($err_str = $value)) {
                $self->sanlog(-msg => "invalid value for \"$key\": $value");
                return $SanUtil::FALSE;
            }
        } else {
            $self->sanlog(-msg => "invalid \"$key\": passed. Valid parameter: -err_msg");
            return $SanUtil::FALSE;
        }
    }

    # if it is an empty array and no error msg passed, simply return PASS
    unless (scalar(@{$self->{LunAttrsArray}})) {
        if (length $err_str) {
            $self->sanlog(-msg => $err_str);
            return $SanUtil::FALSE;
        }
        return $SanUtil::TRUE;
    }

    foreach my $lun_info (@{$self->{LunAttrsArray}}) {
        my $lun_obj = $lun_info->{'lun_obj'};
        my $lun_name = $lun_obj->get_name();
        if ($lun_obj->is_mapped()) {
            my @igroups = $lun_obj->get_maps();
            if ($lun_obj->error()) {
                if ($err_str ne "") {
                    $err_str .= "\n";
                }
                $err_str .= "Failed to get maps on lun($lun_name)\nReason: " . $lun_obj->errmsg();
            }
            foreach my $igroup (@igroups) {
                $self->sanlog(-msg => "about to unmap from $igroup");
                $lun_obj->unmap($igroup);
                if ($lun_obj->error()) {
                    if ($err_str ne "") {
                        $err_str .= "\n";
                    }
                    $err_str .= "Failed to unmap lun($lun_name)\nReason: " . $lun_obj->errmsg();
                }   # end if $lun_obj->error()
            }   # end foreach my $igroup (@igroups)
        }    # end if ($lun_obj->is_mapped())
    }    # end foreach my $lun_info (@lun_objs_array)

    # remove all the lun which are created by this thpl test cases
    foreach my $lun_info (@{$self->{LunAttrsArray}}) {
        my $lun_obj  = $lun_info->{'lun_obj'};
        my $lun_name = $lun_obj->get_name();
        
        $self->sanlog(-msg => "about to call destroy on $lun_name");
        $lun_obj->destroy();
        if ($lun_obj->error()) {
            if ($err_str ne "") {
                $err_str .= "\n";
            }
            $err_str .= "Failed to destroy lun($lun_name)\nReason: " . $lun_obj->errmsg();
        }
        if (defined $lun_info->{'file_path'}) {

            my $container_obj = $lun_obj->get_parent_obj();
            $container_obj->remove_file('-from_file' => $lun_info->{'file_path'});
            if($container_obj->error()) {
                if ($err_str ne "") {
                    $err_str .= "\n";
                }
                $err_str .= "Failed to destroy file($lun_info->{'file_path'})\nReason: " . $container_obj->errmsg();                
            }
        }
    }

    if ($err_str ne "") {
        $self->sanlog(-msg => $err_str);
        return $SanUtil::FALSE;
    }

    @{$self->{LunAttrsArray}} = ();
    return $SanUtil::TRUE;
}

###############################################################################
# Method:    cleanup
# Objective: To cleanup what setup created
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    my ($self, %args) = @_;

    my ($sanmgr_obj);

    $self->_clear_err();
    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return $SanUtil::FALSE;
    }

    # FIXME: unnecessary once burt 437466 is implemented
    if ($self->{'qtree_count'} >= 1) {
        $self->_delete_qtrees();
        if ($self->error()) {
            return $SanUtil::FALSE;
        }
    }

    # if SanMgr is created by LunMgr::setup then call
    # SanMgr::cleanup
    if ($self->{'sanmgr_created'} == $SanUtil::TRUE) {

        # cleanup what SanMgr created
        $sanmgr_obj->cleanup();
        if ($sanmgr_obj->error()) {
            $self->_set_err($sanmgr_obj->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    validate (right now it is place holder for later implementation)
# Objective: To validate the setup for the test suite to run
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate {
    my ($self) = @_;

    my $sanmgr_obj;

    # Need to validate following
    # 1. requested vserver presents
    # 2. iscsi and fcp object created and service enabled
    #    if not,create/enabled it, during cleanup, put back to original state
    # 3. enable required license
    # 4. requested iscsi and fcp lifs presents
    # 5. available fcp target port to create fcp lif
    # 6. available iscsi ports to create iscsi lif

    $self->_clear_err();

    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return $SanUtil::FALSE;
    }

    # when SanMgr support this functionality, implement it
    #     $sanmgr_obj->validate();
    #     if ($sanmgr_obj->error()) {
    #         $self->_set_err($sanmgr_obj->errmsg());
    #         return $SanUtil::FALSE;
    #     }
    return $SanUtil::TRUE;
}

###############################################################################
# Method:    get_sanmgr_obj
# Objective: To get the SanMgr Object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanmgr_obj {
    my ($self) = @_;

    $self->_clear_err();

    # verify the sanmgr object initialized
    unless (ref($self->{'sanmgr_obj'}) =~ /San::SanMgr/) {
        $self->_set_err("LunMgr::setup need to be called prior to this function");
        return {};
    }
    return $self->{'sanmgr_obj'};
}

###############################################################################
# Method:    get_vserver_objs
# Objective: To get the vserver objects that SanMgr::setup create/reuse
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vserver_objs {
    my ($self) = @_;

    my (@vserver_objs, $sanmgr_obj);

    $self->_clear_err();

    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return (wantarray()) ? () : [];
    }

    # retrieve the vserver objects created/reused by SanMgr
    @vserver_objs = $sanmgr_obj->get_objects('-type' => "vserver");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # additional check to make sure we have some objects
    # since get_objects function returns empty array without any error set
    unless (scalar(@vserver_objs)) {
        $self->_set_err("No vserver object found");
        return (wantarray()) ? () : [];
    }
    return (wantarray()) ? @vserver_objs : \@vserver_objs;
}

###############################################################################
# Method:    get_aggr_objs
# Objective: To get the aggregate objects that SanMgr::setup create/reuse
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_aggr_objs {
    my ($self) = @_;

    my (@aggr_objs, $sanmgr_obj);

    $self->_clear_err();

    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return (wantarray()) ? () : [];
    }

    # retrieve the aggr objects created/reused by SanMgr
    @aggr_objs = $sanmgr_obj->get_objects(-type => "aggregate");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # additional check to make sure we have some objects
    # since get_objects function returns empty array without any error set
    unless (scalar(@aggr_objs)) {
        $self->_set_err("No aggr object found");
        return (wantarray()) ? () : [];
    }
    return (wantarray()) ? @aggr_objs : \@aggr_objs;
}

###############################################################################
# Method:    get_volume_objs
# Objective: To get the volume objects that SanMgr::setup create/reuse
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_volume_objs {
    my ($self, %args) = @_;

    my (@volume_objs, $sanmgr_obj, $vs_name, @sanmgr_volume_objs);

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /vserver_name/) {
            if (!length($vs_name = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return (wantarray()) ? () : [];
            }
        } else {
            $self->_set_err("invalid \"-$key\": passed.\nValid parameter: -protocol_type|-vserver_name");
            return (wantarray()) ? () : [];
        }
    }

    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return (wantarray()) ? () : [];
    }

    # retrieve the vserver objects created/reused by SanMgr
    @sanmgr_volume_objs = $sanmgr_obj->get_objects(-type => "volume");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # additional check to make sure we have some objects
    # since get_objects function returns empty array without any error set
    unless (scalar(@sanmgr_volume_objs)) {
        $self->_set_err("No volume object found");
        return (wantarray()) ? () : [];
    }

    # go through all the volume objects and
    # filter it out based on user request
    foreach my $volume_obj (@sanmgr_volume_objs) {
        my $volume_name = $volume_obj->get_name();
        if ($volume_obj->error()) {
            $self->_set_err("Unable to get the name of the volume\nReason: " . $volume_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        # if user wants the volume objects in specified vserver name then
        # get the vserver name from igroup object.
        if ($vs_name) {

            # parent from volume object should be vserver object
            my $vs_obj = $volume_obj->get_parent_obj();
            if (ref($vs_obj) !~ /San::ControllerCluster::Filer10M::Vserver/) {
                $self->_set_err("Unable to get parent (vserver) object of volume");
                return (wantarray()) ? () : [];
            }
            my $volume_vserver_name = $vs_obj->get_name();
            if ($vs_obj->error()) {
                $self->_set_err(  "Unable to get the name of the vserver "
                                . "in the volume($volume_name)\nReason: "
                                . $vs_obj->errmsg());
                return (wantarray()) ? () : [];
            }
            next unless ($volume_vserver_name eq $vs_name);
        }
        push(@volume_objs, $volume_obj);
        $self->sanlog(-msg => "Found volume object $volume_name", -msg_level => 3);
    }

    # make sure we have at least one volume object to return
    # else fail with error message.
    $self->sanlog(-msg => "Found " . scalar(@volume_objs) . " volumes", -msg_level => 3);
    unless (scalar(@volume_objs)) {
        my $err_str = "No volume object found";
        if ($vs_name) {
            $err_str .= " in vserver: $vs_name";
        }
        $self->_set_err($err_str);
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @volume_objs : \@volume_objs;
}

###############################################################################
# Method:    get_igroup_objs
# Objective: To get the igroup objects that SanMgr::setup create/reuse
#            it will return all igroup or specific protocol/ostype igroup objects
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_igroup_objs {
    my ($self, %args) = @_;

    my ($vs_name, @igroup_objs, @sanmgr_igroup_objs, $sanmgr_obj);
    my $protocol_type = "all";
    my $os_type = "all";

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /protocol_type/) {
            if (!length($protocol_type = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }
            if ($protocol_type !~ /all|mixed|fcp|iscsi/) {
                $self->_set_err(
                            "invalid value for \"-$key\": $value\nValid value for protocol_type : all|mixed|fcp|iscsi");
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /vserver_name/) {
            if (!length($vs_name = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /os_type/) {
            if (!length($os_type = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return (wantarray()) ? () : [];
            }
        } else {
            $self->_set_err("invalid \"-$key\": passed.\nValid parameter: -protocol_type|-vserver_name");
            return (wantarray()) ? () : [];
        }
    }

    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return (wantarray()) ? () : [];
    }

    # get all the igroup objects created/reused by SanMgr
    @sanmgr_igroup_objs = $sanmgr_obj->get_objects(-type => "igroup");
    if ($sanmgr_obj->error()) {
        $self->_set_err("Unable to get the igroup objects\nReason: " . $sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # additional check to make sure we have some objects
    # since get_objects function returns empty array without any error set
    unless (scalar(@sanmgr_igroup_objs)) {
        $self->_set_err("No igroup object found in the SanMgr");
        return (wantarray()) ? () : [];
    }

    # if vserver name, protocol_type, and os_type are not specified
    # then simply return all objects
    if (!defined($vs_name) and $protocol_type =~ /all/ and $os_type =~ /all/) {
        return (wantarray()) ? @sanmgr_igroup_objs : \@sanmgr_igroup_objs;
    }

    # go through all the igroup objects and
    # filter it out based on user request
    foreach my $igroup_obj (@sanmgr_igroup_objs) {
        my $igroup_name = $igroup_obj->get_name();
        if ($igroup_obj->error()) {
            $self->_set_err("Unable to get the name of the igroup\nReason: " . $igroup_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        # if user wants the igroup objects in specified vserver name then
        # get the vserver name from igroup object.
        if ($vs_name) {

            # parent from igroup object should be vserver object
            my $vs_obj = $igroup_obj->get_parent_obj();
            if (ref($vs_obj) !~ /San::ControllerCluster::Filer10M::Vserver/) {
                $self->_set_err("Unable to get parent (vserver) object of igroup");
                return (wantarray()) ? () : [];
            }
            my $igroup_vserver_name = $vs_obj->get_name();
            if ($vs_obj->error()) {
                $self->_set_err(  "Unable to get the name of the vserver "
                                . "in the igroup($igroup_name)\nReason: "
                                . $vs_obj->errmsg());
                return (wantarray()) ? () : [];
            }
            next unless ($igroup_vserver_name eq $vs_name);
        }

        # now filter with protocol type
        my $pro_type = $igroup_obj->get_protocol_type();
        if ($igroup_obj->error()) {
            $self->_set_err(
                         "Unable to get the protocol type from igroup($igroup_name)\nReason: " . $igroup_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        # now filter with os type
        my $os = $igroup_obj->get_ostype();
        if ($igroup_obj->error()) {
            $self->_set_err(
                         "Unable to get the os type from igroup($igroup_name)\nReason: " . $igroup_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        # return if requested protocol type matches or
        # if protocol type == "all"
        if ((($pro_type eq $protocol_type) or ($protocol_type =~ /all/)) and
             (($os eq $os_type) or ($os_type =~ /all/))) {
            push(@igroup_objs, $igroup_obj);
            $self->sanlog(-msg => "Found igroup object $igroup_name", -msg_level => 3);
        }
    }

    # make sure we have at least one igroup object to return
    # else fail with error message.
    $self->sanlog(-msg => "Found " . scalar(@igroup_objs) . " $protocol_type $os_type igroups", -msg_level => 3);
    unless (scalar(@igroup_objs)) {
        my $err_str = "No '$protocol_type' '$os_type' igroup object found";
        if ($vs_name) {
            $err_str .= " in vserver: $vs_name";
        }
        $self->_set_err($err_str);
        return (wantarray()) ? () : [];
    }
    return (wantarray()) ? @igroup_objs : \@igroup_objs;
}

###############################################################################
# Method:    get_luns
# Objective: To get the luns either by ostype or vserver or name.
#            These luns may not have been created by SanMgr
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_luns {
    my ($self, %args) = @_;

    my ($vs_name, @lun_attrs_return, $lun_prefix, $mapped, $in_qtree, $volume, $qtree);
    my $os_type = "all";

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        $self->sanlog(-msg => "$key, $value", -msg_level => 3);
        if ($key =~ /vserver_name/) {
            if (!length($vs_name = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /os_type/) {
            if (!length($os_type = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /lun_prefix/) {
            if (!length($lun_prefix = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /mapped/) {
            if (!length($mapped = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /in_qtree/) {
            if (!length($in_qtree = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /volume/) {
            if (!length($volume = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /qtree/) {
            if (!length($qtree = $value)) {
                $self->_set_err("invalid value for \"-$key\": $value");
                return (wantarray()) ? () : [];
            }
        } else {
            $self->_set_err("invalid \"-$key\": passed.\nValid parameter: -os_type|-vserver_name|-lun_prefix");
            return (wantarray()) ? () : [];
        }
    }

    # go through all the lun attributes and filter it out based on user request
    foreach my $lun_attr (@{$self->{LunAttrsArray}}) {
        # if user wants the lun objects in specified vserver name then
        # get the vserver name from lun object.
        if ($vs_name) {
            # get the vserver name
            my $lun_vserver_name = $lun_attr->{'vserver'};
            next unless ($lun_vserver_name eq $vs_name);
        }

        # now filter by volume
        if ($volume) {
            my $lun_path = $lun_attr->{'path'};
            unless ($lun_path =~ /$volume/) {
                # this lun is in a different volume
                next;
            }
            $self->sanlog(-msg => "Found LUN object in volume $lun_attr->{'lun'}",
                          -msg_level => 3);
        }

        # now filter by lun_prefix
        if ($lun_prefix) {
            my $lun_path = $lun_attr->{'path'};
            my $lun_name = File::Basename::basename($lun_path);
            unless ($lun_name =~ /^$lun_prefix/) {
                # this lun has a different lun prefix
                next;
            }
            $self->sanlog(-msg => "Found LUN object with $lun_prefix in $lun_attr->{'lun'}",
                          -msg_level => 3);
        }

        # now filter by qtree
        if (defined $qtree) {
            next unless $qtree eq $lun_attr->{'qtree'};
        }

        # now filter by in_qtree
        if ($in_qtree eq "true") {
            my $lun_path = $lun_attr->{'path'};
            unless ($lun_path =~ /\/vol\/\S+\/\S+\/\S+/) {
                # this lun is in a volume
                next;
            }
            $self->sanlog(-msg => "Found LUN object in qtree $lun_attr->{'lun'}",
                          -msg_level => 3);
        }

        # now filter by mapped
        if ($mapped) {
            my $lun_obj = $lun_attr->{'lun_obj'};
            if ((($mapped eq "false") and ($lun_obj->is_mapped())) or
                (($mapped eq "true") and (!$lun_obj->is_mapped()))) {
                    next;
                }
            $self->sanlog(-msg => "Found mapped LUN object $lun_attr->{'lun'}",
                          -msg_level => 3);
        }

        # now filter with os type
        my $os = $lun_attr->{'ostype'};

        # return if requested os type matches or
        # if os type == "all"
        if ((($os eq $os_type) or ($os_type =~ /all/))) {
            $self->sanlog(-msg => "Found LUN object $lun_attr->{'lun'}", -msg_level => 3);
            push(@lun_attrs_return, $lun_attr);
        }
    }

    # make sure we have at least one lun object to return
    # else fail with error message.
    unless (scalar(@lun_attrs_return)) {
        my $err_str = "No '$os_type' lun object found";
        if ($in_qtree eq "true") {
            $err_str .= " in qtree";
        }
        if ($vs_name) {
            $err_str .= " in vserver: $vs_name";
        }
        $self->_set_err($err_str);
        return (wantarray()) ? () : [];
    }
    return (wantarray()) ? @lun_attrs_return : \@lun_attrs_return;
}

###############################################################################
# Method:    get_qtree_objs
# Objective: Return qtree objects.  We're tracking this ourselves for
#            now.  Once burt 452710 is fixed, we want to retrieve this
#            from the San library.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_qtree_objs {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Placeholders for now.  Eventually, the caller should be given more
    # control over what to return.

    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    while (my ($key, $value) = each(%args)) {
        $self->sanlog(-msg => "$key, $value", -msg_level => 3);
    }

    return (wantarray()) ? @{$self->{'qtree_objs'}} : \@{$self->{'qtree_objs'}};
}

###############################################################################
#                        Private Functions
###############################################################################

###############################################################################
# Method:    _get_xml_filename
# Objective: to get the full path name with extension (xml)
# Parameter:  $xml_file (Optional)   -> user specified xml name
# Expected Output if script location is in
# "/u/muthug/p4/test/san/lun_mgmt/Lun/basic_func.thpl";
#
#  _get_xml_filename() =
#     "/u/muthug/p4/test/san/lun_mgmt/Lun/basic_func.xml";
#  _get_xml_filename("") =
#     "/u/muthug/p4/test/san/lun_mgmt/Lun/basic_func.xml";
#  _get_xml_filename("gmk1.xml.xml") =
#     "/u/muthug/p4/test/san/lun_mgmt/Lun/gmk1.xml.xml";
#  _get_xml_filename("justfile") =
#     "/u/muthug/p4/test/san/lun_mgmt/Lun/justfile.xml";
#  _get_xml_filename("p4/test/san/lun_mgmt/Lun/fullpath.xml") =
#     "p4/test/san/lun_mgmt/Lun/fullpath.xml";
#  _get_xml_filename("/san/lun_mgmt/Lun/fullpathwitoutext") =
#     "/san/lun_mgmt/Lun/fullpathwitoutext.xml";
#  _get_xml_filename("/u/any/junk/folder/name/only/Lun/") =
#     "/u/any/junk/folder/name/only/Lun/basic_func.xml";
#
# Return:  returns a string with full path with filename and extension(.xml)
###############################################################################
sub _get_xml_filename {
    my ($xml_file) = @_;

    my ($sc_file, $sc_dir, $sc_ext);
    my ($file,    $dir,    $ext);

    # if user passed nothing or empty ("") return the script location with name
    unless (length $xml_file) {
        $xml_file = $Tharn::Scriptname;

        # get the last .thpl(or any ext) name and replace with .xml
        $xml_file =~ s/\.[^.]*/.xml/;
        return $xml_file;
    }

    # first get the script location: dir:filename:ext
    ($sc_dir, $sc_file, $sc_ext) = $Tharn::Scriptname =~ m#(\S+)/(\S+)\.(\S+)$#;

    # Get file location info
    ($file, $dir, $ext) = &File::Basename::fileparse($xml_file, qr/\.[^.]*/);

    # if dir is not passed, crab it from script location
    if ($dir =~ /^.\/$|\.$|^\/$|""/) {
        $xml_file = "$sc_dir/";
    } else {
        $xml_file = $dir;
    }

    # if file is not passed, crab it from script name
    unless (length $file) {
        $xml_file .= $sc_file;
    } else {
        $xml_file .= $file;
    }

    # always extension is ".xml"
    $xml_file .= ".xml";
    return $xml_file;
}

###############################################################################
# Method:    _create_qtrees
# Objective: Create qtrees needed by test cases.  Qtrees will be created in
#            every volume that is specified in the XML.  Eventually, qtree
#            creation will be done in SanMgr (burt 437466).
###############################################################################
sub _create_qtrees {
    my ($self) = @_;

    $self->_clear_err();

    if ($self->{'qtree_count'} <= 0) {
        $self->_set_err("Invalid qtree count: $self->{'qtree_count'}; value should be >= 1");
        return $SanUtil::FALSE;
    }

    # create qtrees in each volume
    my $vol_objs = $self->get_volume_objs();
    if ($self->error()) {
        return $SanUtil::FALSE;
    }

    foreach my $vol_obj (@{$vol_objs}) {
        for (my $i = 1; $i <= $self->{'qtree_count'}; $i++) {
            my $qtree_obj = $vol_obj->create_qtree_obj(-name => "qtree$i");
            if ($vol_obj->error()) {
                $self->_set_err($vol_obj->errmsg());
                $self->_delete_qtrees();
                return $SanUtil::FALSE;
            }
            # Tracking qtree objects ourselves due to burt 452710
            push @{$self->{'qtree_objs'}}, $qtree_obj;
        }
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    _delete_qtrees
# Objective: Delete qtrees that were created by test cases.  Eventually, this
#            will be done in SanMgr (burt 437466).
###############################################################################
sub _delete_qtrees {
    my ($self) = @_;

    $self->_clear_err();

    foreach my $qtree_obj (@{$self->{'qtree_objs'}}) {
        $qtree_obj->delete();
        if ($qtree_obj->error()) {
            $self->_set_err($qtree_obj->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
}

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

LunMgr.pm

=head1 Author

Ann Harrold (harrold@netapp.com)

=head1 Description

LunMgr.pm provides an common API interface to test Lun functionality with the
main intent to avoid repeated code from thpl level.  LunMgr is inherited from
LunHelper.pm, which has common Lun utility function. LunMgr.pm uses SanMgr and
San Library. Apart from SanMgr::setup, it will do additional setup and cleanup
need for Lun testing.

Since this package base class is San::Task::LunHelper, From LunMgr object, we
can access the San::Task::LunHelper function directly.


=head1 See Also

Package inherits from the San::Task::LunHelper module. See these San::Task::LunHelper
POD for further interface detail.

B<San::Task::LunHelper>

B<San::SanMgr>

=head1 Attributes

B<sanmgr_obj>

Holds the SanMgr reference. LunMgr creates SanMgr object or user can pass the reference after setup.

B<sanmgr_created>

Either TRUE or FALSE depending on whether SanMgr object has been created.

B<qtree_count>

The number of qtrees that are to be created

B<LunAttrsArray>

An array of lun attribute hashes which represents the LUNs created by setup

=head1 Revision History

Nov-04-2009 harrold@netapp.com: Copied template from Portset
May-10-2010 harrold@netapp.com: modified after initial code review

=head1 Public Methods

=head2 B<new()>

Class Constructor method

=over 4

=item Inputs

None

=item Outputs

Returns a LunMgr blessed object reference.

=item Examples

   #
   # Create a LunMgr object
   #
   my $LUNMgr_obj = new LunMgr();
   if ($LUNMgr_obj->error()) {
      my $err_msg = $LUNMgr_obj->errmsg();
      $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
      return $TCD::FAIL;
   }

=back

=head2 B<setup(%args)>

Create necessary resources on filers and clients using SanMgr library.  It takes
same parameters as SanMgr::setup.  In addition it takes following parameters.
Refer SanMgr Pod for SanMgr::setup parameters.

=over 4

=item Inputs

=item * I<all inputs that SanMgr::setup takes>

=item * I<-file> <optional>

Name of the (param definition) xml file with full path. if xml name not specified,
then it will default to script name(with script path) with .xml

Ex: If script name ~/san/feature/Lun/ps_basic_func_cm.thpl and if you do not
pass the xml file, it will search the file from ~/san/feature/Lun/ps_basic_func_cm.xml.

=item * I<-debug_level> <optional>

Level of debug to set it for San Library

=item * I<-sanmgr_obj> <optional>

Reference to SanMgr object. Object has to be created and setup would have been called on SanMgr.
If this param is not passed, then LunMgr::setup will create SanMgr and do the setup.

=item * I<-qtree_count> <optional>

Controls how many qtrees will be created as part of the config.

=item * I<-vol_state> <optional>

Controls the state of the volume in which to do the testing.  Can be one of
ONLINE, OFFLINE, RESTRICTED, READONLY,

=item * I<-sanmgr_obj> <optional>

Reference to SanMgr object. Object has to be created and setup would have been called on SanMgr.
If this param is not passed, then LunMgr::setup will create SanMgr and do the setup.
=item Outputs

Returns TRUE resource are created successful. If an error occurred the error
flag is set, errmsg is set to the failure reason, and FALSE is passed back as
the return value.

=item Examples

      #
      # Instantiate LunMgr Object

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error()) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $cmd_args{'-file'} = $SanSetupXML;
      $cmd_args{'-debug_level'} = $DebugLevel;
      $cmd_args{'-vol_state'}   = $VolState;
      $cmd_args{'-qtree_count'} = $QtreeCount;

      $LUNMgr_obj->setup(%cmd_args);
      if ($LUNMgr_obj->error()) {
          $Test->nlog($LUNMgr_obj->errmsg());
          return $TCD::FAIL;
      }

=back

=head2 B<cleanup()>

To Cleanup what LunMgr::setup created. If SanMgr created outside of LunMgr,
then it wont cleanup up the SanMgr.

=over 4

=item Inputs

None

=item Outputs

Returns TRUE resource are cleaned up successful. If an error occurred the error
flag is set, errmsg is set to the failure reason, and FALSE is passed back as
the return value.

=item Examples

      #
      # Instantiate LunMgr Object

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup();
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      ...
      #
      # Call cleanup
      #
      $LUNMgr_obj->cleanup();
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<validate()>

To validate the basic requirement of Lun test suite. Apart from SanMgr validate it does following validation

    # Need to validate following
    # 1. requested vserver presents
    # 2. iscsi and fcp object created and service enabled
    #    if not,create/enabled it, during cleanup, put back to original state
    # 3. enable required license
    # 4. requested iscsi and fcp lifs presents
    # 5. available fcp target port to create fcp lif
    # 6. available iscsi ports to create iscsi lif

This is called by LunMgr::Setup, mostly no need to call it.

=over 4

=item Inputs

None

=item Outputs

Returns TRUE resource are created successful. If an error occurred the error
flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

      #
      # Instantiate LunMgr Object (see above examples)
            
      $LUNMgr_obj = new LunMgr();
      ...
      ...
      ...
      #
      # Call validate 
      #
      $LUNMgr_obj->validate();
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_sanmgr_obj()>

To get the SanMgr objects. It will return what Lun::setup created or user passed.

NOTE: This is called by LunMgr::Setup, mostly no need to call it.

=over 4

=item Inputs

None

=item Outputs

Returns a reference to SanMgr object if successful. If an error occurred the
error flag is set, errmsg is set to the failure reason, and {} (empty hash
reference) is passed back as the return value.

=item Examples

      #
      # Instantiate LunMgr Object (see above examples)

      $LUNMgr_obj = new LunMgr();
      ...
      ...
      ...
      #
      # get the SanMgr Object.
      #
      my $SanMgr_obj = $LUNMgr_obj->get_sanmgr_obj();
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_vserver_objs()>

To get the vserver objects that SanMgr::setup created/reused

NOTE;

This functions returns the vserver object which is/are created or reused by San::SanMgr:setup.
You get all the vservers in the controller by using SanFiler::get_vservers().

=over 4

=item Inputs

None

=item Outputs

Returns a reference to array or array of vserver objects if successful. If an
error occurred the error flag is set, errmsg is set to the failure reason,
and empty array ([] or ()) is passed back as the return value.

=item Examples

      #
      # Instantiate LunMgr Object (see above examples)

      $LUNMgr_obj = new LunMgr();
      ...
      ...
      ...
      #
      # get the vserver objects created by setup
      #
      my @Vserver_objects = $LUNMgr_obj->get_vserver_objs();
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_igroup_objs(%args)>

To get the igroup objects that SanMgr::setup created/reused. It will return all
igroup objects or specific protocol/ostype igroup objects on specified vserver

=over 4

=item Inputs

=item * I<-vserver_name> <optional>

Name of the vserver. If you specify this, it returns only the igroups belonging to this vserver

=item * I<-protocol_type> <optional, default: all>

Name of the protocol type: valid inputs are: all|mixed|fcp|iscsi. Returns only
the igroups which matches the specified protocol type.

=item * I<-ostype> <optional, default: all>

Name of the os type: valid inputs are any igroup ostype supported. Returns only
the igroups which matches the specified ostype.

=item Outputs

Returns a reference to array or array of igroup objects if successful. If an
error occurred the error flag is set, errmsg is set to the failure reason,
and empty array ([] or ()) is passed back as the return value.

=item Examples

      #
      # Instantiate LunMgr Object (see above examples)
 
      $LUNMgr_obj = new LunMgr();
      ...
      ...
      ...
      # get the all the igroup objects created by setup
      my @igroup_objects = $LUNMgr_obj->get_igroup_objs();
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # get the all the igroup objects in particular vserver
      # this will return all the igroups on vserver named "ps_vs" 
      my @igroup_objects = $LUNMgr_obj->get_igroup_objs(-vserver_name => "ps_vs");
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # get only fcp igroup objects in particular vserver
      my @igroup_objects = $LUNMgr_obj->get_igroup_objs(-vserver_name => "ps_vs",
                                                        -protocol_type => "fcp");
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # get all iscsi igroup objects
      my @igroup_objects = $LUNMgr_obj->get_igroup_objs(-protocol_type => "iscsi");
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # to use the igroup objects
      foreach my $igrp_obj (@igroup_objects) {
         $igrp_obj->get_name();
         ...
         ...
      }

=back

=head2 B<get_luns(%args)>

To get the luns that SanMgr::setup created/reused. It will return all
luns or specific luns on specified vserver.  The return type is a lun
attribute hash.

=over 4

=item Inputs

=item * I<-vserver_name> <optional>

Name of the vserver. If you specify this, it returns only the luns belonging to this vserver

=item * I<-ostype> <optional, default: all>

Name of the os type: valid inputs are any lun ostype supported. Returns only
the luns which matches the specified ostype.

=item * I<-lun_prefix> <optional>

Prefix of the lun name to use for searching for luns.  Returns only those luns
which start with specified lun_prefix.

=item * I<-mapped> <optional>

Controls whether to search for mapped or unmapped luns.  Valid values are
"true" and "false".

=item * I<-in_qtree> <optional>

Controls whether to search for luns that exist in qtrees.  Valid values are
"true" and "false".

=item * I<-volume> <optional>

Name of the volume - if you specify this, it will return only the luns that exist
in this volume.

=item Outputs

Returns a reference to array or array of lun attribute hashes if successful. If an
error occurred the error flag is set, errmsg is set to the failure reason,
and empty array ([] or ()) is passed back as the return value.

=item Examples

      #
      # Instantiate LunMgr Object (see above examples)

      $LUNMgr_obj = new LunMgr();
      ...
      ...
      ...
      # get the all the lun attr hashes created by setup
      my @lun_attrs = $LUNMgr_obj->get_luns();
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # get the all the lun attr hashes in particular vserver
      # this will return all the luns on vserver named "ps_vs"
      my @lun_attrs = $LUNMgr_obj->get_luns(-vserver_name => "ps_vs");
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # get only the unmapped luns in particular vserver
      @lun_attrs = $LUNMgr_obj->get_luns('-vserver_name'        => $vs_name,
                                         '-mapped'              => "false");
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...

=back

=head2 B<get_volume_objs(%args)>

To get the volume objects that SanMgr::setup created/reused. It will return all
volume objects or the volume objects on specified vserver

=over 4

=item Inputs

=item * I<-vserver_name> <optional>

Name of the vserver. If you specify this, it returns only the volumes belonging to this vserver

=item Outputs

Returns a reference to array or array of volume objects if successful. If an
error occurred the error flag is set, errmsg is set to the failure reason,
and empty array ([] or ()) is passed back as the return value.

=item Examples

      #
      # Instantiate LunMgr Object (see above examples)

      $LUNMgr_obj = new LunMgr();
      ...
      ...
      ...
      # get the all the volume objects created by setup
      my @volume_objects = $LUNMgr_obj->get_volume_objs();
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # get the all the volume objects in particular vserver
      # this will return all the volumes on vserver named "ps_vs"
      my @volume_objects = $LUNMgr_obj->get_volume_objs(-vserver_name => "ps_vs");
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...

=back

=head2 B<create_tc_luns(%args)>

To create the luns that will be used for testing.  By default both mapped
and unmapped luns are created.

=over 4

=item Inputs

=item * I<-volumes> <required>

An array of the volume objects in which to create the LUNs

=item * I<-luns_per_ostype> <optional, default: 1>

How many luns to create of each supported OS type.  The supported OS types
are maintained in this library.  By default, this method creates luns_per_ostype
of each os type

=item * I<-lun_size> <optional, default: 40MB>

The size of the luns to create.

=item * I<-lun_name_prefix> <required>

Prefix of the lun name to use when creating the luns.  Lun names are created
like this:
my $lun_name = $lun_name_prefix . "_unmapped_" . $os_type . $lun_count;
my $lun_name = $lun_name_prefix . "_mapped_" . $os_type . $lun_count;

=item * I<-lun_os_type> <optional>

If this is specified, create luns of only this os type.  Otherwise create luns
of all supported os types.

=item * I<-space-resv> <optional>

Controls whether to create space reserved luns.  Valid values are noreserve

=item * I<-volume_name> <optional>

Name of the volume - if you specify this, it will create luns in this volume only,
instead of the default behavior which is to create luns in all Volumes in -volumes
array.

=item Outputs

Returns a reference to array or array of lun attribute hashes if successful. If an
error occurred the error flag is set, errmsg is set to the failure reason,
and empty array ([] or ()) is passed back as the return value.  (TBD, might
change this to return TRUE or FALSE and have the lun attr array maintained inside
this object...)

=item Examples

      #
      # Instantiate LunMgr Object (see above examples)

      $LUNMgr_obj = new LunMgr();
      ...
      ...
      ...
      @LunAttrsArray =  $LUNMgr_obj->create_tc_luns(-volumes         => \@Volumes,
                                                    -luns_per_ostype => $LunsPerOsType,
                                                    -lun_size        => $LunSize,
                                                    -lun_name_prefix => "lun_move");
      ...
      ...

=back

=head2 B<cleanup_tc_luns()>

To cleanup the luns that were created by create_tc_luns.

=over 4

=item Inputs

None

=item Outputs

Returns TRUE resource are cleaned up successful. If an error occurred the error
flag is set, errmsg is set to the failure reason, and FALSE is passed back as
the return value.

=item Examples

      #
      # Instantiate LunMgr Object

      $LUNMgr_obj = new LunMgr();
      if ($LUNMgr_obj->error) {
         my $err_msg = $LUNMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate LunMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $LUNMgr_obj->setup();
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      ...
      @LunAttrsArray =  $LUNMgr_obj->create_tc_luns(-volumes         => \@Volumes,
                                                    -luns_per_ostype => $LunsPerOsType,
                                                    -lun_size        => $LunSize,
                                                    -lun_name_prefix => "lun_move");
      ...
      ...
      ...
      #
      # Call cleanup
      #
      $LUNMgr_obj->cleanup_tc_luns();
      if ($LUNMgr_obj->error()) {
         $Test->nlog($LUNMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=cut

