# -*- perl -*-
#  $Id: //depot/prod/test/Rsierranevada/lib/San/Application/PortsetMgr.pm#1 $
#
# Copyright (c) 2010 NetApp, Inc.
# All rights reserved.
#
##
## @summary Utility for creating Portset test suite
## It provides following functionality
##  * setup test suite
##  * cleanup test suite
## @author SAN QA (c-dl-ssan-test@netapp.com)
## @status public
## @pod here

package San::Application::PortsetMgr;

# compiler directives
use strict;
use Params::Validate qw( :all );
use San::Application::Portsetconstant;
use San::SanMgr;

# inherit base class
use base qw( San::Task::PortsetHelper );

###############################################################################
# Method:    new
# Objective: Class Constructor.
# Parameter: See below
# Returns:   this class blessed object reference
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my $self = {'sanmgr_obj'     => '',
                'sanmgr_created' => $SanUtil::FALSE};
    @{$self->{'tpgorup_objs'}}   = ();
    @{$self->{'iscsi_lif_objs'}} = ();

    # bless instance into class
    bless $self, $class;
    return $self;
}

###############################################################################
# Method:    setup
# Objective: To create SanMgr Object and create the environment using SanMgr.
# Returns:  If succeeds returns TRUE else FALSE with error message set
# Details:   See POD documentation at the end of this file.
###############################################################################
sub setup {
    my ($self, %args) = @_;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {'file' => {type     => SCALAR,
                               optional => 1},
                    'tcd_test' => {type     => OBJECT,
                                   required => 1,
                                   isa      => 'TCD'},
                    'sanmgr_obj' => {type     => OBJECT,
                                     optional => 1,
                                     isa      => 'San::SanMgr'},
                    'helper_debug_level' => {type     => SCALAR,
                                             optional => 1,
                                             default  => 0,
                                             regex    => qr/^\d+$/}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }

    # set debug level for PortsetMgr.pm and PortsetHelper.pm
    $self->set_debug_level($args{'helper_debug_level'});

    # display argument hash if debug_level is high
    $self->sanlog(-msg => $self->dumper_get(\%args), -msg_level => 3);

    if (defined $args{'sanmgr_obj'}) {
        $self->{'sanmgr_obj'}     = $args{'sanmgr_obj'};
        $self->{'sanmgr_created'} = $SanUtil::FALSE;
    } else {
        my $sanmgr_obj = new San::SanMgr('-file' => $args{'file'});
        if ($sanmgr_obj->error()) {
            $self->_set_err($sanmgr_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # do SanMgr::setup
        $sanmgr_obj->setup('-tcd_test' => $args{'tcd_test'});
        if ($sanmgr_obj->error()) {
            $self->_set_err($sanmgr_obj->errmsg());
            return $SanUtil::FALSE;
        }
        $self->{'sanmgr_obj'}     = $sanmgr_obj;
        $self->{'sanmgr_created'} = $SanUtil::TRUE;
    }
    return $SanUtil::TRUE;
}    # end of sub setup

###############################################################################
# Method:    cleanup
# Objective: To cleanup what setup created
# Details:   See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    my ($self) = @_;

    $self->_clear_err();
    my $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return $SanUtil::FALSE;
    }

    # cleanup what PortsetMgr::setup created...
    # destroy tpgroup if created
    my $errstr   = "";
    my @tpg_objs = @{$self->{'tpgorup_objs'}};
    foreach my $tpg (@tpg_objs) {
        $tpg->destroy();
        if ($tpg->error()) {
            $errstr .= $tpg->errmsg() . "\n";
        }
    }

    # destroy iscsi if created
    my @iscsi_lif_objs = @{$self->{'iscsi_lif_objs'}};
    foreach my $iscsi_lif (@iscsi_lif_objs) {
        $iscsi_lif->destroy();
        if ($iscsi_lif->error()) {
            $errstr .= $iscsi_lif->errmsg() . "\n";
        }
    }

    # if SanMgr is created by PortsetMgr::setup then call
    # SanMgr::cleanup
    if ($self->{'sanmgr_created'} == $SanUtil::TRUE) {

        # cleanup what SanMgr created
        $sanmgr_obj->cleanup();
        if ($sanmgr_obj->error()) {
            $errstr .= "SanMgr::Cleanup failed: " . $sanmgr_obj->errmsg();
        }
    }
    if ($errstr ne "") {
        $self->_set_err($errstr);
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}    # end of sub cleanup

###############################################################################
# Method:    get_sanmgr_obj
# Objective: To get the SanMgr Object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanmgr_obj {
    my ($self) = shift;

    $self->_clear_err();

    # verify the sanmgr object initialized
    unless (ref($self->{'sanmgr_obj'}) =~ /San::SanMgr/) {
        $self->_set_err("PortsetMgr::setup need to be called prior to this function");
        return {};
    }
    return $self->{'sanmgr_obj'};
}    # end of sub get_sanmgr_obj

###############################################################################
# Method:    get_vserver_objs
# Objective: To get the vserver objects that SanMgr::setup create/reuse
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vserver_objs {
    my ($self) = shift;

    my (@vserver_objs, $sanmgr_obj);

    $self->_clear_err();

    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return (wantarray()) ? () : [];
    }

    # retrieve the vserver objects created/reused by SanMgr
    @vserver_objs = $sanmgr_obj->get_objects(-type => "vserver");
    if ($sanmgr_obj->error()) {
        $self->_set_err($sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # additional check to make sure we have some objects
    # since get_objects function returns empty array without any error set
    unless (scalar(@vserver_objs)) {
        $self->_set_err("No vserver object found");
        return (wantarray()) ? () : [];
    }
    return (wantarray()) ? @vserver_objs : \@vserver_objs;
}    # end of sub get_vserver_objs

###############################################################################
# Method:    get_igroup_objs
# Objective: To get the igroup objects that SanMgr::setup create/reuse
#            it will return all igroup or specific protocol igroup objects
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_igroup_objs {
    my ($self, %args) = @_;

    my ($vs_name, $protocol_type, @igroup_objs, @sanmgr_igroup_objs, $sanmgr_obj);
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(PS_PARAM_VSERVER_NAME) => {type => SCALAR, optional => 1},
             (PS_ATTR_PROTOCOL)      => {
                 type      => SCALAR,
                 optional  => 1,
                 default   => 'all',
                 callbacks => {'One of the values: all|fcp|iscsi|mixed' => sub { $_[0] =~ /all|fcp|iscsi|mixed/i }
                              }}

            });
    };
    if ($@) {
        $self->_set_err("Failed to validate parameter: " . $@);
        return (wantarray()) ? () : [];
    }
    if (defined $args{(PS_PARAM_VSERVER_NAME)}) {
        $vs_name = $args{(PS_PARAM_VSERVER_NAME)};
    }
    if (defined $args{(PS_ATTR_PROTOCOL)}) {
        $protocol_type = $args{(PS_ATTR_PROTOCOL)};
    }
    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return (wantarray()) ? () : [];
    }

    # get all the igroup objects created/reused by SanMgr
    @sanmgr_igroup_objs = $sanmgr_obj->get_objects(-type => "igroup");
    if ($sanmgr_obj->error()) {
        $self->_set_err("Unable to get the igroup objects\nReason: " . $sanmgr_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # additional check to make sure we have some objects
    # since get_objects function returns empty array without any error set
    unless (scalar(@sanmgr_igroup_objs)) {
        $self->_set_err("No igroup object found in the SanMgr");
        return (wantarray()) ? () : [];
    }

    # if vserver name and protocol_type is not specified
    # then simply return all objects
    if (!defined($vs_name) and $protocol_type =~ /all/) {
        return (wantarray()) ? @sanmgr_igroup_objs : \@sanmgr_igroup_objs;
    }

    # go through all the igroup objects and
    # filter it out based on user request
    foreach my $igroup_obj (@sanmgr_igroup_objs) {
        my $igroup_name = $igroup_obj->get_name();
        if ($igroup_obj->error()) {
            $self->_set_err("Unable to get the name of the igroup\nReason: " . $igroup_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        # if user wants the igroup objects in specified vserver name then
        # get the vserver name from igroup object.
        if ($vs_name) {

            # parent from igroup object should be vserver object
            my $vs_obj = $igroup_obj->get_parent_obj();
            if (ref($vs_obj) !~ /San::ControllerCluster::Filer10M::Vserver/) {
                $self->_set_err("Unable to get parent (vserver) object of igroup");
                return (wantarray()) ? () : [];
            }
            my $igroup_vserver_name = $vs_obj->get_name();
            if ($vs_obj->error()) {
                $self->_set_err(  "Unable to get the name of the vserver "
                                . "in the igroup($igroup_name)\nReason: "
                                . $vs_obj->errmsg());
                return (wantarray()) ? () : [];
            }
            next unless ($igroup_vserver_name eq $vs_name);
        }

        # now filter with protocol type
        my $pro_type = $igroup_obj->get_protocol_type();
        if ($igroup_obj->error()) {
            $self->_set_err(
                    "Unable to get the protocol type from " . "igroup($igroup_name)\nReason: " . $igroup_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        # return if requested protocol type matches or
        # if protocol type == "all"
        if (($pro_type eq $protocol_type) or $protocol_type =~ /all/) {
            push(@igroup_objs, $igroup_obj);
        }
    }

    # make sure we have at least one igroup object to return
    # else fail with error message.
    unless (scalar(@igroup_objs)) {
        my $err_str = "No '$protocol_type' igroup object found";
        if ($vs_name) {
            $err_str .= " in vserver: $vs_name";
        }
        $self->_set_err($err_str);
        return (wantarray()) ? () : [];
    }
    return (wantarray()) ? @igroup_objs : \@igroup_objs;
}    # end of sub get_igroup_objs

###############################################################################
# Method:    get_protocol_lifs_name_in_vserver
# Objective: to get all the specified 'protocol' lifs name for the given vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_protocol_lifs_name_in_vserver {
    my ($self, %args) = @_;

    my ($vs_obj, $vs_name, $protocol_type, @lif_objs, @lif_names);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) = validate(
            @input_params,
            {(PS_PARAM_VSERVER_OBJ) => {type     => OBJECT,
                                        required => 1,
                                        isa      => 'San::ControllerCluster::Filer10M::Vserver'},
             (PS_ATTR_PROTOCOL) => {
                 type      => SCALAR,
                 required  => 1,
                 callbacks => {'One of the value: fcp|iscsi|mixed' => sub { $_[0] =~ /fcp|iscsi/i }
                              }}});
    };
    if ($@) {
        $self->_set_err($@);
        return undef;
    }
    $protocol_type = $args{(PS_ATTR_PROTOCOL)};

    # get the vserver object and the name of the vserver
    $vs_obj  = $args{(PS_PARAM_VSERVER_OBJ)};
    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get vserver name.\nReason:\n" . $vs_obj->errmsg());
        return undef;
    }

    # get all the specified protocol lif objects, this function returns
    # the fcp lifs in a given vserver, not only the lif created by setup,
    # including already exist on the given vserver
    @lif_objs = $vs_obj->get_lifs(-type => $protocol_type);
    if ($vs_obj->error()) {
        $self->_set_err(
                 "Failed to get all '$protocol_type' lifs from " . "vserver($vs_name).\nReason:\n" . $vs_obj->errmsg());
        return undef;
    }
    unless (scalar(@lif_objs)) {
        $self->_set_err("No '$protocol_type' lif found in the vserver($vs_name)");
        return undef;
    }
    foreach my $lif_obj (@lif_objs) {

        # get name of the lifs
        my $lifname = $lif_obj->get_name();
        if ($lif_obj->error()) {
            $self->_set_err(
                "Failed to get '$protocol_type' lif name from " . "vserver($vs_name).\nReason:\n" . $lif_obj->errmsg());
            return undef;
        }
        push(@lif_names, $lifname);
    }
    unless (scalar(@lif_names)) {
        $self->_set_err("No '$protocol_type' lif found in the vserver($vs_name)");
        return undef;
    }

    # cmode library can not send space separated lif name
    # so join with comma. refer burt:383323 for more details - 10/21/09
    my $lif_names = join ',', @lif_names;
    return $lif_names;
}    # end sub get_protocol_lifs_name_in_vserver

###############################################################################
# Method:    get_fcp_lifs_name_wwpn_in_vserver
# Objective: to get all the fcp lifs name and its wwpn for the given vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fcp_lifs_name_wwpn_in_vserver {
    my ($self, %args) = @_;

    my ($vs_obj,   $vs_name,       $fcp_obj);
    my (@fcp_lifs, @fcp_lif_names, @fcp_wwpn);
    my ($lif_names_param, $lif_wwpn_param);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(PS_PARAM_VSERVER_OBJ) => {type     => OBJECT,
                                               required => 1,
                                               isa      => 'San::ControllerCluster::Filer10M::Vserver'}});
    };
    if ($@) {
        $self->_set_err($@);
        return (undef, undef);
    }

    # get the vserver object and the name of the vserver
    $vs_obj  = $args{(PS_PARAM_VSERVER_OBJ)};
    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get vserver name.\nReason:\n" . $vs_obj->errmsg());
        return (undef, undef);
    }

    # fcp object need to be created by setup or any other function
    # prior to this call.
    # For right now to get the fcp nodename,
    # need to go through fcp object by passing fcp lif name.
    $fcp_obj = $vs_obj->get_fcp_obj();
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get fcp object from " . "vserver($vs_name).\nReason:\n" . $vs_obj->errmsg());
        return (undef, undef);
    }

    # get all the fcp(only) lif objects, this function returns
    # the fcp lifs in a given vserver, not only the lif created by setup,
    # including already exist on the given vserver
    @fcp_lifs = $vs_obj->get_lifs(-type => "fcp");
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get all fcp lifs from " . "vserver($vs_name).\nReason:\n" . $vs_obj->errmsg());
        return (undef, undef);
    }
    unless (scalar(@fcp_lifs)) {
        $self->_set_err("No fcp lif found from vserver($vs_name)");
        return (undef, undef);
    }
    foreach my $fcp_lif (@fcp_lifs) {

        # get name of the lifs
        my $fcp_lifname = $fcp_lif->get_name();
        if ($fcp_lif->error()) {
            $self->_set_err("Failed to get fcp lif name from " . "vserver($vs_name).\nReason:\n" . $fcp_lif->errmsg());
            return (undef, undef);
        }
        push(@fcp_lif_names, $fcp_lifname);

        # get wwpn of fcp lifs, right now you need to access through fcp object
        # hopefully later wwpn can be access from lif object - 10/18/09 - gmk
        my @fcp_wwpn_name = $fcp_obj->get_wwpn(-lif => $fcp_lifname);
        if ($fcp_obj->error()) {
            $self->_set_err("Failed to get fcp lif wwpn from " . "vserver($vs_name).\nReason:\n" . $fcp_obj->errmsg());
            return (undef, undef);
        }
        push(@fcp_wwpn, @fcp_wwpn_name);
    }
    unless (scalar(@fcp_lif_names)) {
        $self->_set_err("No fcp lif name found from vserver($vs_name)");
        return (undef, undef);
    }
    unless (scalar(@fcp_wwpn)) {
        $self->_set_err("No fcp lif wwpn found from vserver($vs_name)");
        return (undef, undef);
    }

    # cmode library can not send space separated lif name/wwpn,
    # so join with comma. refer burt:383323 for more details - 10/21/09 - gmk
    $lif_names_param = join ',', @fcp_lif_names;
    $lif_wwpn_param  = join ',', @fcp_wwpn;
    return ($lif_names_param, $lif_wwpn_param);
}    # end sub get_fcp_lifs_name_wwpn_in_vserver

###############################################################################
# Method:    get_iscsi_lifs_name_tpg_in_vserver
# Objective: to get all the iscsi lif name and default tpg name
#            for the given vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_iscsi_lifs_name_tpg_in_vserver {
    my ($self, %args) = @_;

    my ($vs_obj, $vs_name, $iscsi_obj, @iscsi_lifs_objs, @iscsi_lif_name, @iscsi_lif_tpg_name);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(PS_PARAM_VSERVER_OBJ) => {type     => OBJECT,
                                               required => 1,
                                               isa      => 'San::ControllerCluster::Filer10M::Vserver'}});
    };
    if ($@) {
        $self->_set_err($@);
        return (undef, undef);
    }

    # get the vserver object and the name of the vserver
    $vs_obj  = $args{(PS_PARAM_VSERVER_OBJ)};
    $vs_name = $vs_obj->get_name();
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get vserver name.\nReason:\n" . $vs_obj->errmsg());
        return (undef, undef);
    }

    # iscsi object need to be created by setup or any other function
    # prior to this call.
    # For right now to get the iscsi lif tpg object,
    # need to go through iscsi object.
    # For right now to get the fcp nodename,
    # need to go through fcp object by passing iscsi lif name.
    $iscsi_obj = $vs_obj->get_iscsi_obj();
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get iscsi object from " . "vserver($vs_name).\nReason:\n" . $vs_obj->errmsg());
        return (undef, undef);
    }

    # get all the iscsi(only) lif objects. This function returns
    # the iscsi lifs in a given vserver, not only the lif created by setup,
    # including already exist on the given vserver
    @iscsi_lifs_objs = $vs_obj->get_lifs(-type => "iscsi");
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get all fcp lifs from " . "vserver($vs_name).\nReason:\n" . $vs_obj->errmsg());
        return (undef, undef);
    }
    unless (scalar(@iscsi_lifs_objs)) {
        $self->_set_err("No iscsi lif found from vserver($vs_name)");
        return (undef);
    }
    foreach my $iscsi_lif_obj (@iscsi_lifs_objs) {

        # get name of the iscsi lif
        my $iscsi_lif_name = $iscsi_lif_obj->get_name();
        if ($iscsi_lif_obj->error()) {
            $self->_set_err(
                     "Failed to get iscsi lif name from vserver" . "($vs_name).\nReason:\n" . $iscsi_lif_obj->errmsg());
            return (undef, undef);
        }
        push(@iscsi_lif_name, $iscsi_lif_name);

        # get name of the iscsi lif
        my $tpgtag = $iscsi_lif_obj->get_tpgroup();
        if ($iscsi_lif_obj->error()) {
            $self->_set_err(
                  "Failed to get iscsi lif tpg tag from vserver" . "($vs_name).\nReason:\n" . $iscsi_lif_obj->errmsg());
            return (undef, undef);
        }
        my $tpgroup_obj = $iscsi_lif_obj->get_tpgroup_objs('-tpgtag' => $tpgtag);
        if ($iscsi_obj->error()) {
            $self->_set_err(  "Failed to get tpgroup object from iscsi_lif object"
                            . " on vserver($vs_name).\nReason:\n"
                            . $iscsi_obj->errmsg());
            return (undef, undef);
        }
        my $iscsi_lig_tpg = $tpgroup_obj->get_name();
        if ($tpgroup_obj->error()) {
            $self->_set_err(
                          "Failed to get tpgroup name " . " on vserver($vs_name).\nReason:\n" . $tpgroup_obj->errmsg());
            return (undef, undef);
        }
        push(@iscsi_lif_tpg_name, $iscsi_lig_tpg);
    }
    unless (scalar(@iscsi_lif_name)) {
        $self->_set_err("No iscsi lif name found from vserver($vs_name).");
        return (undef, undef);
    }
    unless (scalar(@iscsi_lif_tpg_name)) {
        $self->_set_err("No iscsi lif default tpg name found from vserver($vs_name).");
        return (undef, undef);
    }

    # cmode library can not send space separated lif name/wwpn,
    # so join with comma. refer burt:383323 for more details - 10/21/09 - gmk

    my $lif_names     = join ',', @iscsi_lif_name;
    my $lif_tpg_names = join ',', @iscsi_lif_tpg_name;
    return ($lif_names, $lif_tpg_names);
}    # end sub get_iscsi_lifs_name_tpg_in_vserver

###############################################################################
# Method:    create_tpgroup_on_vserver
# Objective: to create the tpgroup on given vserver.
#            if you don't pass iscsi_lif_name, it will create empty tpgroup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_tpgroup_on_vserver {
    my ($self, %args) = @_;

    my ($vs_obj, $sanmgr_obj, $tpgroup_name);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(PS_PARAM_VSERVER_OBJ) => {type     => OBJECT,
                                               required => 1,
                                               isa      => 'San::ControllerCluster::Filer10M::Vserver'},
                    (PS_PARAM_TPGROUP_NAME)   => {type => SCALAR, required => 1},
                    (PS_PARAM_ISCSI_LIF_NAME) => {type => SCALAR, optional => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return {};
    }
    $vs_obj = $args{(PS_PARAM_VSERVER_OBJ)};
    my %cmd_args;
    $cmd_args{'-name'} = $args{(PS_PARAM_TPGROUP_NAME)};
    $cmd_args{'-lif'} = $args{(PS_PARAM_ISCSI_LIF_NAME)} if (defined $args{(PS_PARAM_ISCSI_LIF_NAME)});
    my $tpg_obj = $vs_obj->create_tpgroup_obj(%cmd_args);
    if ($vs_obj->error()) {
        $self->_set_err("Unable to create tpgroup object\nReason:\n" . $vs_obj->errmsg());
        return {};
    }
    push(@{$self->{'tpgorup_objs'}}, $tpg_obj);
    return $tpg_obj;
}    # end sub create_tpgroup_on_vserver

###############################################################################
# Method:    is_tpgroup_on_vserver_empty
# Objective: to find the given tpgroup has lif or not on given vserver.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_tpgroup_on_vserver_empty {
    my ($self, %args) = @_;

    my ($vs_name, $vs_obj, $sanmgr_obj, $tpgroup_name, $iscsi_obj);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    eval {

        # validate input parameters
        my @input_params = (%args);
        (%args) =
          validate(@input_params,
                   {(PS_PARAM_VSERVER_NAME) => {type => SCALAR, required => 1},
                    (PS_PARAM_TPGROUP_NAME) => {type => SCALAR, required => 1}});
    };
    if ($@) {
        $self->_set_err($@);
        return $SanUtil::FALSE;
    }
    $sanmgr_obj = $self->get_sanmgr_obj();
    if ($self->error()) {

        # error set in the function just bubble it up
        return $SanUtil::FALSE;
    }

    # get the vserver object
    # retrieve the vserver objects created/reused by SanMgr
    # get_objects always returns in array, so pickup the first one

    $vs_name      = $args{(PS_PARAM_VSERVER_NAME)};
    $tpgroup_name = $args{(PS_PARAM_TPGROUP_NAME)};
    my @vs_objs = $sanmgr_obj->get_objects('-type' => "vserver",
                                           '-name' => $vs_name);
    if ($sanmgr_obj->error()) {
        $self->_set_err("Unable to get the vserver($vs_name) object\nReason:\n" . $sanmgr_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $vs_obj = $vs_objs[0];
    if (ref($vs_obj) !~ /San::ControllerCluster::Filer10M::Vserver/) {
        $self->_set_err("It is not vserver($vs_name) object " . "return from SanMgr: " . ref($vs_obj));
        return $SanUtil::FALSE;
    }
    $iscsi_obj = $vs_obj->get_iscsi_obj();
    if ($vs_obj->error()) {
        $self->_set_err("Failed to get iscsi object from " . "vserver($vs_name).\nReason:\n" . $vs_obj->errmsg());
        return $SanUtil::FALSE;
    }
    my $tpgroup_obj = $iscsi_obj->get_tpgroup_objs('-name' => $tpgroup_name);
    if ($iscsi_obj->error()) {
        $self->_set_err("Failed to get tpgroup($tpgroup_name) object" . $iscsi_obj->errmsg());
        return $SanUtil::FALSE;
    }
    my $lif_names = $tpgroup_obj->get_lif();
    if ($tpgroup_obj->error()) {
        $self->_set_err("Failed to get tpgroup($tpgroup_name) lifs name" . $tpgroup_obj->errmsg());
        return $SanUtil::FALSE;
    }
    $self->sanlog(-msg       => "tpgroup($tpgroup_name) lifs: $lif_names",
                  -msg_level => 3);

    # typically "-" means is empty, but OnTap allow to create lif name with "_",
    # in that case this assumption is wrong, Hope library should handle this
    # since this output may vary for zapi....
    if ($lif_names eq "-") {
        return $SanUtil::TRUE;
    }
    return $SanUtil::FALSE;
}    # end sub is_tpgroup_on_vserver_empty

1;

###############################################################################
#                             POD Documentation
###############################################################################

=pod

=head1 Package Name

PortsetMgr.pm

=head1 Author

Muthukumaran Govindarajan (muthug@netapp.com)

=head1 Description

PortsetMgr.pm provides an common API interface to test portset functionality with the main intent to avoid repeated code from thpl level.
PortsetMgr is inherited from PortsetHelper.pm, which has common portset utility function. PortsetMgr.pm uses SanMgr and San Library. Apart
from SanMgr::setup, it will do additional setup and cleanup need for portset testing.

Since this package base class is San::Task::PortsetHelper, From PortsetMgr object, we can access the San::Task::PortsetHelper function directly.


=head1 See Also

Package inherits from the San::Task::PortsetHelper module. See these San::Task::PortsetHelper POD
for further interface detail.

B<San::Task::PortsetHelper>

B<San::SanMgr>


=head1 Attributes

B<sanmgr_obj> 

Holds the SanMgr reference. PortsetMgr creates SanMgr object or user can pass the reference after setup.

=head1 Revision History

Nov-04-2009 muthug@netapp.com: Original template created
Nov-30-2009 muthug@netapp.com: modified after initial code review
Jan-04-2010 muthug@netapp.com: Modified for new UI portset command
Jul-08-2010 muthug@netapp.com: Modified to use SanMgr

=head1 Public Methods

=head2 B<new()>

Class Constructor method

=over 4

=item Inputs

None

=item Outputs

Returns a PortsetMgr blessed object reference.

=item Examples

   #
   # Create a PortsetMgr object
   #
   my $psmgr_obj = new PortsetMgr();
   if ($psmgr_obj->error()) {
      my $err_msg = $psmgr_obj->errmsg();
      $Test->nlog("Failed to instantiate PortsetMgr object:\nReason: $err_msg");
      return $TCD::FAIL;
   }

=back

=head2 B<setup(%args)>

Create necessary resources on filers and clients using San library.  It takes same parameters as SanMgr::setup.
In addition it takes following parameters. Refer SanMgr Pod for SanMgr::setup parameters.

=over 4

=item Inputs

=item * I<all inputs that SanMgr::setup takes>

=item * I<-file> <optional> 

Name of the (param definition) xml file with full path. if xml name not specified, then it will default to script name(with script path) with .xml

Ex: If script name ~/san/feature/portset/ps_basic_func_cm.thpl and if you do not pass the xml file, it will search the file from
~/san/feature/portset/ps_basic_func_cm.xml.

=item * I<-sanmgr_obj> <optional> 

Reference to SanMgr object. Object has to be created and setup would have been called on SanMgr.
If this param is not passed, then PortsetMgr::setup will create SanMgr and do the setup.

=item Outputs

Returns TRUE resource are created successful. If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

      #
      # Instantiate PortsetMgr Object 
            
      $PSMgr_obj = new PortsetMgr();
      if ($PSMgr_obj->error()) {
         my $err_msg = $PSMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate PortsetMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $PSMgr_obj->setup("-file" => full_path/thplname.xml);
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<cleanup()>

To Cleanup what PortsetMgr::setup created. If SanMgr created outside of PortsetMgr, then it wont cleanup up the SanMgr.

=over 4

=item Inputs

None

=item Outputs

Returns TRUE resource are created successful. If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

      #
      # Instantiate PortsetMgr Object 
            
      $PSMgr_obj = new PortsetMgr();
      if ($PSMgr_obj->error) {
         my $err_msg = $PSMgr_obj->errmsg();
         $Test->nlog("Failed to instantiate PortsetMgr object:\nReason: $err_msg");
         return $TCD::FAIL;
      }
      #
      # Call setup to create environment to testing
      #
      $PSMgr_obj->setup("-file" => full_path/thplname.xml);
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      ...
      #
      # Call cleanup 
      #
      $PSMgr_obj->cleanup();
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_sanmgr_obj()>

To get the SanMgr objects. It will return what Portset::setup created or user passed.

This is called by PortsetMgr::Setup, mostly no need to call it.

=over 4

=item Inputs

None

=item Outputs

Returns a reference to SanMgr object if successful. If an error occurred the error flag is set, errmsg is set to the failure reason,
and {} (empty hash reference) is passed back as the return value.

=item Examples

      #
      # Instantiate PortsetMgr Object (see above examples) 
            
      $PSMgr_obj = new PortsetMgr();
      ...
      ...
      ...
      #
      # get the SanMgr Object. 
      #
      my $SanMgr_obj = $PSMgr_obj->get_sanmgr_obj();
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_vserver_objs()>

To get the vserver objects that SanMgr::setup created/reused

NOTE;

This functions returns the vserver object which is/are created or reused by San::SanMgr:setup.
You get all the vservers in the controller by using SanFiler::get_vservers().
    

=over 4

=item Inputs

None

=item Outputs

Returns a reference to array or array of vserver objects if successful. If an error occurred the error flag is set, errmsg is set to the failure reason,
and empty array ([] or ()) is passed back as the return value.

=item Examples

      #
      # Instantiate PortsetMgr Object (see above examples) 
            
      $PSMgr_obj = new PortsetMgr();
      ...
      ...
      ...
      #
      # get the vserver objects created by setup 
      #
      my @Vserver_objects = $PSMgr_obj->get_vserver_objs();
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_igroup_objs(%args)>

To get the igroup objects that SanMgr::setup created/reused. It will return all igroup objects or specific protocol igroup objects on specified vserver

=over 4

=item Inputs

=item * I<-vserver_name> <optional> 

Name of the vserver. If you specify this, it returns only the igroups belongs to this vserver

=item * I<-protocol_type> <optional, default: all>

Name of the protocol type: valid inputs are: all|mixed|fcp|iscsi. returns only the igroups which matches the specified protocol type.

=item Outputs

Returns a reference to array or array of igroup objects if successful. If an error occurred the error flag is set, errmsg is set to the failure reason,
and empty array ([] or ()) is passed back as the return value.

=item Examples

      #
      # Instantiate PortsetMgr Object (see above examples) 
            
      $PSMgr_obj = new PortsetMgr();
      ...
      ...
      ...
      # get the all the igroup objects created by setup 
      my @igroup_objects = $PSMgr_obj->get_igroup_objs();
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # get the all the igroup objects in particular vserver
      # this will return all the igroups on vserver named "ps_vs" 
      my @igroup_objects = $PSMgr_obj->get_igroup_objs(-vserver_name => "ps_vs");
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # get only fcp igroup objects in particular vserver
      my @igroup_objects = $PSMgr_obj->get_igroup_objs(-vserver_name => "ps_vs",
                                                       -protocol_type => "fcp");
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # get all iscsi igroup objects 
      my @igroup_objects = $PSMgr_obj->get_igroup_objs(-protocol_type => "iscsi");
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # to use the igroup objects
      foreach my $igrp_obj (@igroup_objects) {
         $igrp_obj->get_name();
         ...
         ...   
      }

=back

=head2 B<get_protocol_lifs_name_in_vserver(%args)>

To  get all the specified 'protocol' lifs name for the given vserver

=over 4

=item Inputs

=item * I<-vserver_obj> <required> 

Reference to vserver object.

=item * I<-protocol_type> <required>

Name of the protocol type: valid inputs are: all|mixed|fcp|iscsi.

=item Outputs

Returns a string(with lif name(s)). If more than one lifs found
then it will return comma separated string.If an error occurred the error flag is set, 
errmsg is set to the failure reason, and (undef) is passed back as the return value.

=item Examples

      #
      # Instantiate PortsetMgr Object (see above examples) 
            
      $PSMgr_obj = new PortsetMgr();
      $vs_obj = ...;
      ...
      ...
      ...
      # get the all *fcp* lif names in the specified vserver 
      my $fcp_lif_names = $PSMgr_obj->get_protocol_lifs_name_in_vserver(-vserver_obj => $vs_obj,
                                                                        -protocol_type => "fcp");
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ...
      ...
      # get the all *iscsi* lifs name in the specified vserver 
      my $iscsi_lif_names = $PSMgr_obj->get_protocol_lifs_name_in_vserver(-vserver_obj => $vs_obj,
                                                                        -protocol_type => "iscsi");
      if ($PSMgr_obj->error()) {
         $Test->nlog($PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }

=back

=head2 B<get_fcp_lifs_name_wwpn_in_vserver(%args)>

To get all the fcp lifs name and its wwpn for the given vserver

Info:
 
      1. fcp object need to be created by setup or any other function on 
         the given vserver, prior to this call
      
      2. returns all the fcp lifs name and wwpn on the given vserver,  
              includes existing lifs and setup created lifs.

=over 4

=item Inputs

=item * I<-vserver_obj> <required> 

Reference to vserver object, from which fcp lifs and wwpn will be return

=item Outputs

Returns two strings(one with fcp_lif_name and followed by fcp_wwpn). If more than one lifs/wwpn
then each string will be comma separated string.If an error occurred the error flag is set, 
errmsg is set to the failure reason, and (undef, undef) is passed back as the return value.


=item Examples

      #
      # Instantiate PortsetHelper Object (see above _new() pod)
      # or use PortsetMgr Object (see above create_portset_in_vserver() pod)
      #
      ....
      ....
      ....
      # get vserver object
      $vs_obj = ...;
      
      # get all available fcp lifs
      my ($fc_lifs, $fc_wwpns) = $PSMgr_obj->get_fcp_lifs_name_wwpn_in_vserver(
                                       '-vserver_obj' => $vs_obj,
                                      );
      if ($PSMgr_obj->error()) {
         $Test->nlog("Failed to get fcp lifs and wwpn: " . $PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      
      #
      # create fcp portset with fcp wwpn name
      #
      my $ps_attr_exp = $PSMgr_obj->create_portset_in_vserver(
                                                '-vserver_obj'   => $vs_obj,
                                                '-portset'       => "ps_fcp_empty", 
                                                '-protocol_type' => "fcp",
                                                '-wwpn'        => $fc_wwpns,
               '-expected_fcp_attr' => {
               'lif_name' => $fc_lifs,
               'lif_wwpn' => $fc_wwpns
             } );
      if ($PSMgr_obj->error()) {
         $Test->nlog("Failed to create portset: " . $PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ....
      ....

=back
            
=head2 B<get_iscsi_lifs_name_tpg_in_vserver(%args)>

To get all the iscsi lifs name and its tpg name for the given vserver

Info:
 
      1. iscsi object need to be created by setup or any other function on 
         the given vserver, prior to this call
      
      2. returns all the iscsi lifs tpg name on the given vserver,  
              includes existing lifs and setup created lifs.

=over 4

=item Inputs

=item * I<-vserver_obj> <required> 

Reference to vserver object, from which iscsi lifs tpg name will be return

=item Outputs

Returns two strings(one with iscsi_lif_name and followed by iscsi_lif_tpg_name). If more than one lifs/tpg
then each string will be comma separated string.If an error occurred the error flag is set, 
errmsg is set to the failure reason, and (undef, undef) is passed back as the return value.

=item Examples

      #
      # Instantiate PortsetHelper Object (see above _new() pod)
      # or use PortsetMgr Object (see above create_portset_in_vserver() pod)
      #
      ....
      ....
      ....
      # get vserver object
      $vs_obj = ...;
      
      # get all available fcp lifs
      my ($iscsi_lif_names, $iscsi_lif_tpg_names) = $PSMgr_obj->get_iscsi_lifs_name_tpg_in_vserver(
                                       '-vserver_obj' => $vs_obj,
                                      );
      if ($PSMgr_obj->error()) {
         $Test->nlog("Failed to get iscsi lifs/tpg name: " . $PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      
      #
      # create iscsi portset with iscsi using iscsi lif's tpg name
      #
      my $ps_attr_exp = $PSMgr_obj->create_portset_in_vserver(
                                                '-vserver_obj'   => $vs_obj,
                                                '-portset'       => "ps_iscsi", 
                                                '-protocol_type' => "iscsi",
                                                '-tpg'        => $iscsi_lif_tpg_names);
      if ($PSMgr_obj->error()) {
         $Test->nlog("Failed to create portset: " . $PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ....
      ....

=back

=head2 B<create_tpgroup_on_vserver(%args)>

To create tpgroup on given the given vserver
NOTE: Cleanup will destroy the tpgroup, if created successfully with this method

=over 4

=item Inputs

=item * I<-vserver_obj> <required> 

Reference to vserver object to create tpgroup

=item * I<-tpgroup_name> <required> 

Name of the tpgroup to create on the given vserver


=item * I<-iscsi_lif_name> <optional> 

Name of the iscsi lif name to add to tpgroup when creating. If this is not passed, 
it will create empty tpgroup.

=item Outputs

Returns the tpgroup object, if successful. If an error occurred the error flag is set, 
errmsg is set to the failure reason, and empty reference {} passed back as the return value.

=item Examples

      #
      # Instantiate PortsetHelper Object (see above _new() pod)
      # or use PortsetMgr Object (see above create_portset_in_vserver() pod)
      #
      ....
      ....
      ....
      # get vserver object
      $vs_obj = ...;
      
      # to create empty tpgroup
      my $tpg_obj = $PSMgr_obj->create_tpgroup_on_vserver(
                                       '-vserver_obj' => $vs_obj,
                                       '-tpgroup_name' => "user_empty_tpgroup"
                                      );
      if ($PSMgr_obj->error()) {
         $Test->nlog("Failed to create empty tpgroup: " . $PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      
      # to create tpgroup with iscsi lif
	  my $iscsi_lif_name = ...;
	  
      my $tpg_obj = $PSMgr_obj->create_tpgroup_on_vserver(
                                       '-vserver_obj' => $vs_obj,
                                       '-tpgroup_name' => "user_empty_tpgroup",
                                       '-iscsi_lif_name' => $iscsi_lif_name
									   
                                      );
      if ($PSMgr_obj->error()) {
         $Test->nlog("Failed to create tpgroup with lif: " . $PSMgr_obj->errmsg());
         return $TCD::FAIL;
      }
      ....
      ....

=back

=head2 B<is_tpgroup_on_vserver_empty(%args)>

To determine the tpgroup is an empty tpgroup or not.

=over 4

=item Inputs

=item * I<-vserver_name> <required> 

Name of the vserver

=item * I<-tpgroup_name> <required> 

Name of the tpgroup to create on the given vserver


=item Outputs

Returns TRUE if the given tpgroup has no lif assign to it, in other words if tpgroup is empty. If an error occurred the error flag is set, errmsg is set to the failure reason, and FALSE
is passed back as the return value.

=item Examples

      #
      # Instantiate PortsetHelper Object (see above _new() pod)
      # or use PortsetMgr Object (see above create_portset_in_vserver() pod)
      #
      ....
      ....
      ....
      # get vserver object
      $vs_obj = ...;
	  $vs_name = $vs_obj->get_name();
      
      # to create empty tpgroup
      if( $PSMgr_obj->is_tpgroup_on_vserver_empty(
                                       '-vserver_name' => $vs_name,
                                       '-tpgroup_name' => "user_empty_tpgroup"
                                      ) == FALSE){
          if ($PSMgr_obj->error()) {
             $Test->nlog("error to verify, tpgroup empty: " . $PSMgr_obj->errmsg());
             return $TCD::FAIL;
          }else {
             $Test->nlog("tpgroup is not empty");
          }
      }
      ....
      ....

=back
=cut

