# Copyright (c) 2007-2010 NetApp Inc.
# All rights reserved.
#
## @summary Provides the core set of methods that provide common interface
## as a base class in order to manage system objects
## (ex. filer, aggregate, volume, etc.)
## All UI objects will support:
##  * discovery and caching of related objects
##  * creation and caching of resource objects
##  * retrieval of cached objects
##  * refresh of resource objects
##  * deletion of resource objects
##
## @author korry@netapp.com, dl-san-automation
## @status public

package San::CommonInterface;

# compiler directives
use strict;

# imported modules
use San::ConnectionMgr;
use San::Schema;
use Tharn qw(say param disconnect);
use Error;
use Exporter;
use San::SanLog;
use San::SanUtil;
use San::SanConsts;
use San::SanCache;
use Module::Common qw( validate_mixed );
use Params::Validate qw( :all );
use Params::Validate qw( SCALAR ARRAYREF HASHREF );
use Time::HiRes qw( gettimeofday tv_interval );
use List::MoreUtils;
use File::Basename qw( basename );
use File::Copy qw( copy );
use File::Path qw( make_path remove_tree );
use CmdsDriver;
use Subtest;
use Data::Dumper;

use vars qw( @EXPORT @ISA @Known_cdef_locations %Show_cmds );

@ISA = qw( Error SanLog SanUtil SanCache SanConsts);

# constant variable
use constant DEFAULT_CMDS    => 'lun,igroup,iscsi,fcp';
use constant DEFAULT_VSERVER => 'vs0';
use constant DIAG_USER => 'diag';
use constant DEFAULT_PASSWD => 'netapp1!';
use constant SSAN_CDEF_PATH  => '/x/eng/rtp/san_automation/cdefs/';
use constant RLSE_DOT_PATH   => '/x/eng/rlse/DOT/';
use constant CMD_TIMEOUT => 60;
use constant DISCOVER_TIMEOUT => 300;

# global variables
my $Connection_lost_str = 'Connection reset by peer';
my $Noentry_str = 'There are no entries';
my $Timeout_str = 'Timeout waiting for';
my $Changenum_file = 'P4SYNC';
my $Cmd_obj = undef;
my $Cmd_interface = undef;
my $Debug_level = undef;
my $Enable_cache = undef;
my @Alias_cdefs = ('vserver_iscsi_portal.cdefs');
my @Known_cdef_locations = ('/mgmtgateway/bedrock/obj/x86_64/src/svcmgr/sm/',
                            '/mgmtgateway/bedrock/obj/x86_64/src/tables/',
                            '/smf/bedrock/obj/x86_64/src/snmp/smftables/',
                            '/smf/bedrock/obj/x86_64/src/glue/',
                            '/smf/bedrock/export/common/smf/cdefs/',
                            '/mgmtgateway/bedrock/export/common/smf/cdefs/',
                            '/final/bedrock/export/common/cdefs/',
                            );

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method Name: _new
# Objective: Common new contructor to create object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($class, %args) = @_;

    my ($conn, $debug_level, $self, $schema_obj, $path, @list_fields, $enable_cache);
    my ($username, $password, $cmd_interface, $conn_type, $reuse_conn, $is_cserver);
    my $new_cmds_obj;

    $self = {};
    bless $self, $class;

    $self->_clear_err();

    # default parameters
    $reuse_conn  = $SanUtil::TRUE;
    unless (defined $Debug_level) { $Debug_level = &Tharn::param('DEBUG_LEVEL', '-default', 0); }
    unless (defined $Cmd_interface) { $Cmd_interface = 'cli'; }
    unless (defined $Enable_cache) { $Enable_cache = 0; }
    $self->{is_cserver} = $SanUtil::FALSE;

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # 7mode used $interface param instead of $cmd_interface
    if (exists $args{interface}) {
        $args{cmd_interface} = delete $args{interface};
    }

    # Simple Flag to create a new cmds driver
    if (defined $args{'new_cmds_obj'}) {
        $new_cmds_obj = delete $args{'new_cmds_obj'};
    } else {
        $new_cmds_obj = $SanUtil::FALSE;
    }

    # process arguments
    # name      - object name to be instantiated
    # filer     - controller name
    # buildinfo - controller build information
    while (my ($key, $value) =  each(%args)) {
        if ($key =~ /-?name$/) {
            # name could be empty, so continue on because it will discover all
            # objects such as aggregates, volumes, etc.
            $self->{name} = $value;
            next;
        } elsif ($key =~ /-?filer$/) {
            if (! length($self->{$key} = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
            next;
        } elsif ($key =~ /-?buildinfo$/) {
            if (! length($self->{$key} = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
            next;
        } elsif ($key =~ /-?username$/) {
            if (! length($username = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
            next;
        } elsif ($key =~ /-?password$/) {
            if (! length($password = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
            next;
        } elsif ($key =~ /-?cmd_interface$/) {
            if (! length($cmd_interface = $value) || $value !~ /cli|zapi/i) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
            next;
        } elsif ($key =~ /-?debug_level$/) {
            if (! length($debug_level = $value) || $value !~ /\d+/i) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
            next;
        } elsif ($key =~ /-?enable_cache$/) {
            if (! length($enable_cache = $value) || $value !~ /$SanUtil::TRUE|$SanUtil::FALSE/) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
            next;
        } elsif ($key =~ /-?is_cserver$/) {
            if (! length($self->{is_cserver} = $value) || $value !~ /$SanUtil::TRUE|$SanUtil::FALSE/) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
            next;
        } elsif ($key =~ /-?reuse_conn$/) {
            if (! length($reuse_conn = $value) || $value !~ /$SanUtil::TRUE|$SanUtil::FALSE/) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
            next;
        }
    }

    # set global debug level
    if (defined $debug_level && $debug_level =~ /\d+/) {
        $Debug_level = $debug_level;
    }

    # set debug level of an object
    $self->set_debug_level($Debug_level);

    # set global cmd interface
    if (defined $cmd_interface && $cmd_interface =~ /^(cli|zapi)$/i) {
        $Cmd_interface = $cmd_interface;
    }

    # set cmd_interface
    $self->{cmd_interface} = $Cmd_interface;

    # set global caching flag
    if (defined $enable_cache && $enable_cache =~ /^($SanUtil::TRUE|$SanUtil::FALSE)$/i) {
        $Enable_cache = $enable_cache;
    }

    # set enable_cache flag
    $self->{enable_cache} = $Enable_cache;

    # verify mandatory parameters
    unless (defined $self->{filer} and length $self->{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }

    # dynamically create 'get' accessor public methods for each field
    @list_fields = qw( build branch debug_level filer node release buildinfo cmd_interface cmd_obj path);
    $self->create_accessor_methods(-type => "get_attr", -list => \@list_fields);
    $self->sanlog(-msg_level => 3,
                  -msg       => "Create get accessor methods for: " . join(" ", @list_fields));

    # store elements in the fields list
    foreach my $key (sort keys %args) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # retrieve existing connection object if exist or
    # instantiate a new one
    if (defined $args{conn} and ref($args{conn}) =~ /ConnectionMgr/) {
        $self->{conn_obj} = $args{conn};
    } else {
        $self->{conn_obj} = $self->get_conn_obj(-reuse_conn => $reuse_conn);
        if ($self->error()) {
            $self->_set_err("get connect object failed", $self->errmsg());
            return $self;
        }
    }

    # korry - need to add zapi suport (next phase)

    # if connection does not exist
    # make connection and set environment for SUT to be run as
    unless ($self->{conn_obj}->is_exist(-filer     => $self->get_filer(),
                                        -cmd_interface => $self->get_cmd_interface())) {

        # without specifying username and passwd, _connect will
        # use default setting from tharnhost file.
        # without conn_type, _connect will default to 'ssh'
        $self->_connect($self->get_filer(), $self->get_cmd_interface(), $username, $password, $conn_type);
        if ($self->error()) {
            $self->_set_err("create connection failed", $self->errmsg());
            return $self;
        }

        # this was outside of this block but we may not need after
        # the first call
        # get controller build info and cache it if hasn't
        # for detail of build info, see _get_build_info()
        $self->_get_build_info();
        if ($self->error()) {
            $self->_set_err("get controller build info failed", $self->errmsg());
            return $self;
        }
    }

    # get connection
    $conn = $self->{conn_obj}->get_connection(-filer         => $self->get_filer(),
                                              -cmd_interface => $self->get_cmd_interface());
    unless (ref($conn) =~ /Connectrec/i) {
        $self->_set_err("failed to get connection");
        return $self;
    }

    # create cmdsDriver object and load its neccessary data if it hasn't
    if( !defined($Cmd_obj) || $new_cmds_obj) {
        $Cmd_obj = new CmdsDriver(-cmds => $self->DEFAULT_CMDS,
                                  -buildinfo => $self->{buildinfo},
                                  -connection => $conn,
                                  -cmd_interface => $self->get_cmd_interface(),
                                  -debug_level => $self->get_debug_level());
        if ($Cmd_obj->error()) {
            $self->_set_err("failed to instantiate CmdsDriver", $Cmd_obj->errmsg());
            return $self;
        }

        # set filer version and interface for CmdsDriver
        $Cmd_obj->filer_version($self->{version});
        $Cmd_obj->interface($self->get_cmd_interface());
        $Cmd_obj->log_output(1);
        $Cmd_obj->{conn_obj} = $self->{conn_obj};
    }

    # workaround for a global CmdsDriver object
    unless (defined $Cmd_obj->{cmode_schema}->{$self->get_filer()}) {
        # in order to reuse CmdsDriver and Schema for various nodes, we need to
        # set the schema for this controller/filer
        # NOTE: once we implemented cserver, this probably isn't required
        $Cmd_obj->{cmode_schema}->{$self->get_filer()} = $self->get_branch();
    }

    # store CmdsDriver object
    $self->{cmd_obj} = $Cmd_obj;

    return $self;
}

###############################################################################
# Method Name: create_accessor_methods
# Objective: solution for creating accessor methods based on type.
#            (i.e. type could be 'get_fields', 'get_objs', or 'discover'
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_accessor_methods {
    my ($self, %args) = @_;

    my ($type, $pkg, $list, @list_fields);

    # process pkg and obj_fields
    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);
    #$self->entry(-msg_level => 1);

    # parse and validate arguments
    # type  - type of method to be created
    # pkg   - package/module that calls this method
    # list  - singular or arrayref that contains list of names
    while (my ($key, $value) =  each(%args)) {
        if ($key =~ /-?type$/) {
            if (! length($type = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $SanUtil::FALSE;
            }
        } elsif ($key =~ /-?pkg$/) {
            if (! length($pkg = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $SanUtil::FALSE;
            }
        } elsif ($key =~ /-?list$/) {
            if (! length($list = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $SanUtil::FALSE;
            }
        }
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg =>$self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 7);

    # must have valid pkg to create discover and get obj methods
    if (($type eq "get_obj") and (! defined($pkg) or ! length($pkg))) {
        $self->_set_err("required parameter \"-pkg\" to create get objs");
        return $SanUtil::FALSE;
    }

    # args{list} could be SCALAR for single and ARRAYREF for multiple
    # store in an array
    if (ref($list) eq 'ARRAY') {
        @list_fields = @{$list};
    } else {
        @list_fields = ($list);
    }

    # create 'get' accessor method based on array of fields
    # set 'no strict' due to the dereference used to create
    # accesssor methods dynamically
    if ($type eq "get_attr") {
        foreach my $field (@list_fields) {
            no strict 'refs';

            # for each field, create a get_<field> method
            *{"get_" . lc $field} = sub {
                my ($self) = @_;

                # need to add check for the defined field

                # reset restrict rule
                use strict 'refs';

                # is field exist?
                unless (exists $self->{$field}) {
                    $self->_set_err($field . " does not exist");
                    return "";
                }

                # return array if the value of field is an array
                # otherwise return the value of field
                if (ref($self->{$field}) =~ /ARRAY/) {
                    return (wantarray()) ? @{$self->{$field}} : $self->{$field};
                }
                return $self->{$field};
            };
        }
        # cache the fields array
        push @{$self->{fields}}, @list_fields;

    } elsif ($type eq "get_obj") {
        # at this time, the get_obj will create both discover and get_obj
        # we're think about having discover inside get_obj methods

        # ex: get_agregate_objs, get_vserver_objs, get_volume_objs, etc.
        foreach my $field (@list_fields) {
            # create get_obj based on list_fields
            no strict 'refs';

            # change field to lowevercase
            $field = lc($field);

            # display accessor methods that were created
            $self->sanlog(-msg => "create accessor methods: discover_" . $field,
                          -msg_level => 3);

            # create discover obj for singular
            # ex: discover_aggregate, discover_vserver, discover_disk, etc.
            *{"discover_" . $field} = sub {
                my ($self, %d_args) = @_;

                my ($pkg_name, $new_obj, $cdef);

                $self->_clear_err();

                # reset restrict rule
                use strict 'refs';

                # cache image here to pass in Vserver
                if (defined $self->{buildinfo}) {
                    $d_args{buildinfo} = $self->get_buildinfo();
                }

                # remove dash '-' from input parameters because parameters
                # may be used to create get accessor method on the fly
                %d_args = &SanUtil::strip_dash(%d_args);

                # since this method can be used at multiple levels, we need to
                # add the parent object field. For example, the parent object for
                # volume is vserver so we need to pass in vserver => vserver_name
                SWITCH: {
                    ($field =~ /^(volume|igroup|portset|iscsi|fcp)$/i) && do {
                        # append parent object for volume
                        %d_args = (%d_args,
                                   vserver => $self->get_name());
                        last SWITCH;
                    };
                    ($field =~ /^lif$/i) && do {
                        # for lif, network interface show requires server param
                        # instead of vserver (I think it's a burt)
                        %d_args = (%d_args,
                                   server  => $self->get_name());
                        last SWITCH;
                    };
                    ($field =~ /^(lun|qtree|snapshot)$/i) && do {
                        # append parent object for lun,qtree, and snapshot
                        # lun could be contained in volume or qtree. therefore
                        # we need to determine if the discover is called from qtree
                        # volume object
                        if (ref($self) =~ /Qtree/i) {
                            %d_args = (%d_args,
                                       qtree  => $self->get_name(),
                                       volume  => $self->get_volume(),
                                       vserver => $self->get_vserver());
                        } else {
                            %d_args = (%d_args,
                                       volume  => $self->get_name(),
                                       vserver => $self->get_vserver());
                        }
                        last SWITCH;
                    };
                    ($field =~ /^tpgroup$/i) && do {
                        # append parent object for tpgroup
                        %d_args = (%d_args,
                                   vserver => $self->get_vserver());
                        last SWITCH;
                    };
                }

                # remove Qtree from package name because qtree and lun are at
                # the same level. The reason is both Qtree and Lun modules resided under
                # Volume. So, if Lun is created in Qtree, meaning Lun is sub-component
                # of Qtree, we have to chop off Qtree so we can import Lun module
                # correctly.
                $pkg =~ s/(::Qtree)$//;

                # get the package
                # San::ControllerCluster::Filer10M::Aggregate
                # $pkg   - San::ControllerCluster::Filer10M
                # $field - aggregate
                $pkg_name =  $pkg . "::" . ucfirst($field);

                $self->sanlog(-msg => "import library: " . $pkg_name,
                              -msg_level => 3);
                # create library path and its module to be imported at runtime
                my $lib = $pkg_name;
                # convert from San::ControllerCluster::Filer10M
                # to San/ControllerCluster/Filer10M and import
                $lib =~ s|\:\:|\/|g;
                require $lib . ".pm";

                $self->sanlog(-msg => "d_arg hash discover_" . $field . "s: " .
                              $self->dumper_get(\%d_args, $self->get_debug_level()),
                              -msg_level => 7);

                # param defined: -pkg, -cdef, and other %args
                # discover objs, update object data, and store in parent object
                my ($obj) = $self->discover(-pkg => $pkg_name,
                                            -cdef => $d_args{cdef},
                                            -timeout => $self->DISCOVER_TIMEOUT,
                                            %d_args);
                if ($self->error()) {
                    $self->_set_err("fail to discover object: $field " . $d_args{name}, $self->errmsg());
                    return {};
                }

                # push onto object list
                push @{$self->{$field . "_objs"}}, $obj;
                return $obj;
            };

            # display accessor methods that were created
            $self->sanlog(-msg => "create accessor methods: discover_" . $field . "s",
                          -msg_level => 3);
            # create discover obj
            # ex: discover_aggregates, discover_vservers, discover_disks, etc.
            *{"discover_" . $field . "s"} = sub {
                my ($self, %d_args) = @_;

                my ($pkg_name, $new_obj, $cdef);

                #$self->entry(-msg => "object $field",
                #             -msg_level => 1);

                $self->_clear_err();

                # reset restrict rule
                use strict 'refs';

                # cache image here to pass in Vserver
                if (defined $self->{buildinfo}) {
                    $d_args{buildinfo} = $self->get_buildinfo();
                }

                # remove dash '-' from input parameters because parameters
                # may be used to create get accessor method on the fly
                %d_args = &SanUtil::strip_dash(%d_args);

                # since this method can be used at multiple levels, we need to
                # add the parent object field. For example, the parent object for
                # volume is vserver so we need to pass in vserver => vserver_name
                SWITCH: {
                    ($field =~ /^(volume|igroup|portset|iscsi|fcp)$/i) && do {
                        # append parent object for volume
                        %d_args = (%d_args,
                                   vserver => $self->get_name());
                        last SWITCH;
                    };
                    ($field =~ /^lif$/i) && do {
                        # for lif, network interface show requires server param
                        # instead of vserver (I think it's a burt)
                        %d_args = (%d_args,
                                   server  => $self->get_name());
                        last SWITCH;
                    };
                    ($field =~ /^(lun|qtree|snapshot)$/i) && do {
                        # append parent object for lun,qtree, and snapshot
                        # lun could be contained in volume or qtree. therefore
                        # we need to determine if the discover is called from qtree
                        # volume object
                        if (ref($self) =~ /Qtree/i) {
                            %d_args = (%d_args,
                                       qtree  => $self->get_name(),
                                       volume  => $self->get_volume(),
                                       vserver => $self->get_vserver());
                        } else {
                            %d_args = (%d_args,
                                       volume  => $self->get_name(),
                                       vserver => $self->get_vserver());
                        }
                        last SWITCH;
                    };
                    ($field =~ /^tpgroup$/i) && do {
                        # append parent object for tpgroup
                        %d_args = (%d_args,
                                   vserver => $self->get_vserver());
                        last SWITCH;
                    };
                }

                # remove Qtree from package name because qtree and lun are at
                # the same level. The reason is both Qtree and Lun modules resided under
                # Volume. So, if Lun is created in Qtree, meaning Lun is sub-component
                # of Qtree, we have to chop off Qtree so we can import Lun module
                # correctly.
                $pkg =~ s/(::Qtree)$//;

                # get the package
                # San::ControllerCluster::Filer10M::Aggregate
                # $pkg   - San::ControllerCluster::Filer10M
                # $field - aggregate
                $pkg_name =  $pkg . "::" . ucfirst($field);

                $self->sanlog(-msg => "import library: " . $pkg_name,
                              -msg_level => 3);
                # create library path and its module to be imported at runtime
                my $lib = $pkg_name;
                # convert from San::ControllerCluster::Filer10M
                # to San/ControllerCluster/Filer10M and import
                $lib =~ s|\:\:|\/|g;
                require $lib . ".pm";

                $self->sanlog(-msg => "d_arg hash discover_" . $field . "s: " .
                              $self->dumper_get(\%d_args, $self->get_debug_level()),
                              -msg_level => 7);

                # param defined: -pkg, -cdef, and other %args
                # discover objs, update object data, and store in parent object
                $self->{$field . "_objs"} = $self->discover(-pkg => $pkg_name,
                                                            -cdef => $d_args{cdef},
                                                            -timeout => $self->DISCOVER_TIMEOUT,
                                                            %d_args);
                if ($self->error()) {
                    $self->_set_err("discover objects failed for: " . $field, $self->errmsg());
                    return $SanUtil::FALSE;
                }

                return $SanUtil::TRUE;
            };

            $self->sanlog(-msg => "create accessor methods: get_" . $field . "_objs",
                          -msg_level => 3);
            *{"get_" . $field . "_objs"} = sub {
                my ($self, %go_args) = @_;

                my (@objs, %dis_args, $cdef, $key);

                # reset restrict rule
                use strict 'refs';

                $self->_clear_err();

                # remove dash '-' from input parameters because parameters
                # may be used to create get accessor method on the fly
                %go_args = &SanUtil::strip_dash(%go_args);

                # korry - perhaps we should check for dirty flag and re-discover
                # for now, let discover obj everytime we call get obj
                # set discover arguments
                if (defined $go_args{state}) {
                    $dis_args{state} = delete $go_args{state};
                }
                if (defined $go_args{owner}) {
                    $dis_args{owner} = delete $go_args{owner};
                }

                $self->sanlog(-msg => "dis_args hash in get_" . $field . "_objs: " .
                                      $self->dumper_get(\%dis_args, $self->get_debug_level()),
                              -msg_level => 7);

                # use SWITCH to set discover cdef command based on field
                SWITCH: {
                    ($field =~ /^(aggregate|disk)$/i) && do {
                        # create cdef for aggregate_show or disk_show
                        $cdef = "storage_" . $field . "_show";
                        $key = $field;
                        last SWITCH;
                    };
                    ($field =~ /^(vserver|volume|lun)$/i) && do {
                        # create cdef for vserver_show, volume_show, and
                        # lun_show
                        $cdef = $field . "_show";
                        $key = $field;

                        # if the field is "vserver", we only want
                        # type "cluster" vservers
                        if ($field =~ /vserver/i) {
                            if (exists($go_args{type})) {
                                $dis_args{type} = $go_args{type}
                            } else {
                                $dis_args{type} = "cluster|data";
                            }
                        }

                        last SWITCH;
                    };
                    ($field =~ /^(igroup|portset)$/i) && do {
                        # create cdef for lun_igroup_show
                        $cdef = "lun_" . $field . "_show";
                        $key = $field;
                        last SWITCH;
                    };
                    ($field =~ /^(iscsi|fcp)$/i) && do {
                        # create cdef for vserver_iscsi_show
                        $cdef = "vserver_" . $field . "_show";
                        $key = $field;
                        last SWITCH;
                    };
                    ($field =~ /^lif$/i) && do {
                        # create cdef for vserver_iscsi_show
                        $cdef = "network_interface_show";
                        $key = $field;
                        last SWITCH;
                    };
                    ($field =~ /^tpgroup$/i) && do {
                        # create cdef for vserver_iscsi_tpgroup_show
                        $cdef = "vserver_iscsi_" . $field . "_show";
                        $key = $field;
                        last SWITCH;
                    };
                    ($field =~ /^networkadapter$/i) && do {
                        # create cdef for network_port_show
                        $cdef = "network_port_show";
                        $key = "port";
                        last SWITCH;
                    };
                    ($field =~ /^fcpadapter$/i) && do {
                        # create cdef for fcp_adapter_show
                        $cdef = "network_fcp_adapter_show";
                        $key = "adapter";
                        last SWITCH;
                    };
                    ($field =~ /^(qtree|snapshot)$/i) && do {
                        # create cdef for volume_qtree_show and volume_snapshot_show
                        $cdef = "volume_" . $field . "_show";
                        $key = $field;
                        last SWITCH;
                    };
                }

                # set discover argument for cdef
                $dis_args{cdef} = $cdef;
                my $discover_obj;

                # if name is specified, we discover that object and done
                # if name has regex, we will discover all objects and use the regex
                # to search for matching objects.
                if (exists $go_args{name} and $go_args{name} !~ /(\^|\$)/) {
                    $dis_args{$key} = $go_args{name};
                    $discover_obj = "discover_" . $field;
                    my $obj = $self->$discover_obj(%dis_args);
                    if ($self->error()) {
                        $self->_set_err("failed to discover object: " .
                                        $field, $self->errmsg());
                        return {};
                    }
                    return $obj;
                }

                $discover_obj = "discover_" . $field . "s";

                # check for cache data valid
                # discover if objs field is not defined or empty
                # for disk objects, we will always rediscover regardless
                if (defined @{$self->{$field . "_objs"}} and
                     scalar(@{$self->{$field . "_objs"}} and
                     $self->{enable_cache} == $SanUtil::TRUE and
                     $field ne "disk")) {
                    # object field is available
                    # but does it contain the object we look for?
                    push my @tmp_objs, @{$self->{$field . "_objs"}};

                    # searching for the matching object(s)
                    if (keys %go_args) {
                        @tmp_objs = $self->search(objs => \@tmp_objs, %go_args);
                        # find no matching object in cache, so let's rediscover
                        if (scalar(@tmp_objs) == 0) {
                            $self->$discover_obj(%dis_args);
                            if ($self->error()) {
                                $self->_set_err("failed to discover object: " .
                                                $field, $self->errmsg());
                                return (wantarray()) ? () : [];
                            }
                        }
                    }
                } else {
                    # discover objects
                    # set discover_<field>s and make the call to that
                    # specific obj
                    $self->$discover_obj(%dis_args);
                    if ($self->error()) {
                        $self->_set_err("failed to discover object: " . $field, $self->errmsg());
                        return (wantarray()) ? () : [];
                    }
                }

                # get the list of objects
                # if field is vserver, get the list of vserver_objs
                if (defined @{$self->{$field . "_objs"}}) {
                    push @objs, @{$self->{$field . "_objs"}};
                } else {
                    # log a message regarding no object found
                    $self->sanlog(-msg => "no object found for: " . $field,
                                  -msg_level => 3);
                    return (wantarray()) ? () : [];
                }

                # the search method will find the right object
                if (keys %go_args) {
                    @objs = $self->search(objs => \@objs, %go_args);
                    if ($self->error()) {
                        # error is already set, bubble up
                        return (wantarray()) ? () : [];
                    }
                }

                # verify objs exist
                if (scalar(@objs) == 0) {
                    # log a message regarding no object found with matching
                    # requested criteria
                    $self->sanlog(-msg => "go_args hash in get_" . $field . "_objs: ",
                                  $self->dumper_get(\%go_args, $self->get_debug_level()),
                                  -msg_level => 7);
                    $self->sanlog(-msg => "no object found matching requested criteria for: " . $field,
                                  -msg_level => 3);
                    return (wantarray()) ? () : [];
                }

                # return one object or a list objects based a count of one or
                # a single object in the list
                return $objs[0] if ($go_args{count} == 1 || scalar(@objs) == 1);
                return (wantarray()) ? @objs : \@objs;

            };
        }
    } else {
        $self->_set_err("invalid type to create accessor method: " . $type);
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub create_accessor_methods

###############################################################################
# Method Name: get
# Objective: get accessor methods to object data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get {
    my ($self, $element) = @_;

    $self->_clear_err();

    unless (defined($element)) {
        $self->_set_err("missing element parameter.");
        return "";
    }

    # Todo: May need to revisit this, zapi's put dashes in everything!
    # since all keys and methods are replaced with '_', let's
    # replace '-' with '_'
    $element =~ s/-/_/g;

    if (defined($self->{$element})) {
        return $self->{$element};
    } else {
        $self->_set_err("'$element' element does not exist.");
        return "";
    }
}

###############################################################################
# Method Name: set
# Objective: set accessor methods to object data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set {
    my ($self, %args) = @_;

    $self->_clear_err();

    while (my ($key, $value) =  each(%args)) {
        $self->{$key} = $value;
    }
    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: search
# Objective: find the objects that match the requested criteria
# Details:   See POD documentation at the end of this file.
###############################################################################
sub search {
    my ($self, %args) = @_;

    my ($compile, $count, @objs, @results);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # default to max count if count is not defined
    $count = (defined $args{count}) ? $args{count} : 99999;

    # set error and return if no object list
    @objs = @{$args{objs}};

    # verify mandatory cmd argument
    if (not @objs) {
        $self->_set_err("object list is not availble to search for");
        return (wantarray()) ? () : [];
    }

    # remove unneeded arguments from the argument list
    delete $args{objs};
    delete $args{count};

    # verify arguments from the list exist in the listed fields
    foreach my $key (keys %args) {
        unless (grep { $_ =~ /^$key$/i } $objs[0]->get_fields()) {
            $self->_set_err("invalid input key: " . $key);
            return (wantarray()) ? () : [];
        }
    }

    # search for the matching object based on argument keys
    foreach my $key (keys %args) {
        $compile = qr/$args{$key}/;                              # quote regex
        @results = grep ({$_->{lc $key} =~ /$compile/} @objs);
        @objs    = @results;                                     # matched result
    }

    # verify the found objects meet the requested criteria
    if (not @objs) {
        $self->_set_err("no object found matching requested criteria");
        return (wantarray()) ? () : [];
    } elsif ($count > scalar(@objs) and $count ne 99999) {
        $self->_set_err(  "too few objects found matching requested criteria. "
                        . $count
                        . " objs requested but only "
                        . scalar(@objs)
                        . " found.");
        return (wantarray()) ? () : [];
    }

    # set count to the number of objects found
    $count = scalar(@objs) if ($count == 99999);

    return (wantarray()) ? @objs[0 .. ($count - 1)] : \@objs[0 .. ($count - 1)];
} ## end sub search

###############################################################################
# Method Name: get_fields
# Objective: Gets a list of attributes of an object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fields {
    my ($self) = @_;

    $self->_clear_err();

    return (wantarray()) ? @{$self->{fields}} : $self->{fields};
}

################################################################################
# Method Name: get_field_from_controller
# Objective: Gets a specific field from the controller
################################################################################
sub get_field_from_controller {
    my ($self, %args) = @_;

    my ($current_attr, $field);
    my $expected_attr = "attributes";

    $self->_clear_err();

    # remove the dash '-' from input args
    %args = &SanUtil::strip_dash(%args);

    # make sure we have a valid cdef
    unless (length($args{cmd})) {
        $self->_set_err("required parameter -cmd");
        return "";
    }
    # make sure we have a valid field
    if (length($args{field})) {

        # ontap looks for "fields" param (with plural)
        $args{fields} = delete $args{field};
    } else {
        $self->_set_err("required parameter -field");
        return "";
    }

    if (length($args{attr})) {
        $expected_attr = delete $args{attr};
    }

    my (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return "";
    }

    if (exists $output[0]->{$args{fields}}) {
        return $output[0]->{$args{fields}};
    } else {
        $self->_set_err("output does not contain field: " . $args{fields});
        return "";
    }
}

###############################################################################
# Method Name: get_conn_obj
# Objective: return handle to ConnectionMgr instance
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_conn_obj {
    my ($self, %args) = @_;

    return ConnectionMgr->new(%args);
}

###############################################################################
# Method Name: get_schema_obj
# Objective: return handle to Schema object instance
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_schema_obj {
    my ($self, %args) = @_;

    return Schema->new();
}

###############################################################################
# Method Name: execute
# Objective: execute command and return results as an array of hash records
# Details:   See POD documentation at the end of this file.
###############################################################################
sub execute {
    my ($self, %args) = @_;

    my ($cmd, $conn, $connMgr_obj, $errmsg, @records, $raw_data);
    my ($timeout, $sec_str, $received_prompt, $return_prompt, @prompts, $no_prompt);
    my ($start_time, $elapsed_time);
    my ($nocr, $no_verbose, $ignore_error_check);

    my @additional_say_args = ();
    my $buffer = undef;

    $self->_clear_err();

    # set default value
    $raw_data = $SanUtil::FALSE;
    $timeout  = (exists $self->{timeout}) ? $self->{timeout} : $self->CMD_TIMEOUT;
    $return_prompt = $SanUtil::FALSE;
    $no_prompt = $SanUtil::FALSE;
    $no_verbose = $SanUtil::FALSE;
    $ignore_error_check = $SanUtil::FALSE;

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # process parameters
    while (my ($key, $value) =  each(%args)) {
        if ($key =~ /-?cmd$/) {
            if (! length($cmd = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /-?raw_data$/) {
            if (! length($raw_data = $value) || $value !~ /$SanUtil::TRUE|$SanUtil::FALSE/) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /-?return_prompt$/) {
            if (! length($return_prompt = $value) || $value !~ /$SanUtil::TRUE|$SanUtil::FALSE/) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /-?timeout$/) {
            if (! length($timeout = $value) || $value !~ /\d+/) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /-?no_prompt$/) {
            if (! length($no_prompt = $value) || $value !~ /$SanUtil::TRUE|$SanUtil::FALSE/) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /-?prompts$/) {
            if (ref($value) ne "ARRAY" || ! scalar(@prompts = @{$value})) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /-?no_verbose$/) {
            if (! length($no_verbose = $value) || $value !~ /$SanUtil::TRUE|$SanUtil::FALSE/) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }
        } elsif ($key =~ /-?ignore_error_check$/) {
            if (! length($ignore_error_check = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }
        }
    }

    # verify mandatory cmd argument
    unless (defined $cmd and length $cmd) {
        $self->_set_err("mandatory parameter \"-cmd\" not specified ");
        return (wantarray()) ? () : [];
    }

    # retrieve handle to connectionMgr object
    # this should be get_conn_obj()
    if (defined $self->{conn_obj}) {
        $connMgr_obj = $self->{conn_obj};
    } else {
        $connMgr_obj = $self->get_conn_obj();
    }
    $conn = $connMgr_obj->get_connection(-filer         => $self->get("filer"),
                                         -cmd_interface => $self->get("cmd_interface"));
    unless (ref($conn) =~ /Connectrec/i) {
        $self->_set_err("failed to get connection");
        return (wantarray()) ? () : [];
    }

    # build the additional_say_args array with any other parameters that need to be
    # passed into Tharn::say()
    push(@additional_say_args, "-timeout", $timeout) if (defined($timeout));
    push(@additional_say_args, "-nocr") if ($nocr == $SanUtil::TRUE);
    push(@additional_say_args, "-noprompt") if ($no_prompt == $SanUtil::TRUE);
    push(@additional_say_args, "-noverbose") if ($no_verbose == $SanUtil::TRUE);

    # if requested to search for additional prompts, push them onto the arg list
    if (scalar(@prompts)) {
        push(@additional_say_args, "-prompt");
        push(@additional_say_args, @prompts);
        push(@additional_say_args, "-ignore", "$cmd");
        push(@additional_say_args, "-comment", "$cmd");
    }

    # is connection cserver type?
    if ($conn->hostp->hosttype() =~ /cserver/i) {

        # check connection is alive. if not, reconnect
        unless ($conn->is_responsive()) {
            $conn = $self->reconnect();
            if ($self->error()) {
                $self->_set_err("failed to reconnect ", $self->errmsg());
                return (wantarray()) ? () : [];
            }
        }
    }

    # set start times
    $start_time   = [Time::HiRes::gettimeofday()];

    # issue command and get results.
    eval {
        (undef, undef, $buffer, $received_prompt) = Tharn::say($conn, $cmd,
                                                                  -timeout => $timeout,
                                                                  @additional_say_args);
        $buffer =~ s/\cG//g;
    };
    if ($@) {
        $self->_set_err("failed to execute: " . $cmd , $@);
        return (wantarray()) ? () : [];
    }

    # get elapsed times
    $elapsed_time =  Time::HiRes::tv_interval($start_time);

    # debug traces from returned buffer
    $self->sanlog(-msg => "  --- buffer: $buffer", msg_level => 3);
    $self->sanlog(-msg => "  --- received_prompt: $received_prompt", msg_level => 3);

    # determine singular or plural second
    $sec_str = ($elapsed_time =~ /^(0|1)\./) ? "seconds" : "second";

    # place '<--' after each command execution
    $self->sanlog(-msg => "<-- (cmd: $cmd elapsed: $elapsed_time $sec_str)");

    # determine if the output buffer contains known error strings
    if ($buffer =~ /ERROR:|$Connection_lost_str|$Timeout_str/i and
        $ignore_error_check == $SanUtil::FALSE) {

        # extract only error message and set it
        my (@error_str) = $buffer =~ /[ERROR|Error]: (.*)|($Connection_lost_str)|($Timeout_str)/s;
        $self->_set_err("Known error strings found in output buffer: @error_str");
        return (wantarray()) ? ($buffer) : [$buffer];
    }

    # if raw_data flag is set, just return the output buffer
    if ($raw_data == $SanUtil::TRUE) {
        return (wantarray()) ? ($buffer) : [$buffer];
    }

    # if return prompt is set, return the received prompt as the first element in the return array
    # followed by the output buffer
    if ($return_prompt == $SanUtil::TRUE) {
        return (wantarray()) ? ($received_prompt, $buffer) : [$received_prompt, $buffer];
    }

    # if the execution is a job manager or background actions, there
    # won't be any record to decode
    if ($buffer eq ""
        or $buffer =~ /Job succeeded:/i
        or $buffer =~ /$Noentry_str/i
        or $buffer =~ /^(set |ngsh|exit)/i) {
        return (wantarray()) ? ($buffer) : [$buffer];
    }

    # decode the returned data into array of hash records
    @records = $self->cli_output_decode($buffer);
    if ($self->error()) {
        $self->_set_err("cli output decode failed", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @records : \@records;
} ## end sub execute

###############################################################################
# Method Name: invoke
# Objective: translate cdef schema into actual command and execute
# Details:   See POD documentation at the end of this file.
###############################################################################
sub invoke {
    my ($self, %args) = @_;

    my ($cmd, $cdef, $filer, $schema_obj, @records);
    my (%validate_params);
    my $update_attrs = $SanUtil::TRUE;
    my $raw_data = $SanUtil::FALSE;

    $self->_clear_err();

    # strip out the '-' dash of key arguments
    %args = &SanUtil::strip_dash(%args);
    my @input_params = (%args);

    # validation_options is set to allow extra params not defined in %args to bypass
    Params::Validate::validation_options( allow_extra => 1);

    # validate_params is the arguments to validate
    %validate_params = (
                      cdef         => { type     => SCALAR,
                                        required => 1 },
                      debug_level  => { type     => SCALAR,
                                        optional => 1,
                                        regex    => qr /^\d+$/ },
                      update_attrs => { type     => SCALAR,
                                        optional => 1,
                                        regex    => qr /^\d+$/ },
                      raw_data     => { type     => SCALAR,
                                        optional => 1,
                                        regex    => qr /^\d+$/ }
                      );
    eval {
        (%args) = Params::Validate::validate(@input_params, \%validate_params);
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: ". $@);
        return $SanUtil::FALSE;
    }

    # set and delete %args that CmdsDriver doesn't care
    if (exists $args{raw_data}) { $raw_data = delete $args{raw_data}; }
    if (exists $args{update_attrs}) { $update_attrs = delete $args{update_attrs}; }
    if (exists $args{debug_level}) {
        $self->set_debug_level($args{debug_level});
        delete $args{debug_level};
    }

    # retrieve schema handle
    $schema_obj = $self->get_schema_obj();
    if ($self->error()) {
        $self->_set_err("failed to get schema object", $self->errmsg());
        return (wantarray()) ? (): [];
    }

    # translate cdef schema into command
    $cmd = $schema_obj->translate(filer => $self->get_filer(),
                                  %args);
    if ($schema_obj->error()) {
        $self->_set_err("translate cdef schema failed",
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $schema_obj->errmsg());
        return (wantarray()) ? (): [];
    }

    # execute command
    (@records) = $self->execute(-cmd => $cmd, -raw_data => $raw_data );
    if ($self->error()) {
        # error is already set
        return (wantarray()) ? (): [];
    }

    # if execution was successful then you should be able to
    # assume that the attribute/value pairs passed in are valid.
    # update object metadata with these attribute/value pairs if update_attrs
    # is enabled
    $self->_update(%args) if ($update_attrs == 1);

    return (wantarray()) ? @records: \@records;
} ## end sub invoke

###############################################################################
# Method Name: invoke_cmd_obj
# Objective: build cmd_driver hash, get cmd_obj, and execute cmd via cmd_obj
# Details:   See POD documentation at the end of this file.
###############################################################################
sub invoke_cmd_obj {
    my ($self, %args) = @_;

    my ($cmd, $cmd_obj, @records);
    my (%validate_params);
    my %cmd_args;
    my $update_attrs = $SanUtil::TRUE;
    my $raw_data = $SanUtil::FALSE;

    $self->_clear_err();

    # strip out the '-' dash of key arguments
    %args = &SanUtil::strip_dash(%args);
    my @input_params = (%args);

    # validation_options is set to allow extra params not defined in %args to bypass
    Params::Validate::validation_options( allow_extra => 1);

    # validate_params is the arguments to validate
    %validate_params = (
                      cmd          => { type     => SCALAR,
                                        required => 1 },
                      debug_level  => { type     => SCALAR,
                                        optional => 1,
                                        regex    => qr /^\d+$/ },
                      update_attrs => { type     => SCALAR,
                                        optional => 1,
                                        regex    => qr /^\d+$/ },
                      raw_data     => { type     => SCALAR,
                                        optional => 1,
                                        regex    => qr /^\d+$/ }
                      );
    eval {
        (%args) = Params::Validate::validate(@input_params, \%validate_params);
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: ". $@);
        return $SanUtil::FALSE;
    }

    # set and delete %args that CmdsDriver doesn't care
    if (exists $args{raw_data}) { $raw_data = delete $args{raw_data}; }
    if (exists $args{update_attrs}) { $update_attrs = delete $args{update_attrs}; }
    if (exists $args{debug_level}) {
        $self->set_debug_level($args{debug_level});
        delete $args{debug_level};
    }

    # get cmd driver object
    $cmd_obj = $self->get_cmd_obj();
    if ($self->error()) {
        $self->_set_err("failed to get cmd driver object", $self->errmsg());
        return (wantarray()) ? (): [];
    }

    # get conn rec from ConnectionMgr object
    my $conn = $self->{conn_obj}->get_connection(-filer => $self->get("filer"),
                                                 -cmd_interface => $self->get("cmd_interface"));

    my $output_format = ($raw_data == $SanUtil::TRUE) ? "raw" : "hash";

    # set interface
    my $interface = (exists $args{cmd_interface}) ? delete $args{cmd_interface} : $self->get_cmd_interface();

    # build cmd argument for CmdsDriver::execute_cmd()
    %cmd_args = (-cmd_filer      => $self->get_filer(),
                 -check_result   => 1,
                 -cmd_interface  => $interface,
                 -cmd_log_output => 1,
                 -cmd_connection => $conn,
                 -cmd_output_format => $output_format,
                 %args);

    # execute cmd using CmdsDriver object
    my $result_ref = $cmd_obj->execute_cmd(%cmd_args);
    if ($cmd_obj->error()) {
        $self->_set_err("failed to execute cmd: ",
                        $self->dumper_get(\%cmd_args, $self->get_debug_level()),
                        $cmd_obj->errmsg());
        return (wantarray()) ? (): [];
    }

    # log results if msg_level meets criteria
    $self->log_cmd_results($result_ref);

    # set return output
    my $output = $result_ref->{output};

    # Check the result from CmdsDriver ensure it's ok
    unless ($result_ref->{result_check}) {
        $self->_set_err("failed to check the result from output: ",
                        $output);
        return (wantarray()) ? () : [];
    }

    # if user wants raw_data, return output buffer
    if ($raw_data == $SanUtil::TRUE) {
        return (wantarray()) ? ($output) : [$output];
    }

    # decode result for cli or zapi
    if ($interface =~ /^cli$/i) {

        # if the execution is a job manager or background actions, there
        # won't be any record to decode
        if ($output eq ""
            or $output =~ /Job succeeded:/i
            or $output =~ /$Noentry_str/i
            or $output =~ /^(ngsh|exit)/i) {
            return (wantarray()) ? () : [];
        } else {

            # strip out line with only whitespace
            $output =~ s/\r\s+\r/\r/g;

            # decode cli output into hash
            (@records) = $self->cli_output_decode($output);
            if ($self->error()) {
                $self->_set_err("cli output decode failed", $self->errmsg());
                return (wantarray()) ? () : [];
            }
        }
    } else { #ZAPI
        # decode result from zapi output
        (@records) = $self->zapi_output_decode($args{cmd}, $output);
        if ($self->error()) {
            $self->_set_err("failed to decode zapi output", $self->errmsg());
            return (wantarray()) ? () : [];
        }
    }

    # dumping records for debugging purpose
    foreach my $record (@records) {
        $self->sanlog(-msg => $self->dumper_get($record, $self->get_debug_level()),
                      -msg_level => 7);
    }

    # if execution was successful then you should be able to
    # assume that the attribute/value pairs passed in are valid.
    # update object metadata with these attribute/value pairs if update_attrs
    # is enabled
    $self->_update(%args) if ($update_attrs == $SanUtil::TRUE);

    return (wantarray()) ? @records : \@records;
} ## end sub invoke_cmd_obj

###############################################################################
# Method Name: run_node_scope
# Objective: provide a method to execute dblade (7-mode) commands via run node
#            This run_node_scope method will return output similar to 7-mode
# Details:   See POD documentation at the end of this file.
###############################################################################
sub run_node_scope {
    my ($self, %args) = @_;

    my ($node_scoped_cmd, $ret_buffer);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);
    eval {

        # validate input parameters and their type.
        # cmd   - dblade (7-mode) command to be executed via 'run node'
        # filer - filer that commad will be executed on
        (%args) =
          Module::Common::validate_mixed(@input_params,
                                         [],
                                         {cmd     => {type => SCALAR, required => 1},
                                          timeout => {type => SCALAR,
                                                      optional => 1,
                                                      default  => 300},
                                          priv    => {type => SCALAR,
                                                      optional => 1,
                                                      default  => ""},
                                          filer   => {type     => SCALAR,
                                                      optional => 1,
                                                      default  => $self->get_filer()}});
    };
    if ($@) {
        $self->_set_err("failed to validate parameters: " . $@);
        return (wantarray()) ? () : [];
    }

    # if the command does not contain an embedded priv set level, add the default priv level
    if ($args{cmd} !~ /^\s*priv set.*?;/) {
        $args{cmd} = "priv set -q $args{priv}; " . $args{cmd};
    }

    # create the run node scope command
    $node_scoped_cmd = "run -node " . $args{filer} . " -command \"" . $args{cmd} . "\"";

    # execute command with raw_data flag
    ($ret_buffer) = $self->execute(-cmd      => $node_scoped_cmd,
                                   -raw_data => $SanUtil::TRUE,
                                   -timeout  => $args{timeout});
    if ($self->error()) {
        $self->_set_err("failed to run node scope command: " . $node_scoped_cmd, $self->errmsg());
    }

    return (wantarray()) ? ($ret_buffer) : [$ret_buffer];
} ## end sub run_node_scope

###############################################################################
# Method Name: refresh
# Objective: refresh object based on cdef schema (ie. vserver_show)
#            This means the refresh method wil refresh either Cluster, Aggregate
#            Vserver, Volume, etc. objects
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    my (@records);
    my $record = {};

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd and filer arguments
    unless (defined $args{cmd} and length $args{cmd}) {
        $self->_set_err("mandatory parameter \"-cmd\" not specified ");
        return $SanUtil::FALSE;
    }

    # execute command and retrieve records
    (@records) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # if we found no record, set error and return false
    unless (scalar(@records)) {
        $self->_set_err("record does not exist for " . $args{cmd});
        return $SanUtil::FALSE;
    }

    # we only need one record to obtain its keys
    $record = $records[0];

    # cache record keys per each object
    # so these keys will be unique to a particular object
    foreach my $key (sort keys %{$record}) {
        $self->{lc($key)} = $record->{$key};
        push @{$self->{fields}}, lc($key);
    }

    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: discover
# Objective: discover existing system entities and create objects
# Details:   See POD documentation at the end of this file.
###############################################################################
sub discover {
    my ($self, %args) = @_;

    my ($cmd, $pkg, $filer, $cdef, $buildinfo, $status, @records, @objs);
    my ($parent_obj);

    # process arguments
    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # process arguments
    # pkg       - package name (ie. San::ControllerCluster::Aggregate)
    # filer      - controller name
    # buildinfo - controller build information
    # cdef      - command definition (ie. vserver_show)
   while (my ($key, $value) =  each(%args)) {
        if ($key =~ /-?pkg$/) {
            if (! length($pkg = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }

            # delete args that is not needed for translation
            delete $args{$key};
            next;
        } elsif ($key =~ /-?buildinfo$/) {
            if (! length($self->{$key} = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }

            # delete args that is not needed for translation
            delete $args{$key};
            next;
        } elsif ($key =~ /-?debug_level$/) {
            if (! length($value) || $value !~ /\d+/) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }
            $self->set_debug_level($value);

            # delete args that is not needed for translation
            delete $args{$key};
            next;
        } elsif ($key =~ /-?timeout$/) {
            if (! length($self->{$key} = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return (wantarray()) ? () : [];
            }

            # delete args that is not needed for translation
            delete $args{$key};
            next;
        } elsif ($key =~ /-?parent_obj$/) {
            if (! length($parent_obj = $value)) {
                $self->_set_err("invalid value for \"-$parent_obj\": " . $value);
                return (wantarray()) ? () : [];
            }

            # delete args that is not needed for translation
            delete $args{$key};
            next;
        } elsif ($key =~ /-?conn$/) {

            # we don't connection object anymore here
            # delete connection object prior to trasnlation
            delete $args{$key};
            next;
        }
    }

    # get filer if undefined
    $args{filer} = $self->get_filer() unless (defined($args{filer}));

    # verify mandatory parameter
    unless (defined $pkg and length $pkg) {
        $self->_set_err("mandatory parameter \"-pkg\" not specified ");
        return (wantarray()) ? () : [];
    }
    unless (defined $args{cdef} and length $args{cdef}) {
        $self->_set_err("mandatory parameter \"-cdef\" not specified ");
        return (wantarray()) ? () : [];
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => "CommonInterface::discover() command arguments: " .
                          $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 7);

    # invoke cdef command and ignore update attributes
    (@records) = $self->invoke(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke cdef: " . $args{cdef},
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }
    # if we found no record, return empty list but no error
    if ($records[0] =~ /$Noentry_str/i) {
        return (wantarray()) ? () : [];
    }

    # NOTE for Korry:
    # ok let's cache the records as well using SanCache
    $self->store_cache_data($args{cdef}, @records);

    # create object per each record found
    foreach my $record (@records) {
        # set filer attribute and cache other info for each rec
        $record->{filer}       = $self->get_filer();
        $record->{buildinfo}   = $self->get_buildinfo() if defined($self->{buildinfo});
        $record->{parent_obj}   = $self;
        $record->{initialized} = $SanUtil::TRUE;

        # instantiate new object and polulate record
        my $obj = $pkg->new(%$record);
        if ($obj->error()) {
            $self->_set_err("fail to populate object for " . $pkg, $obj->errmsg());
            return (wantarray()) ? () : [];
        }
        push @objs, $obj;
    }

    return (wantarray()) ? @objs : \@objs;
} ## end sub discover

###############################################################################
# Method Name: disable_req_field_check()
# Objective: clear/disable flag disables the verification of required field check
#            against cdef required fields
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable_req_field_check {
    my ($self) = @_;

    # get the schema object and clear the required field check flag
    $self->get_schema_obj()->disable_req_field_check();
}

###############################################################################
# Method Name: enable_req_field_check()
# Objective: set/enable flag enables the verification of required field check
#            against cdef required fields
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable_req_field_check {
    my ($self) = @_;

    # get the schema object and set the required field check flag
    $self->get_schema_obj()->enable_req_field_check();
}

###############################################################################
# Method Name: get_req_field_check()
# Objective: retrieve state of cmd's input requirement check
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_req_field_check {
    my ($self) = @_;

    # get the schema object and retrieve the state of required field check flag
    $self->get_schema_obj()->get_req_field_check();
}

###############################################################################
# Method Name: get_version
# Objective: get version on the controller (7-mode compatible)
# Details:   none
###############################################################################
sub get_version {
    my ($self) = @_;

    my ($version_str, $output, $cmd);

    $self->_clear_err();

    # execute command to obtain controller build information
    $cmd = "version";
    ($output) = $self->run_node_scope(-filer => $self->get_filer(), -cmd => $cmd);
    if ($self->error()) {
        $self->_set_err("failed to run_node_scope: " . $cmd, $self->errmsg());
        return "";
    }

    if ($output =~ /Release\s+(.*)\:\s+/i) {
        $version_str= $1;
    } else {
        $self->_set_err("output does not contain valid version: " . $version_str);
        return "";
    }

    return $version_str;

} ## end sub get_version

###############################################################################
# Method Name: get_shell_environment
# Objective: get shell environment
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_shell_environment {
    my ($self, %args) = @_;

    my ($conn, $cmd);

    $self->_clear_err();

    if (exists $self->{conn_obj}) {
        $conn = $self->{conn_obj}->get_connection(-filer => $self->get("filer"),
                                                  -cmd_interface => $self->get("cmd_interface"));
    } elsif (exists $self->{connection}) {
        $conn =  $self->{connection};
    } else {
        $self->_set_err("connection manager does not exist");
        return "";
    }

    if (ref($conn) =~ /Connectrec/i) {
        return $conn->{environment};
    } else {
        $self->_set_err("failed to get connection");
        return "";
    }
}

###############################################################################
# Method Name: set_shell_environment
# Objective: set shell environment between  bsd or ngsh
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_shell_environment {
    my ($self, %args) = @_;

    my ($conn, $cmd);

    $self->_clear_err();

    # process pkg and obj_fields
    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);
    eval {

        # validate input parameters and their type.
        # pkg  - package/module that calls this method
        # list - array that contains names of object to be created for
        (%args) =
          Module::Common::validate_mixed(@input_params,
                                         [],
                                         {type  => {type => SCALAR, required => 1},
                                          user => {type => SCALAR,
                                                   optional => 1,
                                                   default => "diag"}});
    };
    if ($@) {
        $self->_set_err("failed to validate parameters: " . $@);
        return $SanUtil::FALSE;
    }
    # get connection
    if (exists $self->{conn_obj}) {
        $conn = $self->{conn_obj}->get_connection(-filer => $self->get("filer"),
                                                  -cmd_interface => $self->get("cmd_interface"));
    } elsif (exists $self->{connection}) {
        $conn =  $self->{connection};
    } else {
        $self->_set_err("connection or connection  manager does not exist");
        return $SanUtil::FAIL;
    }

    # set shell environment if valid connection
    if (ref($conn) =~ /Connectrec/i) {
        if (lc($args{type}) eq lc($conn->{environment})) {
            $self->sanlog(-msg => "no need to switch because current shell environment is: " . $conn->{environment});
        } elsif ($args{type} =~ /^ngsh$/i) {
            if ($conn->{environment} =~ /^bsd$/i) {
                # execute ngsh command and set environment flag to ngsh
                if ($conn->{username} eq "diag") {
                    $cmd = "ngsh";
                    $self->execute(-cmd => $cmd);
                    # set the output format for ngsh environment
                    $self->_set_output_format();
                    if ($self->error()) {
                        $self->_set_err("failed to set output format", $self->errmsg());
                        return $SanUtil::FALSE;
                    }
                } else {
                    $cmd = "exit";
                    $self->execute(-cmd => $cmd);
                    if ($self->error()) {
                        $self->_set_err("Failed to switch to ngsh", $self->errmsg());
                        return $SanUtil::FALSE;
                    }
                }
            }
            if ($conn->{environment} =~ /^node_scope$/i) {
                $self->execute(-cmd => "exit");
            }
            $conn->{environment} = "ngsh";
        } elsif ($args{type} =~ /^bsd$/i) {
            # assume the environment is ngsh
            if ($conn->{username} eq "diag") {
                $self->execute(-cmd => "exit");
                if ($conn->{environment} =~ /^node_scope$/i) {
                    $self->execute(-cmd => "exit");
                }

            } else {
                # get out of node scope if we are currently there
                if ($conn->{environment} =~ /^node_scope$/i) {
                    $self->execute(-cmd => "exit");
                }

                # switch to bsd using the diag user and the password used
                # for the current connection
                my $node_passthru;
                if ($self->get("server_type") eq "node") {
                    $node_passthru = $self->get_filer();
                } else {
                    $node_passthru = $self->get_node();
                }
                $self->_switch_to_bsd(-node => $node_passthru,
                                      -username => DIAG_USER,
                                      -password => $conn->{password});
                if ($self->error()) {
                    $self->_set_err("Failed to switch to bsd", $self->errmsg());
                    return $SanUtil::FALSE;
                }
            }

            $conn->{environment} = "bsd";

        } elsif ($args{type} =~ /^node_scope$/i) {
            if ($conn->{environment} =~ /^bsd$/i) {
                # execute ngsh command and set environment flag to ngsh
                # $cmd = ($args{user} =~ /root/i) ? "ngsh" : "sudo ngsh";
                $cmd = "ngsh";
                $self->execute(-cmd => $cmd);
                if ($self->error()) {
                    $self->_set_err("failed to execute: " . $cmd, $self->errmsg());
                     return $SanUtil::FALSE;
                }
            }
            # if environment is not bsd, assume we are at ngsh
            # run node scope
            $cmd = "run -node " . $self->get_filer();
            my $buf = $self->execute(-cmd => $cmd, -raw_data => 1);
            if ($buf->[0] =~ /Too many users logged in/i) {
                $self->_set_err($buf->[0]);
                return $SanUtil::FAIL;
            }
            $conn->{environment} = "node_scope";
        } else {
            $self->_set_err("environment is " . $conn->{environment} . ". want to switch to " . $args{type});
            return $SanUtil::FAIL;
        }
    } else {
        $self->_set_err("failed to get connection");
        return $SanUtil::FAIL;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: cli_output_decode
# Objective: parse returned CLI output data into an array of hashes
# Details:   none
###############################################################################
sub cli_output_decode {
    my ($self, $output) = @_;

    my ($key, @fields, @lines, @recs, @values);

    # parse and convert output into array of key=>value hash
    # line[0]    - command
    # line[1]    - list of lowercase fields without space
    # line[2]    - fields descriptions
    # line[3 ..] - values of fields in line[1]
    @lines  = split("\r?\n", $output);
    @fields = split(/\|/,   $lines[1]);
    foreach my $line (@lines[3 .. scalar(@lines) - 1]) {
        if ($line !~ /\|/) {
            $self->sanlog(-msg => "cli_output_decode skipping: $line", -msg_level => 3);
            next;
        }
        push @values, split(/\|/, $line);
        next if (scalar(@values) < scalar(@fields));

        # there is a bug in the output so number of fields and values
        # are not equal
        # change to le (<=) until burt311268 is fixed
        # korry - fix the below line when burt311268 is fixed
        if (scalar(@fields) <= scalar(@values)) {

            # based on the index, store key => value pairs per record
            my $rec = {};
            foreach my $idx (0 .. scalar(@fields) - 1) {
                $key = $fields[$idx];

                # replace "-" with "_" because these fields will
                # be used to create get accessor methods on fly
                $key =~ s/-/_/g;
                $rec->{"$key"} = $values[$idx];
            }

            # store each record into array
            push @recs, $rec;
        } else {
            $self->_set_err("cli_output_decode unmatched fields = " . scalar(@fields) . ";values = " . scalar(@values));
            return (wantarray()) ? () : [];
        }

        # clear array values for the next record
        @values = ();
    }

    return (wantarray()) ? @recs : [@recs];
} ## end sub cli_output_decode

###############################################################################
# Method Name: zapi_output_decode
# Objective: parse returned ZAPI output data into an array of hashes that are
#            comparable to CLI output
# Details:   none
###############################################################################
sub zapi_output_decode {
    my ($self, $cmd, $output) = @_;

    # load table
    require "San/AttrTable.pm";

    # verify the command is in the table and obtain its attributes
    if (defined($Show_cmds{$cmd})) {
        # Get the attribute list for that specific command
        my %attrs = %{$Show_cmds{$cmd}};
        # if cmd is a show command, the  output (response) should have an'attributes'
        # tag, so decode it expecting that key.
        if ($cmd =~ /_show/) {
            # First verify that there are records to decode. So, if output
            # returned zero record, just return here
            if (exists $output->{'num-records'} and $output->{'num-records'} == 0) {
                return (wantarray()) ? () : [];
            }
            # Records to decode, sending the zapi response (output) and the info
            # for this specific command in the AttrTable.
            return $self->_zapi_decode_show_cmd($output, \%attrs);
        } else {
        # If cmd is NOT a show command, the structure of the output should not
        # have an 'attributes' tag structure. Need to use a different method
            return $self->_zapi_decode_non_show_cmd($output, \%attrs);
        }
    } else {
        # If it's a show command and is not found, throw an error
        if ($cmd =~ /_show/) {
             $self->_set_err("$cmd not found in the AttrTable.pm");
             return (wantarray()) ? () : [];
        } else {
        #if is NOT a show command, then return the output buffer.
            return (wantarray()) ? ($output) : [$output];
        }
    }
} ## end sub zapi_output_decode

###############################################################################
# Method Name: disconnect_conn
# Objective:   Disconnects the connection on the object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub disconnect_conn {
    my ($self, %args) = @_;

    my $filer = $self->get_filer();
    my $cmd_interface = $self->get_cmd_interface();

    my $conn = $self->get_conn_obj()->get_connection(filer => $filer, cmd_interface => $cmd_interface);

    Tharn::disconnect($conn,'-ok');
}

###############################################################################
# Method Name: reconnect
# Objective:   Creates a new filer connection. This is needed for sitiations
#              like post take/give where a SanFiler object exists but will have
#              a stale connection.  Returns the connection attribute.
# Details:     See POD documentation at the end of this file.
###############################################################################
sub reconnect {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my $conn;
    my $filer = $self->get_filer();
    my $cmd_interface = $self->get_cmd_interface();

    # disconnect the existing connection by default
    my $disconnect_existing_conn = exists($args{disconnect_existing_conn})
                                       ? $args{disconnect_existing_conn} : 1;

    # determine connection
    if (exists($args{connection})) {
        $conn = $args{connection};
    } else {
        $conn = $self->get_conn_obj()->get_connection(filer => $filer, cmd_interface => $cmd_interface);
    }

    if ($disconnect_existing_conn == $SanUtil::TRUE) {
        # disconnect connection
        Tharn::disconnect($conn);
    }

    $self->_connect($filer, $cmd_interface);
    if ($self->error()){
        $self->_set_err("failed to connect to $filer", $self->errmsg());
        return {};
    }

    $conn = $self->get_conn_obj()->get_connection(filer => $filer, cmd_interface => $cmd_interface);
    unless (ref($conn) =~ /Connectrec/i) {
        $self->_set_err("failed to get connection");
        return {};
    }

    # since reconnect can be inherited by CmdsDriver, so determine which
    # class calls reconnect and update CmdsDriver
    if (ref($self) =~/CmdsDriver/i) {
        $self->connection($conn);
    } else {
        # get CmdsDriver object and reset its connection
        my $cmd_obj = $self->get_cmd_obj();
        if ($self->error()) {
            $self->_set_err("failed to get cmd driver object", $self->errmsg());
            return {};
        }

        $cmd_obj->connection($conn);
    }

    return $conn;
}

###############################################################################
# Method Name: create_common_path
# Objective: create a path based on the arguments that pass in
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_common_path {
    my ($self, %args) = @_;

    my ($volume);
    my $path = "";

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # check for mandatory parameters
    if (defined $args{volume} and length $args{volume}) {
        $volume = delete $args{volume};
    } else {
        $self->_set_err("Mandatory parameter \"volume\" not specified ");
        return $path;
    }

    # get qtree param if exist
    my $qtree = (exists $args{qtree}) ? delete $args{qtree} : "";

    # get lun param if exist
    my $lun = (exists $args{lun}) ? delete $args{lun} : "";

    # build a path
    $path = "/vol/" . $volume;
    $path .= "/" . $qtree if (length($qtree) and $qtree ne '""');
    $path .= "/" . $lun if (length($lun) and $lun ne '""');

    return $path;
}

###############################################################################
# Method Name: _switch_to_bsd()
# Objective: Moves the current shell environment into BSD (from ngsh), using systemshell
# Details:     See POD documentation at the end of this file.
###############################################################################
sub _switch_to_bsd {
    my ($self, %args) = @_;

    my $node = exists($args{-node}) ? $args{-node} : $self->get_node();
    my $username = $args{-username};
    my $password = $args{-password};

    # retrieve handle to connectionMgr object
    # this should be get_conn_obj()
    my $connMgr_obj;
    if (defined $self->{conn_obj}) {
        $connMgr_obj = $self->{conn_obj};
    } else {
        $connMgr_obj = $self->get_conn_obj();
    }
    my $conn = $connMgr_obj->get_connection(-filer         => $self->get("filer"),
                                            -cmd_interface => $self->get("cmd_interface"));
    unless (ref($conn) =~ /Connectrec/i) {
        $self->_set_err("failed to get connection");
        return $SanUtil::FALSE;
    }

    my $cmd = "systemshell -node $node";
    my $prompt = "login:";

    # run the system shell command and look for the login prompt
    my (undef, undef, $buffer, $rprompt) = Tharn::say($conn, $cmd,
                                                      '-timeout' => 300,
                                                      '-noprompt',
                                                      '-prompt' => $prompt,
                                                      '-ignore'=> "$cmd",
                                                      '-comment'=> "$cmd");

    # log in using the specified user name
    if ($rprompt =~ $prompt) {
        $cmd = $username;
        $prompt = "Password:";
        (undef, undef, $buffer, $rprompt) = Tharn::say($conn, $cmd,
                                                       '-timeout' => 300,
                                                       '-noprompt',
                                                       '-prompt' => $prompt,
                                                       '-ignore'=> "$cmd",
                                                       '-comment'=> "$cmd");


        # Use the specified password
        if ($rprompt =~ $prompt) {
            $cmd = $password;
            (undef, undef, $buffer, $rprompt) = Tharn::say($conn, $cmd,
                                                           '-timeout' => 300);

        } else {
            $self->_set_err("Failed to switch to bsd");
            return $SanUtil::FALSE;
        }
    } else {
        $self->_set_err("Failed to switch to bsd");
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method Name: _update
# Objective: update object metadata
# Details:   none
###############################################################################
sub _update {
    my ($self, %args) = @_;

    my ($value);

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # update each key => value
    foreach my $key (sort keys %{args}) {

        # ignore fields key. we don't need to update "fields"
        next if ($key =~ /fields/);

        # add quote if value contains space
        $self->{lc($key)} =
          ($args{$key} =~ /\s/ and $args{$key} != /\"/)
          ? "\"" . $args{$key} . "\""
          : $args{$key};
    }

    return $SanUtil::TRUE;
} ## end sub _update

###############################################################################
# Method Name: _get_build_info
# Objective: get controller build info and cache it to reuse
# Details:   none
###############################################################################
sub _get_build_info {
    my ($self) = @_;

    my (@records, $cmd);

    $self->_clear_err();

    # system image command
    # We may need to do some work to get the node in rare cases.
    my $node;
    my $obj = $self;

    # If we don't have the node yet...
    while (!length($node)) {

        # Call get_node
        $node = $obj->get_node();

        # If we found the node, exit the while loop.
        if (length($node)) {
            last;
        }

        # If get_parent_obj does not exists, exit gracefully.
        if (!$obj->can("get_parent_obj")) {
            $self->_set_err("get_parent_obj() not found.  Unable to identify current node.");
            return $SanUtil::FALSE;
        }

        # Else try and get the parent obj.
        $obj = $obj->get_parent_obj();

        # If we could not get the parent obj, return error.
        if (!$obj) {
            $self->_set_err("No parent obj returned? Unable to identify current node.");
            return $SanUtil::FALSE;
        }

    }

    $cmd = "system image show -iscurrent true -node $node";

    # execute command and cache returned info to reuse
    # command output:
    #
    # line[0]: system image show -iscurrent true -node ssan-6070-1
    # line[1]: node|image|isdefault|iscurrent|kernel-path|rootfs-path|version|
    #          installdate|client|change|deltas|build-time|build-type|
    #          build-info|skip-capability-check|
    # line[2]: Node|Image Name|Is Default Image|Is Current Image|
    #          Kernel Image Path|Root Filesystem Image Path|Software Version|
    #          Install Date|Build Client|Build Change Number|
    #          Build Deltas in Client|Build Time|Build Type|Build Information|
    #          Skip Capability Check|
    # line[3]: ssan-6070-1|image1|true|true|
    #          /cfcard/x86_64/freebsd/image1/kernel|
    #          /cfcard/x86_64/freebsd/image1/rootfs.img|
    #          MainNG__10.0.99|"9/5/2008 14:47:22"|korry:sjqa|828006|unknown|
    #          080905091609EDT|REGULAR|"BRANCH: smokejumper_qa "|-|
    unless (defined($self->{buildinfo})) {
        (@records) = $self->execute(-cmd => $cmd);
        if ($self->error()) {
            $self->_set_err("fail to execute \"$cmd\"", $self->errmsg());
            return $SanUtil::FALSE;
        }

        # cache build info to reuse
        $self->{buildinfo} = \@records;
    }

    # populate the build path if we havnt done so yet
    if (not defined($self->{build_path})) {
        if (defined($self->{buildinfo}[0]->{build_path})) {
            $self->{build_path} = $self->{buildinfo}[0]->{build_path};
        } else {
            $self->{build_path} = $self->_get_path_from_bsd();
            if ($self->error()) {
                $self->_set_err("Failed to get build path", $self->errmsg());
                return $SanUtil::FALSE;
            }
            # make sure we are back in ngsh context
            $self->set_shell_environment(-type => "ngsh");
        }
    }

    # if the buildinfo information is not fully populated it means we
    # are not in priv test. The only thing we can do to populate this
    # data with any sort of meaningful information is to parse the build path
    # and pretty much make the rest up our selves
    if (not defined($self->{buildinfo}[0]->{client})) {
        # the client should map to the basename of the build_path
        $self->{buildinfo}[0]->{client} = basename($self->{build_path});
    }
    if (not defined($self->{buildinfo}[0]->{change})) {
        # there is no way to determine the build number if we dont know
        # the change number, use the change number found in the build path
        $self->{buildinfo}[0]->{change} = $self->_get_change_number($self->{build_path});
    }
    if (not defined($self->{buildinfo}[0]->{build_info})) {
        # if we dont have build info, we dont know what branch is being used,
        # lets make up the build info string ourselves. You got any better ideas? Didnt think so.
        $self->{buildinfo}[0]->{build_info} = "BRANCH: DOT/" . basename($self->{build_path}) .
                                              " maxBBChg: " . $self->{buildinfo}[0]->{change} .
                                              " maxChange: " . $self->{buildinfo}[0]->{change} .
                                              " " . $self->{build_path};
    }
    if (not defined($self->{buildinfo}[0]->{build_path})) {
        # if the build path isnt defined, then lets just use the one we have in build path,
        # hopefully it is set
        $self->{buildinfo}[0]->{build_path} = $self->{build_path};
    }

    # cache build information
    # build   = change number
    # branch  = build-info
    # release = version
    $self->{build}      = $self->{buildinfo}[0]->{'change'};
    ($self->{branch})   = $self->{buildinfo}[0]->{'build_info'} =~ /:\s(.*)/;
    ($self->{version})  = $self->{buildinfo}[0]->{'version'} =~ /_(\d+\.\d+\.\d+\.?\d*)/;
    $self->{release}    = $self->{buildinfo}[0]->{'version'};
    $self->{client}     = $self->{buildinfo}[0]->{'client'};

    # strip path at the end and
    # remove DOT/ from DOT/<branch>
    $self->{branch} =~ s/\s+.*//;
    $self->{branch} =~ s/^DOT\///;

    return $SanUtil::TRUE;
} ## end sub _get_build_info

###############################################################################
# Method Name: _get_cdef_path
# Objective: get cdef path based on the installed build
#            CDEF_PATH environment variable can be set to override
# Details:   none
###############################################################################
sub _get_cdef_path {
    my ($self) = @_;

    my ($version_str, $build_client, $build_path, $cmd, $userid);

    $self->_clear_err();

    # if ENV{CDEF_PATH} variable was set, it will be used to override
    # default method
    my $cdef_path        = &Tharn::param('CDEF_PATH', '-default', undef);
    my $use_common_cdefs = &Tharn::param('CDEF_PATH_USE_COMMON', '-default', 0);
    if ($cdef_path) {
        $self->{cdef_path} = $cdef_path;
        $self->sanlog(-msg => "Using cdef path from parameter CDEF_PATH: $cdef_path");
    } elsif ($use_common_cdefs == 1) {
        $self->_get_common_cdef_path();
        $self->sanlog(-msg => "Using common cdef path because parameter "
                            . "CDEF_PATH_USE_COMMON is set: " . $self->{cdef_path});
    } else {
        # only query for build path information if its not already defined
        if (not defined($self->{build_path})) {
            # execute command to obtain controller build information
            if ($self->get_shell_environment() =~ /bsd/i) {
                $build_path = $self->_get_path_from_bsd();
                if ($self->error() || $build_path eq "") {
                    $self->_set_err("failed to _get_path_from_bsd", $self->errmsg());
                    return "";
                }
            } else {
                $build_path = $self->get("build_path");
                # if build_path is not in this build yet, then workaround to get build_path
                unless (length($build_path)) {
                    $build_path = $self->_get_path_from_bsd();
                    if ($self->error() || $build_path eq "") {
                        $self->_set_err("failed to _get_path_from_bsd", $self->errmsg());
                        return "";
                    }
                }
            }
            # make sure we are back in ngsh context
            $self->set_shell_environment(-type => "ngsh");
        } else {
            # otherwise get it from cache
            $build_path = $self->{build_path};
        }
        $self->sanlog(-msg => "BUILDROOT path: " . $build_path,
                      -msg_level => 2);

        # get build client
        $build_client = $self->get("client");
        if ($self->error() || not length($build_client)) {
            # if we can't figure out the build client,
            # assume it is the basename of the build path
            # this conditional used to error out
            $self->_clear_err();
            $build_client = basename($build_path);
        }

        $self->{cdef_path} = $build_path . "cdefs";

        # store the build path since we may need it later
        $self->{build_path} = $build_path;
    }

    # if cdef_path exist, use it. perhaps burt 342734 is fixed
    # if not exist, set path to BUILDROOT and determine ways to:
    #   1) autoload cdef files from installed build if AUTOLOAD_CDEFS set
    #   2) use cdef_path from SSAN_CDEF_PATH/$build_client if exists
    #   3) copy cdef files to SSAN_CDEF_PATH/$build_client if not exists
    unless (-d $self->{cdef_path}) {

        # if BUILDROOT does not exist any more because it has been moved
        # use /x/eng/rlse/DOT
        unless (-d $build_path) {
            $build_path = $self->RLSE_DOT_PATH . $build_client;
            $self->{cdef_path} = $build_path;
        }

        # if environment AUTOLOAD_CDEFS set, Schema::load() will load cdef files
        # from known locations. Otherwise, we will use SSAN_CDEF_PATH/$build_client
        # if exists. if it doesn't, then we will copy cdef files to
        # SSAN_CDEF_PATH/$build_client
        unless (defined $ENV{AUTOLOAD_CDEFS}) {

            # let's try to use cdefs from SSAN_CDEF_PATH/$build_client
            $self->{cdef_path} = $self->SSAN_CDEF_PATH . $build_client;

            # get change from cdef and buildroot path for comparison
            # it will either return valid change number or empty
            my $build_changenum = $self->_get_change_number($build_path);
            my $cdef_changenum = $self->_get_change_number($self->{cdef_path});

            # if build_path exists and change number is same, reuse it
            # otherwise re/copy it
            if (-d $self->{cdef_path} and $build_changenum eq $cdef_changenum) {
                $self->sanlog(-msg => "CDEF path exists and buildpath and cdefpath "
                              . "changenum are the same \($cdef_changenum\). So reuse cdefs from: "
                              . $self->{cdef_path});
            } else {
                #  copy cdef files to SSAN_CDEF_PATH/$build_client
                $self->_copy_cdef_files($build_path);
                if ($self->error()) {
                    $self->_set_err("failed to copy cdef files to " . $self->{cdef_path},
                                    $self->errmsg());
                    return "";
                }
            }
        }
    }

    $self->sanlog(-msg => "Using cdef path: " . $self->{cdef_path});
    return $self->{cdef_path};
} ## end sub _get_cdef_path

###############################################################################
# Method Name: _get_common_cdef_path
# Objective: get cdef path by finding lib/Node/NG/cdefs/main
# Details:   This will set $self->{cdef_path}
###############################################################################
sub _get_common_cdef_path {
    my ($self) = @_;

    $self->_clear_err();

    my $full_path = Tharn::param("TESTNAME");
    my @paths     = split(';', Tharn::param("NATE_LIB"));
    my $cdef_path = undef;
    my $eval_msg = "";

    # Put an eval around the path checks, one run had a runtime error on
    # the regex check.  Since we are checking whatever is found in NATE_LIB
    # we can't control the regex.
    eval {
        foreach my $path (@paths) {
            # If we find the path that's part of the full path, we can
            # use it.
            if($full_path =~ /$path/) {
                $self->sanlog(-msg => "The path $path matches $full_path.  "
                                    . "Verifying that the path " . $path
                                    . "/lib/Node/NG/cdefs/main exists",
                              -msg_level => 2);
                my $temp_path = $path . "/lib/Node/NG/cdefs/main";

                # Check to see if the path we have generated exists as a
                # directory, if it is, use it.
                if (-d $temp_path) {
                    $cdef_path = $temp_path;
                    last;
                }
            } # end if regex match
        } # end foreach
    };
    if($@) {
        $eval_msg = $@;
        $self->_set_err("Error locating cdefs in these paths from NATE_LIB: "
                      . join("\n", @paths));
        return;
    }

    # If the path was found and there was no error, the value will be defined
    # in $cdef_path
    if (defined($cdef_path)) {
        $self->{cdef_path} = $cdef_path;
        $self->sanlog(-msg => "Using common cdef path $cdef_path",
                      -msg_level => 2);
    } else {
        $self->_set_err("Unable to locate cdefs in these paths from NATE_LIB: "
                      . join("\n", @paths));
    }
    return;
} ## end sub _get_common_cdef_path

###############################################################################
# Method Name: _get_path_from_bsd
# Objective: get build path from bsd environment
# Details:   none
###############################################################################
sub _get_path_from_bsd {
    my ($self) = @_;
    my ($path, $version_str);

    $self->_clear_err();

    unless ($self->get_shell_environment() =~ /bsd/i) {
        $self->set_shell_environment(-type => "bsd");
    }

    my $cmd = 'ident /boot/modules/maytag.ko | grep NtapBuildPath';
    ($version_str) = $self->execute(-cmd => $cmd, -raw_data => 1);
    if ($self->error()) {
        $self->_set_err("failed to execute: " . $cmd, $self->errmsg());
        return "";
    }
        # "NetApp Release /x/eng/rtpbld01/scratch/korry/sj: \
        # Tue Dec  2 03:29:31 EST 2008  <1d>"
        # "Data Ontap Release /x/eng/rtpbld01/scratch/korry/sj: \
        # Tue Dec  2 03:29:31 EST 2008  <1d>"

        # another (better way)
        # modvers -v | grep BuildPath | grep maytag
        # maytag.ko, Build , BuildTime , BuildPath \
        #/x/eng/rlse15/ontapng/nightly/mainngN_090430_0100/, BuildOptions x86_64

    if ($version_str =~ /NtapBuildPath:\s+(\/.*)\s+\$/) {
        $path = $1;
    } elsif ($version_str =~ /NtapBuildPath:\s+(\S+)/) {
        # Any other build that doesnt start with /
        # build the release path manually
        my $promo_build = $1;
        $path = "/x/eng/rlse/DOT/$promo_build";
    } else {
        $self->_set_err("output does not contain buildpath: " . $version_str);
        return "";
    }
} ## end sub _get_path_from_bsd

###############################################################################
# Method Name: _copy_cdef_files
# Objective: copy cdef files from BUILDROOT to SSAN_CDEF_PATH/$build_client
# Details:   none
###############################################################################
sub _copy_cdef_files {
    my ($self, $path) = @_;

    $self->_clear_err();

    #  create directory SSAN_PATH/$build_client if it does not exist
    unless (-d $self->{cdef_path}) {
        eval { File::Path::make_path($self->{cdef_path}); };
        if ($@) {
            $self->_set_err("failed to make dir: ", $@);
            return $SanUtil::FALSE;
        }
    }

    # copy alias cdef files that smfGen did not create
    foreach my $alias_file (@Alias_cdefs) {
        my $alias_src = $self->SSAN_CDEF_PATH . $alias_file;
        my $alias_dst = $self->{cdef_path} . '/' . $alias_file;
        eval { File::Copy::copy($alias_src, $alias_dst); };
        if ($@) {
            $self->_set_err("failed to copy: " . $alias_src, $@);
            $self->_remove_cdef_dir();
            return $SanUtil::FALSE;
        }
    }

    # let's copy cdefs from the installed build path to SSAN_PATH/$build_client
    foreach my $location (@Known_cdef_locations) {
        my $src_path = $path . $location;
        # check for valid src cdef path
        if (-d $src_path) {

            # retrieve cdef files
            opendir(DIR, $src_path) or do {
                $self->_set_err("can not open dir: " . $src_path, $!);
                $self->_remove_cdef_dir();
                return $SanUtil::FALSE;
            };
            foreach my $file (readdir(DIR)) {
                # skip files that does not have cdefs extension
                next if ($file !~ /\.cdefs$/);
                my $src_file = $src_path . $file;
                my $dst_file = $self->{cdef_path} . '/' . $file;
                eval { File::Copy::copy($src_file, $dst_file); };
                if ($@) {
                    closedir(DIR);
                    $self->_set_err("failed to copy: " . $src_file, $@);
                    $self->_remove_cdef_dir();
                    return $SanUtil::FALSE;
                }
            }
            closedir(DIR);
        } else {
            $self->sanlog("cdef path: $src_path does not exist");
            # right now, it's possible some pathes do not exist
            # we just log a msg. Once cdefs is consolidated in a common
            # location, we will set_err and remove directory
            # $self->_set_err("invalid cdef path: " . $src_path);
            # $self->_remove_cdef_dir();
            # return $SanUtil::FALSE;
            #
        }
    }

    # copy the new p4sync file with current changenum
    my $changenum_src = $path . '/' . $Changenum_file;
    my $changenum_dst = $self->{cdef_path} . '/' . $Changenum_file;
    eval { File::Copy::copy($changenum_src, $changenum_dst); };
    if ($@) {
        $self->_set_err("failed to copy: " . $changenum_src, $@);
        $self->_remove_cdef_dir();
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub _copy_cdef_files

###############################################################################
# Method Name: _remove_cdef_dir
# Objective: remove cdef directory when copy cdef files encountered error
# Details:   none
###############################################################################
sub _remove_cdef_dir {
    my ($self) = @_;

    $self->_clear_err();

    #  remove directory SSAN_PATH/$build_client and its content
    eval { File::Path::remove_tree($self->{cdef_path}); };
    if ($@) {
        $self->_set_err("failed to remove dir: " . $self->{cdef_path}, $@);
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: _get_change_number
# Objective: Get change number from specific path
# Details:   none
###############################################################################
sub _get_change_number {
    my ($self, $path) = @_;

    $self->_clear_err();

    my $changenum_file = $path . '/' . $Changenum_file;
    if (-f $changenum_file) {
        # read file
        open FILE, "<", $changenum_file;
        my $changenum = <FILE>;
        chomp($changenum);
        close(FILE);
        return $changenum;
    } else {
        return "";
    }
}


###############################################################################
# Method Name: _zapi_decode_non_show_cmd
# Objective: Transforms the returned ZAPI data to CLI like, from commands that
#            does not contain a '_show' nor have a tag 'attributes' in the
#            returning xml.
# Details:   none
###############################################################################
sub _zapi_decode_non_show_cmd {
    my ($self, $output_ref, $attrs_ref) = @_;
    # LOCALS
    my @recs = ();

    # Build output hash comparable to cli output using the attributes in the
    # attributes table and the zapi output.
    #
    # IMPORTANT: need to define 'current_attr' as "non_show". This will enable the
    # method to parse through the returned zapi output. Need to have the same attr
    # defined in the AttrTable.pm to work.
    @recs = $self->_zapi_to_cli_format(attr => $attrs_ref,
                                       current_attr => "non_show",
                                       output => $output_ref);
    if ($self->error()) {
        $self->_set_err("failed to convert zapi to cli format", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @recs : [@recs];
} ## end _zapi_decode_non_show_cmd

###############################################################################
# Method Name: _zapi_decode_show_cmd
# Objective: Transforms the returned ZAPI data, from a show command, into an array
# of hashes that are comparable to CLI output
# Details:   none
###############################################################################
sub _zapi_decode_show_cmd {
    my ($self, $output_ref, $attrs_ref) = @_;
    # Locals
    my @recs = ();
    my $current_attr ='';
    my $current_list = {};

    # get attributes or attributes-list from zapi output hash
    foreach my $key (keys %{$output_ref}) {
        if ($key =~ /attributes/) {
            my $attr_list = $output_ref->{$key};

            # get the next nested tag that contains output data/info
            # there is only one nested key at this level
            foreach my $nested_key (keys %{$attr_list}) {
                $current_attr = $nested_key;
                $current_list = $attr_list->{$nested_key};
                last;
            }
        }
    }

    # if current list is array, we have multiple records
    my $rec = {};
    if (ref($current_list) =~ /ARRAY/i) {
        foreach my $list (@{$current_list}) {
            # build output hash comparable to cli output
            $rec = $self->_zapi_to_cli_format(attr => $attrs_ref,
                                              current_attr => $current_attr,
                                              output => $list);
            if ($self->error()) {
                $self->_set_err("failed to convert zapi to cli format", $self->errmsg());
                return (wantarray()) ? () : [];
            }
            push @recs, $rec;
        }
    } else {
        # build output hash comparable to cli output
        $rec = $self->_zapi_to_cli_format(attr => $attrs_ref,
                                          current_attr => $current_attr,
                                          output => $current_list);
        if ($self->error()) {
            $self->_set_err("failed to convert zapi to cli format", $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # Just like the current_list =~ ARRAY, the actual values are returned in
        # an array inside of an attribute that is part of the current list. The logic
        # in place will only detect arrays that are under the 'attributes' key.
        #
        # All known attributes that hold its records in a nested array and are not under
        # the 'attributes-list' key, should be added to the SanConsts::zapi_nested_attributes hash.
        #
        # The following logic looks for known attributes that has a nested array in the
        # current list, if found one,  gets the attribute (key) and value (array)
        # and transforms it to a cli-format.
        #
        # Example:
        #  $VAR1 = {'status' => 'passed',
        #           'attributes-list' => {'fcp-adapter-initiators-info' => {'adapter' => 'utest_lif_node1',
        #                                                                   'vserver' => 'utest_vserver',
        #                                                                   'fcp-connected-initiators' => {'fcp-connected-initiator-info' => [
        #                                                                                                                                     {'port-name' => '...,
        #                                                                                                                                      'port-address' => '...,
        #                                                                                                                                     },
        #
        #                                                                                                                                     {'port-name' => '...,
        #                                                                                                                                      'port-address' => '...,
        #                                                                                                                                     },
        #
        #                                                                                                                                     {'port-name' => '...,
        #                                                                                                                                      'port-address' => '...,
        #                                                                                                                                     }
        #                                                                                                                                    ]

        my @new_recs = ();
        my %zapi_known_nested_attr = $SanConsts::zapi_nested_attributes;

        foreach my $nested_attr (keys %zapi_known_nested_attr) {
            if ( exists($current_list->{$nested_attr}) ) {
                # With the Key, grab the value (the attribute) that actually holds
                # the array
                $current_attr = $zapi_known_nested_attr{$nested_attr};
                my $array_list_ref = $current_list->{$nested_attr}->{$current_attr};
                my $nested_rec= '';

                # Some attributes could've been found before, so, delete those to
                # avoid overwriting information.
                foreach my $attr_key (keys %{$attrs_ref}) {
                    if ($rec->{$attr_key} ne '-') {
                        delete $attrs_ref->{$attr_key};
                    }
                }

                # Transform to CLI-format, each record on the array
                foreach my $list (@{$array_list_ref}) {
                    $nested_rec = $self->_zapi_to_cli_format(attr => $attrs_ref,
                                                             current_attr => $current_attr,
                                                             output => $list);
                    if ($self->error()) {
                        $self->_set_err("failed to convert zapi to cli format", $self->errmsg());
                        return (wantarray()) ? () : [];
                    }
                    # Merge the common attributes with the returned list of attributes.
                    my %combined_rec = ( %{$rec}, %{$nested_rec} );
                    # Insert into the array the newly created list of all the attributes
                    push (@new_recs, \%combined_rec);
                }
            }
        }

        # if nested attributes were found, the array will not be empty, so, return that array.
        if ( scalar(@new_recs) ) {
            push (@recs, @new_recs);
        } else { #no nested attributes were found, so return the original rec.
            push @recs, $rec;
        }
    }

    return (wantarray()) ? @recs : [@recs];
} ## end _zapi_decode_show_cmd

###############################################################################
# Method Name: _zapi_to_cli_format
# Objective: convert api output to cli hash format
# Details:   none
###############################################################################
sub _zapi_to_cli_format {
    my ($self, %args) = @_;
    my $rec = {};

    $self->_clear_err();

    # strip out the '-' dash of key arguments
    %args = &SanUtil::strip_dash(%args);
    my @input_params = (%args);

    # validation_options is set to allow extra params not defined in %args to bypass
    Params::Validate::validation_options( allow_extra => 1);

    # validate_params is the arguments to validate
    my %validate_params = (
                      attr         => { type     => HASHREF,
                                        required => 1 },
                      current_attr => { type     => SCALAR,
                                        required => 1 },
                      output       => { type     => HASHREF,
                                        required => 1 }
                      );
    eval {
        (%args) = Params::Validate::validate(@input_params, \%validate_params);
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: ". $@);
        return {};
    }

    my $attr_list = $args{attr};
    foreach my $attr (keys %{$attr_list}) {
       # if zapi_attr is not exist, fill 'zapi_na'
       if ($attr_list->{$attr}->{zapi_field} eq "not_exist") {
           $rec->{$attr} = "zapi_na";
       } else {
           # get matched value from attribute list
           $rec->{$attr} = $self->_get_value_from_attr_list(
                                                elem => $args{output},
                                                expected_attr => $attr_list->{$attr}->{attr},
                                                current_attr => $args{current_attr},
                                                field => $attr_list->{$attr}->{zapi_field});

           # if we traverse through and still unable to find the specified attrbute
           # we have to fill the dash "-" similar to CLI output
           if (! length($rec->{$attr})) {
               $rec->{$attr} =  "-";
           }

           # translate value if value key exists
           # example of 'state' attribute of lun_show in AttrTable.pm
           # 'state' => {'zapi_field' => 'online', attr => 'lun-info', 'true' => 'online', 'false' => 'offline'},
           # the example above will convert the value 'true' to 'online' to be consistent
           # with CLI
           if (exists $attr_list->{$attr}->{$rec->{$attr}}) {
               $rec->{$attr} = $attr_list->{$attr}->{$rec->{$attr}};
           }
       }
    }
    return $rec;
}## end _zapi_to_cli_format
###############################################################################
# Method Name: _get_value_from_attr_list
# Objective: get value with matched field from attribute list
# Details:   none
###############################################################################
sub _get_value_from_attr_list {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    my $value = "";
    my $elem  = $args{elem};
    my $expected_attr  = $args{expected_attr};
    my $current_attr  = $args{current_attr};
    my $field = $args{field};

    # traverse attribute list
    foreach (keys %{$elem}) {
        if (ref($elem->{$_}) =~ /HASH/) {
            # traverse down the tree
            $value = $self->_get_value_from_attr_list(elem => $elem->{$_},
                                                      expected_attr => $expected_attr,
                                                      current_attr => $_,
                                                      field => $field);
        }
        elsif ($_ =~ /^($field)$/) {

            # if the field is ARRAY, we need to get all attrs
            # and join them similar to CLI format
            # ei.
            # <portset-port-info>
            #     <portset-port-name>igrp_iscsi_lif_1</portset-port-name>
            #     <portset-port-name>igrp_fcp_lif_1</portset-port-name>
            #  </portset-port-info>
            # convert to:
            # port_name = 'igrp_iscsi_lif_1,igrp_fcp_lif_1'
            if ($elem->{$_} =~ /ARRAY/) {
                # join array into value strings
                $value = join (",", @{$elem->{$_}});
            } else {

                # looking for matching field and return its value
                $value = $elem->{$_};
            }
            # we found the value, is it from the right attribute?
            # if not, reset the value and continue on
            unless ($current_attr =~ /info|^$expected_attr$/) {
                $value = "";
            }
        }
        # if the key (current_attr) that holds the field is an ARRAY, we need
        # to get all attrs values of the field and join them, to return a similar
        # response to CLI.
        #
        # <initiators>
        #    <initiator-info>
        #        <initiator-name>21:00:00:1b:32:0f:cd:8a</initiator-name>
        #    </initiator-info>
        #    <initiator-info>
        #        <initiator-name>21:01:00:1b:32:2f:cd:8a</initiator-name>
        #    </initiator-info>
        #</initiators>
        # convert to:
        # initiator = '21:00:00:1b:32:0f:cd:8a,21:00:00:1b:32:2f:cd:8a'
        #
        elsif ( ($_ =~ /^($expected_attr)$/) and (ref($elem->{$_}) =~ /ARRAY/) ) {
            # Get the value from each element
            foreach my $field_hash ( @{($elem->{$_})} ) {
                # Concatenate all each found value, adding the commas
                $value .= sprintf($field_hash->{$field}).',';
            }
            # Remove the last comma "," from $value if it's not empty.
            ($value = substr($value, 0, -1)) if (length($value)) ;
        }

        # return value if exists
        return $value if (length($value));
    }

    # field is not found in zapi output hash
    $self->sanlog(-msg => "Attribute '".$current_attr."' does not contain field: " . $field,
                  -msg_level => 9);
    return "";
}##end _get_value_from_attr_list

###############################################################################
# Method Name: _connect
# Objective: make connection to 10-mode controller and set its environment
#            to be run either as 'root' or 'admin'
# Details:   none
###############################################################################
sub _connect {
    my ($self, $filer, $cmd_interface, $user, $passwd, $conn_type) = @_;

    my (%args);

    $self->_clear_err();

    # set arguments
    $args{cmd_interface} = $cmd_interface;
    $args{is_cserver} = $self->{is_cserver};
    $args{filer}     = $filer;
    $args{username}  = $user unless ($user eq "");
    $args{password}  = $passwd unless ($passwd eq "");
    $args{conntype}  = "ssh" if ($conn_type eq "");

    # verify connection object has been instantiated
    unless (defined $self->{conn_obj}) {
        $self->_set_err("connection object does not exist");
        return $SanUtil::FALSE;
    }

    # if is_cserver is set, we know cluster_mgmt_ip and node list are set
    # so use them to make connection
    if ($args{is_cserver}) {
        $args{cluster_mgmt_ip} = $self->{cluster_mgmt_ip};
        $args{nodes_in_cluster} = $self->{nodes_in_cluster};
    }

    # make the connection
    # let always make connection as "root" user
    $self->{conn_obj}->make_connect(%args);
    if ($self->{conn_obj}->error()) {
        $self->_set_err("fail to create connection", $self->{conn_obj}->errmsg());
        return $SanUtil::FALSE;
    }

    # get shell environment
    my $shell_env = $self->get_shell_environment();
    if ($self->error()) {
        $self->_set_err("failed to get shell environment", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # if environment is bsd, unset autologout
    # if environment is ngsh, set output format
    if ($shell_env =~ /bsd/i) {
        $self->execute(-cmd => "unset autologout");
    } elsif ($shell_env =~ /ngsh/i) {
        # we're already in ngsh shell because we used 'admin'
        # set the output format for ngsh environment
        $self->_set_output_format();
        if ($self->error()) {
            $self->_set_err("failed to set output format", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

if (0) {
    # let's get cdef path
    $self->_get_cdef_path();
    if ($self->error()) {
        $self->_set_err("failed to get command definition path", $self->errmsg());
        return $SanUtil::FALSE;
    }
}

    # set command to be run as "root", "diag", or "admin"
    # if undefined, use default setting from tharnhost
    # if it was rootor diag user, we're in bsd and need to get to ngsh
    if ($user eq "") {

        my $conn = $self->{conn_obj}->get_connection(-filer => $args{filer},
                                                     -cmd_interface => $args{cmd_interface});
        $user = $conn->{username};

        # enter ngsh if username is root
           unless ($user =~ /admin/i) {
            if ($user =~ /(root|diag)/i) {
                # we're in 'bsd' so set shell environment to 'ngsh'
                $self->set_shell_environment(-type => "ngsh", -user => $user);
                if ($self->error()) {
                    $self->_set_err("failed to set ngsh environment", $self->errmsg());
                    return $SanUtil::FALSE;
                }
                } else {
                    $self->_set_err("Please specify valid username (root/diag) for conntype $args{conntype}");
                    return $SanUtil::FALSE;
                }
        }
    } elsif ($user =~ /(root|diag)/i) {
        # we're in 'bsd' so set shell environment to 'ngsh'
        $self->set_shell_environment(-type => "ngsh", -user => $user);
        if ($self->error()) {
            $self->_set_err("failed to set ngsh environment", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: _set_output_format
# Objective: set the output format in one row with ":" seperator per record
# Details:   none
###############################################################################
sub _set_output_format {
    my ($self) = @_;

    my ($confirm, $allfields, $sys_timeout, $row, $cmd);

    # get the connection object
    my $conn;
    if (exists $self->{conn_obj}) {
        $conn = $self->{conn_obj}->get_connection(-filer => $self->get("filer"),
                                                  -cmd_interface => $self->get("cmd_interface"));
    } elsif (exists $self->{connection}) {
        $conn =  $self->{connection};
    } else {
        $self->_set_err("connection or connection manager does not exist");
        return $SanUtil::FAIL;
    }

    # if we are connected under the diag user, then default to priv test
    # we may want to make this configurable via an input into the constructor
    my $priv;
    if ($conn->{username} eq "diag") {
        $priv = "test";
    } else {
        # everything else, default to priv diag
        $priv = "diag";
    }

    # default commands for environment to be run on
    $confirm     = "set -confirm off;";
    $allfields   = "set -showallfields true -showseparator \"|\" -privilege $priv;";
    $sys_timeout = "system timeout modify -timeout 0;";
    $row         = "row 0";
    $cmd         = $confirm . $allfields . $sys_timeout . $row;

    # execute cmd to set env for system under test
    my @buffer = $self->execute(-cmd      => $cmd,
                                -raw_data => 1);
    if ($self->error()) {
        $self->_set_err("failed to execute: " . $cmd .
                        " buffer: " . $buffer[0] , $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

CommonInterface.pm

=head1 Author

Korry Nguyen (dl-ssan-automation)

=head1 Description

Provides the core set of methods as a base class for all classes (Filer10M, Aggregate, Volume, Lun, etc.) which were being used in the Cmode SAN libraries

=head1 See Also

CommonInterface package is used as a base class for other modules. See these modules for further interface detail.

B<San::ControllerCluster.pm, San::ControllerCluster::Filer10M.pm>

=head1 B<Common Methods>

=head2 B<create_accessor_methods(%args)>

Create accessor methods for attributes and objects dynamically

=over 4

=item Inputs

=item * I<-type> (required)

Type of accessor methods to be created (get_obj | get_attr)

=item * I<-list> (required)

Arrayref of elements to be created

=item * I<-pkg> (required only for get_obj type)

The package where accessor methods are created for

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Create accessor methods (get_build, get_release, get_name)
 my @fields = qw(build release name);
 $self->create_accessor_methods(-type => "get_attr",
                                -list => \@fields);
 if ($self->error()) {
     ...
 }

 # ex. 2: Create accessor methods (get_aggregate_objs, get_disk_objs)
 my @names = qw(aggregate disk);
 $self->create_accessor_methods(-type => "get_obj",
                                -pkg  => __PACKAGE__,
                                -list => \@names);
 if ($self->error()) {
     ...
 }

=back

=head2 B<get($attribute)>

Get attribute of an object

=over 4

=item Inputs

=item * I<attribute> (required)

Attribute to retrieve from an object

=item Outputs

Returns value of the passed in attribute

=item Examples

 # ex. 1: Get filer system ID
 my $filer_obj = SanFiler->new(-filer => $filer_name);

 $filer_obj->get("systemid");
 if ($filer_obj->error()) {
     ...
 }

 # ex. 2: Get vserver rootvolume
 my $vs_obj = $filer_obj->get_vserver_objs(-name => "^my_vserver\$");
 $rootvol = $vs_obj->get("rootvolume");
 if ($vs_obj->error()) {
     ...
 }

=back

=head2 B<get_fields()>

Get a list of attributes that exist in an object

=over 4

=item Inputs

None

=item Outputs

Returns an array or arrayref list of attributes

=item Examples

 # ex. 1: Get all attributes of a filer object
 my @filer_attrs = $filer_obj->get_fields();
 if ($filer_obj->error()) {
     ...
 }

 # ex. 2: Get all attributes of a volume object
 my @vol_attrs = $vol_obj->get_fields();
 if ($vol_obj->error()) {
     ...
 }

=back

=head2 B<get_conn_obj()>

Get connection manager object

=over 4

=item Inputs

None

=item Outputs

Returns connection object

=item Examples

 # ex. 1: Get connection manager object
 my $conn = $self->get_conn_obj();
 if ($self->error()) {
     ...
 }

=back

=head2 B<get_schema_obj()>

Get CDEF schema object

=over 4

=item Inputs

None

=item Outputs

Returns schema object

=item Examples

 # ex. 1: Get the schema object
 my $schema_obj = $self->get_schema_obj();
 if ($self->error()) {
     ...
 }

=back

=head2 B<execute(%args)>

Execute the CLI command on a controller

=over 4

=item Inputs

=item * I<-cmd> (required)

The command to be executed

=item * I<-raw_data> (optional) Default is 0

Flag to return raw data if set.

=item * I<-timeout> (optional) Default is 300 seconds

Time out interval for the executed command

=item Outputs

Returns an array of hash (key => value) records or an array of raw outputs if raw_data flag is set

=item Examples

 # ex. 1: Get raw data for fcp nodename command
 my $cmd = "vserver fcp nodename -vserver vs0";
 my @output = $self->execute(-cmd => $cmd,
                             -raw_data => 1);
 if ($self->error()) {
     ...
 }
 logcomment("Output: " . $output[0]);

 # ex. 2: Get an array of hash for volume show command
 my $cmd = "volume show -vserver vs0";
 my @records = $self->execute(-cmd => $cmd);
 if ($self->error()) {
     ...
 }

=back

=head2 B<invoke(%args)>

Takes CDEF schema and translate into actual command prior to execute

=over 4

=item Inputs

=item * I<-cdef> (required)

Cdef schema that will be translated into CLI command

=item * I<-update_attr> (optional) Default is 1

Flag to return raw data if set.

=item * I<-raw_data> (optional) Default is 0

Flag to return raw data if set. If raw_data is set, update_attr flag must be disabled

=item * I<-timeout> (optional) Default is 300 seconds

Time out interval for the executed command

=item Outputs

Returns an array of hash (key => value) records or an array of raw outputs if raw_data flag is set

=item Examples

 # ex. 1: Get raw data for fcp nodename command
 my $cdef = "vserver_fcp_nodename";
 my @output = $self->invoke(-cdef => $cdef,
                            -vserver => "vs0",
                            -raw_data => 1,
                            -update_attr => 0);
 if ($self->error()) {
     ...
 }
 logcomment("Output: " . $output[0]);

 # ex. 2: Get an array of hash for volume show command
 my $cdef = "volume_show";
 my @records = $self->invoke(-cdef => $cdef,
                             -vserver => "vs0");
 if ($self->error()) {
     ...
 }

=back

=head2 B<run_node_scope(%args)>

Execute 7mode command in 7mode stack

=over 4

=item Inputs

=item * I<-cmd> (required)

The command to be executed

=item Outputs

Returns an array of hash (key => value) records

=item Examples

 # ex. 1: Execute 7mode command in 7mode stack
 my $cmd_cmd = "vol status";
 my @output = $self->run_node_scope(-cmd => $cmd_7m);
 if ($self->error()) {
     ...
 }

=back

=head2 B<refresh(%args)>

Refresh and update object based on cdef command/schema

=over 4

=item Inputs

=item * I<-cdef> (required)

Cdef command for show (ex. vserver_show, volume_show)

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Refresh volume object
 my $cdef = "volume_show";
 $self->refresh(-cdef => $cdef);
 if ($self->error()) {
     ...
 }

=back

=head2 B<get_version()>

Get version string from 7mode stack

=over 4

=item Inputs

None

=item Outputs

Returns version string

=item Examples

 # ex. 1: Get version string in 7mode stack
 my $ver_str = $filer_obj->get_version();
 if ($filer_obj->error()) {
     ...
 }

=back

=head2 B<get_shell_environment()>

Get current shell environment

=over 4

=item Inputs

None

=item Outputs

Returns either bsd, ngsh, or node_scope

=item Examples

 # ex. 1: Determine shell environment
 if ($filer_obj->get_shell_environment() =~ /ngsh/i) {
     # execute Cmode command
 } elsif ($filer_obj->get_shell_environment() =~ /node_scope/i) {
     # execute 7mode command
 } elsif ($filer_obj->get_shell_environment() =~ /bsd/i) {
     # execute command in FreeBSD shell
 } else {
     # set error
 }

=back

=head2 B<set_shell_environment(%args)>

Set shell environment

=over 4

=item Inputs

=item * I<-type> (required)

Type of shell to set (ex. bsd, ngsh, or node_scope)

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Set shell environment to node_scope (7mode)
 $self->set_shell_environment(-type => "node_scope");
 if ($self->error()) {
     ...
 }
 # execute 7mode command

=back

=head2 B<search(%args)>

Search for the matching objects

=over 4

=item Inputs

=item * I<-field> (required at least one)

Field to search. It could be a regex.

=item * I<-count> (optional)

Number of objects to return

=item Outputs

Returns an array of objects that met the criteria

=item Examples

 # ex. 1: Search for 8 spare disks from the entire list of disk objects
 @spare_disks = $self->search(-objs => \@disk_objs,
                              -state => "spare",
                              -count => 8);
 if ($self->error()) {
     ...
 }

 # ex. 2: Search for volumes with matching prefix name
 @vols = $self->search(-objs => \@vol_objs,
                       -name => "^myvol");
 if ($self->error()) {
     ...
 }

=back

=head2 B<cli_output_decode($cli_output)>

Parse returned CLI output data into an array of hashes

=over 4

=item Inputs

=item * <output> (required)

The output from a cli command

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing hashes of the parsed cli output. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # ex. 1:
 (@output) = $self->cli_output_decode($result_ref->{output});
 if ($self->error()) {
     $self->_set_err("cli output decode failed", $self->errmsg());
     return (wantarray()) ? () : [];
 }

=back

=head2 B<zapi_output_decode($cdef_cmd, $zapi_output)>

Parse returned ZAPI output data into an array of hashes

=over 4

=item Inputs

=item * <cdef_cmd> (required)

cdef command

=item * <output> (required)

The zapi output

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing hashes of the parsed zapi output. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # ex. 1:
 (@output) = $self->zapi_output_decode($cdef_cmd, $result_ref->{output});
 if ($self->error()) {
     $self->_set_err("zapi output decode failed", $self->errmsg());
     return (wantarray()) ? () : [];
 }

=back

=head2 B<create_common_path(%args)>

Create volume, qtree, or lun path based on the arguments that passed in

=over 4

=item Inputs

=item * <volume> (required)

Volume name

=item * <qtree> (optional)

Qtree name

=item * <lun> (optional)

Lun name

=item Outputs

Returns a string path that starts with "/vol/".

=item Examples

 # ex. 1: create volume path
 $vol_path = $self->create_common_path(volume => $volume_name);
 if ($self->error()) {
     $self->_set_err("failed to create volume path", $self->errmsg());
     return "";
 }

 # ex. 2: create qtree path
 $qtree_path = $self->create_common_path(volume => $volume_name, qtree => $qtree_name);
 if ($self->error()) {
     $self->_set_err("failed to create qtree path", $self->errmsg());
     return "";
 }

 # ex. 2: create lun path
 $lun_path = $self->create_common_path(volume => $volume_name,
                                       qtree  => $qtree_name,
                                       lun    => $lun_name);
 if ($self->error()) {
     $self->_set_err("failed to create lun path", $self->errmsg());
     return "";
 }

=back
