# Copyright (c) 2007-2010 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to manage a Connection object
##
## @author korry@netapp.com, dl-ssan-test
## @status public

package ConnectionMgr;

# compiler directives
use strict;

# imported modules
use Params::Validate qw( :all );
use Tharn;
use Error;
use Exporter;
use San::SanLog;
use San::SanUtil;
use Connectrec qw ();
use vars qw( @EXPORT @ISA );

@ISA = qw( Error SanLog SanUtil );

# global variables
my $Connection = undef;
my $Connect_id = 0;

###############################################################################
# Method:    new
# Objective: Create new Cluster Node object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    # set default to use existence connection
    my $using_existing_conn = $SanUtil::TRUE;

    my $self = {};

    # let's bless instance into class
    bless $self, $class;

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # process parameters and see if we want to reuse the existing
    # connection object.
    while (my ($key, $value) =  each(%args)) {
        if ($key =~ /-?reuse_conn$/) {
            if (! length($using_existing_conn = $value) || $value !~  /($SanUtil::FALSE|$SanUtil::TRUE)/) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
        }
    }

    # initialize connection retry count to 5
    $self->{connection_retries} = 5;

    # Connection object only instantiated once throughout an application
    # if using_existing_conn flag is true:
    #  1) return Connection if it's already exist
    #  2) if not, set Connection to self and return Connection
    if ($using_existing_conn == $SanUtil::TRUE) {

        # first time creating connection so set Connection to self
        unless (defined $Connection) {
            $Connection = $self;
        }

        return $Connection;
    }

    return $self;
} ## end sub new

###############################################################################
# Method:    make_connect
# Objective: Creates a connection
# Details:   See POD documentation at the end of this file.
###############################################################################
sub make_connect {
    my ($self, %args) = @_;

    my ($conn_str, $connect, $host, $errmsg, $retries);
    my ($username, $password, @nodes, $host_obj, %validate_params);
    my $prompt = "";

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);
    my @input_params = (%args);

    # validation_options is set to allow extra params not defined in %args to bypass
    Params::Validate::validation_options( allow_extra => 1);

    # validate_params is the arguments to validate
    %validate_params = (filer         => { type     => SCALAR,
                                           required => 1 },
                        is_cserver    => { type     => SCALAR,
                                           optional => 1,
                                           default  => $SanUtil::FALSE },
                        cmd_interface => { type     => SCALAR,
                                           optional => 1,
                                           default  => 'cli' },
                        conntype      => { type     => SCALAR,
                                           optional => 1 },
                        password      => { type     => SCALAR,
                                           optional => 1 },
                        username      => { type     => SCALAR,
                                           optional => 1 }
                       );
    eval {
        (%args) = Params::Validate::validate(@input_params, \%validate_params);
    };
    if ($@) {
        $self->_set_err("fail to validate parameters: " . $@);
        return $SanUtil::FALSE;
    }

    # create host obj for cserver or node
    if ($args{is_cserver}) {

        # create host object for cserver
        $host_obj = &SanUtil::create_host_obj(cserver => $args{filer}, %args);
    } else {

        # create host object for node
        $host_obj = &SanUtil::create_host_obj(filer => $args{filer});
    }

    # set connnection field based on hosttype
    # if hosttype is Cmode, use ssh-persist
    # if hosttype is 7mode, use rsh-socket
    my $conn_field = ($host_obj->hosttype() =~ /.*-ng/) ? "ssh" : "rsh";

    $self->{mode} = ($host_obj->hosttype() =~ /.*-ng/) ? "Cmode" : "7mode";

    # 1) use passed in username and password
    # 2) if not, use username and password from nate host setting
    # 3) if not, use default username=diag and password=netapp1!
    if (exists($args{username})) {
        $host_obj->{connection_fields}{$conn_field}{username} = $args{username};
    } else {
        unless (defined $host_obj->{connection_fields}{$conn_field}{username} and
                length($host_obj->{connection_fields}{$conn_field}{username})) {
            $host_obj->{connection_fields}{$conn_field}{username} = "diag";
        }
    }
    if (exists($args{password})) {
        $host_obj->{connection_fields}{$conn_field}{password} = $args{password};
    } else {
        unless (defined $host_obj->{connection_fields}{$conn_field}{password} and
                length($host_obj->{connection_fields}{$conn_field}{password})) {
            $host_obj->{connection_fields}{$conn_field}{password} = "netapp1!";
        }
    }

    # if conntype is passed in, it must be either ssh or rsh
    if (exists($args{conntype})) {
        # we will only support ssh-persist and rsh-socket
        if ($args{conntype} =~ /ssh/i) {
            $host_obj->{connection_fields}{$conn_field}{conntype} = "ssh-persist";
        } elsif ($args{conntype} =~ /rsh/i) {
            $host_obj->{connection_fields}{$conn_field}{conntype} = "rsh-socket";
        } else {
            $self->_set_err("invalid conntype: $args{conntype}. conntype must be \'ssh\' or \'rsh\'");
            return $SanUtil::FALSE;
        }
    } else {
        unless (defined $host_obj->{connection_fields}{$conn_field}{conntype} and
                length($host_obj->{connection_fields}{$conn_field}{conntype})) {
            $host_obj->{connection_fields}{$conn_field}{conntype} =
                ($host_obj->hosttype() =~ /.*-ng/) ? "ssh-persist" : "rsh-socket";
        }
    }

    # initialize connection string
    $conn_str =
        $host_obj->{id} . "#"
      . $host_obj->{connection_fields}{$conn_field}{conntype}
      . $Connect_id++;

    $self->sanlog(-msg => "Connection string is: " . $conn_str);

    my $max_retries = $self->{connection_retries};

    # set mode for connection object
    # retry connection five times
    while (! defined($connect) and $retries < $max_retries) {
        eval {
            $connect = $host_obj->connect(template => $conn_field);
        };
        if ($@) {
            $retries++;
            $connect = undef;

            $errmsg = join("", $@);
            $errmsg =~ s/\n//g;

            # Display an error in case user doesn't include error
            # handler on object instantiation.
            $self->sanlog(-msg => "\n\tWARNING: " . (caller(0))[3] .
                                  " unable to connect to controller '" .
                                  $args{filer} . "': $errmsg\n" .
                                  "Attempt $retries of $max_retries");
            &Tharn::disconnect($host_obj, '-ok');
            &SanUtil::nate_sleep(30);
        }
    }

    # if retry count has been exhausted, set error and return
    if ($retries == $max_retries) {
        $errmsg = 'Perhaps $prompt_norm param is missing from nate_host ' .
                  'or it was set incorrectly. The prompt_norm should be: ' .
                  'controller_name(::[^>#%]*)?\*?[>#%]' . "\n" . $errmsg;
        $self->_set_err($errmsg);
        return $SanUtil::FALSE;
    }

    # connection is successful so set connection environment flag
    # to either "bsd" or "ngsh" depending on username
    if ($host_obj->hosttype() =~ /.*-ng/) {
        $connect->{environment} = ($host_obj->{connection_fields}{$conn_field}{username} =~ /(root|diag)/)
                                  ? "bsd" : "ngsh";
    }

    $self->{$args{filer}}{$args{cmd_interface}} = $connect;

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    make_connect_for_host
# Objective: Creates a connection for host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub make_connect_for_host {
    my ($self, %args) = @_;

    my ($conn_str, $connect, $host, $errmsg, $retries, @validate);
    my ($username, $password, $conn_type);

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    @validate = (%args);
    eval {
        (%args) =
          Module::Common::validate_mixed(@validate,
                                          [],
                                          {client     => {type => SCALAR,
                                                         required => 1},
                                           cmd_interface => {type     => SCALAR,
                                                         optional => 1,
                                                         default  => 'cli'},
                                           conntype  => {type => SCALAR,
                                                         optional => 1},
                                           password  => {type     => SCALAR,
                                                         optional => 1},
                                           username  => {type     => SCALAR,
                                                         optional => 1}});
    };
    if ($@) {
        $self->_set_err("fail to validate parameters: " . $@);
        return $SanUtil::FALSE;
    }

    # set filer object
    $host = &Tharn::host($args{client});

    # default connection based on filer type
    if (defined($args{conntype})) {
        $conn_type = $args{conntype};
    } else {
        $conn_type = ($host->hosttype() =~ /windows/) ? "rsh" : "ssh";
    }

    # strip dash and anything after
    if ($conn_type eq 'ssh-persist') {
        $conn_type = 'ssh';
    } elsif ($conn_type eq 'rsh') {
        $conn_type = 'rsh-persay';
    }

    my $max_retries = $self->{connection_retries};

    # 1) use passed in username and password
    # 2) if not, use username and password from nate host setting
    # 3) if not, use default username=diag and password=netapp1!
    if ($args{username}) {
        $username = $args{username};
    } else {
        if (defined $host->{connection_fields}{$conn_type}{username} and
            length($host->{connection_fields}{$conn_type}{username})) {
            $username = $host->{connection_fields}{$conn_type}{username};
        }
    }
    if ($args{password}) {
        $password = $args{password};
    } else {
        if (defined $host->{connection_fields}{$conn_type}{password} and
            length($host->{connection_fields}{$conn_type}{password})) {
            $password = $host->{connection_fields}{$conn_type}{password};
        }
    }

    # initialize connection string
    $conn_str =
        $args{client} . "#"
      . $conn_type
      . $Connect_id++ . ";"
      . "conntype="
      . $conn_type;

    # append -persist to make conn_type = "ssh-persist"
    $conn_str .= "-persist" if ($conn_type eq 'ssh');

    # append username and password if connect type is not "rsh"
    unless ($conn_type =~ /rsh/) {
        $conn_str .= ";username=" . $username . ";password=" . $password;
    }

    $self->sanlog(-msg => "Connection string is $conn_str");

    # retry connection five times
    while (! defined($connect) and $retries < $max_retries) {
        $connect =
          &Tharn::connect(
            $conn_str,
            -eval=>sub {
                if ($@) {
                    $retries++;
                    $connect = undef;

                    $errmsg = join("", $@);
                    $errmsg =~ s/\n//g;

                    # Display an error in case user doesn't include error
                    # handler on object instantiation.
                    $self->sanlog(-msg => "\n\tWARNING: " . (caller(0))[3] .
                                          " unable to connect to client '" .
                                          $args{client} . "': $errmsg\n" .
                                          "Attempt $retries of $max_retries");
                    &Tharn::disconnect($conn_str, '-ok');
                    &SanUtil::nate_sleep(30);
                }
            });
    }

    $self->{$args{client}}{$args{cmd_interface}} = $connect;
    return $SanUtil::TRUE;
}


###############################################################################
# Method:    get_connection
# Objective: Gets a connection
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_connection {
    my ($self, %args) = @_;

    my (@validate);

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    @validate = (%args);
    eval {
        (%args) =
          Module::Common::validate_mixed(@validate,
                                          [],
                                          {filer     => {type     => SCALAR,
                                                         required => 1},
                                           cmd_interface => {type     => SCALAR,
                                                         optional => 1,
                                                         default  => 'cli'}});
    };
    if ($@) {
        $self->_set_err("fail to validate parameters: " . $@);
        return {};
    }

    # return valid connection
    if (exists $self->{$args{filer}}{$args{cmd_interface}}) {
        return $self->{$args{filer}}{$args{cmd_interface}};
    } else {
        return {};
    }
}

###############################################################################
# Method:    is_exist()
# Objective: Determine if connection object is exist
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_exist {
    my ($self, %args) = @_;

    my (@validate);

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    @validate = (%args);
    eval {
        (%args) =
          Module::Common::validate_mixed(@validate,
                                          [],
                                          {filer     => {type     => SCALAR,
                                                         required => 1},
                                           cmd_interface => {type     => SCALAR,
                                                         optional => 1,
                                                         default  => 'cli'}});
    };
    if ($@) {
        $self->_set_err("fail to validate parameters: " . $@);
        return $SanUtil::FALSE;
    }

    if (exists $self->{$args{filer}}{$args{cmd_interface}}) {
        return $SanUtil::TRUE;
    } else {
        return $SanUtil::FALSE;
    }
}

###############################################################################
#                               Private Methods
###############################################################################

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

ConnectionMgr.pm

=head1 Author

Korry Nguyen (dl-ssan-test)

=head1 Description

ConnectionMgr.pm provides an interface to manage a connection object.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm>

=head1 Attributes

=head2 B<reuse_conn>

boolean = <0 | 1> (default is 1) reuse of existing connection object

=head1 B<Public Methods>

=head2 B<new()>

Create a connectionMgr object. If reuse_conn flag is passed in and connectionMgr object already exist, the existing connectionMgr object will be reused.

=over 4

=item Inputs

=item * I<-reuse_conn> => <0 | 1> (default is 1)

By default, if connectionMgr object exists, it will be reused

=item Outputs

Return a connectionMgr object

=item Examples

 use San::ConnectionMgr;

 # Create connectionMgr object
 # Reuse if connectionMgr object exists
 #
 my $conn_mgr = ConnectionMgr->new();
 if ($conn_mgr->error()) {
     $self->_set_err("can not create connectionMgr object",
                     $conn_mgr->errmsg());
     return $SanUtil::FALSE;
 }

 # Create a new connectionMgr object regardless whether the
 # connectionMgr object exists or not
 #
 my $conn_mgr = ConnectionMgr->new(-reuse_conn => 0);
 if ($conn_mgr->error()) {
     $self->_set_err("can not create connectionMgr object",
                     $conn_mgr->errmsg());
     return $SanUtil::FALSE;
 }

=back

=head2 B<make_connect(%args)>

Make a connect to a filer

=over 4

=item Inputs

=item * I<-filer> => <filer_name>

Name of filer or controller

=item * I<-cmd_interface> => <cli | zapi> (default is 'cli')

Interface type to use when executing commands on the filer
(Currently only support 'cli')

=item * I<-conntype> => <ssh-persist | telnet | console> (default is 'ssh-persist')

Connection type to use when connecting to the filer
(Currently only support 'ssh-persist')

=item * I<-username> => <root | admin> (default is 'root')

User name to use when connecting to the filer

=item * I<-password> => <password for username> (default is password for 'root')

Password to use when connecting to the filer

=item Outputs

Return TRUE | FALSE (1 | 0)

=item Examples

 # Connect to a filer with default parameters
 #
 $conn_mgr->make_connect(-filer => $filer_name);
 if ($conn_mgr->error()) {
     $self->_set_err("can not connect to filer: " . $filer_name,
                     $conn_mgr->errmsg());
     return $SanUtil::FALSE;
 }

 # Connect to a filer with specific cmd_interface and conntype
 #
 $conn_mgr->make_connect(-filer     => $filer_name,
                        -cmd_interface => 'cli',
                        -conntype  => 'telnet');
 if ($conn_mgr->error()) {
     $self->_set_err("can not connect to filer: " . $filer_name,
                     $conn_mgr->errmsg());
     return $SanUtil::FALSE;
 }

=back

=head2 B<get_connection(%args)>

Get the connection object

=over 4

=item Inputs

=item * I<-filer> => <filer_name>

Name of filer or controller

=item * I<-cmd_interface> => <cli | zapi> (default is 'cli')

Interface type to use when executing commands on the filer
(Currently only support 'cli')

=item Outputs

Return connection object or empty hash

=item Examples

 # Get connection object
 #
 my $conn_obj = $conn_mgr->get_connection(-filer => $filer_name);
 if ($conn_mgr->error() or not defined($conn_obj)) {
     $self->_set_err("failed to get connection object", $conn_mgr->errmsg());
     return $SanUtil::FALSE;
 }

=back

=head2 B<is_exist(%args)>

Check to see whether connection object exists or not

=over 4

=item Inputs

=item * I<-filer> => <filer_name>

Name of filer or controller

=item * I<-cmd_interface> => <cli | zapi> (default is 'cli')

Interface type to use when executing commands on the filer
(Currently only support 'cli')

=item Outputs

Return TRUE (1) if connection object exists, otherwise FALSE (0)

=item Examples

 # Check for existing connection obj
 #
 my $conn_obj;
 if ($conn_mgr->is_exist(-filer => $filer_name) {
     # connection object exists, retrieve it
     #
     $conn_obj = $conn_mgr->get_connection(-filer => $filer_name);
 } else {
     # connection does not exist, make a connect
     #
     $conn_mgr->name_connect(-filer => $filer_name);
     if ($conn_mgr->error()) {
         $self->_set_err("failed to make a connect to " . $filer_name,
                         $conn_mgr->errmsg());
         return $SanUtil::FALSE;
     }
     # get connection
     #
     $conn_obj = $conn_mgr->get_connection(-filer => $filer_name);
 }

=cut
