# Copyright (c) 2007-2010 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to manage a cluster
##
## @author korry@netapp.com, dl-ssan-automation
## @status Shared

package San::ControllerCluster;

# compiler directives
use strict;

# imported modules
use San::SanFiler;
use Params::Validate qw( :all );
use ParamsMgr;

# inherit base class
use base qw( San::CommonInterface );

###############################################################################
# Method Name: new
# Objective: Create new Cluster object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($cserver, $host_obj, %validate_params);
    my $is_cserver = $SanUtil::FALSE;

    # parameter initialization
    my $self = {};
    bless $self, $class;

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);
    my @input_params = (%args);

    # validation_options is set to allow extra params not defined in %args to bypass
    Params::Validate::validation_options( allow_extra => 1);

    # validate_params is the arguments to validate
    %validate_params = (filer         => { type     => SCALAR,
                                           optional => 1 },
                        cserver       => { type     => SCALAR,
                                           optional => 1 },
                        cmd_interface => { type     => SCALAR,
                                           optional => 1,
                                           default  => 'cli' },
                        reuse_conn    => { type     => SCALAR,
                                           optional => 1,
                                           default  => 1 },
                        build_nodes   => { type     => SCALAR,
                                           optional => 1,
                                           default  => 1 },
                        debug_level   => { type     => SCALAR,
                                           optional => 1,
                                           default  => 0 },
                        new_cmds_obj  =>{ type     => SCALAR,
                                           optional => 1,
                                           default  => 0 }
                       );
    eval {
        (%args) = Params::Validate::validate(@input_params, \%validate_params);
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $self;
    }

    # we need filer or cserver param
    if (exists($args{filer}) and length($args{filer})) {

        # create host object for node
        $host_obj = &SanUtil::create_host_obj(filer => $args{filer});
    } elsif (exists($args{cserver}) and length($args{cserver})) {

        # create host object for cserver
        $host_obj = &SanUtil::create_host_obj(cserver => $args{cserver});
        $is_cserver = $SanUtil::TRUE;
    } else {
        $self->_set_err("cserver or filer must be specified");
        return $self;
    }

    # create 10-mode specific cluster
    if ($host_obj->hosttype() =~ /^filer-ng|^vsim-ng/i) {

        # use base class to create new cluster object
        $self = $class->SUPER::_new(-name          => $host_obj->hostname(),
                                    -filer         => $host_obj->hostname(),
                                    -node          => $host_obj->{node},
                                    -reuse_conn    => $args{reuse_conn},
                                    -cmd_interface => $args{cmd_interface},
                                    -is_cserver    => $is_cserver,
                                    -debug_level   => $args{debug_level},
                                    -new_cmds_obj  => $args{new_cmds_obj}
                                    );
        if ($self->error()) {
            $self->_set_err("failed to create cluster object", $self->errmsg());
            return $self;
        }

        # refresh to update cluster object
        if ($self->{enable_cache}) {
            $self->refresh();
            if ($self->error()) {
                $self->_set_err("failed to refresh cluster object", $self->errmsg());
                return $self;
            }
        }
    } else {
        $self->_set_err("invalid hosttype: ", $host_obj->hosttype());
        return $self;
    }

    # if build_nodes flag is set, create cserver and node objects in ControllerCluster object
    if ($args{build_nodes} == $SanUtil::TRUE) {
        $self->_discover_cserver();
        if ($self->error()) {
            $self->_set_err("failed to discover object: cserver", $self->errmsg());
            return $self;
        }
        $self->_discover_nodes();
        if ($self->error()) {
            $self->_set_err("failed to discover object: nodes", $self->errmsg());
            return $self;
        }
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    return $self;
} ## end sub new

################################################################################
# Method Name: reboot()
# Objective: Reboots all the nodes in the cluster
# Details: See POD documentation at the end of this file
################################################################################
sub reboot {
    my ($self, %args) = @_;

    my (@node_objs, @subtests);
    my ($node_obj, $parallel, $subtest);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    my @input_params = (%args);
    eval {
        # validate input parameters
        (%args) = validate(@input_params,
                          { parallel   => { type => SCALAR,
                                            optional => 1,
                                            regex => qr /^(0|1)$/,
                                            default => 1},
                            wait_time  => { type => SCALAR,
                                            optional => 1,
                                            regex => qr /^\d+$/,
                                            default => 0},
                          });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    @node_objs = $self->get_node_objs();
    if ($self->error() or scalar (@node_objs) == 0) {
        $self->_set_err ("failed to get node objs",
                         $self->errmsg());
        return $SanUtil::FALSE;
    }

    foreach $node_obj (@node_objs) {
        if ($args{parallel}) {
            $subtest = &Tharn::subtest(sub {
                                            $node_obj->reboot();
                                       },
                                       -runid  => "reboot_" . $node_obj->get_name(),
                                      '-ipc',
                                      '-bg'
                       );
            push(@subtests, $subtest);

            # Sleeping for 5 secs before proceeding to next node
            sleep($args{wait_time});
        } else {
            $node_obj->reboot();
            if ($node_obj->error()) {
                $self->_set_err ("failed to reboot node " .
                                 $node_obj->get_name(),
                                 $node_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }
    }

    if ($args{wait_time}) {
        Subtest::wait_finish(subtests => \@subtests);

	# If parallel is set to 1, the reboot must have reconnected in the subtest
	# This can be handled much cleanly, but rather than doing it here, there
	# is a need for this kind of support all over the libs
	foreach $node_obj (@node_objs) {
	    $node_obj->reconnect();
	    if ($node_obj->error()) {
	        $self->_set_err ("failed to reboot node " . $node_obj->get_name(),
			         $node_obj->errmsg());
	        return $SanUtil::FALSE;
	    }
	}
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: ha_modify()
# Objective: Modifies cluster ha configured state
# Details: See POD documentation at the end of this file
################################################################################
sub ha_modify {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # call cluster_ha_modify
    (@output) = $self->invoke_cmd_obj(%args,
                                      cmd => "cluster_ha_modify");
    if ($self->error()) {
        $self->_set_err ("failed to get node objs",
                         $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: enable_sfo()
# Objective: enable sfo
# Details: See POD documentation at the end of this file
################################################################################
sub enable_sfo {
    my ($self, %args) = @_;

    my ($node_obj);
    my (@node_objs);
    my $wait_time = 5;

    %args = &SanUtil::strip_dash(%args);

    @node_objs = $self->get_node_objs();
    if ($self->error() or scalar (@node_objs) == 0) {
        $self->_set_err ("failed to get node objs",
                         $self->errmsg());
        return $SanUtil::FALSE;
    }

    if ($self->is_sfo_enabled()) {
        $self->sanlog(-msg => "sfo is already enabled",
                      -msg_level => 1);
        return $SanUtil::TRUE;
    }

    # Verify if is_sfo_enabled resulted in any error
    if ($self->error()) {
        $self->_set_err("failed to verify whether sfo is enabled",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # by default no reboot will occur unless the next piece of code fails
    $args{inhibit_reboot} = $SanUtil::TRUE;

    # first see if sfo can be enabled without doing a reboot.
    # this can happen if someone had rebooted their cluster prior to running
    # this script
    foreach $node_obj (@node_objs) {
        # SN requiers a step before enabling the SFO
        my $release_info = $node_obj->get_release();
        if ($node_obj->error()) {
            $node_obj->_set_err("Failed to retrive the release info". $node_obj->errmsg());
            return $SanUtil::FALSE;
        }

        if ($release_info =~ /SierraNevada/) {
            $node_obj->sfo_mode(-mode => "ha");
            if ($node_obj->error()) {
                $node_obj->_set_err("Failed to modify the mode of SFO". $node_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # force the reboot to occur
            $args{inhibit_reboot} = $SanUtil::FALSE;
            next;
        } else {
            $node_obj->sfo_modify(-enabled => "true");
            if ($node_obj->error()) {
                # clear the error on the node and indicate that a
                # reboot must be done
                $node_obj->_clear_err();

                # force the reboot to occur
                $args{inhibit_reboot} = $SanUtil::FALSE;
                next;
            }
        }
    }

    # A reboot is necessary for enabling sfo before running storage failover
    # command. We will wait for 5 seconds between each reboot
    unless ($args{inhibit_reboot}) {
        $self->reboot(-parallel => $SanUtil::TRUE,
                      -wait_time => $wait_time);
        if ($self->error()) {
            $self->_set_err("failed to reboot nodes",
                            $self->errmsg());
            return $SanUtil::FALSE;
        }

        # now try to enable sfo after the reboot
        foreach $node_obj (@node_objs) {
            $node_obj->sfo_modify(-enabled => "true");
            if ($node_obj->error()) {
                $self->_set_err ("failed to modify sf enabled flag to true " .
                                     $node_obj->get_name(),
                                 $node_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }
    }

    # Unidirectional sfo is enabled in 2+ node config. cluster ha command must be
    # run in such a case
    if (scalar @node_objs == 2) {
        $self->ha_modify(-configured => "true");
        if ($self->error()) {
            $self->_set_err ("failed to enable ha", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # Verify if sfo is enabled
    if ($self->is_sfo_enabled()) {
        return $SanUtil::TRUE;
    }

    # Verify if is_sfo_enabled resulted in any error
    if ($self->error()) {
        $self->_set_err("failed to verify whether sfo is enabled",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: disable_sfo()
# Objective: disable sfo
# Details: See POD documentation at the end of this file
################################################################################
sub disable_sfo {
    my ($self, %args) = @_;

    my ($node_obj);
    my (@node_objs);

    %args = &SanUtil::strip_dash(%args);

    @node_objs = $self->get_node_objs();
    if ($self->error() or scalar (@node_objs) == 0) {
        $self->_set_err ("failed to get node objs",
                         $self->errmsg());
        return $SanUtil::FALSE;
    }

    foreach $node_obj (@node_objs) {
        $node_obj->disable_sfo();
        if ($node_obj->error()) {
            $self->_set_err ("failed to disable sfo on node " .
                             $node_obj->get_name(),
                             $node_obj->errmsg());
            return $SanUtil::FALSE;
        }
        if ($node_obj->is_sfo_enabled()) {
            $self->_set_err ("sfo is not disabled for node " .
                             $node_obj->get_name(),
                             $node_obj->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: is_sfo_enabled()
# Objective: verify if sfo is enabled on all the nodes
# Details: See POD documentation at the end of this file
################################################################################
sub is_sfo_enabled {
    my ($self, %args) = @_;

    my (@node_objs);

    %args = &SanUtil::strip_dash(%args);

    @node_objs = $self->get_node_objs();
    if ($self->error() or scalar (@node_objs) == 0) {
        $self->_set_err ("failed to get node objs",
                         $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Verify whether sfo is enabled on all the nodes.
    foreach my $node_obj (@node_objs) {
        if ($node_obj->is_sfo_enabled()) {
            next;
        } else {
            if ($node_obj->error()) {
                $self->_set_err ("failed to verify if sfo is enabled for node ".
                                 $node_obj->get_name(),
                                 $self->errmsg());
                return $SanUtil::FALSE;
            }
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name:     get_cserver_obj()
# Objective:       Get cserver obj from cluster
# Details:         See POD documentation at the end of this file
################################################################################
sub get_cserver_obj {
    my ($self) = @_;

    $self->_clear_err();

    # retrieve node objects from cache if exist
    if (defined $self->{cserver_obj}) {
        return $self->{cserver_obj};
    } else {
        # discover cserver here
        my $c_obj = $self->_discover_cserver();
        if ($self->error()) {
            $self->_set_err("failed to discover object: cserver", $self->errmsg());
            return {};
        }

        return $c_obj;
    }
} ## end sub get_cserver_obj

################################################################################
# Method Name: get_node_objs()
# Objective:       Get node objects from cluster
# Details:         See POD documentation at the end of this file
################################################################################
sub get_node_objs {
    my ($self, %args) = @_;

    my (@objs);

    $self->_clear_err();

    # remove dash '-' from input parameters because parameters may be used
    # to create get accessor method on the fly and easy to process
    %args = &SanUtil::strip_dash(%args);

    # set name to node
    if (exists $args{name}) {
        $args{node} = delete $args{name};
    }

    # retrieve node objects from cache if exist
    if (defined @{$self->{node_objs}}) {
        push @objs, @{$self->{node_objs}};
    } else {
        # discover nodes here
        @objs = $self->_discover_nodes();
        if ($self->error()) {
            $self->_set_err("failed to discover object: nodes", $self->errmsg());
            return (wantarray()) ? () : [];
        }
        $self->_set_err("node objects don't exist");
        return (wantarray()) ? () : [];
    }

    # search and select objects based on arg count
    @objs = $self->search(objs => \@objs, %args) if (keys %args);
    if ($self->error() or scalar(@objs) == 0) {
        $self->_set_err("search found no matching object", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return $objs[0] if ($args{count} == 1);
    return (wantarray()) ? @objs : \@objs;
} ## end sub get_node_objs

################################################################################
# Method Name: get_node_members()
# Objective:       Return a list of node member names that belong to this
#                  cluster
# Details:         See POD documentation at the end of this file
################################################################################
sub get_node_members {
    my ($self) = @_;

    my @node_objs;

    # retrieve node objects from cache if they exist
    if (scalar(@{$self->{node_objs}})) {
        @node_objs = @{$self->{node_objs}};
    } else {
        # if they dont exist, then get the node objects by querying the system
        @node_objs = $self->get_node_objs();
        if ($self->error()) {
            $self->_set_err("Unable to get node objects from the cluster",
                            $self->errmsg());
            return (wantarray()) ? () : [];
        }
    }

    # pull out the names of the nodes and return to the caller
    my @node_names = map {
        $_->get_name();
    } @node_objs;

    return (wantarray()) ? @node_names : \@node_names;
} ## end sub get_node_members

###############################################################################
# Method Name: kcs_list_info()
# Objective: Get kernel service information on all nodes in this cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub kcs_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "cluster_kernel_service_show");

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain igroup object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub kcs_list_info

###############################################################################
# Method Name: refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    my (@tmp_objs, @vol_objs);

    # initialize cmd arguments for cluster show
    %args = (%args, cmd => "cluster_identity_show");

    # hand responsibility to base class to update the object
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        $self->_set_err("failed to refresh cluster object data", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub refresh

################################################################################
# Method Name: get_contact()
# Objective:   Get cluster contact
# Details:     See POD documentation at the end of this file
################################################################################
sub get_contact {
    my ($self) = @_;

    # the get() will handle error
    return $self->get("contact");
}

################################################################################
# Method Name: get_location()
# Objective:   Get cluster location
# Details:     See POD documentation at the end of this file
################################################################################
sub get_location {
    my ($self, %args) = @_;

    $args{field} = "location";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_name()
# Objective:   Get the cluster name, from cache or from the controller
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "name";

    # If the param use_cache is not defined, will default to true.
    #
    # Will return the name in memory if 'use_cache'=True, that is the hostname
    # Will return the name of the cluster in the controller, if 'use_cache'=False.
    unless ( defined($args{use_cache}) ){
        $args{use_cache} = $SanUtil::TRUE;
    }

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_serialnumber()
# Objective:   Get cluster serialnumber
# Details:     See POD documentation at the end of this file
################################################################################
sub get_serialnumber {
    my ($self, %args) = @_;

    $args{field} = "serialnumber";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_uuid()
# Objective:   Get cluster uuid
# Details:     See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self, %args) = @_;

    $args{field} = "uuid";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method Name: shutdown()
# Objective: shutdown a cluster??
# Details:   See POD documentation at the end of this file.
###############################################################################
sub shutdown {
    my ($self, %args) = @_;

    return $SanUtil::FALSE;
}

###############################################################################
# Method Name: unjoin()
# Objective: unjoin a node from the cluster
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unjoin {
    my ($self, %args) = @_;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    my @input_params = (%args);
    eval {
        (%args) =
          Module::Common::validate_mixed(@input_params,
                                         [],
                                         {
                                             node => {type => SCALAR,
                                                      required => 1},
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # Grab the cserver object to execute this command from
    my $cserver_obj = $self->get_cserver_obj();
    if($cserver_obj->error()){
        $self->_set_err("failed get_cserver_obj: " . $cserver_obj->err_msg());
        return $SanUtil::FALSE;
    }

    # build cmd "cluster unjoin"
    %args = (%args, cmd => "cluster_unjoin");

    # invoke cdef command but do not update attributes
    my (@output) = $cserver_obj->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($cserver_obj->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $cserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: get_state()
# Objective: Gets the state of the each node in the cluster.
# Details: See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self, %args) = @_;
    #Locals
    my (@records);

    # get all healthy nodes in a cluster
    %args =('cmd' => "cluster_show",
            'fields' => "health",
            %args
            );

    (@records) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("Fail to get the health of the nodes in a cluster",
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @records : \@records;
} ## end get_state

################################################################################
# Method Name: get_node_state()
# Objective: Gets the state of a particular node in the cluster.
# Details: See POD documentation at the end of this file
################################################################################
sub get_node_state {
    my ($self, %args) = @_;
    #Locals
    my @records = ();

    # verify mandatory field argument
    unless (defined $args{node} and length $args{node}) {
        $self->_set_err("mandatory parameter \"node\" not specified ");
        return "";
    }

    (@records) = $self->get_state(%args);
    if ($self->error()) {
        $self->_set_err("Fail to get the health state of node". $args{'node'}.
                        "in the cluster", $self->errmsg());
        return "";
    }

    return $records[0]->{'health'};
} ## end get_state

################################################################################
# Method Name: intercluster_obj_create()
# Objective: Creates an intercluster object, that is an instance of the Intercluster.pm
# Details: See POD documentation at the end of this file
################################################################################
sub intercluster_obj_create {
    my ($self, %args) = @_;
    $self->_clear_err();
    # Locals
    my $intercluster_obj = undef;

    # Verify mandatory field argument
    unless (defined $args{'CLUSTERS'} and length $args{'CLUSTERS'}) {
        $self->_set_err("Mandatory parameter \"CLUSTERS\" not specified ");
        return undef;
    }

    # Create the object
    eval {
        require Intercluster;
        $intercluster_obj = new Intercluster(%args);
    };
    if ($@) {
        $self->_set_err("FAILED to create an Intercluster object: " . $@);
        return undef;
    }

    return $intercluster_obj;
}

################################################################################
# Method Name: intercluster_connection_create()
# Objective: Creates the connection between two clusters
# Details: See POD documentation at the end of this file
################################################################################
sub intercluster_connection_create {
    my ($self, %args) = @_;
    $self->_clear_err();
    #Locals
    my $intercluster_obj = undef;

    # Verify mandatory field argument
    if (defined $args{'intercluster_obj'} and length $args{'intercluster_obj'}) {
        # If intercluster_obj provided, use it
        $intercluster_obj = $args{'intercluster_obj'};
    } else {
        # If intercluster_obj was not provided, but clusters were, create the
        # intercluster obj
        if (defined $args{'CLUSTERS'} and length $args{'CLUSTERS'}) {
            $intercluster_obj = $self->intercluster_obj_create(%args);
        } else {
            $self->_set_err("Did NOT specified neither \"intercluster_obj\" parameter " .
                            "nor  \"CLUSTERS\" parameter. At least one of them needs to ".
                            "be provided. ");
            return $SanUtil::FALSE;
        }
    }

    # If intercluster connectivity has been established, no work to do...
    unless ($self->intercluster_connection_exists('intercluster_obj' => $intercluster_obj)) {
        # Config the intercluster lifs
        $intercluster_obj->config_intercluster_lifs();
        # Create the intercluster peers using admin
        $intercluster_obj->create_intercluster_peers();
    }

    # Check cluster peer connectivity
    $intercluster_obj->check_peer_connectivity();

    return $SanUtil::TRUE;

} ## end intercluster_create

################################################################################
# Method Name: intercluster_connection_exists()
# Objective: Verifies if the connection between cluster peers exist.
# Details: See POD documentation at the end of this file
################################################################################
sub intercluster_connection_exists {
    my ($self, %args) = @_;
    $self->_clear_err();
    #Locals
    my $answer = $SanUtil::FALSE;
    my $intercluster_obj = undef;
    my @intercluster_keys = ();
    my $peer_cluster = '';

    # verify mandatory field argument
    unless (defined $args{'intercluster_obj'} and length $args{'intercluster_obj'}) {
        $self->_set_err("Mandatory parameter \"intercluster_obj\" not specified ");
        return $SanUtil::FALSE;
    }

    # Get values
    $intercluster_obj = $args{'intercluster_obj'};
    @intercluster_keys = keys(%{$intercluster_obj});
    $peer_cluster = $intercluster_keys[1];

    eval {
         $answer = $intercluster_obj->peer_ping('dest_cluster'=> $peer_cluster);
    };
    if ($@) {
        $self->_set_err("FAIL: Unable to PING cluster peer: /n" . $@);
        return $SanUtil::FALSE;
    }

    return $answer;

} ## end intercluster_create

################################################################################
#                            Private Methods
################################################################################

################################################################################
# Method Name:     _discover_cserver()
# Objective:       Discover cserver and build object
# Details:         See POD documentation at the end of this file
################################################################################
sub _discover_cserver {
    my ($self) = @_;

    my ($status, @records, $cserver);
    my ($cluster_mgmt_ip, @nodes);

    # get cluster name for cserver
    my %args = (cmd => "cluster_identity_show");

    my (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return {};
    }

    # get cluster server (cserver) name
    $cserver = $output[0]->{name};
    $self->sanlog(-msg => "cserver: $cserver", -msg_level => 3);

    # get nodes in cluster
    @nodes = $self->_nodes();
    if ($self->error()) {
        $self->_set_err("fail to get list of nodes in cluster", $self->errmsg());
        return {};
    }

    # get cluster mgmt ip address dynamically
    $cluster_mgmt_ip = $self->_get_clus_mgmt_ip($cserver);
    if ($self->error()) {
        $self->_set_err("fail to get cserver ip", $self->errmsg());
        return {};
    }

    # retrieve buildinfo from the object
    $args{buildinfo}  = $self->get_buildinfo() if (defined($self->{buildinfo}));
    $args{conn}       = $self->{conn_obj} if (defined($self->{conn_obj}));
    $args{timeout}    = $self->{timeout}  if (defined($self->{timeout}));
    $args{debug_level} = $self->get_debug_level();

    # create cserver (SanFiler) instance using host machine
    # connection of same name.
    my $cserver_obj = SanFiler->new(-cserver => $cserver,
                                    -cluster_mgmt_ip => $cluster_mgmt_ip,
                                    -nodes_in_cluster => \@nodes,
                                    -server_type => "cserver",
                                    %args);
    if ($cserver_obj->error() || ref($cserver_obj) !~ /(Filer10M)/) {
        $self->_set_err("fail to instantiate cserver object", $cserver_obj->errmsg());
        return {};
    }

    # cache cserver object
    $self->{cserver_obj} = $cserver_obj;

    return $cserver_obj;
} ## end sub _discover_cserver

################################################################################
# Method Name: _nodes()
# Objective:       Return array of node names
# Details:         See POD documentation at the end of this file
################################################################################
sub _nodes {
    my ($self) = @_;

    my ($status, @records, @nodes);

    # get all healthy nodes in a cluster
    (@records) = $self->execute(-cmd => "cluster show -health true -fields node");
    if ($self->error()) {
        $self->_set_err("fail to discover nodes in a cluster", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # each record is a node
    foreach my $record (@records) {
        push @nodes, $record->{node};
    }

    return (wantarray()) ? @nodes : \@nodes;
} ## end sub _nodes

################################################################################
# Method Name: _get_clus_mgmt_ip()
# Objective:       Get cluster mgmt ip address
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_clus_mgmt_ip {
    my ($self, $cserver) = @_;

    # get cluster management ip address
    my (@records) = $self->execute(-cmd => "network interface show -fields role,address -vserver " . $cserver);
    if ($self->error() or not scalar(@records)) {
        $self->_set_err("Failed to determine the cluster mgmt ip address for the cluster $cserver. " .
                        "Make sure that the cluster $cserver has a cluster mgmt vserver with a pingable " .
                        "cluster mgmt lif.", $self->errmsg());
        return "";
    }

    foreach my $record (@records) {
        if ($record->{role} =~ /cluster-mgmt/i) {
            return $record->{address};
        }
    }

    # if we are here, we did not find the cluster-mgmt lif, so set an error    
    $self->_set_err("Failed to determine the cluster mgmt ip address for the cluster $cserver. " .
                        "Make sure that the cluster $cserver has a cluster mgmt vserver with a pingable " .
                        "cluster mgmt lif.");
    return "";
} ## end sub _get_clus_mgmt_ip

################################################################################
# Method Name: _discover_nodes
# Objective:       return array of node objects
# Details:         See POD documentation at the end of this file
################################################################################
sub _discover_nodes {
    my ($self, %args) = @_;

    my (@nodes, @node_objs);

    # retrieve the list of healthy nodes in cluster
    @nodes = $self->_nodes();
    if ($self->error()) {
        $self->_set_err("fail to get list of nodes in cluster", $self->errmsg());
        return {};
    }

    # retrieve buildinfo from the object
    $args{buildinfo}  = $self->get_buildinfo() if (defined($self->{buildinfo}));
    $args{conn}       = $self->{conn_obj} if (defined($self->{conn_obj}));
    $args{timeout}    = $self->{timeout}  if (defined($self->{timeout}));
    $args{debug_level} = $self->get_debug_level();

    # create node (SanFiler) instance using host machine
    # connection of same name.
    foreach my $node (@nodes) {
        my $node_obj = SanFiler->new(-filer => $node,
                                     -server_type => "node",
                                     %args);
        if ($node_obj->error() || ref($node_obj) !~ /(Filer10M|SanFiler7M)/) {
            $self->_set_err("fail to instantiate SanFiler object", $node_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        # store temporary node object
        push @node_objs, $node_obj;
    }

    # cache node objects
    $self->{node_objs} = \@node_objs;

    return (wantarray()) ? @node_objs : \@node_objs;
} ## end sub discover_nodes

###############################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the cluster
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory field argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # convert common args to args specifically for creating lif
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "cluster_identity_show",
                                                field => $args{field});
    }
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

San::ControllerCluster.pm

=head1 Author

Korry Nguyen (dl-ssan-automation)

=head1 Description

Provides an interface to manage a C-mode cluster

=head1 See Also

Package @ISA base class B<San::CommonInterface> that inherits other modules. See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm, San::SanCache.pm>

=head1 Attributes

=head2 B<cserver>

string = cluster management IP/name

=head2 B<reuse_conn>

boolean = 0 | 1 (default is 1)

=head2 B<debug_level>

interger = 0 - 100 (default is 0)

=head2 B<timeout>

interger = seconds (default is 1800)

=head1 B<Public Methods>

=head2 B<new(%args)>

Create a controller cluster object

=over 4

=item Inputs

=item * I<-cserver> => <cserver IP or name>

Cluster management IP address or name

=item * I<-reuse_conn> => <reuse connection flag>

Reuse connection policy

=item * I<-timeout> => <number of seconds>

Times that take the command to timeout

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. The higher the number it will log more debug messages

=item * I<-new_cmds_obj> => <new CmdsDriver obj flag>

If true (1) CommonInterface will create a new CmdsDriver object for the cluster.
This is usefull when creating ControllerCluster objects with different builds.

=item Outputs

Return controller cluster object

=item Examples

 use San::ControllerCluster;

 my $cserver = "ssan-3170-cluster";

 my $cluster_obj = new San::ControllerCluster(-cserver => $cserver,
                                              -timeout => 180,
                                              -debug_level => 0,
                                              -new_cmds_obj => 0,
                                             );
 if ($cluster_obj->error) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to create cluster object: $err");
     return TCD::FAIL;
 }

=back

=head2 B<reboot()>

Reboots all the nodes in the cluster

=over 4

=item Inputs

=item * I<-parallel> => <1|0>, default=1

If 1, all nodes will be rebooted parallely otherwise sequentially.

=item * I<-wait_time> => <integer>, default=0

If parallel is 1, we start reboots on all nodes at the same time.  This
parameter is the amount of time to sleep before invoking reboot on each node. In
a 2 node cluster, a reboot of node 1 will be followed by wait_time followed by
reboot on node2. Applicable only if parallel is set to 1

=item Outputs

Returns $SanUtil::TRUE if all the nodes are rebooted successfuly otherwise sets
error andreturns $SanUtil::FALSE

=item Examples

 $cluster_obj->reboot();
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to reboot: $err");
     return TCD::FAIL;
 }

=back

=head2 B<ha_modify()>

Modify cluster ha attributes

=over 4

=item Inputs

=item * I<-configured> => <true|false>, default=false

Enables cluster ha if configured is true and disables ha if false

=item Outputs

Returns $SanUtil::TRUE if all the nodes are rebooted successfuly otherwise sets
error andreturns $SanUtil::FALSE

=item Examples

 $cluster_obj->ha_modify(-configured => "true");
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to reboot: $err");
     return TCD::FAIL;
 }

=back

=head2 B<enable_sfo()>

Enable sfo for all nodes in the cluster

=over 4

=item Inputs

=item * I<-inhibit_reboot> => 1|0, default=0

If inhibit_reboot is set to 1, this method will attempt to enable sfo without
reboot. Otherwise a reboot is done before sf is enabled

=item Outputs

Returns $SanUtil::TRUE if sfo is enabled successfuly otherwise sets error and
returns $SanUtil::FALSE

=item Examples

 $cluster_obj->enable_sfo();
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to enable sfo: $err");
     return TCD::FAIL;
 }

=back

=head2 B<disable_sfo()>

disable sfo on all nodes in the cluster

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE if sfo is disabled successfuly otherwise sets error
and returns $SanUtil::FALSE

=item Examples

 $cluster_obj->disable_sfo();
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to enable sfo: $err");
     return TCD::FAIL;
 }

=back

=head2 B<is_sfo_enabled()>

Determine whether sfo is enabled

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE if sfo is enabled, otherwise $SanUtil::FALSE

=item Examples

 # determine whether sfo is enabled
 $is_sfo_enabled = $clus_obj->is_sfo_enabled();
 if ($clus_obj->error()) {
    ...
 }

=back

=head2 B<get_node_objs()>

Get node objects in the cluster

=over 4

=item Inputs

None

=item Outputs

List of node objects

=item Examples

 @node_objs = $cluster_obj->get_node_objs();
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to get node objects: $err");
     return TCD::FAIL;
 }

=back

=head2 B<get_cserver_obj()>

Get cserver object in the cluster

=over 4

=item Inputs

None

=item Outputs

Cserver object

=item Examples

 cserver_obj = $cluster_obj->get_cserver_obj();
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to get cserver object: $err");
     return TCD::FAIL;
 }

=back

=head2 B<get_node_members()>

Get node member names that are in the cluster

=over 4

=item Inputs

None

=item Outputs

List of node name strings.

=item Examples

 @node_names = $cluster_obj->get_node_members();
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to get node names: $err");
     return TCD::FAIL;
 }

=back

=head2 B<refresh()>

Refresh and update cluster environment

=over 4

=item Inputs

None

=item Outputs

Return TRUE | FALSE (1 | 0)

=item Examples

 $cluster_obj->refresh();
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to refresh cluster object: $err");
     return TCD::FAIL;
 }

=back

=head2 B<get_name()>

Will return the name of the cluster that is either in memory (that was used when
creating the object) or in the controller (the cluster identity)

=over 4

=item Inputs

=item * I<-use_cache> => True|False

Optional, if the parameter is not defined, will default to True, hence, will
return the name that is in memory.
If it's set to FALSE, will return the name of the cluster in the controller.

=item Outputs

Return String name

=item Examples

 $cluster_obj->get_name('use_cache'=>$SanUtil::FALSE);
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to get the name of the cluster: $err");
     return TCD::FAIL;
 }

=back

=head2 B<unjoin()>

Unjoin the specified node from the cluster

=over 4

=item Inputs

=item * I<-node> => name of node to unjoin

Controller name to unjoin

=item Outputs

Returns TRUE if the node was unjoined successfully, FALSE otherwise.

=item Examples

 $cluster_obj->unjoin( -node => $node_name );
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to unjoin: $err");
     return $TCD::FAIL;
 }

=back

=head2 B<kcs_list_info()>

Get kernel service info from a cluster

=over 4

=item Inputs

None

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing kernel service info. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my @kcs_info = $cluster_obj->kcs_list_info();
 if ($cluster_obj->error()) {
      logcomment("Failed to get kcs info:", $cluster_obj->errmsg());
      return $TCD::FAIL;
 }

=back

=head2 B<get_state()>

Will return an array with the health (state) of each node in the cluster.

=over 4

=item Inputs

None

=item Outputs

A Hash array with the node name (key node) and the state (key health) of each
node in the array.

=item Examples

 my @states = $cluster_obj->get_state();
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to get the state of all the nodes in the cluster: $err");
     return TCD::FAIL;
 }

=back

=head2 B<get_node_state()>

Will return the health of the node. True (1) if healthy or False (0) if the node is
down.

=over 4

=item Inputs

=item * I<-node> => name of node to get the state. Required

=item Outputs

Boolean string.

=item Examples

 my $state = $cluster_obj->get_node_state('node'=>"ssan-3210-10a");
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to get the state of all the ssan-3210-10a node: $err");
     return TCD::FAIL;
 }

=back

=head2 B<intercluster_obj_create()>

Will create an instance of the class Intercluster.pm used for connecting two clusters

=over 4

=item Inputs

=item * I<-CLUSTERS> => The hash of hashes required by Intercluster.pm

=item Outputs

The Intercluster obj.

=item Examples

 my $intercluster_obj = $cluster_obj->intercluster_obj_create('CLUSTERS'=> \%hash_of_clusters);
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to create an instance of Intercluster class: $err");
     return TCD::FAIL;
 }

=back

=head2 B<intercluster_connection_create()>

Will create the connection between two clusters.

=over 4

=item Inputs

=item * I<-intercluster_obj> => An instance of the Intercluster.pm

=item Inputs

=item * I<-CLUSTERS> =>  The hash of hashes required by Intercluster.pm to create
an instance

=item Outputs

Boolean string.

=item Examples

 my $peer_connection_flag = $cluster_obj->intercluster_connection_create('intercluster_obj'=>$intercluster_obj);
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to create a connection with the provided intercluster object: $err");
     return TCD::FAIL;
 }

=back

=head2 B<intercluster_connection_exists()>

Will verify if there is connectivity between the clusters (cluster peer ping).

=over 4

=item Inputs

=item * I<-intercluster_obj> => An instance of the Intercluster.pm

=item Outputs

Boolean string.

=item Examples

 my $connection_exists_flag = $cluster_obj->intercluster_connection_exists('intercluster_obj'=>$intercluster_obj);
 if ($cluster_obj->error()) {
     $err = $cluster_obj->errmsg();
     logcomment("Failed to verify the intercluster connection: $err");
     return TCD::FAIL;
 }

=cut
