# Copyright (c) 2008-2010 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to 10-mode filer
##
## @author korry@netapp.com, dl-ssan-automation
## @status Shared

package San::ControllerCluster::Filer10M;

# compiler directives
use strict;

# imported modules
use San::ControllerCluster::Filer10M::Aggregate;
use San::ControllerCluster::Filer10M::Vserver;
use San::ControllerCluster::Filer10M::Hwresource;
use Params::Validate qw( SCALAR ARRAYREF HASHREF OBJECT );
use boot qw ( reboot_to boot_state );

# inherit base class
use base qw( San::CommonInterface
             San::ControllerCluster::FilerHelper
             San::ControllerCluster::FilerCompat );

###############################################################################
# Method Name: new
# Objective:   Create new Filer10M object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, @list_names, $host_obj, $hwresource_obj);

    # parameter initialization
    $self = {};
    bless $self, $class;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # process arguments
    while (my ($key, $value) =  each(%args)) {
        if ($key =~ /-?filer$/) {
            if (! length($value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
        } elsif ($key =~ /-?is_cserver$/) {
            if (! length($value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
        } elsif ($key =~ /-?(node|name)$/) {
            if (! length($args{node} = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
        } elsif ($key =~ /-?buildinfo$/) {
            if (! length($value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
        } elsif ($key =~ /-?debug_level$/) {
            if (! length($value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
        } elsif ($key =~ /-?server_type$/) {
            if (! length($value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $self;
            }
        }
    }

    # Check mandatory parameter.
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }
    # use filer if node/name not specified
    unless (exists $args{node}) {
        $args{node} = $args{filer};
    }

    # default the server type to 'node'
    unless (exists $args{server_type}) {
        $args{server_type} = "node";
    }

    # San::CommonInterface is a base (generic, or SUPER) class
    # use base class SUPER::_new to contruct filer (node) object
    $self = $class->SUPER::_new(-name  => $args{filer},
                                -filer => $args{filer},
                                %args);
    if ($self->error()) {
        $self->_set_err("failed in calling SUPER::_new()", $self->errmsg());
        return $self;
    }

    # store the server type (valid values are 'cserver' or 'node')
    $self->{server_type} = $args{server_type};

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # although we already have the build info hash, we need to call
    # _get_build_info in order to set its attributes per node/filer
    $self->_get_build_info();
    if ($self->error()) {
        $self->_set_err("get controller build info failed", $self->errmsg());
        return $self;
    }

    # get hosttype for filer object
    if (defined($self->{conn_obj})) {
        my $conn = $self->{conn_obj}->get_connection(
                                 -filer  => $self->get("filer"),
                                 -cmd_interface => $self->get("cmd_interface"));
        unless (ref($conn) =~ /Connectrec/i) {
            $self->_set_err("failed to get connection");
            return $self;
        }

        $self->{type} = $conn->hostp->hosttype();
    } else {
        if ($args{is_cserver}) {

            # create host object for cserver dynamically via first node
            $host_obj = &SanUtil::create_host_obj(-cserver => $args{filer});
        } else {

            # create host object for cserver/node dynamically via first node
            $host_obj = &SanUtil::create_host_obj(-filer => $args{filer});
        }
        $self->{type} = $host_obj->hosttype();
    }

    # add type to the fields list
    push @{$self->{fields}}, "type";

    # a list of object names that will be created on-the-fly by
    # create_accessor_methods. For each name element, create_accessor_methods
    # will generate its discover and get_objs methods
    @list_names = qw(aggregate disk vserver);

    # create 'get_obj' accessor methods for disk, aggr, and vserver methods
    # ex: get_aggregate_objs, get_disk_objs, get_vserver_objs
    # create 'discover' accessor methods for disk, aggr, and vserver methods
    # ex: discover_aggregates, discover_disks, discover_vservers
    $self->create_accessor_methods(-type => "get_obj",
                                   -pkg  => __PACKAGE__,
                                   -list => \@list_names);

    # for hwresource we don't have a cdef, so we will not be able to use
    # create_accessor_methods of commoninterface.pm
    $hwresource_obj = San::ControllerCluster::Filer10M::Hwresource->new(filer => $self->get_filer());
    if ($hwresource_obj->error()){
        $self->_seterr("failed to create hwresource obj " , $hwresource_obj->errmsg());
        return $self;
    }
    $self->{hwresource_obj} = $hwresource_obj;

    # initialize if it has not been initialized
    unless ($args{initialized}) {
        if ($self->{enable_cache}) {
            $self->refresh();
            if ($self->error()) {
                $self->_set_err("failed to refresh sanfiler object", $self->errmsg());
                return $self;
            }
        }
    }

    return $self;
} ## end sub new

###############################################################################
# Method Name: cluster_join
# Objective:   Tells a node to join a clusterip of another node
# Details:     See POD documentation at the end of this file.
###############################################################################
sub cluster_join {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory clusteripaddr param
    unless (defined $args{clusteripaddr} and length $args{clusteripaddr}) {
        $self->_set_err("mandatory parameter \"-clusteripaddr\" not specified ");
        return $SanUtil::FALSE;
    }

    # build cmd "cluster join"
    %args = (%args, cmd => "cluster_join");

    # invoke cdef command but do not update attributes
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # Error is set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub cluster_join

################################################################################
# Method Name: create_aggregate_obj()
# Objective:   create a new aggregate object
# Details:     See POD documentation at the end of this file
################################################################################
sub create_aggregate_obj {
    my ($self, %args) = @_;

    my $aggr_obj;

    $self->_clear_err();

    # instantiate aggregate object
    $aggr_obj =
      San::ControllerCluster::Filer10M::Aggregate->new(-parent_obj => $self,
                                                       -filer => $self->get_filer(),
                                                       %args);
    if ($aggr_obj->error()) {
        $self->_set_err("instantiate aggr object failed.", $aggr_obj->errmsg());
        return {};
    }

    # is aggr initializing?
    # we're seeing 'aggr create' returned passed; however the aggr show returned nothing
    # meaning aggr create is still in progress, so keep trying
    my $state = $aggr_obj->get_state();
    unless ($state =~ /online/i) {
        # aggr has to zero the disks
        # so await for aggr to online
        $self->wait_for_aggregate_online(-name => $aggr_obj->get_name());
        if ($self->error()) {
            # it's either timed out waiting for aggr to online or
            # no aggr matching the name
            $self->_set_err("still fail after waiting for aggr to online.", $aggr_obj->errmsg());
            return {};
        }
        # we need to refresh the object
        if (! $aggr_obj->refresh()) {
            $self->_set_err("fail to refresh after aggr online.", $aggr_obj->errmsg());
            return {};
        }
    }

    # store aggr object in aggr list
    push @{$self->{aggregate_objs}}, $aggr_obj;

    return $aggr_obj;
} ## end sub create_aggregate_obj

################################################################################
# Method Name: create_vserver_obj()
# Objective:   create a new vserver
# Details:     See POD documentation at the end of this file
################################################################################
sub create_vserver_obj {
    my ($self, %args) = @_;

    my $vserver_obj;

    $self->_clear_err();

    # instantiate vserver object
    $vserver_obj =
      San::ControllerCluster::Filer10M::Vserver->new(-parent_obj => $self,
                                                     -filer => $self->get_filer(),
                                                     %args);

    if ($vserver_obj->error()) {
        $self->_set_err("instantiate vserver object failed.", $vserver_obj->errmsg());
        return {};
    }

    # store vserver object in filer object
    push @{$self->{vserver_objs}}, $vserver_obj;

    return $vserver_obj;
} ## end sub create_vserver_obj

################################################################################
# Method Name: destroy_aggregate_obj()
# Objective:   Destroy aggregate and its object
# Details:     See POD documentation at the end of this file
################################################################################
sub destroy_aggregate_obj {
    my ($self, %args) = @_;

    my ($aggr_obj, @tmp_objs);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert aggregate to name to maintain compatability w/ 7mode
    if (exists $args{aggregate}) {
        $args{name} = delete $args{aggregate};
    }
    # verify mandatory name param
    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return $SanUtil::FALSE;
    }

    # if cache flag is set, get it from cach
    # if it does exist in cache, query the controller for the particular aggregate
    if ($self->{use_cache} && exists $self->{aggregate_objs}) {
        # get aggregate from cache
        ($aggr_obj) = grep ({$_->get_name() =~ /^$args{name}$/} @{$self->{aggregate_objs}});

        # pop matched aggregate object from the list
        @{$self->{aggregate_objs}} = grep ({$_->get_name() !~ /^$args{name}$/} @{$self->{aggregate_objs}});
    } else {
        # discover this particular aggregate on the controller
        # get aggregate objects
        $aggr_obj = $self->get_aggregate_objs(-name => $args{name});
        if ($self->error()) {
            $self->_set_err("failed to get aggregate object: " . $args{name},
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # delete aggregate object that matched and update the list
    if (ref($aggr_obj) =~ /Aggregate/i) {
        $aggr_obj->delete(-parent_obj => $self, -force => 1);
        if ($aggr_obj->error()) {
            $self->_set_err("failed to delete aggregate: " . $args{name},
                            $aggr_obj->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} ## end sub destroy_aggregate_obj

################################################################################
# Method Name: destroy_vserver_obj()
# Objective:   Destroy vserver and its object
# Details:     See POD documentation at the end of this file
################################################################################
sub destroy_vserver_obj {
    my ($self, %args) = @_;

    my ($vserver_obj, @tmp_objs);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert vserver to name to maintain compatability w/ 7mode
    if (exists $args{vserver}) {
        $args{name} = delete $args{vserver};
    }
    # verify mandatory name param
    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return $SanUtil::FALSE;
    }

    # if cache flag is set, get it from cach
    # if it does exist in cache, query the controller for the particular vserver
    if ($self->{use_cache} && exists $self->{vserver_objs}) {
        # get from cache
        ($vserver_obj) = grep ({$_->get_name() =~ /^$args{name}$/} @{$self->{vserver_objs}});

        # pop matched vserver object from the list
        @{$self->{vserver_objs}} = grep ({$_->get_name() !~ /^$args{name}$/} @{$self->{vserver_objs}});
    } else {
        # discover this particular vserver on the controller
        # get vserver objects
        $vserver_obj = $self->get_vserver_objs(-name => $args{name});
        if ($self->error()) {
            $self->_set_err("failed to get vserver object: " . $args{name},
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # delete vserver object that matched and update the list
    if (ref($vserver_obj) =~ /Vserver/i) {
        $vserver_obj->delete(-force => 1);
        if ($vserver_obj->error()) {
            $self->_set_err("failed to delete vserver: " . $args{name},
                            $vserver_obj->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} ## end sub destroy_vserver_obj

################################################################################
# Method Name: get_aggregate_obj()
# Objective:   Get aggregate object.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_aggregate_obj {
    my ($self, %args) = @_;

    my $aggregate_obj = {};

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert aggregate to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{aggregate}) {
        $args{name} = $args{aggregate};
        delete $args{aggregate};
    }

    # verify name param
    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return {};
    }

    # if cache flag is set, try to get it from cache
    # if it does exist in cache, query the controller for the particular aggregate
    if ($self->{use_cache} && exists $self->{aggregate_objs}) {
        # get from cache
        $aggregate_obj = grep ({$_->get_name() =~ /^$args{name}$/} @{$self->{aggregate_objs}});
    } else {
        # discover this particular aggregate on the controller
        # get aggregate objects
        $aggregate_obj = $self->get_aggregate_objs(-name => $args{name});
        if ($self->error()) {
            $self->_set_err("failed to get aggregate object: " . $args{name},
                            $self->errmsg());
            return {};
        }
    }

    unless (ref($aggregate_obj) =~ /Aggregate/i) {
        $self->_set_err("can not find aggregate " . $args{name});
        return {};
    }

    return $aggregate_obj;
} ## end sub get_aggregate_obj

################################################################################
# Method Name: get_vserver_obj()
# Objective:   Get vserver object.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vserver_obj {
    my ($self, %args) = @_;

    my $vserver_obj = {};

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert vserver to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{vserver}) {
        $args{name} = $args{vserver};
        delete $args{vserver};
    }

    if (not exists $args{type}) {
        $args{type} = "cluster|data";
    }

    # verify name param
    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return {};
    }

    # if cache flag is set, get it from cach
    # if it does exist in cache, query the controller for the particular vserver
    # do not use get_vserver_objs because it will discover all vservers in the cluster
    if ($self->{use_cache} && exists $self->{vserver_objs}) {
        # get from cache
        $vserver_obj = grep ({$_->get_name() =~ /^$args{name}$/} @{$self->{vserver_objs}});
    } else {
        # discover this particular vserver on the controller
        # get vserver objects
        $vserver_obj = $self->get_vserver_objs(-name => $args{name},
                                               -type => $args{type});
        if ($self->error()) {
            $self->_set_err("failed to get vserver object: " . $args{name},
                            $self->errmsg());
            return {};
        }
    }

    unless (ref($vserver_obj) =~ /Vserver/i) {
        $self->_set_err("can not find vserver " . $args{name});
        return {};
    }

    return $vserver_obj;
} ## end sub get_vserver_obj

################################################################################
# Method Name: get_epsilon()
# Objective:   Get epsilon
# Details:     See POD documentation at the end of this file
################################################################################
sub get_epsilon {
    my ($self, %args) = @_;

    $args{field} = "epsilon";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_health()
# Objective:   Get health
# Details:     See POD documentation at the end of this file
################################################################################
sub get_health {
    my ($self, %args) = @_;

    $args{field} = "health";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_location()
# Objective:   Get location
# Details:     See POD documentation at the end of this file
################################################################################
sub get_location {
    my ($self, %args) = @_;

    $args{field} = "location";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_model()
# Objective:   Get model
# Details:     See POD documentation at the end of this file
################################################################################
sub get_model {
    my ($self, %args) = @_;

    $args{field} = "model";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_name()
# Objective:   Get filer/cserver name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "name";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_nvramid()
# Objective:   Get nvramid
# Details:     See POD documentation at the end of this file
################################################################################
sub get_nvramid {
    my ($self, %args) = @_;

    $args{field} = "nvramid";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_owner()
# Objective:   Get owner
# Details:     See POD documentation at the end of this file
################################################################################
sub get_owner {
    my ($self, %args) = @_;

    $args{field} = "owner";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_serialnumber()
# Objective:   Get serialnumber
# Details:     See POD documentation at the end of this file
################################################################################
sub get_serialnumber {
    my ($self, %args) = @_;

    $args{field} = "serialnumber";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_systemid()
# Objective:   Get node's system id
# Details:     See POD documentation at the end of this file
################################################################################
sub get_systemid {
    my ($self, %args) = @_;

    $args{field} = "systemid";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_type()
# Objective:   Get filer type
# Details:     See POD documentation at the end of this file
################################################################################
sub get_type {
    my ($self, %args) = @_;

    $args{field} = "type";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_uptime()
# Objective:   Get uptime
# Details:     See POD documentation at the end of this file
################################################################################
sub get_uptime {
    my ($self, %args) = @_;

    $args{field} = "uptime";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_vendor()
# Objective:   Get vendor
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vendor {
    my ($self, %args) = @_;

    $args{field} = "vendor";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_uuid()
# Objective:   Get node uuid
# Details:     See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self, %args) = @_;

    $args{field} = "uuid";
    $args{command} = "cluster_show";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method Name: modify
# Objective:   Modifies a node attribute
# Details:     See POD documentation at the end of this file.
###############################################################################
sub modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # initialize cmd parameters
    %args = (%args,
             cmd => "system_node_modify",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # Need to refresh (which calls net int show) to cache the new values,
    # otherwise, if a user calls an accessor method (i.e. get_admin_status), it
    # will get the previous value.
    $self->refresh();
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub modify

###############################################################################
# Method Name: modify_ipv6
# Objective:   Modifies ipv6 setting
# Details:     See POD documentation at the end of this file.
###############################################################################
sub modify_ipv6 {
    my ($self, %args) = @_;

    $self->_clear_err();

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_options_ipv6_modify");

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub modify_ipv6

################################################################################
# Method Name: aggregate_exist()
# Objective:   Determines if aggregate exists.
# Details:     See POD documentation at the end of this file
################################################################################
sub aggregate_exist {
    my ($self, %args) = @_;

    my (@output, %tmp_args);

    $self->_clear_err();

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|aggregate)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # the key is a scalar
            $tmp_args{aggregate} = $key;
        }
    }

    # set args to %tmp_args
    %args = %tmp_args;

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to aggregate for cdef syntax
    if (exists $args{name}) {
        $args{aggregate} = $args{name};
        delete $args{name};
    }

    # verify mandatory name param
    unless (defined $args{aggregate} and length $args{aggregate}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return $SanUtil::FALSE;
    }

    # build cmd vserver show
    %args = (%args, cmd => "storage_aggregate_show");

    # invoke cdef command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # Check to see if we failed because the aggregate does not exist
        my @errmsg = $self->errmsg();
        my $match  = grep(/does\s+not\s+exist/i, @errmsg);
        if ($match){
            # Clear the error since not existing isn't an error
            $self->_clear_err();
            return $SanUtil::FALSE;
        }

        # If we get here, we got a failure other than one caused by the
        # aggregate not existing, set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub aggregate_exist

################################################################################
# Method Name: vserver_exist()
# Objective:   Determines if vserver exists.
# Details:     See POD documentation at the end of this file
################################################################################
sub vserver_exist {
    my ($self, %args) = @_;

    my (@output, %tmp_args);

    $self->_clear_err();

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|vserver)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # the key was a scalar
            $tmp_args{vserver} = $key;
        }
        last;
    }
    # set tmp_args to args
    %args = %tmp_args;

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to vserver for cdef syntax
    if (exists $args{name}) {
        $args{vserver} = $args{name};
        delete $args{name};
    }

    # verify name param
    unless (defined $args{vserver} and length $args{vserver}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return $SanUtil::FALSE;
    }

    # build cmd vserver show
    %args = (%args, cmd => "vserver_show");

    # invoke cdef command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub vserver_exist

###############################################################################
# Method Name: refresh()
# Objective:   Refresh object's data for Filer10M
# Details:     See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self) = @_;

    # initialize cmd arguments for system node show
    my (%args) = (cmd => "system_node_show",
                  node => $self->get_node());

    # hand of responsibility to base class to refresh aggr object
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        $self->_set_err("failed to refresh Filer10M object data", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: reboot()
# Objective:   Reboots a filer.
# Details:     See POD documentation at the end of this file.
###############################################################################
sub reboot {
    my ($self) = @_;

    $self->halt_node();
    if ($self->error()) {
        $self->_set_err("failed to halt the node", $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->boot_node();
    if ($self->error()) {
        $self->_set_err("failed to boot the node", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Restore the connection before returning the object to the caller
    $self->reconnect();
    if ($self->error()) {
        $self->_set_err("Failed to restore connection after reboot.",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub reboot

###############################################################################
# Method:    sfo_show()
# Objective: Run the storage failover show cdef
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_show {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # if this is not a cserver server type then default some params
    if ($self->{server_type} ne "cserver") {
        # default the node argument to be the current node
        unless (defined $args{node} and length $args{node}) {
            $args{node} = $self->get_name();
        }
    }

    # build cmd sfo show
    %args = (%args, cmd => "storage_failover_show");

    # invoke cdef command but do not update attributes
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    sfo_show_giveback()
# Objective: Run the storage failover show giveback cdef
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_show_giveback {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # default the node argument to be the current node
    unless (defined $args{node} and length $args{node}) {
        $args{node} = $self->get_name();
    }

    # build cmd sfo show giveback cdef
    %args = (%args, cmd => "storage_failover_show_giveback");

    # invoke cdef command but do not update attributes
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    sfo_modify()
# Objective: run the storage failover modify cdef
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # default the node argument to be the current node
    unless (defined $args{node} and length $args{node}) {
        $args{node} = $self->get_node();
    }

    # build cmd sfo modify
    %args = (%args, cmd => "storage_failover_modify");

    # invoke cdef command but do not update attributes
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    sfo_takeover()
# Objective: run the storage failover takeover cdef
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_takeover {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # default the bynode argument to be the current node
    unless (defined $args{bynode} and length $args{bynode}) {
        $args{bynode} = $self->get_name();
    }

    # build cmd sfo takeover
    %args = (%args, cmd => "storage_failover_takeover");

    # invoke cdef command but do not update attributes
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    sfo_giveback()
# Objective: run the storage failover giveback cdef
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_giveback {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # default the fromnode argument to the current node
    unless (defined $args{fromnode} and length $args{fromnode}) {
        $args{fromnode} = $self->get_name();
    }

    # build cmd sfo giveback
    %args = (%args, cmd => "storage_failover_giveback");

    # invoke cdef command but do not update attributes
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    sfo_mode()
# Objective: run the storage modify MODE.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub sfo_mode {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # Get the needed arguments
    my $mode = ( defined($args{'mode'}) ) ? $args{'mode'} : "ha";
    my $node = ( defined($args{'node'}) ) ? $args{'node'} : $self->get_name();

    # Build up the command.
    my %command = ('cmd' => "run -node $node -command options cf.mode $mode");

    # Execute command
    my (@output) = $self->execute(%command);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("Failed to EXECUTE \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    statistics_show_periodic()
# Objective: Run the statistics show periodic cdef
# Details:   See POD documentation at the end of this file.
###############################################################################
sub statistics_show_periodic {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # build cmd statistics show periodic giveback cdef
    %args = (%args, cmd => "statistics_show_periodic");

    # invoke cdef command but do not update attributes.  We are turning off
    # the check for required options because the cdef is incorrectly requiring
    # options that can be defaulted.
    my (@output) = $self->SUPER::invoke_cmd_obj(%args,
                                                -update_attrs    => 0,
                                                -require_options => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    system_node_reboot()
# Objective: run the system node reboot cdef
# Details:   See POD documentation at the end of this file.
###############################################################################
sub system_node_reboot {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # default the node argument to the current node
    unless (defined $args{node} and length $args{node}) {
        $args{node} = $self->get_node();
    }

    # build cmd to reboot specified node
    %args = (%args, cmd => "system_node_reboot");

    # invoke cdef command but do not update attributes
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    get_hwresource_obj()
# Objective: Get hw resource object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_hwresource_obj{
    my ($self) = @_;

    # the get() will handle error
    return $self->get("hwresource_obj");
}

###############################################################################
# Method:    autosupport_modify()
# Objective: Modify autosupport settings
# Details:   See POD documentation at the end of this file.
###############################################################################
sub autosupport_modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # default the node argument to be the current node
    unless (defined $args{node} and length $args{node}) {
        $args{node} = $self->get_name();
    }

    # build cmd sfo show giveback cdef
    %args = (%args, cmd => "system_node_autosupport_modify");

    # invoke cdef command but do not update attributes
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}

##############################################################################
#  DYNAMICALLY CREATED METHODS
##############################################################################

################################################################################
# Method Name: get_aggregate_objs()
# Objective:   Get aggregate objects. This method is created dynamically from
#              create_accessor_methods in the new() constructor. This is a
#              placeholder in case someone is looking for the method.
# Details:     See POD documentation at the end of this file for usage
################################################################################

################################################################################
# Method Name: get_disk_objs()
# Objective:   Get disk objects. This method is created dynamically from
#              create_accessor_methods in the new() constructor. This is a
#              placeholder in case someone is looking for the method.
# Details:     See POD documentation at the end of this file for usage
################################################################################

################################################################################
# Method Name: get_vserver_objs()
# Objective:   Get vserver objects. This method is created dynamically from
#              create_accessor_methods in the new() constructor. This is a
#              placeholder in case someone is looking for the method.
# Details:     See POD documentation at the end of this file for usage
################################################################################

##############################################################################
#  PRIVATE METHODS
##############################################################################

################################################################################
# Method Name: _discover
# Objective:   refresh and update cache of node objects
# Details:     See POD documentation at the end of this file
################################################################################
sub _discover {
    my ($self, %args) = @_;

    my ($filer_objs, @objs, @records, $parent_obj);

    # parameter initialization
    #$self = {};
    #bless $self, $class;

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);
    eval {

        # validate input parameters and their type.
        # filer - filer name
        (%args) =
          Module::Common::validate_mixed(@input_params,
                                         [],
                                         {filer => {type => SCALAR, required => 1},
                                          name => {type => SCALAR, optional => 1},
                                          buildinfo => {type => SCALAR | ARRAYREF, optional => 1},
                                          conn => {type => OBJECT, optional => 1},
                                          timeout => {type => SCALAR,
                                                      optional => 1,
                                                      default => 1800},
                                          debug_level => {type => SCALAR,
                                                          optional => 1 }});
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return {};
    }

    # discover Filer10M object via system node show
    # return arrayref of filers it has discovered
    $filer_objs = $self->SUPER::discover(-pkg  => __PACKAGE__,
                                         -cdef => "system_node_show",
                                         %args);
    if ($self->error()) {
        $self->_set_err("failed to discover filer object", $self->errmsg());
        return {};
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # let's try to obtain filer object that matches with filer name
    # if specified
    if (defined($args{filer})) {
        foreach my $obj (@{$filer_objs}) {
            my $name = $obj->get_name();
            # return specific filer object if requested criteria matched
            #
            # we need to return singular to be compatible with 7-mode
            if ($name =~ /^$args{filer}$/) {
                #return (wantarray()) ? ($obj) : [$obj];
                return $obj;
            }
        }

        # if we get here, we can't find the object specific to filer name
        $self->_set_err("no object found for '" . $args{filer} . "'");
        return {};
    }

} ## end sub _discover

################################################################################
# Method Name: _get_param_defaults
# Objective:   Gets default node parameters
# Details:     See POD documentation at the end of this file.
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    my %defaults_params = (node => $self->get_name(-use_cache => 1));

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Lif
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    my $use_cache =  $SanUtil::FALSE;
    my $command = (defined $args{command}) ? delete $args{command} : "system_node_show";

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # convert common args to args specifically for creating lif
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => $command,
                                                $self->_get_param_defaults(),
                                                %args);
    }
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Filer10M.pm

=head1 Author

Korry Nguyen (dl-ssan-automation)

=head1 Description

Provides an interface to manage a C-mode filer

=head1 See Also

Filer10M package used base class B<San::CommonInterface>, B<San::ControllerCluster::FilerHelper> and B<San::ControllerCluster::FilerCompat> that also inherit other modules. See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm, San::SanCache.pm>

=head1 Attributes

=head2 B<filer>

string = controller name

=head2 B<name>

string = name of controller object to be created

=head2 B<debug_level>

interger = 0 - 100 (default is 0)

=head2 B<timeout>

interger = seconds (default is 1800)

=head1 B<Public Methods>

=head2 B<new(%args)>

Create a controller object

=over 4

=item Inputs

=item * I<-filer> => <controller name>

Controller name

=item * I<-name> => <name of controller object>

Controller object name to be created with

=item * I<-server_type> => <cserver or node>

Specifies what kind of server this is 'cserver' or 'node', defaults to 'node'

=item * I<-timeout> => <number of seconds>

Times that take the command to timeout

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. With higher level, it will log more debug messages

=item Outputs

Return Filer10M object

=item Examples

 use San::SanFiler;

 my $filer = "ssan-3170";

 my $filer_obj = new SanFiler(-filer => $filer,
                              -timeout => 180,
                              -debug_level => 0,
                             );

 if ($filer_obj->error) {
     $err = $filer_obj->errmsg();
     logcomment("Failed to create a SAN Filer object: $err");
     return TCD::FAIL;
 }

=back

=head2 B<cluster_join(%args)>

Join the San::ControllerCluster::Filer10M to specified cluster IP of node
    from the cluster.

=over 4

=item Inputs

=item * I<-clusteripaddr> => <cluster IP address of cluster node>

This should be the IP address of a network interface of role "cluster" that is
already in the cluster.

=item Outputs

Returns TRUE if the node was joined successfully, FALSE and set error otherwise.

=item Examples

    $node_obj->cluster_join(-clusteripaddr => $clusterjoinip);
    if ($node_obj->error()) {
        $Test->nlog("Fail to join " . $node_obj->get_name()
                    . $node_obj->errmsg());
        return $SanUtil::FALSE;
    }

=back

=head2 B<create_aggregate_obj(%args)>

Create an aggregate object. After the object is created, it will be stored in Filer10M object.

=over 4

=item Inputs

=item * I<-name | -aggregate> (required)

The aggregate name to be used

=item * I<-disk_objs> (required either diskcount or disk_objs)

Arrayref of a list of spare disk objects

=item * I<-diskcount> (required either diskcount or disk_objs)

Number of disks

=item * I<-raidtype> (default: raid_dp)

The raidtype of the aggregate, can be raid4 or raid_dp

=item * I<-maxraidsize> (default: default ONTAP raidsize)

The raidsize of the aggregate, can be between 2 and 28

=item Outputs

Returns aggregate object. If error, error message will be pushed onto error stack.

=item Examples

 # ex. 1: Create aggregate with diskcount
 $filer_obj->create_aggregate_obj(-name => test_aggr,
                                  -diskcount => 5);
 if ($filer_obj->error()){
     ...
 }

 # ex. 2: Create aggregate with disk_objs

 # get 5 spare disk objects
 @disk_objs = $filer_obj->get_disk_objs(-state => 'spare',
                                        # disks owned by this filer
                                        -owner => $self->get_node(),
                                        -count => 5);
 if ($filer_obj->error()){
     ...
 }

 # Create aggregate
 $filer_obj->create_aggregate_obj(-name => 'test_aggr',
                                  -disk_obj => [@disk_objs]);
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<create_vserver_obj(%args)>

Create a vserver object. After the object is created, it will be stored in Filer10M object.

=over 4

=item Inputs

=item * I<-name | -vserver> (required)

The vserver name to be used

=item * I<-aggregate> (required)

The aggregate name to put vserver rootvolume

=item * I<-rootvolume> (optional: default is $vserver_name . _root)

The vserver root volume name

=item Outputs

Returns vserver object. If error, error message will be pushed onto error stack.

=item Examples

 # ex. 1: Create vserver with default rootvolume
 $filer_obj->create_vserver_obj(-name => $vs_name,
                                -aggregate => $aggr_name);
 if ($filer_obj->error()){
     ...
 }

 # ex. 2: Create vserver with a specific rootvolume name
 $filer_obj->create_vserver_obj(-name => 'my_vs',
                                -aggregate => $aggr_name,
                                -rootvolume => 'rootvoume_for_my_vs');
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<destroy_aggregate_obj(%args)>

Destroy Aggregate and the object.

=over 4

=item Inputs

=item * I<-name | -aggregate> (required)

The aggregate name to destroy

=item * I<-force> (optional: default is true

Flag to force destroy the aggregate

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Destroy an aggregate and its object
 $filer_obj->destroy_aggregate_obj(-name => $aggr_name);
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<destroy_vserver_obj(%args)>

Destroy vserver and the object.

=over 4

=item Inputs

=item * I<-name | -vserver> (required)

The vserver name to destroy

=item * I<-force> (optional: default is true

Flag to force destroy the vserver

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Destroy an vserver and its object
 $filer_obj->destroy_vserver_obj(-name => $vs_name);
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<get_aggregate_obj(%args)>

Get an aggregate object

=over 4

=item Inputs

=item * I<-name | -aggregate> (required)

The aggregate name to retrieve

=item Outputs

Returns aggregate object. If not exist, set error and returns empty hashref

=item Examples

 # ex. 1: Get aggregate object
 $aggr_obj = $filer_obj->get_aggregate_obj(-name => 'test_aggr');
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<get_aggregate_objs(%args)>

Get multiple aggregate objects

=over 4

=item Inputs

=item * I<-name | -aggregate> (optional)

The regex matching of aggregate name

=item * I<-voltype> (optional)

The volume type of the aggregate

=item * I<-count> (optional)

Number of aggregates

=item Outputs

Returns array of aggregate objects that meet the criteria. If not met, set error and returns empty array

=item Examples

 # ex. 1: Get all aggregate objects
 @aggr_objs = $filer_obj->get_aggregate_objs();
 if ($filer_obj->error()){
     ...
 }

 # ex. 2: Get aggregate objects with the regex matching name
 @aggr_objs = $filer_obj->get_aggregate_objs(-name => "^$aggr_name");
 if ($filer_obj->error()){
     ...
 }

 # ex. 3: Get exactly 2 aggregate objects with matching name and voltype
 @aggr_objs = $filer_obj->get_aggregate_objs(-name => "^$aggr_name",
                                             -voltype => "flex",
                                             -count => 2);
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<get_disk_objs(%args)>

Get multiple disk objects

=over 4

=item Inputs

=item * I<-name | -disk> (optional)

The regex matching of disk name

=item * I<-state> (optional)

The state of the disks

=item * I<-owner> (optional)

The owner of the disks

=item * I<-count> (optional)

Number of disks

=item Outputs

Returns array of disk objects that meet the criteria. If not met, set error and returns empty array

=item Examples

 # ex. 1: Get all disk objects
 @disk_objs = $filer_obj->get_disk_objs();
 if ($filer_obj->error()){
     ...
 }

 # ex. 2: Get all spare disk objects
 @disk_objs = $filer_obj->get_disk_objs(-state => "spare");
 if ($filer_obj->error()){
     ...
 }

 # ex. 3: Get 5 spare disk objects with the same owner
 @disk_objs = $filer_obj->get_disk_objs(-state => "spare",
                                        -owner => $filer_name,
                                        -count => 5);
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<get_vserver_obj(%args)>

Get an vserver object

=over 4

=item Inputs

=item * I<-name | -vserver> (required)

The vserver name to retrieve

=item Outputs

Returns vserver object. If not exist, set error and returns empty hashref

=item Examples

 # ex. 1: Get vserver object
 $aggr_obj = $filer_obj->get_vserver_obj(-name => 'my_vserver');
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<get_vserver_objs(%args)>

Get multiple vserver objects

=over 4

=item Inputs

=item * I<-name | -vserver> (optional)

The regex matching of vserver name

=item * I<-voltype> (optional)

The volume type of the vserver

=item * I<-count> (optional)

Number of vservers

=item Outputs

Returns array of vserver objects that meet the criteria. If not met, set error and returns empty array

=item Examples

 # ex. 1: Get all vserver objects
 @aggr_objs = $filer_obj->get_vserver_objs();
 if ($filer_obj->error()){
     ...
 }

 # ex. 2: Get vserver objects with the regex matching name
 @vs_objs = $filer_obj->get_vserver_objs(-name => "^$vs_name");
 if ($filer_obj->error()){
     ...
 }

 # ex. 2: Get exactly 2 vserver objects with the regex matching name
 @vs_objs = $filer_obj->get_vserver_objs(-name => "^$vs_name",
                                           -count => 2);
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<get_epsilon()>

Get epsilon attribute for the Controller

=over 4

=item Inputs

None

=item Outputs

Returns epsilon

=item Examples

 # ex. 1: Detect if the node holds epsilon
    if($node_obj->get_epsilon() =~ /true/){
        $Test->nlog("Controller has epsilon.");
    }

=back

=head2 B<get_health()>

Get health of the controller

=over 4

=item Inputs

None

=item Outputs

Returns string "true" or "false"

=item Examples

 # ex. 1: Check controller health
 if ($filer_obj->get_health() =~ /true/) {
     # controller is in good health
 } else {
     # controller is in bad health
 }

=back

=head2 B<get_hwresource_obj()>

Get hardware resource object

=over 4

=item Inputs

None

=item Outputs

Returns hwresource object.

=item Examples

 # ex. 1: Get hardware resource obj
 $hw_obj =  $filer_obj->get_hwresource_obj();
 if ($filer_obj->error()) {
     ...
 }

=back

=head2 B<get_location()>

Get location of the controller

=over 4

=item Inputs

None

=item Outputs

Returns location where controller resided

=item Examples

 # ex. 1: Display location
 logcomment("Controller location: " . $filer_obj->get_location());

=back

=head2 B<get_model()>

Get controller model

=over 4

=item Inputs

None

=item Outputs

Returns model number

=item Examples

 # ex. 1: Display model
 logcomment("Controller model: " . $filer_obj->get_model());

=back

=head2 B<get_name()>

Get controller/cserver name

=over 4

=item Inputs

None

=item Outputs

Returns controller/cserver name

=item Examples

 # ex. 1: Display controller/cserver name
 logcomment("Controller name: " . $filer_obj->get_name());

=back

=head2 B<get_node()>

Get node name

=over 4

=item Inputs

None

=item Outputs

Returns controller name

=item Examples

 # ex. 1: Display controller name
 logcomment("Controller name: " . $filer_obj->get_name());

=back

=head2 B<get_nvramid()>

Get controller nvramid

=over 4

=item Inputs

None

=item Outputs

Returns nvramid

=item Examples

 # ex. 1: Display nvramid
 logcomment("Controller nvramid: " . $filer_obj->get_nvramid());

=back

=head2 B<get_owner()>

Get controller owner

=over 4

=item Inputs

None

=item Outputs

Returns owner number

=item Examples

 # ex. 1: Display owner
 logcomment("Controller owner: " . $filer_obj->get_owner());

=back

=head2 B<get_serialnumber()>

Get controller serialnumber

=over 4

=item Inputs

None

=item Outputs

Returns serialnumber

=item Examples

 # ex. 1: Display serialnumber
 logcomment("Controller serialnumber: " . $filer_obj->get_serialnumber());

=back

=head2 B<get_systemid()>

Get controller systemid

=over 4

=item Inputs

None

=item Outputs

Returns systemid number

=item Examples

 # ex. 1: Display systemid
 logcomment("Controller systemid: " . $filer_obj->get_systemid());

=back

=head2 B<get_type()>

Get controller type

=over 4

=item Inputs

None

=item Outputs

Returns type

=item Examples

 # ex. 1: Display type
 logcomment("Controller type: " . $filer_obj->get_type());

=back

=head2 B<get_uptime()>

Get controller uptime

=over 4

=item Inputs

None

=item Outputs

Returns uptime

=item Examples

 # ex. 1: Display uptime
 logcomment("Controller uptime: " . $filer_obj->get_uptime());

=back

=head2 B<get_uuid()>

Get controller uuid

=over 4

=item Inputs

None

=item Outputs

Returns controller uuid

=item Examples

 # ex. 1: Display controller uuid
 logcomment("Controller uuid: " . $filer_obj->get_uuid());

=back

=head2 B<modify()>

Modify attributes of the controller "system node modify -<attribute> <value>"

=over 4

=item Inputs

  [[-owner] <text>]              Owner
  [ -location <text> ]           Location
  [ -assettag <text> ]           Asset Tag
  [ -eligibility {true|false} ]  Eligibility
  [ -epsilon {true|false} ]      *Epsilon

=item Outputs

None

=item Examples

 # ex. 1: Modify epsilon to false
    if($node_obj->get_epsilon() =~ /true/){
        $Test->nlog("Controller has epsilon.");
        $node_obj->modify(-epsilon =>"false");
        if ($node_obj->error()){
            $Test->nlog(" - Error modifying epsilon on Controller" .
                        $node_obj->errmsg());
            return $SanUtil::FAIL;
        } else {
            $Test->nlog(" - Modified epsilon to \"false\" on Controller");
        }
    }

=back

=head2 B<get_vendor()>

Get controller vendor

=over 4

=item Inputs

None

=item Outputs

Returns vendor

=item Examples

 # ex. 1: Display vendor
 logcomment("Controller vendor: " . $filer_obj->get_vendor());

=back

=head2 B<aggregate_exist(%args)>

Determine whether aggregate exists or not.

=over 4

=item Inputs

=item * I<-name | -aggregate> (required)

=item Outputs

Returns TRUE if exist, otherwise FALSE

=item Examples

 # ex. 1: Check whether aggregate exist or not
 if ($filer_obj->aggregate_exist(-name => 'test_aggr')) {
     $aggr_obj = $filer_obj->get_aggregate_obj(-name => 'test_aggr');
     if ($filer_obj->error()){
         ...
     }
 } else {
     $aggr_obj = $filer_obj->create_aggregate_obj(-name => 'test_aggr',
                                                  -diskcount => 3);
     if ($filer_obj->error()){
         ...
     }
 }

=back

=head2 B<vserver_exist(%args)>

Determine whether vserver exists or not.

=over 4

=item Inputs

=item * I<-name | -vserver> (required)

=item Outputs

Returns TRUE if exist, otherwise FALSE

=item Examples

 # ex. 1: Check whether vserver exist or not
 if ($filer_obj->vserver_exist(-name => 'test_aggr')) {
     $aggr_obj = $filer_obj->get_vserver_obj(-name => $vs_name);
     if ($filer_obj->error()){
         ...
     }
 } else {
     $aggr_obj = $filer_obj->create_vserver_obj(-name => $vs_name,
                                                -aggregate => $aggr_name);
     if ($filer_obj->error()){
         ...
     }
 }

=back

=head2 B<refresh()>

Refresh and update Filer10M object

=over 4

=item Inputs

None

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Refresh and update filer object
 $filer_obj->refresh();
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<reboot()>

Reboots a c-mode filer. This method requires a valid tharnhost file, including
console connection parameters.

=over 4

=item Inputs

=item * None

=item Outputs

Returns true if the reboot was successful, false otherwise.

=item Examples

 $filer_obj->reboot();

 if ($filer_obj->error) {
     $err = $filer_obj->errmsg();
     logcomment("Failed to reboot the filer: $err");
     return TCD::FAIL;
 }

=back

=head2 B<sfo_show(%args)>

List out storage failover properties

=over 4

=item Inputs

=item * node

Optional: The node to display sfo properties for.
Default: The current node.

=item Outputs

Returns an array or array reference of the cdef output of
storage_failover_show.

=item Examples

 my ($output) = $filer_obj->sfo_show(-fields => "state,possible");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

 if ($output->{state}    eq "takeover" and
     $output->{possible} eq "false") {
     ...
 }

=back

=head2 B<sfo_show_giveback(%args)>

Show storage failover giveback status

=over 4

=item Inputs

=item * node

Optional: The node to display sfo giveback  properties for.
Default: The current node.

=item Outputs

Returns an array or array reference of the cdef output of
storage_failover_show_giveback

=item Examples

 my ($output) = $filer_obj->sfo_show_giveback(-fields => "giveback-status,aggregate,destination");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

 if ($output->{giveback_status} =~ /No aggregates to give back/i) {
     ...
 }

=back

=head2 B<sfo_modify(%args)>

Make modifications to the SFO properties of a specific node

=over 4

=item Inputs

=item * node

Optional: The node to modify SFO properties for.
Default: The current node.

=item Outputs

Returns an array or array reference of the cdef output of
storage_failover_modify.

=item Examples

 # enable takeover onpanic/onreboot on the node "mypartnername"
 $filer_obj->sfo_modify(-node     => "mypartnername",
                        -onpanic  => "true",
                        -onreboot => "true");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

=back

=head2 B<sfo_takeover(%args)>

Issue a storage failover takeover on the specified node.

=over 4

=item Inputs

=item * node

Optional: The node to issue the sfo takeover on.
Default: The current node.

=item Outputs

Returns $SanUtil::TRUE if successful, $SanUtil::FALSE and sets error if not.

=item Examples

 # Issue a forced takeover on this node
 $filer_obj->sfo_takeover(-option => "immediate");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

=back

=head2 B<sfo_giveback(%args)>

Issue a storage failover giveback on the specified node.

=over 4

=item Inputs

=item * node

Optional: The node to issue the sfo giveback on.
Default: The current node.

=item Outputs

Returns $SanUtil::TRUE if successful, $SanUtil::FALSE and sets error if not.

=item Examples

 # Issue a forced giveback on this node,
 $filer_obj->sfo_giveback('-override-vetoes' => "true",
                          '-require-partner-waiting' => "false");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

=back

=head2 B<sfo_mode(%args)>

Enable the storage administrator to configure HA, Standalone.
IMPORTANT: This method is only needed for Sierra Nevada nodes. It is the user's
responsibility to know the build that the nodes are running.

=over 4

=item Inputs

=item * Mode

This is the mode that the HA registry wants to be set to. It can be set to HA or
to StandAlone (non_ha).
IMPORTANT: This is a SierraNevada feature. For mor information please refere to

http://wikid.netapp.com/w/HANLR/FS

Optional: The mode to which the  SFO properties will be modified.
Default: HA.


=item * node

Optional: The node to modify SFO properties for.
Default: The current node.

=item Outputs

Boolean (true or false)

=item Examples

 # Enable HA
 $filer_obj->sfo_mode(-mode => "ha");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

=back

=head2 B<statistics_show_periodic(%args)>

Run the statistics show-periodic command and return the output

=over 4

=item Inputs

=item Outputs

Returns an array or array reference of the cdef output of
statics_show_periodic.

=item Examples

 my ($output) = $filer_obj->statics_show_periodic(-iterations => "10");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

=back

=head2 B<system_node_reboot(%args)>

Issue a reboot on the specified node, or current node.

=over 4

=item Inputs

=item * node

Optional: The node to issue the reboot on.
Default: The current node.

=item Outputs

Returns $SanUtil::TRUE if successful, $SanUtil::FALSE and sets error if not.

=item Examples

 # Issue a system node reboot on the partner node
 $filer_obj->system_node_reboot(-node => "mypartner");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

=head2 B<modify_ipv6_enabled>(%args)>

Modifies ipv6 options on the current cluster.

=over 4

=item Inputs

=item enabled

Required: The whether ipv6 should be enabled. It can
currently only be enabled and not disabled

=item Outputs

Returns $SanUtil::TRUE if successful, $SanUtil::FALSE and sets error if not.

=item Examples

 # Issue a system node reboot on the partner node
 $filer_obj->modify_ipv6(-enabled => "true");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

=back

=head2 B<autosupport_modify(%args)>

Make modifications to autosupport settings

=over 4

=item Inputs

=item * node

Optional: The node or nodes to modify
Default: The current node.

=item Outputs

None

=item Examples

 # disable auto support on every node
 $filer_obj->sfo_modify(-node     => "*",
                        -state    => "disable",
                        -support  => "disable");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

=back

=cut
