# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to an Aggregate object
##
## @author korry@netapp.com, dl-ssan-test
## @status Shared

package San::ControllerCluster::Filer10M::Aggregate;

# compiler directives
use strict;

# imported modules
use Scalar::Util 'refaddr';

# inherit base class
use base qw( San::CommonInterface );

###############################################################################
# Method Name: new
# Objective:   Create new Aggregate object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }

    # cdef takes aggregate key instead of name
    if (exists $args{name}) {
        $args{aggregate} = delete $args{name};
    }
    # cdef takes nodes key instead of disk_owner
    if (exists $args{disk_owner}) {
        $args{nodes} = delete $args{disk_owner};
    }

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct aggregate object
    $self = $class->SUPER::_new(-name  => $args{aggregate},
                                -filer => $args{filer},
                                %args);
    if ($self->error()) {
        $self->_set_err("failed to contruct aggregate object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # store Aggregate elements on a list
    foreach my $key (sort keys %args) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # parent argument has been stored, delete unused parent argument
    delete $args{parent_obj};

    # if the obj hs been initialized, just refresh to update info
    # otherwise initialize it by create and update its object
    unless ($args{initialized}) {
        $self->_create_aggr(%args);
        if ($self->error()) {
            $self->_set_err("failed to create aggregate object", $self->errmsg());
            return $self;
        }
    }

    return $self;
} ## end sub new

###############################################################################
# Method Name: delete
# Objective:   delete aggregate object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    my ($parent_obj, $force, @vs_objs);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # we want aggr delete to cleanup any element (vserver, volume, etc) that
    # associated/contained in aggregate
    # process arguments
    while (my ($key, $value) =  each(%args)) {
        if ($key =~ /-?parent_obj$/) {
            if (! length($parent_obj = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $SanUtil::FALSE;
            }
            delete $args{parent_obj};
        } elsif ($key =~ /-?force$/) {
            if (! length($force = $value)) {
                $self->_set_err("invalid value for \"-$key\": " . $value);
                return $SanUtil::FALSE;
            }
            delete $args{force};
        }
    }

    if (defined ($force) and $force == $SanUtil::TRUE) {
        # if we have parent object, let's try to cleanup prior to deleteing aggr.
        # otherwise, assume aggregate does not contain anthing, delete it.
        if (defined $parent_obj and ref($parent_obj) =~ /Filer10M/i) {
            # in order to delete aggr, volume and vserver contained in aggr
            # must be deleted
            # since we don't know which vserver contained in the aggr
            # we have to get all vservers
            @vs_objs = $parent_obj->get_vserver_objs();
            if ($parent_obj->error()) {
                $self->_set_err("failed to get vserver objects", $parent_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # find vservers that match with aggregate and cleanup its content
            foreach my $vs_obj (@vs_objs) {

                # if vserver contained in the aggr to be deleted matches
                # force delete vserver. it will delete everything in the vserver
                if ($vs_obj->get_aggregate() eq $self->get_name()) {

                    # force delete vserver
                    $vs_obj->delete(-force => $SanUtil::TRUE);
                    if ($vs_obj->error()) {
                        $self->_set_err("failed to delete vserver", $vs_obj->errmsg());
                        return $SanUtil::FALSE;
                    }
                }
            }
        } else {
            $self->_set_err("need the parent object to force delete aggregate");
            return $SanUtil::FALSE;
        }
    }

    # get nodes this aggregate was created on so we can zero the disks
    # after the aggregate is deleted
    my $node_list = $self->get_nodes();
    if ($self->error()) {
        $self->_set_err("failed to get node list of this aggregate created on", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # initialize cmd arguments for aggregate delete
    %args = (%args,
             cmd => "storage_aggregate_delete",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to delete an Aggregate", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # zero spare disks after aggregate deleted in the background
    $self->zero_spare_disks(nodes => $node_list);
    if ($self->error()) {
        # bubble up the error
        return $SanUtil::FALSE;
    }

    # NOTE: disks are still zeroing at this time so refreshing disks won't have
    # the correct information. Perhaps setting a dirty flag to indicate the
	# disk is currently zeroing. RFE to mark disk zeroed or zeroing for
    # optimization

    # refresh disk objects to obtain their curent state if exist
    if (defined($self->{disk_objs})) {
        foreach my $disk_obj (@{$self->{disk_objs}}) {
            $self->sanlog(-msg_level => 3,
                          -msg       => $disk_obj->get_name() . " is " . refaddr($disk_obj));
            $disk_obj->refresh();
            if ($disk_obj->error()) {
                $self->_set_err("failed to refresh object's data: " . refaddr($disk_obj), $disk_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }
    }

    return $SanUtil::TRUE;
} ## end sub delete

###############################################################################
# Method Name: modify
# Objective: modify aggregate attribute
# Details:   See POD documentation at the end of this file.
###############################################################################
sub modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "storage_aggregate_modify",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub modify

###############################################################################
# Method Name: wafliron_start
# Objective: Start wafliron on the specified aggregate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wafliron_start {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "storage_aggregate_wafliron_start",
             $self->_get_param_defaults());

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain igroup object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub wafliron_start

###############################################################################
# Method Name: wafliron_stop
# Objective: Stop wafliron on the specified aggregate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wafliron_stop {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "storage_aggregate_wafliron_stop",
             $self->_get_param_defaults());

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain igroup object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub wafliron_stop

###############################################################################
# Method Name: wafliron_show()
# Objective: Display wafliron status on an aggregate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wafliron_show {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "storage_aggregate_wafliron_show");

    # by default this method will show a single aggregate's wafliron status
    # show_all can be optionally passed in, so that all wafliron aggregates can be displayed in the same command
    my $show_all = delete($args{show_all});
    if ($show_all == $SanUtil::FALSE) {
        %args = (%args, $self->_get_param_defaults());
    }

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain igroup object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub wafliron_show

###############################################################################
# Method Name: wafliron_review()
# Objective: Display wafliron review on an aggregate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wafliron_review {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "storage_aggregate_wafliron_review",
             $self->_get_param_defaults());

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain igroup object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

} ## end sub wafliron_review

###############################################################################
# Method Name: wafliron_commit()
# Objective: Commit wafliron changes on an aggregate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wafliron_commit {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "storage_aggregate_wafliron_commit",
             $self->_get_param_defaults());

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain igroup object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

} ## end sub wafliron_commit

###############################################################################
# Method Name: wafliron_reject()
# Objective: Reject wafliron changes on an aggregate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wafliron_reject {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "storage_aggregate_wafliron_reject",
             $self->_get_param_defaults());

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain igroup object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

} ## end sub wafliron_reject

###############################################################################
# Method Name: get_list_info()
# Objective: Get aggregate information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "storage_aggregate_show",
             $self->_get_param_defaults());

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain igroup object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return {};
    }

    unless (scalar(@output)) {
        $self->_set_err("'" . $self->get_name() . "' does not exist");
        return {};
    }

    # getting object's attributes list so it will always return one record
    return $output[0];

} ## end sub get_list_info

################################################################################
# Method Name: get_availsize()
# Objective:   Get aggregate availsize
# Details:     See POD documentation at the end of this file
################################################################################
sub get_availsize {
    my ($self, %args) = @_;

    $args{field} = "availsize";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: is_chksumenabled()
# Objective:   Is aggregate chksumenabled
# Details:     See POD documentation at the end of this file
################################################################################
sub is_chksumenabled {
    my ($self, %args) = @_;

    $args{field} = "chksumenabled";

    # the _get_field() will handle error
    if ($self->_get_field(%args) =~ /true/i) {
        return $SanUtil::TRUE;
    } else {
        return $SanUtil::FALSE;
    }
}

################################################################################
# Method Name: get_chksumstatus()
# Objective:   Get aggregate chksumstatus
# Details:     See POD documentation at the end of this file
################################################################################
sub get_chksumstatus {
    my ($self, %args) = @_;

    $args{field} = "chksumstatus";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_chksumstyle()
# Objective:   Get aggregate chksumstyle
# Details:     See POD documentation at the end of this file
################################################################################
sub get_chksumstyle {
    my ($self, %args) = @_;

    $args{field} = "chksumstyle";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_diskcount()
# Objective:   Get aggregate diskcount
# Details:     See POD documentation at the end of this file
################################################################################
sub get_diskcount {
    my ($self, %args) = @_;

    $args{field} = "diskcount";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_disklist()
# Objective:   Get aggregate disklist
# Details:     See POD documentation at the end of this file
################################################################################
sub get_disklist {
    my ($self, %args) = @_;

    $args{field} = "disklist";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_ha_policy()
# Objective:   Get aggregate ha_policy
# Details:     See POD documentation at the end of this file
################################################################################
sub get_ha_policy {
    my ($self, %args) = @_;

    $args{field} = "ha_policy";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_iron_last_start_errno()
# Objective:   Get aggregate iron_last_start_errno
# Details:     See POD documentation at the end of this file
################################################################################
sub get_iron_last_start_errno {
    my ($self, %args) = @_;

    $args{field} = "iron_last_start_errno";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_iron_last_start_error_info()
# Objective:   Get aggregate iron_last_start_error_info
# Details:     See POD documentation at the end of this file
################################################################################
sub get_iron_last_start_error_info {
    my ($self, %args) = @_;

    $args{field} = "iron_last_start_error_info";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_iron_status()
# Objective:   Get aggregate iron_status
# Details:     See POD documentation at the end of this file
################################################################################
sub get_iron_status {
    my ($self, %args) = @_;

    $args{field} = "iron_status";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_iron_scan_percentage()
# Objective:   Get aggregate iron_scan_percentage
# Details:     See POD documentation at the end of this file
################################################################################
sub get_iron_scan_percentage {
    my ($self, %args) = @_;

    $args{field} = "iron_scan_percentage";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_maxraidsize()
# Objective:   Get aggregate maxraidsize
# Details:     See POD documentation at the end of this file
################################################################################
sub get_maxraidsize {
    my ($self, %args) = @_;

    $args{field} = "maxraidsize";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_mirror_count_ls_destinations()
# Objective:   Get aggregate mirror_count_ls_destinations
# Details:     See POD documentation at the end of this file
################################################################################
sub get_mirror_count_ls_destinations {
    my ($self, %args) = @_;

    $args{field} = "mirror_count_ls_destinations";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_mirror_count_dp_destinations()
# Objective:   Get aggregate mirror_count_dp_destinations
# Details:     See POD documentation at the end of this file
################################################################################
sub get_mirror_count_dp_destinations {
    my ($self, %args) = @_;

    $args{field} = "mirror_count_dp_destinations";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_mirror_count_move_destinations()
# Objective:   Get aggregate mirror_count_move_destinations
# Details:     See POD documentation at the end of this file
################################################################################
sub get_mirror_count_move_destinations {
    my ($self, %args) = @_;

    $args{field} = "mirror_count_move_destinations";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_mirror_count_dp_qtree_destinations()
# Objective:   Get aggregate mirror_count_dp_qtree_destinations
# Details:     See POD documentation at the end of this file
################################################################################
sub get_mirror_count_dp_qtree_destinations {
    my ($self, %args) = @_;

    $args{field} = "mirror_count_dp_qtree_destinations";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_raid_block_checksum()
# Objective:   Get aggregate raid_block_checksum
# Details:     See POD documentation at the end of this file
################################################################################
sub get_raid_block_checksum {
    my ($self, %args) = @_;

    $args{field} = "raid_block_checksum";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_name()
# Objective:   Get aggregate name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "aggregate";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_nodes()
# Objective:   Get nodes this aggregate resided on
# Details:     See POD documentation at the end of this file
################################################################################
sub get_nodes {
    my ($self, %args) = @_;

    $args{field} = "nodes";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get parent object of aggregate
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_percent_used()
# Objective:   Get aggregate percent_used
# Details:     See POD documentation at the end of this file
################################################################################
sub get_percent_used {
    my ($self, %args) = @_;

    $args{field} = "percent_used";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_plexes()
# Objective:   Get aggregate plexes
# Details:     See POD documentation at the end of this file
################################################################################
sub get_plexes {
    my ($self, %args) = @_;

    $args{field} = "plexes";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_raid_block_checksum()
# Objective:   Get aggregate raid_block_checksum
# Details:     See POD documentation at the end of this file
################################################################################
sub get_raid_block_checksum {
    my ($self, %args) = @_;

    $args{field} = "raid_block_checksum";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_raid_zoned_checksum()
# Objective:   Get aggregate raid_zoned_checksum
# Details:     See POD documentation at the end of this file
################################################################################
sub get_raid_zoned_checksum {
    my ($self, %args) = @_;

    $args{field} = "raid_zoned_checksum";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_raidgroups()
# Objective:   Get aggregate raidgroups
# Details:     See POD documentation at the end of this file
################################################################################
sub get_raidgroups {
    my ($self, %args) = @_;

    $args{field} = "raidgroups";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_raidstatus()
# Objective:   Get aggregate raidstatus
# Details:     See POD documentation at the end of this file
################################################################################
sub get_raidstatus {
    my ($self, %args) = @_;

    $args{field} = "raidstatus";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_raidtype()
# Objective:   Get aggregate raidtype
# Details:     See POD documentation at the end of this file
################################################################################
sub get_raidtype {
    my ($self, %args) = @_;

    $args{field} =  "raidtype";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: is_snap_autodelete()
# Objective:   Is aggregate snap_autodelete
# Details:     See POD documentation at the end of this file
################################################################################
sub is_snap_autodelete {
    my ($self, %args) = @_;

    $args{field} = "snap_autodelete";

    # the _get_field() will handle error
    if ($self->_get_field(%args) =~ /on/i) {
        $SanUtil::TRUE;
    } else {
        $SanUtil::FALSE;
    }
}

################################################################################
# Method Name: get_size()
# Objective:   Get aggregate size
# Details:     See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self, %args) = @_;

    $args{field} = "size";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_state()
# Objective:   Get aggregate state
# Details:     See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self, %args) = @_;

    $args{field} = "state";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_type()
# Objective:   Get aggregate type
# Details:     See POD documentation at the end of this file
################################################################################
sub get_type {
    my ($self, %args) = @_;

    $args{field} = "type";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_usedsize()
# Objective:   Get aggregate usedsize
# Details:     See POD documentation at the end of this file
################################################################################
sub get_usedsize {
    my ($self, %args) = @_;

    $args{field} = "usedsize";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_uuid()
# Objective:   Get aggregate uuid
# Details:     See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self, %args) = @_;

    $args{field} = "uuid";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_volcount()
# Objective:   Get aggregate volcount
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volcount {
    my ($self, %args) = @_;

    $args{field} = "volcount";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_volcount_quiesced()
# Objective:   Get aggregate volcount_quiesced
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volcount_quiesced {
    my ($self, %args) = @_;

    $args{field} = "volcount_quiesced";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_volcount_not_online()
# Objective:   Get aggregate volcount_not_online
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volcount_not_online {
    my ($self, %args) = @_;

    $args{field} = "volcount_not_online";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_voltype()
# Objective:   Get aggregate voltype
# Details:     See POD documentation at the end of this file
################################################################################
sub get_voltype {
    my ($self, %args) = @_;

    $args{field} = "voltype";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_volume_style()
# Objective:   Get aggregate volume_style
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volume_style {
    my ($self, %args) = @_;

    $args{field} = "volume_style";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method Name: refresh()
# Objective:   Refresh aggregate object's data
# Details:     See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self) = @_;

    $self->_clear_err();

    # initialize cmd arguments for aggregate show
    my (%args) = (cmd => "storage_aggregate_show",
                  $self->_get_param_defaults());

    # use base class (San::COmmonInterface) refresh via SUPER::refresh
    # to update aggregate object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        $self->_set_err("failed to refresh aggregate object data", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: zero_spare_disks()
# Objective:   zero spare disks
# Details:     See POD documentation at the end of this file.
###############################################################################
sub zero_spare_disks {
    my ($self, %args) = @_;
    my $node_list;

    $self->_clear_err();

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # convert common args to args specifically for creating lif
    if (exists $args{nodes}) {
        $node_list = delete $args{nodes};
    } else {
        $node_list = $self->get_node();
    }

    # initialize cmd arguments for zeroing spare disks
    %args = (cmd  => "storage_disk_zerospares");

    foreach my $node (split",", $node_list) {
        %args = (%args, owner => $node);

        # use common invoke to translate cdef into UI cmd and execute it
        $self->SUPER::invoke_cmd_obj(%args);
        if ($self->error()) {
            $self->_set_err("failed to zeroing spare disks", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
}

################################################################################
# Private methods
################################################################################

################################################################################
# Method Name: _discover
# Objective:  return array or arrayref of aggregate objects
# Details:    See POD documentation at the end of this file
################################################################################
sub _discover {
    my ($class, %args) = @_;

    my ($self, $aggr_objs);

    # parameter initialization
    $self = {};
    bless $self, $class;

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parent object was passed in to bubble error up the stack
    my $parent_obj = delete $args{parent_obj};

    # San::CommonInterface is a base, generic, or SUPER class
    # San::ControllerCluster::Filer10M::Aggregate @ISA San::CommonInterface
    # in this case a base class.
    # Aggregate object has been instantiated as
    # San::ControllerCluster::Filer10M::Aggregate.
    # SUPER::new calls the generic new method from base (SUPER) class
    # San::CommonInterface::new. however, the object is still retained its
    # package name (San::ControllerCluster::Filer10M::Aggregate)
    # use common _new method to contruct aggregate object
    $self = $class->SUPER::_new(-name  => $args{aggregate},
                                -filer => $args{filer},
                                %args);
    if ($self->error()) {
        $parent_obj->_set_err("failed to contruct aggregate object", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # use base class (San::CommonInterface) discover to load aggregate objects
    # data and create accessor methods for data fields
    $aggr_objs = $self->SUPER::discover(-pkg  => __PACKAGE__,
                                        -cdef => "storage_aggregate_show",
                                        %args);
    if ($self->error()) {
        $parent_obj->_set_err("failed to discover aggregate object", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @$aggr_objs : $aggr_objs;
}

################################################################################
# Method Name: _get_param_defaults
# Objective:   Gets default aggregate parameters
# Details:     See POD documentation at the end of this file.
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    my %defaults_params = (aggregate => $self->get_name(-use_cache => 1));

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Aggregate
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # convert common args to args specifically for creating lif
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "storage_aggregate_show",
                                                $self->_get_param_defaults(),
                                                %args);
    }
}

###############################################################################
# Method Name: _create_aggr()
# Objective:   Initializes the aggregate object by creating aggregate on the filer
# Details:     See POD documentation at the end of this file.
###############################################################################
sub _create_aggr {
    my ($self, %args) = @_;

    my ($max, @disks, $disk_objs, $retries);

    # retries for verifying aggregate online
    my $max_retries = 5;

    # remove dash '-' from input parameters because parameters may be used
    # to create get accessor method on the fly
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd arguments for aggregate create
    %args = (%args, cmd => "storage_aggregate_create");

    # if disk objects are passed in, get their name
    # and append to disklist
    if (exists $args{disk_objs}) {
        $disk_objs = $args{disk_objs};
        foreach my $disk_obj (@$disk_objs) {
            push @disks, $disk_obj->get_name();
        }

        if ($self->get_cmd_interface() =~ /cli/i) {
            $args{disklist} = join(",", @disks);
        } else {
            $args{disks} = \@disks;
        }

        # delete disk_objs because we now have disklist to create aggr
        delete $args{disk_objs};
    } else {

        # if no disklist and diskcount, we should fail here
        unless ((exists $args{diskcount} and $args{diskcount} =~ /^\d+$/) or
                (exists $args{disklist})) {
            $self->_set_err("required -disk_objs, -disklist or -diskcount parameter to create aggregate");
            return $SanUtil::FALSE;
        }
    }

    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);
    # use common invoke from base class (San::CommonInterface)
    # to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, just bubble up
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # one time refresh to ensure current state
    if ($self->{enable_cache}) {
        $self->refresh(%args);
        if ($self->error()) {
            $self->_set_err("failed to refresh aggregate object", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # since we used these disks to create aggregate, refresh to obtain
    # their current state
    if (defined $disk_objs and scalar(@$disk_objs) >= 1) {
        foreach my $disk_obj (@$disk_objs) {
            $disk_obj->refresh();
            if ($disk_obj->error()) {
                $self->_set_err("failed to refresh disk object: " . refaddr($disk_obj), $disk_obj->errmsg());
                return $SanUtil::FALSE;
            }
            $self->sanlog(-msg_level => 3,
                          -msg       => $disk_obj->get_name() . " is " . $disk_obj->get_state());
        }
    } else {
        # NOTE: korry - need to implement
        # need to handle the diskcount param and refresh them
        # to obtain disk objects
        # perhaps $self->refresh() is already updated disk objects

    }

    return $SanUtil::TRUE;
} ## end sub _create_aggr

1;


#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Aggregate.pm

=head1 Author

Korry Nguyen (dl-san-automation)

=head1 Description

Provides an interface to manage an Aggregate

=head1 See Also

Aggregate package uses base class B<San::CommonInterface>  See these modules for further interface detail:

B<Error.pm, San::SanLog.pm, San::SanUtil.pm, San::SanCache.pm>

=head1 Attributes

=head1 B<Public Methods>

=head2 B<modify(%args)>

Modify an aggregate

=over 4

=item Inputs

=item * I<-state>

State to transition to: online|offline

=item * I<-raidtype>

Raid format of aggr i.e. "raid_dp"

=item * I<-maxraidsize>

Maximum raid size i.e. "16"

=item * I<-hybrid-enabled>

Enable hybrid aggregate: true:false

=item * I<-snapshot-autodelete>

Snapshot Autodelete: on|off

=item * I<-ha-policy>

High Availability Policy i.e. "sfo"

=item * I<-free-space-realloc>

Reallocate free space: on|off

=item * I<-snapshot-reserve>

Percent of snapshot reserve i.e. "10%"

=item Outputs

On success, returns $SanUtil::TRUE. On failre, returns $SanUtil::FALSE and sets the error flag

=item Examples

 # Offline an aggregate
 $aggr_obj->modify(-state => "offline");
 if ($aggr_obj->error()) {
    ...
 }

=back

=head2 B<wafliron_start(%args)>

Start wafliron on the current aggregate

=over 4

=item Inputs

=item * I<-aggregate>

The name of the aggregate to run walfiron against. Defaults to the current aggr object.

=item Outputs

On success, returns $SanUtil::TRUE. On failure, returns $SanUtil::FALSE and sets the error flag

=item Examples

 # start wafliron
 $aggr_obj->wafliron_start();
 if ($aggr_obj->error()) {
    ...
 }

=back

=head2 B<wafliron_stop(%args)>

Stop wafliron on the current aggregate

=over 4

=item Inputs

=item * I<-aggregate>

The name of the aggregate to stop walfiron against. Defaults to the current aggr object.

=item Outputs

On success, returns $SanUtil::TRUE. On failure, returns $SanUtil::FALSE and sets the error flag

=item Examples

 # stop wafliron
 $aggr_obj->wafliron_stop();
 if ($aggr_obj->error()) {
    ...
 }

=back

=head2 B<wafliron_commit(%args)>

Commit wafliron changes on the current aggregate

=over 4

=item Inputs

=item * I<-aggregate>

The name of the aggregate to commit walfiron changes against. Defaults to the current aggr object.

=item Outputs

On success, returns $SanUtil::TRUE. On failure, returns $SanUtil::FALSE and sets the error flag

=item Examples

 # commit wafliron changes
 $aggr_obj->wafliron_commit();
 if ($aggr_obj->error()) {
    ...
 }

=back

=head2 B<wafliron_reject(%args)>

Reject wafliron changes on the current aggregate

=over 4

=item Inputs

=item * I<-aggregate>

The name of the aggregate to reject walfiron changes against. Defaults to the current aggr object.

=item Outputs

On success, returns $SanUtil::TRUE. On failure, returns $SanUtil::FALSE and sets the error flag

=item Examples

 # reject wafliron changes
 $aggr_obj->wafliron_reject();
 if ($aggr_obj->error()) {
    ...
 }

=back

=head2 B<wafliron_review(%args)>

Review wafliron changes on the current aggregate

=over 4

=item Inputs

=item * I<-aggregate>

The name of the aggregate to review walfiron changes against. Defaults to the current aggr object.

=item Outputs

On success, returns the output of the cdef "storage_aggregate_wafliron_review"

=item Examples

 # obtain wafliron review
 my @output = $aggr_obj->wafliron_review();
 if ($aggr_obj->error()) {
    ...
 }

=back

=head2 B<wafliron_show(%args)>

Show wafliron changes on the current aggregate

=over 4

=item Inputs

=item * I<-aggregate>

The name of the aggregate to review walfiron changes against. Defaults to the current aggr object.

=item * I<-show_all>

If set to 1, then all aggregates will be displayed from one wafliron show command rather than the default behavior
of displaying only the current working aggregate.

=item * I<additional_inputs>

Inputs into the cdef "storage_aggregate_wafliron_show"

=item Outputs

On success, returns the output of the cdef "storage_aggregate_wafliron_show"

=item Examples

 # Show wafliron status for all aggrs
 my @output = $aggr_obj->wafliron_show(-show_all => 1);
 if ($aggr_obj->error()) {
    ...
 }

 # Show wafliron status for only aggr_obj
 my @output = $aggr_obj->wafliron_show();
 if ($aggr_obj->error()) {
    ...
 }

=back

=cut
