# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to a Hwresource object
##
## @author vchintal@netapp.com, dl-ssan-test
## @status public

package San::ControllerCluster::Filer10M::Hwresource;

# compiler directives
use strict;

use Params::Validate qw( :all );

# inherit base class
use base qw( San::CommonInterface );

###############################################################################
# Method:    new
# Objective: Create new Hwresource object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, @list_names);

    # initialize parameter
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # validate input params
    my @input_params = (%args);
    (%args) = validate(@input_params,
                       {filer   => { type => SCALAR, required => 1 }
                       });

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct a Hwresource object
    $self = $class->SUPER::_new(-filer => $args{filer});
    if ($self->error()) {
        $self->_set_err("failed to create a Hwresource object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # create get object accessor method for adapter
    @list_names = qw( fcpadapter networkadapter);
    $self->create_accessor_methods(-type => "get_obj",
                                   -pkg  => __PACKAGE__,
                                   -list => \@list_names);

    return $self;
}

###############################################################################
# Method:    get_filer()
# Objective: Get filer name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_filer{
    my ($self) = @_;

    # the get() will handle error
    return $self->get("filer");
}

1;

################################################################################
# D Y N A M I C A L L Y   G E N E R A T E D   M E T H O D S
################################################################################

################################################################################
# Method Name: get_fcpadapter_objs()
# Objective:   Get fcpadapter objects. This method is created dynamically from
#              create_accessor_methods in the new() constructor. This is a
#              placeholder in case someone is looking for the method.
# Details:     See POD documentation at the end of this file for usage
################################################################################

################################################################################
# Method Name: get_networkadapter_objs()
# Objective:   Get networkadapter objects. This method is created dynamically
#              from create_accessor_methods in the new() constructor. This is a
#              placeholder in case someone is looking for the method.
# Details:     See POD documentation at the end of this file for usage
################################################################################

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Hwresource.pm

=head1 Author

Manoj Chintala (dl-ssan-test)

=head1 Description

Provides an interface to all hardware resources

=head1 See Also

Hwresource package used base class B<San::CommonInterface> that also inherit other modules. See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm>

=head1 Attributes

=head2 B<filer>

string = controller name

=head2 B<debug_level>

interger = 0 - 100 (default is 0)

=head1 B<Public Methods>

=head2 B<new(%args)>

Creates hardware resource object

=over 4

=item Inputs

=item * I<-filer> => <controller name>

Controller name

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. With higher level, it will log more debug messages

=item Outputs

Return Hwresource object

=item Examples

 # ex. 1: Create hwresource object directly
 use San::ControllerCluster::Filer10M::Hwresource;

 my $filer = "ssan-3170";

 my hwresource_obj = new Hwresource(-filer => $filer);
 if ($hwresource_obj->error()) {
    logcomment("failed to create hwresource object" . $hwresource->errmsg());
    return $TCD::FAIL;
 }

 # ex. 2: Create hwresource object using filer_obj
 use San::ControllerCluster::Filer10M;

 my $filer = "ssan-3170";

 my $filer_obj = new SanFiler (-filer => $filer);
 if ($filer_obj->error){
    logcomment("cannot instantiate filer object for $filer", "$filer_obj->errmsg()");
    return $TCD::FAIL;
 }

 $hwresource_obj = $filer_obj->get_hwresource_obj();
 if ($filer_obj->error()) {
    $logcomment("failed to get hardware resource object", $filer_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_fcpadapter_objs(%args)>

Get multiple fcpadapter objects

=over 4

=item Inputs

=item * I<-adapter | -name> (optional)

The regex matching of fcp adapter name

=item * I<-state> (optional)

The state of the adapter

=item * I<-adaptertype> (optional)

The type of the adapter

=item Outputs

Returns array of fcpadapter objects that meet the criteria. If not met, set error and returns empty array

=item Examples

 # ex. 1: Get all fcpadapter objects
 @fcpadapter_objs = $hwresource_obj->get_fcpadapter_objs();
 if ($hwresource_obj->error()){
     ...
 }

 # ex. 2: Get fcpadapter objects with the regex matching name
 @fcpadapter_objs = $hwresource_obj->get_fcpadapter_objs(-name => "^0");
 if ($hwresource_obj->error()){
     ...
 }

 # ex. 3: Get all online fcpadapter objects
 @fcpadapter_objs = $filer_obj->get_fcpadapter_objs(-state => "online");
 if ($hwresource_obj->error()){
     ...
 }

=back

=head2 B<get_networkadapter_objs(%args)>

Get multiple networkadapter objects

=over 4

=item Inputs

=item * I<-port | -name> (optional)

The regex matching of network adapter name

=item * I<-role> (optional)

The role of the adapter

=item * I<-mtu> (optional)

The mtu of the adapter

=item Outputs

Returns array of networkadapter objects that meet the criteria. If not met, set
error and returns empty array

=item Examples

 # ex. 1: Get all networkadapter objects
 @networkadapter_objs = $hwresource_obj->get_networkadapter_objs();
 if ($hwresource_obj->error()){
     ...
 }

 # ex. 2: Get networkadapter objects with the regex matching name
 @networkadapter_objs = $hwresource_obj->get_networkadapter_objs(-name => "^e");
 if ($hwresource_obj->error()){
     ...
 }

 # ex. 3: Get all data networkadapter objects
 @networkadapter_objs = $cserver_obj->get_networkadapter_objs();
 if ($hwresource_obj->error()){
     ...
 }

=back
