# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to a fcpadapter object
##
## @author vchintal@netapp.com, dl-ssan-test
## @status public

package San::ControllerCluster::Filer10M::Hwresource::Fcpadapter;

# compiler directives
use strict;

# imported modules
use Params::Validate qw( :all );

# inherit base class
use base qw( San::CommonInterface );

###############################################################################
# Method:    new
# Objective: Create new fcpadapter object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, @fields);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # cdef takes adapter key instead of name
    if (exists $args{name}) {
        $args{adapter} = delete $args{name};
    }

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct a adapter object
    $self = $class->SUPER::_new(-name  => $args{adapter},
                                -filer => $args{filer},
                                %args);
    if ($self->error()) {
        $self->_set_err("failed to create a adapter object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # If initialized is set to true like CommonInterface.pm's discover() does, we
    # already have the output of show command and so we don't need to call the show
    # command.
    unless ($args{initialized}) {
        my @input_params = (%args);
        eval {
            # validate input parameters
            (%args) = validate(@input_params,
                              {filer   => { type => SCALAR, required => 1 },
                               adapter => { type => SCALAR, required => 1 },
                              });
        };
        if ($@) {
            $self->_set_err("failed to validate parameter: " . $@);
            return $self;
        }

        # append cmd for fcp adapter show.
        %args = (%args, cmd => "network_fcp_adapter_show");

        # use common invoke from base class (San::CommonInterface)
        # to translate cdef into UI cmd and execute it
        $self->SUPER::invoke_cmd_obj(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
            return $self;
        }

        if ($self->{enable_cache}) {
            $self->refresh(%args);
            if ($self->error()) {
                # set error and bubble up
                $self->_set_err("failed to refresh " . $args{name}, $self->errmsg());
                return $self;
            }
        }

        $self->{initialized} = $SanUtil::TRUE;
    }

    # store args in adapter object and push the keys to list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @fields, lc($key);
    }

    return $self;
} ## end sub new

###############################################################################
# Method:    refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_fcp_adapter_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update Fcpadapter object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: modify()
# Objective:   Modify the fcp adapter
# Details:     See POD Documentation at the end of this file
###############################################################################
sub modify {
    my ($self, %args) = @_;

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    if ($self->get_cmd_interface() =~ /zapi/i) {
        if ($args{'state'} =~ /up/i) {
            delete $args{'state'} if (exists $args{'state'});
            # initialize cmd parameters
            %args = (%args,
                     cmd                  => "network_fcp_adapter_modify_up",
                     donot_append_vserver => 1,
                     $self->_get_param_defaults());
        }
        elsif ($args{'state'} =~ /down/i) {
            delete $args{'state'} if (exists $args{'state'});
            # initialize cmd parameters
            %args = (%args,
                     cmd                  => "network_fcp_adapter_modify_down",
                     donot_append_vserver => 1,
                     $self->_get_param_defaults());
        }
        elsif ($args{'speed'}) {
            # initialize cmd parameters
            %args = (%args,
                     cmd                  => "network_fcp_adapter_modify_speed",
                     donot_append_vserver => 1,
                     $self->_get_param_defaults());
        }
    }
    else {
        # Initialize args parameters
        %args = (%args,
                 cmd                  => "network_fcp_adapter_modify",
                 donot_append_vserver => 1,
                 $self->_get_param_defaults());
    }

    # Get a CmdsDriver object and executes the actual command
    $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Update $self after the set comment
    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh fcp adapter object:", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub modify()

###############################################################################
# Method:    get_list_info()
# Objective: get fcp adapter information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_fcp_adapter_show",
             $self->_get_param_defaults());

    # invoke cmd command and update the attrs
    my @records = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return {};
    } else {
        return $records[0];
    }
} ## end sub get_list_info

###############################################################################
# Method:    get_adapter()
# Objective: Get fcp adapter name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_adapter{
    my ($self, %args) = @_;

    $args{field} = "adapter";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_desciption()
# Objective: Get fcp adapter desciption
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_desciption{
    my ($self, %args) = @_;

    $args{field} = "description";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_physical_protocol()
# Objective: Get fcp adapter physical_protocol
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_physical_protocol{
    my ($self, %args) = @_;

    $args{field} = "physical_protocol";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_max_speed()
# Objective: Get fcp adapter max_speed
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_max_speed{
    my ($self, %args) = @_;

    $args{field} = "max_speed";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_nodename()
# Objective: Get fcp adapter nodename
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nodename{
    my ($self, %args) = @_;

    return $self->get_fc_wwnn(%args);
}

###############################################################################
# Method:    get_portname()
# Objective: Get fcp adapter portname
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_portname{
    my ($self, %args) = @_;

    return $self->get_fc_wwpn(%args);
}

###############################################################################
# Method:    get_status()
# Objective: Get fcp adapter status
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_status{
    my ($self, %args) = @_;

    $args{field} = "status";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_substatus()
# Objective: Get fcp adapter substatus
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_substatus{
    my ($self, %args) = @_;

    $args{field} = "substatus";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_portaddr()
# Objective: Get fcp adapter portaddr
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_portaddr{
    my ($self, %args) = @_;

    $args{field} = "portaddr";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_firmware_rev()
# Objective: Get fcp adapter firmware_rev
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_firmware_rev{
    my ($self, %args) = @_;

    $args{field} = "firmware_rev";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_pci_bus_width()
# Objective: Get fcp adapter pci_bus_width
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_pci_bus_width{
    my ($self, %args) = @_;

    $args{field} = "pci_bus_width";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_pci_clock_speed()
# Objective: Get fcp adapter pci_clock_speed
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_pci_clock_speed{
    my ($self, %args) = @_;

    $args{field} = "pci_clock_speed";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_cache_line_size()
# Objective: Get fcp adapter cache_line_size
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_cache_line_size{
    my ($self, %args) = @_;

    $args{field} = "cache_line_size";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_packet_size()
# Objective: Get fcp adapter packet_size
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_packet_size{
    my ($self, %args) = @_;

    $args{field} = "packet_size";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_sram_parity_enabled()
# Objective: Get fcp adapter sram_parity_enabled
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sram_parity_enabled{
    my ($self, %args) = @_;

    $args{field} = "sram_parity_enabled";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_external_gbic_enabled()
# Objective: Get fcp adapter external_gbic_enabled
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_external_gbic_enabled{
    my ($self, %args) = @_;

    $args{field} = "external_gbic_enabled";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_data_link_rate()
# Objective: Get fcp adapter data_link_rate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_data_link_rate{
    my ($self, %args) = @_;

    $args{field} = "data_link_rate";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_fabric_established()
# Objective: Get fcp adapter fabric_established
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fabric_established{
    my ($self, %args) = @_;

    $args{field} = "fabric_established";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_conn_established()
# Objective: Get fcp adapter conn_established
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_conn_established{
    my ($self, %args) = @_;

    $args{field} = "conn_established";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_media_type()
# Objective: Get fcp adapter media_type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_media_type{
    my ($self, %args) = @_;

    $args{field} = "media_type";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_speed()
# Objective: Get fcp adapter speed
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_speed{
    my ($self, %args) = @_;

    $args{field} = "speed";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_state()
# Objective: Get fcp adapter state
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_state{
    my ($self, %args) = @_;

    $args{field} = "state";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_lif_capable()
# Objective: Get fcp adapter lif_capable
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lif_capable{
    my ($self, %args) = @_;

    $args{field} = "lif_capable";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_tha_index()
# Objective: Get fcp adapter tha_index
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_tha_index{
    my ($self, %args) = @_;

    $args{field} = "tha_index";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_handle()
# Objective: Get fcp adapter handle
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_handle{
    my ($self, %args) = @_;

    $args{field} = "handle";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_pslot()
# Objective: Get fcp adapter pslot
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_pslot{
    my ($self, %args) = @_;

    $args{field} = "pslot";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_orig_wwnn()
# Objective: Get fcp adapter orig_wwnn
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_orig_wwnn{
    my ($self, %args) = @_;

    $args{field} = "orig_wwnn";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_fc_wwnn()
# Objective: Get fcp adapter fc_wwnn
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fc_wwnn{
    my ($self, %args) = @_;

    $args{field} = "fc_wwnn";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_fc_wwpn()
# Objective: Get fcp adapter fc_wwpn
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_fc_wwpn{
    my ($self, %args) = @_;

    $args{field} = "fc_wwpn";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_queue_depth()
# Objective: Get fcp adapter queue_depth
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_queue_depth{
    my ($self, %args) = @_;

    $args{field} = "queue_depth";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    enable_sram_dumps()
# Objective: enables SRAM dump generation on this adapter
# Details:   See POD documentation at the end of this file.
###############################################################################
sub enable_sram_dumps {
    my ($self) = @_;

    # this procedure can only be done with node scoped commands
    # fcp dump -e <adapter name>
    my $adapter_name = $self->get_adapter();

    $self->run_node_scope(-cmd => "fcp dump -e $adapter_name",
                          -priv => "diag");
    if ($self->error()) {
        $self->_set_err("Failed to enable SRAM dumps on adapter '$adapter_name'", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}


###############################################################################
# Method:    disable_sram_dumps()
# Objective: disables SRAM dump generation on this adapter
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disable_sram_dumps {
    my ($self) = @_;

    # this procedure can only be done with node scoped commands
    # fcp dump -d <adapter name>
    my $adapter_name = $self->get_adapter();

    $self->run_node_scope(-cmd => "fcp dump -d $adapter_name",
                          -priv => "diag");
    if ($self->error()) {
        $self->_set_err("Failed to disable SRAM dumps on adapter '$adapter_name'", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method:    _get_param_defaults
# Objective: Gets default adapter parameters
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    # default hash parameters to be used in adapter object
    my %defaults_params = (adapter => $self->{adapter},
                           node    => $self->{filer});

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Fcpadapter
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # convert common args to args specifically for creating lif
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "network_fcp_adapter_show",
                                                field => $args{field},
                                                $self->_get_param_defaults());
    }
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Fcpadapter.pm

=head1 Author

Manoj Chintala (dl-ssan-test)

=head1 Description

Provides an interface to fcpadapter

=head1 See Also

Fcpadapter package used base class B<San::CommonInterface> that also inherit other modules. See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm>

=head1 Attributes

=head2 B<filer>

string = controller name

=head2 B<adapter>

string = fcp adapter name

=head2 B<debug_level>

interger = 0 - 100 (default is 0)

=head1 B<Public Methods>

=head2 B<new(%args)>

Create a fcpadapter object

=over 4

=item Inputs

=item * I<-filer> => <controller name>

Controller name

=item * I<-name | -adapter> => <name of fcpadapter>

Fcp adapter name

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. With higher level, it will log more debug messages

=item Outputs

Return fcpadapter object

=item Examples

 # ex. 1: Create fcpadapter object directly
 use San::ControllerCluster::Filer10M::Hwresource::Fcpadapter;

 my $adapter = "0c";
 my fcpadapter_obj = new Fcpadapter(-filer => $filer,
                                    -adapter => $adapter);
 if($fcpadapter_obj->error()) {
     logcomment("Failed to create fcpadapter object" . $fcpadapter_obj->errmsg());
     return $TCD::FAIL;
 }

 # ex. 2: Create fcpadapter object from filer object
 use San::ControllerCluster::Filer10M;

 my $filer = "ssan-3170";
 my $adapter = "0c";

 my $filer_obj = new SanFiler (-filer => $filer);
 if ($filer_obj->error()){
     logcomment("Cannot instantiate filer object for $filer", "$filer_obj->errmsg()");
     return $TCD::FAIL;
 }

 my $fcpadapter_obj = $filer_obj->get_fcpadapter_objs(-adapter => $adapter);
 if ($filer_obj->error()) {
     $logcomment("FAIL: Failed to get fcp adapter objects ", $filer_obj->errmsg());
     return $TCD::FAIL;
 }

=back

=head2 B<refresh()>

Refresh and update fcpadapter object

=over 4

=item Inputs

None

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Refresh and update fcpadapter object
 $fcpadapter_obj->refresh();
 if ($fcpadapter_obj->error()){
     ...
 }

=back

=head2 B<modify()>

Modifies the fcpadapter object

=over 4

=item Inputs

=item * state: (optional) "up" or "down" to change the state of the adapter.

=item * speed: (optional) 1|2|4|8|10|auto Speed of the adapter.

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Modify fcpadapter object
 $fcpadapter_obj->modify(-state   => "down");
 if ($fcpadapter_obj->error()){
     ...
 }

=back

=head2 B<get_adapter()>

Get the fcp adapter name

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter name

 # ex. 1: Display fcp adapter name
 logcomment("Fcp adapter name " . $fcpadapter_obj->get_adapter());

=back

=head2 B<get_description()>

Get the fcp adapter description

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter description

 # ex. 1: Display fcp adapter description
 logcomment("Fcp adapter description " . $fcpadapter_obj->get_description());

=back

=head2 B<get_physical_protocol()>

Get the fcp adapter physical protocol

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter physical protocol

 # ex. 1: Display fcp adapter physical protocol
 logcomment("Fcp adapter physical protocol " . $fcpadapter_obj->get_physical_protocol());

=back

=head2 B<get_max_speed()>

Get the fcp adapter max speed

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter max speed

 # ex. 1: Display fcp adapter max speed
 logcomment("Fcp adapter max speed " . $fcpadapter_obj->get_max_speed());

=back

=head2 B<get_nodename()>

Get the fcp adapter nodename

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter nodename

 # ex. 1: Display fcp adapter nodename
 logcomment("Fcp adapter nodename " . $fcpadapter_obj->get_nodename());

=back

=head2 B<get_portname()>

Get the fcp adapter portname

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter portname

 # ex. 1: Display fcp adapter portname
 logcomment("Fcp adapter portname " . $fcpadapter_obj->get_portname());

=back

=head2 B<get_status()>

Get the fcp adapter status

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter status

 # ex. 1: Display fcp adapter status
 logcomment("Fcp adapter status " . $fcpadapter_obj->get_status());

=back

=head2 B<get_substatus()>

Get the fcp adapter substatus

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter substatus

 # ex. 1: Display fcp adapter substatus
 logcomment("Fcp adapter substatus " . $fcpadapter_obj->get_substatus());

=back

=head2 B<get_portaddr()>

Get the fcp adapter portaddr

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter portaddr

 # ex. 1: Display fcp adapter portaddr
 logcomment("Fcp adapter portaddr " . $fcpadapter_obj->get_portaddr());

=back

=head2 B<get_firmware_rev()>

Get the fcp adapter firmware_rev

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter firmware_rev

 # ex. 1: Display fcp adapter firmware_rev
 logcomment("Fcp adapter firmware_rev " . $fcpadapter_obj->get_firmware_rev());

=back

=head2 B<get_pci_bus_width()>

Get the fcp adapter pci_bus_width

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter pci_bus_width

 # ex. 1: Display fcp adapter pci_bus_width
 logcomment("Fcp adapter pci_bus_width " . $fcpadapter_obj->get_pci_bus_width());

=back

=head2 B<get_pci_clock_speed()>

Get the fcp adapter pci_clock_speed

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter pci_clock_speed

 # ex. 1: Display fcp adapter pci_clock_speed
 logcomment("Fcp adapter pci_clock_speed " . $fcpadapter_obj->get_pci_clock_speed());

=back

=head2 B<get_cache_line_size()>

Get the fcp adapter cache_line_size

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter cache_line_size

 # ex. 1: Display fcp adapter cache_line_size
 logcomment("Fcp adapter cache_line_size " . $fcpadapter_obj->get_cache_line_size());

=back

=head2 B<get_packet_size()>

Get the fcp adapter packet_size

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter packet_size

 # ex. 1: Display fcp adapter packet_size
 logcomment("Fcp adapter packet_size " . $fcpadapter_obj->get_packet_size());

=back

=head2 B<get_sram_parity_enabled()>

Get the fcp adapter sram_parity_enabled

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter sram_parity_enabled

 # ex. 1: Display fcp adapter sram_parity_enabled
 logcomment("Fcp adapter sram_parity_enabled " . $fcpadapter_obj->get_sram_parity_enabled());

=back

=head2 B<get_external_gbic_enabled()>

Get the fcp adapter external_gbic_enabled

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter external_gbic_enabled

 # ex. 1: Display fcp adapter external_gbic_enabled
 logcomment("Fcp adapter external_gbic_enabled " . $fcpadapter_obj->get_external_gbic_enabled());

=back

=head2 B<get_data_link_rate()>

Get the fcp adapter data_link_rate

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter data_link_rate

 # ex. 1: Display fcp adapter data_link_rate
 logcomment("Fcp adapter data_link_rate " . $fcpadapter_obj->get_data_link_rate());

=back

=head2 B<get_fabric_established()>

Get the fcp adapter fabric_established

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter fabric_established

 # ex. 1: Display fcp adapter fabric_established
 logcomment("Fcp adapter fabric_established " . $fcpadapter_obj->get_fabric_established());

=back

=head2 B<get_conn_established()>

Get the fcp adapter conn_established

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter conn_established

 # ex. 1: Display fcp adapter conn_established
 logcomment("Fcp adapter conn_established " . $fcpadapter_obj->get_conn_established());

=back

=head2 B<get_media_type()>

Get the fcp adapter media_type

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter media_type

 # ex. 1: Display fcp adapter media_type
 logcomment("Fcp adapter media_type " . $fcpadapter_obj->get_media_type());

=back

=head2 B<get_speed()>

Get the fcp adapter speed

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter speed

 # ex. 1: Display fcp adapter speed
 logcomment("Fcp adapter speed " . $fcpadapter_obj->get_speed());

=back

=head2 B<get_state()>

Get the fcp adapter state

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter state

 # ex. 1: Display fcp adapter state
 logcomment("Fcp adapter state " . $fcpadapter_obj->get_state());

=back

=head2 B<get_lif_capable()>

Get the fcp adapter lif_capable

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter lif_capable

 # ex. 1: Display fcp adapter lif_capable
 logcomment("Fcp adapter lif_capable " . $fcpadapter_obj->get_lif_capable());

=back

=head2 B<get_tha_index()>

Get the fcp adapter tha_index

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter tha_index

 # ex. 1: Display fcp adapter tha_index
 logcomment("Fcp adapter tha_index " . $fcpadapter_obj->get_tha_index());

=back

=head2 B<get_handle()>

Get the fcp adapter handle

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter handle

 # ex. 1: Display fcp adapter handle
 logcomment("Fcp adapter handle " . $fcpadapter_obj->get_handle());

=back

=head2 B<get_pslot()>

Get the fcp adapter pslot

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter pslot

 # ex. 1: Display fcp adapter pslot
 logcomment("Fcp adapter pslot " . $fcpadapter_obj->get_pslot());

=back

=head2 B<get_orig_wwnn()>

Get the fcp adapter original world wide node name

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter inal world wide node name

 # ex. 1: Display fcp adapter orig_wwnn
 logcomment("Fcp adapter orig_wwnn " . $fcpadapter_obj->get_orig_wwnn());

=back

=head2 B<get_fc_wwnn()>

Get the fcp adapter world wide node name

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter world wide node name

 # ex. 1: Display fcp adapter fc_wwnn
 logcomment("Fcp adapter fc_wwnn " . $fcpadapter_obj->get_fc_wwnn());

=back

=head2 B<get_fc_wwpn()>

Get the fcp adapter world wide port name

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter world wide port name

 # ex. 1: Display fcp adapter fc_wwpn
 logcomment("Fcp adapter fc_wwpn " . $fcpadapter_obj->get_fc_wwpn());

=back

=head2 B<get_queue_depth()>

Get the fcp adapter queue_depth

=over 4

=item Inputs

None

=item Outputs

Returns fcp adapter queue_depth

 # ex. 1: Display fcp adapter queue_depth
 logcomment("Fcp adapter queue_depth " . $fcpadapter_obj->get_queue_depth());

=back

=head2 B<enable_sram_dumps()>

Enable SRAM Core dumps on this FCP adapter.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE if successful, $SanUtil::FALSE if not and sets error.

=item Example

 # enable sram dumps
 $fcpadapter_obj->enable_sram_dumps();
 if ($fcpadapter_obj->error()) {
     ...
 }

=back

=head2 B<disable_sram_dumps()>

Disable SRAM Core dumps on this FCP adapter.

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE if successful, $SanUtil::FALSE if not and sets error.

=item Example

 # disable sram dumps
 $fcpadapter_obj->disable_sram_dumps();
 if ($fcpadapter_obj->error()) {
     ...
 }

=back
