# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to a Networkadapter object
##
## @author vchintal@netapp.com, dl-ssan-automation
## @status public

package San::ControllerCluster::Filer10M::Hwresource::Networkadapter;

# compiler directives
use strict;

# imported modules
use Params::Validate qw( :all );

# inherit base class
use base qw( San::CommonInterface );

###############################################################################
# Method:    new
# Objective: Create new Networkadapter object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, @fields);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # cdef takes port key instead of name
    if (exists $args{name}) {
        $args{port} = delete $args{name};
    }

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct a adapter object
    $self = $class->SUPER::_new(-name  => $args{port},
                                -filer => $args{filer},
                                %args);
    if ($self->error()) {
        $self->_set_err("failed to create object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # If initialized is set to true like CommonInterface.pm's discover() does, we
    # already have the output of show command and so we don't need to call the show
    # command.
    unless ($args{initialized}) {
        my @input_params = (%args);
        eval {
            # validate input parameters
            (%args) = validate(@input_params,
                              {filer   => { type => SCALAR, required => 1 },
                               port => { type => SCALAR, required => 1 },
                              });
        };
        if ($@) {
            $self->_set_err("failed to validate parameter: " . $@);
            return $self;
        }

        if ($self->{enable_cache}) {
            $self->refresh(%args);
            if ($self->error()) {
                # set error and bubble up
                $self->_set_err("failed to refresh " . $args{name}, $self->errmsg());
                return $self;
            }
        }

        $self->{initialized} = $SanUtil::TRUE;
    }

    # store args in adapter object and push the keys to list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @fields, lc($key);
    }

    return $self;
} ## end sub new

###############################################################################
# Method:    refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_port_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update networkadapter object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: modify()
# Objective:   Modify the network adapter
# Details:     See POD Documentation at the end of this file
###############################################################################
sub modify {
    my ($self, %args) = @_;

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Initialize args parameters
    %args = (%args,
             cmd                  => "network_port_modify",
             donot_append_vserver => 1,
             $self->_get_param_defaults());

    # Get a CmdsDriver object and executes the actual command
    $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Update $self after the set comment
    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh network adapter object:",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub modify()

###############################################################################
# Method Name: get_list_info()
# Objective: Get network adapter information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "network_port_show");

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke_cmd_obj to obtain adapter object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return {};
    }

    unless (scalar(@output)) {
        return {};
    }

    return $output[0];
} ## end sub get_list_info

###############################################################################
# Method:    get_node()
# Objective: Get networkapter node name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_node{
    my ($self, %args) = @_;

    $args{field} = "node";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_port()
# Objective: Get networkapter name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_port{
    my ($self, %args) = @_;

    $args{field} = "port";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_mac()
# Objective: Get mac address assigned
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_mac{
    my ($self, %args) = @_;

    $args{field} = "mac";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_mtu()
# Objective: Get network adapter mtu
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_mtu{
    my ($self, %args) = @_;

    $args{field} = "mtu";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_role()
# Objective: Get network adapter role
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_role{
    my ($self, %args) = @_;

    $args{field} = "role";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_link()
# Objective: Get network adapter link
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_link{
    my ($self, %args) = @_;

    $args{field} = "link";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method:    _get_param_defaults
# Objective: Gets default adapter parameters
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    # default hash parameters to be used in adapter object
    my %defaults_params = (port => $self->get_port(),
                           node => $self->get_node());

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for
#            Networkadapter
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # convert common args to args specifically for creating lif
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "network_port_show",
                                                field => $args{field},
                                                $self->_get_param_defaults());
    }
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Networkadapter.pm

=head1 Author

Manoj Chintala (dl-ssan-automation)

=head1 Description

Provides an interface to Networkadapter

=head1 See Also

Networkadapter package used base class B<San::CommonInterface> that also inherit
other modules. See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm>

=head1 Attributes

=head2 B<filer>

string = controller name

=head2 B<port>

string = port name

=head2 B<debug_level>

interger = 0 - 100 (default is 0)

=head1 B<Public Methods>

=head2 B<new(%args)>

Create a networkadapter object

=over 4

=item Inputs

=item * I<-node> => <controller name>

Controller name

=item * I<-name | -port> => <name of port>

Port name

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. With higher level, it will log more debug messages

=item Outputs

Return networkadapter object

=item Examples

 # ex. 1: Create networkadapter object directly
 use San::ControllerCluster::Filer10M::Hwresource::Networkadapter;

 my $port = "e0c";
 my networkadapter_obj = new Networkadapter(-port => $port);
 if($networkadapter_obj->error()) {
     logcomment("Failed to create networkadapter object" . $networkadapter_obj->errmsg());
     return $TCD::FAIL;
 }

 # ex. 2: Create networkadapter object from cserver object
 use San::ControllerCluster;

 my $port = "e0c";

 my $cserver_obj = $Clus_Obj->get_cserver_obj();
 if ($cserver_obj->error()){
     logcomment("Cannot get cserver object", "$cserver_obj->errmsg()");
     return $TCD::FAIL;
 }

 my $networkadapter_obj = $cserver_obj->get_networkadapter_objs(-port => $port);
 if ($cserver_obj->error()) {
     $logcomment("FAIL: Failed to get networkadapter objects ",
                 $cserver_obj->errmsg());
     return $TCD::FAIL;
 }

=back

=head2 B<refresh()>

Refresh and update networkadapter object

=over 4

=item Inputs

None

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Refresh and update networkadapter object
 $networkadapter_obj->refresh();
 if ($networkadapter_obj->error()){
     ...
 }

=back

=head2 B<modify()>

Modifies the networkadapter object

=over 4

=item Inputs

=item * port: (optional) port name

=item * role: (optional) data|cluster|node-mgmt|intercluster
              role of the adapter

=item * mtu: (optional) mtu size

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Modify networkadapter object
 $networkadapter_obj->modify(-port   => "e0c");
 if ($networkadapter_obj->error()){
     ...
 }

=back

=head2 B<get_list_info()>

Get networkadapter information. This method invokes a network_port_show cdef.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns a reference to a hash containing the output of "network port show"
command. If it fails, it will return a null and set the error flag.

=item Examples

 # Display network adapter info
 $info = $networkadapter_obj->get_list_info();
 if ($networkadapter_obj->error()) {
    logcomment("Failed to get network adapter info" .
               $networkadapter_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_port()>

Get the network adapter name

=over 4

=item Inputs

None

=item Outputs

Returns network port name

 # ex. 1: Display network port name
 logcomment("Network port name " . $networkadapter_obj->get_port());

=back

=head2 B<get_node()>

Get the node name

=over 4

=item Inputs

None

=item Outputs

Returns network adapter mac address

 # ex. 1: Display network adapter mac address
 logcomment("Network adapter mac address " . $networkdapter_obj->get_mac());

=back

=head2 B<get_mtu()>

Get the network port mtu

=over 4

=item Inputs

None

=item Outputs

Returns network port mtu

 # ex. 1: Display network port mtu
 logcomment("Network port mtu " . $networkadapter_obj->get_mtu());

=back

=head2 B<get_role()>

Get the network port role

=over 4

=item Inputs

None

=item Outputs

Returns network port role name

 # ex. 1: Display network port role name
 logcomment("Network port name " . $networkadapter_obj->get_port());

=back

=head2 B<get_link()>

Get the network port link status

=over 4

=item Inputs

None

=item Outputs

Returns network port link status

 # ex. 1: Display network port link status
 logcomment("Network port link status " . $networkadapter_obj->get_link());

=back
