# Copyright (c) 2008-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to a Vserver object
##
## @author korry@netapp.com, dl-ssan-automation
## @status Shared

package San::ControllerCluster::Filer10M::Vserver;

# compiler directives
use strict;

# imported modules
use San::ControllerCluster::Filer10M::Vserver::Volume;
use San::ControllerCluster::Filer10M::Vserver::Igroup;
use San::ControllerCluster::Filer10M::Vserver::Portset;
use San::ControllerCluster::Filer10M::Vserver::Lif;
use San::ControllerCluster::Filer10M::Vserver::Fcp;
use San::ControllerCluster::Filer10M::Vserver::Fcptarget;
use San::SanClient::Fcptarget;
use San::ControllerCluster::Filer10M::Vserver::Iscsi;
use San::ControllerCluster::Filer10M::Vserver::Iscsitarget;
use San::SanClient::Iscsitarget;
use Params::Validate qw( SCALAR ARRAYREF );

# Constants
use constant MAX_ZAPI_RECORDS => 1024;

# inherit base class
use base qw( San::CommonInterface
             San::ControllerCluster::Filer10M::VserverHelper);

# list of sub components
my @Sub_components = qw( volume igroup iscsi lif fcp portset );

###############################################################################
# Method Name: new
# Objective:   Create new Cluster Vserver object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self);

    # initialize parameter
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # use base class to contruct vserver object
    $self = $class->SUPER::_new(-name  => $args{vserver},
                                -filer => $args{filer},
                                %args);
    if ($self->error()) {
        $self->_set_err("failed to contruct Vserver object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # set and store all Vserver elements
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # create get object accessor method for volume
    $self->create_accessor_methods(-type => "get_obj",
                                   -pkg  => __PACKAGE__,
                                   -list => \@Sub_components);

    # parent argument has been stored, delete unused parent argument now
    delete $args{parent_obj};
    delete $args{debug_level};

    # initialize object by creating and update it has not been initialized
    unless ($args{initialized}) {
        $self->_create_vserver(%args);
        if ($self->error()) {
            # set error and bubble up
             $self->_set_err("failed to _create_vserver() ", $self->errmsg());
            return $self;
        }
    }

    # set place holder for iscsi and fcp lif data
    $self->{fcp_lif_data} = [];
    $self->{iscsi_lif_data} = [];

    return $self;
} ## end sub new

###############################################################################
# Method Name: create_igroup_obj
# Objective:   Create Igroup object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_igroup_obj {
    my ($self, %args) = @_;

    my ($igroup_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # append aggregate name parameter to argument list
    %args = (%args, $self->_get_param_defaults());

    # instantiate igroup object
    $igroup_obj =
      San::ControllerCluster::Filer10M::Vserver::Igroup->new(
          -parent_obj => $self,
          -filer  => $self->get_filer(),
          %args);
    if ($igroup_obj->error()) {
        $self->_set_err("failed to instantiate igroup object", $igroup_obj->errmsg());
        return {};
    } else {
        # store volume object in filer object
        push @{$self->{igroup_objs}}, $igroup_obj;
    }
    return $igroup_obj;
} ## end sub create_igroup_obj

###############################################################################
# Method Name: create_iscsi_obj
# Objective:   Create Iscsi object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_iscsi_obj {
    my ($self, %args) = @_;

    my ($iscsi_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # append vserver name parameter to argument list
    %args = (%args, vserver => $self->{vserver});

    # instantiate volume object
    $iscsi_obj =
      San::ControllerCluster::Filer10M::Vserver::Iscsi->new(
          -parent_obj => $self,
          -filer  => $self->get_filer(),
          %args);
    if ($iscsi_obj->error()) {
        $self->_set_err("failed to instantiate iscsi object", $iscsi_obj->errmsg());
        return {};
    } else {
        # store iscsi object in filer object
        $self->{iscsi_obj} = $iscsi_obj;
    }
    return $iscsi_obj;
} ## end sub create_iscsi_obj

###############################################################################
# Method Name: create_fcp_obj
# Objective:   Create Fcp object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_fcp_obj {
    my ($self, %args) = @_;

    my ($fcp_obj);

    $self->_clear_err();

    # remove the dash from argument list which will be used to create 'get'
    # methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # append vserver name parameter to argument list
    %args = (%args, vserver => $self->{vserver});

    # instantiate fcp object
    $fcp_obj =
        San::ControllerCluster::Filer10M::Vserver::Fcp->new(
            -parent_obj => $self,
            -filer  => $self->get_filer(),
            %args);

    if ($fcp_obj->error()) {
        $self->_set_err("failed to instantiate fcp object", $fcp_obj->errmsg());
        return {};
    } else {
        # store fcp object in filer object
        $self->{fcp_obj} = $fcp_obj;
    }

    return $fcp_obj;
} ## end sub create_fcp_obj

###############################################################################
# Method Name: create_lif_obj
# Objective:   Create logical i/f object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_lif_obj {
    my ($self, %args) = @_;

    my ($lif_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    if (exists $args{name}) {
        $args{lif} = $args{name};
        delete $args{name};
    }
    # append vserver name parameter to argument list
    # use reference variable for vserver so when vserver is renamed
    # we don't have to update subcomponents
    %args = (%args, server => $self->{vserver});

    # instantiate volume object
    $lif_obj =
      San::ControllerCluster::Filer10M::Vserver::Lif->new(
          -parent_obj => $self,
          -filer  => $self->get_filer(),
          %args);
    if ($lif_obj->error()) {
        $self->_set_err("failed to instantiate lif object", $lif_obj->errmsg());
        return {};
    } else {
        # store lif object in filer object
        push @{$self->{lif_objs}}, $lif_obj;
    }
    return $lif_obj;
} ## end sub create_lif_obj

###############################################################################
# Method Name: create_portset_obj
# Objective:   Create Portset object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_portset_obj {
    my ($self, %args) = @_;

    my ($portset_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # append aggregate name parameter to argument list
    %args = (%args, $self->_get_param_defaults());

    # instantiate portset object
    $portset_obj =
      San::ControllerCluster::Filer10M::Vserver::Portset->new(
          -parent_obj => $self,
          -filer => $self->get_filer(),
          %args);
    if ($portset_obj->error()) {
        $self->_set_err("failed to instantiate portset object", $portset_obj->errmsg());
        return {};
    } else {
        # store volume object in filer object
        push @{$self->{portset_objs}}, $portset_obj;
    }
    return $portset_obj;
} ## end sub create_portset_obj

###############################################################################
# Method Name: create_volume_obj
# Objective:   Create Volume object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_volume_obj {
    my ($self, %args) = @_;

    my ($vol_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # append aggregate name parameter to argument list
    %args = (%args, $self->_get_param_defaults());

    # instantiate volume object
    $vol_obj =
      San::ControllerCluster::Filer10M::Vserver::Volume->new(
          -parent_obj => $self,
          -filer  => $self->get_filer(),
          %args);
    if ($vol_obj->error()) {
        $self->_set_err("failed to instantiate volume object", $vol_obj->errmsg());
        return {};
    } else {
        # store volume object in filer object
        push @{$self->{volume_objs}}, $vol_obj;
    }
    return $vol_obj;
} ## end sub create_volume_obj

###############################################################################
# Method Name: delete()
# Objective:   Deletes a Vserver object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($fcp_obj, $iscsi_obj, $root_vol);
    my (@vol_objs, @routing_groups, @lif_objs)= ();
    my $force = $SanUtil::FALSE;

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Users can specify to delete a vserver by force
    if (exists($args{force})) {
        $force = delete($args{force});
    }

    # If users are force destroying then make sure it wipes everything, starting
    # by stopping services, root volumes and routing groups.
    if ($force == $SanUtil::TRUE) {
        # Stop fcp service if needed
        if ($self->fcp_exist()) {
            $fcp_obj = $self->get_fcp_obj();
	    if ($fcp_obj->get_status() eq "up") {
            $fcp_obj->stop();
	    }
            if ($fcp_obj->error()) {
                $self->_set_err("failed to stop fcp while force destroying a vserver",
                                $fcp_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # Stop iscsi service if needed
        if ($self->iscsi_exist()) {
            $iscsi_obj = $self->get_iscsi_obj();
	    if ($iscsi_obj->get_status() eq "up") {
            $iscsi_obj->stop();
	    }
            if ($iscsi_obj->error()) {
                $self->_set_err("failed to stop iscsi while force destroying a vserver",
                                $iscsi_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }

        @lif_objs = $self->get_lif_objs();
        foreach my $lif (@lif_objs) {
            next if ($lif->get_name() =~ /_DELETE_/i);
            $lif->delete();
            if ($lif->error()) {
                $self->_set_err("failed to delete lif objects while force " .
                                "destroying the vserver" , $lif->errmsg());
                return $SanUtil::FALSE;
            }
        }

        my @igroups = $self->get_igroup_objs();
        if ($self->error()) {
            $self->_set_err("Unable to get igroups", $self->errmsg());
            return $SanUtil::FALSE;
        }

        foreach my $igroup (@igroups) {
            $igroup->delete(force => 'true');
            if ($igroup->error()) {
                $self->_set_err("failed to delete igroup", $igroup->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # Get volumes and delete them
        @vol_objs = $self->get_volume_objs();
        foreach my $volume (@vol_objs) {
            next if ($volume->get_name() =~ /_DELETE_/i);
            # Can't delete a root volume unless all 'normal' volumes are deleted
            # first. So, save the object, skip it and delete it later.
            if ($volume->get_name() =~ /rootvol/i) {
                $root_vol = $volume;
                next;
            }
            $volume->delete(-force => 1);
            if ($volume->error()) {
                $self->_set_err("failed to delete volume objects while force " .
                                "destroying the vserver" , $volume->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # Get all routing groups and delete them
        my @routing_groups = $self->get_routing_groups();
        if ($self->error()) {
            $self->_set_err("Unable to get routing groups", $self->errmsg());
            return $SanUtil::FALSE;
        }

        foreach my $rg (@routing_groups) {
            $self->delete_routing_group('routing-group' => $rg);
            if ($self->error()) {
                $self->_set_err("failed to delete routing group",
                                $self->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # Delete the root volume
        if (defined($root_vol)) {
            $root_vol->delete(-force => 1);
            if ($root_vol->error()) {
                $self->_set_err("failed to delete the root volume ",
                                $root_vol->errmsg());
                return $SanUtil::FALSE;
            }
        }
    }

    %args = (%args,
             cmd => "vserver_delete",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to delete a Vserver", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: get_igroup_obj()
# Objective:   Get a specific igroup object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_igroup_obj {
    my ($self, %args) = @_;

    my ($igroup_obj);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert igroup to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{igroup}) {
        $args{name} = $args{igroup};
        delete $args{igroup};
    }

    # verify name param
    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return {};
    }

    # if cache flag is set, get it from cache
    # if it does exist in cache, query the controller for the particular igroup
    if ($self->{use_cache} && exists $self->{igroup_objs}) {
        # get from cache
        ($igroup_obj) = grep ({$_->get_name() =~ /^$args{name}$/} @{$self->{igroup_objs}});
    } else {
        # discover this particular igroup on the controller
        # get igroup objects
        $igroup_obj = $self->get_igroup_objs(-name => $args{name});
        if ($self->error()) {
            $self->_set_err("failed to get igroup object: " . $args{name},
                            $self->errmsg());
            return {};
        }
    }

    unless (ref($igroup_obj) =~ /Igroup/i) {
        $self->_set_err("can not find igroup " . $args{name});
        return {};
    }

    return $igroup_obj;
} ## end sub get_igroup_obj

################################################################################
# Method Name: get_iscsi_obj()
# Objective:   get iscsi object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_iscsi_obj {
    my ($self) = @_;

    my $iscsi_obj = {};

    $self->_clear_err();

    # if iscsi object does exist in cache, get it
    # otherwise, call get_iscsi_objs
    if (defined $self->{iscsi_obj} and ref($self->{iscsi_obj}) =~ /iscsi/i) {
        $iscsi_obj = $self->{iscsi_obj};
    } else {
        # get iscsi obj if it exists, otherwise return empty hash
        if ($self->iscsi_exist()) {
            # get iscsi object
            $iscsi_obj = $self->get_iscsi_objs();
            if ($self->error()) {
                $self->_set_err("failed to get iscsi object", $self->errmsg());
                return {};
            }
        }
    }
    return $iscsi_obj;
} ## end sub get_iscsi_obj

################################################################################
# Method Name: get_fcp_obj
# Objective:   get fcp object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_fcp_obj {
    my ($self) = @_;

    my $fcp_obj = {};

    $self->_clear_err();

    # if fcp object does exist in cache, get it otherwise, call get_fcp_objs
    if (defined $self->{fcp_obj} and ref($self->{fcp_obj}) =~ /fcp/i) {
        $fcp_obj = $self->{fcp_obj};
    } else {
        # get fcp obj if it exists, otherwise return empty hash
        if ($self->fcp_exist()) {
            # get fcp object
            $fcp_obj = $self->get_fcp_objs();
            if ($self->error()) {
                $self->_set_err("failed to get fcp object", $self->errmsg());
                return {};
            }
        }
    }
    return $fcp_obj;
} ## end sub get_fcp_obj

################################################################################
# Method Name: get_lif_obj()
# Objective:   Get a specific lif object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_lif_obj {
    my ($self, %args) = @_;

    my ($lif_obj);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert lif to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{lif}) {
        $args{name} = $args{lif};
        delete $args{lif};
    }

    # verify name param
    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return {};
    }

    # if cache flag is set, get it from cache
    # if it does exist in cache, query the controller for the particular lif
    if ($self->{use_cache} && exists $self->{lif_objs}) {
        # get from cache
        ($lif_obj) = grep ({$_->get_name() =~ /^$args{name}$/} @{$self->{lif_objs}});
    } else {
        # discover this particular lif on the controller
        # get lif objects
        $lif_obj = $self->get_lif_objs(-name => $args{name});
        if ($self->error()) {
            $self->_set_err("failed to get lif object: " . $args{name},
                            $self->errmsg());
            return {};
        }
    }
    
    unless (ref($lif_obj) =~ /Lif/i) {
        $self->_set_err("can not find lif " . $args{name});
        return {};
    }

    return $lif_obj;
} ## end sub get_lif_obj

################################################################################
# Method Name: get_portset_obj()
# Objective:   Get a specific portset object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_portset_obj {
    my ($self, %args) = @_;

    my ($portset_obj);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert portset to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{portset}) {
        $args{name} = $args{portset};
        delete $args{portset};
    }

    # verify name param
    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return {};
    }

    # if cache flag is set, get it from cache
    # if it does exist in cache, query the controller for the particular portset
    if ($self->{use_cache} && exists $self->{portset_objs}) {
        # get from cache
        ($portset_obj) = grep ({$_->get_name() =~ /^$args{name}$/} @{$self->{portset_objs}});
    } else {
        # discover this particular portset on the controller
        # get portset objects
        $portset_obj = $self->get_portset_objs(-name => $args{name});
        if ($self->error()) {
            $self->_set_err("failed to get portset object: " . $args{name},
                            $self->errmsg());
            return {};
        }
    }
    
    unless (ref($portset_obj) =~ /Portset/i) {
        $self->_set_err("can not find portset " . $args{name});
        return {};
    }

    return $portset_obj;
} ## end sub get_portset_obj

################################################################################
# Method Name: get_volume_obj()
# Objective:   Get a specific volume object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volume_obj {
    my ($self, %args) = @_;

    my ($volume_obj);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert volume to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{volume}) {
        $args{name} = $args{volume};
        delete $args{volume};
    }

    # verify name param
    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return {};
    }

    # if cache flag is set, get it from cache
    # if it does exist in cache, query the controller for the particular volume
    if ($self->{use_cache} && exists $self->{volume_objs}) {
        # get from cache
        ($volume_obj) = grep ({$_->get_name() =~ /^$args{name}$/} @{$self->{volume_objs}});
    } else {
        # discover this particular volume on the controller
        # get volume objects
        $volume_obj = $self->get_volume_objs(-name => $args{name});
        if ($self->error()) {
            $self->_set_err("failed to get volume object: " . $args{name},
                            $self->errmsg());
            return {};
        }
    }
   
    unless (ref($volume_obj) =~ /Volume/i) {
        $self->_set_err("can not find volume " . $args{name});
        return {};
    }

    return $volume_obj;
} ## end sub get_volume_obj

###############################################################################
# Method:    get_list_info()
# Objective: get vserver information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_show",
             $self->_get_param_defaults());

    # invoke cmd command and update the attrs
    my @records = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return {};
    } else {
        return $records[0];
    }
} ## end sub get_list_info

################################################################################
# Method Name: get_admin_state()
# Objective:   Get vserver admin state
# Details:     See POD documentation at the end of this file
################################################################################
sub get_admin_state {
    my ($self, %args) = @_;

    $args{field} = "admin_state";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_aggregate()
# Objective:   Get vserver aggregate
# Details:     See POD documentation at the end of this file
################################################################################
sub get_aggregate {
    my ($self, %args) = @_;

    $args{field} = "aggregate";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_language()
# Objective:   Get vserver language
# Details:     See POD documentation at the end of this file
################################################################################
sub get_language {
    my ($self, %args) = @_;

    $args{field} = "language";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_ldap_client()
# Objective:   Get vserver ldap_client
# Details:     See POD documentation at the end of this file
################################################################################
sub get_ldap_client {
    my ($self, %args) = @_;

    $args{field} = "ldap_client";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_name()
# Objective:   Get vserver name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "vserver";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_nisdomain()
# Objective:   Get vserver nisdomain
# Details:     See POD documentation at the end of this file
################################################################################
sub get_nisdomain {
    my ($self, %args) = @_;

    $args{field} = "nisdomain";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_nm_switch()
# Objective:   Get vserver nm_switch
# Details:     See POD documentation at the end of this file
################################################################################
sub get_nm_switch {
    my ($self, %args) = @_;

    $args{field} = "nm_switch";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_ns_switch()
# Objective:   Get vserver ns switch
# Details:     See POD documentation at the end of this file
################################################################################
sub get_ns_switch {
    my ($self, %args) = @_;

    $args{field} = "ns_switch";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get parent object of vserver
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_quota_policy()
# Objective:   Get vserver quota_policy
# Details:     See POD documentation at the end of this file
################################################################################
sub get_quota_policy {
    my ($self, %args) = @_;

    $args{field} = "quota_policy";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_rootvolume()
# Objective:   Get vserver rootvolume
# Details:     See POD documentation at the end of this file
################################################################################
sub get_rootvolume {
    my ($self, %args) = @_;

    $args{field} = "rootvolume";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_rootvolume_security_style()
# Objective:   Get vserver rootvolume_security_style
# Details:     See POD documentation at the end of this file
################################################################################
sub get_rootvolume_security_style {
    my ($self, %args) = @_;

    $args{field} = "rootvolume_security_style";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_snapshot_policy()
# Objective:   Get vserver snapshot_policy
# Details:     See POD documentation at the end of this file
################################################################################
sub get_snapshot_policy {
    my ($self, %args) = @_;

    $args{field} = "snapshot_policy";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_type()
# Objective:   Get vserver type
# Details:     See POD documentation at the end of this file
################################################################################
sub get_type {
    my ($self, %args) = @_;

    $args{field} = "type";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_uuid()
# Objective:   Get vserver uuid
# Details:     See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self, %args) = @_;

    $args{field} = "uuid";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: igroup_exist()
# Objective:   Determines if igroup exists.
# Details:     See POD documentation at the end of this file
################################################################################
sub igroup_exist {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to igroup to maintain consistency
    if (exists $args{name}) {
        $args{igroup} = delete $args{name};
    }

    # verify igroup param
    unless (defined $args{igroup} and length $args{igroup}) {
        $self->_set_err("mandatory parameter \"-igroup\" not specified ");
        return $SanUtil::FALSE;
    }

    # build cmd igroup show
    %args = (%args,
             cmd => "lun_igroup_show",
             vserver => $self->get_name());

    # invoke cdef command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: igroup_list_info()
# Objective: Get igroup information from this vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub igroup_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_igroup_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke_cmd_obj to obtain igroup info from this vserver
    my @records = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # error is already set, bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Using List::MoreUtils::any to find string in an array
    if (List::MoreUtils::any
        {$_ =~ /There are no entries matching/i} @records) {
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @records : \@records;
} ## end sub igroup_list_info

################################################################################
# Method Name: iscsi_exist()
# Objective:   Determines if iscsi exists.
# Details:     See POD documentation at the end of this file
################################################################################
sub iscsi_exist {
    my ($self) = @_;

    my (@output, %args);

    $self->_clear_err();

    # build cmd vserver iscsi show
    %args = (cmd => "vserver_iscsi_show",
             $self->_get_param_defaults());

    # invoke cdef command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: fcp_exist()
# Objective:   Determines if fcp exists.
# Details:     See POD documentation at the end of this file
################################################################################
sub fcp_exist {
    my ($self) = @_;

    my (@output, %args);

    $self->_clear_err();

    # build cmd vserver fcp show
    %args = (cmd => "vserver_fcp_show",
             $self->_get_param_defaults());

    # invoke cdef command but do not update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: lif_exist()
# Objective:   Determines if lif exists.
# Details:     See POD documentation at the end of this file
################################################################################
sub lif_exist {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to lif to maintain consistency
    if (exists $args{name}) {
        $args{lif} = delete $args{name};
    }

    # verify lif param
    unless (defined $args{lif} and length $args{lif}) {
        $self->_set_err("mandatory parameter \"-lif\" not specified ");
        return $SanUtil::FALSE;
    }

    # build cmd network interface show
    %args = (%args,
             cmd => "network_interface_show",
             server => $self->get_name());

    # invoke cdef command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: lif_list_info()
# Objective: Get lif information from this vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub lif_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_show",
             $self->_get_param_defaults());

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    # use base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke_cmd_obj to obtain lif info from this vserver
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # error is already set, bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    unless (scalar(@output)) {
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;
} ## end sub lif_list_info

################################################################################
# Method Name: portset_exist()
# Objective:   Determines if portset exists.
# Details:     See POD documentation at the end of this file
################################################################################
sub portset_exist {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to portset to maintain consistency
    if (exists $args{name}) {
        $args{portset} = delete $args{name};
    }

    # verify portset param
    unless (defined $args{portset} and length $args{portset}) {
        $self->_set_err("mandatory parameter \"-portset\" not specified ");
        return $SanUtil::FALSE;
    }

    # build cmd portset show
    %args = (%args,
             cmd => "lun_portset_show",
             vserver => $self->get_name());

    # invoke cdef command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: portset_list_info()
# Objective: Get portset information from this vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub portset_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_portset_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke_cmd_obj to obtain portset info from this vserver
    my @records = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # error is already set, bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Using List::MoreUtils::any to find string in an array
    if (List::MoreUtils::any
        {$_ =~ /There are no entries matching/i} @records) {
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @records : \@records;
} ## end sub portset_list_info

################################################################################
# Method Name: volume_exist()
# Objective:   Determines if volume exists.
# Details:     See POD documentation at the end of this file
################################################################################
sub volume_exist {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to volume to maintain consistency
    if (exists $args{name}) {
        $args{volume} = delete $args{name};
    }

    # verify volume param
    unless (defined $args{volume} and length $args{volume}) {
        $self->_set_err("mandatory parameter \"-volume\" not specified ");
        return $SanUtil::FALSE;
    }

    # build cmd volume show
    %args = (%args,
             cmd => "volume_show",
             vserver => $self->get_name());

    # invoke cdef command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: volume_list_info()
# Objective: Get volume information from this vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub volume_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke_cmd_obj to obtain volume info from this vserver
    my @records = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # error is already set, bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Using List::MoreUtils::any to find string in an array
    if (List::MoreUtils::any
        {$_ =~ /There are no entries matching/i} @records) {
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @records : \@records;
} ## end sub volume_list_info

###############################################################################
# Method Name: refresh()
# Objective:   Refresh object's data
# Details:     See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    my (@tmp_objs, @vol_objs);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd arguments for aggregate show
    %args = (%args,
             cmd => "vserver_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update vserver object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        $self->_set_err("failed to refresh vserver object data", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # get all volume objects
    @vol_objs = $self->get_volume_objs();
    if ($self->error()) {
        $self->_set_err("failed to get volume objects", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # refresh all volumes from this vserver
    foreach my $vol_obj (@vol_objs) {
        $vol_obj->refresh();
        if ($vol_obj->error()) {
            $vol_obj->_set_err("failed to refresh volume", $vol_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # add to the list
        push @tmp_objs, $vol_obj;
    }

    # store all volume objects from this vserver
    $self->{volume_objs} = \@tmp_objs;

    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: rename
# Objective: Rename this vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub rename {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify newname param
    unless (defined $args{newname} and length $args{newname}) {
        $self->_set_err("mandatory parameter \"-newname\" not specified ");
        return $SanUtil::FALSE;
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_rename",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    } else {
        # rename was successful so we need to update the object with newname
        $self->set(vserver => $args{newname});

    }

    return $SanUtil::TRUE;
} ## end sub rename

###############################################################################
# Method Name: start
# Objective: Start this vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub start {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # if it's running, we don't need to start
    if ($self->get_admin_state() =~ /running/i) {
        return $SanUtil::TRUE;
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_start",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub start

###############################################################################
# Method Name: stop
# Objective: stop this vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub stop {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # if it's already stopped, we don't need to stop
    if ($self->get_admin_state() =~ /stopped/i) {
        return $SanUtil::TRUE;
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_stop",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub stopped

###############################################################################
# Method Name: lun_mapped_list_info()
# Objective: Get a list of mapped luns.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub lun_mapped_list_info {
    my ($self, %args) = @_;
    my (@output);
    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_mapped_show",
             -vserver => $self->get_name());

    @output = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",$self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Using List::MoreUtils::any to find string in an array
    if (List::MoreUtils::any
        {$_ =~ /Table is currently empty/i} @output) {
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## lun_mapped_list_info
##############################################################################
#  DYNAMICALLY CREATED METHODS
##############################################################################

################################################################################
# Method Name: get_igroup_objs()
# Objective:   Get igroup objects. This method is created dynamically from
#              create_accessor_methods in the new() constructor. This is a
#              placeholder in case someone is looking for the method.
# Details:     See POD documentation at the end of this file for usage
################################################################################

################################################################################
# Method Name: get_lif_objs()
# Objective:   Get lif objects. This method is created dynamically from
#              create_accessor_methods in the new() constructor. This is a
#              placeholder in case someone is looking for the method.
# Details:     See POD documentation at the end of this file for usage
################################################################################

################################################################################
# Method Name: get_portset_objs()
# Objective:   Get portset objects. This method is created dynamically from
#              create_accessor_methods in the new() constructor. This is a
#              placeholder in case someone is looking for the method.
# Details:     See POD documentation at the end of this file for usage
################################################################################

################################################################################
# Method Name: get_volume_objs()
# Objective:   Get volume objects. This method is created dynamically from
#              create_accessor_methods in the new() constructor. This is a
#              placeholder in case someone is looking for the method.
# Details:     See POD documentation at the end of this file for usage
################################################################################

##############################################################################
#  PRIVATE
##############################################################################

################################################################################
# Method Name: _get_param_defaults
# Objective:   Gets default Vserver parameters
# Details:     See POD documentation at the end of this file.
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    my %defaults_params = (vserver => $self->{vserver});

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Lif
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # convert common args to args specifically for creating lif
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "vserver_show",
                                                field => $args{field},
                                                $self->_get_param_defaults());
    }
}

###############################################################################
# Method Name: _create_vserver()
# Objective:   Initializes the object by creating vserver on the filer
# Details:     See POD documentation at the end of this file.
###############################################################################
sub _create_vserver {
    my ($self, %args) = @_;

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);
    eval {

        # validate input parameters and their type.
        # filer   - filer name (ssan-6070-1)
        # inherit - flag to inherit existence ConnectionMgr object
        (%args) =
          Module::Common::validate_mixed(@input_params,
                                         [],
                                         {filer   => {type => SCALAR, required => 1},
                                          vserver => {type => SCALAR, required => 1},
                                          aggregate => {type => SCALAR, required => 1},
                                          'rootvolume' => {type => SCALAR,
                                                           optional => 1,
                                                           default  => $args{vserver} . "_rootvol"},
                                          'rootvolume-security-style' => {type => SCALAR,
                                                           optional => 1,
                                                           default  => "mixed"},
                                          'ns-switch'  => {type     => SCALAR,
                                                           optional => 1,
                                                           default  => "file"}});
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # There could be cases where the root volume exists. The library needs to
    # check this before attempting to create the vserver or else ONTAP
    # will complain about duplicate entries.

    #if ($self->volume_exist(volume => $args{rootvolume})) {
    #    $self->_set_err("The root volume " . $args{rootvolume} . " exists." .
    #                    "Please delete it or specify a different root volume.");
    #    return $SanUtil::FALSE;
    #}

    # initialize cmd parameters
    %args = (%args, cmd => "vserver_create");

    # use common invoke from base class (San::CommonInterface)
    # to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # bubble up error
        $self->_set_err("failed to invoke " . $args{cdef},
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # call vserver refresh if enable_cache is set
    if ($self->{enable_cache}) {
        $self->refresh($self->_get_param_defaults());
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to refresh " . $args{vserver}, $self->errmsg());
            return $SanUtil::FALSE;
        }
    }


    # set intitialized flag
    $self->{initialized} = $SanUtil::TRUE;

    return $SanUtil::TRUE;
} ## end sub _create_vserver

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Vserver.pm

=head1 Author

Korry Nguyen (dl-ssan-automation)

=head1 Description

Provides an interface to manage a Vserver

=head1 See Also

Vserver package used base class B<San::CommonInterface> and B<San::ControllerCluster::Filer10M::VserverHelper> that also inherit other modules. See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm, San::SanCache.pm>

=head1 Attributes

=head2 B<filer> (required)

string = controller name

=head2 B<vserver> (required)

string = name of vserver object to be created

=head2 B<aggregate> (required)

string = aggregate name to create vserver on

=head2 B<rootvolume>

string = root volume name for vserver (default is $vserver . "_rootvol")

=head2 B<rootvolume-security-style>

string = security style for vserver root volume (default is "mixed")

=head2 B<ns-switch>

string = Name service switch (default is "file")

=head2 B<debug_level>

interger = 0 - 100 (default is 0)

=head1 B<Public Methods>

=head2 B<new(%args)>

Create a vserver object

=over 4

=item Inputs

=item * I<-filer> => <controller name>

Controller name

=item * I<-vserver> => <name for vserver object>

Vserver name to be created with

=item * I<-aggregate> => <the aggregate name>

Aggregate name to create vserver on

=item * I<-rootvolume> => <vserver root volume name>

Vserver root volume name to be created with

=item * I<-rootvolume> => <vserver root volume name>

Vserver root volume name to be created with

=item * I<'-rootvolume-security-style'> => <security style for vserver root volume>

Vserver root volume security style

=item * I<'-ns-switch'> => <name service switch>

Name service switch type

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. With higher level, it will log more debug messages

=item Outputs

Return Vserver object. If error, error message will pushed onto error stack

=item Examples

 use San::ControllerCluster::Filer10M::Vserver;

 my $filer = "ssan-3170";

 my $vserver_obj = San::ControllerCluster::Filer10M::Vserver->new(
                                                   -filer => $filer,
                                                   -vserver => $vs_name,
                                                   -aggregate => $aggr_name,
                                                   );

 if ($vserver_obj->error) {
     logcomment("Failed to create a vserver object: " . $vserver->errmsg());
     return TCD::FAIL;
 }

=back

=head2 B<create_fcp_obj(%args)>

Create a fcp object. After the object is created, it will be stored in Vserver object.

=over 4

=item Inputs

None

=item Outputs

Returns fcp object. If error, error message will be pushed onto error stack.

=item Examples

 # ex. 1: Create fcp fcp object w/out portset binding or initiator
 $fcp_obj = $vs_obj->create_fcp_obj();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<create_igroup_obj(%args)>

Create a igroup object. After the object is created, it will be stored in Vserver object.

=over 4

=item Inputs

=item * I<-name | -igroup> (required)

The igroup name to be used

=item * I<-ostype> (required)

OS type for the igroup

=item * I<'-protocol-type'> (optional) Default is mixed

Protocol type of the igroup (mixed | iscsi | fcp)

=item * I<-portset> (optional)

Portset binding to this igroup

=item * I<-initiator> (optional)

Initiator nodename

=item Outputs

Returns igroup object. If error, error message will be pushed onto error stack.

=item Examples

 # ex. 1: Create fcp igroup object w/out portset binding or initiator
 $igrp_obj = $vs_obj->create_igroup_obj(-name => $igrp_name,
                                        -ostype => $ostype,
                                       '-protocol-type' => "fcp");
 if ($vs_obj->error()){
     ...
 }

 # ex. 2: Create fcp igroup object with portset binding and initiator
 $igrp_obj = $vs_obj->create_igroup_obj(-name => $igrp_name,
                                        -ostype => $ostype,
                                       '-protocol-type' => "fcp",
                                       -portset' => $ps,
                                       -initiator => "10:00:00:00:c9:43:f3:5c");
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<create_iscsi_obj(%args)>

Create a iscsi object. After the object is created, it will be stored in Vserver object.

=over 4

=item Inputs

None

=item Outputs

Returns iscsi object. If error, error message will be pushed onto error stack.

=item Examples

 # ex. 1: Create fcp iscsi object w/out portset binding or initiator
 $iscsi_obj = $vs_obj->create_iscsi_obj();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<create_lif_obj(%args)>

Create a lif object. After the object is created, it will be stored in Vserver object.

=over 4

=item Inputs

=item * I<-name | -lif> (required)

The lif name to be used

=item * I<-interface | '-home-port'> (required)

Home port for the lif

=item * I<'-home-node'> (optional)

Home node for the lif (default is local controller)

=item * I<-protocol | '-data-protocol'> (optional)

Data protocol for the lif (default is fcp)

=item * I<-address | -ip> (required if protocol is iscsi)

IP address for iscsi lif

=item * I<-netmask> (required if protocol is iscsi)

Netmask for iscsi lif

=item Outputs

Returns lif object. If error, error message will be pushed onto error stack.

=item Examples

 # ex. 1: Create fcp lif object
 $lif_obj = $vs_obj->create_lif_obj(-name => $lif_name,
                                    -protocol => "fcp",
                                    -interface => "5a");
 if ($vs_obj->error()){
     ...
 }

 # ex. 2: Create iscsi lif object
 $lif_obj = $vs_obj->create_lif_obj(-name => $lif_name,
                                    -protocol => "iscsi",
                                    -interface => "e0b",
                                    -ip => $address,
                                    -netmask => $netmask);
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<create_portset_obj(%args)>

Create a portset object. After the object is created, it will be stored in Vserver object.

=over 4

=item Inputs

=item * I<-name | -portset> (required)

The portset name to be used

=item * I<'-protocol-type'> (required) (default is mixed)

Protocol type for the portset (mixed | fcp | iscsi)

=item * I<-tpg> (optional)

iSCSI target portal group

=item * I<-wwpn> (optional)

FC world wide portname

=item * I<'-fc-lif'> (optional)

FC data lif

=item Outputs

Returns portset object. If error, error message will be pushed onto error stack.

=item Examples

 # ex. 1: Create iSCSI portset object
 $vol_obj = $vs_obj->create_portset_obj(-name => $ps_name,
                                        '-protocol-type' => "iscsi",
                                        -tpg => $tpgroup);
 if ($vs_obj->error()){
     ...
 }

 # ex. 2: Create FCP portset object with wwpn
 $vol_obj = $vs_obj->create_portset_obj(-name => $ps_name,
                                        '-protocol-type' => "fcp",
                                        -wwpn => $wwpn);
 if ($vs_obj->error()){
     ...
 }

 # ex. 3: Create FCP portset object with fc lif
 $vol_obj = $vs_obj->create_portset_obj(-name => $ps_name,
                                        '-protocol-type' => "fcp",
                                        '-fc-lif' => $fcp_lif);
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<create_volume_obj(%args)>

Create a volume object. After the object is created, it will be stored in Vserver object.

=over 4

=item Inputs

=item * I<-name | -volume> (required)

The volume name to be used

=item * I<-aggregate> (required)

The aggregate name to create volume on

=item * I<-size> (required)

Volume size

=item * I<'-space-guarantee'> (optional) (default is volume)

Volume space guarantee style (none | volume)

=item * I<'-snapshot-policy'> (optional) (default is default)

Volume snapshot policy

=item Outputs

Returns volume object. If error, error message will be pushed onto error stack.

=item Examples

 # ex. 1: Create volume object
 $vol_obj = $vs_obj->create_volume_obj(-name => $vol_name,
                                       -aggregate => $aggr_name,
                                       -size => "100g");
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<delete()>

Delete this vserver

=over 4

=item Inputs

=item * I<-force> (optional)

Force delete all content within the vserver object

=item Outputs

Returns TRUE if successful. Otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Delete vserver object
 $vol_obj = $vs_obj->delete();
 if ($vs_obj->error()){
     ...
 }

 # ex. 2: Force delete vserver object
 $vol_obj = $vs_obj->delete(-force => 1);
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<get_igroup_obj(%args)>

Get a specific igroup object

=over 4

=item Inputs

=item * I<-name | -igroup> (required)

The igroup name to retrieve

=item Outputs

Returns a igroup object. If error, sets error and returns empty hashref

=item Examples

 # ex. 1: Get igroup object with specific name
 $igroup_obj = $vserver_obj->get_igroup_obj(-name => $igroup_name);
 if ($vserver_obj->error()){
     ...
 }

=back

=head2 B<get_igroup_objs(%args)>

Get multiple igroup objects

=over 4

=item Inputs

=item * I<-name | -igroup> (optional)

The regex matching of igroup name

=item * I<-ostype> (required)

OS type for the igroup

=item * I<'-protocol-type'> (optional)

Protocol type of the igroup (mixed | iscsi | fcp)

=item * I<-portset> (optional)

Portset binding to this igroup

=item * I<-count> (optional)

Number of igroups to retrieve

=item Outputs

Returns array of igroup objects that meet the criteria. If not met, set error and returns empty array

=item Examples

 # ex. 1: Get all igroup objects
 @igroup_objs = $vserver_obj->get_igroup_objs();
 if ($vserver_obj->error()){
     ...
 }

 # ex. 2: Get igroup objects with ostype "windows"
 @igroup_objs = $vserver_obj->get_igroup_objs(-ostype => "windows");
 if ($vserver_obj->error()){
     ...
 }

 # ex. 3: Get exactly 2 igroup objects with the regex matching name
 @igroup_objs = $vserver_obj->get_igroup_objs(-name => "^$igroup_name",
                                          -count => 2);
 if ($vserver_obj->error()){
     ...
 }

=back

=head2 B<get_iscsi_obj()>

Get iscsi object from this vserver

=over 4

=item Inputs

None

=item Outputs

Returns iscsi object. If error, error message will be pushed onto error stack.

=item Examples

 # ex. 1: Get iscsi service object
 $iscsi_obj = $vs_obj->get_iscsi_obj();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<get_fcp_obj()>

Get fcp object from this vserver

=over 4

=item Inputs

None

=item Outputs

Returns fcp object. If error, error message will be pushed onto error stack.

=item Examples

 # ex. 1: Get fcp service object
 $fcp_obj = $vs_obj->get_fcp_obj();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<get_lif_obj(%args)>

Get a specific lif object

=over 4

=item Inputs

=item * I<-name | -lif> (required)

The lif name to retrieve

=item Outputs

Returns a lif object. If error, sets error and returns empty hashref

=item Examples

 # ex. 1: Get lif object with specific name
 $lif_obj = $vserver_obj->get_lif_obj(-name => $lif_name);
 if ($vserver_obj->error()){
     ...
 }

=back

=head2 B<get_lif_objs(%args)>

Get multiple lif objects

=over 4

=item Inputs

=item * I<-name | -lif> (optional)

The regex matching of lif name

=item * I<'-status-admin'> (optional)

The status admin of lifs (up | down)

=item * I<'-status-oper'> (optional)

The status operation of lifs (up | down)

=item * I<'-curr-port'> (optional)

Current port for lif

=item * I<-count> (optional)

Number of lifs to retrieve

=item Outputs

Returns array of lif objects that meet the criteria. If not met, set error and returns empty array

=item Examples

 # ex. 1: Get all lif objects
 @lif_objs = $vserver_obj->get_lif_objs();
 if ($vserver_obj->error()){
     ...
 }

 # ex. 2: Get lif objects with status operation "up"
 @lif_objs = $vserver_obj->get_lif_objs('-status-oper' => "up");
 if ($vserver_obj->error()){
     ...
 }

 # ex. 3: Get exactly 2 lif objects with the regex matching name
 @lif_objs = $vserver_obj->get_lif_objs(-name => "^$lif_name",
                                        -count => 2);
 if ($vserver_obj->error()){
     ...
 }

=back

=head2 B<get_portset_obj(%args)>

Get a specific portset object

=over 4

=item Inputs

=item * I<-name | -portset> (required)

The portset name to retrieve

=item Outputs

Returns a portset object. If error, sets error and returns empty hashref

=item Examples

 # ex. 1: Get portset object with specific name
 $ps_obj = $vserver_obj->get_portset_obj(-name => $ps_name);
 if ($vserver_obj->error()){
     ...
 }

=back

=head2 B<get_portset_objs(%args)>

Get multiple portset objects

=over 4

=item Inputs

=item * I<-name | -portset> (optional)

The regex matching of portset name

=item * I<'-protocol-type'> (optional)

Protocol type for the portset (mixed | fcp | iscsi)

=item * I<-tpg> (optional)

iSCSI target portal group

=item * I<-wwpn> (optional)

FC world wide portname

=item * I<'-fc-lif'> (optional)

FC data lif

=item * I<-count> (optional)

Number of portsets to retrieve

=item Outputs

Returns array of portset objects that meet the criteria. If not met, set error and returns empty array

=item Examples

 # ex. 1: Get all portset objects
 @ps_objs = $vserver_obj->get_portset_objs();
 if ($vserver_obj->error()){
     ...
 }

 # ex. 2: Get portset objects with protocol type "fcp"
 @ps_objs = $vserver_obj->get_portset_objs('-protocol-type' => "fcp");
 if ($vserver_obj->error()){
     ...
 }

 # ex. 3: Get exactly 2 portset objects with the regex matching name
 @ps_objs = $vserver_obj->get_portset_objs(-name => "^$ps_name",
                                           -count => 2);
 if ($vserver_obj->error()){
     ...
 }

=back

=head2 B<get_volume_obj(%args)>

Get a specific volume object

=over 4

=item Inputs

=item * I<-name | -volume> (required)

The volume name to retrieve

=item Outputs

Returns a volume object. If error, sets error and returns empty hashref

=item Examples

 # ex. 1: Get volume object with specific name
 $vol_obj = $vserver_obj->get_volume_obj(-name => $vol_name);
 if ($vserver_obj->error()){
     ...
 }

=back

=head2 B<get_volume_objs(%args)>

Get multiple volume objects

=over 4

=item Inputs

=item * I<-name | -volume> (optional)

The regex matching of volume name

=item * I<-aggregate> (optional)

The aggregate name to create volume on

=item * I<-size> (optional)

Volume size

=item * I<'-space-guarantee'> (optional)

Volume space guarantee style (none | volume)

=item * I<'-snapshot-policy'> (optional)

Volume snapshot policy

=item * I<-count> (optional)

Number of volumes to retrieve

=item Outputs

Returns array of volume objects that meet the criteria. If not met, set error and returns empty array

=item Examples

 # ex. 1: Get all volume objects
 @vol_objs = $vserver_obj->get_volume_objs();
 if ($vserver_obj->error()){
     ...
 }

 # ex. 2: Get volume objects that were created w/ specific aggregate
 @vol_objs = $vserver_obj->get_volume_objs(-aggregate => $aggr_name);
 if ($vserver_obj->error()){
     ...
 }

 # ex. 3: Get exactly 2 volume objects with the regex matching name
 @vol_objs = $vserver_obj->get_volume_objs(-name => "^$vol_name",
                                          -count => 2);
 if ($vserver_obj->error()){
     ...
 }

=back

=head2 B<get_list_info()>

Returns hash reference to vserver show command with key=>value data

=over 4

=item Inputs

None

=item Outputs

Returns hash ref if successful, otherwise sets error and returns empty hash ref

=item Examples

 my $result_ref = $vs_obj->get_list_info();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<get_aggregate()>

Get the aggregate this vserver were created with

=over 4

=item Inputs

None

=item Outputs

Returns aggregate name where the rootvolume of this vserver resided

=item Examples

 # ex. 1: Display aggregate name where this vserver was created with
 logcomment("Aggregate: " . $vs_obj->get_aggregate());

=back

=head2 B<get_language()>

Get vserver language

=over 4

=item Inputs

None

=item Outputs

Returns vserver language

=item Examples

 # ex. 1: Display vserver language
 logcomment("Vserver language: " . $vs_obj->get_language());

=back

=head2 B<get_ldap_client()>

Get vserver ldap_client

=over 4

=item Inputs

None

=item Outputs

Returns vserver ldap_client

=item Examples

 # ex. 1: Display vserver ldap_client
 logcomment("Vserver ldap_client: " . $vs_obj->get_ldap_client());

=back

=head2 B<get_name()>

Get vserver name

=over 4

=item Inputs

None

=item Outputs

Returns vserver name

=item Examples

 # ex. 1: Display vserver name
 logcomment("Vserver name: " . $vs_obj->get_name());

=back

=head2 B<get_nisdomain()>

Get vserver nisdomain

=over 4

=item Inputs

None

=item Outputs

Returns vserver nisdomain

=item Examples

 # ex. 1: Display vserver nisdomain
 logcomment("Vserver nisdomain: " . $vs_obj->get_nisdomain());

=back

=head2 B<get_nm_switch()>

Get vserver name mapping switch

=over 4

=item Inputs

None

=item Outputs

Returns vserver nm_switch

=item Examples

 # ex. 1: Display vserver nm_switch
 logcomment("Vserver nm_switch: " . $vs_obj->get_nm_switch());

=back

=head2 B<get_ns_switch()>

Get vserver name service switch

=over 4

=item Inputs

None

=item Outputs

Returns vserver ns_switch

=item Examples

 # ex. 1: Display vserver ns_switch
 logcomment("Vserver ns_switch: " . $vs_obj->get_ns_switch());

=back

=head2 B<get_parent_obj()>

Get parent object of this vserver

=over 4

=item Inputs

None

=item Outputs

Returns parent object

=item Examples

 # ex. 1: Get parent object of this vserver
 my $filer_obj = $vs_obj->get_parent_obj();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<get_quota_policy()>

Get vserver quota_policy

=over 4

=item Inputs

None

=item Outputs

Returns vserver quota_policy

=item Examples

 # ex. 1: Display vserver quota_policy
 logcomment("Vserver quota_policy: " . $vs_obj->get_quota_policy());

=back

=head2 B<get_rootvolume()>

Get vserver rootvolume

=over 4

=item Inputs

None

=item Outputs

Returns vserver rootvolume

=item Examples

 # ex. 1: Display vserver rootvolume
 logcomment("Vserver rootvolume: " . $vs_obj->get_rootvolume());

=back

=head2 B<get_rootvolume_security_style()>

Get vserver rootvolume_security_style

=over 4

=item Inputs

None

=item Outputs

Returns vserver rootvolume_security_style

=item Examples

 # ex. 1: Display vserver rootvolume_security_style
 logcomment("Vserver rootvolume_security_style: " . $vs_obj->get_rootvolume_security_style());

=back

=head2 B<get_snapshot_policy()>

Get vserver snapshot_policy

=over 4

=item Inputs

None

=item Outputs

Returns vserver snapshot_policy

=item Examples

 # ex. 1: Display vserver snapshot_policy
 logcomment("Vserver snapshot_policy: " . $vs_obj->get_snapshot_policy());

=back

=head2 B<get_type()>

Get vserver type

=over 4

=item Inputs

None

=item Outputs

Returns vserver type

=item Examples

 # ex. 1: Display vserver type
 logcomment("Vserver type: " . $vs_obj->get_type());

=back

=head2 B<get_uuid()>

Get vserver uuid

=over 4

=item Inputs

None

=item Outputs

Returns vserver uuid

=item Examples

 # ex. 1: Display vserver uuid
 logcomment("Vserver uuid: " . $vs_obj->get_uuid());

=back

=head2 B<get_admin_state()>

Get vserver admininstration state

=over 4

=item Inputs

None

=item Outputs

Returns vserver admin state

=item Examples

 # ex. 1: Display vserver admin state
 logcomment("Vserver admin-state: " . $vs_obj->get_admin_state());

=back

=head2 B<fcp_exist()>

Determine whether fcp service exists or not.

=over 4

=item Inputs

None

=item Outputs

Returns TRUE if exists, othervise FALSE

=item Examples

 # ex. 1: Check whether fcp service exists or not
 if ($vs_obj->fcp_exist()) {
     $fcp_obj = $vs_obj->get_fcp_obj();
     if ($vs_obj->error()){
         ...
     }
 } else {
     # Create fcp service
     $fcp_obj = $vs_obj->create_fcp_obj();
     if ($vs_obj->error()){
         ...
     }
 }

=back

=head2 B<iscsi_exist()>

Determine whether iscsi service exists or not.

=over 4

=item Inputs

None

=item Outputs

Returns TRUE if exists, othervise FALSE

=item Examples

 # ex. 1: Check whether iscsi service exists or not
 if ($vs_obj->iscsi_exist()) {
     $fcp_obj = $vs_obj->get_iscsi_obj();
     if ($vs_obj->error()){
         ...
     }
 } else {
     # Create iscsi service
     $fcp_obj = $vs_obj->create_iscsi_obj();
     if ($vs_obj->error()){
         ...
     }
 }

=back

=head2 B<igroup_exist()>

Determine whether igroup exists or not.

=over 4

=item Inputs

=item * I<-name | -igroup> (required)

=item Outputs

Returns TRUE if exists, othervise FALSE

=item Examples

 # ex. 1: Check whether igroup exists or not
 if ($vs_obj->igroup_exist(-name => $igrp_name)) {
     $igrp_obj = $vs_obj->get_igroup_obj(-name => $igrp_name);
     if ($vs_obj->error()){
         ...
     }
 } else {
     $igrp_obj = $vs_obj->create_igroup_obj(-name => $igrp_name,
                                            -ostype => 'linux',
                                            '-protocol-type' => 'fcp');
     if ($vs_obj->error()){
         ...
     }
 }

=head2 B<lif_exist()>

Determine whether lif exists or not.

=over 4

=item Inputs

=item * I<-name | -lif> (required)

=item Outputs

Returns TRUE if exists, othervise FALSE

=item Examples

 # ex. 1: Check whether lif exists or not
 if ($vs_obj->lif_exist(-name => $lif)) {
     $lif_obj = $vs_obj->get_lif_obj(-name => $lif);
     if ($vs_obj->error()){
         ...
     }
 } else {
     # Create fcp lif
     $lif_obj = $vs_obj->create_lif_obj(-name => $lif,
                                        -protocol => 'fcp');
                                        -interface => '0c');
     if ($vs_obj->error()){
         ...
     }
 }

=back

=head2 B<portset_exist()>

Determine whether portset exists or not.

=over 4

=item Inputs

=item * I<-name | -portset> (required)

=item Outputs

Returns TRUE if exists, othervise FALSE

=item Examples

 # ex. 1: Check whether portset exists or not
 if ($vs_obj->portset_exist(-name => $igrp_name)) {
     $ps_obj = $vs_obj->get_portset_obj(-name => $ps_name);
     if ($vs_obj->error()){
         ...
     }
 } else {
     $ps_obj = $vs_obj->create_portset_obj(-name => $ps_name,
                                            '-protocol-type' => 'fcp');
     if ($vs_obj->error()){
         ...
     }
 }

=back

=head2 B<volume_exist()>

Determine whether volume exists or not.

=over 4

=item Inputs

=item * I<-name | -volume> (required)

=item Outputs

Returns TRUE if exists, othervise FALSE

=item Examples

 # ex. 1: Check whether volume exists or not
 if ($vs_obj->volume_exist(-name => $igrp_name)) {
     $vol_obj = $vs_obj->get_volume_obj(-name => $vol_name);
     if ($vs_obj->error()){
         ...
     }
 } else {
     $vol_obj = $vs_obj->create_volume_obj(-name => $vol_name,
                                           -aggregate => $aggr_name,
                                           -size => "100g");
     if ($vs_obj->error()){
         ...
     }
 }

=back

=head2 B<igroup_list_info()>

Get a list of igroup records on this vserver

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing igroup records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # ex. 1: Get a list of igroup records
 my @igroup_records = $vs_obj->igroup_list_info();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<lif_list_info()>

Get a list of lif records on this vserver

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing lif records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # ex. 1: Get a list of lif records
 my @lif_records = $vs_obj->lif_list_info();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<portset_list_info()>

Get a list of portset records on this vserver

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing portset records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # ex. 1: Get a list of portset records
 my @ps_records = $vs_obj->portset_list_info();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<volume_list_info()>

Get a list of volume records on this vserver

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing volume records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # ex. 1: Get a list of volume records
 my @vol_records = $vs_obj->volume_list_info();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<start()>

Start this vserver. If vserver is running, return TRUE.

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns TRUE if successful. Otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Start vserver
 $vs_obj->start();
 if ($vs_obj->error()){
     ...
 }

=back

=head2 B<stop()>

Stop this vserver. If vserver already stopped, return TRUE.

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns TRUE if successful. Otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Stop vserver
 $vs_obj->stop();
 if ($vs_obj->error()){
     ...
 }

 =back

 =head2 B<lun_mapped_list_info(%args)>

Retrieves an array of hashes with the information of each LUN
associated to this particular vserver.

=over 4

=item Inputs

=item * I<-igroup> (optional)

The name of the igroup to which the LUN (or LUNS) are mapped to.
The igroup name will be use to filter the returning information.

=item * I<-volume> (optional)

The name of the volume to which the LUN (or LUNS) belong to.
The volum name will be use to filter the returning information.

=item Outputs

Return an array of hashes containing each lun information.

=item Examples

# example: Retrieve all the LUN information for this Vserver

  @mapped_luns_array = $Vserver_Obj->lun_mapped_list_info();
  if ($Vserver_Obj->error()) {
         ...
  }

  #Go through each element [hash] in the returned array
  foreach $lun_hash_ref (@mapped_luns_array) {
        if (exists $lun_hash_ref->{lun}) {
            Print ("Lun Name :" . $lun_hash_ref->{lun});
            ...
        }
    }

=back
