# Copyright (c) 2009-2010 NetApp Inc.
# All rights reserved.
#
## @summary provides an Fcp object with generic methods that related to fcp
##          commands in C-Mode
## @author jose.porrua@netapp.com (dl-ssan-automation)
## @status Shared

package San::ControllerCluster::Filer10M::Vserver::Fcp;

# Compiler Directives
use strict;

# Imports
use Params::Validate qw( SCALAR ARRAYREF HASHREF OBJECT );

# Inherit base class
use base qw( San::CommonInterface );

###############################################################################
#  PUBLIC METHODS
###############################################################################

###############################################################################
# Method Name: new
# Objective: Create new FCP object (Constructor)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    # Parameter initialization and blessing
    my $self = {};
    bless $self, $class;

    # Remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # Verify mandatory cmd argument
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }

    # Use the default vserver if users did not specify one
    unless (defined $args{vserver} and length $args{vserver}) {
        $args{vserver} = $self->DEFAULT_VSERVER;
    }

    # Use base class (San::CommonInterface) via SUPER::_new method
    # to contruct an FCP object
    $self = $class->SUPER::_new(-name => $args{fcp},
                                -filer => $args{filer},
                                %args);
    if ($self->error()) {
        $self->_set_err("failed to create an FCP object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # Store FCP elements on a list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # Store parent because we will need it to for the fcp lifs
    if (exists $args{parent_obj}) {
        $self->{parent_obj} = delete $args{parent_obj};
    }

    # delete unused argument
    delete $args{debug_level};

    # if the obj has been initialized, just refresh to update info
    # otherwise initialize it by create and update its object
    unless ($args{initialized}) {
        # If an fcp service exists for that vserver and we try to create a
        # it again, the command will fail and invoke will set the flag. We don't
        # want that to happen so we need to make sure it doesn't exist.
        if ($self->_exists(%args)) {
            $self->{initialized} = $SanUtil::TRUE;
        } else {
            # append cdef for fcp create to argument list
            %args = (%args, cmd => "vserver_fcp_create");

            # use common invoke from base class (San::CommonInterface)
            # to translate cdef into UI cmd and execute it
            $self->SUPER::invoke_cmd_obj(%args);
            if ($self->error()) {
                # set error and bubble up
                $self->_set_err("failed to invoke " . $args{cmd},
                                $self->errmsg());
                return $self;
            }

            # call refresh to populate object if enable_cache is set
            if ($self->{enable_cache}) {
                $self->refresh(%args);
                if ($self->error()) {
                    # set error and bubble up
                    $self->_set_err("failed to refresh " . $args{fcp},
                                    $self->errmsg());
                    return $self;
                }
            }

            # Set the flag
            $self->{initialized} = $SanUtil::TRUE;
        }
    }

    return $self;
} # End of new()

###############################################################################
# Method Name: start
# Objective: Starts FCP service on a vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub start {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             raw_data => 1,
             cmd => "vserver_fcp_start",
             vserver => $self->{vserver});

    # Use invoke from common interface to translate cdef into UI
    # cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error() || (scalar(@output) == 0)) {
        $self->_set_err("Failed to invoke " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    if ($self->get_status() =~ /up|true/) {
        # service has started so we need to
        # refresh all iscsi lifs
        $self->_refresh_lifs();
        if ($self->error()) {
            # error has been set, just bubble up
            return $SanUtil::FALSE;
        }
    } else {
        $self->_set_err("failed to start FCP service" .
                        $args{cmd},
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End of start()

###############################################################################
# Method Name: stop
# Objective: Stops FCP service on a vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub stop {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    # Initialize cmd parameters
    %args = (%args,
             raw_data => 1,
             cmd => "vserver_fcp_stop",
             vserver => $self->{vserver});

    # Use invoke from common interface to translate cdef into UI
    # cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error() || (scalar(@output) == 0)) {
        $self->_set_err("Failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    if ($self->get_status() =~ /down|false/) {
        # service has started so we need to
        # refresh all iscsi lifs
        $self->_refresh_lifs();
        if ($self->error()) {
            # error has been set, just bubble up
            return $SanUtil::FALSE;
        }
    } else {
        $self->_set_err("failed to stop FCP service" .
                        $args{cmd},
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End of stop()

###############################################################################
# Method Name: get_initiators
# Objective: Get FCP initiators
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_initiators {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    %args = (%args,
             cmd => "vserver_fcp_initiator_show",
             $self->_get_param_defaults());

    # WORKAROUND.
    # CmdsDrvier is timing out. Code will increase the timeout just in this
    # instruction. Will return to the previous timeout setting.
    #
    # Will:
    # 1. Retrieve the CmdsDriver from CommonInterface.
    # 2. Get the Current TimeOut time
    # 3. Modify the TimeOut time
    # 4. Execute
    # 5. Return to the original TimeOut time.

    # Get cmd driver object
    my $cmd_obj = undef;
    my $org_timeout = '';
    my $new_timeout = '';

    # 1.
    $cmd_obj = $self->SUPER::get_cmd_obj();
    if ($self->error()) {
        $self->_set_err("Failed to retrieve the CmdsDriver object from CommonInterface" .
                        $self->errmsg()
                    );
        return (wantarray()) ? (): [];
    }

    # 2.
    $org_timeout = $cmd_obj->timeout();
    # 3
    $new_timeout = $cmd_obj->timeout("1800");
    $self->sanlog(-msg => "WORKAROUND: CmdDrives Obj TimeOut attribute changed from " .
                  $org_timeout/60 . " mins to ". ($cmd_obj->timeout())/60 . " mins"
                );

    # 4.
    # Use invoke from common interface to translate cdef into UI
    # cmd and execute it
    @output = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("Failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # 5.
    $new_timeout = $cmd_obj->timeout($org_timeout);
    $self->sanlog(-msg => "WORKAROUND: CmdDrives Obj TimeOut attribute changed back to ".
                  ($cmd_obj->timeout())/60 . " mins"
                );

    unless (scalar(@output)) {
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub get_initiators

###############################################################################
# Method Name: get_lifs
# Objective: get fcp logical interfaces
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lifs {
    my ($self) = @_;

    my (@lif_objs);

    $self->_clear_err();

    # we need discover and get lif object using parent object
    if (defined $self->{parent_obj}) {
        # get fcp lifs
        @lif_objs = $self->{parent_obj}->get_lifs(-type => "fcp");
        if ($self->{parent_obj}->error()) {
            unless ($self->{parent_obj}->errmsg() =~ /no lif exists in vserver/i) {
                $self->_set_err("failed to get lif objects",
                                $self->{parent_obj}->errmsg());
            }
            return (wantarray()) ? (): [];
        }
        # cache fcp lifs in fcp object
        $self->{lif_objs} = \@lif_objs;
    } else {
        # no parent object, set error
        $self->_set_err("no parent object in order to get lif objects");
        return (wantarray()) ? (): [];
    }

    return (wantarray()) ? @lif_objs: \@lif_objs;
} ## end sub get_lifs

###############################################################################
# Method Name: nodename
# Objective: Get the FCP nodename on a vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub nodename {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    # Initialize cmd parameters
    %args = (%args,
             raw_data => 1,
             cmd => "vserver_fcp_nodename_get",
             vserver => $self->{vserver});

    # Use invoke from common interface to translate cdef into UI
    # cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error() || (scalar(@output) == 0)) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return '';
    }

    # Traverse the array and get the FCP nodename
    foreach my $record (@output) {
        if ($record =~ /Node Name: (\S+)/i) {
            return $1;
        } elsif ($record =~ /ERROR: command failed:.*/i) {
            $self->_set_err("failed to get FCP nodename" .
                            $args{cmd},
                            $self->errmsg());
            return '';
        }
    }
} # End of nodename()

###############################################################################
# Method Name: status
# Objective: Get the current FCP status
# Details:   See POD documentation at the end of this file.
###############################################################################
sub status {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd     => "vserver_fcp_status",
             $self->_get_param_defaults());

    # Use invoke from common interface to translate cdef into UI
    # cmd and execute it
    my (@records) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # set error, bubble up
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return {};
    }

    # If more than 1 item was returned for this vserver, that's an error.
    if (scalar(@records) > 1) {
        $self->_set_err("Got back more than 1 array element for fcp status!");
        return {};
    }

    # The cli can take status-admin as a filter.  ZAPI cannot do this.
    # Therefore, we check for a filter after the zapi has been called.
    if ($self->get_cmd_interface() =~ /zapi/i and defined $args{'status-admin'})
    {

        # Note that ONTAP verifies the CLI input for status-admin is valid.
        # ZAPI does not do this processing since it does not take status-admin
        # as an input.  So we will process the arg here.
        if ($args{'status-admin'} !~ /^(up|down)$/i) {
            $self->_set_err("status-admin must be 'up' or 'down'");
            return {};
        }

        # If caller is filtering on status-admin up, verify the status is up.
        # Or if caller is filtering on status-admin down, verify the status is down.
        # If the filter is a match, return output[0]
        if ((    ($args{'status-admin'} =~ /^up$/i)
             and ($records[0]->{status_admin} =~ /up/i))
            or (    ($args{'status-admin'} =~ /^down$/i)
                and ($records[0]->{status_admin} =~ /down/i))
          ) {
            return $records[0];
        } else {

            # Else the filter did not match, return nothing.
            return {};
        }

    } ## end if ($self->get_cmd_interface...

    # Else cmd was cli, or zapi without a status-admin filter.
    return $records[0];
} # End of status()

###############################################################################
# Method Name: create
# Objective: Create FCP service on a vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             raw_data => 1,
             cmd => "vserver_fcp_create",
             vserver => $self->{vserver});

    # Use invoke from common interface to translate cdef into UI
    # cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End of create()

###############################################################################
# Method Name: delete
# Objective: Deletes FCP service on a vsserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             raw_data => 1,
             cmd => "vserver_fcp_delete",
             vserver => $self->{vserver});

    # Use invoke from common interface to translate cdef into UI
    # cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Since we deleted the object mark the attribute name as deleted.
    $self->{name} = "_DELETE_";

    return $SanUtil::TRUE;
} # End of delete

###############################################################################
# Method Name: destroy
# Objective: destroy fcp object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self, %args) = @_;

    return $self->delete(%args);
}

###############################################################################
# Method Name: wwpnalias_get
# Objective: Retrieves wwpn by alias or by -wwpn
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wwpnalias_get {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Local variables
    my $alias;
    my $wwpn;

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameters
    foreach (keys %args) {
        if (/-?alias/) {
            if (not length($alias = $args{$_})) {
                $self->_set_err("invalid value for argument '-alias'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?wwpn/) {
            if (not length($wwpn = $args{$_})) {
                $self->_set_err("invalid value for argument '-wwpn'");
                return (wantarray()) ? () : [];
            }
            next;
        } else {
            $self->_set_err("invalid parameter specified \'$_\'");
            return (wantarray()) ? () : [];
        }
    }

    # Can't have both alias and wwpn params
    if (defined($wwpn) and defined($alias)) {
        $self->_set_err("both -alias and -wwpn were specified");
        return (wantarray()) ? () : [];
    }

    %args = (%args,
             cmd => "vserver_fcp_wwpn_alias_show",
             vserver => $self->{vserver});

    # Use invoke from common interface to translate cdef into UI
    # cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} # End of wwpnalias_get()

###############################################################################
# Method Name: wwpnalias_set
# Objective: Sets a new wwpn
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wwpnalias_set {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Local variables
    my $alias;
    my $wwpn;
    my $force;

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameters
    foreach (keys %args) {
        if (/-?alias/) {
            if (not length($alias = $args{$_})) {
                $self->_set_err("invalid value for argument '-alias'");
                return $SanUtil::FALSE;
            }
            next;
        } elsif (/-?wwpn/) {
            if (not length($wwpn = $args{$_})) {
                $self->_set_err("invalid value for argument '-wwpn'");
                return $SanUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("invalid value for argument '-force'");
                return $SanUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    }
    # must have both alias and wwpn params
    if (not defined($wwpn) or not defined($alias)) {
        $self->_set_err("both -alias and -wwpm must be specified");
        return $SanUtil::FALSE;
    }

    %args = (%args,
             cmd => "vserver_fcp_wwpn_alias_set",
             vserver => $self->{vserver});

    # Use invoke from common interface to translate cdef into UI
    # cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End of wwpn_set()

###############################################################################
# Method Name: wwpnalias_remove
# Objective: Removes a wwpn
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wwpnalias_remove {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Local variables
    my $alias;
    my $wwpn;

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameters
    foreach (keys %args) {
        if (/-?alias/) {
            if (not length($alias = $args{$_})) {
                $self->_set_err("invalid value for argument '-alias'");
                return $SanUtil::FALSE;
            }
            next;
        } elsif (/-?wwpn/) {
            if (not length($wwpn = $args{$_})) {
                $self->_set_err("invalid value for argument '-wwpn'");
                return $SanUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    }

    # Can't have both alias and wwpn params
    if (defined($wwpn) and defined($alias)) {
        $self->_set_err("both -alias and -wwpn were specified");
        return $SanUtil::FALSE;
    }

    %args = (%args,
             cmd => "vserver_fcp_wwpn_alias_remove",
             vserver => $self->{vserver});

    # Use invoke from common interface to translate cdef into UI
    # cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End of wwpn_remove()

###############################################################################
# Method Name: wwpnalias_exists
# Objective: Determines if an alias exists or not
# Details:   See POD documentation at the end of this file.
###############################################################################
sub wwpnalias_exists {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Local variables
    my $alias;
    my $wwpn;
    my $force;
    my @response;

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameters
    foreach (keys %args) {
        if (/-?alias/) {
            if (not length($alias = $args{$_})) {
                $self->_set_err("invalid value for argument '-alias'");
                return $SanUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    }

    # Alias is a mandatory parameter
    unless (defined($alias)) {
        $self->_set_err("-alias was not specified");
        return $SanUtil::FALSE;
    }

    @response = $self->wwpnalias_get(-alias=>$alias);

    unless (scalar(@response)) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End of wwpn_exists

###############################################################################
# Method Name: modify
# Objective: Modifies FCP status on a vserver by changing it to up/down
# Details:   See POD Documentation at the end of this file
###############################################################################
sub modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Local Variables
    my $status;

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    # regarding which interface is specified, this method is only for cli
    $args{cmd_interface} = "cli";

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_fcp_modify",
             vserver => $self->{vserver});

    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # we just modified the fcp service; therefore we must refresh its lifs
    # if there are any
    $self->_refresh_lifs();
    if ($self->error()) {
        # error has been set, just bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End of modify

###############################################################################
# Method Name: set_nodename
# Objective: Sets Fcp node name
# Details:   See POD Documentation at the end of this file
###############################################################################
sub set_nodename {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Local Variables
    my ($status, $previous_status, $nodename);

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # convert nodename to target_name
    if (exists $args{nodename}) {
       $args{target_name} = delete $args{nodename};
    }

    # A nodename must have been passed in either from the -nodename or the
    # -target_name parameter
    unless (exists($args{target_name})) {
        $self->_set_err("target_name was not specified");
        return $SanUtil::FALSE;
    }

    # Get the current Fcp status
    $previous_status = $self->get_status();
    if ($self->error()) {
        $self->_set_err("failed to get current Fcp status: " . $self->errmsg());
        return $SanUtil::FALSE;
    }

    if ($previous_status eq 'up') {
        # Need to stop the service before attempting to change the node name.
        $status = $self->stop();
        if ($self->error()) {
            $self->_set_err("failed to stop fcp status.");
            return $SanUtil::FALSE;
        }
    } elsif ($previous_status eq 'unknown') {
        $self->_set_err("fcp status is unknown.");
        return $SanUtil::FALSE;
    }

    # Build the parameters we'll pass to the command
    %args = (%args,
             cmd => "vserver_fcp_nodename",
             cmd_privilege => "advanced",
             vserver => $self->{vserver});

    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd},
			$self->errmsg());
        return $SanUtil::FALSE;
    }

    # Restart FCP service if required
    if ($previous_status eq 'up') {
	$self->sanlog(-msg => "Start FCP Service",
		      -msg_level => 2);
	$self->start();
	if ($self->error()) {
	    $self->_set_err("failed to re-start fcp service ",
			    $self->errmsg());
	    return $SanUtil::FALSE;
	}
    }
    return $SanUtil::TRUE;
} # End of set_nodename

###############################################################################
# Method Name: set_portname
# Objective: Sets the portname
# Details:   See POD Documentation at the end of this file
###############################################################################
sub set_portname {
    my ($self, %args) = @_;

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Check for the required comment argument
    unless (defined $args{wwpn} and length $args{wwpn}) {
        $self->_set_err("Mandatory parameter \"wwpn\" not specified ");
        return $SanUtil::FALSE;
    }
	unless (defined $args{lif} and length $args{lif}) {
        $self->_set_err("Mandatory parameter \"lif\" not specified ");
        return $SanUtil::FALSE;
    }

    # Initialize args parameters
    %args = (%args,
             cmd => "vserver_fcp_portname_set",
             $self->_get_param_defaults());

    # for zapi, remove the force flag attribute until burt486680 is fixed
    # for cli, add cmd_privilege
    if ($self->get_cmd_interface() =~ /cli/i) {
        $args{'cmd_privilege'} = "advanced";
    } else {
        # delete force flag
        delete $args{force} if (exists $args{force});
    }

    # calls invoke_cmd_obj and executes the actual command
    $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Update $self after the set comment
    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh fcp object:", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub set_portname()

###############################################################################
# Method Name: get_portnames
# Objective: Gets the portnames
# Details:   See POD Documentation at the end of this file
###############################################################################
sub get_portnames {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_fcp_portname_show",
             $self->_get_param_defaults());

    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub get_portnames

###############################################################################
# Method Name: get_ports
# Objective: Wrapper for get_interfaces
# Details:   See POD Documentation at the end of this file
###############################################################################
sub get_ports {
    my ($self, %args) = @_;

    return $self->get_interfaces(%args);
}

###############################################################################
# Method Name: get_interfaces
# Objective: Gets the list of interfaces
# Details:   See POD Documentation at the end of this file
###############################################################################
sub get_interfaces {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_fcp_interface_show",
             $self->_get_param_defaults());

    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub get_interfaces

###############################################################################
# Method Name: get_wwpn
# Objective: Gets the wwpn
# Details:   See POD Documentation at the end of this file
###############################################################################
sub get_wwpn {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Local Variables
    my ($adapter_name, $used);
    my @wwpns = ();

    # Remove "-" from arguments
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameters
    foreach (keys %args) {
        if (/adapter|lif/) {
            unless (length($adapter_name = $args{$_})) {
                $self->_set_err("invalid value for argument $_");
                return (wantarray()) ? () : [];
            }
        } elsif (/used/) {
            unless (length($used = $args{$_})) {
                $self->_set_err("invalid value for argument $_");
                return (wantarray()) ? () : [];
            }
            unless (($used == $SanUtil::TRUE) or ($used == $SanUtil::FALSE)) {
                $self->_set_err("invalid value for argument $_: $used");
                return (wantarray()) ? () : [];
            }
        }
        else {
            $self->_set_err("invalid parameter: $_");
            return (wantarray()) ? () : [];
        }
    }

    # Change adapter for lif just in case the users specified it
    if (defined($args{adapter})) {
        $args{lif} = delete $args{adapter};
    }

    # Build the parameters we'll pass to the command
    %args = (%args,
             raw_data => $SanUtil::TRUE,
             cmd      => "vserver_fcp_portname_show",
             vserver  => $self->{vserver});

    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke " . $args{cmd}, $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Parse the output and see if the table is empty.
    if (List::MoreUtils::any
        {$_ =~ /no entries matching/i} @output) {
        return (wantarray()) ? () : [];
    }

    # right now, @output is a single element array, we need to convert it to
    # string then back to an array in order for every line to come out properly
    # as individual elements in the array.
    @output = split(/\n+/, join("\n", @output));

    # Get the wwpns from the list and push it into an array.
    List::MoreUtils::apply { push(@wwpns, /([0-9a-f]{2}:.+[0-9a-f]{2})/)
                           } @output;

    return (wantarray()) ? @wwpns : \@wwpns;
} # End of get_wwpn

################################################################################
# Method Name: get_nodename()
# Objective:   wrapper to fcp target name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_nodename {
    my ($self, %args) = @_;

    return $self->get_target_name(%args);
}

################################################################################
# Method Name: get_target_name()
# Objective:   Get fcp nodename/target name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_target_name {
    my ($self, %args) = @_;

    $args{field} = "target_name";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get parent object of this Fcp
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get parent object from cache
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_status()
# Objective:   Get fcp status
# Details:     See POD documentation at the end of this file
################################################################################
sub get_status {
    my ($self, %args) = @_;

    $args{field} = "status_admin";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_vserver()
# Objective:   Get vserver this fcp belong to
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vserver {
    my ($self, %args) = @_;

    $args{field} = "vserver";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method Name: refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # we don't need args{fcp} here
    delete $args{fcp} if (exists $args{fcp});

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_fcp_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update fcp object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: get_stats
# Objective: Retrieves all or the specified statistic information in an array
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_stats {
    my ($self, %args) = @_;
    my (@stats);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd_privilege => "diagnostic",
             cmd => "network_fcp_adapter_stats");

    (@stats) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj: " . $args{cmd} . "\n" .
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @stats : \@stats;

} ## end get_stats

###############################################################################
# PRIVATE METHODS
###############################################################################

###############################################################################
# Method Name: _exists
# Objective: Determines if the fcp service exists
# Details:   See POD.
###############################################################################
sub _exists {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Get FCP status
    my $status =  $self->status(%args);

    # Anything other than down or up means that it doesn't exist yet
    if ($status eq 'down' or $status eq 'up') {
        return $SanUtil::TRUE;
    } else {
        # Clear the error flag here because $self->status would set it
        $self->_clear_err();
        return $SanUtil::FALSE;
    }
} # End of _exists

###############################################################################
# Method Name: _refresh_lifs
# Objective: Refresh lif objects. This will be called when service is started/
#            stopped/modified
# Details:   See POD.
###############################################################################
sub _refresh_lifs {
    my ($self) = @_;

    $self->_clear_err();

    # Get FCP lifs
    my @lifs =  $self->get_lifs();
    if ($self->error()) {
        # if the error was set because there is no lif, let's ignore
        unless ($self->errmsg() =~ /no lif exists in vserver/i) {
            $self->_set_err("failed to get lif objects ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # Let's refresh each lif
    foreach my $lif (@lifs) {
        # ignore lifs that are already been deleted
        next if ($lif->get_name() =~ /_DELETE_/i);

        $lif->refresh();
        if ($lif->error()) {
            # if the error was set because there is no lif, let's ignore
            unless ($lif->errmsg() =~ /record does not exist/i) {
                $self->_set_err("failed to refresh lif ", $lif->errmsg());
                return $SanUtil::FALSE;
            }
        }
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: _get_param_defaults
# Objective: Gets default igroup parameters
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    # default hash parameters to be used in igroup object
    # get accessor method must obtain from cache to prevent infinite loop
    my %defaults_params = (vserver => $self->get_vserver(-use_cache => 1));

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get Fcp attribute from the object cache or the controller
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # convert common args to args specifically for creating lif
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "vserver_fcp_show",
                                                field => $args{field},
                                                $self->_get_param_defaults());
    }
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Fcp.pm

=head1 Author

Jose Porrua (dl-ssan-automation)

=head1 Description

Fcp.pm provides an API for FCP commands in C-Mode

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<San::CommonInterface>

=head1 Attributes

=head2 B<filer> (required)

string = controller name

=head2 B<vserver> (required)

string = vserver name to create Fcp

=head1 Public Methods

=head2 B<new(%args)>

Creates a new FCP object.  Uses San::CommonInterface to create the object. Users
may pass as a parameter a vserver. If a vserver is not specified, it uses 'vs0'
as default.

=over 4

=item Inputs

=item * -filer

Name of filer

=item * I<-vserver> (default: 'vs0').

Name of the vserver to create Fcp.

=item Outputs

Returns an FCP object. The error flag will be set if something fails.

=item Examples

 # Create an FCP object using vserver 'vs1'

 $fcp_obj = new San::ControllerCluster::Filer10M::Vserver::Fcp(filer => $filer,
                                                               vserver=> 'vs1');

 if ($fcp_obj->error()) {
    my $errmsg = join "\n", $fcp_obj->errmsg();
    $Test->nlog("ERROR: Failed to instantiate FCP object " . $errmsg);
    return $TCD::FAIL;
 }

=back

=head2 B<start()>

Starts the FCP service on the vserver. This method will set the error flag
if the command failed.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the service started successfully, FALSE otherwise.

=item Examples

 my $status = $fcp_obj->start();

 if ($fcp_obj->error()) {
    my $errmsg = join "\n", $fcp_obj->errmsg();
    $Test->nlog("ERROR: Failed to start FCP service " . $errmsg);
    return $TCD::FAIL;
 }

=back

=head2 B<stop()>

Stops the FCP service on the vserver. This method will set the error flag
if the command failed.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the service stopped successfully, FALSE otherwise.

=item Examples

 my $status = $fcp_obj->stop();

 if ($fcp_obj->error()) {
    my $errmsg = join "\n", $fcp_obj->errmsg();
    $Test->nlog("ERROR: Failed to stop FCP service " . $errmsg);
    return $TCD::FAIL;
 }

=back

=head2 B<get_initiators()>

Gets a list of fcp initiators records on the vserver.

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing initiators records. If an error occurs, sets the error flag and
returns an empty list.

=item Example1:

    my @initiators = $fcp_obj->get_initiators();

    if ($fcp_obj->error()) {
        $err = $fcp_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=item Example2:

    my $initiators = $fcp_obj->get_initiators();
    my @records = @$initiators;

    if ($fcp_obj->error()) {
        $err = $fcp_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<status()>

Get the status of the FCP service on a vserver.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the service is running or FALSE if it's stopped.

=item Examples

 my $status = $fcp_obj->status();

 if ($fcp_obj->error()) {
    my $errmsg = join "\n", $fcp_obj->errmsg();
    $Test->nlog("ERROR: Failed to get the FCP status" . $errmsg);
    return $TCD::FAIL;
 }

 Tharn::logcomment("FCP STATUS: ". $status);

=back

=head2 B<nodename()>

Get the FCP node name on a vserver.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the node name or an empty string if it doesn't exist.

=item Examples

  my $nodename = $fcp_obj->nodename();

  if ($fcp_obj->error()) {
      my $errmsg = join "\n", $fcp_obj->errmsg();
      $Test->nlog("ERROR: Failed to get FCP nodename " . $errmsg);
      return $TCD::FAIL;
  }

  Tharn::logcomment("NODE NAME: ". $nodename);

=back

=head2 B<create()>

Creates an FCP service on a vserver. If the service exists, the command will
fail and hence the error flag will be set.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the FCP service was created successfully, FALSE otherwise.

=item Examples

  my $output = $fcp_obj->create();

  if ($fcp_obj->error()) {
        my $errmsg = join "\n", $fcp_obj->errmsg();
        $Test->nlog("ERROR: Failed to create FCP " . $errmsg);
        return $TCD::FAIL;
  }

  Tharn::logcomment("FCP Create: ". Dumper($output));

=back

=head2 B<delete()>

Deletes an FCP service on a vserver. If the service does not exist, the command
will fail and hence the error flag will be set.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the fcp service was deleted successfully, FALSE otherwise.

=item Examples

  my $output = $fcp_obj->delete();

  if ($fcp_obj->error()) {
        my $errmsg = join "\n", $fcp_obj->errmsg();
        $Test->nlog("ERROR: Failed to delete FCP " . $errmsg);
        return $TCD::FAIL;
  }

  Tharn::logcomment("FCP delete: ". Dumper($output));

=back

=head2 B<destroy()>

Deletes an FCP service on a vserver. If the service does not exist, the command
will fail and hence the error flag will be set.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the fcp service was deleted successfully, FALSE otherwise.

=item Examples

  my $output = $fcp_obj->destroy();

  if ($fcp_obj->error()) {
        my $errmsg = join "\n", $fcp_obj->errmsg();
        $Test->nlog("ERROR: Failed to delete FCP " . $errmsg);
        return $TCD::FAIL;
  }

  Tharn::logcomment("FCP destroy: ". Dumper($output));

=back

=head2 B<wwpnalias_get()>

Get WWPN alias

=over 4

=item Inputs

=item * I<-alias> (optional)

Getting WWPN that mapped to the specified alias

=item * I<-wwpn> (optional)

Get alias(es) that associated with specified wwpn

=item Outputs

Return either an array or a reference to an array (depending on the context
it is called) containing the wwpn mapped to a specified alias

=item Examples

  my $output = $fcp_obj->fwwpnalias_get();

  if ($fcp_obj->error()) {
        my $errmsg = join "\n", $fcp_obj->errmsg();
        $Test->nlog("ERROR: Failed to get wwpn " . $errmsg);
        return $TCD::FAIL;
  }

  Tharn::logcomment("FCP get wwpn: ". Dumper($output));

=back

=head2 B<wwpnalias_set()>

Create an alias and map to the given wwpn

=over 4

=item Inputs

=item * I<-alias> (optional)

Name of a alias.

=item * I<-wwpn> (optional)

Name of a wwpn.

=item Outputs

Returns TRUE if it was successful or FALSE otherwise.

=item Examples

  my $output = $fcp_obj->fwwpnalias_set(-alias => "test");

  if ($fcp_obj->error()) {
        my $errmsg = join "\n", $fcp_obj->errmsg();
        $Test->nlog("ERROR: Failed to set alias " . $errmsg);
        return $TCD::FAIL;
  }

  Tharn::logcomment("FCP set wwpn: ". Dumper($output));

=back

=head2 B<wwpnalias_remove()>

Removes an alias and unmaps it from a wwpn

=over 4

=item Inputs

=item * I<-alias> (optional)

Name of a alias.

=item * I<-wwpn> (optional)

Name of a wwpn.

=item Outputs

Returns TRUE if it was successful or FALSE otherwise.

=item Examples

  my $output = $fcp_obj->fwwpnalias_remove(-alias => "test");

  if ($fcp_obj->error()) {
        my $errmsg = join "\n", $fcp_obj->errmsg();
        $Test->nlog("ERROR: Failed to remove alias " . $errmsg);
        return $TCD::FAIL;
  }

  Tharn::logcomment("FCP remove wwpn: ". $output);

=back

=head2 B<wwpnalias_exists()>

Determines if an alias exists

=over 4

=item Inputs

=item * I<-alias> (optional)

Name of a alias.

=item Outputs

Returns TRUE if it was successful or FALSE otherwise.

=item Examples

  my $output = $fcp_obj->wwpnalias_exists(-alias => "test");

  if ($fcp_obj->error()) {
        my $errmsg = join "\n", $fcp_obj->errmsg();
        $Test->nlog("ERROR: Failed to set alias " . $errmsg);
        return $TCD::FAIL;
  }

  Tharn::logcomment("FCP exists: ". $output);

=back

=head2 B<modify()>

Modifies Fcp Object on vserver.
See http://wikid.netapp.com/w/Smokejumper/Bcom/Documents/Protocols_Mgmt_FS#fcp_modify for details.

=over 4

=item Inputs

=item * status_admin: (optional) "up" or "down" to change the status of the service.

=item * nodename: (optional) The nodename of the FCP service (advanced priv level)

=item * force: (optional) An indicator to accept the WWNN if it deviates from NetApp format standards. (advanced priv level)

=item Outputs

Returns TRUE if fcp was modified successfully, FALSE otherwise.

=item Examples

  my $status = $fcp_obj->modify(-status_admin => "up");

  if ($fcp_obj->error()) {
        my $errmsg = join "\n", $fcp_obj->errmsg();
        $Test->nlog("ERROR: Failed to bring FCP up" . $errmsg);
        return $TCD::FAIL;
  }

  Tharn::logcomment("Fcp status was changed successfully);

=back

=head2 B<set_nodename()>

Changes the fcp nodename. This function will stop fcp service to change the
node name and will restore the service to its initial state once it's done.
It will also set the advanced priviledges to run the command and switch back
to "test" priviledges when it's done.

=over 4

=item Inputs

=item * target_name: A valid FCP nodename.

=item Outputs

Returns TRUE if fcp nodename was changes successfully, FALSE otherwise.

=item Examples

  my $status = $fcp_obj->set_nodename(target_name => "21:12:00:a0:98:0d:d3:84");

  if ($fcp_obj->error() or $status == 0) {
        my $errmsg = join "\n", $fcp_obj->errmsg();
        $Test->nlog("ERROR: Failed to set a new Fcp nodename: " . $errmsg);
        return $TCD::FAIL;
  }

  Tharn::logcomment("Fcp nodename was changed successfully);

=back

=head2 B<set_portname()>

Sets the fcp portname. This function will offline the lif to change the
port name.
It will also set the advanced priviledges to run the command and switch back
to "test" priviledges when it's done.

=over 4

=item Inputs

=item * lif: A lif name.

=item * wwpn: A valid wwpn.

=item * force (optional): Forces the set, even if wwpn is invalid..

=item Outputs

Returns TRUE if fcp portname was changed successfully, FALSE otherwise.

=item Examples

  my $status = $fcp_obj->set_portname(wwpn  => "23:77:00:0c:29:c1:59:ca",
                                      lif   => $lif_name,
                                      force => "true");

  if ($fcp_obj->error() or $status == 0) {
        my $errmsg = join "\n", $fcp_obj->errmsg();
        $Test->nlog("ERROR: Failed to set a new Fcp portname: " . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<get_portnames()>

Gets a list of WWPNs used by local Fibre Channel target LIFs and
names of the corresponding LIFs.

=over 4

=item Inputs

=item * vserver (optional): virtual server.

=item * lif (optional): A lif name.

=item * wwpn (optional): wwpn.

=item Outputs

Returns a list of WWPNs used by fcp LIFs.

=item Examples

  my @portnames = $fcp_obj->get_portnames();
  if ($fcp_obj->error() or $status == 0) {
        $Test->nlog("ERROR: Failed to get portnames ");
        return $TCD::FAIL;
  }

=back

=head2 B<get_interfaces()>

Displays FCP LIFs

=over 4

=item Inputs

=item * vserver (optional): Virtual Server name.

=item * lif (optional): LIF name.

=item * wwpn (optional): WWPN.

=item * wwnn (optional): WWNN.

=item * status-admin (optional): Admin Status.

=item * status-oper (optional): Operational Status.

=item * curr-node (optional): Current node.

=item * curr-port (optional): Current port.

=item * is-home (optional): Is home.

=item Outputs

Returns a list of adapters and their information.

=item Examples

  my @adapters = $fcp_obj->get_interfaces();
  if ($fcp_obj->error() or $status == 0) {
        $Test->nlog("ERROR: Failed to get ports ");
        return $TCD::FAIL;
  }

=back

=head2 B<get_wwpn()>

Gets the wwpns on a vserver by executing fcp potname show and parsing the
output. If a user specifies -lif, it will only return the wwpn corresponding to
that wwpn.

=over 4

=item Inputs

=item * lif (optional): A lif name.

* used (optional): If a user specifies this flag, it returns those wwpns
that are being used.

=item Outputs

Returns a list of a reference to a list containing wwpn.

=item Examples

  my @wwpns = $fcp_obj->get_wwpn();
  if ($fcp_obj->error() or $status == 0) {
        $Test->nlog("ERROR: Failed to set the wwpn ");
        return $TCD::FAIL;
  }

=back

=head2 B<get_nodename()>

Get fcp nodename from object cache or controller.

=over 4

=item Inputs

=item * I<-use_cache> (optional)

Flag to retrieve nodename from cache if set to true. Default is false.

=item Outputs

Returns nodename (ex. 29:26:00:a0:98:08:68:92)

=item Examples

 # ex. 1: Display nodename
 logcomment("Fcp nodename: " . $fcp_obj->get_nodename());

=back

=head2 B<get_target_name()>

Get fcp target name from object cache or controller.

=over 4

=item Inputs

=item * I<-use_cache> (optional)

Flag to retrieve fcp target name from cache if set to true. Default is false.

=item Outputs

Returns target name (ex. 29:26:00:a0:98:08:68:92)

=item Examples

 # ex. 1: Display target name
 logcomment("Fcp target name: " . $fcp_obj->get_target_name());

=back

=head2 B<get_status()>

Get fcp status from object cache or controller.

=over 4

=item Inputs

=item * I<-use_cache> (optional)

Flag to retrieve status from cache if set to true. Default is false.

=item Outputs

Returns status

=item Examples (ex. up or down)

 # ex. 1: Display status
 logcomment("Fcp status: " . $fcp_obj->get_status());

=back

=head2 B<get_vserver()>

Get vserver this fcp belongs to from object cache or controller.

=over 4

=item Inputs

=item * I<-use_cache> (optional)

Flag to retrieve vserver name from cache if set to true. Default is false.

=item Outputs

Returns vserver name

=item Examples

 # ex. 1: Display vserver name
 logcomment("Vserver: " . $fcp_obj->get_vserver());

=back

=head2 B<get_stats()>

Retrieves the statistics for the Fiber Channel Protocol

=over 4

=item Inputs

=item * I<-vserver> (optional)

Specify the vserver that the FC stats will be returned

=item * I<-fields> (optional)

Specify what information is to be returned.

=item Outputs

Retrieves an Array with a Hash containing all the statistics of the FCP
or just the specified by the input 'fields'.

=item Example1:

    my @stats = $fcp_obj->get_stats();
    if ($fcp_obj->error()) {
        ...
    }

=back

=cut
