# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to manage FCP Target object
##
## @author korry@netapp.com (dl-ssan-test)
## @status public

package San::ControllerCluster::Filer10M::Vserver::Fcptarget;

# compiler directives
use strict;

use Params::Validate qw( SCALAR ARRAYREF HASHREF OBJECT );

# inherit base class
use base qw( San::CommonInterface );

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Method Name: get_fcp_obj()
# Objective:   Get fcp_obj
# Details:     See POD documentation at the end of this file
################################################################################
sub get_fcp_obj {
    my ($self) = @_;

    # the get() will handle error
    return $self->get("fcp_obj");
}

################################################################################
# Method Name: get_name()
# Objective:   Get name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    # the get() will handle error
    return $self->get("name");
}

################################################################################
# Method Name: get_nodename()
# Objective:   Get nodename
# Details:     See POD documentation at the end of this file
################################################################################
sub get_nodename {
    my ($self) = @_;

    # the get() will handle error
    return $self->get("nodename");
}

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get parent object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self) = @_;

    # the get() will handle error
    return $self->get("parent_obj");
}

################################################################################
# Method Name: get_state()
# Objective:   Get state
# Details:     See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;

    # the get() will handle error
    return $self->get("state");
}

################################################################################
# Method Name: is_mgmt_port()
# Objective:   determine Fcptarget management port?
# Details:     See POD documentation at the end of this file
################################################################################
sub is_mgmt_port {
    my ($self) = @_;

    # the get() will handle error
    return $self->get("is_mgmt_port");
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

###############################################################################
# Method Name: _new
# Objective: Create new FCP Target object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, @inits);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{filer} and length $args{filer}) {
        if (defined $args{parent_obj}) {
            $args{filer} = $args{parent_obj}->get_filer();
        } else {
            $self->_set_err("missing \"-parent\" parameters");
            return $self;
        }
    }

    # support both name and fcptarget
    # switch to fcptarget now because SUPER::_new is being used by all objects
    # and cdef requires fcptarget instead of name
    if (exists $args{name}) {
        $args{fcptarget} = $args{name};
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => $self->dumper_get(\%args),
                  -msg_level => 3);

    # use base class (San::CommonInterface) via SUPER::_new method 
    # to contruct an fcptarget object
    $self = $class->SUPER::_new(-name  => $args{fcptarget},
                                -filer => $args{filer});

    if ($self->error()) {
        $self->_set_err("failed to create an fcp target object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);


    # store/initalize fcp target elements
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # now we store and create accessor methods, delete the one we don't need
    delete $args{name};
    delete $args{parent_obj};

    # get initiators
    @inits = $self->_get_initiators();
    if ($self->error()) {
        $self->_set_err("failed to get initiators", $self->errmsg());
        return $self;
    }
    $self->{inits} = \@inits;
    $self->{initcnt} = scalar(@inits);

    return $self;
} ## end sub _new

################################################################################
# Method Name: _get_param_defaults
# Objective: Gets default fcp target parameters 
# Details:   See POD documentation at the end of this file. 
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    # default hash parameters to be used in igroup object
    my %defaults_params = (vserver => $self->{parent_obj}->get_name(-use_cache => 1));

    return %defaults_params;
}

################################################################################
# Method Name: _get_initiators
# Objective: Gets initiators to into fcp target object
# Details:   See POD documentation at the end of this file. 
################################################################################
sub _get_initiators {
    my ($self) = @_;

    my (@inits, @tmp_inits);
 
    @inits = ();

    # we need to call initiator show to get all initiators that tags to 
    # target portal group
    if (defined $self->{fcp_obj} and ref($self->{fcp_obj}) =~ /fcp/i) {
        # get an list of initiators and its key/value attributes
        @tmp_inits = $self->{fcp_obj}->get_initiators();
        if ($self->{fcp_obj}->error()) {
            $self->_set_err("failed to get initiators from fcp object",
                            $self->{fcp_obj}->errmsg());
            return (wantarray()) ? () : [];
        }

        # traverse through all and get the initiator name
        foreach my $init_hash (@tmp_inits) {
            if (defined $init_hash->{initiator} and $init_hash->{initiator} =~ /^\d\d:\d\d:/) {
                push @inits, $init_hash->{initiator};
            }
        }
    } else {
        $self->sanlog(-msg => "no fcp object to get initiators");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @inits : \@inits;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Fcptarget.pm

=head1 Author

Korry Nguyen (dl-ssan-automation)

=head1 Description

Provides an interface to manage an Fcp target. The package should be instantiated via get_fcp_targets().

=head1 See Also

Fcptarget package used base class B<San::CommonInterface> that also inherits other modules. See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm, San::SanCache.pm>

=head1 Attributes

=head2 B<name>

string = Name of the lif target

=head2 B<parent_obj>

string = Parent object

=head2 B<fcp_obj>

string = FCP object

=head2 B<nodename>

string = Name of the target nodename

=head2 B<state>

string = state of the lif

=head1 B<Public Methods>

=head2 B<get_name()>

Get lif target name

=over 4

=item Inputs

None

=item Outputs

Return target lif name

=item Examples

 # ex. 1: get fcp target name
 $target_name = $fcptarget_obj->get_name();
 if (fcptarget_obj->error()) {
     ...
 }

=back

=head2 B<get_parent_obj()>

Get parent object

=over 4

=item Inputs

None

=item Outputs

Return parent object of Fcptarget (Vserver object)

=item Examples

 # ex. 1: get parent object
 $vs_obj = $fcptarget_obj->get_parent_obj();
 if (fcptarget_obj->error()) {
     ...
 }

=back

=head2 B<get_fcp_obj()>

Get FCP object

=over 4

=item Inputs

None

=item Outputs

Return FCP object

=item Examples

 # ex. 1: get fcp object
 $fcp_obj = $fcptarget_obj->get_fcp_obj();
 if (fcptarget_obj->error()) {
     ...
 }

=back

=head2 B<get_nodename()>

Get FCP nodename

=over 4

=item Inputs

None

=item Outputs

Return FCP nodename

=item Examples

 # ex. 1: get fcp nodename
 $nodename = $fcptarget_obj->get_nodename();
 if (fcptarget_obj->error()) {
     ...
 }

=back

=head2 B<get_state()>

Get state of the target lif

=over 4

=item Inputs

None

=item Outputs

Return state of target lif (ex. up or down)

=item Examples

 # ex. 1: get state of target lif
 $state = $fcptarget_obj->get_state();
 if (fcptarget_obj->error()) {
     ...
 }

=back

