# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to an Igroup object
##
## @author dl-ssan-test
## @status public

package San::ControllerCluster::Filer10M::Vserver::Igroup;

# compiler directives
use strict;

# inherit base class
use base qw( San::CommonInterface );

###############################################################################
# Method Name: new
# Objective: Create new IGroup object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, @nodes, $client_obj, $ostype);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    if (exists $args{name}) {
        $args{igroup} = $args{name};
        delete $args{name};
    }

    # verify mandatory cmd argument
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }

    # display argument hash if debug_level is high
    $self->sanlog(-msg => $self->dumper_get(\%args),
                  -msg_level => 3);

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct an igroup object
    $self = $class->SUPER::_new(-name  => $args{igroup},
                                -filer => $args{filer});

    if ($self->error()) {
        $self->_set_err("failed to create an igroup object", $self->errmsg());
        return $self;
    }

    # dump its hash parameters if debug_level is set
    $self->sanlog(-msg => $self->dumper_get(\%args),
                  -msg_level => 3);

    # if client handle or init objs are passed in, we need to store the
    # necessary parameters
    if (defined $args{initObjs}) {
        $client_obj = $args{initObjs}->[0]->{parent};

        unless (defined $client_obj and length $client_obj) {
            $self->_set_err("failed to obtain client object from init obj" , $self->errmsg());
            return $self;
        }
        # store client obj
        $self->{client} = $client_obj;

        # let's set the required params for igroup create
        $args{'protocol'} = ($self->{client}->get_protocol() =~ /fcp/i) ? "fcp" : "iscsi";
        # we need to match hosttype and set the correct os so filer can
        # interpret
        $ostype = $self->{client}->hosttype();
        if ($ostype =~ /solaris/i) {
            $args{ostype} = "solaris";
        } elsif ($ostype =~ /windows/i) {
            # we need to handle windows_gpt,etc.
            $args{ostype} = "windows";
        } elsif ($ostype =~ /linux/i) {
            $args{ostype} = "linux";
        }

        foreach my $init_obj (@{$args{initObjs}}) {
            my $nodename;
            if ($args{'protocol'} =~ /fcp/i) {
                $nodename = $init_obj->get_portname();
            } elsif ($args{'protocol'} =~ /iscsi|mixed/i) {
                $nodename = $init_obj->get_nodename();
            }
            if ($init_obj->error()) {
                $self->_set_err("failed to get nodename ", $init_obj->errmsg());
                return $self;
            }
            push(@nodes, $nodename);
        }
        $args{initiator} = join(',', @nodes);

        # delete initObjs param that we don't need
        delete $args{initObjs};
    }

    # store iGroup elements on a list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # we may need the parent object later
    # at this time, we pass it in but unused
    # delete unneeded argument from the list
    delete $args{parent_obj};

    # if the obj has been initialized, elements has been updated so just return $self
    unless ($args{initialized}) {
        # send create igroup command to system and refresh object
        $self->_create_igroup(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to _create_igroup() ", $self->errmsg());
            return $self;
        }
        $self->{initialized} = $SanUtil::TRUE;
    }

    return $self;
} ## end sub new

###############################################################################
# Method Name: add
# Objective: add initiator to an igroup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # just ensure initiator is passed in. we don't care about format
    unless (defined $args{initiator} and length($args{initiator})) {
        $self->_set_err("invalid -initiator param");
        return $SanUtil::FALSE;
    }


    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_igroup_add",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub destroy

###############################################################################
# Method Name: delete
# Objective: delete an igroup object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_igroup_delete",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    $self->{name} = "_DELETE_";
    $self->{igroup} = "_DELETE_";

    return $SanUtil::TRUE;
} ## end sub delete

###############################################################################
# Method Name: destroy
# Objective: destroy an igroup object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self, %args) = @_;

    return $self->delete(%args);
}

###############################################################################
# Method Name: refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_igroup_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update igroup object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: remove()
# Objective: remove initiator from mapping to an igroup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_igroup_remove",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub remove

###############################################################################
# Method Name: rename()
# Objective: rename igroup to a new name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub rename {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # replace "-" with "_" 
    if (exists $args{'new_name'}) {
        $args{'new-name'} = delete $args{'new_name'};
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_igroup_rename",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # update the new name
    $self->{'igroup'} = $args{'new-name'};
    return $SanUtil::TRUE;
} ## end sub rename

###############################################################################
# Method Name: set()
# Objective: set igroup attributes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # to make compatible with 7mode library, 
    #    replace 'attr' with 'attribute' 
    if (exists $args{'attr'}) {
        $args{'attribute'} = delete $args{'attr'};
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_igroup_set",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub set

###############################################################################
# Method Name: modify()
# Objective: to modify the portset and/or initiator of igroup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_igroup_modify",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub modify

###############################################################################
# Method Name: display
# Objective: 7mode compatibility
# Details:   See POD documentation at the end of this file.
##############################################################################
sub display {
    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: add_nodes
# Objective: 7mode compatibility
# Details:   See POD documentation at the end of this file.
##############################################################################
sub add_nodes {
    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: bind
# Objective: bind portset to an igroup
# Details:   See POD documentation at the end of this file.
##############################################################################
sub bind {
    my ($self, %args) = @_;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    if (exists $args{portset_name}) {
        $args{portset} = delete $args{portset_name};
    }

    unless (defined $args{portset} and length $args{portset}) {
        $self->_set_err("mandatory parameter \"-portset\" is not passed");
        return $SanUtil::FALSE;
    }

    # initialize cmd params
    %args = (%args, $self->_get_param_defaults(), cmd => "lun_igroup_bind");

    # use common invoke to translate cdef into UI cmd and execute
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # Error is already set in SUPER invoke, just bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub bind

###############################################################################
# Method Name: unbind
# Objective: unbind portset from an igrouop
# Details:   See POD documentation at the end of this file.
##############################################################################
sub unbind {
    my ($self, %args) = @_;

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd params
    %args = ($self->_get_param_defaults(), cmd => "lun_igroup_unbind");

    # use common invoke to translate cdef into UI cmd and execute
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # Error is already set in SUPER invoke, just bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub unbind

###############################################################################
# Method Name: get_list_info()
# Objective: Get igroup information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_igroup_show",
             $self->_get_param_defaults());
    
    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain igroup object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return {};
    }

    unless (scalar(@output)) {
        return {};
    }

    return $output[0];

} ## end sub get_list_info


###############################################################################
# Method Name: has_maps()
# Objective: Checks whether an igroup has one or more LUNs mapped to it
# Details:   See POD documentation at the end of this file.
###############################################################################
sub has_maps {
    my ($self) = @_;
    #Locals
    my @lun_names = ();
    #Clear erros
    $self->_clear_err();

    #Call private method
    @lun_names = $self->_get_luns_mapped_to_igroup();
    if ($self->error()) {
        $self->_set_err("Failed to get the luns mapped to this igroup" .
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    if (scalar(@lun_names)) {
        return $SanUtil::TRUE;
    } else {
        return $SanUtil::FALSE;
    }

} ## end sub has_maps

###############################################################################
# Method Name: get_maps()
# Objective: Gets a list of lun names mapped to this igroup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_maps {
    my ($self) = @_;
    #Loclas
    my @lun_names = ();
    #Clear errors
    $self->_clear_err();

    #Call private method
    @lun_names = $self->_get_luns_mapped_to_igroup();
    if ($self->error()) {
        $self->_set_err("Failed to get the lun mapped to this igroup" .
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return return (wantarray()) ? @lun_names : \@lun_names;

} ## end sub get_maps

################################################################################
# A C C E S S O R   M E T H O D S
################################################################################

################################################################################
# Method Name: get_alua()
# Objective:   Get igroup alua
# Details:     See POD documentation at the end of this file
################################################################################
sub get_alua {
    my ($self, %args) = @_;

    $args{field} = "alua";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_initiator()
# Objective:   Get igroup initiator
# Details:     See POD documentation at the end of this file
################################################################################
sub get_initiator {
    my ($self, %args) = @_;

    $args{field} = "init_details";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_name()
# Objective:   Get igroup name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "igroup";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get igroup parent object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_portset()
# Objective:   Get igroup portset
# Details:     See POD documentation at the end of this file
################################################################################
sub get_portset {
    my ($self, %args) = @_;

    $args{field} = "portset";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}
################################################################################
# Method Name: get_protocol_type()
# Objective:   Wrapper to get_protocol
# Details:     See POD documentation at the end of this file
################################################################################
sub get_protocol_type {
    my ($self, %args) = @_;

    # wrapper to get_protocol
    return $self->get_protocol(%args);
}

################################################################################
# Method Name: get_protocolx()
# Objective:   Get igroup protocol type
# Details:     See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self, %args) = @_;

    $args{field} = "protocol";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_ostype()
# Objective:   Get igroup ostype
# Details:     See POD documentation at the end of this file
################################################################################
sub get_ostype {
    my ($self, %args) = @_;

    $args{field} = "ostype";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_throttle_resv()
# Objective:   Get igroup throttle_resv
# Details:     See POD documentation at the end of this file
################################################################################
sub get_throttle_resv {
    my ($self, %args) = @_;

    $args{field} = "throttle_resv";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_throttle_borrow()
# Objective:   Get igroup throttle_borrow
# Details:     See POD documentation at the end of this file
################################################################################
sub get_throttle_borrow {
    my ($self, %args) = @_;

    $args{field} = "throttle_borrow";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_uuid()
# Objective:   Get igroup uuid
# Details:     See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self, %args) = @_;

    $args{field} = "uuid";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}
################################################################################
# Method Name: get_vserver()
# Objective:   Get vserver this igroup belongs to
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vserver {
    my ($self, %args) = @_;

    $args{field} = "vserver";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method Name: _get_param_defaults
# Objective: Gets default igroup parameters
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    # default hash parameters to be used in igroup object
    my %defaults_params = (igroup => $self->get_name(-use_cache => 1),
                           vserver => $self->get_vserver(-use_cache => 1));

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Igroup
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # set use_cache flag if exists
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "lun_igroup_show",
                                                field => $args{field},
                                                $self->_get_param_defaults());
    }
}

################################################################################
# Method Name: _create_igroup
# Objective: Gets default igroup parameters
################################################################################
sub _create_igroup {
    my ($self, %args) = @_;

    # append cdef for igroup create to argument list
    %args = (%args, cmd => "lun_igroup_create");

    # if initiator objs are passed in, use it to obtain parameters to create
    # igroup and add its nodes
    if (defined $self->{initobjs}) {
        # get client handle from the first initobj
    }

    # use common invoke from base class (San::CommonInterface)
    # to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj " . $args{cdef}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # call igroup refresh to populate object
    if ($self->{enable_cache} == $SanUtil::TRUE) {
        $self->refresh();
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to refresh " . $args{igroup}, $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    return $SanUtil::TRUE;
}

################################################################################
# Method Name: _get_luns_mapped_to_igroup
# Objective: Gets the list of luns mapped to a given igroup
################################################################################
sub _get_luns_mapped_to_igroup {
    my ($self) = @_;
    # Locals
    my $igroup_name = '';
    my $node_obj = undef;
    my (@mapped_luns, @lun_names, %info) = ();
    #Clear error
    $self->_clear_err();

    # Get the Node Object
    $node_obj =  $self->get_parent_obj();
    if ($self->error()) {
        $self->_set_err("Failed to retrieve the parent object" . $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get all the information for each mapped Lun for this igroup
    @mapped_luns = $node_obj->lun_mapped_list_info(-igroup => $self->get_name());
    if ($node_obj->error()) {
        $self->_set_err("Failed to get the list of mapped luns" . $node_obj->errmsg());
        return (wantarray()) ? () : [];
    }

    # Go through each element [hash] and get the name of the lun
    foreach my $info_ref (@mapped_luns) {
        if (exists $info_ref->{lun}) {
            push(@lun_names, $info_ref->{lun});
        }
    }

    return (wantarray()) ? @lun_names : \@lun_names;
}

1;

=pod

=head1 Package Name

Igroup.pm

-head1 Author

Manoj Chintala

=head1 Description

Provides an interface to Igroup object

=over 4

=item Examples

$igroup_obj = San::ControllerCluster::Filer10M::Vserver::Igroup->new(-name => $igroup,
                                                                    -filer => $filer,
                                                                     %args);
if ($igroup_obj->error()) {
    $self->_set_err("failed to instantiate igroup object", $igroup_obj->errmsg());
    return {};
}

$igroup_obj = San::ControllerCluster::Filer10M::Vserver->create_igroup_obj(-name => $igroup,
                                                                     %args);
if ($igroup_obj->error()) {
    $self->_set_err("failed to instantiate igroup object", $igroup_obj->errmsg());
    return {};
}
=back

=head1 Public Methods  

=head2 B<set>

set igroup attributes. Call the function to set the igroup attribute one at a time.

Params:
    attribute <Mandatory> The name of the attribute to be set on the igorup
           currently supported attribute: os-type|throttle-reserve|throttle-borrow|alua
    value  <Mandatory> Value for the attribute specified in the "attribute" param.
    force  <optional>  true. Forcibly set the attribute, disabling conflict checks where and when applicable
    
=over 4

=item Example:

    # to change the ostype of the igroup               
    $igroup_obj->set(-attribute => "os_type", -value => "solaris");
    if ($igroup_obj->error()) {
        $self->_set_err("failed to set ostype attributte on igroup", $igroup_obj->errmsg());
        return {};
    }
    
    # to change the alua of the igroup               
    $igroup_obj->set(-attribute => "alua", -value => "yes");
    if ($igroup_obj->error()) {
        $self->_set_err("failed to set alua attributte on igroup", $igroup_obj->errmsg());
        return {};
    }
    
=back

=head2 B<modify>

To modify the portset and/or initiator of igroup.
Function can be used to modify igroup by calling both param at once or one at a time.

Params:
    portset <optional> The name of the portset to bind with igroup
    initiator <optional> List of initiators, if more than one, comma separated list
    
=over 4

=item Example:

    # to modify the portset binding of the igroup               
    $igroup_obj->modify(-portset => "ps1");
    if ($igroup_obj->error()) {
        $self->_set_err("failed to modify the portset name on igroup", $igroup_obj->errmsg());
        return {};
    }
    
    # to modify the portset binding and initiator of the igroup               
    $igroup_obj->modify(-portset => "ps1", 
                        -initiator => "01:02:03:04:05:06:07:08, 01:02:03:04:05:06:07:FF");
    if ($igroup_obj->error()) {
        $self->_set_err("failed to modify the igroup", $igroup_obj->errmsg());
        return {};
    }
    
=back

=head2 B<bind>

Bind the portset to the igroup.

Params:
    portset_name <Mandatory> The name of the portset to be bind

=over 4

=item Example:

    $igroup_obj->bind(-portset_name => $portset_name);
    if ($igroup_obj->error()) {
        $self->_set_err("failed to bind portset $portset_name to igroup", $igroup_obj->errmsg());
        return {};
    }
=back

=head2 B<unbind()>

Unbind the portset from the igroup.

=over 4

=item Example:

$igroup_obj->unbind();
if ($igroup_obj->error()) {
    $self->_set_err("failed to unbind portset $portset_name from igroup", $igroup_obj->errmsg());
    return {};
}

=back

=head2 B<get_list_info()>

Gets the igroup info from igroup object

=over 4

=item Example:
#
# The rslt_hash_ref will contain as keys all of the fields returned by the show
# command. For example if you want the ostype, you have to do $rslt_hash_ref->{ostype}
#
$rslt_hash_ref = $igroup_obj->get_list_info();
if ($igroup_obj->error()) {
    $self->_set_err("failed to get igroup info", $igroup_obj->errmsg());
    return {};
}

=back

=head2 B<has_maps()>

Returns TRUE if there is at least one lun mapped to this igroup.
Returns FALSE if no lun mapped is found.

=over 4

=item Example:

$is_mapped = $igroup_obj->has_maps();
if ($igroup_obj->error()) {
    ...
}

if ($is_mapped) {
    ...
}


=back

=head2 B<get_maps()>

Returns a list (array) of strings (names of the luns) that are mapped to this
igroup.

=over 4

=item Example:

@lun_names= $igroup_obj->get_maps();
if ($igroup_obj->error()) {
    ...
}

foreach $name (@lun_names) {
    ...
}





=back
