# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to an iSCSI object
##
## @author korry@netapp.com (dl-ssan-automation)
## @status public

package San::ControllerCluster::Filer10M::Vserver::Iscsi;

# compiler directives
use strict;

# imported methods
use San::ControllerCluster::Filer10M::Vserver::Iscsi::Tpgroup;
use Data::Dumper;

use Params::Validate qw( :all );
use San::SanLog;

# Constants
use constant MAX_ZAPI_RECORDS => 1024;

# inherit base class
use base qw( San::CommonInterface );

###############################################################################
# Method Name: new
# Objective: Create new iSCSI object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, @list_names);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }

    # set priv level advanced if setting a parm that requires it.
    if (   defined $args{'target-name'}
        or defined $args{'max-error-recovery-level'}
        or defined $args{'retain-timeout'}
        or defined $args{'login-timeout'}
        or defined $args{'max-conn-per-session'}
        or defined $args{'max-ios-per-session'}
        or defined $args{'tcp-window-size'}) {
        $args{cmd_privilege} = "advanced";
    }

    # If the force parm was sent, then we must run as CLI.
    # Zapi does not support the force parm.
    $args{cmd_interface} = "cli" if (defined $args{force});

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct an igroup object
    $self = $class->SUPER::_new(-name  => $args{iscsi},
                                -filer => $args{filer},
                                %args);

    if ($self->error()) {
        $self->_set_err("failed to create an iscsi object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # store iscsi elements on a list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # store parent object (vserver) because we will need it to iscsi lifs
    if (exists $args{parent_obj}) {
        $self->{parent_obj} = delete $args{parent_obj};
    }

    # delete unused argument
    delete $args{debug_level};

    # create get object accessor method for Iscsi
    @list_names = qw( tpgroup );
    $self->create_accessor_methods(-type => "get_obj",
                                  -pkg  => __PACKAGE__,
                                  -list => \@list_names);

    # if the obj has been initialized, just refresh to update info
    # otherwise initialize it by create and update its object
    unless ($args{initialized}) {
        # iscsi has not been initialized/created, so we need to create it
        $self->_create_iscsi(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to _create_iscsi() ", $self->errmsg());
            return $self;
        }
    }

    return $self;
} ## end sub new

###############################################################################
# Method Name: get_initiators
# Objective: get iscsi initiators
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_initiators {
    my ($self) = @_;

    my (@init_records, %args, $schema_obj, $cmd);

    $self->_clear_err();

    %args = (cmd => "vserver_iscsi_initiator_show",
             $self->_get_param_defaults());

    # if initiators have been cached, get it
    if ($self->is_cache_data_valid($args{cmd})) {
        @init_records = $self->get_cache_data($args{cmd});
    } else {

        # nothing is cached, we need to get initiators
        # invoke_cmd_obj
        (@init_records) = $self->SUPER::invoke_cmd_obj(%args);
        if ($self->error()) {

            # set error and bubble up
            $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # if no initiator, return empty list
        if ($init_records[0] =~ /There are no entries/i) {
            return (wantarray()) ? () : [];
        }

        # let's sancache it
        $self->store_cache_data($args{cmd}, @init_records);
    }

    return (wantarray()) ? @init_records : \@init_records;
} ## end sub get_initiators

###############################################################################
# Method Name: interface_accesslist_list_info
# Objective: get iscsi initiator_auths
# Details:   See POD documentation at the end of this file.
###############################################################################
sub interface_accesslist_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_interface_accesslist_show",
             $self->_get_param_defaults());

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;
} ## end sub interface_accesslist_list_info

###############################################################################
# Method Name: initiator_auth_list_info
# Objective: get iscsi initiator_auths
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initiator_auth_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_security_show",
             $self->_get_param_defaults());

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;

} ## end sub initiator_auth_list_info

###############################################################################
# Method Name: interface_list_info
# Objective: get iscsi interfaces
# Details:   See POD documentation at the end of this file.
###############################################################################
sub interface_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_interface_show",
             $self->_get_param_defaults());

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;

} ## end sub interface_list_info

###############################################################################
# Method Name: isns_list_info
# Objective: get isns information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub isns_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_isns_show",
             $self->_get_param_defaults());

    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    if (scalar(@output) > 1) {
        $self->_set_err("Got back more than 1 array element for isns show!");
        return {};
    }

    return $output[0];

} ## end sub isns_list_info

###############################################################################
# Method Name: isns_create
# Objective: Creates isns record
# Details:   See POD documentation at the end of this file.
###############################################################################
sub isns_create {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{'address'} and length $args{'address'}) {
        $self->_set_err("mandatory parameter \"-address\" not specified ");
        return $SanUtil::FALSE;
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_isns_create",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub isns_create

###############################################################################
# Method Name: isns_delete
# Objective: deletes isns record
# Details:   See POD documentation at the end of this file.
###############################################################################
sub isns_delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_isns_delete",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub isns_delete

###############################################################################
# Method Name: isns_start
# Objective: starts isns service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub isns_start {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_isns_start",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub isns_start

###############################################################################
# Method Name: isns_stop
# Objective: stops isns service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub isns_stop {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_isns_stop",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub isns_stop

###############################################################################
# Method Name: isns_update
# Objective: sends updated config to iSNS server
# Details:   See POD documentation at the end of this file.
###############################################################################
sub isns_update {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_isns_update",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub isns_update

###############################################################################
# Method Name: isns_modify
# Objective: modifies iSNS record
# Details:   See POD documentation at the end of this file.
###############################################################################
sub isns_modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_isns_modify",
             $self->_get_param_defaults());

    # If using CLI, call isns modify as normal.
    if ($self->get_cmd_interface() =~ /cli/i) {

        # use common invoke to translate cdef into UI cmd and execute it
        $self->SUPER::invoke_cmd_obj(%args);
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj " . $args{cmd},
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    } else {

        # Else it's a zapi.  Note the CLI cmd for isns modify allows the user
        # to specify an ip address and/or a status-admin state.  But the zapi
        # only allows you to set the ip address.  So we need to check the
        # args and call the proper zapis that are needed.

        # If 'address' is an arg, then we need to call the isns modify zapi.
        if (defined $args{'address'}) {

            # use common invoke to translate cdef into UI cmd and execute it
            $self->SUPER::invoke_cmd_obj(%args);
            if ($self->error()) {
                $self->_set_err("failed to invoke_cmd_obj " . $args{cmd},
                                $self->errmsg());
                return $SanUtil::FALSE;
            }
        } ## end if (defined $args{'address'...

        # If 'status-admin' is an arg, then we need to call either start or stop
        if (defined $args{'status-admin'}) {

            # If status-admin is up, call the start zapi
            if ($args{'status-admin'} =~ /^up$/i) {

                $self->isns_start();
                if ($self->error()) {
                    $self->_set_err("isns start failed", $self->errmsg());
                    return $SanUtil::FALSE;
                }
            } elsif ($args{'status-admin'} =~ /^down$/i) {

                # elsif status-admin is down, call the stop zapi
                $self->isns_stop();
                if ($self->error()) {
                    $self->_set_err("isns stop failed", $self->errmsg());
                    return $SanUtil::FALSE;
                }
            } else {
                $self->_set_err(  "isns modify called with status-admin '"
                                . $args{'status-admin'}
                                . "'.  Expecting up or down.");
                return $SanUtil::FALSE;
            }
        } ## end if (defined $args{'status-admin'...
    } ## end else [ if ($self->get_cmd_interface...

    return $SanUtil::TRUE;
} ## end sub isns_modify

###############################################################################
# Method Name: session_list_info
# Objective: get iscsi sessions
# Details:   See POD documentation at the end of this file.
###############################################################################
sub session_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_session_show",
             $self->_get_param_defaults());

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;

} ## end sub session_list_info

###############################################################################
# Method Name: connection_list_info
# Objective: get iscsi connections
# Details:   See POD documentation at the end of this file.
###############################################################################
sub connection_list_info {
    my ($self, %args) = @_;

    my (@output, $init_nodename);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # iscsi connection show used to take an -initiator-name flag.
    # But this was removed.  Our testcases found this option useful.
    # So I have added in code to this function to support the -initiator-name flag.
    # In this case, we call session_list_info with the inititor-name parm.
    # Then take the tsih and tpgroup-tag from each session to discover the
    # connections. (vserver parm is added by default since we are in Iscsi.pm.)
    # Also added session-type filter.  This filter is also used first on iscsi
    # session show as a "pre-filter" for iscsi connection show.

    # If the caller passed in initiator-name or session-type, then call
    # session show first as a pre-filter.
    if (exists $args{'initiator-name'} or exists $args{'session-type'}) {

        # User asked for initiator-name and/or session-type filter.  We need to
        # call session_list_info first in order to then call
        # connection_list_info with the session's tsih.
        # (vserver is automatically appended by _get_param_defaults)
        my %session_args = ($self->_get_param_defaults());

        # If the user specified initiator-name, add it to the %session_args.
        # Also delete it from the %args.
        if (defined $args{'initiator-name'} and length $args{'initiator-name'}) {
            %session_args =
              (%session_args, 'initiator-name' => $args{'initiator-name'});
            delete $args{'initiator-name'};
        }

        # If the user specified session-type, add it to the %session_args.
        # Also delete it from the %args.
        if (defined $args{'session-type'} and length $args{'session-type'}) {
            %session_args =
              (%session_args, 'session-type' => $args{'session-type'});
            delete $args{'session-type'};
        }

        # If the user asked for a tsih filter, let's add that as well
        if (defined $args{'tsih'} and length $args{'tsih'}) {
            %session_args = (%session_args, tsih => $args{tsih});

            # delete the tsih from args.  It is added back in below from the
            # session data. (In case the user did not specify this arg.)
            delete $args{tsih};
        }

        # If the user asked for a tpgroup-tag filter, let's add that as well
        if (defined $args{'tpgroup-tag'} and length $args{'tpgroup-tag'}) {
            %session_args = (%session_args, 'tpgroup-tag' => $args{'tpgroup-tag'});

            # delete the tpgroup-tag from args.  It is added back in below from the
            # session data. (In case the user did not specify this arg.)
            delete $args{'tpgroup-tag'};
        }

        # Now call session list info with the args we built
        my @sessions = $self->session_list_info(%session_args);
        if ($self->error()) {
            $self->_set_err("Failed: session_list_info with args:",
                            Dumper(%session_args),
                            $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # Now scan each session to find connections
        foreach my $session (@sessions) {

            # Add parms from the session to the conn arg list.
            %args = (%args,
                     cmd => "vserver_iscsi_connection_show",
                     $self->_get_param_defaults(),
                     tsih => $session->{tsih},
                     'tpgroup-tag' => $session->{'tpgroup_tag'});

            # If using ZAPI and user did not set max-records, set max-records
            if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
            {
                $args{'max-records'} = MAX_ZAPI_RECORDS;
            }

            my @conns = $self->invoke_cmd_obj(%args);
            if ($self->error()) {
                $self->_set_err("failed to invoke_cmd_obj " .
                                $self->dumper_get(\%args, $self->get_debug_level()),
                                $self->errmsg());
                return (wantarray()) ? () : [];
            }

            # If using zapi and max-records were returned, issue a warning.
            # There could be more items.
            if (($self->get_cmd_interface() =~ /zapi/i)
                and scalar(@conns) == $args{'max-records'}) {
                &SanUtil::nate_log(
                    '-warn',
                    $args{'max-records'}
                      . " items returned from the zapi call. "
                      . "Call currently uses a max-records value of "
                      . $args{'max-records'}
                      . ", so there may have been more. "
                      . "You may need to increase the max-records value.");
            } ## end if (($self->get_cmd_interface...

            # Save the found connections to the output
            push(@output, @conns);

        } ## end foreach my $session (@sessions)

    } else {

        # Else user did not want any special session show filters.
        # Just call connection show normally.
        %args = (%args,
                 cmd => "vserver_iscsi_connection_show",
                 $self->_get_param_defaults());

        # If using ZAPI and user did not set max-records, set max-records
        if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
        {
            $args{'max-records'} = MAX_ZAPI_RECORDS;
        }

        (@output) = $self->invoke_cmd_obj(%args);
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj " .
                            $self->dumper_get(\%args, $self->get_debug_level()),
                            $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # If using zapi and max-records were returned, issue a warning.
        # There could be more items.
        if (($self->get_cmd_interface() =~ /zapi/i)
            and scalar(@output) == $args{'max-records'}) {
            &SanUtil::nate_log(
                '-warn',
                $args{'max-records'}
                  . " items returned from the zapi call. "
                  . "Call currently uses a max-records value of "
                  . $args{'max-records'}
                  . ", so there may have been more. "
                  . "You may need to increase the max-records value.");
        } ## end if (($self->get_cmd_interface...
    } ## end else [ if (!(defined $args{'initiator-name'...

    return (wantarray()) ? @output : \@output;
} ## end sub connection_list_info

###############################################################################
# Method Name: portal_list_info
# Objective: get iscsi portals
# Details:   See POD documentation at the end of this file.
###############################################################################
sub portal_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # vserver_iscsi_portal_show cdef doesn't exist.
    # burt443724 provided a work-around for the non-existing cdef.
    # burt430590 asks for cdefs for portal show (an aliased cmd)
    %args = (%args,
             cmd => "vserver_iscsi_portal_show",
             $self->_get_param_defaults());

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;

} ## end sub portal_list_info

###############################################################################
# Method:    get_list_info()
# Objective: get iscsi information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_show",
             $self->_get_param_defaults());

    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return {};
    }

    if (scalar(@output) > 1) {
        $self->_set_err("Got back more than 1 array element for iscsi show!");
        return {};
    }

    return $output[0];
} ## end sub get_list_info

###############################################################################
# Method Name: get_lifs
# Objective: get iscsi logical interfaces
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lifs {
    my ($self) = @_;

    my (@lif_objs);

    $self->_clear_err();

    # we need discover and get lif object using parent object
    if (defined $self->{parent_obj}) {
        # get iscsi lifs
        @lif_objs = $self->{parent_obj}->get_lifs(-type => "iscsi");
        if ($self->{parent_obj}->error()) {
            unless ($self->{parent_obj}->errmsg() =~ /no lif exists in vserver/i) {
                $self->_set_err("failed to get lif objects",
                                $self->{parent_obj}->errmsg());
            }
            return (wantarray()) ? (): [];
        }

        # cache iscsi lifs in iscsi object
        $self->{lif_objs} = \@lif_objs;
    } else {
        # no parent object, set error
        $self->_set_err("no parent object in order to get lif objects");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @lif_objs : \@lif_objs;
} ## end sub get_lifs

###############################################################################
# Method Name: set_nodename
# Objective: Sets iSCSI target nodename when -nodename parameter is supplied,
#            using cdef vserver_iscsi_nodename_set.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_nodename {
    my ($self, %args) = @_;

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    $self->_clear_err();

    my $nodename;    # nodename arg
    my $stop_iscsi = 0;
                     # 1 = iscsi stop / set nodename / iscsi start
                     # 0 = just set iscsi nodename (default)

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Parse arguments
    foreach (keys %args) {
        if (/nodename/) {
            $nodename = $args{$_};
            if (not length($nodename)) {
                $self->_set_err("Invalid value for argument '-nodename'");
                return $SanUtil::FALSE;
            }
        } elsif (/stop_iscsi/) {
            $stop_iscsi = $args{$_};
            if (not length($stop_iscsi) or $stop_iscsi !~ /^0$|^1$/) {
                $self->_set_err("Invalid value for argument '-stop_iscsi'");
                return $SanUtil::FALSE;
            }

            # Remove stop_iscsi from the args so cdef translation will work
            delete $args{"stop_iscsi"};
        }
    }    # end foreach arg

    # check for existance of required arguments
    unless (defined $nodename) {
        $self->_set_err("Missing required argument '-nodename'");
        return $SanUtil::FALSE;
    }

    # If stop_iscsi is 1, then stop the iscsi service.
    if ($stop_iscsi) {

        # If the service is already stopped turn off the stop_iscsi flag so the
        # service will not be re-started after the set.  This preserves the
        # initial state of iscsi.
        if ($self->get_status() eq "down") {
            $stop_iscsi = 0;
        } else {

            # Else service is online, go ahead and stop iscsi.
            $self->stop();
            if ($self->error()) {
                $self->_set_err("iscsi->stop failed", $self->errmsg());
                return $SanUtil::FALSE;
            }
        }
    } ## end if ($stop_iscsi)

    # If the force parm was sent, then we must run as CLI.
    # Zapi does not support the force parm.
    $args{cmd_interface} = "cli" if (defined $args{force});

    # set the nodename
    $self->_nodename(%args);
    if ($self->error()) {

        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # If stop_iscsi is 1, then start the iscsi service.
    if ($stop_iscsi) {
        $self->start();
        if ($self->error()) {
            $self->_set_err("iscsi->start failed", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} ## end sub set_nodename

###############################################################################
# Method Name: get_nodename
# Objective: By default, returns the iSCSI target nodename returned from
#            vserver_nodename or, alternatively, returns the iSCSI target
#            nodename stored in the iSCSI object if the -use_cache flag
#            is set to 1
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_nodename {
    my ($self, %args) = @_;

    my ($controller, $nodename);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify optional parameter
    if (defined $args{'use_cache'}) {
        unless (length $args{'use_cache'}) {
            $self->_set_err("invalid parameter value for \"-use_cache\" " .
                            "specified ");
            return $SanUtil::FALSE;
        }
    }

    if ($args{'use_cache'} == $SanUtil::TRUE) {
        $nodename = $self->get("target-name");
        if ($self->error() or $nodename eq "") {
            # set error and bubble up
            $self->_set_err("failed to get iscsi nodename ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    } else {
        $nodename = $self->_nodename();
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to get iscsi nodename ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $nodename;
} ## end sub get_nodename

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get parent object of this Iscsi
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get parent object from cache
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_status()
# Objective:   Get iscsi status
# Details:     See POD documentation at the end of this file
################################################################################
sub get_status {
    my ($self, %args) = @_;

    $args{field} = "status_admin";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_vserver()
# Objective:   Get vserver this iscsi belong to
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vserver {
    my ($self, %args) = @_;

    $args{field} = "vserver";

    # we always want to get vserver name from cache

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method Name: start
# Objective: start iscsi service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub start {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             raw_data => 1,
             cmd => "vserver_iscsi_start",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error() or (scalar(@output) == 0)) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless ($self->get_status() =~ /up|true/) {
        $self->_set_err("failed to start iSCSI service" . $args{cmd},
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub start

###############################################################################
# Method Name: delete
# Objective: deletes iscsi service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_delete",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {

        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # Since we deleted the object mark the attribute as deleted.
    $self->{name} = "_DELETE_";

    return $SanUtil::TRUE;
} ## end sub delete

###############################################################################
# Method Name: destroy
# Objective: destroy iSCSI object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self, %args) = @_;

    return $self->delete(%args);
}

###############################################################################
# Method Name: modify
# Objective: modifies iscsi service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # regarding which interface is specified, this method is only for cli
    $args{cmd_interface} = "cli";

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_modify",
             $self->_get_param_defaults());

    # set priv level advanced if setting a parm that requires it.
    if (   defined $args{'target-name'}
        or defined $args{'max-error-recovery-level'}
        or defined $args{'retain-timeout'}
        or defined $args{'login-timeout'}
        or defined $args{'max-conn-per-session'}
        or defined $args{'max-ios-per-session'}
        or defined $args{'tcp-window-size'}) {
        $args{cmd_privilege} = "advanced";
    }

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # refresh iscsi obj to re-populate current states
    $self->refresh();
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub modify

###############################################################################
# Method Name: status
# Objective: status iscsi service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub status {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_status",
             $self->_get_param_defaults());

    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj "
                          . $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return {};
    }

    # If more than 1 item was returned for this vserver, that's an error.
    if (scalar(@output) > 1) {
        $self->_set_err("Got back more than 1 array element for iscsi status!");
        return {};
    }

    # The cli can take status-admin as a filter.  ZAPI cannot do this.
    # Therefore, we check for a filter after the zapi has been called.
    if ($self->get_cmd_interface() =~ /zapi/i and defined $args{'status-admin'})
    {

        # Note that ONTAP verifies the CLI input for status-admin is valid.
        # ZAPI does not do this processing since it does not take status-admin
        # as an input.  So we will process the arg here.
        if ($args{'status-admin'} !~ /^(up|down)$/i) {
            $self->_set_err("status-admin must be 'up' or 'down'");
            return {};
        }

        # If caller is filtering on status-admin up, verify the status is up.
        # Or if caller is filtering on status-admin down, verify the status is down.
        # If the filter is a match, return output[0]
        if ((    ($args{'status-admin'} =~ /^up$/i)
             and ($output[0]->{status_admin} =~ /up/i))
            or (    ($args{'status-admin'} =~ /^down$/i)
                and ($output[0]->{status_admin} =~ /down/i))
          ) {
            return $output[0];
        } else {

            # Else the filter did not match, return nothing.
            return {};
        }

    } ## end if ($self->get_cmd_interface...

    # Else cmd was cli, or zapi without a status-admin filter.
    return $output[0];
} ## end sub status

###############################################################################
# Method Name: stop
# Objective: stop iscsi service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub stop {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             raw_data => 1,
             cmd => "vserver_iscsi_stop",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error() or (scalar(@output) == 0)) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Verify that the service (status) is down
    unless ($self->get_status() =~ /down|false/) {
        $self->_set_err("failed to stop iSCSI service" . $args{cmd},
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub stop

###############################################################################
# Method Name: interface_disable
# Objective: interface_disable iscsi service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub interface_disable {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless ((defined $args{'lif'} and length $args{'lif'})
            xor($args{'all'} =~ /^true$/i)) {
        $self->_set_err(
                      "Must specify \"-lif\" or \"-all true\".  But not both.");
        return $SanUtil::FALSE;
    }

    # Note: If the user passes in -all false and specifies a -lif, both CLI and
    # zapi will catch this and return an error.

    # If using zapi, and -all true is specified, then intercept this.
    # zapi does not support -all true, so we have to have an exception for it.
    if (($self->get_cmd_interface() =~ /zapi/i) and ($args{all} =~ /true/i)) {

        # grab list of all iscsi interfaces that are enabled.
        my @ints = $self->interface_list_info(-enabled => "true");
        if ($self->error()) {
            $self->_set_err("interface_list_info failed", $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Now scan the list of interfaces, and disable each one.
        foreach my $int (@ints) {

            $self->interface_disable(-lif => $int->{lif});
            if ($self->error()) {
                $self->_set_err("interface_disable failed", $self->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # If we get here, then all interfaces were disabled.
        return $SanUtil::TRUE;
    } ## end if (($self->get_cmd_interface...

    # Next, if user requested zapi and passed in multiple lifs, then we need
    # to convert these into individual zapi calls because zapi only allows 1
    # lif at a time.
    if ($self->get_cmd_interface() =~ /zapi/i) {

        # Let's see how many lifs were passed in. (separated by comma)
        my @lifs = split(/,/, $args{lif});

        # If more than 1 lif, then we have work to do.
        if (scalar(@lifs) > 1) {

            # Scan each lif
            foreach my $lif (@lifs) {

                $self->interface_disable(-lif => $lif);
                if ($self->error()) {
                    $self->_set_err("interface_disable failed",
                                    $self->errmsg());
                    return $SanUtil::FALSE;
                }
            }

            # If we get here, then all lifs were disabled.
            return $SanUtil::TRUE;
        } ## end if (scalar(@lifs) > 1)
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_interface_disable",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd},
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub interface_disable

###############################################################################
# Method Name: interface_enable
# Objective: interface_enable iscsi service
# Details:   See POD documentation at the end of this file.
###############################################################################
sub interface_enable {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless ((defined $args{'lif'} and length $args{'lif'})
            xor ($args{'all'} =~ /^true$/i)) {
        $self->_set_err(
                      "Must specify \"-lif\" or \"-all true\".  But not both.");
        return $SanUtil::FALSE;
    }

    # Note: If the user passes in -all false and specifies a -lif, both CLI and
    # zapi will catch this and return an error.

    # If using zapi, and -all true is specified, then intercept this.
    # zapi does not support -all true, so we have to have an exception for it.
    if (($self->get_cmd_interface() =~ /zapi/i) and ($args{all} =~ /true/i)) {

        # grab list of all iscsi interfaces that are disabled.
        my @ints = $self->interface_list_info(-enabled => "false");
        if ($self->error()) {
            $self->_set_err("interface_list_info failed", $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Now scan the list of interfaces, and enable each one.
        foreach my $int (@ints) {

            $self->interface_enable(-lif => $int->{lif});
            if ($self->error()) {
                $self->_set_err("interface_enable failed", $self->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # If we get here, then all interfaces were disabled.
        return $SanUtil::TRUE;
    } ## end if (($self->get_cmd_interface...

    # Next, if user requested zapi and passed in multiple lifs, then we need
    # to convert these into individual zapi calls because zapi only allows 1
    # lif at a time.
    if ($self->get_cmd_interface() =~ /zapi/i) {

        # Let's see how many lifs were passed in. (separated by comma)
        my @lifs = split(/,/, $args{lif});

        # If more than 1 lif, then we have work to do.
        if (scalar(@lifs) > 1) {

            # Scan each lif
            foreach my $lif (@lifs) {

                $self->interface_enable(-lif => $lif);
                if ($self->error()) {
                    $self->_set_err("interface_enable failed",
                                    $self->errmsg());
                    return $SanUtil::FALSE;
                }
            }

            # If we get here, then all lifs were enabled.
            return $SanUtil::TRUE;
        } ## end if (scalar(@lifs) > 1)
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_interface_enable",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub interface_enable

################################################################################
# Subroutine Name: tpgroup_exist
# Objective:       Determines if tpgroup exists.
# Details:         See POD documentation at the end of this file
################################################################################
sub tpgroup_exist {
    my ($self, %args) = @_;

    my (@tpgroup_records);

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);
    if (defined $args{name} and length $args{name}){
        $args{tpgroup} = delete $args{name};
    }

    # build cmd vserver iscsi show
    %args = (%args,
             cmd => "vserver_iscsi_tpgroup_show",
             $self->_get_param_defaults());

    # invoke cdef command but do not update attributes
    (@tpgroup_records) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # if no entries, return FALSE
    if ($tpgroup_records[0] =~ /There are no entries/i or scalar(@tpgroup_records) == 0) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: create_tpgroup_obj
# Objective:   Create tpgroup object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_tpgroup_obj {
    my ($self, %args) = @_;

    my ($tpgroup_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # append aggregate name parameter to argument list
    %args = (%args, $self->_get_param_defaults());

    # instantiate tpgroup object
    $tpgroup_obj =
      San::ControllerCluster::Filer10M::Vserver::Iscsi::Tpgroup->new(
                                                        -parent_obj => $self,
                                                        -filer => $self->get_filer(),
                                                        %args);
    if ($tpgroup_obj->error()) {
        $self->_set_err("failed to instantiate tpgroup object", $tpgroup_obj->errmsg());
        return {};
    } else {
        # store volume object in filer object
        push @{$self->{tpgroup_objs}}, $tpgroup_obj;
    }
    return $tpgroup_obj;
} ## end sub create_tpgroup_obj

################################################################################
# Method Name: get_tpgroup_obj()
# Objective:   Get a specific tpgroup object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_tpgroup_obj {
    my ($self, %args) = @_;

    my (@tpgroup_objs);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to tpgroup to maintain consistency
    # and compatibility w/ 7-mode
    if (defined $args{name} and length $args{name}) {
        $args{tpgroup} = delete $args{name};
    }

    # verify tpgroup param
    unless (defined $args{tpgroup} and length $args{tpgroup}) {
        $self->_set_err("mandatory parameter \"-tpgroup\" not specified ");
        return {};
    }

    # get tpgroup objects
    @tpgroup_objs = $self->get_tpgroup_objs();
    if ($self->error()) {
        $self->_set_err("failed to get tpgroup objects", $self->errmsg());
        return {};
    }

    # find a specific tpgroup object
    foreach my $tpgroup_obj (@tpgroup_objs) {
        return $tpgroup_obj if ($tpgroup_obj->get_name() =~ /^$args{tpgroup}$/);
    }

    # no tpgroup object that matches the criteria
    $self->_set_err("cannot find tpgroup " . $args{tpgroup});

    return {};
} ## end sub get_tpgroup_obj

###############################################################################
# Method Name: refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # we don't need args{iscsi} here
    delete $args{iscsi} if (exists $args{iscsi});

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update igroup object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {

        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub refresh

################################################################################
# Method Name: get_alias()
# Objective:   Get iscsi alias
# Details:     See POD documentation at the end of this file
################################################################################
sub get_alias {
    my ($self, %args) = @_;
    my ($vserver);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # validate use_alias_cmd flag
    unless ($args{use_alias_cmd} == $SanUtil::FALSE or
            $args{use_alias_cmd} == $SanUtil::TRUE) {
        $self->_set_err("failed to validate parameter: use_alias_cmd");
        return "";
    }

    if (delete $args{use_alias_cmd} == $SanUtil::FALSE) {
        $args{field} = "target-alias";
        # the _get_field() will handle error
        return $self->_get_field(%args);
    }

    # initialize cmd parameters
    #
    # For zapi we need a different command to be able to diferentiate between
    # get and set.
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{'cmd'} = "vserver_iscsi_alias_get";
    } else {
        $args{'cmd'} = "vserver_iscsi_alias";
    }
    %args = (%args,
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("Failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return "";
    }

    if (scalar(@output) > 1) {
        $self->_set_err("Got back more than 1 array element for iscsi alias!");
        return "";
    }

    # Verify we got something back.
    if (exists $output[0]->{'target_alias'} and length($output[0]->{'target_alias'})) {
        return $output[0]->{'target_alias'};
    }

    $self->_set_err("Failed to get iscsi alias from output:\n" . Dumper($output[0]) );
    return "";

}

################################################################################
# Method Name: set_alias()
# Objective:   set iscsi alias
# Details:     See POD documentation at the end of this file
################################################################################
sub set_alias {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # validate doesn't understand $SanUtil:: in its callback.
    # So, have to do this.
    unless ($args{use_alias_cmd} == $SanUtil::FALSE or
            $args{use_alias_cmd} == $SanUtil::TRUE) {
        $self->_set_err("failed to validate parameter: use_alias_cmd");
        return "";
    }

    if (exists($args{alias})) {
        $args{'target-alias'} = delete $args{alias};
    }

    unless (defined $args{'target-alias'} and length $args{'target-alias'}) {
        $self->_set_err("failed to validate parameter target-alias:" .
                        $args{'target-alias'});
        return $SanUtil::FALSE;
    }

    if (delete $args{use_alias_cmd} == $SanUtil::FALSE) {
        # call modify() to set alias
        return $self->modify(%args);
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_alias",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # refresh iscsi obj to re-populate current alias
    $self->refresh();
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: clear_alias()
# Objective:   clear iscsi alias
# Details:     See POD documentation at the end of this file
################################################################################
sub clear_alias {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    #
    # For zapi we need a different command to be able to call the correct instruction.
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{'cmd'} = "vserver_iscsi_alias_clear";
    } else {
        $args{'cmd'}  = "vserver_iscsi_alias";
        $args{'clear'}= "true";
    }
    %args = (%args,
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # refresh iscsi obj to re-populate alias
    $self->refresh();
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: tpgroup_list_info
# Objective: get tpgroup list
# Details:   See POD documentation at the end of this file.
###############################################################################
sub tpgroup_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_tpgroup_show",
             $self->_get_param_defaults());

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;

} ## end sub tpgroup_list_info

###############################################################################
# Method Name: initiator_list_info
# Objective: get initiator list
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initiator_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # If the caller passed in session-type, then call session show first as a
    # pre-filter.
    if (exists $args{'session-type'}) {

        # User asked for session-type filter.  We need to call
        # session_list_info first in order to then call initiator_list_info
        # with the session's tsih and tpgroup-tag.
        # (vserver is automatically appended by _get_param_defaults)
        my %session_args = ($self->_get_param_defaults());

        # If the user specified initiator-name, add it to the %session_args.
        # Don't delete initiator from the args, we'll use it when we call
        # initiator show below.
        if (defined $args{'initiator-name'} and length $args{'initiator-name'}) {
            %session_args =
              (%session_args, 'initiator-name' => $args{'initiator-name'});
        }

        # If the user specified session-type, add it to the %session_args.
        # Also delete it from the %args.
        if (defined $args{'session-type'} and length $args{'session-type'}) {
            %session_args =
              (%session_args, 'session-type' => $args{'session-type'});
            delete $args{'session-type'};
        }

        # If the user asked for a tsih filter, let's add that as well
        if (defined $args{'tsih'} and length $args{'tsih'}) {
            %session_args = (%session_args, tsih => $args{tsih});

            # delete the tsih from args.  It is added back in below from the
            # session data. (In case the user did not specify this arg.)
            delete $args{tsih};
        }

        # If the user asked for a tpgroup-tag filter, let's add that as well
        if (defined $args{'tpgroup-tag'} and length $args{'tpgroup-tag'}) {
            %session_args = (%session_args, 'tpgroup-tag' => $args{'tpgroup-tag'});

            # delete the tpgroup-tag from args.  It is added back in below from the
            # session data. (In case the user did not specify this arg.)
            delete $args{'tpgroup-tag'};
        }

        # Now call session list info with the args we built
        my @sessions = $self->session_list_info(%session_args);
        if ($self->error()) {
            $self->_set_err("Failed: session_list_info with args:",
                            Dumper(%session_args),
                            $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # Now scan each session to find initiators
        foreach my $session (@sessions) {

            # Add tsih from the session to the arg list.
            %args = (%args,
                     cmd => "vserver_iscsi_initiator_show",
                     $self->_get_param_defaults(),
                     tsih => $session->{tsih},
                     'tpgroup-tag' => $session->{'tpgroup_tag'});

            # If using ZAPI and user did not set max-records, set max-records
            if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
            {
                $args{'max-records'} = MAX_ZAPI_RECORDS;
            }

            my @inits = $self->invoke_cmd_obj(%args);
            if ($self->error()) {
                $self->_set_err("failed to invoke_cmd_obj " .
                                $self->dumper_get(\%args, $self->get_debug_level()),
                                $self->errmsg());
                return (wantarray()) ? () : [];
            }

            # If using zapi and max-records were returned, issue a warning.
            # There could be more items.
            if (($self->get_cmd_interface() =~ /zapi/i)
                and scalar(@inits) == $args{'max-records'}) {
                &SanUtil::nate_log(
                    '-warn',
                    $args{'max-records'}
                      . " items returned from the zapi call. "
                      . "Call currently uses a max-records value of "
                      . $args{'max-records'}
                      . ", so there may have been more. "
                      . "You may need to increase the max-records value.");
            } ## end if (($self->get_cmd_interface...

            # Save the found initiators to the output
            push(@output, @inits);

        } ## end foreach my $session (@session

    } else {

        # Else user did not want any special session show filters.
        # Just call initiator show normally.
        %args = (%args,
                 cmd => "vserver_iscsi_initiator_show",
                 $self->_get_param_defaults());

        # If using ZAPI and user did not set max-records, set max-records
        if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
        {
            $args{'max-records'} = MAX_ZAPI_RECORDS;
        }

        (@output) = $self->invoke_cmd_obj(%args);
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj " .
                            $self->dumper_get(\%args, $self->get_debug_level()),
                            $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # If using zapi and max-records were returned, issue a warning.
        # There could be more items.
        if (($self->get_cmd_interface() =~ /zapi/i)
            and scalar(@output) == $args{'max-records'}) {
            &SanUtil::nate_log(
                '-warn',
                $args{'max-records'}
                  . " items returned from the zapi call. "
                  . "Call currently uses a max-records value of "
                  . $args{'max-records'}
                  . ", so there may have been more. "
                  . "You may need to increase the max-records value.");
        } ## end if (($self->get_cmd_interface...

    } ## end else [ if (exists $args{'session-type'...

    return (wantarray()) ? @output : \@output;

} ## end sub initiator_list_info

###############################################################################
# Method Name: initiator_create_auth
# Objective: Creates an initiators authentication policy
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initiator_create_auth {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{'auth-type'} and length $args{'auth-type'}) {
        $self->_set_err("mandatory parameter \"-auth-type\" not specified ");
        return $SanUtil::FALSE;
    }
    unless (defined $args{'initiator-name'} and length $args{'initiator-name'}) {
        $self->_set_err("mandatory parameter \"-initiator-name\" not specified ");
        return $SanUtil::FALSE;
    }

    # if we are using zapi, we must encrypt the chap passwords
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{'password'} = $self->_convert_chap_password(
            -password => $args{'password'}) if length ($args{'password'});
        if ($self->error()) {
            $self->_set_err("failed to convert_chap_password", $self->errmsg());
            return $SanUtil::FALSE;
        }

        $args{'outbound-password'} = $self->_convert_chap_password(
            -password => $args{'outbound-password'})
            if length ($args{'outbound-password'});
        if ($self->error()) {
            $self->_set_err("failed to convert_chap_password", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_security_create",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub initiator_create_auth

###############################################################################
# Method Name: initiator_set_default_auth
# Objective: Sets the default authentication policy
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initiator_set_default_auth {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{'auth-type'} and length $args{'auth-type'}) {
        $self->_set_err("mandatory parameter \"-auth-type\" not specified ");
        return $SanUtil::FALSE;
    }

    # if we are using zapi, we must encrypt the chap passwords
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{'password'} = $self->_convert_chap_password(
            -password => $args{'password'}) if length ($args{'password'});
        if ($self->error()) {
            $self->_set_err("failed to convert_chap_password", $self->errmsg());
            return $SanUtil::FALSE;
        }

        $args{'outbound-password'} = $self->_convert_chap_password(
            -password => $args{'outbound-password'})
            if length ($args{'outbound-password'});
        if ($self->error()) {
            $self->_set_err("failed to convert_chap_password", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_security_default",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub initiator_set_default_auth

###############################################################################
# Method Name: initiator_get_default_auth
# Objective: Gets the default authentication policy
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initiator_get_default_auth {
    my ($self, %args) = @_;

    return $self->initiator_auth_list_info(%args, '-initiator-name' => 'default');
} ## end sub initiator_get_default_auth

###############################################################################
# Method Name: initiator_delete_auth
# Objective: delete an initiators authentication policy
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initiator_delete_auth {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{'initiator-name'} and length $args{'initiator-name'}) {
        $self->_set_err("mandatory parameter \"-initiator-name\" not specified ");
        return $SanUtil::FALSE;;
    }

    # initialize cmd parameters
    %args = (%args,
             cmd       => "vserver_iscsi_security_delete",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub initiator_delete_auth

###############################################################################
# Method Name: initiator_modify_chap_params
# Objective: Modify an initiators authentication policy
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initiator_modify_chap_params {
    my ($self, %args) = @_;

    return $self->initiator_modify_auth(%args);
} ## end sub initiator_modify_chap_params

###############################################################################
# Method Name: initiator_modify_auth
# Objective: Modify an initiators authentication policy
# Details:   See POD documentation at the end of this file.
###############################################################################
sub initiator_modify_auth {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{'initiator-name'} and length $args{'initiator-name'})
    {
        $self->_set_err(
                      "mandatory parameter \"-initiator-name\" not specified ");
        return $SanUtil::FALSE;
    }

    # CLI - 'iscsi security modify' can modify any parms, including auth-type.
    # ZAPI - can only modify CHAP parms. (iscsi-initiator-modify-chap-params)
    # We want thpls to be ok if user calls using cli or zapi.  Therefore, in zapi
    # case we may need some extra processing.
    if ($self->get_cmd_interface() =~ /zapi/i) {

        # if we are using zapi, we must encrypt the chap passwords
        $args{'password'} = $self->_convert_chap_password(
            -password => $args{'password'}) if length ($args{'password'});
        if ($self->error()) {
            $self->_set_err("failed to convert_chap_password", $self->errmsg());
            return $SanUtil::FALSE;
        }

        $args{'outbound-password'} = $self->_convert_chap_password(
            -password => $args{'outbound-password'})
            if length ($args{'outbound-password'});
        if ($self->error()) {
            $self->_set_err("failed to convert_chap_password", $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Check current security for this initiator.
        my @auth = $self->initiator_auth_list_info(
                                  '-initiator-name' => $args{'initiator-name'});
        if ($self->error()) {
            $self->_set_err("initiator_auth_list_info failed", $self->errmsg());
            return $SanUtil::FALSE;
        }

        # If the auth entry for the initiator was not found, return an error.
        if (scalar(@auth) == 0) {
            $self->_set_err(
                "Attempting to modify auth for an initiator without an auth entry",
                $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Check for error condition.  There is a special case that we want to
        # catch and return error.  The case: current auth method is chap AND
        # user requests none or deny AND user asked for a username or password
        # or outbound user-name or outbound password.  This would fall through
        # and get sent to TranslationTable.  TranslationTable would drop
        # auth-method and modify the chap parms, instead of returning an error
        # like the cli.  So I will check for this and catch it here.
        if (    ($auth[0]->{'auth_type'} =~ /chap/i)
            and ($args{'auth-type'} =~ /(none|deny)/i)
            and (   defined $args{'user-name'}
                 or defined $args{'password'}
                 or defined $args{'outbound-user-name'}
                 or defined $args{'outbound-password'})
          ) {
            $self->_set_err(  "Bad argument list. Cannot set auth-type to "
                            . $args{'auth-type'}
                            . " and also request chap parms");
            return $SanUtil::FALSE;
        } ## end if (($auth[0]->{'auth_type'...

        # There is a very specific case where we want the zapi to fall through
        # and run the same cmd as cli.  For this case, the TranslationTable
        # will call iscsi-initiator-modify-chap-params.  The case where we want
        # this to happen:
        #    - current auth-type is CHAP
        #    - new auth-type is CHAP OR user passed in user-name, password,
        #      outbound-user-name, or outbound-password.
        # Note in this case if the cmd falls through, the TranslationTable
        # will remove the auth-type parm since the zapi won't take it.
        # So the negative of this query is what we do intercept.  Here is the
        # query we intercept:
        if (!( ($auth[0]->{'auth_type'} =~ /chap/i)
               and (   ($args{'auth-type'} =~ /chap/i)
                    or defined $args{'user-name'}
                    or defined $args{'password'}
                    or defined $args{'outbound-user-name'}
                    or defined $args{'outbound-password'}))
          ) {

            # Since modify zapi cannot do what we want, we first delete the
            # auth method for the initiator.
            $self->initiator_delete_auth(
                                  '-initiator-name' => $args{'initiator-name'});
            if ($self->error()) {
                $self->_set_err("initiator_delete_auth failed",
                                $self->errmsg());
                return $SanUtil::FALSE;
            }

            # Now create the new auth with needed parms.
            $self->initiator_create_auth(%args);
            if ($self->error()) {
                $self->_set_err("initiator_create_auth failed",
                                $self->errmsg());
                return $SanUtil::FALSE;
            }

            # If we get here, then return success
            return $SanUtil::TRUE;
        } ## end if (!(($auth[0]->{'auth_type'...

    } ## end if ($self->get_cmd_interface...

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_security_modify",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd},
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub initiator_modify_auth

###############################################################################
# Method Name: auth_generate_chap_password
# Objective: Generate a CHAP password using cdef:vserver_iscsi_security_generate
# Details:   See POD documentation at the end of this file.
###############################################################################
sub auth_generate_chap_password {
    my ($self) = @_;
    $self->_clear_err();

    # initialize cmd parameters
    my %args = (cmd                  => "vserver_iscsi_security_generate",
                donot_append_vserver => $SanUtil::TRUE,
                raw_data             => $SanUtil::TRUE);

    # use common invoke to translate cdef into UI cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return "";
    }

    if (scalar(@output) > 1) {
        $self->_set_err("Got back more than 1 array element for iscsi security generate!");
        return "";
    }

    my $generated_password;
    if ($output[0]=~ /Generated Random Secret: (\S*)/) {
        $generated_password = $1;
    } else {
        $self->_set_err("Unexpected output from iscsi security generate command: ".
                        $output[0]);
        return "";
    }

    return $generated_password;
} ## end sub auth_generate_chap_password

###############################################################################
# Method Name: interface_accesslist_add
# Objective: Creates an initiators authentication policy
# Details:   See POD documentation at the end of this file.
###############################################################################
sub interface_accesslist_add {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless ((defined $args{'lif'} and length $args{'lif'})
            xor ($args{'all'} =~ /^true$/i)) {
        $self->_set_err(
                      "Must specify \"-lif\" or \"-all true\".  But not both.");
        return $SanUtil::FALSE;
    }
    unless (defined $args{'initiator-name'} and length $args{'initiator-name'}) {
        $self->_set_err("mandatory parameter \"-initiator-name\" not specified ");
        return $SanUtil::FALSE;;
    }

    # Note: If the user passes in -all false and specifies a -lif, both CLI and
    # zapi will catch this and return an error.

    # If using zapi, and -all true is specified, then intercept this.
    # zapi does not support -all true, so we have to have an exception for it.
    if (($self->get_cmd_interface() =~ /zapi/i) and ($args{all} =~ /true/i)) {

        # grab list of all iscsi interfaces
        my @ints = $self->interface_list_info();
        if ($self->error()) {
            $self->_set_err("interface_list_info failed", $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Now scan the list of interfaces, and add each one.
        foreach my $int (@ints) {

            $self->interface_accesslist_add(
                                    -lif              => $int->{lif},
                                    '-initiator-name' => $args{'initiator-name'}
            );
            if ($self->error()) {
                $self->_set_err("interface_accesslist_add failed",
                                $self->errmsg());
                return $SanUtil::FALSE;
            }
        } ## end foreach my $int (@ints)

        # If we get here, then all interfaces were added.
        return $SanUtil::TRUE;
    } ## end if (($self->get_cmd_interface...

    # Next, if user requested zapi and passed in multiple lifs, then we need
    # to convert these into individual zapi calls because zapi only allows 1
    # lif at a time.
    if ($self->get_cmd_interface() =~ /zapi/i) {

        # Let's see how many lifs were passed in. (separated by comma)
        my @lifs = split(/,/, $args{lif});

        # If more than 1 lif, then we have work to do.
        if (scalar(@lifs) > 1) {

            # Scan each lif
            foreach my $lif (@lifs) {

                $self->interface_accesslist_add(
                                   -lif              => $lif,
                                   '-initiator-name' => $args{'initiator-name'}
                );
                if ($self->error()) {
                    $self->_set_err("interface_accesslist_add failed",
                                    $self->errmsg());
                    return $SanUtil::FALSE;
                }
            }

            # If we get here, then all lifs were added.
            return $SanUtil::TRUE;
        } ## end if (scalar(@lifs) > 1)
    }

    # initialize cmd parameters
    %args = (%args,
             cmd       => "vserver_iscsi_interface_accesslist_add",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub interface_accesslist_add

###############################################################################
# Method Name: interface_accesslist_remove
# Objective: Creates an initiators authentication policy
# Details:   See POD documentation at the end of this file.
###############################################################################
sub interface_accesslist_remove {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless ((defined $args{'lif'} and length $args{'lif'})
            xor ($args{'all'} =~ /^true$/i)) {
        $self->_set_err(
                      "Must specify \"-lif\" or \"-all true\".  But not both.");
        return $SanUtil::FALSE;
    }
    unless (defined $args{'initiator-name'} and length $args{'initiator-name'}) {
        $self->_set_err("mandatory parameter \"-initiator-name\" not specified ");
        return $SanUtil::FALSE;;
    }

    # Note: If the user passes in -all false and specifies a -lif, both CLI and
    # zapi will catch this and return an error.

    # If using zapi, and -all true is specified, then intercept this.
    # zapi does not support -all true, so we have to have an exception for it.
    if (($self->get_cmd_interface() =~ /zapi/i) and ($args{all} =~ /true/i)) {

        # Pull a list of all interfaces in the accesslist for this iniator.
        my @ints = $self->interface_accesslist_list_info(
                                  '-initiator-name' => $args{'initiator-name'});
        if ($self->error()) {
            $self->_set_err("interface_list_info failed", $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Now scan the list of interfaces, and remove each one.
        foreach my $int (@ints) {

            $self->interface_accesslist_remove(
                                   -lif              => $int->{lif},
                                   '-initiator-name' => $args{'initiator-name'}
            );
            if ($self->error()) {
                $self->_set_err("interface_accesslist_remove failed",
                                $self->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # If we get here, then all interfaces were removed.
        return $SanUtil::TRUE;
    } ## end if (($self->get_cmd_interface...

    # Next, if user requested zapi and passed in multiple lifs, then we need
    # to convert these into individual zapi calls because zapi only allows 1
    # lif at a time.
    if ($self->get_cmd_interface() =~ /zapi/i) {

        # Let's see how many lifs were passed in. (separated by comma)
        my @lifs = split(/,/, $args{lif});

        # If more than 1 lif, then we have work to do.
        if (scalar(@lifs) > 1) {

            # Scan each lif
            foreach my $lif (@lifs) {

                $self->interface_accesslist_remove(
                                   -lif              => $lif,
                                   '-initiator-name' => $args{'initiator-name'}
                );
                if ($self->error()) {
                    $self->_set_err("interface_accesslist_remove failed",
                                    $self->errmsg());
                    return $SanUtil::FALSE;
                }
            }

            # If we get here, then all lifs were removed.
            return $SanUtil::TRUE;
        } ## end if (scalar(@lifs) > 1)
    }

    # initialize cmd parameters
    %args = (%args,
             cmd       => "vserver_iscsi_interface_accesslist_remove",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub interface_accesslist_remove

###############################################################################
# Method Name: get_stats
# Objective: Retrieves all or the specified statistic information in an array
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_stats {
    my ($self, %args) = @_;
    my (@stats);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd_privilege => "diagnostic",
             cmd => "vserver_iscsi_stats");


    @stats = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj: " . $args{cmd} . "\n" .
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @stats : \@stats;

} ## end get_stats

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method Name: _get_param_defaults
# Objective: Gets default igroup parameters
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    # default hash parameters to be used in igroup object
    # the get accessor method will retrieve from cache to prevent infinite loop
    my %defaults_params = (vserver => $self->get_vserver(-use_cache => 1));

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get Iscsi attribute from the object cache or the controller
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # convert common args to args specifically for creating lif
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "vserver_iscsi_show",
                                                field => $args{field},
                                                $self->_get_param_defaults());
    }
}

###############################################################################
# Method Name: _refresh_lifs
# Objective: Refresh lif objects. This will be called when service is started/
#            stopped/modified
# Details:   See POD.
###############################################################################
sub _refresh_lifs {
    my ($self) = @_;

    $self->_clear_err();

    # Get iSCSI lifs
    my @lifs =  $self->get_lifs();
    if ($self->error()) {
        # if the error was set because there is no lif, let's ignore
        unless ($self->errmsg() =~ /no lif exists in vserver/i) {
            $self->_set_err("failed to get lif objects ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # let's refresh each lif
    foreach my $lif (@lifs) {
        # ignore lifs that are already been deleted
        next if ($lif->get_name() =~ /_DELETE_/i);

        $lif->refresh();
        if ($lif->error()) {
            # if the error was set because there is no lif, let's ignore
            unless ($lif->errmsg() =~ /record does not exist/i) {
                $self->_set_err("failed to refresh lif ", $lif->errmsg());
                return $SanUtil::FALSE;
            }
        }
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: _nodename
# Objective: Gets iSCSI target nodename by default or sets iSCSI target
#            nodename if the -nodename parameter is supplied, using cdef
#            vserver_iscsi_nodename.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _nodename {
    my ($self, %args) = @_;

    my ($vserver, @output, $nodename);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Convert nodename to target-name
    if (exists($args{nodename})) {
        $args{'target-name'} = delete $args{nodename};
    }

    # Set cmd.  Will update below if needed.  This is used for ZAPI set and
    # CLI set/get.
    $args{'cmd'} = "vserver_iscsi_nodename";

    # verify optional parameter.  If parm is present, we are doing a set.
    if (defined $args{'target-name'}) {

        # Verify the parm is not empty.
        unless (length $args{'target-name'}) {
            $self->_set_err(
                         "parameter value for \"-target-name\" not specified ");
            return $SanUtil::FALSE;
        }

        # set priv level advanced if setting the nodename
        $args{cmd_privilege} = "advanced";

    } elsif ($self->get_cmd_interface() =~ /zapi/i) {

        # Else a target-name is not set.  If interface is zapi, set proper
        # cmd for get.
        $args{'cmd'} = "vserver_iscsi_nodename_get";
    }

    # initialize cmd parameters
    %args = (%args, $self->_get_param_defaults());

    # display argument hash if debug_level is high
    $self->sanlog(-msg       => $self->dumper_get(\%args),
                  -msg_level => 2);

    # use common invoke to translate cdef into UI cmd and execute it
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {

        # ONTAP returns error if you try to set the iSCSI nodename to the same
        # value as the current nodename.  If we did NOT get this error, then
        # continue to return FALSE.  If we did get the error, the function will
        # continue as normal.
        if ($self->errmsg() !~ /Specified nodename is same as current nodename/)
        {

            # error is already set, bubble up
            return $SanUtil::FALSE;
        }
    } ## end if ($self->error())

    # if -target-name is specified, then we set the nodename,
    # else get and return the nodename
    if (defined $args{'target-name'}) {

        # refresh iscsi obj to re-populate current nodename
        if ($self->{enable_cache}) {
            $self->refresh(%args);
            if ($self->error()) {

                # error is already set, bubble up
                return $SanUtil::FALSE;
            }
        }

        # If code reaches here, set was successful.
        return $SanUtil::TRUE;

    } else {

        # Else we performed a get operation.  Verify only 1 item returned.
        if (scalar(@output) > 1) {
            $self->_set_err("Got back more than 1 array element for iscsi nodename!");
            return "";
        }

        if (exists $output[0]->{'target_name'} and length($output[0]->{'target_name'})) {
            return $output[0]->{'target_name'};
        }

        # If code reaches here, there was a problem.
        $self->_set_err("failed to get iscsi nodename " . $output[0]);
        return "";
    } ## end else [ if (defined $args{'target-name'...
} ## end sub _nodename

################################################################################
# Method Name: _create_iscsi
# Objective: create iscsi
################################################################################
sub _create_iscsi {
    my ($self, %args) = @_;

    unless (exists $args{'no-license'}) {
        # Make sure iscsi license is installed before creating iscsi
        $self->get_parent_obj()->get_parent_obj()->license_add(-service => "iscsi");
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to add iscsi license ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    delete $args{'no-license'} if exists $args{'no-license'};

    # append cmd for iscsi create to argument list
    %args = (%args, cmd => "vserver_iscsi_create");

    # use common invoke from base class (San::CommonInterface)
    # to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # call refresh to populate object if enable_cache is set
    if ($self->{enable_cache}) {
        $self->refresh(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to refresh " . $args{iscsi},
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # set initialized flag
    $self->{initialized} = $SanUtil::TRUE;

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: _convert_chap_password
# Objective: Converts a CHAP password into encrypted form
###############################################################################
sub _convert_chap_password {
    my ($self, %args) = @_;

    my @output;

    # append cmd for password encryption to argument list
    %args = (%args, cmd => "test_password_get");

    # use common invoke from base class (San::CommonInterface)
    # to translate cdef into UI cmd and execute it
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd},
            $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Verify only 1 item returned, and return the output
    if (scalar(@output) > 1) {
        $self->_set_err("Got back more than 1 array element for password!");
        return "";
    }

    if (exists $output[0]->{'password'} and length($output[0]->{'password'})) {
        return $output[0]->{'password'};
    }

    $self->_set_err("no encrypted password returned");
    return "";
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Iscsi.pm

=head1 Author

Liz Haring (dl-ssan-automation)

=head1 Description

Provides an Iscsi interface to a C-mode filer

=head1 See Also

Iscsi.pm package used base class B<San::CommonInterface> that also inherits other modules. See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm, San::SanCache.pm>

=head1 Attributes

=head2 B<filer> (required)

string = controller name

=head2 B<vserver> (required)

string = vserver name to create Iscsi

=head1 B<Public Methods>

=head2 B<new(%args)>

Create an iscsi object

=over 4

=item Inputs

=item * I<-filer>

Controller name

=item * I<-vserver> (default: 'vs0').

Name of the vserver to create Iscsi

=item Outputs

Return Iscsi object

=item Examples

 # ex. 1: create iscsi object on vserver 'vs1'
 use San::ControllerCluster::Filer10M::Vserver::Iscsi;

 $iscsi_obj =
      San::ControllerCluster::Filer10M::Vserver::Iscsi->new(-filer  => $filer,
                                                            -vserver => "vs1");
 if ($iscsi_obj->error()) {
    ...
 }

=back

=head2 B<get_initiators()>

Gets a list of iscsi initiators records on the vserver.

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing initiators records. If an error occurs, sets the error flag and
returns an empty list.

=item Example1:

    my @initiators = $Iscsi_obj->get_initiators();

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=item Example2:

    my $initiators = $Iscsi_obj->get_initiators();
    my @records = @$initiators;

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<initiator_auth_list_info()>

Gets a list of iscsi security records on the vserver

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing hashes of iscsi security records. If an error occurs, sets the error flag and
returns an empty list.

=item Example1:

    my @iscsi_securities = $iscsi_obj->initiator_auth_list_info();
    if ($iscsi_obj->error()){
     ...
    }
    $Test->nlog(Dumper(\@iscsi_securities));

=item Example2:

    my $iscsi_securities = $iscsi_obj->initiator_auth_list_info();
    my @records = @$iscsi_securities;

    if ($iscsi_obj->error()) {
     ...
    }
    $Test->nlog(Dumper($iscsi_securities));

=back

=head2 B<interface_list_info()>

Gets a list of iscsi interface records on the vserver

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing hashes of iscsi interface records. If an error occurs, sets the error flag and
returns an empty list.

=item Example1:

    my @iscsi_interfaces = $iscsi_obj->interface_list_info();
    if ($iscsi_obj->error()){
     ...
    }
    $Test->nlog(Dumper(\@iscsi_interfaces));

=item Example2:

    my $iscsi_interfaces = $iscsi_obj->interface_list_info();
    my @records = @$iscsi_interfaces;

    if ($iscsi_obj->error()) {
     ...
    }
    $Test->nlog(Dumper($iscsi_interfaces));

=back

=head2 B<session_list_info()>

Gets a list of iscsi session records on the vserver

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi session records. If an error occurs, sets the error flag and
returns an empty list.

=item Example1:

    my @iscsi_sessions = $iscsi_obj->session_list_info();
    if ($iscsi_obj->error()){
     ...
    }

=item Example2:

    my $iscsi_sessions = $iscsi_obj->session_list_info();
    my @records = @$iscsi_sessions;

    if ($iscsi_obj->error()) {
     ...
    }

=back

=head2 B<connection_list_info()>

Gets a list of iscsi connection records on the vserver

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
connection show command.  For a list of current options, try 'iscsi connection
show ?'.

=item * I<initiator-name>

initiator-name used to be a filter in iscsi connection show.  It was removed.
However, this function has been updated so that you can still filter by
nodename.  It does this by first querying the sessions with the nodename, then
queries all connections associated with the sessions.

=item * I<session-type>

Like initiator-name, session-type can be applied as a filter.  It first calls
session_list_info to filter the sessions, then calls connection show to filter
the connections.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi connection records. If an error occurs, sets the error flag and
returns an empty list.

=item Example1:

    my @iscsi_connections = $iscsi_obj->connection_list_info();
    if ($iscsi_obj->error()){
     ...
    }

=item Example2:

    my $iscsi_connections = $iscsi_obj->connection_list_info();
    my @records = @$iscsi_connections;

    if ($iscsi_obj->error()) {
     ...
    }

=item Example3:

    my @iscsi_connections =
        $iscsi_obj->connection_list_info('-initiator-name' => "iqn.1986-03.com.sun:01:0003ba4154df.4cd2e60a");
    if ($iscsi_obj->error()){
     ...
    }

=item Example4:

    my @iscsi_connections =
        $iscsi_obj->connection_list_info('-session-type' => "Normal");
    if ($iscsi_obj->error()){
     ...
    }

=back

=head2 B<portal_list_info()>

Gets a list of iscsi portal records on the vserver

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi portal records. If an error occurs, sets the error flag and
returns an empty list.

=item Example1:

    my @iscsi_portals = $iscsi_obj->portal_list_info();
    if ($iscsi_obj->error()){
     ...
    }

=item Example2:

    my $iscsi_portals = $iscsi_obj->portal_list_info();
    my @records = @$iscsi_portals;

    if ($iscsi_obj->error()) {
     ...
    }

=back

=head2 B<get_list_info(%args)>

Returns hash reference to iscsi show command with keys being fields
and values being their values in show command

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns hash ref if successful, otherwise sets error and returns empty hash ref

=item Examples

 my $result_ref = $iscsi_obj->get_list_info();
 if ($iscsi_obj->error()){
     ...
 }

=back

=head2 B<isns_list_info(%args)>

Returns hash reference to iscsi isns show command with keys being fields
and values being their values in show command

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi isns
show command. For a list of current options, try 'iscsi isns show ?'.

=item Outputs

Returns hash ref if successful, otherwise sets error and returns empty hash ref

=item Examples

 my $result_ref = $iscsi_obj->isns_list_info();
 if ($iscsi_obj->error()){
     ...
 }

=back

=head2 B<isns_create(%args)>

Create an isns on the vserver for the iscsi object

=over 4

=item Inputs

=item * I<address>

the address of the isns

=item * <status-admin>

the status admin of the isns

=item * <force>

force creation even if isns server cannot be pinged. set to true or false.

=item Outputs

Returns true, otherwise sets error and returns false

=item Examples

 my $result_ref = $iscsi_obj->isns_create('-address' => '10.228.136.188');
 if ($iscsi_obj->error()){
     ...
 }

=back

=head2 B<isns_delete(%args)>

Deletes an isns

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns true if successful, otherwise sets error and returns false

=item Examples

 my $result_ref = $iscsi_obj->isns_delete();
 if ($iscsi_obj->error()){
     ...
 }

=back

=head2 B<isns_start(%args)>

Starts an isns

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns hash true, otherwise sets error and returns false

=item Examples

 my $result_ref = $iscsi_obj->isns_start();
 if ($iscsi_obj->error()){
     ...
 }

=back

=head2 B<isns_stop(%args)>

Stops an isns

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns true if successful, otherwise sets error and returns false

=item Examples

 my $result_ref = $iscsi_obj->isns_stop();
 if ($iscsi_obj->error()){
     ...
 }

=back

=head2 B<isns_update(%args)>

Updates an isns

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns true if successful, otherwise sets error and returns false

=item Examples

 my $result_ref = $iscsi_obj->isns_update();
 if ($iscsi_obj->error()){
     ...
 }

=back

=head2 B<isns_modify(%args)>

Modifies an isns

=over 4

=item Inputs

=item * I<status-admin>

Sets the status admin of the isns to up|down

=item * I<address>

Sets the address of isns server

=item * <force>

force modify even if isns server cannot be pinged. set to true or false.

=item Outputs

Returns true if successful, otherwise sets error and returns false

=item Examples

 my $result_ref = $iscsi_obj->isns_modify('-status-admin' => 'down');
 if ($iscsi_obj->error()){
     ...
 }

=back

=head2 B<get_lifs()>

Gets a list of iscsi data lifs on a vserver.

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi lif records. If an error occurs, sets the error flag and
returns an empty list.

=item Example1:

    my @lifs = $Iscsi_obj->get_lifs();

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=item Example2:

    my $lifs = $Iscsi_obj->get_lifs();
    my @lifs = @$lifs;

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<get_status()>

Gets the status of the iscsi service from object cache or controller.

=over 4

=item Inputs

=item * I<-use_cache> (optional)

Flag to retrieve status from cache if set to true. Default is false.

=item Outputs

Returns the iscsi status: "online" or "offline"

=item Example1:

    my $status = $Iscsi_obj->get_status();

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<get_parent_obj()>

Get parent (Vserver) object of Iscsi object

=over 4

=item Inputs

None

=item Outputs

Return parent object of Iscsi object (Vserver object)

=item Examples

 # ex. 1: get parent object
 $vs_obj = $Iscsi_obj->get_parent_obj();
 if ($Iscsi_obj->error()) {
     ...
 }

=back

=head2 B<get_vserver()>

Gets the vserver name this iscsi belongs to from cache object or controller

=over 4

=item Inputs

=item * I<-use_cache> (optional)

Flag to retrieve vserver name from cache if set to true. Default is false.

=item Outputs

Returns the vserver name of iscsi service.

=item Examples

 # ex. 1: get vserver object
 my $vs_name = $Iscsi_obj->get_vserver();

 if ($Iscsi_obj->error()) {
     $err = $Iscsi_obj->errmsg();
     logcomment("$err");
     return TCD::FAIL;
 }

=back

=head2 B<get_nodename()>

This method, by default, will return the iSCSI target nodename of the vserver. Alternatively, this method will return the iscsi target nodename of the vserver from the cache of the iscsi object when the -use_cache is set to 1.

=over 4

=item Inputs

=item * I<-use_cache> (optional)

If -use_cache is set to true, then retrieve the value of the iscsi target nodename
from the cache of the iscsi object, otherwise return the value returned by the
vserver_iscsi_nodename cdef.

=item Outputs

Returns iscsi nodename if successful, empty string otherwise.

=item Example1:

    my $nodename_from_cache=$Iscsi_obj->get_nodename();

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=item Example2:

    my $nodename_from_controller=$Iscsi_obj->get_nodename(-use_cache=> 1);

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<set_nodename()>

Sets iSCSI target nodename of vserver.

=over 4

=item Inputs

=item * I<-nodename>

(Required) Name of the iscsi nodename for the vserver.

=item * <-force>

Forces ONTAP to accept a non-NTAP or IBM target name.

=item * <-stop_iscsi>

(Optional) If set to 1, the iSCSI service will be stopped if needed.  Then the
nodename will be set.  The iSCSI service will exit the function in the same
state as it entered the function. (up or down)
If set to 0 (default), the iscsi service status will not be altered.

=item Outputs

Sets the iSCSI target nodename using the vserver_iscsi_nodename cdef.
If an error occurs the error flag is set, errmsg is set to
the failure reason, and false is passed back as the return value.

=item Examples

    $Iscsi_obj->set_nodename(-nodename => "iqn.1992-08.com.netapp:new");

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

    $iscsi->set_nodename(-nodename   => $new_iscsi_nodename,
                         -stop_iscsi => 1);
    if ($iscsi->error()) {
        $Test->nlog("set_nodename failed", $iscsi->errmsg);
        return $TCD::FAIL;
    }

=back

=head2 B<start()>

Starts the iSCSI service of vserver using the vserver_iscsi_start cdef.

=over 4

=item Inputs

=item * I<None>

=item Outputs

Returns either true or false if the operation was successful or not.
If an error occurs the error flag is set, errmsg is set to the failure reason,
and 0 is passed back as the return value.

=item Examples

    $Iscsi_obj->start();

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<status()>

Gets the status of the iSCSI service of vserver
cdef.

=over 4

=item Inputs

=item * I<None>

=item Outputs

Returns either "online" or "offline" if the operation was successful.
If an error occurs the error flag is set, errmsg is set to the failure reason,
and false is passed back as the return value.

=item Examples

    $status = $Iscsi_obj->status();

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<stop()>

Stops the iSCSI service of vserver

=over 4

=item Inputs

=item * I<None>

=item Outputs

Returns either true or false if the operation was successful or not.
If an error occurs the error flag is set, errmsg is set to the failure reason,
and false is passed back as the return value.

=item Examples

    $Iscsi_obj->stop();

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<interface_disable()>

Disables the specified iscsi interface

=over 4

=item Inputs

=item * <-lif>

the name of the lif for the interface to disable

=item Outputs

Returns either true or false if the operation was successful or not.
If an error occurs the error flag is set, errmsg is set to the failure reason,
and false is passed back as the return value.

=item Examples

    my $output = $iscsi_obj->interface_disable(-lif => $lif_name);
    if ($iscsi_obj->error()) {
        $Test->nlog("FAIL: failed to disable ", $iscsi_obj->errmsg());
        return $TCD::FAIL;
    }

=back

=head2 B<interface_enable()>

Enables the specified iscsi interface

=over 4

=item Inputs

=item * <-lif>

the name of the lif for the interface to enable

=item Outputs

Returns either true or false if the operation was successful or not.
If an error occurs the error flag is set, errmsg is set to the failure reason,
and false is passed back as the return value.

=item Examples

    my $output = $iscsi_obj->interface_enable(-lif => $lif_name);
    if ($iscsi_obj->error()) {
        $Test->nlog("FAIL: failed to enable ", $iscsi_obj->errmsg());
        return $TCD::FAIL;
    }

=back

=head2 B<delete()>

Deletes the iSCSI service of vserver

=over 4

=item Inputs

=item * I<None>

Set this value to if you want to force delete the iSCSI service.

=item Outputs

Returns either true or false if the operation was successful or not.
If an error occurs the error flag is set, errmsg is set to the failure reason,
and false is passed back as the return value.

=item Examples

    $Iscsi_obj->delete();

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<destroy()>

Deletes the iSCSI service of vserver

=over 4

=item Inputs

=item * I<None>

=item Outputs

Returns either true or false if the operation was successful or not.
If an error occurs the error flag is set, errmsg is set to the failure reason,
and false is passed back as the return value.

=item Examples

    $Iscsi_obj->destroy();

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<modify()>

Modifies iSCSI service of vserver

=over 4

=item Inputs

=item * I<-nodename nodename> (optional)

A string corresponding to a new nodename. Use this option to modify the
iscsi nodename.

=item * I<-alias alias> (optional)

A string corresponding to a new alias. Use this option to modify an
iscsi alias.

=item * I<-status_admin status> (optional)

A string (up|down) corresponding to the new status. Use this option to
modify the iscsi status.

=item Outputs

Returns either true or false if the operation was successful or not.
If an error occurs the error flag is set, errmsg is set to the failure reason,
and false is passed back as the return value.

=item Examples

    $Iscsi_obj->modify(-status_admin => "down");

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<refresh()>

Refreshes the iscsi object attributes. Essentially, it executes a show command
and updates the cached data.

=over 4

=item Inputs

=item * I<None>

=item Outputs

Returns either true or false if the operation was successful or not.
If an error occurs the error flag is set, errmsg is set to the failure reason,
and false is passed back as the return value.

=item Examples

    $Iscsi_obj->refresh();

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<tpgroup_exist()>

Determines if the tpgroup exists

=over 4

=item Inputs

=item * I<-name tpgroup_name>

The name of the tpgroup.

=item Outputs

Returns either true or false if the tpgroup exists or not.
If an error occurs the error flag is set, errmsg is set to the failure reason,
and false is passed back as the return value.

=item Examples

    $Iscsi_obj->tpgroup_name(-name => "tpgroup1");

    if ($Iscsi_obj->error()) {
        $err = $Iscsi_obj->errmsg();
        logcomment("$err");
        return TCD::FAIL;
    }

=back

=head2 B<create_tpgroup_obj(%args)>

Create a tpgroup object. After the tpgroup object is created, it will be
stored in iscsi object.

=over 4

=item Inputs

=item * I<-name> (required)

The tpgroup name to be used

=item * I<-lif> (optional)

The lif name with which this tpgroup has to be created

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. With higher level, it will log more debug messages

=item Outputs

Return Tpgroup object

=item Examples

 # ex. 1: Create tpgroup object
 $tpgroup_obj = $iscsi_obj->create_tpgroup_obj(-name => $tpgroup_name);
 if ($iscsi_obj->error()){
     ...
 }

 # ex. 2: Create tpgroup object with a lif
 $tpgroup_obj = $iscse_obj->create_tpgroup_obj(-name => $tpgroup_name,
                                              -lif => "iscsi_data_lif");
 if ($iscsi_obj->error()){
     ...
 }

=back

=head2 B<get_alias()>

This method, by default, will return the iSCSI alias of the vserver from iscsi
object. Alternatively, this method will return the iscsi alias from iscsi alias
show command if use_alias_cmd flag is set to true

=over 4

=item Inputs

=item * I<-use_alias_cmd> (optional)

If -use_alias_cmd is set to false, then retrieve the value of the iscsi alias
from the cache of the iscsi object, otherwise return the value returned by the
vserver_iscsi_alias cdef.

=item Outputs

Returns iscsi alias name if successful.  For cleared alias, returns "".  For error, returns "" and set_err.

=item Examples:

    # ex. 1: Get alias from iscsi object
    my $alias = $iscsi_obj->get_alias();
    if ($iscsi_obj->error()) {
        ...
    }

    # ex. 2: Get alias from iscsi alias
    my $alias = $iscsi_obj->get_alias(-use_alias_cmd => 1);
    if ($iscsi_obj->error()) {
        ...
    }

=back

=head2 B<set_alias()>

Sets iSCSI alias.

=over 4

=item Inputs

=item * I<-alias> (required)

Name of the iscsi alias.

=item * I<-use_alias_cmd> (optional)

If -use_alias_cmd is set to false, then sets the value of the iscsi alias
from the vserver_iscsi_modify cdef , if its true sets the value using
vserver_iscsi_alias cdef.

=item Outputs

Sets the iSCSI alias using the vserver_iscsi_alias cdef.
If an error occurs the error flag is set, errmsg is set to
the failure reason, and SanUtil::FALSE is passed back as the return value.

=item Examples

    # ex. 1: Sets the alias using iscsi_obj modify()
    $iscsi_obj->set_alias(alias => "new_alias");
    if ($iscsi_obj->error()) {
        ...
    }

    # ex. 2: Sets the alias using iscsi_alias cdef
    $iscsi->set_alias(-alias => "new_alias,
                      -use_alias_cmd => 1);
    if ($iscsi->error()) {
        ...
    }

=back

=head2 B<clear_alias()>

Clear iSCSI alias.

=over 4

=item Inputs

None

=item Outputs

Clears the iSCSI alias. If an error occurs the error flag is set, errmsg is set
to the failure reason, and SanUtil::FALSE is passed back as the return value.

=item Examples

    # ex. 1: Clear the alias
    $iscsi_obj->clear_alias();
    if ($iscsi_obj->error()) {
        ...
    }

=back

=head2 B<tpgroup_list_info()>

Gets a list of tpgroup records on the vserver

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing tpgroup records. If an error occurs, sets the error flag and
returns an empty list.

=item Example1:

    my @tpgroups = $iscsi_obj->tpgroup_list_info();
    if ($iscsi_obj->error()){
     ...
    }

=item Example2:

    my $tpgroups = $iscsi_obj->tpgroup_list_info();
    my @records = @$tpgroups;

    if ($iscsi_obj->error()) {
     ...
    }

=back

=head2 B<initiator_list_info()>

Gets a list of initiator records on the vserver

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
initiator show command.  For a list of current options, try 'iscsi initiator
show ?'.

=item * I<session-type>

session-type can be applied as a filter.  It first calls session_list_info to
filter the sessions, then calls initiator show to filter the initiators.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing initiator records. If an error occurs, sets the error flag and
returns an empty list.

=item Example1:

    my @inits = $iscsi_obj->initiator_list_info();
    if ($iscsi_obj->error()){
     ...
    }

=item Example2:

    my $inits = $iscsi_obj->initiator_list_info();
    my @records = @inits;

    if ($iscsi_obj->error()) {
     ...
    }

=item Example3:

    my @inits = $iscsi_obj->initiator_list_info('-session-type' => "Normal");
    if ($iscsi_obj->error()){
     ...
    }

=back

=head2 B<initiator_create_auth()>

Creates an authentication policy for an initiator

=over 4

=item Inputs

=item * <-vserver>

Virtual Server

=item * <-auth-type>

Authentication method

=item * <-initiator-name>

Initiator Name

=item * I<-user-name> (optional unless CHAP is auth-type)

Inbound CHAP username

=item * I<-password> (optional unless CHAP is auth-type)

Inbound CHAP password

=item * I<-radius> (optional true|false)

Use configured RADIUS server

=item * I<-outbound-user-name> (optional)

Outbound CHAP username

=item * I<-outbound-password> (optional)

Outbound CHAP password

=item Outputs

Creates an authentication policy for an initiator. If an error occurs the
error flag is set, errmsg is set to the failure reason, and
SanUtil::FALSE is passed back as the return value.

=item Example1:

    my $result = $iscsi_obj->initiator_create_auth('-auth-type'       => 'None',
                                                   '-initiator-name'  => 'iqn.1994-05.com.redhat:ssan-rhel5u4-11' );
    if ($iscsi_obj->error()) {
        ...
    }

=back

=head2 B<initiator_modify_auth()>

Modifies an authentication policy for an initiator. Note for cli it just calls iscsi
security modify.  For ZAPI, it may call the zapi for iscsi-initiator-modify-chap-params
or it may call zapis to iscsi-initiator-delete-auth followed by iscsi-initiator-add-auth.

=over 4

=item Inputs

=item * <-vserver> (required)

Virtual Server

=item * <-auth-type> (optional)

Authentication method

=item * <-initiator-name> (required)

Initiator Name

=item * I<-user-name> (optional)

Inbound CHAP username

=item * I<-password> (optional)

Inbound CHAP password

=item * I<-outbound-user-name> (optional)

Outbound CHAP username

=item * I<-outbound-password> (optional)

Outbound CHAP password

=item Outputs

Modifies an authentication policy for an initiator. If an error occurs the
error flag is set, errmsg is set to the failure reason, and
SanUtil::FALSE is passed back as the return value.

=item Example1:

    my $result = $iscsi_obj->initiator_modify_auth('-auth-type'       => 'None',
                                                   '-initiator-name'  => 'iqn.1994-05.com.redhat:ssan-rhel5u4-11' );
    if ($iscsi_obj->error()) {
        ...
    }

=back

=head2 B<initiator_set_default_auth()>

Sets the default authentication policy for a vserver

=over 4

=item Inputs

=item * <-vserver>

Virtual Server

=item * <-auth-type>

Authentication method

=item * I<-user-name> (optional unless CHAP is auth-type)

Inbound CHAP username

=item * I<-password> (optional unless CHAP is auth-type)

Inbound CHAP password

=item * I<-radius> (optional true|false)

Use configured RADIUS server

=item * I<-outbound-user-name> (optional)

Outbound CHAP username

=item * I<-outbound-password> (optional)

Outbound CHAP password

=item * I<-remove-outbound> (optional true|false)

Clear outbound CHAP parameters

=item Outputs

Sets a default authentication policy for a vserver. If an error occurs the
error flag is set, errmsg is set to the failure reason, and
SanUtil::FALSE is passed back as the return value.

=item Example1:

    my $result = $iscsi_obj->initiator_set_default_auth('-auth-type' => 'None' );
    if ($iscsi_obj->error()) {
        ...
    }

=back

=head2 B<initiator_get_default_auth>

Gets a list of iscsi security records on the vserver for default
authentication policies.

=over 4

=item Inputs

=item * I<none>

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing hashes of iscsi security records for default authentication
policies. If an error occurs, sets the error flag and returns an empty list.

=item Example1:

    my @iscsi_securities = $iscsi_obj->initiator_get_default_auth();
    if ($iscsi_obj->error()){
     ...
    }
    $Test->nlog(Dumper(\@iscsi_securities));

=item Example2:

    my $iscsi_securities = $iscsi_obj->initiator_get_default_auth();
    my @records = @$iscsi_securities;

    if ($iscsi_obj->error()) {
     ...
    }
    $Test->nlog(Dumper($iscsi_securities));

=back

=head2 B<initiator_delete_auth()>

Deletes an authentication policy for an initiator

=over 4

=item Inputs

=item * <-vserver>

Virtual Server

=item * <-initiator-name>

Initiator Name

=item Outputs

Deletes an authentication policy for an initiator. If an error occurs the
error flag is set, errmsg is set to the failure reason, and
SanUtil::FALSE is passed back as the return value.

=item Example1:

    my $result = $iscsi_obj->initiator_delete_auth( '-initiator-name' => 'iqn.1994-05.com.redhat:ssan-rhel5u4-11' );
    if ($iscsi_obj->error()) {
        ...
    }

=back

=head2 B<initiator_modify_chap_params()>

This method now just calls initiator_modify_auth. Please read definition for initiator_modify_auth.

=back

=head2 B<auth_generate_chap_password()>

Generate a CHAP password

=over 4

=item Inputs

=item * I<none>

=item Outputs

Generates a CHAP password and returns a string containing the password.
If an error occurs the error flag is set, errmsg is set to the failure
reason, and [] is passed back as the return value.

=item Example1:

    my $password = $iscsi_obj->auth_generate_chap_password();
    if ($iscsi_obj->error()) {
        ...
    }

=back

=head2 B<get_stats()>

Retrieves the statistics for the iSCSI service

=over 4

=item Inputs

=item * I<-vserver> (optional)

Specify the vserver that the iSCSI services stats will be returned

=item * I<-fields> (optional)

Specify what information is to be returned.

=item Outputs

Retrieves a Hash containing all the statistics for the iSCSI service
or just the specified by the input 'fields'.

=item Example1:

    my @stats = $iscsi_obj->get_stats();
    if ($iscsi_obj->error()) {
        ...
    }

=back

=cut
