# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to a tpgroup object
##
## @author vchintal@netapp.com, dl-ssan-test
## @status public

package San::ControllerCluster::Filer10M::Vserver::Iscsi::Tpgroup;

# compiler directives
use strict;

# imported modules
use Params::Validate qw( :all );
# inherit base class
use base qw( San::CommonInterface );

###############################################################################
# Method:    new
# Objective: Create new Tpgroup object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    if (exists $args{name}) {
        $args{tpgroup} = delete $args{name};
    }

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct a tpgroup object
    $self = $class->SUPER::_new(-tpgroup  => $args{tpgroup},
                                -filer => $args{filer});
    if ($self->error()) {
        $self->_set_err("failed to create a tpgroup object", $self->errmsg());
        return $self;
    }

    # dump its hash parameters if debug_level is set
    $self->sanlog(-msg => $self->dumper_get(\%args),
                  -msg_level => 3);

    # store args in tpgroup object and push the keys to list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # we can now remove parent param from the args list
    delete $args{parent_obj};

    # If initialized is set to true like CommonInterface.pm's discover() does, we
    # already have the output of show command and so we don't need to call the create
    # command. Looping through all the args should be good enough in that case
    unless ($args{initialized}) {
        my @input_params = (%args);
        eval {
            # validate input parameters
            (%args) = validate(@input_params,
                              {filer   => { type => SCALAR, required => 1 },
                               vserver => { type => SCALAR, required => 1 },
                               tpgroup => { type => SCALAR, required => 1 },
                               lif     => { type => SCALAR, required => 1,
                                            default => '-' },
                               force   => { type => SCALAR, optional => 1,
                                        regex => /true/ },
                               tpgtag  => { type => SCALAR, optional => 1,
                                            regex => /^\d+$/,
                                            callbacks => {
                                                {'should be with 1..256' => sub { $_[0] >= 1 and $_[0] <= 256 }}
                                            }
                                          }
                              });
        };
        if ($@) {
            $self->_set_err("failed to validate parameter: " . $@);
            return $self;
        }

        # create tp group
        $self->_create_tpgroup(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to _create_tpgroup() ", $self->errmsg());
            return $self;
        }
        $self->{initialized} = $SanUtil::TRUE;
    }

    return $self;
} ## end sub new

###############################################################################
# Method:    delete
# Objective: delete a tpgroup object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    my @input_params = (%args);
    eval {
        # validate input parameters
        (%args) = validate(@input_params,
                          {force   => { type => SCALAR, optional => 1,
                                        regex => qr/true/i, }
                           });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_tpgroup_delete",
             $self->_get_param_defaults());

    # this is a special case between zapi and cli so we have to handle it here
    # zapi used tpgroup-tag instead of tpgroup-name to destroy tpgroup
    # so we ned to get tpgroup tag if cmd interface is zapi
    if ($self->get_cmd_interface() =~ /zapi/i and defined($args{'tpgroup'})) {
        # Delete force flag, no force available with ZAPI
        delete($args{'force'});

        $args{'tpgroup-tag'} = $self->get_tpgroup_tag();
        if ($self->error()) {
            $self->_set_err("failed to get tpgroup tag: ", $self->errmsg());
            return $SanUtil::FALSE;
        }
        delete $args{'tpgroup'};
    }

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    $self->{tpgroup} = "_DELETED_";

    return $SanUtil::TRUE;
} ## end sub delete

###############################################################################
# Method:    destroy
# Objective: wrapper to delete()
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $self->delete(%args);
}

###############################################################################
# Method:    add
# Objective: add iscsi lif to tpgroup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);
    my @input_params = (%args);

    # validation_options is set to allow extra params not defined in %args to bypass
    Params::Validate::validation_options( allow_extra => 1);

    eval {
        # validate input parameters
        (%args) = validate(@input_params,
                          {lif     => { type => SCALAR, required => 1},
                           force   => { type => SCALAR, optional => 1,
                                        regex => qr/true/, }
                           });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_tpgroup_add",
             $self->_get_param_defaults());

    # this is a special case between zapi and cli so we have to handle here
    # tpgroup-tag is needed for zapi so we have to delete tpgroup-name
    if ($self->get_cmd_interface() =~ /zapi/i and defined($args{'tpgroup'})) {
        delete $args{'tpgroup'};
    }

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    $self->refresh();
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to refresh " . $self->errmsg());
        return $self;
    }

    return $SanUtil::TRUE;
} ## end sub add

###############################################################################
# Method:    remove()
# Objective: remove iscsi lif from tpgroup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    my @input_params = (%args);

    # validation_options is set to allow extra params not defined in %args to bypass
    Params::Validate::validation_options( allow_extra => 1);

    eval {
        # validate input parameters
        (%args) = validate(@input_params,
                          {lif     => { type => SCALAR, required => 1},
                           force   => { type => SCALAR, optional => 1,
                                        regex => qr/true/, }
                           });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_tpgroup_remove",
             $self->_get_param_defaults());

    # this is a special case between zapi and cli so we have to handle here
    # tpgroup-tag is needed for zapi so we have to delete tpgroup-name
    if ($self->get_cmd_interface() =~ /zapi/i and defined($args{'tpgroup'})) {
        delete $args{'tpgroup'};
    }

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    $self->refresh();
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to refresh " . $self->errmsg());
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub remove

###############################################################################
# Method:    get_list_info()
# Objective: Get tpgroup information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_tpgroup_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update tpgroup object data
    my @records = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return {};
    } else {
        return $records[0];
    }
} ## end sub get_list_info

###############################################################################
# Method:    modify()
# Objective: modify tpgroup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    my @input_params = (%args);
    eval {
        # validate input parameters
        (%args) = validate(@input_params,
                          {lif     => { type => SCALAR, optional => 1},
                           tpgtag  => { type => SCALAR, optional => 1,
                                        regex => qr/^\d+$/,
                                        callbacks => {
                                            {'should be with 1..256' => sub { $_[0] >= 1 and $_[0] <= 256 }}
                                        }
                                      },
                           force   => { type => SCALAR, optional => 1,
                                        regex => qr/true/, }
                           });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_tpgroup_modify",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    $self->refresh();
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to refresh " . $self->errmsg());
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub modify

###############################################################################
# Method:    refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_iscsi_tpgroup_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update tpgroup object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method:    get_name()
# Objective: Get tpgroup name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "tpgroup";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_parent_obj()
# Objective: Get parent object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_vserver()
# Objective: Get vserver name
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_vserver {
    my ($self, %args) = @_;

    $args{field} = "vserver";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_tpgroup_tag()
# Objective: Get tpgroup tag
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_tpgroup_tag {
    my ($self, %args) = @_;

    $args{field} = "tpgroup_tag";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

###############################################################################
# Method:    get_lif()
# Objective: Get lifs
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lif {
    my ($self, %args) = @_;

    $args{field} = "lif";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method:    _get_param_defaults
# Objective: Gets default portset parameters
################################################################################
sub _get_param_defaults {
    my ($self, %args) = @_;

    # default hash parameters to be used in portset object
    my %defaults_params = (%args, tpgroup => $self->get_name(-use_cache => 1),
                           vserver => $self->get_vserver(-use_cache => 1));

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Tpgroup
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # set use_cache flag if exists
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "vserver_iscsi_tpgroup_show",
                                                $self->_get_param_defaults(),
                                                %args);
    }
}

################################################################################
# Method:    _create_tpgroup
# Objective: create tpgroup
################################################################################
sub _create_tpgroup {
    my ($self, %args) = @_;

    # append cmd for tpgroup create to argument list
    %args = (%args, cmd => "vserver_iscsi_tpgroup_create");

    my $lif;
    # zapi creates tpgroup without LIF. it must be added after tpgroup is created
    # so for zapi, we need to save lif param and add afterward
    if (exists($args{lif}) and $self->get_cmd_interface() =~ /zapi/i) {
        $lif = delete $args{lif};
    }

    # use common invoke from base class (San::CommonInterface)
    # to translate cdef into UI cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # add lif/interface for zapi
    if ($self->get_cmd_interface() =~ /zapi/i and $lif ne "-") {

        # it needs tpgroup-tag to add interface
        # get it from @output
        my $tp_tag;
        if (exists($output[0]->{'tpgroup-tag'})) {
            $tp_tag = $output[0]->{'tpgroup-tag'};
        } else {
            $self->_set_err("tpgroup create did not return tpgroup-tag ");
            return $SanUtil::FALSE;
        }

        # Have to add one LIF at a time with ZAPI calls
        foreach my $lif_single (split(/\,/, $lif)) {
            $self->add(-lif => $lif_single, "-tpgroup-tag" => $tp_tag);
            if ($self->error()) {
                # set error and bubble up
                $self->_set_err("failed to add lif ($lif_single) to tpgroup ", $self->errmsg());
                return $SanUtil::FALSE;
            }
        }
    }

    # call tpgroup refresh to populate object. This is required since we
    # don't have all the params populated at this point. refresh runs show
    # command and populates the rest of params too
    # call volume refresh if enable cache is set
    if ($self->{enable_cache}) {
        $self->refresh();
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to refresh " . $args{tpgroup}, $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    return $SanUtil::TRUE;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Tpgroup.pm

=head1 Author

Manoj Chintala (dl-ssan-test)

=head1 Description

Provides an interface to tpgroup

=head1 See Also

Tpgroup package used base class B<San::CommonInterface> that also inherit other modules. See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm>

=head1 Attributes

=head2 B<filer>

string = controller name

=head2 B<name>

string = name of tpgroup object to be created

=head2 B<debug_level>

interger = 0 - 100 (default is 0)

=head1 B<Public Methods>

=head2 B<new(%args)>

Create a tpgroup object

=over 4

=item Inputs

=item * I<-filer> => <controller name>

Tpgroup name

=item * I<-name|-tpgroup> => <name of tpgroup object>

Tpgroup object name to be created

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. With higher level, it will log more debug messages

=item Outputs

Return Tpgroup object

=item Examples

 # ex. 1: Create tpgroup directly, make sure you have lif, vserver
 use San::ControllerCluster::Filer10M::Vserver::Iscsi::Tpgroup;

 my $filer = "ssan-3170";
 my $lif = "iscsi_data_lif";
 my $vserver = "vs0";

 my tpgroup_obj = new Tpgroup(-filer => $filer,
                              -tpgroup => $tpgroup,
                              -vserver => $vserver,
                              -lif => $lif,
                              -debug_level => 0);
 if ($tpgroup_obj->error()) {
    logcomment("Failed to create tpgroup object" . $tpgroup->errmsg());
    return $TCD::FAIL;
 }

 # ex. 2: Create tpgroup using Iscsi.pm for existing tpgroup. We populate the
 data from tpgroup show and exit in this case instead of running tpgroup
 create command. Lets assume that auto_tpgroup exists on the controller

 my $filer = "ssan-3170";
 my $lif = "iscsi_data_lif";
 my $vserver = "vs0";
 my $tpgroup = "auto_tpgroup";

 my $iscsi_obj = new Iscsi(-filer => $filer,
                           -vserver => $vserver);

 if ($iscsi_obj->tpgroup_exist(-tpgroup => $tpgroup)) {
    my $tpgroup_obj = $iscsi_obj->get_tpgroup_objs(-tpgroup => $tpgroup);
    if ($tpgroup_obj->error()) {
        logcomment("Failed to create tpgroup object" . $tpgroup->errmsg());
        return $TCD::FAIL;
    }
 }

=back

=head2 B<add(%args)>

Add lif to a target object

=over 4

=item Inputs

=item * I<lif> (required)

The lif to be added. Multiple lifs can be passed with by separating them with comma

=item * I<-force>

Used to specify to go ahead with adding lifs to tpgroup even if there any active
sessions

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 my $lif1 = "auto_lif1";
 $tpgroup_obj->add(-lif => $lif1);
 if ($tpgroup_obj->error()){
     ...
 }

=back

=head2 B<remove(%args)>

Remove lif from a target object

=over 4

=item Inputs

=item * I<lif> (required)

The lif to be removed. Multiple lifs can be passed with by separating them with comma

=item * I<-force>

Used to specify to go ahead with removing lifs from tpgroup even if there any active
sessions

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 my $lif1 = "auto_lif1";
 $tpgroup_obj->remove(-lif => $lif1);
 if ($tpgroup_obj->error()){
     ...
 }

=back

=head2 B<modify(%args)>

Modify target object

=over 4

=item Inputs

=item * I<lif>

lif name

=item * I<-force>

Used to specify to go ahead with modifying tpgroup even if there any active
sessions

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 $tpgroup_obj->modify(-lif => $lif1);
 if ($tpgroup_obj->error()){
     ...
 }

=back

=head2 B<get_list_info(%args)>

Returns hash reference to tpgroup show command with keys being fields
and values being their values in show command

=over 4

=item Inputs

None

=item Outputs

Returns hash ref if successful, otherwise sets error and returns empty hash ref

=item Examples

 my $rslt_ref = $tpgroup_obj->get_list_info();
 if ($tpgroup_obj->error()){
     ...
 } else {
     $Test->nlog("-- Display tpgroup info using get_list_info");
        $Test->nlog("\tName:  " . $rslt_ref->{tpgroup},
                   "\tTPG Tag:  " . $rslt_ref->{tpgtag},
                   "\tLif: " . $rslt_ref->{lif}
                   );
 }

=back

=head2 B<refresh()>

Refresh and update Tpgroup object

=over 4

=item Inputs

None

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Refresh and update tpgroup object
 $tpgroup_obj->refresh();
 if ($tpgroup_obj->error()){
     ...
 }

=back

=head2 B<delete(%args)>

Delete target object

=over 4

=item Inputs

=item * I<-force>

Used to specify to go ahead with destroying tpgroup object even if there any active
sessions

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 $tpgroup_obj->delete();
 if ($tpgroup_obj->error()){
     ...
 }

=back

=head2 B<get_name()>

Get tpgroup name

=over 4

=item Inputs

None

=item Outputs

Returns tpgroup name

=item Examples

 # ex. 1: Display tpgroup name
 logcomment("Tpgroup name: " . $tpgroup_obj->get_name());

=back

=head2 B<get_vserver()>

Get the containing vserver name of the tpgroup

=over 4

=item Inputs

None

=item Outputs

Returns containing vserver name of the tpgroup

=item Examples

 # ex. 1: Display tpgroup name
 logcomment("Tpgroup name: " . $tpgroup_obj->get_name());

=back

=head2 B<get_lif()>

Get lifs of the tpgroup

=over 4

=item Inputs

None

=item Outputs

Returns lifs of this tpgroup

=item Examples

 # ex. 1: Display lifs of the tpgroup
 logcomment("Lifs assinged to the tpgroup: " . $tpgroup_obj->get_lif());

=back

=head2 B<get_tpgroup_tag()>

Get the tpgroup tag

=over 4

=item Inputs

None

=item Outputs

Returns tpgroup tag

=item Examples

 # ex. 1: Display tpgroup tag
 logcomment("Tpgroup tag of the tpgroup: " . $tpgroup_obj->get_tpgroup_tag());

=back

