# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to a Lif (Logical I/F) object
##
## @author dl-ssan-test
## @status public

package San::ControllerCluster::Filer10M::Vserver::Lif;

# compiler directives
use strict;
use Data::Dumper;
use Params::Validate qw( SCALAR );
use Params::Validate qw( :all );

# inherit base class
use base qw( San::CommonInterface );

###############################################################################
# Method Name: new
# Objective: Create new Logical I/F object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self);
    my (%validate_params);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }

    # convert common args to args specifically for creating lif
    if (exists $args{name}) {
        $args{lif} = delete $args{name};
    }
    if (exists $args{server}) {
        $args{vserver} = delete $args{server};
    }
    if (exists $args{interface}) {
        $args{'home-port'} = delete $args{interface};
    }
    if (exists $args{ip}) {
        $args{address} = delete $args{ip};
    }
    if (exists $args{protocol}) {
        $args{'data-protocol'} = delete $args{protocol};
    }

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct a lif object
    $self = $class->SUPER::_new(-name  => $args{lif},
                                -filer => $args{filer},
                                %args);

    if ($self->error()) {
        $self->_set_err("failed to create a lif object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # store lif elements on a list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # we may need the parent object later
    # at this time, we pass it in but unused
    # delete unneeded argument from the list
    delete $args{parent_obj};
    delete $args{debug_level};

    # if the obj has been initialized, just refresh to populate object
    # otherwise initialize it by create and then update its object
    unless ($args{initialized}) {
        my @input_params = (%args);

        # validation_options is set to allow extra params not defined in %args to bypass
        Params::Validate::validation_options( allow_extra => 1);

        # validate_params is the arguments to validate
        %validate_params = (
                          filer        => { type     => SCALAR,
                                            required => 1 },
                          vserver      => { type     => SCALAR,
                                            required => 1 },
                          lif          => { type     => SCALAR,
                                            required => 1 },
                          'home-port'  => { type     => SCALAR,
                                            required => 1 },
                          'home-node'  => { type     => SCALAR,
                                            optional => 1,
                                            default => $args{filer} },
                          address      => { type     => SCALAR,
                                            optional => 1 },
                          netmask      => { type     => SCALAR,
                                            optional => 1 },
                          role         => { type     => SCALAR,
                                            optional => 1,
                                            default => "data" },
                          'data-protocol'  => { type     => SCALAR,
                                            required => 1 },
                          );
        eval {
            (%args) = Params::Validate::validate(@input_params, \%validate_params);
        };
        if ($@) {
            $self->_set_err("failed to validate parameter: ". $@);
            return $self;
        }

        # Both netmask and address are required if the protocol is iSCSI or FCOE
        if ($args{'data-protocol'} =~ /iscsi|fcoe/i) {
            unless (defined $args{address} and length $args{address}) {
                $self->_set_err("mandatory parameter \"address\" not specified ");
                return $self;
            }
            unless (defined $args{netmask} and length $args{netmask}) {
                 delete($args{netmask});
            }
        } elsif ($args{'data-protocol'} =~ /fcp/i) {
            # in a fibrechannel case, we dont need the netmask or address fields
            # delete them from the args hash, because they may have been passed in
            # with null values.
            delete($args{netmask});
            delete($args{address});
        }

        # append cmd for network interface create to argument list
        %args = (%args, cmd => "network_interface_create");

        # use common invoke from base class (San::CommonInterface)
        # to translate cdef into UI cmd and execute it
        $self->SUPER::invoke_cmd_obj(%args);

        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
            return $self;
        }

        # call refresh to populate object if enable_cache is set
        if ($self->{enable_cache}) {
            $self->refresh(%args);
            if ($self->error()) {
                # set error and bubble up
                $self->_set_err("failed to refresh object " . $args{lif},
                                $self->errmsg());
                return $self;
            }
        }

        # set intitialized flag
        $self->{initialized} = $SanUtil::TRUE;
    }

    return $self;
} ## end sub new

###############################################################################
# Method Name: delete
# Objective: Delete a logical interface.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_delete",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # something to indicate that this object has been deleted
    $self->{lif} = "_DELETE_";

    return $SanUtil::TRUE;
} ## end sub delete

###############################################################################
# Method Name: refresh
# Objective: Display logical interfaces
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_show",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub show

###############################################################################
# Method Name: migrate
# Objective: Migrate a logical interface to a different port
# Details:   See POD documentation at the end of this file.
###############################################################################
sub migrate {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_migrate",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub migrate

###############################################################################
# Method Name: rename
# Objective: Rename a logical interface
# Details:   See POD documentation at the end of this file.
###############################################################################
sub rename {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # regarding which interface is specified, this method is only for cli
    $args{cmd_interface} = "cli";

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_rename",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    } else {
        # rename was successful so we need to update the object with newname
        $self->set(lif => $args{newname});
    }

    return $SanUtil::TRUE;
} ## end sub rename

###############################################################################
# Method Name: revert
# Objective: Revert a logical interface to its home port
# Details:   See POD documentation at the end of this file.
###############################################################################
sub revert {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_revert",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub revert

###############################################################################
# Method Name: show_routing_group
# Objective: A wrapper for get_routing_group to support 7-mode like calls
# Details:   See POD documentation at the end of this file.
###############################################################################
sub show_routing_group {
    my ($self, %args) = @_;

    return $self->get_routing_group(%args);
} ## end sub show_routing_group

################################################################################
# Method Name: route_exist
# Objective: Determines if a route to given destination
# Details:   See POD documentation at the end of this file.
################################################################################
sub route_exist {
    my ($self, %args) = @_;

    my (@route_records);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);
    # initialize cmd parameters
    %args = (%args,
             cmd => "network_routing_groups_route_show",
             routing_group => $self->get_routing_group(),
             vserver => $self->get_vserver(-use_cache => 1));

    (@route_records) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke cmd obj ",
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    if (scalar(@route_records)) {
        return $SanUtil::TRUE;
    } else {
        return $SanUtil::FALSE;
    }
}

################################################################################
# Method Name: create_route
# Objective: Creates route
# Details:   See POD documentation at the end of this file.
################################################################################
sub create_route {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory gateway argument
    unless (defined $args{gateway} and length $args{gateway}) {
        $self->_set_err("mandatory parameter \"-gateway\" not specified ");
        return $SanUtil::FALSE;
    }

    if ((not defined (defined $args{destination})) or
         not length($args{destination})){
        # Check for ipv6 or ipv4 routes

        if ($args{gateway} =~ /:/){
            $args{destination} = '::/0';
            if ($self->route_exist(-destination => $args{destination})) {
                return $SanUtil::TRUE;
            }
            if ($self->error()) {
                $self->_set_err("failed to verify if ipv6 route exists ",
                            $self->dumper_get(\%args, $self->get_debug_level()),
                            $self->errmsg());
                return $SanUtil::FALSE;
            }
        }else{
            $args{destination} = '0.0.0.0/0';
            # verify if ipv4 route to the destination exists
            if ($self->route_exist(-destination => $args{destination})) {
                return $SanUtil::TRUE;
            }
            if ($self->error()) {
                $self->_set_err("failed to verify if ipv4 route exists ",
                            $self->dumper_get(\%args, $self->get_debug_level()),
                            $self->errmsg());
                return $SanUtil::FALSE;
            }
        }
    } else {
        if ($self->route_exist(-destination => $args{destination})) {
            return $SanUtil::TRUE;
        }
        if ($self->error()) {
          $self->_set_err("failed to verify if route exists ",
                            $self->dumper_get(\%args, $self->get_debug_level()),
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # initialize cmd parameters
    %args = (%args,
             cmd => "network_routing_groups_route_create",
             routing_group => $self->get_routing_group(),
             vserver => $self->get_vserver(-use_cache => 1));

    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke cmd obj ",
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: get_active_connections
# Objective: Gets the list of active connections
# Details:   See POD Documentation at the end of this file
###############################################################################
sub get_active_connections {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "network_connections_active_show",
             donot_append_vserver => 1,
             $self->_get_param_defaults());

    delete $args{vserver} if (exists $args{vserver});
    delete $args{lif} if (exists $args{lif});


    my $cmd_obj = $self->get_cmd_obj();
    if ($self->error()) {
        $self->_set_err("failed to get cmd object ", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get a CmdsDriver object and execute the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub get_active_connections

###############################################################################
# Method Name: get_active_connections_clients
# Objective: Displays client connection info
# Details:   See POD Documentation at the end of this file
###############################################################################
sub get_active_connections_clients {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "network_connections_active_show_clients",
             donot_append_vserver => 1,
             $self->_get_param_defaults());

    delete $args{vserver} if (exists $args{vserver});
    delete $args{lif} if (exists $args{lif});


    my $cmd_obj = $self->get_cmd_obj();
    if ($self->error()) {
        $self->_set_err("failed to get cmd object ", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get a CmdsDriver object and execute the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub get_active_connections_clients

###############################################################################
# Method Name: get_active_connections_protocols
# Objective: Displays info on protocols being used on active connections
# Details:   See POD Documentation at the end of this file
###############################################################################
sub get_active_connections_protocols {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "network_connections_active_show_protocols",
             donot_append_vserver => 1,
             $self->_get_param_defaults());

    delete $args{vserver} if (exists $args{vserver});
    delete $args{lif} if (exists $args{lif});

    my $cmd_obj = $self->get_cmd_obj();
    if ($self->error()) {
        $self->_set_err("failed to get cmd object ", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get a CmdsDriver object and execute the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub get_active_connections_protocols

###############################################################################
# Method Name: get_active_connections_services
# Objective: Displays info on services being used on active connections
# Details:   See POD Documentation at the end of this file
###############################################################################
sub get_active_connections_services {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "network_connections_active_show_services",
             donot_append_vserver => 1,
             $self->_get_param_defaults());

    delete $args{vserver} if (exists $args{vserver});
    delete $args{lif} if (exists $args{lif});

    my $cmd_obj = $self->get_cmd_obj();
    if ($self->error()) {
        $self->_set_err("failed to get cmd object ", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get a CmdsDriver object and execute the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub get_active_connections_services

###############################################################################
# Method Name: get_active_connections_lifs
# Objective: Displays info on virtual interfaces and nodes or virtual servers
# being used by active connections
# Details:   See POD Documentation at the end of this file
###############################################################################
sub get_active_connections_lifs {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "network_connections_active_show_lifs",
             donot_append_vserver => 1,
             $self->_get_param_defaults());

    delete $args{lif} if (exists $args{lif});
    delete $args{vserver} if (exists $args{vserver});

    my $cmd_obj = $self->get_cmd_obj();
    if ($self->error()) {
        $self->_set_err("failed to get cmd object ", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get a CmdsDriver object and execute the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub get_active_connections_lifs

###############################################################################
# Method Name: get_listening_connections
# Objective: Displays information on listening network connections
# Details:   See POD Documentation at the end of this file
###############################################################################
sub get_listening_connections {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "network_connections_listening_show",
             donot_append_vserver => 1,
             $self->_get_param_defaults());

    delete $args{vserver} if (exists $args{vserver});
    delete $args{lif} if (exists $args{lif});

    my $cmd_obj = $self->get_cmd_obj();
    if ($self->error()) {
        $self->_set_err("failed to get cmd object ", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Get a CmdsDriver object and execute the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub get_listening_connections

###############################################################################
# Method Name: get_zones
# Objective: Gets the DNS zone name of each lif for each virtual server
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_zones {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_show_zones",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? (): [];
    }

    # If there are no entries, return an empty list.
    if (($output[0] =~ /There are no entries/i) or
        ($output[0]->{dns_zone} =~ /none/i)) {
        return (wantarray()) ? (): [];
    }

    my @dns_zones = ();
    foreach my $dns_zone (@output) {
        if (defined ($dns_zone->{dns_zone})) {
            push (@dns_zones, $dns_zone->{dns_zone});
        }
    }

    return (wantarray()) ? @dns_zones: \@dns_zones;
} ## end sub get_zones

###############################################################################
# Method Name: show_zones
# Objective: A wrapper for get_zones to support 7-mode like calls
# Details:   See POD documentation at the end of this file.
###############################################################################
sub show_zones {
    my ($self, %args) = @_;

    return $self->get_zones(%args);
} ## end sub show_zones

###############################################################################
# Method Name: start_cluster_check
# Objective: Start the cluster check function
# Details:   See POD documentation at the end of this file.
###############################################################################
sub start_cluster_check {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_start_cluster_check",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub start_cluster_check

###############################################################################
# Method Name: get_failover_rules
# Objective: Get logical interface failover rules
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_failover_rules {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_show",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    my (@failover_rules) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? (): [];
    }

    # If there are no entries, return an empty list.
    if ($failover_rules[0] =~ /There are no entries/i) {
        return (wantarray()) ? (): [];
    }

    return (wantarray()) ? @failover_rules: \@failover_rules;
} ## end sub get_failover_rules

###############################################################################
# Method Name: failover_show
# Objective: 7-mode wrapper for get_failover_rules
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_show {
    my ($self, %args) = @_;

    return $self->get_failover_rules(%args);
} ## end sub failover_show

###############################################################################
# Method Name: failover_create
# Objective:  Create a logical interface failover rule
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_create {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_create",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub failover_create

###############################################################################
# Method Name: failover_append
# Objective: Append a logical interface failover rule
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_append {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_append",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub failover_append

###############################################################################
# Method Name: failover_delete
# Objective: Append a logical interface failover rule
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_delete",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub failover_delete

###############################################################################
# Method Name: failover_change_priority
# Objective: Change the priority of a logical interface's failover rule
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_change_priority {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_change_priority",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub failover_change_priority

###############################################################################
# Method Name: failover_insert
# Objective: Insert a logical interface failover rule
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_insert{
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_insert",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub failover_insert

###############################################################################
# Method Name: failover_modify
# Objective: Modify a logical interface failover rule
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_modify{
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_modify",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub failover_modify

###############################################################################
# Method Name: failover_groups_create
# Objective: Create a new failover group or add an entry to an existing group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_groups_create {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_groups_create",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub failover_groups_create

###############################################################################
# Method Name: failover_groups_remove
# Objective: Remove a port from a failover group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_groups_remove {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_groups_remove",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub failover_groups_remove

###############################################################################
# Method Name: failover_groups_remove
# Objective: Rename a network interface failover group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_groups_rename {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_groups_rename",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub failover_groups_rename

###############################################################################
# Method Name: get_failover_groups
# Objective: Get network interface failover groups
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_failover_groups {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_failover_groups_show",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    my (@failover_groups) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? (): [];
    }

    # If there are no entries, return an empty list.
    if ($failover_groups[0] =~ /There are no entries/i) {
        return (wantarray()) ? (): [];
    }

    return (wantarray()) ? @failover_groups: \@failover_groups;
} ## end sub get_failover_groups

###############################################################################
# Method Name: failover_groups_show
# Objective: 7-mode wrapper for get_failover_groups
# Details:   See POD documentation at the end of this file.
###############################################################################
sub failover_groups_show {
    my ($self, %args) = @_;

    return $self->get_failover_groups;
} ## end sub failover_groups_show

###############################################################################
# Method Name: modify
# Objective:   Modifies an iscsi or fcp Lif
# Details:     See POD documentation at the end of this file.
###############################################################################
sub modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_modify",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # Need to refresh (which calls net int show) to cache the new values,
    # otherwise, if a user calls an accessor method (i.e. get_admin_status), it
    # will get the previous value.
    $self->refresh();
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub modify

################################################################################
# Method Name: get_address()
# Objective:   Get address
# Details:     See POD documentation at the end of this file
################################################################################
sub get_address {
    my ($self, %args) = @_;

    $args{field} = "address";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_name()
# Objective:   Get lif name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "lif";
    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_netmask()
# Objective:   Get netmask
# Details:     See POD documentation at the end of this file
################################################################################
sub get_netmask {
    my ($self, %args) = @_;

    $args{field} = "netmask";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_netmask_length()
# Objective:   Get netmask_length
# Details:     See POD documentation at the end of this file
################################################################################
sub get_netmask_length {
    my ($self, %args) = @_;

    $args{field} = "netmask_length";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_curr_node()
# Objective:   Get current node
# Details:     See POD documentation at the end of this file
################################################################################
sub get_curr_node {
    my ($self, %args) = @_;

    $args{field} = "curr_node";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_curr_port()
# Objective:   Get curr_port
# Details:     See POD documentation at the end of this file
################################################################################
sub get_curr_port {
    my ($self, %args) = @_;

    $args{field} = "curr_port";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_data_protocol()
# Objective:   Get data_protocol
# Details:     See POD documentation at the end of this file
################################################################################
sub get_data_protocol {
    my ($self, %args) = @_;

    $args{field} = "data_protocol";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_home_node()
# Objective:   Get home_node
# Details:     See POD documentation at the end of this file
################################################################################
sub get_home_node {
    my ($self, %args) = @_;

    $args{field} = "home_node";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_home_port()
# Objective:   Get home_port
# Details:     See POD documentation at the end of this file
################################################################################
sub get_home_port {
    my ($self, %args) = @_;

    $args{field} = "home_port";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get parent object (Vserver) from cache
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get parent object from cache
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_role()
# Objective:   Get role
# Details:     See POD documentation at the end of this file
################################################################################
sub get_role {
    my ($self, %args) = @_;

    $args{field} = "role";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_routing_group()
# Objective:   Get routing_group
# Details:     See POD documentation at the end of this file
################################################################################
sub get_routing_group {
    my ($self, %args) = @_;

    $args{field} = "routing_group";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_status_oper()
# Objective:   Get status_oper
# Details:     See POD documentation at the end of this file
################################################################################
sub get_status_oper {
    my ($self, %args) = @_;

    $args{field} = "status_oper";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_tpgroup()
# Objective:   Get tpgroup
# Details:     See POD documentation at the end of this file
################################################################################
sub get_tpgroup {
    my ($self, %args) = @_;

    # numberic ID is tpgroup
    $args{field} = "numeric_id";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_vserver()
# Objective:   Get vserver this lif belong to
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vserver {
    my ($self, %args) = @_;

    $args{field} = "vserver";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_admin_status()
# Objective:   Get the admin status of a lif
# Details:     See POD documentation at the end of this file
################################################################################
sub get_admin_status {
    my ($self, %args) = @_;

    $args{field} = "status_admin";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: is_at_home()
# Objective:   Is lif at home port?
# Details:     See POD documentation at the end of this file
################################################################################
sub is_at_home {
    my ($self, %args) = @_;

    $args{field} = "is_home";

    # the _get_field() will handle error
    if ($self->_get_field(%args) =~ /true/i) {
        return $SanUtil::TRUE;
    } else {
        return $SanUtil::FALSE;
    }
}

################################################################################
# Method Name: delete_routing_group
# Objective: Deletes a routing group
################################################################################
sub delete_routing_group {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters. Note that we don't need the lif key so
    # don't call _get_param_defaults
    %args = (%args,
             cmd => "network_routing_groups_delete",
             cmd_interface => "cli",
             server => $self->get_vserver());

    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: get_wwpn()
# Objective:   Get wwpn
# Details:     See POD documentation at the end of this file
################################################################################
sub get_wwpn {
    my ($self, %args) = @_;

    $args{field} = "wwpn";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method Name: _get_param_defaults
# Objective: Gets default lif parameters
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    # default hash parameters to be used in Lif object
    # the get accessor method must obtain from cache to prevent infinite loop
    my %defaults_params = (lif => $self->get_name(-use_cache => 1),
                           vserver => $self->get_vserver(-use_cache => 1));

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Lif
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # convert common args to args specifically for creating lif
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "network_interface_show",
                                                field => $args{field},
                                                $self->_get_param_defaults());
    }
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Lif.pm

=head1 Author

dl-ssan-automation

=head1 Description

Lif.pm provides an interface to manage a lif in Clustered environment

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<San::CommonInterface>

=head1 Attributes

=head2 B<filer> (required)

string = controller name

=head2 B<vserver> (required)

string = vserver name to create Lif

=head2 B<name | lif> (required)

string = lif name to be created

=head2 B<interface> (required)

string = interface to be used to create lif

=head2 B<'data-protocol'> (optional)

string = iscsi or fcp. Default is fcp

=head2 B<ip | address> (required for iscsi Lif)

string = IP address

=head2 B<netmask> (required for iscsi Lif)

string = netmask

=head1 Public Methods

=head2 B<new(%args)>

Creates a new Lif object.  Uses San::CommonInterface to create the object. Users
may pass as a parameter a vserver. If a vserver is not specified, it uses 'vs0'
as default.

=over 4

=item Inputs

=item * I<-filer>

Name of filer

=item * I<-vserver> (default: 'vs0').

Name of the vserver to create Lif.

=item * I<-name | -lif>

Name of lif to be created

=item * I<-interface>

Physical interface to map Lif to

=item * I<'-data-protocol'>

Protocol type for the Lif to be created

=item Outputs

Returns an Lif object. The error flag will be set if something fails.

=item Examples

 # ex 1: create an fcp Lif object using vserver 'vs1'

 $lif_obj = new San::ControllerCluster::Filer10M::Vserver::Lif(filer => $filer,
                                                               name => $lif_name,
                                                               interface => 5a,
                                                               'data-protocol' => "fcp",
                                                               vserver=> 'vs1');
 if ($lif_obj->error()) {
    $Test->nlog("ERROR: Failed to instantiate lif object " . $lif_obj->errmsg());
    return $TCD::FAIL;
 }

 # ex 2: create an iscsi Lif object using default vserver

 $lif_obj = new San::ControllerCluster::Filer10M::Vserver::Lif(filer => $filer,
                                                               name => $lif_name,
                                                               interface => 5a,
                                                               'data-protocol' => "fcp",
                                                               address => '9.0.120.8',
                                                               netmask => '255.252.0.0');
 if ($lif_obj->error()) {
    $Test->nlog("ERROR: Failed to instantiate lif object " . $lif_obj->errmsg());
    return $TCD::FAIL;
 }

=head2 B<delete()>

=back

=head2 B<route_exist(%args)>

Returns 1 if the route exists to given destination, otherwise 0

=over 4

=item Inputs

=item * I<destination>

destination in CIDF (<IpAddress/Mask>) format

=item Outputs

Returns 1 if route to given destination exists, otherwise 0

=item Examples

 my $route_exists = $lif_obj->route_exist(-destination => "0.0.0.0/0");
 if ($lif_obj->error()){
     ...
 }

=back

=head2 B<create_route(%args)>

Returns 1 if the route is created successfully, otherwise 0

=over 4

=item Inputs

=item * I<gateway>

gateway for the route

=item * I<destination>

destination in CIDF (<IpAddress/Mask>) format

=item Outputs

Returns 1 if route is successfully created, otherwise 0

=item Examples

 $lif_obj->create_route(-gateway => "10.225.114.1");
 if ($lif_obj->error()){
     ...
 }

=back

=head2 B<get_active_connections()>

Gets the list of active connections

=over 4

=item Inputs

=item * node (optional):

Node name.

=item * cid (optional):

Connection Id.

=item * vserver (optional):

Vserver.

=item * lif-name (optional):

Logical Interface Name.

=item * local-address (optional):

Local IP address.

=item * local-port (optional):

Local port.

=item * remote-ip (optional):

Remote Ip Address.

=item * remote-host (optional):

Remote Host.

=item * remote-port (optional):

Remote Port.

=item * proto (optional):

Protocol.

=item * lifid (optional):

Logical Interface Id.

=item * service (optional):

Protocol Service.

=item * lru (optional):

Least Recently Used.

=item * blocks-lb (optional):

Connection Blocks LB Migrate.


=item Outputs

Returns a list of active connections and their information.

=item Examples

  my @act_cons = $fcp_obj->get_active_connections();
  if ($fcp_obj->error() or $status == 0) {
        $Test->nlog("ERROR: Failed to get active connections ");
        return $TCD::FAIL;
  }

=back

=head2 B<get_active_connections_clients()>

Displays client connection info

=over 4

=item Inputs

=item * node (optional):

Node name.

=item * remote-address (optional):

Remote Ip Address.

=item * count (optional):

Client Count.


=item Outputs

Returns a list of clients and their information.

=item Examples

  my @act_clients = $fcp_obj->get_active_connections_clients();
  if ($fcp_obj->error() or $status == 0) {
        $Test->nlog("ERROR: Failed to get clients with active connections");
        return $TCD::FAIL;
  }

=back

=head2 B<get_active_connections_protocols()>

Displays info on protocols being used on active connections

=over 4

=item Inputs

=item * node (optional):

Node name.

=item * proto (optional):

Protocol.

=item * count (optional):

Client Count.


=item Outputs

Returns a list of protocols and their information.

=item Examples

  my @act_protos = $fcp_obj->get_active_connections_protocols();
  if ($fcp_obj->error() or $status == 0) {
        $Test->nlog("ERROR: Failed to get protocols with active connections ");
        return $TCD::FAIL;
  }

=back

=head2 B<get_active_connections_services()>

Displays info on services being used on active connections

=over 4

=item Inputs

=item * node (optional):

Node name.

=item * service (optional):

Protocol Service.

=item * count (optional):

Client Count.


=item Outputs

Returns a list of services and their information.

=item Examples

  my @act_services = $fcp_obj->get_active_connections_services();
  if ($fcp_obj->error() or $status == 0) {
        $Test->nlog("ERROR: Failed to get services with active connections ");
        return $TCD::FAIL;
  }

=back

=head2 B<get_active_connections_lifs()>

Displays info on virtual interfaces and nodes or virtual servers with active connections

=over 4

=item Inputs

=item * node (optional):

Node name.

=item * vserver (optional):

Vserver.

=item * lif-name (optional):

Logical Interface Name.

=item * count (optional):

Client Count.

=item * blocked-count (optional):

LB Blocking Count.


=item Outputs

Returns a list of lifs and their information.

=item Examples

  my @list_cons = $fcp_obj->get_active_connections_lif();
  if ($fcp_obj->error() or $status == 0) {
        $Test->nlog("ERROR: Failed to get lifs with active connections ");
        return $TCD::FAIL;
  }

=back

=head2 B<get_listening_connections()>

Displays information on listening network connections

=over 4

=item Inputs

=item * node (optional):

Node name.

=item * mgmt-cid (optional):

Management Connection Id.

=item * vserver (optional):

Vserver.

=item * cid (optional):

System Connection Id.

=item * lif-name (optional):

logical Interface name.

=item * local-address (optional):

Local IP address.

=item * local-port (optional):

Local port.

=item * remote-ip (optional):

Remote Ip Address.

=item * remote-host (optional):

Remote Host.

=item * remote-port (optional):

Remote Port.

=item * proto (optional):

Protocol.

=item * lifid (optional):

Logical Interface Id.

=item * service (optional):

Protocol Service.

=item * lru (optional):

Least Recently Used.


=item Outputs

Returns a list of listening network connections and their information.

=item Examples

  my @list_cons = $fcp_obj->get_listening_connections();
  if ($fcp_obj->error() or $status == 0) {
        $Test->nlog("ERROR: Failed to get listening connections ");
        return $TCD::FAIL;
  }

=back

=head2 B<get_wwpn()>

Gets the wwpn of the lif from object cache or controller.

=over 4

=item Inputs

=item * I<-use_cache> (optional)

Flag to retrieve wwpn from cache if set to true. Default is false.

=item Outputs

Returns the lif's wwpn.

=item Example1:

    my $wwpn = $lif_obj->get_wwpn();
    if ($lif_obj->error()){
     ...
    }

=back
