# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to a Portset object
##
## @author dl-ssan-automation
## @status public

package San::ControllerCluster::Filer10M::Vserver::Portset;

# compiler directives
use strict;

# inherit base class
use base qw( San::CommonInterface );

###############################################################################
# Method Name: new
# Objective: Create new Portset object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, @nodes, $client_obj, $ostype);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    if (exists $args{name}) {
        $args{portset} = $args{name};
        delete $args{name};
    }

    # verify mandatory cmd argument
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct a portset object
    $self = $class->SUPER::_new(-name  => $args{portset},
                                -filer => $args{filer},
                                %args);

    if ($self->error()) {
        $self->_set_err("failed to create a portset object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # store portset elements on a list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # we may need the parent object later
    # at this time, we pass it in but unused
    # delete unneeded argument from the list
    delete $args{parent_obj};
    delete $args{debug_level};

    # if command had been created, skip executing the create command
    unless ($args{initialized}) {
        # use common invoke from base class (San::CommonInterface)
        # to translate cdef into UI cmd and execute it
        $self->_create_portset(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to _create_portset() ", $self->errmsg());
            return $self;
        }
    }

    return $self;
} ## end sub new

###############################################################################
# Method Name: add
# Objective: add iscsi tpgroup or fcp wwpn/lif to a portset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub add {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_portset_add",
             $self->_get_param_defaults());

    # invoke command via cmd object
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub add

###############################################################################
# Method Name: delete
# Objective: delete a portset object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmdsDriver parameters
    %args = (%args,
             cmd => "lun_portset_delete",
             $self->_get_param_defaults());

    # invoke command via cmd object
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # set portset/name attributes to indicate portset has been deleted
    $self->{name} = "_DELETE_";
    $self->{portset} = "_DELETE_";

    return $SanUtil::TRUE;
} ## end sub delete

###############################################################################
# Method Name: destroy
# Objective: destroy portset object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self, %args) = @_;

    return $self->delete(%args);
}

###############################################################################
# Method Name: get_list_info()
# Objective: Get portset information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_portset_show",
             $self->_get_param_defaults());

    # invoke command via cmd object
    my @records = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return {};
    } else {
        return $records[0];
    }
} ## end sub get_list_info

###############################################################################
# Method Name: refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_portset_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update portset object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: remove()
# Objective: remove iscsi tpgroup or fcp wwpn/lif from a portset
# Details:   See POD documentation at the end of this file.
###############################################################################
sub remove {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmdsDriver parameters
    %args = (%args,
             cmd => "lun_portset_remove",
             $self->_get_param_defaults());

    # invoke command via cmd object
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub remove

################################################################################
# Method Name: get_fc_lif()
# Objective:   Get fc_lif
# Details:     See POD documentation at the end of this file
################################################################################
sub get_fc_lif {
    my ($self, %args) = @_;

    $args{field} = "fc_lif";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_name()
# Objective:   Get name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "portset";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get parent object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_protocol()
# Objective:   Get protocol
# Details:     See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self, %args) = @_;

    $args{field} = "protocol";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_protocol_type()
# Objective:   Wrapper to get_protocol
# Details:     See POD documentation at the end of this file
################################################################################
sub get_protocol_type {
    my ($self, %args) = @_;

    # wrapper to get_protocol
    return $self->get_protocol(%args);
}

################################################################################
# Method Name: get_tpg()
# Objective:   Get target portal group
# Details:     See POD documentation at the end of this file
################################################################################
sub get_tpg {
    my ($self, %args) = @_;

    $args{field} = "tpg";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_vserver()
# Objective:   Get vserver
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vserver {
    my ($self, %args) = @_;

    $args{field} = "vserver";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_wwpn()
# Objective:   Get wwpn
# Details:     See POD documentation at the end of this file
################################################################################
sub get_wwpn {
    my ($self, %args) = @_;

    $args{field} = "wwpn";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method Name: _get_param_defaults
# Objective: Gets default portset parameters
################################################################################
sub _get_param_defaults {
    my ($self) = @_;

    # default hash parameters to be used in portset object
    my %defaults_params = (portset => $self->get_name(-use_cache => 1),
                           vserver => $self->get_vserver(-use_cache => 1));

    return %defaults_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Portset
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache =  $SanUtil::FALSE;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # set use_cache flag if exists
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => "lun_portset_show",
                                                field => $args{field},
                                                $self->_get_param_defaults());
    }
}

################################################################################
# Method Name: _create_portset
# Objective: create portset
################################################################################
sub _create_portset {
    my ($self, %args) = @_;

    # append cmd for volume create to argument list
    %args = (%args, cmd => "lun_portset_create");

    my $port_names = "";
    # zapi creates portsets without LIF's. it must be added after portset is created
    # For zapi, we need to save lif argument and add afterwards
    if (exists($args{'port-name'}) and $self->get_cmd_interface() =~ /zapi/i) {
        $port_names = delete $args{'port-name'};
    }

    # invoke command via cmd object
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Add in the port-names specified
    if ($port_names ne "" and $self->get_cmd_interface() =~ /zapi/i) {
        # Protocol is not needed when adding LIF's
        delete($args{'protocol'});

        # Have to add a port name one at a time
        foreach my $port (split(/\,/, $port_names)) {
            $self->add(%args, "port-name" => $port);
            if ($self->error()) {
                $self->_set_err("failed to add $port to portset " . $args{'portset'});
                return $SanUtil::FALSE;
            }
        }
    }

    # call portset refresh to populate object if use_cache flag is enabled
    if ($self->{enable_cache}) {
        $self->refresh();
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to refresh " . $args{portset}, $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # set initialized flag
    $self->{initialized} = $SanUtil::TRUE;

    return $SanUtil::TRUE;
}

1;
