# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to a Volume object
##
## @author korry@netapp.com, dl-ssan-test
## @status Shared

package San::ControllerCluster::Filer10M::Vserver::Volume;

# Compiler directives
use strict;

# Imported modules
use San::ControllerCluster::Filer10M::Vserver::Volume::Lun;
use San::ControllerCluster::Filer10M::Vserver::Volume::Qtree;
use San::ControllerCluster::Filer10M::Vserver::Volume::Snapshot;
use Data::Dumper;

# Inherit base class
use base qw( San::CommonInterface
             San::ControllerCluster::Filer10M::Vserver::VolumeCompat
             San::ControllerCluster::Filer10M::Vserver::VolumeHelper);

my @Sub_components = qw( lun qtree snapshot );

###############################################################################
# Method Name: new
# Objective: Create new Volume object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self, @list_names);

    # Parameter initialization
    $self = {};

    # Bless instance into class
    bless $self, $class;

    # Remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # Verify mandatory cmd argument
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }

    # Use base class (San::CommonInterface) via SUPER::_new method
    # to contruct volume object
    $self = $class->SUPER::_new(-name  => $args{volume},
                                -filer => $args{filer},
                                %args);
    if ($self->error()) {
        $self->_set_err("failed to contruct volume object", $self->errmsg());
        return $self;
    }

    # Dump only first depth of hash if debug_level < 3
    # Dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # Store Volume elements on a list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # Create get object accessor method for lun
    $self->create_accessor_methods(-type => "get_obj",
                                   -pkg  => __PACKAGE__,
                                   -list => \@Sub_components);

    # We may need the parent object later
    # at this time, we pass it in but delete
    # delete unneeded argument from the list
    delete $args{parent_obj};
    delete $args{debug_level};

    # If the obj has been initialized, just refresh to update info
    # otherwise initialize it by create and update its object
    unless ($args{initialized}) {
        # Volume has not been initialized/created, so we need to create it
        $self->_create_volume(%args);
        if ($self->error()) {
            # Set error and bubble up
            $self->_set_err("failed to _create_volume() ", $self->errmsg());
            return $self;
        }
    }

    return $self;
} ## end sub new

###############################################################################
# Method Name: copy
# Objective: copy volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub copy {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{'destination-volume'} and length $args{'destination-volume'}) {
        $self->_set_err("Mandatory parameter \"-destination-volume\" not specified ");
        return $self;
    }
    my $dst_vol = $args{'destination-volume'};

    my $dst_aggr = (defined $args{'destination-aggregate'}) ? $args{'destination-aggregate'} : $self->get_aggregate();

    # initialize cmd parameters
    %args = (%args,
             cmd                    => "volume_copy_start",
             'destination-volume'    => $dst_vol,
             'destination-aggregate' => $dst_aggr,
             #may need to update to allow override if operation take a long time.
             'foreground'            => 'true',
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    delete $args{volume} if $args{volume};
    delete $args{vserver} if $args{vserver};
    delete $args{parent_obj} if $args{parent_obj};
    $self->sanlog(-msg => "Volume:Copy - Copy a Volume " ,
                      -msg_level => 2);
    my $vol_copy_obj = San::ControllerCluster::Filer10M::Vserver::Volume->new(
        volume     => $dst_vol,
        filer      => $self->get_filer(),
        vserver    => $self->get_vserver(),
        parent_obj => $self->get_parent_obj());

    $vol_copy_obj->refresh();
    if ($vol_copy_obj->error()) {
            # set error and bubble up
            $self->_set_err("Failed to refresh " . $vol_copy_obj->get_name(), $vol_copy_obj->errmsg());
            return $SanUtil::FALSE;
        }
    $self->refresh();
    if ($self->error()) {
            # set error and bubble up
            $self->_set_err("Failed to refresh " . $self->get_name(), $self->errmsg());
            return $SanUtil::FALSE;
        }
    return $vol_copy_obj;
} ## end sub copy

###############################################################################
# Method Name: create_lun
# Objective:  Create lun. This method maps one-to-one with 7-mode
# Details:    See POD documentation at the end of this file.
###############################################################################
sub create_lun {
    my ($self, %args) = @_;

    return ($self->create_lun_obj(%args));
}

###############################################################################
# Method Name: create_lun_obj
# Objective:  Create lun object
# Details:    See POD documentation at the end of this file.
###############################################################################
sub create_lun_obj {
    my ($self, %args) = @_;

    my ($lun_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # convert parameters to maintain compatibility w/ 7mode
    if (exists $args{type}) {
        $args{ostype} = $args{type};
        delete $args{type};
    }
    if (exists $args{name}) {
        $args{lun} = $args{name};
        delete $args{name};
    }
    # if reserve param is set to 0, set space-resv to noreserve
    # NOTE: this feature has not yet implemented, ignore for now
    #if (exists $args{reserve} and $args{reserve} == 0) {
    #    $args{"space-resv"} = "noreserve";
    #    delete $args{reserve};
    #}
    if (exists $args{reserve}) {
        delete $args{reserve};
    }

    # append vserver and volume parameters to argument list
    %args = (%args, $self->_get_param_defaults());

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # instantiate lun object
    $lun_obj =
      San::ControllerCluster::Filer10M::Vserver::Volume::Lun->new(-parent_obj => $self,
                                                                  -filer  => $self->get_filer(),
                                                                  %args);
    if ($lun_obj->error()) {
        $self->_set_err("failed to instantiate lun object", $lun_obj->errmsg());
        return {};
    } else {
        # store lun object in filer object
        push @{$self->{lun_objs}}, $lun_obj;
    }
    return $lun_obj;
} ## end sub create_lun_obj

###############################################################################
# Method Name: create_qtree_obj
# Objective:  Create qtree object
# Details:    See POD documentation at the end of this file.
###############################################################################
sub create_qtree_obj {
    my ($self, %args) = @_;

    my ($qtree_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # convert parameters to maintain compatibility w/ 7mode
    $args{qtree} = delete $args{name} if (exists $args{name});

    # Check for the required qtree argument
    unless (defined $args{qtree} and length $args{qtree}) {
        $self->_set_err("Mandatory parameter \"qtree\" not specified ");
        return {};
    }
    # append vserver and volume parameters to argument list
    %args = (%args,
             $self->_get_param_defaults());

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # instantiate qtree object
    $qtree_obj =
      San::ControllerCluster::Filer10M::Vserver::Volume::Qtree->new(-parent_obj => $self,
                                                                    -filer  => $self->get_filer(),
                                                                    %args);
    if ($qtree_obj->error()) {
        $self->_set_err("failed to instantiate qtree object", $qtree_obj->errmsg());
        return {};
    } else {
        # store qtree object in filer object
        push @{$self->{qtree_objs}}, $qtree_obj;
    }
    return $qtree_obj;
} ## end sub create_qtree_obj

###############################################################################
# Method Name: create_snapshot_obj
# Objective:   Create snapshot object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_snapshot_obj {
    my ($self, %args) = @_;

    my ($snapshot_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # convert parameters to maintain compatibility w/ 7mode
    $args{snapshot} = delete $args{name} if (exists $args{name});

    # Check for the required snapshot argument
    unless (defined $args{snapshot} and length $args{snapshot}) {
        $self->_set_err("Mandatory parameter \"snapshot\" not specified ");
        return {};
    }
    # append vserver and volume parameters to argument list
    %args = (%args,
             $self->_get_param_defaults());

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # instantiate snapshot object
    $snapshot_obj =
      San::ControllerCluster::Filer10M::Vserver::Volume::Snapshot->new(-parent_obj => $self,
                                                                       -filer      => $self->get_filer(),
                                                                       %args);
    if ($snapshot_obj->error()) {
        $self->_set_err("failed to instantiate snapshot object", $snapshot_obj->errmsg());
        return {};
    } else {
        # store snapshot object in filer object
        push @{$self->{snapshot_objs}}, $snapshot_obj;
    }
    return $snapshot_obj;
} ## end sub create_snapshot_obj

###############################################################################
# Method Name: create_file
# Objective:  Create file on volume
# Details:    See POD documentation at the end of this file.
###############################################################################
sub create_file {
    my ($self, %args) = @_;

    my $size;
    my $file_name;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # check for required arguments
    unless (defined $args{size} and length $args{size}){
        $self->_set_err("Mandatory parameter \"size\" not specified ");
        return $SanUtil::FALSE;
    }
    $size = delete $args{size};

    unless (defined $args{from_file} and length $args{from_file}){
        $self->_set_err("Mandatory parameter \"from_file\" not specified ");
        return $SanUtil::FALSE;
    }
    $file_name = delete $args{from_file};

    # Get the node obj that is the parent of this volume
    # First get_parent gets the vserver. The second gets
    # the node from the vserver
    my $parent_vserver = $self->get_parent_obj();
    if ($self->error()) {
        $self->_set_err("failed to get parent vserver object ", $self->errmsg());
        return $SanUtil::FALSE;
    }
    my $parent_node = $parent_vserver->get_parent_obj();
    if ($parent_vserver->error()) {
        $self->_set_err("failed to get parent node object ", $parent_vserver->errmsg());
        return $SanUtil::FALSE;
    }

    # Get the aggregate object that this volume belongs to.
    # Need the aggregate to figure out which node owns the volume
    my $aggr_obj = $parent_node->get_aggregate_obj(name => $self->get_aggregate());
    if ($parent_node->error()) {
        $self->_set_err("failed to get aggregate object ", $parent_node->errmsg());
        return $SanUtil::FALSE;
    }

    # Finally, get the node that owns the volume from the aggr
    #### When striped multiple nodes will get returned.  Code may no longer work,
    #### Need to eventually find a cmode context to mkfile and rm file
    my $node = $aggr_obj->get_nodes();
    if ($aggr_obj->error()) {
        $self->_set_err("failed to get nodes ", $aggr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Create a command line
    my $cmd = "mkfile ".$size." /vol/".$self->get_name()."/".$file_name;

    # Run the command in node scope (cli only, no zapi) to create the file
    $self->SUPER::run_node_scope(-filer => $node,
                                 -priv  => "diag",
                                 -cmd   => $cmd,
                                 %args);
    if ($self->error()) {
        $self->_set_err("failed to run_node_scope " . $cmd,
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub create_file

###############################################################################
# Method Name: remove_file
# Objective:  remove file on volume
# Details:    See POD documentation at the end of this file.
###############################################################################
sub remove_file {
    my ($self, %args) = @_;

    my $file_name;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # check for required arguments
    unless (defined $args{from_file} and length $args{from_file}){
        $self->_set_err("Mandatory parameter \"from_file\" not specified ");
        return $SanUtil::FALSE;
    }
    $file_name = delete $args{from_file};

    # Get the node obj that is the parent of this volume
    # First get_parent gets the vserver. The second gets
    # the node from the vserver
    my $parent_vserver = $self->get_parent_obj();
    if ($self->error()) {
        $self->_set_err("failed to get parent vserver object ", $self->errmsg());
        return $SanUtil::FALSE;
    }
    my $parent_node = $parent_vserver->get_parent_obj();
    if ($parent_vserver->error()) {
        $self->_set_err("failed to get parent node object ", $parent_vserver->errmsg());
        return $SanUtil::FALSE;
    }

    # Get the aggregate object that this volume belongs to.
    # Need the aggregate to figure out which node owns the volume
    my $aggr_obj = $parent_node->get_aggregate_obj(name => $self->get_aggregate());
    if ($parent_node->error()) {
        $self->_set_err("failed to get aggregate object ", $parent_node->errmsg());
        return $SanUtil::FALSE;
    }

    # Finally, get the node that owns the volume from the aggr
    #### When striped multiple nodes will get returned.  Code may no longer work,
    #### Need to eventually find a cmode context to mkfile and rm file
    my $node = $aggr_obj->get_nodes();
    if ($aggr_obj->error()) {
        $self->_set_err("failed to get nodes ", $aggr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Create a command line
    my $cmd = "rm /vol/".$self->get_name()."/".$file_name;

    # Run the command in node scope (cli only, no zapi) to remove the file
    $self->SUPER::run_node_scope(-filer => $node,
                                 -priv  => "diag",
                                 -cmd   => $cmd,
                                 %args);
    if ($self->error()) {
        $self->_set_err("failed to run_node_scope " . $cmd,
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub remove_file

###############################################################################
# Method Name: delete
# Objective: delete volume object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    my $force = $SanUtil::FALSE;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # determine whether force flag exist or not
    if (exists $args{force}) {
        $force = $SanUtil::TRUE;
        delete $args{force};
    }

    # offline volume first if it's force delete
    if ($force) {
        $self->offline();
        if ($self->error()) {
            # error is already set, bubble up
            return $SanUtil::FALSE;
        }
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_delete",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # korry - volume is deleted but the object is still exist
    # vserver object must re-discover or refresh to update volume objects
    # set volume to "_DELETE_" to indicate volume has been deleted
    $self->{volume} = "_DELETE_";

    return $SanUtil::TRUE;
} ## end sub delete

###############################################################################
# Method Name: destroy
# Objective:   destroy volume object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self, %args) = @_;

    # provide compatible method to 7-mode
    return $self->delete(%args);
}

################################################################################
# Method Name: destroy_lun()
# Objective:   Destroy lun
# Details:     See POD documentation at the end of this file
################################################################################
sub destroy_lun {
    my ($self, %args) = @_;

    my (@lun_objs, @tmp_objs);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    if ((defined $args{force}) and $args{force} =~ /(1|true)/) {
        $args{force} = "true";
    }

    # convert lun to name to maintain compatability w/ 7mode
    if (exists $args{name}) {
        $args{lun} = delete $args{name};
    }

    # push default parameters
    %args = ($self->_get_param_defaults(%args));

    # verify mandatory name param
    unless (defined $args{lun} and length $args{lun}) {
        $self->_set_err("mandatory parameter \"-lun\" not specified ");
        return $SanUtil::FALSE;
    }

    # get lun objs from parent
    @lun_objs = $self->get_luns();
    if ($self->error()) {
        $self->_set_err("failed to find lun objs ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # delete lun object that matches and update the list
    foreach my $lun_obj (@lun_objs) {
        # delete un that matches
        if ($lun_obj->get_name() =~ /^$args{lun}$/) {
            $lun_obj->destroy(%args);
            if ($lun_obj->error()) {
                $self->_set_err("failed to destroy lun: " . $args{lun},
                                $lun_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {
            push (@tmp_objs, $lun_obj);
        }
    }

    # store the list back to the parent
    $self->{lun_objs} = \@tmp_objs;

    return $SanUtil::TRUE;
} ## end sub destroy_lun

################################################################################
# Method Name: destroy_qtree_obj()
# Objective:   Destroy qtree
# Details:     See POD documentation at the end of this file
################################################################################
sub destroy_qtree_obj {
    my ($self, %args) = @_;

    my (@qtree_objs, @tmp_objs);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    if ((defined $args{force}) and $args{force} =~ /(1|true)/) {
        $args{force} = "true";
    }

    # convert qtree to name to maintain compatability w/ 7mode
    if (exists $args{name}) {
        $args{qtree} = delete $args{name};
    }

    # push default parameters
    %args = ($self->_get_param_defaults(%args));

    # verify mandatory name param
    unless (defined $args{qtree} and length $args{qtree}) {
        $self->_set_err("mandatory parameter \"-qtree\" not specified ");
        return $SanUtil::FALSE;
    }

    # get qtree objs
    if (defined $self->{qtree_objs} and scalar(@{$self->{qtree_objs}})) {
        @qtree_objs = @{$self->{qtree_objs}};
    } else {
        # discover and get qtree objects
        @qtree_objs = $self->get_qtree_objs();
        if ($self->error()) {
            $self->_set_err("failed to get qtree objects", $self->errmsg());
            return {};
        }
    }

    # delete qtree object that matches and update the list
    foreach my $qtree_obj (@qtree_objs) {
        # delete qtree that matches
        if ($qtree_obj->get_name() =~ /^$args{qtree}$/) {
            $qtree_obj->delete(%args);
            if ($qtree_obj->error()) {
                $self->_set_err("failed to delete qtree: " . $args{qtree},
                                $qtree_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {
            push (@tmp_objs, $qtree_obj);
        }
    }

    # store the list back to the parent
    $self->{qtree_objs} = \@tmp_objs;

    return $SanUtil::TRUE;
} ## end sub destroy_qtree_obj

################################################################################
# Method Name: destroy_snapshot_obj()
# Objective:   Destroy snapshot
# Details:     See POD documentation at the end of this file
################################################################################
sub destroy_snapshot_obj {
    my ($self, %args) = @_;

    my (@snapshot_objs, @tmp_objs);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert snapshot to name to maintain compatability w/ 7mode
    if (exists $args{name}) {
        $args{snapshot} = delete $args{name};
    }

    # push default parameters
    %args = ($self->_get_param_defaults(%args));

    # verify mandatory snapshot param
    unless (defined $args{snapshot} and length $args{snapshot}) {
        $self->_set_err("mandatory parameter \"-snapshot\" not specified ");
        return $SanUtil::FALSE;
    }

    # get snapshot objs
    if (defined $self->{snapshot_objs} and scalar(@{$self->{snapshot_objs}})) {
        @snapshot_objs = @{$self->{snapshot_objs}};
    } else {
        # discover and get snapshot objects
        @snapshot_objs = $self->get_snapshot_objs();
        if ($self->error()) {
            $self->_set_err("failed to get snapshot objects", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # delete snapshot object that matches and update the list
    foreach my $snapshot_obj (@snapshot_objs) {
        # delete snapshot that matches
        if ($snapshot_obj->get_name() =~ /^$args{snapshot}$/) {
            $snapshot_obj->delete(%args);
            if ($snapshot_obj->error()) {
                $self->_set_err("failed to delete snapshot: " . $args{snapshot},
                                $snapshot_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {
            push (@tmp_objs, $snapshot_obj);
        }
    }

    # store the list back to the parent
    $self->{snapshot_objs} = \@tmp_objs;

    return $SanUtil::TRUE;
} ## end sub destroy_snapshot_obj

################################################################################
# Method Name: get_lun()
# Objective:   Get lun object.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self, %args) = @_;

    my ($lun_obj, %tmp_args);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Hate this crap but we have to make it compatible w/ 7-mode. This means
    # we have to take scalar as param instead of hash

    # To maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    # Allows us to take a scalar and use it as a hash, or just take a hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|lun)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # this is a scalar
            $tmp_args{name} = $key;
        }
        last;
    }

    # set tmp args back to args
    %args = %tmp_args;

    if (exists $args{lun}) {
        $args{name} = $args{lun};
        delete $args{lun};
    }

    # check if full lun path include filer name was
    # given.  If so extract the lun name.
    if ($args{name} =~ /^\//) {
        $args{name} = File::Basename::basename($args{name});
    }

    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"\$name or -name => \$name\" is required");
        return {};
    }

    # if cache flag is set, get it from cache
    # if it does exist in cache, query the controller for the particular volume
    if ($self->{use_cache} && exists $self->{lun_objs}) {
        # get from cache
        ($lun_obj) = grep ({$_->get_name() =~ /^$args{name}$/} @{$self->{lun_objs}});
    } else {
        # Discover this particular volume on the controller
        # Get lun objects
        $lun_obj = $self->get_lun_objs(-name => $args{name});
        if ($self->error()) {
            $self->_set_err("failed to get lun object: " . $args{name},
                            $self->errmsg());
            return {};
        }
    }

    unless (ref($lun_obj) =~ /lun/i) {
        $self->_set_err("can not find lun " . $args{name});
        return {};
    }

    return $lun_obj;
} ## end sub get_lun

################################################################################
# Method Name: get_qtree_obj()
# Objective:   Get qtree object.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_qtree_obj {
    my ($self, %args) = @_;

    my (@qtree_objs, %tmp_args);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Hate this crap but we have to make it compatible w/ 7-mode. This means
    # we have to take scalar as param instead of hash

    # To maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    # Allows us to take a scalar and use it as a hash, or just take a hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|qtree|use_cache|cmd_interface)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # this is a scalar
            $tmp_args{name} = $key;
        }
        last;
    }

    # set tmp args back to args
    %args = %tmp_args;

    if (exists $args{qtree}) {
        $args{name} = $args{qtree};
        delete $args{qtree};
    }

    # check if full qtree path include filer name was
    # given.  If so extract the qtree name.
    if ($args{name} =~ /^\//) {
        $args{name} = File::Basename::basename($args{name});
    }

    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"\$name or -name => \$name\" is required");
        return {};
    }

    # discover qtrees and retrieve specific qtree object
    if (defined $self->{qtree_objs} and scalar(@{$self->{qtree_objs}})) {
        @qtree_objs = @{$self->{qtree_objs}};
    } else {
        # discover and get qtree objects
        @qtree_objs = $self->get_qtree_objs();
        if ($self->error()) {
            $self->_set_err("failed to get qtree objects", $self->errmsg());
            return {};
        }
    }

    # find specific qtree object
    foreach my $qtree_obj (@qtree_objs) {

        return $qtree_obj
          if ($qtree_obj->get_name() =~ /^$args{name}$/);
    }

    # no qtree object that matches the criteria
    $self->_set_err("can not find qtree " . $args{name});
    return {};
} ## end sub get_qtree_obj

################################################################################
# Method Name: get_snapshot_obj()
# Objective:   Get snapshot object.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_snapshot_obj {
    my ($self, %args) = @_;

    my (@snapshot_objs, %tmp_args);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # hate this crap but we have to make it compatible w/ 7-mode. This means
    # we have to take scalar as param instead of hash

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    # Allows us to take a scalar and use it as a hash, or just take a hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|snapshot)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
            last;
        } else {
            # this is a scalar
            $tmp_args{name} = $key;
        }
        last;
    }

    # set tmp args back to args
    %args = %tmp_args;

    if (exists $args{snapshot}) {
        $args{name} = $args{snapshot};
        delete $args{snapshot};
    }

    # check if full snapshot path include filer name was
    # given.  If so extract the snapshot name.
    if ($args{name} =~ /^\//) {
        $args{name} = File::Basename::basename($args{name});
    }

    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"\$name or -name => \$name\" is required");
        return {};
    }

    # discover snapshots and retrieve specific snapshot object
    if (defined $self->{snapshot_objs} and scalar(@{$self->{snapshot_objs}})) {
        @snapshot_objs = @{$self->{snapshot_objs}};
    } else {
        # discover and get snapshot objects
        @snapshot_objs = $self->get_snapshot_objs();
        if ($self->error()) {
            $self->_set_err("failed to get snapshot objects", $self->errmsg());
            return {};
        }
    }

    # find specific snapshot object
    foreach my $snapshot_obj (@snapshot_objs) {
        return $snapshot_obj
          if ($snapshot_obj->get_name() =~ /^$args{name}$/);
    }

    # no snapshot object that matches the criteria
    $self->_set_err("can not find snapshot " . $args{name});
    return {};
} ## end sub get_snapshot_obj

################################################################################
# Method Name: get_lun_maxsize()
# Objective:   Get lun maxsize
# Details:     See POD documentation at the end of this file
################################################################################
sub get_lun_maxsize {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # get volume path
    $args{path} = $self->create_common_path(volume => $self->get_name());
    if ($self->error()) {
        $self->_set_err("Could not get volume path", $self->errmsg());
        return {};
    }

    # check for mandatory parameters
    unless (defined $args{ostype} and length $args{ostype}) {
        $self->_set_err("Mandatory parameter \"ostype\" not specified ");
        return {};
    }

    # Initialize cmd parameters
    %args = (%args,
             cmd => "lun_maxsize",
             $self->_get_param_defaults());

    # delete name and volume since we will be using path
    delete $args{name} if ($args{name});
    delete $args{volume} if ($args{volume});

    # Get a CmdsDriver object and executes the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return {};
    }

    if (scalar(@output) == 1) {
        return $output[0];
    } else {
        return {};
    }
} ## end sub get_lun_maxsize

################################################################################
# Method Name: get_lun_minsize()
# Objective:   Get lun minsize
# Details:     See POD documentation at the end of this file
################################################################################
sub get_lun_minsize {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # check for mandatory parameters
    unless (defined $args{ostype} and length $args{ostype}) {
        $self->_set_err("Mandatory parameter \"ostype\" not specified ");
        return "";
    }

    # Initialize cmd parameters
    %args = (%args,
             cmd => "lun_minsize",
             $self->_get_param_defaults());

    # delete name and volume since we will be using path
    delete $args{name} if ($args{name});
    delete $args{volume} if ($args{volume});

    # regarding which interface is specified, this method is only for zapi
    $args{cmd_interface} = "zapi";

    # Get a CmdsDriver object and executes the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    if (defined $output[0]->{'min-size'}) {
        return $output[0]->{'min-size'};
    } else {
        return "";
    }
} ## end sub get_lun_minsize

################################################################################
# Method Name: get_luns
# Objective:   return array of lun objects
# Details:     See POD documentation at the end of this file
################################################################################
sub get_luns {
    my ($self) = @_;

    $self->_clear_err();

    # discover and get lun objects
    my (@lun_objs) = $self->get_lun_objs();
    if ($self->error()) {
        $self->_set_err("failed to get lun objects", $self->errmsg());
        return (wantarray()) ? (): [];
    }

    return (wantarray()) ? @lun_objs : \@lun_objs;
} ## end sub get_luns

################################################################################
# Method Name: lun_exist()
# Objective:   Determines if lun exists. This maps one-to-one
#              to 7-mode
# Details:     See POD documentation at the end of this file
################################################################################
sub lun_exist {
    my ($self, %args) = @_;

    my (@output, %tmp_args);

    # let' not clear_err here because if lun does not exist, we still
    # have the original error
    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|lun)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # this is a scalar
            $tmp_args{lun} = $key;
        }
        last;
    }

    # set tmp_args back to args
    %args = %tmp_args;

    # convert vserver to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{name}) {
        $args{lun} = $args{name};
        delete $args{name};
    }

    # verify name param
    unless (defined $args{lun} and length $args{lun}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return $SanUtil::FALSE;
    }


    # build cmd vserver show
    %args = (%args,
             cmd => "lun_show",
             volume => $self->get_name(),
             $self->_get_param_defaults());

    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{path} = '/vol/' . delete $args{volume};
        if (exists($args{qtree}) and $args{qtree} ne '""') {
            $args{path} .= '/' . delete $args{qtree};
        }
        $args{path} .= '/' . delete $args{lun};
    }

    # invoke cmd command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke cmd \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub lun_exist

################################################################################
# Method Name: qtree_exist()
# Objective:   Determines if qtree exists. This maps one-to-one
#              to 7-mode
# Details:     See POD documentation at the end of this file
################################################################################
sub qtree_exist {
    my ($self, %args) = @_;

    my (@output, %tmp_args);
    my ($qtree, $volume);

    # Let's not clear_err here because if qtree does not exist, we still
    # have the original error
    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # To maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|qtree)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
            last;
        } else {
            # this is a scalar
            $tmp_args{qtree} = $key;
        }
        last;
    }

    # Set tmp_args back to args
    %args = %tmp_args;

    # Convert vserver to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{name}) {
        $args{qtree} = $args{name};
        delete $args{name};
    }

    # Verify name param
    unless (defined $args{qtree} and length $args{qtree}) {
        $self->_set_err("mandatory parameter \"-qtree\" not specified ");
        return $SanUtil::FALSE;
    }

    # Build cmd qtree show
    %args = (%args,
             cmd => "volume_qtree_show",
             $self->_get_param_defaults());

    #  The ZAPI wants a path not a volume and qtree.
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $qtree = $args{qtree};
        $volume = $args{volume};
        $args{path} = '/vol/' . delete $args{volume};
        if (exists($args{qtree}) and $args{qtree} ne '""') {
            $args{path} .= '/' . delete $args{qtree};
        }
    }

    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # If the interface is zapi, you will need to do a foreach
    # and check the output and match the qtree/volume. Steal the next if
    # code from fetch qtree.
    if ($self->get_cmd_interface() =~ /zapi/i) {
        foreach my $entry (@output) {
            next if ($entry->{"qtree-id"} == 0); # skip volumes
            if (($entry->{qtree} eq $qtree) and ($entry->{volume} eq $volume)) {
                return $SanUtil::TRUE;
            }
        } # end foreach entry
        return $SanUtil::FALSE;
    } else {
        if (scalar @output) {
            return $SanUtil::TRUE;
        } else {
            return $SanUtil::FALSE;
        }
    }
} ## end sub qtree_exist

################################################################################
# Method Name: snapshot_exist()
# Objective:   Determines if snapshot exists. This maps one-to-one
#              to 7-mode
# Details:     See POD documentation at the end of this file
################################################################################
sub snapshot_exist {
    my ($self, %args) = @_;

    my (@output, $cdef, %tmp_args);

    # let' not clear_err here because if snapshot does not exist, we still
    # have the original error
    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|snapshot)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
            last;
        } else {
            # this is a scalar
            $tmp_args{snapshot} = $key;
        }
        last;
    }

    # set tmp_args back to args
    %args = %tmp_args;

    # convert vserver to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{name}) {
        $args{snapshot} = $args{name};
        delete $args{name};
    }

    # verify name param
    unless (defined $args{snapshot} and length $args{snapshot}) {
        $self->_set_err("mandatory parameter \"-snapshot\" not specified ");
        return $SanUtil::FALSE;
    }

    # build cmd snapshot show
    %args = (%args,
             cmd => "volume_snapshot_show",
             $self->_get_param_defaults());

    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::FALSE unless(scalar(@output));

    return $SanUtil::TRUE;
} ## end sub snapshot_exist

###############################################################################
# Method Name: lun_list_info()
# Objective: Get lun information from this volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub lun_list_info {
    my ($self, %args) = @_;

    my @output;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_show",
             volume => $self->get_name(),
             $self->_get_param_defaults());

    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{path} = '/vol/' . delete $args{volume};
        if (exists($args{qtree}) and $args{qtree} ne '""') {
            $args{path} .= '/' . delete $args{qtree};
        }
        if (exists($args{lun}) and $args{lun} ne '""') {
            $args{path} .= '/' . delete $args{lun};
        } else {
            $args{path} .= '/*';
        }
    }


    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub lun_list_info

###############################################################################
# Method Name: qtree_list_info()
# Objective: Get qtree information from this volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub qtree_list_info {
    my ($self, %args) = @_;

    my @output;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_qtree_show",
             $self->_get_param_defaults());

    #  The ZAPI wants a path not a volume and qtree.
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{path} = '/vol/' . delete $args{volume};
        if (exists($args{qtree}) and $args{qtree} ne '""') {
            $args{path} .= '/' . delete $args{qtree};
        }
    }

    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub qtree_list_info

###############################################################################
# Method Name: snapshot_list_info()
# Objective:   Get snapshot information from this volume
# Details:     See POD documentation at the end of this file.
###############################################################################
sub snapshot_list_info {
    my ($self, %args) = @_;

    my @output;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_snapshot_show",
             $self->_get_param_defaults());

    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub snapshot_list_info

###############################################################################
# Method Name: efficiency_list_info()
# Objective:   Get efficiency information from this volume
# Details:     See POD documentation at the end of this file.
###############################################################################
sub efficiency_list_info {
    my ($self, %args) = @_;

    my @output;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_efficiency_show",
             $self->_get_param_defaults());

    # regardless which interface is specified, this method is only for cli
    # TODO: need to add to AttrTable for common output
    $args{cmd_interface} = "cli";

    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub efficiency_list_info

###############################################################################
# Method Name: efficiency_off()
# Objective:   Turn efficiency off for this volume
# Details:     See POD documentation at the end of this file.
###############################################################################
sub efficiency_off {
    my ($self, %args) = @_;

    my @output;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_efficiency_off",
             $self->_get_param_defaults());

    # regardless which interface is specified, this method is only for cli
    # TODO: sis-disable is already added to Translation Table. We need to change
    # SanMgr::Controller::setup to ensure it handles when error occurred correctly
    # then we can remove the below line
    $args{cmd_interface} = "cli";

    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub efficiency_off

###############################################################################
# Method Name: efficiency_on()
# Objective:   Turn efficiency on for this volume
# Details:     See POD documentation at the end of this file.
###############################################################################
sub efficiency_on {
    my ($self, %args) = @_;

    my @output;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_efficiency_on",
             $self->_get_param_defaults());

    # regardless which interface is specified, this method is only for cli
    # TODO: sis-enable is already added to Translation Table. We need to change
    # SanMgr::Controller::setup to ensure it handles when error occurred correctly
    # then we can remove the below line
    $args{cmd_interface} = "cli";

    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub efficiency_on

###############################################################################
# Method Name: efficiency_modify()
# Objective:   modify efficiency for this volume
# Details:     See POD documentation at the end of this file.
###############################################################################
sub efficiency_modify {
    my ($self, %args) = @_;

    my @output;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_efficiency_modify",
             $self->_get_param_defaults());

    # regardless which interface is specified, this method is only for cli
    # TODO: sis-set-config is already added to Translation Table. We need to change
    # SanMgr::Controller::setup to ensure it handles when error occurred correctly.
    # then we can remove the below line
    $args{cmd_interface} = "cli";

    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub efficiency_modify

###############################################################################
# Method Name: efficiency_start()
# Objective:   Start efficiency for this volume
# Details:     See POD documentation at the end of this file.
###############################################################################
sub efficiency_start {
    my ($self, %args) = @_;

    my @output;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd_interface => "cli",
             cmd => "volume_efficiency_start",
             $self->_get_param_defaults());

    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub efficiency_start

###############################################################################
# Method Name: efficiency_stop()
# Objective:   Stop efficiency for this volume
# Details:     See POD documentation at the end of this file.
###############################################################################
sub efficiency_stop {
    my ($self, %args) = @_;

    my @output;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd_interface => "cli",
             cmd => "volume_efficiency_stop",
             $self->_get_param_defaults());

    # Invoke CmdsDriver object and executes the actual command
    (@output) = $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub efficiency_stop

################################################################################
# Method Name: qtree_statistics_reset()
# Objective:   Resets Qtree Statistics.
# Details:     See POD documentation at the end of this file
################################################################################
sub qtree_statistics_reset {
    my ($self, %args) = @_;

    my (@output, $cmd);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # build cmd qtree show
    $cmd = "volume_qtree_statistics_reset";
    %args = (%args,
             cmd => $cmd,
             $self->_get_param_defaults());

    # Invoke CmdsDriver object and executes the actual command
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub qtree_statistics_reset

###############################################################################
# Method Name: modify
# Objective: modify volume attribute
# Details:   See POD documentation at the end of this file.
###############################################################################
sub modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_modify",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub modify

###############################################################################
# Method Name: mount
# Objective: mount volume junction
# Details:   See POD documentation at the end of this file.
###############################################################################
sub mount {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_mount",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub mount

###############################################################################
# Method Name: move
# Objective: move volume to new aggregate destination
# Details:   See POD documentation at the end of this file.
###############################################################################
sub move {
    my ($self, %args) = @_;

    $self->_clear_err();

    # initialize cmd parameters
    %args = (%args,
             cmd       => "volume_move_start",
             foreground => "true",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub move

###############################################################################
# Method Name: refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    my (@lun_objs, @tmp_objs) = ();

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_show",
             $self->_get_param_defaults());

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update volume object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # get lun objects
    @lun_objs = $self->get_lun_objs();
    if ($self->error()) {
        $self->_set_err("failed to get lun objects", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # store lun objects in this volume
    if (scalar(@lun_objs)) {
        $self->{lun_objs} = \@lun_objs;
    }

    return $SanUtil::TRUE;
} ## end sub refresh


###############################################################################
# Method Name: online
# Objective: online volume object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub online {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_online",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}


###############################################################################
# Method Name: offline
# Objective: offline volume object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub offline {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_offline",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: unmount
# Objective: unmount volume junction
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unmount {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_unmount",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub unmount

################################################################################
# Method Name: file_clone_create()
# Objective:   Create a file clone
# Details:     See POD documentation at the end of this file
################################################################################
sub file_clone_create {
    my ($self, %args) = @_;

    my $cdef;
    my $sync = 0;
    my %sync_args;

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify name param
    unless (defined $args{'source-path'} and length $args{'source-path'}) {
        $self->_set_err("mandatory parameter \"-source-path\" not specified ");
        return $SanUtil::FALSE;
    }

    $sync = delete $args{'sync'} if exists $args{'sync'};
    $sync_args{path} = "/vol/".$self->get_name()."".$args{'source-path'};
    $sync_args{timeout} = delete $args{timeout} if exists $args{timeout};
    $sync_args{interval} = delete $args{interval} if exists $args{interval};

    # For CLI interface the RANGE argument is expected to be a coma-separrated list
    # like this: -range 0:0:8, 24:24:8
    # But on ZAPI interface RANGE is supposed to be an array of hashes:
    # block-range => block-range [
    #                              {destination-block-number => 0,
    #                               source-block-number => 0,
    #                               number-of-blocks => 8,
    #                              },
    #                              {destination-block-number => 24,
    #                               source-block-number => 24,
    #                               number-of-blocks => 8,
    #                              }
    #                             ]
     if ($self->get_cmd_interface() =~ /zapi/i) {
        if ( defined($args{'range'}) ) {
            my @block_array = ();
            my @param_ranges = split(',', delete($args{'range'}) );
            #For each block, get the three elements
            foreach my $range_set (@param_ranges){
                my %block_range = ();
                my @block = split(':', $range_set);
                unless( scalar(@block) ){
                    $self->_set_err("parameter \"-range\" must have all three block " .
                                    "ranges (block-count:destination-block-number:source-block-number " .
                                    $self->errmsg()
                                    );
                    return $SanUtil::FALSE;
                }
                # Tag the values
                $block_range{'block-count'} = $block[0];
                $block_range{'destination-block-number'} = $block[1];
                $block_range{'source-block-number'} = $block[2];
                # Add them to the array
                push (@block_array, \%block_range)
            }
        # Get the array inside of the correct ZAPI attribute structure.
        $args{'block-ranges'}{'block-range'} = \@block_array;
        }
    }

    # build cmd qtree show
    %args = (%args,
             cmd => "volume_file_clone_create",
             $self->_get_param_defaults());

    # Invoke CmdsDriver object and executes the actual command
    $self->invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    if($sync) {
        $self->wait_for_clone_completion(%sync_args);
    }
    if ($self->error()){
        $self->_set_err("clone failed to complete." .
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub file_clone_create

###############################################################################
# Method Name: rename
# Objective: Rename this volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub rename {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify newname param
    unless (defined $args{newname} and length $args{newname}) {
        $self->_set_err("mandatory parameter \"-newname\" not specified ");
        return $SanUtil::FALSE;
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_rename",
             $self->_get_param_defaults());

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    } else {
        # rename was successful so we need to update the object with newname
        $self->set(volume => $args{newname});
    }

    return $SanUtil::TRUE;
} ## end sub rename

################################################################################
# A C C E S S O R   M E T H O D S
################################################################################

################################################################################
# Method Name: is_space_guarantee_enabled()
# Objective:   Is space guarantee enabled
# Details:     See POD documentation at the end of this file
################################################################################
sub is_space_guarantee_enabled {
    my ($self, %args) = @_;

    $args{field} = "space_guarantee_enabled";

    # the _get_field() will handle error
    if ($self->_get_field(%args) =~ /true/i) {
        return $SanUtil::TRUE;
    } else {
        return $SanUtil::FALSE;
    }
}

################################################################################
# Method Name: get_aggregate()
# Objective:   Get aggregate where this volume live
# Details:     See POD documentation at the end of this file
################################################################################
sub get_aggregate {
    my ($self, %args) = @_;

    $args{field} = "aggregate";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_autosize()
# Objective:   Get volume autosize
# Details:     See POD documentation at the end of this file
################################################################################
sub get_autosize {
    my ($self, %args) = @_;

    $args{field} = "autosize";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_autosize_increment()
# Objective:   Get volume autosize_increment
# Details:     See POD documentation at the end of this file
################################################################################
sub get_autosize_increment {
    my ($self, %args) = @_;

    $args{field} = "autosize_increment";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_available()
# Objective:   Get volume available
# Details:     See POD documentation at the end of this file
################################################################################
sub get_available {
    my ($self, %args) = @_;

    $args{field} = "available";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_dsid()
# Objective:   Get volume dsid
# Details:     See POD documentation at the end of this file
################################################################################
sub get_dsid {
    my ($self, %args) = @_;

    $args{field} = "dsid";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_group()
# Objective:   Get volume group
# Details:     See POD documentation at the end of this file
################################################################################
sub get_group {
    my ($self, %args) = @_;

    $args{field} = "group";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_junction_active()
# Objective:   Get volume junction_active
# Details:     See POD documentation at the end of this file
################################################################################
sub get_junction_active {
    my ($self, %args) = @_;

    $args{field} = "junction_active";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_junction_path()
# Objective:   Get volume junction_path
# Details:     See POD documentation at the end of this file
################################################################################
sub get_junction_path {
    my ($self, %args) = @_;

    $args{field} = "junction_path";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_junction_path_source()
# Objective:   Get volume junction_path_source
# Details:     See POD documentation at the end of this file
################################################################################
sub get_junction_path_source {
    my ($self, %args) = @_;

    $args{field} = "junction_path_source";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_max_autosize()
# Objective:   Get volume max_autosize
# Details:     See POD documentation at the end of this file
################################################################################
sub get_max_autosize {
    my ($self, %args) = @_;

    $args{field} = "max_autosize";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_msid()
# Objective:   Get volume msid
# Details:     See POD documentation at the end of this file
################################################################################
sub get_msid {
    my ($self, %args) = @_;

    $args{field} = "msid";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_name()
# Objective:   Get volume name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "volume";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get volume parent object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_percent_snapshot_space()
# Objective:   Get volume percent_snapshot_space
# Details:     See POD documentation at the end of this file
################################################################################
sub get_percent_snapshot_space {
    my ($self, %args) = @_;

    $args{field} = "percent_snapshot_space";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_percent_used()
# Objective:   Get volume percent_used
# Details:     See POD documentation at the end of this file
################################################################################
sub get_percent_used {
    my ($self, %args) = @_;

    $args{field} = "percent_used";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_policy()
# Objective:   Get volume policy
# Details:     See POD documentation at the end of this file
################################################################################
sub get_policy {
    my ($self, %args) = @_;

    $args{field} = "policy";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_security_style()
# Objective:   Get volume security_style
# Details:     See POD documentation at the end of this file
################################################################################
sub get_security_style {
    my ($self, %args) = @_;

    $args{field} = "security_style";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_size()
# Objective:   Get volume size
# Details:     See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self, %args) = @_;

    $args{field} = "size";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_snapshot_space_used()
# Objective:   Get volume snapshot_space_used
# Details:     See POD documentation at the end of this file
################################################################################
sub get_snapshot_space_used {
    my ($self, %args) = @_;

    $args{field} = "snapshot_space_used";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_snapshot_policy()
# Objective:   Get volume snapshot_policy
# Details:     See POD documentation at the end of this file
################################################################################
sub get_snapshot_policy {
    my ($self, %args) = @_;

    $args{field} = "snapshot_policy";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_space_guarantee()
# Objective:   Get volume space_guarantee
# Details:     See POD documentation at the end of this file
################################################################################
sub get_space_guarantee {
    my ($self, %args) = @_;

    $args{field} = "space_guarantee";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_state()
# Objective:   Get volume state
# Details:     See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self, %args) = @_;

    $args{field} = "state";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_stripe_count()
# Objective:   Get volume stripe_count
# Details:     See POD documentation at the end of this file
################################################################################
sub get_stripe_count {
    my ($self, %args) = @_;

    $args{field} = "stripe_count";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_stripe_width()
# Objective:   Get volume stripe_width
# Details:     See POD documentation at the end of this file
################################################################################
sub get_stripe_width {
    my ($self, %args) = @_;

    $args{field} = "stripe_width";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_stripe_epoch()
# Objective:   Get volume stripe_epoch
# Details:     See POD documentation at the end of this file
################################################################################
sub get_stripe_epoch {
    my ($self, %args) = @_;

    $args{field} = "stripe_epoch";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_total()
# Objective:   Get volume total
# Details:     See POD documentation at the end of this file
################################################################################
sub get_total {
    my ($self, %args) = @_;

    $args{field} = "total";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_type()
# Objective:   Get volume type
# Details:     See POD documentation at the end of this file
################################################################################
sub get_type {
    my ($self, %args) = @_;

    $args{field} = "type";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_unix_permissions()
# Objective:   Get volume unix_permissions
# Details:     See POD documentation at the end of this file
################################################################################
sub get_unix_permissions {
    my ($self, %args) = @_;

    $args{field} = "unix_permissions";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_used()
# Objective:   Get volume used
# Details:     See POD documentation at the end of this file
################################################################################
sub get_used {
    my ($self, %args) = @_;

    $args{field} = "used";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_user()
# Objective:   Get volume user
# Details:     See POD documentation at the end of this file
################################################################################
sub get_user {
    my ($self, %args) = @_;

    $args{field} = "user";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_uuid()
# Objective:   Get volume uuid
# Details:     See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self, %args) = @_;

    $args{field} = "uuid";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_volume_ownership()
# Objective:   Get volume volume_ownership
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volume_ownership {
    my ($self, %args) = @_;

    $args{field} = "volume_ownership";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_volume_style()
# Objective:   Get volume volume_style
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volume_style {
    my ($self, %args) = @_;

    $args{field} = "volume_style";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_vserver()
# Objective:   Get volume vserver
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vserver {
    my ($self, %args) = @_;

    $args{field} = "vserver";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_vsroot()
# Objective:   Get volume vsroot
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vsroot {
    my ($self, %args) = @_;

    $args{field} = "vsroot";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_node()
# Objective:   Get node name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_node {
    my ($self, %args) = @_;

    $args{field} = "node";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method Name: _get_param_defaults
# Objective: Gets default volume parameters
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_param_defaults {
    my ($self, %args) = @_;

    # always append vserver
    my %default_params = (%args, vserver => $self->{vserver});

    # if path is not provided, we need volume params
    unless (exists $args{'path'}) {
        %default_params = (%default_params, volume  => $self->{volume});
    }

    return %default_params;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Qtree
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache = $SanUtil::FALSE;

     # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    my $command
        = (defined $args{command}) ? $args{command} : "volume_show";

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # set use_cache flag if exists
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {

        # the get() will handle error
        return $self->get($args{field});
    } else {

        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(
            cmd  => $command,
            field => $args{field},
            $self->_get_param_defaults()
        );
    } ## end else [ if ($use_cache == $SanUtil::TRUE)
} ## end sub _get_field


################################################################################
# Method Name: _create_volume
# Objective: create volume
################################################################################
sub _create_volume {
    my ($self, %args) = @_;

    $self->_clear_err();

    # append cmd for volume create to argument list
    %args = (%args, cmd => "volume_create");

    # use common invoke from base class (San::CommonInterface)
    # to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # call volume refresh if enable cache is set
    if ($self->{enable_cache}) {
        $self->refresh(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to refresh " . $args{volume}, $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # set initialized flag
    $self->{initialized} = $SanUtil::TRUE;

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: _get_luns_from_qtree
# Objective: Gets lun objects from qtree objects
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_luns_from_qtree {
    my ($self, %args) = @_;

    my (@lun_objs, @qtree_objs);

    $self->_clear_err();

    # is qtree object exist?
    if (defined $self->{qtree_objs} and scalar(@{$self->{qtree_objs}})) {
        @qtree_objs = @{$self->{qtree_objs}};
    } else {
        # discover and get qtree objects
        @qtree_objs = $self->get_qtree_objs();
        if ($self->error()) {
            $self->_set_err("failed to get qtree objects", $self->errmsg());
            return (wantarray()) ? (): [];
        }
    }

    # return empty list if qtree does not exist
    unless (scalar(@qtree_objs)) {
        return (wantarray()) ? (): [];
    }

    # get luns from qtree objects
    foreach my $qtree_obj (@qtree_objs) {
        # make sure object is valid
        next unless (ref($qtree_obj) =~ /qtree/i);

        my @luns = $qtree_obj->get_lun_objs();
        if ($qtree_obj->error()) {
            $self->_set_err("failed to get lun objects", $qtree_obj->errmsg());
            return (wantarray()) ? (): [];
        }

        push @lun_objs, @luns;
    }

    return (wantarray()) ? (@lun_objs): [@lun_objs];
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Volume.pm

=head1 Author

Manoj Chintala (dl-ssan-automation)

=head1 Description

Provides an interface to manage a Volume

=head1 See Also

Vserver package used base class B<San::CommonInterface>  See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm, San::SanCache.pm>

=head1 Attributes

=head2 B<filer> (required)

string = controller name

=head2 B<vserver> (required)

string = vserver name to create volume

=head2 B<aggregate> (required)

string = aggregate name to be used for volume creation

=head2 B<size>

string = size of the volume

=head2 B<state>

string = volume's state.

=head2 B<policy>

string = ID number of the export policy associated with the volume

=head2 B<user>

string = name or ID of the user that is set as the owner of the
volume's root

=head2 B<group>

string = name or ID of the group that is set as the owner of the
volume's root

=head2 B<security-style>

string = security style for the volume

=head2 B<unix-permissions>

string = default UNIX permissions for files on the volume

=head2 B<junction-path>

string = volume's junction path

=head1 B<Public Methods>

=head2 B<new(%args)>

Create a volume object

=over 4

=item Inputs

=item * I<-filer>

Controller name

=item * I<-vserver> (default: 'vs0').

Name of the vserver to create volume

=item * I<-size> (default: 20MB)

Size of the volume. The value should be a number followed by a unit
designation. The unit can be specified in MB/m for mega bytes, GB/g for giga bytes,
TB/t for tera bytes. If the unit designation is not specified, bytes are used as
the unit, and the specified number is rounded up to the nearest 4 KB.
The minimum size for a volume is 20 MB (the default), and the maximum size is 16 TB.

=item * I<-state> (default: online)

Allowed values are {online | restricted | offline}
A restricted volume does not provide client access to data but is available for
administrative operations.

=item * I<-policy>

ID number of the export policy associated with the volume

=item * I<-user>

name or ID of the user that is set as the owner of the volume's root

=item * I<-group>

name or ID of the group that is set as the owner of the volume's root

=item * I<-security-style>

Possible values include unix (for UNIX mode bits), ntfs (for CIFS ACLs), and
mixed (for mixed NFS and CIFS permissions). Regardless of the security style,
both NFS and CIFS clients can read from and write to the volume

=item * I<-unix-permissions>

UNIX permissions can be specified either as a four-digit octal value (for example, 0700) or in the
style of the UNIX ls command (for example, -rwxr-x---)

=item * I<-junction-path>

volume's junction path

=item Outputs

Return volume object

=item Examples

 # ex. 1: create a volume object with size 100MB on vserver 'vs1'.
 use San::ControllerCluster::Filer10M::Vserver::Volume;

 $vol_obj =
      San::ControllerCluster::Filer10M::Vserver::Volume->new(-filer => $filer,
                                                             -vserver => "vs1"
                                                             -name => "testvol"
                                                             -size => "100m");
 if ($vol_obj->error()) {
    ...
 }

=back

=head2 B<copy()>

Copy the volume to a new destination

=over 4

=item Inputs

=item * <-destination-volume>

Name for the copy of the volume

=item * I<-destination-aggregate> (default: current aggregate)

Name of the aggregate to put the copy on.

=item Outputs

Returns a volume object representing the newly created volume

=item Examples

 # ex. 1: Copy a volume to $Aggr and name the volume "copy_vol".
 use San::ControllerCluster::Filer10M::Vserver::Volume;

 my $vol_copy_obj = $vol_obj->copy('destination-volume' => 'copy_vol',
                                    'destination-aggregate' => "$AGGR" );
 if ($vol_obj->error()) {
    ...
 }

=back

=head2 B<get_available>

Get available volume size in bytes.

=over 4

=item Inputs

None

=item Output

Returns the number of bytes available. The unit will not be suffixed at the end.
The output will just be a number

=item Examples

 # ex. 1: Get available size of the volume
 $available_size = $vol_obj->get_available();
 if ($vol_obj->error()) {
    ...
 }

=back

=head2 B<get_lun()>

Gets a lun object.

=over 4

=item Inputs

=item * name|lun: name of the lun.

=item Outputs

lun object.

=item Examples

  my $lun_obj = $vol_obj->get_lun(name => "lun_name");

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to get lun obj" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<get_qtree_obj()>

Gets a qtree object.

=over 4

=item Inputs

=item * name|qtree: name of the qtree.

=item Outputs

Qtree object.

=item Examples

  my $qtree_obj = $vol_obj->get_qtree_obj(name => "qtree_name");

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to get qtree obj" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<get_snapshot_obj()>

Gets a snapshot object.

=over 4

=item Inputs

=item * name|snapshot: name of the snapshot.

=item Outputs

snapshot object.

=item Examples

  my $snapshot_obj = $vol_obj->get_snapshot_obj(name => "snapshot_name");

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to get snapshot obj" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<create_snapshot_obj()>

Creates snapshot object.

=over 4

=item Inputs

=item * name|snapshot: name of the snapshot.

=item Outputs

Returns snapshot object.

=item Examples

  my $status = $vol_obj->create_snapshot_obj(name => "snapshot_name");

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to create snapshot object" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<qtree_list_info()>

Gets the qtree info on this volume.

=over 4

=item Inputs

none

=item Outputs

Array of output

=item Examples

  my @ouput = $vol_obj->qtree_list_info();

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to get qtree list info" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<snapshot_list_info()>

Gets the snapshot info on this volume.

=over 4

=item Inputs

none

=item Outputs

Array of output

=item Examples

  my @ouput = $vol_obj->snapshot_list_info();

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to get snapshot list info" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<efficiency_list_info()>

Gets the efficiency info on this volume.

=over 4

=item Inputs

none

=item Outputs

Array of output

=item Examples

  my @ouput = $vol_obj->efficiency_list_info();

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to get snapshot list info" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<efficiency_on()>

Gets the efficiency info on this volume.

=over 4

=item Inputs

none

=item Outputs

Array of output

=item Examples

  my @ouput = $vol_obj->efficiency_on();

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to get snapshot list info" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<efficiency_off()>

Gets the efficiency info on this volume.

=over 4

=item Inputs

none

=item Outputs

Array of output

=item Examples

  my @ouput = $vol_obj->efficiency_off();

  if ($vol_obj->error()) {
        ...
  }

=back

=head2 B<efficiency_modify()>

Gets the efficiency info on this volume.

=over 4

=item Inputs

none

=item Outputs

Array of output

=item Examples

  my @ouput = $vol_obj->efficiency_modify('-compression' => 'true');

  if ($vol_obj->error()) {
  ...
  }

=back

=head2 B<efficiency_start()>

Starts storage efficiency on this volume.

=over 4

=item Inputs

none

=item Outputs

True if efficiency started successfully, False otherwise

=item Examples

  $vol_obj->efficiency_start('scan-old-data' => 'true');
  if ($vol_obj->error()) {
  ...
  }

=back

=head2 B<efficiency_stop()>

Stop storage efficiency on this volume.

=over 4

=item Inputs

none

=item Outputs

True if efficiency stopped successfully, False otherwise

=item Examples

  $vol_obj->efficiency_stop('force => 'true');
  if ($vol_obj->error()) {
  ...
  }

=back

=head2 B<qtree_exist()>

Determines if qtree exists.

=over 4

=item Inputs

=item * name|qtree: name of the qtree.

=item Outputs

Returns True if qtree exists, else returns false.

=item Examples

  my $status = $vol_obj->qtree_exist(name => "qtree_name");

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to check if qtree exists" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<snapshot_exist()>

Determines if snapshot exists.

=over 4

=item Inputs

=item * name|snapshot: name of the snapshot.

=item Outputs

Returns True if snapshot exists, else returns false.

=item Examples

  my $status = $vol_obj->snapshot_exist(name => "snapshot_name");

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to check if snapshot exists" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<destroy_qtree_obj()>

Deletes qtree obj.

=over 4

=item Inputs

=item * name|qtree: name of the qtree.

=item Outputs

Returns True if qtree deleted.

=item Examples

  $vol_obj->destroy_qtree_obj(name => "qtree_name");

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to check if qtree exists" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<destroy_snapshot_obj()>

Deletes snapshot obj.

=over 4

=item Inputs

=item * name|snapshot: name of the snapshot.

=item Outputs

Returns True if snapshot deleted.

=item Examples

  $vol_obj->destroy_snapshot_obj(name => "snapshot_name");

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to check if snapshot exists" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<qtree_statistics_reset()>

Resets qtree statistics on a volume.

=over 4

=item Inputs

None

=item Outputs

Returns True on success.

=item Examples

  $vol_obj->qtree_statistics_reset();

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to reset qtree statistics" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<file_clone_create()>

Create a file clone on the volume.

=over 4

=item Inputs

=item * source-path: Filepath to be cloned.

=item * destination-path (optional): Destination path of the clone.

=item Outputs

Returns True on success.

=item Examples

  $vol_obj->file_clone_create();

  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to clone the file" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<get_lun_maxsize(%args)>

Get lun maxsize for this volume

=over 4

=item Inputs

=item * I<ostype> => <windows|linux|solaris|hyper_v|xen|aix|etc.> (required)

The operating system type

=item Outputs

Returns a hash of maxsize with/out snapshot. Set error if command fails.

=item Examples

  # ex. 1: get lun maxsize for windows ostype
  my %win_maxsize = $vol_obj->get_lun_maxsize(-ostype => "windows");
  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to get lun maxsize for windows" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<get_lun_minsize(%args)>

Get lun maxsize for this volume

=over 4

=item Inputs

=item * I<ostype> => <windows|linux|solaris|hyper_v|xen|aix|etc.> (required)

The operating system type

=item Outputs

Returns a string number of lun minsize. Sets error if command fails and returns empty string.

=item Examples

  # ex. 1: get lun minsize for windows ostype
  my $lun_minsize = $vol_obj->get_lun_minsize(-ostype => "linux");
  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to get lun minsize" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<rename(%args)>

Rename a volume

=over 4

=item Inputs

=item * I<newname> => "new_name" (required)

The new name for the volume.

=item Outputs

Returns TRUE if the volume was renamed successfully, FALSE otherwise.

=item Examples

  # ex. 1: Rename a volume
  $vol_obj->rename(-newname => "foo_vol");
  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("Failed to rename the volume" . $errmsg);
        return $TCD::FAIL;
  }

=back

=head2 B<get_node()>

Gets the node name where the volume was created.

=over 4

=item Inputs

=item * None.

=item Outputs

=item * A string containing the node's name.

=item Examples

  my $node_name = $vol_obj->get_node();
  if ($vol_obj->error()) {
        my $errmsg = join "\n", $vol_obj->errmsg();
        $Test->nlog("ERROR: Failed to get node name" . $errmsg);
        return $TCD::FAIL;
  }

=back
