# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to a Lun object
##
## @author korry@netapp.com, dl-ssan-test
## @status Shared

package San::ControllerCluster::Filer10M::Vserver::Volume::Lun;

# compiler directives
use strict;
use Data::Dumper;

# inherit base class
use base qw( San::CommonInterface
             San::ControllerCluster::Filer10M::Vserver::Volume::LunCompat );

###############################################################################
# Method Name: new
# Objective: Create new Lun object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    if (exists $args{name}) {
        $args{lun} = delete $args{name};
    }

    # verify mandatory cmd argument
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }
    unless (defined $args{parent_obj} and length $args{parent_obj}) {
        $self->_set_err("mandatory parameter \"-parent_obj\" not specified ");
        return $self;
    }

    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # since path and lun name could be used to create lun, we need to
    # get its name for the SUPER::_new()
    if  (!(exists $args{'lun'}) or ($args{'lun'} eq "zapi_na") ) {
        # make sure lun path exists
        if (exists $args{'path'}) {
            # strip path
            $args{lun} = File::Basename::basename($args{'path'});
        } else {
            $self->_set_err("either \"-lun or -path\" must be specified ");
            return $self;
        }
    }

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct lun object
    $self = $class->SUPER::_new(-name  => $args{lun},
                                -filer => $args{filer},
                                %args);
    if ($self->error()) {
        $self->_set_err("failed to contruct lun object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # to be compatible w/ 7-M, we set these default params
    $args{igroup} = "";
    $args{lunid} = "";

    # Add qtree so it gets populated into the fields
    $args{qtree} = "" unless exists $args{qtree};

    # store Lun elements on a list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # we can now remove params from the args list that we dont need
    delete $args{parent_obj};
    delete $args{debug_level};
    delete $args{igroup};
    delete $args{lunid};
    delete $args{qtree} if ($args{qtree} eq "");

    # for zapi, we need to convert to path
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{'path'} = $self->_get_lun_path(%args);
        delete $args{volume};
        delete $args{qtree};
        delete $args{lun};
    } else {
        # delete lun path if initialized flag is set
        delete $args{path} unless $self->{initialized};
    }

    # if the obj hs been initialized, just refresh to update info
    # otherwise initialize it by create and update its object
    unless ($args{initialized}) {
        # lun has not been initialized/created, so we need to create it
        $self->_create_lun(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to _create_lun() ", $self->errmsg());
            return $self;
        }
    }

    return $self;
} ## end sub new

###############################################################################
# Method Name: delete
# Objective: delete lun object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_delete");

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # vserver object must re-discover or refresh to update lun objects
    # set name to "_DELETE_" to indicate lun has been deleted
    $self->{name} = "_DELETE_";
    $self->{lun} = "_DELETE_";

    return $SanUtil::TRUE;
} ## end sub delete

###############################################################################
# Method Name: destroy
# Objective: destroy lun object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self, %args) = @_;

    return $self->delete(%args);
}

################################################################################
# Method Name: is_alua_enabled()
# Objective:   Returns true if mapped lun is alua enabled, otherwise false
# Details:     See POD documentation at the end of this file
################################################################################
sub is_alua_enabled {
    my ($self) = @_;

    $self->_clear_err();

    # call get_alua() to determine whether lun is alua enabled or not
    return ($self->get_alua() =~ /true/i) ? $SanUtil::TRUE : $SanUtil::FALSE;
}

################################################################################
# Method Name: is_readonly()
# Objective:   Returns true if the lun is read-only, otherwise false
# Details:     See POD documentation at the end of this file
################################################################################
sub is_readonly {
    my ($self) = @_;

    $self->_clear_err();

    # call get_readonly() to determine whether the lun is read-only or not
    return ($self->get_readonly() =~ /true/i) ? $SanUtil::TRUE : $SanUtil::FALSE;
}

###############################################################################
# Method Name: is_offline
# Objective: check to see if lun state is offline
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_offline {
    my ($self) = @_;

    $self->_clear_err();

    if ($self->is_online()) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: is_online()
# Objective:   Returns true if lun state is online otherwise false
# Details:     See POD documentation at the end of this file
################################################################################
sub is_online {
    my ($self, %args) = @_;

    $self->_clear_err();

    # call get_isonline() to determine whether lun is online or not
    return $self->get_isonline(%args);
}

###############################################################################
# Method Name: modify
# Objective: modify lun attribute
# Details:   See POD documentation at the end of this file.
###############################################################################
sub modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_modify");

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub modify

###############################################################################
# Method Name: lun_get_select_attribute
# Objective: get lun select attribute
# Details:   See POD documentation at the end of this file.
###############################################################################

sub lun_get_select_attribute {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $path = $self->get_path();
    if ($self->error()) {
        $self->_set_err("Could not get lun path", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_select");

    # use common invoke to translate cdef into UI cmd and execute it
    my @output = $self->SUPER::invoke_cmd_obj(%args);

    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? () : [];
    }

    unless (scalar(@output)) {
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub lun_get_select_attribute

###############################################################################
# Method Name: map
# Objective: map lun to an igroup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub map {
    my ($self, %args) = @_;
    my (%tmp_args);

    $self->_clear_err();

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|igroup|lun-id|lun_id|force)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # this is a scalar
            $tmp_args{igroup} = $key;
        }
        last;
    }

    # set tmp args back to args
    %args = %tmp_args;

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to igroup to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{name}) {
        $args{igroup} = $args{name};
        delete $args{name};
    }
    unless (defined $args{igroup} and length $args{igroup}) {
        $self->_set_err("mandatory parameter \"-igroup\" not specified ");
        return $SanUtil::FALSE;
    }
    $self->{igroup} = $args{igroup};

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_map");

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub map

###############################################################################
# Method Name: unmap
# Objective: unmap lun from an igroup
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unmap {
    my ($self, %args) = @_;
    my (%tmp_args);

    $self->_clear_err();

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|igroup|lun-id)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # this is a scalar
            $tmp_args{igroup} = $key;
        }
        last;
    }

    # set tmp args back to args
    %args = %tmp_args;

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to igroup to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{name}) {
        $args{igroup} = $args{name};
        delete $args{name};
    }
    unless (defined $args{igroup} and length $args{igroup}) {
        $self->_set_err("mandatory parameter \"-igroup\" not specified ");
        return $SanUtil::FALSE;
    }
    $self->{igroup} = $args{igroup};

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_unmap");

    # use common invoke to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub unmap

###############################################################################
#Method Name: getbind
#Objective: get bindings of a vvol type lun. No support needed
#           for 7 mode controllers.
#Details: See POD documentation at the end of this file.
#         This may be used in the future for now it is not needed.
#         Leaving it here in case someone else wants to use it.
###############################################################################
sub get_bind {

    my ($self, %args) = @_;

    $self->_clear_err();
    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_bind_show");

    my $vvol = delete $args{lun};
    my $volume = delete $args{volume};
    my $qtree = delete $args{qtree};

    #build up the vvol path
    my $vvol_path;
    if (defined($qtree)){
        $vvol_path = "/vol/$volume/$qtree/$vvol";
    } else {
        $vvol_path = "/vol/$volume/$vvol";
    }

    $args{'vvol-path'} = $vvol_path;
    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke_cmd_obj to obtain lun object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return {};
    }

    #build up an array of exsiting pep bindings
    my @peparray = ();
    foreach my $elem(@output){
        push (@peparray,$elem->{protocol_endpoint_path});
    }
    unless (scalar(@output)) {
        return {};
    }
    return @peparray;

} ## end sub get_bind

###############################################################################
# Method Name: offline
# Objective: offline a lun
# Details:   See POD documentation at the end of this file.
###############################################################################
sub offline {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $path = $self->get_path();
    if ($self->error()) {
        $self->_set_err("Could not get lun path", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Initialize cmd parameters
    %args = (%args,
             cmd        => "lun_offline",
             path       => $path);
    # Get default params based on initialized Parameters
    %args = ($self->_get_param_defaults(%args) );


    # The lun and volume arguments can't be used with the path argument
    delete $args{name} if ($args{name});
    delete $args{lun} if ($args{lun});
    delete $args{qtree} if ($args{qtree});
    delete $args{volume} if ($args{volume});

    $self->sanlog(-msg => "Lun::offline() - offline lun $path",
                  -msg_level => 2);

    # Get a CmdsDriver object and executes the actual command
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Update $self after the offline
    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh lun object:", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ##end of offline

###############################################################################
# Method Name: online
# Objective: online a lun
# Details:   See POD documentation at the end of this file.
###############################################################################
sub online {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $path = $self->get_path();
    if ($self->error()) {
        $self->_set_err("Could not get lun path", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Initialize cmd parameters
    %args = (%args,
             cmd        => "lun_online",
             path       => $path);
    # Get default params based on initialized Parameters
    %args = ($self->_get_param_defaults(%args) );

    # The lun and volume arguments can't be used with the path argument
    delete $args{name} if ($args{name});
    delete $args{lun} if ($args{lun});
    delete $args{qtree} if ($args{qtree});
    delete $args{volume} if ($args{volume});

    $self->sanlog(-msg => "Lun::online() - online a lun",
                  -msg_level => 2);

    # Get a CmdsDriver object and executes the actual command
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Update $self after the online
    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh lun object:", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ##end of online


###############################################################################
# Method Name: refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # we don't need the size parameter for lun show, so remove it
    if (exists $args{size}) {
        delete $args{size};
    }

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_show");

    $self->sanlog(-msg => $self->dumper_get(\%args),
                  -msg_level => 3);

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update lun object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return $SanUtil::FALSE;
    }

    # for zapi we need to split new path to polulate lun, qtree, and volume
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $self->_split_path_into_names();
    }
    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: is_mapped
# Objective: Determines if a lun is mapped to an igroup.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub is_mapped {
    my ($self, %args) = @_;
    my (%tmp_args);

    $self->_clear_err();

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(igroup)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # this is a scalar
            $tmp_args{igroup} = $key;
        }
        last;
    }

    # set tmp args back to args
    %args = %tmp_args;

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_mapped_show");

    # Invoke the command
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar(@output)) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: get_list_info()
# Objective: Get lun information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_show");
    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke_cmd_obj to obtain lun object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return {};
    }

    unless (scalar(@output)) {
        return {};
    }

    return $output[0];
} ## end sub get_list_info

###############################################################################
# Method Name: map_list_info()
# Objective: Get a list of mapped lun info
# Details:   See POD documentation at the end of this file.
###############################################################################
sub map_list_info {
    my ($self, %args) = @_;

    my @igroups = ();

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_mapped_show");

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke_cmd_obj to obtain lun object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? () : [];
    }

    unless (scalar(@output)) {
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub map_list_info

###############################################################################
# Method Name: get_maps()
# Objective: A wrapper to map_list_info to get a list of igroups that
#            mapped to this lun
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_maps {
    my ($self, %args) = @_;

    my @igroups = ();

    $self->_clear_err();

    # get the map list info from this lun
    my @records = $self->map_list_info();
    if ($self->error() or scalar(@records) == 0) {
        # error is already set, bubble up
        return (wantarray()) ? () : [];
    }

    # get the igroup name that mapped to this lun
    foreach my $record (@records) {
        if (exists($record->{igroup})) {
            push (@igroups, $record->{igroup});
        }
    }

    return (wantarray()) ? @igroups : \@igroups;
} ## end sub get_maps

###############################################################################
# Method Name: move()
# Objective: Move a lun (basicaly this is a rename)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub move {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $path = $self->get_path();
    if ($self->error()) {
        $self->_set_err("Could not get lun path", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Check for the required path argument
    unless (defined $args{path} and length $args{path}) {
        $self->_set_err("Mandatory parameter \"path\" not specified ");
        return $SanUtil::FALSE;
    }
    my $new_path = $args{path};

    # Initialize cmd parameters
    %args = (%args,
             cmd        => "lun_move",
             path       => $path,
             "new-path" => "$new_path");
    # Get default params based on initialized Parameters
    %args = ($self->_get_param_defaults(%args) );

    # The lun and volume arguments can't be used with the path argument
    delete $args{name} if ($args{name});
    delete $args{lun} if ($args{lun});
    delete $args{qtree} if ($args{qtree});
    delete $args{volume} if ($args{volume});

    $self->sanlog(-msg => "Lun::move() - move a lun from $path to $new_path",
                  -msg_level => 1);

    # Get a CmdsDriver object and executes the actual command
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Update our changed attributes after the move and refresh
    my $new_name = File::Basename::basename($new_path);
    $self->{lun}  = $new_name;
    $self->{name} = $new_name;
    $self->{path} = $new_path;
    $self->_split_path_into_names();

    # refresh all attribute if flag enable_cache is set
    if ($self->{enable_cache}) {
        $self->refresh(path => $new_path);
        if ($self->error()) {
            $self->_set_err("Failed to refresh lun object:", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} ## end sub move()

###############################################################################
# Method Name: resize()
# Objective: Resize a lun
# Details:   See POD documentation at the end of this file.
###############################################################################
sub resize {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $path = $self->get_path();
    if ($self->error()) {
        $self->_set_err("Could not get lun path", $self->errmsg());
        return $SanUtil::FALSE;
    }


    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Check for the required size argument
    unless (defined $args{size} and length $args{size}) {
        $self->_set_err("Mandatory parameter \"size\" not specified ");
        return $SanUtil::FALSE;
    }
    my $new_size = $args{size};

    # Initialize cmd parameters
    %args = (%args,
             cmd        => "lun_resize",
             path       => $path,
             "size"     => "$new_size");
    # Get default params based on initialized Parameters
    %args = ($self->_get_param_defaults(%args) );

    # The lun and volume arguments can't be used with the path argument
    delete $args{name} if ($args{name});
    delete $args{lun} if ($args{lun});
    delete $args{qtree} if ($args{qtree});
    delete $args{volume} if ($args{volume});

    $self->sanlog(-msg => "Lun::resize() - resize a lun to the size $new_size",
                  -msg_level => 2);

    # Get a CmdsDriver object and executes the actual command
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Update $self after the resize
    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh lun object:", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub resize()


###############################################################################
# Method Name: set_comment()
# Objective: set the comment of a lun
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_comment {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $path = $self->get_path();
    if ($self->error()) {
        $self->_set_err("Could not get lun path", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Check for the required comment argument
    unless (defined $args{comment} and length $args{comment}) {
        $self->_set_err("Mandatory parameter \"comment\" not specified ");
        return $SanUtil::FALSE;
    }

    my $new_comment = $args{comment};

    # Initialize cmd parameters
    #
    # For zapi we need a different command to be able to translate it.
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{'cmd'} = "lun_set_comment";
    } else {
        $args{'cmd'} = "lun_comment",
    }

    %args = (%args,
             'path'   => $path,
             'comment'=> $new_comment);

        # Get default params based on initialized Parameters
    %args = ($self->_get_param_defaults(%args) );

    # The lun and volume arguments can't be used with the path argument
    delete $args{name} if ($args{name});
    delete $args{lun} if ($args{lun});
    delete $args{qtree} if ($args{qtree});
    delete $args{volume} if ($args{volume});

    $self->sanlog(-msg => "Lun::set_comment() - change comment in a lun to the $new_comment",
                  -msg_level => 2);

    # Get a CmdsDriver object and executes the actual command
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Update $self after the set comment
    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh lun object:", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub set_comment()


###############################################################################
# Method Name: set_serial()
# Objective: set_serial a lun
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_serial {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $path = $self->get_path();
    if ($self->error()) {
        $self->_set_err("Could not get lun path", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Check for the required serial argument
    unless (defined $args{serial} and length $args{serial}) {
        $self->_set_err("Mandatory parameter \"serial\" not specified ");
        return $SanUtil::FALSE;
    }

    my $new_serial_number = $args{serial};

    # Initialize cmd parameters
    #
    # For zapi we need a different command to be able to translate it.
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{'cmd'} = "lun_set_serial";
    } else {
        $args{'cmd'} = "lun_serial",
    }

    %args = (%args,
             'path'  => $path,
             'serial'=> $new_serial_number);
    # Get default params based on initialized Parameters
    %args = ($self->_get_param_defaults(%args) );

    # The lun and volume arguments can't be used with the path argument
    delete $args{name} if ($args{name});
    delete $args{lun} if ($args{lun});
    delete $args{qtree} if ($args{qtree});
    delete $args{volume} if ($args{volume});

    $self->sanlog(-msg => "Lun::set_serial() - set the serial of a lun to $new_serial_number",
                  -msg_level => 2);

    # Get a CmdsDriver object and executes the actual command
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Update $self after the setting the serial
    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh lun object:", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub set_serial()


###############################################################################
# Method Name: set_device_id
# Objective: set_device_id a lun
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_device_id {

    my ($self, %args) = @_;
    my $device_id;
    my $type_flag;

    $self->_clear_err();

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Check for the required device-id argument
    unless (defined $args{'device-id'} and length $args{'device-id'}) {
        $self->_set_err("Mandatory parameter \"device-id\" not specified ");
        return $SanUtil::FALSE;
    }
    $device_id = $args{'device-id'};

    # Check for the required type argument
    unless (defined $args{'type'} and length $args{'type'}) {
        $self->_set_err("Mandatory parameter \"type\" not specified ");
        return $SanUtil::FALSE;
    }

    if ($args{'type'} eq "binary") {
        $type_flag = "-device-binary-id";
    } elsif ($args{'type'} eq "text") {
        $type_flag = "-device-text-id";
    } elsif ($args{'type'} eq "legacy") {
        $type_flag = "-device-legacy-id";
    } else {
        $self->_set_err("Invalid \"type\" specified: $args{'type'} ");
        return $SanUtil::FALSE;
    }

    if ($self->get_cmd_interface() =~ /zapi/i) {
        # initialize cmd parameters
        %args = ($self->_get_param_defaults(%args), cmd => "lun_set_device_id");

        # use common invoke to translate cdef into UI cmd and execute it
        $self->SUPER::invoke_cmd_obj(%args);
    } else {
        $self->modify($type_flag => $device_id);
    }
    if ($self->error()) {
        $self->_set_err("Could not modify device id", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;

} ## end of set_device_id

###############################################################################
# Method Name: unset_device_id
# Objective: unset device id of a lun
# Details:   See POD documentation at the end of this file.
###############################################################################
sub unset_device_id {

    my ($self, %args) = @_;
    my $type_flag;

    $self->_clear_err();

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    if ($self->get_cmd_interface() =~ /zapi/i) {
        delete $args{'type'} if (exists $args{'type'});

        # initialize cmd parameters
        %args = ($self->_get_param_defaults(%args), cmd => "lun_unset_device_id");

        # use common invoke to translate cdef into UI cmd and execute it
        $self->SUPER::invoke_cmd_obj(%args);
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj " .
                            $self->dumper_get(\%args, $self->get_debug_level()),
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    } else {
        # Check for the required type argument
        unless (defined $args{'type'} and length $args{'type'}) {
            $self->_set_err("Mandatory parameter \"type\" not specified ");
            return $SanUtil::FALSE;
        }

        if ($args{'type'} eq "binary") {
            $type_flag = "-clear-binary-id";
        } elsif ($args{'type'} eq "text") {
            $type_flag = "-clear-text-id";
        } else {
            $self->_set_err("Invalid \"type\" specified: $args{'type'} ");
        }

        $self->modify($type_flag => 'true');
        if ($self->error()) {
            $self->_set_err("Could not unset device id", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;

} ##end of unset_device_id

###############################################################################
# Method Name: set_space_allocation
# Objective: set_space_allocation for a lun
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_space_allocation {

    my ($self, %args) = @_;
    my $space_allocation;

    $self->_clear_err();

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Check for the required space-allocation argument
    unless (defined $args{'space-allocation'} and length $args{'space-allocation'}) {
        $self->_set_err("Mandatory parameter \"space_allocation\" not specified ");
        return $SanUtil::FALSE;
    }
    $space_allocation = $args{'space-allocation'};

    $self->modify('-space-allocation' => $space_allocation);
    if ($self->error()) {
        $self->_set_err("Could not modify space allocation", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;

} ## end of set_space_allocation


###############################################################################
# Method Name: set_space_resrv
# Objective: Enables or disables the space reservation on the LUN
# Details:   See POD documentation at the end of this file.
###############################################################################
sub set_space_resrv {

    my ($self, %args) = @_;

    $self->_clear_err();

    my $path = $self->get_path();
    if ($self->error()) {
        $self->_set_err("Could not get lun path", $self->errmsg());
        return $SanUtil::FALSE;
    }
    my $space_reserv;

    # Remove the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Check for the required space-reserve argument
    unless (defined $args{'space-reserve'} and length $args{'space-reserve'}) {
        $self->_set_err("Mandatory parameter \"space-reserve\" not specified ");
        return $SanUtil::FALSE;
    }

    # For zapi we need to translate the value of the attribute to boolean.
    if ($self->get_cmd_interface() =~ /zapi/i) {
        if ( lc($args{'space-reserve'}) eq "enabled") {
            $args{'space-reserve'} = "true";
        } else {
            $args{'space-reserve'} = "false";
        }
    }

    # Initialize cmd parameters
    %args = (%args,
             cmd        => 'lun_set_reservation',
             path       => $path);
    # Get default params based on initialized Parameters
    %args = ($self->_get_param_defaults(%args) );

    # The lun and volume arguments can't be used with the path argument
    delete $args{name} if ($args{name});
    delete $args{lun} if ($args{lun});
    delete $args{qtree} if ($args{qtree});
    delete $args{volume} if ($args{volume});

    $self->sanlog(-msg => "Lun::set_space_resrv() - set_space_resrv a lun",
                  -msg_level => 2);

    # Get a CmdsDriver object and executes the actual command
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Update $self after the move
    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh lun object:", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;

} ##end of set_space_resrv

###############################################################################
# Method Name: get_lun_geometry()
# Objective: Get lun geometry information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lun_geometry {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_geometry");

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke_cmd_obj to obtain lun object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return {};
    }

    unless (scalar(@output)) {
        return {};
    }

    return $output[0];
} ## end sub get_lun_geometry

###############################################################################
# Method Name: get_lun_stats()
# Objective: Get lun stats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lun_stats {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "lun_stats");

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke_cmd_obj to obtain lun object data
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return {};
    }

    unless (scalar(@output)) {
        return {};
    }

    return $output[0];
} ## end sub get_lun_stats

################################################################################
# A C C E S S O R   M E T H O D S
################################################################################

################################################################################
# Method Name: get_alua()
# Objective:   Get alua that mapped to this lun
# Details:     See POD documentation at the end of this file
################################################################################
sub get_alua {
    my ($self, %args) = @_;

    $args{field} = "alua";
    $args{-command} = "lun_mapped_show";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_igroup()
# Objective:   Get igroup that mapped to this lun
# Details:     See POD documentation at the end of this file
################################################################################
sub get_igroup {
    my ($self, %args) = @_;

    $args{field} = "igroup";
    $args{-command} = "lun_mapped_show";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_igroup_type()
# Objective:   Get igroup type/protocol that mapped to this lun
# Details:     See POD documentation at the end of this file
################################################################################
sub get_igroup_type {
    my ($self, %args) = @_;

    $args{field} = "protocol";
    $args{-command} = "lun_mapped_show";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_initiators()
# Objective:   Get initiators that mapped to this lun
# Details:     See POD documentation at the end of this file
################################################################################
sub get_initiators {
    my ($self, %args) = @_;

    $args{field} = "initiators";
    $args{-command} = "lun_mapped_show";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_isonline()
# Objective:   Get is-online field
# Details:     See POD documentation at the end of this file
################################################################################
sub get_isonline {
    my ($self, %args) = @_;

    $args{field} = "state";

    # the _get_field() will handle error
    my $state = $self->_get_field(%args);

    if ($state =~ /online|true/i) {
        return $SanUtil::TRUE;
    } else {
        return $SanUtil::FALSE;
    }
}

################################################################################
# Method Name: get_readonly()
# Objective:   Get read-only field
# Details:     See POD documentation at the end of this file
################################################################################
sub get_readonly {
    my ($self, %args) = @_;

    $args{field} = "read_only";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_lun_id()
# Objective:   Get LU number ID
# Details:     See POD documentation at the end of this file
################################################################################
sub get_lun_id {
    my ($self, %args) = @_;

    $args{field} = "lun_id";
    $args{-command} = "lun_mapped_show";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_luno()
# Objective:   Get LU number
# Details:     See POD documentation at the end of this file
################################################################################
sub get_luno {
    my ($self, %args) = @_;

    # wrapper call to get_lun_id
    return $self->get_lun_id(%args);
}

################################################################################
# Method Name: get_path()
# Objective:   Get lun_path
# Details:     See POD documentation at the end of this file
################################################################################
sub get_path {
    my ($self, %args) = @_;

    $args{field} = "path";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_portset()
# Objective:   Get portset that mapped to this lun
# Details:     See POD documentation at the end of this file
################################################################################
sub get_portset {
    my ($self, %args) = @_;

    $args{field} = "portset";
    $args{-command} = "lun_mapped_show";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_name()
# Objective:   Get lun name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "lun";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_class()
# Objective:   Get lun class
# Details:     See POD documentation at the end of this file
################################################################################
sub get_class {
    my ($self, %args) = @_;

    $args{field} = "class";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get lun parent object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_ostype()
# Objective:   Get lun ostype
# Details:     See POD documentation at the end of this file
################################################################################
sub get_ostype {
    my ($self, %args) = @_;

    $args{field} = "ostype";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_text_device_id()
# Objective:   Get text device id
# Details:     See POD documentation at the end of this file
################################################################################
sub get_text_device_id {
    my ($self, %args) = @_;

    $args{field} = "device_text_id";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_binary_device_id()
# Objective:   Get binary device id
# Details:     See POD documentation at the end of this file
################################################################################
sub get_binary_device_id {
    my ($self, %args) = @_;

    $args{field} = "device_binary_id";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_legacy_device_id()
# Objective:   Get legacy device id
# Details:     See POD documentation at the end of this file
################################################################################
sub get_legacy_device_id {
    my ($self, %args) = @_;

    $args{field} = "device_legacy_id";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_serial()
# Objective:   Get lun serial
# Details:     See POD documentation at the end of this file
################################################################################
sub get_serial {
    my ($self, %args) = @_;

    $args{field} = ($self->get_cmd_interface() =~ /cli/) ? "serial" : "serial-number";
    $args{-command} = "lun_serial";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_size()
# Objective:   Get lun size
# Details:     See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self, %args) = @_;

    $args{field} = "size";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_space_resv()
# Objective:   Get lun space_resv
# Details:     See POD documentation at the end of this file
################################################################################
sub get_space_resv {
    my ($self, %args) = @_;

    # For CLI use the usual method get_field
    if ($self->get_cmd_interface() =~ /cli/) {
        $args{'field'} = "space_reserve";
        $args{'command'} = "lun_set_reservation";

        # the _get_field() will handle error
        return $self->_get_field(%args);

    } else { # cmd_interface = zapi
        # For ZAPI, need to directly execute the command and transform the result
        # Initialize the command
        $args{'cmd'} = 'lun_get_reservation';

        # Get default params based; the required parameter PATH will be added here.
        %args = ($self->_get_param_defaults(%args) );

        # Executes the actual command
        my @output = $self->SUPER::invoke_cmd_obj(%args);
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj " .
                            $self->dumper_get(\%args, $self->get_debug_level()),
                            $self->errmsg());
            return undef;
        }
        # Now translate the ZAPI output to CLI readable
        if (exists $output[0]->{'is-enabled'}) {
            if ( lc($output[0]->{'is-enabled'}) eq "true") {
                return "enabled";
            } else {
                return "disabled";
            }
        } else {
            $self->_set_err("output does not contain field 'is-enabled'",
                            $self->errmsg());
            return undef;
        }
    }
}

################################################################################
# Method Name: get_space_allocation()
# Objective:   Get lun space_allocation
# Details:     See POD documentation at the end of this file
################################################################################
sub get_space_allocation {
    my ($self, %args) = @_;

    $args{field} = "space_allocation";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_suffix_size()
# Objective:   Get prefix size
# Details:     See POD documentation at the end of this file
################################################################################
sub get_suffix_size {
    my ($self, %args) = @_;

    $args{field} = "suffix_size";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_comment()
# Objective:   Get lun comment
# Details:     See POD documentation at the end of this file
################################################################################
sub get_comment {
    my ($self, %args) = @_;

    $args{field} = "comment";
    $args{-command} = "lun_comment";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_prefix_size()
# Objective:   Get prefix size
# Details:     See POD documentation at the end of this file
################################################################################
sub get_prefix_size {
    my ($self, %args) = @_;

    $args{field} = "prefix_size";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_uuid()
# Objective:   Get lun uuid
# Details:     See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self, %args) = @_;

    $args{field} = "uuid";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_volume()
# Objective:   Get volume this lun created on
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volume {
    my ($self, %args) = @_;

    $args{field} = "volume";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_vserver()
# Objective:   Get lun vserver
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vserver {
    my ($self, %args) = @_;

    $args{field} = "vserver";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# Method Name: get_qtree()
# Objective:   Get lun vserver
# Details:     See POD documentation at the end of this file
################################################################################
sub get_qtree {
    my ($self, %args) = @_;

    $args{field} = "qtree";

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method Name: _get_param_defaults
# Objective: Gets default lun parameters
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_param_defaults {
    my ($self, %args) = @_;

    my %default_params = (%args, vserver => $self->{vserver});

    # if path is not provided, we need volume, qtree and lun params
    unless (exists $args{'path'}) {
        if (!defined($self->{qtree}) and length($self->{qtree}) == 0){
            $self->{qtree} = "\"\"";
        }

        my $lun     = $self->{lun};
        my $volume  = $self->{volume};
        my $qtree   = $self->{qtree};

        # zapi required lun path so we build it here
        if ($self->get_cmd_interface() =~ /zapi/i) {
            $default_params{'path'} = $self->_get_lun_path(-volume => $volume,
                                                           -qtree => $qtree,
                                                           -lun => $lun);
        } else {
            $default_params{'lun'} = $lun;
            $default_params{'volume'} = $volume;
            $default_params{'qtree'} = $qtree if (length($qtree));;
        }
    }

    return %default_params;
}

################################################################################
# Method Name: _get_lun_path
# Objective: Gets lun path from %args
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_lun_path {
    my ($self, %args) = @_;

    my ($volume, $lun);
    my $lun_path = "";

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # check for mandatory parameters
    if (defined $args{volume} and length $args{volume}) {
        $volume = delete $args{volume};
    } else {
        $self->_set_err("Mandatory parameter \"volume\" not specified ");
        return $lun_path;
    }
    if (defined $args{lun} and length $args{lun}) {
        $lun = delete $args{lun};
    } else {
        $self->_set_err("Mandatory parameter \"lun\" not specified ");
        return $lun_path;
    }

    # get qtree param if exist
    my $qtree = (exists $args{qtree}) ? delete $args{qtree} : "";

    # build lun path
    $lun_path = "/vol/" . $volume;
    $lun_path .= "/" . $qtree if (length($qtree) and (($qtree ne '""') and ($qtree ne "-")));
    $lun_path .= "/" . $lun;

    return $lun_path;
}

################################################################################
# Method Name: _split_path_into_names
# Objective: Split path into names and populate them
# Details:   See POD documentation at the end of this file.
################################################################################
sub _split_path_into_names {
    my ($self) = @_;

    # populate volume and qtree attributes
    my @fields = split(/\//, $self->{path});
    $self->{volume} = $fields[2];
    if (scalar(@fields) == 5) {
        # lun is in qtree
        $self->{qtree}  = $fields[3];
        $self->{lun}    = $fields[4];
    } else {
        $self->{qtree}  = "";
        $self->{lun}    = $fields[3];
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Lun
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    my $use_cache =  $SanUtil::FALSE;
    my $command = (defined $args{command}) ? delete $args{command} : "lun_show";

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # set use_cache flag if exists
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {
        # the get() will handle error
        return $self->get($args{field});
    } else {
        # the get_field_from_controller() will handle error
        return $self->get_field_from_controller(cmd => $command,
                                                $self->_get_param_defaults(),
                                                %args);
    }
}

################################################################################
# Method Name: _create_lun
# Objective: create lun
################################################################################
sub _create_lun {
    my ($self, %args) = @_;

    # set cmd based on interface and args
    my $cmd = ($self->get_cmd_interface() =~ /zapi/i and
                defined($args{'file-path'}))
         ? 'lun_create_from_file'
         : 'lun_create';

    %args = (%args, cmd => $cmd);

    # use common invoke from base class (San::CommonInterface)
    # to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # call lun refresh if enable cache is set
    if ($self->{enable_cache}) {
        $self->refresh(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to refresh lun " . $args{lun}, $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # set initialized flag
    $self->{initialized} = $SanUtil::TRUE;

    return $SanUtil::TRUE;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Lun.pm

=head1 Author

Korry Nguyen (dl-ssan-automation)

=head1 Description

Provides an interface to create and manage C-mode luns

=head1 See Also

Lun.pm inherits methods from B<San::CommonInterface> and from its compatibility
layer B<San::ControllerCluster::Filer10M::Vserver::Volume::LunCompat>. See these
methods for further interface details.

=head1 Attributes

=head2 B<filer>

string = controller name

=head2 B<name>

string = name of lun object to be created

=head2 B<debug_level>

interger = 0 - 100 (default is 0)

=head1 B<Public Methods>

=head2 B<new(%args)>

Create a lun object

=over 4

=item Inputs

=item * I<-filer> => <controller name>

Controller name

=item * I<-name> => <name of lun object>

Lun object name to be created

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. With higher level, it will log more debug messages

=item Outputs

Returns a lun object

=item Examples

 # ex. 1: Create lun directly, make sure you have a vserver and a volume

 use San::ControllerCluster::Filer10M::Vserver::Volume::Lun;

 my $vol = "vs0_vol";
 my $vserver = "vs0";
 my $lun = "lun_0";

 # Create filer object
 $filer_obj = new SanFiler();
 if ($filer->error()) {
    logcomment("Failed to create filer object" . $filer_obj->errmsg());
    return $TCD::FAIL;
 }

 # Get the vol object. Assuming the vol exists on that node.
 $vol_obj = $filer_obj->get_vol(volume  => $vol,
                                vserver => $vserver);
 if ($vol_obj->error()) {
    logcomment("Failed to get vol object" . $vol_obj->errmsg());
    return $TCD::FAIL;
 }

 # Create the lun
 my $lun_obj = new Lun(-filer => $filer,
                       -name => $lun,
                       -parent_obj => $vol_obj,
                       -vserver => $vserver,
                       -volume => $vol,
                       -ostype => "windows",
                       -size => $size,
                       -debug_level => 0);
 if ($lun_obj->error()) {
    logcomment("Failed to create lun object" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

 # ex. 2: Create lun from a volume object

 my $lun_obj = $volume_obj->create_lun(...);
 if ($lun_obj->error()) {
    logcomment("Failed to create lun object" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<destroy(%args)>

Destroys a lun object by invoking "lun_destroy" CDEF.

=over 4

=item Inputs

=item * I<-force> => <1|0> (optional)

Force destroy a lun. Same as passing -force to a lun destroy command. Default
value is FALSE.

=item Outputs

Returns TRUE if the lun was destroyed successfully. Returns FALSE and sets the
error flag if something failed.

=item Examples

 # Destroy a lun
 my $lun_obj->destroy();
 if ($lun_obj->error()) {
    logcomment("Failed to destroy lun object" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<is_offline()>

Check to see if lun state is 'offline'.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the Lun is offline and FALSE otherwise. This method does not
set an error flag.

=item Examples

 # Destroy a lun
 if ($lun_obj->is_offline()) {
     # Do something.
 }

=back

=head2 B<is_readonly()>

Check to see if lun is read only .

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the Lun is read only and FALSE otherwise. This method does not
set an error flag.

=item Examples

 # Check if the lun is read only
 if ($lun_obj->is_readonly()) {
     # Do something.
 }

=back

=head2 B<modify()>

Modifies Lun attributes. It invokes lun_modify CDEF

=over 4

=item Inputs

=item * I<-size> => <size of lun> (optional)

Modifies the size of the lun.

=item * I<-state> => <offline|online> (optional)

Modifies the state of a lun. Valid values are offline and online.

=item Outputs

Returns TRUE if the Lun was modified successfully and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Modify state
 $lun_obj->modify(-state => 'online');
 if ($lun_obj->error()) {
    logcomment("Failed to modify lun object" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=head2<lun_get_select_attribute>
Returns the Lun select attribute of the lun like creation time, 
previous serial number (serial number of parent)

=item Examples

 $hash_ref = $lun_obj->lun_get_select_attribute();
 if ($lun_obj->error()) {
    logcomment("Failed to get select attribute" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

 $Test->nlog("LUN select Attr: " . $hash_ref->{'select-attribute'});
 $Test->nlog("Creation Time  : " . $hash_ref->{'creation-time'});
 $Test->nlog("Previous Serial: " . $hash_ref->{'previous-serial'});

=back

=head2 B<map()>

Maps a Lun to an iGroup.

=over 4

=item Inputs

=item * I<-name> => <name>.

The name of the lun.

=item * I<-igroup> => <igroup_name>.

The name of the igroup.

=item * I<-force> => <1|0> (optional)

Same as passing -force to a lun map command. Default value is FALSE.

=item Outputs

Returns TRUE if the Lun was mapped successfully and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Map lun
 $lun_obj->map(-igroup=> $igroup_name);
 if ($lun_obj->error()) {
    logcomment("Failed to map lun object" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<unmap()>

Un maps a Lun from an iGroup.

=over 4

=item Inputs

=item * I<-name> => <name>.

The name of the lun.

=item * I<-igroup> => <igroup_name>.

The name of the igroup.

=item * I<-force> => <1|0> (optional)

Same as passing -force to a lun map command. Default value is FALSE.

=item Outputs

Returns TRUE if the Lun was unmapped successfully and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Unmap lun
 $lun_obj->unmap(-igroup=> $igroup_name);
 if ($lun_obj->error()) {
    logcomment("Failed to unmap lun object" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<online()>

Brings a lun online.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the Lun's state was changed to online and FALSE otherwise. This
will set an error flag if something fails.

=item Examples

 # Bring Lun online
 $lun_obj->online();
 if ($lun_obj->error()) {
    logcomment("Failed to bring a lun online" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<offline()>

Brings a lun offline.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the Lun's state was changed to offline and FALSE otherwise. This
will set an error flag if something fails.

=item Examples

 # Bring Lun offline
 $lun_obj->offline();
 if ($lun_obj->error()) {
    logcomment("Failed to bring a lun offline" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<is_alua_enabled()>

Determines if alua os enabled on this mapped lun.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if alua is enabled, FALSE otherwise. This
will set an error flag if something fails.

=item Examples

 # Check to see if the mapped lun is alua enabled
 $alua_enabled = $lun_obj->is_alua_enabled();
 if ($lun_obj->error()) {
    logcomment("Failed to check if alua is enabled" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<is_mapped()>

Determines if a lun is mapped to an igroup.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns TRUE if the is mapped to an iGroup and FALSE otherwise. This
will set an error flag if something fails.

=item Examples

 # Check to see if a lun is mapped
 $is_mapped = $lun_obj->is_mapped();
 if ($lun_obj->error()) {
    logcomment("Failed to check if a lun is mapped" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_list_info()>

Get lun information. This method invokes a basic lun_show cdef on a filer.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns a reference to a hash containing the output of "lun show" command. If
it fails, it will return a null and set the error flag.

=item Examples

 # Check to see if a lun is mapped
 $lun_info = $lun_obj->get_list_info();
 if ($lun_obj->error()) {
    logcomment("Failed to get lun information" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_maps()>

A wrapper to map_list_info to get a list of igroups that map to this lun

=over 4

=item Inputs

=item * None.

=item Outputs

Returns a reference to a list containing the output of "lun show map" command.
If it fails, it will return a null or an empty list and set the error flag.

=item Examples

 # Get a list of mapped lun info
 $lun_info = $lun_obj->get_maps();
 if ($lun_obj->error()) {
    logcomment("Failed to get a list of mapped lun info" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<move()>

Move a lun giving it a new name

=over 4

=item Inputs

=item * I<-name> => <name>.

The new name of the lun.

=item Outputs

Returns TRUE if the Lun was moved successfully and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Move lun
 $lun_obj->move(name=> $new_lun_name);
 if ($lun_obj->error()) {
    logcomment("Failed to move lun " . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<resize()>

 Resize a lun.

=over 4

=item Inputs

=item * I<-size> => <size>.

The new size of the lun.

=item Outputs

Returns TRUE if the resize was successfully and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Set comment
 $lun_obj->resize(size=> $new_size);
 if ($lun_obj->error()) {
    logcomment("Failed to resize the lun " . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<set_comment()>

Set the comment for a lun.

=over 4

=item Inputs

=item * I<-comment> => <comment>.

The new comment of the lun.

=item Outputs

Returns TRUE if the comment was successfully set and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Set comment
 $lun_obj->set_comment(comment=> $new_comment);
 if ($lun_obj->error()) {
    logcomment("Failed to set comment of lun " . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<set_serial()>

Set the serial number for a lun.

=over 4

=item Inputs

=item * I<-serial> => <serial>.

The new serial number of the lun.

=item Outputs

Returns TRUE if the serial number was successfully set and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Set serial number
 $lun_obj->set_serial(serial=> $new_serial_number);
 if ($lun_obj->error()) {
    logcomment("Failed to set serial of lun " . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<set_device_id()>

Set the device-id for a lun.

=over 4

=item Inputs

=item * I<-device-id> => <device-id>.

The new device id of the lun.

=item Outputs

Returns TRUE if the device id was successfully set and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Set device id
 $lun_obj->set_device_id('device-id'=> $new_device_id);
 if ($lun_obj->error()) {
    logcomment("Failed to set device-id of lun " . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<unset_device_id()>

Unset the device-id for a lun.

=over 4

=item Inputs

=item Outputs

Returns TRUE if the device id was successfully unset and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Unset device id
 $lun_obj->unset_device_id();
 if ($lun_obj->error()) {
    logcomment("Failed to unset device-id of lun " . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<set_space_resrv()>

Set the space reservation for a lun.

=over 4

=item Inputs

=item * vserver:

Vserver Name.

=item * volume:

Volume Name.

=item * lun:

LUN Name.

=item * space-reserve:

Enable/Disable the space reservation on this LUN.

=item * path (optional):

Lun Path.

=item * qtree (optional):

Qtree Name.


=item Outputs

Returns TRUE if the space reservation was successfully changed and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Unset device id
 $lun_obj->unset_device_id();
 if ($lun_obj->error()) {
    logcomment("Failed to change space reservation of lun " . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<set_space_allocation()>

Set the space allocation for a lun.

=over 4

=item Inputs

=item * vserver:

Vserver Name.

=item * volume:

Volume Name.

=item * lun:

LUN Name.

=item * space-allocation:

Enable/Disable the space allocation on this LUN.

=item * path (optional):

Lun Path.

=item * qtree (optional):

Qtree Name.


=item Outputs

Returns TRUE if the space allocation was successfully set and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Enabled Space Allocation
 $lun_obj->set_space_allocation(-space-allocation => 'enabled');
 if ($lun_obj->error()) {
    logcomment("Failed to change the space-allocation of lun " . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_lun_geometry()>

Get lun geometry information. It invokes lun_geometry cdef.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns a reference to a hash containing the output of the "lun geometry"
command.  If it fails, it will return a null and set the error flag.

=item Examples

 # Get lun geometry info
 $lun_geometry_info = $lun_obj->get_lun_geometry();
 if ($lun_obj->error()) {
    logcomment("Failed to get lun geometry information" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=head2 B<get_lun_stats()>

Get lun stats. It invokes lun_stats cdef.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns a reference to a hash containing the output of the "lun stats"
command.  If it fails, it will return a null and set the error flag.

=item Examples

 # Get lun stats
 $lun_stats = $lun_obj->get_lun_stats();
 if ($lun_obj->error()) {
    logcomment("Failed to get lun stats" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_alua()>

Gets alua from this  mapped lun.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the alua attribute on this object..

=item Examples

 # Get alua
 $lun_obj->get_alua();
 if ($lun_obj->error()) {
    logcomment("Failed to get alua" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_readonly()>

Gets the read-only attribute from a lun object.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the read-only attribute on this object.

=item Examples

 # Get read-only
 $attr = $lun_obj->get_readonly();
 if ($lun_obj->error()) {
    logcomment("Failed to get read-only attribute" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_igroup()>

Gets igroup that mapped to this lun.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the igroup attribute on this object..

=item Examples

 # Get igroup
 $lun_obj->get_igroup();
 if ($lun_obj->error()) {
    logcomment("Failed to get igroup" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_igroup_type()>

Gets igroup type of this mapped lun.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the igroup type attribute on this object..

=item Examples

 # Get igroup type
 $lun_obj->get_igroup_type();
 if ($lun_obj->error()) {
    logcomment("Failed to get igroup type" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_initiators()>

Gets initiators that mapped to this lun.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the initiators attribute on this object..

=item Examples

 # Get initiators that mapped to this lun
 # it's a scalar value
 $initiators = $lun_obj->get_initiators();
 if ($lun_obj->error()) {
    logcomment("Failed to get initiator" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_lun_id()>

Gets the LU number (ID) of this mapped lun.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the lun id attribute on this object.

=item Examples

 # Get LUN ID
 my $lunid = $lun_obj->get_lun_id();
 if ($lun_obj->error()) {
    logcomment("Failed to get lun id" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_name()>

Gets the name of this lun.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the name of this lun.

=item Examples

 # Get lunid
 my $lun = $lun_obj->get_name();
 if ($lun_obj->error()) {
    logcomment("Failed to get name" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_parent_obj()>

Gets the parent object of this lun (A volume object).

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the parent object of a lun.

=item Examples

 # Get parent object
 my $vol_obj = $lun_obj->get_parent_obj();
 if ($lun_obj->error()) {
    logcomment("Failed to get parent object" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_ostype()>

Gets the ostype of this lun

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the ostype of a lun.

=item Examples

 # Get ostype
 my $vol_obj = $lun_obj->get_ostype();
 if ($lun_obj->error()) {
    logcomment("Failed to get ostype" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_text_device_id()>

Gets the text device id of this lun

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the text device id of a lun.

=item Examples

 # Get text_device_id
 my $text_device_id = $lun_obj->get_text_device_id();
 if ($lun_obj->error()) {
    logcomment("Failed to get text device id" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_binary_device_id()>

Gets the binary device id of this lun

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the binary device id of a lun.

=item Examples

 # Get binary_device_id
 my $binary_device_id = $lun_obj->get_binary_device_id();
 if ($lun_obj->error()) {
    logcomment("Failed to get binary device id" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_legacy_device_id()>

Gets the legacy device id of this lun

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the legacy device id of a lun.

=item Examples

 # Get legacy_device_id
 my $legacy_device_id = $lun_obj->get_legacy_device_id();
 if ($lun_obj->error()) {
    logcomment("Failed to get legacy device id" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_prefix_size()>

Gets the prefix-size attribute from a lun object.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the prefix-size attribute on this object.

=item Examples

 # Get read-only
 $attr = $lun_obj->get_prefix_size();
 if ($lun_obj->error()) {
    logcomment("Failed to get prefix-size attribute" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_serial()>

Gets the serial number of this lun

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the serial number of a lun.

=item Examples

 # Get serial
 my $serial = $lun_obj->get_serial();
 if ($lun_obj->error()) {
    logcomment("Failed to get serial" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_size()>

Gets the size of this lun

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the lun size.

=item Examples

 # Get lun size
 my $serial = $lun_obj->get_size();
 if ($lun_obj->error()) {
    logcomment("Failed to get size" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_space_resv()>

Gets the space reservation of this lun

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the space reservation size.

=item Examples

 # Get space resv
 my $space_resv = $lun_obj->get_space_resv();
 if ($lun_obj->error()) {
    logcomment("Failed to get space resv" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_space_allocation()>

Gets the space allocation of this lun

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the space allocation boolean (disabled|enabled).

=item Examples

 # Get space resv
 my $space_allocation = $lun_obj->get_space_allocation();
 if ($lun_obj->error()) {
    logcomment("Failed to get space allocation" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_suffix_size()>

Gets the suffix-size attribute from a lun object.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the suffix-size attribute on this object.

=item Examples

 # Get read-only
 $attr = $lun_obj->get_suffix_size();
 if ($lun_obj->error()) {
    logcomment("Failed to get suffix-size attribute" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_state()>

Gets the Lun state

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the lun state: online or offline

=item Examples

 # Get lun state
 my $state= $lun_obj->get_state();
 if ($lun_obj->error()) {
    logcomment("Failed to get state" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_uuid()>

Gets the Lun uuid

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the lun uuid attribute

=item Examples

 # Get uuid comment
 my $uuid = $lun_obj->get_uuid();
 if ($lun_obj->error()) {
    logcomment("Failed to get uuid" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_comment()>

Gets the Lun comment

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the lun comment attribute

=item Examples

 # Get lun comment
 my $comment = $lun_obj->get_comment();
 if ($lun_obj->error()) {
    logcomment("Failed to get comment" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_vserver()>

Gets the vserver owning this Lun.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the vserver attribute

=item Examples

 # Get vserver
 my $vserver = $lun_obj->get_vserver();
 if ($lun_obj->error()) {
    logcomment("Failed to get vserver" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_qtree()>

Gets the qtree owning this Lun.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the qtree attribute

=item Examples

 # Get qtree
 my $qtree = $lun_obj->get_qtree();
 if ($lun_obj->error()) {
    logcomment("Failed to get qtree" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_path()>

Get the lun path.

=over 4

=item Inputs

None

=item Outputs

The path string of the lun

=item Examples

    if ($SanFilerObj->lun_exist($lun_obj->get_path())) {
        $Test->nlog("-- cleanup lun=" . $lun_obj->get_path() . " this test created");
        $lun_obj->destroy();
    }


=back

=head2 B<get_class()>

Gets the class of the lun. Vvol, PEP, or Regular.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the class of the lun

=item Examples

 # Get qtree
 my $lunclass = $lun_obj->get_class();
 if ($lun_obj->error()) {
    logcomment("Failed to get class" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_bind()>

Gets the list of peps a vvol is bound to.

=over 4

=item Inputs

=item * None

=item Outputs

Returns all of the peps a vvol is bound to

=item Examples

 # Get qtree
 my @peps = $lun_obj->get_binds();
 if ($lun_obj->error()) {
    logcomment("Failed to get binds" . $lun_obj->errmsg());
    return $TCD::FAIL;
 }

=back