# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to a Qtree object
##
## @author mcclella@netapp.com, dl-ssan-test
## @status Shared

#  Code formatted using perltidy command line:
#  "perltidy --perl-best-practices --cuddled-else --nooutdent-keywords --notabs
#  --closing-side-comments --output-line-ending=unix --nostandard-output
#  --nostandard-error-output --paren-tightness=2 --square-bracket-tightness=2
#  --brace-tightness=2"

package San::ControllerCluster::Filer10M::Vserver::Volume::Qtree;

# compiler directives
use strict;

# imported modules
use San::ControllerCluster::Filer10M::Vserver::Volume::Lun;
use Data::Dumper;

# inherit base class
use base qw( San::CommonInterface
             San::ControllerCluster::Filer10M::Vserver::Volume::QtreeCompat );

my @Sub_components = qw( lun );

###############################################################################
# Method Name: new
# Objective: Create new Qtree object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($self);

    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    if (exists $args{name}) {
        $args{qtree} = delete $args{name};
    }

    # verify mandatory cmd argument
    unless (defined $args{filer} and length $args{filer}) {
        $self->_set_err("mandatory parameter \"-filer\" not specified ");
        return $self;
    }
    unless (defined $args{parent_obj} and length $args{parent_obj}) {
        $self->_set_err("mandatory parameter \"-parent_obj\" not specified ");
        return $self;
    }

    $self->sanlog(
        -msg       => $self->dumper_get(\%args, $self->get_debug_level()),
        -msg_level => 2
    );

    # since path and qtree name could be used to create qtree, we need to
    # get its name for the SUPER::_new()
    unless (exists $args{'qtree'}) {

        # make sure qtree path exists
        if (exists $args{'qtree-path'}) {

            # strip path
            $args{qtree} = File::Basename::basename($args{'qtree-path'});
        } else {
            $self->_set_err(
                "either \"-qtree or -qtree-path\" must be specified");
            return $self;
        }
    } ## end unless (exists $args{'qtree'...

    # use base class (San::CommonInterface) via SUPER::_new method
    # to contruct qtree object
    $self = $class->SUPER::_new(
        -name  => $args{qtree},
        -filer => $args{filer},
        %args
    );
    if ($self->error()) {
        $self->_set_err("failed to construct qtree object", $self->errmsg());
        return $self;
    }

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(
        -msg       => $self->dumper_get(\%args, $self->get_debug_level()),
        -msg_level => 2
    );

    # store Qtree elements on a list
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }

    # create get object accessor method for lun
    $self->create_accessor_methods(-type => "get_obj",
                                   -pkg  => __PACKAGE__,
                                   -list => \@Sub_components);

    # we can now remove params from the args list that we dont need
    delete $args{parent_obj};
    delete $args{debug_level};

    # delete qtree path if initialized flag is set
    delete $args{path} unless $self->{initialized};

    # if the obj has been initialized, just refresh to update info
    # otherwise initialize it by create and update its object
    unless ($args{initialized}) {

        # qtree has not created, so we need to create it
        $self->_create_qtree(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to _create_qtree() ", $self->errmsg());
            return $self;
        }
    } ## end unless ($args{initialized})

    return $self;
} ## end sub new

###############################################################################
# Method Name: create_file
# Objective:  Create file on qtree
# Details:    See POD documentation at the end of this file.
###############################################################################
sub create_file {
my ($self, %args) = @_;

     my $parent;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # check for required arguments
    unless (defined $args{size} and length $args{size}){
        $self->_set_err("Mandatory parameter \"size\" not specified ");
        return $SanUtil::FALSE;
    }
    unless (defined $args{from_file} and length $args{from_file}){
        $self->_set_err("Mandatory parameter \"from_file\" not specified ");
        return $SanUtil::FALSE;
    }
    $args{from_file} = $self->get_name().'/'.$args{from_file};

     # Get the parent volume to call create method from
     $parent = $self->get_parent_obj();
     if ($self->error()) {
        $self->_set_err("failed to get parent ", $self->errmsg());
        return $SanUtil::FALSE;
    }

     # Call Volume Create Method with correct qtree file path
     $parent->create_file(%args);
     if ($self->error()) {
        $self->_set_err("failed to create file properly ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::True;
} ## end sub create_file

###############################################################################
# Method Name: remove_file
# Objective:  remove file on qtree
# Details:    See POD documentation at the end of this file.
###############################################################################
sub remove_file {
    my ($self, %args) = @_;

    my $parent;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # check for required arguments
    unless (defined $args{from_file} and length $args{from_file}){
        $self->_set_err("Mandatory parameter \"from_file\" not specified ");
        return $SanUtil::FALSE;
    }
    $args{from_file} = $self->get_name().'/'.$args{from_file};

     # Get the parent volume to call remove method from
    $parent = $self->get_parent_obj();
    if ($self->error()) {
        $self->_set_err("failed to get parent ", $self->errmsg());
        return $SanUtil::FALSE;
    }

     # Call Volume Remove Method with correct qtree file path
     $parent->remove_file(%args);
     if ($self->error()) {
        $self->_set_err("failed to create file properly ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::True;
} ## end sub remove_file

###############################################################################
# Method Name: delete
# Objective: delete qtree object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub delete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "volume_qtree_delete");

    # for zapi, qtree is qtree path
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{qtree} =  $self->create_common_path(volume => $args{volume},
                                                  qtree  => $args{qtree});
        delete $args{volume} if ($args{volume});
        delete $args{name} if ($args{name});
    }

    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err(
            "failed to invoke_cmd_obj "
                . $self->dumper_get(\%args, $self->get_debug_level()),
            $self->errmsg()
        );
        return $SanUtil::FALSE;
    } ## end if ($self->error())

    # set name to "_DELETE_" to indicate qtree has been deleted
    $self->{name}  = "_DELETE_";
    $self->{qtree} = "_DELETE_";

    return $SanUtil::TRUE;
} ## end sub delete

###############################################################################
# Method Name: destroy
# Objective: destroy qtree object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub destroy {
    my ($self, %args) = @_;

    return $self->delete(%args);
}

###############################################################################
# Method Name: modify
# Objective: modify qtree attribute
# Details:   See POD documentation at the end of this file.
###############################################################################
sub modify {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "volume_qtree_modify");

    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err(
            "failed to invoke_cmd_obj "
                . $self->dumper_get(\%args, $self->get_debug_level()),
            $self->errmsg()
        );
        return $SanUtil::FALSE;
    } ## end if ($self->error())

    return $SanUtil::TRUE;
} ## end sub modify

###############################################################################
# Method Name: refresh()
# Objective: Refresh object's data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub refresh {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Initialize cmd parameters
    %args = ($self->_get_param_defaults(%args), cmd => "volume_qtree_show");

    # For ZAPI, qtree is qtree path
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{path} =  $self->create_common_path(volume => $args{volume},
                                                 qtree  => $args{qtree});
        delete $args{volume} if ($args{volume});
        delete $args{name}   if ($args{name});
        delete $args{qtree}  if ($args{qtree});
    }

    $self->sanlog(
        -msg       => "args to volume_qtree_show", $self->dumper_get(\%args),
        -msg_level => 3
    );

    # use base class (San::CommonInterface) refresh via SUPER::refresh
    # to update qtree object data
    $self->SUPER::refresh(%args);
    if ($self->error()) {

        # error is already set, bubble up
        return $SanUtil::FALSE;
    }
    return $SanUtil::TRUE;
} ## end sub refresh

###############################################################################
# Method Name: get_list_info()
# Objective: Get qtree information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (
        %args,
        cmd => "volume_qtree_show",
        $self->_get_param_defaults()
    );

    # For ZAPI, qtree is qtree path
    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{path} =  $self->create_common_path(volume => $args{volume},
                                                 qtree  => $args{qtree});
        delete $args{volume} if ($args{volume});
        delete $args{name}   if ($args{name});
        delete $args{qtree}  if ($args{qtree});
    }

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err(
            "failed to invoke_cmd_obj "
                . $self->dumper_get(\%args, $self->get_debug_level()),
            $self->errmsg()
        );
        return (wantarray()) ? () : [];
    } ## end if ($self->error())

    return (wantarray()) ? @output : \@output;
} ## end sub get_list_info

###############################################################################
# Method Name: get_statistics()
# Objective: Get qtree statistics information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_statistics {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (
        %args,
        cmd => "volume_qtree_statistics",
        $self->_get_param_defaults()
    );

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err(
            "failed to invoke_cmd_obj "
                . $self->dumper_get(\%args, $self->get_debug_level()),
            $self->errmsg()
        );
        return (wantarray()) ? () : [];
    } ## end if ($self->error())

    return (wantarray()) ? @output : \@output;
} ## end sub get_statistics

################################################################################
# Method Name: get_lun_maxsize()
# Objective:   Get lun maxsize
# Details:     See POD documentation at the end of this file
################################################################################
sub get_lun_maxsize {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # get qtree path
    $args{path} = $self->create_common_path(volume => $self->get_volume(),
                                            qtree => $self->get_name());
    if ($self->error()) {
        $self->_set_err("failed to get qtree path", $self->errmsg());
        return {};
    }

    # Initialize cmd parameters
    %args = (
        %args,
        cmd => "lun_maxsize",
        $self->_get_param_defaults()
    );

    # delete name and volume since we will be using path
    delete $args{name} if ($args{name});
    delete $args{volume} if ($args{volume});
    delete $args{qtree} if ($args{qtree});

    # Get a CmdsDriver object and executes the actual command
    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err(
            "failed to invoke_cmd_obj "
                . $self->dumper_get(\%args, $self->get_debug_level()),
            $self->errmsg()
        );
        return (wantarray()) ? () : [];
    } ## end if ($self->error())

    return (wantarray()) ? @output : \@output;
} ## end sub get_lun_maxsize

###############################################################################
# Method Name: rename()
# Objective: Move a qtree (basicaly this is a rename)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub rename {
    my ($self, %args) = @_;
    $self->_clear_err();

    my $path = $self->get_path();
    if ($self->error()) {
        $self->_set_err("Could not get qtree path", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Rerename the optional dash '-' prefix from the input args
    %args = &SanUtil::strip_dash(%args);

    # Check for the required path argument
    unless (defined $args{newname} and length $args{newname}) {
        $self->_set_err("Mandatory parameter \"newname\" not specified ");
        return $SanUtil::FALSE;
    }
    my $new_name = $args{newname};

    # Initialize cmd parameters
    %args = (
        %args,
        cmd          => "volume_qtree_rename",
        'qtree-path' => $path,
        "newname"    => "$new_name",
        $self->_get_param_defaults()
    );

    # The qtree and volume arguments can't be used with the path argument
    delete $args{name}   if ($args{name});
    delete $args{qtree}  if ($args{qtree});
    delete $args{volume} if ($args{volume});

    $self->sanlog(
        -msg => "Qtree::rename() - rename a qtree from $path to $new_name",
        -msg_level => 1
    );

    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err(
            "failed to invoke_cmd_obj "
                . $self->dumper_get(\%args, $self->get_debug_level()),
            $self->errmsg()
        );
        return (wantarray()) ? () : [];
    } ## end if ($self->error())

    # Update our changed attributes after the rename and refresh
    my $new_path = File::Basename::dirname($path) . '/' . $new_name;
    $self->{qtree} = $new_name;
    $self->{name}  = $new_name;
    $self->refresh();
    if ($self->error()) {
        $self->_set_err("Failed to refresh qtree object:", $self->errmsg());
        return $SanUtil::FALSE;
    }
    $self->{'qtree-path'} = $new_path;

    return $SanUtil::TRUE;
} ## end sub rename

###############################################################################
# Method Name: create_lun_obj
# Objective:  Create lun object
# Details:    See POD documentation at the end of this file.
###############################################################################
sub create_lun_obj {
    my ($self, %args) = @_;

    my ($lun_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # convert parameters to maintain compatibility w/ 7mode
    if (exists $args{type}) {
        $args{ostype} = $args{type};
        delete $args{type};
    }
    if (exists $args{name}) {
        $args{lun} = $args{name};
        delete $args{name};
    }
    # if reserve param is set to 0, set space-resv to noreserve
    # NOTE: this feature has not yet implemented, ignore for now
    #if (exists $args{reserve} and $args{reserve} == 0) {
    #    $args{"space-resv"} = "noreserve";
    #    delete $args{reserve};
    #}
    if (exists $args{reserve}) {
        delete $args{reserve};
    }

    # append vserver, volume and qtree parameters to argument list
    %args = (%args, $self->_get_param_defaults());

    # dump only first depth of hash if debug_level < 3
    # dump everything of debug_level => 3
    $self->sanlog(-msg => $self->dumper_get(\%args, $self->get_debug_level()),
                  -msg_level => 2);

    # instantiate lun object
    $lun_obj =
      San::ControllerCluster::Filer10M::Vserver::Volume::Lun->new(-parent_obj => $self,
                                                                  -filer  => $self->get_filer(),
                                                                  %args);
    if ($lun_obj->error()) {
        $self->_set_err("failed to instantiate lun object", $lun_obj->errmsg());
        return {};
    } else {
        # store lun object in filer object
        push @{$self->{lun_objs}}, $lun_obj;
    }
    return $lun_obj;
} ## end sub create_lun_obj

################################################################################
# Method Name: get_lun()
# Objective:   Get lun object.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self, %args) = @_;

    my ($lun_obj, %tmp_args);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # hate this crap but we have to make it compatible w/ 7-mode. This means
    # we have to take scalar as param instead of hash

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|lun)$/ and length $args{$key}) {

            # assume it's a hash
            %tmp_args = %args;
        } else {

            # this is a scalar
            $tmp_args{name} = $key;
        }
        last;
    } ## end foreach my $key (keys %args)

    # set tmp args back to args
    %args = %tmp_args;

    if (exists $args{lun}) {
        $args{name} = $args{lun};
        delete $args{lun};
    }

    # check if full lun path include filer name was
    # given.  If so extract the lun name.
    if ($args{name} =~ /^\//) {
        $args{name} = File::Basename::basename($args{name});
    }

    unless (defined $args{name} and length $args{name}) {
        $self->_set_err(
            "mandatory parameter \"\$name or -name => \$name\" is required");
        return {};
    }

    # get_param_defaults will put the qtree and volume into the args
    %args = (%args,
             $self->_get_param_defaults()
            );

    # If cache flag is set, get it from cache
    # If it does exist in cache, query the controller for the particular volume
    if ($self->{use_cache} && exists $self->{lun_objs}) {
        # Get data from cache
        ($lun_obj) = grep ({$_->get_name() =~ /^$args{name}$/} @{$self->{lun_objs}});
    } else {
        # Discover this particular qtree on the controller
        # Get lun objects

        $lun_obj = $self->get_lun_objs(-name => $args{name});
        if ($self->error()) {
            $self->_set_err("failed to get lun object: " . $args{name},
                            $self->errmsg());
            return {};
        }
    }

    unless (ref($lun_obj) =~ /lun/i) {
        $self->_set_err("can not find lun " . $args{name});
        return {};
    }

    return $lun_obj;
} ## end sub get_lun

################################################################################
# Method Name: get_luns
# Objective:   return array of lun objects
# Details:     See POD documentation at the end of this file
################################################################################
sub get_luns {
    my ($self) = @_;

    $self->_clear_err();

    # discover and get lun objects
    my (@lun_objs) = $self->get_lun_objs();
    if ($self->error()) {
        $self->_set_err("failed to get lun objects", $self->errmsg());
        return (wantarray()) ? () : [];
    }
    return (wantarray()) ? @lun_objs : \@lun_objs;
} ## end sub get_luns

################################################################################
# Method Name: lun_exist()
# Objective:   Determines if lun exists. This maps one-to-one
#              to 7-mode
# Details:     See POD documentation at the end of this file
################################################################################
sub lun_exist {
    my ($self, %args) = @_;

    my (@output, %tmp_args);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    # If the user passed in a scalar (string) then the args hash will contain
    # one entry that is keyed by that string.  We set the name to that
    # key/string.
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|lun)$/ and length $args{$key}) {

            # assume it's a hash
            %tmp_args = %args;
        } else {

            # this is a scalar
            $tmp_args{lun} = $key;
        }
        last;
    } ## end foreach my $key (keys %args)

    # set tmp_args back to args
    %args = %tmp_args;

    # convert vserver to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{name}) {
        $args{lun} = $args{name};
        delete $args{name};
    }

    # verify name param
    unless (defined $args{lun} and length $args{lun}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return $SanUtil::FALSE;
    }

    # build cmd vserver show
    %args = (
        %args,
        cmd => "lun_show",
        $self->_get_param_defaults()
    );

    if ($self->get_cmd_interface() =~ /zapi/i) {
        $args{path} = '/vol/' . delete $args{volume};
        if (exists($args{qtree}) and $args{qtree} ne '""') {
            $args{path} .= '/' . delete $args{qtree};
        }
        $args{path} .= '/' . delete $args{lun};
    }

    # invoke cmd command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke cmd \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} ## end sub lun_exist

###############################################################################
# Method Name: lun_list_info()
# Objective: Get lun information from this qtree
# Details:   See POD documentation at the end of this file.
###############################################################################
sub lun_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (
        %args,
        cmd    => "lun_show",
        $self->_get_param_defaults()
    );

    if ($self->get_cmd_interface() =~ /zapi/i) {
        # build qtree path
        $args{path} = $self->create_common_path(volume => $self->get_volume(),
                                                qtree => $self->get_name());
        if ($self->error()) {
            $self->_set_err("failed to get qtree path", $self->errmsg());
            return {};
        }
        # append lun to path
        if (exists($args{lun}) and $args{lun} ne '""') {
            $args{path} .= '/' . delete $args{lun};
        } else {
            $args{path} .= '/*';
        }
    }

    # use base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain lun info from this volume
    my @records = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err(
            "failed to invoke_cmd_obj "
                . $self->dumper_get(\%args, $self->get_debug_level()),
            $self->errmsg()
        );
        return (wantarray()) ? () : [];
    } ## end if ($self->error())

    return (wantarray()) ? @records : \@records;
} ## end sub lun_list_info

################################################################################
# A C C E S S O R   M E T H O D S
################################################################################

################################################################################
# Method Name: get_path()
# Objective:   Get qtree_path
# Details:     See POD documentation at the end of this file
################################################################################
sub get_path {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);
    if (($args{use_cache}) and defined($self->{qtree_path})) {
        return $self->{qtree_path}
    }

    $args{field} = "qtree_path";

    # the _get_field() will handle error
    return $self->_get_field(%args);
} ## end sub get_path

################################################################################
# Method Name: get_name()
# Objective:   Get qtree name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self, %args) = @_;

    $args{field} = "qtree";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
} ## end sub get_name

################################################################################
# Method Name: get_parent_obj()
# Objective:   Get qtree parent object
# Details:     See POD documentation at the end of this file
################################################################################
sub get_parent_obj {
    my ($self, %args) = @_;

    $args{field} = "parent_obj";

    # let's get from cache for name field
    $args{use_cache} = $SanUtil::TRUE;

    # the _get_field() will handle error
    return $self->_get_field(%args);
} ## end sub get_parent_obj

################################################################################
# Method Name: get_volume()
# Objective:   Get volume this qtree created on
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volume {
    my ($self, %args) = @_;

    $args{field} = "volume";

    # the _get_field() will handle error
    return $self->_get_field(%args);
} ## end sub get_volume

################################################################################
# Method Name: get_vserver()
# Objective:   Get qtree vserver
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vserver {
    my ($self, %args) = @_;

    $args{field} = "vserver";

    # the _get_field() will handle error
    return $self->_get_field(%args);
} ## end sub get_vserver

################################################################################
# Method Name: get_security()
# Objective:   Get qtree security style
# Details:     See POD documentation at the end of this file
################################################################################
sub get_security {
    my ($self, %args) = @_;

    $args{field} = "security_style";

    # the _get_field() will handle error
    return $self->_get_field(%args);
} ## end sub get_security

################################################################################
# Method Name: get_oplocks_mode()
# Objective:   Get qtree oplocks mode
# Details:     See POD documentation at the end of this file
################################################################################
sub get_oplocks_mode {
    my ($self, %args) = @_;

    $args{field} = "oplock_mode";

    # the _get_field() will handle error
    return $self->_get_field(%args);
} ## end sub get_oplocks_mode

################################################################################
# Method Name: get_unix_perm()
# Objective:   Get qtree unix permissions
# Details:     See POD documentation at the end of this file
################################################################################
sub get_unix_perm {
    my ($self, %args) = @_;

    $args{field} = "unix_permissions";

    # the _get_field() will handle error
    return $self->_get_field(%args);
} ## end sub get_unix_perm

################################################################################
# Method Name: get_id()
# Objective:   Get qtree id
# Details:     See POD documentation at the end of this file
################################################################################
sub get_id {
    my ($self, %args) = @_;

    $args{field} = "qtree_id";

    # the _get_field() will handle error
    return $self->_get_field(%args);
} ## end sub get_id

################################################################################
# Method Name: get_status()
# Objective:   Get qtree status
# Details:     See POD documentation at the end of this file
################################################################################
sub get_status {
    my ($self, %args) = @_;

    $args{field} = "status";

    # the _get_field() will handle error
    return $self->_get_field(%args);
} ## end sub get_status

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Method Name: _get_param_defaults
# Objective: Gets default qtree parameters
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_param_defaults {
    my ($self, %args) = @_;

    my %default_params
        = (%args, vserver => $self->{vserver});

    # if path is not provided, we need volume and qtree params
    unless (exists $args{'qtree-path'}) {
        %default_params = (
            %default_params,
            qtree  => $self->{qtree},
            volume => $self->{volume}
        );
    } ## end unless (exists $args{'qtree-path'...

    return %default_params;
} ## end sub _get_param_defaults

################################################################################
# Method Name: _get_field
# Objective: Get attribute from the object cache or the controller for Qtree
################################################################################
sub _get_field {
    my ($self, %args) = @_;

    my $use_cache = $SanUtil::FALSE;
    my $command
        = (defined $args{-command}) ? $args{-command} : "volume_qtree_show";

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # verify mandatory cmd argument
    unless (defined $args{field} and length $args{field}) {
        $self->_set_err("mandatory parameter \"-field\" not specified ");
        return "";
    }

    # set use_cache flag if exists
    if (exists $args{use_cache}) {
        $use_cache = delete $args{use_cache};
    }

    # get attribute from object cache or on the controller
    if ($use_cache == $SanUtil::TRUE) {

        # the get() will handle error
        return $self->get("$args{field}");
    } else {
        # get_param_defaults will put the qtree and volume into the args
        %args = (
            %args,
            $self->_get_param_defaults()
        );

        # For ZAPI, we can't just use the standard get_field_from_controller
        # method because the list of qtrees comes out as volumes and qtrees.
        # So we will get the output and find the data for this Qtree
        if (($command =~ /volume_qtree_show/) and
            ($self->get_cmd_interface() =~ /zapi/i)) {

            $args{"path"} =  $self->create_common_path(volume => $args{volume},
                                                       qtree  => $args{qtree});

            if ($self->error()) {
                # error is already set, bubble up
                return "";
            }
            my $current_vol   = delete $args{volume} if ($args{volume});
            my $current_qtree = delete $args{qtree}  if ($args{qtree});
            delete $args{name} if ($args{name});

            # Ontap looks for "fields" param (with plural)
            $args{fields} = delete $args{field};
            my (@output) = $self->invoke_cmd_obj(-cmd => $command, %args);
            if ($self->error()) {
                # error is already set, bubble up
                return "";
            }

            # The command for zapi returns volumes and qtrees. The generic
            # get_field_from_controller() call won't work for volume qtree
            # show (it always grabs the first entry in the list).
            foreach my $entry (@output) {
                next if $entry->{"qtree-id"} eq "0"; # skip entries that are vols

                # If the qtree name and volume match, return the field
                if ($entry->{qtree} eq "$current_qtree" and
                    $entry->{volume} eq "$current_vol") {

                    # If the user is trying to get the path to the qtree, they will
                    # fail for ZAPI because the ZAPI output doesn't contain the
                    # path.  Since we have already created the path from the
                    # volume and qtree name, just return the path
                    if ($args{fields} eq "qtree_path" or
                        $args{fields} eq "qtree-path") {
                        return $args{path};
                    }
                    if (exists $entry->{$args{fields}}) {
                        return $entry->{$args{fields}};
                    } else {
                        $self->_set_err("output does not contain field: "
                                      . $args{fields});
                        return "";
                    }
                }
            } # end foreach output entry
            $self->_set_err("Qtree " . $args{path} . " not found");
            return "";
        } else {
            # CLI case is generic
            # the get_field_from_controller() will handle error
            return $self->get_field_from_controller(cmd   => $command,
                                                    field => $args{field},
                                                    %args
                                                   );
        }
    } ## end else [ if ($use_cache == $SanUtil::TRUE)
} ## end sub _get_field

################################################################################
# Method Name: _create_qtree
# Objective: create qtree
################################################################################
sub _create_qtree {
    my ($self, %args) = @_;

    $self->_clear_err();

    # append cmd for volume qtree create to argument list
    %args = (%args, cmd => "volume_qtree_create");

    # use common invoke from base class (San::CommonInterface)
    # to translate cdef into UI cmd and execute it
    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj " . $args{cmd}, $self->errmsg());
        return $SanUtil::FALSE;
    }

    # call volume refresh if enable cache is set
    if ($self->{enable_cache}) {
        $self->refresh(%args);
        if ($self->error()) {
            # set error and bubble up
            $self->_set_err("failed to refresh " . $args{qtree}, $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # set initialized flag
    $self->{initialized} = $SanUtil::TRUE;

    return $SanUtil::TRUE;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Qtree.pm

=head1 Author

Korry Nguyen (dl-ssan-automation)

=head1 Description

Provides an interface to create and manage C-mode qtrees

=head1 See Also

Qtree.pm inherits methods from B<San::CommonInterface>. See these
methods for further interface details.

=head1 Attributes

=head2 B<filer>

string = controller name

=head2 B<name>

string = name of qtree object to be created

=head2 B<debug_level>

interger = 0 - 100 (default is 0)

=head2 B<security_style>

string = type of security in the qtree (unix|ntfs|mixed)

=head2 B<oplock_mode>

string = the oplocks mode (enable|disable)

=head2 B<unix_permissions>

string = the unix permissions on the qtree

=head2 B<volume>

string = the name of the volume the qtree is in

=head2 B<qtree>

string = the name of the qtree

=head2 B<qtree_path>

string = the path of the qtree

=head2 B<qtree_id>

integer = the id of the qtree

=head2 B<status>

string = the status of the qtree

=head1 B<Public Methods>

=head2 B<new(%args)>

Create a qtree object

=over 4

=item Inputs

=item * I<-filer> => <controller name>

Controller name

=item * I<-name> => <name of qtree object>

Qtree object name to be created

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. With higher level, it will log more debug messages

=item * I<-volume> => <volume qtree is on>

Volume qtree is created on

=item * I<-qtree-path> => <path to create qtee> (either this or name and volume are required)

Path qtree is created at

=item * I<-security-style> => <unix|ntfs|mixed>

Security style of the qtree

=item * I<-oplock-mode> => <enable|disable>

Sets the oplock mode of the qtree to enable or disable

=item * I<-unix-permissions> => <unix perm>

Sets the unix permissions of the qtree

=item Outputs

Returns a qtree object

=item Examples

 # ex. 1: Create qtree directly, make sure you have a vserver and a volume

 use San::ControllerCluster::Filer10M::Vserver::Volume::Qtree;

 my $vol = "vs0_vol";
 my $vserver = "vs0";
 my $qtree = "qtree_0";

 # Create filer object
 $filer_obj = new SanFiler();
 if ($filer->error()) {
    logcomment("Failed to create filer object" . $filer_obj->errmsg());
    return $TCD::FAIL;
 }

 # Get the vol object. Assuming the vol exists on that node.
 $vol_obj = $filer_obj->get_vol(volume  => $vol,
                                vserver => $vserver);
 if ($vol_obj->error()) {
    logcomment("Failed to get vol object" . $vol_obj->errmsg());
    return $TCD::FAIL;
 }

 # Create the qtree
 my $qtree_obj = new Qtree(-filer => $filer,
                           -name => $qtree,
                           -parent_obj => $vol_obj,
                           -vserver => $vserver,
                           -volume => $vol,
                           -debug_level => 0);
 if ($qtree_obj->error()) {
    logcomment("Failed to create qtree object" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

 # ex. 2: Create qtree from a volume object
 my $qtree_obj = $volume_obj->create_qtree(...);
 if ($qtree_obj->error()) {
    logcomment("Failed to create qtree object" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<create_file(%args)>

Creates a file on a Qtree.

=over 4

=item Inputs

=item * I<-size> (required)

The size of the file to be created

=item * I<-from_file> (required)

The name of the file to be created.

=item Outputs

Returns TRUE if the file was created successfully. Returns FALSE and sets the
error flag if something failed.

=item Examples

 $qtree_obj->create_file(size => "50m",
                         from_file => $file_name);
    if ($self->error()) {
        ...
    }

=back

=head2 B<remove_file(%args)>

Removes a file on a Qtree.

=over 4

=item Inputs

=item * I<-size> (required)

The size of the file to be removed

=item * I<-from_file> (required)

The name of the file to be removed.

=item Outputs

Returns TRUE if the file was removed successfully. Returns FALSE and sets the
error flag if something failed.

=item Examples

 $qtree_obj->remove_file(from_file => $file_name);
    if ($self->error()) {
        ...
    }

=back

=head2 B<destroy(%args)>

Destroys a qtree object by invoking "qtree_destroy" CDEF.

=over 4

=item Inputs

=item * I<-force> => <1|0> (optional)

Force destroy a qtree. Same as passing -force to a qtree destroy command. Default
value is FALSE.

=item Outputs

Returns TRUE if the qtree was destroyed successfully. Returns FALSE and sets the
error flag if something failed.

=item Examples

 # Destroy a qtree
 my $qtree_obj->destroy();
 if ($qtree_obj->error()) {
    logcomment("Failed to destroy qtree object" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<modify()>

Modifies Qtree attributes. It invokes qtree_modify CDEF

=over 4

=item Inputs

=item * I<-oplock-mode> => <enable|disable> (optional)

Modifies the oplocks mode of the qtree.

=item * I<-security-style> => <unix|ntfs|mixed> (optional)

Modifies the security style of a qtree.

=item * I<-unix-permissions> => <unix perm> (optional)

Modifies the unix permissions of a qtree.

=item Outputs

Returns TRUE if the Qtree was modified successfully and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Modify state
 $qtree_obj->modify(-oplock-mode => 'disable');
 if ($qtree_obj->error()) {
    logcomment("Failed to modify qtree object" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_list_info()>

Get qtree information. This method invokes a basic qtree_show cdef on a filer.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns a reference to a hash containing the output of "qtree show" command. If
it fails, it will return a null and set the error flag.

=item Examples

 $qtree_info = $qtree_obj->get_list_info();
 if ($qtree_obj->error()) {
    logcomment("Failed to get qtree information" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_lun_maxsize()>

Gets the maxsize lun for this qtree.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the maxsizes of the lun this qtree can hold in a hash.
If an error occurs it return an empty hash and sets an error
flag.

=item Examples

 # Get maxsize for a lun in the qtree
 my $maxsize = $qtree_obj->get_lun_maxsize();
 if ($qtree_obj->error()) {
    logcomment("Failed to get maxsize" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<rename()>

Rename a qtree

=over 4

=item Inputs

=item * <-newname> => <newname>.

The new name of the qtree.

=item Outputs

Returns TRUE if the Qtree was renamed successfully and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 # Rename qtree
 $qtree_obj->rename(name=> $new_qtree_name);
 if ($qtree_obj->error()) {
    logcomment("Failed to rename qtree " . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<create_lun_obj()>

create a lun on a qtree

=over 4

=item Inputs

=item * I<-path> => <path>.

The path of the lun.

=item * <-size> => <size>.

The size of the lun.

=item * I<-file-path> => <text (size 1..255)>.

The path of the file to create the lun from.

=item * <-ostype> => <vmware|hyper_v|windows_2008|windows_gpt|windows|linux|xen|solaris|solaris_efi|hpux|aix|netware|openvms>.

The OS type of the lun.

=item * I<-option> => <option>.

The space reservation option of the lun.

=item Outputs

Returns the lun_obj if the lun was creates successfully and an empty hash otherwise. This
method will set an error flag if the command fails.

=item Examples

 my $lun_obj = $qtree_obj->create_lun_obj(
               parent_obj => $qtree_obj,
               name       => $lun_name,
               ostype     => "windows",
               size       => "50m",
           );
   if ($qtree_obj->error()) {
       $Test->nlog("FAIL: qtree_lun_create_get() Failed to create lun"
               . $qtree_obj->errmsg());
       $cleanup_status = _cleanup();
       return $TCD::FAIL;
   }

=back

=head2 B<get_lun>

get a lun from the qtree

=over 4

=item Inputs

=item * <lun|name> => <name of lun>.

The name of the lun to get.

=item Outputs

Returns the lun_obj if the lun was retrieved successfully and an empty hash otherwise. This
method will set an error flag if the command fails.

=item Examples

 my $lun_get_obj = $qtree_obj->get_lun(lun => $lun_name);
 if ($qtree_obj->error()) {
    $Test->nlog("FAIL: qtree_lun_create_get() Failed to get lun"
                . $qtree_obj->errmsg());
    $cleanup_status = _cleanup();
    return $TCD::FAIL;
 }

=back

=head2 B<get_luns>

get all the luns from the qtree

=over 4

=item Inputs

=item * None.

=item Outputs

Returns an array of lun_objs if the luns were gathered successfully and an empty array otherwise. This
method will set an error flag if the command fails.

=item Examples

 @lun_objs = $self->get_luns();
 if ($self->error()) {
    $self->_set_err("failed to get lun objects ", $self->errmsg());
    return {};
 }

=back

=head2 B<lun_exist>

check to see if a lun exists in the qtree

=over 4

=item Inputs

=item * <name> => <name of lun>.

The name of the lun to get.

=item Outputs

Returns the TRUE if the lun exists and FALSE otherwise. This
method will set an error flag if the command fails.

=item Examples

 if ($qtree_obj->lun_exist(-name => $lun)) {
    $Test->nlog("   - Retrieving lun $lun");
    $lun_obj = $qtree_obj->get_lun(name => $lun);
    if ($qtree_obj->error()) {
        $Test->nlog("FAIL: Failed to get lun object " . $qtree_obj->errmsg());
        return $TCD::FAIL;
 }

=back

=head2 B<lun_list_info()>

Get lun information. This method invokes a basic lun_show cdef on a filer with qtree parameters passed in.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns a reference to a hash containing the output of "lun show" command. If
it fails, it will return a null and set the error flag.

=item Examples

 # Get the info on the lun
 $lun_info = $qtree_obj->lun_list_info();
 if ($qtree_obj->error()) {
    logcomment("Failed to get lun information" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_name()>

Gets the name of this qtree.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the name of this qtree.

=item Examples

 # Get qtree name
 my $qtree = $qtree_obj->get_name();
 if ($qtree_obj->error()) {
    logcomment("Failed to get name" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_parent_obj()>

Gets the parent object of this qtree (A volume object).

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the parent object of a qtree.

=item Examples

 # Get parent object
 my $vol_obj = $qtree_obj->get_parent_obj();
 if ($qtree_obj->error()) {
    logcomment("Failed to get parent object" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_vserver()>

Gets the vserver owning this Qtree.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the vserver attribute

=item Examples

 # Get vserver
 my $vserver = $qtree_obj->get_vserver();
 if ($qtree_obj->error()) {
    logcomment("Failed to get vserver" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_security()>

Gets the security style of this Qtree.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the security style attribute

=item Examples

 # Get security style
 my $security = $qtree_obj->get_security();
 if ($qtree_obj->error()) {
    logcomment("Failed to get security style" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_oplock_mode()>

Gets the oplocks mode of this Qtree.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the oplocks mode attribute

=item Examples

 # Get oplocks mode
 my $oplock_mode = $qtree_obj->get_oplock_mode();
 if ($qtree_obj->error()) {
    logcomment("Failed to get oplock_mode" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_unix_perm()>

Gets the unix permissions of this Qtree.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the unix permissions attribute

=item Examples

 # Get unix permissions
 my $unix_permissions = $qtree_obj->get_unix_perm();
 if ($qtree_obj->error()) {
    logcomment("Failed to get unix_perm" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_id()>

Gets the id of this Qtree.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the id attribute

=item Examples

 # Get id
 my $id = $qtree_obj->get_id();
 if ($qtree_obj->error()) {
    logcomment("Failed to get id" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back

=head2 B<get_status()>

Gets the status of this Qtree.

=over 4

=item Inputs

=item * None.

=item Outputs

Returns the status attribute

=item Examples

 # Get status
 my $status = $qtree_obj->get_status();
 if ($qtree_obj->error()) {
    logcomment("Failed to get status" . $qtree_obj->errmsg());
    return $TCD::FAIL;
 }

=back
