# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary provides 7mode compatible methods for Volume class
##
## @author korry@netapp.com, dl-ssan-automation
## @status Shared

package San::ControllerCluster::Filer10M::Vserver::VolumeCompat;

# compiler directives
use strict;
use lib();

# imported modules;
use Params::Validate qw( SCALAR ARRAYREF HASHREF OBJECT );

###############################################################################
# Method Name: create_lun_from_file
# Objective:  Create lun from a file created during the method.
# Details:    See POD documentation at the end of this file.
###############################################################################
sub create_lun_from_file {
    my ($self, %args) = @_;
    my $lunObj;
    my %create_file_args;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # check for required arguments
    unless (defined $args{size} and length $args{size}){
        $self->_set_err("Mandatory parameter \"size\" not specified ");
        return {};
    }

    # Get the arguments for file create and then create the file.
    $create_file_args{size} = delete $args{size};

    unless (defined $args{from_file} and length $args{from_file}){
        $self->_set_err("Mandatory parameter \"from_file\" not specified ");
        return {};
    }
    my $file_name = $args{from_file};
    $create_file_args{from_file} = delete $args{from_file};

    $create_file_args{timeout} = delete $args{timeout} if exists $args{timeout};

    $self->create_file(%create_file_args);
    if ($self->error()) {
        $self->_set_err("Failed to create file: " .
                        $self->errmsg());
        return {};
    }

    # Now that the file has been created, create a lun from the file.
    $lunObj = $self->create_lun_obj(%args,
                                    'file-path' => "/vol/".$self->get_name().
                                                   "/".$file_name);
    if ($self->error()) {
        $self->_set_err("Failed to execute create lun command: " .
                        $self->errmsg());
        return {};
    }

    # This is cut because currently the file cannot be removed
    # if a lun has been created from the file.  There is a burt open
    # for it. BURT# 428434
=cut
    $self->remove_file(%args);
    if ($self->error()) {
        $self->_set_err("Failed to remove file: " .
                        $self->errmsg());
        return {};
    }
=cut

    #
    # Add the lun object created to our list of cached luns
    #
    push @{$self->{luns}}, $lunObj;

    return $lunObj;
} # end create_lun_from_file()

################################################################################
# Method Name: get_freespace()
# Objective:   get the freespace on the volume. This method map one-to-one to 7-mode
# Details:     See POD documentation at the end of this file
################################################################################
sub get_freespace {
    my ($self, %args) = @_;

    return ($self->get_available());
}

################################################################################
# Method Name: get_space()
# Objective:   get the space on the volume. This method map one-to-one to 7-mode
# Details:     See POD documentation at the end of this file
################################################################################
sub get_space {
    my ($self, %args) = @_;

    # remove the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    $args{field} = $args{type};

    # the _get_field() will handle error
    return $self->_get_field(%args);
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

VolumeCompat.pm

=head1 Author

Ann Harrold (dl-ssan-automation)

=head1 Description

VolumeCompat provides 7mode compatible and helper methods for Volume. This package maintains separately from Volume for maintainability purpose while preventing convoluting Volume object model.

=head1 See Also

VolumeCompat package is used as part of the base class by B<San::ControllerCluster::Filer10M::Vserver::Volume>. See these modules for further interface detail.

B<San::ControllerCluster::Filer10M.pm, San::CommonInterface.pm>

=head1 B<Public Methods>

=head2 B<create_lun_from_file()>

Creates a file from the size adn file name passed in.  Creates a Lun from the file,
then removes the file.

=over 4

=item Inputs

=item * <size>

Size of the file to make

=item * <from_file>

Name of the file to make.

=item * <parent_obj>

The parent object of the lun

=item * <name>

Name of the lun to make.

=item * <ostype>

The operating system of the lun to make.

=item Outputs

Returns a lun object representing the newly created lun

=item Examples

 my $lun_obj = $vol_obj->create_lun_from_file(parent_obj => $vol_obj,
                                              name       => $lun_name,
                                              ostype     => "windows",
                                              size       => "50m",
                                              from_file  => $file_name);
 if ($vol_obj->error()) {
    ...
 }

=back

=head2 B<get_freespace()>

Get available volume size in bytes.  (See get_available in Volume.pm)

=over 4

=item Inputs

None

=item Outputs

Returns the number of bytes available. The unit will not be suffixed at the end.
The output will just be a number

=item Examples

 # ex. 1: Get available size of the volume
 $vol_freespace = $vol_obj->get_freespace();
 if ($vol_obj->error()) {
    ...
 }

=head2 B<get_space()>

Get volume space attributes in bytes.

=over 4

=item Inputs

=item * I<-type> (total, available, or used)

The space attribute type you wish to retrieve.  This can be one of total
available, or used.

=item Outputs

Returns the space attribute required. The unit will not be suffixed at the end.
The output will just be a number

=item Examples

 # ex. 1: Get the used space on the volume
 my $used_space = $vol_obj->get_space(-type => "used");
 if ($vol_obj->error()) {
    ...
 }

 =back
