# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary provides 7mode compatible methods for Volume class
##
## @author korry@netapp.com, dl-ssan-automation
## @status Shared

package San::ControllerCluster::Filer10M::Vserver::VolumeHelper;

# compiler directives
use strict;
use lib();
use Time::HiRes qw( gettimeofday );

# imported modules;
use Params::Validate qw( SCALAR ARRAYREF HASHREF OBJECT );

# inherit base class
use base qw( San::CommonInterface);

###############################################################################
# Method Name: wait_for_clone_completion
# Objective:  Wait for a clone completion.
# Details:    See POD documentation at the end of this file.
###############################################################################
sub wait_for_clone_completion {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);
    eval {

        # validate input parameters and their type.
        # list - array that contains names of object to be created
         (%args) =
          Module::Common::validate_mixed(@input_params,
                                         [],
                                         {path => {type => SCALAR, required => 1},
                                          timeout => {type => SCALAR,
                                                      optional => 1,
                                                      default => 3600,
                                                      regex => qr/\d+/ },
                                          interval => {type => SCALAR,
                                                       optional => 1,
                                                       default => 60,
                                                       regex => qr/\d+/ }});
    };

    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # Get the node obj that is the parent of this volume
    # First get_parent gets the vserver. The second gets
    # the node from the vserver
    my $parent_vserver = $self->get_parent_obj();
    if ($self->error()) {
        $self->_set_err("failed to get parent vserver object ", $self->errmsg());
        return $SanUtil::FALSE;
    }
    my $parent_node = $parent_vserver->get_parent_obj();
    if ($parent_vserver->error()) {
        $self->_set_err("failed to get parent node object ", $parent_vserver->errmsg());
        return $SanUtil::FALSE;
    }

    # Get the aggregate object that this volume belongs to.
    # Need the aggregate to figure out which node owns the volume
    my $aggr_obj = $parent_node->get_aggregate_obj(name => $self->get_aggregate());
    if ($parent_node->error()) {
        $self->_set_err("failed to get aggregate object ", $parent_node->errmsg());
        return $SanUtil::FALSE;
    }

    # Finally, get the node that owns the volume from the aggr
    #### When striped multiple nodes will get returned.  Code may no longer work,
    #### Need to eventually find a cmode context to mkfile and rm file
    my $node = $aggr_obj->get_nodes();
    if ($aggr_obj->error()) {
        $self->_set_err("failed to get nodes ", $aggr_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # set start and current times
    my $start_time   = int(gettimeofday);
    my $current_time = $start_time;

    # check for clone completion
    while (($current_time - $start_time) < $args{timeout}) {
        $current_time = int(gettimeofday);

        # Need to check if clone has finished
        # initialize cmd parameters
        my %cmd1 = ( cmd => "system_node_run",
                    node => $node,
                    command => "clone list  -p ".$args{path},
                    donot_append_vserver => 1,
                    raw_data => 1);

        my (@output) = $self->SUPER::invoke_cmd_obj(%cmd1);
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj " .
                            $self->dumper_get(\%cmd1, $self->get_debug_level()),
                            $self->errmsg());
            return $SanUtil::FALSE;
        }

        $self->sanlog(-msg => "The output from the system node run command is: ". $output[0],
                  -msg_level => 3);

        # clone is complete is not entries found
        return $SanUtil::TRUE if ($output[0] =~ /No entries found!/i);

        # wait for the next interval to check clone state
        $self->sanlog(-msg => $args{name} . " is " . $aggr_obj->get_state() .
                      ". Raidstatus is " . $aggr_obj->get_raidstatus() .
                      ". Waiting \"$args{interval}\" seconds",
                      -msg_level => 3);
        &SanUtil::nate_sleep($args{interval});
    }

    # clone has not finish and timeout expired
    $self->_set_err("unable to finish cloning before time expired");
    return $SanUtil::FALSE;
} ## end sub wait_for_clone_completion

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

VolumeCompat.pm

=head1 Author

Philip McClellan (dl-ssan-automation)

=head1 Description

VolumeHelper provides 7mode compatible and helper methods for Volume. This package maintains separately from Volume for maintainability purpose while preventing convoluting Volume object model.

=head1 See Also

VolumeHelper package is used as part of the base class by B<San::ControllerCluster::Filer10M::Vserver::Volume>. See these modules for further interface detail.

B<San::ControllerCluster::Filer10M.pm, San::CommonInterface.pm>

=head1 B<Public Methods>

=head2 B<wait_for_clone_completion()>

Waits for a clone operation to complete based on path passed in.

=over 4

=item Inputs

=item * <path>

Path of the Clone

=item * <timeout>

Timeout for waiting.

=item * <interval>

Interval between tries.

=item Outputs

Returns a whether the clone has complete or if the wait process timed out

=item Examples

 if($sync) {
        $self->wait_for_clone_completion(%sync_args);
    }
 if ($self->error()){
        $self->_set_err("clone failed to complete." .
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

=back

