# Copyright (c) 2008-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provides helper methods for the Vserver object
##
## @author korry@netapp.com, dl-ssan-automation
## @status Shared

package San::ControllerCluster::Filer10M::VserverHelper;

# compiler directives
use strict;
use lib();

# imported modules
use Params::Validate qw( validate SCALAR ARRAYREF );

# inherit base class
use base qw( San::CommonInterface );

################################################################################
# Method Name: get_lifs()
# Objective:       get lif objects either from cach or discover on system
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lifs {
    my ($self, %args) = @_;

    my (@lif_objs, @matched_lifs, @tmp_lif_objs);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # discover and get lif objects
    @tmp_lif_objs = $self->get_lif_objs();
    if ($self->error()) {
        $self->_set_err("failed to get lif objects", $self->errmsg());
        return (wantarray()) ? (): [];
    }

    # Make sure all lif objects exist and have not been marked as deleted
    foreach my $lif_obj (@tmp_lif_objs) {
        unless ($lif_obj->get_name() =~ /_DELETE_/i) {
            push @lif_objs, $lif_obj;
        }
    }

    # if there are no lifs set the error...
    unless (scalar(@lif_objs)) {
        $self->sanlog(-msg => "no lif in vserver " . $self->get_name(),
                      -msg_level => 0);
        return (wantarray()) ? (): [];
    }

    if (defined $args{type}) {
        # get the right lif type
        foreach my $lif_obj (@lif_objs) {
            if ($lif_obj->get_data_protocol() =~ /$args{type}/i) {
                push @matched_lifs, $lif_obj;
            }
        }

        # no match lif found
        unless (scalar(@matched_lifs) >= 1) {
            $self->sanlog(-msg => "no matched lif in vserver " . $self->get_name() . " for type " . $args{type},
                          -msg_level => 0);
            return (wantarray()) ? (): [];
        }
        return (wantarray()) ? @matched_lifs: \@matched_lifs;
    }

    return (wantarray()) ? @lif_objs: \@lif_objs;
} ## end sub get_lifs

###############################################################################
# Method Name: create_tpgroup_obj
# Objective:   Create tpgroup object
# Details:     See POD documentation at the end of this file.
###############################################################################
sub create_tpgroup_obj {
    my ($self, %args) = @_;

    my ($iscsi_obj, $tpgroup_obj);

    $self->_clear_err();

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # append aggregate name parameter to argument list
    %args = (%args, $self->_get_param_defaults());

    # get iscsi object. it will create if it does not exist
    $iscsi_obj = $self->get_iscsi_obj();
    if ($self->error()) {
        $self->_set_err("failed to get iscsi object ", $self->errmsg());
        return {};
    }

    # instantiate tpgroup object using iscsi_obj's create_tpgroup_obj method
    $tpgroup_obj = $iscsi_obj->create_tpgroup_obj(%args);
    if ($iscsi_obj->error()) {
        $self->_set_err("failed to instantiate tpgroup object using iscsi obj", $iscsi_obj->errmsg());
        return {};
    }

    return $tpgroup_obj;
} ## end sub create_tpgroup_obj

###############################################################################
# Method Name: load_fcp_targets()
# Objective:   Load important information about vserver fcp targets
# Details:     See POD documentation at the end of this file.
###############################################################################
sub load_fcp_targets {
    my ($self, %args) = @_;

    my ($fcp_obj, @fcp_lif_objs, $lif_obj, $host_obj);
    my ($nodename, $state, $filer_name);
    my ($is_mgmt_port, $interface);
    my (@lif_data);

    # korry - the right thing would be:
    # 1) get current existing lif data in vserver
    # 2) is the exist lifs adequate w/ the request to load for use?
    # 3) if not, get fcp adapters that are online
    # 4) create additional lifs

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # if vserver has fcp lif data, it will take precedent
    if (scalar(@{$self->{fcp_lif_data}})) {
        @lif_data = @{$self->{fcp_lif_data}};
    } elsif ($args{use_nate_params} == $SanUtil::TRUE) {
        # use parameters from nate host file
        # VS0.FCP.PREFIX=vs0_fcp_lif
        # VS0.FCP.NUM_LIFS=2
        # VS0.FCP.PORTS=5a:5b
        # VS1.FCP.PREFIX=vs1_fcp_lif
        # VS1.FCP.NUM_LIFS=2
        # VS1.FCP.PORTS=5a:5b

        # the dot notation does not work so we are using "_" instead
        # VS0_FCP_PREFIX=vs0_fcp_lif
        # VS0_FCP_NUM_LIFS=2
        # VS0_FCP_PORTS=5a:5b
        # VS1_FCP_PREFIX=vs1_fcp_lif
        # VS1_FCP_NUM_LIFS=2
        # VS1_FCP_PORTS=5a:5b
        @lif_data = $self->_get_nate_params(-type => "fcp");
        if ($self->error()) {
            $self->_set_err("failed to get nate params for fcp lif ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # get fcp object. it will create if it does not exist
    $fcp_obj = $self->get_fcp_obj();
    if ($self->error()) {
        $self->_set_err("failed to get fcp object ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # get/create lif objects
    foreach my $lif_info (@lif_data) {
        # create lif if it does not exist
        if ($self->lif_exist(-name => $lif_info->{lif})) {
            $lif_obj = $self->get_lif_obj(-name => $lif_info->{lif});
            if ($self->error()) {
                $self->_set_err("failed to get lif object ", $self->errmsg());
                return $SanUtil::FALSE;
            }
        } else {
            $lif_obj = $self->create_lif_obj(-name => $lif_info->{lif},
                                             -protocol => "fcp",
                                             -interface => $lif_info->{port});
            if ($self->error()) {
                $self->_set_err("failed to create lif object ", $self->errmsg());
                return $SanUtil::FALSE;
            }
        }
    }

    # get fcp nodename if fcp object exists
    if (ref($fcp_obj) =~ /FCP/i) {
        $nodename = $fcp_obj->get_nodename();
        if ($fcp_obj->error()) {
            $self->_set_err("failed to get fcp nodename", $fcp_obj->errmsg());
            return $SanUtil::FALSE;
        }
    } else {

        # since we don't have iscsi object, set nodename to empty string
        $nodename = "";
    }

    # if we don't have the list of lifs, try to get all fcp lifs
    unless (scalar(@fcp_lif_objs)) {
        @fcp_lif_objs = $self->get_lifs(-type => "fcp");
        if ($self->error()) {
            $self->_set_err("failed to get fcp lifs", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # create and load target for each lif
    foreach $lif_obj (@fcp_lif_objs) {
        $is_mgmt_port = 0;
        $state = $lif_obj->get_status_oper();
        if ($lif_obj->error()) {
            $self->_set_err("failed to get lif status operation", $lif_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Create fcp target object (if interface is enabled for fcp)
        my $targObj =
            San::SanClient::Fcptarget->_new(
                                        -parent_obj       => $self,
                                        -fcp_obj      => $fcp_obj,
                                        -name         => $lif_obj->get_name(),
                                        -state        => $state,
                                        -is_mgmt_port => $is_mgmt_port,
                                        -nodename     => $nodename);
        if ($targObj->error()) {
             $self->_set_err($targObj->errmsg());
             return $SanUtil::FALSE;
        }

        push(@{$self->{fcp_targets}}, $targObj);
    }
    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: load_iscsi_targets()
# Objective:   Return important information about filer iscsi targets
# Details:     See POD documentation at the end of this file.
###############################################################################
sub load_iscsi_targets {
    my ($self, %args) = @_;

    my ($iscsi_obj, @iscsi_lif_objs, $lif_obj, $host_obj);
    my ($nodename, $state, $ipv6, $ipv4, $tpgn, $tpgt);
    my ($filer_name, $interface, $ip, $netmask);
    my ($ipv4_netmask, $ipv6_netmask_length, $mac, $is_mgmt_port);
    my (@lif_data);

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # if vserver has iscsi lif data, it will take precedent
    if (scalar(@{$self->{iscsi_lif_data}})) {
        @lif_data = @{$self->{iscsi_lif_data}};
    } elsif ($args{use_nate_params} == $SanUtil::TRUE) {
        # use parameters from nate host file in order to porvide 7mode compatibility
        # VS0.ISCSI.PREFIX=vs0_iscsi_lif
        # VS0.ISCSI.NUM_LIFS=2
        # VS0.ISCSI.PORTS=e0b:e0c
        # VS0.ISCSI.IP_RANGE=9.0.120.[7-8]
        # VS0.ISCSI.NETMASK=255.252.0.0
        # VS1.ISCSI.PREFIX=vs1_iscsi_lif
        # VS1.ISCSI.NUM_LIFS=2
        # VS1.ISCSI.PORTS=e0b:e0c
        # VS1.ISCSI.IP_RANGE=9.0.121.[7-8]
        # VS1.ISCSI.NETMASK=255.252.0.0

        # the dot notation does not work so we are using "_" instead
        # VS0_ISCSI_PREFIX=vs0_iscsi_lif
        # VS0_ISCSI_NUM_LIFS=2
        # VS0_ISCSI_PORTS=e0b:e0c
        # VS0_ISCSI_IP_RANGE=9.0.120.[7-8]
        # VS0_ISCSI_NETMASK=255.252.0.0
        @lif_data = $self->_get_nate_params(-type => "iscsi");
        if ($self->error()) {
            $self->_set_err("failed to get nate params for iscsi lif ", $self->errmsg());
            return $SanUtil::FALSE;
        }

    }

    # get iscsi object. it will create if it does not exist
    $iscsi_obj = $self->get_iscsi_obj();
    if ($self->error()) {
        $self->_set_err("failed to get iscsi object ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # get/create lif objects
    foreach my $lif_info (@lif_data) {
        # create lif if it does not exist
        if ($self->lif_exist(-name => $lif_info->{lif})) {
            $lif_obj = $self->get_lif_obj(-name => $lif_info->{lif});
            if ($self->error()) {
                $self->_set_err("failed to get lif object ", $self->errmsg());
                return $SanUtil::FALSE;
            }
        } else {
            $lif_obj = $self->create_lif_obj(-name => $lif_info->{lif},
                                             -protocol => "iscsi",
                                             -interface => $lif_info->{port},
                                             -ip => $lif_info->{address},
                                             -netmask => $lif_info->{netmask});
            if ($self->error()) {
                $self->_set_err("failed to create lif object ", $self->errmsg());
                return $SanUtil::FALSE;
            }
        }
        push (@iscsi_lif_objs, $lif_obj);
    }

    # get iscsi nodename if iscsi object exists
    if (ref($iscsi_obj) =~ /iSCSI/i) {
        $nodename = $iscsi_obj->get_nodename();
        if ($iscsi_obj->error()) {
            $self->_set_err("failed to get iscsi nodename", $iscsi_obj->errmsg());
            return $SanUtil::FALSE;
        }
    } else {

        # since we don't have iscsi object, set nodename to empty string
        $nodename = "";
    }

    # if we don't have the list of lifs, try to discover existing iscsi lifs
    unless (scalar(@iscsi_lif_objs)) {
        @iscsi_lif_objs = $self->get_lifs(-type => "iscsi");
        if ($self->error()) {
            $self->_set_err("failed to get iscsi lifs", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # create and load target for each lif
    foreach $lif_obj (@iscsi_lif_objs) {
        $ipv4 = $ipv6 = $tpgt = $tpgn = "not_set";
        $ipv4_netmask = $ipv6_netmask_length = $mac = "not_set";
        $is_mgmt_port = 0;
        my $address = $lif_obj->get_address();

        if ($lif_obj->error()) {
                $self->_set_err("failed to get lif address", $lif_obj->errmsg());
                return $SanUtil::FALSE;
        }

        if ($address =~ /:/){
            $ipv6 = $address;

            $ipv6_netmask_length = $lif_obj->get_netmask_length();
            if ($lif_obj->error()) {
                $self->_set_err("failed to get lif netmask", $lif_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {
            $ipv4 = $address;

            $ipv4_netmask = $lif_obj->get_netmask();
            if ($lif_obj->error()) {
                $self->_set_err("failed to get lif netmask", $lif_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }

        $tpgn = $lif_obj->get_name();
        if ($lif_obj->error()) {
            $self->_set_err("failed to get lif name", $lif_obj->errmsg());
            return $SanUtil::FALSE;
        }
        $tpgt = $lif_obj->get_tpgroup();
        if ($lif_obj->error()) {
            $self->_set_err("failed to get lif tpgroup", $lif_obj->errmsg());
            return $SanUtil::FALSE;
        }
        $state = $lif_obj->get_status_oper();
        if ($lif_obj->error()) {
            $self->_set_err("failed to get lif status operation", $lif_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Create iscsi target object (if interface is enabled for iSCSI)
        my $targObj =
            San::SanClient::Iscsitarget->_new(
                                        -parent_obj   => $self,
                                        -iscsi_obj    => $iscsi_obj,
                                        -name         => $lif_obj->get_name(),
                                        -state        => $state,
                                        -ip           => $address,
                                        -ipv4         => $ipv4,
                                        -ipv4_netmask => $ipv4_netmask,
                                        -ipv6         => $ipv6,
                                        -ipv6_netmask_length => $ipv6_netmask_length,
                                        -tpgt         => $tpgt,
                                        -tpgn         => $tpgn,
                                        -mac          => $mac,
                                        -is_mgmt_port => $is_mgmt_port,
                                        -nodename     => $nodename);
        if ($targObj->error()) {
             $self->_set_err($targObj->errmsg());
             return $SanUtil::FALSE;
        }

        push(@{$self->{iscsi_targets}}, $targObj);
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: get_routing_groups
# Objective: Gets all the routing groups on a vserver
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_routing_groups {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_routing_groups_show",
             cmd_interface => "cli",
             vserver => $self->get_name());

    # use common invoke to translate cdef into UI cmd and execute it
    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? (): [];
    }

    # if there are no entries, return an empty list
    if ( scalar(@output) == 0 ) {
        return (wantarray()) ? (): [];
    }

    # Get the routing group field. Invoke returns a hash: r_name, server, lif.
    my @routing_groups = ();
    foreach my $routing_group (@output) {
        if (defined ($routing_group->{routing_group})) {
            push (@routing_groups, $routing_group->{routing_group});
        }
    }

    return (wantarray()) ? @routing_groups: \@routing_groups;
} ## end sub get_routing_groups

###############################################################################
# Method Name: get_routes
# Objective: Gets all the destination of routes on a routing-group
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_routes {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "network_routing_groups_route_show",
             vserver => $self->get_name());

    my (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke cmd obj ",
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? (): [];
    }

    # if there are no entries, return an empty list
    if ($output[0] =~ /There are no entries/i) {
        return (wantarray()) ? (): [];
    }

    # Get the destination field.
    my @destinations = ();
    foreach my $route (@output) {
        if (defined ($route->{destination})) {
            push (@destinations, $route->{destination});
        }
    }

    return (wantarray()) ? @destinations: \@destinations;
} ## end sub get_routes

################################################################################
# Method Name: delete_routing_group
# Objective: Deletes a routing group
################################################################################
sub delete_routing_group {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify if mandatory parameter routing-group exists
    unless (defined $args{'routing-group'} and length $args{'routing-group'}) {
        $self->_set_err("mandatory parameter \"-routing-group\" not specified ");
        return $SanUtil::FALSE;
    }

    # Get all route destinations for this routing group and delete them
    my @destinations = $self->get_routes('routing-group' => $args{'routing-group'});
    if ($self->error()) {
        $self->_set_err("Unable to get route destinations", $self->errmsg());
        return $SanUtil::FALSE;
    }

    foreach my $destination (@destinations) {
        $self->delete_route('routing-group' => $args{'routing-group'},
                            destination => $destination);
        if ($self->error()) {
            $self->_set_err("failed to delete route", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    %args = (%args,
             cmd => "network_routing_groups_delete",
             cmd_interface => "cli",
             vserver => $self->get_name());

    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke cmd obj ",
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: delete_route
# Objective: Deletes a route
################################################################################
sub delete_route {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # verify if mandatory parameter destination exists
    unless (defined $args{destination} and length $args{destination}) {
        $self->_set_err("mandatory parameter \"-route\" not specified ");
        return $SanUtil::FALSE;
    }

    # verify if mandatory parameter routing-group exists
    unless (defined $args{'routing-group'} and length $args{'routing-group'}) {
        $self->_set_err("mandatory parameter \"-routing-group\" not specified ");
        return $SanUtil::FALSE;
    }

    # Initialize args and call invoke_cmd_obj
    %args = (%args,
             cmd => "network_routing_groups_route_delete",
             vserver => $self->get_name(),
             'routing-group' => $args{'routing-group'});

    $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke cmd obj ",
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

##############################################################################
#  P R I V A T E
##############################################################################

###############################################################################
# Method Name: _get_nate_params()
# Objective:   get nate parameters for either iscsi or fcp targets
# Details:     See POD documentation at the end of this file.
###############################################################################
sub _get_nate_params {
    my ($self, %args) = @_;

    my ($prefix_name, $num_lifs, $ports, $ip_range, $netmask);
    my ($filer_name, $host_obj);
    my (@lif_data, @ips, @ports) = ();

    # strip the dash '-' from input argument list
    %args = &SanUtil::strip_dash(%args);

    # validate input parameter
    my @input_params = (%args);
    eval {
        # validate input parameters
        (%args) = validate(@input_params,
                          {type => { type => SCALAR, required => 1 }});
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return (wantarray()) ? (): [];
    }

    # make sure we have a valid type
    unless ($args{type} =~ /iscsi|fcp/i) {
        $self->_set_err("required \"iscsi\" or \"fcp\" type. invalid type: " .
                        $args{type});
        return (wantarray()) ? (): [];
    }

    # use parameters from nate host file
    $filer_name = $self->{filer};
    $host_obj = $self->{conn_obj}{$filer_name}{cli}{hostp};
    my $vserver = $self->get_name();

    # VS0_ISCSI_PREFIX=vs0_iscsi_lif
    # VS0_ISCSI_NUM_LIFS=2
    # VS0_ISCSI_PORTS=e0b:e0c
    # VS0_ISCSI_IP_RANGE=9.0.120.[7-8]
    # VS0_ISCSI_NETMASK=255.252.0.0

    # we need to use Options package to take in cmdline parameters as well
    # verify we have the correct parameter
    my $prefix_vsparam = uc($vserver) . '_' . uc($args{type}) . '_';
    my $prefix_param = $prefix_vsparam;

    if (exists $host_obj->{$prefix_param . "PREFIX"}) {
        $prefix_name = $host_obj->{$prefix_param . "PREFIX"};
    } else {
        # if it's iscsi, we will use the default setting
        # else fail
        if ($args{type} =~ /iscsi/i) {
            $prefix_param = "DATA_";
            $prefix_name = "iscsi_lif";
        } else {
            $self->_set_err("failed to find valid parameters for: " . $args{type} .
                            " Please refer to perldoc for usage");
            return (wantarray()) ? (): [];
        }
    }

    if (exists $host_obj->{$prefix_param . "NUM_LIFS"}) {
        $num_lifs = $host_obj->{$prefix_param . "NUM_LIFS"};
    } else {
        $num_lifs = 1;
    }

    if (exists $host_obj->{$prefix_param. "PORTS"}) {
        @ports = split(/:/, $host_obj->{$prefix_param . "PORTS"});
    } elsif (exists $host_obj->{$prefix_param . "PORT1"}) {
        $ports[0] = $host_obj->{$prefix_param . "PORT1"};
    } else {
        $self->_set_err("need  \"" . $prefix_vsparam . "PORTS\" or \"" .
                        $prefix_param . "PORTS\"to be defined in nate host");
        return (wantarray()) ? (): [];
    }


    # get additional parameters for iscsi type
    if ($args{type} =~ /iscsi/i) {
        if (exists $host_obj->{$prefix_param . "IP_RANGE"}) {
            $ip_range = $host_obj->{$prefix_param . "IP_RANGE"};
            # does it has the range? if not, we have a single ip
            if ($ip_range =~ /\[|\]/) {
                # split up ip_range
                my ($first, $second, $third, $fourth) = split(/\./, $ip_range);
                # find the range in the third and fourth byte
                if ($third =~ /\[(\d+)-(\d+)\]/) {
                    for (my $i = $1; $i <= $2; $i++) {
                        # do we have the range in the forth byte as well?
                        if ($fourth =~ /\[(\d+)-(\d+)\]/) {
                            for (my $j = $1; $j <= $2; $j++) {
                                push(@ips, "$first.$second.$i.$j");
                            }
                        } else {
                            # range in third byte only
                            push(@ips, "$first.$second.$i.$fourth");
                        }
                    }
                }
                # range in fourth byte only
                if ($fourth =~ /\[(\d+)-(\d+)\]/) {
                    for (my $i = $1; $i <= $2; $i++) {
                        push(@ips, "$first.$second.$third.$i");
                    }
                }
            } else {
                # assuming we have a single ip
                push @ips, $ip_range;
            }
        } elsif (exists $host_obj->{$prefix_param . "IP1"}) {
            push @ips, $host_obj->{$prefix_param . "IP1"};
        } else {
            $self->_set_err("need  \"" . $prefix_vsparam . "IP_RANGE\" or \"" .
                            $prefix_param . "IP1\"to be defined in nate host");
            return (wantarray()) ? (): [];
        }
        if (exists $host_obj->{$prefix_param . "NETMASK"} or exists $host_obj->{$prefix_param . "MASK1"}) {
            $netmask = $host_obj->{$prefix_param . "NETMASK"} || $host_obj->{$prefix_param . "MASK1"};
        } else {
            $self->_set_err("need  \"" . $prefix_vsparam . "NETMASK\" or \"" .
                            $prefix_param . "MASK1\"to be defined in nate host");
            return (wantarray()) ? (): [];
        }
    }

    # build an array of lif data from the nate params
    for (my $i = 1; $i <= $num_lifs; $i++) {
        my %lif_attributes;
        $lif_attributes{lif} = $prefix_name . $i;

        # round robin the ports
        $lif_attributes{port} = $ports[$i % scalar(@ports)];

        # add ip and netmask for iscsi
        if ($args{type} =~ /iscsi/i) {
            # round robin the ips
            $lif_attributes{address} = $ips[$i % scalar(@ips)];
            $lif_attributes{netmask} = $netmask;
        }
        push @lif_data, \%lif_attributes;
    }

    return (wantarray()) ? @lif_data : \@lif_data;
}

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

VserverHelper.pm

=head1 Author

Korry Nguyen (dl-ssan-automation)

=head1 Description

Provides helper methods for Vserver package.  This package maintains separately from Vserver for maintainability purpose while preventing convoluting Vserver object model.

=head1 See Also

VserverHelper package is used as part of the base class for B<San::ControllerCluster::Filer10M::Vserver>. See these modules for further interface detail.

B<San::CommonInterface.pm, San::ControllerCluster::Vserver.pm>

=head1 B<Public Methods>

=head2 B<get_lifs(%args)>

Get available lifs on this vserver

=over 4

=item Inputs

=item * I<-type> (optional)

Type of lifs to retrieve (iscsi or fcp)

=item Outputs

Returns an array of lif objects. If error, error message is pushed onto the error stack

=item Examples

 # ex. 1: Get iSCSI lif objects
 @iscsi_lifs = $vserver_obj->get_lifs(-type => "iscsi");
 if ($vserver_obj->error){
     ...
 }

=back

=head2 B<load_fcp_targets(%args)>

Load fcp targets for this vserver. Without %args parameters, it will used parameters (PREFIX, NUM_LIFS, PORTS) from the testbed nate host file.

=over 4

=item Inputs

=item * I<-prefix> (optional)

Prefix name for target lif

=item * I<-num_lifs> (optional)

Number of lifs to load/create

=item * I<-ports> (optional)

List of physical ports to create lifs with. It will be round-robin.

=item Outputs

Returns TRUE if successful. Otherwise sets error and returns FALSE

=item Examples

 # example of nate host
 VS0_FCP_PREFIX=vs0_fcp_lif
 VS0_FCP_NUM_LIFS=2
 VS0_FCP_PORTS=5a:5b

 # ex. 1: Load fcp targets using parameters from nate host
 $vserver_obj->load_fcp_targets();
 if ($vserver_obj->error){
     ...
 }

 # ex. 2: Load fcp targets with passed in parameters.
 # if fcp targets do not exist, create and load them
 $vserver_obj->load_fcp_targets(-prefix => "fcp_lif",
                                -num_lifs => 2,
                                -ports => \@fcp_ports);
 if ($vserver_obj->error){
     ...
 }

=back

=head2 B<load_iscsi_targets(%args)>

Load iscsi targets for this vserver. Without %args parameters, it will used parameters (PREFIX, NUM_LIFS, PORTS, IP_RANGE, NETMASK) from the testbed nate host file.

=over 4

=item Inputs

=item * I<-prefix> (optional)

Prefix name for target lif

=item * I<-num_lifs> (optional)

Number of lifs to load/create

=item * I<-ports> (optional)

List of physical ports to create lifs with. It will be round-robin.

=item * I<-ips> (optional)

List of IP addresses to be used to create lifs

=item * I<-netmask> (optional)

Netmask to be used to create lifs

=item Outputs

Returns TRUE if successful. Otherwise sets error and returns FALSE

=item Examples

 # example of nate host
 VS0_ISCSI_PREFIX=iscsi_lif
 VS0_ISCSI_NUM_LIFS=2
 VS0_ISCSI_PORTS=e0b:e0c
 VS0_ISCSI_IP_RANGE=9.0.[120-121].7
 VS0_ISCSI_NETMASK=255.252.0.0

 # ex. 1: Load iscsi targets using parameters from nate host
 $vserver_obj->load_iscsi_targets();
 if ($vserver_obj->error){
     ...
 }

 # ex. 2: Load iscsi targets with passed in parameters.
 # if iscsi targets do not exist, create and load them
 $vserver_obj->load_iscsi_targets(-prefix => "iscsi_lif",
                                  -num_lifs => 2,
                                  -ips => \@ip_addresses,
                                  -netmask => "255.252.0.0",
                                  -ports => \@iscsi_ports);
 if ($vserver_obj->error){
     ...
 }

=back

=head2 B<get_routing_groups()>

Get routing groups from this vserver

=over 4

=item Inputs

None

=item Outputs

Returns an array list of routing groups if successful. Otherwise sets error and returns empty array

=item Examples

 # ex. 1: Get the available routing groups in this vserver
 @rg = $vserver_obj->get_routing_groups();
 if ($vserver_obj->error){
     ...
 }

=back

=head2 B<get_routes()>

Get destination of the routes in the routing group

=over 4

=item Inputs

=item * I<'-routing-group'> (required)

Routing group that belongs to this vserver

=item Outputs

Returns an array list of destination of the routes if successful. Otherwise sets
error and returns empty array

=item Examples

 # ex. 1: Get the destination of the routes
 @route_destinations = $vserver_obj->get_routes();
 if ($vserver_obj->error){
     ...
 }

=back

=head2 B<delete_routing_group(%args)>

Delete routing group from this vserver

=over 4

=item Inputs

=item * I<'-routing-group'> (required)

Routing group that belongs to this vserver

=item Outputs

Returns TRUE if successful. Otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Delete routing group from this vserver
 $vserver_obj->delete_routing_group('-routing-group' => $rg);
 if ($vserver_obj->error){
     ...
 }

=back

=head2 B<delete_route(%args)>

Delete route from routing group

=over 4

=item Inputs

=item * I<'-routing-group'> (required)

Routing group that belongs to this vserver

=item * I<-destination> (required)

destination of the route to be deleted

=item Outputs

Returns TRUE if successful. Otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Delete route from routing group $rg and destination 0.0.0.0/0
 $vserver_obj->delete_route('-routing-group' => $rg,
                            -destination => '0.0.0.0/0');
 if ($vserver_obj->error){
     ...
 }

=back

=head2 B<create_tpgroup_obj(%args)>

Create a tpgroup object. Iscsi service will be started on this vserver before
creating tpgroup object if its not already started. After the tpgroup object is
created, it will be stored in Vserver object.

=over 4

=item Inputs

=item * I<-name> (required)

The tpgroup name to be used

=item * I<-lif> (optional)

The lif name with which this tpgroup has to be created

=item * I<-debug_level> => <level of debug>

Debug level from 0 - 100. With higher level, it will log more debug messages

=item Outputs

Return Tpgroup object

=item Examples

 # ex. 1: Create tpgroup object
 $tpgroup_obj = $vs_obj->create_tpgroup_obj(-name => $tpgroup_name);
 if ($vs_obj->error()){
     ...
 }

 # ex. 2: Create tpgroup object with a lif
 $tpgroup_obj = $vs_obj->create_tpgroup_obj(-name => $tpgroup_name,
                                            -lif => "iscsi_data_lif");
 if ($vs_obj->error()){
     ...
 }
=back
