# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary provides 7mode compatible methods for Filer10M class
##
## @author korry@netapp.com, dl-ssan-automation
## @status Shared

package San::ControllerCluster::FilerCompat;

# compiler directives
use strict;
use lib();

# imported modules;
use Params::Validate qw( SCALAR ARRAYREF HASHREF OBJECT );

# inherit base class
use base qw( San::CommonInterface );

################################################################################
# Method Name: aggr_create()
# Objective:   create a new aggregate. This method map one-to-one to 7-mode
# Details:     See POD documentation at the end of this file
################################################################################
sub aggr_create {
    my ($self, %args) = @_;

    return ($self->aggregate_create(%args));
}

################################################################################
# Method Name: aggr_destroy()
# Objective:   Delete aggregate. This maps one-to-one
#              to 7-mode
# Details:     See POD documentation at the end of this file
################################################################################
sub aggr_destroy {
    my ($self, $name) = @_;

    return ($self->destroy_aggregate_obj(-name => $name));
}

################################################################################
# Method Name: aggr_exist()
# Objective:   Determines if aggregate exists. This maps one-to-one
#              to 7-mode
# Details:     See POD documentation at the end of this file
################################################################################
sub aggr_exist {
    my ($self, $name) = @_;

    return ($self->aggregate_exist(-name => $name));
}

################################################################################
# Method Name: aggr_offline()
# Objective:   Offline an aggregate. This maps one-to-one
#              to 7-mode
# Details:     See POD documentation at the end of this file
################################################################################
sub aggr_offline {
    my ($self, $name) = @_;

    return ($self->aggregate_offline(-name => $name));
}

###############################################################################
# Method Name: connection
# Objective:   Get/Set filer connection
# Details:     See POD documentation at the end of this file.
###############################################################################
sub connection {
    my ($self, $value) = @_;

    my $filer   = $self->get_name();
    my $cmd_interface = $self->get_cmd_interface();

    if (defined($value)) {
        $self->{conn_obj}->{$filer}{$cmd_interface} = $value;
    }
    my $connection = $self->{conn_obj}->get_connection(-filer => $filer,
                                                       -cmd_interface => $cmd_interface);
    if ($self->{conn_obj}->error()) {
        $self->_set_err("fail to get_connection: " . $self->{conn_obj}->errmsg());
    }
    return $connection;
}

################################################################################
# Method Name: create_igroup()
# Objective:   create and return an igroup object. This maps one-to-one to 7mode
# Details:     See POD documentation at the end of this file
################################################################################
sub create_igroup {
    my ($self, %args) = @_;

    my ($igroup_obj, $vs_obj);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert igroup to name to maintain compatability w/ 7mode
    if (exists $args{name}) {
        $args{igroup} = delete $args{name};
    }
    if (exists $args{hosttype}) {
        $args{ostype} = delete $args{hosttype};
    }
    if (exists $args{'protocol-type'}) {
        $args{'protocol'} = delete $args{'protocol-type'};
    }
    if (exists $args{'vserver_obj'}) {
        $vs_obj = delete $args{'vserver_obj'};
    }

    # set default vserver if it's not specified
    unless ($args{vserver}) {
        $args{vserver} = $self->DEFAULT_VSERVER;
    }

    # discover and retrieve vserver object (unless it was passed to us)
    if (not defined($vs_obj) or ref($vs_obj) !~ /Vserver/) {
        $vs_obj = $self->get_vserver_obj(-name  => $args{vserver});
        if ($self->error() or ref($vs_obj) !~ /Vserver/) {
            $self->_set_err("can not find vserver: " . $args{vserver},
                            $self->errmsg());
            return {};
        }
    }

    # create an igroup object
    $igroup_obj = $vs_obj->create_igroup_obj(%args);
    if ($vs_obj->error()) {
        $self->_set_err("fail to create igroup object " . $args{name}, $vs_obj->errmsg());
        return {};
    }

    return $igroup_obj;
} ## end sub create_igroup

################################################################################
# Method Name: create_vol()
# Objective:   create volume object. This maps one-to-one to 7-mode.
# Details:     See POD documentation at the end of this file
################################################################################
sub create_vol {
    my ($self, %args) = @_;

    return ($self->create_volume(%args));
}

################################################################################
# Method Name: destroy_igroup()
# Objective:   Destroy igroup. This maps one-to-one to 7mode
# Details:     See POD documentation at the end of this file
################################################################################
sub destroy_igroup {
    my ($self, %args) = @_;

    my (@igroup_objs, @tmp_objs, $vs_obj, %tmp_args);

    $self->_clear_err();

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|igroup)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # this is a scalar
            $tmp_args{igroup} = $key;
        }
    }
    # set tmp_args to args
    %args = %tmp_args;

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert igroup to name to maintain compatability w/ 7mode
    if (exists $args{name}) {
        $args{igroup} = $args{name};
        delete $args{name};
    }
    # verify mandatory name param
    unless (defined $args{igroup} and length $args{igroup}) {
        $self->_set_err("mandatory parameter \"-igroup\" not specified ");
        return $SanUtil::FALSE;
    }

    # set default vserver if it's not specified
    unless ($args{vserver}) {
        $args{vserver} = $self->DEFAULT_VSERVER;
    }

    # get igroups from parent obj.
    @igroup_objs = $self->get_igroups(-vserver => $args{vserver});

    # delete igroup object that matches and update the list
    foreach my $igroup_obj (@igroup_objs) {
        # get specific igroup object and delete it
        if ($igroup_obj->get_name() =~ /^$args{igroup}$/) {
            $igroup_obj->destroy(%args);
            if ($igroup_obj->error()) {
                $self->_set_err("failed to destroy igroup: " . $args{igroup},
                                $igroup_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {
            push (@tmp_objs, $igroup_obj);
        }
    }

    # store the list back to the parent
    $self->{igroup_objs} = \@tmp_objs;

    # store the list back to the vserver obj as well
    $vs_obj = $self->get_vserver_obj(-name => $args{vserver});
    if ($self->error() or ref($vs_obj) !~ /Vserver/) {
     $self->_set_err("can not find vserver: " . $args{vserver}, $vs_obj->errmsg());
     return $SanUtil::FALSE;
    }
    $vs_obj->{igroup_objs} = \@tmp_objs;

    return $SanUtil::TRUE;
} ## end sub destroy_igroup

################################################################################
# Method Name: destroy_volume()
# Objective:   Destroy volume
# Details:     See POD documentation at the end of this file
################################################################################
sub destroy_volume {
    my ($self, %args) = @_;

    my (@volume_objs, @tmp_objs, $vs_obj);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert volume to name to maintain compatability w/ 7mode
    if (exists $args{name}) {
        $args{volume} = $args{name};
        delete $args{name};
    }
    # verify mandatory name param
    unless (defined $args{volume} and length $args{volume}) {
        $self->_set_err("mandatory parameter \"-volume\" not specified ");
        return $SanUtil::FALSE;
    }

    # set default vserver if it's not specified
    unless ($args{vserver}) {
        $args{vserver} = $self->DEFAULT_VSERVER;
    }

    # get volumes from parent obj.
    @volume_objs = $self->get_volumes(-vserver => $args{vserver});

    # delete volume object that matches and update the list
    foreach my $volume_obj (@volume_objs) {
        # get specific volume object and delete it
        if ($volume_obj->get_name() =~ /^$args{volume}$/) {
            $volume_obj->destroy(%args);
            if ($volume_obj->error()) {
                $self->_set_err("failed to destroy volume: " . $args{volume},
                                $volume_obj->errmsg());
                return $SanUtil::FALSE;
            }
        } else {
            push (@tmp_objs, $volume_obj);
        }
    }

    # store the list back to the parent
    $self->{volume_objs} = \@tmp_objs;

    return $SanUtil::TRUE;
} ## end sub destroy_volume

###############################################################################
# Method Name: fcp_start
# Objective:   A wrapper to support 7-mode users. Starts fcp service.
# Details:     See POD documentation at the end of this file.
###############################################################################
sub fcp_start {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    my $status = $SanUtil::FALSE;

    # Get the fcp object to call the Fcp.pm API
    my $fcp_obj = $self->_get_fcp_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    my $service_status = $fcp_obj->status(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp status", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Start FCP if needed
    if ($service_status eq 'offline') {
        $status = $fcp_obj->start(%args);
        if ($fcp_obj->error()) {
            $self->_set_err("failed to start fcp ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    } elsif ($service_status eq 'online') {
        return $SanUtil::TRUE;
    }

    return $status;
}

###############################################################################
# Method Name: fcp_stop
# Objective:   A wrapper to support 7-mode users. Stop fcp service.
# Details:     See POD documentation at the end of this file.
###############################################################################
sub fcp_stop {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();
    my $status;

    # Get the fcp object to call the Fcp.pm API
    my $fcp_obj = $self->_get_fcp_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    my $service_status = $fcp_obj->status(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp status", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Stop FCP
    if ($service_status eq 'online') {
        $status = $fcp_obj->stop(%args);
        if ($fcp_obj->error()) {
            $self->_set_err("failed to stop fcp ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    } elsif ($service_status eq 'offline') {
        return $SanUtil::TRUE;
    }

    return $status;
}

###############################################################################
# Method Name: fcp_adapter_online
# Objective:   A wrapper to support 7-mode users. fcp_adapter_online.
# Details:     See POD documentation at the end of this file.
###############################################################################
sub fcp_adapter_online {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    # Get the fcp object to call the Fcp.pm API
    my $fcp_obj = $self->_get_fcp_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # bring the adapter online
    # NOTE: Not implemented yet in SJ4 so adapter_online is a no-op!
    my $status = $fcp_obj->adapter_online(%args);
    if ($fcp_obj->error()) {
        $self->_set_err("failed to fcp adapter online ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $status;
}

###############################################################################
# Method Name: fcp_wwpnalias_get
# Objective:   A wrapper to support 7-mode users. Same as wwpnalias_get.
# Details:     See wwpnalias_get POD.
###############################################################################
sub fcp_wwpnalias_get {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    # Get the fcp object to call the Fcp.pm API
    my $fcp_obj = $self->_get_fcp_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp object", $self->errmsg());
        return {};
    }

    # Get wwpnalias
    my @wwpnaliases = $fcp_obj->wwpnalias_get(%args);
    if ($fcp_obj->error()) {
        $self->_set_err("failed to get wwpnalias ", $self->errmsg());
        return {};
    }

    return (wantarray()) ? @wwpnaliases : \@wwpnaliases;
} # End of fcp_wwpnalias_get()

###############################################################################
# Method Name: fcp_wwpnalias_remove
# Objective:   A wrapper to support 7-mode users. Same as wwpnalias_remove.
# Details:     See wwpnalias_remove POD.
###############################################################################
sub fcp_wwpnalias_remove {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    # Get the fcp object to call the Fcp.pm API
    my $fcp_obj = $self->_get_fcp_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp object", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Remove wwpnalias
    my $status= $fcp_obj->wwpnalias_remove(%args);
    if ($fcp_obj->error()) {
        $self->_set_err("failed to remove wwpnalias ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $status;
} # End of fcp_wwpnalias_remove()

###############################################################################
# Method Name: fcp_wwpnalias_set
# Objective:   A wrapper to support 7-mode users. Same as wwpnalias_set.
# Details:     See wwpnalias_set POD.
###############################################################################
sub fcp_wwpnalias_set {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    # Get the fcp object to call the Fcp.pm API
    my $fcp_obj = $self->_get_fcp_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Set wwpnalias
    my $status= $fcp_obj->wwpnalias_set(%args);
    if ($fcp_obj->error()) {
        $self->_set_err("failed to set wwpnalias object", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $status;
} # End of fcp_wwpnalias_set

################################################################################
# Method Name: get_aggrs()
# Objective:   Get a list of aggregate names. This method is 7mode compatible
# Details:     See POD documentation at the end of this file
################################################################################
sub get_aggrs {
    my ($self, %args) = @_;

    my (@aggregate_objs, @aggregate_names);

    $self->_clear_err();

    # if aggregate objects has been cached, get it
    # otherwise, call get_aggregate_objs
    if (defined $self->{aggregate_objs} and scalar(@{$self->{aggregate_objs}})) {
        @aggregate_objs = @{$self->{aggregate_objs}};
    } else {
        # discover and get aggregate objects
        @aggregate_objs = $self->get_aggregate_objs();
        if ($self->error()) {
            $self->_set_err("failed to get aggregate objects", $self->errmsg());
            return (wantarray()) ? (): [];
        }
    }

    # get a list of aggregate names
    foreach my $aggr_obj (@aggregate_objs) {
        push(@aggregate_names, $aggr_obj->get_name());
    }

    return (wantarray()) ? @aggregate_names: \@aggregate_names;
} ## end sub get_aggrs

################################################################################
# Method Name: get_cfmode()
# Objective:   get cfmode of the filer. Clustered ONTAP has no such cfmode
#              To be compatible w/ 7-M, This method will always return SSI
# Details:     See POD documentation at the end of this file
################################################################################
sub get_cfmode {
    my ($self) = @_;

    # stub to return single_image
    my $cfmode = "single_image";

    return $cfmode;
}

################################################################################
# Method Name: get_partner()
# Objective:   Get the HA partner node name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_partner {
    my ($self, %args) = @_;

    # default the node to the current node
    if (not exists($args{node})) {
        $args{node} = $self->get_name();
    }

    my ($output) = $self->sfo_show(-fields => "partner-name",
                                   %args);
    if ($self->error()) {
        $self->_set_err("Failed to get HA partner name", $self->errmsg());
        return "";
    }

    return $output->{partner_name};
}

################################################################################
# Method Name: load_sram_dump_list()
# Objective:   load in the initial sram dump list to validate against.
# Details:     See POD documentation at the end of this file
################################################################################
sub load_sram_dump_list {
    my ($self) = @_;

    # switch shell context to freebsd
    $self->set_shell_environment(-type => "bsd");
    if ($self->error()) {
        $self->_set_err("Unable to change shell to 'bsd'", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # check /mroot/etc/log for any sram dumps and store it in the sram list
    my @output = $self->execute(-cmd => "ls /mroot/etc/log | cat | grep ispfct");
    if ($self->error()) {
        $self->_set_err("Unable to list SRAM dumps", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # store the sram list
    $self->set_sram_dump_list(@output);

    # switch shell context back to ngsh
    $self->set_shell_environment(-type => "ngsh");
    if ($self->error()) {
        $self->_set_err("Unable to change shell to 'ngsh'", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: get_sram_dump_list()
# Objective:   return the sram dump list that is currently stored
# Details:     See POD documentation at the end of this file
################################################################################
sub get_sram_dump_list {
    my ($self) = @_;

    if (not defined($self->{sram_list})) {
        return (wantarray()) ? () : [];
    } else {
        return (wantarray()) ? @{$self->{sram_list}} : $self->{sram_list};
    }
}

################################################################################
# Method Name: set_sram_dump_list()
# Objective:   set the sram dump list with the passed in value
# Details:     See POD documentation at the end of this file
################################################################################
sub set_sram_dump_list {
    my ($self, @sram_list) = @_;

    $self->{sram_list} = \@sram_list;

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: enable_fcp_adapter_sram_dumps()
# Objective:   Enable sram dumps on all fcp target ports
# Details:     See POD documentation at the end of this file
################################################################################
sub enable_fcp_adapter_sram_dumps {
    my ($self) = @_;

    # get all fcp adapter objects and enable sram dump generation on each one
    my @fcpadapter_objs = $self->get_fcpadapter_objs(-node => $self->get_name());
    if ($self->error()) {
        $self->_set_err("Failed to get FCP adapters", $self->errmsg());
        return $SanUtil::TRUE;
    }

    foreach my $fcp_adapter (@fcpadapter_objs) {
        $fcp_adapter->enable_sram_dumps();
        if ($fcp_adapter->error()) {
            $self->_set_err("Failed to enable sram dumps", $fcp_adapter->errmsg());
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: get_fcp_nodename
# Objective:   A wrapper to support 7-mode users. Gets FCP node name
# Details:     See POD documentation at the end of this file.
###############################################################################
sub get_fcp_nodename {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    # Get the fcp object to call the Fcp.pm API
    my $fcp_obj = $self->_get_fcp_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp object", $self->errmsg());
        return "";
    }

    # Get the nodename
    my $name = $fcp_obj->nodename(%args);
    if ($fcp_obj->error()) {
        $self->_set_err("failed to get fcp nodename", $self->errmsg());
        return "";
    }

    return $name
}

###############################################################################
# Method Name: get_fcp_status
# Objective:   A wrapper to support 7-mode users. Gets the status of FCP service.
# Details:     See POD documentation at the end of this file.
###############################################################################
sub get_fcp_status {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    # Get the fcp object to call the Fcp.pm API
    my $fcp_obj = $self->_get_fcp_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp object", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Get the status of FCP
    my $status = $fcp_obj->status(%args);
    if ($fcp_obj->error()) {
        $self->_set_err("failed to fcp adapter online", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $status;
}

################################################################################
# Method Name: get_fcp_targets()
# Objective:   get fcp target objects
# Details:     See POD documentation at the end of this file
################################################################################
sub get_fcp_targets {
    my ($self, %args) = @_;

    my (@target_objs, $vs_obj);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Verify if the node is a vsim, if so return
    if ($self->is_vsim()) {
        $self->sanlog(-msg => "Fcp is not supported on vsims ",
                      -msg_level => 0);
        return (wantarray()) ? () : [];
    }

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);
    eval {

        # validate input parameters and their type.
        # list - array that contains names of object to be created
        (%args) =
          Module::Common::validate_mixed(@input_params,
                                         [],
                                         {use_nate_params => {type => SCALAR,
                                                       optional => 1,
                                                       default => 0},
                                          vserver => {type => SCALAR,
                                                      optional => 1,
                                                      default =>  $self->DEFAULT_VSERVER}});
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return (wantarray()) ? (): [];
    }

    # discover and retrieve vserver object
    $vs_obj = $self->get_vserver_obj(-name  => $args{vserver});
    if ($self->error() or ref($vs_obj) !~ /Vserver/) {
        $self->_set_err("can not find vserver: " . $args{vserver},
                        $self->errmsg());
        return (wantarray()) ? (): [];
    }

    # Load vserver fcp targets only if vserver type is cluster
    if (defined $vs_obj->get_type() =~ /cluster/ and $vs_obj->get_aggregate() !~ /\-/) {
            $vs_obj->{fcp_targets} = [];
            $vs_obj->load_fcp_targets(use_nate_params => $args{use_nate_params});
            if ($vs_obj->error()) {
                $self->_set_err("failed to load fcp target", $vs_obj->errmsg());
                return (wantarray()) ? (): [];
            }
    }
    # at this point, we should have at least one fcp targets in vserver object
    # if not, set error and return
    if (defined($vs_obj->{fcp_targets}) and scalar(@{$vs_obj->{fcp_targets}})) {
        @target_objs = @{$vs_obj->{fcp_targets}};
    } else {
        $self->sanlog(-msg => "no FCP target in " . $args{vserver} . " vserver",
                      -msg_level => 0);
        return (wantarray()) ? () : [];
    }

     return (wantarray()) ? @target_objs : \@target_objs;
} ## end sub get_fcp_targets

################################################################################
# Method Name: get_igroup()
# Objective:   Get a specific igroup
# Details:     See POD documentation at the end of this file
################################################################################
sub get_igroup {
    my ($self, %args) = @_;

    my (@igroup_objs);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert igroup to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{igroup}) {
        $args{name} = $args{igroup};
        delete $args{igroup};
    }

    # verify name param
    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return {};
    }

    # set default vserver if it's unspecified
    $args{vserver} = $self->DEFAULT_VSERVER unless (defined $args{vserver});

    # get igroup objects
    @igroup_objs = $self->get_igroups(-vserver => $args{vserver});
    if ($self->error()) {
        $self->_set_err("failed to get igroup objects from " . $args{vserver}, $self->errmsg());
        return {};
    }

    # find a specific igroup object
    foreach my $igroup_obj (@igroup_objs) {
        return $igroup_obj
          if ($igroup_obj->get_name() =~ /^$args{name}$/);
    }

    # no igroup object that matches the criteria
    $self->_set_err("can not find igroup " . $args{name});
    return {};
} ## end sub get_igroup

################################################################################
# Method Name: get_igroups()
# Objective:   get a list of igroup objects
# Details:     See POD documentation at the end of this file
################################################################################
sub get_igroups {
    my ($self, %args) = @_;

    my (@igroup_objs, $vs_obj);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # set default vserver if it's not specified
    $args{vserver} = $self->DEFAULT_VSERVER unless (defined $args{vserver});

    # discover and retrieve vserver object
    $vs_obj = $self->get_vserver_obj(-name  => $args{vserver});
    if ($self->error() or ref($vs_obj) !~ /Vserver/) {
        $self->_set_err("can not find vserver: " . $args{vserver},
                        $self->errmsg());
        return (wantarray()) ? (): [];
    }

    # if igroup objects has been cached, get it
    # otherwise, call get_igroup_objs
    if (defined $vs_obj->{igroup_objs} and scalar(@{$vs_obj->{igroup_objs}})) {
        @igroup_objs = @{$vs_obj->{igroup_objs}};
    } else {
        # discover and get igroup objects
        @igroup_objs = $vs_obj->get_igroup_objs();
        if ($vs_obj->error()) {
            $self->_set_err("failed to get igroup objects", $vs_obj->errmsg());
            return (wantarray()) ? (): [];
        }
    }
    return (wantarray()) ? @igroup_objs: \@igroup_objs;
} ## end sub get_igroups

################################################################################
# Method Name: get_iscsi_targets()
# Objective:   get iscsi target objects
# Details:     See POD documentation at the end of this file
################################################################################
sub get_iscsi_targets {
    my ($self, %args) = @_;

    my (@target_objs, @filtered_objs, @added_ips, $vs_obj);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);
    eval {

        # validate input parameters and their type.
        # list - array that contains names of object to be created
        (%args) =
          Module::Common::validate_mixed(@input_params,
                                         [],
                                         {state => {type => SCALAR,
                                                    optional => 1,
                                                    default => "up"},
                                          valid_ip => {type => SCALAR,
                                                       optional => 1,
                                                       default => 1},
                                          use_nate_params => {type => SCALAR,
                                                       optional => 1,
                                                       default => 0},
                                          vserver => {type => SCALAR,
                                                      optional => 1,
                                                      default =>  $self->DEFAULT_VSERVER}});
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return {};
    }

    # change state to up
    $args{state} = "up" if ($args{state} =~ /enabled/i);

    # discover and retrieve vserver object
    $vs_obj = $self->get_vserver_obj(-name  => $args{vserver});
    if ($self->error() or ref($vs_obj) !~ /Vserver/) {
        $self->_set_err("can not find vserver: " . $args{vserver},
                        $self->errmsg());
        return (wantarray()) ? (): [];
    }

    # Load vserver iscsi targets only if vserver type is cluster
    if (defined $vs_obj->get_type() =~ /cluster/ and $vs_obj->get_aggregate() !~ /\-/) {
            $vs_obj->{iscsi_targets} = [];
            $vs_obj->load_iscsi_targets(use_nate_params => $args{use_nate_params});
            if ($vs_obj->error()) {
                $self->_set_err("failed to load iscsi target", $vs_obj->errmsg());
                return (wantarray()) ? (): [];
            }
    }
    # if iscsi object does exist in cache, get it
    # otherwise, call get_iscsi_objs
    if (defined($vs_obj->{iscsi_targets}) and scalar(@{$vs_obj->{iscsi_targets}})) {
        @target_objs = @{$vs_obj->{iscsi_targets}};

        # get targets with valid ip addresses and state if defined
        foreach my $target (@target_objs) {
            # get ip address from target
            my $ip = $target->get_ip();

            # filtered targets by valid ip
            if ($args{valid_ip} == 1) {
                unless ($ip =~ /not_set/i) {
                    push(@filtered_objs, $target);
                    # keep track of what is being pushed onto the array
                    # so we don't do it twice
                    push(@added_ips, $ip);
                }
            }
            # filtered targets by state
            if (defined($args{state})) {
                if ($target->get_state() =~ /$args{state}/i) {
                    # avoid pushing target object twice
                    unless (List::MoreUtils::any
                        {$_ =~ /$ip/} @added_ips) {
                        push(@filtered_objs, $target);
                    }
                }
            }
        }

        # reset target list to filtered list
        @target_objs = @filtered_objs;
    } else {
        $self->sanlog(-msg => "no iSCSI target in " . $args{vserver} . " vserver",
                      -msg_level => 0);
        return (wantarray()) ? () : [];
    }

     return (wantarray()) ? @target_objs : \@target_objs;
} ## end sub get_iscsi_targets

################################################################################
# Method Name: get_targets()
# Objective:   wrapper to get_fcp_target. This maps one-to-one to 7-mode.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_targets {
    my ($self, %args) = @_;

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # set default vserver if it's unspecified
    $args{vserver} = $self->DEFAULT_VSERVER unless (defined $args{vserver});

    return ($self->get_fcp_targets(%args));
}
################################################################################
# Method Name: get_vol()
# Objective:   Get volume object. This maps one-to-one to 7-mode.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vol {
    my ($self, %args) = @_;

    my (%tmp_args, $vs_obj, @volume_objs);

    $self->_clear_err();

    # hate this crap but we have to make it compatible w/ 7-mode. This means
    # we have to take scalar as param instead of hash

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|volume|vserver)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
            last;
        } else {
            # this is a scalar
            $tmp_args{name} = $key;
        }
    }
    # set tmp_args to args
    %args = %tmp_args;
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert volume to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{volume}) {
        $args{name} = $args{volume};
        delete $args{volume};
    }

    # verify name param
    unless (defined $args{name} and length $args{name}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return {};
    }

    # set default vserver if it's unspecified
    $args{vserver} = $self->DEFAULT_VSERVER unless (defined $args{vserver});

    # get volume objects
    @volume_objs = $self->get_volumes(-vserver => $args{vserver});
    if ($self->error()) {
        $self->_set_err("failed to get volume objects from " . $args{vserver}, $self->errmsg());
        return {};
    }

    # find a specific volume object
    foreach my $volume_obj (@volume_objs) {
        return $volume_obj
          if ($volume_obj->get_name() =~ /^$args{name}$/);
    }

    # no volume object that matches the criteria
    $self->_set_err("can not find volume " . $args{name});
    return {};
} ## end sub get_vol

################################################################################
# Method Name: igroup_exist()
# Objective:   Determines if igroup exists.
# Details:     See POD documentation at the end of this file
################################################################################
sub igroup_exist {
    my ($self, %args) = @_;

    my (@output, %tmp_args);

    $self->_clear_err();

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|igroup|vserver)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
            last;
        } else {
            # this is a scalar
            $tmp_args{igroup} = $key;
        }
    }
    # set tmp_args to args
    %args = %tmp_args;

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to igroup to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{name}) {
        $args{igroup} = $args{name};
        delete $args{name};
    }

    # verify name param
    unless (defined $args{igroup} and length $args{igroup}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return $SanUtil::FALSE;
    }

    # set default vserver if it's unspecified
    $args{vserver} = $self->DEFAULT_VSERVER unless (defined $args{vserver});

    # build cmd for lun igroup show
    %args = (%args, cmd => "lun_igroup_show");

    # invoke cdef command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: iscsi_start
# Objective:   A wrapper to support 7-mode users. Starts iscsi service.
# Details:     See POD documentation at the end of this file.
###############################################################################
sub iscsi_start {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    # Get the iscsi object to call the Iscsi.pm API
    my $iscsi_obj = $self->_get_iscsi_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get iscsi ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    my $service_status = $iscsi_obj->status(%args);
    if ($self->error()) {
        $self->_set_err("failed to get iscsi status", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Start iscsi if needed
    if ($service_status eq 'offline') {
        $iscsi_obj->start(%args);
        if ($iscsi_obj->error()) {
            $self->_set_err("failed to start iscsi ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    } elsif ($service_status eq 'online') {
        return $SanUtil::TRUE;
    } else {
        # if we get here, the status is neither online or offline
        $self->_set_err("unknown iscsi status: ", $service_status);
        return $SanUtil::FALSE;
    }
}

###############################################################################
# Method Name: iscsi_stop
# Objective:   A wrapper to support 7-mode users. Stop iscsi service.
# Details:     See POD documentation at the end of this file.
###############################################################################
sub iscsi_stop {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    # Get the iscsi object to call the Iscsi.pm API
    my $iscsi_obj = $self->_get_iscsi_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get iscsi ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    my $service_status = $iscsi_obj->status(%args);
    if ($self->error()) {
        $self->_set_err("failed to get iscsi status", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Stop iscsi
    if ($service_status eq 'online') {
        $iscsi_obj->stop(%args);
        if ($iscsi_obj->error()) {
            $self->_set_err("failed to stop iscsi ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    } elsif ($service_status eq 'offline') {
        return $SanUtil::TRUE;
    } else {
        # if we get here, the status is neither online or offline
        $self->_set_err("unknown iscsi status: ", $service_status);
        return $SanUtil::FALSE;
    }
}

################################################################################
# Method Name: interface()
# Objective:   Get interface type (cli or zapi). This maps one-to-one to 7-mode.
# Details:     See POD documentation at the end of this file
################################################################################
sub interface {
    my ($self, $value) = @_;

    # just a stub right now to support 7mode
    return "cli";
}

################################################################################
# Method Name: lun_exist()
# Objective:   Determines if lun-path exists. This maps one-to-one to 7-mode.
# Details:     See POD documentation at the end of this file
################################################################################
sub lun_exist {
    my ($self, $lunpath, $vserver) = @_;

    my (@output, %args);

    $self->_clear_err();

    # verify name param
    unless (defined $lunpath and length $lunpath) {
        $self->_set_err("mandatory parameter \"\$lunpath\" not specified ");
        return $SanUtil::FALSE;
    }

    # set default vserver if it's unspecified
    $vserver = $self->DEFAULT_VSERVER unless (defined $vserver and length $vserver);
    # build up parameter for cdef
    $args{vserver} = $vserver;
    $args{path} = $lunpath;

    # build cmd lun show
    %args = (%args, cmd => "lun_show");

    # invoke cdef command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;

}

###############################################################################
# Method Name: set_fcp_nodename
# Objective:   A wrapper to support 7-mode users. Gets FCP node name
# Details:     See POD documentation at the end of this file.
###############################################################################
sub set_fcp_nodename {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    # Get the fcp object to call the Fcp.pm API
    my $fcp_obj = $self->_get_fcp_object(%args);
    if ($self->error()) {
        $self->_set_err("failed to get fcp object", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Set the node name
    my $status = $fcp_obj->set_fcp_nodename(%args);
    if ($fcp_obj->error()) {
        $self->_set_err("failed to set fcp nodename", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $status;
}

################################################################################
# Method Name: vol_status()
# Objective:   Get the volume information for the specified volume.
# Details:     See POD documentation at the end of this file
################################################################################
sub vol_status {
    my ($self, $vol_name, $vserver) = @_;

    my (@vol_records, %args);

    $self->_clear_err();

    # verify name param
    unless (defined $vol_name and length $vol_name) {
        $self->_set_err("mandatory parameter \"\$vol_name\" not specified ");
        return (wantarray()) ? () : [];
    }

    # set default vserver if it's unspecified
    $vserver = $self->DEFAULT_VSERVER unless (defined $vserver and length $vserver);
    # build up parameter for cdef
    $args{vserver} = $vserver;
    # set the volume argument
    $args{volume} = $vol_name;

    # build cmd volume show
    %args = (%args, cmd => "volume_show");

    # invoke cdef command but donot update attributes
    # 7mode vol_status returns an array, so use raw_data to do the same
    (@vol_records) = $self->SUPER::invoke_cmd_obj(%args, -raw_data => 1, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # if no entries, return FALSE
    if ($vol_records[0] =~ /There are no entries/i) {
        $self->sanlog(-msg => "volume $args{volume} does not exist",
                      -msg_level => 0);
        $self->_set_err("volume $args{volume} does not exist");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @vol_records : \@vol_records;
}

################################################################################
# Method Name: get_volnames()
# Objective:   Get the volume names on this controller/vserver
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volnames {
    my ($self, $vserver) = @_;

    my (@vol_records, @vol_names, %args);

    $self->_clear_err();

    # set default vserver if it's unspecified
    $vserver = $self->DEFAULT_VSERVER unless (defined $vserver and length $vserver);
    # build up parameter for cdef
    $args{vserver} = $vserver;

    # build cmd volume show
    %args = (%args, cmd => "volume_show");

    # invoke cdef command but donot update attributes
    @vol_records = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",$self->errmsg());
        return (wantarray()) ? () : [];
    }

    # 7mode vol_status returns an array, so convert hash to array
    foreach my $vol_record (@vol_records) {
        my $vol_name = $vol_record->{volume};
        push @vol_names, $vol_name;
    }
    return (wantarray()) ? @vol_names : \@vol_names;
} #end get_volnames

################################################################################
# Method Name: get_lun_minsize()
# Objective:   Get the minimum size for a lun on this controller
# Details:     See POD documentation at the end of this file
################################################################################
sub get_lun_minsize {
    my ($self, %args) = @_;

    my $type;

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?type/) {
            if (not length($type= $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return "";
            }
        }  else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return "";
        }
    }

    unless(defined $type and length $type){
            $self->_set_err("type is a required argument");
            return "";
    }

    #hard coding this value for now, not supported in cmode
    my $lun_minsize = "32901120";
    return $lun_minsize;
}

################################################################################
# Method Name: vol_exist()
# Objective:   Determines if volume exists. This maps one-to-one to 7-mode.
# Details:     See POD documentation at the end of this file
################################################################################
sub vol_exist {
    my ($self, $name, $vserver) = @_;

    return ($self->volume_exist(-name => $name, -vserver => $vserver));
}

################################################################################
# Method Name: wait_for_aggr_online()
# Objective:   wait for aggr to be online. This maps one-to-one to 7-mode.
# Details:     See POD documentation at the end of this file
################################################################################
sub wait_for_aggr_online {
    my ($self, %args) = @_;

    return ($self->wait_for_aggregate_online(%args));
}

###############################################################################
# Method:    sysstat()
# Objective: Collect system statistics on the specified node
# Details:     See POD documentation at the end of this file
###############################################################################
sub sysstat {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($stats);

    %args = &SanUtil::strip_dash(%args);

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             filer => {type => SCALAR,
                                                       optional => 1,
                                                       default => $self->get_name()},
                                             interval => {type => SCALAR,
                                                          optional => 1,
                                                          default => 10},
                                             count => {type => SCALAR,
                                                       optional => 1,
                                                       default => 1},
                                             pass_thru => {type => SCALAR,
                                                           optional => 1,
                                                           default => "-b"}
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return {};
    }

    # run a node scoped command to collect san statistics
    my $output = $self->run_node_scope(-cmd => "sysstat " . $args{pass_thru} . " -c " . $args{count} . " " . $args{interval},
                                       -filer => $args{filer});
    if ($self->error()) {
        $self->_set_err("Failed to collect SAN sysstats", $self->errmsg());
        return {};
    }

    # Save the sysstats raw output as part of the return
    $stats->{_raw_output} = @{$output}[0];

    # Example sysstat output
    #  CPU     FCP   iSCSI Partner  Total     FCP   kB/s   iSCSI   kB/s Partner   kB/s    Disk   kB/s    CP  CP  Disk
    #                                          in    out      in    out      in    out    read  write  time  ty  util
    #  16%       0       0       0      0       0      0       0      0       0      0      94  16729   32%  Zs   23%
    my @header = ("cpu", "fcp", "iscsi", "partner", "total", "fcp_in", "fcp_out", "iscsi_in",
                  "iscsi_out", "partner_in", "partner_out", "disk_read", "disk_write", "cp_time", "cp_ty", "disk_util");
    my @output = split("\n", @{$output}[0]);

    # Get rid of any empty lines
    @output = grep(!/^\s*$/, @output);

    # Get rid of the command line
    @output = grep(!/priv set/, @output);

    # Pull off the header
    @output = grep(!/^\s*CPU/, @output);
    @output = grep(!/\s*in\s+out/, @output);

    my $data_line = shift(@output);
    # Strip off white space at the beginning of the data
    $data_line =~ s/^\s*//g;
    my @data = split(/\s+/, $data_line);

    # Build out the stats hash mapping one-to-one header to data
    for (my $i = 0; $i < @header; $i++) {
        $data[$i] =~ s/\%//g;
        $stats->{$header[$i]} = $data[$i];
    }

    return $stats;
}

###############################################################################
# Method:    syslog_msg()
# Objective: syslog a message using the node scoped command 'logger'
# Details:     See POD documentation at the end of this file
###############################################################################
sub syslog_msg {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             msg  => {type => SCALAR,
                                                     required => 1},
                                             filer => {type => SCALAR,
                                                       optional => 1,
                                                       default => $self->get_name()}
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # run a node scoped command to log the specified message
    # command format: logger <string>
    # this will store a logged message in the following files on the given node:
    #    /mroot/etc/log/mlog/messages.log
    #    /mroot/etc/log/ems
    $self->run_node_scope(-cmd => "logger " . $args{msg},
                          -filer => $args{filer},
                          -priv => "diag");
    if ($self->error()) {
        $self->_set_err("Failed to syslog message $args{msg}", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    ntp_sync()
# Objective: NTP sync this node to the specified ntp server
# Details:     See POD documentation at the end of this file
###############################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             ntp_time_zone  => {type => SCALAR,
                                                                required => 1},
                                             ntp_server => {type => SCALAR,
                                                            required => 1},
                                             filer => {type => SCALAR,
                                                       optional => 1,
                                                       default => $self->get_name()}
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    my $node = $args{filer};
    my $server = $args{ntp_server};
    my $timezone = $args{ntp_time_zone};

    # Modify the time zone to the specified TZ
    %args = (timezone => $timezone, cmd => "timezone");
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj  \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Add the ntp server
    # build cdef to create the ntp server
    %args = (node => $node, server => $server, cmd => "system_services_ntp_server_create");
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        if ($self->errmsg() !~ /duplicate entry/i) {
            # set error and bubble up
            $self->_set_err("failed to invoke_cmd_obj  \"$args{cmd}\"", $self->errmsg());
            return $SanUtil::FALSE;
        } else {
            # allow duplicate entries
            $self->_clear_err();
        }
    }

    # enable NTP via system services
    %args = (enabled => "true", cmd => "system_services_ntp_config_modify");
    $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    $self->invoke(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj  \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: cf_disable
# Objective:   A wrapper to support 7-mode users. Disables the Storage FailOver
# Details:     See POD documentation at the end of this file.
###############################################################################
sub cf_disable {
    my ($self, %args) = @_;
    my ($ha_flag, $sfo_flag);

    # Clear the error flag
    $self->_clear_err();

    # Verify that the FailOver is enabled
    if ($self->is_sfo_enabled()){
          # Disable cluster HA
          $ha_flag = $self->cluster_ha_disable();
          if ($self->error()) {
              $self->_set_err("failed to disable the cluster HA before ".
                            "attempting to disable SFO using 7Mode wrapper.",
                            $self->errmsg());
              return $SanUtil::FALSE;
          }

          # Disable SFO
          $sfo_flag = $self->disable_sfo(%args);
          if ($self->error()) {
              $self->_set_err("failed to disable SFO using 7Mode wrapper.",
                            $self->errmsg());
              return $SanUtil::FALSE;
          }
     }

    return ($ha_flag&&$sfo_flag)? $SanUtil::TRUE : $SanUtil::FALSE;
} # end cf_disable

##############################################################################
#  P R I V A T E
##############################################################################

################################################################################
# Method Name: _get_fcp_object
# Objective:   A helper method to get an fcp_object for 7-mode callers
# Details:     See POD documentation at the end of this file
################################################################################
sub _get_fcp_object {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    my %tmp_args;

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|vserver)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # the key was a scalar
            $tmp_args{vserver} = $key;
        }
    }
    # set tmp_args to args
    %args = %tmp_args;

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to vserver to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{name}) {
        $args{vserver} = $args{name};
        delete $args{name};
    }

    # Use the default vserver as a last resort
    unless (defined $args{vserver} and length $args{vserver}) {
        $args{vserver} = $self->DEFAULT_VSERVER;
    }

    # discover and retrieve vserver object
    my $vs_obj = $self->get_vserver_obj(-name  => $args{vserver});
    if ($self->error() or ref($vs_obj) !~ /Vserver/) {
        $self->_set_err("can not find vserver: " . $args{vserver},
                        $self->errmsg());
        return {};
    }

    # Try to get the fcp object corresponding to this vserver.
    my $fcp_obj = $vs_obj->get_fcp_obj(%args);
    if ($vs_obj->error()) {
        $self->_set_err("can not find fcp object: ", $vs_obj->errmsg());
        return {};
    }

    return $fcp_obj;
}

################################################################################
# Method Name: _get_iscsi_object
# Objective:   A helper method to get an iscsi_object for 7-mode callers
# Details:     See POD documentation at the end of this file
################################################################################
sub _get_iscsi_object {
    my ($self, %args) = @_;

    # Clear error flag.
    $self->_clear_err();

    my %tmp_args;

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|vserver)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
        } else {
            # the key was a scalar
            $tmp_args{vserver} = $key;
        }
    }
    # set tmp_args to args
    %args = %tmp_args;

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to vserver to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{name}) {
        $args{vserver} = $args{name};
        delete $args{name};
    }

    # Use the default vserver as a last resort
    unless (defined $args{vserver} and length $args{vserver}) {
        $args{vserver} = $self->DEFAULT_VSERVER;
    }

    # discover and retrieve vserver object
    my $vs_obj = $self->get_vserver_obj(-name  => $args{vserver});
    if ($self->error() or ref($vs_obj) !~ /Vserver/) {
        $self->_set_err("can not find vserver: " . $args{vserver},
                        $self->errmsg());
        return {};
    }

    # Try to get the fcp object corresponding to this vserver.
    my $iscsi_obj = $vs_obj->get_iscsi_obj(%args);
    if ($vs_obj->error()) {
        $self->_set_err("can not find iscsi object: ", $vs_obj->errmsg());
        return {};
    }

    return $iscsi_obj;
} # End _get_iscsi_object


1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

FilerCompat.pm

=head1 Author

Korry Nguyen (dl-ssan-automation)

=head1 Description

Filer10M provides 7mode compatible and helper methods for Filer10M. This package maintains separately from Filer10M for maintainability purpose while preventing convoluting Filer10M object model.

=head1 See Also

FilerCompat package is used as part of the base class by B<San::ControllerCluster::Filer10M>. See these modules for further interface detail.

B<San::ControllerCluster::Filer10M.pm, San::CommonInterface.pm>

=head1 B<Public Methods>

=head2 B<aggr_create(%args)> (This method is 7mode compatible)

Create an aggregate.

=over 4

=item Inputs

=item * I<-name> (default: test_aggr)

The aggregate name to use

=item * I<-num_disks> (default: 3)

The number of disks in the aggregate

=item * I<-timeout> (default: 3600)

How long to wait for aggregate to come online

=item * I<-raidtype> (default: raid_dp)

The raidtype of the aggregate, can be raid4 or raid_dp

=item * I<-raidsize> (default: default ONTAP raidsize)

The raidsize of the aggregate, can be between 2 and 28

=item Outputs

Returns true if the aggregate create was successful, false otherwise.

=item Examples

 # Create aggregate (one-to-one 7mode compatible)
 $Filer->aggr_create(-name => test_aggr,
                     -blocktype => 64,
                     -num_disks => 5);
 if ($Filer->error){
     ...
 }

=back

=head2 B<aggr_destroy($name, $force)> This method is 7mode compatible

Destroy an aggregate

=over 4

=item Inputs

=item * I<name> (required)

Aggregate name to destroy

=item * I<force> <0|1> default is 1

Force the destroy

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Destroy an aggregate
 $filer_obj->aggr_destroy("test_aggr");
 if ($filer_obj->error()) {
     ...
 }

=back

=head2 B<aggr_exist($name)> This method is 7mode compatible

Determine whether an aggregate exists or not

=over 4

=item Inputs

=item * I<name> (required)

Aggregate name to use

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Determine if aggregate exists
 if ($filer_obj->aggr_exist("test_aggr")) {
     # get the aggregate
 } else {
     # create the aggregate
 }

=back

=head2 B<aggr_offline($name)> This method is 7mode compatible

Offline an aggregate

=over 4

=item Inputs

=item * I<name> (required)

Aggregate name to offline

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Offline an aggregate
 $filer_obj->aggr_offline("test_aggr");
 if ($filer_obj->error()) {
     ...
 }

=back

=head2 B<connection($value)> This method is 7mode compatible

Get or Set the connection attribute.  If a value is given,
the value is set.  The value is always returned.

=over 4

=item Inputs

=item * I<value> (optional)

connectrec object

=item Outputs

Returns connectrec object

=item Examples

 # ex. 1: Set connection
 my $filer_conn = connect($FILER_CONN);
 $filer_obj->connection($filer_conn);
 if ($filer_obj->error()) {
     ...
 }

 # ex. 2: Get connection
 $conn_obj = $filer_obj->connection($filer_conn);
 if ($filer_obj->error()) {
     ...
 }

=back

=head2 B<create_igroup(%args)> This method is 7mode compatible

Create an igroup object

=over 4

=item Inputs

=item * I<-name> (required)

Igroup name to create

=item * I<-initObjs> (optional)

An arrayref of initiator objects or a single initiator object

=item Outputs

Returns an igroup object. If error, error message is pushed onto the error stack.

=item Examples

 # ex. 1: Create an igroup
 $filer_obj->create_igroup(-name => "my_igroup");
 if ($filer_obj->error()) {
     ...
 }

 # ex. 2: Create an igroup with initiator objects
 # Retrieve list of client initiator objects
 @initObjs=$client_obj->get_initiators();
 if ($client_obj->error){
        $Test->nlog($Client->errmsg);
        return $TCD::FAIL;
 }

 # Create igroup
 $filer_obj->create_igroup(-name => "my_igroup",
                           -initObjs => [@initObjs]);
 if ($filer_obj->error()) {
     ...
 }

=back

=head2 B<create_vol(%args)> This method is 7mode compatible

Create a volume object

=over 4

=item Inputs

=item * I<-name> (required)

Volume name to create

=item * I<-aggr> (required)

Aggregate that the volume will be created on

=item * I<-size> (required)

Volume size to be created

=item Outputs

Returns a volume object. If error, error message is pushed onto the error stack.

=item Examples

 # ex. 1: Create a volume
 $filer_obj->create_vol(-name => "my_vol",
                        -aggr => $aggr,
                        -size => "100g");
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<destroy_igroup(%args)> This method is 7mode compatible

Destroy an igroup object

=over 4

=item Inputs

=item * I<-name> (required)

Igroup name to destroy

=item * I<-force> <0|1> default is 1

Force the destroy

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Destroy an igroup (7mode compatibility)
 $filer_obj->destroy_igroup("my_igroup");
 if ($filer_obj->error()) {
     ...
 }

 # ex. 2: Destroy an igroup with hash
 $filer_obj->destroy_igroup(-name => "my_igroup");
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<destroy_volume(%args)>

Destroy a volume object

=over 4

=item Inputs

=item * I<-name> (required)

Volume name to destroy

=item * I<-force> <0|1> default is 1

Force the destroy

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Destroy a volume
 $filer_obj->destroy_volume(-name => "my_vol");
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<fcp_start(%args)> This method is 7mode compatible

Start fcp service

=over 4

=item Inputs

=item * I<-name | -vserver> (optional)

Vserver name to start fcp service. If not specified, fcp service will be stopped on the default vserver (vs0)

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Start fcp service (7mode compatibility)
 $filer_obj->fcp_start();
 if ($filer_obj->error()){
     ...
 }

 # ex. 2: Start fcp service on vserver for Cmode
 $filer_obj->fcp_start(-vserver => "my_vserver");
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<fcp_stop(%args)> This method is 7mode compatible

Stop fcp service

=over 4

=item Inputs

=item * I<-name | -vserver> (optional)

Vserver name to stop fcp service. If not specified, fcp service will be stopped on the default vserver (vs0)

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Stop fcp service (7mode compatibility)
 $filer_obj->fcp_stop();
 if ($filer_obj->error()){
     ...
 }

 # ex. 2: Stop fcp service on vserver for Cmode
 $filer_obj->fcp_stop(-vserver => "my_vserver");
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<fcp_adapter_online(%args)> This method is 7mode compatible

Online a fcp adapter

=over 4

=item Inputs

=item * I<-adapter> (required)

Adapter to be online

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Online fcp adapter (7mode compatibility)
 $filer_obj->fcp_adapter_online(-adapter => "0c");
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<fcp_wwpnalias_get(%args)> This method is 7mode compatible

Get WWPN alias

=over 4

=item Inputs

=item * I<-alias> (optional)

Getting WWPN that mapped to the specified alias

=item * I<-wwpn> (optional)

Get alias(es) that associated with specified wwpn

=item Outputs

Returns a hash. The hash contains {alias_name => wwpn_name}.
Failure or no existing alias will return empty hash.

 single alias output:
   alias_name => 10:00:00:00:c9:48:ce:9f

 multi aliases output:
   alias_name1 => 10:00:00:00:c9:48:ce:9f
   alias_name2 => 10:00:00:00:c9:48:ce:9f
   alias_name3 => 10:00:00:00:c9:48:ce:9f
   alias_name4 => 10:00:00:00:c9:48:ce:9f
   alias_name5 => 10:00:00:00:c9:48:ce:9f

=item Examples

 # ex. 1:  Get alias(es) that mapped to a wwpn (7mode compatibility)
 $hash_ref = $filer_obj->fcp_wwpnalias_get(-wwpn => $wwpn_name);
 if ($filer_obj->error())
 {
    ...
 }

=back

=head2 B<fcp_wwpnalias_remove(%args)> This method is 7mode compatible

Remove WWPN alias
If alias was given, remove that alias which mapped to a WWPN
If wwpn was given, remove alias(es) that mapped to the given WWPN

=over 4

=item Inputs

=item * I<-alias | -wwpn> (required either)

Name of alias or wwpn

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1:  Remove given alias that mapped to a wwpn (7mode compatibility)
 $filer_obj->fcp_wwpnalias_remove(-alias => $alias_name);
 if ($filer_obj->error())
 {
    ...
 }

=back

=head2 B<fcp_wwpnalias_set(%args)> This method is 7mode compatible

Create an alias and map to the given wwpn

=over 4

=item Inputs

=item * I<-alias> (required)

Name of alias

=item * I<-wwpn> (required)

Name of wwpn

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1:  Create an alias and map to a given wwpn (7mode compatibility)
 $filer_obj->fcp_wwpnalias_set(-alias => $alias_name,
                               -wwpn => $wwpn_name);
 if ($filer_obj->error())
 {
    ...
 }

=back

=head2 B<get_aggrs()> This method is 7mode compatible

Get a list of aggregate name

=over 4

=item Inputs

None

=item Outputs

Returns an array of aggregate names. If error, error message is pushed onto the error stack.

=item Examples

 # ex. 1:  Get a list of aggregate names (7mode compatibility)
 @aggr_names = $filer_obj->get_aggrs();
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_cfmode()> This method is 7mode compatible

Get cfmode. Clustered ONTAP has no such cfmode so this method is provided for 7mode compatibility

=over 4

=item Inputs

None

=item Outputs

Returns "single_image"

=item Examples

 # ex. 1:  Get aggregate objects (7mode compatibility)
 $cfmode = $filer_obj->get_cfmode();

=back

=head2 B<get_fcp_nodename(%args)> This method is 7mode compatible

Get fcp nodename

=over 4

=item Inputs

=item * I<-name | -vserver> (optional)

Vserver name to get fcp nodename from. If not specified, the nodename will be retrieved from the default vserver (vs0)

=item Outputs

Returns the fcp nodename. If error, error message is pushed onto the error stack and return empty string.

=item Examples

 # ex. 1:  Get fcp nodename (7mode compatibility)
 $nodename = $filer_obj->get_fcp_nodename();
 if ($filer_obj->error()) {
    ...
 }

 # ex. 2:  Get fcp nodename on specific vserver
 $nodename = $filer_obj->get_fcp_nodename(-vserver => $vs_name);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_fcp_status(%args)> This method is 7mode compatible

Get fcp status

=over 4

=item Inputs

=item * I<-name | -vserver> (optional)

Vserver name to get fcp status from. If not specified, the status will be retrieved from the default vserver (vs0)

=item Outputs

Returns the fcp status. If error, error message is pushed onto the error stack and
 return empty string.

=item Examples

 # ex. 1:  Get fcp status (7mode compatibility)
 $status = $filer_obj->get_fcp_status();
 if ($filer_obj->error()) {
    ...
 }

 # ex. 2:  Get fcp status on specific vserver
 $status = $filer_obj->get_fcp_status(-vserver => $vs_name);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_fcp_targets(%args)>

=head2 B<get_targets(%args)> This method is 7mode compatible

Get fcp target objects

=over 4

=item Inputs

=item * I<-name | -vserver> (optional)

Vserver name to get fcp targets from. If not specified, targets will be retrieved from the default vserver (vs0)

=item Outputs

Returns an array of fcp target objects. If error, error message is pushed onto the error stack and return empty array.

=item Examples

 # ex. 1:  Get fcp target objects (7mode compatibility)
 @target_objs = $filer_obj->get_targets();
 if ($filer_obj->error()) {
    ...
 }

 # ex. 2:  Get fcp target objects on specific vserver
 @target_objs = $filer_obj->get_fcp_targets(-vserver => $vs_name);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_igroup(%args)> This method is 7mode compatible

Get igroup object

=over 4

=item Inputs

=item * I<-name | -igroup> (required)

Igroup name to retrieve

=item * I<-vserver> (optional)

Vserver name to get igroup object from. If not specified, targets will be retrieved from the default vserver (vs0)

=item Outputs

Returns igroup object. If error, error message is pushed onto the error stack and return empty hashref.

=item Examples

 # ex. 1:  Get igroup objects (7mode compatibility)
 $igrp_obj = $filer_obj->get_igroup();
 if ($filer_obj->error()) {
    ...
 }

 # ex. 2:  Get igroup objects on specific vserver
 $igrp_obj = $filer_obj->get_igroup(-vserver => $vs_name);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_igroups(%args)> This method is 7mode compatible

Get igroup objects

=over 4

=item Inputs

=item * I<-vserver> (optional)

Vserver name to get igroup objects from. If not specified, targets will be retrieved from the default vserver (vs0)

=item Outputs

Returns an array of igroup objects. If error, error message is pushed onto the error stack and return empty array.

=item Examples

 # ex. 1:  Get igroup objects (7mode compatibility)
 @igrp_objs = $filer_obj->get_igroups();
 if ($filer_obj->error()) {
    ...
 }

 # ex. 2:  Get igroup objects on specific vserver
 @igrp_objs = $filer_obj->get_igroups(-vserver => $vs_name);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_iscsi_targets(%args)> This method is 7mode compatible

Get iscsi target objects

=over 4

=item Inputs

=item * I<-state> (optional)

State of iscsi target. Default is "up"

=item * I<-valid_ip> (optional)

Valid IP flag. Default is "1"

=item * I<-name | -vserver> (optional)

Vserver name to get iscsi targets from. If not specified, targets will be retrieved from the default vserver (vs0)

=item Outputs

Returns an array of iscsi target objects. If error, error message is pushed onto the error stack and return empty array.

=item Examples

 # ex. 1:  Get iscsi target objs (7mode compatibility)
 @target_objs = $filer_obj->get_iscsi_targets();
 if ($filer_obj->error()) {
    ...
 }

 # ex. 2:  Get iscsi target objects on specific vserver
 @target_objs = $filer_obj->get_iscsi_targets(-vserver => $vs_name);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_vol(%args)> This method is 7mode compatible

Get volume object

=over 4

=item Inputs

=item * I<-name | -volume> (required)

Volume name to retrieve

=item * I<-vserver> (optional)

Vserver name to get volume object from. If not specified, portsets will be retrieved from the default vserver (vs0)

=item Outputs

Returns volume object. If error, error message is pushed onto the error stack and return empty hashref.

=item Examples

 # ex. 1:  Get volume object (7mode compatibility)
 $vol_obj = $filer_obj->get_vol($vol_name);
 if ($filer_obj->error()) {
    ...
 }

 # ex. 2:  Get volume object on specific vserver
 $vol_obj = $filer_obj->get_vol(-name => $vol_name,
                                -vserver => $vs_name);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<igroup_exist($name)> This method is 7mode compatible

Determine whether an igroup exists or not

=over 4

=item Inputs

=item * I<name> (required)

Igroup name to use

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Determine if igroup exists
 if ($filer_obj->igroup_exist("my_igroup")) {
     # get igroup object
 } else {
     # create igroup object
 }

=back

=head2 B<iscsi_start(%args)> This method is 7mode compatible

Start iscsi service

=over 4

=item Inputs

=item * I<-name | -vserver> (optional)

Vserver name to start iscsi service. If not specified, iscsi service will be stopped on the default vserver (vs0)

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Start iscsi service (7mode compatibility)
 $filer_obj->iscsi_start();
 if ($filer_obj->error()){
     ...
 }

 # ex. 2: Start iscsi service on vserver for Cmode
 $filer_obj->iscsi_start(-vserver => "my_vserver");
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<iscsi_stop(%args)> This method is 7mode compatible

Stop iscsi service

=over 4

=item Inputs

=item * I<-name | -vserver> (optional)

Vserver name to stop iscsi service. If not specified, iscsi service will be stopped on the default vserver (vs0)

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Stop iscsi service (7mode compatibility)
 $filer_obj->iscsi_stop();
 if ($filer_obj->error()){
     ...
 }

 # ex. 2: Stop iscsi service on vserver for Cmode
 $filer_obj->iscsi_stop(-vserver => "my_vserver");
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<lun_exist($name)> This method is 7mode compatible

Determine whether a lun exists or not

=over 4

=item Inputs

=item * I<name> (required)

Lun name to use

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Determine if lun exists
 if ($filer_obj->lun_exist("my_lun")) {
     # get lun object
 } else {
     # create lun object
 }

=back

=head2 B<get_lun_minsize> This method is 7mode compatible

Get the lun minsize for a particular ostype

=over 4

=item Inputs

=item * I<ostype> (required)

Ostype to get minimum size for

=item Outputs

Returns minimum size if successful, otherwise sets error and returns FALSE

=item Examples

   my $min_size = $San_Filer_Obj->get_lun_minsize(-type => $os_types[$i]);

=back

=head2 B<reconnect()> This method is 7mode compatible

Reconnect to the controller and store connection

=over 4

=item Inputs

None

=item Outputs

Returns connection object. If error, error message is pushed onto error stack and returns empty hashref

=item Examples

 # ex. 1: Reconnect to controller
 $conn_obj = $filer_obj->reconnect();
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<set_fcp_nodename()> This method is 7mode compatible

Set fcp nodename

=over 4

=item Inputs

=item * I<-name | -nodename> (required)

Fcp nodename to be set

=item * I<-vserver> (optional)

Vserver name to get fcp nodename from. If not specified, the nodename will be retrieved from the default vserver (vs0)

=item Outputs

Returns the fcp nodename. If error, error message is pushed onto the error stack and return empty string.

=item Examples

 # ex. 1:  Set fcp nodename (7mode compatibility)
 $filer_obj->set_fcp_nodename($nodename);
 if ($filer_obj->error()) {
    ...
 }

 # ex. 2:  Set fcp nodename on specific vserver
 $filer_obj->set_fcp_nodename(-nodename => $nodename,
                              -vserver => $vs_name);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<vol_exist()> This method is 7mode compatible

Determine whether the volume exists or not

=over 4

=item Inputs

=item * I<-name | -volume> (required)

Volume name to use

=item * I<-vserver> (optional)

Vserver name to get fcp nodename from. If not specified, the nodename will be retrieved from the default vserver (vs0)

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Determine if volume exists (7mode compatibility)
 if ($filer_obj->vol_exist("my_vol")) {
     # get vol object
 } else {
     # create vol object
 }

 # ex. 2: Determine if volume exists on a specified vserver (7mode style)
 if ($filer_obj->vol_exist("my_vol", "my_vserver")) {
     # get vol object
 } else {
     # create vol object
 }

=back

=head2 B<vol_status()> This method is 7mode compatible

Get the status for the specified volume

=over 4

=item Inputs

=item * I<-name | -volume> (required)

Volume name to use

=item Outputs

Returns array containing output from volume show <name> if successful,
otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Determine if volume exists (7mode compatibility)
 @output = $filer_obj->vol_status("my_vol")
 if ("@output" !~ m/trad/i) {
    $Test->nlog("Flex volume already exists with $VolName, please remove and rerun ");

=back

=head2 B<get_volnames()> This method is 7mode compatible

Get the names of the Volumes on the system

=over 4

=item Inputs

None

=item Outputs

Returns array containing the names of the Volumes on the system
otherwise sets error and returns an empty array

=item Examples

    my @Vol_Names = $filer_obj->get_volnames();
    if ( $file_obj->error ) {
        $Test->nlog($filer_obj->errmsg());
        return $TCD::CONFIG;
    }

    foreach my $name (@Vol_Names) {
        if ($name =~ /san_vol/) {
            $Vol_Name=$name;
            $Test->nlog("Volume Name = $Vol_Name");
        }
    }

=back

=head2 B<wait_for_aggr_online(%args)> This method is 7mode compatible

Wait for aggregate to be online

=over 4

=item Inputs

=item * I<-name | -aggregate> (required)

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Create aggregate
 $aggr_obj = $filer_obj->aggr_create(-name => $aggr_name);
 if ($filer_obj->error()) {
    ...
 }

 # Wait for aggregate to go online
 $filer_obj->wait_for_aggr_online(-name => $aggr_name);
 if ($filer_obj->error()) {
     ...
 }

=back

=head2 B<get_partner()> This method is 7mode compatible

Obtain the HA partner name of this node

=over 4

=item Inputs

=item * (same inputs into $node->sfo_show()), however, -node defaults to the current node that is being invoked.

=item Outputs

Returns the partner name if successful, "" if not and sets error.

=item Examples

 my $partner_name = $node->get_partner();
 if ($node->error()) {
     ...
 }

=back

=head2 B<load_sram_dump_list()> This method is 7mode compatible

Query the system for the current SRAM dump list and store it in memory. (requires freebsd access)

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE if successful, $SanUtil::FALSE if not and sets error.

=item Examples

 # load the current sram dump list into memory
 $node->load_sram_dump_list();
 if ($node->error()) {
     ...
 }

=back

=head2 B<get_sram_dump_list()> This method is 7mode compatible

Obtain the current SRAM dump list that is stored in memory.

=over 4

=item Inputs

None

=item Outputs

Returns an array or array reference of sram dump files

=item Examples

 # get the sram list that is stored in memory
 my (@sram_list) = $node->get_sram_dump_list();
 if ($node->error()) {
     ...
 }

=back

=head2 B<set_sram_dump_list(@array)> This method is 7mode compatible

Modify the current SRAM dump list that is stored in memory.

=over 4

=item Inputs

=item * I<ARRAY>

The new sram dump list to store in memory.

=item Outputs

returns $SanUtil::TRUE

=item Examples

 # modify the sram dump list in memory
 $node->set_sram_dump_list(@new_sram_list);

=back

=head2 B<syslog_msg(%args)> This method is 7mode compatible

Write a message to the syslogs & ems logs on the specified node

=over 4

=item Inputs

=item * I<-msg>

Required: The message to log to the syslogs.

=item * I<-filer>

Optional: The node to log the message on.
Default: The current node.

=item Outputs

returns $SanUtil::TRUE if successful, $SanUtil::FALSE if not and sets error.

=item Examples

 # log message on this node
 $node->syslog_msg(-msg => "Performing takeover on this node");
 if ($node->error()) {

 }

 # log message on the HA partner node
 $node->syslog_msg(-msg => "The other node is taking over this node",
                   -filer => "partnernode");
 if ($node->error()) {

 }

=back

=head2 B<ntp_sync(%args)> This method is 7mode compatible

NTP Sync the controller to a specified time server and set the time zone

=over 4

=item Inputs

=item * I<-filer>

Optional: The node to ntp sync
Default: The current node.


=item * I<-ntp_server>

Required: The message to log to the syslogs.
Example: ntp0.eng.netapp.com

=item * I<-ntp_time_zone>

Required: The timezone to use.
Example: EST5EDT

=item Outputs

returns $SanUtil::TRUE if successful, $SanUtil::FALSE if not and sets error.

=item Examples

 # ntp time synchronize this node
 $node->ntp_sync(-ntp_server => "ntp0.eng.netapp.com",
                 -ntp_time_zone => "EST5EDT");
 if ($node->error()) {

 }

=back

=head2 B<sysstat(%args)> This method is 7mode compatible

Collect Sysstats over a specified interval

=over 4

=item Inputs

=item * I<-filer>

Optional: The node to perform the operation against
Default: The current node

=item * I<-interval>

Optional: The interval period to collect stats over
Default: 10 (seconds)

=item * I<-count>

 Optional: The number of samples to collect
Default: 1

=item * I<-pass_thru>

Optional: Additional parameters to pass into the sysstat command
Default: "-b"

=item Outputs

returns a hash containing the parsed information from the output.
Note, if anything other then the default is used, parsing of the output may not work.
This will leave only the raw output available within the returned hash.

=item Examples

 # Collect three 5-second samples of SAN data on the current node
 my $stats = $node->sysstat(-count => 3,
                            -interval => 5,
                            -pass_thru => "-b");
 if ($node->error()) {
 }

 my @output = @{$stats->{_raw_output}};

=back

=head2 B<cf_disable()> This method is 7mode compatible

Disables the StorageFailOver (konw in 7Mode as Cluster Failover) when enabled. Also, disconfigure the HA pair before disabling the SFO.

=over 4

=item Inputs

None

=item Outputs

returns $SanUtil::TRUE if successful, $SanUtil::FALSE if not and sets error

=item Examples

 # disable the SFO and HA on a cluster pair
 my $enable_flag = $node->cf_disable();
 if ($node->error()) {
     ...
 }

=back

=cut
