# Copyright (c) 2009-2010 NetApp Inc.
# All rights reserved.
#
## @summary provides helper methods for Filer10M class
##
## @author korry@netapp.com, dl-ssan-automation
## @status Shared

package San::ControllerCluster::FilerHelper;

# compiler directives
use strict;
use lib();

# imported modules;
use Params::Validate qw( SCALAR ARRAYREF HASHREF OBJECT );
use Time::HiRes qw( gettimeofday );
use boot;
use Data::Dumper;
use San::ControllerCluster::Filer10M::Vserver::Volume::Lun;
use NACL::Transit;
use NATE::BaseException qw (:try);

# Constants
use constant MAX_ZAPI_RECORDS => 1024;

# inherit base class
use base qw( San::CommonInterface );

# global variables
my $Return_index = 3;

################################################################################
# Method Name: aggregate_create()
# Objective:   create a new aggregate.
# Details:     See POD documentation at the end of this file
################################################################################
sub aggregate_create {
    my ($self, %args) = @_;

    my ($aggr_obj, @disk_objs, %validate_params);
    my $use_disk_objs = $SanUtil::FALSE;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to aggregate to maintain consistency
    if (exists $args{name}) {
        $args{aggregate} = delete $args{name};
    }
    if (exists $args{num_disks}) {
        $args{diskcount} = delete $args{num_disks};
    }

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);

    # validation_options is set to allow extra params not defined in %args to bypass
    Params::Validate::validation_options( allow_extra => 1);

    # TODO - we need to handle larger aggr and balance raidgroup
    # via interlace disks

    # validate input parameters and their type.
    # aggregate - aggr name (default = test_aggr)
    # num_disks - number of disks (default = 3)
    # raidtype - raid type (default = raid_dp)
    # timeout - default = 3600
    # validate_params is the arguments to validate
    %validate_params = (
                      aggregate      => { type     => SCALAR,
                                          optional => 1,
                                          default  => "test_aggr"},
                      diskcount      => {type      => SCALAR,
                                         optional  => 1,
                                         default   => 3,
                                         regex     => qr/\d+/},
                      raidtype       => {type      => SCALAR,
                                         optional  => 1,
                                         default   => "raid_dp"},
                      use_disk_objs  => {type      => SCALAR,
                                         optional  => 1},
                      disk_owner     => {type      => SCALAR,
                                         optional  => 1},
                      timeout        => {type      => SCALAR,
                                         optional  => 1,
                                         default   => 3600,
                                         regex     => qr /^\d+$/ }
                      );
    eval {
        (%args) = Params::Validate::validate(@input_params, \%validate_params);
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return {};
    }

    # if disk_owner is passed in, we need to get spare disks from the specified disk_owner
    # otherwise, we just use diskcount
    if (exists $args{use_disk_objs}) { $use_disk_objs = delete $args{use_disk_objs}; }

    if ($use_disk_objs) {
        # discover all spares disks in the system
        # retrieve number of spare disks
        @disk_objs = $self->get_disk_objs(-state => 'spare',
                                          -owner => $args{disk_owner},
                                          -count => $args{diskcount});
        if ($self->error()) {
            $self->_set_err("failed to get spare disks ", $self->errmsg());
            return {};
        }
        # set diskobj and remove diskcount args
        $args{disk_objs} = [@disk_objs];
        delete $args{diskcount};
    }

    # timeout is not being used at this time, so delete it
    delete $args{timeout};

    # disable required field check if it's set
    $self->disable_req_field_check();

    # Create aggregate object
    $aggr_obj = $self->create_aggregate_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to create aggr: " . $args{aggregate} . " ", $self->errmsg());
        return {};
    }

    # enable required field check
    $self->enable_req_field_check();

    return $aggr_obj;

} ## end sub aggregate_create

################################################################################
# Method Name: aggregate_offline()
# Objective:   Take an aggregate ofline
# Details:     See POD documentation at the end of this file
################################################################################
sub aggregate_offline {
    my ($self, %args) = @_;

    # this is a stub for compatibility with 7-mode
    # 10-mode aggr does not require to be offlined
    # clear the error here.
    $self->_clear_err();
    return $SanUtil::TRUE;
}

################################################################################
# Method Name: create_cluster_lifs()
# Objective:   Create the cluster lifs for a node
# Details:     See POD documentation at the end of this file
################################################################################
sub configure_cluster_lifs {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $num_clus_lifs = 2;
    my $host = &Tharn::host($self->get_name());

    for (my $lif_cnt = 1; $lif_cnt <= $num_clus_lifs; $lif_cnt++) {
        # Build lif args from hostrec hash.
        my %lif_args = (
            vserver   => $self->get_name(),
            lif       => 'clus' . $lif_cnt,
            role      => 'cluster',
            home_node => $self->get_name(),
            home_port => $host->{'CLUSTER_PORT' . $lif_cnt},
            address   => $host->{'CLUSTER_IP' . $lif_cnt},
            netmask   => $host->{'CLUSTER_MASK' . $lif_cnt},
            gateway   => $host->{'CLUSTER_GATEWAY' . $lif_cnt}
        );

        $self->configure_lif(%lif_args);
        if ($self->error()) {
            # Error set return false
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} ## end sub configure_cluster_lifs

###############################################################################
# Method Name: configure_lif
# Objective: Generic configuration of a network interface for a node/filer.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub configure_lif {
    my ($self, %args) = @_;

    # Locals
    my @lif_objs = ();
    my $lif_obj = undef;
    my %route_args = ();

    # Clear error
    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    # Get parameters
    if ( defined($args{gateway}) ) {
        $route_args{gateway} = delete $args{gateway};
    }
    if ( defined($args{destination}) ) {
        $route_args{destination} = delete $args{destination};
    }

    # Build lif args from remaining args
    my %lif_args = %args;

    # Done here if the cluster lif already exists.
    @lif_objs = $self->fetch_lif_objs(-lif     => $args{lif},
                                      -vserver => $args{vserver});
    if ($self->error()) {
        # Error is set, bubble up
        return $SanUtil::FALSE;
    }
    if (scalar(@lif_objs)) {
        # There can be only one.
        $lif_obj = $lif_objs[0];

        # Validate that the args we need match those we got.
        foreach my $arg (keys(%lif_args)) {
            unless (exists($lif_obj->{$arg}) and ($lif_obj->{$arg} eq $lif_args{$arg})) {
                $self->_set_err(
                    "The lif $args{lif} already exists on vserver "
                    . $self->get_name() ."but $arg differs: "
                    . "Actual: " . $lif_obj->{$arg}
                    . "Requested: " . $lif_args{$arg}
                );
            }
        }
        if ($self->error()) {
            return $SanUtil::FALSE;
        }
        $self->sanlog(-msg       => "cluster lif $args{lif} exists, reuse..",
                      -msg_level => $SanLog::LOG_QUIET);
    } else {
        # Create a lif object to use later.
        $lif_obj = San::ControllerCluster::Filer10M::Vserver::Lif->new(
            -parent_obj  => $self,
            -filer       => $self->get_name(),
            -initialized => 1,
            %lif_args
        );
        if ($lif_obj->error()) {
            $self->_set_err(
                "failed to instantiate lif object", $lif_obj->errmsg()
            );
            return $SanUtil::FALSE;
        }

        # Add the command to the lif args
        $lif_args{cmd} = "network_interface_create";

        # use common invoke from base class (San::CommonInterface)
        # to translate cdef into UI cmd and execute it
        $lif_obj->San::CommonInterface::invoke_cmd_obj(%lif_args);
        if ($lif_obj->error()) {
            # set error and bubble up
            $self->_set_err(
                "failed to invoke_cmd_obj " . $args{cmd}
                . $lif_obj->errmsg()
            );
            return $SanUtil::FALSE;
        }
    }

    # Create cserver lif route...
    $lif_obj->create_route( %route_args );
    if ($lif_obj->error()) {
        # set error and bubble up
        $self->_set_err(
            "failed to create lif route" . $lif_obj->errmsg()
        );
        return $SanUtil::FALSE;
    }

}# end sub configure_lif

################################################################################
# Method Name: create_volume()
# Objective:   create and return a volume object
# Details:     See POD documentation at the end of this file
################################################################################
sub create_volume {
    my ($self, %args) = @_;

    my ($aggr_obj, $vol_obj, $vs_obj);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert aggr to aggregate to maintain compatability w/ 7mode
    if (exists $args{aggr}) {
        $args{aggregate} = delete $args{aggr};
    }
    # convert name to volume to maintain consistency
    if (exists $args{name}) {
        $args{volume} = delete $args{name};
    }

    # validation_options is set to allow extra params not defined in %args to bypass
    Params::Validate::validation_options( allow_extra => 1);

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);
    eval {
        (%args) = Params::Validate::validate(@input_params,
                                            {volume => {type => SCALAR, required => 1},
                                             aggregate => {type => SCALAR, required => 1},
                                             "junction-path" => {type => SCALAR, optional=> 1},
                                             "space-guarantee" => {type => SCALAR, optional=> 1},
                                             "snapshot-policy" => {type => SCALAR, optional=> 1},
                                             type => {type => SCALAR, optional => 1},
                                             state => {type => SCALAR, optional => 1},
                                             vserver => {type => SCALAR,
                                                         optional => 1,
                                                         default => $self->DEFAULT_VSERVER},
                                             size => {type => SCALAR,
                                                      required => 1,
                                                      regex => qr/\d+[kmgt]?/ },
                                             verify_vol => {type => SCALAR,
                                                            optional => 1,
                                                            default => 1},
                                             vserver_obj => {optional => 1,
                                                             default => undef}});
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return {};
    }

    # remove the vserver object from the input
    $vs_obj = delete($args{vserver_obj});

    # check if user wants us to skip the verification step
    my $verify_vol = delete($args{verify_vol});

    # strip /vol/ from volume name
    $args{volume} =~ s/\/vol\/// if ($args{volume} =~ /^\/vol\//);

    # if a vserver was passed in, don't discover it again
    if (not defined($vs_obj) or ref($vs_obj) !~ /Vserver/i) {
        # korry - discover vservers and then get vserver on specified aggregate
        # perhaps we need to discover aggregate first, then vserver.
        # verify vserver is contained from the matching aggregate
        #
        # make sure we have an aggregate. if not create one.
        if ($self->vserver_exist(-name => $args{vserver})) {
            # retrieve vserver object
            $vs_obj = $self->get_vserver_obj(-name  => $args{vserver});
            if ($self->error()) {
                $self->_set_err("can not get vserver " . $args{vserver}, $self->errmsg());
                return {};
            }
        } else {
            # create vserver object
            $vs_obj = $self->create_vserver_obj(-vserver   => $args{vserver},
                                                -aggregate => $args{aggregate});
            if ($self->error()) {
                $self->_set_err("can not create vserver " . $args{vserver}, $self->errmsg());
                return {};
            }
        }
    }

    # only do this if the user asked us to verify the volume create process
    if ($verify_vol) {
        # does the volume exist in vserver?
        if ($self->volume_exist(-name => $args{volume},
                                -vserver => $args{vserver})) {
            $self->sanlog(-msg => "volume $args{volume} exists. get vol object.",
                          -msg_level => 0);
            $vol_obj = $self->get_vol(-name => $args{volume},
                                      -vserver => $args{vserver});
            if ($self->error()) {
                $self->_set_err("can not get volume " . $args{volume}, $self->errmsg());
                return {};
            }
            # is the volume was created on the aggregate we wanted?
            # if so, it is the one we wanted
            if ($vol_obj->get_aggregate() eq $args{aggregate}) {
                $self->sanlog(-msg => "found vol obj: " . $args{volume},
                              -msg_level => 0);
                return $vol_obj;
            } else {
                # NOTE: we need to add policy to ensure we get the exact volume
                # we asked for (size, type, etc.)
                # the existing volume is on the wrong aggregate
                # delete it
                $self->destroy_volume(-name => $args{volume},
                                      -vserver => $args{vserver});
                if ($self->error()) {
                    $self->_set_err("can not destroy volume " . $args{volume},
                                    $self->errmsg());
                    return {};
                }
            }
        }
        # if volume does not exist, it could left behind an error message, clear it now
        $self->_clear_err();
    }

    # Deleting vserver from the args because we already have a vserver object
    delete $args{vserver};

    # we got here either because volume doesn't exist or volume is on a different
    # aggregate than we asked for.
    # create a volume object on the aggregate we want
    $vol_obj = $vs_obj->create_volume_obj(%args);
    if ($vs_obj->error()) {
        $self->_set_err("fail to create volume object " . $args{volume}, $vs_obj->errmsg());
        return {};
    }

    return $vol_obj;
} ## end sub create_volume

################################################################################
# Method Name: get_aggregate()
# Objective:   Get aggregate object.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_aggregate {
    my ($self, %args) = @_;

    # wrapper for get_aggregate_obj
    return $self->get_aggregate_obj(%args);

} ## end sub get_aggregate

################################################################################
# Method Name: get_aggregates()
# Objective:   Get aggregate from cache or discover.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_aggregates {
    my ($self, %args) = @_;

    my (@aggregate_objs);

    $self->_clear_err();

    # if aggregate objects has been cached, get it
    # otherwise, call get_aggregate_objs
    if (defined $self->{aggregate_objs} and scalar(@{$self->{aggregate_objs}})) {
        @aggregate_objs = @{$self->{aggregate_objs}};
    } else {
        # discover and get aggregate objects
        @aggregate_objs = $self->get_aggregate_objs();
        if ($self->error()) {
            $self->_set_err("failed to get aggregate objects", $self->errmsg());
            return (wantarray()) ? (): [];
        }
    }
    return (wantarray()) ? @aggregate_objs: \@aggregate_objs;
} ## end sub get_aggregates

###############################################################################
# Method Name: get_fcpadapter_objs()
# Objective:   Get fcpadapter objects.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_fcpadapter_objs {
    my ($self, %args) = @_;

    my ($hwresource_obj, @fcpadapter_objs);

    $self->_clear_err();

    $hwresource_obj = $self->{hwresource_obj} if exists ($self->{hwresource_obj});
    @fcpadapter_objs = $hwresource_obj->get_fcpadapter_objs(%args);
    if ($hwresource_obj->error()) {
        $self->_set_err("failed to get fcpadapter objects", $hwresource_obj->errmsg());
        return (wantarray()) ? (): [];
    }
    return (wantarray()) ? @fcpadapter_objs: \@fcpadapter_objs;
} ## end sub get_fcpadapter_objs

###############################################################################
# Method Name: get_ip_addresses
# Objective:   Get ip addresses of the filer.
# Details:     It is a wrapper around the 'get_network_interfaces' method
###############################################################################
sub get_ip_addresses {
    my ($self, %args) = @_;
    my (@ip_records, @ip_addresses);
    $self->_clear_err();

    # Call existing function.
    @ip_records = $self->get_network_interfaces(%args);
    # Catch any error.
    if ($self->error()) {
        $self->_set_err("failed to get IP addresses ", $self->errmsg());
        return (wantarray()) ? (): [];
    }
    # Get IPs from records and dump them to an array.
    foreach my $record (@ip_records){
        push(@ip_addresses, $record->{address});
    }
    return (wantarray()) ? @ip_addresses: \@ip_addresses;
} ## end get_ip_addresses

###############################################################################
# Method Name: get_network_interfaces
# Objective:   Get network interface information on a filer
# Details:     See POD documentation at the end of this file.
###############################################################################
sub get_network_interfaces {
    my ($self, %args) = @_;

    my @network_records;

    $self->_clear_err();

    # initialize cmd parameters
    %args = (%args, cmd => "network_interface_show");

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    # use common invoke to translate cdef into UI cmd and execute it
    @network_records = $self->SUPER::invoke_cmd_obj(%args);

    # Set the error just in case something fails
    if ($self->error()) {
        $self->_set_err("Failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return (wantarray()) ? (): [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@network_records) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @network_records: \@network_records;
} ## end get_network_interfaces

################################################################################
# Method Name: get_portsets()
# Objective:   get a list of available portset objects
# Details:     See POD documentation at the end of this file
################################################################################
sub get_portsets {
    my ($self, %args) = @_;

    my (@portset_objs, $vs_obj);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # set default vserver if it's not specified
    $args{vserver} = $self->DEFAULT_VSERVER unless (defined $args{vserver});

    # discover and retrieve vserver object
    $vs_obj = $self->get_vserver_obj(-name  => $args{vserver});
    if ($self->error() or ref($vs_obj) !~ /Vserver/) {
        $self->_set_err("can not find vserver: " . $args{vserver},
                        $self->errmsg());
        return (wantarray()) ? (): [];
    }

    # if portset objects has been cached, get it
    # otherwise, call get_portset_objs
    if (defined $vs_obj->{portset_objs} and scalar(@{$vs_obj->{portset_objs}})) {
        @portset_objs = @{$vs_obj->{portset_objs}};
    } else {
        # discover and get portset objects
        @portset_objs = $vs_obj->get_portset_objs();
        if ($vs_obj->error()) {
            $self->_set_err("failed to get portset objects", $vs_obj->errmsg());
            return (wantarray()) ? (): [];
        }
    }
    return (wantarray()) ? @portset_objs: \@portset_objs;
} ## end sub get_portsets

################################################################################
# Method Name: get_volumes()
# Objective:   Get volume objects.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_volumes {
    my ($self, %args) = @_;

    my ($vs_obj, @volume_objs);

    $self->_clear_err();

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # set default vserver if it's not specified
    $args{vserver} = $self->DEFAULT_VSERVER unless (defined $args{vserver});

    # discover and retrieve vserver object
    $vs_obj = $self->get_vserver_obj(-name  => $args{vserver});
    if ($self->error() or ref($vs_obj) !~ /Vserver/) {
        $self->_set_err("can not find vserver: " . $args{vserver},
                        $self->errmsg());
        return (wantarray()) ? (): [];
    }

    # if volume objects has been cached, get it
    # otherwise, call get_volume_objs
    if (defined $vs_obj->{volume_objs} and scalar(@{$vs_obj->{volume_objs}})) {
        @volume_objs = @{$vs_obj->{volume_objs}};
    } else {
        # discover and get volume objects
        @volume_objs = $vs_obj->get_volume_objs();
        if ($vs_obj->error()) {
            $self->_set_err("failed to get volume objects", $vs_obj->errmsg());
            return (wantarray()) ? (): [];
        }
    }
    return (wantarray()) ? @volume_objs: \@volume_objs;
} ## end sub get_volumes

################################################################################
# Method Name: get_vservers()
# Objective:   Get vserver objects.
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vservers {
    my ($self) = @_;

    my (@vserver_objs);

    $self->_clear_err();

    # if vserver objects has been cached, get it
    # otherwise, call get_vserver_objs
    if (defined $self->{vserver_objs} and scalar(@{$self->{vserver_objs}})) {
        @vserver_objs = @{$self->{vserver_objs}};
    } else {
        # discover and get vserver objects
        @vserver_objs = $self->get_vserver_objs();
        if ($self->error()) {
            $self->_set_err("failed to get vserver objects", $self->errmsg());
            return (wantarray()) ? (): [];
        }
    }
    return (wantarray()) ? @vserver_objs: \@vserver_objs;
} ## end sub get_vservers

################################################################################
# Method Name: interlace_disks()
# Objective:   Interlace disk objects
# Details:     See POD documentation at the end of this file
################################################################################
sub interlace_disks {
    my ($self, %args) = @_;

    my ($id, $more, @objs, @interlaces, $hsh, $cnt);

    $self->_clear_err();

    # default variable
    $more = 1;
    $cnt  = 0;
    # get all disk objects from the list
    if (defined @{$self->{disk_objs}}) {
        push @objs, @{$self->{disk_objs}};
    } else {
        $self->_set_err("disk object does not exist");
        return $SanUtil::FALSE;
    }

    # get disk object ID
    foreach my $obj (@objs) {
        $id = $obj->get_home();
        push @{$hsh->{$id}}, $obj;
    }

    # sort the list and interlace it
    while ($more) {
        $more = 0;
        foreach my $key (sort keys %{$hsh}) {
            $self->sanlog(-msg_level => 3,
                          -msg       => ++$cnt . ": KEY = " . $key);
            if (@{$hsh->{$key}}) {
                push @interlaces, shift @{$hsh->{$key}};
                $more++;
            }
        }
    }

    @{$self->{disk_objs}} = @interlaces;

    return $SanUtil::TRUE;

} ## end sub interlace_disks

###############################################################################
# Method Name: volume_exist()
# Objective:   Determines if volume exists.
# Details:     See POD documentation at the end of this file
################################################################################
sub volume_exist {
    my ($self, %args) = @_;

    my (@output, %tmp_args);

    $self->_clear_err();

    # to maintain compatibility w/ 7 mode, we will take scalar or hash
    # verify whether pass-in arg is scalar or hash
    foreach my $key (keys %args) {
        if ($key =~ /^-?(name|volume|vserver)$/ and length $args{$key}) {
            # assume it's a hash
            %tmp_args = %args;
            last;
        } else {
            # this is a scalar
            $tmp_args{volume} = $key;
        }
    }
    # set tmp_args to args
    %args = %tmp_args;

    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert name to volume to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{name}) {
        $args{volume} = $args{name};
        delete $args{name};
    }

    # verify name param
    unless (defined $args{volume} and length $args{volume}) {
        $self->_set_err("mandatory parameter \"-name\" not specified ");
        return $SanUtil::FALSE;
    }

    # use default vserver if unspecified
    $args{vserver} = $self->DEFAULT_VSERVER unless (defined $args{vserver});

    # strip /vol/ from volume name
    $args{volume} =~ s/\/vol\/// if ($args{volume} =~ /^\/vol\//);

    $self->sanlog(-msg => $self->dumper_get(\%args),
                  -msg_level => 0);

    # build cmd volume show
    %args = (%args, cmd => "volume_show");

    # invoke cdef command but donot update attributes
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"", $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless (scalar @output) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;

} ## end sub volume_exist

################################################################################
# Method Name: wait_for_wafliron_to_complete()
# Objective:   wait for wafliron to complete on a set of aggregates
# Details:     See POD documentation at the end of this file
################################################################################
sub wait_for_wafliron_to_complete {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             aggr_objs => { type     => ARRAYREF,
                                                            optional => 1,
                                                            default  => [$self->get_aggregates()]},
                                             timeout => { type => SCALAR,
                                                          optional => 1,
                                                          default => 7200 }
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return "";
    }

    my @aggr_objs = @{$args{aggr_objs}};
    my @aggr_names = map { $_->get_name(); } @aggr_objs;
    my @completed_aggrs;
    my @iron_restricted_aggrs;
    my $complete = 1;

    # default the timeout to 2 hours if a bad value was passed in
    if ($args{timeout} < 0) {
        $args{timeout} = 7200;
    }
    my $timeout_time = time + $args{timeout};

    # keep running wafliron show until all the aggrs complete wafliron or enter iron restricted state
    while (scalar(@aggr_names) > (scalar(@completed_aggrs) + scalar(@iron_restricted_aggrs))) {
        # show wafliron status on all aggrs
        my @output = $aggr_objs[0]->wafliron_show(-show_all => $SanUtil::TRUE);
        if ($aggr_objs[0]->error()) {
            $self->_set_err("Failed to display wafliron status", $aggr_objs[0]->errmsg());
            return $SanUtil::FALSE;
        }

        # go through the list of aggr names we have and check if the status is
        # done
        foreach my $aggr_name (@aggr_names) {
            # skip this aggr if it has already completed, or it is in iron_restricted state
            next if (grep(/^$aggr_name$/, @completed_aggrs) or
                     grep(/^$aggr_name$/, @iron_restricted_aggrs));

            foreach my $line (@output) {
                if ($line->{aggregate} eq $aggr_name) {
                    if ($line->{iron_status} =~ /not_running/i) {
                        # wafliron on this aggr is complete
                        # mark it down and go on to the next aggr
                        push(@completed_aggrs, $aggr_name);
                        last;
                    } elsif ($line->{iron_status} =~ /iron_restricted/i) {
                        # this aggr encountered wafliron issues, flag it
                        # and go on to the next aggr
                        push(@iron_restricted_aggrs, $aggr_name);
                        last;
                    } else {
                        # this aggr is not complete,
                        # unset the complete flag so that we sleep and try again
                        $complete = 0;
                        last;
                    }
                }
            }
        }

        # if we exhausted our timeout, error out
        if ($timeout_time < time) {
            $self->_set_err("Wafliron completion detection has timed out after $args{timeout} seconds.");
            return $SanUtil::FALSE;
        }

        # in this case there is still an aggregate with wafliron running,
        # sleep sometime and redo
        if ($complete == 0) {
            # sleep
            &SanUtil::nate_sleep(20);

            # reset complete to 1
            $complete = 1;
            next;
        }
    }

    # at this point, review if wafliron has any changes that need to be committed
    # this only applies to aggrs that are currently in 'iron_restricted' state
    foreach my $aggr_obj (@aggr_objs) {
        my @output = $aggr_obj->wafliron_review();
        if ($aggr_obj->error()) {
            $self->_set_err("Failed to review wafliron changes on aggregate", $aggr_obj->error());
            return $SanUtil::FALSE;
        }
    }

    # Print out summaries of the aggrs that ran wafliron
    if (scalar(@completed_aggrs)) {
        $self->sanlog(-msg => "The following aggregates had no wafliron changes:\n\t" . join("\n\t", @completed_aggrs));
    }
    if (scalar(@iron_restricted_aggrs)) {
        $self->sanlog(-msg => "The following aggregates had wafliron changes:\n\t" . join("\n\t", @iron_restricted_aggrs));
        $self->_set_err("Detected wafliron changes on aggregates (@iron_restricted_aggrs)");
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

################################################################################
# Method Name: wait_for_aggregate_online()
# Objective:   wait for aggregate to be online
# Details:     See POD documentation at the end of this file
################################################################################
sub wait_for_aggregate_online {
    my ($self, %args) = @_;

    my ($aggr_obj);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # in order to use validate_mixed routine to validate parameter types
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # convert aggregate to name to maintain consistency
    # and compatibility w/ 7-mode
    if (exists $args{aggregate}) {
        $args{name} = $args{aggregate};
        delete $args{aggregate};
    }

    # parse and validate arguments
    # using Params::Validate perl module to export constants as part of
    # type checking.
    # Module::Common::validate_mixed is an internal module which is an
    # easier method to parse and validate mandatory and optional argument list
    # at the same time.
    my @input_params = (%args);
    eval {

        # validate input parameters and their type.
        # list - array that contains names of object to be created
        (%args) =
          Module::Common::validate_mixed(@input_params,
                                         [],
                                         {name => {type => SCALAR, required => 1},
                                          timeout => {type => SCALAR,
                                                      optional => 1,
                                                      default => 3600,
                                                      regex => qr/\d+/ },
                                          interval => {type => SCALAR,
                                                       optional => 1,
                                                       default => 60,
                                                       regex => qr/\d+/ }});
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # set start and current times
    my $start_time   = int(gettimeofday);
    my $current_time = $start_time;

    # discover and retrieve aggr
    $aggr_obj = $self->get_aggregate_obj(-name => $args{name});
    unless (ref($aggr_obj) and $aggr_obj->get_name() =~ /^$args{name}$/) {
        $self->_set_err("no aggregate found matching " . $args{name});
        return $SanUtil::FALSE;
    }

    # check for onlined aggregate
    while (($current_time - $start_time) < $args{timeout}) {
        $current_time = int(gettimeofday);

        # refresh and update it attributes
        $aggr_obj->refresh();
        if ($aggr_obj->error()) {
            $self->_set_err("failed to refresh aggregate information " . $args{name}, $aggr_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # aggregate is online
        return $SanUtil::TRUE if ($aggr_obj->get_state() =~ /^online$/i);

        # wait for the next interval to check aggr state
        $self->sanlog(-msg => $args{name} . " is " . $aggr_obj->get_state() .
                      ". Raidstatus is " . $aggr_obj->get_raidstatus() .
                      ". Waiting \"$args{interval}\" seconds");
        &SanUtil::nate_sleep($args{interval});
    }

    # aggr have not onlined prior to timeout expired
    $self->_set_err("unable to online aggregate before time expired");
    return $SanUtil::FALSE;
} ## end sub wait_for_aggregate_online

###############################################################################
# Method:    is_sfo_enabled()
# Objective: return true if sfo is enable else false
# Details:   See POD documentation at the end of this file
###############################################################################
sub is_sfo_enabled {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    # call sfo_show()  and verify if sfo is enabled
    (@output) = $self->sfo_show();
    if ($self->error()) {
        $self->_set_err("failed in sfo_show()",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }
    if ($output[0]->{possible} eq "true") {
        # already enabled, no need to do anything
        return $SanUtil::TRUE;
    }

    return $SanUtil::FALSE;
}

###############################################################################
# Method:    enable_sfo()
# Objective: enable sfo on the node if its not enabled
# Details:   See POD documentation at the end of this file
###############################################################################
sub enable_sfo {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    if ($self->is_sfo_enabled()) {
        return $SanUtil::TRUE;
    }

    # is_sfo_enabled might set error when it returns false, we verify that
    # before proceeding
    if ($self->error()) {
        $self->_set_err("failed to verify whether sfo is enabled",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->halt_node();
    if ($self->error()) {
        $self->_set_err("failed to halt the node",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }
    $self->boot_node();
    if ($self->error()) {
        $self->_set_err("failed to boot the node",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Restore the connection before returning the object to the caller
    $self->reconnect();
    if ($self->error()) {
        $self->_set_err("Failed to restore connection after reboot.",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # A reboot followed by sfo_modify is required
    $self->sfo_modify(-enabled => "true");
    if ($self->error()) {
        $self->_set_err("Failed to modify sfo enabled status",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # verify if sfo is enabled
    if ($self->is_sfo_enabled()) {
        return $SanUtil::TRUE;
    }

    return $SanUtil::FALSE;
}

###############################################################################
# Method:    disable_sfo()
# Objective: disable sfo on the node if its  enabled
# Details:   See POD documentation at the end of this file
###############################################################################
sub disable_sfo {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    unless ($self->is_sfo_enabled()) {
        return $SanUtil::TRUE;
    }

    # disable sfo
    (@output) = $self->SUPER::invoke_cmd_obj(%args,
                                      cmd => "storage_failover_modify",
                                      enabled => "false",
                                      $self->_get_param_defaults());
    if ($self->error()) {
        $self->_set_err("failed to invoke cmd obj" .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    detect_sfo_takeover()
# Objective: Detect within a given timeout, how long it takes for a SFO
#            takeover to occur.
# Details:     See POD documentation at the end of this file
###############################################################################
sub detect_sfo_takeover {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             takeover_node   => {type => SCALAR,
                                                                 optional => 1,
                                                                 default => $self->get_name()},
                                             sfo_timeout     => {type => SCALAR,
                                                                 optional => 1,
                                                                 default => 300},
                                             sfo_max_timeout => {type => SCALAR,
                                                                 optional => 1,
                                                                 default => 900},
                                             cmd_timeout     => {type => SCALAR,
                                                                 optional => 1,
                                                                 default => undef}
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return "";
    }

    # calculate the start and timeout times for the sfo to occur
    my $stime = time();
    my $sfo_timeout_time = $stime + $args{sfo_timeout};

    # default the sfo max timeout to be the same value as sfo_timeout,
    # if it is less than sfo_timeout
    if ($args{sfo_max_timeout} < $args{sfo_timeout}) {
        $args{sfo_max_timeout} = $args{sfo_timeout};
    }

    # calcualte the upper limit timeout time, this will be the time we
    # will report a failure if the sfo has not occurred.
    my $sfo_max_timeout_time = $stime + $args{sfo_max_timeout};

    my $output = {};
    $output->{state} = "";
    $output->{possible} = "";
    $output->{enabled} = "";

    my $timeout_hit = $SanUtil::FALSE;
    my $max_timeout_hit = $SanUtil::FALSE;

    # populate the additional args hash
    my %additional_args;
    if (defined($args{cmd_timeout})) {
        $additional_args{-cmd_timeout} = $args{cmd_timeout};
    }

    # keep sleeping until we have entered the takeover state
    until ($output->{state}    eq "takeover" and
           $output->{possible} eq "false" and
           $output->{enabled}  eq "true") {

        &SanUtil::nate_sleep(1);

        # if we have exceeded the sfo timeout, then set a flag that we failed SFO
        if (time() > $sfo_timeout_time) {
            $timeout_hit = $SanUtil::TRUE;
        }

        # if we have exceeded the sfo max timeout, then break out now
        if (time() > $sfo_max_timeout_time) {
            $max_timeout_hit = $SanUtil::TRUE;
            last;
        }

        # check the SFO status
        ($output) = $self->sfo_show(-node => $args{takeover_node},
                                    -fields => "state,possible,enabled,time-until-takeover",
                                    %additional_args);
        if ($self->error() or not defined($output)) {
            $self->_set_err("failed to show SFO state for " . $args{takeover_node}, $self->errmsg());
            return "";
        }

        # burt 635152: check if the takeover failed, if so, set an error and return.
        if ($output->{state} =~ /takeover_failed/i) {
            $self->_set_err("Detected that the SFO Takeover failed, output state: " . $output->{state});
            return "";
        }
    }

    # calculate the amount of time it took for the takeover to complete
    my $sfo_total_time = time() - $stime;

    # if the timeout was hit, then set an error.
    if ($timeout_hit == $SanUtil::TRUE) {
        if ($max_timeout_hit == $SanUtil::TRUE) {
            # sfo never completed
            $self->_set_err("SFO takeover never occurred and exceeded SFO " .
                            "max timeout " . $args{sfo_max_timeout} .
                            " seconds. ($sfo_total_time seconds)");
        } else {
            # sfo completed, but took too long
            return $sfo_total_time;
        }
    } else {
        # success case
        $self->sanlog(-msg => "SFO takeover took $sfo_total_time seconds to complete " .
                              "(timeout = " . $args{sfo_timeout} . ")");
    }

    return $sfo_total_time;
}

###############################################################################
# Method:    detect_sfo_giveback()
# Objective: Detect within a given timeout, how long it takes for a SFO
#            giveback to occur.
# Details:     See POD documentation at the end of this file
###############################################################################
sub detect_sfo_giveback {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             giveback_node => {type => SCALAR,
                                                               optional => 1,
                                                               default => $self->get_name()},
                                             sfo_timeout   => {type => SCALAR,
                                                               optional => 1,
                                                               default => 300},
                                             sfo_max_timeout => {type => SCALAR,
                                                                 optional => 1,
                                                                 default => 900},
                                             wait_for_aggrs_to_sendhome => {type => SCALAR,
                                                                            optional => 1,
                                                                            default => 1},
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return 0;
    }

    # store the start and end times for the SFO to occur between
    my $stime = time();
    my $sfo_timeout_time = $stime + $args{sfo_timeout};

    # default the sfo max timeout to be the same value as sfo_timeout,
    # if it is less than sfo_timeout
    if ($args{sfo_max_timeout} < $args{sfo_timeout}) {
        $args{sfo_max_timeout} = $args{sfo_timeout};
    }

    # calculate the upper limit timeout time, this will be the time we
    # will report a failure if the sfo has not occurred.
    my $sfo_max_timeout_time = $stime + $args{sfo_max_timeout};

    my $timeout_hit = $SanUtil::FALSE;
    my $max_timeout_hit = $SanUtil::FALSE;

    # initialize the sfo show output hash
    my $sfo_show_output = {};
    $sfo_show_output->{reason} = "";
    $sfo_show_output->{state} = "";
    $sfo_show_output->{possible} = "";
    $sfo_show_output->{enabled} = "";

    # if requested to wait for aggrs to sendhome, then initialize the giveback
    # status value to null
    my @sfo_show_giveback_output = ();
    if ($args{wait_for_aggrs_to_sendhome} == 1) {
        $sfo_show_giveback_output[0]->{giveback_status} = "";
    } else {
        # otherwise, hardcode the giveback status so that the until statement
        # can evaluate to true if not requested to wait for aggrs to sendhome
        $sfo_show_giveback_output[0]->{giveback_status} = "No aggregates to give back";
    }

    # detect when the giveback has completed
    until ($sfo_show_output->{state}      eq "connected" and
           $sfo_show_output->{enabled}    eq "true" and
           (($sfo_show_output->{possible} eq "true" and $sfo_show_output->{reason} eq "-") or
            ($sfo_show_output->{reason} =~ /The version of software/i and $sfo_show_output->{reason} !~ /Failover partner node is booting/i)) and
           $sfo_show_giveback_output[0]->{giveback_status} =~ /No aggregates to give back/i) {

        &SanUtil::nate_sleep(1);

        # if we have exceeded the sfo timeout, then set a flag that the giveback took
        # too long
        if (time() > $sfo_timeout_time) {
            $timeout_hit = $SanUtil::TRUE;
        }

        # if we have exceeded the sfo max timeout, then break out now
        if (time() > $sfo_max_timeout_time) {
            $max_timeout_hit = $SanUtil::TRUE;
            last;
        }

        # check SFO status
        ($sfo_show_output) = $self->sfo_show(-node => $args{giveback_node},
                                             -fields => "state,possible,enabled,reason");
        if ($self->error()) {
            $self->_set_err("failed to show SFO state for " . $args{giveback_node}, $self->errmsg());
            return 0;
        }

        # if requested to wait for aggregates to be sent home, detect that here
        # the purpose of this is to verify that the Aggrs are actually getting sent home
        # properly, also for logging purposes to illustrate which aggrs are being
        # sent home
        if ($args{wait_for_aggrs_to_sendhome} == 1) {
            (@sfo_show_giveback_output) = $self->sfo_show_giveback(-node => $args{giveback_node},
                                                                   -fields => "giveback-status,aggregate,destination");
            if ($self->error()) {
                $self->_set_err("failed to show SFO giveback status for " .
                                $args{giveback_node}, $self->errmsg());
                return 0;
            }
            foreach my $elm (@sfo_show_giveback_output) {
                if ($elm->{giveback_status} =~ /Failed,\s*\"module:\s*(\S+?)\"/i) {
                    my $module = $1;
                    my $aggr = $elm->{aggregate};
                    $aggr =~ s/\"//g;
                    my $return_data = {};
                    $return_data->{module} = $module;
                    $return_data->{aggr} = $aggr;
                    $self->sanlog(-msg => "SFO Giveback failed in module '$module' for aggregate '$aggr'");
                    return -1;
                }
            }
        }
    }

    # calculate how much time it took the SFO giveback to take
    my $sfo_total_time = time() - $stime;

    # if the timeout was hit, then set an error.
    if ($timeout_hit == $SanUtil::TRUE) {
        if ($max_timeout_hit == $SanUtil::TRUE) {
            # sfo never completed in time
            $self->_set_err("SFO giveback never occurred and SFO max timeout (" .
                             $args{sfo_max_timeout} . " seconds) was exhausted.");
        } else {
            # sfo completed but took too long
            return $sfo_total_time;
        }
    } else {
        # success case
        $self->sanlog(-msg => "SFO giveback took $sfo_total_time seconds to complete " .
                              "(timeout = " . $args{sfo_timeout} . ")");
    }

    return $sfo_total_time;
}

###############################################################################
# Method:    detect_waiting_for_giveback()
# Objective: Detect when this node reaches the waiting for giveback prompt
# Details:     See POD documentation at the end of this file
###############################################################################
sub detect_waiting_for_giveback {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my %console_connect_args;
    $console_connect_args{node} = $args{node} if (exists($args{node}));
    $console_connect_args{conn_type} = delete($args{conn_type}) if (exists($args{conn_type}));
    $console_connect_args{no_login} = delete($args{no_login}) if (exists($args{no_login}));
    $console_connect_args{conn_timeout} = delete($args{conn_timeout}) if (exists($args{conn_timeout}));
    $console_connect_args{username} = delete($args{username}) if (exists($args{username}));
    $console_connect_args{password} = delete($args{password}) if (exists($args{password}));

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             node => {type => SCALAR,
                                                      optional => 1,
                                                      default => $self->get_name()},
                                             take_over_node => {type => SCALAR,
                                                               optional => 1,
                                                               default => ""},
                                             timeout => {type => SCALAR,
                                                         optional => 1,
                                                         default => 300}
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return "";
    }

    # if the node arg doesn't represent the current object
    # and the user specified a take over node, then perform the sfo show check
    # against the takeover node
    if ($args{node} ne $self->get_name() and $args{take_over_node} ne "") {
        # run sfo show to see if the taken over node is in a waiting for giveback state
        my ($output) = $self->sfo_show(-node => $args{take_over_node},
                                       -fields => "state,partner-state");
        if ($self->error()) {
            $self->_set_err("Unable to determine if $args{node} is in the waiting for giveback state",
                            $self->errmsg());
            return "";
        }

        # if the HA partner is already in a waiting for giveback state, then just return
        if ($output->{partner_state} =~ /Waiting for giveback \(HA Mailboxes\)/i) {
            return 0;
        }
    }

    # otherwise, connect to the console for the specified node
    # and poll until it hits the waiting for giveback prompt
    my ($conn) = $self->console_connect(-login => 0,
                                        %console_connect_args);
    if ($self->error()) {
        $self->_set_err("Failed to connect to node console ", $self->errmsg());
        return "";
    }

    my $state = "";
    my $stime = time();
    my $timeout_time = time() + $args{timeout};

    # keep checking the current state until we have reached the
    # taken over waiting ontap state
    until ($state eq "TAKEN_OVER_WAITING_ONTAP") {
        eval {
            $state = &boot::boot_state($conn);
        };
        if ($@) {
            $self->_set_err("Failed to obtain boot state: $@");
            &Tharn::disconnect($conn);
            return "";
        }

        # if we have exhausted the timeout for detecting wfgiveback error out
        if (time() > $timeout_time) {
            $self->_set_err("Exceeded waiting for giveback timeout ($args{timeout} seconds)");
            &Tharn::disconnect($conn);
            return "";
        }
    }

    # disconnect from the console
    &Tharn::disconnect($conn);

    # if the node arg doesn't represent the current object
    # and the user specified a take over node, then perform the sfo show check
    # against the takeover node after we have hit a semi waiting for giveback state
    if ($args{node} ne $self->get_name() and $args{take_over_node} ne "") {
        # keep running sfo show to validate that the taken over node is in a waiting for giveback state
        my $output = {};
        until ($output->{partner_state} =~ /Waiting for giveback \(HA Mailboxes\)/i) {
            ($output) = $self->sfo_show(-node => $args{take_over_node},
                                        -fields => "state,partner-state");
            if ($self->error()) {
                $self->_set_err("Unable to determine if $args{node} is in the waiting for giveback state",
                                $self->errmsg());
                return "";
            }
            # if we have exhausted the timeout for detecting wfgiveback error out
            if (time() > $timeout_time) {
                $self->_set_err("Exceeded waiting for giveback timeout ($args{timeout} seconds)");
                return "";
            }
            &SanUtil::nate_sleep(1);
        }
    }

    # return back how long it took to detect
    return time() - $stime;
}

###############################################################################
# Method:    wait_for_node_to_boot()
# Objective: Wait for a specified node to boot
# Details:     See POD documentation at the end of this file
###############################################################################
sub wait_for_node_to_boot {
    my ($self, %args) = @_;

    # connect to the console for the specified node
    my ($conn) = $self->console_connect(-login => 0,
                                        %args);
    if ($self->error()) {
        $self->_set_err("Failed to connect to node console ", $self->errmsg());
        return "";
    }

    %args = &SanUtil::strip_dash(%args);

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             node => {type => SCALAR,
                                                      optional => 1,
                                                      default => $self->get_name()},
                                             timeout => {type => SCALAR,
                                                         optional => 1,
                                                         default => 600}
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return "";
    }

    my $state = "";
    my $stime = time();
    my $timeout_time = time() + $args{timeout};

    # keep checking the current state until we have reached the up state
    until ($state eq "UP" or $state eq "USERNAME" or $state eq "PASSWORD") {
        eval {
            $state = &boot::boot_state($conn);
        };
        if ($@) {
            $self->_set_err("Failed to obtain boot state: $@");
            &Tharn::disconnect($conn);
            return "";
        }

        # if we have exhausted the timeout for detecting boot
        if (time() > $timeout_time) {
            $self->_set_err("Exceeded waiting for boot timeout ($args{timeout} seconds)");
            &Tharn::disconnect($conn);
            return "";
        }
    }

    # disconnect from the console
    &Tharn::disconnect($conn);

    # return back how long it took to detect
    return time() - $stime;
}

###############################################################################
# Method:    panic_remote_node()
# Objective: panic a remote node using a node scoped command
# Details:     See POD documentation at the end of this file
###############################################################################
sub panic_remote_node {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             node => {type => SCALAR,
                                                      required => 1}
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # this method cannot be used to panic the current node
    # it can only be used panic a remote node
    my $node = $self->get_name();
    if ($args{node} eq $node) {
        $self->_set_err("This method cannot panic the current node.");
        return $SanUtil::FALSE;
    }

    # run a node scoped command to panic the specified node
    $self->run_node_scope(-cmd => "panic",
                          -timeout => 10,
                          -priv => "diag",
                          -filer => $args{node});

    # we are expecting this command to timeout in a few seconds
    # so clear the error on the object if we see the error "failed to execute"
    if ($self->error() and grep(/failed to execute:/, $self->errmsg())) {
        # this error was expected, clear it
        $self->_clear_err();
    } elsif ($self->error()) {
        # otherwise the error was something other than timing out
        # treat it as a real error
        $self->_set_err("Failed to panic node $args{node}", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # we need to clear out the prompt because the node scoped command will timeout and leave the
    # cli in a state that requires a ctrl c.
    $self->execute(-cmd => "\cc");

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    boot_node()
# Objective: Boot ontap on the specified node
# Details:     See POD documentation at the end of this file
###############################################################################
sub boot_node {
    my ($self, %args) = @_;

    # connect to the console on the requested node
    # dont login because we expect to not be at a login prompt
    my ($conn) = $self->console_connect(-login => 0,
                                        %args);
    if ($self->error()) {
        $self->_set_err("Failed to connect to node console", $self->errmsg());
        return "";
    }

    my $state = "";

    # boot the node
    try {
        my $transit = NACL::Transit->new(name => $conn);
        my $state = $transit->change_state(to => "UP_OR_TAKEN_OVER");
    } otherwise {
        my $e = shift;
        $self->_set_err("Exception caught: " . $e->text());
    };

    # disconnect form the console
    &Tharn::disconnect($conn);

    return $state;
}

###############################################################################
# Method:    halt_node()
# Objective: Halt the specifed node down to the loader prompt.
# Details:     See POD documentation at the end of this file
###############################################################################
sub halt_node {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my %console_connect_args;
    $console_connect_args{node} = $args{node} if (exists($args{node}));
    $console_connect_args{conn_type} = delete($args{conn_type}) if (exists($args{conn_type}));
    $console_connect_args{login} = delete($args{login}) if (exists($args{login}));
    $console_connect_args{no_login} = delete($args{no_login}) if (exists($args{no_login}));
    $console_connect_args{conn_timeout} = delete($args{conn_timeout}) if (exists($args{conn_timeout}));
    $console_connect_args{username} = delete($args{username}) if (exists($args{username}));
    $console_connect_args{password} = delete($args{password}) if (exists($args{password}));

    # connect to the console on the specified node
    my ($conn) = $self->console_connect(%console_connect_args);
    if ($self->error()) {
        $self->_set_err("Failed to connect to node console", $self->errmsg());
        return $SanUtil::FALSE;
    }

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             node => {type => SCALAR,
                                                      optional => 1,
                                                      default => $self->get_name()},
                                             halt_timeout => {type => SCALAR,
                                                              optional => 1,
                                                              default => 300},
                                             inhibit_takeover => {type => SCALAR,
                                                                  optional => 1,
                                                                  default => "false"},
                                             dump => {type => SCALAR,
                                                      optional => 1,
                                                      default => "false"},
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # halt the specified node via the console
    my $expected_prompt = "WARNING: Are you sure you want to halt the node? {y|n}:";
    my (undef, undef, undef, $received_prompt) =
        &Tharn::say($conn, "system node halt -node $args{node} -inhibit-takeover $args{inhibit_takeover} -dump $args{dump}",
                    '-prompt', [$expected_prompt, "CFE\>|LOADER\>|LOADER\-[AB]\>"],
                    '-timeout', $args{halt_timeout},
                    '-ignore', "echo ignore",
                    '-comment', "Attempted to halt node");
    if ($received_prompt !~ /$expected_prompt/) {
    &Tharn::disconnect($conn);
        return $SanUtil::TRUE;
    }

    # answer the expected question and return once we hit the loader prompt
    $expected_prompt = "CFE\>|LOADER\>|LOADER\-[AB]\>";
    (undef, undef, undef, $received_prompt) =
        &Tharn::say($conn, "y",
                    '-noprompt', '-prompt', $expected_prompt,
                    '-timeout', $args{halt_timeout},
                    '-ignore', "echo ignore",
                    '-comment', "Attempted to halt node");
    if ($received_prompt !~ /$expected_prompt/) {
        $self->_set_err("Did not receive expected prompt ($expected_prompt)");
        return $SanUtil::FALSE;
    }

    # disconnect from the console
    &Tharn::disconnect($conn);

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    disable_auto_giveback()
# Objective: Disable the storage failover option "auto-giveback"
# Details:     See POD documentation at the end of this file
###############################################################################
sub disable_auto_giveback {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             node => {type => SCALAR,
                                                      optional => 1,
                                                      default => $self->get_name()},
                                         });
    };
    if ($@) {
        $self->_set_err("Failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # disable the auto giveback mechanism, this asks a question so we need to respond
    # to the specific prompt
    my ($received_prompt, @output) =
        $self->execute(-cmd => "storage failover modify -node " . $args{node} . " -auto-giveback false",
                       -return_prompt => 1,
                       -prompts => ["\\{y\\|n\\}"]);
    if ($self->error()) {
        # set error and bubble up
        $self->_set_err("Failed to invoke 'storage failover modify'", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # if the received prompt was what we expected then send a "y" character
    if ($received_prompt =~ /\{y\|n\}/) {
        $self->execute(-cmd => "y");
        if ($self->error()) {
            $self->_set_err("Failed to invoke 'y'", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # in RR.1 and later, there is a new setting to disable auto-giveback-after-panic
    # check the cdefs to see if the storage failover modify cdef contains this modify option
    if ($self->{release} !~ /RollingRock__8\.1\.0|BoilerMaker__8\.0/) {
        ($received_prompt, @output) =
            $self->execute(-cmd => "storage failover modify -node " . $args{node} . " -auto-giveback-after-panic false",
                           -return_prompt => 1,
                           -prompts => ["\\{y\\|n\\}"]);
        if ($self->error()) {
            # ignore the error for now because this change has not made it into SN.0 yet
            $self->_clear_err();
            return $SanUtil::TRUE;
        } else {
            # if the received prompt was what we expected then send a "y" character
            if ($received_prompt =~ /\{y\|n\}/) {
                $self->execute(-cmd => "y");
                if ($self->error()) {
                    $self->_set_err("Failed to invoke 'y'", $self->errmsg());
                    return $SanUtil::FALSE;
                }
            }
        }
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:    console_connect()
# Objective: Connect to a node's console port and return the connection
# Details:     See POD documentation at the end of this file
###############################################################################
sub console_connect {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    # flip from no_login to login for backward compat
    if (exists($args{no_login})) {
        $args{login} = !$args{no_login};
        delete $args{no_login};
    }

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             node         => {type => SCALAR,
                                                              optional => 1,
                                                              default => $self->get_name()},
                                             conn_type    => {type => SCALAR,
                                                              optional => 1,
                                                              default => "console"},
                                             login        => {type => SCALAR,
                                                              optional => 1,
                                                              default => 1},
                                             conn_timeout => {type => SCALAR,
                                                              optional => 1,
                                                              default => 300},
                                             username     => {type => SCALAR,
                                                              optional => 1},
                                             password     => {type => SCALAR,
                                                              optional => 1}
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return {};
    }

    # Grab the host object so we can modify if needed
    my $host_obj = &Tharn::host($args{node});

    # Create options to pass to the Hosttrec::connect
    # We are not connecting (connect => false) for now
    # because we will call connect with login later.
    # This is a work around until NATE4 because the
    # Connectrec->new can't accept login option this is
    # tracked by burt 462461 this means if the connection
    # is not logged in it will fail if the paramter is not
    # passed, and since we CAN'T pass it without failing
    # parameter validation it will FAIL.
    my %options = ( template => $args{conn_type},
                    connect => $SanUtil::FALSE);
    $options{username} = $args{username} if defined($args{username});
    $options{password} = $args{password} if defined($args{password});

    my $conn = undef;
    eval {
        $conn = $host_obj->connect(%options);

        # We have to pass this seperate for now because it is not available
        # to the Connectrec::new until NATE4
        $conn->connect(login => $args{login});
    };
    if ($@) {
           my $errmsg = "console_connect ERROR: " . join("", $@);
           $errmsg =~ s/\n//g;
           $self->_set_err($errmsg);
           $self->sanlog(-msg => $errmsg, -msg_level => 0);
           return {};
    }
    return $conn;
}

################################################################################
# Method:    console_execute()
# Objective: make console connect and execute command based on shell_type
# Details:   See POD documentation at the end of this file.
################################################################################
sub console_execute {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    my $buffer = undef;
    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             cmd         => {type => SCALAR,
                                                              required => 1},
                                             shell_type  => {type => SCALAR,
                                                              optional => 1,
                                                              default => "bsd"},
                                             timeout     => {type => SCALAR,
                                                              optional => 1,
                                                              default => 120}
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return {};
    }

    # Connect to the console on the specified node
    my ($conn) = $self->console_connect(-username => $self->DIAG_USER,
                                        -password => $self->DEFAULT_PASSWD);
    if ($self->error()) {
        $self->_set_err("failed to connect to node console",
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    $conn = $self->transit_console_to(-conn => $conn,
                                      -shell_type => $args{shell_type});
    if ($self->error()) {
        $self->_set_err("failed to transit console to $args{shell_type}",
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }


    # execute command on the node via the console
    eval {
        (undef, undef, $buffer, undef) = Tharn::say($conn, $args{cmd},
                                                        -timeout => $args{timeout});
        $buffer =~ s/\cG//g;
    };
    if ($@) {
        $self->_set_err("failed to execute: " . $args{cmd} , $@);
        return (wantarray()) ? () : [];
    }

    # disconnect and logout from the console
    &Tharn::disconnect($conn, -logout);

    return (wantarray()) ? ($buffer) : [$buffer];
} # end sub console_execute

################################################################################
sub transit_console_to {
    my ($self, %args) = @_;

    $self->_clear_err();

    %args = &SanUtil::strip_dash(%args);

    my $output;
    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             conn        => {type => OBJECT,
                                                             required => 1},
                                             shell_type  => {type => SCALAR,
                                                             optional => 1,
                                                             default => "bsd"}
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return {};
    }

    my $conn = $args{conn};

    # Find out which prompt we're connected to
    eval {
        $output = (Tharn::say($conn, " ", -timeout, $self->CMD_TIMEOUT))[$Return_index];
    };
    if ($@) {
        $self->_set_err("failed to call say(): ", $@);
        return {};
    }

    $self->sanlog(msg => "Prompt is $output", -msg_level => 7);

    # get controller name
    my $node = $self->get_node();

    # If the connection was in ngsh attempt to reconnect to the console
    if ($output =~ /.+\:\:.*>/) {
        $output = $self->_reset_connection($conn);
        if ($self->error()) {
            $self->_set_err("failed to reset connection", $self->errmsg());
            return {};
        }

        # Check to see if it's back to bsd
        if ($output =~ /$node%/) {
            return $conn;
        } else {
            $self->_set_err("Lost prompt, user intervention required");
            return {};
        }
    # If the connection was in node run, logout and attempt to reconnect twice
    } elsif ($output =~ /.+\>|.+\*\>/) {
        $self->_reset_connection($conn);

        $output = $self->_reset_connection($conn);
        if ($self->error()) {
            $self->_set_err("failed to reset connection", $self->errmsg());
            return {};
        }

        # Check if it's back in bsd shell
        if ($output =~ /$node%/) {
            return $conn;
        } else {
            $self->_set_err("Lost prompt, user intervention required");
            return {};
        }
    # If the connection was already in bsd then return
    } elsif ($output =~ /$node%/) {
        return $conn;
    }

    # If we reached this point it means something is very messed up
    $self->_set_err("failed to transit console to: $args{type}");
    return {};
} # end transit_console_to_bsd

###############################################################################
# Method:    iscsi_list_info()
# Objective: get iscsi information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_show",
             -donot_append_vserver => 1);

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_list_info

###############################################################################
# Method:    iscsi_isns_list_info()
# Objective: get iscsi isns information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_isns_list_info {
    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_isns_show",
             -donot_append_vserver => 1);

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_isns_list_info

###############################################################################
# Method:    iscsi_status_list_info()
# Objective: get iscsi status information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_status_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Note: force this to CLI.  This zapi is only supported at vserver level,
    # not at cluster level.
    %args = (%args,
             cmd => "vserver_iscsi_status",
             cmd_interface         => "cli",
             -donot_append_vserver => 1);

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_status_list_info

###############################################################################
# Method Name: iscsi_connection_list_info
# Objective: get iscsi connections
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_connection_list_info {
    my ($self, %args) = @_;

    my (@output, $init_nodename);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # iscsi connection show used to take an -initiator-name flag.
    # But this was removed.  Our testcases found this option useful.
    # So I have added in code to this function to support the -initiator-name flag.
    # In this case, we call session_list_info with the initiator-name parm.
    # Then take the tsih, tpgroup-tag, and vserver from each session to discover
    # the connections.
    # Also added session-type filter.  This filter is also used first on iscsi
    # session show as a "pre-filter" for iscsi connection show.

    # If the caller passed in initiator-name or session-type, then call
    # session show first as a pre-filter.
    if (exists $args{'initiator-name'} or exists $args{'session-type'}) {

        # User asked for initiator-name and/or session-type filter.  We need to
        # call session_list_info first in order to then call
        # connection_list_info with the session's tsih.
        my %session_args = (donot_append_vserver => 1);

        # If the user specified initiator-name, add it to the %session_args.
        # Also delete it from the %args.
        if (defined $args{'initiator-name'} and length $args{'initiator-name'}) {
            %session_args =
              (%session_args, 'initiator-name' => $args{'initiator-name'});
            delete $args{'initiator-name'};
        }

        # If the user specified session-type, add it to the %session_args.
        # Also delete it from the %args.
        if (defined $args{'session-type'} and length $args{'session-type'}) {
            %session_args =
              (%session_args, 'session-type' => $args{'session-type'});
            delete $args{'session-type'};
        }

        # If the user asked for a tsih filter, let's add that as well
        if (defined $args{'tsih'} and length $args{'tsih'}) {
            %session_args = (%session_args, tsih => $args{tsih});

            # delete the tsih from args.  It is added back in below from the
            # session data. (In case the user did not specify this arg.)
            delete $args{tsih};
        }

        # If the user asked for a tpgroup-tag filter, let's add that as well
        if (defined $args{'tpgroup-tag'} and length $args{'tpgroup-tag'}) {
            %session_args = (%session_args, 'tpgroup-tag' => $args{'tpgroup-tag'});

            # delete the tpgroup-tag from args.  It is added back in below from the
            # session data. (In case the user did not specify this arg.)
            delete $args{'tpgroup-tag'};
        }

        # If the user asked for a vserver filter, let's add that as well
        if (defined $args{'vserver'} and length $args{'vserver'}) {
            %session_args = (%session_args, vserver => $args{vserver});

            # delete the vserver from args.  It is added back in below from the
            # session data. (In case the user did not specify this arg.)
            delete $args{vserver};
        }

        # Now call session list info with the args we built
        my @sessions = $self->iscsi_session_list_info(%session_args);
        if ($self->error()) {
            $self->_set_err("Failed: iscsi_session_list_info with args:",
                            Dumper(%session_args),
                            $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # Now scan each session to find connections
        foreach my $session (@sessions) {

            # Add parms from the session to the conn arg list.
            %args = (%args,
                     cmd                  => "vserver_iscsi_connection_show",
                     tsih                 => $session->{tsih},
                     'tpgroup-tag'        => $session->{'tpgroup_tag'},
                     vserver              => $session->{vserver},
                     donot_append_vserver => 1);

            # If using ZAPI and user did not set max-records, set max-records
            if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
            {
                $args{'max-records'} = MAX_ZAPI_RECORDS;
            }

            my @conns = $self->SUPER::invoke_cmd_obj(%args);
            if ($self->error()) {
                $self->_set_err("failed to invoke_cmd_obj " .
                                $self->dumper_get(\%args, $self->get_debug_level()),
                                $self->errmsg());
                return (wantarray()) ? () : [];
            }

            # If using zapi and max-records were returned, issue a warning.
            # There could be more items.
            if (($self->get_cmd_interface() =~ /zapi/i)
                and scalar(@conns) == $args{'max-records'}) {
                &SanUtil::nate_log(
                    '-warn',
                    $args{'max-records'}
                      . " items returned from the zapi call. "
                      . "Call currently uses a max-records value of "
                      . $args{'max-records'}
                      . ", so there may have been more. "
                      . "You may need to increase the max-records value.");
            } ## end if (($self->get_cmd_interface...

            # Save the found connections to the output
            push(@output, @conns);

        } ## end foreach my $session (@sessions)

    } else {

        # Else user did not want any special session show filters.
        # Just call connection show normally.
        %args = (%args,
                 cmd                  => "vserver_iscsi_connection_show",
                 donot_append_vserver => 1);

        # If using ZAPI and user did not set max-records, set max-records
        if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
        {
            $args{'max-records'} = MAX_ZAPI_RECORDS;
        }

        (@output) = $self->SUPER::invoke_cmd_obj(%args);
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj " .
                            $self->dumper_get(\%args, $self->get_debug_level()),
                            $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # If using zapi and max-records were returned, issue a warning.
        # There could be more items.
        if (($self->get_cmd_interface() =~ /zapi/i)
            and scalar(@output) == $args{'max-records'}) {
            &SanUtil::nate_log(
                '-warn',
                $args{'max-records'}
                  . " items returned from the zapi call. "
                  . "Call currently uses a max-records value of "
                  . $args{'max-records'}
                  . ", so there may have been more. "
                  . "You may need to increase the max-records value.");
        } ## end if (($self->get_cmd_interface...

    } ## end else [ if (exists $args{'initiator-name'...

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_connection_list_info

###############################################################################
# Method Name: iscsi_session_list_info
# Objective: get iscsi sessions
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_session_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_session_show",
             -donot_append_vserver => 1);

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_session_list_info

###############################################################################
# Method Name: iscsi_nodename_list_info
# Objective: get iscsi nodename(s)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_nodename_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Note: force this to CLI.  This zapi is only supported at vserver level,
    # not at cluster level.
    %args = (%args,
             cmd => "vserver_iscsi_nodename",
             cmd_interface         => "cli",
             -donot_append_vserver => 1);

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_nodename_list_info

###############################################################################
# Method Name: iscsi_alias_list_info
# Objective: get iscsi alias(es)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_alias_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Note: force this to CLI.  This zapi is only supported at vserver level,
    # not at cluster level.
    %args = (%args,
             cmd                   => "vserver_iscsi_alias",
             cmd_interface         => "cli",
             -donot_append_vserver => 1);

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj "
                          . $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_alias_list_info

###############################################################################
# Method Name: iscsi_initiator_list_info
# Objective: get iscsi initiator(s)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_initiator_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # If the caller passed in session-type, then call session show first as a
    # pre-filter.
    if (exists $args{'session-type'}) {

        # User asked for session-type filter.  We need to call
        # session_list_info first in order to then call initiator_list_info
        # with the session's tsih and tpgroup-tag.
        my %session_args = (donot_append_vserver => 1);

        # If the user specified initiator-name, add it to the %session_args.
        # Don't delete initiator from the args, we'll use it when we call
        # initiator show below.
        if (defined $args{'initiator-name'} and length $args{'initiator-name'}) {
            %session_args =
              (%session_args, 'initiator-name' => $args{'initiator-name'});
        }

        # If the user specified session-type, add it to the %session_args.
        # Also delete it from the %args.
        if (defined $args{'session-type'} and length $args{'session-type'}) {
            %session_args =
              (%session_args, 'session-type' => $args{'session-type'});
            delete $args{'session-type'};
        }

        # If the user asked for a tsih filter, let's add that as well
        if (defined $args{'tsih'} and length $args{'tsih'}) {
            %session_args = (%session_args, tsih => $args{tsih});

            # delete the tsih from args.  It is added back in below from the
            # session data. (In case the user did not specify this arg.)
            delete $args{tsih};
        }

        # If the user asked for a vserver filter, let's add that as well
        if (defined $args{'vserver'} and length $args{'vserver'}) {
            %session_args = (%session_args, vserver => $args{vserver});

            # delete the vserver from args.  It is added back in below from the
            # session data. (In case the user did not specify this arg.)
            delete $args{vserver};
        }

        # If the user asked for a tpgroup-tag filter, let's add that as well
        if (defined $args{'tpgroup-tag'} and length $args{'tpgroup-tag'}) {
            %session_args = (%session_args, 'tpgroup-tag' => $args{'tpgroup-tag'});

            # delete the tpgroup-tag from args.  It is added back in below from the
            # session data. (In case the user did not specify this arg.)
            delete $args{'tpgroup-tag'};
        }

        # Now call session list info with the args we built
        my @sessions = $self->iscsi_session_list_info(%session_args);
        if ($self->error()) {
            $self->_set_err("Failed:iscsi_session_list_info with args:",
                            Dumper(%session_args),
                            $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # Now scan each session to find initiators
        foreach my $session (@sessions) {

            # Add tsih from the session to the arg list.
            %args = (%args,
                     cmd                  => "vserver_iscsi_initiator_show",
                     tsih                 => $session->{tsih},
                     'tpgroup-tag'        => $session->{'tpgroup_tag'},
                     vserver              => $session->{vserver},
                     donot_append_vserver => 1);

            # If using ZAPI and user did not set max-records, set max-records
            if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
            {
                $args{'max-records'} = MAX_ZAPI_RECORDS;
            }

            my @inits = $self->SUPER::invoke_cmd_obj(%args);
            if ($self->error()) {
                $self->_set_err("failed to invoke_cmd_obj " .
                                $self->dumper_get(\%args, $self->get_debug_level()),
                                $self->errmsg());
                return (wantarray()) ? () : [];
            }

            # If using zapi and max-records were returned, issue a warning.
            # There could be more items.
            if (($self->get_cmd_interface() =~ /zapi/i)
                and scalar(@inits) == $args{'max-records'}) {
                &SanUtil::nate_log(
                    '-warn',
                    $args{'max-records'}
                      . " items returned from the zapi call. "
                      . "Call currently uses a max-records value of "
                      . $args{'max-records'}
                      . ", so there may have been more. "
                      . "You may need to increase the max-records value.");
            } ## end if (($self->get_cmd_interface...

            # Save the found initiators to the output
            push(@output, @inits);

        } ## end foreach my $session (@session

    } else {

        # Else user did not want any special session show filters.
        # Just call initiator show normally.
        %args = (%args,
                 cmd => "vserver_iscsi_initiator_show",
                 -donot_append_vserver => 1);

        # If using ZAPI and user did not set max-records, set max-records
        if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
        {
            $args{'max-records'} = MAX_ZAPI_RECORDS;
        }

        (@output) = $self->SUPER::invoke_cmd_obj(%args);
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj " .
                            $self->dumper_get(\%args, $self->get_debug_level()),
                            $self->errmsg());
            return (wantarray()) ? () : [];
        }

        # If using zapi and max-records were returned, issue a warning.
        # There could be more items.
        if (($self->get_cmd_interface() =~ /zapi/i)
            and scalar(@output) == $args{'max-records'}) {
            &SanUtil::nate_log(
                '-warn',
                $args{'max-records'}
                  . " items returned from the zapi call. "
                  . "Call currently uses a max-records value of "
                  . $args{'max-records'}
                  . ", so there may have been more. "
                  . "You may need to increase the max-records value.");
        } ## end if (($self->get_cmd_interface...

    } ## end else [ if (exists $args{'session-type'...

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_initiator_list_info

###############################################################################
# Method Name: iscsi_tpgroup_list_info
# Objective: get iscsi tpgroup(s)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_tpgroup_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_tpgroup_show",
             -donot_append_vserver => 1);

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_tpgroup_list_info

###############################################################################
# Method Name: iscsi_interface_list_info
# Objective: get iscsi interface(s)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_interface_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_interface_show",
             -donot_append_vserver => 1);

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_interface_list_info

###############################################################################
# Method Name: iscsi_interface_accesslist_list_info
# Objective: get iscsi interface accesslist(s)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_interface_accesslist_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_interface_accesslist_show",
             -donot_append_vserver => 1);

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_interface_accesslist_list_info

###############################################################################
# Method Name: iscsi_portal_list_info
# Objective: get iscsi portal(s)
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_portal_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # vserver_iscsi_portal_show cdef doesn't exist.
    # burt443724 provided a work-around for the non-existing cdef.
    # burt430590 asks for cdefs for portal show (an aliased cmd)
    %args = (%args,
             cmd => "vserver_iscsi_portal_show",
             -donot_append_vserver => 1);

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_portal_list_info

###############################################################################
# Method Name: iscsi_initiator_auth_list_info
# Objective: get iscsi security entries
# Details:   See POD documentation at the end of this file.
###############################################################################
sub iscsi_initiator_auth_list_info {

    my ($self, %args) = @_;

    my (@output);

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    %args = (%args,
             cmd => "vserver_iscsi_security_show",
             -donot_append_vserver => 1);

    # If using ZAPI and user did not set max-records, set max-records
    if ($self->get_cmd_interface() =~ /zapi/i and !exists $args{'max-records'})
    {
        $args{'max-records'} = MAX_ZAPI_RECORDS;
    }

    (@output) = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%args, $self->get_debug_level()),
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    # If using zapi and max-records were returned, issue a warning.
    # There could be more items.
    if (($self->get_cmd_interface() =~ /zapi/i)
        and scalar(@output) == $args{'max-records'}) {
        &SanUtil::nate_log(
            '-warn',
            $args{'max-records'}
              . " items returned from the zapi call. "
              . "Call currently uses a max-records value of "
              . $args{'max-records'}
              . ", so there may have been more. "
              . "You may need to increase the max-records value.");
    } ## end if (($self->get_cmd_interface...

    return (wantarray()) ? @output : \@output;
} ## end sub iscsi_initiator_auth_list_info

###############################################################################
# Method Name: kcs_list_info()
# Objective: Get kernel service information on this node
# Details:   See POD documentation at the end of this file.
###############################################################################
sub kcs_list_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    #Forcing this to cli, equivalent zapi unavailable.
    %args = (%args,
             cmd => "cluster_kernel_service_show",
             cmd_interface => "cli");
    unless (defined($args{node})){
        %args = (%args,
                 $self->_get_param_defaults());
    }

    # using base class (San::CommonInterface) to invoke a command via
    # SUPER::invoke to obtain igroup object data
    my @output = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        # error is already set, bubble up
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

} ## end sub kcs_list_info

###############################################################################
# Method Name: license_add
# Objective: install license
# Details:   See POD documentation at the end of this file.
###############################################################################
sub license_add {
    my ($self, %args) = @_;

    my (%license_codes, %cmd) = ();
    my (@licenses, @services) = ();
    my ($license, $service, $serial_number) = '';
    my $release_info='';

    # Clear errors
    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    # validate if service is a scalar or arrayref
    unless (defined $args{service} || defined $args{'license-code'}) {
        $self->_set_err("Mandatory parameter service and/or license-code not specified");
        return $SanUtil::FALSE;
    }

    # If license-code exists then add license-cdoes to license-code array
    if(defined $args{'license-code'}) {
        # If $args{'license-code'} is array ref then dereference it and install all
        # licenses for all services. Otherwise treat it like a service name
        if (ref($args{'license-code'}) eq "ARRAY") {
            @licenses = @{$args{'license-code'}};
        } else {
            push (@licenses, $args{'license-code'});
        }
    }

    # TODO, uncomment the following code line and get rid of the call to the
    # private method '_get_license_release()'.
    # Method '_get_license_release' was created as a work arround to the fact
    # that not all versions of SierraNevada had implemented the V2 licenses
    # strategy.
    # $release_info = $self->get_release();
    $release_info = $self->_get_license_release();
    if ($self->error()) {
        $self->_set_err("Failed to retrieve the release info". $self->errmsg());
        return $SanUtil::FALSE;
    }

    # If service exists then translate services to license-codes and add
    # to license-code array
    if(defined $args{service}) {

        # Get the serial number of the cluster
        if ( defined($args{serial_number}) ) {
            $serial_number = (delete ($args{serial_number}) );
        } elsif ($release_info =~ /SierraNevada_LicenseV2/){
            $serial_number =  "1-81-000000";
        } else {
            $serial_number = $self->_get_serial_number();
        }
        # Now, get the licenses using the serial number.
        %license_codes = $self->license_codes('serial_number'=>$serial_number);
        if ($self->error()) {
            $self->_set_err("\nFailed to retrieve licenses codes for serial " .
                            $serial_number . ": " . $self->errmsg());
            return $SanUtil::FALSE;
        }

        # If $args{service} is array ref then dereference it and install all
        # licenses for all services. Otherwise treat it like a service name
        if (ref($args{service}) eq "ARRAY") {
            @services = @{$args{service}};
        } else {
            push (@services, $args{service});
        }

        # Add the license for each service that has been passed in to the
        # license code array
        foreach $service (@services) {
            # If the service is snapmirror_db and the release is SN LicV2 convert to snapmirror
            if ($release_info =~ /SierraNevada_LicenseV2/ and $service eq "snapmirror_dp") {
                $service = "snapmirror";
            }
            unless (exists $license_codes{lc($service)}) {

                # If license code for given service is not found, set the error
                # and continue. At the end, we report all errors and return
                $self->_set_err("License code for specified service " . $service .
                                " not found");
                next;
            }

            next if ($self->license_exist(-service => $service));
            if ($self->error()) {
                # Error already set
                $self->_set_err("failed to verify if license is installed ",
                                $self->errmsg());
                return $SanUtil::FALSE;
            }

            push (@licenses, $license_codes{lc($service)});
        }
    }

    # Add the license for each license code that has been put in the license code array
    foreach $license (@licenses) {
        %cmd = (cmd            => "system_license_add",
                'license-code' => $license
                );
        $self->SUPER::invoke_cmd_obj(%cmd);
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj for adding the license " .
                            $self->dumper_get(\%cmd, $self->get_debug_level()),
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    if ($self->error()) {
        $self->_set_err("failed to add all given licenses successfully " .
                        $self->dumper_get(\%cmd, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: license_exist
# Objective: Return true if license for service exists otherwise false
# Details:   See POD documentation at the end of this file.
###############################################################################
sub license_exist {
    my ($self, %args) = @_;

    my (%cmd);
    my (@records);
    my $attribute = '';
    my $release_info = '';

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # validate if service is a scalar or arrayref
    unless (defined $args{service}) {
        $self->_set_err("Mandatory parameter service not specified");
        return $SanUtil::FALSE;
    }

    %cmd = (cmd => "system_license_show",
            "feature" => $args{service},
             cmd_interface => "cli");
    # FOR SN, LICENSES API CHANGE IN AN ATTRIBUTE NAME, FROM 'FEATURE' TO'PACKAGE'.
    # So, let's get the release
    #
    # TODO, uncomment the following code line and get rid of the call to the
    # private method '_get_license_release()'.
    # Method '_get_license_release' was created as a work arround to the fact
    # that not all versions of SierraNevada had implemented the V2 licenses
    # strategy.
    # $release_info = $self->get_release();

    $release_info = $self->_get_license_release();
    if ($self->error()) {
        $self->_set_err("Failed to retrieve the release info". $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Create the command
    %cmd = (cmd => "system_license_show");

    # Modify the command accordingly
    if ($release_info =~ /SierraNevada_LicenseV2/) {
        # Must use CLI because the not all ZAPI for SN License V2 is not supported or available
        my $output = $self->execute(-cmd => "system license show -package " . $args{service});

        # If an error occurs, the license is not installed
        if (grep(/There are no entries matching your query/i, @{$output})) {
            return $SanUtil::FALSE;
        } elsif (grep(/^Error:/, @{$output})) {
            $self->_set_err("Failed to retrieve the license info for the package " . $args{service});
            return $SanUtil::FALSE;
        } else {
            return $SanUtil::TRUE;
        }
    } else {
        $cmd{'feature'} = $args{service};
    }

    (@records) = $self->SUPER::invoke_cmd_obj(%cmd);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj " .
                        $self->dumper_get(\%cmd, $self->get_debug_level()),
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    # IF NO record was found with the 'service' name, the license for that
    # feature/package has not been added, so, return FALSE
    return  (scalar(@records)) ? $SanUtil::TRUE : $SanUtil::FALSE;
}

###############################################################################
# Method Name: license_delete()
# Objective: Deletes the license of a Service/Feature.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub license_delete {
    my ($self, %args) = @_;
    # Locals
    my (%cmd) = ();
    my $license = '';
    my $release_info = '';

    # Clear error
    $self->_clear_err();

    # Remove dashes
    %args = &SanUtil::strip_dash(%args);

    # Validate input parameters. Supporting 'service' because add and exist methods support it
    unless (defined $args{service} || defined $args{feature}|| defined $args{'package'}) {
        $self->_set_err("Mandatory parameter 'feature' or 'service' or 'package' not specified");
        return $SanUtil::FALSE;
    }

    # Backwards compatibility
    if ( defined($args{'feature'}) ) {
        $license = $args{'feature'};
    }
    elsif ( defined($args{'package'}) ) {
        $license = $args{'package'};
    } else {
        $license = $args{'service'};
    }

    if ( $self->license_exist(service => $license) ) {
        # FOR SN, LICENSES API CHANGE IN AN ATTRIBUTE NAME, FROM 'FEATURE' TO'PACKAGE'.
        # So, let's get the release
        #
        # TODO, uncomment the following code line and get rid of the call to the
        # private method '_get_license_release()'.
        # Method '_get_license_release' was created as a work arround to the fact
        # that not all versions of SierraNevada had implemented the V2 licenses
        # strategy.
        #
        #$release_info = $self->get_release();
        $release_info = $self->_get_license_release();
        if ($self->error()) {
            $self->_set_err("Failed to retrieve the release info". $self->errmsg());
            return $SanUtil::FALSE;
        }

        # Create the command
        %cmd = (cmd => "system_license_delete");

        # Modify the command accordingly
        if ($release_info =~ /SierraNevada_LicenseV2/) {
            $cmd{'package'} = $license;
        } else {
            $cmd{'feature'} = $license;
        }

        $self->SUPER::invoke_cmd_obj(%cmd);
        if ($self->error()) {
            $self->_set_err("Failed to invoke_cmd_obj " .
                            $self->dumper_get(\%cmd, $self->get_debug_level()),
                            $self->errmsg());
            return $SanUtil::FALSE;
        }
    } else {
        $self->_set_err("ERROR: Trying to delete the license '$license' that is ".
                        "not set\n");
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;

}##End license_delete

###############################################################################
# Method Name: license_codes
# Objective: Returns the licenses codes on a hash for the provided serial number
# Details:
###############################################################################
sub license_codes {
    my ($self, %args) = @_;

    # Verify that the required param has been provided.
    unless (defined $args{'serial_number'} ) {
        $self->_set_err("Failed to retrieve the licenses hash, mandatory " .
                        "parameter 'serial_number' missing " . $self->errmsg());
        return ();
    }
    # Locals
    my $licenses = undef;
    # TODO, uncomment the following code line and get rid of the call to the
    # private method '_get_license_release()'.
    # Method '_get_license_release' was created as a work arround to the fact
    # that not all versions of SierraNevada had implemented the V2 licenses
    # strategy.
    #
    #my $release_info = $self->get_release();
    my $release_info = $self->_get_license_release();
    if ($self->error()) {
        $self->_set_err("Failed to retrieve the release info". $self->errmsg());
        return ();
    }

    if ($release_info =~ /SierraNevada_LicenseV2/) {
        $licenses = SanConsts::SierraNevada_Licenses->{$args{serial_number}};
    } else {
        $licenses = SanConsts::RollingRock_Licenses->{$args{serial_number}};
    }

    # If LICNESES is empty, a wrong serial number was provided.
    unless($licenses) {
        $self->_set_err("Failed to retrieve the licenses hash, provided an invalid " .
                        "'serial_number': " . $args{'serial_number'});
        return ();
    }

    return (wantarray()) ? %{$licenses} : $licenses;
}## license_codes

###############################################################################
# Method Name: get_hostrec
# Objective:   Get filer hostrec
# Details:     See POD documentation at the end of this file.
###############################################################################
sub get_hostrec {
    my ($self, $value) = @_;

    $self->_clear_err();

    my $hostrec = $self->connection()->hostp();

    unless (ref($hostrec) =~ /HostRec/i) {
        $self->_set_err("failed to get hostrec: " . $self->errmsg());
        return {};
    }
    return $hostrec;
}

###############################################################################
# Method Name: is_vsim
# Objective:   Return true if the node is vsim
# Details:     See POD documentation at the end of this file.
###############################################################################
sub is_vsim {
    my ($self, $value) = @_;

    $self->_clear_err();

    my $host_obj = $self->get_hostrec();
    if ($self->error()) {
        $self->_set_err("fail to get hostrec: " . $self->errmsg());
        return $SanUtil::FALSE;
    }

    unless ($host_obj->hosttype() =~ /^vsim-ng/i) {
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: fetch_lif_objs
# Objective: Fetch a list of LIF objects from the system for the caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fetch_lif_objs {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "network_interface_show",
             cmd_interface => "cli");

    my @output = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",$self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Using List::MoreUtils::any to find string in an array
    if (List::MoreUtils::any
        {$_ =~ /Table is currently empty/i} @output) {
        return (wantarray()) ? () : [];
    }

    my @lif_objs = ();

    # build a lun object for every lun element returned by the command
    foreach my $elm (@output) {
        my $lif_obj = new San::ControllerCluster::Filer10M::Vserver::Lif(%{$elm},
                                                                         -filer       => $self->get_name(),
                                                                         -parent_obj  => $self,
                                                                         -initialized => 1);
        if ($lif_obj->error()) {
            $self->_set_err("Failed to populate LIF object", $lif_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        push(@lif_objs, $lif_obj);
    }

    return (wantarray()) ? @lif_objs : \@lif_objs;
} ## fetch_lif_objs

###############################################################################
# Method Name: fetch_lun_objs
# Objective: Fetch a list of LUN objects from the system for the caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fetch_lun_objs {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_show",
             cmd_interface => "cli");

    my @output = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",$self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Using List::MoreUtils::any to find string in an array
    if (List::MoreUtils::any
        {$_ =~ /Table is currently empty/i} @output) {
        return (wantarray()) ? () : [];
    }

    my @lun_objs = ();

    # build a lun object for every lun element returned by the command
    foreach my $elm (@output) {
        my $lun_obj = new San::ControllerCluster::Filer10M::Vserver::Volume::Lun(%{$elm},
                                                                                 -filer => $self->get_name(),
                                                                                 -parent_obj => $self,
                                                                                 -initialized => $SanUtil::TRUE);
        if ($lun_obj->error()) {
            $self->_set_err("Failed to populate LUN object", $lun_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        push(@lun_objs, $lun_obj);
    }

    return (wantarray()) ? @lun_objs : \@lun_objs;
} ## fetch_lun_objs

###############################################################################
# Method Name: fetch_volume_objs
# Objective: Fetch a list of Volume objects from the system for the caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fetch_volume_objs {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_show",
             cmd_interface => "cli");

    my @output = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",$self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Using List::MoreUtils::any to find string in an array
    if (List::MoreUtils::any
        {$_ =~ /Table is currently empty/i} @output) {
        return (wantarray()) ? () : [];
    }

    my @volume_objs = ();

    # build a vol object for every vol element returned by the command
    foreach my $elm (@output) {
        my $volume_obj = new San::ControllerCluster::Filer10M::Vserver::Volume(%{$elm},
                                                                               -filer => $self->get_name(),
                                                                               -parent_obj => $self,
                                                                               -initialized => $SanUtil::TRUE);
        if ($volume_obj->error()) {
            $self->_set_err("Failed to populate Volume object", $volume_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        push(@volume_objs, $volume_obj);
    }

    return (wantarray()) ? @volume_objs : \@volume_objs;
} ## fetch_volume_objs

###############################################################################
# Method Name: fetch_vserver_objs
# Objective: Fetch a list of Vserver objects from the system for the caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fetch_vserver_objs {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "vserver_show",
             cmd_interface => "cli");

    # default 'type' to be cluster, since vserver show also displays cservers and nodes
    # as well as vservers, this will scope it to only vserver output by default
    if (not exists($args{type})) {
        $args{type} = "cluster|data";
    }

    my @output = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",$self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Using List::MoreUtils::any to find string in an array
    if (List::MoreUtils::any
        {$_ =~ /Table is currently empty/i} @output) {
        return (wantarray()) ? () : [];
    }

    my @vserver_objs = ();

    # build a vserver object for every vserver element returned by the command
    foreach my $elm (@output) {
        my $vserver_obj = new San::ControllerCluster::Filer10M::Vserver(%{$elm},
                                                                        -filer => $self->get_name(),
                                                                        -parent_obj => $self,
                                                                        -initialized => $SanUtil::TRUE);
        if ($vserver_obj->error()) {
            $self->_set_err("Failed to populate Vserver object", $vserver_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        push(@vserver_objs, $vserver_obj);
    }

    return (wantarray()) ? @vserver_objs : \@vserver_objs;
} ## fetch_vserver_objs

###############################################################################
# Method Name: fetch_igroup_objs
# Objective: Fetch a list of Igroup objects from the system for the caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fetch_igroup_objs {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_igroup_show",
             cmd_interface => "cli");

    my @output = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",$self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Using List::MoreUtils::any to find string in an array
    if (List::MoreUtils::any
        {$_ =~ /Table is currently empty/i} @output) {
        return (wantarray()) ? () : [];
    }

    my @igroup_objs = ();

    # build a igroup object for every igroup element returned by the command
    foreach my $elm (@output) {
        my $igroup_obj = new San::ControllerCluster::Filer10M::Vserver::Igroup(%{$elm},
                                                                               -filer => $self->get_name(),
                                                                               -parent_obj => $self,
                                                                               -initialized => $SanUtil::TRUE);
        if ($igroup_obj->error()) {
            $self->_set_err("Failed to populate Igroup object", $igroup_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        push(@igroup_objs, $igroup_obj);
    }

    return (wantarray()) ? @igroup_objs : \@igroup_objs;
} ## fetch_igroup_objs

###############################################################################
# Method Name: fetch_qtree_objs
# Objective: Fetch a list of QTREE objects from the system for the caller
# Details:   See POD documentation at the end of this file.
###############################################################################
sub fetch_qtree_objs {
    my ($self, %args) = @_;

    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);

    # initialize cmd parameters
    %args = (%args,
             cmd => "volume_qtree_show",
             cmd_interface => "cli");

    my @output = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",$self->errmsg());
        return (wantarray()) ? () : [];
    }

    # Using List::MoreUtils::any to find string in an array
    if (List::MoreUtils::any
        {$_ =~ /Table is currently empty/i} @output) {
        return (wantarray()) ? () : [];
    }

    my @qtree_objs = ();

    # build a qtree object for every qtree element returned by the command
    foreach my $elm (@output) {
        my $id_field = (defined $elm->{"qtree_id"}) ? "qtree_id" : "qtree-id";

        # Make sure the qtree id is not zero otherwise we are not
        # really looking at a qtree (might be a volume).
        next if ($elm->{"$id_field"} == 0);

        # In zapi the path comes back as "-" which is useless.  If it doesn't
        # include the qtree lets set it properly.
        if ($elm->{path} !~ /$elm->{qtree}/) {
            $elm->{path} = "/vol/" . $elm->{volume} . "/" . $elm->{qtree};
            $elm->{"qtree_path"} = $elm->{path};
            delete $elm->{"qtree-path"} if ($elm->{"qtree-path"});
        }
        my $qtree_obj = new San::ControllerCluster::Filer10M::Vserver::Volume::Qtree(
                                            %{$elm},
                                            -filer => $self->get_name(),
                                            -parent_obj => $self,
                                            -initialized => 1);
        if ($qtree_obj->error()) {
            $self->_set_err("Failed to populate QTREE object",
                            $qtree_obj->errmsg());
            return (wantarray()) ? () : [];
        }

        push(@qtree_objs, $qtree_obj);
    }

    return (wantarray()) ? @qtree_objs : \@qtree_objs;
} ## fetch_qtree_objs

################################################################################
# Method Name: _reset_connection
# Objective: Connects to a node, executes a command and disconnects.
# Params: connection object
# Returns: A string containing the output returned by the command.
################################################################################
sub _reset_connection {
    my ($self, $conn) = @_;

    my $output = "";

    $self->_clear_err();

    # Make sure we have a node object
    unless (defined($conn)) {
        $self->_set_err("mandatory parameter conn is missing!");
        return "";
    }

    # disconnect and logout console connection
    &Tharn::disconnect($conn, -logout);

    # Now try to connect to the node
    $conn = $self->console_connect(-username => $self->DIAG_USER,
                                   -password => $self->DEFAULT_PASSWD);
    if ($self->error()) {
        $self->_set_err("failed to connect to node console" .
                        $self->errmsg());
        return "";
    }

    # Execute a dummy command and detect the prompt output
    eval {
        $output = (Tharn::say($conn, " ", -timeout, $self->CMD_TIMEOUT))[$Return_index];
    };
    if ($@) {
        $self->_set_err("failed to call say(): ", $@);
        return "";
    }
    $self->sanlog(msg => "prompt is $output", -msg_level => 7);

    return $output;
} # End of _reset_connection


###############################################################################
# Method:    cluster_ha_show()
# Objective: Will return an array with the Cluster HA show output hash.
# Details:   See POD documentation at the end of this file
###############################################################################
sub cluster_ha_show{
    my ($self, %args) = @_;
    my (@output);
    # Clear the error flag
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);
    %args = (%args, cmd => "cluster_ha_show");

    # call cluster_ha_show
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",
                        $self->errmsg());
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;
}# End cluster_ha_show

###############################################################################
# Method:    cluster_ha_modify()
# Objective: Modifies Cluster HA configured state.
# Details:   See POD documentation at the end of this file
###############################################################################
sub cluster_ha_modify{
    my ($self, %args) = @_;
    my (@output);
    # Clear the error flag
    $self->_clear_err();
    # Verify the input has been defined and is either True or False
    %args = &SanUtil::strip_dash(%args);
    unless (defined $args{configured}) {
        $self->_set_err("Mandatory parameter 'configured' not specified");
        return $SanUtil::FALSE;
    }
    unless ( (lc($args{configured}) eq "true")||(lc($args{configured}) eq "false") ){
        $self->_set_err("Parameter 'configured' can only be true|false");
        return $SanUtil::FALSE;
    }
    %args = (%args, cmd => "cluster_ha_modify");

    # Execute command
    (@output) = $self->SUPER::invoke_cmd_obj(%args, -update_attrs => 0);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
}# End cluster_ha_modify


###############################################################################
# Method:    is_cluster_ha_configured()
# Objective: Returns TRUE if the Cluster HA has been properly configured
# Details:   See POD documentation at the end of this file
###############################################################################
sub is_cluster_ha_configured {
    my ($self, %args) = @_;
    my (@output);
    # Clear the error flag
    $self->_clear_err();

    # call cluster_ha_modify
    (@output) = $self->cluster_ha_show();
    if ($self->error()) {
        $self->_set_err ("failed in show the cluster HA", $self->errmsg());
        return $SanUtil::FALSE;
    }

    if ($output[0]->{configured} eq "true") {
        return $SanUtil::TRUE;
    }else{
        return $SanUtil::FALSE;
    }

}# End is_cluster_ha_configured

###############################################################################
# Method:    cluster_ha_disable()
# Objective: Deconfigures cluster HA if it was configured
# Details:   See POD documentation at the end of this file
###############################################################################
sub cluster_ha_disable{
    my ($self, %args) = @_;
    my (@output);
    # Clear the error flag
    $self->_clear_err();
    %args = &SanUtil::strip_dash(%args);

    unless($self->is_cluster_ha_configured()) {
        return $SanUtil::TRUE;
    }

    # De-configure Cluster HA
    $self->cluster_ha_modify(-configured => "false");
    if ($self->error()) {
        $self->_set_err("Failed to deconfigure cluster HA",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;

} #End cluster_ha_disable

###############################################################################
# Method Name: lun_serials_are_unique
# Objective: Return TRUE if all the LUNs serial numbers are unique, FALSE if not.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub lun_serials_are_unique {
    my ($self, %args) = @_;
    # Locals
    my (@output,@serial_nums) = ();
    my $duplicates_ref = [];
    # Clear error
    $self->_clear_err();

    # remove the dash '-' from input args if exist
    %args = &SanUtil::strip_dash(%args);
    # Get parameters
    if ( defined($args{'fields'}) && ($args{'fields'} ne "") ) {
        $args{'fields'} = $args{'fields'} . ",serial";
    } else {
        $args{'fields'} = "serial";
    }
    # Delete the reference for the array where serials will be returned, in case
    # duplicates are found.
    if ( defined($args{'duplicates'}) ) {
        $duplicates_ref = delete $args{'duplicates'};
    }

    # initialize cmd parameters
    %args = (%args,
             cmd => "lun_show");

    @output = $self->SUPER::invoke_cmd_obj(%args);
    if ($self->error()) {
        $self->_set_err("failed to invoke_cmd_obj \"$args{cmd}\"",$self->errmsg());
        return $SanUtil::FALSE;
    }

    # Verify if the returned array has elements.
    unless(scalar(@output)) {
        # No records found, means no LUNs available for the search criteria.
        $self->sanlog(msg => "\n- LUN query returned empty, could NOT verify the ".
                      "uniqueness of the serial numbers");
        return $SanUtil::FALSE;
    }

    # Build an array with all the found LUN serial numbers
    foreach my $lun_record (@output) {
        push(@serial_nums, $lun_record->{serial});
    }

    # Get the duplicated elements of the array
    @{$duplicates_ref} = &SanUtil::array_get_duplicates(@serial_nums);

    if ( scalar(@{$duplicates_ref})>0) {
        $self->sanlog(msg => "\n-Found duplicated serials: " .
                      join(',', (@{$duplicates_ref}))
                    );
        return $SanUtil::FALSE;
    }

    #No duplicates
    return $SanUtil::TRUE;
}# lun_serials_are_unique

###############################################################################
# Method Name: terminate_savecore_process
# Objective: Given a Node that is currently in a takeover state. terminate any savecore process.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub terminate_savecore_process {
    my ($self, %args) = @_;

    %args = &SanUtil::strip_dash(%args);

    my @args = (%args);

    eval {
        (%args) =
          Module::Common::validate_mixed(@args,
                                         [],
                                         {
                                             node => {type => SCALAR,
                                                      optional => 1,
                                                      default => $self->get_name()},
                                         });
    };
    if ($@) {
        $self->_set_err("failed to validate parameter: " . $@);
        return $SanUtil::FALSE;
    }

    # initialize the core name attribute if its not defined
    if (not defined($self->{_core_name})) {
        $self->{_core_name} = "";
    }

    # Check to make sure the filer is in a failover state
    my $cmd = "cf status";
    my $output = $self->run_node_scope(-cmd => $cmd,
                                       -filer => $args{node});
    if ($self->error()) {
        $self->_set_err("Failed to run command '$cmd'", $self->errmsg());
        return $SanUtil::FALSE;
    }

    if ($output =~ /is up/) {
        # in this case, the controller ha pair is not in a failover state
        $self->_set_err("The controller HA pair is not in a failover state.");
        return $SanUtil::FALSE;
    }

    # check to see if partner has completed core and collect the core name
    $cmd = "partner savecore -s";
    $output = $self->run_node_scope(-cmd => $cmd,
                                    -filer => $args{node});
    if ($self->error()) {
        $self->_set_err("Failed to run command '$cmd'", $self->errmsg());
        return $SanUtil::FALSE;
    }

    if (ref($output) eq "ARRAY") {
        $output = $output->[0];
    }

    if ($output =~ /Saving\s(.+)\:\s/i && $self->{_core_name} eq "") {
        $self->{_core_name} = $1;
        $self->sanlog(-msg => "Core file found " . $self->{_core_name});
        return $SanUtil::TRUE;
    } elsif ($output =~ /Saving\s(.+)\:\s/i && $self->{_core_name} ne "") {
        $output =~ /Saving\s(.+)\:\s/;
        if ($self->{_core_name} eq $1) {
            $self->sanlog(-msg => "Found automation induced core that is still saving, will attempt to kill it.");

            # Have to kill the core, and go on
            $cmd = "partner savecore -k";
            $self->run_node_scope(-cmd => $cmd,
                                  -filer => $args{node});
            if ($self->error()) {
                $self->_set_err("Failed to run command '$cmd'", $self->errmsg());
                return $SanUtil::FALSE;
            }

            $cmd = "partner savecore -s";
            $output = $self->run_node_scope(-cmd => $cmd,
                                            -filer => $args{node});
            if ($self->error()) {
                $self->_set_err("Failed to run command '$cmd'", $self->errmsg());
                return $SanUtil::FALSE;
            }

            if (ref($output) eq "ARRAY") {
                $output = $output->[0];
            }

            if ($output =~ /Saving\s(.+):\s/i) {
                &SanUtil::nate_log('-warn', "Core deletion failed");
                &SanUtil::nate_sleep(10);
                return $SanUtil::TRUE;
            } else {
                $self->sanlog(-msg => "Core deletion completed");
            }

        } else {
            # in this case we found a new core that was not induced by the automation, this should be a failure
            $self->_set_err("Detected a new core that was not induced by the automation: '$1'");
            return $SanUtil::FALSE;
        }
    } elsif ($output =~ /No core need|No savecore in progress/i) {
        $self->sanlog(-msg => "Did not find a core file being saved");
        return $SanUtil::TRUE;
    } elsif ($output =~ /Generating mini-core/) {
        $self->sanlog(-msg => "Found mini-core: $output");
        # Sleep 60 seconds, it usually takes 40-50 seconds to complete
        # the mini-core generation
        &SanUtil::nate_sleep(60);
        return $SanUtil::TRUE;
    } elsif ($output =~ /Preparing to save core/i) {
        $self->sanlog(-msg => "Nodes are preparing to save core");
        # Sleep 15 seconds to allow the filers to start creating the core,
        # then return 0
        &SanUtil::nate_sleep(15);
        return $SanUtil::TRUE;
    } elsif ($output =~ /Waiting for partner dump to complete/i) {
        $self->sanlog(-msg => "Nodes are waiting for partner to dump core");
        # Sleep 60 seconds to allow the filers to start creating the core,
        # then return 0
        &SanUtil::nate_sleep(60);
        return $SanUtil::TRUE;
    } else {
        $self->sanlog(-msg => "Did not find a core file being saved");
    }

    # All went well, disconnect and return a good status
    return $SanUtil::TRUE;
}

###############################################################################
# Method Name: _get_serial_number
# Objective: Retreives the current license or the default one
# Details:
###############################################################################
sub _get_serial_number {
    my ($self, %args) = @_;
    # Locals
    my @serials=();
    my %cmd = {};
    my $default_serial = SanConsts::DEFAULT_SERIAL_NUMBER;
    my $use_default = defined($args{'use_default'}) ? $args{'use_default'} : $SanUtil::FALSE;

    #If requested to get the default.
    unless ($use_default){
        # If any licenses have been added with a serial, return it
        %cmd = ('cmd'   => "system_license_show",
                'fields'=> "serial-number");
        @serials = $self->SUPER::invoke_cmd_obj(%cmd);
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj " .
                            $self->dumper_get(\%cmd, $self->get_debug_level()),
                            $self->errmsg());
            return undef;
        }

        if ( scalar(@serials) ) {
            foreach my $serial (@serials) {
                if ( $serial->{feature} eq "Base") {
                    return $serial->{serial_number};
                }
            }
        }
    }

    # If no serial was provided or found or user asked for the default, return the
    # default serial number
    return $default_serial;
}

#===============================================================================
# Method Name: _get_license_release
# Objective: Return a name with LICENSE V2 if that releas is found
#===============================================================================
sub _get_license_release {
    my ($self, %args) = @_;
    # Locals

    if ( defined($self->{'license_release'}) ){
        return $self->{'license_release'};
    } else {
        my @records = $self->SUPER::invoke_cmd_obj(cmd => "system_license_show");
        if ($self->error()) {
            $self->_set_err("failed to invoke_cmd_obj 'system_license_show'" .
                            $self->errmsg());
            return undef;
        }
        #If attribute PACKAGE is found, then it's a V2 License Release
        if (exists($records[0]->{'package'}) ){
            return $self->{'license_release'} = "SierraNevada_LicenseV2";
        } else {
            #If not, just return a name, will not be actually reading this one.
            return $self->{'license_release'} = "RollingRock__LicenseV1";
        }
    }
}## _get_license_release

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

FilerHelper.pm

=head1 Author

Korry Nguyen (dl-ssan-automation)

=head1 Description

FilerHelper.pm provides helper methods for Filer10M. This package maintains separately from Filer10M for maintainability purpose while preventing convoluting Filer10M object model.

=head1 See Also

FilerHelper package is used as part of the base class by B<San::ControllerCluster::Filer10M>. See these modules for further interface detail.


B<San::ControllerCluster::Filer10M.pm, San::CommonInterface.pm>

=head1 B<Public Methods>

=head2 B<aggregate_create(%args)>

Create an aggregate.

=over 4

=item Inputs

=item * I<-name> (default: test_aggr)

The aggregate name to use

=item * I<-num_disks> (default: 3)

The number of disks in the aggregate

=item * I<-timeout> (default: 3600)

How long to wait for aggregate to come online

=item * I<-raidtype> (default: raid_dp)

The raidtype of the aggregate, can be raid4 or raid_dp

=item * I<-raidsize> (default: default ONTAP raidsize)

The raidsize of the aggregate, can be between 2 and 28

=item Outputs

Returns aggregate object. if error, sets error and returns empty hash.

=item Examples

 # ex. 1: Create aggregate
 my $aggr_obj = $filer_obj->aggregate_create(-name => test_aggr,
                                             -num_disks => 5);
 if ($filer_obj->error){
     ...
 }

=back

=head2 B<aggregate_offline(%args)>

Offline an aggregate

=over 4

=item Inputs

=item * I<-name> (required)

Aggregate name to offline

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Offline an aggregate
 $filer_obj->aggregate_offline(-name => "test_aggr");
 if ($filer_obj->error()) {
     ...
 }

=back

=head2 B<configure_lif(%args)>

Configure a generic lif on a Filer10M object including the gateway.

=over 4

=item Inputs

Parameters required to configure a lif this includes the gateway ip.  See example below.

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

    # Build lif args from hostrec hash.
    my %lif_args = (
        vserver   => $clus_name,
        lif       => "cluster_mgmt",
        role      => "cluster-mgmt",
        home_node => $node_obj->get_name(),
        home_port => $host->{CLUSTER_MGMT_PORT},
        address   => $host->{CLUSTER_MGMT_IP},
        netmask   => $host->{CLUSTER_MGMT_MASK},
        gateway   => $host->{CLUSTER_MGMT_GATEWAY}
    );

    $node_obj->configure_lif(%lif_args);
    if ($node_obj->error()) {
        # error set return
        return undef;
    }

=back

=head2 B<configure_cluster_lifs(%args)>

Configure the 2 required cluster lifs for a node to join a cluster.

=over 4

=item Inputs

None

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

    $Node_Obj->configure_cluster_lifs();
    if ($Node_Obj->error()) {
        $Test->nlog("Failed to configure cluster lifs on $node."
                    . $Node_Obj->errmsg());
        return $TCD::FAIL;
    }

=back

=head2 B<create_volume(%args)>

Create a volume object

=over 4

=item Inputs

=item * I<-name> (required)

Volume name to create

=item * I<-aggr> (required)

Aggregate that the volume will be created on

=item * I<-size> (required)

Volume size to be created

=item Outputs

Returns a volume object. If error, sets error and returns empty hash.

=item Examples

 # ex. 1: Create a volume
 my $vol_obj = $filer_obj->create_volume(-name => "my_vol",
                                         -aggr => $aggr,
                                         -size => "100g");
 if ($filer_obj->error()){
     ...
 }

=back

=head2 B<get_aggregate(%args)>

Get specific aggregate object

=over 4

=item Inputs

=item * I<-name> (required)

The name of the aggregate to retrieve

=item Outputs

Returns an aggregate object. If error, error message is pushed onto the error stack.

=item Examples

 # ex. 1:  Get aggregate object
 $aggr_obj = $filer_obj->get_aggregate(-name => "test_aggr");
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_aggregates()>

Get aggregate objects

=over 4

=item Inputs

None

=item Outputs

Returns an array of aggregate objects. If error, error message is pushed onto the error stack.

=item Examples

 # ex. 1:  Get aggregate objects
 @aggr_objs = $filer_obj->get_aggregates();
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_fcpadapter_objs()>

Get fcp adapter objects

=over 4

=item Inputs

None

=item Outputs

Returns an array of fcp adapter objects. If error, error message is pushed onto the error stack.

=item Examples

 # ex. 1:  Get fcp adapter objects
 @aggr_objs = $filer_obj->get_fcpadapter_objs();
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_network_interfaces()>

Get network interfaces

=over 4

=item Inputs

None

=item Outputs

Returns an array network interfaces. If error, error message is pushed onto the error stack and return empty array.

=item Examples

 # ex. 1:  Get network interface
 @net_interfaces = $filer_obj->get_network_interfaces();
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_portsets(%args)>

Get portset objects

=over 4

=item Inputs

=item * I<-vserver> (optional)

Vserver name to get portset objects from. If not specified, portsets will be retrieved from the default vserver (vs0)

=item Outputs

Returns an array of portset objects. If error, error message is pushed onto the error stack and return empty array.

=item Examples

 # ex. 1:  Get portset objects
 @ps_objs = $filer_obj->get_portsets();
 if ($filer_obj->error()) {
    ...
 }

 # ex. 2:  Get portset objects on specific vserver
 @ps_objs = $filer_obj->get_portsets(-vserver => $vs_name);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_volumes(%args)>

Get volume objects

=over 4

=item Inputs

=item * I<-vserver> (optional)

Vserver name to get volume objects from. If not specified, targets will be retrieved from the default vserver (vs0)

=item Outputs

Returns an array of volume objects. If error, error message is pushed onto the error stack and return empty array.

=item Examples

 # ex. 1:  Get volume objects
 @vol_objs = $filer_obj->get_volumes();
 if ($filer_obj->error()) {
    ...
 }

 # ex. 2:  Get volume objects on specific vserver
 @vol_objs = $filer_obj->get_volumes(-vserver => $vs_name);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<get_vservers(%args)>

Get vserver objects

=over 4

=item Inputs

None

=item Outputs

Returns an array of vserver objects. If error, error message is pushed onto the error stack and return empty array.

=item Examples

 # ex. 1:  Get vserver objects
 @vs_objs = $filer_obj->get_vservers();
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<interlace_disks()>

Interlace all available disks

=over 4

=item Inputs

None

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Interlace the disks
 ($filer_obj->interlace_disks());
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<volume_exist()>

Determine whether the volume exists or not

=over 4

=item Inputs

=item * I<-name | -volume> (required)

Volume name to use

=item * I<-vserver> (optional)

Vserver name to get fcp nodename from. If not specified, the nodename will be retrieved from the default vserver (vs0)

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Determine if volume exists (7mode compatibility)
 if ($filer_obj->vol_exist("my_vol")) {
     # get vol object
 } else {
     # create vol object
 }

 # ex. 2: Determine if volume exists on a specified vserver (7mode style)
 if ($filer_obj->vol_exist("my_vol", "my_vserver")) {
     # get vol object
 } else {
     # create vol object
 }

 # ex. 3: Determine if volume exists on a specified vserver
 if ($filer_obj->volume_exist(-name => "my_vol",
                              -vserver => "my_vserver")) {
     # get vol object
 } else {
     # create vol object
 }

=back

=head2 B<wait_for_aggregate_online(%args)>

Wait for aggregate to be online

=over 4

=item Inputs

=item * I<-name | -aggregate> (required)

=item Outputs

Returns TRUE if successful, otherwise sets error and returns FALSE

=item Examples

 # ex. 1: Create aggregate
 $aggr_obj = $filer_obj->aggr_create(-name => $aggr_name);
 if ($filer_obj->error()) {
    ...
 }

 # Wait for aggregate to go online
 $filer_obj->wait_for_aggr_online(-name => $aggr_name);
 if ($filer_obj->error()) {
     ...
 }

=back

=head2 B<is_sfo_enabled()>

Determine whether sfo is enabled

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE if sfo is enabled, otherwise $SanUtil::FALSE

=item Examples

 # determine whether sfo is enabled
 $is_sfo_enabled = $filer_obj->is_sfo_enabled();
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<enable_sfo()>

Enable sfo if its not enabled

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE if sfo is enabled successfully otherwise sets error and
returns $SanUtil::FALSE

=item Examples

 # enable sfo
 $filer_obj->enable_sfo();
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<disable_sfo()>

Disable sfo if its not disabled

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE if sfo is disabled successfully otherwise sets error and
returns $SanUtil::FALSE

=item Examples

 # disable sfo
 $filer_obj->disable_sfo();
 if ($filer_obj->error()) {
    ...
 }

=back


=head2 B<detect_sfo_takeover(%args)>

Detect how long it takes for an SFO takeover to complete.

=over 4

=item Inputs

=item * I<-takeover_node> (optional)

The node to perform takeover detection on.
Default: the current node.

=item * I<-sfo_timeout> (optional)

If the takeover does not occur within this many seconds, then it will keep detecting
until sfo_max_timeout is exhausted.
Default: 300 (seconds)

=item * I<-sfo_max_timeout> (optional)

The maximum amount of time to keep detecting whether the sfo takeover completes or not.
If the takeover does not occur within this time frame, an error will be set.
Default: 900 (seconds)

=item * I<-cmd_timeout> (optional)

If the cmd used to detect the takeover timesout within this time, then error out
Default: the default cdef timeout set on the object

=item Outputs

Returns the amount of time it took for the takeover to occur,
$SanUtil::FALSE and sets error otherwise.

=item Examples

 # detect how long the takeover took
 $time = $filer_obj->detect_sfo_takeover(-sfo_timeout => 100);
 if ($filer_obj->error()) {
    ...
 }

 if ($time < 60) {
     print "good job!";
 } else {
     print "hmmm...";
 }

=back

=head2 B<detect_sfo_giveback(%args)>

Detect how long it takes for an SFO giveback to complete.

=over 4

=item Inputs

=item * I<-giveback_node> (optional)

The node to perform giveback detection on.
Default: the current node.

=item * I<-sfo_timeout> (optional)

If the giveback does not occur within this many seconds, then it will keep detecting
until sfo_max_timeout is exhausted.
Default: 300 (seconds)

=item * I<-sfo_max_timeout> (optional)

The maximum amount of time to keep detecting whether the sfo giveback completes or not.
If the giveback does not occur within this time frame, an error will be set.
Default: 900 (seconds)

=item Outputs

Returns the amount of time it took for the giveback to occur,
$SanUtil::FALSE and sets error otherwise.

=item Examples

 # issue a giveback
 $filer_obj->sfo_giveback();
 if ($filer_obj->error()) {
    ...
 }

 # detect how long the giveback took
 $time = $filer_obj->detect_sfo_giveback(-sfo_timeout => 100, -sfo_max_timeout => 300);
 if ($filer_obj->error()) {
    ...
 }

 if ($time < 100) {
     print "good job!";
 } else {
     print "hmmm...";
 }

=back

=head2 B<detect_waiting_for_giveback(%args)>

Detect how long it takes for a controller to reach the waiting for giveback prompt.
Assuming it is currently being booted or was recently power on and is in a takeover state.

=over 4

=item Inputs

=item * I<-node> (optional)

The node to detect for the waiting for giveback prompt.
Default: the current node

=item * I<-take_over_node> (optional)

The node that has performed the take over.
Default: empty

=item * I<-timeout> (optional)

If this amount of time has elapsed and -node has still not reached the prompt, error out
Default: 300 (seconds)

=item Outputs

Returns the amount of time it took to reach the waiting for giveback state.
$SanUtil::FALSE and sets error otherwise.

=item Examples

 # detect how long it took for filer_obj to reach the waiting for giveback prompt
 $time = $filer_obj->detect_waiting_for_giveback(-timeout => 200);
 if ($filer_obj->error()) {
    ...
 }

 # detect how long it took for node_a to reach the waiting for giveback prompt
 $time = $filer_obj->detect_waiting_for_giveback(-node => "node_a",
                                                 -take_over_node => "node_b");
 if ($filer_obj->error()) {
    ...
 }

 # detect how long it took for node_a to reach the waiting for giveback prompt
 $time = $filer_obj->detect_waiting_for_giveback(-node => "node_a");
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<wait_for_node_to_boot(%args)>

Wait for the specified node to boot.

=over 4

=item Inputs

=item * I<-node> (optional)

The node to detect for boot
Default: the current node.

=item * I<-timeout> (optional)

If this amount of time has elapsed and -node has still not reached the prompt, error out
Default: 600 (seconds)

=item Outputs

Returns the amount of time it took to boot the node
$SanUtil::FALSE and sets error otherwise.

=item Examples

 # detect how long it takes to boot the current node
 $time = $filer_obj->wait_for_node_to_boot(-timeout => 200);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<panic_remote_node(%args)>

Issue a panic on a remote node (not the current node).

=over 4

=item Inputs

=item * I<-node> (required)

The node to issue the panic against. (Cannot be the current node)

=item Outputs

Returns $SanUtil::TRUE if succesful, $SanUtil::FALSE and sets error otherwise.

=item Examples

 # panic the HA partner
 $filer_obj->panic_remote_node(-node => $filer_obj->get_partner());
 if ($filer_obj->error()) {
    ...
 }

 # detect how long it took for the takeover to occur
 $filer_obj->detect_sfo_takeover(-sfo_timeout => 120);
 ...

=back

=head2 B<boot_node(%args)>

Boot the specified node over a console connection.

=over 4

=item Inputs

=item * I<-node> (optional)

The node to boot.
Default: the current node.

=item Outputs

Returns the boot state that the specified node is currently at.

=item Examples

 # boot the ha partner node
 $filer_obj->boot_node(-node => $filer_obj->get_partner());
 if ($filer_obj->error()) {
    ...
 }

 # detect how long it took for the partner to reach the waiting for giveback prompt
 $filer_obj->detect_waiting_for_giveback(-timeout => 300,
                                         -node => $filer_obj->get_partner());
 ...

=back

=head2 B<halt_node(%args)>

Halt the specified node over a console connection.

=over 4

=item Inputs

=item * I<-node> (optional)

The node to halt.
Default: the current node.

=item * I<-halt_timeout> (optional)

The timeout to pass into the halt command
Default: 300 (seconds).

=item * I<-inhibit_takeover> (optional)

Control whether to inhibit a takeover or not. By default a takeover will occur.
Default: false.

=item * I<-dump> (optional)

Control whether to create a core dump.
Default: false.

=item Outputs

Returns $SanUtil::TRUE if succesful, $SanUtil::FALSE and sets error otherwise.

=item Examples

 # halt the ha partner node
 $filer_obj->halt_node(-node => $filer_obj->get_partner());
 if ($filer_obj->error()) {
    ...
 }

 # detect how long it took for the takeover to occur
 # Note: it is recommended that the detection and halt action be done in parallel to
 # achieve accurate results.
 $filer_obj->detect_sfo_takeover(-sfo_timeout => 120);
 ...

=back

=head2 B<disable_auto_giveback(%args)>

Disable the auto giveback feature on the specified node

=over 4

=item Inputs

=item * I<-node> (optional)

The node to make the SFO modification.
Default: the current node.

=item Outputs

Returns $SanUtil::TRUE if succesful, $SanUtil::FALSE and sets error otherwise.

=item Examples

 # disable auto giveback on this node and its HA partner
 $filer_obj->disable_auto_giveback();
 if ($filer_obj->error()) {
    ...
 }
 $filer_obj->disable_auto_giveback(-node => $filer_obj->get_partner());
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<console_connect(%args)>

Connect over a given connection type (defaults to the console) on the specified node.

=over 4

=item Inputs

=item * I<-node> (optional)

The node to connect to.
Default: the current node.

=item * I<-conn_type> (optional)

The type of connection
Default: console

=item * I<-login> (optional)

Controls whether to perform a login or not
Default: 0 (perform the login)

=item * I<-conn_timeout> (optional)

The timeout to use when creating the connection.
Default: 300 (seconds)

=item Outputs

Returns a valid ConnectRec handle if successful, sets error otherwise.

=item Examples

 # connect to this node's console
 my $conn = $filer_obj->console_connect();
 if ($filer_obj->error()) {
    ...
 }

 ...

 # disconnect when done.
 &Tharn::disconnect($conn);

=back

=head2 B<fetch_lun_objs(%args)>

Fetch a list of LUN objects for the caller to use by discovering them on the system.

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the lun show command.
For a list of current options, try 'lun show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing LUN objects. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # return all lun objects on vs0
 my @lun_objs = $node_obj->fetch_lun_objs(-vserver => "vs0");
 if ($node_obj->error()){
     ...
 }

 # return all lun objects on vs0 in volume 'myvol'
 my @lun_objs = $node_obj->fetch_lun_objs(-vserver => "vs0",
                                          -volume => "myvol");
 if ($node_obj->error()){
     ...
 }

 # return all lun objects in the cluster
 my @lun_objs = $node_obj->fetch_lun_objs();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<fetch_lif_objs(%args)>

Fetch a list of LIF objects for the caller to use by discovering them on the system.

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the network interfce show
command.  For a list of current options, try 'network interface show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing LIF objects. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # return all lif objects on vs0
 my @lif_objs = $controller_cluster_obj->fetch_lif_objs(-vserver => "vs0");
 if ($controller_cluster_obj->error()){
     ...
 }

 # return all lif objects on homed to node1 with data-protocol fcp
 my @lun_objs = $controller_cluster_obj->fetch_lif_objs(-home-node     => "node1",
                                                        -data-protocol => "fcp");
 if ($controller_cluster_obj->error()){
     ...
 }

 # return all lif objects in the cluster
 my @lif_objs = $controller_cluster_obj->fetch_lif_objs();
 if ($controller_cluster_obj->error()){
     ...
 }

=back

=head2 B<fetch_volume_objs(%args)>

Fetch a list of VOLUME objects for the caller to use by discovering them on the system.

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the volume show command.
For a list of current options, try 'volume show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing VOLUME objects. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # return all volume objects on vs0
 my @volume_objs = $node_obj->fetch_volume_objs(-vserver => "vs0");
 if ($node_obj->error()){
     ...
 }

 # return all volume objects in the cluster
 my @volume_objs = $node_obj->fetch_volume_objs();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<fetch_vserver_objs(%args)>

Fetch a list of VSERVER objects for the caller to use by discovering them on the system.

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the vserver show command.
For a list of current options, try 'vserver show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing VSERVER objects. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # return all vserver objects in the cluster
 my @vserver_objs = $node_obj->fetch_vserver_objs();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<fetch_igroup_objs(%args)>

Fetch a list of IGROUP objects for the caller to use by discovering them on the system.

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the igroup show command.
For a list of current options, try 'igroup show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing IGROUP objects. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # return all igroup objects on vs0
 my @igroup_objs = $node_obj->fetch_igroup_objs(-vserver => "vs0");
 if ($node_obj->error()){
     ...
 }

 # return all igroup objects in the cluster
 my @igroup_objs = $node_obj->fetch_igroup_objs();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<fetch_qtree_objs(%args)>

Fetch a list of qtree objects for the caller to use by discovering them on the system.

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the qtree show command.
For a list of current options, try 'qtree show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing qtree objects. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 # return all qtree objects on vs0
 my @qtree_objs = $node_obj->fetch_qtree_objs(-vserver => "vs0");
 if ($node_obj->error()){
     ...
 }

 # return all qtree objects in the cluster
 my @qtree_objs = $node_obj->fetch_qtree_objs();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_list_info(%args)>

Gets a list of iscsi show records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
show command.  For a list of current options, try 'iscsi show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi show records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_isns_list_info(%args)>

Gets a list of iscsi isns show records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
isns show command.  For a list of current options, try 'iscsi isns show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi isns show records. If an error occurs, sets the error flag
and returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_isns_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_status_list_info(%args)>

Gets a list of iscsi status records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
status command.  For a list of current options, try 'vserver iscsi status ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi status records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_status_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_session_list_info(%args)>

Gets a list of iscsi session records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
session show command.  For a list of current options, try 'iscsi session show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi session records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_session_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_connection_list_info(%args)>

Gets a list of iscsi connection records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
connection show command.  For a list of current options, try 'iscsi connection
show ?'.

=item * I<initiator-name>

initiator-name used to be a filter in iscsi connection show.  It was removed.
However, this function has been updated so that you can still filter by
nodename.  It does this by first querying the sessions with the nodename, then
queries all connections associated with the sessions.

=item * I<session-type>

Like initiator-name, session-type can be applied as a filter.  It first calls
session_list_info to filter the sessions, then calls connection show to filter
the connections.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi connection records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_connection_list_info();
 if ($node_obj->error()){
     ...
 }

 my $result_ref = $node_obj->iscsi_connection_list_info('initiator-name' => "iqn.1986-03.com.sun:01:0003ba4154df.4cd2e60a");
 if ($node_obj->error()){
     ...
 }

 my $result_ref = $node_obj->iscsi_connection_list_info('session-type' => "Normal");
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_nodename_list_info(%args)>

Gets a list of iscsi nodename records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
nodename command.  For a list of current options, try 'vserver iscsi nodename ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi nodename records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_nodename_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_alias_list_info(%args)>

Gets a list of iscsi alias records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
alias command.  For a list of current options, try 'vserver iscsi alias ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi alias records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_alias_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_initiator_list_info(%args)>

Gets a list of iscsi initiator records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
initiator show command.  For a list of current options, try 'iscsi initiator
show ?'.

=item * I<session-type>

session-type can be applied as a filter.  It first calls session_list_info to
filter the sessions, then calls initiator show to filter the initiators.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi inititor records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_initiator_list_info();
 if ($node_obj->error()){
     ...
 }

 my $result_ref = $node_obj->iscsi_initiator_list_info('session-type' => "Normal");
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_tpgroup_list_info(%args)>

Gets a list of iscsi tpgroup records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
tpgroup show command.  For a list of current options, try 'iscsi tpgroup
show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi tpgroup records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_tpgroup_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_interface_list_info(%args)>

Gets a list of iscsi interface records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
interface show command.  For a list of current options, try 'iscsi interface
show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi interface records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_interface_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_interface_accesslist_list_info(%args)>

Gets a list of iscsi interface accesslist records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
interface acceslist show command.  For a list of current options, try 'iscsi
interface accesslist show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi interface accesslist records. If an error occurs, sets the
error flag and returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_interface_accesslist_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_portal_list_info(%args)>

Gets a list of iscsi portal records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
portal show command.  For a list of current options, try 'iscsi portal
show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi portal records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_portal_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<iscsi_initiator_auth_list_info(%args)>

Gets a list of iscsi security records on the cluster

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the iscsi
security show command.  For a list of current options, try 'iscsi security
show ?'.

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing iscsi security records. If an error occurs, sets the error flag and
returns an empty list.

=item Examples

 my $result_ref = $node_obj->iscsi_initiator_auth_list_info();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<license_add(%args)>

Adds a license for the specified services

=over 4

=item Inputs

=item * I<service>

If service  is a scalar, a license for this service will be added. Alternatively
service can be an array ref to an array of services. If so, element of the array
will be treated as a service and an attempt will be made to add license for each

=item * I<license-code>

If license-code is a scalar, a license for this license-code will be added. Alternatively
license-code can be an array ref to an array of license-codes. If so, element of the array
will be treated as a license-code and an attempt will be made to add license for each

=item Outputs

Returns $SanUtil::TRUE if the licenses are added successfully otherwise
$SanUtil::FALSE

=item Examples

 # ex. 1: Install one service. service will be a scalar in this case
 $node_obj->license_add(-service => "iscsi");
 if ($node_obj->error()){
     ...
 }

 # ex. 2: Install multiple services. service will be an arrayref in this case
 my @services = qw (nfs cifs iscsi fcp);
 $node_obj->license_add(-service => \@services);
 if ($node_obj->error()){
     ...
 }

 # ex. 3: Install one service. license will be a scalar in this case
 my $license = 'HVWKXWGLYOMIWB';
 $node_obj->license_add('license-code' => $license);
 if ($node_obj->error()){
     ...
 }

 # ex. 4: Install multiple services. license-code will be an arrayref in this case
 my @licenses = ('HVWKXWGLYOMIWB', 'FKSOYWGLYOMIWB');
 $node_obj->license_add('license-code' => \@licenses);
 if ($node_obj->error()){
     ...
 }

 # ex. 5: Install multiple services. service and license-code will be scalars in this case
 #        service and or license-code could have also been arrayrefs
 my $license = 'HVWKXWGLYOMIWB';
    my $service = 'fcp';
 $node_obj->license_add('license-code' => $license, service => $service);
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<license_exist(%args)>

Returns 1 if the given license exists, otherwise 0

=over 4

=item Inputs

=item * I<service>

Service name for which the existence of the license has to be checked

=item Outputs

Returns 1 if license for given service exists, otherwise 0

=item Examples

 my $iscsi_license_exists = $node_obj->license_exist(-service => "iscsi");
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<get_hostrec()>

Returns hostrec object for the filer

=over 4

=item Inputs

=item * I<None>

=item Outputs

Returns hostrec object

=item Examples

 my $host_obj = $filer_obj->get_hostrec();
 if ($filer_obj->error()) {
     ...
 }

=back

=head2 B<is_vsim()>

Returns 1 if the node is a vsim, otherwise 0

=over 4

=item Inputs

=item * I<None>

=item Outputs

Returns 1 if the node is a vsim, otherwise 0

=item Examples

 my $is_vsim = $node_obj->is_vsim();
 if ($node_obj->error()){
     ...
 }

=back

=head2 B<console_execute(%args)>

Execute command over a console connection.

=over 4

=item Inputs

=item * I<-cmd> (required)

Command to execute

=item * I<-shell_type> (optional)

Shell type for command to execute on
Default: bsd

=item * I<-timeout> (optional)

The timeout to use when executing the command
Default: 60 (seconds)

=item Outputs

Returns output from command executed, sets error otherwise.

=item Examples

 # execute bsd command over console connection
 my (@output) = $filer_obj->console_execute(-cmd => $cmd);
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<cluster_ha_show(%args)>

List out the cluster HA properties

=over 4

=item Inputs

=item * node

Optional: The node to display the cluster HA properties for.
Default: The current node.

=item Outputs

Returns an array with the propierties and values of the cluster HA.

=item Examples

 my ($output) = $filer_obj->cluster_ha_show();
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

 if ($output->{configured}     eq "false" and
     $output->{sfo_configured} eq "false") {
     ...
 }

=back


=head2 B<cluster_ha_modify(%args)>

Make modifications to the cluster HA properties of a specific node

=over 4

=item Inputs

=item * configured

Mandatory: Either True or False

=item Outputs

Returns $SanUtil::TRUE if successful, $SanUtil::FALSE and sets error if not.

=item Examples

 $ha_flag = $filer_obj->cluster_ha_modify(-configured => "true");
 if ($filer_obj->error) {
     return TCD::FAIL;
 }

=back

=head2 B<is_cluster_ha_configured()>

Determine whether the cluster HA has been configured

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE if cluster HA is enabled, otherwise $SanUtil::FALSE

=item Examples

 # determine whether sfo is enabled
 $ha_flag = $filer_obj->is_cluster_ha_configured();
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<cluster_ha_disable()>

Disables the cluster HA, if its not disabled

=over 4

=item Inputs

None

=item Outputs

Returns $SanUtil::TRUE if the cluster HA is disabled successfully otherwise
sets error and returns $SanUtil::FALSE

=item Examples

 # disable sfo
 $disabled_flag = $filer_obj->cluster_ha_disable();
 if ($filer_obj->error()) {
    ...
 }

=back

=head2 B<kcs_list_info()>

Get kernel service info from this node

=over 4

=item Inputs

None

=item Outputs

Returns an array or a reference to an array (depending on how it's called)
containing hash(s) with kernel service info for self or list of nodes passed. If
an error occurs, sets the error flag and returns an empty list.

=item Examples

  my $kcs_info = ($filer_obj->kcs_list_info())[0];
  if ($filer_obj->error()) {
      ...
  }

  my $node_list = "$FilerA,$FilerB";
  @kcs_records = $filer_obj->kcs_list_info(-node => $node_list);
  if ($node_obj ->error()) {
      ...
  }
=back

=head2 B<license_delete(%args)>

Deletes a specific license for a services.

=over 4

=item Inputs

=item * I<feature>

The name of the feature (service) that the license will be deleted.

=item Outputs

Returns $SanUtil::TRUE if the license was properly deleted, otherwise $SanUtil::FALSE

=item Examples

 # example
 $node_obj->license_delete(-feature => "iscsi");
 if ($node_obj->error()){
     ...
 }

=back


=head2 B<lun_serials_are_unique(%args)>

Verifies that all the existing LUN serial numbers are unique.

=over 4

=item Inputs

=item * I<none>

Note: You may optionally pass in any filters that are valid for the lun show
command.  For a list of current options, try 'lun show ?'.

=item * I<duplicates>

Optional: You could pass a REFERENCE to an array called (key)'duplicates' where
all the duplicated serial numbers will be stored.

=item Outputs

Returns $SanUtil::TRUE if all the LUN serial numbers are unique.
Returns $SanUtil::FALSE if any duplicated LUN serial number is found and if provided,
modifies the 'duplicates' array reference to point to an array holding the
duplicated serial numbers.

=item Examples
    my ($answer) = '';
    my (@duplicates) = ();


# Example 1 - Simple call, with no arguments:

    $answer = $Node_Obj->lun_serials_are_unique();
    if ($Node_Obj->error()) {
        ...;
    }

# Example 2 - Verifying the LUN serial numbers on a specific vserver and providing
              the array reference to get the duplicated serial nums:

    $answer = $Node_Obj->lun_serials_are_unique('vserver'=> 'utest_vserver',
                                                'duplicates'=> \@duplicates);
    if ($Node_Obj->error()) {
        ...
    }

# Example 3 - Verifying the LUN serial numbers on a specific vserver and volume,
              using the 'fields' argument. Also provided the array reference to
              get the duplicated serial nums:

    $answer = $Node_Obj->lun_serials_are_unique('fields'=> 'vserver,volume',
                                                'duplicates'=> \@duplicates);
    if ($Node_Obj->error()) {
        ...
    }

# Example on how to use the returned values:

    if ($answer){
        $Test->nlog("\t LUN Serials are Unique !");
    } else {
        $Test->nlog("\t Found LUN serials duplicates: " . Dmper(@duplicates) );
    }

=back

=cut
