# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary A Utility Module providing Clam Validation
## @author mhester@netapp.com
## @status private
## @pod here

package ClamValidation;

# Perl Packages
use List::MoreUtils qw( any firstidx );
use POSIX qw( ceil );
use Data::Dumper;

# NATE Packages
use Tharn qw( $Log param);

# San packages
use San::SanUtil;
use ParamsMgr;

# Clam Back Off Timer Constants
use constant {
    CMBO_CONST_P => 1000, # Clam HB Interval (ms)
    CMBO_CONST_T => 1,    # Clam HB control multilier?
    CMBO_CONST_t => 10,   # Clam HB control multiplier?
    CLAM_NODE_ID_BASE => 1000,
    OOQ_INTERVAL => 9,
    RDB_SITELIST_UPDATE_INTERVAL => 30,
    CLAM_ONLINE_BUFFER_AFTER_GIVEBACK => 150,
    CLAM_OFFLINE_BUFFER_AFTER_TAKEOVER => 30,
    CLAM_OFFLINE_BUFFER_AFTER_UNJOIN => 30
};

# Initialize Global Variables
my $OOQ_Interval = &Tharn::param("CLAM_OOQ_INTERVAL", "-default", OOQ_INTERVAL);

###############################################################################
#                               Public Methods
###############################################################################
###############################################################################
# Method:    validate_clam_is_healthy()
# Objective: Issue clam validation assuming all nodes healthy and active.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_clam_is_healthy {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my @nodes           = @{&ParamsMgr::get_arg(-arg      => "nodes",
                                                -ref_type => "ARRAY",
                                                -required => $SanUtil::TRUE)};

    my $original_master =   &ParamsMgr::get_arg(-arg      => "original_master",
                                                -required => $SanUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # Verify clam master
    unless (&ClamValidation::_verify_clam_master(
                                        -nodes           => \@nodes,
                                        -original_master => $original_master,
                                        -changed         => $SanUtil::FALSE)) {
        # Error messages should already be logged.
        return $SanUtil::FALSE;
    }

    # Loop through the nodes and verify their clam instances also checks the
    # node's own clam instance.
    foreach my $node (@nodes) {
        unless (&ClamValidation::_verify_clam_cluster_state(
                                            -nodes      => \@nodes,
                                            -clam_node  => $node,
                                            -in_quorum  => $SanUtil::TRUE,
                                            -in_cluster => $SanUtil::TRUE)) {
            # Error messages should already be logged.
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} # End validate_clam_is_healthy

###############################################################################
# Method:    validate_clam_after_takeover()
# Objective: Issue clam validation assuming a takeover was issued on
#            storage
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_clam_after_takeover {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my @nodes           = @{&ParamsMgr::get_arg(-arg      => "nodes",
                                                -ref_type => "ARRAY",
                                                -required => $SanUtil::TRUE)};

    my $fault_node      =   &ParamsMgr::get_arg(-arg      => "fault_node",
                                                -required => $SanUtil::TRUE);

    my $original_master =   &ParamsMgr::get_arg(-arg      => "original_master",
                                                -required => $SanUtil::TRUE);

    my $take_start_time =   &ParamsMgr::get_arg(-arg      => "take_start_time",
                                                -required => $SanUtil::FALSE,
                                                -default  => time);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # Locals
    $take_start_time = $take_start_time + CLAM_OFFLINE_BUFFER_AFTER_TAKEOVER;
    my $wait_time = $take_start_time - time;
    unless ($wait_time <= 0) {
        $Log->comment(
            "Sleeping for $wait_time seconds to allow clam to go offline "
            . "on faulted node after takeover."
        );
        sleep($wait_time);
    }

    # Did the original clam master change?
    my $master_change = $SanUtil::FALSE;
    if (&ClamValidation::_are_nodes_equal($fault_node, $original_master)) {
        $master_change = $SanUtil::TRUE;
    }

    # We want to remove the faulted node as it is out of service and cannot
    # be used for verification.
    @nodes = &ClamValidation::_remove_node_from_list(
        -remove_node => $fault_node,
        -nodes       => \@nodes
    );
    unless (scalar(@nodes)) {
        $Log->error(
            "There are no node_objs left in \"node_objs\" array "
            . "after attempting to remove \"fault_node\" $fault_node."
        );
        return $SanUtil::FALSE;
    }

    # Verify clam master
    unless (
        &ClamValidation::_verify_clam_master(
            -nodes           => \@nodes,
            -original_master => $original_master,
            -changed         => $master_change,
            -change_time     => $take_start_time)
    ) {
        # Error messages should already be logged.
        return $SanUtil::FALSE;
    }

    # Wait for OOQ_Interval Before Proceeding to check OOQ Events.
    $Log->comment(
            "Sleeping for $OOQ_Interval seconds to allow clam to go "
            . " OOQ on faulted node ($fault_node) after takeover."
    );
    sleep($OOQ_Interval);

    # Verify the clam state on the faulted node via other node objects.
    unless (
        &ClamValidation::_verify_clam_cluster_state(
            -nodes      => \@nodes,
            -clam_node  => $fault_node,
            -in_quorum  => $SanUtil::FALSE,
            -in_cluster => $SanUtil::TRUE)
    ) {
        # We have already logged the error, just return FALSE
        return $SanUtil::FALSE;
    }

    # Lets loop through the remaining "active" nodes and verify their clam
    # instances. This will check the node against its ownself as well.
    foreach my $node (@nodes) {
        unless (
            &ClamValidation::_verify_clam_cluster_state(
                -nodes      => \@nodes,
                -clam_node  => $node,
                -in_quorum  => $SanUtil::TRUE,
                -in_cluster => $SanUtil::TRUE)
        ) {
            # We have already logged the error, just return FALSE
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} # End validate_clam_after_takeover

###############################################################################
# Method:    validate_clam_after_giveback()
# Objective: Issue clam validation assuming a giveback was issued on
#            storage
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_clam_after_giveback {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my @nodes           = @{&ParamsMgr::get_arg(-arg      => "nodes",
                                                -ref_type => "ARRAY",
                                                -required => $SanUtil::TRUE)};

    my $original_master =   &ParamsMgr::get_arg(-arg      => "original_master",
                                                -required => $SanUtil::TRUE);

    my $give_start_time =   &ParamsMgr::get_arg(-arg      => "give_start_time",
                                                -required => $SanUtil::FALSE,
                                                -default  => time);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    my $wait_time = $give_start_time + CLAM_ONLINE_BUFFER_AFTER_GIVEBACK - time;
    unless ($wait_time <= 0) {
        $Log->comment(
            "Sleeping for $wait_time seconds to allow giveback node to come "
            . "online for validation after giveback is issued."
        );
        sleep($wait_time);
    }

    return &ClamValidation::validate_clam_is_healthy(
        -nodes           => \@nodes,
        -original_master => $original_master
    );
}

###############################################################################
# Method:    validate_clam_after_unjoin()
# Objective: Issue clam validation assuming a unjoin was issued on
#            storage
# Details:   See POD documentation at the end of this file.
###############################################################################
sub validate_clam_after_unjoin {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my @nodes             = @{&ParamsMgr::get_arg(-arg      => "nodes",
                                                  -ref_type => "ARRAY",
                                                  -required => $SanUtil::TRUE)};

    my $unjoin_node       =   &ParamsMgr::get_arg(-arg      => "unjoin_node",
                                                  -required => $SanUtil::TRUE);

    my $original_master   =   &ParamsMgr::get_arg(-arg      => "original_master",
                                                  -required => $SanUtil::TRUE);

    my $unjoin_start_time =   &ParamsMgr::get_arg(-arg      => "unjoin_start_time",
                                                  -required => $SanUtil::FALSE,
                                                  -default  => time);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # Locals
    $unjoin_start_time = $unjoin_start_time + CLAM_OFFLINE_BUFFER_AFTER_UNJOIN;
    my $wait_time = time - $unjoin_start_time;
    unless ($wait_time <= 0) {
        $Log->comment(
            "Sleeping for $wait_time seconds to allow clam to go offline "
            . "on faulted node after unjoin."
        );
        sleep($wait_time);
    }

    # Did the original clam master change?
    my $master_change = $SanUtil::FALSE;
    if (&ClamValidation::_are_nodes_equal($unjoin_node, $original_master)) {
        $master_change = $SanUtil::TRUE;
    }

    # We want to remove the faulted node as it is out of service and cannot
    # be used for verification.
    @nodes = &ClamValidation::_remove_node_from_list(-remove_node => $unjoin_node,
                                                     -nodes       => \@nodes);
    if(!@nodes) {
        $Log->error(
            "There are no node_objs left in \"node_objs\" array
             after attempting to remove \"unjoin_node\" $unjoin_node."
        );
        return $SanUtil::FALSE;
    }

    # Verify clam master
    unless (
        &ClamValidation::_verify_clam_master(
            -nodes           => \@nodes,
            -original_master => $original_master,
            -changed         => $master_change)
    ) {
        # Error messages should already be logged.
        return $SanUtil::FALSE;
    }

    # RDB may take some time to update depending on what is happening to/on
    # the cluster, wait for a few seconds.
    sleep(RDB_SITELIST_UPDATE_INTERVAL);

    # Verify the clam state on the faulted node via other node objects.
    unless (
        &ClamValidation::_verify_clam_cluster_state(
            -nodes      => \@nodes,
            -clam_node  => $unjoin_node,
            -in_cluster => $SanUtil::FALSE)
    ) {
        # We have already logged the error, just return FALSE
        return $SanUtil::FALSE;
    }

    # Lets loop through the remaining "active" nodes and verify their clam
    # instances. This will check the node against its ownself as well.
    foreach my $node (@nodes) {
        unless (
            &ClamValidation::_verify_clam_cluster_state(
                -nodes      => \@nodes,
                -clam_node  => $node,
                -in_quorum  => $SanUtil::TRUE,
                -in_cluster => $SanUtil::TRUE)
        ) {
            # We have already logged the error, just return FALSE
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} # End _validate_clam_after_unjoin

###############################################################################
# Method:     get_clam_master():
# Objective:  Detemermines which node in the list is clam_master.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_clam_master {
    my (%args) = @_;
    my $node_obj = undef;
    my $clam_master = "";
    my @node_names = ();
    my @kcs_records = ();
    my $want_object = $SanUtil::FALSE;

    &ParamsMgr::process_args(%args);

    # parse out the nodes param
    my @nodes = @{&ParamsMgr::get_arg(-arg      => "nodes",
                                      -ref_type => "ARRAY",
                                      -required => $SanUtil::TRUE)};

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return "";
    }

    foreach my $node (@nodes) {
        # if reference is Filer10M treat the element as
        # an object just store the name.
        # NOTE: This will always use the last node in the list to
        # execute the kcs show command on.
        if (ref($node) =~ /Filer10M/) {
            $want_object = $SanUtil::TRUE;
            $node_obj = $node;
            push(@node_names, $node_obj->get_name());
        } else {
            push(@node_names, $node);
        }
    }

    unless ($node_obj) {
        # initialize node_obj to last element of array to use same behavior
        # as foreach above.
        $Log->debug("Creating the Filer10M Object for " . $nodes[-1]);
        $node_obj = San::ControllerCluster::Filer10M->new(-filer => $nodes[-1]);
        if ($node_obj->error()){
            $Log->error("Creating the Filer10M Object for " . $nodes[-1]
                        . $node_obj->errmsg());
            return "";
        }
    }

    my $node_list = join(',', @node_names);
    @kcs_records = $node_obj->kcs_list_info(-node => $node_list);
    if ($node_obj ->error()) {
        $Log->error($node_obj->errmsg());
        return "";
    }

    foreach my $kcs_info (@kcs_records) {
        if (defined($kcs_info->{fsm})) {
            my $node_name = $kcs_info->{node} if defined($kcs_info->{node});
            #-------------------------------------------
            # Clam master will report "state=CLAM_CORE_ST_MASTER_..."
            # Clam non-master will report "state=CLAM_CORE_ST_NONMASTER_..."
            #-------------------------------------------
            if ($kcs_info->{fsm} =~ /CLAM_CORE_ST_MASTER/) {
                if ($clam_master) {
                    $Log->error(
                        "-$node_name and " . $kcs_info->{fsm}
                        . " both reported as clam master."
                    );
                    return "";
                }
                $Log->log("-$node_name IS clam master.");
                $clam_master = $node_name;
            } elsif ($kcs_info->{fsm} =~ /CLAM_CORE_ST_NONMASTER/) {
                $Log->log("-$node_name is NOT clam master.");
            } else {
                $Log->warn(
                    "-$node_name is not one of the defined states: "
                    . "CLAM_CORE_ST_MASTER or CLAM_CORE_ST_NONMASTER."
                    . "Instead it shows state: " . $kcs_info->{fsm}
                );
            }
        } else {
            $Log->error("Could not parse kcs records from kcs_list_info()");
            return "";
        }
    }

    if ($want_object) {
        my $cm_idx = firstidx { $_->get_name() eq $clam_master } @nodes;
        if ($cm_idx == -1) {
            $Log->warn("None of the nodes in the cluster are clam master.");
            return "";
        }
        $clam_master = $nodes[$cm_idx];
    }

    return $clam_master;
} # end get_clam_master

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:     _is_clam_master():
# Objective:  Verifies whether or not the node is the clam
#             master or not.
#
# Required Params:
#    * node - Name of node to verify from   OR
#             SAN::ControllerCluster::Filer10M Object for node to verify.
#
# Returns:
#    * TRUE if the node is Clam Master
#    * FALSE otherwise
###############################################################################
sub _is_clam_master {
    my (%args) = @_;
    my ($node_obj, $name, $kcs_info) = '';

    &ParamsMgr::process_args(%args);

    # parse out the node param
    my $node = &ParamsMgr::get_arg(-arg      => "node",
                                   -required => $SanUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # initialize local params
    # if we see a Filer10M assume they passed the the obj, else create one
    if (ref($node) =~ /Filer10M/) {
        $node_obj = $node;
        $name = $node_obj->get_name();
    } else {
        $Log->debug("Creating the Filer10M Object for $node ");
        $node_obj = San::ControllerCluster::Filer10M->new(-filer => $node);
        if ($node_obj->error()){
            $Log->error(
                "Creating the Filer10M Object for $node "
                . $node_obj->errmsg()
            );
            return $SanUtil::FALSE;
        }
    }

    $kcs_info = ($node_obj->kcs_list_info())[0];
    if ($node_obj ->error()) {
        $Log->error($node_obj->errmsg());
        return $SanUtil::FALSE;
    }
    #-------------------------------------------
    # Clam master will report "state=CLAM_CORE_ST_MASTER_..."
    # Clam non-master will report "state=CLAM_CORE_ST_NONMASTER_..."
    #-------------------------------------------
    if ($kcs_info->{fsm} =~ /CLAM_CORE_ST_MASTER/) {
        $Log->log("-$name IS clam master.");
        return $SanUtil::TRUE;
    } elsif ($kcs_info->{fsm} =~ /CLAM_CORE_ST_NONMASTER/) {
        $Log->log("-$name IS NOT clam master.");
        return $SanUtil::FALSE;
    } else {
        $Log->warn(
            "-$name is not one of the defined states: "
            . "CLAM_CORE_ST_MASTER or CLAM_CORE_ST_NONMASTER."
            . "Instead it shows state: " . $kcs_info->{fsm}
        );
        return $SanUtil::FALSE;
    }
} # end _is_clam_master

###############################################################################
# Method:     _verify_clam_master():
# Objective:  Verifies whether or not the node is the clam
#             master or not.
#
# Required Params:
#    * node_objs       - List of Names of nodes to verify from   OR
#                        SAN::ControllerCluster::Filer10M Object for node
#                        to verify.
#    * original_master - Original master before event. Name of node  OR
#                        SAN::ControllerCluster::Filer10M Object for node.
#    * changed         - true if the original master should have changed, false
#                        otherwise
#
# Returns:
#    * TRUE if the node is Clam Master
#    * FALSE otherwise
###############################################################################
sub _verify_clam_master {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    # parse out the node param
    my @nodes           = @{&ParamsMgr::get_arg(-arg      => "nodes",
                                                -ref_type => "ARRAY",
                                                -required => $SanUtil::TRUE)};

    my $original_master =   &ParamsMgr::get_arg(-arg      => "original_master",
                                                -required => $SanUtil::TRUE);

    my $changed         =   &ParamsMgr::get_arg(-arg      => "changed",
                                                -required => $SanUtil::FALSE,
                                                -default  => $SanUtil::FALSE);

    my $change_time     =   &ParamsMgr::get_arg(-arg      => "change_time",
                                                -required => $SanUtil::FALSE,
                                                -default  => time);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # Locals
    my $clam_master = "";
    my $orig_master_name = (ref($original_master) =~ /Filer10M/) ? $original_master->get_name() : $original_master;

    # Verify that the original clam master is still the master, unless master
    # was changed.
    if ($changed) {
        # Need to determine if there is a new clam master
        $Log->comment(
            "The original master should have changed, "
            . "determining if a new one has been selected."
        );

        my $clam_master = undef;
        my $cmbo_timer = 0;
        my $cmbo_wait = 0;
        my $cmbo_idx = 0;
        my @backoff_times = &ClamValidation::_get_clam_master_backoff_times(
            -nodes => \@nodes
        );

        # Try until a clam master is returned or all backoff timers expire
        while ($clam_master == undef) {
            $cmbo_timer = ceil(time - $change_time);
            if ($cmbo_timer > $backoff_times[$cmbo_idx]) {
                $Log->warn(
                    "clam master back off interval "
                    . $backoff_times[$cmbo_idx] . " elapsed"
                    . " and there is still no clam master. Current time since"
                    . " change event is: $cmbo_timer seconds."
                );
                $cmbo_idx++;
                unless ($cmbo_idx < scalar(@nodes)) {
                    $Log->error(
                        "No clam master was selected after waiting "
                        . "for all of the CMBO times:"
                        . join(',', @backoff_times)
                    );
                    return $SanUtil::FALSE;
                }
            }
            $cmbo_wait = $backoff_times[$cmbo_idx] - $cmbo_timer;
            unless ($cmbo_wait <= 0) {
                $Log->comment(
                    "Sleeping for $cmbo_wait seconds to allow back off timer to"
                    . " expire before attempting to get master new clam master."
                );
                sleep($cmbo_wait);
            }
            $clam_master = &ClamValidation::get_clam_master(-nodes => \@nodes);
        }
        my $cmbo_reselect_timer = time - $change_time;
        $Log->comment(
            "Clam master was selected within $cmbo_reselect_timer seconds."
        );

        # Make sure the clam master actually changed.
        $clam_master = (ref($clam_master) =~ /Filer10M/) ? $clam_master->get_name() : $clam_master;
        if ($orig_master_name eq $clam_master){
            $Log->error(
                "The original clam master $orig_master_name "
                . "should have changed, but the new master "
                . "$clam_master is the same."
            );
            return $SanUtil::FALSE;
        }
        $Log->comment("Determined $clam_master is the new clam master");

    } else {
        $Log->comment("Verify the original clam master is STILL clam master.");
        if (&ClamValidation::_is_clam_master(-node => $original_master)) {
            $Log->comment("Verified that $orig_master_name is clam master.");
        } else {
            $Log->error("Failed to verify $orig_master_name is clam master.");
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
} # End _verify_clam_master

###############################################################################
# _verify_clam_cluster_state():
#                       Verifies that clam knows the node's status in the
#                       cluster.  It will verify if the node is HA or not,
#                       in quorum or not, and if the node holds epsilon or not.
#                       It will return true if all the the input parameters
#                       match the clam dump status, false if any do not match
#
# Required Params:
#    * nodes - ArrayRef of node names or San::ControllerCluster:Filer10M objects
#    * clam_node - node we are checking the status of
#                 (can be different than the node in the node_object)
#    * in_cluster - true, false
#    * in_quorum - true, false
#
# Returns:
#    * TRUE if the verification was successful
#    * FALSE otherwise
###############################################################################
sub _verify_clam_cluster_state {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my @nodes      = @{&ParamsMgr::get_arg(-arg      => "nodes",
                                           -ref_type => "ARRAY",
                                           -required => $SanUtil::TRUE)};

    my $clam_node  =  &ParamsMgr::get_arg(-arg      => "clam_node",
                                          -required => $SanUtil::TRUE);

    my $in_cluster =  &ParamsMgr::get_arg(-arg      => "in_cluster",
                                          -required => $SanUtil::FALSE,
                                          -default  => undef);

    my $in_quorum  =  &ParamsMgr::get_arg(-arg      => "in_quorum",
                                          -required => $SanUtil::FALSE,
                                          -default  => undef);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return $SanUtil::FALSE;
    }

    # Locals
    my $fail_cnt = 0;
    my ($kcs_info, $quorum_idx) = "";
    my (@node_names, @node_list, @kcs_records, @quorum_list) = ();
    my $nodename = (ref($clam_node) =~ /Filer10M/) ? $clam_node->get_name() : $clam_node;
    $Log->comment("Setting clam node to $nodename.");

    # Build array of node_names for validation.
    foreach my $node (@nodes) {
        # if reference is Filer10M treat the element as
        # an object just store the name.
        # NOTE: This will always use the last node in the list to
        # execute the kcs show command on.
        if (ref($node) =~ /Filer10M/) {
            $node_obj = $node;
            push(@node_names, $node_obj->get_name());
        } else {
            push(@node_names, $node);
        }
    }

    # If there isn't a node object, pick one node to run the validation from...
    unless ($node_obj) {
        # initialize node_obj to last element of array to use same behavior
        # as foreach above.
        $Log->debug("Creating the Filer10M Object for " . $nodes[-1]);
        $node_obj = San::ControllerCluster::Filer10M->new(-filer => $nodes[-1]);
        if ($node_obj->error()){
            $Log->error(
                "Creating the Filer10M Object for " . $nodes[-1]
                . $node_obj->errmsg()
            );
            return $SanUtil::FALSE;
        }
    }

    # Gather the clam information from the node object for all nodes in the list
    my $node_name_list = join(',', @node_names);
    @kcs_records = $node_obj->kcs_list_info(-node => $node_name_list);
    if ($node_obj->error()) {
        $Log->error($node_obj->error());
        return $SanUtil::FALSE;
    }

    # Loop through each record which is a representation of clam on each node
    # in $node_name_list above.
    foreach my $kcs_info (@kcs_records) {
        $Log->comment(
            "Verification completed via node: " . $kcs_info->{node}
        );

        @node_list = split(',', $kcs_info->{nodelist});
        @quorum_list = split(',', $kcs_info->{quorum});

        if ($in_cluster) {
            if (any { $_ eq $nodename } @node_list) {
                $Log->comment("Pass: Clam sees $nodename in cluster.");
            } else {
                $Log->error(
                    "Clam does not recognize $nodename is in the cluster: "
                    . "Nodelist: " . $kcs_info->{nodelist}
                );
                $fail_cnt++;
            }
        } elsif (defined($in_cluster) && !$in_cluster) {
            if (any { $_ eq $nodename } @node_list) {
                $Log->error("Clam recognizes $nodename as in the cluster");
                $fail_cnt++;
            } else {
                $Log->comment(
                    "Pass: Clam does not recognize $nodename as in the cluster."
                );
            }
        }
        if ($in_quorum) {
            $quorum_idx = firstidx { $_ eq $nodename } @node_list;
            if ($quorum_idx < 0) {
                $Log->error(
                    "Did not find $nodename in nodelist: "
                    . $kcs_info->{nodelist}
                );
                return $SanUtil::FALSE;
            }
            if ($quorum_list[$quorum_idx] =~ /INQ"$/) {
                $Log->comment("Pass: $nodename IS in clam quorum.");
            } else {
                $Log->error(
                    "$nodename IS NOT in clam quorum: "
                    . "Nodelist: " . $kcs_info->{nodelist}
                    . "Quorum: " . $kcs_info->{quorum}
                );
                $fail_cnt++;
            }
        } elsif (defined($in_quorum) && !$in_quorum) {
            $quorum_idx = firstidx { $_ eq $nodename } @node_list;
            if ($quorum_idx < 0){
                $Log->error(
                    "Did not find $nodename in nodelist"
                    . $kcs_info->{nodelist}
                );
                return $SanUtil::FALSE;
            }
            if ($quorum_list[$quorum_idx] =~ /OOQ"$/) {
                $Log->comment("Pass: $nodename IS NOT clam quorum.");
            } else {
                $Log->error(
                    "$nodename IS in clam quorum. \n"
                    . "Nodelist: " . $kcs_info->{nodelist} . "\n"
                    . " Quorum: " . $kcs_info->{quorum}
                );
                $fail_cnt++;
            }
        }
    }

    return $fail_cnt ? $SanUtil::FALSE : $SanUtil::TRUE;
} # End _verify_clam_cluster_state

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# _remove_node_from_list():
#                       Removes a single node from a list of provided nodes
#
# Required Params:
#  Array containing a single hash ref with these keys:
#    * nodes - ArrayRef of node names or San::ControllerCluster:Filer10M objects
#    * remove_node - Single node name or San::ControllerCluster::Filer10M object
#                    that is to be removed from the list
#
# Returns:
#    * Array or ArrayRef of new list minus remove node or Empty Array/ArrayRef
###############################################################################
sub _remove_node_from_list {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my @nodes       = @{&ParamsMgr::get_arg(-arg      => "nodes",
                                            -ref_type => "ARRAY",
                                            -required => $SanUtil::TRUE)};

    my $rm_node     =   &ParamsMgr::get_arg(-arg      => "remove_node",
                                            -required => $SanUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return wantarray() ? () : [];
    }

    # Locals
    my @new_nodes = ();

    my $remove_name = (ref($rm_node) =~ /Filer10M/) ? $rm_node->get_name() : $rm_node;

    $Log->comment("Removing $remove_name from list of nodes.");
    foreach my $node (@nodes) {
        if (ref($node) =~ /Filer10M/) {
            next if ($node->get_name() eq $remove_name);
            $Log->debug(
                "Keeping " . $node->get_name() . " in the list of node_objs"
            );
        } else {
            next if ($node eq $remove_name);
            $Log->debug(
                "Keeping " . $node . " in the list of node_objs"
            );
        }
        push(@new_nodes, $node);
    }

    return wantarray() ? @new_nodes : \@new_nodes;
} # End _remove_node_from_list

###############################################################################
# _are_nodes_equal(): Compare two nodes for names.
#
# Required Params:
#  Array of:
#    * node1 - Single node name or San::ControllerCluster::Filer10M object
#                    that is to be compared to the second
#    * node2 - Single node name or San::ControllerCluster::Filer10M object
#                    that is to be compared to the first
#
# Returns:
#    * 0 or 1 (not equal or equal).
###############################################################################
sub _are_nodes_equal {
    my ($node1, $node2) = @_;

    my $nodename1 = (ref($node1) =~ /Filer10M/) ? $node1->get_name() : $node1;
    my $nodename2 = (ref($node2) =~ /Filer10M/) ? $node2->get_name() : $node2;

    return ($nodename1 eq $nodename2);
} # End _are_nodes_equal

###############################################################################
# _calculate_clam_master_backoff(): Calculate the backoff time based on node_id.
#
# Required Params:
#  Array of:
#    * node_id - The Site ID of the node obtained from the following cmd:
#                "debug smdb table rdbsitelist show"
#
# Returns:
#    * Integer value for number of seconds or -1 if there is an error.
###############################################################################
sub _calculate_clam_master_backoff {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my $node_id = &ParamsMgr::get_arg(-arg      => "node_id",
                                      -required => $SanUtil::TRUE);

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return -1;
    }

    # CMBO interval = (1+T)P + tPi
    # T and t are constants,
    # P is the basic CLAM transaction interval, and
    # i is the ordinal number of the node in the cluster sitelist
    $Log->comment("Node site id is: $node_id");
    my $i = int($node_id) - CLAM_NODE_ID_BASE;
    $Log->comment("Node ordinal number is: $i");
    my $cmbo = (1 + CMBO_CONST_T) * CMBO_CONST_P + (CMBO_CONST_t * CMBO_CONST_P * $i);

    return $cmbo/1000;
} # End _calculate_clam_master_backoff
###############################################################################
# _get_clam_master_backoff_times(): Get the back off time for clam master
#                                   reselection from each of the nodes in list
#
# Required Params:
#  Array of hash ref with key:
#    * node - Array reference containing list of nodes to get CMBO interval for.
#
# Returns:
#    * ArrayRef or Array based on caller conaining a sorted list of backoff
#      times from low to high.
###############################################################################
sub _get_clam_master_backoff_times {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);

    my @nodes   = @{&ParamsMgr::get_arg(-arg      => "nodes",
                                        -ref_type => "ARRAY",
                                        -required => $SanUtil::TRUE)};

    # check for error in parsing the parameters
    if (&ParamsMgr::error()) {
        $Log->error(&ParamsMgr::get_error());
        return wantarray() ? () : [];
    }

    #Locals
    my ($node_obj) = "";
    my @cmbo_intervals = ();

    # Build array of node_names for validation.
    foreach my $node (@nodes) {
        # if reference is Filer10M treat the element as
        # an object just store the name.
        # NOTE: This will always use the last node in the list to
        # execute the kcs show command on.
        if (ref($node) =~ /Filer10M/) {
            $node_obj = $node;
            push(@node_names, $node_obj->get_name());
        } else {
            push(@node_names, $node);
        }
    }

    # If there isn't a node object, pick one node to run the validation from...
    unless ($node_obj) {
        # initialize node_obj to last element of array to use same behavior
        # as foreach above.
        $Log->debug("Creating the Filer10M Object for " . $nodes[-1]);
        $node_obj = San::ControllerCluster::Filer10M->new(-filer => $nodes[-1]);
        if ($node_obj->error()){
            $Log->error(
                "Creating the Filer10M Object for " . $nodes[-1]
                . $node_obj->errmsg()
            );
            return wantarray() ? () : [];
        }
    }

    my @records = $node_obj->execute(
        -cmd =>"debug smdb table rdbsitelist show"
    );
    if ($node_obj->error()) {
        $Log->error($node_obj->errmsg());
        return wantarray() ? () : [];
    }

    foreach my $record (@records) {
        if (any { $_ eq $record->{node} } @node_names) {
            my $cmbo = _calculate_clam_master_backoff(
                -node_id => $record->{sid}
            );
            push(@cmbo_intervals, $cmbo);
        }
    }

    @cmbo_intervals = sort { $a <=> $b } @cmbo_intervals;

    return wantarray() ? @cmbo_intervals : \@cmbo_intervals;
}

1;

=pod

=head1 Package Name

San::FIT::ClamValidation.pm

=head1 Author

Michael Hester

=head1 Description

A validation module designed to provide helper
methods to perform cluster kernel-service state checking
during various node state change scenarios (ie takover/giveback,join/unjoin)


=head1 B<Public Methods>

=head2 B<get_clam_master()>

Returns the current clam master.  It will give node name or
the San::ControllerCluster::Filer10M object based on
what the caller passes in.

=over 4

=item Inputs

=item *I<-nodes>

Required: A reference to an array of node names or
          San::ControllerCluster::Filer10M objects.  If making repeated calls
          it is recommended to use objects, as this function will build an
          object each time it is needed for validation.

=item Outputs

The current clam master.  It will give node name or
the San::ControllerCluster::Filer10M object based on
what the caller passes in. If a master cannot be determined it will return an
empty string.  An error will appear in the logs.

=item Examples

    # This example will return a clam master object as a reference to
    #    an array of node objects is passed in.
    $clam_master_obj = &ClamValidation::get_clam_master(-nodes => \@node_objs);
    unless ($clam_master_obj) {
        $Test->nlog("Error: determining a clam master");
        return $SanUtil::FAIL;
    }
    $clam_master = $clam_master_obj->get_name();
    $Test->nlog("Current clam master is: $clam_master.");

=back

=head2 B<validate_clam_is_healthy()>

Performs clam validation for all nodes resources in the list
provided by the caller.  It will asssume that all nodes are in the same
cluster and are all in clam quorum.  It will validate that the original
clam master is still clam master.

=over 4

=item Inputs

=item *I<-nodes>

Required: A reference to an array of node names or
          San::ControllerCluster::Filer10M objects.  If making repeated calls
          it is recommended to use objects, as this function will build an
          object each time it is needed for validation.

=item *I<-original_master>

Required: The name or San::ControllerCluster::Filer10M object that should
          be validated as clam_master.

=item Outputs

$SanUtil::TRUE if successful, $SanUtil::FALSE if not.  Errors, are recorded
in the logs directory.

=item Examples

    # This example will validate that $FilerA and $FilerB are in clam quorum
    #   and in the cluster.  The validation is performed on from both nodes.
    # This will also validate that $FilerA is still the clam master.
    if (&ClamValidation::validate_clam_is_healthy(
                                -nodes           => [$FilerA,$FilerB],
                                -original_master => $FilerA) {
        $Test->nlog("Verified that clam is aware of $FilerA being in the "
                    . "cluster and in clam quorum.");
    } else {
        $Test->nlog("Failed verification of clam cluster state on $FilerA.");
        return $TCD::FAIL;
    }

=back

=head2 B<validate_clam_after_giveback()>

Performs clam validation for all nodes resources in the list
provided by the caller.  It will assume that all provided nodes are in the same
cluster and are all in clam quorum.  It will validate that the original
clam master is still clam master.  The function will wait for a period of time
defined in the module for the node being givenback to come online.

=over 4

=item Inputs

=item *I<-nodes>

Required: A reference to an array of node names or
          San::ControllerCluster::Filer10M objects.  If making repeated calls
          it is recommended to use objects, as this function will build an
          object each time it is needed for validation.

=item *I<-original_master>

Required: The name or San::ControllerCluster::Filer10M object that should
          be validated as clam_master.

=item *I<-give_start_time>

Optional: The exact time the giveback command completes on the cli.

Default: Perl built-in function 'time' that this function is called.

=item Outputs

$SanUtil::TRUE if successful, $SanUtil::FALSE if not.  Errors, are recorded
in the logs directory.

=item Examples

    # This example will validate that $FilerA and $FilerB are in clam quorum
    #   and in the cluster.  The validation is performed on from both nodes.
    # This will also validate that $FilerA is still the clam master.
    if (&ClamValidation::validate_clam_after_giveback(
                                -nodes           => [$FilerA,$FilerB],
                                -original_master => $FilerA,
                                -give_start_time => time) {
        $Test->nlog("Verified that clam is aware of $FilerA being in the "
                    . "cluster and in clam quorum.");
    } else {
        $Test->nlog("Failed verification of clam cluster state on $FilerA.");
        return $TCD::FAIL;
    }

=back

=head2 B<validate_clam_after_takeover()>

Performs clam validation for all nodes resources in the list
provided by the caller.  It will asssume that all nodes are in the same
cluster and are all in clam quorum.  It will validate that the original
clam master is still clam master, unless the fault_node is the original clam
master.  This will tell it to verify that a new clam_master has been selected.

=over 4

=item Inputs

=item *I<-nodes>

Required: A reference to an array of node names or
          San::ControllerCluster::Filer10M objects.  If making repeated calls
          it is recommended to use objects, as this function will build an
          object each time it is needed for validation.

=item *I<-fault_node>

Required: The name or San::ControllerCluster::Filer10M object that should
          has been takenover, or is in a faulted state.

=item *I<-original_master>

Required: The name or San::ControllerCluster::Filer10M object that should
          be validated as clam_master.

=item Outputs

$SanUtil::TRUE if successful, $SanUtil::FALSE if not.  Errors, are recorded
in the logs directory.

=item Examples

    # This example will validate that $FilerB is in clam quorum
    #   and in the cluster and that $FilerA is out of quorum and in the cluster
    # The validation is performed on from only $FilerB as $FilerA should be
    # takenover, and inactive.
    # This will also validate that a new clam master is selected.
    if (&ClamValidation::validate_clam_after_giveback(
                                -nodes           => [$FilerA,$FilerB],
                                -fault_node      => $FilerA,
                                -original_master => $FilerA) {
        $Test->nlog("Verified that clam is aware of $FilerA being in the "
                    . "cluster and in clam quorum.");
    } else {
        $Test->nlog("Failed verification of clam cluster state on $FilerA.");
        return $TCD::FAIL;
    }

=back

=head2 B<validate_clam_after_unjoin()>

Performs clam validation for all nodes resources in the list
provided by the caller.  It will asssume that all nodes are in the same
cluster and are all in clam quorum except for the unjoin node which it will
evaluate as not in the cluster.  It will validate that the original
clam master is still clam master, unless the unjoin_node is the original clam
master.  This will tell it to verify that a new clam_master has been selected.

=over 4

=item Inputs

=item *I<-nodes>

Required: A reference to an array of node names or
          San::ControllerCluster::Filer10M objects.  If making repeated calls
          it is recommended to use objects, as this function will build an
          object each time it is needed for validation.

=item *I<-unjoin_node>

Required: The name or San::ControllerCluster::Filer10M object that
          has been unjoined.

=item *I<-original_master>

Required: The name or San::ControllerCluster::Filer10M object that should
          be validated as clam_master.

=item Outputs

$SanUtil::TRUE if successful, $SanUtil::FALSE if not.  Errors, are recorded
in the logs directory.

=item Examples

    # This example will validate that $FilerB is in clam quorum
    #   and in the cluster and that $FilerA is out of the cluster
    # The validation is performed on from only $FilerB as $FilerA should be
    # unjoined, and inactive in the cluster.
    # This will also validate that a new clam master is selected if unjoin node
    # Is equal to fault_node.
    if (&ClamValidation::validate_clam_after_giveback(
                                -nodes           => [$FilerA,$FilerB],
                                -fault_node      => $FilerA,
                                -original_master => $FilerA) {
        $Test->nlog("Verified that clam is aware of $FilerA being in the "
                    . "cluster and in clam quorum.");
    } else {
        $Test->nlog("Failed verification of clam cluster state on $FilerA.");
        return $TCD::FAIL;
    }

=back

=cut
