# Copyright (c) 2010 - 2011 NetApp Inc.
# All rights reserved.
#
## @summary Definition of all I/O presets used by SAN FIT tests
## @author dl-san-fit@netapp.com
## @status private

use strict;

# Package Name
package IOPresets;

# Module imports
use San::SanUtil;
use Tharn;

# Constants
use constant {
    DEFAULT_IO_BS => "4k",
    DEFAULT_IO_START_BYTE => "1m",
    DEFAULT_IO_LIMIT      => "50m",
    DEFAULT_IO_MIN_MICRODELAY => "10000",
    DEFAULT_IO_MAX_MICRODELAY => "80000",
    DEFAULT_IO_READ_MIN_MICRODELAY => "10000",
    DEFAULT_IO_READ_MAX_MICRODELAY => "80000",
    DEFAULT_IO_WRITE_MIN_MICRODELAY => "10000",
    DEFAULT_IO_WRITE_MAX_MICRODELAY => "80000",
    DEFAULT_ONCERR => "abort",
    DEFAULT_TOOL_VERSION => "current",
    DEFAULT_DEVICES_PER_TOOL => 1,
    DEFAULT_IO_DT_NOPROGT => 15,
    DEFAULT_IO_DT_NOPROGTT => 130,
    DEFAULT_IO_DT_ALARM => 3,
    DEFAULT_IO_DT_SLICES => 3,
    DEFAULT_IO_DT_PATTERN => "iot",
    DEFAULT_IO_DT_DSIZE => "4k",
    DEFAULT_IO_DT_PREFIX => '%d@%h',
    DEFAULT_IO_DT_ENABLE => "microdelay,timestamp,syslog,nate",
    DEFAULT_IO_DT_IOTUNE => "dt.iotune",
    DEFAULT_IO_DT_KEEPALIVET => 60,
};

# Globals
use vars qw (
    @Presets
);

################################################################################
# Description: Initialize IO presets by appending IO limits and start bytes.
# Params: A hash, -io_limits and -start_byte
# Returns: A Boolean
################################################################################
sub init_io_presets {
    my (%args) = @_;

    # Locals
    my $io_limit   = DEFAULT_IO_LIMIT;
    my $start_byte = DEFAULT_IO_START_BYTE;
    my $min_microdelay = DEFAULT_IO_MIN_MICRODELAY;
    my $max_microdelay = DEFAULT_IO_MAX_MICRODELAY;
    my $oncerr = DEFAULT_ONCERR;
    my $tool_version = DEFAULT_TOOL_VERSION;
    my $devices_per_tool = DEFAULT_DEVICES_PER_TOOL;

    if (defined $args{io_limit}) {
        $io_limit = $args{io_limit};
    }

    if (defined $args{start_byte}) {
        $start_byte = $args{start_byte};
    }

    if (defined $args{min_microdelay}) {
        $min_microdelay = $args{min_microdelay};
    }

    if (defined $args{max_microdelay}) {
        $max_microdelay = $args{max_microdelay};
    }

    if (defined $args{oncerr}) {
        $oncerr = $args{oncerr};
    }

    if (defined $args{tool_version} && $args{tool_version} !~ "") {
        $tool_version = $args{tool_version};
    }

    if (defined $args{devices_per_tool} && $args{devices_per_tool} !~ "") {
        $devices_per_tool = $args{devices_per_tool};
    }

    #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
    #                             SIO LOADS
    #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
    my %default_sio_load = (
        id          => "sio_default_load",
        description => "SAN FIT's default I/O load",
        io_tool     => "sio",
        iotype      => "random",
        read_perc   => 50,
        random_perc => 100,
        start_byte  => $start_byte,
        limit       => $io_limit,
        bs          => "4k",
        pass_thru   => "-rep_latency -60"
    );
    push(@Presets, \%default_sio_load);

    # DSS type of workload
    my %sio_large_sequential_read = (
        id          => "Sio_Large_Sequential_Read",
        description => "Large sequential reads simulating a DSS environment",
        io_tool     => "sio",
        iotype      => "sequential",
        read_perc   => 100,
        random_perc => 0,
        start_byte  => $start_byte,
        limit       => $io_limit,
        bs          => "128k",
        pass_thru   => "-rep_latency -60"
    );
    push(@Presets, \%sio_large_sequential_read);

    # Transactional type of workload
    my %sio_small_random_read = (
        id          => "Sio_Small_Random_Read",
        description => "Small random reads simulating a typical database workload",
        io_tool     => "sio",
        iotype      => "random",
        read_perc   => 100,
        random_perc => 100,
        threads     => 1,
        start_byte  => $start_byte,
        limit       => $io_limit,
        bs          => "4k",
        pass_thru   => "-rep_latency -60"
    );
    push(@Presets, \%sio_small_random_read);

    # Small Random I/Os with 70/30 read/Write Ratio. Typical OLTP environment
    my %sio_small_random_70_read = (
        id          => "Sio_Small_Random_70_Read",
        description => "Small random reads and writes simulating a typical OLTP",
        io_tool     => "sio",
        iotype      => "random",
        read_perc   => 70,
        random_perc => 100,
        threads     => 1,
        start_byte  => $start_byte,
        limit       => $io_limit,
        bs          => "4k",
        pass_thru   => "-rep_latency -60"
    );
    push(@Presets, \%sio_small_random_70_read);

    # Large sequential I/Os with 90/10 read/Write Ratio. Stress IO Bandwidth
    my %sio_large_sequential_90_read = (
        id          => "Sio_Large_Sequential_90_Read",
        description => "Large sequential reads simulating lots of IO Bandwidth",
        io_tool     => "sio",
        iotype      => "sequential",
        read_perc   => 90,
        random_perc => 0,
        threads     => 1,
        start_byte  => $start_byte,
        limit       => $io_limit,
        bs          => "128k",
        pass_thru   => "-rep_latency -60"
    );
    push(@Presets, \%sio_large_sequential_90_read);

    #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
    #                             DT Loads
    #+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#

    # Random Read/Write using small block sizes
    my %default_dt_load = (
        id               => "dt_default_load",
        description      => "SAN FIT's default I/O load using DT",
        io_tool          => "dt",
        write            => 1,
        noprogt          => 15,
        noprogtt         => 130,
        alarm            => 3,
        iotype           => "random",
        pattern          => "iot",
        enable           => "raw",
        #limit        => $io_limit,
        bs               => "4k",
        tool_version     => $tool_version,
        devices_per_tool => $devices_per_tool,
        pass_thru        => "dsize=4k enable=microdelay rdelay=$min_microdelay"
                            . " wdelay=$min_microdelay oncerr=$oncerr"
                            . " prefix=\%d@%h\ enable=timestamp"
    );
    push(@Presets, \%default_dt_load);

    # DSS type of workload
    my %dt_large_sequential_read = (
        id               => "Dt_Large_Sequential_Read",
        description      => "Large sequential reads simulating a DSS environment",
        io_tool          => "dt",
        write            => 1,
        noprogt          => 15,
        noprogtt         => 130,
        alarm            => 3,
        pattern          => "iot",
        #limit        => $io_limit,
        bs               => "512k",
        tool_version     => $tool_version,
        devices_per_tool => $devices_per_tool,
        pass_thru        => "dsize=4k enable=microdelay rdelay=$max_microdelay"
                            . " wdelay=$max_microdelay oncerr=$oncerr"
                            . " prefix=\%d@%h\ enable=timestamp"
    );
    push(@Presets, \%dt_large_sequential_read);

    # Transactional type of workload
    my %dt_small_random_read = (
        id               => "Dt_Small_Random_Read",
        description      => "Small random reads simulating a typical database workload",
        io_tool          => "dt",
        write            => 1,
        noprogt          => 15,
        noprogtt         => 130,
        alarm            => 3,
        iotype           => "random",
        slices           => 6,
        pattern          => "iot",
        #limit            => $io_limit,
        bs               => "4k",
        tool_version     => $tool_version,
        devices_per_tool => $devices_per_tool,
        pass_thru        => "dsize=4k enable=microdelay rdelay=$min_microdelay"
                            . " wdelay=$min_microdelay oncerr=$oncerr"
                            . " prefix=\%d@%h\ enable=timestamp"
    );
    push(@Presets, \%dt_small_random_read);

    # Small Random I/Os with 70/30 read/Write Ratio. Typical OLTP environment
    my %dt_small_random_70_read = (
        id               => "Dt_Small_Random_70_Read",
        description      => "Small random reads and writes simulating a typical OLTP",
        io_tool          => "dt",
        write            => 1,
        noprogt          => 15,
        noprogtt         => 130,
        alarm            => 3,
        iotype           => "random",
        slices           => 3,
        pattern          => "iot",
        #limit               => $io_limit,
        bs               => "4k",
        tool_version     => $tool_version,
        devices_per_tool => $devices_per_tool,
        pass_thru        => "dsize=4k enable=microdelay rdelay=$min_microdelay"
                            . " wdelay=$min_microdelay oncerr=$oncerr"
                            . " prefix=\%d@%h\ enable=timestamp"
    );
    push(@Presets, \%dt_small_random_70_read);

    # Large sequential I/Os with 90/10 read/Write Ratio. Stress IO Bandwidth,
    # default for dt is sequential
    my %dt_large_sequential_90_read = (
        id               => "Dt_Large_Sequential_90_Read",
        description      => "Large sequential reads simulating lots of IO Bandwidth",
        io_tool          => "dt",
        write            => 1,
        noprogt          => 15,
        noprogtt         => 130,
        alarm            => 3,
        slices           => 3,
        pattern          => "iot",
        #limit            => $io_limit,
        bs               => "512k",
        tool_version     => $tool_version,
        devices_per_tool => $devices_per_tool,
        pass_thru        => "dsize=4k enable=microdelay rdelay=$max_microdelay"
                            . " wdelay=$max_microdelay oncerr=$oncerr"
                            . " prefix=\%d@%h\ enable=timestamp"
    );
    push(@Presets, \%dt_large_sequential_90_read);

    # Trickle IO so as not to overload the cluster to allow for other ops
    my %dt_trickle = (
        id           => "Dt_Trickle",
        description  => "Trickle the IO to maixmize number of LUN ops",
        io_tool      => "dt",
        write        => 1,
        noprogt      => 15,
        noprogtt     => 130,
        alarm        => 3,
        slices       => 3,
        pattern      => "iot",
        #limit        => $io_limit,
        bs           => "random",
        tool_version => $tool_version,
        devices_per_tool => $devices_per_tool,
        pass_thru    => "dsize=4k"
                       .' prefix="%d@%h"'
                       ." enable=microdelay"
                       ." rdelay=500000"
                       ." wdelay=500000"
                       ." oncerr=$oncerr"
                       ." enable=timestamp"
                       ." enable=syslog"
                       ." keepalivet=60"
    );
    push(@Presets, \%dt_trickle);

    # Highly sequential pattern, consisting of small writes
    my %dt_log_workload = (
        id           => "Dt_Log_Workload",
        description  => "Highly sequential pattern, consisting of small writes",
        io_tool      => "dt",
        write        => 1,
        noprogt      => 15,
        noprogtt     => 130,
        alarm        => 3,
        slices       => 3,
        iotype       => "sequential",
        pattern      => "iot",
        bs           => "8k",
        tool_version => $tool_version,
        devices_per_tool => $devices_per_tool,
        pass_thru    => "dsize=4k"
                       .' prefix="%d@%h"'
                       ." enable=microdelay"
                       ." rdelay=500000"
                       ." wdelay=500000"
                       ." oncerr=$oncerr"
                       ." enable=timestamp"
                       ." enable=syslog"
                       ." keepalivet=60"
    );
    push(@Presets, \%dt_log_workload);

    # MS Exchange load simulating a.edb file, random reads, 32kb size
    my %dt_ms_exchange_workload = (
        id           => "Dt_Ms_Exchange_Workload",
        description  => "Simulate Exchange's database file. Random reads",
        io_tool      => "dt",
        write        => 1,
        noprogt      => 15,
        noprogtt     => 130,
        alarm        => 3,
        slices       => 3,
        iotype       => "random",
        pattern      => "iot",
        bs           => "32k",
        tool_version => $tool_version,
        devices_per_tool => $devices_per_tool,
        pass_thru    => "dsize=4k"
                       .' prefix="%d@%h"'
                       ." enable=microdelay"
                       ." rdelay=500000"
                       ." wdelay=500000"
                       ." oncerr=$oncerr"
                       ." enable=timestamp"
                       ." enable=syslog"
                       ." keepalivet=60"
    );
    push(@Presets, \%dt_ms_exchange_workload);

    # DT FileSystem Workload
    my %dt_default_filesystem = (
        id           => "Dt_Default_FileSystem",
        description  => "Default Filesystem Workload bypasses host cache.",
        io_tool      => "dt",
        write        => 1,
        noprogt      => 15,
        alarm        => 3,
        procs        => 1, # Use one proc to limit the number of files dt uses.
                           # by default each file is the size of "limit".
        pattern      => "iot",
        bs           => "random",
        limit        => $io_limit, # Limit is required for FS workloads.
        tool_version => $tool_version,
        devices_per_tool => $devices_per_tool,
        pass_thru    => "dsize=4k"
                       .' prefix="%d@%h"'
                       ." enable=microdelay"
                       ." rdelay=$min_microdelay"
                       ." wdelay=$min_microdelay"
                       ." oncerr=$oncerr"
                       ." enable=timestamp"
                       ." enable=syslog"
                       ." keepalivet=60"
                       ." bufmodes=unbuffered" # We are telling dt to bypass
                                               # the Filesystem cache.
    );
    push(@Presets, \%dt_default_filesystem);

    # DT Custom pulls values from fit.config or uses defaults
    my $dtc_noprogt   = param("FIT_DT_CUSTOM_NOPROGT", "-default", DEFAULT_IO_DT_NOPROGT);
    my $dtc_noprogtt  = param("FIT_DT_CUSTOM_NOPROGTT", "-default", DEFAULT_IO_DT_NOPROGTT);
    my $dtc_alarm     = param("FIT_DT_CUSTOM_ALARM", "-default", DEFAULT_IO_DT_ALARM);
    my $dtc_slices    = param("FIT_DT_CUSTOM_SLICES", "-default", DEFAULT_IO_DT_SLICES);
    my $dtc_pattern   = param("FIT_DT_CUSTOM_PATTERN", "-default", DEFAULT_IO_DT_PATTERN);
    my $dtc_bs        = param("FIT_DT_CUSTOM_BLOCKSIZE", "-default", DEFAULT_IO_BS);
    my $dtc_tool_ver  = param("FIT_DT_CUSTOM_TOOL_VERSION", "-default", $tool_version);
    my $dtc_dper_tool = param("FIT_DT_CUSTOM_DEVICES_PER_TOOL", "-default", $devices_per_tool);
    my $dtc_pt_dsize  = param("FIT_DT_CUSTOM_PASS_THRU_DSIZE", "-default", DEFAULT_IO_DT_DSIZE);
    my $dtc_pt_prefix = param("FIT_DT_CUSTOM_PASS_THRU_PREFIX", "-default", DEFAULT_IO_DT_PREFIX);
    my $dtc_pt_enable = param("FIT_DT_CUSTOM_PASS_THRU_ENABLE", "-default", DEFAULT_IO_DT_ENABLE);
    my $dtc_pt_rdelay = param("FIT_DT_CUSTOM_PASS_THRU_RDELAY", "-default", DEFAULT_IO_READ_MAX_MICRODELAY);
    my $dtc_pt_wdelay = param("FIT_DT_CUSTOM_PASS_THRU_WDELAY", "-default", DEFAULT_IO_WRITE_MAX_MICRODELAY);
    my $dtc_pt_iotune = param("FIT_DT_CUSTOM_PASS_THRU_IOTUNE", "-default", DEFAULT_IO_DT_IOTUNE);
    my $dtc_pt_oncerr = param("FIT_DT_CUSTOM_PASS_THRU_ONCERR", "-default", $oncerr);
    my $dtc_pt_keepat = param("FIT_DT_CUSTOM_PASS_THRU_KEEPALIVET", "-default", DEFAULT_IO_DT_KEEPALIVET);
    my $dtc_pt_custom = param("FIT_DT_CUSTOM_PASS_THRU_CUSTOM", "-default", "");
    my $dtc_pass_thru = "dsize=$dtc_pt_dsize"
                       ." prefix=$dtc_pt_prefix"
                       ." enable=$dtc_pt_enable"
                       ." rdelay=$dtc_pt_rdelay"
                       ." wdelay=$dtc_pt_wdelay"
                       ." iotune=$dtc_pt_iotune"
                       ." oncerr=$dtc_pt_oncerr"
                       ." keepalivet=$dtc_pt_keepat"
                       ." $dtc_pt_custom";
    my %dt_custom = (
        id               => "Dt_Custom",
        description      => "DT Custom IO based on values from fit.config",
        io_tool          => "dt",
        write            => 1,
        noprogt          => $dtc_noprogt,
        noprogtt         => $dtc_noprogtt,
        alarm            => $dtc_alarm,
        slices           => $dtc_slices,
        pattern          => $dtc_pattern,
        bs               => $dtc_bs,
        tool_version     => $dtc_tool_ver,
        devices_per_tool => $dtc_dper_tool,
        pass_thru        => $dtc_pass_thru,
    );
    push(@Presets, \%dt_custom);

    return (wantarray()) ? @Presets : \@Presets;
}

################################################################################
# Objective: Returns a San FIT IO preset.
# Params:    An optional hash containing:
#              * -id: The id corresponding to a SAN Fit preset. .
# Returns:   A hash reference containing an I/O preset
################################################################################
sub get_io_preset {
    my (%args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # Locals
    my $load = undef;
    my $id   = "";

    # Validate params
    if (defined ($args{id})) {
        $id = $args{id};
    } else {
        return $load;
    }

    foreach my $preset (@Presets) {
        my %load = %{$preset};
        my $preset_id = $load{id};
        if ($id =~ /^$preset_id$/i) {
            return \%load;
        }
    }

    return $load;
} # End get_io_preset

################################################################################
# Objective: This method returns an I/O load selected randomly from SAN FIT's
#            load presets.
# Params:  None.
# Returns: A hash reference containing an I/O preset
################################################################################
sub get_random_io_preset {
    my (%args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # Locals
    my $load   = {};
    my %preset = ();

    $load = $Presets[int(rand(@Presets))];
    %preset = %{$load};

    return \%preset;
} # End get_random_io_preset

1;
