# Copyright (c) 2010 - 2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide a set of lun operations for SAN FIT thpls and automation
## @author dl-san-fit@netapp.com
## @status private
# ntest -plain /u/carteram/p4/test/san/fit/scenarios/sanops/lunops.pm

use strict;

# Package Name
package LunOps;

# NetApp Module Imports
use San::SanUtil;
use SubIPC;
use Tharn;
use IOMan;
use TCD;
use San::ControllerCluster::Filer10M;
use San::FIT::SanFitUtil;

# Perl's Module Imports
use List::MoreUtils qw(uniq);
use Data::Dumper;

# Globals
use vars qw (
    $Test
    @Lunops
    $Num_Vols
    $LunOps_Enabled
    $Num_LunOps_Runs
    $ControllerCluster_Obj
);

# Constants
use constant {
    DEFAULT_NUM_LUNOPS_RUNS              => 10,
    DEFAULT_LUN_SIZE                     => '40m',
    DEFAULT_LUN_OPS_SLEEP_AFTER_CREATE   => 10,
    DEFAULT_NUM_VOLS                     => 5
};

################################################################################
# Objective: To create, map, unmap and delete luns.  This will run in the
# background while other test functions are being ran.
# Params: $Test - this is the param for the tests to be ran.
#         $c_obj - this is the cserver object.
# Returns: True or False depending if all the lun maps/unmaps were successful
################################################################################
sub lun_ops {
    my (%args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # local variables
    my ($subtest,$lun_size) = undef;
    my (@results) = ();
    my ($fail_cnt)       = 0;
    my (@volumes)        = ();
    my (@sources)        = ();
    my ($cserver_obj)    = undef;
    my ($num_runs)       = 0;
    my ($lunops_timeout) = 0;

    if (defined ($args{Test})) {
        $Test = $args{Test};
    } else {
        $Test->nlog("A TCD object must be passed into the test script");
        return $SanUtil::FALSE;
    }

    if (defined ($args{c_obj})){
        $ControllerCluster_Obj = $args{c_obj};
    } else {
        $Test->nlog("A Cluster object was not defined");
        return $SanUtil::FALSE;
    }

    $lunops_timeout =
        $Test->get_param("LUNOPS_TIMEOUT", DEFAULT_LUN_OPS_SLEEP_AFTER_CREATE);
    $num_runs = $Test->get_param("NUM_LUNOPS_RUNS",
        DEFAULT_LUN_OPS_SLEEP_AFTER_CREATE);
    $Num_Vols = $Test->get_param("NUM_VOLS",
        DEFAULT_NUM_VOLS);
    $cserver_obj = $ControllerCluster_Obj->get_cserver_obj();
    $lun_size = $Test->get_param("LUN_SIZE", DEFAULT_LUN_SIZE);

    @volumes = $cserver_obj->fetch_volume_objs();
    if ($cserver_obj->error()) {
        $Test->nlog("Failed fetching local vols: ", $cserver_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Since fetch vols returns ALL volumes, filter out the rootvols, vol0s
    # and sanblaze vservers
    foreach my $vol (@volumes) {
        if ($Num_Vols > 0) {
            # ignore rootvols or sanblaze vols
            next if ($vol->get_name() =~ /rootvol|vol0|sanblaze/i);
            next if ($vol->get_vserver() =~ /sanblaze/i);
            $Test->nlog("Luns will be saved to this volume: " . $vol->get_name());
            push(@sources, $vol);
            $Num_Vols--;
        }
    }

    #Run number of times specified by user
    push (@Lunops, subtest(
        sub {
            while ($num_runs > 0) {
                # Create LUNs
                $Test->step("Run number " . $num_runs .
                    "\nCreating new LUNs on source volumes and Mapping them");
                    my $luns = &_add_multiflavored_luns(vols     => \@sources,
                                                        lun_size => $lun_size);
                    unless (scalar (@{$luns})) {
                        Tharn::logresult(
                            "FAIL",
                            msg => "Failed to create and map new LUNs");
                        return $SanUtil::FALSE;
                    }

                sleep($lunops_timeout);

                #Delete the luns
                $Test->step("Deleting ALL LUNs on destination volumes");
                my $status = &_delete_multiflavored_luns(
                    vols => \@sources,
                    cmd    => "all",
                    filter => "lunops*"
                );
                unless ($status) {
                    Tharn::logresult(
                        "FAIL",
                        msg => "Failed to delete and unmap new LUNs");
                    return $SanUtil::FALSE;
                }
                $num_runs--;

                sleep($lunops_timeout);
            }
        },
        -runid => "lun_ops",
        '-bg',
        -onexit,
            sub {
                my ($subtest) = @_;
                # Check the results
                @results = $subtest->results();
                foreach my $result (@results) {
                    next if $result->type()=~/WARN/;
                        if ($result->severity() >= 0) {
                            $Test->nlog("   - " . $result->type() .
                                "(" . $result->severity . ") RESULT" . " for " .
                            $subtest->runid(), "   - msg: " . $result);
                        if ($result->type() =~ /FAIL|SCRIPT|CONFIG/i) {
                            $fail_cnt++;
                        }
                    }
                }
                if ($fail_cnt > 0){
                    $Test->nlog("Lunops failed $fail_cnt times.");
                    return $SanUtil::FALSE;
                } else{
                    $Test->nlog("Lunops PASSED");
                    return $SanUtil::TRUE;
                }
            }
        )
    );
    return $SanUtil::TRUE
}

#===============================================================================
# Description: Creates different types of LUNs
# Params:
#    Required:
#        vols => list of volume objects.
#    Returns:
#      * A reference to a list containing LUN objects
#===============================================================================
sub _add_multiflavored_luns {
    my (%args) = @_;
    %args = &SanUtil::strip_dash(%args);

    # locals
    my $lun = undef;
    my (@luns, @nodes) = ();
    my ($fail_count,$lun_num) = (0,0);
    my %seen = ();
    my @volumes = @{$args{vols}};
    my $prefix = "lunops_";
    my @ostypes = ("windows","linux","solaris");
    my @space_reservation = ("enabled", "disabled");
    my $lun_size = $args{lun_size};

    # Create space-reserved and non-reserved LUNs, different OS types on
    # every Volume
    foreach my $vol (@volumes) {
        my ($vserver_obj, @igroup_objs) = undef;
        my $node = $vol->get_parent_obj();
        my $igroup_name = "";

        if (ref ($node) =~ /vserver/i) {
            $node = $node->get_parent_obj();
        }

        if (!$seen{$node->get_name()}) {
            $node->reconnect();
            $seen{$node->get_name()} = 1;
        }

        # Get the vserver object
        $vserver_obj = $node->get_vserver_obj(vserver => $vol->get_vserver());
        @igroup_objs = $vserver_obj->get_igroup_objs();
        if ($vserver_obj->error()) {
            $Test->nlog("Could not get igroups associated to this vserver",
                        $vserver_obj->errmsg());
            $fail_count++;
            next;
        }

        # Use the first igroup on each vserver
        if (scalar (@igroup_objs)) {
            $igroup_name = $igroup_objs[0]->get_name();
        } else {
            $Test->nlog("Could not find any igroups!");
            next;
        }

        foreach my $ostype (@ostypes){
            foreach my $space_reservation (@space_reservation) {
                my $lun_name = $prefix . $ostype . "_" . $space_reservation;
                if (!$vol->lun_exist(lun => $lun_name)){
                    $lun = $vol->create_lun_obj(
                        lun => $prefix . $ostype . "_" . $space_reservation,
                        size   => $lun_size,
                        ostype => $ostype,
                        'space-reserve' => $space_reservation
                    );
                if ($vol->error()) {
                    $Test->nlog('Could not create ' . $lun_name,
                        $vol->errmsg());
                        $fail_count++;
                }

                # Map it and cache it
                $lun->map(igroup => $igroup_name);
                if ($lun->error()) {
                    $Test->nlog('Failed to map' . $lun_name .
                                ' to $igroup_name',
                                $lun->errmsg());
                        $fail_count++;
                    }
                } else {
                    $lun =
                        $vol->get_lun(name => $lun_name);
                }
                push (@luns, $lun);
            }
        }
    }
    return \@luns;
} # End of _add_multiflavored_luns

#===============================================================================
# Description: Unmaps and Deletes LUNs based on a list of volumes
# Params:
#    Required:
#        cmd => unmap|delete|all.
#    Returns:
#      * TRUE or FALSE depending on whether the LUNs were deleted or not
#===============================================================================
sub _delete_multiflavored_luns {
    my (%args) = @_;

    my @vols = $args{vols} if defined ($args{vols});
    my (@lun_objs, @maps, @nodes) = ();
    my $lun_filter                = "lunops*";
    my $node_obj                  = undef;
    my %seen                      = ();
    my $cmd                       = "all";
    my @volumes                   = @{$args{vols}};

    if (defined ($args{cmd})) {
        $cmd = $args{cmd};
    }

    if (defined ($args{filter})) {
        $lun_filter = $args{filter};
    }

    # Get all the node objects used by this group and reconnect them
    # or else the subtests will fail
    foreach my $vol (@volumes) {
        my $node = $vol->get_parent_obj();
        if (ref ($node) =~ /vserver/i) {
            $node = $node->get_parent_obj();
        }
        if (!$seen{$node->get_name()}) {
            $node->reconnect();
            $seen{$node->get_name()} = 1;
        }
        push (@lun_objs, $vol->get_lun_objs());
        push (@nodes, $node);
    }

    # Get all lunops LUN objects, unmap them and destroy them
    unless ($lun_filter eq "*") {
        @lun_objs = $nodes[0]->fetch_lun_objs(-lun => $lun_filter,
                                               -volume => "sanfit_*");
        if ($nodes[0]->error()) {
            $Test->nlog("Failed to get a list of lunops LUNs",
                        $nodes[0]->errmsg());
            return $SanUtil::FALSE;
        }
    }

    foreach my $lun (@lun_objs) {
        if ($cmd =~ /unmap|all/i) {
            @maps = $lun->get_maps();
            if ($lun->error()) {
                $Test->nlog("Failed to get list of igroups", $lun->errmsg());
                return $SanUtil::FALSE;
            }
            foreach my $map (@maps) {
                $lun->unmap(igroup => $map);
                if ($lun->error()) {
                    if ($lun->errmsg() =~ /no entries matching your query/i) {
                        $Test->nlog("LUN " . $lun->get_name() . " was not found");
                    } else {
                        $Test->nlog("Failed to unmap " . $lun->get_name(),
                                    $lun->errmsg());
                        return $SanUtil::FALSE;
                    }
                }
            }
        }
        if ($cmd =~ /delete|all/i) {
            unless ($lun->is_readonly()){
                $lun->delete();
                if ($lun->error()) {
                    if ($lun->errmsg() =~ /no entries matching your query/i) {
                        $Test->nlog("LUN " . $lun->get_name() . " was not found");
                    } else {
                        $Test->nlog("Failed to delete Lun " . $lun->get_name(),
                                    $lun->errmsg());
                        return $SanUtil::FALSE;
                    }
                }
            }
        }
    }

    return $SanUtil::TRUE;
} # End of _delete_multiflavored_luns

1;