# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Base module for all Misfit components. Parent class
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Core::Base
# Parent class for all Misfit Feature modules.

package San::FIT::Misfit::Core::Base;

# Compiler Directives
use strict;
use warnings;
use Data::Dumper;

# Imported modules
use Tharn;
use ParamsMgr;
use NATE::ParamSet;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::MessageBroker;

use constant {
    SEVERITY_NA => 2,
};

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Creates new base object.  Base is the parent class for all Misfit feature
#   modules.  It will initialize all default Misfit features for the
#   feature modules such as creating queues and exchanges.  Also provides
#   access to common functions for Misfit modules.
#
#   Exchanges:
#       Module ID - ie., San::FIT::Misfit::Features::Snapmirror
#       errors - Exchange for error messages.  Only one errors exchange is
#           created per test by default.
#       critical - Exchange for critical messages.  Only one critical exchange
#           is created per test by default.
#
#   Queues:
#       Module ID - ie., San::FIT::Misfit::Features::Snapmirror
#       errors - Queue for error messages.  One errors queue is subscribed to
#           the errors exchange by default.
#       critical - Exchange for critical messages.  One critical queue is
#           subscribed to the critical exchange per test by default
#
#   Returns:
#       Base object
#===============================================================================
sub new {
    my ($class, %args) = @_;

    my $self = bless {}, $class;

    &ParamsMgr::process_args(%args);
    my $start_handlers = &ParamsMgr::get_arg(
        arg      => "start_handlers",
        required => FALSE,
        default  => FALSE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    # Each testbed has its own virtual host so set it based on the parameter
    my $vhost = param("TESTBED", '-default', '/');
    $self->{vhost} = $vhost;

    # Initialize and cache a message broker
    $self->{message_broker} = San::FIT::Misfit::Core::MessageBroker->new(
        vhost => $vhost
    );
    unless (ref($self->{message_broker}) =~ /San::FIT::Misfit::Core::MessageBroker/ ) {
        return {};
    }

    # Create a vhost if needed
    $self->{message_broker}->create_vhost(vhost => $vhost);

    # Declare exchanges
    $self->_create_default_exchanges();

    # Declare queues
    $self->_create_default_queues();

    if ($start_handlers) {
        $self->start_default_handlers();
    }

    return $self;
} # End of new

#===============================================================================
# Function: Execute
#   Execute a series of commands from a single feature.
#
#   Parameters:
#       operations - Hash containing the operation and any specific parameters
#           for that operation
#       params - Hash of parameters that will be used in each operation.  This
#           is where the list of objects for the commands to operate on.
#       background - TRUE: runs parallely/FALSE: runs serially
#
#   Returns:
#       TRUE|FALSE
#===============================================================================
sub execute {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $operations = &ParamsMgr::get_arg(
        'arg'      => "operations",
        'required' => TRUE
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {}
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => FALSE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my $status = TRUE;
    my $index = 0;

    foreach my $operation (@{$operations}) {
        if (defined($operation->{params})) {
            while (my ($key, $value) = each %{$operation->{params}}) {
                if (defined($params->{objects})) {
                    foreach my $object (@{$params->{objects}}) {
                        $object->{$key} = $value;
                    }
                }
            }
        }

        if(defined($operation->{background}) and exists($operation->{background})){
            $background = $operation->{background};
        }

        my $command = $operation->{command};
        $status = $self->$command(
            %{$params},
            id         => $self->{id},
            background => $background,
            runid      => $index . "_" . $self->{id}
        );
        if ($status != TRUE) {
            return FALSE;
        }
        $index++;
    }

    return $status;
} # End of execute

#===============================================================================
# Function: _run_workflow
#   Allows sanfit blocks to execute a series of commands by running
#   workflows.
#
#   Returns:
#       TRUE|FALSE
#===============================================================================
sub _run_workflow {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $objects = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => TRUE,
    );
    my $operation = &ParamsMgr::get_arg(
        'arg'      => "operation",
        'required' => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    my $msg_data = &ParamsMgr::get_arg(
        'arg'      => "msg_data",
        'required' => FALSE,
        'default'  => 1,
    );
    my $procs = &ParamsMgr::get_arg(
        'arg'      => "max_procs",
        'required' => FALSE,
        'default'  => 0,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $max_procs = param('MAX_PROCS', '-default', $procs);

    if ($max_procs == 0) {
        $max_procs = scalar(@{$objects});
    }
    my $fail_count = 0;

    $self->{message_broker}->publish_message(
        exchange => $self->{id},
        message  => {
            msg  => $operation . ".started",
            data => $msg_data
        }
    );

    # Replacing ':' with '_' to prevent NATE from messing log results
    my $nate_safe_id = $self->{id} . "__" . $operation;
    $nate_safe_id =~ s/\:/_/g;

    while (scalar(@{$objects})) {
        my @processes = splice(@{$objects}, 0, $max_procs);
        $self->{workflow} = San::FIT::Misfit::Core::Workflow->new(
            steps      => \@processes,
            iterations => $iterations,
            runid      => $nate_safe_id,
            background => $background,
        );
        try {
            $self->{workflow}->start();
            $self->{workflow}->wait() if $background;
            my $result = $self->{workflow}->{workflow}->worst_result();
            if(defined($result)){
                if ($result->severity() > SEVERITY_NA) {
                    $fail_count++;
                }
            }
            $self->{workflow}->destroy();
        } otherwise {
            $fail_count++;
        };
    }
    if (!$fail_count) {
        $self->{message_broker}->publish_message(
            exchange => $self->{id},
            message  => {
                msg  => $operation . ".finished",
                data => $msg_data
            }
        );
        return TRUE;
    } else {
        $self->{message_broker}->publish_message(
            exchange => $self->{id},
            message  => {
                msg  => $operation . ".failed",
                data => $msg_data
            }
        );
        return FALSE;
    }

    $Log->exit();
} # End of _run_workflow

#===============================================================================
# Description: NOTE: Abstract class (Perlish way...) Your module should
# implement an _operation module.
#===============================================================================
sub _operation {
    my ($self, %args) = @_;

    return TRUE;
} # End of _operation

#===============================================================================
# Function: get_default_queues
#   Get a list of default queues.
#
#   Returns:
#       Array or reference to an array containing the default queues.
#===============================================================================
sub get_default_queues {
    my ($self, %args) = @_;

    my @queues = ();
    if (defined($self->{default_queues})) {
        foreach (@{$self->{default_queues}}) {
            push (@queues, $_);
        }
    }

    return wantarray ? @queues: \@queues;
} # End of get_default_queues

#===============================================================================
# Function: local_message
#   Publishes a message. If no exchange is used, it sends it to
#   its default module exchange
#
#   Parameters:
#       exchange - Exchange to publish message to. Defaults to module ID.
#       routing_key - For topic exchanges with routing keys. Defaults to '#'
#       message - Message to be published.  Can be a string, hash or array.
#       vhost - Name of rabbit virtual host where exchanges are.  Defaults to
#           the testbed name
#
#   Returns:
#       TRUE
#===============================================================================
sub local_message {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $exchange = &ParamsMgr::get_arg(
        arg      => "exchange",
        required => FALSE,
        default => $self->{id}
    );
    my $routing_key = &ParamsMgr::get_arg(
        arg      => "routing_key",
        required => FALSE,
        default => ''
    );
    my $message = &ParamsMgr::get_arg(
        arg      => "message",
        required => TRUE,
    );
    my $vhost = &ParamsMgr::get_arg(
        arg      => "vhost",
        required => FALSE,
        default => $self->{vhost}
    );

    $self->{message_broker}->publish_message(
        vhost       => $vhost,
        exchange    => $exchange,
        routing_key => $routing_key,
        message     => $message
    );

    return TRUE;
} # End of local_message

#===============================================================================
# Function: local_get_message
#   Wait for message on module queue
#
#   Parameters:
#       message - String of message to wait for
#
#   Returns:
#       Received message
#===============================================================================
sub local_get_message {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $queue= &ParamsMgr::get_arg(
        arg      => "queue",
        required => FALSE,
        default => (caller(1))[0]
    );
    my $message = &ParamsMgr::get_arg(
        arg      => "message",
        required => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return {};
    }

    my $received = $self->{message_broker}->wait_for_message(
        message  => $message,
        queue    => $queue,
        vhost    => $self->{vhost}
    );

    return $received;
} # End of local_get_message

#===============================================================================
# Function: broadcast_error
#   Publishes a message to the testbed's error exchange
#
#   Parameters:
#       message - Message to be sent.  Can be string, hash, or array.
#
#   Returns:
#       TRUE
#===============================================================================
sub broadcast_error {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $message = &ParamsMgr::get_arg(
        arg      => "message",
        required => FALSE,
        default => (caller(1))[0] . '.unknown_error'
    );

    $self->local_message('exchange' => 'errors', 'message' => $message);

    return TRUE;
} # End of broadcast_error

#===============================================================================
# Function: broadcast_critical
#   Publishes a message to the testbed's critical exchange
#
#   Parameters:
#       message - Message to be sent.  Can be string, hash, or array.
#
#   Returns:
#       TRUE
#===============================================================================
sub broadcast_critical {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $message = &ParamsMgr::get_arg(
        arg      => "message",
        required => FALSE,
        default => (caller(1))[0] . '.unknown_critical'
    );

    $self->local_message('exchange' => 'critical', 'message' => $message);

    return TRUE;
} # End of broadcast_critical

#===============================================================================
# Description: Starts a handler critical messages
# XXX - This needs to be designed and discussed...
#===============================================================================
sub start_default_handlers {
    my ($self, %args) = @_;

    my @queues = $self->get_default_queues();
    my @subtests = ();
    my $module_name = (caller(1))[0];

    foreach my $queue (@queues) {
        # Focusing on critical messages for now
        next unless ($queue->{queue} =~ /critical/i);
        my $subtest = subtest(
            sub {
                $self->{message_broker}->wait_for_message(
                    message => 'exit',
                    queue   => $module_name .'.critical'
                );
                $self->_stop();
            },
            -runid => 'handler_' . $queue->{queue},
            -bg,
            -onexit => sub {
                Tharn::logresult("PASS", msg => "Exit handler");
            }
        );
        push (@subtests, $subtest);
    }

    $self->{default_handlers} = \@subtests;

    return TRUE;
} # End of start_default_handlers

#===============================================================================
# Description: Stops handlers for errors and critical messages
#===============================================================================
sub stop_default_handlers {
    my ($self, %args) = @_;

    if (defined ($self->{default_handlers})) {
        foreach my $subtest (@{$self->{default_handlers}}) {
            $subtest->kill();
        }
    }

    return TRUE;
} # End of stop_default_handlers

#===============================================================================
# Function: get_default_exchanges
#   Get a list of default exchanges.
#
#   Returns:
#       Array or reference to an array containing the default exchanges.
#===============================================================================
sub get_default_exchanges {
    my ($self, %args) = @_;

    my @exchanges = ();
    if (defined($self->{default_exchanges})) {
        foreach (@{$self->{default_exchanges}}) {
            push (@exchanges, $_);
        }
    }

    return wantarray ? @exchanges : \@exchanges;
} # End of get_default_exchanges

#===============================================================================
# Function: whoami
#   Returns:
#       Name of this package
#===============================================================================
sub whoami {
    my ($self, %args) = @_;
    return (caller(1))[0];
}

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Functions: _create_default_exchanges.
#   If needed, create default exchanges for a Misfit block.
#
#   Returns:
#       TRUE
#===============================================================================
sub _create_default_exchanges {
    my ($self, %args) = @_;

    $self->{default_exchanges} = [
        {
            exchange => (caller(1))[0],
            type     => 'topic'
        },
        {
            exchange => 'errors',
            type     => 'fanout'
        },
        {
            exchange => 'critical',
            type     => 'fanout'
        }
    ];

    # Default exchange for Misfit
    $self->{message_broker}->create_exchange(
        exchanges => $self->{default_exchanges},
        vhost     => $self->{vhost},
        mode      => "http"
    );

    return TRUE;
} # End of _create_default_exchange

#===============================================================================
# Function: _create_default_queues
#   Create default queues for a Misfit block.
#
#   Returns:
#       TRUE
#===============================================================================
sub _create_default_queues {
    my ($self, %args) = @_;

    $self->{default_queues} = [
        {
            queue       => (caller(1))[0].'.errors',
            exchange    => 'errors',
            routing_key => ''
        },
        {
            queue       => (caller(1))[0].'.critical',
            exchange    => 'critical',
            routing_key => ''
        },
        {
            queue       => (caller(1))[0],
            exchange    => (caller(1))[0],
            routing_key => '#'
        }
    ];

    # Default queues in Misfit
    $self->{message_broker}->create_queue(
        queues => $self->{default_queues},
        vhost  => $self->{vhost},
        mode   => "http"
    );

    return TRUE;
} # End of _create_default_queues

#===============================================================================
# Description: Handles messages related to critical events on the testbed
#===============================================================================
sub _stop {
    my ($self, %args) = @_;

    my @subtests = Subtest::get_live();
    foreach my $subtest (@subtests) {
        $subtest->kill();
    }

    return TRUE;
} # End of _stop

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/core/Base>

=cut

1;