# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Message Broker. RabbitMQ
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Core::MessageBroker
# MessageBroker functions for modules to interact with RabbitMQ.

package San::FIT::Misfit::Core::MessageBroker;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use JSON;
use Tharn;
use English qw($OSNAME);
use ParamsMgr;
use Data::Dumper;
use HTTP::Request;
use HTTP::Response;
use LWP::UserAgent;
use NATE::ParamSet;
use POSIX qw(strftime);
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use NATE::Events qw(listener_add timer_add);

use constant {
    BROKER     => "/usr/software/test/noarch/dev/san/fit/MessageBrokerCLI.py",
    BROKER_WIN => 'X:\\noarch\\Rsierranevada\\san\\fit\\MessageBrokerCLI.py',
    URL        => "http://sanfit.gdl.englab.netapp.com",
    PORT       => 55672,
    VHOST      => "/",
    USER       => "guest",
    PASSWORD   => "guest",
    AGENT      => "Misfit/v1.0",
    DELAY      => 1,
    TRUNCATE   => 50000,
    TIMEOUT    => 3600,
    PYTHON_BIN => "/usr/software/bin/python2.6 ",
    PYTHON_WIN => "X:\\ActiveState\\Python\\Python26\\python.exe ",
    WINDOWS    => "MSWin32",
};

#===============================================================================
# Constructor: new
#   Message Broker constructor.  Creates the vhost and url data for
#   communicating with the RabbitMQ server.
#
#   Parameters:
#       vhost - Name of vhost to create for test. Default is the testbed name
#       url - URL of RabbitMQ server
#       port - Port of RabbitMQ service on server
#       mode - Many messaging features are supported via HTTP or CLI interface.
#           Defaults to HTTP.
#       user - RabbitMQ username for interacting with the server. Defaults to
#           guest.
#       password - Password for user. Defaults to guest password.
#       purge - Delete the vhost before attempting to create one. Defaults to
#           FALSE
#       broker - override the default messagebroker.
#           Default: /usr/software/test/noarch/dev/san/fit/MessageBrokerCLI.py
#           Windows Default: X:\noarch\Rsierranevada\san\fit\MessageBrokerCLI.py
#
#   Returns:
#       MessageBroker object
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;

    &ParamsMgr::process_args(%args);
    $self->{vhost} = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => param("TESTBED", '-default', '/')
    );
    $self->{url} = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    $self->{port} = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => PORT
    );
    $self->{mode} = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => "http"
    );
    $self->{user} = &ParamsMgr::get_arg(
        'arg'      => "user",
        'required' => FALSE,
        'default'  => USER
    );
    $self->{password} = &ParamsMgr::get_arg(
        'arg'      => "password",
        'required' => FALSE,
        'default'  => PASSWORD
    );
    my $purge = &ParamsMgr::get_arg(
        'arg'      => "purge",
        'required' => FALSE,
        'default'  => FALSE
    );
    my $broker = &ParamsMgr::get_arg(
        'arg'      => "broker",
        'required' => FALSE,
        'default'  => _os_is_windows() ? BROKER_WIN : BROKER,
    );

    $self->{broker} = $broker;
    $self->{python} = _os_is_windows() ? PYTHON_WIN : PYTHON_BIN;

    my $url = $self->{url} . ":" . $self->{port} ."/api/";
    $self->{http_methods} = {
        "aliveness"   => $url . "aliveness-test",
        "nodes"       => $url . "nodes",
        "extensions"  => $url . "extensions",
        "definitions" => $url . "definitions",
        "connections" => $url . "connections",
        "permissions" => $url . "permissions",
        "channels"    => $url . "channels",
        "exchanges"   => $url . "exchanges",
        "queues"      => $url . "queues",
        "bindings"    => $url . "bindings",
        "vhosts"      => $url . "vhosts",
        "whoami"      => $url . "whoami",
        "users"       => $url . "users",
    };
    $self->{id} = "San::FIT::Misfit::Core::MessageBroker";

    # Create a vhost if there isn't one
    if ($self->vhost_exists(vhost => $self->{vhost})) {
        if ($purge) {
            $self->delete_vhost(vhost => $self->{vhost});
        }
    }
    unless ($self->vhost_exists(vhost => $self->{vhost})) {
        $self->create_vhost(vhost => $self->{vhost});
        # Set permissions on the vhost
        $self->set_permission(
            user      => $self->{user},
            password  => $self->{password},
            vhost     => $self->{vhost},
            configure => ".*",
            read      => ".*",
            write     => ".*",
        );
    }
    unless ($self->is_alive(vhost => $self->{vhost})) {
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Function: create_exchange
#   Creates RabbitMQ exchanges. An exchange receives messages from
#   producers and the other side it pushes them to queues.
#
#   Parameters:
#       exchanges - Reference to a list of exchange names.
#       vhost - Name of vhost to create exchanges on. Defaults to testbed name.
#       mode - Create exchanges via HTTP or CLI.  Defaults to mode set in
#           MessageBroker constructor.
#
#   Returns:
#       TRUE|FALSE
#===============================================================================
sub create_exchange {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $exchanges = &ParamsMgr::get_arg(
        'arg'      => "exchanges",
        'required' => FALSE,
        'default'  => []
    );
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => $self->{mode}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $errors = 0;
    foreach my $exchange (@{$exchanges}) {
        &ParamsMgr::process_args(%{$exchange});
        my $exchange = &ParamsMgr::get_arg(
            'arg'      => "exchange",
            'required' => TRUE
        );
        my $type = &ParamsMgr::get_arg(
            'arg'      => "type",
            'required' => FALSE,
            'default'  => 'topic'
        );
        my $durable = &ParamsMgr::get_arg(
            'arg'      => "durable",
            'required' => FALSE,
            'default'  => TRUE
        );
        if (&ParamsMgr::error()) {
            NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
            $Log->exit();
            return FALSE;
        }

        if ($self->exchange_exists(exchange => $exchange)) {
            # Exchange exists, skipping this one
            next;
        }

        if ($mode eq "cli") {
            my $cmd = "-v $vhost -c exchange_declare -e $exchange -t $type ";
            my $response = $self->_execute(
                cmd => $cmd,
            );
            unless ($self->_decode_cli_response(
                response => $response, return_boolean => TRUE)) {
                $Log->comment("Failed to create exchange $exchange.");
                $errors++;
            }
            sleep(DELAY);
        } elsif ($mode eq "http") {
            my $response_data = $self->_http_PUT(
                url => $self->{http_methods}->{exchanges} .
                       "/" . $vhost . "/" . $exchange,
                content => JSON::XS->new->utf8->encode({
                        type => $type,
                        durable => 'true',
                    }
                ),
            );
            if ($response_data->{rc} ne '204'){
                $Log->comment("Failed to create exchange: " . $response_data->{msg});
                $errors++;
            }
        }
    }
    $Log->exit();

    if ($errors > 0) {
        return FALSE;
    } else {
        return TRUE;
    }
} # End of create_exchange

#===============================================================================
# Function: create_queue
#   Creates RabbitMQ queues. A queue is bound to an exchange and
#   received messages routed by the broker.
#
#   Parameters:
#       queues - Reference to a list of hashes containing settings for creating
#           each queue. These include exchange to bind queue to, a routing key
#           for binding to the exchange, the queue name, and whether or not to
#           make the queue durable.
#       vhost - Name of vhost to create queues on. Defaults to testbed name.
#       mode - Create queues via HTTP or CLI.  Defaults to mode set in
#           MessageBroker constructor.
#
#   Returns:
#       TRUE|FALSE
#===============================================================================
sub create_queue {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $queues = &ParamsMgr::get_arg(
        'arg'      => "queues",
        'required' => TRUE
    );
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => $self->{mode}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $errors = 0;
    foreach my $queue (@{$queues}) {
        &ParamsMgr::process_args(%{$queue});
        my $exchange = &ParamsMgr::get_arg(
            'arg'      => "exchange",
            'required' => TRUE
        );
        my $routing_key = &ParamsMgr::get_arg(
            'arg'      => "routing_key",
            'required' => FALSE,
            'default'  => '#'
        );
        my $queue = &ParamsMgr::get_arg(
            'arg'      => "queue",
            'required' => TRUE
        );
        my $durable = &ParamsMgr::get_arg(
            'arg'      => "durable",
            'required' => FALSE,
            'default'  => TRUE
        );
        if (&ParamsMgr::error()) {
            NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
            $Log->exit();
            return FALSE;
        }

        if ($self->queue_exists(queue => $queue)) {
            # Queue exists, skipping this one
            next;
        }

        if ($mode eq "cli") {
            my $cmd = "-v $vhost -c queue_declare -q $queue -e $exchange " .
                      "-r $routing_key";
            my $response = $self->_execute(
                cmd => $cmd,
            );
            unless ($self->_decode_cli_response(
                response => $response, return_boolean => TRUE)) {
                $errors++;
            }
            sleep(DELAY);
        } elsif ($mode eq "http") {
            # Allow users to bind a queue to multiple exchanges
            my @exchanges = split(/,/, $exchange);
            foreach my $exch (@exchanges) {
                my $response_data = $self->_http_PUT(
                    url => $self->{http_methods}->{queues} . "/"
                            . $vhost . "/" . $queue,
                    content => JSON::XS->new->utf8->encode(
                        {
                            durable     => 'true',
                            exchange    => $exch,
                            routing_key => $routing_key,
                        }
                    ),
                );
                if ($response_data->{rc} ne '204'){
                    $Log->comment($response_data->{msg});
                    return FALSE;
                }
                # Bind the queue
                my $status = $self->create_binding(
                    vhost   => $vhost,
                    queues  => [{
                        queue       => $queue,
                        exchange    => $exch,
                        routing_key => $routing_key,
                        durable     => TRUE,
                    }],
                    mode => $mode,
                );
                unless ($status == TRUE) {
                    $Log->comment(
                        "Failed binding $queue to $exch using $routing_key"
                    );
                    return FALSE;
                }
            }
        }
    }
    $Log->exit();

    if ($errors > 0) {
        return FALSE;
    } else {
        return TRUE;
    }
} # End of create_queue

#===============================================================================
# Function: delete_exchange
#   Deletes exchanges.
#
#   Parameters:
#       exchanges - Reference to a list of exchanges to be deleted
#       vhost - Name of vhost to delete exchanges from.  Defaults to the vhost
#           set in MessageBroker constructor.
#       mode - Delete exchanges via HTTP or CLI.  Defaults to mode set in
#           MessageBroker constructor.
#
#   Returns:
#       TRUE|FALSE
#===============================================================================
sub delete_exchange {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $exchanges = &ParamsMgr::get_arg(
        'arg'      => "exchanges",
        'required' => FALSE,
        'default'  => []
    );
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => $self->{mode}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $errors = 0;
    foreach my $exchange (@{$exchanges}) {
        &ParamsMgr::process_args(%{$exchange});
        my $exchange = &ParamsMgr::get_arg(
            'arg'      => "exchange",
            'required' => TRUE
        );
        if (&ParamsMgr::error()) {
            NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
            $Log->exit();
            return FALSE;
        }
        unless ($self->exchange_exists(exchange => $exchange)) {
            next;
        }
        if ($mode eq "cli") {
            my $cmd = "-v $vhost -c exchange_delete -e $exchange ";
            my $response = $self->_execute(
                cmd => $cmd,
            );
            unless ($self->_decode_cli_response(
                response => $response, return_boolean => TRUE)) {
                $errors++;
            }
        } elsif ($mode eq "http") {
            my $out = $self->_http_DELETE(
                url => $self->{http_methods}->{exchanges} .
                       "/" . $vhost . "/" . $exchange,
            );
            if ($out->{rc} ne '204') {
                $Log->comment($out->{msg});
                return FALSE;
            }
        }
    }
    $Log->exit();

    if ($errors > 0) {
        return FALSE;
    } else {
        return TRUE;
    }
} # End of delete_exchange

#===============================================================================
# Function: delete_queue
#   Deletes queues.
#
#   Parameters:
#       queues- Reference to a list of queues to be deleted
#       vhost - Name of vhost to delete queues from.  Defaults to the vhost
#           set in MessageBroker constructor.
#       mode - Delete queues via HTTP or CLI.  Defaults to mode set in
#           MessageBroker constructor.
#
#   Returns:
#       TRUE|FALSE
#===============================================================================
sub delete_queue {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $queues = &ParamsMgr::get_arg(
        'arg'      => "queues",
        'required' => TRUE
    );
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => $self->{mode}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $errors = 0;
    foreach my $queue (@{$queues}) {
        &ParamsMgr::process_args(%{$queue});
        my $queue = &ParamsMgr::get_arg(
            'arg'      => "queue",
            'required' => TRUE
        );
        if (&ParamsMgr::error()) {
            NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
            $Log->exit();
            return FALSE;
        }
        unless ($self->queue_exists(queue => $queue)) {
            next;
        }
        if ($mode eq "cli") {
            my $cmd = "-v $vhost -c queue_delete -q $queue ";
            my $response = $self->_execute(
                cmd => $cmd
            );
            unless ($self->_decode_cli_response(
                response => $response, return_boolean => TRUE)) {
                $errors++;
            }
        } elsif ($mode eq "http") {
            my $out = $self->_http_DELETE(
                url => $self->{http_methods}->{queues} . "/"
                        . $vhost . "/" . $queue,
            );
            if ($out->{rc} ne '204') {
                $Log->comment($out->{msg});
                return FALSE;
            }
        }
    }
    $Log->exit();

    if ($errors > 0) {
        return FALSE;
    } else {
        return TRUE;
    }
} # End of delete_queue

#===============================================================================
# Function: publish_message
#   Publishes a message to an exchange.
#
#   Parameters:
#       vhost - Name of vhost exchange is on.  Defaults to vhost set in
#           constructor.
#       exchange - Name of exchange to publish message to.
#       routing_key - Routing key for message.  Defaults to '#'.
#       message - Message to be published.  May be string, array, or hash.
#       immediate - TRUE or FALSE telling to prioritize the message and skip
#           ACK to it, making its delivery faster but unreliable
#       mode - Message can be published via HTTP or CLI.  Defaults to mode set
#           in constructor.
#       properties - Extra RabbitMQ properties
#       payload_encoding - What format is the message being sent in.  Defaults
#           to "string".
#
#       Returns:
#           TRUE
#===============================================================================
sub publish_message {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $exchange = &ParamsMgr::get_arg(
        'arg'      => "exchange",
        'required' => FALSE,
        'default'  => ''
    );
    my $routing_key = &ParamsMgr::get_arg(
        'arg'      => "routing_key",
        'required' => FALSE,
        'default'  => '#'
    );
    my $message = &ParamsMgr::get_arg(
        'arg'      => "message",
        'required' => TRUE
    );
    my $immediate = &ParamsMgr::get_arg(
        'arg'      => "immediate",
        'required' => FALSE,
        'default'  => TRUE
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => $self->{mode}
    );
    my $properties = &ParamsMgr::get_arg(
        'arg'      => "properties",
        'required' => FALSE,
        'default'  => {},
    );
    my $payload_encoding = &ParamsMgr::get_arg(
        'arg'      => "payload_encoding",
        'required' => FALSE,
        'default'  => "string",
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    # If the exchange does not exist, create one on the fly, logging a message
    # letting users know that one will be created for them. We do not expect
    # Misfit users to do this, but we've seen a situation with people not
    # following best practices, where this simple check is needed.

    unless ($self->exchange_exists(exchange => $exchange)) {
        $Log->comment(
            "Exchange '$exchange' does not exist on '$vhost'. Misfit will create " .
            "it for you. It is not recommended to publish messages on exchanges " .
            "that do not exist!"
        );
        my $status = $self->create_exchange(
            exchanges => [{exchange => $exchange}],
            mode      => $mode,
            vhost     => $vhost,
        );
        if ($status == FALSE) {
            $Log->comment(
                "The exchange '$exchange' did not exist. Misfit tried to create " .
                "it for you and failed. Please make sure vhost '$vhost' is up " .
                "and operational. Your message cannot and will not be published."
            );
            return FALSE;
        }
    }

    my $coder = JSON::XS->new->allow_nonref;
    my $encoded_message = $coder->encode($message);

    if ($mode eq "cli") {
        chomp($encoded_message);
        $Log->comment("Message sent: $encoded_message to exchange $exchange");
        if (($encoded_message =~ /^{\".*\"}$/) or
            ($encoded_message =~ /^\[{\".*\"}\]$/)) {
            $encoded_message = "'$encoded_message'";
            $encoded_message = $coder->encode($encoded_message);
        }
        my $cmd = "-v $vhost -c message_publish -e $exchange -r ".
                  "'$routing_key' -m $encoded_message";
        my $response = $self->_execute(
            cmd => $cmd
        );
        if ($self->_decode_cli_response(
            response => $response, return_boolean => TRUE)) {
            return TRUE;
        }
        return FALSE;
    } elsif ($mode eq "http") {
        $Log->comment("Message sent: $encoded_message to exchange $exchange");
        my $out = $self->_http_POST(
            url => $self->{http_methods}->{exchanges} . "/"
                   . $vhost . "/" . $exchange . "/publish",
            content => JSON::XS->new->utf8->encode(
                {
                    routing_key      => $routing_key,
                    payload          => $encoded_message,
                    properties       => $properties,
                    payload_encoding => $payload_encoding,
                }
            ),
        );
        if ($out->{rc} ne "200") {
            $Log->comment($out->{msg});
            return FALSE;
        }
    }
    $Log->exit();

    return TRUE;
} # End of publish_message

#===============================================================================
# Function: create_vhost
#   Creates a virtual host on the RabbitMq broker.
#
#   Parameters:
#       vhost - Name of vhost to create.  Defaults to default vhost set in
#           the constructor.
#       mode - Vhost can be created via HTTP or CLI.  Defaults to default mode
#           set in the constructor.
#
#   Returns:
#       TRUE
#===============================================================================
sub create_vhost {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => $self->{mode}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $errors = 0;
    if ($mode eq "cli") {
        my $cmd = "-c vhost_declare -v $vhost";
        my $response = $self->_execute(
            cmd => $cmd,
        );
        unless ($self->_decode_cli_response(
            response => $response, return_boolean => TRUE)) {
            $Log->comment("Failed to create vhost $vhost.");
            $errors++;
        }
    } elsif ($mode eq "http") {
        my $out = $self->_http_PUT(
            url => $self->{http_methods}->{vhosts} . "/" . $vhost
        );
        if ($out->{rc} ne '204'){
            $Log->comment($out->{msg});
            $errors++;
        }
    }
    $Log->exit();

    if ($errors > 0) {
        return FALSE;
    } else {
        return TRUE;
    }
} # End of create_vhost

#===============================================================================
# Function: delete_vhost
#   Deletes a virtual host on the RabbitMq broker.
#
#   Parameters:
#       vhost - Name of vhost to delete.  Defaults to default vhost set in the
#           constructor.
#       mode - Vhosts can be deleted via HTTP or CLI.  Defaults to default mode
#           set in constructor.
#
#   Returns:
#       TRUE
#===============================================================================
sub delete_vhost {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => $self->{mode}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $errors = 0;
    if ($mode eq "cli") {
        my $cmd = "-c vhost_delete -v $vhost ";
        my $response = $self->_execute(
            cmd => $cmd,
        );
        unless ($self->_decode_cli_response(
            response => $response, return_boolean => TRUE)) {
            $Log->comment("Failed to delete vhost $vhost.");
            $errors++;
        }
    } elsif ($mode eq "http") {
        my $out = $self->_http_DELETE(
            url => $self->{http_methods}->{vhosts} . "/" . $vhost
        );
        if ($out->{rc} ne '204') {
            $Log->comment($out->{msg});
            $errors++;
        }
    }
    $Log->exit();

    if ($errors > 0) {
        return FALSE;
    } else {
        return TRUE;
    }
} # End of delete_vhost

#===============================================================================
# Function: wait_for_message
#   Waits or blocks until a message arrives to a queue.
#
#   Parameters:
#       queue - Name of queue to digest messages from.
#       message - String to match in message.
#       timeout - Amount of time in seconds to wait for message before moving
#           on.
#       vhost - Name of vhost where queue is.
#       ack_all - If 'true', the wait_for_message in the CLI interface will
#           acknowledge every message received and remove them from the queue.
#           Setting to 'false' will only acknowledge the message that matches
#           the message string.  Defaults to 'true'.
#       verbose - If 'true' prints all messages, expected or not, that arrive to
#           the given queue. Defaults to 'true'.
#
#   Returns:
#       Message that matches 'message' parameter.
#===============================================================================
sub wait_for_message {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $queue = &ParamsMgr::get_arg(
        'arg'      => "queue",
        'required' => TRUE
    );
    my $message = &ParamsMgr::get_arg(
        'arg'      => "message",
        'required' => TRUE
    );
    my $timeout = &ParamsMgr::get_arg(
        'arg'      => "timeout",
        'required' => FALSE,
        'default'  => TIMEOUT
    );
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $ack_all = &ParamsMgr::get_arg(
        'arg'      => "ack_all",
        'required' => FALSE,
        'default'  => "true" # set 'false' to only ack the found message
    );
    my $verbose = &ParamsMgr::get_arg(
        'arg'      => "verbose",
        'required' => FALSE,
        'default'  => "true"
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return '';
    }

    # Make sure the queue exists. If it does not, log a message and skip the
    # waiting block

    unless ($self->queue_exists(queue => $queue)) {
        $Log->comment(
            "\n**************************************************************"  .
            "\n* The queue where you're expecting to receive the message does"  .
            "\n* not exist. This could happen under certain circumstances"      .
            "\n* where the caller script is waiting for a message from another" .
            "\n* script that is not running or ready to send messages. The most".
            "\n* common case is a test script with test feature interaction "   .
            "\n* methods waiting for messages that will never be received. "    .
            "\n* If this happens the wait message will be bypassed and the "    .
            "\n* interaction becomes a no-op. Queue: $queue" .
            "\n**************************************************************"  .
            "\n\n"
        );
        $Log->exit();
        return {
            msg =>"Queue $queue does not exist. Skipping wait for $message",
            data => []
        };
    }

    $Log->comment("Waiting for: $message");
    if (($message =~ /^{\".*\"}$/) or
        ($message =~ /^\[{\".*\"}\]$/)) {
         $message = "'$message'";
    }
    my $cmd =
        "-v $vhost -c message_wait -q $queue -t $timeout -a $ack_all " .
        "-p $verbose -m \"$message\"";
    my $response = $self->_execute(cmd => $cmd);
    my $decoded_message = $self->_decode_cli_response(response => $response);

    # Return data directly by removing the extra status key that's only
    # used internally and not by common misfit users

    if (defined ($decoded_message->{msg}) and defined ($decoded_message->{status})) {
        if (ref($decoded_message->{msg}) =~ /HASH/i) {
            $decoded_message->{msg}->{status} = $decoded_message->{status};
            $Log->exit();
            return $decoded_message->{msg};
        }
    }
    $Log->exit();

    return $decoded_message;
} # End of wait_for_message

#===============================================================================
# Function: on_message
#   Executes a callback when a message gets to a queue
#
#   Parameters:
#       queue - Name of queue to digest messages from.
#       message - String to match in message.
#       timeout - Amount of time in seconds to wait for message before moving
#           on.
#       vhost - Name of vhost where queue is.
#       callback - Reference to method to be executed on receiving matching
#           message.
#       params - Parameters for callback method.
#       wait_for_callback - If TRUE, execution of the callback will block
#           the parent process from continuing.  If FALSE, the parent process
#           will continue executing while the callback executes.
#
#   Returns:
#       Pointer to the NATE process object of the callback.
#===============================================================================
sub on_message {
    my ($self, %args) = @_;

    $Log->enter();
    my $process = undef;
    &ParamsMgr::process_args(%args);
    my $queue = &ParamsMgr::get_arg(
        'arg'      => "queue",
        'required' => TRUE
    );
    my $message = &ParamsMgr::get_arg(
        'arg'      => "message",
        'required' => TRUE
    );
    my $callback = &ParamsMgr::get_arg(
        'arg'      => "callback",
        'required' => TRUE
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => undef
    );
    my $timeout = &ParamsMgr::get_arg(
        'arg'      => "timeout",
        'required' => FALSE,
        'default'  => TIMEOUT
    );
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost},
    );
    my $wait_for_callback = &ParamsMgr::get_arg(
        'arg'      => "wait_for_callback",
        'required' => FALSE,
        'default'  => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return undef;
    }

    # Wait for the message
    my $cmd = "-v $vhost -c message_wait -q $queue -t $timeout -m $message ";
    my $data = $self->_execute(
        cmd => $cmd
    );
    $Log->comment('# Received message, executing callback');

    $process = subtest(
        sub {
            if (defined ($params)) {
                $callback->(%{$params}, 'data' => $data);
            } else {
                $callback->('data' => $data);
            }
        },
        -runid => 'on_message_' . $message,
        -bg,
        -onexit => sub {
            Tharn::logresult("PASS", msg => "Message Received");
        }
    );

    if($wait_for_callback){
        $process->wait_finish();
    }
    $Log->exit();

    return $process;
} # End of on_message

#===============================================================================
# Function: is_alive
#   Checks the status of a vhost on the RabbitMQ server.
#
#   Parameters:
#       vhost - Name of vhost to check.  Defaults to default vhost set in
#           constructor.
#
#   Returns:
#       TRUE if vhost is alive or FALSE otherwise.
#===============================================================================
sub is_alive {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost},
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $response = $self->_http_GET(
        url => $self->{http_methods}->{aliveness} . "/$vhost",
    );
    if ($response->{rc} eq '200' and $response->{content}->{status} eq 'ok') {
        $Log->exit();
        return TRUE;
    } else {
        $Log->comment("Rabbit is not alive! " . $response->{msg});
        return FALSE;
    }
} # End of is_alive

#===============================================================================
# Function: get_nodes
#   Requests list of nodes from RabbitMQ cluster.
#
#   Parameters:
#       node - Name of node to check for (Optional).
#
#   Returns:
#       A list of nodes in the RabbitMQ cluster.
#===============================================================================
sub get_nodes {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $node = &ParamsMgr::get_arg(
        'arg'      => "node",
        'required' => FALSE,
        'default'  => ""
    );

    my $url = ($node ne "") ? $self->{http_methods}->{nodes} . "/"
            . $node : $self->{http_methods}->{nodes};
    my $nodes = $self->_http_GET(url => $url);
    if (($nodes->{msg} eq "OK") and ($nodes->{rc} eq "200")) {
        delete $nodes->{rc};
        delete $nodes->{msg};
    } else {
        $Log->comment("Failed to get nodes: " . $nodes->{msg});
        NATE::BaseException->throw("Failed to get nodes: " . $nodes->{msg});
        return wantarray ? () : [];
    }
    $Log->exit();

    return wantarray ? @{$nodes->{content}} : $nodes->{content};
} # End of get_nodes

#===============================================================================
# Function: get_extensions
#
#   Returns:
#       A list of Broker extensions to the management plugin
#===============================================================================
sub get_extensions {
    my ($self, %args) = @_;

    $Log->enter();
    my $extensions = $self->_http_GET(
        url => $self->{http_methods}->{extensions}
    );
    if (($extensions->{msg} eq "OK") and ($extensions->{rc} eq "200")) {
        delete $extensions->{rc};
        delete $extensions->{msg};
    } else {
        $Log->comment("Failed to get extensions: " . $extensions->{msg});
        NATE::BaseException->throw(
            "Failed to get extensions: " . $extensions->{msg}
        );
        return wantarray ? () : [];
    }
    $Log->exit();

    return wantarray ? @{$extensions->{content}} : $extensions->{content};
} # End of get_extensions

#===============================================================================
# Function: get_definitions
#
#   Returns:
#       A list of server definitions (exchanges, queues, etc)
#===============================================================================
sub get_definitions {
    my ($self, %args) = @_;

    $Log->enter();
    my $definitions = $self->_http_GET(url => $self->{http_methods}->{definitions});
    if (($definitions->{msg} eq "OK") and ($definitions->{rc} eq "200")) {
        delete $definitions->{rc};
        delete $definitions->{msg};
    } else {
        $Log->comment("Failed to get definitions: " . $definitions->{msg});
        NATE::BaseException->throw(
            "Failed to get definitions: ". $definitions->{msg}
        );
        return wantarray ? () : [];
    }
    $Log->exit();

    return $definitions->{content};
} # End of get_definitions

#===============================================================================
# Function: get_connections
#
#   Returns:
#       A list of all open connections
#===============================================================================
sub get_connections {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $name = &ParamsMgr::get_arg(
        'arg'      => "name",
        'required' => FALSE,
        'default'  => ""
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $url = ($name ne "") ? $self->{http_methods}->{connections} . "/"
                            . $name : $self->{http_methods}->{connections};
    my $connections = $self->_http_GET(url => $url);
    if (($connections->{msg} eq "OK") and ($connections->{rc} eq "200")) {
        delete $connections->{rc};
        delete $connections->{msg};
    } else {
        $Log->comment("Failed to get connections: " . $connections->{msg});
        NATE::BaseException->throw("Failed to get connections: " . $connections->{msg});
        return wantarray ? () : [];
    }
    $Log->exit();

    return wantarray ? @{$connections->{content}} : $connections->{content};
} # End of get_connections

#===============================================================================
# Function: get_channels
#
#   Parameters:
#       vhost - vhost to see open channels on.  Defaults to default vhost set
#           in constructor.
#
#   Returns:
#       A list of open channels.
#===============================================================================
sub get_channels {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $channel = &ParamsMgr::get_arg(
        'arg'      => "channel",
        'required' => FALSE,
        'default'  => ""
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $url = $self->{http_methods}->{channels};
    if ($channel ne '') {
        $url = $url . "/" . $channel;
    }

    my $channels = $self->_http_GET(url => $url);
    if (($channels->{msg} eq "OK") and ($channels->{rc} eq "200")) {
        delete $channels->{rc};
        delete $channels->{msg};
    } else {
        $Log->comment("Failed to get channels: " . $channels->{msg});
        NATE::BaseException->throw(
            "Failed to get channels: " . $channels->{msg}
        );
        return wantarray ? () : [];
    }
    $Log->exit();

    return wantarray ? @{$channels->{content}} : $channels->{content};
} # End of get_channels

#===============================================================================
# Function: get_exchanges
#
#   Get a list of all exchanges in RabbitMQ or a particular vhost.
#
#   Parameters:
#       vhost - Name of vhost to list exchanges from.  Defaults to empty string.
#       exchange - To check if exchange exists, pass the exchange name here,
#           along with appropriate vhost.  The function should return a list
#           with only the specified exchange.  This signifies the exchange
#           exists.
#
#   Note: To list all exchanges on the RabbitMQ server/cluster, leave
#           parameters blank.
#   Returns:
#       A list of all exchanges or in a given vhost
#===============================================================================
sub get_exchanges {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => ""
    );
    my $exchange = &ParamsMgr::get_arg(
        'arg'      => "exchange",
        'required' => FALSE,
        'default'  => ""
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $url = $self->{http_methods}->{exchanges};
    if ($vhost ne "") {
        if ($exchange ne '') {
            $url = $url . "/" . $vhost . "/" . $exchange;
        } else {
            $url = $url . "/" . $vhost;
        }
    }

    my $exchanges = $self->_http_GET(url => $url);
    if (($exchanges->{msg} eq "OK") and ($exchanges->{rc} eq "200")) {
        delete $exchanges->{rc};
        delete $exchanges->{msg};
    } else {
        $Log->comment("Failed to get exchanges: " . $exchanges->{msg});
        NATE::BaseException->throw(
            "Failed to get exchanges: " . $exchanges->{msg}
        );
        return wantarray ? () : [];
    }
    $Log->exit();

    return wantarray ? @{$exchanges->{content}} : $exchanges->{content};
} # End of get_exchanges

#===============================================================================
# Function: get_queues
#
#   Get a list of all queues in RabbitMQ or a particular vhost.
#
#   Parameters:
#       vhost - Name of vhost to list queues from.  Defaults to empty string.
#       exchange - To check if queue exists, pass the queue name here,
#           along with appropriate vhost.  The function should return a list
#           with only the specified queue.  This signifies the queue
#           exists.
#
#   Note: To list all queues on the RabbitMQ server/cluster, leave
#           parameters blank.
#
#   Returns:
#       A list of all queues or in a given vhost
#===============================================================================
sub get_queues {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => ""
    );
    my $queue = &ParamsMgr::get_arg(
        'arg'      => "queue",
        'required' => FALSE,
        'default'  => ""
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $url = $self->{http_methods}->{queues};
    if ($vhost ne "") {
        if ($queue ne "") {
            $url = $url . "/" . $vhost . "/" . $queue;
        } else {
            $url = $url . "/" . $vhost;
        }
    }

    my $queues = $self->_http_GET(url => $url);
    if (($queues->{msg} eq "OK") and ($queues->{rc} eq "200")) {
        delete $queues->{rc};
        delete $queues->{msg};
    } else {
        $Log->comment("Failed to get queues: " . $queues->{msg});
        NATE::BaseException->throw(
            "Failed to get queues: " . $queues->{msg}
        );
        return wantarray ? () : [];
    }
    $Log->exit();

    return wantarray ? @{$queues->{content}} : $queues->{content};
} # End of get_queues

#===============================================================================
# Function: empty_queue
#   Empty contents of a queue
#
#   Parameters:
#       vhost - Name of vhost where queue is.  Defaults to default vhost set in
#           constructor.
#       queue - Name of queue to purge messages from.
#
#   Returns:
#       TRUE|FALSE
#===============================================================================
sub empty_queue {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $queue = &ParamsMgr::get_arg(
        'arg'      => "queue",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $url = $self->{http_methods}->{queues};
    if ($vhost ne "") {
        if ($queue ne "") {
            $url = $url . "/" . $vhost . "/" . $queue . "/contents";
        } else {
            $url = $url . "/" . $vhost . "/contents";
        }
    }
    my $response = $self->_http_DELETE(url => $url);
    if (($response->{msg} eq "No Content") and ($response->{rc} eq "204")) {
        $Log->exit();
        return TRUE;
    } else {
        $Log->comment(
            "Failed purging contents of $queue: " . $response->{msg}
        );
        NATE::BaseException->throw("Could not purge queue: $response->{msg}");
        return FALSE;
    }
} # End of empty_queue

#===============================================================================
# Function: create_binding
#   Binds queue to an exchange
#
#   Parameters:
#       vhost - Name of vhost to create bindings on.  Defaults to default vhost
#           set in constructor.
#       queues - List of hashes containing data about each binding to be
#           created.
#
#   Queue Parameters:
#       exchange - Exchange to bind queue to.
#       routing_key - Routing_key to bind queue to exchange with.  Defaults to
#           '#'.
#       queue - Name of queue to be bound.
#===============================================================================
sub create_binding {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $queues = &ParamsMgr::get_arg(
        'arg'      => "queues",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    foreach my $binding (@{$queues}) {
        &ParamsMgr::process_args(%{$binding});
        my $exchange = &ParamsMgr::get_arg(
            'arg'      => "exchange",
            'required' => TRUE
        );
        my $routing_key = &ParamsMgr::get_arg(
            'arg'      => "routing_key",
            'required' => FALSE,
            'default'  => '#'
        );
        my $queue = &ParamsMgr::get_arg(
            'arg'      => "queue",
            'required' => TRUE
        );
        if ($exchange eq "" or $queue eq "") {
            return FALSE;
        }
        my $out = $self->_http_POST(
            url => $self->{http_methods}->{bindings} . "/" . $vhost
                   . "/e/" . $exchange. "/q/" . $queue,
            content => JSON::XS->new->utf8->encode(
                {routing_key => $routing_key}
            ),
        );
        if ($out->{rc} ne "201") {
            $Log->comment("Failed binding $queue: " . $out->{msg});
            NATE::BaseException->throw("Failed binding $queue: $out->{msg}");
            return FALSE;
        }
    }
    $Log->exit();

    return TRUE;
} # End of create_binding

#===============================================================================
# Function: get_message
#   Gets n messages from a queue.
#
#   Parameters:
#       vhost - Vhost where queue exists.  Defaults to default vhost set in
#           constructor.
#       queue - Name of queue to get messages from.
#       requeue - If 'true', Rabbit will keep the downloaded messages in queue.
#           If 'false' (default), n messages will be removed from queue.
#       count - Number of messages to download. Defaults to 1.
#       truncate - Don't remember what this does.
#
#   Returns:
#       List of downloaded messages.
#===============================================================================
sub get_message {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $queue = &ParamsMgr::get_arg(
        'arg'      => "queue",
        'required' => TRUE,
    );
    my $requeue = &ParamsMgr::get_arg(
        'arg'      => "requeue",
        'required' => FALSE,
        'default'  => 'false'
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1
    );
    my $truncate = &ParamsMgr::get_arg(
        'arg'      => "truncate",
        'required' => FALSE,
        'default'  => TRUNCATE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $coder = JSON::XS->new->allow_nonref;
    my $out = $self->_http_POST(
        url => $self->{http_methods}->{queues} . "/"
               . $vhost . "/" . $queue . "/get",
        content => JSON::XS->new->utf8->encode(
            {
                count => $count,
                requeue => $requeue,
                encoding => "auto",
                truncate => $truncate,
            }
        ),
    );
    if ($out->{rc} ne "200") {
        $Log->comment("Failed getting message from $queue: " . $out->{msg});
        NATE::BaseException->throw(
            "Could not get message from  $queue: ". $out->{msg}
        );
        return wantarray ? (): [];
    }

    my @payloads = ();
    local  $Data::Dumper::Indent = 0;
    foreach my $message (@{$out->{content}}) {
        if ($message->{payload} ne "") {
            try {
                my $decoded_message = $coder->decode($message->{payload});
                push (@payloads, $decoded_message);
            } otherwise {
                push(@payloads, $message->{payload});
            };
            $Log->debug("- Message received: " . Dumper($message));
        }
    }
    $Log->exit();

    return wantarray ? @payloads : \@payloads;
} # End of get_message

#===============================================================================
# Function: get_bindings
#   Get a list of bindings on the RabbitMQ cluster, a given vhost, or queue on
#   a vhost.
#
#   Parameters:
#       vhost - Name of vhost to get bindings on.
#       queue - Name of queeu to get bindings on.
#
#   Note!! Leaving both fields blank will return all bindings in cluster.
#
#   Returns:
#       A list of bindings.
#===============================================================================
sub get_bindings {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => ""
    );
    my $queue = &ParamsMgr::get_arg(
        'arg'      => "queue",
        'required' => FALSE,
        'default'  => ''
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $url = $self->{http_methods}->{bindings};
    if ($vhost ne "") {
        if ($queue ne '') {
            $url = $url . "/" . $vhost . "/" . $queue;
        } else {
            $url = $url . "/" . $vhost;
        }
    }

    my $bindings = $self->_http_GET(url => $url);
    if (($bindings->{msg} eq "OK") and ($bindings->{rc} eq "200")) {
        delete $bindings->{rc};
        delete $bindings->{msg};
    } else {
        $Log->comment("Failed to get bindings: " . $bindings->{msg});
        NATE::BaseException->throw(
            "Failed to get bindings: " . $bindings->{msg}
        );
        return wantarray ? () : [];
    }
    $Log->exit();

    return wantarray ? @{$bindings->{content}} : $bindings->{content};
} # End of get_bindings

#===============================================================================
# Function: get_vhosts
#   Gets a list of vhosts
#
#   Returns:
#       List of vhosts in RabbitMQ cluster.
#===============================================================================
sub get_vhosts {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => ""
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $url = ($vhost ne "") ? $self->{http_methods}->{vhosts} . "/"
                        . $vhost : $self->{http_methods}->{vhosts};
    my $vhosts = $self->_http_GET(url => $url);
    if (($vhosts->{msg} eq "OK") and ($vhosts->{rc} eq "200")) {
        delete $vhosts->{rc};
        delete $vhosts->{msg};
    } else {
        $Log->comment("Failed to get vhosts: " . $vhosts->{msg});
        NATE::BaseException->throw("Failed to get vhosts: " . $vhosts->{msg});
        return wantarray ? () : [];
    }
    $Log->exit();

    return wantarray ? @{$vhosts->{content}} : $vhosts->{content};
} # End of get_vhosts

#===============================================================================
# Function: vhost_exists
#   Check with RabbitMQ to see if specified vhost exists.
#
#   Parameters:
#       vhost - Name of vhost.  Defaults to default vhost set in constructor.
#
#   Returns:
#       TRUE if a vhost exists, FALSE otherwise
#===============================================================================
sub vhost_exists {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => $self->{mode},
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $errors = 0;
    if ($mode eq "cli") {
        my $cmd = "-c vhost_show -v $vhost";
        my $response = $self->_execute(
            cmd => $cmd
        );
        unless ($self->_decode_cli_response(
            response => $response, return_boolean => TRUE)) {
            $errors++;
        }
    } elsif ($mode eq "http") {
        my $url = $self->{http_methods}->{vhosts} . "/" . $vhost;
        $Log->comment("$url");
        my $response = $self->_http_GET(url => $url, user => $self->{user});
        if (($response->{'msg'} =~ /Object Not Found/i) or
            ($response->{rc} eq "404")) {
            $errors++;
        }
    }
    $Log->exit();

    if ($errors > 0) {
        $Log->debug("Vhost '$vhost' does not exist.");
        return FALSE;
    } else {
        $Log->debug("Vhost '$vhost' exists.");
        return TRUE;
    }
} # End of vhost_exists

#===============================================================================
# Function: queue_exists
#   Check with RabbitMQ to see if specified queue exists.
#
#   Parameters:
#       vhost - Name of vhost.  Defaults to default vhost set in constructor.
#       queue - Name of queue.
#
#   Returns:
#       TRUE if the queue exists, FALSE otherwise
#===============================================================================
sub queue_exists {
    my ($self, %args) = @_;

    $Log->enter();
    my @queues = ();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $queue = &ParamsMgr::get_arg(
        'arg'      => "queue",
        'required' => TRUE,
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => $self->{mode},
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $errors = 0;
    if ($mode eq "cli") {
        my $cmd = "-c queue_show -v $vhost -q $queue";
        my $response = $self->_execute(
            cmd => $cmd
        );
        unless ($self->_decode_cli_response(
            response => $response, return_boolean => TRUE)) {
            $errors++;
        }
    } elsif ($mode eq "http") {
        try {
            @queues = $self->get_queues (
                vhost => $vhost,
            );
        } otherwise {
            $Log->comment("Vhost '$vhost' does not exist.");
            $errors++;
        };

        my $found = FALSE;
        foreach my $temp_queue (@queues) {
            if ($temp_queue->{name} eq $queue) {
                $found = TRUE;
                last;
            }
        }
        unless ($found) {
            $errors++;
        }
    }
    $Log->exit();

    if ($errors > 0) {
        $Log->debug("Queue '$queue' does not exist.");
        return FALSE;
    } else {
        $Log->debug("Queue '$queue' exists.");
        return TRUE;
    }
} # End of queue_exists

#===============================================================================
# Function: exchange_exists
#   Check with RabbitMQ to see if specified exchange exists.
#
#   Parameters:
#       vhost - Name of vhost.  Defaults to default vhost set in constructor.
#       exchange - Name of exchange.
#
#   Returns:
#       TRUE if the exchange exists, FALSE otherwise
#===============================================================================
sub exchange_exists {
    my ($self, %args) = @_;

    my @exchanges = ();
    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => $self->{vhost}
    );
    my $exchange = &ParamsMgr::get_arg(
        'arg'      => "exchange",
        'required' => TRUE,
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => $self->{mode},
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $errors = 0;
    if ($mode eq "cli") {
        my $cmd = "-c exchange_show -v $vhost -e $exchange";
        my $response = $self->_execute(
            cmd => $cmd
        );
        unless ($self->_decode_cli_response(
            response => $response, return_boolean => TRUE)) {
            $errors++;
        }
    } elsif ($mode eq "http") {
        try {
            @exchanges = $self->get_exchanges (
                vhost => $vhost,
            );
        } otherwise {
            $Log->comment("Vhost '$vhost' does not exist.");
            $errors++;
        };

        my $found = FALSE;
        foreach my $temp_exchange (@exchanges) {
            if ($temp_exchange->{name} eq $exchange) {
                $found = TRUE;
                last;
            }
        }
        unless ($found) {
            $errors++;
        }
    }
    $Log->exit();

    if ($errors > 0) {
        $Log->debug("Exchange '$exchange' does not exist.");
        return FALSE;
    } else {
        $Log->debug("Exchange '$exchange' exists.");
        return TRUE;
    }
} # End of exchange_exists

#===============================================================================
# Function: get_users
#   Get a list of all users in RabbitMQ cluster.
#
#   Returns:
#       A list of all users.
#===============================================================================
sub get_users {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $user = &ParamsMgr::get_arg(
        'arg'      => "user",
        'required' => FALSE,
        'default'  => ''
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $url = ($user ne '') ? $self->{http_methods}->{users} . "/"
                            . $user : $self->{http_methods}->{users};
    my $users = $self->_http_GET(url => $url);
    if (($users->{msg} eq "OK") and ($users->{rc} eq "200")) {
        delete $users->{rc};
        delete $users->{msg};
    } else {
        $Log->comment("Failed to get users: " . $users->{msg});
        NATE::BaseException->throw("Failed to get users: " . $users->{msg});
        return wantarray ? () : [];
    }
    $Log->exit();

    return wantarray ? @{$users->{content}} : $users->{content};
} # End of get_users

#===============================================================================
# Function: create_user
#   Creates a user for the message broker
#
#   Parameters:
#       user - Name of user.
#       password - Password for new user.
#       role - Array containing role information for the new user.
#
#   Returns:
#       TRUE if successful, FALSE otherwise
#===============================================================================
sub create_user {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $user = &ParamsMgr::get_arg(
        'arg'      => "user",
        'required' => TRUE,
    );
    my $password = &ParamsMgr::get_arg(
        'arg'      => "password",
        'required' => TRUE,
    );
    my @role = &ParamsMgr::get_arg(
        'arg'      => "role",
        'required' => FALSE,
        'default'  => ()
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $url = $self->{http_methods}->{users} . "/" . $user;
    my $out = $self->_http_PUT(
        url => $url,
        content => JSON::XS->new->utf8->encode(
            {password => $password, tags => @role})
    );

    if ($out->{rc} eq "204"){
        return TRUE;
    }
    $Log->exit();

    return FALSE;
} # End of create_user

#===============================================================================
# Function: delete_user
#   Deletes a user of the Message Broker.
#
#   Parameters:
#       user - Name of user to delete.
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub delete_user {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $user = &ParamsMgr::get_arg(
        'arg'      => "user",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $out = $self->_http_DELETE(
        url => $self->{http_methods}->{users} . "/" . $user
    );
    if ($out->{rc} ne '204') {
        $Log->comment($out->{msg});
        return FALSE;
    }
    $Log->exit();

    return TRUE;
} # End of delete_user

#===============================================================================
# Function: set_permission
#   Sets permissions for user on vhost
#
#   Parameters:
#       user - Name of user.
#       password - Password for user.
#       vhost - Name of vhost to set permissions on.
#       read - Set read permission 'TRUE' (default) | 'FALSE'
#       write - Set write permission 'TRUE' (default) | 'FALSE'
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub set_permission {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $user = &ParamsMgr::get_arg(
        'arg'      => "user",
        'required' => TRUE,
    );
    my $password = &ParamsMgr::get_arg(
        'arg'      => "password",
        'required' => TRUE,
    );
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => TRUE,
    );
    my $configure = &ParamsMgr::get_arg(
        'arg'      => "configure",
        'required' => FALSE,
        'default'  => ""
    );
    my $read = &ParamsMgr::get_arg(
        'arg'      => "read",
        'required' => TRUE,
    );
    my $write = &ParamsMgr::get_arg(
        'arg'      => "write",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $out = $self->_http_PUT(
        url => $self->{http_methods}->{permissions} . "/"
                . $vhost . "/" . $user,
        content => JSON::XS->new->utf8->encode({
                configure => $configure,
                read      => $read,
                write     => $write
            }
        ),
    );
    if ($out->{rc} ne "204") {
        return FALSE;
    }
    $Log->exit();

    return TRUE;
} # End of set_permission

#===============================================================================
#  PRIVATE METHODS
#===============================================================================
#===============================================================================
# Description: Executes a command on the broker
#===============================================================================
sub _execute {
    my ($self, %args) = @_;

    $Log->enter();
    my @params = ();
    my $out = '';
    my $log = TRUE;

    unless (defined ($args{cmd})) {
        $Log->comment("Mandatory parameter 'cmd' is missing");
        $Log->exit();
        return $out;
    }

    # Build parameters for CLI broker
    my $cmd = $self->{python} . $self->{broker} . " ";
    $cmd .= $args{cmd};

    if (defined($args{log})) {
        $log = delete ($args{'log'});
    }

    print "+" x 124 . "\n" if $log;
    print "- SAN FIT's Message Broker\n" if $log;
    print "- CLI: $cmd\n";
    print "- Output: ";
    $out = qx($cmd);
    print $out;
    print "+" x 124 . "\n" if $log;
    $Log->exit();

    return $out;
} # End of _execute

#===============================================================================
# Description: Cleans CLI output and prepares the string for JSON decoding
#===============================================================================
sub _decode_cli_response {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $response = &ParamsMgr::get_arg(
        'arg'      => "response",
        'required' => TRUE,
    );
    my $return_boolean = &ParamsMgr::get_arg(
        'arg'      => "return_boolean",
        'required' => FALSE,
        'default'  => FALSE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        if ($return_boolean) {
            return FALSE;
        } else {
            return {};
        }
    }

    my $json = JSON->new->utf8;
    #  Get rid of an annoying extra character that contributed to my extra hours
    chomp($response);
    # Get rid of the enclosing quotes we get for mixing Python and Perl
    my $clean_response = substr $response, 1, -1;
    # Get rid of forward splashes so that we can decode the JSON string
    $clean_response =~ s/\\//g;
    # Decode the JSON string
    my $decoded_response = $json->decode($clean_response);
    $Log->debug(Dumper($decoded_response));

    if ($return_boolean) {
        if ($decoded_response->{status} =~ /error/i) {
            $Log->comment("The CLI command failed! Please read the logs");
            return FALSE;
        } elsif ($decoded_response->{status} =~ /success/i) {
            return TRUE;
        } else {
            $Log->comment("The response contained an unknown status!");
            return FALSE;
        }
    }
    $Log->exit();

    return $decoded_response;
} # End of _decode_cli_response

#===============================================================================
# Description: Executes an HTTP GET to the message broker
#===============================================================================
sub _http_GET {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $user = &ParamsMgr::get_arg(
        'arg'      => "user",
        'required' => FALSE,
        'default'  => USER
    );
    my $password = &ParamsMgr::get_arg(
        'arg'      => "password",
        'required' => FALSE,
        'default'  => PASSWORD
    );

    # Set user agent details and authentication to the broker
    my $user_agent = LWP::UserAgent->new();
    $user_agent->agent(AGENT);
    $user_agent->credentials(
        $url, 'Misfit Authenticated Data', $user, $password
    );

    # Set HTTP request details and authentication
    my $request = HTTP::Request->new();
    $request->method('GET');
    $request->uri($url);
    $request->header('Content-Type' => 'application/json');
    $request->authorization_basic($user, $password);

    # Make request to broker for content
    my $response = $user_agent->request($request);
    $Log->debug($response->as_string);

    # Parse returned JSON string and return as hash reference
    my $content_response = "{}";
    if ($response->content ne '' and !($response->content =~ /</)) {
        $content_response = $response->content;
    }
    if ($response->{_rc} eq '404') {
        return {
            msg     => "One of the parameters posted to the Message Broker " .
                       "is incorrect or missing. Make sure queues, exchanges, " .
                       "vhosts used in the URL exist and are valid: \n $url",
            rc      => $response->{_rc},
            content => {}
        };
    } elsif ($response->{_rc} eq '501') {
        return {
            msg     => "The server does not recognize the request method. " .
                       "Make sure the RabbitMQ address is correct: $url",
            rc      => $response->{_rc},
            content => {}
        };
    }
    my $out = {};
    try {
        $out = {
            content => JSON::XS->new->utf8->decode($content_response),
            msg     => $response->{_msg},
            rc      => $response->{_rc},
        };
    } otherwise {
        my $e = shift;
        $Log->debug("http GET exception: " . $e->text());
        $out = {
            content => {},
            rc      => '500', # Server error
            msg     => $e->text(),
        };
        $Log->error(
            "Failed to decode the response from the broker: " .
            Dumper($content_response)
        );
    };
    $Log->exit();

    return $out;
} # End of _http_GET

#===============================================================================
# Description: Executes an HTTP PUT to the message broker
#===============================================================================
sub _http_PUT {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $user = &ParamsMgr::get_arg(
        'arg'      => "user",
        'required' => FALSE,
        'default'  => USER
    );
    my $password = &ParamsMgr::get_arg(
        'arg'      => "password",
        'required' => FALSE,
        'default'  => PASSWORD
    );
    my $content = &ParamsMgr::get_arg(
        'arg'      => "content",
        'required' => FALSE,
        'default'  => ''
    );

    my $user_agent = LWP::UserAgent->new();
    $user_agent->agent(AGENT);
    $user_agent->credentials(
        $url, 'Misfit Authenticated Data', $user, $password
    );

    # Set HTTP request details and authentication
    my $request = HTTP::Request->new();
    $request->method('PUT');
    $request->uri($url);
    $request->header('Content-Type' => 'application/json');
    $request->content($content);
    $request->authorization_basic($user, $password);

    # Make request to broker for content
    my $response = $user_agent->request($request);
    $Log->debug($response->as_string);

    # Parse returned JSON string and return as hash reference
    my $content_response = "{}";
    if (exists($response->{_content}) && $response->{'_content'} ne '') {
        $content_response = $response->content;
    }
    if ($response->{_rc} eq '404') {
        return {
            msg     => "One of the parameters posted to the Message Broker " .
                       "is incorrect or missing. Make sure queues, exchanges, " .
                       "vhosts used in the URL exist and are valid: \n $url",
            rc      => $response->{_rc},
            content => {}
        };
    } elsif ($response->{_rc} eq '501') {
        return {
            msg     => "The server does not recognize the request method. " .
                       "Make sure the RabbitMQ address is correct: $url",
            rc      => $response->{_rc},
            content => {}
        };
    }

    my $out = {};
    try {
        $out = {
            content => JSON::XS->new->utf8->decode($content_response),
            msg     => $response->{_msg},
            rc      => $response->{_rc},
        };
    } otherwise {
        my $e = shift;
        $Log->debug("http PUT exception: " . $e->text());
        $out = {
            content => {},
            rc      => '500', # Server error
            msg     => $e->text(),
        };
        $Log->error(
            "Failed to decode the response from the broker: " .
            Dumper($content_response)
        );
    };
    $Log->exit();

    return $out;
} # End of _http_PUT

#===============================================================================
# Description: Executes an HTTP DELETE to the message broker
#===============================================================================
sub _http_DELETE {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $user = &ParamsMgr::get_arg(
        'arg'      => "user",
        'required' => FALSE,
        'default'  => $self->{user}
    );
    my $password = &ParamsMgr::get_arg(
        'arg'      => "password",
        'required' => FALSE,
        'default'  => $self->{password}
    );

    # Set user agent details and authentication to the broker
    my $user_agent = LWP::UserAgent->new();
    $user_agent->agent(AGENT);
    $user_agent->credentials(
        $url, 'Misfit Authenticated Data', $user, $password
    );

    # Set HTTP request details and authentication
    my $request = HTTP::Request->new(DELETE => $url);
    $request->authorization_basic($user, $password);

    # Make request to broker for content
    my $response = $user_agent->request($request);
    $Log->debug($response->as_string);

    # Parse returned JSON string and return as hash reference
    my $content_response = "{}";
    if ($response->content ne "" and !($response->content =~ /</)) {
        $content_response = $response->content;
    }
    if ($response->{_rc} eq '404') {
        return {
            msg     => "One of the parameters posted to the Message Broker " .
                       "is incorrect or missing. Make sure queues, exchanges, " .
                       "vhosts used in the URL exist and are valid: \n $url",
            rc      => $response->{_rc},
            content => {}
        };
    } elsif ($response->{_rc} eq '501') {
        return {
            msg     => "The server does not recognize the request method. " .
                       "Make sure the RabbitMQ address is correct: $url",
            rc      => $response->{_rc},
            content => {}
        };
    }

    my $out = {};
    try {
        $out = {
            content => JSON::XS->new->utf8->decode($content_response),
            msg     => $response->{_msg},
            rc      => $response->{_rc},
        };
    } otherwise {
        my $e = shift;
        $Log->debug("http DEL exception: " . $e->text());
        $out = {
            content => {},
            rc      => '500', # Server error
            msg     => $e->text(),
        };
        $Log->error(
            "Failed to decode the response from the broker: " .
            Dumper($content_response)
        );
    };
    $Log->exit();

    return $out;
} # End of _http_DELETE

#===============================================================================
# Description: Executes an HTTP POST to the message broker
#===============================================================================
sub _http_POST {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $user = &ParamsMgr::get_arg(
        'arg'      => "user",
        'required' => FALSE,
        'default'  => $self->{user}
    );
    my $password = &ParamsMgr::get_arg(
        'arg'      => "password",
        'required' => FALSE,
        'default'  => $self->{password}
    );
    my $content = &ParamsMgr::get_arg(
        'arg'      => "content",
        'required' => FALSE,
        'default'  => ''
    );

    my $user_agent = LWP::UserAgent->new();
    $user_agent->agent(AGENT);
    $user_agent->credentials(
        $url, 'Misfit Authenticated Data', $user, $password
    );

    # Set HTTP request details and authentication
    my $request = HTTP::Request->new();
    $request->method('POST');
    $request->uri($url);
    $request->header('Content-Type' => 'application/json');
    $request->content($content);
    $request->authorization_basic($user, $password);

    # Make request to broker for content
    my $response = $user_agent->request($request);
    $Log->debug($response->as_string);

    # Parse returned JSON string and return as hash reference
    my $content_response = "{}";
    if (exists($response->{_content}) and $response->{'_content'} ne '') {
        $content_response = $response->content;
    }

    # Avoid passing a malformed JSON string when the content is not found
    # this will happen if a user tries to publish a message to an exchange that
    # does not exist (among many other combinations)
    if ($response->{_rc} eq '404') {
        return {
            msg     => "One of the parameters posted to the Message Broker " .
                       "is incorrect or missing. Make sure queues, exchanges, " .
                       "vhosts used in the URL exist and are valid: \n $url",
            rc      => $response->{_rc},
            content => {}
        };
    } elsif ($response->{_rc} eq '501') {
        return {
            msg     => "The server does not recognize the request method. " .
                       "Make sure the RabbitMQ address is correct: $url",
            rc      => $response->{_rc},
            content => {}
        };
    }

    my $out = {
        content => JSON::XS->new->utf8->decode($content_response),
        msg     => $response->{_msg},
        rc      => $response->{_rc},
    };
    $Log->exit();

    return $out;
} # End of _http_POST

#===============================================================================
# Description: Detect if the os this code is running on is windows or not.
#===============================================================================
sub _os_is_windows {
    $Log->debug("Detected OS: $OSNAME");
    return $OSNAME eq WINDOWS;
} # End of _os_is_windows

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://misfitapi.gdl.englab.netapp.com/files/Core/MessageBroker-pm.html

=cut

1;
