# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Cron-tab style scheduler for Misfit
## @author dl-san-fit@netapp.com
## @status Private

package San::FIT::Misfit::Core::Scheduler;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use ParamsMgr;
use Time::ParseDate;
use NATE::Process;
use NATE::Util::DataTypes qw(TRUE FALSE);

#===============================================================================
# Description: Scheduler constructor
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;

    &ParamsMgr::process_args(%args);
    $self->{schedule} = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return undef;
    }

    $args{background} = TRUE;
    delete $args{schedule};
    $self->{process} = NATE::Process->new(%args);
    $Log->exit();

    return $self;
} # End of new

#===============================================================================
# Description: Starts executing the scheduler
#===============================================================================
sub start {
    my ($self, %args) = @_;

    if ($self->{schedule} eq '-') {
        return NATE::Result::Pass->new('Done');
    }

    $Log->comment('SCHEDULE IS:  ' . $self->{schedule});
    while ($self->{schedule} ne '-') {
        my $time = $self->_calculate_next($self->{schedule});
        my $now = time;
        my $sleep = 0;
        if ($time < $now) {
            next;
        } else {
            $sleep = $time - $now;
        }
        $Log->comment('Next run will happen at ' . scalar(localtime($time)));
        $Log->comment('Sleeping for: ' . $sleep);
        sleep($sleep);
        $self->{process}->start();
        $self->{process}->wait();
        last if ($self->_calculate_next($self->{schedule} eq ''));
    }
    $self->{process}->stop();
    $self->{process}->destroy();

    return TRUE;
}

#===============================================================================
# Description: Stops the scheduler
#===============================================================================
sub stop {
    my ($self, %args) = @_;

    $Log->comment('######## \n Received stop message \n ###########');
    $self->{schedule} = '-';
    $self->{process}->stop();
    $self->{process}->destroy();
    return TRUE;
}

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#==============================================================================
# Description: Clears schedule stopping all scheduled jobs
#==============================================================================
sub clear_schedule {
    my ($self, %args) = @_;

    $Log->enter();
    $self->{schedule} = '';
    $Log->exit();

    return TRUE;
} # End of clear_schedule

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Calculate next scheduled time
#===============================================================================
sub _calculate_next {
    my ($self, $schedule, $time) = @_;

    # Locals
    my @expanded = ();
    my $w = '';

    # Ranges (minutes, hours, days, months, day of week, seconds)
    my @ranges = ([0,59], [0,23], [0,31], [0,12], [0,7], [0,60]);
    my @date_conversions = ({}, {}, {},
        {qw(jan 1 feb 2 mar 3 apr 4 may 5 jun 6 jul 7 aug 8 sep 9 oct 10 nov 11 dec 12)},
        {qw(sun 0 mon 1 tue 2 wed 3 thu 4 fri 5 sat 6)}, {}
    );

    my @low_mappings = ({}, {}, {0 => 1}, {0 => 1}, {7 => 0}, {},);
    my $entry = [split /\s+/, $schedule];

    # Make sure the syntax is correct (5 or 6 columns only)
    if (scalar(@$entry) != 4 && scalar(@$entry) != 5) {
        return '';
    }

    for my $i (0..scalar(@$entry)) {
        my @e = split /,/, $entry->[$i];
        my @res;
        my $t;

        while (defined($t = shift @e)) {
            # Change the string from "*/5" to "0-59/5"
            $t =~ s|^\*(/.+)$|$ranges[$i][0]."-".$ranges[$i][1].$1|e;
            if ($t =~ m|^([^-]+)-([^-/]+)(/(.*))?$|) {
                my ($low, $high, $step) = ($1, $2, $4);
                $step = 1 unless $step;
                if ($low !~ /^(\d+)/) {
                    $low = $date_conversions[$i]{lc $low};
                }
                if ($high !~ /^(\d+)/) {
                    $high = $date_conversions[$i]{lc $high};
                }
                if (!defined($low) || !defined($high) ||
                    $low > $high || $step !~ /^\d+$/) {
                    return '';
                }
                my $j;
                for ($j = $low; $j <= $high; $j += $step) {
                    push (@e, $j);
                }
            } else {
                $t = $date_conversions[$i]{lc $t} if $t !~ /^(\d+|\*)$/;
                $t = $low_mappings[$i]{$t} if exists($low_mappings[$i]{$t});
                if (!defined($t) || ($t ne '*' &&
                                    ($t < $ranges[$i][0] ||
                                     $t > $ranges[$i][1]))) {
                    return '';
                }
                push (@res, $t);
            }
        }
        push (@expanded, [sort {int($a) <=> int($b)} @res]);
    }

    # Calculate the time and make your head explode
    my $now = $time || time;
    if ($expanded[2]->[0] ne '*' && $expanded[4]->[0] ne '*') {
        my @bak = @{$expanded[4]};
        $expanded[4] = ['*'];
        my $t1 = $self->_calculate($now, \@expanded);
        $expanded[4] = \@bak;
        $expanded[2] = ['*'];
        my $t2 = $self->_calculate($now, \@expanded);
        return $t1 < $t2 ? $t1 : $t2;
    } else {
        return $self->_calculate($now, \@expanded);
    }
} # End of _calculate_next

#===============================================================================
# Description: Calculate execution time
#===============================================================================
sub _calculate {
    my ($self, $now, $expanded) = @_;

    my @days = qw(Sunday Monday Tuesday Wednesday Thursday Friday Saturday Sunday);
    my $offset = ($expanded->[5] ? 1 : 60);

    # Get current time
    my ($now_sec, $now_min, $now_hour, $now_mday, $now_mon, $now_wday, $now_year) =
        (localtime($now+$offset))[0, 1, 2, 3, 4, 6, 5];
    $now_mon++;
    $now_year += 1900;

    my ($dest_mon, $dest_mday, $dest_wday, $dest_hour, $dest_min, $dest_sec, $dest_year) =
        ($now_mon, $now_mday, $now_wday, $now_hour, $now_min, $now_sec, $now_year);

    while ($dest_year <= $now_year + 1) {
        # Month:
        if ($expanded->[3]->[0] ne '*') {
            unless (defined ($dest_mon = $self->_get_next($dest_mon, $expanded->[3]))) {
                $dest_mon = $expanded->[3]->[0];
                $dest_year++;
            }
        }
        # Day (Month):
        if ($expanded->[2]->[0] ne '*') {
            if ($dest_mon != $now_mon) {
                $dest_mday = $expanded->[2]->[0];
            } else {
                unless (defined ($dest_mday = $self->_get_next($dest_mday, $expanded->[2]))) {
                    $dest_mday = $expanded->[2]->[0];
                    $dest_mon++;
                    if ($dest_mon > 12) {
                        $dest_mon = 1;
                        $dest_year++;
                    }
                    next;
                }
            }
        } else {
            $dest_mday = ($dest_mon == $now_mon ? $dest_mday : 1);
        }
        # Day (week):
        if ($expanded->[4]->[0] ne '*') {
            $dest_wday = $self->_get_next($dest_wday, $expanded->[4]);
            $dest_wday = $expanded->[4]->[0] unless $dest_wday;

            my ($mon, $mday, $year);
            $dest_mday = 1 if $dest_mon != $now_mon;
            my $t = parsedate(sprintf("%4.4d/%2.2d/%2.2d",$dest_year, $dest_mon, $dest_mday));
            ($mon, $mday, $year) = (localtime(
                parsedate("$days[$dest_wday]", PREFER_FUTURE => 1, NOW => $t-1))
            )[4,3,5];
            $mon++;
            $year += 1900;

            if ($mon != $dest_mon || $year != $dest_year) {
                $dest_mon = $mon;
                $dest_year = $year;
                $dest_mday = 1;
                $dest_wday = (localtime(
                    parsedate(
                        sprintf("%4.4d/%2.2d/%2.2d", $dest_year, $dest_mon, $dest_mday)
                    ))
                )[6];
                next;
            }
            $dest_mday = $mday;
        } else {
            unless ($dest_mday) {
                $dest_mday = ($dest_mon == $now_mon ? $dest_mday : 1);
            }
        }
        # Hour:
        if ($expanded->[1]->[0] ne '*') {
            if ($dest_mday != $now_mday) {
                $dest_hour = $expanded->[1]->[0];
            }
            else {
                unless (defined ($dest_hour = $self->_get_next($dest_hour, $expanded->[1]))) {
                    $dest_hour = $expanded->[1]->[0];
                    my $t = parsedate(
                        sprintf("%2.2d:%2.2d:%2.2d %4.4d/%2.2d/%2.2d",
                                $dest_hour, $dest_min, $dest_sec, $dest_year,
                                $dest_mon, $dest_mday
                        )
                    );
                    ($dest_mday, $dest_mon, $dest_year, $dest_wday) = (localtime(
                        parsedate("+ 1 day", NOW => $t)))[3, 4, 5, 6];
                    $dest_mon++;
                    $dest_year += 1900;
                    next;
                }
            }
        } else {
            $dest_hour = ($dest_mday == $now_mday ? $dest_hour : 0);
        }
        # Minute
        if ($expanded->[0]->[0] ne '*') {
            if ($dest_hour != $now_hour) {
                $dest_min = $expanded->[0]->[0];
            } else {
                unless (defined ($dest_min = $self->_get_next($dest_min, $expanded->[0]))) {
                    $dest_min = $expanded->[0]->[0];
                    my $t = parsedate(
                        sprintf("%2.2d:%2.2d:%2.2d %4.4d/%2.2d/%2.2d",
                                $dest_hour, $dest_min, $dest_sec,
                                $dest_year, $dest_mon, $dest_mday
                        )
                    );
                    ($dest_hour,$dest_mday,$dest_mon,$dest_year,$dest_wday) =
                      (localtime(parsedate(" + 1 hour", NOW => $t)))  [2, 3, 4, 5, 6];
                    $dest_mon++;
                    $dest_year += 1900;
                    next;
                }
            }
        } else {
            $dest_min = ($dest_hour == $now_hour ? $dest_min : 0);
        }
        # Seconds
        if ($expanded->[5]) {
            if ($expanded->[5]->[0] ne '*') {
                if ($dest_min != $now_min) {
                    $dest_sec = $expanded->[5]->[0];
                } else {
                    unless (defined ($dest_sec = $self->_get_next($dest_sec, $expanded->[5]))) {
                        $dest_sec = $expanded->[5]->[0];
                        my $t = parsedate(
                            sprintf("%2.2d:%2.2d:%2.2d %4.4d/%2.2d/%2.2d",
                                    $dest_hour,$dest_min,$dest_sec,
                                    $dest_year,$dest_mon,$dest_mday
                            )
                        );
                        ($dest_min, $dest_hour, $dest_mday, $dest_mon,
                         $dest_year, $dest_wday) = (localtime(
                            parsedate(" + 1 minute", NOW => $t))
                         )[1, 2, 3, 4, 5, 6];
                        $dest_mon++;
                        $dest_year += 1900;
                        next;
                    }
                }
            } else {
                $dest_sec = ($dest_min == $now_min ? $dest_sec : 0);
            }
        } else {
            $dest_sec = 0;
        }

        my $date = sprintf(
            "%2.2d:%2.2d:%2.2d %4.4d/%2.2d/%2.2d",
            $dest_hour, $dest_min, $dest_sec, $dest_year, $dest_mon, $dest_mday
        );
        return parsedate($date);
    }
} # End _calculate

#===============================================================================
# Description: Get next job to execute based on the schedule.
#===============================================================================
sub _get_next {
    my ($self, $current, $to_check) = @_;

    foreach my $i (0 .. $#$to_check) {
        if ($$to_check[$i] >= $current) {
            return $$to_check[$i] ;
        }
    }

    return undef;
} # End of _get_next

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/core/scheduler>

=cut

1;