# Copyright (c) 2001-2013 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Utility module for all Misfit blocks.
## @author dl-san-fit@netapp.com
## @status Private - For SAN FIT use only

# Package: San::FIT::Misfit::Core::Util
# A set of utility methods to interact with MISFIT. Provides shortcuts to
# common asynchronous messaging operations.

package San::FIT::Misfit::Core::Util;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NATE::Process;
use NATE::ParamSet;
use File::Basename;
use NATE::ProcManager;
use POSIX qw/strftime/;
use San::FIT::Misfit::Services::IO;
use San::FIT::Misfit::Core::Workflow;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::MessageBroker;

use constant {
    URL                            => "http://sanfit.gdl.englab.netapp.com",
    BYE                            => "__bye__",
    PORT                           => 55672,
    DELAY                          => 60,
    COLUMN                         => 100,
    MIN_WAIT                       => 3600,
    TIMEOUT                        => 14400,
    IO_DURATION                    => 300,
    SHORT_TIMEOUT                  => 5,
    SIXTY_SECONDS                  => 60,
    DEFAULT_IO_TOOL                => "dt",
    DEFAULT_P4_PATH                => "/usr/software/test/noarch/agnostic",
    INTERACTION_BYE                => "interaction.finished",
    DEFAULT_TRY_LIMIT              => 10,
    QUEUE_SUFFIX_LIMIT             => 10,
    MAX_MESSAGES_PER_QUEUE         => 10000,
    INTERACTIONS_RUNNING_WAIT      => 5,
    DEFAULT_INTERACTION_ITERATIONS => 1,
};

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Function: use_misfit
#   Initializes all the necessary resources to use Misfit. Creates default
#   queues and exchanges required to interact with a Message Broker and other
#   scripts.
#
#   Parameters:
#      vhost (optional) - String representing the name of the vhost to create or
#          use for messaging. Defaults to the name of the testbed.
#      queues (optional) - Reference to an array containing the names of extra
#          queues to be created for the script.
#      url (optional) - String representing the url of the message broker server.
#      port (optional) - String representing the port number of the message broker
#          server. Defaults to the default port of RabbitMQ: 55672.
#      mode (optional) - "http" or "cli". Defaults to "http".
#      interactions (optional) - A list of feature interactions (see below)
#      reset_environment (optional) - Deletes any Misfit artifact from the broker
#          before initializing the environment. Defaults to FALSE (0)
#      use_resource_reservation_service (optional) - Tells the broker to use
#          SAN FIT's reservation system. Defaults to FALSE (0). Use it only
#          if you know what this is and how it works!
#
#      Syntax for interactions:
#      method - (Required) The callback triggered when a particular message is
#           received.
#      message - (Required) The message that the interaction will wait for
#           before starting.
#      sender - (Optional) The name of the Misfit module,thpl or an exchange
#           that will route the particular message the interaction is waiting
#           for. While it is optional, you MUST specify a queue if you do NOT
#           specify the sender.
#      queue - (Optional) Mutually exclusive with the 'sender' param.  Sender
#           has precedence over queue.  If queue is specified it is possible to
#           re-use an existing queue that is already created.  This is for
#           for advanced use only.  Having multiple proesses monitoring the
#           same queue has been known to cause issues. Only use if you know what
#           you are doing.
#      name - (Optional) This allows you to name the interaction.  It will
#           affect the name of the queue created as well as the runid for the
#           processes in the logs.  This is useful if you have different
#           interaction messages triggering the same method.
#      iterations - (Optional) The iterations parameter allows the user to
#           specify whether to run an interaction for a set number of times or
#           until a terminating message is received.  Iterations can be set in
#           two ways.
#           Default: 1 iteration.
#           To execute n times, specify an integer n > 0.
#           For continuous interaction mode, n = -1.
#      verbose - (Optional) Set to "false" or 0 to avoid logging messages other
#           than the expected one during a wait_for_message call. Default: "true"
#
#           > interactions => [
#           >     { method => 'my_method', message => 'hello', queue => 'queue'}
#           > ]
#           >
#           > interactions => [
#           >      {
#           >         method  => 'enable_efficiency',
#           >         message => 'vol.move.finished',
#           >         sender  => 'SAN::FIT::Misfit::Features::VolOps'
#           >      }
#           > ]
#           >
#           > interactions => [
#           >      {
#           >         method     => 'enable_efficiency',
#           >         message    => 'vol.move.finished',
#           >         sender     => 'SAN::FIT::Misfit::Features::VolOps',
#           >         iterations => -1, # Interact with every vol.move.finished
#           >         verbose    => 0,
#           >      }
#           > ]
#           >
#           > interactions => [
#           >      {
#           >         method     => 'enable_efficiency',
#           >         message    => 'vol.move.finished',
#           >         sender     => 'SAN::FIT::Misfit::Features::VolOps',
#           >         iterations => 5, # Interact the first 5 vol.move.finished
#           >      }
#           > ]
#           >
#           > interactions => [
#           >      {
#           >         method  => 'enable_efficiency',
#           >         message => 'vol.move.finished',
#           >         name    => 'enable_efficiency_on_vol_move',
#           >         sender  => 'SAN::FIT::Misfit::Features::VolOps'
#           >      },
#           >      {
#           >         method  => 'enable_efficiency',
#           >         message => 'vol.copy.finished',
#           >         name    => 'enable_efficiency_on_vol_copy',
#           >         sender  => 'SAN::FIT::Misfit::Features::VolOps'
#           >      }
#           > ]
#
#  Exchanges created by this method:
#    - errors          (fanout)
#    - critical        (fanout)
#    - features        (direct)
#    - interactions    (direct)
#    - script's name   (topic)
#
#  Queues created by this method:
#   -  script_name.errors bound to errors by '#'
#   -  script_name.critical bound to critical by '#'
#   -  script_name.features bound to features by '#'
#   -  script_name.interactions bound to interactions by '#'
#   -  script_name.io bound to San::FIT::Misfit::Services::IO by '#'
#   -  script_name.io bound to San::FIT::Misfit::Services::IO by '#'
#   -  script_name.reservation bound to resource by 'script_name'
#   -  script_name.feature1 bound to feature1 by '#'
#   -  ...
#   -  script_name.featureN bound to featureN by '#'
#
#  Returns:
#      ID. A string representing the name of the script using misfit
#
#  Example:
#
#      > San::FIT::Misfit::Core::Util::use_misfit(
#      >    interactions => \@Interactions,
#      >    reset_environment => 0,
#      > );
#===============================================================================
sub use_misfit {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => param("TESTBED", '-default', '/')
    );
    my $extra_queues = &ParamsMgr::get_arg(
        'arg'      => "queues",
        'required' => FALSE,
        'default'  => []
    );
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => PORT
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => "http"
    );
    my $interactions = &ParamsMgr::get_arg(
        'arg'      => "interactions",
        'required' => FALSE,
        'default'  => []
    );
    my $purge = &ParamsMgr::get_arg(
        'arg'      => "reset_environment",
        'required' => FALSE,
        'default'  => FALSE
    );
    my $use_resource_reservation_service = &ParamsMgr::get_arg(
        'arg'      => "use_resource_reservation_service",
        'required' => FALSE,
        'default'  => FALSE
    );

    # Name of the script using it.
    my $id = whoami();
    $id =~ s/\.[^.]+$//;

    # Create message broker
    my $misfit = San::FIT::Misfit::Core::MessageBroker->new(
        vhost => $vhost,
        url   => $url,
        port  => $port,
        mode  => $mode,
        purge => $purge,
    );

    # If the attempt to connect via CLI fails, try to do it through the HTTP API
    if ($misfit == {}) {
        if ($mode =~ /cli/i) {
            $mode = "http";
        } elsif ($mode =~ /http/i) {
            $mode = "cli";
        }
        $misfit = San::FIT::Misfit::Core::MessageBroker->new(
            vhost => $vhost,
            url   => $url,
            port  => $port,
            mode  => $mode,
            purge => $purge,
        );
        if ($misfit == {}) {
            $Log->comment("Something is wrong with the Rabbit Server!");
            return $id;
        }
    }

    # Get list of services and features used by this misfit user. These params
    # are generic params passed to independent thpls or stest files

    my $blocks = San::FIT::Misfit::Core::Util::_get_list_of_misfit_blocks(
        id => $id
    );

    # If the Reservation module is being used and specified as a parameter
    # to use_misfit then create reservation queues

    if ($use_resource_reservation_service) {
        my @reservation_queues =
            San::FIT::Misfit::Core::Util::_create_reservation_queues(
                id     => $id,
                blocks => $blocks,
        );
        push(@{$extra_queues}, @reservation_queues);
    }

    # Create a set of default exchanges and queues for the caller of use_misfit
    # These will be used to provide messaging and interaction capabilities.

    $misfit->create_exchange(
        exchanges => [
            { exchange => $id,            type => 'topic'  },
            { exchange => 'errors',       type => 'fanout' },
            { exchange => 'critical',     type => 'fanout' },
            { exchange => 'features',     type => 'direct' },
            { exchange => 'interactions', type => 'direct' },
        ],
    );

    my $queues = [
        {
            exchange    => 'errors',
            queue       => $id . '.errors',
            routing_key => '#'
        },
        {
            exchange    => 'critical',
            queue       => $id . '.critical',
            routing_key => '#'
        },
        {
            exchange    => 'features',
            queue       => 'features',
            routing_key => '#'
        },
        {
            exchange    => 'interactions',
            queue       => 'interactions',
            routing_key => '#'
        }
    ];

    # If the I/O module is being used then create a queue for notifications
    # such as io started, io finished

    if (grep (/^io$/i, @{$blocks})) {
        push(@{$queues}, {
            exchange    => 'San::FIT::Misfit::Services::IO',
            queue       => "$id.io",
            routing_key => "#",
        });
    }

    # If a user passes a list of interactions, extract the sender (exchange)
    # and build the queues needed to interact and communicate.

    if (scalar(@{$interactions})) {
        my $interaction_queues =
            San::FIT::Misfit::Core::Util::extract_interaction_info(
                id           => $id,
                interactions => $interactions
        );
        push(@{$extra_queues}, @{$interaction_queues->{queue_info}});
        $interactions = $interaction_queues->{interactions};
    }

    # Send a message to an exchange that will be used exclusively to track
    # what features or 'blocks' are still running on the system. This message
    # registers the feature with the scenario. Note that I/O, as a service,
    # needs to be treated differently and will not be tracked as a feature.

    unless ($id =~ /^io$|^San::FIT::Misfit::Services::IO$/i) {
        $misfit->publish_message(
            vhost       => $vhost,
            exchange    => 'features',
            immediate   => TRUE,
            routing_key => '#',
            message     => $id
        );
    }

    # Create any extra exchanges and queues specified by the user or discovered
    # by the scenario based on the list of misfit features/blocks/services

    my @exchanges = ();
    foreach my $extra_queue (@{$extra_queues}) {
        if ($misfit->queue_exists(queue => $extra_queue->{queue})) {
            next;
        }
        if ($misfit->exchange_exists(exchange => $extra_queue->{exchange})) {
            push(@{$queues}, $extra_queue);
        } else {
            $misfit->create_exchange(
                mode => 'cli',
                exchanges => [
                    { exchange => $extra_queue->{exchange}, type => 'topic' }
                ]
            );
            push(@{$queues}, $extra_queue);
        }
        push(@exchanges, $extra_queue->{exchange});
    }

    $misfit->create_queue(
        queues => $queues,
        mode   => "cli"
    );

    # Bind the queues to a 'critical' notification exchange. This means that
    # every queue in misfit will receive any critical error broadcasted by
    # any other feature in the scenario. Skip 'interactions' because that's a
    # special queue for tracking purposes only. Same thing for 'features'

    foreach my $queue (@{$queues}) {
        next if $queue->{queue} =~ /^features$|^interactions$|\.errors/i;
        $misfit->create_binding(
            queues => [{queue => $queue->{queue}, exchange => 'critical'}],
        );
    }

    # If users pass a list of interactions, enable them and store the
    # interaction handler so that it can be cleaned when users call
    # cleanup_misfit

    if (scalar(@{$interactions})) {
        $Log->comment("Enabling feature interaction");
        my $interaction_mgr = enable_interactions(
            interactions => $interactions,
        );
    }
    $Log->exit();

    return $id;
} # End of use_misfit

#===============================================================================
# Function: cleanup_misfit
#   Cleans up all Misfit resources initialized by the script
#
#   Parameters:
#       vhost (optional) -  String representing the name of the vhost to create or
#           use for messaging. Defaults to the name of the testbed.
#       queues (optional) -  Reference to an array containing the names of extra
#           queues to be created for the script.
#       url (optional) - String representing the url of the message broker server.
#       port (optional) - String representing the port number of the message broker
#           server. Defaults to the default port of RabbitMQ: 55672.
#       mode (optional) - "http" or "cli". Defaults to "http".
#       delete_vhost (optional) - Defaults to false. Delete the vhost corresponding
#           to the testbed.
#
#   Returns:
#       TRUE
#
#   Example:
#
#       > $status = San::FIT::Misfit::Core::Util::cleanup_misfit();
#===============================================================================
sub cleanup_misfit {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => param("TESTBED", '-default', '/')
    );
    my $extra_queues = &ParamsMgr::get_arg(
        'arg'      => "queues",
        'required' => FALSE,
        'default'  => []
    );
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => PORT
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => "http"
    );
    my $delete_vhost = &ParamsMgr::get_arg(
        'arg'      => "delete_vhost",
        'required' => FALSE,
        'default'  => FALSE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    # Create message broker
    my $misfit = San::FIT::Misfit::Core::MessageBroker->new(
        vhost => $vhost,
        url   => $url,
        port  => $port,
        mode  => $mode
    );

    # Name of the script using it.
    my $id = whoami();
    $id =~ s/\.[^.]+$//;

    my $queues = [
        {
            exchange    => 'errors',
            queue       => $id . '.errors',
            routing_key => '#'
        },
        {
            exchange    => 'critical',
            queue       => $id . '.critical',
            routing_key => '#'
        },
    ];
    if ($extra_queues != []) {
        push(@{$queues}, @{$extra_queues});
    }



    # Remove the feature/service ID from the interactions queue. This will
    # allow us to query this exchange later and determine if a feature
    # associated to an interaction is no longer running

    unless ($id =~ /^io$|^San::FIT::Misfit::Services::IO$/i) {
        San::FIT::Misfit::Core::Util::_wait_for_features_to_finish(
            broker => $misfit,
            id     => $id
        );
    }

    # Signal end of thpl
    $misfit->publish_message(
        exchange    => "resource",
        message     => {
            msg  => $id . '.finished',
        },
    );

    # If there is any interaction process currently running (not waiting),
    # wait for it to finish. Note that this could be a process spawned by a
    # completely different caller
    San::FIT::Misfit::Core::Util::_wait_for_interactions_to_finish(
        broker => $misfit
    );

    if (get_number_of_features_running() > 0) {
        # Wait for critical announcement telling all features are done
        $misfit->wait_for_message(
            message => BYE,
            queue   => $id . '.critical',
        );
    } else {
        $Log->comment(
            "No other features registered, sending interaction bye " .
            "to signal interactions to exit gracefully."
        );
        $misfit->publish_message(
            exchange    => 'critical',
            immediate   => TRUE,
            routing_key => '#',
            message     => INTERACTION_BYE,
        );
    }

    # Now that we have told all the interaction handlers to exit gracefully
    # give them time.  These processes are (waiting) for messages.
    San::FIT::Misfit::Core::Util::_wait_for_interaction_handlers_to_finish(
        broker => $misfit
    );

    # Delete queues created by use_misfit or passed by the caller
    $misfit->delete_queue(queues => $queues);

    # Delete the main exchange created for this caller
    $misfit->delete_exchange(
        exchanges => [
            { exchange => $id, type => 'topic' },
        ],
    );

    # If users want to completely clean the testbed broker space, delete vhost
    if ($delete_vhost) {
        $misfit->delete_vhost();
    }
    $Log->exit();

    return TRUE;
} # End of cleanup_misfit

#===============================================================================
# Function: wait_for_message
#   Waits or blocks until a message arrives to a queue.
#
#   Parameters:
#       queue - Name of queue to digest messages from.
#       message - String to match in message.
#       timeout - Amount of time in seconds to wait for message before moving
#           on.
#       vhost - Name of vhost where queue is.
#       ack_all - If 'true', the wait_for_message in the CLI interface will
#           acknowledge every message received and remove them from the queue.
#           Setting to 'false' will only acknowledge the message that matches
#           the message string.  Defaults to 'true'.
#       port - (Optional) RabbitMQ port, defaults to PORT.
#       url  - (Optional) URL of the RabbitMQ server defaults to URL.
#       mode - (Optional) Mode for the API defaults to 'http'.
#
#   Returns:
#       Data associated to the message matching the 'message' parameter.
#
#   Example:
#       > $msg = San::FIT::Misfit::Core::Util->wait_for_message(
#       >     queue => 'my_queue',
#       >     message => 'hello_world',
#       > );
#===============================================================================
sub wait_for_message {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => param("TESTBED", '-default', '/')
    );
    my $queue = &ParamsMgr::get_arg(
        'arg'      => "queue",
        'required' => TRUE
    );
    my $message = &ParamsMgr::get_arg(
        'arg'      => "message",
        'required' => TRUE
    );
    my $timeout = &ParamsMgr::get_arg(
        'arg'      => "timeout",
        'required' => FALSE,
        'default'  => TIMEOUT
    );
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => PORT
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => "http"
    );
    my $ack_all = &ParamsMgr::get_arg(
        'arg'      => "ack_all",
        'required' => FALSE,
        'default'  => "true"
    );
    my $verbose = &ParamsMgr::get_arg(
        'arg'      => "verbose",
        'required' => FALSE,
        'default'  => "true"
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return {};
    }

    # Create message broker
    my $misfit = San::FIT::Misfit::Core::MessageBroker->new(
        vhost => $vhost,
        url   => $url,
        port  => $port,
        mode  => $mode,
    );

    # Blocking call
    my $data = $misfit->wait_for_message(
        message => $message,
        queue   => $queue,
        timeout => $timeout,
        ack_all => $ack_all,
        verbose => $verbose,
    );

    $Log->exit();
    return $data;
} # End of wait_for_message

#===============================================================================
# Function: on_message
#   Executes a callback when a message gets to a queue
#
#   Parameters:
#       queue - Name of queue to digest messages from.
#       message - String to match in message.
#       timeout - Amount of time in seconds to wait for message before moving
#           on.
#       vhost - Name of vhost where queue is.
#       callback - Reference to method to be executed on receiving matching
#           message.
#       params - Parameters for callback method.
#       port - (Optional) RabbitMQ port, defaults to PORT.
#       url  - (Optional) URL of the RabbitMQ server defaults to URL.
#       mode - (Optional) Mode for the API defaults to 'http'.
#
#   Returns:
#       Pointer to the NATE process object of the callback.
#
#   Example:
#       > San::FIT::Misfit::Core::Util->on_message(
#       >     queue => 'my_queue',
#       >     message => 'hello_world',
#       >     callback => \&foo,
#       >     params => {param1 => 'a', param2 => 'b'},
#       >     wait_for_callback => TRUE
#       > );
#===============================================================================
sub on_message {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => param("TESTBED", '-default', '/')
    );
    my $queue = &ParamsMgr::get_arg(
        'arg'      => "queue",
        'required' => TRUE
    );
    my $message = &ParamsMgr::get_arg(
        'arg'      => "message",
        'required' => TRUE
    );
    my $timeout = &ParamsMgr::get_arg(
        'arg'      => "timeout",
        'required' => FALSE,
        'default'  => TIMEOUT
    );
    my $callback = &ParamsMgr::get_arg(
        'arg'      => "callback",
        'required' => TRUE
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => undef
    );
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => PORT
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => "http"
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return undef;
    }

    # Create message broker
    my $misfit = San::FIT::Misfit::Core::MessageBroker->new(
        vhost => $vhost,
        url   => $url,
        port  => $port,
        mode  => $mode
    );

    # Blocking call
    my $process = $misfit->on_message(
        message  => $message,
        queue    => $queue,
        timeout  => $timeout,
        callback => $callback,
        params   => $params
    );
    $Log->exit();

    return $process;
} # End of on_message

#===============================================================================
# Function: publish_message
#   Shortcut for publishing a message to an exchange.
#
#   Parameters:
#       vhost - Name of vhost exchange is on.  Defaults to vhost set in
#           constructor.
#       exchange - Name of exchange to publish message to.
#       routing_key - Routing key for message.  Defaults to '#'.
#       message - Message to be published.  May be string, array, or hash.
#       immediate - not sure what this is.
#       port - (Optional) RabbitMQ port, defaults to PORT.
#       url  - (Optional) URL of the RabbitMQ server defaults to URL
#       mode - (Optional) Mode for the API defaults to 'http'
#
#   Returns:
#       TRUE
#
#   Example:
#       > my $status = San::FIT::Misfit::Core::Util->publish_message(
#       >     exchange    => 'my_exchange',
#       >     routing_key => "#",
#       >     mode        => 'http',
#       >     message     => "Semper ubi sub ubi via http",
#       > );
#===============================================================================
sub publish_message {
    my (%args) = @_;

    $Log->enter();

    &ParamsMgr::process_args(%args);
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => param("TESTBED", '-default', '/')
    );
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => PORT
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => "http"
    );
    my $message = &ParamsMgr::get_arg(
        'arg'      => "message",
        'required' => TRUE,
    );
    my $exchange = &ParamsMgr::get_arg(
        'arg'      => "exchange",
        'required' => TRUE,
    );
    my $routing_key = &ParamsMgr::get_arg(
        'arg'      => "routing_key",
        'required' => FALSE,
        'default'  => '#'
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    # Create message broker
    my $misfit = San::FIT::Misfit::Core::MessageBroker->new(
        vhost => $vhost,
        url   => $url,
        port  => $port,
        mode  => $mode
    );
    $misfit->publish_message(
        vhost       => $vhost,
        message     => $message,
        routing_key => $routing_key,
        exchange    => $exchange
    );
    $Log->exit();

    return TRUE;
} # End of publish_message

#===============================================================================
# Function: publish_error
#   Shortcut for publishing a message to an error exchange.
#
#   Parameters:
#       vhost - Name of vhost exchange is on.  Defaults to vhost set in
#           constructor.
#       message - Message to be published.  May be string, array, or hash.
#       mode - Message can be published via HTTP or CLI.  Defaults to HTTP
#       properties - Extra RabbitMQ properties
#       port - (Optional) RabbitMQ port, defaults to PORT.
#       url  - (Optional) URL of the RabbitMQ server defaults to URL
#
#   Returns:
#       TRUE
#
#   Example:
#
#       > my $status = San::FIT::Misfit::Core::Util->publish_error(
#       >     mode    => 'http',
#       >     message => "Semper ubi sub ubi via http",
#       >);
#===============================================================================
sub publish_error {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => PORT
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => "http"
    );
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => param("TESTBED", '-default', '/')
    );
    my $message = &ParamsMgr::get_arg(
        'arg'      => "message",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    # Create message broker
    my $misfit = San::FIT::Misfit::Core::MessageBroker->new(
        vhost => $vhost,
        url   => $url,
        port  => $port,
        mode  => "http"
    );
    $misfit->publish_message(
        vhost       => $vhost,
        exchange    => 'errors',
        immediate   => TRUE,
        routing_key => '',
        message     => $message
    );
    $Log->exit();

    return TRUE;
} # End of publish_error

#===============================================================================
# Function: publish_critical
#   Shortcut for publishing a message to a critical exchange.
#
#   Parameters:
#       vhost - Name of vhost exchange is on.  Defaults to vhost set in
#           constructor.
#       message - Message to be published.  May be string, array, or hash.
#       mode - Message can be published via HTTP or CLI.  Defaults to HTTP
#       properties - Extra RabbitMQ properties
#       port - (Optional) RabbitMQ port, defaults to PORT.
#       url  - (Optional) URL of the RabbitMQ server defaults to URL
#
#   Returns:
#       TRUE
#
#   Example:
#
#       > my $status = San::FIT::Misfit::Core::Util->publish_critical(
#       >     mode    => 'http',
#       >     message => "Semper ubi sub ubi via http",
#       >);
#===============================================================================
sub publish_critical {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $url = &ParamsMgr::get_arg(
        'arg'      => "url",
        'required' => FALSE,
        'default'  => URL
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => PORT
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => FALSE,
        'default'  => "http"
    );
    my $vhost = &ParamsMgr::get_arg(
        'arg'      => "vhost",
        'required' => FALSE,
        'default'  => param("TESTBED", '-default', '/')
    );
    my $message = &ParamsMgr::get_arg(
        'arg'      => "message",
        'required' => FALSE,
        'default'  => ""
    );

    # Create message broker
    my $misfit = San::FIT::Misfit::Core::MessageBroker->new(
        vhost => $vhost,
        url   => $url,
        port  => $port,
        mode => $mode
    );

    $misfit->publish_message(
        vhost       => $vhost,
        exchange    => 'critical',
        immediate   => TRUE,
        routing_key => '',
        message     => $message
    );
    $Log->exit();

    return TRUE;
} # End of publish_critical

#===============================================================================
# Function: whoami
#   Shortcut for getting the name of the script calling this method.
#
#   Parameters:
#       None
#
#   Returns:
#       The name of the script
#
#   Example:
#
#       > my $id = San::FIT::Misfit::Core::Util->whoami();
#===============================================================================
sub whoami {
    my $id = basename($0);
    chop($id);

    return $id;
} # End of whoami

#===============================================================================
# Function: listutils
#   A helper method to get intersection, union or differences between 2 lists.
#
#   Parameters:
#       list1 - A reference to a list
#       list2 - A reference to a list
#       mode - One of the following strings: intersection, union, difference
#
#   Returns:
#       A list
#
#   Example:
#
#       > my @result = San::FIT::Misfit::Core::Util->listutils(
#       >     list1 => ['item1', 'item2'],
#       >     list2 => ['item1', 'item2'],
#       >     mode  => "difference",
#       > );
#===============================================================================
sub listutils {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $list1 = &ParamsMgr::get_arg(
        'arg'      => "list1",
        'required' => TRUE,
    );
    my $list2 = &ParamsMgr::get_arg(
        'arg'      => "list2",
        'required' => TRUE,
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        $Log->comment("This method requires two array references");
        return wantarray ? () : [];
    }

    my @intersection = ();
    my @difference = ();
    my @union = ();

    my %count = ();
    foreach my $element (@{$list1}, @{$list2}) { $count{$element}++ }
    foreach my $element (keys %count) {
        push @union, $element;
        push @{ $count{$element} > 1 ? \@intersection : \@difference }, $element;
    }

    if ($mode eq 'intersection') {
        return wantarray ? @intersection : \@intersection;
    } elsif ($mode eq 'union') {
        return wantarray ? @union: \@union;
    } elsif ($mode eq 'difference') {
        return wantarray ? @difference : \@difference;
    } else {
        $Log->comment("Unknown mode: $mode");
        return wantarray ? () : [];
    }
} # End listutils

#===============================================================================
# Function: get_percentage_of_list
#   A helper method that return a percentage of elements from a list. Useful for
#   cases where a user has a list of 'N' resources and wants only a percentage.
#
#   Parameters:
#       list1 - A reference to a list
#       percent - A number from 0 to 100
#
#   Returns:
#       A reference to list
#
#   Example:
#
#       > my @result = San::FIT::Misfit::Core::Util->get_percentage_of_list(
#       >     list   => ['item1', 'item2'],
#       >     percent => 20
#       > );
#===============================================================================
sub get_percentage_of_list {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $list = &ParamsMgr::get_arg(
        'arg'      => "list",
        'required' => TRUE,
    );
    my $percent = &ParamsMgr::get_arg(
        'arg'      => "percent",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        return wantarray ? () : [];
    }

    unless (ref($list) =~ /ARRAY/i) {
        $Log->comment("List must be a reference to an array");
        return wantarray ? () : [];
    }

    unless (scalar(@{$list})) {
        $Log->comment("List is empty!");
        return wantarray ? () : [];
    }

    my @original = @{$list};
    my $total    = scalar(@original);
    my $limit    = int (($percent * $total) / 100);
    $Log->comment(
        "You are using $limit volumes of $total for your interactions."
    );

    if (($limit == 0) and ($total > 0)) {
        $Log->comment(
            "Either the list is empty of the percentage is too small. " .
            "Only one element will be returned."
        );
        return wantarray ? ($original[0]) : [$original[0]];
    }

    my @new_list = @original[0..$limit];

    return wantarray ? @new_list : \@new_list;
} # End of get_percentage_of_list


#===============================================================================
# Function: extract_interaction_info
#   Takes the list of interaction details passed to use_misfit and uses it to
#   create queues and connect features interacting with other features
#
#   Parameters:
#       id - Unique id of the caller
#       interactions - A reference to an array of interaction information. This
#       is the same format as the interactions hash passed to use_mifsit
#
#   Returns:
#       A hash with interaction data
#
#   Examples:
#
#        > my $interaction_data = extract_interaction_info(
#        >    id => 'test',
#        >    interactions => [
#        >        { method => 'my_method', message => 'hello', queue => 'queue'}
#        > ]);
#
#        > my $interaction_data = extract_interaction_info(
#        >    id => 'test',
#        >    interactions => [
#        >         {
#        >            method  => 'enable_efficiciency',
#        >            message => 'vol.move.finished',
#        >            name    => 'enable_efficiency_on_vol_move',
#        >            sender  => 'SAN::FIT::Misfit::Features::VolOps'
#        >         }
#        > ]);
#
#        > my $interaction_data = extract_interaction_info(
#        >    id => 'test',
#        >    interactions => [
#        >         {
#        >            method  => 'enable_efficiciency',
#        >            message => 'vol.move.finished',
#        >            name    => 'enable_efficiency_on_vol_move',
#        >            sender  => 'SAN::FIT::Misfit::Features::VolOps'
#        >         },
#        >         {
#        >            method  => 'enable_efficiciency',
#        >            message => 'vol.copy.finished',
#        >            name    => 'enable_efficiency_on_vol_copy',
#        >            sender  => 'SAN::FIT::Misfit::Features::VolOps'
#        >         }
#        > ]);
#===============================================================================
sub extract_interaction_info {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $id = &ParamsMgr::get_arg(
        'arg'      => "id",
        'required' => TRUE,
    );
    my $interactions = &ParamsMgr::get_arg(
        'arg'      => "interactions",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        return {};
    }

    unless (ref($interactions) =~ /ARRAY/) {
        return {};
    }

    my $unique_queues = TRUE;
    my $queue_limit_reached = FALSE;
    my %seen_queue = ();
    my @queue_info = ();
    my @omitions = ();
    my @interaction_info = ();
    foreach my $interaction (@{$interactions}) {
        my $queue = $id;
        my $verbose = TRUE;
        my $timeout = TIMEOUT;
        my %interaction_spec = %{$interaction};
        if (defined $interaction->{name}) {
            $queue .= "." . $interaction->{name};
        } else {
            $queue .= "." . $interaction->{method};
        }
        if (defined $interaction->{timeout}) {
            $timeout = $interaction->{timeout};
        }
        $interaction_spec{timeout} = $timeout;
        if (defined $interaction->{verbose}) {
            $verbose = $interaction->{verbose};
        }
        $interaction_spec{verbose} = $verbose;
        if (defined($interaction->{sender})) {
            $queue .= "." . $interaction->{sender};
            my $i = 1;
            my $new_queue = $queue;
            while ($seen_queue{$new_queue} && $i <= QUEUE_SUFFIX_LIMIT) {
                $new_queue = $queue . "_" . $i;
                $i++;
            }
            # Set a flag if we reached the suffix counter limit.
            if ($i > QUEUE_SUFFIX_LIMIT) {
                push(@omitions,
                     "\tsender:  " . $interaction->{sender} . "\n" .
                     "\tmethod:  " . $interaction->{method} . "\n" .
                     "\tmessage: " . $interaction->{message} . "\n");
                $queue_limit_reached = TRUE;
                next;
            }
            # Set a flag if we had to make the queues unique for the user.
            unless ($i == 1) {
                $unique_queues = FALSE;
            }
            $queue = $new_queue;
            $seen_queue{$queue} = TRUE;
            push (@queue_info, {
                exchange => $interaction->{sender},
                queue    => $queue,
            });
            delete $interaction_spec{sender};
            $interaction_spec{queue} = $queue;
            if (defined $interaction->{iterations}) {
                $interaction_spec{iterations} = $interaction->{iterations};
            }
            push (@interaction_info, \%interaction_spec);
        } else {
            if (defined $interaction->{iterations}) {
                $interaction_spec{iterations} = $interaction->{iterations};
            }
            push (@interaction_info, \%interaction_spec);
        }
    }

    unless ($unique_queues) {
        $Log->warn(
            "When using the same callback method AND sender multiple times,\n" .
            "for different interactions it is recommended that you pass in \n" .
            "the \"name\" parameter with each interaction construct and \n" .
            "make the values unique. This will allow both logs and " .
            "messaging to be more readable.\n" .
            "Please refer to API docs for more information: \n".
            "http://misfitapi.gdl.englab.netapp.com/" .
            "files/Core/Util-pm.html#San.FIT.Misfit.Core.Util.use_misfit"
        );
    }

    if ($queue_limit_reached) {
        $Log->warn(
            "You have used the same method and sender combination " .
            "more than " . QUEUE_SUFFIX_LIMIT . " times.\nNo more " .
            "unique queues will be generated and some interactions " .
            "will not be run in order to guarantee proper behavior.\n" .
            "Omitions:\n" . join("\n",@omitions)
        );
    }

    my $interaction_data = {
        interactions => \@interaction_info,
        queue_info   => \@queue_info
    };

    return $interaction_data;
} # End of extract_interaction_info

#===============================================================================
# Function: enable_interactions
#   Enables feature interaction mode.
#
#   Parameters:
#       interactions - Reference to an array of hashes containing interaction
#       data.
#
#   Returns:
#       A NATE process manager
#
#   Example:
#
#       > my $interaction_mgr = enable_interactions(
#       >    interactions => [
#       >        {
#       >           method     => 'my_method',
#       >           message    => 'hello',
#       >           queue      => 'queue',
#       >           iterations => 1,
#       >        }
#       > ]);
#===============================================================================
sub enable_interactions {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $interactions = &ParamsMgr::get_arg(
        'arg'      => "interactions",
        'required' => FALSE,
        'default'  => [],
    );
    if (&ParamsMgr::error()) {
        $Log->comment("A list of feature interactions is needed");
        return undef;
    }

    my @processes = ();
    foreach my $interaction (@{$interactions}) {
        # Recent NATE checkin breaks runids using ':', so replace them with '_'
        my $nate_safe_id = $interaction->{queue};
        $nate_safe_id =~ s/\:|\./_/g;
        $interaction->{runid} = $nate_safe_id;
        my $process = NATE::Process->new(
            codespec   =>
                "{ San::FIT::Misfit::Core::Util::_interaction_handler(); }",
            params     => {
                interaction => $interaction,
            },
            runid      => $nate_safe_id,
            background => TRUE,
            onexit     => \&NATE::Process::on_exit_propagate_worst_result
        );
        push(@processes, $process);
    }

    my $interaction_mgr = NATE::ProcManager->new(
        runid       => 'Interactions',
        background  => TRUE,
        parallel    => TRUE,
        independent => FALSE,
        proc_info   => \@processes,
        onexit      => \&NATE::Process::on_exit_propagate_worst_result,
    );

    $interaction_mgr->start();
    return $interaction_mgr;
} # End of enable_interactions

#===============================================================================
# Function: diff_vols
#   This compares two lists of hashes containing volume information to remove
#   volumes already claimed by list2
#
#   Parameters:
#       list1 - A reference to an array
#       list2 - A reference to an array
#
#   Returns:
#       An array containing a list of unique elements
#
#   Example:
#
#       > my @list = diff_vols(
#       >     list1 => ['a', 'b'],
#       >     list2 => ['c', 'b', 'd']
#       > );
#===============================================================================
sub diff_vols {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $list1 = &ParamsMgr::get_arg(
        arg      => "list1",
        required => TRUE,
    );
    my $list2 = &ParamsMgr::get_arg(
        arg      => "list2",
        required => TRUE,
    );
    if (&ParamsMgr::error()){
        $Log->comment('Did not provide a list');
        return ();
    }

    my @intersection = ();
    my @difference = ();
    my @union = ();

    my %count = ();
    foreach my $element (@{$list1}, @{$list2}) {
        $count{$element->{volume}}++;
    }
    foreach my $element (@{$list1}) {
        push (@{$count{$element->{volume}} > 1 ?
            \@intersection : \@difference }, $element
        );
    }

    return @difference;
}

#===============================================================================
# Function: reserve_resources
#   Helper methods for scripts using Resource.pm and a resource.thpl to control
#   resource reservations. It makes sure each thpl running on a systemic test
#   gets resources based on their criteria and needs
#
#   Parameters:
#       request - Hash containing reservation details
#
#   Returns:
#       A reference to an array containing reserved resources
#
#   Example:
#
#       > $Vols = San::FIT::Misfit::Core::Util::reserve_resources(
#       >     request =>{
#       >         type  => 'volume',
#       >         total => $Num_Vols,
#       >     }
#       > );
#===============================================================================
sub reserve_resources {
    my (%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $request = &ParamsMgr::get_arg(
        arg      => "request",
        required => TRUE,
    );
    if (&ParamsMgr::error()){
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return [];
    }

    my $message_broker = San::FIT::Misfit::Core::MessageBroker->new(
        mode => "http"
    );

    my $id = whoami();
    $id =~ s/\.[^.]+$//;
    my $queue = $id . '.reservation';

    unless ($message_broker->queue_exists(queue => $queue)) {
        $Log->comment(
            "Queue '$queue' does not exist. Cannot return " .
            "resources to a queue that does not exist."
        );
        return [];
    }

    # Wait for resource service to call us
    $Log->comment("Waiting for resource service to attend request from $queue");
    wait_for_message(
        queue   => $queue,
        message => 'resource.reserve.ready',
    );

    # Request resources
    publish_message(
        exchange    => "resource",
        message     => {
            msg  => "resource.reserve.start",
            data => $request
        },
    );

    my $resources = wait_for_message(
        queue   => $queue,
        message => 'resource.reserve.finished',
    );

    return $resources->{data};
} # End of reserve_resources

#===============================================================================
# Function: wait_for_all_reservations_to_complete
#   Used by any script using the Misfit Resource model. Waits for a message
#   from the Resource.pm module telling the script that all resource
#   reservations are complete.
#
#   Parameters:
#       None
#
#   Returns:
#       TRUE
#
#   Example:
#
#       > San::FIT::Misfit::Core::Util->wait_for_all_reservation_to_complete();
#===============================================================================
sub wait_for_all_reservations_to_complete {
    my (%args) = @_;

    my $id = whoami();
    $id =~ s/\.[^.]+$//;

    wait_for_message(
        message => "resource.reserve.complete",
        queue => $id . ".critical_reservation",
    );

    return TRUE;
} # End of wait_for_all_reservations_to_complete

#===============================================================================
# Function: wait_for_all_io_to_start
#   Used by any script using the Misfit IO module. Waits for a message
#   from the IO.pm module telling the script that I/O has started
#
#   Parameters:
#       None.
#
#   Returns:
#       TRUE
#
#   Example:
#
#       > San::FIT::Misfit::Core::Util->wait_for_all_io_to_start();
#===============================================================================
sub wait_for_all_io_to_start {
    my (%args) = @_;

    my $id = whoami();
    $id =~ s/\.[^.]+$//;

    my $blocks = San::FIT::Misfit::Core::Util::_get_list_of_misfit_blocks(
        id => $id
    );
    unless (grep (/^io$/i, @{$blocks})) {
        $Log->comment("I/O is not one of the blocks running in this scenario.");
        return TRUE;
    }

    # Matching both start and failed because we don't want to deadlock if
    # client failed to start

    San::FIT::Misfit::Core::Util::wait_for_message(
        message => "io.started|io.failed|io.finished",
        queue   => "$id.io",
        timeout => MIN_WAIT,
    );

    return TRUE;
} # End of wait_for_all_io_to_start

#===============================================================================
# Function: get_p4_dir
#   Get the path to the p4 directory used by the test
#
#   Parameters:
#       None
#
#   Returns:
#       Path to P4 directory
#
#   Example:
#
#       > my $path = San::FIT::Misfit::Core::Util->get_p4_dir();
#===============================================================================
sub get_p4_dir {
    # Locals
    my $p4_dir = DEFAULT_P4_PATH;
    my @natelib = ();

    # Try to determine from NATE_LIB/NATE_LIB_UNIX
    if (exists($ENV{"NATE_LIB"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB"});
    }
    if (not scalar(@natelib) and exists($ENV{"NATE_LIB_UNIX"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB_UNIX"});
    }

    # Walk through all the natelib paths until we find one that works
    foreach my $natelib_path (@natelib) {
        $p4_dir = $natelib_path;
        if (-d $p4_dir) {
            return $p4_dir;
        }
    }

    return $p4_dir;
} # End of get_p4_dir

#===============================================================================
# Function: parse_path
#   Extracts cluster, vserver and volume names from a string representing a
#   path.
#
#   Parameters:
#       A String - cluster://vserver/volume
#
#   Returns:
#       A hash - (vserver => '', volume => '', cluster => '')
#
#   Example:
#
#       > my $path = San::FIT::Misfit::Core::Util->parse_path("cluster://vs/vol1");
#===============================================================================
sub parse_path {
    my ($path) = @_;

    if ($path =~ /(\S+):\/\/(\S+)\/(\S+)/) {
        return (vserver => $2, cluster => $1, volume => $3);
    } elsif ($path =~ /(\S+):(\S+)/) {
        return (vserver => $1, volume => $2);
    } else {
        $Log->comment('Error: Could not parse path: ' . $path);
        return ();
    }
} # End of parse_path

#===============================================================================
# Function: start_scenario
#   Starts the execution of a scenario using misfit and a handler for
#   interactions. Note that this method will block by default.
#
#   Parameters:
#       scenario - A path to a scenario or stest file
#       background (optional) - Defaults to FALSE. Set to TRUE if you don't
#           want this call to block.
#       id - A string representing the name of the scenario (process id)
#
#   Returns:
#       TRUE
#
#   Example:
#
#       > San::FIT::Misfit::Core::Util::start_scenario(
#       >     id       => 'Cool Scenario',
#       >     scenario => '/u/cool/test/scenario.stest'
#       > );
#===============================================================================
sub start_scenario {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $scenario = &ParamsMgr::get_arg(
        arg      => "scenario",
        required => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        arg      => "background",
        required => FALSE,
        default  => TRUE,
    );
    my $id = &ParamsMgr::get_arg(
        arg      => "id",
        required => FALSE,
        default  => 'scenario',
    );
    if (&ParamsMgr::error()){
        return FALSE;
    }

    my @services = split(",", param("SERVICES", '-default', ""));
    my @features = split(",", param("FEATURES", '-default', ""));
    my $blocks = [@services, @features];

    # Start two (and possibly three in the near future) processes. One to
    # execute the scenario/stest itself and another one to track which features
    # are in progress. The third one could be used as a 'serial killer' process
    # that can kill every process if a fatal event happens..

    my $scenario_process = NATE::Process->new(
        codespec  => '{
            use San::FIT::Misfit::Core::Util;
            San::FIT::Misfit::Core::Util::_execute_stest();
        }',
        runid          => "../$id",
        inherit_env    => TRUE,
        inherit_params => TRUE,
        params         => {
            scenario => $scenario,
        }
    );

    my $caller = whoami();
    $caller =~ s/\.[^.]+$//;

    my $interaction_watchdog = NATE::Process->new(
        codespec  => '{
            use San::FIT::Misfit::Core::Util;
            San::FIT::Misfit::Core::Util::_interaction_watchdog();
        }',
        runid  => "../$id",
        params => {
            blocks => $blocks,
            id     => $caller,
        }
    );

    my $process_manager = NATE::ProcManager->new(
        proc_info  => [$scenario_process, $interaction_watchdog],
        runid      => $id,
        background => $background,
        parallel   => $background,
    );

    $process_manager->start();
    $process_manager->wait();
    $process_manager->destroy();

    return TRUE;
} # End of start_scenario

#===============================================================================
# Function: dispatch_resource_requests
#   Responsible for distributing objects across a scenario based on requests
#   from individual thpls. IMPORTANT: This should be used by only ONE of the
#   thpls in a scenario (i.e. resource.thpl), otherwise you'll end up with
#   'two or more managers' and your script will be as confused as Adam or Eve
#   on Mother's day... Ideally, this call should be right after the code
#   responsible for setting up the scenario
#
#   Parameters:
#       resource - A San::FIT::Misfit::Services::Resource object or a similar
#           'resource' object capable of tracking resource utilization.
#
#   Returns:
#       TRUE once all reservations have been dispatched
#
#   Example:
#
#       > San::FIT::Misfit::Core::Util::dispatch_resource_requests(
#       >     resource => $Resource,
#       > );
#===============================================================================
sub dispatch_resource_requests {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $resource_obj = &ParamsMgr::get_arg(
        arg      => "resource",
        required => FALSE,
        default  => undef,
    );
    if (&ParamsMgr::error()){
        return FALSE;
    }

    my @services = split(",", param("SERVICES", '-default', ""));
    my @features = split(",", param("FEATURES", '-default', ""));
    my $blocks = [@services, @features];

    if (ref($resource_obj) !~ /San::FIT::Misfit::Services::Resource/i) {
        $resource_obj = San::FIT::Misfit::Services::Resource->new();
        $resource_obj->_gather_resources();
    }
    my $misfit = San::FIT::Misfit::Core::MessageBroker->new(mode => "http");
    foreach my $block (@{$blocks}) {
        if ($block =~ /^resource$/) {
            next;
        }
        unless ($misfit->exchange_exists(exchange => $block)) {
            next;
        }
        $Log->comment("Reserving resources for $block ");
        San::FIT::Misfit::Core::Util::publish_message(
            exchange    => $block,
            message     => "resource.reserve.ready",
            routing_key => $block
        );

        my $request = San::FIT::Misfit::Core::Util::wait_for_message(
            queue   => "resource.local",
            message => "resource.reserve.start",
            timeout => SIXTY_SECONDS,
        );

        # If the message was not received on time it means that either something
        # is wrong with the 'block' or that it's running but does not care about
        # reservations.

        my $resources = [];
        if (($request->{status} =~ /error/i) or
            ($request->{msg} =~ /Abort initiated after waiting/i)) {
            $Log->comment(
                "It looks like '$block' is not interested in reserving objects " .
                "or something is not working. Skipping the reservation process ".
                "for this block"
            );
        } else {
            if (($block =~ /^io$/) and defined($resource_obj->{client_device_list})) {
                $resources = $resource_obj->{client_device_list};
            } else {
                $resources = $resource_obj->reserve(%{$request->{data}});
            }
        }

        San::FIT::Misfit::Core::Util::publish_message(
            exchange => $block,
            message  => {
                msg  => "resource.reserve.finished",
                data => $resources
            },
            routing_key => $block
        );

        my $header = "\n" . "=" x COLUMN . "\n" .
            "Resources reserved for $block \n" . "=" x COLUMN;
        my @body = ($header);
        my $row  = "";
        if (ref($resources) =~ /ARRAY/i) {
            foreach my $resource (@{$resources}) {
                $row = "";
                while (my ($key, $value) = each %{$resource}) {
                    next unless (defined($value));
                    $row = sprintf("%10s => %10s", $key, $value);
                    push(@body, $row);
                }
                push(@body, "_" x COLUMN . "\n");
            }
        } elsif (ref($resources) =~ /HASH/i) {
            while (my ($key, $value) = each %{$resources}) {
                next unless (defined($value));
                $row = sprintf("%10s => %10s", $key, $value);
                push(@body, $row);
            }
            push(@body, "_" x COLUMN . "\n");
        }
        push (@body, "\n");

        # Yes, using print on purpose to get a clean format in the logs
        print join("\n", @body);
    }

    # Let everyone know that resource reservation is complete
    San::FIT::Misfit::Core::Util::publish_message(
        exchange => ref($resource_obj),
        message  => "resource.reserve.complete",
    );

    return TRUE;
} # End of dispatch_resource_requests

#===============================================================================
# Function: run_io
#   Simple wrapper around IO tools allowing users to run, monitor, wait and
#   stop I/O services with a single command. I/O will run and will be stopped
#   once the last feature and feature interaction running in the scenario
#   is done.
#
#   Parameters:
#       hosts - An array containing host names.
#       presets - An array containing SAN FIT IOPresets.
#       io_module - Currently a no-op. Defaults to Misfit I/O. This will allow
#           us to pass a different I/O tool or strategy
#       pass_thru - Currently a no-op. This will allow user to pass specific
#           parameters to their io_module (in case they're not using Misfit IO)
#       duration - (optional) parameter to be used as a common runtime value.
#           defaults to the testbed parameter IO_RUNTIME. If this parameter was
#           not passed, it will run for 5 minutes
#
#   Returns:
#       TRUE once I/O is done running. FALSE if I/O fails to start
#
#   Example:
#
#       > San::FIT::Misfit::Core::Util::run_io(
#       >     hosts => $hosts,
#       > );
#===============================================================================
sub run_io {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $hosts_ref = &ParamsMgr::get_arg(
        'arg'      => "hosts",
        'required' => FALSE,
        'default'  => (),
    );
    my $presets_ref = &ParamsMgr::get_arg(
        'arg'      => "presets",
        'required' => FALSE,
        'default'  => (),
    );
    my $io_module = &ParamsMgr::get_arg(
        'arg'      => "io_module",
        'required' => FALSE,
        'default'  => "misfit",
    );
    my $pass_thru = &ParamsMgr::get_arg(
        'arg'      => "pass_thru",
        'required' => FALSE,
        'default'  => {},
    );
    my $duration = &ParamsMgr::get_arg(
        'arg'      => "duration",
        'required' => FALSE,
        'default'  => undef,
    );
    my $try_minutes = &ParamsMgr::get_arg(
        'arg'      => "try_minutes",
        'required' => FALSE,
        'default'  => DEFAULT_TRY_LIMIT,
    );
    if (&ParamsMgr::error()){
        return FALSE;
    }

    my @hosts = @{$hosts_ref};
    my @presets = @{$presets_ref};
    my $misfit = San::FIT::Misfit::Core::MessageBroker->new();

    if ($io_module =~ /misfit/i) {
        unless (scalar(@hosts) and scalar(@presets)) {
            $Log->comment(
                "The lists of hosts or presets are empty. I/O will not run."
            );
            return FALSE;
        }

        unless (defined($duration)) {
            $duration = param("IO_DURATION", '-default', IO_DURATION);
        }

        if (scalar(@hosts) > scalar(@presets)) {
            my $i = 0;
            while (scalar(@presets) < scalar(@hosts)) {
                push(@presets, $presets[$i]);
                $i++;
            }
        }

        # I'm not sure if I should be proud or ashamed for this nasty clever
        # Perl trick! This will turn two lists (hosts and presets) into a
        # nice hash: (host_1 => preset_1, host_2 => preset_2...)

        my %host_preset = ();
        $Log->comment(Dumper(\@hosts));
        @host_preset{@hosts} = @presets;
        $Log->comment("Host Preset\n" . Dumper(\%host_preset));
        my @io_data = ();

        while (my ($key, $value) = each %host_preset) {
            my $tool = DEFAULT_IO_TOOL;
            # Unless the client is sanblaze, change tool type
            if ($key =~ /sanblaze/i) {
                $tool = "sanblaze";
            }
            my $data = {
                host      => $key,
                workload  => $value,
                duration  => $duration,
                result_id => $key,
                tool      => $tool,
            };
            push(@io_data, $data);
        }

        # Send IO start
        my $io = San::FIT::Misfit::Services::IO->new();
        if (not defined($io)) {
            return FALSE;
        }

        $Log->comment("IO data\n" . Dumper(\@io_data));

        # Start the I/O keeper
        my $status = $io->start_keeper();
        if (!$status) {
            return FALSE;
        }

        my @attempted_start_hosts = ();
        my @started_hosts = ();
        my @attempted_stop_hosts = ();
        my @stopped_hosts = ();

        # Start IO on hosts
        $Log->comment("Start IO on hosts provided in IO data.");
        foreach my $data (@io_data) {
            $io->start(
                host      => $data->{host},
                tool      => $data->{tool},
                workload  => $data->{workload},
                duration  => $data->{duration},
                result_id => $data->{host},
            );
            # Track started IO on hosts
            push(@attempted_start_hosts, $data->{host});
        }

        # Confirm IO has started
        @started_hosts = $io->get_successful_io_results_for_hosts(
            hosts       => \@attempted_start_hosts,
            try_minutes => $try_minutes,
        );

        # Fail if IO never started
        if (scalar(@started_hosts) == 0) {
            $Log->comment(
                "I/O failed to start on any clients. Check the clients " .
                "and make sure they're configured and operational."
            );
            $io->local_message(
                message     => "io.failed",
                routing_key => "#",
            );
            return FALSE;
        }

        # Announce that IO has begun
        $Log->comment("IO started on hosts: ". Dumper(@started_hosts));
        $io->local_message(
            message     => "io.started",
            routing_key => "#",
        );

        # Wait for I/O processes to either finish or be terminated. I/O will
        # stop when either all features finish executing or when the runtime
        # expires

        San::FIT::Misfit::Core::Util::wait_for_message(
            queue   => 'io.critical',
            message => 'io.stop',
            timeout => $duration,
            ack_all => 'false',
        );

        # Stop IO on all hosts
        $Log->comment("Stop IO on hosts: ". Dumper(@started_hosts));
        foreach my $host (@started_hosts) {
            # Send command to stop IO
            $io->stop(
                result_id => $host,
            );
            # Track host as stopped
            push(@attempted_stop_hosts, $host);
        }

        # Confirm IO has stopped
        @stopped_hosts = $io->get_successful_io_results_for_hosts(
            hosts       => \@attempted_stop_hosts,
            try_minutes => $try_minutes,
        );

        # Announce that IO has completed
        $Log->comment("IO stopped on hosts: ". Dumper(@stopped_hosts));
        $io->local_message(
            message     => "io.finished",
            routing_key => "#",
        );

        $io->stop_keeper();
    } else {
        # Any other I/O tool we want to support? Old IOMan, etc.
        # Use the pass_thru param to pass parameters to a specific tool/code
        $Log->comment("'$io_module' is not supported by this Utility package");
        return FALSE;
    }

    return TRUE
} # End of run_io

#===============================================================================
# Function: get_number_of_features_running
#   Returns the number registered blocks that are currently running. Every
#   script using misfit will add itself to the list when calling use_misfit and
#   will remove itself by calling cleanup_misfit. If the returned number is 1
#   it means that the caller is the only registered block active.
#
#   Parameters:
#       None
#
#   Returns:
#       Number of active misfit features or scripts
#
#   Example:
#
#       > $num_features = get_number_of_features_running();
#===============================================================================
sub get_number_of_features_running {
    my (%args) = @_;

    my $misfit = San::FIT::Misfit::Core::MessageBroker->new(mode => "http");

    my @messages = ();
    try {
        # Get the list of features currently running
        @messages = $misfit->get_message(
            requeue => "true",
            count   => MAX_MESSAGES_PER_QUEUE,
            queue   => "features"
        );
    } otherwise {
        my $e = shift();
        # Catch the case where any part of the url is missing and assume
        # there are no features. For any other error, re-throw the exception.
        if ($e->text() =~ /used in the URL exist and are valid/) {
            return scalar(@messages);
        } else {
            $e->throw();
        }
    };

    return scalar(@messages);
} # End get_number_of_features_running

#===============================================================================
# Function: execute_independent_script
#   Executes an external, non-misfit script or thpl as a NATE process.
#
#   Parameters:
#       script - A path to the thpl to execute
#       runid  - (optional) An identifier for the process
#       iterations - Number of times to execute. Defaults to "0" which will
#           execute the script multiple times until the last misfit feature
#           finishes.
#
#   Returns:
#       TRUE or FALSE depending on whether the processes ran successfully or not
#
#   Example:
#
#       > $status = execute_independent_script(
#       >     script => "path/to/my/thpl",
#       >     runid  => "my_thpl",
#       >     iterations => 1, # Run just once
#       > );
#===============================================================================
sub execute_independent_script {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $script = &ParamsMgr::get_arg(
        'arg'      => "script",
        'required' => TRUE,
    );
    my $runid = &ParamsMgr::get_arg(
        'arg'      => "runid",
        'required' => FALSE,
        'default'  => ''
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 0
    );
    if (&ParamsMgr::error()){
        return FALSE;
    }

    my $id = whoami();
    $id =~ s/\.[^.]+$//;
    my $fail_count = 0;

    # If the runid was not passed, use the script's name removing the extension
    if ($runid eq '') {
        my $script_name = $script;
        $script_name =~ s/\.[^.]+$//;
        $runid = $script_name
    }

    if ($iterations == 0) {
        my $i = 1;
        while (San::FIT::Misfit::Core::Util::get_number_of_features_running() > 1) {
            my $process = NATE::Process->new(
                codespec       => $script,
                runid          => $i . "_" . $runid,
                background     => TRUE,
                onexit         => \&NATE::Process::on_exit_propagate_worst_result,
                inherit_params => TRUE,
                inherit_env    => TRUE,
            );
            $process->start();
            San::FIT::Misfit::Core::Util::publish_message(
                message  => $id . ":started. Iteration $i",
                exchange => $id
            );
            $process->wait();

            # Get exit codes: Anything other than PASS (0), WARN (250)
            # or NA (251) should be considered an error.

            my $exit_code = $process->worst_result()->exit_code();
            if ($exit_code =~ /0|250|251/) {
                San::FIT::Misfit::Core::Util::publish_message(
                    message  => $id . ":finished. Iteration $i",
                    exchange => $id
                );
            } else {
                my @error_msgs = $process->worst_result()->messages();
                foreach my $error (@error_msgs) {
                    San::FIT::Misfit::Core::Util::publish_message(
                        message  => $error,
                        exchange => 'errors'
                    );
                    $fail_count++;
                }
            }
            $process->destroy();
            $i++;
        }
    } else {
        for (my $j = 0; $j < $iterations; $j++) {
            my $process = NATE::Process->new(
                codespec       => $script,
                runid          => $j . "_" . $runid,
                background     => TRUE,
                onexit         => \&NATE::Process::on_exit_propagate_worst_result,
                inherit_params => TRUE,
                inherit_env    => TRUE,
            );
            $process->start();
            San::FIT::Misfit::Core::Util::publish_message(
                message  => $id . ":started. Iteration $j",
                exchange => $id
            );
            $process->wait();
            my $exit_code = $process->worst_result()->exit_code();
            if ($exit_code =~ /0|250|251/) {
                San::FIT::Misfit::Core::Util::publish_message(
                    message  => $id . ":finished. Iteration $j",
                    exchange => $id
                );
            } else {
                my @error_msgs = $process->worst_result()->messages();
                foreach my $error (@error_msgs) {
                    San::FIT::Misfit::Core::Util::publish_message(
                        message  => $error,
                        exchange => 'errors'
                    );
                    $fail_count++;
                }
            }
            $process->destroy();
        }
    }

    if ($fail_count > 0) {
        return FALSE;
    } else {
        return TRUE;
    }
} # End of execute_independent_script

################################################################################
# PRIVATE METHODS
################################################################################

#===============================================================================
# Description: Executes an stest process and starts the scenario
#===============================================================================
sub _execute_stest {
    use NATE::ParamSet;
    use NATE::Parser::Stest;

    # Get subprocess parameters
    my $params = NATE::ParamSet->get_global();
    my $scenario = $params->get("scenario");

    # Parse an stest file and extract its content
    my $stest = NATE::Parser::Stest->new();
    my $stest_content = Stest::get_file_content($scenario);
    $Log->comment($stest_content);

    # Parse the contents of the stest file
    my $stest_parser = $stest->parse(content => $stest_content);
    $stest_parser->start();

    return TRUE;
} # End of _execute_stest

#===============================================================================
# Description: Listens for interaction messages and executes the desired method
#   or terminates and exits.
#===============================================================================
sub _interaction_handler {
    require NATE::Process;
    my $params = NATE::ParamSet->get_global();
    my $interaction = $params->get('interaction');

    # Initialize variable for number of times to execute the interaction
    # n > 0 executes n interactions, n < 0 will continuously listen for new
    # interactions.
    my $iterations = DEFAULT_INTERACTION_ITERATIONS;
    if (defined $interaction->{iterations}) {
        $iterations = $interaction->{iterations};
    }

    my $verbose = TRUE;
    if (defined($interaction->{verbose})) {
        $verbose = $interaction->{verbose};
    }

    # Print a message to the log to let user know how many iterations will
    # execute
    if ($iterations < 0) {
        $Log->comment(
            "This interaction will execute until a terminating message is " .
            "received."
        );
    } elsif ($iterations > 0) {
        $Log->comment(
            "This interaction will be limited to $iterations iterations."
        );
    } else {
        return NATE::Result::NA->new(
            "No interactions will excecute. Iterations was set to 0."
        )->log();
    }

    # Publish message to interactions queue to let the interaction watchdog
    # know that an interaction handler is listening for messages.  This prevents
    # cleanup_misfit from blindly killing these processes before they have
    # finished closing. "Warning test killed because ancestor stopped.. etc"
    publish_message(
        exchange    => 'interactions',
        routing_key => '#',
        message     => 'waiting',
    );

    my @processes = ();
    my $id_count = 0;

    while ($iterations) {
        # Wait for terminating or interaction triggering string in message
        # on specified queue
        my $data = San::FIT::Misfit::Core::Util::_execute_interaction(
            queue   => $interaction->{queue},
            message => $interaction->{message} . "|" . BYE . "|" .
                       INTERACTION_BYE,
            verbose => $verbose,
            timeout => $interaction->{timeout}
        );
        # End loop if a terminating message was received
        if ($data =~ /terminated|skipped|timedout/i) {
           last;
        }

        # Increment counter for making a runid
        $id_count++;
        try {
            # Create a new NATE Process for executing the interaction
            # method in the background so we can continue waiting for new
            # interactions
            # Note that each interaction publishes a 'running' message to the
            # interactions queue and then waits for that message once the
            # interaction has completed.  This allows cleanup_misfit to block
            # while there are running interaction before telling the handlers
            # to exit and stop receiving more messages. It also allows the
            # interaction_watchdog to wait for these running interactions so
            # that it wont tell IO to stop before the interactions are complete.
            my $interaction_process = NATE::Process->new(
                codespec   => "{ ".
                    "San::FIT::Misfit::Core::Util::publish_message(".
                        "exchange    => 'interactions'," .
                        "routing_key => '#', " .
                        "message     => 'running'," .
                    ");" .
                    "my \$params = NATE::ParamSet->get_global();" .
                    "my \$data   = \$params->get('data');" .
                    $interaction->{method} . "(\%{\$data});" .
                    "San::FIT::Misfit::Core::Util::wait_for_message(" .
                        "message => 'running'," .
                        "queue   => 'interactions'," .
                        "ack_all => 'false'," .
                        "timeout => 5," .
                    ");" .
                "}",
                params     => {data => $data},
                runid      => $interaction->{runid} . "_" . $id_count,
                background => TRUE,
            );
            # Start process and add it to list of processes for keeping
            # track of still running interactions
            $interaction_process->start();
            push(@processes, $interaction_process);
        } otherwise {
            my $e = shift;
            $Log->comment('Interaction terminated' . $e->text());
        };
        $iterations--;
    }

    # When the interaction listener loop exits, wait for and pull results from
    # each process. Use NATE::Result to determine the worst from all the
    # process results.
    my @results = ();
    foreach my $proc (@processes) {
        $proc->wait();
        push(@results, $proc->worst_result());
        $proc->destroy();
    }
    my $worst_result = NATE::Result->worst(@results);

    # Remove a running message from "interactions" queue to signal that this
    # interaction listner has completed and return the worst result from the
    # interaction processes
    San::FIT::Misfit::Core::Util::wait_for_message(
        message => 'waiting',
        queue   => 'interactions',
        ack_all => 'false',
        timeout => INTERACTIONS_RUNNING_WAIT,
    );

    $worst_result->log() if defined $worst_result;

} # End of _interaction_handler

#===============================================================================
# Description: Responsible for making sure interactions wait for them to be
# dispatched or to end them gracefully
#===============================================================================
sub _interaction_watchdog {
    use NATE::ParamSet;

    # Get subprocess parameters
    my $params = NATE::ParamSet->get_global();
    my $blocks = $params->get("blocks");
    my $id     = $params->get("id");

    # In order to ensure that interactions wait long enough for other features
    # to finish, wait for all of them to notify this main watchdog and then
    # broadcast a 'BYE' message to terminate those that were not triggered.

    my $misfit = San::FIT::Misfit::Core::MessageBroker->new(mode => "http");

    # TODO: We need to come up with a less-hacky approach for this. Using a
    # a lame sleep to delay the executing of this subprocess and give the
    # main process time to create the features queue. An easy fix is to do it
    # via messaging

    sleep(DELAY);

    my @messages = $misfit->get_message(
        requeue => "true",
        count   => MAX_MESSAGES_PER_QUEUE,
        queue   => "features",
    );

    while (scalar(@messages)) {
        @messages = $misfit->get_message(
            requeue => "true",
            count   => MAX_MESSAGES_PER_QUEUE,
            queue   => "features",
        );
        sleep(DELAY / 2);
    }

    # If we reach this point it means that all features have 'deregistered'
    # from misfit. Let any pending interaction finish

    San::FIT::Misfit::Core::Util::_wait_for_interactions_to_finish(
        broker => $misfit
    );

    $Log->comment("All features are de-registered, sending critical bye.");
    $misfit->publish_message(
        exchange    => 'critical',
        immediate   => TRUE,
        routing_key => '#',
        message     => BYE,
    );

    # At this point all interaction handlers that are "waiting" should have
    # received the BYE message to exit gracefully.  Now, we just need to wait
    # for them to finish.

    San::FIT::Misfit::Core::Util::_wait_for_interaction_handlers_to_finish(
        broker => $misfit
    );


    # If there's an IO module and it happens to be listening to the critical
    # queue, here's where we can tell it to stop whatever it's doing

    $misfit->publish_message(
        exchange => 'critical',
        message  => 'io.stop',
    );

    # If there were errors reported by any of the misfit blocks then get them
    # from the queue before it gets deleted. This will contain useful triage
    # data

    San::FIT::Misfit::Core::Util::_get_error_messages(id => $id);

    return TRUE;
} # End of _interaction_watchdog

#==============================================================================
# Description: Creates queues for every feature and service included in a
# given secnario for resource reservation purposes
#==============================================================================
sub _create_reservation_queues {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $blocks = &ParamsMgr::get_arg(
        arg      => "blocks",
        required => TRUE,
    );
    my $id = &ParamsMgr::get_arg(
        arg      => "id",
        required => TRUE,
    );
    if (&ParamsMgr::error()) {
        return wantarray ? () : [];
    }

    my @queues = ();
    push(@queues,(
        {
            exchange    => $id,
            queue       => $id . '.reservation',
            routing_key => $id,
        },
        {
            exchange    => 'resource',
            queue       => 'resource.local',
            routing_key => '#',
        },
        {
            exchange    => 'San::FIT::Misfit::Services::Resource',
            queue       => $id . '.critical_reservation',
            routing_key => '#',
        })
    );

    return wantarray ? @queues : \@queues;
} # End of _create reservation queues

#===============================================================================
# Description: Waits for specific messages before dispatching an interaction
#===============================================================================
sub _execute_interaction {
    my (%args) = @_;

    my $queue   = $args{queue};
    my $message = $args{message};
    my $verbose = $args{verbose};
    my $timeout = $args{timeout};

    my $data = wait_for_message(
        message => $message,
        queue   => $queue,
        timeout => $timeout,
        ack_all => 'false',
        verbose => $verbose,
    );

    my $termination = BYE . "|" . INTERACTION_BYE;
    if ($data->{msg} =~ /$termination/i) {
        return "terminated";
    } elsif ($data->{msg} =~ /does not exist. Skipping wait/i) {
        return "skipped";
    } elsif ($data->{msg} =~ /Abort initiated after waiting./i) {
        return "timedout";
    } else {
        return $data;
    }
} # End of _execute_interaction

#===============================================================================
# Description: Waits for interactions to finish (processes)
#===============================================================================
sub _wait_for_interactions_to_finish {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $misfit = &ParamsMgr::get_arg(
        arg      => "broker",
        required => TRUE,
    );
    if (&ParamsMgr::error()) {
        return FALSE;
    }

    if ($misfit->queue_exists(queue => "interactions")) {
        my @messages = $misfit->get_message(
            requeue => "true",
            count   => MAX_MESSAGES_PER_QUEUE,
            queue   => "interactions",
        );

        while (scalar(@messages)) {
            @messages = $misfit->get_message(
                requeue => "true",
                count   => MAX_MESSAGES_PER_QUEUE,
                queue   => "interactions",
            );
            last unless(grep(/running/,@messages));
            sleep(DELAY);
        }
    }

    return TRUE;
} # End of wait_for_interactions_to_finish

#===============================================================================
# Description: Waits for interaction handlers to finish (processes)
#===============================================================================
sub _wait_for_interaction_handlers_to_finish {
        my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $misfit = &ParamsMgr::get_arg(
        arg      => "broker",
        required => TRUE,
    );
    if (&ParamsMgr::error()) {
        return FALSE;
    }

    if ($misfit->queue_exists(queue => "interactions")) {
        my @messages = $misfit->get_message(
            requeue => "true",
            count   => MAX_MESSAGES_PER_QUEUE,
            queue   => "interactions",
        );

        while (scalar(@messages)) {
            @messages = $misfit->get_message(
                requeue => "true",
                count   => MAX_MESSAGES_PER_QUEUE,
                queue   => "interactions",
            );
            last unless(grep(/waiting/,@messages));
            sleep(DELAY);
        }
    }

    return TRUE;
}

#===============================================================================
# Description: Waits for features to finish executing
#===============================================================================
sub _wait_for_features_to_finish {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $misfit = &ParamsMgr::get_arg(
        arg      => "broker",
        required => TRUE,
    );
    my $id = &ParamsMgr::get_arg(
        arg      => "id",
        required => TRUE,
    );
    if (&ParamsMgr::error()) {
        return FALSE;
    }

    if ($misfit->queue_exists(queue => "features")) {
        $misfit->wait_for_message(
            message => $id,
            queue   => "features",
            ack_all => "false",
            timeout => SHORT_TIMEOUT,
        );
    }

    return TRUE;
} # End of wait_for_features_to_finish

#===============================================================================
# Description: Get the list of services and features specified in the SERVICES
# and FEATURES variables passed to the scenario or thpl
#===============================================================================
sub _get_list_of_misfit_blocks {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $id = &ParamsMgr::get_arg(
        arg      => "id",
        required => TRUE,
    );

    # Get list of services and features used by this misfit user. These params
    # are generic params passed to independent thpls or stest files

    my @services = split(",", param("SERVICES", '-default', ""));
    my @features = split(",", param("FEATURES", '-default', ""));
    my $blocks;

    # Add the misfit user to the list of blocks to avoid creating unnecessary
    # queues.

    push(@{$blocks}, ($id, @services, @features));
    my @no_duplicates = List::MoreUtils::uniq(@{$blocks});
    $blocks = \@no_duplicates;

    return $blocks
} # End of _get_list_of_misfit_blocks

#===============================================================================
# Description: Download all messages from error queue to show errors from
# scenario
#===============================================================================
sub _get_error_messages {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $id = &ParamsMgr::get_arg(
        arg      => "id",
        required => FALSE,
        default  => undef,
    );

    unless (defined($id)) {
        $id = whoami();
        $id =~ s/\.[^.]+$//;
    }

    my @messages = ();
    my $misfit = San::FIT::Misfit::Core::MessageBroker->new();
    if ($misfit->queue_exists(queue => $id . '.errors')) {
        @messages = $misfit->get_message(
            queue => $id . '.errors',
            count => MAX_MESSAGES_PER_QUEUE,
        );
    } else {
        $Log->comment(
            "Doesn't look like your queue exists! " . $id . ".errors"
        );
        return;
    }

    if (!scalar(@messages)) {
        $Log->comment(
            "It looks like there were no errors pushed to the Misfit error "   .
            "tracker. This means that either the test completed successfully " .
            "or the errors were thrown by non-misfit scripts or libraries. "   .
            "Please review the logs and make sure there were no errors."
        );
        return;
    }

    my $log_out = "\n";

    $log_out = $log_out . "*" x COLUMN . "\n" .
    "* This Log file downloads all errors encountered by misfit-enabled\n" .
    "* modules. This file should be used to help quickly triage errors\n" .
    "* encountered during a SAN FIT test.\n" .
    "*\n" .
    "* From here you can find features that published errors and look for\n" .
    "* more specific data in the logs for that feature.\n".
    "*" x COLUMN . "\n";

    $Log->redirect(
        path => File::Spec->catdir(
            Tharn::param('TOP_LOGDIR'),
            Tharn::param('RUNID'),
            'triage_log.log',
        )
    );

    foreach my $message (@messages) {
        $log_out = $log_out . "_" x COLUMN . "\n";
        if (ref($message) =~ /HASH/i) {
            $log_out = $log_out .
                sprintf("%15s => %s", "error", $message->{msg} . "\n");
            while (my ($key, $value) = each %{$message->{data}}) {
                $log_out = $log_out . sprintf("\t%20s => %s\n", $key, $value);
            }
        } else {
            $log_out = $log_out . "\n" .
                sprintf("%15s => %s", "error", $message . "\n");
        }
        $log_out = $log_out . "_" x COLUMN . "\n";
    }
    $Log->comment($log_out);

    $Log->reset_path();
} # End of _get_error_messages

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/core/Util>

=cut

1;
