# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Workflow class to wrap Process Manager code
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Core::Workflow
# Wrapper for NATE Process Manager code

package San::FIT::Misfit::Core::Workflow;


# Compiler Directives
use strict;
use warnings;
use Data::Dumper;

# NetApp's modules
use NATE::Util::DataTypes qw(TRUE FALSE);
use NATE::BaseException qw(:try);
use NATE::Log qw(log_global);
use NATE::ProcManager;
use NATE::Process;
use ParamsMgr;
use constant {
    DEFAULT_WORKFLOW_NAME => "workflow",
};

#===============================================================================
# Constructor: new
#	Workflow constructor
#
#	Parameters:
#		background - If FALSE, wait for child process to complete.  If TRUE,
#			continue execution of parent.
#		independent - TODO
#		runid - NATE runid of workflow process.
#		steps - Code to executed in subprocesses.
#
#	Returns:
#		New workflow object
#===============================================================================
sub new {
    my ($class, %args) = @_;

    my $self = bless {}, $class;
    $self->{log} = log_global();
    $self->{log}->enter();

    # Parse parameters
    &ParamsMgr::process_args(%args);
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => FALSE
    );
    my $independent = &ParamsMgr::get_arg(
        'arg'      => "independent",
        'required' => FALSE,
        'default'  => FALSE
    );
    my $runid = &ParamsMgr::get_arg(
        'arg'      => "runid",
        'required' => FALSE,
        'default'  => DEFAULT_WORKFLOW_NAME
    );
    my $steps = &ParamsMgr::get_arg(
        'arg'      => "steps",
        'required' => FALSE,
        'default'  => []
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $self->{log}->exit();
        return undef;
    }

    # Workflow steps
    $self->{steps} = $steps;

    # Create process

    $self->{workflow} = NATE::ProcManager->new(
        runid       => $runid,
        background  => $background,
        independent => $independent,
        proc_info   => $self->{steps},
        onexit      => \&NATE::Process::on_exit_propagate_worst_result,
    );

    $self->{log}->exit();
    return $self;
} # End of new

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Function: start
#	Starts workflow
#
#	Parameters:
#		on_start - TODO
#
#	Returns:
#		TRUE if successful, FALSE otherwise
#===============================================================================
sub start {
    my ($self, %args) = @_;

    $self->{log}->enter();
    &ParamsMgr::process_args(%args);
    my $on_start = &ParamsMgr::get_arg(
        'arg'      => "on_start",
        'required' => FALSE,
        'default'  => undef
    );
    if (defined ($self->{workflow}) or scalar(@{$self->{steps}})) {
        if (defined(($on_start))) {
            $on_start->();
        }
        try {
            $self->{workflow}->start();
        }
        otherwise {
            my $ex = shift;
            $self->{log}->error($ex->text());
            NATE::BaseException->throw($ex);
        };
    } else {
        $self->{log}->error("Workflow doesn't exist or it's empty");
        return FALSE;
    }

    $self->{log}->exit();
    return TRUE;
} # End of start

#===============================================================================
# Function: stop
#	Stops workflow
#
#	Parameters:
#		on_stop - TODO
#
#	Returns:
#		TRUE if successful, FALSE otherwise
#===============================================================================
sub stop {
    my ($self, %args) = @_;

    $self->{log}->enter();
    &ParamsMgr::process_args(%args);
    my $on_stop = &ParamsMgr::get_arg(
        'arg'      => "on_stop",
        'required' => FALSE,
        'default'  => undef
    );
    if (defined ($self->{workflow}) or scalar(@{$self->{steps}})) {
	try{
           $self->{workflow}->stop();
        }
        otherwise {
           my $ex = shift;
           $self->{log}->error("Caught in workflow->stop()");
           NATE::BaseException->throw($ex);
        };

        if (defined(($on_stop))) {
           $on_stop->();
        }
    } else {
        $self->{log}->error("Workflow doesn't exist or it's empty");
        return FALSE;
    }

    $self->{log}->exit();
    return TRUE;
} # End of stop

#===============================================================================
# Function: wait
#	Wait for a workflow to finish
#
#	Parameters:
#		after_wait - TODO
#
#	Returns:
#		TRUE if successful, FALSE otherwise
#===============================================================================
sub wait {
    my ($self, %args) = @_;

    $self->{log}->enter();
    &ParamsMgr::process_args(%args);
    my $after_wait = &ParamsMgr::get_arg(
        'arg'      => "after_wait",
        'required' => FALSE,
        'default'  => undef
    );

    if (defined ($self->{workflow}) or scalar(@{$self->{steps}})) {
       try {
           $self->{workflow}->wait();
       }
       otherwise {
           my $ex = shift;
           $self->{log}->error("Caught in workflow->wait()");
           NATE::BaseException->throw($ex);
       };
        if (defined(($after_wait))) {
            $after_wait->();
        }
     } else {
        $self->{log}->error("Workflow doesn't exist or it's empty");
        return FALSE;
     }

    $self->{log}->exit();
    return TRUE;
} # End of wait

#===============================================================================
# Function: is_running
#	Checks if a workflow is running
#
#	Returns:
#		TRUE if running, FALSE otherwise
#===============================================================================
sub is_running {
    my ($self, %args) = @_;

    $self->{log}->enter();
    my $status = FALSE;
    if (defined ($self->{workflow}) or scalar(@{$self->{steps}})) {
        $status = $self->{workflow}->is_running();
    } else {
        $self->{log}->error("Workflow doesn't exist or it's empty");
        return $status;
    }

    $self->{log}->exit();
    return $status;
} # End of is_running

#===============================================================================
# Function: destroy
#	Destroy workflow
#
#	Returns:
#		TRUE if successful, FALSE otherwise
#===============================================================================
sub destroy {
    my ($self, %args) = @_;

    $self->{log}->enter();
    if (defined ($self->{workflow})) {
       try {
           $self->{workflow}->destroy();
       }
       otherwise {
           my $ex = shift;
           $self->{log}->error("Caught in workflow->destroy()");
           NATE::BaseException->throw($ex);
       };
    } else {
           $self->{log}->error("Workflow doesn't exist");
           return FALSE;
    }
    $self->{log}->exit();
    return TRUE;
} # End of destroy

#===============================================================================
## Function: add($self, %steps)
#	Adds a step to the workflow
#
#	Parameters:
#		%steps - Hash containing NATE process information
#
#	Returns:
#		TRUE
#===============================================================================
sub add {
    my ($self, %args) = @_;

    $self->{log}->enter();
    my $process = NATE::Process->new(%args);
    push(@{$self->{steps}}, $process);

    $self->{log}->exit();
    return TRUE;
} # End of add

#===============================================================================
# Function: results
#	Gets results from workflow
#
#	Returns:
#		Workflow results
#===============================================================================
sub results {
    my ($self, %args) = @_;

    # XXX: Add mechanism to summarize and display results inteligently.

    return $self->{workflow}->results();

} # End of results

#===============================================================================
# Function: worst_result
#	Gets worst result from workflow
#
#	Returns:
#		Worst NATE result of workflow
#===============================================================================
sub worst_result {
    my ($self, %args) = @_;
    return $self->{workflow}->worst_result();
} # End of worst_result

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/core/Workflow>

=cut

1;
