# Copyright (c) 2012-2020 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Aggregate Operations module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Features::AggrOps
# Wrapper for NACL aggregate operations with Misfit messaging
#

package San::FIT::Misfit::Features::AggrOps;

# Compiler Directives
use strict;
use warnings;

# Import Libraries
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::Cluster;
use NACL::C::Aggregate;
use San::FIT::Misfit::Core::Workflow;
use NACL::C::StorageAggregateWafliron;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);

# This package inherits methods and attributes from Misfit's Base class
use base qw( San::FIT::Misfit::Core::Base );

use constant {
    TIMEOUT => 360
};

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#
#   Parameters:
#       cluster - name of cluster to perform operation on.
#
#   Returns:
#       New AggrOps object
#
#   Example:
#>      my $AggrOps = San::FIT::Misfit::Features::AggrOps->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Features::AggrOps";
    my $interface = NACL::C::Node->find();

    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );

    my $node = undef;
    my $node_cli = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
        $node_cli = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
        $node_cli = NACL::C::Node->find();
    }

    $node->apiset_must(interface => 'ZAPI');
    $node->apiset();
    $self->{node} = $node;
    $self->{node_cli} = $node_cli;
    $self->{node_cli}->apiset_must(interface => 'CLI');
    $self->{node_cli}->apiset();
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#==============================================================================
# Description: Executes a specific aggregate operation
#==============================================================================

#==============================================================================
# Function: create
# Create a new aggregate.  Create will loop through the list of objects provided
# and create each aggregate with the specified settings.
#
# Parameters:
#   objects - Array of hashes containing information for creating each aggregate
#   background - Option to specify whether to perform each create
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the creates
#
# Aggregate create attributes:
#   aggregate - Name of aggregate (required)
#   diskcount - Number of disks to use in the aggregate (required)
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $AggrOps->create(
# >    objects    => [{
# >        aggregate => 'sanfit_test_aggr',
# >        diskcount => 3,
# >    ]},
# >    background => TRUE,
# >);
#

#==============================================================================

sub create {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $aggregates = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $node = &ParamsMgr::get_arg(
        'arg'      => "node",
        'required' => FALSE,
        'default'  => $self->{node},
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @aggregate_info = ();
    my @msg_data = ();
    foreach my $aggregate (@{$aggregates}) {
        my $runid = '';
        if (defined ($aggregate->{id})) {
            $runid = $aggregate->{id} . "_create";
        } else {
            $runid = $aggregate->{aggregate} . "_create";
        }
        push(@msg_data, {aggregate => $aggregate->{aggregate}});
        push (@aggregate_info, {
            codespec => "{" . $self->{id} . "::_create(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$aggregate},
                id    => $self->{id},
                node  => Clone::clone($node),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@aggregate_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Aggregate Create failed");
        return FALSE;
    }
} # End of create

#==============================================================================
# Function: wafliron_start
# Start wafliron on an aggregate
#
# Parameters:
#   objects - Array of hashes containing information for starting wafliron on
#       each aggregate
#   background - Option to specify whether to perform each start
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the starts
#
# Wafliron start attributes:
#   aggregate - Name of aggregate (required)
#   include-mirrors - indicate if snapmirrors need to be checked (optional).
#       TRUE (default) | FALSE
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $AggrOps->wafliron_start(
# >    objects    => [{
# >        aggregate => 'sanfit_test_aggr',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub wafliron_start {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $aggregates = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $node = &ParamsMgr::get_arg(
        'arg'      => "node",
        'required' => FALSE,
        'default'  => $self->{node_cli},
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @aggregate_info = ();
    my @msg_data = ();
    foreach my $aggregate (@{$aggregates}) {
        my $runid = '';
        if (defined ($aggregate->{id})) {
            $runid = $aggregate->{id} . "_wafliron_start";
        } else {
            $runid = $aggregate->{aggregate} . "_wafliron_start";
        }
        push(@msg_data, {aggregate => $aggregate->{aggregate}});
        push (@aggregate_info, {
            codespec => "{" . $self->{id} . "::_wafliron_start(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$aggregate},
                id    => $self->{id},
                node  => Clone::clone($node),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@aggregate_info,
        operation  => "wafliron_start",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Aggregate wafliron start failed");
        return FALSE;
    }
} # End of wafliron_start

#==============================================================================
# Function: wafliron_stop
# Stop wafliron on an aggregate
#
# Parameters:
#   objects - Array of hashes containing information for stopping wafliron on
#       each aggregate
#   background - Option to specify whether to perform each stop
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the stops
#
# Wafliron stop attributes:
#   aggregate - Name of aggregate (required)
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $AggrOps->wafliron_stop(
# >    objects    => [{
# >        aggregate => 'sanfit_test_aggr',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub wafliron_stop {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $aggregates = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $node = &ParamsMgr::get_arg(
        'arg'      => "node",
        'required' => FALSE,
        'default'  => $self->{node_cli},
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @aggregate_info = ();
    my @msg_data = ();
    foreach my $aggregate (@{$aggregates}) {
        my $runid = '';
        if (defined ($aggregate->{id})) {
            $runid = $aggregate->{id} . "_wafliron_stop";
        } else {
            $runid = $aggregate->{aggregate} . "_wafliron_stop";
        }
        push(@msg_data, {aggregate => $aggregate->{aggregate}});
        push (@aggregate_info, {
            codespec => "{" . $self->{id} . "::_wafliron_stop(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$aggregate},
                id    => $self->{id},
                node  => Clone::clone($node),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@aggregate_info,
        operation  => "wafliron_stop",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Aggregate wafliron stop failed");
        return FALSE;
    }
} # End of wafliron_stop

#==============================================================================
# PRIVATE METHODS
#==============================================================================

#==============================================================================
# Description: Subprocess responsible for creating aggregates
#==============================================================================
sub _create {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $nodes = $params->get('nodes');
    my $aggr = $params->get('aggregate');
    my $diskcount = $params->get('diskcount');
    my $disklist = $params->get('disklist');
    my $status = FALSE;

    $Log->comment("Global Paramset Currently Contains:\n" . Dumper($params));

    my %args = (
        aggregate => $aggr
    );
    if (defined($nodes)) {
        $args{nodes} = $nodes;
    };
    if (defined($diskcount)) {
        $args{diskcount} = $diskcount;
    };
    if (defined($disklist)) {
        $args{disklist} = $disklist;
    };

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::AggrOps',
            routing_key => '#',
            message     => {
                msg  => 'AggrOps.create.started',
                data => { %args, node => $node->{node} },
            }
        );
        NACL::C::Aggregate->create(
            %args,
            command_interface => $node
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::AggrOps',
            routing_key => '#',
            message     => {
                msg  => 'AggrOps.create.finished',
                data => { %args, node => $node->{node} },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed to create aggregates');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'AggrOps.create.' . $exception->text(),
                data => { %args, node => $node->{node} },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Aggregate create successful');
    } else {
        return NATE::Result::Fail->new('Aggregate create failed');
    }
} # End of _create

#==============================================================================
# Description: Subprocess responsible for starting wafliron on aggregates
#==============================================================================
sub _wafliron_start {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $include_mirrors = $params->get('include_mirrors', default => 'true');
    my $aggr = $params->get('aggregate');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::AggrOps',
            routing_key => '#',
            message     => {
                msg  => 'aggrops.wafliron_start.started',
                data => { aggregate => $aggr, node => $node->{node} },
            }
        );
        NACL::C::StorageAggregateWafliron->start(
            aggregate         => $aggr,
            'include-mirrors' => $include_mirrors,
            command_interface => $node,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::AggrOps',
            routing_key => '#',
            message     => {
                msg  => 'aggrops.wafliron_start.finished',
                data => { aggregate => $aggr, node => $node->{node} },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed to start wafliron on aggregate');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'aggrops.wafliron_start.' . $exception->text(),
                data => { aggregate => $aggr, node => $node->{node} },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Aggregate wafliron start successful');
    } else {
        return NATE::Result::Fail->new('Aggregate wafliron start failed');
    }
} # End of _wafliron_start

#==============================================================================
# Description: Subprocess responsible for stopping wafliron on aggregates
#==============================================================================
sub _wafliron_stop {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $include_mirrors = $params->get('include_mirrors', default => 'true');
    my $aggr = $params->get('aggregate');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::AggrOps',
            routing_key => '#',
            message     => {
                msg  => 'aggrops.wafliron_stop.started',
                data => { aggregate => $aggr, node => $node->{node} },
            }
        );
        NACL::C::StorageAggregateWafliron->stop(
            aggregate         => $aggr,
            command_interface => $node,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::AggrOps',
            routing_key => '#',
            message     => {
                msg  => 'aggrops.wafliron_stop.finished',
                data => { aggregate => $aggr, node => $node->{node} },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /wafliron is not currently active/){
            $Log->comment('wafliron is not currently active, continuing...');
            $status = TRUE;
        } elsif ($exception->text() =~ /Wafliron is finishing/) {
            $Log->comment('Stop aborted, Wafliron is finishing.  Continuing...');
            $status = TRUE;
        } else {
            $Log->comment('Failed to stop wafliron on aggregate');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg  => 'aggrops.wafliron_stop.' . $exception->text(),
                    data => { aggregate => $aggr, node => $node->{node} },
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Aggregate wafliron stop successful');
    } else {
        return NATE::Result::Fail->new('Aggregate wafliron stop failed');
    }
} # End of _wafliron_stop

#===============================================================================
# HANDLERS
#===============================================================================
sub _disruptive_handler {
    my ($self, $queue) = @_;

    $self->broadcast_critical(message => 'stop');
    Tharn::logresult("PASS", msg => "Handler completed");

    return TRUE;
} # End of _disruptive_handler

1;

#===============================================================================
# Message: aggrops.create.started
#   Exchange:
#       San::FIT::Misfit::Features::AggrOps
#
#   Message Structure:
# >     {
# >         "msg": "aggrops.create.started",
# >         "data" : {
# >             "aggregate": "sanfit_test_aggr"
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: aggrops.create.finished
#   Exchange:
#       San::FIT::Misfit::Features::AggrOps
#
#   Message Structure:
# >     {
# >         "msg": "aggrops.create.finished",
# >         "data" : {
# >             "aggregate": "sanfit_test_aggr"
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: aggrops.create.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "aggrops.create.Error Text",
# >         "data" : {
# >             "aggregate": "sanfit_test_aggr"
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: aggrops.wafliron_start.started
#   Exchange:
#       San::FIT::Misfit::Features::AggrOps
#
#   Message Structure:
# >     {
# >         "msg": "aggrops.wafliron_start.started",
# >         "data" : {
# >             "aggregate": "sanfit_test_aggr"
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: aggrops.wafliron_start.finished
#   Exchange:
#       San::FIT::Misfit::Features::AggrOps
#
#   Message Structure:
# >     {
# >         "msg": "aggrops.wafliron_start.finished",
# >         "data" : {
# >             "aggregate": "sanfit_test_aggr"
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: aggrops.wafliron_start.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "aggrops.wafliron_start.Error Text",
# >         "data" : {
# >             "aggregate": "sanfit_test_aggr"
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#===============================================================================

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/Misfit/features/AggrOps>

=cut

1;
