# Copyright (c) 2001-2013 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Ctran Synthetic Client Operations module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Features::CtranOps
# Ctran Synthetic Client Operations module for Misfit

package San::FIT::Misfit::Features::CtranOps;

# Compiler Directives
use strict;
use warnings;

# Perl Libraries
use List::MoreUtils qw(zip);

# Imported modules
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NATE::ParamSet;
use NACL::CS::Node;
use NACL::C::Cluster;
use NACL::C::DebugSmdb;
use NACL::CS::DebugSmdb;
use NACL::C::SystemNode;
use San::FIT::Misfit::Core::Workflow;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);

# Inherit base class - Misfit Base
use base qw(San::FIT::Misfit::Core::Base);

# Main broker used by this module
my $MessageBroker;

# Constants
use constant {
    MAX_RETRIES_PROPOSAL_WAIT => 18, # Multiply by DELAY for total retry time.
    TRANSACTION_TYPE          => 'spp',
    FAS2K_PROP_LIMIT          => 8000,  # 16000 - Dev suggested
    FAS3K_PROP_LIMIT          => 8000,  # 32000 - Dev suggested
    FAS6K_PROP_LIMIT          => 64000, # 64000 - Dev suggested
    CTRAN_MODIFY              => 'debug smdb table ctran_ut modify ',
    CTRAN_SHOW                => 'debug smdb table ctran_ut show ',
    DELAY                     => 5,
};

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a Misfit Ctran Operations object
#
#   Parameters:
#       cluster - name of cluster to perform operation on.
#       (Defaults to testbed parameter CLUSTER_NAME)
#       proposal_count - number of proposals to generate per invocation to
#       each node. (Default: max for the type of platform)
#       transaction_type - the type of transactions to (Defaults to 'spp')
#
#
#   Returns:
#     New Ctran object
#
#   Example:
#     > my $CtranOps = San::FIT::Misfit::Features::CtranOps->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Features::CtranOps";

    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    $self->{proposal_count} = &ParamsMgr::get_arg(
        'arg'      => "proposal_count",
        'required' => FALSE,
        'default'  => undef,
    );
    $self->{transaction_type} = &ParamsMgr::get_arg(
        'arg'      => "transaction_type",
        'required' => FALSE,
        'default'  => TRANSACTION_TYPE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return {};
    }

    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }

    # Create the cluster interface for this test.
    if (defined ($cluster_name)) {
        $self->{cluster} = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $Log->comment("Cluster_Name parameter must be defined for this module.");
        return {};
    }
    unless (defined($self->{cluster})) {
        $Log->comment('Failed to create a cluster command interface object!');
        return {};
    }
    # Store the list of nodes from this cluster.
    @{$self->{nodes}} = $self->{cluster}->nodes();
    $Log->comment("Nodes in cluster: " . join(',', @{$self->{nodes}}));
    # In order to execute debug smdb commands we need an apiset for raw commands
    $self->{cluster_apiset} = $self->{cluster}->apiset(
        category  => "Node",
        interface => "CLI",
        set       => "CMode",
    );

    # Set the kenv bootargs to enable the synthetic client.
    $self->enable_ctran_ut();

    # Get/Load proposal limits for this cluster
    $self->{proposal_count} = $self->get_proposal_limit();

    # Initialize the run_id counters
    $self->{join_id}    = 0;
    $self->{propose_id} = 0;
    $self->{leave_id}   = 0;

    $Log->exit();
    return $self;
} # End of new

#==============================================================================
# Function: join
#   Spawns CTRAN Synthetic client workloads concurrently on the list of nodes
#   specified.
#
#   Parameters:
#     nodes - Arrayref of node names to start the join proposals on.
#     (Defaut: all nodes)
#     transaction_type - The type of ctran transaction (Default: spp).
#     proposal_count - The number join proposals to generate per node.
#     (Default: platform maximum)
#     background - Flag to specify running the join in parallel to each node
#     or serially (Default: TRUE)
#     iterations - number of times to run the join on each node. (Default:1)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#   Example:
#     >  my $status = $CtranOps->join(
#     >      nodes          => ['node1','node2'],
#     >      proposal_count => 16000,
#     > );
#==============================================================================
sub join {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my @nodes = &ParamsMgr::get_arg(
        'arg'      => "nodes",
        'required' => FALSE,
        'default'  => $self->{nodes},
    );
    my $transaction = &ParamsMgr::get_arg(
        'arg'      => "transaction_type",
        'required' => FALSE,
        'default'  => $self->{transaction_type},
    );
    my $proposal_count = &ParamsMgr::get_arg(
        'arg'      => "proposal_count",
        'required' => FALSE,
        'default'  => $self->{proposal_count},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    $Log->comment("Executing join on " . join(',',@nodes));
    my @ctran_procs = ();
    my @msg_data = ();
    foreach my $node (@nodes) {
        push(@msg_data,
            {
                node           => $node,
                transaction    => $transaction,
                proposal_count => $proposal_count,
            }
        );
        push(@ctran_procs, {
            codespec => "{" . $self->{id} . "::_ctran_ut_modify(); }",
            runid    => sprintf("%03d", $self->{join_id}).
                        "_ctran_join_workload_on_" . $node,
            params   => {
                action           => 'join',
                filer_id         => $node,
                transaction_type => $transaction,
                proposal_count   => $proposal_count,
                apiset           => Clone::clone($self->{cluster_apiset}),
            },
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@ctran_procs,
        operation  => 'join',
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );
    $self->{join_id}++;
    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Ctran Join Proposals failed.");
        return FALSE;
    }
} # End join

#==============================================================================
# Function: propose
#   Spawns CTRAN Synthetic client propose workloads concurrently on the list of
#   nodes specified.
#
#   Parameters:
#     nodes - Arrayref of node names to start the propose proposals on.
#     (Defaut: all nodes)
#     transaction_type - The type of ctran transaction (Default: spp).
#     proposal_count - The number join proposals to generate per node.
#     (Default: platform maximum)
#     background - Flag to specify running the propose in parallel to each node
#     or serially (Default: TRUE)
#     iterations - number of times to run the propose on each node. (Default:1)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#   Example:
#     >  my $status = $CtranOps->join(
#     >      nodes          => ['node1','node2'],
#     >      proposal_count => 16000,
#     > );
#==============================================================================
sub propose {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my @nodes = &ParamsMgr::get_arg(
        'arg'      => "nodes",
        'required' => FALSE,
        'default'  =>  $self->{nodes},
    );
    my $transaction = &ParamsMgr::get_arg(
        'arg'      => "transaction_type",
        'required' => FALSE,
        'default'  => $self->{transaction_type},
    );
    my $proposal_count = &ParamsMgr::get_arg(
        'arg'      => "proposal_count",
        'required' => FALSE,
        'default'  => $self->{proposal_count},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    $Log->comment("Executing propose on " . join(',',@nodes));
    my @ctran_procs = ();
    my @msg_data = ();
    foreach my $node (@nodes) {
        push(@msg_data,
            {
                node             => $node,
                transaction_type => $transaction,
                proposal_count   => $proposal_count,
            }
        );
        push(@ctran_procs, {
            codespec => "{" . $self->{id} . "::_ctran_ut_modify(); }",
            runid    => sprintf("%03d", $self->{propose_id}).
                        "_ctran_propose_workload_on_" . $node,
            params   => {
                action           => 'propose',
                filer_id         => $node,
                transaction_type => $transaction,
                proposal_count   => $proposal_count,
                apiset           => Clone::clone($self->{cluster_apiset}),
            },
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@ctran_procs,
        operation  => 'propose',
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );
    $self->{propose_id}++;
    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Ctran Propose Proposals failed.");
        return FALSE;
    }
} # End propose

#==============================================================================
# Function: leave
#   Spawns CTRAN Synthetic client leave workloads concurrently on the list of
#   nodes specified.
#
#   Parameters:
#     nodes - Arrayref of node names to start the leave proposals on.
#     (Defaut: all nodes)
#     transaction_type - The type of ctran transaction (Default: spp).
#     proposal_count - The number join proposals to generate per node.
#     (Default: platform maximum)
#     background - Flag to specify running the leave in parallel to each node
#     or serially (Default: TRUE)
#     iterations - number of times to run the leave on each node. (Default:1)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#   Example:
#     >  my $status = $CtranOps->leave(
#     >      nodes          => ['node1','node2'],
#     >      proposal_count => 16000,
#     > );
#==============================================================================
sub leave {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my @nodes = &ParamsMgr::get_arg(
        'arg'      => "nodes",
        'required' => FALSE,
        'default'  =>  $self->{nodes},
    );
    my $transaction = &ParamsMgr::get_arg(
        'arg'      => "transaction",
        'required' => FALSE,
        'default'  => $self->{transaction_type},
    );
    my $proposal_count = &ParamsMgr::get_arg(
        'arg'      => "proposal_count",
        'required' => FALSE,
        'default'  => $self->{proposal_count},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    $Log->comment("Executing leave on " . join(',',@nodes));
    my @ctran_procs = ();
    my @msg_data = ();
    foreach my $node (@nodes) {
        push(@msg_data,
            {
                node             => $node,
                transaction_type => $transaction,
                proposal_count   => $proposal_count,
            }
        );
        push(@ctran_procs, {
            codespec => "{" . $self->{id} . "::_ctran_ut_modify(); }",
            runid    => sprintf("%03d", $self->{leave_id}) .
                        "_ctran_leave_workload_on_" . $node,
            params   => {
                action   => 'leave',
                filer_id         => $node,
                transaction_type => $transaction,
                proposal_count   => $proposal_count,
                apiset           => Clone::clone($self->{cluster_apiset}),
            },
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@ctran_procs,
        operation  => 'leave',
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );
    $self->{leave_id}++;
    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Ctran Leave Proposals failed.");
        return FALSE;
    }
} # End leave

#==============================================================================
# Function: enable_ctran_ut
#   Enables the synthetic client using kenv bootargs from the systemshell.
#
#   Parameters:
#     nodes - Arrayref of node names to start the join proposals on.
#             (Defaut: all nodes)
#
#   Returns:
#     Nothing (NACL Exception should fire if a Failure occurs.)
#
#
#   Example:
#     >  my $status = $CtranOps->enable_ctran_ut(
#     >      nodes          => ['node1','node2'],
#     > );
#==============================================================================
sub enable_ctran_ut {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my @nodes = &ParamsMgr::get_arg(
        'arg'      => "nodes",
        'required' => FALSE,
        'default'  =>  $self->{nodes},
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my $status = TRUE;
    my $node_obj = undef;
    foreach my $node (@nodes) {
        try {
            $node_obj = NACL::C::SystemNode->new(node => $node);
            $node_obj->systemshell(
                command_interface => $node_obj,
                node              => $node,
                command           => '"kenv bootarg.kcs.ctran.ut=1"',
            );
        } otherwise {
            my $e = shift();
            $status = FALSE;
            $e->throw();
        }
    }

    # Attempt to disconnect this nodes connection gracefully before it goes
    # out of scope.
    if (defined($node_obj)) {
        my $apiset = $node_obj->apiset();
        $apiset->cleanup_APISet();
    }

    return $status;
} # End enable_ctran_ut

#==============================================================================
# Function: get_ctran_ut_proposal_limit
#   Determines the proposal limit for nodes in the cluster and returns it.
#
#   Parameters:
#     None.
#
#   Returns:
#     Integer value representing the proposal limit for the cluster.  The limit
#     will be either the proposal count (if the user defined it in the new
#     constructor) or the default max for the platforms in the cluster.
#     0 on error.
#
#   Example:
#     >  my $num_props = $CtranOps->get_ctran_ut_proposal_limit();
#
#==============================================================================
sub get_proposal_limit {
    my ($self, %args) = @_;
    $Log->enter();

    my $num_props = 0;
    if (defined($self->{proposal_count})) {
        $num_props = $self->{proposal_count};
    } else {
        # Get the model number of a node in the cluster
        my $node = NACL::CS::Node->fetch(command_interface => $self->{cluster});
        if ($node->{model} =~ /FAS2/) {
            $num_props = FAS2K_PROP_LIMIT;
        } elsif ($node->{model} =~ /FAS3/) {
            $num_props = FAS3K_PROP_LIMIT;
        } elsif ($node->{model} =~ /FAS6/) {
            $num_props = FAS6K_PROP_LIMIT;
        } else {
            $Log->warn("This platfrom model is unsupported.");
            $num_props = 0;
        }
    }

    return $num_props;
} # End get_ctran_ut_proposal_limit

#===============================================================================
#  PRIVATE METHODS
#===============================================================================
#==============================================================================
# Description: This method will execute the debug command to start a proposal
#              workload using the synthetic client.  It will wait for it to
#              finish.
#==============================================================================
sub _ctran_ut_modify {
    my (%args) = @_;
    my $params = NATE::ParamSet->get_global();

    &ParamsMgr::process_args(%args);
    my $action = &ParamsMgr::get_arg(
        'arg'      => "action",
        'required' => FALSE,
        'default'  => $params->get('action'),
    );
    my $filer_id = &ParamsMgr::get_arg(
        'arg'      => "filer_id",
        'required' => FALSE,
        'default'  => $params->get('filer_id'),
    );
    my $transaction = &ParamsMgr::get_arg(
        'arg'      => "transaction_type",
        'required' => FALSE,
        'default'  => $params->get('transaction_type'),
    );
    my $proposal_count = &ParamsMgr::get_arg(
        'arg'      => "proposal_count",
        'required' => FALSE,
        'default'  => $params->get('proposal_count'),
    );
    my $apiset = &ParamsMgr::get_arg(
        'arg'      => "apiset",
        'required' => FALSE,
        'default'  => $params->get('apiset'),
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    my $status = TRUE;
    try {
        $apiset->api_reconnect_session();
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::CtranSyntheticOps',
            routing_key => '#',
            message     => {
                msg  => "CtranSyntheticOps.$action.started",
                data => {
                    node             => $filer_id,
                    transaction_type => $transaction,
                    proposal_count   => $proposal_count,
                },
            },
        );
        my $result = $apiset->execute_raw_command(
            command => CTRAN_MODIFY .
                       "-filerid $filer_id " .
                       "-action $action " .
                       "-ut $transaction " .
                       "-count $proposal_count",
        );
        if ($result =~ /Error:/) {
            $status = FALSE;
            NATE::BaseException->throw($result);
            return; # Since try{} is an anonymous subroutine, we can just exit.
        }
        $status &= _wait_for_proposals_complete(
            apiset   => $apiset,
            filer_id => $filer_id,
            action   => $action,
        );
        # Guarantee we close the connections in time.
        $apiset->cleanup_APISet();
    } otherwise {
        my $exception = shift;
        $Log->comment("Ctran $action proposals failed");
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg     => "CtranSyntheticOps.$action." . $exception->text(),
                data => {
                    node             => $filer_id,
                    transaction_type => $transaction,
                    proposal_count   => $proposal_count,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::CtranSyntheticOps',
            routing_key => '#',
            message     => {
                msg  => "CtranSyntheticOps.$action.finished",
                data => {
                    node             => $filer_id,
                    transaction_type => $transaction,
                    proposal_count   => $proposal_count,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new("Ctran $action proposals successful."));
    } else {
        $Log->log(NATE::Result::Fail->new("Ctran $action proposals failed."));
    }
} # End _ctran_ut_modify

#==============================================================================
# Description: This method waits for the ctran ut proposals to finish on a
#              particular node.
#==============================================================================
sub _wait_for_proposals_complete {
    my (%args) = @_;
    my $params = NATE::ParamSet->get_global();
    &ParamsMgr::process_args(%args);
    my $filer_id = &ParamsMgr::get_arg(
        'arg'      => "filer_id",
        'required' => TRUE,
    );
    my $action = &ParamsMgr::get_arg(
        'arg'      => "action",
        'required' => TRUE,
    );
    my $transaction = &ParamsMgr::get_arg(
        'arg'      => "transaction",
        'required' => FALSE,
        'default'  => $params->get('transaction_type'),
    );
    my $apiset = &ParamsMgr::get_arg(
        'arg'      => "apiset",
        'required' => FALSE,
        'default'  => $params->get('apiset'),
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $iteration = 0;
    while ($iteration < MAX_RETRIES_PROPOSAL_WAIT) {
        my $output = $apiset->execute_raw_command(
            command => CTRAN_SHOW .
                       "-filerid $filer_id " .
                       "-ut *$transaction* ",
        );
        if ($output =~ /Error:/) {
            return FALSE;
        }
        my @ctran_tables = _format_output(output => $output);
        $Log->debug("Debug smdb fetch returned: " , Dumper(@ctran_tables));
        if ($ctran_tables[0]->{pending} == 0) {
            return TRUE;
        }
        $iteration++;
        sleep(DELAY);
    }

    NATE::BaseException->throw(
        "The $action operation did not complete in " .
        DELAY * $iteration . " seconds."
    );
    return FALSE;
}

#==============================================================================
# Description: This method will return an array/ref of hashrefs for each row of
#              data keyed by the field names.
#==============================================================================
sub _format_output {
    my (%args) = @_;
    my $output = $args{output} if defined $args{output};

    return wantarray ? () : [] unless length $output;

    my @lines = split('\n', $output);
    # The first line contains the field names.
    my @fields = split("X;X", shift(@lines));
    shift(@lines); # Discard the second row, it is useless.

    my @records = ();
    foreach my $line (@lines) {
        my @values = split("X;X", $line);
        next unless scalar(@values); # Skip empty lines
        my %hash = zip @fields, @values; # Magic function to enter key,value
        push(@records, \%hash);
    }

    return wantarray ? @records : \@records;
}

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://http://misfitapi.gdl.englab.netapp.com/files/Features/CtranOps-pm.html>

=cut

1;
