# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Storage Efficiency module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Features::Efficiency
# Wrapper for NACL storage efficiency with Misfit messaging

package San::FIT::Misfit::Features::Efficiency;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use Clone;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::Volume;
use NACL::C::Cluster;
use NACL::C::VolumeEfficiency;
use NACL::STask::VolumeEfficiency;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;
use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

use constant {
    EFFICIENCY_WAIT => 180,
    TIMEOUT         => 360,
};

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a Misfit efficiency object
#
#   Parameters:
#       cluster - name of cluster to perform operation on.
#
#   Returns:
#     New Misfit Efficiency object
#
#   Example:
#     > my $Efficiency = San::FIT::Misfit::Features::Efficiency->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id}   = "San::FIT::Misfit::Features::Efficiency";
    my $interface = NACL::C::Node->find();
    $interface->apiset_must(interface => 'ZAPI');

    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );

    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }

    $node->apiset_must(interface => 'ZAPI');
    $node->apiset();
    $self->{node} = $node;
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#==============================================================================
# Function: on
#   Turns on storage efficiency. On will loop through the list of objects provided
#   and enable efficiency with the specified parameters. Provides a wrapper
#   to NACL::STask::VolumeEfficiency->on function
#
#   Parameters:
#     objects - Array of hashes containing information for volumes
#     background - Option to specify whether to perform each 'on'
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to turn efficiency on
#     filter - If objects is not passed and a count is specified, use this filter
#         to filter volumes based on attributes
#
#     Efficiency attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $Efficiency->on(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub on {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = {$volume->{id}};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_on(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "on",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Efficiency on failed");
        return FALSE;
    }
} # End of on

#==============================================================================
# Function: off
#   Turns off storage efficiency. Off will loop through the list of objects provided
#   and turn off efficiency with the specified parameters. Provides a wrapper
#   to NACL::STask::VolumeEfficiency->off function
#
#   Parameters:
#     objects - Array of hashes containing information for volumes
#     background - Option to specify whether to perform each 'off'
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to turn efficiency off
#     filter - If objects is not passed and a count is specified, use this filter
#         to filter volumes based on attributes
#
#     Efficiency attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $Efficiency->off(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub off {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = {$volume->{id}};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_off(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "off",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Efficiency off failed");
        return FALSE;
    }
} # End of off

#==============================================================================
# Function: start
#   Start storage efficiency. Start will loop through the list of objects provided
#   and start efficiency with the specified parameters. Provides a wrapper
#   to NACL::STask::VolumeEfficiency->start function
#
#   Parameters:
#     objects - Array of hashes containing information for volumes
#     background - Option to specify whether to perform each 'start'
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to start scanning
#     filter - If objects is not passed and a count is specified, use this filter
#         to filter volumes based on attributes
#
#     Efficiency attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#       'dedupe' - TRUE or FALSE to start this operation. Defaults to TRUE
#       'compression' - TRUE or FALSE to start this operation. Defaults to TRUE
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $Efficiency->start(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >        'dedupe'    => FALSE
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub start {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = {$volume->{id}};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_start(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "start",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Efficiency start failed");
        return FALSE;
    }
} # End of start

#==============================================================================
# Function: config
#   Set efficiency attributes. Configure will loop through the list of objects
#   provided and configure efficiency and dedupe. Provides a wrapper
#   to NACL::STask::VolumeEfficiency->config function
#
#   Parameters:
#     objects - Array of hashes containing information for volumes
#     background - Option to specify whether to perform each 'configure'
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to configure
#     filter - If objects is not passed and a count is specified, use this filter
#         to filter volumes based on attributes
#
#   Efficiency attributes:
#     'volume' - Volume name (required)
#     'vserver' - Vserver name (required)
#     'inline-compression' - TRUE or FALSE to set this mode. Defaults to TRUE
#     'backhround-compression' - TRUE or FALSE to set this mode. Defaults to TRUE
#     'compression' - TRUE or FALSE to set this mode. Defaults to TRUE
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $Efficiency->config(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >        'inline-compression' => TRUE,
#     >        'background-compression' => TRUE,
#     >        'compression' => TRUE,
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub config {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = {$volume->{id}};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_config(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "config",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Efficiency config failed");
        return FALSE;
    }
} # End of config

#==============================================================================
# Function: stop
#   Stop storage efficiency. Stop will loop through the list of objects provided
#   and stop efficiency with the specified parameters. Provides a wrapper
#   to NACL::STask::VolumeEfficiency->start function
#
#   Parameters:
#     objects - Array of hashes containing information for volumes
#     background - Option to specify whether to perform each 'stop'
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to stop scanning
#     filter - If objects is not passed and a count is specified, use this filter
#         to filter volumes based on attributes
#
#     Efficiency attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $Efficiency->stop(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub stop {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = {$volume->{id}};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_stop(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "stop",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Efficiency stop failed");
        return FALSE;
    }
} # End of stop

#==============================================================================
# Function: check
#   Check storage efficiency. Check will loop through the list of objects provided
#   and check efficiency with the specified parameters. Provides a wrapper
#   to NACL::STask::VolumeEfficiency->check function
#
#   Parameters:
#     objects - Array of hashes containing information for volumes
#     background - Option to specify whether to perform each 'check'
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to check
#     filter - If objects is not passed and a count is specified, use this filter
#         to filter volumes based on attributes
#
#   Efficiency attributes:
#     'volume' - Volume name (required)
#     'vserver' - Vserver name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $Efficiency->check(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub check {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = {$volume->{id}};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_check(); }",
            runid => $runid,
            background => FALSE,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "check",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Efficiency check failed");
        return FALSE;
    }
} # End of check

#==============================================================================
# Function: stat
#   Gather efficiency stats. Stat will loop through the list of objects provided
#   and gather statistics. Provides a wrapper
#   to NACL::STask::VolumeEfficiency->stat function
#
#   Parameters:
#     objects - Array of hashes containing information for volumes
#     background - Option to specify whether to perform each 'stat'
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to gather stats from
#     filter - If objects is not passed and a count is specified, use this filter
#         to filter volumes based on attributes
#
#     Efficiency attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#       'dedupe' - TRUE or FALSE to start this operation. Defaults to TRUE
#       'compression' - TRUE or FALSE to start this operation. Defaults to TRUE
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#   > my $status = $Efficiency->stat(
#   >    objects    => [{
#   >        'volume'    => 'my_volume',
#   >        'vserver'   => 'vs1',
#   >    ]},
#   >    background => TRUE,
#   > );
#==============================================================================
sub stat {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = {$volume->{id}} . "_stat";
        } else {
            $runid = $volume->{volume} . "_stat";
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_stat(); }",
            runid => $runid,
            background => FALSE, # Stat is CLI only. Forcing it to be serial
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "stat",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Efficiency stat failed");
        return FALSE;
    }
} # End of stat

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Responsible for generating a list of volumes matching a criteria
#===============================================================================
sub _get_volume_list {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $operation = &ParamsMgr::get_arg(
        'arg'      => "operation",
        'required' => FALSE,
        'default'  => '',
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @volumes = ();
    my @all_volumes = NACL::C::Volume->find(
        command_interface => $self->{node},
        filter => $filter
    );

    if ($count =~ /^all$/i) {
        $count = scalar(@all_volumes);
    }

    for (my $i = 0; $i < $count; $i++) {
        my $vserver = $all_volumes[$i]->get_vserver();
        $volumes[$i] =  {
            vserver  => $vserver->name(),
            volume   => $all_volumes[$i]->volume(),
            %{$params}
        };
    }

    return wantarray ? @volumes : \@volumes;
} # End of _get_volume_list

#===============================================================================
# Description: Turn on storage efficiency, compression and dedupe
#===============================================================================
sub _on {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.on.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        NACL::STask::VolumeEfficiency->on(
            command_interface => $node,
            volume            => $volume,
            vserver           => $vserver,
            ignore_enabled    => TRUE,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.on.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() !~ /Operation has already been enabled/i) {
            $Log->error('Failed to turn on efficiency');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'Efficiency.on.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume
                    }
                }
            );
            NATE::BaseException->throw($exception);
        } else {
            $status = TRUE;
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Efficiency on');
    } else {
        return NATE::Result::Fail->new('Efficiency on failed');
    }
} # End of _on

#===============================================================================
# Description: Turn off Storage Efficiency
#===============================================================================
sub _off {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.off.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        NACL::STask::VolumeEfficiency->off(
            command_interface => $node,
            volume            => $volume,
            vserver           => $vserver,
            ignore_disabled   => TRUE,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.off.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() !~ /Operation has already been disabled/i) {
            $Log->error('Failed to turn off efficiency');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'Efficiency.off.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume
                    }
                }
            );
            NATE::BaseException->throw($exception);
        } else {
            $status = TRUE;
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Efficiency started');
    } else {
        return NATE::Result::Fail->new('Efficiency start failed');
    }
} # End of _off

#===============================================================================
# Description: Start volume efficiency
#===============================================================================
sub _start {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $compression = $params->get('compression');
    my $dedupe = $params->get('dedupe');
    my $scan = $params->get('scan');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.start.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        NACL::STask::VolumeEfficiency->start(
            command_interface => $node,
            volume            => $volume,
            vserver           => $vserver,
            nacltask_wait     => 1,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.start.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /No matching volume efficiency/i) {
            $Log->comment("Efficiency has started but the status is unknown");
            $status = TRUE;
        } else {
            $Log->error('Failed to start efficiency');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'Efficiency.start.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Efficiency started');
    } else {
        return NATE::Result::Fail->new('Efficiency start failed');
    }
} # End of _start

#===============================================================================
# Description: Configures storage efficiency on volumes
#===============================================================================
sub _config {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $compression = $params->get('compression');
    my $inline_compression = $params->get('inline-compression');
    my $background_compression = $params->get('background-compression');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
             message => {
                msg => 'Efficiency.config.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        NACL::STask::VolumeEfficiency->config(
            command_interface    => $node,
            volume               => $volume,
            vserver              => $vserver,
            'inline-compression' => $inline_compression,
            'background-compression' => $background_compression,
            compression => $compression,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
             message => {
                msg => 'Efficiency.config.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->error('Failed to config efficiency');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message => {
                msg => 'Efficiency.config.' . $exception->text(),
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Efficiency configured');
    } else {
        return NATE::Result::Fail->new('Efficiency config failed');
    }
} # End of _config

#===============================================================================
# Description: Stop volume efficiency
#===============================================================================
sub _stop {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.stop.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        NACL::STask::VolumeEfficiency->stop(
            command_interface => $node,
            volume            => $volume,
            vserver           => $vserver,
            all               => 'true'
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.stop.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        return NATE::Result::Pass->new('Efficiency stopped');
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /Operation is currently idle/i) {
            $MessageBroker->publish_message(
                exchange    => 'San::FIT::Misfit::Features::Efficiency',
                routing_key => '#',
                message => {
                    msg => 'Efficiency.stop.finished',
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume
                    }
                }
            );
            $status = TRUE;
        } else {
            $Log->error('Failed to stop efficiency');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'Efficiency.stop.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume
                    }
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        return NATE::Result::Pass->new('Efficiency stopped');
    } else {
        return NATE::Result::Fail->new('Efficiency stop failed');
    }
} # End of _stop

#===============================================================================
# Description: Check volume efficiency
#===============================================================================
sub _check {
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status = FALSE;

    try {
        # This method supports CLI only.
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.check.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        NACL::STask::VolumeEfficiency->check(
            command_interface => $node,
            volume            => $volume,
            vserver           => $vserver,
        );

        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /Operation is currently active/){
            sleep(EFFICIENCY_WAIT);
            $Log->comment(
                'Efficiency operation currently in progress... retrying'
            );
            NACL::STask::VolumeEfficiency->check(
                command_interface => $node,
                volume            => $volume,
                vserver           => $vserver,
            );
            $status = TRUE;
        } else {
            $Log->error('Failed to check efficiency');
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message => {
                    msg => 'Efficiency.check.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume
                    }
                }
            );
            $status = FALSE;
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.check.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        return NATE::Result::Pass->new('Efficiency checked');
    } else {
        return NATE::Result::Fail->new('Efficiency check failed');
    }
} # End of _check

#===============================================================================
# Description: Gather storage efficiency statistics
#===============================================================================
sub _stat{
    my ($class) = @_;

    # Global params
    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status = FALSE;

    try {
        # This method supports CLI only.
        $node->apiset_must(interface => 'CLI');
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.stat.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                }
            }
        );
        NACL::C::VolumeEfficiency->stat(
            command_interface => $node,
            filter => {
                volume => $volume,
                vserver => $vserver
            },
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Efficiency',
            routing_key => '#',
            message => {
                msg => 'Efficiency.stat.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                }
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->error('Failed to get statistics');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message => {
                msg => 'Efficiency.stat.' . $exception->text(),
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume
                }
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Efficiency stat passed');
    } else {
        return NATE::Result::Fail->new('Efficiency stat failed');
    }

} # End of _stat

#===============================================================================
# MESSAGE CATALOG
#===============================================================================

#===============================================================================
# Message: efficiency.on.started
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.on.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.on.finished
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.on.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.on.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.on.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.off.started
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.off.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.off.finished
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.off.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.off.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.off.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.start.started
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.start.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.start.finished
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.start.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.start.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.start.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.config.started
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.config.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.config.finished
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.config.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.config.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.config.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.stop.started
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.stop.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.stop.finished
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.stop.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.stop.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.stop.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.check.started
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.check.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.check.finished
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.check.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.check.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.check.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.stat.started
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.stat.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.stat.finished
#   Exchange:
#       San::FIT::Misfit::Features::Efficiency
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.stat.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
# Message: efficiency.stat.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "efficiency.stat.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >          }
# >     }
#
#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/misfit/feature/Efficiency>

=cut

1;