# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Igroup Operations module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

package San::FIT::Misfit::Features::Igroup;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::LunIgroup;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;

# Inherit base class - Misfit Base
use base qw(San::FIT::Misfit::Core::Base);

# Main broker used by this module
my $MessageBroker;

#===============================================================================
# Description: Igroup Operations constructor
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Features::Igroup";
    $self->{node} = NACL::C::Node->find();
    $self->{node}->apiset_must(interface => 'ZAPI');
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#===============================================================================
#  PUBLIC METHODS
#===============================================================================
#===============================================================================
# Description: Creates different types and numbers  of Igroups that are specified
#   by the user.
#===============================================================================
sub create {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $igroups = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @igroup_info = ();
    my @msg_data = ();
    foreach my $igroup (@{$igroups}) {
        my $runid = '';
        if (defined ($igroup->{id})) {
            $runid = $igroup->{id} . "_create";
        } else {
            $runid = $igroup->{igroup} . "_create";
        }
        push(@msg_data, {vserver => $igroup->{vserver},
                         igroup => $igroup->{igroup}});
        push (@igroup_info, {
            codespec => "{" . $self->{id} . "::_create(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$igroup},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@igroup_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Igroup create failed");
        return FALSE;
    }
} # End of create

#===============================================================================
# Description: Binds different types and numbers  of Igroups that are specified
#   by the user.
#===============================================================================
sub bind {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $igroups = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @igroup_info = ();
    my @msg_data = ();
    foreach my $igroup (@{$igroups}) {
        my $runid = '';
        if (defined ($igroup->{id})) {
            $runid = $igroup->{id} . "_bind";
        } else {
            $runid = $igroup->{igroup} . "_bind";
        }
        push(@msg_data, {vserver => $igroup->{vserver},
                         igroup => $igroup->{igroup}});
        push (@igroup_info, {
            codespec => "{" . $self->{id} . "::_bind(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$igroup},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@igroup_info,
        operation  => "bind",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Igroup bind failed");
        return FALSE;
    }
} # End of bind

#===============================================================================
# Description: Removes different types and numbers  of Igroups that are specified
#   by the user.
#===============================================================================
sub remove {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $igroups = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @igroup_info = ();
    my @msg_data = ();
    foreach my $igroup (@{$igroups}) {
        my $runid = '';
        if (defined ($igroup->{id})) {
            $runid = $igroup->{id} . "_remove";
        } else {
            $runid = $igroup->{igroup} . "_remove";
        }
        push(@msg_data, {vserver => $igroup->{vserver},
                         igroup => $igroup->{igroup}});
        push (@igroup_info, {
            codespec => "{" . $self->{id} . "::_remove(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$igroup},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@igroup_info,
        operation  => "remove",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Igroup remove failed");
        return FALSE;
    }
} # End of remove

#===============================================================================
# Description: Adds different types and numbers  of Igroups that are specified
#   by the user.
#===============================================================================
sub add {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $igroups = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @igroup_info = ();
    my @msg_data = ();
    foreach my $igroup (@{$igroups}) {
        my $runid = '';
        if (defined ($igroup->{id})) {
            $runid = $igroup->{id} . "_add";
        } else {
            $runid = $igroup->{igroup} . "_add";
        }
        push(@msg_data, {vserver => $igroup->{vserver},
                         igroup => $igroup->{igroup}});
        push (@igroup_info, {
            codespec => "{" . $self->{id} . "::_add(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$igroup},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@igroup_info,
        operation  => "add",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Igroup add failed");
        return FALSE;
    }
} # End of add

#===============================================================================
# Description: Unbinds different types and numbers  of Igroups that are specified
#   by the user.
#===============================================================================
sub unbind {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $igroups = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @igroup_info = ();
    my @msg_data = ();
    foreach my $igroup (@{$igroups}) {
        my $runid = '';
        if (defined ($igroup->{id})) {
            $runid = $igroup->{id} . "_unbind";
        } else {
            $runid = $igroup->{igroup} . "_unbind";
        }
        push(@msg_data, {vserver => $igroup->{vserver},
                         igroup => $igroup->{igroup}});
        push (@igroup_info, {
            codespec => "{" . $self->{id} . "::_unbind(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$igroup},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@igroup_info,
        operation  => "unbind",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Igroup unbind failed");
        return FALSE;
    }
} # End of unbind

#===============================================================================
# Description: Deletes different types and numbers  of Igroups that are specified
#   by the user.
#===============================================================================
sub delete {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $igroups = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @igroup_info = ();
    my @msg_data = ();
    foreach my $igroup (@{$igroups}) {
        my $runid = '';
        if (defined ($igroup->{id})) {
            $runid = $igroup->{id} . "_delete";
        } else {
            $runid = $igroup->{igroup} . "_delete";
        }
        push(@msg_data,
             {vserver => $igroup->{vserver}, igroup => $igroup->{igroup}}
        );
        push (@igroup_info, {
            codespec => "{" . $self->{id} . "::_delete(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$igroup},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@igroup_info,
        operation  => "delete",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Igroup delete failed");
        return FALSE;
    }
} # End of delete

#===============================================================================
#  PRIVATE METHODS
#===============================================================================
#==============================================================================
# Description: Subprocess responsible for creating Igroups
#==============================================================================
sub _create {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $igroup = $params->get('igroup');
    my $ostype = $params->get('ostype');
    my $protocol = $params->get('protocol', default => 'mixed');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.create.started',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NACL::C::LunIgroup->create(
            command_interface => $node,
            vserver           => $vserver,
            igroup            => $igroup,
            ostype            => $ostype,
            protocol          => $protocol
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.create.finished',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,,
                },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed to create Igroups');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.create.' . $exception->text(),
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Igroup create successful');
    } else {
        return NATE::Result::Fail->new('Igroup create failed');
    }
} # End of _create

#==============================================================================
# Description: Subprocess responsible for removing initiators from an Igroup
#==============================================================================
sub _remove {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $igroup = $params->get('igroup');
    my $initiator = $params->get('initiator');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.remove.started',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NACL::C::LunIgroup->remove(
            command_interface => $node,
            vserver           => $vserver,
            igroup            => $igroup,
            initiator         => $initiator
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.remove.finished',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,,
                },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed to remove initiators');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.remove.' . $exception->text(),
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Igroup remove successful');
    } else {
        return NATE::Result::Fail->new('Igroup remove failed');
    }
} # End of _igroup_remove_codestring

#==============================================================================
# Description: Subprocess responsible for binding Igroups
#==============================================================================
sub _bind {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $igroup = $params->get('igroup');
    my $portset = $params->get('portset');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.bind.started',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NACL::C::LunIgroup->bind(
            command_interface => $node,
            vserver           => $vserver,
            igroup            => $igroup,
            portset           => $portset,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.bind.finished',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,,
                },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed to create Igroups');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.bind.' . $exception->text(),
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Igroup bind successful');
    } else {
        return NATE::Result::Fail->new('Igroup bind failed');
    }
} # End of _bind

#==============================================================================
# Description: Subprocess responsible for adding initiators to an Igroup
#==============================================================================
sub _add {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $igroup = $params->get('igroup');
    my $initiator = $params->get('initiator');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.add.started',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NACL::C::LunIgroup->add(
            command_interface => $node,
            vserver           => $vserver,
            igroup            => $igroup,
            initiator         => $initiator
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.add.finished',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed to add Igroups');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.add.' . $exception->text(),
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Igroup add successful');
    } else {
        return NATE::Result::Fail->new('Igroup add failed');
    }
} # End of _add

#==============================================================================
# Description: Subprocess responsible for unbinding Igroups
#==============================================================================
sub _unbind {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $igroup = $params->get('igroup');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.unbind.started',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NACL::C::LunIgroup->unbind(
            command_interface => $node,
            vserver           => $vserver,
            igroup            => $igroup,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.unbind.finished',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,,
                },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed to unbind Igroups');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.unbind.' . $exception->text(),
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Igroup unbind successful');
    } else {
        return NATE::Result::Fail->new('Igroup unbind failed');
    }
} # End of _unbind

#==============================================================================
# Description: Subprocess responsible for deleting Igroups
#==============================================================================
sub _delete {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $igroup = $params->get('igroup');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.delete.started',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NACL::C::LunIgroup->delete(
            command_interface => $node,
            vserver           => $vserver,
            igroup            => $igroup,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Igroup',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.delete.finished',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,,
                },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed to delete Igroups');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'Igroup.delete.' . $exception->text(),
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    igroup    => $igroup,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Igroup delete successful');
    } else {
        return NATE::Result::Fail->new('Igroup delete failed');
    }
} # End of _delete

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/Misfit/features/Igroup>

=cut

1;
