# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary LUN Operations module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Features::LunOps
# LUN Operations module for Misfit

package San::FIT::Misfit::Features::LunOps;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Lun;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::Cluster;
use NACL::CS::LunIgroup;
use San::FIT::Misfit::Core::Workflow;
use NATE::Exceptions::Argument qw(:try);
use San::FIT::Misfit::Services::Resource;
use NATE::Util::DataTypes qw(TRUE FALSE);

# Inherit base class - Misfit Base
use base qw(San::FIT::Misfit::Core::Base);

# Main broker used by this module
my $MessageBroker;

use constant {
    TIMEOUT => 300,
};

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a Misfit LUN operations object
#
#   Parameters:
#       cluster - name of cluster to perform operation on.
#
#   Returns:
#     New LUN Operations object
#
#   Example:
#     > my $LunOps = San::FIT::Misfit::Features::LunOps->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Features::LunOps";
    my $interface = NACL::C::Node->find();
    $interface->apiset_must(interface => 'ZAPI');

    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );

    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }

    $node->apiset_must(interface => 'ZAPI');
    $node->apiset();
    $self->{node} = $node;
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#==============================================================================
# Function: create
#   Create n LUNs. Create will loop through the list of objects provided
#   and create each LUN with the specified parameters. Provides a wrapper
#   to NACL::STask::Lun->create function
#
#   Parameters:
#     objects - Array of hashes containing information for creating each
#         LUN.
#     background - Option to specify whether to perform each create
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of LUNs to create
#
#     LUN create attributes:
#       'lun' - LUN name (required)
#       'vserver' - Vserver name (required)
#       'volume' - Volume name (required)
#       'size' - Size of the LUN (required)
#       'ostype' - String, linux, windows, etc (optional)
#       'space-reserved' - TRUE or FALSE
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $LunOps->create(
#     >      objects    => [{
#     >          'volume'    => 'my_volume',
#     >          'lun'       => 'my_lun',
#     >          'vserver'   => 'vs1',
#     >          'ostype'      => 'windows',
#     >          'size'      => '500M'
#     >      ]},
#     >      background => TRUE,
#     > );
#==============================================================================
sub create {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $luns = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @lun_info = ();
    my @msg_data = ();
    foreach my $lun (@{$luns}) {
        my $runid = '';
        if (defined ($lun->{id}) and exists($lun->{id})) {
            $runid = $lun->{id};
        } else {
            $runid = $lun->{lun};
        }
        push(@msg_data,
             {
                vserver => $lun->{vserver},
                volume => $lun->{volume},
                lun => $lun->{lun}
            }
        );
        push (@lun_info, {
            codespec => "{" . $self->{id} . "::_create(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$lun},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@lun_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Lun create failed");
        return FALSE;
    }

}

#==============================================================================
# Function: delete
#   Delete n LUNs. Delete will loop through the list of objects provided
#   and delete each LUN. Provides a wrapper to NACL::STask::Lun->delete()
#
# Parameters:
#   objects - Array of hashes containing information for deleting each
#       LUN
#   background - Option to specify whether to perform each delete
#       serially (FALSE) or parallely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   count - The number of LUNs to delete
#
#   LUN delete attributes:
#     'volume' - Volume name (required)
#     'vserver' - Vserver name (required)
#     'lun' - LUN name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $LunOps->delete(
#     >     objects    => [{
#     >        'volume'  => 'my_volume',
#     >        'vserver' => 'vs1',
#     >        'lun'     => 'my_lun'
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub delete {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $luns = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    my $prefix = &ParamsMgr::get_arg(
        'arg'      => "prefix",
        'required' => FALSE,
        'default'  => 'lun',
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @lun_info = ();
    my @msg_data = ();
    foreach my $lun (@{$luns}) {
        my $runid = '';
        if (defined ($lun->{id}) and exists($lun->{id})) {
            $runid = $lun->{id};
        } else {
            $runid = $lun->{lun};
        }
        push(@msg_data,
             {
                vserver => $lun->{vserver},
                volume => $lun->{volume},
                lun => $lun->{lun}
            }
        );
        push (@lun_info, {
            codespec => "{" . $self->{id} . "::_delete(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$lun},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@lun_info,
        operation  => "delete",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Lun delete failed");
        return FALSE;
    }
}

#==============================================================================
# Function: map
#   Map n LUNs. Map will loop through the list of objects provided
#   and map each LUN with the specified parameters. Provides a wrapper
#   to NACL::STask::Lun->map function
#
#   Parameters:
#     objects - Array of hashes containing information for mapping each
#         LUN.
#     background - Option to specify whether to perform each LUN map
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of LUNs to map
#
#     LUN map attributes:
#       'lun' - LUN name (required)
#       'vserver' - Vserver name (required)
#       'volume' - Volume name (required)
#       'igroup' - Igroup name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $LunOps->map(
#     >      objects    => [{
#     >          'volume'  => 'my_volume',
#     >          'lun'     => 'my_lun',
#     >          'vserver' => 'vs1',
#     >          'igroup'  => 'igroup_1',
#     >      ]},
#     >      background => TRUE,
#     > );
#==============================================================================
sub map {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $luns = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @lun_info = ();
    my @msg_data = ();
    foreach my $lun (@{$luns}) {
        my $runid = '';
        if (defined ($lun->{id}) and exists($lun->{id})) {
            $runid = $lun->{id};
        } else {
            $runid = $lun->{lun};
        }
        push(@msg_data,
             {
                vserver => $lun->{vserver},
                volume  => $lun->{volume},
                lun     => $lun->{lun},
                igroup  => $lun->{igroup},
            }
        );
        push (@lun_info, {
            codespec => "{" . $self->{id} . "::_map(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$lun},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@lun_info,
        operation  => "map",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Lun map failed");
        return FALSE;
    }
}

#==============================================================================
# Function: unmap
#   Unmap n LUNs. Unmap will loop through the list of objects provided
#   and unmap each LUN with the specified parameters. Provides a wrapper
#   to NACL::STask::Lun->umap function
#
#   Parameters:
#     objects - Array of hashes containing information for unmapping each
#         LUN.
#     background - Option to specify whether to perform each LUN unmap
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of LUNs to unmap
#
#     LUN unmap attributes:
#       'lun' - LUN name (required)
#       'vserver' - Vserver name (required)
#       'volume' - Volume name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $LunOps->unmap(
#     >      objects    => [{
#     >          'volume'  => 'my_volume',
#     >          'lun'     => 'my_lun',
#     >          'vserver' => 'vs1',
#     >      ]},
#     >      background => TRUE,
#     > );
#==============================================================================
sub unmap {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $luns = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @lun_info = ();
    my @msg_data = ();
    foreach my $lun (@{$luns}) {
        my $runid = '';
        if (defined ($lun->{id}) and exists($lun->{id})) {
            $runid = $lun->{id};
        } else {
            $runid = $lun->{lun};
        }
        push(@msg_data,
             {
                vserver => $lun->{vserver},
                volume  => $lun->{volume},
                lun     => $lun->{lun},
                igroup  => $lun->{igroup},
            }
        );
        push (@lun_info, {
            codespec => "{" . $self->{id} . "::_unmap(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$lun},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@lun_info,
        operation  => "unmap",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Lun unmap failed");
        return FALSE;
    }
}

#==============================================================================
# Function: resize
#   Resize n LUNs. Resize will loop through the list of objects provided
#   and resize each LUN with the specified parameters. Provides a wrapper
#   to NACL::C::Lun->resize function
#
#   Parameters:
#     objects - Array of hashes containing information for resizing each
#         LUN.
#     background - Option to specify whether to perform each LUN resize
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of LUNs to resize
#
#     LUN resize attributes:
#       'lun' - LUN name (required)
#       'vserver' - Vserver name (required)
#       'volume' - Volume name (required)
#       'size' - New size (Required unless percent is specified)
#       'percent' - Resize a LUN by 'percent' (Required if size is not specifiec)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $LunOps->resize(
#     >      objects    => [{
#     >          'volume'  => 'my_volume',
#     >          'lun'     => 'my_lun',
#     >          'vserver' => 'vs1',
#     >          'size'    => '300M'
#     >      ]},
#     >      background => TRUE,
#     > );
#==============================================================================
sub resize {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $luns = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @lun_info = ();
    my @msg_data = ();
    foreach my $lun (@{$luns}) {
        my $runid = '';
        if (defined($lun->{id}) and exists($lun->{id})) {
            $runid = $lun->{id};
        } else {
            $runid = $lun->{lun};
        }
        my $percent = 0;
        if (defined ($lun->{percent})) {
            $percent = $lun->{percent};
        }
        push(@msg_data,
             {
                vserver => $lun->{vserver},
                volume => $lun->{volume},
                lun => $lun->{lun}
            }
        );
        push (@lun_info, {
            codespec => "{" . $self->{id} . "::_resize(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$lun},
                percent => $percent,
                id      => $self->{id},
                node    => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@lun_info,
        operation  => "resize",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Lun resize failed");
        return FALSE;
    }
}

#==============================================================================
# Function: clone
#   Clone n LUNs. Clone will loop through the list of objects provided
#   and resize each LUN with the specified parameters. Provides a wrapper
#   to NACL::C::Lun->clone function
#
#   Parameters:
#     objects - Array of hashes containing information for cloning each
#         LUN.
#     background - Option to specify whether to perform each LUN clone
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of LUNs to clone
#
#     LUN clone attributes:
#       'lun' - LUN name (required)
#       'vserver' - Vserver name (required)
#       'volume' - Volume name (required)
#       'destination' - Destination for the LUN clone (required)
#       'Snapshot' - Snapshot name for cloning from snapshot (optional)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $LunOps->clone(
#     >      objects    => [{
#     >          'volume'  => 'my_volume',
#     >          'lun'     => 'my_lun',
#     >          'vserver' => 'vs1',
#     >          'destination' => '/vs0/vol1/lun_clone'
#     >      ]},
#     >      background => TRUE,
#     > );
#==============================================================================
sub clone {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $luns = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @lun_info = ();
    my @msg_data = ();
    foreach my $lun (@{$luns}) {
        my $runid = '';
        if (defined($lun->{id}) and exists($lun->{id})) {
            $runid = $lun->{id};
        } else {
            $runid = $lun->{lun};
        }
        push(@msg_data,
             {
                vserver => $lun->{vserver},
                volume => $lun->{volume},
                lun => $lun->{lun}
            }
        );
        push (@lun_info, {
            codespec => "{" . $self->{id} . "::_clone(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$lun},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@lun_info,
        operation  => "clone",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Lun clone failed");
        return FALSE;
    }
} # End of clone

#==============================================================================
# Description: Method used in swimlanes for creating array containing data
# for executing default lun lifecycles
#==============================================================================
sub create_test_luns {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vols = &ParamsMgr::get_arg(
        'arg'      => "vols",
        'required' => TRUE,
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $prefix = &ParamsMgr::get_arg(
        'arg'      => "prefix",
        'required' => FALSE,
        'default'  => 'lunops.lun_',
    );
    my $size = &ParamsMgr::get_arg(
        'arg'      => "size",
        'required' => FALSE,
        'default'  => '250M',
    );
    my $igroup = &ParamsMgr::get_arg(
        'arg'      => "igroup",
        'required' => FALSE,
        'default'  => undef,
    );
    my $ostype = &ParamsMgr::get_arg(
        'arg'      => "ostype",
        'required' => FALSE,
        'default'  => undef,
    );

    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return ();
    }

    my @ostypes = (
        'windows', 'linux', 'solaris', 'aix', 'hpux', 'vmware', 'xen', 'openvms'
    );
    if (defined $ostype) {
        @ostypes = ($ostype);
    }
    my @space_reserve = ('true', 'false');

    # This way we only initialize Resource once...
    my $resource = undef;
    if (defined($self->{resource})) {
        $resource = $self->{resource};
    } else {
        $self->{resource} = San::FIT::Misfit::Services::Resource->new();
        $resource = $self->{resource};
    }

    my @vols = @{$vols};
    my @luns = ();
    my %vserver = ();
    my $lun_count = 0;
    foreach my $vol (@vols) {
        if (!defined($vserver{$vol->{vserver}})) {
            my $filter = {vserver => $vol->{vserver}};
            $filter->{igroup} = $igroup if defined $igroup;
            if (defined $ostype) {
                $filter->{ostype} = $ostype;
                # There is no windows_2008 ostype for igroup.
                $filter->{ostype} =~ s/_2008//;
            }
            $vserver{$vol->{vserver}} = $resource->get_objects(
                type => 'igroup',
                count => 1,
                filter => $filter,
                requested_fields => ['igroup']
            );
            # Make sure we found some matching igroups otherwise return empty.
            return () unless (scalar(@{$vserver{$vol->{vserver}}}));
            $vol->{igroup} = $vserver{$vol->{vserver}};
        }
        for (my $i = 0; $i < $count; $i++) {
            my $index = $lun_count % scalar(@ostypes);
            my $reserve_type = $lun_count % scalar(@space_reserve);
            my $ostype = $ostypes[$index];
            my $lun_data = {};
            my $lun_name = $prefix . $lun_count;
            $lun_data = {
                lun             => $lun_name,
                vserver         => $vol->{vserver},
                volume          => $vol->{volume},
                size            => $size,
                ostype          => $ostype,
                igroup          => $vserver{$vol->{vserver}}[0]->{igroup},
                'space-reserve' => $space_reserve[$reserve_type],
                destination     => $lun_name. "__clone",
            };
            $lun_count++;
            push(@luns, $lun_data);
        }
    }

    return @luns;
}

#===============================================================================
#  PRIVATE METHODS
#===============================================================================
#==============================================================================
# Description: Subprocess responsible for creating LUNs
#==============================================================================
sub _create {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $lun = $params->get('lun');
    my $volume = $params->get('volume');
    my $size = $params->get('size');
    my $ostype = $params->get('ostype');
    my $if_exists = $params->get('if_exists', default => 'reuse');
    my $space_reserve = $params->get('space-reserve', default => 'false');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            routing_key => '#',
            message     => {
                msg  => 'LunOps.create.started',
                data => {
                    node              => $node->{node},
                    lun               => $lun,
                    volume            => $volume,
                    vserver           => $vserver,
                },
            },
        );
        NACL::STask::Lun->create(
            command_interface => $node,
            lun               => $lun,
            volume            => $volume,
            vserver           => $vserver,
            size              => $size,
            ostype            => $ostype,
            if_exists         => $if_exists,
            'space-reserve'   => $space_reserve,
            'method-timeout'  => TIMEOUT,
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Lun was not created');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg      => 'LunOps.create.' . $exception->text(),
                    data => {
                        node              => $node->{node},
                        lun               => $lun,
                        volume            => $volume,
                        vserver           => $vserver,
                    },
                }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            routing_key => '#',
            message     => {
                msg  => 'LunOps.create.finished',
                data => {
                    node              => $node->{node},
                    lun               => $lun,
                    volume            => $volume,
                    vserver           => $vserver,
                },
            }
        );
        return NATE::Result::Pass->new('Lun create successful');
    } else {
        return NATE::Result::Fail->new('Lun create failed');
    }
} # End of _create

#==============================================================================
# Description: Subprocess responsible for mapping LUNs
#==============================================================================
sub _map {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $lun = $params->get('lun');
    my $volume = $params->get('volume');
    my $igroup = $params->get('igroup');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            routing_key => '#',
            message     => {
                msg  => 'LunOps.map.started',
                data => {
                    node     => $node->{node},
                    lun      => $lun,
                    volume   => $volume,
                    vserver  => $vserver,
                    igroup   => $igroup
                },
            }
        );
        NACL::STask::Lun->map(
            command_interface => $node,
            lun               => $lun,
            volume            => $volume,
            vserver           => $vserver,
            igroup            => $igroup,
            'method-timeout'  => TIMEOUT,
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        if ($exception->text() =~ /already mapped to this/i) {
            $Log->comment("$lun is already mapped to $igroup - Skipping");
            $status = TRUE;
        } elsif ($exception->text() =~ /No such LUN exists/i) {
            $Log->comment(
                "$lun does not exist. Review the logs and see if this " .
                "was expected (i.e the LUN was never created)"
            );
            $status = TRUE;
        } else {
            $Log->comment(
                "Error mapping Lun: lun => $lun, vol => $volume, vs => $vserver"
            );
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg  => 'LunOps.map.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        lun     => $lun,
                        volume  => $volume,
                        vserver => $vserver,
                        igroup  => $igroup
                    },
                 }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            routing_key => '#',
            message     => {
                msg  => 'LunOps.map.finished',
                data => {
                    node    => $node->{node},
                    lun     => $lun,
                    volume  => $volume,
                    vserver => $vserver,
                    igroup  => $igroup
                },
            }
        );
        return NATE::Result::Pass->new('Lun map successful');
    } else {
        return NATE::Result::Fail->new('Lun map failed');
    }
} # End of _map

#==============================================================================
# Description: Subprocess responsible for unmapping LUNs
#==============================================================================
sub _unmap {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $lun = $params->get('lun');
    my $volume = $params->get('volume');
    my $igroup = $params->get('igroup');
    my $status = FALSE;

    try {
        # Making the same assumption as map, where the it's very likely that
        # the caller didn't care about the igroup
        unless (defined($igroup)) {
            my @igroups = NACL::CS::LunIgroup->fetch(
                command_interface => $node,
                filter => {
                    vserver => $vserver,
                },
                requested_fields => ['igroup']
            );
            if (scalar(@igroups)) {
                $igroup = $igroups[0]->{igroup};
            }
        }
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            routing_key => '#',
            message     => {
                msg  => 'LunOps.unmap.started',
                data => {
                    node    => $node->{node},
                    lun     => $lun,
                    volume  => $volume,
                    vserver => $vserver,
                    igroup  => $igroup
                },
            }
        );
        NACL::STask::Lun->unmap(
            command_interface => $node,
            lun               => $lun,
            volume            => $volume,
            vserver           => $vserver,
            igroup            => $igroup,
            'method-timeout'  => TIMEOUT,
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        if ($exception->text() =~ /not mapped to this/i) {
            $Log->comment("$lun is not mapped to $igroup - Skipping");
            $status = TRUE;
        } elsif ($exception->text() =~ /No such LUN exists/i) {
            $Log->comment(
                "$lun does not exist. Review the logs and see if this " .
                "was expected (i.e the LUN was never created)"
            );
            $status = TRUE;
        } else {
            $Log->comment(
                "Error unmapping Lun: lun => $lun, vol => $volume, vs => $vserver"
            );
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'LunOps.umap.' . $exception->text(),
                data => {
                    node    => $node->{node},
                    lun     => $lun,
                    volume  => $volume,
                    vserver => $vserver,
                    igroup  => $igroup
                },
            }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            routing_key => '#',
            message     => {
                msg  => 'LunOps.unmap.finished',
                data => {
                    node    => $node->{node},
                    lun     => $lun,
                    volume  => $volume,
                    vserver => $vserver,
                    igroup  => $igroup
                },
            }
        );
        return NATE::Result::Pass->new('Lun unmap successful');
    } else {
        return NATE::Result::Fail->new('Lun unmap failed');
    }
} # End of _unmap

#==============================================================================
# Description: Subprocess responsible for deleting LUNs
#==============================================================================
sub _delete {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $lun = $params->get('lun');
    my $volume = $params->get('volume');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            routing_key => '#',
            message     => {
                msg  => 'LunOps.delete.started',
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    lun      => $lun,
                },
            }
        );

        NACL::STask::Lun->purge(
            command_interface => $node,
            vserver           => $vserver,
            volume            => $volume,
            lun               => $lun,
            'method-timeout'  => TIMEOUT,
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        if ($exception->text() =~ /entry doesn\'t exist/i) {
            $Log->comment("$lun was not found -- Skipping");
            $status = TRUE;
        } else {
            $Log->comment(
                "Error deleting Lun: lun => $lun, vol => $volume, vs => $vserver"
            );
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg  => 'LunOps.delete.' . $exception->text(),
                    data => {
                        node     => $node->{node},
                        vserver  => $vserver,
                        volume   => $volume,
                        lun      => $lun,
                    },
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            routing_key => '#',
            message     => {
                msg  => 'LunOps.delete.finished',
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    lun      => $lun,
                },
            }
        );
        return NATE::Result::Pass->new('Lun delete successful');
    } else {
        return NATE::Result::Fail->new('Lun delete failed');
    }
} # End of _delete

#==============================================================================
# Description: Subprocess responsible for resizing LUNs
#==============================================================================
sub _resize {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $lun = $params->get('lun');
    my $volume = $params->get('volume');
    my $size = $params->get('size');
    my $status = FALSE;
    my $percent = $params->get('percent');

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            routing_key => '#',
            message     => {
                msg  => 'LunOps.resize.started',
                data => {
                    node    => $node->{node},
                    lun     => $lun,
                    volume  => $volume,
                    vserver => $vserver,
                },
            }
        );

        # Users may pass a percentage value instead of a size. This allows
        # them to 'increase the size of the LUN by 20%'

        if ($percent > 0) {
            my $tmp_lun = NACL::CS::Lun->fetch(
                command_interface => $node,
                filter => {
                    vserver => $vserver,
                    path    => "/vol/$volume/$lun"
                },
                requested_fields => ['size']
            );
            my $newsize = int((1 + ($percent / 100)) * $tmp_lun->{'size'});
            $Log->comment(
                "$lun will be increased by $percent%. New size will be $newsize"
            );
            $size = $newsize;
        }

        NACL::C::Lun->resize(
            command_interface => $node,
            lun               => $lun,
            volume            => $volume,
            vserver           => $vserver,
            size              => $size,
            'method-timeout'  => TIMEOUT,
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment(
            "Error resizing Lun: lun => $lun, vol => $volume, vs => $vserver " .
            $exception
        );
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'LunOps.resize.' . $exception->text(),
                data => {
                    node    => $node->{node},
                    lun     => $lun,
                    volume  => $volume,
                    vserver => $vserver,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            message     => {
                msg  => 'LunOps.resize.finished',
                data => {
                    node    => $node->{node},
                    lun     => $lun,
                    volume  => $volume,
                    vserver => $vserver,
                },
            }
        );
        return NATE::Result::Pass->new('Lun resize successful');
    } else {
        return NATE::Result::Fail->new('Lun resize failed');
    }
} # End of _resize

#==============================================================================
# Description: Subprocess responsible for cloning LUNs
#==============================================================================
sub _clone {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $lun = $params->get('lun');
    my $volume = $params->get('volume');
    my $snapshot = $params->get('snapshot');
    my $destination = $params->get('destination');
    my $status = FALSE;

    unless (defined($destination)) {
        $destination = $lun . '.clone';
    }

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            routing_key => '#',
            message     => {
                msg  => 'LunOps.clone.started',
                data => {
                    node    => $node->{node},
                    lun     => $lun,
                    volume  => $volume,
                    vserver => $vserver,

                },
            }
        );
        my %params = (
            command_interface  => $node,
            volume             => $volume,
            "source-path"      => $lun,
            "destination-path" => $destination,
            vserver            => $vserver,
            'method-timeout'  => TIMEOUT,
        );
        if (defined($snapshot)) {
            %params = (%params, "snapshot-name" => $snapshot);
        }
        NACL::C::Lun->clone(%params);
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        if ($exception =~ /Volume is full or Volume crossed autodelete/i) {
            $Log->comment("Volume is full and the clone operation failed");
            $status = TRUE;
        } elsif ($exception =~ /Clone file exists/i) {
            $Log->comment("LUN clone exists. Skipping");
            $status = TRUE;
        } else {
            $Log->comment(
                "Error cloning Lun: lun => $lun, vol => $volume, vs => $vserver"
            );
            $MessageBroker->publish_message(
                    exchange    => 'errors',
                    routing_key => '#',
                    message     => {
                    msg  => 'LunOps.clone.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        lun     => $lun,
                        volume  => $volume,
                        vserver => $vserver,
                    },
                }
            );
            NATE::BaseException->throw($exception);
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::LunOps',
            message     => {
                msg  => 'LunOps.clone.finished',
                data => {
                    node    => $node->{node},
                    lun     => $lun,
                    volume  => $volume,
                    vserver => $vserver,
                },
            }
        );
        return NATE::Result::Pass->new('Lun clone was successful');
    } else {
        return NATE::Result::Fail->new('Lun clone failed');
    }
} # End of _clone

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/Misfit/features/LunOps>

=cut

1;
