# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Snapmirror module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Features::Snapmirror
# Wrapper for NACL snapmirror operations with Misfit messaging

package San::FIT::Misfit::Features::Snapmirror;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Cluster;
use NACL::C::Snapmirror;
use NACL::STask::Cluster;
use NACL::STask::Snapmirror;
use NACL::STask::VserverPeer;
use NATE::BaseException;
use San::FIT::Misfit::Core::Util;
use San::FIT::Misfit::Core::Workflow;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::MessageBroker;
use NACL::C::Exceptions::VserverPeer::AlreadyExists;

# Inherit base class - Misfit Base
use base qw(San::FIT::Misfit::Core::Base);

use constant {
    TIMEOUT      => 360,
    INIT_TIMEOUT => 600
};

# Main broker used by this module
my $MessageBroker;

#===============================================================================
# Constructor: new
#
# Returns:
# New Snapmirror object
#
# Example:
#> my $Snapmirror = San::FIT::Misfit::Features::Snapmirror->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Features::Snapmirror";
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->comment("Failed to create a message broker for this module");
        return {};
    }
    # Initialize self with the testbed's cluster information.
    $self->_discover_clusters();

    $Log->exit();
    return $self;
} # End of new

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#==============================================================================
# Function: create
# Create n snapmirrors.  Create will loop through the list of objects provided
# and create each snapmirror with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for creating each
#       snapmirror
#   background - Option to specify whether to perform each create
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the creates
#
# Snapmirror create attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#   type - Type of snapmirror relationship.  Valid options are DP (default) or
#       LS.  For XDP snapmirrors, please use the Vault module.
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $snapmirror->create(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
#          type               => 'DP'
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub create {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapmirrors = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @snapmirror_info = ();
    my @msg_data = ();
    foreach my $snapmirror (@{$snapmirrors}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $snapmirror->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $snapmirror->{'destination-path'}
        });
        if (defined ($snapmirror->{id})) {
            $runid = $snapmirror->{id} . "_snapmirror_create";
        } else {
            $runid = $path{volume} . "_snapmirror_create";
        }
        push (@snapmirror_info, {
            codespec => "{" . $self->{id} . "::_create(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapmirror},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapmirror_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapmirror create failed");
        return FALSE;
    }
} # End of create

#==============================================================================
# Function: initialize
# Initialize n snapmirrors.  Initialize will loop through the list of objects provided
# and initialize each snapmirror with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for initializing each
#       snapmirror
#   background - Option to specify whether to perform each initialize
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the initializes
#
# Snapmirror initialize attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $snapmirror->initialize(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub initialize {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapmirrors = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @snapmirror_info = ();
    my @msg_data = ();
    foreach my $snapmirror (@{$snapmirrors}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $snapmirror->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $snapmirror->{'destination-path'}
        });
        if (defined ($snapmirror->{id})) {
            $runid = $snapmirror->{id} . "_snapmirror_initialize";
        } else {
            $runid = $path{volume} . "_snapmirror_initialize";
        }
        push (@snapmirror_info, {
            codespec => "{" . $self->{id} . "::_initialize(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapmirror},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapmirror_info,
        operation  => "initialize",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapmirror initialize failed");
        return FALSE;
    }
} # End of initialize

#==============================================================================
# Function: initialize_ls_set
# Initialize n load sharing snapmirrors.  Initialize_ls_set will loop through the list
# of objects provided and initialize each load sharing snapmirror with the
# specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for initializing each
#       load sharing snapmirror
#   background - Option to specify whether to perform each initialize
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the initializes
#
# Snapmirror create attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $snapmirror->initialize_ls_set(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub initialize_ls_set {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapmirrors = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @snapmirror_info = ();
    my @msg_data = ();
    foreach my $snapmirror (@{$snapmirrors}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $snapmirror->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $snapmirror->{'destination-path'}
        });
        if (defined ($snapmirror->{id})) {
            $runid = $snapmirror->{id} . "_snapmirror_initialize_ls_set";
        } else {
            $runid = $path{volume} . "_snapmirror_initialize_ls_set";
        }
        push (@snapmirror_info, {
            codespec => "{" . $self->{id} . "::_initialize_ls_set(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapmirror},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_home}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapmirror_info,
        operation  => "initialize_ls_set",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapmirror initialize_ls_set failed");
        return FALSE;
    }
} # End of initialize_ls_set

#==============================================================================
# Function: update
# Update n snapmirrors.  Update will loop through the list of objects provided
# and update each snapmirror with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for updating each
#       snapmirror
#   background - Option to specify whether to perform each update
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the updates
#
# Snapmirror update attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $snapmirror->update(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub update {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapmirrors = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @snapmirror_info = ();
    my @msg_data = ();
    foreach my $snapmirror (@{$snapmirrors}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $snapmirror->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $snapmirror->{'destination-path'}
        });
        if (defined ($snapmirror->{id})) {
            $runid = $snapmirror->{id} . "_snapmirror_udate";
        } else {
            $runid = $path{volume} . "snapmirror_update";
        }
        push (@snapmirror_info, {
            codespec => "{" . $self->{id} . "::_update(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapmirror},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapmirror_info,
        operation  => "update",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapmirror update failed");
        return FALSE;
    }
} # End of initialize

#==============================================================================
# Function: delete
# Create n snapmirrors.  Delete will loop through the list of objects provided
# and delete each snapmirror with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for deleting each
#       snapmirror
#   background - Option to specify whether to perform each delete
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the deletes
#
# Snapmirror delete attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $snapmirror->delete(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub delete {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapmirrors = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @snapmirror_info = ();
    my @msg_data = ();
    foreach my $snapmirror (@{$snapmirrors}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $snapmirror->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $snapmirror->{'destination-path'}
        });
        if (defined ($snapmirror->{id})) {
            $runid = $snapmirror->{id} . "_snapmirror_delete";
        } else {
            $runid = $path{volume} . "_snapmirror_delete";
        }
        push (@snapmirror_info, {
            codespec => "{" . $self->{id} . "::_delete(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapmirror},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapmirror_info,
        operation  => "delete",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapmirror delete failed");
        return FALSE;
    }
} # End of delete

#==============================================================================
# Function: release
# Release n snapmirrors.  Release will loop through the list of objects provided
# and release each snapmirror with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for releasing each
#       snapmirror
#   background - Option to specify whether to perform each release
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the releases
#
# Snapmirror release attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $snapmirror->release(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub release {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapmirrors = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @snapmirror_info = ();
    my @msg_data = ();
    foreach my $snapmirror (@{$snapmirrors}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $snapmirror->{'source-path'}
        );
        push (@msg_data, {
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $snapmirror->{'destination-path'}
        });
        if (defined ($snapmirror->{id})) {
            $runid = $snapmirror->{id} . "_snapmirror_release";
        } else {
            $runid = $path{volume} . "_snapmirror_release";
        }
        push (@snapmirror_info, {
            codespec => "{" . $self->{id} . "::_release(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapmirror},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_home}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapmirror_info,
        operation  => "release",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapmirror release failed");
        return FALSE;
    }
} # End of release

#==============================================================================
# Function: promote
# Promote n snapmirrors.  Promote will loop through the list of objects
# provided and promote each snapmirror with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for promoting each
#       snapmirror
#   background - Option to specify whether to perform each promote
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the promotes
#
# Snapmirror promote attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $snapmirror->promote(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub promote {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapmirrors = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($snapmirrors)) {
        $snapmirrors = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @snapmirror_info = ();
    my @msg_data = ();
    foreach my $snapmirror (@{$snapmirrors}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $snapmirror->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $snapmirror->{'destination-path'}
        });
        if (defined ($snapmirror->{id})) {
            $runid = $snapmirror->{id} . "_snapmirror_promote";
        } else {
            $runid = $path{volume} . "_snapmirror_promote";
        }
        push (@snapmirror_info, {
            codespec => "{" . $self->{id} . "::_promote(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapmirror},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_home}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapmirror_info,
        operation  => "promote",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapmirror promote failed");
        return FALSE;
    }
} # End of promote

#==============================================================================
# Function: break
# Break n snapmirrors.  Break will loop through the list of objects provided
# and break each snapmirror with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for breaking each
#       snapmirror
#   background - Option to specify whether to perform each break
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the breaks
#
# Snapmirror create attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $snapmirror->break(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub break {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapmirrors = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($snapmirrors)) {
        $snapmirrors = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @snapmirror_info = ();
    my @msg_data = ();
    foreach my $snapmirror (@{$snapmirrors}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $snapmirror->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $snapmirror->{'destination-path'}
        });
        if (defined ($snapmirror->{id})) {
            $runid = $snapmirror->{id} . "_snapmirror_break";
        } else {
            $runid = $path{volume} . "_snapmirror_break";
        }
        push (@snapmirror_info, {
            codespec => "{" . $self->{id} . "::_break(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapmirror},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapmirror_info,
        operation  => "break",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapmirror break failed");
        return FALSE;
    }
} # End of break

#==============================================================================
# Function: resync
# Resync n snapmirrors.  Resync will loop through the list of objects provided
# and resync each snapmirror with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for resyncing each
#       snapmirror
#   background - Option to specify whether to perform each resync
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the resyncs
#
# Snapmirror resync attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $snapmirror->resync(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub resync {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapmirrors = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($snapmirrors)) {
        $snapmirrors = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @snapmirror_info = ();
    my @msg_data = ();
    foreach my $snapmirror (@{$snapmirrors}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $snapmirror->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $snapmirror->{'destination-path'}
        });
        if (defined ($snapmirror->{id})) {
            $runid = $snapmirror->{id} . "_snapmirror_resync";
        } else {
            $runid = $path{volume} . "_snapmirror_resync";
        }
        push (@snapmirror_info, {
            codespec => "{" . $self->{id} . "::_resync(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapmirror},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapmirror_info,
        operation  => "resync",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapmirror resync failed");
        return FALSE;
    }
} # End of resync

#==============================================================================
# Function: build_snapmirror_relationships
# Given a list of source volume hashes and a list of destination volume hashes,
# build_snapmirror_relationships parses the volume objects and creates source and
# destination paths for use in the other functions of the Misfit Snapmirror
# module.
#
# Parameters:
#   source_volumes - array of hashes containing information
#       for each source volume.
#   destination_volumes - array of hashes containing
#       information for each destination volume.
#   type - Type of snapmirror relationships to be created (For XDP
#       relationships, use the Vault module). Defaults to DP
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   Array of Snapmirror relationships containing paths for source and
#   destination
#
#
# Example:
# >my @snapmirrors = $Snapmirror->build_snapmirror_relationships(
# >    source_volumes      => \@Source_Volumes,
# >    destination_volumes => \@Destination_Volumes,
# >    type                => 'DP',
# >);
#
#==============================================================================
sub build_snapmirror_relationships {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $src_volumes = &ParamsMgr::get_arg(
        'arg'      => "source_volumes",
        'required' => TRUE,
    );
    my $dst_volumes = &ParamsMgr::get_arg(
        'arg'      => "destination_volumes",
        'required' => TRUE,
    );
    my $type = &ParamsMgr::get_arg(
        'arg'      => "type",
        'required' => FALSE,
        'default'  => 'DP',
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @snapmirror_relationships = ();
    my @dsts = @{$dst_volumes};
    my $index = 0;
    foreach my $src (@{$src_volumes}) {
        my $dst = $dsts[$index];
        if (defined($dst)) {
            my $source_path = $self->{cluster_home_name} . '://' .
                $src->{vserver} . '/' . $src->{volume};
            my $dest_path = $self->{cluster_peer_name}  . '://' . $dst->{vserver}
                . '/' . $dst->{volume};
            push(@snapmirror_relationships, {
                'destination-path' => $dest_path,
                'source-path'      => $source_path,
                type               => $type,
            });
        }
        $index++;
    }

    return wantarray ? @snapmirror_relationships : \@snapmirror_relationships;
}

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#==============================================================================
# Description: Subprocess responsible for creating snapmirrors
#==============================================================================
sub _create {
    my ($class) = @_;

    my $params     = NATE::ParamSet->get_global();
    my $source     = $params->get('source-path');
    my $dest       = $params->get('destination-path');
    my $node       = $params->get('node');
    my $type       = $params->get('type');
    my $vserver    = $params->get('vserver');
    my $volume     = $params->get('volume');
    my $if_exists  = $params->get('if_exists', default => 'reuse');
    my $status     = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.create.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NACL::STask::Snapmirror->create(
            command_interface   => $node,
            do_initialize       => FALSE,
            'source-path'       => $source,
            'destination-path'  => $dest,
            'type'              => $type,
            if_exists           => $if_exists
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.create.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error creating Snapmirror');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'Snapmirror.create.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $Log->comment($exception->text());
    };

    if ($status) {
        return $Log->log(NATE::Result::Pass->new('Snapmirror create successful'));
    } else {
        $Log->comment('Snapmirror create failed');
    }
} # End of _create

#==============================================================================
# Description: Subprocess responsible for initializing Snapmirrors
#==============================================================================
sub _initialize {
    my ($class) = @_;

    my $params     = NATE::ParamSet->get_global();
    my $source     = $params->get('source-path');
    my $dest       = $params->get('destination-path');
    my $node       = $params->get('node');
    my $vserver    = $params->get('vserver');
    my $volume     = $params->get('volume');
    my $if_exists  = $params->get('if_exists', default => 'reuse');
    my $status     = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.initialize.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NACL::STask::Snapmirror->initialize(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest,
            if_exists          => $if_exists
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.initialize.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error initializing Snapmirror');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
            msg  => 'Snapmirror.initialize.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Snapmirror initialize successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Snapmirror initialize failed'));
    }
} # End of _initialize

#==============================================================================
# Description: Subprocess responsible for initializing LoadSharing Snapmirrors
#==============================================================================
sub _initialize_ls_set {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.initialize_ls_set.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NACL::C::Snapmirror->initialize_ls_set(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest
        );
        NACL::C::Snapmirror->wait_for_completion(
            command_interface  => $node,
            'method-timeout'   => INIT_TIMEOUT,
            'source-path'      => $source,
            'destination-path' => $dest,
            attribute_to_check => 'state',
            till_value         => 'snapmirrored'
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.initialize_ls_set.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error initialize_ls_set Snapmirror');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
            msg  => 'Snapmirror.initialize_ls_set.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Snapmirror initialize_ls_set successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Snapmirror initialize_ls_set failed'));
    }
} # End of _initialize_ls_set

#==============================================================================
# Description: Subprocess responsible for updating Snapmirrors
#==============================================================================
sub _update {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.update.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NACL::STask::Snapmirror->update(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.update.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error updating Snapmirror');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
            msg  => 'Snapmirror.update.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Snapmirror update successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Snapmirror update failed'));
    }
} # End of _update

#==============================================================================
# Description: Subprocess responsible for deleting Snapmirrors
#==============================================================================
sub _delete {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.delete.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        my $job;
        NACL::C::Snapmirror->delete(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest,
            job_component      => \$job
        );
        if (defined ($job)) {
            $job->wait_on_job('method-timeout' => TIMEOUT);
        }
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.delete.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error deleting Snapmirror');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'Snapmirror.delete.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Snapmirror delete successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Snapmirror delete failed'));
    }
} # End of _delete

#==============================================================================
# Description: Subprocess responsible for releasing Snapmirrors
#==============================================================================
sub _release {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.release.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        my $job;
        NACL::C::Snapmirror->release(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest,
            job_component      => \$job
        );
        if (defined ($job)) {
            $job->watch_progress('method-timeout' => TIMEOUT);
        }
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.release.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error releasing Snapmirror');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'Snapmirror.release.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Snapmirror release successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Snapmirror release failed'));
    }
} # End of _release

#==============================================================================
# Description: Subprocess responsible for promoting Snapmirrors
#==============================================================================
sub _promote {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.promote.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );

        my $job;
        NACL::C::Snapmirror->promote(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest,
            job_component      => \$job
        );
        if (defined ($job)) {
            $job->watch_progress('method-timeout' => TIMEOUT);
        }
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.promote.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error promoting Snapmirror');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'Snapmirror.promote.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Snapmirror promote successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Snapmirror promote failed'));
    }
} # End of _promote

#==============================================================================
# Description: Break Snapmirror
#==============================================================================
sub _break {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.break.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        my $job;
        NACL::C::Snapmirror->break(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest,
            job_component      => \$job
        );
        if (defined ($job)) {
            $job->watch_progress('method-timeout' => TIMEOUT);
        }
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.break.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error breaking Snapmirror');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'Snapmirror.break.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Snapmirror break successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Snapmirror break failed'));
    }
} # End of _break

#==============================================================================
# Description: Resync Snapmirror
#==============================================================================
sub _resync {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.resync.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NACL::STask::Snapmirror->resync(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapmirror',
            routing_key => '#',
            message     => {
                msg  => 'Snapmirror.resync.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error resyncing Snapmirror');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'Snapmirror.resync.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Snapmirror resync successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Snapmirror resync failed'));
    }
} # End of _resync

#==============================================================================
# Description: Discover and get cluster information for Snapmirrors tests.
#==============================================================================
sub _discover_clusters {
    my ($self, %args) = @_;
    $Log->enter();


    my $cluster_name = param("CLUSTER_NAME", "-default", "");
    my $peer_name    = param("PEER_CLUSTER", "-default", $cluster_name);

    $self->{cluster_home_name} = $cluster_name;
    $self->{cluster_home}      = NACL::C::Cluster->new(name => $cluster_name);
    $self->{cluster_home}->apiset_must(interface => 'ZAPI');
    $self->{cluster_home}->apiset();
    $self->{cluster_home_cli} = NACL::C::Cluster->new(name => $cluster_name);
    $self->{cluster_home_cli}->apiset_must(interface => 'CLI');
    $self->{cluster_home_cli}->apiset();

    $self->{cluster_peer_name} = $peer_name;
    $self->{cluster_peer}      = NACL::C::Cluster->new(name => $peer_name);
    $self->{cluster_peer}->apiset_must(interface => 'ZAPI');
    $self->{cluster_peer}->apiset();
    $self->{cluster_peer_cli} = NACL::C::Cluster->new(name => $peer_name);
    $self->{cluster_peer_cli}->apiset_must(interface => 'CLI');
    $self->{cluster_peer_cli}->apiset();

   $Log->exit();
}

#==============================================================================
# Description: Create and accept a vserver peer relationship for SN clusters
#==============================================================================
sub _create_vserver_peer {
    my ($self, %args) = @_;
    $Log->enter();

    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $peer_vserver = &ParamsMgr::get_arg(
        'arg'      => "peer_vserver",
        'required' => TRUE,
    );
    my $peer_cluster = &ParamsMgr::get_arg(
        'arg'      => "peer_cluster",
        'required' => FALSE,
        'default'  => $self->{cluster_peer_name}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $status = TRUE;
    try {
        NACL::STask::VserverPeer->create(
            command_interface  => $self->{cluster_home},
            vserver            => $vserver,
            'peer-vserver'     => $peer_vserver,
            'peer-cluster'     => $peer_cluster,
            applications       => ['snapmirror'],
            nacltask_if_exists => 'reuse'
        );
    } otherwise {
        my $exception = shift;
        $Log->comment($exception->text());
        $status = FALSE;
    };

    try {
        NACL::STask::VserverPeer->accept(
            command_interface  => $self->{cluster_peer},
            'peer-vserver'     => $vserver,
            vserver            => $peer_vserver,
        );
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /Invalid peering state "peered" on the local cluster./i) {
            $Log->comment("Vservers already peered... Continuing...");
            $status = TRUE;
        } elsif ($exception->text() =~ /Invalid operation on intra cluster Vserver peering/i) {
            $Log->comment(
                'Cannot accept peer on Intracluster peering, continuing...'
            );
            $status = TRUE;
        } elsif ($exception->text() =~ /Invalid operation on intra cluster Vserver peer relationship./i) {
            $Log->comment(
                'Cannot accept peer on Intracluster peering, continuing...'
            );
            $status = TRUE;
        } else {
            $Log->comment($exception->text());
            $status = FALSE;
        }
    };

    return $status;
}

#==============================================================================
# Description: Delete a vserver peer relationship for SN clusters
#==============================================================================
sub _delete_vserver_peer {
    my ($self, %args) = @_;
    $Log->enter();

    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $peer_vserver = &ParamsMgr::get_arg(
        'arg'      => "peer_vserver",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $status = TRUE;

    try {
        NACL::STask::VserverPeer->purge(
            command_interface  => $self->{cluster_peer},
            vserver            => $peer_vserver,
            'peer-vserver'     => $vserver,
            nacltask_if_purged => 'pass',
        );
    } otherwise {
        $status = FALSE;
    };

    try {
        NACL::STask::VserverPeer->purge(
            command_interface  => $self->{cluster_home},
            vserver            => $vserver,
            'peer-vserver'     => $peer_vserver,
            nacltask_if_purged => 'pass',
        );
    } otherwise {
        $status = FALSE;
    };

    if ($status) {
        return TRUE;
    } else {
        return FALSE;
    }
} # End of _delete_vserver_peer

1;

#===============================================================================
# Message: snapmirror.create.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.create.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.create.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.create.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.create.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.create.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.initialize.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.initialize.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.initialize.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.initialize.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.initialize.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.initialize.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.initialize_ls_set.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.initialize_ls_set.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.initialize_ls_set.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.initialize_ls_set.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.initialize_ls_set.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.initialize_ls_set.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.update.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.update.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.update.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.update.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.update.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.update.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.delete.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.delete.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.delete.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.delete.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.delete.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.delete.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.release.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.release.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.release.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.release.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.release.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.release.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.promote.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.promote.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.promote.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.promote.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.promote.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.promote.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.break.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.break.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.break.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.break.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.break.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.create.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.resync.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.resync.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.resync.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.resync.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: snapmirror.resync.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.resync.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#===============================================================================

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/Misfit/features/Snapmirror>

=cut

1;
