# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Snapshot module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Features::Snapshot
# Wrapper for NACL snapshot operations with Misfit messaging

package San::FIT::Misfit::Features::Snapshot;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use ParamsMgr;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::Cluster;
use NACL::C::VolumeSnapshot;
use NACL::STask::VolumeSnapshot;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Workflow;

# Inherit base class - Misfit Base
use base qw(San::FIT::Misfit::Core::Base);

use constant {
    TIMEOUT => 360
};

# Main broker used by this module
my $MessageBroker;

#===============================================================================
# Constructor: new
#
#   Parameters:
#       cluster - name of cluster to perform operation on.
#
#   Returns:
#       New Snapshot object
#
#   Example:
# >     my $Snapshot = San::FIT::Misfit::Features::Snapshot->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Features::Snapshot";
    my $interface = NACL::C::Node->find();
    $interface->apiset_must(interface => 'ZAPI');

    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );

    my $node = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
    }

    unless (defined($node)) {
        $Log->comment('Failed to create a command interface object!');
        return {};
    }

    $node->apiset_must(interface => 'ZAPI');
    $node->apiset();
    $self->{node} = $node;
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#==============================================================================
# Function: create
# Create n snapshots.  Create will loop through the list of objects provided
# and create each snapshot with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for creating each snapshot
#   background - Option to specify whether to perform each create
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the creates
#
# Snapshot create attributes:
#   snapshot - Name of snapshot (required)
#   volume - Name of volume to create snapshot (required)
#   vserver - Name of vserver where volume is located (required)
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $Snapshot->create(
# >    objects    => [{
# >        snapshot => 'test_vol_snapshot',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub create {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapshots = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($snapshots)) {
        $snapshots = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @snapshot_info = ();
    my @msg_data = ();
    foreach my $snapshot (@{$snapshots}) {
        my $runid = '';
        if (defined ($snapshot->{id})) {
            $runid = $snapshot->{id} . "_create";
        } else {
            $runid = $snapshot->{snapshot} . "_create";
        }
        push(@msg_data, {vserver => $snapshot->{vserver}, volume => $snapshot->{volume}});
        push (@snapshot_info, {
            codespec => "{" . $self->{id} . "::_create(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapshot},
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapshot_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapshot create failed");
        return FALSE;
    }
} # End of create

#==============================================================================
# Function: restore
# Restore n snapshots.  Restore will loop through list of objects and perform
# a restore on the specified volume from the specified snapshot
#
# Parameters:
#   objects - Array of hashes containing information for restoring each
#       snapshot
#   background - Option to specify whether to perform each restore
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the restores
#
# Snapshot restore attributes:
#   snapshot - Name of snapshot (required)
#   volume - Name of volume to restore snapshot on (required)
#   vserver - Name of vserver where volume is located
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $Snapshot->restore(
# >    objects    => [{
# >        snapshot => 'test_vol_snapshot',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub restore {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapshots = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @snapshot_info = ();
    my @msg_data = ();
    foreach my $snapshot (@{$snapshots}) {
        my $runid = '';
        if (defined ($snapshot->{id})) {
            $runid = $snapshot->{id} . "_restore";
        } else {
            $runid = $snapshot->{snapshot} . "_restore";
        }
        push(@msg_data, {vserver => $snapshot->{vserver}, volume => $snapshot->{volume}});
        push (@snapshot_info, {
            codespec => "{" . $self->{id} . "::_restore(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapshot},
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapshot_info,
        operation  => "restore",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapshot create failed");
        return FALSE;
    }
} # End of restore

#==============================================================================
# Function: delete
# Delete n snapshots.  Restore will loop through list of objects and perform
# a delete on the specified snapshot
#
# Parameters:
#   objects - Array of hashes containing information for deleting each
#       snapshot
#   background - Option to specify whether to perform each delete
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the deletes
#
# Snapshot restore attributes:
#   snapshot - Name of snapshot (required)
#   volume - Name of volume to delete snapshot from (required)
#   vserver - Name of vserver where volume is located
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $Snapshot->delete(
# >    objects    => [{
# >        snapshot => 'test_vol_snapshot',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub delete {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $snapshots = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @snapshot_info = ();
    my @msg_data = ();
    foreach my $snapshot (@{$snapshots}) {
        my $runid = '';
        if (defined ($snapshot->{id})) {
            $runid = $snapshot->{id} . "_delete";
        } else {
            $runid = $snapshot->{snapshot} . "_delete";
        }
        push(@msg_data, {vserver => $snapshot->{vserver}, volume => $snapshot->{volume}});
        push (@snapshot_info, {
            codespec => "{" . $self->{id} . "::_delete(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$snapshot},
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@snapshot_info,
        operation  => "delete",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Snapshot create failed");
        return FALSE;
    }
} # End of delete

#===============================================================================
#  PRIVATE METHODS
#===============================================================================

#==============================================================================
# Description: Subprocess responsible for creating a snapshot
#==============================================================================
sub _create {
    my ($class) = @_;
    my $params = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $snapshot = $params->get('snapshot');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapshot',
            routing_key => '#',
            message     => {
                msg  => 'snapshot.create.started',
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    snapshot => $snapshot,
                },
            }
        );
        NACL::STask::VolumeSnapshot->create(
            command_interface  => $node,
            volume             => $volume,
            vserver            => $vserver,
            snapshot           => $snapshot,
            nacltask_if_exists => "purge",
            nacltask_wait      => TRUE,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapshot',
            routing_key => '#',
            message     => {
                msg  => 'snapshot.create.finished',
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    snapshot => $snapshot,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->error('Error creating Snapshot');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'snapshot.create.' . $exception->text(),
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    snapshot => $snapshot,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Snapshot create successful');
    } else {
        return NATE::Result::Fail->new('Snapshot create failed');
    }
} # End of _create

#==============================================================================
# Description: Subprocess responsible for restoring a snapshot
#==============================================================================
sub _restore {
    my ($class) = @_;
    my $params = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $snapshot = $params->get('snapshot');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapshot',
            routing_key => '#',
            message     => {
                msg  => 'snapshot.restore.started',
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    snapshot => $snapshot,
                },
            }
        );
        my $job;
        NACL::C::VolumeSnapshot->restore(
            command_interface => $node,
            volume            => $volume,
            vserver           => $vserver,
            snapshot          => $snapshot,
            job_component     => \$job
        );
        if (defined ($job)) {
            $job->watch_progress('method-timeout' => TIMEOUT);
        }
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapshot',
            routing_key => '#',
            message     => {
                msg  => 'snapshot.restore.finished',
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    snapshot => $snapshot,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->error('Error restoring Snapshot');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'snapshot.restore.' . $exception->text(),
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    snapshot => $snapshot,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Snapshot restore successful');
    } else {
        return NATE::Result::Fail->new('Snapshot restore failed');
    }
} # End of _restore

#==============================================================================
# Description: Subprocess responsible for deleting Snapshots
#==============================================================================
sub _delete {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $snapshot = $params->get('snapshot');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapshot',
            routing_key => '#',
            message     => {
                msg  => 'snapshot.delete.started',
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    snapshot => $snapshot,
                },
            }
        );
        NACL::STask::VolumeSnapshot->purge(
            command_interface  => $node,
            volume             => $volume,
            vserver            => $vserver,
            snapshot           => $snapshot,
            nacltask_if_exists => "purge",
            nacltask_wait      => TRUE,
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Snapshot',
            routing_key => '#',
            message     => {
                msg  => 'snapshot.delete.finished',
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    snapshot => $snapshot,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error deleting Snapshot');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'snapshot.delete.' . $exception->text(),
                data => {
                    node     => $node->{node},
                    vserver  => $vserver,
                    volume   => $volume,
                    snapshot => $snapshot,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Snapshot delete successful');
    } else {
        return NATE::Result::Fail->new('Snapshot delete failed');
    }
} # End of _delete

#===============================================================================
# Message: snapshot.create.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapshot
#
#   Message Structure:
# >     {
# >         "msg": "snapshot.create.started",
# >         "data" : {
# >             "snapshot": "sanfit_snapshot",
# >             "volume": "sanfit_vol1",
# >             "vserver": "sanfit_test_vs",
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: snapshot.create.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapshot
#
#   Message Structure:
# >     {
# >         "msg": "snapshot.create.finished",
# >         "data" : {
# >             "snapshot": "sanfit_snapshot",
# >             "volume": "sanfit_vol1",
# >             "vserver": "sanfit_test_vs",
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: snapshot.create.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapshot.create.Error Text",
# >         "data" : {
# >             "snapshot": "sanfit_snapshot",
# >             "volume": "sanfit_vol1",
# >             "vserver": "sanfit_test_vs",
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: snapshot.restore.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapshot
#
#   Message Structure:
# >     {
# >         "msg": "snapshot.restore.started",
# >         "data" : {
# >             "snapshot": "sanfit_snapshot",
# >             "volume": "sanfit_vol1",
# >             "vserver": "sanfit_test_vs",
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: snapshot.restore.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapshot
#
#   Message Structure:
# >     {
# >         "msg": "snapshot.restore.finished",
# >         "data" : {
# >             "snapshot": "sanfit_snapshot",
# >             "volume": "sanfit_vol1",
# >             "vserver": "sanfit_test_vs",
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: snapshot.restore.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapshot.restore.Error Text",
# >         "data" : {
# >             "snapshot": "sanfit_snapshot",
# >             "volume": "sanfit_vol1",
# >             "vserver": "sanfit_test_vs",
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
# Message: snapshot.delete.started
#   Exchange:
#       San::FIT::Misfit::Features::Snapshot
#
#   Message Structure:
# >     {
# >         "msg": "snapshot.delete.started",
# >         "data" : {
# >             "snapshot": "sanfit_snapshot",
# >             "volume": "sanfit_vol1",
# >             "vserver": "sanfit_test_vs",
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: snapshot.delete.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapshot
#
#   Message Structure:
# >     {
# >         "msg": "snapshot.delete.finished",
# >         "data" : {
# >             "snapshot": "sanfit_snapshot",
# >             "volume": "sanfit_vol1",
# >             "vserver": "sanfit_test_vs",
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#
# Message: snapshot.delete.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "snapshot.delete.Error Text",
# >         "data" : {
# >             "snapshot": "sanfit_snapshot",
# >             "volume": "sanfit_vol1",
# >             "vserver": "sanfit_test_vs",
# >             "node": "sanfit_cluster_1"
# >         }
# >     }
#===============================================================================

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/Misfit/features/Snapshot>

=cut

1;