# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Storage Failover module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Features::StorageFailover
# Library to perform Storage Failover operations using Misfit messaging

package San::FIT::Misfit::Features::StorageFailover;

# Compiler Directives
use strict;
use warnings;

# Import Libraries
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::MTask::HA;
use NACL::C::Cluster;
use NACL::STask::Node;
use NACL::C::StorageFailover;
use NACL::STask::StorageFailover;
use San::FIT::Misfit::Core::Workflow;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);

# This package inherits methods and attributes from Misfit's Base class
use base qw( San::FIT::Misfit::Core::Base );

use constant {
    ID                       => "San::FIT::Misfit::Features::StorageFailover",
    DELAY                    => 600,
    TIMEOUT                  => 300,
    VERIFY_EMS               => TRUE,
    VERIFY_ASUP              => TRUE,
    NACL_VERIFY              => TRUE,
    RESET_FLAGS              => TRUE,
    POLL_INTERVAL            => 10,
    VERIFY_ARL_EMS           => TRUE,
    FAILURE_REASON           => 'triggered_by_misfit_automation',
    DETECTION_TIME           => 15,
    MESSAGE_TIMEOUT          => 14400,
    OVERRIDE_VETOES          => 'true',
    BOOT_AFTER_HALT          => TRUE,
    VERIFY_SFO_STATE         => TRUE,
    WAIT_FOR_PARTNER         => FALSE,
    INHIBIT_TAKEOVER         => TRUE,
    TAKEOVER_EXPECTED        => TRUE,
    GIVEBACK_EXPECTED        => TRUE,
    SKIP_AUTO_GIVEBACK       => TRUE,
    ONLY_CFO_AGGREGATES      => 'false',
    BYPASS_OPTIMIZATION      => 'false',
    VERIFY_CROSS_CONNECT     => FALSE,
    VERIFY_DISK_OWNERSHIP    => TRUE,
    DEFAULT_TAKEOVER_TYPE    => 'takeover_command',
    DEFAULT_METHOD_TIMEOUT   => 600,
    AUTO_GIVEBACK_EXPECTED   => FALSE,
    VERIFY_AGGR_STATE_OWNER  => TRUE,
    DEFAULT_TAKEOVER_OPTION  => 'normal',
    REQUIRE_PARTNER_WAITING  => 'false',
    VERIFY_AGGR_STATE_OWNER  => TRUE,
    WAIT_FOR_ASUP_GENERATION => TRUE,
};

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Create a StorageFailover object using Misfit services
#
#   Parameters:
#     cluster - Name of the cluster to perform operations on. Optional as long
#       as CLUSTER_NAME is defined or passed to the script.
#     initialize_ha_across_cluster - Initialize and configure the list of HA
#       pairs. An array reference containing node/partner names:
#       [{ node => 'node_1' , partner => 'node_2' }]
#
#   Returns:
#     New object for Storage Failover operations using misfit
#
#   Example:
#     > my $Sfo = San::FIT::Misfit::Features::StorageFailover->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $initialize_ha_across_cluster = &ParamsMgr::get_arg(
        'arg'      => "initialize_ha_across_cluster",
        'required' => FALSE,
        'default'  => [],
    );
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    # Instantiate the class
    my $self       = bless {}, $class;
    $self          = $class->SUPER::new();
    $self->{id}    = ID;
    $self->{ha}    = {};
    $MessageBroker = $self->{message_broker};

    # Try using the cluster management interface if possible
    my $cmd_interface = undef;
    if (defined ($cluster_name)) {
        $cmd_interface = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $cmd_interface = NACL::C::Node->find();
    }
    $Log->comment(
        "Using " . $cmd_interface->node() . " as the main command interface"
    );

    unless (defined($cmd_interface)) {
        $Log->comment(
            "Failed to create a command interface object. Make sure " .
            "CLUSTER_NAME is defined on this testbed or that the cluster name ".
            "passed to this method is correct."
        );
        return {};
    }

    $cmd_interface->apiset_must(interface => 'CLI');
    $cmd_interface->apiset();
    $self->{cmd_interface} = $cmd_interface;

    # If users specify to initialize Storage Failover then do it by going
    # through every node in the cluster and setting the appropriate (best
    # practices) settings for SFO testing

    if (scalar(@{$initialize_ha_across_cluster})) {
        my $status = $self->initialize(
            objects => $initialize_ha_across_cluster,
        );
        unless ($status) {
            $Log->comment(
                "Failed initializing High-Availability across all nodes. " .
                "Storage operations may not behave as expected. This " .
                "verification includes enabling HA, checkingaggregates and " .
                "node states before performing any Storage Failover commands."
            );
        }
    }

    return $self;
} # End of new

#===============================================================================
# Function: initialize
#   Verifies that disk shelves are cross-connected, storage failover is
#   enabled, all nodes are in a known state (no partial givebacks) and that the
#   event monitoring is reset.
#
#   Parameters:
#     verify_cross_connect - FALSE or TRUE (default). Ensures disk shelves are
#       properly connected for reliable Storage Failover operations
#     reset_flags - FALSE or TRUE (default). Reset storage failover settings on
#       a node.
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $Sfo->initialize();
#===============================================================================
sub initialize {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $verify_cross_connect = &ParamsMgr::get_arg(
        'arg'      => "verify_cross_connect",
        'required' => FALSE,
        'default'  => VERIFY_CROSS_CONNECT,
    );
    my $reset_flags = &ParamsMgr::get_arg(
        'arg'      => "reset_flags",
        'required' => FALSE,
        'default'  => RESET_FLAGS,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @nodes = NACL::C::Node->find(
        command_interface => $self->{cmd_interface}
    );

    my $fail_cnt = 0;
    foreach my $node (@nodes) {
        my $node_name = $node->node();
        my $status = $self->_initialize(
            node                 => $node,
            node_name            => $node_name,
            reset_flags          => $reset_flags,
            verify_cross_connect => $verify_cross_connect,
        );
        unless ($status) {
            $fail_cnt++;
        }
    }

    if ($fail_cnt == 0) {
        return TRUE;
    } else {
        NATE::BaseException->throw(
            "Failed to enable and verify HA across one or more HA pairs"
        );
        return FALSE;
    }
} # End of initialize

#===============================================================================
# Function: enable
#   Enables Storage Failover on a given node(s)
#
#   Parameters:
#     objects - Array of hashes containing node names (See attributes)
#     background - Enable storage failover on sevaral nodes in parallel.
#       Defaults to TRUE
#     iterations - Enable SFO 'n' number of times. Defaults to 1
#
#     Objects attributes:
#       'node' - Name of the node (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $Sfo->enable(
#     >      objects    => [{
#     >          'node'    => 'ssan-2240-2a',
#     >      ]},
#     > );
#===============================================================================
sub enable {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $nodes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @node_info = ();
    my @msg_data  = ();
    foreach my $node (@{$nodes}) {
        my $skip      = FALSE;
        my $node_obj  = undef;
        my $node_name = '';
        try {
            $node_obj = NACL::C::Node->new(
                node              => $node->{node},
                command_interface => $self->{cmd_interface}
            );
            $node_obj->apiset_must(interface => 'CLI');
            $node_obj->apiset();
            $node_name = $node->{node};
        } otherwise {
            $Log->comment(
                "Node " . $node->{node} . " is not available and will be skipped"
            );
            $skip = TRUE;
        };

        unless ($skip) {
            push(@msg_data, {node => $node_name});
            my $runid = $node_name . "_sfo_enable";
            push(@node_info, {
                codespec   => "{" . $self->{id} . "::_enable(); }",
                runid      => $runid,
                background => $background,
                params     => {
                    %{$node},
                    id                => $self->{id},
                    node_obj          => Clone::clone($node_obj),
                    node_name         => $node_name,
                    command_interface => Clone::clone($self->{cmd_interface}),
                }
            });
        }
    }

    my $status = $self->_run_workflow(
        objects    => \@node_info,
        operation  => "enable",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Failed to enabling storage failover");
        return FALSE;
    }
} # End of enable

#===============================================================================
# Function: disable
#   Disables Storage Failover on a given node(s)
#
#   Parameters:
#     objects - Array of hashes containing node names (See attributes)
#     background - Disable storage failover on sevaral nodes in parallel.
#       Defaults to TRUE
#     iterations - Disable SFO 'n' number of times. Defaults to 1
#
#     Objects attributes:
#       'node' - Name of the node (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $Sfo->disable(
#     >      objects    => [{
#     >          'node'    => 'ssan-2240-2a',
#     >      ]},
#     > );
#===============================================================================
sub disable {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $nodes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @node_info = ();
    my @msg_data  = ();
    foreach my $node (@{$nodes}) {
        my $skip      = FALSE;
        my $node_obj  = undef;
        my $node_name = '';
        try {
            $node_obj = NACL::C::Node->new(
                node              => $node->{node},
                command_interface => $self->{cmd_interface},
            );
            $node_obj->apiset_must(interface => 'CLI');
            $node_obj->apiset();
            $node_name = $node->{node};
        } otherwise {
            $Log->comment(
                "Node " . $node->{node} . "is not available and will be skipped"
            );
            $skip = TRUE;
        };

        unless ($skip) {
            push(@msg_data, {node => $node_name});
            my $runid = $node_name . "_sfo_disable";
            push(@node_info, {
                codespec   => "{" . $self->{id} . "::_disable(); }",
                runid      => $runid,
                background => $background,
                params     => {
                    %{$node},
                    id                => $self->{id},
                    node_obj          => Clone::clone($node_obj),
                    node_name         => $node_name,
                    command_interface => Clone::clone($self->{cmd_interface}),
                }
            });
        }
    }

    my $status = $self->_run_workflow(
        objects    => \@node_info,
        operation  => "disable",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Failed to disabling storage failover");
        return FALSE;
    }
} # End of disable

#===============================================================================
# Function: modify
#   Modify the specified storage failover attributes
#
#   Parameters:
#     objects - Array of hashes containing node names and SFO attributes
#     background - Modifies Storage Failover on sevaral nodes in parallel.
#       Defaults to TRUE
#     iterations - Modify SFO 'n' number of times. Defaults to 1
#
#     Modify attributes:
#       node -  Name of the node
#       onfailure - (Optional) 'true' or 'false'
#       onpanic - (Optional) 'true' or 'false'
#       onreboot - (Optional) 'true' or 'false'
#       auto-giveback - (Optional) 'true' or 'false'
#       check-partner - (Optional) 'true' or 'false'
#       detection-time - (Optional) Defaults to 15.
#       delay-seconds - (Optional) Defaults to 600
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $Sfo->modify(
#     >      objects    => [{
#     >          'node'      => 'ssan-2240-2a',
#     >          'onfailure' => 'true',
#     >          'onpanic'   => 'false',
#     >      ]},
#     >      background => TRUE,
#     > );
#===============================================================================
sub modify {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $nodes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @node_info = ();
    my @msg_data  = ();
    foreach my $node (@{$nodes}) {
        my $node_obj = NACL::C::Node->new(
            node              => $node->{node},
            command_interface => $self->{cmd_interface},
        );
        $node_obj->apiset_must(interface => 'CLI');
        $node_obj->apiset();
        my $node_name = $node->{node};
        push(@msg_data, {node => $node_name});
        my $runid = $node_name . "_sfo_modify";
        push(@node_info, {
            codespec   => "{" . $self->{id} . "::_modify(); }",
            runid      => $runid,
            background => $background,
            params     => {
                %{$node},
                id                => $self->{id},
                node_obj          => Clone::clone($node_obj),
                node_name         => $node_name,
                command_interface => Clone::clone($self->{cmd_interface}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@node_info,
        operation  => "modify",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Failed modifying SFO attributes on the node");
        return FALSE;
    }
} # End of modify

#===============================================================================
# Function: check_state
#   Checks the storage failover state on a given node(s). Note, this method
#   requires both nodes to be operational
#
#   Parameters:
#     objects - Array of hashes containing node names and SFO states
#     background - Check the SFO states on sevaral nodes in parallel.
#       Defaults to TRUE
#     iterations - Check states 'n' number of times. Defaults to 1
#
#     State attributes:
#       'node'          - name of the node
#       'node_state'    - Expected state for the main node
#       'partner_state' - Expected state for the partner node
#
#       Valid states:
#         'up'               - When the node is up with the OS booted
#         'takeover'         - When the node is in takeover mode
#         'givewait'         - When the node is at 'Press ctrl-C for maintainance mode'
#            storage-failover aggregates with storage-failover policy to the partner
#         'any_takeover'     - When the node is any takeover state
#         'wfg'              - When the node is in "waiting for giveback" prompt
#         'sparecore'        - When partner selects sparecore for dumping core
#         'dumpcore'         - When partner selects spraycore for dumping core
#         'node_unreachable' - When the partner is unreachable
#         'wfg_or_givewait'  - When the node is in *any* of the "waiting for
#           giveback" prompts.
#         'partial_giveback' - When node has not sent all of the
#         'takeover_autogiveback_scheduled' - State when the node is in takeover
#           and autogiveback is scheduled
#         'system_id_changed_on_partner'- State when the partner system id is
#           changed after headswap
#
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $Sfo->check_state(
#     >      objects    => [{
#     >          'node'          => 'ssan-2240-2a',
#     >          'node_state'    => 'up',
#     >          'partner_state' => 'up',
#     >      ]},
#     >      background => TRUE,
#     > );
#===============================================================================
sub check_state {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $nodes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @node_info = ();
    my @msg_data  = ();
    foreach my $node (@{$nodes}) {
        my $node_obj     = undef;
        my $node_name    = '';
        my $partner_obj  = undef;
        my $partner_name = '';
        my $skip         = FALSE;
        try {
            $node_obj = NACL::C::Node->new(
                node              => $node->{node},
                command_interface => $self->{cmd_interface},
            );
            $node_obj->apiset_must(interface => 'CLI');
            $node_obj->apiset();
            $partner_obj = NACL::STask::Node->get_partner_obj(
                command_interface => $node_obj,
            );
            $partner_name = $partner_obj->node();
            $node_name    = $node->{node};
        } catch NACL::Exceptions::HighAvailabilityError with {
            my $exception = shift;
            $Log->comment(
                "HA Pair is not available: " . $exception->text()
            );
            $skip = TRUE;
        } otherwise {
            my $exception = shift;
            $Log->comment($exception->text());
            $skip = TRUE;
        };

        unless ($skip) {
            push(@msg_data, {
                node    => $node_name,
                partner => $partner_name,
            });
            my $runid = $node_name . "_sfo_check_state";
            push(@node_info, {
                codespec   => "{" . $self->{id} . "::_check_state(); }",
                runid      => $runid,
                background => $background,
                params     => {
                    %{$node},
                    id                => $self->{id},
                    node_obj          => Clone::clone($node_obj),
                    partner_obj       => Clone::clone($partner_obj),
                    command_interface => Clone::clone($self->{cmd_interface}),
                }
            });
        }
    }

    my $status = $self->_run_workflow(
        objects    => \@node_info,
        operation  => "check_state",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        $Log->comment(
            "The current state of a node or its partner is not the same as the " .
            "expected state."
        );
        return FALSE;
    }
} # End of check_state

#===============================================================================
# Function: wait_for_state
#   This method is used to wait on the state of the SFO operation on the given
#   node or node pair.
#
#   Parameters:
#     objects - Array of hashes containing node names and SFO states
#     background - Wait for nodes to reach a state on sevaral nodes in parallel.
#       Defaults to TRUE
#     iterations - Wait for an SFO state 'n' number of times. Defaults to 1
#
#     Wait for state attributes:
#       'node'          - name of the node
#       'node_state'    - Expected state for the main node
#       'partner_state' - Expected state for the partner node
#
#       Valid states:
#         'up'               - When the node is up with the OS booted
#         'takeover'         - When the node is in takeover mode
#         'givewait'         - When the node is at 'Press ctrl-C for maintainance mode
#           storage-failover aggregates with storage-failover policy to the partner
#         'any_takeover'     - When the node is any takeover state
#         'wfg'              - When the node is in "waiting for giveback" prompt
#         'sparecore'        - When partner selects sparecore for dumping core
#         'dumpcore'         - When partner selects spraycore for dumping core
#         'node_unreachable' - When the partner is unreachable
#         'wfg_or_givewait'  - When the node is in *any* of the "waiting for
#           giveback" prompts.
#         'partial_giveback' - When node has not sent all of the
#         'takeover_autogiveback_scheduled' - State when the node is in takeover
#           and autogiveback is scheduled
#         'system_id_changed_on_partner'- State when the partner system id is
#           changed after headswap
#
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $Sfo->wait_for_state(
#     >      objects    => [{
#     >          'node'          => 'ssan-2240-2a',
#     >          'node_state'    => 'up',
#     >          'partner_state' => 'up',
#     >      ]},
#     >      background => TRUE,
#     > );
#===============================================================================
sub wait_for_state {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $nodes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @node_info = ();
    my @msg_data  = ();
    foreach my $node (@{$nodes}) {
        my $node_obj     = undef;
        my $node_name    = '';
        my $partner_obj  = undef;
        my $partner_name = '';
        my $skip         = FALSE;
        try {
            $node_obj = NACL::C::Node->new(
                node              => $node->{node},
                command_interface => $self->{cmd_interface},
            );
            $node_obj->apiset_must(interface => 'CLI');
            $node_obj->apiset();
            $partner_obj = NACL::STask::Node->get_partner_obj(
                command_interface => $node_obj,
            );
            $partner_name = $partner_obj->node();
            $node_name    = $node->{node};
        } catch NACL::Exceptions::HighAvailabilityError with {
            my $exception = shift;
            $Log->comment(
                "HA Pair is not available: " . $exception->text()
            );
            $skip = TRUE;
        } otherwise {
            my $exception = shift;
            $Log->comment($exception->text());
            $skip = TRUE;
        };
        push(@msg_data, {node => $node_name, partner => $partner_name});
        my $runid = $node_name . "_sfo_wait_for_state";

        unless ($skip) {
            push(@node_info, {
                codespec   => "{" . $self->{id} . "::_wait_for_state(); }",
                runid      => $runid,
                background => $background,
                params     => {
                    %{$node},
                    id                => $self->{id},
                    node_name         => $node_name,
                    partner_name      => $partner_name,
                    node_obj          => Clone::clone($node_obj),
                    partner_obj       => Clone::clone($partner_obj),
                    command_interface => Clone::clone($self->{cmd_interface}),
                }
            });
        }
    }

    my $status = $self->_run_workflow(
        objects    => \@node_info,
        operation  => "wait_for_state",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Wait for storage failover state failed");
        return FALSE;
    }
} # End of wait_for_state

#===============================================================================
# Function: takeover
#   Initiate takeover. This method provides additional services that are
#   inherited from NACL::C::StorageFailover to perform takeover with the
#   various options provided.
#
#   Parameters:
#     objects - Array of hashes containing node names and takeover options
#     background - Execute takeover on sevaral nodes in parallel.
#       Defaults to TRUE
#     iterations - Execute takeover 'n' number of times. Defaults to 1
#
#     Takeover options:
#       'node' - Name of the node
#       'partner - (Optional) Name of the partner node
#       'type' - Type of takeover to perform. Possible values are
#         "takeover_command", "panic", "reboot", "halt", "system_powercycle",
#         "system_power_off, "system_reset", "watchdog_reset".
#         Default is "takeover_command".
#       'option' - (Optional) Takeover options. Possible values are "normal",
#         "immediate", "force", "allow-version-mismatch",
#         "allow-disk-inventory-mismatch".
#         Default is "normal".
#       'polling_interval' - (Optional) This is the interval (in seconds)
#         after which the status of TO will be polled. If not specified it
#         defaults to 10 seconds.
#       'method_timeout' - (Optional) Time in seconds to wait for the completion
#         of the Takeover operation. Default is 600.
#       'takeover_expected' - (Optional) Boolean value. Defaults to 1.
#         Option to indicate if takeover is expected to happen or not. Based
#         on this option the states of the node and its partner are set.
#         By default it takes the value as 1 , indicating that takeover is
#         expected to happen.
#       'failure_reason' - (Optional) The reason for which the takeover is
#         expected to fail. Options include: "version_mismatch",
#         "nvram_logs_unsynced".
#       'nacl_verify' - (Optional) Boolean value. Defaults to 0.
#         If 1, then check_state() will be invoked to verify complete takeover
#         scenario.
#       'nacltask_inhibit_takeover' - (Optional) Boolean value. Defaults to 1.
#         If the value is set to 1, then a reboot command is sent with
#         inhibit-takeover flag set to true.
#       'nacltask_boot_after_halt' - (Optional) Option to indicate if the partner
#         node needs to be booted using boot_ontap after issuing a halt.
#         Default is 1 which means that the partner will be waiting in the
#         Firmware prompt.
#       'nacltask_verify_node_sfo_state' - (Optional) If 1, then verify the state
#         of the local node. Default is 1.
#       'nacltask_verify_aggr_state_owner' - (Optional)  If 1, then verify aggr
#         ownership. Default is 1.
#       'nacltask_verify_ems' - (Optional) If 1 verify that that ems
#         messages are generated and giveback duration is within limits.
#         Default is 1.
#
#     Assumptions:
#       Takeover can occur in the following cases:
#          a. takeover due to panic
#          b. takeover due to reboot
#          c. takeover due to halt
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $Sfo->takeover(
#     >      objects    => [{
#     >          'node' => 'ssan-2240-2a',
#     >          'type' => 'reboot',
#     >      ]},
#     >      background => TRUE,
#     > );
#===============================================================================
sub takeover {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $nodes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @node_info = ();
    my @msg_data  = ();
    foreach my $node (@{$nodes}) {
        # If HA was initialized, it means we have HA objects to operate and
        # there's no need to create and discover partner objects.
        my $runid  = $node->{node} . "_takeover";
        my $ha_obj = undef;
        my $skip   = FALSE;
        if (defined($self->{ha}->{$node->{node}})) {
            $ha_obj = $self->{ha}->{$node->{node}};
        } else {
            try {
                my $node_obj = NACL::C::Node->new(
                    node              => $node->{node},
                    command_interface => $self->{cmd_interface},
                );
                $node_obj->apiset_must(interface => 'CLI');
                $node_obj->apiset();
                $ha_obj = NACL::MTask::HA->new(
                    node                 => $node_obj,
                    reset_flags          => FALSE,
                    verify_cross_connect => FALSE,
                );
            } otherwise {
                my $exception = shift;
                $Log->comment($exception->text());
                $skip = TRUE;
            }
        }

        unless ($skip) {
            my %msg_args = (node => $node->{node});
            if (defined($node->{partner})) {
                $msg_args{partner} = $node->{partner};
            }
            push(@msg_data, \%msg_args);
            push(@node_info, {
                codespec   => "{" . $self->{id} . "::_takeover(); }",
                runid      => $runid,
                background => $background,
                params     => {
                    %{$node},
                    id                => $self->{id},
                    ha_obj            => Clone::clone($ha_obj),
                    command_interface => Clone::clone($self->{cmd_interface}),
                }
            });
        }
    }

    my $status = $self->_run_workflow(
        objects    => \@node_info,
        operation  => "takeover",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Storage Takeover failed");
        return FALSE;
    }
} # End of takeover

#===============================================================================
# Function: giveback
#   Initiate giveback. This method provides additional services that are
#   inherited from NACL::C::StorageFailover to perform giveback based on
#   a certain storage-failover flags
#
#   Parameters:
#     objects - Array of hashes containing node names and giveback options
#     background - Execute giveback on sevaral nodes in parallel. Defaults to TRUE
#     iterations - Execute giveback 'n' number of times. Defaults to 1
#
#     Giveback options:
#      'node' - Name of the node
#      'partner' - (Optional) Name of the partner node
#      'polling_interval' - (Optional) This is the interval (in seconds)
#        after which the status of TO will be polled. If not specified it
#        defaults to 10 seconds.
#      'method_timeout' - (Optional) Time in seconds to wait for the completion
#        of the Takeover operation. Default : 600.
#      'giveback_expected' - (Optional) Boolean value. If 1 then giveback is
#        expected to complete successfully. It will wait for the nodes to come
#        back up. Defaults : 1.
#      'auto_giveback_expected' - (Optional) User input flag that is used to
#        decide whether an auto giveback is expected or not. Example, if
#        takeover happens due to partner reboot then even if the autogiveback
#        flag is disabled, a autogiveback is expected to happen. Default : 0
#      'failure_reason' - (Optional) User input flag that is used to exit
#        giveback in case of failure. Options include: "autosupport_vetoed".
#      'wait_for_asup_generation' - (Optional) user input flag that is used to
#        wait as along as autosupport is being generated if giveback is
#        vetoed due to autosupport.  Default : 'false'
#      'require_partner_waiting' - (Optional) User input flag that makes the
#        takeover node giveback only after the partner is at waiting for
#        giveback prompt. Default : 'false'
#      'wait_for_partner' - (Optional) Boolean. If 1, the library will not send
#        the giveback command until the node is wfg. Default : 0.
#      'override_vetoes' - (Optional) user input flag that allows giveback to
#        happen even when giveback is vetoed due to any reason Default : 'true'
#      'only_cfo_aggregates' - (Optional) user input flag that is used to
#        giveback only cfo style aggregates. Default : 'true'
#      'wait_for_partner' - (Optional) user input flag that is used to wait
#        for partner to come to wfg state before issuing a giveback. Default : 1
#      'nacltask_verify_arl_ems' - (Optional) Boolean. If 1 verify  temporary.
#        aggregate relocation ems. Default : 1
#      'nacltask_verify_ems' - (Optional) Boolean. If 1 verify  ems messages
#        generated and giveback duration limits. Default : 1
#      'nacltask_verify_ems' - (Optional) Boolean. If 1 verify  ems messages
#        generated and giveback duration limits. Default : 1
#      'nacltask_verify_node_sfo_state' - (Optional) Boolean. If 1, then verify
#        the state of the local node. Default : 1.
#      'nacltask_verify_aggr_state_owner' - (Optional) Boolean.  If 1, then verify
#        aggr ownership. Default : 1.
#      'nacltask_wait_for_asup_generation' - (Optional) If 1, that the appropriate
#        asp's are generated. Default : 1.
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $Sfo->giveback(
#     >      objects    => [{
#     >          'node'    => 'ssan-2240-2a',
#     >      ]},
#     >      background => TRUE,
#     > );
#===============================================================================
sub giveback {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $nodes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @node_info = ();
    my @msg_data  = ();
    foreach my $node (@{$nodes}) {
        my $runid  = $node->{node} . "_giveback";
        my $ha_obj = undef;
        my $skip   = FALSE;
        if (defined($self->{ha}->{$node->{node}})) {
            $ha_obj = $self->{ha}->{$node->{node}};
        } else {
            try {
                my $node_obj = NACL::C::Node->new(
                    node              => $node->{node},
                    command_interface => $self->{cmd_interface},
                );
                $node_obj->apiset_must(interface => 'CLI');
                $node_obj->apiset();
                $ha_obj = NACL::MTask::HA->new(
                    node                 => $node_obj,
                    reset_flags          => FALSE,
                    verify_cross_connect => FALSE,
                );
            } otherwise {
                my $exception = shift;
                $Log->comment($exception->text());
                $skip = TRUE;
            }
        }

        unless ($skip) {
            my %msg_args = (node => $node->{node});
            if (defined($node->{partner})) {
                $msg_args{partner} = $node->{partner};
            }
            push(@msg_data, \%msg_args);
            push(@node_info, {
                codespec   => "{" . $self->{id} . "::_giveback(); }",
                runid      => $runid,
                background => $background,
                params     => {
                    %{$node},
                    id                => $self->{id},
                    ha_obj            => Clone::clone($ha_obj),
                    command_interface => Clone::clone($self->{cmd_interface}),
                }
            });
        }
    }

    my $status = $self->_run_workflow(
        objects    => \@node_info,
        operation  => "giveback",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Storage Giveback failed");
        return FALSE;
    }
} # End of giveback

#===============================================================================
# Function: takeover_giveback
#   SFO Ultimate combo. Perform a takeover followed by a giveback, allowing you
#   to execute them back-to-back, with a delay in between, or with a wait for
#   message in between.
#
#   Parameters:
#     objects - Array of hashes containing node names, takeover and giveback
#       options.
#     iterations - The number of takeover/giveback combos to run.
#     background - Execute takeover/giveback on sevaral nodes in parallel.
#       Defaults to TRUE
#     message - (Optional) A string representing the message you'd like to
#       wait for before triggering the giveback operation. Message should be
#       sent to the module's exchange.
#     message_timeout - (Optional) An integer specifying the number of seconds
#       to wait for the message that triggers the giveback.  After the time
#       expires, the wait will abort and the giveback will happen. Defaults to
#       14400 seconds (4 hours)
#     delay - (Optional) An integer representing the amount of seconds you'd
#       like to sleep between a takeover and a giveback. Defaults to 0. The
#       giveback command will be triggered as soon as the partner is wfg.
#    takeover_params - A hash ref containing any of the valid params for the
#        takeover method. See takeover's API for details
#     giveback_params - A hash ref containing any of the valid params for the
#        takeover method. See giveback's API for details
#
#     Takeover/Giveback options:
#      'node' - Name of the node
#      'partner' - (Optional) Name of the partner node
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  # My insane sfo scenario
#     >  my $status = $Sfo->takeover_giveback(
#     >      objects    => [{
#     >          'node'    => 'ssan-2240-2a',
#     >      ]},
#     >      background => TRUE,
#     >      iterations => 99,
#     >      message    => 'give_em_back!'
#     >     takeover_params => { type => 'reboot' },
#     >     giveback_params => { override_vetoes => 'true' }
#     > );
#===============================================================================
sub takeover_giveback {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $nodes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    my $message = &ParamsMgr::get_arg(
        'arg'      => "message",
        'required' => FALSE,
        'default'  => undef,
    );
    my $message_timeout = &ParamsMgr::get_arg(
        'arg'      => "message_timeout",
        'required' => FALSE,
        'default'  => MESSAGE_TIMEOUT,
    );
    my $delay = &ParamsMgr::get_arg(
        'arg'      => "delay",
        'required' => FALSE,
        'default'  => 0,
    );
    my $takeover_params = &ParamsMgr::get_arg(
        'arg'      => "takeover_params",
        'required' => TRUE,
    );
    my $giveback_params = &ParamsMgr::get_arg(
        'arg'      => "giveback_params",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    # Disable autogiveback to control when we want to execute the giveback cmd
    my $status = TRUE;
    try {
        my @tmp_nodes = @{$nodes};
        foreach my $node (@tmp_nodes) {
            $node->{'auto-giveback'}             = "false";
        }
        $self->modify(objects => \@tmp_nodes);
    } otherwise {
        my $exception = shift;
        $Log->comment(
            "Failed to disable autogiveback. This will cause the nodes to " .
            "giveback aggregates automatically, overriding any delay or " .
            "wait for message attributes. " . $exception->text()
        );
    };

    # Extract the set of parameters for takegive and giveback
    my @nodes_takeover = @{$nodes};
    foreach my $node (@nodes_takeover) {
        while (my ($key, $value) = each (%{$takeover_params})) {
            $node->{$key} = $value;
        }
    }
    my @nodes_giveback = @{$nodes};
    foreach my $node (@nodes_giveback) {
        while (my ($key, $value) = each (%{$giveback_params})) {
            $node->{$key} = $value;
        }
    }

    try {
        for (my $i = 0; $i < $iterations; $i++) {
            $Log->comment("Takeover Giveback - Iteration $i");
            $MessageBroker->publish_message(
                exchange    => ID,
                routing_key => '#',
                message     => {
                    msg  => 'takeover_giveback.started',
                    data => {
                        iteration => $i,
                    },
                }
            );
            $self->takeover(objects => \@nodes_takeover);
            unless ($delay == 0) {
                $Log->comment(
                    "Takeover completed successfully. Sleeping $delay seconds " .
                    "before executing a giveback."
                );
                sleep($delay);
            }

            # If there's a message, wait for it to arrive before giveback
            if (defined($message)) {
                $Log->comment(
                    "Waiting for $message before issuing a giveback command."
                );
                $MessageBroker->wait_for_message(
                    message => $message,
                    queue   => ID,
                    timeout => $message_timeout,
                );
            }

            $self->giveback(objects => \@nodes_giveback);
            $MessageBroker->publish_message(
                exchange    => ID,
                routing_key => '#',
                message     => {
                    msg  => 'takeover_giveback.finished',
                    data => {
                        iteration => $i,
                    },
                }
            );
        }
    } otherwise {
        my $exception = shift;
        $Log->comment($exception->text());
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg  => 'takeover_giveback.failed',
                    data => {
                        error => $exception->text()
                    },
                }
            );
        $status = FALSE;
    };

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Takeover/Giveback failed");
        return FALSE;
    }
} # End of takeover_giveback

#===============================================================================
# Function: get_ha_pairs
#   Returns a list of HA pairs containing node names.
#
#   Parameters:
#     unique - Return unique pairs. i.e {node => A, partner => B},
#       {node => C, partner => D}, and NOT {node => B, partner => A} ...
#
#   Returns:
#     [{node => 'node_name', partner => 'partner_name}, {...}, {...}]
#
#
#   Example:
#     >  my @nodes = $Sfo->get_ha_pairs();
#===============================================================================
sub get_ha_pairs {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $unique = &ParamsMgr::get_arg(
        'arg'      => "unique",
        'required' => FALSE,
        'default'  => FALSE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return wantarray ? () : [];
    }

    my @ha_pairs = ();
    my @nodes    = NACL::C::Node->find(
        command_interface => $self->{cmd_interface}
    );

    my %seen = ();
    foreach my $node (@nodes) {
        try {
            my $node_name = $node->node();
            unless (defined($seen{$node_name})) {
                my $partner   = NACL::STask::Node->get_partner_obj(
                    command_interface => $node,
                );
                my $partner_name = $partner->node();
                %seen = (%seen, "$partner_name" => TRUE) if $unique;
                push(@ha_pairs, {node => $node_name, partner => $partner_name});
            }
        } catch NACL::Exceptions::HighAvailabilityError with {
            my $exception = shift;
            $Log->comment(
                "HA Pair is not available: " . $exception->text()
            );
        } otherwise {
            my $exception = shift;
            $Log->comment($exception->text());
        };
    }

    return wantarray ? @ha_pairs : \@ha_pairs;
} # End of get_ha_pairs

#===============================================================================
# PRIVATE METHODS
#===============================================================================

#===============================================================================
# Description: Subprocess responsible for enabling storage failover
#===============================================================================
sub _enable {
    my ($class) = @_;

    my $status    = FALSE;
    my $params    = NATE::ParamSet->get_global();
    my $node_obj  = $params->get('node_obj');
    my $node_name = $params->get('node_name', default => $node_obj->node());
    my $command_interface =  $params->get('command_interface');

    try {
        $command_interface->refresh_command_interface(no_version_change => TRUE);
        $node_obj->refresh_command_interface(no_version_change => TRUE);
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.enable.started',
                data => {
                    node => $node_name,
                },
            }
        );
        NACL::STask::StorageFailover->enable(
            command_interface => $command_interface,
            node              => $node_obj,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed enabling SFO on this node');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'sfo.enable.failed',
                data => {
                    node  => $node_name,
                    error => $exception->text(),
                },
            }
        );
        $Log->log(NATE::Result::Fail->new($exception->text()));
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.enable.finished',
                data => {
                    node => $node_name,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Enabled failover successfully'));
    } else {
        $Log->log(NATE::Result::Fail->new('Failed to enable failover'));
    }
} # End of _enable

#===============================================================================
# Description: Subprocess responsible for disabling storage failover
#===============================================================================
sub _disable {
    my ($class) = @_;

    my $status    = FALSE;
    my $params    = NATE::ParamSet->get_global();
    my $node_obj  = $params->get('node_obj');
    my $node_name = $params->get('node_name', default => $node_obj->node());
    my $command_interface =  $params->get('command_interface');

    $MessageBroker->publish_message(
        exchange    => ID,
        routing_key => '#',
        message     => {
            msg  => 'sfo.disable.started',
            data => {
                node => $node_name,
            },
        }
    );

    try {
        $command_interface->refresh_command_interface(no_version_change => TRUE);
        $node_obj->refresh_command_interface(no_version_change => TRUE);
        NACL::STask::StorageFailover->disable(
            command_interface => $command_interface,
            node              => $node_obj,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed disabling SFO on this node');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'sfo.disable.failed',
                data => {
                    node => $node_name,
                    error => $exception->text(),
                },
            }
        );
        $Log->log(NATE::Result::Fail->new($exception->text()));
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.disable.finished',
                data => {
                    node => $node_name,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Disables failover successfully'));
    } else {
        $Log->log(NATE::Result::Fail->new('Failed to disable failover'));
    }
} # End of _disable

#===============================================================================
# Description: Subprocess responsible for initializing storage failover
#===============================================================================
sub _initialize {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $node = &ParamsMgr::get_arg(
        'arg'      => "node",
        'required' => TRUE,
    );
    my $node_name = &ParamsMgr::get_arg(
        'arg'      => "node_name",
        'required' => FALSE,
        'default'  => $node->node(),
    );
    my $reset_flags = &ParamsMgr::get_arg(
        'arg'      => "reset_flags",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $verify_cross_connect = &ParamsMgr::get_arg(
        'arg'      => "verify_cross_connect",
        'required' => FALSE,
        'default'  => TRUE,
    );
    if (&ParamsMgr::error()) {
        return FALSE;
    }

    my $status = FALSE;
    my $ha     = undef;
    $MessageBroker->publish_message(
        exchange    => ID,
        routing_key => '#',
        message     => {
            msg  => 'sfo.initialize.started',
            data => {
                node    => $node_name,
            },
        }
    );
    try {
        $node->refresh_command_interface(no_version_change => TRUE);
        $ha = NACL::MTask::HA->new(
            node                 => $node,
            reset_flags          => $reset_flags,
            verify_cross_connect => FALSE,
       );
        $status = TRUE;
        $self->{ha}->{$node_name} = $ha;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /HA partner not available/i) {
            $Log->comment(
                "$node_name's partner is down or unreachable. " .
                "Skipping initialization"
            );
            $status = TRUE;
        } elsif ($exception->text() =~ /Could not create connection/i) {
            $Log->comment(
                "$node_name is not reachable. Skipping initialization."
            );
            $status = TRUE;
        } else {
            $Log->comment($exception->text());
            $status = FALSE;
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg  => 'sfo.initialize.failed',
                    data => {
                        node  => $node_name,
                        error => $exception->text(),
                    }
                }
            );
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.initialize.finished',
                data => {
                    node => $node_name,
                },
            }
        );
        return TRUE;
    } else {
        $Log->comment('SFO initialization failed');
        return FALSE;
    }
} # End of _initialize

#===============================================================================
# Description: Subprocess responsible for modifying storage failover
#===============================================================================
sub _modify {
    my ($class) = @_;

    my $params             = NATE::ParamSet->get_global();
    my $onpanic            = $params->get('onpanic',        default => 'true');
    my $node_obj           = $params->get('node_obj',       default => undef);
    my $onreboot           = $params->get('onreboot',       default => 'true');
    my $node_name          = $params->get('node_name',      default => $node_obj->node());
    my $onfailure          = $params->get('onfailure',      default => 'true');
    my $auto_giveback      = $params->get('auto-giveback',  default => 'true');
    my $check_partner      = $params->get('check-partner',  default => 'true');
    my $delay_seconds      = $params->get('delay-seconds',  default => DELAY);
    my $detection_time     = $params->get('detection-time', default => DETECTION_TIME);
    my $command_interface  = $params->get('command_interface');
    my $auto_give_on_panic = $params->get('auto-giveback-after-panic', default => 'true');
    my $status = FALSE;

    $MessageBroker->publish_message(
        exchange    => ID,
        routing_key => '#',
        message     => {
            msg  => 'sfo.modify.started',
            data => {
                node                        => $node_name,
                onpanic                     => $onpanic,
                onreboot                    => $onreboot,
                onfailure                   => $onfailure,
                'auto-giveback'             => $auto_giveback,
                'delay-seconds'             => $delay_seconds,
                'check-partner'             => $check_partner,
                'detection-time'            => $detection_time,
                'auto-giveback-after-panic' => $auto_give_on_panic,
            },
        }
    );

    try {
        $command_interface->refresh_command_interface(no_version_change => TRUE);
        NACL::STask::StorageFailover->modify(
            node                        => $node_obj,
            onpanic                     => $onpanic,
            onreboot                    => $onreboot,
            onfailure                   => $onfailure,
            'auto-giveback'             => $auto_giveback,
            'delay-seconds'             => $delay_seconds,
            'check-partner'             => $check_partner,
            'detection-time'            => $detection_time,
            command_interface           => $command_interface,
            'auto-giveback-after-panic' => $auto_give_on_panic,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed enabling SFO on this node');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'sfo.modify.failed',
                data => {
                    node                        => $node_name,
                    onpanic                     => $onpanic,
                    onreboot                    => $onreboot,
                    onfailure                   => $onfailure,
                    'auto-giveback'             => $auto_giveback,
                    'delay-seconds'             => $delay_seconds,
                    'check-partner'             => $check_partner,
                    'detection-time'            => $detection_time,
                    'auto-giveback-after-panic' => $auto_give_on_panic,
                    error => $exception->text(),
                },
            }
        );
        $Log->log(NATE::Result::Fail->new($exception->text()));
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.modify.finished',
                data => {
                    node                        => $node_name,
                    onpanic                     => $onpanic,
                    onreboot                    => $onreboot,
                    onfailure                   => $onfailure,
                    'auto-giveback'             => $auto_giveback,
                    'delay-seconds'             => $delay_seconds,
                    'check-partner'             => $check_partner,
                    'detection-time'            => $detection_time,
                    'auto-giveback-after-panic' => $auto_give_on_panic,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Modified SFO successfully'));
    } else {
        $Log->log(NATE::Result::Fail->new('Failed modifying SFO'));
    }
} # End of _modify

#===============================================================================
# Description: Subprocess responsible for modifying storage failover
#===============================================================================
sub _check_state {
    my ($class) = @_;

    my $params            = NATE::ParamSet->get_global();
    my $node_obj          = $params->get('node_obj');
    my $partner_obj       = $params->get('partner_obj');
    my $node_state        = $params->get('node_state',    default => 'up');
    my $partner_state     = $params->get('partner_state', default => 'up');
    my $command_interface = $params->get('command_interface');
    my $status            = FALSE;
    my $node_name         = $node_obj->node();
    my $partner_name      = $partner_obj->node();

    $MessageBroker->publish_message(
        exchange    => ID,
        routing_key => '#',
        message     => {
            msg  => 'sfo.check_state.started',
            data => {
                node => $node_name,
                node_state => $node_state,
                partner_state => $partner_state,
            },
        }
    );

    try {
        $command_interface->refresh_command_interface(no_version_change => TRUE);
        $node_obj->refresh_command_interface(no_version_change => TRUE);
        $partner_obj->refresh_command_interface(no_version_change => TRUE);
        NACL::STask::StorageFailover->check_state(
            command_interface => $command_interface,
            node              => $node_obj,
            partner           => $partner_obj,
            node_state        => $node_state,
            partner_state     => $partner_state,
        );
        $status = TRUE;
    } catch NACL::STask::Exceptions::UnexpectedSFOState with {
        my $exception = shift;
        $Log->comment("Unexpected state: " . $exception->text());
        $status = FALSE;
    } otherwise {
        my $exception = shift;
        $Log->comment(
            'Failed checking node state on this HA pair ' . $exception->text()
        );
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'sfo.check_state.failed',
                data => {
                    node          => $node_obj->node(),
                    node_state    => $node_state,
                    partner_state => $partner_state,
                    error         => $exception->text(),
                },
            }
        );
        $Log->log(NATE::Result::Fail->new($exception->text()));
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.check_state.finished',
                data => {
                    node          => $node_obj->node(),
                    node_state    => $node_state,
                    partner_state => $partner_state,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('HA pair is in an expected state'));
    } else {
        $Log->log(NATE::Result::Fail->new(
            "'$node_name' is not '$node_state' or '$partner_name' is not " .
            "'$partner_state'"
        ));
    }
} # End of _check_state

#===============================================================================
# Description: Subprocess responsible for waiting for an HA to reach a state
#===============================================================================
sub _wait_for_state {
    my ($class) = @_;

    my $params            = NATE::ParamSet->get_global();
    my $timeout           = $params->get('method-timeout',    default => TIMEOUT);
    my $node_obj          = $params->get('node_obj',          default => undef);
    my $partner_obj       = $params->get('partner_obj',       default => undef);
    my $node_state        = $params->get('node_state',        default => undef);
    my $partner_state     = $params->get('partner_state',     default => undef);
    my $command_interface = $params->get('command_interface', default => $node_obj);
    my $status            = FALSE;
    my $node_name         = $node_obj->node();

    $MessageBroker->publish_message(
        exchange    => ID,
        routing_key => '#',
        message     => {
            msg  => 'sfo.wait_for_state.started',
            data => {
                node          => $node_name,
                node_state    => $node_state,
                partner_state => $partner_state,
            },
        }
    );

    try {
        $command_interface->refresh_command_interface(no_version_change => TRUE);
        $node_obj->refresh_command_interface(no_version_change => TRUE);
        $partner_obj->refresh_command_interface(no_version_change => TRUE);
        my %args = (
            node              => $node_obj,
            'method-timeout'  => $timeout,
            command_interface => $command_interface,
        );
        if (defined($node_state)) {
            %args = (%args, node_state => $node_state);
        }
        if (defined($partner_state)) {
            %args = (%args, partner_state => $partner_state);
        }
        if (defined($partner_obj)) {
            %args = (%args, partner => $partner_obj);
        }
        NACL::STask::StorageFailover->wait_for_state(%args);
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment("Failed waiting for the HA pair to reach $node_state");
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'sfo.wait_for_state.failed',
                data => {
                    node          => $node_name,
                    node_state    => $node_state,
                    partner_state => $partner_state,
                    error         => $exception->text(),
                },
            }
        );
        $Log->log(NATE::Result::Fail->new($exception->text()));
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.wait_for_state.finished',
                data => {
                    node          => $node_name,
                    node_state    => $node_state,
                    partner_state => $partner_state,
                },
            }
        );
        $Log->log(
            NATE::Result::Pass->new('HA pair has reached an expected state'
        ));
    } else {
        $Log->log(NATE::Result::Fail->new(
            'Failed waiting for the HA pair to reach a state'
        ));
    }
} # End of _wait_for_state

#===============================================================================
# Description: Subprocess responsible for initiating a takeover operation
#===============================================================================
sub _takeover {
    my ($class) = @_;

    my $params       = NATE::ParamSet->get_global();
    my $ha_obj       = $params->get('ha_obj',  default => undef);
    my $node_name    = $params->get('node',    default => '');
    my $partner_name = $params->get('partner', default => '');
    my $status       = FALSE;

    unless (defined($ha_obj)) {
        $Log->log(NATE::Result::Fail->new(
            "An HA object is required for takeover. Takeover will not run."
        ));
        return FALSE;
    }

    my %opts = _get_takeover_params(paramset => $params);
    try {
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.takeover.start',
                data => {
                    node    => $node_name,
                    partner => $partner_name,
                },
            }
        );
        $ha_obj->{node}->refresh_command_interface(no_version_change => TRUE);
        $ha_obj->takeover(%opts);
        $status = TRUE;
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.takeover.finished',
                data => {
                    node    => $node_name,
                    partner => $partner_name,
                },
            }
        );
    } otherwise {
        my $exception = shift;
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'sfo.takeover.failed',
                data => {
                    node    => $node_name,
                    partner => $partner_name,
                    error   => $exception->text(),
                },
            }
        );
        $Log->comment($exception);
        $status = FALSE;
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Takeover completed successfully'));
    } else {
        $Log->log(NATE::Result::Fail->new(
            'Takeover did not complete successfully'
        ));
    };
} # End of _takeover

#===============================================================================
# Description: Subprocess responsible for initiating a giveback operation
#===============================================================================
sub _giveback {
    my ($class) = @_;

    my $params       = NATE::ParamSet->get_global();
    my $ha_obj       = $params->get('ha_obj',  default => undef);
    my $node_name    = $params->get('node',    default => '');
    my $partner_name = $params->get('partner', default => '');
    my $status       = FALSE;

    unless (defined($ha_obj)) {
        $Log->log(NATE::Result::Fail->new(
            "An HA object is required for giveback. Giveback will not run."
        ));
        return FALSE;
    }

    my %opts = _get_giveback_params(paramset => $params);
    try {
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.giveback.start',
                data => {
                    node    => $node_name,
                    partner => $partner_name,
                },
            }
        );
        $ha_obj->{node}->refresh_command_interface(no_version_change => TRUE);
        $ha_obj->giveback(%opts);
        $status = TRUE;
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.giveback.finished',
                data => {
                    node    => $node_name,
                    partner => $partner_name,
                },
            }
        );
    } catch NACL::STask::Exceptions::UnexpectedSFOState with {
        my $exception = shift;
        $Log->comment(
            "'$partner_name' is not waiting for giveback, so the operation " .
            "will not be executed."
        );
        $MessageBroker->publish_message(
            exchange    => ID,
            routing_key => '#',
            message     => {
                msg  => 'sfo.giveback.skipped',
                data => {
                    node    => $node_name,
                    partner => $partner_name,
                },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'sfo.giveback.failed',
                data => {
                    node    => $node_name,
                    partner => $partner_name,
                    error   => $exception->text(),
                },
            }
        );
        $Log->comment($exception);
        $status = FALSE;
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Giveback completed successfully'));
    } else {
        $Log->log(NATE::Result::Fail->new(
            'Giveback did not complete successfully'
        ));
    };
} # End of _giveback

#===============================================================================
# HELPER METHODS
#===============================================================================

#===============================================================================
# Description: Extract and fill in the parameters required for the takeover
# API.
#===============================================================================
sub _get_takeover_params {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $paramset = &ParamsMgr::get_arg(
        'arg'      => "paramset",
        'required' => TRUE,
    );

    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return ();
    }

    # Start extracting parameters from the horrible paramset
    my %takeover_params = ();

    $takeover_params{type} = _extract_param(
        paramset => $paramset,
        arg      => "type",
        default  => DEFAULT_TAKEOVER_TYPE,
    );
    $takeover_params{option} = _extract_param(
        paramset => $paramset,
        arg      => "option",
        default  => DEFAULT_TAKEOVER_OPTION,
    );
    $takeover_params{polling_interval} = _extract_param(
        paramset => $paramset,
        arg      => "polling_interval",
        default  => POLL_INTERVAL,
    );
    $takeover_params{'method-timeout'} = _extract_param(
        paramset => $paramset,
        arg      => "method_timeout",
        default  => DEFAULT_METHOD_TIMEOUT,
    );
    $takeover_params{takeover_expected} = _extract_param(
        paramset => $paramset,
        arg      => "takeover_expected",
        default  => TAKEOVER_EXPECTED,
    );
    $takeover_params{failure_reason} = _extract_param(
        paramset => $paramset,
        arg      => "failure_reason",
        default  => FAILURE_REASON,
    );
    $takeover_params{nacl_verify} = _extract_param(
        paramset => $paramset,
        arg      => "nacl_verify",
        default  => TRUE,
    );
    $takeover_params{nacltask_verify_node_sfo_state} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_verify_node_sfo_state",
        default  => VERIFY_SFO_STATE,
    );
    $takeover_params{nacltask_verify_aggr_state_owner} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_verify_aggr_state_owner",
        default  => VERIFY_AGGR_STATE_OWNER,
    );
    $takeover_params{nacltask_verify_ems} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_verify_ems",
        default  => VERIFY_EMS,
    );
    $takeover_params{nacltask_inhibit_takeover} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_inhibit_takeover",
        default  => INHIBIT_TAKEOVER,
    );
    $takeover_params{nacltask_boot_after_halt} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_boot_after_halt",
        default  => BOOT_AFTER_HALT,
    );

    foreach my $key (keys(%takeover_params)) {
        unless (defined($takeover_params{$key})) {
            delete $takeover_params{$key};
        }
    }

    $Log->comment(
        "Parameters used for the takeover: \n" . Dumper(\%takeover_params)
    );

    return %takeover_params;
}  # End of _get_takeover_params

#===============================================================================
# Description: Extract and fill in the parameters required for the giveback
# API.
#===============================================================================
sub _get_giveback_params {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $paramset = &ParamsMgr::get_arg(
        'arg'      => "paramset",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return ();
    }

    # Start extracting parameters from the horrible paramset
    my %giveback_params = ();
    $giveback_params{polling_interval} = _extract_param(
        paramset => $paramset,
        arg      => "polling_interval",
        default  => POLL_INTERVAL,
    );
    $giveback_params{'method_timeout'} = _extract_param(
        paramset => $paramset,
        arg      => "method_timeout",
        default  => DEFAULT_METHOD_TIMEOUT,
    );
    $giveback_params{auto_giveback_expected} = _extract_param(
        paramset => $paramset,
        arg      => "auto_giveback_expected",
        default  => AUTO_GIVEBACK_EXPECTED,
    );
    $giveback_params{giveback_expected} = _extract_param(
        paramset => $paramset,
        arg      => "giveback_expected",
        default  => GIVEBACK_EXPECTED,
    );
    $giveback_params{'failure_reason'} = _extract_param(
        paramset => $paramset,
        arg      => "failure_reason",
        default  => FAILURE_REASON,
    );
    $giveback_params{'nacltask_wait_for_asup_generation'} = _extract_param(
        paramset => $paramset,
        arg      => "wait_for_asup_generation",
        default  => WAIT_FOR_ASUP_GENERATION,
    );
    $giveback_params{'require-partner-waiting'} = _extract_param(
        paramset => $paramset,
        arg      => "require_partner_waiting",
        default  => REQUIRE_PARTNER_WAITING,
    );
    $giveback_params{'nacltask_wait_for_partner'} = _extract_param(
        paramset => $paramset,
        arg      => "wait_for_partner",
        default  => WAIT_FOR_PARTNER,
    );
    $giveback_params{'only-cfo-aggregates'} = _extract_param(
        paramset => $paramset,
        arg      => "only_cfo_aggregates",
        default  => ONLY_CFO_AGGREGATES,
    );
    $giveback_params{'override-vetoes'} = _extract_param(
        paramset => $paramset,
        arg      => "override_vetoes",
        default  => OVERRIDE_VETOES,
    );
    $giveback_params{'nacltask_verify_arl_ems'} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_verify_arl_ems",
        default  => VERIFY_ARL_EMS,
    );
    $giveback_params{'nacltask_verify_ems'} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_verify_ems",
        default  => VERIFY_EMS,
    );
    $giveback_params{'nacltask_verify_sfo_state'} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_verify_sfo_state",
        default  => VERIFY_SFO_STATE,
    );
    $giveback_params{'nacltask_verify_aggr_state_owner'} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_verify_aggr_state_owner",
        default  => VERIFY_AGGR_STATE_OWNER,
    );
    $giveback_params{'nacltask_wait_for_asup_generation'} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_wait_for_asup_generation",
        default  => WAIT_FOR_ASUP_GENERATION,
    );
    $giveback_params{'nacltask_skip_auto_giveback'} = _extract_param(
        paramset => $paramset,
        arg      => "nacltask_skip_auto_giveback",
        default  => SKIP_AUTO_GIVEBACK,
    );

    foreach my $key (keys(%giveback_params)) {
        unless (defined($giveback_params{$key})) {
            delete $giveback_params{$key};
        }
    }

    $Log->comment(
        "Parameters used for giveback: \n" . Dumper(\%giveback_params)
    );

    return %giveback_params;
}  # End of _get_giveback_params

#===============================================================================
# Description: Helper function to extract arguments and values quickly
#===============================================================================
sub _extract_param {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $paramset = &ParamsMgr::get_arg(
        'arg'      => "paramset",
        'required' => TRUE,
    );
    my $arg = &ParamsMgr::get_arg(
        'arg'      => "arg",
        'required' => TRUE,
    );
    my $default = &ParamsMgr::get_arg(
        'arg'      => "default",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return undef;
    }

    my $value = $paramset->get($arg);
    if (defined($value)) {
        return $value;
    } else {
        return $default;
    }
} # End of _extract_param

#===============================================================================
# Description: Verify shelves are cross connected and configured for SFO ops
#===============================================================================
sub _verify_disk_connectivity {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $node_name = &ParamsMgr::get_arg(
        'arg'      => "node_name",
        'required' => TRUE,
    );
    my $partner_name = &ParamsMgr::get_arg(
        'arg'      => "partner_name",
        'required' => TRUE,
    );
    my $command_interface = &ParamsMgr::get_arg(
        'arg'      => "command_interface",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my $status = FALSE;
    try {
        $Log->comment(
            "Verifying that all disk shelves are cross-connected. Note: this " .
            "required both nodes in an HA pair to be up."
        );
        NACL::STask::StorageDisk->check_cross_connect(
            command_interface => $command_interface,
            node1             => $node_name,
            node2             => $partner_name,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment(
            "Disk shelves are either not connected properly or '$node_name' " .
            "or 'partner_name' is down. " . $exception->text()
        );
        $status = FALSE;
    };

    if ($status) {
        $Log->comment(
            "Disk shelves are connected properly and it's ready for Storage " .
            "Failover testing."
        );
        return TRUE;
    } else {
        $Log->comment(
            "Disk shelf connectivity verification failed. Initialization " .
            "cannot continue. Please review the logs and make sure all nodes " .
            "are operational."
        );
        return FALSE;
    }
} # End of _verify_disk_connectivity

#===============================================================================
# MESSAGE CATALOG
#===============================================================================

#===============================================================================
# Message: enable.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "enable.started",
# >         "data" : [{"node": "ssan-2240-2a"}, {"node": "ssan-2240-2b"}],
# >     }
#
# Message: enable.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "enable.finished",
# >         "data" : [{"node": "ssan-2240-2a"}, {"node": "ssan-2240-2b"}],
# >     }
#
# Message: enable.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "enable.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: sfo.enable.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.enable.started",
# >         "data" : { "node": "ssan-2240-2a" },
# >     }
#
# Message: sfo.enable.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.enable.finished",
# >         "data" : { "node": "ssan-2240-2a" },
# >     }
#
# Message: sfo.enable.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "sfo.enable.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: disable.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "disable.started",
# >         "data" : [{"node": "ssan-2240-2a"}, {"node": "ssan-2240-2b"}],
# >     }
#
# Message: disable.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "disable.finished",
# >         "data" : [{"node": "ssan-2240-2a"}, {"node": "ssan-2240-2b"}],
# >     }
#
# Message: disable.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "disable.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: sfo.disable.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.disable.started",
# >         "data" : { "node": "ssan-2240-2a" },
# >     }
#
# Message: sfo.disable.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.disable.finished",
# >         "data" : { "node": "ssan-2240-2a" },
# >     }
#
# Message: sfo.disable.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "sfo.disable.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: initialize.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "initialize.started",
# >         "data" : [{"node": "ssan-2240-2a"}, {"node": "ssan-2240-2b"}],
# >     }
#
# Message: initialize.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "initialize.finished",
# >         "data" : [{"node": "ssan-2240-2a"}, {"node": "ssan-2240-2b"}],
# >     }
#
# Message: initialize.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "initialize.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: sfo.initialize.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.initialize.started",
# >         "data" : { "node": "ssan-2240-2a" },
# >     }
#
# Message: sfo.initialize.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.initialize.finished",
# >         "data" : { "node": "ssan-2240-2a" },
# >     }
#
# Message: sfo.initialize.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "sfo.initialize.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: modify.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "modify.started",
# >         "data" : [{"node": "ssan-2240-2a"}, {"node": "ssan-2240-2b"}],
# >     }
#
# Message: modify.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "modify.finished",
# >         "data" : [{"node": "ssan-2240-2a"}, {"node": "ssan-2240-2b"}],
# >     }
#
# Message: modify.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "modify.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: sfo.modify.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.modify.started",
# >         "data" : {
# >             "node": "ssan-2240-2a",
# >             "onpanic": "true",
# >             "onreboot": "true",
# >             "onfailure": "true",
# >             "auto-giveback": "true";
# >             "delay-seconds": "500",
# >             "check-partner":  => 1,
# >             "detection-time": => "300",
# >         },
# >     }
#
# Message: sfo.modify.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.modify.finished",
# >         "data" : {
# >             "node": "ssan-2240-2a",
# >             "onpanic": "true",
# >             "onreboot": "true",
# >             "onfailure": "true",
# >             "auto-giveback": "true";
# >             "delay-seconds": "500",
# >             "check-partner":  => 1,
# >             "detection-time": => "300",
# >         },
# >     }
#
# Message: sfo.modify.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "sfo.modify.failed,
# >         "data" : {
# >             "node": "ssan-2240-2a",
# >             "onpanic": "true",
# >             "onreboot": "true",
# >             "onfailure": "true",
# >             "auto-giveback": "true";
# >             "delay-seconds": "500",
# >             "check-partner":  => 1,
# >             "detection-time": => "300",
# >         },
# >         "error" : <exception_text>
# >     }
#
# Message: check_state.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "check_state.started",
# >         "data" : [
# >             {"node": "ssan-2240-2a", "node_state": "up", "partner_state": "up"},
# >             {"node": "ssan-2240-2b", "node_state": "up", "partner_state": "up"},
# >         ]
# >     }
#
# Message: check_state.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "check_state.finished",
# >         "data" : [
# >             {"node": "ssan-2240-2a", "node_state": "up", "partner_state": "up"},
# >             {"node": "ssan-2240-2b", "node_state": "up", "partner_state": "up"},
# >         ]
# >     }
#
# Message: check_state.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "check_state.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: sfo.check_state.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.check_state.started",
# >         "data" : {"node": "ssan-2240-2a", "node_state": "up", "partner_state": "up"},
# >     }
#
# Message: sfo.check_state.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.check_state.finished",
# >         "data" : {"node": "ssan-2240-2a", "node_state": "up", "partner_state": "up"},
# >     }
#
# Message: sfo.check_state.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "sfo.check_state.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: wait_for_state.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "wait_for_state.started",
# >         "data" : [
# >             {"node": "ssan-2240-2a", "node_state": "up", "partner_state": "up"},
# >             {"node": "ssan-2240-2b", "node_state": "up", "partner_state": "up"},
# >         ]
# >     }
#
# Message: wait_for_state.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "wait_for_state.finished",
# >         "data" : [
# >             {"node": "ssan-2240-2a", "node_state": "up", "partner_state": "up"},
# >             {"node": "ssan-2240-2b", "node_state": "up", "partner_state": "up"},
# >         ]
# >     }
#
# Message: wait_for_state.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "wait_for_state.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: sfo.wait_for_state.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.wait_for_state.started",
# >         "data" : {"node": "ssan-2240-2a", "node_state": "up", "partner_state": "up"},
# >     }
#
# Message: sfo.wait_for_state.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.wait_for_state.finished",
# >         "data" : {"node": "ssan-2240-2a", "node_state": "up", "partner_state": "up"},
# >     }
#
# Message: sfo.wait_for_state.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "sfo.wait_for_state.failed,
# >         "data" : { "node": "ssan-2240-2a" },
# >         "error" : <exception_text>
# >     }
#
# Message: takeover.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "takeover.started",
# >         "data" : [
# >             {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"},
# >             {"node": "ssan-2240-3a", "partner": "ssan-2240-3b"},
# >         ]
# >     }
#
# Message: takeover.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "takeover.finished",
# >         "data" : [
# >             {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"},
# >             {"node": "ssan-2240-3a", "partner": "ssan-2240-3b"},
# >         ]
# >     }
#
# Message: takeover.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "takeover.failed,
# >         "data" : [
# >             {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"},
# >             {"node": "ssan-2240-3a", "partner": "ssan-2240-3b"},
# >         ]
# >         "error" : <exception_text>
# >     }
#
# Message: sfo.takeover.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.takeover.started",
# >         "data" : {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"}
# >     }
#
# Message: sfo.takeover.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.takeover.finished",
# >         "data" : {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"}
# >     }
#
# Message: sfo.takeover.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "sfo.takeover.failed,
# >         "data" : {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"}
# >         "error" : <exception_text>
# >     }
#
# Message: giveback.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "giveback.started",
# >         "data" : [
# >             {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"},
# >             {"node": "ssan-2240-3a", "partner": "ssan-2240-3b"},
# >         ]
# >     }
#
# Message: giveback.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "giveback.finished",
# >         "data" : [
# >             {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"},
# >             {"node": "ssan-2240-3a", "partner": "ssan-2240-3b"},
# >         ]
# >     }
#
# Message: giveback.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "giveback.failed,
# >         "data" : [
# >             {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"},
# >             {"node": "ssan-2240-3a", "partner": "ssan-2240-3b"},
# >         ]
# >         "error" : <exception_text>
# >     }
#
# Message: sfo.giveback.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.giveback.started",
# >         "data" : {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"}
# >     }
#
# Message: sfo.giveback.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.giveback.finished",
# >         "data" : {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"}
# >     }
#
# Message: sfo.giveback.skipped
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.giveback.skipped",
# >         "data" : {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"}
# >     }
#
# Message: sfo.giveback.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "sfo.giveback.failed,
# >         "data" : {"node": "ssan-2240-2a", "partner": "ssan-2240-2b"}
# >         "error" : <exception_text>
# >     }
#
# Message: takeover_giveback.started
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.takeover_giveback.started",
# >         "data" : {"iteration": "1"}
# >     }
#
# Message: takeover_giveback.finished
#   Exchange:
#       San::FIT::Misfit::Features::StorageFailover
#
#   Message Structure:
# >     {
# >         "msg": "sfo.giveback.finished",
# >         "data" : {"iteration": "1"}
# >     }
#
# Message: takeover_giveback.failed
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "sfo.takeover_giveback.failed,
# >         "data" : {"iteration" : "1"}
# >         "error" : <exception_text>
# >     }
#

1;