# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Vault module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Features::Vault
#   Wrapper for NACL snapmirror operations with Misfit messaging for XDP/Vault
#   snapmirror relationships
#
#   To use this module you must specify the CLUSTER_NAME parameter.  For
#   intercluster functionality, you must additionally add the PEER_CLUSTER
#   parameter.

package San::FIT::Misfit::Features::Vault;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Cluster;
use NACL::C::Snapmirror;
use NACL::STask::Cluster;
use NACL::STask::Snapmirror;
use NACL::STask::VserverPeer;
use NATE::BaseException;
use San::FIT::Misfit::Core::Workflow;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::MessageBroker;

# Inherit base class - Misfit Base
use base qw(San::FIT::Misfit::Core::Base);

use constant {
    TIMEOUT      => 360,
    INIT_TIMEOUT => 600
};

# Main broker used by this module
my $MessageBroker;

#===============================================================================
# Constructor: new
#
# Returns:
# New Vault object
#
# Example:
#> my $Vault = San::FIT::Misfit::Features::Vault->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Features::Vault";
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    # Initialize self with the testbed's cluster information.
    $self->_discover_clusters();

    $Log->exit();
    return $self;
} # End of new

#===============================================================================
#  PUBLIC METHODS
#===============================================================================
#==============================================================================
# Function: create
# Create n vaults.  Create will loop through the list of objects provided
# and create each vault with the specified parameters.  Vaulted snapmirrors
# are specifically created with a type XDP.
#
# Parameters:
#   objects - Array of hashes containing information for creating each
#       vault
#   background - Option to specify whether to perform each create
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the creates
#
# Vault create attributes:
#   'destination-path' - Path to vault destination (required)
#   'source-path' - Path to vault source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $Vault->create(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub create {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vaults = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vault_info = ();
    my @msg_data = ();
    foreach my $vault (@{$vaults}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $vault->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $vault->{'destination-path'}
        });
        if (defined ($vault->{id})) {
            $runid = $vault->{id} . "_vault_create";
        } else {
            $runid = $path{volume} . "_vault_create";
        }
        push (@vault_info, {
            codespec => "{" . $self->{id} . "::_create(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vault},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vault_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Vault create failed");
        return FALSE;
    }
} # End of create

#==============================================================================
# Function: initialize
# Initialize n vaults.  Initialize will loop through the list of objects provided
# and initialize each vault with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for initializing each
#       vault
#   background - Option to specify whether to perform each initialize
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the initializes
#
# Vault initialize attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $Vault->initialize(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub initialize {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vaults = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vault_info = ();
    my @msg_data = ();
    foreach my $vault (@{$vaults}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $vault->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $vault->{'destination-path'}
        });
        if (defined ($vault->{id})) {
            $runid = $vault->{id} . "_vault_initialize";
        } else {
            $runid = $path{volume} . "_vault_initialize";
        }
        push (@vault_info, {
            codespec => "{" . $self->{id} . "::_initialize(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vault},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vault_info,
        operation  => "initialize",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Vault initialize failed");
        return FALSE;
    }
} # End of initialize

#==============================================================================
# Function: update
# Update n vaults.  Update will loop through the list of objects provided
# and update each vault with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for updating each
#       vault
#   background - Option to specify whether to perform each update
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the updates
#
# Vault update attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $Vault->update(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub update {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vaults = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vault_info = ();
    my @msg_data = ();
    foreach my $vault (@{$vaults}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $vault->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $vault->{'destination-path'}
        });
        if (defined ($vault->{id})) {
            $runid = $vault->{id} . "_vault_udate";
        } else {
            $runid = $path{volume} . "_vault_update";
        }
        push (@vault_info, {
            codespec => "{" . $self->{id} . "::_update(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vault},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vault_info,
        operation  => "update",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Vault update failed");
        return FALSE;
    }
} # End of update

#==============================================================================
# Function: delete
# Create n vaults.  Delete will loop through the list of objects provided
# and delete each vault with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for deleting each
#       vault
#   background - Option to specify whether to perform each delete
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the deletes
#
# Vault delete attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $Vault->delete(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub delete {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vaults = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vault_info = ();
    my @msg_data = ();
    foreach my $vault (@{$vaults}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $vault->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $vault->{'destination-path'}
        });
        if (defined ($vault->{id})) {
            $runid = $vault->{id} . "_vault_delete";
        } else {
            $runid = $path{volume} . "_vault_delete";
        }
        push (@vault_info, {
            codespec => "{" . $self->{id} . "::_delete(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vault},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vault_info,
        operation  => "delete",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Vault delete failed");
        return FALSE;
    }
} # End of delete

#==============================================================================
# Function: release
# Release n vaults.  Release will loop through the list of objects provided
# and release each vault with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for releasing each
#       vault
#   background - Option to specify whether to perform each release
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the releases
#
# Vault release attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $Vault->release(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub release {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vaults = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vault_info = ();
    my @msg_data = ();
    foreach my $vault (@{$vaults}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $vault->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $vault->{'destination-path'}
        });
        if (defined ($vault->{id})) {
            $runid = $vault->{id} . "_vault_release";
        } else {
            $runid = $path{volume} . "_vault_release";
        }
        push (@vault_info, {
            codespec => "{" . $self->{id} . "::_release(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vault},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_home}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vault_info,
        operation  => "release",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Vault release failed");
        return FALSE;
    }
} # End of release

#==============================================================================
# Function: promote
# Promote n vaults.  Promote will loop through the list of objects
# provided and promote each vault with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for promoting each
#       vault
#   background - Option to specify whether to perform each promote
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the promotes
#
# Vault promote attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $Vault->promote(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub promote {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vaults = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($vaults)) {
        $vaults = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @vault_info = ();
    my @msg_data = ();
    foreach my $vault (@{$vaults}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $vault->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $vault->{'destination-path'}
        });
        if (defined ($vault->{id})) {
            $runid = $vault->{id} . "_vault_promote";
        } else {
            $runid = $path{volume} . "_vault_promote";
        }
        push (@vault_info, {
            codespec => "{" . $self->{id} . "::_promote(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vault},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_home}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vault_info,
        operation  => "promote",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Vault promote failed");
        return FALSE;
    }
} # End of promote

#==============================================================================
# Function: resync
# Resync n vaults.  Resync will loop through the list of objects provided
# and resync each vault with the specified parameters.
#
# Parameters:
#   objects - Array of hashes containing information for resyncing each
#       vault
#   background - Option to specify whether to perform each resync
#       serially (FALSE) or paralllely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   iterations - Mumber of times to perform the resyncs
#
# Vault resync attributes:
#   'destination-path' - Path to snapmirror destination (required)
#   'source-path' - Path to snapmirror source (required)
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   TRUE if successful, FALSE if unsuccessful
#
#
# Example:
# >my $status = $Vault->resync(
# >    objects    => [{
# >        'destination-path' => 'ssan-cluster1://sanfit_vs/sanfit_vol',
# >        'source-path'      => 'ssan-cluster2://sanfit_vs/sanfit_vol',
# >    ]},
# >    background => TRUE,
# >);
#
#==============================================================================
sub resync {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vaults = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($vaults)) {
        $vaults = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @vault_info = ();
    my @msg_data = ();
    foreach my $vault (@{$vaults}) {
        my $runid = '';
        my %path = San::FIT::Misfit::Core::Util::parse_path(
            $vault->{'source-path'}
        );
        push (@msg_data, {
            cluster            => $path{cluster},
            vserver            => $path{vserver},
            volume             => $path{volume},
            'destination-path' => $vault->{'destination-path'}
        });
        if (defined ($vault->{id})) {
            $runid = $vault->{id} . "_vault_resync";
        } else {
            $runid = $path{volume} . "_vault_resync";
        }
        push (@vault_info, {
            codespec => "{" . $self->{id} . "::_resync(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vault},
                %path,
                vhost => $self->{vhost},
                id    => $self->{id},
                node  => Clone::clone($self->{cluster_peer}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vault_info,
        operation  => "resync",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Vault resync failed");
        return FALSE;
    }
} # End of resync

#==============================================================================
# Function: build_vault_relationships
# Given a list of source volume hashes and a list of destination volume hashes,
# build_vault_relationships parses the volume objects and creates source and
# destination paths for use in the other functions of the Misfit Vault module.
#
# Parameters:
#   source_volumes - array of hashes containing information
#       for each source volume.
#   destination_volumes - array of hashes containing
#       information for each destination volume.
#
# Syntax for SN Paths:
#   cluster://vserver/volume
#
# Returns:
#   Array of Vault relationships containing paths for source and destination
#
#
# Example:
# >my @vaults = $Vault->build_vault_relationships(
# >    source_volumes      => \@Source_Volumes
# >    destination_volumes => \@Destination_Volumes
# >);
#
#==============================================================================
sub build_vault_relationships {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $src_volumes = &ParamsMgr::get_arg(
        'arg'      => "source_volumes",
        'required' => FALSE,
        'default'  => (),
    );
    my $dst_volumes = &ParamsMgr::get_arg(
        'arg'      => "destination_volumes",
        'required' => FALSE,
        'default'  => (),
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vault_relationships = ();
    my @dst_array = @{$dst_volumes};

    my $index = 0;
    foreach my $src (@{$src_volumes}) {
        my $dst = $dst_array[$index];
        if (defined($dst)) {
            my $source_path = $self->{cluster_home_name} . '://' .
                $src->{vserver} . '/' . $src->{volume};
            my $dest_path = $self->{cluster_peer_name}  . '://' . $dst->{vserver}
                . '/' . $dst->{volume};
            push(@vault_relationships, {
                'destination-path' => $dest_path,
                'source-path'      => $source_path,
                type               => 'XDP',
            });
        }
        $index++;
    }

    return wantarray ? @vault_relationships : \@vault_relationships;
}

#===============================================================================
#  PRIVATE METHODS
#===============================================================================
#==============================================================================
# Description: Subprocess responsible for creating vaults
#==============================================================================
sub _create {
    my ($class) = @_;

    my $params     = NATE::ParamSet->get_global();
    my $source     = $params->get('source-path');
    my $dest       = $params->get('destination-path');
    my $node       = $params->get('node');
    my $type       = 'XDP';
    my $vserver    = $params->get('vserver');
    my $volume     = $params->get('volume');
    my $if_exists  = $params->get('if_exists', default => 'reuse');
    my $status     = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.create.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NACL::STask::Snapmirror->create(
            command_interface   => $node,
            do_initialize       => FALSE,
            'source-path'       => $source,
            'destination-path'  => $dest,
            'type'              => $type,
            if_exists           => $if_exists
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.create.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error creating Vault');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'vault.create.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $Log->comment($exception->text());
    };

    if ($status) {
        return NATE::Result::Pass->new('Vault create successful');
    } else {
        return NATE::Result::Fail->new('Vault create failed');
    }
} # End of _create

#==============================================================================
# Description: Subprocess responsible for initializing Vaults
#==============================================================================
sub _initialize {
    my ($class) = @_;

    my $params     = NATE::ParamSet->get_global();
    my $source     = $params->get('source-path');
    my $dest       = $params->get('destination-path');
    my $node       = $params->get('node');
    my $vserver    = $params->get('vserver');
    my $volume     = $params->get('volume');
    my $if_exists  = $params->get('if_exists', default => 'reuse');
    my $status     = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.initialize.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NACL::STask::Snapmirror->initialize(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest,
            if_exists          => $if_exists
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.initialize.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error initializing Vault');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
            msg  => 'vault.initialize.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Vault initialize successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Vault initialize failed'));
    }
} # End of _initialize

#==============================================================================
# Description: Subprocess responsible for updating Vaults
#==============================================================================
sub _update {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.update.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NACL::STask::Snapmirror->update(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.update.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error updating Vault');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
            msg  => 'vault.update.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Vault update successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Vault update failed'));
    }
} # End of _update

#==============================================================================
# Description: Subprocess responsible for deleting Vaults
#==============================================================================
sub _delete {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.delete.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        my $job;
        NACL::C::Snapmirror->delete(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest,
            job_component      => \$job
        );
        if (defined ($job)) {
            $job->wait_on_job('method-timeout' => TIMEOUT);
        }
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.delete.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error deleting Vault');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'vault.delete.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Vault delete successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Vault delete failed'));
    }
} # End of _delete

#==============================================================================
# Description: Subprocess responsible for releasing Vaults
#==============================================================================
sub _release {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.release.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        my $job;
        NACL::C::Snapmirror->release(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest,
            job_component      => \$job
        );
        if (defined ($job)) {
            $job->watch_progress('method-timeout' => TIMEOUT);
        }
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.release.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error releasing Vault');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'vault.release.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Vault release successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Vault release failed'));
    }
} # End of _release

#==============================================================================
# Description: Subprocess responsible for promoting Vaults
#==============================================================================
sub _promote {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.promote.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );

        my $job;
        NACL::C::Snapmirror->promote(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest,
            job_component      => \$job
        );
        if (defined ($job)) {
            $job->watch_progress('method-timeout' => TIMEOUT);
        }
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.promote.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error promoting Vault');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'vault.promote.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw();
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Vault promote successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Vault promote failed'));
    }
} # End of _promote

#==============================================================================
# Description: Resync Vault
#==============================================================================
sub _resync {
    my ($class) = @_;

    my $params  = NATE::ParamSet->get_global();
    my $source  = $params->get('source-path');
    my $dest    = $params->get('destination-path');
    my $node    = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume  = $params->get('volume');
    my $status  = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.resync.started',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NACL::STask::Snapmirror->resync(
            command_interface  => $node,
            'source-path'      => $source,
            'destination-path' => $dest
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::Vault',
            routing_key => '#',
            message     => {
                msg  => 'vault.resync.finished',
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Error resyncing Vault');
        $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                msg  => 'vault.resync.' . $exception->text(),
                data => {
                    node                => $node->{node},
                    vserver             => $vserver,
                    volume              => $volume,
                    'destination-path'  => $dest,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        $Log->log(NATE::Result::Pass->new('Vault resync successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Vault resync failed'));
    }
} # End of _resync

#==============================================================================
# Description: Discover and get cluster information for Vault tests.
#==============================================================================
sub _discover_clusters {
    my ($self, %args) = @_;
    $Log->enter();


    my $cluster_name = param("CLUSTER_NAME", "-default", "");
    my $peer_name    = param("PEER_CLUSTER", "-default", "");

    if ($peer_name eq "") {
        $peer_name = $cluster_name;
    }

    $self->{cluster_home_name} = $cluster_name;
    $self->{cluster_home}      = NACL::C::Cluster->new(name => $cluster_name);
    $self->{cluster_home}->apiset_must(interface => 'ZAPI');
    $self->{cluster_home}->apiset();
    $self->{cluster_home_cli} = NACL::C::Cluster->new(name => $cluster_name);
    $self->{cluster_home_cli}->apiset_must(interface => 'CLI');
    $self->{cluster_home_cli}->apiset();

    $self->{cluster_peer_name} = $peer_name;
    $self->{cluster_peer}      = NACL::C::Cluster->new(name => $peer_name);
    $self->{cluster_peer}->apiset_must(interface => 'ZAPI');
    $self->{cluster_peer}->apiset();
    $self->{cluster_peer_cli} = NACL::C::Cluster->new(name => $peer_name);
    $self->{cluster_peer_cli}->apiset_must(interface => 'CLI');
    $self->{cluster_peer_cli}->apiset();

    $Log->exit();
}

#==============================================================================
# Description: Create and accept a vserver peer relationship for SN clusters
#==============================================================================
sub _create_vserver_peer {
    my ($self, %args) = @_;
    $Log->enter();

    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $peer_vserver = &ParamsMgr::get_arg(
        'arg'      => "peer_vserver",
        'required' => TRUE,
    );
    my $peer_cluster = &ParamsMgr::get_arg(
        'arg'      => "peer_cluster",
        'required' => FALSE,
        'default'  => $self->{cluster_peer_name}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $status = TRUE;
    try {
        NACL::STask::VserverPeer->create(
            command_interface  => $self->{cluster_home},
            vserver            => $vserver,
            'peer-vserver'     => $peer_vserver,
            'peer-cluster'     => $peer_cluster,
            applications       => ['snapmirror'],
            nacltask_if_exists => 'reuse',
        );
    } otherwise {
        my $exception = shift;
        $Log->comment($exception->text());
        $status = FALSE;
    };

    try {
        NACL::STask::VserverPeer->accept(
            command_interface  => $self->{cluster_peer},
            'peer-vserver'     => $vserver,
            vserver            => $peer_vserver,
        );
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /Invalid peering state "peered" on the local cluster./) {
            $Log->comment("Vservers already peered... Continuing...");
            $status = TRUE;
        } elsif ($exception->text() =~ /Invalid operation on intra cluster Vserver peering/) {
            $Log->comment(
                'Cannot accept peer on Intracluster peering, continuing...'
            );
        } elsif ($exception->text() =~ /Invalid operation on intra cluster Vserver peer relationship./i) {
            $Log->comment(
                'Cannot accept peer on Intracluster peering, continuing...'
            );
            $status = TRUE;
        } else {
            $Log->comment($exception->text());
            $status = FALSE;
        }
    };

    return $status;
}

#==============================================================================
# Description: Delete a vserver peer relationship for SN clusters
#==============================================================================
sub _delete_vserver_peer {
    my ($self, %args) = @_;
    $Log->enter();

    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $peer_vserver = &ParamsMgr::get_arg(
        'arg'      => "peer_vserver",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $status = TRUE;

    try {
        NACL::STask::VserverPeer->purge(
            command_interface  => $self->{cluster_peer},
            vserver            => $peer_vserver,
            'peer-vserver'     => $vserver,
            nacltask_if_purged => 'pass',
        );
    } otherwise {
        $status = FALSE;
    };

    try {
        NACL::STask::VserverPeer->purge(
            command_interface  => $self->{cluster_home},
            vserver            => $vserver,
            'peer-vserver'     => $peer_vserver,
            nacltask_if_purged => 'pass',
        );
    } otherwise {
        $status = FALSE;
    };

    if ($status) {
        return TRUE;
    } else {
        return FALSE;
    }

} # End of _delete_vserver_peer

#===============================================================================
# Message: vault.create.started
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.create.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.create.finished
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.create.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.create.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "vault.create.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.initialize.started
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.initialize.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.initialize.finished
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.initialize.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.initialize.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "vault.initialize.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.update.started
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.update.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.update.finished
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.update.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.update.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "vault.update.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.delete.started
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.delete.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.delete.finished
#   Exchange:
#       San::FIT::Misfit::Features::Snapmirror
#
#   Message Structure:
# >     {
# >         "msg": "snapmirror.delete.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.delete.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "vault.delete.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.release.started
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.release.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.release.finished
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.release.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.release.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "vault.release.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.promote.started
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.promote.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.promote.finished
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.promote.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.promote.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "vault.promote.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.resync.started
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.resync.started",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.resync.finished
#   Exchange:
#       San::FIT::Misfit::Features::Vault
#
#   Message Structure:
# >     {
# >         "msg": "vault.resync.finished",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#
# Message: vault.resync.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "vault.resync.Error Text",
# >         "data" : {
# >             "destination-path": "sanfit_cluster1://vserver/volume",
# >             "source-path": "sanfit_cluster2://vserver/volume",
# >          }
# >     }
#===============================================================================

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/Misfit/features/Vault>

=cut

1;
