# Copyright (c) 2012-2020 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Volume Operations module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Features::VolOps
# Wrapper for NACL volume operations with Misfit messaging

package San::FIT::Misfit::Features::VolOps;

# Compiler Directives
use strict;
use warnings;

# Import Libraries
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::Volume;
use NACL::C::Cluster;
use NACL::STask::Volume;
use NACL::C::VolumeMove;
use NACL::CS::VolumeMove;
use NACL::STask::VolumeCopy;
use NACL::STask::VolumeMove;
use NACL::STask::VolumeClone;
use San::FIT::Misfit::Core::Workflow;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);

# This package inherits methods and attributes from Misfit's Base class
use base qw( San::FIT::Misfit::Core::Base );

use constant {
    TIMEOUT      => 600,
    DELAY_JOB    => 60,
    MOVE_TIMEOUT => 1800,
};

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Constructor: new
#   Creates a Misfit Volume Object
#
#   Parameters:
#       cluster - name of cluster to perform operation on.
#
#   Returns:
#     New VolOps object
#
#   Example:
#     > my $VolOps = San::FIT::Misfit::Features::VolOps->new();
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Features::VolOps";

    &ParamsMgr::process_args(%args);
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );

    my $node = undef;
    my $node_cli = undef;
    if (defined ($cluster_name)) {
        $node = NACL::C::Cluster->new(name => $cluster_name);
        $node_cli = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $node = NACL::C::Node->find();
        $node_cli = NACL::C::Node->find();
    }

    $node->apiset_must(interface => 'ZAPI');
    $node->apiset();
    $self->{node} = $node;
    $self->{node_cli} = $node_cli;
    $self->{node_cli}->apiset_must(interface => 'CLI');
    $self->{node_cli}->apiset();
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#==============================================================================
# Function: create
#   Create n volumes. Create will loop through the list of objects provided
#   and create each volume with the specified parameters. Provides a wrapper
#   to NACL::STask::Volume->create() function
#
#   Parameters:
#     objects - Array of hashes containing information for creating each
#         volume.
#     background - Option to specify whether to perform each create
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to create
#
#     Volume create attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#       'type' - Type of volume (optional. Defaults to RW)
#       'aggregate' - Name of the aggregate (required)
#       'size' - Size of the volume
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >  my $status = $VolOps->create(
#     >      objects    => [{
#     >          'volume'    => 'my_volume',
#     >          'vserver'   => 'vs1',
#     >          'type'      => 'DP',
#     >          'aggregate' => 'aggr_1'
#     >          'size'      => '500M'
#     >      ]},
#     >      background => TRUE,
#     > );
#==============================================================================
sub create {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = $volume->{id};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_create(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Volume Create failed");
        return FALSE;
    }
} # End of create

#==============================================================================
# Function: clone
#   Clone n volumes. Clone will loop through the list of objects provided
#   and create each volume with the specified parameters. Provides a wrapper
#   to NACL::C::VolumeClone->create function
#
#   Parameters:
#     objects - Array of hashes containing information for cloning each
#         volume.
#     background - Option to specify whether to perform each clone
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to create
#
#     Volume clone attributes:
#       'vserver' - Name of the vserver (required)
#       'volume' - Name of the volume to clone (required)
#       'flexclone' - Name of the cloned volume (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >my $status = $VolOps->clone(
#     >    objects    => [{
#     >        'volume'   => 'my_volume',
#     >        'vserver'  => 'vs1',
#     >        'flexvol'  => 'my_volume_clone'
#     >    ]},
#     >    background => TRUE,
#     >);
#==============================================================================
sub clone {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = $volume->{id};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_clone(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id        => $self->{id},
                node      => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "clone",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Volume clone failed");
        return FALSE;
    }
} # End of clone

#==============================================================================
# Function: split
#   Split 'n' volumes clones. Split will loop through the list of objects provided
#   and split each volume with the specified parameters. Provides a wrapper
#   to NACL::STask::VolumeClone->split() function
#
#   Parameters:
#     objects - Array of hashes containing information for splitting each
#         volume.
#     background - Option to specify whether to perform each split
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to create
#
#     Volume split attributes:
#       'vserver' - Name of the vserver (required)
#       'flexclone' - Name of the cloned volume (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     >my $status = $VolOps->split(
#     >    objects    => [{
#     >        'volume'   => 'my_volume',
#     >        'vserver'  => 'vs1',
#     >        'flexvol'  => 'my_volume_clone'
#     >    ]},
#     >    background => TRUE,
#     >);
#==============================================================================
sub split {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = $volume->{id};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_clone_split(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "split",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Volume split failed");
        return FALSE;
    }
} # End of clone split

#==============================================================================
# Function: delete
#   Delete n volumes. Create will loop through the list of objects provided
#   and delete each volume with the specified parameters. Provides a wrapper
#   to NACL::STask::Volume->delete() function
#
# Parameters:
#   objects - Array of hashes containing information for deleting each
#       volume
#   background - Option to specify whether to perform each delete
#       serially (FALSE) or parallely (TRUE, default)
#   node - NACL node object for the node to perform the operation on
#   count - The number of volumes to delete
#
#   Volume delete attributes:
#     'volume' - Volume name (required)
#     'vserver' - Vserver name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $VolOps->delete(
#     >     objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub delete {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = $volume->{id};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_delete(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "delete",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Volume delete failed");
        return FALSE;
    }
} # End of delete

#==============================================================================
# Function: copy
#   Copy n volumes. Copy will loop through the list of objects provided
#   and copy each volume with the specified parameters. Provides a wrapper
#   to NACL::STask::VolumeCopy->start() function
#
#   Parameters:
#     objects - Array of hashes containing information for copying each
#         volume
#     background - Option to specify whether to perform each create
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to create
#
#     Volume delete attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#       'destination-volume' - Destination Volume name (required)
#       'destination-aggregate' - Destination Aggregate name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $VolOps->copy(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >        'destination-volume' => 'my_volume_copy',
#     >        'destination-aggregate' => 'aggr_3',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub copy {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = $volume->{id};
        } else {
            $runid = $volume->{volume};
        }
        push(@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_copy(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "copy",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Volume copy failed");
        return FALSE;
    }
} # End of copy

#==============================================================================
# Function: move
#   Move 'n' volumes. Move will loop through the list of objects provided
#   and move each volume with the specified parameters. Provides a wrapper
#   to NACL::STask::VolumeMove->start() function
#
#   Parameters:
#     objects - Array of hashes containing information for moving each
#         volume.
#     background - Option to specify whether to perform each move
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     iterations - Number of times to perform the moves
#     count - The number of volumes to create
#
#     Volume create attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#       'destination-aggregate' - Name of the aggregate where the vol will move (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $VolOps->move(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >        'destination-aggregate' => 'aggr_3',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub move {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = $volume->{id};
        } else {
            $runid = $volume->{volume};
        }
        push (@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_move(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "move",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Volume move failed");
        return FALSE;
    }
} # End of move

#==============================================================================
# Function: online
#   Online 'n' volumes. Online will loop through the list of objects provided
#   and online each volume with the specified parameters. Provides a wrapper
#   to NACL::C::Volume->online function
#
#   Parameters:
#     objects - Array of hashes containing information for onlining each
#         volume
#     background - Option to specify whether to perform each create
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to create
#
#     Volume online attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $VolOps->online(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub online {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = $volume->{id};
        } else {
            $runid = $volume->{volume};
        }
        push (@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_online(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "online",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Volume online failed");
        return FALSE;
    }
} # End of online

#==============================================================================
# Function: offline
#   Offline 'n' volumes. Offline will loop through the list of objects provided
#   and offline each volume with the specified parameter. Provides a wrapper
#   to NACL::C::Volume->offline function
#
#   Parameters:
#     objects - Array of hashes containing information for offlining each
#         volume
#     background - Option to specify whether to perform each create
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     count - The number of volumes to create
#
#   Volume offline attributes:
#     'volume' - Volume name (required)
#     'vserver' - Vserver name (required)
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $VolOps->offline(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub offline {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = $volume->{id};
        } else {
            $runid = $volume->{volume};
        }
        push (@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_offline(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "offline",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Volume Offline failed");
        return FALSE;
    }
} # End of offline

#==============================================================================
# Function: resize
#   Resize 'n' volumes. Resize will loop through the list of objects provided
#   and resize each volume with the specified parameter. Provides a wrapper
#   to NACL::C::Volume->size function
#
#   Parameters:
#     objects - Array of hashes containing information for resizing each
#         volume
#     background - Option to specify whether to perform each resize
#         serially (FALSE) or parallely (TRUE, default)
#     node - NACL node object for the node to perform the operation on
#     iterations - Number of times to perform the resize
#     count - The number of volumes to create
#
#     Volume resize attributes:
#       'volume' - Volume name (required)
#       'vserver' - Vserver name (required)
#       'new-size' - New volume size (optional)
#       'percent' - Increase the volume size by 'n' percent
#
#   Returns:
#     TRUE if successful, FALSE if unsuccessful
#
#
#   Example:
#     > my $status = $VolOps->resize(
#     >    objects    => [{
#     >        'volume'    => 'my_volume',
#     >        'vserver'   => 'vs1',
#     >        'new-size'  => '1G',
#     >    ]},
#     >    background => TRUE,
#     > );
#==============================================================================
sub resize {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $volumes = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    unless (defined($volumes)) {
        $volumes = $self->_get_volume_list(
            count  => $count,
            filter => $filter,
            params => $params
        );
    }

    my @volume_info = ();
    my @msg_data = ();
    foreach my $volume (@{$volumes}) {
        my $runid = '';
        if (defined ($volume->{id})) {
            $runid = $volume->{id};
        } else {
            $runid = $volume->{volume};
        }
        my $percent = 0;
        if (defined ($volume->{percent})) {
            $percent = $volume->{percent};
        }
        push (@msg_data, {vserver => $volume->{vserver}, volume => $volume->{volume}});
        push (@volume_info, {
            codespec => "{" . $self->{id} . "::_resize(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$volume},
                percent => $percent,
                id      => $self->{id},
                node    => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@volume_info,
        operation  => "resize",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Volume resize failed");
        return FALSE;
    }
} # End of resize

#==============================================================================
# PRIVATE METHODS
#==============================================================================
#===============================================================================
# Description: Responsible by creating a data structure used to determine
# volume move pattern
#===============================================================================
sub _build_node_tree {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $vols_ref = &ParamsMgr::get_arg(
        'arg'      => "vols",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return ();
    }

    my @vols = @{$vols_ref};
    my @nodes = NACL::C::Cluster->nodes(command_interface => $self->{node_cli});
    my @return = ();

    require San::FIT::Misfit::Services::Resource;
    my $resource = San::FIT::Misfit::Services::Resource->new();

    # Loop through of each of the nodes provided for the cluster
    foreach my $node (@nodes) {
        # Create temporary hash to push into node_tree
        my %resource_info = ();
        $resource_info{node} = $node;
        $resource_info{objects} = [];
        $resource_info{commands} = [];
        $resource_info{volumes} = [];
        $resource_info{path} = [];
        $resource_info{aggregates} = [];
        $resource_info{total_size} = 0;
        # Get aggregates per node
        my @aggrs = $resource->get_objects(
            type        => 'aggregate',
            search_node => $node,
            filter      => {
                nodes       => [$node],
                'ha-policy' => 'sfo'
            },
            requested_fields => [
                'size','owner-name','aggregate', 'availsize'
            ],
        );
        next if (scalar(@aggrs) <= 0);
        # Loop through volumes provided to function and add them to array under
        # node. Also will set a total_size for the volume group on the node.
        foreach my $vol (@vols) {
            next if (not($vol->{node} eq $node));
            $resource_info{total_size} = $resource_info{total_size} +
                $vol->{size};
            $vol->{orig_aggr} = $vol->{aggregate};
            push(@{$resource_info{volumes}}, $vol);
        }
        push(@{$resource_info{aggregates}}, @aggrs);
        push(@return, \%resource_info);
    }

    my $index = 0;
    # In this loop, a path is decided for the movement of volumes across
    # the cluster.  This is necessary for certain volop testcases.  Also will
    # create array for the move commands to be used in an execute call.
    foreach my $node (@return) {
        my $j = $index + 1;
        if ($j >= scalar(@return)) {
                $j = 0;
        }
        while ($j != $index) {
            foreach my $aggr (@{$return[$j]->{aggregates}}) {
                if ($aggr->{availsize} >= $node->{total_size}) {
                    push(@{$node->{path}}, $aggr->{aggregate});
                    last;
                }
            }
            $j++;
            if ($j >= scalar(@return)){
                $j = 0;
            }
        }
        $index++;
        $j = 0;
        foreach my $aggr (@{$node->{path}}) {
            if ($j-1 < 0) {
                push(@{$node->{commands}}, {
                    command => 'move',
                    params  => {
                        'destination-aggregate' => $node->{path}[$j],
                    }
                });
            } else {
                push(@{$node->{commands}}, {
                    command => 'move',
                    params  => {
                        'destination-aggregate' => $node->{path}[$j],
                    }
                });
            }
            $j++;
        }
        # Creates objects list for execution
        foreach my $vol (@{$node->{volumes}}) {
            push(@{$node->{objects}}, {
                volume                  => $vol->{volume},
                vserver                 => $vol->{vserver},
                'destination-aggregate' => $vol->{orig_aggr}
            });
        }
    }

    return @return;
}

#===============================================================================
# Description: Responsible for generating a list of volumes matching a criteria
#===============================================================================
sub _get_volume_list {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $operation = &ParamsMgr::get_arg(
        'arg'      => "operation",
        'required' => FALSE,
        'default'  => '',
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @volumes = ();
    my @all_volumes = NACL::C::Volume->find(
        command_interface => $self->{node},
        filter => $filter
    );

    if ($count =~ /^all$/i) {
        $count = scalar(@all_volumes);
    }

    for (my $i = 0; $i < $count; $i++) {
        my $vserver = $all_volumes[$i]->get_vserver();
        $volumes[$i] =  {
            vserver  => $vserver->name(),
            volume   => $all_volumes[$i]->volume(),
            %{$params}
        };
    }

    return wantarray ? @volumes : \@volumes;
} # End of _get_volume_list

#==============================================================================
# Description: Subprocess responsible for creating volumes
#==============================================================================
sub _create {
    my ($class) = @_;

    my $params    = NATE::ParamSet->get_global();
    my $node      = $params->get('node');
    my $vserver   = $params->get('vserver');
    my $volume    = $params->get('volume');
    my $agg       = $params->get('aggregate');
    my $type      = $params->get('type');
    my $if_exists = $params->get('if_exists', default => 'purge');
    my $autosize  = $params->get('autosize', default => 'true');
    my $status  = FALSE;

    my %attributes = (
        nacltask_if_exists => $if_exists,
        volume             => $volume,
        vserver            => $vserver,
        aggregate          => $agg,
        autosize           => $autosize,
    );
    if (defined($type)) {
        $attributes{type} = $type;
    }

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.create.started',
                data => {
                    node      => $node->{node},
                    %attributes
                },
            }
        );
        NACL::STask::Volume->create(
            command_interface => $node,
            %attributes
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed to create volumes');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.create.' . $exception->text(),
                data => {
                    node => $node->{node},
                    %attributes
                },
            }
        );
        $Log->log(NATE::Result::Fail->new($exception->text()));
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.create.finished',
                data => {
                    node => $node->{node},
                    %attributes
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Volume create successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Volume create failed'));
    }
} # End of _create

#==============================================================================
# Description: Subprocess responsible for cloning volumes
#==============================================================================
sub _clone {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume = $params->get('volume');
    my $flexclone = $params->get('flexclone');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.clone.started',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    volume    => $volume,
                    flexclone => $flexclone,
                },
            }
        );
        NACL::STask::VolumeClone->create(
            command_interface  => $node,
            'parent-volume'    => $volume,
            vserver            => $vserver,
            flexclone          => $flexclone,
            nacltask_wait      => TRUE,
            nacltask_if_exists => "purge",
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.clone.' . $exception->text(),
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    volume    => $volume,
                    flexclone => $flexclone,
                },
            }
        );
        $Log->log(NATE::Result::Fail->new($exception->text()));
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.clone.finished',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    volume    => $volume,
                    flexclone => $flexclone,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Volume clone successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Volume clone failed'));
    }
} # End of _volume_clone

#==============================================================================
# Description: Subprocess responsible for doing a clone split on a volume
#==============================================================================
sub _clone_split {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $flexclone = $params->get('flexclone');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.clonesplit.started',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    flexclone => $flexclone,
                },
            }
        );
        NACL::STask::VolumeClone->split(
            command_interface => $node,
            vserver           => $vserver,
            flexclone         => $flexclone,
            nacltask_wait     => TRUE,
            'method-timeout'  => 800,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed in volume clone split operation. Volume: ' .
            $flexclone . ' Vserver: ' . $vserver
        );
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.clonesplit.' . $exception->text(),
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    flexclone => $flexclone,
                },
            }
        );
        $Log->log(NATE::Result::Fail->new($exception->text()));
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.clonesplit.finished',
                data => {
                    node      => $node->{node},
                    vserver   => $vserver,
                    flexclone => $flexclone,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Volume clone split successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Volume clone split failed'));
    }
} # End of _clone_split

#==============================================================================
# Description: Subprocess responsible for deleting volumes
#==============================================================================
sub _delete {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume = $params->get('volume');
    my $dest_agg = $params->get('destination-aggregate');
    my $dest_vol = $params->get('destination-volume');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.delete.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );
        NACL::STask::Volume->purge(
            command_interface  => $node,
            nacltask_if_purged => 'pass',
            volume             => $volume,
            vserver            => $vserver,
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Failed in volume delete operation. Volume:' .
            $volume . ' Vserver: ' . $vserver
        );
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.delete.' . $exception->text(),
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );
        $Log->log(NATE::Result::Fail->new($exception->text()));
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.delete.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Volume delete successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Volume delete failed'));
    }
} # End of _delete

#==============================================================================
# Description: Subprocess responsible for copying volumes
#==============================================================================
sub _copy {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume = $params->get('volume');
    my $dest_agg = $params->get('destination-aggregate');
    my $dest_vol = $params->get('destination-volume');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.copy.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );
        NACL::STask::VolumeCopy->start(
            command_interface       => $node,
            volume                  => $volume,
            vserver                 => $vserver,
            'destination-volume'    => $dest_vol,
            'destination-aggregate' => $dest_agg,
            nacltask_wait           => TRUE,
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /existing destination volume/i) {
            $Log->comment(
                "$volume exists on the destination aggregate, skipping."
            );
            $status = TRUE;
        } else {
            $Log->comment('Failed in volume copy operation. Volume: '
                . $volume . ' Vserver: ' . $vserver
            );
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg  => 'VolOps.copy.' . $exception->text(),
                    data => {
                        node        => $node->{node},
                        vserver     => $vserver,
                        volume      => $volume,
                        dest_volume => $dest_vol,
                        dest_agg    => $dest_agg,
                    },
                }
            );
            $Log->log(NATE::Result::Fail->new($exception->text()));
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.copy.finished',
                data => {
                    node        => $node->{node},
                    vserver     => $vserver,
                    volume      => $volume,
                    dest_volume => $dest_vol,
                    dest_agg    => $dest_agg,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Volume copy successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Volume copy failed'));
    }
} # End of _copy

#==============================================================================
# Description: Subprocess responsible for moving volumes
#==============================================================================
sub _move {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node   = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume = $params->get('volume');
    my $dest = $params->get('destination-aggregate');
    my $timeout = $params->get('method-timeout', default => MOVE_TIMEOUT);
    my $status = FALSE;
    my $trigger_cutover = FALSE;

    $MessageBroker->publish_message(
        exchange    => 'San::FIT::Misfit::Features::VolOps',
        routing_key => '#',
        message     => {
            msg  => 'VolOps.move.started',
            data => {
                node    => $node->{node},
                vserver => $vserver,
                volume  => $volume,
                'destination-aggregate' => $dest,
            },
        }
    );
    try {
        NACL::STask::VolumeMove->start(
            command_interface       => $node,
            volume                  => $volume,
            vserver                 => $vserver,
            'destination-aggregate' => $dest,
            'method-timeout'        => $timeout,
            nacltask_wait           => TRUE,
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        if ($exception->text() =~ /cannot be moved because it is locked by job/i){
            $Log->comment("$volume currently locked for move.  Waiting to continue...");
            sleep(DELAY_JOB);
            $trigger_cutover = TRUE;
        } elsif ($exception->text() =~ /volume does not exist for the operation/){
            $status = TRUE;
        } else {
            $Log->comment($exception->text());
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg  => 'VolOps.move.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume,
                        'destination-aggregate' => $dest,
                    },
                }
            );
            $Log->log(NATE::Result::Fail->new($exception->text()));
        }
    };

    if ($trigger_cutover) {
        $Log->comment('Attempting to manually cutover volume move... ' .
            'checking for errors.'
        );
        my @errors = ();
        try {
            @errors = NACL::CS::VolumeMove->fetch(
                command_interface => $node,
                filter            => {
                    vserver => $vserver,
                    volume  => $volume,
                },
                requested_fields => ['prior-issues'],
            );
        } otherwise {
            $Log->comment('No prior issues with vol move found, triggering' .
                ' cutover'
            );
        };

        if (scalar(@errors)) {
            my $output = "\n" . "_" x 100 .
                "\nIssues moving $vserver:$volume:\n" .
                "_" x 100;

            foreach my $error (@errors) {
                $output = $output . $error->{prior_issues} . "\n" . "_" x 100 .
                    "\n";
            }
            $Log->comment($output);
        }

        try {
            $MessageBroker->publish_message(
                exchange    => 'San::FIT::Misfit::Features::VolOps',
                routing_key => '#',
                message     => {
                    msg  => 'VolOps.cutover.started',
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume,
                        'destination-aggregate' => $dest,
                    },
                }
            );
            NACL::C::VolumeMove->trigger_cutover(
                command_interface => $node,
                volume            => $volume,
                vserver           => $vserver,
                force             => 'true',
            );
        } otherwise {
            my $exception = shift;
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg  => 'VolOps.move.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume,
                        'destination-aggregate' => $dest,
                    },
                }
            );
            $Log->log(NATE::Result::Fail->new($exception->text()));
        };
    }

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.move.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                    'destination-aggregate' => $dest,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Volume move successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Volume move failed'));
    }
} # End of _move

#==============================================================================
# Description: Subprocess responsible for bringing volumes online
#==============================================================================
sub _online {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume = $params->get('volume');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.online.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );
        NACL::C::Volume->online(
            command_interface => $node,
            volume            => $volume,
            vserver           => $vserver,
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        if ($exception->text() =~ /is already online/i) {
            $status = TRUE;
        } else {
            $Log->comment('Failed in volume online operation. Volume: ' .
                $volume . ' Vserver: ' . $vserver
            );
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg  => 'VolOps.online.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume,
                    },
                }
            );
            $Log->log(NATE::Result::Fail->new($exception->text()));
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.online.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Volume online successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Volume online failed'));
    }
} # End of _online

#==============================================================================
# Description: Subprocess responsible for offlining volumes
#==============================================================================
sub _offline {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume = $params->get('volume');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.offline.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );
        NACL::STask::Volume->modify(
            command_interface => $node,
            volume            => $volume,
            vserver           => $vserver,
            state             => 'offline',
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        if ($exception->text() =~ /is already offline/i) {
            $status = TRUE;
        } else {
            $Log->comment('Failed in volume offline operation. Volume: ' .
                $volume . ' Vserver: ' . $vserver
            );
            $MessageBroker->publish_message(
                exchange    => 'errors',
                routing_key => '#',
                message     => {
                    msg  => 'VolOps.offline.' . $exception->text(),
                    data => {
                        node    => $node->{node},
                        vserver => $vserver,
                        volume  => $volume,
                    },
                }
            );
            $Log->log(NATE::Result::Fail->new($exception->text()));
        }
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.offline.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Volume offline successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Volume offline failed'));
    }
} # End of _offline

#==============================================================================
# Description: Subprocess responsible for resizing volumes
#==============================================================================
sub _resize {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vserver = $params->get('vserver');
    my $volume = $params->get('volume');
    my $newsize = $params->get('new-size');
    my $percent = $params->get('percent');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.resize.started',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );

        # Users may pass a percentage value instead of a size. This allows
        # them to 'increase the size of the volume by 20%'

        if ($percent > 0) {
            my $tmp_vol = NACL::CS::Volume->fetch(
                command_interface => $node,
                filter => {
                    volume  => $volume,
                    vserver => $vserver,
                }
            );
            my $size = $tmp_vol->{'size'};
            $newsize = int((1 + ($percent / 100)) * $size);
            $Log->comment("
                $volume will be increased by $percent%. New size will be $newsize"
            );
        }

        NACL::C::Volume->size(
            command_interface => $node,
            volume            => $volume,
            vserver           => $vserver,
            'new-size'        => $newsize,
            'method-timeout'  => TIMEOUT,
        );
        $status = TRUE;
    }
    otherwise {
        my $exception = shift;
        $Log->comment('Failed in volume resize operation. Volume: ' .
            $volume . ' Vserver: ' . $vserver
        );
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.resize.' . $exception->text(),
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );
        $Log->comment($exception->text());
    };

    if ($status) {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VolOps',
            routing_key => '#',
            message     => {
                msg  => 'VolOps.resize.finished',
                data => {
                    node    => $node->{node},
                    vserver => $vserver,
                    volume  => $volume,
                },
            }
        );
        $Log->log(NATE::Result::Pass->new('Volume resize successful'));
    } else {
        $Log->log(NATE::Result::Fail->new('Volume resize failed'));
    }
} # End of _resize

#===============================================================================
# MESSAGE CATALOG
#===============================================================================

#===============================================================================
# Message: volops.create.started
#   Exchange:
#       San::FIT::Misfit::Features::Volops
#
#   Message Structure:
# >     {
# >         "msg": "volops.create.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.create.finished
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.create.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.create.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "volops.create.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.delete.started
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.delete.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.delete.finished
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.delete.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.delete.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "volops.delete.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.clone.started
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.clone.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "flexclone": "vol_1_flexclone",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.clone.finished
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.clone.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >             "flexclone": "vol_1_flexclone",
# >          }
# >     }
#
# Message: volops.clone.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "volops.clone.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >             "flexclone": "vol_1_flexclone",
# >          }
# >     }
#
# Message: volops.split.started
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.split.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "flexclone": "vol_1_flexclone",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.split.finished
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.split.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >             "flexclone": "vol_1_flexclone",
# >          }
# >     }
#
# Message: volops.split.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "volops.split.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >             "flexclone": "vol_1_flexclone",
# >          }
# >     }
#
# Message: volops.online.started
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.online.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.online.finished
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.online.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.online.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "volops.online.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.offline.started
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.offline.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.offline.finished
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.offline.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.offline.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "volops.offline.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.resize.started
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.resize.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.resize.finished
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.resize.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.resize.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "volops.resize.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggregate": "aggr_1",
# >          }
# >     }
#
# Message: volops.move.started
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.move.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggreagte": "aggr1",
# >             "destination-aggregate": "aggr_2",
# >          }
# >     }
#
# Message: volops.move.finished
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.move.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggreagte": "aggr1",
# >             "destination-aggregate": "aggr_2",
# >          }
# >     }
#
# Message: volops.move.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "volops.move.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "aggreagte": "aggr1",
# >             "destination-aggregate": "aggr_2",
# >          }
# >     }
#
# Message: volops.copy.started
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.copy.started",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "dest_volume": "dest_vol",
# >             "dest_agg": "dest_agg",
# >          }
# >     }
#
# Message: volops.copy.finished
#   Exchange:
#       San::FIT::Misfit::Features::VolOps
#
#   Message Structure:
# >     {
# >         "msg": "volops.copy.finished",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "dest_volume": "dest_vol",
# >             "dest_agg": "dest_agg",
# >          }
# >     }
#
# Message: volops.copy.Error Text
#   Exchange:
#       errors
#
#   Message Structure:
# >     {
# >         "msg": "volops.move.Error Text",
# >         "data" : {
# >             "vserver": "vs_1",
# >             "volume": "vol_1",
# >             "dest_volume": "dest_vol",
# >             "dest_agg": "dest_agg",
# >          }
# >     }
#
#===============================================================================

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/Misfit/features/VolOps>

=cut

1;
