# Copyright (c) 2012-2020 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Vserver Operations module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

package San::FIT::Misfit::Features::VserverOps;

# Compiler Directives
use strict;
use warnings;

# Import Libraries
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NATE::ParamSet;
use NACL::C::Vserver;
use San::FIT::Misfit::Core::Workflow;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);

# This package inherits methods and attributes from Misfit's Base class
use base qw( San::FIT::Misfit::Core::Base );

use constant {
    TIMEOUT => 360
};

# Main broker used by this module
my $MessageBroker;

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Description: Vservers Ops constructor
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Features::VserverOps";
    $self->{node} = NACL::C::Node->find();
    $self->{node}->apiset_must(interface => 'ZAPI');
    $MessageBroker = $self->{message_broker};
    unless (defined ($MessageBroker)) {
        $Log->error("Failed to create a message broker for this module");
        return {};
    }
    $Log->exit();

    return $self;
} # End of new

#==============================================================================
# Description: Executes a specific vserver operation
#==============================================================================

#==============================================================================
# Description: Creates 'n' number of vservers
#==============================================================================
sub create {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vservers = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => FALSE,
        'default'  => undef
    );
    my $schedule = &ParamsMgr::get_arg(
        'arg'      => "schedule",
        'required' => FALSE,
        'default'  => '',
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $params = &ParamsMgr::get_arg(
        'arg'      => "params",
        'required' => FALSE,
        'default'  => {},
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $iterations = &ParamsMgr::get_arg(
        'arg'      => "iterations",
        'required' => FALSE,
        'default'  => 1,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my @vserver_info = ();
    my @msg_data = ();
    foreach my $vserver (@{$vservers}) {
        my $runid = '';
        if (defined ($vserver->{id})) {
            $runid = $vserver->{id} . "_create";
        } else {
            $runid = $vserver->{vserver} . "_create";
        }
        push(@msg_data, {vserver => $vserver->{vserver}, vserver => $vserver->{vserver}});
        push (@vserver_info, {
            codespec => "{" . $self->{id} . "::_create(); }",
            runid => $runid,
            background => $background,
            params => {
                %{$vserver},
                id    => $self->{id},
                node  => Clone::clone($self->{node}),
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@vserver_info,
        operation  => "create",
        background => $background,
        iterations => $iterations,
        msg_data   => \@msg_data
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw("Vserver Create failed");
        return FALSE;
    }
} # End of create

#==============================================================================
# PRIVATE METHODS
#==============================================================================

#==============================================================================
# Description: Subprocess responsible for creating vservers
#==============================================================================
sub _create {
    my ($class) = @_;

    my $params = NATE::ParamSet->get_global();
    my $node = $params->get('node');
    my $vs = $params->get('vserver');
    my $status = FALSE;

    try {
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VserverOps',
            routing_key => '#',
            message     => {
                msg  => 'VserverOps.create.started',
                data => {
                    node      => $node->{node},
                    vserver => $vs,
                },
            }
        );
        NACL::STask::Vserver->create(
            command_interface => $node,
            vserver           => $vs
        );
        $MessageBroker->publish_message(
            exchange    => 'San::FIT::Misfit::Features::VserverOps',
            routing_key => '#',
            message     => {
                msg  => 'VserverOps.create.finished',
                data => {
                    node      => $node->{node},
                    vserver => $vs,
                },
            }
        );
        $status = TRUE;
    } otherwise {
        my $exception = shift;
        $Log->comment('Failed to create vservers');
        $MessageBroker->publish_message(
            exchange    => 'errors',
            routing_key => '#',
            message     => {
                msg  => 'VserverOps.create.' . $exception->text(),
                data => {
                    node      => $node->{node},
                    vserver => $vs,
                },
            }
        );
        NATE::BaseException->throw($exception);
    };

    if ($status) {
        return NATE::Result::Pass->new('Vserver create successful');
    } else {
        return NATE::Result::Fail->new('Vserver create failed');
    }
} # End of _create

#===============================================================================
# HANDLERS
#===============================================================================
sub _disruptive_handler {
    my ($self, $queue) = @_;

    $self->broadcast_critical(message => 'stop');
    Tharn::logresult("PASS", msg => "Handler completed");

    return TRUE;
} # End of _disruptive_handler

1;

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/Misfit/features/VserverOps>

=cut

1;
