# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary IO module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Services::IO
# The IO module provides an interface to NetApp IO tools.

package San::FIT::Misfit::Services::IO;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use TCD;
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NACL::C::Node;
use NACL::C::Lun;
use NACL::CS::Lun;
use NATE::Process;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::NotProcManager;
use San::FIT::Misfit::Core::MessageBroker;
use San::FIT::Misfit::Core::Util;
use San::FIT::Misfit::Services::IO::Sanblaze;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Constants
use constant {
    HEADER => "
        use JSON;
        use NATE::Result;
        use Data::Dumper;
        use NATE::Util::DataTypes qw(TRUE FALSE);
        use San::FIT::Misfit::Core::MessageBroker;
        use San::FIT::Misfit::Services::IO;
        use San::FIT::Misfit::Services::IO::Sanblaze;
        use IOMan;
    ",
    RUNTIME                    => 3600,
    MAX_RECORDS                => 999999,
    TWO_SECONDS                => 2,
    SIXTY_SECONDS              => 60,
    SHORT_TIMEOUT              => 5,
    DEFAULT_TIMEOUT            => 10,
    DEFAULT_IO_TOOL            => 'dt',
    KILL_DT_COMMAND            => 'cmd.exe /c "taskkill /f /im dt.exe 2>&1"',
    DEFAULT_TRY_LIMIT          => 10,
    INFINITE_DURATION          => 99999999, # Near-infinite, aka 3.17 years
    DEFAULT_ITERATIONS         => 10,
    DEFAULT_IO_WORKLOAD        => 'trickle',
    DEFAULT_IO_DURATION        => 1800,
    DEFAULT_WAIT_TIMEOUT       => 30,
    DEFAULT_DT_THPL_PATH       => "lib/San/FIT/Misfit/Services/IO/io_dt.thpl",
    DEFAULT_IO_LAB_DOMAIN      => 'gdl.englab.netapp.com',
    DEFAULT_DT_INITIALIZE_TIME => 20, # Takes 20 secs on avg for DT to begin IO
};

#===============================================================================
# Constructor: new
#   IO Service constructor
#
#   Returns:
#       New IO object
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Services::IO";
    $self->{keeper} = undef;
    $self->{keeper_id} = 'io_keeper';
    $self->{io_running} = ();

    # Setup additional exchange for Keeper
    $self->{task_exchanges} = [
        {
            exchange => $self->{keeper_id},
            type     => 'topic'
        },
    ];
    my $exchange = $self->{message_broker}->create_exchange(
        exchanges => $self->{task_exchanges},
    );
    if (!$exchange) {
        $Log->error("Failed to setup exchange for Keeper.");
        $Log->exit();
        return {};
    }

    # Setup additional queues for Keeper
    $self->{operation_queues} = [
        {
            queue       => $self->{keeper_id}.'.start',
            exchange    => $self->{keeper_id},
            routing_key => 'start'
        },
        {
            queue       => $self->{keeper_id}.'.stop',
            exchange    => $self->{keeper_id},
            routing_key => 'stop'
        },
        {
            queue       => $self->{keeper_id}.'.result',
            exchange    => $self->{keeper_id},
            routing_key => 'result'
        },
    ];
    my $queues = $self->{message_broker}->create_queue(
        queues => $self->{operation_queues},
    );
    if (!$queues) {
        $Log->error("Failed to setup queues for Keeper.");
        $self = {};
    }
    $Log->exit();

    return $self;
} # End of new

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Function: start_keeper
#   Start IO keeper
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub start_keeper {
    my ($self, %args) = @_;

    my $runid = "keeper";
    my $paramset = NATE::ParamSet->get_global();
    my $parent_runid = $paramset->get('TOP_LOGDIR').'/'.$paramset->get('RUNID');
    my $logfile = $parent_runid.'/'.$runid.'.log';
    $self->{keeper} = NATE::Process->new(
        codespec     => "{ " . $self->{id} . "::keeper('new'); }",
        runid        => $runid,
        logfile      => $logfile,
        background   => 1,
        independent  => 0,
        inherit_env  => 1,
    );
    $self->{keeper}->start();
    until ($self->{keeper}->is_running()) {
        sleep(1);
        $Log->comment("Keeper exit status: " . $self->{keeper}->{proc}->exit_status());
    }
    $Log->comment($self->{keeper}->{proc}->exit_status());
    if (!$self->{keeper}->is_running()) {
        $Log->exit();
        return FALSE;
    }
    return TRUE;
}

#===============================================================================
# Function: stop_keeper
#   Stop IO keeper
#
#   Returns:
#       TRUE
#===============================================================================
sub stop_keeper {
    my ($self, %args) = @_;

    # Wait for keeper to finish up, then kill and destroy its memory
    # $keeper->wait();
    # Not waiting right now, kill Keeper
    $self->{keeper}->stop();
    $self->{keeper}->results();
    $self->{keeper}->destroy();

    return TRUE;
}

#===============================================================================
# Function: start_io_on_hosts
#   Start IO on specified hosts
#
#   Parameters:
#       hosts - Array of hosts to start IO on.
#       tool - NetApp IO tool to be used for IO.
#       workload - Type of IO to send to controller.
#       duration - Length of time to execute IO sequence.
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub start_io_on_hosts {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my @hosts = &ParamsMgr::get_arg(
        'arg'      => "hosts",
        'required' => TRUE, # array @
    );
    my $tool = &ParamsMgr::get_arg(
        'arg'      => "tool",
        'required' => FALSE,
        'default'  => 'dt', # string ''
    );
    my $workload = &ParamsMgr::get_arg(
        'arg'      => "workload",
        'required' => FALSE,
        'default'  => 'default', # string ''
    );
    my $duration = &ParamsMgr::get_arg(
        'arg'      => "duration",
        'required' => FALSE,
        'default'  => DEFAULT_IO_DURATION, # integer 0
    );
    my $lab_domain = &ParamsMgr::get_arg(
        'arg'      => "lab_domain",
        'required' => FALSE,
        'default'  => DEFAULT_IO_LAB_DOMAIN, # string ''
    );
    my $try_minutes = &ParamsMgr::get_arg(
        'arg'      => "try_minutes",
        'required' => FALSE,
        'default'  => DEFAULT_ITERATIONS, # integer 0
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    unless ($self->{message_broker}->queue_exists(queue => 'io_keeper.result')) {
        $Log->comment("The io_keeper.result queue is missing. I/O will not start");
        $self->local_message(
            message => "io.failed",
            exchange => $self->{id},
            routing_key => "#"
        );
        return FALSE;
    }

    # Send IO start
    $Log->comment("Starting IO on these hosts: ".Dumper(\@hosts));
    foreach my $host (@hosts) {
        my $result_id = $tool . "_" . $host;
        $self->start(
            host       => $host,
            tool       => $tool,
            workload   => $workload,
            duration   => $duration,
            lab_domain => $lab_domain,
            result_id  => $result_id,
        );
    }

    # Confirm IO started
    my @started_hosts = ();
    while ($try_minutes) {
        @started_hosts = $self->{message_broker}->get_message(
            requeue => 'true',
            count   => 10000,
            queue   => 'io_keeper.result',
        );

        if (scalar(@started_hosts) == (scalar(@hosts))) {
            $Log->comment('All hosts have reported, continuing');
            last;
        }
        sleep(SIXTY_SECONDS);
        $try_minutes--;
        $Log->comment($try_minutes . ' minutes until test force-continues. ' .
            scalar(@started_hosts) . ' out of ' .
            scalar(@hosts) . ' hosts have reported.'
        );
    }

    $self->{message_broker}->empty_queue(
        queue => 'io_keeper.result',
    );

    # Fail if IO never started
    if (scalar(@started_hosts) == 0) {
        # Announce that IO never started
        $self->local_message(
            message => "io.failed",
            exchange => $self->{id},
            routing_key => "#"
        );

        $Log->exit();
        return FALSE;
    }

    $self->{io_running} = @started_hosts;

    $Log->comment("IO started on hosts: ".Dumper($self->{started_hosts}));
    # Announce that IO is running
    $self->local_message(
        message  => "io.started",
        exchange => $self->{id},
        routing_key => "#"
    );
    $Log->exit();
    return TRUE;
}

#===============================================================================
# Function: stop_io_on_hosts
#   Stop IO on specified hosts
#
#   Parameters:
#       hosts - List of hosts to stop IO on.
#       tool - NetApp tool that is running IO.
#
#   Returns:
#       TRUE
#===============================================================================
sub stop_io_on_hosts {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my @hosts = &ParamsMgr::get_arg(
        'arg'      => "hosts",
        'required' => TRUE,
    );
    my $tool = &ParamsMgr::get_arg(
        'arg'      => "tool",
        'required' => FALSE,
        'default'  => 'dt', # string ''
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    # Stop IO if not stopped yet
    foreach my $host (@hosts) {
        my $result_id = $tool . "_" . $host;
        # Check quickly in case already stopped
        my $result = $self->result(
            result_id => $result_id,
            timeout   => 2,
        );
        if (defined($result->{contents})) {
            if ($result->{contents} =~ /success/i) {
                $Log->comment($result->{contents});
                next;
            } elsif ($result->{contents} =~ /error/i) {
                $Log->error($result->{contents});
                next;
            }
        }
        # Send command to stop IO if not already stopped
        $self->stop(
            tool => $tool,
            result_id => $result_id,
        );
        # Confirm IO stopped
        while (TRUE) {
            my $result = $self->result(
                result_id => $result_id,
                timeout   => DEFAULT_TIMEOUT,
            );
            if (!defined($result->{contents})) {
                next;
            }
            if ($result->{contents} =~ /success/i) {
                $Log->comment($result->{contents});
                last;
            } elsif ($result->{contents} =~ /error/i) {
                $Log->error($result->{contents});
                last;
            }
        }
    }

    # Announce that IO has completed
    $self->local_message(
        message => "io.finished",
        exchange => $self->{id},
        routing_key => "#"
    );

    $Log->exit();
    return TRUE;
}

#===============================================================================
# Function: keeper
#   Keeper process
#
#   Parameters:
#       duration - Length of time to run keeper for.
#
#   Returns:
#       TRUE
#===============================================================================
sub keeper {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $duration = &ParamsMgr::get_arg(
        'arg'      => "duration",
        'required' => FALSE,
        'default'  => INFINITE_DURATION,
    );
    if ($duration == 0) { # 0 is run until stopped
        $duration = INFINITE_DURATION;
    }
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }



    $Log->comment('Creating ProcManager for IO Keeper...');
    my $paramset = NATE::ParamSet->get_global();
    my $pm = San::FIT::Misfit::Core::NotProcManager->new(
        runid       => "../keeper",
        background  => 1,
        independent => 0,
        proc_info   => [],
    );

    if ($self =~ /new/i) {
        $self = San::FIT::Misfit::Services::IO->new();
    }

    # Start Keeper
    try {
        $self->local_message(
            routing_key => '',
            exchange    => $self->{id},
            message     => {
                keeper => {
                    status => "Started: Creating Keeper with an expected " .
                              "runtime of $duration seconds.",
                }
            }
        );
        my $task = {};
        my $tick = 0;
        my $new_tasks = 0;

        my $start_time = time;
        my $close_time = $start_time + $duration;
        my $actual_duration = 0;
        my $stop_message = "";

        # Collect more start messages
        while (TRUE) {
            # Check if runtime duration has been reached
            if (time >= $close_time) {
                last;
            }
            $Log->comment("Look for a start operation...");
            $task = $self->_receive_operation(
                operation => 'start',
                queue     => $self->{keeper_id},
                timeout   => DEFAULT_WAIT_TIMEOUT,
            );
            # If found start, add a new task
            if (keys(%$task)) {
                if ($task->{message}) {
                    $Log->comment("Invalid message: ".$task->{message});
                    next;
                }
                $tick++;
                $Log->comment("Adding task $tick.");
                # Adding new IO task based on given info
                my $runid = sprintf("../%03d",$tick) . "_" . $task->{result_id};
                $pm->add(
                    proc_info => [{
                        runid   => $runid,
                        codespec => $self->_task(
                            %args,
                            tick    => $tick,
                            details => $task,
                        ),
                    }],
                );
                $new_tasks++;
                # Look for another start task
                next;
            # When no more start tasks are found, start all at once
            } else {
                if ($new_tasks > 0) {
                    # Start new tasks
                    $Log->comment("Starting the new tasks.");
                    $pm->start();
                    # Reset counter to 0
                    $new_tasks = 0;
                } else {
                    # No new tasks, move on
                    $Log->comment(
                        "There were no new tasks in the last " .
                        DEFAULT_WAIT_TIMEOUT . " seconds."
                    );
                }
                # Look for another start task
                next;
            }
        }
        $Log->comment("Wait for everyone to terminate...");
        $pm->wait();
        $Log->comment('Stopping and destroying...');
        $pm->stop();
        $Log->debug(Dumper($pm->results()));
        $pm->destroy();

        $actual_duration = time - $start_time;
        $stop_message = "Closing Keeper, ran for $actual_duration of requested $duration seconds.";
        $Log->comment($stop_message);
        $self->local_message(
            routing_key => '',
            exchange    => $self->{id},
            message     => {
                keeper  => {
                    status  => "Stopped: $stop_message",
                }
            }
        );
    } otherwise {
        my $exception = shift;
        $self->broadcast_error(
            message => {keeper => {status => 'Exception: ' . $exception->text()}}
        );
        $exception->throw();
    };
    $Log->exit();

    return TRUE;
} # End of keeper

#===============================================================================
# Function: start
#   Starts IO of any type
#
#   Returns:
#       TRUE if successful, FALSE otherwise
#===============================================================================
sub start {
    my ($self, %args) = @_;

    $Log->enter();
    my $response = FALSE;

    $response = $self->_send_operation(
        %args,
        operation => 'start',
        id        => $self->{keeper_id},
    );
    $Log->exit();

    return $response;
} # End of start

#===============================================================================
# Function: stop
#   Stops IO of any type
#
#   Returns:
#       TRUE if successful, FALSE otherwise
#===============================================================================
sub stop {
    my ($self, %args) = @_;

    $Log->enter();
    my $response = FALSE;

    $response = $self->_send_operation(
        %args,
        operation => 'stop',
        id        => $self->{keeper_id},
    );
    $Log->exit();

    return $response;
} # End of stop

#===============================================================================
# Function: result
#   Gets Result of IO of any type
#
#   Returns:
#       result
#===============================================================================
sub result {
    my ($self, %args) = @_;

    $Log->enter();

    my $result = $self->_receive_operation(
        %args,
        operation => 'result',
        id        => $self->{keeper_id},
    );

    $Log->exit();

    return $result;
} # End of result

#===============================================================================
# Function: get_successful_io_results_for_hosts
#   Waits for and gets IO results for given array of hosts
#
#   Parameters:
#       hosts - array reference of hostnames for expected results
#       try_minutes - number of minutes to block while waiting for results
#
#   Returns:
#       array of successful hosts
#===============================================================================
sub get_successful_io_results_for_hosts {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $hosts_ref = &ParamsMgr::get_arg(
        'arg'      => "hosts",
        'required' => TRUE,
        'default'  => [],
    );
    my $try_minutes = &ParamsMgr::get_arg(
        'arg'      => "try_minutes",
        'required' => FALSE,
        'default'  => DEFAULT_ITERATIONS, # integer 0
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return ();
    }

    my @hosts = @{$hosts_ref};
    my $expected_hosts = scalar(@hosts);
    my @responded_hosts = ();
    my @check_hosts_again = ();
    my @successful_hosts = ();

    if ($expected_hosts <= 0) {
        $Log->comment("There are no hosts expected to return results.");
        return ();
    }

    # Loop while there is still time
    my $start_time = time;
    my $close_time = $start_time + ($try_minutes * SIXTY_SECONDS);
    while (TRUE) {

        # Cycle through all hosts looking for results
        foreach my $host (@hosts) {

            # Collect result
            my $result = $self->result(
                result_id => $host,
                timeout   => SIXTY_SECONDS,
            );

            # Check for result contents
            if (!defined($result->{contents})) {
                $Log->debug("No result received for $host.");
                push(@check_hosts_again, $host);
                next;
            }

            # Mark response for the host
            push(@responded_hosts, $host);
            $Log->comment(
                "Result received for host $host: " . $result->{contents}
            );

            # Mark successful response for the host
            if ($result->{contents} =~ /success/i) {
                push(@successful_hosts, $host);
            }
            last;
        }

        # Exit while loop if all hosts have responded
        if (scalar(@responded_hosts) == $expected_hosts) {
            $Log->comment("All $expected_hosts hosts have reported results.");
            last;
        }

        # This is some perl magic I found for diffing lists
        # This stores the items in hosts not found in responded_hosts
        # as the keys to the hash.
        my %diff1;
        @diff1{ @hosts } = @hosts;
        delete(@diff1{@responded_hosts});
        @hosts = keys(%diff1);

        $Log->comment(
            "There are " . scalar(@responded_hosts) . " reported results " .
            "of an expected " . $expected_hosts . " responses."
        );

        # Check if try_minutes has been reached
        if (time >= $close_time) {
            $Log->comment(
                "Time ran out for responses, moving on with the test."
            );
            last;
        }

    } # while
    $Log->exit();

    return @successful_hosts;
} # End of get_successful_io_results_for_hosts

#===============================================================================
# Function: send_result
#   Sends Result of IO of any type
#
#   Returns:
#       TRUE if successful, FALSE otherwise
#===============================================================================
sub send_result {
    my ($self, %args) = @_;

    $Log->enter();

    my $response = FALSE;
    $response = $self->_send_operation(
        %args,
        operation => 'result',
        id        => $self->{keeper_id},
    );
    $Log->exit();

    return $response;
} # End of send_result

#===============================================================================
#  PRIVATE METHODS
#===============================================================================
#===============================================================================
# Description: Responsible for sending IO operations to the keeper
#===============================================================================
sub _send_operation {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = undef;
    my $tool = undef;
    my $workload = undef;
    my $duration = undef;
    my $filter = undef;
    my $count = undef;
    my $contents = undef;
    my $lab_domain = undef;
    # Parse args for GENERIC
    my $operation = &ParamsMgr::get_arg(
        'arg'      => "operation",
        'required' => TRUE # values: start, stop...
    );
    my $id = &ParamsMgr::get_arg(
        'arg'      => "id",
        'required' => FALSE,
        'default'  => (caller(0))[3],
    );
    my $targets = &ParamsMgr::get_arg(
        'arg'      => "targets",
        'required' => FALSE,
        'default'  => undef, # array []
    );
    my $devices = &ParamsMgr::get_arg(
        'arg'      => "devices",
        'required' => FALSE,
        'default'  => undef, # array []
    );
    my $result_id = &ParamsMgr::get_arg(
        'arg'      => "result_id",
        'required' => FALSE,
        'default'  => undef,
    );
    # Parse args for START
    if ($operation =~ /(start)/i) {
        $host = &ParamsMgr::get_arg(
            'arg'      => "host",
            'required' => TRUE, # string ''
        );
        $tool = &ParamsMgr::get_arg(
            'arg'      => "tool",
            'required' => TRUE, # string ''
        );
        $workload = &ParamsMgr::get_arg(
            'arg'      => "workload",
            'required' => FALSE,
            'default'  => 'default', # string ''
        );
        $duration = &ParamsMgr::get_arg(
            'arg'      => "duration",
            'required' => FALSE,
            'default'  => DEFAULT_IO_DURATION, # integer 0
        );
        $filter = &ParamsMgr::get_arg(
            'arg'      => "filter",
            'required' => FALSE,
            'default'  => undef, # hash {}
        );
        $count = &ParamsMgr::get_arg(
            'arg'      => "count",
            'required' => FALSE,
            'default'  => 0, # integer 0
        );
        $lab_domain = &ParamsMgr::get_arg(
            'arg'      => "lab_domain",
            'required' => FALSE,
            'default'  => DEFAULT_IO_LAB_DOMAIN # string ''
        );
    # Parse args for STOP
    } elsif ($operation =~ /(stop)/i) {
        $host = &ParamsMgr::get_arg(
            'arg'      => "host",
            'required' => FALSE,
            'default'  => undef, # string ''
        );
        $tool = &ParamsMgr::get_arg(
            'arg'      => "tool",
            'required' => FALSE,
            'default'  => undef, # string ''
        );
        $lab_domain = &ParamsMgr::get_arg(
            'arg'      => "lab_domain",
            'required' => FALSE,
            'default'  => DEFAULT_IO_LAB_DOMAIN, # string ''
        );
    # Parse args for RESULT
    } elsif ($operation =~ /(result)/i) {
        $contents = &ParamsMgr::get_arg(
            'arg'      => "contents",
            'required' => TRUE,
        );
    }
    # Parse args for ROUTING
    my $routing_key = &ParamsMgr::get_arg(
        'arg'      => "routing_key",
        'required' => FALSE,
        'default'  => $operation,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $msg = {};

    if (defined($routing_key)) {
        $msg->{routing_key} = $routing_key;
    }
    if (defined($host)) {
        $msg->{host} = $host;
    }
    if (defined($tool)) {
        $msg->{tool} = $tool;
    }
    if (defined($workload)) {
        $msg->{workload} = $workload;
    }
    if (defined($duration)) {
        $msg->{duration} = $duration;
    }
    if (defined($targets)) {
        $msg->{targets} = $targets;
    }
    if (defined($devices)) {
        $msg->{devices} = $devices;
    }
    if (defined($filter)) {
        $msg->{filter} = $filter;
    }
    if (defined($count)) {
        $msg->{count} = $count;
    }
    if (defined($operation)) {
        $msg->{operation} = $operation;
    }
    if (defined($result_id)) {
        $msg->{result_id} = $result_id;
    }
    if (defined($lab_domain)) {
        $msg->{lab_domain} = $lab_domain;
    }
    if (defined($contents)) {
        $msg->{contents} = $contents;
    }

    $self->{message_broker}->publish_message(
        exchange    => $self->{keeper_id},
        routing_key => $routing_key,
        message     => {
            msg  => $result_id,
            data => {
                keeper => {$operation => $msg}},
            },
    );

    $Log->exit();
    return TRUE;
} # End of _send_operation

#===============================================================================
# Description: Responsible for receiving IO operations for the keeper
#===============================================================================
sub _receive_operation {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $operation = &ParamsMgr::get_arg(
        'arg'      => "operation",
        'required' => TRUE # values: start, stop...
    );
    my $id = &ParamsMgr::get_arg(
        'arg'      => "id",
        'required' => FALSE,
        'default'  => $self->{keeper_id},
    );
    my $timeout = &ParamsMgr::get_arg(
        'arg'      => "timeout",
        'required' => FALSE,
        'default'  => DEFAULT_WAIT_TIMEOUT,
    );
    my $result_id = &ParamsMgr::get_arg(
        'arg'      => "result_id",
        'required' => FALSE,
        'default'  => ".*",
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return undef;
    }
    my $msg = '';
    my $message = '';
    $msg = $self->{message_broker}->wait_for_message(
        verbose => FALSE,
        queue   => $id.".$operation",
        timeout => $timeout,
        message => $result_id,
        ack_all => 'false',
    );
    # Verify success or failure
    if ($msg->{status} =~ /success/i) {
        $Log->debug("Found a message: ".Dumper($msg));
    } else {
        $Log->debug("No messages after waiting: ".Dumper($msg));
        $msg = undef;
    }
    # Verify message exists for operation
    if (!defined($msg)) {
        # No messages
        $message = undef;
    } elsif (defined($msg->{data}->{keeper}->{$operation})) {
        $message = $msg->{data}->{keeper}->{$operation};
        $Log->debug("Message contents are: ".Dumper($message));
    } else {
        $Log->debug("Message is not an operation: ".Dumper($msg));
        $message = undef;
    }
    $Log->exit();
    return $message;
} # End of _receive_operation

#===============================================================================
# Description: Responsible for validating IO operations before execution
#===============================================================================
sub _validate_operation {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $operation = &ParamsMgr::get_arg(
        'arg'      => "operation",
        'required' => TRUE
    );
    my $id = &ParamsMgr::get_arg(
        'arg'      => "id",
        'required' => FALSE,
        'default'  => (caller(0))[3],
    );
    my $result_id = &ParamsMgr::get_arg(
        'arg'      => "result_id",
        'required' => FALSE,
        'default'  => 'generic_result',
    );
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => '',
    );
    my $tool = &ParamsMgr::get_arg(
        'arg'      => "tool",
        'required' => FALSE,
        'default'  => '',
    );
    my $workload = &ParamsMgr::get_arg(
        'arg'      => "workload",
        'required' => FALSE,
        'default'  => 'default',
    );
    my $duration = &ParamsMgr::get_arg(
        'arg'      => "duration",
        'required' => FALSE,
        'default'  => DEFAULT_IO_DURATION,
    );
    my $targets = &ParamsMgr::get_arg(
        'arg'      => "targets",
        'required' => FALSE,
        'default'  => undef,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => undef,
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 0,
    );
    my $randomize_order = &ParamsMgr::get_arg(
        'arg'      => "randomize_order",
        'required' => FALSE,
        'default'  => 0,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return undef;
    }

    my $method = '';
    my $target_data = undef;
    my $client_data = undef;

    # Users may pass a list of targets to use IO or simply a total
    # count and a filter to apply
    if (($operation =~ /start/i) && (defined($targets) || defined($filter))) {
        $method = '_start';
        my $result = '';
        my @found = undef;
        my $count_found = 0;
        my $max_records = MAX_RECORDS;
        if (defined($targets)) {
            $Log->comment("STATUS: Targets defined, fetching target data.");
            $Log->comment(Dumper($targets));
            $count = @$targets;
        } elsif (defined($filter)) {
            $Log->comment("STATUS: Filter defined, fetching targets and data.");
            $Log->comment(Dumper($filter));
            if ($count > 0) {
                $max_records = $count;
            }
        } else {
            NATE::Exceptions::Argument->throw('Unknown targets and filter');
            $Log->exit();
            return undef;
        }
        my $node = NACL::C::Node->find();
        $node->apiset_must(interface => 'ZAPI');
        if (defined($filter)) {
            @found = NACL::C::Lun->find(
                command_interface => $node,
                filter => $filter,
                randomize_order => $randomize_order,
                allow_empty => 1
            );
            $count_found = @found;
            if ($count_found < $count) {
                $result = "STATUS: You asked for $max_records targets, " .
                          "but only $count_found targets exist.";
                $Log->error($result);
            } else {
                $result = "STATUS: Found $count_found targets, " .
                          "and you wanted $max_records targets.";
                $Log->comment($result);
            }
            $self->local_message(
                routing_key => 'result',
                exchange => $self->{id},
                message => {
                    validate => {
                        result_id => $result_id,
                        result    => $result,
                    }
                }
            );
        }
        for (my $i = 0; $i < $count; $i++) {
            if (defined($filter) && ($count_found == 0)) {
                last;
            }
            if (defined($targets)) {
                @found = NACL::C::Lun->find(
                    command_interface => $node,
                    filter => $targets->[$i],
                    allow_empty => 1
                );
                $count_found = @found;
            }
            if ($count_found > 0) {
                my $lun = $found[0]->state();
                $target_data->[$i] =  {
                    vserver  => $lun->vserver,
                    volume   => $lun->volume,
                    lun      => (split('/', $lun->path))[-1], # take end of path
                    lun_real => $lun->lun, # get name from nacl
                    path     => $lun->path,
                    serial   => $lun->serial,
                };
                # example lun output: {
                #   'vserver'  => 'vs0',
                #   'volume'   => 'vol0',
                #   'lun'      => 'lun0',  <- this is parsing end of path
                #   'lun-real' => '-',  <- this is with $lun->lun
                #   'path'     => '/vol/vol0/lun0'
                #   'serial'   => '2Fgnl$B5Q//H',
                # }
                if (defined($targets)) {
                    $result = "STATUS: Target $i found.";
                    $Log->comment($result);
                    $count_found += 1;
                }
            } else {
                $result = "STATUS: Target $i missing.";
                $Log->error($result);
            }
            if (defined($targets)) {
                $self->local_message(
                    routing_key => 'result',
                    exchange => $self->{id},
                    message => {
                        validate => {
                            result_id => $result_id,
                            result    => $result,
                        }
                    }
                );
            }
        }
        if ($count_found >= $count) {
            $result = "SUCCESS: All ($count) targets found and validated.";
            $Log->comment($result);
        } else {
            $result = "ERROR: Failed to find and validate " .
                      "(" . ($count - $count_found) . ") of ($count) targets.";
            $Log->error($result);
            $target_data = undef;
        }
        $self->local_message(
            routing_key => 'result',
            exchange => $self->{id},
            message => {
                validate => {
                    result_id => $result_id,
                    result    => $result,
                }
            }
        );

    } elsif ($operation =~ /start/i) {
        $method = '_start';
        $target_data = $self->_validate_host_devices(
            host      => $host,
            id        => $id,
            result_id => $result_id,
        );

    } elsif ($operation =~ /stop/i) {
        $method = '_stop';
        # FUTURE USE

    } elsif ($operation =~ /status/i) {
        $method = '_status';
        # FUTURE USE

    } elsif ($operation =~ /result/i) {
        $method = '_result';
        # FUTURE USE?

    } else {
        NATE::Exceptions::Argument->throw('Unknown operation');
        $Log->exit();
        return undef;
    }
    $Log->debug('Dump of Final Targets: '.Dumper($target_data));

    $self->{targets} = $target_data;

    $Log->exit();
    return $target_data;
} # End of _validate_operation


#===============================================================================
# Description: Responsible for validating host devices against targets
#===============================================================================
sub _validate_host_devices {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => TRUE,
    );
    my $target_data = &ParamsMgr::get_arg(
        'arg'      => "target_data",
        'required' => FALSE,
        'default'  => undef,
    );
    my $id = &ParamsMgr::get_arg(
        'arg'      => "id",
        'required' => FALSE,
        'default'  => (caller(0))[3],
    );
    my $result_id = &ParamsMgr::get_arg(
        'arg'      => "result_id",
        'required' => FALSE,
        'default'  => 'generic_result',
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    # Temporarily abort if the name contains "sanblaze" so that time is not
    # wasted trying to check Windows RSH commands against a Linux machine
    # that has absolutely no idea what cmd.exe does.
    # TODO: Fix this function properly to confirm devices regardless of host
    if ($host =~ /sanblaze/i) {
        $Log->comment("Sanblaze client detected, skipping device validation.");
        $Log->exit();
        return {};
    }
    # this is done quickly and simply without using normal objects
    # Logic copied from: lib/WinClient/Host.pm
    # mpc show disks first
    my @client_disks = `rsh $host cmd.exe /c "%SystemDrive%\\TOOLS\\mpc.exe show disks 2>&1"`;
    $Log->comment('Dump of Host Disks:'.Dumper(@client_disks));
    my $disks = {};
    foreach my $line (@client_disks) {
        my $disk = undef;
        if ($line =~ /^Disk(\d+)\s+/i) {
            $disk = $1;
        } else {
            next;
        }
        # Collect disk devices in the client.
        if ($line =~ m{        # match
                ^Disk          # line must start with "Disk"
                (\d+)\s+       # Disk#:    1(digit)space          "1 "
                (\d+)\s+       # LUN:      2(digit)space          "0 "
                (\d+\s+\w+)\s+ # Size:     3(digitspaceword)space "1027 M "
                (\w+)\s+       # State:    4(word)space           "OK "
                (\w+)\s+       # Boot:     5(word)space           "False "
                (\w+)\s+       # Vendor:   6(word)space           "NETAPP "
                (\w+)\s+       # Style:    7(word)space           "MBR "
                (\w+)\s+       # Type:     8(word)space           "None"
                (\w+)\s+       # Letter:   9(word)space           "None"
                (\w+)\s+       # FileSys: 10(word)space           "None"
                 \[            #            [                     "["
                (\w+)          # Label:   11(word)                "None"
                 \]\s+         #            ]                     "] "
                (\w+)\s+       # Quorum:  12(word)space           "False "
                (\w+)\s+       # Cluster: 13(word)space           "None "
                (\w+)\s*       # Node:    14(word)                "None "
                $              # line ending
                }isx           # case-insensitive, single-line, allow-comments
            ) {
            $disks->{$disk}->{lun}         = $2;
            $disks->{$disk}->{lunSize}     = $3;
            $disks->{$disk}->{lunState}    = $4;
            $disks->{$disk}->{boot}        = $5;
            $disks->{$disk}->{vendor}      = $6;
            $disks->{$disk}->{partStyle}   = $7;
            $disks->{$disk}->{diskType}    = $8;
            $disks->{$disk}->{mount}       = $9;
            $disks->{$disk}->{fsType}      = $10;
            $disks->{$disk}->{fsLabel}     = $11;
            $disks->{$disk}->{isQuorum}    = $12;
            $disks->{$disk}->{clusterName} = $13;
            $disks->{$disk}->{clusterNode} = $14;
        } else {
            $Log->comment("Client response was unexpected: " . $line);
            $Log->exit();
            return {};
        }
    }
    # mpc show luns second
    my @client_luns = `rsh $host cmd.exe /c "%SystemDrive%\\TOOLS\\mpc.exe show luns 2>&1"`;
    $Log->comment('Dump of Host LUNs: ' . Dumper(@client_luns));
    foreach my $line (@client_luns) {
        my $disk = undef;
        if ($line =~ /^Disk(\d+)\s+/i) {
            $disk = $1;
        } else {
            next;
        }
        # Collect cluster values for devices.
        # Serial numbers can contain: a-zA-Z0-9_$%&*+-<=>?@[/\]^~; according to:
        # http://wikid.netapp.com/w/Smokejumper/Bcom/Documents/Serial_Number_Generation
        # Have to avoid using the ] in regex comment due to processing error
        # Have to avoid using the - in regex comment due to processing error
        # Have to take fewest +? of Serial and Path chars otherwise greedy
        if ($line =~ m{    # match
                ^Disk      # line must start with "Disk"
                (\d+)\s+   # Disk#:    1(digit)space   "1 "
                (\d+)\s+   # LUN:      2(digit)space   "0 "
                (\w+)\s+   # DSM:      3(word)space    "ontap "
                (\w+)\s+   # Policy:   4(word)space    "DLQD "
                (\d+)\s+   # Paths:    5(digit)space   "6"
                ([         # Serial:   6(charset
                    \w     #             word
                    \$     #             $
                    \%     #             %
                    \&     #             &
                    \*     #             *
                    \+     #             +
                    \-     #             dash character
                    \<     #             <
                    \=     #             =
                    \>     #             >
                    \?     #             ?
                    \@     #             @
                    \[     #             [
                    \]     #             close square bracket character
                    \^     #             ^
                    \~     #             ~
                    \;     #             ;
                    \/     #             /
                    \\     #             \
                 ]+?)\s+   #             charset)space "7S4iL+BL5em/ "
                (\w+)\s+   # CFMode:   7(word)space    "SSI "
                (\w+)\s+   # Filer:    8(word)space    "vs_4n_iscsi_3 "
                (\w+)\s+   # Proto:    9(word)space    "iSCSI "
                (\w+)\s+   # LunType: 10(word)space    "windows "
                ([         # Path:    11(charset
                    \w     #             word
                    \.     #             .
                    \-     #             dash character
                    \/     #             /
                 ]+?)\s*   #             charset)space "/vol/vl_35/ops.lun_16 "
                $          # line ending
                }isx       # case-insensitive, single-line, allow-comments
            ){
            $disks->{$disk}->{lun}      = $2;
            $disks->{$disk}->{dsm}      = $3;
            $disks->{$disk}->{policy}   = $4;
            $disks->{$disk}->{paths}    = $5;
            $disks->{$disk}->{serial}   = $6;
            $disks->{$disk}->{cfMode}   = $7;
            $disks->{$disk}->{vserver}  = $8;
            $disks->{$disk}->{protocol} = $9;
            $disks->{$disk}->{lunType}  = $10;
            $disks->{$disk}->{path}     = $11;
        } else {
            $Log->comment("Client response was unexpected: " . $line);
            $Log->exit();
            return {};
        }
    }
    $Log->debug("Dump of Compiled Host Disk Information:" . Dumper($disks));

    my $device_data = [];
    my $lun_serials = '';
    if (defined($target_data)) {
        foreach my $target (@$target_data) {
            $lun_serials .= ' '.$target.' ';
        }
    }
    foreach my $disk (keys %$disks) {
        my $serial = $disks->{$disk}->{serial};
        my $device = "\\\\.\\PHYSICALDRIVE".$disk;
        if ( !defined($serial) || (index($lun_serials, $serial) == 0)) {
            $Log->comment(
                "Serial $serial for device $device was not a requested target."
            );
            next;
        }
        my $path = $disks->{$disk}->{path};
        # Take count of array and use it as new index for next device
        $device_data->[scalar @$device_data] = {
            vserver   => $disks->{$disk}->{vserver},
            volume    => (split('/', $path))[-2], # take middle of path
            lun       => (split('/', $path))[-1], # take end of path
            path      => $path,
            num_paths => $disks->{$disk}->{paths},
            serial    => $disks->{$disk}->{serial},
            device    => $device,
        };
        $Log->comment("Serial $serial was found as device $device.");
    }

    $Log->exit();
    return $device_data;
} # End of _validate_host_devices


#===============================================================================
# Description: Generates a codestring for an IO task
#===============================================================================
sub _task {
    my ($self, %args) = @_;

    $Log->enter();

    $Log->comment("Creating new task.");

        # Removing this temporarily until I have better control over exchanges
        #\$io->_create_exchanges_and_queue(
        #    targets => \$targets,
        #    host => '".$args{details}->{host}."',
        #    tool => '".$args{details}->{tool}."',
        #);

        # Temporarily removing this unless needed later
        #\$io->{message_broker} = San::FIT::Misfit::Core::MessageBroker->new(
        #    vhost => '".$self->{vhost}."',
        #);
        #\$io->local_message(
        #    exchange    => '".$self->{id}."',
        #    routing_key => 'result',
        #    message     => {".deref($message)."},
        #);

    my $codestring = "{" . HEADER . "
        # Create new IO object for task
        my \$io = San::FIT::Misfit::Services::IO->new();

        # Validate and return targets for IO
        my \$targets = \$io->_validate_operation(
            ".deref($args{details})."
        );

        \$io->_create_task_queue(
            host => '".$args{details}->{host}."',
            tool => '".$args{details}->{tool}."',
            result_id => '".$args{details}->{result_id}."',
        );

        # DT TASK
        if ('dt' =~ /".$args{details}->{tool}."/i) {
            my \$dt = \$io->_run_dt(
                devices => \$targets,
                host => '".$args{details}->{host}."',
                lab_domain => '".$args{details}->{lab_domain}."',
                workload => '".$args{details}->{workload}."',
                duration => ".$args{details}->{duration}.",
                result_id => '".$args{details}->{result_id}."',
            );

        # SANBLAZE TASK
        } elsif ('sanblaze' =~ /".$args{details}->{tool}."/i) {
            my \$sanblaze = \$io->_run_sanblaze(
                luns => \$targets,
                host => '".$args{details}->{host}."',
                lab_domain => '".$args{details}->{lab_domain}."',
                workload => '".$args{details}->{workload}."',
                duration => ".$args{details}->{duration}.",
                result_id => '".$args{details}->{result_id}."',
            );
        } else {
            return NATE::Result::Fail->new('IO Task is using an unknown tool.');
        }
        return NATE::Result::Pass->new('IO Task Completed.');
    }";

    $Log->debug("Task codestring is as follows:" . Dumper($codestring));
    $Log->exit();

    return $codestring;
} # End of _task

#===============================================================================
# Description: Creates and runs an DT object with IOMan
#===============================================================================
sub _run_dt {
    my ($self, %args) = @_;
    my $response = $TCD::FAIL;

    $Log->enter();

    my $luns        = $args{luns};
    my $host        = $args{host};
    my $workload    = $args{workload};
    my $duration    = $args{duration};
    my $result_id   = $args{result_id};
    my $lab_domain  = $args{lab_domain};
    my $device_data = $args{devices};

    my $devices = [];
    foreach my $device (@$device_data) {
        $devices->[scalar @$devices] = $device->{device};
    }

    $Log->debug('devices_list: '.Dumper($devices));

    if (scalar(@{$devices}) <= 0) {
        $self->send_result(
            result_id => $args{result_id},
            contents  => 'ERROR: IO not started with DT, there were no devices.',
        );
        $Log->error("There are no devices to start on the client.");
        $Log->exit();
        return $TCD::FAIL;
    }

    my $testcase = "";
    if ($workload =~ /^default$/i) {
        $testcase = "test_dt_default_load";
    }
    elsif ($workload =~ /^dss$/i) {
        $testcase = "test_dt_dss_load";
    }
    elsif ($workload =~ /^database$/i) {
        $testcase = "test_dt_database_load";
    }
    elsif ($workload =~ /^oltp$/i) {
        $testcase = "test_dt_oltp_load";
    }
    elsif ($workload =~ /^large_seq_reads$/i) {
        $testcase = "test_dt_large_sequential_reads";
    }
    elsif ($workload =~ /^trickle$/i) {
        $testcase = "test_dt_trickle";
    }
    elsif ($workload =~ /^log$/i) {
        $testcase = "test_dt_log_load";
    }
    elsif ($workload =~ /^exchange$/i) {
        $testcase = "test_dt_ms_exchange_load";
    }
    elsif ($workload =~ /^custom$/i) {
        $testcase = "test_dt_custom";
    }

    my $runid = "task_run_dt";
    my $paramset = NATE::ParamSet->get_global();
    my $parent_runid = $paramset->get('TOP_LOGDIR').'/'.$paramset->get('RUNID');
    my $logfile = $parent_runid.'/'.$runid.'.log';

    my $dt = NATE::Process->new(
        codespec    => DEFAULT_DT_THPL_PATH,
        runid       => $runid,
        logfile     => $logfile,
        background  => 1,
        independent => 0,
        inherit_env => 0,
        params      => {
            DEVICES             => {devices => $devices},
            CLIENTS             => $host,
            WORKLOAD            => $workload,
            LOG_STATISTICS      => 0,
            RUNTIME             => $duration,
            TESTCASES           => $testcase,
            IO_DEVICES_PER_TOOL => "all",
            IO_TOOL_VERSION     => "staging",
        },
    );

    # Start DT process
    $dt->start();
    if ($dt->is_running()) {
        $self->send_result(
            result_id => $args{result_id},
            contents  => 'SUCCESS: Started IO using DT',
        );
    } else {
        $self->send_result(
            result_id => $args{result_id},
            contents  => 'ERROR: IO not started with DT',
        );
        $Log->exit();
        return $TCD::FAIL;
    }

    # Loop while waiting to Stop IO
    my $start_time = time;
    my $close_time = $start_time + $duration;
    while (TRUE) {

        $Log->comment("Waiting for message to stop DT IO.");
        my $stop = $self->_receive_operation(
            operation => 'stop',
            queue     => $self->{keeper_id},
            timeout   => DEFAULT_WAIT_TIMEOUT,
            result_id => $args{result_id},
        );

        # Check if stop message was received or runtime expired
        if (defined($stop) or (time >= $close_time)) {
            my $reason = '';
            if (defined($stop)) {
                $reason = "Stopping DT IO, STOP message: " . Dumper($stop);
            } else {
                $reason = "Stopping DT IO, runtime ($duration) expired."
            }
            $Log->comment($reason);
            last;
        }
    }

    $Log->comment("Stopping and destroying DT process after completion.");
    $dt->stop();

    # Allow time to stop...
    sleep(SHORT_TIMEOUT);

    # Kills dt process on current host
    my $command = "rsh $args{host} " . KILL_DT_COMMAND;
    my $output = `$command`;
    $Log->comment($output);

    $dt->destroy();

    if (!$dt->is_running()) {
        $self->send_result(
            result_id => $args{result_id},
            contents  => 'SUCCESS: Stopped IO that used DT',
        );
        $response = $TCD::PASS;
    } else {
        $self->send_result(
            result_id => $args{result_id},
            contents  => 'ERROR: IO might not be stopped that used DT',
        );
    }

    $Log->exit();

    return $response;
} # End of _run_dt

#===============================================================================
# Function: _run_sanblaze
#   Communicates to a Sanblaze server to start and stop IO load.
#
#   Parameters:
#       host - Hostname for Sanblaze server
#       lab_domain - Remainder of FQDN for Sanblaze
#       workload - Type of IO load to execute
#       duration - Time in seconds to run IO
#       result_id - Response queue to send messages
#       luns - (Not yet implemented) Hash of LUNs to run IO
#
#   Returns:
#       TCD::PASS or TCD::FAIL
#
#   Example:
#       > my $sanblaze = San::FIT::Misfit::Services::IO->_run_sanblaze(
#       >     luns => $targets,
#       >     host => "ssan-sanblaze-2",
#       >     lab_domain => "gdl.englab.netapp.com",
#       >     workload => "oltp",
#       >     duration => 300,
#       >     result_id => "ssan-sanblaze-2_oltp",
#       > );
#
#===============================================================================
sub _run_sanblaze {
    my ($self, %args) = @_;

    my $status = FALSE;
    my $response = $TCD::FAIL;

    # Initialize workload variables with acceptable defaults
    # TODO: Move these into a separate library for Workloads documentation
    # TODO: Move testcases for sanblaze into new Workload library

    my $test_type = "Compare";

    # Variable: threads
    #   Number of threads to start per connected LUN:
    #       1 - single IO thread per lun
    #     512 - maximum threads per lun per port
    my $threads = 1;

    # Variable: blocks
    #   Number of blocks (LBNs) per IO, aka, size of IO in blocks:
    #   Note: smaller blocksize increases CPU, decreases throughput
    #   and larger blocksize decreases CPU, but increases throughput.
    #     512 - good SANFIT approved default
    #      -1 - random blocksize
    my $blocks = 512;

    # Variable: ios
    #   Number of IOs to run per thread:
    #       5 - stops thread after 5 IOs are done
    #       0 - unlimited, runs until manually stopped
    my $ios = 0;

    # Variable: pass_ios
    #   Required for "Compare" test: number of IOs to write before reading back
    #       4 - performs 4 writes first, then 4 reads
    my $pass_ios = 4;

    # Variable: seek_type
    #   IO access:
    #       0 - sequential
    #       1 - random
    #       2 - min/max
    #       3 - butterfly
    my $seek_type = 1;

    # Variable: opcode
    #   IO opcode to use for read/write commands:
    #       0 - random (no reason to change this)
    #       6, 10, 12, 16 - available options that we don't use
    my $opcode = 0;

    # Variable: paused
    #   Start test in paused state:
    #       0 - no, start test immediately
    #       1 - yes, pause test until manually started
    my $paused = 0;

    # Variable: initiator
    #   Number of initiator alias to use for the test:
    #       5 - use only initiator alias with value of 5
    #      -1 - all initiators
    my $initiator = -1;

    # Variable: multipath_mode
    #   If multiple paths are detected, use:
    #       0 - default setting
    #       1 - this path
    #       2 - one path
    #       3 - all paths
    #       4 - act/opt paths
    #       5 - act/non-opt paths
    my $multipath_mode = 4;

    # Variable: pattern
    #   Pattern used for writes:
    #       0 - random pattern
    #       1 - 0x00ff00ff
    #       2 - 0x55aa55aa
    #       3 - 8-bit incrementing
    #       4 - 8-bit walking 1,0
    #       5 - 0x0000ffff
    #       6 - 0x5555aaaa
    #       7 - 16-bit incrementing
    #       8 - 16-bit walking 1,0
    #       9 - 32-bit incrementing
    #      -1 - custom (uses dir: /virtualun/lundata/initiatorpatternfiles/)
    my $pattern = 2;

    $Log->enter();

    my $luns       = $args{luns};
    my $host       = $args{host};
    my $workload   = $args{workload};
    my $duration   = $args{duration};
    my $result_id  = $args{result_id};
    my $lab_domain = $args{lab_domain};

    my $ports        = 0;
    my $host_address = "";

    if (($workload =~ /^test_compare_io_with_12_ports$/i) or
        ($workload =~ /^oltp$/i) or
        ($workload =~ /^dss$/i) or
        ($workload =~ /^log$/i) or
        ($workload =~ /^ms$/i)
        ) {
        $Log->comment("Sanblaze workload '$workload' is now being activated.");
        $ports = 12;
        $test_type      = "Compare";  # compare
        $threads        = 1;          # 1: single thread
        $blocks         = 512;        # 512 blocks per io
        $ios            = 0;          # 0: unlimited
        $pass_ios       = 4;          # 4: writes before reads
        $seek_type      = 1;          # 1: random
        $opcode         = 0;          # 0: random
        $paused         = 0;          # 0: not paused
        $initiator      = -1;         # -1: all initiators
        $multipath_mode = 4;          # 4: active/optimized
        $pattern        = 2;          # 2: 0x55aa55aa
        $host_address = "http://" . $host . "." . $lab_domain;
    }
    else {
        $Log->error("Sanblaze workload '$workload' does not exist.");
    }

    # If there are no ports provided, we cannot start IO and must exit
    if ($ports <= 0) {
        $Log->error("There were no ports provided to start Sanblaze IO.");
        $self->send_result(
            result_id => $result_id,
            contents  => 'ERROR: Failed starting IO using Sanblaze, no ports.',
        );
        $Log->exit();
        return $response;
    }

    # Create Sanblaze object
    my $sanblaze = {};
    $sanblaze = San::FIT::Misfit::Services::IO::Sanblaze->new(
        host       => $host_address,
        port_range => undef,
    );

    # Start IO on ports
    my $start = undef;
    for (my $port = 0; $port < $ports; $port++) {
        my $result = $sanblaze->sanblaze_start_test(
            host           => $host_address,
            port           => $port,
            test_type      => $test_type,
            threads        => $threads,
            blocks         => $blocks,
            ios            => $ios,
            seek_type      => $seek_type,
            opcode         => $opcode,
            paused         => $paused,
            initiator      => $initiator,
            multipath_mode => $multipath_mode,
            pattern        => $pattern,
            pass_ios       => $pass_ios,
        );

        # Track and report if successful to start IO
        if ($result != 0) {
            $Log->comment("Started IO on Port $port");
            $start->{$port} = $result;
        } else {
            $Log->comment("Failed to start IO on Port $port");
        }

        # Delay implemented so Sanblaze is not overwhelmed (this happens)
        sleep(TWO_SECONDS);
    }

    # If IO did not start on any port, then report it and exit
    unless (defined($start)) {
        $Log->error("There was no IO started on any Sanblaze port.");
        $self->send_result(
            result_id => $result_id,
            contents  => 'ERROR: Failed starting any IO using Sanblaze.',
        );
        $Log->exit();
        return $response;
    }

    # Announce that IO started
    $Log->comment("Sending result proclaiming IO started on Sanblaze.");
    $self->send_result(
        result_id => $result_id,
        contents  => 'SUCCESS: Started IO using Sanblaze',
    );

    # Loop while waiting to Stop IO and grab statistics periodically
    my $start_time = time;
    my $close_time = $start_time + $duration;
    while (TRUE) {

        # No ports were started, don't loop
        if ($ports <= 0) {
            last;
        }

        # Collect IO statistics for each port
        $Log->comment("Collecting stats for each port.");
        for (my $port = 0; $port < $ports; $port++) {

            # Abort checking statistics if port was not started
            if (not defined($start->{$port})) {
                $Log->comment("Statistics collection failed for port $port");
                next;
            }

            # Get status for given test_id
            my $test_id = $start->{$port}->{test}->{test_id};
            my $test_status = $sanblaze->sanblaze_test_status(
                host    => $host_address,
                port    => $port,
                test_id => $test_id,
            );

            # Abort output of statistics if test_bytes is not defined or is 0
            if (not defined($test_status->{test_bytes})) {
                $Log->comment("Port $port is reporting no test status.");
                $Log->comment(Dumper($test_status));
                next;
            } elsif ($test_status->{test_bytes} == 0) {
                $Log->comment("Port $port is reporting 0 bytes of IO.");
                $Log->comment(Dumper($test_status));
                next;
            }

            # Report IO statistics for this port/test_id
            $Log->comment(
                "Port: ".$port.", ".
                "TestID: ".$test_status->{test_id}.", ".
                "MBytes: ".int($test_status->{test_bytes}/1048576).", ".
                "IOs: ".$test_status->{test_ios}.", ".
                "Time: ".$test_status->{test_time}.", ".
                "LUNs: ".$test_status->{num_test_luns}
            );
        }

        # Check for a Stop message
        my $stop = undef;
        $Log->comment("Waiting for message to stop Sanblaze IO.");
        $stop = $self->_receive_operation(
            operation => 'stop',
            queue     => $self->{keeper_id},
            timeout   => DEFAULT_WAIT_TIMEOUT,
            result_id => $args{result_id},
        );

        # Check if stop message was received
        if (defined($stop)) {
            $Log->comment(
                "Stopping Sanblaze IO due to STOP message: " . Dumper($stop)
            );
            last;
        }
        # Check if runtime duration has been reached
        if (time >= $close_time) {
            $Log->comment(
                "Stopping Sanblaze IO, runtime ($duration) has been reached."
            );
            last;
        }
    }

    # Stop IO on ports
    for (my $port = 0; $port < $ports; $port++) {

        # Abort stopping if port was not started
        if (not defined($start->{$port})) {
            $Log->comment("No test to stop on Port $port");
            next;
        }

        # Stop test on specific port
        $Log->comment("Stopping Sanblaze Test on Port $port");
        my $result = $sanblaze->sanblaze_stop_test(
            host    => $host_address,
            port    => $port,
            test_id => $start->{$port}->{test}->{test_id},
        );
        $Log->comment(Dumper($result));

        # Delay implemented so Sanblaze is not overwhelmed (this happens)
        sleep(TWO_SECONDS);
    }

    $status = TRUE;

    if ($status) {
        $Log->comment("Sending result proclaiming IO stopped on Sanblaze.");
        $self->send_result(
            result_id => $args{result_id},
            contents  => 'SUCCESS: Stopped IO that used Sanblaze',
        );
        $response = $TCD::PASS;
    } else {
        $Log->comment("Sending result proclaiming IO unknown on Sanblaze.");
        $self->send_result(
            result_id => $args{result_id},
            contents  => 'ERROR: IO might not be stopped that used Sanblaze',
        );
    }

    $Log->exit();

    return $response;
} # End of _run_sanblaze

#===============================================================================
# Description: Creates the requested exchanges and queue for a task
#===============================================================================
sub _create_task_queue {
    my ($self, %args) = @_;
    my $response = FALSE;

    $Log->enter();

    if (!defined $args{result_id}) {
        $Log->exit();
        return $response;
    }
    # Setup additional task queue with bindings
    my $task_queue = $self->{keeper_id}.'.'.$args{result_id};
    $self->{task_queues} = [
        {
            queue       => $task_queue,
            exchange    => $self->{keeper_id},
            routing_key => 'stop'
        },
        {
            queue       => $task_queue,
            exchange    => $self->{keeper_id},
            routing_key => 'result'
        },
        {
            queue       => $task_queue,
            exchange    => $self->{keeper_id},
            routing_key => $args{tool}
        },
        {
            queue       => $task_queue,
            exchange    => $self->{keeper_id},
            routing_key => $args{host}
        },
        {
            queue       => $task_queue,
            exchange    => $self->{keeper_id},
            routing_key => $args{result_id}
        },
    ];
    # Default queue for IO task
    my $create = $self->{message_broker}->create_queue(
        queues => $self->{task_queues},
    );
    if ($create) {
        $response = TRUE;
    } else {
        $Log->error("Failed to create queues for task.");
    }
    $Log->exit();

    return $response;
} # End of _create_task_queue

#===============================================================================
# Function: _create_exchanges_and_queue
#   Creates the requested exchanges and queue for a task
#===============================================================================
sub _create_exchanges_and_queue {
    my ($self, %args) = @_;

    $Log->enter();
    my $response = FALSE;
    if (!defined $args{targets}) {
        $Log->exit();
        return $response;
    }
    # Variable: task_exchanges
    #   Setup additional operational exchanges
    $self->{task_exchanges} = [
        {
            exchange => 'io_tasks',
            type     => 'fanout'
        },
        {
            exchange => 'io_tool::'.$args{tool},
            type     => 'fanout'
        },
        {
            exchange => 'io_host::'.$args{host},
            type     => 'fanout'
        },
    ];
    foreach my $target (@{$args{targets}}) {
        # I'll include this later, right now the name is really long
        # $Log->comment('count is: '.scalar @{$self->{task_exchanges}});
        # $self->{task_exchanges}->[scalar @{$self->{task_exchanges}}] = {
        #     exchange => 'io_target::'.$target->{vserver}.'::'.$target->{path},
        #     type     => 'fanout'
        # };
        $Log->comment('count is: '.scalar @{$self->{task_exchanges}});
        $self->{task_exchanges}->[scalar @{$self->{task_exchanges}}] = {
            exchange => 'io_vserver::'.$target->{vserver},
            type     => 'fanout'
        };
        $Log->comment('count is: '.scalar @{$self->{task_exchanges}});
        $self->{task_exchanges}->[scalar @{$self->{task_exchanges}}] = {
            exchange => 'io_volume::'.$target->{volume},
            type     => 'fanout'
        };
        $Log->comment('count is: '.scalar @{$self->{task_exchanges}});
        $self->{task_exchanges}->[scalar @{$self->{task_exchanges}}] = {
            exchange => 'io_lun::'.$target->{lun},
            type     => 'fanout',
        };
    }

    # TODO NEED BINDINGS FOR EXCHANGES

    $Log->comment(Dumper($self->{task_exchanges}));

    # Default exchange for Misfit
    my $create = $self->{message_broker}->create_exchange(
        exchanges => $self->{task_exchanges},
    );
    if ($create) {
        $response = TRUE;
    } else {
        $Log->error("Failed to create exchanges for task queue.");
    }
    $Log->exit();

    return $response;
} # End of _create_exchanges_and_queue

#===============================================================================
# Description: Completely dereference using Dumper
#===============================================================================
sub deref {
    my $ref = shift;

    $Log->enter();
    local $Data::Dumper::Terse = 1;
    local $Data::Dumper::Indent = 0;

    $Log->debug($ref);
    my $not_ref = Dumper($ref);
    # strip { and } from begin and end of string
    $not_ref = substr($not_ref, 1, (length($not_ref) - 2));
    $Log->debug($not_ref);
    $Log->exit();

    return $not_ref;
} # End of deref

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://wikid.netapp.com/w/Misfit/services/IO>

=cut

1;
