# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary SANBlaze IO API
## @author dl-san-fit@netapp.com

package San::FIT::Misfit::Services::IO::Sanblaze;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NATE::Util::DataTypes qw(TRUE FALSE);
use NATE::Exceptions::Argument qw(:try);
use LWP::UserAgent;
use XML::Simple;

use constant {
    BLOCKS_RANDOM => -1,
    OPCODE_RANDOM => 0,
    PATTERN_RANDOM => 0,
    MULTIPATH_DEFAULT => 0,
    ALL_TEST_IDS => 'ALL',
    ALL_INITIATORS => -1,
    ALL_LUNS => 2
};

###############################################################################
# Public Methods
###############################################################################

###############################################################################
# Description:
#    Sanblaze Constructor
###############################################################################
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    my $self = bless {}, $class;
    &ParamsMgr::process_args(%args);
    $self->{host} = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => TRUE
    );
    $self->{port_range} = &ParamsMgr::get_arg(
        'arg'      => "port_range",
        'required' => FALSE,
        'default'  => undef
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    #Remove trailing '/' if it is in the hostname
    if(substr($self->{host},-1,1) eq '/') {
        chop($self->{host});
    }

    # If host does not contain the protocol 'http://', add it
    if ($self->{host} !~ /^http\:\/\//) {
        $self->{host} = "http://" . $self->{host};
    }

    $Log->debug("[INFO] Sanblaze Object Setup");
    $Log->exit();
    return $self;
}

###############################################################################
# Description:
#    vlunSystemInfo API Call. Return system software revision information.
# Parameters:
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_system_info {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Getting System Info @ $host");
    my $result = $self->_call (
        host   => $host,
        method => "vlunSystemInfo"
    );
    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunPortList API Call. Returns list of all physical ports in a given
#    system.
# Parameters:
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_port_list {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return [];
    }
    $Log->debug("[INFO] Getting Port List @ $host");
    my $result = $self->_call (
        host   => $host,
        method => "vlunPortList"
    );
    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunPortStatus API Call. Return port level status and performance data
#    for a given port.
# Parameters:
#    Required:
#        port = physical port number being used
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_port_status {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Getting Port $port Status @ $host");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunPortStatus",
        data     => "port=$port"
    );
    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunTargetList API Call. Return a list of targets on a give port with
#    links to their status pages.
# Parameters:
#    Required:
#        port = physical port number being used
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_target_list {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return [];
    }
    $Log->debug("[INFO] Getting Target List from Port $port @ $host");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunTargetList",
        data     => "port=$port"
    );
    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunTargetStatus API Call. Returns status information for a given target
#    and a list of luns under that target.
# Parameters:
#    Required:
#        port = physical port number being used
#        target = target number the lun is under
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_target_status {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $target = &ParamsMgr::get_arg(
        'arg'      => "target",
        'required' => TRUE
        );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Getting Target $target Status on Port $port @ $host");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunTargetStatus",
        data     => "port=$port&target=$target"
    );
    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunLunSearch API Call. Returns link to lun status for a given lun serial.
# Parameters:
#    Required:
#        lun_serial = serial number of lun
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_lun_search {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $lun_serial = &ParamsMgr::get_arg(
        'arg'      => "lun_serial",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    $Log->comment("[INFO] Performing Lun Search for Serial $lun_serial @ $host");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunLunSearch",
        data     => "serial_number=$lun_serial"
    );
    if ($result) {
        $Log->comment("[SUCCESS] Lun $lun_serial Found. ");
        $Log->exit();
        return {
            port => $result->{port},
            target => $result->{target},
            lun => $result->{luns}->{lun}->{id},
        };
    } else {
        $Log->comment("[ERROR] Lun Serial $lun_serial not Found (sanblaze_lun_status)");
        $Log->exit();
        return FALSE;
    }
    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunLunStatus API Call. Returns status and performance information for
#    a given lun.
# Parameters:
#    Required:
#        port = physical port number being used
#        target = target number the lun is under
#        lun = lun id to get status for
#        OR
#        lun_serial = serial number of lun
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_lun_status {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => ''
    );
    my $target = &ParamsMgr::get_arg(
        'arg'      => "target",
        'required' => FALSE,
        'default'  => ''
    );
    my $lun = &ParamsMgr::get_arg(
        'arg'      => "lun",
        'required' => FALSE,
        'default'  => ''
    );
    my $lun_serial = &ParamsMgr::get_arg(
        'arg'      => "lun_serial",
        'required' => FALSE,
        'default'  => undef
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    if(defined($lun_serial)) {
        my $lun_search_result = $self->sanblaze_lun_search(
            host => $host,
            lun_serial => $lun_serial
        );
        if ($lun_search_result) {
            $port = $lun_search_result->{port};
            $target = $lun_search_result->{target};
            $lun = $lun_search_result->{lun};
        } else {
            $Log->debug("[ERROR] Lun Serial $lun_serial not Found (sanblaze_lun_status)");
            $Log->exit();
            return FALSE;
        }
    }
    $Log->debug("[INFO] Getting Lun $lun Status on Port $port Target $target @ $host");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunLunStatus",
        data     => "port=$port&target=$target&lun=$lun"
    );
    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunTestList API Call. Return a list of tests for a given port.
#    Includes both currently active tests and completed tests.
# Parameters:
#    Required:
#        port = physical port number being used
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_test_list {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Getting Test List on Port $port @ $host");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunTestList",
        data     => "port=$port"
    );
    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunTestStatus API Call. Return status and statistics for a given
#    test-id.
# Parameters:
#    Required:
#        port = physical port number being used
#        OR
#        lun_serial = serial number of lun the test is being run on
#        test_id = test-id string of test being requested
#   Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_test_status {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => ''
    );
    my $lun_serial = &ParamsMgr::get_arg(
        'arg'      => "lun_serial",
        'required' => FALSE,
        'default'  => undef
    );
    my $test_id = &ParamsMgr::get_arg(
        'arg'      => "test_id",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    if(defined($lun_serial)) {
        my $lun_search_result = $self->sanblaze_lun_search(
            host => $host,
            lun_serial => $lun_serial
        );
        if ($lun_search_result) {
            $port = $lun_search_result->{port};
        } else {
            $Log->debug("[ERROR] Lun Serial $lun_serial not Found (sanblaze_test_status)");
            $Log->exit();
            return FALSE;
        }
    }

    $Log->debug("[INFO] Getting Test $test_id Status on Port $port @ $host");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunTestStatus",
        data     => "port=$port&test_id=$test_id"
    );
    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunGetPortConfig API Call. Report port level settings.
# Parameters:
#    Required:
#        port = physical port number being used
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_get_port_config {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Getting Port $port Config @ $host");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunGetPortConfig",
        data     => "port=$port"
    );
    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunGetPortTestParameters API Call. Report port level test parameter
#    values.
# Parameters:
#    Required:
#        port = physical port number being used
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_get_port_test_parameters {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Getting Port $port Test Parameters @ $host");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunGetPortTestParameters",
        data     => "port=$port"
    );
    $Log->exit();
    return $result;
}

###############################################################################
# The following API calls will convert the given data into XML before it is
# passed to the SANBlaze Web API.
###############################################################################

###############################################################################
# Description:
#    vlunStartTest API Call. Start a new test to a given port, target or lun.
# Parameters:
#    Required:
#        port = physical port number being used
#        OR
#        lun_serial = serial number of lun to start test on
#    Optional:
#        host = SANBlaze hostname
#        target = target number to run test on
#        lun = lun number to run test on
#        test_type = (Read,Write,Compare,Verify)
#        threads = number of test threads to run
#        blocks = number of blocks per IO (-1 for random)
#        ios = number of IOs to do (0 for unlimited)
#        pass_ios = for compare tests the number of ios to write before reading back
#        seek_type = io access type
#            0 - Sequential
#            1 - Random
#            2 - Min/Max
#            3 - Butterfly
#        opcode = opcode to use (6,10,12,16, 0 (random))
#        paused = start test in paused sate (0,1)
#        initiator = initiator test is being run from (-1 for all)
#        pattern = pattern being used for writes
#            0 - random
#            1 - 0x00ff00ff
#            2 - 0x55aa55aa
#            3 - 8-bit incr
#            4 - 8-bit walking 1,0
#            5 - 0x0000ffff
#            6 - 0x5555aaaa
#            7 - 16-bit incr
#            8 - 16-bit walking 1,0
#            9 - 32-bit incr
#            -1 - custom
#        multipath_mode = if multiple paths to lun are detected, how to handle them
#            0 - default
#            1 - this path
#            2 - one path
#            3 - all paths
#            4 - act/opt paths
#            5 - act/non-opt paths
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_start_test {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => ''
    );
    my $target = &ParamsMgr::get_arg(
        'arg'      => "target",
        'required' => FALSE,
        'default'  => ''
    );
    my $lun = &ParamsMgr::get_arg(
        'arg'      => "lun",
        'required' => FALSE,
        'default'  => ''
    );
    my $lun_serial = &ParamsMgr::get_arg(
        'arg'      => "lun_serial",
        'required' => FALSE,
        'default'  => undef
    );
    my $test_type = &ParamsMgr::get_arg(
        'arg'      => "test_type",
        'required' => FALSE,
        'default'  => ''
    );
    my $threads = &ParamsMgr::get_arg(
        'arg'      => "threads",
        'required' => FALSE,
        'default'  => ''
    );
    my $blocks = &ParamsMgr::get_arg(
        'arg'      => "blocks",
        'required' => FALSE,
        'default'  => BLOCKS_RANDOM
    );
    my $ios = &ParamsMgr::get_arg(
        'arg'      => "ios",
        'required' => FALSE,
        'default'  => ''
    );
    my $pass_ios = &ParamsMgr::get_arg(
        'arg'      => "pass_ios",
        'required' => FALSE,
        'default'  => ''
    );
    my $seek_type = &ParamsMgr::get_arg(
        'arg'      => "seek_type",
        'required' => FALSE,
        'default'  => ''
    );
    my $opcode = &ParamsMgr::get_arg(
        'arg'      => "opcode",
        'required' => FALSE,
        'default'  => OPCODE_RANDOM
    );
    my $paused = &ParamsMgr::get_arg(
        'arg'      => "paused",
        'required' => FALSE,
        'default'  => ''
    );
    my $initiator = &ParamsMgr::get_arg(
        'arg'      => "initiator",
        'required' => FALSE,
        'default'  => ''
        );
    my $pattern = &ParamsMgr::get_arg(
        'arg'      => "pattern",
        'required' => FALSE,
        'default'  => PATTERN_RANDOM
    );
    my $multipath_mode = &ParamsMgr::get_arg(
        'arg'      => "multipath_mode",
        'required' => FALSE,
        'default'  => MULTIPATH_DEFAULT
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    if(defined($lun_serial)) {
        my $lun_search_result = $self->sanblaze_lun_search(
            host => $host,
            lun_serial => $lun_serial
        );
        if ($lun_search_result) {
            $port = $lun_search_result->{port};
            $target = $lun_search_result->{target};
            $lun = $lun_search_result->{lun};
        } else {
            $Log->debug("[ERROR] Lun Serial $lun_serial not Found (sanblaze_lun_status)");
            $Log->exit();
            return {};
        }
    }

    my %parameters = (
        port => $port,
        target => $target,
        lun => $lun,
        test_type => $test_type,
        threads => $threads,
        blocks => $blocks,
        ios => $ios,
        pass_ios => $pass_ios,
        seek_type => $seek_type,
        opcode => $opcode,
        paused => $paused,
        initiator => $initiator,
        pattern => $pattern,
        multipath_mode => $multipath_mode
    );
    # TEMPORARY UNTIL SANBLAZE XML API FIX
    # my $xml_string = $self->_build_xml(%parameters);
    my $xml_string = $self->_build_url(%parameters);

    $Log->debug("[INFO] Starting IO with XML:\n$xml_string");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunStartTest",
        data     => $xml_string,
        is_action => 1
    );

    if (($result == 0) || ($result->{test}->{status} != 0)) {
        $Log->error("[ERROR] Status Return - Fail");
        $result = {};
    } else {
        $Log->debug("[SUCCESS] Status Return - Success");
    }

    $Log->exit();
    return $result;
}

###############################################################################
# Description:
#    vlunStopTest API Call. Stop a currently running test.
# Parameters:
#    Required:
#        port = physical port number being used
#        OR
#        lun_serial = serial number of lun to stop test on
#    Optional:
#        host = SANBlaze hostname
#        target = target number to run test on
#        lun = lun number to run test on
#        test_id = test id of the test that is being stopped
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_stop_test {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => ''
    );
    my $target = &ParamsMgr::get_arg(
        'arg'      => "target",
        'required' => FALSE,
        'default'  => ''
    );
    my $lun = &ParamsMgr::get_arg(
        'arg'      => "lun",
        'required' => FALSE,
        'default'  => ''
    );
    my $lun_serial = &ParamsMgr::get_arg(
        'arg'      => "lun_serial",
        'required' => FALSE,
        'default'  => undef
    );
    my $test_id = &ParamsMgr::get_arg(
        'arg'      => "test_id",
        'required' => FALSE,
        'default'  => ALL_TEST_IDS
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    if(defined($lun_serial)) {
        my $lun_search_result = $self->sanblaze_lun_search(
            host => $host,
            lun_serial => $lun_serial
        );
        if ($lun_search_result) {
            $port = $lun_search_result->{port};
            $target = $lun_search_result->{target};
            $lun = $lun_search_result->{lun};
        } else {
            $Log->debug("[ERROR] Lun Serial $lun_serial not Found (sanblaze_lun_status)");
            $Log->exit();
            return {};
        }
    }

    my %parameters = (
        port => $port,
        target => $target,
        lun => $lun,
        test_id => $test_id
    );
    # TEMPORARY UNTIL SANBLAZE XML API FIX
    # my $xml_string = $self->_build_xml(%parameters);
    my $xml_string = $self->_build_url(%parameters);

    $Log->debug("[INFO] Stopping IO with XML:\n$xml_string");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunStopTest",
        data     => $xml_string,
        is_action => 1
    );
    if(($result == 0) || ($result->{status} != 0)) {
        $Log->debug("[ERROR] Status Return - Fail");
        $Log->exit();
        return {};
    } else {
        $Log->debug("[SUCCESS] Status Return - Success");
        $Log->exit();
        return $result;
    }
}

###############################################################################
# Description:
#    vlunLinkReset API Call. Link reset a particular port.
# Parameters:
#    Required:
#        port = physical port number being used
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_link_reset {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    my $xml_string = "<port>$port</port>";

    $Log->debug("[INFO] Sending Link Reset with XML:\n$xml_string");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunLinkReset",
        data     => $xml_string,
        is_action => 1
    );
    if($result->{status} == 0) {
        $Log->debug("[SUCCESS] Status Return - Success");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Status Return - Fail");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    vlunInitReprobe API Call. Reprobe for connected devices on a given port.
# Parameters:
#    Required:
#        port = physical port number being used
#    Optional:
#        host = SANBlaze hostname
#        probe_type = type of reprobe to do
#            1 - lun 0 only
#            2 - All luns
#            3 - Report luns
#        initiator = initiator to reprobe from
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_init_reprobe {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $probe_type = &ParamsMgr::get_arg(
        'arg'      => "probe_type",
        'required' => FALSE,
        'default'  => ALL_LUNS
    );
    my $initiator = &ParamsMgr::get_arg(
        'arg'      => "initiator",
        'required' => FALSE,
        'default'  => ALL_INITIATORS
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    my %parameters = (
        port => $port,
        probe_type => $probe_type,
        initiator => $initiator
    );
    my $xml_string = $self->_build_xml(%parameters);

    $Log->debug("[INFO] Initializing Reprobe with XML:\n$xml_string");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunInitReprobe",
        data     => $xml_string,
        is_action => 1
    );
    if($result->{status} == 0) {
        $Log->debug("[SUCCESS] Status Return - Success");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Status Return - Fail");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    vlunTaskManagement API Call. Send task management commands from a port
#    to a give target or lun.
# Parameters:
#    Required:
#        port = physical port number being used
#        OR
#        lun_serial = serial number of lun
#    Optional:
#        host = SANBlaze hostname
#        target = target id to send task management to
#        lun = lun id to send task management to
#        count = number of commands to abort (for AbortTask only)
#        command = type of task management to send
#            BusReset
#            TargetReset
#            LogicialUnitReset
#            ClearTaskSet
#            AbortTaskSet
#            AbortTask
#        probe_type = type of reprobe to do
#            1 - lun 0 only
#            2 - All luns
#            3 - Report luns
#        initiator = initiator to send command from from
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_task_management {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => FALSE,
        'default'  => ''
    );
    my $target = &ParamsMgr::get_arg(
        'arg'      => "target",
        'required' => FALSE,
        'default'  => ''
    );
    my $lun = &ParamsMgr::get_arg(
        'arg'      => "lun",
        'required' => FALSE,
        'default'  => ''
    );
    my $lun_serial = &ParamsMgr::get_arg(
        'arg'      => "lun_serial",
        'required' => FALSE,
        'default'  => undef
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => ''
    );
    my $command = &ParamsMgr::get_arg(
        'arg'      => "command",
        'required' => FALSE,
        'default'  => ''
    );
    my $probe_type = &ParamsMgr::get_arg(
        'arg'      => "probe_type",
        'required' => FALSE,
        'default'  => ALL_LUNS
    );
    my $initiators = &ParamsMgr::get_arg(
        'arg'      => "initiator",
        'required' => FALSE,
        'default'  => ALL_INITIATORS
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    if(defined($lun_serial)) {
        my $lun_search_result = $self->sanblaze_lun_search(
            host => $host,
            lun_serial => $lun_serial
        );
        if ($lun_search_result) {
            $port = $lun_search_result->{port};
            $target = $lun_search_result->{target};
            $lun = $lun_search_result->{lun};
        } else {
            $Log->debug("[ERROR] Lun Serial $lun_serial not Found (sanblaze_lun_status)");
            $Log->exit();
            return {};
        }
    }

    my %parameters = (
        port => $port,
        target => $target,
        lun => $lun,
        count => $count,
        command => $command,
        probe_type => $probe_type,
        initiators => $initiators
    );
    my $xml_string = $self->_build_xml(%parameters);

    $Log->debug("[INFO] Sending Task Management Command with XML:\n$xml_string");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunTaskManagement",
        data     => $xml_string,
        is_action => 1
    );
    if($result->{status} == 0) {
        $Log->debug("[SUCCESS] Status Return - Success");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Status Return - Fail");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    vlunErrorAction API Call. Inject certain error condition on a given
#    port. Depending on the error, the possible arguments can differ.
# Parameters:
#    Required:
#        port = physical port number being used
#        action = type of error injection to do. Action arguments can differ
#            BadDA - send frame with bad destination address
#                count - number of frames to corrupt
#                address - MAC address to inject (ie 00:11:22:33:44:55)
#            BadSA - send frame with source address
#                count - number of frames to corrupt
#                address - MAC address to inject
#            BadVLAN - send frame with bad vlan tag
#                count - number of frames to corrupt
#                priority - vlan priority to use ( 0 - 7 )
#                id - vlan id to use
#            BadType - send frame with bad ethernet type
#                count - number of frames to corrupt
#                eth_type - ethernet type to use (4 hex digits, 8906 for FCoE)
#            BadSOF - send frame with bad SOF
#                count - number of frames to corrupt
#                sof - sof to use
#            BadEOF - send frame with bad EOF
#                count - number of frames to corrupt
#                eof - eof to use
#            BadCRCOut - send frame with bad CRC
#                count - number of frames to corrupt
#            BadCRCIn - receive frame with bad CRC
#                count - number of frames to corrupt
#            BadLenUnder - send frame with underrun
#                count - number of frames to corrupt
#                length - length to shorten frame by
#            BadLenOver - send frame with overrun
#                count - number of frames to corrupt
#                length - length to extend frame by
#            FCoEDrop - drop a specific frame
#                count - number of frames to corrupt
#                seq_cnt - frame sequence count (FFFF for don't care)
#                offset - frame offset (FFFFFFFF for don't care)
#                type - type of frame to drop:
#                    1 - Don't send/recv FCP_XFER_RDY
#                    2 - Don't send FCP_DATA, any frame of seq
#                    3 - Don't recv FCP_DATA, any frame of seq
#                    4 - Don't send/recv FCP_RSP
#                    5 - Don't send/recv FCP_CONF
#                    6 - Don't send/recv FCP_CMND (not task mgmt)
#                    7 - Don't send FCP_DATA, last frame of seq
#                    8 - Don't recv FCP_DATA, last frame of seq
#                    9 - Don't send/recv FCP_CMND (task mgmt)
#                    10 - Don't send FCP_DATA, first frame of seq
#                    11 - Don't recv FCP_DATA, first frame of seq
#                    12 - Don't send FCP_DATA, middle frame of seq
#                    13 - Don't recv FCP_DATA, middle frame of seq
#                    14 - Don't recv SRR
#                    15 - Don't recv ADISC
#                    16 - Don't recv FDISC
#                    17 - Ignore recv'd FCP_CMND (not task mgmt)
#                    18 - Reject recv'd FCP_CMND (not task mgmt)
#                    19 - Don't send FCP_DATA, specific SeqCnt/Offset
#                    20 - Don't recv FCP_DATA, specific SeqCnt/Offset
#            FCoEDelay - delay a specific frame
#                count - number of frames to corrupt
#                seq_cnt - frame sequence count (FFFF for don't care)
#                offset - frame offset (FFFFFFFF for don't care)
#                delay - amout to delay in milliseconds
#                type - type of delay to inject:
#                    1 - Delay sending FCP_XFER_RDY
#                    2 - Delay sending FCP_DATA, any frame of seq
#                    4 - Delay sending FCP_RSP
#                    5 - Delay sending FCP_CONF
#                    7 - Delay sending FCP_DATA, last frame of seq
#                    10 - Delay sending FCP_DATA, first frame of seq
#                    12 - Delay sending FCP_DATA, middle frame of seq
#                    19 - Delay sending FCP_DATA, specific SeqCnt/Offset
#            FIPDrop - drop a specific FIP frame
#                count - number of frames to drop
#                port_id - specific port_id to use
#                type - type of frame to drop
#                    1 - Drop all Keep Alives
#                    2 - Drop controller Keep Alives
#                    3 - Drop port Keep Alives
#                    4 - Drop port Keep Alives for a specific PortID
#            SendPause - send a standard pause frame
#                count - number of frames to send
#                interval - interval between frames in milliseconds
#                quanta - time quanta to use in frames (0 -ffff)
#            SendPFCPause - send a PFC pause frame
#                count - number of frames to send
#                interval - interval between frames in milliseconds
#                quanta - time quanta to use in frames ( 0 - ffff)
#                class - priority class to pause (0 - 7 )
#            ForceBadDIFGuardIn - insert bad t10 DIF guard tag to incoming data
#                count - number of frames to inject
#            ForceBadDIFGuardOut - insert bad t10 DIF guard to outgoing data
#                count - number of frames to inject
#            ForceBadDIFRefTagIn - insert bad t10 DIF ref tag to incoming data
#                count - number of frames to inject
#            ForceBadDIFRefTagOut - insert bad t10 DIF ref tag to outgoing data
#                count - number of frames to inject
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_error_action {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $action = &ParamsMgr::get_arg(
        'arg'      => "action",
        'required' => TRUE
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => ''
    );
    my $address = &ParamsMgr::get_arg(
        'arg'      => "address",
        'required' => FALSE,
        'default'  => ''
    );
    my $priority = &ParamsMgr::get_arg(
        'arg'      => "priority",
        'required' => FALSE,
        'default'  => ''
    );
    my $id = &ParamsMgr::get_arg(
        'arg'      => "id",
        'required' => FALSE,
        'default'  => ''
    );
    my $eth_type = &ParamsMgr::get_arg(
        'arg'      => "eth_type",
        'required' => FALSE,
        'default'  => ''
    );
    my $sof = &ParamsMgr::get_arg(
        'arg'      => "sof",
        'required' => FALSE,
        'default'  => ''
    );
    my $length = &ParamsMgr::get_arg(
        'arg'      => "length",
        'required' => FALSE,
        'default'  => ''
    );
    my $seq_cnt = &ParamsMgr::get_arg(
        'arg'      => "seq_cnt",
        'required' => FALSE,
        'default'  => ''
    );
    my $offset = &ParamsMgr::get_arg(
        'arg'      => "offset",
        'required' => FALSE,
        'default'  => ''
    );
    my $type = &ParamsMgr::get_arg(
        'arg'      => "type",
        'required' => FALSE,
        'default'  => ''
    );
    my $delay = &ParamsMgr::get_arg(
        'arg'      => "delay",
        'required' => FALSE,
        'default'  => ''
    );
    my $port_id = &ParamsMgr::get_arg(
        'arg'      => "port_id",
        'required' => FALSE,
        'default'  => ''
    );
    my $interval = &ParamsMgr::get_arg(
        'arg'      => "interval",
        'required' => FALSE,
        'default'  => ''
    );
    my $quanta = &ParamsMgr::get_arg(
        'arg'      => "quanta",
        'required' => FALSE,
        'default'  => ''
    );
    my $class = &ParamsMgr::get_arg(
        'arg'      => "class",
        'required' => FALSE,
        'default'  => ''
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    my %parameters = (
        port => $port,
        action => $action,
        count => $count,
        address => $address,
        priority => $priority,
        id => $id,
        eth_type => $eth_type,
        sof => $sof,
        length => $length,
        seq_cnt => $seq_cnt,
        offset => $offset,
        type => $type,
        delay => $delay,
        port_id => $port_id,
        interval => $interval,
        quanta => $quanta,
        class => $class
    );
    my $xml_string = $self->_build_xml(%parameters);

    $Log->debug("[INFO] Sending On Error Action with XML:\n$xml_string");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunErrorAction",
        data     => $xml_string,
        is_action => 1
    );
    if($result->{status} == 0) {
        $Log->debug("[SUCCESS] Status Return - Success");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Status Return - Fail");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    Saves the port config of the given port to the given destination file.
# Parameters:
#    Required:
#        port = Sanblaze port number
#        target_xml = path to target XML file where config will be saved
#    Optional:
#        host = hostname for Sanblaze
# Return Value:
#    TRUE if success or FALSE if error
###############################################################################
sub sanblaze_save_config {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $target_xml = &ParamsMgr::get_arg(
        'arg'      => "target_xml",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }
    $Log->debug("[INFO] Saving Config from Port $port to $target_xml @ $host");
    my $port_config = $self->sanblaze_get_port_config(
        host => $host,
        port => $port
    );
    if($port_config) {
        my $config_str;
        $self->_for_hash($port_config, sub {
            my ($key, $value) = @_;
            $config_str .= "<$key>$value</$key>\n";
        });
        open(XML, ">$target_xml") or ($Log->debug("[ERROR] Failed to Open File $target_xml for Write. Returning...")
            and return FALSE);
        print XML $config_str;
        close XML;
        $Log->debug("[SUCCESS] Config Saved");
        $Log->exit();
        return TRUE;
    } else {
        $Log->debug("[ERROR] Failed to Get Port Config. Returning...");
        $Log->exit();
        return FALSE;
    }
}

###############################################################################
# Description:
#    Sets port config based on given XML file.
# Parameters:
#    Required:
#        port = Sanblaze port number
#        source_xml = path to XML file containing config
#    Optional:
#        host = hostname for Sanblaze
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_load_config {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $source_xml = &ParamsMgr::get_arg(
        'arg'      => "source_xml",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    $Log->debug("[INFO] Loading Config from $source_xml onto Port $port @ $host");
    open(XML, $source_xml) or ($Log->debug("[ERROR] Failed to Find File at $source_xml. Returning...")
        and return FALSE);
    my $xml = join("", <XML>);
    close XML;

    #If the xml string has a port tag, override it with the
    #port parameter given. If tag doesn't exist, add it in.
    if($xml =~ s/<port>\d<\/port>/<port>$port<\/port>/g) {
    } else {
        $xml = "<port>$port</port>\n$xml";
    }

    my $result = $self->sanblaze_set_port_config(
        host => $host,
        xml  => $xml
    );
    if($result) {
        $Log->debug("[SUCCESS] Config File Loaded");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Failed to Load Config File. Returning...");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    vlunSetPortConfig API Call. Change port level settings. Only arguments
#    that need to be changed need to be passed in. See API manual for XML
#    that can be passed in.
# Parameters:
#    Required:
#        xml = string of XML to send to web API
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
#
# NOTE:
#    This method requires actual XML string to be passed due to the large
#    number (30+) of possible parameters that can be passed to it. See
#    SANBlaze API manual for the XML format.
###############################################################################
sub sanblaze_set_port_config {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $xml = &ParamsMgr::get_arg(
        'arg'      => "xml",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Setting Port Config with XML:\n$xml");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunSetPortConfig",
        data     => $xml,
        is_action => 1
    );
    if($result->{status} == 0) {
        $Log->debug("[SUCCESS] Status Return - Success");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Status Return - Fail");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    Changes number of initiators on given port using vlunSetPortConfig
#    API Call.
# Parameters:
#    Required:
#        port = physical port number being used
#        initiators = number of initiators to set
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_set_initiators {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $initiators = &ParamsMgr::get_arg(
        'arg'      => "initiators",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Setting Number of Initiators on Port $port to $initiators @ $host");
    my $result = $self->sanblaze_set_port_config (
        host     => $host,
        xml      => "<port>$port</port>\n<num_initiators>$initiators</num_initiators>"
    );
    if($result) {
        $Log->debug("[SUCCESS] Number of Initiators Set");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Failed to Set Initiators. Returning...");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    Changes state of given port using vlunSetPortConfig API Call.
# Parameters:
#    Required:
#        port = physical port number being used
#        state = "Online" or "Offline"
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_set_state {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $state = &ParamsMgr::get_arg(
        'arg'      => "state",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Setting State of Port $port to $state @ $host");
    my $result = $self->sanblaze_set_port_config (
        host     => $host,
        xml      => "<port>$port</port>\n<state>$state</state>"
    );
    if($result) {
        $Log->debug("[SUCCESS] State Set");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Failed to Set State. Returning...");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    Changes mode of given port using vlunSetPortConfig API Call.
# Parameters:
#    Required:
#        port = physical port number being used
#        mode = "Targ" or "Init"
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_set_mode {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $mode = &ParamsMgr::get_arg(
        'arg'      => "mode",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Setting Mode of Port $port to $mode @ $host");
    my $result = $self->sanblaze_set_port_config (
        host     => $host,
        xml      => "<port>$port</port>\n<mode>$mode</mode>"
    );
    if($result) {
        $Log->debug("[SUCCESS] Mode Set");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Failed to Set Mode. Returning...");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    Changes base wwnn to seed all initiators from of given port using
#    vlunSetPortConfig API Call.
# Parameters:
#    Required:
#        port = physical port number being used
#        wwnn
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
#
# NOTE:
#    FC Only
###############################################################################
sub sanblaze_set_base_wwnn {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $wwnn = &ParamsMgr::get_arg(
        'arg'      => "wwnn",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Setting Base WWNN of Port $port to $wwnn @ $host");
    my $result = $self->sanblaze_set_port_config (
        host     => $host,
        xml      => "<port>$port</port>\n<base_wwnn>$wwnn</base_wwnn>"
    );
    if($result) {
        $Log->debug("[SUCCESS] Base WWNN Set");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Failed to Set Base WWNN. Returning...");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    Changes base wwpn to seed all initiators from of given port using
#    vlunSetPortConfig API Call.
# Parameters:
#    Required:
#        port = physical port number being used
#        wwpn
#    Optional:
#        host = SANBlaze hostname
# Return Value:
#    Hash_ref
#
# NOTE:
#    FC Only
###############################################################################
sub sanblaze_set_base_wwpn {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $wwpn = &ParamsMgr::get_arg(
        'arg'      => "wwpn",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }
    $Log->debug("[INFO] Setting Base WWPN of Port $port to $wwpn @ $host");
    my $result = $self->sanblaze_set_port_config (
        host     => $host,
        xml      => "<port>$port</port>\n<base_wwpn>$wwpn</base_wwpn>"
    );
    if($result) {
        $Log->debug("[SUCCESS] Base WWPN Set");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Failed to Set Base WWPN. Returning...");
        $Log->exit();
        return {};
    }
}

###############################################################################
# Description:
#    vlunSetPortTestParameters API Call. Change port level test parameters.
#    Only arguments that need to be changed need to be passed in.
# Parameters:
#    Required:
#        port = physical port number being used
#    Optional:
#        host = SANBlaze hostname
#        general_timeout = timeout for general commands
#        rw_timeout = timeout for read/write commands
#        taskmgmt_timeout = timeout for task management commands
#        busy_delay = delay time before retrying busy commands
#        onerror_action = action to take when error is encountered
#            Continue - ignore error and keep test running
#            StopThread - stop current test thread
#            StopTest - stop all threads under this test
#            StopAllTests - stop all tests
#        retry_count = number of retries to attempt before logging failure
# Return Value:
#    Hash_ref
###############################################################################
sub sanblaze_set_port_test_parameters {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => FALSE,
        'default'  => $self->{host}
    );
    my $port = &ParamsMgr::get_arg(
        'arg'      => "port",
        'required' => TRUE
    );
    my $general_timeout = &ParamsMgr::get_arg(
        'arg'      => "general_timeout",
        'required' => FALSE,
        'default'  => ''
    );
    my $rw_timeout = &ParamsMgr::get_arg(
        'arg'      => "rw_timeout",
        'required' => FALSE,
        'default'  => ''
    );
    my $taskmgmt_timeout = &ParamsMgr::get_arg(
        'arg'      => "taskmgmt_timeout",
        'required' => FALSE,
        'default'  => ''
    );
    my $busy_delay = &ParamsMgr::get_arg(
        'arg'      => "busy_delay",
        'required' => FALSE,
        'default'  => ''
    );
    my $onerror_action = &ParamsMgr::get_arg(
        'arg'      => "onerror_action",
        'required' => FALSE,
        'default'  => ''
    );
    my $retry_count = &ParamsMgr::get_arg(
        'arg'      => "retry_count",
        'required' => FALSE,
        'default'  => ''
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    my %parameters = (
        port => $port,
        general_timeout => $general_timeout,
        rw_timeout => $rw_timeout,
        taskmgmt_timeout => $taskmgmt_timeout,
        busy_delay => $busy_delay,
        onerror_action => $onerror_action,
        retry_count => $retry_count
    );
    my $xml_string = $self->_build_xml(%parameters);

    $Log->debug("[INFO] Setting Port Test Parameters with XML:\n$xml_string");
    my $result = $self->_call (
        host     => $host,
        method   => "vlunSetPortTestParameters",
        data     => $xml_string,
        is_action => 1
    );
    if($result->{status} == 0) {
        $Log->debug("[SUCCESS] Status Return - Success");
        $Log->exit();
        return $result;
    } else {
        $Log->debug("[ERROR] Status Return - Fail");
        $Log->exit();
        return {};
    }
}


###############################################################################
# Private Methods
###############################################################################

###############################################################################
# Description:
#    Private method used to send the HTTP request to the SANBlaze Web API.
# Parameters:
#    Required:
#        host, method
#    Optional:
#        data
# Return Value:
#    Hash_ref
###############################################################################
sub _call {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $host = &ParamsMgr::get_arg(
        'arg'      => "host",
        'required' => TRUE
        );
    my $method = &ParamsMgr::get_arg(
        'arg'      => "method",
        'required' => TRUE
        );
    my $data = &ParamsMgr::get_arg(
        'arg'      => "data",
        'required' => FALSE,
        'default'  => undef
        );
    #Sanblaze returns malformed XML when sending action commands.
    #This is resolved by fixing XML before it gets parsed into hash.
    my $is_action = &ParamsMgr::get_arg(
        'arg'      => "is_action",
        'required' => FALSE,
        'default'  => undef
        );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    # Create a user agent object
    my $ua = new LWP::UserAgent;
    $ua->agent("AgentName/0.1 $ua->agent");

    # Create a request
    my $req;
    if (defined($data)) {
        $req = new HTTP::Request POST => "$host/goform/XmlApi?op=$method";
        $req->content($data);
    } else {
        $req = new HTTP::Request GET => "$host/goform/XmlApi?op=$method";
    }

    $Log->debug($req->as_string);
    # Pass request to the user agent and get a response back
    my $res = $ua->request($req);

    # Check the outcome of the response
    if ($res->is_success) {
        $Log->debug("[SUCCESS] $method - HTTP Request Successful");
        if(defined($is_action)) {
            return XMLin("<xml>".$res->content."</xml>");
        } else {
            if ( length($res->content) < length("<?xml version='1.0' ?>") ) {
                $Log->comment("[WARNING] $method - HTTP Return Malformed");
                return {};
            }
            return XMLin($res->content);
        }
    } else {
        $Log->error("[ERROR] $method - HTTP Request Failed: ".$res->status_line);
        return {};
    }
}

###############################################################################
# Description:
#    Private method used to build string of XML based on parameter hash given.
# Parameters:
#    Required:
#        parameters
# Return Value:
#    String
###############################################################################
sub _build_xml {
    my ($self, %parameters) = @_;

    my $xml_string = "";
    for my $key ( keys %parameters ) {
        my $value = "";
        $value = $parameters{$key};
        if ($value ne '') {
            $xml_string = "$xml_string<$key>$value</$key>\n";
        }
    }
    return $xml_string;
}

###############################################################################
# Description:
#    Private method used to build string of URL based on parameter hash given.
# Parameters:
#    Required:
#        parameters
# Return Value:
#    String
###############################################################################
sub _build_url {
    my ($self, %parameters) = @_;

    my $url_string = "";
    for my $key ( keys %parameters ) {
        my $value = "";
        $value = $parameters{$key};
        if ($value ne '') {
            $url_string = "$url_string&$key=$value";
        }
    }
    return $url_string;
}

###############################################################################
# Description:
#    Private method used to iterate through hash and execute a function.
# Parameters:
#    Required:
#        hash, fn
# Return Value:
#    None
###############################################################################
sub _for_hash {
    my ($self, $hash, $fn) = @_;
    while (my ($key, $value) = each %$hash) {
        if ('HASH' eq ref $value) {
            $self->for_hash($value, $fn);
        }
        else {
            $fn->($key, $value);
        }
    }
}

1;
