# Copyright (c) 2013-2020 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary  Definition of all I/O presets used by MISFIT tests
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Services::IO::Workload

use strict;

# Package Name
package San::FIT::Misfit::Services::IO::Workload;

# Module imports
use Tharn;
use ParamsMgr;
use NATE::Util::DataTypes qw(TRUE FALSE);

# Constants
use constant {
};

# Globals
use vars qw (
    $Workloads
);

#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
#                             Workloads
#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#

$Workloads = {
    log => {
        description => "Highly sequential pattern, consisting of small writes",
        dt          => {
            noprogt    => 15,
            noprogtt   => 130,
            alarm      => 3,
            slices     => 3,
            iotype     => "sequential",
            pattern    => "iot",
            bs         => "8k",
            dsize      => "4k",
            prefix     => '%d@%h',
            enable     => "microdelay,timestamp,syslog,raw",
            rdelay     => "500000",
            wdelay     => "500000",
            oncerr     => "abort",
            keepalivet => "60"
        },
        sanblaze   => {
            pattern        => "1",     # 1 = 0x00FF00FF, pattern for writes
            seek_type      => "0",     # 0 = sequential
            test_type      => "Write",
            blocks         => "8",
            multipath_mode => "4",     # 3 = All paths
        },
    }, # End log
    ms_exchange => {
        description  => "Simulate Exchange's database file. Random reads",
        dt           => {
            noprogt    => 15,
            noprogtt   => 130,
            alarm      => 3,
            slices     => 3,
            iotype     => "random",
            pattern    => "iot",
            bs         => "32k",
            dsize      => "4k",
            prefix     => '%d@%h',
            enable     => "microdelay,timestamp,syslog,raw",
            rdelay     => "500000",
            wdelay     => "500000",
            oncerr     => "abort",
            keepalivet => "60"
        },
        sanblaze   => {
            seek_type      => "1",    # 1 = Random
            test_type      => "Read",
            blocks         => "32",
            multipath_mode => "4",    # 3 = All paths
        },
    }, # End ms_exchange
    oltp => {
        description => "Online Transaction Processing (OLTP). DB workload",
        dt          => {
            noprogt    => 15,
            noprogtt   => 130,
            alarm      => 3,
            slices     => 3,
            iotype     => "sequential",
            pattern    => "iot",
            bs         => "8k",
            dsize      => "4k",
            prefix     => '%d@%h',
            enable     => "microdelay,timestamp,syslog,raw",
            rdelay     => "500000",
            wdelay     => "500000",
            oncerr     => "abort",
            keepalivet => "60"
        },
        sanblaze   => {
            pattern        => "7",     # 7 = 16-bit increment for writes
            seek_type      => "0",     # 0 = sequential
            test_type      => "Write",
            blocks         => "8",
            multipath_mode => "4",     # 3 = All paths
        },
    }, # End oltp
    dss => {
        description  => "Larger I/O block size. Data warehouse workload",
        dt           => {
            noprogt    => 15,
            noprogtt   => 130,
            alarm      => 3,
            slices     => 3,
            iotype     => "random",
            pattern    => "iot",
            bs         => "32k",
            dsize      => "4k",
            prefix     => '%d@%h',
            enable     => "microdelay,timestamp,syslog,raw",
            rdelay     => "500000",
            wdelay     => "500000",
            oncerr     => "abort",
            keepalivet => "60"
        },
        sanblaze   => {
            pattern        => "0",     # 0 = Random pattern for writes
            seek_type      => "0",     # 0 = sequential
            test_type      => "Read",
            blocks         => "512",
            multipath_mode => "4",     # 3 = All paths
        },
    }, # End dss
};

#===============================================================================
# Function: get_workload
#   Returns a MISFIT IO Workload.
#
#   Parameters:
#       id - The id corresponding to a MISFIT preset.
#
#   Returns:
#       A hash reference containing an I/O workload or undef if the id
#       is not found.
#
#   For example:
#
#   dss => {
#        description  => "Larger I/O block size. Data warehouse workload",
#        dt           => {
#            write      => 1,
#            noprogt    => 15,
#            noprogtt   => 130,
#            alarm      => 3,
#            slices     => 3,
#            iotype     => "random",
#            pattern    => "iot",
#            bs         => "32k",
#            dsize      => "4k",
#            # prefix     => '%d@%h',
#            enable     => "microdelay,timestamp,syslog,raw",
#            rdelay     => "500000",
#            wdelay     => "500000",
#            oncerr     => "abort",
#            keepalivet => "60"
#        },
#        sanblaze   => {
#            pattern        => "0",     # 0 = Random pattern for writes
#            seek_type      => "0",     # 0 = sequential
#            test_type      => "Read",
#            blocks         => "512",
#            multipath_mode => "3",     # 3 = All paths
#        }
#===============================================================================
sub get_workload {
    my (%args) = @_;
    &ParamsMgr::process_args(%args);
    my $id = &ParamsMgr::get_arg(
        'arg'      => "id",
        'required' => FALSE,
        'default'  => "",
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return undef;
    }

    # Locals
    my $workload = undef;

    # Validate params
    if ($id eq "") {
        my @ids = keys(%$Workloads);
        $id = $ids[int(rand(@ids))];
        $Log->comment("No specific id requested, selecting a random id: $id");
    }

    if (defined($Workloads->{$id})) {
        $workload = $Workloads->{$id};
    } else {
        $Log->comment(
            "There is no Workload ID defined for the requested ID $id"
        );
    }

    return $workload;
} # End get_workload

1;
