# Copyright (c) 2013-2020 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Module responsible for handling Initiator operations and target
##     discovery services
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Services::InitiatorOps
# Library to perform Initiator and I/O services using Misfit messaging

package San::FIT::Misfit::Services::InitiatorOps;

# Compiler Directives
use strict;
use warnings;

# Nate Libs
use Tharn;
use NATE::Process;
use NATE::ParamSet;
use NATE::ProcManager;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);

# Test Libs
use ParamsMgr;
use NACL::C::Node;
use San::SanClient;
use NACL::C::Client;
use NACL::C::Cluster;
use NACL::CS::LunIgroup;
use San::FIT::Misfit::Core::Workflow;

# Perl Libs
use DBI;
use threads;
use Data::Dumper;
use List::MoreUtils qw(any);

# Inherit Misfit base class
use base qw( San::FIT::Misfit::Core::Base );

use constant {
    ID                     => "San::FIT::Misfit::Services::InitiatorOps",
    HEADER                 => "",
    DB_FILE                => "discovery.sqlite",
    DB_TABLE               => "devices",
    LOCAL_HOST             => 'localhost',
    SEVERITY_NA            => 2,
    CLIENTS_FILE           => "CLIENTS.txt",
    CLIENT_WORKER		   => 'lib/San/FIT/Misfit/Services/IO/misfit_io.thpl',
    NATE_TEST_CMD          => "hostname",
    START_STAF_WIN         => 'START X:\\Staf\\netapp_staf.bat',
    SANBLAZE_WORKER        => 'lib/San/FIT/Misfit/Services/IO/sanblaze.thpl',
    LOG_MNT_WIN_RTP        =>
        'net use \\\\rtptestbeds.rtp.netapp.com\\rtpeng_testbeds netapp1! /USER:minwin',
    HOME_MNT_WIN_RTP       =>
        'net use \\\\rtpusers.eng.netapp.com\\HOME qa4testing /USER:rtpusers\\minwin',
    DB_CREATE_COLUMNS      => [
        "lun_serial UNIQUE", "lun", "volume", "vserver", "ontap_path",
        "hostname", "hosttype", "mount_point", "device_name", "protocol"
    ],
    DB_SELECT_COLUMNS      => [
        "lun_serial", "lun", "volume", "vserver", "ontap_path",
        "hostname", "hosttype", "mount_point", "device_name", "protocol"
    ],
    NATE_READY_RETRY_LIMIT => 2,
    IGROUP_INITIATOR_FILTER_LIMIT => 10,
};

#===============================================================================
# Constructor: new
#   Initialize attributes needed for initiator discovery and operations. These
#   parameters will have to be either passed to the scripts or defined in
#   the testbed configuration files.
#
#   Parameters:
#       discover - If TRUE, perform an initial client/target discovery.
#           Please note that the time required for this process to complete is
#           proportional to the size of the configuration. Valid values: 1|0.
#           Defaults to 0.
#       initialize - If TRUE, initializes processes on all initiators,
#           starts and verifies local services, prepares clients to be ready for
#           dispatching I/O requests. Valid values: 1|0, Defaults to 0.
#
#   Returns:
#       An InitiatorOps object.
#===============================================================================
sub new {
    my ($class, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $discover = &ParamsMgr::get_arg(
        'arg'      => "discover",
        'required' => FALSE,
        'default'  => FALSE,
    );
    my $initialize = &ParamsMgr::get_arg(
        'arg'      => "initialize",
        'required' => FALSE,
        'default'  => FALSE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return {};
    }

    my $self = bless {}, $class;
    $self = $class->SUPER::new();
    $self->{id}  = ID;

    # Get global parameters and initialize attributes
    my $clients = "";
    $clients = [split(/:/, param('CLIENTS', '-default', ''))];
    if (!scalar(@{$clients})) {
        $clients = $self->_get_clients_from_text_file();
    }

    $self->{clients} = $clients;
    $Log->comment("Client list: " . join(', ', @{$self->{clients}}));
    $self->{cluster_name} = param('CLUSTER_NAME', '-default', '');

    # Need to see if this is needed here...
    $self->{dbh} = _database_handle();
    unless (defined($self->{dbh})) {
        $Log->comment(
            "Failed connecting to the DB responsible for storing initiator " .
            "and target details. Make sure " . DB_FILE .
            " exists (SQLite)"
        );
        return {};
    }

    $self->initialize() if $initialize;
    $self->discover_devices() if $discover;

    return $self;
} # End of new

#===============================================================================
# Function: initialize
#   Runs a quick verification process on each client to ensure NATE and all the
#   'good' stuff required for NetApp's automation infrastructure to work, are
#   in place.
#
#   Parameters:
#       clients - An array reference containing a list of clients. It defaults
#       to the ones included in CLIENTS.txt. This is an optional parameter.
#
#   Returns:
#       An InitiatorOps object.
#
#   Example:
#   > my @ready = $InitiatorOps->initialize(
#   >     clients => ['client1','client2']
#   > );
#===============================================================================
sub initialize {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $clients = &ParamsMgr::get_arg(
        'arg'      => "clients",
        'required' => FALSE,
        'default'  => $self->{clients},
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @ready = $self->get_nate_ready_clients(clients => $clients);
    $self->{nate_ready_clients} = \@ready;

    return TRUE;
} # End of initialize

#===============================================================================
# Function: start_staf
#   This function will attempt to execute the batch file that STAF on the list
#   of clients provided.
#
#   Parameters:
#       clients - arrayref of client names (Default: CLIENTS testbed param)
#
#   Returns:
#     Array or arrayref of clients that started STAF. Empty array or arrayref
#     if None were able to start STAF.
#
#   Example:
#   > my @ready = InitiatorOps->start_staf(
#   >     client => ['client1','client2']
#   > );
#===============================================================================
sub start_staf {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $clients = &ParamsMgr::get_arg(
        'arg'      => "clients",
        'required' => FALSE,
        'default'  => $self->{clients},
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return wantarray ? () : [];
    }

    my @started = ();
    my @procs = ();

    foreach my $client (@{$clients}) {
        next if ($client =~ /sanblaze/);
        push (@procs, {
            codespec   => "{" .
                "use " . __PACKAGE__ . ";" .
                __PACKAGE__ . "::_start_staf();" .
            "}",
            params     => { client => $client },
            runid      => $client,
            background => TRUE,
            onexit     => sub {
                my $proc = shift();
                my $result = $proc->own_worst_result();
                if (!defined($result)) {
                    push(@started, $client);
                }
            },
        });
    }

    my $procman = NATE::ProcManager->new(
        runid       => "start_staf",
        background  => FALSE,
        proc_info   => \@procs,
    );
    $procman->start();
    $procman->wait();

    return wantarray ? @started : \@started;
} # End of start_staf

#===============================================================================
# Function: setup_nate_mounts
#   This function will attempt to mount the windows logging share for NATE as
#   well as the homedir directory for faster and more reliable access to tests
#   and test libs.  Currently, defaults to http for test and testlib.
#
#   Parameters:
#       clients - arrayref of client names (Default: CLIENTS testbed param)
#
#   Returns:
#     Array or arrayref of clients that started STAF. Empty array or arrayref
#     if None are NATE ready.
#
#   Example:
#   > my @ready = $InitiatorOps->setup_nate_mounts(
#   >     client => ['client1','client2']
#   > );
#===============================================================================
sub setup_nate_mounts {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $clients = &ParamsMgr::get_arg(
        'arg'      => "clients",
        'required' => FALSE,
        'default'  => $self->{clients},
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return wantarray ? () : [];
    }

    my @procs = ();
    my @mounted = ();
    foreach my $client (@{$clients}) {
        next if $client =~ /sanblaze/i;
        push (@procs, {
            codespec   => "{" .
                "use " . __PACKAGE__ . ";" .
                __PACKAGE__ . "::_mount_required_nate_shares();" .
            "}",
            runid      => $client,
            params     => { client => $client},
            background => TRUE,
            onexit     => sub {
                my $proc = shift();
                my $result = $proc->own_worst_result();
                if (!defined($result)) {
                    push(@mounted, $client);
                }
            },
        });
    }

    my $procman = NATE::ProcManager->new(
        runid       => "setup_nate_mounts",
        background  => FALSE,
        proc_info   => \@procs,
    );
    $procman->start();
    $procman->wait();
    $Log->comment("Mounted properly on: " . join(",",@mounted));

    return wantarray ? @mounted : \@mounted;
} # End of setup_nate_mounts

#===============================================================================
# Function: start_workers
#   This method will attempt to start a single instance of 'worker' on each
#   client.
#
#   Parameters:
#       clients - arrayref of client names (Default: CLIENTS testbed param)
#       worker  - Any executable code that will run on the request client
#           can be .thpl, .exe, codespec, etc...
#       args_to_worker - A hash reference containg optional params to the
#           worker thpls.
#
#   Returns:
#     Hashref of SanClient Objects keyed by hostname.
#
#   Example:
#   > InitiatorOps->start_workers(
#   >     client => [$client],
#   >     worker => "{print \"Hello $client\";}",
#   > );
#===============================================================================
sub start_workers {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $clients = &ParamsMgr::get_arg(
        'arg'      => "clients",
        'required' => FALSE,
        'default'  => $self->{clients},
    );
    my $worker = &ParamsMgr::get_arg(
        'arg'      => "worker",
        'required' => FALSE,
        'default'  => CLIENT_WORKER,
    );
    my $sanblaze_worker = &ParamsMgr::get_arg(
        'arg'      => "sanblaze_worker",
        'required' => FALSE,
        'default'  => SANBLAZE_WORKER,
    );
    my $args_to_worker = &ParamsMgr::get_arg(
        'arg'      => "args_to_worker",
        'required' => FALSE,
        'default'  => {},
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @workers = ();
    foreach my $client (@{$clients}) {
        $Log->comment("Creating process for $client");
        my $worker_info = {};
        my $runid = "";
        $runid = $client . "_worker";
        $worker_info = {
            runid       => $runid,
            background  => TRUE,
            inherit_env => TRUE,
            inherit_params => TRUE,
            params      => $args_to_worker,
        };
        $Log->comment(Dumper($worker_info));
        # SANBlaze workers need to run locally for now, since STAF is not
        # running on those workers.
        if ($client =~ /sanblaze/i) {
            $worker_info->{codespec} = $sanblaze_worker;
            $worker_info->{client}   = LOCAL_HOST;
        } else {
            $worker_info->{codespec} = $worker;
            $worker_info->{client}   = $client;
        }
        push (@workers, $worker_info);
    }

    my $procman = NATE::ProcManager->new(
        runid       => "initiator_manager",
        proc_info   => \@workers,
        onexit      => \&NATE::Process::on_exit_propagate_worst_result,
    );

    $procman->start();

    return $procman;
} # End of start_workers

#===============================================================================
# Function: start_workers
#   This method will attempt to start a single instance of 'worker' on each
#   client.
#
#   Parameters:
#       clients - arrayref of client names (Default: CLIENTS testbed param)
#       worker  - Any executable code that will run on the request client
#           can be .thpl, .exe, codespec, etc...
#       args_to_worker - A hash reference containg optional params to the
#           worker thpls.
#
#   Returns:
#     Hashref of SanClient Objects keyed by hostname.
#
#   Example:
#   > InitiatorOps->start_workers(
#   >     client => [$client],
#   >     worker => "{print \"Hello $client\";}",
#   > );
#===============================================================================
sub stop_workers {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $clients = &ParamsMgr::get_arg(
        'arg'      => "clients",
        'required' => FALSE,
        'default'  => $self->{clients},
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my @workers = ();
    foreach my $client (@{$clients}) {
        $Log->comment("Stopping worker process for $client");
        my $routing_key = $client;
        if ($routing_key =~ /sanblaze/) {
            $routing_key = 'sanblaze';
        }
        $self->local_message(
            exchange => 'io_exchange',
            routing_key => $routing_key,
            message => {msg => '__bye__'},
        );
    }

    my $procman = NATE::ProcManager->new(
        runid       => "initiator_manager",
        proc_info   => \@workers,
        onexit      => \&NATE::Process::on_exit_propagate_worst_result,
    );

    return TRUE;
} # End of stop_workers

#===============================================================================
# Function: get_nate_ready_clients
#   This method will check if hosts are NATE ready by attempting to start a
#   remote process on each host
#
#   Parameters:
#       clients - An arrayref of client names (Default: CLIENTS testbed param)
#
#   Returns:
#     Array or arrayref of clients that ARE NATE ready. Empty array or arrayref
#     if None are NATE ready.
#
#   Example:
#   > my @ready = $InitiatorOps->get_nate_ready_clients(
#   >     clients => ['client1','client2']
#   > );
#===============================================================================
sub get_nate_ready_clients {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $clients = &ParamsMgr::get_arg(
        'arg'      => "clients",
        'required' => FALSE,
        'default'  => $self->{clients},
    );
    my $retry = &ParamsMgr::get_arg(
        'arg'      => "retry",
        'required' => FALSE,
        'default'  => 0,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return wantarray ? (): [];
    }

    # Recursive method. 'retry' will be increased and passed to its recursive
    # call
    if ($retry > NATE_READY_RETRY_LIMIT) {
        return wantarray ? (): [];
    } elsif ($retry > 0) {
        $Log->comment("Attempting retry: $retry to make clients nate ready");
    }
    $retry++;

    my $cmd          = NATE_TEST_CMD;
    my %procs        = ();
    my $status       = TRUE;
    my @no_staf      = ();
    my @no_mount     = ();
    my @corrected    = ();
    my @not_ready    = ();
    my @staf_started = ();

    foreach my $client (@{$clients}) {
        if ($client =~ /sanblaze/i) {
            $Log->comment("Skipping SANBlaze Client $client");
            next;
        }
        try {
            $procs{$client} = NATE::Process->new(
                codespec   => "{" .
                    "use " . __PACKAGE__ . ";" .
                    __PACKAGE__ . "::_execute_local();" .
                "}",
                client     => $client,
                runid      => "local_nate_ready/" . $client,
                params     => { cmd => $cmd },
                background => TRUE,
            );
            $procs{$client}->start();
            push(@staf_started, $client);
        } otherwise {
            my $e = shift();
            if (($e->messages())[0] =~ /Result code: 16 \(NoPathToMachine\)/) {
                $Log->comment(
                    "Staf is not running on $client:\n" . Dumper($e->messages())
                );
                push (@no_staf, $client);
            } elsif (($e->messages())[0] =~ /Could not create STDOUT directory|Error creating stdout file/) {
                $Log->comment(
                    "LOGDIR unavailable for $client:\n" . Dumper($e->messages())
                );
                push (@no_mount, $client);
            } else {
                $Log->comment(
                    "$client is not staf ready: " . Dumper($e->messages())
                );
                push(@not_ready, $client);
            }
        };
    }

    # Start staf on any clients that were not running
    if (scalar(@no_staf)) {
        my @started = $self->start_staf(clients => \@no_staf);
        if (scalar(@no_staf) != scalar(@started)) {
            my @unique = _unique(@started, @no_staf);
            $Log->warn(
                "The following clients are having issues starting STAF " .
                "and will be omitted: " . join(",", @unique)
            );
        }
        push(@corrected, @started);
    }

    # Mount nate logdir on any clients which are not mounted.
    if (scalar(@no_mount)) {
        my @mounted = $self->setup_nate_mounts(clients => \@no_mount);
        if (scalar(@no_mount) != scalar(@mounted)) {
            my @unique = _unique(@mounted, @no_mount);
            $Log->warn(
                "The following clients are having mounting LOGDIR " .
                "and will be omitted: " . join(",", @unique)
            );
        }
        push(@corrected, @mounted);
    }

    # Check the results on any of the clients where staf procs actually started.
    my @nate_ready = ();
    foreach my $client (@staf_started) {
        try {
            $procs{$client}->wait();
            my @results = $procs{$client}->own_results();
            $Log->debug("My Results $client: " . Dumper(@results));
            push(@nate_ready,$client);
        } otherwise {
            my $e = shift();
            $Log->comment(
                "Remote process failed on $client: " . Dumper($e->messages())
            );
            push (@not_ready, $client);
        };
    }

    # Recursively call self to retry the "corrected" clients
    if (scalar(@corrected)) {
        push(@nate_ready, $self->get_nate_ready_clients(
            clients => \@corrected,
            retry   => $retry
        ));
    }

    return wantarray ? @nate_ready : \@nate_ready;
} # End of get_nate_ready_clients

#===============================================================================
# Function: reset_database
#   Resets the database table associated to its corresponding testbed
#
#   Parameters:
#       None
#
#   Returns:
#       TRUE or FALSE
#===============================================================================
sub reset_database {
    my ($self, %args) = @_;

    # Drop the table if it exists, attempt a rollback, return FALSE if failure
    eval {
        $Log->comment("Resetting the database...");
        my $drop = "DROP TABLE IF EXISTS " . DB_TABLE . " ;";
        $Log->comment("SQLITE: " . $drop);
        $self->{dbh}->do($drop) or die();
    };
    if ($@) {
        $Log->comment("Drop table failed, rolling back.");
        eval {
            $self->{dbh}->rollback() or die();
        };
        return FALSE;
    } else {
        $self->{dbh}->commit();
        $Log->comment("Table dropped.");
    }

    # Create the table, attempt a rollback, return FALSE if failure
    eval {
        my $cols   = DB_CREATE_COLUMNS;
        my $create = "CREATE TABLE " . DB_TABLE .
            " ( id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL UNIQUE, " .
            join(", ", @$cols) . " ) ;";
        $Log->comment("SQLITE: " . $create);
        $self->{dbh}->do($create) or die();
    };
    if ($@) {
        eval {
            $Log->comment("Create table failed, rolling back.");
            $self->{dbh}->rollback() or die();
        };
        return FALSE;
    } else {
        $self->{dbh}->commit();
        $Log->comment("Database Table " . DB_TABLE . " created.");
    }

    return TRUE;
} # End of reset_database

#===============================================================================
# Function: discover_devices
#   Discover devices related to Ontap objects on specified hosts.
#
#   Parameters:
#       clients - A reference to an array of hostnames for discovery (optional
#           defaults to CLIENTS.txt)
#       cluster - A cluster name (optional, defaults to CLUSTER_NAME)
#
#   Returns:
#       Array of discovered information about the requested hosts.
#===============================================================================
sub discover_devices {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $clients = &ParamsMgr::get_arg(
        'arg'      => "clients",
        'required' => FALSE,
        'default'  => $self->{clients},
    );
    my $cluster_name = &ParamsMgr::get_arg(
        'arg'      => "cluster",
        'required' => FALSE,
        'default'  => param("CLUSTER_NAME", "-default", undef),
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    $Log->comment("Starting discovery on these clients: " . Dumper($clients));
    my @client_info = ();
    foreach my $client (@{$clients}) {
        my $runid = "discovery_$client";
        push (@client_info, {
            codespec   => "{ " . __PACKAGE__ . "::_discover_devices(); }",
            runid      => $runid,
            background => TRUE,
            params     => {
                hostname     => $client,
                cluster_name => $cluster_name,
            }
        });
    }

    my $status = $self->_run_workflow(
        objects    => \@client_info,
        operation  => "discovery",
        background => TRUE,
    );

    if ($status) {
        return TRUE;
    } else {
        NATE::BaseException->throw(
            "Failed discovering one or more devices from one or more clients"
        );
        return FALSE;
    }
} # End of discover_devices

#===============================================================================
# Function: get_devices
#   Retrieve devices based on provided filters.
#
#   Parameters:
#       clients - Array of clients to filter (overwrites $hostnames)
#       hostnames - Array of hostnames to filter
#       hosttypes - Array of hosttypes to filter
#       devices_names - Array of devices_names to filter
#       mount_points - Array of mount_points to filter
#       vservers - Array of vservers to filter
#       volumes - Array of volumes to filter
#       luns - Array of luns to filter
#       ontap_paths - Array of ontap_paths to filter
#       lun_serials - Array of lun_serials to filter
#       protocols - Array of protocols to filter
#
#   Returns:
#       Array of devices as filtered by the parameters.
#===============================================================================
sub get_devices {
    my ($self, %args) = @_;

    # For compatibility, use client if exists and no hostname was passed.
    if (defined($args{clients}) and not defined($args{hostnames})) {
	    %args = (%args, hostnames => $args{clients});
	}

    my @select_columns = ();
    my @select_values = ();

    # Initialize the host name based on the object's attribute (list of clients)
    # String of (column=? OR column=?) for each value
    while (my ($key, $value) = each(%args)) {
        if (defined($args{$key}) and exists($args{$key})) {
            # Special case for serial numbers, which contain weird characters
            # not supported by LIKE
            if ($key =~ /lun_serials/i) {
                push(@select_columns,
                    "(" . join(" OR ", map substr($key, 0, -1) .
                    " IS ?", @{$args{$key}}) .
                    ")"
                );
            } else {
                push(@select_columns,
                    "(" . join(" OR ", map substr($key, 0, -1) .
                    " LIKE ?", @{$args{$key}}) .
                    ")"
                );
            }
            push(@select_values, @{$args{$key}});
        }
    }

    # Number of rows selected is placed here
    my $device_results = [];

    # Create select where column1=?, column2=?
    my $select = "";
    my $columns = DB_SELECT_COLUMNS;
    $select .= "SELECT " . join(", ", @$columns) . " ";
    $select .= "FROM " . DB_TABLE . " ";
    # If there are no filters, don't use WHERE in query
    if (scalar(@select_columns) > 0) {
        $select .= "WHERE " . join(' AND ', @select_columns) . " ";
    }
    $select .= ";";
    $Log->comment("SQLITE:\n" . $select);
    $Log->comment("...values: " . join(', ', @select_values));

    # Update the table
    eval {
        my $sth = $self->{dbh}->prepare($select);
        $sth->execute(@select_values);
        $device_results = $sth->fetchall_arrayref();
        $self->{dbh}->commit();
    };
    if ($@) {
        $Log->error("Select failed: " . $self->{dbh}->errstr);
    }

    $Log->debug("Rows: " . Dumper($device_results));

    # Assign devices to devices array
    my @devices = ();
    foreach my $device_result (@$device_results) {
        my $device = {};
        foreach my $column (@$columns) {
            $device->{$column} = shift(@$device_result);
        }
        push(@devices, $device);
    }

    return wantarray ? @devices : \@devices;
} # End of get_devices

################################################################################
# P R I V A T E  M E T H O D S
################################################################################

#===============================================================================
# Description: Connect and return a database handle
#===============================================================================
sub _database_handle {
    my (%args) = @_;

    my $dsn      = "dbi:SQLite:dbname=" . DB_FILE;
    my $username = "";
    my $password = "";
    my $dbh      = undef;

    eval {
        $dbh = DBI->connect($dsn, $username, $password, {
            RaiseError => TRUE,
            PrintError => TRUE,
            AutoCommit => FALSE,
        }) or die ("Can't connect to the database: " . DB_FILE . "\n");
        chmod(0766, DB_FILE);
    };
    if ($@) {
        $Log->error("Can't open the database handle: " . DBI->errstr);
        return undef;
    } else {
        $Log->comment("Connected to database: " . DB_FILE);
    }

    return $dbh;
} # End of _database_handle

#===============================================================================
# Description: Discovers devices related to Ontap objects for specified host
#===============================================================================
sub _discover_devices {
    my ($class, %args) = @_;

    my $params       = NATE::ParamSet->get_global();
    my $status       = FALSE;
    my $hostname     = $params->get('hostname');
    my $cluster_name = $params->get('cluster_name');
    my $dbh          = _database_handle();

    unless (defined($dbh)) {
        $Log->comment(
            "It looks like the DB connection failed. Discovery is aborting " .
            "because the devices cannot be stored for later retrieval."
        );
        return FALSE;
    }

    my $client = {};
    $client->{hostname}   = $hostname;
    $client->{hosttype}   = "";
    $client->{devices}    = [];
    $client->{initiators} = [];
    my $cmd_interface = undef;

    if (defined ($cluster_name)) {
        $cmd_interface = NACL::C::Cluster->new(name => $cluster_name);
    } else {
        $cmd_interface = NACL::C::Node->find();
    }

    $cmd_interface->apiset_must(interface => 'CLI');
    $cmd_interface->apiset();

    # Create SANClient object for initiator discovery
    my $hostrec = Tharn::host($hostname);
    my $protocol = "";

    if ($hostrec->{CONFIGS} =~ /fcp/i) {
        $Log->comment(
            "Making the assumption that host $hostname is FCP as it was " .
            "detected in the CONFIGS param in the NATE_HOST file: "
        );
        $protocol = "fcp";
    } elsif ($hostrec->{CONFIGS} =~ /iscsi/i) {
        $Log->comment(
            "Making the assumption that host $hostname is iSCSI as it was " .
            "detected in the CONFIGS param in the nate_host file."
        );
        $protocol = "iscsi";
    } else {
        $Log->log(NATE::Result::Fail->new(
            "Neither iscsi nor fcp were defined in the CONFIGS param " .
            "in the NATE_HOST file for host $hostname.\n" .
            $hostrec->{CONFIGS})
        );
    }
    $Log->comment("Creating SanClient Object for " . $hostname);
    my $client_obj = new SanClient(
        host     => $hostname,
        protocol => $protocol,
    );
    if ($client_obj->error()) {
        $Log->log(NATE::Result::Fail->new(
            "There is an error with the SanClient object, please investigate." .
            $client_obj->errmsg())
        );
    }

    # Collect hosttype
    $client->{hosttype} = $client_obj->hosttype();

    # Collect initiator addresses
    my @initiators = ();
    my @client_inits = $client_obj->get_initiators();

    $Log->debug("Client Initiators: " . join(', ', @client_inits));
    foreach my $client_init (@client_inits) {
        my $initiator;
        if ($protocol =~ /fcp/i) {
            $initiator = $client_init->get_portname();
        } else {
            $initiator = $client_init->get_nodename();
        }

        # Fix WWPNS since SANBLAZE provided values don't have colons
        if (($initiator !~ /:/) and ($protocol =~ /fcp/i)) {
            # Join every two characters with a colon
            $initiator = join(":", ($initiator =~ m/.{2}/g));
        }

        push(@initiators, $initiator);
    }
    # Speed lookups done later, convert initiators into a hash
    my %initiators_hash = map( {$_ => 1 } @initiators );
    $client->{initiators} = \@initiators;
    $Log->comment("Host Initiators: " . join(', ', @{$client->{initiators}}));

    # Collect Vserver names based on iGroups/initiators
    my @vserver_names = ();
    $cmd_interface->apiset_must(interface => 'CLI');
    $cmd_interface->apiset();

    if (scalar(@initiators)) {
        my %fetch_attrs = (
            command_interface => $cmd_interface,
            allow_empty       => TRUE,
            filter            => { initiator => \@initiators },
            requested_fields  => ['vserver'],
        );
        # If there are lots of inits, don't filter or CLI command breaks
        if (scalar(@initiators) > IGROUP_INITIATOR_FILTER_LIMIT) {
            delete($fetch_attrs{filter});
        }
        # Collect the Igroups
        my @igroups = NACL::CS::LunIgroup->fetch(%fetch_attrs);
        # Save all the Vserver names
        foreach my $igroup (@igroups) {
            my $vserver = $igroup->{vserver};
            push(@vserver_names, $vserver);
        }
    }
    $Log->comment("Host Vservers: " . join(', ', @vserver_names));

    # Hold all discovered device records here for future database update
    my @records = ();

    # Discover Windows devices
    if ($client->{hosttype} =~ /windows/i) {
        # Collect MPDevMgr details
        my $dev_mgr = $client_obj->create_mpdevmgr(
            -nacl_cluster_objs => [$cmd_interface],
            -vservers          => \@vserver_names,
        );
        if ($dev_mgr->error()) {
            $Log->log(NATE::Result::Fail->new(
                "There was an error with MPDevMgr object, " .
                "please investigate. " . $dev_mgr->errmsg()
            ));
        }

        # Filter devices by NETAPP luns
        my @netapp_devices = $dev_mgr->get_filtered_devices(
            -device_vendor => 'NETAPP'
        );

        foreach my $netapp_device (@netapp_devices) {
            # temporary fix to add mount_point value
            my $mount_point = $netapp_device->{mp_device_path};
            $mount_point =~ s/Disk/PHYSICALDRIVE/g;
            my $device_name = $mount_point;
            $device_name =~ s/\\|\.//g;
            # Save this device record to be updated/inserted into the db
            push(@records, {
                hostname    => $client->{hostname},
                hosttype    => $client->{hosttype},
                device_name => $device_name,
                mount_point => $mount_point,
                vserver     => $netapp_device->{ontap_vserver},
                volume      => $netapp_device->{ontap_vol_name},
                lun         => $netapp_device->{ontap_lun_name},
                ontap_path  => $netapp_device->{ontap_lun_path},
                lun_serial  => $netapp_device->{os_specific_data}->{serial},
                protocol    => $protocol,
            });
        }
        # Discover SANBlaze devices
    } elsif ($client->{hosttype} =~ /sanblaze/i) {
        # Abort if no Vservers are related
        unless (scalar(@vserver_names)) {
            $Log->error("There are no Vservers associated to this host.");
            return FALSE;
        }

        # Collect serial numbers of luns and their initiator mappings
        my @luns = ();
        my @mapped_luns = ();

        foreach my $vserver_name (@vserver_names) {
            push(@luns, NACL::CS::Lun->fetch(
                command_interface => $cmd_interface,
                allow_empty       => TRUE,
                filter            => {
                    vserver => $vserver_name,
                    mapped  => 'mapped'
                },
                requested_fields  => ['volume','vserver','lun','path','serial'],
            ));
            # Collecting inititators related to the Luns
            push(@mapped_luns, NACL::CS::LunMapped->fetch(
                command_interface => $cmd_interface,
                allow_empty       => TRUE,
                filter            => { vserver => $vserver_name },
                requested_fields  => ['path','initiators'],
            ));
        }

        $Log->debug("Vserver Luns: " . Dumper(\@luns));
        $Log->debug("Mapped Luns: " . Dumper(\@mapped_luns));

        # Create hash with path as key and initiators as value
        my %mapped_path_initiator = map(
            { $_->path() => $_->initiators() } @mapped_luns
        );
        $Log->debug(
            "Mapped Paths to Initiators: " . Dumper(\%mapped_path_initiator)
        );

        # Open connection to SANBlaze
        my $sanblaze = San::FIT::Misfit::Services::IO::Sanblaze->new(
            host => $client->{hostname},
        );

        # Analyze all mapped luns and map against sanblaze ports/targets
        foreach my $lun (@luns) {
            # Confirm it's a mapped lun with a matching path
            unless (exists $mapped_path_initiator{$lun->path()}) {
                $Log->debug(
                    "Lun path " . $lun->path() . "is not mapped to an " .
                    "initiator, checking next Lun path."
                );
                next;
            }

            # Confirm that the mapped initiator matches the lun path
            my $mapped_initiator = $mapped_path_initiator{ $lun->path() };
            if (exists $initiators_hash{ $mapped_initiator }) {
                # Add a new device if mapped and matching
                my $device = {};
                $device->{vserver}     = $lun->vserver();
                $device->{volume}      = $lun->volume();
                $device->{lun}         = $lun->lun(); # NOTE breaks with ZAPI.
                $device->{ontap_path}  = $lun->path();
                $device->{lun_serial}  = $lun->serial();
                $device->{device_name} = $lun->serial();
                $device->{mount_point} = $lun->serial();

                # Save this device record to be updated/inserted into the db
                push(@records, {
                    hostname    => $client->{hostname},
                    hosttype    => $client->{hosttype},
                    device_name => $device->{device_name},
                    mount_point => $device->{mount_point},
                    vserver     => $device->{vserver},
                    volume      => $device->{volume},
                    lun         => $device->{lun},
                    ontap_path  => $device->{ontap_path},
                    lun_serial  => $device->{lun_serial},
                    protocol    => $protocol,
                });
            } else {
                $Log->debug(
                    "Lun path " . $lun->path() . " mapped inititator " .
                    $mapped_initiator . " does not belong to this client."
                );
            }
        }
    }

    # Update or Insert database with client devices
    my $update_or_insert = _update_or_insert_records(
        dbh => $dbh,
        records => \@records
    );
    if ($update_or_insert) {
        $Log->comment(
            "Updated database with " . scalar(@records) . " discovered devices."
        );
    } else {
        $Log->comment(
            "Database update failed while trying to update " .
            scalar(@records) . " discovered devices."
        );
        return FALSE;
    }
    $Log->log(NATE::Result::Pass->new(
        "Done discovering devices for $hostname."
    ));

    return TRUE;
} # End of _discover_devices

#===============================================================================
# Description: Update or Insert records, matching on lun_serial
#===============================================================================
sub _update_or_insert_records {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $dbh = &ParamsMgr::get_arg(
        'arg'      => "dbh",
        'required' => TRUE,
    );
    my $records = &ParamsMgr::get_arg(
        'arg'      => "records",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    foreach my $record (@{$records}) {
        my @update_columns = ();
        my @update_values = ();
        my @insert_columns = ();
        my @insert_values = ();
        my $lun_serial = $record->{lun_serial};

        while (my ($key, $value) = each(%{$record})) {
            if (defined($record->{$key}) and exists($record->{$key})) {
                push(@update_columns, "$key=?");
                push(@update_values, $value);
                push(@insert_columns, "$key");
                push(@insert_values, $value);
            }
        }

        # lun_serial is required for insert AND update
        push(@update_values, $lun_serial);
        push(@insert_columns, "lun_serial");
        push(@insert_values, $lun_serial);

        # Number of rows updated is placed here
        my $rows;

        # Create update column1=?, column2=?
        my $update = "UPDATE " . DB_TABLE .
            " SET " . join(', ', @update_columns) . " WHERE lun_serial=? ;";
        $Log->comment("SQLITE:\n" . $update);
        $Log->comment("...values: " . join(', ', @update_values));

        # Update the table
        eval {
            $rows = $dbh->do($update, undef, @update_values);
        };
        if ($@) {
            $Log->comment("Update failed. " . $dbh->errstr);
        }

        # If a row was updated, skip to next record
        if ($rows > 0) {
            $Log->comment("Update successful.");
            next;
        } else {
            $Log->comment("Device not found, inserting new device.");
        }

        # Create insert () values ()
        my $insert = "INSERT INTO " . DB_TABLE .
            " ( " . join(', ', @insert_columns) . " ) VALUES" .
            " ( " . join(', ', map "?", @insert_values) . " ) ;";
        $Log->comment("SQLITE:\n" . $insert);
        $Log->comment("...values: " . join(', ', @insert_values));

        # Insert into the table
        eval {
            $dbh->do($insert, undef, @insert_values);
        };
        if ($@) {
            $Log->comment("Insert failed. " . $dbh->errstr);
        } else {
            $Log->comment("Insert successful.");
        }
    }

    eval {
        $dbh->commit();
    };
    if ($@) {
        $Log->comment("Commit failed, rolling back. " . $dbh->errstr);
        eval {
            $dbh->rollback()
        };
        return FALSE;
    }

    return TRUE;
} # End of _update_or_insert_records

#===============================================================================
# Description: Update or Insert records, matching on lun_serial
#===============================================================================
sub _get_clients_from_text_file {
    my ($self, %args) = @_;

    &ParamsMgr::process_args(%args);
    my $file_name = &ParamsMgr::get_arg(
        'arg'      => "file_name",
        'required' => FALSE,
        'default'  => "CLIENTS.txt",
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return wantarray ? () : [];
    }

    my @clients = ();
    open(CLIENT_FILE, "<", $file_name) or die ("Bad... $!");

    while (<CLIENT_FILE>) {
        chomp $_;
        next if ($_ =~ /#/);
        push(@clients, $_);
    }

    close(CLIENT_FILE);

    return wantarray ? @clients : \@clients;
} # End of _get_clients_dot_txt

#===============================================================================
# Description: Subprocess responsible for executing a command local to a client
#===============================================================================
sub _execute_local {
    my ($class, %args) = @_;

    my $params = NATE::ParamSet->get_global();
    &ParamsMgr::process_args(%args);
    my $cmd = &ParamsMgr::get_arg(
        'arg'      => "cmd",
        'required' => FALSE,
        'default'  => $params->get('cmd'),
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    $Log->comment("localhost>> $cmd");
    my $result = `$cmd`;
    $Log->comment("localhost<< $result");

    return TRUE;
} # End of _execute_local

#===============================================================================
# Description: Subprocess responsible for starting staf on windows
#===============================================================================
sub _start_staf {
    my ($class, %args) = @_;

    my $params = NATE::ParamSet->get_global();
    &ParamsMgr::process_args(%args);
    my $client = &ParamsMgr::get_arg(
        'arg'      => "client",
        'required' => FALSE,
        'default'  => $params->get('client'),
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    my $client_obj = undef;
    my $output = "";
    try {
        $client_obj = NACL::C::Client->new(name => $client);
        $output = $client_obj->execute(
            command => START_STAF_WIN,
        );
    } otherwise {
        my $e = shift;
        $e->throw();
    };

    if ($output =~ /is not recognized as an internal or external command/) {
        NATE::BaseException->throw($output);
    }

    return TRUE;
} # End of _start_staff

#===============================================================================
# Description: Subprocess responsible for mapping the windows shares for NATE
#===============================================================================
sub _mount_required_nate_shares {
    my ($class, %args) = @_;

    my $params = NATE::ParamSet->get_global();
    &ParamsMgr::process_args(%args);
    my $client = &ParamsMgr::get_arg(
        'arg'      => "client",
        'required' => FALSE,
        'default'  => $params->get('client'),
    );

    # Setup the client connection we will be working with.
    my $client_obj = undef;
    try {
        $client_obj = NACL::C::Client->new(name => $client);
    } otherwise {
        my $e = shift;
        $e->throw();
    };

    # Attempt to mount the log directory on the clients so that it can report
    # its own results rather than relying on the parent.
    my $output = "";
    my $success = "The command completed successfully";
    my $status = TRUE;

    try {
        $output = $client_obj->execute(command => LOG_MNT_WIN_RTP);
    } otherwise {
        my $e = shift;
        $e->throw();
    };

    # Validate the result output
    unless ($output =~ /$success/) {
        NATE::Result::Fail->new("$client failed logdir")->log();
        $status = FALSE;
    }

    # TODO: still need to setup p4 clients on scratch space and use those
    # servers here.  Unless we can get a system user created for the
    # rtphomedirs servers.
    # Now we need to mount the user home directory so we can access user copies
    # of tests and libs.
    $output = "";
    try {
        $output = $client_obj->execute(command => LOG_MNT_WIN_RTP);
    } otherwise {
        my $e = shift;
        $e->throw();
    };
    # Validate the result output
    unless ($output =~ /$success/) {
        NATE::Result::Fail->new("$client failed homedir")->log();
        $status = FALSE;
    }

    unless ($status) {
        NATE::Result::Fail->new("$client")->log();
    }
} # End of _mount_required_nate_shares

#===============================================================================
# Description: Given two arrays, lista and listb, this method returns the
#   elements from lista that are not in listb
#===============================================================================
sub _unique() {
    my (@lista, @listb) = @_;

    my @return_array = ();
    foreach my $list_item (@lista) {
        if (grep($list_item, @listb)) {
            push(@return_array, $list_item);
        }
    }

    return wantarray ? @return_array : \@return_array;
}

return 1;