# Copyright (c) 2001-2013 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary IO module for Misfit
## @author dl-san-fit@netapp.com
## @status Private

# Package: San::FIT::Misfit::Services::MisfitIO
# The IO module provides an interface to NetApp IO tools.

package San::FIT::Misfit::Services::MisfitIO;

# Compiler Directives
use strict;
use warnings;

# Imported modules
use JSON;
use TCD;
use Tharn;
use ParamsMgr;
use Data::Dumper;
use NATE::Result;
use San::FIT::Misfit::Core::Util;
use NATE::Exceptions::Argument qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::MessageBroker;
use San::FIT::Misfit::Services::InitiatorOps;
use San::FIT::Misfit::Services::IO::Workload;

# Inherit base class - Misfit Base
use base qw( San::FIT::Misfit::Core::Base );

# Constants
# Message broker doesn't like '\'.  '%7' is a hackaround
# assumes windows initially
use constant {
    DT_PATH  => 'x:%7%7SAN%7%7bin%7%7common%7%7dt.staging%7%7dt.exe',
    EXCHANGE => 'io_exchange',
};

#===============================================================================
#  PUBLIC METHODS
#===============================================================================

#===============================================================================
# Function: start
#   Start I/O on one or more luns.  This function will accept any combination
#   of vserver, volume, or lun names and start I/O on all luns that correspond
#   to the set of parameters.
#
#   For example:
#
#   luns     => [lun1, lun2, lun3]
#       start I/O on these 3 luns (or possibly more if multiple vservers
#       and/or volumes have luns with matching names)
#
#   luns     => [lun1, lun2, lun3]
#   volumes  => [vol1]
#       start I/O on these 3 luns within vol1
#
#   volumes  => [vol1, vol2]
#       start I/O on all luns in vol1 and vol2 (any vserver)
#
#   vservers => [vs1]
#       start I/O on all luns in this vserver
#
#   vservers => [vs1]
#   volumes  => [vol1]
#   luns     => [lun1]
#       start I/O on vs1:/vol/vol1/lun1
#
#   Parameters:
#       vservers  - list of vservers
#       volumes   - list of volumes
#       luns      - list of luns
#       perc      - use this percentage of luns that match the filter(s),
#                   e.g. perc => 40
#       discover_on_fail - if no devices are found, discover devices?
#                   Defaults to TRUE.
#       workload_id - name of a pre-defined set of I/O params; workloads will
#                   correspond to those defined based on real-world profiles
#                   (e.g. oltp, dss, etc) For a complete list see:
#                   San::FIT::Misfit::Services::IO::Workload.pm
#       tool      - I/O tool to use (if non-sanblaze lun); defaults to dt
#       io_params - key/value pairs that will be mapped to the particular
#                   tool's keywords or api
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub start {
    my (%args) = @_;

    $args{operation} = 'io_start';
    my $rc = _send_operation(%args);

    return $rc;
}

#===============================================================================
# Function: stop
#   Stop IO on one or more luns
#
#   Parameters:
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub stop {
    my (%args) = @_;

    $args{operation} = 'io_stop';
    my $rc = _send_operation(%args);

    return $rc;
}

#===============================================================================
# Function: pause
#   Pause IO on one or more luns
#
#   Parameters:
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub pause {
    my (%args) = @_;

    $args{operation} = 'io_pause';
    my $rc = _send_operation(%args);

    return $rc;
}

#===============================================================================
# Function: resume
#   Resume IO on one or more luns
#
#   Parameters:
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub resume {
    my (%args) = @_;

    $args{operation} = 'io_start';
    my $rc = _send_operation(%args);

    return $rc;
}

#===============================================================================
# Function: modify
#   Modify IO parameters
#
#   Parameters:
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub modify {
    my (%args) = @_;

    # placeholder

    return TRUE;
}

#===============================================================================
# Function: toggle
#   Toggle which paths are used for I/O.  This needs to be flushed out.
#
#   Parameters:
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub toggle {
    my (%args) = @_;

    # placeholder

    return TRUE;
}

#==============================================================================
#  PRIVATE METHODS
#==============================================================================

#==============================================================================
# Description: Process an IO operation for a set of luns.
#==============================================================================
sub _send_operation {
    my (%args) = @_;
    my %filter;

    &ParamsMgr::process_args(%args);
    my $operation = &ParamsMgr::get_arg(
        'arg'      => "operation",
        'required' => TRUE,
    );
    $filter{vservers} = &ParamsMgr::get_arg(
        'arg'      => "vservers",
        'required' => FALSE,
        'default'  => undef,
    );
    $filter{volumes} = &ParamsMgr::get_arg(
        'arg'      => "volumes",
        'required' => FALSE,
        'default'  => undef,
    );
    $filter{luns} = &ParamsMgr::get_arg(
        'arg'      => "luns",
        'required' => FALSE,
        'default'  => undef,
    );
    my $discover_on_fail = &ParamsMgr::get_arg(
        'arg'      => "discover_on_fail",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $workload_id = &ParamsMgr::get_arg(
        'arg'      => "workload_id",
        'required' => FALSE,
        'default'  => "",
    );
    my $tool = &ParamsMgr::get_arg(
        'arg'      => "tool",
        'required' => FALSE,
        'default'  => "dt",
    );
    my $io_params = &ParamsMgr::get_arg(
        'arg'      => "io_params",
        'required' => FALSE,
        'default'  => undef,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return FALSE;
    }

    # At least one of 'vservers', 'volumes', or 'luns' must be specified
    unless (
        defined($filter{volumes}) ||
        defined($filter{luns}) ||
        defined($filter{vservers})
    ) {
        $Log->comment("Performing $operation on ALL luns!");
        $filter{vservers} = ['%'];
    }

    # Get the requested workload.


    my $initiator_ops = San::FIT::Misfit::Services::InitiatorOps->new();

    # Get devices corresponding to luns.  If no devices are found,
    # conditionally do a discover and retry.
    my @devices;
    while (TRUE) {
        @devices = $initiator_ops->get_devices(%filter);
        if (!scalar(@devices)) {
            if ($discover_on_fail) {
                $Log->comment("No devices found.  Attempting discover");
                if (!$initiator_ops->discover_devices()) {
                    $Log->error("discover_devices() failed");
                    return FALSE;
                }
                $discover_on_fail = FALSE;
                next;
            } else {
                $Log->error(
                    "No devices found. Discover on fail was not requested"
                );
                return FALSE;
            }
        }
        last;
    }

    # Push one message per device
    foreach my $dev (@devices) {
        my $workload = San::FIT::Misfit::Services::IO::Workload::get_workload(
            id => $workload_id
        );
        # Build the io exchange message
        my $msg = _build_message(
            operation => $operation,
            workload  => $workload,
            device    => $dev
        );

        # Publish the message
        San::FIT::Misfit::Core::Util::publish_message(
            routing_key => $msg->{routing_key},
            exchange    => $msg->{exchange},
            message     => $msg->{message},
        );
    }

    return TRUE;
}
#==============================================================================
# Description: Build message
#==============================================================================
sub _build_message {
    my (%args) = @_;
    my %msg;

    &ParamsMgr::process_args(%args);
    my $operation = &ParamsMgr::get_arg(
        'arg'      => "operation",
        'required' => TRUE,
    );
    my $device = &ParamsMgr::get_arg(
        'arg'      => "device",
        'required' => TRUE,
    );
    my $workload = &ParamsMgr::get_arg(
        'arg'      => "workload",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return {};
    }

    $msg{exchange} = 'io_exchange';
    if ($device->{hosttype} =~ 'windows') {
        my $command = _build_dt_command(
            device => $device->{device_name},
            workload => $workload,
        );
        $msg{routing_key} = $device->{hostname};
        $msg{message} = {
            msg => $operation,
            command => $command,
            device => $device->{device_name},
        };

    } elsif ($device->{hosttype} =~ 'sanblaze') {
        $msg{routing_key} = 'sanblaze';
        $msg{message} = {
            msg      => $operation,
            workload => $workload->{sanblaze},
            device   => $device->{lun_serial},
        };
    } else { # need to add linux
        $Log->error("unexpected hosttype: " . $device->{hosttype});
        return {};
    }

    return \%msg;
}

#==============================================================================
# Description: Build a dt command line
#==============================================================================
sub _build_dt_command {
    my (%args) = @_;

    &ParamsMgr::process_args(%args);
    my $device = &ParamsMgr::get_arg(
        'arg'      => "device",
        'required' => TRUE,
    );
    my $workload = &ParamsMgr::get_arg(
        'arg'      => "workload",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        return "";
    }

    # Attempt to retrieve the requested workload
    my $options = "";
    my $dt_workload = $workload->{dt};
    foreach my $key (keys(%$dt_workload)) {
        $options .= " $key=$dt_workload->{$key}";
    }
    if ($options eq "") {
        $Log->error("failed to get dt options");
        return "";
    }

    return DT_PATH . " of=%7%7%7%7.%7%7$device $options";
}

#===============================================================================
#  POD
#===============================================================================

=head1 Documentation and examples:

L<http://misfitapi.gdl.englab.netapp.com/>

=cut

1;
