# Copyright (c) 2001-2012 NetApp, Inc. All Rights Reserved
# Any use, modification, or distribution is prohibited
# without prior written consent from NetApp, Inc.
#
## @summary Common configuration methods for ONTAP objects.
## @author dl-san-fit@netapp.com
## @status Private
## @pod No thanks

# Package: San::FIT::Misfit::Services::Resource
# Module in charge of provisioning cluster and monitoring resource
# utilization in Misfit test scenario.

package San::FIT::Misfit::Services::Resource;

use strict;
use warnings;

# Perl's public modules
use List::MoreUtils qw(all first_value apply first_index);
use List::Util qw(sum);
use Compress::Zlib;
use Data::Dumper;
use File::Spec;
use JSON;
use Clone qw(clone);

# NetApp Modules
use Tharn;
use ParamsMgr;
use San::SanMgr;
use NATE::Time qw(timestamp);
use NATE::BaseException qw(:try);
use NATE::Util::DataTypes qw(TRUE FALSE);
use San::FIT::Misfit::Core::Util;
use San::FIT::Misfit::Core::Workflow;
use San::FIT::Misfit::Features::AggrOps;
use San::FIT::Misfit::Core::MessageBroker;
use San::FIT::Misfit::Features::VserverOps;
use San::FIT::Misfit::Features::VolOps;
use San::FIT::Misfit::Features::LunOps;
use San::FIT::Misfit::Features::Igroup;
use NACL::STask::LunIgroup;
use NACL::STask::Vserver;
use NACL::STask::Cluster;
use NACL::CS::Aggregate;
use NACL::STask::Volume;
use NACL::C::Cluster;
use NACL::STask::Lun;
use NACL::C::Node;


# Use the Misfit Core Base Class
use base qw( San::FIT::Misfit::Core::Base );

# Constants
use constant {
    DEFAULT_DEBUG_LEVEL      => 1,
    CONFIGURE_METHOD_NAME    => "setup",
    DECONFIGURE_METHOD_NAME  => "cleanup",
    MAX_PERCENT_RESOURCE_GET => 100,
    INTERCLUSTER_TIMEOUT     => 180,
};

#===============================================================================
# Constructor: new
#   Create and initialize a new SANFIT Configure Object
#
#   Parameters:
#       debug_level - TODO Not sure what this does
#
#   Returns:
#       New Resource object.
#===============================================================================
sub new {
    my ($class, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $debug_level = &ParamsMgr::get_arg(
        'arg'      => "debug_level",
        'required' => FALSE,
        'default'  => DEFAULT_DEBUG_LEVEL
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    my $self = $class->SUPER::new();
    $self->{id} = "San::FIT::Misfit::Services::Resource";
    $self->{debug_level} = $debug_level;
    $self->{client_device_list} = [];
    $self->{intercluster_nodes} = [];
    $self->{default_node} = NACL::C::Node->find();
    $self->{default_node}->apiset_must(interface => 'ZAPI');
    #_discover_clusters();
    my $cluster_name = param("CLUSTER_NAME", "-default", "");
    my $peer_name    = param("PEER_CLUSTER", "-default", $cluster_name);

    if ($cluster_name eq "") {
        $Log->comment("CLUSTER_NAME was not provided, this script will fail");
        return {};
    }

    $self->{cluster_home_name} = $cluster_name;
    $self->{cluster_home} = NACL::C::Cluster->new(name => $cluster_name);
    $self->{cluster_home}->apiset_must(interface => 'ZAPI');
    $self->{cluster_home}->apiset();
    $self->{cluster_home_cli} = NACL::C::Cluster->new(name => $cluster_name);
    $self->{cluster_home_cli}->apiset_must(interface => 'CLI');
    $self->{cluster_home_cli}->apiset();

    $self->{cluster_peer_name} = $peer_name;
    $self->{cluster_peer} = NACL::C::Cluster->new(name => $peer_name);
    $self->{cluster_peer}->apiset_must(interface => 'ZAPI');
    $self->{cluster_peer}->apiset();
    $self->{cluster_peer_cli} = NACL::C::Cluster->new(name => $peer_name);
    $self->{cluster_peer_cli}->apiset_must(interface => 'CLI');
    $self->{cluster_peer_cli}->apiset();

    bless($self, $class);

    $Log->exit();
    return $self;
} # End new

#===============================================================================
# Function: configure
#   Configure an xml with SanMgr or EnGenX (future) with and XML or
#   resource_list (future)
#
#   Parameters:
#       xml_files - List of xmls for each cluster to provision
#       config_files - List of config files for each cluster to provision
#       background - Run workflow serially or in parallel
#       id - TODO
#       intercluster - TRUE, if interclustering needs to be configured, FALSE
#           otherwise.
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub configure {
    my ($self, %args) = @_;

    return $self->_operate(
        %args,
        operation => CONFIGURE_METHOD_NAME,
        id => (caller(0))[3]
    );
} # End configure

#===============================================================================
# Function: deconfigure
#   Removes the configuration that was configured within the current
#   Configure object or any XML or resource_list that is passed in.
#
#   Parameters:
#       xml_files - List of xmls for each cluster to provision
#       config_files - List of config files for each cluster to provision
#       background - Run workflow serially or in parallel
#       id - TODO
#       intercluster - TRUE, if interclustering needs to be configured, FALSE
#           otherwise.
#
#   Returns:
#       TRUE if successful, FALSE otherwise.
#===============================================================================
sub deconfigure {
    my ($self, %args) = @_;

    return $self->_operate(
        %args,
        operation => DECONFIGURE_METHOD_NAME,
        id => (caller(0))[3]
    );
} # End deconfigure

#===============================================================================
# Function: get_objects
#   Retrieve ONTAP objects specifications or NACL objects.
#
#   NOTE!! This method should not be used to reserve objects for Misfit
#       swimlanes.  Use <San::FIT::Misfit::Core::Util::reserve_resources>
#       instead.
#
#   Parameters:
#       type - Type of objects to get.  Valid types are 'volume', 'aggregate',
#           'vserver', 'lun', 'igroup', and 'client'.
#       count - Number of objects to return.  Defaults to 'all'.
#       filter - Filter to search for objects.
#       node - Node to search for objects on.
#       nacl_objects - If TRUE, return NACL objects.  If FALSE (default), return
#           barebone object information.
#
#   Returns:
#       List of objects meeting criteria above.
#===============================================================================
sub get_objects {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $type = &ParamsMgr::get_arg(
        'arg'      => "type",
        'required' => TRUE
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 'all',
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => {},
    );
    my $node = &ParamsMgr::get_arg(
        'arg'      => "search_node",
        'required' => FALSE,
        'default'  => undef,
    );
    my $nacl_objects = &ParamsMgr::get_arg(
        'arg'      => "nacl_objects",
        'required' => FALSE,
        'default'  => FALSE,
    );
    my $requested_fields = &ParamsMgr::get_arg(
        'arg'      => "requested_fields",
        'required' => FALSE,
        'default'  => undef,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return wantarray ? () : [];
    }

    my $nacl_module = "";
    my @objs = ();
    my @results = ();
    if (defined($node)) {
        $node = NACL::C::Node->new(node => $node);
        $node->apiset_must(interface => 'ZAPI');
    } else {
        $node = $self->{default_node};
    }

    # Determine the NACL module we need to use.
    if ($type =~ /aggregate/i) {
        $nacl_module = "NACL::CS::Aggregate";
    } elsif ($type =~ /vserver/i) {
        $nacl_module = "NACL::CS::Vserver";
    } elsif ($type =~ /volume/i) {
        $nacl_module = "NACL::CS::Volume";
    } elsif ($type =~ /lun/i) {
        $nacl_module = "NACL::CS::Lun";
    } elsif ($type =~ /igroup/i) {
        $nacl_module = "NACL::CS::LunIgroup";
    } elsif ($type =~ /client/i) {
        return wantarray ? @{$self->{client_device_list}} : $self->{client_device_list};
    } else {
        NATE::Exceptions::Argument->throw(
            "The value: $type passed for the parameter " .
            "'type' is not supported.");
        $Log->exit();
        return wantarray ? @results : \@results;
    }

    # No need to catch exceptions yet
    if (defined($requested_fields)) {
        @objs = $nacl_module->fetch(
            command_interface => $node,
            filter => $filter,
            allow_empty => TRUE,
            requested_fields => $requested_fields,
        );
    } else {
        @objs = $nacl_module->fetch(
            command_interface => $node,
            filter => $filter,
            allow_empty => TRUE,
        );
    }

    # Done if no objects were returned.
    unless (scalar(@objs)){
        $Log->exit();
        return wantarray ? () : [];
    }

    # Shrink the list to the number of items specified in count unless the user
    # requests all.  If they specify a count make sure it is not greater than
    # the number of objects requested.
    unless ($count eq 'all') {
        if ($count > scalar(@objs)) {
            $count = scalar(@objs);
        }
        @objs = @objs[0 .. $count-1];
    }

    # Here we convert the objects to hash refs containing the attributes
    # we need in Misfit, unless the user explicitly requests the nacl objects.
    if ($nacl_objects) {
        @results = @objs;
    } else {
        foreach my $obj (@objs) {
            my $attrs = $self->get_object_type_and_attrs(object=>$obj);
            push(@results, $attrs->{misfit_attributes});
        }
    }

    $Log->exit();
    return wantarray ? @results : \@results;
} # End get_objects

#===============================================================================
# Function: create_objects
#   Creates ONTAP objects based on specifications and returns the
#   specifications for those created.
#
#   Parameters:
#       type - Type of objects to get.  Valid types are 'volume', 'aggregate',
#           'vserver', 'lun', 'igroup', and 'client'.
#       count - Number of objects to create.
#       specs - This contains the information needed to create the objects
#            Example: lun
#            [
#               {
#                   vserver => 'vs1',
#                   volume  => 'vol2',
#                   ostype  => 'windows',
#                   size    => '1g'
#               },
#               {
#                   vserver => 'vs2',
#                   volume  => 'vol6',
#                   ostype  => 'linux',
#                   size    => '1g'
#               },
#            ]
#       prefix - Prefix for naming the objects to be created.
#       node - Node to create objects on.
#
#   Note: if 'specs' and 'count' are both specified, the specs will be iterated
#         in a round-robin fashion until the number of objects specified in
#         'count' is reached.
#
#   Returns:
#       List of objects matching the prefix parameter.
#
#   Example:
#    try {
#        my $spec = [
#            {
#                aggregate => 'aggr_node1'
#            },
#            {
#                aggregate => 'aggr_node2'
#            }
#        ];
#        my @vs = $Resource_Obj->create_objects(
#            type   => 'vserver',
#            prefix => 'vs_2node_iscsi_',
#            specs  => $spec,
#            count  => 20
#        );
#    } otherwise {
#        my $exception = shift();
#        $Test->nlog($exception);
#        $status = FALSE;
#    };
#
#===============================================================================
sub create_objects {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $type = &ParamsMgr::get_arg(
        'arg'      => "type",
        'required' => TRUE
    );
    my $count = &ParamsMgr::get_arg(
        'arg'      => "count",
        'required' => FALSE,
        'default'  => 1,
    );
    my @specs = &ParamsMgr::get_arg(
        'arg'      => "specs",
        'required' => FALSE,
        'default'  => (),
    );
    my $prefix = &ParamsMgr::get_arg(
        'arg'      => "prefix",
        'required' => FALSE,
        'default'  => "",
    );
    my $node = &ParamsMgr::get_arg(
        'arg'      => "node",
        'required' => FALSE,
        'default'  => undef,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return wantarray() ? () : [];
    }

    # Locals
    my $timestamp = timestamp;
    my @new_specs = ();
    my @objs = ();
    my @results = ();
    my ($nacl_module, $misfit_module, $object_type, $misfit_objects) = "";
    my ($object_id) = "";
    if (defined($node)) {
        $node = NACL::C::Node->new(node => $node);
        $node->apiset_must(interface => 'ZAPI');
    } else {
        $Log->comment(
            "No node was specified for executing create $object_type, " .
            "using the default node " . $self->{default_node}->{node});
        $node = $self->{default_node};
    }

    if ($type =~ /volume/i) {
        $misfit_module = "San::FIT::Misfit::Features::VolOps";
        $nacl_module = "NACL::C::Volume";
        $object_type = "volume";
        $object_id = "volume";
    } elsif ($type =~ /lun/i) {
        $misfit_module = "San::FIT::Misfit::Features::LunOps";
        $nacl_module = "NACL::C::Lun";
        $object_type = "lun";
        $object_id = "path";
    } elsif ($type =~ /igroup/i) {
        $misfit_module = "San::FIT::Misfit::Features::Igroup";
        $nacl_module = "NACL::C::LunIgroup";
        $object_type = "igroup";
        $object_id = "igroup";
    } elsif ($type =~ /aggregate/i) {
        $misfit_module = "San::FIT::Misfit::Features::AggrOps";
        $nacl_module = "NACL::STask::Aggregate";
        $object_type = "aggregate";
        $object_id = "aggregate";
    } elsif ($type =~ /vserver/i) {
        $misfit_module = "San::FIT::Misfit::Features::VserverOps";
        $nacl_module = "NACL::STask::Vserver";
        $object_type = "vserver";
        $object_id = "vserver";
    } else {
        NATE::Exceptions::Argument->throw(
            "The value: $type passed for the parameter 'type' is not supported."
        );
        return wantarray() ? () : [];
    }

    # Increment names for each of the new objects to be created.
    my $spec_idx = 0;
    for (my $i = 0; $i < $count; $i++) {
        # Rotate through specs for each iteration of count.
        $spec_idx = $i % scalar(@specs);
        my $spec = $specs[$spec_idx];

        # Generate a prefix if one is not provided.
        if ($prefix eq "") {
            $timestamp =~ s/\s/_/g;
            $prefix = "Resource_" . $timestamp . "_";
            if (exists($spec->{$type})) {
                $prefix .= $spec->{$object_type} . "_";
            } else {
                $prefix .= $object_type . "_";
            }
        }

        my %new_spec = %{$spec};
        $new_spec{$object_type} = $prefix . $i;
        push(@new_specs, \%new_spec);
    }

    $Log->debug(
        "Created the following specs for Resources:\n" . Dumper(@new_specs)
    );
    my $misfit = $misfit_module->new();
    unless (ref($misfit) eq $misfit_module) {
        $Log->error("Failed to instantiate $misfit_module");
        $Log->exit();
        return wantarray() ? () : [];
    }
    $misfit->create(
        objects => \@new_specs,
        node    => $node
    );
    @objs = $nacl_module->find(
        command_interface => $node,
        filter => {$object_id => "*$prefix*"}
    );

    # Here we convert the objects to hash refs containing the attributes
    # we need in Misfit.
    foreach my $obj (@objs) {
        my $attrs = $self->get_object_type_and_attrs(object=>$obj);
        push(@results, $attrs->{misfit_attributes});
    }

    $Log->exit();
    return wantarray() ? @results : \@results;
} # End create_objects

#===============================================================================
# Function: get_object_type_and_attrs
#   For a given NACL object, returns the type of the object as well
#   as the unique identifier attributes.
#
#   Parameters:
#       object - NACL object to get data from.
#
#   Returns:
#       Hash with specific data gathered from NACL object.
#===============================================================================
sub get_object_type_and_attrs {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $object = &ParamsMgr::get_arg(
        'arg'      => "object",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return {};
    }

    my %standard_attrs = (
        vserver => $object->{vserver}
    );

    my $type_attrs = {};
    if (ref($object) =~ /Lun$/) {
        $type_attrs = {
            type => 'lun',
            attributes => {
                %standard_attrs,
                path  => $object->{path}
            },
            misfit_attributes => {
                %standard_attrs,
                path => $object->{path},
                %{$self->_decode_lun_path(path => $object->{path})}
            }
        };
    } elsif (ref($object) =~ /Volume$/ ){
        $type_attrs = {
            type => 'volume',
            attributes => {
                %standard_attrs,
                'volume'  => $object->{volume}
            },
            misfit_attributes => {
                %standard_attrs,
                volume    => $object->{volume},
                aggregate => $object->{aggregate},
                size      => $object->{size},
                type      => $object->{type},
            }
        };
    } elsif (ref($object) =~ /Igroup$/) {
        $type_attrs = {
            type => 'igroup',
            attributes => {
                %standard_attrs,
                igroup => $object->{igroup}
            },
            misfit_attributes => {
                %standard_attrs,
                igroup => $object->{igroup},
                initiators => $object->{initiators},
                igroup  => $object->{igroup},
                ostype  => $object->{ostype},
                protocol => $object->{protocol}
            }
        };
    } elsif (ref($object) =~ /Vserver$/) {
        $type_attrs = {
            type => 'vserver',
            attributes => {
                %standard_attrs
            },
            misfit_attributes => {
                %standard_attrs
            }
        };
    } elsif (ref($object) =~ /Aggregate$/) {
        $type_attrs = {
            type => 'aggregate',
            attributes => {
                aggregate => $object->{aggregate}
            },
            misfit_attributes => {
                aggregate => $object->{aggregate},
                nodes => $object->{node},
                availsize => $object->{availsize},
                owner_name => $object->{owner_name},
                size      => $object->{size},
            }
        };
    } else {
        NATE::Exceptions::Argument->throw(
            "The object type: \"" . ref($object)
            . "\" is not currently handled."
            . Dumper($object)
        );
    }

    $Log->exit();
    return $type_attrs;
} # End get_object_type_and_attrs

#===============================================================================
# Function: destroy_objects
#   Destroy objects.
#
#   Parameters:
#       objects - List of objects to destroy.
#
#   Returns:
#       TRUE
#===============================================================================
sub destroy_objects {
    my ($self,%args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my @objects = &ParamsMgr::get_arg(
        'arg'      => "objects",
        'required' => TRUE
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    foreach my $obj (@objects) {
        my $attrs = $self->get_object_type_and_attrs(object => $obj);
        $attrs->{attributes}->{command_interface} = $obj->{command_interface};
        my $nacl_module = ref($obj);
        # If the user passed and STask object, call purge directly,
        # otherwise, we will call the STask purge as a class method, and pass
        # the required params.
        if ($nacl_module =~ /STask/) {
            $Log->debug(
                "User passed STask object: " . ref($obj) .
                ", calling purge directly."
            );
            $obj->purge();
        } else {
            $Log->debug(
                "User passed a non-STask object: " . ref($obj) .
                ", calling the STask class method for purge."
            );
            $nacl_module =~ s/::C::|::CS::/::STask::/;
            $nacl_module->purge(%{$attrs->{attributes}});
        }
    }

    $Log->exit();
    return TRUE;
} # End destroy_objects

#===============================================================================
# Function: reserve
#   Function for reserving volumes from the cluster for specific
#   operation.
#
#   Parameters:
#       total - Total number of objects to reserve.  Depending on the
#           balanced_distribution parameter, this will gather objects in a
#           round robin fashion per node, the first available objects.
#       type - Type of objects to get.  Valid types are 'volume',
#           'aggregate', 'vserver', 'lun', 'igroup', and 'client'.
#       max_per_node - Maximum number of objects to reserve per node.
#       filter - TODO
#       balanced_distribution - If TRUE, reserve will reserve resources in
#           a round robin fashion based on the node.  If FALSE, it will reserve
#           the first available.
#
#       NOTE!! total and max_per_node are mutually exclusive parameters.
#           max_per_node takes priority over total.
#
#   Returns:
#       List of reserved objects.
#===============================================================================
sub reserve {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $total_objs = &ParamsMgr::get_arg(
        'arg'      => "total",
        'required' => FALSE,
        'default'  => undef,
    );
    my $type = &ParamsMgr::get_arg(
        'arg'      => "type",
        'required' => TRUE
    );
    my $percent = &ParamsMgr::get_arg(
        'arg'      => "percent",
        'required' => FALSE,
        'default'  => undef,
    );
    my $max_per_node = &ParamsMgr::get_arg(
        'arg'      => "max_per_node",
        'required' => FALSE,
        'default'  => undef,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => undef,
    );
    my $balanced = &ParamsMgr::get_arg(
        'arg'      => "balanced_distribution",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $create_dests = &ParamsMgr::get_arg(
        'arg'      => "create_dests",
        'required' => FALSE,
        'default'  => FALSE,
    );
    my $dst_data = &ParamsMgr::get_arg(
        'arg'      => "dst_data",
        'required' => FALSE,
        'default'  => undef,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return wantarray ? () : [];
    }

    $Log->comment('Reserving ' . $type . ' for feature');

    # Make sure resource list exists
    if (!(defined($self->{resources}) and exists($self->{resources}))) {
        $self->_gather_resources();
    }

    # Set node filter
    my $node_filter = '*';
    if (defined($filter)) {
        $node_filter = $filter->{node};
    }

    my $cur_obj = 0;
    my %node = ();
    my @objects = ();
    my $total_type_count = $self->{'num_' . $type};

    # XXX: For now, treat clients as get_objects (no reservation). Fix this
    # once we have everything working and the need for reserving clients
    if ($type =~ /client/i) {
        return wantarray ? @{$self->{client_device_list}} : $self->{client_device_list};
    }

    if (defined($percent)) {
        $total_objs = int($self->{'total_' . $type} * ($percent/100));
        $max_per_node = undef;
    }

    # Gather objects per node
    if (defined($max_per_node)) {
        $Log->comment('Reserving ' . $max_per_node . ' ' . $type . 's per node');
        foreach my $cur_node (@{$self->{resources}}) {
            for(my $i = 0; $i < $max_per_node; $i++) {
                $Log->debug(Dumper($cur_node));
                last if (scalar(@{$cur_node->{$type}}) == 0);
                my $object = pop(@{$cur_node->{$type}});
                if (defined($object)) {
                    push(@objects, $object);
                    $self->{'num_' . $type}--;
                }
            }
        }
    } elsif (defined($total_objs)) {
        if ($balanced){
            my $cur_node = 0;
            # Gather available objects across nodes in cluster
            $Log->comment('Reserving ' . $total_objs . ' ' . $type . '(s) across the cluster...');
            while (scalar(@objects) < $total_objs && $cur_obj < $total_type_count) {
                $Log->debug($self->{'num_' . $type} . " $type(s) remaining.");
                last if ($self->{'num_' . $type} == 0);
                my $object = undef;
                my $index = $cur_node % scalar(@{$self->{resources}});
                my @node_info = @{$self->{resources}};
                $Log->debug(Dumper($node_info[$index]->{$type}));
                my $object_index = 0;
                foreach my $obj (@{$node_info[$index]->{$type}}) {
                    my $match = $self->_filter_check(
                        object => $obj,
                        filter => $filter
                    );
                    if ($match) {
                        $object = splice(@{$node_info[$index]->{$type}},$object_index,1);
                        last;
                    }
                    $object_index++;
                    $cur_obj++;
                }

                if (defined($object)){
                    push(@objects, $object);
                    $self->{'num_' . $type}--;
                    $Log->debug('adding ' . Dumper($object));
                }
                $cur_node++;
            }
        } else {
            # Find group of objects on one node.  If filter->{node} is defined
            # only return objects from that node.
            # The first attempt at this tries to find a node that can hold the
            # the total number of volumes requested.  If there aren't enough
            # left, then just pick one node with volumes left to give
            foreach my $cur_node (@{$self->{resources}}) {
                next if (!($cur_node->{node} =~ /$node_filter/));
                next if (scalar(@{$cur_node->{$type}}) < $total_objs);
                for (my $i = 0; $i < $total_objs; $i++) {
                    my $object = pop(@{$cur_node->{$type}});
                    if (defined($object)) {
                        push(@objects, $object);
                        $self->{'num_' . $type}--;
                        $Log->debug('adding ' . Dumper($object));
                    }
                }
                last if ($cur_node->{node} =~ /$node_filter/);
            }
            if(scalar(@objects) == 0){
                foreach my $cur_node (@{$self->{resources}}) {
                    next if (!($cur_node->{node} =~ /$node_filter/));
                    next if (scalar(@{$cur_node->{$type}}) <= 0);
                    for (my $i = 0; $i < scalar(@{$cur_node->{$type}}); $i++) {
                        my $object = pop(@{$cur_node->{$type}});
                        if (defined($object)) {
                            push(@objects, $object);
                            $self->{'num_' . $type}--;
                            $Log->debug('adding ' . Dumper($object));
                        }
                    }
                    last if ($cur_node->{node} =~ /$node_filter/);
                }
            }
        }
    }

    if ($create_dests) {
        my %return_objects = ();
        $return_objects{source_volumes} = \@objects;
        $return_objects{destination_volumes} = $self->create_destination_volumes(
            %{$dst_data},
            source_vols => \@objects
        );
        return \%return_objects;
    }

    return wantarray ? @objects : \@objects;
} # End reserve

#===============================================================================
# Function: get_resources
#   Function for getting resources from the cluster based on certain filters.
#   Currently ignores any outstanding reservations as it uses a different list.
#
#   Parameters:
#       type - (Required) Type of objects to retrieve.  Valid types are
#           'volume', 'aggregate', 'vserver', 'lun', 'igroup', and 'client'.
#       total - (Optional, Default: Undef - all available objects)
#           Total number of resources to retrieve.  Depending on the
#           balanced_distribution parameter, this will gather resources in a
#           round robin fashion per node, or the first available resources.
#       percent - (Optional, Default: 100%)
#           This will return a percentage of the total number
#           of the type of object requested that does not exceed the
#           requested percentage.  If there are less volumes the
#           percentage will be less.
#       max_per_node - (Optional, Default: Undef - unlimited)
#           Maximum number of resources to retrieve per node.
#           Note: this can limit total count.
#       max_per_vserver - (Optional, Default: Undef - unlimited)
#           Maximum number of resources to retrieve per vserver.
#           Note: this can limit total count.
#       max_vservers - (Optional, Default: Undef - unlimited)
#           The number of vservers to use when retreiving resources.
#           Note: this can limit total count.
#       balanced_distribution - (Optional, Default: TRUE)
#           If TRUE, the method will retrieve resources in
#           a round robin fashion based on the node.  If FALSE, it will reserve
#           the first available.
#       filter - (Optional, Default: Undef - unlimited)
#           Operates on the fields returned by Misfit.  Not all
#           fields are returned by Misfit.  Only if all of the fields in
#           filter exist AND match the fields returned, will the object
#           will be returned.
#
#   Returns:
#       List of objects.
#===============================================================================
sub get_resources {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $total_objs = &ParamsMgr::get_arg(
        'arg'      => "total",
        'required' => FALSE,
        'default'  => undef,
    );
    my $type = &ParamsMgr::get_arg(
        'arg'      => "type",
        'required' => TRUE
    );
    my $percent = &ParamsMgr::get_arg(
        'arg'      => "percent",
        'required' => FALSE,
        'default'  => MAX_PERCENT_RESOURCE_GET,
    );
    my $max_per_node = &ParamsMgr::get_arg(
        'arg'      => "max_per_node",
        'required' => FALSE,
        'default'  => undef,
    );
    my $max_per_vserver = &ParamsMgr::get_arg(
        'arg'      => "max_per_vserver",
        'required' => FALSE,
        'default'  => undef,
    );
    my $max_vservers = &ParamsMgr::get_arg(
        'arg'      => "max_vservers",
        'required' => FALSE,
        'default'  => undef,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => undef,
    );
    my $balanced = &ParamsMgr::get_arg(
        'arg'      => "balanced_distribution",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $create_dests = &ParamsMgr::get_arg(
        'arg'      => "create_dests",
        'required' => FALSE,
        'default'  => FALSE,
    );
    my $dst_data = &ParamsMgr::get_arg(
        'arg'      => "dst_data",
        'required' => FALSE,
        'default'  => undef,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return wantarray ? () : [];
    }

    $Log->comment(
        'Retrieving resources using the following options: ' .
        Dumper(\%args)
    );

    # Check to make sure the user did not pass 0's for any of the fields
    if (defined($max_vservers)) {
        return wantarray ? () : [] unless ($max_vservers > 0);
    }
    if (defined($max_per_vserver)) {
        return wantarray ? () : [] unless ($max_per_vserver > 0);
    }
    if (defined($max_per_node)) {
        return wantarray ? () : [] unless ($max_per_node > 0);
    }

    # Make sure resource list exists
    if (!(defined($self->{resources}) and exists($self->{resources}))) {
        $self->_gather_resources();
    }
    # Create a clone of the global resource list so we can make changes to the
    # structure, without corrupting the original.
    my %resources = %{Clone::clone($self->{resources_node_vs})};

    # XXX: For now, treat clients as get_objects (no reservation). Fix this
    # once we have everything working and the need for reserving clients
    if ($type =~ /client/i) {
        return wantarray ? @{$self->{client_device_list}} : $self->{client_device_list};
    }

    # If the user specified a node or vserver in the filter, remove it
    # and shorten the list of nodes/vservers to iterate in the resource
    # hash.  The resources are keyed with {node}->{vserver} so we can look at
    # just those references.  Delete the key as it is no longer needed for the
    # filter check later.
    my @nodes = keys(%resources);
    my $vserver_filter = undef;
    if (defined($filter)) {
        @nodes = delete($filter->{node}) if defined($filter->{node});
        $vserver_filter = delete($filter->{vserver}) if defined($filter->{vserver});
    }

    # If the user specifies percent use the number of the objects currently
    # NOT reserved, otherwise use all the objects currently NOT reserved.
    $total_objs = int($self->{'num_' . $type} * ($percent/100));

    # Array holding the resources to be returned from the method
    my @objects = ();
    # Nodes currently returning resources
    my %used_node = ();
    # Vservers currently returning resources
    my %used_vs = ();

    # Keep looping while found objects are less than total object count AND
    # there are nodes.
    my $need_more_objects = TRUE;
    while ($need_more_objects) {
        $need_more_objects = scalar(@objects) < $total_objs && scalar(@nodes);
        $Log->debug(
            "Requested: " . scalar(@objects) . " of $total_objs\n" .
            "Nodes:     " . scalar(@nodes) . "\n"
        );

        foreach my $node (@nodes) {
            $Log->debug("Current used_node: " . Dumper(\%used_node));
            # If we've reached the maximum resources on this node we need to
            # remove the node from the list of nodes and increment the analyzed
            # counter.
            if (defined($max_per_node)) {
                if ($used_node{$node} >= $max_per_node) {
                    $Log->debug("Max per node has been reached for $node");
                    my $node_id = first_index {$_ eq $node} @nodes;
                    $node = splice(@nodes, $node_id, 1) if defined($node_id);
                    $Log->debug(
                        "Max per node limit was reached so, removed $node " .
                        "from list of nodes to search."
                    );
                    last;
                }
            }

            # If there is a vserver in filter, remove any vservers not matching
            # that filter.  If that vserver is not found then it means this
            # vserver is not eligible for this node, so remove the node as well.
            my @vservers = keys(%{$resources{$node}});
            if (defined($vserver_filter)) {
                my $vs_regex = $vserver_filter;
                # Remove wildcards
                $vs_regex =~ s/\*//g;
                if ($vs_regex =~ s/!//) {
                    @vservers = grep(!/$vs_regex/, @vservers);
                } else {
                    @vservers = grep(/$vs_regex/, @vservers);
                }
                unless (scalar(@vservers)) {
                    my $node_id = first_index {$_ eq $node} @nodes;
                    $node = splice(@nodes, $node_id, 1) if defined($node_id);
                    $Log->debug(
                        "Vserver: $vserver_filter was specified as a filter, " .
                        "but this vserver has no $type resources on $node. " .
                        "So, removed $node from list of nodes to search."
                    );
                    last;
                }
            }

            # If we've reached the max vservers to get resources from,
            # remove any vservsers that have not already been used.
            if (defined($max_vservers)) {
                if (scalar(keys(%used_vs)) >= $max_vservers) {
                    my @temp = ();
                    foreach my $vs (@vservers) {
                        push(@temp, $vs) if ($used_vs{$vs});
                    }
                    @vservers = @temp;
                    $Log->debug(
                        "Max vservers limit was reached, so only the " .
                        "following vservers will continue to be used:\n" .
                        join(',', @vservers)
                    );
                }
            }

            # Remove any vservers that have been used the max amount of times.
            if (defined($max_per_vserver)) {
                my @temp = ();
                foreach my $vs (@vservers) {
                    if (defined($used_vs{$vs})) {
                        push(@temp, $vs) if ($used_vs{$vs} < $max_per_vserver);
                    } else {
                        push(@temp, $vs);
                    }
                }
                @vservers = @temp;
                $Log->debug(
                    "The the following vservers have not reached " .
                    "the max_per_vserver_limit:\n". join(', ', @vservers)
                );
            }

            # Remove any vservers that have no resources left.
            my @temp = ();
            foreach my $vs (@vservers) {
                push(@temp, $vs) if (scalar(@{$resources{$node}{$vs}{$type}}));
            }
            @vservers = @temp;
            $Log->debug(
                "The the following vservers have resources left on $node:\n"
                . join(', ', @vservers)
            );

            # If there are no eligible vservers left for this node, remove
            # the node from the node list so we dont look here again.
            unless (scalar(@vservers)) {
                $Log->debug(
                    "There are no vservers to search for $type (s) on $node."
                );
                my $node_id = first_index {$_ eq $node} @nodes;
                $node = splice(@nodes, $node_id, 1) if defined($node_id);
                $Log->debug("So, removed $node from list of nodes to search.");
                last;
            }

            # Now that we've reduced the loops by nodes and vservers look at
            # what is left.
            foreach my $vs (@vservers) {
                # Make sure we need more objects before continuing...
                last unless $need_more_objects;
                # Reset the object reference to determine if a new object is
                # found.
                my $obj = undef;

                # We only need to iterate objects if there are keys remaining
                # in filter, otherwise we can just grab groups from the list.
                if (keys(%{$filter})) {

                    # Return the first object index that matches the filter,
                    # returns -1 if no match.
                    my $match_id = first_index
                        {
                            $self->_filter_check(
                                object => $_,
                                filter => $filter
                            );
                        } @{$resources{$node}{$vs}{$type}};
                    if ($match_id > -1) {
                        $obj = splice(@{$resources{$node}{$vs}{$type}},$match_id,1);
                    } else {
                        # None of the objects match so empty the list.
                        $Log->debug(
                            "None of the resources in $vs match the requested" .
                            " filter options: " . Dumper($filter)
                        );
                        @{$resources{$node}{$vs}{$type}} = ();
                    }
                } else {
                    # No filter was used, so we can just take the first object
                    # from this node->vserver.
                    $obj = pop(@{$resources{$node}{$vs}{$type}});
                } # End if filter else

                # If we found an object it will be defined now.
                if (defined($obj)) {
                    $Log->debug("Add filtered resource: " . Dumper($obj));
                    push (@objects, $obj);
                    # Initialize/increment the node and vserver seen counters.
                    $used_vs{$vs}++;
                    $used_node{$node}++;
                    $need_more_objects = scalar(@objects) < $total_objs && scalar(@nodes);

                    # If the user requested balanced for nodes, we need to exit
                    # the vservers for this node and move to the next node.
                    # Else we can just move to the next vserver.
                    if ($balanced) {
                        $Log->debug("Balanced specified so move to next node.");
                        last;
                    } else {
                        $Log->debug(
                            "Balanced NOT specified to move to next vserver."
                        );
                        next;
                    }
                }
            } # End foreach vserver
        } # End foreach node
    }

    return wantarray ? @objects : \@objects;
} # End get_resources

#===============================================================================
# Function: create_destination_volumes
#   Operations such as Vault and Snapmirror require destination volumes for
#   their successful operation.  This function takes a list of source volumes
#   and creates a destination vserver, destination volumes, and vserver peering.
#
#   Parameters:
#       vserver - Name of the destination vserver to be created.
#       source_vols - List of source volumes.
#       dest_cluster - NACL object for destination cluster
#
#   Returns:
#       List of destination volumes.
#===============================================================================
sub create_destination_volumes {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $source_vols = &ParamsMgr::get_arg(
        'arg'      => "source_vols",
        'required' => FALSE,
        'default'  => [],
    );
    my $node_name = &ParamsMgr::get_arg(
        'arg'      => "cluster_name",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return wantarray ? () : [];
    }

    my @src_vols = @{$source_vols};
    my @cluster_tree = ();
    my %vserver_peers = ();
    my $index = 0;
    my %src_tree = ();
    my $cluster = $self->{cluster_peer};

    # Create new volops object to create volumes on dest cluster
    my $volops = San::FIT::Misfit::Features::VolOps->new(
        cluster => $node_name
    );



    # Get list of aggregates for each node in dest cluster to pass to volume
    # create.
    foreach my $node ($self->{cluster_peer_cli}->nodes()) {
        # Gather aggregates for that node.
        $cluster_tree[$index] = [];
        my @aggrs = NACL::CS::Aggregate->fetch(
            command_interface  => $self->{cluster_peer},
            filter => {
                'ha-policy' => 'sfo',
                nodes       => [$node],
            },
            requested_fields => ['aggregate'],
            allow_empty      => TRUE,
        );
        unless(scalar(@aggrs)) {
            $Log->comment("No aggrs for $node");
            next;
        }
        foreach my $aggr (@aggrs) {
            push(@{$cluster_tree[$index]}, $aggr->{aggregate});
        }
        $index++;
    }

    # Create destination vserver
    $Log->comment("Creating vserver for feature");
    try {
        NACL::STask::Vserver->create(
            command_interface  => $self->{cluster_peer},
            vserver            => $vserver,
            rootvolume         => $vserver . '_rootvol',
            nacltask_if_exists => 'reuse',
        );
    } otherwise {
        my $exception = shift;
        $Log->comment($exception->text());
        return wantarray ? () : [];
    };

    # Create data structure for src_vols to distribute destinations
    # across each node in the cluster
    foreach my $src (@src_vols) {
        if (!defined($src_tree{$src->{node}})) {
            $src_tree{$src->{node}} = [];
        }
        push(@{$src_tree{$src->{node}}}, $src);
    }

    my @dst_vols = ();
    $self->{peers} = [];

    # Loop through each node and create a destination volume for each src_vol
    # reserved for the feature.  Also peer vserver if not already done.
    while (my ($node, $src_volumes) = each(%src_tree)) {
        my $i = 0;
        foreach my $src (@{$src_volumes}) {
            my $node_index = $i % scalar(@cluster_tree);
            my @dst_aggrs = @{$cluster_tree[$node_index]};
            foreach my $dst_aggr (@dst_aggrs) {
                push(@dst_vols, {
                    volume             => $src->{volume} . '_dp',
                    aggregate          => $dst_aggr,
                    size               => $src->{size},
                    nacltask_if_exists => 'purge',
                    type               => 'DP',
                    vserver            => $vserver,
                });

                if (grep(/$src->{vserver}/, @{$self->{peers}})) {
                    $Log->comment(
                        'Skipping ' . $src->{vserver} . '. Already peered.'
                    );
                    last;
                }
                my $result = $self->_create_vserver_peer(
                    vserver      => $src->{vserver},
                    peer_vserver => $vserver,
                );
                push(@{$self->{peers}}, $src->{vserver});
                last;
            }
            $i++;
        }
    }

    my $status = FALSE;

    try {
        $status = $volops->create(
            objects => \@dst_vols
        );
    } otherwise {
        my $exception = shift;
        $Log->comment($exception->text());
    };

    if ($status) {
        return wantarray ? @dst_vols : \@dst_vols;
    } else {
        return wantarray ? () : [];
    }

} # End of create_destination_volumes

#===============================================================================
# Private Methods
#===============================================================================

#===============================================================================
# Description: Operate on an xml with SanMgr or EnGenX (future) with and XML or
# resource_list (future)
#===============================================================================
sub _operate {
    my ($self, %args) = @_;

    $Log->enter();
    &ParamsMgr::process_args(%args);
    my @xml_files = &ParamsMgr::get_arg(
        'arg'      => "xml_files",
        'required' => TRUE,
    );
    my @config_files = &ParamsMgr::get_arg(
        'arg'      => "config_files",
        'required' => FALSE,
        'default'  => [],
    );
    my $operation = &ParamsMgr::get_arg(
        'arg'      => "operation",
        'required' => TRUE,
    );
    my $background = &ParamsMgr::get_arg(
        'arg'      => "background",
        'required' => FALSE,
        'default'  => TRUE,
    );
    my $id = &ParamsMgr::get_arg(
        'arg'      => "id",
        'required' => FALSE,
        'default'  => $operation,
    );
    my $intercluster = &ParamsMgr::get_arg(
        'arg'      => "intercluster",
        'required' => FALSE,
        'default'  => FALSE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    # This array contains all of the info for individual NATE::Process specs.
    my @process_info = ();
    my @intercluster_nodes = ();
    my $status = TRUE;

    for (my $index = 0; $index < scalar(@xml_files); $index++) {
        $Log->comment(
            "Creating NATE::Process for $xml_files[$index] and $config_files[$index]"
        );
        my $config_file_params = $self->_encode_config(
            config_file => $config_files[$index]
        );

        my $runid = $config_file_params->{CLUSTER_NAME} . "_" .
                    (split('/',$xml_files[$index]))[-1] . "_" .
                    (split('/',$config_files[$index]))[-1];

        push(@process_info,
            {
                codespec    => "{" . $self->{id} . "::_sanmgr(); }",
                params      => {
                    self        => $self,
                    operation   => $operation,
                    xml_file    => $xml_files[$index],
                    debug_level => $self->{debug_level},
                    %{$config_file_params}
                },
                runid       => $runid,
                background  => $background,
                inherit_env => TRUE,
                inherit_params => TRUE,
            }
        );
    }

    # For some reason SanMgr converts it's run_ids from "::" to "__", so I'm
    # going to conform as well because I'm tired of having 2 folders per
    # workflow.
    my $workflow_id = $id;
    $workflow_id =~ s/::/__/g;
    $self->{workflow} = San::FIT::Misfit::Core::Workflow->new(
        steps      => \@process_info,
        iterations => 1,
        runid      => $workflow_id,
        background => $background,
    );

    try {
        $self->{workflow}->start();
        $self->{workflow}->wait();
        my @results = $self->{workflow}->{workflow}->own_results();
        $Log->debug("Results from _sanmgr: \n" . Dumper(@results));
        foreach my $result (@results) {
            if (defined($result)) {
                if ($result->type() =~ /pass/i) {
                    my $json = Compress::Zlib::uncompress($result->misc());
                    my $device = JSON::from_json($json);
                    $Log->debug("Client devices are: " . Dumper($device));
                    if (keys(%{$device})) {
                        $Log->comment('adding device: ' . $device);
                        push(@{$self->{client_device_list}}, $device);
                    }
                } else {
                    $status = FALSE;
                }
            } else {
                $Log->comment("Did not receive any results from _sanmgr.");
                $status = FALSE;
            }
        }
        $self->{workflow}->destroy();
    } otherwise {
        my $exception = shift;
        $exception->throw();
        $status = FALSE;
    };

    unless ($status) {
        NATE::BaseException->throw("Failed configuring testbed.");
        return FALSE;
    }

    if ($intercluster) {
        unless (scalar(@{$self->{interclustered_nodes}}) > 1) {
            NATE::BaseException->throw("Intercluster requires at least two clusters");
            return FALSE;
        }
        my %intercluster_params = (
            home_node   => $self->{interclustered_nodes}->[0],
            remote_node => $self->{interclustered_nodes}->[1]
        );
        if ($operation eq "setup") {
            $self->_initialize_intercluster(%intercluster_params);
        } elsif ($operation eq "cleanup") {
            $self->_uninitialize_intercluster(%intercluster_params);
        }
    }

    # If setting up, we need to create structure that contains all objects in
    # the cluster
    if($operation eq CONFIGURE_METHOD_NAME){
        $self->_gather_resources();
    }

    $Log->exit();
    return $status;
} # end _operate

#===============================================================================
# Description: The function of this method is to setup and cleanup
# SanMgr XML files. It should be called from NATE::Process or
# Misfit::Core::Workflow.
#===============================================================================
sub _sanmgr {

    my $global_params      = NATE::ParamSet->get_global();
    my $self               = $global_params->get('self');
    my $operation          = $global_params->get('operation');
    my $xml_file           = $global_params->get('xml_file');
    my $debug_level        = $global_params->get('debug_level');
    my $client_device_list = {};
    my @client_objs        = ();

    # Create a SanMgr object
    my $sanmgr_obj = San::SanMgr->new(
        -xml_file    => $xml_file,
        -debug_level => $debug_level
    );
    if ($sanmgr_obj->error()) {
        $Log->error(
            'Failed to create a SanMgr object: ' . $sanmgr_obj->errmsg()
        );
        return NATE::Result::Fail->new($sanmgr_obj->errmsg());
    }
    $sanmgr_obj->$operation();
    if ($sanmgr_obj->error()) {
        NATE::BaseException->throw(
            'SanMgr Error is un-recoverable: ' . $sanmgr_obj->errmsg()
        );
        return NATE::Result::Fail->new($sanmgr_obj->errmsg());
    }

    # Parse out the client dev list if it is stored on the sanmgr object
    $client_device_list = $sanmgr_obj->get_client_dev_list();
    @client_objs = $sanmgr_obj->get_objects( -type => "client");
    if ($sanmgr_obj->error()) {
        if ($sanmgr_obj->errmsg() =~ /Client device list is not defined!/) {
            $Log->comment("Client Device List was not defined.");
        } else {
            $Log->log(
                NATE::Result::Fail->new(
                    "Failed to get the list of devices from clients:" .
                    $sanmgr_obj->errmsg()
                )
            );
        }
    }

    # For each client, retrieve only the paths to the devices from their
    # device lists.
    my $dev_data = {};
    foreach my $client (keys %{$client_device_list}) {
        my @dev_paths = ();
        foreach my $cdl_entry (@{$client_device_list->{$client}}) {
            my $path = $cdl_entry->{path};
            push (@dev_paths, $cdl_entry->{path});
        }

        $Log->comment("Device Paths for $client:", Dumper(\@dev_paths));
        $dev_data->{$client} = \@dev_paths;
    }

    # In the event client setup was skipped, device lists will not contain that
    # client, find any clients not in the list and add them with an empty device
    # list
    foreach my $client (@client_objs) {
        my $client_name = $client->{host};
        if (!defined($dev_data->{$client_name})) {
            $Log->comment("No Device Paths for $client_name, added empty list");
            $dev_data->{$client_name} = [];
        }
    }

    # There appears to be a truncation in NATE at 65535 bytes for the misc
    # attribute, we are compressing the JSON string to send all of the data.
    my $compressed_data = Compress::Zlib::compress(JSON::to_json($dev_data));
    my $result_obj = NATE::Result::Pass->new("SanMgr Completed Successfully");
    $result_obj->misc($compressed_data);
    $Log->log($result_obj);
} # End _sanmgr

#===============================================================================
# Description: The function of this method is to read in a specified config file
# and convert to a hash of parameters to use for NATE::Process param attributes.
#===============================================================================
sub _encode_config {
    my ($self, %args) = @_;

    # Locals
    my %params;
    my $lines = ();
    my $config_file = $args{config_file};

    eval {
        open (FILE, "<", $config_file) or die "Cannot Open File: $config_file";
        local $/;
        $lines = <FILE>;
        close(FILE);
    };
    if ($@) {
        NATE::Exceptions::Argument->throw(
            "Config File Does NOT Exist: \"$config_file\""
        );
        return \%params;
    }
    foreach my $line (split('\n',$lines)) {
        # Skip any lines without an '='
        unless ($line =~ '=') {
            next;
        }
        # Skip any commented lines
        if ($line =~ /^#/) {
            next;
        }
        # Remove trailing spaces and everything after
        $line =~ s/\s.*$//;
        # Store the first node to use for cluster peering operations
        if ($line =~ /FIT_NODE_1=(.*)/i) {
            push(@{$self->{interclustered_nodes}}, $1);
        }
        # Split the line on the equal
        my ($param,$value) = split('=',$line);

        # Check for variable references and resolve referenced lists
        if ($value =~ /^\$/) {
            $params{$param} = '';

            # Remove '$' from string and split value into array
            $value =~ s/\$//g;
            my @vars = split(':', $value);
            # Loop through variable references and add values to params
            # Will append current list for $param
            foreach my $key (@vars) {
                if ($params{$param} eq '') {
                    $params{$param} = $params{$key};
                } else {
                    $params{$param} = $params{$param} . ':' . $params{$key};
                }
            }
        } else {
            $params{$param} = $value;
        }
    }

    return \%params;
} # End _encode_config

#===============================================================================
# Description: Initialize intercluster connections by creating LIFs if needed
# and creating peer clusters.
#===============================================================================
sub _initialize_intercluster {
    my ($self,%args) = @_;

    my $home_node = $args{home_node};
    my $remote_node = $args{remote_node};

    my $home_cluster_interface = NACL::C::Node->new(node=>$home_node);
    my $remote_cluster_interface = NACL::C::Node->new(node=>$remote_node);
    my $was_created = FALSE;
    my $clusterpeer = undef;

    try {
        require NACL::MTask::ClusterPeer;
        $clusterpeer = NACL::MTask::ClusterPeer->create(
              home_cluster_interface   => $home_cluster_interface,
              remote_cluster_interface => $remote_cluster_interface,
              nacltask_if_exists_peer  => 'reuse',        # default 'die'
              _was_created             => \$was_created,   # optional
              timeout                  => INTERCLUSTER_TIMEOUT,
        );
    } otherwise {
        my $e = shift();
        if ($e->text() =~ /Host .+ is unreachable/) {
            $Log->comment(
                "The clusters cannot communicate over their intercluster lifs" .
                " attempting to create routes for those lifs."
            );
            _create_intercluster_routes(%args);
            # Retry the create command here, if it fails this time, we are out
            # of luck.
            $clusterpeer = NACL::MTask::ClusterPeer->create(
                home_cluster_interface   => $home_cluster_interface,
                remote_cluster_interface => $remote_cluster_interface,
                nacltask_if_exists_peer  => 'reuse',        # default 'die'
                _was_created             =>  \$was_created,   # optional
            );
        } else {
            $e->throw();
        }
    };

    return $clusterpeer;
} # End of _initialize_intercluster

#===============================================================================
# Description: Create routes required for Intercluster LIFs.
#===============================================================================
sub _create_intercluster_routes {
    my (%args) = @_;

    require NACL::C::NetworkRoutingGroups;
    require NACL::C::NetworkRoutingGroupsRoute;
    my $home_node = $args{home_node};
    my $remote_node = $args{remote_node};
    my $home_cluster = NACL::C::Cluster->new(node => $home_node);
    my $remote_cluster = NACL::C::Cluster->new(node => $remote_node);
    my @nodes = $home_cluster->nodes();
    push(@nodes, $remote_cluster->nodes());

    foreach my $node_name (@nodes) {
        # NOTE: NOT forcing ZAPI here because the RoutingGroup requires CLI
        my $node = NACL::C::Node->new(node => $node_name);

        # This function collects the LIF information from the tharnhost files.
        my $host_lif_info = NACL::STask::Node->get_hostrec_lif_info(
            command_interface  => $node,
            roles => ["INTERCLUSTER"]
        );
        $Log->comment("Host lif info:\n" . Dumper($host_lif_info));

        # There should be only one lif per node, but we only need a single
        # route anyway, so grab the first from the list.
        my $route_info = $host_lif_info->{INTERCLUSTER}->[0];

        # Attempt to get the intercluster Routing group that should have been
        # created with the LIF create.
        # NOTE: this function requires the command interface to be CLI
        my $routing_group = NACL::C::NetworkRoutingGroups->find(
            command_interface => $node,
            filter => {
                vserver => $node_name,
                role    => 'intercluster'
            }
        );
        $Log->comment(
            "Creating Intercluster Route: ".
            " vserver: $node_name \n" .
            " routing-group: " . $routing_group->{routing_group} . "\n" .
            " gateway: " . $route_info->{GATEWAY}
        );

        # Finally we will create the route for this node.
        try {
            my $route = NACL::C::NetworkRoutingGroupsRoute->create(
                command_interface  => $node,
                vserver            => $node_name,
                'routing-group'    => $routing_group->{routing_group},
                gateway            => $route_info->{GATEWAY},
                destination        => '0.0.0.0/0'
            );
        } otherwise {
            my $e = shift();
            if ($e->text() =~ /duplicate entry/) {
                # Don't care, we want to reuse. BTW, the STask supports this,
                # but it is acting weird, so we are going to use Component for
                # now.
            } else {
                $e->throw();
            }
        };
    }

} # End _create_intercluster_routes

#===============================================================================
# Description: Initialize intercluster connections by creating LIFs if needed
# and creating peer clusters.
#===============================================================================
sub _uninitialize_intercluster {
    my ($self,%args) = @_;

    my $home_node = $args{home_node};
    my $remote_node = $args{remote_node};

    my $home_cluster_interface = NACL::C::Node->new(node => $home_node);
    my $remote_cluster_interface = NACL::C::Node->new(node => $remote_node);
    my $home_cluster = NACL::C::Cluster->new(node => $home_node);
    my $remote_cluster = NACL::C::Cluster->new(node => $remote_node);

    require NACL::MTask::ClusterPeer;
    NACL::MTask::ClusterPeer->purge(
        home_cluster_interface => $home_cluster_interface,
        remote_cluster_interface => $remote_cluster_interface,
        unique_key_for_home_cluster => { 'cluster' => $remote_cluster->{name} },
        unique_key_for_remote_cluster => { 'cluster' => $home_cluster->{name} },
    );

    return TRUE;
} # End _initialize_intercluster

#===============================================================================
# Description: Initialize intercluster connections by creating LIFs if needed
# and creating peer clusters.
#===============================================================================
sub _decode_lun_path {
    my ($self,%args) = @_;

    my ($empty,$prefix,$volume,$lun) = split('/',$args{path});

    if (length($volume) && length($lun)) {
        return {
            volume => $volume,
            lun    => $lun
        };
    }

    return {};
} # End _decode_lun_path

#===============================================================================
# Description: Create table of volumes, aggregates, and vservers by node after
# setup
#===============================================================================
sub _gather_resources {
    my ($self, %args) = @_;

    # This is a very chatty method, this call will redirect all the output
    # to a seperate file to clear the "crud".
    $Log->redirect(
        path => File::Spec->catdir(
            Tharn::param('TOP_LOGDIR'),
            Tharn::param('RUNID'),
            '_gather_resources.log'
        )
    );
    # Create instance variables for certain attributes required to perform
    # reserve operations
    $self->{resources} = [];
    $self->{clusters} = {};
    $self->{num_volume} = 0;
    $self->{total_volume} = 0;
    $self->{num_aggregate} = 0;
    $self->{num_vserver} = 0;

    my %nodes = ();
    my %node_map = ();
    my %nodes_vs = ();

    my @aggrs = $self->get_objects(
        type => 'aggregate',
        filter => {
            'ha-policy' => 'sfo'
        },
        requested_fields =>
            ['aggregate', 'owner-name', 'size'],
    );

    foreach my $aggr (@aggrs) {
        $node_map{$aggr->{aggregate}} = $aggr->{owner_name};
    }

    my @volumes = $self->get_objects(
        type   => 'volume',
        filter => {
            'is-cluster-volume' => 'true',
        },
        requested_fields => ['volume','aggregate','size','vserver','type'],
    );

    foreach my $vol (@volumes) {
        my $node = $node_map{$vol->{aggregate}};
        $vol->{node} = $node;
        if (!defined($nodes{$node})) {
            $nodes{$node} = {};
            $nodes{$node}->{node} = $vol->{node};
        }
        if ($vol->{volume} =~ /root/) {

        } else {
            if (!defined($nodes{$node}->{volume})) {
                $nodes{$node}->{volume} = [];
            }

            # Leaving the original for backwards compatiblility with reserve
            push(@{$nodes{$node}->{volume}}, $vol);
            $self->{num_volume}++;
            $self->{total_volume}++;

            # Adding the vserver as a sub key to node for the volume location
            my $vserver = $vol->{vserver};
            unless (defined($nodes_vs{$node}{$vserver}{volume})) {
                $nodes_vs{$node}{$vserver}{volume} = [];
            }
            push(@{$nodes_vs{$node}{$vserver}{volume}}, $vol);
        }
    }

    # Existing resources data structure
    while (my ($node, $data) = each %nodes) {
        push(@{$self->{resources}}, $data);
    }

    # New implementation with vservers as subkeys
    $self->{resources_node_vs} = \%nodes_vs;
    $Log->debug("resource node vs structure: " . Dumper(\%nodes_vs));

    # Reset the logging to what it was previously.
    $Log->reset_path();
    $Log->comment('RESOURCES HAVE BEEN GATHERED');
} # End _gather_resources

#===============================================================================
# Description: Uses regex to compare the values of the filter with the values
# of the object that is passed in.  Returns true if all filter values are
# equivalent to the object values.  Returns false if any of the filter values
# are not equivalent to the object values.
#===============================================================================
sub _filter_check {
    my ($self, %args) = @_;
    $Log->enter();
    &ParamsMgr::process_args(%args);
    my $object = &ParamsMgr::get_arg(
        'arg'      => "object",
        'required' => FALSE,
        'default'  => undef,
    );
    my $filter = &ParamsMgr::get_arg(
        'arg'      => "filter",
        'required' => FALSE,
        'default'  => undef,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return wantarray ? () : [];
    }

    my ($key, $value) = "";
    while ( ($key, $value) = each %{$filter} ) {
        $Log->debug(
            "Comparing key: $key value: $value against object:" .
            $object->{$key}
        );
        unless ($object->{$key} =~ /$value/i){
            return FALSE;
        }
    }

    return TRUE;
} # End _filter_check

#==============================================================================
# Description: Create and accept a vserver peer relationship for SN clusters
#==============================================================================
sub _create_vserver_peer {
    my ($self, %args) = @_;
    $Log->enter();

    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $peer_vserver = &ParamsMgr::get_arg(
        'arg'      => "peer_vserver",
        'required' => TRUE,
    );
    my $peer_cluster = &ParamsMgr::get_arg(
        'arg'      => "peer_cluster",
        'required' => FALSE,
        'default'  => $self->{cluster_peer_name}
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $status = TRUE;
    try {
        NACL::STask::VserverPeer->create(
            command_interface  => $self->{cluster_home_cli},
            vserver            => $vserver,
            'peer-vserver'     => $peer_vserver,
            'peer-cluster'     => $peer_cluster,
            applications       => ['snapmirror'],
            nacltask_if_exists => 'reuse'
        );
    } otherwise {
        my $exception = shift;
        $Log->comment($exception->text());
        $status = FALSE;
    };

    try {
        NACL::STask::VserverPeer->accept(
            command_interface  => $self->{cluster_peer_cli},
            'peer-vserver'     => $vserver,
            vserver            => $peer_vserver,
        );
    } otherwise {
        my $exception = shift;
        if ($exception->text() =~ /Invalid peering state "peered" on the local cluster./) {
            $Log->comment("Vservers already peered... Continuing...");
            $status = TRUE;
        } elsif ($exception->text() =~ /Invalid operation on intra cluster Vserver peering/) {
            $Log->comment(
                'Cannot accept peer on Intracluster peering, continuing...'
            );
            $status = TRUE;
        } elsif ($exception->text() =~ /Invalid operation on intra cluster Vserver peer relationship./i) {
            $Log->comment(
                'Cannot accept peer on Intracluster peering, continuing...'
            );
            $status = TRUE;
        } else {
            $Log->comment($exception->text());
            $status = FALSE;
        }
    };

    return $status;
}

#==============================================================================
# Description: Delete a vserver peer relationship for SN clusters
#==============================================================================
sub _delete_vserver_peer {
    my ($self, %args) = @_;
    $Log->enter();

    &ParamsMgr::process_args(%args);
    my $vserver = &ParamsMgr::get_arg(
        'arg'      => "vserver",
        'required' => TRUE,
    );
    my $peer_vserver = &ParamsMgr::get_arg(
        'arg'      => "peer_vserver",
        'required' => TRUE,
    );
    if (&ParamsMgr::error()) {
        NATE::Exceptions::Argument->throw(&ParamsMgr::get_error());
        $Log->exit();
        return FALSE;
    }

    my $status = TRUE;

    try {
        NACL::STask::VserverPeer->purge(
            command_interface  => $self->{cluster_peer},
            vserver            => $peer_vserver,
            'peer-vserver'     => $vserver,
            nacltask_if_purged => 'pass',
        );
    } otherwise {
        $status = FALSE;
    };

    try {
        NACL::STask::VserverPeer->purge(
            command_interface  => $self->{cluster_home},
            vserver            => $vserver,
            'peer-vserver'     => $peer_vserver,
            nacltask_if_purged => 'pass',
        );
    } otherwise {
        $status = FALSE;
    };

    if ($status) {
        return TRUE;
    } else {
        return FALSE;
    }
} # End of _delete_vserver_peer

#==============================================================================
# Description: Discover and get cluster information for Snapmirrors tests.
#==============================================================================
sub _discover_clusters {
    my ($self, %args) = @_;
    $Log->enter();


    my $cluster_name = param("CLUSTER_NAME", "-default", "");
    my $peer_name    = param("PEER_CLUSTER", "-default", $cluster_name);

    # Get cluster objects so we can find the nodes easily for each cluster.
    my @clusters = NACL::STask::Cluster->find_unique_clusters();

    foreach my $cluster (@clusters) {
        my $name = $cluster->get_one_identity_attr('name');
        my $node = $cluster->{node};
        if ($name eq $cluster_name) {
            $self->{cluster_home_name} = $name;
            $Log->comment('Cluster home is:' . $name);
            $self->{cluster_home} = NACL::C::Cluster->new(node => $node);
            $self->{cluster_home}->apiset_must(interface => 'ZAPI');
            $self->{cluster_home}->apiset();
            $self->{cluster_home_cli} = NACL::C::Cluster->new(node => $node);
            $self->{cluster_home_cli}->apiset_must(interface => 'CLI');
            $self->{cluster_home_cli}->apiset();
        } elsif ($name eq $peer_name) {
            $self->{cluster_peer_name} = $name;
            $self->{cluster_peer} = NACL::C::Cluster->new(node => $node);
            $self->{cluster_peer}->apiset_must(interface => 'ZAPI');
            $self->{cluster_peer}->apiset();
            $self->{cluster_peer_cli} = NACL::C::Cluster->new(node => $node);
            $self->{cluster_peer_cli}->apiset_must(interface => 'CLI');
            $self->{cluster_peer_cli}->apiset();
        }
    }

    # Set the Peer Cluster equal to the Cluster Name if its not defined.
    if ($peer_name eq $cluster_name) {
        #$peer_name = $cluster_name;
        $self->{cluster_peer_name} = $self->{cluster_home_name};
        $self->{cluster_peer} = Clone::clone($self->{cluster_home});
        $self->{cluster_peer_cli} = Clone::clone($self->{cluster_home_cli});
        $Log->comment('Peer cluster:' . $self->{cluster_peer_name});
        $Log->comment(
            "Cluster and Peer Cluster have been set to the same name " .
            "so all operations will be Intracluster"
        );
    }

    $Log->comment(
        " - Using the following cluster(s): \n" .
        "  * Cluster Home: $cluster_name: " .
        join(',',$self->{cluster_home_cli}->nodes()) . "\n" .
        "  * Cluster Peer: $peer_name: " .
        join(',',$self->{cluster_peer_cli}->nodes()) . "\n"
    );

    $Log->exit();
}

1;