# Copyright (c) 2010 - 2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide a common set of methods for SAN FIT thpls and automation
## @author dl-san-fit@netapp.com
## @status private

use strict;

# Package Name
package SanFitUtil;

# NetApp Module Imports
use San::FIT::IOPresets;
use NATE::ParamSet;
use San::SanUtil;
use San::SanMgr;
use EventMan;
use SubIPC;
use Tharn;
use IOMan;
use TCD;
use San::ControllerCluster::Filer10M;

# Perl's Module Imports
use List::MoreUtils qw(uniq);
use Data::Dumper;
use Cwd;

# Globals
use vars qw (
    $Test
    @Stats
    $IOMan_Obj
    $SanMgr_Obj
    $EventMan_Obj
    $LifV_Abort_On_Error
    $ClamV_Abort_On_Error
    $PathV_Abort_On_Error
    $PortV_Abort_On_Error
    $Sanfit_Abort_On_Error
    $ControllerCluster_Obj
    $Use_Devices_From_File
    $ControllerV_Abort_On_Error
    $Controller_Scraper_Counter
    $Controller_Scraper_Is_Running
);

# Constants
use constant {
    DEFAULT_CONFIG_PATH          => "/san/fit/config/",
    DEFAULT_FIT_CONFIG           => "fit.config",
    DEFAULT_P4_PATH              => "/usr/software/test/noarch/agnostic",
    DEBUG_LEVEL                  => 1,
    DEFAULT_SLEEP                => 10,
    SUBTEST_TIMEOUT              => 2000,
    DEFAULT_SYSSTAT_COUNT        => 1000,
    DEFAULT_SYSSTAT_INTERVAL     => 5,
    DEFAULT_CLUSTER_STAT_TIMEOUT => 600,
    QOS_SCRIPT => "/x/eng/localtest/noarch/main/cluster/enterprise/config_qos_workload.thpl",
    RECOVERY_CO_SCRIPT => "/x/eng/localtest/noarch/main/cluster/enterprise/recovery_co_test.thpl",
    CONTROLLER_SCRAPER => "/x/eng/localtest/noarch/main/san/fit/ControllerScraper.py",
    RECOVERY_CO_TIMEOUT          => "TIMEOUT=1200",
    RECOVERY_CO_TESTCASES        => "TESTCASES=randomize_events*1000",
    RECOVERY_CO_EXIT_ON_ERROR    => "EXIT_ON_ERROR=1",
    RECOVERY_CO_RAND_SLEEP_MAX   => "RAND_SLEEP_MAX=240",
    RECOVERY_CO_RAND_SLEEP_MIN   => "RAND_SLEEP_MIN=120",
    RECOVERY_CO_FIND_ALL_NODES   => "FIND_ALL_NODES=1",
    RECOVERY_CO_NODES_ON_DEMAND  => "NODES_ON_DEMAND=1",
    VALIDATE_CLAM_HEALTH_SLEEP   => 10,
    DEFAULT_IO_TOOL_VERSION      => "current",
    DEFAULT_IO_DEVICES_PER_TOOL  => 1,
};

#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
#                               Public Methods
#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#

################################################################################
# Description: Initializes params used by SAN FIT, creates SanMgr, IOMan and
#              EventMan objects.
# Params: A TCD object, create_sanmgr, create_eventman, create_ioman flags
# Returns: A hash containing a ClusterObj, Clients, Devices, SanMgr, IOMan
#          EventMan object or Node objects if specified.
################################################################################
sub init_sanfit {
    my (%args) = @_;

    # Local variables
    my ($config_description, $param_string, $xml_file, $sanfit_config) = '';
    my ($create_sanmgr, $create_eventman, $create_ioman) = $SanUtil::TRUE;
    my ($init_node_objs, $create_cluster, $add_cluster) = $SanUtil::FALSE;
    my ($down_cluster_ports, $clobber_csm, $process_kill) = $SanUtil::FALSE;
    my ($validate_clam_is_healthy) = $SanUtil::FALSE;
    my ($out, $devices_from_file) = undef;
    my ($filer, $xml_root_dir, $io_tool_version, $io_devices_per_tool) = '';
    my $log_statistics = $SanUtil::TRUE;
    my $debug_level = DEBUG_LEVEL;
    my @print_stack = ();
    my $collect_asup = $SanUtil::FALSE;
    $Controller_Scraper_Counter = 0;

    %args = &SanUtil::strip_dash(%args);

    if (defined ($args{Test})) {
        $Test = $args{Test};
    } else {
        $Test->nlog("A TCD object must be passed into the test script");
        return undef;
    }

    if (defined ($args{create_sanmgr})) {
        $create_sanmgr = $args{create_sanmgr};
    }

    if (defined ($args{create_eventman})) {
        $create_eventman = $args{create_eventman};
    }

    if (defined ($args{create_ioman})) {
        $create_ioman = $args{create_ioman};
    }

    if (defined ($args{init_node_objs})) {
        $init_node_objs = $args{init_node_objs};
    }

    if (defined ($args{create_cluster})) {
        $create_cluster = $args{create_cluster};
    }

    if (defined ($args{add_cluster})) {
        $add_cluster = $args{add_cluster};
    }

    if (defined ($args{log_statistics})) {
        $log_statistics = $args{log_statistics};
    }

    # Get mandatory test parameters
    unless ($filer = $Test->get_param("FILER")) {
        $Test->nlog("A parameter FILER must be passed into the test script");
        return undef;
    }

    # Get default script params and initialize with default values
    $debug_level              = $Test->get_param("DEBUG_LEVEL", DEBUG_LEVEL);
    $log_statistics           = $Test->get_param("LOG_STATISTICS",
                                                 $log_statistics);
    $clobber_csm              = $Test->get_param("CLOBBER_CSM",
                                                 $SanUtil::FALSE);
    $process_kill             = $Test->get_param("PROCESS_KILL",
                                                 $SanUtil::FALSE);
    $down_cluster_ports       = $Test->get_param("DOWN_CLUSTER_PORTS",
                                                 $SanUtil::FALSE);
    $validate_clam_is_healthy = $Test->get_param("VALIDATE_CLAM_IS_HEALTHY",
                                                 $SanUtil::FALSE);
    $collect_asup             = $Test->get_param("COLLECT_ASUP",
                                                 $SanUtil::FALSE);

    # Users skipping SanMgr would need to decide if they want to reuse devices
    $Use_Devices_From_File = $Test->get_param("REUSE_DEVICES", $SanUtil::TRUE);

    # Get the right config for this script. This is the convention we're
    # following:
    #
    #  a) 2s -> 2-node small config
    #  b) 2m -> 2-node medium config
    #  c) 2l -> 2-node large config
    #  d) 4s -> 4-node small config
    #  .
    #  ..

    if ($xml_file = $Test->get_param("XML_FILE")) {
        $config_description = "Using a customized SAN FIT Config!";
    } else {
        # Defaults to a 2-node cluster, small config. Check SAN FIT TDS for details
        $sanfit_config = $Test->get_param("SANFIT_CONFIG", '2s');

        # Get the path to SANFIT's config files
        $xml_root_dir = &_get_p4_dir() . DEFAULT_CONFIG_PATH;
        $Test->nlog("Looking for XML files in: $xml_root_dir");

        # Get the path to the XML file describing specific configurations
        if ($sanfit_config =~ /2s/i) {
            $config_description = "Two-node cluster running SAN FIT small config.";
            $xml_file = $xml_root_dir . "sanfit_two_node_small.xml";
        } elsif ($sanfit_config =~ /2m/i)  {
            $config_description = "Two-node cluster running SAN FIT medium config.";
            $xml_file = $xml_root_dir . "sanfit_two_node_medium.xml";
        } elsif ($sanfit_config =~ /2l/i)  {
            $config_description = "Two-node cluster running SAN FIT large config.";
            $xml_file = $xml_root_dir . "sanfit_two_node_large.xml";
        } elsif ($sanfit_config =~ /4s/i)  {
            $config_description = "Four-node cluster running SAN FIT small config.";
            $xml_file = $xml_root_dir . "sanfit_four_node_small.xml";
        } elsif ($sanfit_config =~ /4m/i)  {
            $config_description = "Four-node cluster running SAN FIT medium config.";
            $xml_file = $xml_root_dir . "sanfit_four_node_medium.xml";
        } elsif ($sanfit_config =~ /4l/i)  {
            $config_description = "Four-node cluster running SAN FIT large config.";
            $xml_file = $xml_root_dir . "sanfit_four_node_large.xml";
        } else {
            # This should never happen...
            $config_description = "Unknown SAN FIT Config!";
            $Test->nlog("An XML configuration file is needed!");
            return undef;
        }
    }

    #+#+#+#+#+#+#+#+#+#+#+#
    #  EVENTMAN SETTINGS  #
    #+#+#+#+#+#+#+#+#+#+#+#
    # The order is important, please don't re-arrange unless something changed.
    $Sanfit_Abort_On_Error = $Test->get_param(
        "SANFIT_ABORT_ON_ERROR",
        $SanUtil::TRUE
    );
    $ClamV_Abort_On_Error = $Test->get_param(
        "CLAM_VALIDATION_ABORT_ON_ERROR",
        $Sanfit_Abort_On_Error
    );
    $PathV_Abort_On_Error = $Test->get_param(
        "PATH_VALIDATION_ABORT_ON_ERROR",
        $Sanfit_Abort_On_Error
    );
    $PortV_Abort_On_Error = $Test->get_param(
        "PORT_VALIDATION_ABORT_ON_ERROR",
        $Sanfit_Abort_On_Error
    );
    $LifV_Abort_On_Error = $Test->get_param(
        "LIF_VALIDATION_ABORT_ON_ERROR",
        $Sanfit_Abort_On_Error
    );
    $ControllerV_Abort_On_Error = $Test->get_param(
        "CONTROLLER_VALIDATION_ABORT_ON_ERROR",
        $Sanfit_Abort_On_Error
    );

    #+#+#+#+#+#+#+#+#+
    #  I/O SETTINGS  #
    #+#+#+#+#+#+#+#+#+
    $io_tool_version = $Test->get_param("IO_TOOL_VERSION",
                                        DEFAULT_IO_TOOL_VERSION);
    $io_devices_per_tool = $Test->get_param("IO_DEVICES_PER_TOOL",
                                            DEFAULT_IO_DEVICES_PER_TOOL);
    IOPresets::init_io_presets(io_limit   => $args{io_limit},
                               start_byte => $args{io_start_byte},
                               tool_version => $io_tool_version,
                               devices_per_tool => $io_devices_per_tool);

    # for each environment param, push its value onto a print stack
    my $params = NATE::ParamSet->get_global();
    my @paramset = $params->walk();
    push (@print_stack, "Test Default Parameters:\n==================");
    foreach my $hash (@paramset) {
        push (@print_stack, $hash->{key} . " : " . $hash->{value});
    }

    # print out the print stack once it has been populated.
    $param_string = join ("\n", @print_stack);
    $param_string = $param_string . "\n  SAN FIT Config : $config_description";

    # Make sure to log every single param this script expects
    $Log->comment($param_string);

    if ($create_sanmgr) {
        $Test->nlog("\n###############################\n" .
                    " Creating a SanMgr object." .
                    "\n###############################\n");

        # Make sure the configuration file exists
        unless (-e $xml_file) {
            $Test->nlog("The XML configuration file does not exist!");
            return undef;
        }

        # Create a SanMgr object
        $SanMgr_Obj = San::SanMgr->new(-xml_file => $xml_file,
                                       -debug_level => $debug_level,);
        if ($SanMgr_Obj->error()) {
            $Test->nlog("Failed to create a SanMgr object: \n",
                        $SanMgr_Obj->errmsg());
            return undef;
        }

        $out->{sanmgr_obj} = $SanMgr_Obj;
    }

    if ($create_cluster) {
        $Test->nlog("\n###############################\n" .
                    " Creating a Cluster object."    .
                    "\n###############################\n");

        $ControllerCluster_Obj = San::ControllerCluster->new(
            -cserver => $filer,
            -debug_level => $debug_level,
            -new_cmds_obj => 1
        );
        if ($ControllerCluster_Obj->error()) {
            $Test->nlog("Failed to create Controller Cluster object",
                        $ControllerCluster_Obj->errmsg());
            return undef;
        }

        $out->{cluster_obj} = $ControllerCluster_Obj;

        # If users want to run I/O and don't want to use SanMgr then they'll
        # need to get them from somewhere else.

        if ($Use_Devices_From_File) {
            $devices_from_file = &_get_devices_from_file();
            unless (defined ($devices_from_file)) {
                $Test->nlog("Failed to get devices from SAN FIT's config file.");
                return undef;
            }

            $out->{clients} = $devices_from_file->{clients};
            $out->{devices} = $devices_from_file->{devices};
        }
    }

    if ($create_eventman) {
        $Test->nlog("\n###############################\n" .
                    " Creating an EventMan object." .
                    "\n###############################\n");
        # Create an EventMan Object
        $EventMan_Obj = new EventMan(
            -debug_level => $debug_level,
            -clam_validation_abort_on_error       => $ClamV_Abort_On_Error,
            -path_validation_abort_on_error       => $PathV_Abort_On_Error,
            -port_validation_abort_on_error       => $PortV_Abort_On_Error,
            -lif_validation_abort_on_error        => $LifV_Abort_On_Error,
            -controller_validation_abort_on_error => $ControllerV_Abort_On_Error
        );
        if ($EventMan_Obj->error()) {
            $Test->nlog("Failed to create an EventMan object", $EventMan_Obj->errmsg());
            return undef;
        }

        unless (defined ($ControllerCluster_Obj)) {
            $ControllerCluster_Obj = San::ControllerCluster->new(
                -cserver => $filer,
                -debug_level => $debug_level,
                -new_cmds_obj => 1
            );
            if ($ControllerCluster_Obj->error()) {
                $Test->nlog("Failed to create a cluster object",
                            $ControllerCluster_Obj->errmsg());
                return undef;
            }
        }

        $Test->nlog("Adding controller cluster to EventMan");
        if (defined ($ControllerCluster_Obj)) {
            $EventMan_Obj->add_resource(-name => $ControllerCluster_Obj->get_name(),
                                        -obj  => $ControllerCluster_Obj,
                                        -type => "controllerclusters");
            if ($EventMan_Obj->error()) {
                $Test->nlog("Failed to add resource to EventMan object",
                            $EventMan_Obj->errmsg());
                return undef;
            }
        }

        $out->{eventman_obj} = $EventMan_Obj;
    }

    if ($create_ioman) {
        $Test->nlog("\n###############################\n" .
                    " Creating an IOMan object."      .
                    "\n###############################\n");
        # Create an IOMan Object
        $IOMan_Obj = new IOMan(-debug_level    => $debug_level,
                               -failure_policy => 'stop_all_client');
        if ($IOMan_Obj->error()) {
            $Test->nlog("Failed to create an IOMan object", $IOMan_Obj->errmsg());
            return undef;
        }

        $out->{ioman_obj} = $IOMan_Obj;
    }

    if ($init_node_objs) {
        my %node_objs;
        my @nodes = @{$init_node_objs};

        foreach my $node (@nodes) {
            $Test->nlog("\n#####################################\n" .
                        " Creating a Filer10M Object for $node" .
                        "\n#####################################\n");
            $node_objs{$node} = San::ControllerCluster::Filer10M->new(-filer => $node,
                                                                      -debug_level => $debug_level);
            if ($node_objs{$node}->error()){
                $Test->nlog("Creating the Filer10M Object for $node",
                            $node_objs{$node}->errmsg());
                return undef;
             }

             $out->{node_objs} = \%node_objs;
        }
    }

    if ($log_statistics) {
        # Start cluster and sysstat monitoring in the background
        start_sysstat(filer => $filer, debug_level => $debug_level);
        # Cache all stats
        $out->{stats} = \@Stats;
    }

    # Recovery CO background tasks for random cluster faults.
    if ($down_cluster_ports) {
        start_recovery_co(event => 'down_cluster_ports');
    }
    if ($clobber_csm) {
        start_recovery_co(event => 'clobber_csm');
    }
    if ($process_kill) {
        start_recovery_co(event => 'process_kill');
    }

    # Clam validation background task
    if ($validate_clam_is_healthy) {
        start_validate_clam_is_healthy(filer => $filer);
    }

    # Kick off a background ASUP collection
    if ($collect_asup) {
        my $interval   = $Test->get_param("ASUP_INTERVAL", 0);
        my $iterations = $Test->get_param("ASUP_ITERATIONS", 1);
        generate_asup(interval => $interval, iterations => $iterations);
    }

    return $out;
} # End init_sanfit

################################################################################
# Description: Gets a list of devices from a clients created by SanMgr.
# Params: None.
# Returns: A hash containing a list of device paths.
################################################################################
sub get_device_list {
    my (%args) = @_;

    # Local variables
    my $client_device_list = undef;
    my $final_client_dev_list = {};

    %args = &SanUtil::strip_dash(%args);

    unless (defined ($SanMgr_Obj)) {
        $Test->nlog("Failed to get the default SanMgr object!");
        return undef;
    }

    # Parse out the client dev list if it is stored on the sanmgr object
    $client_device_list = $SanMgr_Obj->get_client_dev_list();
    if ($SanMgr_Obj->error()) {
        $Test->nlog("Failed to get the list of devices from clients:",
                    $SanMgr_Obj->errmsg());
        return undef;
    }

    $Test->nlog("Devices returned by SanMgr:\n", Dumper ($final_client_dev_list));

    foreach my $client (keys %{$client_device_list}) {
        my @dev_paths = ();
        foreach my $cdl_entry (@{$client_device_list->{$client}}) {
            my $path = $cdl_entry->{path};
            push (@dev_paths, $cdl_entry->{path});
        }

        $Test->nlog("Device Paths:", Dumper(\@dev_paths));

        $final_client_dev_list->{$client} = \@dev_paths;
    }

    $Test->nlog("List of client devices:\n", Dumper ($final_client_dev_list));

    return $final_client_dev_list;
} # End get_device_list

################################################################################
# Objective: This method appends a list of clients and client devices to a
#            fit.config file. This is used to allow users to skip SanMgr and
#            still be able to run I/O on a previously configured host.
# Params: None.
# Returns: Path to a p4 directory.
################################################################################
sub save_client_devices {
    my (%args) = @_;

    # Locals
    my ($old, $new, $bak, $device_str) = "";
    my (@device_paths) = ();
    my $devices = {};

    if (defined ($args{devices})) {
        $devices = $args{devices};
    } else {
        $Test->nlog("Mandatory parameter 'devices' is not defined.");
        return $SanUtil::FALSE;
    }

    # Build strings from the hash where the keys represent the clients
    # and the values their corresponding devices. Use the following
    # syntax: client_name#device1,device2...

    while (my ($key, $value) = each (%{$devices})) {
        push (@device_paths, $key . '#' . join (',', @{$value}));
    }

    # Now use a semicolon to separate each record
    $device_str = join (';', @device_paths);

    # SanMgr does not like to be hacked... Screw it, hack it anyway.
    # I'm not too proud about this one though.
    # TODO: This is for windows clients. Linux should be different
    $device_str =~ s/\\\\.\\/\\\\\\\\.\\\\/g;

    # Open the existing fit.config file or create a new one if it does not
    # exist. Create a new file to add the config info and a backup file.

    $old = DEFAULT_FIT_CONFIG;
    $new = "$old.tmp.$$";
    $bak = "$old.bak";

    # Open the old file as a READ only and the new file as a WRITE
    eval {
        open(OLD, "< $old");
    };
    if ($@) {
        $Test->nlog("Failed to open SAN FIT configuration file (fit.config) $!");
        return $SanUtil::FALSE;
    }

    eval {
        open(NEW, "> $new");
    };
    if ($@) {
        $Test->nlog("Failed to create a temporary configuration file $!");
        close OLD;
        return $SanUtil::FALSE;
    }

    # Look for the cached params and append the client info
    eval {
        while (<OLD>) {
            s/FIT_SAVED_DEVICES.*/FIT_SAVED_DEVICES=$device_str/i;
            (print NEW $_);
        }
    };
    if ($@) {
        $Test->nlog("Failed to save changes to fit.config $!");
        close NEW;
        close OLD;
        return $SanUtil::FALSE;
    }

    # Close and cleanup configuration files
    close(OLD);
    close(NEW);

    eval {
        rename($old, $bak);
        rename($new, $old);
        unlink($bak);
    };
    if ($@) {
        $Test->nlog("Failed to cleanup SAN FIT configuration files $!");
        return $SanUtil::FALSE;
    }

    return $SanUtil::TRUE;
} # End of save_client_devices

################################################################################
# Objective: Returns a San FIT IO preset.
# Params:    An optional hash containing:
#              * -id: The id corresponding to a SAN Fit preset. .
# Returns:   A hash reference containing an I/O preset
################################################################################
sub get_io_preset {
    my (%args) = @_;
    %args = &SanUtil::strip_dash(%args);

    # Locals
    my %load = {};
    my $id   = "";

    # Validate params
    if (defined ($args{id})) {
        $id = $args{id};
    } else {
        $Test->nlog("Mandatory param 'id' is missing.");
        return \%load;
    }

    %load = %{IOPresets::get_io_preset(id => $id)};
    unless (%load) {
        $Test->nlog("Couldn't find an IO preset with that ID.");
    }

    $Test->nlog("  Selected load:    " . delete ($load{id}));
    $Test->nlog("  Load Description: " . delete ($load{description}));
    while (my ($key, $value) = each (%load)) {
        $Test->nlog("    $key: $value");
    }

    return \%load;
} # End get_io_preset

################################################################################
# Objective: This method returns an I/O load selected randomly from SAN FIT's
#            load presets.
# Params:  None.
# Returns: A hash reference containing an I/O preset
################################################################################
sub get_random_io_preset {
    my (%args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # Locals
    my ($id, $description) = "";
    my $load               = {};
    my %preset             = ();

    %preset = %{IOPresets::get_random_io_preset()};

    # Delete extra fields such as id and description
    $id = delete $preset{id};
    $description = delete $preset{description};
    $Test->nlog("  Selected load: $id");
    $Test->nlog("  Load Description: $description");
    while (my ($key, $value) = each (%{$load})) {
        $Test->nlog("    $key: $value");
    }
    $Test->nlog("\n");

    return \%preset;
} # End get_random_io_preset

################################################################################
# Objective: Start running sysstat on each node in the cluster,
# Params: A reference to a list containing the nodes in the cluster
# Returns: TRUE if sysstats run successfully, FALSE otherwise
################################################################################
sub start_sysstat {
    my (%args) = @_;

    # Locals
    my $eventman_obj = undef;
    my $debug_level  = DEBUG_LEVEL;

    if (defined $args{debug_level}) {
        $debug_level = $args{debug_level};
    }

    unless (defined ($ControllerCluster_Obj)) {
        $ControllerCluster_Obj =
            San::ControllerCluster->new(-cserver     => $args{filer},
                                        -debug_level => $debug_level,
                                        -new_cmds_obj => 1);
        if ($ControllerCluster_Obj->error()) {
            $Test->nlog("Failed to create Controller Cluster object",
                        $ControllerCluster_Obj->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # We don't want to reuse the EventMan object exported to our tests
    # so create a new one just for sysstats monitoring
    $eventman_obj = new EventMan(-debug_level => $debug_level);
    if ($eventman_obj->error()) {
        $Test->nlog("Failed to create an EventMan object", $eventman_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Need to add the cluster as a resource
    $eventman_obj->add_resource(-name => $ControllerCluster_Obj->get_name(),
                                -obj  => $ControllerCluster_Obj,
                                -type => "controllerclusters");
    if ($eventman_obj->error()) {
        $Test->nlog("Failed to add resource to EventMan object",
                    $eventman_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Add the event responsible for collecting stats.
    $eventman_obj->add_event(-event_name => "collect_controller_stats",
                             -controller_stats_iteration_count => DEFAULT_SYSSTAT_COUNT,
                             -controller_stats_interval_time   => DEFAULT_SYSSTAT_INTERVAL);
    if ($eventman_obj->error()) {
        $Test->nlog("Failed to add systat events to EventMan object",
                    $eventman_obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Create a separate subtest to run sysstats
    push (@Stats, subtest(
        sub {
            $eventman_obj->run_events();
            if ($eventman_obj->error()) {
                $Test->nlog("Failed to add systat events to EventMan object",
                            $eventman_obj->errmsg());
                return $SanUtil::FALSE;
            }

            return $SanUtil::TRUE;
        },
        -runid => "cluster_stats",
        -bg,
        -onexit,
            sub {
                my ($subtest) = @_;
                if ($eventman_obj->is_running()) {
                    $eventman_obj->abort_events(-graceful => $SanUtil::TRUE);
                }

                if ($eventman_obj->error()) {
                    $Test->nlog("Failed to stop sysstat event in EventMan object",
                                $eventman_obj->errmsg());
                    $subtest->exithandler($SanUtil::FALSE);
                }

                $subtest->exithandler($SanUtil::TRUE);
            }
        )
    );

    return $SanUtil::TRUE;
} # End start_sysstat

################################################################################
# Objective: Gets a snapshot of CPU, memstats and disk utilization.
# Params: A cluster object
# Returns: A hash containing measured cpu, mem usage and disk util
################################################################################
sub measure_cluster_load {
    my (%args) = @_;

    # Locals
    my ($measured_cpu_io, $measured_disk_io, $io_ops, $sample_length) = 0;
    my $cluster_obj = undef;
    my %stats = ();

    if (defined ($args{cluster_obj})) {
        $cluster_obj = $args{cluster_obj};
    } else {
        if (defined ($ControllerCluster_Obj)) {
            $cluster_obj = $ControllerCluster_Obj
        } else {
            $Test->nlog("Could not find a Cluster Object!");
            return undef;
        }
    }

    if (defined ($args{sample_length})) {
        $sample_length = $args{sample_length};
    } else {
        $sample_length = DEFAULT_SYSSTAT_COUNT;
    }

    ($measured_cpu_io, $measured_disk_io, $io_ops) =
        &SanUtil::measure_cluster_io_load(-clusters      => [$cluster_obj],
                                          -sample_length => $sample_length);

    $stats{measured_cpu}  = $measured_cpu_io;
    $stats{measured_disk} = $measured_disk_io;
    $stats{measured_iops} = $io_ops;

    $Test->nlog("Stats: ", Dumper (\%stats));

    return \%stats;
} # End of measure_cluster_load

################################################################################
# Objective: Checks the results of the requested IOMan_Obj and
# Params: Optional - An IOMan object
# Returns: A true if results have no errors, returns false if there are errors.
################################################################################
sub check_ioman_results {
    my %args = @_;
    %args = &SanUtil::strip_dash(%args);

    # Locals
    my $ioman_obj = undef;
    my @results = ();

    if (defined($args{ioman_obj})) {
        $ioman_obj = $args{ioman_obj};
    } else {
        $ioman_obj = $IOMan_Obj if (defined($IOMan_Obj));
    }
    unless (defined($ioman_obj)) {
        $Test->nlog(
            "Error: An IOMan Object is required to check_ioman_results IO.",
            "It must have been created using init_sanfit function " .
            "in this utility module, or passed in by the calling code"
        );
        return $SanUtil::FALSE;
    }

    @results = $IOMan_Obj->get_result(-wait_finish => $SanUtil::FALSE);
    if ($IOMan_Obj->error()) {
        $Test->nlog("Error getting results from IOMan: "
                    . $IOMan_Obj->errmsg());
    }
    foreach my $result (@results){
        if ($result->{error_count}) {
            $Test->nlog("IO Failure detected for id "
                        . $result->{id} . "\n" . $result->{error});
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
}

################################################################################
# Objective: Executes ControllerScraper.py as a subtest.
# Params: Required - nodes: a comma seperated list of nodenames to run on.
#         Optional - pass through key => values where key is any parameter
#                    supported by ControllerScraper and value is any appropriate
#                    value for that key.
#         examples:
#             $Controller_Scraper_Subtest = &SanFitUtil::controller_scraper(
#                 -nodes      =>  join(',',$Cluster_Obj->get_node_members()),
#                 -target_dir =>  "/x/eng/cores/" . getlogin() .
#                                 "/ioman_failure_" . time(),
#                 -show_cmds  =>  'vserver,volume,port,"fcp adapter",lif,' .
#                                 '"fcp initiator","iscsi initiator"'
#              );
# Returns: A subtest object.
################################################################################
sub controller_scraper {
    my (%args) = @_;

    %args = &SanUtil::strip_dash(%args);

    # Locals
    my $nodes = "";
    my $param_prefix = '-';
    my $subtest = undef;
    my @subtest_params = ();

    unless (defined($args{nodes})) {
        $Test->nlog(
            "Error: A list of nodes is required to controller_scraper."
        );
        return $SanUtil::FALSE;
    }
    # Remove this required param from the rest of the arguments so it won't
    # be added twice.
    $nodes = delete($args{nodes});

    @subtest_params = (
        CONTROLLER_SCRAPER,
        -runid  => $Controller_Scraper_Counter . "_controller_scraper",
        -onexit => sub {
            $Controller_Scraper_Is_Running = $SanUtil::FALSE;
            $Test->nlog("Done Executing Controller Scraper");
        },
        '-bg',
        '--',
        "-n  $nodes"
    );

    # This keeps us from having to update this function everytime a new
    # paramter arrives in ControllerScraper.
    foreach my $key (keys(%args)) {
        if (length($key) > 1) {
            $param_prefix = '--';
        }
        push(@subtest_params, $param_prefix . $key . " " . $args{$key});
    }

    # If any other instances are running, wait for them to finish.
    while($Controller_Scraper_Is_Running) {
        sleep(DEFAULT_SLEEP);
    }

    # Kick-off ControllerScraper script as a subtest
    $Controller_Scraper_Is_Running = $SanUtil::TRUE;
    $Controller_Scraper_Counter++;
    return subtest(@subtest_params);
}

################################################################################
# Objective: Runs qos script to create workloads for every volume
# Params: None.
# Returns: True if the script completes successfully, False otherwise
################################################################################
sub generate_qos_workload {
    $Test->step("Creating QoS workloads");

    # Locals
    my ($subtest)  = undef;
    my (@results)  = ();
    my ($fail_cnt) = 0;

    # Kick-off QoS script as a subtest
    $subtest = subtest(
        QOS_SCRIPT,
        -runid  => "qos_workload_setup",
        -onexit => sub {
            $Test->nlog("Done creating QoS workloads");
        },
        '-nobg',
        '-timeout' => SUBTEST_TIMEOUT
    );

    # Wait for it to finish (timeout if it doesn't finish in SUBTEST_TIMEOUT sec)
    while ($subtest->is_alive()) {
        Tharn::snooze(DEFAULT_SLEEP);
    }

    # Check the results
    @results = $subtest->results();
    foreach my $result (@results) {
        next if $result->type()=~/WARN/;
        if ($result->severity() >= 0) {
            $Test->nlog("   - " . $result->type() .
                        "(" . $result->severity . ") RESULT" . " for " .
                        $subtest->runid(), "   - msg: " . $result);
            if ($result->type() =~ /FAIL|SCRIPT|CONFIG/i) {
                $fail_cnt++;
            }
        }
    }

    if ($fail_cnt == 0) {
        $Test->step("Finished creating qos workloads for SAN FIT scenario");
        return $SanUtil::TRUE;
    } else {
        return $SanUtil::FALSE;
    }
} # End of generate_qos_workload

################################################################################
# Objective: Runs recovery co script to down cluster lifs randomly for duration
#                of test.
# Params: None.
# Returns: True if the script starts successfully, False otherwise
################################################################################
sub start_recovery_co {
    my (%args) = @_;

    my $event = $args{event} if defined($args{event});
    unless ($event) {
        $Test->nlog("Required parameter for start_recover_co \"event\" is "
                    . "missing.");
        return $SanUtil::FALSE;
    }

    $Test->step("Starting Recovery CO: $event");

    # Locals
    my $subtest  = undef;

    # Kick-off recovery co script as a subtest
    $subtest = subtest(
        RECOVERY_CO_SCRIPT,
        "EVENTS=" . $event,
        RECOVERY_CO_TIMEOUT,
        RECOVERY_CO_EXIT_ON_ERROR,
        RECOVERY_CO_RAND_SLEEP_MAX,
        RECOVERY_CO_RAND_SLEEP_MIN,
        RECOVERY_CO_FIND_ALL_NODES,
        RECOVERY_CO_NODES_ON_DEMAND,
        RECOVERY_CO_TESTCASES,
        -runid  => "recover_co:$event",
        -onexit => sub {
            my @subtest_results = $subtest->own_results(result_type => "FAIL");
            if (scalar(@subtest_results) > 0) {
                $Log->error(
                    "There were errors while running recover_co:$event."
                );
            }
        },
        '-bg'
    );

    if ($subtest->is_alive()) {
        $Test->step("Recovery CO: $event is running");
        return $SanUtil::TRUE;
    } else {
        $Test->step("ERROR: Recovery CO: $event is NOT running");
        return $SanUtil::FALSE;
    }
} # End of start_recovery_co

################################################################################
# Objective: Runs ClamValidation::validate_clam_is_healthy in a loop until
#                 parent exits..
# Params: Optional - filer - name of filer to instantiate a Cluster Object with.
#                            if no global ControllerCluster Object exists.
# Returns: True if the script starts successfully, False otherwise
################################################################################
sub start_validate_clam_is_healthy {
    my (%args) = @_;

    $Test->step("Starting Clam Validation: Clam stays healthy.");

    # Locals
    my $debug_level = defined($args{debug_level}) ? $args{debug_level} : DEBUG_LEVEL;
    my $subtest  = undef;
    my $fail_cnt = 0;
    unless (defined ($ControllerCluster_Obj)) {
        $ControllerCluster_Obj =
            San::ControllerCluster->new(-cserver     => $args{filer},
                                        -debug_level => $debug_level,
                                        -new_cmds_obj => 1);
        if ($ControllerCluster_Obj->error()) {
            $Test->nlog("Failed to create Controller Cluster object",
                        $ControllerCluster_Obj->errmsg());
            return $SanUtil::FALSE;
        }
    }

    my @nodes = $ControllerCluster_Obj->get_node_objs();
    if ($ControllerCluster_Obj->error()) {
        $Test->nlog("Failed to get node objects from Controller Cluster object",
                    $ControllerCluster_Obj->errmsg());
        return $SanUtil::FALSE;
    }

    # Kick-off recovery co script as a subtest
    $subtest = subtest(
        sub {
            # Reconnect all node objects so they wont step on other connections.
            foreach my $n_obj (@nodes) {
                $n_obj->reconnect(-disconnect_existing_conn => $SanUtil::FALSE);
                if ($n_obj->error()) {
                    $Test->nlog("Failed to reconnect to " . $n_obj->get_name()
                                . $n_obj->errmsg());
                    return $fail_cnt++;
                }
            }

            my $clam_master = &ClamValidation::get_clam_master(-nodes => \@nodes);
            unless ($clam_master) {
                $Test->nlog("Failed to determine a clam master");
                return $fail_cnt++;
            }

            # Let the parent kill this subtest.
            while (1) {
                unless(
                    &ClamValidation::validate_clam_is_healthy(
                        -nodes => \@nodes,
                        -original_master => $clam_master
                    )
                ) {
                    $fail_cnt++;
                }
                &Tharn::snooze(VALIDATE_CLAM_HEALTH_SLEEP);
            }
        },
        -runid  => "validate_clam_is_healthy",
        -onexit => sub {
            unless ($fail_cnt == 0) {
                $Log->error(
                    "Clam Validation Failed $fail_cnt times during testing."
                );
                die fatal_result(
                    "Clam Validation Failed $fail_cnt times during testing."
                );
            }
        },
        '-bg'
    );

    if ($subtest->is_alive()) {
        $Test->step("Validate Clam is Healthy is running");
        return $SanUtil::TRUE;
    } else {
        $Test->nlog("ERROR: Validate Clam is Healthy is NOT running");
        return $SanUtil::FALSE;
    }
} # End sub start_validate_clam_is_healthy

################################################################################
# Objective: Generates and collects ASUP data during a SAN FIT test
# Params:
#   * interval (optional): Number of seconds between ASUP calls
#   * iterations (optional): Number of ASUP calls to execute during the test
# Returns: True if the ASUP data was collected successfully, False otherwise
################################################################################
sub generate_asup {
    my (%args) = @_;

    $Test->step("Generating ASUP data");

    # Locals
    my $debug_level = defined($args{debug_level}) ? $args{debug_level} : DEBUG_LEVEL;
    my $iterations  = defined($args{iterations}) ? $args{iterations} : 1;
    my $interval    = defined($args{interval}) ? $args{interval} : 0;
    my $runtime     = $Test->get_param("RUNTIME", 1);
    my $dir         = getcwd();

    # If users specify iterations, calculate the interval based on the runtime
    if ($iterations > 1) {
        # Weird case where no I/O is running but we want to collect ASUP
        # (i.e. RUNTIME=1 or RUNTIME=0)
        if ($runtime <= $iterations) {
            $interval = 1;
        } else {
            $interval = int($runtime / $iterations);
        }
    } else {
        if ($runtime < $interval) {
            $interval = $runtime;
        } else {
            $iterations = int($runtime / $interval) if ($interval != 0);
        }
    }

    subtest(
        sub {
            unless (defined ($ControllerCluster_Obj)) {
                $ControllerCluster_Obj =
                    San::ControllerCluster->new(-cserver     => $args{filer},
                                                -debug_level => $debug_level,
                                                -new_cmds_obj => 1);
                if ($ControllerCluster_Obj->error()) {
                    $Test->nlog("Failed to create Controller Cluster object",
                                $ControllerCluster_Obj->errmsg());
                    return $SanUtil::FALSE;
                }
            }

            my @nodes = $ControllerCluster_Obj->get_node_objs();
            if ($ControllerCluster_Obj->error()) {
                $Test->nlog(
                    "Failed to get node objects from Controller Cluster object",
                    $ControllerCluster_Obj->errmsg()
                );
                return $SanUtil::FALSE;
            }

            for (my $i = 0; $i < $iterations; $i++) {
                # Kick off controllerscraper.
                my $controller_scraper_subtest = &SanFitUtil::controller_scraper(
                    -nodes      => join(',',
                                        $ControllerCluster_Obj->get_node_members()
                                   ),
                    -target_dir => Tharn::param('LOGDIR')
                                    . "/asup_load_generation/asup_data/0" . $i
                                    . "_iteration",
                    -options    => 'asup',
                    -verbose    => ''
                );
                $controller_scraper_subtest->wait_finish();
                sleep($interval);
            }
        },
        -runid => "asup_load_generation",
        -onexit => sub {
            $Test->nlog("Completed ASUP Load Generation.");
        },
        '-bg'
    );

    return $SanUtil::TRUE;
} # End generate_asup

#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#
#                              Private Methods
#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#+#

################################################################################
# Objective: This function will dynamically obtain the p4 directory of the test
# Params: None.
# Returns: Path to a p4 directory.
################################################################################
sub _get_p4_dir {
    # Locals
    my $p4_dir = DEFAULT_P4_PATH;
    my @natelib = ();

    # Try to determine from NATE_LIB/NATE_LIB_UNIX
    if (exists($ENV{"NATE_LIB"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB"});
    }
    if (not scalar(@natelib) and exists($ENV{"NATE_LIB_UNIX"})) {
        @natelib = split(/;/, $ENV{"NATE_LIB_UNIX"});
    }

    # Walk through all the natelib paths until we find one that works
    foreach my $natelib_path (@natelib) {
        $p4_dir = $natelib_path;
        if (-d $p4_dir) {
            return $p4_dir;
        }
    }

    return $p4_dir;
} # End of _get_p4_dir

################################################################################
# Objective: Looks for a list of client/devices in the testbed config file
#            and caches it in global variables. This list exists only if
#            users ran SanMgr's setup successfully and specified the parameter
#            to cache the list.
# Params: None.
# Returns: A hash containing two hash references with clients and devices
################################################################################
sub _get_devices_from_file {
    # Locals
    my (@clients, @devices, @client_dev_pairs) = ();
    my $file = DEFAULT_FIT_CONFIG;
    my $device_str = "";
    my $out = {};

    $device_str = Tharn::param("FIT_SAVED_DEVICES","-default", "");

    # Open the file as a READ only
    eval {
        open(CONFIG, "< $file");
    };
    if ($@) {
        $Log->comment("Failed to open SAN FIT configuration file " . DEFAULT_FIT_CONFIG . " $!");
        return $out;
    }

    # Look for the device and client parameter
    eval {
        while (<CONFIG>) {
            /^FIT_SAVED_DEVICES=(.+)/;
            $device_str = $1;
        }
    };
    if ($@) {
        $Log->comment("Failed to read " . DEFAULT_FIT_CONFIG . " $!");
        close CONFIG;
        return $out;
    }

    # Close the file
    close CONFIG;

    if ($device_str eq "") {
        $Log->comment("There are no devices in your SAN FIT config file.\n" .
                    "This could happen if SanMgr's setup didn't run ".
                    "successfully or you don't have a fit.config file in your ".
                    "testbed.");

        return $out;
    }

    # The syntax for clients/devices is:
    # client1#device1,device2,..,deviceN;client2#device1,device2,..,deviceN
    # Start by splitting each client/device pair.

    @client_dev_pairs = split (';', $device_str);

    # Go through the list and extract the clients and the devices
    foreach my $pair (@client_dev_pairs) {
        $pair =~ s/(.*)#//;
        push (@clients, $1);
        # Apparently reading from files appends even more \\ so let's remove.
        $pair =~ s/\\\\\\\\.\\\\/\\\\.\\/g;
        my @device = split (',', $pair);
        push (@devices, \@device);
    }

    # Now store both clients and devices
    $out->{devices} = \@devices;
    $out->{clients} = \@clients;

    return $out;
} # End of _get_devices_from_file

################################################################################
# Objective: Just a fun moo-moo message to cheer up SAN FIT'ers
# Params: None.
# Returns: A string
################################################################################
sub say_moo {
    my $str = "";
    my $file = &_get_p4_dir() . DEFAULT_CONFIG_PATH . "holycow";

    eval { open(MOO, "< $file"); };
    if ($@) { return ""; }
    while (<MOO>) { $str .= $_; }
    close MOO;

    return $str;
}

1;
