# Copyright (c) 2009 - 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provides a framework to build test scenarios for SSAN FIT.
## @author jose.porrua@netapp.com
## @status Private
## @pod Not here yet.

package San::FitMgr;

# Compiler Directives
use strict;

# Module Import
use San::SanUtil;
use San::SanLog;
use Error;
use San::SanFiler;
use San::SanClient;
use XML::LibXML;
use Data::Dumper;
use Tharn;

use base qw(Error SanLog SanUtil);

# Global Variables. Yes - I'm planning to move it inside the object and define
# it as an instance variable.
my @Lun_Object_Pool = ();
my @Devices = ();
my @Lun_Map_Info = ();

###############################################################################
# P U B L I C   M E T H O D S
###############################################################################

###############################################################################
# Method:     new
# Objective:  Creates new FitMgr object and starts the creation of SAN objs.
# Parameters: An XML file describing a FIT scenario.
# Returns:    A reference to the object.
# Details:    See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    # Object Initialization
    my $self = {
        clusters   => '',
        nodes      => '',
        vservers   => '',
        aggregates => '',
        volumes    => '',
        lifs       => '',
        luns       => '',
        igroups    => '',
        portsets   => '',
	iscsi      => '',
	fcp        => '',
	tpgroup    => '',
        xml_data   => '',
        file       => '',
        tcd_test   => '',
        cleanup    => $SanUtil::FALSE,
        clean_all  => $SanUtil::FALSE,
        rescan_timeout => 3,
        cmd_interface => 'cli',
    };

    # set log handle for stdout
    if (defined($args{loghandle}) and length($args{loghandle})) {
        $self->{log} = $args{loghandle};
    }

    # Create an instance of the class
    bless $self, $class;

    return $self;
} # End of new

###############################################################################
# Method:     setup
# Objective:  Gets the elements from the XML file and starts creating objects
# Parameters: None.
# Returns:    True if everything succeeded, False otherwise (and sets error)
###############################################################################
sub setup {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my ($xml_data, $xml_root_node, $clean_if_error, $clean_all);
    my $status = $SanUtil::FALSE;
    my $debug_level = 0;
    my (@vservers, @nodes,    @aggregates,
        @volumes,  @portsets, @lifs,
        @luns,     @igroups,  @clients,
        @potsets, @iscsi, @fcp, @tpgroups) = ();

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating objects."));

    # Remove the dash '-' from input args if it exists
    %args = &SanUtil::strip_dash(%args);

    # Parse parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /file/) {
            unless($value) {
                $self->_set_err("Invalid value for \"-$key\": " . $value);
                return $SanUtil::FALSE;
            }
            $self->{file} = $value;
        } elsif ($key =~ /debug_level/) {
            if ($value =~ /\d+/) {
                $debug_level = $value;
            }
        } elsif ($key =~ /loghandle/) {
            if (length($value)) {
                $self->{log} = $value;
            }
        } elsif ($key =~ /tcd_test/) {
            if (length($value)) {
                $self->{tcd_test} = $value;
            }
        } elsif ($key =~ /cmd_interface/) {
            if ($value =~ /cli|zapi/i) {
                $self->{cmd_interface} = $value;
            }
        } else {
            $self->_set_err("Invalid Parameter. ");
            return $SanUtil::FALSE;
        }
    }

    # set debug level
    $self->set_debug_level($debug_level);

    # Validate that the file exists, is well-formed and follows the Schema
    $self->_validate_xml();
    if ($self->error()) {
        $self->_set_err("Failed to validate XML Scenario: ", $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Get the xml data. Make sure there's info in there (_validate_xml will
    # update this attribute)
    $xml_data = $self->{xml_data};
    unless ($xml_data) {
        $self->_set_err("There is no data on this XML file.");
        return $SanUtil::FALSE;
    }

    # Get the root of the XML Document and keep it as an object's instance var.
    $xml_root_node = $xml_data->getDocumentElement;
    $self->{xml_root} = $xml_root_node;

    # We use an optional parameter to determine if we want cleanup to run
    # automatically if something fails
    $clean_if_error = $self->_get_val($xml_root_node->getAttribute('cleanup_if_error'));
    if ($clean_if_error) {
        $self->{cleanup} = $SanUtil::TRUE;
    }

    # There may be cases where users don't want to reuse existing objects
    # this is useful when you want the cleanup to kick in and delete everything
    $clean_all = $self->_get_val($xml_root_node->getAttribute('clean_all'));
    if ($clean_all) {
        $self->{clean_all} = $SanUtil::TRUE;
    }

    # Get Every node and put them in their corresponding "buckets"
    @nodes      = $self->_get_nodes(id => 'node');
    @aggregates = $self->_get_nodes(id => 'aggr');
    @vservers   = $self->_get_nodes(id => 'vserver');
    @lifs       = $self->_get_nodes(id => 'lif');
    @volumes    = $self->_get_nodes(id => 'vol');
    @luns       = $self->_get_nodes(id => 'lun');
    @igroups    = $self->_get_nodes(id => 'igroup');
    @portsets   = $self->_get_nodes(id => 'portset');
    @iscsi      = $self->_get_nodes(id => 'iscsi');
    @fcp        = $self->_get_nodes(id => 'fcp');
    @tpgroups   = $self->_get_nodes(id => 'tpgroup');
    @clients    = $self->_get_nodes(id => 'client');

    # Build Cluster and Nodes (San Filer objects)
    if (scalar(@nodes)) {
        $self->_create_nodes(nodes => \@nodes);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more nodes based on " .
                            "the XML data. ", $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # Build Aggregates
    if (scalar(@aggregates)) {
        $self->_create_aggregates(aggregates => \@aggregates);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more aggrs based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # Build Vservers
    if (scalar(@vservers)) {
        $self->_create_vservers(vservers => \@vservers);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more vserver based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # Build LIFS
    if (scalar(@lifs)) {
        $self->_create_lifs(lifs => \@lifs);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more LIFs based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # Build Volumes
    if (scalar(@volumes)) {
        $self->_create_volumes(volumes => \@volumes);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more volume based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # Build Luns
    if (scalar(@luns)) {
        $self->_create_luns(luns => \@luns);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more luns based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # Build iGroups
    if (scalar(@igroups)) {
        $self->_create_igroups(igroups => \@igroups);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more igroups based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # Build Portsets
    if (scalar(@portsets)) {
        $self->_create_portsets(portsets => \@portsets);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more portset based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # Build iSCSI Service
    if (scalar(@iscsi)) {
        $self->_create_iscsi(iscsi => \@iscsi);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more iSCSI service(s) based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # Build FCP Service
    if (scalar(@fcp)) {
        $self->_create_fcp(fcp => \@fcp);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more FCP service(s) based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }
    # Build tpgroups
    if (scalar(@tpgroups)) {
        $self->_create_tpgroup(tpgroup => \@tpgroups);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more TPGroup(s) based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # Build Clients
    if (scalar(@clients)) {
        $self->_create_clients(clients => \@clients);
        if ($self->error()) {
            $self->_set_err("Failed to create one or more clients based on " .
                            "the XML data. " , $self->errmsg());
            return $SanUtil::FALSE;
        }
    }

    # If something blew up and we want to cleanup the mess.
    if ($self->error()) {
        $self->_set_err("Setup failed to create one or more objects" ,
                        $self->errmsg());
        unless ($self->_clean_if_error()) {
            $self->_set_err("Setup detected an error, attempted to cleanup " .
                            "the system and failed." , $self->errmsg());
        }
        return $SanUtil::FALSE;
    }
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating objects."));

    return $SanUtil::TRUE;
} # End of setup()

###############################################################################
# Method:     cleanup
# Objective:  Cleans or destroys all SAN objects created by setup.
# Parameters: None
# Returns:    True if it succeeded, false otherwise
# Details:    See POD documentation at the end of this file.
###############################################################################
sub cleanup {
    my ($self, %args) = @_;

    $self->_clear_err();

    # Local Variables & Initialization
    my ($aggr_ref, $vserver_ref, $vol_ref, $lun_ref, $igroup_ref, $lif_ref);

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start cleaning up."));

    # New will build a scenario and a SAN object for each element described
    # in the XML file. Start cleaning up the scenario by destroying elements
    # in reverse order in which they were created: luns, igroups, lifs, vols,
    # vsrevers, aggregates.

    $self->_destroy(elements => "clients", object_id => "client_obj");
    if ($self->error()) {
        $self->_set_err("Failed to deconfigure clients: " , $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->_destroy(elements => "portsets", object_id => "portset_obj");
    if ($self->error()) {
        $self->_set_err("Failed to destroy portsets: " , $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->_destroy(elements => "tpgroups", object_id => "tpgroup_obj");
    if ($self->error()) {
        $self->_set_err("Failed to destroy tpgroups: " , $self->errmsg());
        return $SanUtil::FALSE;
    }

    $self->_destroy(elements => "luns", object_id => "lun_obj");
    if ($self->error()) {
        $self->_set_err("Failed to destroy luns: " , $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Now try destroying the igroups
    $self->_destroy(elements => "igroups", object_id => "igroup_obj");
    if ($self->error()) {
        $self->_set_err("Failed to destroy igroups: " , $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Now try destroying the vols
    $self->_destroy(elements => "volumes", object_id => "volume_obj");
    if ($self->error()) {
        $self->_set_err("Failed to destroy volumes: " , $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Now try destroying the lifs
    $self->_destroy(elements => "lifs", object_id => "lif_obj");
    if ($self->error()) {
        $self->_set_err("Failed to destroy lifs: " , $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Now try destroying the vservers (and its root volumes)
    $self->_destroy(elements => "vservers", object_id => "vserver_obj");
    if ($self->error()) {
        $self->_set_err("Failed to destroy vservers: " , $self->errmsg());
        return $SanUtil::FALSE;
    }

    # Finally, destroy the aggregates
    $self->_destroy(elements => "aggregates", object_id => "aggr_obj");
    if ($self->error()) {
        $self->_set_err("Failed to destroy aggregates: " , $self->errmsg());
        return $SanUtil::FALSE;
    }
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done cleaning up."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     get_filers
# Objective:  Gets filer objects created by Fit Setup.
# Parameters: An optional hash: -name to get a specific filer
# Returns:    A list of filer objects
# Details:    See POD documentation at the end of this file.
###############################################################################
sub get_filers {
    my ($self, %args) = @_;

    # Locals
    my ($filer_obj_ref, $filer_name);
    my (@filers, @filer_objs, @matched_filer_obj);

    $self->_clear_err();

    # If users specify an optional parameters "name", this method will only
    # return the filer object whose name matches this argument.
    if (defined($args{name})) {
        $filer_name = $args{name};
    }

    # If the build process was successful, all filer objects will be in the
    # "nodes" bucket. Get the reference to it.
    if (defined($self->{nodes})) {
        $filer_obj_ref = $self->{nodes};
    } else {
        return (wantarray()) ? (): [];
    }
    @filers = @$filer_obj_ref;

    # @ filers contains hashes. Traverse the list of hashes and get the
    # filer object data. Users do not cara about the rest.
    foreach my $filer (@filers) {
	# Make sure we've got a ref to a hash here or else it won't work.
        unless (defined ($filer)) {
            next;
        }
        if ($filer_name) {
            # If using name, return a list of one element (to keep this API
            # consistent i.e. always return a list or a ref to one)
            if ($filer->{name} =~ /$filer_name/) {
                push (@matched_filer_obj, $filer->{filer_obj});
                return (wantarray()) ? @matched_filer_obj: \@matched_filer_obj;
            }
        }
	push(@filer_objs, $filer->{filer_obj});
    }

    # At this point, if a user specified a filer name and we have not returned
    # yet, it means we didn't find it
    if ($filer_name) {
        return (wantarray()) ? (): [];
    }

    return (wantarray()) ? @filer_objs: \@filer_objs;
}

###############################################################################
# Method:     get_objects
# Objective:  Gets all the objects of a specific 'type' created by Fit Setup.
# Parameters: An ash: -type (Mandatory) specifying type of object
#                     -name (optional) return object matching a name.
# Returns:    A list of objects.
# Details:    See POD at the end of this file
###############################################################################
sub get_objects {
    my ($self, %args) = @_;

    # Start with a cleared flag
    $self->_clear_err();

    # Locals
    my ($obj_ref, $name, $type, $obj_id);
    my (@fit_objs, @objs);

    # Remove the dash '-' from input args if it exists
    %args = &SanUtil::strip_dash(%args);

    # Parse and validate params
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /type$/i) {
            unless($value) {
                $self->_set_err("Invalid value for \"-$key\": " . $value);
                return (wantarray()) ? (): [];
            }
            $type = $value;
        } elsif ($key =~ /name$/i) {
            unless($value) {
                $self->_set_err("Invalid value for \"-$key\": " . $value);
                return (wantarray()) ? (): [];
            }
            $name = $value;
        } else {
            $self->_set_err("Invalid Parameter. ");
            return (wantarray()) ? (): [];
        }
    }

    # Type is a mandatory param
    unless ($type) {
        $self->_set_err("Please specify an object type.");
        return (wantarray()) ? (): [];
    }

    # Tweak the type name to create a valid obj key required to access the objects
    if ($type =~ /filers?|nodes?/i) {
        $type = "nodes";
        $obj_id = "filer_obj";
    } elsif ($type =~ /clients?/i) {
        $type = "clients";
        $obj_id = "client_obj";
    } elsif ($type =~ /vservers?/i) {
        $type = "vservers";
        $obj_id = "vserver_obj";
    } elsif ($type =~ /aggrs?|aggregates?/i) {
        $type = "aggregates";
        $obj_id = "aggregate_obj";
    } elsif ($type =~ /volumes?|vols?/i) {
        $type = "volumes";
        $obj_id = "volume_obj";
    } elsif ($type =~ /luns?/i) {
        $type = "luns";
        $obj_id = "lun_obj";
    } elsif ($type =~ /lifs?/i) {
        $type = "lifs";
        $obj_id = "lif_obj";
    } elsif ($type =~ /igroups?/i) {
        $type = "igroups";
        $obj_id = "igroup_obj";
    } elsif ($type =~ /portsets?/i) {
        $type = "portsets";
        $obj_id = "portset_obj";
    } elsif ($type =~ /portsets?/i) {
        $type = "portsets";
        $obj_id = "portset_obj";
    } else {
        $self->_set_err("Invalid type: $type.");
        return (wantarray()) ? (): [];
    }

    # If the build process was successful, all objects will be in the
    # "object" bucket. Get the reference to it.
    if (defined($self->{$type})) {
        $obj_ref = $self->{$type};
    } else {
        $self->_set_err("No entries matching $type.");
        return (wantarray()) ? (): [];
    }
    @fit_objs = @$obj_ref;

    # @fit_objs contains hashes. Traverse the list of hashes and get the
    # object data by building a new list.
    foreach my $object (@fit_objs) {
        # Make sure we've got a ref to a hash here or else it won't work.
        unless (defined ($object)) {
            next;
        }
        if ($name) {
            # If using name, return a list of one element (to keep this API
            # consistent i.e. always return a list or a ref to one)
            if ($object->{name} =~ /^$name$/) {
                push (@objs, $object->{$obj_id});
                return (wantarray()) ? @objs: \@objs;
            }
        } else {
            push(@objs, $object->{$obj_id});
        }
    }

    return (wantarray()) ? @objs: \@objs;
}

###############################################################################
# Method:     get_devices
# Objective:  Gets device objects if users specified initiators in the XML file.
# Parameters: An hash: -client specifies the client to get the devices from
#                      -vserver tells the lib to return those devices which
#                      corrspond to a given vserver.
#                      -path_only will return a list of paths instead of devs.
# Returns:    A list of devices
# Details:    See POD documentation at the end of this file.
###############################################################################
sub get_devices {
    my ($self, %args) = @_;

    # Locals
    my ($client_name, $vserver_name, $device_ref, $client_type, $client_obj);
    my (@devices, @all_devices, @paths, @client_objs) = ();
    my ($path_only, $refresh, $skip_discover) = $SanUtil::FALSE;
    my %device_data;

    # Clear the error flag
    $self->_clear_err();

    # Validate parameters
    while (my ($key, $value) = each(%args)) {
        if ($key =~ /client/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"-$key\": " . $value);
                return (wantarray()) ? (): [];
            }
            $client_name = $value;
        } elsif ($key =~ /vserver/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"-$key\": " . $value);
                return (wantarray()) ? (): [];
            }
            $vserver_name = $value;
        } elsif ($key =~ /path_only/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"-$key\": " . $value);
                return (wantarray()) ? (): [];
            }
            $path_only = $value;
        } elsif ($key =~ /refresh/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"-$key\": " . $value);
                return (wantarray()) ? (): [];
            }
            $refresh = $value;
        } elsif ($key =~ /skip_discover/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"-$key\": " . $value);
                return (wantarray()) ? (): [];
            }
            $skip_discover = $value;
        } elsif ($key =~ /rescan_timeout/) {
            unless ($value) {
                $self->_set_err("Invalid value for \"-$key\": " . $value);
                return (wantarray()) ? (): [];
            }
            $self->{rescan_timeout} = $value;
        } else {
            $self->_set_err("Invalid Parameter \"-$key\" ");
            return (wantarray()) ? (): [];
        }
    }

    # If the users wants an updated list of Devices (not cached), initialize
    # the list and call discover and get devices
    if ($refresh == $SanUtil::TRUE) {
        my @client_objs = $self->get_objects(type => "clients",
                                             name => $client_name);
        if ($self->error()) {
            # Bubble up
            return (wantarray()) ? (): [];
        }
        # We only expect and care about one client.
        $client_obj = $client_objs[0];
        @Devices = ();
        $self->_discover_devices(client_obj     => $client_obj,
                                 skip_discover  => $skip_discover,
                                 rescan_timeout => $self->{rescan_timeout});
        if ($self->error()) {
            # Bubble up
            return (wantarray()) ? (): [];
        }
    }

    # Now loop through the global list of devices and get them (list of hashes)
    foreach my $device (@Devices) {
        %device_data = %$device;
        $device_ref = $device_data{devices};
        @all_devices = @$device_ref;
        $client_type = $device_data{client_type};

        # If no client or vserver is specified return all devices
        unless ($client_name or $vserver_name) {
            @devices = @all_devices;
            next;
        }
        # Specified both client name and vserver name
        if ($client_name and $vserver_name) {
            foreach my $dev (@all_devices) {
                unless ($device_data{client_name} =~ /^$client_name$/i) {
                    next;
                }
                # Windows and Linux hosts will piggyback a filer name, which
                # we'll use to filter those devices.
                unless ($dev->{filer_name} =~ /^$vserver_name$/i) {
                    next;
                }

                # need the client_type to build paths properly if path-only
                if($path_only){
                    $dev->{client_type} = $client_type;
                }
                push (@devices, $dev);
            }
            next;
        }
        # Specified a client only
        if ($client_name) {
            foreach my $dev (@all_devices) {
                unless ($device_data{client_name} =~ /^$client_name$/i) {
                    next;
                }

                # need the client_type to build paths properly if path-only
                if($path_only){
                    $dev->{client_type} = $client_type;
                }
                push (@devices, $dev);
            }
            next;
        }
        # Specified a vserver only
        if ($vserver_name) {
            foreach my $dev (@all_devices) {
                unless ($dev->{filer_name} =~ /^$vserver_name$/i) {
                    next;
                }
                push (@devices, $dev);
            }
            next;
        }
    } # End looping the global list of devices

    # If users call path_only and the host is not Windows, append the device
    # name to the end.
    if ($path_only) {
        foreach my $device (@devices) {
            # if I have some sort of Dev object I'm assuming FS for now
            # so need to call get_mount to get the drive letter or mountpoint
            if( ref($device) =~ /Dev/){
                push(@paths, $device->get_mount());
                next;
            }
            if ($device->{client_type}  =~ /windows/i) {
                push(@paths, $device->{path});
            } else {
                my $dev = $device->{path} . "/" . $device->{name};
                push(@paths, $dev);
            }
        }
        return (wantarray()) ? @paths : \@paths;
    }

    return (wantarray()) ? @devices : \@devices
}

################################################################################
# Method Name: logout_clients
# Objective:   Logs out of all @clients in the FitMgr object and erases the
#              callback.
# Parameters: -refresh  1 (default) = call get_devices with refresh => 1 after
#                                     logout
#                       0 = don't call get_devices
# Returns:    True if it succeeded, False otherwise.
################################################################################
sub logout_clients {
    my ($self, %args) = @_;
    my @clients;      # FitMgr object's clients
    my @callbacks;    # Callbacks for the client
    my $refresh = 1;  # Refresh means call get_devices after logout

    $self->_clear_err();

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start logout_clients."));

    # remove the dash '-' from input args if exist
    # SanUtil::strip_dash
    %args = &SanUtil::strip_dash(%args);

    # Parse arguments
    foreach (keys %args) {
        if (/refresh/) {
            $refresh = $args{$_};
            if (not length($refresh) or $refresh !~ /^0$|^1$/) {
                $self->_set_err("Invalid value for argument '-refresh'");
                return $SanUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $SanUtil::FALSE;
        }
    }    # end foreach arg

    my $reference = $self->{clients};

    # Make sure we can get an array. If there are no elements, it would be a
    # string ""
    if (ref($reference) eq "ARRAY") {
        @clients = @$reference;
    }

    # Scan each client object
    foreach my $client (@clients) {

        # Grab the callback array from the client.
        # During login, FitMgr saves a 'callback' which can be used to
        # logout of the client.  It calls $client->deconfigure().
        # This code simply retrieves the saved callbacks and executes them.
        @callbacks = @{$client->{deconfigure}};

        # Execute each callback to deconfigure the client
        foreach my $callback (@callbacks) {
            if ($callback == {}) {
                next;
            } else {
                $callback->();
                if ($self->error()) {
                    $self->_set_err("Deconfigure callback failed",
                                    $self->errmsg());
                    return $SanUtil::FALSE;
                }
            }
        } ## end foreach my $callback (@callbacks)

        # Remove the callbacks from the client, they're no longer valid
        $client->{deconfigure} = [];

        # If user specified a refresh, then refresh the devices.
        if ($refresh) {
            $self->get_devices(-client  => $client->{name},
                               -refresh => 1);
            if ($self->error()) {
                $self->_set_err("Failed to get_devices", $self->errmsg());
                return $SanUtil::FALSE;
            }
        }

    } ## end foreach my $client (@clients)

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done logout_clients."));

    return $SanUtil::TRUE;
} ## end sub logout_clients

###############################################################################
# P R I V A T E   M E T H O D S
###############################################################################

###############################################################################
# Method:     _validate_xml
# Objective:  Validates an XML file. (Exists, well formed, SCHEMA)
# Parameters: None.
# Returns:    True if the document is valid, false otherwise.
###############################################################################
sub _validate_xml {
    my ($self, %args) = @_;

    # Local Variables.
    my $file = $self->{file};
    my $parser;
    my $xml_data;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start validating XML."));
    # First validation: Ensure the file exists.
    unless (-e $file) {
        $self->_set_err("The XML file does not exist: " . $file);
        return $SanUtil::FALSE;
    }

    # Second validation: Check whether an XML file is well formed or not.
    # We don't give a hoot about missing or incorrect elements/attributes here.
    $parser = new XML::LibXML;

    # Try to parse it and set the error if the document is invalid.
    $xml_data = eval { $parser->parse_file($file); };
    if ($@) {
        $self->_set_err("XML file is malformed. Check for missing tags.");
        return $SanUtil::FALSE;
    }

    # Third validation: TODO-> I'm still defining the schema: element names,
    # attribute values, etc. Once we have that I'll uncomment these lines and
    # put the path to the schema file.
    # $xml_data = $parser->parse_file($file);
    # $xml_schema = XML::LibXML::Schema->new( location => $schema_file );
    # eval { $xmlschema->validate($xml_data); };

    # Now that we've validated the XML, store the data in the object's attribute
    $self->{xml_data} = $xml_data;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done validating XML."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_nodes
# Objective:  Creates San Filer on the XML data.
# Parameters: A hash with one mandatory key: a list of XML nodes.
# Returns:    True if it succeeded, False otherwise.
###############################################################################
sub _create_nodes {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my ($filer_obj, $node_data, $cluster_data);
    my ($parent_node, $parent_name, $xml_node_ref);
    my (@clusters, @nodes, @node_xml_nodes) = ();
    my %attributes = ();

   $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating nodes."));

    # Get the XML nodes describing filer nodes or filers. No need to verify
    # because this method will always be called only if $args{nodes} is defined
    $xml_node_ref = $args{nodes};
    @node_xml_nodes = @$xml_node_ref;

    # Go through each node, convert the XML attributes into valid parameters for
    # the libraries. Change the name of the key to filer because the lib expects
    # -filer and not -name. NOTE: We're using name in XML to simplify things.

    foreach my $node (@node_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $node);

        if (defined($attributes{name})) {
            $attributes{filer} = $attributes{name};
            delete($attributes{name});
        }
        # set debug level attr
        unless (exists ($attributes{debug_level})) {
            $attributes{debug_level} = $self->get_debug_level();
        }
        # set cmd interface
        unless (exists ($attributes{cmd_interface})) {
            $attributes{cmd_interface} = $self->{cmd_interface};
        }

        # Create a SanFiler obj for each node.
        $filer_obj = new SanFiler(%attributes);
        if ($filer_obj->error()) {
            $self->_set_err("Failed to create a San Filer object. " ,
                            $filer_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # We need to see which cluster is the parent XML node of this filer
        $parent_node = $node->parentNode;
        $parent_name = $parent_node->getAttribute('name');

        # Put the node/filer data together and push it into a list.
        $node_data = {filer_obj => $filer_obj,
                      name      => $attributes{filer},
                      xml_node  => $node};
        push (@nodes, $node_data);

        # Do the same thing with the cluster. Append the node data here as well
        $cluster_data = {name  => $parent_name,
                         nodes => $node_data};
        push (@clusters, $cluster_data);
    }

    # Update the object's instance variable with the nodes. This is important to
    # get this data easily.
    $self->{clusters} = \@clusters;
    $self->{nodes} = \@nodes;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating nodes."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_aggregates
# Objective:  Creates aggregates based on the XML data.
# Parameters: A hash with the 'aggregates' key containing all aggr XML nodes.
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_aggregates {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@aggregate_xml_nodes, @disk_objs, @aggregates) = ();
    my ($filer_obj, $aggr_name, $num_disks, $state, $owner, $disk_type);
    my ($aggregate_obj, $aggregate_data, $xml_node_ref);
    my (%attributes);
    my $reused_obj = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating aggrs."));

    # Similar to what we did with nodes, get the list of XML nodes for aggrs.
    $xml_node_ref = $args{aggregates};
    @aggregate_xml_nodes = @$xml_node_ref;

    # Go through each aggregate, convert the XML attributes into valid
    # parameters for the libraries. There are several attributes needed to get
    # disk objects but that are not valid as parameters for the aggregate,
    # remove them if needed.

    foreach my $aggregate (@aggregate_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $aggregate);
        # An aggregate will be contained inside a node, get the filer object
        # from the parent node because it will be used to create the aggr.
        # Skip this (avoid the error!) if we cannot find a filer obj!
        $filer_obj = $self->_get_filer_object(xml_node => $aggregate);
        unless ($filer_obj) { next; }

        # Aggregate will complain if we use name, so change it to aggregate.
        if (defined($attributes{name})) {
            $attributes{aggregate} = $attributes{name};
            $aggr_name = $attributes{name};
            delete($attributes{name});
        }

        # This element can be cloned (i.e create n aggrs with the same
        # attributes. Remove it from the parameters.
        if (defined($attributes{loop})) {
            delete($attributes{loop});
        }

        # Remove the index attribute because it's no longer needed here.
        if (defined($attributes{index})) {
            delete($attributes{index});
        }

        # Get the number of disks and remove it from the params
        if (defined($attributes{num_disks})) {
            $num_disks = $attributes{num_disks};
            delete($attributes{num_disks});
        }

        # Get the state of the disks
        if (defined($attributes{state})) {
            $state = $attributes{state};
            delete($attributes{state});
        }

        # Get the owner of the disks (by default it should be this node but we
        # don't need to limit the users... asuming Mike knows what he's doing!)
        if (defined($attributes{owner})) {
            $owner = $attributes{owner};
            delete($attributes{owner});
        } else {
            $owner = $filer_obj->get_name();
        }

        # Get the type of the disks
        if (defined($attributes{disk_type})) {
            $disk_type = $attributes{disk_type};
            delete($attributes{disk_type});
        }

        # Check if this aggregate exists. If it does reuse it
        if ($filer_obj->aggr_exist($aggr_name)) {
            # Get the aggregate object and cache it
            $aggregate_obj = $filer_obj->get_aggregate(
                aggregate => $attributes{aggregate});
            if ($filer_obj->error()) {
                $self->_set_err("Failed to get the aggregate for Fit." ,
                                $filer_obj->errmsg());
                return $SanUtil::FALSE;
            }
            # Since we are reusing things we didn't create, set a flag to skip
            # cleanup
            $reused_obj = $SanUtil::TRUE;
        } else {
            my %disk_args = (state => $state,
                             owner => $owner,
                             count => $num_disks);

            # Add disk_type if user specifies
            %disk_args = (%disk_args, type => $disk_type) if (defined($disk_type));

            # Get the disk objects
            @disk_objs = $filer_obj->get_disk_objs(%disk_args);
            if ($filer_obj->error() or (scalar(@disk_objs) == 0)) {
                $self->_set_err("Failed to get disks for this aggr." . $filer_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # If there are no disks available, we cannot create an aggr.
            unless (scalar(@disk_objs)) {
                $self->_set_err("No disks available for this aggregate.");
                return $SanUtil::FALSE;
            }

            # Create an aggregate object. Catch any errors
            $aggregate_obj = $filer_obj->create_aggregate_obj(%attributes,
                disk_objs => [@disk_objs]);
            if ($filer_obj->error()) {
                $self->_set_err("Failed to create the aggregate for Fit." ,
                                $filer_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }
        # Cache aggregate data
        $aggregate_data = {filer_obj  => $filer_obj,
                           name       => $attributes{aggregate},
                           xml_node   => $aggregate,
                           aggr_obj   => $aggregate_obj,
                           reused_obj => $reused_obj};
        push (@aggregates, $aggregate_data);
    }

    # Update the object's instance variable.
    $self->{aggregates} = \@aggregates;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating aggrs."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_vservers
# Objective:  Creates vservers based on the XML data.
# Parameters: A hash with 'vservers' as a key containing a list of vserver XML
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_vservers {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@vserver_xml_nodes, @vservers) = ();
    my ($filer_obj, $vserver_name, $aggregate_name, $aggregate_node);
    my ($filer_name, $vserver_obj, $vserver_data, $iscsi_enabled, $iscsi_obj);
    my ($fcp_enabled, $fcp_obj, $xml_node_ref);
    my %attributes;
    my $reused_obj = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating vservers."));

    # Similar to what we did with nodes, get the list of XML nodes for vservers.
    $xml_node_ref = $args{vservers};
    @vserver_xml_nodes = @$xml_node_ref;

    # Go through each vserver, convert the XML attributes into valid
    # parameters for the libraries.

    foreach my $vserver (@vserver_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $vserver);
        # Initialize iSCSI and FCP services here to avoid enabling services on
        # incorrect vservers
        $fcp_enabled = $SanUtil::FALSE;
        $iscsi_enabled = $SanUtil::FALSE;
        # Users may specify the aggr as an attribute for the vserver node.
        # However, the most common case it to get it from the outer aggr node.
        # So make sure we handle both cases.
        unless (defined ($attributes{aggregate})) {
            $aggregate_node = $vserver->parentNode;
            $aggregate_name = $aggregate_node->getAttribute('name');
            $attributes{aggregate} = $aggregate_name;
        }

        # We'll need a filer object, get it form the parent node.
        $filer_obj = $self->_get_filer_object(xml_node => $aggregate_node);
        unless ($filer_obj) { next; }

        # Append the name of the filer
        $attributes{filer} = $filer_obj->get_name();

        # Change name for vserver.
        if (defined($attributes{name})) {
            $attributes{vserver} = $attributes{name};
            $vserver_name = $attributes{vserver};
            delete($attributes{name});
        }

        # This element can be cloned (i.e create n vservers with the same
        # attributes. Remove it from the parameters.
        if (defined($attributes{loop})) {
            delete($attributes{loop});
        }

        # Remove the index attribute because it's no longer needed here.
        if (defined($attributes{index})) {
            delete($attributes{index});
        }

        # See if they want to create iscsi service
        if (defined($attributes{iscsi_enabled})) {
            $iscsi_enabled = $attributes{iscsi_enabled};
            delete($attributes{iscsi_enabled});
        }

        # See if they want to create fcp service
        if (defined($attributes{fcp_enabled})) {
            $fcp_enabled = $attributes{fcp_enabled};
            delete($attributes{fcp_enabled});
        }

        # Check if it already exists. If it does we'll reuse it.
        if ($filer_obj->vserver_exist(name => $vserver_name)) {
            $vserver_obj = $filer_obj->get_vserver_obj(
                vserver => $attributes{vserver});
            if ($filer_obj->error()) {
                $self->_set_err("Failed to get the vserver for Fit." ,
                                $filer_obj->errmsg());
                return $SanUtil::FALSE;
            }
            # Set the flag to indicate it's being reused
            $reused_obj = $SanUtil::TRUE;
        } else {
            # Create a vserver object. Catch any errors
            $vserver_obj = $filer_obj->create_vserver_obj(%attributes);
            if ($filer_obj->error()) {
                $self->_set_err("Failed to create a vserver for Fit." ,
                                $filer_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # If a user specified iscsi_enabled, turn it on here.
        if ($iscsi_enabled) {
            unless ($vserver_obj->iscsi_exist()) {
                $iscsi_obj = $vserver_obj->create_iscsi_obj(
                    vserver => $attributes{vserver});
                if ($vserver_obj->error()) {
                    $self->_set_err("Failed to create iSCSI service." ,
                                    $vserver_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            }
        }

        # Same thing for FCP
        if ($fcp_enabled) {
            unless ($vserver_obj->fcp_exist()) {
                $fcp_obj = $vserver_obj->create_fcp_obj(
                    vserver => $attributes{vserver});
                if ($vserver_obj->error()) {
                    $self->_set_err("Failed to create FCP service." ,
                                    $vserver_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            }
        }

        # Cache vserver data
        $vserver_data = {filer_obj   => $filer_obj,
                         name        => $attributes{vserver},
                         xml_node    => $vserver,
                         vserver_obj => $vserver_obj,
                         reused_obj  => $reused_obj};
        push (@vservers, $vserver_data);
    }

    # Update the object's instance variable.
    $self->{vservers} = \@vservers;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating vservers."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_lifs
# Objective:  Creates lifs based on the XML data.
# Parameters: A hash with 'lifs' as a key containing a list of LIF XML
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_lifs {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@lif_xml_nodes, @lifs, @addresses, @ports, @lif_names, @netmasks, @ips) = ();
    my ($filer_obj, $lif_name, $vserver_name, $vserver_node, $xml_node_ref);
    my ($vserver_obj, $home_port, $address, $data_protocol, $loop, $lif_obj);
    my ($lif_data, $port);
    my (%attributes);
    my $reused_obj = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating LIFs"));

    # Get the LIF nodes.
    $xml_node_ref = $args{lifs};
    @lif_xml_nodes = @$xml_node_ref;

    foreach my $lif (@lif_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $lif);

        # We need the vserver name
        unless (defined ($attributes{vserver})) {
            $vserver_node = $lif->parentNode;
            $vserver_name = $vserver_node->getAttribute('name');
            $attributes{vserver} = $vserver_name;
        }

        # Default to data LIF if it has not been defined.
        unless (defined ($attributes{role})) {
            $attributes{role} = 'data';
        }

        # Get the data protocol for this LIF.
        if (defined ($attributes{'data-protocol'})) {
            $data_protocol = $attributes{'data-protocol'};
        }

        # Get the data port for this LIF.
        if (defined ($attributes{'port'})) {
            $port = $attributes{'port'};
        }

        # Get the number of loops and delete it from the args
        if (defined ($attributes{loop})) {
            $loop = $attributes{loop};
            delete($attributes{loop});
        }

        # Get the addresses for iSCSI LIFs
        if (defined ($attributes{address})) {
            $address = $attributes{address};
            @addresses = split(/,/, $address);
            # If there's more than 1 IP address, use the first as default value
            if (scalar(@addresses) > 1) {
                if ($loop > scalar(@addresses)) {
                    $self->_set_err("Not enough IP addresses to create LIFs.");
                    return $SanUtil::FALSE;
                }
                $attributes{address} = $addresses[0];
            } else {
                $attributes{address} = $address;
            }
            push (@ips, $attributes{address});
        }

        # Get the netmask fof iscsi
        if (defined ($attributes{netmask})) {
            push (@netmasks, $attributes{netmask});
        }

        # We'll need a filer object, get it form the parent node.
        $filer_obj = $self->_get_filer_object(xml_node => $lif);
        unless ($filer_obj) { next; }

        # Change name for lif.
        if (defined($attributes{name})) {
            $attributes{lif} = $attributes{name};
            $lif_name = $attributes{lif};
            push(@lif_names, $lif_name);
            delete($attributes{name});
        }

        # Get the vserver object
        $vserver_obj = $filer_obj->get_vserver_obj(vserver => $attributes{vserver});
        if ($filer_obj->error()) {
            $self->_set_err("Failed to get the vserver." , $filer_obj->errmsg());
            return $SanUtil::FALSE;
        }

        if ($vserver_obj->lif_exist(name => $lif_name)) {
            $lif_obj = $vserver_obj->get_lif_obj(lif => $lif_name);
            if ($vserver_obj->error()) {
                $self->_set_err("Failed to get the lif." , $vserver_obj->errmsg());
                return $SanUtil::FALSE;
            }
            # Mark it as reused
            $reused_obj = $SanUtil::TRUE;
        } else {
            $lif_obj = $vserver_obj->create_lif_obj(%attributes);
            if ($vserver_obj->errmsg() =~ /duplicate ip address/i) {
                if (scalar(@addresses) > 1) {
                    for (my $i = 1; $i < scalar(@addresses); $i++) {
                        $attributes{address} = $addresses[$i];
                        $lif_obj = $vserver_obj->create_lif_obj(%attributes);
                        # try the next ip
                        next if ($vserver_obj->error());
                        # we got here that means the creation is ok
                        # clear error and get out of this loop
                        $vserver_obj->_clear_err();
                        last;
                    }
                }
            }
            if ($vserver_obj->error()) {
                $self->_set_err("Failed to create a lif." , $vserver_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # Store the LIF attributes, either FCP or iSCSI, in the vserver object
        # This is needed in order to load fcp or iscsi targets using the right
        # data.
        if ($data_protocol =~ /fcp/i) {
            # force copy the hash
            my %fcp_attributes = %attributes;
            push (@{$vserver_obj->{fcp_lif_data}}, \%fcp_attributes);
        } elsif ($data_protocol =~ /iscsi/i) {
            my %iscsi_attributes = %attributes;
            push (@{$vserver_obj->{iscsi_lif_data}}, \%iscsi_attributes);
        }

        # Now cache the LIF data
        $lif_data = {filer_obj  => $filer_obj,
                     name       => $attributes{lif},
                     xml_node   => $lif,
                     lif_obj    => $lif_obj,
                     reused_obj => $reused_obj};
        push (@lifs, $lif_data);
    }

    # Update the object's instance variable.
    $self->{lifs} = \@lifs;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating LIFs."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_volumes
# Objective:  Creates volumesbased on the XML data.
# Parameters: A hash with 'volumes' as a key containing a list of volume XML
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_volumes {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@volume_xml_nodes, @volumes) = ();
    my ($filer_obj, $vserver_name, $volume_name, $volume_node, $aggregate_name);
    my ($filer_name, $volume_obj, $volume_data, $vserver_node, $aggregate_node);
    my ($percent_snapshot_space);
    my ($xml_node_ref);
    my %attributes;
    my $reused_obj = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating volumes."));

    # Get all volume nodes in the XML.
    $xml_node_ref = $args{volumes};
    @volume_xml_nodes = @$xml_node_ref;

    # Go through each volume, convert the XML attributes into valid
    # parameters for the volume library

    foreach my $volume (@volume_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $volume);
        # We'll need both a vserver and an aggregate name. See if they were
        # defined in the XML. If they were not then get then from the parent and
        # grandparent nodes respectively.
        unless (defined ($attributes{vserver})) {
            $vserver_node = $volume->parentNode;
            $vserver_name = $vserver_node->getAttribute('name');
            $attributes{vserver} = $vserver_name;
        }

        unless (defined ($attributes{aggregate})) {
            $vserver_node = $volume->parentNode;
            $aggregate_node = $vserver_node->parentNode;
            $aggregate_name = $aggregate_node->getAttribute('name');
            $attributes{aggregate} = $aggregate_name;
        }

        # We'll need a filer object. Locate the filer node in the XML file and
        # use it
        $filer_obj = $self->_get_filer_object(xml_node => $volume);
        unless ($filer_obj) { next; }

        # Change name for volume
        if (defined($attributes{name})) {
            $attributes{volume} = $attributes{name};
            $volume_name = $attributes{volume};
            delete($attributes{name});
        }

        # This element can be cloned (i.e create n vols with the same
        # attributes. Remove it from the parameters.
        if (defined($attributes{loop})) {
            delete($attributes{loop});
        }

        # Remove the index attribute because it's no longer needed here.
        if (defined($attributes{index})) {
            delete($attributes{index});
        }

        # Save/Remove the percent_snapshot_space attribute because it's no longer needed here.
        if (defined($attributes{percent_snapshot_space})) {
            $percent_snapshot_space = $attributes{percent_snapshot_space};
            delete($attributes{percent_snapshot_space});
        }

        # Check if it already exists. If it does we'll reuse it.
        if ($filer_obj->volume_exist(name    => $attributes{volume},
                                     vserver => $attributes{vserver})) {
            # Get the volume object
            $volume_obj = $filer_obj->get_vol(volume => $attributes{volume},
                                              vserver => $attributes{vserver});
            if ($filer_obj->error()) {
                $self->_set_err("Failed to get the volume for Fit." ,
                                $filer_obj->errmsg());
                return $SanUtil::FALSE;
            }
            # Set the reused flag
            $reused_obj = $SanUtil::TRUE;
        } else {
            # Create a volume object. Catch any errors (no need to append err)
            $volume_obj = $filer_obj->create_volume(%attributes);
            if ($filer_obj->error()) {
                $self->_set_err("Failed to create a volume for Fit." ,
                                $filer_obj->errmsg());
                return $SanUtil::FALSE;
            }
            # If the user specified snapshot % for the volume
            # we will modify this attribute here since that attribute
            # cannot be specified at the time the volume is created.
            if(defined($percent_snapshot_space)){
                $volume_obj->modify( percent_snapshot_space => "$percent_snapshot_space");
                if ($volume_obj->error()) {
                    $self->_set_err("Failed to modify volume $attributes{volume} for Fit." ,
                                    $volume_obj->errmsg());
                    return $SanUtil::FALSE;
                }
            }
        }

        # Cache volume data
        $volume_data = {filer_obj  => $filer_obj,
                        name       => $attributes{volume},
                        xml_node   => $volume,
                        volume_obj => $volume_obj,
                        reused_obj => $reused_obj};
        push (@volumes, $volume_data);
    }

    # Update the object's instance variable.
    $self->{volumes} = \@volumes;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating volumes."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_igroups
# Objective:  Creates igroups based on the XML data.
# Parameters: A hash with 'igroups' as a key containing a list of igroups XML
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_igroups {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@igroup_xml_nodes, @igroups) = ();
    my ($filer_obj, $vserver_name, $igroup_name, $igroup_node, $lun_count);
    my ($filer_name, $igroup_obj, $igroup_data, $vserver_node, $exists);
    my ($xml_node_ref, $vserver_obj);
    my %attributes;
    my $reused_obj = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating igroups."));

    # Get all igroup nodes in the XML.
    $xml_node_ref = $args{igroups};
    @igroup_xml_nodes = @$xml_node_ref;

    # Go through each igroup, convert the XML attributes into valid
    # parameters for the igroup library

    foreach my $igroup (@igroup_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $igroup);

        # We'll need the vserver (parent node).
        unless (defined ($attributes{vserver})) {
            $vserver_node = $igroup->parentNode;
            $vserver_name = $vserver_node->getAttribute('name');
            $attributes{vserver} = $vserver_name;
        }

        # We'll need a filer object. Locate the filer node in the XML file and
        # use it
        $filer_obj = $self->_get_filer_object(xml_node => $igroup);
        unless ($filer_obj) { next; }

        # Change name for igroup
        if (defined($attributes{name})) {
            $attributes{igroup} = $attributes{name};
            $igroup_name = $attributes{igroup};
            delete($attributes{name});
        }

        # This element can be cloned (i.e create n igroups with the same
        # attributes. Remove it from the parameters.
        if (defined($attributes{loop})) {
            delete($attributes{loop});
        }

        # Remove the index attribute because it's no longer needed here.
        if (defined($attributes{index})) {
            delete($attributes{index});
        }

        # Get the number of LUNS per igroup (this is optional)
        if (defined($attributes{lun_count})) {
            $lun_count = $attributes{lun_count};
            delete($attributes{lun_count});
        }

        # Check if it already exists. If it does we'll reuse it.
        $exists = $filer_obj->igroup_exist(igroup  => $attributes{igroup},
                                           vserver => $attributes{vserver});

        # Check if it already exists. If it does we'll reuse it.
        if ($exists) {
            # Get the igroup object
            $igroup_obj = $filer_obj->get_igroup(
                igroup  => $attributes{igroup},
                vserver => $attributes{vserver});
            if ($filer_obj->error()) {
                $self->_set_err("Failed to get the igroup for Fit." ,
                                $filer_obj->errmsg());
                return $SanUtil::FALSE;
            }
            # Set reused flag
            $reused_obj = $SanUtil::TRUE;
        } else {
            # Create an igroup object. Catch any errors (no need to append err)
            $vserver_obj = $filer_obj->get_vserver_obj(name => $attributes{vserver});
            $igroup_obj = $vserver_obj->create_igroup_obj(%attributes);
            if ($vserver_obj->error()) {
                $self->_set_err("Failed to create an igroup for Fit." ,
                                $vserver_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # Map the luns only if we have enough LUNs available and a user
        # specified it in the XML file. Otherwise, just create the igroups
        # and forget about the mapping
        my $lun_map_data = {igroup_name  => $attributes{igroup},
                            lun_count    => $lun_count,
                            vserver_name => $attributes{vserver},
                            xml_node     => $igroup};
        push(@Lun_Map_Info, $lun_map_data);

        # Cache igroup data
        $igroup_data = {filer_obj  => $filer_obj,
                        name       => $attributes{igroup},
                        xml_node   => $igroup,
                        igroup_obj => $igroup_obj,
                        reused_obj => $reused_obj};
        push (@igroups, $igroup_data);
    }

    # Update the object's instance variable.
    $self->{igroups} = \@igroups;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating iGroups."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_portsets
# Objective:  Creates portsets based on the XML data.
# Parameters: A hash with 'portsets' as a key containing a list of portset XML
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_portsets {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@portset_xml_nodes, @portsets) = ();
    my ($filer_obj, $vserver_name, $portset_node, $portset_data);
    my ($filer_name, $portset_obj, $vserver_node, $vserver_obj);
    my ($igroup, $igroup_obj);
    my %attributes;
    my $reused_obj = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating portsets."));

    # Get all portset nodes in the XML.
    @portset_xml_nodes = @{$args{portsets}};

    # Go through each portset, convert the XML attributes into valid
    # parameters for the portset library

    foreach my $portset (@portset_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $portset);

        # We'll need the vserver (parent node).
        unless (defined ($attributes{vserver})) {
            $vserver_node = $portset->parentNode;
            $vserver_name = $vserver_node->getAttribute('name');
            $attributes{vserver} = $vserver_name;
        }

        # We'll need a filer object. Locate the filer node in the XML file and
        # use it
        $filer_obj = $self->_get_filer_object(xml_node => $portset);
        unless ($filer_obj) { next; }

        # Change name for portset
        if (defined($attributes{name})) {
            $attributes{portset} = $attributes{name};
            delete($attributes{name});
        }

        # This element can be cloned (i.e create n portsets with the same
        # attributes. Remove it from the parameters.
        if (defined($attributes{loop})) {
            delete($attributes{loop});
        }

        # Remove the index attribute because it's no longer needed here.
        if (defined($attributes{index})) {
            delete($attributes{index});
        }

        # if they specify an igroup, we'll bind it later and need to remove it
        # from the args
        if (defined($attributes{igroup})) {
            $igroup = delete ($attributes{igroup});
        }

        $vserver_obj = $filer_obj->get_vserver_obj(name => $attributes{vserver});
        if ($filer_obj->error()) {
            $self->_set_err("Failed to get vserver object to create a portset.",
                             $filer_obj->errmsg());
            return $SanUtil::FALSE;
        }

        if ($vserver_obj->portset_exist(portset => $attributes{portset})) {
            # Get the portset object
            $portset_obj = $vserver_obj->get_portset_obj(portset => $attributes{portset});
            if ($vserver_obj->error()) {
                $self->_set_err("Failed to get the portset for Fit: ",
                                $vserver_obj->errmsg());
                return $SanUtil::FALSE;
            }
            # Set reused flag
            $reused_obj = $SanUtil::TRUE;
        } else {
            # Create a portset object. Catch any errors (no need to append err)
            $portset_obj = $vserver_obj->create_portset_obj(%attributes);
            if ($vserver_obj->error()) {
                $self->_set_err("Failed to create a portset for Fit." ,
                                $vserver_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # If users specify an igroup name in the XML, bind it
        if ($igroup) {
            # Try to get an igroup object first
            $igroup_obj = $filer_obj->get_igroup(igroup  => $igroup,
                                                 vserver => $attributes{vserver});
            if ($filer_obj->error()) {
                $self->_set_err("Failed to get an igroup and bind it to a portset." ,
                                $filer_obj->errmsg());
                return $SanUtil::FALSE;
            }

            # Now try binding it
            $igroup_obj->bind(portset_name => $attributes{portset});
            if ($igroup_obj->error()) {
                $self->_set_err("Failed to bind an igroup to a portset in FIT Mgr." ,
                                $igroup_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # Cache portset data
        $portset_data = {filer_obj   => $filer_obj,
                         name        => $attributes{portset},
                         xml_node    => $portset,
                         portset_obj => $portset_obj,
                         reused_obj  => $reused_obj};
        push (@portsets, $portset_data);
    }

    # Update the object's instance variable.
    $self->{portsets} = \@portsets;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating portsets."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_luns
# Objective:  Creates LUNs based on the XML data.
# Parameters: A hash with the 'luns' key containing all LUN XML nodes.
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_luns {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@lun_xml_nodes, @luns) = ();
    my ($filer_obj, $vserver_name, $volume_name, $volume_node, $lun_name);
    my ($filer_name, $lun_obj, $lun_data, $vserver_node, $lun_node);
    my ($volume_obj, $exists, $xml_node_ref);
    my %attributes;
    my $reused_obj = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating LUNs."));

    # Get all LUN nodes in the XML.
    $xml_node_ref = $args{luns};
    @lun_xml_nodes = @$xml_node_ref;

    # Go through each LUN, convert the XML attributes into valid
    # parameters for the LUN library
    foreach my $lun (@lun_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $lun);
        # For LUNs we'll need the vserver name and a volume name.
        unless (defined ($attributes{vserver})) {
            $vserver_node = $lun->parentNode->parentNode;
            $vserver_name = $vserver_node->getAttribute('name');
            $attributes{vserver} = $vserver_name;
        }

        unless (defined ($attributes{vol})) {
            $volume_node = $lun->parentNode;
            $volume_name = $volume_node->getAttribute('name');
            $attributes{volume} = $volume_name;
        }

        # We'll need a filer object. Locate the filer node in the XML file and
        # use it
        $filer_obj = $self->_get_filer_object(xml_node => $lun);
        unless ($filer_obj) { next; }

        # Change name for LUN
        if (defined($attributes{name})) {
            $attributes{lun} = $attributes{name};
            $lun_name = $attributes{lun};
            delete($attributes{name});
        }

        # This element can be cloned (i.e create n igroups with the same
        # attributes. Remove it from the parameters.
        if (defined($attributes{loop})) {
            delete($attributes{loop});
        }

        # Remove the index attribute because it's no longer needed here.
        if (defined($attributes{index})) {
            delete($attributes{index});
        }

        # To create a LUN, we must invoke it from a vol object, so get one
        $volume_obj = $filer_obj->get_vol(volume  => $attributes{volume},
                                          vserver => $attributes{vserver});
        if ($filer_obj->error()) {
            $self->_set_err("Failed to get the volume for Fit." ,
                            $filer_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # Check if it already exists. If it does we'll reuse it.
        $exists = $volume_obj->lun_exist(lun     => $attributes{lun},
                                         volume  => $attributes{volume},
                                         vserver => $attributes{vserver});

        if ($exists) {
            $lun_obj = $volume_obj->get_lun(lun     => $attributes{lun},
                                            volume  => $attributes{volume},
                                            vserver => $attributes{vserver});
            if ($volume_obj->error()) {
                $self->_set_err("Failed to get a LUN for Fit." ,
                                $volume_obj->errmsg());
                return $SanUtil::FALSE;
            }
            $reused_obj = $SanUtil::TRUE;
        } else {
            # Create a LUN object. Catch any errors (no need to append err)
            $lun_obj = $volume_obj->create_lun(%attributes);
            if ($volume_obj->error()) {
                $self->_set_err("Failed to create a LUN for Fit." ,
                                $volume_obj->errmsg());
                return $SanUtil::FALSE;
            }
        }

        # Put the new LUNs in the pool and cache them
        push(@Lun_Object_Pool, $lun_obj);
        $lun_data = {filer_obj  => $filer_obj,
                     name       => $attributes{lun},
                     xml_node   => $lun,
                     is_mapped  => $SanUtil::FALSE,
                     lun_obj    => $lun_obj,
                     reused_obj => $reused_obj};
        push (@luns, $lun_data);
    }
    # Update the object's instance variable and the global pool (used to map)
    $self->{luns} = \@luns;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating LUNS."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_iscsi
# Objective:  Creates iscsi based on the XML data.
# Parameters: A hash with 'iscsi' as a key containing a list of vserver XML
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_iscsi {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@vserver_xml_nodes, @iscsi) = ();
    my ($filer_obj, $vserver_name, $aggregate_name, $aggregate_node);
    my ($filer_name, $vserver_obj, $vserver_data, $iscsi_enabled, $iscsi_obj);
    my ($fcp_enabled, $fcp_obj, $xml_node_ref);
    my %attributes;
    my $reused_obj = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating iscsi."));

    # Similar to what we did with nodes, get the list of XML nodes for iscsi.
    $xml_node_ref = $args{iscsi};
    @vserver_xml_nodes = @$xml_node_ref;

    # Go through each vserver, convert the XML attributes into valid
    # parameters for the libraries.

    foreach my $vserver (@vserver_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $vserver);

        # We'll need a filer object, get it form the parent node.
        $filer_obj = $self->_get_filer_object(xml_node => $vserver);
        unless ($filer_obj) { next; }

        # Append the name of the filer
        $attributes{filer} = $filer_obj->get_name();

        # Change name for vserver.
        if (defined($attributes{name})) {
            $attributes{vserver} = $attributes{name};
            $vserver_name = $attributes{vserver};
            delete($attributes{name});
        }

	$vserver_obj = $filer_obj->get_vserver_obj(
	    vserver => $attributes{vserver});
	if ($filer_obj->error()) {
	    $self->_set_err("Failed to get the vserver for Fit." ,
			    $filer_obj->errmsg());
	    return $SanUtil::FALSE;
	}

	unless ($vserver_obj->iscsi_exist()) {
	    $iscsi_obj = $vserver_obj->create_iscsi_obj(
		vserver => $attributes{vserver});
	    if ($vserver_obj->error()) {
		$self->_set_err("Failed to create iSCSI service." ,
				$vserver_obj->errmsg());
		return $SanUtil::FALSE;
	    }
	}
    }

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating iscsi."));
    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_fcp
# Objective:  Creates fcp based on the XML data.
# Parameters: A hash with 'fcp' as a key containing a list of vserver XML
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_fcp {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@vserver_xml_nodes, @fcp) = ();
    my ($filer_obj, $vserver_name, $aggregate_name, $aggregate_node);
    my ($filer_name, $vserver_obj, $vserver_data);
    my ($fcp_enabled, $fcp_obj, $xml_node_ref);
    my %attributes;
    my $reused_obj = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating fcp."));

    # Similar to what we did with nodes, get the list of XML nodes for fcp.
    $xml_node_ref = $args{fcp};
    @vserver_xml_nodes = @$xml_node_ref;

    # Go through each vserver, convert the XML attributes into valid
    # parameters for the libraries.

    foreach my $vserver (@vserver_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $vserver);

        # We'll need a filer object, get it form the parent node.
        $filer_obj = $self->_get_filer_object(xml_node => $aggregate_node);
        unless ($filer_obj) { next; }

        # Append the name of the filer
        $attributes{filer} = $filer_obj->get_name();

        # Change name for vserver.
        if (defined($attributes{name})) {
            $attributes{vserver} = $attributes{name};
            $vserver_name = $attributes{vserver};
            delete($attributes{name});
        }

	    $vserver_obj = $filer_obj->get_vserver_obj(
	        vserver => $attributes{vserver});
	    if ($filer_obj->error()) {
	        $self->_set_err("Failed to get the vserver for Fit." ,
			        $filer_obj->errmsg());
	        return $SanUtil::FALSE;
	    }

	    unless ($vserver_obj->fcp_exist()) {
	        $fcp_obj = $vserver_obj->create_fcp_obj(
		    vserver => $attributes{vserver});
	        if ($vserver_obj->error()) {
		        $self->_set_err("Failed to create iSCSI service." ,
			    $vserver_obj->errmsg());
		        return $SanUtil::FALSE;
	        }
	    }
    }

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating fcp."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_tpgroup
# Objective:  Creates tpgroup based on the XML data.
# Parameters: A hash with 'tpgroup' as a key containing a list of vserver XML
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_tpgroup {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@tpgroup_xml_nodes, @tpgroup) = ();
    my ($filer_obj, $vserver_name, $tpgroup_data);
    my ($filer_name, $tpgroup_obj, $vserver_node, $vserver_obj, $iscsi_obj);
    my %attributes;
    my $reused_obj = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating tpgroup."));

    # Get all tpgroup nodes in the XML.
    @tpgroup_xml_nodes = @{$args{tpgroup}};

    # Go through each tpgroup, convert the XML attributes into valid
    # parameters for the tpgroup library

    foreach my $tpgroup (@tpgroup_xml_nodes) {
        %attributes = $self->_get_xml_attributes(xml_node => $tpgroup);

        # We'll need the vserver (parent node).
        unless (defined ($attributes{vserver})) {
            $vserver_node = $tpgroup->parentNode;
            $vserver_name = $vserver_node->getAttribute('name');
            $attributes{vserver} = $vserver_name;
        }

        # We'll need a filer object. Locate the filer node in the XML file and
        # use it
        $filer_obj = $self->_get_filer_object(xml_node => $tpgroup);
        unless ($filer_obj) { next; }

        # Change name for tpgroup
        if (defined($attributes{name})) {
            $attributes{tpgroup} = $attributes{name};
            delete($attributes{name});
        }

        # This element can be cloned (i.e create n tpgroup with the same
        # attributes. Remove it from the parameters.
        if (defined($attributes{loop})) {
            delete($attributes{loop});
        }

        # Remove the index attribute because it's no longer needed here.
        if (defined($attributes{index})) {
            delete($attributes{index});
        }

        $vserver_obj = $filer_obj->get_vserver_obj(name => $attributes{vserver});
        if ($filer_obj->error()) {
            $self->_set_err("Failed to get vserver object to create a tpgroup.",
                             $filer_obj->errmsg());
            return $SanUtil::FALSE;
        }

	unless ($vserver_obj->iscsi_exist()) {
	    $self->_set_err("iSCSI needed for tpgroups." ,
			    $vserver_obj->errmsg());
	    return $SanUtil::FALSE;
	} else {
	    $iscsi_obj = $vserver_obj->get_iscsi_obj();
	    if ($iscsi_obj->tpgroup_exist(tpgroup => $attributes{tpgroup})) {
		# Get the tpgroup object
		$tpgroup_obj = $iscsi_obj->get_tpgroup_obj(tpgroup => $attributes{tpgroup});
		if ($iscsi_obj->error()) {
		    $self->_set_err("Failed to get the tpgroup for Fit: ",
				    $iscsi_obj->errmsg());
		    return $SanUtil::FALSE;
		}
		# Set reused flag
		$reused_obj = $SanUtil::TRUE;
	    } else {
		# Create a tpgroup object. Catch any errors (no need to append err)
		$tpgroup_obj = $iscsi_obj->create_tpgroup_obj(%attributes);
		if ($iscsi_obj->error()) {
		    $self->_set_err("Failed to create a tpgroup for Fit." ,
				    $iscsi_obj->errmsg());
		    return $SanUtil::FALSE;
		}
	    }

	    # Cache tpgroup data
	    $tpgroup_data = {filer_obj   => $filer_obj,
			     name        => $attributes{tpgroup},
			     xml_node    => $tpgroup,
			     tpgroup_obj => $tpgroup_obj,
			     reused_obj  => $reused_obj};
	    push (@tpgroup, $tpgroup_data);
	}
    }

    # Update the object's instance variable.
    $self->{tpgroup} = \@tpgroup;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating tpgroup."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _create_clients
# Objective:  Creates clientsbased on the XML data.
# Parameters: A hash with the 'clients' key containing all client XML nodes.
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _create_clients {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local Variables & Initialization
    my (@client_xml_nodes, @clients, @initiators, @portnames, @filers) = ();
    my ($client_obj, $client_name, $xml_node_ref, $client_data, $filer_name);
    my (%attributes, $target_data, $filers_ref, $protocol, $device_type, $callback);
    my $skip_config = $SanUtil::FALSE;
    my $initiator_protocol;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start creating clients."));

    # Get all client nodes in the XML.
    $xml_node_ref = $args{clients};
    @client_xml_nodes = @$xml_node_ref;

    # Go through each client, convert the XML attributes into valid
    # parameters for the SanClient library
    foreach my $client (@client_xml_nodes) {
        my @callbacks = ();
        @portnames = ();
        %attributes = $self->_get_xml_attributes(xml_node => $client);

        # Need to change name for host to avoid problems
        if (defined($attributes{name})) {
            $attributes{host} = $attributes{name};
            delete($attributes{name});
        }

        # Users may want to skip client configuration
        if (defined($attributes{skip_config})) {
            $skip_config = $attributes{skip_config};
            delete($attributes{skip_config});
        }

        if (defined($attributes{rescan_timeout})) {
            $self->{rescan_timeout} = delete($attributes{rescan_timeout});
        }

        if (defined($attributes{device_type})) {
            $device_type = $attributes{device_type};
        }

        if (defined($attributes{protocol})) {
            $protocol = $attributes{protocol};
        }

        # Our mapping works like this:
        #     target="filer_name:vserver_name:igroup_name"
        # let _extract_target_info do all the magic.
        if (defined($attributes{target})) {
            $target_data = $self->_extract_target_info($attributes{target});
            if ($self->error()) {
                $self->_set_err("Failed to extract target data." ,
                                $self->errmsg());
                return $SanUtil::FALSE;
            }
            $filers_ref = $target_data->{filers};
            @filers = @$filers_ref;
            delete($attributes{target});
        }

        # Set the debug level for the new SanClient object using the same level
        # as the FitMgr object.
        unless (defined($attributes{debug_level})) {
            $attributes{debug_level} = $self->get_debug_level();
        }

        # Create the client object
        $client_obj = new SanClient(%attributes);
        if ($client_obj->error()) {
            $self->_set_err("Failed to create a SanClient object for Fit.",
                            $client_obj->errmsg());
            return $SanUtil::FALSE;
        }

        # If we have target data then try to configure clients and add inits.
        if ($target_data) {
            unless ($skip_config) {
                # Get the initiators of the client
                @initiators = $client_obj->get_initiators();
                if ($client_obj->error()) {
                    $self->_set_err("Failed to get initiators for Fit.",
                                    $client_obj->errmsg());
                    return $SanUtil::FALSE;
                }

                foreach my $initiator (@initiators) {

                    # What protocol is this initiator?
                    $initiator_protocol = $initiator->get_protocol();

                    # If it's fcp, get the portname.
                    if ($initiator_protocol eq "fcp") {
                        push(@portnames, $initiator->get_portname());
                    } elsif ($initiator_protocol eq "iscsi") {

                        # Else if it's iscsi, get the nodename.
                        push(@portnames, $initiator->get_nodename());
                    } else {

                        # Else unknown protocol, return error
                        $self->_set_err(
                              "Initiator $initiator is not using iscsi or fcp");
                        return $SanUtil::FALSE;
                    }
                }

                # Add the initiators
                $target_data->{initiators} = \@portnames;
                $self->_add_initiators_to_igroup($target_data);
                if ($self->error()) {
                    $self->_set_err("Failed to add initiators to the igroup.",
                                    $self->errmsg());
                    return $SanUtil::FALSE;
                }

                # Map the LUNs. We do this here to avoid the infamous lun-map
                # id conflic issue.
                $self->_finalize_lun_mapping();
                if ($self->error()) {
                    $self->_set_err("Failed to finalize LUN mapping." ,
                                    $self->errmsg());
                }

                # Configure and discover
                foreach my $filer (@filers) {
                    # configure_client returns a deconfigure_client callback.
                    # make sure we collect each of these and use them for
                    # cleanup.
                    $callback = $self->_configure_client(client_obj => $client_obj,
                                                         filer_name => $filer,
                                                         protocol   => $protocol,
                                                         vservers   => $target_data->{vservers},
                                                         rescan_timeout => $self->{rescan_timeout});
                    push (@callbacks, $callback);
                    if ($self->error()) {
                        $self->_set_err("Failed to configure $attributes{host}",
                                        $self->errmsg());
                        return $SanUtil::FALSE;
                    }
                }
            }
        }

        $client_data = {client_obj  => $client_obj,
                        name        => $attributes{host},
                        xml_node    => $client,
                        skip_config => $skip_config,
                        deconfigure => \@callbacks};
        push (@clients, $client_data);
    }

    # Update the object's instance
    $self->{clients} = \@clients;
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done creating clients."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _map_luns
# Objective:  Maps LUNs to igroups according to the XML file.
# Parameters: A hash with 'igroup_name' 'lun_count' 'vserver'.
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _map_luns {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Local variables
    my ($igroup_name, $vserver_name, $lun);
    my ($mapped, $index, $lun_count) = 0;
    my @lun_objs;

    # Get the arguments
    if (defined($args{igroup_name})) {
        $igroup_name = $args{igroup_name};
    } else {
        $self->_set_err("iGroup Name is mandatory");
        return $SanUtil::FALSE;
    }

    if (defined($args{vserver_name})) {
        $vserver_name = $args{vserver_name};
    } else {
        $self->_set_err("The vserver name is mandatory");
        return $SanUtil::FALSE;
    }

    if (defined($args{lun_count})) {
        $lun_count = $args{lun_count};
    }

    # We can only map those luns that belong to this igroup's vserver so
    # iterate over the LUN pool, check if they belong to the same vserver and
    # remove them. NOTE: I'm being extra careful here because removing
    # elements (splice) while iterating the list can cause problems.

    while ($mapped < $lun_count) {
        $index = 0;
        while ($index <= $#Lun_Object_Pool) {
            if ($Lun_Object_Pool[$index]->{vserver} =~ /$vserver_name/) {
                push(@lun_objs, (splice @Lun_Object_Pool, $index, 1));
                $mapped++;
                last;
            } else {
                $index++;
            }
        }
    }

    # Now map them.
    foreach my $lun (@lun_objs) {
        unless ($lun->is_mapped()) {
            $lun->map($igroup_name);
            if ($lun->error()) {
                $self->_set_err("Failed to map the LUN.", $lun->errmsg());
                return $SanUtil::FALSE;
            }
        }
    }

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _finalize_lun_mapping
# Objective:  Maps Luns.
# Parameters: None.
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _finalize_lun_mapping {
    my ($self) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Locals
    my ($lun_count, $lun_map_ref);
    my (@lun_map_info) = ();
    my %args;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start mapping LUNs."));

    # Go through the list of lun map info and call the method to map them
    foreach my $lun_map (@Lun_Map_Info) {
        %args = %$lun_map;
        if (defined($args{lun_count})) {
            $lun_count = $args{lun_count};
        } else {
            $lun_count = 0;
        }
        if ($lun_count and $lun_count <= scalar(@Lun_Object_Pool)) {
            $self->_map_luns(%args);
        }
    }
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done mapping LUNs."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _get_xml_attributes
# Objective:  Gets a list of XML attributes on a given XML node.
# Parameters: A hash containing the key 'xml_node', with a valid XML node.
# Returns:    A hash containing key/value pairs of attributes.
###############################################################################
sub _get_xml_attributes {
    my ($self, %args) = @_;

    # Local Variables
    my $attr_name = "";
    my %params = ();
    my $xml_node = $args{xml_node};
    my @attributes = $xml_node->getAttributes();

    # Get all attributes on this node and build a hash.
    foreach my $attribute (@attributes) {
        $attr_name = $attribute->getName();
        $params{$attr_name} = $xml_node->getAttribute($attr_name);
    }

    return %params;
}

###############################################################################
# Method:     _get_filer_object
# Objective:  Gets the filer object corresponding to a given name
# Parameters: A hash containing the key 'xml_node', with a parent node.
# Returns:    A hash containing key/value pairs of attributes.
###############################################################################
sub _get_filer_object {
    my ($self, %args) = @_;

    # This method will be called after build_nodes. Inner nodes will use it to
    # get the filer object created on top of them.
    my $node = $args{xml_node};
    my ($filer_obj, $nodes_ref, $parent_node_name);
    my @nodes = ();
    my $found_it = $SanUtil::FALSE;

    # Every bucket will have a copy of the filer object. If the user specifies
    # one, use it. If not then default to the "nodes" bucket.
    if (defined($args{bucket})) {
        $nodes_ref = $self->{$args{bucket}};
    } else {
        $nodes_ref = $self->{nodes};
    }
    @nodes = @$nodes_ref;

    # Make sure we get to the "node" element and get the filer name
    while (1) {
        if ($node->getName =~ /node/i) {
            last;
        } else {
            $node = $node->parentNode;
        }
    }
    $parent_node_name = $node->getAttribute('name');

    # Traverse the filer objects (nodes) and look for matching names
    foreach my $element (@nodes) {
        my %elements = %$element;
        while (my ($key, $value) = each(%elements)) {
            if ($value eq $parent_node_name) {
                $filer_obj = $elements{filer_obj};
                $found_it = $SanUtil::TRUE;
                last;
            }
        }
        # If we found it, there's no need to spin cycles here so let's exit.
        last if ($found_it);
    }

    return $filer_obj;
}

###############################################################################
# Method:     _get_nodes
# Objective:  Gets the nodes and creates clones based on the XML document
# Parameters: A hash with the 'id' key defining the type of element.
# Returns:    True if it was successful, False otherwise
###############################################################################
sub _get_nodes{
    my ($self, %args) = @_;

    # Local Variables
    my ($clone, $name, $loop, $xml_root_node, $id, $attr_name, $value);
    my (@nodes, @new_nodes, @attributes) = ();
    my ($index, $offset) = 0;

    # Get the elements and attributes
    $xml_root_node = $self->{xml_root};
    $id = $args{id};
    @nodes = $xml_root_node->getElementsByTagName($id);

    # Go through the list of nodes, if the "loop" attribute is defined, create
    # a clone, rename it by appending _$i to the name and add the clone to the
    # DOM tree

    foreach my $node (@nodes) {
        @attributes = $node->getAttributes();
        # Get all attribute values
        foreach my $attribute (@attributes) {
            $attr_name = $attribute->getName();
            $value = $node->getAttribute($attr_name);
            # Substitute the real attribute value
            $attribute->setValue($self->_parse_attr($value));
        }

        $loop = $node->getAttribute('loop');
        $index = $node->getAttribute('index');
        unless (defined($index)) {
            $index = 0;
        }
        if (defined($loop) and $loop > 1) {
            # Modify the original's node name and clone the rest
            $name = $node->getAttribute('name');
            $offset = $index;
            $node->setAttribute('name', $name . "$offset");
            push (@new_nodes, $node);
            for (my $i = 1; $i < $loop; $i++) {
                $clone = $node->cloneNode($SanUtil::TRUE); #1 = Clone childs too
                $offset = $index + $i;
                $clone->setAttribute('name', $name . "$offset");
                $node->addSibling($clone);
                push (@new_nodes, $clone);
            }
            # Need to remove the original node from the DOM otherwise it'll
            # stay there and we'll end up with a mess
            $xml_root_node->removeChild($node);
        } else {
            push (@new_nodes, $node);
        }
    }

    $self->{xml_root} = $xml_root_node;
    return @new_nodes;
}

###############################################################################
# Method:     _destroy
# Objective:  Destroys SAN objects
# Parameters: None.
# Returns:    True if everything succeeded, False otherwise (and sets error)
###############################################################################
sub _destroy {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Locals
    my ($element_id, $object_id, $reference, $filer_obj, $fcp_obj, $iscsi_obj);
    my (@elements, @volume_objs, @callbacks) = ();
    my $callback;

    # Initialize variables
    $element_id = $args{elements};
    $object_id = $args{object_id};
    $reference = $self->{$element_id};

    # Make sure we can get an array. If there are no elements, it would be a
    # string "" and this would blow!
    if (ref($reference) eq "ARRAY") {
        @elements  = @$reference;
    }

    # Go through each element and delete them. NOTE: For LUNS we need to do a
    # force delete because we don't know if some of them and mapped or not. For
    # vservers, delete the default root volume before deleting the vserver.

    foreach my $element (@elements) {
        # Skip cleanup if this element was recycled/resued. Only destroy what
        # we created.
        unless ($self->{clean_all}) {
            if (exists($element->{reused_obj})) {
                if ($element->{reused_obj} == $SanUtil::TRUE) {
                    next;
                }
            }
        }

        if ($element_id =~ /volumes/) {
            $element->{$object_id}->delete(-force => 1);
            if ($element->{$object_id}->error()) {
                $self->_set_err("Unable to destroy " . $element_id,
                                $element->{$object_id}->errmsg());
                next;
            }
        } elsif ($element_id =~ /luns/) {
            # Check if the lun is mapped
            my @maps = $element->{$object_id}->get_maps();
            if ($element->{$object_id}->error()) {
                $self->_set_err("Unable to get lun maps for " . $element_id,
                                $element->{$object_id}->errmsg());
                next;
            }
            foreach my $map (@maps) {
                $element->{$object_id}->unmap(-igroup => $map);
                if ($element->{$object_id}->error()) {
                    $self->_set_err("Unable to unmap " . $element_id,
                                    $element->{$object_id}->errmsg());
                    next;
                }
            }
            $element->{$object_id}->destroy();
            if ($element->{$object_id}->error()) {
                $self->_set_err("Unable to destroy " . $element_id,
                                $element->{$object_id}->errmsg());
                next;
            }
        } elsif ($element_id =~ /igroups/) {
            $element->{$object_id}->destroy();
            if ($element->{$object_id}->error()) {
                $self->_set_err("Unable to destroy " . $element_id,
                                $element->{$object_id}->errmsg());
                next;
            }
        } elsif ($element_id =~ /lifs|aggregates/) {
            $element->{$object_id}->delete();
            if ($element->{$object_id}->error()) {
                $self->_set_err("Unable to destroy " . $element_id,
                                $element->{$object_id}->errmsg());
                next;
            }
        } elsif ($element_id =~ /vservers/) {
            $element->{$object_id}->delete(-force => 1);
            if ($element->{$object_id}->error()) {
                $self->_set_err("Unable to destroy " . $element_id,
                                $element->{$object_id}->errmsg());
                next;
            }
        } elsif ($element_id =~ /portsets/) {
            $element->{$object_id}->destroy(-force => '');
            if ($element->{$object_id}->error()) {
                $self->_set_err("Unable to destroy " . $element_id,
                                $element->{$object_id}->errmsg());
                next;
            }
        } elsif ($element_id =~ /tpgroups/) {
            $element->{$object_id}->destroy(-force => 1);
            if ($element->{$object_id}->error()) {
                $self->_set_err("Unable to destroy " . $element_id,
                                $element->{$object_id}->errmsg());
                next;
            }
        } elsif ($element_id =~ /clients/) {
            unless ($element->{skip_config}) {
                $self->sanlog(-msg =>
                    $self->_pretty_log("FIT MGR: Start Deconfiguring clients")
                );
                @callbacks = @{$element->{deconfigure}};
                foreach my $callback (@callbacks) {
                    if ($callback == {}) {
                        next;
                    } else {
                        $callback->();
                    }
                }
                $self->sanlog(-msg =>
                    $self->_pretty_log("FIT MGR: Done Deconfiguring clients")
                );
            }
        }
        # Now clear the object's attribute. This means that once the objects
        # have been destroyed, Fit Setup will be empty. This will prevent
        # users from doing something stupid - hopefully.
        $self->{$element_id} = '';
    }

    # This method will attempt to destroy as much as it can, if it hits an issue
    # it will skip that SAN object and move on to the rest so here is where we
    # check if something went really wrong and return False.
    if ($self->error()) {
        return $SanUtil::FALSE;
    }
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done destroying $element_id."));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _configure_client
# Objective:  Calls configure and discover on a SanClient.
# Parameters: A client object, and a filer name.
# Returns:    True if everything succeeded, False otherwise (and sets error)
###############################################################################
sub _configure_client {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Locals
    my ($filer_name, $client_obj, $filer_obj, $cf_mode, $protocol, $vserver_ref);
    my (@filer_targets, @filer_objs, @vservers) = ();
    my $status = $SanUtil::FALSE;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start configuring clients."));

    if (defined($args{client_obj})) {
        $client_obj = $args{client_obj};
    } else {
        $self->_set_err("Client object is a mandatory parameter");
        return {};
    }

    # Get the protocol type, defaults to fcp.
    if (defined($args{protocol})) {
        $protocol = $args{protocol};
    } else {
        $protocol = "fcp";
    }

    # Get rescan timeout value. By default it will be 3 secs
    if (defined($args{rescan_timeout})) {
        $self->{rescan_timeout} = $args{rescan_timeout};
    }

    # Get the vservers that will be used for getting target info
    if (defined($args{vservers})) {
        $vserver_ref = $args{vservers};
        @vservers = @$vserver_ref;
        delete $args{vservers};
    }

    # Get the filer object
    @filer_objs = $self->get_filers(name => $args{filer_name});
    if ($self->error()) {
        return {};
    }

    # There's only one filer obj in filer_objs.
    $filer_obj = $filer_objs[0];

    # Get filer targets (It's always the first one in the array because we
    # specified its name)
    if ($protocol =~ /fcp/i) {
        foreach my $vserver (@vservers) {
            $self->sanlog(-msg =>
                $self->_pretty_log("FIT MGR: Getting fcp targets for $vserver."));
            push(@filer_targets, $filer_obj->get_fcp_targets(vserver => $vserver));
            if ($filer_obj->error()) {
                $self->_set_err("Failed to get fcp targets from filer. ",
                                $filer_obj->errmsg());
                return {};
            }
            $self->sanlog(-msg =>
                $self->_pretty_log("FIT MGR: Done getting fcp targets for $vserver."));
        }
    } elsif ($protocol =~ /iscsi/i) {
        foreach my $vserver (@vservers) {
            $self->sanlog(-msg =>
                $self->_pretty_log("FIT MGR: Getting iscsi targets for $vserver."));
            push(@filer_targets, $filer_obj->get_iscsi_targets(vserver => $vserver));
            if ($filer_obj->error()) {
                $self->_set_err("Failed to get iscsi targets from filer. ",
                                $filer_obj->errmsg());
                return {};
            }
            $self->sanlog(-msg =>
                $self->_pretty_log("FIT MGR: Done getting iscsi targets for $vserver."));
        }
    }

    $cf_mode = $filer_obj->get_cfmode();
    if ($filer_obj->error()) {
        $self->_set_err("Failed to get cfmode from filer. ",
                        $filer_obj->errmsg());
        return {};
    }

    # Configure the client based on the targets
    $client_obj->configure(-cfo_mode => $cf_mode,
                           -targets  => \@filer_targets);
    if ($client_obj->error()) {
        $self->_set_err("Failed to configure the client based on targets "
                          . "and cf_mode",
                        $client_obj->errmsg());
        return {};
    }
    $self->sanlog(-msg =>
        $self->_pretty_log("FIT MGR: Done config  using: cf_mode -> $cf_mode" .
                           "Num Targets -> " . scalar(@filer_targets)));

    $self->_discover_devices(client_obj     => $client_obj,
                             rescan_timeout => $self->{rescan_timeout});
    if ($self->error()) {
        $self->_set_err("Failed to discover San Client.");
        return {};
    }

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done configuring clients"));
    # Return a deconfiguration callback for cleanup
    return sub {
        $client_obj->deconfigure(-targets => \@filer_targets);
        if ($client_obj->error()) {
            $self->_set_err("Failed to deconfigure client.",
                            $client_obj->errmsg());
            return {};
        }
    };
}

###############################################################################
# Method:     _discover_devices
# Objective:  Rescan on the client to find any new devices.
# Parameters: A client object.
# Returns:    True if everything succeeded, False otherwise (and sets error)
###############################################################################
sub _discover_devices {
    my ($self, %args) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Locals
    my ($client_obj, $device_data, $devmgr_obj, $msg);
    my $skip_discover = $SanUtil::FALSE;
    my (@devices, @dev_names) =();

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start discovering devices."));

    if (defined($args{client_obj})) {
        $client_obj = $args{client_obj};
    } else {
        $self->_set_err("Client object is a mandatory parameter");
        return $SanUtil::FALSE;
    }

    # If users -for some strange reason- decide to skip discover
    if (defined($args{skip_discover})) {
        $skip_discover = $args{skip_discover};
    }

    if (defined($args{rescan_timeout})) {
        $self->{rescan_timeout} = $args{rescan_timeout};
    }

    unless ($skip_discover == $SanUtil::TRUE) {
        $client_obj->discover(-rescan_timeout => $self->{rescan_timeout});
    }

    # Create a device manager object
    $devmgr_obj = $client_obj->create_devmgr();
    if ($client_obj->error) {
        $self->_set_err("Failed to create device manager obj" . $client_obj->errmsg());
        return $SanUtil::FALSE;
    }
    # Rescan on the client to find any new devices.
    # all get_devices if windows
    @devices = $client_obj->get_devices(-refresh => 0);
    if ($client_obj->error()) {
        $self->_set_err("Failed to get_devices ", $client_obj->errmsg());
        return $SanUtil::FALSE;
    }
    # :::NOTE::: fs devices will change the list of returned
    # devices to device objects not hashes
    if ($client_obj->{device_type} =~ /fs/i) {
        $self->sanlog(-msg => "Creating filesystem on devices.");
        # Get the device list from DevMgr
        @devices = $devmgr_obj->get_devices();

        if (scalar(@devices) > 0) {
            foreach my $devObj (@devices) {
                push (@dev_names, $devObj->{name});
            }
            $msg = join("\n", @dev_names);
            $self->sanlog(-msg => "List of discovered raw devices includes:\n" . $msg);

            # format the devices within the @dev_names array
            $client_obj->format(-devices => \@dev_names);
            if ($client_obj->error()) {
                $self->_set_err("Client device format() failed");
                return $SanUtil::FALSE;
            }

            foreach my $device (@devices) {
                # For windows, format already does the format() call,
                # so there is no need to do again.
                if ($client_obj->hosttype() !~ /windows/i) {
                    # this empty create_fs() call will use default
                    # filesystem types for non windows hosts
                    $device->create_fs();
                    if ($device->error()) {
                        $self->_set_err("device create_fs() failed" . $device->errmsg());
                        return $SanUtil::FALSE;
                    }
                }

                $device->mount_fs();
                if ($device->error()) {
                    $self->_set_err("device mount_fs() failed" . $device->errmsg());
                    return $SanUtil::FALSE;
                }

                $device->load();
                if ($device->error()) {
                    $self->_set_err("device load() failed" . $device->errmsg());
                    return $SanUtil::FALSE;
                }
            }

        } else {
            $self->_set_err("No valid devices are available to use");
            return $SanUtil::FALSE;
        }
    } elsif($client_obj->{device_type} =~ /raw/i){
        if( $client_obj->hosttype() =~ /windows/i) {
            # Call raw_device_setup where on windows 2008. This will online the disks
            $client_obj->raw_device_setup();
            if ($client_obj->error()) {
                return $SanUtil::FALSE;
            }
        }
    }

    $self->sanlog(-msg => scalar(@devices) . " devices found");

    # Put the devices in the Device Pool
    $device_data = {
        client_type => $client_obj->hosttype(),
        client_name => $client_obj->hostname(),
        devices => \@devices
    };
    push (@Devices, $device_data);

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done discovering devices."));
    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _add_initiators_to_igroup
# Objective:  Adds initiators (client)to igroups
# Parameters: Filer name, vserver name, igroup name and a list of initiators
# Returns:    True if everything succeeded, False otherwise (and sets error)
###############################################################################
sub _add_initiators_to_igroup {
    my ($self, $init) = @_;

    # Make sure we start with a cleared error flag
    $self->_clear_err();

    # Locals
    my ($filer_ref, $vserver_ref, $igroup_ref, $initiator_ref, $igroup_obj);
    my (@initiators, @filers, @vservers, @igroups, @filer_obj) =();
    my %args = %$init;

    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Start adding initiators."));

    # Get the data. No need to validate here since we already did that before
    # calling this method
    $initiator_ref = $args{initiators};
    @initiators = @$initiator_ref;
    $filer_ref = $args{filers};
    @filers = @$filer_ref;
    $vserver_ref = $args{vservers};
    @vservers = @$vserver_ref;
    $igroup_ref = $args{igroups};
    @igroups = @$igroup_ref;

    # This is needed when users passed the magic keywords and decided to loop
    # We basically add initiators to any filer, vserver and igroup according
    # to the XML file.
    foreach my $filer (@filers) {
        foreach my $vserver (@vservers) {
            foreach my $igroup (@igroups) {
                # Get filer object and then igroup objects
                @filer_obj = $self->get_filers(name => $filer);
                $igroup_obj = $filer_obj[0]->get_igroup(name  => $igroup,
                                                      vserver => $vserver);
                if ($filer_obj[0]->error()) {
                    return $SanUtil::FALSE;
                }
                foreach my $initiator (@initiators) {
                    $igroup_obj->add(initiator => $initiator);
                    if ($igroup_obj->error()) {

                        # If the initiator is already in the igroup, goto next
                        # intiator.
                        if ($igroup_obj->errmsg() =~ /(Duplicate entry|Initiator group already contains node)/) {
                            next;
                        }
                        $self->_set_err("Failed to add $initiator to $igroup",
                                        $igroup_obj->errmsg());
                        next;
                    } ## end if ($igroup_obj->error...

                }
            }
        }
    }

    # Catch the error just in case any igroup add failed.
    if ($self->error()) {
        $self->_set_err("Failed to add one or more initiators",
                        $self->errmsg());
        return $SanUtil::FALSE;
    }
    $self->sanlog(-msg => $self->_pretty_log("FIT MGR: Done adding initiators"));

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _extract_target_info
# Objective:  Extracts all target info from the XML client tag
# Parameters: A string. The "target" attribute: filer:vserver:igroup
# Returns:    An array containing the info needed to add targets.
###############################################################################
sub _extract_target_info {
    my ($self, $targets) = @_;

    # Clear the error flag
    $self->_clear_err();

    # Locals
    my ($filer, $vserver, $igroup);
    my (@filers, @vservers, @igroups, @data) = ();
    my $target_data;

    # Extract data from the input
    @data = split (/:/, $targets);

    # We need three things in there, if they are not... well, err.
    unless (scalar(@data) == 3) {
        $self->_set_err("Target info on the client XML tag is wrong!",
                        $self->errmsg());
        return ();
    }

    # Get the three strings: filers, vservers, igroups
    $filer   = $data[0];
    $vserver = $data[1];
    $igroup  = $data[2];

    # Users have three options:
    # Simple: filer_name:vserver_name:igroup_name
    # All: filer_name[all]:vserver_name[all]:igroup_name[all]
    # Range: filer_name[1-n]:vserver_name[1-n]:igroup_name[1-n]

    # Start with filer info
    if ($filer =~ /(.+)\[all\]/i) {
        push (@filers, $1);
    } elsif ($filer =~ /(.+)\[(.+)-(.+)\]/i) {
        for (my $i = $2 ; $i <= $3; $i++) {
            push(@filers, "$1$i");
        }
    } else {
        push(@filers, $filer);
    }

    # Same thing for vservers
    if ($vserver =~ /(.+)\[all\]/i) {
        push (@vservers, $1);
    } elsif ($vserver =~ /(.+)\[(.+)-(.+)\]/i) {
        for (my $i = $2 ; $i <= $3; $i++) {
            push(@vservers, "$1$i");
        }
    } else {
        push(@vservers, $vserver);
    }

    # Same thing for igroups
    if ($igroup =~ /(.+)\[all\]/i) {
        push (@igroups, $1);
    } elsif ($igroup =~ /(.+)\[(.+)-(.+)\]/i) {
        for (my $i = $2 ; $i <= $3; $i++) {
            push(@igroups, "$1$i");
        }
    } else {
        push(@igroups, $igroup);
    }

    # Now pack everything
    $target_data = { filers   => \@filers,
                     vservers => \@vservers,
                     igroups  => \@igroups };

    return $target_data;
}

###############################################################################
# Method:     _clean_if_error  (Helper function)
# Objective:  Calls cleanup methods. Created to avoid duplicating the same code.
# Parameters: None.
# Returns:    True if everything succeeded, False otherwise (and sets error)
###############################################################################
sub _clean_if_error {
    my ($self) = @_;

    $self->_clear_err();

    # we will cleanup anyway.
    #if ($self->{cleanup}) {
    #    $self->cleanup();
    #    if ($self->error()) {
    #        $self->_set_err("Failed to cleanup one of more elements. " .
    #                        $self->errmsg());
    #    }
    #    return $SanUtil::FALSE;
    #}

    return $SanUtil::TRUE;
}

###############################################################################
# Method:     _get_val
# Objective:  Gets attribute values from XML. If the value is a variable it
#             tries to get it from the PARAM. It's up to the user to make
#             sure all PARAM names are correct.
# Parameters: A string containing the XML attribute value.
# Returns:    A string representing the value to use. If it fails it will
#             return an empty string and log a message. It DOES NOT set an
#             error flag; it's up to the caller to set it.
###############################################################################
sub _get_val {
    my ($self, $value) = @_;

    my ($var_name, $default_name, $range);

    $self->_clear_err();

    # First check if the value is a normal string or a variable. If it's a
    # value, return it. If it's a variable, make sure we've got a valid TCD obj
    # which should be passed by users to the constructor, get the PARAM and
    # return the string.

    # Variables start with "$". Users can define a "default value", this allows
    # them to use both variables/values: $AGGR,test_aggr.
    if ($value =~ /^.*\$([0-9a-zA-Z_\-:;.]*),?([0-9a-zA-Z_\-:;.]*)(\[.*\])?/) {
        $var_name = $1;
        $default_name = $2 if $2;
        $range = $3 if $3;
    } else {
        return $value;
    }

    # Now make sure we have a Test object
    unless (defined($self->{tcd_test}) and length($self->{tcd_test})) {
        $self->sanlog(-msg => $self->_pretty_log("You are using variable names " .
                              "in your XML but did not pass a Test object to " .
                              "FitMgr::new."));
        $self->_set_err("It looks like you forgot to pass a TCD object!");
        return undef;
    }

    # Now try getting a param from the Test object.
    $value = $self->{tcd_test}->get_param($var_name);

    # If we didn't get a value, check if there's a default value, if ther isn't
    # log a message and return an empty string.
    unless (defined($value) and length($value)) {
        if ($default_name) {
            $default_name = $default_name . $range if ($range);
            return $default_name;
        }
        $self->sanlog(-msg => $self->_pretty_log("get_param did not return any " .
                              "value for $var_name. Check your XML variables " .
                              "and make sure they are defined in your environment"));
        $self->_set_err("Failed to get a value from $var_name!");
        return undef;
    }

    if ($range) {
        $value = $value . $range;
    }

    return $value;
}

###############################################################################
# Method:     _parse_attr
# Objective:  Parses and makes variables substitutionsin on an XML attribute
# Parameters: A string to parse.
# Returns:   A parsed string
###############################################################################
sub _parse_attr {
    my ($self, $value) = @_;

    my @entries = split(/:/, $value);
    my @parsed_entries = ();
    my $parsed_value;

    foreach my $entry (@entries) {
        push(@parsed_entries, $self->_get_val($entry));
    }

    $parsed_value = join(":", @parsed_entries);

    return $parsed_value;
}

###############################################################################
# Method:     _print_log (Helper function)
# Objective:  Adds a simple box around a message for logging purposes.
# Parameters: None.
# Returns:   The pretty message
###############################################################################
sub _pretty_log {
    my ($self, $msg) = @_;

    my $header  = "\n" x 2 . "#-" x 39 . "#\n";
    my $body    = "#  " . $msg . "\n";
    my $footer  = "\#-" x 39 . "#\n";

    # Prints something like (easier to identify FIT entries in the logs...):
    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
    #  Some message
    #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#
    return ($header . $body . $footer);
}

1;
