# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to object caching methods
## @author hrivera@netapp.com
## @status public
## @requirements Inheriting object must inherit Error.pm

package SanCache;

use lib();
use San::SanUtil;

# exported global vars
$SanCache::VALID    = 1;
$SanCache::INVALID  = 0;
$SanCache::ENABLED  = -1;
$SanCache::DISABLED = -2;

################################################################################
# Subroutine Name: _initialize_cache()
# Objective:       Initialize the cache attribute
################################################################################
sub _initialize_cache {
    my ($self) = @_;

    $self->{sancache} = {};

    # sancache will be enabled during initialization only if it was not
    # explicitly disabled (IE: the method "disable_sancache()" was not invoked)
    if ($self->is_sancache_enabled() != $SanCache::DISABLED) {
        $self->enable_sancache();
    }

    return $SanUtil::TRUE;
}

################################################################################
# Subroutine Name: is_sancache_enabled()
# Objective:       Determine if sancaching is enabled
################################################################################
sub is_sancache_enabled {
    my ($self) = @_;

    return $self->{is_sancache_enabled};
}

################################################################################
# Subroutine Name: enable_sancache()
# Objective:       Enable the sancache methods
################################################################################
sub enable_sancache {
    my ($self, %args) = @_;

    my $invalidate_cache = $SanUtil::FALSE;

    # parse arguments
    foreach (keys %args) {
        if (/-?invalidate_cache/) {
            if (not length($invalidate_cache = $args{$_})) {
                $self->_set_err("Invalid value for -invalidate_cache");
                return $SanUtil::FALSE;
            }
            next;
        }
    }

    # keep track of the history of whether it was enabled
    # this will allow us to revert to the previous setting
    $self->{was_sancache_enabled} = $self->is_sancache_enabled();

    $self->{is_sancache_enabled} = $SanCache::ENABLED;

    if ($invalidate_cache) {
        return $self->invalidate_all_cache_data();
    }

    return $SanUtil::TRUE;
}

################################################################################
# Subroutine Name: disable_sancache()
# Objective:       Disable the sancache methods
################################################################################
sub disable_sancache {
    my ($self, %args) = @_;

    my $invalidate_cache = $SanUtil::FALSE;

    # parse arguments
    foreach (keys %args) {
        if (/-?invalidate_cache/) {
            if (not length($invalidate_cache = $args{$_})) {
                $self->_set_err("Invalid value for -invalidate_cache");
                return $SanUtil::FALSE;
            }
            next;
        }
    }

    # keep track of the history of whether it was enabled
    # this will allow us to revert to the previous setting
    $self->{was_sancache_enabled} = $self->is_sancache_enabled();

    $self->{is_sancache_enabled} = $SanCache::DISABLED;

    if ($invalidate_cache) {
        return $self->invalidate_all_cache_data();
    }

    return $SanUtil::TRUE;
}

################################################################################
# Subroutine Name: revert_sancache()
# Objective:       Revert to the previous setting of sancache (enabled|disabled)
################################################################################
sub revert_sancache {
    my ($self) = @_;

    # if the previous value was enabled, then enable it
    if ($self->{was_sancache_enabled} == $SanCache::ENABLED) {
        $self->enable_sancache();
        if ($self->error()) {
            return $SanUtil::FALSE;
        }
    } else {
        # otherwise the previous value was disabled, so disable it
        $self->disable_sancache();
        if ($self->error()) {
            return $SanUtil::FALSE;
        }
    }

    return $SanUtil::TRUE;
}

################################################################################
# Subroutine Name: get_cache_data()
# Objective:       Return the cache data for the specified item
################################################################################
sub get_cache_data {
    my ($self, $item) = @_;

    # if the cache is not defined, then this is an error
    if (not defined($self->{sancache})) {
        $self->_initialize_cache();
    }

    # if the item that was requested is not defined then set an error
    if (not defined($item)) {
        $self->_set_err("A cache item was not specified.");
        return (wantarray()) ? () : [];
    }

    # determine if the cache item exists
    if (not defined($self->{sancache}->{$item})) {
        $self->_set_err("The cache item '$item' does not exist.");
        return (wantarray()) ? () : [];
    }

    my @cache_data = @{$self->{sancache}->{$item}->{data}};

    return (wantarray()) ? @cache_data : \@cache_data;
}

################################################################################
# Subroutine Name: store_cache_data()
# Objective:       Store the data into the item's cache data
################################################################################
sub store_cache_data {
    my ($self, $item, @data) = @_;

    # if the cache is not defined, then this is an error
    if (not defined($self->{sancache})) {
        $self->_initialize_cache();
    }

    # if the item that was requested is not defined then set an error
    if (not defined($item)) {
        $self->_set_err("A cache item was not specified.");
        return $SanUtil::FALSE;
    }

    $self->{sancache}->{$item}->{data} = \@data;
    $self->{sancache}->{$item}->{is_valid} = $SanCache::VALID;

    return $SanUtil::TRUE;
}

################################################################################
# Subroutine Name: is_cache_data_valid()
# Objective:       Return 0 if the cache data is invalid, 1 if valid.
################################################################################
sub is_cache_data_valid {
    my ($self, $item) = @_;

    # if the cache is not defined, then this is an error
    if (not defined($self->{sancache})) {
        $self->_initialize_cache();
    }

    # if caching is disabled, return immediately with an invalid flag
    if ($self->is_sancache_enabled() == $SanCache::DISABLED) {
        return $SanCache::INVALID;
    }

    # if the item that was requested is not defined then set an error
    if (not defined($item)) {
        $self->_set_err("A cache item was not specified.");
        return $SanCache::INVALID;
    }

    # determine if the cache item exists
    if (not defined($self->{sancache}->{$item})) {
        return $SanCache::INVALID; # cache item does not exist, thus it is invalid
    } else {
        # return the value of is_valid for this cache item.
        return $self->{sancache}->{$item}->{is_valid};
    }
}

################################################################################
# Subroutine Name: invalidate_items_cache_data()
# Objective:       Mark the items cache as invalid
################################################################################
sub invalidate_items_cache_data {
    my ($self, $item) = @_;

    if (not defined($self->{sancache})) {
        $self->_initialize_cache();
    }

    # mark this cache item to be invalid
    $self->{sancache}->{$item}->{is_valid} = $SanCache::INVALID;
    $self->{sancache}->{$item}->{data} = [];

    return $SanUtil::TRUE;
}

################################################################################
# Subroutine Name: invalidate_all_cache_data()
# Objective:       Mark all items in the cache as invalid
################################################################################
sub invalidate_all_cache_data {
    my ($self, %args) = @_;

    my $match;
    my $doesnt_match;

    # parse arguments
    foreach (keys %args) {
        if (/-?match/) {
            if (not length($match = $args{$_})) {
                $self->_set_err("Invalid value for -match");
                return $SanUtil::FALSE;
            }
            next;
        } elsif (/-?doesnt_match/) {
            if (not length($doesnt_match = $args{$_})) {
                $self->_set_err("Invalid value for -doesnt_match");
                return $SanUtil::FALSE;
            }
            next;
        }
    }

    # initialize the cache if it has not been.
    if (not defined($self->{sancache})) {
        $self->_initialize_cache();
    }

    foreach my $item (keys %{$self->{sancache}}) {
        # filter out items that dont match the '$match' regex
        if (defined($match)) {
            next if ($item !~ $match);
        }

        # filter out items that match the '$doesnt_match' regex
        if (defined($doesnt_match)) {
            next if ($item =~ $doesnt_match);
        }

        # invalidate this item
        $self->invalidate_items_cache_data($item);
    }

    return $SanUtil::TRUE;
}

1;
