#
# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author knopp@netapp.com
## @status private
## @pod here

package SanClient;

# Compiler directives.
use strict;

# Module import.
use File::Basename;
use File::Compare;
use Time::HiRes qw(gettimeofday);

use Tharn;
use San::SanLog;
use San::SanCache;
use San::SanClient::Lvm;
use San::SanClient::DevMgr;
use San::SanClient::MPDevMgr;
use San::SanClient::Sanlun;
use San::SanClient::ClientUtil;
use San::SanClient::Initiator;
use San::SanClient::MP;
use San::SanClient::HU;
use WinClient;
use Utils qw(make_host make_conn);
use Error;
use Data::Dumper;
use Exporter;
use vars qw($_CLIENT_CNT
            @EXPORT
            @ISA);

@ISA = qw(Error SanLog SanCache);

no warnings qw(redefine);

###############################################################################
#                               Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new Host object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($proto, %args) = @_;

    my ($subr, $execute, $host, $connection, $conn_type, $log, $passwd);
    my ($protocol, $rc, $show_cmd, $timeout, $user,   $multipath, $debug_level);
    my ($winclient_obj, $device_type, $sanblaze_ports, $initiator_count);
    my ($alua_support, $mp_type, $load_inits, $test_bin_dir, $pnate_environment);
    my ($connection_retries, @protocols, %hArgs, $is_virtual, $hypervisor_type, $any_iscsi);
    my ($override_vm_manufacturer, $vma_host, $iscsi_type);
    my $self = {};
    my $sanlun_obj = {};
    my $obj_template = undef;

    my $class = ref($proto) || $proto;
    bless ($self, $class);

    # init variables
    $protocol = "nfs";
    $multipath = 1;
    $debug_level = 0;
    $device_type = "raw";
    $timeout = 1800;
    $mp_type = "none";
    $load_inits = 1;
    $test_bin_dir = "NOT_SET";
    $pnate_environment = 0;
    $connection_retries = 5;
    $is_virtual = 0;
    $hypervisor_type = "none";
    $override_vm_manufacturer = 1;
    $vma_host = "";
    $iscsi_type = undef;

    # Process arguments.  If an error occurs during the arg processing,
    # The error flag and errmsg values are set and the object is returned.
    # It is up to the caller to check for errors after the new call.
    foreach (keys %args) {
        if (/^-?host$/) {
            length($host = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"host\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/^-?connection$/) {
            length($connection = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"connection\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/^-?conn_type$/) {
            length($conn_type = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"conn_type\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/^-?user$/) {
            length($user = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"user\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/^-?passwd$/) {
            length($passwd = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"passwd\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?timeout$/) {
            length($timeout = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"timeout\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?conn_retries$/) {
            length($connection_retries = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"conn_retries\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?debug_level$/) {
            length($debug_level = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"debug_level\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/^-?vma_host$/) {
            $vma_host = $args{$_};
        } elsif (/^-?iscsi_type$/) {
            $iscsi_type = $args{$_};
        } elsif (/-?device_type$/) {
            if (not length($device_type = $args{$_}) or $device_type !~ /raw|lv|fs/i) {
                $self->_set_err("Invalid value for \"device_type\":".$args{$_});
                return $self;

            }
            next;
        } elsif (/-?mp_type$/) {
            if (not length($mp_type = $args{$_})) {
                $self->_set_err("Invalid value for \"mp_type\":".$args{$_});
            }
            next;
        } elsif (/-?load_inits$/) {
            if (not length($load_inits = $args{$_}) or $load_inits !~ /1|0/) {
                $self->_set_err("Invalid value for \"load_inits\":".$args{$_});
            }
            next;
        } elsif (/-?pnate_environment$/) {
            if (not length($pnate_environment = $args{$_}) or $pnate_environment !~ /1|0/) {
                $self->_set_err("Invalid value for \"pnate_environment\":".$args{$_});
            }
            next;
        }
        elsif (/-?test_bin_dir$/) {
            if (not length($test_bin_dir = $args{$_})) {
                $self->_set_err("Invalid value for \"test_bin_dir\":" . $args{$_});
            }
            next;
        } elsif (/-?protocol$/) {
            # Default to nfs if exist and not user defined
            # (Note: Undefined get_parameter returns 0)
            $protocol = $args{$_};
            if (exists $args{$_} and $args{$_} =~ /0/) {
                $protocol = "nfs";
            } else {
                unless (exists $args{$_} and $args{$_} =~ /nfs|fcp|iscsi/i) {
                    $self->_set_err("Invalid value for \"protocol\":" . $args{$_});
                    return $self;

                }
            }
        } elsif (/-?show_cmd$/) {
            length($show_cmd = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"show_cmd\"",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?loghdl$/) {
            length($log = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for",
                                    "\"log\" $args{$_}");
                    return $self;
                };
        } elsif (/-?multipath$/) {
            if (not length($multipath = $args{$_}) or $multipath !~ /no|yes|1|0|off|on/i) {
                $self->_set_err("Invalid value for",
                                "\"multipath\" $args{$_}");
                return $self;
            } else {
                if ($multipath =~ /yes|1|on/) {
                    $multipath=1;
                } else {
                    $multipath=0;
                }
            }
            next;
        } elsif (/-?_obsolete|-?parent/) {
            next;

        } elsif (/-?sanblaze_ports/) {
            length($sanblaze_ports = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"sanblaze_ports\"",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?initiator_count/) {
            length($initiator_count = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"initiator_count\"",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?hypervisor_type/) {
            length($hypervisor_type = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"hypervisor_type\"",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?override_vm_manufacturer/) {
            length($override_vm_manufacturer = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"override_vm_manufacturer\"",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?alua_support$/) {
            if (not length($alua_support = $args{$_}) or $alua_support !~ /no|yes|1|0|off|on/i) {
                $self->_set_err("Invalid value for",
                                "\"alua_support\" $args{$_}");
                return $self;
            } else {
                if ($alua_support =~ /yes|1|on/) {
                    $alua_support=1;
                } else {
                    $alua_support=0;
                }
            }
            next;
        } elsif (/-?any_iscsi/) {
            if (length($args{$_})){
                $any_iscsi = $args{$_};
            }else{
                $any_iscsi = undef;
            }
        } elsif (/-?obj_template/) {
            if (defined($args{$_})){
                $obj_template = $args{$_};
            }else{
                $obj_template = undef;
            }
        }
    }

    # create a lightweight sanclient object if an existing template was passed in
    if (defined($obj_template)) {
        eval {
            $self = $self->_create_lite_sanclient_object(-obj_template => $obj_template);
        };
        if ($@) {
            $self->_set_err("SanClient lite constructor failed for host '" . $obj_template->{host} . "'", "$@");
            return $self;
        }
        return $self;
    }

    # Verify mandatory parameters.
    unless (defined $host and length $host) {
        $self->_set_err("Mandatory parameter \"host\"",
                        "not specified");
        return $self;
    }

    # create a host object for $host
    my $hostrec = Utils::make_host($host);

    $self->sanlog(-msg => "Creating a SanClient Object for client: '$host' (protocol:$protocol)");

    # reset the vma_host parameter if this is not a VMware host
    if ($hostrec->hosttype() !~ /linux\-vmware/) {
        $vma_host = "";
    }

    # Set the timeout if one is given.
    $self->timeout($timeout) if (defined $timeout);

    # Set the debug_level if one is given.
    $self->set_debug_level($debug_level);

    # initialize the number of connect retries on the object
    $self->{connection_retries} = $connection_retries;

    # Determine if a connection type has been provided
    if (!defined($connection)) {
        if ($vma_host ne "") {
            # Connect to the host and store the connection obj
            $connection = $self->_connect($vma_host, $user, $passwd, $timeout, $conn_type);
        } else {
            $connection = $self->_connect($host, $user, $passwd, $timeout, $conn_type);
        }
        if ($self->error) {
            # Error message is already set, return object so caller
            # can get the error.
            return $self;
        }
    }

    $self->{conn} = $connection;

    # Create the WinClient Object if we are using a windows host.
    $winclient_obj = undef;
    if ($self->hosttype() =~ /windows/) {
        # retry the creation of the WinClient object 5 times before issuing an error
        my $retry_count = 5;
        while (1) {
            $winclient_obj = WinClient->new(-connection         => $connection,
                                            -timeout            => $timeout,
                                            -protocol           => $protocol,
                                            -multipath          => $multipath,
                                            -hypervisor_type    => $hypervisor_type,
                                            -pnate_environment  => $pnate_environment);
            if ($winclient_obj->error() and $retry_count == 0) {
                # the retry count was exhausted, set and error and return.
                $self->_set_err("Failed to properly create",
                                "WinClient object - please check connection type and connectivity (RSH) to host",
                                $winclient_obj->errmsg());
                return $self;
            } elsif ($winclient_obj->error()) {
                # sleep ten seconds and try creating the winclient object again
                $retry_count--;
                &SanUtil::nate_sleep(10);
                next;
            } else {
                # otherwise the winclient object was created successfully, break out.
                last;
            }
        }
    }

    # Get OS Version
    if ($self->hosttype() =~ /windows/) {
        $self->{conn}{hostp}{version} = $winclient_obj->{hostver};
    } elsif ($self->hosttype() =~ /aix/) {
        my $ver = $self->execute("uname -v", -hide => 1, -_hostcmd=>1)->[0];
        my $rel = $self->execute("uname -r", -hide => 1, -_hostcmd=>1)->[0];
        $self->{conn}{hostp}{version} = "$ver.$rel";
    } else {
        $self->{conn}{hostp}{version} = $self->execute("uname -r", -hide => 1, -_hostcmd => 1)->[0];
    }

    return $self if ($self->error);

    # if specified to not override the vm_manufacturer parameters in
    # the tharnhost files, then use them (see burts 423619/366564)
    # by default we override the vm manufacturer with
    # the values passed into the constructor
    if ($override_vm_manufacturer == 0) {
        $hypervisor_type = $self->{conn}{hostp}{MANUFACTURER}
            if defined $self->{conn}{hostp}{MANUFACTURER};
        $hypervisor_type = $self->{conn}{hostp}{VM_MANUFACTURER}
            if defined $self->{conn}{hostp}{VM_MANUFACTURER};
    }

    # determine the right platform and require/import/call the appropriate
    # OS object's new call.
    my $hosttype = $self->hosttype();

    if ($hosttype =~ /unix-solaris/i) {
        if ($hypervisor_type =~ /VMware|HyperV|Microsoft|Citrix|KVM|OVM/i) {
            $is_virtual = 1;
            require San::SanClient::Host::VSolarisHost;
            $self = San::SanClient::Host::VSolarisHost->_new(-conn => $self->get_connection());
        } else {
            require San::SanClient::Host::SolarisHost;
            $self = San::SanClient::Host::SolarisHost->_new(-conn => $self->get_connection());
        }
    } elsif ($hosttype =~ /unix-hpux/i) {
        require San::SanClient::Host::HPUXHost;
        $self = San::SanClient::Host::HPUXHost->_new(-conn => $self->get_connection(), %args);
    } elsif ($hosttype =~ /unix-aix/i) {
        require San::SanClient::Host::AixHost;
        $self = San::SanClient::Host::AixHost->_new(-conn => $self->get_connection());
    } elsif ($hosttype =~ /unix-linux-\w+-kvm/i){
        require San::SanClient::Host::KVMHost;
        $self = San::SanClient::Host::KVMHost->_new(-conn => $self->get_connection(), %args);
    } elsif ($hosttype =~ /unix-linux-\w+-ovm/i){
        require San::SanClient::Host::OVMHost;
        $self = San::SanClient::Host::OVMHost->_new(-conn => $self->get_connection(), %args);
    } elsif ($hosttype =~ /unix-linux.*-citrix/i){
        require San::SanClient::Host::CXSHost;
        $self = San::SanClient::Host::CXSHost->_new(-conn => $self->get_connection(), %args);
    } elsif ($hosttype =~ /unix-linux-sanblaze/i) {
        require San::SanClient::Host::SanBlazeHost;
        $self = San::SanClient::Host::SanBlazeHost->_new(-conn => $self->get_connection(), %args);
    } elsif ($hosttype =~ /unix-linux/i) {
        if ($hypervisor_type =~ /VMware|HyperV|Microsoft|Citrix|KVM|OVM/i) {
            $is_virtual = 1;
            require San::SanClient::Host::VLinuxHost;
            $self = San::SanClient::Host::VLinuxHost->_new(-conn => $self->get_connection());
        } else {
            require San::SanClient::Host::LinuxHost;
            $self = San::SanClient::Host::LinuxHost->_new(-conn => $self->get_connection(), %args);
        }
    } elsif ($hosttype =~ /hyper-v/i) {
        $args{winclient_obj} = $winclient_obj;
        require San::SanClient::Host::HyperVHost;
        $self = San::SanClient::Host::HyperVHost->_new(-conn => $self->get_connection(), %args);
    } elsif ($hosttype =~ /windows/i) {
        if ($hypervisor_type =~ /VMware|HyperV|Microsoft|Citrix|KVM|OVM/i) {
            $is_virtual = 1;
            require San::SanClient::Host::VWinHost;
            $self = San::SanClient::Host::VWinHost->_new(-conn => $self->get_connection());
            # force the WinClient protocol to be "none" in the virtual context
            $winclient_obj->{protocol} = "none";
        } else {
            require San::SanClient::Host::WinHost;
            $self = San::SanClient::Host::WinHost->_new(-conn => $self->get_connection());
        }
    } elsif ($hosttype =~ /linux-vmware/i) {
        if ($vma_host ne "") {
            require San::SanClient::Host::ESXiHost;
            $self = San::SanClient::Host::ESXiHost->_new(-conn => $self->get_connection(),
                                                         %args,
                                                         -phy_host => $host);
        } else{
            require San::SanClient::Host::ESXHost;
            $self = San::SanClient::Host::ESXHost->_new(-conn => $self->get_connection(), %args);
        }
    } elsif ($hosttype =~ /sanblaze/i) {
        require San::SanClient::Host::SanBlazeHost;
        $self = San::SanClient::Host::SanBlazeHost->_new(-conn => $self->get_connection(), %args);
        return $self if ($self->error);
    }   else {
        $self->_set_err("Host platform '$hosttype' not",
                        "currently supported by SanClient object");
        return $self;
    }

    # check if there were any errors from per platform constructor
    if ($self->error()) {
        $self->_set_err("Error from per-platform ($hosttype) SanClient constructor", $self->errmsg());
        return $self;
    }

    @ISA = qw(Error SanLog San::SanClient);

    if (not defined($self->{arch})) {
        $self->sanlog(-msg => "Host Architecture is not set",
                      -msg_level => 3);
    }

    # save the connection again
    $self->{conn} = $connection;

    # Store connection parameters in object.
    $self->{host} =  $self->{conn}->{hostp}->{hostname};
    $self->{user} =  $self->{conn}->{username};
    $self->{pass} =  $self->{conn}->{password};

    # initialize the sub object hash
    $self->{obj} = {};

    # save value of conn_type in host object
    $self->{conn_type} = $conn_type;

    # Show command mode.  By default when printing, the command
    # issued will be displayed.  This can be turned off by the user.
    $self->{show_cmd} = defined $show_cmd ? $show_cmd : 1;

    $self->{show_data} = 1;
    $self->{protocol}  = $protocol;
    $self->{multipath} = $multipath;

    # set client device type; default is "raw"
    $self->{device_type} = $device_type;

    # Allow CLI commands to be formed, but not executed.
    $self->{execute} = defined $execute ? $execute : 1;

    # Allow CLI commands to be logged;
    $self->{log} = $log if (defined $log);

    # Set the timeout if one is given.
    $self->timeout($timeout) if (defined $timeout);

    # Set the WinClient object, will be undef unless windows
    $self->{WinClient} = $winclient_obj;

    if ($self->hosttype() =~ /windows/) {
        # add some attributes on SanClient objects to stay
        # consistent with UNIX clients
        $self->{revision} = $self->{WinClient}->{hostver};
        $self->{minor_revision} = "NA";
        $self->{arch} = $self->{WinClient}->{arch};
        $self->{dsms} = $self->{WinClient}->{dsms};
    }

    # Set the test_bin_dir if provided
    if ($test_bin_dir ne "NOT_SET") {
        $self->{TEST_BIN} = $test_bin_dir;
    }

    # Set the pnate_environment flag
    $self->{pnate_environment} = $pnate_environment;

    # initialize the connect retries on the object
    $self->{connection_retries} = $connection_retries;

    # Setting the multipathing type if given, check to see if it's already
    # set (possibly by platform specific _new())
    if (not defined ($self->{mp_type})) {
        $self->{mp_type} = $mp_type;
    } else {
        # if the default mp_type was set in the host contructor and
        # it is not the same as what was passed (and it not "none")
        # use what was passed in (instead of the default set in the
        # contructor).
        if (($self->{mp_type} ne $mp_type) and ($mp_type ne "none")) {
            $self->{mp_type} = $mp_type;
        }
    }

    # Set the debug_level if one is given.
    $self->set_debug_level($debug_level);

    # Set the alua_support if one is given, else default to off (0)
    if (defined($alua_support)) {
        $self->{alua_support} = $alua_support;
    } else {
        $self->{alua_support} = 0;
    }

    # store the hypervisor settings
    $self->{is_virtual} = $is_virtual;
    $self->{hypervisor_type} = $hypervisor_type;

    $self->{iscsi_type} = $iscsi_type;

    # Load configuration params for iSCSI, fcp, and hu
    $self->_load_config_params();

    # Check if client supports user specified protocol.
    # Retrieve list of supported protocols and compare
    # to what was  passed in
    @protocols = $self->get_supported_protocols();
    $protocol = $self->get_protocol();

    unless ("@protocols" =~ /$protocol/i) {
        $self->_set_err("$protocol protocol on " .
                        $self->hosttype() .
                        " is currently not supported by SanClient object");
        return $self;
    }
    $self->{load_inits} = $load_inits;
    if ($protocol =~ /fcp|iscsi/i && $self->{load_inits}) {
        $hArgs{-loghdl} = $self->{log} if (defined $self->{log});
        if ($self->hosttype() =~ /windows/) {
            $hArgs{-WinClient} = $self->{WinClient} if (defined $self->{WinClient});
        }
        # Create a Sanlun object with the parent being passed to get protocol and conn
        $sanlun_obj = Sanlun->new(-parent => $self,
                                  %hArgs);
        if ($sanlun_obj->error) {
            $self->_set_err($sanlun_obj->errmsg());
            return $self;
        }
        $sanlun_obj->set_debug_level($self->get_debug_level());
        $self->{sanlun} = $sanlun_obj;
        if ($self->hosttype() =~ /unix-linux-sanblaze/) {
            # For the SanBlaze host, undef the initiators attribute
            $self->{initiators} = undef;
        } else {
            # Load initiators
            $self->sanlog(-msg => "Loading Initiators",
                          -msg_level => 3);
            $self->_load_initiators($self->get_protocol(), $any_iscsi);
            return $self if ($self->error);
        }
    } else {
        $self->sanlog(-msg => "Load Initiators skipped due to (-load_inits=>" .
                              $self->{load_inits} . ") being defined",
                      -msg_level => 3);
    }

    # Create the multipathing object if mp_type is specified
    if ($self->{mp_type} ne "" && $self->{mp_type} !~ /^none$/i) {
        $self->create_mp(-mp_type => $self->{mp_type});
        if ($self->error) {
            return $self;
        }
    }
    $self->{mp}->{mp_type} = $self->{mp_type};

    return $self;
}

###############################################################################
# Method:    options
# Objective: set user definable options for client object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub options {
    my ($self, %args) = @_;

    my $error = 0;
    my $errstr ="";

    foreach my $key (keys %args) {
        if ($key =~ /^-?(.+)$/) {
            my $option_key = $1;
            if (not length($self->{$option_key} = $args{$key})) {
                $errstr .= "Invalid value for argument $option_key\n";
                $error++;
            }

            # also, set the attributes on the WinClient object.
            # This is a hack.
            if ($self->hosttype() =~ /windows/i) {
                my $winclient = $self->{WinClient};
                $winclient->{$option_key} = $args{$key};
            }
        }
    }

    if ($error > 0) {
        $self->_set_err($errstr);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: detect_reboot ()
# Objective:       Detect whether host rebooted or not
# Details:         See POD documentation at the end of this file
################################################################################
sub detect_reboot {
    my($self, %args) = @_;

    my ($retry, $alive, $num_tries, $machine, $machine_ip);
    my ($shutdown_timeout, $bootup_timeout, $bootup_sleep);

    $machine = $self->hostname();

    foreach (keys %args) {
        if (/-?shutdown_timeout/) {
            $shutdown_timeout = $args{$_};
            next;
        } elsif (/-?bootup_timeout/) {
            $bootup_timeout = $args{$_};
            next;
        } elsif (/-?bootup_sleep/) {
            $bootup_sleep = $args{$_};
        }
    }

    # Assigning default timeout values
    if (!defined($shutdown_timeout)) {
        $shutdown_timeout = 300;
    }
    if (!defined($bootup_timeout)) {
        $bootup_timeout = 900;
    }
    if (!defined($shutdown_timeout)) {
        $bootup_sleep = 180;
    }

    # We will ping the host for a total of ($shutdown_timeout)
    # seconds before we fail out if host didnt stop pinging. This means we are
    # unable to soft reboot the host. Hard reboot is required
    $alive = 1;
    while ($shutdown_timeout > 0) {
        # ping the host to see if it is dead or not
        my $temp_time = time();
        if (&SanUtil::ping_host($machine_ip, $machine) == $ClientUtil::FALSE) {
            $alive = 0;
            $self->sanlog(-msg => "($machine) shutdown successfully");
            last;
        } else {
            $self->sanlog(-msg => "($machine) still alive. sleeping for 5 seconds before trying to ping again");
            &SanUtil::nate_sleep(5);
            $shutdown_timeout = $shutdown_timeout - (time() - $temp_time);
        }
    }

    # check the status of shutdown and report
    if ($alive) {
        $self->_set_err("Not able to soft reboot the Machine ($machine). " .
                        "User should manually try to hard reboot the machine");
        return $ClientUtil::FALSE;
    }

    # sleep just enough to prime the reboot process
    &SanUtil::nate_sleep($bootup_sleep);

    # We will ping host for a total of ($bootup_timeout + $bootup_sleep)
    # seconds before we fail out if host still didnt ping. This means that host
    # didnt come up after shutdown. Hard reboot is required
    $alive = 0;
    while ($bootup_timeout > 0) {
        # ping host to see if it is up
        my $temp_time = time();
        if (&SanUtil::ping_host($machine_ip, $machine) == $ClientUtil::TRUE) {
            $alive = 1;
            $self->sanlog(-msg => "($machine) rebooted successfully");
            last;
        } else {
            $self->sanlog(-msg => "($machine) not up sleeping 30 seconds");
            &SanUtil::nate_sleep(30);
            $bootup_timeout = $bootup_timeout - (time() - $temp_time);
        }
    }

    # check status of reboot and report
    if ($alive == 0) {
        $self->_set_err("Machine ($machine) never came back up after reboot. " .
                        "User should manually try to hard reboot the machine");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy ()
# Objective:       destroy client object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;

    $self->disconnect_connection();

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    create_lvm()
# Objective: Create and pass back a lvm handle
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_lvm {
    my($self, %args) = @_;

    my ($lvm, $lvmObj, $loadcfg, %hArgs);

    $self->_clear_err();

    $loadcfg = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_})) {
                $self->_set_err("Invalid value for argument '-lvm'");
                return $self;
            }
            next;
        } elsif (/-?loadcfg/) {
            if (not length($loadcfg = $args{$_}) or $loadcfg !~ /no|yes|1|0|off|on/i) {
                $loadcfg=1;
            }
            next;
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $self;
        }
    }

    # check for existance of required -lvm argument
    unless (defined $lvm and length $lvm) {
        $self->_set_err("Missing required argument '-lvm'");
        return $self;
    }


    $hArgs{-loghdl} = $self->{log} if (defined $self->{log});
    $lvmObj = Lvm->_new(-parent => $self,
                        -lvm => $lvm,
                        -loadcfg => $loadcfg,
                        %hArgs);
    if ($lvmObj->error) {
        $self->_set_err($lvmObj->errmsg());
        return $self;
    }

    # Keep pointer to first snapdrive object created by client
    $self->{obj}{lvm} = $lvmObj if (not defined $self->{obj}{lvm}); # Keep list of child objects created

    return $lvmObj;
}

###############################################################################
# Method:    create_devmgr()
# Objective: Create and pass back a devmgr handle
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_devmgr {
    my ($self, %args) = @_;

    my ($dev, $dmObj, %hArgs);

    $self->_clear_err();

    $hArgs{-loghdl} = $self->{log} if (defined $self->{log});
    $dmObj = DevMgr->_new(-parent => $self, %hArgs, %args);
    if ($dmObj->error) {
        $self->_set_err($dmObj->errmsg());
        return $self;
    }

    # Keep pointer to first snapdrive object created by client
    $self->{obj}{devmgr} = $dmObj if (not defined $self->{obj}{devmgr});
    # Device manager object created successfully..returning to THPL
    # Keep list of child objects created
    return $dmObj;
}

###############################################################################
# Method:    create_mpdevmgr()
# Objective: Create an MPDevMgr object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_mpdevmgr {
    my ($self, %args) = @_;

    $self->_clear_err();

    # instantiate an MPDevMgr object
    my $mp_devmgr_obj = San::SanClient::MPDevMgr->new(-sc_obj => $self,
                                                      %args);
    if ($mp_devmgr_obj->error) {
        $self->_set_err("Failed to create an MPDevMgr object", $mp_devmgr_obj->errmsg());
        return $self;
    }

    # Keep pointer to the mpdevmgr object
    $self->{obj}{mpdevmgr} = $mp_devmgr_obj;

    return $mp_devmgr_obj;
}

###############################################################################
# Method:    create_hu()
# Objective: Create and pass back a hu object
###############################################################################
sub create_hu {
    my ($self, %args) = @_;

    my ($huobj);

    $self->_clear_err();

    $self->sanlog(-msg => "Creating HU Object",
                  -msg_level => 3);
    $huobj = HU->new(-parent => $self,
                     %args);
    if ($huobj->error) {
        $self->_set_err($huobj->errmsg());
        return $huobj;
    }

    $self->{huobj} = $huobj;

    return $huobj;
}

###############################################################################
# Method:    get_hu_toolkit_version()
# Objective: return san toolkit revision from HU object
###############################################################################
sub get_hu_toolkit_version {
    my ($self) = @_;

    my ($version, $huobj);

    $huobj = $self->{huobj};
    $version = $huobj->{toolkit_version};

    return($version);

}

###############################################################################
# Method:    get_hu_sanlun_version()
# Objective: return sanlun revision from HU object
###############################################################################
sub get_hu_sanlun_version {
    my ($self) = @_;

    my ($version, $huobj);

    $huobj = $self->{huobj};
    $version = $huobj->{sanlun_version};

    return($version);

}

###############################################################################
# Method:    get_hu_bin_path()
# Objective: return HU path for binaries
###############################################################################
sub get_hu_bin_path {
    my ($self) = @_;

    my ($path, $huobj);

    $huobj = $self->{huobj};
    $path = $huobj->{hu_bin_path};

    return($path);

}

###############################################################################
# Method:    get_mp_type()
# Objective: return multipathing type
###############################################################################
sub get_mp_type {
    my ($self) = @_;

    my ($mp_type);

    $mp_type = $self->{mp_type};

    return($mp_type);

}

###############################################################################
# Method:    mp()
# Objective: return the multipathing object
###############################################################################
sub mp {
    my ($self) = @_;

    return $self->{mp};
}

###############################################################################
# Method:    mpdevmgr()
# Objective: return the MPDevMgr object
###############################################################################
sub mpdevmgr {
    my ($self) = @_;

    return $self->{obj}{mpdevmgr};
}

###############################################################################
# Method:    execute
# Objective: Issue arbitrary command to host and get results.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub execute {
    my ($self, $cmd, %args) = @_;
    my (@arr, $buffer, $hide, @rows, $cmd_fmt, $hostcmd, $errmsg, $t0, $t1);

    $hide = 0;

    foreach (keys %args) {
        if (/-?hide/) {
            length($hide = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"hide\":",
                                    "$args{$_}");
                    return $self;
                };
        } elsif (/-?_hostcmd/) {
            length($hostcmd = $args{$_}) or
                do {
                    $self->_set_err("Invalid value for \"-_hostcmd\":",
                                    "$args{$_}");
                    return $self;
                };
        } else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $self;
        }
    }

    # Clear errors.
    $self->_clear_err;

    unless (length $cmd) {
        $self->_set_err("Invalid CLI command");
        return [];
    }

    if (!$hide and $self->{show_cmd}) {
        $self->sanlog(-msg => $cmd, -use_loghdl => 0);
    }

    return (wantarray()) ? () : [] unless ($self->{execute});

    # Init execution timer
    $t0 = gettimeofday;

    # Execute the command
    my $status;
    my @output;
    my $tmo     = $self->timeout();

    if ($self->hosttype() =~ /windows/) {
        ($status, @output) = $self->{WinClient}->execute($cmd);
    } elsif(defined $tmo && length $tmo) {
        ($status, @output) = &ClientUtil::execute($self->{conn},
                                                  $cmd,
                                                  $tmo);
    } else {
        ($status, @output) = &ClientUtil::execute($self->{conn},$cmd);
    }

    # Get the post execution time and save the command time
    $t1 = gettimeofday;
    $self->{command_time} = ($t1 - $t0);

    # Check for standard set of errors returned from the CLI.
    if ($status) {
        $self->_set_err("Error: ",@output);
        return (wantarray()) ? @output : \@output;
    }

    return (wantarray()) ? @output : \@output;

}
################################################################################
# Subroutine Name: get_initiators
# Objective:       return array of initiator objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_initiators {
    my ($self) = @_;

    my (@initObjs);

    $self->_clear_err();
    if (defined $self->{initiators}) {
        @initObjs = @{$self->{initiators}};
    } else {
        # If the hosttype is unix-linux-sanblaze, try to load the initiators
        if ($self->hosttype() =~ /unix-linux-sanblaze/) {
            $self->_load_initiators($self->get_protocol(), 0);
            # Either an error occurred or the initiators list is empty
            # set error and return back empty array
            if ($self->error() or not defined ($self->{initiators})) {
                $self->_set_err("Error loading SanBlaze initiators");
                return ();
            } else {
                @initObjs = @{$self->{initiators}};
            }
        } else {
            $self->_set_err("SanClient object has no initiators");
            return ();
        }
    }

    return (wantarray()) ? @initObjs : \@initObjs;
}

###############################################################################
# Method:    get_sanlun()
# Objective: Get sanlun object handle
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_sanlun {
    my ($self, %args) = @_;

    $self->_clear_err();

    if (not defined $self->{sanlun}) {
        $self->_set_err("Sanlun object was not created.  Client protocol " .
                        "must be set to either 'fcp' or 'iscsi'");
        return {};
    }

    return $self->{sanlun};

}

###############################################################################
# Method:    get_connection()
# Objective: Get nate connection object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_connection {
    my ($self) = @_;

    return $self->{conn};
}

###############################################################################
# Method:    get_exec_time()
# Objective: Get execution time of snap command statement
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_exec_time {
    my($self) = @_;

    if (length $self->{command_time}) {
        return sprintf "%.4f", $self->{command_time};
    }

    return "";
}

################################################################################
# Subroutine Name: get_protocol
# Objective:       Return protocol type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{protocol}) {
        return $self->{protocol};
    } else {
        $self->_set_err("SanClient object is missing protocol type");
        return "";
    }
}

################################################################################
# Subroutine Name: get_WinClient
# Objective:       Return WinClient object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_WinClient {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{WinClient}) {
        return $self->{WinClient};
    } else {
        $self->_set_err("SanClient object is missing WinClient object");
        return "";
    }
}

###############################################################################
# Method:    hostname()
# Objective: Get client host name from the host record
# Details:   See POD documentation at the end of this file.
###############################################################################
sub hostname {
    my ($self) = @_;

    # need to return the physical host name if there is a vMA host
    if ($self->has_vma_host()) {
        return $self->{phy_host};
    } else {
        return $self->{conn}{hostp}{hostname};
    }
}

###############################################################################
# Method:    hosttype()
# Objective: Get client host type from the host record
# Details:   See POD documentation at the end of this file.
###############################################################################
sub hosttype {
    my ($self) = @_;

    return $self->{conn}{hostp}{hosttype};
}

###############################################################################
# Method:    os_version()
# Objective: get the os version from the host record
# Details:   See POD documentation at the end of this file.
###############################################################################
sub os_version {
    my ($self) = @_;

    return $self->{conn}{hostp}{version};
}

###############################################################################
# Method:    disconnect_connection()
# Objective: disconnect the connection to the client
# Details:   See POD documentation at the end of this file.
###############################################################################
sub disconnect_connection {
    my ($self) = @_;

    if (defined($self->{conn})) {
        # disconnect the internal connection
        Tharn::disconnect($self->{conn}, '-ok');
    }

    # unset the connected flag
    $self->{connected} = 0;

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    reconnect()
# Objective: reconect to host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub reconnect {
    my ($self, %args) = @_;

    # default action is to disconnect the connect that currently exists
    my $disconnect_existing_conn = exists($args{-disconnect_existing_conn}) ? $args{-disconnect_existing_conn} : 1;

    $self->_clear_err();
    if (defined($self->{conn}) and $disconnect_existing_conn == 1) {
        Tharn::disconnect($self->{conn}, '-ok');
    }

    $self->sanlog(-msg => "Reconnecting to host machine '" . $self->hostname() . "'");
    $self->{conn} = $self->_connect($self->{host},
                                    $self->{user},
                                    $self->{pass},
                                    $self->{timeout},
                                    $self->{conn_type});

    if ($self->error) {
        my @error = $self->errmsg();
        my $errmsg  = join('\n', @error);
        $self->_set_err("connect failed: $errmsg");
        return ($ClientUtil::FALSE);
    }

    # update the connection on the sub objects if they exist
    if (exists($self->{mp})) {
        $self->{mp}->{conn} = $self->{conn};
    }
    if (exists($self->{sanlun})) {
        $self->{sanlun}->{conn} = $self->{conn};
    }
    if (exists($self->{obj})) {
        if (exists($self->{obj}{lvm})) {
            $self->{obj}{lvm}->{conn} = $self->{conn};
        }
        if (exists($self->{obj}{devmgr})) {
            $self->{obj}{devmgr}->{conn} = $self->{conn};
        }
    }
    if (exists($self->{huobj})) {
        $self->{huobj}->{conn} = $self->{conn};
    }

    # for ESXi hosts, we need to re-set the vMA context
    if ($self->has_vma_host()) {
        $self->_set_vma_context();
        if ($self->error()) {
            $self->_set_err("Failed to set vMA context", $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    timeout
# Objective: set or get telnet timeout
# Details:   See POD documentation at the end of this file.
###############################################################################
sub timeout {
    my ($self, $timeout) = @_;

    if (defined $timeout and length $timeout) {
        $self->{timeout} = $timeout;
    } else {
        return $self->{timeout};
    }
}

###############################################################################
# Method:    get_testbin_path
# Objective: get the TEST_BIN attribute
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_testbin_path {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{TEST_BIN}) and
        length($self->{TEST_BIN}) and
        $self->{TEST_BIN} !~ /NOT_SET/i) {
        return $self->{TEST_BIN};
    } else {
        $self->_set_err("SanClient object has not set the TEST_BIN attribute");
        return "";
    }
}

###############################################################################
# Method:    create_mp
# Objective: create a MP object object corresponding to the passed in MP_TYPE
#            and associate it on the SanClient object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_mp {
    my ($self, %args) = @_;

    $self->_clear_err();

    # mp_type is a required parameter
    if (not exists($args{-mp_type})) {
        $self->_set_err("Missing required parameter '-mp_type'");
        return {};
    }

    my ($mpObj);

    $self->sanlog(-msg => "Creating MP Object",
                  -msg_level => 3);

    # create MP object
    $mpObj = MP->new(-hosttype => $self->hosttype(),
                     -parent => $self,
                     %args);
    if ($mpObj->error) {
        $self->_set_err("Unable to create MP object", $mpObj->errmsg());
        return $mpObj;
    }

    # reset MP object on SanClient and reset MP_TYPE
    $self->{mp} = $mpObj;
    $self->{mp}->{mp_type} = $args{-mp_type};
    $self->{mp_type} = $args{-mp_type};

    return $mpObj;
}

###############################################################################
# Method:    display_info
# Objective: dump client information for diagnostic/audit purposes
# Details:   See POD documentation at the end of this file.
###############################################################################
sub display_info {
    my ($self) = @_;

    my ($msg);

    $self->_clear_err();

    # gather general information
    $msg = "Client Information:\n" .
           "Name:                     " . $self->hostname() . "\n" .
           "Type:                     " . $self->hosttype() . "\n" .
           "Major OS version:         " . $self->{revision} . "\n" .
           "Minor OS version:         " . $self->{minor_revision} . "\n" .
           "Machine architecture:     " . $self->{arch} . "\n" .
           "Host Utilities version:   " . $self->get_hu_toolkit_version() . "\n" .
           "Sanlun version:           " . $self->get_hu_sanlun_version() . "\n" .
           "Multipathing type:        " . $self->get_mp_type() . "\n" .
           "Protocol in use by tests: " . $self->get_protocol() . "\n" .
           "ALUA specified by tests:  " . $self->get_alua_support() . "\n";

    # get adapter information (if loaded)
    if (defined($self->{sanlun})) {
        my @output = $self->{sanlun}->display_adapters($self->get_protocol());
        my $str = join("\n", @output);
        if ($str ne "") {
            $msg .= "\nAdapter Information:\n" . $str . "\n";
        }
    }

    # If this OS type can get iscsi info, then do so now.
    # this will allow a get iscsi info method for each OS
    if ($self->can("get_iscsi_info")) {
        my $iscsi_hash = $self->get_iscsi_info();
        if ($self->error()) {
            $self->_clear_err();
        } else {
            my $str = Dumper($iscsi_hash);
            if ($str ne "") {
                $msg .= "\nIscsi Information:\n" . $str . "\n";
            }
        }
    }

    # log the information
    $self->sanlog(-msg => $msg);

    # If this OS type can display additional info, then do so now.
    # this will allow a specific display method for each OS
    if ($self->can("display_additional_info")) {
        $self->display_additional_info();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rotate_generic_log()
# Objective:       Rotate a generic log
################################################################################
sub rotate_generic_log {
    my ($self, %args) = @_;

    my ($cmd, $status, @output, $filename, $decoded_filename, $minsize, $keep);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?minsize/) {
            if (not length($minsize = $args{$_})) {
                $self->_set_err("Invalid value for argument '-minsize'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?keep/) {
            if (not length($keep = $args{$_})) {
                $self->_set_err("Invalid value for argument '-keep'");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?file/) {
            if (not length($filename = $args{$_})) {
                $self->_set_err("Invalid value for argument '-file'");
                return $ClientUtil::FALSE;
            }
        }
    }

    # Check to make sure a file name was passed in as an argument
    if (defined($filename)) {
        if ($filename eq "") {
            $self->_set_err("Invalid value for argument '-file'");
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Missing required argument '-file'");
        return $ClientUtil::FALSE;
    }

    # Default minsize to 1024 if it isn't passed in as an argument
    if (not defined($minsize) or $minsize eq "") {
        $minsize = 1024;
    }

    # Default $keep to 5 if it isn't passed in as an argument
    if (not defined($keep) or $keep eq "") {
        $keep = 5;
    }

    # Try to decode the file $filename in case it is a symlink
    my $file_exists = &ClientUtil::remote_file_exists($self->{conn}, $filename);
    if ($file_exists == $ClientUtil::TRUE) {
        $decoded_filename = &ClientUtil::decode_remote_symlink($self->{conn}, $filename);
        if (not defined($decoded_filename)) {
            $self->_set_err("Unable to determine whether the file $filename is a symlink or not");
            return $ClientUtil::FALSE;
        }

        # Now that we know what file to look for and we know it exists, get the filesize
        # so we can determine if we want to rotate it.
        my $size = &ClientUtil::remote_get_filesize($self->{conn}, $decoded_filename);
        if (not defined($size) or $size eq "") {
            $self->_set_err("Unable to determine size of file $decoded_filename");
            return $ClientUtil::FALSE;
        } else {
            if ($size <= $minsize) {
                # Looks like the size is less than or equal to $minsize so we won't bother
                # with it. Just return TRUE.
                return $ClientUtil::TRUE;
            }
        }

        # Now that we know that the file needs rotated, go ahead and rotate
        # the files.  We will keep $keep copies (the original plus 0 -> $keep - 1).
        # We need to define $index_start as keep - 2 since we index at 0 and we also
        # need to start copying logs that are one less than the max number thus we would
        # need a total of 2 less. Keeping 5 logs would mean keeping numbers 0-4 (keep - 1) and
        # since we start at one less than the final number, we need ((keep - 1) - 1) or keep - 2.
        my $index_start = $keep - 2;
        for (my $count = $index_start; $count >= 0; $count--) {
            my $next = $count + 1;
            my $logfile = $decoded_filename . "." . $count;
            my $new_logfile = $decoded_filename . "." . $next;
            my $file_exists = &ClientUtil::remote_file_exists($self->{conn}, $logfile);
            if ($file_exists == $ClientUtil::TRUE) {
                my $cmd = "mv $logfile $new_logfile";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("($cmd) returned a bad status ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }

        # Now that we have moved filename.0 thru filename.($keep - 1), we now need to copy (not move)
        # filename to filename.0 and then cat /dev/null into filename.
        my @cmds = ("cp $decoded_filename ${decoded_filename}.0",
                    "cat /dev/null > $decoded_filename");
        foreach my $cmd (@cmds) {
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: has_vma_host()
# Objective:       Return true or false depending on whether this host is under
#                  vMA control or not.
################################################################################
sub has_vma_host {
    my ($self) = @_;

    if (defined($self->{vma_host}) and $self->{vma_host} ne "") {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

###############################################################################
#                               Private Methods
###############################################################################

###############################################################################
# Method:    _connect
# Objective: Create telnet connection to host
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _connect {
    my ($self, $hostname, $user, $passwd, $timeout, $conn_type) = @_;

    my ($connect, $conn_str, $errmsg, $host, $id);
    my ($hostInfo, $retries, @preferred_conns);

    # Variable init.
    $self->{connected} = 0;

    $id = $SanClient::_CLIENT_CNT++;

    $host = Utils::make_host($hostname);
    my $max_retries = $self->{connection_retries};

    if ($conn_type eq "THARNHOST") {
        $conn_type = "";

        # get list of preferred connection types for this
        # platforms. No error to check in this context.
        @preferred_conns = $self->_get_preferred_conn_types($host);

        # Check to see if any connection types are defined, if not return the default value
        if (!defined($host->{connection_fields}) or
            scalar(keys(%{$host->{connection_fields}})) == 0) {
            # Set the $conn_type to the default of telnet
            $conn_type = "telnet";
            if ($host->hosttype() =~ /windows/) {
                $conn_type = "rsh";
            }
        } elsif (scalar(@preferred_conns) > 0) {
            # If we find preferred connection types in the
            # for this platform, try each of those to see if they
            # are defined, if one is, use it.
            foreach my $type (@preferred_conns) {
                if (defined($host->{connection_fields}->{$type})) {
                    $conn_type = $type;
                    last;
                }
            }
        }

        # Go through each remainder field and set $conn_type
        # IF rsh is defined, set that. This means that if the
        # conn type of rsh is defined in the tharnhost file,
        # we set $conn_type to that. To not use rsh, comment
        # out the conn type of rsh from the tharnhost file.
        # This will then check to see if the next available
        # conn type is defined in the thatnhost file.
        # The order of connection type is:
        # 1) rsh
        # 2) ssh
        # 3) telnet
        # 4) rlogin
        # 5) console
        if ($conn_type eq "") {
            if (defined($host->{connection_fields}->{rsh})) {
                $conn_type = "rsh";
            } elsif (defined($host->{connection_fields}->{"rsh-socket"})) {
                $conn_type = "rsh-socket";
            } elsif (defined($host->{connection_fields}->{"libssh-persay"})) {
                $conn_type = "libssh-persay";
            } elsif (defined($host->{connection_fields}->{"staf-persay"})) {
                $conn_type = "staf-persay";
            } elsif (defined($host->{connection_fields}->{ssh})) {
                $conn_type = "ssh-persist";
            } elsif (defined($host->{connection_fields}->{telnet})) {
                $conn_type = "telnet";
            } elsif (defined($host->{connection_fields}->{rlogin})) {
                $conn_type = "rlogin";
            } elsif (defined($host->{connection_fields}->{console})) {
                $conn_type = "console";
            }
        }

        # If conn type is still not defined, default.
        if ($conn_type eq "") {
            $conn_type = "telnet";
            if ($host->hosttype() =~ /windows/) {
                $conn_type = "rsh";
            }
        }
    } elsif ($conn_type eq "") {
        $conn_type = "telnet";
        if ($host->hosttype() =~ /windows/) {
            $conn_type = "rsh";
        }
    } else {
        # Default $conn_type to the specified value user passed in
    }

    unless (defined($timeout) and length($timeout)) {
        $self->timeout($host->{connection_fields}{$conn_type}{timeout});
    }

    if (defined($user) and length($user)) {
        $host->{connection_fields}{$conn_type}{username} = $user;
    }

    if (defined($passwd) and length($passwd)) {
        $host->{connection_fields}{$conn_type}{password} = $passwd;
    }

    if ($conn_type eq "ssh" or $conn_type eq "ssh-root") {
        $conn_str = $hostname . "#ssh" . "_" . $id . ";" .
            "conntype=ssh-persist;" .
            "username=" . $host->{connection_fields}{$conn_type}{username} . ";" .
            "password=" . $host->{connection_fields}{$conn_type}{password} . ";" .
            "timeout="  . $self->{timeout} . ";inherit=0;";
    } else {
        $conn_str = $hostname . "#" . $conn_type . ";" .
            "username=" . $host->{connection_fields}{$conn_type}{username} . ";" .
            "password=" . $host->{connection_fields}{$conn_type}{password} . ";" .
            "timeout="  . $self->{timeout} . ";inherit=0;";
    }

    if(defined $host->{connection_fields}->{$conn_type}->{shell_type}) {
        $conn_str = $conn_str . "shell_type=" . $host->{connection_fields}->{$conn_type}->{shell_type} . ";";
    }

    $self->sanlog(-msg_level => 0,
                  -msg => "Connection string: $conn_str");
    # if we are performing a connection test, hardcode the max retries to 0
    $retries = 0;
    while (!defined($connect) and $retries <= $max_retries) {
        $connect = &Tharn::connect(
            $conn_str,
            -eval=>sub {
                if (@_) {
                    &Tharn::disconnect($conn_str, '-ok');
                    $connect = undef;
                    $errmsg = join("", @_);
                    $errmsg =~ s/\n//g;

                    # Display an error in case user doesn't include error
                    # handler on object instantiation.
                    $self->sanlog(-msg => "WARNING: " .
                                          "unable to connect to client '" .
                                          $hostname."': $errmsg\n" .
                                          "Attempt $retries of $max_retries");
                    $retries++;
                    # don't sleep if we have no retries left
                    return if ($retries > $max_retries);

                    &SanUtil::nate_sleep(5);
                }
            });
    }

    # if retry count has been exhausted, set error and return
    if ($retries > $max_retries) {
        $self->_set_err($errmsg);
        return undef;
    }

    # we are now connected
    $self->{connected} = 1;

    # return the connection and the conn type
    return $connect;
}

################################################################################
# Subroutine Name: _get_objs
# Objective:       return array of children objects
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_objs {
    my ($self, %args) = @_;
    my (@objs);

    $self->_clear_err();

    if (defined $self->{objs}) {
        @objs = @{$self->{objs}};
    } else {
        $self->_set_err("No children objects to be logged");
        return  (wantarray()) ? () : [];
    }

    return (wantarray()) ? @objs : \@objs;

}
###############################################################################
# Method:    _load_initiators()
# Objective: load initiators
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _load_initiators {
    my ($self, $protocol, $any_iscsi) = @_;

    my ($cmd, $key, $name, $value, %hsh, @output, @slots);

    # Select specific fields from the output for each
    # initiator and store in hash
    #
    # We have two formats to support for getting the adapter name and WWPN:
    # lpfc0              WWPN:10000000c948d2e4
    #
    # adapter name:      td0
    # WWPN:              50060b0000bd73f6\r\n
    #

    # handle fcp initiators
    if ($protocol =~ /fcp/i) {
        # retrieve all initiator info from sanlun for fcp
        #
        # We are looking for fcp initiators like this:
        # @output[0] = "lpfc0              WWPN:10000000c92d617e";
        # @output[1] = "lpfc1              WWPN:10000000c938edc6";
        #
        @output = $self->{sanlun}->display_adapters("fcp");
        if ($self->{sanlun}->error) {
            $self->_set_err($self->{sanlun}->errmsg);
            return $ClientUtil::FALSE;
        }

        my $adapter_name = "xxx";
        foreach my $line (@output) {
            next unless ($line =~ /WWPN\:|^WWNN\:|^model\:|model description\:/ or
                         $line =~ /^driver version\:|^firmware version|^adapter name\:/i);

            #adapter name:      emlxs
            if ($line =~ /^adapter\s+name\:\s+(.+)/i) {
                $adapter_name = $1;
                $hsh{$adapter_name}{name} = $adapter_name;
                $key = "name"; $value = "$adapter_name";
            } elsif ($line =~ /^WWPN\:\s+(.+)/i) {
                $value = $1;
                $hsh{$adapter_name}{"WWPN"} = $value;
                $key = "WWPN";
            } elsif ($line =~ /WWPN/) {
                ($adapter_name, $key, $value) = $line =~ /(\S+)\s+(\w+)\:(\S+)/;
                $hsh{$adapter_name}{name} = $adapter_name;
            } else {
                ($key,$value) = $line =~ /(.*):\s+(.*)/i;
            }

            $hsh{$adapter_name}{$key} = $value;

            # store the protocol for this adpater
            $hsh{$adapter_name}{-protocol} = "fcp";
        }
    }

    # handle iSCSI initiators
    if ($protocol =~ /iscsi/i) {
        # display iscsi initiators
        @output = $self->{sanlun}->display_adapters("iscsi");
        if ($self->{sanlun}->error) {
            $self->_set_err($self->{sanlun}->errmsg);
            return $ClientUtil::FALSE;
        }

        foreach my $line (@output) {
            # Remove goofy control chars in output
            $line =~ s/\cI\cI/ /g;

            next unless ($line =~ /initiator name\:|model\:|model description\:/ or
                         $line =~ /driver version\:|WWPN\:|initiator node name/i or
                         $line =~ /firmware version\:/);

            if ($line =~ /initiator name/i) {
                ($name, $value) = $line =~ /(\S+)\s+initiator name:\s+(\S+)/;
                $key = "WWNN";  # change the key name
                $hsh{$name}{name} = $name;
            } elsif ($line =~ /Initiator\s+node\s+name:\s+/i) {
                ($value) = $line =~ /Initiator\s+node\s+name:\s+(.+)/i;
                $name++;
                $key = "WWNN";  # change the key name
                $hsh{$name}{name} = $name;
            } else {
                ($key,$value) = $line =~ /(.*):\s+(.*)/i;
            }
            if (length($any_iscsi) && length($value) && $key =~ /WWNN/){
                $self->sanlog(-msg => "Changing initiator IQN to special any_iscsi IQN");
                #we're using any_iscsi, we have
                #to trick the system into putting
                #a special iqn for the initiator
                #that any_iscsi will use
                $value = $any_iscsi;
            }
            $hsh{$name}{$key} = $value;

            # store the protocol for this adapter
            $hsh{$name}{-protocol} = "iscsi";
        }

        # Make NA the remainder attributes
        # $hsh{$name}{"WWPN"}="NA";
        if (not defined($hsh{$name}{"firmware version"}) or $hsh{$name}{"firmware version"} eq "") {
            $hsh{$name}{"firmware version"} = "NA";
        }
    }

    # Set the load_inits flag to show that the initiators have been loaded
    $self->{load_inits} = 1;

    # init stored hash since we don't want to keep adding to it
    @{$self->{initiators}} = ();

    # Walk hash and create an initiator object for
    # each initiator
    @slots = (sort keys %hsh);
    foreach my $slot (@slots) {
        $hsh{$slot}{-loghdl} = $self->{log} if (defined $self->{log});
        $hsh{$slot}{-parent} = $self;

        # Create initiator object
        my $initiatorObj = Initiator->_new($hsh{$slot});
        if ($initiatorObj->error) {
            $self->_set_err($initiatorObj->errmsg());
            return $ClientUtil::FALSE;
        }

        push @{$self->{initiators}}, $initiatorObj;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: is_reboot()
# Objective:       reboot host
# Details:         See POD documentation at the end of this file
################################################################################
sub is_reboot {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{reboot} and $self->{reboot} =~ /1|0/) {
        return $self->{reboot};
    } else {
        $self->{reboot} = 0;
    }

    return $self->{reboot};
}

################################################################################
# Subroutine Name: get_alua_support
# Objective:       Return alua_support boolean
# Details:         See POD documentation at the end of this file
################################################################################
sub get_alua_support {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{alua_support}) {
        return $self->{alua_support};
    } else {
        $self->_set_err("SanClient object is missing alua_support type");
        return "";
    }
}

################################################################################
# Subroutine Name: _load_config_params
# Objective:       This will check the input parameters and determine whether to
#                  load parameters for iscsi, fcp, and general
################################################################################
sub _load_config_params {
    my ($self, %args) = @_;

    # add the iscsi options into the SanClient object
    if ($self->{protocol} =~ /iscsi/i) {
        $self->sanlog(-msg => "Loading iSCSI Config Parameters",
                      -msg_level => 3);
        my $iscsi_options = $self->_load_config_iscsi_params();
        $self->options(-iscsi => $iscsi_options);
    }

    # add the fcp options into the SanClient object
    if ($self->{protocol} =~ /fcp/i) {
        $self->sanlog(-msg => "Loading FCP Config Parameters",
                      -msg_level => 3);
        my $fcp_options = $self->_load_config_fcp_params();
        $self->options(-fcp => $fcp_options);
    }

    # add general parameters
    $self->sanlog(-msg => "Loading General Config Parameters",
                  -msg_level => 3);
    my $general_options = $self->_load_config_general_params();
    $self->options(-general => $general_options);

    # create HU object (only if not using vma host)
    if (not $self->has_vma_host()) {
        $self->create_hu();
    }
}

################################################################################
# Subroutine Name: _load_config_general_params
# Objective:       This will take all the NATE params that are general
#                  and build a hash. It will then return this hash
#                  as a reference.
# Details:         See POD documentation at the end of this file
################################################################################
sub _load_config_general_params {
    my ($self) = @_;

    # pull out all the nate params
    &Tharn::param("OEM_REBRAND", "-default", "netapp");
    &Tharn::param("CONFIG_WIN_LBPOLICY", "-default", "0");

    my $general_href = {};

    # store the values into the datastructure
    $general_href->{OEM_REBRAND} = &Tharn::param("OEM_REBRAND");
    $general_href->{CONFIG_WIN_LBPOLICY} = &Tharn::param("CONFIG_WIN_LBPOLICY");

    return $general_href;
}

################################################################################
# Subroutine Name: _load_config_iscsi_params
# Objective:       This will take all the NATE params of type CONFIG_ISCSI
#                  and build a hash. It will then return this hash
#                  as a reference.
# Details:         See POD documentation at the end of this file
################################################################################
sub _load_config_iscsi_params {
    my ($self) = @_;

    # pull out all the nate params
    &Tharn::param("CONFIG_ISCSI_DISCOVERY", "-default", 1);
    &Tharn::param("CONFIG_ISCSI_DISCOVERY_PRECLEANUP", "-default", 2);
    &Tharn::param("CONFIG_ISCSI_SUBNET_MATCHING", "-default", 1);
    &Tharn::param("CONFIG_ISCSI_USE_MGMT_PORT", "-default", 0);
    &Tharn::param("CONFIG_ISCSI_DISCOVERY_TYPE", "-default", "sendtargets");
    &Tharn::param("CONFIG_ISCSI_TYPE", "-default", "software");
    &Tharn::param("CONFIG_ISCSI_HARDWARE_VENDOR", "-default", "");
    &Tharn::param("CONFIG_ISCSI_CHAP_MODE", "-default", "none");
    &Tharn::param("CONFIG_ISCSI_CHAP_USERNAME", "-default", "");
    &Tharn::param("CONFIG_ISCSI_CHAP_OUT_USERNAME", "-default", "");
    &Tharn::param("CONFIG_ISCSI_CHAP_IN_PASSWORD", "-default", "NATE12345678");
    &Tharn::param("CONFIG_ISCSI_CHAP_OUT_PASSWORD", "-default", "87654321ETAN");
    &Tharn::param("CONFIG_ISCSI_DATA_DIGEST", "-default", 0);
    &Tharn::param("CONFIG_ISCSI_HEADER_DIGEST", "-default", 0);
    &Tharn::param("CONFIG_ISCSI_ERL", "-default", 0);
    &Tharn::param("CONFIG_ISCSI_ISNS_SERVER", "-default", "");
    &Tharn::param("CONFIG_ISCSI_IPVERSION", "-default", "ipv4");
    &Tharn::param("CONFIG_ISCSI_IPV6_EXCLUDE_LINK_LOCAL", "-default", 1);
    &Tharn::param("CONFIG_ISCSI_DEFAULT_TIME2WAIT", "-default", "");
    &Tharn::param("CONFIG_ISCSI_DEFAULT_TIME2RETAIN", "-default", "");
    &Tharn::param("CONFIG_ISCSI_UDO_INITIAL_R2T", "-default", 0);
    &Tharn::param("CONFIG_ISCSI_UDO_IMMEDIATE_DATA", "-default", 0);
    &Tharn::param("CONFIG_ISCSI_WIN_LBPOLICY", "-default", 0);
    &Tharn::param("CONFIG_ISCSI_WIN_NUM_SESSIONS", "-default", 1);
    &Tharn::param("CONFIG_ISCSI_WIN_MCS_NUM_CONNS_PER_SESSION", "-default", 1);
    &Tharn::param("CONFIG_ISCSI_WIN_MPIO", "-default", 1);
    &Tharn::param("CONFIG_ISCSI_WIN_PERSISTENT", "-default", 0);
    &Tharn::param("CONFIG_ISCSI_WIN_KEEP_TARGET_PORTALS", "-default", 0);
    &Tharn::param("CONFIG_ISCSI_WIN_MAX_CONNECTIONS", "-default", "");

    my $iscsi_href = {};

    # store the values into the datastructure
    $iscsi_href->{CONFIG_ISCSI_DISCOVERY} =
        &Tharn::param("CONFIG_ISCSI_DISCOVERY");
    $iscsi_href->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} =
        &Tharn::param("CONFIG_ISCSI_DISCOVERY_PRECLEANUP");
    $iscsi_href->{CONFIG_ISCSI_SUBNET_MATCHING} =
        &Tharn::param("CONFIG_ISCSI_SUBNET_MATCHING");
    $iscsi_href->{CONFIG_ISCSI_USE_MGMT_PORT} =
        &Tharn::param("CONFIG_ISCSI_USE_MGMT_PORT");
    $iscsi_href->{CONFIG_ISCSI_DISCOVERY_TYPE} =
        &Tharn::param("CONFIG_ISCSI_DISCOVERY_TYPE");
    $iscsi_href->{CONFIG_ISCSI_TYPE} =
        &Tharn::param("CONFIG_ISCSI_TYPE");
    $iscsi_href->{CONFIG_ISCSI_HARDWARE_VENDOR} =
        &Tharn::param("CONFIG_ISCSI_HARDWARE_VENDOR");
    $iscsi_href->{CONFIG_ISCSI_CHAP_MODE} =
        &Tharn::param("CONFIG_ISCSI_CHAP_MODE");
    $iscsi_href->{CONFIG_ISCSI_CHAP_IN_PASSWORD} =
        &Tharn::param("CONFIG_ISCSI_CHAP_IN_PASSWORD");
    $iscsi_href->{CONFIG_ISCSI_CHAP_USERNAME} =
        &Tharn::param("CONFIG_ISCSI_CHAP_USERNAME");
    $iscsi_href->{CONFIG_ISCSI_CHAP_OUT_USERNAME} =
        &Tharn::param("CONFIG_ISCSI_CHAP_OUT_USERNAME");
    $iscsi_href->{CONFIG_ISCSI_CHAP_OUT_PASSWORD} =
        &Tharn::param("CONFIG_ISCSI_CHAP_OUT_PASSWORD");
    $iscsi_href->{CONFIG_ISCSI_DATA_DIGEST} =
        &Tharn::param("CONFIG_ISCSI_DATA_DIGEST");
    $iscsi_href->{CONFIG_ISCSI_HEADER_DIGEST} =
        &Tharn::param("CONFIG_ISCSI_HEADER_DIGEST");
    $iscsi_href->{CONFIG_ISCSI_ERL} =
        &Tharn::param("CONFIG_ISCSI_ERL");
    $iscsi_href->{CONFIG_ISCSI_ISNS_SERVER} =
        &Tharn::param("CONFIG_ISCSI_ISNS_SERVER");
    $iscsi_href->{CONFIG_ISCSI_IPVERSION} =
        &Tharn::param("CONFIG_ISCSI_IPVERSION");
    $iscsi_href->{CONFIG_ISCSI_IPV6_EXCLUDE_LINK_LOCAL} =
        &Tharn::param("CONFIG_ISCSI_IPV6_EXCLUDE_LINK_LOCAL");
    $iscsi_href->{CONFIG_ISCSI_DEFAULT_TIME2WAIT} =
        &Tharn::param("CONFIG_ISCSI_DEFAULT_TIME2WAIT");
    $iscsi_href->{CONFIG_ISCSI_DEFAULT_TIME2RETAIN} =
        &Tharn::param("CONFIG_ISCSI_DEFAULT_TIME2RETAIN");
    $iscsi_href->{CONFIG_ISCSI_UDO_INITIAL_R2T} =
        &Tharn::param("CONFIG_ISCSI_UDO_INITIAL_R2T");
    $iscsi_href->{CONFIG_ISCSI_UDO_IMMEDIATE_DATA} =
        &Tharn::param("CONFIG_ISCSI_UDO_IMMEDIATE_DATA");
    $iscsi_href->{CONFIG_ISCSI_WIN_LBPOLICY} =
        &Tharn::param("CONFIG_ISCSI_WIN_LBPOLICY");
    $iscsi_href->{CONFIG_ISCSI_WIN_NUM_SESSIONS} =
        &Tharn::param("CONFIG_ISCSI_WIN_NUM_SESSIONS");
    $iscsi_href->{CONFIG_ISCSI_WIN_MCS_NUM_CONNS_PER_SESSION} =
        &Tharn::param("CONFIG_ISCSI_WIN_MCS_NUM_CONNS_PER_SESSION");
    $iscsi_href->{CONFIG_ISCSI_WIN_MPIO} =
        &Tharn::param("CONFIG_ISCSI_WIN_MPIO");
    $iscsi_href->{CONFIG_ISCSI_WIN_PERSISTENT} =
        &Tharn::param("CONFIG_ISCSI_WIN_PERSISTENT");
    $iscsi_href->{CONFIG_ISCSI_WIN_MAX_CONNECTIONS} =
        &Tharn::param("CONFIG_ISCSI_WIN_MAX_CONNECTIONS");

    # if an iscsi type was passed in from the constructor, then load that value in
    # instead of the tharn param
    if (defined($self->{iscsi_type})) {
        $iscsi_href->{CONFIG_ISCSI_TYPE} = $self->{iscsi_type};
    }

    # If we have values in $self->{iscsi} add them to the $iscsi_href stack
    # so they can later be saved
    if (defined($self->{iscsi})) {
        foreach my $key (keys(%{$self->{iscsi}})) {
            $iscsi_href->{$key} = $self->{iscsi}->{$key};
        }
    }

    return $iscsi_href;
}

################################################################################
# Subroutine Name: _load_config_fcp_params
# Objective:       This will take all the NATE params of type CONFIG_FCP
#                  and build a hash. It will then return this hash
#                  as a reference.
# Details:         See POD documentation at the end of this file
################################################################################
sub _load_config_fcp_params {
    my ($self) = @_;

    # pull out all the nate params
    &Tharn::param("CONFIG_FCP_DRIVER_TYPE",          "-default", "");

    # AIX specific FCP parameters
    &Tharn::param("CONFIG_FCP_AIX_NUM_CMD_ELEMS",    "-default", "");
    &Tharn::param("CONFIG_FCP_AIX_DYNAMIC_TRACKING", "-default", "");
    &Tharn::param("CONFIG_FCP_AIX_ERROR_RECOVERY",   "-default", "");

    my $fcp_href = {};

    # store the values into the datastructure
    $fcp_href->{CONFIG_FCP_DRIVER_TYPE} =
        &Tharn::param("CONFIG_FCP_DRIVER_TYPE");

    $fcp_href->{CONFIG_FCP_AIX_NUM_CMD_ELEMS} =
        &Tharn::param("CONFIG_FCP_AIX_NUM_CMD_ELEMS");

    $fcp_href->{CONFIG_FCP_AIX_DYNAMIC_TRACKING} =
        &Tharn::param("CONFIG_FCP_AIX_DYNAMIC_TRACKING");

    $fcp_href->{CONFIG_FCP_AIX_ERROR_RECOVERY} =
        &Tharn::param("CONFIG_FCP_AIX_ERROR_RECOVERY");

    # If we have values in $self->{fcp} add them to the $fcp_href stack
    # so they can later be saved
    if (defined($self->{fcp})) {
        foreach my $key (keys(%{$self->{fcp}})) {
            $fcp_href->{$key} = $self->{fcp}->{$key};
        }
    }

    return $fcp_href;
}

################################################################################
# Subroutine Name: client_lun_filter
# Objective:       Filter client device list as specified by user
################################################################################
sub client_lun_filter {
    my ($self, $client_lun_filter, @devices) = @_;

    # If client lun filter is not set, return device list
    if ($client_lun_filter eq "") {
        return (wantarray()) ? @devices : \@devices;
    }

    # Create a temporary device array
    my @temp_devices = ();

    # If lun filter is a single number the user just wants X number
    # of luns to be used.
    my ($client_lun_filter_num);
    if ($client_lun_filter =~ /^(\d+)(\w*)$/) {
        # Valid options for sort_type is (r)andom or (s)equential
        # Random will grab at random devices from the device array
        # Sequential will take the 0-X devices from the device array
        $client_lun_filter_num = $1;
        my $sort_type = lc($2);
        if ($sort_type eq "") {
            $sort_type = "r";
        } elsif ($sort_type !~ /^r|s$/i) {
            $self->_set_err("Invalid value specified for sort type ($sort_type) provided " .
                            "with CLIENT_LUN_FILTER");
            return  (wantarray()) ? () : [];
        }

        my $num_devices = scalar(@devices);
        if ($client_lun_filter_num > $num_devices) {
            $self->_set_err("Number of devices ($num_devices) is less than the requested number of devices " .
                            "to filter to ($client_lun_filter)");
            return  (wantarray()) ? () : [];
        }

        # Loop through devices list and select the amount of devices
        # user wants to use
        for (my $i = 0; $i < $client_lun_filter_num; $i++) {
            my $index = -1;
            if ($sort_type eq "r") {
                # Randomaly select an index from the devices array
                $index = int(rand(scalar(@devices)));
            } elsif ($sort_type eq "s") {
                $index = $i;
            }

            # If the value of the index is undefined, skip it and redo the selection process
            if ($devices[$index] eq undef or $index < 0) {
                redo;
            }

            # Push value of device list onto temp device list and set
            # the device list index to undefined
            push @temp_devices, $devices[$index];
            $devices[$index] = undef;
        }

        $self->sanlog(-msg => "Filtered supplied lun list (total luns: " .
                              scalar(@devices) . ", filtered list: " . scalar(@temp_devices) . ") for a requested " .
                              "total of ($client_lun_filter)");
    } else {
        $self->_set_err("Invalid value specified for CLIENT_LUN_FILTER ($client_lun_filter)");
        return  (wantarray()) ? () : [];
    }

    return (wantarray()) ? @temp_devices : \@temp_devices;
}

################################################################################
# Subroutine Name: _get_preferred_conn_types
# Objective:       get the preferred list of connection types to use on a per
#                  platform basis. Cannot be in platform host objects,
#                  because _connect is called before per-platform contructors
################################################################################
sub _get_preferred_conn_types {
    my ($self, $host) = @_;

    my ($host_type);

    $host_type = $host->{hosttype};

    if ($host_type =~ /aix/i) {
        return("telnet", "ssh-root", "ssh", "rsh");
    } elsif ($host_type =~ /linux-vmware/i) {
        return("ssh-root", "ssh");
    } elsif ($host_type =~ /unix-linux/i) {
        return("ssh-root", "ssh", "rsh", "telnet");
    } elsif ($host_type =~ /hpux/i) {
        return("ssh-root", "rsh", "telnet", "ssh");
    } elsif ($host_type =~ /solaris/i) {
        return("libssh-persay", "rsh", "telnet", "ssh-root", "ssh");
    } elsif ($host_type =~ /windows/i) {
        return("rsh");
    } else {
        return undef;
    }

}

################################################################################
# Subroutine Name: filter_targets_by_ipversion()
# Objective:       Remove targets from the @targets array that doesn't have
#                  valid ip address of IP version which user requested
################################################################################
sub filter_targets_by_ipversion {
    my ($self, %args) = @_;

    my (@targets, @final_targets_list, $ip_version, $ip, @final_targets_summary);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # check for required params
    if (not scalar(@targets)) {
        $self->_set_err("Argument '-targets' is required");
        return (wantarray()) ? () : [];
    }

    # Get the IP version which user requested
    $ip_version = $self->{iscsi}->{CONFIG_ISCSI_IPVERSION};

    foreach my $target (@targets) {
        if ($ip_version =~ /ipv4/i) {
            $ip = $target->get_ipv4();
        } elsif ($ip_version =~ /ipv6/i) {
            $ip = $target->get_ipv6();
        } elsif ($ip_version =~ /auto/i) {
            $ip = $target->get_ip();
        } else {
            $self->_set_err("Invalid value for CONFIG_ISCSI_IPVERSION: $ip_version");
            return (wantarray()) ? () : [];
        }

        # if the target has valid IP address, add it to final target list
        if (defined($ip) and $ip ne "not_set") {
            push(@final_targets_list, $target);
            push(@final_targets_summary, "  $ip\n");
        }
    }

    $self->sanlog(-msg => "The following iSCSI targets have valid $ip_version addresses:\n " .
                          "@final_targets_summary");

    return (wantarray()) ? @final_targets_list : \@final_targets_list;
}

################################################################################
# Subroutine Name: _filter_targets_by_subnet()
# Objective:       Remove targets from the @targets array that are inaccessible
#                  from this hosts interfaces and are not on the same subnets
################################################################################
sub _filter_targets_by_subnet {
    my ($self, %args) = @_;

    my (@targets, @final_targets_list, @matched_subnets_summary, @unmatched_subnets_summary);
    my (@management_ports_summary);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # check for required params
    if (not scalar(@targets)) {
        $self->_set_err("Invalid value for argument '-targets'");
        return (wantarray()) ? () : [];
    }

    # get the interface information for this host.
    my $ip_info_hsh = $self->get_ip_info();
    if ($self->error()) {
        return (wantarray()) ? () : [];
    }

    # search through the targets and compare the subnets with each interface
    foreach my $target (@targets) {
        my ($check_ipv4_subnet, $check_ipv6_subnet) = (0,0);

        # check if this target has an ipv4 address
        my $target_ipv4 = $target->get_ipv4();
        my $target_ipv4_subnet;
        if (not $target->error()) {
            $check_ipv4_subnet = 1;
            $target_ipv4_subnet = $target->get_ipv4_subnet();
            if ($target->error()) {
                $self->_set_err("Could not obtain ipv4 subnet: " . $target->errmsg());
                return (wantarray()) ? () : [];
            }
        } else {
            $target->_clear_err();
        }

        # check if this target has an ipv6 address
        my $target_ipv6 = $target->get_ipv6();
        my $target_ipv6_subnet;
        if (not $target->error()) {
            $check_ipv6_subnet = 1;
            $target_ipv6_subnet = $target->get_ipv6_subnet();
            if ($target->error()) {
                $self->_set_err("Could not obtain ipv6 subnet: " . $target->errmsg());
                return (wantarray()) ? () : [];
            }
        } else {
            $target->_clear_err();
        }

        # skip this target if its a management interface and
        # iscsi options are configured to not use management
        # interfaces as targets
        if (($self->{iscsi}->{CONFIG_ISCSI_USE_MGMT_PORT} == 0) and ($target->is_mgmt_port() == 1)) {
            push(@management_ports_summary, "  $target_ipv4 $target_ipv6\n");
            next;
        }

        my $subnet_match = 0;

        # if requested to do subnet matching, do so now
        if ($self->{iscsi}->{CONFIG_ISCSI_SUBNET_MATCHING} == 1) {
            # check if the subnets of any of the interfaces on this host match the
            # subnets of this target for both ipv4 and ipv6
            foreach my $interface (values %{$ip_info_hsh}) {
                if ($check_ipv4_subnet and $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~ /ipv4|auto/i) {
                    if ($interface->{ipv4_subnet} eq $target_ipv4_subnet) {
                        $subnet_match = 1;
                        last;
                    }
                }

                if ($check_ipv6_subnet and $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~ /ipv6|auto/i) {
                    # if this is a link local ipv6 address, skip it if specified
                    if ($interface->{ipv6_subnet} =~ /^fe80:/ and
                        $self->{iscsi}->{CONFIG_ISCSI_IPV6_EXCLUDE_LINK_LOCAL} == 1) {
                        next;
                    }
                    if ($interface->{ipv6_subnet} eq $target_ipv6_subnet) {
                        $subnet_match = 1;
                        last;
                    }
                }
            }
        } else {
            # otherwise ignore the subnet matching.
            $subnet_match = 1;
        }

        # if a subnet matched this target, push it onto the final array
        if ($subnet_match) {
            push(@final_targets_list, $target);
            push(@matched_subnets_summary,   "  $target_ipv4 $target_ipv6\n");
        } else {
            push(@unmatched_subnets_summary, "  $target_ipv4 $target_ipv6\n");
        }
    }

    my @summary;
    # print out the summary of what was filtered out of the targets list
    if (scalar(@matched_subnets_summary)) {
        push(@summary, "The following iSCSI targets are accessible by this initiator:\n");
        push(@summary, @matched_subnets_summary);
    }
    if (scalar(@unmatched_subnets_summary)) {
        push(@summary, "\nThe following iSCSI targets are NOT accessible by this initiator " .
                       "(and have been excluded from the target discovery list):\n");
        push(@summary, @unmatched_subnets_summary);
    }
    if (scalar(@management_ports_summary)) {
        push(@summary, "\nThe following iSCSI targets are the management ports of the controllers " .
                       "(and have been excluded from the target discovery list):\n");
        push(@summary, @management_ports_summary);
    }

    $self->sanlog(-msg => "@summary");

    # if no targets are available, set error.
    if (scalar(@final_targets_list) == 0) {
        $self->_set_err("There are no targets accessible by any of this host's initiator ports.");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @final_targets_list : \@final_targets_list;
}

##############################################################################
# Method: get_isns_server()
# Objective: Get iSNS server object handle
# Details: See POD documentation at the end of this file.
##############################################################################
sub get_isns_server {
    my ($self) = @_;

    # Get iSNS server object
    if ($self->{isns_server}) {
        return $self->{isns_server};
    } else {
        $self->_set_err("Unable to get isns server object");
        return {};
    }

} ## end sub get_isns_server

##############################################################################
# Method:  create_isns_server()
# Objective: Create iSNS server object handle
# Details: See POD documentation at the end of this file.
##############################################################################
sub create_isns_server {
    my ($self) = @_;

    $self->sanlog(-msg => "Creating iSNS server object",
                  -msg_level => 3);

    my $isns_server_obj = San::SanClient::IsnsServer->_new(-parent => $self);
    if ($isns_server_obj->error) {
        $self->_set_err("Failed to create Isns server Object",
                        $isns_server_obj->errmsg());
        return $isns_server_obj;
    }

    $self->{isns_server} = $isns_server_obj;

    return $isns_server_obj;
} ## end sub create_isns_server

################################################################################
# Subroutine Name: get_interface_name()
# Objective:       get the name of the interface based on the ip address passed
#                  in.
################################################################################
sub get_interface_name {
    my ($self, %args) = @_;

    my ($ip_address, $ip_info, $interface_name);

    # parse arguments
    foreach (keys %args) {
        if (/-?ip_address/) {
            unless (defined ($ip_address = $args{'ip_address'}) and length ($ip_address)) {
                $self->_set_err("Invalid value for argument '-ip_address'");
                return "";
            }
            next;
        }
    }

    # check for required params
    unless (defined ($ip_address) and length ($ip_address)) {
        $self->_set_err("Invalid value for argument '-ip_address'");
        return "";
    }

    $ip_info = $self->get_ip_info();
    if($self->error) {
        $self->_set_err("Failed to get ip information" , $self->errmsg());
        return "";
    }

    # Searching through $ip_info hash to find the interface with
    # the same ip address and the passed in ip address
    foreach my $interface (keys %$ip_info) {
        if($ip_info->{$interface}->{ipv4} eq $ip_address) {
            return $ip_info->{$interface}->{interface_name};
        }
    }

    $self->_set_err("Failed to locate interface with ip \'$ip_address\'");
    return "";
}

################################################################################
# Subroutine Name: _create_lite_sanclient_object()
# Objective:       Given an object template, build a lightweight chatter-less
#                  SanClient object
################################################################################
sub _create_lite_sanclient_object {
    my ($self, %args) = @_;

    my $obj_template = $args{-obj_template};

    # first figure out which constructor to call. for the client object
    my $base_class = ref($obj_template);

    # if windows, rebuild the WinClient object
    if (defined($obj_template->{WinClient})) {
        $obj_template->{WinClient} = WinClient->new(-obj_template => $obj_template->{WinClient});
        if ($obj_template->{WinClient}->error()) {
            $self->_set_err("Failed to reconstruct WinClient object", $obj_template->{WinClient}->errmsg());
            return $self;
        }
    }

    # call the base class constructor in a light weight fashion
    if ($base_class =~ /Host::LinuxHost/i) {
        require San::SanClient::Host::LinuxHost;
    } elsif ($base_class =~ /Host::VLinuxHost/) {
        require San::SanClient::Host::VLinuxHost;
    } elsif ($base_class =~ /Host::WinHost/) {
        require San::SanClient::Host::WinHost;
    } elsif ($base_class =~ /Host::VWinHost/) {
        require San::SanClient::Host::VWinHost;
    } elsif ($base_class =~ /Host::ESXHost/) {
        require San::SanClient::Host::ESXHost;
    } elsif ($base_class =~ /Host::ESXiHost/) {
        require San::SanClient::Host::ESXiHost;
    } elsif ($base_class =~ /Host::SolarisHost/) {
        require San::SanClient::Host::SolarisHost;
    } elsif ($base_class =~ /Host::VSolarisHost/) {
        require San::SanClient::Host::VSolarisHost;
    } elsif ($base_class =~ /Host::AixHost/) {
        require San::SanClient::Host::AixHost;
    } elsif ($base_class =~ /Host::HPUXHost/) {
        require San::SanClient::Host::HPUXHost;
    } elsif ($base_class =~ /Host::KVMHost/) {
        require San::SanClient::Host::KVMHost;
    } elsif ($base_class =~ /Host::OVMHost/) {
        require San::SanClient::Host::OVMHost;
    } elsif ($base_class =~ /Host::CXSHost/) {
        require San::SanClient::Host::CXSHost;
    } elsif ($base_class =~ /Host::SanBlazeHost/) {
        require San::SanClient::Host::SanBlazeHost;
    } elsif ($base_class =~ /Host::HyperVHost/) {
        require San::SanClient::Host::HyperVHost;
    } else {
        $self->_set_err("SanClient Lite constructor does not support the base class '$base_class'");
        return $self;
    }

    # reinstantiate the base class
    $self = $base_class->_new(-obj_template => $obj_template);
    if ($self->error()) {
        $self->_set_err("Failed to instantiate base class $base_class", $self->errmsg());
        return $self;
    }

    # copy all the keys from the template
    foreach my $key (keys %{$obj_template}) {
        $self->{$key} = $obj_template->{$key};
    }

    # rebuild the connections hostrec object
    my $hostrec = Tharn::host($self->{conn}->{hostp}->{hostname});
    my $old_hostrec = $self->{conn}->{hostp};
    foreach (keys %{$old_hostrec}) {
        $hostrec->{$_} = $old_hostrec->{$_};
    }
    $self->{conn}->{hostp} = $hostrec;

    # fix all the initiators parents to point to $self
    # burt621713: only on non SanBlaze hosts
    if ($self->hosttype() !~ /unix-linux-sanblaze/i and $self->{protocol} =~ /iscsi|fcp/i) {
        my @objs = $self->get_initiators();
        foreach (@objs) {
            $_->{parent} = $self;
        }
    }

    # rebuild the sanlun object
    if ($self->{protocol} =~ /fcp|iscsi/i) {
        my %extra_args;
        $extra_args{WinClient} = $self->{WinClient} if (defined($self->{WinClient}));
        $self->{sanlun} = Sanlun->new(-obj_template => $self->{sanlun},
                                      -parent => $self,
                                      %extra_args);
        if ($self->{sanlun}->error()) {
            $self->_set_err("Failed to re-create the Sanlun object", $self->{sanlun}->errmsg());
            return $self;
        }
    }

    # rebuild the mp object if there is an mp type
    if ($self->{mp_type} ne "" and $self->{mp_type} !~ /^none$/i) {
        $self->create_mp(-mp_type => $self->{mp_type},
                         -obj_template => $self->{mp});
        if ($self->error()) {
            $self->_set_err("Failed to re-create the MP object", $self->errmsg());
            return $self;
        }
    }

    # this might be needed
    $self->{mp}->{mp_type} = $self->{mp_type};

    # rebuild the HU object
    if (defined($self->{huobj})) {
        $self->create_hu(-obj_template => $self->{huobj});
        if ($self->error()) {
            $self->_set_err("Failed to re-create the HU object", $self->errmsg());
            return $self;
        }
        # fix the parent link
        $self->{huobj}->{parent} = $self;
    }

    # rebuild the MPDevMgr if it exists
    if (defined($self->mpdevmgr())) {
        $self->create_mpdevmgr(-obj_template => $self->mpdevmgr());
        if ($self->error()) {
            $self->_set_err("Failed to re-create the MPDevMgr object");
            return $self;
        }
    }

    return $self;
}

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

SanClient.pm

=head1 Author

Gary Knopp

=head1 Description

SanClient.pm provides an API interface to a host machine through the
use of the Telnet library.

=head1 See Also

Package inherits from the following modules. See these modules for further
interface detail.

B<San::SanClient::Host::[I<SolarisHost|HPUXHost|AixHost|LinuxHost|WinHost>]>

B<Error.pm>

=head1 Attributes

=head2 B<connected>

boolean = <1 | 0>

=head2 B<show_cmd>

boolean = <1 | 0> (default: 1 (true) By default log commands executed)

=head2 B<show_data>

boolean = <1 | 0>

=head2 B<protocol>

string = <fcp | iscsi> (default: fcp)

=head2 B<execute>

boolean = <1 | 0> (default: 1)

=head2 B<log>

file handle = File handle used for logging

=head2 B<connect>

connection object = Connection to the host (as returned from Tharn::connect())

=head2 B<conn>

connection object = Connection to the host

=head2 B<host>

string = Host name, connection parameter

=head2 B<user>

string = Username, connection parameter

=head2 B<pass>

string = Password, connection parameter

=head2 B<sanlun>

sanlun object = Host sanlun object

=head2 B<obj>

hash = Hash of children objects (devmgr, lvm)

=head2 B<command_time>

integer = time in seconds of the last command executed

=head2 B<initiators>

array = List of initiator objects

=head2 B<timeout>

integer = timeout to use for say calls

=head2 B<debug_level>

integer = Current debug level

=head2 B<no_indent>

boolean = <1 | undef> Do not indent log entries when set

=head1 Public Methods

=head2 B<new(%args)>

Creates a new SanClient object.  Uses the specified IP address or machine name
to create a telnet connection. If the connection is successful, the calling
program will be logged into the CLIENT using either the default user/password
specified in the NATE_HOSTS file or a user and password supplied by the caller.

=over 4

=item Inputs

=item * -host

Machine name or IP address of the remote machine.

=item * I<-user> (default: Tharnhost username entry).

Override default username for host login.

=item * I<-passwd> (default: Tharnhost password entry).

Override default password for host login.

=item * I<-timeout> (default: Tharnhost telnet timeout entry).

Telnet timeout in seconds

=item * I<-debug_level> (default: 0).

Verbosity level for Sanlog::sanlog()

=item * I<-loghdl>

Filehandle to log output too.

Override the Telnet timeout value.

=item * I<-show_cmd> <1 | 0> (default: 1 (true) ).

By default commands executed will be logged   This can be
turned off by passing the -show_cmd with a value of 0.

=item * I<-connection>

If specified, a active/live connection must be passed in, the &connect function
must be called prior to calling the SanClient::new function

=item * I<-conn_type> <THARNHOST | rsh | telnet | ssh | rlogin | console> (NO DEFAULT).

This parameter will determine what connection type will be used for the client.
THIS CAN NOT BE USED IN CONJUNCTION WITH '-connection'. This parameter will NOT
override '-connection'.

=item Outputs

Returns a SanClient object.  If a successful Telnet connection and login was
achieved, an open Telnet connection is maintained by the object until
the calling program exits, or the B<disconnect> method is explicitly
called.

=item Examples

 # Create a Snapdrive  object with a connection to a remote machine.

 $Client = new SanClient(
        -host => '192.168.x.y',
        -user => 'smith',
        -passwd  => 'mypassword');

  if ($Client->error) {
     $Test->nlog($Client->errmsg());
     return $TCase::FAIL;
 }

=back

=head2 B<create_devmgr(%args)>

Creates a new DevMgr object

=over 4

=item Inputs

None

=item Outputs

Returns a DevMgr object.  If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Examples

 # Create a DevMgr object

  $DevMgr = $Client=>create_devmgr();
  if ($Client->error) {
     $Test->nlog($Client->errmsg());
     return $TCase::FAIL;
 }

=back

=head2 B<create_lvm(%args)>

Creates a new Lvm object

=over 4

=item Inputs

=item * -lvm

=item Outputs

Returns a Lvm object.  If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Examples

 # Create a Lvm object

  $Lvm = $Client=>create_lvm(
        -lvm=>'vxvm'
        );
  if ($Client->error) {
     $Test->nlog($Client->errmsg());
     return $TCase::FAIL;
 }

=back

=head2 B<execute($cmd,I<%args>)>

Executes command

=over 4

=item Inputs

=item * $cmd

Command to execute on the host.

=item * I<$hide> default: 0

Hide the command from STDOUT

=item Outputs

Returns an array or an array reference to the data returned by executing the command.
If an error occurred, the error flag is set, errmsg is set to the failure reason, and
an empty array reference is passed back as the return value.

=item Examples

 # Display command output from host
 $buf=$Client->execute("ls -l");
 if ($Client->error) {
     $Test->nlog($Client->errmsg());
     return $TCase::FAIL;
 }


 foreach $line (@$buf){
     print $line."\n";
 }

=back

=head2 B<get_connection()>

This method returns the nate conenction object

=over 4

=item Outputs

Returns a connection object. If an error occured the error flag
is set, errmsg is set to the failure reason

=item Examples

 #Retrieve execution time
 $connObj = $obj->get_connection();

=back

=head2 B<get_exec_time()>

This method returns the execution time in seconds of the last snap command statement executed.  Its the
actual time it took to process the statement and does not include any post processing of the data. The
timer used is a high resolution timer with millisecond accuracy.

=over 4

=item Outputs

Returns a scalar value representing the execution time in seconds.

=item Examples

 # Retrieve execution time

 $time = $obj->get_exec_time();
 print "Execution Time = $time seconds\n";

=back

=head2 B<get_initiators()>

Returns an array of initiator objects

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of Initiator objects.  If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Examples

 # Create a Lvm object
 @InitObjs = $Client=>get_initiators();
 if ($Client->error) {
     $Test->nlog($Client->errmsg);
     return 0;
 }

=back

=head2 B<get_protocol()>

Retrieve protocol used by client

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the protocol. If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Examples

 # Get protocol
 $protocol = $Client=>get_protocol();

 $Test->nlog("Protocol: ".$protocol);

=back

=head2 B<get_sanlun()>

Retrieve Sanlun object handle

=over 4

=item Inputs

none

=item Outputs

Returns a Sanlun object.  If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Examples

 # Create a Lvm object

 $obj = $Client=>get_sanlun();

 if ($obj->error) {
     print $obj->errmsg;
     return 0;
 }

=back

=head2 B<get_WinClient()>

Retrieve WinClient object used by client

=over 4

=item Inputs

none

=item Outputs

Returns the WinClient object. If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Examples

 # Get WinClient
 $WinClient_obj = $Client=>get_WinClient();

 # Update WinClients device (disk) list
 $WinClient_obj->getDevices();

=back

=head2 B<get_interface_name()>

Retrieve interface name of an interface on a client based on ip.

=over 4

=item Inputs

=item * <ip_address> the ip address of the interface

=item Outputs

Returns the name of the interface. If an error occured the error flag
is set, errmsg is set to the failure reason.

=item Examples

 #Get the name of the interface
 my $interface_name = $client->get_interface_name(ip_address => '10.225.115.238');
 if ($client->error) {
     &tstlog("-fatal", "Could not get the interface name: " . $client->errmsg());
 }

=back

=head2 B<hostname()>

This method allows you to retrieve the hostname of the client machine

=over 4

=item Outputs

Returns a scalar value containing the hostname.  If an error occured the error flag
is set, errmsg is set to the failure reason, and 0\'s are passed back as the return values.

=item Examples

 # Get build_dt
 $hostname= $Obj->hostname();
 if ($Obj->error){
     ...
 }

 print "The hostname is $hostname.\n";

=back

=head2 B<hosttype()>

This method allows you to retrieve the hosttype of the client machine

=over 4

=item Outputs

Returns a scalar value containing the hosttype.  If an error occured the error flag
is set, errmsg is set to the failure reason, and 0\'s are passed back as the return values.

=item Examples

 # Get hosttype
 $hosttype= $Obj->hosttype();
 if ($Obj->error){
     ...
 }

 print "The hosttype is $hosttype.\n";

=back

=head2 B<os_version()>

This method allows you to retrieve the os_version of the client machine

=over 4

=item Outputs

Returns a scalar value containing the os_version.  If an error occured the error flag
is set, errmsg is set to the failure reason, and 0\'s are passed back as the return values.

=item Examples

 # Get os version
 $os_ver= $Obj->os_version();
 if ($Obj->error){
     ...
 }

 print "The version is $os_ver.\n";

=back

=head2 B<timeout($sec)>

This method allows you to change the Telnet timeout.  The default value
for timeout is the value specified in the Tharnhost file.

=over 4

=item Inputs

=item * $sec

Specify number of seconds for timeout value.

=item Outputs

If successful, the timeout in the Telnet object is set to the specified
value and the previously set value is returned.  If an error occurred,
the error flag is set, errmsg is set to the failure reason returned
from the Telnet module and 0 is passed back as the return value.

=item Examples

 # Set Telnet timeout value to 45 seconds.
 $prev = $obj->timeout(45);
 <check for errors>

=back

=head2 B<destroy()>

Destroy client object (disconnect from client).

=over 4

=item Examples

 # Get os version
 $Obj->destroy();

=back

=head2 B<reconnect()>

This method allows you to reconnect to the client machine

=over 4

=item Outputs

Returns 1 on success. On failure, 0 will be set and the objects error flag will be set.

=item Examples

 # Reconnect to filer
 $Obj->reconnect();
 if ($Obj->error){
     ...
 }

=back

=head2 B<get_testbin_path()>

This method allows you to retrieve the test bin directory of the client machine

=over 4

=item Inputs

None

=item Outputs

Returns a string with full path for the "test bin" directory. If an error occurred the error flag is set, errmsg is set to the failure reason.

=item Examples

 my $testpath = $client_obj->get_testbin_path();

 if ($client_obj->error) {
     my $errmsg = $client_obj->errmsg();
     return 0;
 }

 print "The test bin dir is $testpath.\n";

=back

=head2 B<rotate_generic_log()>

Rotate a generic logfile.

=over 4

=item Inputs

=item * -file

String containing the name of the file to rotate.

=item * -minsize

Minimum filesize of the the log in bytes before it will be rotated.

=item * -keep

The number of rotated copies to keep.  Logs will be named .0 -> .<keep-1>

=item Outputs

Returns $ClientUtil::TRUE on success.  If an error occurs, the error flag is set, the errmsg is set to the failure
reason and $ClientUtil::FALSE is returned.

=item Example

# Rotate the log /etc/vx/dmpevents.log
$status = $self->rotate_generic_log(-file => "/etc/vx/dmpevents.log", -minsize => "1024", -keep => "5");
if ($status == $ClientUtil::TRUE) {
    return $ClientUtil::TRUE;
} else {
    if (not $self->error) {
        $self->_set_err("An error was received while attempting to rotate the Veritas DMP logs");
    }
    return $ClientUtil::FALSE;
}

=back

=head2 B<get_isns_server()>

Get iSNS Server object handle

=over 4

=item Inputs

none

=item Outputs

Returns the iSNS Server object. If an error occured the error flag
is set, returns an empty hash.

=item Examples

 # Get iSNS server object handle
 $isns_obj = $Client->get_isns_server();

=back

=head2 B<create_isns_server()>

Create iSNS server object handle

=over 4

=item Inputs

none

=item Outputs

Returns the iSNS Server object. If an error occured the error flag
is set, returns iSNS server object.

=item Examples

 # Create iSNS_Server
 $isns_server_obj = $Client->create_isns_server();

=back

=cut

use warnings qw(redefine);

1;
