# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provides common host methods
## @author knopp@netapp.com latif@netapp.com thomasm@netapp.com
## @status private
## @pod here

use strict;

# Package Name
##############
package ClientUtil;

# Package Requirements
######################
use POSIX;
use Config;
use Tharn;
use San::SanUtil;
use NATE::Log qw(log_global);

# Exported global variables
###########################
$ClientUtil::TRUE       = 1;
$ClientUtil::FALSE      = 0;
$ClientUtil::PER_CLIENT = "PC_";

my $Log = log_global();

###############################################################################
# Subroutine Name: execute
# Objective: Execute command on host and return output
#            On Windows, execute will be vectored to _execute_windows
###############################################################################
sub execute {
    my ($connection, $cmd, $timeout, %extra_args) = @_;

    my ($env_output, $output, @output, $status);
    my ($match, $before, $after, $sentinel, $default_timeout);
    my ($cmdlog);

    $default_timeout = "600";
    $cmdlog = 1;
    @output = ();

    if (!defined($cmd) or $cmd =~ /^\s*$/) {
        &SanUtil::nate_log('-warn', "Command passed to ClientUtil::execute was empty");
        $status = 1;
        return ($status, @output);
    }

    # set the default timeout if not specified
    $timeout = $default_timeout if (!defined($timeout) or $timeout =~ m/^\s*$/);

    # If we are on Windows, execute separate logic (from WinClient)
    if ($connection->{hostp}->{hosttype} =~ /windows/i) {
        ($status, @output) = &_execute_windows($connection, $cmd, $timeout);
        return($status, @output);
    }

    # If we are on HP make some settings on the shell that will turn of
    # of command processing in the SHELL and give true output.
    if (($connection->{hostp}->{hosttype} =~ /unix-hpux/) or
        ($connection->{hostp}->{hosttype} =~ /unix-aix/)) {
        &_set_term_setting($connection);
    }

    # Adding 2>&1 to the end of the command string for all UNIX plaforms
    # This is needed because sometimes the STDERR does not print before
    # the rstatus=#. This forces the STRERR to show up in the STDOUT
    if ($connection->{hostp}->{hosttype} =~ /unix|vmware/) {
        &Tharn::param("CLIENT_UTIL_EXECUTE_BEHAVIOR", "-default", 0);
        if (&Tharn::param("CLIENT_UTIL_EXECUTE_BEHAVIOR") == 1) {
            $cmd = $cmd . " 2>&1";
        } else {
            if ($extra_args{-dont_wrap_in_parens} == 1) {
                $cmd = $cmd . " 2>&1";
            } else {
                $cmd = "(" . $cmd . ")" . " 2>&1";
            }
        }
    }

    my %extra_say_args;

    if (exists($extra_args{-maxbufsize})) {
        $extra_say_args{-maxbufsize} = $extra_args{-maxbufsize};
    } else {
        # hardcode the maxbufsize so that output doesn't get overwritten for large output commands
        $extra_say_args{-maxbufsize} = "40000000";
    }

    # add to the command the echo of status to get that information from
    # the unix shell. Also add the sentinel string DoNe to command so that
    # we can match on that string.
    $sentinel = 'DoNe' . time();
    $cmd = $cmd . ";echo \"rstatus=\$\?\" ;echo $sentinel";

    # libssh/libpersay do not print out normal headers letting you know
    # command being executed. Add these.
    my $hostname = $connection->{hostp}->hostname();
    my $conntype = $connection->conntype();
    if ($conntype =~ /libssh/i) {
        &SanUtil::nate_log('-comment', ">>" . $hostname . " #" . $conntype . " CMD $cmd ");
    }

    # execute the passed command. Use just timeout since we check earlier
    # now if it was not passed and set the default if that is the case.
    ($match, $output, $after) =  &Tharn::say($connection, $cmd, $sentinel,
                                             '-swallow',
                                             '-timeout', $timeout,
                                             '-ignore', "echo $sentinel",
                                             '-comment', "Attempted '$cmd'",
                                             %extra_say_args);

    # strip output and remove CR/LF
    chomp($output);
    $output =~ s/\r//g;

    # remove first line
    $cmd =~ s/"rstatus=\$\?" ;echo $sentinel//g;
    $output =~ s/"rstatus=\$\?" ;//g;

    # strip sentinel and then status from output
    $output =~ s/\n$sentinel.*//;

    # strip any telnet ctrl characters that get
    # added to the output (this problem seems
    # to only occur with certain platforms).
    $output =~ s/ \cH//g;

    # strip out any lines containing "You have new mail in"
    $output =~ s/You have new mail in\S+\n//ig;

    # locate status.  Remember that commands
    # echo themselves so make sure you don't
    # get the command line
    ($status) = $output =~ /rstatus=(\d+)/;

    # handle case where rstatus is not in the output
    # stream. One case is when NATE times out the command
    # In case of filer, do not do this since rstatus only
    # exists for clients.
    if (not defined $status and
        $connection->{hostp}->{hosttype} =~ /unix|vmware/) {
        $status = 1;
    }

    # Check to see if we got a failure executing the command
    # and we have a persistant connection type (ssh/telnet).
    # If so, need to clear out connection and possible previous
    # command that may be running still
    if ($status != 0 and
        $output !~ /rstatus/i and
        $conntype =~ /ssh|telnet/i and
        $conntype !~ /persay/i) {
        # Had a failure while trying to execute command
        # Send <CTRL>-C to clear out running command.
        for (1..3) {
            $connection->send_intr();
        }

        # Execute some command on the connection to just clear
        # out any extra output from the buffer.
        for (1..3) {
            &Tharn::say($connection, 'echo CtRlC', "CtRlC",
                                     '-swallow',
                                     '-timeout', "30",
                                     '-ignore', "echo CtRlC",
                                     '-comment',
                                     "Attempted '<CTRL>-C'");
        }
    }

    $output =~ s/rstatus=\d+//;

    # create an array from the output
    @output = split(/\n/, $output);

    ## We need to determine if command is multiline
    ## and exists in the output.  IMPORTANT: Not all
    ## commands echo back their input so don't forget to
    ## check before modifying the output.
    my @cmds = split(/\n/, $cmd);

    ## If output contains second line of cmd
    ## then cmd was echoed to output and is
    ## multiline
    if (defined $cmds[1] and length $cmds[1]) {
        ## Remove continuation character and parens (from group commands
        ## added recently)
        $cmds[1] =~ s/\\//g;
        $cmds[1] =~ s/\)//g;
        $cmds[1] =~ s/\(//g;
        if ($output[0] =~ /$cmds[1]/) {
          ## determine size of command and strip
          ## from there.
          @output = @output[@cmds-1..@output-1];

        }
    }

    ## Go through the output and remove any line that has the command in it
    ## We do this, because none of the function that call execute expect
    ## to have to filter out their cmd from the returned output.
    my @output_sorted;
    foreach my $line (@output) {
        my $match = 0;
        foreach my $cmd (@cmds) {
            if ($cmd eq $line) {
                $match = 1;
                last;
            }
        }

        if ($match == 0) {
            push (@output_sorted, $line);
        }
    }

    @output = @output_sorted;

    # case with dt that run w/aios flag - error status from dt is not reliable
    # add this case to catch/compensate
    $status = 1 if ($output =~ /Error number \d+ occurred on/i);

    return($status, @output);

}

################################################################################
# Subroutine Name: execute_retry
# Objective:       This function will execute a command on a client (or filer)
#                  as many times at set by the $retry_count, if it fails it
#                  will fail the execute attempt. By default $retry_count is set
#                  to 6
# Inputs:          connection handle
#                  command to execute
#                  timeout (in seconds - default to ClientUil::execute's value)
#                  number of times to retry
#                  error type on failure (default is -warn)
#                  exptect return code (0 by default)
################################################################################
sub execute_retry {
    my ($client_connection, $command, $time_in_seconds_for_timeout,
        $retry_count, $err_type, $return_check) = @_;

    my ($status, @output);

    # if retry count is not specified, use 6
    if (not defined $retry_count or $retry_count eq "") {
        $retry_count = 6;
    }

    # if retry check is not specified, use 0
    if (not defined $return_check or $return_check eq "") {
        $return_check = 0;
    }

    # for the number of request retries, execute the command in eval block
    # to catch errors.
    while ($retry_count > 0) {
        eval {
            ($status, @output) = &ClientUtil::execute($client_connection,
                                                      $command,
                                                      $time_in_seconds_for_timeout);
            if ($status > $return_check && $status != 254) {
                &SanUtil::nate_log('-comment',
                                   "ClientUtil::execute_retry() Attempted command ($command) and failed with\n" .
                                   "status ($status = expected $return_check)\n" .
                                   "and output (@output)");
                $status = -2;
                @output = ();
            }
        };

        # check if command died abmormally and was caught
        if ($@) {
            &SanUtil::nate_log('-comment',
                               "ClientUtil::execute_retry() Command ($command) failed with\n" .
                               "output of ($@) and status ($status)\n" .
                               "(@output)");
            $status = -2;
            @output = ();
        }

        # if command completed, check status
        if (($status > $return_check or $status == -2) and $status != 254) {
            &SanUtil::nate_log('-comment',
                               "ClientUtil::execute_retry() Attempted command ($command) and failed with\n" .
                               "status ($status)\n" .
                               "and output (@output)\n" .
                               "Attempts left (" . ($retry_count - 1) . ")");
            $retry_count--;
            # Sleep some time before attempting to execute the command again
            &SanUtil::nate_sleep(10);
            next;
        } else {
            &SanUtil::nate_log('-comment',
                               "ClientUtil::execute_retry() Command ($command) executed properly with (" . ($retry_count - 1) . ") attempts left");
            return ($status, @output);
        }
    }

    # If we get here, we could not execute the command . . . return the last status and output
    if (defined $err_type and $err_type ne "") {
        &SanUtil::nate_log($err_type,
                           "ClientUtil::execute_retry() Unable to execute command ($command)");
    } else {
        &SanUtil::nate_log('-warn',
                           "ClientUtil::execute_retry() Unable to execute ($command)");
    }

    return ($status, @output);

}

###############################################################################
# Subroutine Name: remove_remote_file
# Objective:       delete file on remote machine
###############################################################################
sub remove_remote_file {
    my ($connection, $rem_file, $ignore_err) = @_;

    my (@output, $status);

    if (not defined($ignore_err)) {
        $ignore_err=0;
    }

    # Issue different remove command based on the client platform
    if ($connection->{hostp}->{hosttype} =~ /windows/i) {
        ($status, @output) = &execute($connection, "del $rem_file");
    }
    else{
        ($status, @output) = &execute($connection, "rm $rem_file");
    }

    if ($status != 0 and !$ignore_err) {
        &SanUtil::nate_log('-fatal', "Cannot remove remote file: $rem_file");
    }

}

###############################################################################
# Subroutine Name: create_remote_file
# Objective:       create file on remote machine
###############################################################################
sub create_remote_file {
    my ($connection, $rem_file, @filebuf) = @_;

    my (@output, $status, $line);

    if (not defined($rem_file)) {
        &SanUtil::nate_log('-fatal', "Remote file name not specified.");
    }

    &execute($connection, "touch $rem_file");
    my $i = 0;
    foreach $line (@filebuf) {
        my $oper;
        if ($i == 0) {
            # trucate the file on the first echo
            $oper = '>';
        } else {
            # append for the rest
            $oper = '>>';
        }

        $line =~ s/\$/\\\$/g;
        $line =~ s/\"/\\\"/g;
        # need to handle Windows differently, it echo's the double
        # quotes on the string and UNIX does not.
        if ($connection->{hostp}->{hosttype} =~ /windows/i) {
            ($status, @output) = &execute($connection,
                                          "echo $line $oper $rem_file");
        } else {
            ($status, @output) = &execute($connection,
                                          "echo \"$line\" $oper $rem_file");
        }
        if ($status != 0) {
            &SanUtil::nate_log('-fatal', "Cannot populate remote file: $rem_file");
        }
        $i++;
    }

    return($ClientUtill::TRUE);

}

###############################################################################
# Subroutine Name: _set_term_setting
# Objective:       This routine will setup the HP and AIX POSIX shell (sh) so
#                  that it does not do processing of commands in terms of
#                  buffering and logical command placement.
###############################################################################
sub _set_term_setting {
    my ($connection) = @_;

    # Check to see if the connection type is rsh, if so DON'T run this
    if ($connection->{id} =~ /\#rsh/i) {
        return($ClientUtill::TRUE);
    }

    my ($cmd, $sentinel);
    my ($match, $before, $output);

    # set COLUMN width to 2048 to remove the command line processing
    # in HP's POSiX SHELL.
    $cmd = "COLUMNS=2048;export COLUMNS;unset EDITOR";

    # execute the command
    $sentinel = 'TeRm_CoMpLeTe';
    $cmd = $cmd . " 2>&1";
    $cmd = $cmd . ";echo $sentinel";

    # do command in eval block not to kill tests.
    eval {
        ($match, $before, $output) = &Tharn::say($connection, $cmd, $sentinel,
                                                 '-swallow',
                                                 '-timeout', 30);
    };

    return($ClientUtill::TRUE);
}

################################################################################
# Subroutine Name: convert_to_bytes()
# Objective:       convert size strings to bytes of unit k,m,g,t.
# Details:         See POD documentation at the end of this file
################################################################################
sub convert_to_bytes {
    my ($size, $unit_to) = @_;

    my ($bytes, $nbr, $unit_from);

    # if nothing was passed return 0 (bytes)
    if (not defined($size)) {
        return (0);
    }

   # parse out the number and the unit, this will handle integers and non-integers
    ($nbr, $unit_from) = $size =~ /(\d+\.?\d*)(\w*)/;

    if ($unit_from eq $unit_to) {
        return $nbr;
    }

    if ($unit_from eq "") {
        # if $unit is null, assume $size is in bytes
        $bytes = $nbr;
    } elsif ($unit_from =~ /k/i) {
        $bytes = $nbr * 1024;
    } elsif ($unit_from =~ /m/i) {
        $bytes = $nbr * 1024 * 1024;
    } elsif ($unit_from =~ /g/i) {
        $bytes = $nbr * 1024 * 1024 * 1024;
    } elsif ($unit_from =~ /t/i) {
        $bytes = $nbr * 1024 * 1024 * 1024 * 1024;
    }

    if ($unit_to eq "") {
        # if $unit_to is null, assume $size is converted to bytes
    } elsif ($unit_to =~ /k/i) {
        $bytes = $bytes / ( 1024 );
    } elsif ($unit_to =~ /m/i) {
        $bytes = $bytes / ( 1024 * 1024 );
    } elsif ($unit_to =~ /g/i) {
        $bytes = $bytes / ( 1024 * 1024 * 1024 );
    } elsif ($unit_to =~ /t/i) {
        $bytes = $bytes / ( 1024 * 1024 * 1024 * 1024 );
    }

    return $bytes;

}

################################################################################
# Subroutine Name: convert_to_seconds()
# Objective:       convert size strings to bytes.
# Details:         See POD documentation at the end of this file
################################################################################
sub convert_to_seconds {
    my ($time_to_convert) = @_;

    my $converted_time = 0;

    # Time is already in seconds, return it back
    if ($time_to_convert =~ /^\d+$/) {
        return ($time_to_convert);
    }

    # Handle some cases of bad user input
    $time_to_convert =~ s/\s+//g;
    $time_to_convert =~ s/([a-zA-Z]+)(\d+)/$1 $2/g;

    # Foreach piece (e.g. 10h4m6s) calculate the time
    foreach my $chunk (split(/\s+/, $time_to_convert)) {

        # Convert years
        if ($chunk =~ /(\d+)Y/) {
            $converted_time += $1 * 365 * 24 * 60 * 60;
            # Convert months
        } elsif ($chunk =~ /(\d+)M/) {
            $converted_time += $1 * 30 * 24 * 60 * 60;
            # Convert days
        } elsif ($chunk =~ /(\d+)d/) {
            $converted_time += $1 * 24 * 60 * 60;
            # Convert hours
        } elsif ($chunk =~ /(\d+)h/) {
            $converted_time += $1 * 60 * 60;
            # Convert minutes
        } elsif ($chunk =~ /(\d+)m/) {
            $converted_time += $1 * 60;
            # Convert seconds
        } elsif ($chunk =~ /(\d+)s/) {
            $converted_time += $1;
        } else {
            # Error unsuported time unit specified
            return (undef);
        }
    }

    return ($converted_time);

}

################################################################################
# Subroutine Name: _execute_windows
# Objective:       Windows specific execute method
# Details:         See POD documentation at the end of this file
################################################################################
sub _execute_windows {
    my ($connection, $cmd, $timeout, $allowTimeout) = @_;

    my ($status, $output, @output, $after);
    my ($sentinel, $match, $tmp);

    # if the connection is via telnet on windows handle execution differently
    my $conntype = $connection->conntype();
    if ($conntype =~ /telnet/i) {
        ($status, @output) = &_execute_windows_telnet($connection, $cmd, $timeout);
        return ($status, @output);
    }

    # set initial timeOut to 0
    my $timeOut = 0;

    # allow timeout by default, if not overriden
    $allowTimeout = 0 if (!defined($allowTimeout));

    # append status at the end of command and prepend command.com
    # re-direct STDOUT and STDERR to same channel
    $cmd = "cmd.exe /c \"$cmd 2>\&1\" & echo rstatus=\%ERRORLEVEL\%";

    # execute the command
    ($match, $output, $after) = &Tharn::say($connection,
                                            $cmd,
                                            'rstatus=\d+',
                                            '-timeout', $timeout,
                                            "-comment", "Attempted '$cmd'");

    if (defined($match)) {
        ($status) = ($match =~ /rstatus=(\d+)/);
    } else {
        if ($allowTimeout) {
            $timeOut = 1;
        } else {
            $status = 1;
            push(@output, "Command did not complete: $cmd");
            goto execute_EXIT;
        }
    }

    # Check to make sure that $status is not an empty string, this could
    # happen if there is NO rstatus in the return string because the
    # connection hung and timed out.
    if ($status eq "") {
        $status = 1;
        push(@output, "Command did not complete: $cmd");
        goto execute_EXIT;
    }

    # remove the first line if it is an echo of the input
    $output =~ s/.*rstatus=\%ERRORLEVEL\%\"?\r*\n//;

    # strip output and remove CR/LF
    chomp($output);
    $output =~ s/\r//g;

    # TODO: move the io specific errors to new sections below
    # parse the output and look for standard error messages
    # return 1 (shell bad status) as opposed to ClientUtil::FALSE (0)
    $status = 1 if ($output =~ /is not recognized as an/i or
                    $output =~ /no such file/i            or
                    $output =~ /cannot find the file/i    or
                    $output =~ /cannot find the path/i    or
                    $output =~ /aborting child process/i  or
                    $output =~ /STATUS_ACCESS_VIOLATION/i or
                    $output =~ /sio:.*error/i             or
                    $output =~ /error accessing file/i    or
                    $output =~ /data compare error/i      or
                    $output =~ /error number/i            or
                    $output =~ /permission denied/i       or
                    $output =~ /file not found/i          or
                    $output =~ /syntax is incorrect/i     or
                    $output =~ /path is invalid/i         or
                    $output =~ /cannot open volume/i      or
                    $output =~ /could not assign/i        or
                    $output =~ /access is denied/i        or
                    $output =~ /did not see expected waitstrings/i or
                    $output =~ /Windows DT: The type of error/i or
                    $output =~ /exiting because of signal 11/i or
                    $output =~ /errno \=/i or
                    $output =~ /failed with error/i or
                    $output =~ /The system cannot find the drive specified/i or
                    $output =~ /Failed\s+to\s+complete\s+the\s+operation/i  or
                    $output =~ /Trigger exited with status 1\!/i or
                    $output =~ /Terminating\.\.\./i or
                    $output =~ /The device is not ready/i or
                    $output =~ /is no longer available/i or
                    $output =~ /The device is not connected/i or
                    $output =~ /Invalid option.*specified/i or
                    $output =~ /TerminateProcess\(\)/i);

    # dsmtest errors
    $status = 1 if ($output =~ /^ERROR:/i);

    # create an array of the output
    @output = split(/\n/, $output);

    if ($timeOut or !defined($status)) {
        $status = 1;
        push(@output, "Command did not complete: $cmd");
    }

execute_EXIT:

    return ($status, @output);

} # END _execute_windows

################################################################################
# Subroutine Name: _execute_windows_telnet
# Objective:       Windows specific execute method w/telnet connection due to
#                  ERRORLEVEL not being accurate over that connection
# Details:         See POD documentation at the end of this file
################################################################################
sub _execute_windows_telnet {
    my ($connection, $cmd, $timeout) = @_;

    my ($status, $output, @output, $sentinel);
    my ($match, $after);

    # re-direct STDOUT and STDERR to same channel
    $cmd = "$cmd 2>\&1";

    # Add the sentinel string DoNe to command so that we can match on that string.
    $sentinel = 'DoNe' . time();
    $cmd = $cmd . " \& echo $sentinel";

    $status = 0;
    # execute the passed command.
    ($match, $output, $after) =  &Tharn::say($connection, $cmd, $sentinel,
                                             '-swallow',
                                             '-timeout', $timeout,
                                             '-ignore', " \& echo $sentinel",
                                             '-comment', "Attempted '$cmd'");

    # strip output and remove CR/LF
    chomp($output);
    $output =~ s/\r//g;

    # remove first line (cmd)
    $cmd =~ s/\s+\&\s+echo $sentinel//;
    $cmd = quotemeta($cmd);
    $output =~ s/$cmd//;
    $output =~ s/^\s+//;

    # strip sentinel and then status from output
    $output =~ s/\n$sentinel.*//;

    # TODO: move the io specific errors to new sections below
    # parse the output and look for standard error messages
    # return 1 (shell bad status) as opposed to ClientUtil::FALSE (0)
    $status = 1 if ($output =~ /is not recognized as an/i or
                    $output =~ /no such file/i            or
                    $output =~ /cannot find the file/i    or
                    $output =~ /cannot find the path/i    or
                    $output =~ /aborting child process/i  or
                    $output =~ /STATUS_ACCESS_VIOLATION/i or
                    $output =~ /sio:.*error/i             or
                    $output =~ /error accessing file/i    or
                    $output =~ /data compare error/i      or
                    $output =~ /error number/i            or
                    $output =~ /permission denied/i       or
                    $output =~ /file not found/i          or
                    $output =~ /syntax is incorrect/i     or
                    $output =~ /path is invalid/i         or
                    $output =~ /cannot open volume/i      or
                    $output =~ /could not assign/i        or
                    $output =~ /access is denied/i        or
                    $output =~ /did not see expected waitstrings/i or
                    $output =~ /Windows DT: The type of error/i or
                    $output =~ /exiting because of signal 11/i or
                    $output =~ /errno \=/i or
                    $output =~ /failed with error/i or
                    $output =~ /The system cannot find the drive specified/i or
                    $output =~ /Failed\s+to\s+complete\s+the\s+operation/i  or
                    $output =~ /Trigger exited with status 1\!/i or
                    $output =~ /Terminating\.\.\./i or
                    $output =~ /The device is not ready/i or
                    $output =~ /is no longer available/i or
                    $output =~ /The device is not connected/i or
                    $output =~ /Invalid option.*specified/i or
                    $output =~ /Access is denied/i or
                    $output =~ /error occurred:/i or
                    $output =~ /TerminateProcess\(\)/i);

    # dsmtest errors
    $status = 1 if ($output =~ /^ERROR:/i);

    # create an array of the output
    @output = split(/\n/, $output);

    if (!defined($status)) {
        $status = 1;
        push(@output, "Command did not complete: $cmd");
    }

    return ($status, @output);

} # END _execute_windows_telnet

################################################################################
# Subroutine Name: load_client_params
# Objective:       This will load individual client params into the NTEST
#                  Environment by accessing the tharnhost file of the client
#                  passed into the function.
# Details:         See POD documentation at the end of this file
################################################################################
sub load_client_params {
    my ($client, $parse_flag, $quiet) = @_;

    if (not defined($quiet)) {
        $quiet = 1;
    }

    # check if this option is enabled from the calling THPL
    my $enable_pcp = &Tharn::param("ENABLE_PER_CLIENT_PARAMS", "-default", 0);
    if ($enable_pcp == 0) {
        if ($quiet) {
            &SanUtil::nate_log('-comment', "ENABLE_PER_CLIENT_PARAMS was not set.");
        }
        return;
    }

    # obtain the clients tharnhost hash
    my $tharnhost_hsh = host($client);
    my $change_count = 0;
    my $old_environment_hshref;
    my (@output_results) = ();

    if ($parse_flag eq "") {
        $parse_flag = $ClientUtil::PER_CLIENT;
    }

    # iterate over each param in the tharnhost file that is prepended by $parse_flag
    foreach my $param (sort keys %{$tharnhost_hsh}) {
        if ($param =~ /^$parse_flag(\S+)/) {
            my ($param_name) = $1;

            # if this parameter is already defined in the environment,
            # then override it with the value in the tharnhost file.
            if (defined &Tharn::param("$param_name")) {

                # first store the original value in a hash
                $old_environment_hshref->{$param_name} = &Tharn::param("$param_name");

                # update the environment with the new value
                &Tharn::param("$param_name", -set, $tharnhost_hsh->{$param});

                $change_count++;
                push(@output_results, "$param_name=" . $tharnhost_hsh->{$param} . "\n");
            }
        }
    }

    if ($change_count > 0 and $quiet) {
        # chomp the last new line in the array.
        chomp($output_results[$#output_results]);

        &SanUtil::nate_log('-comment', "ENABLE_PER_CLIENT_PARAMS is set.\n" .
                           "The following environment parameters will be overridden ".
                           "for this client: '$client'\n @output_results");
    }

    # return the old environment to the caller.
    return $old_environment_hshref;
}

################################################################################
# Subroutine Name: read_client_params
# Objective:       This will read individual client params and return a hash
#                  containing the read parameters
# Details:         See POD documentation at the end of this file
################################################################################
sub read_client_params {
    my ($client, $parse_flag) = @_;

    # obtain the clients tharnhost hash
    my $tharnhost_hsh = host($client);
    my $environment_hshref;

    if ($parse_flag eq "") {
        $parse_flag = $ClientUtil::PER_CLIENT;
    }

    # iterate over each param in the tharnhost file that is prepended by $parse_flag
    foreach my $param (sort keys %{$tharnhost_hsh}) {
        if ($param =~ /^$parse_flag(\S+)/) {
            my ($param_name) = $1;

            # if this parameter is already defined in the environment,
            # then override it with the value in the tharnhost file.
            if (defined &Tharn::param("$param_name")) {

                # first store the original value in a hash
                $environment_hshref->{$param_name} = $tharnhost_hsh->{$param};
            }
        }
    }

    # return the new environment to the caller.
    return $environment_hshref;
}

################################################################################
# Subroutine Name: restore_original_params
# Objective:       This will restore the environment parameter values of all
#                  the keys in the $old_environment_hshref variable
# Details:         See POD documentation at the end of this file
################################################################################
sub restore_original_params {
    my ($old_environment_hshref) = @_;

    # iterate through each parameter and restore the value in the environment.
    foreach my $param_name (sort keys %{$old_environment_hshref}) {
        my $param_value = $old_environment_hshref->{$param_name};
        &Tharn::param("$param_name", -set, $param_value);
    }

    return;
}

################################################################################
# Subroutine Name: kill_process_instances
# Objective:       Kill all instances of the passed program name on the
#                  specified host
################################################################################
sub kill_process_instances {
    my ($client, $exe_to_kill, $level, $signal) = @_;

    my ($connection, $status, @output);
    my ($ps_line, $kill_line, $loop_retry);
    my ($retries, $num_retries, $done, $search_string);
    my (@ps_output, @kill_output, $rsh_failures);
    my (@ver_output, $cmd, $kill_cmd);

    # set default level
    if (not defined($level) or $level eq "") {
        $level = '-fatal';
    }

    # set default for signal
    if (not defined($signal) or $signal eq "") {
        $signal = "9";
    }

    my $host_rec = &Tharn::host($client);
    my $h_type = $host_rec->{hosttype};
    my $client_conn = &ClientUtil::get_device_connection($client);

    # skip this method for ESX hosts, there are no processes to kill on ESX hosts
    if ($h_type =~ /linux\-vmware/) {
        &SanUtil::nate_log('-comment',
                           "ClienUtil::kill_process_instances() Skipping host (because it is ESX)");
        return $ClientUtil::TRUE;
    }

    eval {
        if ($client_conn !~ /rsh/) {
            $connection = &Tharn::connect($client . $client_conn . ";inherit=0");
        } else {
            $connection = &Tharn::connect($client . $client_conn);
        }
    };

    if ($@) {
        &SanUtil::nate_log('-warn',
                           "ClienUtil::kill_program_instances() Client $client had a failure attempting to connect to it using connection type of " .
                           $client_conn);
        return $ClientUtil::FALSE;
    }

    &SanUtil::nate_log('-comment',
                       "ClienUtil::kill_program_instances() Starting process to kill all instances of ($exe_to_kill) on the host");

    if ($h_type =~ /windows/) {
        # append .exe to program
        if ($exe_to_kill !~ /\.exe/i) {
            $exe_to_kill = $exe_to_kill . ".exe";
        }

        # determine version of windows
        my $cmd = "ver";
        ($status, @ver_output) = &ClientUtil::execute($connection,
                                                      $cmd, 20);

        # On win2k, taskkill does not exist, if output from taskkill
        # shows program is not found, try to use pskill utility from
        # the X: (MinWin) drive. string string is set to look for the
        # tool specific string that says there are not more processes
        # to kill by that name.
        if (grep(/Windows 2000/i, @ver_output) > 0) {
            # check if this is a pNATE environment
            if (&ClientUtil::_is_PNATE_ENVIRONMENT()) {
                $kill_cmd = "C:\\NETAPP\\pskill /accepteula $exe_to_kill";
            } else {
                $kill_cmd = "X:\\SAN\\bin\\win2k\\pskill /accepteula $exe_to_kill";
            }
            $search_string = "Process does not exist";
        } else {
            $kill_cmd = "taskkill /f /im $exe_to_kill";
            $search_string = "not found";
        }

        # try to kill (first time to perform initial kill)
        ($status, @output) = &ClientUtil::execute($connection,
                                                  $kill_cmd,
                                                  20);

        # try to kill again (and do this in a loop until there are no more
        # processes running or retry count has been hit).
        $retries = 0;
        $num_retries = 100;
        $rsh_failures = 0;
        $done = 0;
        while ($retries < $num_retries and $done == 0) {
            if (grep(/$search_string/i, @output) == 1) {
                $done = 1;
                last;
            }

            # sleep to allows kills to catch up.
            &SanUtil::nate_sleep(4);
            ($status, @output) = &ClientUtil::execute($connection,
                                                      $kill_cmd,
                                                      20);

            # check status, if we have 3 failures to RSH to box (meaning
            # status comes back as non-zero), then don't continue to try.
            if ($status != 0) {
                 $rsh_failures++;
                 if ($rsh_failures == 3) {
                     &SanUtil::nate_log('-comment',
                                        "ClienUtil::kill_program_instances() (3) failures to connect via RSH to Windows box, stopping attempts to kill I/O");
                     $retries = $num_retries;
                 }
            } else {
                 $rsh_failures = 0;
            }

            $retries++;
        }
    } else {
        # get the list of processes to kill

        # check if this is a pNATE environment (don't want to look for localtest)
        if (&ClientUtil::_is_PNATE_ENVIRONMENT()) {
            $ps_line = "ps -eaf | grep -v grep | grep -i NETAPP | grep $exe_to_kill | awk '{print \$2}'";
        } else {
            $ps_line = "ps -eaf | grep -v grep | grep localtest | grep $exe_to_kill | awk '{print \$2}'";
        }
        ($status, @ps_output) = &ClientUtil::execute($connection,
                                                     $ps_line, 20);

        # try to kill again (and do this in a loop until there are no more
        # processes running or retry count has been hit).
        $retries = 0;
        $num_retries = 50;
        $done = 0;
        while ($retries < $num_retries and $done == 0) {
            $loop_retry = 0;
            foreach my $line (@ps_output) {
                if ($line =~ /^\d+$/ and $line !~ /^1$/) {
                    $kill_line = "kill -" . $signal . " $line";
                } else {
                    &SanUtil::nate_log('-comment',
                                       "ClienUtil::kill_program_instances() Illegal process ID to kill ($line)");
                    next;
                }

                ($status, @kill_output) = &ClientUtil::execute($connection,
                                                               $kill_line, 20);
                if ($status > 0) {
                    &SanUtil::nate_log('-comment',
                                       "ClienUtil::kill_program_instances() Could not kill process ID ($line)");
                    $loop_retry = 1;
                } else {
                    &SanUtil::nate_log('-comment',
                                       "ClienUtil::kill_program_instances(): Killed process ID ($line)");
                }
            }

            # If we failed to kiill any thread in list of processes, we
            # want to retry to the whole loop again. If inner loop killed
            # everything, set done and finish.
            if ($loop_retry == 1) {
                $retries++;

                &SanUtil::nate_sleep(2);

                # check if this is a pNATE environment (don't want to look for localtest)
                if (&ClientUtil::_is_PNATE_ENVIRONMENT()) {
                    $ps_line = "ps -eaf | grep -v grep | grep -i NETAPP | grep $exe_to_kill | awk '{print \$2}'";
                } else {
                    $ps_line = "ps -eaf | grep -v grep | grep localtest | grep $exe_to_kill | awk '{print \$2}'";
                }
                ($status, @ps_output) = &ClientUtil::execute($connection,
                                                             $ps_line, 20);
            } elsif ($loop_retry == 0) {
                $done = 1;
            }
        }
    }

    if ($retries == $num_retries) {
        &SanUtil::nate_log($level,
                           "ClienUtil::kill_program_instances() Error: Problems running Host I/O process/taskkill script\n" .
                           "Not all $exe_to_kill processes could be killed after ($num_retries) retries");
    } else {
        &SanUtil::nate_log('-comment', "ClienUtil::kill_program_instances() Successfully ran Host process/taskkill script\n" .
                           "on host $client for program ($exe_to_kill)\n" .
                           "Took ($retries) calls to rocess/taskkill script to kill all instances");
    }

    &Tharn::disconnect($connection);

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: count_process_instances
# Objective:       returns back the number of instances of an exe program that
#                  is currently running on the $client
################################################################################
sub count_process_instances {
    my ($client, $exe_to_get, $timeout, $connection) = @_;

    my ($status, $ps_cmd, @ps_output);
    my $host_rec = &Tharn::host($client);
    my $h_type = $host_rec->{hosttype};

    # default the timeout value
    $timeout = 60 if (not defined $timeout);

    if ($h_type =~ /windows/) {
        # WINDOWS
        # remove .exe from program name
        $exe_to_get =~ s/\.exe$//;
        # remove version numbers from dt
        $exe_to_get =~ s/^dt\..+/dt/;

        # check if this is a pNATE environment
        if (&ClientUtil::_is_PNATE_ENVIRONMENT()) {
            $ps_cmd = "C:\\NETAPP\\pslist /accepteula";
        } else {
            $ps_cmd = "X:\\SAN\\bin\\win2k\\pslist /accepteula";
        }

        # run pslist
        ($status, @ps_output) = &ClientUtil::execute($connection,
                                                     $ps_cmd,
                                                     $timeout);
        if ($status > 0 or (grep(/performance object not found/, @ps_output) > 0)) {
            &SanUtil::nate_log('-comment',
                               "ClienUtil::count_process_instances() Could not execute '$ps_cmd'\n" .
                               "Attempting to run lodctr /r to correct the issue");

            # Run lodctr /r for the user, don't bother to check results (pslist working will
            # be verification)
            ($status, @ps_output) = &ClientUtil::execute($connection, "lodctr /r", $timeout);

            # Run PS again
            ($status, @ps_output) = &ClientUtil::execute($connection,
                                                         $ps_cmd,
                                                         $timeout);
            if ($status > 0 or (grep(/performance object not found/, @ps_output) > 0)) {
                &SanUtil::nate_log('-comment',
                                   "ClienUtil::count_process_instances() Could not execute '$ps_cmd'\n" .
                                   "even after running lodctr /r");
                return -1;
            }
        }

        # search for the instances of $exe_to_get
        @ps_output = grep(/^$exe_to_get /, @ps_output);

    } else {
        # We need to check whether the connetion type is telnet. Because, output of the below $ps_cmd running on client
        # getting truncated when using connection type as telnet
        my $conntype = $connection->conntype();

        # UNIX
        # check if this is a pNATE environment (don't want to look for localtest)
        if (&ClientUtil::_is_PNATE_ENVIRONMENT()) {
            if ($conntype =~ /telnet/i) {
                # Output of the cmd running on client getting truncated when using connection type as telnet,
                # hence exporting 512 columns
                $ps_cmd = "export COLUMNS=512; ps -eaf | grep -v grep | grep -i NETAPP";
            } else {
                $ps_cmd = "ps -eaf | grep -v grep | grep -i NETAPP";
            }
        } else {
            if ($conntype =~ /telnet/i) {
                # Output of the cmd running on client getting truncated when using connection type as telnet,
                # hence exporting 512 columns
                $ps_cmd = "export COLUMNS=512; ps -eaf | grep -v grep | grep localtest";
            } else {
                $ps_cmd = "ps -eaf | grep -v grep | grep localtest";
            }
        }
        ($status, @ps_output) = &ClientUtil::execute_retry($connection,
                                                           $ps_cmd,
                                                           $timeout, 2);
        if ($status != 0) {
            &SanUtil::nate_log('-comment',
                               "ClienUtil::count_process_instances() Could not execute '$ps_cmd'");
            return -1;
        }

        # search for the instances of $exe_to_get
        @ps_output = grep(/\/$exe_to_get /, @ps_output);

    }

    # return the number of instances that are running
    return scalar(@ps_output);
}

################################################################################
# Subroutine Name: get_device_connection
# Objective:       Given a device, this function can return the proper
#                  connection type from the tharnhosts file
################################################################################
sub get_device_connection {
    my ($device, $def_connection) = @_;

    my $connection;

    # Get the device data handle from the host function
    my $ddh = &Tharn::host($device);
    my $host_type = $ddh->{hosttype};

    # Check to see if any connection types are defined, if not return the default value
    if (!defined($ddh->{connection_fields})) {
        if ($host_type =~ /windows/i) {
            $connection = "#rsh";
        } else {
            $connection = $def_connection;
        }
    } elsif (defined($ddh->{connection_fields}->{rsh})) {
        $connection = "#rsh";
    } elsif (defined($ddh->{connection_fields}->{"rsh-socket"})) {
        $connection = "#rsh-socket";
    } elsif (defined($ddh->{connection_fields}->{"libssh-persay"})) {
        $connection = "#libssh-persay";
    } elsif (defined($ddh->{connection_fields}->{"staf-persay"})) {
        $connection = "#staf-persay";
    } elsif (defined($ddh->{connection_fields}->{telnet})) {
        $connection = "#telnet";
    } elsif (defined($ddh->{connection_fields}->{ssh})) {
        $connection = "#ssh";
    } elsif (defined($ddh->{connection_fields}->{rlogin})) {
        $connection = "#rlogin";
    } elsif (defined($ddh->{connection_fields}->{console})) {
        $connection = "#console";
    }
    # Problem, unable to determine the connection type from the tharnhosts file
    # See if this is windows since testbed_setup sometimes does not
    # creation the conn for RSH. Else we should
    # Return an undefined value, this will cause a failure, as needed
    else {
        if ($host_type =~ /windows/i) {
            $connection = "#rsh";
        } else {
            $connection = undef;
        }
    }

    return $connection;
}

################################################################################
# Subroutine Name: convert_to_mb
# Objective:       Given a size (k,m,t) convert that value to MB
################################################################################
sub convert_to_mb {
    my ($size) = @_;

    my ($nbr, $unit, $mb);

    # if nothing was passed return 0 (bytes)
    if (not defined($size)) {
        return (0);
    }

    #  If no units, just return.
    if ($size =~ /^\d+$/) {
        return($size);
    }

    ($nbr, $unit) = $size =~ /(\d+)(\w+)/;

    if ($unit =~ /k/i) {
        $mb = int($nbr / 1024);
        if ($mb <= 0) {
            $mb = 1;
        }
    } elsif ($unit =~ /m/i) {
        $mb = $nbr;
    } elsif ($unit =~ /g/i) {
        $mb = $nbr * 1024;
    } elsif ($unit =~ /t/i) {
        $mb = $nbr * 1024 * 1024;
    }

    return ($mb);

}

################################################################################
# Subroutine Name: set_pnate_environment
# Objective:       Sets a environment for pNATE
################################################################################
sub set_pnate_environment {
    my ($quiet) = @_;

    # check if this option is enabled from the calling THPL
    if (not &ClientUtil::_is_PNATE_ENVIRONMENT()) {
        if (!$quiet) {
            &SanUtil::nate_log('-comment', "PNATE_ENVIRONMENT was not set.");
        }
        return $ClientUtil::TRUE;
    }

    if ($quiet eq "") {
        $quiet = 1;
    }

    my ($pnate_location, $pnate_version) = ("/opt/NETAPP/PNATE/p4", "");

    # Override params hash, simple key value pair.
    # Place params that need to be changed for pNATE environment
    # within this hash.
    my %override = ("TEST_BIN_DIR"                  => "/opt/NETAPP/bin/",
                    "P4_DIR"                        => "/opt/NETAPP/PNATE/p4/bin/",
                    "SLA_MEASUREMENT"               => "0",
                    "IO_DUMP_FILER_CONFIG_ON_ERROR" => "0",
                    "COLLECT_PERFSTATS"             => "0",
                   );

    foreach my $param (sort(keys(%override))) {
        # If the param is defined in the environment, set its value to requested
        # in override hash.
        if (defined &Tharn::param("$param")) {
            &Tharn::param("$param", -set, $override{$param});
            if (!$quiet) {
                &SanUtil::nate_log('-comment',
                                   "PNATE_ENVIRONMENT parameter override ($param) is set to " . $override{$param});
            }
        }
    }

    $pnate_version = `ls $pnate_location/.version_build_*`;
    chomp($pnate_version);
    $pnate_version =~ s/^.+\.version_build_//;

    if ($pnate_version ne "" and $pnate_version !~ /^\s+$/) {
        &SanUtil::nate_log('-comment', "*************** PNATE version currently running: ($pnate_version) ***************");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: randomize_array
# Objective:       Randomizes an array in place (fisher yates shuffle)
################################################################################
sub randomize_array {
    my ($array) = @_;
    my ($i);

    for ($i = scalar(@{$array}) - 1; $i >= 0; $i--) {
        # generate a random index between [0, $i+1)
        my $j = int rand ($i + 1);

        # dont perform the swap if they are the same
        next if ($i == $j);

        # swap the element in $i with the element in $j
        # this will update the array reference in place.
        @$array[$i, $j] = @$array[$j, $i];
    }
}

###############################################################################
# Subroutine Name: log
# Objective:       log cmds issued to client and filer (deprecated)
#                  Used to write to sanaudit log which was removed on agreement
#                  of team. Still need entry point not to break others.
###############################################################################
sub log {
  my (@line) = @_;

  return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: remote_file_exists
# Objective:       returns true if a file exists on the client
################################################################################
sub remote_file_exists {
    my ($connection, $rem_file) = @_;

    my (@output, $status, $line, $cmd);

    if (not defined($connection)) {
        &SanUtil::nate_log('-fatal', "Connection information not specified.");
        return $ClientUtil::FALSE;
    }

    if (not defined($rem_file)) {
        &SanUtil::nate_log('-fatal', "Remote file name not specified.");
        return $ClientUtil::FALSE;
    }

    if ($connection->{hostp}->{hosttype} =~ /windows/i) {
        $cmd = "dir " . $rem_file;
    } else {
        $cmd = "/bin/ls " . $rem_file;
    }

    ($status, @output) = &execute($connection, $cmd);
    if ($status != 0) {
        return $ClientUtil::FALSE;
    }
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remote_get_filesize
# Objective:       returns size of remote file or undef if an error occurs
################################################################################
sub remote_get_filesize {
    my ($connection, $rem_file) = @_;

    my (@output, $status, $line, $cmd, $size, $grepfile);

    if (not defined($connection)) {
        &SanUtil::nate_log('-fatal', "Connection information not specified.");
        return undef;
    }

    if (not defined($rem_file)) {
        &SanUtil::nate_log('-fatal', "Remote file name not specified.");
        return undef;
    }

    if ($connection->{hostp}->{hosttype} =~ /windows/i) {
        $cmd = "dir " . $rem_file;

        # We need to define $grepfile so it doesn't include the directory
        # since that won't be displayed in the line with the file name.
        my @split = split(/\\/, $rem_file);
        $grepfile = pop(@split);
    } else {
        $cmd = "/bin/ls -l " . $rem_file;

        # We will set $grepfile to $rem_file since the "ls" will properly
        # display the directory on the same output line.
        $grepfile = $rem_file;
    }

    ($status, @output) = &ClientUtil::execute($connection, $cmd);
    if ($status > 0) {
        &SanUtil::nate_log('-fatal', "($cmd) returned a bad status ($status)");
        return undef;
    }

    # Try to determine the filesize
    my @lsout = grep(/$grepfile/, @output);
    if (scalar(@lsout) == 1) {
        # We should only have 1 line in the output since we are pulling it
        # out with the grep statement above.
        if ($connection->{hostp}->{hosttype} =~ /windows/i) {
            # For windows systems, the output of dir should look like this:
            #
            # 06/10/2009  05:42 PM                24 autoexec.bat
            #
            # We want index 3 when indexing starting at 0
            $size = (split(/\s+/, $lsout[0]))[3];
        } else {
            # For non-windows systems, the output of ls -l should look like this:
            #
            # -rw-r--r--   1 root     sys          968 Oct  4  2009 /etc/passwd
            #
            # We want index 4 when indexing starting at 0
            $size = (split(/\s+/, $lsout[0]))[4];
        }

        if (defined($size) and $size ne "") {
            # Remove any commas from the output
            $size =~ s/\,//g;
            return $size;
        } else {
            &SanUtil::nate_log("Unable to determine size of file $rem_file");
            return undef;
        }
    } else {
        &SanUtil::nate_log('-fatal', "Unexpected output returned from command ($cmd)");
        return undef;
    }
}

################################################################################
# Subroutine Name: decode_remote_symlink
# Objective:       Take in a fully qualified file name and checks to see if it
#                  is a symlink.  If it is, it returns the file that the link is
#                  pointing at.  If it isn't a symlink, it just returns the original
#                  file.  For windows hosts, just return the filename since.
################################################################################
sub decode_remote_symlink {
    my ($connection, $rem_file) = @_;

    if (not defined($connection)) {
        &SanUtil::nate_log('-fatal', "Connection information not specified.");
        return $ClientUtil::FALSE;
    }

    # Return immediately for windows hosttype.
    if ($connection->{hostp}->{hosttype} =~ /windows/i) {
        # We have a windows host so just return the file.
        return $rem_file;
    }

    if (not defined($rem_file)) {
        &SanUtil::nate_log('-fatal', "Remote file name not specified.");
        return undef;
    }

    # Check to make sure that $file is fully qualified
    if ($rem_file !~ /^\//) {
        &SanUtil::nate_log('-fatal', "Filename specified ($rem_file) is not fully qualified");
        return undef;
    }

    my $cmd = "/bin/ls -l " . $rem_file;
    my ($status, @output) = &ClientUtil::execute($connection, $cmd);
    if ($status > 0) {
        # For this case, the "ls" failed so just return undef but don't set an error.  We will
        # assume it failed since the file doesn't exist.
        return undef;
    }

    # We found the file so we need to check the output.
    if (scalar(@output) > 0) {
        my @lsout = grep(/${rem_file}/, @output);
        if (scalar(@lsout) == 1) {
            if ($lsout[0] =~ /${rem_file} -> (\S+)$/) {
                # Looks like we have a symlink so return what it is linked to.
                return $1;
            } elsif ($lsout[0] =~ /^-.* ${rem_file}$/) {
                # We found the original file and it is not a symlink so return the original
                return $rem_file;
            }
        } else {
            &SanUtil::nate_log('-fatal', "Unexpected output returned from command ($cmd)");
            return undef;
        }
    } else {
        &SanUtil::nate_log('-fatal', "No output was returned from command ($cmd)");
        return undef;
    }
}

################################################################################
# Subroutine Name: exit_sysrq_dump_script_on_NCS
# Objective:       Will kill the sysrq dump script on Nate Controller Server
# Details:         See POD documentation at the end of this file
################################################################################
sub exit_sysrq_dump_script_on_NCS {
    my ($hostname) = @_;
    my (@output, $status, $cmd, $pid, $proc_info);

    # Get the pid from the file /var/tmp/<hostname>-sysrq-pid
    unless (open(PID_FILE, "</var/tmp/$hostname-sysrq-pid")){
        &SanUtil::nate_log('-warn', "Unable to open file /var/tmp/$hostname-sysrq-pid");
        return $ClientUtil::FALSE;
    }

    if (defined($pid = <PID_FILE>)) {
        close(PID_FILE);

        # Get proc info of the sysrq dump script running on Nate Control Server
        #
        # Here we are grepping for PID(which is a number) but this number
        # could occur in someone else proc at some other place. So we need
        # to make sure that we get actual proc. Also grepping with the
        # spaces around PID to make sure that this PID is not a substring
        # of another proc's PID
        #
        $cmd = "ps -elf | grep \" $pid \" | grep -v grep";
        &SanUtil::nate_log('comment', "Running cmd $cmd on NATE control server");
        ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
        if (scalar (@output) == 0) {
            &SanUtil::nate_log('comment', "SysRq dump trigger script is not running on NATE control server");
            return $ClientUtil::TRUE;
        }
        foreach my $line (@output) {
            my ($current_pid) = ($line =~ /^\s*\w+\s+\w+\s+\w+\s+(\d+)\s+.+/);
            if ($current_pid eq $pid) {
                $proc_info = $line;
                last;
            }
        }
        &SanUtil::nate_log('comment', "Proc info for sysrq dump script is $proc_info and PID is $pid");

        # Exit the script if its running
        if (defined ($proc_info)) {
            $cmd = "sudo kill -INT $pid";
            &SanUtil::nate_log('comment', "Running cmd $cmd on NATE control server");
            ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
            if ($status > 0) {
                &SanUtil::nate_log('warn', "Unable to kill sysrq dump script on Nate Control Server @output");
                return $ClientUtil::FALSE;
            }
        } else {
            &SanUtil::nate_log('comment', "SysRq dump script for client $hostname is not running on NATE control server");
            }
    } else {
        &SanUtil::nate_log('comment', "SysRq dump script for client $hostname is not running on NATE control server");
    }

    # Get rid of the temp files that are use to store the port numbers used by nc
    $cmd = "ls /var/tmp/nc/";
    ($status, @output) = &SanUtil::nate_server_execute(-cmd => $cmd);
    if ($status == 0) {
        foreach my $port_file (@output) {
            $cmd = "cat /var/tmp/nc/$port_file";
            my ($status, @cat_output) = &SanUtil::nate_server_execute(-cmd => $cmd);
            if ($hostname =~ /$cat_output[0]/) {
                # Delete the port file that belong to this host. It means that this port
                # is free and can be used by nc to create a connection with other hosts
                $cmd = "sudo rm -f /var/tmp/nc/$port_file";
                my ($status, @tmp_output) = &SanUtil::nate_server_execute(-cmd => $cmd);
                if ($status > 0) {
                    &SanUtil::nate_log('fail', "Failed to run command \'$cmd\' on NATE server");
                    return $ClientUtil::FALSE;
                }

                # Get all the process using this port and kill them
                my @ps_output;
                $cmd = "sudo ps -elf | grep \"nc -u -l $port_file\" | grep -v grep";
                ($status, @ps_output) = &SanUtil::nate_server_execute(-cmd => $cmd);
                foreach my $ps_info (@ps_output) {
                    my (undef, undef, undef, $pid) = split (/\s+/, $ps_info);
                    $cmd = "sudo kill -9 $pid";
                    my ($status, @ps_kill_output) = &SanUtil::nate_server_execute(-cmd => $cmd);
                    if ($status > 0) {
                        # If we fail here it is important that we know about this so that we can
                        # manually kill this process. Otherwise it will keep running and it might create
                        # huge log file
                        &SanUtil::nate_log('fail', "Failed to kill the process with pid '$pid' on NATE Server. " .
                                                   "Please inform NATE Server maintainer about this issue");
                        return $ClientUtil::FALSE;
                    }
                }
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: scp_file
# Objective:       SCP a file to the remote SanClient object (must have ssh)
################################################################################
sub scp_file {
    my ($self, %args) = @_;

    # verify that self is a sanclient object
    if (ref($self) !~ /SanClient/i) {
        return $ClientUtil::FALSE;
    }

    # obtain scp credentials and the local file / remote file names
    my $hostrec = &Tharn::host($self->{host});
    my $ip_addr = $hostrec->{default_ip};
    my $user = $self->{user};
    my $pass = $self->{pass};
    my $local_file = exists($args{-local_file}) ? $args{-local_file} : undef;
    if (not defined($local_file)) {
        $self->_set_err("Missing required parameter '-local_file'");
        return $ClientUtil::FALSE;
    }
    my $remote_file = exists($args{-remote_file}) ? $args{-remote_file} : undef;
    if (not defined($remote_file)) {
        $self->_set_err("Missing required parameter '-remote_file'");
        return $ClientUtil::FALSE;
    }
    my $mode = exists($args{-mode}) ? $args{-mode} : "send";

    # transfer the dmmp discovery script to the remote host.
    require Expect;

    # build the command and transfer the file
    my $cmd;
    if ($mode =~ /send/i) {
        # send a file
        $cmd = "scp -o \"StrictHostKeyChecking=no\" -o \"UserKnownHostsFile=/dev/null\" " .
            "${local_file} ${user}\@${ip_addr}:${remote_file}";
    } else {
        # receive a file
        $cmd = "scp -o \"StrictHostKeyChecking=no\" -o \"UserKnownHostsFile=/dev/null\" " .
            "${user}\@${ip_addr}:${remote_file} ${local_file}";
    }
    $Log->comment("Transferring file: $cmd");

    my $scp = Expect->spawn("$cmd");

    # hide stdout logging from expect
    $scp->log_stdout(0);

    # look for the password prompt
    eval {
        $scp->expect(30,
                     [ qr/ssword:/ =>
                           sub {
                               $Log->debug("Received password prompt\n");
                               my $e = shift;
                               # answer the password
                               print $e "$pass" . "\n";
                               $e->exp_continue;
                           }
                       ],
                     [EOF => sub {$Log->debug("Hit EOF"); die "Error: Hit EOF";}],
                     [timeout => sub {$Log->debug("Timed out"); die "Error: Timed out"}],
                     '-re', '\$');
    };
    if ($@) {
        $self->_set_err($@);
        return $ClientUtil::FALSE;
    }

    $scp->soft_close();

    $Log->comment("Transferring complete");

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: _is_PNATE_ENVIRONMENT
# Objective:       Determine if PNATE environment is enabled, return true/false
###############################################################################
sub _is_PNATE_ENVIRONMENT {
    my $pnate = &Tharn::param("PNATE_ENVIRONMENT");
    if (not defined($pnate)) {
        return $ClientUtil::FALSE;
    } elsif ($pnate == 1) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

1;

=pod

=head1 Package Name

    ClientUtil.pm

=cut
