# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to DevMgr on host machine
## @author lippe@netapp.com
## @status private
## @pod here

package DevMgr;

use strict;

use Error;
use San::SanLog;

use vars qw( @ISA);

@ISA = qw(Error SanLog);

sub _new {
    my ($class, %args) = @_;
    my ($self, $hosttype);

    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self,$class;

    # check for existance of required -parent argument
    if (grep ( /^-?parent$/, (keys %args) )) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Error: DevMgr::_new() " .
                            "Invalid value for argument '-parent'");
            return $self;
        }
        $hosttype=$args{-parent}->get_connection()->{hostp}{hosttype};
    } else {
        $self->_set_err("Error: DevMgr::_new() " .
                        "Missing required argument '-parent'");
        return $self;
    }

    # determine if this is a virtual machine
    my $hypervisor_type;
    if ($args{-parent}->{is_virtual} == 1) {
        $hypervisor_type = $args{-parent}->{hypervisor_type};
    }

    # load the appropriate DevMgr for the hosttype
    if ($hosttype =~ /unix-solaris/i) {
        if ($hypervisor_type =~ /VMware|Microsoft|HyperV|Citrix|KVM|OVM/i) {
            require San::SanClient::DevMgr::VSolarisDevMgr;
            $self = San::SanClient::DevMgr::VSolarisDevMgr->_new(%args);
        } else {
            require San::SanClient::DevMgr::SolarisDevMgr;
            $self = San::SanClient::DevMgr::SolarisDevMgr->_new(%args);
        }
    } elsif ($hosttype =~ /unix-hpux/i) {
        require San::SanClient::DevMgr::HPUXDevMgr;
        $self = San::SanClient::DevMgr::HPUXDevMgr->_new(%args);
    } elsif ($hosttype =~ /unix-aix/i) {
        require San::SanClient::DevMgr::AixDevMgr;
        $self = San::SanClient::DevMgr::AixDevMgr->_new(%args);
    } elsif ($hosttype =~ /unix-linux-\w+-kvm/i) {
        require San::SanClient::DevMgr::KVMDevMgr;
        $self = San::SanClient::DevMgr::KVMDevMgr->_new(%args);
    } elsif ($hosttype =~ /unix-linux-\w+-ovm/i) {
        require San::SanClient::DevMgr::OVMDevMgr;
        $self = San::SanClient::DevMgr::OVMDevMgr->_new(%args);
    } elsif ($hosttype =~ /unix-linux/i) {
        if ($hypervisor_type =~ /VMware|Microsoft|HyperV|Citrix|KVM|OVM/i) {
            require San::SanClient::DevMgr::VLinuxDevMgr;
            $self = San::SanClient::DevMgr::VLinuxDevMgr->_new(%args);
        } else {
            require San::SanClient::DevMgr::LinuxDevMgr;
            $self = San::SanClient::DevMgr::LinuxDevMgr->_new(%args);
        }
    } elsif ($hosttype =~ /windows-hyper-v/i) {
        require San::SanClient::DevMgr::HyperVDevMgr;
        $self = San::SanClient::DevMgr::HyperVDevMgr->_new(%args);
    } elsif ($hosttype =~ /windows/i) {
        require San::SanClient::DevMgr::WinDevMgr;
        $self = San::SanClient::DevMgr::WinDevMgr->_new(%args);
    } elsif ($hosttype =~ /linux-vmware/i) {
        require San::SanClient::DevMgr::ESXDevMgr;
        $self = San::SanClient::DevMgr::ESXDevMgr->_new(%args);
    } else {
        $self->_set_err("Error: DevMgr::_new() Host type '".$hosttype."' " .
                        "not currently supported with this library");
        return $self;
    }

    $self->{parent} = $args{-parent};
    $self->set_debug_level($self->{parent}->get_debug_level());

    #
    # By returning $self here we are returning a parent class such
    # as SolarisDevMgr, not a DevMgr object.  This is OK because
    # the OS specific DevMgr::XXXDevMgr classes add DevMgr to @ISA
    # so DevMgr is the base class for the parent
    #
    return $self;
}
1;


#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

DevMgr.pm

=head1 Author

Kris Lippe

=head1 Description

Provides an abstract API interface to allow operations on host raw physical
volumes.  This object is used mainly in conjunction with raw device operations
as it allows operations not possible due to the VG-centric nature of the Lvm.pm
module.  The DevMgr object works as a container for individual device objects.  

=head1 See Also

Package inherits from the following modules. See modules for interface detail

B<Error.pm>

=head1 Attributes

=head2 B<parent>

SanClient object = Reference to a parent SanClient object

=head1 Public Methods

=head2 B<get_device($name)>

Returns device object with given name.  Returns error if none exists.

=over 4

=item Inputs

=item * -name

The name of the physical device (e.g on solaris c21t0d0)

=item Outputs

Returns a  device object. If an error occurs, the error flag is set and errmsg is
set to the failure reason.  

=item Examples

    # Return a device object (solaris).
    $Test->log("-- Get device object");
    my $devObj = $DevMgr->get_device("c21t0d0");
    if ($DevMgr->error){
        $Test->log($DevMgr->errmsg());    
        return 0;
    }

=back


=head2 B<get_devices()>

Returns an array of all configured device objects.

=over 4

=item Outputs

Returns an array of device objects.  If none were configured, the empty set will
be returned.

=item Examples

    # Return all configured device objects.
    $Test->log("-- Get device objects");
    if (not $DevMgr->get_devices){
        $Test->log($DevMgr->errmsg());
        return 0;
    }

=back


=head2 B<create_device(I<%args>)>

Returns a device object.

=over 4

=item Inputs

=item * -name

The name of the physical device (e.g c21t0d0)

=item Outputs

Creates a device object by calling the Dev object constructor. If an error occurs, 
the error flag is set and errmsg is set to the failure reason.  

=item Examples

    # Create a device manager object.
    $Test->log("-- Create device manager object");
    $DevMgr = $Client->create_devmgr();
    if ($Client->error){
        $Test->log($Client->errmsg());    
        return 0;
    }

    # Get free disk for device object creation.
    $Test->log("-- Get free device");
    $device = $DevMgr->get_free_devices(-cnt=>1);
    if ($DevMgr->error){
        $Test->log($DevMgr->errmsg());    
        return 0;
    }

    # Create a device object.
    $Test->log("-- Create device object");
    my $devObj = $DevMgr->create_device($device);
    if ($devObj->error){
        $Test->log($devObj->errmsg());    
        return 0;
    }

=back


=head2 B<cleanup()>

Destroys all device objects and cleans up mount points.  Since a device object is 
built on the assumption that physical devices already exist on the system, cleanup
of the devmgr object DOES NOT remove physical devices from the system.  It merely 
removes the object reference.

=over 4

=item Outputs

Destroys a device object. If an error occurs, the error flag is set and errmsg is
set to the failure reason.  

=item Examples

    # Destroy a device object.
    $Test->log("-- Cleaning up devmgr object");
    $DevMgr->cleanup();
    if ($DevMgr->error){
        $Test->log($DevMgr->errmsg());    
        return 0;
    }

=back


=head2 B<get_free_devices(I<%args>)>

Returns a list of devices that are not being used by other Device or LVM objects.  The 
list of devices can be selected by any combination of filer, volume, and lun.

=over 4

=item Inputs

=item * I<-cnt> 

The number of free devices to retrieve.  If the number of free devices found is less 
than the count than the error flag is set.

=item * I<-filer>

Retrieve host devices belonging to a specific filer. 

=item * I<-vol>    (ex: vol2)

Retrieve host devices belonging to a specific filer volume.  Only supply the filer 
volume name and not the full volume path.

=item * I<-lun> (ex: mylun)

Retrieve the host device mapped to a specific lun.  Only supply the lun name and not 
the full path

=item * I<-refresh>    (default: 0)

Forces the host device to filers maps to be refeshed.  This is automatically done the 
first time the method is called.  It needs to be called whenever a test program 
dynamically creates and discovers new luns in a test method. This will cause the map 
to be updated and correctly reflect device metadata on the host.

=item Outputs

Returns an array or array reference of free devices.  If an error occured the error flag 
is set and an errmsg is set to the failure reason.       

=item Examples

    $Test->log("Get 3 free devices from filer 'myfiler'");
    @devs = $DevMgr->get_free_devices(-cnt=>3, -filer=>'myfiler');
    if ($DevMgr->error){
        $Test->log($DevMgr->errmsg());    
        return 0;
    }

=back

1;




