# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Devices on Aix
## @author thomasm@netapp.com
## @status private
## @pod ./Dev.pod

# Package Name
##############
package San::SanClient::DevMgr::AixDev;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: load
# Objective:       sets parameters based on devices current mount state
# Details:         See POD documentation at the end of this file
################################################################################
sub load {
    my ($self) = @_;

    my ($cmd, $name, $partition, $phash, $status, $type, @output);

    $self->_clear_err();

    $self->sanlog(-msg => "Populating $self->{name} data object");
  
    # save response to mkfs prompt
    $cmd = "mount | grep /dev/$self->{name} | awk '{print \$2}{print \$3}'";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);  
    if ($status > 0) {
    $self->_set_err("Error: AixDev::load Mount command failed", @output); 
        return $ClientUtil::FALSE;                            
    }

    $self->sanlog(-msg => "Loading $self->{name}  Mount: $output[0] FS: $output[1]");

    $self->{mount} = $output[0];
    $self->{fs} = $output[1];

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: create_fs
# Objective:       create a filesystem on the device object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_fs {
    my ($self, %args) = @_;

    my ($cmd, $name, $partition, $phash, $status, $type, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
      if (/-?type$/) {
          $type = $args{$_};
          next;
      } 
    } 
    
    # if type was not passed, then default to jfs2
    if (not defined $type or $type eq "") {
        $type = "jfs2";
    }
    
    if ($type =~ /^jfs2$/i) {
        # Since this is FS on raw, all logs must be inline
        $cmd = "yes | mkfs -V $type -o log=INLINE /dev/$self->{name}";

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status > 0) {
            $self->_set_err("Error: AixDev::create_fs Unable to create $type filesystem on volume '$name'",
                            @output);
            return $ClientUtil::FALSE;  
        } 
    } else {
        $self->_set_err("Error: AixDev::create_fs()",
                        "Invalid Filesystem type '$type'");
        return $ClientUtil::FALSE;  
    }

    $self->{fs} = $type;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: mount_fs
# Objective:       mount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_fs {
    my ($self, %args) = @_;

    my ($cmd, $directio, $type, $vgname, $mount, $name, $status, @output);

    # Get attributes 
    $directio = 0;
    $name = $self->get_name();
    $type = $self->get_fs();
    $mount = undef;
    
    # check if volume is unformatted
    unless (length $type) {
        $self->_set_err("Error: AixDev::mount_fs()",
                        "No filesystem detected");
        return $ClientUtil::FALSE; 
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?path$/) {
            if (not length($mount = $args{$_})) {
                $self->_set_err("Error: AixDev::mount_fs()",
                                "Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }        
            next;
        } elsif (/-?directio$/) {
            if (not length($directio = $args{$_}) or $args{$_} !~ /0|1/) {
                $self->_set_err("Error: Dev::mount_fs()",
                                "Invalid value for argument '-directio'");
                return $ClientUtil::FALSE;
            }        
            next;
        } 
    }

    # check if the volume is already mounted
    if ($self->is_mounted()){
        $self->_set_err("Error: AixDev::mount_fs() device '$name' is already mounted");                    
        return $ClientUtil::FALSE;                            
    }
    
    # Use default path if not defined
    unless (defined $mount) {
        $mount = "/mnt/qa/$name/";
        $self->sanlog(-msg => "using default mount pt '$mount'");      
    } else {
        $self->sanlog(-msg => "using mount pt '$mount'"); 
    }
    
    $self->{mount} = $mount;

    # Create the mount point and open up dir privs 
    $cmd = "mkdir -p $mount; chmod 777 $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: AixDev::mount_fs() Unable to create mount point directory for volume '$name'",
                        @output);
        return $ClientUtil::FALSE;                            
    } 

    $cmd = ($directio) 
        ? "mount -V $type -o log=INLINE,dio /dev/$name $mount"
        : "mount -V $type -o log=INLINE /dev/$name $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: AixDev::mount_fs() Unable to mount volume filesystem", @output);
        return $ClientUtil::FALSE;                            
    } 

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, %args) = @_;

    my ($cmd, $type, $vgname, $fs, $mount, $name, $status, $force);
    my ($option, @output);

    $self->_clear_err();

    $option = "";
    
    # parse hash
    foreach (keys %args) {
        if (/-?force$/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: AixDev::unmount_fs()",
                                "Invalid value for argument '-force'");
                return $self;
            }        
            next;
        } 
    }  

    $option = "-f" if (defined $force);

    # Get attributes 
    $name = $self->get_name();
    $mount = $self->get_mount();
    $fs = $self->get_fs();

    # check if device contains a filesystem
    unless (length $fs) {
        $self->_set_err("Error: AixDev::unmount_fs()",
                        "No filesystem detected");
        return $ClientUtil::FALSE; 
    }

    # check if volume is mounted
    unless (length $mount) {
        $self->_set_err("Error: AixDev::unmount_fs()",
                        "Cannot unmount a filesystem that is not mounted");
        return $ClientUtil::FALSE; 
    } 

    # Unmount the volume's filesystem
    $cmd = "umount $option /dev/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: AixDev::unmount_fs() Unable to unmount volume's filesystem",
                        @output);
        return $ClientUtil::FALSE;                            
    } 

    # Remove mount pt metadata from object
    $self->{mount} = "";

    $self->sanlog(-msg => "removing mount point '$mount'");    
 
    # Remove directory head
    my ($base, $parent) = fileparse($mount);
    if ($mount ne "/"){
        $cmd = "rm -rf $mount";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: AixDev::unmount_fs() Unable to remove mount point '$mount'", 
                            @output);
            return $ClientUtil::FALSE;                            
        } 
    } else {
        $self->_set_err("Error: AixDev::unmount_fs() Will not remove mount point of \"/\"");
        return $ClientUtil::FALSE;                            
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if filesystem is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{mount} and length $self->{mount}) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: is_partitioned
# Objective:       returns ClientUtil::TRUE if partitioned. 
#                  This is a noop on AIX
# Details:         See POD documentation at the end of this file
################################################################################
sub is_partitioned {
    my ($self) = @_;
    
    $self->_clear_err();

    return $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: partition
# Objective:       create a partition on the raw device
#                  This is a noop on AIX
# Details:         See POD documentation at the end of this file
################################################################################
sub partition {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: destroy_partition
# Objective:       destroy a partition on the raw device
#                  This is a noop on AIX
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_partition {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: get_uuid
# Objective:       Returns the UUID of the configured device.
#                  This function is a noop on AIX
# Details:         See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self) = @_;
    
    $self->_clear_err();

    return "";

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{size})){
        return $self->{size};
    } else {
        $self->_set_err("Error: AixDev::get_size() Vol object is missing size");
        return (0);
    }

}

################################################################################
# Subroutine Name: get_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined($self->{name}) and length($self->{name})) {
        return $self->{name};
    } else {
        $self->_set_err("Error: AixDev::get_name() Dev object is missing name");
        return "";
    }

}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       returns the raw device name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;
    
    # simply call get_name. Needed for Windows compatibility.
    return ($self->get_name());

}

################################################################################
# Subroutine Name: get_mount
# Objective:       returns mount point of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined($self->{mount}) and length($self->{mount})) {
        return $self->{mount};
    } else {
        $self->_set_err("Error: AixDev::get_mount() Dev object is missing mount");
        return "";
    }
}

################################################################################
# Subroutine Name: get_fs
# Objective:       returns fs of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined($self->{fs}) and length($self->{fs})) {
        return $self->{fs};
    } else {
        $self->_set_err("Error: AixDev::get_fs() Dev object missing fs");
        return "";
    }

}

################################################################################
# Subroutine Name: get_lun
# Objective:       Return lun name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{lun}) and length($self->{lun})) {
        return $self->{lun};
    } else {
        $self->_set_err("Error: AixDev::get_lun() Dev object is missing lun name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_partition
# Objective:       Returns the name of the configured partition.
#                  This is a noop on AIX
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition {
    my ($self) = @_;
    
    $self->_clear_err();

    return "";

}

################################################################################
# Subroutine Name: get_partition_size
# Objective:       Returns the name size of a particular partition in bytes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition_size {
    my ($self, $partition_num) = @_;
    
    $self->_clear_err();

    # partitions are not support on hdisks, simply
    # return the entire size of the device. Error will
    # be set by get_size if this fails.
    my $size = $self->get_size();

    return $size;

}

################################################################################
# Subroutine Name: get_label
# Objective:       Returns the name of the configured partition's label.
#                  This is a noop on AIX
# Details:         See POD documentation at the end of this file
################################################################################
sub get_label {
    my ($self) = @_;
    
    $self->_clear_err();

    return "";

}

################################################################################
# Subroutine Name: refresh_geometry
# Objective:       refresh the geomery on device. Usefull to call after a 
#                  resize on controller. NO-OP on this platform
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh_geometry {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new dev object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $conn, $log, $name, $parent, $psize, $self, $status, @output);

    $self = {};

    bless ($self, $class);    

    # parse arguments
    foreach (keys %args) {
        if (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: AixDev::new()",
                               "Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixDev::new()",
                                "Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/){
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: AixDev::new()",
                                "Missing value for argument '-parent'");
            return $self;
            }
            next;
        }
    }  
    
    # check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: AixDev::new()",
                        "Missing required argument '-name'");
        return $self; 
    }

    # check for existance of required -parent argument
    unless (defined $parent) {
        $self->_set_err("Error: AixDev::new()",
                        "Missing required argument '-parent'");
        return $self; 
    }

    # check for existance of required -conn argument
    if (grep ( /^-?conn$/, (keys %args) )) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: AixDev::new()",
                            "Invalid value for argument '-conn'");
            return $self; 
        }    
    } else {
        $self->_set_err("Error: AixDev::new()",
                        "Missing required argument '-conn'");
        return $self; 
    }

    # initilize the object members
    $self->{conn} = $conn;
    $self->{parent} = $parent;
    $self->{name} = $name;

    # get filer lun associated with physical dev
    $self->{lun} = $parent->{sanlun}->get_lun_path(-device => $name);
    if ($parent->{sanlun}->error) {
        $self->_set_err($parent->{sanlun}->errmsg());
        return $self; 
    }

    $self->{fs} = "";
    $self->{mount} = "";
    $self->{size} = $self->_init_size();

    return $self;

}

################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing disks
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self) = @_;
    
    my($cmd, $size, @output, $status);
    
    $size = "";
    
    # formulate bootinfo command
    $cmd = "bootinfo -s $self->{name}";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixDev::_init_size() Unable to retrieve size (bytes) for disk \'$self->{name}\'", 
                        @output);
        return "";
    }

    # get the size in bytes returned by the bootinfo
    $size = &ClientUtil::convert_to_bytes($output[0]  . "m");
    
    return $size;    
  
}

1;
