# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to DevMgr for Aix host
## @author thomasm@netapp.com 
## @status private
## @pod ./DevMgr.pod

# Package Name
##############
package San::SanClient::DevMgr::AixDevMgr;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanClient::DevMgr;
use San::SanClient::DevMgr::AixDev;
use San::SanLog;
use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog DevMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_device
# Objective:       return a device object based on the device name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device {
    my ($self, $name) = @_;
    
    $self->_clear_err();

    unless(defined $name and length $name){
        $self->_set_err("Error: AixDevMgr::get_device() name is a required argument");
        return (wantarray()) ? () : [];                          
    }

    foreach my $obj ($self->get_devices){
        if ($obj->get_name() =~ /^$name$/){
            return $obj;
        }
    }
    
    $self->_set_err("Error: AixDevMgr::get_device() Unable to find device '$name'");
    return (wantarray()) ? () : [];                          
    
}

################################################################################
# Subroutine Name: get_devices
# Objective:       return all currently configured device objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;

    my (@devices);
    
    $self->_clear_err();

    if (@{$self->{devices}}) {
        @devices = @{$self->{devices}};
    } else {
        return ();        
    }

    return (wantarray()) ? @devices : \@devices;  

} 

################################################################################
# Subroutine Name: create_device
# Objective:       create a device object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_device {
    my ($self, %args) = @_;

    my ($devObj);

    $self->_clear_err();

    $devObj = San::SanClient::DevMgr::AixDev->_new(-parent => $self,
                           -conn => $self->{conn},
                           %args);

    if ($devObj->error) {
        $self->_set_err($devObj->errmsg());
        return $self;
    }
    
    push @{$self->{devices}}, $devObj;

    return ($devObj);  

} 

################################################################################
# Subroutine Name: get_dev_id
# Objective:       returns unique device name(with full path) of passed device's
#                  short name to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_id {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: get_dev_name
# Objective:       returns device short name of passed device's unique name(long name)
#                  to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_name {
    my ($self) = @_;

    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: cleanup
# Objective:       remove device objects and remove mount point entries 
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup {
    my ($self) = @_;

    my ($devObj, $fail_cnt);

    $self->_clear_err();

    $fail_cnt = 0;

    # remove mount point entry for each device passed in
    foreach $devObj ($self->get_devices) {
        
        $self->sanlog(-msg => 
                      "Cleaning up mount points for device: $devObj->{name}");
        
        $devObj->unmount_fs() if $devObj->is_mounted;
        if ($devObj->error) {
            $self->_set_err($devObj->errmsg);
            $fail_cnt++;
        }
        
        # remove physical devices from system
        $self->_delete_devices($devObj->{name});
        if ($self->error) {
            $fail_cnt++;
        }
    }
    
    $self->{devices} = [];

    return ($fail_cnt == 0) ? $ClientUtil::TRUE : $ClientUtil::FALSE;

} 

################################################################################
# Subroutine Name: get_free_devices
# Objective:       retrieve list of free host volumes 
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices { 
    my ($self, %args) = @_;

    my ($bCnt, $cmd, $cnt, $device, $filer, $filername, $lun, $refresh);
    my ($status, $total, $vol, %hDevs, @all_devices, @attr, @devices);
    my (@disk_devices, @output, @used, @new_devices);
    my ($bootdevice, $filers, $protocol);

    $total = 0;
    @{$filers} = ();

    $self->_clear_err();

    # Asssume user supplied -cnt argument
    $bCnt = 1;
    $refresh = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Error: AixDevMgr::get_free_devices() Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [];
            }        
            next;
        } elsif (/-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Error: AixDevMgr::get_free_devices() Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Error: AixDevMgr::get_free_devices() Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Error: AixDevMgr::get_free_devices() Invalid value for argument '-vol'");
                return (wantarray()) ? () : [];
            }        
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Error: AixDevMgr::get_free_devices() Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];
            }        
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: AixDevMgr::get_free_devices() Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [];
            }        
            next;
        } elsif (/^-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: AixDevMgr::get_free_devices() Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }     
    
    # Check if user wants free devices from a specified filer, volume, or lun
    if (defined $filer or defined $vol or defined $lun or defined $protocol or scalar(@{$filers}) != 0 or $refresh == 1) {
        # Check if a device to filer map 'dfmap' has already been created or
        # refresh it if requested.
        if (not defined $self->{dfmap} or $refresh eq 1) {
            $self->sanlog(-msg => "Creating translation maps; device to filer, vol, and lun");
            @output = $self->{sanlun}->lun_show();
            if ($self->{sanlun}->error) {
                $self->_set_err($self->{sanlun}->errmsg());
                return (wantarray()) ? () : [];
            }

            foreach my $line (@output) {
                # skip first line (header)
                if ($line =~ /lun-pathname/i) {
                    next;
                }

                # depending on filername length sanlun may not have display blank space before name
                my ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp);
                if ($line  =~ /\s{0}(\S+):*\s+\/vol\/(\w+)\/(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+/) {
                    # The line is in the following format
                    # p720_aix03c1_data:  /vol/p720_aix03c1_data32_0/p720_aix03c1_data32_lun_0  hdisk1           fcs0     FCP            2g (2147483648)     GOOD
                    # p720_aix03c1_data:  /vol/p720_aix03c1_data64_1/p720_aix03c1_data32_lun_0  hdisk2           fcs0     FCP            2g (2147483648)     GOOD
                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp) = ($1, $2, $3, $4, lc($5));
                } elsif ($line =~ /\s{0}(\S+):*\s+\/vol\/(\w+)\/(\S+)\s+(\S+)\s+/) {
                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp) = ($1, $2, $3, $4);
                }

                if ($device_tmp eq "" or $device_tmp =~ /^\s*$/) {
                    $self->sanlog(-msg => "Output from sanlun is incorrect:\n$line\n");
                    next;
                }

                my $dev_tmp = basename($device_tmp);

                $hDevs{$dev_tmp} = {filer    => $filer_tmp,
                                    vol      => $vol_tmp,
                                    lun      => $lun_tmp,
                                    protocol => $protocol_tmp};

                $self->sanlog(-msg => "Device ($dev_tmp) information:\n\tFiler: $filer_tmp\n\tVolume: $vol_tmp\n" .
                                      "\tLUN: $lun_tmp\n\tProtocol: $protocol_tmp\n",
                              -msg_level => 3);

            }

            # save map for next call of function
            $self->{dfmap} = \%hDevs;

        }
    }

    # If not -cnt then assume all free disks 
    # under lvm control are wanted
    unless (defined $cnt and length($cnt) and $cnt > 0) {
        $cnt = 10000;
        $bCnt = 0;
    }

    # Get list of devices. Make sure you take in account deported vgs.
    if ($self->{parent}->get_protocol() =~ /fcp/i) {
        if ($self->{parent}->{mp_type} =~ /sanpath/i) {
            $cmd = "setsp | egrep  'spd[0-999]' | grep 'NETAPP' | awk '{print \$3;}'";
        } elsif ($self->{parent}->{mp_type} =~ /mpio/i) {
            $cmd = "lsdev -C -c disk | egrep \"\(MPIO)\|\(Virtual\)\" | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\) FCP\" | awk '{print \$1;}'"; 
        } else {
            $cmd = "lsdev -C -c disk | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\)\" | awk '{print \$1}'";
        }

        ($status, @disk_devices) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status gt 0) {
            $self->_set_err("Error: AixHost::get_free_devices() Command ($cmd) returned a bad status ($status)");
            return($ClientUtil::FALSE);                             
        }

        # determine if there were any sanpath devices available
        if ($self->{parent}->{mp_type} =~ /sanpath/i) {
            if (grep(/no spd devices found/i, @disk_devices) > 0 ) {
                $self->_set_err("Error: AixHost::get_free_devices() Command ($cmd) failed to find Sanpath devices.  Check license sppath -v.");
                return($ClientUtil::FALSE);
            }
        }
    } elsif ($self->{parent}->get_protocol() =~ /iscsi/i) {
        if ($self->{parent}->{mp_type} =~ /mpio/i) {
            $cmd = "lsdev -C -c disk | egrep \"\(MPIO)\|\(Virtual\)\" | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\) iSCSI\" | awk '{print \$1;}'"; 
        } else {
            $cmd = "lsdev -C -c disk | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\)\" | awk '{print \$1}'";
        }
        ($status, @disk_devices) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status gt 0) {
            $self->_set_err("Error: AixHost::get_free_devices() Command ($cmd) returned a bad status ($status)");
            return($ClientUtil::FALSE);                             
        }
    }

    # Get list of all free devices and paths
    $cmd = "lspv | grep -i none | awk '{print \$1}'";  
    ($status, @all_devices) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: AixHost::get_free_devices() Command ($cmd) returned a bad status ($status)");
        return($ClientUtil::FALSE);                             
    } 

    # get the boot device - do not attempt to rmdev this - could
    # be sanbooted netapp device.
    $cmd = "bootinfo -b";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: AixDevMgr::get_free_devices() Command ($cmd) returned a bad status ($status)");
        return($ClientUtil::FALSE);                             
    }
    $bootdevice = $output[0];
    
    # iterate through the devices and make sure that none of the disks 
    # are being used other LVM disk object.  
    foreach my $device (@disk_devices) {

        # skip the boot device
        if ($device eq $bootdevice) {
            $self->sanlog(-msg => "\nFilter bootdevice: $bootdevice\nDevice: $device",
                          -msg_level => 3);
            next;
        }

        # If  the device is being used by other VGs, skip it
        if (grep(/^$device$/, @all_devices) == 0) {
            $self->sanlog(-msg => "\nFilter all devices: " . join(" " , @all_devices) . "\nDevice: $device",
                          -msg_level => 3);
            next;
        }

        # Apply filters to any of the free devices found
        if (defined $filer and length $filer) {
            $self->sanlog(-msg => "\nFilter filer: $filer\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i);
        }

        if (defined $vol and length $vol) {
            $self->sanlog(-msg => "\nFilter volume: $vol\nDevice: $device\nDevice volume: " . $self->{dfmap}{$device}{vol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i);
        }

        if (defined $lun and length $lun) {
            $self->sanlog(-msg => "\nFilter lun: $lun\nDevice: $device\nDevice lun: " . $self->{dfmap}{$device}{lun},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{lun} !~ /^$lun$/i);
        }

        if (defined $protocol and length $protocol) {
            $self->sanlog(-msg => "\nFilter protocol: $protocol\nDevice: $device\nDevice protocol: " . $self->{dfmap}{$device}{protocol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{protocol} !~ /$protocol/i);
        }

        # Filter on a filer list if provided by caller
        if (defined $filers and scalar(@{$filers}) != 0) {
            my $found = 0;
            # Look through each filer in the filers list
            foreach my $filer_f (@{$filers}) {
                $self->sanlog(-msg => "\nFilter filer(s): $filer_f\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                              -msg_level => 3);
                # If filer matches devices filer, set found to 1
                if ($self->{dfmap}{$device}{filer} =~ /^$filer_f$/i) {
                    $found = 1;
                }
            }

            # If the device is not part of the filers list caller requested skip it.
            if ($found == 0) {
                next;
            }
        }

        push @devices, $device;
    }

    # If user supplied cnt and free devices fell short
    # generate error
    if ($bCnt == 1 and scalar(@devices) < $cnt){
        $self->_set_err("Error: AixDevMgr::get_free_devices() ",
                        "Only " . scalar(@devices) .
                        " of $cnt requested devices were found");
    }

    $self->sanlog(-msg => "Devices returned: @devices");

    return (wantarray()) ? @devices : \@devices; 

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new devices object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $dev, $devObj, $log, $loadcfg, $lvm, $parent, $hosttype);
    my ($self, $status, @devices, @output, $load_devices); 

    $self = {};

    bless ($self, $class);    

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: AixDevMgr::_new() Invalid value for argument '-parent'");
                return ($self);
            }
            next;
        } elsif (/-?load_devices/) {
            if (not length($load_devices = $args{$_})) {
                $self->_set_err("Error: AixDevMgr::_new() Invalid value for argument '-load_devices'");
                return ($self);
            }
            next;
        }
    }     

    # set default for load_devices (true)
    if (not defined($load_devices)) {
        $load_devices = 1;
    }
    
    # check for existance of required -parent argument
    if (grep ( /^-?parent$/, (keys %args) )) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Error: AixDevMgr::_new()",
                            "Invalid value for argument '-parent'");
            return $self; 
        }
        $hosttype = $args{-parent}->{hostp}{hosttype};
    } else {
        $self->_set_err("Error: AixDevMgr::_new()",
                        "Missing required argument '-parent'");
        return $self; 
    }

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{devices} = [];

    # Set debug level on this object to whatever the parents level is
    $self->set_debug_level($parent->get_debug_level());

    if ($load_devices == 1) {
        @devices = $self->get_free_devices(%args);
        return($self) if ($self->error);

        foreach $dev (@devices) {
            # initialize device object from the contents of the device manager
            # configuration file.  Sometimes stale entries can appear from
            # previously failed runs.  If this happens, don't bother loading
            # the device.
            $devObj = $self->create_device(-name => $dev);
            unless ($devObj->error) {
                $devObj->load();
            }
        }
    }
    
    return($self);

}

################################################################################
# Subroutine Name: _delete_devices
# Objective:       delete devices from host
# Details:         See POD documentation at the end of this file
################################################################################
sub _delete_devices {
    my ($self, @devices) = @_;

    my ($cmd, $status, @output); 

    $self->_clear_err();

    # Remove devices from host.  This is because orphaned
    # devices maintain metadata of prior lun
    if ($self->{parent}->get_protocol() =~ /fcp/) {
        foreach my $device (@devices) {
            $self->sanlog(-msg => "device: " . $device);

            if ($self->{parent}->{mp_type} =~ /sanpath/i) {
                # Determine spd id for master device name
                my $spd_id = $self->_get_spd_id($device);
                $self->sanlog(-msg => $self->errmsg()) if ($self->error);
                $self->sanlog(-msg => "spd_id: " . $spd_id);
        
                # Retrieve all device paths for spd_id
                my @alldevpaths = $self->_get_spd_devices($spd_id);
                $self->sanlog(-msg => $self->errmsg()) if ($self->error);
        
                # Stop master spd device
                $self->_stop_spd_id($spd_id);
                $self->sanlog(-msg => $self->errmsg()) if ($self->error);
    
                # Remove master spd device
                $self->_remove_spd_id($spd_id);
                $self->sanlog(-msg => $self->errmsg()) if ($self->error);

                # Remove device from host
                foreach my $device (@alldevpaths) {
                    $cmd = "rmdev -d -l " . $device;
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status gt 0) {
                        $self->_set_err("Error: AixDevMgr::destroy() Unable to remove device '" . 
                                        $device . "'", @output);
                        
                        return $ClientUtil::FALSE;             
                    } 
                }
            } elsif ($self->{parent}->{mp_type} =~ /mpio/i) {
                $cmd = "rmdev -d -l " . $device;
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
                if ($status gt 0) {
                    $self->_set_err("Error: AixDevMgr::destroy() Unable to remove device '" .
                                    $device . "'", @output);
                        
                    return($ClientUtil::FALSE);                       
                }
            }
        }
    } else {
        foreach my $device (@devices){
            $cmd = "rmdev -d -l " . $device;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status gt 0) {
                $self->_set_err("Error: AixDevMgr::destroy() Unable to remove device '" .
                                $device . "'", @output);
                
                return($ClientUtil::FALSE);                       
            }
        }
    }

    return($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: _get_spd_devices
# Objective:       get all device id paths
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_spd_devices {
    my ($self, $spd_id) = @_;
    my ($cmd, $id, $status, @devices, @output);
    
    $cmd = "setsp -l" . $spd_id;      
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
        $self->_set_err("Error: AixDevMgr::_get_spd_devices() Unable to get devices for spd_id \'$spd_id\'",
                        @output);
        return (wantarray()) ? () : [];                          
    }

    foreach my $line (@output) {
        next if ($line !~ /hdisk\d+\(/i);
        push @devices, $line =~ /(hdisk\d+)\(/;;
    }
    
    return (wantarray()) ? @devices : \@devices; 

}

################################################################################
# Subroutine Name: _get_spd_id
# Objective:       get sanpath device id
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_spd_id {
    my ($self, $device) = @_;

    my ($cmd, $id, $status, @output);

    $cmd = "cat \/etc\/sppath.conf | egrep \':$device\$\' | awk '{print substr(\$1,5)}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
        $self->_set_err("Error: AixDevMgr::_get_spd_id() Unable to retrieve sanpath id for device \'$device\'", 
                        @output);
    
        return ""                          
    }

    $id = $output[0];

    return $id;

}

################################################################################
# Subroutine Name: _stop_spd_id
# Objective:       stop sanpath device
# Details:         See POD documentation at the end of this file
################################################################################
sub _stop_spd_id {
    my ($self, $spd_id) = @_;

    my ($cmd, $id, $status, @output);
    
    $cmd = "setsp -T -l" . $spd_id;      
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
        $self->_set_err("Error: AixDevMgr::_stop_spd_id() Unable to stop sanpath device id\'$spd_id\'", 
                        @output);
        return($ClientUtil::FALSE);                       
    }

    return($ClientUtil::TRUE);
}

################################################################################
# Subroutine Name: _remove_spd_id
# Objective:       remove sanpath device
# Details:         See POD documentation at the end of this file
################################################################################
sub _remove_spd_id {
    my ($self, $spd_id) = @_;

    my ($cmd, $id, $status, @output);
    
    $cmd = "setsp -x -l" . $spd_id;      
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
        $self->_set_err("Error: AixDevMgr::_remove_spd_id() Unable to remove sanpath device id\'$spd_id\'",
                        @output);
        return($ClientUtil::FALSE);                       
    }

    return($ClientUtil::TRUE);
}

1;

