# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Devices on ESX
## @author hrivera@netapp.com
## @status private
## @pod ./Dev.pod


# Package Name
##############
package ESXDev;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanLog;

use Error;

use vars qw( @ISA );

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: load
# Objective:       sets parameters based on devices current mount state
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub load {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: create_fs
# Objective:       create and return a filesystem object
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub create_fs {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: mount_fs
# Objective:       mount filesystem
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_fs {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if filesystem is mounted
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: is_partitioned
# Objective:       returns ClientUtil::FALSE if partitioned
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub is_partitioned {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

################################################################################
# Subroutine Name: partition
# Objective:       create a partition on the raw device
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub partition {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: destroy_partition
# Objective:       destroy a partition on the raw device
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_partition {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{lun_size};
}

################################################################################
# Subroutine Name: get_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    return ($self->{raw_device_name}, $self->{vmkdisk_name});
}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       returns the raw device name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    # simply call get_name. Needed for Windows compatibility.
    return ($self->get_name());

}

################################################################################
# Subroutine Name: get_mount
# Objective:       returns mount point of device object
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

################################################################################
# Subroutine Name: get_fs
# Objective:       returns fs of device object
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

################################################################################
# Subroutine Name: get_lun
# Objective:       Return lun path
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{lun_path};
}

################################################################################
# Subroutine Name: get_partition
# Objective:       Returns the name of the configured partition.
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

################################################################################
# Subroutine Name: get_partition_size
# Objective:       Returns the size of the specified partition in bytes
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition_size {
    my ($self, $partition_num) = @_;

    $self->_clear_err();

    return "";

}

################################################################################
# Subroutine Name: get_label
# Objective:       Returns the name of the configured partition's label.
#                  NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub get_label {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

################################################################################
# Subroutine Name: refresh_geometry
# Objective:       refresh the geomery on device. Usefull to call after a
#                  resize on controller. NO-OP on this platform
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh_geometry {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: set_device_state
# Objective:       Resets the device state
# Details:         See POD documentation at the end of this file
################################################################################
sub set_device_state {
    my ($self, $state) = @_;

    $self->_clear_err();

    $self->{device_is_free} = $state;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: is_device_free
# Objective:       Returns 0 if the device is not free, 1 if free
# Details:         See POD documentation at the end of this file
################################################################################
sub is_device_free {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{device_is_free};
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new dev object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self, $vmkdisk_name, $filer_name,
        $lun_path, $raw_device_name, $adapter, $dont_query_ostype,
        $protocol, $lun_size, $lun_state, $filer_id, $parent);

    $self = {};

    bless $self, $class;

    $dont_query_ostype = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?vmkdisk_name$/) {
            if (not length($vmkdisk_name = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-vmkdisk_name'");
                return $self;
            }
            next;
        } elsif (/-?filer_name$/) {
            if (not length($filer_name = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-filer_name'");
                return $self;
            }
            next;
        } elsif (/-?lun_path$/) {
            if (not length($lun_path = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-lun_path'");
                return $self;
            }
            next;
        } elsif (/-?raw_device_name$/) {
            if (not length($raw_device_name = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-raw_device_name'");
                return $self;
            }
            next;
        } elsif (/-?adapter$/) {
            if (not length($adapter = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-adapter'");
                return $self;
            }
            next;
        } elsif (/-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-protocol'");
                return $self;
            }
            next;
        } elsif (/-?lun_size$/) {
            if (not length($lun_size = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-lun_size'");
                return $self;
            }
            next;
        } elsif (/-?lun_state$/) {
            if (not length($lun_state = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-lun_state'");
                return $self;
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?filer_id$/) {
            if (not length($filer_id = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-filer_id'");
                return $self;
            }
            next;
        } elsif (/-?dont_query_ostype$/) {
            if (not length($dont_query_ostype = $args{$_})) {
                $self->_set_err("Error: ESXDev::_new()",
                                "Invalid value for argument '-dont_query_ostype'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -vmkdisk_name argument
    unless (defined $vmkdisk_name) {
        $self->_set_err("Error: ESXDev::_new()",
                        "Missing required argument '-vmkdisk_name'");
        return $self;
    }
    # check for existance of required -filer_name argument
    unless (defined $filer_name) {
        $self->_set_err("Error: ESXDev::_new()",
                        "Missing required argument '-filer_name'");
        return $self;
    }
    # check for existance of required -lun_path argument
    unless (defined $lun_path) {
        $self->_set_err("Error: ESXDev::_new()",
                        "Missing required argument '-lun_path'");
        return $self;
    }
    # check for existance of required -raw_device_name argument
    unless (defined $raw_device_name) {
        $self->_set_err("Error: ESXDev::_new()",
                        "Missing required argument '-raw_device_name'");
        return $self;
    }
    # check for existance of required -adapter argument
    unless (defined $adapter) {
        $self->_set_err("Error: ESXDev::_new()",
                        "Missing required argument '-adapter'");
        return $self;
    }
    # check for existance of required -protocol argument
    unless (defined $protocol) {
        $self->_set_err("Error: ESXDev::_new()",
                        "Missing required argument '-protocol'");
        return $self;
    }
    # check for existance of required -lun_size argument
    unless (defined $lun_size) {
        $self->_set_err("Error: ESXDev::_new()",
                        "Missing required argument '-lun_size'");
        return $self;
    }
    # check for existance of required -lun_state argument
    unless (defined $lun_state) {
        $self->_set_err("Error: ESXDev::_new()",
                        "Missing required argument '-lun_state'");
        return $self;
    }
    # check for existance of required -lun_state argument
    unless (defined $filer_id) {
        $self->_set_err("Error: ESXDev::_new()",
                        "Missing required argument '-filer_id'");
        return $self;
    }

    # initilize the object members
    $self->{vmkdisk_name} = $vmkdisk_name;
    $self->{filer_name} = $filer_name;
    $self->{lun_path} = $lun_path;
    $self->{raw_device_name} = $raw_device_name;
    $self->{adapter} = $adapter;
    $self->{protocol} = $protocol;
    $self->{lun_size} = $lun_size;
    $self->{lun_state} = $lun_state;
    $self->{device_is_free} = 1;
    $self->{filer_id} = $filer_id;

    # determine if this lun is vmfs or rdm
    my ($status, @output);
    my $os_type = "";

    if ($dont_query_ostype == 0) {
        if ($parent->get_os_version() =~ /^3\./) {
            # for 3.5 or earlier, sg_raw is unavailable, we need to parse the lun path directly
            if ($lun_path =~ /_vmware_/) {
                $os_type = "vmware";
            } elsif ($lun_path =~ /_(windows_2008|windows_gpt|windows|linux|aix|solaris|hpux)_/) {
                $os_type = $1;
            }
        } else {
            # for 4.0 and later, use sg_raw to query page 18 of the GVA on the scsi device
            my $cmd = "/usr/bin/sg_raw -n -r 1k $raw_device_name C0 00 02 0A 98 0A 18 00 FF 00";
            ($status, @output) = &ClientUtil::execute($parent->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Could not LUN OS Type of ${raw_device_name}");
                return $self;
            }

            foreach my $line (@output) {
                if ($line =~ /\s+(\S+)\s*$/) {
                    $os_type .= $1;
                }
            }
        }
    }

    if ($os_type =~ /vmware/) {
        $self->{lun_type} = "vmfs";
        $self->{hosttype} = "vmware";
    } elsif ($os_type =~ /(windows_2008|windows_gpt|windows|linux|aix|solaris|hpux)/) {
        $self->{lun_type} = "rdm";
        $self->{hosttype} = $1;
    } else {
        # default if lun path cannot be parsed
        $self->{lun_type} = "other";
        $self->{hosttype} = "unknown";
    }

    return $self;
}

1;
