# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to DevMgr for ESX host
## @author hrivera@netapp.com
## @status private
## @pod ./DevMgr.pod

# Package Name
##############
package San::SanClient::DevMgr::ESXDevMgr;

use strict;

# Package Requirements
######################
use POSIX;
use San::SanClient::ClientUtil;
use San::SanClient::DevMgr;
use San::SanClient::DevMgr::ESXDev;

use Error;
use San::SanLog;

use vars qw( @ISA );

@ISA = qw(Error SanLog DevMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_device
# Objective:       return a device object based on the device name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device {
    my ($self, $name) = @_;

    $self->_clear_err();

    # scan through the device lists until we find the device
    my @device_list = ($self->get_devices('vmfs'), $self->get_devices('rdm'));

    foreach my $device (@device_list) {
        my ($raw_name, $vmkdisk_name) = $device->get_name();
        if (($raw_name eq $name) or ($vmkdisk_name eq $name)) {
            return $device;
        }
    }

    $self->_set_err("Error: ESXDevMgr::get_device() ".
                    "Could not find device: '$name'.");
    return "";
}

################################################################################
# Subroutine Name: get_devices
# Objective:       return all currently configured device objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self, $device_type) = @_;

    $self->_clear_err();

    my (@device_list);

    if ($device_type =~ /vmfs/i) {
        @device_list = @{$self->{vmfs_devices}};
    } elsif ($device_type =~ /rdm/i) {
        @device_list = @{$self->{rdm_devices}};
    } elsif ($device_type =~ /other/i) {
        @device_list = @{$self->{other_devices}};
    } else {
        $self->_set_err("Error: ESXDevMgr::get_devices() ".
                        "bad device_type: '$device_type'");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @device_list : \@device_list;
}

################################################################################
# Subroutine Name: create_device
# Objective:       create a device object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_device {
    my ($self, $lun) = @_;

    my ($filer_name,
        $lun_path,
        $raw_device_name,
        $adapter,
        $protocol,
        $lun_size_in_bytes,
        $lun_state,
        $vmkdisk_name);

    $self->_clear_err();

    my $lun_path_regex_filter = $self->{lun_path_regex_filter};

    # if $lun contains the header of sanlun lun show, just return
    if ($lun =~ /lun-pathname/i or $lun =~ /lun state/i) {
        return $ClientUtil::TRUE;
    }

    # if the lunpath contains the word "boot", ignore it
    if ($lun =~ /boot/i and $self->{all_devices} == 0) {
        return $ClientUtil::TRUE;
    }

    # if the lunpath does not match the lun path regex filter, ignore it
    if ($lun !~ $lun_path_regex_filter and $self->{all_devices} == 0) {
        return $ClientUtil::TRUE;
    }

    # if the lunpath contains the string "<unknown>" then it is an orphaned lun
    # meaning that it was deleted from the filer. Skip these.
    if ($lun =~ /\<unknown\>/i) {
        return $ClientUtil::TRUE;
    }

    # if the lun doesnt match one of the internally stored filer names to filter against
    # then skip it
    if (scalar(@{$self->{filers}}) and $self->{all_devices} == 0) {
        my $matched = 0;
        foreach my $filer (@{$self->{filers}}) {
            if ($lun =~ /^\s*$filer\:/) {
                # then this lun matched one of the specified filers, break out
                $matched = 1;
                last;
            }
        }
        # if this lun didnt match any of the filers
        if ($matched == 0) {
            # then skip the lun
            return $ClientUtil::TRUE;
        }
    }

    # if the adapter of the lun does not match the adapter stored in the
    # SanClient object, then ignore the lun. This will protect us against
    # creating devices for software iscsi luns when we are supposed to do
    # hardware and vice versa.
    my @adapters = @{$self->{parent}->{vmhba_adapter}};
    my $lun_is_valid = 0;
    foreach my $vmhba_adapter (@adapters) {
        # if this lun matches one of the adapters, then let it through
        if ($lun =~ /$vmhba_adapter/) {
            $lun_is_valid = 1;
            last;
        }
    }

    # if this lun did not match one of the valid adapters, ignore it
    if ($lun_is_valid != 1 and $self->{all_devices} == 0) {
        return $ClientUtil::TRUE;
    }

    # if $lun has reached this point, it has passed all the filter tests.
    # parse out all the information from the line and build an ESXDev object.
    ($filer_name,
     $lun_path,
     $raw_device_name,
     $adapter,
     $protocol,
     $lun_size_in_bytes,
     $lun_state,
     $vmkdisk_name) = ($lun =~ /^\s*(\S+?):\s*(\S+?)\s*(\S+?)\s*(\S+?)\s*(\S+?)\s*\S+\s*\((.+?)\)\s*(\S+?)\s*(vmhba\S+|naa\S+)/);

    my $filer_id = $self->_convert_to_filer_id($filer_name);

    # dont query the ostype when asking for all devices
    my $dont_query_ostype = 0;
    if ($self->{all_devices} == 1) {
        $dont_query_ostype = 1;
    }

    # create the ESXDev object
    my $dev_obj = ESXDev->_new(-vmkdisk_name => $vmkdisk_name,
                               -filer_name => $filer_name,
                               -lun_path => $lun_path,
                               -raw_device_name => $raw_device_name,
                               -adapter => $adapter,
                               -protocol => $protocol,
                               -lun_size => $lun_size_in_bytes,
                               -lun_state => $lun_state,
                               -filer_id => $filer_id,
                               -parent => $self->{parent},
                               -dont_query_ostype => $dont_query_ostype);
    if ($dev_obj->error()) {
        $self->_set_err($dev_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # push the object onto the vmfs device array or the rdm device array.
    if ($dev_obj->{lun_type} eq "vmfs") {
        push(@{$self->{vmfs_devices}}, $dev_obj);
        $self->{vmfs_device_count} += 1;
    } elsif ($dev_obj->{lun_type} eq "rdm") {
        push(@{$self->{rdm_devices}}, $dev_obj);
        $self->{rdm_device_count} += 1;
    } else {
        push(@{$self->{other_devices}}, $dev_obj);
        $self->{other_device_count} += 1;
    }

    return $ClientUtil::TRUE;
}    

################################################################################
# Subroutine Name: get_dev_id
# Objective:       returns unique device name(with full path) of passed device's
#                  short name to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_id {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: get_dev_name
# Objective:       returns device short name of passed device's unique name(long name)
#                  to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_name {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: cleanup
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup {
    my ($self) = @_;

    $self->_clear_err();

    $self->_set_err("Error: ESXDevMgr::cleanup() function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($device_type, $num_devices, $hosttype, $hostname);
    my (@filers) = ();
    my (@device_list) = ();
    my $lun_path_regex_filter = qr/.*/;
    my $protocol = qr/.*/;

    # parse args
    foreach (keys %args) {
        if (/-?device_type/) {
            if (not length($device_type = $args{$_})) {
                $self->_set_err("Error: ESXDevMgr::get_free_devices() ".
                                "Invalid value for argument " .
                                "'-device_type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?num_devices/) {
            if (not length($num_devices = $args{$_})) {
                $self->_set_err("Error: ESXDevMgr::get_free_devices() ".
                                "Invalid value for argument " .
                                "'-num_devices'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?hosttype/) {
            if (not length($hosttype = $args{$_})) {
                $self->_set_err("Error: ESXDevMgr::get_free_devices() ".
                                "Invalid value for argument " .
                                "'-hosttype'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?filers/) {
            if (ref($args{$_}) ne "ARRAY" or not scalar(@filers = @{$args{$_}})) {
                $self->_set_err("Error: ESXDevMgr::get_free_devices() ".
                                "Invalid value for argument " .
                                "'-filer_name'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?hostname/) {
            if (not length($hostname = $args{$_})) {
                $self->_set_err("Error: ESXDevMgr::get_free_devices() ".
                                "Invalid value for argument " .
                                "'-hostname'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?lun_path_regex_filter/) {
            if (not length($lun_path_regex_filter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-lun_path_regex_filter'");
                return $self;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-protocol'");
                return $self;
            }
            next;
        }
    }

    # check for required arguments
    if (not defined($device_type) or $device_type !~ /vmfs|rdm/i) {
        $self->_set_err("Error: ESXDevMgr::get_free_devices() ".
                        "Missing required parameter '-device_type'" .
                        " Use 'vmfs' or 'rdm'.");
        return (wantarray()) ? () : [];
    }
    if (not defined($num_devices)) {
        # default value is -1, meaning: "all available devices"
        $num_devices = -1;
    }
    if (defined($hosttype) and
        $hosttype !~ /^(vmware|linux|solaris|windows_2008|windows_gpt|windows|hpux|netware)$/i) {
        $self->_set_err("Invalid value for parameter'-hosttype'");
        return (wantarray()) ? () : [];
    }

    # first thing to do: get the current device list of vmfs or rdm
    @device_list = $self->get_devices($device_type);

    # We are now going to filter all devices from @device_list that do not
    # match the following:
    # (1) the device is free.
    # (2) $hostname
    # (3) filers from @filers
    # (4) protocol
    # (5) $hosttype
    # (6) no boot luns
    # (7) $num_devices

    my @filtered_device_list = ();

    foreach my $device (@device_list) {
        # (1) filter out devices that are not free
        next if ($device->{device_is_free} != 1);

        # (2) filter out devices that do not match $hostname
        if (defined($hostname)) {
            next if ($device->{lun_path} !~ /_($hostname)_/);
        }

        if (defined($protocol)) {
            next if ($device->{protocol} !~ $protocol);
        }

        # filter out by the specified lun path if there is one specified
        if (defined($lun_path_regex_filter)) {
            next if ($device->{lun_path} !~ $lun_path_regex_filter);
        }

        # (3) iterate through each filer in @filers and only let the ones that match
        # a filer pass.
        my $matched = 0;
        foreach my $filer_name (@filers) {
            $matched = 1 if ($device->{filer_name} =~ /^($filer_name)$/);
        }
        next if ($matched != 1);

        # (4) filter out devices that do not match the specified protocol
        my $protocol = $self->{parent}->get_protocol();
        next if ($device->{protocol} !~ /$protocol/i);

        # (5) filter out devices that do not match $hosttype
        # first fix the hosttypes
        if (defined($hosttype)) {
            $hosttype =~ s/linux-vmware/vmware/;
            $hosttype =~ s/unix-//;
            $hosttype =~ s/linux.*/linux/;  # Drop linux variations
            $hosttype =~ s/.*solaris.*/solaris/;
            $hosttype =~ s/.*hpux.*/hpux/;
            $hosttype =~ s/.*netware.*/netware/;
            next if ($device->{hosttype} ne $hosttype);
        }

        # (6) filter out devices that are boot luns
        next if ($device->{lun_path} =~ /boot/i);

        # this device passed (1) -> (6), add this device to the filtered list
        push(@filtered_device_list , $device);
    }

    # The next task is to make @filtered_device_list balanced.
    # filer_hash will be a hash of filer name keys. Each key will have an
    # array of device objects that belong to that filer.
    my %filer_hash = ();

    foreach my $device (@filtered_device_list) {
        # add this device into that filer's key array
        push(@{$filer_hash{$device->{filer_name}}}, $device);
    }


    # now balance out the device list so that the final list we return
    # to the caller has alternating devices from different filers/controllers.
    my @balanced_device_list = ();
    while(scalar(keys %filer_hash) > 0) {
        # for every filer key, shift off one of the devices and push it into the final array
        foreach my $filer_array (keys %filer_hash) {
            push(@balanced_device_list, shift(@{$filer_hash{$filer_array}}));
            if (scalar(@{$filer_hash{$filer_array}}) == 0) {
                delete $filer_hash{$filer_array};
            }
        }
    }

    # Since the ordering for @balanced_device_list is consistent across
    # multiple ESX Hosts, it may occur where all devices returned are heavily
    # biased towards one storage controller. Therefore we provide the ability
    # to reverse the @balanced_device_list if the value passed into $self->{_balance}
    # is an odd number.
    if (($self->{_balance} % 2) == 1) {
        # reverse the device list
        @balanced_device_list = reverse @balanced_device_list;
    }

    # (7) if the calling function requested a specific amount of devices,
    # pull some out of the balanced device list using some clever math.
    # store the final list in @final_device_list
    my @final_device_list = ();
    if ($num_devices != -1) {
        my $total_devices = scalar(@balanced_device_list);
        my $difference = $total_devices - $num_devices;
        if ($difference < 0) {
            $self->_set_err("Error: ESXDevMgr::get_free_devices() Requested ".
                            "($num_devices) devices and only " .
                            "($total_devices) devices are available.");
            return (wantarray()) ? () : [];
        }

        if ($difference == 0) {
            @final_device_list = @balanced_device_list;
        } else {
            # pull out $difference devices from @balanced_device_list
            for (my $count = 0; $count < $num_devices; $count++) {
                my $next_device = $self->{_next_device} % $total_devices;
                $self->{_next_device}++;

                # mark this device as not free
                ${balanced_device_list[$next_device]}->{device_is_free} = 0;
                push(@final_device_list, $balanced_device_list[$next_device]);
            }
        }
    } else {
        # otherwise return all available devices
        @final_device_list = @balanced_device_list;
    }

    # if our final device list is empty at this point, then error out
    if (scalar(@final_device_list) < 1) {
        $self->_set_err("Error: ESXDevMgr::get_free_devices() ".
                        "No $device_type devices available!");
        return (wantarray()) ? () : [];
    }

    # if device type was vmfs, reset the next device attribute, also reverse the final dev list
    # so that we at least have a balance between vmfs and rdm.
    if ($device_type =~ /vmfs/i) {
        $self->{_next_device} = 0;
        @final_device_list = reverse @final_device_list;
    }

    # now return the device list
    return (wantarray()) ? @final_device_list : \@final_device_list;
}


################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new devices object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $balance, $self, $all_devices, $lun_path_regex_filter);
    my ($filers);
    $self = {};

    bless $self, $class;

    $self->sanlog(-msg => "Creating an ESXDevMgr Object.");

    # by default device restrictions are enabled
    $all_devices = 0;

    # default lun path filter to all LUNs
    $lun_path_regex_filter = qr/.*/i;
    $filers = [];

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: ESXDevMgr::_new() Invalid value for ".
                                "argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?balance/) {
            if (not length($balance = $args{$_})) {
                $self->_set_err("Error: ESXDevMgr::_new() Invalid value for ".
                                "argument '-balance'");
                return $self;
            }
            next;
        } elsif (/-?all_devices/) {
            if (not length($all_devices = $args{$_})) {
                $self->_set_err("Error: ESXDevMgr::_new() Invalid value for ".
                                "argument '-all_devices'");
                return $self;
            }
            next;
        } elsif (/-?lun_path_regex_filter/) {
            if (not length($lun_path_regex_filter = $args{$_})) {
                $self->_set_err("Invalid value for ".
                                "argument '-lun_path_regex_filter'");
                return $self;
            }
            next;
        } elsif (/-?filers/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Invalid value for ".
                                "argument '-filers'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    if (defined($parent)) {
        unless (ref($parent) =~ "SanClient") {
            $self->_set_err("Error: ESXDevMgr::_new()",
                            "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: ESXDevMgr::_new()",
                        "Missing required argument '-parent'");
        return $self;
    }

    # default $balance to 0 if not defined.
    $balance = 0 if (not defined($balance));

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{vmfs_devices} = [];
    $self->{rdm_devices} = [];
    $self->{other_devices} = [];
    $self->{vmfs_device_count} = 0;
    $self->{rdm_device_count} = 0;
    $self->{other_device_count} = 0;
    $self->{_next_device} = 0;
    $self->{_balance} = $balance;
    $self->{_filer_id_hash} = {};
    $self->{_filer_id_hash}->{id_count} = 0;
    $self->{all_devices} = $all_devices;
    $self->{lun_path_regex_filter} = $lun_path_regex_filter;
    $self->{filers} = $filers;

    # call sanlun lun show
    my @sanlun_output = $self->{sanlun}->lun_show();
    if ($self->{sanlun}->error()) {
        $self->_set_err($self->{sanlun}->errmsg);
        return $self;
    }

    # create a device object for each lun. create_device() will filter out
    # any luns that are not part of the test.
    foreach my $lun (@sanlun_output) {
        $self->create_device($lun);
        if ($self->error()) {
            return $self;
        }
    }

    # dont print out the summary when all devices is enabled
    if ($all_devices == 1) {
        return $self;
    }

    # obtain the lun paths for our vmfs and rdm luns.
    my @vmfs_lun_paths =
        map {
            $_->{vmkdisk_name} . "\t\t" . $_->{filer_name} . ": " . $_->{lun_path} . "\n";
        } $self->get_devices("vmfs");
    my @rdm_lun_paths =
        map {
            $_->{vmkdisk_name} . "\t\t" . $_->{filer_name} . ": " . $_->{lun_path} . "\n";
        } $self->get_devices("rdm");

    # print out which luns we will be using.
    $self->sanlog(-msg => "We will be considering the following VMFS luns:\n @vmfs_lun_paths");
    $self->sanlog(-msg => "We will be considering the following RDM luns:\n @rdm_lun_paths");
    $self->sanlog(-msg => "Some of these VMFS and RDM luns may be data luns for other ESX hosts ".
                          "if VM_CONFIG_FOR_VMOTION is set to 1");

    return $self;
}

################################################################################
# Subroutine Name: _convert_to_filer_id()
# Objective:       Given a filer name, convert to a filer id
# Details:         See POD documentation at the end of this file
################################################################################
sub _convert_to_filer_id {
    my ($self, $filer_name) = @_;

    # if this filer does not have an id assigned to it yet,
    # generate one and  increment the id count

    if (not exists($self->{_filer_id_hash}->{filers}->{$filer_name})) {
        $self->{_filer_id_hash}->{filers}->{$filer_name} =
            $self->{_filer_id_hash}->{id_count};
        $self->{_filer_id_hash}->{id_count} += 1;
    }

    # return the id
    return $self->{_filer_id_hash}->{filers}->{$filer_name};
}

1;
