# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Devices on HPUX
## @author lippe@netapp.com, latif@netapp.com
## @status private
## @pod ./Dev.pod


# Package Name
##############
package HPUXDev;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use Error;

use vars qw($_NAME_CNT @ISA);

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: load
# Objective:       sets parameters based on devices current mount state
# Details:         See POD documentation at the end of this file
################################################################################
sub load {
    my ($self) = @_;

    my ($cmd, $name, $partition, $phash, $status, $type, @output);

    $self->_clear_err();

    $self->sanlog(-msg => "Populating $self->{name} data object");
  
    # save response to mkfs prompt
    $cmd = "/usr/sbin/mount | grep /dev/dsk/$self->{name} | awk '{print \$1}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);  
    if ($status > 0) {
        $self->_set_err("Error: HPUXDev::load Mount command failed", @output); 
        return $ClientUtil::FALSE;                            
    }

    $self->{mount} = $output[0];

    # save response to mkfs prompt
    $cmd = "/usr/sbin/fstyp /dev/dsk/$self->{name}";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);  
    if ($status > 0) {
    $self->_set_err("Error: HPUXDev::load Fstyp command failed", @output); 
        return $ClientUtil::FALSE;                            
    }

    $self->{fs} = $output[0];

    $self->sanlog(-msg => "Loading $self->{name}  Mount: $self->{mount} FS: $self->{fs}");

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: create_fs
# Objective:       create and return a filesystem object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_fs {
    my ($self, %args) = @_;

    my ($cmd, @output, $partition, $phash, $status, $type);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?type$/) {
            $type = $args{$_};
            next;
        }
    } 

    # check for existance of required -type argument
    if (not defined $type or $type eq "") {
        $type = "vxfs";
    }
    
    # create the filesystem
    $cmd = "/usr/sbin/mkfs -F $type -o largefiles /dev/rdsk/$self->{name}";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
         $self->_set_err("Error: HPUXDev::create_fs() Unable to create $type filesystem on dev '$self->{name}'",
                         @output);
         return $ClientUtil::FALSE;                            
    } 

    $self->{fs} = $type;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: mount_fs
# Objective:       mount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_fs {
    my ($self, %args) = @_;

    my ($cmd, $directio, $options, $type, $vgname, $mount);
    my ($name, $status, @output);

    # Get attributes 
    $directio = 0;
    $name = $self->get_name();
    $options = "";  # option default 
    $type = $self->get_fs();
    $mount = undef;
    
    # check if volume is unformatted
    unless (length $type) {
        $self->_set_err("Error: HPUXDev::mount_fs()",
                        "No filesystem detected");
        return $ClientUtil::FALSE; 
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?path$/) {
            if (not length($mount = $args{$_})) {
                $self->_set_err("Error: HPUXDev::mount_fs()",
                                "Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }        
            next;
        } elsif (/-?directio$/) {
            if (not length($directio = $args{$_}) or $args{$_} !~ /0|1/) {
                $self->_set_err("Error: HPUXDev::mount_fs()",
                                "Invalid value for argument '-directio'");
                return $ClientUtil::FALSE;
            }        
            next;
        }
    }

    # check if the volume is already mounted
    if ($self->is_mounted()){
       $self->_set_err( "Error: HPUXDev::mount_fs() device '$name' is already mounted");                    
       return $ClientUtil::FALSE;                            
    }
    
    # Use default path if not defined
    unless (defined $mount) {
        $mount = "/mnt/qa/$name/";
        $self->sanlog(-msg => "using default mount pt '$mount'");      
    } else {
        $self->sanlog(-msg => "using mount pt '$mount'"); 
    }
    
    $self->{mount} = $mount;

    # Create the mount point and open up dir privs 
    $cmd = "mkdir -p $mount; chmod 777 $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: HPUXDev::mount_fs() Unable to create mount ".
                        "point directory for volume '$name'", @output);
        return $ClientUtil::FALSE;                            
    } 

    # Add directio option if requested and it was not
    # passed in as part of the options string
    if ($directio and $options !~ /convosync=direct/i) {        
        $options .= (defined $options and length $options)
        ? ",convosync=direct"
        : "convosync=direct";
    }

    # Remove option param if user passed it in as part of the 
    # options string. Lib should add the tag.
    $options =~ s/-o //g;

    # Mount the volume's filesystem
    $cmd = ($options)
      ? "/usr/sbin/mount -F $type -o $options /dev/dsk/$name $mount"
      : "/usr/sbin/mount -F $type /dev/dsk/$name $mount";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: HPUXDev::mount_fs() Unable to mount volume filesystem",
                        @output);
        return $ClientUtil::FALSE;                            
    } 

=comment
    # For now lets not add the entry to the vfstab file
    # Get handle to client object 

    # Get handle to client object 
    $self->{client}=$self->{parent}{parent};    

    $self->{client}->mount_entry_add(
                     -device_mount=>"/dev/dsk/$name",
                     -mount_pt=>$mount,
                     -fs_type=>$type,
                     -mount_options=>"log,largefiles");

    if ($self->{client}->error) {
        $self->_set_err($self->{client}->errmsg());
        return $ClientUtil::FALSE;
    }   

=cut

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, %args) = @_;

    my ($cmd, $type, $vgname, $fs, $mount, $name, $status);
    my ($force, $option, @output);

    $self->_clear_err();

    $option = "";
    
    # parse hash
    foreach (keys %args) {
        if (/-?force$/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: HPUXDev::unmount_fs()",
                                "Invalid value for argument '-force'");
                return $self;
            }        
            next;
        }
    }  

    $option = "-f" if (defined $force);

    # Get attributes 
    $name = $self->get_name();
    $mount = $self->get_mount();
    $fs = $self->get_fs();

    # check if device contains a filesystem
    unless (length $fs) {
        $self->_set_err("Error: HPUXDev::unmount_fs()",
                        "No filesystem detected");
        return $ClientUtil::FALSE; 
    }

    # check if volume is mounted
    unless (length $mount) {
        $self->_set_err("Error: HPUXDev::unmount_fs()",
                        "Cannot unmount a filesystem that is not mounted");
        return $ClientUtil::FALSE; 
    } 

    # Unmount the volume's filesystem
    $cmd = "/usr/sbin/umount $option /dev/dsk/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: HPUXDev::unmount_fs() Unable to unmount volume's filesystem",
                        @output);
        return $ClientUtil::FALSE;                            
    } 

    # Remove mount pt metadata from object
    $self->{mount} = "";

=comment
    
    # Add mount entry to vfstab file has been commented out,
    # so there should be no need to delete it.

    # Clear vfstab mount entry
    # client = lvm <-- vg <-- vol
    $self->{client}=$self->{parent}{parent};

    $self->{client}->mount_entry_delete("/dev/dsk/$name");
    if ($self->{client}->error) {
        $self->_set_error($self->{client}->errmsg());
        return $ClientUtil::FALSE;
    }   

=cut

    $self->sanlog(-msg => "removing mount point '$mount'");    
 
    # Remove directory head
    my ($base, $parent) = fileparse($mount);
    $cmd = "rm -rf $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXDev::unmount_fs() Unable to remove mount point '$mount'", @output);
    } 

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if filesystem is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{mount} and length $self->{mount}) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: is_partitioned
# Objective:       returns ClientUtil::TRUE if partitioned
# Details:         See POD documentation at the end of this file
################################################################################
sub is_partitioned {
    my ($self) = @_;
    
    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: partition
# Objective:       create a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub partition {
    my ($self, %args) = @_;

    $self->_clear_err();
    
    # noop on HPUX

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: destroy_partition
# Objective:       destroy a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_partition {
    my ($self, %args) = @_;

    $self->_clear_err();
    
    # noop on HPUX

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_uuid
# Objective:       Returns the UUID of the configured device.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self) = @_;
    
    $self->_clear_err();

    # This function is only valid on HPUX
    return undef;
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{size})) {
        return $self->{size};
    } else {
        $self->_set_err("Error: Vol::get_size() Vol object is missing size");
        return undef;
    }

}

################################################################################
# Subroutine Name: get_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined($self->{name}) and length($self->{name})) {
        return $self->{name};
    } else {
        $self->_set_err("Error: HPUXDev::get_name() Dev object is missing name");
        return undef;
    }
}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       returns the raw device name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    # simply call get_name. Needed for Windows compatibility.
    return ($self->get_name());

}

################################################################################
# Subroutine Name: get_mount
# Objective:       returns mount point of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined($self->{mount}) and length($self->{mount})) {
        return $self->{mount};
    } else {
        $self->_set_err("Error: HPUXDev::get_mount() Dev object is missing mount");
        return undef;
    }
}

################################################################################
# Subroutine Name: get_fs
# Objective:       returns fstype of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined($self->{fs}) and length($self->{fs})) {
        return $self->{fs};
    } else {
        $self->_set_err("Error: HPUXDev::get_fs() Dev object missing fstype");
        return undef;
    }
}

################################################################################
# Subroutine Name: get_lun
# Objective:       Return lun name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{lun}) and length($self->{lun})) {
        return $self->{lun};
    } else {
        $self->_set_err("Error: HPUXDev::get_lun() Dev object is missing lun name");
        return undef;
    }
}

################################################################################
# Subroutine Name: get_partition
# Objective:       Returns the name of the configured partition.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition {
    my ($self) = @_;
    
    $self->_clear_err();

    return undef;
}

################################################################################
# Subroutine Name: get_partition_size
# Objective:       Returns the name size of a particular partition in bytes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition_size {
    my ($self, $partition_num) = @_;

    $self->_clear_err();

    # partitions are not support on HP, simply
    # return the entire size of the device. Error will
    # be set by get_size if this fails.
    my $size = $self->get_size();

    return $size

}

################################################################################
# Subroutine Name: get_label
# Details:         See POD documentation at the end of this file
################################################################################
sub get_label {
    my ($self) = @_;
    
    $self->_clear_err();

    return undef;
}

################################################################################
# Subroutine Name: refresh_geometry
# Objective:       refresh the geomery on device. Usefull to call after a 
#                  resize on controller. NO-OP on this platform
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh_geometry {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new dev object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $conn, $log, $name, $parent, $psize, $self, $status);
    my (@output);

    $self = {};

    bless $self, $class;    

    # parse arguments
    foreach (keys %args) {
        if (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: HPUXDev::_new()",
                                "Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXDev::_new()",
                                "Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/){
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: HPUXDev::_new()",
                                "Missing value for argument '-parent'");
                return $self;
            }
            next;
        } 
    }  
    
    # check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: HPUXDev::_new()",
                        "Missing required argument '-name'");
        return $self; 
    }

    # check for existance of required -parent argument
    unless (defined $parent) {
        $self->_set_err("Error: HPUXDev::_new()",
                        "Missing required argument '-parent'");
        return $self; 
    }

    # check for existance of required -conn argument
    if (grep ( /^-?conn$/, (keys %args) )) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: HPUXDev::_new()",
                            "Invalid value for argument '-conn'");
            return $self; 
        }    
    } else {
        $self->_set_err("Error: HPUXDev::_new()",
                        "Missing required argument '-conn'");
        return $self; 
    }

    # initilize the object members
    $self->{conn} = $conn;
    $self->{parent} = $parent;
    $self->{name} = $name;

    # If the device is a CTD device, get lun path info, otherwise leave it blank
    if ($self->{name} =~ /c\d+t\d+d\d+$/) {
        # get filer lun associated with physical dev
        $self->{lun} = $parent->{sanlun}->get_lun_path(-device => $name);
        if ($parent->{sanlun}->error) {
            $self->_set_err($parent->{sanlun}->errmsg());
            return $self; 
        }
    } else {
        $self->{lun} = "";
    }

    $self->{fs} = "";
    $self->{mount} = "";
    $self->{size} = $self->_init_size();

    return $self;
}

################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing disk
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self) = @_;
    
    my ($cmd, $size, $name, @output, $status);
    
    $size = "";
    $name = "/dev/rdsk/" . $self->{name};

    $cmd = "/usr/sbin/diskinfo -b $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    # check status
    if ($status > 0) {
        $self->_set_err("Error: HPUXDev::_init_size() Unable to retrieve size (bytes) for disk \'$name\'", @output);
        return "";
    }

    # get the size in bytes 
    $size = &ClientUtil::convert_to_bytes($output[0]  . "k");
    
    return $size;    
  
}

1;
