# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to DevMgr for HPUX host
## @author lippe@netapp.com knopp@netapp.com latif@netapp.com
## @status private
## @pod ./DevMgr.pod

# Package Name
##############
package San::SanClient::DevMgr::HPUXDevMgr;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::DevMgr;
use San::SanClient::DevMgr::HPUXDev;
use File::Basename;

use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog DevMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_device
# Objective:       return a device object based on the device name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device {
    my ($self, $name) = @_;

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: HPUXDevMgr::get_device() name is a required argument");
        return {};
    }

    foreach my $obj ($self->get_devices) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: HPUXDevMgr::get_device() Unable to find device '$name'");

    return {};
}

################################################################################
# Subroutine Name: get_devices
# Objective:       return all currently configured device objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;
    my (@devices);

    $self->_clear_err();

    if (@{$self->{devices}}) {
        @devices = @{$self->{devices}};
    } else {
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: create_device
# Objective:       create a device object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_device {
    my ($self, %args) = @_;
    my ($devObj);

    $self->_clear_err();

    $devObj = HPUXDev->_new(-parent => $self,
                            -conn   => $self->{conn},
                            %args);

    if ($devObj->error) {
        $self->_set_err($devObj->errmsg());
        return {};
    }

    push @{$self->{devices}}, $devObj;

    return ($devObj);
}

################################################################################
# Subroutine Name: get_dev_id
# Objective:       returns unique device name(with full path) of passed device's
#                  short name to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_id {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: get_dev_name
# Objective:       returns device short name of passed device's unique name(long name)
#                  to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_name {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: cleanup
# Objective:       remove device objects and remove mount point entries
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup {
    my ($self) = @_;
    my ($devObj, $fail_cnt);

    $self->_clear_err();

    # remove mount point entry for each device passed in
    foreach $devObj ($self->get_devices) {
        $self->sanlog(-msg => 
                      "Cleaning up mount points for device: $devObj->{name}");
        $devObj->unmount_fs() if $devObj->is_mounted;
        if ($devObj->error) {
            $self->_set_err($devObj->errmsg);
            $fail_cnt++;
        }
    }

    $self->{devices} = [];

    return ($fail_cnt == 0) ? $ClientUtil::TRUE : $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       retrieve list of free host volumes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;
    my ($bCnt,    $cmd,     $cnt,       $device,      $filer,         $filername,   $lun,
        $lun_tmp, $refresh, $status,    $total,       $type,          $vol,         %hDevs,
        @attr,    @devices, @output,    @used,        $done,          @split,       $dev,
        $line,    @array,   @tmp_array, @tmp_devices, @sanlun_output, @cur_devices, @new_devices,
        $filers,  $protocol);

    $type  = ".*";
    $total = 0;
    @{$filers} = ();

    $self->_clear_err();

    # Asssume user supplied -cnt argument
    $bCnt    = 1;
    $refresh = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Error: HPUXDevMgr::get_free_devices() Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: HPUXDevMgr::get_free_devices() Invalid value for argument '-type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Error: HPUXDevMgr::get_free_devices() Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Error: HPUXDevMgr::get_free_devices() Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Error: HPUXDevMgr::get_free_devices() Invalid value for argument '-vol'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Error: HPUXDevMgr::get_free_devices() Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: HPUXDevMgr::get_free_devices() Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: HPUXDevMgr::get_free_devices() Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Check if user wants free devices from a specified filer,volume, or lun
    if (defined $filer or defined $vol or defined $lun or defined $protocol or scalar(@{$filers}) != 0 or $refresh == 1) {
        # Check if a device to filer map 'dfmap' has already been created or
        # refresh it if requested.
        if (not defined $self->{dfmap} or $refresh == 1) {
            $self->sanlog(-msg => "refreshing device to filer, vol and lun map");
            @output = $self->{sanlun}->lun_show();
            if ($self->{sanlun}->error) {
                $self->_set_err($self->{sanlun}->errmsg());
                return $ClientUtil::FALSE;
            }

            foreach my $line (@output) {
                # skip first line (header)
                if ($line =~ /lun-pathname|WARNING/i) {
                    next;
                }

                # depending on filername length sanlun may not have display blank space before name
                my ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp);
                if ($line  =~ /^\s*(\S+?):?\s+\/vol\/(\w+)\/(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+/) {
                    # The line is in the following format
                    #sanmgr_vserv_0:  /vol/vol_1/i_5_0  /dev/rdsk/c5t0d0   lan2     iSCSI          3g (3221225472)     GOOD
                    #sanmgr_vserv_1:  /vol/vol_1/f_5_0  /dev/rdsk/c19t0d0  td0      FCP            3g (3221225472)     GOOD

                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp) = ($1, $2, $3, $4, lc($5));
                } elsif ($line =~ /\s{0}(\S+):\s+\/vol\/(\w+)\/(\S+)\s+\/dev\/rdsk\/(\w+)/) {
                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp) = ($1, $2, $3, $4);
                }

                if ($device_tmp eq "" or $device_tmp =~ /^\s*$/) {
                    $self->sanlog(-msg => "Output from sanlun is incorrect:\n$line\n");
                    next;
                }

                my $dev_tmp = basename($device_tmp);

                $hDevs{$dev_tmp} = {filer    => $filer_tmp,
                                    vol      => $vol_tmp,
                                    lun      => $lun_tmp,
                                    protocol => $protocol_tmp};

                $self->sanlog(-msg => "Device ($dev_tmp) information:\n\tFiler: $filer_tmp\n\tVolume: $vol_tmp\n" .
                                      "\tLUN: $lun_tmp\n\tProtocol: $protocol_tmp\n",
                              -msg_level => 3);

            }

            # save map for next call of function
            $self->{dfmap} = \%hDevs;
        }
    }

    # If not -cnt then assume all free disks are wanted
    unless (defined $cnt and length($cnt) and $cnt > 0) {
        $cnt  = 10000;
        $bCnt = 0;
    }

    my @hDevices = ();
    if (defined $self->{parent}->{mp} and $self->{parent}->{mp_type} !~ /none/i and
        $self->{parent}->{mp}->{primary_paths} == 1) {
        # If there is an MP object defined, use the get_mp_devices call to get the list of MP devices
        # to traverse through when filtering
        @hDevices = $self->{parent}->{mp}->get_mp_devices(-conn => $self->{parent}->get_connection());
        if ($self->{parent}->{mp}->error()) {
            my $error = join("\n", $self->{parent}->{mp}->errmsg());
            $self->_set_err("Error: HPUXDevMgr::get_free_devices() Unable to get list of devices from MP stack " . $error );
            return (wantarray()) ? () : [];
        }
    } else {
        # Grab a list of the devices visible from the host
        @hDevices = $self->{parent}->_get_devices();
        if ($self->{parent}->error()) {
            my $error = join("\n", $self->{parent}->errmsg());
            $self->_set_err("Error: HPUXDevMgr::get_free_devices() Unable to get list of devices from system " . $error );
            return (wantarray()) ? () : [];
        }
    }

    # get boot device (to exclude)
    my $boot_device = $self->{parent}->_get_boot_device();
    # Get the basename of the boot device, no need for the /dev/rdsk path
    $boot_device = basename($boot_device);

    foreach my $dev (@hDevices) {
        if ($dev eq "") {
            $self->sanlog(-msg => "HPUXDevMgr::get_free_devices() - Unable to locate " .
                                  "$dev within hDevices list from HPUXHost::_get_devices()");
            next;
        }

        # run pvdisplay to make sure this disk is free
        $cmd = "/usr/sbin/pvdisplay " . $dev->{dev_path} . "/" . $dev->{name_ctd};
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        if (grep(/Cannot display physical volume/i, @output) == 0 and grep(/PV Status\s+unavailable/i, @output) == 0) {
            next;
        }

        # skip the boot device
        if ($boot_device eq $dev->{name}) {
            $self->sanlog(-msg =>
                          "HPUXDevMgr::_get_free_devices() - skipping the boot device ($boot_device)");
            next;
        }

        # Apply filters to any of the free devices found
        my $device = $dev->{name_ctd};

        if (defined $filer and length $filer) {
            $self->sanlog(-msg => "\nFilter filer: $filer\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i);
        }

        if (defined $vol and length $vol) {
            $self->sanlog(-msg => "\nFilter volume: $vol\nDevice: $device\nDevice volume: " . $self->{dfmap}{$device}{vol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i);
        }

        if (defined $lun and length $lun) {
            $self->sanlog(-msg => "\nFilter lun: $lun\nDevice: $device\nDevice lun: " . $self->{dfmap}{$device}{lun},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{lun} !~ /^$lun$/i);
        }

        if (defined $protocol and length $protocol) {
            $self->sanlog(-msg => "\nFilter protocol: $protocol\nDevice: $device\nDevice protocol: " . $self->{dfmap}{$device}{protocol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{protocol} !~ /$protocol/i);
        }

        # Filter on a filer list if provided by caller
        if (defined $filers and scalar(@{$filers}) != 0) {
            my $found = 0;
            # Look through each filer in the filers list
            foreach my $filer_f (@{$filers}) {
                $self->sanlog(-msg => "\nFilter filer(s): $filer_f\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                              -msg_level => 3);
                # If filer matches devices filer, set found to 1
                if ($self->{dfmap}{$device}{filer} =~ /^$filer_f$/i) {
                    $found = 1;
                }
            }

            # If the device is not part of the filers list caller requested skip it.
            if ($found == 0) {
                next;
            }
        }

        push @devices, $device if (++$total <= $cnt);
    }

    # generate error if user supplied cnt and free devices fell short
    if ($bCnt == 1 and scalar(@devices) < $cnt) {
        $self->_set_err("Error: HPUXDevMgr::get_free_devices() ",
                        "Only " . scalar(@devices) . " of $cnt requested devices were found");
        return (wantarray()) ? () : [];
    }

    $self->sanlog(-msg => "Devices returned: @devices");
    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new devices object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $dev, $devObj, $log, $lvm, $parent,
        $hosttype, $load_devices, $self, $status, @devices, @output);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: HPUXDevMgr::_new() Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?load_devices/) {
            if (not length($load_devices = $args{$_})) {
                $self->_set_err("Error: HPUXDevMgr::_new() Invalid value for argument '-load_devices'");
                return ($self);
            }
            next;
        }
    }

    # set default for load_devices (true)
    if (not defined($load_devices)) {
        $load_devices = 1;
    }
    
    # Allow CLI commands to be logged;
    $self->{log} = $log if (defined $log);

    # check for existance of required -parent argument
    if (grep(/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Error: HPUXDevMgr::_new()", "Invalid value for argument '-conn'");
            return $self;
        }
        $hosttype = $args{-parent}->{hostp}{hosttype};
    } else {
        $self->_set_err("Error: HPUXDevMgr::_new()", "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn}   = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{devices} = [];

    if ($load_devices == 1) {
        @devices = $self->get_free_devices(-refresh => 1, %args);
        if ($self->error) {
            return $self;
        }
    
        foreach $dev (@devices) {
            # initialize device object from the contents of the device manager
            # configuration file. Sometimes stale entries can appear from
            # previously failed runs. If this happens, don't bother loading
            # the device.
            $devObj = $self->create_device(-name => $dev);
            unless ($self->error) {
                $devObj->load();
            } else {
                # We don't care if there is an error on self, clear it
                $self->_clear_err();
            }
        }
    }

    return $self;
}

1;
