# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Devices on HyperV
## @author hrivera@netapp.com
## @status private

# Package Name
##############
package HyperVDev;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use vars qw( @ISA );

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: load
# Objective:       sets parameters based on devices current mount state
#                  NOOP for HyperV
# Details:         See POD documentation at the end of this file
################################################################################
sub load {
    my ($self) = @_;

    $self->_clear_err();

    # load all the existing vhd files into the object
    if ($self->get_device_type() =~ /vhd/i) {
        $self->get_vhd_files();
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_fs
# Objective:       create a filesystem on the disk object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_fs {
    my ($self, $label, $type) = @_;

    # this should be called on all VHD devices

    $self->_clear_err();

    # partition the device
    $self->partition();
    if ($self->error()) {
        $self->_set_err("Partition failed.\n" .
                        "Reason: " . $self->error());

        return $ClientUtil::FALSE;
    }

    # assign a drive letter
    $self->mount_fs($type);
    if ($self->error()) {
        $self->_set_err("Mount FS failed.\n" .
                        "Reason: " . $self->error());

        return $ClientUtil::FALSE;
    }

    # format with a label
    $self->format($label);
    if ($self->error()) {
        $self->_set_err("Format failed.\n" .
                        "Reason: " . $self->error());

        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: mount_fs
# Objective:       assign a drive letter to this disk object
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_fs {
    my ($self, $type) = @_;

    $self->_clear_err();

    $type = "letter" if (not defined($type));

    $self->{disk_obj}->assign(-type => $type);
    if ($self->{disk_obj}->error()) {
        $self->_set_err("Could not assign drive letter to " . $self->get_disk_id() . "\n" .
                        "Reason: " . $self->{disk_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{mount_type} = $type;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
#                  NOOP for HyperV
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self) = @_;

    $self->_clear_err();
    $self->_set_err("unmount_fs() function is not implemented");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if filesystem is mounted
#                  NOOP for HyperV
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    $self->_clear_err();

    if ($self->get_mount() eq "") {
        return $ClientUtil::FALSE;
    } else {
        return $ClientUtil::TRUE;
    }
}

################################################################################
# Subroutine Name: is_partitioned
# Objective:       returns
# Details:         See POD documentation at the end of this file
################################################################################
sub is_partitioned {
    my ($self) = @_;

    $self->_clear_err();

    if ($self->{disk_obj}->{isParted}) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: partition
# Objective:       create a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub partition {
    my ($self, %args) = @_;

    $self->{disk_obj}->partition();
    if ($self->{disk_obj}->error()) {
        $self->_set_err("Could not partition " . $self->get_disk_id() . "\n" .
                        "Reason: " . $self->{disk_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: format
# Objective:       format the disk object
# Details:         See POD documentation at the end of this file
################################################################################
sub format {
    my ($self, $label) = @_;

    my %format_args = ();
    if (defined($label)) {
        $format_args{-label} = $label;
    }

    $self->{disk_obj}->format(%format_args);
    if ($self->{disk_obj}->error()) {
        $self->_set_err("Could not format " . $self->get_disk_id() . "\n" .
                        "Reason: " . $self->{disk_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy_partition
# Objective:       destroy a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_partition {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->{disk_obj}->destroy();
    if ($self->{disk_obj}->error()) {
        $self->_set_err("Could not destroy " . $self->get_disk_id() . "\n" .
                        "Reason: " . $self->{disk_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: offline_disk
# Objective:       offline this device
# Details:         See POD documentation at the end of this file
################################################################################
sub offline_disk {
    my ($self, %args) = @_;

    $self->_clear_err();

    # should be called on all passthrough devices
    $self->{disk_obj}->offline();
    if ($self->{disk_obj}->error()) {
        $self->_set_err("Could not offline " . $self->get_disk_id() . "\n" .
                        "Reason: " . $self->{disk_obj}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();

    my $size = $self->{disk_obj}->{lunSize};
    $size =~ s/\s+//g;
    return $size;
}

################################################################################
# Subroutine Name: get_filer_name
# Objective:       Obtain the name of the filer
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filer_name {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{disk_obj}->{filer};
}

################################################################################
# Subroutine Name: get_filer_name
# Objective:       Obtain the name of the filer
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{disk_obj}->{protocol};
}

################################################################################
# Subroutine Name: get_controller_type
# Objective:       Obtain the type of the disk controller (ide|scsi)
# Details:         See POD documentation at the end of this file
################################################################################
sub get_controller_type {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{controller_type};
}

################################################################################
# Subroutine Name: get_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{disk_obj}->{rawName};
}

################################################################################
# Subroutine Name: refresh_geometry
# Objective:       refresh geometry of the device (NOOP)
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh_geometry {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_partition_size
# Objective:       Determine the size of the active partition (NOOP)
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition_size {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_vhd_files
# Objective:       returns all the VHD files on this device.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vhd_files {
    my ($self) = @_;

    $self->_clear_err();

    if ($self->is_partitioned() == $ClientUtil::FALSE
        or $self->is_mounted() == $ClientUtil::FALSE) {
        $self->{vhd_count} = 0;
        return (wantarray() ? () : []);
    } else {
        my @vhd_list;
        my $mount_point = $self->get_mount();
        my $cache_id = "vhd_list_" . $mount_point;
        if ($self->{host_obj}->is_cache_data_valid($cache_id)) {
            @vhd_list = $self->{host_obj}->get_cache_data($cache_id);
        } else {
            # issue a dir on this devices drive letter vhd directory
            # return the listing of all the files. store in an array
            if ($self->{mount_type} !~ /mount/i and $mount_point !~ /\:/) {
                $mount_point .= ":";
            }

            # sample dir output:
            # Volume in drive G is NATE_VHD_LUN_fas270-rtp07
            # Volume Serial Number is DC1B-20CE
            #
            # Directory of G:\\
            #
            #03/17/2009  02:09 PM     2,578,658,816 nate_vhd_fixed_fas270_rtp07_0.vhd
            #                1 File(s)  2,578,658,816 bytes
            #                0 Dir(s)     599,429,120 bytes free


            my $cmd = "dir $mount_point";
            my ($status, @output) = &ClientUtil::execute($self->{parent}->{parent}->{conn},
                                                         $cmd,
                                                         60);
            if (grep(/cannot_find/, @output) or scalar(@output) == 0) {
                $self->_set_err("Cannot find mount point '$mount_point'");
                return (wantarray() ? () : []);
            }

            # pick apart the output of Dir and parse out all the VHD
            @vhd_list = grep(/\.vhd\s*$/i, @output);

            # modify the list of vhd files in place so that they contain the mount point prepended to it
            @vhd_list = map {
                my $line = $_;
                my $vhd_file = "";
                if ($line =~ /\d+\s+(\S+\.vhd)\s*$/i) {
                    $vhd_file = $1;
                }
                # prepend the mount point
                $mount_point . "\\" . $vhd_file;
            } @vhd_list;

            # store the information in the cache.
            $self->{host_obj}->store_cache_data($cache_id, @vhd_list);

            # update the number VHD files in this device
            $self->{vhd_count} = scalar(@vhd_list);
        }

        # return the list of vhd files
        return (wantarray() ? @vhd_list : \@vhd_list);
    }
}

################################################################################
# Subroutine Name: create_vhd()
# Objective:       create a VHD file in it of the specified size and name.
# Returns:         return back the filepath
################################################################################
sub create_vhd {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($vhd_type, $vhd_size_bytes, $vhd_path);
    my ($space_allocation_percentage) = 0.80;

    # parse arguments
    foreach (keys %args) {
        if (/-?vhd_type/) {
            if (not length($vhd_type = $args{$_})) {
                $self->_set_err("Invalid value for -vhd_type");
                return "";
            }
            next;
        } elsif (/-?vhd_size_bytes/) {
            if (not length($vhd_size_bytes = $args{$_})) {
                $self->_set_err("Invalid value for -vhd_size_bytes");
                return "";
            }
            next;
        } elsif (/-?vhd_path/) {
            if (not length($vhd_path = $args{$_})) {
                $self->_set_err("Invalid value for -vhd_path");
                return "";
            }
        } elsif (/-?space_allocation_percentage/) {
            if (not length($space_allocation_percentage = $args{$_})) {
                $self->_set_err("Invalid value for -space_allocation_percentage, '$space_allocation_percentage'");
                return (wantarray() ? () : []);
            }
        }
    }

    # check for required parameters
    if (not defined($vhd_type) or $vhd_type !~ /fixed|dynamic/i) {
        $self->_set_err("Missing or Invalid parameter -vhd_type, valid values are 'fixed' or 'dynamic'.");
        return "";
    }
    if (not defined($vhd_path)) {
        $self->_set_err("Missing required parameter -vhd_path");
        return "";
    }

    # if vhd_name does not contain the full path to the mount point then prepend it
    my $dev_mount = $self->get_mount();
    if ($self->{mount_type} =~ /letter/i) {
        if ($vhd_path !~ /^\s*$dev_mount\:\\/i) {
            $vhd_path = $dev_mount . "\:\\" . $vhd_path;
        }
    } elsif ($self->{mount_type} =~ /mount/i) {
        my $mntpoint = $dev_mount;
        $mntpoint =~ s/\//\\\//g;
        if ($vhd_path !~ /^\s*$dev_mount\\/i) {
            $vhd_path = $dev_mount . "\\". $vhd_path;
        }
    }

    # if a size is not defined, then consume the entire device (leave 20% free)
    # If we run out of space on any VHD device, Hyper-V will suspend the guests that are using
    # the disk (IE: have VHD files originating on this LUN)
    if (not defined($vhd_size_bytes)) {
        $vhd_size_bytes = int(&ClientUtil::convert_to_bytes($self->get_size()) * $space_allocation_percentage);
    }

    # create the VHD file
    my ($status, @output) = $self->{host_obj}->hvcontrol(-create_vhd => 1,
                                                         -vhd_path => $vhd_path,
                                                         -vhd_type => $vhd_type,
                                                         -vhd_size_bytes => $vhd_size_bytes,
                                                         -timeout => 600);
    if ($self->{host_obj}->error()) {
        $self->_set_err("Could not create VHD ($vhd_path [$vhd_type]) of size ($vhd_size_bytes)\n" .
                        "Reason: " . $self->{host_obj}->errmsg());
        return "";
    }

    # increment the vhd count on this device
    $self->{vhd_count}++;

    # invalidate the cache of the vhd_list for this device object
    # since we now have modified the list.
    my $cache_id = "vhd_list_" . $dev_mount;
    $self->{host_obj}->invalidate_items_cache_data($cache_id);

    # return back the path to the VHD
    return $vhd_path;
}

################################################################################
# Subroutine Name: destroy_vhd
# Objective:       delete a VHD file from this device
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vhd {
    my ($self, %args) = @_;

    my $vhd_path;

    # parse arguments
    foreach (keys %args) {
        if (/-?vhd_path/) {
            if (not length($vhd_path = $args{$_})) {
                $self->_set_err("Invalid value for -vhd_path");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not defined($vhd_path) or $vhd_path eq "") {
        $self->_set_err("Missing required parameter -vhd_path");
        return $ClientUtil::FALSE;
    }

    if ($vhd_path !~ /\.vhd\s*$/i) {
        $self->_set_err("Invalid VHD file, $vhd_path does not have a .vhd extension");
        return $ClientUtil::FALSE;
    }

    # delete the vhd file
    $self->{host_obj}->{WinClient}->deleteRemoteFile($vhd_path);

    # decrement the vhd_count on this vhd device
    $self->{vhd_count}--;

    # invalidate the cache of the vhd_list for this device object
    # since we now have modified the list.
    my $cache_id = "vhd_list_" . $self->get_mount();
    $self->{host_obj}->invalidate_items_cache_data($cache_id);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy_all_vhds
# Objective:       removes all VHD files from the VHD device.
################################################################################
sub destroy_all_vhds {
    my ($self) = @_;

    # get all the vhd files in this device
    my (@vhd_files) = $self->get_vhd_files();
    if ($self->error()) {
        $self->_set_err("Could not obtain VHD files\n" .
                        "Reason" . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # destroy all the vhd files on this device.
    foreach my $vhd_file (@vhd_files) {
        $self->destroy_vhd(-vhd_path => $vhd_file);
        if ($self->error()) {
            $self->_set_err("Could not destroy VHD file '$vhd_file'\n" .
                            "Reason" . $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    $self->{vhd_count} = 0;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_device_type
# Objective:       returns the type of this device (vhd,passthrough)
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device_type {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{device_type};
}

################################################################################
# Subroutine Name: get_lun_type
# Objective:       returns the lun type of this lun
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun_type {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{disk_obj}->{lunType};
}

################################################################################
# Subroutine Name: set_lbpolicy
# Objective:       set the load balancing policy on this device
################################################################################
sub set_lbpolicy {
    my ($self) = @_;

    $self->_clear_err();

    $self->{host_obj}->_set_lbpolicy($self->{disk_obj});
    if ($self->{host_obj}->error()) {
        $self->_set_err($self->{host_obj}->error());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       returns the raw device name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{disk_obj}->{rawName};
}

################################################################################
# Subroutine Name: get_mount
# Objective:       returns mount point of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{disk_obj}->{mount};
}

################################################################################
# Subroutine Name: get_fs
# Objective:       returns the fstype of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{disk_obj}->{fsType};
}

################################################################################
# Subroutine Name: get_lun
# Objective:       Return lun path
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{disk_obj}->{lunPath};
}

################################################################################
# Subroutine Name: get_disk_id
# Objective:       Return the disk id for this device
# Details:         See POD documentation at the end of this file
################################################################################
sub get_disk_id {
    my ($self) = @_;

    $self->_clear_err();

    return "Disk " . $self->{disk_obj}->{device};
}

################################################################################
# Subroutine Name: get_partition
# Objective:       Returns the name of the configured partition.
#                  NOOP for HyperV
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

################################################################################
# Subroutine Name: is_device_free()
# Objective:       Returns 0 if device is not free, 1 if free (not in use)
# Details:         See POD documentation at the end of this file
################################################################################
sub is_device_free {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{device_is_free};
}

################################################################################
# Subroutine Name: set_device_state()
# Objective:       Resets the device is free attribute to the value passed in
# Details:         See POD documentation at the end of this file
################################################################################
sub set_device_state {
    my ($self, $state) = @_;

    $self->_clear_err();

    $self->{device_is_free} = $state;
    return;
}

################################################################################
# Subroutine Name: get_label
# Objective:       Returns the name of the configured partition's label.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_label {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{disk_obj}->{fsLabel};
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new dev object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self, $disk_obj, $parent);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?disk_obj$/) {
            if (not length($disk_obj = $args{$_})) {
                $self->_set_err("Invalid value for argument '-disk_obj'");
                return $self;
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for required arguments
    unless (defined $disk_obj) {
        $self->_set_err("Missing required argument '-disk_obj'");
        return $self;
    }
    unless (defined $parent) {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{parent} = $parent;
    $self->{host_obj} = $parent->{parent};
    $self->{disk_obj} = $disk_obj;

    # need to differentiate between vhd and passthrough, use the parents vhd_lun_path_regex attribute if available
    my $lun_path = $self->get_lun();
    my $lun_type = $self->get_lun_type();
    my $regex = $self->{parent}->{vhd_lun_path_regex};

    if (defined($regex)) {
        # use the regex defined in the devmgr to differentiate between vhd and passthrough
        if ($lun_path !~ /$regex/) {
            $self->{device_type} = "passthrough";
        } else {
            $self->{device_type} = "vhd";
        }
    } else {
        # otherwise default to "hyper_v" luntypes as being default luntype for VHD devices
        if ($lun_type !~ /hyper_v/i) {
            $self->{device_type} = "passthrough";
        } else {
            $self->{device_type} = "vhd";
        }
    }

    # mark this device as free
    $self->set_device_state(1);

    # load the device's current state
    $self->load();

    return $self;
}

1;
