# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to DevMgr for Hyper-V host
## @author hrivera@netapp.com
## @status private
## @pod ./DevMgr.pod

# Package Name
##############
package San::SanClient::DevMgr::HyperVDevMgr;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::DevMgr;
use San::SanClient::DevMgr::HyperVDev;

use Error;
use San::SanLog;

use vars qw( @ISA );

@ISA = qw(Error SanLog DevMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_device_by_disk_id()
# Objective:       return a device object based on the device name
################################################################################
sub get_device_by_disk_id {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $disk_id;

    # parse arguments
    foreach (keys %args) {
        if (/-?disk_id/) {
            if (not length($disk_id = $args{$_})) {
                $self->_set_err("Invalid value for -disk_id");
                return {};
            }
            next;
        }
    }

    # get all the devices to search through
    my @all_devices = $self->get_devices(%args);

    # scan through the device list until we find the device
    foreach my $dev_obj (@all_devices) {
        if ($dev_obj->get_disk_id() eq $disk_id) {
            return $dev_obj;
        }
    }

    $self->_set_err("Could not find device by disk id: '$disk_id'");
    return {};
}

################################################################################
# Subroutine Name: get_devices
# Objective:       return all currently configured device objects
################################################################################
sub get_devices {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $device_type;

    # parse arguments
    foreach (keys %args) {
        if (/-?device_type/) {
            if (not length($device_type = $args{$_})) {
                $self->_set_err("Invalid value for -device_type");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    my (@device_list);

    if ($device_type =~ /vhd/i) {
        @device_list = @{$self->{vhd_devices}};
    } elsif ($device_type =~ /passthrough/i) {
        @device_list = @{$self->{passthrough_devices}};
    } elsif (not defined($device_type)) {
        # otherwise return all of them
        @device_list = (@{$self->{vhd_devices}},
                        @{$self->{passthrough_devices}});
    } else {
        $self->_set_err("Unknown device_type: '$device_type'");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @device_list : \@device_list;
}

################################################################################
# Subroutine Name: create_device
# Objective:       create a device object
################################################################################
sub create_device {
    my ($self, %args) = @_;

    my $disk_obj;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?disk_obj/) {
            if (not length($disk_obj = $args{$_})) {
                $self->_set_err("Invalid value for -disk_obj");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not defined($disk_obj)) {
        $self->_set_err("Missing required parameter -disk_obj");
        return $ClientUtil::FALSE;
    }

    my $disk_id = $disk_obj->{device};

    # create the HyperVDev Object
    my $dev_obj = HyperVDev->_new(-disk_obj => $disk_obj,
                                  -parent => $self);
    if ($dev_obj->error()) {
        $self->_set_err("Failed to create a HyperVDev Object for 'Disk $disk_id'\n" .
                        "Reason: " . $dev_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # determine the type of this dev object and add to the appropriate device list array
    my $device_type = $dev_obj->get_device_type();
    if ($device_type =~ /vhd/i) {
        push (@{$self->{vhd_devices}}, $dev_obj);
        $self->{vhd_device_count}++;
    } elsif ($device_type =~ /passthrough/i) {
        push (@{$self->{passthrough_devices}}, $dev_obj);
        $self->{passthrough_device_count}++;
    } else {
        $self->_set_err("Unknown device_type: '$device_type'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_dev_id
# Objective:       returns unique device name(with full path) of passed device's
#                  short name to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_id {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: get_dev_name
# Objective:       returns device short name of passed device's unique name(long name)
#                  to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_name {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: cleanup
# Objective:       NOOP for HyperV
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup {
    my ($self) = @_;

    $self->_clear_err();
}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;

    $self->_clear_err();

    my (@filers, $num_devices, $device_type, $lun_ostype_regex, $lun_path_regex, $protocol);
    $num_devices = -1; # default is to return all devices found

    # parse arguments
    foreach (keys %args) {
        if (/-?filers/) {
            if (not scalar(@filers = @{$args{$_}})) {
                $self->_set_err("Invalid value for -filers");
                return (wantarray() ? () : []);
            }
            next;
        } elsif (/-?num_devices/) {
            if (not length($num_devices = $args{$_})) {
                $self->_set_err("Invalid value for -num_devices");
                return (wantarray() ? () : []);
            }
            next;
        } elsif (/-?device_type/) {
            if (not length($device_type = $args{$_}) or $device_type !~ /vhd|passthrough/i) {
                $self->_set_err("Invalid value for -device_type");
                return (wantarray() ? () : []);
            }
            next;
        } elsif (/-?lun_path_regex/) {
            if (not length($lun_path_regex = $args{$_})) {
                $self->_set_err("Invalid value for -lun_path_regex");
                return (wantarray() ? () : []);
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_}) or $protocol !~ /fcp|iscsi/i) {
                $self->_set_err("Invalid value for -protocol");
                return (wantarray() ? () : []);
            }
            next;
        } elsif (/-?lun_ostype_regex/) {
            if (not length($lun_ostype_regex = $args{$_})) {
                $self->_set_err("Invalid value for -lun_ostype_regex");
                return (wantarray() ? () : []);
            }
            next;
        }
    }

    my @device_list = $self->get_devices();
    my @filtered_device_list;

    # default is to return all devices that are marked free.
    foreach my $device (@device_list) {
        # (1) filter out devices that are not free
        next if ($device->is_device_free() != 1);

        # (2) filter out devices that do not match the lun path filter
        if (defined($lun_path_regex)) {
            next if ($device->get_lun() !~ $lun_path_regex);
        }

        # (2a) filter out devices that do not match the lun ostype filter
        if (defined($lun_ostype_regex)) {
            next if ($device->get_lun_type() !~ $lun_ostype_regex);
        }

        # (3) filter out devices that do not match any of the specified filers
        if (scalar(@filers)) {
            my $matched = 0;
            foreach my $filer_name (@filers) {
                $matched = 1 if ($device->get_filer_name() =~ /^($filer_name)$/);
            }
            next if ($matched != 1);
        }

        # (4) filter out devices that do not match the protocol
        if (defined($protocol)) {
            next if ($device->get_protocol() !~ /$protocol/i);
        }

        # (5) filter out boot luns
        next if ($device->get_lun() =~ /boot/i);

        # (6) filter out device type (vhd/passthrough)
        if (defined($device_type)) {
            next if ($device->get_device_type() !~ /$device_type/i);
        }

        # this device passed filters 1 through 6
        push(@filtered_device_list, $device);
    }

    # The next task is to make @filtered_device_list balanced.
    # filer_hash will be a hash of filer name keys. Each key will have an
    # array of device objects that belong to that filer.
    my %filer_hash = ();

    foreach my $device (@filtered_device_list) {
        # add this device into that filer's key array
        push(@{$filer_hash{$device->get_filer_name()}}, $device);
    }

    # now balance out the device list so that the final list we return
    # to the caller has alternating devices from different filers/controllers.
    my @balanced_device_list = ();
    while(scalar(keys %filer_hash) > 0) {
        # for every filer key, shift off one of the devices and push it into the final array
        foreach my $filer_array (keys %filer_hash) {
            push(@balanced_device_list, shift(@{$filer_hash{$filer_array}}));
            if (scalar(@{$filer_hash{$filer_array}}) == 0) {
                delete $filer_hash{$filer_array};
            }
        }
    }

    # Since the ordering for @balanced_device_list is consistent across
    # multiple HyperV Hosts, it may occur where all devices returned are heavily
    # biased towards one storage controller. Therefore we provide the ability
    # to reverse the @balanced_device_list if the value passed into $self->{_balance}
    # is an odd number.
    if (($self->{_balance} % 2) == 1) {
        # reverse the device list
        @balanced_device_list = reverse @balanced_device_list;
    }

    # if the calling function requested a specific amount of devices,
    # pull some out of the balanced device list using some clever math.
    # store the final list in @final_device_list
    my @final_device_list = ();
    if ($num_devices != -1) {
        my $total_devices = scalar(@balanced_device_list);
        my $difference = $total_devices - $num_devices;
        if ($difference < 0) {
            $self->_set_err("Requested ($num_devices) devices but only " .
                            "($total_devices) devices are available.");
            return (wantarray()) ? () : [];
        } elsif ($difference == 0) {
            # in this case the num devices is equal to the number of devices available, return all of them
            @final_device_list = @balanced_device_list;
        } else {
            # pull out $num_devices devices from @balanced_device_list
            for (my $count = 0; $count < $num_devices; $count++) {
                my $next_device = $self->{_next_device} % $total_devices;
                $self->{_next_device}++;

                # push it on the final list
                push(@final_device_list, $balanced_device_list[$next_device]);
            }
        }
    } else {
        # otherwise return all available devices because the caller did not
        # specify a specific amount of devices to get
        @final_device_list = @balanced_device_list;
    }

    # if our final device list is empty at this point, then error out
    if (scalar(@final_device_list) < 1) {
        $self->_set_err("No '$device_type' devices available!");
        return (wantarray()) ? () : [];
    }

    # if we are obtaining vhd devices, reverse the list so that
    # passthrough disk controller origins are out of phase with vhd devices.
    if ($device_type =~ /vhd/i) {
        $self->{_next_device} = 0;
        @final_device_list = reverse @final_device_list;
    }

    # mark the remaining device objects as not free
    foreach my $dev_obj (@final_device_list) {
        $dev_obj->set_device_state(0);
    }

    return (wantarray() ? @final_device_list : \@final_device_list);
}

################################################################################
# Subroutine Name: create_vhd_files()
# Objective:       Create VHD files on the specified devices.
################################################################################
sub create_vhd_files {
    my ($self, %args) = @_;

    my (@vhd_devices);
    my $num_vhds_per_device = 1; # defaults to one vhd file per vhd device
    my @vhd_types = ("fixed"); # default to fixed vhd type
    my $fs_label;
    my $mount_type;
    my $space_allocation_percentage = 0.80; # default 80% of device consumed for vhd files

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vhd_devices/) {
            if (not scalar(@vhd_devices = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vhd_devices");
                return (wantarray() ? () : []);
            }
            next;
        } elsif (/-?num_vhds_per_device/) {
            if (not length($num_vhds_per_device = $args{$_}) or $num_vhds_per_device < 1) {
                $self->_set_err("Invalid value for -num_vhds_per_device, '$num_vhds_per_device'");
                return (wantarray() ? () : []);
            }
            next;
        } elsif (/-?vhd_types/) {
            if (not scalar(@vhd_types = @{$args{$_}}) or not grep(/^fixed|dynamic$/i, @vhd_types)) {
                $self->_set_err("Invalid value for -vhd_types, '@vhd_types'");
                return (wantarray() ? () : []);
            }
            next;
        } elsif (/-?fs_label/) {
            if (not length($fs_label = $args{$_})) {
                $self->_set_err("Invalid value for -fs_label, '$fs_label'");
                return (wantarray() ? () : []);
            }
        } elsif (/-?mount_type/) {
            if (not length($mount_type = $args{$_})) {
                $self->_set_err("Invalid value for -mount_type, '$mount_type'");
                return (wantarray() ? () : []);
            }
        } elsif (/-?space_allocation_percentage/) {
            if (not length($space_allocation_percentage = $args{$_})) {
                $self->_set_err("Invalid value for -space_allocation_percentage, '$space_allocation_percentage'");
                return (wantarray() ? () : []);
            }
        }
    }

    # check for required parameters
    if (not scalar(@vhd_devices)) {
        $self->_set_err("Missing required parameter -vhd_devices");
        return (wantarray() ? () : []);
    }
    if (not scalar(@vhd_types)) {
        $self->_set_err("Missing required parameter -vhd_types");
        return (wantarray() ? () : []);
    }

    # this is the counter we will use for our balance hash
    my $vhd_balance_hash = {};
    my $vhd_device_count = 0;

    foreach my $vhd_device (@vhd_devices) {
        # determine the fs_label to use if none were specified
        if ($fs_label eq "") {
            $fs_label = "NATE_VHD_LUN_" . $vhd_device->get_filer_name();
            
            # the label can be no more than 30 characters
            $fs_label =~ s/^(\S{30}).+$/$1/;
        }

        # create a filesystem on this VHD device
        $vhd_device->create_fs($fs_label, $mount_type);
        if ($vhd_device->error()) {
            $self->_set_err("Could not create filesystem with label ($fs_label)\n" .
                            "Reason: " . $vhd_device->errmsg());
            return (wantarray() ? () : []);
        }

        # determine the size of this vhd device in bytes, shave off 20% of the total size
        # If we run out of space on any VHD device, Hyper-V will suspend the guests that are using
        # the disk (IE: have VHD files originating on this LUN)
        my $size_of_vhd_lun_bytes = int(&ClientUtil::convert_to_bytes($vhd_device->get_size()) * $space_allocation_percentage);

        # calculate the size of each vhd file
        my $size_of_each_vhd_file_bytes = int($size_of_vhd_lun_bytes / $num_vhds_per_device);

        my @vhd_files_for_this_device = ();
        my $vhd_type = "fixed";

        # create the VHD file(s) for this device
        for (my $i = 0; $i < $num_vhds_per_device; $i++) {
            # determine the vhd type for this vhd from the vhd types list.
            # if there are none left, then it will use the last value it was set to.
            if (scalar(@vhd_types)) {
                $vhd_type = shift(@vhd_types);
            }

            # create the VHD file
            my $vhd_file_name = "nate_vhd_" . $vhd_type . "_" . $vhd_device->get_filer_name() . "_" . $i  . ".vhd";
            $vhd_file_name =~ s/\-/_/g;
            my $vhd_path = $vhd_device->create_vhd(-vhd_path => $vhd_file_name,
                                                   -vhd_size_bytes => $size_of_vhd_lun_bytes,
                                                   -vhd_type => $vhd_type,
                                                   -space_allocation_percentage => $space_allocation_percentage);
            if ($vhd_device->error()) {
                $self->_set_err("Could not create VHD file '$vhd_file_name'\n" .
                                "Reason: " . $vhd_device->errmsg());
                return (wantarray() ? () : []);
            }

            # push the vhd path onto an array that represents all the vhd files
            # on this device. We will need to balance this out once we have created
            # all vhd's for all devices.
            push(@vhd_files_for_this_device, $vhd_path);
        }

        # store the vhd files for this device in the balance_hash.
        # we are expecting the array @vhd_devices to be an alternating list
        # between all controller heads.
        @{$vhd_balance_hash->{$vhd_device_count}} = @vhd_files_for_this_device;
        $vhd_device_count++;
        @vhd_files_for_this_device = ();
    }

    my @balanced_vhd_list;

    # Shift one vhd file out of each vhd_device and push it on the balanced list
    # until all the vhd files have been added to the balanced list.
    while (keys %{$vhd_balance_hash}) {
        my @keys = keys %{$vhd_balance_hash};
        foreach my $vhd_device_id (sort {$a <=> $b} @keys) {
            my $vhd_file = shift(@{$vhd_balance_hash->{$vhd_device_id}});
            # add this vhd file to the balanced list.
            push(@balanced_vhd_list, $vhd_file);

            # if this vhd_device is empty, then remove it from the hash.
            if (scalar(@{$vhd_balance_hash->{$vhd_device_id}}) <= 0) {
                delete($vhd_balance_hash->{$vhd_device_id});
            }
        }
    }

    return (wantarray() ? @balanced_vhd_list : \@balanced_vhd_list);
}

################################################################################
# Subroutine Name: configure_passthrough_disks()
# Objective:       Offlines the specified Passthrough HyperVDev objects
################################################################################
sub configure_passthrough_disks {
    my ($self, %args) = @_;

    $self->_clear_err();

    my (@passthrough_devices, @controller_types);

    # parse arguments
    foreach (keys %args) {
        if (/-?passthrough_devices/) {
            if (not scalar(@passthrough_devices = @{$args{$_}})) {
                $self->_set_err("Invalid value for -passthrough_devices");
                return (wantarray() ? () : []);
            }
            next;
        } elsif (/-?controller_types/) {
            if (not scalar(@controller_types = @{$args{$_}})) {
                $self->_set_err("Invalid value for -controller_types");
                return (wantarray() ? () : []);
            }
            next;
        }
    }

    # by default configure all passthrough devices
    if (not scalar(@passthrough_devices)) {
        @passthrough_devices = $self->get_devices(-device_type => "passthrough");
    }

    # check for required parameters
    if (not scalar(@controller_types)) {
        $self->_set_err("Missing required parameter -controller_types");
        return (wantarray() ? () : []);
    }

    my $controller_type;

    # offline all the passthrough devices
    foreach my $passthrough_dev (@passthrough_devices) {
        $passthrough_dev->offline_disk();
        if ($passthrough_dev->error()) {
            $self->_set_err("Offline disk failed\n" .
                            "Reason: " . $passthrough_dev->errmsg());
            return (wantarray() ? () : []);
        }

        # assign a controller type to this passthrough device (scsi|ide)
        if (scalar(@controller_types)) {
            $controller_type = shift(@controller_types);
        }
        $passthrough_dev->{controller_type} = $controller_type;
    }

    # return back the array of dev objects
    return (wantarray() ? @passthrough_devices : \@passthrough_devices);
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new Device Manager Object for Hyper-V
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self, $balance);
    my ($lun_path_regex) = qr//;
    my ($vhd_lun_path_regex) = qr//;
    my (@target_list);

    $self = {};

    bless $self, $class;

    $self->sanlog(-msg => "Creating a HyperVDevMgr Object.");

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/^-?lun_path_regex$/) {
            $lun_path_regex = $args{$_};
            next;
        } elsif (/^-?target_list$/) {
            if (ref($args{$_}) ne "ARRAY") {
                $self->_set_err("Invalid value for argument '-target_list'");
                return $self;
            }
            @target_list = @{$args{$_}};
            next;
        } elsif (/^-?vhd_lun_path_regex$/) {
            $vhd_lun_path_regex = $args{$_};
            next;
        } elsif (/-?balance/) {
            if (not length($balance = $args{$_})) {
                $self->_set_err("Invalid value for argument '-balance'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    if (defined($parent)) {
        unless (ref($parent) =~ "SanClient") {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # if a balance was not passed in, assume its zero.
    $balance = 0 if (not defined($balance));

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{vhd_devices} = [];
    $self->{passthrough_devices} = [];
    $self->{vhd_device_count} = 0;
    $self->{passthrough_device_count} = 0;
    $self->{lun_path_regex} = $lun_path_regex;
    $self->{vhd_lun_path_regex} = $vhd_lun_path_regex;
    $self->{target_list} = \@target_list;

    $self->{_balance} = $balance;
    $self->{_next_device} = 0;

    # update all the drive letters
    $self->{parent}->{WinClient}->updateLetters();
    if ($self->{parent}->{WinClient}->error()) {
        $self->_set_err("Could not update drive letters.\n" .
                        "Reason: " . $self->{WinClient}->errmsg());
        return $self;
    }

    # pull out all the disk objects from the WinClient object.
    my $disk_objects = $self->{parent}->{WinClient}->{disks};

    # create each HyperVDev object from each WinClient Disk object.
    foreach my $disk_object (values %{$disk_objects}) {
        # skip boot luns, quorum disks, and cluster disks owned by other hosts
        if ($disk_object->{boot}) {
            next;
        }
        if ($disk_object->{isQuorum} == 1) {
            next;
        }
        if ($disk_object->{lunState} =~ /Clus/i) {
            next;
        }

        # skip devices that do not match the lun_path naming convention (if specified)
        if ($lun_path_regex ne "" and $disk_object->{lunPath} !~ $lun_path_regex) {
            next;
        }

        # skip devices that do not match any controller/vserver in the target list specified
        if (scalar(@target_list)) {
            my $target = $disk_object->{filer};
            if (not grep(/^$target$/, @target_list)) {
                next;
            }
        }

        # create the device object for this disk object
        $self->create_device(-disk_obj => $disk_object);
        if ($self->error()) {
            $self->_set_err("Failed to create device");
            return $self;
        }
    }

    return $self;
}

1;
