# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to DevMgr for Solaris host
## @author rawat@netapp.com
## @status private
## @pod ./DevMgr.pod

# Package Name
##############
package San::SanClient::DevMgr::KVMDevMgr;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::DevMgr;
use San::SanClient::DevMgr::KVMDev;

use San::SanLog;
use Error;

use vars qw( @ISA );

@ISA = qw(Error SanLog DevMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_device
# Objective:       return a device object based on the device name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device {
    my ($self, $name) = @_;

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Name is a required argument");
        return {};
    }

    foreach my $obj ($self->get_devices) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Unable to find device '$name'");

    return {};
}

################################################################################
# Subroutine Name: get_devices
# Objective:       return all currently configured device objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self, %args) = @_;
    my (@devices);

    $self->_clear_err();

    my $device_type;

    # parse arguments
    foreach (keys %args) {
        if (/-?device_type/) {
            if (not length($device_type = $args{$_})) {
                $self->_set_err("Invalid value for -device_type");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    if ($device_type =~ /vhd/i) {
        @devices = @{$self->{vhd_devices}};
    } elsif ($device_type =~ /passthrough/i) {
        @devices = @{$self->{passthrough_devices}};
    } elsif (not defined($device_type)) {
        # otherwise return all of them
        @devices = (@{$self->{vhd_devices}},
                    @{$self->{passthrough_devices}});
    } else {
        $self->_set_err("Unknown device_type: '$device_type'");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: create_device
# Objective:       create a device object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_device {
    my ($self, %args) = @_;
    my ($devObj);

    $self->_clear_err();

    $devObj = KVMDev->_new(-parent => $self,
                           -conn   => $self->{conn},
                           %args);

    if ($devObj->error) {
        $self->_set_err($devObj->errmsg());
        return $ClientUtil::FALSE;
    }

    # determine the type of this dev object and add to the appropriate device list array
    my $device_type = $devObj->get_device_type();
    if ($device_type =~ /vhd/i) {
        push (@{$self->{vhd_devices}}, $devObj);
        $self->{vhd_device_count}++;
    } elsif ($device_type =~ /passthrough/i) {
        push (@{$self->{passthrough_devices}}, $devObj);
        $self->{passthrough_device_count}++;
    } else {
        $self->_set_err("Unknown device_type: '$device_type'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: cleanup
# Objective:       remove device objects and remove mount point entries
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup {
    my ($self, @devices) = @_;
    my ($devObj, $fail_cnt);

    $self->_clear_err();

    # remove mount point entry for each device passed in
    foreach $devObj (@{$self->{devices}}) {

        $self->sanlog(-msg => "Cleaning up mount points for device: $devObj->{name}", -msg_level => 2);

        $devObj->unmount_fs() if $devObj->is_mounted;
        if ($devObj->error) {
            $self->_set_err($devObj->errmsg);
            return $ClientUtil::FALSE;
        }

        $devObj->destroy_partition() if $devObj->is_partitioned;
        if ($devObj->error) {
            $self->_set_err($devObj->errmsg);
            return $ClientUtil::FALSE;
        }
    }

    $self->{devices} = [];

    return $ClientUtil::TRUE ;
}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       retrieve list of free host volumes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;
    my ($bCnt,    $cmd,     $num_devices, $filername,   $lun, $device_type,
        $refresh, $status,  $total,       $type,   $vol,   %hDevs,       @all_devices,
        @attr,    @devices, @lvm_devices, @output, @used,  @free_devices, @Filers,
        $lun_path_regex);

    $type  = ".*";
    $total = 0;

    $self->_clear_err();

    # Asssume user supplied -count argument
    $num_devices = -1;
    $refresh = 0;

    # parse arguments

    foreach (keys %args) {
        if (/-?num_devices/) {
            if (not length($num_devices = $args{$_})) {
                $self->_set_err("Invalid value for argument '-num_devices'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?device_type/) {
            if (not length($device_type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-device_type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?filers/) {
            if (not scalar(@Filers = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?lun_path_regex/) {
            if (not length($lun_path_regex = $args{$_})) {
                $self->_set_err("Invalid value for argument '-lun_path_regex'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # It will return the devices depending upon whether multipath
    # is enabled or not
    @devices = $self->get_devices();
    if ($self->error) {
        $self->_set_err($self->errmsg);
        return (wantarray()) ? () : [];
    }

    my @filtered_device_list;
    # iterate through the devices and make sure that none of the disks
    # are being used other LVM disk object.
    foreach my $device (@devices) {

        # Filter out devices that do not match any of the specified filers
        if (scalar(@Filers) > 0) {
            my $matched = 0;
            foreach my $filer_name (@Filers) {
                $matched = 1 if ($device->get_filer_name() =~ /^($filer_name)$/);
            }
            next if ($matched != 1);
        }

        # Filter out device type (vhd/passthrough)
        if (defined($device_type)) {
            next if ($device->get_device_type() !~ /$device_type/i);
        }

        # Filter out devices that do not match the lun path filter
        if (defined($lun_path_regex)) {
            next if ($device->get_lun() !~ $lun_path_regex);
        }

        push (@filtered_device_list, $device);
    }

    # The next task is to make @filtered_device_list balanced.
    # filer_hash will be a hash of filer name keys. Each key will have an
    # array of device objects that belong to that filer.
    my %filer_hash = ();

    foreach my $device (@filtered_device_list) {
        # add this device into that filer's key array
        push(@{$filer_hash{$device->get_filer_name()}}, $device);
    }

    my (@final_device_list);
    if ($num_devices != -1) {
        my $total_devices = scalar(@filtered_device_list);
        my $difference = $total_devices - $num_devices;
        if ($difference < 0) {
            $self->_set_err("Requested ($num_devices) devices but only " .
                            "($total_devices) devices are available.");
            return (wantarray()) ? () : [];
        } elsif ($difference == 0) {
            # in this case the num devices is equal to the number of devices available, return all of them
            @final_device_list = @filtered_device_list;
        } else {
            # pull out $num_devices devices from @filtered_device_list
            for (my $count = 0; $count < $num_devices; $count++) {
                my $next_device = $self->{_next_device} % $total_devices;
                $self->{_next_device}++;

                # push it on the final list
                push(@final_device_list, $filtered_device_list[$next_device]);
            }
        }
    } else {
        # otherwise return all available devices because the caller did not
        # specify a specific amount of devices to get
        @final_device_list = @filtered_device_list;
    }

    # if our final device list is empty at this point, then error out
    if (scalar(@final_device_list) < 1) {
        $self->_set_err("No '$device_type' devices available!");
        return (wantarray()) ? () : [];
    }

    $self->sanlog(-msg => "Devices returned: @final_device_list", -msg_level=>2);
    return (wantarray()) ? @final_device_list : \@final_device_list;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new devices object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $dev, $devObj, $log, $lvm);
    my ($load_devices, $parent, $hosttype, $self, $status, @devices, @output);
    my ($lun_path_regex) = qr//;
    my ($vhd_lun_path_regex) = qr//;

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/^-?lun_path_regex$/) {
            $lun_path_regex = $args{$_};
            next;
        } elsif (/^-?vhd_lun_path_regex$/) {
            $vhd_lun_path_regex = $args{$_};
            next;
        } elsif (/-?load_devices/) {
            if (not length($load_devices = $args{$_})) {
                $self->_set_err("Invalid value for argument '-load_devices'");
                return ($self);
            }
            next;
        }
    }

    # set default for load_devices (true)
    if (not defined($load_devices)) {
        $load_devices = 1;
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
        $hosttype = $args{-parent}->{hostp}{hosttype};
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn}   = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{devices} = [];
    $self->{vhd_devices} = [];
    $self->{passthrough_devices} = [];
    $self->{vhd_device_count} = 0;
    $self->{passthrough_device_count} = 0;
    $self->{lun_path_regex} = $lun_path_regex;
    $self->{vhd_lun_path_regex} = $vhd_lun_path_regex;

    if ($load_devices == 1) {
        @devices = $self->{parent}->_get_unlabeled_devices();
        if ($self->error) {
            return $self;
        }
        foreach $dev (@devices) {
            # initialize device object from the contents of the device manager
            # configuration file.  Sometimes stale entries can appear from
            # previously failed runs.  If this happens, don't bother loading
            # the device.
            $self->create_device(-name => $dev);
            if($self->error){
                return $self;
            }
        }
    }

    return ($self);
}

1;
