# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to DevMgr for Solaris host
## @author lippe@netapp.com, knopp@netapp.com, rawat@netapp.com
## @status private
## @pod ./DevMgr.pod

# Package Name
##############
package San::SanClient::DevMgr::LinuxDevMgr;

use strict;

# Package Requirements
######################
use File::Basename;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanClient::DevMgr::LinuxDev;
use San::SanClient::DevMgr;
use San::SanLog;

use Error;

use vars qw( @ISA );

@ISA = qw(Error SanLog DevMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_device
# Objective:       return a device object based on the device name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device {
    my ($self, $name) = @_;

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: DevMgr::get_device() name is a required argument");
        return {};
    }

    foreach my $obj ($self->get_devices) {
        if ($obj->get_name() =~ /^$name$/ or
            $obj->get_device_path() =~ /^$name$/ or
            $obj->get_raw_device() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: DevMgr::get_device() Unable to find device '$name'");

    return {};
}

################################################################################
# Subroutine Name: get_devices
# Objective:       return all currently configured device objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;
    my (@devices);

    $self->_clear_err();

    if (@{$self->{devices}}) {
        @devices = @{$self->{devices}};
    } else {
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: get_dev_id
# Objective:       returns unique device name(with full path) of passed device's
#                  short name to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_id {
    my ($self, %args) = @_;

    my ($devObj, $dev_name);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?dev_name$/) {
            if (not length($dev_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-dev_name'");
                return "";
            }
            next;
        }
    }

    my $mp_obj = $self->{parent}->{mp};

    if ($mp_obj->{mp_type} =~ /none/i) {
        # if we pass device name to LinuxDevMgr::get_device(), it will return device object
        $devObj = $self->get_device($dev_name);
        # LinuxDev::get_device_path() will return device id with full path
        if ($self->error) {
            $self->_set_err("Unable to find device '$dev_name'");
            return "";
        }
        return $devObj->get_device_path();
    } else {
        while (my ($key, $value) = each (%{$mp_obj->{mapping}})) {
            # If $value->{sdX} was device name (for example 'sdr') we have, then $key would be
            # equivalent device id (for example 360a98000572d436f4b6f5a4c6b7a2d71 or 3600a098000572d436f4b6f5a4c6b7a2d)
            # which we require.
            if ($value->{sdX} eq $dev_name) {
                # if we pass device id to LinuxDevMgr::get_device(), it will return device object
                my $devObj = $self->get_device($key);
                # LinuxDev::get_device_path() will return device id with full path
                if ($self->error) {
                    $self->_set_err("Unable to find device '$key'");
                    return "";
                }
                return $devObj->get_device_path();
            }
        }
    }

    $self->_set_err("Unable to find device '$dev_name'");
    return "";
}

################################################################################
# Subroutine Name: get_dev_name
# Objective:       returns device short name of passed device's unique name(long name)
#                  to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_name {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $dev_id;

    foreach (keys %args) {
        if (/-?dev_id$/) {
            if (not length($dev_id = $args{$_})) {
                $self->_set_err("Invalid value for argument '-dev_id'");
                return "";
            }
            next;
        }
    }

    my $mp_obj = $self->{parent}->{mp};

    foreach my $devObj ($self->get_devices) {
        if ($devObj->get_device_path() =~ /$dev_id/) {
            # For example, $dev_id would be 360a98000572d436f4b6f5a4c6b7a2d71 or 3600a098000572d436f4b6f5a4c6b7a2d
            # if multipath is enabled.
            # scsi-360a9800050334b4f4e342f5044476666 (7-mode) if multipath is not enabled.
            # or
            # scsi-3600a098000572d436f4b6f5a4c6b7a2d (c-mode) if multipath is not enabled.
            my $name = $devObj->get_name();
            # LinuxDev::get_name() returns device name if mp_type is none, else device id
            if ($mp_obj->{mp_type} =~ /none/i) {
                return $name;
            } else {
                return $mp_obj->{mapping}->{$name}->{sdX};
            }
        }
    }

    $self->_set_err("Unable to find device '$dev_id'");
    return "";
}

################################################################################
# Subroutine Name: create_device
# Objective:       create a device object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_device {
    my ($self, %args) = @_;
    my ($devObj);

    $self->_clear_err();

    $devObj = LinuxDev->_new(-parent => $self,
                             -conn   => $self->{conn},
                             %args);

    if ($devObj->error) {
        $self->_set_err($devObj->errmsg());
        return {};
    }

    push @{$self->{devices}}, $devObj;

    return $devObj;
}

################################################################################
# Subroutine Name: cleanup
# Objective:       remove device objects and remove mount point entries
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup {
    my ($self, @devices) = @_;
    my ($devObj, $fail_cnt);

    $self->_clear_err();

    # remove mount point entry for each device passed in
    foreach $devObj (@{$self->{devices}}) {

        $self->sanlog(-msg => 
                      "Cleaning up mount points for device: $devObj->{name}", -msg_level => 2);

        $devObj->unmount_fs() if $devObj->is_mounted;
        if ($devObj->error) {
            $self->_set_err($devObj->errmsg);
            return $ClientUtil::FALSE;
        }

        $devObj->destroy_partition() if $devObj->is_partitioned;
        if ($devObj->error) {
            $self->_set_err($devObj->errmsg);
            return $ClientUtil::FALSE;
        }
    }

    $self->{devices} = [];

    return $ClientUtil::TRUE ;
}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       retrieve list of free host volumes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;
    my ($bCnt,    $cmd,     $cnt,         $device, $filer, $filername,   $lun,
        $refresh, $status,  $total,       $type,   $vol,   %hDevs,       @all_devices,
        @attr,    @devices, @lvm_devices, @output, @used,  @free_devices, $filers,
        $protocol);

    $type  = ".*";
    $total = 0;
    @{$filers} = ();

    $self->_clear_err();

    # Asssume user supplied -cnt argument
    $bCnt    = 1;
    $refresh = 0;

    # parse arguments

    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Error: DevMgr::get_free_devices() Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: DevMgr::get_free_devices() Invalid value for argument '-type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Error: DevMgr::get_free_devices() Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Error: DevMgr::get_free_devices() Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Error: DevMgr::get_free_devices() Invalid value for argument '-vol'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Error: DevMgr::get_free_devices() Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: DevMgr::get_free_devices() Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: NativeLvm::get_free_devices() Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Check if user wants free devices from a specified filer,volume, or lun
    if (defined $filer or defined $vol or defined $lun or defined $protocol or scalar(@{$filers}) != 0) {

        # Check if a device to filer map 'dfmap' has already been created or
        # refresh it if requested.
        if (not defined $self->{dfmap} or $refresh == 1) {
            $self->sanlog(-msg => 
                          "Creating translation maps; device to filer, vol, and lun");
            @output = $self->{sanlun}->lun_show();
            if ($self->{sanlun}->error) {
                $self->_set_err($self->{sanlun}->errmsg());
                return (wantarray()) ? () : [];
            }

            $cmd = "mount";
            my ($status_mount, @mount_output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status_mount > 0) {
                $self->_set_err("mount cmd failed");
                return  (wantarray()) ? () : [];
            }

            foreach my $line (@output) {
                my $dev_tmp = "";

                # skip first line (header)
                if ($line =~ /lun-pathname/i) {
                    next;
                }

                # depending on filername length sanlun may not have display blank space before name
                my ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp);

                # remove leading white spaces if any
                $line =~ s/^\s*//;

                if ($line  =~ /(\S+)\s+\/vol\/(\w+)\/(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+/) {
                    # old sanlun kit output looks like below,
                    #    controller:          lun-pathname        device filename  adapter  protocol          lun size         lun state
                    # mml_sanmgr_vserv:  /vol/mml_vol_2/mml_lun_0  /dev/sda         host37   iSCSI          3g (3221225472)     GOOD
                    #
                    # new sanlun kit output looks like below,
                    # controller(7mode)/                                                 device          host                  lun
                    # vserver(Cmode)       lun-pathname                                  filename        adapter    protocol   size    mode
                    # ---------------------------------------------------------------------------------------------------------------------
                    # fas3170-201-76       /vol/nate_1_1_1/nate_IBMblade-200-218_linux_0 /dev/sde        host22     iSCSI      4g      7

                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp) = ($1, $2, $3, $4, lc($5));
                } elsif ($line =~ /(\S+)\s+\/vol\/(\w+)\/(\S+)\s+(\S+)\s+/) {
                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp) = ($1, $2, $3, $4);
                }

                # Remove colon in the end of filer/vserver name if presents
                $filer_tmp =~ s/:$//;

                next unless (defined $vol_tmp and length($vol_tmp));    # skip orphaned luns

                # skip the lun if it is mounted on root '/'
                my @tmp_array = grep(/$device_tmp/, @mount_output);
                my @root_arr = split(/\s+/, $tmp_array[0]);
                # call pvdisplay to see if device is mounted in LV - don't check status
                $cmd = "pvdisplay | grep $device_tmp";
                my ($pv_status, @pv_output) = &ClientUtil::execute($self->{conn}, $cmd);

                # skip if device is the root filesystem (on raw device or on LVM device)
                next if (($root_arr[2] eq "/") or (grep(/PV\s+Name\s+$device_tmp\b/, @pv_output) > 0));

                if ($self->{parent}->{mp_type} =~ /dmmp/i) {
                    # We need to grep all the dm devices from /dev/mapper. Netapp device starts
                    # with 3600?a0?9. Output of the following cmd will look like this
                    # $cmd = "multipath -l -d $device_tmp |grep -P "360a9|3600a09"";
                    #####################################################################
                    # 360a98000486e574c6734467a334e3366dm-2 NETAPP,LUN
                    # [size=3.0G][features=1 queue_if_no_path][hwhandler=0]
                    # \_ round-robin 0 [prio=0][active]
                    #  \_ 49:0:0:0 sdc  8:32   [failed][undef]
                    #  \_ 50:0:0:0 sdi  8:128  [failed][undef]
                    #####################################################################
                    $cmd = "multipath -l -d $device_tmp |grep -P \"360a9|3600a09\"";
                    my ($status, @device_id) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status > 0) {
                        $self->_set_err(
                                    "Error: NativeLvm::get_free devices() Unable to get lun id for device $device_tmp");
                        return (wantarray()) ? () : [];
                    }

                    # make sure We are not doing same work for the same device again.
                    next if ($dev_tmp =~ /$device_id[0]/);

                    # extracting the 33 char long SCSI ID. This is how the $device_tmp will
                    # look like
                    # e.g. 360a98000486e574d4b6f467a334e4356
                    my @lun_id = split(//, $device_id[0]);
                    my $i = 0;
                    while ($lun_id[$i] != 3) {
                        $i++;
                    }
                    my @dev;
                    for (my $j = 0; $j < 33; $j++) {
                        $dev[$j] = $lun_id[$i++];
                    }

                    $device_tmp = join("", @dev);
                    
                }

                @attr = split(/\s+/, $line);
                $dev_tmp = basename($device_tmp);

                $hDevs{$dev_tmp} = {filer    => $filer_tmp,
                                    vol      => $vol_tmp,
                                    lun      => $lun_tmp,
                                    protocol => $protocol_tmp};

                $self->sanlog(-msg => "Device ($dev_tmp) information:\n\tFiler: $filer_tmp\n\tVolume: $vol_tmp\n" .
                                      "\tLUN: $lun_tmp\n\tProtocol: $protocol_tmp\n",
                              -msg_level => 5);
            }

            # save map for next call of function
            $self->{dfmap} = \%hDevs;

        }
    }

    # If not -cnt then assume all free disks
    # under lvm control are wanted
    unless (defined $cnt and length($cnt) and $cnt > 0) {
        $cnt  = 10000;
        $bCnt = 0;
    }

    @lvm_devices = $self->{parent}->_get_unlabeled_devices();
    if ($self->{parent}->error) {
        $self->_set_err($self->{parent}->errmsg);
        return (wantarray()) ? () : [];
    }

    # iterate through the devices and make sure that none of the disks
    # are being used other LVM disk object.
    foreach my $device (@lvm_devices) {
        if (defined $filer and length $filer) {
            $self->sanlog(-msg => "\nFilter filer: $filer\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i);
        }

        if (defined $vol and length $vol) {
            $self->sanlog(-msg => "\nFilter volume: $vol\nDevice: $device\nDevice volume: " . $self->{dfmap}{$device}{vol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i);
        }

        if (defined $lun and length $lun) {
            $self->sanlog(-msg => "\nFilter lun: $lun\nDevice: $device\nDevice lun: " . $self->{dfmap}{$device}{lun},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{lun} !~ /^$lun$/i);
        }

        if (defined $protocol and length $protocol) {
            $self->sanlog(-msg => "\nFilter protocol: $protocol\nDevice: $device\nDevice protocol: " . $self->{dfmap}{$device}{protocol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{protocol} !~ /$protocol/i);
        }

        # Filter on a filer list if provided by caller
        if (defined $filers and scalar(@{$filers}) != 0) {
            my $found = 0;
            # Look through each filer in the filers list
            foreach my $filer_f (@{$filers}) {
                $self->sanlog(-msg => "\nFilter filer(s): $filer_f\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                              -msg_level => 3);
                # If filer matches devices filer, set found to 1
                if ($self->{dfmap}{$device}{filer} =~ /^$filer_f$/i) {
                    $found = 1;
                }
            }

            # If the device is not part of the filers list caller requested skip it.
            if ($found == 0) {
                next;
            }
        }

        push @devices, $device if (++$total <= $cnt);
    }

    $cmd = "mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: DevMgr::get_free_devices() mount cmd failed");
        return  (wantarray()) ? () : [];
    }
   
    foreach my $device (@devices) {
        my $device_not_free = 0;
        my @tmp_array = grep(/\/dev\/$device/, @output);
        my @root_arr = split(/\s+/, $tmp_array[0]);

        # call pvdisplay to see if devices in in LV - don't check status
        $cmd = "pvdisplay | grep /dev\/$device";
        my ($pv_status, @pv_output) = &ClientUtil::execute($self->{conn}, $cmd);

        # skip if device is the root filesystem (on raw device or on LVM device)
        next if (($root_arr[2] eq "/") or (grep(/PV\s+Name\s+\/dev\/$device.*$/, @pv_output) > 0));

        if ($self->{parent}->{mp_type} =~ /dmmp/i) {
            #################################################################################
            # We need to remove devices which have partitions, like Root LUN etc                                                            
            # we get the follwoing outupt if we have partitions on device
            # 
            # fdisk -l /dev/mapper/mpath0
            # Disk /dev/mapper/mpath0: 18.2 GB, 18254659584 bytes
            # 255 heads, 63 sectors/track, 2219 cylinders
            # Units = cylinders of 16065 * 512 = 8225280 bytes
            #
            # Device Boot      Start         End      Blocks   Id  System
            # /dev/mapper/mpath0p1   *           1          13      104391   83  Linux
            # /dev/mapper/mpath0p2              14        2219    17719695   8e  Linux LVM
            #
            # If device is raw then
            # fdisk -l /dev/mapper/mpath7
            # Disk /dev/mapper/mpath7: 5368 MB, 5368709120 bytes
            # 255 heads, 63 sectors/track, 652 cylinders
            # Units = cylinders of 16065 * 512 = 8225280 bytes
            # 
            # Disk /dev/mapper/mpath7 doesn't contain a valid partition table
            ##################################################################################
            $cmd = "fdisk -l /dev/mapper/$device";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: LinuxDevMgr::get_free_devices() $cmd failed");
                return  (wantarray()) ? () : [];
            }
            if (grep(/Device\s+Boot\s+Start/i, @output)) {
                $device_not_free = 1;
            }
            if ($device_not_free != 1) {
                push (@free_devices, "/dev/mapper/" . $device);
            }
        } elsif ($self->{parent}->{mp_type} =~ /dmp/i) {
            push (@free_devices, "/dev/vx/dmp/" . $device);
        } else {
            push (@free_devices, "/dev/" . $device) ;
        }
    }

    # If user supplied cnt and free devices fell short
    # generate error
    if ($bCnt == 1 and scalar(@free_devices) < $cnt) {
        $self->_set_err("Error: LinuxDevMgr::get_free_devices() ",
                        "Only " . scalar(@free_devices) . " of $cnt requested devices were found");
        return (wantarray()) ? () : [];
    }
    $self->sanlog(-msg => "Devices returned: @free_devices", -msg_level=>2);
    return (wantarray()) ? @free_devices : \@free_devices;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new devices object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $dev, $devObj, $log, $lvm);
    my ($load_devices, $parent, $hosttype, $self, $status, @devices, @output);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: LinuxDevMgr::_new() Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?load_devices/) {
            if (not length($load_devices = $args{$_})) {
                $self->_set_err("Error: LinuxDevMgr::_new() Invalid value for argument '-load_devices'");
                return ($self);
            }
            next;
        }
    }

    # set default for load_devices (true)
    if (not defined($load_devices)) {
        $load_devices = 1;
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Error: LinuxDevMgr::_new()", "Invalid value for argument '-parent'");
            return $self;
        }
        $hosttype = $args{-parent}->{hostp}{hosttype};
    } else {
        $self->_set_err("Error: LinuxDevMgr::_new()", "Missing required argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn}   = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{devices} = [];

    # adding support for CentOS
    $cmd = "cat /etc/issue";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxDevMgr::_new() failed to start udev");
        return $self;
    } 

    # Add udev rule to get persistent entries for /dev/sdX devices
    # This rule need to be added just for rhel4 because all the other 
    # OSs (RHEL5, SLES9 and SLES10, OEL5) has this rule added to the list
    # of udev rules by default.
    if ($self->{parent}->{major_revision} =~ /rhel4/i or $output[0] =~ /CentOS\s+release\s+4/i) {
        
        $self->_update_scsi_id_conf_file();
        if ($self->error) {
            return $self;
        }

        $self->_add_udev_rule();
        if ($self->error) {
            return $self;
        } 

        # Start udevd 
        $cmd = "/sbin/start_udev";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: LinuxDevMgr::_new() failed to start udev");
            return $self;
        } else {
            $self->sanlog(-msg => "udev started on this client");
        }
    }

    if ($load_devices == 1) {
        @devices = $self->get_free_devices(%args);
        if ($self->error) {
            return $self;
        }
        foreach $dev (@devices) {
            # initialize device object from the contents of the device manager
            # configuration file.  Sometimes stale entries can appear from
            # previously failed runs.  If this happens, don't bother loading
            # the device.
            $devObj = $self->create_device(-name => $dev);
            if($self->error){
                return $self;
            }
            unless ($devObj->error) {
                $devObj->load();
                if ($devObj->error) {
                    $self->_set_err($devObj->errmsg);
                    return $self;
                }
            }
        }
    }

    return ($self);
}

################################################################################
# Subroutine Name: _add_udev_rule
# Objective:       Creates a temporary definition file in /tmp listing all
#                  used physical volumes
# Details:         See POD documentation at the end of this file
################################################################################
sub _add_udev_rule {
    my ($self) = @_;
    my ($cmd, $dev, $status, @output, $rule);

    # Most probably this (/etc/udev/rules.d/49_nate_by-id.rules) file will be unique
        $cmd = "cat /etc/udev/rules.d/49_nate_by-id.rules";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0 or scalar (@output) == 0) {
            $cmd = "touch /etc/udev/rules.d/49_nate_by-id.rules";
            my ($status, @output_tmp) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: LinuxDevMgr::_add_udev_rule() failed to create udev rule file");
                return $ClientUtil::FALSE;
            }

            # defining the rule. This rule says if kernel name for the device is sdX then create
            # scsi_id symlink for the device in /dev/disk/by-id dir
            my $rule = "KERNEL=\"sd*[!0-9]\", BUS=\"scsi\", PROGRAM=\"/sbin/scsi_id\", SYMLINK=\"disk/by-id/scsi-%c\"\n" .
                       "KERNEL=\"sd*[0-9]\", BUS=\"scsi\", PROGRAM=\"/sbin/scsi_id\", SYMLINK=\"disk/by-id/scsi-%c-part%n\"";
            $cmd = "echo '$rule' > /etc/udev/rules.d/49_nate_by-id.rules";
            ($status, @output_tmp) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: LinuxDevMgr::_add_udev_rule() failed to add rule \n$rule\n to udev rule file");
                return $ClientUtil::FALSE;
            }
        } elsif (@output > 0) {
            
            # if rule file exists but one or both required rules are missing from it, it will add missing rule(s).
            my $pattern1 = 'KERNEL="sd\*\[!0-9\]", BUS="scsi", PROGRAM="/sbin/scsi_id", SYMLINK="disk/by-id/scsi-%c"';
            my $pattern2 = 'KERNEL="sd\*\[0-9\]", BUS="scsi", PROGRAM="/sbin/scsi_id", SYMLINK="disk/by-id/scsi-%c-part%n"';
            if (not grep (/$pattern1/, @output)) {
                my $rule1 = 'KERNEL="sd*[!0-9]", BUS="scsi", PROGRAM="/sbin/scsi_id", SYMLINK="disk/by-id/scsi-%c"';
                $cmd = "echo '$rule1' >> /etc/udev/rules.d/49_nate_by-id.rules";
                my ($status, @output_tmp) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Error: LinuxDevMgr::_add_udev_rule() failed to add rule \n$pattern1\n to udev rule file");
                    return $ClientUtil::FALSE;
                }
            }
            if (not grep (/$pattern2/, @output)) {
                my $rule2 = 'KERNEL="sd*[0-9]", BUS="scsi", PROGRAM="/sbin/scsi_id", SYMLINK="disk/by-id/scsi-%c-part%n"';
                $cmd = "echo '$rule2' >> /etc/udev/rules.d/49_nate_by-id.rules";
                my ($status, @output_tmp) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Error: LinuxDevMgr::_add_udev_rule() failed to add rule \n$pattern2\n to udev rule file");
                    return $ClientUtil::FALSE;
                }
            }
        }
        return $ClientUtil::TRUE;
            
}    

################################################################################
# Subroutine Name: _update_scsi_id_conf_file
# Objective:       Creates a temporary definition file in /tmp listing all
#                  used physical volumes
# Details:         See POD documentation at the end of this file
################################################################################
sub _update_scsi_id_conf_file {
    my ($self) = @_;
    my ($cmd, $dev, $status, @output);
    
    $cmd = "touch out1 out2";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxDevMgr::_update_scsi_id_conf_file unable to create out1 out2 tmp files");
        return $ClientUtil::FALSE;
    }

    $cmd = "/bin/sed 's/^options=\\s*-b/\#\#\ &/' < /etc/scsi_id.config > out1";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxDevMgr::_update_scsi_id_conf_file unable to remove -b option");
        return $ClientUtil::FALSE;
    }
    
    $cmd = "/bin/sed 's/^vendor=.*\$/vendor=NETAPP\,\ options=-g/' < out1 > out2";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxDevMgr::_update_scsi_id_conf_file unable to add -g and vendor options");
        return $ClientUtil::FALSE;
    }

    $cmd = "cp -f out2 /etc/scsi_id.config";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxDevMgr::_update_scsi_id_conf_file unable to modify /etc/scsi_id.conf file");
        return $ClientUtil::FALSE;
    }

    $cmd = "rm -f out1 out2";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxDevMgr::_update_scsi_id_conf_file unable to delete out1 out2 tmp files");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _update_definition_file
# Objective:       Creates a temporary definition file in /tmp listing all
#                  used physical volumes
# Details:         See POD documentation at the end of this file
################################################################################
sub _update_definition_file {
    my ($self) = @_;
    my ($cmd, $dev, $status, @output);

    $self->_clear_err();

    $cmd = 'echo "';
    foreach $dev ($self->get_devices) {
        $cmd .= $dev->get_name;
        if ($dev->is_partitioned) {
            $cmd .= ":1";
        }
        $cmd .= "\n";
    }
    chomp($cmd);
    $cmd .= '" > /tmp/devmgr.cfg';
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: DevMgr::_update_definition_file Unable to create definition file", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}
################################################################################
# Subroutine Name: _get_definition_file
# Objective:       Returns array of device entries found in /tmp def file
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_definition_file {
    my ($self) = @_;
    my ($cmd, $dev, $status, @output);

    $self->_clear_err();

    $cmd = 'cat /tmp/devmgr.cfg';
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: DevMgr::_get_definition_file Definition file(/tmp/devmgr.cfg) is not present", @output);
        return ();
    }

    return @output;
}

1;
