# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Devices on Linux
## @author lippe@netapp.com knopp@netapp.com, rawat@netapp.com
## @status private
## @pod ./Dev.pod

# Package Name
##############
package OVMDev;

use strict;

# Package Requirements
######################
use File::Basename;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanLog;

use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: load
# Objective:       sets parameters based on devices current mount state
# Details:         See POD documentation at the end of this file
################################################################################
sub load {
    my ($self) = @_;
    my ($cmd, $name, $status, @output, $path);

    $self->_clear_err();

    $self->sanlog(-msg => "Populating $self->{name} data object");
    $name = $self->get_name();

    if ($self->{parent}->{parent}->{mp_type} =~ /none/i) {
        $cmd = "ls -l " . $self->{device_path};
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0 or scalar (@output == 0)) {
            $self->_set_err("Device $name doesnt exist");
            return $ClientUtil::FALSE;
        }
        my @sd_dev_info = split(/\s+/, $output[0]);

        # pull out the last column and split on slashes
        my @sd_dev_arr = split(/\//, $sd_dev_info[-1]);

        my $sd_dev = pop(@sd_dev_arr);
        $cmd = "mount |grep /dev/$sd_dev";
    } elsif($self->{parent}->{parent}->{mp_type} =~ /dmmp|dmp/i) {
        $cmd = "mount | grep $self->{device_path}";
    } else {
        $self->_set_err("Unsupported MP type specified");
        return $ClientUtil::FALSE;
    }

    # save response to mkfs prompt
    # will get mountpt and fstype and save it for future
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {

        # mount cmd will not fail. We are in this block it means that grep has failed
        # it means that devices is not mounted. using the other method to get fstype
        # of the device
        $cmd = "mount -f --guess-fstype $self->{device_path}";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            if ($output[0] =~ /unknown/) {
                $self->{fs} = "";
            } else {
                $self->_set_err("Recieved unexpected output", @output);
                return $ClientUtil::FALSE;
            }
        } else {
            $self->{fs} = $output[0];
        }
        $self->sanlog(-msg => "Loading " . $self->{device_path} ." \'mount --guess-fstype\' FS: " . $output[0] .
                              " , the device is not mounted so mount point cannot be determined");
        $self->{mount} = "";
    } else {
        my @dev_mount_info = split(/\s+/,$output[0]);
        $self->sanlog(-msg => "Loading " . $self->{device_path} .
                              " Mount: " . $dev_mount_info[2] . " FS: " . $dev_mount_info[4] . " ");
        $self->{mount} = $dev_mount_info[2];
        $self->{fs}    = $dev_mount_info[4];
        $self->sanlog(-msg => "Loading " . $self->{name} . " Mount: " . $self->{mount} .
                              " FS: " . $self->{fs} . " ");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_device_type
# Objective:       returns the type of this device (vhd,passthrough)
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device_type {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{device_type};
}

###############################################################################
# Method:    create_fs
# Objective: create a filesystem on a volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_fs {
    my ($self, %args) = @_;
    my ($cmd, $label, $partition, $type, $vgname, $name, $status, @output);

    $name = $self->get_name;

    # parse arguments
    foreach (keys %args) {
        if (/-?type$/) {
            $type = $args{$_};
            next;
        } elsif (/-?label$/) {
            if (not length($label = $args{$_})) {
                $self->_set_err("Invalid value for argument '-label'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -type argument
    if (not defined $type or $type eq "") {
        $type = "ext3";
    }

    # if -label is passed, make sure device is partitioned
    if (defined $label) {
        unless ($self->is_partitioned) {
            $self->_set_err("Device must contain partition to label");
            return $self;
        }
    }

    # set partition name if necessary
    if ($self->is_partitioned) {
        $name = $self->get_partition();
    }

    if ($type =~ /^ext2$|^ext3$/i) {
        # Create a filesystem
        $cmd = " yes | mke2fs ";
        $cmd .= "-j " if ($type =~ /ext3/i);
        $cmd .= $self->{device_path};
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to create '$type' filesystem on '$name'", @output);
            return $self;
        }

    } elsif ($type =~ /^reiserfs/) {
        $cmd = "mkfs.reiserfs " . $self->{device_path};
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to create '$type' filesystem on '$name'", @output);
            return $self;
        }
    } else {
        $self->_set_err("Invalid Filesystem type '$type'");
        return $self;
    }

    $self->{fs} = $type;

    # label the partition if the user passed -label
    if (defined $label) {
        $cmd = "/sbin/e2label " . $self->{device_path} . " $label";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to label partition", @output);
            return $self;
        }

        $self->{label} = $label;
    }

    return $self;
}

################################################################################
# Subroutine Name: mount_fs
# Objective:       mount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_fs {
    my ($self, %args) = @_;
    my ($cmd, $type, $vgname, $label, $mount, $name, $status, @output, $partition, @partitions);

    # parse arguments
    foreach (keys %args) {
        if (/-?path$/) {
            if (not length($mount = $args{$_})) {
                $self->_set_err("Invalid value for argument '-path'");
                return $self;
            }
            next;
        } elsif (/-?directio$/) {
            $self->sanlog(-msg => "Info: OVMDev::mount_fs() -directio option not available on linux");
            next;
        }
    }

    # Get attributes
    $name = $self->get_name;
    $type = $self->get_fs;

    if ($self->is_partitioned) {
        $name  = $self->get_partition();
        $label = $self->get_label();
    }

    # check if volume is unformatted
    unless (length $type) {
        $self->_set_err("No filesystem $type detected for device $name");
        return $self;
    }

    # check if the volume is already mounted
    if ($self->is_mounted()) {
        $self->_set_err("Device '$name' is already mounted");
        return $self;
    }

    # Use default path if not defined
    unless (defined $mount) {
        $mount = "/mnt/qa/" . $name . "/";
        $self->sanlog(-msg => "using default mount pt '$mount'");
    } else {
        $self->sanlog(-msg => "using mount pt '$mount'");
    }

    $self->{mount} = $mount;

    $cmd = "ls $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $cmd = "mkdir -m 777 -p $mount";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to create mount point directory for volume '$name'",
                            @output);
            return $self;
        }
    }

    # Mount the volume's filesystem
    $cmd = "mount -t $type " . $self->{device_path} . " $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to mount volume filesystem", @output);
        return $self;
    }

    # Open up permissions after the mount
    $cmd = "chmod 777 $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to change permissions on $mount", @output);
        return $self;
    }

    return $self;
}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, %args) = @_;
    my ($cmd, $vgname, $mount, $status, $force, $option, $partition, @output);

    $self->_clear_err();

    $option = "";

    # parse hash
    foreach (keys %args) {
        if (/-?force$/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $self;
            }
            next;
        }
    }

    $option = "-f" if (defined $force);

    # Get attributes
    $mount = $self->get_mount();

    # check if volume is mounted
    unless (length $mount) {
        $self->_set_err("Cannot unmount a filesystem that is not mounted");
        return $self;
    }

    # Unmount the volume's filesystem
    $cmd = "umount $option $self->{mount}";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to unmount volume's filesystem", @output);
        return $self;
    }

    $self->sanlog(-msg => "removing mount point '$mount'", -msg_level => 3);

    # Remove directory head
    $cmd = "rm -rf $self->{mount}";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to remove mount point '$mount'", @output);
        return $self;
    }

    # Remove mount pt metadata from object
    $self->{mount} = "";

    return $self;
}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if filesystem is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self, %args) = @_;
    my ($partition);

    $self->_clear_err();

    if (defined $self->{mount} and length $self->{mount}) {
        return 1;
    } else {
        return 0;
    }
}

################################################################################
# Subroutine Name: is_partitioned
# Objective:       returns ClientUtil::TRUE if partitioned
# Details:         See POD documentation at the end of this file
################################################################################
sub is_partitioned {
    my ($self) = @_;

    $self->_clear_err();

    if ($self->{partition}) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: partition
# Objective:       create a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub partition {
    my ($self) = @_;
    my ($cmd, $name, $num, $parObj, $size, $status, @output);

    $self->_clear_err();

    # check to see if device has already been partitioned, if so exit
    if ($self->{partition}) {
        $self->_set_err("'$self->{name}' already partitioned", @output);
        return $self;
    }

    # text file holds input to fdisk command
    $cmd = 'printf "n\np\n1\n\n\nw\n"' . " > /tmp/part.txt";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to write to /tmp/part.txt", @output);
        return $self;
    }

    # create partition using fdsisk
    $cmd = "fdisk " . $self->{device_path} . " < /tmp/part.txt";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to create partition on volume '$self->{name}'", @output);
        return $self;
    }

    # flag that device is partitioned
    $self->{partition} = $ClientUtil::TRUE;

    # set the name of the partition
    $self->{pname} = $self->get_name . "1";

    return $self;
}

################################################################################
# Subroutine Name: destroy_partition
# Objective:       destroy a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_partition {
    my ($self) = @_;
    my ($cmd, $name, $num, $parObj, $size, $status, @output);

    $self->_clear_err();

    # check to see if device has already been partitioned, if so exit
    unless ($self->is_partitioned) {
        $self->_set_err("'$self->{name}' is not partitioned");
        return $ClientUtil::FALSE;
    }

    # text file holds input to fdisk command
    $cmd = 'printf "d\nw"' . " > /tmp/part.txt";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to write to /tmp/part.txt", @output);
        return $ClientUtil::FALSE;
    }

    # deleting partition using fdisk
    $cmd = "fdisk " . $self->{device_path} . " < /tmp/part.txt";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to create partition on volume '$self->{name}'", @output);
        return $ClientUtil::FALSE;
    }

    # flag that device is not partitioned
    $self->{partition} = $ClientUtil::FALSE;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_uuid
# Objective:       Returns the UUID of the configured device.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self) = @_;
    my ($cmd, $status, $name, @output);

    $self->_clear_err();

    $name = $self->is_partitioned ? $self->{pname} : $self->{name};
    $cmd = "dumpe2fs -h " . $self->{device_path} . " | grep \"Filesystem UUID\" | awk '{print \$3}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to get UUID", @output);
        return "";
    }

    return $output[0];
}

################################################################################
# Subroutine Name: get_partition
# Objective:       Returns the name of the configured partition.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition {
    my ($self) = @_;

    $self->_clear_err();

    if ($self->is_partitioned) {
        return $self->{pname};
    } else {
        return undef;
    }
}

################################################################################
# Subroutine Name: get_partition_size
# Objective:       Returns the size of the specified partition in bytes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition_size {
    my ($self, $partition_num) = @_;

    my ($cmd, $status, @output, $name);

    $self->_clear_err();

    # check required partition number
    if (!defined($partition_num) or $partition_num !~ /\d+/) {
        $self->_set_err("partition number not passed to this method or format is incorrect");
        return $ClientUtil::FALSE;
    }

    # generate path to device via raw device.
    $name = "/dev/" . $self->get_name();

    # Get device information
    $cmd = "/sbin/fdisk -l $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to get device information for device ($name)");
        return $ClientUtil::FALSE;
    }

    # add partition
    $name .= $partition_num;

    # find partition in fdisk output
    my $fdisk_line = (grep(/^$name\s+/i, @output))[0];
    $fdisk_line =~ s/\*//;

    # check to see if you got a valid partion line
    if (!defined($fdisk_line) or $fdisk_line eq "") {
        $self->_set_err("Unable to get partion information for device ($name)");
        return $ClientUtil::FALSE;
    }

    # fdisk line:
    #    Device Boot      Start         End      Blocks   Id  System
    # /dev/sda1               1        1011     1048376+  83  Linux
    my ($partition, $start, $end, $size, $id, $system) = split(/\s+/, $fdisk_line);
    $size =~ s/\+//;
    my $size_bytes = $size * 1024;

    return $size_bytes;

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $refresh = 0;

    foreach (keys %args) {
        if (/-?refresh$/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Invalid value for argument '-refresh'");
                return 0;
            }
            next;
        }
    }

    # if $refresh is 1, _init_size() function will be called to initialize size of device again.
    # It is useful when size of device modified after it is created.
    if ($refresh == 1) {
        $self->{size} = $self->_init_size();
        return $self->{size};
    } else {
        if (defined($self->{size})) {
            return $self->{size};
        } else {
            $self->{size} = $self->_init_size();
            return $self->{size};
        }
    }
}

################################################################################
# Subroutine Name: get_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{name}) and length($self->{name})) {
        return $self->{name};
    } else {
        return "";
    }
}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       returns the raw device name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    # simply call get_name. Needed for Windows compatibility.
    return ($self->get_name());

}

################################################################################
# Subroutine Name: get_label
# Objective:       returns partition label of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_label {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{label}) and length($self->{label})) {
        return $self->{label};
    } else {
        return "";
    }
}

################################################################################
# Subroutine Name: get_mount
# Objective:       returns mount point of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{mount}) and length($self->{mount})) {
        return $self->{mount};
    } else {
        return "";
    }
}

################################################################################
# Subroutine Name: get_fs
# Objective:       returns fs of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{fs}) and length($self->{fs})) {
        return $self->{fs};
    } else {
        return "";
    }
}

################################################################################
# Subroutine Name: get_filer_name
# Objective:       Return lun name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_filer_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{filer}) and length($self->{filer})) {
        return $self->{filer};
    } else {
        return "";
    }
}

################################################################################
# Subroutine Name: get_lun
# Objective:       Return lun name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{lun}) and length($self->{lun})) {
        return $self->{lun};
    } else {
        return "";
    }
}

################################################################################
# Subroutine Name: refresh_geometry
# Objective:       refresh the geomery on device. Usefull to call after a
#                  resize on controller. NO-OP on this platform
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh_geometry {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new dev object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($cmd, $conn, $log, $mp_obj, $mount, $name, $parent, $psize, $self, $status, @output);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Missing value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Missing required argument '-name'");
        return $self;
    }

    # check for existance of required -parent argument
    unless (defined $parent) {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn}   = $conn;
    $self->{parent} = $parent;
    $mp_obj = $parent->{parent}->{mp};

    # "fdisk -l" is used to find out the partitions of the device
    # Each line starting with /dev/sd will be a partition of the device
    # Sample output of "fdisk -l <device>"
    #---------------------------------------------
    #   Disk /dev/sda: 73.4 GB, 73407488000 bytes
    #   255 heads, 63 sectors/track, 8924 cylinders
    #   Units = cylinders of 16065 * 512 = 8225280 bytes

    #   Device Boot      Start         End      Blocks   Id  System
    #   /dev/sda1   *           1          13      104391   83  Linux
    #   /dev/sda2              14        8924    71577607+  8e  Linux LVM
    #--------------------------------------------------------------------

    if ($parent->{parent}->{mp_type} =~ /none/i) {
        $name =~ s/\/dev\///;

        # Get all the partition existing on the disk
        $cmd = "fdisk -l /dev/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to retrieve partitions of the dev /dev/$name",
                            @output);
            return $self;
        }

        # this will give us the number of partitions
        my $no_of_partitions = grep (/^\/dev\/sd/ , @output);
        if ($no_of_partitions == 0) {
            $self->{name}      = $name;
            $self->{partition} = $ClientUtil::FALSE;

        } elsif ($no_of_partitions == 1) {
            $self->{name}      = $name;
            $self->{partition} = $ClientUtil::TRUE;
            $self->{pname}     = $name . "1";

        } elsif ($no_of_partitions > 1) {
            $self->_set_err("Currently we only support devices with single partition" .
                            " and the device $name has $no_of_partitions partitions");
            return $self;
        }

        if ($self->{partition}) {
            $name = $self->{pname};
        }

        # Store id of the device which is persistent across iscsi restart(host reboots)
        $cmd = "ls -l /dev/disk/by-id |grep '\/$name'";
        my ($status, @disk_by_id) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("No persistent entry found " .
                            "for device /dev/$name in /dev/disk/by-id");
            return $self;
        }
        # pull off the third from last column from the ls -l output and store it in disk_by_id
        # this is the scsi-\d+ id
        my @disk_info_arr = split (/\s+/, $disk_by_id[0]);
        $self->{disk_by_id} = $disk_info_arr[-3];
        $self->{device_path} = "\/dev\/disk\/by-id\/" . $self->{disk_by_id};
    } elsif ( $parent->{parent}->{mp_type} =~ /dmp/i) {
        ($self->{name}) = ($name =~ /\/dev\/vx\/dmp\/(\w+)/);
        $self->{device_path} = $name;
    } else {
        ($self->{name}) = ($name =~ /(360a9\w+|3600a09\w+|mpath\w+)/i);
        $self->{device_path} = "\/dev\/mapper\/" . $self->{name};
    }

    # get filer lun associated with physical dev
    # get_lun_path expects input to be a sdX device
    # but if multipath is enabled then the device is mpath(lun id starting with 360a9 or 3600a09)
    # -ed device So we have stored the mapping between mpath device and a path of that
    # device. We will pass the path as device if my_type is not none.
    if ($parent->{parent}->{mp_type} =~ /none/i) {
        $self->{lun} = $parent->{sanlun}->get_lun_path(-device => "/dev/$name");

        # Store Filer name for the device
        $self->{filer} = $parent->{sanlun}->get_filer(-device => "/dev/$name");
    } else {
        my $mp_dev = $self->{name};
        my $sd_dev = $mp_obj->{mapping}->{$mp_dev}->{sdX};
        $self->{lun} = $parent->{sanlun}->get_lun_path(-device => "/dev/$sd_dev");
        if ($parent->{sanlun}->error) {
            return $self;
        }

        # Store Filer name for the device
        $self->{filer} = $parent->{sanlun}->get_filer(-device => "/dev/$sd_dev");
        if ($parent->{sanlun}->error) {
            return $self;
        }
    }

    my $lun_path = $self->{lun};
    my $regex = $parent->{vhd_lun_path_regex};

    if (defined($regex)) {
        # use the regex defined in the devmgr to differentiate between vhd and passthrough
        if ($lun_path !~ /$regex/) {
            $self->{device_type} = "passthrough";
        } else {
            $self->{device_type} = "vhd";
        }
    } else {
        # otherwise default to "OVM" luntypes as being default luntype for VHD devices
        if ($parent->{parent}->hosttype() !~ /ovm/i) {
            $self->{device_type} = "passthrough";
        } else {
            $self->{device_type} = "vhd";
        }
    }

    if ($parent->{sanlun}->error) {
        $self->_set_err($parent->{sanlun}->errmsg());
        return $self;
    }

    $self->{fs}    = "";
    $self->{mount} = "";

    $self->{size} = $self->_init_size();

    return $self;
}

################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing disks
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self) = @_;

    my ($cmd, $size, $name, @output, $sectors, $status);

    $size = "";

    my $path = $self->{parent}->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show $self->{lun}";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to retrieve size (bytes) for disk \'$self->{lun}\'",
                        @output);
        return 0;
    }

    # get the size in bytes
    my $lun = $output[1];
    $lun =~ /\((\d+)\)/;
    $size = $1;

    return $size;
}

1;
