# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Devices on Solaris
## @author thomasm@netapp.com
## @status private
## @pod ./Dev.pod

# Package Name
##############
package SolarisDev;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanUtil;
use Error;

use vars qw($NEW_DEV_LABEL @ISA);

$NEW_DEV_LABEL = "NATE";

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: load
# Objective:       sets parameters based on devices current mount state
# Details:         See POD documentation at the end of this file
################################################################################
sub load {
    my ($self) = @_;

    my ($cmd, $name, $partition, $phash, $status, $type, @output);

    $self->_clear_err();

    $self->sanlog(-msg => "Populating $self->{name} data object");

    ### save response to mkfs prompt
    $cmd = "mount | grep /dev/dsk/$self->{name} | awk '{print \$1}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Dev::load Mount command failed", @output);
        return $ClientUtil::FALSE;
    }

    $self->{mount} = $output[0];

    ### save response to mkfs prompt
    $cmd = "fstyp /dev/dsk/$self->{name}";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Dev::load Fstyp command failed", @output);
        return $ClientUtil::FALSE;
    }

    $self->{fs} = $output[0];

    $self->sanlog(-msg => "Loading $self->{name}  Mount: $self->{mount} FS: $self->{fs}");

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: create_fs
# Objective:       create and return a filesystem object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_fs {
    my ($self, %args) = @_;

    my ($cmd, @output, $partition, $phash, $status, $type);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?type$/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: Dev::create_fs()", "Invalid value for argument '-type'");
                return $self;
            }
            next;
        } 
    }

    ### check for existance of required -type argument
    unless (defined $type) {
        $self->_set_err("Error: Dev::create_fs()", 
                        "Missing required argument '-type'");
        return $self;
    }

    # if type was not passed, then default to ufs
    if (not defined $type or $type eq "") {
        $type = "ufs";
    }

    if ($type =~ /vxfs/i) {
        $cmd = "/usr/sbin/mkfs -F $type -o largefiles /dev/rdsk/$self->{name}";
    } elsif ($type =~ /ufs/i) {
        $cmd = "yes | /usr/sbin/newfs /dev/rdsk/$self->{name}";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Dev::create_fs Unable to create $type filesystem on volume '$self->{name}'",
                        @output);
        return $ClientUtil::FALSE;
    }

    $self->{fs} = $type;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: mount_fs
# Objective:       mount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_fs {
    my ($self, %args) = @_;

    my ($cmd, $directio, $type, $vgname, $mount, $options, $name);
    my ($status, @output);

    ### Get attributes
    $directio = 0;
    $name     = $self->get_name();
    $type     = $self->get_fs();
    $mount    = undef;

    ### check if volume is unformatted
    unless (length $type) {
        $self->_set_err("Error: Dev::mount_fs()", "No filesystem detected");
        return $ClientUtil::FALSE;
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?path$/) {
            if (not length($mount = $args{$_})) {
                $self->_set_err("Error: Dev::mount_fs()", "Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?directio$/) {
            if (not length($directio = $args{$_}) or $args{$_} !~ /0|1/) {
                $self->_set_err("Error: Dev::mount_fs()", "Invalid value for argument '-directio'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### check if the volume is already mounted
    if ($self->is_mounted()) {
        $self->_set_err("Error: Dev::mount_fs() device '$name' is already mounted");
        return $ClientUtil::FALSE;
    }

    ### Use default path if not defined
    unless (defined $mount) {
        $mount = "/mnt/qa/$name";
        $self->sanlog(-msg => "using default mount pt '$mount'");
    } else {
        $self->sanlog(-msg => "using mount pt '$mount'");
    }

    $self->{mount} = $mount;

    ### Create the mount point and open up dir privs
    $cmd = "mkdir -p $mount; chmod 777 $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Dev::mount_fs() Unable to create mount point directory for volume '$name'", @output);
        return $ClientUtil::FALSE;
    }

    $options = "largefiles";
    if ($directio) {
        if ($type =~ /ufs/i) {
            $options .= ",forcedirectio";
        } else {
            $options .= ",convosync=direct";
        }
    }

    ### Mount the volume's filesystem
    $cmd = "/usr/sbin/mount -F $type -o $options /dev/dsk/$name $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        ### device needs to be fsck
        if ("@output" =~ /fsck/i) {
            $self->sanlog(-msg => "device needed to be fsck'd");
            ($status, @output) = &ClientUtil::execute($self->{conn}, "fsck -y /dev/dsk/" . $self->get_name());
        }

        ### try to mount it again
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Dev::mount_fs() Unable to mount volume filesystem", @output);
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, %args) = @_;

    my ($cmd, $type, $vgname, $fs, $mount, $name, $status, $force);
    my ($option, @output);

    $self->_clear_err();

    $option = "";

    # parse hash
    foreach (keys %args) {
        if (/-?force$/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Dev::unmount_fs()", "Invalid value for argument '-force'");
                return $self;
            }
            next;
        } 
    }

    $option = "-f" if (defined $force);

    ### Get attributes
    $name  = $self->get_name();
    $mount = $self->get_mount();
    $fs    = $self->get_fs();

    ### check if device contains a filesystem
    unless (length $fs) {
        $self->_set_err("Error: Dev::unmount_fs()", 
                        "No filesystem detected");
        return $ClientUtil::FALSE;
    }

    ### check if volume is mounted
    unless (length $mount) {
        $self->_set_err("Error: Dev::unmount_fs()", 
                        "Cannot unmount a filesystem that is not mounted");
        return $ClientUtil::FALSE;
    }

    ### Unmount the volume's filesystem
    $cmd = "umount $option /dev/dsk/$name";

    my $loop = 60;
    foreach my $tries (1 .. $loop) {
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Dev::unmount_fs() Unable to unmount volume's filesystem", @output);
            if ("@output" =~ /busy/i and $tries < $loop) {
                if ($tries == 1) {
                    $self->sanlog(-msg => "Info: device is busy, I'll try to unmount for another $loop seconds");
                }
                &SanUtil::nate_sleep(1);
                next;
            }
            return $ClientUtil::FALSE;
        }
        last;
    }

    ### Remove mount pt metadata from object
    $self->{mount} = "";

    $self->sanlog(-msg => "removing mount point '$mount'");

    ### Remove directory head
    my ($base, $parent) = fileparse($mount);
    if ($mount ne "/") {
        $cmd = "rm -rf $mount";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Dev::unmount_fs() Unable to remove mount point '$mount'", @output);
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if filesystem is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{mount} and length $self->{mount}) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: is_partitioned
# Objective:       returns ClientUtil::TRUE if partitioned
# Details:         See POD documentation at the end of this file
################################################################################
sub is_partitioned {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: partition
# Objective:       create a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub partition {
    my ($self, %args) = @_;

    $self->_clear_err();

    ### noop on solaris

    return $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: destroy_partition
# Objective:       destroy a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_partition {
    my ($self, %args) = @_;

    $self->_clear_err();

    ### noop on solaris

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return size of the entire LUN
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{size})) {
        return $self->{size};
    } else {
        $self->_set_err("Error: Vol::get_size() Vol object is missing size");
        return undef;
    }

}

################################################################################
# Subroutine Name: get_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{name}) and length($self->{name})) {
        return $self->{name};
    } else {
        $self->_set_err("Error: Dev::get_name() Dev object is missing name");
        return undef;
    }

}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       returns the raw device name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    # Get device name, check if an error occurred and return undef
    # if there was an error.
    my $raw_device_name = $self->get_name();
    if ($self->error()) {
        return undef;
    }

    # Prepend /dev/rdsk/ to the device name, this provides the raw
    # path to the device under Solaris.
    $raw_device_name = "/dev/rdsk/" . $raw_device_name;

    return $raw_device_name;
}

################################################################################
# Subroutine Name: get_mount
# Objective:       returns mount point of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{mount}) and length($self->{mount})) {
        return $self->{mount};
    } else {
        $self->_set_err("Error: Dev::get_mount() Dev object is missing mount");
        return undef;
    }

}

################################################################################
# Subroutine Name: get_fs
# Objective:       returns fs of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{fs}) and length($self->{fs})) {
        return $self->{fs};
    } else {
        $self->_set_err("Error: Dev::get_fs() Dev object missing fs");
        return undef;
    }

}

################################################################################
# Subroutine Name: get_lun
# Objective:       Return lun name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{lun}) and length($self->{lun})) {
        return $self->{lun};
    } else {
        $self->_set_err("Error: Dev::get_lun() Dev object is missing lun name");
        return undef;
    }

}

################################################################################
# Subroutine Name: get_partition
# Objective:       Returns the name of the configured partition.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition {
    my ($self) = @_;

    $self->_clear_err();

    return undef;

}

################################################################################
# Subroutine Name: get_label
# Objective:       Returns the name of the configured partition's label.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_label {
    my ($self) = @_;

    $self->_clear_err();

    return undef;

}

################################################################################
# Subroutine Name: get_partition_size
# Objective:       Returns the size of the specified partition in bytes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition_size {
    my ($self, $partition_num) = @_;

    my ($cmd, $status, @output, $name);

    $self->_clear_err();

    # check required partition number
    if (!defined($partition_num) or
        ($partition_num !~ /\d+/ and $partition_num !~ /s\d+/)) {
        $self->_set_err("partition number not passed to this method or format is incorrect");
        return $ClientUtil::FALSE;
    }

    # generate path to device via raw device. Exclude
    # "s2" since we are going to be looking at the passed
    # in slice (partition)
    $name = "/dev/rdsk/" . $self->get_name();
    $name =~ s/s2$//;

    # allow of the user to pass either <num> or "s<num>" 
    # syntax
    if ($partition_num =~ /^s/) {
        $name .= $partition_num;
    } else {
        $name .= "s" . $partition_num;
    }

    # Get device information
    $cmd = "/usr/sbin/devinfo -p $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to get device information for device ($name)");
        return $ClientUtil::FALSE;
    }

    # devinfo returns:
    # devinfo -p /dev/rdsk/c3t20000004CF6221BDd0s0
    # /dev/rdsk/c3t20000004CF6221BDd0s0  76      c0   4194828 23920920  0  2
    my ($partition, $major, $minor, $start, $size, $flag, $tag) = split(/\s+/, $output[0]); 
    my $size_bytes = $size * 512;

    return $size_bytes;

}

################################################################################
# Subroutine Name: refresh_geometry
# Objective:       refresh the geomery on device. Usefull to call after a 
#                  resize on controller.
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh_geometry {
    my ($self) = @_;

    my ($cmd, $status, @output, $name, @format_cmds);

    $self->_clear_err();

    # get device name
    $name = $self->get_name();

    # define location of some temporary files to conduct formats on.
    my $filename = "/usr/tmp/format_input." . $$;

    # specify commands to use. Add label of device after reset of
    # geometry
    push(@format_cmds, "type");
    push(@format_cmds, "0");
    push(@format_cmds, "volname $NEW_DEV_LABEL");
    push(@format_cmds, "label");
    push(@format_cmds, "quit");

    # create file with commands to execute to format
    &ClientUtil::create_remote_file($self->{conn}, $filename, @format_cmds);

    # format device
    $cmd = "NOINUSE_CHECK=1;export NOINUSE_CHECK;/usr/sbin/format -f $filename $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    &SanUtil::nate_sleep(1);
    if ($status > 0) {
        $self->_set_err("Error: SolarisDev::refresh_geometry() ", 
                        "format returned a bad status ($status))");
        ### Cleanup temporary format input files
        &ClientUtil::remove_remote_file($self->{conn}, $filename);
        return $ClientUtil::FALSE;
    } else {
        ### Cleanup temporary format inout files
        &ClientUtil::remove_remote_file($self->{conn}, $filename);
    }

    # update size attribute for newly changed size
    $self->{size} = $self->_init_size();
    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new dev object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $conn, $log, $name, $parent, $psize, $self, $status, @output);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: Dev::new()", "Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: Dev::new()", "Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Dev::new()", "Missing value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    ### check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: Dev::new()", 
                        "Missing required argument '-name'");
        return $self;
    }

    ### check for existance of required -parent argument
    unless (defined $parent) {
        $self->_set_err("Error: Dev::new()", 
                        "Missing required argument '-parent'");
        return $self;
    }

    ### check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: Dev::new()", 
                            "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: Dev::new()", 
                        "Missing required argument '-conn'");
        return $self;
    }

    $self->{log} = $log if (defined $log);

    ### initilize the object members
    $self->{conn}   = $conn;
    $self->{parent} = $parent;

    $self->{name} = $name;
    $name =~ s/s2//;

    ### get filer lun associated with physical dev
    $self->{lun} = $parent->{sanlun}->get_lun_path(-device => $name);
    if ($parent->{sanlun}->error) {
        $self->_set_err($parent->{sanlun}->errmsg());
        return $self;
    }

    $self->{fs}    = "";
    $self->{mount} = "";
    $self->{size}  = $self->_init_size();

    return $self;

}

################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing disks
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self) = @_;

    my ($cmd, $size, $name, @output, $status);

    $size = "";

    my $path = $self->{parent}->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show $self->{lun}";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Dev::_init_size() Unable to retrieve size (bytes) for disk \'$self->{lun}\'", @output);
        return "";
    }

    # get the size in bytes
    my $lun = $output[1];
    $lun =~ /\((\d+)\)/;
    $size = $1;

    return $size;

}

1;
