# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to DevMgr for Solaris host
## @author lippe@netapp.com knopp@netapp.com
## @status private
## @pod ./DevMgr.pod

# Package Name
##############
package San::SanClient::DevMgr::SolarisDevMgr;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::DevMgr;
use San::SanClient::DevMgr::SolarisDev;
use Error;
use San::SanLog;
use Data::Dumper;

use vars qw(@ISA);

@ISA = qw(Error SanLog DevMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_device
# Objective:       return a device object based on the device name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device {
    my ($self, $name) = @_;

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: SolarisDevMgr::get_device() name is a required argument");
        return;
    }

    foreach my $obj ($self->get_devices) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: SolarisDevMgr::get_device() Unable to find device '$name'");

    return;
}

################################################################################
# Subroutine Name: get_devices
# Objective:       return all currently configured device objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;

    my (@devices);

    $self->_clear_err();

    if (@{$self->{devices}}) {
        @devices = @{$self->{devices}};
    } else {
        return ();
    }

    return (wantarray()) ? @devices : \@devices;

}

################################################################################
# Subroutine Name: create_device
# Objective:       create a device object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_device {
    my ($self, %args) = @_;

    my ($devObj);

    $self->_clear_err();

    $devObj = SolarisDev->_new(-parent => $self,
                               -conn   => $self->{conn},
                               -loghdl => $self->{log},
                               %args);

    if ($devObj->error) {
        $self->_set_err($devObj->errmsg());
        return $self;
    }

    push @{$self->{devices}}, $devObj;

    return ($devObj);

}

################################################################################
# Subroutine Name: get_dev_id
# Objective:       returns unique device name(with full path) of passed device's
#                  short name to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_id {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: get_dev_name
# Objective:       returns device short name of passed device's unique name(long name)
#                  to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_name {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: cleanup
# Objective:       remove device objects and remove mount point entries
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup {
    my ($self, @devices) = @_;

    my ($devObj, $fail_cnt);

    $self->_clear_err();

    # remove mount point entry for each device passed in
    foreach $devObj ($self->get_devices) {

        $self->sanlog(-msg => "Cleaning up mount points for device: $devObj->{name}");

        $devObj->unmount_fs() if $devObj->is_mounted;
        if ($devObj->error) {
            $self->_set_err($devObj->errmsg);
            $fail_cnt++;
        }
    }

    $self->{devices} = [];

    return ($fail_cnt == 0) ? $ClientUtil::TRUE : $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;

    my ($bCnt, $cmd, $cnt, $dev, $device, $done, $filer, $filers, $protocol);
    my ($filername,  $line,  $lun, $lun_tmp, $refresh, $status, $total);
    my ($vol, %hDevs, @attr, @buffer, @devices, $device_tmp);
    my (@sanlun_output, @split, @tmp_array, @tmp_devices);
    my (@new_devices, @cur_devices);

    $total = 0;

    $self->_clear_err();

    # Asssume user supplied -cnt argument
    $bCnt    = 1;
    $refresh = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Error: SolarisDevMgr::get_free_devices() Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Error: SolarisDevMgr::get_free_devices() Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Error: SolarisDevMgr::get_free_devices() Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: SolarisDevMgr::get_free_devices() Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {

                $self->_set_err("Error: SolarisDevMgr::get_free_devices() Invalid value for argument '-vol'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Error: SolarisDevMgr::get_free_devices() Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: SolarisDevMgr::get_free_ evices() Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Check if a device to filer map 'dfmap' has already been created or
    # refresh it if requested.
    if (not defined $self->{raw_device_output} or not defined $self->{dfmap} or $refresh eq 1) {
        $self->sanlog(-msg => "Refreshing device to filer, vol and lun map");
        if ($self->{parent}->{mp}->{mp_type} =~ /dmp/i) {
            $self->_load_dfmap_dmp();
            if ($self->error()) {
                return (wantarray()) ? () : [];
            }
        } elsif ($self->{parent}->{mp}->{mp_type} =~ /mpxio/i) {
            $self->_load_dfmap_mpxio();
            if ($self->error()) {
                return (wantarray()) ? () : [];
            }
        } else {
            $self->_set_err("Error: SolarisDevMgr::get_free_devices() Invalid value for MP type " .
                            "of '$self->{parent}->{mp}->{mp_type}' should be 'dmp' or 'mpxio'");
            return (wantarray()) ? () : [];
        }
    }

    # If not -cnt then assume all free disks are wanted
    unless (defined $cnt and length($cnt) and $cnt > 0) {
        $cnt  = 10000;
        $bCnt = 0;
    }

    # Both _load_dfmap_dmp and _load_dfmap_mpxio set raw_device_output so
    # we should not need to get the sanlun output again.
    @sanlun_output = @{$self->{raw_device_output}};

    # traverse through output and get a unique path to
    # each <Filer>:<LUN> grouping.
    @tmp_devices = ();

    while (scalar(@sanlun_output)) {
        $line = shift(@sanlun_output);

        # grab first line and extract the lun path
        # and the device entry.

        # Make sure the line is what we expect it to be:
        # filer:/vol/vol_path/lun (LUN #) dev_path
        # filer:  /vol/vol_path/lun  dev_path for MPXIO
        # Also remove a known string that sanlun is prepending
        # to the first lun it finds (hack).
        $line =~ s|Report bugs to <bug-findutils\@gnu.org>.||;
        unless ($line =~ /^(.*:)\s*(\S*)\s+\(*.*\)*\s+\/dev\/rdsk\/(\w*)\s?.*$/ or $line =~/(\S+)\s+(\S+)\s+\/dev\/rdsk\/(\S+)s2\s+\S+\s+(\S+)\s+/) {
            next;
        }

        my $filer_name = $1 ;
        my $lun_path   = $2;
        my $lun_name   = "$filer_name" . "$lun_path";

        my $device     = $3;
        $device =~ s|/dev/rdsk/||;

        # reduce array (removing paths to that lun path)
        @sanlun_output = grep(!/$filer_name\s*$lun_path\s+/, @sanlun_output);
        push(@tmp_devices, $device);

        next;
    }

    # iterate through the devices and make sure that none of the disks
    # are being used an LVM disk object.
    foreach $dev (@tmp_devices) {
        $self->sanlog(-msg       => "Apply filters to any of the free devices found",
                      -msg_level => 3);

        if ($dev) {
            my $hash_dev = $dev;
            $hash_dev =~ s/s2//;

            $self->sanlog(-msg       => "Filtering device:\n" .
                                        "filer:    $self->{dfmap}{$hash_dev}{filer}\n" .
                                        "volume:   $self->{dfmap}{$hash_dev}{vol}\n"   .
                                        "lun:      $self->{dfmap}{$hash_dev}{lun}\n",
                                        "protocol: $self->{dfmap}{$hash_dev}{protocol}\n",
                          -msg_level => 3);
            if (defined $filer) {
                next if ($self->{dfmap}{$hash_dev}{filer} !~ /^$filer$/i);
                $self->sanlog(-msg       => "Filer ($filer) matched",
                              -msg_level => 3);
            }
            if (defined $vol) {
                next if ($self->{dfmap}{$hash_dev}{vol} !~ /^$vol$/i);
                $self->sanlog(-msg       => "Volume ($vol) matched",
                              -msg_level => 3);
            }

            if (defined $lun) {
                next if ($self->{dfmap}{$hash_dev}{lun} !~ /^$lun$/i);
                $self->sanlog(-msg       => "Lun ($lun) matched",
                              -msg_level => 3);
            }

            if (defined $protocol) {
                next if ($self->{dfmap}{$hash_dev}{protocol} !~ /$protocol/i);
                $self->sanlog(-msg       => "Protocol ($protocol) matched",
                              -msg_level => 3);
            }

            # Filter on a filer list if provided by caller
            if (defined $filers and scalar(@{$filers}) != 0) {
                my $found = 0;
                # Look through each filer in the filers list
                foreach my $filer_f (@{$filers}) {
                    if ($self->{dfmap}{$hash_dev}{filer} =~ /^$filer_f$/i) {
                        $self->sanlog(-msg       => "Filer ($filer_f) matched",
                                      -msg_level => 3);
                        # If filer matches devices filer, set found to 1
                        $found = 1;
                    }
                }

                # If the device is not part of the filers list caller requested skip it.
                if ($found == 0) {
                    next;
                }
            }
        }

        if ($self->{parent}->{mp_type} !~ /mpxio/i) {
            # Check if device is valid. Veritas 3.5MP3 does not properly clean
            # up devices that are removed.
            $cmd = "vxdisk list " . $dev . " | grep \"group:\"";
            ($status, @buffer) = &ClientUtil::execute($self->{conn}, $cmd);

            # If the disk isn't visible to Veritas, go ahead and use it
            if ($status) {
                push @devices, $dev if (++$total <= $cnt);
                next;
            } else {
                # Check to see if the disk use to be in a VG.  A disk that was removed
                # from a VG will still list the "group:" field, but the name will be blank"
                # vxdisk list das_vs00_4|grep group:
                # group:     name= id=1291853032.17.auto-sunv440-01

                # If the disk isn't a member of a Veritas VG, use it
                if ((grep /group:\s+name=\s+id/i, @buffer) > 0) {
                    # The disk is no longer in use
                    push @devices, $dev if (++$total <= $cnt);
                    next;
                } elsif ((grep /not in the configuration/i, @buffer) > 0) {
                    push @devices, $dev if (++$total <= $cnt);
                    next;
                }
            }
        } else {
            push @devices, $dev if (++$total <= $cnt);
            next;
        }
    }

    # If user supplied cnt and free devices fell short
    # generate error
    if ($bCnt == 1 and scalar(@devices) < $cnt) {
        $self->_set_err("Error: SolarisDevMgr::get_free_devices() ",
                        "Only " . scalar(@devices) .
                        " of $cnt requested devices were found");
    }

    my $msg = join("\n", @devices);
    $self->sanlog(-msg => "Devices returned:\n" . $msg);

    return (wantarray()) ? @devices : \@devices;

}


################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new devices object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $dev, $devObj, $log, $loadcfg, $lvm, $parent);
    my ($load_devices, $hosttype, $self, $status, @devices, @output);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: SolarisDevMgr::_new() Invalid value for argument '-parent'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?load_devices/) {
            if (not length($load_devices = $args{$_})) {
                $self->_set_err("Error: SolarisDevMgr::_new() Invalid value for argument '-load_devices'");
                return ($self);
            }
            next;
        }
    }

    # Allow CLI commands to be logged;
    $self->{log} = $log if (defined $log);

    # set default for load_devices (true)
    if (not defined($load_devices)) {
        $load_devices = 1;
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanClient/i) {
            $self->_set_err("Error: SolarisDevMgr::_new()",
                            "Invalid value for argument '-parent'");
            return $self;
        }
        $hosttype = $args{-parent}->{hostp}{hosttype};
    } else {
        $self->_set_err("Error: SolarisDevMgr::_new()",
                        "Missing required argument '-parent'");

        return $self;
    }

    # initilize the object members
    $self->{conn}   = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{devices} = [];

    if ($load_devices == 1) {
        @devices = $self->get_free_devices(%args);
        return($self) if ($self->error);

        foreach $dev (@devices) {
            # configuration file. Sometimes stale entries can appear from
            # previously failed runs. If this happens, don't bother loading
            # the device.
            $devObj = $self->create_device(-name => $dev);
            unless ($devObj->error) {
                $devObj->load();
            }
        }
    }

    # We need to undefine $self-{dfmap} because at this point it will not be fully
    # populated and most methods won't have expected it to be defined at this point.
    $self->{dfmap} = undef;

    return ($self);

}

################################################################################
# Subroutine Name: _load_dfmap_mpxio
# Objective:       load the dfmap for mp_type mpxio.
################################################################################
sub _load_dfmap_mpxio {
    my ($self, %args) = @_;

    my (@sanlun_output,%hDevs);
    # MPXIO only needs the sanlun lun show all output
    @sanlun_output = $self->{sanlun}->lun_show(-_primary_paths_only => 1,
                                               -mp_type => $self->{parent}->{mp}->{mp_type});
    if ($self->{sanlun}->error) {
        $self->_set_err("Error: SolarisDevMgr::_load_dfmap_mpxio() ".$self->{sanlun}->errmsg());
        return (wantarray()) ? () : [];
    }

    $self->{raw_device_output} = \@sanlun_output;

    foreach my $line (@sanlun_output) {
        # Also remove a known string that sanlun is prepending
        # to the first lun it finds (hack).
        $line =~ s|Report bugs to <bug-findutils\@gnu.org>.||;
        $line =~ s/^\s+//g;

        $self->sanlog(-msg       => "Parsing Sanlun line: $line",
                      -msg_level => 3);

        # Figure out what we are dealing with from sanlun
        # old sanlun kit output looks like below,
        # controller:                 lun-pathname                                 device filename                     adapter   protocol          lun size         lun state
        # iscsi_vs:  /vol/iscsi_vol_1/iscsi_auto-sunx2200-01_0  /dev/rdsk/c0t600A0980324668762D3F41734170386Fd0s2  scsi_vhci0  iSCSI          5g (5368709120)     GOOD
        #
        # new sanlun kit output looks like below,
        # controller(7mode)/                                             device                                            host                  lun
        # vserver(Cmode)       lun-pathname                              filename                                          adapter    protocol   size    mode
        # ---------------------------------------------------------------------------------------------------------------------------------------------------
        # iscsi_vs             /vol/iscsi_vol_1/iscsi_auto-sunx2200-01_0 /dev/rdsk/c0t600A0980324668762D3F41734170386Fd0s2 scsi_vhci0 iSCSI      5g      C
        my ($filer, $vol, $lun, $device, $protocol);
        if ($line =~ /(\S+)\s+\/vol\/(\S+)\/(\S+)\s+\/dev\/rdsk\/(\S+)s2\s+\S+\s+(\S+)\s+/) {
            ($filer,$vol,$lun,$device,$protocol) = ($1,$2,$3,$4,lc($5));
        } elsif ($line =~ /(\S+)\s+\/vol\/(\S+)\/(\S+)\s+\/dev\/rdsk\/(\S+)s2/) {
            ($filer,$vol,$lun,$device,$protocol) = ($1,$2,$3,$4,"");
        } else {
            next;
        }

        # Remove colon in the end of filer/vserver name if presents
        $filer =~ s/:$//;

        $hDevs{$device} = {
                                filer    => $filer,
                                vol      => $vol,
                                lun      => $lun,
                                protocol => $protocol
                              };
    }

    # save map for next call of function
    $self->{dfmap} = \%hDevs;
}

################################################################################
# Subroutine Name: _load_dfmap_dmp
# Objective:       load the dfmap for mp_type dmp.
################################################################################
sub _load_dfmap_dmp {
    my ($self, %args) = @_;

    my ($cmd,$status,@output,%hDevs);

    # add rescan - needed on some system after format
    # need to run the vxdctl script after the new VTOC is down
    # to get Veritas to see the disk again (properly) .. not completely
    # sure why
    $cmd  = "vxdctl enable";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SolarisDevMgr::_load_dfmap_dmp() Command ($cmd) returned a bad status ($status)");
        return;
    }

    my $vxdisk_hash = {};
    $self->sanlog(-msg=>"creating translation maps:");

    # Grab the vxdisk -o alldgs list output
    # DEVICE       TYPE            DISK         GROUP        STATUS
    # das_fcp_0_0  auto            -            -            error
    # das_fcp_0_1  auto            -            -            error
    # das_fcp_0_2  auto            -            -            error
    # das_fcp_0_3  auto            -            -            error
    # das_iscs0_0  auto            -            -            error
    # das_iscs0_1  auto            -            -            error
    # das_iscs0_2  auto            -            -            error
    # das_iscs0_3  auto            -            -            error
    # sda          auto:none       -            -            online invalid
    my @vxdisk_list_output;
    $cmd = "/usr/sbin/vxdisk -o alldgs list";
    ($status, @vxdisk_list_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SolarisDevMgr::_load_dfmap_dmp() Command ($cmd) returned a bad status ($status)");
        return;
    }

    # Store the data
    $self->{raw_vxdisk_list_output} = \@vxdisk_list_output;

    # Build a relationship hash keyed by device that looks like this:
    # vxdisk_hash => {device} => {
    #                               dmpnode    = "node_name",
    #                               primary    = "0|1",
    #                               init_state = "online"
    #                            }
    foreach my $line (sort @vxdisk_list_output) {
        $line =~ s/^\s+//g;
        next if ($line =~ /^DEVICE/);

        # Skip any devices flagged as boot, root, or reserved
        next if ($line =~ /boot|root|reserved/i);

        my ($dmpnode, $type, $disk, $diskgroup, @init_state) = split(/\s+/, $line);
        # Update $diskgroup to remove any ()
        $diskgroup =~ s/\(|\)//;

        # Skip any disks who have a disk group defined that isn't "-";
        if ($diskgroup ne "-") {
            $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since the it is used in diskgroup '$diskgroup'");
            next;
        }

        # Make sure type is "auto"
        unless ($type =~ /auto/) {
            $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since the type isn't 'auto'");
            next;
        }

        # Now that we have the dmpnode, run vxdisk list on that node.
        # and find out primary/secondary status
        my @vxdisk_list_node_output;
        $cmd = "/usr/sbin/vxdisk list " . $dmpnode;
        ($status, @vxdisk_list_node_output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: SolarisDevMgr::_load_dfmap_dmp() Command ($cmd) returned a bad status ($status)");
            return;
        }

        # Output should be
        # [root@auto-rhel5u6-02 ~]# vxdisk list das_fcp_0_0
        # Device:    das_fcp_0_0
        # devicetag: das_fcp_0_0
        # type:      auto
        # flags:     error private autoconfig
        # pubpaths:  block=/dev/vx/dmp/das_fcp_0_0 char=/dev/vx/rdmp/das_fcp_0_0
        # guid:      -
        # udid:      NETAPP%5FLUN%20C-Mode%5F2FiIn%3F-lCptj%5F2FiIn%3F-lCptm
        # site:      -
        # Multipathing information:
        # numpaths:   4
        # sdj             state=enabled   type=primary
        # sdn             state=enabled   type=secondary
        # sdb             state=enabled   type=secondary
        # sdf             state=enabled   type=primary
        foreach my $line (@vxdisk_list_node_output) {
            chomp($line);
            if ($line =~ /^udid\:\s+(.+)/) {
                # Make sure this is a netapp disk
                my $udid = $1;
                unless ($udid =~ /netapp/i) {
                    $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since it isn't a NETAPP lun");
                    last;
                }
            } elsif ($line =~ /^flags\:\s+(.+)/i) {
                my $flags = $1;
                if ($flags =~ /coordinator/i) {
                    # Disk has coordinator flag, which means it is a fencing device, skip it
                    $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since it is a fenced device");
                    last;
                }
            } elsif ($line =~ /^(\S+)\s+state=(\S+)\s+type=(\S+)/) {
                my $device_tmp = $1;
                my $state = $2;
                my $type = $3;

                # Skip the device if it isn't enabled
                if ($state !~ /enabled/i) {
                    $self->sanlog(-msg => "Skipping device '$device_tmp' since the path isn't enabled");
                    next;
                }

                # remove s2 from the end so this will match the later parsed output from Sanlun
                $device_tmp =~ s/s\d+$//;

                # Now let's add it to $vxdisk_hash
                $vxdisk_hash->{$device_tmp}->{"dmpnode"}  = $dmpnode;
                $vxdisk_hash->{$device_tmp}->{"init_state"} = join(" ", @init_state);
                if ($type =~ /primary/i) {
                    $vxdisk_hash->{$device_tmp}->{"primary"} = 1;
                } else {
                    $vxdisk_hash->{$device_tmp}->{"primary"} = 0;
                }
            } elsif ($line =~ /^(\S+)\s+state=(\S+)$/) {
                my $device_tmp = $1;
                my $state = $2;

                # Skip the device if it isn't enabled
                if ($state !~ /enabled/i) {
                    $self->sanlog(-msg => "Skipping device '$device_tmp' since the path isn't enabled");
                    next;
                }

                # remove s2 from the end so this will match the later parsed output from Sanlun
                $device_tmp =~ s/s\d+$//;

                # Now let's add it to $vxdisk_hash.  Since we don't have a "type", we will set
                # it as a primary path
                $vxdisk_hash->{$device_tmp}->{"dmpnode"} = $dmpnode;
                $vxdisk_hash->{$device_tmp}->{"primary"} = 1;
                $vxdisk_hash->{$device_tmp}->{"init_state"} = join(" ", @init_state);
            } ##END if ($line =~ /^udid\:\s+(.+)/) {
        } ##END foreach my $line (@vxdisk_list_output) {
    } ##END foreach my $device (sort keys %$vxdisk_hash) {

    # make sure our $vxdisk_hash ended up with something in it.
    unless (scalar(keys %$vxdisk_hash)) {
        $self->_set_err("Error: SolarisDevMgr::_load_dfmap_dmp() The vxdisk relationship hash is empty. Unable to continue.");
        return;
    }

    # Collect sanlun so we can later sort out internal devices
    my @sanlun_output = $self->{sanlun}->lun_show();

    if ($self->{sanlun}->error) {
        $self->_set_err("Error: SolarisDevMgr::_load_dfmap_dmp() ".$self->{sanlun}->errmsg());
        return;
    }

    # Store the raw sanlun output for later
    $self->{raw_device_output} = \@sanlun_output;

    foreach my $line (@sanlun_output) {
        # remove a known string that sanlun is prepending
        # to the first lun it finds (hack).
        $line =~ s|Report bugs to <bug-findutils\@gnu.org>.||;
        $line =~ s/^\s+//g;

        # Example Output depending on architecture
        # controller:                lun-pathname                device filename  adapter  protocol          lun size         lun state
        # auto-fas3020-03:  /vol/nate_2_1_auto_sunv440_01/nate_auto-sunv440-01_solaris_1  /dev/rdsk/c2t500A09849667B1D1d3s2  emlxs0   FCP           15g (16106127360)    GOOD
        if ($line =~ /(\S+)\s+\/vol\/(\S+)\/(\S+)\s+\/dev\/rdsk\/(\S+)s2\s+\S+\s+(\S+)\s+/) {
            # old sanlun kit output looks like below,
            # controller:                 lun-pathname                                 device filename                     adapter   protocol          lun size         lun state
            # iscsi_vs:  /vol/iscsi_vol_1/iscsi_auto-sunx2200-01_0  /dev/rdsk/c0t600A0980324668762D3F41734170386Fd0s2  scsi_vhci0  iSCSI          5g (5368709120)     GOOD
            #
            # new sanlun kit output looks like below,
            # controller(7mode)/                                             device                                            host                  lun
            # vserver(Cmode)       lun-pathname                              filename                                          adapter    protocol   size    mode
            # ---------------------------------------------------------------------------------------------------------------------------------------------------
            # iscsi_vs             /vol/iscsi_vol_1/iscsi_auto-sunx2200-01_0 /dev/rdsk/c0t600A0980324668762D3F41734170386Fd0s2 scsi_vhci0 iSCSI      5g      C
            my $filer_tmp    = $1;
            my $vol_tmp      = $2;
            my $lun_tmp      = $3;
            my $device_tmp   = $4;
            my $protocol_tmp = $5;

            # Remove colon in the end of filer/vserver name if presents
            $filer_tmp =~ s/:$//;

            # Now we want to look in the $vxdisk_hash for the device.  If it is reserved then we will
            # skip it.  If it isn't reserved, we will grab the dmpnode and add it to the hash with the
            # rest of the information.
            if (exists($vxdisk_hash->{$device_tmp})) {
                # Store all of the collected information
                $hDevs{$device_tmp} = {
                                        filer      => $filer_tmp,
                                        vol        => $vol_tmp,
                                        lun        => $lun_tmp,
                                        protocol   => $protocol_tmp,
                                        dmpnode    => $vxdisk_hash->{$device_tmp}->{'dmpnode'},
                                        init_state => $vxdisk_hash->{$device_tmp}->{'init_state'},
                                      };
            }
        }
    }

    # Now we want to walk through the $hDevs hash and eliminate any duplicate paths
    # for dmpnodes.  We likely have multiple primary and secondary paths for each dmpnode
    # and we only want to represent one primary path for each dmpnode in the final hash.
    my $temp_dmpnode_hash = {};
    my $hDevs_primary = {};
    foreach my $device (sort keys %hDevs) {
        # Check the primary/secondary status for the device in the
        # previously created vxdisk_hash.
        if (exists($vxdisk_hash->{$device})) {
            if (defined($vxdisk_hash->{$device}->{'primary'})
                and $vxdisk_hash->{$device}->{'primary'} == 1) {
                # Looks like this is a primary path so see if
                # the dmpnode has already been represented.
                my $dmpnode = $hDevs{$device}{'dmpnode'};
                unless (exists($temp_dmpnode_hash->{$dmpnode})) {
                    # This dmpnode has not been represented yet.
                    $temp_dmpnode_hash->{$dmpnode}->{$device} = $hDevs{$device};
                }
            } ##END if (defined($vxdisk_hash->{$device}->{'primary'}
        } ##END if (exists($vxdisk_hash->{$device})) {
    } ##END foreach my $device (sort keys %hDevs) {

    # Now take all of the primary paths and store them in a new hDevs_primary hash
    foreach my $dmpnode (sort keys %$temp_dmpnode_hash) {
        foreach my $device (sort keys %{$temp_dmpnode_hash->{$dmpnode}}) {
            $hDevs_primary->{$device} = $temp_dmpnode_hash->{$dmpnode}->{$device};
        }
    }

    # Make sure we found veritas luns.  If we didn't, then stop now.
    unless (scalar(keys %$hDevs_primary)) {
        $self->_set_err("Error: SolarisDevMgr::_load_dfmap_dmp() Failed to find any valid devices under Veritas control");
        return;
    }

    # save map for next call of function
    $self->{dfmap} = $hDevs_primary;
}

1;
