# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Devices on Virtual Linux
## @author hrivera@netapp.com
## @status private
## @pod ./Dev.pod


# Package Name
##############
package VLinuxDev;

use strict;

# Package Requirements
######################
use File::Basename;
use San::SanClient::ClientUtil;
use Error;
use San::SanLog;

use vars qw( @ISA );

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: load
# Objective:       sets parameters based on devices current mount state
# Details:         See POD documentation at the end of this file
################################################################################
sub load {
    my ($self) = @_;
    my ($dev_name) = $self->get_name();

    my ($partition) = $dev_name . "1";

    my ($cmd) = "/bin/mount | grep '$partition on'";
    my ($status, @output) = &ClientUtil::execute($self->{conn} , $cmd);
    if (scalar(@output) == 0) {
        # then this device is not mounted, probably not partitioned either.
        $self->{is_mounted} = $ClientUtil::FALSE;
        $self->{mount} = "";
        $self->{partition} = $ClientUtil::FALSE;
        $self->{pname} = "";
        $self->{device_is_free} = 1;
    } elsif (scalar(@output) > 1) {
        $self->_set_err("Error: VLinuxDev::load() $partition is mounted on " .
                        "on more than one mountpoint");
        return $ClientUtil::FALSE;
    } else {
        # the device is mounted, pull out the mount point
        $self->{is_mounted} = $ClientUtil::TRUE;
        ($self->{mount}, $self->{fs_type}) = ($output[0] =~ /$partition on (\S+) type (\S+)/);
        $self->{partition} = $ClientUtil::TRUE;
        $self->{pname} = $partition;
        $self->{device_is_free} = 0;
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    partition
# Objective: create a filesystem on a volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub partition {
    my ($self) = @_;

    my $dev_name = $self->get_name();
    my $cmd;

    $self->_clear_err();

    # check to see if this is already partitioned.
    if ($self->{partition}) {
        $self->_set_err("Error: VLinuxDev::partition() ".
                        "'$dev_name' is already partitioned.");
        return $ClientUtil::FALSE;
    }

    # fdisk the device with an alignment of 128 if the device is a VMFS/VHD lun
    if ($self->{vendor} =~ /VMware|Msft/i) {
        $cmd = "echo 'd\n1\nn\np\n1\n\n\nt\n83\nx\nb\n1\n128\nw\n' | /sbin/fdisk $dev_name";
    } else {
        $cmd = "echo 'd\n1\nn\np\n1\n\n\nt\n83\nw\n' | /sbin/fdisk $dev_name";
    }

    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        # this might be a false positive, check to see if the partition
        # exists
        $cmd = "ls ${dev_name}1";
        ($status) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Could not fdisk '$dev_name'");
            return $ClientUtil::FALSE;
        }
    }

    # update the partition values
    $self->{partition} = $ClientUtil::TRUE;
    $self->{pname} = $dev_name . "1";

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    create_fs
# Objective: create a filesystem on a volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_fs {
    my ($self, %args) = @_;

    my ($type);
    my $dev_name = $self->get_name();
    my ($cmd, $status, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?type/) {
            $type = $args{$_};
            next;
        }
    }

    # get client object
    my $client_obj = $self->{parent}->{parent};

    # check for existance of -type argument
    if (not defined $type or $type =~ /^\s*$/) {
        if ($client_obj->{major_revision} =~ /rhel6|rhel7|ol6|fedora/i) {
            # From RHEL6 and OL6 onwards, the preferred file system is ext4.
            # So, we are creating ext4 file system by default.
            $type = "ext4";
        } else {
            $type = "ext3";
        }
    }

    # Remove leading and trailing white spaces
    $type =~ s/^\s*//;
    $type =~ s/\s*$//;

    # make sure the device is partitioned
    if ($self->is_partitioned() == $ClientUtil::FALSE) {
        $self->_set_err("Error: VLinuxDev::create_fs() $dev_name is not ".
                        "yet partitioned");
        return $ClientUtil::FALSE;
    }

    # make the filesystem on the partitioned device
    my $partition_name = $self->get_partition();

    $self->sanlog(-msg => "Creating '$type' filesystem on '$partition_name'");

    if ($type =~ /^ext2$|^ext3$|^ext4$/i) {
        # Create a filesystem
        $cmd = " yes | mke2fs ";
        $cmd .= "-t $type " if ($client_obj->{major_revision} =~ /rhel6|rhel7|ol6|sles11|fedora/i);
        $cmd .= "-j " if ($type =~ /ext3/i);
        $cmd .= $partition_name;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Unable to create '$type' filesystem on '$partition_name'", @output);
            return $ClientUtil::FALSE;
        }
    } elsif ($type =~ /^xfs$/i) {
        $cmd = "yes | mkfs.xfs " . $partition_name;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Unable to create $type filesystem on '$partition_name'. " .
                            "Make sure that xfsprogs package is installed", @output);
            return $ClientUtil::FALSE;
        }
    } elsif ($type =~ /^btrfs$/i) {
        $cmd = "yes | mkfs.btrfs " . $partition_name;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Unable to create $type filesystem on '$partition_name'. " .
                            "Make sure that btrfsprogs package is installed", @output);
            return $ClientUtil::FALSE;
        }
    } elsif ($type =~ /^reiserfs/) {
        $cmd = "mkfs.reiserfs " . $partition_name;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to create '$type' filesystem on '$partition_name'", @output);
            return $self;
        }
    } else {
        $self->_set_err("Invalid Filesystem type '$type'");
        return $self;
    }

    $self->{fs_type} = $type;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: mount_fs
# Objective:       mount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_fs {
    my ($self) = @_;

    my ($cmd, $status, @output);
    my $dev_name = $self->get_name();
    my $partition = $self->get_partition();

    # make sure this is not already mounted
    if ($self->is_mounted() == $ClientUtil::TRUE) {
        $self->_set_err("Error: VLinuxDev::mount_fs() $dev_name is ".
                        "already mounted.");
        return $ClientUtil::FALSE;
    }

    $partition = basename($partition);
    my $mount_point = "/mnt/qa/nate_" . $partition;

    $self->sanlog(-msg => "Mounting '/dev/$partition' on mount point '$mount_point'");

    # create the mount point
    $cmd = "mkdir -p $mount_point";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error:VLinuxDev::mount_fs() Could not create mount ".
                        "directory '$mount_point'");
        return $ClientUtil::FALSE;
    }

    # mount the partition
    $cmd = "/bin/mount /dev/$partition $mount_point";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error:VLinuxDev::mount_fs() Could not mount '$partition'".
                        "to '$mount_point'");
        return $ClientUtil::FALSE;
    }

    # update the current mount values
    $self->{is_mounted} = 1;
    $self->{mount} = $mount_point;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, %args) = @_;

    my $dev_name = $self->get_name();

    # first make sure that this is device mounted
    if ($self->is_mounted() == $ClientUtil::FALSE) {
        $self->_set_err("Error: VLinuxDev::unmount_fs() $dev_name is ".
                        "not mounted");
        return $ClientUtil::FALSE;
    }

    my $mount_point = $self->get_mount();
    my $partition = $self->get_partition();

    # make sure this isnt the root mount
    if ($mount_point !~ /^\/mnt\/qa\//) {
        $self->_set_err("Error:VLinuxDev::unmount_fs() the mount point '$mount_point'".
                        " is not recognized by the automation");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Unmounting '$partition' from '$mount_point'");

    # umount the partition
    my $cmd = "/bin/umount $mount_point";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error:VLinuxDev::unmount_fs() Could not unmount '$partition'".
                        "from '$mount_point'");
        return $ClientUtil::FALSE;
    }

    # remove the mount point
    $cmd = "rm -rf $mount_point";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error:VLinuxDev::unmount_fs() Could not remove ".
                        "'$mount_point'");
        return $ClientUtil::FALSE;
    }

    $self->{is_mounted} = $ClientUtil::FALSE;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if the filesystem is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    if (defined($self->{is_mounted})) {
        return $self->{is_mounted};
    }

    my $partition = $self->get_partition();

    # check if this partition is mounted
    my $cmd = "/bin/mount | grep \'$partition on\'";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (scalar(@output) > 0) {
        # this device is mounted
        $self->{is_mounted} = $ClientUtil::TRUE;
        return $ClientUtil::TRUE;
    } else {
        $self->{is_mounted} = $ClientUtil::FALSE;
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: is_partitioned
# Objective:       returns ClientUtil::TRUE if partitioned
# Details:         See POD documentation at the end of this file
################################################################################
sub is_partitioned {
    my ($self, %args) = @_;

    return $self->{partition};
}

################################################################################
# Subroutine Name: destroy_partition
# Objective:       destroy a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_partition {
    my ($self) = @_;

    my $dev_name = $self->get_name();
    my $cmd = "echo 'd\n1\n' | /sbin/fdisk $dev_name";

    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: VLinuxDev::destroy_partition() ",
                        "Could not fdisk '$dev_name'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_uuid
# Objective:       Returns the UUID of the configured device.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self) = @_;

    return "";
}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       returns the raw device name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    # get the full device path. Needed for Windows compatibility.
    return ($self->get_name());
}

################################################################################
# Subroutine Name: get_partition
# Objective:       Returns the name of the configured partition.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition {
    my ($self) = @_;

    return $self->{pname};
}

################################################################################
# Subroutine Name: get_partition_size
# Objective:       Returns the size of the specified partition in bytes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition_size {
    my ($self, $partition_num) = @_;

    my ($cmd, $status, @output, $name);

    $self->_clear_err();

    # check required partition number
    if (!defined($partition_num) or $partition_num !~ /\d+/) {
        $self->_set_err("partition number not passed to this method or format is incorrect");
        return $ClientUtil::FALSE;
    }

    # generate path to device via raw device.
    $name = "/dev/" . $self->get_name();

    # Get device information
    $cmd = "/sbin/fdisk -l $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to get device information for device ($name)");
        return $ClientUtil::FALSE;
    }

    # add partition
    $name .= $partition_num;

    # find partition in fdisk output
    my $fdisk_line = (grep(/^$name\s+/i, @output))[0];
    $fdisk_line =~ s/\*//;

    # check to see if you got a valid partion line
    if (!defined($fdisk_line) or $fdisk_line eq "") {
        $self->_set_err("Unable to get partion information for device ($name)");
        return $ClientUtil::FALSE;
    }

    # fdisk line:
    #    Device Boot      Start         End      Blocks   Id  System
    # /dev/sda1               1        1011     1048376+  83  Linux
    my ($partition, $start, $end, $size, $id, $system) = split(/\s+/, $fdisk_line);
    $size =~ s/\+//;
    my $size_bytes = $size * 1024;

    return $size_bytes;

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    return $self->{lun_size_in_bytes};
}

################################################################################
# Subroutine Name: get_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{name};
}

################################################################################
# Subroutine Name: get_raw_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_name {
    my ($self) = @_;

    return $self->get_name();
}

################################################################################
# Subroutine Name: get_label
# Objective:       returns partition label of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_label {
    my ($self) = @_;

    return "";
}

################################################################################
# Subroutine Name: get_mount
# Objective:       returns mount point of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    return $self->{mount};
}

################################################################################
# Subroutine Name: get_fs
# Objective:       returns fs of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    return $self->{fs_type};
}

################################################################################
# Subroutine Name: get_lun
# Objective:       Return lun name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self) = @_;

    return $self->{lun_path};
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new dev object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($filer_name, $lun_path, $raw_device_name, $vendor, $protocol,
        $lun_size_in_bytes, $lun_state, $conn, $parent);

    my $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?filer_name/) {
            if (not length($filer_name = $args{$_})) {
                $self->_set_err("Error: VLinuxDev::_new() Invalid ".
                                "value for argument '-filer_name'");
                return $self;
            }
            next;
        } elsif (/-?lun_path/) {
            if (not length($lun_path = $args{$_})) {
                $self->_set_err("Error: VLinuxDev::_new() Invalid ".
                                "value for argument '-lun_path'");
                return $self;
            }
            next;
        } elsif (/-?raw_device_name/) {
            if (not length($raw_device_name = $args{$_})) {
                $self->_set_err("Error: VLinuxDev::_new() Invalid ".
                                "value for argument '-raw_device_name'");
                return $self;
            }
            next;
        } elsif (/-?vendor/) {
            if (not length($vendor = $args{$_})) {
                $self->_set_err("Error: VLinuxDev::_new() Invalid ".
                                "value for argument '-vendor'");
                return $self;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: VLinuxDev::_new() Invalid ".
                                "value for argument '-protocol'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: VLinuxDev::_new() Invalid ".
                                "value for argument '-parent'");
                return $self;
            }
            next;
       } elsif (/-?lun_size_in_bytes/) {
            if (not length($lun_size_in_bytes = $args{$_})) {
                $self->_set_err("Error: VLinuxDev::_new() Invalid ".
                                "value for argument '-lun_size_in_bytes'");
                return $self;
            }
            next;
       } elsif (/-?lun_state/) {
            if (not length($lun_state = $args{$_})) {
                $self->_set_err("Error: VLinuxDev::_new() Invalid ".
                                "value for argument '-lun_state'");
                return $self;
            }
            next;
       } elsif (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: VLinuxDev::_new() Invalid ".
                                "value for argument '-conn'");
                return $self;
            }
            next;
        }
    }

    # store the variables in the object
    $self->{filer_name} = $filer_name;
    $self->{lun_path} = $lun_path;
    $self->{name} = $raw_device_name;
    $self->{vendor} = $vendor;
    $self->{protocol} = $protocol;
    $self->{lun_size_in_bytes} = $lun_size_in_bytes;
    $self->{lun_state} = $lun_state;
    $self->{conn} = $conn;
    $self->{device_is_free} = 1;
    $self->{partition} = $ClientUtil::FALSE;
    $self->{pname} = "";
    $self->{is_mounted} = $ClientUtil::FALSE;
    $self->{mount} = "";
    $self->{fs_type} = "";
    $self->{parent} = $parent;

    # load this device's current partition/mount status on the system
    $self->load();

    return $self;
}

1;
