# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to DevMgr for a Virtual Linux Host
## @author hrivera@netapp.com
## @status private
## @pod ./DevMgr.pod

# Package Name
##############
package San::SanClient::DevMgr::VLinuxDevMgr;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::DevMgr;
use San::SanClient::DevMgr::VLinuxDev;
use Error;
use San::SanLog;

use vars qw( @ISA );

@ISA = qw(Error SanLog DevMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_device
# Objective:       return a device object based on the device name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device {
    my ($self, $name) = @_;
    
    $self->_clear_err();

    my @devices = $self->get_devices();
    
    foreach my $device (@devices) {
        if ($device->get_name() =~ /$name$/) {
            return $device;
        } 
    }
    
    $self->_set_err("Error: VLinuxDevMgr::get_device() No such device '$name'");
    return {};
}

################################################################################
# Subroutine Name: get_devices
# Objective:       return all currently configured device objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;

    $self->_clear_err();

    if (scalar(@{$self->{devices}}) == 0) {
        return (wantarray()) ? () : [];
    }
    
    my @devices = @{$self->{devices}};

    return (wantarray()) ? @devices : \@devices;
} 

################################################################################
# Subroutine Name: create_device
# Objective:       create a device object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_device {
    my ($self, $lun) = @_;
    
    $self->_clear_err();

    # if $lun contains the header of sanlun output, return
    if ($lun =~ /lun-pathname/i or $lun =~ /lun state/i) {
        return $ClientUtil::TRUE;
    }

    # parse the lun show output
    my ($filer_name, 
        $lun_path, 
        $raw_device_name, 
        $vendor, 
        $protocol, 
        $lun_size_in_gbytes,
        $lun_size_in_bytes,
        $lun_state) = ($lun =~ /^\s*(\S+?):\s*(\S+?)\s*(\S+?)\s*(\S+?)\s*(\S+?)\s*(\S+?)\s*\((.+?)\)\s*(\S+)/);

    # create the dev object
    my $dev_obj = VLinuxDev->_new(-filer_name => $filer_name,
                                  -lun_path => $lun_path,
                                  -raw_device_name => $raw_device_name,
                                  -vendor => $vendor,
                                  -protocol => $protocol,
                                  -lun_size_in_bytes => $lun_size_in_bytes,
                                  -lun_state => $lun_state,
                                  -parent => $self,
                                  -conn => $self->{conn});

    if ($dev_obj->error()) {
        $self->_set_err($dev_obj->errmsg());
        return $ClientUtil::FALSE;
    }

    # push it on the device list
    push(@{$self->{devices}}, $dev_obj);
    $self->{device_count}++;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_dev_id
# Objective:       returns unique device name(with full path) of passed device's
#                  short name to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_id {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: get_dev_name
# Objective:       returns device short name of passed device's unique name(long name)
#                  to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_name {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: cleanup
# Objective:       remove device objects and remove mount point entries 
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup {
    my ($self, @devices) = @_;

    $self->_clear_err();

    # unmount each device
    foreach my $device (@devices) {
        $device->unmount_fs();
        if ($device->error()) {
            $self->_set_err($device->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
} 

################################################################################
# Subroutine Name: get_free_devices
# Objective:       retrieve list of free host volumes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices { 
    my ($self) = @_;

    $self->_clear_err();

    # obtain all the devices
    my @devices = $self->get_devices();
    
    # filter out the devices that are not free
    @devices = grep { 
        my $dev_obj = $_;
        if ($dev_obj->{device_is_free} == 1) {
            1;
        } else {
            0;
        }
    } @devices;

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new devices object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($parent, $load_devices);

    my $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: VLinuxDevMgr::_new() Invalid ".
                                "value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?load_devices/) {
            if (not length($load_devices = $args{$_})) {
                $self->_set_err("Error: VLinuxDevMgr::_new() Invalid value for argument '-load_devices'");
                return ($self);
            }
            next;
        }
    }     

    # set default for load_devices (true)
    if (not defined($load_devices)) {
        $load_devices = 1;
    }
   
    # check for existance of required -parent argument
    if (grep ( /^-?parent$/, (keys %args) )) {
        unless (ref($args{-parent}) =~ /SanClient/) {
            $self->_set_err("Error: VLinuxDevMgr::_new()",
                            "Invalid value for argument '-parent'");
            return $self; 
        }
    } else {
        $self->_set_err("Error: VLinuxDevMgr::_new()",
                        "Missing required argument '-parent'");
        return $self; 
    }

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{devices} = [];
    $self->{device_count} = 0;

    # call lun show from the sanlun object
    my (@sanlun_output) = $self->{sanlun}->lun_show();
    if ($self->{sanlun}->error()) {
        $self->_set_err($self->{sanlun}->errmsg);
        return $self;
    }

    # cache the sanlun output
    $self->{sanlun_output} = @sanlun_output;

    if ($load_devices == 1) {
        # create a device object for each lun
        foreach my $lun (@sanlun_output) {
            $self->create_device($lun);
            if ($self->error()) {
                return $self;
            }
        }
    }

    return $self;
}


1;

