# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Devices on a virtual Solaris machine
## @author hrivera@netapp.com
## @status private
## @pod ./Dev.pod


# Package Name
##############
package VSolarisDev;

use strict;

# Package Requirements
######################
use File::Basename;
use San::SanClient::ClientUtil;
use Error;
use San::SanLog;
use San::SanUtil;

use vars qw( @ISA );

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: load
# Objective:       sets parameters based on devices current mount state
# Details:         See POD documentation at the end of this file
################################################################################
sub load {
    my ($self) = @_;
    my ($dev_name) = $self->get_name();
    # use slice 6
    $dev_name =~ s/s2/s6/;
    my ($dev_name_only) = basename($dev_name);


    $self->sanlog(-msg => "Loading '$dev_name' from its current state");

    my ($cmd) = "/usr/sbin/mount | grep 'on /dev/dsk/$dev_name_only ' | awk '{print \$1}'";
    my ($status, @output) = &ClientUtil::execute($self->{conn} , $cmd);
    if (scalar(@output) == 0) {
        # then this device is not mounted, probably not partitioned either.
        $self->{is_mounted} = $ClientUtil::FALSE;
        $self->{mount} = "";
        $self->{device_is_free} = 1;
    } elsif (scalar(@output) > 1) {
        $self->_set_err("Error: VSolarisDev::load() $dev_name is mounted on " .
                        "on more than one mountpoint");
        return $ClientUtil::FALSE;
    } else {
        # the device is mounted, pull out the mount point
        $self->{is_mounted} = $ClientUtil::TRUE;
        $self->{mount} = $output[0];
        $self->{partition} = $ClientUtil::TRUE;
        $self->{pname} = $dev_name;
        $self->{device_is_free} = 0;

        # obtain the fs type
        $cmd = "/usr/sbin/fstyp $dev_name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: VSolarisDev::load() Fstyp command failed:", @output);
            return $ClientUtil::FALSE;
        }
        $self->{fs_type} = $output[0];

    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    partition
# Objective: partition the device
# Details:   See POD documentation at the end of this file.
###############################################################################
sub partition {
    my ($self) = @_;

    $self->_clear_err();
    my ($cmd, @output, $status);

    my $partition_name = $self->get_partition();
    my $dev_name = $self->get_name();

    # fdisk the partition
    $cmd = "/usr/sbin/fdisk -B $partition_name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: VSolarisDev::partition() could not fdisk $partition_name");
        return $ClientUtil::FALSE;
    }

    # burt750020: partition the device so that it is properly alligned based on RDM or VMFS
    my $label_name = basename($dev_name);
    my $file_name = "/tmp/" . $label_name . ".format";

    my (@format_cmds) = ();
    if ($self->{vendor} =~ /NETAPP/i) {
        # RDM (burt750020)
        push(@format_cmds, "type 0");
        push(@format_cmds, "label");
        push(@format_cmds, "partition");
        push(@format_cmds, "modify");
        push(@format_cmds, "1");
        push(@format_cmds, "yes");
        push(@format_cmds, "6");
        push(@format_cmds, "0");
        push(@format_cmds, "0");
        push(@format_cmds, "0");
        push(@format_cmds, "0");
        push(@format_cmds, "0");
        push(@format_cmds, "0");
        push(@format_cmds, "yes");
        push(@format_cmds, "NATE");
        push(@format_cmds, "q");
        push(@format_cmds, "q");
    } else {
        # VMFS (burt750020)
        push(@format_cmds, "type 0");
        push(@format_cmds, "label");
        push(@format_cmds, "partition");
        push(@format_cmds, "modify");
        push(@format_cmds, "1");
        push(@format_cmds, "yes");
        push(@format_cmds, "6");
        push(@format_cmds, "0");
        push(@format_cmds, "0");
        push(@format_cmds, "0");
        push(@format_cmds, "0");
        push(@format_cmds, "0");
        push(@format_cmds, "0");
        push(@format_cmds, "yes");
        push(@format_cmds, "NATE");
        push(@format_cmds, "6");
        push(@format_cmds, "unassigned");
        push(@format_cmds, "wm");
        push(@format_cmds, "7");
        push(@format_cmds, '$');
        push(@format_cmds, "label");
        push(@format_cmds, "q");
        push(@format_cmds, "q");
    }

    &ClientUtil::create_remote_file($self->{conn}, $file_name, @format_cmds);
    $cmd = "/usr/sbin/format -f $file_name $partition_name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: VSolarisDev::partition() could not partition the device $dev_name ($partition_name)");
        return $ClientUtil::FALSE;
    }
    &ClientUtil::remove_remote_file($self->{conn}, $file_name);

    $self->{partition} = $ClientUtil::TRUE;

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    create_fs
# Objective: create a filesystem on a volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_fs {
    my ($self, %args) = @_;

    my ($type) = "ufs";
    my $dev_name = $self->get_name();

    # use slice 6 of the device to create the fs on
    $dev_name =~ s/s2/s6/;

    # parse arguments
    foreach (keys %args) {
        if (/-?type/) {
            if (not length($type = $args{$_})) {
                # default $type to ufs if it wasnt provided.
                $type = "ufs";
            }
            next;
        }
    }

    # check for invalid fs types
    if ($type !~ /^(ufs|vxfs)$/) {
        $self->_set_err("Error: VSolarisDev::create_fs() '$type' is not a ".
                        "supported fs type'");
        return $ClientUtil::FALSE;
    }

    # make the filesystem on the device
    $self->sanlog(-msg => "Creating '$type' filesystem on '$dev_name'");

    my $cmd = "yes y | /usr/sbin/newfs $dev_name";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: VSolarisDev::create_fs() Unable to create $type ".
                        "filesystem on volume '$dev_name' ",
                        @output);

        return $ClientUtil::FALSE;
    }

    $self->{fs_type} = $type;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: mount_fs
# Objective:       mount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_fs {
    my ($self) = @_;

    my $dev_name = $self->get_name();
    # use slice 6 of the device
    $dev_name =~ s/s2/s6/;

    my $dev_name_only = basename($dev_name);

    my $fs_type = $self->get_fs();
    my ($cmd, $status, @output);

    my $mount_point = "/mnt/qa/nate_" . $dev_name_only;

    $self->sanlog(-msg => "Mounting '$dev_name' at mount point '$mount_point'");

    # make sure this is not already mounted
    if ($self->is_mounted() == $ClientUtil::TRUE) {
        $self->_set_err("Error: VSolarisDev::mount_fs() $dev_name is ".
                        "already mounted.");
        return $ClientUtil::FALSE;
    }

    # create the mount point
    $cmd = "mkdir -p $mount_point; chmod 777 $mount_point";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error:VSolarisDev::mount_fs() Could not create mount ".
                        "Directory '$mount_point'");
        return $ClientUtil::FALSE;
    }

    # mount the partition
    $cmd = "/usr/sbin/mount /dev/dsk/$dev_name_only $mount_point";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        ### device needs to be fsck
        if ("@output" =~ /fsck/i) {
            $self->sanlog(-msg => "Device '$dev_name' needs to be fsck'd");
            ($status, @output) = &ClientUtil::execute($self->{conn}, "fsck -y $dev_name");
        }

        ### try to mount it again
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error:VSolarisDev::mount_fs() Could not mount '$dev_name'".
                            "to '$mount_point'");
            return $ClientUtil::FALSE;
        }
    }

    $self->{is_mounted} = 1;
    $self->{mount} = $mount_point;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, %args) = @_;

    my $dev_name = $self->get_name();

    # use slice 6 of the device
    $dev_name =~ s/s2/s6/;

    my ($status, @output);

    # make sure that this is mounted
    if ($self->is_mounted() == $ClientUtil::FALSE) {
        $self->_set_err("Error: VSolarisDev::unmount_fs() $dev_name is ".
                        "not mounted");
        return $ClientUtil::FALSE;
    }

    my $mount_point = $self->get_mount();

    # make sure this isnt the root mount
    if ($mount_point !~ /^\/mnt\/qa\//) {
        $self->_set_err("Error: VSolarisDev::unmount_fs() mount point ".
                        "'$mount_point' is not recognized by the automation");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Unmounting '$dev_name' from '$mount_point'");

    my $dev_name_only = basename($dev_name);

    # umount the device, try 60 times
    my $cmd = "/usr/sbin/umount /dev/dsk/$dev_name_only";

    my $loop = 60;
    foreach my $tries (1 .. $loop) {
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            if ("@output" =~ /busy/i and $tries < $loop) {
                if ($tries == 1) {
                    $self->sanlog(-msg => "Info: device is busy: retry count ($tries/$loop)");
                }
                &SanUtil::nate_sleep(1);
                next;
            }
            $self->_set_err("Error:VSolarisDev::unmount_fs() Could not unmount '$dev_name'".
                            "from '$mount_point'");
            return $ClientUtil::FALSE;
        }
        last;
    }

    # remove the mount point
    $cmd = "rm -rf $mount_point";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error:VSolarisDev::unmount_fs() Could not remove ".
                        "'$mount_point'");
        return $ClientUtil::FALSE;
    }

    $self->{is_mounted} = $ClientUtil::FALSE;
    $self->{mount} = "";

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if filesystem is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    my $dev_name = $self->get_name();

    # use slice 6 of the device
    $dev_name =~ s/s2/s6/;
    $dev_name =~ s/rdsk/dsk/;

    # determine if this device is mounted
    my $cmd = "/usr/sbin/mount | grep $dev_name | awk '{print \$1}'";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (scalar(@output) > 0) {
        # this device is mounted
        # ignoring status because it can return a false status.
        $self->{is_mounted} = $ClientUtil::TRUE;
        return $ClientUtil::TRUE;
    } else {
        $self->{is_mounted} = $ClientUtil::FALSE;
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: is_partitioned
# Objective:       returns ClientUtil::TRUE if partitioned
# Details:         See POD documentation at the end of this file
################################################################################
sub is_partitioned {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: destroy_partition
# Objective:       destroy a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_partition {
    my ($self) = @_;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_uuid
# Objective:       Returns the UUID of the configured device.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self) = @_;

    return "";
}

################################################################################
# Subroutine Name: get_partition
# Objective:       Returns the name of the configured partition.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition {
    my ($self) = @_;

    return $self->{pname};
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    return $self->{lun_size_in_bytes};
}

################################################################################
# Subroutine Name: get_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{name};
}

################################################################################
# Subroutine Name: get_raw_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_name {
    my ($self) = @_;

    return $self->get_name();
}

################################################################################
# Subroutine Name: get_label
# Objective:       returns partition label of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_label {
    my ($self) = @_;

    return "";
}

################################################################################
# Subroutine Name: get_mount
# Objective:       returns mount point of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    return $self->{mount};
}

################################################################################
# Subroutine Name: get_fs
# Objective:       returns fs of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    return $self->{fs_type};
}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       returns the raw device name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    # get the full device path. Needed for Windows compatibility.
    my $raw_dev_name = $self->get_name();
    $raw_dev_name =~ s/s2/s6/;

    return $raw_dev_name;
}

################################################################################
# Subroutine Name: get_lun
# Objective:       Return lun name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self) = @_;

    return $self->{lun_path};
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new dev object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($filer_name, $lun_path, $raw_device_name, $vendor, $protocol,
        $lun_size_in_bytes, $lun_state, $conn, $device_number, $partition_name);
    my $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?filer_name/) {
            if (not length($filer_name = $args{$_})) {
                $self->_set_err("Error: VSolarisDev::_new() Invalid ".
                                "value for argument '-filer_name'");
                return $self;
            }
            next;
        } elsif (/-?lun_path/) {
            if (not length($lun_path = $args{$_})) {
                $self->_set_err("Error: VSolarisDev::_new() Invalid ".
                                "value for argument '-lun_path'");
                return $self;
            }
            next;
        } elsif (/-?raw_device_name/) {
            if (not length($raw_device_name = $args{$_})) {
                $self->_set_err("Error: VSolarisDev::_new() Invalid ".
                                "value for argument '-raw_device_name'");
                return $self;
            }
            next;
        } elsif (/-?vendor/) {
            if (not length($vendor = $args{$_})) {
                $self->_set_err("Error: VSolarisDev::_new() Invalid ".
                                "value for argument '-vendor'");
                return $self;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: VSolarisDev::_new() Invalid ".
                                "value for argument '-protocol'");
                return $self;
            }
            next;
       } elsif (/-?lun_size_in_bytes/) {
            if (not length($lun_size_in_bytes = $args{$_})) {
                $self->_set_err("Error: VSolarisDev::_new() Invalid ".
                                "value for argument '-lun_size_in_bytes'");
                return $self;
            }
            next;
       } elsif (/-?lun_state/) {
            if (not length($lun_state = $args{$_})) {
                $self->_set_err("Error: VSolarisDev::_new() Invalid ".
                                "value for argument '-lun_state'");
                return $self;
            }
            next;
       } elsif (/-?partition_name/) {
            if (not length($partition_name = $args{$_})) {
                $self->_set_err("Error: VSolarisDev::_new() Invalid ".
                                "value for argument '-partition_name'");
                return $self;
            }
            next;
       } elsif (/-?device_number/) {
            if (not length($device_number = $args{$_})) {
                $self->_set_err("Error: VSolarisDev::_new() Invalid ".
                                "value for argument '-device_number'");
                return $self;
            }
            next;
       } elsif (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: VSolarisDev::_new() Invalid ".
                                "value for argument '-conn'");
                return $self;
            }
            next;
        }
    }

    # store the variables in the object
    $self->{filer_name} = $filer_name;
    $self->{lun_path} = $lun_path;
    $self->{name} = $raw_device_name;
    $self->{vendor} = $vendor;
    $self->{protocol} = $protocol;
    $self->{lun_size_in_bytes} = $lun_size_in_bytes;
    $self->{lun_state} = $lun_state;
    $self->{conn} = $conn;
    $self->{device_is_free} = 1;
    $self->{partition} = $ClientUtil::FALSE;
    $self->{pname} = $partition_name;
    $self->{device_number} = $device_number;
    $self->{is_mounted} = $ClientUtil::FALSE;
    $self->{mount} = "";
    $self->{fs_type} = "";

    $self->load();

    return $self;
}

1;
