# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Devices on Win
## @author pressley@netapp.com
## @status public
## @pod ./Dev.pod

# Package Name
##############
package WinDev;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanLog;
use San::SanClient::ClientUtil;

use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# P U B L I C   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: load
# Objective:       sets parameters based on devices current mount state
# Details:         See POD documentation at the end of this file
################################################################################
sub load {
    my ($self) = @_;

    my $mount = $self->{diskObj}{mount};
    my $fs    = $self->{diskObj}{fsType};

    $self->_clear_err();

    $self->sanlog(-msg => "Populating $self->{name} data object");
    $self->sanlog(-msg => "Loading $self->{name}  Mount: $mount FS: $fs");

    $self->{mount} = $mount;
    $self->{fs}    = $fs;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_fs
# Objective:       create a filesystem on the device object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_fs {
    my ($self, %args) = @_;

    my $type;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
      if (/-?type$/) {
          $type = $args{$_};
          next;
      } 
    } 

    if (not defined($type) or $type eq "") {
        ### Default -type to value of ntfs
        $type = "ntfs";
    }                                                                                                                        
    #
    # Find the diskObject and execute format to create a fs on the disk.
    #
    my $diskObj = $self->{diskObj};
    $diskObj->format(-type => $type);
    if ($diskObj->error) {
        my $tempmsg = $diskObj->errmsg();
        $self->_set_err("Error: WinDev::create_fs() error from " .
                        "WinClient::Disk::format(): $tempmsg");
        return ($ClientUtil::FALSE); 
    }
    
    $self->{fs} = $type;

    return $ClientUtil::TRUE;
} 

################################################################################
# Subroutine Name: mount_fs
# Objective:       mount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_fs {
    my ($self, %args) = @_;

    $self->_clear_err();
    
    #
    # noop on Windows
    # On windows you create a filesystem on an assigned and thus moutned 
    # dev so it doesn't make any sense to mount a filesystem.
    #

    return $ClientUtil::TRUE;
} 

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, %args) = @_;

    $self->_clear_err();
    
    #
    # noop on Windows
    # On windows you create a filesystem on an assigned and thus moutned
    # dev so it doesn't make any sense to mount a filesystem.
    #

    return $ClientUtil::TRUE;
} 

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if filesystem is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{mount} and length $self->{mount}) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
} 

################################################################################
# Subroutine Name: is_partitioned
# Objective:       returns ClientUtil::TRUE if partitioned
# Details:         See POD documentation at the end of this file
################################################################################
sub is_partitioned {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined $self->{diskObj}{isParted} and
        length $self->{diskObj}{isParted}  and
        $self->{diskObj}{isParted}) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
} 

################################################################################
# Subroutine Name: partition
# Objective:       create a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub partition {
    my ($self, %args) = @_;

    my $diskObj = $self->{diskObj};

    $self->_clear_err();

    #
    # Just get out of there if the device is already partitioned because
    # trying to partition the device over again will cause and error
    #
    if ($self->is_partitioned()) {
        return $ClientUtil::TRUE;
    }

    $diskObj->partition();
    if ($diskObj->error) {
        my $tempmsg = $diskObj->errmsg();
        $self->_set_err("Error: WinDev::partition() error from " .
                        "WinClient::Disk::partition(): $tempmsg");
        return ($ClientUtil::FALSE); 
    }

    return $ClientUtil::TRUE;
} 

################################################################################
# Subroutine Name: destroy_partition
# Objective:       destroy a partition on the raw device
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_partition {
    my ($self, %args) = @_;

    $self->_clear_err();
    
    # TBD

    return $ClientUtil::TRUE;
} 

################################################################################
# Subroutine Name: get_uuid
# Objective:       Returns the UUID of the configured device.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_uuid {
    my ($self) = @_;
    
    $self->_clear_err();

    # This function is only valid on Linux
    return undef;

} 

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{size})) {
        return $self->{size};
    } else {
        $self->_set_err("Error: WinDev::get_size() Dev object is missing size");
        return undef;
    }
} 

################################################################################
# Subroutine Name: get_name
# Objective:       returns name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined($self->{name}) and length($self->{name})) {
        return $self->{name};
    } else {
        $self->_set_err("Error: WinDev::get_name() Dev object is missing name");
        return undef;
    }
} 

################################################################################
# Subroutine Name: get_raw_device
# Objective:       returns the raw device name of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    my $disk_num = $self->get_name();
    return "" if ($self->error);

    my $physical_drive_path = "\\\\.\\PHYSICALDRIVE" . $disk_num;
    return ($physical_drive_path);

}

################################################################################
# Subroutine Name: get_mount
# Objective:       returns mount point of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined($self->{mount}) and length($self->{mount})) {
        return $self->{mount};
    } else {
        $self->_set_err("Error: WinDev::get_mount() Dev object is missing mount");
        return undef;
    }
} 

################################################################################
# Subroutine Name: get_fs
# Objective:       returns fs of device object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined($self->{fs}) and length($self->{fs})) {
        return $self->{fs};
    } else {
        $self->_set_err("Error: WinDev::get_fs() Dev object missing fs");
        return undef;
    }
} 

################################################################################
# Subroutine Name: get_lun
# Objective:       Return lun name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lun {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{lun}) and length($self->{lun})) {
        return $self->{lun};
    } else {
        $self->_set_err("Error: WinDev::get_lun() Dev object is missing lun name");
        return undef;
    }
}

################################################################################
# Subroutine Name: get_partition
# Objective:       Returns the name of the configured partition.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition {
    my ($self) = @_;
    
    $self->_clear_err();

    # TBD

    return undef;
} 

################################################################################
# Subroutine Name: get_partition_size
# Objective:       Returns the name size of a particular partition in bytes
# Details:         See POD documentation at the end of this file
################################################################################
sub get_partition_size {
    my ($self, $partition_num) = @_;

    $self->_clear_err();

    # in automation, we only create a single partition,
    # so simply return the size in bytes.
    my $size = $self->get_size();

    # lunSize is in MB, convert to bytes
    $size = &ClientUtil::convert_to_bytes($size . "m");
    
    return ($size);

}

################################################################################
# Subroutine Name: get_label
# Objective:       Returns the name of the configured partition's label.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_label {
    my ($self) = @_;
    
    $self->_clear_err();

    if (defined($self->{diskObj}{fsLabel}) and length($self->{lun}{fsLabel})) {
        return $self->{diskObj}{fsLabel};
    } else {
        $self->_set_err("Error: WinDev::get_lun() Dev object is missing label");
        return undef;
    }
}

################################################################################
# Subroutine Name: refresh_geometry
# Objective:       refresh the geomery on device. Usefull to call after a 
#                  resize on controller. NO-OP on this platform
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh_geometry {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new dev object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my $self;
    my $conn;
    my $log;
    my $name;
    my $parent;

    $self = {};

    $self->{debug_level} = 10; # TBD 10 or 1?

    bless $self, $class;    

    #
    # parse arguments
    #
    foreach (keys %args) {
        if (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: WinDev::new()",
                                "Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: WinDev::new()",
                                "Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/){
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: WinDev::new()",
                                "Missing value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?loghdl$/) {
            $log = $args{$_} if (defined $args{$_} and length($args{$_}));            
            next;
        }
    }
    
    #
    # check for existance of required -name argument
    #
    unless (defined $name) {
        $self->_set_err("Error: WinDev::new()",
                        "Missing required argument '-name'");
        return $self; 
    }

    #
    # check for existance of required -parent argument
    #
    unless (defined $parent) {
        $self->_set_err("Error: WinDev::new()",
                        "Missing required argument '-parent'");
        return $self; 
    }

    #
    # check for existance of required -conn argument
    #
    if (grep ( /^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: WinDev::new()",
                            "Invalid value for argument '-conn'");
            return $self; 
        }    
    } else {
        $self->_set_err("Error: WinDev::new()",
                        "Missing required argument '-conn'");
        return $self; 
    }

    $self->{log} = $log if (defined $log);

    #
    # Set the debug level to that of the parent if you can
    #
    $self->{debug_level} = $parent->{debug_level}
        if (defined $parent->{debug_level});
    
    $self->sanlog(-msg_level => 3,
                  -msg       => "Instantiating a WinDev Object ($name)");

    #
    # Find the disk object
    #
    $self->{diskObj} = $parent->{WinClient}->{disks}{$name};
    if (not (ref($self->{diskObj}) eq "WinClient::Disk")) {
        $self->_set_err("Error: WinDev::_new() Invalid data type for disk object");
         return $self;
    } 
    
    #
    # Show the user the info for this disk.
    #
    my $debugMsg = "WinDev::_new() diskObj info:\n";
    foreach my $key (keys %{$self->{diskObj}}) {
        $debugMsg .= "Disk Field $key is " . $self->{diskObj}{$key} . "\n";
    }
    $self->sanlog(-msg => $debugMsg, -msg_level => 7);

    #
    # initilize the object members
    #
    $self->{conn}   = $conn;
    $self->{parent} = $parent;
    $self->{name}   = $name;

    #
    # get filer lun associated with physical dev
    #
    $self->{lun}   = $self->{diskObj}{filer} . ":" . $self->{diskObj}{lunPath};
    $self->{fs}    = "";
    $self->{mount} = "";
    $self->{size}  = $self->{diskObj}{lunSize};
    $self->{size}  =~ s/\s+//;   # remove white space
    $self->{serial} = $self->{diskObj}{serial};

    return $self;
} 

1;

