# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to DevMgr for Windows host
## @author pressley@netapp.com
## @status public
## @pod ./DevMgr.pod

# Package Name
##############
package San::SanClient::DevMgr::WinDevMgr;

use strict;

# Package Requirements
######################
use San::SanLog;
use San::SanClient::ClientUtil; 
use San::SanClient::DevMgr;
use San::SanClient::DevMgr::WinDev;

use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog DevMgr);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: get_device
# Objective:       return a device object based on the device name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device {
    my ($self, $name) = @_;
    
    $self->_clear_err();

    unless(defined $name and length $name) {
        $self->_set_err("Error: WinDevMgr::get_device() name is a required argument");
        return;
    }

    # if we get in a physicaldrive representation, strip it off
    if ($name =~ /PHYSICALDRIVE/i) {
        $name =~ s/\\\\\.\\PHYSICALDRIVE//i;
    }

    #
    # Check out each object we know about and if the device name matches
    # return that object
    #
    foreach my $obj ($self->get_devices) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }
    
    $self->_set_err("Error: WinDevMgr::get_device() Unable to find device '$name'");

    return;
}

################################################################################
# Subroutine Name: get_devices
# Objective:       return all currently configured device objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;

    my @devices;
    
    $self->_clear_err();

    if (@{$self->{devices}}) {
        @devices = @{$self->{devices}};
    } else {
        return ();        
    }

    return (wantarray()) ? @devices : \@devices;  
} 

################################################################################
# Subroutine Name: create_device
# Objective:       create a device object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_device {
    my ($self, %args) = @_;

    my $devObj;

    $self->_clear_err();

    $devObj = WinDev->_new(-parent=>$self,
                           -conn=>$self->{conn},
                           -loghdl=>$self->{log},
                           %args);

    if ($devObj->error) {
        my @tempmsg = $devObj->errmsg();
        $self->_set_err("WinDevMgr::create_device: "             .
                        "Could not create a new device object: " .
                        "@tempmsg");
        return $self;
    }
    
       
    $self->sanlog(-msg_level => 7,
                  -msg       => "Pushing device on devices stack");
    push @{$self->{devices}}, $devObj;

    return ($devObj);  
}

################################################################################
# Subroutine Name: get_dev_id
# Objective:       returns unique device name(with full path) of passed device's
#                  short name to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_id {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: get_dev_name
# Objective:       returns device short name of passed device's unique name(long name)
#                  to this function
# Details:         See POD documentation at the end of this file
################################################################################
sub get_dev_name {
    my ($self) = @_;

    # It is supported only in Linux Host
    $self->_clear_err();

    $self->_set_err("This method is not supported on this platform");

    return "";

}

################################################################################
# Subroutine Name: cleanup
# Objective:       remove device objects and remove mount point entries 
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup {
    my ($self) = @_;

    my $devObj;
    my $fail_cnt = 0;

    $self->_clear_err();

    #
    # remove mount point entry for each device passed in
    #
    foreach $devObj ($self->get_devices) {
        $self->sanlog(-msg =>"Cleaning up mount points for device: " .
                             $devObj->{name});
    
        $devObj->unmount_fs() if $devObj->is_mounted;
        if ($devObj->error) {
            $self->_set_err($devObj->errmsg);
            $fail_cnt++;
        }
    
        #
        # remove physical devices from system
        #
        $self->_delete_devices($devObj->{name});
        if ($self->error) {
            $fail_cnt++;
        }
    } 
    
    $self->{devices} = [];

    return ($fail_cnt == 0) ? $ClientUtil::TRUE : $ClientUtil::FALSE;
} # end cleanup()

################################################################################
# Subroutine Name: get_free_devices
# Objective:       retrieve list of free host volumes 
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices { 
    my ($self, %args) = @_;

    my $bCnt    = 1;   # Asssume user supplied -cnt argument
    my $total   = 0;
    my $refresh = 0;
    my $cnt     = 0;
    my $type    = ".*";
    my $filer;
    my $vol;
    my $lun;
    my %hDevs;
    my @devices;
    my $filers;
    my $protocol;

    @{$filers} = ();

    $self->_clear_err();

    #
    # parse arguments
    #
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Error: WinDevMgr::get_free_devices() " .
                                "Invalid value for argument '-cnt'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: WinDevMgr::get_free_devices() " .
                                "Invalid value for argument '-type'");
                return ($ClientUtil::FALSE); 
            }
            next;
        } elsif (/^-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Error: WinDevMgr::get_free_devices() " .
                                "Invalid value for argument '-filer'");
                return ($ClientUtil::FALSE); 
            }
            next;
        } elsif (/^-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Error: WinDevMgr::get_free_devices() " .
                                "Invalid value for argument '-filers'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Error: WinDevMgr::get_free_devices() " .
                                "Invalid value for argument '-vol'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Error: WinDevMgr::get_free_devices() " .
                                "Invalid value for argument '-lun'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: WinDevMgr::get_free_devices() " .
                                "Invalid value for argument '-refresh'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        } elsif (/^-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: WinDevMgr::get_free_devices() " .
                                "Invalid value for argument '-protocol'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    } 
    
    #
    # Check if a device to filer map 'dfmap' has already been created or 
    # refresh it if requested and refresh if necessary.
    #
    if (not defined $self->{dfmap} or $refresh == 1) {
        $self->sanlog(-msg => "Creating translation maps; " .
                              "device to filer, vol, and lun");
        my @output=$self->{sanlun}->lun_show();
        if ($self->{sanlun}->error) {
            $self->_set_err($self->{sanlun}->errmsg());
            return $ClientUtil::FALSE;
        }

        #
        # We are expecting @output to look like the following for physical windows machines:
        #
        #   filer:         lun-pathname         device filename          adapter        protocol        lun size        lun state
        # fas270-rtp08: /vol/nate_1_1/nate_dell2650-rtp8_windows_0   \\\\.\\PHYSICALDRIVE1         TBD       iscsi         2104M (2206203904)     OK
        # fas270-rtp07: /vol/nate_1_1/nate_dell2650-rtp8_windows_0   \\\\.\\PHYSICALDRIVE2         TBD       iscsi         2104M (2206203904)     OK
        #
        # but @output can also look like this for virtual windows machines:
        #
        #   filer:         lun-pathname         device filename          adapter        protocol        lun size        lun state
        # fas270-rtp07: /vol/nate_3_1_ibm3550_rtp01_guest02/nate_ibm3550-rtp01-guest02_windows_0   \\\\.\\PHYSICALDRIVE4         TBD       none         4200M (4404019200)     OK
        # :    \\\\.\\PHYSICALDRIVE1         TBD       none         3092M (3242196992)     OK
        # fas270-rtp08: /vol/nate_3_1_ibm3550_rtp01_guest02/nate_ibm3550-rtp01-guest02_windows_0   \\\\.\\PHYSICALDRIVE3         TBD       none         4200M (4404019200)     OK
        # :    \\\\.\\PHYSICALDRIVE2         TBD       none         3092M (3242196992)     OK
        #
        
        #
        # Look at each line from sanlun and create a hash for each lun
        # Take this hash and add it to a dfmap.
        #
        foreach my $line (@output) {
            #
            # skip first line (header)
            #
            next if ($line =~ /lun-pathname/i);

            #
            # extract the info from the lun show line
            #
            my ($filer_name, $lunpath, $device_filename, $device_protocol) = 
                ($line =~ /(.*?):(.*?)([\\\.]*PHYSICALDRIVE\d+)\s+\S+\s+(\S+)\s+/);

            #
            # eliminate whitespace in the lun path
            #
            $lunpath =~ s/\s//g;
            
            #
            # extract the vol name and lun name.
            #
            my ($vol_name, $lun_name) = ($lunpath =~ /^\/vol\/(.*?)\/(.*)/);
        
            #
            # pull out the drive number
            #
            my ($drive_number) = ($device_filename =~ /PHYSICALDRIVE(\d+)/);

            $hDevs{$device_filename} = { filer        => $filer_name,
                                         vol          => $vol_name,
                                         lun          => $lun_name,
                                         drive_number => $drive_number,
                                         protocol     => $device_protocol };
            
            $self->sanlog(-msg_level => 3,
                          -msg       => "\nDevice Info for ($device_filename) dfmap:\n" .
                                        "  filer    = $filer_name\n"                    .
                                        "  vol      = $vol_name\n"                      .
                                        "  lun      = $lun_name\n"                      .
                                        "  protocol = $device_protocol\n");
        }
        
        #
        # save map for next call of function
        #
        $self->{dfmap} = \%hDevs;
    } 

    #
    # If not -cnt then assume all free disks 
    # under lvm control are wanted
    #
    unless (defined $cnt and length($cnt) and $cnt > 0) {
        $cnt = 10000;
        $bCnt = 0;
    }

    #
    # We are not currently supporting LVM for windows.  What this means is
    # that at this point we are going to check the filer|vol|lun filters and
    # return the matching devices.
    # 
    foreach my $device (keys %{$self->{dfmap}}) {
        #
        # Apply filters to any of the free devices found
        #
        if (defined $filer and length $filer) {
            $self->sanlog(-msg => "\nFilter filer: $filer\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i);
        }

        if (defined $vol and length $vol) {
            $self->sanlog(-msg => "\nFilter volume: $vol\nDevice: $device\nDevice volume: " . $self->{dfmap}{$device}{vol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i);
        }

        if (defined $lun and length $lun) {
            $self->sanlog(-msg => "\nFilter lun: $lun\nDevice: $device\nDevice lun: " . $self->{dfmap}{$device}{lun},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{lun} !~ /^$lun$/i);
        }

        if (defined $protocol and length $protocol) {
            $self->sanlog(-msg => "\nFilter protocol: $protocol\nDevice: $device\nDevice protocol: " . $self->{dfmap}{$device}{protocol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{protocol} !~ /$protocol/i);
        }

        # Filter on a filer list if provided by caller
        if (defined $filers and scalar(@{$filers}) != 0) {
            my $found = 0;
            # Look through each filer in the filers list
            foreach my $filer_f (@{$filers}) {
                $self->sanlog(-msg => "\nFilter filer(s): $filer_f\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                              -msg_level => 3);
                # If filer matches devices filer, set found to 1
                if ($self->{dfmap}{$device}{filer} =~ /^$filer_f$/i) {
                    $found = 1;
                }
            }

            # If the device is not part of the filers list caller requested skip it.
            if ($found == 0) {
                next;
            }
        }

        push @devices, $device if (++$total <= $cnt);
    } 

    #
    # If user supplied cnt and free devices fell short
    # generate error
    #
    if ($bCnt eq 1 and scalar(@devices) < $cnt) {
        $self->_set_err(
            "Error: WinDevMgr::get_free_devices() ",
            "Only ".scalar(@devices)." of $cnt requested devices were found");
    }

    $self->sanlog(-msg => "Devices returned: @devices");
    
    return (wantarray()) ? @devices : \@devices; 
} 

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new devices object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $load_devices, $devObj, $log, $parent, $hosttype);

    $self = {};

    bless $self, $class;    

    $self->{debug_level} = 10; # TBD 10 or 1?

    #
    # parse arguments
    #
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: WinDevMgr::_new() " .
                                "Invalid value for argument '-parent'");
                return ($ClientUtil::FALSE); 
            }
            next;
        } elsif (/-?loghdl/) {
            if (not length($log = $args{$_})) {
                $self->_set_err("Error: WinDevMgr::_new() " .
                                "Invalid value for argument '-loghdl'");
                return ($ClientUtil::FALSE); 
            }
            next;
        } elsif (/-?load_devices/) {
            if (not length($load_devices = $args{$_})) {
                $self->_set_err("Error: WindowsDevMgr::_new() Invalid value for argument '-load_devices'");
                return ($self);
            }
            next;
        }
    } 

    # set default for load_devices (true)
    if (not defined($load_devices)) {
        $load_devices = 1;
    }
    
    #
    # check for existance of required -parent argument
    #
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Error: WinDevMgr::_new()",
                            "Invalid value for argument '-conn'");
            return $self; 
        }
        $hosttype = $args{-parent}->{hostp}{hosttype};
    } else {
        $self->_set_err("Error: WinDevMgr::_new()",
                        "Missing required argument '-parent'");
        return $self; 
    }

    #
    # initilize the object members
    #
    $self->{conn}      = $parent->get_connection();
    $self->{sanlun}    = $parent->get_sanlun();
    $self->{WinClient} = $parent->{WinClient};
    $self->{parent}    = $parent;

    #
    # Allow CLI commands to be logged;
    #
    $self->{log} = $log if (defined $log);

    #
    # Set the debug level to that of the parent if you can
    #
    $self->{debug_level} = $parent->{debug_level}
        if (defined $parent->{debug_level});
    
    $self->sanlog(-msg_level => 3,
                  -msg       => "Instantiating a WinDevMgr Object");
    #
    # Find the devices on the client.
    #
    $self->{devices} = [];

    #
    # We should have called getDevices and it filled the WinClient
    # disks element with a hash of disk objects keyed by disk number
    #
    if ($load_devices == 1) {
        foreach my $key (keys %{$self->{WinClient}->{disks}}) {
            my $diskObj = $self->{WinClient}->{disks}{$key};
    
            #
            # Don't mess with the boot disk
            #
            next if ($diskObj->{boot});
    
            #
            # Don't mess with the quorum disk
            #
            next if ($diskObj->{isQuorum});
    
            #
            # Don't mess with LUNs not owned by node
            #
            next if ($diskObj->{lunState} =~ /Clus/i);
    
            $self->sanlog(-msg_level => 3,
                          -msg       => "Adding Device "     .
                                        $diskObj->{rawName} .
                                        "\n");
    
            #
            # Create a device object
            #
            $devObj = $self->create_device(-name => $diskObj->{device});
            if ($devObj->error()) {
                my @tempmsg = $devObj->errmsg();
                $self->_set_err("Error: WinDevMgr::_new()"        .
                                "Unable to create device object:" .
                                "@tempmsg");
                return $self; 
            }
    
            #
            # Add info to the device object.  We will basicaly be copying
            # from the WinClient:Disk object for this.
            #
            $devObj->load();
        } 
    }
    
    return($self);
} 

################################################################################
# Subroutine Name: _delete_devices
# Objective:       delete devices from host
# Details:         See POD documentation at the end of this file
################################################################################
sub _delete_devices {
    my ($self, @devices) = @_;

    #
    # TBD
    #
    return $ClientUtil::TRUE;
} 

1;

