# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary Provides an interface to manage FCP Target object
##          Modeled after lib\San\ControllerCluster\Filer10M\Vserver\Fcptarget.pm
##          but stripped to only what SanClient Uses.
##
## @author dl-san-automation
## @status public

package San::SanClient::Fcptarget;

# compiler directives
use strict;

use Tharn qw(say param disconnect $Log);
use Params::Validate qw( SCALAR ARRAYREF HASHREF OBJECT );
use Data::Dumper;
use Error;
use San::SanLog;
use San::SanUtil;
use San::SanConsts;
use San::SanCache;

use vars qw(@ISA);
@ISA = qw( Error SanLog SanUtil SanCache SanConsts);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Method Name: get()
# Objective:   Get field from self
# Details:     See POD documentation at the end of this file
################################################################################
sub get {
    my ($self, $field) = @_;
    $Log->enter();
    $Log->exit();
    # the get() will handle error
    return $self->{$field};
} ## end sub get

################################################################################
# Method Name: get_name()
# Objective:   Get name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;
    $Log->enter();
    $Log->exit();
    # the get() will handle error
    return $self->get("name");
} ## end sub get_name

################################################################################
# Method Name: get_nodename()
# Objective:   Get nodename
# Details:     See POD documentation at the end of this file
################################################################################
sub get_nodename {
    my ($self) = @_;
    $Log->enter();
    $Log->exit();
    # the get() will handle error
    return $self->get("nodename");
} ## end sub get_nodename

################################################################################
# Method Name: get_state()
# Objective:   Get state
# Details:     See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;
    $Log->enter();
    $Log->exit();
    # the get() will handle error
    return $self->get("state");
} ## end sub get_state

################################################################################
# Method Name: get_vserver_name()
# Objective:   Get vserver_name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_vserver_name {
    my ($self) = @_;
    $Log->enter();
    $Log->exit();
    # the get() will handle error
    return $self->get("vserver_name");
} ## end sub get_vserver_name

################################################################################
# Method Name: get_cluster_name()
# Objective:   Get cluster_name
# Details:     See POD documentation at the end of this file
################################################################################
sub get_cluster_name {
    my ($self) = @_;
    $Log->enter();
    $Log->exit();
    # the get() will handle error
    return $self->get("cluster_name");
} ## end sub get_cluster_name

################################################################################
# Method Name: is_mgmt_port()
# Objective:   determine Fcptarget management port?
# Details:     See POD documentation at the end of this file
################################################################################
sub is_mgmt_port {
    my ($self) = @_;
    $Log->enter();
    $Log->exit();
    # the get() will handle error
    return $self->get("is_mgmt_port");
} ## end sub is_mgmt_port

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

###############################################################################
# Method Name: _new
# Objective: Create new FCP Target object
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, @inits);
    $Log->enter();
    # parameter initialization
    $self = {};

    # bless instance into class
    bless $self, $class;

    # remove the dash '-' from input argument list which will be used to
    # create 'get' accessor methods on the fly
    %args = &SanUtil::strip_dash(%args);

    # support both name and fcptarget
    # switch to fcptarget now because SUPER::_new is being used by all objects
    # and cdef requires fcptarget instead of name
    if (exists $args{name}) {
        $args{fcptarget} = $args{name};
    }

    # Set up the base arguments of the Object.
    $self->{name} = $args{fcptarget};
    $self->{debug_level} = &Tharn::param('DEBUG_LEVEL', '-default', 0);

    # dump the argument hash
    $self->sanlog(
        -msg       => Dumper(\%args),
        -msg_level => 2
    );

    # store/initalize fcp target elements
    foreach my $key (keys(%args)) {
        $self->{lc($key)} = $args{$key};
        push @{$self->{fields}}, lc($key);
    }
    $Log->exit();
    return $self;
} ## end sub _new

1;

#############################################################################
#                             POD Documentation
#############################################################################
## @pod here

=pod

=head1 Package Name

Fcptarget.pm

=head1 Author

dl-san-automation

=head1 Description

Provides an interface to manage an Fcp target. The package should be instantiated via get_fcp_targets().

=head1 See Also

Fcptarget package used base class B<San::CommonInterface> that also inherits other modules. See these modules for further interface detail.

B<Error.pm, San::SanLog.pm, San::SanUtil.pm, San::SanCache.pm>

=head1 Attributes

=head2 B<name>

string = Name of the lif target

=head2 B<parent_obj>

string = Parent object

=head2 B<fcp_obj>

string = FCP object

=head2 B<nodename>

string = Name of the target nodename

=head2 B<state>

string = state of the lif

=head1 B<Public Methods>

=head2 B<get_name()>

Get lif target name

=over 4

=item Inputs

None

=item Outputs

Return target lif name

=item Examples

 # ex. 1: get fcp target name
 $target_name = $fcptarget_obj->get_name();
 if (fcptarget_obj->error()) {
     ...
 }

=back

=head2 B<get_parent_obj()>

Get parent object

=over 4

=item Inputs

None

=item Outputs

Return parent object of Fcptarget (Vserver object)

=item Examples

 # ex. 1: get parent object
 $vs_obj = $fcptarget_obj->get_parent_obj();
 if (fcptarget_obj->error()) {
     ...
 }

=back

=head2 B<get_fcp_obj()>

Get FCP object

=over 4

=item Inputs

None

=item Outputs

Return FCP object

=item Examples

 # ex. 1: get fcp object
 $fcp_obj = $fcptarget_obj->get_fcp_obj();
 if (fcptarget_obj->error()) {
     ...
 }

=back

=head2 B<get_nodename()>

Get FCP nodename

=over 4

=item Inputs

None

=item Outputs

Return FCP nodename

=item Examples

 # ex. 1: get fcp nodename
 $nodename = $fcptarget_obj->get_nodename();
 if (fcptarget_obj->error()) {
     ...
 }

=back

=head2 B<get_state()>

Get state of the target lif

=over 4

=item Inputs

None

=item Outputs

Return state of target lif (ex. up or down)

=item Examples

 # ex. 1: get state of target lif
 $state = $fcptarget_obj->get_state();
 if (fcptarget_obj->error()) {
     ...
 }

=back

=head2 B<get_vserver_name()>

Get FCP vserver name

=over 4

=item Inputs

None

=item Outputs

Return FCP vserver name

=item Examples

 # ex. 1: get fcp vserver name
 $vserver_name = $fcptarget_obj->get_vserver_name();
 if (fcptarget_obj->error()) {
     ...
 }

=back

=head2 B<get_cluster_name()>

Get FCP cluster name

=over 4

=item Inputs

None

=item Outputs

Return FCP cluster name

=item Examples

 # ex. 1: get fcp cluster name
 $cluster_name = $fcptarget_obj->get_cluster_name();
 if (fcptarget_obj->error()) {
     ...
 }

=back
