# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to Host Utilities
## @author ljimmy@netapp.com
## @status private
## @pod here

# Package Name
##############
package HU;

# Package Requirements
######################
use Tharn;
use San::SanLog;
use Error;

# Other Definitions
###################
use strict;
use vars qw( @ISA );
@ISA = qw(Error SanLog);

###############################################################################
# P U B L I C    M E T H O D S
###############################################################################

###############################################################################
# Method:    new
# Objective: HU object constructor.
###############################################################################
sub new {
    my ($class, %args) = @_;
    my ($self, $hosttype);
    
    $self = {};

    ## bless the object upfront in case an error occurs.
    bless $self, $class;

    # check for existance of required -parent argument
    if (defined($args{-parent})) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Error: " . ref($self) . "::new() " .
                            "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->sanlog("-msg", "-parent not defined!!!");
        $self->_set_err("Error: " . ref($self) . "::new() " . 
                        "Missing required argument '-parent'");
        return $self;
    }
    
    $hosttype = $args{-parent}->hosttype();

    ## load the appropriate HU for the hosttype
    if ($hosttype =~ /unix-solaris/i) {
        require San::SanClient::HU::SolarisHU;   
        $self = San::SanClient::HU::SolarisHU->_new(%args);
    } elsif ($hosttype =~ /unix-hpux/i) {
        require San::SanClient::HU::HPUXHU;   
        $self = San::SanClient::HU::HPUXHU->_new(%args);
    } elsif ($hosttype =~ /unix-aix/i) {
        require San::SanClient::HU::AixHU;   
        $self = San::SanClient::HU::AixHU->_new(%args);
    } elsif ($hosttype =~ /unix-linux/i) {
        require San::SanClient::HU::LinuxHU;   
        $self = San::SanClient::HU::LinuxHU->_new(%args);
    } elsif ($hosttype =~ /windows/i) {
        require San::SanClient::HU::WinHU;   
        $self = San::SanClient::HU::WinHU->_new(%args);
    } elsif ($hosttype =~ /linux-vmware/i) {
        require San::SanClient::HU::ESXHU;
        $self = San::SanClient::HU::ESXHU->_new(%args);
    } else {
        $self->_set_err("Error: " . ref($self) . "::new() Platform '$hosttype' " .
                        "not currently supported by HU object");
        return $self;                           
    }

    #
    # By returning $self here we are returning a parent class such
    # as SolarisHU, not a HU object.  This is OK because
    # the OS specific HU::XXXHU classes add HU in @ISA
    # so HU is the base class for the parent
    #
    return $self;
}

###############################################################################
# Subroutine Name: san_version
# Objective: Returns the host utility san version
###############################################################################
sub san_version {
    my ($self) = @_;
    return $self->{san_version};
}

###############################################################################
# Subroutine Name: generate_switch_info_log
# Objective: Determine the type of the switch and generate info log using
#            corresponding info script.
###############################################################################
sub generate_switch_info_log {
    my ($self, $switch) = @_;
    my ($login, $pass, $type, @locations);

    ## determine usrname/password/type of the switch
    my $hostrec = host($switch);
    my $conntype = &ClientUtil::get_device_connection($switch);
    $conntype =~ s/#//;
    $login = $hostrec->{connection_fields}->{$conntype}->{username};
    $pass  = $hostrec->{connection_fields}->{$conntype}->{password};
    $type  = $hostrec->{hosttype};

    ## generate the log
    if ($type =~ /brocade/i) {
        @locations = $self->generate_brocade_info_log($switch, $login, $pass);
        if ($self->error()) {
            return @locations;
        } 
    } elsif ($type =~ /cisco/i) {
        @locations = $self->generate_cisco_info_log($switch, $login, $pass);
        if ($self->error()) {
            return @locations;
        }
    } elsif ($type =~ /qlogic/i) {
        @locations = $self->generate_qlogic_info_log($switch, $login, $pass);
        if ($self->error()) {
            return @locations;
        }
    } elsif ($type =~ /mcdata/i) {
        @locations = $self->generate_mcdata_info_log($switch, $login, $pass);
        if ($self->error()) {
            return @locations;
        }
    } else {
        $self->_set_err("Error: " . ref($self) . "::collect_switch_info " .
                        "Unsupported switch type => $type");
        return ();
    }
    
    # return an array consisting of full path name of tar ball and subdirectory
    # to be created in the remote FTP directory.
    return @locations;
}

1;
