# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to the AIX Host Utilities
## @author Robert Hunter - hrivera@netapp.com
## @status private
## @pod here

# Package Name
##############
package San::SanClient::HU::AixHU;

# Package Requirements
######################
use San::SanClient;
use San::SanLog;
use Tharn;
use Error;
use Digest::MD5;

# Other Definitions
###################
use strict;
use vars qw( @ISA );
@ISA = qw(HU Error SanLog);

####################################################
# Method:    _new
# Objective: AixHU object constructor
####################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self, $oem_rebrand, $hu_bin_path, $obj_template);

    $self = {};
    bless $self, $class;

    ## parse the arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length ($parent = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-parent => $parent");
                return $self;
            }
            last;
        } elsif (/-?obj_template/) {
            if (not length ($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            last;
        }
    }

    if (defined($obj_template)) {
        # copy all the keys and return
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        return $self;
    }

    ## check for required parameters
    if (not defined($parent)) {
        $self->_set_err("Error: " . ref($self) . "::_new() -parent not set");
        return $self;
    }

    ## check for OEM rebrand
    $oem_rebrand = $parent->{general}->{OEM_REBRAND};
    if ($oem_rebrand =~ /netapp/i) {
        $oem_rebrand = "NetApp";
    } else {
        ## this will be different on each platform
        $oem_rebrand = "Ontap";
    }
    $hu_bin_path   = "/opt/" . $oem_rebrand . "/santools/bin";

    ## set the attributes in the object
    $self->{oem_rebrand} = $oem_rebrand;
    $self->{parent} = $parent;
    $self->{client_name} = $self->{parent}->hostname();
    $self->{hu_bin_path} = $hu_bin_path;

    ## store san_version -v output
    $self->{san_version} = $self->get_san_version();

    return $self;
}

###############################################################################
# Subroutine Name: set_info_params
# Objective: Load the collector specific parameters
###############################################################################
sub set_info_params {
    my ($self, %args) = @_;
    my ($ftp_server, $ftp_login, $ftp_pass, $ftp_dir);
    my ($log_dir, $ftp_script, $timestamp);

    ## init variables
    $log_dir    = "/tmp/logviewer_";
    $ftp_script = "/x/eng/localtest/noarch/bin/san_collector_ftp.pl";

    ## parse the arguments
    foreach (keys %args) {
        if (/-?ftp_server/) {
            if (not length ($ftp_server = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-ftp_server => $ftp_server");
                return $self;
            }
            next;
        } elsif (/-?ftp_login/) {
            if (not length ($ftp_login = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-ftp_login => $ftp_login");
                return $self;
            }
            next;
        } elsif (/-?ftp_pass/) {
            if (not length ($ftp_pass = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-ftp_pass => $ftp_pass");
                return $self;
            }
            next;
        } elsif (/-?ftp_dir/) {
            if (not length ($ftp_dir = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-ftp_dir => $ftp_dir");
                return $self;
            }
        } elsif (/-?log_dir/) {
            if (not length ($log_dir = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-log_dir => $log_dir");
                return $self;
            }
            next;
        } elsif (/-?ftp_script/) {
            if (not length ($ftp_script = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-ftp_script => $ftp_script");
                return $self;
            }
            next;
        } elsif (/-?timestamp/) {
            if (not length ($timestamp = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-timestamp => $timestamp");
                return $self;
            }
            next;
        }
    }

    ## set the attributes in the object
    $self->{ftp_server} = $ftp_server;
    $self->{ftp_login} = $ftp_login;
    $self->{ftp_pass} = $ftp_pass;
    $self->{ftp_dir} = $ftp_dir;
    $self->{ftp_script} = $ftp_script;
    $self->{log_dir} = $log_dir . $timestamp;

    return;
}

###############################################################################
# Subroutine Name: get_san_version
# Objective: Run san_version to see the  version of host utility being used.
###############################################################################
sub get_san_version {
    my ($self) = @_;

    ## run san_version
    my $cmd = "$self->{hu_bin_path}/san_version -v";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);

    # Disabling status check until all teams have a valid host utilities
    ## toolkit setup on client machines
    #if ($status > 0) {
    #    $self->_set_err("Error: " . ref($self) .
    #                    "::get_san_version() command '$cmd' failed");
    #    return ();
    #}

    # get toolkit version and store as attribute on object
    my $toolkit_version = $output[0];
    $toolkit_version =~ s/.*version //;
    $self->{toolkit_version} = $toolkit_version;

    # get sanlun version and store as attribute on object
    my $sanlun_version = $output[1];
    $sanlun_version =~ s/sanlun version\s+//;
    $self->{sanlun_version} = $sanlun_version;

    return @output;
}

###############################################################################
# Subroutine Name: generate_client_info_log
# Objective: Run client info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "client" for remote subdirectory to transfer.
###############################################################################
sub generate_client_info_log {
    my ($self) = @_;
    my $log_path = $self->{log_dir} . "/ntap_aix_" . $self->{client_name} . ".tar.Z";

    ## run hpux_info
    my $cmd = "$self->{hu_bin_path}/aix_info -f -d $self->{log_dir} -n $self->{client_name}";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd, 2400);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_client_log() command '$cmd' failed");
        return ($log_path);
    }

    ## check to see if hpux_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /AIX system info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_client_info_log() log generation failed");
        return ($log_path);
    }

    ## return location of generated tar ball
    return ($log_path, "client/aix");
}

###############################################################################
# Subroutine Name: generate_filer_info_log
# Objective: Run filer info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "filer" for remote subdirectory to transfer.
###############################################################################
sub generate_filer_info_log {
    my ($self, $filer) = @_;
    my $log_path = $self->{log_dir} . "/netapp_filer_" . $filer . ".tar.Z";

    ## run filer_info
    my $cmd = "$self->{hu_bin_path}/filer_info -f -d $self->{log_dir} -n $filer";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_filer_log() command '$cmd' failed");
        return ($log_path);
    }

    ## Check to see if filer_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Filer $filer info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_filer_info_log() log generation failed");
        return ($log_path);
    }

    return ($log_path, "filer");
}

###############################################################################
# Subroutine Name: generate_brocade_info_log
# Objective: Run brocade info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "switch" for remote subdirectory to transfer.
###############################################################################
sub generate_brocade_info_log {
    my ($self, $switch, $login, $pass) = @_;
    my $log_path = $self->{log_dir} . "/brocade_switch_" . $switch . ".tar.Z";

    ## Run brocade_info
    my $cmd = "$self->{hu_bin_path}/brocade_info -f -d $self->{log_dir} -n $switch -l $login:$pass";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_brocade_info_log() command '$cmd' failed");
        return ($log_path);
    }

    ## Check to see if brocade_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Switch $switch info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_brocade_info_log() log generation failed");
        return ($log_path);
    }

    return ($log_path, "switch/brocade");
}

###############################################################################
# Subroutine Name: generate_cisco_info_log
# Objective: Run brocade info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "switch" for remote subdirectory to transfer.
###############################################################################
sub generate_cisco_info_log {
    my ($self, $switch, $login, $pass) = @_;
    my $log_path = $self->{log_dir} . "/cisco_switch_" . $switch . ".tar.Z";

    ## Run cisco_info
    my $cmd = "$self->{hu_bin_path}/cisco_info -f -d $self->{log_dir} -n $switch -l $login:$pass";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_cisco_info_log() command '$cmd' failed");
        return ($log_path);
    }

    ## Check to see if cisco_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Switch $switch info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_cisco_info_log() log generation failed");
        return ($log_path);
    }

    return ($log_path, "switch/cisco");
}

###############################################################################
# Subroutine Name: generate_mcdata_info_log
# Objective: Run brocade info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "switch" for remote subdirectory to transfer.
###############################################################################
sub generate_mcdata_info_log {
    my ($self, $switch, $login, $pass) = @_;
    my $log_path = $self->{log_dir} . "/mcdata_switch_" . $switch . ".tar.Z";

    ## Run brocade_info
    my $cmd = "$self->{hu_bin_path}/mcdata_info -f -d $self->{log_dir} -n $switch -l $login:$pass";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_mcdata_info_log() command '$cmd' failed");
        return ($log_path);
    }

    ## Check to see if mcdata_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Switch $switch info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_mcdata_info_log() log generation failed");
        return ($log_path);
    }

    return ($log_path, "switch/mcdata");
}

###############################################################################
# Subroutine Name: generate_qlogic_info_log
# Objective: Run brocade info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "switch" for remote subdirectory to transfer.
###############################################################################
sub generate_qlogic_info_log {
    my ($self, $switch, $login, $pass) = @_;
    my $log_path = $self->{log_dir} . "/qlogic_switch_" . $switch . ".tgz";

    ## Run qlogic_info
    my $cmd = "$self->{hu_bin_path}/qlogic_info -f -d $self->{log_dir} -n $switch -l $login:$pass";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_qlogic_info_log() command '$cmd' failed");
        return ($log_path);
    }

    ## Check to see if qlogic_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Switch $switch info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_qlogic_info_log() log generation failed");
        return ($log_path);
    }

    return ($log_path, "switch/qlogic");
}

###############################################################################
# Subroutine Name: collect_info_log
# Objective: Run the collector_ftp perl script on x drive to collect the tar
#            ball info logs to the controller.
###############################################################################
sub collect_info_log {
    my ($self, $file_path, $sub_dir) = @_;

    ## run FTP script
    ## subdirectories "client", "file", or "switch" will be created depending
    ## on value of $sub_dir
    my $cmd = "$self->{ftp_script} " .
                "-user $self->{ftp_login} " .
                "-pass $self->{ftp_pass} " .
                "-tb " . $self->{ftp_dir} . "/" . $sub_dir . " " .
                "-srv $self->{ftp_server} " .
                "-file $file_path";

    $self->sanlog("-msg", "Transferring $file_path via FTP to " .
                  "$self->{ftp_server}:$self->{ftp_dir}/$sub_dir");
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::collect_info_log() command '$cmd' failed");
        return $ClientUtil::FALSE;
    }

    ## check to see if FTP successful or not
    if ($output[0] =~ /FTP Transfer Complete/i) {

        ## extract file name from "/tmp/logviewer/xxxx.tar.gz"
        ## combine with ftp_dir to form path to file on x drive
        my $file = $file_path;
        $file =~ s/^\/(.+)\///;
        $file = "$self->{ftp_dir}/" . $sub_dir . "/$file";

        ## generate MD5 digest
        if(open (FILE, "$file") <= 0) {
            $self->_set_err("Error: " . ref($self) . "::collect_info_log() Failed to open $file");
            return $ClientUtil::FALSE;
        }
        binmode(FILE);
        my $md5 = Digest::MD5->new;
        my $digest = $md5->addfile(*FILE)->hexdigest;

        ## compare digest with the log (third line of output is MD5 digest)
        if ($output[2] eq $digest) {
            $self->sanlog("-msg", "FTP Successful");
            return $ClientUtil::TRUE;
        } else {
            $self->_set_err("Error: " . ref($self) . "MD5 mismatch $output[2] != $digest");
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Error: " . ref($self) . "FTP Failed");
        return $ClientUtil::FALSE;
    }
}

###############################################################################
# Subroutine Name: clean_up_log_dir
# Objective: Remove the temporary log directory used to generate tar balls
###############################################################################
sub clean_up_log_dir {
    my ($self) = @_;

    ## double check that log_dir is not "/"
    if ($self->{log_dir} =~ /^\s*\/\s*$/) {
        $self->_set_err("Error: " . ref($self) . "::clean_up_log_dir() " .
                        "log_dir cannot be '/'");
        return $ClientUtil::FALSE;
    }

    ## force remove log directory
    my $cmd = "rm -rf $self->{log_dir}";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::clean_up_log_dir() command '$cmd' failed");
        return $ClientUtil::FALSE;
    }

    $self->sanlog("-msg", "Log Cleanup Complete");
    return $ClientUtil::TRUE;
}

1;
