# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to the Windows Host Utilities
## @author Robert Hunter - hrivera@netapp.com
## @status private
## @pod here

# Package Name
##############
package San::SanClient::HU::WinHU;

# Package Requirements
######################
use San::SanClient;
use San::SanLog;
use Tharn;
use Error;
use Digest::MD5;

# Other Definitions
###################
use strict;
use vars qw( @ISA );
@ISA = qw(HU Error SanLog);

####################################################
# Method:    _new
# Objective: WinHU object constructor
####################################################
sub _new {
    my ($class, %args) = @_;
    my ($parent, $self, $oem_rebrand, $hu_bin_path);
    my ($obj_template);

    $self = {};
    bless $self, $class;

    ## parse the arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length ($parent = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-parent => $parent");
                return $self;
            }
            last;
        } elsif (/-?obj_template/) {
            if (not length ($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            last;
        }
    }

    if (defined($obj_template)) {
        # copy all the keys and return
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        return $self;
    }

    ## check for required parameters
    if (not defined($parent)) {
        $self->_set_err("Error: " . ref($self) . "::_new() -parent not set");
        return $self;
    }

    ## check for OEM rebrand
    $oem_rebrand = $parent->{general}->{OEM_REBRAND};
    if ($oem_rebrand =~ /netapp/i) {
        $oem_rebrand = "NetApp";
    } else {
        ## this will be different on each platform
        $oem_rebrand = "ontap";
    }
    $hu_bin_path   = "c:\\PROGRA~1\\" . $oem_rebrand . "\\WINDOW~1";

    ## set the attributes in the object
    $self->{oem_rebrand} = $oem_rebrand;
    $self->{parent} = $parent;
    $self->{client_name} = $self->{parent}->hostname();
    $self->{hu_bin_path} = $hu_bin_path;

    ## store san_version -v output
    $self->{san_version} = $self->get_san_version();

    return $self;

}

###############################################################################
# Subroutine Name: set_info_params
# Objective: Load the collector specific parameters
###############################################################################
sub set_info_params {
    my ($self, %args) = @_;
    my ($ftp_server, $ftp_login, $ftp_pass, $ftp_dir);
    my ($log_dir, $ftp_script, $perl_dir, $hu_bin_path, $timestamp, $cifs_dir);
    my ($file_transport_type) = "ftp";

    ## init variables
    $log_dir    = "c:\\temp\\logviewer_";
    $ftp_script = "x:\\SAN\\bin\\common\\san_collector_ftp.pl";
    $perl_dir   = "x:\\perl\\bin\\perl";

    ## parse the arguments
    foreach (keys %args) {
        if (/-?ftp_server/) {
            if (not length ($ftp_server = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-ftp_server => $ftp_server");
                return $self;
            }
            next;
        } elsif (/-?ftp_login/) {
            if (not length ($ftp_login = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-ftp_login => $ftp_login");
                return $self;
            }
            next;
        } elsif (/-?ftp_pass/) {
            if (not length ($ftp_pass = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-ftp_pass => $ftp_pass");
                return $self;
            }
            next;
        } elsif (/-?ftp_dir/) {
            if (not length ($ftp_dir = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-ftp_dir => $ftp_dir");
                return $self;
            }
        } elsif (/-?log_dir/) {
            if (not length ($log_dir = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-log_dir => $log_dir");
                return $self;
            }
            next;
        } elsif (/-?ftp_script/) {
            if (not length ($ftp_script = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-ftp_script => $ftp_script");
                return $self;
            }
            next;
        } elsif (/-?timestamp/) {
            if (not length ($timestamp = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-timestamp => $timestamp");
                return $self;
            }
            next;
        } elsif (/-?cifs_dir/) {
            if (not length ($cifs_dir = $args{$_})) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-cifs_dir => $cifs_dir");
                return $self;
            }
            next;
        } elsif (/-?file_transport_type/) {
            if (not length ($file_transport_type = $args{$_}) or $file_transport_type !~ /ftp|cifs/i) {
                $self->_set_err("Error: " . ref($self) . "::_new() Invalid value for " .
                                "-file_transport_type => $file_transport_type");
                return $self;
            }
            next;
        }
    }

    ## set the attributes in the object
    $self->{ftp_server} = $ftp_server;
    $self->{ftp_login} = $ftp_login;
    $self->{ftp_pass} = $ftp_pass;
    $self->{ftp_dir} = $ftp_dir;
    $self->{ftp_script} = $ftp_script;
    $self->{log_dir} = $log_dir . $timestamp;
    $self->{perl_dir} = $perl_dir;
    $self->{cifs_dir} = $cifs_dir;
    $self->{file_transport_type} = $file_transport_type;

    return;
}

###############################################################################
# Subroutine Name: get_san_version
# Objective: Run san_version to see the  version of host utility being used.
###############################################################################
sub get_san_version {
    my ($self) = @_;

    ## run san_version
    my $cmd = "$self->{hu_bin_path}\\san_version -v";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);

    # if this failed, then most likely HU is not installed, just initializing the versions to null
    if ($status > 0) {
        $self->{toolkit_version} = "NOT_INSTALLED";
        $self->{sanlun_version} = "NOT_INSTALLED";
        return ();
    }

    # get toolkit version and store as attribute on object
    my $toolkit_version = $output[0];
    $toolkit_version =~ s/.*version //;
    $self->{toolkit_version} = $toolkit_version;

    # get sanlun version and store as attribute on object
    my $sanlun_version = $output[1];
    $sanlun_version =~ s/sanlun version\s+//;
    $self->{sanlun_version} = $sanlun_version;

    return @output;
}

###############################################################################
# Subroutine Name: generate_client_info_log
# Objective: Run client info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "client" for remote subdirectory to transfer.
###############################################################################
sub generate_client_info_log {
    my ($self) = @_;
    my $log_path = $self->{log_dir} . "\\windows_info_" . $self->{client_name} . ".tar.gz";

    ## run windows_info
    my $cmd = "$self->{hu_bin_path}\\windows_info -f -d $self->{log_dir} -n $self->{client_name}";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd, 2400);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_client_log() command '$cmd' failed");
        return ($log_path);
    }

    ## check to see if windows_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Host_$self->{client_name} info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_client_info_log() log generation failed");
        return ($log_path);
    }

    ## return location of generated tar ball
    return ($log_path, "client/windows");
}

###############################################################################
# Subroutine Name: generate_filer_info_log
# Objective: Run filer info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "filer" for remote subdirectory to transfer.
###############################################################################
sub generate_filer_info_log {
    my ($self, $filer) = @_;
    my $log_path = $self->{log_dir} . "\\netapp_filer_" . $filer . ".tar.gz";

    ## run filer_info
    my $cmd = "$self->{hu_bin_path}/filer_info -f -d $self->{log_dir} -n $filer";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_filer_log() command '$cmd' failed");
        return ($log_path);
    }

    ## Check to see if filer_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Filer $filer info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_filer_info_log() log generation failed");
        return ($log_path);
    }

    return ($log_path, "filer");
}

###############################################################################
# Subroutine Name: generate_brocade_info_log
# Objective: Run brocade info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "switch" for remote subdirectory to transfer.
###############################################################################
sub generate_brocade_info_log {
    my ($self, $switch, $login, $pass) = @_;
    my $log_path = $self->{log_dir} . "\\brocade_switch_" . $switch . ".tar.gz";

    ## Run brocade_info
    my $cmd = "$self->{hu_bin_path}/brocade_info -f -d $self->{log_dir} -n $switch -l $login:$pass";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_brocade_info_log() command '$cmd' failed");
        return ($log_path);
    }

    ## Check to see if brocade_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Switch $switch info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_brocade_info_log() log generation failed");
        return ($log_path);
    }

    return ($log_path, "switch/brocade");
}

###############################################################################
# Subroutine Name: generate_cisco_info_log
# Objective: Run brocade info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "switch" for remote subdirectory to transfer.
###############################################################################
sub generate_cisco_info_log {
    my ($self, $switch, $login, $pass) = @_;
    my $log_path = $self->{log_dir} . "\\cisco_switch_" . $switch . ".tar.gz";

    ## Run cisco_info
    my $cmd = "$self->{hu_bin_path}/cisco_info -f -d $self->{log_dir} -n $switch -l $login:$pass";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_cisco_info_log() command '$cmd' failed");
        return ($log_path);
    }

    ## Check to see if cisco_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Switch $switch info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_cisco_info_log() log generation failed");
        return ($log_path);
    }

    return ($log_path, "switch/cisco");
}

###############################################################################
# Subroutine Name: generate_mcdata_info_log
# Objective: Run brocade info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "switch" for remote subdirectory to transfer.
###############################################################################
sub generate_mcdata_info_log {
    my ($self, $switch, $login, $pass) = @_;
    my $log_path = $self->{log_dir} . "\\mcdata_switch_" . $switch . ".tar.gz";

    ## Run brocade_info
    my $cmd = "$self->{hu_bin_path}/mcdata_info -f -d $self->{log_dir} -n $switch -l $login:$pass";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_mcdata_info_log() command '$cmd' failed");
        return ($log_path);
    }

    ## Check to see if mcdata_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Switch $switch info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_mcdata_info_log() log generation failed");
        return ($log_path);
    }

    return ($log_path, "switch/mcdata");
}

###############################################################################
# Subroutine Name: generate_qlogic_info_log
# Objective: Run brocade info script on the client to generate the tar ball for
#            collection. Returns path to tar ball generated and directory name
#            "switch" for remote subdirectory to transfer.
###############################################################################
sub generate_qlogic_info_log {
    my ($self, $switch, $login, $pass) = @_;
    my $log_path = $self->{log_dir} . "\\qlogic_switch_" . $switch . ".tgz";

    ## Run qlogic_info
    my $cmd = "$self->{hu_bin_path}/qlogic_info -f -d $self->{log_dir} -n $switch -l $login:$pass";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: " . ref($self) . "::generate_qlogic_info_log() command '$cmd' failed");
        return ($log_path);
    }

    ## Check to see if qlogic_info ran successfully
    my $program_ran = 1;
    foreach my $line (@output) {
        if ($line =~ /Switch $switch info is in/) {
            $program_ran = 0;
        }
    }

    if ($program_ran == 0) {
        $self->sanlog("-msg", "Log generation successful");
    } else {
        $self->_set_err("Error: " . ref($self) . "::generate_qlogic_info_log() log generation failed");
        return ($log_path);
    }

    return ($log_path, "switch/qlogic");
}

###############################################################################
# Subroutine Name: transfer_file
# Objective: Transfer the given file to the destination sub_dir via ftp/cifs
###############################################################################
sub transfer_file {
    my ($self, %args) = @_;

    my ($file_path, $sub_dir);

    # parse args
    foreach (keys %args) {
        if (/-?file_path/) {
            if (not length ($file_path = $args{$_})) {
                $self->_set_err("Invalid value for -file_path");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?destination_dir/) {
            if (not length ($sub_dir = $args{$_})) {
                $self->_set_err("Invalid value for -destination_dir");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not defined($file_path) or not defined($sub_dir)) {
        $self->_set_err("Missing required parameters -file_path or -destination_dir");
        return $ClientUtil::FALSE;
    }

    if ($self->{file_transport_type} =~ /ftp/i) {
        ## run FTP script
        ## subdirectories "client", "file", or "switch" will be created depending
        ## on value of $sub_dir
        my $cmd = "$self->{perl_dir} $self->{ftp_script} " .
                  "-user $self->{ftp_login} " .
                  "-pass $self->{ftp_pass} " .
                  "-tb " . $self->{ftp_dir} . "/" . $sub_dir . " " .
                  "-srv $self->{ftp_server} " .
                  "-file $file_path";

        $self->sanlog("-msg", "Transferring $file_path via FTP to " .
                              "$self->{ftp_server}:$self->{ftp_dir}/$sub_dir");
        my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command '$cmd' failed with status ($status)");
            return $ClientUtil::FALSE;
        }

        ## check to see if FTP successful or not
        if ($output[0] =~ /FTP Transfer Complete/i) {

            ## extract file name from "c:\temp\logviewer\xxxx.tar.gz"
            ## combine with ftp_dir to form path to file on x drive
            my $file = $file_path;
            $file =~ s/^c:\\(.*)\\//i;
            $file = "$self->{ftp_dir}/" . $sub_dir . "/$file";

            ## generate MD5 digest
            if(open (FILE, "$file") <= 0) {
                $self->_set_err("Failed to open $file");
                return $ClientUtil::FALSE;
            }
            binmode(FILE);
            my $md5 = Digest::MD5->new;
            my $digest = $md5->addfile(*FILE)->hexdigest;

            ## compare digest with the log (third line of output is MD5 digest)
            if ($output[2] eq $digest) {
                $self->sanlog("-msg", "FTP Successful");
                return $ClientUtil::TRUE;
            } else {
                $self->_set_err("MD5 mismatch $output[2] != $digest");
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_set_err("FTP transfer failed");
            return $ClientUtil::FALSE;
        }
    } elsif ($self->{file_transport_type} =~ /cifs/i) {
        # convert forward slashes to back slashes
        $sub_dir =~ s/\//\\/g;

        # find out windows system drive
        my $systemdrive_cmd = "echo %systemdrive%";
        my ($systemdrive_status, $systemdrive) = &ClientUtil::execute($self->{parent}->{conn}, $systemdrive_cmd);
        if ($systemdrive_status > 0) {
            $self->_set_err("Command ($systemdrive_cmd) failed with status ($systemdrive_status)");
            return $ClientUtil::FALSE;
        }

        # find out windows system root directory
        my $systemroot_cmd  = "echo %systemroot%";
        my ($systemroot_status, $systemroot) = &ClientUtil::execute($self->{parent}->{conn}, $systemroot_cmd);
        if ($systemroot_status > 0) {
            $self->_set_err("Command ($systemroot_cmd) failed with status ($systemroot_status)");
            return $ClientUtil::FALSE;
        }

        # build the root dir regex string
        my $root_dir = "C:";
        $root_dir .= "|$systemdrive" if length($systemdrive);
        $root_dir .= "|$systemroot"  if length($systemroot);

        # dont play with the system drive.
        if ($self->{cifs_dir} =~ /^$root_dir/i) {
            $self->_set_err("Invalid CIFS dir: '$self->{cifs_dir}'");
            return $ClientUtil::FALSE;
        }

        # delete the directory in the cifs share just to be sure theres nothing left over
        my $cmd = "rmdir /S /Q $self->{cifs_dir}\\$sub_dir";
        my ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, "$cmd");
        if ($status != 0 and not grep(/system cannot find the path|system cannot find the file/i, @output)) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }

        # make the sub directory in the cifs share
        $cmd = "mkdir $self->{cifs_dir}\\$sub_dir";
        ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, "$cmd");
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }

        # move the file to the intended location
        $cmd = "move /Y $file_path $self->{cifs_dir}\\$sub_dir";
        ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, "$cmd");
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: clean_up_log_dir
# Objective: Remove the temporary log directory used to generate tar balls
###############################################################################
sub clean_up_log_dir {
    my ($self, $log_dir) = @_;
    my ($status, $systemdrive, $systemroot, @output);

    ## find out windows system drive
    my $systemdrive_cmd = "echo %systemdrive%";
    ($status, $systemdrive) = &ClientUtil::execute($self->{parent}->{conn}, $systemdrive_cmd);
    if ($status > 0) {
        $self->_set_err("Command '$systemdrive_cmd' failed");
        return $ClientUtil::FALSE;
    }

    ## find out windows system root directory
    my $systemroot_cmd  = "echo %systemroot%";
    ($status, $systemroot) = &ClientUtil::execute($self->{parent}->{conn}, $systemroot_cmd);
    if ($status > 0) {
        $self->_set_err("Command '$systemroot_cmd' failed");
        return $ClientUtil::FALSE;
    }

    ## double check that log_dir is not systemroot or systemdrive
    if ($log_dir =~ /^$systemroot$/i) {
        $self->_set_err("log_dir cannot be '$systemroot'");
        return $ClientUtil::FALSE;
    }

    if ($log_dir =~ /^$systemdrive$/i) {
        $self->_set_err("log_dir cannot be '$systemdrive'");
        return $ClientUtil::FALSE;
    }

    ## force remove log directory
    my $cmd = "rd /s /q ${log_dir}";
    ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0 and not grep(/system cannot find the path|system cannot find the file/i, @output)) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    $self->sanlog("-msg", "Log Cleanup Complete");
    return $ClientUtil::TRUE;
}

1;
