# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author thomasm@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::AixHost;

# Package Requirements
######################
use File::Basename;

use San::SanClient;
use San::SanLog;
use San::SanUtil;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new AixHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($device_str, @devices);
    my ($self, $conn, $nfs_mount, $build, $obj_template);
    my ($cmd, $status, @output);

    $self = {};

    bless ($self, $class);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixHost::_new() Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Error: AixHost::_new() Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Error: AixHost::_new()",
                        "Invalid value for argument '-conn'");
        return $self;
    }

    $self->{conn} = $conn;
    $self->{gdb} = "/usr/local/bin/gdb-6.3";
    $self->{TEST_BIN} = "/x/eng/localtest/arch/powerpc-ibm-aix5.1/bin";

    # get architecture
    $cmd = "/usr/bin/uname -p";
    ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::_new() Command ($cmd) returned a bad status ($status)");
        return $self;
    } else {
        $self->{arch} = $output[0];
        $self->sanlog(-msg => "AIX Host Architecture is $self->{arch}");
    }

    # get client OS revision (major/minor)
    $cmd = "/usr/bin/oslevel -r";
    ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::_new() Command ($cmd) returned a bad status ($status)");
        return $self;
    }
    my ($os_major, $os_minor) = split(/\-/, $output[0]);
    $self->{revision} = $os_major;
    $self->{minor_revision} = $os_minor;

    $self->sanlog(-msg => "Major Revision: $self->{revision}, Minor Revision: $self->{minor_revision}");

    $self->{nfs_mount} = undef;

    # flag to note that configuration has not happened. Needed to
    # support Blocks QA libs that might not call configure()
    $self->{configured} = 0;

    # check to see if mp_type is set to NONE (no MP). If it is not
    # choose a default (MPIO). Needed for Blocks QA compliance. MP is
    # always needed on AIX to survive CFO (for FCP). For iSCSI we need
    # to allow no MP needed with iSCSI single path.
    if (($self->{mp_type} eq "none" or $self->{mp_type} eq "") and
         $self->{protocol} =~ /fcp/i) {
         $self->{mp_type} = "mpio";
    }

    # obtain memory information for this machine.
    $self->_obtain_memory_info();
    if ($self->error()) {
        # error is already set.
        return $self;
    }

    return $self;
}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            if (not defined ($cfmode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    # if mp_type is not set, do not validate currently
    my $mp_type = $self->{mp_type};
    if (not defined($mp_type) or $mp_type eq "") {
        $self->sanlog(-msg => "Multipathing type (mp_type) is not set. Not validating configuration");
        $self->{valid} = 1;
        return $ClientUtil::TRUE;
    }

    # get the MP stack info
    my $veritas_version    = "";
    my $veritas_version_sp = "";
    if ($mp_type =~ /mpio/i) {
        $mp_type = "mpio";
    } elsif ($mp_type =~ /sanpath/i) {
        $mp_type = "sanpath";
    } elsif ($mp_type =~ /dmp|vxvm|vrts|veritas/i) {
        $mp_type = "veritasdmp";
        # Get the MP object and get the veritas version from it.
        if (exists($self->{mp}) and ref($self->{mp}) =~ /dmp/i) {
            $veritas_version = $self->{mp}->{version};

            # See if we have an SP installed
            if (defined($self->{mp}->{sp}) and $self->{mp}->{sp} ne "") {
                $veritas_version_sp = $self->{mp}->{sp};
                $veritas_version_sp =~ s/SP//g;
            } else {
                $veritas_version_sp = 0;
            }
        } else {
            $self->_set_err("MP object either doesn't exists or is not a DMP object");
            $self->{valid} = 0;
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Unknown mp_type ($mp_type) for AIX. Failing validation");
        $self->{valid} = 0;
        return $ClientUtil::FALSE;
    }

    # Make sure the the restore daemon is running for veritas configs only
    if ($mp_type eq "veritasdmp") {
        if ($veritas_version >= 5.1) {
            # We want to look at the vxdmpadm gettune all output
            my $cmd  = "/usr/sbin/vxdmpadm gettune all";
            my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }

            # Now let's check for some settings
            # vxdmpadm gettune all
            # Tunable                           Current Value  Default Value
            # ------------------------------    -------------  -------------
            # dmp_restore_state                   enabled          enabled
            # dmp_fast_recovery                        on               on
            # dmp_log_level                             4                1
            # dmp_lun_retry_timeout                     0                0
            # dmp_path_age                            120              300
            # dmp_probe_idle_lun                      off               on
            # dmp_restore_interval                     60              300
            # dmp_restore_policy             check_disabled   check_disabled
            # dmp_monitor_fabric                      off               on
            my @errors_found = ();
            foreach my $line (@output) {
                if ($line =~ /^dmp_restore_state\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_restore_state'") if ($1 ne "enabled");
                } elsif ($line =~ /^dmp_fast_recovery\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_fast_recovery'") if ($1 ne "on");
                } elsif ($line =~ /^dmp_log_level\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_log_level'") if ($1 != 4);
                } elsif ($line =~ /^dmp_lun_retry_timeout\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_lun_retry_timeout'") if ($1 != 0);
                } elsif ($line =~ /^dmp_resore_interval\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_resore_interval'") if ($1 != 60);
                } elsif ($line =~ /^dmp_restore_policy\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_restore_policy'") if ($1 ne "check_disabled");
                } elsif ($line =~ /^dmp_probe_idle_lun\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_probe_idle_lun'") if ($1 ne "off");
                } elsif ($line =~ /^dmp_monitor_fabric\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_monitor_fabric'") if ($1 ne "off");
                }
            }

            # Now check out the settings that are 5.1SP1 or later
            if ($veritas_version >= 6.0 or ($veritas_version == 5.1 and $veritas_version_sp > 0)) {
                # Tunable                           Current Value  Default Value
                # ------------------------------    -------------  -------------
                # dmp_path_age                            120              300
                foreach my $line (@output) {
                    if ($line =~ /^dmp_path_age\s+(\S+)\s+/) {
                        push(@errors_found, "Invalid value '$1' for 'dmp_path_age'") if ($1 != 120);
                    }
                }
            }

            if (scalar(@errors_found)) {
                my $errors = join("\n", @errors_found);
                $self->sanlog(-msg => $errors);
                $self->_set_err("There were one or more tuneables set improperly, please verify the " .
                                "settings in 'vxdmpadm gettune all'");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }
        } else {
            # Run vxdmpadm stat restored
            my $cmd  = "/usr/sbin/vxdmpadm stat restored";
            my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }

            # We want to check the output and look for the following:
            # "The number of daemons running : 1".  The number should not be 0.
            if (grep(/daemons running\s+\:\s+0$/, @output) > 0) {
                $self->_set_err("Veritas Restore Daemon does not appear to be running");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->{valid} = 1;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: configure
# Objective: configure the AIX machine for either iSCSI or FCP access
# Details: See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args) = @_;

    my ($cfo_mode, @targets);
    my ($cmd, $status, @output);
    my (@iscsi_targets);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for -targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?cfo_mode$/) {
            if (not defined ($cfo_mode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?iscsi_targets$/) {
            if (not (@iscsi_targets = ${args{$_}})) {
                $self->_set_err("Invalid value for " . "-iscsi_targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Configure POSIX AIO (only enable for 5x versions of OS)
    if ($self->{revision} =~ /^5/) {
        $self->sanlog(-msg => "Enabling Legacy AIO");
        $cmd = "mkdev -l aio0";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # determine if mkdev command worked
        if ($status > 0) {
            $self->_set_err("Error: AixHost::configure() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        $cmd = "chdev -l aio0 -a autoconfig='available' -P";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: AixHost::configure() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "Enabling POSIX AIO");
        $cmd = "mkdev -l posix_aio0";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # determine if mkdev command worked
        if ($status > 0) {
            $self->_set_err("Error: AixHost::configure() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        $cmd = "chdev -l posix_aio0 -a autoconfig='available' -P";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: AixHost::configure() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
    }

    # configure for protocol iscsi
    if ($self->{protocol} =~ /iscsi/i) {
        # Need to add discovery address to iscsiadm
        if (scalar(@iscsi_targets) == 0) {
            $self->_iscsi_configure(-targets => @targets);
        } else {
            $self->_iscsi_configure(-targets => @iscsi_targets);
        }

        if ($self->error) {
            $self->sanlog(-msg => "Error: AixHost::configure() SanClient iSCSI configuration failed");
            return $ClientUtil::FALSE;
        }
    }

    # configure FCP
    if ($self->{protocol} =~ /fcp/i) {
        $self->_fcp_configure();

        if ($self->error) {
            $self->sanlog(-msg => "Error: AixHost::configure() SanClient FCP configuration failed");
            return $ClientUtil::FALSE;
        }
    }

    # Configure the MP Object
    if (defined $self->{mp} and $self->{mp_type} !~ /none/i) {
        $self->_MP_configure();
        return $ClientUtil::FALSE if ($self->error);
    }

    $self->sanlog(-msg => "AIX SanClient configuration has completed");

    $self->{configured} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: deconfigure
# Objective: deconfigure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub deconfigure {
    my ($self, %args) = @_;

    my (@targets);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not (@targets = ${args{$_}})) {
                $self->_set_err("Error: AixHost::deconfigure() Invalid value for " . "-targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # call appropriate deconfigure methods
    if ($self->{protocol} =~ /fcp/i) {
        $self->_fcp_deconfigure();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    } elsif ($self->{protocol} =~ /iscsi/i) {
        $self->_iscsi_deconfigure(-targets => @targets);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

}

################################################################################
# Subroutine Name  gather_data
# Objective:       Gather the data on the client. This data will mostly be
#                  used for debug purposes. Currently no-op
# Details:         See POD documentation at the end of this file
################################################################################
sub gather_data {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
 ################################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: collect_sysrq_dump_periodic
# Objective:       N/A for Aix, provide for compatibility with Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_sysrq_dump_periodic {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: discover
# Objective: Discover devices and label if new.
# Details: See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self) = @_;

    my ($cmd, @devices, $status, @output, @hDevices);

    $self->sanlog(-msg => "Discovering all devices");

    # execute cfgmgr
    $cmd = "cfgmgr";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::discover() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Locate devices that are not part of a volume group",
                  -msg_level => 3);

    @devices = $self->_get_unlabeled_devices();
    $self->sanlog(-msg => scalar(@devices) .
                          " Devices are not part of a volume group",
                  -msg_level => 3);

    # execute _config_paths which on non-ALUA configs with
    # MPIO will run dotpaths utility
    $status = $self->_config_paths();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: format
# Objective:       N/A for AIX
# Details:         See POD documentation at the end of this file
################################################################################
sub format {
    my ($self, %args) = @_;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: sync()
# Objective:       flush buffer cache
# Details:         See POD documentation at the end of this file
################################################################################
sub sync {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    $cmd = "sync";

    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::sync() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: get_mounted_device
# Objective:       returns a mounted device name when passed a filesystem
###############################################################################
sub get_mounted_device {
    my ($self, $mount) = @_;

    my ($cmd, $line, $status, @output);

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount =~ s/\//\\\//g;

    # execute the command
    $cmd = "mount | grep $mount | awk '{print \$1}'";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 or scalar(@output) == 0) {
        $self->_set_err("Error: AIXHost::get_mounted_device(): Mount command failed!",
                        @output);
        return undef;
    }

    if (scalar(@output) > 1) {
        $self->_set_err("Error: AIXHost::get_mounted_device(): Multiple mount entries found!",
                        @output);
        return undef;
    }

    return $output[0];

}

###############################################################################
# Method:    get_lvm_default
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lvm_default {
    my ($self) = @_;

    $self->_clear_err();

    return qw(nativelvm);

}

################################################################################
# Subroutine Name: sanlun_version()
# Objective:       get sanlun version
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;

    my ($cmd, @output, $status, $version);

    $self->_clear_err();

##### Need to make this work with rebranded Kit!!!

    $cmd = "lslpp -l NetApp* | grep 'NetApp.SAN_toolkit.sanlun' | awk '{print \$2}'";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::sanlun_version() Command ($cmd) returned a bad status ($status)");
        return "";
    }

    unless (defined $output[0] and length $output[0]) {
        $self->_set_err("Error: AixHost::sanlun_version(): Unable to get sanlun version");
        return "";
    }

    $version = $output[0];

    return $version;

}

###############################################################################
# Method:    info
# Objective: Get client information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;

    my ($hsh, $protocol, @output);

    $protocol = $self->get_protocol();

    push @output,
         "Host name:      " . $self->hostname(),
         "Host type:      " . $self->hosttype(),
         "OS version:     " . $self->os_version();

    push @output,
         "Sanlun version: " . $self->sanlun_version()
            if ($protocol =~ /fcp|iscsi/i);

    push @output, "Protocol:       " . $protocol, "\n";


    return (wantarray()) ? @output: \@output;

}

################################################################################
# Subroutine Name: iostat
# Objective:       Format devices with a label
# Details:         See POD documentation at the end of this file
################################################################################
sub iostat {
    my ($self, %args) = @_;

    my ($interval, $cmd, $status, @output, $cnt, $count, @buf);

    $self->_clear_err();

    # Default interval is 2 seconds
    $interval = 2;

    # Default count is 1
    $count = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?interval/) {
            if (not length($interval = $args{$_})) {
                $self->_set_err("Error: AixHost::iostat() Invalid value for argument '-interval'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?count/) {
            if (not length($count = $args{$_})) {
                $self->_set_err("Error: AixHost::iostat() Missing value for argument '-count'");
                return (wantarray()) ? () : [];
            }
            next;
        } else {
            $self->_set_err("Error: AixHost::iostat() Invalid parameter specified \'$_\'");
            return (wantarray()) ? () : [];
        }
    }

    # Remember that the first iostat cycle is wrong so add an extra cycle.
    # Minimize the output return since 'z' flag doesn't kick in until second cycle.
    $count += 1;
    $cmd = "iostat $interval $count | grep -v '0.0       0.0       0.0          0         0'";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::iostat() ",
                        "iostat returned a bad status ($status))");

        return (wantarray()) ? () : [];
    }

    # Clean up the output by removing the first iostat cycle

    $cnt = 0;
    foreach my $line (@output) {
        if ($line =~ /disk history|tty:/i) {
            $cnt++;
            next;
        }

        if ($cnt == 1) {
            push (@buf, $line);
        }
    }

    push (@buf, "\n\n");

    return (wantarray()) ? @buf : \@buf;

}

################################################################################
# Subroutine Name: get_nfs_mounts
# Objective:       get NFS mounts currently mounted on the system
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_mounts {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    my $targets = undef;
    my @nfs_mounts;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not length($targets = $args{$_})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($targets)) {
        $self->_set_err("Missing required argument '-targets'");
        return (wantarray()) ? () : [];
    }

    $cmd = "/usr/sbin/mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to execute command: " . $cmd);
        return (wantarray()) ? () : [];
    }

    # Capture the output of mount for each filer
    # and parse out the shares that are assigned to $self.
    foreach my $target (@{$targets}) {
        # Look through each line of output and determint if there is a filer in the $targets list mounted to this host
        my @mounts = grep(/$target/i, @output);
        if (scalar(@mounts) == 0) {
            next;
        }

        # Go through the found mounts and populate the return array
        foreach my $mount (@mounts) {
            # fas2050-rtp03 /vol/nate_1_1_p615_aix03/nate_qtree_p615-aix03_aix_0 /mnt/qa/fas2050-rtp03/nate_qtree_p615-aix03_aix_0 nfs3
            my ($target, $nfs_share, $nfs_mount, $nfs_options) = split(/\s+/, $mount);
            $nfs_share = $target . ":" . $nfs_share;

            my $nfs_pt;
            $nfs_pt->{mount} = $nfs_mount;
            $nfs_pt->{share} = $nfs_share;
            $nfs_pt->{options} = $nfs_options;

            push(@nfs_mounts, $nfs_pt);
        }
    }

    if (scalar(@nfs_mounts) == 0) {
        $self->_set_err("Did not find any mounts associated with targets (" . join(" ", @{$targets}) . ")");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @nfs_mounts : \@nfs_mounts;
}

################################################################################
# Subroutine Name: get_nfs_share
# Objective:       get available NFS shares volumes via showmount -e
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_share {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    my $targets = undef;
    my $interfaces = undef;
    my @nfs_shares;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not length($targets = $args{$_})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?interfaces/) {
            if (not length($interfaces = $args{$_})) {
                $self->_set_err("Invalid value for argument '-interfaces'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($targets)) {
        $self->_set_err("Missing required argument '-targets'");
        return (wantarray()) ? () : [];
    }

    # If the interface list is not passed in, get it
    if (not defined($interfaces)) {
        $interfaces = $self->get_ip_info();
    }

    # Capture the output of showmount -e for each filer
    # and parse out the shares that are assigned to $self.
    foreach my $target (@{$targets}) {
        # Collect showmount -e
        $cmd = "/usr/bin/showmount -e " . $target;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # Look for the two following erros, they failures.
        # showmount: <device_name>: RPC: Program not registered
        # showmount: <device_name>: RPC: Unknown host
        if (grep(/showmount\:.+RPC\:/, @output)) {
            $self->_set_err("RPC error occurred during command: " . $cmd);
            return (wantarray()) ? () : [];
        }

        if ($status > 0) {
            $self->_set_err("Unable to execute command: " . $cmd);
            return (wantarray()) ? () : [];
        }

        # Grep through the output and only get the ones that have an ip address in the interface list
        foreach my $interface (keys(%{$interfaces})) {
            my $ipv4 = $interfaces->{$interface}->{ipv4};
            my $ipv6 = $interfaces->{$interface}->{ipv6};
            my @results = ();

            # If we have an ipv4 address specified, add it to the results array
            if (defined($ipv4) and $ipv4 =~ /\d+\.\d+\.\d+\.\d+/) {
                push(@results, grep(/$ipv4/i, @output));
            }

            # If we have an ipv6 address specified, add it to the results array
            if (defined($ipv6)) {
                push(@results, grep(/$ipv6/i, @output));
            }

            if (scalar(@results) == 0) {
                $self->sanlog(-msg => "Did not find any shares for interface: $ipv4 $ipv6");
                next;
            }

            # Parse the output and populate nfs_mounts
            foreach my $line (@results) {
                # /vol/nate_1_1_hp2620_rtp11   10.60.242.103
                # /vol/nate_1_1_hprx2600_rtp14 10.60.242.105
                my ($vol) = split(/\s+/, $line);

                my $nfs_share;
                $nfs_share->{volume} = basename($vol);
                $nfs_share->{path} = $vol;
                $nfs_share->{ipv4_address} = $ipv4;
                $nfs_share->{ipv6_address} = $ipv6;
                $nfs_share->{target} = $target;

                push(@nfs_shares, $nfs_share);
            }
        }
    }

    # See if no NFS shares were found
    if (scalar(@nfs_shares) == 0) {
        $self->_set_err("No NFS shares found for targets (" . join(" ", @{$targets}) . ")");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @nfs_shares : \@nfs_shares;
}

################################################################################
# Subroutine Name: nfs_mount
# Objective:       NFS mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_mount {
    my ($self, %args) = @_;

    my ($cmd, $directio, $hide, $resource, $mount_pt, $options);
    my ($nodename, $vol_name, $name, $status, @output);

    $self->_clear_err();
    $directio = 0;
    $options = "";
    $hide = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?resource/) {
            if (not length($resource = $args{$_})) {
                $self->_set_err("Error: AIXHost::nfs_mount() Invalid value for argument '-resource'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?directio/) {
            if (not length($directio = $args{$_} or $directio !~ /1|0/)) {
                $self->_set_err("Error: AixHost::nfs_mount() Missing value for argument '-directio'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Error: AIXHost::nfs_mount() Missing value for argument '-mount_pt'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?options/) {
            if (not length($options = $args{$_})) {
                $self->_set_err("Error: AIXHost::nfs_mount() Missing value for argument '-options'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }


    # Check that required parameter '-resource' was provided
    unless (defined $resource and length $resource) {
        $self->_set_err("Error: AIXHost::nfs_mount() ",
                        "Missing required argument '-resource'");
        return $ClientUtil::FALSE;
    }

    #  Check that required parameter '-mount_pt'
    unless (defined $mount_pt and length $mount_pt) {
        #  Use default path if not defined
        ($nodename = $resource) =~ s/(^\S+):\S+/$1/;

        # Check for empty value
        if ($nodename eq "") {
            $self->_set_err("Error: AIXHost::nfs_mount() Unable to determine filer name from device mount '$resource'",
                            @output);
            return $ClientUtil::FALSE;
        }

        $vol_name = basename($resource);
        $mount_pt = "/mnt/qa/" . $nodename . "/" . $vol_name;
        $self->sanlog(-msg => "Using default mount pt '$mount_pt' for $resource");
    }

    # check if the volume is already mounted
    $self->mount_table_verify($mount_pt);
    if ($self->error) {
        $self->_clear_err();

        # Create the mount point and open up dir privs
        $cmd = "mkdir -p $mount_pt; chmod -R 777 $mount_pt";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->sanlog(-msg => "WARNING: AIXHost::nfs_mount() Unable to create mount " .
                                  "point directory for volume '$resource'",
                                  @output);
        }

        # Add direction option if requested and it was not
        # passed in as part of the options string
        if ($directio and $options !~ /dio/i) {
            $options .= (defined $options and length $options)
                ? ",dio"
                : "dio";
        }

        # Remove option param if user passed it in as part of the
        # options string. Lib should add the tag.
        $options =~ s/-o //g;

        # Create the NFS mount command.
        # mount -o rw,hard,intr  gin:/vol/vol1/ /mnt
        if (defined($options) and $options ne "") {
            $cmd = "/usr/sbin/mount -o $options $resource $mount_pt";
        } else {
            $cmd = "/usr/sbin/mount $resource $mount_pt";
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: AIXHost::nfs_mount() Unable to NFS mount volume",
                            @output);
            return $ClientUtil::FALSE;
        }
    } else {
        $self->sanlog(-msg => "Warning: AIXHost::nfs_mount() volume ($mount_pt) is already mounted");
        $self->_clear_err();
        return $ClientUtil::TRUE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: nfs_unmount()
# Objective:       unmount NFS volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_unmount {
    my ($self, %args) = @_;

    my($cmd, $hide, $mount_pt, $nfs_mount, $status, @output);

    $self->_clear_err();

    $hide = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?mount_pt/) {
            # Make sure mount point exists and is not '/'.
            if (not length($mount_pt = $args{$_}) and ($args{$_}!= "/")) {
                $self->_set_err("Error: AixHost::nfs_unmount() Missing value for argument '-mount_pt'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Error: AixHost::nfs_unmount() Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }

    # Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt) {
        $self->_set_err("Error: AixHost::nfs_unmount() ",
                        "Missing required argument '-mount_pt'");
        return $ClientUtil::FALSE;
    }

    $cmd = "umount " . $mount_pt;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AIXHost::nfs_unmount() ",
                        "Unable to unmount NFS volume " . $mount_pt . " ",
                        @output);
        return $ClientUtil::FALSE;
    }

    # Create the command to remove the NFS mount point
    if ($mount_pt ne "/") {
        $cmd = "rm -Rf " . $mount_pt;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Warning: AIXHost::nfs_unmount() Unable to remove ",
                            "mount point directory '" . $mount_pt . "'",
                            @output);

            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: mount_entry_add()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_add {
    my ($self, %args) = @_;

    my ($cmd, $device_mount, $mount_pt, $fs_type, $mount_options);
    my ($nodename, $line, $status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?device_mount/) {
            if (not length($device_mount = $args{$_})) {
                $self->_set_err("Error: AIXHostHost::mount_entry_add() Invalid value for argument '-device_mount'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Error: AIXHost::mount_entry_add() Missing value for argument '-mount_pt'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?fs_type/) {
            if (not length($fs_type = $args{$_})) {
                $self->_set_err("Error: AIXHost::mount_entry_add() Missing value for argument '-fs_type'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?mount_options/) {
            if (not length($mount_options = $args{$_})) {
                $self->_set_err("Error: AIXHost::mount_entry_add() Missing value for argument '-mount_options'");
                return $ClientUtil::FALSE;
            }
            next;
        } else {
            $self->_set_err("Error: AIXHost::mount_entry_add() Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }

    # Check that required parameter '-device' was provided
    unless (defined $device_mount and length $device_mount) {
        $self->_set_err("Error: AIXHost::mount_entry_add() ",
                        "Missing required argument '-device_mount'");
        return $ClientUtil::TRUE;
    }

    if ($fs_type =~ /^nfs$/i) {
        ($nodename = $device_mount) =~ s#(^\S+):\S+#$1#;
        $device_mount =~ s#^\S+:##;
    }

    # Check that required parameter '-device' was provided
    unless (defined $mount_pt and length $mount_pt) {
        $self->_set_err("Error: AIXHost::mount_entry_add() ",
                        "Missing required argument '-mount_pt'");
        return $ClientUtil::TRUE;
    }

    # Check that required parameter '-device' was provided
    unless (defined $fs_type and $fs_type =~ /^jfs$|^jfs2$|^nfs$/i) {
        $self->_set_err("Error: AIXHost::mount_entry_add() ",
                        "Missing required argument '-fs_type'");
        return $ClientUtil::TRUE;
    }

    # Check that required parameter '-device' was provided
    unless (defined $mount_options and length $mount_options) {
        $self->_set_err("Error: AIXHost::mount_entry_add() ",
                        "Missing required argument '-mount_options'");
        return $ClientUtil::TRUE;
    }

    if ($fs_type =~ /^jfs$|^jfs2$/i) {
        #  /mnt/qa/dg1/vol25:
        #      dev             = /dev/dg1_vol25
        #      vfs             = jfs
        #      log             = /dev/dg1_log
        #      mount           = false
        #      check           = false

            $line = sprintf("\n%s:\n%8.8s%-16.16s= %s\n%8.8s%-16.16s= %s\n%8.8s%-16.16s= %s\n%8.8s%-16.16s= %s\n%8.8s%-16.16s= %s",
                      $mount_pt,
                      "", "dev", $device_mount,
                      "", "vfs", $fs_type,
                      "", "log", $mount_options,
                      "", "mount", "false",
                      "", "check", "false");
    } elsif ($fs_type =~ /^nfs$/i) {
        # /usr/local:
        #        dev             = "/vol/local/powerpc-ibm-aix5"
        #        vfs             = nfs
        #        nodename        = orca7.nane.netapp.com
        #        mount           = true
        #        type            = nfs
        #        options         = ro,bg,hard,intr
        #        account         = false

        $line = sprintf("\n%s:\n%8.8s%-16.16s= %s\n%8.8s%-16.16s= %s\n%8.8s%-16.16s= %s\n%8.8s%-16.16s= %s\n%8.8s%-16.16s= %s\n%8.8s%-16.16s= %s\n%8.8s%-16.16s= %s",
                      $mount_pt,
                      "", "dev",'"' . $device_mount . '"',
                      "", "vfs", $fs_type,
                      "", "nodename", $nodename,
                      "", "mount", "true",
                      "", "type", "nfs",
                      "", "options", $mount_options,
                      "", "account", "false");

    }


    $self->sanlog(-msg => "Adding ($fs_type) fs entry to mount table ($mount_pt)");

    $cmd = "echo '" . $line . "' >> /etc/filesystems";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AIXHost::mount_entry_add Unable to add mount entry to /etc/filesystems file",
                        @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, $mntpt) = @_;

    my ($cmd, $hide, $mount, $status, @output);

    $self->_clear_err();

    $hide = 0 ;

    # Unmount the volume's filesystem
    $cmd = "umount $mntpt";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error:AixHost ::unmount_fs() Unable to unmount filesystem",
                        @output);
    }

    $self->mount_entry_delete($mntpt);

    $self->sanlog(-msg => "Removing mount point '$mntpt'");

    # Remove directory head
    if ($self->{mount} ne "/") {
        $cmd = "rm -rf $self->{mount}";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: AixHost::unmount_fs() Unable to remove mount point '$mount'",
                            @output);
        }
    } else {
       $self->_set_err("Error: AixHost::unmount_fs() Mount Point not apporpriate for removal ($self->{mount})");
    }

    return $self;

}

################################################################################
# Subroutine Name: mount_entry_delete()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_delete {
    my ($self, $mount) = @_;

    my ($cmd, $extra, $status, @output);

    $self->_clear_err();

    # Clear /etc/filesystem mount entry
    $self->sanlog(-msg => "Clear /etc/filesystems mount entry");

    unless (defined $mount and length($mount)) {
        $self->_set_err("Error: AIXHost::mount_entry_delete Missing required paremeter '$mount'",
                        @output);
        return $ClientUtil::FALSE;
    }

    # BURT 162373: The '$extra' variable is to remove an invalid second mount entry created
    # by the Snap Connect' command.  This should be removed after bug is fixed.
    $extra = "/fs/";
    $extra =~ s/\//\\\//g;
    $mount =~ s/\//\\\//g;

    if ($mount =~ /^\\\/dev/i) {
        # Search and delete extra mount entry created by 'snap connect'
        # bug as well ($mount and /fs/$mount).
        $cmd = "perl -e\'".
            "open(FH,\"< \/etc\/filesystems\") or exit(1);".
                "open(FD,\"> \/etc\/filesystems_new\") or exit(1);".
                    "\$bFound=0;".
                        "\@buf=<FH>;".
                            "close FH;".
                                "for (\$i=0;\$i<scalar(\@buf);\$i++){".
                                    "  if (\$buf[\$i+1] =~ /$mount/){".
                                        "  print \"FOUND: \$buf[\$i]\n\";".
                                            "    \$bFound=1;".
                                                "    next;".
                                                    "  }".
                                                        "  \$bFound=0 if (\$buf[\$i]=~/:/);".
                                                            "  print FD \$buf[\$i] if (\$bFound==0);".
                                                                "}".
                                                                    "close FD;".
                                                                        "rename(\"\/etc\/filesystems_new\", \"\/etc\/filesystems\") or exit(1)".
                                                                            "\' ";

    } else {

        # Search and delete extra mount entry created by 'snap connect'
        # bug as well ($mount and /fs/$mount).
        $cmd = "perl -e\'".
            "open(FH,\"< \/etc\/filesystems\") or exit(1);".
                "open(FD,\"> \/etc\/filesystems_new\") or exit(1);".
                    "\$bFound=0;".
                        "while(<FH>){".
                            "  if (\$_ =~ /\^$mount:\$|\^$extra/){".
                                "  print \"FOUND: \$_\n\";".
                                    "    \$bFound=1;".
                                        "    next;".
                                            "  }".
                                                "  \$bFound=0 if (\$_=~/:/);".
                                                    "  print FD \$_ if (\$bFound==0);".
                                                        "}\\\n".
                                                            "close FD;".
                                                                "close FH;".
                                                                    "rename(\"\/etc\/filesystems_new\", \"\/etc\/filesystems\") or exit(1)".
                                                                        "\'";

    }

    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AIXHost::mount_entry_delete Unable to remove mount entry from /etc/fstab file",
                        @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: mount_entry_verify
# Objective:      verify mount entries
###############################################################################
sub mount_entry_verify {
    my ($self, $mount) = @_;

    my ($cmd, $entry, $idx, $status, $value, @values);
    my (@params, @output, @buf);

    $self->_clear_err();

    # initialize
    $entry = "";
    $idx = 0;

    $mount =~ s/\//\\\//g;

    $cmd = "perl -e\'".
        "open(FH,\"< \/etc\/filesystems\") or exit(1);".
            "\$bFound=0;".
                "while(<FH>){".
                    "  if (\$_ =~ /$mount:/){".
                        "  print \"\$_\";".
                            "    \$bFound=1;".
                                "    next;".
                                    "  }".
                                        "  \$bFound=0 if (\$_=~/:/);".
                                            "  print \$_ if (\$bFound==1);".
                                                "}\\\n".
                                                    "close FH;".
                                                        "\'";



    # execute the command
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    # remove line continuation prompt
    $output[0] =~ s/\> //;
    if ($status > 0 or scalar(@output) == 0) {
        $self->_set_err("Error: AIXHost::mount_entry_verify(): Unable to verify mount file entry",
                        @output);
        return $entry;
    }

    # Need to parse the mount into some useful structure.  For now just show
    # that something was found.

    return (wantarray()) ? @output : \@output;

}

###############################################################################
# Subroutine Name: mount_table_verify
# Objective:       verify mount table
###############################################################################
sub mount_table_verify {
    my ($self, $mount) = @_;

    my ($status, @output);

    $self->_clear_err();

    if ($mount =~ /\/$/) {
        chop($mount);
    }

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, "mount");
    if ($status > 0 or scalar(@output) == 0) {
        $self->_set_err("Error: AIXHost::mount_table_verify(): Unable to verify mount table",
                        @output);

        return (wantarray()) ? @output : \@output;
    }

    unless (grep /$mount/, @output) {
        $self->_set_err("Error: AIXHost::mount_table_verify(): Unable to verify mount table - mount ($mount) not found",
                        @output);
        return (wantarray()) ? @output : \@output;
    }

    # Need to parse the mount into some useful structure.
    # For now just show that something was found.

    return (wantarray()) ? @output : \@output;

}

###############################################################################
# Subroutine Name: mount_exists
# Objective:       validates a mount point is correctly mounted and does exist
###############################################################################
sub mount_exists {
    my ($self, $mount) = @_;

    $self->_clear_err();

    my ($cmd, $status, @output);

    $cmd = "/usr/sbin/mount | grep $mount";
    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 30);

    # Check the status of the mount command
    if ($status != 0 or scalar(@output) == 0) {
        $self->_set_err("Error: AixHost::mount_exists(): Unable to execute command ($cmd)");
        return $ClientUtil::FALSE;
    }

    # Join all lines in output with spaces then split on all spaces.
    # This puts all pieces as elements of the output array.
    @output = split(/\s+/, join(" ", @output));

    # Look for the mount in the output, if it exists return true
    if (grep(/^$mount$/, @output)) {
        return $ClientUtil::TRUE;
    }

    # Could not find the mount
    $self->_set_err("Error: AixHost::mount_exists(): Unable to locate mount point ($mount)");
    return $ClientUtil::FALSE;
}

###############################################################################
# Method:    get_supported_protocols
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_protocols {
    my ($self) = @_;

    $self->_clear_err();

    return qw(fcp iscsi nfs);

}

###############################################################################
# Method:    get_supported_fs_types
# Objective: Get list of supported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_fs_types {
    my ($self) = @_;

    return qw(jfs2 jfs vxfs);

}

################################################################################
# Subroutine Name: config_paths
# Objective:       Need for compat with HP-UX
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self) = @_;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: raw_device_setup
# Objective:       Do whatever we need to do to use raw devices successfully
#                  after the call to discover().
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_setup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       Do whatever we need to do to make sure the host is more
#                  happy when luns are removed.  For linux, we want to do
#                  an iscsi stop.
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For AixHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       When the luns are removed from the filer, we will have
#                  orphaned luns.  This causes big problems.
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;

    my ($cmd, $device, $status, $hide, @output);
    my ($bootdevice, @disk_info);
    my ($b_output, @b_output, @rm_output, @vx_output, @c_output);

    $self->_clear_err();
    $hide = 0;

    # remove spd (sanpath) devices from conf file
    if ($self->{mp_type} =~ /sanpath/i) {
        $cmd = "cat /dev/null > /etc/spd.conf";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: AixHost::orphaned_device_cleanup() Command ($cmd) returned a bad status ($status)");
        }
    }

    # execute cfgmgr
    $cmd = "cfgmgr";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::orphaned_device_cleanup() Command ($cmd) returned a bad status ($status)");
    }

    # find all instances devices
    if ($self->{mp_type} =~ /mpio/i) {
        if ($self->get_protocol() =~ /fcp/i) {
            $cmd = "lsdev -C -c disk | egrep \"\(MPIO)\|\(Virtual\)\" | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\) FCP\" | awk '{print \$1}'";
        } else {
            $cmd = "lsdev -C -c disk | egrep \"\(MPIO)\|\(Virtual\)\" | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\) iSCSI\" | awk '{print \$1}'";
        }
    } elsif ($self->{mp_type} =~ /sanpath/i) {
        # find all instances of SanPath devices
        $cmd = "lsdev -C | grep spd | grep -v ctrl | awk '{print  \$1}'";
    } else {
        $cmd = "lsdev -C -c disk | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\)\" | awk '{print \$1}'";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # perform rmdisk on left over VRTS devices (or rmdev on
    # hdisk devcies will fail)
    if ($self->{mp_type} =~ /dmp|vxvm/i) {
        $self->sanlog(-msg => "removing orphaned vxdisk devices");

        # HUP VRTS, sometimes not all devices are visible
        $cmd = "/usr/sbin/vxdctl enable";
        ($status, @vx_output) = &ClientUtil::execute($self->{conn}, $cmd);

        $cmd = "/usr/sbin/vxdisk list";
        ($status, @vx_output) = &ClientUtil::execute($self->{conn}, $cmd);

        foreach my $line (@vx_output) {
            if ($line =~  /^DEVICE/ or $line =~ /^Disk_/) {
                next;
            } else {
                @disk_info = split(/\s+/, $line);
                my $dev = $disk_info[0];
                # if group and disk are "-", no DG's using this.
                # also check to make sure device is not "-" which could
                # be the case in faulted state.
                if ($disk_info[2] eq "-" and $disk_info[3] eq "-" and
                    $dev ne "-") {
                    $cmd = "/usr/sbin/vxdisk rm $dev";
                    ($status, @rm_output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status > 0) {
                        $self->_set_err("Error: AixHost::orphaned_device_cleanup() Command ($cmd) returned a bad status ($status)");
                    }
                }
            }
        }
    }

    # get the boot device - do not attempt to rmdev this - could
    # be sanbooted netapp device.
    $cmd = "bootinfo -b";
    ($status, @b_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::orphaned_device_cleanup() Command ($cmd) returned a bad status ($status)");
    }
    $bootdevice = $b_output[0];

    # just ignore status and plow thru all devices (ignore boot device)
    foreach $device (@output) {
        # ignore bootdevice
        if ($device eq $bootdevice) {
             next;
        }

        # remove all instances of MPIO devices
        $cmd = "rmdev -dl $device";
        ($status, @c_output) = &ClientUtil::execute($self->{conn}, $cmd);
    }

    # find all instances of hdisk devices not part of a vg (rootvg)
    $cmd = "lspv | grep -i none | awk '{print  \$1}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::orphaned_device_cleanup() Command ($cmd) returned a bad status ($status)");
    }

    # just ignore status and plow thru all devices
    foreach $device (@output) {
        # ignore bootdevice
        if ($device eq $bootdevice) {
             next;
        }

        # remove all instances of found hdisk devices
        $cmd = "rmdev -dl $device";
        ($status, @c_output) = &ClientUtil::execute($self->{conn}, $cmd);
    }

    # Reset vxconfigd to release the adapters
    if ($self->{mp_type} =~ /dmp|vxvm/i) {
        $cmd = "/usr/sbin/vxconfigd -kr reset";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: shutdown()
# Objective:       shutdown host
# Details:         See POD documentation at the end of this file
################################################################################
sub shutdown {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "/usr/sbin/shutdown -F now";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    # sleep 1 minute to make sure shutdown occurs
    &SanUtil::nate_sleep(60);

    # Invalidate the connection on the object
    &Tharn::disconnect($self->{conn}, '-ok');

    # Ping the host every 10 seconds until it has completely shutdown
    my $alive = 1;

    for (my $cnt = 0; $cnt < 36; $cnt++) {
        my $status = &SanUtil::ping_host($self->hostname());
        if ($status == $ClientUtil::FALSE) {
            $self->sanlog(-msg => "OS is shutdown");
            $alive = 0;
            last;
        } else {
            $self->sanlog(-msg => "Ping was successful",
                          -msg_level => 7);
            &SanUtil::nate_sleep(10);
        }
    }

    # Machine is still alive, shutdown did not work
    if ($alive) {
        $self->_set_err("OS never shutdown");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reboot()
# Objective:       reboot host
# Details:         See POD documentation at the end of this file
################################################################################
sub reboot {
    my ($self) = @_;
    my (
        $cmd,
        @output,
        $status,
        );

    $self->_clear_err();

    $cmd = "shutdown -F -r now";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');
    $self->sanlog(-msg => "$cmd (CMD=DONE)",
                  -msg_level => 7);

    foreach my $line (@output){
        $self->sanlog(-msg => "line=$line\n",
                      -indent    => 4,
                      -msg_level => 7);
    }

    # detect if the host actually rebooted or not
    $self->detect_reboot(-shutdown_timeout => 180,
                         -bootup_timeout   => 360,
                         -bootup_sleep     => 180);
    return $ClientUtil::FALSE if ($self->error);

    # sleep for a while to make sure the host is all the
    # way back up.  Sometimes ping will work before telnet
    # will work.
    &SanUtil::nate_sleep(120);

    # reconnect to the client
    $self->reconnect();
    return $ClientUtil::FALSE if ($self->error);

    # run uname on the client so we know the connection/client are happy
    ($status, @output) = &ClientUtil::execute($self->{conn}, "uname -a");
    if ($status > 0) {
        $self->_set_err("Error: AixHost::reboot() Command ($cmd) " .
                        "returned a bad status ($status)");
        return "";
    }

    my $tempmsg = "Uname output = ";
    $tempmsg .= join "\n", @output;
    $self->sanlog(-msg => $tempmsg);

    $self->{reboot} = 0;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: ntp_sync
# Objective:       Sync the host with the RTP NTP Time Server
#                  (timehost1.eng.netapp.com )
# Details:         See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my ($ntp_server, $ntp_time_zone);

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: AIXHost::ntp_sync() Invalid value for argument '-ntp_sync'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: AIXHost::ntp_sync() Invalid value for argument '-ntp_time_zone'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # do not check status of the stop since if it not already running this might
    # return non-zero
    $self->sanlog(-msg => "Stopping NTP time service");
    $cmd = "stopsrc -s xntpd";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    $self->sanlog(-msg => "Populating /etc/ntp.conf file");
    $cmd = "echo \"server $ntp_server\" > /etc/ntp.conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::ntp_sync Could not execute $cmd");
        return ($ClientUtil::FALSE);
    }

    $cmd = "echo \"driftfile /etc/ntp.drift\" >> /etc/ntp.conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::ntp_sync Could not execute $cmd");
        return ($ClientUtil::FALSE);
    }

    $cmd = "echo \"tracefile /etc/ntp.trace\" >> /etc/ntp.conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::ntp_sync Could not execute $cmd");
        return ($ClientUtil::FALSE);
    }

    $self->sanlog(-msg => "Setting the Time Zone (TZ) to $ntp_time_zone");
    $cmd = "TZ=$ntp_time_zone; export TZ";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::ntp_sync Could not execute $cmd",
                        @output);
        return ($ClientUtil::FALSE);
    }

    $cmd = "chtz $ntp_time_zone";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::ntp_sync Could not execute $cmd",
                        @output);
        return ($ClientUtil::FALSE);
    }

    $self->sanlog(-msg => "Running ntdate with $ntp_server");
    $cmd = "ntpdate $ntp_server";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        # If already managed by NTP, ignore.
        if (grep(/NTP socket is in use/i, @output) == 0) {
            $self->_set_err("Error: AixHost::ntp_sync Could not execute $cmd",
                            @output);
            return ($ClientUtil::FALSE);
        }
    }

    $self->sanlog(-msg => "Starting NTP time service");
    $cmd = "startsrc -s xntpd";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::ntp_sync Could not execute $cmd",
                        @output);
        return ($ClientUtil::FALSE);
    }

    $cmd = "date";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::ntp_sync Could not execute $cmd",
                        @output);
        return ($ClientUtil::FALSE);
    } else {
        $self->sanlog(-msg => "Local time is: $output[0]");
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: syslog_msg
# Objective:       Send message to AIX syslog
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg {
    my ($self, @message) = @_;

    my ($cmd, @date_output, @output, $status, $date);

    foreach my $msg (@message) {
        # get date on AIX system to match to system time (should be sync)
        # AIX team want the string w/date because of how errpt shows the
        # data.
        $cmd = "date '+%m/%d/%y - %H:%M:%S'";
        ($status, @date_output) = &ClientUtil::execute($self->{conn}, $cmd, 5);
        $date = $date_output[0];

        $cmd = "errlogger \"**NATE** ($date): $msg\"";
        ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd, 10);
        if ($status > 0) {
            $self->_set_err("Error: AixHost::syslog_msg(): Error running errlogger command ($cmd)");
            return $ClientUtil::FALSE;
        }
    }
}

################################################################################
# Method:    get_system_stats()
# Objective: Get the cpu/memory usage percentage over a specified interval
################################################################################
sub get_system_stats {
    my ($self, $interval) = @_;

    my $system_stats = {};
    my $cpu_samples = 0;
    my $mem_samples = 0;
    my $total_memory = 0;
    my $page_size = $self->{memory_page_size};
    my $num_cpus = 1;
    my $cpu_mem_flag = 0;

    # collect system statistics for $interval length,
    my $cmd = "/usr/bin/vmstat 1 $interval";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, ($interval + 120));
    if ($status > 1) {
        $self->_set_err("Error: AixHost::get_system_stats(): failed to collect system info.");
        return $ClientUtil::FALSE;
    }

    # @output could look like this:
    #kthr    memory              page              faults        cpu
    #----- ----------- ------------------------ ------------ -----------
    # r  b   avm   fre  re  pi  po  fr   sr  cy  in   sy  cs us sy id wa
    # 0  0 111878 363006   0   0   0   0    0   0   4   66  62  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0  10   10  58  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0   4   13  58  0  1 99  0
    # 0  0 111878 363006   0   0   0   0    0   0   4  180  86  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0   3    8  54  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0   3   11  60  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0   2    8  62  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0   2   15  62  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0  27    8  52  0  1 99  0
    # 0  0 111878 363006   0   0   0   0    0   0  26   10  58  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0  16    8  56  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0   2   11  58  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0   2   13  58  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0  12   10  66  0  0 99  0
    # 0  0 111878 363006   0   0   0   0    0   0   2    8  54  0  1 99  0

    # parse out all cpu and memory samples and tally them up
    foreach my $line (@output) {
        # skip unnessesary lines
        next if ($line =~ /^\s*$/);
        next if ($line =~ /--/);
        next if ($line =~ /faults/);
        next if ($line =~ /avm/);

        # parse out the total memory and num cpus
        if ($line =~ /cpu=(\d+)\s+mem=(\d+[KMGT]?)B/i) {
            $total_memory = &ClientUtil::convert_to_bytes($2);
            $num_cpus = $1;
        } elsif ($line =~ /^\s*\d+\s+\d+\s+\d+\s+/) {
            if ($cpu_mem_flag == 0) {
                # First reading of CPU and memory gives an average since the last reboot. So, we need to neglect this one.
                $cpu_mem_flag = 1;
                next;
            } else {
                # remove any spaces at the beginning of the string
                $line =~ s/^\s+//g;

                # split up all the numbers
                my @numbers = split(/\s+/, $line);

                # parse the free memory pages, 4th column
                my $mem_pages_free = int($numbers[3]);
                my $mem_free_percent = int(100 * (($mem_pages_free * $page_size) / $total_memory));
                $system_stats->{mem_usage_percent} += (100 - $mem_free_percent);
                $mem_samples++;

                # parse the idle CPU percentage, 16th column
                my $cpu_idle = int($numbers[15]);
                $system_stats->{cpu_usage_percent} += (100 - $cpu_idle);
                $cpu_samples++;
            }
        }
    }

    # average out cpu usage and mem usage from all the tallied samples.
    if ($cpu_samples > 0) {
        $system_stats->{cpu_usage_percent} = int($system_stats->{cpu_usage_percent} / $cpu_samples);
    }
    if ($mem_samples > 0) {
        $system_stats->{mem_usage_percent} = int($system_stats->{mem_usage_percent} / $mem_samples);
    }

    # store num cpus
    $system_stats->{num_cpus} = $num_cpus;

    return $system_stats;
}

################################################################################
# Method:    get_ip_info()
# Objective: Get the IP information for this system and return hash with that
#            data
################################################################################
sub get_ip_info {
    my ($self) = @_;

    my ($mac, $ipv4, $ipv6, $description, $gateway, $subnet_mask);
    my (@e_out, $set_interface, $mtusize, $broadcast, $ipv6_prefix_len);

    my $start = 0;
    my $finished = 0;
    my $interface = 0;
    my $interface_name = 'NA';
    my $ip_hash = {};

    # get all IP information from host via ifconfig
    my $cmd = "ifconfig -au";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::get_ip_info(): Error running command ($cmd)");
        return {};
    }

    my $cnt = 0;
    my $num_lines = scalar(@output);
    foreach my $line (@output) {
        chomp($line);
        $cnt++;

        # get IPv4 address
        if ($line =~ /inet (\S+)/) {
            $ipv4 = $1;
        }

        # get IPv6 address
        if ($line =~ /inet6 (\S+)\/(\d+)/) {
            $ipv6 = $1;
            $ipv6_prefix_len = $2;
        }

        # get MAC address
        if ($line =~ /ether (\S+)/i) {
            $mac = $1;
            $mac =~ s/\://g;
        }

        # no gateway for UNIX
        $gateway = "";

        # get subnet mask
        if ($line =~ /netmask (\S+)/i) {
            $subnet_mask = $1;
        }

        # get broadcast
        if ($line =~ /broadcast (\S+)/i) {
            $broadcast = $1;
        }

        # get mtusize
        if ($line =~ /mtu (\S+)/i) {
            $mtusize = $1;
        }

        # Look for unique sections describing interface.
        if ($line =~ /^(\S+)\:/ or $cnt == $num_lines) {
            if ($start == 1) {
                $finished = 1;
                $set_interface = $interface;
                $interface = $1;
            } else {
                $interface = $1;
            }
            $start = 1;
        }

        if ($start == 1 and $finished == 1) {
            # don't push in loopback
            if ($set_interface !~ /lo0/) {
                $ip_hash->{$set_interface}->{interface_name} = $interface_name;
                $ip_hash->{$set_interface}->{description} = "NIC";
                $ip_hash->{$set_interface}->{gateway} = $gateway;
                $ip_hash->{$set_interface}->{broadcast} = $broadcast;
                $ip_hash->{$set_interface}->{subnet_mask} = $subnet_mask;
                $ip_hash->{$set_interface}->{mtusize} = $mtusize;
                $ip_hash->{$set_interface}->{ipv4} = $ipv4;
                $ip_hash->{$set_interface}->{ipv6} = $ipv6;

                # if this is the management interface of the client
                # set an attribute
                my $host_rec = &Tharn::host($self->hostname());
                my $host_ip_default = $host_rec->{'default_ip'};
                if ($ipv4 eq $host_ip_default) {
                    $ip_hash->{$set_interface}->{is_mgmt_interface} = 1;
                } else {
                    $ip_hash->{$set_interface}->{is_mgmt_interface} = 0;
                }

                # use entstat to get the MAC address
                $cmd = "entstat $set_interface";
                ($status, @e_out) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Error: AixHost::get_ip_info(): Error running command ($cmd)");
                    return {};
                }

                # grab the MAC
                $mac = (grep(/Hardware Address/i, @e_out))[0];
                chomp($mac);
                $mac =~ s/Hardware Address\s*//i;
                $mac =~ s/://g;
                $mac =~ s/\s+//g;
                $ip_hash->{$set_interface}->{mac} = $mac;

                # calculate the ipv4 subnet
                if ($ipv4 ne "" and $subnet_mask ne "") {
                    $ip_hash->{$interface}->{ipv4_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv4,
                                                                                    -ip_version => "ipv4",
                                                                                    -ipv4_netmask => $subnet_mask);
                } else {
                    $ip_hash->{$interface}->{ipv4_subnet} = "";
                }

                # calculate the ipv6 subnet
                if ($ipv6 ne "" and $ipv6_prefix_len ne "") {
                    $ip_hash->{$interface}->{ipv6_prefix_len} = $ipv6_prefix_len;
                    $ip_hash->{$interface}->{ipv6_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv6,
                                                                                    -ip_version => "ipv6",
                                                                                    -ipv6_prefix_len => $ipv6_prefix_len);
                } else {
                    $ip_hash->{$interface}->{ipv6_prefix_len} = "";
                    $ip_hash->{$interface}->{ipv6_subnet} = "";
                }
            }

            $finished = 0;

            # initialize values
            $mac = $description = $mtusize = $broadcast = $gateway = $subnet_mask = "";
            $ipv4 = $ipv6 = "";
        }
    }

    # in addition to NICs, get information the iSCSI HW TOEs.

    # get the list of AIX hardware
    $cmd = "lscfg";
    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 120);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::get_ip_info(): Error running command ($cmd)");
        return {};
    }

    # grep for ics which tells us we found iSCSI HBAs
    my @hbas = grep(/ics/, @output);

    if (@hbas) {
        # on each adapter, run lscfg -vl and get wwpns.
        foreach my $adapter (@hbas) {
            # get the actual device name
            $adapter =~ s/.*\s+(ics\d+)\s+.*/$1/;

            $cmd = "lscfg -vl " . $adapter;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 120);
            if ($status > 0) {
                $self->_set_err("Error: AixHost::get_ip_info(): Error running command ($cmd)");
                return {};
            }

            # get the line with the MAC address
            my $port_line = (grep (/Network Address/i, @output))[0];
            # get the value
            my $mac = (split(/\.+/, $port_line))[1];

            # get IP address of card
            $cmd = "lsattr -El $adapter | grep \"host_addr\" | awk '{print \$2}'";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($output[0] =~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/) {
                $ip_hash->{$adapter}->{ipv4} = $output[0];
            } else {
                $ip_hash->{$adapter}->{ipv4} = "";
            }

            $ip_hash->{$adapter}->{mac} = $mac;
            $ip_hash->{$adapter}->{description} = "iSCSI HW ADAPTER";
            $ip_hash->{$adapter}->{ipv6} = "";
            $ip_hash->{$adapter}->{gateway} = "";
            $ip_hash->{$adapter}->{broadcast} = "";
            $ip_hash->{$adapter}->{subnet_mask} = "";
            $ip_hash->{$adapter}->{mtusize} = "";
        }
    }

    # return info hash
    return ($ip_hash);
}

################################################################################
# Subroutine Name: set_ip_address()
# Objective:       Sets the IP address and subnet mask for an interface
# Details:         See POD documentation at the end of this file
################################################################################
sub set_ip_address {
    my ($self) = @_;

    $self->_set_err("This function is not yet implemented");
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################
sub get_default_mp_type {
    my ($self) = @_;

    return("MPIO");

}

################################################################################
# Subroutine Name: rotate_logs()
# Objective:       rotate the log files for Aix
################################################################################
sub rotate_logs {
    my ($self) = @_;

    $self->_clear_err();

    # rotate the Veritas logs
    my $status = $self->rotate_generic_log(-file => "/etc/vx/dmpevents.log", -minsize => "1024", -keep => "5");
    if ($status == $ClientUtil::TRUE) {
        return $ClientUtil::TRUE;
    } else {
        if (not $self->error) {
            $self->_set_err("An error was received while attempting to rotate the Veritas DMP logs");
        }
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: get_root_dev()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_root_dev {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: update_partition_table()
# Objective:       NOOP For AIX
# Details:         See POD documentation at the end of this file
################################################################################
sub update_partition_table {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_devices()
# Objective:       wrapper to call _get_devices() routine
################################################################################
sub get_devices {
    my ($self) = @_;

    return ($self->_get_devices());

}

################################################################################
# Subroutine Name: _get_devices
# Objective:       Get list of all devices on host
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices() {
    my ($self) = @_;

    my ($cmd, $status, @output, @hDevices);
    my ($bootdevice, @b_output);

    if ($self->get_protocol() =~ /fcp/i) {
        if ($self->{mp_type} =~ /mpio/i) {
            $cmd = "lsdev -C -c disk | egrep \"\(MPIO)\|\(Virtual\)\" | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\) FCP\" | awk '{print \$1;}'";
        } elsif ($self->{mp_type} =~ /sanpath/i) {
            $cmd = "setsp | egrep 'spd[0-999]' | grep 'NETAPP' | awk '{print \$3;}'";
        } else {
            $cmd = "lsdev -C -c disk | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\)\" | awk '{print \$1;}'";
        }
    } elsif ($self->get_protocol() =~ /iscsi/i) {
        if ($self->{mp_type} =~ /mpio/i) {
            $cmd = "lsdev -C -c disk | egrep \"\(MPIO)\|\(Virtual\)\" | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\) iSCSI\" | awk '{print \$1;}'";
        } else {
            $cmd = "lsdev -C -c disk | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\)\" | awk '{print \$1;}'";
        }
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::_get_devices() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # get the boot device - do not attempt to rmdev this - could
    # be sanbooted netapp device.
    $cmd = "bootinfo -b";
    ($status, @b_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::_get_devices() Command ($cmd) returned a bad status ($status)");
    }
    $bootdevice = $b_output[0];

    foreach my $device (@output) {
        chomp $device;

        # skip the boot device
        if ($device eq $bootdevice) {
            next;
        }

        my $hDevs = {};
        $hDevs->{path} = "/dev";
        $hDevs->{name} = $device;
        $hDevs->{type} = "NETAPP";
        $hDevs->{label} = "NA";
        $hDevs->{id} = $device;

        push (@hDevices, $hDevs);
    }

    $self->sanlog(-msg => scalar(@hDevices) . " devices found ");

    return (wantarray()) ? @hDevices : \@hDevices;

}

################################################################################
# Subroutine Name: _get_unlabeled_devices
# Objective:       Retrieve devices that do not have a label
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_unlabeled_devices {
    my ($self) = @_;

    my ($cmd, $status, @hDevices, @all_devices, @devices);

    # Retrieve all host devices
    @hDevices = $self->_get_devices();
    return 0 if ($self->error);

    # Get list of all devices and paths
    $cmd = "lspv | grep -i none | awk '{print \$1}'";
    ($status, @all_devices) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::_get_unlabeled_devices() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }


    # Determine which devices are not being used by a vg
    foreach my $device (@hDevices) {
        if (grep /^$device->{name}$/, @all_devices) {
            push (@devices, $device->{name});
        }
    }

    # Format the list of devices with label
    return (wantarray()) ? @devices : \@devices;

}

################################################################################
# Subroutine Name: _config_paths
# Objective:       For AIX and MPIO (w/o ALUA) need to run dotpaths
# Details:         See POD documentation at the end of this file
################################################################################
sub _config_paths {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    # check to see if ALUA support
    if (($self->get_alua_support() == 0) and
        ($self->get_protocol() =~ /fcp/i)) {
        if ($self->{mp_type} =~ /mpio/i) {
            $self->sanlog(-msg => "Running dotpaths to prioritize paths (Non-ALUA)");
            $cmd = "dotpaths -v";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

            # determine if Dotpaths command worked
            if ($status > 0) {
                $self->_set_err("Error: AixHost::_config_paths() Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _iscsi_configure()
# Objective:       Configure iscsiadm to include the correct targets for the
#                  filers.
#                  Assumptions: a targets list is passed into this function with
#                  the IP's of all the ports on each filer.
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _iscsi_configure {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Error: AixHost::iscsi_configure() Invalid ".
                                "value for argument '-targets' => %args{$_}");
            }
            next;
        } else {
            $self->_set_err("Error: AixHost::_iscsi_configure() Invalid ".
                            " parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg =>
                  "AixHost::_iscsi_configure: Starting iSCSI configuration");

    # check to make sure the HW/SW initiators are properly setup.
    # Also, store in the object the iSCSI SW initiator devices and
    # the iSCSI HW adapter devices.
    if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /hardware/) {
        $self->_check_and_set_iscsi_hw();
        if ($self->error) {
            $self->sanlog(-msg => "Error: AixHost::configure() Problems checking the iSCSI HW adapter(s)");
            return $ClientUtil::FALSE;
        }
    } elsif ($self->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /software/) {
        $self->_check_and_set_iscsi_sw();
        if ($self->error) {
            $self->sanlog(-msg => "Error: AixHost::configure() Problems checking the iSCSI SW initiator");
            return $ClientUtil::FALSE;
        }
    }

    # Skip iSCSI discover and configuration if requested
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY} == 0) {
         $self->sanlog(-msg => "AixHost::_iscsi_configure: iSCSI configuration skipped, " .
                               "CONFIG_ISCSI_DISCOVERY = $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY}");
         return();
    }

    # configure the discover information for HW and SW separately.
    if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /hardware/) {
        $self->_configure_hw_config(@targets);
        if ($self->error) {
            $self->sanlog(-msg => "Error: AixHost::configure() Problems configuring the iSCSI HW adapter(s)");
            return $ClientUtil::FALSE;
        }
    } elsif ($self->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /software/) {
        $self->_configure_sw_config(@targets);
        if ($self->error) {
            $self->sanlog(-msg => "Error: AixHost::configure() Problems configuring the iSCSI SW initiator");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _fcp_configure()
# Objective:       Configure FCP on the host.
# Details:         See POD documentation at the end of this file
################################################################################
sub _fcp_configure {
    my ($self) = @_;

    my ($status, @output, $cmd, @cmds);
    my (@adapter_output);

    $self->sanlog(-msg =>
                  "AixHost::_fcp_configure: Starting FCP configuration");

    $cmd = "lscfg | grep fcs | awk '{print \$2}'";
    ($status, @adapter_output) = &ClientUtil::execute($self->get_connection(), $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::_fcp_configure() Could not get list of FCP adapters",
                        @output);
        return $ClientUtil::FALSE;
    } else {
        # store the adapter info in the host object
        $self->{fc_adapters} = \@adapter_output;
        $self->sanlog(-msg => "AixHost::_fcp_configure: FC Adapters are @adapter_output");
    }

    my $num_cmd_elems = $self->{fcp}->{CONFIG_FCP_AIX_NUM_CMD_ELEMS};
    my $dyntrk = $self->{fcp}->{CONFIG_FCP_AIX_DYNAMIC_TRACKING};
    my $fc_err_recov = $self->{fcp}->{CONFIG_FCP_AIX_ERROR_RECOVERY};

    if (defined($num_cmd_elems) and length($num_cmd_elems)) {
        $self->sanlog(-msg => "AixHost::_fcp_configure: Setting num_cmd_elems=${num_cmd_elems} on the FC adapters");
    }

    if (defined($dyntrk) and length($dyntrk)) {
        $self->sanlog(-msg => "AixHost::_fcp_configure: Setting 'Dymanic Tracking=${dyntrk}' on the FC adapters");
    }

    if (defined($fc_err_recov) and length($fc_err_recov)) {
        $self->sanlog(-msg => "AixHost::_fcp_configure: Setting 'FC Error Recovery=${fc_err_recov}' on the FC adapters");
    }

    # removing the devices and readjusting the settings per adapter
    foreach my $adapter (@adapter_output) {
        # get the protocol device from the adapter
        $cmd = "lsdev -p $adapter | grep fscsi | awk '{print \$1}'";
        ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
        if ($status > 0) {
           $self->_set_err("Error: AixHost::_fcp_configure() Error executing command ($cmd)", @output);
           return $ClientUtil::FALSE;
        }

        my $protocol_device = $output[0];
        chomp($protocol_device);

        @cmds = ("rmdev -l $protocol_device -R",
                 "rmdev -l $adapter");

        # set num_cmd_elems if specified
        if (defined($num_cmd_elems) and length($num_cmd_elems)) {
            push(@cmds, "chdev -l $adapter -a num_cmd_elems=${num_cmd_elems}");
        }

        # set fc_err_recov if specified
        if (defined($fc_err_recov) and length($fc_err_recov)) {
            push(@cmds, "chdev -l $protocol_device -a fc_err_recov=${fc_err_recov}");
        }

        # set dyntrk if specified
        if (defined($dyntrk) and length($dyntrk)) {
            push(@cmds, "chdev -l $protocol_device -a dyntrk=${dyntrk}");
        }

        # add commands to re-run cfgmgr for the device and make the devices
        # again
        push(@cmds, "cfgmgr -l $adapter");
        push(@cmds, "mkdev -l $adapter");
        push(@cmds, "mkdev -l $protocol_device");

        foreach $cmd (@cmds) {
            ($status, @output) = &ClientUtil::execute($self->get_connection(),
                                                      $cmd);
            if ($status > 0) {
                $self->_set_err("Error: AixHost::_fcp_configure() Error executing command ($cmd)",
                                @output);
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _fcp_deconfigure()
# Objective:       Deconfigures fcp
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _fcp_deconfigure {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure()
# Objective:       Deconfigures iscsi
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _iscsi_deconfigure {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _check_and_set_iscsi_hw
# Objective:       Check the HW iSCSI card to make sure it is properly setup
#                  and so that it is worth while moving forward with config
# Details:         See POD documentation at the end of this file
################################################################################
sub _check_and_set_iscsi_hw {
    my ($self) = @_;

    my ($status, @output, $output, $cmd, $adapter);
    my (@a_output);

    # let's determine if the iSCSI HW adapter(s) are in the system
    # and store in object
    $cmd = "lsdev -Cc adapter | grep \'iSCSI Adapter\' | awk '{print \$1}'";
    ($status, @a_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 0) {
        $self->{hw_iscsi_adapters} = \@a_output;
    } else {
        $self->_set_err("Error: AixHost::_check_and_set_iscsi_hw() No HW adapters found in the system");
        return ($ClientUtil::FALSE);
    }

    # let' check to make sure host_addr (IP address of card) and
    # initiator_name (IQN name) are set. If not, kick out.
    foreach $adapter (@{$self->{hw_iscsi_adapters}}) {
        $cmd = "lsattr -El $adapter | grep \"host_addr\" | awk '{print \$2}'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($output[0] eq "" or $output[0] !~ /(\d+)\.(\d+)\.(\d+)\.(\d+)/) {
            $self->_set_err("Error: AixHost::_check_and_set_iscsi_hw() $adapter host_addr attribute has not been set properly ($output[0])");
            return ($ClientUtil::FALSE);
        }

        $cmd = "lsattr -El $adapter | grep \"initiator_name\" | awk '{print \$2}'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($output[0] eq "" or $output[0] !~ /iqn/i) {
            $self->_set_err("Error: AixHost::_check_and_set_iscsi_hw() $adapter initiator_name attribute has not been set properly ($output[0])");
            return ($ClientUtil::FALSE);
        }
    }

}

################################################################################
# Subroutine Name: _check_and_set_iscsi_sw
# Objective:       Check the SW iSCSI initiator to make sure it is properly
#                  setup and so that it is worth while moving forward with
#                  config
# Details:         See POD documentation at the end of this file
################################################################################
sub _check_and_set_iscsi_sw {
    my ($self) = @_;

    my ($status, @output, $output, $cmd);
    my ($dev, $location_code);

    # let's determine if the iSCSI SW intiator protocol device
    # and store in object. Method we are using is to look lsdev
    # output and check the location code of iSCSI Protocol devices.
    # For the iSCSI SW initiator the protocol device will be empty.
    $cmd = "lsdev | grep \'iSCSI Protocol Device\'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixHost::_check_and_set_iscsi_sw() No SW Initiator Protocol Device found in the system");
        return ($ClientUtil::FALSE);
    } else {
        foreach $dev (@output) {
            my $location_code = (split(/\s+/, $dev))[2];
            if ($location_code =~ /iSCSI/) {
                $self->{sw_initiator} = (split(/\s+/, $dev))[0];
            }
        }
    }

    # let' check to make sure the # initiator_name (IQN name) is set.
    # If not, kick out.
    $cmd = "lsattr -El $self->{sw_initiator} | grep \"initiator_name\" | awk '{print \$2}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($output[0] eq "" or $output[0] !~ /iqn/i) {
        $self->_set_err("Error: AixHost::_check_and_set_iscsi_sw() $self->{sw_initiator} initiator_name attribute has not been set properly ($output[0])");
        return ($ClientUtil::FALSE);
    }

}

################################################################################
# Subroutine Name: _configure_sw_config
# Objective:       Configure the iSCSI SW initiator config file relative
#                  to the controllers.
#                  Note: iSCSI SW does not allow the disc_filename to be
#                  changed, so /etc/iscsi/targets will always be the location
# Details:         See POD documentation at the end of this file
################################################################################
sub _configure_sw_config {
    my ($self, @targets) = @_;

    my ($status, @output, $cmd, @cmds, $t_file);
    my ($target);

    # this cannot change since disc_filename cannot be modified.
    $t_file = "/etc/iscsi/targets";

    # determine if the iscsi/targets file is the original IBM version. If
    # so, save, if not, we are going blow away and re-create.
    $cmd = "cat $t_file | grep \"Licensed Materials - Property of IBM\"";
    ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
    if ($status eq 0) {
        $cmd = "mv -f $t_file ${t_file}.ibm.save";
        ($status, @output) = &ClientUtil::execute($self->get_connection(),
                                                  $cmd);
        if ($status > 0) {
            $self->_set_err("Error: AixHost::iscsi_configure() Error backing up $t_file",
                            @output);
            return ($ClientUtil::FALSE);
        }
    }

    # remove the targets file (if present -f flag)
    $cmd = "rm -f $t_file";
    ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
    if ($status ne 0) {
        $self->_set_err("Error: AixHost::iscsi_configure() Error removing $t_file", @output);
        return ($ClientUtil::FALSE);
    }

    # In C-mode, targets will have either ipv4 or ipv6 address unlike 7-mode.
    # Hence, remove the targets from the target list which don't have valid IP address
    # of IP version requested by user.
    @targets = $self->filter_targets_by_ipversion(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # generate list of commands to create the targets file
    foreach my $target (@targets) {
        $cmd = "echo \"# Controller: $target->{filer_name}\" >> $t_file";
        push (@cmds, $cmd);
        $cmd = "echo \"$target->{ip} 3260 $target->{nodename}\" >> $t_file";
        push (@cmds, $cmd);
        $cmd = "echo \" \" >> $t_file";
        push (@cmds, $cmd);
    }

    # generate the targets file
    foreach $cmd (@cmds) {
        ($status, @output) = &ClientUtil::execute($self->get_connection(),
                                                  $cmd);
        if ($status ne 0) {
            $self->_set_err("Error: AixHost::iscsi_configure() Error building $t_file",
                            @output);
            return ($ClientUtil::FALSE);
        }
    }

}

################################################################################
# Subroutine Name: _configure_hw_config
# Objective:       Configure the iSCSI SW initiator config file relative
#                  to the controllers.
#                  changed, so /etc/iscsi/targets will always be the location
# Details:         See POD documentation at the end of this file
################################################################################
sub _configure_hw_config {
    my ($self, @targets) = @_;

    my ($status, @output, $cmd, @cmds, $t_file);
    my ($target, $adapter, $a_subnet, $t_subnet);
    my ($protocol_device);

    foreach $adapter (@{$self->{hw_iscsi_adapters}}) {

        $t_file = "/etc/iscsi/targets.hw.nate.${adapter}";

        # get subnet info for adapter (to make sure it's on
        # the sam subnet as target
        $cmd = "lsattr -El $adapter | grep \"host_addr\" | awk '{print \$2}'";
        ($status, @output) = &ClientUtil::execute($self->get_connection(),
                                                  $cmd);
        if ($status ne 0) {
            $self->_set_err("Error: AixHost::iscsi_configure() Error removing $t_file", @output);
            return ($ClientUtil::FALSE);
        } else {
            $output[0] =~ m/(\d+)\.(\d+)\.(\d+)\.(\d+)/;
            $a_subnet = "$1\.$2\.$3";
        }

        # remove the targets file (if present -f flag)
        $cmd = "rm -f $t_file";
        ($status, @output) = &ClientUtil::execute($self->get_connection(),
                                                  $cmd);
        if ($status ne 0) {
            $self->_set_err("Error: AixHost::iscsi_configure() Error removing $t_file",
                            @output);
            return ($ClientUtil::FALSE);
        }

        # generate list of commands to create the targets file
        foreach my $target (@targets) {

            # get subnet of target to compare versus adapter
            $t_subnet = $target->{ip};
            $t_subnet =~ m/(\d+)\.(\d+)\.(\d+)\.(\d+)/;
            $t_subnet = "$1\.$2\.$3";

            if ($t_subnet eq $a_subnet) {
               $cmd = "echo \"# Controller: $target->{filer_name}\" >> $t_file";
               push (@cmds, $cmd);
               $cmd = "echo \"$target->{ip} 3260 $target->{nodename}\" >> $t_file";
               push (@cmds, $cmd);
               $cmd = "echo \" \" >> $t_file";
               push (@cmds, $cmd);
            } else {
               $self->sanlog(-msg => "Skipping add of iSCSI target ($t_subnet}, not on same subnet as adapter ($a_subnet)");
            }
        }

        # generate (write out) the targets file
        foreach $cmd (@cmds) {
            ($status, @output) = &ClientUtil::execute($self->get_connection(),
                                                      $cmd);
            if ($status ne 0) {
                $self->_set_err("Error: AixHost::iscsi_configure() Error building $t_file",
                                @output);
                return ($ClientUtil::FALSE);
            }
        }

        # get protocol device from the adapter
        $cmd = "lsdev -p $adapter | awk '{print \$1}'";
        ($status, @output) = &ClientUtil::execute($self->get_connection(),
                                                  $cmd);
        if ($status ne 0) {
            $self->_set_err("Error: AixHost::iscsi_configure() Could not get protocol device for $adapter");
            return ($ClientUtil::FALSE);
        } else {
            $protocol_device = $output[0];
        }

        # execute the set of commands to set the disc_filename and
        # disc_policy for the adapter
        @cmds = ("rmdev -l $protocol_device",
                 "rmdev -l $adapter",
                 "chdev -l $adapter -a disc_policy=file",
                 "chdev -l $adapter -a disc_filename=$t_file",
                 "cfgmgr -l $adapter",
                 "mkdev -l $adapter",
                 "mkdev -l $protocol_device");

        foreach $cmd (@cmds) {
            ($status, @output) = &ClientUtil::execute($self->get_connection(),
                                                      $cmd);
            if ($status ne 0) {
                $self->_set_err("Error: AixHost::iscsi_configure() Error executing command ($cmd)");
                return ($ClientUtil::FALSE);
            }
        }
    }

}

################################################################################
# Subroutine Name: _MP_configure
# Objective:       Configures MP on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _MP_configure {
    my ($self) = @_;

    # verify that all mp pkgs are installed
    $self->{mp}->verify(-conn => $self->{conn});
    if ($self->{mp}->error) {
        $self->_set_err($self->{mp}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "MP is verified", -msg_level => 2);

    # configure MP
    $self->{mp}->config(-conn => $self->{conn},
                        -protocol => $self->{protocol});
    if ($self->{mp}->error) {
        $self->_set_err($self->{mp}->errmsg());
        return $ClientUtil::FALSE;
    }

    # enable MP
    $self->{mp}->enable(-conn => $self->{conn});
    if ($self->{mp}->error) {
        $self->_set_err($self->{mp}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "MP is configured", -msg_level => 2);

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _obtain_memory_info()
# Objective:       Store the page size of this system on the object
################################################################################
sub _obtain_memory_info {
    my ($self) = @_;
    my ($cmd, $status, @output);

    # obtain the page size, should only be one lined entry in the output
    $cmd = "/usr/bin/pagesize";
    ($status, @output) = &ClientUtil::execute($self->get_connection(), $cmd);
    if ($status > 0 or scalar(@output) != 1) {
        $self->_set_err("Error: AixHost::new() Failed to obtain the page size");
        return $ClientUtil::FALSE;
    }

    # remove spaces.
    $output[0] =~ s/\s+//g;

    # store the page size
    $self->{memory_page_size} = $output[0];
    return $ClientUtil::TRUE;
}

1;
