# Copyright (c) 2012 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author mathanku@netapp.com
## @status private
## @pod ../Host.pod

use strict;
use warnings;

# Package Name
##############
package San::SanClient::Host::CXSHost;

# Package Requirements
######################
use File::Basename;
use Data::Dumper;
use San::SanClient;
use San::SanClient::Host::LinuxHost;
use San::SanClient::ClientUtil;

use San::SanLog;
use Error;
use vars qw( @ISA );

@ISA = qw(San::SanClient::Host::LinuxHost Error SanLog SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new Citrix XenServer object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($device_str, @devices, $virtualization);
    my ($self, $conn, $nfs_mount, $cmd, $status, $hosttype, $protocol, @output, $service, $obj_template);

    $self = {};
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for -protocol: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Invalid value for argument '-conn'");
        return $self;
    }

    if ($protocol !~ /fcp|iscsi/i) {
        $self->_set_err("Invalid value for argument '-protocol'." .
                        " It can be either fcp or iscsi");
        return $self;
    }

    $self->{protocol} = $protocol;
    $self->{conn}     = $conn;
    $self->{virtualization}     = $virtualization;
    $self->{nfs_mount} = undef;

    $self->store_OS_major_and_minor_revisions();
    if ($self->error) {
        return $self;
    }

    $self->sanlog(-msg => "OS version: " . $self->{major_revision});

    # store TEST_BIN directory
    $self->store_test_bin_dir();
    if ($self->error) {
        return $self;
    }

    $self->{iscsi}->{iscsi_version} = "open-iscsi";

    # store client kernel version
    $cmd = "uname -r";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    $self->{revision} = $output[0];

    # get host uuid
    my $host_uuid = "";
    $cmd = "xe host-list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    foreach my $line (@output) {
        if ($line =~ /uuid.*:\s*(\S+)/i) {
            $host_uuid = $1;
        } elsif ($line =~ /name-label.*:(.+)/i) {
            my $name_label = $1;
            $name_label =~ s/^\s*//;
            my $host = $self->{conn}->{hostp}->{hostname};
            if ($name_label =~ /$host/i) {
                $self->{uuid} = $host_uuid;
                $self->sanlog(-msg => "Hostname: " . $host);
                $self->sanlog(-msg => "Host UUID: " . $self->{uuid});
                last;
            } else {
                $host_uuid = "";
            }
        }
    }

    # need to load config params right now so we can access them
    # for decision on software or hardware initiator
    $self->_load_config_params();
    if ($self->error()) {
        return $self;
    }

    return $self;
}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            if (not defined ($cfmode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->{valid} = 1;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: xecontrol()
# Objective:       Will perform the requested action on VMs
################################################################################
sub xecontrol {
    my ($self, %args) = @_;

    my ($action, $vm_name, $cmd, $status, @output);
    my $options = "";

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?action/) {
            if (not length($action = $args{$_})) {
                $self->_set_err("Invalid value for -action => $action");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name => $vm_name");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?options/) {
            if (not length($options = $args{$_})) {
                $self->_set_err("Invalid value for -options => $options");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    unless (defined($action)) {
        $self->_set_err("Missing required argument 'action'");
        return $ClientUtil::FALSE;
    }

    unless (defined($vm_name)) {
        $self->_set_err("Missing required argument 'vm_name'");
        return $ClientUtil::FALSE;
    }

    $cmd = "xe $action vm=$vm_name";
    $cmd .= " $options" if $options !~ /^\s*$/;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 900);
    if ($status != 0) {
        $self->_set_err("Failed to perform action $action on VM $vm_name");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: domain_management()
# Objective:       method to allow domain management for virtual machines
################################################################################
sub domain_management {
    my ($self, %args) = @_;

    my ($action);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?action/) {
            if (not length($action = $args{$_})) {
                $self->_set_err("Invalid value for -action => $action");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if ($action =~ /^resume$/i) {
        $self->_resume_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^resume_suspended$/i) {
        $self->_resume_all_suspended_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^suspend$/i) {
        $self->_suspend_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^suspend_all$/i) {
        $self->_suspend_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_on$/i) {
        $self->_power_on_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_on_all$/i) {
        $self->_power_on_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_off$/i) {
        $self->_power_off_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_off_all$/i) {
        $self->_power_off_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^shutdown$/i) {
        $self->_shutdown_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^shutdown_all$/i) {
        $self->_shutdown_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_cycle$/i) {
        $self->_power_cycle_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_cycle_all$/i) {
        $self->_power_cycle_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################
# Subroutine Name: get_vm_list()
# Objective:       Returns a hash ref which has info of all VMs
################################################################
sub get_vm_list {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my ($vm_info) = {};
    my $refresh = 0;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Invalid value for -refresh => $refresh");
                return $vm_info;
            }
            next;
        }
    }

    if ($refresh =~ /0|n/i and defined($self->{vm_info})) {
        return $self->{vm_info};
    }

    $cmd = "xe vm-list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to get guest VMs list");
        return $vm_info;
    }

    ###########################################################################
    # Here is the output of above command
    ###########################################################################
    # [root@IBMblade-210-172 ~]# xe vm-list
    # uuid ( RO)           : 5794daca-4260-43af-b8ae-811cb66af8e1
    #      name-label ( RW): VM2
    #     power-state ( RO): running
    ###########################################################################

    my $uuid = "";
    my $found = 0;
    my $vm_name = "";
    foreach my $line (@output) {
        # Skip empty lines
        next if ($line =~ /^\s*$/);

        if ($line =~ /uuid.*:\s*(\S+)/i) {
            $uuid = $1;
            $found = 1;
        } elsif ($line =~ /name-label.*:(.+)/i and $found == 1) {
            $vm_name = $1;
            $vm_name =~ s/^\s*//;
            if ($vm_name =~ /Control\s+domain\s+on\s+host/i) {
                # Skip the step for hypervisor
                $found = 0;
                next;
            }

            $vm_info->{$vm_name}->{uuid} = $uuid;
        } elsif ($line =~ /power-state.*:\s*(\S+)/i and $found == 1) {
            $vm_info->{$vm_name}->{state} = $1;
            $found = 0;
            $vm_name = "";
        }
    }

    $self->{vm_info} = $vm_info;

    return $vm_info;
}

################################################################################
# Subroutine Name: orphaned_device_cleanup()
# Objective:       When the luns are removed from the filer, we will have stale
#                  entries of sdX devices. These devices should be removed.
################################################################################
sub orphaned_device_cleanup {
    my ($self, %args) = @_;

    my ($targets);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not ($targets = $args{$_})) {
                $self->_set_err("Invalid value for " . "-targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # call discover() to remove all orphaned devices
    $self->discover(-targets => $targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    # NOOP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: deconfigure
# Objective:       NOOP
################################################################################
sub deconfigure {
    my ($self) = @_;

    $self->_clear_err();

    # NOOP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: add_devices_to_vm()
# Objective:       Add disks to VM
################################################################################
sub add_devices_to_vm {
    my ($self, %args) = @_;

    my (@devices) = ();
    my $vm_name;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not scalar(@devices = @{$args{$_}})) {
                $self->_set_err("Invalid value for -devices => @devices");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name => $vm_name");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@devices)) {
        $self->_set_err("Missing or invalid value for -devices");
        return $ClientUtil::FALSE;
    }

    unless (defined($vm_name)) {
        $self->_set_err("Missing required argument 'vm_name'");
        return $ClientUtil::FALSE;
    }

    my $vm_info = $self->get_vm_list();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    if (not exists($vm_info->{$vm_name})) {
        $self->_set_err("VM ($vm_name) is not found in the vm-list. Make sure you have provided " .
                        "correct VM name.");
        return $ClientUtil::FALSE;
    }

    my ($status, $cmd, @output);
    my $device_num = 1;
    my $vm_uuid = $vm_info->{$vm_name}->{uuid};

    foreach my $device (@devices) {
        my $vdi_uuid = $device->{uuid};
        my $num_retry = 1;
        my $attached = 0;

        while ($num_retry <= 10) {
            $cmd = "xe vbd-create vm-uuid=$vm_uuid device=$device_num vdi-uuid=$vdi_uuid";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status == 0) {
                $attached = 1;
                last;
            } else {
                if ($num_retry == 5) {
                    # Just to make sure that if many devices are already attached
                    $device_num += 10;
                } else {
                    $device_num++;
                }
                $num_retry++;
            }
        }

        if ($attached != 1) {
            $self->_set_err("Failed to create VBD for VM ($vm_name)");
            return $ClientUtil::FALSE;
        }

        $device_num++;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_devices_from_vm()
# Objective:       Get the devices attached to a specific VM
################################################################################
sub get_devices_from_vm {
    my ($self, %args) = @_;

    my $vm_name;
    my ($status, $cmd, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name => $vm_name");
                return ();
            }
            next;
        }
    }

    unless (defined($vm_name)) {
        $self->_set_err("Missing required argument 'vm_name'");
        return $ClientUtil::FALSE;
    }

    my $vm_info = $self->get_vm_list();
    if ($self->error) {
        return ();
    }

    if (not exists($vm_info->{$vm_name})) {
        $self->_set_err("VM ($vm_name) is not found in the vm-list. Make sure you have provided " .
                        "the correct VM name.");
        return ();
    }

    my $vdi_hash;
    my $vm_uuid = $vm_info->{$vm_name}->{uuid};

    $cmd = "xe vbd-list vm-uuid=$vm_uuid params=uuid,vdi-uuid,vdi-name-label";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to list VBD attached to VM ($vm_name)");
        return ();
    }

    my $vbd_uuid = "";
    my $vdi_uuid = "";
    my $found = 0;
    foreach my $line (@output) {
        if ($line =~ /^\s*uuid.*:\s*(\S+)/i) {
            $vbd_uuid = $1;
            $found = 1;
        } elsif ($line =~ /vdi-uuid.*:\s*(\S+)/i and $found == 1) {
            $vdi_uuid = $1;
            $vdi_hash->{$vdi_uuid}->{vdi_uuid} = $vdi_uuid;
            $vdi_hash->{$vdi_uuid}->{vbd_uuid} = $vbd_uuid;
        } elsif ($line =~ /vdi-name-label.*:(.+)/i and $found == 1) {
            my $name_label = $1;
            $name_label =~ s/^\s*//;
            $vdi_hash->{$vdi_uuid}->{vdi_name_label} = $name_label if exists $vdi_hash->{$vdi_uuid};

            $found = 0;
            $vdi_uuid = "";
            $vbd_uuid = "";
        }
    }

    return values %{$vdi_hash};
}

################################################################################
# Subroutine Name: remove_devices_from_vm()
# Objective:       Revome the devices attached to specific VM
################################################################################
sub remove_devices_from_vm {
    my ($self, %args) = @_;

    my (@devices) = ();
    my $vm_name;
    my ($status, $cmd, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not scalar(@devices = @{$args{$_}})) {
                $self->_set_err("Invalid value for -devices => @devices");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name => $vm_name");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@devices)) {
        $self->_set_err("Missing or invalid value for -devices");
        return $ClientUtil::FALSE;
    }

    unless (defined($vm_name)) {
        $self->_set_err("Missing required argument 'vm_name'");
        return $ClientUtil::FALSE;
    }

    foreach my $device (@devices) {
        my $vdi_uuid = $device->{vdi_uuid};
        my $vbd_uuid = $device->{vbd_uuid};
        my $vdi_name_label = $device->{vdi_name_label};

        if ($vdi_name_label !~ /NATE_VDI/i) {
            next;
        }

        $cmd = "xe vbd-unplug uuid=$vbd_uuid";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to unplug VBD ($vbd_uuid)");
            return $ClientUtil::FALSE;
        }

        $cmd = "xe vdi-destroy uuid=$vdi_uuid";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to destroy VDI ($vdi_uuid)");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

############################################################
# Subroutine Name: discover
# Objective: discovers newly added devices to the host
############################################################
sub discover {
    my ($self, %args) = @_;

    my (@output, $status, $cmd);
    my ($targets);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not ($targets = $args{$_})) {
                $self->_set_err("Invalid value for " . "-targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "Discovering newly added devices");

    my $host_uuid = $self->{uuid};

    if ($self->{protocol} =~ /iscsi/i) {
        foreach my $target (@{$targets}) {
            my $filer_ip = $target->{ip};
            my $target_iqn = $target->{nodename};

            $cmd = "xe sr-probe host-uuid=$host_uuid  type=lvmoiscsi device-config:target=$filer_ip ";
            $cmd .= "device-config:targetIQN=$target_iqn";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

            # Not checking the status, because there is an issue with the command which returns
            # error status even if command passes
        }
    } elsif ($self->{protocol} =~ /fcp/i) {
        $cmd = "xe sr-probe host-uuid=$host_uuid type=lvmohba";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # Not checking the status, because there is an issue with the command which returns
        # error status even if command passes
    }

    return $ClientUtil::TRUE;
}

########################################################
# Subroutine Name: configure
# Objective: configures the host machine
########################################################
sub configure {
    my ($self, %args) = @_;

    $self->sanlog(-msg => "Configure");

    if ($self->{protocol} =~ /iscsi/i) {
        $self->_iscsi_configure();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($self->{protocol} =~ /fcp/i) {
        $self->_fcp_configure();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    # Configure multipath
    $self->_MP_configure();
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    create_vdi_disks()
# Objective: Creates VDI disks on the given SRs
###############################################################################
sub create_vdi_disks {
    my ($self, %args) = @_;

    # parse the arguments
    my $storage_repos = exists($args{-storage_repos}) ? $args{-storage_repos} : undef;
    if (not scalar(@{$storage_repos})) {
        $self->_set_err("Missing required -storage_repos parameter");
        return ();
    }

    my $num_vdi_per_lun = exists($args{-num_vdi_per_lun}) ? $args{-num_vdi_per_lun} : 1;
    my ($cmd, @output, $status);
    my $vdi_hash;

    # based on mp type, perform discovery
    foreach my $sr (@{$storage_repos}) {
        my $sr_uuid = $sr->{uuid};

        # convert size into mega bytes
        my $sr_size = &ClientUtil::convert_to_bytes($sr->{size}, "m");

        # we can't use the complete size, need to exclude some space for meta data usage
        # 100MB has been excluded here
        $sr_size -= 100;

        # calculate vdi size from SR size based on number VDIs need to be created per LUN/SR
        my $vdi_size = $sr_size / $num_vdi_per_lun;

        for (my $i = 0; $i < $num_vdi_per_lun; $i++) {
            my $vdi_name = "NATE_VDI_" . $sr->{name} . "_$i";
            $cmd = "xe vdi-create sr-uuid=$sr_uuid type=user virtual-size=${vdi_size}MiB name-label=$vdi_name";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return ();
            }

            my $vdi_uuid = $output[0];
            $vdi_hash->{$vdi_uuid}->{name} = $vdi_name;
            $vdi_hash->{$vdi_uuid}->{size} = $vdi_size;
            $vdi_hash->{$vdi_uuid}->{uuid} = $vdi_uuid;
        }
    }

    return values %{$vdi_hash};
}

###############################################################################
# Method:    destroy_storage_repos()
# Objective: Destroys empty storage repositories
###############################################################################
sub destroy_storage_repos {
    my ($self) = @_;

    my ($cmd, @output, $status);

    my @storage_repos = $self->_get_sr_list();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # get mpdevmgr object
    if (not defined($self->{obj}{mpdevmgr})) {
        $self->_set_err("An MPDevMgr instance has not been created for client " . $self->hostname());
        return $ClientUtil::FALSE;
    }
    my $mpdevmgr = $self->{obj}{mpdevmgr};

    foreach my $sr (@storage_repos) {
        my $name_label = $sr->{name_label};

        if ($name_label !~ /NATE_SR/i) {
            # skip if the SR is not created by automation
            next;
        }

        # unplug PBDs
        foreach my $pbd (@{$sr->{pbds}}) {
            $cmd = "xe pbd-unplug uuid=$pbd";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }

        $cmd = "xe sr-forget uuid=" . $sr->{uuid};
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to destroy SR ($name_label)\n");
            return $ClientUtil::FALSE;
        }

        # Get mpdevobj for this SR
        my ($mpdev_obj) = $mpdevmgr->get_filtered_devices(-os_specific_data_filter => { sr_label => $name_label });
        if ($mpdevmgr->error() or not defined($mpdev_obj)) {
            $self->_set_err("Failed to obtain MPDev device for SR '$name_label' from MPDevMgr", $mpdevmgr->errmsg());
            return $ClientUtil::FALSE;
        }

        # obtain the mp_device name for this SR
        my $mp_device_name = $mpdev_obj->mp_device_name();

        # load the device's current state into memory
        $mpdev_obj->load();
        if ($mpdev_obj->error()) {
            $self->_set_err("Failed to load $mp_device_name", $mpdev_obj->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    _get_sr_list()
# Objective: returns a hash with information of all SRs
###############################################################################
sub _get_sr_list {
    my ($self) = @_;

    my ($cmd, @output, $status);
    my $sr_hash;

    $cmd = "xe sr-list params=uuid,name-label,VDIs,PBDs,sm-config";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return ();
    }

    my $uuid = "";
    foreach my $line (@output) {
        if ($line =~ /uuid.*:\s*(\S+)/i) {
            $sr_hash->{$1}->{uuid} = $1;
            $uuid = $1;
        } elsif ($line =~ /name-label.*:(.+)/i) {
            my $name_label = $1;
            $name_label =~ s/^\s*//;
            $sr_hash->{$uuid}->{name_label} = $name_label if $uuid !~ /^\s*$/;
        } elsif ($line =~ /VDIs.*:(.+)/i) {
            my $vdi_string = $1;
            $vdi_string =~ s/\s+//g;
            my @vdis = split(/,/, $vdi_string);
            $sr_hash->{$uuid}->{vdis} = \@vdis if $uuid !~ /^\s*$/;
        } elsif ($line =~ /PBDs.*:(.+)/i) {
            my $pbd_string = $1;
            $pbd_string =~ s/\s+//g;
            my @pbds = split(/,/, $pbd_string);
            $sr_hash->{$uuid}->{pbds} = \@pbds if $uuid !~ /^\s*$/;
        } elsif ($line =~ /sm-config.*:(.+)/i) {
            if ($1 =~ /scsi-([\S]{33})/i) {
                $sr_hash->{$uuid}->{scsi_id} = $1 if $uuid !~ /^\s*$/;
            }
            $uuid = "";
        }
    }

    return values %{$sr_hash};
}

###############################################################################
# Method:    create_storage_repos()
# Objective: Creates storage repositories for the given LUNs/devices
###############################################################################
sub create_storage_repos {
    my ($self, %args) = @_;

    # parse the arguments
    my $devices = exists($args{-devices}) ? $args{-devices} : undef;
    if (not scalar(@{$devices})) {
        $self->_set_err("Missing required -devices parameter");
        return ();
    }

    my ($cmd, @output, $status);
    my $sr_hash;

    foreach my $device (@{$devices}) {
        my $host_uuid = $self->{uuid};
        my $scsi_id = $device->{mp_device_name};
        my $sr_name = "NATE_SR_" . $device->{protocol} . "_" . $scsi_id;
        my $execute_flag = 0;

        if ($device->{protocol} =~ /iscsi/i) {
            my $type = "lvmoiscsi";
            my $target;
            my $target_iqn;

            foreach my $mpdevpath_obj ($device->get_current_path_state()->get_path_objs()) {
                $target_iqn = $mpdevpath_obj->target_iqn();
                $target = $mpdevpath_obj->target_ip_address();
                last;
            }

            $cmd = "xe sr-create content-type=user device-config:SCSIid=$scsi_id host-uuid=$host_uuid";
            $cmd .= " name-label=$sr_name shared=true type=$type device-config:target=$target device-config:targetIQN=$target_iqn";
            $execute_flag = 1;
        } elsif ($device->{protocol} =~ /fcp/i) {
            my $type = "lvmohba";
            $cmd = "xe sr-create content-type=user device-config:SCSIid=$scsi_id host-uuid=$host_uuid";
            $cmd .= " name-label=$sr_name shared=true type=$type";
            $execute_flag = 1;
        }

        if ($execute_flag == 1) {
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return ();
            }

            my $sr_uuid = $output[0];
            $sr_hash->{$sr_uuid}->{name} = $sr_name;
            $sr_hash->{$sr_uuid}->{size} = $device->{device_size};
            $sr_hash->{$sr_uuid}->{uuid} = $sr_uuid;

            $execute_flag = 0;
        }
    }

    return values %{$sr_hash};
}


##########################################################
# Subroutine Name: _iscsi_configure
# Objective:   configures iscsi on the host
##########################################################
sub _iscsi_configure {
    my ($self, %args) = @_;

    $self->sanlog(-msg => "Configuring iSCSI");

    if (defined $self->{mp} and $self->{mp_type} =~ /dmmp/i) {
        # modify session replacement timeout in the /etc/iscsi/iscsid.conf file
        $self->_change_parameter_value_in_file("node.session.timeo.replacement_timeout", "5", "/etc/iscsi/iscsid.conf");
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

##############################################################
# Subroutine Name: _power_on_vms()
# Objective:       Powers on the specified virtual machines
##############################################################
sub _power_on_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my (@waiting_to_boot_vm_list);
    my $force = 0;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for -force => $force");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    if ($force =~ /1|y|true/i) {
        $force = 1;
    } else {
        $force = 0;
    }

    # get VM information
    my $vm_info = $self->get_vm_list(-refresh => 1);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # power on each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {
        if (not exists($vm_info->{$vm_name})) {
            $self->_set_err("VM ($vm_name) is not found in the vm-list. Make sure you have provided " .
                            "correct VM name.");
            return $ClientUtil::FALSE;
        }

        # get the state of this virtual machine, if its not halted, then skip
        my $vm_state = $vm_info->{$vm_name}->{state};
        if ($vm_state !~ /halted/i) {
            $self->sanlog(-msg => "Virtual Machine ($vm_name) is not halted");
            next;
        }

        $self->sanlog(-msg => "Powering on ($vm_name).");

        # power on this virtual machine
        my ($status, @output) = $self->xecontrol(-action => "vm-start",
                                                 -vm_name => $vm_name,
                                                 -options => "force=$force");
        if ($self->error()) {
            $self->_set_err("Could not power on $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # push this virtual machine on the waiting to boot vm list
        push(@waiting_to_boot_vm_list, $vm_name);
    }

    # wait for the virtual machines to boot
    $self->_wait_for_vms_to_boot(@waiting_to_boot_vm_list);
    if ($self->error()) {
        $self->_set_err("Failed while waiting for virtual machines to boot.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

##########################################################################
# Subroutine Name: _power_on_all_vms()
# Objective:       Powers on all the virtual machines present on the host
##########################################################################
sub _power_on_all_vms {
    my ($self, %args) = @_;

    my $force = 0;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for -force => $force");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # obtain all virtual machines on the system
    my ($vm_info) = $self->get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    my @vm_list = keys %{$vm_info};

    # turn on all the virtual machines.
    $self->_power_on_vms(-vm_list => \@vm_list,
                         -force => $force);
    if ($self->error()) {
        $self->_set_err("Could not power on all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

###############################################################
# Subroutine Name: _power_off_vms()
# Objective:       Powers off the specified virtual machines
###############################################################
sub _power_off_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # get VM information
    my $vm_info = $self->get_vm_list(-refresh => 1);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # power off each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {
        if (not exists($vm_info->{$vm_name})) {
            $self->_set_err("VM ($vm_name) is not found in the vm-list. Make sure you have provided " .
                            "correct VM name.");
            return $ClientUtil::FALSE;
        }

        # get the state of this virtual machine, if its halted then skip
        my $vm_state = $vm_info->{$vm_name}->{state};
        if ($vm_state =~ /halted/i) {
            $self->sanlog(-msg => "Virtual Machine ($vm_name) is already halted");
            next;
        }

        $self->sanlog(-msg => "Powering off ($vm_name)");

        # power off this virtual machine
        $self->xeontrol(-action => "vm-shutdown",
                        -vm_name => $vm_name,
                        -options => "force=1");
        if ($self->error()) {
            $self->_set_err("Could not power off $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

############################################################################
# Subroutine Name: _power_off_all_vms()
# Objective:       Powers off all the virtual machines present on the host
############################################################################
sub _power_off_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    # obtain all virtual machines on the system
    my ($vm_info) = $self->get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    my @vm_list = keys %{$vm_info};

    # turn off all the virtual machines.
    $self->_power_off_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not power off all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

#######################################################################
# Subroutine Name: _shutdown_vms()
# Objective:       Gracefully Shutdown the specified virtual machines
#######################################################################
sub _shutdown_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # get VM information
    my $vm_info = $self->get_vm_list(-refresh => 1);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # Shutdown each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {
        if (not exists($vm_info->{$vm_name})) {
            $self->_set_err("VM ($vm_name) is not found in the vm-list. Make sure you have provided " .
                            "correct VM name.");
            return $ClientUtil::FALSE;
        }

        # get the state of this virtual machine, if its not running then skip
        my $vm_state = $vm_info->{$vm_name}->{state};
        if ($vm_state !~ /running/i) {
            $self->sanlog(-msg => "Virtual Machine ($vm_name) is not running.");
            next;
        }

        $self->sanlog(-msg => "Shutting down ($vm_name)");

        # shutdown this virtual machine
        $self->xecontrol(-action => "vm-shutdown",
                            -vm_name => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not shutdown $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

###########################################################################
# Subroutine Name: _shutdown_all_vms()
# Objective:       Shutsdown all the virtual machines present on the host
###########################################################################
sub _shutdown_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    # obtain all virtual machines on the system
    my ($vm_info) = $self->get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    my @vm_list = keys %{$vm_info};

    # shutdown all the virtual machines.
    $self->_shutdown_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not shutdown all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

##########################################################################
# Subroutine Name: _power_cycle_vms()
# Objective:       Power off/Power on the specified virtual machines
##########################################################################
sub _power_cycle_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my ($soft_shutdown) = "no";

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }  elsif (/-?soft_shutdown/) {
            if (not length($soft_shutdown = $args{$_})) {
                $self->_set_err("Invalid value for -soft_shutdown => $soft_shutdown");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "Power Cycling the following Virtual machines (graceful: $soft_shutdown):\n " .
                          join("\n ", @vm_list));

    # power off/shutdown the specified VMs
    if ($soft_shutdown =~ /no|false|0/i) {
        $self->_power_off_vms(-vm_list => \@vm_list);
    } else {
        $self->_shutdown_vms(-vm_list => \@vm_list);
    }
    if ($self->error()) {
        $self->_set_err("power_off_vms/shutdown_vms failed\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # power on the specified VMs
    $self->_power_on_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("power_on_vms failed\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Power Cycle Complete");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_cycle_all_vms()
# Objective:       Power off/Power on all virtual machines
################################################################################
sub _power_cycle_all_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # obtain all virtual machines on the system
    my ($vm_info) = $self->get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    @vm_list = keys %{$vm_info};

    # power cycle all the virtual machines
    $self->_power_cycle_vms(-vm_list => \@vm_list, %args);
    if ($self->error()) {
        $self->_set_err("Power Cycle failed\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

#########################################################################
# Subroutine Name: _wait_for_vms_to_boot()
# Objective:       This will wait until all VMs have finished booting
#########################################################################
sub _wait_for_vms_to_boot {
    my ($self, @vm_list) = @_;

    my (@ip_list) = ();

    # go through each vm and obtain its ip.
    foreach my $vm (@vm_list) {
        my $tharnhost_hsh = &main::host($vm);
        my $ip = $tharnhost_hsh->{default_ip};
        push(@ip_list, $ip);
    }

    my $num_retries = 10;
    my $num_ips = scalar(@ip_list);

    $self->sanlog(-msg => "Waiting for the following Virtual Machines to boot.\n " .
                          join("\n ", @vm_list));

    # cycle through the ip_list, pinging them until they boot up
    while (scalar(@ip_list)) {

        # if our retries are up, then break out of this loop
        if ($num_retries <= 0) {
            last;
        }

        # shift off an ip
        my $ip = shift @ip_list;
        my $vm = shift @vm_list;
        $num_ips -= 1;

        # ping the ip
        my $ping_results = &SanUtil::ping_host($ip);

        # if this ip could not be pinged, push it back on the list.
        if ($ping_results == $SanUtil::FALSE) {
            $self->sanlog(-msg => "The virtual machine '$vm' has not booted yet.");
            push (@ip_list, $ip);
            push (@vm_list, $vm);
        } else {
            $self->sanlog(-msg => "The virtual machine '$vm' is pingable.");
        }

        # if num_ips has gone to zero, then update its value to the size
        # of the ip list. Decrement the retry count, and sleep.
        if ($num_ips <= 0) {
            $num_ips = scalar(@ip_list);
            $num_retries -= 1;
            &SanUtil::nate_sleep(20);
        }
    }

    # if no ips are left in the list, all vms have booted.
    if (scalar(@ip_list) == 0) {
        $self->sanlog(-msg => "All VMs have booted.");
    } else {
        # then these vm's did not finish booting in time, display a warning
        &SanUtil::nate_log(-warn => "The following VMs took too long ".
                              "to boot: |@vm_list|. Continuing... ");
    }

    return $ClientUtil::TRUE;
}

1;
