# Copyright (c) 2010,2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to an ESXi host
## @author hrivera@netapp.com, arrance@netapp.com
## @status Private
## @pod none

# Package Name
##############
package San::SanClient::Host::ESXiHost;

# Package Requirements
######################
use POSIX;
use File::Basename;
use Config;
use San::SanClient;
use San::SanClient::ClientUtil;
use San::SanLog;
use San::SanCache;
use San::SanUtil;
use Error;
use Tharn;
use San::SanClient::Host::ESXHost;
use Data::Dumper;
use warnings;
use strict;
use Connectrec;
use Hostrec;

use vars qw( @EXPORT @ISA );

@ISA = qw(Error SanLog SanCache SanClient San::SanClient::Host::ESXHost);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new ESXHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn, $protocol, $phy_host, $vma_host, $iscsi_type, $obj_template);

    $self = {};
    # bless
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for -conn");
                return $self;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for -protocol");
                return $self;
            }
            next;
        } elsif (/-?phy_host/) {
            if (not length($phy_host = $args{$_})) {
                $self->_set_err("Invalid value for -phy_host");
                return $self;
            }
            next;
        } elsif (/-?vma_host/) {
            if (not length($vma_host = $args{$_})) {
                $self->_set_err("Invalid value for -vma_host");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        } elsif (/-?iscsi_type/) {
            $iscsi_type = $args{$_};
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    # check for required conn argument
    unless (defined $conn and ref($conn) eq "Connectrec") {
        $self->_set_err("Invalid value for argument '-conn'");
        return $self;
    }

    # check for required protocol argument
    unless (defined $protocol) {
        $self->_set_err("Invalid value or missing value for '-protocol'");
        return $self;
    }

    # check for required phy_host argument
    unless (defined $phy_host) {
        $self->_set_err("Invalid value or missing value for '-phy_host'");
        return $self;
    }

    $self->{conn} = $conn;
    $self->{protocol} = $protocol;
    $self->{phy_host} = $phy_host;
    $self->{vma_host} = $vma_host;
    $self->{NETAPP_VMA_BIN} = "/opt/netapp/vma/bin/";

    #set context
    $self->_set_vma_context();
    if ($self->error()) {
        $self->_set_err("Failed to set context", $self->errmsg());
        return $self;
    }

    # need to load config params right now so we can access them
    $self->{iscsi_type} = $iscsi_type;
    $self->_load_config_params();
    if ($self->error()) {
        return $self;
    }

    # update the vmhba adapters in the host object with those currently
    # present on the system
    $self->sanlog(-msg => "Updating vmhba adapters...");
    $self->_update_vmhba_adapters();
    if ($self->error()) {
        return $self;
    }

    # set tools bin path for ESX (really pointer to Linux binaries)
    $self->{TEST_BIN} = "/x/eng/localtest/arch/i386-redhat-linux8/bin";

    # determine if this host is capable of direct ssh access
    # by performing a very brief connection test to the physical esxi host
    my $host_obj = Hostrec->new(id => $self->{phy_host});
    my $conn_temp;
    eval {
        my $pw = $self->{phy_host_password};
        $pw =~ s/\\//g;
        $conn_temp = Connectrec->new(host => $host_obj,
                                     id => $self->{phy_host},
                                     prompts => ['~ #'],
                                     timeout => 10,
                                     username=> $self->{phy_host_username},
                                     password => $pw,
                                     conntype => "ssh-persist",
                                     shell_type => "none",
                                     template => "");
    };
    if ($@) {
        $self->sanlog(-msg => "ESXi host does not have SSH enabled, disabling direct execute flag");
        $self->{esxi_direct_execute} = 0;
    } else {
        $self->sanlog(-msg => "ESXi host has SSH enabled, enabling direct execute flag");
        $self->{esxi_direct_execute} = 1;
        $conn_temp->disconnect()
    }

    return $self;
}

################################################################################
# Subroutine Name: config_paths
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self) = @_;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
################################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: ntp_sync
# Objective:       Sync the host with the RTP NTP Time Server
#                  (timehost1.eng.netapp.com )
# Details:         See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my ($ntp_server);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Invalid value for argument '-ntp_sync'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # sync with the time server
    $self->sanlog(-msg => "Syncing the time with local NTP server: $ntp_server");

    # add the time server to the conf file and step-tickers file
    $cmd = "/usr/bin/esxcfg-ntp --add $ntp_server";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Could not add the NTP server.");
        return $ClientUtil::FALSE;
    }

    # stop the ntp service
    $cmd = "/usr/bin/esxcfg-ntp --stop";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);

    # start the ntp service
    $cmd = "/usr/bin/esxcfg-ntp --start";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Could not start the ntpd service");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Local time is: $output[0]");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rotate_logs()
# Objective:       rotate the log files for ESXi
################################################################################
sub rotate_logs {
    my ($self) = @_;

    # Noop for ESXi

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: restart_mgmt_vmware_service()
# Objective:  NOOP for ESXi
################################################################################
sub restart_mgmt_vmware_service {
    my ($self) = @_;

    # Noop for ESXi

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: vmfs_rescan()
# Objective:
################################################################################
sub vmfs_rescan {
    my ($self) = @_;

    if ($self->{vma_installed} and $self->{vma_version} =~ /5\./) {
        # force a VMFS rescan
        my $cmd = "vmkfstools --rescanvmfs";
        my ($status, @output) = $self->_vmkfstools_cmd($cmd);
        if ($status > 0) {
            $self->_set_err("Could not rescan VMFS devices");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}


################################################################################
# Subroutine Name: lun_target_reset()
# Objective:       Will do target reset or lun reset for a given Lun.
# Details:         See POD documentation at the end of this file
################################################################################
sub lun_target_reset{
    my ($self, %args) = @_;
    my (@lun_naa_ids, $reset_count, $reset_type, $conn_temp, $vma_host,);
    my ($cmd, $status, @output);
    # Default value for reset count
    $reset_count = 1;
    # parse arguments
    foreach (keys %args) {
        if (/-?lun_naa_ids/) {
            if (scalar(@{$args{$_}}) < 1) {
                $self->_set_err("Invalid value for -lun_naa_ids => $args{$_}. Expecting an array of lun naa ids");
                return $ClientUtil::FALSE;
            }
            @lun_naa_ids = @{$args{$_}};
            next;
        }
        if (/-?reset_count/) {
            if (not $args{$_} =~ /\d/ || $args{$_} < 1) {
                $self->_set_err("Invalid value for -reset_count => $args{$_}. reset_count value should be 1 or above");
                return $ClientUtil::FALSE;
            }
            $reset_count = $args{$_};
            next;
        }
        if (/-?reset_type/) {
            if (($args{$_} ne "targetreset") and ($args{$_} ne "lunreset")) {
                $self->_set_err("Invalid value for -reset_type => $args{$_}. Expected values are targetreset|lunreset");
                return $ClientUtil::FALSE;
            }
            $reset_type = $args{$_};
            next;
        }
    }
    if ($reset_type !~ /targetreset|lunreset/) {
        $self->_set_err("Invalid reset type - $_");
        return $ClientUtil::FALSE;
    }

    # verifying scalar lun naa ids
    if (!scalar (@lun_naa_ids)) {
        $self->_set_err("Missing argument -lun_naa_ids");
        return $ClientUtil::FALSE;
    }

    $vma_host = $self->{vma_host};
    # Checking if vma is configured to decide on the connection to ESXi/ESX
    if($vma_host ne "") {
        # Checking if ESX 4.x or ESXi 5.x
        if ($self->get_os_version() =~ /^4\./){
            $conn_temp=  $self->_connect($self->{phy_host}, $self->{phy_host_username}, $self->{phy_host_password}, 1800, "THARNHOST");
        }else{
            my $host_obj = Hostrec->new(id => $self->{phy_host});
            $conn_temp= Connectrec->new(host => $host_obj,id => $self->{phy_host}, username=> $self->{phy_host_username}, password => $self->{phy_host_password},conntype => "ssh-persist", shell_type => "none", template => "");
        }
    }else{
        # Connecton is already made directly to ESX/ESXi
        $conn_temp= $self->{conn}
    }

    $self->sanlog(-msg => "Starting $reset_type.");
    # LUN/target reset operation on ESX/ESXi
    for (my $count = 1 ; $count <= $reset_count; $count++){
        $self->sanlog(-msg => "Running $reset_type iteration ($count/$reset_count).");
        #Run through each Lun naa id
        foreach (@lun_naa_ids) {
            $cmd = "vmkfstools -L $reset_type /vmfs/devices/disks/" . $_;
            $self->sanlog(-msg => "$reset_type for LUN $_.");
            ($status, @output) = &ClientUtil::execute($conn_temp, $cmd);
            if ($status != 0) {
                $self->_set_err("Error running command ($cmd) : ERROR : @output");
                return $ClientUtil::FALSE;
            }
            #Sleep before next reset operation
            sleep 10;
        }
    }

    &Tharn::disconnect($conn_temp, '-ok');

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: svmotion()
# Objective:       Perform a storage VMotion operation.
################################################################################
sub svmotion {
    my ($self, %args) = @_;

    my ($datacenter, $vcenter_server, $vm_name, $destination_datastore, $vmdk_spec);
    my ($cmd, $status, @output);
    my $svmotion_timeout = 5400;

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?datacenter/) {
            if (not length($datacenter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-datacenter'");
                return {};
            }
            next;
        } elsif (/-?vcenter_server/) {
            if (not length($vcenter_server = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vcenter_server'");
                return {};
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vm_name'");
                return {};
            }
            next;
        } elsif (/-?destination_datastore/) {
            if (not length($destination_datastore = $args{$_})) {
                $self->_set_err("Invalid value for argument '-destination_datastore'");
                return {};
            }
            next;
        } elsif (/-?vmdk_spec/) {
            if (not length($vmdk_spec = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vmdk_spec'");
                return {};
            }
            next;
        } elsif (/-?timeout/) {
            if (not length($svmotion_timeout = $args{$_})) {
                $self->_set_err("Invalid value for argument '-timeout'");
                return {};
            }
            next;
        }
    }

    if (not defined($vcenter_server)) {
        $self->_set_err("-vcenter_server is a required parameter");
        return {};
    }

    if (defined($vmdk_spec) and (ref($vmdk_spec) ne "HASH" or
                                 not scalar(keys %{$vmdk_spec}))) {
        $self->_set_err("-vmdk_spec is not in a valid format, must be: hash{'disk_path} => datastore");
        return {};
    }

    # obtain vcenter credentials from the vcenter server tharnhost file
    my $hostrec = &Tharn::host($vcenter_server);
    if ($hostrec->hosttype() !~ /vcenter/) {
        $self->_set_err("$vcenter_server does not have the proper hosttype: 'vcenter'");
        return {};
    }
    my $vcenter_password = $hostrec->{default_password};
    my $vcenter_username = $hostrec->{default_username};
    my $vcenter_server_fqdn = $hostrec->{net}->[0]->{hostname};
    if (not defined($vcenter_server_fqdn) or not length($vcenter_server_fqdn)) {
        # handle non DNS environments
        $vcenter_server_fqdn = $hostrec->{default_ip};
        $cmd = "host " . $hostrec->{default_ip};
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            # then this IP is in DNS, extract the hostname from the output
            @output = grep { /pointer/ } @output;
            # store it
            ($vcenter_server_fqdn) = ($output[0] =~ /pointer\s+(\S+)\.\s*$/);
        }
    } else {
        chomp($vcenter_server_fqdn);
    }

    if (not defined($datacenter)) {
        # dynamically determine the datacenter from the vcenter
        $cmd = "vifs --listdc --server ${vcenter_server_fqdn} --username ${vcenter_username} --password ${vcenter_password} --listdc";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0 or not scalar(@output)) {
            $self->_set_err("Error running command ($cmd) (status=$status), or no output seen");
            return {};
        }

        # this may return a list of datacenters... we're just gonna grab the first one in the list and assume its that one
        $datacenter = $output[0];
    }

    # convert vm_name to its vmx file path, then convert the vmx file path to datastore format '[...] path/to/vmx'
    my $vm_config_file_path = $self->_get_vm_config_file(-vm_name => $vm_name);
    if ($self->error()) {
        return {};
    }
    my $vmx_path = $self->_convert_to_datastore_path(-file_path => $vm_config_file_path,
                                                     -format => "datastore_path",
                                                     -use_uuid => 0);
    if ($self->error()) {
        return {};
    }

    # if the user did not specify a destination datastore, then keep the virtual machine in the same place
    my ($current_datastore_name) = ($vmx_path =~ /^\s*\[(.+?)\]/);
    if (not defined($destination_datastore)) {
        $destination_datastore = $current_datastore_name;
    }

    # issue the command to perform the svmotion
    $cmd = "svmotion --server ${vcenter_server_fqdn} " .
                    "--username ${vcenter_username} " .
                    "--password ${vcenter_password} " .
                    "--datacenter ${datacenter} " .
                    "--vm '${vmx_path}':'${destination_datastore}'";

    if (ref($vmdk_spec) eq "HASH" and scalar(keys %{$vmdk_spec})) {
        # build the vmdk spec
        my $disks = "";
        foreach my $disk (keys %{$vmdk_spec}) {
            my $dstore = $vmdk_spec->{$disk};
            $disks .= "'${disk}':'${dstore}',";
        }
        chop($disks);
        $cmd .= " --disks ${disks}";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, $svmotion_timeout);
    if ($status > 0) {
        $self->_set_err("Error running command ($cmd)");
        return {};
    }

    # verify that the old vmx path no longer exists. If it does, then delete it if it has no files
    my ($old_vmx_path_dir) = $self->_parse_vmdk_file_path(-vmdk_file_path => $vmx_path);
    $cmd = "vifs --dir '$old_vmx_path_dir'";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if (not grep(/Can not list directory/i, @output)) {
        # the directory exists, delete it if it is empty
        @output = grep { !/^-+$|^\s*$|^\s*Content\s*Listing\s*$/i } @output;
        # delete the directory if the output is empty
        if (scalar(@output) == 0) {
            $cmd = "vifs --force --rm '$old_vmx_path_dir'";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Error running command ($cmd)");
                return {};
            }
        }
    }

    # construct the translation between that maps the original vmx path to the new vmx path
    my $return_hash;
    my $new_vmx_path = $vmx_path;
    $new_vmx_path =~ s/\[.+?\]/\[$destination_datastore\]/;
    $return_hash->{$vm_name}->{new_vmx_path} = $new_vmx_path;
    $return_hash->{$vm_name}->{new_datastore} = $destination_datastore;
    $return_hash->{$vm_name}->{old_vmx_path} = $vmx_path;
    $return_hash->{$vm_name}->{old_datastore} = $current_datastore_name;

    # if $vmx_path is not the same as $new_vmx_path, then
    # invalidate the vmware_cmd_l cache in sancache
    if ($vmx_path ne $new_vmx_path) {
        $self->invalidate_items_cache_data("vmware_cmd_l");
    }

    return $return_hash;
}
################################################################################
# Subroutine Name: vmotion()
# Objective:       Perform a VMotion operation.
################################################################################
sub vmotion {
    my ($self, %args) = @_;

    my ($target_host, $vm_name, $vcenter_server, $vmotion_priority);
    my ($cmd, $status, @output);
    my ($vcenter_password, $vcenter_username);
    my $timeout_in_seconds = 600;
    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?target_host/) {
            if (not length($target_host = $args{$_})) {
                $self->_set_err("Invalid value for argument '-target_host'");
                return $ClientUtil::FALSE ;
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vm_name'");
                return $ClientUtil::FALSE ;
            }
            next;
        } elsif (/-?vcenter_server/) {
            if (not length($vcenter_server = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vcenter_server'");
                return $ClientUtil::FALSE ;
            }
            next;
        } elsif (/-?vmotion_priority/) {
            if (not length($vmotion_priority = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vmotion_priority'");
                return $ClientUtil::FALSE ;
            }
            next;
        }
    }

    if (not defined($vcenter_server)or not defined($target_host)or not defined($vm_name)) {
        $self->_set_err("-vcenter_server, -target_host, -vm_name are required parameters");
        return $ClientUtil::FALSE ;
    }

    # obtain vcenter credentials from the vcenter server tharnhost file
    my $hostrec = &Tharn::host($vcenter_server);
    if ($hostrec->hosttype() !~ /vcenter/) {
        $self->_set_err("$vcenter_server does not have the proper hosttype: 'vcenter'");
        return $ClientUtil::FALSE ;
    }

    if( not defined ($vcenter_username = $hostrec->default_username()) or not defined ($vcenter_password = $hostrec->default_password())) {
        $self->_set_err("Could not fetch Username or Password  of vcenter server. Check '$vcenter_server' tharnhost file.");
        return $ClientUtil::FALSE;
    }
    my $vcenter_server_fqdn = $hostrec->{net}->[0]->{hostname};
    if (not defined($vcenter_server_fqdn) or not length($vcenter_server_fqdn)) {
        # handle non DNS environments
        $vcenter_server_fqdn = $hostrec->{default_ip};
        $cmd = "host " . $hostrec->{default_ip};
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            # then this IP is in DNS, extract the hostname from the output
            @output = grep { /pointer/ } @output;
            # store it
            ($vcenter_server_fqdn) = ($output[0] =~ /pointer\s+(\S+)\.\s*$/);
        }
    } else {
        chomp($vcenter_server_fqdn);
    }

    # Ensure that we have the fully qualified domain name
    # Examples
    # host auto-esxi5u1-02
    # auto-esxi5u1-02.gdl.englab.netapp.com has address 10.228.181.196
    # host auto-esxi5u1-02.gdl.englab.netapp.com
    # auto-esxi5u1-02.gdl.englab.netapp.com has address 10.228.181.196
    $cmd = "host " . $target_host;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 0) {
        # then this IP is in DNS, extract the hostname from the output
        @output = grep { /has address/ } @output;
        # store it
        ($target_host) = ($output[0] =~ /^(\S+)\s*has\s*address\s*\S+\s*$/);
    } # else assume the user passed the value in correctly but cant find via host.

    # issue the command to perform the vmotion
    $cmd = "esx_vmotion.pl --server ${vcenter_server_fqdn} " .
                    "--username ${vcenter_username} " .
                    "--password ${vcenter_password} " .
                    "--destinationhost ${target_host} " .
                    "--vmname ${vm_name}";

    if (defined($vmotion_priority)) {
        $cmd .= " --priority ${vmotion_priority}";
    }

    $self->sanlog(-msg => "Starting Vmotion of '$vm_name' VM.");
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, $timeout_in_seconds);
    if ($status > 0) {
        $self->_set_err("Error running command\n($cmd)\nERROR : @output");
        return $ClientUtil::FALSE;
    } else {
        $self->sanlog(-msg => "@output");
        return $ClientUtil::TRUE ;
    }
}

################################################################################
# Subroutine Name: cleanup_vm_config_file()
# Objective:       ESX only subroutine. Will remove vmdks from
#                  each vm in @vm_lists.
# Details:         See POD documentation at the end of this file
################################################################################
sub cleanup_vm_config_file {
    my ($self, %args) = @_;

    my ($status, @output, $cmd);
    my (@vm_list, $vmdk_filepath_regex_filter);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-vm_list' ".
                                "=> @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vmdk_filepath_regex_filter/) {
            if (not length($vmdk_filepath_regex_filter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vmdk_filepath_regex_filter'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not @vm_list) {
        $self->_set_err("Missing required argument list: -vm_list");
        return $ClientUtil::FALSE;
    }

    # clean up each vm
    foreach my $vm_name (@vm_list) {
        my $vmdk_disk_hash = $self->get_vmdks_from_vm(-vm_name => $vm_name,
                                                      -vmdk_filepath_regex_filter => $vmdk_filepath_regex_filter,
                                                      -return_hash => "disk_type");
        if ($self->error()) {
            $self->_set_err("Failed to get VMDKs from VM", $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # build the args hash for the remove vmdks method
        my $rdm_vmdks  = exists($vmdk_disk_hash->{RDM})  ? $vmdk_disk_hash->{RDM}  : undef;
        my $vmfs_vmdks = exists($vmdk_disk_hash->{VMFS}) ? $vmdk_disk_hash->{VMFS} : undef;
        my %args;
        if (defined($rdm_vmdks)) {
            $args{-rdm_vmdk_files} = $rdm_vmdks;
        }
        if (defined($vmfs_vmdks)) {
            $args{-vmfs_vmdk_files} = $vmfs_vmdks;
        }

        if (not defined($rdm_vmdks) and not defined($vmfs_vmdks)) {
            $self->sanlog(-msg => "There are no VMDKs to cleanup for VM '$vm_name'");
            return $ClientUtil::TRUE;
        }

        # now remove the vmdks
        $self->remove_vmdks_from_vm(-vm_name => $vm_name,
                                    %args);
        if ($self->error()) {
            $self->_set_err("Failed to remove vmdks from $vm_name", $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    get_ip_info()
# Objective: Get the IP information for this system and return hash with that
#            data
################################################################################
sub get_ip_info {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($mac, $ipv4, $ipv6, $ipv4_gateway, $ipv6_gateway, $subnet_mask, $ipv6_prefix_len);
    my ($set_interface, $mtusize, $broadcast, $iscsi_capable);

    foreach (keys %args) {
        if (/-?iscsi_capable/) {
            if (not length($iscsi_capable = $args{$_})) {
                $self->_set_err("Invalid value for -iscsi_capable");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my $ip_hash = {};

    # get all IP information from host via ifconfig
    my $cmd = $self->{NETAPP_VMA_BIN} . "esx_network_info.pl";
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Error running command ($cmd)");
        return {};
    }

    # parse the output of the network info script
    foreach my $line (@output) {
        if ($line =~ /^\s+(\S+)\.IPv4\.IP_address=(.*)$/i) {
            $ip_hash->{$1}->{ipv4} = $2;
        } elsif ($line =~ /^\s+(\S+)\.IPv4\.Subnet_mask=(.*)$/i) {
            $ip_hash->{$1}->{subnet_mask} = $2;
        } elsif ($line =~ /^\s+(\S+)\.IPv4\.Gateway=(.*)$/i) {
            $ip_hash->{$1}->{ipv4_gateway} = $2;
        } elsif ($line =~ /^\s+(\S+)\.IPv6\.IP_address\[(\d+)\]=(.*)$/i) {
            my $interface = $1;
            my $ip_num = $2;
            my $addr = $3;
            if ($addr =~ /^fe80/) {
                # skip link local addresses
                next;
            } else {
                my @arr = split(/;/, $addr);
                $ip_hash->{$interface}->{ipv6} = $arr[0];
                $ip_hash->{$interface}->{ipv6_prefix_len} = $arr[1];
            }
        } elsif ($line =~ /^\s+(\S+)\.MAC_address=(.*)$/i) {
            $ip_hash->{$1}->{mac} = $2;
        } elsif ($line =~ /^\s+(\S+)\.IPv6\.Gateway=(.*)$/i) {
            $ip_hash->{$1}->{ipv6_gateway} = $2;
        } elsif ($line =~ /^\s+(\S+)\.MTU_size=(.*)$/i) {
            $ip_hash->{$1}->{mtusize} = $2;
        }
    }

    my $host_rec = &Tharn::host($self->{phy_host});
    my $host_ip_default = $host_rec->{default_ip};

    # go through each interface discovered, and calculate subnets/broadcasts etc...
    foreach my $interface (keys %{$ip_hash}) {
        # generate the ipv4 subnet
        if (defined($ip_hash->{$interface}->{ipv4}) and
            defined($ip_hash->{$interface}->{subnet_mask})) {
            $ip_hash->{$interface}->{broadcast} =
                &SanUtil::get_ip_broadcast(-ip_address   => $ip_hash->{$interface}->{ipv4},
                                           -ipv4_netmask => $ip_hash->{$interface}->{subnet_mask});
            $ip_hash->{$interface}->{ipv4_subnet} =
                &SanUtil::get_ip_subnet(-ip_address   => $ip_hash->{$interface}->{ipv4},
                                        -ipv4_netmask => $ip_hash->{$interface}->{subnet_mask},
                                        -ip_version   => "ipv4");
        } else {
            $ip_hash->{$interface}->{broadcast} = "";
            $ip_hash->{$interface}->{ipv4_subnet} = "";
        }

        # generate the ipv6 subnet
        if (defined($ip_hash->{$interface}->{ipv6}) and
            defined($ip_hash->{$interface}->{ipv6_prefix_len})) {
            $ip_hash->{$interface}->{ipv6_subnet} =
                &SanUtil::get_ip_subnet(-ip_address   => $ip_hash->{$interface}->{ipv6},
                                        -ipv6_prefix_len => $ip_hash->{$interface}->{ipv6_prefix_len},
                                        -ip_version   => "ipv6");
        } else {
            $ip_hash->{$interface}->{ipv6_subnet} = "";
        }

        # determine if this is the management interface
        # and if this is an iscsi capable interface
        if ($ip_hash->{$interface}->{ipv4} eq $host_ip_default) {
            $ip_hash->{$interface}->{is_mgmt_interface} = 1;
            $ip_hash->{$interface}->{description} = "NIC";
            $ip_hash->{$interface}->{iscsi_capable} = "0";

        } else {
            $ip_hash->{$interface}->{is_mgmt_interface} = 0;
            $ip_hash->{$interface}->{description} = "vmkNIC";
            $ip_hash->{$interface}->{iscsi_capable} = "1";
        }
    }

    # get hardware iscsi adapter interface data as well
    if (scalar(@{$self->{vmhba_hwiscsi_adapters}})) {

        # if on ESXi5 or later, gather the network portal list data up front
        my @iscsi_adapter_info = ();
        if ($self->get_os_version() =~ /^(5|6)\./) {
            my $cmd = "esxcli iscsi networkportal list";
            ($status, @iscsi_adapter_info) = $self->_esx_execute_cmd(-cmd => "esxcli iscsi networkportal list",
                                                                     -include_host_name => 1);
            if ($status > 0 ) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return {};
            }
        }

        foreach my $vmhba_hwiscsi_adapter (@{$self->{vmhba_hwiscsi_adapters}}) {

            # skip this section if this is a broadcom iSCSI hw adapter
            # it should have been covered by the section above
            my @raw_vmhba_list = $self->get_cache_data("vmhba_raw_output");
            next if (grep(/^$vmhba_hwiscsi_adapter .*?Broadcom iSCSI/i, @raw_vmhba_list));

            my $ipv4_addr;
            my $ipv6_addr;
            my $subnet_mask;
            my $ipv6_prefix;
            my $mtusize;
            my $default_gateway;
            my $mac;

            # if on ESXi4, do it the old way
            if ($self->get_os_version() =~ /^(4)\./) {
                # get ip information
                #[root@vmshe-ibm3550-02 ~]# vmkiscsi-tool vmhba1 -i
                #IP Address   : 192.168.9.50
                #[root@vmshe-ibm3550-02 ~]# vmkiscsi-tool vmhba2 -i
                #IP Address   : 192.168.8.50
                #IP Address   : fe80::2a0:98ff:fe02:ff22
                #IP Address   : fc00:cccc:bbbb:1232:1::
                #IP Address   : ::

                $cmd = "/usr/bin/esxcfg-iscsi --network --list $vmhba_hwiscsi_adapter";
                ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
                if ($status > 0) {
                    $self->_set_err("Command ($cmd) failed with status ($status)");
                    return {};
                }

                # grep ipv4 information
                my @ipv4_output = grep { /IP Address\s+:\s+\d+\.\d+\.\d+\.\d+/i } @output;
                ($ipv4_addr) = ($ipv4_output[0] =~ /IP Address\s*:\s*(\d+\.\d+\.\d+\.\d+)/i);

                # grep ipv6 information
                my @ipv6_output = grep { /IP Address\s+:\s+.*:/i } @output;
                ($ipv6_addr) = ($ipv6_output[0] =~ /IP Address\s*:\s*(\S+):/i);
                $ipv6_prefix = 64; # hack

                # parse out the subnet line
                my @subnet_mask = grep { /SUBNET MASK\s+:\s+\d+\.\d+\.\d+\.\d+/i } @output;
                ($subnet_mask) = ($subnet_mask[0] =~ /SUBNET MASK\s*:\s*(\d+\.\d+\.\d+\.\d+)/i);

                # parse out the mtu size
                my @mtusize = grep { /MTU\sSIZE/i } @output;
                ($mtusize) = ($mtusize[0] =~ /MTU\sSIZE\s+\:\s(\d+)/i);

                # parse out the gateway
                my @default_gateway = grep { /DEFAULT\sGATEWAY\s+\:\s/i } @output;
                ($default_gateway) = ($default_gateway[0] =~ /DEFAULT\sGATEWAY\s+\:\s(\d+\.\d+\.\d+\.\d+)/i);

                # get the mac address
                $cmd = "/usr/bin/esxcfg-iscsi --pnp --list $vmhba_hwiscsi_adapter";
                ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
                if ($status > 0) {
                    $self->_set_err("Command ($cmd) failed with status ($status)");
                    return {};
                }

                # OUTPUT:
                #MAC ADDRESS SETTABLE                    :
                #MAX TRANSFER RATE (Mbps)                : 1024
                #CURRENT TRANSFER RATE (Mbps)            : 1024
                #MAX FRAME SIZE                          : 1500
                #MAC ADDRESS                             : 00:c0:dd:09:58:3a

                # parse out the mac address
                my @mac_address = grep { /MAC\s\S+\s+\:/i } @output;
                ($mac) = ($mac_address[0] =~ /MAC\sADDRESS\s+\:\s+(\S+\:\S+\:\S+\:\S+\:\S+\:\S+)/i);

            } elsif ($self->get_os_version() =~ /^(5|6)\./) {
                # parse the  @iscsi_adapter_info output
                #vmhba33
                #    Adapter: vmhba33
                #    Vmknic: vmk1
                #    MAC Address: 78:e7:d1:94:18:1c
                #    MAC Address Valid: true
                #    IPv4: 192.168.1.170
                #    IPv4 Subnet Mask: 255.255.255.0
                #    IPv6: fe80::250:56ff:fe70:31f8/64 fec0::1:1/64
                #    MTU: 1500
                #    Vlan Supported: true
                #    Vlan ID: 0
                #    Reserved Ports: 63488~65536
                #    TOE: false
                #    TSO: true
                #    TCP Checksum: false
                #    Link Up: true
                #    Current Speed: 1000
                #    Rx Packets: 1857329985
                #    Tx Packets: 2612198145
                #    NIC Driver: bnx2
                #    NIC Driver Version: 2.0.15g.v50.11-5vmw
                #    NIC Firmware Version: bc 5.2.2 NCSI 2.0.6
                #    Compliant Status: compliant
                #    NonCompliant Message:
                #    NonCompliant Remedy:
                #    Vswitch: vSwitch1
                #    PortGroup: SAN iSCSI VMkernel 1
                #    VswitchUuid:
                #    PortGroupKey:
                #    PortKey:
                #    Duplex:
                #    Path Status: active
                #
                #vmhba34
                #    Adapter: vmhba34
                #    Vmknic: vmk2
                #    MAC Address: 78:e7:d1:94:18:1e
                #    MAC Address Valid: true
                #    IPv4: 192.168.1.171
                #    IPv4 Subnet Mask: 255.255.255.0
                #    IPv6: fe80::250:56ff:fe74:e297/64 fec0::1:2/64
                #    MTU: 1500
                #    Vlan Supported: true
                #    Vlan ID: 0
                #    Reserved Ports: 63488~65536
                #    TOE: false
                #    TSO: true
                #    TCP Checksum: false
                #    Link Up: true
                #    Current Speed: 1000
                #    Rx Packets: 1852040711
                #    Tx Packets: 2607474185
                #    NIC Driver: bnx2
                #    NIC Driver Version: 2.0.15g.v50.11-5vmw
                #    NIC Firmware Version: bc 5.2.2
                #    Compliant Status: compliant
                #    NonCompliant Message:
                #    NonCompliant Remedy:
                #    Vswitch: vSwitch2
                #    PortGroup: SAN iSCSI VMkernel 2
                #    VswitchUuid:
                #    PortGroupKey:
                #    PortKey:
                #    Duplex:
                #    Path Status: active

                my @block_output = split("\n\n", join("\n", @iscsi_adapter_info));
                foreach my $block (@block_output) {
                    # skip if this is not the right adapter block
                    next if ($block !~ /\b${vmhba_hwiscsi_adapter}\b/);

                    # dont have a way to figure out the gateway yet
                    $default_gateway = "NA";

                    foreach my $line (split("\n", $block)) {
                        if ($line =~ /MAC Address:\s+(\S+)/i) {
                            $mac = $1;
                        } elsif ($line =~ /MTU:\s+(\S+)/i) {
                            $mtusize = $1;
                        } elsif ($line =~ /IPv4:\s+(\S+)/) {
                            $ipv4_addr = $1;
                        } elsif ($line =~ /IPv6:\s+(.+)/) {
                            my $ipv6_data = $1;
                            my @addrs = split(/\s+/, $ipv6_data);
                            # grab the last one. hack
                            $ipv6_addr = $addrs[-1];
                            # parse the prefix length
                            ($ipv6_prefix) = ($ipv6_addr =~ /\/(\d+)/);
                        } elsif ($line =~ /IPv4 Subnet Mask:\s+(\S+)/i) {
                            $subnet_mask = $1;
                        }
                    }
                    last;
                }

            } else {
                # abort for any other ESX version that is unsupported
                last;
            }

            # store all the values on the ip hash for this hw iscsi interface
            if (defined($ipv4_addr)) {
                $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4} = $ipv4_addr;
                if (defined($subnet_mask)) {
                    $ip_hash->{$vmhba_hwiscsi_adapter}->{subnet_mask} = $subnet_mask;
                    # generate the broadcast ip for ipv4 and store it
                    $ip_hash->{$vmhba_hwiscsi_adapter}->{broadcast} =
                        &SanUtil::get_ip_broadcast(-ip_address => $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4},
                                                   -ipv4_netmask => $ip_hash->{$vmhba_hwiscsi_adapter}->{subnet_mask});
                } else {
                    $ip_hash->{$vmhba_hwiscsi_adapter}->{subnet_mask} = "";
                }
            } else {
                $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4} = "";
                $ip_hash->{$vmhba_hwiscsi_adapter}->{subnet_mask} = "";
            }

            if (defined($ipv6_addr)) {
                $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6} = $ipv6_addr;
                $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6_prefix_len} = $ipv6_prefix;

                # calculate the ipv6 subnet
                $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6_subnet} =
                    &SanUtil::get_ip_subnet(-ip_address      => $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6},
                                            -ip_version      => "ipv6",
                                            -ipv6_prefix_len => $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6_prefix_len});
            } else {
                $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6} = "";
                $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv6_prefix_len} = "";
            }

            # indicate that this adapter is iscsi capable
            $ip_hash->{$vmhba_hwiscsi_adapter}->{iscsi_capable} = 1;

            # store these values in the ip hash
            ($ip_hash->{$vmhba_hwiscsi_adapter}->{mac})       = $mac;
            $ip_hash->{$vmhba_hwiscsi_adapter}->{description} = "HW iSCSI HBA";
            ($ip_hash->{$vmhba_hwiscsi_adapter}->{gateway})   = $default_gateway;
            $ip_hash->{$vmhba_hwiscsi_adapter}->{broadcast}   = "NA";
            ($ip_hash->{$vmhba_hwiscsi_adapter}->{mtusize})   = $mtusize;

            # update the subnet values for ipv4
            if ($ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4}) {
                $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4_subnet} =
                    &SanUtil::get_ip_subnet(-ip_address   => $ip_hash->{$vmhba_hwiscsi_adapter}->{ipv4},
                                            -ip_version   => "ipv4",
                                            -ipv4_netmask => $ip_hash->{$vmhba_hwiscsi_adapter}->{subnet_mask});
            }
        }
    }

    # if requesting only iscsi capable interfaces, filter them out here
    if (defined($iscsi_capable) and $iscsi_capable == 1) {
        my $tmp_ip_hash;
        foreach my $int (keys %{$ip_hash}) {
            if ($ip_hash->{$int}->{iscsi_capable} == 1) {
                # copy this interface's data into the tmp ip hash
                my %data = %{$ip_hash->{$int}};
                $tmp_ip_hash->{$int} = \%data;
            }
        }

        # reassign the new ip hash
        $ip_hash = $tmp_ip_hash;
    }

    # return info hash
    return ($ip_hash);
}

################################################################################
# Subroutine Name: add_vmdks_to_vm()
# Objective:       ESXi subroutine. Will assign all vmdk files in
#                  @vmfs_vmdk_files and @rdm_vmdk_files  to $vm's
#                  configuration file.
# Details:         See POD documentation at the end of this file
################################################################################
sub add_vmdks_to_vm {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @before, @after);
    my ($vm_name, @vmfs_vmdk_files, @rdm_vmdk_files, $controller);

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for argument -vm_name");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vmfs_vmdk_files/) {
            @vmfs_vmdk_files = @{$args{$_}};
            next;
        } elsif (/-?rdm_vmdk_files/) {
            @rdm_vmdk_files = @{$args{$_}};
            next;
        } elsif (/-?controller/) {
            if (not length($controller = $args{$_})) {
                $self->_set_err("Invalid value for argument -controller");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }
    # check for required parameters
    if (not defined($vm_name)) {
        $self->_set_err("Missing required parameter '-vm_name'");
        return $ClientUtil::FALSE;
    }
    if (not @vmfs_vmdk_files and
        not @rdm_vmdk_files) {
        $self->_set_err("You must provide at least one of the " .
                        "following (-vmfs_vmdk_files | -rdm_vmdk_files)");
        return $ClientUtil::FALSE;
    }
    if (not defined($controller)) {
        $self->_set_err("You must provide a -controller parameter," .
                        "use either lsilogic | buslogic | lsisas1068 | pvscsi");
        return $ClientUtil::FALSE;
    }

    # convert all the vmdk file paths to "datastore_path" format if they are not in it already
    if (scalar(@vmfs_vmdk_files)) {
        @vmfs_vmdk_files = map { $self->_convert_to_datastore_path(-file_path => $_,
                                                                   -format => "datastore_path",
                                                                   -use_uuid => 0); } @vmfs_vmdk_files;
        if ($self->error()) {
            $self->_set_err("VMFS vmdk conversion to datastore path failed", $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }
    if (scalar(@rdm_vmdk_files)) {
        @rdm_vmdk_files = map { $self->_convert_to_datastore_path(-file_path => $_,
                                                                  -format => "datastore_path",
                                                                  -use_uuid => 0); } @rdm_vmdk_files;
        if ($self->error()) {
            $self->_set_err("RDM vmdk conversion to datastore path failed", $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    my $num_vmfs_vmdks = scalar(@vmfs_vmdk_files);
    my $num_rdm_vmdks = scalar(@rdm_vmdk_files);

    $self->sanlog(-msg => "Adding VMDK Files (VMFS:$num_vmfs_vmdks, RDM:$num_rdm_vmdks) to $vm_name.\n" .
                          "VMFS:\n " . join("\n ", @vmfs_vmdk_files) . "\n" .
                          "RDM:\n "  . join("\n ", @rdm_vmdk_files));

    my $rdmlist = join(',', @rdm_vmdk_files);
    my $vmfslist = join(',', @vmfs_vmdk_files);
    my $rdm = "";
    my $vmfs = "";

    if ($num_rdm_vmdks > 0) {
        $rdm = "--rdm_files '$rdmlist'";
    }
    if ($num_vmfs_vmdks > 0) {
        $vmfs = "--vmfs_files '$vmfslist'";
    }

    my $retry_count = 5;
    my $success = 0;

    while ($retry_count > 0) {
        $cmd = $self->{NETAPP_VMA_BIN} . "esx_get_vmdks.pl --vm $vm_name";
        ($status, @before) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            if (grep(/Can\'t open session file|Session file ended unexpectedly at/i, @before) and $retry_count > 0) {
                # command failed, try it again after sleeping 10 seconds
                $retry_count--;
                &SanUtil::nate_sleep(10);
                next;
            } else {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }
        } else {
            last;
        }
    }

    $retry_count = 5;

    for (my $i = 0; $i < 2; $i++) {
        # add the disks to the vm
        while ($retry_count > 0) {
            $cmd = $self->{NETAPP_VMA_BIN} . "esx_add_vmdks.pl --vm $vm_name --controller $controller $rdm $vmfs";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                if (grep(/Can\'t open session file|Session file ended unexpectedly at|Killed/i, @output) and $retry_count > 0) {
                    # command failed, try it again after sleeping 10 seconds
                    $retry_count--;
                    &SanUtil::nate_sleep(10);
                    next;
                } else {
                    $self->_set_err("Command ($cmd) failed with status ($status)");
                    return $ClientUtil::FALSE;
                }
            } else {
                last;
            }
        }

        $retry_count = 5;

        # double check to see if the add was successful
        while ($retry_count > 0) {
            $cmd = $self->{NETAPP_VMA_BIN} . "esx_get_vmdks.pl --vm $vm_name";
            ($status, @after) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }

            for (my $i = 0; $i < scalar(@after); $i++) {
                if ($i >= scalar(@before) or $after[$i] ne $before[$i]) {
                    $success = 1;
                    last;
                } else {
                    $success = 0;
                    next;
                }
            }

            last if ($success);
            $retry_count--;
            &SanUtil::nate_sleep(10);
        }

        last if ($success);
    }

    if (not $success) {
        $self->_set_err("Failed to add VMDKs to vm $vm_name after 4 retries");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_vmdks_from_vm()
# Objective:       ESXi subroutine. Will return a list of vmdk files
#                  in use by the indicated -vm_name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vmdks_from_vm {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($cmd, $status, @output);
    my ($vm_name, @vmdk_files, $vmdk_filepath_regex_filter);
    my $return_hash = undef;
    my $normalize_file_paths = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vm_name' ".
                                "=> $vm_name");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vmdk_filepath_regex_filter/) {
            if (not length($vmdk_filepath_regex_filter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vmdk_filepath_regex_filter' ".
                                "=> $vmdk_filepath_regex_filter");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?return_hash/) {
            if (not length($return_hash = $args{$_})) {
                $self->_set_err("Invalid value for argument '-return_hash' ".
                                "=> $return_hash");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?normalize_file_paths/) {
            if (not length($normalize_file_paths = $args{$_})) {
                $self->_set_err("Invalid value for argument '-normalize_file_paths' ".
                                "=> $normalize_file_paths");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # check for required arguments
    if (not defined($vm_name)) {
        $self->_set_err("Missing required argument -vm_name");
        return {};
    }

    my $retry_count = 4;

    while ($retry_count > 0) {
        $cmd = $self->{NETAPP_VMA_BIN} . "esx_get_vmdks.pl --vm $vm_name";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            if (grep(/Can\'t open session file|Session file ended unexpectedly at/i, @output) and $retry_count > 0) {
                # command failed, try it again after sleeping 10 seconds
                $retry_count--;
                &SanUtil::nate_sleep(10);
                next;
            } else {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return (wantarray()) ? () : [];
            }
        } else {
            last;
        }
    }

    my $disk_type_hash;
    my $scsi_id_hash;

    foreach my $line (@output) {
        if ($line =~ /^\s*\(?(scsi\d+:\d+)?\)?\s*Disk\stype:\s+(VMFS|RDM),?\s+VMDK path:\s+(.+)$/i) {
            my $scsi_id = $1;
            my $disk_type = $2;
            my $path = $3;

            if (defined($vmdk_filepath_regex_filter) and $path !~ $vmdk_filepath_regex_filter) {
                next;
            } else {
                # normalize the file path if requested to
                if ($normalize_file_paths) {
                    $path = $self->_convert_to_datastore_path(-file_path => $path,
                                                              -format => "file_path",
                                                              -use_uuid => 0);
                    if ($self->error()) {
                        return (wantarray()) ? () : [];
                    }
                }

                # store the path in the three return formats
                push (@vmdk_files,                      $path);
                push (@{$disk_type_hash->{$disk_type}}, $path);
                if (defined($scsi_id)) {
                    $scsi_id_hash->{$scsi_id} = $path;
                }
            }
        }
    }

    # return the hash if requested, otherwise return the list of disks
    if (not defined($return_hash)) {
        return (wantarray()) ? @vmdk_files : \@vmdk_files;
    } elsif ($return_hash =~ /disk_type/) {
        return $disk_type_hash;
    } elsif ($return_hash =~ /scsi_id/) {
        return $scsi_id_hash;
    }
}

################################################################################
#Subroutine Name: remove_vmdks_from_vm()
#Details:         Remove the vmdks from the vm but do not destroy them.
################################################################################
sub remove_vmdks_from_vm {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd);
    my ($vm_name, @vmfs_vmdk_files, @rdm_vmdk_files);

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for argument -vm_name");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vmfs_vmdk_files/) {
            @vmfs_vmdk_files = @{$args{$_}};
            next;
        } elsif (/-?rdm_vmdk_files/) {
            @rdm_vmdk_files = @{$args{$_}};
            next;
        }
    }
    # check for required parameters
    if (not defined($vm_name)) {
        $self->_set_err("Missing required parameter '-vm_name'");
        return $ClientUtil::FALSE;
    }
    if (not @vmfs_vmdk_files and
        not @rdm_vmdk_files) {
        $self->_set_err("You must provide at least one of the " .
                        "following (-vmfs_vmdk_files | -rdm_vmdk_files)");
        return $ClientUtil::FALSE;
    }

    my $num_vmfs_vmdks = scalar(@vmfs_vmdk_files);
    my $num_rdm_vmdks = scalar(@rdm_vmdk_files);

    $self->sanlog(-msg => "Removing VMDK Files (VMFS:$num_vmfs_vmdks, RDM:$num_rdm_vmdks) from $vm_name.\n" .
                          "VMFS:\n " . join("\n ", @vmfs_vmdk_files) . "\n" .
                          "RDM:\n "  . join("\n ", @rdm_vmdk_files));

    my $rdmlist = join(',', @rdm_vmdk_files);
    my $vmfslist = join(',', @vmfs_vmdk_files);
    my $rdm = "";
    my $vmfs = "";

    if ($num_rdm_vmdks > 0) {
        $rdm = "--rdm_files '$rdmlist'";
    }
    if ($num_vmfs_vmdks > 0) {
        $vmfs = "--vmfs_files '$vmfslist'";
    }

    $cmd = $self->{NETAPP_VMA_BIN} . "esx_remove_vmdks.pl --vm $vm_name $rdm $vmfs";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: syslog_msg
# Objective:       Send message to ESX syslog
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg {
    my ($self, @message) = @_;

    # syslogging in rCLI only works for ESXi 5.0.0 and later
    if ($self->get_os_version() =~ /^(5|6)\./) {
        foreach my $msg (@message) {
            $msg =~ s/\(//g;
            $msg =~ s/\)//g;
            $msg =~ s/'//g;
            $msg =~ s/"//g;
            my $cmd = "esxcli system syslog mark --message=\"$msg\"";
            my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                            -include_host_name => 1);
            if ($status > 0 and not grep(/Unknown command/i, @output)) {
                $self->_set_err("Command ($cmd) failed with status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display_additional_info()
# Objective:       dump additional OS specific information
################################################################################
sub display_additional_info {
    my ($self) = @_;

    my $os_version = $self->get_os_version();

    my ($cmd, $status, @output);

    my $esxcli_nmp_namespace = $self->{esxcli_nmp_namespace};

    $cmd = "esxcli ${esxcli_nmp_namespace} device list";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                 -include_host_name => 1);
    if ($status > 0) {
        $self->_set_err("($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # print out hardware details
    $cmd = "esxcfg-hostops -o info";
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);

    # get list of out-of-box "async" drivers that are installed
    if ($os_version =~ /^4\./) {
        $cmd = "esxupdate query"; # -v?
        ($status, @output) = $self->_esxi_direct_execute(-cmd => $cmd,
                                                         -timeout => 240);
        $cmd = "vihostupdate -q";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);

        $cmd = "esxcli ${esxcli_nmp_namespace} satp listrules | grep NETAPP";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -timeout => 240);
    } elsif ($os_version =~ /^(5|6)\./) {
        $cmd = "esxcli software vib list";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -include_host_name => 1,
                                                     -timeout => 240);
        $cmd = "esxcli storage ${esxcli_nmp_namespace} satp rule list | grep NETAPP";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -timeout => 240);
        $cmd = "/usr/lib/vmware/vmkmgmt_keyval/vmkmgmt_keyval -a";
        ($status, @output) = $self->_esxi_direct_execute(-cmd => $cmd,
                                                         -timeout => 240);
    }

    # gather module data
    $cmd = "esx_module_info.pl";
    my @modoutput;
    ($status, @modoutput) = $self->_esx_execute_cmd(-cmd => $cmd);
    # if we detect VxDMP installed, then run some additional commands
    if (grep(/vxdmp/i, @output) or grep(/vxdmp/i, @modoutput)) {

        # list the tuning settings for this vxdmp host
        $self->_vxdmpadm_execute_cmd(-args => "gettune");

        # list all dmp nodes
        $self->_vxdmpadm_execute_cmd(-args => "list dmpnode all");

        ($status, @output) = $self->_vxdmpadm_execute_cmd(-args => "listenclosure");
        # ENCLR_NAME        ENCLR_TYPE     ENCLR_SNO            STATUS     ARRAY_TYPE LUN_COUNT
        # ==================================================================================
        # fas30500          FAS3050        1070882              CONNECTED  A/A        4
        # fas30200          FAS3020        1071883              CONNECTED  A/A        2
        foreach my $line (@output) {
            # skip bad lines
            next if ($line =~ /ENCLR_NAME|\=\=\=\=\=\=\=/);
            if ($line =~ /^\s*(\S+)\s/) {
                my $enclosure_name = $1;
                # display the attributes for this enclosure
                $self->_vxdmpadm_execute_cmd(-args => "getattr enclosure $enclosure_name all");
            }
        }
    }

    # Brocade Broadcom Intel Emulex Qlogic
    $cmd = "cat /proc/scsi/bfa/*; " .
           "cat /proc/scsi/bnx2fc/*; " .
           "cat /proc/scsi/fcoe/*; " .
           "cat /proc/scsi/lpfc*/*; " .
           "cat /proc/scsi/qla*/*";
    ($status, @output) = $self->_esxi_direct_execute(-cmd => $cmd,
                                                     -timeout => 240);
    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For ESXiHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       Clean up any orphaned devices
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    if ($self->get_os_version() =~ /^(5|6)\./) {
        # flush all the vmhba adapters
        $self->_flush_vmhba_adapters();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # remove all devices from the persistent device detach list
        # only on ESXi5
        my $corestorage_ns = $self->{esxcli_corestorage_namespace};
        my $cmd = "esxcli $corestorage_ns device detached list";
        my ($status, @dev_list) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                          -include_host_name => 1);
        if ($status > 0) {
            $self->_set_err("Failed to list persistently detached device list");
            return $ClientUtil::FALSE;
        }

        # find all the devices that are detached and remove them from the list
        # this will remove the blacklisting of all devices to this host.
        # This will allow this host to "auto attach" devices that were unmapped and remapped
        # to the host to properly handle a "semi" PDL (permament device loss) condition
        # if this is not done, then LUNs that are remapped to the host will not
        # be auto-attached after a rescan, making them inaccessible.
        foreach my $line (@dev_list) {
            # burt620758 - can't assume the first column is an "naa.****" device with VxDMP
            if ($line =~ /^(\S+)\s+(off|on)/i) {
                my $naa_device = $1;
                $cmd = "esxcli $corestorage_ns device detached remove --device=\"${naa_device}\"";
                my ($remove_status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                                       -include_host_name => 1);
                if ($status > 0) {
                    $self->_set_err("Failed to remove device '$naa_device' from the " .
                                        "persistent detach list");
                    return $ClientUtil::FALSE;
                }
            }
        }
    } else {
        # invoke the parents orphaned device cleanup routine
        # because ESXi4 and ESX4 share the same logic
        $self->SUPER::orphaned_device_cleanup();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _set_vma_context
# Objective:       This will set the context of the vma. The context refers to
#                  the physical host the vma is pointed at. If the context is
#                  not set properly commands can not be run on the target ESXi
#                  host.
################################################################################
sub _set_vma_context {
    my ($self) = @_;

    # get the tharn info for the physical host
    my $phy_host = $self->{phy_host};
    my $esxi_host = &Tharn::host($phy_host);
    my $hosttype = $esxi_host->{hosttype};
    my @prompts = [];
    my ($cmd, $status, @output);

    # store the IP address of the vMA host for later use
    my $vma_host_rec = &Tharn::host($self->{vma_host});
    my $vma_ip = $vma_host_rec->{default_ip};
    $self->{vma_ip_address} = $vma_ip;

    # make sure the physical host is of vmware hosttype
    if ($hosttype !~ /vmware/i) {
        $self->_set_err("Invalid hosttype for physical host '$phy_host', " .
                        "it must be 'hosttype=linux-vmware'");
        return $ClientUtil::FALSE;
    }

    # get the connection info for the physical host
    my $esxi_ip = $esxi_host->{default_ip};
    my $esxi_password;
    if (defined($esxi_host->{conn})) {
        $esxi_password = $esxi_host->{conn}->{ssh}->{password};
    } else {
        $esxi_password = $esxi_host->{connection_fields}->{ssh}->{password};
    }
    chomp($esxi_password);
    $esxi_password =~ s/!/\\!/g;
    my $esxi_user;
    if (defined($esxi_host->{conn})) {
        $esxi_user = $esxi_host->{conn}->{ssh}->{username};
    } else {
        $esxi_user = $esxi_host->{connection_fields}->{ssh}->{username};
    }
    chomp($esxi_user);

    my $esxi_host_fqdn = $esxi_host->{net}->[0]->{hostname};

    # we need to handle non DNS environments with the vMA
    # we also need to handle the case where the esxi_host_fqdn parameter is not a proper FQDN
    if (not defined($esxi_host_fqdn) or not length($esxi_host_fqdn) or $esxi_host_fqdn !~ /\./) {
        $esxi_host_fqdn = $esxi_ip;

        # its possible the vMA might have DNS set up, see what the vMA says
        # about this ip
        $cmd = "host $esxi_ip";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            # then this IP is in DNS, extract the hostname from the output
            @output = grep { /pointer/ } @output;
            # store it
            ($esxi_host_fqdn) = ($output[0] =~ /pointer\s+(\S+)\.\s*$/);
        }
    } else {
        chomp($esxi_host_fqdn);
    }

    # set these in the self object for easier access
    $self->{phy_host_username} = $esxi_user;
    $self->{phy_host_password} = $esxi_password;
    $self->{esxi_host_fqdn} = $esxi_host_fqdn;

    # first check to see if this is a real vMA host
    # if it is, then set context, if it isnt, then dont
    $cmd = "cat /etc/vma-release | grep vMA";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # is vMA installed?
    if ($status == 0) {
        # this is a vMA host if status == 0
        $self->{vma_installed} = $ClientUtil::TRUE;
        $self->{vma_version} = $output[0];

        # see if the server is in the vma's context list
        $cmd = "vifp listservers";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0 and not grep(/No targets have been defined/i, @output)) {
            $self->_set_err("Command ($cmd) failed. Status = ($status)");
            return $ClientUtil::FALSE;
        }

        # if this server is not in the list, add it to the vi fast pass
        if (not grep(/^\s*${esxi_host_fqdn}\s+ESX/i, @output)) {
            # If the server is not in the context list add the server to the context list and
            # then change to it.
            $cmd = "vifp addserver $esxi_host_fqdn --username $esxi_user --password $esxi_password";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Command ($cmd) failed. Status = ($status)");
                return $ClientUtil::FALSE;
            }
        }

        # update the prompts for both the IP addr and the host name
        # when we change context to the ESX Host, the prompt will
        # change, we don't know for certain whether the prompt will
        # contain the host's fqdn, or host name, or ip address
        # so add all of them as possible prompts
        my @org_prompts = $self->{conn}->prompts();
        my @new_prompts = ();
        foreach my $prompt_host ($esxi_host_fqdn, $phy_host, $esxi_ip) {
            my $new_prompt = $org_prompts[0];
            $new_prompt =~ s/_>\s+$/_\\\[$prompt_host\\\]> /;
            push(@new_prompts, $new_prompt);
        }
        $self->{conn}->prompts(@new_prompts, @org_prompts);

        # set the context to the ESX host
        $cmd = "source /opt/vmware/vma/bin/vifptarget -s $esxi_host_fqdn";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, undef,
                                                  -dont_wrap_in_parens => 1);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed. Status = ($status)");
            return $ClientUtil::FALSE;
        }
    } else {
        $self->{vma_installed} = $ClientUtil::FALSE;
    }

    # Get the build info
    $cmd = $self->{NETAPP_VMA_BIN} . "esx_remote_info.pl";
    if ($self->is_cache_data_valid("$cmd")) {
        # get the data from cache
        @output = $self->get_cache_data("$cmd");
    } else {
        # run the remote info script
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed. Status = ($status)");
            return $ClientUtil::FALSE;
        }
    }

    # update the cache for this command
    $self->store_cache_data("$cmd", @output);

    # Set the version in the log
    my ($ESX, $major_version, $build_number) = ($output[0] =~ /^\s*\S+\s(\S+)\s(\d+\.\d+\.\d+)\s+build-(\d+)/);
    $self->sanlog(-msg => "ESX Version: " . $output[0]);

    # store the version & build number on the object
    $self->{os_version} = $major_version;
    $self->{os_build_number} = $build_number;
    $self->{revision} = $major_version;
    $self->{minor_revision} = $build_number;
    $self->{version} = $ESX;

    # some esxcli namespaces are different with older versions of ESX
    if ($major_version =~ /^4/ or ($major_version =~ /^5\.0\.0/ and
                                   $build_number < 357874)) {
        $self->{esxcli_nmp_namespace} = "nmp";
    } else {
        # this is the namespace moving forward for newer versions of ESXi5
        $self->{esxcli_nmp_namespace} = "storage nmp"
    }

    # some esxcli namespaces are different with older versions of ESX
    if ($major_version =~ /^4/ or ($major_version =~ /^5\.0\.0/ and
                                   $build_number < 357874)) {
        $self->{esxcli_corestorage_namespace} = "corestorage";
    } else {
        # this is the namespace moving forward for newer versions of ESXi5
        $self->{esxcli_corestorage_namespace} = "storage core"
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _flush_vmhba_adapters
# Objective:       This will flush all vmhba adapters
#                  information.
# Details:         See POD documentation at the end of this file
################################################################################
sub _flush_vmhba_adapters {
    my ($self) = @_;

    my (@cmds) = ();

    $self->_clear_err();

    my $corestorage_ns = $self->{esxcli_corestorage_namespace};

    # rescan each fcp adapter
    foreach my $vmhba_adapter (@{$self->{vmhba_fcp_adapters}}) {
        next if ($vmhba_adapter !~ /vmhba/);
        if ($self->get_os_version() =~ /^(5|6)\./) {
            push (@cmds, "esxcli ${corestorage_ns} adapter rescan --adapter " . $vmhba_adapter);
        } else {
            push (@cmds, "esxcfg-rescan " . $vmhba_adapter);
        }
    }

    # rescan each hw iscsi adapter
    foreach my $vmhba_adapter (@{$self->{vmhba_hwiscsi_adapters}}) {
        next if ($vmhba_adapter !~ /vmhba/);
        if ($self->get_os_version() =~ /^(5|6)\./) {
            push (@cmds, "esxcli ${corestorage_ns} adapter rescan --adapter " . $vmhba_adapter);
        } else {
            push (@cmds, "esxcfg-rescan " . $vmhba_adapter);
        }
    }

    # rescan each sw iscsi adapter
    foreach my $vmhba_adapter (@{$self->{vmhba_swiscsi_adapters}}) {
        next if ($vmhba_adapter !~ /vmhba/);

        # we won't flush the sw iscsi adapter if the sw initiator is
        # not enabled
        my $cmd;
        if ($self->get_os_version() =~ /^(5|6)\./) {
            $cmd = "esxcli iscsi software get";
        } else {
            $cmd = "esxcfg-iscsi --swiscsi --list";
        }
        my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                        -include_host_name => 1);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # if its enabled, then setup the command to rescan this adapter.
        # otherwise, forget about flushing sw iscsi adapters.
        if (grep(/Software iSCSI is enabled|true/i, @output)) {
            if ($self->get_os_version() =~ /^(5|6)\./) {
                push (@cmds, "esxcli ${corestorage_ns} adapter rescan --adapter " . $vmhba_adapter);
            } else {
                push (@cmds, "esxcfg-rescan " . $vmhba_adapter);
            }
        } else {
            last;
        }
    }

    my $retry_count = 2;
    $self->sanlog(-msg => "Flushing all vmhba adapters...");
    foreach my $cmd (@cmds) {
        my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                        -include_host_name => 1);
        if ($status > 0) {
            if (grep(/Scan operation failed/i, @output) and $retry_count > 0) {
                $retry_count--;
                &SanUtil::nate_sleep(5);
                redo;
            }
            $self->_set_err("Command '$cmd' returned a bad status ($status)");
            return $ClientUtil::FALSE;
        } else {
            $retry_count = 2;
        }
    }
    $self->sanlog(-msg => "All vmhba adapters have been flushed.");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure()
# Objective:       Configure iscsi to include the targets for the filers.
#                  the IP's of all the ports on each filer.
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_configure {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets);
    # validate that iscsi parameters before continuing.
    $self->_iscsi_configure_validate_params();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # if the user does not want to configure iscsi, then leave now.
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY} != 1) {
        $self->sanlog(-msg => "iSCSI Configure: Bypassing iSCSI Configuration by user request (CONFIG_ISCSI_DISCOVERY=0)");
        return $ClientUtil::TRUE;
    }

    # parse arguments
    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        }
        if (/^-?iscsi_targets$/) {
            @targets = @{$args{$_}};
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }
    $self->sanlog(-msg => "iSCSI Configure: Configuring iSCSI");

    # if requested to do subnet matching, filter the targets list
    # so that it contains targets that are accessible by this host
    @targets = $self->_filter_targets_by_subnet(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # In C-mode, targets will have either ipv4 or ipv6 address unlike 7-mode.
    # Hence, remove the targets from the target list which don't have valid IP address
    # of IP version requested by user.
    @targets = $self->filter_targets_by_ipversion(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # check what ip version is being requested and set the ip mode on all the target objects
    if ($self->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~ /^ipv4$|^ipv6$|^auto$/i) {
        foreach my $target (@targets) {
            # set the ip mode to be either ipv4 or ipv6
            $target->set_ip_mode($self->{iscsi}->{CONFIG_ISCSI_IPVERSION});
            if ($target->error()) {
                $self->_set_err("Could not set the ip mode of target $target->{nodename} to " .
                                 $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} . "\n" .
                                "Reason: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    } else {
       $self->_set_err("Invalid value for CONFIG_ISCSI_IPVERSION: " .
                        $self->{iscsi}->{CONFIG_ISCSI_IPVERSION});
       return $ClientUtil::FALSE;
    }

    # print out a summary of what we are going to do for iSCSI Configure
    $self->_iscsi_configure_print_summary(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # deconfigure iscsi if requested to do so before configuring
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} > 0) {
        $self->_iscsi_deconfigure(-targets => \@targets,
                                  -clean_level => $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP});
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }
    # configure each iSCSI vmhba adapter
    foreach my $iscsi_adapter (@{$self->{vmhba_iscsi_adapters}}) {
        # if we are doing sw iscsi, enable the software initiator.
        if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} =~ /^software$/i) {
            &SanUtil::nate_sleep(1);
            # check if the sw iscsi initiator is enabled
            my $cmd;
            my $start_cmd;
            if ($self->get_os_version() =~ /^(5|6)\./) {
                $cmd = "esxcli iscsi software get";
                $start_cmd = "esxcli iscsi softare set --enabled=\"true\"";
            } else {
                $cmd = "/usr/bin/esxcfg-iscsi --swiscsi --list";
                $start_cmd = "/usr/bin/esxcfg-iscsi --swiscsi --enable";
            }
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                         -include_host_name => 1);
            if ($status > 0) {
                $self->_set_err("Could not query the software iscsi service");
                return $ClientUtil::FALSE;
            }

            if (grep { /Software iSCSI is not enabled|false/i } @output) {
                # then enable the iscsi software initiator

                ($status, @output) = $self->_esx_execute_cmd(-cmd => $start_cmd);
                if ($status > 0) {
                    $self->_set_err("Could not enable the software iscsi service");
                    return $ClientUtil::FALSE;
                }
            }
        }

        # configure the iSCSI option stacking features
        $self->_iscsi_configure_options(-iscsi_adapter => $iscsi_adapter);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # configure for sendtargets or static discovery depending on what was passed in.
        if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^sendtargets$/i) {
            $self->_iscsi_configure_sendtargets_discovery(-iscsi_adapter => $iscsi_adapter,
                                                          -targets => \@targets);
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
        } elsif ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^static$/i) {
            $self->_iscsi_configure_static_discovery(-iscsi_adapter => $iscsi_adapter,
                                                     -targets => \@targets);
            if ($self->error()) {
                return $ClientUtil::FALSE;
            }
        } else {
            $self->_set_err("Invalid ISCSI Discovery type '" . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} . "'");
            return $ClientUtil::FALSE;
        }

    }

    $self->sanlog(-msg => "iSCSI Configure: Complete.");
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_options()
# Objective:       Set up iSCSI Option stacking features.
################################################################################
sub _iscsi_configure_options {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, $iscsi_adapter);

    # parse arguments
    foreach (keys %args) {
        if (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring iSCSI options.");

    # set the iscsi alias for this adapter to be the hostname, command may not work in esx3.x
    # dont fatal out if it doesnt work.
    if ($self->get_os_version() =~ /^(5|6)\./) {
        $cmd = "esxcli iscsi adapter set --adapter=$iscsi_adapter --alias=" . $self->{phy_host} . "_" . $iscsi_adapter;
    } else {
        $cmd = "/usr/bin/esxcfg-iscsi --iscsiname --alias " . $self->{phy_host} . "_${iscsi_adapter} ${iscsi_adapter}";
    }
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                 -include_host_name => 1);

    # enable CHAP if requested
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /(uni|bi)directional/i) {
        # obtain the CHAP Username & CHAP Password
        my $nodename = $self->_get_iscsi_node_name_by_adapter($iscsi_adapter);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        my $chap_username;
        if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} ne "") {
            $chap_username = $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME};
        } else {
            # default the chap username to the initiator nodename of this adapter
            $chap_username = $nodename;
        }
        my $chap_out_username;
        if (defined($self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME}) and $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME} ne "") {
            $chap_out_username = $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME};
        } else {
            # default the chap out username to the initiator nodename of this adapter
            $chap_out_username = $nodename;
        }

        my $chap_password = $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD};
        my $chap_out_password = $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD};

        # if the user requested bidirectional chap then add it to the command
        my $bidirectional_params = "";
        if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /bidirectional/i) {
            if ($self->get_os_version() =~ /^(5|6)\./) {
                # enable bidirectional chap for ESXi5 and later
                $cmd = "esxcli iscsi adapter auth chap set --adapter=${iscsi_adapter} --direction=mutual --level=required --authname=$chap_out_username --secret=$chap_out_password";
                ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                             -include_host_name => 1);
                if ($status > 0) {
                    $self->_set_err("Could not enable CHAP on the iscsi adapter '$iscsi_adapter'");
                    return $ClientUtil::FALSE;
                }
            } else {
                # Older ESX versions
                $bidirectional_params = "--mchap_username $chap_out_username --mchap_password $chap_out_password --mutual";
            }
        }

        # enable CHAP on this adapter
        if ($self->get_os_version() =~ /^(5|6)\./) {
            # this will enable unidirectional chap for esxi5 and later
            $cmd = "esxcli iscsi adapter auth chap set --adapter=${iscsi_adapter} --direction=uni --level=required --authname=$chap_username --secret=$chap_password"
        } else {
            $cmd = "/usr/bin/esxcfg-iscsi --authentication --level chapRequired --method CHAP --chap_username $chap_username --chap_password $chap_password $bidirectional_params $iscsi_adapter";
        }
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -include_host_name => 1);
        if ($status > 0) {
            $self->_set_err("Could not enable CHAP on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }
    }

    # set HeaderDigest (esx4+ only)
    if ($self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST} == 1) {
        # emable header digest
        if ($self->get_os_version() =~ /^(5|6)\./) {
            $cmd = "esxcli iscsi adapter param set --adapter=$iscsi_adapter --key=HeaderDigest --value=required";
        } else {
            $cmd = "/usr/bin/esxcfg-iscsi --parameter --set \"headerDigestType=digestRequired\" $iscsi_adapter";
        }
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Could not enable Header Digest on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }
    }

    # set up Data Digest (esx4+ only)
    if ($self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST} == 1) {
        # enable data digest
        if ($self->get_os_version() =~ /^(5|6)\./) {
            $cmd = "esxcli iscsi adapter param set --adapter=$iscsi_adapter --key=DataDigest --value=required";
        } else {
            $cmd = "/usr/bin/esxcfg-iscsi --parameter --set \"dataDigestType=digestRequired\" $iscsi_adapter";
        }
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status > 0) {
            $self->_set_err("Could not enable Data Digest on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_options()
# Objective:       Deconfigure iSCSI Option stacking features.
################################################################################
sub _iscsi_deconfigure_options {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, $iscsi_adapter);

    # parse arguments
    foreach (keys %args) {
        if (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        }

    }

    # check for required parameters
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring iSCSI options.");

    # disable Header/Data Digest (esx4 only)
    # disable header digest
    if ($self->get_os_version() =~ /^(5|6)\./) {
        $cmd = "esxcli iscsi adapter param set --adapter=$iscsi_adapter --key=HeaderDigest --value=prohibited";
    } else {
        $cmd = "/usr/bin/esxcfg-iscsi --parameter --set \"headerDigestType=digestProhibited\" $iscsi_adapter";
    }
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                 -include_host_name => 1);
    # keeps returning a false BAD error (for esxi4)
    if (($status != 0) and ($output[1] ne "headerDigestSettable is 0")) {
        $self->_set_err("Could not disable Header Digest on the iscsi adapter '$iscsi_adapter'");
        return $ClientUtil::FALSE;
    }

    # disable data digest
    if ($self->get_os_version() =~ /^(5|6)\./) {
        $cmd = "esxcli iscsi adapter param set --adapter=$iscsi_adapter --key=DataDigest --value=prohibited";
    } else {
        $cmd = "/usr/bin/esxcfg-iscsi --parameter --set \"dataDigestType=digestProhibited\" $iscsi_adapter";
    }
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                 -include_host_name => 1);
    # keeps returning a false BAD error
    if (($status != 0) and ($output[1] ne "dataDigestSettable is 0")) {
        $self->_set_err("Could not disable Data Digest on the iscsi adapter '$iscsi_adapter'");
        return $ClientUtil::FALSE;
    }

    # disable CHAP
    if ($self->get_os_version() =~ /^(5|6)\./) {
        $cmd = "esxcli iscsi adapter auth chap set --adapter=$iscsi_adapter --direction=uni --default";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -include_host_name => 1);
        if ($status != 0) {
            $self->_set_err("Could not disable CHAP on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }
        $cmd = "esxcli iscsi adapter auth chap set --adapter=$iscsi_adapter --direction=mutual --default";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -include_host_name => 1);
        if ($status != 0) {
            $self->_set_err("Could not disable CHAP on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }
    } else {
        $cmd = "/usr/bin/esxcfg-iscsi --authentication --method CHAP --level chapProhibited --chap_username xxx --chap_password yyy --mchap_username xxx --mchap_password zzz $iscsi_adapter";
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
        if ($status != 0) {
            $self->_set_err("Could not disable CHAP on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_static_discovery()
# Objective:       Set up static discovery for the specified targets.
################################################################################
sub _iscsi_configure_static_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, $iscsi_adapter, @list);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring 'Static' discovery.");

    # list the static iscsi targets that are currently added
    if ($self->get_os_version() =~ /^(6)\./) {
        $cmd = "esxcli iscsi adapter discovery statictarget list --adapter=$iscsi_adapter";
    } else {
        $cmd = "/usr/bin/esxcfg-iscsi --static --list $iscsi_adapter";
    }
    ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                 -include_host_name => 1);

    # not checking for error because this command can do the
    # right thing but still have a bad return status.

    # add a target address for each filer ip in @targets
    foreach my $target (@targets) {
        my $ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not get ip from target object: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }
        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not get nodename from target object: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # check if the target already exists.
        if (grep {/^\s*ADDRESS\s*?:\s*($ip)(:3260)?/i } @list) {
            # then this target already exists, skip it
            $self->sanlog(-msg => "iSCSI Configure: Skipping static target '$ip $nodename', it is already added.");
            next;
        }

        $self->sanlog(-msg => "iSCSI Configure: Adding Static Target '$ip $nodename'");

        # add the target ip (this command can take some time...)
        if ($self->get_os_version() =~ /^(5|6)\./) {
            # for esxi5, use esxcli (because it supports ipv6 addresses)
            $cmd = "/usr/bin/esxcli iscsi adapter discovery statictarget add --adapter=\"${iscsi_adapter}\" --address=\"${ip}\" --name=\"${nodename}\"";
        } else {
            $cmd = "/usr/bin/esxcfg-iscsi --static --add --ip \"${ip}:3260\" --name \"${nodename}\" $iscsi_adapter";
        }
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -include_host_name => 1);
        if ($status != 0) {
            $self->_set_err("Could not add the ip '$ip' as a static target ".
                            "on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }

    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_static_discovery()
# Objective:       Deconfigure static discovery for the specified targets.
################################################################################
sub _iscsi_deconfigure_static_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, $iscsi_adapter, @list);
    my $clean_level = 1;

    # static discovery is only supported on esx4 and later, don't set error for non 4.x
    if ($self->get_os_version() !~ /^(4|5|6)\./) {
        return $ClientUtil::TRUE;
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring 'Static' discovery.");

    # list the static iscsi targets that are currently added\
    if ($self->get_os_version() =~ /^(6)\./) {
        $cmd = "esxcli iscsi adapter discovery statictarget list --adapter=$iscsi_adapter";
    } else {
        $cmd = "/usr/bin/esxcfg-iscsi --static --list $iscsi_adapter";
    }
    ($status, @list) = $self->_esx_execute_cmd(-cmd => $cmd,
                                               -include_host_name => 1);
    # not checking for error because this command can do the
    # right thing but still have a bad return status.


    # if told to clean up everything then remove all targets that are in the static list
    if ($clean_level == 2) {
        my $target_iqn;

        foreach my $line (@list) {
            if ($line =~ /\bNAME\s*?:\s*(\S+)/i) {
                $target_iqn = $1;
                next;
            }

            if ($line =~ /(?:ADDRESS|HOSTNAME)\s*?:\s*(\S+)(:3260)?/i) {
                my $ip = $1;

                # concat the port to the ip if its not already there
                $ip .= ":3260" if ($ip !~ /:3260$/);

                # remove the static target ip (this command can take some time...)
                $cmd = "/usr/bin/esxcfg-iscsi --static --remove --ip \"$ip\" --name \"$target_iqn\" $iscsi_adapter";
                ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                             -include_host_name => 1);
                if ($status > 0) {
                    $self->_set_err("Could not remove the ip '$ip' from the static discovery target list ".
                                    "on the iscsi adapter '$iscsi_adapter'");
                    return $ClientUtil::FALSE;
                }
            }
            if ($line =~ /^\s*${iscsi_adapter}\s+(\S+)\s+(\S+)/) {
                $target_iqn = $1;
                my $ip = $2;

                # concat the port to the ip if its not already there
                $ip .= ":3260" if ($ip !~ /:3260$/);

                # remove the static target ip (this command can take some time...)
                # for esxi5, use esxcli (because it supports ipv6 addresses)
                $cmd = "/usr/bin/esxcli iscsi adapter discovery statictarget remove --adapter=\"${iscsi_adapter}\" --address=\"${ip}\" --name=\"${target_iqn}\"";
                ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                             -include_host_name => 1);
                if ($status > 0) {
                    $self->_set_err("Could not remove the ip '$ip' from the static discovery target list ".
                                    "on the iscsi adapter '$iscsi_adapter'");
                    return $ClientUtil::FALSE;
                }
            }
        }

    } elsif ($clean_level == 1) {
        # check for required parameters
        if (not scalar(@targets)) {
            $self->_set_err("Missing or invalid '-targets' argument, required if '-clean_level' == 1");
            return $ClientUtil::FALSE;
        }

        # remove a target address for each filer ip in @targets
        foreach my $target (@targets) {
            my $target_ip = $target->get_ip();
            if ($target->error()) {
                $self->_set_err("Could not get ip from target object: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }
            my $target_iqn = $target->get_nodename();
            if ($target->error()) {
                $self->_set_err("Could not get nodename from target object: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # check if the static target exists
            if (not grep {/^\s*ADDRESS\s*?:\s*($target_ip)(:3260)?|${target_iqn}\s+${target_ip}:3260/i } @list) {
                # then this target does not exist, skip it
                next;
            }

            # remove the static target ip (this command can take some time...)
            if ($self->get_os_version() =~ /^(5|6)\./) {
                # for esxi5, use esxcli (because it supports ipv6 addresses)
                $cmd = "/usr/bin/esxcli iscsi adapter discovery statictarget remove --adapter=\"${iscsi_adapter}\" --address=\"${target_ip}:3260\" --name=\"${target_iqn}\"";
            } else {
                $cmd = "/usr/bin/esxcfg-iscsi --static --remove --ip \"${target_ip}:3260\" --name \"$target_iqn\" $iscsi_adapter";
            }
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                         -include_host_name => 1);
            if ($status > 0) {
                $self->_set_err("Could not remove the ip '$target_ip' from the static discovery target list ".
                                "on the iscsi adapter '$iscsi_adapter'");
                return $ClientUtil::FALSE;
            }
        }
    } else {
        $self->_set_err("Invalid clean_level: '$clean_level'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_sendtargets_discovery()
# Objective:       Set up sendtargets discovery for the specified targets.
################################################################################
sub _iscsi_configure_sendtargets_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, $iscsi_adapter, @list);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid '-targets' argument");
        return $ClientUtil::FALSE;
    }
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring 'Sendtargets' Discovery");

    # list the iscsi targets that are currently added
    if ($self->get_os_version() =~ /^(5|6)\./) {
        $cmd = "esxcli iscsi adapter discovery sendtarget list --adapter=$iscsi_adapter";
    } else {
        $cmd = "/usr/bin/esxcfg-iscsi --discovery --list $iscsi_adapter";
    }
    ($status, @list) = $self->_esx_execute_cmd(-cmd => $cmd,
                                               -include_host_name => 1);

    # not checking for error because this command can do the
    # right thing but still have a bad return status.

    # add a target address for each filer ip in @targets
    foreach my $target (@targets) {
        my $ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not get ip from target object: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }
        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not get nodename from target object: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }

        # check if the target already exists.
        if (grep {/^\s*DISCOVERY ADDRESS\s*?:\s*($ip)(:3260)?|^${iscsi_adapter}\s+${ip}\:3260/i } @list) {
            # then this target already exists, skip it
            $self->sanlog(-msg => "iSCSI Configure: Skipping sendtargets target '$ip $nodename', it is already added.");
            next;
        }

        $self->sanlog(-msg => "iSCSI Configure: Adding SendTargets Target '$ip $nodename'");
        # add the target ip (this command can take some time...)
        if ($self->get_os_version() =~ /^(5|6)\./) {
            # for esxi5, use esxcli (because it supports ipv6 addresses)
            $cmd = "/usr/bin/esxcli iscsi adapter discovery sendtarget add --adapter=\"${iscsi_adapter}\" --address=\"${ip}:3260\"";
        } else {
            $cmd = "/usr/bin/esxcfg-iscsi --discovery --add --ip ${ip}:3260 --name $nodename $iscsi_adapter";
        }
        ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                     -include_host_name => 1);
        if ($status != 0) {
            $self->_set_err("Could not add the ip '$ip' as a sendtargets discovery target ".
                            "on the iscsi adapter '$iscsi_adapter'");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure_sendtargets_discovery()
# Objective:       Deconfigure sendtargets discovery for the specified targets.
################################################################################
sub _iscsi_deconfigure_sendtargets_discovery {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($status, @output, $cmd, @targets, $iscsi_adapter, @list);
    my $clean_level = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?iscsi_adapter/) {
            if (not length($iscsi_adapter = $args{$_})) {
                $self->_set_err("Invalid value for argument '-iscsi_adapter' => $iscsi_adapter");
                return $ClientUtil::FALSE;
            }
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $clean_level");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not defined($iscsi_adapter)) {
        $self->_set_err("Missing or invalid '-iscsi_adapter' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring 'Sendtargets' Discovery.");

    # list the iscsi targets that are currently added
    if ($self->get_os_version() =~ /^(5|6)\./) {
        $cmd = "esxcli iscsi adapter discovery sendtarget list --adapter=$iscsi_adapter";
    } else {
        $cmd = "/usr/bin/esxcfg-iscsi --discovery --list $iscsi_adapter";
    }
    ($status, @list) = $self->_esx_execute_cmd(-cmd => $cmd,
                                               -include_host_name => 1);
    # not cecking for error because this command can do the
    # right thing but still have a bad return status.

    # if told to clean up everything then remove all targets that are in the sendtargets list
    if ($clean_level == 2) {
        foreach my $line (@list) {
            if ($line =~ /^\s*DISCOVERY ADDRESS\s*?:\s*(\S+)(:3260)?/i or
                $line =~ /^$iscsi_adapter\s+(\S+)(:3260)?/) {
                my $ip = $1;

                # concat the port if its not there already
                $ip .= ":3260" if ($ip !~ /:3260/);

                # remove the target ip
                if ($self->get_os_version() =~ /^(5|6)\./) {
                    # for esxi5, use esxcli (because it supports ipv6 addresses)
                    $cmd = "/usr/bin/esxcli iscsi adapter discovery sendtarget remove --adapter=\"${iscsi_adapter}\" --address=\"${ip}\"";
                } else {
                    $cmd = "/usr/bin/esxcfg-iscsi --discovery --remove --ip $ip $iscsi_adapter ";
                }
                ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                             -include_host_name => 1);
                # only checking for error on esx4
                if ($self->get_os_version() =~ /^4\./ and $status != 0) {
                    $self->_set_err("Failed to remove $ip from sendtargets list'");
                    return $ClientUtil::FALSE;
                }
            }
        }
    } elsif ($clean_level == 1) {
        # check for required parameters
        if (not scalar(@targets)) {
            $self->_set_err("Missing or invalid '-targets' argument, required if '-clean_level' == 1");
            return $ClientUtil::FALSE;
        }

        # remove each target address for each filer ip in @targets
        foreach my $target (@targets) {
            my $target_ip = $target->get_ip();
            if ($target->error()) {
                $self->_set_err("Could not get ip from target object: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }

            # check if the target already exists.
            if (not grep {/^\s*DISCOVERY ADDRESS\s*?:\s*($target_ip)(:3260)?|^$iscsi_adapter\s+${target_ip}(:3260)?/i } @list) {
                # then this target does not exist, skip it
                next;
            }

            # remove the target ip (this command can take some time...)
            # only checking for error on esx4
            if ($self->get_os_version() =~ /^(5|6)\./) {
                # for esxi5, use esxcli (because it supports ipv6 addresses)
                $cmd = "/usr/bin/esxcli iscsi adapter discovery sendtarget remove --adapter=\"${iscsi_adapter}\" --address=\"${target_ip}:3260\"";
            } else {
                $cmd = "/usr/bin/esxcfg-iscsi --discovery --remove --ip ${target_ip}:3260 ${iscsi_adapter}";
            }
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                         -include_host_name => 1);
            if ($self->get_os_version() =~ /^(4|5|6)\./ and $status != 0) {
                $self->_set_err("Failed to remove $target_ip from sendtargets list for iSCSI adapter ${iscsi_adapter}");
                return $ClientUtil::FALSE;
            }
        }
    } else {
        $self->_set_err("Invalid clean_level: '$clean_level'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_iscsi_node_name_by_adapter
# Objective:       this will return the nodename of the iscsi adapter
################################################################################
sub _get_iscsi_node_name_by_adapter {
    my ($self, $adapter) = @_;

    # obtain the initiator node name
    my $cmd;
    if ($self->get_os_version() =~ /^6\./) {
        $cmd = "esxcli iscsi adapter list --adapter=$adapter";
    } else {
        $cmd = "/usr/bin/esxcfg-iscsi --iscsiname --list $adapter";
    }
    my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                    -include_host_name => 1);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) failed. Status = ($status)");
        return "";
    }

    # parse out the nodename and return it
    foreach my $line (@output) {
        if ($line =~ /iSCSI Node Name\s+:\s+(iqn\.\S+)/i or
            $line =~ /^${adapter}\s+\S+\s+\S+\s+(iqn\.\S+)/) {
            return $1;
        }
    }

    # otherwise set an error
    $self->_set_err("Could not determine iSCSI Node name for adapter: $adapter");
    return "";
}

################################################################################
# Subroutine Name: _logging_configure()
# Objective:       This will configure all logging on the ESX Host
# Details:         See POD documentation at the end of this file
################################################################################
sub _logging_configure {
    my ($self) = @_;

    # set the SCSI log settings
    $self->_set_scsi_logging_settings();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # set up logging if vma is installed and this is the 5.0 vma and a 5.0 ESXi host
    if ($self->{vma_installed} and $self->{vma_version} =~ /(5|6)\./) {
        if ($self->get_os_version() =~ /^(5|6)\./ and $self->{version} =~ /ESXi/) {

            $self->sanlog(-msg => "Setting up the vMA host for remote SysLogging");

            # enable the remote syslog client in the firewall
            # esxcli network firewall ruleset set --ruleset-id="syslog" --enabled="true"
            my $cmd = "esxcli network firewall ruleset set --ruleset-id=\"syslog\" --enabled=\"true\"";
            my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                            -include_host_name => 1);
            if ($status != 0) {
                $self->_set_err("Command '$cmd' failed with status ($status)");
                return $ClientUtil::FALSE;
            }

            # establish the vma host as the syslog server
            # esxcfg-advcfg --set <ip_of_vma_host> /Syslog/global/logHost
            my $vma_ip = $self->{vma_ip_address};
            $cmd = "esxcfg-advcfg --set $vma_ip /Syslog/global/logHost";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status != 0 and not grep(/SOAP Fault/i, @output)) {
                $self->_set_err("Command '$cmd' failed with status ($status)");
                return $ClientUtil::FALSE;
            }

            # reload the syslog service
            $cmd = "esxcli system syslog reload";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd,
                                                         -include_host_name => 1);
            if ($status != 0) {
                $self->_set_err("Command '$cmd' failed with status ($status)");
                return $ClientUtil::FALSE;
            }

            # this ESXi host should now be ready to relay all syslogs to the
            # vMA host
        } elsif ($self->get_os_version() =~ /^4\./ and $self->{version} =~ /ESXi/) {
            # set up remote logging for ESXi 4.* hosts to relay logs to the vMA
            my $vma_ip = $self->{vma_ip_address};
            my $cmd = "esxcfg-advcfg --set $vma_ip /Syslog/Remote/Hostname";
            my ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status != 0 and not grep(/SOAP Fault/i, @output)) {
                $self->_set_err("Command '$cmd' failed with status ($status)");
                return $ClientUtil::FALSE;
            }
            $cmd = "esxcfg-advcfg --set 514 /Syslog/Remote/Port";
            ($status, @output) = $self->_esx_execute_cmd(-cmd => $cmd);
            if ($status != 0 and not grep(/SOAP Fault/i, @output)) {
                $self->_set_err("Command '$cmd' failed with status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _set_scsi_logging_settings()
# Objective:       Will enable logging for scsi errors in ESX
################################################################################
sub _set_scsi_logging_settings {
    my ($self) = @_;

    # set the scsi logging settings for esx4 or esx3
    # setting esx4 scsi logging may get blocked by some asynchronous lock file, retry after
    # sleeping for a bit
    my $retry_count = 0;
    while (1) {
        $self->_set_esx4_scsi_logging_settings();
        if ($self->error()) {
            $retry_count++;

            # kick out if we've exhausted the retry count
            last if ($retry_count >= 3);
            $self->sanlog(-msg => "Retrying ESX4 SCSI Logging settings ($retry_count / 2)");

            # clear the error and sleep before trying again
            $self->_clear_err();
            &SanUtil::nate_sleep(10);
        } else {
            # the command worked, kick out
            last;
        }
    }

    # check if the command failed
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _recover_from_missing_vm_error()
# Objective:       Sometimes ESX loses track of its vm's and needs it's
#                  management service reloaded.
# Details:         See POD documentation at the end of this file
################################################################################
sub _recover_from_missing_vm_error {
    my ($self) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _parse_vmdk_file_path()
# Objective: given a vmdk file path, parse out the datastore/subdirectory
#            and vmdk file name
################################################################################
sub _parse_vmdk_file_path {
    my ($self, %args) = @_;

    my $vmdk_file_path;

    # parse arguments
    foreach (keys %args) {
        if (/-?vmdk_file_path/) {
            if (not length($vmdk_file_path = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vmdk_file_path'");
                return (undef);
            }
        }
    }

    my ($datastore, $vmdk_path) = ($vmdk_file_path =~ /^\s*(\[.+?\])\s+(.+)/);
    my $sub_dir = dirname($vmdk_path);
    my $directory;
    if ($sub_dir =~ /^\.$/) {
        $directory = "$datastore";
    } else {
        $directory = "$datastore $sub_dir";
    }

    my $vmdk_filename = basename($vmdk_path);

    # escape spaces and parens from vmdk file name
    $vmdk_filename =~ s# #\\ #g;
    $vmdk_filename =~ s#\(#\\\(#g;
    $vmdk_filename =~ s#\)#\\\)#g;

    return ($directory, $vmdk_filename);
}

################################################################################
# Subroutine Name: _vxdmpadm_execute_cmd()
# Objective:       Wrapper to issue commands using vxdmpadm
################################################################################
sub _vxdmpadm_execute_cmd {
    my ($self, %args) = @_;

    my $args = "";
    my $timeout = 300;

    # parse arguments
    foreach (keys %args) {
        if (/^-?args$/) {
            $args = $args{-args};
        } elsif (/^-?timeout$/) {
            $timeout = $args{-timeout};
        }
    }

    # generate the credentials for this ESXi host
    my $esxi_host_fqdn = $self->{esxi_host_fqdn};
    my $username = $self->{phy_host_username};
    my $password = $self->{phy_host_password};
    my $credentials = "host=${esxi_host_fqdn} user=${username} passwd=${password}";
    my $cmd = "/opt/VRTSdmpadm/bin/vxdmpadm " . $credentials . " " . $args;

    # execute the command and return the output
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, $timeout);

    if (grep(/No such file or directory/, @output)) {
        $self->_set_err("Failed to execute /opt/VRTSdmpadm/bin/vxdmpadm, did you install the vxdmpadm RPM package on the vMA?");
    }

    return ($status, @output);
}
################################################################################
# Subroutine Name: unmap()
# Objective:       Runs the unmap command on different versions of esxi
################################################################################
sub unmap {
    my ($self, %args) = @_;
    my ($unmap_cmd, $unmap_directory);
    my $reclaim_percentage = 100;

    foreach (keys %args) {
        if (/^-?reclaim_percentage$/) {
            $reclaim_percentage = $args{reclaim_percentage};
        } elsif (/^-?unmap_directory$/) {
            $unmap_directory = $args{unmap_directory};
        }
    }

    if (not defined $unmap_directory) {
        $self->_set_err("Unmap directory not defined");
        return $ClientUtil::FALSE;
    }

    if ($self->get_os_version() =~ /^(5\.5)|^6\./) {
        $unmap_cmd = "esxcli storage vmfs unmap -l $unmap_directory";
    } else {
        $unmap_cmd = "vmkfstools -y $reclaim_percentage $unmap_directory";
    }

    my ($status, @output) = $self->_esxi_direct_execute(-cmd => $unmap_cmd);
    if ($status > 0) {
        if (not (grep(/VMFS3\+/, @output))) {
            $self->_set_err("Error running command esx unmap($unmap_cmd)");
            return $ClientUtil::FALSE;
        }
        &SanUtil::nate_log('-warn', "Skipped unmap because VMFS version < 3");
    }
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _esxi_direct_execute()
# Objective:       Execute a command directly(no VMA) on the ESXi Host. Creates
#                  an on-demand connection and disconnects after executing the
#                  command.
################################################################################
sub _esxi_direct_execute {
    my ($self, %args) = @_;

    my $cmd = "";
    my $retries = undef;
    my $retry_failure_type = "-warn";
    my $timeout = 10;
    my $reconnect_on_failure = 0;

    # parse arguments
    foreach (keys %args) {
        if (/^-?cmd$/) {
            $cmd = $args{-cmd};
        } elsif (/^-?timeout$/) {
            $timeout = $args{-timeout};
        } elsif (/^-?retries$/) {
            $retries = $args{-retries};
        } elsif (/^-?retry_failure_type$/) {
            $retry_failure_type = $args{-retry_failure_type};
        }
    }

    if (!$self->{esxi_direct_execute}) {
        $self->sanlog(-msg => "ESXi SSH access is not enabled. Not executing command ($cmd)");
        return (0);
    }

    my (@cmds) = split(/\s+/, $cmd);
    my $new_cmd;
    $new_cmd = shift(@cmds);

    foreach my $val (@cmds) {
        $new_cmd .= " " .$val
    }
    $cmd = $new_cmd;
    my ($status, @output);
    my $password = $self->{phy_host_password};
    $password =~ s/\\//g;

    # Attempt to create a connection to the esxi host
    my $direct_mgmt_conn = $self->_connect($self->{phy_host}, $self->{phy_host_username}, $password, $timeout, $self->{conn_type});
    if ($self->error()) {
        $self->_set_err("Could not open connection to the esxi host. Check if ssh is enabled on the host.", $self->errmsg());
        $status = "255"; #bad status
        return ($status, @output);
    }

    # if requested to retry the command on failure, use execute_retry()
    if (defined($retries)) {
        ($status, @output) = &ClientUtil::execute_retry($direct_mgmt_conn, $cmd, $timeout, $retries, $retry_failure_type);
    } else {
        # otherwise default to regular execute()
        ($status, @output) = &ClientUtil::execute($direct_mgmt_conn, $cmd, $timeout);
    }

    Tharn::disconnect($direct_mgmt_conn, '-ok');

    return ($status, @output);
}

1;


=item Event Parameters

    Required Event Parameters:
    lun_target_reset()

=item Examples

  # get all netapp test lun naa ids.first the function will collect all scsi luns in the server ( esxcfg-scsidevs -c )
  # then it will filter naa ids of devices which is of NETAPP type and return the value to lun_target_reset()
  # lun_target_reset() funtion will check for the reset type ( lun reset or target reset ) . If reset type is other than
  # lun reset or target reset function will throw error.
  # get_os_version() will check for ESX/ESXi version and will connect to ESX/ESXi host directly and will start reset operation


=back

=cut
