# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author knopp@netapp.com latif@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::HPUXHost;

# Package Requirements
######################
use File::Basename;

use San::SanClient;
use San::SanLog;
use San::SanUtil;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new HPUXHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($device_str, @devices, $obj_template);
    my ($self, $conn, $nfs_mount, $protocol);

    $self = {};
    bless $self, $class;

    # parse arguments
    foreach (keys (%args)) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXHost::new() Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Error: HPUXHost::new() Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: HPUXHost::new() Invalid value for -protocol: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Error: HPUXHost::new()", "Invalid value for argument '-conn'");
        return $self;
    }

    $self->{conn} = $conn;

    $self->{nfs_mount} = undef;

    my $cmd = "uname -m";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    foreach my $line (@output) {
        $self->sanlog(-msg_level => 3, -msg => "HP-UX Host Architecture is: $line.");
        $self->{arch} = $line;
    }

    $cmd = "uname -r";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    foreach my $line (@output) {
        $self->sanlog(-msg_level => 3, -msg => "HP-UX Host OS level is: $line.");
        $self->{os_version} = $line;
        $self->{revision} = $line;
    }

    if ($self->{arch} ne "") {
        $self->{TEST_BIN} = "/x/eng/localtest/arch/hppa-hp-hpux11/bin";
    } else {
        $self->_set_err("Error: HPUXHost::new() Invalid parameter specified for -arch");
        return $self;
    }

    # If the protocol is NFS, check to see if the NFS_CLIENT="1" in
    # /etc/rc.config.d/nfsconf file.
    if ($protocol =~ /nfs/i) {
        my $file = "/etc/rc.config.d/nfsconf";
        my $cmd = "cat $file |grep NFS_CLIENT";
        my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        @output = grep(/^NFS_CLIENT/, @output);

        if (scalar(@output) == 0) {
            $self->_set_err("Error: HPUXHost::new() The parameter NFS_CLIENT was not found\n" .
                            "in $file, please read the setup guide for HPUX and set this parameter\n" .
                            "to one '1'");
            return $self;
        } elsif (grep(/=.*0.*/, @output) != 0) {
            $self->_set_err("Error: HPUXHost::new() The parameter NFS_CLIENT was not set properly\n" .
                            "in $file, please read the setup guide for HPUX and set this parameter\n" .
                            "to one '1'");
            return $self;
        } else {
            $self->sanlog(-msg => "HP-UX NFS_CLIENT is set properly");
        }
    }

    return ($self);

}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            if (not defined ($cfmode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    # if mp_type is not set, do not validate currently
    my $mp_type = $self->{mp_type};
    if (not defined($mp_type) or $mp_type eq "") {
        $self->sanlog(-msg => "Multipathing type (mp_type) is not set. Not validating configuration");
        $self->{valid} = 1;
        return $ClientUtil::TRUE;
    }

    # get the MP stack info
    my $veritas_version    = "";
    my $veritas_version_sp = "";
    if ($mp_type =~ /mass/i) {
        $mp_type = "mass";
    } elsif ($mp_type =~ /pvlinks/i) {
        $mp_type = "pvlinks";
    } elsif ($mp_type =~ /dmp|vxvm|vrts|veritas/i) {
        $mp_type = "veritasdmp";
        # Get the MP object and get the veritas version from it.
        if (exists($self->{mp}) and ref($self->{mp}) =~ /dmp/i) {
            $veritas_version = $self->{mp}->{version};

            # See if we have an SP installed
            if (defined($self->{mp}->{sp}) and $self->{mp}->{sp} ne "") {
                $veritas_version_sp = $self->{mp}->{sp};
                $veritas_version_sp =~ s/SP//g;
            } else {
                $veritas_version_sp = 0;
            }
        } else {
            $self->_set_err("MP object either doesn't exists or is not a DMP object");
            $self->{valid} = 0;
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Unknown mp_type ($mp_type) for Linux. Failing validation");
        $self->{valid} = 0;
        return $ClientUtil::FALSE;
    }

    # Make sure the the restore daemon is running for veritas configs only
    if ($mp_type eq "veritasdmp") {
        if ($veritas_version >= 5.1) {
            # We want to look at the vxdmpadm gettune all output
            my $cmd  = "/usr/sbin/vxdmpadm gettune all";
            my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }

            # Now let's check for some settings
            # vxdmpadm gettune all
            # Tunable                           Current Value  Default Value
            # ------------------------------    -------------  -------------
            # dmp_restore_state                   enabled          enabled
            # dmp_fast_recovery                        on               on
            # dmp_log_level                             4                1
            # dmp_lun_retry_timeout                     0                0
            # dmp_path_age                            120              300
            # dmp_probe_idle_lun                      off               on
            # dmp_restore_interval                     60              300
            # dmp_restore_policy             check_disabled   check_disabled
            # dmp_monitor_fabric                       on               on
            my @errors_found = ();
            foreach my $line (@output) {
                if ($line =~ /^dmp_restore_state\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_restore_state'") if ($1 ne "enabled");
                } elsif ($line =~ /^dmp_fast_recovery\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_fast_recovery'") if ($1 ne "on");
                } elsif ($line =~ /^dmp_log_level\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_log_level'") if ($1 != 4);
                } elsif ($line =~ /^dmp_lun_retry_timeout\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_lun_retry_timeout'") if ($1 != 0);
                } elsif ($line =~ /^dmp_resore_interval\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_resore_interval'") if ($1 != 60);
                } elsif ($line =~ /^dmp_restore_policy\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_restore_policy'") if ($1 ne "check_disabled");
                } elsif ($line =~ /^dmp_probe_idle_lun\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_probe_idle_lun'") if ($1 ne "off");
                } elsif ($line =~ /^dmp_monitor_fabric\s+(\S+)\s+/) {
                    push(@errors_found, "Invalid value '$1' for 'dmp_monitor_fabric'") if ($1 ne "on");
                }
            }

            # Now check out the settings that are 5.1SP1 or later
            if ($veritas_version >= 6.0 or ($veritas_version == 5.1 and $veritas_version_sp > 0)) {
                # Tunable                           Current Value  Default Value
                # ------------------------------    -------------  -------------
                # dmp_path_age                            120              300
                foreach my $line (@output) {
                    if ($line =~ /^dmp_path_age\s+(\S+)\s+/) {
                        push(@errors_found, "Invalid value '$1' for 'dmp_path_age'") if ($1 != 120);
                    }
                }
            }

            if (scalar(@errors_found)) {
                my $errors = join("\n", @errors_found);
                $self->sanlog(-msg => $errors);
                $self->_set_err("There were one or more tuneables set improperly, please verify the " .
                                "settings in 'vxdmpadm gettune all'");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }
        } else {
            # Run vxdmpadm stat restored
            my $cmd  = "/usr/sbin/vxdmpadm stat restored";
            my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }

            # We want to check the output and look for the following:
            # "The number of daemons running : 1".  The number should not be 0.
            if (grep(/daemons running\s+\:\s+0$/, @output) > 0) {
                $self->_set_err("Veritas Restore Daemon does not appear to be running");
                $self->{valid} = 0;
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->{valid} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: configure
# Objective: configure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($cfo_mode, @targets, @iscsi_targets, @fcp_targets, %hsh1, %hsh2);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not (@targets = ${args{$_}})) {
                $self->_set_err("Invalid value for " . "-targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?iscsi_targets$/) {
            if (not (@iscsi_targets = ${args{$_}})) {
                $self->_set_err("Invalid value for " . "-iscsi_targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?fcp_targets$/) {
            if (not (@fcp_targets = ${args{$_}})) {
                $self->_set_err("Invalid value for " . "-fcp_targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?cfo_mode/) {
            if (not defined($cfo_mode = $args{$_})) {
                $self->_set_err("Error: HPUXHost::configure() Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if ($self->{protocol} =~ /iscsi/i) {
        # Configure for protocol iscsi.
        # Need to add discovery address to iscsi services.
        if (scalar(@iscsi_targets) == 0) {
            $self->_iscsi_configure(-targets => @targets);
        } else {
            $self->_iscsi_configure(-targets => @iscsi_targets);
        }
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    # Configure the MP Object
    if (defined $self->{mp} && $self->{mp_type} !~ /none/i) {
        $self->_MP_configure();
        return $ClientUtil::FALSE if ($self->error);
    }

    $self->sanlog(-msg => "SanClient Configuration is completed", -msg_level => 3);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: deconfigure
# Objective: deconfigure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub deconfigure {
    my ($self, %args) = @_;

    my (@targets);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not (@targets = ${args{$_}})) {
                $self->_set_err("Error: HPUXHost::deconfigure() Invalid value for " . "-targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # call appropriate deconfigure methods
    if ($self->{protocol} =~ /fcp/i) {
        $self->_fcp_deconfigure();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }
    if ($self->{protocol} =~ /iscsi/i) {
        $self->_iscsi_deconfigure(-targets => @targets);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_root_dev()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_root_dev {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: update_partition_table()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub update_partition_table {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  gather_data
# Objective:       Gather the data on the client. This data will mostly be
#                  used for debug purposes. Currently no-op
# Details:         See POD documentation at the end of this file
################################################################################
sub gather_data {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
################################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: collect_sysrq_dump_periodic
# Objective:       N/A for HPUX, provide for compatibility with Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_sysrq_dump_periodic {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: discover
# Objective: Discover devices and label if new.
# Details: See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self) = @_;
    my ($cmd, @devices, $status, @output, @hDevices, @no_hw_list);

    $self->_clear_err();

    $self->sanlog(-msg => "Discovering all devices");

    # execute ioscan (as root)
    $cmd = "/usr/sbin/ioscan -fn -C disk";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # determine if ioscan worked
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::discover() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # Only perform the NO_HW state check if the os version is 11iv3
    if ($self->{os_version} =~ /B\.11\.3/i) {
        # execute ioscan to collect all lun paths
        $cmd = "/usr/sbin/ioscan -fNC lunpath";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # determine if ioscan worked
        if ($status > 0) {
            $self->_set_err("Error: HPUXHost::discover() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # Go through the output of ioscan and locate any devices that are in NO_HW state.
        # Once found, run scsimgr command to validate binding of LUN path.
        @no_hw_list = grep(/NO_HW/, @output);
    }

    if (scalar(@no_hw_list) != 0) {
        $self->sanlog(-msg => "HPUXHost::discover() Found the following devices that have a state of NO_HW\n" .
                              join("\n", @no_hw_list));

        # Clear out error from host object
        $self->_clear_err();

        my @vol_subtests = ();
        my @orig_subtests = ();
        my $subtest_count = 0;

        # Calculate what 20% of the devices in the no_hw_list array is
        my $number_no_hw_list = int(.20 * scalar(@no_hw_list));

        # Go through all of the devices listed in the no_hw_list array and run scsimgr on each
        while (scalar(@no_hw_list) != 0) {

            # Grab the first group of devices and store them on a temporary array
            my @temp_working_array = ();
            for (my $i = 0; $i <= $number_no_hw_list; $i++) {
                push(@temp_working_array, shift(@no_hw_list));
            }

            # Kick off subtest for each list of devices
            my $subtest_id = "Device_cleanup_" . $subtest_count;
            my $subtest = &Tharn::subtest(sub { $self->_scsimgr(@temp_working_array) },
                                          -runid, $subtest_id, -bg);

            push @vol_subtests, $subtest;
            push @orig_subtests, $subtest;

            $subtest_count++;

            # We wait until one subtest exits before we create a new subtest
            # AS long as there are no more then 5 subtest running at once.
            while (scalar(@vol_subtests) >= 5) {
                my $subtest = shift(@vol_subtests);
                if ($subtest->is_alive()) {
                    push @vol_subtests, $subtest;
                } else {
                    # Check to see if the subtest failed, if so, set error and quit loop
                    if ($subtest->exit_status() != 0) {
                        $self->_set_err("Error: HPUXHost::discover() Failed to run _scsimgr on some of the devices (see subtests for error)");
                        last;
                    }
                }
                &SanUtil::nate_sleep(1);
            }
        }

        # Wait for all of the subtests to complete
        while (scalar(@vol_subtests) > 0) {
            my $subtest = shift(@vol_subtests);
            if ($subtest->is_alive()) {
                 push @vol_subtests, $subtest;
            } else {
                # Check to see if the subtest failed, if so, set error and quit loop
                if ($subtest->exit_status() != 0) {
                    $self->_set_err("Error: HPUXHost::discover() Failed to run _scsimgr on some of the devices (see subtests for error)");
                }
            }
            &SanUtil::nate_sleep(5);
        }

        &SanUtil::nate_wait_finish(@orig_subtests);

        # Check to see if any of the subtests had an error
        # If so, return a fail to the caller
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        # Rerun ioscan and look for any other NO_HW device states
        $cmd = "/usr/sbin/ioscan -fNC lunpath";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # determine if ioscan worked
        if ($status > 0) {
            $self->_set_err("Error: HPUXHost::discover() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # Go through the output of ioscan and locate any devices that are in NO_HW state.
        @no_hw_list = grep(/NO_HW/, @output);

        # If devices still exist with state of NO_HW, reboot the machine
        if (scalar(@no_hw_list) != 0) {
            $self->reboot();
        }

        # Rerun ioscan and look for any other NO_HW device states, if there are fail out
        $cmd = "/usr/sbin/ioscan -fNC lunpath";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # determine if ioscan worked
        if ($status > 0) {
            $self->_set_err("Error: HPUXHost::discover() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # Go through the output of ioscan and locate any devices that are in NO_HW state.
        @no_hw_list = grep(/NO_HW/, @output);

        if (scalar(@no_hw_list) != 0) {
            $self->_set_err("HPUXHost::discover() Found the following devices that have a state of NO_HW AFTER running " .
                            "scsimgr to clean devices. Possible solution is to reboot this machine\n" .
                            join("\n", @no_hw_list));
            return $ClientUtil::FALSE;
        }
    }

    # execute ioinit to create devices entries
    $cmd = "/usr/sbin/ioinit -i";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::discover() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # execute insf to create devices entries
    $cmd = "/usr/sbin/insf -e";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::discover() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # Enable the MP Object
    if (defined $self->{mp} && $self->{mp_type} !~ /none/i) {
        $self->{mp}->enable(-conn => $self->{conn});
        if ($self->{mp}->error) {
            $self->_set_err($self->{mp}->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg       => "Locate devices that are not part of a volume group",
                  -msg_level => 3);

    @devices = $self->_get_unlabeled_devices();
    return $ClientUtil::FALSE if ($self->error);

    $self->sanlog(-msg       => scalar(@devices) . " Devices are not part of a volume group",
                  -msg_level => 3);

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: format
# Objective:       Format devices with a label
# Details:         See POD documentation at the end of this file
################################################################################
sub format {
    my ($self, %args) = @_;
    my ($cmd, $devices, $label, $status, @format_cmds, @used, @output, @hDevices);

    $self->_clear_err();

    # parse arguments

    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: HPUXHost::format() Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    #  Check that required parameter '-devices' was provided
    unless (defined $devices and ref($devices) eq "ARRAY") {
        $self->_set_err("Error: HPUXHost::format() ", "Missing required argument '-device'");

        return $ClientUtil::FALSE;
    }

    @hDevices = $self->_get_devices();
    if ($self->error()) {
        my @error = $self->errmsg();
        $self->_set_err("Error: HPUXHost::format() Unable to get list of devices from system " . @error );
    }

    # iterate through the devices
    foreach my $device (@$devices) {

        my $hdev = "";
        foreach $hdev (@hDevices) {
            if ($hdev->{name} eq $device) {
                last;
            }
        }

        if ($hdev eq "") {
            $self->sanlog(-msg => "HPUXHost::format() - Unable to locate " .
                                  "$device within hDevices list from HPUXHost::_get_devices()");
            next;
        }

        $self->sanlog(-msg => "Running pvcreate on device '" . $hdev->{path} . "/$device'");

        # formulate the pvcreate command
        $cmd = "/usr/sbin/pvcreate -f " . $hdev->{path} . "/$device";

        # run pvcreate on the device
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: HPUXHost::format() ", "pvcreate returned a bad status ($status))");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: sanlun_version()
# Objective:       get sanlun version
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;
    my ($cmd, @output, $status, $version);

    $self->_clear_err();

    $cmd = "/usr/sbin/swlist | grep -i santools | awk '{print \$2}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::sanlun_version() Command ($cmd) returned a bad status ($status)");
        return "";
    }

    unless (defined $output[0] and length $output[0]) {
        $self->_set_err("Error: HPUXHost::sanlun_version(): Unable to get sanlun version");
        return "";
    }

    $version = $output[0];

    return $version;

}

###############################################################################
# Method:    info
# Objective: Get client information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;
    my ($hsh, @output);

    $self->_clear_err();

    push @output, "Host name:      " . $self->hostname(), "Host type:      " . $self->hosttype(),
      "OS version:     " . $self->os_version(),   "Sanlun version: " . $self->sanlun_version(),
      "Protocol:       " . $self->get_protocol(), "\n";

    return (wantarray()) ? @output : \@output;

}

###############################################################################
# Subroutine Name: get_mounted_device
# Objective:       returns a mounted device name when passed a filesystem
###############################################################################
sub get_mounted_device {
    my ($self, $mount) = @_;
    my ($cmd, $line, $status, @output);

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount =~ s/\//\\\//g;

    # execute the command
    $cmd = "/usr/sbin/mount | grep $mount | awk '{print \$3}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 or scalar(@output) == 0) {
        $self->_set_err("Error: HPUXHost::get_mounted_device(): Mount command failed!", @output);
        return undef;
    }

    if (scalar(@output) > 1) {
        $self->_set_err("Error: HPUXHost::get_mounted_device(): Multiple mount entries found!", @output);
        return undef;
    }

    return $output[0];
}

###############################################################################
# Method:    get_lvm_default
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lvm_default {
    my ($self) = @_;

    $self->_clear_err();

    return qw(nativelvm);
}

###############################################################################
# Method:    get_supported_protocols
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_protocols {
    my ($self) = @_;

    $self->_clear_err();

    return qw(fcp iscsi nfs);
}

###############################################################################
# Method:    get_supported_fs_types
# Objective: Get list of supported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_fs_types {
    my ($self) = @_;

    $self->_clear_err();

    return qw(vxfs jfs);
}

###############################################################################
# Method:    get_unsupported_fs
# Objective: Get list of unsupported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_unsupported_fs {
    my ($self) = @_;

    $self->_clear_err();

    return ();
}

################################################################################
# Subroutine Name: iostat
# Objective:       Format devices with a label
# Details:         See POD documentation at the end of this file
################################################################################
sub iostat {
    my ($self, %args) = @_;
    my ($interval, $cmd, $status, @output, $cnt, $count, @buf);

    $self->_clear_err();

    # Default interval is 2 seconds
    $interval = 2;

    # Default count is 1
    $count = 2;

    # parse arguments
    foreach (keys %args) {
        if (/-?interval/) {
            if (not length($interval = $args{$_})) {
                $self->_set_err("Error: HPUXHost::iostat() Invalid value for argument '-interval'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?count/) {
            if (not length($count = $args{$_})) {
                $self->_set_err("Error: HPUXHost::iostat() Missing value for argument '-count'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # iostat command for HP-UX
    $cmd = "/usr/bin/iostat $interval $count";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::iostat() ", "iostat returned a bad status ($status))");
        return (wantarray()) ? () : [];
    }

    push @output, "\n\n";
    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: get_nfs_mounts
# Objective:       get NFS mounts currently mounted on the system
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_mounts {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    my $targets = undef;
    my @nfs_mounts;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not length($targets = $args{$_})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($targets)) {
        $self->_set_err("Missing required argument '-targets'");
        return (wantarray()) ? () : [];
    }

    $cmd = "/usr/sbin/mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to execute command: " . $cmd);
        return (wantarray()) ? () : [];
    }

    # Capture the output of mount for each filer
    # and parse out the shares that are assigned to $self.
    foreach my $target (@{$targets}) {
        # Look through each line of output and determint if there is a filer in the $targets list mounted to this host
        my @mounts = grep(/$target/i, @output);
        if (scalar(@mounts) == 0) {
            next;
        }

        # Go through the found mounts and populate the return array
        foreach my $mount (@mounts) {
            # /mnt/qa/nate_1_1_hp2620_rtp11 on fas2050-rtp03:/vol/nate_1_1_hp2620_rtp11 rsize=32768,wsize=32768,NFSv3,dev=2000004 on Tue Jul 14 12:57:19 2009
            my ($nfs_mount, undef, $nfs_share, $nfs_options) = split(/\s+/, $mount);
            my $nfs_pt;
            $nfs_pt->{mount} = $nfs_mount;
            $nfs_pt->{share} = $nfs_share;
            $nfs_pt->{options} = $nfs_options;

            push(@nfs_mounts, $nfs_pt);
        }
    }

    if (scalar(@nfs_mounts) == 0) {
        $self->_set_err("Did not find any mounts associated with targets (" . join(" ", @{$targets}) . ")");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @nfs_mounts : \@nfs_mounts;
}

################################################################################
# Subroutine Name: get_nfs_share
# Objective:       get available NFS shares volumes via showmount -e
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_share {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    my $targets = undef;
    my $interfaces = undef;
    my @nfs_shares;

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not length($targets = $args{$_})) {
                $self->_set_err("Invalid value for argument '-targets'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?interfaces/) {
            if (not length($interfaces = $args{$_})) {
                $self->_set_err("Invalid value for argument '-interfaces'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Validate that targets are provided
    if (not defined($targets)) {
        $self->_set_err("Missing required argument '-targets'");
        return (wantarray()) ? () : [];
    }

    # If the interface list is not passed in, get it
    if (not defined($interfaces)) {
        $interfaces = $self->get_ip_info();
    }

    # Capture the output of showmount -e for each filer
    # and parse out the shares that are assigned to $self.
    foreach my $target (@{$targets}) {
        # Collect showmount -e
        $cmd = "/usr/sbin/showmount -e " . $target;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # Look for the two following erros, they failures.
        # showmount: <device_name>: RPC: Program not registered
        # showmount: <device_name>: RPC: Unknown host
        if (grep(/showmount\:.+RPC\:/, @output)) {
            $self->_set_err("RPC error occurred during command: " . $cmd);
            return (wantarray()) ? () : [];
        }

        if ($status > 0) {
            $self->_set_err("Unable to execute command: " . $cmd);
            return (wantarray()) ? () : [];
        }

        # Grep through the output and only get the ones that have an ip address in the interface list
        foreach my $interface (keys(%{$interfaces})) {
            my $ipv4 = $interfaces->{$interface}->{ipv4};
            my $ipv6 = $interfaces->{$interface}->{ipv6};
            my @results = ();

            # If we have an ipv4 address specified, add it to the results array
            if (defined($ipv4) and $ipv4 =~ /\d+\.\d+\.\d+\.\d+/) {
                push(@results, grep(/$ipv4/i, @output));
            }

            # If we have an ipv6 address specified, add it to the results array
            if (defined($ipv6)) {
                push(@results, grep(/$ipv6/i, @output));
            }

            if (scalar(@results) == 0) {
                $self->sanlog(-msg => "Did not find any shares for interface: $ipv4 $ipv6");
                next;
            }

            # Parse the output and populate nfs_mounts
            foreach my $line (@results) {
                # /vol/nate_1_1_hp2620_rtp11   10.60.242.103
                # /vol/nate_1_1_hprx2600_rtp14 10.60.242.105
                my ($vol) = split(/\s+/, $line);

                my $nfs_share;
                $nfs_share->{volume} = basename($vol);
                $nfs_share->{path} = $vol;
                $nfs_share->{ipv4_address} = $ipv4;
                $nfs_share->{ipv6_address} = $ipv6;
                $nfs_share->{target} = $target;

                push(@nfs_shares, $nfs_share);
            }
        }
    }

    # See if no NFS shares were found
    if (scalar(@nfs_shares) == 0) {
        $self->_set_err("No NFS shares found for targets (" . join(" ", @{$targets}) . ")");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @nfs_shares : \@nfs_shares;
}

################################################################################
# Subroutine Name: nfs_mount
# Objective:       NFS mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_mount {
    my ($self, %args) = @_;

    my ($cmd, $directio, $hide, $resource, $fs_type, $mount_pt, $options, $nodename, $vol_name, $name, $status, @output);

    $self->_clear_err();
    $fs_type  = "nfs";
    $directio = 0;
    $hide     = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?resource/) {
            if (not length($resource = $args{$_})) {
                $self->_set_err("Error: HPUXHost::nfs_mount() Invalid value for argument '-resource'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?directio/) {
            if (not length($directio = $args{$_}) or $directio !~ /1|0/i) {
                $self->_set_err("Error: HPUXHost::nfs_mount() Missing value for argument '-directio'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Error: HPUXHost::nfs_mount() Missing value for argument '-mount_pt'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?options/) {
            if (not length($options = $args{$_})) {
                $self->_set_err("Error: HPUXHost::nfs_mount() Missing value for argument '-options'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    #  Check that required parameter '-resource' was provided
    unless (defined $resource and length $resource) {
        $self->_set_err("Error: HPUXHost::nfs_mount() ", "Missing required argument '-resource'");
        return $ClientUtil::FALSE;
    }

    #  Check that required parameter '-mount_pt'
    unless (defined $mount_pt and length $mount_pt) {
        #  Use default path if not defined
        ($nodename = $resource) =~ s/(^\S+):\S+/$1/;

        #Check for empty value
        if ($nodename eq "") {
            $self->_set_err("Error: HPUXHost::nfs_mount() Unable to determine filer name from device mount '$resource'",
                            @output);
            return $ClientUtil::FALSE;
        }

        $vol_name = basename($resource);
        $mount_pt = "/mnt/qa/" . $nodename . "/" . $vol_name;
        $self->sanlog(-msg => "Using default mount pt '$mount_pt' for $resource");
    }

    #  check if the volume is already mounted
    $self->mount_table_verify($mount_pt);
    if ($self->error) {
        # It's not mounted so we need to do it now
        # Create the mount point and open up dir privs
        $cmd = "/usr/bin/mkdir -p $mount_pt; chmod -R 777 $mount_pt";

        @output = $self->execute($cmd, -hide => $hide);
        if ($self->error) {
            $self->sanlog(-msg => "WARNING: HPUXHost::nfs_mount() Unable to create mount point " .
                                  "directory for volume '$resource'" . join("\n", @output));
            $self->_clear_err();
        }

        # Add directio option if requested and it was not
        # passed in as part of the options string
        if ($directio and $options !~ /convosync=direct/i) {
            $options .= (defined $options and length $options) ? ",convosync=direct" : "convosync=direct";
        }

        # Remove option param if user passed it in as part of the
        # options string. Lib should add the tag.
        $options =~ s/-o //g;

        # Mount the volume's filesystem
        $cmd = ($options) ? "/usr/sbin/mount -o $options $resource $mount_pt" : "/usr/sbin/mount $resource $mount_pt";

        #  Create the NFS mount command.
        #mount -o rw,hard,intr  gin:/vol/vol1/ /mnt
        @output = $self->execute($cmd, -hide => $hide);
        if ($self->error) {
            $self->_set_err("Error: HPUXHost::nfs_mount() Unable to NFS mount volume", @output);
            return $ClientUtil::FALSE;
        }
    } else {
        $self->sanlog(-msg => "Warning: HPUXHost::nfs_mount() volume ($mount_pt) is already mounted");
        $self->_clear_err();
        return $ClientUtil::TRUE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: nfs_unmount()
# Objective:       unmount NFS volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_unmount() {
    my ($self, %args) = @_;

    my ($cmd, $hide, $mount_pt, $status, @output);

    $self->_clear_err();
    $hide = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?mount_pt/) {
            #  Make sure mount point exists and is not '/'.
            if (not length($mount_pt = $args{$_}) and ($args{$_} != "/")) {
                $self->_set_err("Error: HPUXHost::nfs_unmount() Missing value for argument '-mount_pt'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Check that required parameter '-mount_pt' was provided
    unless (defined $mount_pt and length $mount_pt) {
        $self->_set_err("Error: HPUXHost::nfs_mount() ", "Missing required argument '-mount_pt'");
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/sbin/umount " . $mount_pt;
    @output = $self->execute($cmd, -hide => $hide);
    if ($self->error) {
        $self->_set_err("Error: HPUXHost::nfs_unmount() " . "Unable to unmount NFS volume " . $mount_pt . " ", @output);
        return $ClientUtil::FALSE;
    }

    # Verify that $mount_pt is set and set to root
    if ($mount_pt =~ /^\s*\/\s*$/) {
        $self->_set_err("Error: HPUXHost::nfs_unmount() " . "Unable to umount ($mount_pt), can't use \"/\" as mount point");
        return $ClientUtil::FALSE;
    }

    #  Create the command to remove the NFS mount point
    $cmd = "/usr/bin/rm -rf " . $mount_pt;
    @output = $self->execute($cmd, -hide => $hide);
    if ($self->error) {
        $self->_set_err(
                     "HPUXHost::nfs_unmount() Unable to remove " . "mount point directory '" . $mount_pt . "'",
                     @output);

        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: mount_entry_add()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_add {
    my ($self, %args) = @_;
    my ($cmd, $device_mount, $device_fsck, $mount_pt, $fs_type, $mount_options, $status, @output);

    $self->_clear_err();

    #  initialize variables
    $mount_options = "log,largefiles";

    # parse arguments
    foreach (keys %args) {
        if (/-?device_mount/) {
            if (not length($device_mount = $args{$_})) {
                $self->_set_err("Error: HPUXHost::mount_entry_add() Invalid value for argument '-device_mount'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?device_fsck/) {
            if (not length($device_fsck = $args{$_})) {
                $self->_set_err("Error: HPUXHost::mount_entry_add() Missing value for argument '-device_fsck'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?mount_pt/) {
            if (not length($mount_pt = $args{$_})) {
                $self->_set_err("Error: HPUXHost::mount_entry_add() Missing value for argument '-mount_pt'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?fs_type/) {
            if (not length($fs_type = $args{$_})) {
                $self->_set_err("Error: HPUXHost::mount_entry_add() Missing value for argument '-fs_type'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?mount_options/) {
            if (not length($mount_options = $args{$_})) {
                $self->_set_err("Error: HPUXHost::mount_entry_add() Missing value for argument '-mount_options'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    #  Check that required parameter '-device' was provided
    unless (defined $device_mount and length $device_mount) {
        $self->_set_err("Error: HPUXHost::mount_entry_add() ", "Missing required argument '-device_mount'");
        return $ClientUtil::TRUE;
    }

    #  Check that required parameter '-device' was provided
    unless (defined $mount_pt and length $mount_pt) {
        $self->_set_err("Error: HPUXHost::mount_entry_add() ", "Missing required argument '-mount_pt'");
        return $ClientUtil::TRUE;
    }

    #  Check that required parameter '-device' was provided
    unless (defined $fs_type and $fs_type =~ /vxfs|hfs|nfs/i) {
        $self->_set_err("Error: HPUXHost::mount_entry_add() ", "Missing required argument '-fs_type'");
        return $ClientUtil::TRUE;
    }

    $cmd = "echo "
      . $device_mount . "\t"
      . $mount_pt . "\t"
      . $fs_type . "\t"
      . $mount_options . "\t" . "0" . "\t" . "0" . "\t"
      . ">> /etc/fstab";

    $self->sanlog(-msg => "Adding ($fs_type) fs entry to mount table");

    #  execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::mount_entry_add Unable to add mount entry to /etc/fstab file", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, $mntpt) = @_;
    my ($cmd, $mount, $status, @output);

    $self->_clear_err();

    #  Unmount the volume's filesystem
    $cmd = "/usr/sbin/umount $mntpt";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status >= 0) {
        $self->_set_err("Error:HPUXHost ::unmount_fs() Unable to unmount filesystem", @output);
        return $ClientUtil::FALSE;
    }

    # Verify that $mntpt is set and set to root
    if ($mntpt =~ /^\s*\/\s*$/) {
        $self->_set_err("Error: HPUXHost::unmount_fs() " . "Unable to umount ($mntpt), can't use \"/\" as mount point");
        return $ClientUtil::FALSE;
    }

    $self->mount_entry_delete($mntpt);

    $self->sanlog(-msg => "Removing mount point '$mntpt'");

    #  Remove directory head
    $cmd = "/usr/bin/rm -rf $self->{mount}";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::unmount_fs() Unable to remove mount point '$mount'", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: mount_entry_delete()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_delete {
    my ($self, $mount) = @_;
    my ($cmd, $status, @output);

    $self->_clear_err();

    #  Clear fstab mount entry
    $self->sanlog(-msg => "Clear /etc/fstab mount entry");

    unless (defined $mount and length($mount)) {
        $self->_set_err("Error: HPUXHost::mount_entry_delete Missing required paremeter '$mount'", @output);
        return $ClientUtil::FALSE;
    }

    $mount =~ s/\//\\\//g;

    $cmd =
        "perl -e\'"
      . "open(FH,\"< \/etc\/fstab\") or exit(1);"
      . "open(FD,\"> \/etc\/fstab_new\") or exit(1);"
      . "while(<FH>){\\\n"
      . "print FD \$_ if (\$_ !~ /$mount\\s/);" . "}\\\n"
      . "close FD;"
      . "close FH;\\\n"
      . "rename(\"/etc/fstab_new\", \"/etc/fstab\") or exit(1);" . "\' ";

    #  execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::mount_entry_delete Unable to remove mount entry from /etc/fstab file",
                        @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: mount_entry_verify
# Objective:      get_mount_entries
###############################################################################
sub mount_entry_verify {
    my ($self, $mount) = @_;
    my ($cmd, $entry, $idx, $status, $value, @values, @params, @output, @buf);

    $self->_clear_err();

    # initialize
    $entry = "";
    $idx   = 0;

    #$mount=~s/\//\\\//g;
    #  Check for mount point prifixed by tab or space
    $cmd = "cat /etc/fstab | grep -v '^#' | egrep '    $mount    | $mount'";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::mount_entry_verify(): Unable to verify mount file entry", @output);
        return (wantarray()) ? () : [];
    }

    if (scalar(@output) == 0) {
        $self->_set_err("Error: HPUXHost::mount_entry_verify(): No mount file entry exists");
        return (wantarray()) ? () : [];
    } elsif (scalar(@output) > 1) {
        $self->_set_err("Error: HPUXHost::mount_entry_verify(): Multiple mount file entries exist", @output);
        return (wantarray()) ? () : [];
    } else {
        @params = ("Device to mount:", "Mount pt:", "Fs type:", "Mount Options:");

        @values = split(/\s+/, $output[0]);

        foreach $idx (0 .. 3) {
            push @buf, $params[$idx] . " " . $values[$idx];
        }
    }

    return (wantarray()) ? @buf : \@buf;

}

###############################################################################
# Subroutine Name: mount_exists
# Objective:       validates a mount point is correctly mounted and does exist
###############################################################################
sub mount_exists {
    my ($self, $mount) = @_;

    $self->_clear_err();

    my ($cmd, $status, @output);

    $cmd = "/usr/sbin/mount | grep $mount";
    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 30);

    # Check the status of the mount command
    if ($status != 0 or scalar(@output) == 0) {
        $self->_set_err("Error: HPUXHost::mount_exists(): Unable to execute command ($cmd)");
        return $ClientUtil::FALSE;
    }

    # Join all lines in output with spaces then split on all spaces.
    # This puts all pieces as elements of the output array.
    @output = split(/\s+/, join(" ", @output));

    # Look for the mount in the output, if it exists return true
    if (grep(/^$mount$/, @output)) {
        return $ClientUtil::TRUE;
    }

    # Could not find the mount
    $self->_set_err("Error: HPUXHost::mount_exists(): Unable to locate mount point ($mount)");
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: shutdown()
# Objective:       shutdown host
# Details:         See POD documentation at the end of this file
################################################################################
sub shutdown {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "/usr/sbin/shutdown -h now";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    # sleep 1 minute to make sure shutdown occurs
    &SanUtil::nate_sleep(60);

    # Invalidate the connection on the object
    &Tharn::disconnect($self->{conn}, '-ok');

    # Ping the host every 10 seconds until it has completely shutdown
    my $alive = 1;

    for (my $cnt = 0; $cnt < 36; $cnt++) {
        my $status = &SanUtil::ping_host($self->hostname());
        if ($status == $ClientUtil::FALSE) {
            $self->sanlog(-msg => "OS is shutdown");
            $alive = 0;
            last;
        } else {
            $self->sanlog(-msg => "Ping was successful",
                          -msg_level => 7);
            &SanUtil::nate_sleep(10);
        }
    }

    # Machine is still alive, shutdown did not work
    if ($alive) {
        $self->_set_err("OS never shutdown");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reboot()
# Objective:       reboot host
# Details:         See POD documentation at the end of this file
################################################################################
sub reboot {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "/usr/sbin/reboot";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    foreach my $line (@output) {
        $self->sanlog(-msg       => "line=$line\n",
                      -indent    => 4,
                      -msg_level => 7);
    }

    # detect if the host actually rebooted or not
    $self->detect_reboot(-shutdown_timeout => 240,
                         -bootup_timeout   => 360,
                         -bootup_sleep     => 240);
    return $ClientUtil::FALSE if ($self->error);

    # wait an initial 60 seconds for telnet/rsh/ssh service to come
    # up after reboot.
    &SanUtil::nate_sleep(60);

    # reconnect to the client and retry this operation for 10 minutes
    # before giving up and erroring out.
    $self->reconnect();

    my $retries = 10;
    while ($self->error and $retries > 0) {
       $self->sanlog(-msg =>
                     "HPUXHost::reboot() reconnect failed, still ($retries) retries left");
       &SanUtil::nate_sleep(60);
       $retries--;
       $self->reconnect();
    }

    # if we still have an error after 10 minutes of
    # retries, error out.
    return $ClientUtil::FALSE if ($self->error);

    # run uname on the client so we know the connection/client are happy
    ($status, @output) = &ClientUtil::execute($self->{conn}, "uname -a");
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::reboot() Command ($cmd) " .
                        "returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    my $tempmsg = "Uname output = ";
    $tempmsg .= join "\n", @output;
    $self->sanlog(-msg => $tempmsg);

    $self->{reboot} = 0;

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: mount_table_verify
# Objective:       verify mount table
###############################################################################
sub mount_table_verify {
    my ($self, $mount) = @_;
    my ($status, @output);

    $self->_clear_err();

    chop($mount) if ($mount =~ /\/$/);

    $mount =~ s/\//\\\//g;

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, "/usr/sbin/mount");
    if ($status > 0 or scalar(@output) == 0) {
        $self->_set_err("Error: HPUXHost::mount_table_verify(): Unable to verify mount table", @output);
        return (wantarray()) ? @output : \@output;
    }

    unless (grep(/$mount /, @output)) {
        $self->_set_err("Error: HPUXHost::mount_table_verify(): Unable to verify mount table", @output);
        return (wantarray()) ? @output : \@output;
    }

    #  Need to parse the mount into some usefule structure.  For now just show
    #  that something was found.

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: config_paths
# Objective:       To run ntap_config_paths to setup proper path ordering
#                  for HP-UX.
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self) = @_;
    my ($cmd, $status, @output);

    $self->_clear_err();

    my $path = $self->get_hu_bin_path();

    if ($self->{os_version} =~ /B\.11\.3/i) {
        $self->sanlog(-msg => "Executing enable_ontap_pvlinks");

        # Change the option for the enable_ontap_pvlinks script to unset or set based
        # on mp_type. For MASS stack, unset it for everything else set it
        my $option = "set";
        if ($self->{mp_type} =~ /mass/i) {
            $option = "unset";
        }

        # This script helps in enabling the active/passive multipathing policy for
        # PVLinks with Data ONTAP LUNs. The enable_ontap_pvlinks script is
        # applicable to HP-UX 11iv3 only.
        $cmd = $path . "/enable_ontap_pvlinks $option";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        if ($status > 0 and $status ne 203) {
            $self->_set_err("Error: Vg::_create Unable to run enable_ontap_pvlinks", @output);
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "Executing ntap_config_paths");

    # Run ntap_config_paths to setup proper path ordering
    $cmd = "export PATH=\$PATH:/usr/sbin; " . $path . "/ntap_config_paths";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 1800);

    if ($status > 0 and $status ne 203) {
        $self->_set_err("Error: Vg::_create Unable to run ntap_config_paths", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}
################################################################################
# Subroutine Name: raw_device_setup
# Objective:       Do whatever we need to do to use raw devices successfully
#                  after the call to discover().
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_setup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       Do whatever we need to do to make sure the host is more
#                  happy when luns are removed.  For linux, we want to do
#                  an iscsi stop.
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For HPUXHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       When the luns are removed from the filer, we will have
#                  orphaned luns.  This causes big problems. If we do not
#                  remove these special files we often fail setup in san Libs.
#
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;
    my ($cmd, $device, $status, $hide, @output);

    $self->_clear_err();
    $hide = 0;

    # unconfigure devices
    $cmd = "/usr/sbin/ioscan -fn";
    &ClientUtil::execute($self->{conn}, $cmd);

    # find all instances of unused NETAPP special files and, get the hardware path
    $cmd = "/usr/sbin/ioscan -fC disk | grep NETAPP | grep NO_HW | awk '{print  \$3}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # just ignore status and plow thru all special files
    foreach $device (@output) {
        # remove all instances of unused NETAPP special files
        $cmd = "/usr/sbin/rmsf -H $device";
        &ClientUtil::execute($self->{conn}, $cmd);
    }

    # Only run rmsf -x on HPUX 11.iv3
    if ($self->{os_version} =~ /B\.11\.3/i) {
        # Run rmsf -x to clear all the stale dsf's
        $cmd = "/usr/sbin/rmsf -x";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: HPUXHost::orphaned_device_cleanup - Could not execute $cmd",
                            @output);
            return $ClientUtil::FALSE;
        }
    }

    # perform rmdisk on left over VRTS devices (or rmdev on
    # hdisk devcies will fail)
    if ($self->{mp_type} =~ /dmp|vxvm/i) {
        $self->sanlog(-msg => "Removing orphaned vxdisk devices");

        $cmd = "/usr/sbin/vxdisk -o alldgs list";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # Example Output
        # DEVICE       TYPE            DISK         GROUP        STATUS
        # das_fcp_0_0  auto:none       -            -            online invalid
        # das_fcp_0_1  auto:none       -            -            online invalid
        # das_iscs0_0  auto:none       -            -            online invalid
        # das_iscs0_1  auto:none       -            -            online invalid
        foreach my $line (@output) {
            next if ($line =~  /^DEVICE/ or $line =~ /^Disk_/);

            my @disk_info = split(/\s+/, $line);
            my $dev   = $disk_info[0];
            my $disk  = $disk_info[2];
            my $group = $disk_info[3];
            # if group and disk are "-", no DG's using this.
            # also check to make sure device is not "-" which could
            # be the case in faulted state.
            if ($disk eq "-" and $group eq "-" and $dev ne "-") {
                $cmd = "/usr/sbin/vxdisk rm $dev";
                my ($status, @rm_output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Command ($cmd) returned a bad status ($status)");
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: ntp_sync
# Objective:       Sync the host with the RTP NTP Time Server
#                  (timehost1.eng.netapp.com )
# Details:         See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my ($ntp_server, $ntp_time_zone);

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: HPUXHost::ntp_sync() Invalid value for argument '-ntp_sync'");
                return $ClientUtil::FALSE;
            }
            next;

        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: HPUXHostHost::ntp_sync() Invalid value for argument '-ntp_time_zone'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "Setting the Time Zone (TZ) to $ntp_time_zone");
    $cmd = "TZ=$ntp_time_zone; export TZ";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::ntp_sync Could not execute $cmd",
                        @output);
        return ($ClientUtil::FALSE);
    }

    $self->sanlog(-msg => "Syncing the time with local NTP server: $ntp_server");
    $cmd = "/usr/sbin/ntpdate $ntp_server";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    # We dont want to fail if ntpd is already running
    if ($status > 0 and (not grep(/the\s+NTP\s+socket\s+is\s+in\s+use/i, @output))) {
        $self->_set_err("Error: HPUXHost::ntp_sync Could not execute $cmd", @output);
        return $ClientUtil::FALSE;
    }

    $cmd = "date";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::ntp_sync Could not execute $cmd", @output);
        return $ClientUtil::FALSE;
    } else {
        $self->sanlog(-msg => "Local time is: $output[0]");
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: syslog_msg
# Objective:       Send message to HP-UX syslog
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg {
    my ($self, @message) = @_;

    my ($cmd, @output, $status);

    foreach my $msg (@message) {
        $cmd = "logger -t \"**NATE**\" -p user.error \"$msg\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 10);
        if ($status > 0) {
            $self->_set_err("Error: HPUXHost::syslog_msg(): Error running logger command ($cmd)");
            return $ClientUtil::FALSE;
        }
    }
}

################################################################################
# Method:    get_system_stats()
# Objective: Get the cpu/memory usage percentage over a specified interval
################################################################################
sub get_system_stats {
    my ($self, $interval) = @_;

    my $system_stats = {};
    my $cpu_samples = 0;
    my $mem_samples = 0;
    my ($status, @mem_output1, @mem_output2, @cpu_output, @mem_output);
    my $cpu_flag = 0;

    # collect the first memory sample
    my $cmd = "/usr/sbin/swapinfo -M";
    ($status, @mem_output1) = &ClientUtil::execute($self->{conn}, $cmd, $interval + 30);
    if ($status > 1) {
        $self->_set_err("Error: HPUXHost::get_system_stats(): failed to collect a memory usage sample.");
        return $ClientUtil::FALSE;
    }

    # collect system statistics for $interval length, it is run asynchronously,
    # so we need to sleep during the process.
    $cmd = "/usr/bin/vmstat 1 $interval";
    ($status, @cpu_output) = &ClientUtil::execute($self->{conn}, $cmd, $interval + 30);
    # not checking for error, since $status could be greater than 0 yet the command successfully ran.
    if (scalar(@cpu_output) == 0) {
        $self->_set_err("Error: HPUXHost::get_system_stats(): failed to collect cpu usage samples.");
        return $ClientUtil::FALSE;
    }

    # collect the second memory sample
    $cmd = "/usr/sbin/swapinfo -M";
    ($status, @mem_output2) = &ClientUtil::execute($self->{conn}, $cmd, $interval + 30);
    if ($status > 1) {
        $self->_set_err("Error: HPUXHost::get_system_stats(): failed to collect a memory usage sample.");
        return $ClientUtil::FALSE;
    }

    # @cpu_output looks like this:
    # vmstat 1 3
    #         procs           memory                   page                              faults       cpu
    #    r     b     w      avm    free   re   at    pi   po    fr   de    sr     in     sy    cs  us sy id
    #    1     1     0    78743  267995    0    0     0    0     0    0     0    435   3960    47   0  0 99
    #    1     1     0    78743  267995    0    0     0    0     0    0     0    436   2456    38   0  0 100
    #    1     1     0    78743  267911    0    0     0    0     0    0     0    443   1985    37   0  0 100

    # @mem_output1, @mem_output2 could look like this:
    # swapinfo -M
    #             Kb      Kb      Kb   PCT
    #TYPE      AVAIL    USED    FREE  USED
    #memory  1984836  569676 1415160   29%

    # parse out the CPU output and tally up the average
    foreach my $line (@cpu_output) {
        if ($line =~ /\s(\d+)\s*$/i) {
            if ($cpu_flag == 0) {
                # First reading of CPU gives an average since the last reboot. So, we need to neglect this one.
                $cpu_flag = 1;
                next;
            } else {
                # parse out the idle cpu and subtract from 100 to get used cpu %
                my $cpu_idle = int($1);
                $system_stats->{cpu_usage_percent} += 100 - $cpu_idle;
                $cpu_samples++;
            }
        }
    }

    # parse out the memory output and tally up the damples
    @mem_output = (@mem_output1, @mem_output2);
    foreach my $line (@mem_output) {
        if ($line =~ /\s+(\d+)\%/i) {
            # parse the used memory percentage
            my $mem_used = int($1);
            $system_stats->{mem_usage_percent} += $mem_used;
            $mem_samples++;
        }
    }

    # average out cpu usage and mem usage from all the tallied samples.
    if ($cpu_samples > 0) {
        $system_stats->{cpu_usage_percent} = int($system_stats->{cpu_usage_percent} / $cpu_samples);
    }
    if ($mem_samples > 0) {
        $system_stats->{mem_usage_percent} = int($system_stats->{mem_usage_percent} / $mem_samples);
    }

    return $system_stats;
}

################################################################################
# Method:    get_ip_info()
# Objective: Get the IP information for this system and return hash with that
#            data
################################################################################
sub get_ip_info {
    my ($self) = @_;

    my ($mac, $ipv4, $ipv6, $ipv6_prefix_len, $gateway, $subnet_mask);
    my ($broadcast, $status, @output, @ifconfig);

    my $interface;
    my $interface_name = 'NA';
    my $ip_hash = {};

    # this will show you all the interfaces
    my $cmd = "/usr/sbin/lanscan";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::get_ip_info(): Error running command ($cmd)");
        return {};
    }

    foreach my $int_line (@output) {
        chomp($int_line);
        # skip header
        if ($int_line =~ /^Hardware/ or $int_line =~ /^Path/) {
            next;
        }

        # get interface names
        if ($int_line =~ /ETHER/ and $int_line =~ /\s+UP\s+/i) {
            # get the mac address
            $interface = (split(/\s+/, $int_line))[4];
            $mac = (split(/\s+/, $int_line))[1];
            $mac =~ s/0x//;

            # get an individual interface - details
            $cmd = "/usr/sbin/ifconfig $interface";
            my ($status, @ifconfig) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                # if the failure is due to the interface not being available,
                # simply go on to the next interface
                if (grep(/no such interface/, @ifconfig) > 0) {
                    next;
                } else {
                    $self->_set_err("Error: HPUXHost::get_ip_info(): Error running command ($cmd)");
                    return {};
                }
            }

            foreach my $line (@ifconfig) {
                # get ipv4 address
                if ($line =~ /inet\s+(\S+)/) {
                    $ipv4 = $1;
                }

                # get ipv6 address
                if ($line =~ /inet6\s+(\S+)\/(\d+)/) {
                    $ipv6 = $1;
                    $ipv6_prefix_len = $2;
                }

                # get broadcast address
                if ($line =~ /broadcast\s+(\S+)/) {
                    $broadcast = $1;
                }

                # get subnet mask
                if ($line =~ /netmask\s+(\S+)/) {
                    $subnet_mask = $1;
                }
            }

            # set fields in the hash
            $ip_hash->{$interface}->{interface_name} = $interface_name;
            $ip_hash->{$interface}->{mac} = $mac;
            $ip_hash->{$interface}->{description} = "NIC";
            $ip_hash->{$interface}->{gateway} = "";
            $ip_hash->{$interface}->{broadcast} = $broadcast;
            $ip_hash->{$interface}->{subnet_mask} = $subnet_mask;
            $ip_hash->{$interface}->{mtusize} = "";
            $ip_hash->{$interface}->{ipv4} = $ipv4;
            $ip_hash->{$interface}->{ipv6} = $ipv6;

            # if this is the management interface of the client
            # set an attribute
            my $host_rec = &Tharn::host($self->hostname());
            my $host_ip_default = $host_rec->{'default_ip'};
            if ($ipv4 eq $host_ip_default) {
                $ip_hash->{$interface}->{is_mgmt_interface} = 1;
            } else {
                $ip_hash->{$interface}->{is_mgmt_interface} = 0;
            }

            # calculate the ipv4 subnet
            if ($ipv4 ne "" and $subnet_mask ne "") {
                $ip_hash->{$interface}->{ipv4_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv4,
                                                                                -ip_version => "ipv4",
                                                                            -ipv4_netmask => $subnet_mask);
            } else {
                $ip_hash->{$interface}->{ipv4_subnet} = "";
            }

            # calculate the ipv6 subnet
            if ($ipv6 ne "" and $ipv6_prefix_len ne "") {
                $ip_hash->{$interface}->{ipv6_prefix_len} = $ipv6_prefix_len;
                $ip_hash->{$interface}->{ipv6_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv6,
                                                                                -ip_version => "ipv6",
                                                                                -ipv6_prefix_len => $ipv6_prefix_len);
            } else {
                $ip_hash->{$interface}->{ipv6_prefix_len} = "";
                $ip_hash->{$interface}->{ipv6_subnet} = "";
            }
        }
    }

    # return info hash
    return ($ip_hash);
}

################################################################################
# Subroutine Name: set_ip_address()
# Objective:       Sets the IP address and subnet mask for an interface
# Details:         See POD documentation at the end of this file
################################################################################
sub set_ip_address {
    my ($self) = @_;

    $self->_set_err("This function is not yet implemented");
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################
sub get_default_mp_type {
    my ($self) = @_;

    return("PVLINKS");

}

################################################################################
# Subroutine Name: rotate_logs()
# Objective:       rotate the log files for HPUX
################################################################################
sub rotate_logs {
    my ($self) = @_;

    $self->_clear_err();

    # rotate the Veritas logs
    my $status = $self->rotate_generic_log(-file => "/etc/vx/dmpevents.log", -minsize => "1024", -keep => "5");
    if ($status == $ClientUtil::TRUE) {
        return $ClientUtil::TRUE;
    } else {
        if (not $self->error) {
            $self->_set_err("An error was received while attempting to rotate the Veritas DMP logs");
        }
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: get_devices()
# Objective:       wrapper to call _get_devices() routine
################################################################################
sub get_devices {
    my ($self) = @_;

    return ($self->_get_devices());

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _get_devices
# Objective:       Get list of all devices on host
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices {
    my ($self) = @_;

    my ($cnt, $cmd, $dev_str, $device, $status, @output, $found, @hDevices);

    # Determine if we have MP type set, if so call function to get the proper device list
    if (defined $self->{mp} and $self->{mp_type} !~ /none/i) {
        @hDevices = $self->{mp}->get_mp_devices(-conn => $self->{conn});
        if ($self->{mp}->error()) {
            my $error = join("\n", $self->{mp}->errmsg());
            $self->sanlog(-msg => "Error HPUXHost::_get_devices() $error");
            return (wantarray()) ? () : [];
        } else {
            return (wantarray()) ? @hDevices : \@hDevices;
        }
    }

    # run ioscan getting information from the kernel (since this is allot
    # faster, and ioscan -fn -C disk should have already been run earlier)
    $cmd = "/usr/sbin/ioscan -fnk -C disk";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::_get_devices() Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    # get boot device (to exclude)
    my $boot_device = $self->_get_boot_device();
    if ($self->error) {
        return (wantarray()) ? () : [];
    }

    $found = $ClientUtil::FALSE;
    $cnt   = 0;
    foreach my $line (@output) {
        if ($line =~ /netapp/i and $line =~ /claimed/i) {
            $found = $ClientUtil::TRUE;
            next;
        }

        if ($found) {
            $device = (split(' ', $line))[1];
            $device =~ s#/dev/rdsk/##;

            # skip boot device
            if ($boot_device eq $device) {
                $found = $ClientUtil::FALSE;
                next;
            }

            # create hash of lun descriptions
            my $hDevs = {};
            $hDevs->{dev_path}  = "/dev/dsk";
            $hDevs->{path}  = "/dev/rdsk";
            $hDevs->{name}  = $device;
            $hDevs->{name_ctd}  = $device;
            $hDevs->{type}  = "NETAPP";
            $hDevs->{label} = "NA";
            $hDevs->{id}    = $cnt;

            push @hDevices, $hDevs;
            $cnt++;
        }
        $found = $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "HPUXHost::_get_devices() Found " . scalar(@hDevices) .
                          " devices (includes multiple paths to devices)");

    return (wantarray()) ? @hDevices : \@hDevices;

}

################################################################################
# Subroutine Name: _get_unlabeled_devices
# Objective:       Retrieve devices that do not have a label
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_unlabeled_devices {
    my ($self) = @_;

    my (@hDevices, @devices, $cmd, $status, @output);

    #  Retrieve all host devices
    @hDevices = $self->_get_devices();
    return (wantarray()) ? () : [] if ($self->error);

    #  Determine which devices do not have
    #  a label
    foreach my $device (@hDevices) {

        # Skip if this is not a NETAPP device
        next if ($device->{type} !~ /NETAPP/);

        # If the device is not in a VG (which is what
        # pvdisplay will how), mark it as "unlabeled"
        # and hence available.
        $cmd = "/usr/sbin/pvdisplay " . $device->{dev_path} . "/" . $device->{name};
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if (grep(/Cannot display physical volume/i, @output) > 0) {
            push(@devices, $device->{name});
        }
    }

    #  Format the list of devices with label
    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: _iscsi_configure()
# Objective:       Configure iscsi to include the
#                  correct targets for the filers.
#                  Assumptions: a targets list is passed into this function with
#                  the IP's of all the ports on each filer.
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _iscsi_configure {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err(
                         "Error: HPUXHost::_iscsi_configure() Invalid " . "value for argument '-targets' => %args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY} == 0) {
        # Ignore _iscsi_configure, we are asked to NOT configure anything
        return $ClientUtil::TRUE;
    }

    # Don't check status - run blind.
    $self->_iscsi_deconfigure();
    # Since we are running _iscsi_deconfigure blindly, clear out any error
    # that may occur.
    $self->_clear_err();

    # Have to see if the host IQN is set, if not set it
    $cmd = "/opt/iscsi/bin/iscsiutil -l|grep 'Initiator Name'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (grep(/iqn/, @output) == 0) {
        # IQN is not defined have to set it
        $cmd = "/opt/iscsi/bin/iscsiutil -i -N iqn.1986-03.com.hp:" . $self->hostname() . ".p0";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Error: HPUXHost::_iscsi_configure() Error while " . " setting host IQN");
            return $ClientUtil::FALSE;
        }
    }

    # In C-mode, targets will have either ipv4 or ipv6 address unlike 7-mode.
    # Hence, remove the targets from the target list which don't have valid IP address
    # of IP version requested by user.
    @targets = $self->filter_targets_by_ipversion(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    foreach my $target (@targets) {
        # Add targets to iscsiadm
        $cmd = "/opt/iscsi/bin/iscsiutil -a -I $target->{ip}";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Error: HPUXHost::_iscsi_configure() cmd ($cmd) " .
                            "returned a status of ($status)");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure()
# Objective:       Deconfigure iscsi, remove targets from discovery list
#                  Assumptions: a targets list is passed into this function with
#                  the IP's of all the ports on each filer.
#                  IF a targets list is not provided, all targets on machine are removed
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _iscsi_deconfigure {
    my ($self, %args) = @_;

    my ($status, @output, $cmd, @targets);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err(
                       "Error: HPUXHost::_iscsi_deconfigure() Invalid " . "value for argument '-targets' => %args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (scalar(@targets) == 0) {

        # No targets provided, need to get list
        # iscsiadm list discovery-address|awk '{print $3}'
        $cmd = "/opt/iscsi/bin/iscsiutil -p -D | egrep \"IP|iSCSI Portal\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Error: HPUXHost::_iscsi_deconfigure() cmd ($cmd) " . "returned a status of ($status)");
            return $ClientUtil::FALSE;
        }

        my %tgts;
        my $current_target = "";
        foreach my $line (@output) {
            if ($line =~ /^\s+IP Address\s+\:\s+(\S+)/i) {
                $current_target = $1;
            } elsif ($line =~ /^\s+iSCSI Portal Group Tag\s+\:\s+(\S+)/i) {
                $tgts{$current_target} = $1;
            }
        }

        @targets = ();
        foreach my $target (keys(%tgts)) {
            my $tgt;
            $tgt->{ip}  = $target;
            $tgt->{pgt} = $tgts{$target};
            push @targets, $tgt;
        }
    }

    foreach my $target (@targets) {
        # Check to see if the portal group tag is defined on the target
        # for C-mode, this key is not 'pgt' but 'tpgt'.
        if (not defined($target->{pgt}) or $target->{pgt} eq "") {
            $target->{pgt} = $target->{tpgt};
        }

        # remove targets from iscsiadm
        $cmd = "/opt/iscsi/bin/iscsiutil -d -I $target->{ip} -M $target->{pgt}";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # Check to see if the removal of the target worked, if the target does not
        # exist, the call will state it is not present in the driver repository.
        if (grep(/is not present in the driver repository/, @output)) {
            next;
        } elsif ($status != 0) {
            $self->_set_err("Error: HPUXHost::_iscsi_deconfigure() cmd ($cmd) " . "returned a status of ($status)");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _fcp_deconfigure()
# Objective:       Deconfigures fcp
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _fcp_deconfigure {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _MP_configure
# Objective:       Configures MP on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _MP_configure() {
    my ($self) = @_;

    # verify that all mp pkgs are installed
    $self->{mp}->verify(-conn => $self->{conn});
    if ($self->{mp}->error) {
        $self->_set_err($self->{mp}->errmsg());
        return $ClientUtil::FALSE;
    }
    $self->sanlog(-msg => "MP is verified", -msg_level => 2);

    # configure MP file
    $self->{mp}->config(-conn => $self->{conn});
    if ($self->{mp}->error) {
        $self->_set_err($self->{mp}->errmsg());
        return $ClientUtil::FALSE;
    }
    $self->sanlog(-msg => "MP is configured", -msg_level => 2);

    $self->{mp}->enable(-conn => $self->{conn});
    if ($self->{mp}->error) {
        $self->_set_err($self->{mp}->errmsg());
        return $ClientUtil::FALSE;
    }
    $self->sanlog(-msg => "MP enabled", -msg_level => 2);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _scsimgr
# Objective:       Runs scsimgr to bind new lun path
# Details:         See POD documentation at the end of this file
################################################################################
sub _scsimgr () {
    my ($self, @devices) = @_;
    my ($cmd, $status, @output);
    my ($connection);
    my ($device_retry);

    # Create new connection to run scsimgr on
    $connection = $self->_connect($self->hostname(), "", "", "", "THARNHOST");

    # Check for errors on connection
    if ($self->error()) {
        $self->_set_err("Error: HPUXHost::_scsimgr() Could not create connection to client " . $self->hostname());
        return $ClientUtil::FALSE;
    }

    # Loop through device list and run scsimgr on it
    # If the command scsimgr fails, retry the scsimgr command up to 5 more times
    while (scalar(@devices) != 0) {
        my $device = shift(@devices);
        if ($device =~ /lunpath\s+\d+\s+(\S+\.\S+)\s+/) {
            my $device_id = $1;

            $cmd = "/usr/sbin/scsimgr -f replace_wwid -H $device_id";
            ($status, @output) = &ClientUtil::execute($connection, $cmd);

            # Check to see if the device failed, and if it has remaining attempts
            if (($status > 0 or grep(/ERROR: Failed to remove the LUN path/i, @output))
                 and $device_retry->{$device_id} != 5) {
                # scsimgr failed and have not retried the device, set the retry flag
                # and push the device back on the list to retry
                push(@devices, $device);
                $device_retry->{$device_id}++;
                &SanUtil::nate_sleep(5);
            } elsif ($status > 0 or grep(/ERROR/i, @output)) {
                &Tharn::disconnect($connection);
                # scsimgr failed, set error and return failure
                $self->_set_err("Error: HPUXHost::_scsimgr() Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    &Tharn::disconnect($connection);

    # Return success
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_boot_device
# Objective:       Runs boot device to user
################################################################################
sub _get_boot_device {
   my ($self) = @_;

   my ($cmd, $status, @bootinfo, $primary_boot);

    # get the booted device and exclude (sanboot concern)
    $cmd = "/usr/sbin/setboot";
    ($status, @bootinfo) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::_get_boot_device() Command ($cmd) returned a bad status ($status)");
        return "";
    }

    # parse out primary boot path information.
    $primary_boot = (grep(/^Primary bootpath/i, @bootinfo))[0];
    $primary_boot =~ s/^.*\:\s+//;

    # determine if we got back the new agile /dev/rdisk entry
    # if we did, return that entry.
    if ($primary_boot =~ /\(\S+\)/ and $primary_boot =~ /rdisk/) {
        $primary_boot =~ s/\S+\s+\(//;
        $primary_boot =~ s/\)//;
        return($primary_boot);
    }

    # legacy path with no-agile /dev/rdsk entries
    # lookup to get device from HW Path
    $cmd = "/usr/sbin/ioscan -fnk -H \"$primary_boot\"";
    ($status, @bootinfo) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXHost::_get_boot_device() Command ($cmd) returned a bad status ($status)");
        return "";
    }

    # find the device names for the HW PATH
    foreach my $line (@bootinfo) {
        # skip header and HW path line
        if ($line =~ /^Class|===|^disk/i) {
            next;
        }

        # only need first entry (since we don't
        # care about slices)
        $line =~ s/^\s+//;
        my ($dsk, $rdsk) = split(/\s+/, $line);
        $rdsk =~ s#/dev/rdsk/##;
        $rdsk =~ s#s\d+##;

        $self->sanlog(-msg => "Boot Device was determined to be $rdsk");

        return($rdsk);

    }

    $self->_set_err("Error: HPUXHost::_get_boot_device() Could not determine boot device");
    return "";

}

1;

