# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a Hyper-V machine
## @author hrivera@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::HyperVHost;

# Package Requirements
######################
use San::SanClient;
use San::SanClient::ClientUtil;
use San::SanClient::Host::WinHost;
use San::SanLog;
use San::SanCache;
use Error;
use vars qw( @EXPORT @ISA );

@ISA = qw(San::SanClient::Host::WinHost Error SanLog SanCache SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new HyperVHost object
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn, $winclient_obj, $obj_template);

    $self = {};
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?winclient_obj/) {
            if (not length($winclient_obj = $args{$_})) {
                $self->_set_err("Invalid value for -winclient_obj: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Missing required argument '-conn'");
        return $self;
    }

    $self->{TEST_BIN} = 'X:' . "\\SAN\\bin\\common\\";
    $self->{conn} = $conn;
    $self->{WinClient} = $winclient_obj;

    # display the version of the hvcontrol utility
    $self->hvcontrol(-version => 1);
    if ($self->error()) {
        return $self;
    }

    return $self;
}

################################################################################
# Method: discover
# Objective: Discover devices and label if new.
# Details: See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $num_expected_luns;

    foreach (keys %args) {
        if (/-?num_expected_luns/) {
            if (not length($num_expected_luns = $args{$_})) {
                $self->_set_err("Invalid value for -num_expected_luns");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # discover with WinClient's methods
    $self->{WinClient}->getDevices($num_expected_luns);
    if ($self->{WinClient}->error) {
        my @tmp_msg = $self->{WinClient}->errmsg();
        $self->_set_err("WinClient failed to discover new devices");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: format()
# Objective:       method to add the specified devices to a virtual machine
#                  NOOP for HyperV
################################################################################
sub format {

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
##############################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: add_passthrough_disks_to_vm()
# Objective:       method to add the specified devices to a virtual machine
################################################################################
sub add_passthrough_disks_to_vm {
    my ($self, %args) = @_;

    my (@passthrough_devices, $vm_name);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?passthrough_devices/) {
            if (not scalar(@passthrough_devices = @{$args{$_}})) {
                $self->_set_err("Invalid value for -passthrough_devices");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if (not scalar(@passthrough_devices)) {
        $self->_set_err("Missing required parameter -passthrough_devices");
        return $ClientUtil::FALSE;
    }
    if (not defined($vm_name)) {
        $self->_set_err("Missing required parameter -vm_name");
        return $ClientUtil::FALSE;
    }

    # obtain the virtual machine's full name from the Hyper V host
    my $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
    if ($self->error()) {
        $self->_set_err("Could not find $vm_name.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # make sure this virtual machine is powered off
    my (@vm_state) = $self->_get_vm_state(-full_vm_name => $full_vm_name);
    if (not grep(/Stopped/i, @vm_state)) {
        $self->_set_err("Cannot add disks to virtual machine '$full_vm_name'. It is not powered off.");
        return $ClientUtil::FALSE;
    }

    # add each passthrough device to this vm
    foreach my $passthrough_dev (@passthrough_devices) {
        my $disk_id = $passthrough_dev->get_disk_id();
        my $controller = $passthrough_dev->get_controller_type();

        my ($status, @output) = $self->hvcontrol(-vm_name => $full_vm_name,
                                                 -add_passthrough_disk => 1,
                                                 -disk_id => $disk_id,
                                                 -controller => $controller);
        if ($self->error()) {
            $self->_set_err("Could not add passthrough disk '" . $disk_id . " ($controller)' to virtual machine '$full_vm_name'\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: add_vhd_files_to_vm()
# Objective:       method to add the specified vhd files to a virtual machine
################################################################################
sub add_vhd_files_to_vm {
    my ($self, %args) = @_;

    my (@vhd_files, @controller_types, $vm_name);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vhd_files/) {
            if (not scalar(@vhd_files = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vhd_files");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?controller_types/) {
            if (not scalar(@controller_types = @{$args{$_}})) {
                $self->_set_err("Invalid value for -controller_types");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vhd_files)) {
        $self->_set_err("Missing required parameter -vhd_files");
        return $ClientUtil::FALSE;
    }
    if (not scalar(@controller_types)) {
        $self->_set_err("Missing required parameter -controller_types");
        return $ClientUtil::FALSE;
    }
    if (not defined($vm_name)) {
        $self->_set_err("Missing required parameter -vm_name");
        return $ClientUtil::FALSE;
    }

    # obtain the virtual machine's full name from the Hyper V host
    my $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
    if ($self->error()) {
        $self->_set_err("Could not find $vm_name.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # make sure this virtual machine is powered off
    my (@vm_state) = $self->_get_vm_state(-full_vm_name => $full_vm_name);
    if (not grep(/Stopped/i, @vm_state)) {
        $self->_set_err("Cannot add disks to virtual machine '$full_vm_name'. It is not powered off.");
        return $ClientUtil::FALSE;
    }

    my $controller = "";

    # add each vhd device to this vm
    foreach my $vhd_file (@vhd_files) {

        # if there are still controller types in the array shift one off
        # otherwise the last controller type will be used
        if (scalar(@controller_types) != 0) {
            $controller = shift(@controller_types);
        }
        if ($controller eq "") {
            $self->_set_err("Could not add VHD '" . $vhd_file . "' to virtual machine " .
                            "'$full_vm_name'. No controller type was specified.");
            return $ClientUtil::FALSE;
        }

        # add the vhd file to the vm
        my ($status, @output) = $self->hvcontrol(-vm_name => $full_vm_name,
                                                 -add_vhd => 1,
                                                 -vhd_path => $vhd_file,
                                                 -controller => $controller);
        if ($self->error()) {
            $self->_set_err("Could not add VHD '" . $vhd_file . " ($controller)' " .
                            "to virtual machine '$full_vm_name'\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_vhd_devices_from_vm()
# Objective:       Obtain all VHD devices that are being used by the specified
#                  Virtual Machine
# Returns:         Returns a hash relating vhd_file_path (key) with
#                  vhd_device (value)
################################################################################
sub get_vhd_devices_from_vm {
    my ($self, %args) = @_;

    my ($vm_name, @vhd_devices);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name");
                return {};
            }
        }
    }

    # verify mandatory parameters
    if (not defined($vm_name)) {
        $self->_set_err("Missing required parameter -vm_name");
        return {};
    }

    # verify that this SanClient object contains a DevMgr
    my $devmgr_obj = $self->{obj}{devmgr};
    if (not defined($devmgr_obj)) {
        $self->_set_err("SanClient object does not contain a valid DevMgr object.");
        return {};
    }

    # obtain all the vhd devices
    @vhd_devices = $devmgr_obj->get_devices(-device_type => "vhd");
    if ($devmgr_obj->error()) {
        $self->_set_err("Could not obtain VHD list from the DevMgr object\n" .
                        "Reason: " . $devmgr_obj->errmsg());
        return {};
    }

    # obtain the virtual machine's full name from the Hyper V host
    my $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
    if ($self->error()) {
        $self->_set_err("Could not find $vm_name.\n" .
                        "Reason: " . $self->errmsg());
        return {};
    }

    # get the vhd list from this machine
    my ($status, @vhd_list) = $self->hvcontrol(-vm_name => $full_vm_name,
                                               -get_vhd_list => 1);
    if ($self->error()) {
        $self->_set_err("Could not list VHDs from Virtual machine '$full_vm_name'\n" .
                        "Reason: " . $self->errmsg());
        return {};
    }

    my $vhd_hash = {};

    # if there are no VHD disks, then there is nothing to get
    if (grep(/\[No VHD Disks\]/i, @vhd_list)) {
        return {};
    }

    # find the associate vhd device for each vhd file in use by this VM.
    # if a vhd_file does not have an associate $vhd_device, then it will be
    # excluded.
    foreach my $vhd_file (@vhd_list) {
        # parse out the controller type and vhd path
        # $vhd_file looks like this for ide:
        #(ide) E:\win2k3_32_hv_image\win2k3_32_hv_image.vhd
        # or this for scsi:
        #(scsi) E:\win2k3_32_hv_image\win2k3_32_hv_image.vhd
        my ($controller_type, $vhd_path) = ($vhd_file =~ /\((scsi|ide)\) (\S+)/);

        foreach my $vhd_device_obj (@vhd_devices) {
            my @vhd_files_from_dev = $vhd_device_obj->get_vhd_files();
            my $matched = 0;

            foreach my $vhd_dev_file (@vhd_files_from_dev) {
                if ($vhd_dev_file eq $vhd_path) {
                    $matched = 1;
                    last;
                }
            }

            if ($matched) {
                # associate the vhd path with the VHD device object that it belongs to
                $vhd_hash->{$vhd_path} = $vhd_device_obj;
                last;
            }
        }
    }

    # return the hash that relates the vhd_paths with their respective vhd_device_obj
    return $vhd_hash;
}

################################################################################
# Subroutine Name: get_passthrough_devices_from_vm()
# Objective:       Obtain all Passthrough devices that are being used by the specified
#                  Virtual Machine
# Returns:         An array of HyperVDev objects
################################################################################
sub get_passthrough_devices_from_vm {
    my ($self, %args) = @_;

    my ($vm_name, @passthrough_devices);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name");
                return (wantarray() ? () : []);
            }
        }
    }

    # verify mandatory parameters
    if (not defined($vm_name)) {
        $self->_set_err("Missing required parameter -vm_name");
        return (wantarray() ? () : []);
    }

    # verify that this SanClient object contains a DevMgr
    my $devmgr_obj = $self->{obj}{devmgr};
    if (not defined($devmgr_obj)) {
        $self->_set_err("SanClient object does not contain a valid DevMgr object.");
        return (wantarray() ? () : []);
    }

    # obtain the virtual machine's full name from the Hyper V host
    my $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
    if ($self->error()) {
        $self->_set_err("Could not find $vm_name.\n" .
                        "Reason: " . $self->errmsg());
        return (wantarray() ? () : []);
    }

    # get the passthrough list from this machine
    my ($status, @passthrough_list) = $self->hvcontrol(-vm_name => $full_vm_name,
                                                       -get_passthrough_list => 1);
    if ($self->error()) {
        $self->_set_err("Could not list VHDs from Virtual machine '$full_vm_name'\n" .
                        "Reason: " . $self->errmsg());
        return (wantarray() ? () : []);
    }

    # if there are no pass-through disks, then there is nothing to get
    if (grep(/\[No Pass-Through Disks\]/i, @passthrough_list)) {
        return (wantarray() ? () : []);
    }

    my @return_devices;

    # find the associate passthrough HyperVDev device for each disk listed in the passthrough list
    # that is in use by this VM.
    # if a $disk_id does not have an associate $passthrough_device, then it will be
    # excluded.
    foreach my $disk_line (@passthrough_list) {
        # parse out the controller type and disk number
        # $disk_line looks like this for ide:
        #(ide) Disk 2
        # or this for scsi:
        #(scsi) Disk 2
        next if ($disk_line !~ /scsi|ide/i);
        my ($controller_type, $disk_id) = ($disk_line =~ /\((scsi|ide)\) (Disk \d+)/i);

        # get the passthrough device by disk id
        my $dev_obj = $devmgr_obj->get_device_by_disk_id(-disk_id => $disk_id,
                                                         -device_type => "passthrough");
        if ($devmgr_obj->error()) {
            # could not find a HyperVDev object for this device, skip to the next device in the
            # list
            $devmgr_obj->_clear_err();
            next;
        }

        # store the device object to be returned later
        push (@return_devices, $dev_obj);
    }

    # return the list of passthrough HyperVDev objects in use by this VM
    return (wantarray() ? @return_devices : \@return_devices);
}

################################################################################
# Subroutine Name: remove_vhd_files_from_vm()
# Objective:       method to remove the specified vhd files from this Virtual
#                  Machine.
################################################################################
sub remove_vhd_files_from_vm {
    my ($self, %args) = @_;

    my (@vhd_files, $vm_name);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vhd_files/) {
            if (not scalar(@vhd_files = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vhd_files");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name");
                return $ClientUtil::FALSE;
            }
        }
    }

    # verify mandatory parameters
    if (not scalar(@vhd_files)) {
        $self->_set_err("Missing required parameter -vhd_files");
        return $ClientUtil::FALSE;
    }
    if (not defined($vm_name)) {
        $self->_set_err("Missing required parameter -vm_name");
        return $ClientUtil::FALSE;
    }

    # obtain the virtual machine's full name from the Hyper V host
    my $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
    if ($self->error()) {
        $self->_set_err("Could not find $vm_name.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # make sure this virtual machine is powered off
    my (@vm_state) = $self->_get_vm_state(-full_vm_name => $full_vm_name);
    if (not grep(/Stopped/i, @vm_state)) {
        $self->_set_err("Cannot add disks to virtual machine '$full_vm_name'. It is not powered off.");
        return $ClientUtil::FALSE;
    }

    # remove all @vhd_files from the virtual machine
    foreach my $vhd_file (@vhd_files) {
        my ($status, @output) = $self->hvcontrol(-vm_name => $full_vm_name,
                                                 -remove_vhd => 1,
                                                 -vhd_path => $vhd_file);
        if ($self->error()) {
            $self->_set_err("Could not remove VHD '" . $vhd_file . "' from virtual machine '$full_vm_name'\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remove_passthrough_devices_from_vm()
# Objective:       method to remove passthrough devices from a virtual machine
################################################################################
sub remove_passthrough_devices_from_vm {
    my ($self, %args) = @_;

    my (@passthrough_devices, $vm_name);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?passthrough_devices/) {
            if (not scalar(@passthrough_devices = @{$args{$_}})) {
                $self->_set_err("Invalid value for -passthrough_devices");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name");
                return $ClientUtil::FALSE;
            }
        }
    }

    # verify mandatory parameters
    if (not scalar(@passthrough_devices)) {
        $self->_set_err("Missing required parameter -passthrough_devices");
        return $ClientUtil::FALSE;
    }
    if (not defined($vm_name)) {
        $self->_set_err("Missing required parameter -vm_name");
        return $ClientUtil::FALSE;
    }

    # obtain the virtual machine's full name from the Hyper V host
    my $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
    if ($self->error()) {
        $self->_set_err("Could not find $vm_name.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # make sure this virtual machine is powered off
    my (@vm_state) = $self->_get_vm_state(-full_vm_name => $full_vm_name);
    if (not grep(/Stopped/i, @vm_state)) {
        $self->_set_err("Cannot add disks to virtual machine '$full_vm_name'. It is not powered off.");
        return $ClientUtil::FALSE;
    }

    # remove all passthrough_devices from the virtual machine
    foreach my $passthrough_dev (@passthrough_devices) {
        my $disk_id = $passthrough_dev->get_disk_id();

        my ($status, @output) = $self->hvcontrol(-vm_name => $full_vm_name,
                                                 -remove_passthrough_disk => 1,
                                                 -disk_id => $disk_id);
        if ($self->error()) {
            $self->_set_err("Could not remove passthrough disk '" . $disk_id . "' from virtual machine '$full_vm_name'\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remove_empty_drive_slots_from_vms()
# Objective:       Remove any empty drive slots in -vm_list
################################################################################
sub remove_empty_drive_slots_from_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # remove all empty drive slots for all the vm's in the list
    foreach my $vm_name (@vm_list) {
        # obtain the virtual machine's full name from the Hyper V host
        my $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not find $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # get the state of this virtual machine, if its not powered off, then skip it
        my (@vm_state) = $self->_get_vm_state(-full_vm_name => $full_vm_name);
        if (not grep(/Stopped/i, @vm_state)) {
            next;
        }

        # clean up any empty drive slots.
        my ($status, @output) = $self->hvcontrol(-remove_empty_drive_slots => 1,
                                                 -vm_name  => $full_vm_name);
        if ($self->error()) {
            $self->_set_err("Could not remove empty drive slots for $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: power_management()
# Objective:       method to allow power management for virtual machines
################################################################################
sub power_management {
    my ($self, %args) = @_;

    my ($action);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?action/) {
            if (not length($action = $args{$_})) {
                $self->_set_err("Invalid value for -action => $action");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if ($action =~ /^power_on$/i) {
        $self->_power_on_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_on_all$/i) {
        $self->_power_on_all_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_off$/i) {
        $self->_power_off_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_off_all$/i) {
        $self->_power_off_all_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^shutdown$/i) {
        $self->_shutdown_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^shutdown_all$/i) {
        $self->_shutdown_all_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_cycle$/i) {
        $self->_power_cycle_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_cycle_all$/i) {
        $self->_power_cycle_all_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Invalid value for -action: ($action)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: set_ip_address()
# Objective:       Sets the IP address and subnet mask for an interface
# Details:         See POD documentation at the end of this file
################################################################################
sub set_ip_address {
    my ($self) = @_;

    $self->_set_err("This function is not yet implemented");
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################
sub get_default_mp_type {
    my ($self) = @_;

    return("MPIO");
}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For HyperVHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    orphaned_device_cleanup
# Objective: NOOP for hyperv host
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_on_vms()
# Objective:       Power on the specified virtual machines
################################################################################
sub _power_on_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my (@waiting_to_boot_vm_list);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # power on each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {

        # obtain the virtual machine's full name from the Hyper V host
        my $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not find $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # get the state of this virtual machine, if its not powered off, then skip
        my (@vm_state) = $self->_get_vm_state(-full_vm_name => $full_vm_name);
        if (not grep(/Stopped/i, @vm_state)) {
            $self->sanlog(-msg => "Virtual Machine ($full_vm_name) is already powered on.");
            next;
        }

        $self->sanlog(-msg => "Powering on ($full_vm_name).");

        # power on this virtual machine
        my ($status, @output) = $self->hvcontrol(-power_on => 1,
                                                 -vm_name  => $full_vm_name);
        if ($self->error()) {
            $self->_set_err("Could not power on $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # push this virtual machine on the waiting to boot vm list
        push(@waiting_to_boot_vm_list, $vm_name);
    }

    # wait for the virtual machines to boot
    $self->_wait_for_vms_to_boot(@waiting_to_boot_vm_list);
    if ($self->error()) {
        $self->_set_err("Failed while waiting for virtual machines to boot.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_on_all_vms()
# Objective:       Power on every virtual machine.
################################################################################
sub _power_on_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # turn on all the virtual machines.
    $self->_power_on_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not power on all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_off_vms()
# Objective:       Power on the specified virtual machines
################################################################################
sub _power_off_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my (@waiting_to_shutdown_vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # power off each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {

        # obtain the virtual machine's full name from the Hyper V host
        my $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not find $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # get the state of this virtual machine, if its not running then skip
        my (@vm_state) = $self->_get_vm_state(-full_vm_name => $full_vm_name);
        if (not grep(/Running/i, @vm_state)) {
            $self->sanlog(-msg => "Virtual Machine ($full_vm_name) is already powered off.");
            next;
        }

        $self->sanlog(-msg => "Powering off ($full_vm_name)");

        # power off this virtual machine
        $self->hvcontrol(-power_off => 1,
                         -vm_name   => $full_vm_name);
        if ($self->error()) {
            $self->_set_err("Could not power off $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # push this virtual machine on the waiting to shutdown vm list
        push(@waiting_to_shutdown_vm_list, $full_vm_name);
    }

    # wait for the virtual machines to shutdown
    $self->_wait_for_vms_to_shutdown(@waiting_to_shutdown_vm_list);
    if ($self->error()) {
        $self->_set_err("Failed while waiting for virtual machines to power off.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }


    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_off_all_vms()
# Objective:       Power off every virtual machine.
################################################################################
sub _power_off_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # turn off all the virtual machines.
    $self->_power_off_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not power off all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _shutdown_vms()
# Objective:       Gracefully Shutdown the specified virtual machines
################################################################################
sub _shutdown_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my (@waiting_to_shutdown_vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # Shutdown each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {

        # obtain the virtual machine's full name from the Hyper V host
        my $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not find $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # get the state of this virtual machine, if its not running then skip
        my (@vm_state) = $self->_get_vm_state(-full_vm_name => $full_vm_name);
        if (not grep(/Running/i, @vm_state)) {
            $self->sanlog(-msg => "Virtual Machine ($full_vm_name) is already powered off.");
            next;
        }

        $self->sanlog(-msg => "Shutting down ($full_vm_name)");

        # shutdown this virtual machine
        $self->hvcontrol(-shutdown => 1,
                         -vm_name   => $full_vm_name);
        if ($self->error()) {
            $self->_set_err("Could not power off $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # push this virtual machine on the waiting to shutdown vm list
        push(@waiting_to_shutdown_vm_list, $full_vm_name);
    }

    # wait for the virtual machines to shutdown
    $self->_wait_for_vms_to_shutdown(@waiting_to_shutdown_vm_list);
    if ($self->error()) {
        $self->_set_err("Failed while waiting for virtual machines to power off.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _shutdown_all_vms()
# Objective:       Shutdown every virtual machine.
################################################################################
sub _shutdown_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # shutdown all the virtual machines.
    $self->_shutdown_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not shutdown all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_cycle_vms()
# Objective:       Power off/Power on the specified virtual machines
################################################################################
sub _power_cycle_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my ($soft_shutdown) = "no";

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }  elsif (/-?soft_shutdown/) {
            if (not length($soft_shutdown = $args{$_})) {
                $self->_set_err("Invalid value for -soft_shutdown => $soft_shutdown");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "Power Cycling the following Virtual machines (graceful: $soft_shutdown):\n " .
                          join("\n ", @vm_list));

    # power off/shutdown the specified VMs
    if ($soft_shutdown eq "no") {
        $self->_power_off_vms(-vm_list => \@vm_list);
    } else {
        $self->_shutdown_vms(-vm_list => \@vm_list);
    }
    if ($self->error()) {
        $self->_set_err("power_off_vms/shutdown_vms failed\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # power on the specified VMs
    $self->_power_on_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("power_on_vms failed\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Power Cycle Complete");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_cycle_all_vms()
# Objective:       Power off/Power on all virtual machines
################################################################################
sub _power_cycle_all_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # power cycle all the virtual machines
    $self->_power_cycle_vms(-vm_list => \@vm_list, %args);
    if ($self->error()) {
        $self->_set_err("Power Cycle failed\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_vm_name()
# Objective:       given a virtual machine dns name, determine the name of the
#                  virtual machine relative to the hyper-v machine
################################################################################
sub _get_vm_name {
    my ($self, %args) = @_;

    my ($vm_name);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name => $vm_name");
                return "";
            }
            next;
        }
    }

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine name for ($vm_name)\n" .
                        "Reason: " . $self->errmsg());
        return "";
    }

    # search through the output until we find the virtual machine we are looking for
    # A successful match is if $name =~ /$vm_name\s*$/
    foreach my $name (@vm_list) {
        if ($name =~ /$vm_name\s*$/) {
            return $name;
        }
    }

    $self->_set_err("Could not obtain virtual machine name for ($vm_name)");
    return "";
}

################################################################################
# Subroutine Name: _get_vm_list()
# Objective:       return back the list of virtual machines
################################################################################
sub _get_vm_list {
    my ($self) = @_;

    my ($status, @output);

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    # from the cache if possible
    if ($self->is_cache_data_valid("get_vm_list")) {
        @output = $self->get_cache_data("get_vm_list");
    } else {
        ($status, @output) = $self->hvcontrol(-get_vm_list => 1);
        if ($self->error()) {
            $self->_set_err("Could not obtain virtual machine list.\n" .
                            "Reason: " . $self->errmsg());
            return (wantarray() ? @output : \@output);
        }

        $self->store_cache_data("get_vm_list", @output);
    }

    return (wantarray() ? @output : \@output);
}

################################################################################
# Subroutine Name: _get_vm_state()
# Objective:       Given a virtual machines name, determine its current state.
################################################################################
sub _get_vm_state {
    my ($self, %args) = @_;

    my ($vm_name, $full_vm_name);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name => $vm_name");
                return (wantarray() ? () : []);
            }
            next;
        } elsif (/-?full_vm_name/) {
            if (not length($full_vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -full_vm_name => $full_vm_name");
                return (wantarray() ? () : []);
            }
            next;
        }
    }

    # one of these options must be specified
    if (not defined ($vm_name) and not defined($full_vm_name)) {
        $self->_set_err("Parameters -vm_name or -full_vm_name must be specified.");
        return (wantarray() ? () : []);
    }

    # if the full vm name was not passed in then make the translation.
    if (defined($vm_name) and not defined($full_vm_name)) {
        # obtain the virtual machine's full name from the Hyper V host
        $full_vm_name = $self->_get_vm_name(-vm_name => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not find virtual machine ($vm_name).\n" .
                            "Reason: " . $self->errmsg());
            return (wantarray() ? () : []);
        }
    }

    # determine the state of this virtual machine
    my ($status, @output) = $self->hvcontrol(-vm_name => $full_vm_name,
                                             -get_state => 1);
    if ($self->error()) {
        $self->_set_err("Failed to get the state of virtual machine ($full_vm_name).\n" .
                        "Reason: " . $self->errmsg());
        return (wantarray() ? () : []);
    }

    # return the output of the command.
    return (wantarray() ? @output : \@output);
}

################################################################################
# Subroutine Name: hvcontrol()
# Objective:       interface to the hvcontrol.exe utility in WinClient
################################################################################
sub hvcontrol {
    my ($self, %args) = @_;

    $self->_clear_err();

    my @commands;
    my ($hv_host, $vhd_path, $vhd_type, $vhd_size_bytes, $controller,
        $vm_name, $disk_id, $timeout);
    $timeout = 600;

    # parse arguments
    foreach (keys %args) {
        if (/^-?hv_host$/) {
            if (not length($hv_host = $args{$_})) {
                $self->_set_err("Invalid value for -hv_host => '$hv_host'");
                return ($ClientUtil::FALSE);
            } else {
                push(@commands, "/hv_host=\"$hv_host\"");
            }
            next;
        } elsif (/^-?get_vm_list$/) {
            push(@commands, "/get_vm_list");
            next;
        } elsif (/^-?power_on_all_vms$/) {
            push(@commands, "/power_on_all_vms");
            next;
        } elsif (/^-?power_off_all_vms$/) {
            push(@commands, "/power_off_all_vms");
            next;
        } elsif (/^-?shutdown_all_vms$/) {
            push(@commands, "/shutdown_all_vms");
            next;
        } elsif (/^-?get_state_all_vms$/) {
            push(@commands, "/get_state_all_vms");
            next;
        } elsif (/^-?create_vhd$/) {
            push(@commands, "/create_vhd");
            next;
        } elsif (/^-?vhd_path$/) {
            if (not length($vhd_path = $args{$_})) {
                $self->_set_err("Invalid value for -vhd_path => '$vhd_path'");
                return ($ClientUtil::FALSE);
            } else {
                push(@commands, "/vhd_path=\"$vhd_path\"");
            }
            next;
        } elsif (/^-?vhd_type$/) {
            if (not length($vhd_type = $args{$_})) {
                $self->_set_err("Invalid value for -vhd_type => '$vhd_type'");
                return ($ClientUtil::FALSE);
            } else {
                push(@commands, "/vhd_type=\"$vhd_type\"");
            }
            next;
        } elsif (/^-?vhd_size_bytes$/) {
            if (not length($vhd_size_bytes = $args{$_})) {
                $self->_set_err("Invalid value for -vhd_size_bytes => '$vhd_size_bytes'");
                return ($ClientUtil::FALSE);
            } else {
                push(@commands, "/vhd_size_bytes=\"$vhd_size_bytes\"");
            }
            next;
        } elsif (/^-?vm_name$/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name => '$vm_name'");
                return ($ClientUtil::FALSE);
            } else {
                push(@commands, "/vm_name=\"$vm_name\"");
            }
            next;
        }  elsif (/^-?get_state$/) {
            push(@commands, "/get_state");
            next;
        }  elsif (/^-?power_on$/) {
            push(@commands, "/power_on");
            next;
        }  elsif (/^-?power_off$/) {
            push(@commands, "/power_off");
            next;
        }  elsif (/^-?shutdown$/) {
            push(@commands, "/shutdown");
            next;
        }  elsif (/^-?get_vhd_list$/) {
            push(@commands, "/get_vhd_list");
            next;
        }  elsif (/^-?get_passthrough_list$/) {
            push(@commands, "/get_passthrough_list");
            next;
        }  elsif (/^-?create_add_vhd$/) {
            push(@commands, "/create_add_vhd");
            next;
        }  elsif (/^-?controller$/) {
            if (not length($controller = $args{$_})) {
                $self->_set_err("Invalid value for -controller => '$controller'");
                return ($ClientUtil::FALSE);
            } else {
                push(@commands, "/controller=\"$controller\"");
            }
            next;
        }  elsif (/^-?add_vhd$/) {
            push(@commands, "/add_vhd");
            next;
        }  elsif (/^-?remove_vhd$/) {
            push(@commands, "/remove_vhd");
            next;
        }  elsif (/^-?add_passthrough_disk$/) {
            push(@commands, "/add_passthrough_disk");
            next;
        }  elsif (/^-?remove_passthrough_disk$/) {
            push(@commands, "/remove_passthrough_disk");
            next;
        }  elsif (/^-?remove_empty_drive_slots$/) {
            push(@commands, "/remove_empty_drive_slots");
            next;
        }  elsif (/^-?disk_id$/) {
            if (not length($disk_id = $args{$_})) {
                $self->_set_err("Invalid value for -disk_id => '$disk_id'");
                return ($ClientUtil::FALSE)
            } else {
                push(@commands, "/disk_id=\"$disk_id\"");
            }
            next;
        }  elsif (/^-?version$/) {
            push(@commands, "/version");
            next;
        }  elsif (/^-?timeout$/) {
            if (not length($timeout = $args{$_})) {
                $self->_set_err("Invalid value for -timeout => '$timeout'");
                return ($ClientUtil::FALSE)
            }
            next;
        }
    }

    my $cmd = "";

    # build out the command line from all the commands
    foreach my $command (@commands) {
        $cmd .= $command . " ";
    }
    chop($cmd);

    # invoke the WinClient interface to the hvcontrol utility
    my ($status, @output) = $self->{WinClient}->hvcontrol($cmd, $timeout);
    if ($self->{WinClient}->error()) {
        my $errmsg = $self->{WinClient}->errmsg();
        $self->_set_err("WinClient::hvcontrol() failed.\n" .
                        "Reason: $errmsg");
        return ($status, @output);
    } elsif ($status != 0) {
        $self->_set_err("WinClient::hvcontrol() failed. Status was non-zero (status=$status)");
        return ($status, @output);
    }

    # need this because hvcontrol.exe has a new line at the beginning
    # of the program, need to remove from C# code at some point
    shift(@output);

    # if we see a known error in the output of hvcontrol.exe then set an error.
    if (grep(/System\.Exception|ERROR|Invalid option|Insufficient or invalid arguments/i, @output)) {
        $self->_set_err("The command '$cmd' generated a known error string: (@output)");
        return ($status, @output);
    }

    return ($status, @output);
}

################################################################################
# Subroutine Name: _wait_for_vms_to_shutdown()
# Objective:       Wait for the specified virtual machines to shutdown.
################################################################################
sub _wait_for_vms_to_shutdown {
    my ($self, @waiting_list_vms) = @_;

    $self->_clear_err();

    # number of times to check state until deciding to force a shutdown
    my $num_retries = 10;
    my $num_vms = scalar(@waiting_list_vms);

    # iterate through the waiting_list_vms until either the retry count is
    # zero, or all VMs have been shutdown.
    while (scalar @waiting_list_vms) {

        # if the number of retries is 0
        if ($num_retries <= 0) {
            # then force shutdown on all config files remaining
            # in @waiting_list_vms
            last;
        }

        # pull this vm off the config file list
        my $full_vm_name = shift @waiting_list_vms;
        $num_vms -= 1;

        # Check the state of this VM
        my (@vm_state_output) = $self->_get_vm_state(-full_vm_name => $full_vm_name);
        if ($self->error()) {
            $self->_set_err("Could not get the state of virtual machine ($full_vm_name).\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # if this machine is still on, push it back on the list
        # if this machine is not on, then don't push it
        if (not grep(/Stopped/i, @vm_state_output)) {
            push(@waiting_list_vms, $full_vm_name);
        }

        # if num_vms has gone to zero, then update its value, decrement the retry
        # count, and sleep 20 seconds before checking the state of all vms still
        # on the list.
        if ($num_vms <= 0) {
            $num_vms = scalar(@waiting_list_vms);
            $num_retries -= 1;
            &SanUtil::nate_sleep(20);
        }
    }

    # force shutdown any VMs that are still in the list
    foreach my $full_vm_name (@waiting_list_vms) {

        my ($status, @output) = $self->hvcontrol(-power_off => 1,
                                                 -vm_name   => $full_vm_name);
        if ($self->error()) {
            # if the force shutdown failed, then make sure the vm is powered off.
            my $errmsg = $self->errmsg();
            $self->_clear_err();

            # get the state
            my (@vm_state_output) = $self->_get_vm_state(-full_vm_name => $full_vm_name);
            if ($self->error()) {
                $self->_set_err("Could not get the state of virtual machine ($full_vm_name).\n" .
                                "Reason: " . $self->errmsg());
                return $ClientUtil::FALSE;
            }

            # if the virtual machine is not powered off then this is a failure.
            if (not grep(/Stopped/, @vm_state_output)) {
                $self->_set_err("Could not force shutdown virtual machine '$full_vm_name'\n" .
                                "Reason: $errmsg\n" . $self->errmsg());
                return $ClientUtil::FALSE;
            } else {
                # clear the error and go to the next vm that is in the shutdown list
                $self->_clear_err();
                next;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _wait_for_vms_to_boot()
# Objective:       This will sleep until all VMs have finished booting.
################################################################################
sub _wait_for_vms_to_boot {
    my ($self, @vm_list) = @_;

    my (@ip_list) = ();

    # go through each vm and obtain its ip.
    foreach my $vm (@vm_list) {
        my $tharnhost_hsh = &main::host($vm);
        my $ip = $tharnhost_hsh->{default_ip};
        push(@ip_list, $ip);
    }

    my $num_retries = 10;
    my $num_ips = scalar(@ip_list);

    $self->sanlog(-msg => "Waiting for the following Virtual Machines to boot.\n " .
                          join("\n ", @vm_list));

    # cycle through the ip_list, pinging them until they boot up
    while (scalar(@ip_list)) {

        # if our retries are up, then break out of this loop
        if ($num_retries <= 0) {
            last;
        }

        # shift off an ip
        my $ip = shift @ip_list;
        my $vm = shift @vm_list;
        $num_ips -= 1;

        # ping the ip
        my $ping_results = &SanUtil::ping_host($ip);

        # if this ip could not be pinged, push it back on the list.
        if ($ping_results == $SanUtil::FALSE) {
            $self->sanlog(-msg => "The virtual machine '$vm' has not booted yet.");
            push (@ip_list, $ip);
            push (@vm_list, $vm);
        } else {
            $self->sanlog(-msg => "The virtual machine '$vm' is pingable.");
        }

        # if num_ips has gone to zero, then update its value to the size
        # of the ip list. Decrement the retry count, and sleep.
        if ($num_ips <= 0) {
            $num_ips = scalar(@ip_list);
            $num_retries -= 1;
            &SanUtil::nate_sleep(20);
        }
    }

    # if no ips are left in the list, all vms have booted.
    if (scalar(@ip_list) == 0) {
        $self->sanlog(-msg => "All VMs have booted.");
    } else {
        # then these vm's did not finish booting in time, display a warning
        $self->sanlog(-msg => "WARNING: The following VMs took too long ".
                              "to boot: |@vm_list|. Continuing... ");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_nfs_mounts
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_mounts {
    my ($self, %args) = @_;

    $self->_set_err("Not supported by WinHost");
    return (wantarray()) ? () : [];
}

################################################################################
# Method:    deconfigure
# Objective: See WinHost.pm::deconfigure()
# Details:   See POD documentation at the end of this file
################################################################################

################################################################################
# Subroutine Name: get_nfs_share
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_share {
    my ($self, %args) = @_;

    $self->_set_err("Not supported by WinHost");
    return (wantarray()) ? () : [];
}


1;
