# Copyright (c) 2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author hrivera@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::KVMHost;

# Package Requirements
######################
use File::Basename;
use San::SanClient;
use San::SanClient::Host::LinuxHost;
use San::SanClient::ClientUtil;

use San::SanLog;
use Error;
use vars qw( @ISA );

@ISA = qw(San::SanClient::Host::LinuxHost Error SanLog SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new LinuxHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($device_str, @devices, %hRaw, $virtualization, $obj_template);
    my ($self, $conn, $nfs_mount, $cmd, $status, $hosttype, $protocol, @output, $service);

    $self = {};
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for -protocol: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Invalid value for argument '-conn'");
        return $self;
    }

    if ($protocol != /fcp|iscsi/i) {
        $self->_set_err("Invalid value for argument '-protocol'." .
                        " It can be either fcp or iscsi");
        return $self;
    }

    $self->{protocol} = $protocol;
    $self->{conn}     = $conn;
    $self->{virtualization}     = $virtualization;
    $self->{nfs_mount} = undef;

    $self->store_OS_major_and_minor_revisions();
    if ($self->error) {
        return $self;
    }

    $self->sanlog(-msg => "OS version: " . $self->{major_revision});

    # store TEST_BIN directory
    $self->store_test_bin_dir();
    if ($self->error) {
        return $self;
    }

    # There is difference between iscsi CMDs for rhel4,ol4 & sles9 and
    # sles10,ol5 and rhel5 so We need to store iSCSI
    # version in a variable to differentiate between the iSCSI versions.
    # Depending on the OS version we can use
    # the iscsi CMDs. before sles10, ol5 and rhel5 we are going
    # to use commands for "linux-iscsi" and from rhel5,ol5
    # sles10 and we are going to use cammands "open-iscsi".
    # If OS version are before sles10, ol5 and rhel5 the following
    # variable will have value "linux-iscsi" else it will have value "open-iscsi"
    if ($self->{major_revision} =~ /rhel4|sles9/) {
        $self->{iscsi}->{iscsi_version} = "linux-iscsi";
    } else {
        $self->{iscsi}->{iscsi_version} = "open-iscsi";
    }
    my $iscsi_version = $self->{iscsi}->{iscsi_version};

    # By default 'raw' module is not loaded on SLES. We need to load
    # it. It will create master raw device '/dev/rawctl'.
    # On Redhat there is no raw module and device /dev/rawctl will already
    # be there
    if(($self->hosttype() =~ /suse|sles/i)) {
        $cmd = "modprobe raw";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return $self;
        }
    }

    #  Create, load, and manage hash of raw device bindings.
    #  Keep load order.
    tie %hRaw, "Tie::IxHash";

    $self->_load_raw_bindings(\%hRaw);
    if ($self->error) {
        return $self;
    }
    $self->{raw_bindings} = \%hRaw;

    # store client kernel version
    $cmd = "uname -r";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    $self->{revision} = $output[0];

    # need to load config params right now so we can access them
    # for decision on software or hardware initiator
    $self->_load_config_params();
    if ($self->error()) {
        return $self;
    }

    if ($self->{protocol} =~ /iscsi/i and $self->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /hardware/i) {
        my $running = $self->_iscsi_service(-action => "status");
        if ($self->error) {
            return $self;
        }

        if ($running =~ /offline/) {
            # iscsi will not start if iscsi.conf doesnt contain anything which is usually the case with newly install system
            # HACK : Add a non-empty line (unique) to /etc/iscsi.conf file and iscsi will start
            # We will remove this line during iscsi configure.
            # This is only for linux-iscsi because in open-iscsi case iscsi.conf will not be empty even
            # on the newly installed system
            if ($self->{iscsi}->{iscsi_version} =~ /linux-iscsi/) {
                $cmd = "cat /etc/iscsi.conf |grep -v \"^#\"|grep -v \"^\$\"";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if (scalar(@output) == 0) {
                    my $line = "remove_this_line_during_iscsi_configuration";
                    $cmd = "echo $line >> /etc/iscsi.conf";
                    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status > 0) {
                        $self->_set_err("Unable to add $line to /etc/iscsi.conf file");
                        return $self;
                    }
                }
            }

            $self->_iscsi_service(-action => 'start');
            if ($self->error) {
                return $self;
            }
        }
    }

    return $self;
}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            if (not defined ($cfmode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->{valid} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: configure
# Objective: NOOP for VLinuxHost
# Details: See POD documentation at the end of this file
################################################################################

################################################################################
# Subroutine Name: deconfigure
# Objective: NOOP for VLinuxHost
# Details: See POD documentation at the end of this file
################################################################################

################################################################################
# Subroutine Name: discover
# Objective: Discover devices
# Details: See POD documentation at the end of this file
################################################################################

################################################################################
# Subroutine Name: format
# Objective: format devices
# Details: See POD documentation at the end of this file
################################################################################
sub format {
    my ($self, %args) = @_;
    my ($devices);

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Check that required parameter '-devices' was provided
    unless (defined $devices and ref($devices) eq "ARRAY"){
        $self->_set_err("Missing required argument '-devices'");
        return $ClientUtil::FALSE;
    }

    # pull out the dev manager object
    my $devmgr_obj = $self->{obj}{devmgr};

    # iterate through the devices and partition each one
    foreach my $dev_name (@{$devices}) {

        # convert the device name to the device object
        my $dev_obj = $devmgr_obj->get_device($dev_name);
        if ($devmgr_obj->error()) {
            $self->_set_err($devmgr_obj->errmsg());
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "Partitioning device '$dev_name'");

        $dev_obj->partition();
        if ($dev_obj->error()) {
            $self->_set_err($dev_obj->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_vhd_file
# Objective:       Will create a file that will be exported to VMs
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vhd_file {
    my ($self, %args) = @_;

    my ($conn, $cmd, $status, $hosttype, $protocol, @output, $service);
    my ($location, $name, $size);

    # parse arguments
    foreach (keys %args) {
        if (/-?location/) {
            if (not length($location = $args{$_})) {
                $self->_set_err("Invalid value for -location: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Invalid value for -name: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?size/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Invalid value for -size: $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    my $of = $location . "/" . $name;
    $cmd = "touch $of";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 900);
    if ($status > 0) {
        $self->_set_err("Failed to create VHD file in dir $location");
        return $ClientUtil::FALSE;
    }

    my $bin_dir = $self->{TEST_BIN};
    $cmd = $bin_dir . "/dt of=$of limit=$size iotype=sequential bs=1m oflags=trunc oncerr=abort disable=pstats";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 900);
    if ($status > 0) {
        $self->_set_err("Failed to create VHD file in dir $location");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: virshcontrol()
# Objective:       Will perform passed action on VMs
################################################################################
sub virshcontrol {
    my ($self, %args) = @_;

    my ($action, $vm_name, $cmd, $status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?action/) {
            if (not length($action = $args{$_})) {
                $self->_set_err("Invalid value for -action => $action");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name => $vm_name");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "virsh $action $vm_name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 900);
    if ($status > 0) {
        $self->_set_err("Failed to perform action $action on VM $vm_name");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: domain_management()
# Objective:       method to allow domain management for virtual machines
################################################################################
sub domain_management {
    my ($self, %args) = @_;

    my ($action);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?action/) {
            if (not length($action = $args{$_})) {
                $self->_set_err("Invalid value for -action => $action");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if ($action =~ /^undefine$/i) {
        $self->_undefine_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^undefine_inactive$/i) {
        $self->_undefine_all_inactive_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^destroy$/i) {
        $self->_destroy_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^destroy_all$/i) {
        $self->_destroy_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^resume$/i) {
        $self->_resume_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^resume_suspended$/i) {
        $self->_resume_all_suspended_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^suspend$/i) {
        $self->_suspend_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^suspend_all$/i) {
        $self->_suspend_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_on$/i) {
        $self->_power_on_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_on_all$/i) {
        $self->_power_on_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_off$/i) {
        $self->_power_off_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_off_all$/i) {
        $self->_power_off_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^shutdown$/i) {
        $self->_shutdown_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^shutdown_all$/i) {
        $self->_shutdown_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_cycle$/i) {
        $self->_power_cycle_vms(%args);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    } elsif ($action =~ /^power_cycle_all$/i) {
        $self->_power_cycle_all_vms();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_vm_list()
# Objective:       Will return VM list.
################################################################################
sub _get_vm_list {
    my ($self) = @_;

    my ($cmd, $status, @output);
    my (@vm_list) = ();

    if (defined $self->{vm_list}) {
        @vm_list = @{$self->{vm_list}};
        if (scalar(@vm_list) > 0) {
            return (wantarray() ? @vm_list : \@vm_list);
        }
    }

    $cmd = "virsh list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to list guests");
        return (wantarray() ? () : []);
    }
    ###########################################################################
    # Here is the output of above command
    ###########################################################################
    #[root@IBMx3550-200-133 ~]# virsh list
    # Id Name                 State
    #----------------------------------
    # 16 rhel5_5              paused
    ###########################################################################
    foreach my $line (@output) {
        # Skip first 2 lines
        next if ($line =~ /^\s*Id\s+Name/i);
        next if ($line =~ /^\-+$/);
        # Skip empty lines
        next if ($line =~ /^$/);

        my ($domain) = ($line =~ /^\s*\d+\s+(\S+)\s+/);
        push (@{$self->{vm_list}}, $domain);
    }

    if (defined $self->{vm_list}) {
        @vm_list = @{$self->{vm_list}};
    }
    return (wantarray() ? @vm_list : \@vm_list);
}

################################################################################
# Subroutine Name: _define_vms()
# Objective:       Define domains from XML(file) list. The domains' definition are
#                  registered but not started.
################################################################################
sub _define_vms {
    my ($self, %args) = @_;

    my (@xml_file_list) = ();
    my ($cmd ,$status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?xml_file_list/) {
            if (not scalar(@xml_file_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -xml_file_list => @xml_file_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@xml_file_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # Defining VM from xml file
    foreach my $xml_file_name (@xml_file_list) {
        $self->sanlog(-msg => "Defining VM from xml file ($xml_file_name).");
        $cmd = "virsh define $xml_file_name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to Define VM from xml file ($xml_file_name).");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: dumpxml_vms_to_file()
# Objective:       Will Dump VM configuration to an xml file
################################################################################
sub dumpxml_vms_to_file {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # Make sure following location exists
    $cmd = "mkdir -p /var/tmp/nate_kvm_guest_xml";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to create dir /var/tmp/nate_kvm_guest_xml");
        return $ClientUtil::FALSE;
    }

    foreach my $vm_name (@vm_list) {
        $self->sanlog(-msg => "Dumping VM configuration to an xml file (/var/tmp/nate_kvm_guest_xml/$vm_name).");

        # VM's datastructure will be dumped into following XML file
        $cmd = "touch /var/tmp/nate_kvm_guest_xml/$vm_name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to create xmldump for the guest $vm_name");
            return $ClientUtil::FALSE;
        }

        # Will override the file if already exists
        $cmd = "virsh dumpxml $vm_name > /var/tmp/nate_kvm_guest_xml/$vm_name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to Create VM from xml file (/var/tmp/kvm_guest_xml/$vm_name).");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: dumpxml_all_vms_to_file()
# Objective:       Power on every virtual machine.
################################################################################
sub dumpxml_all_vms_to_file {
    my ($self) = @_;

    $self->_clear_err();

    # Obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # Dump VM's datastructure into XML files
    $self->dumpxml_vms_to_file(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not dump configuration of all virtual machines to an xml file\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _resume_vms()
# Objective:       Power on the specified virtual machines
################################################################################
sub _resume_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # power on each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {

        # get the state of this virtual machine, if its paused, then resume
        my $vm_state = $self->_get_vm_state(-vm_name => $vm_name);
        if ($vm_state =~ /paused/i) {
            $self->sanlog(-msg => "Resuming VM ($vm_name).");

            my ($status, @output) = $self->virshcontrol(-action => "resume",
                                                     -vm_name  => $vm_name);
            if ($self->error()) {
                $self->_set_err("Could not resume VM $vm_name.\n" .
                                "Reason: " . $self->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            $self->sanlog(-msg => "VM $vm_name is not paused");
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _resume_all_suspended_vms()
# Objective:       Power on every virtual machine.
################################################################################
sub _resume_all_suspended_vms {
    my ($self) = @_;

    my (@vm_list_suspended);
    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    foreach my $vm_name (@vm_list) {
        my $vm_state = $self->_get_vm_state(-vm_name => $vm_name);
        if ($vm_state =~ /paused/i) {
            push (@vm_list_suspended, $vm_name);
        }
    }

    # turn on all the virtual machines.
    $self->_resume_vms(-vm_list => \@vm_list_suspended);
    if ($self->error()) {
        $self->_set_err("Could not resume all suspened virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _suspend_vms()
# Objective:       Power on the specified virtual machines
################################################################################
sub _suspend_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # power on each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {

        # get the state of this virtual machine, if its not powered off, then skip
        my $vm_state = $self->_get_vm_state(-vm_name => $vm_name);
        if ($vm_state =~ /paused|shut\s+off/i) {
            $self->sanlog(-msg => "Virtual Machine ($vm_name) is already suspended or destroyed.");
            next;
        }

        $self->sanlog(-msg => "Suspending ($vm_name).");

        my ($status, @output) = $self->virshcontrol(-action => "suspend",
                                                 -vm_name  => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not suspend VM $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _suspend_all_vms()
# Objective:       Power on every virtual machine.
################################################################################
sub _suspend_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # turn on all the virtual machines.
    $self->_suspend_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not suspend all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _destroy_vms()
# Objective:       Power on the specified virtual machines
################################################################################
sub _destroy_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # power off each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {

        # get the state of this virtual machine, if its not powered off, then skip
        my $vm_state = $self->_get_vm_state(-vm_name => $vm_name);
        if ($vm_state =~ /shut\s+off/i) {
            $self->sanlog(-msg => "Virtual Machine ($vm_name) is already destroyed.");
            next;
        }

        $self->sanlog(-msg => "Destroying ($vm_name).");

        my ($status, @output) = $self->virshcontrol(-action => "destroy",
                                                    -vm_name  => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not destroy VM $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _destroy_all_vms()
# Objective:       Power on every virtual machine.
################################################################################
sub _destroy_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # turn on all the virtual machines.
    $self->_destroy_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not destroy all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _undefine_vms()
# Objective:       Power on the specified virtual machines
################################################################################
sub _undefine_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # power on each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {

        # get the state of this virtual machine, if its not powered off, then skip
        my $vm_state = $self->_get_vm_state(-vm_name => $vm_name);
        if ($vm_state =~ /running|paused/i) {
            $self->_ser_err("Can not undefine active/paused Domain($vm_name) " .
                            "Make sure domain is either destroyed or shutdown");
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "Undefining ($vm_name).");

        my ($status, @output) = $self->virshcontrol(-action => "undefine",
                                                 -vm_name  => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not undefine VM $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _undefine_all_inactive_vms()
# Objective:       Power on every virtual machine.
################################################################################
sub _undefine_all_inactive_vms {
    my ($self) = @_;

    my (@vm_list_inactive);
    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    foreach my $vm_name (@vm_list) {
        my $vm_state = $self->_get_vm_state(-vm_name => $vm_name);
        if ($vm_state =~ /shut\s+off/i) {
            push (@vm_list_inactive, $vm_name);
        }
    }

    # turn on all the virtual machines.
    $self->_undefine_vms(-vm_list => \@vm_list_inactive);
    if ($self->error()) {
        $self->_set_err("Could not undefine all inactive virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_on_vms()
# Objective:       Power on the specified virtual machines
################################################################################
sub _power_on_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my (@waiting_to_boot_vm_list);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # power on each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {

        # get the state of this virtual machine, if its not powered off, then skip
        my $vm_state = $self->_get_vm_state(-vm_name => $vm_name);
        if ($vm_state =~ /running/i) {
            $self->sanlog(-msg => "Virtual Machine ($vm_name) is already powered on.");
            next;
        }

        $self->sanlog(-msg => "Powering on ($vm_name).");

        # power on this virtual machine
        my ($status, @output) = $self->virshcontrol(-action => "start",
                                                 -vm_name  => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not power on $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # push this virtual machine on the waiting to boot vm list
        push(@waiting_to_boot_vm_list, $vm_name);
    }

    # wait for the virtual machines to boot
    $self->_wait_for_vms_to_boot(@waiting_to_boot_vm_list);
    if ($self->error()) {
        $self->_set_err("Failed while waiting for virtual machines to boot.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_on_all_vms()
# Objective:       Power on every virtual machine.
################################################################################
sub _power_on_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # turn on all the virtual machines.
    $self->_power_on_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not power on all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_off_vms()
# Objective:       Power off the specified virtual machines
################################################################################
sub _power_off_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my (@waiting_to_shutdown_vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # power off each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {

        # get the state of this virtual machine, if its not running then skip
        my $vm_state = $self->_get_vm_state(-vm_name => $vm_name);
        if ($vm_state =~ /shut\s+off/i) {
            $self->sanlog(-msg => "Virtual Machine ($vm_name) is already powered off.");
            next;
        }

        $self->sanlog(-msg => "Powering off ($vm_name)");

        # power off this virtual machine
        $self->virshcontrol(-action => "destroy",
                         -vm_name   => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not power off $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # push this virtual machine on the waiting to shutdown vm list
        push(@waiting_to_shutdown_vm_list, $vm_name);
    }

    # wait for the virtual machines to shutdown
    $self->_wait_for_vms_to_shutdown(@waiting_to_shutdown_vm_list);
    if ($self->error()) {
        $self->_set_err("Failed while waiting for virtual machines to power off.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }


    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_off_all_vms()
# Objective:       Power off every virtual machine.
################################################################################
sub _power_off_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # turn off all the virtual machines.
    $self->_power_off_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not power off all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _shutdown_vms()
# Objective:       Gracefully Shutdown the specified virtual machines
################################################################################
sub _shutdown_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my (@waiting_to_shutdown_vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vm_list)) {
        $self->_set_err("Missing or invalid value for -vm_list");
        return $ClientUtil::FALSE;
    }

    # Shutdown each virtual machine in the @vm_list array
    foreach my $vm_name (@vm_list) {

        # get the state of this virtual machine, if its not running then skip
        my $vm_state = $self->_get_vm_state(-vm_name => $vm_name);
        if ($vm_state =~ /shut\s+off/i) {
            $self->sanlog(-msg => "Virtual Machine ($vm_name) is already powered off.");
            next;
        }

        $self->sanlog(-msg => "Shutting down ($vm_name)");

        # shutdown this virtual machine
        $self->virshcontrol(-action => "shutdown",
                         -vm_name   => $vm_name);
        if ($self->error()) {
            $self->_set_err("Could not power off $vm_name.\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # push this virtual machine on the waiting to shutdown vm list
        push(@waiting_to_shutdown_vm_list, $vm_name);
    }

    # wait for the virtual machines to shutdown
    $self->_wait_for_vms_to_shutdown(@waiting_to_shutdown_vm_list);
    if ($self->error()) {
        $self->_set_err("Failed while waiting for virtual machines to power off.\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _shutdown_all_vms()
# Objective:       Shutdown every virtual machine.
################################################################################
sub _shutdown_all_vms {
    my ($self) = @_;

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    my (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }
    # shutdown all the virtual machines.
    $self->_shutdown_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("Could not shutdown all virtual machines\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_cycle_vms()
# Objective:       Power off/Power on the specified virtual machines
################################################################################
sub _power_cycle_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();
    my ($soft_shutdown) = "no";

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }  elsif (/-?soft_shutdown/) {
            if (not length($soft_shutdown = $args{$_})) {
                $self->_set_err("Invalid value for -soft_shutdown => $soft_shutdown");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->sanlog(-msg => "Power Cycling the following Virtual machines (graceful: $soft_shutdown):\n " .
                          join("\n ", @vm_list));

    # power off/shutdown the specified VMs
    if ($soft_shutdown eq "no") {
        $self->_power_off_vms(-vm_list => \@vm_list);
    } else {
        $self->_shutdown_vms(-vm_list => \@vm_list);
    }
    if ($self->error()) {
        $self->_set_err("power_off_vms/shutdown_vms failed\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # power on the specified VMs
    $self->_power_on_vms(-vm_list => \@vm_list);
    if ($self->error()) {
        $self->_set_err("power_on_vms failed\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Power Cycle Complete");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _power_cycle_all_vms()
# Objective:       Power off/Power on all virtual machines
################################################################################
sub _power_cycle_all_vms {
    my ($self, %args) = @_;

    my (@vm_list) = ();

    $self->_clear_err();

    # obtain all the names of all virtual machines on the system
    (@vm_list) = $self->_get_vm_list();
    if ($self->error()) {
        $self->_set_err("Could not obtain virtual machine list\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # power cycle all the virtual machines
    $self->_power_cycle_vms(-vm_list => \@vm_list, %args);
    if ($self->error()) {
        $self->_set_err("Power Cycle failed\n" .
                        "Reason: " . $self->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_vm_state()
# Objective:       Given a virtual machines name, determine its current state.
################################################################################
sub _get_vm_state {
    my ($self, %args) = @_;

    my ($vm_name, $cmd, $status, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name => $vm_name");
                return "";
            }
            next;
        }
    }

    # one of these options must be specified
    if (not defined ($vm_name)) {
        $self->_set_err("Parameter -vm_name must be specified.");
        return "";
    }

    $cmd = "virsh domstate $vm_name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to create xmldump for the guest $vm_name");
        return "";
    }

    return $output[0];
}

################################################################################
# Subroutine Name: _wait_for_vms_to_shutdown()
# Objective:       Wait for the specified virtual machines to shutdown.
################################################################################
sub _wait_for_vms_to_shutdown {
    my ($self, @waiting_list_vms) = @_;

    $self->_clear_err();

    # number of times to check state until deciding to force a shutdown
    my $num_retries = 10;
    my $num_vms = scalar(@waiting_list_vms);

    # iterate through the waiting_list_vms until either the retry count is
    # zero, or all VMs have been shutdown.
    while (scalar @waiting_list_vms) {

        # if the number of retries is 0
        if ($num_retries <= 0) {
            # then force shutdown on all config files remaining
            # in @waiting_list_vms
            last;
        }

        # pull this vm off the config file list
        my $full_vm_name = shift @waiting_list_vms;
        $num_vms -= 1;

        # Check the state of this VM
        my $vm_state = $self->_get_vm_state(-vm_name => $full_vm_name);
        if ($self->error()) {
            $self->_set_err("Could not get the state of virtual machine ($full_vm_name).\n" .
                            "Reason: " . $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # if this machine is still on, push it back on the list
        # if this machine is not on, then don't push it
        if ($vm_state =~ /running/i) {
            push(@waiting_list_vms, $full_vm_name);
        }

        # if num_vms has gone to zero, then update its value, decrement the retry
        # count, and sleep 20 seconds before checking the state of all vms still
        # on the list.
        if ($num_vms <= 0) {
            $num_vms = scalar(@waiting_list_vms);
            $num_retries -= 1;
            &SanUtil::nate_sleep(20);
        }
    }

    # force shutdown any VMs that are still in the list
    foreach my $full_vm_name (@waiting_list_vms) {

        my ($status, @output) = $self->virshcontrol(-action => "destroy",
                                                 -vm_name   => $full_vm_name);
        if ($self->error()) {
            # if the force shutdown failed, then make sure the vm is powered off.
            my $errmsg = $self->errmsg();
            $self->_clear_err();

            # get the state
            my $vm_state = $self->_get_vm_state(-vm_name => $full_vm_name);
            if ($self->error()) {
                $self->_set_err("Could not get the state of virtual machine ($full_vm_name).\n" .
                                "Reason: " . $self->errmsg());
                return $ClientUtil::FALSE;
            }

            # if the virtual machine is not powered off then this is a failure.
            if ($vm_state =~ /running/i) {
                $self->_set_err("Could not force shutdown virtual machine '$full_vm_name'\n" .
                                "Reason: $errmsg\n" . $self->errmsg());
                return $ClientUtil::FALSE;
            } else {
                # clear the error and go to the next vm that is in the shutdown list
                $self->_clear_err();
                next;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _wait_for_vms_to_boot()
# Objective:       This will sleep until all VMs have finished booting.
################################################################################
sub _wait_for_vms_to_boot {
    my ($self, @vm_list) = @_;

    my (@ip_list) = ();

    # go through each vm and obtain its ip.
    foreach my $vm (@vm_list) {
        my $tharnhost_hsh = &main::host($vm);
        my $ip = $tharnhost_hsh->{default_ip};
        push(@ip_list, $ip);
    }

    my $num_retries = 10;
    my $num_ips = scalar(@ip_list);

    $self->sanlog(-msg => "Waiting for the following Virtual Machines to boot.\n " .
                          join("\n ", @vm_list));

    # cycle through the ip_list, pinging them until they boot up
    while (scalar(@ip_list)) {

        # if our retries are up, then break out of this loop
        if ($num_retries <= 0) {
            last;
        }

        # shift off an ip
        my $ip = shift @ip_list;
        my $vm = shift @vm_list;
        $num_ips -= 1;

        # ping the ip
        my $ping_results = &SanUtil::ping_host($ip);

        # if this ip could not be pinged, push it back on the list.
        if ($ping_results == $SanUtil::FALSE) {
            $self->sanlog(-msg => "The virtual machine '$vm' has not booted yet.");
            push (@ip_list, $ip);
            push (@vm_list, $vm);
        } else {
            $self->sanlog(-msg => "The virtual machine '$vm' is pingable.");
        }

        # if num_ips has gone to zero, then update its value to the size
        # of the ip list. Decrement the retry count, and sleep.
        if ($num_ips <= 0) {
            $num_ips = scalar(@ip_list);
            $num_retries -= 1;
            &SanUtil::nate_sleep(20);
        }
    }

    # if no ips are left in the list, all vms have booted.
    if (scalar(@ip_list) == 0) {
        $self->sanlog(-msg => "All VMs have booted.");
    } else {
        # then these vm's did not finish booting in time, display a warning
        $self->sanlog(-msg => "WARNING: The following VMs took too long ".
                              "to boot: |@vm_list|. Continuing... ");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remove_devices_from_vms()
# Objective:       Will add VHD files and PT devices to VM
# Details:         See POD documentation at the end of this file
################################################################################
sub remove_devices_from_vms {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my (@vm_list) = ();

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_list/) {
            if (not scalar(@vm_list = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vm_list => @vm_list");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    foreach my $vm_name (@vm_list) {
        $self->remove_devices_from_vm(-vm_name => $vm_name);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remove_devices_from_vm()
# Objective:       Will add VHD files and PT devices to VM
# Details:         See POD documentation at the end of this file
################################################################################
sub remove_devices_from_vm {
    my ($self, %args) = @_;

    my ($cmd, $target_node, $timeout, $status, @output, $vm_xml_dump, $initial_vm_config);
    my ($vm_name, @all_devices);

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    unless (defined $vm_name and length $vm_name) {
        $self->_set_err("Invalid value for argument '-vm_name'");
        return $ClientUtil::FALSE;
    }

    my (@vm_list) = ();
    push(@vm_list, $vm_name);

    # Dump VM's xml to a file
    $self->dumpxml_vms_to_file(-vm_list => \@vm_list);
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    # We need to set options in this way so that we get exactly what we want
    my $xs = new XML::Simple(keeproot => 1,suppressempty => 0,forcearray => 1,keyattr =>"");

    # Its not possible to read xml file directly into xml data structure. Populate xml DS with
    # some initial data
    $initial_vm_config = "<domain alias=\"$vm_name\" config_version=\"0\" name=\"$vm_name\"> </domain>";
    my $vm_domain_config = $xs->XMLin("$initial_vm_config");

    $cmd = "ls /var/tmp/nate_kvm_guest_xml/$vm_name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("xmldump file doesnt exists for guest $vm_name");
        return $ClientUtil::FALSE;
    }

    $cmd = "cat /var/tmp/nate_kvm_guest_xml/$vm_name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to display xmldump file for guest $vm_name");
        return $ClientUtil::FALSE;
    }
    foreach my $line (@output) {
        $vm_xml_dump .= $line . " ";
    }

    # Read the guest xml file into data structure
    $vm_domain_config = $xs->XMLin("$vm_xml_dump");

    my $devices = $vm_domain_config->{domain}->[0]->{devices}->[0];
    my $device_num = 0;
    my @used_device_names;
    while (defined $devices->{disk}->[$device_num]) {
        my $dev_name = $devices->{disk}->[$device_num]->{target}->[0]->{dev};

        # Remove the disk from VM if added by automation
        if ($dev_name =~ /nate$/i) {
            # We only need to remove the files that we created on the device.
            # Rest of the devices will be removed during client cleanup
            if ($devices->{disk}->[$device_num]->{type} eq "file") {
                my $file = $devices->{disk}->[$device_num]->{source}->[0]->{file};

                # Remove the file
                $cmd = "rm -f $file";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Failed to delete file $file");
                    return $ClientUtil::FALSE;
                }
            }

            # Remove device from xml file
            delete($devices->{disk}->[$device_num]);
        }
        $device_num++;
    }

    $vm_domain_config->{domain}->[0]->{devices}->[0] = $devices;

    my $vm_xml_ds = $xs->XMLout($vm_domain_config);
    $cmd = "echo '$vm_xml_ds' > /var/tmp/nate_kvm_guest_xml/$vm_name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 900);
    if ($status > 0) {
        $self->_set_err("Failed to populate xml file for guest $vm_name");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: add_devices_to_vm()
# Objective:       Will add VHD files and PT devices to VM
# Details:         See POD documentation at the end of this file
################################################################################
sub add_devices_to_vm {
    my ($self, %args) = @_;

    my ($cmd, $target_node, $timeout, $status, @output, $vm_xml_dump, $initial_vm_config);
    my ($vm_name, @all_devices);

    # parse arguments
    foreach (keys %args) {
        if (/-?all_devices/) {
            if (not scalar(@all_devices = @{$args{$_}})) {
                $self->_set_err("Invalid value for -vhd_files: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for -vm_name: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    unless (defined $vm_name and length $vm_name) {
        $self->_set_err("Invalid value for argument '-vm_name'");
        return $ClientUtil::FALSE;
    }

    my (@vm_list) = ();
    push(@vm_list, $vm_name);

    # Dump VM's xml to a file
    $self->dumpxml_vms_to_file(-vm_list => \@vm_list);
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    # We need to set options in this way so that we get exactly what we want
    my $xs = new XML::Simple(keeproot => 1,suppressempty => 0,forcearray => 1,keyattr =>"");

    # Its not possible to read xml file directly into xml data structure. Populate xml DS with
    # some initial data
    $initial_vm_config = "<domain alias=\"$vm_name\" config_version=\"0\" name=\"$vm_name\"> </domain>";
    my $vm_domain_config = $xs->XMLin("$initial_vm_config");

    $cmd = "ls /var/tmp/nate_kvm_guest_xml/$vm_name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("xmldump file doesnt exists for guest $vm_name");
        return $ClientUtil::FALSE;
    }

    $cmd = "cat /var/tmp/nate_kvm_guest_xml/$vm_name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to display xmldump file for guest $vm_name");
        return $ClientUtil::FALSE;
    }
    foreach my $line (@output) {
        $vm_xml_dump .= $line . " ";
    }

    # Read the guest xml file into data structure
    $vm_domain_config = $xs->XMLin("$vm_xml_dump");

    my $devices = $vm_domain_config->{domain}->[0]->{devices}->[0];
    my $device_num = 0;
    my @used_device_names;
    my $ide_devices = 0;
    my $virtio_devices = 0;
    my $scsi_devices = 0;

    # Count how many devices total and of each type (ide, virtio and scsi)
    # are already attached to the guest host
    while (defined $devices->{disk}->[$device_num]) {
        push (@used_device_names, $devices->{disk}->[$device_num]->{target}->[0]->{dev});
        if ($devices->{disk}->[$device_num]->{target}->[0]->{bus} eq "ide") {
            $ide_devices++;
        } elsif ($devices->{disk}->[$device_num]->{target}->[0]->{bus} eq "virtio") {
            $virtio_devices++;
        } else {
            $scsi_devices++;
        }
        $device_num++;
    }

    foreach my $device (@all_devices) {

        # Determine device name that we are going to use next
        my ($dev_name, $quotient, $reminder, $tmp_devices);
        if ($device->{controller_type} eq "ide") {
            $ide_devices++;
            $tmp_devices = $ide_devices;
            $dev_name = "hd";
        } elsif ($device->{controller_type} eq "virtio") {
            $virtio_devices++;
            $tmp_devices = $virtio_devices;
            $dev_name = "xvd";
        } else {
            $scsi_devices++;
            $tmp_devices = $scsi_devices;
            $dev_name = "sd";
        }

        # Format next device name to be attached
        my $tmp_dev_name = $dev_name;
        while (grep(/$tmp_dev_name/, @used_device_names)) {
            print "already used device is $tmp_dev_name\n";
            $tmp_dev_name = $dev_name;
            $quotient = int ($tmp_devices / 26);
            $reminder = $tmp_devices % 26;
            for (my $i = 0; $i < $quotient; $i++) {
                $tmp_dev_name .= "a";
            }
            $tmp_dev_name .= chr(ord('a') + $reminder - 1);
            $tmp_devices++;
        }
        $dev_name = $tmp_dev_name;

        # Atmost 4 ide type devices are support per guest in KVM
        if ($ide_devices > 4) {
            $self->_set_err("Atmost 4 ide type devices are support per guest in KVM");
        }

        push(@used_device_names, $dev_name);

        $devices->{disk}->[$device_num]->{target}->[0]->{bus} = $device->{controller_type};

        # Append 'nate' to differentiate from devices that were already
        # attached to guest host. Automation use this device name
        # to remove the disks from guests during cleanup
        $dev_name .= "nate";

        $devices->{disk}->[$device_num]->{target}->[0]->{dev} = $dev_name;

        # If device is attached as a file to guest
        if ($device->{type} eq "file") {
            $devices->{disk}->[$device_num]->{driver}->[0]->{name} = "qemu";
            $devices->{disk}->[$device_num]->{driver}->[0]->{cache} = "none";
            $devices->{disk}->[$device_num]->{source}->[0]->{file} = $device->{name};
            $devices->{disk}->[$device_num]->{type} = "file";
        } else {
            $devices->{disk}->[$device_num]->{source}->[0]->{dev} = $device->{name};
            $devices->{disk}->[$device_num]->{type} = "block";
        }

        # Need to specify these extra fields if device is exported ide
        if ($device->{controller_type} eq "ide") {
            $devices->{disk}->[$device_num]->{address}->[0]->{type} = "drive";
            $devices->{disk}->[$device_num]->{address}->[0]->{controller} = 0;

            $devices->{disk}->[$device_num]->{address}->[0]->{bus} = int(($ide_devices - 1) / 2);
            $devices->{disk}->[$device_num]->{address}->[0]->{unit} = ($ide_devices - 1) % 2;
        }

        $devices->{disk}->[$device_num]->{device} = "disk";
        $device_num++;
    }

    $vm_domain_config->{domain}->[0]->{devices}->[0] = $devices;

    # Dump the guest datastructure into an XML file
    my $vm_xml_ds = $xs->XMLout($vm_domain_config);
    $cmd = "echo '$vm_xml_ds' > /var/tmp/nate_kvm_guest_xml/$vm_name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 900);
    if ($status > 0) {
        $self->_set_err("Failed to populate xml file for guest $vm_name");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For KVMHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
################################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: sanlun_version()
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################

################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################

################################################################################
# Subroutine Name: disable_dmmp
# Objective:       NOOP for VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################

################################################################################
# Subroutine Name: get_supported_protocols
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: get_supported_protocols
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################

1;
