# Copyright (c) 2008 NetApp Inc.
# All rights reserved.
# $Id: //depot/prod/test/Rsierranevada/lib/San/SanClient/Host/SanBlazeHost.pm#2 $

## @summary Provide an interface to a SANBlaze machine
## @author evan@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::SanBlazeHost;

# Package Requirements
######################

use San::SanClient;
use San::SanLog;
use Error;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new SanBlazeHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($device_str, @devices);
    my ($self, $conn, $initiator_count, $sanblaze_ports, $obj_template);
    my ($cmd, $status, @output);

    $self = {};

    bless ($self, $class);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SanBlazeHost::_new() Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?sanblaze_ports/) {
            if (not length($sanblaze_ports = $args{$_})) {
                $self->_set_err("Error: SanBlazeHost::_new() Invalid value for -sanblaze_ports: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?initiator_count/) {
            if (not length($initiator_count = $args{$_})) {
                $self->_set_err("Error: SanBlazeHost::_new() Invalid value for -initiator_count: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Error: SanBlazeHost::_new() Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Error: SanBlazeHost::_new()",
                        "Invalid value for argument '-conn'");
        return $self;
    }

    $self->{ports} = defined($sanblaze_ports) ? $sanblaze_ports : 0;
    $self->{conn} = $conn;

    # Configuration of initiator count is actually by number of aliases so
    # if the user asks for five initiators, there is really four aliases in
    # additon to the original single initiator
    $self->{aliases} = defined($initiator_count) ? $initiator_count - 1 : 0;

    &Tharn::say($self->{conn}, "su", -prompt, "assword:", -timeout, 180);
    &Tharn::say($self->{conn}, "SANBlaze", -timeout, 180);


    # get architecture
    $cmd = "uname -p";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SanBlazeHost::_new() Command ($cmd) returned a bad status ($status)");
    } else {
        $self->{arch} = $output[0];
        $self->sanlog(-msg => "SanBlaze Host Architecture is $self->{arch}");
    }

    if ($self->{ports} != 0) {
        $self->configure();
        if ($self->error) {
            return $self;
        }
    }

    return $self;
}

################################################################################
# Subroutine Name: set_initiator_count
# Objective: set the number of initiators to emulate
# Details: See POD documentation at the end of this file
################################################################################
sub set_initiator_count {
	my ($self, $count) = @_;

	if ($count == 0 || !defined($count)) {
		$self->_set_err("SanBlazeHost::set_initiator_count() invalid number of initiators specified");
	}
	$self->{aliases} = $count - 1;

	$self->configure();
}

################################################################################
# Subroutine Name: configure
# Objective: configure the SanBlaze machine
# Details: See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    foreach my $port (split(/,/,$self->{ports})) {
        $cmd = "cat /iport$port/port | grep NumberOfInitiators";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: SanBlazeHost::configure() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
        my $initcount = (split(/=/, $output[0]))[1];

        if ($initcount != ($self->{aliases} + 1)) {
            $cmd = "echo NumAliases=$self->{aliases} > /port$port/port";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: SanBlazeHost::configure() Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->sanlog(-msg => "SanBlaze configuration has completed");

    $self->{configured} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: discover
# Objective: Discover devices and label if new.
# Details: See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self) = @_;

    my ($status, @output);

    $self->sanlog(-msg => "Discovering all devices");

    foreach my $port (split(/,/,$self->{ports})) {
        # Send a ALL_LUNS (2) through all initiators (-1)
	my @cmds = (#"echo AutoScan=1 > /iport$port/port",
	            "echo Offline > /port$port/port",
		    "echo Online > /port$port/port",
                    "echo Probe=2,-1 > /iport$port/port");
	foreach my $cmd (@cmds) {
		($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
		sleep(10);
		#if ($status > 0) {
		#	$self->_set_err("Error: SanBlazeHost::discover() Command ($cmd) returned a bad status ($status)");
		#	return $ClientUtil::FALSE;
		#}
	}
    }
    sleep(60);
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: format
# Objective:       N/A for SanBlaze
# Details:         See POD documentation at the end of this file
################################################################################
sub format {
    my ($self, %args) = @_;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: sync()
# Objective:       flush buffer cache
# Details:         See POD documentation at the end of this file
################################################################################
sub sync {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    $cmd = "sync";

    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SanBlazeHost::sync() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: get_mounted_device
# Objective:       N/A for SanBlaze
###############################################################################
sub get_mounted_device {
    my ($self, $mount) = @_;

    return undef;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    get_lvm_default
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lvm_default {
    my ($self) = @_;

    return qw(none);

}

################################################################################
# Subroutine Name: get_root_dev()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_root_dev {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: update_partition_table()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub update_partition_table {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: sanlun_version()
# Objective:       get sanlun version
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;

    return "";
}

###############################################################################
# Method:    info
# Objective: Get client information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;

    my (@output);

    push @output,
         "Host name:      " . $self->hostname(),
         "Host type:      " . $self->hosttype(),
         "OS version:     " . $self->os_version();

    return (wantarray()) ? @output: \@output;

}

################################################################################
# Subroutine Name: iostat
# Objective:       Get I/O statistics
# Details:         See POD documentation at the end of this file
################################################################################
sub iostat {
    my ($self, %args) = @_;

    $self->_clear_err();

    return undef;
}

################################################################################
# Subroutine Name: nfs_mount
# Objective:       NFS mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_mount {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: nfs_unmount()
# Objective:       unmount NFS volume
# Details:         See POD documentation at the end of this file
################################################################################
sub nfs_unmount {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: mount_entry_add()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_add {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unmount_fs
# Objective:       unmount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, $mntpt) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: mount_entry_delete()
# Objective:       delete entry from mount file
# Details:         See POD documentation at the end of this file
################################################################################
sub mount_entry_delete {
    my ($self, $mount) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

###############################################################################
# Subroutine Name: mount_entry_verify
# Objective:      verify mount entries
###############################################################################
sub mount_entry_verify {
    my ($self, $mount) = @_;

    $self->_clear_err();

    return undef;
}

###############################################################################
# Subroutine Name: mount_table_verify
# Objective:       verify mount table
###############################################################################
sub mount_table_verify {
    my ($self, $mount) = @_;

    $self->_clear_err();

    return undef;
}

###############################################################################
# Method:    get_supported_protocols
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_protocols {
    my ($self) = @_;

    $self->_clear_err();

    return qw(fcp);

}

###############################################################################
# Method:    get_supported_fs_types
# Objective: Get list of supported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_fs_types {
    my ($self) = @_;

    return "";

}

################################################################################
# Subroutine Name: collect_sysrq_dump_periodic
# Objective:       N/A for SanBlaze, provide for compatibility with Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_sysrq_dump_periodic {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: config_paths
# Objective:       Need for compat with HP-UX
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self) = @_;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: raw_device_setup
# Objective:       Do whatever we need to do to use raw devices successfully
#                  after the call to discover().
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_setup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       Do whatever we need to do to make sure the host is more
#                  happy when luns are removed.  For linux, we want to do
#                  an iscsi stop.
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For SanBlazeHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
################################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       When the luns are removed from the filer, we will have
#                  orphaned luns.  This causes big problems.
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: reboot()
# Objective:       reboot host
# Details:         See POD documentation at the end of this file
################################################################################
sub reboot {
    my ($self, $force) = @_;
    my (
        $cmd,
        @output,
        $status,
        );

    $self->_clear_err();

    $cmd = "reboot";

    @output = &Tharn::say($self->{conn}, $cmd, '-blind');
    $self->sanlog(-msg => "$cmd (CMD=DONE)",
                  -msg_level => 7);

    foreach my $line (@output){
        $self->sanlog(-msg => "line=$line\n",
                      -indent    => 4,
                      -msg_level => 7);
    }

    ### sleep 3 minutes to make sure reboot has occurred
    sleep(180);

    # Ping the host every 10 seconds until it comes back
    my $alive = 0;

    $cmd = "ping -c 1 " . $self->hostname();
    foreach (my $cnt = 0; $cnt < 36; $cnt++){
        $output[0] = `$cmd`;
        $self->sanlog(-msg => "Ping output = " . $output[0],
                      -msg_level => 7);
        if ($output[0] =~ /(alive|\s0% packet loss)/) {
            $self->sanlog(-msg => "SanBlazeHost::reboot() Machine is alive after reboot");
            $alive = 1;
            last;
        } else {
            sleep(10);
        }
    }

    unless($alive){
        $self->_set_err("Error: SanBlazeHost::reboot() Machine did not reboot successfully");
        return $ClientUtil::FALSE;
    }

    # sleep for a while to make sure the host is all the
    # way back up.  Sometimes ping will work before telnet
    # will work.

    sleep(120);

    # reconnect to the client
    $self->reconnect();
    return $ClientUtil::FALSE if ($self->error);

    # run uname on the client so we know the connection/client are happy
    ($status, @output) = &ClientUtil::execute($self->{conn}, "uname -a");
    if ($status > 0) {
        $self->_set_err("Error: SanBlazeHost::reboot() Command ($cmd) " .
                        "returned a bad status ($status)");
        return "";
    }

    my $tempmsg = "Uname output = ";
    $tempmsg .= join "\n", @output;
    $self->sanlog(-msg => $tempmsg);

    $self->{reboot} = 0;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: ntp_sync
# Objective:       Sync the host with the RTP NTP Time Server
#                  (timehost1.eng.netapp.com )
# Details:         See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    deconfigure
# Objective: NOOP
# Details:   See POD documentation at the end of this file
################################################################################
sub deconfigure {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: syslog_msg
# Objective:       Send message to system log
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg {
    my ($self, @message) = @_;

    return $ClientUtil::TRUE;
}

1;
