# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author hrivera@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::VLinuxHost;

# Package Requirements
######################
use File::Basename;
use San::SanClient;
use San::SanClient::Host::LinuxHost;
use San::SanClient::ClientUtil;

use San::SanLog;
use Error;
use vars qw( @ISA );

@ISA = qw(San::SanClient::Host::LinuxHost Error SanLog SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new VLinuxHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($device_str, @devices, %hRaw);
    my ($self, $conn, $nfs_mount, $cmd, $status, @output, $obj_template);
    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: VLinuxHost::new()" .
                                " Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Error: VLinuxHost::new()" .
                                " Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

        # if an object template was specified just return
    if (defined($obj_template)) {
        return $self;
    }



    unless (defined $conn and length $conn) {
        $self->_set_err("Error: VLinuxHost::new()",
                        "Invalid value for argument '-conn'");
        return $self;
    }

    $self->{conn} = $conn;

    $self->{nfs_mount} = undef;

    # store system information from dmidecode
    $cmd = "dmidecode |  grep -i -A3 \"System Information\"";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (scalar(@output)) {
        $self->{dmidecode_system_info} = \@output;
    } else {
        $self->{dmidecode_system_info} = [];
    }

    $self->store_OS_major_and_minor_revisions();
    if ($self->error) {
        return $self;
    }

    $self->sanlog(-msg => "OS version: " . $self->{major_revision});

    # store TEST_BIN directory
    $self->store_test_bin_dir();
    if ($self->error) {
        return $self;
    }

    return ($self);
}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            if (not defined ($cfmode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->{valid} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: configure
# Objective: NOOP for VLinuxHost
# Details: See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: deconfigure
# Objective: NOOP for VLinuxHost
# Details: See POD documentation at the end of this file
################################################################################
sub deconfigure {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
################################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: discover
# Objective: Discover devices
# Details: See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: format
# Objective: format devices
# Details: See POD documentation at the end of this file
################################################################################
sub format {
    my ($self, %args) = @_;
    my ($devices);

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: VLinuxHost::format() " .
                                "Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Check that required parameter '-devices' was provided
    unless (defined $devices and ref($devices) eq "ARRAY"){
        $self->_set_err("Error: VLinuxHost::format() ",
                        "Missing required argument '-devices'");
        return $ClientUtil::FALSE;
    }

    # pull out the dev manager object
    my $devmgr_obj = $self->{obj}{devmgr};

    # iterate through the devices and partition each one
    foreach my $dev_name (@{$devices}) {

        # convert the device name to the device object
        my $dev_obj = $devmgr_obj->get_device($dev_name);
        if ($devmgr_obj->error()) {
            $self->_set_err($devmgr_obj->errmsg());
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "Partitioning device '$dev_name'");

        $dev_obj->partition();
        if ($dev_obj->error()) {
            $self->_set_err($dev_obj->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: ntp_sync
# Objective:       turn off all ntp syncing for Linux VMs
# Details:         See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my ($ntp_time_zone, $ntp_daemon);

    foreach (keys %args) {
        if (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: VLinuxHost::ntp_sync() Invalid value for argument '-ntp_time_zone'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # sles and redhat have different ntp service daemons.
    if ($self->{major_revision} =~ /sles/i) {
        $ntp_daemon = "ntp";
        if ($self->{major_revision} =~ /sles9/) {
            $ntp_daemon = "xntpd";
        }
    } else {
        $ntp_daemon = "ntpd";
    }

    # check if ntpd is running
    if ($self->{major_revision} =~ /rhel7|fedora/i) {
        $cmd = "systemctl status $ntp_daemon" . ".service";
    } else {
        $cmd = "/etc/init.d/$ntp_daemon status";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # if it is, stop it.
    if (grep {/running/i} @output) {
        if ($self->{major_revision} =~ /rhel7|fedora/i) {
            $cmd = "systemctl stop $ntp_daemon" . ".service";
        } else {
            $cmd = "/etc/init.d/$ntp_daemon stop";
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Error: VLinuxHost::ntp_sync() ".
                            "Could not stop the ntp service.");
            return $ClientUtil::FALSE;
        }
    }

    # set the time zone
    $self->sanlog(-msg => "Setting the Time Zone (TZ) to $ntp_time_zone");
    $cmd = "echo -e \"11\n" . $ntp_time_zone . "\n1\" | tzselect";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: LinuxHost::ntp_sync() ".
                        "Could not set the timezone to '$ntp_time_zone'.");
        return $ClientUtil::FALSE;
    }

    # print out the local time
    $cmd = "date";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->sanlog(-msg => "Could not query for local time.");
    } else {
        $self->sanlog(-msg => "Local time is: $output[0]");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: sanlun_version()
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}


################################################################################
# Subroutine Name: raw_device_cleanup
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################
sub raw_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}



################################################################################
# Subroutine Name: disable_dmmp
# Objective:       NOOP for VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################
sub disable_dmmp {
    my($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: orphaned_device_cleanup
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_supported_protocols
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################
sub get_supported_protocols {
    my ($self) = @_;

    $self->_clear_err();

    return qw (fcp iscsi nfs);
}

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################
sub get_default_mp_type {
    my ($self) = @_;

    # Don't use MP in this virtual context. This could be untrue if this is
    # iSCSI and SW initiator, but this is only the default (and could be
    # overriden)
    return("NONE");

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: get_supported_protocols
# Objective:       NOOP For VLinuxHost
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_unlabeled_devices {
    my ($self) = @_;

    return (wantarray()) ? () : [];
}


1;
