# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author hrivera@netapp.com
## @status private
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::VSolarisHost;

# Package Requirements
######################
use POSIX;
use San::SanClient;
use San::SanClient::Host::SolarisHost;
use San::SanClient::ClientUtil;

use San::SanLog;
use Error;
use vars qw( @ISA );

@ISA = qw(San::SanClient::Host::SolarisHost Error SanLog SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: new()
# Objective:       Create a new VSolarisHost object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($device_str, @devices, $obj_template);
    my ($self, $conn);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: VSolarisHost::new() Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        }
        elsif (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Error: VSolarisHost::new() Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Error: VSolarisHost::new()",
                        "Invalid value for argument '-conn'");
        return $self;
    }

    $self->{conn} = $conn;
    $self->{gdb} = "/usr/local/bin/gdb";

    # execute "uname -p" and "uname -r" to check the OS version
    $self->_determine_os_version();

    # obtain information regarding the memory on this system.
    $self->_obtain_memory_info();

    return $self;
}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            if (not defined ($cfmode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->{valid} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: asl_version()
# Objective:       get asl version
# Details:         See POD documentation at the end of this file
################################################################################
sub asl_version {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

################################################################################
# Subroutine Name: ntp_sync()
# Objective:       turn off all ntp syncing for Solaris VMs
# Details:         See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    # Turn off ntp services for all Solaris Virtual Machines
    if ($self->{revision} =~ /5\.8/ or $self->{revision} =~ /5\.9/) {
        # stop the ntp service daemon - do not check return status
        # not valid - Solaris bug
        $cmd = "/etc/init.d/xntpd stop";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    } else {
        # based on current state of service, stop it
        $cmd = "svcadm disable svc:/network/ntp:default";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    }

    # print date to have in log
    $cmd = "date";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SolarisHost::ntp_sync() Could not execute $cmd",
                        @output);
        return $ClientUtil::FALSE;
    } else {
        $self->sanlog(-msg => "Local time is: $output[0]");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: configure
# Objective: configure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args) = @_;

    $self->_clear_err();
    $self->{protocol} = "none";

    # start the yp service (hack to get x/eng/rtp loaded)
    my $cmd = "/usr/lib/netsvc/yp/ypstart";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: VSolarisHost::configure() failed to start yp");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "SanClient Configuration is completed", -msg_level => 3);
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: deconfigure
# Objective: configure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub deconfigure {
    my ($self, %args) = @_;

    $self->_clear_err();

    # No-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: discover
# Objective: rescan devfsadm
# Details: See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $cmd = "/usr/sbin/devfsadm -C";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: VSolarisHost::configure() failed to start yp");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}


################################################################################
# Subroutine Name: sanlun_version()  [noop]
# Objective:       get sanlun version
# Details:         See POD documentation at the end of this file
################################################################################
sub sanlun_version {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

###############################################################################
# Subroutine Name: discoverDevAluaFCP [NOOP]
# Objective:       Discover LUNs which are mapped to ALUA enabled Solaris10
#                  host. This routine is specific for FCP only and is a place
#                  holder untill Interop can provide official support for
#                  Solaris10 MPxIO.
###############################################################################
sub discoverDevAluaFCP {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
################################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: format
# Objective:       Format devices with a label
# Details:         See POD documentation at the end of this file
################################################################################
sub format {
    my ($self, %args) = @_;

    $self->_clear_err();

    my @devices = ();

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not scalar(@devices = @{$args{$_}})) {
                $self->_set_err("Error: VSolarisHost::format() Invalid value for -devices: $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # validate that devices is not empty
    if (not scalar(@devices)) {
        $self->_set_err("Error: VSolarisHost::format() -devices is required");
        return $ClientUtil::FALSE;
    }

    # pull out the dev manager object
    my $devmgr_obj = $self->{obj}{devmgr};

    # partition each device.
    foreach my $dev_name (@devices) {
        # convert the device name to the device object
        my $dev_obj = $devmgr_obj->get_device($dev_name);
        if ($devmgr_obj->error()) {
            $self->_set_err($devmgr_obj->errmsg());
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "Partitioning device '$dev_name'");

        $dev_obj->partition();
        if ($dev_obj->error()) {
            my $error = join("\n", $dev_obj->errmsg());
            $self->_set_err("Error: VSolarisHost::format() partition failed.\n$error");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################
sub get_default_mp_type {
    my ($self) = @_;

    # Don't use MP in this virtual context. This could be untrue if this is
    # iSCSI and SW initiator, but this is only the default (and could be
    # overriden)
    return("NONE");

}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For VSolarisHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    orphaned_device_cleanup
# Objective: NOOP for solaris virtual machines not during SAN
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

1;
