# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Empty Shell to override NOOPd functions for virtual hosts.
## @author hrivera@netapp.com
## @status public
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::VWinHost;

# Package Requirements
######################
use San::SanClient;
use San::SanClient::ClientUtil;
use San::SanClient::Host::WinHost;
use San::SanLog;
use San::SanCache;
use Error;
use WinClient::ISCSI;
use vars qw( @EXPORT @ISA );

@ISA = qw(San::SanClient::Host::WinHost Error SanLog SanCache SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Method:    _new()
# Objective: Create a new VWinHost object
# Details:   See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $conn, $obj_template);

    $self = {};
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: VWinHost::new()",
                                "Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Error: VWinHost::new()",
                                "Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified just return
    if (defined($obj_template)) {
        return $self;
    }

    unless (defined $conn and length $conn) {
        $self->_set_err("Error: VWinHost::new()",
                        "Invalid value for argument '-conn'");
        return $self;
    }

    $self->{TEST_BIN} = 'X:' . "\\SAN\\bin\\common\\";
    $self->{conn} = $conn;

    return $self;
}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            if (not defined ($cfmode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->{valid} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Method:    configure
# Objective: NOOP configure for virtual machines
# Details:   See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args) = @_;

    $self->_clear_err();

    # burt490057, for win2k8 hosts, set SAN Policy to OnlineAll
    $self->_configure_disk_san_policy();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # reset the WinClient's protocol to nothing
    $self->{WinClient}->{protocol} = "none";

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Method:    deconfigure
# Objective: NOOP deconfigure for virtual machines
# Details:   See POD documentation at the end of this file
################################################################################
sub deconfigure {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
################################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Method:    ntp_sync
# Objective: NOOP for virtual machines
# Details:   See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self) = @_;

    # display the date
    my $cmd = "time /T";
    my ($status, @output) = $self->{WinClient}->execute($cmd);
    if ($status > 0) {
        $self->_set_err("Error: VWinHost::ntp_sync() ".
                        "Command ($cmd) failed.");
        return $ClientUtil::FALSE;
    } else {
        $self->sanlog(-msg => "Local time is: $output[0]");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################
sub get_default_mp_type {
    my ($self) = @_;

    # Don't use MP in this virtual context. This could be untrue if this is
    # iSCSI and SW initiator, but this is only the default (and could be
    # overriden)
    return("NONE");
}

################################################################################
# Method:    _partition
# Objective: We override this function because we have to manually align
#            vmfs disks by 32k in windows.
# Details:   See POD documentation at the end of this file
################################################################################
sub _partition
{
    my ($self, $diskObj) = @_;
    my ($status, @output);

    # check the version of windows
    if($self->{WinClient}->{hostver} !~ /2000/) {
        my $cmd = "systeminfo";
        ($status, @output) = $self->{WinClient}->execute($cmd);
        if ($status > 0) {
            my @tempmsg = $diskObj->errmsg();
            $self->_set_err("Error: VWinHost::_partition() error from " .
                            "WinClient::Disk::partition(): @tempmsg");
            return $ClientUtil::FALSE;
        }
    }

    # if this is a VMFS lun then align it by 32k so it is aligned with WAFL
    # Dont do alignment if this is Windows 2000 or XP
    if ($diskObj->{vendor} =~ /VMware|Msft/i and grep(/OS Name:.*(2003|2008)/, @output)) {
        $diskObj->partition("32");
    } else {
        $diskObj->partition();
    }

    if ($diskObj->error)
    {
        my @tempmsg = $diskObj->errmsg();
        $self->_set_err("Error: VWinHost::_partition() error from " .
                        "WinClient::Disk::partition(): @tempmsg");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    _set_lbpolicy
# Objective: NOOP for windows guest OS's
# Details:   See POD documentation at the end of this file
################################################################################
sub _set_lbpolicy
{
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For VWinHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    orphaned_device_cleanup
# Objective: NOOP for windows virtual machines not during SAN
################################################################################
sub orphaned_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    _WinClient_getDevices()
# Objective: function to invoke WinClient::getDevices()
################################################################################
sub _WinClient_getDevices {
    my ($self, %args) = @_;
    my $refresh = 0;

    foreach (keys %args) {
        if (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: VWinHost::_WinClient_getDevices() Invalid value ".
                                "for argument '-refresh'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Look to see if WinClient's Disk Cache is defined and use if defined
    # If refresh is set, always call, which will refresh the cache.
    # Invoke getDevices so that it skips "mpc show luns" because it
    # may cause virtual machines to hang when running hardware iscsi RDMs.
    if (not defined($self->{WinClient}->{disks}) or $refresh == 1) {
        $self->{WinClient}->getDevices(undef, 1);
    }

    return $ClientUtil::TRUE;
}

1;

