# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to a HOST machine
## @author pressley@netapp.com, hrivera@netapp.com, thomasm@netapp.com
## @status public
## @pod ../Host.pod

use strict;

# Package Name
##############
package San::SanClient::Host::WinHost;

# Package Requirements
######################
use POSIX;
use Config;
use Net::Ping;
use Net::IP;

use San::SanCache;
use San::SanClient;
use San::SanClient::ClientUtil;
use San::SanLog;
use San::SanUtil;
use Error;
use Data::Dumper;

use vars qw(@EXPORT @ISA);

@ISA = qw(Error SanLog SanCache SanClient);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Method:    _new()
# Objective: Create a new WinHost object
# Details:   See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($device_str, @devices);
    my ($self,$conn, $obj_template);

    $self = {};
    bless $self,$class;

    #
    # parse arguments
    #
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: WinHost::new()",
                                "Invalid value for -conn: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?self/) {
            if (not length($self = $args{$_})) {
                $self->_set_err("Error: WinHost::new()",
                                "Invalid value for -self: $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for -obj_template");
                return $self;
            }
            next;
        }
    }

    # if an object template was specified, just return
    if (defined($obj_template)) {
        return $self;
    }

    # set TEST_BIN for location of DT/SIO
    $self->{TEST_BIN} = 'X:' . "\\SAN\\bin\\common\\";

    if (defined($conn)) {
        $self->{conn} = $conn;
    }

    return $self;
}

################################################################################
# Subroutine Name: get_registry_path_entries
# Objective:       Works as a wrapper around WinClient::Utils.pm::getRegistryPathEntries
# Details:         See POD documentation at the end of this file
################################################################################
sub get_registry_path_entries {
    my ($self, %args) = @_;

    $self->{WinClient}->getRegistryPathEntries(%args);
    if ($self->{WinClient}->error) {
        $self->_set_err($self->{WinClient}->errmsg());
        return $ClientUtil::FALSE;
    }

    $self->{reg} = $self->{WinClient}->{reg};
    foreach my $path (keys %{$self->{reg}}) {
        $self->sanlog(-msg => "Path:\t $path", -log_level => 2);
        $self->sanlog(-msg => "KeyName\t\t\tKeyTyep\t\t\tKeyValue", -log_level => 2);
        foreach my $name (keys %{$self->{reg}->{$path}}) {
            my $type = $self->{reg}->{$path}->{$name}->{type};
            my $value = $self->{reg}->{$path}->{$name}->{value};
            $self->sanlog(-msg => "$name\t, $type\t, $value", -log_level => 2);
        }
    }
    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: set_registry_dev_entries
# Objective:       Will set registry key's value/type for each instance of a dev
#                  will set same value of the passed key for all instances of
#                  the device
# Details:         See POD documentation at the end of this file
################################################################################
sub set_registry_dev_entries {
    my ($self, %args) = @_;

    my ($path, $cmd, $status, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?path/) {
            if (not defined ($path = $args{$_})) {
                $self->_set_err("Invalid value for -path => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    # Check that required parameter '-path' was provided
    if (not defined $path and $path ne "") {
        $self->_set_err("Missing required argument '-path'");
        return $ClientUtil::FALSE;
    }

    $cmd = "reg query $path";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 30);
    if ($status > 0 or (grep(/ERROR:/, @output))) {
        $self->_set_err("Failed to get resitry entry for path $path");
        return $ClientUtil::FALSE;
    }

    # Get the registry keys' values for each instance of the device
    foreach my $dev_instance_path (@output) {
        # will populate hash for each instance
        my @path_arr = split (/\\/, $path);
        my @instance_arr = split (/\\/, $dev_instance_path);
        if ($instance_arr[0] eq $path_arr[0] and scalar(@instance_arr) > scalar(@path_arr)) {
            $args{-path} = $dev_instance_path;
            $self->set_registry_entry(%args);
            return $ClientUtil::FALSE if ($self->error);
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_fs_disk_space()
# Objective: returns hash reference containing file system disk space usage
################################################################################
sub get_fs_disk_space {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: set_registry_entry
# Objective:       Will work as wrapper for San::WinClient::writeRegistryEntry
# Details:         See POD documentation at the end of this file
################################################################################
sub set_registry_entry {
    my ($self, %args) = @_;

    $self->{WinClient}->writeRegistryKey(%args);
    if ($self->{WinClient}->error) {
        $self->_set_err($self->{WinClient}->errmsg());
        return $ClientUtil::FALSE;
    }
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_registry_entry
# Objective:       Will work as wrapper for San::WinClient::readRegistryEntry
# Details:         See POD documentation at the end of this file
################################################################################
sub get_registry_entry {
    my ($self, %args) = @_;

    my ($registry);
    $registry = $self->{WinClient}->readRegistryKey(%args);
    if ($self->{WinClient}->error) {
        $self->_set_err($self->{WinClient}->errmsg());
        return "";
    }

    return $registry;
}

################################################################################
# Subroutine Name: validate
# Objective:       validates the state of the configuration and returns status
#                  on it's validity to the rules.
# Details:         See POD documentation at the end of this file
################################################################################
sub validate {
    my ($self, %args) = @_;

    my ($cfmode);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?cfo_mode/) {
            if (not defined ($cfmode = $args{$_})) {
                $self->_set_err("Invalid value for -cfo_mode => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->{valid} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Method:    asl_version()
# Objective: N/A for Windows, provide for compatibility with UNIX hosts
# Details:   See POD documentation at the end of this file
################################################################################
sub asl_version {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Method:    configure()
# Objective: Configure iscsi, fcp, and multipathing.
# Details:   See POD documentation at the end of this file
################################################################################
sub configure {
    my ($self, %args) = @_;

    $self->_clear_err();

    my (@targets, @fcp_targets, @iscsi_targets);

    # parse arguments
    foreach (keys %args) {
        if (/^-?targets$/) {
            if (not (@targets = ${args{$_}})) {
                $self->_set_err("Invalid value for " . "-targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?iscsi_targets$/) {
            if (not (@iscsi_targets = ${args{$_}})) {
                $self->_set_err("Invalid value for " . "-iscsi_targets => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    #
    # Configure iSCSI protocol
    #
    if ($self->get_protocol() =~ /iscsi/i) {
        # configure for protocol iscsi
        if (scalar(@iscsi_targets) == 0) {
            $self->_iscsi_configure(-targets => @targets);
        } else {
            $self->_iscsi_configure(-targets => @iscsi_targets);
        }
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    #
    # FCP Config is not supported yet or N/A
    #

    #
    # Multipathing Config is not supported yet or N/A
    #

    # burt490057, for win2k8 hosts, set SAN Policy to OnlineAll
    $self->_configure_disk_san_policy();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: deconfigure
# Objective: deconfigure host machine
# Details: See POD documentation at the end of this file
################################################################################
sub deconfigure {
    my ($self, %args) = @_;

    my (@targets, $clean_level);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            @targets = @{$args{$_}};
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # call appropriate deconfigure methods
    if ($self->{protocol} =~ /fcp/i) {
        $self->_fcp_deconfigure();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }
    if ($self->{protocol} =~ /iscsi/i) {
        # Build the args
        my %cmd_args;

        # Add in targets if defined.
        if (scalar(@targets) != 0) {
            $cmd_args{"-targets"} = \@targets;
        }

        # Add in clean_level if defined.
        if (defined($clean_level)) {
            $cmd_args{"-clean_level"} = $clean_level;
        } else {
            # default it to 1
            $cmd_args{"-clean_level"} = 1;
        }

        # return if clean level is 0
        if ($cmd_args{"-clean_level"} == 0) {
            $self->sanlog(-msg => "iSCSI Deconfigure: Not performing iSCSI deconfiguration because clean_level is '0'");
            return $ClientUtil::TRUE;
        }

        $self->_iscsi_deconfigure(%cmd_args);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_root_dev()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_root_dev {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: update_partition_table()
# Objective:       NOOP
# Details:         See POD documentation at the end of this file
################################################################################
sub update_partition_table {

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name  gather_data
# Objective:       Gather the data on the client. This data will mostly be
#                  used for debug purposes. Currently no-op
# Details:         See POD documentation at the end of this file
################################################################################
sub gather_data {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_device_size
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_device_size {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_scsi_devices
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_scsi_devices {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: collect_sysrq_dump_periodic
# Objective:       N/A for Windows, provide for compatibility with Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub collect_sysrq_dump_periodic {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_blocks_limits_vpd
# Objective:       Verify BLOCK LIMITS VPD (0xb0) details are reflected on
#                  hosts for NetApp LUNs
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_blocks_limits_vpd {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: reclaim_space_on_mount_points
# Objective:       Reclaim space on all mount points
# Details:         See POD documentation at the end of this file
################################################################################
sub reclaim_space_on_mount_points {
    my ($self) = @_;

    # NO-OP
    return $ClientUtil::TRUE;
}

################################################################################
# Method:    config_paths
# Objective: N/A for Windows, provide for compatibility with HP-UX
# Details:   See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Method: discover
# Objective: Discover devices and label if new.
# Details: See POD documentation at the end of this file
################################################################################
sub discover {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->sanlog(-msg => "Discovering all devices");

    #
    # Rescan on the client to find any new devices.
    # Rescan will be called in WinClient->get_devices
    # This will also populate the WinClient->{disks} hash
    # of device entries.
    #
    my @device_array = $self->get_devices(%args,
                                          -refresh => 1);
    return $ClientUtil::FALSE if ($self->error);

    # call raw_device_setup where on windows 2008
    # this will online the disks
    $self->raw_device_setup();
    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;

}

################################################################################
# Method:    format
# Objective: Format devices with a label
# Details:   See POD documentation at the end of this file
################################################################################
sub format {
    my ($self, %args) = @_;

    my $devices;
    my $label;
    my $use_label      = 0;
    my $change_occured = 0;
    my $refresh        = 0;
    my $type           = "";
    my $fs_rotate      = 0;
    my @affected_devs;

    #
    # parse arguments
    #
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: WinHost::format() Invalid value for argument '-devices'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?label/) {
            if (not length($label = $args{$_})) {
                $self->_set_err("Error: WinHost::format() Missing value for argument '-label'");
                return ($ClientUtil::FALSE);
            }
            $use_label = 1;
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: WinHost::format() Missing value for argument '-refresh'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $type ="ntfs";
            }
            next;
        } elsif (/-?fs_rotate/) {
            if (not length($fs_rotate = $args{$_})) {
                $self->_set_err("Error: WinHost::format() Missing value for argument '-fs_rotate'");
                return ($ClientUtil::FALSE);
            }
        }
    }

    #
    # Check that required parameter '-devices' was provided
    #
    unless (defined $devices and ref($devices) eq "ARRAY") {
        $self->_set_err("Error: WinHost::format() ",
                        "Missing required argument '-device'");
        return ($ClientUtil::FALSE);
    }
    #
    # Go get the device info from the client
    #
    my @device_array = $self->get_devices(-refresh => $refresh);

    return $ClientUtil::FALSE if ($self->error);

    #
    # Determine which devs to format
    # This is a 2 pass algorithm.  First we get the devices, and run
    # Disk::destroy. Then we run WinClient::UpdateDriveLetters
    # then we build up the device (partition, assign, format).
    #
    foreach my $device_hash (@device_array) {
        foreach my $key (keys %$device_hash) {
            $self->sanlog(-msg=> "format() device hash: " .
                                 "key is $key value is "  .
                                 $device_hash->{$key} . "\n",
                          -msg_level => 7);
        }

        #
        # Skip non ontap devices.
        #
        next if ($device_hash->{type} !~ /NETAPP|IBM-Nseries/);

        #
        # The device name field holds an integer that is the device ID.
        # We are expecting the devices array to be a list of device
        # ID's.  If this device is in that list, format this device.
        #
        foreach my $device (@$devices) {
            #
            # take out the PHYSICALDRIVE bit, if it's there
            #
            if ($device =~ /PHYSICALDRIVE(\d+)/) {
                $device = $1;
            }

            if ($device == $device_hash->{name}) {
                $self->sanlog(-msg => "WinHost::format() changing device " .
                                      $device_hash->{path});

                #
                # First destroy (oh and push the diskObj to be
                # used later.
                #
                push (@affected_devs, $device_hash);

                my $diskObj = $device_hash->{diskObj};
                $diskObj->destroy();
                if ($diskObj->error) {
                    my @tempmsg = $diskObj->errmsg();
                    $self->_set_err("Error: WinHost::format() error from " .
                                    "WinClient::Disk::destroy(): @tempmsg");
                    return ($ClientUtil::FALSE);
                }
            }
        }
    }

    #
    # Update the assigned drive letters
    #
    $self->{WinClient}->updateLetters();
    if ($self->{WinClient}->error) {
        my @tempmsg = $self->{WinClient}->errmsg();
        $self->_set_err("Error: WinHost::format() error from " .
                        "WinClient::updateLetters(): @tempmsg");
        return $ClientUtil::FALSE;
    }

    #
    # For each affected device, partition, assign, format
    #
    foreach my $device_hash (@affected_devs) {
        my $diskObj = $device_hash->{diskObj};

        #
        # Partition the device
        #
        $self->sanlog(-msg => "WinHost::format() partition device " .
                              $device_hash->{path});
        $self->_partition($diskObj);
        if ($self->error()) {
            return ($ClientUtil::FALSE);
        }

        #
        # Assign a drive letter
        #
        $self->sanlog(-msg => "WinHost::format() assign device " .
                              $device_hash->{path});
        $diskObj->assign(-type => "letter");
        if ($diskObj->error) {
            my @tempmsg = $diskObj->errmsg();
            $self->_set_err("Error: WinHost::format() error from " .
                            "WinClient::Disk::assign(): @tempmsg");
            return ($ClientUtil::FALSE);
        }

        #
        # Format the drive
        #
        $self->sanlog(-msg => "WinHost::format() format device " .
                              $device_hash->{path});
        if ($fs_rotate) {
            if ($type =~ /ntfs/i) {
                $type = "refs";
            } else {
                $type = "ntfs";
            }
        }
        if ($type ne "") {
            $diskObj->format(-type => $type);
        } else {
            $diskObj->format();
        }

        if ($diskObj->error) {
            my @tempmsg = $diskObj->errmsg();
            $self->_set_err("Error: WinHost::format() error from " .
                            "WinClient::Disk::format(): @tempmsg");
            return ($ClientUtil::FALSE);
        }

        # if a load balancing policy was set, then set it on the device
        # if it is an iscsi lun.
        #
        $self->_set_lbpolicy($diskObj);
        if ($self->error()) {
            return ($ClientUtil::FALSE);
        }

    }

    return $ClientUtil::TRUE;

}

###############################################################################
# Method:    get_mounted_device
# Objective: returns a mounted device name when passed a filesystem
###############################################################################
sub get_mounted_device {
    my ($self,$mount) = @_;

    $self->_clear_err();
    $self->_set_err("Error: WinHost::get_mounted_device() Not",
                    "currently supported by WinHost object");

    return undef;

}

###############################################################################
# Method:    get_lvm_default
# Objective: Return the default lvm for this OS
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_lvm_default {
    my ($self) = @_;

    $self->_clear_err();

    return qw(vxvm);

}

###############################################################################
# Method:    get_supported_protocols
# Objective: Get list of supported protocols for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_protocols {
    my ($self) = @_;

    $self->_clear_err();

    return qw(fcp iscsi cifs nfs);

}

###############################################################################
# Method:    get_supported_fs_types
# Objective: Get list of supported file systems for this platform
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_supported_fs_types {
    my ($self) = @_;

    return qw(NTFS FAT32);
}

###############################################################################
# Method:    info
# Objective: Get client information
# Details:   See POD documentation at the end of this file.
###############################################################################
sub info {
    my ($self) = @_;

    my @output;

    push(@output, "Host name:      " . $self->hostname(),
                  "Host type:      " . $self->hosttype(),
                  "OS version:     " . $self->os_version(),
                  "Sanlun version: " . $self->sanlun_version(),
                  "Protocol:       " . $self->get_protocol(),
                  "\n");

    return (wantarray()) ? @output: \@output;

}

################################################################################
# Method:    iostat
# Objective: Get information on I/O per device
# Details:   See POD documentation at the end of this file
################################################################################
sub iostat {
    my ($self, %args) = @_;

    my ($interval, $count);

    $self->_clear_err();

    ### Default interval is 2 seconds
    $interval = 2;

    ### Default count is 1
    $count = 1;

    # parse arguments
    foreach (keys %args) {
        if (/-?interval/) {
            if (not length($interval = $args{$_})) {
                $self->_set_err("Error: WinHost::iostat() Invalid value for argument '-interval'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?count/) {
            if (not length($count = $args{$_})) {
                $self->_set_err("Error: WinHost::iostat() Missing value for argument '-count'");
                return ($ClientUtil::FALSE);
            }
            next;
        } else {
            $self->_set_err("Error: WinHost::iostat() Invalid parameter specified \'$_\'");
            return ($ClientUtil::FALSE);
        }
    }

    $self->_set_err("Error: WinHost::iostat() Not",
                     "currently supported by WinHost object");

    return ($ClientUtil::FALSE);

}

################################################################################
# Subroutine Name: get_nfs_mounts
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_mounts {
    my ($self, %args) = @_;

    $self->_set_err("Not supported by WinHost");
    return (wantarray()) ? () : [];
}

################################################################################
# Subroutine Name: get_nfs_share
# Objective:       NO-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nfs_share {
    my ($self, %args) = @_;

    $self->_set_err("Not supported by WinHost");
    return (wantarray()) ? () : [];
}

################################################################################
# Method:    mount_entry_add()
# Objective: N/A for Windows, provide for compatibility with HP-UX
################################################################################
sub mount_entry_add {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Method:     unmount_fs
# Objective:  N/A for Windows, provide for compatibility with UNIX
# Details:    See POD documentation at the end of this file
################################################################################
sub unmount_fs {
    my ($self, $mntpt) = @_;

    $self->_clear_err();

    return $self;

}

################################################################################
# Method:    mount_entry_delete()
# Objective: N/A for Windows, provide for compatibility with UNIX
# Details:   See POD documentation at the end of this file
################################################################################
sub mount_entry_delete {
    my ($self, $mount) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

###############################################################################
# Method:    mount_entry_verify
# Objective: N/A for Windows, provide for compatibility with UNIX
###############################################################################
sub mount_entry_verify {
    my ($self, $mount) = @_;
    my @buf;

    $self->_clear_err();

    return (wantarray()) ? @buf : \@buf;
}

###############################################################################
# Method:    mount_table_verify
# Objective: N/A for Windows, provide for compatibility with UNIX
###############################################################################
sub mount_table_verify {
    my ($self, $mount) = @_;
    my @output;

    $self->_clear_err();

    return (wantarray()) ? @output : \@output;
}

################################################################################
# Method:    nfs_mount
# Objective: NFS mount volume. N/A for Windows, provide for UNIX compatibility
################################################################################
sub nfs_mount {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Method: nfs_unmount()
# Objective: umount NFS volume. N/A for Windows, provide for UNIX compatibility
################################################################################
sub nfs_unmount {
    my ($self, %args) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

###############################################################################
# Subroutine Name: mount_exists
# Objective:       validates a mount point is correctly mounted and does exist
###############################################################################
sub mount_exists {
    my ($self, $mount) = @_;

    my ($cmd, $status, @output);

    # See if drive path has : on it, if not add it
    if ($mount !~ /\:/) {
        $mount .= ":";
    }

    $cmd = "dir $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 30);

    # Check the status of the mount command
    if (grep(/cannot find/, @output) or scalar(@output) == 0) {
        $self->_set_err("Error: WinHost::mount_exists(): Unable to execute command ($cmd)");
        return $ClientUtil::FALSE;
    } else {
        # dir command did work, means that drive exists
        return $ClientUtil::TRUE;
    }
}

################################################################################
# Subroutine Name: shutdown()
# Objective:       shutdown host
# Details:         See POD documentation at the end of this file
################################################################################
sub shutdown {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "shutdown /s /m \\\\127.0.0.1";
    @output = &Tharn::say($self->{conn}, $cmd, '-blind');

    # sleep 1 minute to make sure shutdown occurs
    &SanUtil::nate_sleep(60);

    # Invalidate the connection on the object
    &Tharn::disconnect($self->{conn}, '-ok');

    # Ping the host every 10 seconds until it has completely shutdown
    my $alive = 1;

    for (my $cnt = 0; $cnt < 36; $cnt++) {
        my $status = &SanUtil::ping_host($self->hostname());
        if ($status == $ClientUtil::FALSE) {
            $self->sanlog(-msg => "OS is shutdown");
            $alive = 0;
            last;
        } else {
            $self->sanlog(-msg => "Ping was successful",
                          -msg_level => 7);
            &SanUtil::nate_sleep(10);
        }
    }

    # Machine is still alive, shutdown did not work
    if ($alive) {
        $self->_set_err("OS never shutdown");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    reboot()
# Objective: reboot host
################################################################################
sub reboot {
    my ($self) = @_;

    $self->_clear_err();

    #
    # Call the WinClient reboot function.  This will reboot the host and wait
    # for it to come back up.
    #
    $self->{WinClient}->reboot();
    if ($self->{WinClient}->error) {
        $self->_set_err("Machine never rebooted");
        return $ClientUtil::FALSE;
    } else {
        $self->sanlog(-msg => "Machine is alive after reboot");
    }

    #
    # Sleep for a while to make sure the host is all the
    # way back up.
    #
    &SanUtil::nate_sleep(120);

    #
    # reconnect to the client
    #
    $self->reconnect();
    if ($self->error) {
        my @tempmsg = $self->errmsg();
        $self->_set_err("WinHost::reboot() reconnect failed: @tempmsg");
        return ($ClientUtil::FALSE);
    }

    $self->{reboot} = 0;

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    sanlun_version()
# Objective: get sanlun version (sanlun does not exist on windows)
################################################################################
sub sanlun_version {
    my ($self) = @_;

    $self->_clear_err();

    return "0.0";
}

################################################################################
# Method:    raw_device_setup
# Objective: For windows 2008, bring the disks online
#            after the call to discover().
################################################################################
sub raw_device_setup {
    my ($self) = @_;

    my (@diskObjs);

    $self->_clear_err();

    # only perform if win2008/win7, other platforms
    # disks will already be online
    if ($self->{WinClient}->{hostver} !~ /2008/ and
        $self->{WinClient}->{hostver} !~ /windows server 7/) {
        return $ClientUtil::TRUE;
    }

    @diskObjs = $self->get_devices(-refresh => 0);
    return $ClientUtil::FALSE if ($self->error);

    # interate through disk hashes and call online
    # on WinClient disk object
    foreach my $device_hash (@diskObjs) {
        my $diskObj = $device_hash->{diskObj};
        $diskObj->online();
        if ($diskObj->error()) {
            my @error = $diskObj->errmsg();
            $self->_set_err("Could properly online disk\n" .
                            join('\n', @error));
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    raw_device_cleanup
# Objective: Do whatever we need to do to make sure the host is more
#            happy when luns are removed.
################################################################################
sub raw_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    #
    # Update devices
    #
    $self->_WinClient_getDevices(-refresh => 1);
    if ($self->{WinClient}->error) {
        my @tmp_msg = $self->{WinClient}->errmsg();
        $self->_set_err("SanClient::raw_device_cleanup(). WinClient failed in" .
                        "getDevices: @tmp_msg");
        return $ClientUtil::FALSE;
    }

    #
    # When we called getDevices, it filled the WinClient disks element with a
    # hash of disk objects keyed by disk number
    #
    foreach my $key (keys %{$self->{WinClient}->{disks}}) {
        my $diskObj = $self->{WinClient}->{disks}{$key};

        #
        # Don't unconfigure the boot disk
        #
        next if ($diskObj->{boot});

        #
        # Don't unconfigure the quorum disk
        #
        next if ($diskObj->{isQuorum});

        #
        # Don't use LUNs not owned by node
        #
        next if ($diskObj->{lunState} =~ /Clus/i);

        #
        # Update devices
        #
        $diskObj->remove();
        if ($diskObj->error) {
            my @errmsg = $diskObj->errmsg();
            $self->_set_err("SanClient::raw_device_cleanup(). " .
                            "WinClient::Disk failed to remove: ".
                            "@errmsg");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: multipath_device_cleanup
# Objective:       NOOP For WindowsHost
# Details:         See POD documentation at the end of this file
################################################################################
sub multipath_device_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    orphaned_device_cleanup
# Objective: When the luns are removed from the filer, we will have
#            orphaned luns. This method will rescan the storage
#            only when mp_type is MPIO and volume manager is VxVM. For
#            any other case its no-op
################################################################################
sub orphaned_device_cleanup {
    my ($self, %args) = @_;

    my ($mp_type, $lvm_type, $cmd, $status, @output, $client_vm);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?client_vm/) {
            if (not length($client_vm = $args{$_})) {
                $self->_set_err("Invalid value for argument '-client_vm'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }

    ($mp_type, $lvm_type) = ($client_vm =~ /^(\S+)\-(.+$)/);

    # If we are not veritas DMP, go to the next client
    if ($mp_type =~ /MPIO/i and $lvm_type =~ /vxvm/i) {
        $cmd = "vxassist rescan";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: failed to rescan Veritas disks.");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: ntp_sync
# Objective:       Sync the host with the RTP NTP Time Server
#                  (timehost1.eng.netapp.com )
# Details:         See POD documentation at the end of this file
################################################################################
sub ntp_sync {
    my ($self, %args) = @_;

    my ($cmd, $status, @output);
    my ($ntp_server, $ntp_time_zone);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?ntp_server/) {
            if (not length($ntp_server = $args{$_})) {
                $self->_set_err("Error: WinHost::ntp_sync() Invalid value ".
                                "for argument '-ntp_sync'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?ntp_time_zone/) {
            if (not length($ntp_time_zone = $args{$_})) {
                $self->_set_err("Error: WinHost::ntp_sync() Invalid value ".
                                "for argument '-ntp_time_zone'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # set the timezone
    # there is no known solution on setting the time zone remotely.

    # sync with the time server
    $self->sanlog(-msg => "Syncing the time with local NTP server: $ntp_server");

    $cmd = "net time /setsntp:$ntp_server";
    ($status, @output) = $self->{WinClient}->execute($cmd);
    if ($status > 0) {
        $self->_set_err("Error: WinHost::ntp_sync() ".
                        "Command ($cmd) failed.");
        return $ClientUtil::FALSE;
    }

    # if the net time command output says to use w32tm.exe, then use that command
    if (grep(/use w32tm/i, @output)) {

        # un register and re register the w32tm service, not checking for error intentionally. because these command may fail
        $cmd = "net stop w32time";
        ($status, @output) = $self->{WinClient}->execute($cmd);
        $cmd = "w32tm /unregister";
        ($status, @output) = $self->{WinClient}->execute($cmd);
        $cmd = "w32tm /register";
        ($status, @output) = $self->{WinClient}->execute($cmd);
        $cmd = "net start w32time";
        ($status, @output) = $self->{WinClient}->execute($cmd);

        # update the ntp sever
        $cmd = "w32tm /config /update /manualpeerlist:$ntp_server /syncfromflags:manual";
        ($status, @output) = $self->{WinClient}->execute($cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) failed.");
            return $ClientUtil::FALSE;
        }

        # issue a resync to update that time from the ntp server
        $cmd = "w32tm /resync";
        ($status, @output) = $self->{WinClient}->execute($cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) failed.");
            return $ClientUtil::FALSE;
        }

        # display the updated ntp source configuration
        $cmd = "w32tm /query /source";
        ($status, @output) = $self->{WinClient}->execute($cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) failed.");
            return $ClientUtil::FALSE;
        }
    }

    # display the date
    $cmd = "time /T";
    ($status, @output) = $self->{WinClient}->execute($cmd);
    if ($status gt 0) {
        $self->_set_err("Error: WinHost::ntp_sync() ".
                        "Command ($cmd) failed.");
        return $ClientUtil::FALSE;
    } else {
        $self->sanlog(-msg => "Local time is: $output[0]");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display_time()
# Objective:       display the current system time
# Details:         See POD documentation at the end of this file
################################################################################
sub display_time {
    my ($self) = @_;

    my $cmd = "time /T";
    my ($status, @output) = $self->{WinClient}->execute($cmd);
    if ($status == 0) {
        $self->sanlog(-msg => "Local time is: $output[0]");
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display_additional_info()
# Objective:       dump additional OS specific information
################################################################################
sub display_additional_info {
    my ($self) = @_;

    my $cmd = "$self->{WinClient}->{cmd_mpc} show dsms";
    my $connection = $self->{conn};
    my ($status, @output) = &ClientUtil::execute($connection, $cmd, 30);

    my @cmds = ("$self->{WinClient}->{cmd_mpc} show version",
                "$self->{WinClient}->{cmd_mpc} mpio show",
                "$self->{WinClient}->{cmd_mpc} mpio show summary",
                "$self->{WinClient}->{cmd_mpc} show disks",
                "$self->{WinClient}->{cmd_mpc} show luns",
                "$self->{WinClient}->{cmd_mpc} iscsi show connections",
                "$self->{WinClient}->{cmd_mpc} iscsi show sessions",
                "$self->{WinClient}->{cmd_mpc} iscsi show targets",
                "$self->{WinClient}->{cmd_mpcdata}");

    # Run the following commands only if ontapdsm is installed. These are not valid
    # on MSDSM
    if (grep (/ontapdsm/i, @output)) {
        push (@cmds, "$self->{WinClient}->{cmd_mpc} ontapdsm show");
    }

    # Run all the commands we saved up.
    foreach my $cmd (@cmds) {
        &ClientUtil::execute($connection, $cmd, 180);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: load_nsanity()
# Objective:       load the nSANity binary
################################################################################
sub load_nsanity {
    my ($self) = @_;

    # invoke the load_nsanity function on the winclient object
    $self->{WinClient}->load_nsanity();
    if ($self->{WinClient}->error()) {
        $self->_set_err("Failed to load nsanity\nReason: " . $self->{WinClient}->errmsg());
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: run_nsanity()
# Objective:       run nSANity on this host
################################################################################
sub run_nsanity {
    my ($self, %args) = @_;

    my ($credentials, $filename_tag, $destination_dir);

    my $timeout = 7200;
    my $survey_mode = 0;
    my $use_internal_commands = 0;

    # parse params
    foreach (keys %args) {
        if (/-?credentials/) {
            if (not length($credentials = $args{$_})) {
                $self->_set_err("Invalid value for -credentials");
                return "";
            }
        } elsif (/-?filename_tag/) {
            if (not length($filename_tag = $args{$_})) {
                $self->_set_err("Invalid value for -filename_tag");
                return "";
            }
        } elsif (/-?destination_dir/) {
            if (not length($destination_dir = $args{$_})) {
                $self->_set_err("Invalid value for -destination_dir");
                return "";
            }
        } elsif (/-?timeout/) {
            if (not length($timeout = $args{$_})) {
                $self->_set_err("Invalid value for -timeout");
                return "";
            }
        } elsif (/-?survey_mode/) {
            if (not length($survey_mode = $args{$_})) {
                $self->_set_err("Invalid value for -survey_mode");
                return "";
            }
        } elsif (/-?use_internal_commands/) {
            if (not length($use_internal_commands = $args{$_})) {
                $self->_set_err("Invalid value for -use_internal_commands");
                return "";
            }
        }
    }

    # check for required params
    if (not defined($credentials) or
        not defined($filename_tag) or
        not defined($destination_dir)) {
        $self->_set_err("Missing required params -credentials, -filename_tag, or -destination_dir");
        return "";
    }

    # see if we need to enable survey mode
    my $survey_mode_flag = "";
    if ($survey_mode == 1) {
        $survey_mode_flag = " -S survey";
    }

    # create the destination dir just in case
    $self->{WinClient}->execute("mkdir ${destination_dir}");

    # run nsanity to collect the data (use a 2 hour timeout, logfiles may be very large)
    my $nsanity_cmd = "${credentials}${survey_mode_flag} -d ${destination_dir} -f ${filename_tag}";
    my ($status, @output) = $self->{WinClient}->nsanity($nsanity_cmd,
                                                        $use_internal_commands,
                                                        $timeout);
    if ($self->{WinClient}->error()) {
        $self->_set_err("nSANity command ($nsanity_cmd) failed.\nReason: " .
                        $self->{WinClient}->errmsg());
        return "";
    }

    # parse out the filename of the file that was generated by nsanity
    my $nsanity_file;
    foreach my $line (@output) {
        if ($line =~ /Collected data saved to file,\s+(\S+)/i) {
            $nsanity_file = $1;
            last;
        }
    }

    # make sure its not null
    if ($nsanity_file =~ /^\s*$/) {
        $self->_set_err("Could not parse nSANity filename from the output (@output)");
        return "";
    }

    return $nsanity_file;
}

################################################################################
# Subroutine Name: cifs_mount()
# Objective:       Mount a cifs drive at the specified location
################################################################################
sub cifs_mount {
    my ($self, %args) = @_;

    my ($cifs_server, $cifs_dir, $username, $password, $reuse_existing_mount);
    my ($cmd, $status, @output, $cifs_drive_letter, $persistent, $domain);

    $cifs_drive_letter = "";
    $persistent = "NO";
    $reuse_existing_mount = 0;

    $self->_clear_err();

    # parse params
    foreach (keys %args) {
        if (/-?cifs_server/) {
            if (not length($cifs_server = $args{$_})) {
                $self->_set_err("Invalid value for -cifs_server");
                return "";
            }
            next;
        } elsif (/-?cifs_dir/) {
            if (not length($cifs_dir = $args{$_})) {
                $self->_set_err("Invalid value for -cifs_dir");
                return "";
            }
            next;
        } elsif (/-?domain/) {
            if (not length($domain = $args{$_})) {
                $self->_set_err("Invalid value for -domain");
                return "";
            }
            next;
        } elsif (/-?username/) {
            if (not length($username = $args{$_})) {
                $self->_set_err("Invalid value for -username");
                return "";
            }
            next;
        } elsif (/-?password/) {
            if (not length($password = $args{$_})) {
                $self->_set_err("Invalid value for -password");
                return "";
            }
            next;
        } elsif (/-?persistent/) {
            if (not length($persistent = $args{$_})) {
                $self->_set_err("Invalid value for -persistent");
                return "";
            }
            next;
        } elsif (/-?reuse_existing_mount/) {
            if (not length($reuse_existing_mount = $args{$_})) {
                $self->_set_err("Invalid value for -reuse_existing_mount");
                return "";
            }
            next;
        }
    }

    # check for required params
    if (not defined($cifs_server) or
        not defined($cifs_dir) or
        not defined($username)) {
        $self->_set_err("Missing required parameters -cifs_server or -cifs_dir or -username");
        return "";
    }

    # check if we already have a mount point
    $cmd = "net use";
    ($status, @output) = &ClientUtil::execute($self->{conn}, "$cmd");
    if ($status != 0) {
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return "";
    }

    # if asked to reuse an existing mount then check here
    if ($reuse_existing_mount == 1) {
        foreach my $line (@output) {
            $line =~ s/\\/_/g;
            my $dir = $cifs_dir;
            $dir =~ s/\\/_/g;
            if ($line =~ /^\s*OK\s+([A-Z]:)\s+${dir}\s*$/) {
                # store the drive letter thats being used and return
                $cifs_drive_letter = $1;
                $self->sanlog(-msg => "Reusing CIFS drive ($cifs_drive_letter) for mount point ($cifs_dir)");
                return $cifs_drive_letter;
            }
        }
    }

    # default the domain to the cifs server
    if (!$domain) {
        $domain = $cifs_server;
    }

    # otherwise we have to mount a new cifs share
    # mount the cifs share path in its place
    $cmd = "net use * $cifs_dir /u:$domain\\$username $password /persistent:$persistent";

    ($status, @output) = &ClientUtil::execute($self->{conn}, "$cmd");
    if ($status != 0) {
        $self->_set_err("Command ($cmd) failed with status ($status)");
        return "";
    }

    # parse out the drive letter that was used
    foreach my $line (@output) {
        if ($line =~ /Drive (\S+:) is now connected/i) {
            $cifs_drive_letter = $1;
            last;
        }
    }

    # if the cifs drive letter is empty, set error
    if ($cifs_drive_letter eq "") {
        $self->_set_err("Failed to mount a cifs drive letter at location $cifs_dir");
        return "";
    }

    return $cifs_drive_letter;
}

################################################################################
# Subroutine Name: cifs_unmount()
# Objective:       Unmount a cifs drive at the specified drive letter
################################################################################
sub cifs_unmount {
    my ($self, %args) = @_;

    my ($cmd, $status, @output, $cifs_drive_letter, $cifs_dir, $recursive);
    $recursive = 0;

    $self->_clear_err();

    # parse params
    foreach (keys %args) {
        if (/-?cifs_drive_letter/) {
            if (not length($cifs_drive_letter = $args{$_})) {
                $self->_set_err("Invalid value for -cifs_drive_letter");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?cifs_dir/) {
            if (not length($cifs_dir = $args{$_})) {
                $self->_set_err("Invalid value for -cifs_dir");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?recursive/) {
            if (not length($recursive = $args{$_})) {
                $self->_set_err("Invalid value for -recursive");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not defined($cifs_drive_letter) and not defined($cifs_dir)) {
        $self->_set_err("Missing required parameters -cifs_drive_letter or -cifs_dir");
        return $ClientUtil::FALSE;
    }

    # unmount the drive letter
    if (defined($cifs_drive_letter)) {
        $cmd = "net use $cifs_drive_letter /d";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }

        # check if the umount failed
        if (grep(/The network connection could not be found/i, @output)) {
            $self->_set_err("Failed to unmount cifs drive $cifs_drive_letter");
            return $ClientUtil::FALSE;
        }
    } elsif (defined($cifs_dir)) {
        # find all mount points that match $cifs_dir and unmount them
        $cmd = "net use";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Command ($cmd) failed with status ($status)");
            return $ClientUtil::FALSE;
        }

        foreach my $line (@output) {
            # convert all backslashes to forward slashes
            # so that the regex doesnt get messed up
            $line =~ s/\\/\//g;
            $cifs_dir =~ s/\\/\//g;

            if ($recursive == 1) {
                # then unmount any drive that matches cifs_dir
                if ($line =~ /^\s*(Disconnected|OK)?\s+([A-Za-z]:)\s+${cifs_dir}/i) {
                    $cmd = "net use $2 /delete";
                    ($status) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status != 0) {
                        $self->_set_err("Command ($cmd) failed with status ($status)");
                        return $ClientUtil::FALSE;
                    }
                }
            } else {
                # then unmount only the drive that is directly mounted
                # on cifs_dir, anchor the regex
                if ($line =~ /^\s*(Disconnected|OK)?\s+([A-Za-z]:)\s+${cifs_dir}\s*$/i) {
                    $cmd = "net use $2 /delete";
                    ($status) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status != 0) {
                        $self->_set_err("Command ($cmd) failed with status ($status)");
                        return $ClientUtil::FALSE;
                    }
                }
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: syslog_msg
# Objective:       Send message to Windows event viewer
# Details:         See POD documentation at the end of this file
################################################################################
sub syslog_msg {
    my ($self, @message) = @_;

    my ($cmd, @output, $status);

    # eventcreate.exe does not exists on Windows 2000. Not going back
    # to add equivalent.
    if ($self->{WinClient}->{hostver} =~ /2000/) {
        return $ClientUtil::TRUE;
    }

    foreach my $msg (@message) {
        $cmd = "eventcreate /t information /l system /id 999 /d \"**NATE**: $msg\"";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd, 10);
        if ($status > 0) {
            $self->_set_err("Error: WinHost::syslog_msg(): Error running evencreate command ($cmd)");
            return $ClientUtil::FALSE;
        }
    }
}

################################################################################
# Method:    get_devices
# Objective: Get list of all devices on host. Either extract from cache
#            (from WinClient) or get from explicit call (from WinClient) to
#            refresh the list.
# Details:   TBD - private methods aren't in POD, add details
################################################################################
sub get_devices {
    my ($self, %args) = @_;

    my @devices; # array of hash
    my $refresh = 0;
    my $r_timeout = -1;
    my $num_expected_luns = undef;

    foreach (keys %args) {
        if (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Invalid value for argument '-refresh'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?rescan_timeout/) {
            if (not length($r_timeout = $args{$_})) {
                $self->_set_err("Invalid value for argument '-rescan_timeout'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?num_expected_luns/) {
            $num_expected_luns = $args{$_};
        }
    }

    #
    # Get a list of devices
    #
    $self->_WinClient_getDevices(-refresh => $refresh,
                                 -rescan_timeout => $r_timeout,
                                 -num_expected_luns => $num_expected_luns);
    if ($self->{WinClient}->error) {
        $self->_set_err("WinClient failed in getDevices",
                        $self->{WinClient}->errmsg());
        return $ClientUtil::FALSE;
    }

    #
    # When we called getDevices, it filled the WinClient disks element with
    # a hash of disk objects keyed by disk number
    #
    foreach my $key (keys %{$self->{WinClient}->{disks}}) {
        my $diskObj = $self->{WinClient}->{disks}{$key};

        #
        # Don't unconfigure the boot disk
        #
        next if ($diskObj->{boot});

        #
        # Don't unconfigure the quorum disk
        #
        next if ($diskObj->{isQuorum});

        #
        # Don't unconfigure LUNs not owned by node
        #
        next if ($diskObj->{lunState} =~ /Clus/i);

        #
        # Create a hash for this device, fill it with wanted data,
        # push it on the devices stack
        #
        my $device_hash={};
        $device_hash->{path}        = $diskObj->{rawName};
        $device_hash->{name}        = $diskObj->{device};
        $device_hash->{type}        = "NETAPP"; # TBD
        $device_hash->{label}       = $diskObj->{fsLabel};
        $device_hash->{id}          = $diskObj->{device};
        $device_hash->{fstype}      = $diskObj->{fstype};
        $device_hash->{filer_name}  = $diskObj->{filer};
        $device_hash->{diskObj}     = $diskObj;
        push(@devices, $device_hash);

        my $tempmsg = "get_devices found device:\n"              .
                      "filer  = " . $device_hash->{filer_name}   . "\n" .
                      "path   = " . $device_hash->{path}         . "\n" .
                      "name   = " . $device_hash->{name}         . "\n" .
                      "type   = " . $device_hash->{type}         . "\n" .
                      "label  = " . $device_hash->{label}        . "\n" .
                      "id     = " . $device_hash->{id}           . "\n" .
                      "fstype = " . $device_hash->{fstype}       . "\n\n" ;
        $self->sanlog(-msg => $tempmsg,
                      -msg_level => 1);
    }

    $self->sanlog(-msg => scalar(@devices)." devices found (may include multiple paths to devices)");

    return (wantarray()) ? @devices : \@devices;

}

################################################################################
# Method:    get_system_stats()
# Objective: Get the cpu/memory usage percentage over a specified interval
################################################################################
sub get_system_stats
{
    my ($self, $interval) = @_;

    $self->_clear_err();
    $self->{WinClient}->_clear_err();

    # collect the CPU usage over the desired interval (seconds)
    my @output = $self->{WinClient}->getSystemStats($interval);
    if (scalar(@output) == 0 or $self->{WinClient}->error()) {
        my $errmsg = join("\n", $self->{WinClient}->errmsg());
        $self->_set_err("Error: WinHost::get_system_stats() getSystemStats() failed.\n$errmsg");
        return {};
    }

    # Possible output can look like:
    #  C:\>win_system_stats.exe 60
    # Avg. CPU Util (%)     : 10%
    # Number of CPUs        : 1
    # Avg. Mem Used (%)     : 45%
    # Avg. Mem Free (kbytes) : 477862775
    # Avg. Mem Used (kbytes) : 595879049
    # Total Memory (kbytes)  : 1073741824
    # Page Faults (per sec) : 3990
    # Sample Interval (sec) : 60

    my $system_stats = {};

    # parse the output for the CPU utilization
    foreach my $line (@output) {
        if ($line =~ /Avg\. CPU Util.+?:\s+(\d+)\%/i) {
            # return the percentage to the caller
            $system_stats->{cpu_usage_percent} = $1;
        } elsif ($line =~ /Number of CPUs.+?:\s+(\d+)\%/i) {
            # return the percentage to the caller
            $system_stats->{num_cpus} = $1;
        } elsif ($line =~ /Avg\. Mem Used \(\%\).+?:\s+(\d+)\%/i) {
            # return the percentage to the caller
            $system_stats->{mem_usage_percent} = $1;
        } elsif ($line =~ /Page Faults.+?:\s+(\d+)\%/i) {
            # return the percentage to the caller
            $system_stats->{page_faults_per_sec} = $1;
        }
    }

    # return the system stats
    return $system_stats;
}

################################################################################
# Method:    get_ip_info()
# Objective: Get the IP information for this system and return hash with that
#            data
################################################################################
sub get_ip_info {
    my ($self) = @_;

    my ($mac, $ipv6, $description, $gateway);

    my @ipv4_array = ();
    my @subnet_mask_array = ();
    my $start = 0;
    my $finished = 0;
    my $interface = 0;
    my $interface_name = 'NA';
    my $ip_hash = {};

    # get route information for ipv6
    my $cmd = "netsh interface ipv6 show route";
    my ($route_status, @route_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($route_status > 0) {
        $self->_set_err("Error running command ($cmd)");
        return {};
    }

    # get all IP information from host via ipconfig
    $cmd = "ipconfig /all";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error running command ($cmd)");
        return {};
    }

    my $cnt = 0;
    my $num_lines = scalar(@output);
    foreach my $line (@output) {
        chomp($line);
        $cnt++;

        # Look for unique sections describing interface.
        if ($line =~ /^\s*Ethernet adapter/i) {
            if ($start == 1) {
                $finished = 1;
            }
            $start = 1;
        }

        # sentinel to know when we are done with an adapter
        if (($line =~ /^\w+.*\:$/ or $num_lines == $cnt) and
             $line !~ /^\s*Ethernet adapter/i) {
            $finished = 1;
        }

        if ($start == 1 and $finished == 1) {
            # only push in local interfaces
            if ($interface_name !~ /NA/i) {
                my $once_through = 0;

                # if ips exist in ipv4 array then create an interface for each otherwise
                # only create one interface using the ipv6 address
                while (scalar(@ipv4_array) or (length($ipv6) and not $once_through)) {
                    $once_through = 1;
                    my $ipv4 = "";
                    my $subnet_mask = "";
                    $ipv4 = pop(@ipv4_array) if(scalar(@ipv4_array));
                    $subnet_mask = pop(@subnet_mask_array) if(scalar(@subnet_mask_array));
                    $ip_hash->{$interface}->{interface_name} = $interface_name;
                    $ip_hash->{$interface}->{mac} = $mac;
                    $ip_hash->{$interface}->{description} = $description;
                    $ip_hash->{$interface}->{gateway} = $gateway;
                    $ip_hash->{$interface}->{subnet_mask} = $subnet_mask;
                    $ip_hash->{$interface}->{ipv4} = $ipv4;

                    $ip_hash->{$interface}->{ipv6} = $ipv6;

                    # calculate the ipv4 subnet
                    if ($ipv4 ne "" and $subnet_mask ne "") {
                        $ip_hash->{$interface}->{ipv4_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv4,
                                                                                        -ip_version => "ipv4",
                                                                                        -ipv4_netmask => $subnet_mask);
                    } else {
                        $ip_hash->{$interface}->{ipv4_subnet} = "";
                    }

                    # calculate the ipv6 subnet only if an ipv6 ip is defined and we have ipv6 route information
                    if ($ipv6 ne "" and not grep(/IPv6 is not installed|No entries were found/i, @route_output)) {
                        # parse the route output for the ipv6 route that our ipv6 address is using
                        my $ipv6_subnet;
                        my $ipv6_prefix_len;

                        #Publish  Type      Met  Prefix                    Idx  Gateway/Interface Name
                        #-------  --------  ---  ------------------------  ---  ------------------------
                        #No       Manual    256  ::1/128                     1  Loopback Pseudo-Interface 1
                        #No       Manual    256  fe80::/64                  17  Teredo Tunneling Pseudo-Interface
                        #No       Manual    256  fe80::/64                  12  Local Area Connection 2
                        #No       Manual    256  fe80::100:7f:fffe/128      17  Teredo Tunneling Pseudo-Interface
                        #No       Manual    256  fe80::5efe:10.60.242.119/128   13  isatap.{BB82DC48-E1C4-42DE-962E-E681419AA2BA}
                        #No       Manual    256  fe80::5efe:192.168.1.105/128   14  isatap.{5623BE34-6FA7-4981-B4AA-50D1AD18114D}
                        #No       Manual    256  fe80::d45f:f777:f28c:24e3/128   12  Local Area Connection 2
                        #No       Manual    256  ff00::/8                    1  Loopback Pseudo-Interface 1
                        #No       Manual    256  ff00::/8                   17  Teredo Tunneling Pseudo-Interface
                        #No       Manual    256  ff00::/8                   12  Local Area Connection 2

                        # filter out any lines in the route_output that have prefix length of /128
                        # because they do not represent subnets
                        @route_output = grep { !/\/128/ } @route_output;
                        @route_output = grep { !/\/0/ } @route_output;

                        foreach my $line (@route_output) {
                            # if we matched an ipv6 route line, parse out the ip and prefix length
                            # of this route
                            if ($line =~ /^\s*\S+\s+\S+\s+\S+\s+(\S+)\/(\d+)/) {
                                my $ipv6_route = $1;
                                $ipv6_prefix_len = $2;

                                # check if this route's subnet matches the subnet of our ipv6 address
                                # using the prefix length of this route
                                my $route_subnet = &SanUtil::get_ip_subnet(-ip_address => $ipv6_route,
                                                                           -ip_version => "ipv6",
                                                                           -ipv6_prefix_len => $ipv6_prefix_len);
                                $ipv6_subnet     = &SanUtil::get_ip_subnet(-ip_address => $ipv6,
                                                                           -ip_version => "ipv6",
                                                                           -ipv6_prefix_len => $ipv6_prefix_len);

                                # if the subnets match, then this is the prefix length we need to store
                                if ($route_subnet eq $ipv6_subnet) {
                                    last;
                                } else {
                                    $ipv6_subnet = "";
                                    $ipv6_prefix_len = "";
                                }
                            }
                        }

                        # store the ipv6 subnet and prefix len we discovered
                        $ip_hash->{$interface}->{ipv6_prefix_len} = $ipv6_prefix_len;
                        $ip_hash->{$interface}->{ipv6_subnet} = $ipv6_subnet;

                    } elsif ($ipv6 ne "") {
                        # default the prefix length to 64 if we can't figure it out
                        # from the netsh interface ipv6 show route command (HACK)
                        # 64 is the most common prefix length anyway.
                        $ip_hash->{$interface}->{ipv6_prefix_len} = "64";
                        $ip_hash->{$interface}->{ipv6_subnet} = &SanUtil::get_ip_subnet(-ip_address => $ipv6,
                                                                                        -ip_version => "ipv6",
                                                                                        -ipv6_prefix_len => "64");
                    }

                    $ip_hash->{$interface}->{mtusize} = "";
                    $ip_hash->{$interface}->{broadcast} = "";

                    # if this is the management interface of the client
                    # set an attribute
                    my $host_rec = &Tharn::host($self->hostname());
                    my $host_ip_default = $host_rec->{'default_ip'};
                    if ($ipv4 eq $host_ip_default) {
                        $ip_hash->{$interface}->{is_mgmt_interface} = 1;
                    } else {
                        $ip_hash->{$interface}->{is_mgmt_interface} = 0;
                    }

                    $interface++;
                }
            }

            $finished = 0;

            # initialize values
            $interface_name = "NA";
            $mac = $description = $gateway = $ipv6 = "";
            @ipv4_array = ();
            @subnet_mask_array = ();
        }

    # Sample network connection record from the ipconfig /all command that gets the ip information

    #Ethernet adapter Local Area Connection 2:
    #
    #   Connection-specific DNS Suffix  . :
    #   Description . . . . . . . . . . . : Intel(R) PRO/1000 MT Network Connection #
    #2
    #   Physical Address. . . . . . . . . : 00-50-56-AE-22-D5
    #   DHCP Enabled. . . . . . . . . . . : No
    #   Autoconfiguration Enabled . . . . : Yes
    #   IPv6 Address. . . . . . . . . . . : fd20:8b1e:b255:8319:1997:b0d9:4cac:b5a0(P
    #referred)
    #   Link-local IPv6 Address . . . . . : fe80::1997:b0d9:4cac:b5a0%11(Preferred)
    #   IPv4 Address. . . . . . . . . . . : 10.225.115.238(Preferred)
    #   Subnet Mask . . . . . . . . . . . : 255.255.254.0
    #   Default Gateway . . . . . . . . . : fe80::226:98ff:fe00:5c1%11
    #   DNS Servers . . . . . . . . . . . : fec0:0:0:ffff::1%1
    #                                       fec0:0:0:ffff::2%1
    #                                       fec0:0:0:ffff::3%1
    #   NetBIOS over Tcpip. . . . . . . . : Enabled

        # start pushing values into hash
        if ($start == 1) {
            # get IPv4/v6 address - different formats w2k3/2k8
            if ($line =~ /^\s+IP Address|IPv4 Address|IPv6 Address/i) {
                my $ip = $line;
                $ip =~ s/.*Address.*\:\s+//;
                $ip =~ s/\s*\(.*\)\s*//;
                if ($ip =~ /\d+\.\d+\.\d+\.\d+/) {
                    unshift (@ipv4_array, $ip);
                } else {
                    if (($line =~ /Link-local/i) and ($self->{iscsi}->{CONFIG_ISCSI_IPV6_EXCLUDE_LINK_LOCAL} == 1)){
                        next;
                    }
                    # get rid of the scope id from the ipv6 address
                    $ip =~ s/\%\d+//g;
                    $ipv6 = $ip;
                }
            }

           # get interface name
            if ($line =~ /^Ethernet adapter/i) {
                $interface_name = $line;
                $interface_name =~ s/^\s*Ethernet adapter\s*//i;
                $interface_name =~ s/\:\s*$//;
            }

            # get MAC address
            if ($line =~ /^\s+Physical Address/i) {
                $mac = $line;
                $mac =~ s/\s+Physical Address.*\: //i;
                $mac =~ s/\-//g;
            }

            # get description of interface
            if ($line =~ /^\s+Description/i) {
                $description = $line;
                $description =~ s/\s+Description.*\: //i;
            }

            # get subnet mask
            if ($line =~ /^\s+Subnet Mask/i) {
                $line =~ s/\s+Subnet Mask.*\: //i;
                unshift (@subnet_mask_array, $line);
            }

            # get gateway
            if ($line =~ /^\s+Default Gateway/i) {
                $gateway = $line;
                $gateway =~ s/\s+Default Gateway.*\: //i;
            }
        }

    }

    # return info hash
    return ($ip_hash);

}

################################################################################
# Subroutine Name: set_ip_address()
# Objective:       Sets the IP address and subnet mask for an interface
# Details:         See POD documentation at the end of this file
################################################################################
sub set_ip_address {
    my ($self, %args) = @_;

    my ($interface, $ip_address, $subnet_mask, $ip_info);

    # parse arguments
    foreach (keys %args) {
        if (/-?interface/) {
            unless (defined ($interface = $args{'interface'}) and length ($interface)) {
                $self->_set_err("Invalid value for argument '-interface'");
                return "";
            }
            next;
        }
        if (/-?ip_address/) {
            unless (defined ($ip_address = $args{'ip_address'}) and length ($ip_address)) {
                $self->_set_err("Invalid value for argument '-ip_address'");
                return "";
            }
            next;
        }
        if (/-?subnet_mask/) {
            unless (defined ($subnet_mask = $args{'subnet_mask'}) and length ($subnet_mask)) {
                $self->_set_err("Invalid value for argument '-subnet_mask'");
                return "";
            }
            next;
        }
    }

    # check for required params
    unless (defined ($interface) and length ($interface)) {
        $self->_set_err("Invalid value for argument '-interface'");
        return "";
    }
    unless (defined ($ip_address) and length ($ip_address)) {
        $self->_set_err("Invalid value for argument '-ip_address'");
        return "";
    }
    unless (defined ($subnet_mask) and length ($subnet_mask)) {
        # get_ip_info so we can find the current subnet mask
        # for the interface since one was not passed in.
        $ip_info = $self->get_ip_info();
        if($self->error) {
            $self->_set_err("Failed to get ip information" , $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # Set subnet mask to error incase the interface passed in is not found
        # if the interface doesn't exist then the subnet mask will remain error
        # and an error will be returned after searching through the ip info
        $subnet_mask = "error";
        foreach my $interface_name (keys %$ip_info) {
            if(lc($ip_info->{$interface_name}->{interface_name}) eq lc($interface)) {
                $subnet_mask = $ip_info->{$interface_name}->{subnet_mask};
                last;
            }
        }
        if ($subnet_mask eq "error") {
            $self->_set_err("Error: Initiator::set_ip() Failed to find the interface $interface");
            return $ClientUtil::FALSE;
        }
    }

    # set ip address based on passed in parameters
    my $cmd = "netsh interface ip set address source=static name=\"$interface\" addr=$ip_address mask=$subnet_mask";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error running command ($cmd)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_default_mp_type
# Objective:       return default multipathing stack type to use for this
#                  host platform
# Details:         See POD documentation at the end of this file
################################################################################
sub get_default_mp_type {
    my ($self) = @_;

    return("MPIO");
}

################################################################################
# Subroutine Name: rotate_logs()
# Objective:       rotate the log files for Windows
################################################################################
sub rotate_logs {
    my ($self) = @_;

    # NOOP

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    get_iscsi_info
# Objective: Collects iscsi info from the client.
#            Packages the information into a complex data structure.
#            Builds a hash of target nodenames.
#            Under each target nodename, there is a hash for each iscsi cmd:
#            alias      - contains scalar value (Can't get for windows)
#            connection - contains an array of hashes
#            initiator  - contains an array of hashes
#            interface  - contains an array of hashes
#            nodename   - contains scalar
#            security   - contains a hash
#            session    - contains an array of hashes
#            show       - contains a hash
#            status     - contains a hash
#            tpgroup    - contains an array of hashes
#            Each hash contains keys which correspond to the c-mode cmd options
#            Also added this hash entry which does not map to a command:
#            discovery_type - will return sendtargets or isns
# Details:
#     Inputs  - none
#     Returns $iscsi_hash - hash of client iscsi info
#             empty hash (do not set_err)  - no sessions or target portals
#             empty hash (set_err) - error condition
################################################################################
sub get_iscsi_info {
    my ($self) = @_;

    my $iscsi_hash;        # Hash to store iscsi info
    my $cmd;               # Cmd to execute
    my $status;            # Return status of cmd
    my @output;            # Output of the cmd
    my $init_node_name;    # iscsi node name of initiator

    $self->_clear_err();

    # Grab the WinClient object so we can execute cmds
    my $winclient_obj = $self->{WinClient};

    # use mpc to extract the nodename in win2k3
    if ($winclient_obj->{hostver} =~ /2003/) {
        my ($adapter_status, @adapter_output) = $winclient_obj->iscsiShowAdapters();
        if ($adapter_status != 0 or $winclient_obj->error()) {
            $self->_set_err("iscsiShow adapters failed", $winclient_obj->errmsg());
            return {};
        }

        # mpc iscsi show adapters
        #
        # iSCSINodeName: iqn.1991-05.com.microsoft:ibm3550-rtp14.rtp.netapp.com
        #
        # AdapterName     Index   Tag     Type    Address
        # -----------     -----   ---     ----    -------
        # MS-iSCSI-0      0       0       IPv4    10.60.242.119
        # MS-iSCSI-1      1       0       IPv4    192.168.1.189
        #
        # Success: iscsi show adapters

        foreach my $line (@adapter_output) {
            next if ($line =~ /AdapterName/i)
                 or ($line =~ /----/)
                 or ($line =~ /^\s*$/)
                 or ($line =~ /Success/i)
                 or ($line =~ /rstatus/i);

            # extract the initiator nodename
            if ($line =~ /iSCSINodeName: (\S+)/) {
                $init_node_name = $1;
                last;
            }
        }
    } else {
        # use wmic to grab the nodename in all other windows versions

        # Phase 1: call wmi to grab the initiator node name.
        $cmd = "wmic /NAMESPACE:\\\\root\\wmi PATH MSiSCSIInitiator_MethodClass";

        # execute the command
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0 or not grep(/iSCSINodeName/i, @output)) {
            $self->_set_err("wmic MSiSCSIInitiator_MethodClass failed.");
            return {};
        }
        # Expected output:
        # C:\Users\Administrator>wmic /NAMESPACE:\\root\wmi PATH MSiSCSIInitiator_MethodClass
        # iSCSINodeName
        # iqn.1991-05.com.microsoft:ssan-w2k8-7.smokejumper.rtp.netapp.com

        # Save the initiator iqn into the security section of the hash.  Do this in
        # case there are no sessions.  If there are no sessions, you can't get the
        # initiator iqn later.
        foreach my $line (@output) {
            if ($line =~ /(iqn\S+)/) {
                $init_node_name = $1;
                last;
            }
        }
    }

    # If iqn was not found, exit.
    if (not length($init_node_name)) {
        $self->_set_err("Unable to get initiator IQN");
        return {};
    }

    # Phase 2: call iscsicli listtargets to get target node names
    $cmd = "$winclient_obj->{cmd_iscsicli} listtargets";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0 or not grep(/The operation completed successfully./i, @output)) {
        $self->_set_err("$self->{cmd_iscsicli} listtargets failed.");
        return {};
    }

    # Expected output:
    # Microsoft iSCSI Initiator Version 6.0 Build 6000
    #
    # Targets List:
    #     iqn.1992-08.com.netapp:0
    # The operation completed successfully.

    # Check each line, searching for targets
    foreach my $line (@output) {

        # This specific pattern will yield a target nodename
        my ($target_node_name) = ($line =~ /^\s+(iqn\S+)$/);

        # If target_nodename was found, save the target/initiator nodenames.
        # Also initialize arrays.  May need to return some empty arrays.
        if ($target_node_name) {

            $iscsi_hash->{"$target_node_name"}->{"nodename"} =
              $target_node_name;
            $iscsi_hash->{"$target_node_name"}->{"show"}->{"target-name"} =
              $target_node_name;
            $iscsi_hash->{"$target_node_name"}->{"security"}->{"initiator-name"} =
              $init_node_name;
            $iscsi_hash->{"$target_node_name"}->{"connection"} = [];
            $iscsi_hash->{"$target_node_name"}->{"session"}    = [];
            $iscsi_hash->{"$target_node_name"}->{"initiator"}  = [];
            $iscsi_hash->{"$target_node_name"}->{"interface"}  = [];
            $iscsi_hash->{"$target_node_name"}->{"tpgroup"}    = [];
            $iscsi_hash->{"$target_node_name"}->{"status"}     = {};

        } ## end if ($target_node_name)

    } ## end foreach my $line (@output)

    # Phase 3 - Call iscsicli sessionlist to get session/connection info

    # build the command in iscsicli
    $cmd = "$winclient_obj->{cmd_iscsicli} sessionlist";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0 or not grep(/The operation completed successfully./i, @output)) {
        $self->_set_err("$self->{cmd_iscsicli} sessionlist failed.");
        return {};
    }

    # Expected output:
    # Microsoft iSCSI Initiator Version 6.0 Build 6000
    #
    # Total of 3 sessions
    #
    # Session Id             : fffffa80035cc438-400001370000032b
    # Initiator Node Name    : iqn.1991-05.com.microsoft:ssan-w2k8-7.smokejumper.rtp.netapp.com
    # Target Node Name       : (null)
    # Target Name            : iqn.1992-08.com.netapp:0
    # ISID                   : 40 00 01 37 00 00
    # TSID                   : 2c 03
    # Number Connections     : 1
    #
    #     Connections:
    #
    #         Connection Id     : fffffa80035cc438-347
    #         Initiator Portal  : 9.0.10.239/29133
    #         Target Portal     : 9.0.120.30/3260
    #         CID               : 01 00
    #
    # Devices:
    #     Device Type            : Disk
    #     Device Number          : 1
    #     Storage Device Type    : 7
    #     Partition Number       : 0
    #     Friendly Name          : NETAPP LUN SCSI Disk Device
    #     Device Description     : Disk drive
    #     Reported Mappings      : Port 2, Bus 0, Target Id 0, LUN 0
    #     Location               : Bus Number 0, Target Id 0, LUN 0
    #     Initiator Name         : Root\ISCSIPRT\0000_0
    #     Target Name            : iqn.1992-08.com.netapp:0
    #     Device Interface Name  : \\?\mpio#disk&ven_netapp&prod_lun&rev_0.2_#1&7f6ac24&0&36304139383030353732443435333334343441353533353533343937393331#{53f56307-b6bf-11d0-94f2-00a0c91efb8b}
    #     Legacy Device Name     : \\.\PhysicalDrive1
    #     Device Instance        : 0x82c
    #
    # Expected Result if there are 0 connections (iscsi down on target)
    # C:\Users\Administrator>iscsicli sessionlist
    # Microsoft iSCSI Initiator Version 6.0 Build 6000
    #
    # Total of 2 sessions
    #
    # Session Id             : fffffa80035f4438-400001370000017a
    # Initiator Node Name    : iqn.1991-05.com.microsoft:ssan-w2k8-7.smokejumper.rtp.netapp.com
    # Target Node Name       : (null)
    # Target Name            : iqn.1992-08.com.netapp:0
    # ISID                   : 40 00 01 37 00 00
    # TSID                   : 96 00
    # Number Connections     : 0
    #
    # Session Id             : fffffa80035f4438-400001370000017c
    # Initiator Node Name    : iqn.1991-05.com.microsoft:ssan-w2k8-7.smokejumper.rtp.netapp.com
    # Target Node Name       : (null)
    # Target Name            : iqn.1992-08.com.netapp:0
    # ISID                   : 40 00 01 37 00 01
    # TSID                   : 97 00
    # Number Connections     : 0

    # While there are still lines to process
    while (@output) {

        # Is this the start of a Session block?  If not throw the line away
        # and continue.  Else we have a session block and work to do.
        if ($output[0] !~ /^Session Id/) {
            shift (@output);
        } else {

            # Make hash to store session data.
            my %session;

            # Set some parms we know.
            # Session Type is always Normal.  MS initiator doesn't set alias.
            $session{"session-type"}    = "Normal";
            $session{"initiator-alias"} = "-";

            # Remove Session Id line
            shift (@output);

            # Next line contains Initiator Node Name, save it
            if ($output[0] !~ /^Initiator Node Name\s+:\s+(\S+)/) {
                $self->_set_err(  "$self->{cmd_iscsicli} sessionlist failed "
                                . "on Initiator Node Name line: $output[0]");
                return {};
            }
            $session{"initiator-name"} = $1;

            # Remove Initiator Node Name & Target Node Name lines
            for (1 .. 2) {
                shift (@output);
            }

            # Next line contains Target Name, save it
            if ($output[0] !~ /^Target Name\s+:\s+(\S+)/) {
                $self->_set_err(  "$self->{cmd_iscsicli} sessionlist failed "
                                . "on Target Name line: $output[0]");
                return {};
            }
            my $target_node_name = $1;

            # Let's make sure the target node name was previously found when
            # we called listtargets.  If the target node name was
            # not found earlier, we error out.
            if (not exists $iscsi_hash->{"$target_node_name"}) {
                $self->_set_err("Found a session with $target_node_name, "
                       . "but this target was not found by iscsicli listtargets"
                );
                return {};
            }

            # Remove Target Name line
            shift (@output);

            # Next line contains ISID, convert to ONTAP format and save it
            if ($output[0] !~ /^ISID\s+:\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/) {
                $self->_set_err(  "$self->{cmd_iscsicli} sessionlist failed "
                                . "on ISID line: $output[0]");
                return {};
            }
            $session{"isid"} = "$1:$2:$3:$4:$5:$6";

            # Remove ISID line
            shift (@output);

            # Next line contains TSID, save it
            if ($output[0] !~ /^TSID\s+:\s+(\S+)\s+(\S+)/) {
                $self->_set_err(  "$self->{cmd_iscsicli} sessionlist failed "
                                . "on TSID line: $output[0]");
                return {};
            }

            # Note windows prints tsid in hex, and byte swapped.  So we swap and
            # then convert to decimal to agree with Data ONTAP.
            # Also: Windows calls the parm 'tsid' while Data ONTAP uses 'tsih'.
            # Save as 'tsih' so verification routines work properly.
            $session{"tsih"} = hex("$2$1");

            # Remove TSID, Number Connections, and Blank Line
            for (1 .. 3) {
                shift (@output);
            }

            # If current line has Connections:, remove Connections: and blank
            # line.  $output[0] will then be pointing to the connection id line.
            if ($output[0] =~ /Connections:/) {
                for (1 .. 2) {
                    shift (@output);
                }
            }

            # Now process the connections.  While there is another connection:
            while ($output[0] =~ /^\s+Connection Id\s+:/) {

                # Hash to store connection data
                my %connection;

                # Copy some items from the session hash
                $connection{"tsih"} = $session{"tsih"};

                # Set some other parms we know
                $connection{"connection-state"} = "Full_Feature_Phase";
                $connection{"has-session"}      = "true";

                # Remove Connection Id line
                shift (@output);

                # Next line contains Initiator Portal (IP + Port), save it
                # Ex: Initiator Portal  : fd20:8b1e:b255:840b:89b7:faeb:cfb3:41df%0/37332'
                if ($output[0] !~ /^\s+Initiator Portal\s+:\s+(\S+)\/(\S+)/) {
                    $self->_set_err("$self->{cmd_iscsicli} sessionlist failed "
                                    . "on Initiator Portal line: $output[0]");
                    return {};
                }

                #check if ipv6 address, if so we
                #have to strip the CIDR number and compress
                #to match ONTAP formatting standard
                my $remote_address = $1;
                my $remote_port    = $2;
                $remote_address =~ s/\%\d+//g;
                if (&Net::IP::ip_is_ipv6($remote_address)){
                    $remote_address = &Net::IP::ip_compress_address($remote_address,6);
                }
                $connection{"remote-address"} = $remote_address;

                # Need to convert the initiator port due to burt 393716.  This
                # code can be updated if/when the burt is fixed.
                # Step 1 - Convert to hex
                my $val = sprintf("%x", $remote_port);

                # If the length of $val is not 4, we need to pad it with leading
                # zeroes.
                while (length $val < 4) {
                    $val = "0$val";
                }

                # Step 2 - Swap the bytes
                $val =~ s/(..)(..)/$2$1/;

                # Step 3 - convert to decimal and save
                $connection{"remote-ip-port"} = hex($val);

                # Remove Initiator Portal line
                shift (@output);

                # Next line contains Target Portal (IP + Port), save it
                if ($output[0] !~ /^\s+Target Portal\s+:\s+(\S+)\/(\S+)/) {
                    $self->_set_err("$self->{cmd_iscsicli} sessionlist failed "
                                    . "on Target Portal line: $output[0]");
                    return {};
                }

                #check if ipv6 address, if so we
                #have to strip the CIDR number and compress
                #to match ONTAP formatting standard
                my $local_address = $1;
                $connection{"local-ip-port"} = $2;
                $local_address =~ s/\%\d+//g;
                if (&Net::IP::ip_is_ipv6($local_address)){
                    $local_address = &Net::IP::ip_compress_address($local_address,6);
                }
                $connection{"local-address"} = $local_address;

                # Remove Target Portal line
                shift (@output);

                # Next line contains CID, save it
                if ($output[0] !~ /^\s+CID\s+:\s+(\S+) (\S+)/) {
                    $self->_set_err(  "$self->{cmd_iscsicli} sessionlist failed"
                                    . "on CID line: $output[0]");
                    return {};
                }

                # Note windows prints CID in hex, and byte swapped.  So we swap
                # and then convert to decimal to agree with Data ONTAP
                $connection{"connection-id"} = hex("$2$1");

                # Remove CID line
                shift (@output);

                # Save the CID into the session's connection-ids.
                # If the string is currently empty, add the first conn-id.
                if (!length($session{"connection-ids"})) {
                    $session{"connection-ids"} = $connection{"connection-id"};
                } else {

                    # Else, add a comma and the next conn-id.
                    $session{"connection-ids"} =
                        $session{"connection-ids"} . "," . $connection{"connection-id"};
                }

                # Save this connection
                push(@{$iscsi_hash->{"$target_node_name"}->{"connection"}},
                     \%connection);

            } ## end while ($output[0] =~ /^\s+Connection Id\s+:/)

            # If we found connections, we can populate these items:
            if (scalar @{$iscsi_hash->{"$target_node_name"}->{"connection"}} > 0) {

                # iscsi session update
                push(@{$iscsi_hash->{"$target_node_name"}->{"session"}}, \%session);

                # iscsi show / iscsi status update
                $iscsi_hash->{"$target_node_name"}->{"show"}->{"status-admin"} = "up";
                $iscsi_hash->{"$target_node_name"}->{"status"}->{"is-available"} = "true";
                $iscsi_hash->{"$target_node_name"}->{"status"}->{"status-admin"} = "up";

                # iscsi initiator update
                my %initiator;
                $initiator{"tsih"}            = $session{"tsih"};
                $initiator{"initiator-name"}  = $session{"initiator-name"};
                $initiator{"isid"}            = $session{"isid"};
                $initiator{"initiator-alias"} = "-";
                push(@{$iscsi_hash->{"$target_node_name"}->{"initiator"}},
                     \%initiator);

            }

        } ## end else [ if ($output[0] !~ /^Session Id/)

    } ## end while (@output)

    # Phase 4 - Call iscsicli targetinfo to get portal/interface/tpgroup info
    # We need target info for each target node name that we have previously
    # found.
    foreach my $target (keys %{$iscsi_hash}) {

        # build the command in iscsicli
        $cmd = "$winclient_obj->{cmd_iscsicli} targetinfo $target";

        # execute the command
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0 or not grep(/The operation completed successfully./i, @output)) {
            $self->_set_err("$self->{cmd_iscsicli} targetinfo failed.");
            return {};
        }

        # Expected output:
        # C:\Users\Administrator>iscsicli targetinfo iqn.1992-08.com.netapp:0
        # Microsoft iSCSI Initiator Version 6.0 Build 6000
        #
        #     Discovery Mechanisms :
        # ++++ "SendTargets:*9.0.120.30 0003260 Root\ISCSIPRT\0000_0 9.0.10.239"
        #
        # The operation completed successfully.
        # isns example:
        # C:\Users\Administrator>iscsicli targetinfo iqn.1992-08.com.netapp:sn.118047048
        # Microsoft iSCSI Initiator Version 6.0 Build 6000
        #
        #     Discovery Mechanisms :
        # ++++ "iSNS:9.0.10.239"
        #
        # The operation completed successfully.

        # If this target has been discovered by both sendtargets and iSNS, let's
        # raise an error.  This will confuse the verification methods.
        if (grep(/ \"SendTargets:/i, @output) and grep(/ \"iSNS:/i, @output)) {
            $self->_set_err(
                       "Target was discovered by both sendtargets and isns. "
                     . "Please only use 1 discovery method.  Target = $target");
            return {};
        }

        # Scan the output
        foreach my $line (@output) {

            # Search for a Discovery Mechanism line - enclosed in ""
            my ($discovery_mechanism) = ($line =~ /\s+(".+")/);

            if ($discovery_mechanism) {

                # Check for discovery mechanism and save it
                if ($discovery_mechanism =~ /sendtargets/i) {
                    $iscsi_hash->{"$target"}->{"discovery_type"} =
                      "sendtargets";
                } elsif ($discovery_mechanism =~ /isns/i) {
                    $iscsi_hash->{"$target"}->{"discovery_type"} = "isns";

                    # Fix for burt 560365.  So if the target was discovered with
                    # isns and it has zero connections, we want to delete this
                    # target from the hash.  It's not relevant and breaks
                    # IscsiHelperCM verify routines.  We then use the last
                    # command to exit the foreach my $line loop and go process
                    # the next target.
                    if (scalar(@{$iscsi_hash->{"$target"}->{"connection"}}) eq
                        0) {
                        delete($iscsi_hash->{"$target"});
                        last;
                    }
                } else {
                    $self->_set_err("Could not determine discovery type!");
                    return {};
                }

                # build the command in iscsicli
                $cmd = "$winclient_obj->{cmd_iscsicli} targetinfo $target "
                  . "$discovery_mechanism";

                # execute the command
                my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0 or not grep(/The operation completed successfully./i, @output)) {
                    $self->_set_err("$self->{cmd_iscsicli} targetinfo "
                                      . "<discovery mechanism> failed.");
                    return {};
                }

                # Expected sample output:
                # C:\Users\Administrator>iscsicli targetinfo iqn.1992-08.com.netapp:0  "SendTargets:*9.0.120.30 0003260 Root\ISCSIPRT\0000_0 "
                # Microsoft iSCSI Initiator Version 6.0 Build 6000
                #
                #     Discovery Mechanisms :
                # ++++ "SendTargets:*9.0.120.30 0003260 Root\ISCSIPRT\0000_0 "
                #     Protocol Type        : iSCSI TCP Protocol
                #     Target Alias         :
                #     PortalGroups         : 3 portal groups
                #         Group 0 has 1 portals
                #             Address and Socket    : 9.0.120.30 3260
                #             Symbolic Name         :
                #         Group 1 has 2 portals
                #             Address and Socket    : 9.0.124.30 3260
                #             Symbolic Name         :
                #             Address and Socket    : 9.0.123.30 3260
                #             Symbolic Name         :
                #         Group 2 has 1 portals
                #             Address and Socket    : 9.0.121.30 3260
                #             Symbolic Name         :
                #     Initiator Name       : Root\ISCSIPRT\0000_0
                #     Target Flags         : 0x0
                # ++Version              : 0
                # ++Information Specified: 0x0
                # ++Login Flags          : 0x0
                #     Target Mappings Failed : This function is not supported on this system.
                #
                # The operation completed successfully.

                # IPv6 appears like this:
                #         Group 0 has 1 portals
                #             Address and Socket    : [fd20:8b1e:b255:8320::450] 3260
                #             Symbolic Name         :

                # While there are still lines to process
                while (@output) {

                    # Is this the start of a Group block?  If not throw the line
                    # away and continue.  Else we have a Group block.
                    if ($output[0] !~ /^\s+Group/) {
                        shift (@output);
                    } else {

                        # We have a Group to process. Create hash.
                        my %tpgroup;

                        # Remove Group line
                        shift (@output);

                        # While we keep finding tpgroup members, grab ip and port.
                        # Use the ? to make it non-greedy (otherwise the ] will
                        # get captured).  Note [] appears for ipv6 only.
                        while ($output[0] =~ /^\s+Address and Socket\s+:\s+\[*(.+?)\]*\s+(\S+)/) {

                            # Update the tpgroup and interface
                            push(@{$tpgroup{"lif"}}, $1);
                            my %interface;
                            $interface{"address"} = $1;
                            $interface{"ip-port"} = $2;
                            # Leaving is-home set to true for now.  There is no
                            # lif failover in rollingrock, should always be true.
                            $interface{"is-home"} = "true";
                            $interface{"enabled"} = "true";

                            # Save interface.  Don't save tpgroup yet
                            # because there may be more members.
                            push(@{$iscsi_hash->{"$target"}->{"interface"}}, \%interface);

                            # Remove Address and Symbolic Name line
                            for (1 .. 2) {
                                shift (@output);
                            }

                        } ## end while ($output[0] =~ ...

                        # We now have all of the tpgroup data, so save it.
                        # Can't get tpgroup tag from MS iscsi initiator.
                        push(@{$iscsi_hash->{"$target"}->{"tpgroup"}}, \%tpgroup);

                    } ## end else [ if ($output[0] !~ /^\s+Group/)

                } ## end while (@output)

                # Break out of the for loop.  Even if there are multiple
                # discovery mechanisms for the same target, they should be
                # returning the same target portal information.  If we want to
                # test sendtargets to multiple portals on the same target, that
                # can be done with a thpl which performs sendtargets to portal
                # 1, saves the data, removes portal 1, then performs sendtargets
                # to portal 2.
                last;

            } ## end if ($discovery_mechanism)

        } ## end foreach my $line (@output)

    } ## end foreach my $target (keys %{...

    # Phase 5 - Call iscsicli listtargetportals to get some session/connection/
    # security info
    $cmd = "$winclient_obj->{cmd_iscsicli} listtargetportals";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0 or not grep(/The operation completed successfully./i, @output)) {
        $self->_set_err("$self->{cmd_iscsicli} listtargetportals failed.");
        return {};
    }

    # Expected output:
    # C:\Users\Administrator>iscsicli listtargetportals
    # Microsoft iSCSI Initiator Version 6.0 Build 6000
    #
    # Total of 1 portals are persisted:
    #
    #     Address and Socket   : 9.0.120.30 3260
    #     Symbolic Name        :
    #     Initiator Name       : Root\ISCSIPRT\0000_0
    #     Port Number          : 4
    # ++Security Flags        : 0x0
    # ++Version              : 0
    # ++Information Specified: 0x1f
    # ++Header Digest        : CRC-32C
    # ++Data Digest          : CRC-32C
    # ++Maximum Connections  : 32
    # ++Default Time 2 Wait  : 0
    # ++Default Time 2 Retain: 20
    # ++Login Flags          : 0x2
    # ++                       Multipath Enabled
    # ++Authentication Type  : CHAP
    # ++Username             :
    # +++++++++++++6d 69 6b 65                                         mike
    # ++Password             : <is established>

    # While there are still lines to process
    while (@output) {
        $self->sanlog(-msg       => "Found output: $output[0]",
                      -msg_level => 7);

        # Is this the start of an Address block?  If not throw the line away
        # and continue.  Else we have an Address block and work to do.
        my ($ip_addr) = ($output[0] =~ /^\s+Address and Socket\s+:\s+(\S+)/);
        if (not defined($ip_addr)) {
            shift (@output);
            $self->sanlog(-msg       => "Throw out line.",
                          -msg_level => 7);
        } else {

            $self->sanlog(-msg => "listtargetportals processing: $ip_addr",
                          -msg_level => 7);
            my $target;

            # Need to figure out which target this ip represents
            # Start by scanning each target we have saved
            foreach my $target_to_scan (keys %{$iscsi_hash}) {

                $self->sanlog(
                             -msg => "Scanning $target_to_scan for an ip match",
                             -msg_level => 7);

                # Then scan each interface
                foreach my $int (@{$iscsi_hash->{"$target_to_scan"}->{"interface"}}) {

                    my $ip_to_check = $int->{"address"};

                    $self->sanlog(-msg       => "Checking $ip_to_check",
                                  -msg_level => 7);

                    # If we find a match, save the name of the target
                    if ($ip_addr eq $ip_to_check) {
                        $target = $target_to_scan;
                        $self->sanlog(-msg       => "Found a match! $ip_addr belongs to $target",
                                      -msg_level => 7);

                        # Exit the foreach interface loop since we have a match
                        last;
                    } ## end if ($ip_addr eq $ip_to_check)
                } ## end foreach my $int (@{$iscsi_hash...

                # If we have a target, exit the foreach $target_to_scan loop
                if ($target) {
                    $self->sanlog(-msg       => "Exit outer loop",
                                  -msg_level => 7);
                    last;
                }

            } ## end foreach my $target_to_scan ...

            # Check to make sure we found a target.  If not, raise a warning
            # and keep processing the rest of the output.
            if (not defined($target)) {
                &SanUtil::nate_log('-warn',
                        "iscsicli listtargetportals: Couldn't determine "
                      . "which target $ip_addr belongs to.  Make sure this\n"
                      . "is expected for your test case.  An example of where "
                      . "this may be expected: Perform sendtargets discovery\n"
                      . "then stop iscsi service on the target.  If you then "
                      . "login to another vserver, mpc iscsi show targets will\n"
                      . "be called, which forces a refresh.  You will then "
                      . "have the ip address show up in listtargetportals, but\n"
                      . "the target will no longer show up in listtargets.");
                shift (@output);
                next;
            } ## end if (not defined($target...

            # Throw away Adress and Socket, Symbolic Name, Initiator Name, and
            # Port Number lines, not needed
            for (1 .. 4) {
                shift (@output);
            }

            # Varibales for info we are searching for
            my $header_digest;
            my $data_digest;
            my $auth_method;
            my $chap_user_name;

            # While the next line is not empty.
            # An empty line will signal the end of a targetportal block.
            while ($output[0] !~ /^$/) {
                $self->sanlog(-msg       => "Found output: $output[0]",
                              -msg_level => 7);

                if ($output[0] =~ /Header Digest\s+:\s+CRC-32C/) {
                    $header_digest = "true";
                    $self->sanlog(-msg       => "Set Header Digest to true",
                                  -msg_level => 7);
                }
                if ($output[0] =~ /Data Digest\s+:\s+CRC-32C/) {
                    $data_digest = "true";
                    $self->sanlog(-msg       => "Set Data Digest to true",
                                  -msg_level => 7);
                }

                if ($output[0] =~ /Authentication Type\s+:\s+(\S+)/) {
                    if ($1 =~ /CHAP|Mutual/) {
                        $auth_method = "CHAP";
                        $self->sanlog(-msg       => "Set Auth Type to CHAP",
                                      -msg_level => 7);
                    }
                }
                if ($output[0] =~ /Username\s+:/) {
                    if ($output[1] =~ /\s(\S+)$/) {
                        $chap_user_name = $1;
                        $self->sanlog(
                                     -msg => "Set User Name to $chap_user_name",
                                     -msg_level => 7);
                    } else {
                        $self->_set_err(
                                     "$self->{cmd_iscsicli} listtargetportals: "
                                       . "Couldn't determine CHAP user name.");
                        return {};
                    }
                } ## end if ($output[0] =~ /User Name\s+:\s+(\S+)/)

                # NOTE: Used to search for Maximum Connections, Default Time 2
                # Wait, and Default Time 2 Retain.  These lines only appear if
                # automation (MPC) is used.  I believe they only display values
                # which were passed to MPC, not the true negotiated values.

                # Throw out line, goto next one.
                shift (@output);

            } ## end while ($output[0] !~ /^$/)

            # If we didn't set the auth method to CHAP, then set it to none.
            # Also, if auth method is none, then we know the chap user names
            # are not configured.
            if (!$auth_method) {
                $auth_method = "none";
                $iscsi_hash->{"$target"}->{"security"}->{"user-name"} = '-';
                $iscsi_hash->{"$target"}->{"security"}->{"outbound-user-name"} = '-';
            }

            # If we didn't set the digests, then they are false
            if (!$header_digest) {
                $header_digest = "false";
            }
            if (!$data_digest) {
                $data_digest = "false";
            }

            # Save the security data we gathered.  Note we can only populate 1
            # security entry for the initiator, it's not an array.
            $iscsi_hash->{"$target"}->{"security"}->{"auth-type"} = $auth_method;
            if ($chap_user_name) {
                $iscsi_hash->{"$target"}->{"security"}->{"user-name"} = $chap_user_name;
            }

            # Populate connection info.  Note that we are assuming each
            # connection is inheriting the properties from the discovery
            # session - target portal.  This is true when using SanClient
            # libraries.  Note that it is possible to establish a connection
            # with custom parms that are different from the listed targetportal
            # parms.  I don't believe the automation currently allows this
            # though.  An example would be: perform sendtargets with CHAP.  Then
            # perform a login without CHAP.  Note that other initiators like
            # Solaris/Linux do a better job of displaying these parms per
            # connection.
            foreach my $conn (@{$iscsi_hash->{"$target"}->{"connection"}}) {
                $conn->{"authentication-method"} = $auth_method;
                $conn->{"header-digest-enabled"} = $header_digest;
                $conn->{"data-digest-enabled"}   = $data_digest;
            }

            # Here, we used to save Max Connections, Default Time 2 Wait, and
            # Default Time 2 Retain into each session.
            # Please see comments above for why we are no longer processing
            # these parms!

            $self->sanlog(-msg => "Completed scanning of target portal $ip_addr",
                          -msg_level => 7);

        } ## end else [ if not defined ($ip_addr)

    } ## end while (@output)

    # Future Possibility: Look at registry settings for iscsi.  Some of them
    # may be valid.  However, many are negotiated and would not necessarily
    # be correct.

    return $iscsi_hash;
} ## end sub get_iscsi_info

################################################################################
# Method:    get_fcp_info
# Objective: Collects fcp info from the client.
#            Packages the information into a complex data structure.
#            Builds a hash of target nodenames.
#            Under each target nodename, there is a hash for each fcp cmd:
#            discoveredports      - contains a hash.
#            Each hash contains keys which correspond to the c-mode cmd options
#            Also added this hash entry which does not map to a command:
#            discovery_type - will return sendtargets or isns
# Details:
#     Inputs  - none
#     Returns $fcp_info_hash - hash of client iscsi info
#             empty hash (do not set_err)  - no sessions or target portals
#             empty hash (set_err) - error condition
################################################################################
sub get_fcp_info {
    my ($self, %args) = @_;
    my %fcp_info_hash = ();

    # First get the adapters information
    my $fcp_show_adapters_hash = $self->_mpc_fcp_show_adapters();
    return {} if($self->error());

    # Put the adapters info into the fcp_info_hash
    foreach my $adaptername (keys %{$fcp_show_adapters_hash}) {
        $fcp_info_hash{$adaptername}->{adapters} = $fcp_show_adapters_hash->{$adaptername};
    }

    # Second get the discoveredports information
    my $fcp_show_discoveredports_hash = $self->_mpc_fcp_show_discoveredports();
    return {} if($self->error());

    # Put the discoveredports info into the fcp_info_hash
    foreach my $adaptername (keys %{$fcp_show_discoveredports_hash}) {
        $fcp_info_hash{$adaptername}->{discoveredports} = $fcp_show_discoveredports_hash->{$adaptername};
    }

    return \%fcp_info_hash;
} ## end sub get_fcp_info

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _mpc_fcp_show_adapters()
# Objective:       Calls mpc fcp show adapters
# Details:         Make as call to mpc fcp show adapters and formats the data
#                  in a hash.
################################################################################
sub _mpc_fcp_show_adapters {
    my ($self, %args) = @_;

    my $cmd;               # Cmd to execute
    my $status;            # Return status of cmd
    my @output;            # Output of the cmd
    my %return_hash = ();  # Return of the method

    $self->_clear_err();

    # Call mpc fcp show adapters command.
    $cmd = "mpc fcp show adapters";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0 or not grep(/Success: fcp show adapters/i, @output)) {
        $self->_set_err("mpc fcp show adapters failed.");
        return {};
    }

    #         The output array looks like the following
    #         $VAR1 = [
    #         '',
    #         '',
    #         'AdapterName          State           Type            NodeName                PortName',
    #         '-----------          -----           ----            --------                --------',
    #         'com.qlogic-QLE2462-0 Operational     Fabric  20:00:00:e0:8b:9c:5c:c2 21:00:00:e0:8b:9c:5c:c2',
    #         'com.qlogic-QLE2462-1 Operational     Fabric  20:01:00:e0:8b:bc:5c:c2 21:01:00:e0:8b:bc:5c:c2',
    #         '',
    #         'Success: fcp show adapters'
    #          ]

    # While there are still lines to process
    while (@output) {
        $self->sanlog(-msg       => "Found output: $output[0]",
                      -msg_level => 7);

        # Is this the start of an info block?  If not throw the line away
        # and continue.  Else we have an info block and work to do.
        # The Info block will have a NODENAME which has a WWPN
        my ($info_line) = shift (@output);

        if (not  ($info_line =~ /([0-9a-fA-F][0-9a-fA-F]:){7}[0-9a-fA-F][0-9a-fA-F]/)) {
            $self->sanlog(-msg       => "Throw out line.",
                          -msg_level => 7);
        } else {
            my ($AdapterName,$State,$Type,$NodeName,$PortName) = split(/\s+/,$info_line);
            $return_hash{$AdapterName} = {state => $State,
                                          type  => $Type,
                                          nodename => $NodeName,
                                          portname => $PortName};
        }
    }
    return \%return_hash;
}

################################################################################
# Subroutine Name: _mpc_fcp_show_discoveredports()
# Objective:       Calls mpc fcp show discoveredports
# Details:         Make as call to mpc fcp show discoveredports and formats the data
#                  in a hash.
################################################################################
sub _mpc_fcp_show_discoveredports {
    my ($self, %args) = @_;

    my $cmd;               # Cmd to execute
    my $status;            # Return status of cmd
    my @output;            # Output of the cmd
    my %return_hash = ();  # Return of the method

    $self->_clear_err();

    # Call mpc fcp show discoveredports command.
    $cmd = "mpc fcp show discoveredports";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0 or not grep(/Success: fcp show discoveredports/i, @output)) {
        $self->_set_err("mpc fcp show discoveredports failed.");
        return {};
    }

    #         The output array looks like the following
    #$VAR1 = [
    #      '',
    #      '',
    #      'AdapterName          State           Type            NodeName                PortName',
    #      '-----------          -----           ----            --------                --------',
    #      'com.qlogic-QLE2462-0 Operational     Public_Loop     20:01:00:a0:98:26:f7:2a 20:09:00:a0:98:26:f7:2a',
    #      'com.qlogic-QLE2462-0 Operational     Public_Loop     20:01:00:a0:98:26:f7:2a 20:08:00:a0:98:26:f7:2a',
    #      '',
    #      'AdapterName          State           Type            NodeName                PortName',
    #      '-----------          -----           ----            --------                --------',
    #      'com.qlogic-QLE2462-1 Operational     Public_Loop     20:01:00:a0:98:26:f7:2a 20:06:00:a0:98:26:f7:2a',
    #      'com.qlogic-QLE2462-1 Operational     Public_Loop     20:01:00:a0:98:26:f7:2a 20:07:00:a0:98:26:f7:2a',
    #      '',
    #      'Success: fcp show discoveredports'
    #    ];

    # While there are still lines to process
    while (@output) {
        $self->sanlog(-msg       => "Found output: $output[0]",
                      -msg_level => 7);

        # Is this the start of an info block?  If not throw the line away
        # and continue.  Else we have an info block and work to do.
        # The Info block will have a NODENAME which has a WWPN
        my ($info_line) = shift (@output);

        if (not  ($info_line =~ /([0-9a-fA-F][0-9a-fA-F]:){7}[0-9a-fA-F][0-9a-fA-F]/)) {
            $self->sanlog(-msg       => "Throw out line.",
                          -msg_level => 7);
        } else {
            my ($AdapterName,$State,$Type,$NodeName,$PortName) = split(/\s+/,$info_line);
            push (@{$return_hash{$AdapterName}},{state => $State,
                                              type  => $Type,
                                              nodename => $NodeName,
                                              portname => $PortName});
        }
    }
    return \%return_hash;
}

################################################################################
# Subroutine Name: _fcp_deconfigure()
# Objective:       Deconfigures fcp
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _fcp_deconfigure {
    my ($self, %args) = @_;

    # No-op (for now)
    return $ClientUtil::TRUE;
}

################################################################################
# Method:    _get_devices
# Objective: Wrapper to allow backwards compatibility with get_devices
################################################################################
sub _get_devices {
    my ($self, %args) = @_;

    return ($self->get_devices(%args));
}

################################################################################
# Method:    _get_unlabeled_devices
# Objective: Retrieve devices that do not have a label. Since this is called
#            from discover, always call get_devices with refresh = 1 to get
#            from system and not cache.
# Details:   TBD - private methods aren't in POD, add details
################################################################################
sub _get_unlabeled_devices {
    my ($self) = @_;

    my @devices;

    my @device_array = $self->get_devices(-refresh => 1);
    return $ClientUtil::FALSE if ($self->error);

    #
    # Determine which devices do not have a label
    #
    foreach my $device_hash (@device_array) {
        foreach my $key (keys %$device_hash) {
            $self->sanlog(-msg=> "_get_unlabelled_devices() device hash: " .
                                 "key is $key value is "                   .
                                 $device_hash->{$key} . "\n",
                          -msg_level => 7);
        }

        #
        # Skip non ontap devices.
        # Unlabeled devices will have a RAW file system type
        #
        next if ($device_hash->{type} !~ /NETAPP|IBM-Nseries/);

        # next unless ($device_hash->{fstype} =~ /RAW/i);
        next if (defined($device_hash->{fstype}));

        $self->sanlog(-msg => "_get_unlabelled_devices() pushing device" .
                              $device_hash->{path} . "\n",
                      -msg_level => 7);
        push (@devices, $device_hash);
    }

    return (wantarray()) ? @devices : \@devices;
}


################################################################################
# Subroutine Name: _iscsi_configure()
# Objective:       Configure iscsi to include the targets for the filers.
#                  Assumptions: a targets list is passed into this function with
#                  the IP's of all the ports on each filer.
# Details:         See POD documentation at the end of this file
# Notes:           The IP addresses of the targets array must be strings.
################################################################################
sub _iscsi_configure {
    my ($self, %args) = @_;

    my ($status, @output, @targets, $num_logins_to_make);

    my $cache_string = "iscsi_show_targets";

    $self->_clear_err();

    # check if iscsi discovery is false, if so then exit from the function.
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY} != 1) {
        $self->sanlog(-msg => "iSCSI Configure: Bypassing iSCSI Configuration by user request (CONFIG_ISCSI_DISCOVERY=0)");
        return $ClientUtil::TRUE;
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => %args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@targets)) {
        $self->_set_err("Missing or invalid \'targets\' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring iSCSI");

    # make sure the sw iscsi service is enabled (only on win2k8+)
    if ($self->{WinClient}->{hostver} =~ /2008|windows server 7/) {
        ($status, @output) = &ClientUtil::execute($self->{conn}, "net start \"Microsoft iSCSI Initiator Service\"");
        if ($status != 0 or not grep(/was started successfully|has already been started/i, @output)) {
            $self->_set_err("Failed to start the iSCSI service on the client");
            return $ClientUtil::FALSE;
        }
    }

    # if requested to do subnet matching, filter the targets list
    # so that it contains targets that are accessible by this host
    @targets = $self->_filter_targets_by_subnet(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # In C-mode, targets will have either ipv4 or ipv6 address unlike 7-mode.
    # Hence, remove the targets from the target list which don't have valid IP address
    # of IP version requested by user.
    @targets = $self->filter_targets_by_ipversion(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # check what ip version is being requested and set the ip mode on all the target objects
    if ($self->{iscsi}->{CONFIG_ISCSI_IPVERSION} =~ /^ipv4$|^ipv6$|^auto$/i) {
        foreach my $target (@targets) {
            # set the ip mode to be either ipv4 or ipv6
            $target->set_ip_mode($self->{iscsi}->{CONFIG_ISCSI_IPVERSION});
            if ($target->error()) {
                $self->_set_err("Could not set the ip mode of target $target->{nodename} to " .
                                 $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} . "\n" .
                                "Reason: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    } else {
       $self->_set_err("Invalid value for CONFIG_ISCSI_IPVERSION: " .
                        $self->{iscsi}->{CONFIG_ISCSI_IPVERSION});
       return $ClientUtil::FALSE
    }

    # validate all the iscsi configuration parameters on the {iscsi} options hash
    $self->_iscsi_configure_validate_params();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # build the initiator hash
    my ($initiator_hsh, $init_node_name) = $self->_iscsi_configure_build_initiator_hash();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # determine whether or not default the chap username to the initiator name of this host
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} eq "") {
        # default the chap username to the initiator nodename
        $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} = $init_node_name;
    }
    $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_USERNAME} = $init_node_name;

    # print out a summary of what we are going to do for iSCSI Configure
    $self->_iscsi_configure_print_summary(-targets => \@targets);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # deconfigure iscsi if requested to do so before configuring.
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} > 0) {
        $self->_iscsi_deconfigure(-targets => \@targets,
                                  -clean_level => $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP});
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    my $winclient_obj = $self->{WinClient};

    # set up a chap mutual secret if we are doing bidirectional chap
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /bidirectional/i) {

        $self->sanlog(-msg => "iSCSI Configure: Setting up CHAP Mutual Secret.");

        ($status, @output) = $winclient_obj->iscsiSetMutualCHAPSecret(-secret => $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD});
        if ($status != 0 or $winclient_obj->error()) {
            my @errmsg = $winclient_obj->errmsg();
            $self->_set_err("iscsiSetMutualCHAPSecret failed: @errmsg");
            return $ClientUtil::FALSE;
        }
    }

    $self->sanlog(-msg => "iSCSI Configure: Configuring '" . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} . "' Discovery.");

    # At this point add the iSNS server if it was requested to be done
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^isns$/i) {
        $self->_iscsi_configure_isns_discovery();
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }
    }

    my $round_robin_count = 0;
    my $loop_counter = 0;

    # iterate through the targets array with the intention of discovering each target and
    # creating sessions for it.
    foreach my $target (@targets) {

        # Determine which initiator index to use to log in to this target by matching the
        # subnets between the initiator and the target.
        my $initiator_index = 0;
        my $initiator_ip = "";

        # obtain the target's attributes
        my $target_subnet = $target->get_subnet();
        if ($target->error()) {
            $self->_set_err("Could not obtain subnet from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }
        my $target_ip = $target->get_ip();
        if ($target->error()) {
            $self->_set_err("Could not obtain ip from target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }
        my $nodename = $target->get_nodename();
        if ($target->error()) {
            $self->_set_err("Could not get nodename of target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }
        my $target_mode = $target->get_mode();
        if ($target->error()) {
            $self->_set_err("Could not get mode of target: " . $target->errmsg());
            return $ClientUtil::FALSE;
        }
        # if doing subnet matching, choose the initiator that matches this subnet
        if ($self->{iscsi}->{CONFIG_ISCSI_SUBNET_MATCHING} == 1) {
            foreach my $init_ip (keys %{$initiator_hsh}) {
                if ($initiator_hsh->{$init_ip}->{subnet} eq $target_subnet) {
                    # we matched the subnets between the initiator and target
                    $initiator_index = $initiator_hsh->{$init_ip}->{index};
                    $initiator_ip = $init_ip;
                    last;
                }
            }
        } else {
            # otherwise don't do subnet matching, selectively choose round robin which initiator to pick
            # to login
            my $ip_version = $self->{iscsi}->{CONFIG_ISCSI_IPVERSION};
            my @inits = sort keys %{$initiator_hsh};
            $initiator_ip = $inits[$round_robin_count % scalar(@inits)];
            $round_robin_count++;
            # make sure that the initiator port that we selected round robin is
            # running the same ip version as the ip version that was selected by the test
            if ($initiator_hsh->{$initiator_ip}->{ip_version} =~ /$target_mode/i) {
                $initiator_index = $initiator_hsh->{$initiator_ip}->{index};

                # reset inf loop counter
                $loop_counter = 0;
            } else {
                # otherwise try again with the next initiator in the list.
                # check if inf loop counter is equal to scalar(@inits)
                # if equal set err and return
                if ($loop_counter == scalar(@inits)) {
                    $self->_set_err("Round robin selection of initiator ports failed for target ip '$target_ip'");
                    return $ClientUtil::FALSE;
                } else {
                    # else incrment inf loop counter here and redo
                    $loop_counter++;
                    redo;
                }
            }
        }

        # if there is no initiator that matches the subnet of this target, then
        # this is a fatal failure. we can't login to this target!
        if ($initiator_ip eq "") {
            $self->_set_err("Could not find an initiator interface that matches subnet '$target_subnet'" .
                            "of the target '$target_ip' " . Dumper($initiator_hsh));
            return $ClientUtil::FALSE;
        }

        # check to see if this target ip is already added
        if ($self->is_cache_data_valid($cache_string)) {
            @output = $self->get_cache_data($cache_string);
        } else {
            ($status, @output) = $winclient_obj->iscsiShow(-object => "targets");
            if ($status != 0 or $winclient_obj->error()) {
                my @errmsg = $winclient_obj->errmsg();
                $self->_set_err("iscsiShow failed: @errmsg");
                return $ClientUtil::FALSE;
            }

            #update the cache with the new data for this file
            $self->store_cache_data($cache_string, @output);
        }

        # if the target portal for this target is not added yet, then add one
        if (not grep(/($nodename)\s+\d+\:\d+\s+\[?($target_ip)\]?\s+3260/, @output)) {

            # If the discovery type is isns, the target portal should have already been
            # added. If it hasn't, something did not go right and we have to exit with
            # an error status
            if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^isns$/i) {
                $self->_set_err("The isns target discovery failed. No target portals have " .
                                "been found for node $nodename on target $target_ip.");
                return $ClientUtil::FALSE;
            }

            # add a target portal for $target_ip
            ($status, @output) =
                $winclient_obj->iscsiAddTargetPortal(-filer => $target_ip,
                                                     -mpio => $self->{iscsi}->{CONFIG_ISCSI_WIN_MPIO},
                                                     -datadigest => $self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST},
                                                     -headerdigest => $self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST},
                                                     -initiator_index => $initiator_index,
                                                     -maxconnections => $self->{iscsi}->{CONFIG_ISCSI_WIN_MAX_CONNECTIONS},
                                                     -auth_type => $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE},
                                                     -chap_username => $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME},
                                                     -chap_password => $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD},
                                                     -default_time2wait => $self->{iscsi}->{CONFIG_ISCSI_DEFAULT_TIME2WAIT},
                                                     -default_time2retain => $self->{iscsi}->{CONFIG_ISCSI_DEFAULT_TIME2RETAIN});
            if ($status != 0 or $winclient_obj->error()) {
                my @errmsg = $winclient_obj->errmsg();
                $self->_set_err("iscsiAddTargetPortal failed: @errmsg");
                return $ClientUtil::FALSE;
            }

            # display the targets table again
            ($status, @output) = $winclient_obj->iscsiShow(-object => "targets");
            if ($status != 0 or $winclient_obj->error()) {
                my @errmsg = $winclient_obj->errmsg();
                $self->_set_err("iscsiShow failed: @errmsg");
                return $ClientUtil::FALSE;
            }

            #update the cache with the new data for this file
            $self->store_cache_data($cache_string, @output);
        }

        # match the target ip with the target nodename and the target index.
        my ($t_name, $t_index, $t_ip);
        foreach my $line (@output) {
            #
            # I'm expecting @output to look like this
            #
            # TargetName                              Index   IP Addr                    Port    Auth    DDigest HDigest MaxConn Type
            # ----------                              -----   -------                    ----    ----    ------- ------- ------- ----
            # iqn.1992-08.com.netapp:sn.33616586      0:0     192.168.1.61               3260    None    None    None    0       SendTargets
            # iqn.1992-08.com.netapp:sn.33616584      0:0     192.168.2.61               3260    None    None    None    0       SendTargets
            # iqn.1992-08.com.netapp:sn.33453267      0:0     192.168.3.61               3260    None    None    None    0       iSNS
            # iqn.1992-08.com.netapp:sn.33453279      0:0     192.168.4.61               3260    None    None    None    0       iSNS
            # iqn.1992-08.com.netapp:sn.84175766      0:1     [fe80::2a0:98ff:fe01:a390] 3260    None    None    None    0       SendTargets
            ($t_name, $t_index, $t_ip) = ($line =~ /^(\S+)\s+(\S+)\s+(\S+)/);
            $t_ip =~ s/\[|\]//g; # remove brackets from ipv6 addresses
            if ($t_ip eq $target_ip and
                $t_name eq $nodename) {
                # store the target index on the target object.
                # we will need this information when we need to login to the target
                $target->{index} = $t_index;
                last;
            } else {
                # skip this line
                next;
            }
        }

        # if there is no target index for this target then we were unsuccessful in adding it
        # discovering the target
        if (not defined($target->{index})) {
            $self->_set_err("No target was discovered for '$target_ip' on iscsi target '$nodename', " .
                            "please verify network connectivity between the initiator: '$initiator_ip' " .
                            "and the target: '$target_ip', or no target was found matching the expected " .
                            "name $nodename.");
            return $ClientUtil::FALSE;
        }

        # We now have to login to this target, There may already be logins(sessions) to this target,
        # so we need to take those into account when we add additional sessions

        # Sample output: @output
        #Session ID                         Connection ID                     SrcAddr        SrcPort DstAddr        DstPort
        #----------                         -------------                     -------        ------- -------        -------
        #fffffa8031410428-400001370000003f  fffffa8031410428-000000000000003e 192.168.1.105  62918   192.168.1.61   3260
        #fffffa8031410428-4000013700000051  fffffa8031410428-00000000000001f6 192.168.1.105  51654   192.168.1.62   3260
        #fffffa8031410428-4000013700000057  fffffa8031410428-00000000000001fd 192.168.1.105  62150   192.168.1.61   3260
        #fffffa8031410428-4000013700000059  fffffa8031410428-00000000000001ff 192.168.1.105  62662   192.168.1.61   3260
        #fffffa8031410428-400001370000005a  fffffa8031410428-0000000000000200 192.168.1.105  51910   192.168.1.62   3260
        #fffffa8031410428-400001370000005b  fffffa8031410428-0000000000000201 192.168.1.105  52166   192.168.1.62   3260
        #fffffa8031410428-400001370000005c  fffffa8031410428-0000000000000202 192.168.1.105  51142   192.168.1.62   3260
        #fffffa8031410428-400001370000005f  fffffa8031410428-000000000000062b 192.168.1.105  63174   192.168.1.61   3260
        #fffffa80312c2018-4000013700000015  fffffa80312c2018-0000000000000af6 fe80::d45f:f777:f28c:24e3  1995  fe80::2a0:98ff:fe01:70ab  3260
        ($status, @output) = $winclient_obj->iscsiShowConnections();
        if ($status != 0 or $winclient_obj->error()) {
            $self->_set_err("Could not obtain iSCSI connection information");
            return $ClientUtil::FALSE;
        }

        # only look at connections that are going to this target IP
        my (@connections_for_this_target) = grep { /\s$target_ip\s/ } @output;

        # build a hash of unique session id's, where the key is the session id and
        # value is the number of connections it has.
        my (%unique_sessions);
        foreach my $connection_line (@connections_for_this_target) {
            # parse out the session id from the connection line
            if ($connection_line =~ /^\s*(\S+)\s+/) {
                my $session_id = $1;
                $unique_sessions{$session_id}++;

                # if this session is doing both MCS and MPIO then this is a config error.
                # A session is doing MCS if it has more than one connection to it.
                # This is a redundant configuration because it is doing "multipathing on top of multipathing".
                if ($unique_sessions{$session_id} > 1 and
                    $self->{iscsi}->{CONFIG_ISCSI_WIN_MPIO} == 1) {
                    $self->_set_err("The MPIO enabled session ($session_id) to target ($target_ip) has more than one MCS connection to it. " .
                                    "This is a possible configuration error since it is essentially doing 'multipathing over multipathing'. " .
                                    "This is not a supported configuration.");
                    return $ClientUtil::FALSE;
                }
            }
        }

        # determine how many sessions this target has and figure out how many additional
        # logins we have to perform.
        my $num_sessions_for_this_target = scalar(keys %unique_sessions);
        my $max_num_logins = $self->{iscsi}->{CONFIG_ISCSI_WIN_NUM_SESSIONS};
        my $num_logins_to_make = $max_num_logins - $num_sessions_for_this_target;

        # call iscsiLogin $num_logins_to_make times to login the initiator into the target
        for (my $i = 1; $i <= $num_logins_to_make; $i++) {
            ($status, @output) =
                $winclient_obj->iscsiLogin(-targetname     => $nodename,
                                           -targetindex    => $target->{index},
                                           -initiatorindex => $initiator_index,
                                           -mpio           => $self->{iscsi}->{CONFIG_ISCSI_WIN_MPIO},
                                           -datadigest     => $self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST},
                                           -headerdigest   => $self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST},
                                           -persistent     => $self->{iscsi}->{CONFIG_ISCSI_WIN_PERSISTENT},
                                           -maxconnections => $self->{iscsi}->{CONFIG_ISCSI_WIN_MAX_CONNECTIONS},
                                           -auth_type      => $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE},
                                           -chap_username  => $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME},
                                           -chap_password  => $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD},
                                           -default_time2wait => $self->{iscsi}->{CONFIG_ISCSI_DEFAULT_TIME2WAIT},
                                           -default_time2retain => $self->{iscsi}->{CONFIG_ISCSI_DEFAULT_TIME2RETAIN});
            if ($status != 0 or $winclient_obj->error()) {
                my @errmsg = $winclient_obj->errmsg();
                $self->_set_err("iscsiLogin (login number $i) failed: @errmsg");
                return $ClientUtil::FALSE;
            }

            # parse out the sessionID of the login we just made and increment the number of connections for that session
            # in the unique sessions
            foreach my $line (@output) {
                if ($line =~ /SessionID:\s+(\S+)/i) {
                    my $session_id = $1;
                    $unique_sessions{$session_id}++;
                    last;
                }
            }
        }

        # if the num logins we were supposed to make is negative that means there are too many logins.
        if ($num_logins_to_make < 0) {
            $self->_set_err("Initiator ($initiator_ip) has more than $max_num_logins session(s) to target ($target_ip).");
            return $ClientUtil::FALSE;
        } else {
            $self->sanlog(-msg => "iSCSI Configure: Initiator ($initiator_ip) " .
                                  "has $max_num_logins session(s) to target ($target_ip)");
        }

        # if requested to do MCS (multi connected sessions), do so now
        if ($self->{iscsi}->{CONFIG_ISCSI_WIN_MCS_NUM_CONNS_PER_SESSION} > 1) {
            # for every unique session, determine how many additional connections
            # we have to make and add them
            foreach my $session_id (keys %unique_sessions) {
                my $num_current_connections = $unique_sessions{$session_id};
                my $total_connections_requested = $self->{iscsi}->{CONFIG_ISCSI_WIN_MCS_NUM_CONNS_PER_SESSION};
                my $num_connections_to_make = $total_connections_requested - $num_current_connections;

                # create additional connections as needed
                for (my $i = 0; $i < $num_connections_to_make; $i++) {
                    ($status, @output) =
                        $winclient_obj->iscsiAddConnection(-sessionid => $session_id,
                                                           -targetindex => $target->{index},
                                                           -initiatorindex => $initiator_index,
                                                           -mpio => $self->{iscsi}->{CONFIG_ISCSI_WIN_MPIO},
                                                           -datadigest => $self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST},
                                                           -headerdigest => $self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST},
                                                           -persistent => $self->{iscsi}->{CONFIG_ISCSI_WIN_PERSISTENT},
                                                           -maxconnections => $self->{iscsi}->{CONFIG_ISCSI_WIN_MAX_CONNECTIONS},
                                                           -auth_type =>  $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE},
                                                           -chap_username => $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME},
                                                           -chap_password => $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD},
                                                           -default_time2wait => $self->{iscsi}->{CONFIG_ISCSI_DEFAULT_TIME2WAIT},
                                                           -default_time2retain => $self->{iscsi}->{CONFIG_ISCSI_DEFAULT_TIME2RETAIN});
                    if ($status != 0 or $winclient_obj->error()) {
                        my @errmsg = $winclient_obj->errmsg();
                        $self->_set_err("iscsiAddConnection (connection number $i) failed: @errmsg");
                        return $ClientUtil::FALSE;
                    }
                }

                # if there are too many connections issue a warning
                if ($num_connections_to_make < 0) {
                    $self->sanlog(-msg => "iSCSI Configure: WARNING: Session ($session_id) " .
                                          "has more than $total_connections_requested connections to target ($target_ip).");
                } else {
                    $self->sanlog(-msg => "iSCSI Configure: Session ($session_id) " .
                                          "has $total_connections_requested connections to target ($target_ip)");
                }
            }
        }
    }

    # display the final list of iSCSI connections
    ($status, @output) = $winclient_obj->iscsiShowConnections();
    if ($status != 0 or $winclient_obj->error()) {
        # this is informative information, clear the error
        my @errmsg = $winclient_obj->errmsg();
        $self->_set_err("Could not show iSCSI Connections: @errmsg");
        return $ClientUtil::FALSE;
    }

    $self->invalidate_items_cache_data($cache_string); # disable cache at end

    $self->sanlog(-msg => "iSCSI Configure: iSCSI Configuration Complete.");
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_validate_params()
# Objective:       validate all the iscsi parameters
################################################################################
sub _iscsi_configure_validate_params {
    my ($self, %args) = @_;

    # if the user specified to login more than 32 times, fail out
    if ($self->{iscsi}->{CONFIG_ISCSI_WIN_NUM_SESSIONS} > 32) {
        $self->_set_err("User requested for this client to login to each iSCSI target " . $self->{iscsi}->{CONFIG_ISCSI_WIN_NUM_SESSIONS} .
                        " times. The max is 32.\nSet CONFIG_ISCSI_WIN_NUM_SESSIONS to a value less than or equal to 32");
        return $ClientUtil::FALSE;
    }

    # validate that MPIO is enabled if doing more than one session,
    # fail out if not doing mpio and sessions is > 1
    if ($self->{iscsi}->{CONFIG_ISCSI_WIN_MPIO} == 0 and
        $self->{iscsi}->{CONFIG_ISCSI_WIN_NUM_SESSIONS} > 1) {
        $self->_set_err("Requested to not do MPIO but CONFIG_ISCSI_WIN_NUM_SESSIONS is greater than 1.");
        return $ClientUtil::FALSE;
    }

    # validate that MPIO and MCS are not both enabled
    if ($self->{iscsi}->{CONFIG_ISCSI_WIN_MPIO} == 1 and
        $self->{iscsi}->{CONFIG_ISCSI_WIN_MCS_NUM_CONNS_PER_SESSION} != 1) {
        $self->_set_err("Requested to do MPIO but CONFIG_ISCSI_WIN_MCS_NUM_CONNS_PER_SESSION is not set to 1");
        return $ClientUtil::FALSE;
    }

    # MCS is not supported in automation yet. Remove once support has been added.
    if ($self->{iscsi}->{CONFIG_ISCSI_WIN_MCS_NUM_CONNS_PER_SESSION} > 1) {
        $self->_set_err("MCS (Multi-Connected Sessions) is not yet supported in the automation. " .
                        "Set CONFIG_ISCSI_WIN_MCS_NUM_CONNS_PER_SESSION to 1.");
        return $ClientUtil::FALSE;
    }

    # validate that if isns is the discovery type, that an isns server was specified
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^isns$/i and
        $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER} eq "") {
        $self->_set_err("Requested to do iSNS discovery but CONFIG_ISCSI_ISNS_SERVER was not specified.");
        return $ClientUtil::FALSE;
    }

    # validate that if doing bidirectional CHAP, a mutual secret must be specified
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD} eq "" and
        $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /bidirectional/i) {
        $self->_set_err("Requested to do bidirectional CHAP but CONFIG_ISCSI_CHAP_OUT_PASSWORD was not specified");
        return $ClientUtil::FALSE;
    }

    # hardware iscsi is not supported in the automation
    if ($self->{iscsi}->{CONFIG_ISCSI_TYPE} !~ /software/i) {
        $self->_set_err("CONFIG_ISCSI_TYPE '" . $self->{iscsi}->{CONFIG_ISCSI_TYPE} .
                        "' is not supported, use 'software'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_print_summary()
# Objective:       print out a summary of what configuration will be done
################################################################################
sub _iscsi_configure_print_summary {
    my ($self, %args) = @_;

    my (@targets);

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => @targets");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # print out a summary of what we are going to do for iSCSI Configuration
    my @target_ips = map { "\n\t" . $_->{ip} } @targets;
    my @config_summary;
    push(@config_summary, "iSCSI Configure: Configuring iSCSI with the following options:\n");
    push(@config_summary, "Targets : @target_ips\n");
    push(@config_summary, "IP Version            : " . $self->{iscsi}->{CONFIG_ISCSI_IPVERSION} . "\n");
    push(@config_summary, "iSCSI Type            : " . $self->{iscsi}->{CONFIG_ISCSI_TYPE} . "\n");
    push(@config_summary, "iSCSI Discovery Type  : " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} . "\n");
    if ($self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_TYPE} =~ /^isns$/i) {
        push(@config_summary, "iSNS Server           : " . $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER} . "\n");
    }
    push(@config_summary, "iSCSI Pre-Clean Level : " . $self->{iscsi}->{CONFIG_ISCSI_DISCOVERY_PRECLEANUP} . "\n");
    push(@config_summary, "iSCSI Subnet Matching : " . $self->{iscsi}->{CONFIG_ISCSI_SUBNET_MATCHING} . "\n");
    push(@config_summary, "iSCSI Use Mgmt Port   : " . $self->{iscsi}->{CONFIG_ISCSI_USE_MGMT_PORT} . "\n");
    push(@config_summary, "CHAP Mode             : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} . "\n");
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /bidirectional/i) {
        push(@config_summary, "CHAP Mutual Secret    : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_OUT_PASSWORD} . "\n");
    }
    if ($self->{iscsi}->{CONFIG_ISCSI_CHAP_MODE} =~ /(uni|bi)directional/i) {
        push(@config_summary, "CHAP Username         : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_USERNAME} . "\n");
        push(@config_summary, "CHAP Password         : " . $self->{iscsi}->{CONFIG_ISCSI_CHAP_IN_PASSWORD} . "\n");
    }
    push(@config_summary, "Header Digest         : " . $self->{iscsi}->{CONFIG_ISCSI_HEADER_DIGEST} . "\n");
    push(@config_summary, "Data Digest           : " . $self->{iscsi}->{CONFIG_ISCSI_DATA_DIGEST} . "\n");
    push(@config_summary, "Num Sessions (Logins) : " . $self->{iscsi}->{CONFIG_ISCSI_WIN_NUM_SESSIONS} . "\n");
    push(@config_summary, "MPIO                  : " . $self->{iscsi}->{CONFIG_ISCSI_WIN_MPIO} . "\n");
    push(@config_summary, "Persistent Sessions   : " . $self->{iscsi}->{CONFIG_ISCSI_WIN_PERSISTENT} . "\n");
    push(@config_summary, "Maximum Connections   : " . $self->{iscsi}->{CONFIG_ISCSI_WIN_MAX_CONNECTIONS} . "\n");
    push(@config_summary, "DefaultTime2Wait      : " . $self->{iscsi}->{CONFIG_ISCSI_DEFAULT_TIME2WAIT} . "\n");
    push(@config_summary, "DefaultTime2Retain    : " . $self->{iscsi}->{CONFIG_ISCSI_DEFAULT_TIME2RETAIN} . "\n");

    $self->sanlog(-msg => "@config_summary");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_configure_build_initiator_hash()
# Objective:       build out a hash of intiator port information for the host
################################################################################
sub _iscsi_configure_build_initiator_hash {
    my ($self) = @_;

    my ($init_node_name);
    my $winclient_obj = $self->{WinClient};

    # obtain the ip information for this host
    my ($interface_hsh) = $self->get_ip_info();
    if ($self->error()) {
        return ({}, "");
    }

    # obtain the initiator adapter information
    my ($adapter_status, @adapter_output) = $winclient_obj->iscsiShowAdapters();
    if ($adapter_status != 0 or $winclient_obj->error()) {
        my @errmsg = $winclient_obj->errmsg();
        $self->_set_err("iscsiShow adapters failed: @errmsg");
        return ({}, "");
    }

    # mpc iscsi show adapters
    #
    # iSCSINodeName: iqn.1991-05.com.microsoft:ibm3550-rtp14.rtp.netapp.com
    #
    # AdapterName     Index   Tag     Type    Address
    # -----------     -----   ---     ----    -------
    # MS-iSCSI-0      0       0       IPv4    10.60.242.119
    # MS-iSCSI-1      1       0       IPv4    192.168.1.189
    #
    # Success: iscsi show adapters

    # store the adapter info in this hash:
    my $initiator_hsh = {};
    foreach my $line (@adapter_output) {
        next if ($line =~ /AdapterName/i)
             or ($line =~ /----/)
             or ($line =~ /^\s*$/)
             or ($line =~ /Success/i)
             or ($line =~ /rstatus/i);

        # extract the initiator nodename
        if ($line =~ /iSCSINodeName: (\S+)/) {
            $init_node_name = $1;
            next;
        }

        # pull out the initiator index and initiator ip
        my ($adaptername, $init_index, $init_tag, $ip_version, $init_ip, $init_ipv6_full) = split(/\s+/, $line);

        # get rid of characters after % because they are not part of the ipv6 address
        # (this won't affect ipv4 addresses)
        $init_ip =~ s/\%.+$//;
        # store the initiator information in the initiator hash for this adapter
        foreach my $key (keys %$interface_hsh) {
            if ($self->{iscsi}->{CONFIG_ISCSI_USE_MGMT_PORT} == 1) {
                # if requested to only use the mgmt port, then next if its not the mgmt port
                next if ($interface_hsh->{$key}->{is_mgmt_interface} != 1);
            } else {
                # skip the management interface because it should never have SAN traffic running over it
                next if ($interface_hsh->{$key}->{is_mgmt_interface} == 1);
            }

            if ($ip_version =~ /IPv4/i and
                $interface_hsh->{$key}->{ipv4} eq $init_ip) {
                # handle ipv4 initiators

                # if we could not determine the ipv4 subnet of this initiator ip then error out
                if ($interface_hsh->{$key}->{ipv4_subnet} eq "") {
                    $self->_set_err("Could not determine the subnet for IPv4 initiator address: '$init_ip'");
                    return ({}, "");
                }

                # store the details for this initiator ip in the hash
                $initiator_hsh->{$init_ip}->{ip} = $init_ip;
                $initiator_hsh->{$init_ip}->{subnet} = $interface_hsh->{$key}->{ipv4_subnet};
                $initiator_hsh->{$init_ip}->{index} = $init_index;
                $initiator_hsh->{$init_ip}->{ip_version} = $ip_version;
                next;
            } elsif ($ip_version =~ /IPv6/i and length $interface_hsh->{$key}->{ipv6}) {
                # remove the scope from the address
                $init_ipv6_full =~ s/\%.+$//;

                # handle ipv6 initiators
                my $o1 =  new Net::IP($interface_hsh->{$key}->{ipv6});
                my $o2 = new Net::IP($init_ipv6_full);

                # compare the binary representation of the two ipv6 addresses
                if ($o1->binip() eq $o2->binip()) {
                    # if we could not determine the subnet for the ipv6 address, then error out
                    if ($interface_hsh->{$key}->{ipv6_subnet} eq "") {
                        $self->_set_err("Could not determine the subnet for IPv6 initiator address: '$init_ip'");
                        return ({}, "");
                    }

                    # store the details for this initiator ip in the hash
                    $initiator_hsh->{$interface_hsh->{$key}->{ipv6}}->{ip} = $interface_hsh->{$key}->{ipv6};
                    $initiator_hsh->{$interface_hsh->{$key}->{ipv6}}->{subnet} = $interface_hsh->{$key}->{ipv6_subnet};
                    $initiator_hsh->{$interface_hsh->{$key}->{ipv6}}->{index} = $init_index;
                    $initiator_hsh->{$interface_hsh->{$key}->{ipv6}}->{ip_version} = $ip_version;
                    next;
                }
            }

        }
    }

    # if there are no initiators in the initiator_hsh, then this is an error
    if (scalar(keys %{$initiator_hsh}) == 0) {
        $self->_set_err("There are no available initiator IP addresses to use to create " .
                        "iSCSI sessions with (not including the management port)");
        return ({}, "");
    }

    # return the initiator hash and the nodename of this initiator
    return ($initiator_hsh, $init_node_name);
}

################################################################################
# Subroutine Name: _iscsi_configure_isns_discovery()
# Objective:       Adds the isns server to the host.
################################################################################
sub _iscsi_configure_isns_discovery() {
    my ($self) = @_;

    my $isns_server = $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER};
    my $winclient_obj = $self->{WinClient};

    # add the isns server
    my $result = $winclient_obj->iscsiAddiSNSServer(-ip_address => $isns_server);
    if ($result eq 0 or $winclient_obj->error()) {
        my @errmsg = $winclient_obj->errmsg();
        $self->_set_err("iscsiAddiSNSServer failed: @errmsg");
        return $ClientUtil::FALSE;
    }

    # display the results
    my ($status, @output) = $winclient_obj->iscsiShow(-object => "isnsservers");
    if ($status != 0 or $winclient_obj->error()) {
        my @errmsg = $winclient_obj->errmsg();
        $self->_set_err("iscsiShow failed: @errmsg");
        return $ClientUtil::FALSE;
    }

    # verify that the isns server was added properly.
    if (not grep(/^iSNSServerAddress=${isns_server}\s*$/i, @output) and
        not grep(/^\s*${isns_server}\s*$/, @output)) {
        $self->_set_err("iSNS Server ($isns_server) was not added properly.");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_deconfigure()
# Objective:       Deconfigures iscsi
# Details:         See POD documentation at the end of this file
################################################################################
sub _iscsi_deconfigure {
    my ($self, %args) = @_;

    my (@targets, $cmd, @output, $status, @targets_output,
        @isns_servers, @isns_servers_to_remove);

    $self->_clear_err();

    my $clean_level = 1;
    my $winclient_obj = $self->{WinClient};

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?clean_level/) {
            if (not length($clean_level = $args{$_})) {
                $self->_set_err("Invalid value for argument '-clean_level' => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    # check for required parameters
    if ($clean_level < 2 and not scalar(@targets)) {
        $self->_set_err("Missing or invalid \'targets\' argument");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: Deconfiguring iSCSI (clean_level = $clean_level)");

    # store the targets table in @targets_output for reference later on
    # Example output:
    #TargetName                                      Index   IP Addr                         Port    Auth    DDigest HDigest MaxConn Type
    #----------                                      -----   -------                         ----    ----    ------- ------- ------- ----
    #iqn.1992-08.com.netapp:sn.84175766              0:0     192.168.1.62                    3260    None    None    None    0       SendTargets
    #iqn.1992-08.com.netapp:sn.84175766              0:1     [fe80::2a0:98ff:fe01:a390]      3260    None    None    None    0       SendTargets
    #iqn.1992-08.com.netapp:sn.84174947              0:0     192.168.1.61                    3260    None    None    None    0       iSNS
    #iqn.1992-08.com.netapp:sn.84174947              0:1     [fe80::2a0:98ff:fe01:70ab]      3260    None    None    None    0       iSNS

    ($status, @targets_output) = $winclient_obj->iscsiShow(-object => "targets");
    if ($status != 0 or $winclient_obj->error()) {
        my @errmsg = $winclient_obj->errmsg();
        $self->_set_err("iscsiShow failed: @errmsg");
        return $ClientUtil::FALSE;
    }

    # obtain the list of isns servers that are currently on the system
    ($status, @isns_servers) = $winclient_obj->iscsiShow(-object => "isnsservers");
    if ($status != 0 or $winclient_obj->error()) {
        my @errmsg = $winclient_obj->errmsg();
        $self->_set_err("iscsiShow failed: @errmsg");
        return $ClientUtil::FALSE;
    }

    # parse out the addresses of the isns servers and push on the isns servers array
    # we will walk through this array and remove them later on.
    if ($winclient_obj->{hostver} =~ /windows server 7|2008/i) {
        # output is in the WMIC format
        @isns_servers = grep { /iSNSServerAddress=/ } @isns_servers;
        @isns_servers = map {
            if ($_ =~ /iSNSServerAddress=(\S+)/) {
                $1;
            }
        } @isns_servers;
    } else {
        # output is in the MPC format
        @isns_servers = grep {/\d+\.\d+\.\d+\.\d+/} @isns_servers;
    }

    # if clean level is '2' then we need to clean up everything,
    if ($clean_level == 2) {
        # modify the targets array so that all targets present on the system are in the array.
        @targets = ();
        foreach my $line (@targets_output) {
            # parse out the IP and store in the targets array
            if ($line =~ /iqn\S+\s+\S+\s+(\S+)/) {
                my $target = {};
                $target->{ip} = $1;
                # get rid of brackets (IPv6 addresses)
                $target->{ip} =~ s/\[|\]//g;
                push(@targets, $target);
            } else {
                next;
            }
        }

        # need to remove all isns servers if clean level == 2
        @isns_servers_to_remove = @isns_servers;
    }

    # print out which targets we will be removing
    my (@target_ips) = map { $_->{ip} . "\n" } @targets;
    if (scalar(@target_ips)) {
        $self->sanlog(-msg => "iSCSI Deconfigure: Removing the following target portals:\n @target_ips");
    }

    # iterate over all the targets
    foreach my $target (@targets) {

        # logout of this target
        $self->_iscsi_logout_target(-target => $target);
        if ($self->error()) {
            return $ClientUtil::FALSE;
        }

        my $t_ip = $target->{ip};

        # retain the target portals if requested to not clean them
        if ($self->{iscsi}->{CONFIG_ISCSI_WIN_KEEP_TARGET_PORTALS} == 1) {
            next;
        }

        # don't remove the target portal if it has a discovery type of iSNS
        # removing the isns server will handle that for us.
        if (grep(/\s\[?$t_ip\]?\s.+?iSNS/i, @targets_output)) {
            next;
        }

        # at this point, all sessions have been logged out, remove this targetportal
        ($status, @output) = $winclient_obj->iscsiRemoveTargetPortal(-filer => $t_ip);
        if ($status != 0 or $winclient_obj->error()) {
            $self->_set_err("Could not remove iSCSI targetportal '$t_ip'");
            return $ClientUtil::FALSE;
        }
    }

    if (not scalar(@isns_servers_to_remove) and $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER} ne "") {
        push(@isns_servers_to_remove, $self->{iscsi}->{CONFIG_ISCSI_ISNS_SERVER});
    }

    # remove the isns servers from the list if there are any
    if (scalar(@isns_servers_to_remove) and scalar(@isns_servers)) {
        $self->sanlog(-msg => "iSCSI Deconfigure: Removing the following iSNS Servers:\n @isns_servers_to_remove");
        foreach my $isns_server (@isns_servers_to_remove) {
            # remove this isns server if it exists on the system
            if (grep(/^$isns_server$/, @isns_servers)) {
                ($status, @output) = $winclient_obj->iscsiRemoveiSNSServer(-ip_address => $isns_server);
                if ($status != 0 or $winclient_obj->error()) {
                    $self->_set_err("Could not remove iSNS Server '$isns_server'");
                    return $ClientUtil::FALSE;
                }
            }
        }
        # if isns servers must be removed, you have to run mpc show
        # targets to refresh the target list after they are removed
        ($status, @output) = $winclient_obj->iscsiShow(-object => "targets");
        if ($status != 0 or $winclient_obj->error()) {
            $self->_set_err("Could not refresh iscsi target list after removing" .
                            " iSNS servers.");
            return $ClientUtil::FALSE;
        }
    }

    if ($clean_level == 1) {
        $self->_iscsi_orphaned_sessions_cleanup(-targets => \@targets);
        if ($self->error()) {
            $self->_set_err("Failed to cleanup orphaned iSCSI sessions", $self->errmsg());
            return $ClientUtil::FALSE;
        }
    } elsif ($clean_level == 2) {

        # if clean level is '2' then we need to clean up everything

        $self->_iscsi_orphaned_sessions_cleanup();
        if ($self->error()) {
            $self->_set_err("Failed to cleanup orphaned iSCSI sessions", $self->errmsg());
            return $ClientUtil::FALSE;
        }

        # retain the target portals if requested to not clean them
        unless ($self->{iscsi}->{CONFIG_ISCSI_WIN_KEEP_TARGET_PORTALS} == 1) {
            $self->_iscsi_orphaned_targetportals_cleanup();
            if ($self->error()) {
                $self->_set_err("Failed to cleanup orphaned iSCSI targetportals", $self->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    }

    $self->sanlog(-msg => "iSCSI Deconfigure: iSCSI has been deconfigured.");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_logout_target()
# Objective:       Log out from the specfied iSCSI target
################################################################################
sub _iscsi_logout_target {
    my ($self, %args) = @_;

    my $target = exists($args{-target}) ? $args{-target} : undef;
    if (not defined($target)) {
        $self->_set_err("Missing required parameter -target");
        return $ClientUtil::FALSE;
    }

    my $winclient_obj = $self->{WinClient};
    my $t_ip = $target->{ip};

    # foreach target, remove all connections for it

    # Sample output: @output
    #Session ID                         Connection ID                     SrcAddr        SrcPort DstAddr        DstPort
    #----------                         -------------                     -------        ------- -------        -------
    #fffffa8031410428-400001370000003f  fffffa8031410428-000000000000003e 192.168.1.105  62918   192.168.1.61   3260
    #fffffa8031410428-4000013700000051  fffffa8031410428-00000000000001f6 192.168.1.105  51654   192.168.1.62   3260
    #fffffa8031410428-4000013700000057  fffffa8031410428-00000000000001fd 192.168.1.105  62150   192.168.1.61   3260
    #fffffa8031410428-4000013700000059  fffffa8031410428-00000000000001ff 192.168.1.105  62662   192.168.1.61   3260
    #fffffa8031410428-400001370000005a  fffffa8031410428-0000000000000200 192.168.1.105  51910   192.168.1.62   3260
    #fffffa8031410428-400001370000005b  fffffa8031410428-0000000000000201 192.168.1.105  52166   192.168.1.62   3260
    #fffffa8031410428-400001370000005c  fffffa8031410428-0000000000000202 192.168.1.105  51142   192.168.1.62   3260
    #fffffa8031410428-400001370000005f  fffffa8031410428-000000000000062b 192.168.1.105  63174   192.168.1.61   3260

    my ($status, @output) = $winclient_obj->iscsiShowConnections();
    if ($status != 0 or $winclient_obj->error()) {
        $self->_set_err("Could not obtain iSCSI connection information");
        return $ClientUtil::FALSE;
    }

    # only look at connections that are going to this target IP
    my (@connections_for_this_target) = grep { /\s$t_ip\s/ } @output;

    # if there are connections to this target, remove them and logout from the target.
    if (scalar(@connections_for_this_target)) {
        my $session_hash;
        foreach my $connection_line (@connections_for_this_target) {
            # parse out the session and connection id for this connection
            my ($session_id, $connection_id) = ($connection_line =~ /^(\S+)\s+(\S+)/);
            push(@{$session_hash->{$session_id}}, $connection_id);
        }

        foreach my $session (keys %{$session_hash}) {
            # if there is more than one connection to this session, then
            # remove all the connections but the first one
            if (scalar(@{$session_hash->{$session}}) > 1) {
                # first sort the array of connection id's because the leading connection must be at the beginning
                # of the array
                @{$session_hash->{$session}} = sort(@{$session_hash->{$session}});

                # remove the leading connection from the array list
                shift(@{$session_hash->{$session}});

                # close all the connections but the leading connection
                foreach my $connection (@{$session_hash->{$session}}) {
                    ($status, @output) = $winclient_obj->iscsiRemoveConnection(-connectionid => $connection,
                                                                               -retry_count => 3);
                    if ($status != 0 or $winclient_obj->error()) {
                        $self->_set_err("Could not remove iSCSI connection '$connection' for target ($t_ip)");
                        return $ClientUtil::FALSE;
                    }
                }
            }

            # after removing all connections, logout from this session
            ($status, @output) = $winclient_obj->iscsiLogout(-sessionid => $session,
                                                             -retry_count => 3);
            if ($status != 0 or $winclient_obj->error()) {
                $self->_set_err("Could not logout from iSCSI session '$session' for target ($t_ip)");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_orphaned_sessions_cleanup()
# Objective:       Cleanup sessions which do not have any connections
################################################################################
sub _iscsi_orphaned_sessions_cleanup {
    my ($self, %args) = @_;

    my (@targets);

    $self->_clear_err();

    my $winclient_obj = $self->{WinClient};

    # parse arguments
    foreach (keys %args) {
        if (/-?targets/) {
            if (not scalar(@targets = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-targets' => $args{$_}");
                return $ClientUtil::FALSE;
            }
        }
    }

    # Get all iSCSI sessions
    my ($status, @output) = $winclient_obj->iscsiShow(-object => "sessions");
    if ($status != 0 or $winclient_obj->error()) {
        my @errmsg = $winclient_obj->errmsg();
        $self->_set_err("iscsiShow failed: @errmsg");
        return $ClientUtil::TRUE;
    }

    # Get all the target nodenames
    my @target_nodenames;
    if (scalar(@targets)) {
        foreach my $target (@targets) {
            my $target_nodename = $target->get_nodename();
            if ($target->error()) {
                $self->_set_err("Could not get nodename of target: " . $target->errmsg());
                return $ClientUtil::FALSE;
            }
            push (@target_nodenames, $target_nodename);
        }
    }

    # Look for session IDs.
    # mpc iscsi show sessions has the following output. Need to parse that
    # for session ids
    #
    #
    # Session ID              Target Name             # Conn  # Devs  TSID    ISID
    # ----------              -----------             ------  ------  ----    ----
    # fffffa8002bc4438-40000137000000a7   iqn.1992-08.com.netapp:sn.386dfe68fd6e11df8498123478563412:vs.3 0   7f00    400001370000
    #
    # Success: iscsi show sessions
    #
    my @sessions;
    foreach my $line (@output) {
        # Convert all occurances of multiple white space into a single space
        $line =~ s/\s+/ /g;
        next if ($line =~ /----/)
                or ($line =~ /^\s*$/)
                or ($line =~ /Success/i)
                or ($line =~ /rstatus/i)
                or ($line =~ /Session/i);
        my @words = split (/\s/, $line);

        if (scalar(@targets)) {
            push (@sessions, $words[0]) if (grep(/^$words[1]$/, @target_nodenames));
        } else {
            # Session ID will be the first word in the line. Push all of them
            push (@sessions, $words[0]);
        }
    }

    # Return if no sessions found
    unless (scalar(@sessions)) {
        $self->sanlog(-msg => "No sessions found. Skipping orphaned sessions cleanup", -msg_level => 1);
        return $ClientUtil::TRUE;
    }

    $self->sanlog(-msg => "Cleaning orphaned sessions", -msg_level => 1);
    foreach my $session (@sessions) {

        # Do iscsi logout to remove the session from the initiator
        ($status, @output) = $winclient_obj->iscsiLogout(-sessionid => $session, -retry_count => 3);
        if ($status != 0 or $winclient_obj->error()) {
            my @errmsg = $winclient_obj->errmsg();
            $self->_set_err("iscsiLogout failed: @errmsg");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _iscsi_orphaned_targetportals_cleanup()
# Objective:       Cleanup targetportals that are no longer needed
################################################################################
sub _iscsi_orphaned_targetportals_cleanup {
    my ($self) = @_;

    $self->_clear_err();

    my $winclient_obj = $self->{WinClient};

    # call iscsicli listtargetportals to get target IPs
    my $cmd = "$winclient_obj->{cmd_iscsicli} listtargetportals";

    # execute the command
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0 or not grep(/The operation completed successfully./i, @output)) {
        $self->_set_err("$self->{cmd_iscsicli} listtargetportals failed.");
        return $ClientUtil::FALSE;
    }

    my @targetportals;
    foreach my $line (@output) {
        #iscsicli listtargetportals
        #Microsoft iSCSI Initiator Version 6.0 Build 6000
        #
        #Total of 1 portals are persisted:
        #
        #    Address and Socket   : 1.2.3.4 3260
        #        Symbolic Name        :
        #            Initiator Name       : Root\ISCSIPRT\0000_0
        #                Port Number          : <Any Port>
        #                ??Security Flags        : 0x0
        #                ??Version              : 0
        #                ??Information Specified: 0x1c
        #                ??Maximum Connections  : 32
        #                ??Default Time 2 Wait  : 0
        #                ??Default Time 2 Retain: 20
        #                ??Login Flags          : 0x0
        #
        #                The operation completed successfully.
        if ($line =~ /Address and Socket/i) {
            # remove leading white space from line
            $line =~ s/^\s+//;

            # Convert all occurances of multiple white space into a single space
            $line =~ s/\s+/ /g;
            my @words = split (/\s/, $line);
            push (@targetportals, $words[4]);
        }
    }

    unless (scalar(@targetportals)) {
        $self->sanlog(-msg => "No targetportals found. Skipping orphaned targetportals cleanup");
        return $ClientUtil::TRUE;
    }

    $self->sanlog(-msg => "Cleaning orphaned targetportals", -msg_level => 1);
    foreach my $t_ip (@targetportals) {
        ($status, @output) = $winclient_obj->iscsiRemoveTargetPortal(-filer => $t_ip);
        if ($status != 0 or $winclient_obj->error()) {
            $self->_set_err("Could not remove iSCSI targetportal '$t_ip'");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _partition()
# Objective:       Partition individual disk objects by calling the WinClient
#                  Disk::partition() subroutine.
# Details:         See POD documentation at the end of this file
################################################################################
sub _partition {
    my ($self, $diskObj) = @_;

    $diskObj->partition();
    if ($diskObj->error) {
        my @tempmsg = $diskObj->errmsg();
        $self->_set_err("Error: WinHost::_partition() error from " .
                        "WinClient::Disk::partition(): @tempmsg");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _set_lbpolicy()
# Objective:       set the load balancing policy for WinClient disk objects.
# Details:         See POD documentation at the end of this file
################################################################################
sub _set_lbpolicy {
    my ($self, $diskObj) = @_;

    # need to support the old parameter CONFIG_ISCSI_WIN_LBPOLICY
    # if the old parameter is set...
    if (defined($self->{iscsi}->{CONFIG_ISCSI_WIN_LBPOLICY}) and
        $self->{iscsi}->{CONFIG_ISCSI_WIN_LBPOLICY} =~ /^[12345678]$/) {
        # and the new parameter is not set to a value...
        if (not defined $self->{general}->{CONFIG_WIN_LBPOLICY} or
            $self->{general}->{CONFIG_WIN_LBPOLICY} !~ /^[1234568]$/) {

            # store the value of the old deprecated parameter in the new parameter
            $self->{general}->{CONFIG_WIN_LBPOLICY} = $self->{iscsi}->{CONFIG_ISCSI_WIN_LBPOLICY};
        }
    }

    # if a load balancing policy was requested, then set it on this disk.
    if (defined $self->{general}->{CONFIG_WIN_LBPOLICY} and
        $self->{general}->{CONFIG_WIN_LBPOLICY} =~ /^[1234568]$/) {
        $diskObj->setLBPolicy(-lbpolicy => $self->{general}->{CONFIG_WIN_LBPOLICY});
        if ($diskObj->error) {
            my @tempmsg = $diskObj->errmsg();
            $self->_set_err("Error from WinClient::Disk::setLBPolicy(): @tempmsg");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    _WinClient_getDevices()
# Objective: function to invoke WinClient::getDevices()
################################################################################
sub _WinClient_getDevices {
    my ($self, %args) = @_;

    my $refresh = 0;
    my $r_timeout = -1;
    my $num_expected_luns = undef;
    foreach (keys %args) {
        if (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Invalid value for argument '-refresh'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?rescan_timeout/) {
            if (not length($r_timeout = $args{$_})) {
                $self->_set_err("Invalid value for argument '-rescan_timeout'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?num_expected_luns/) {
            $num_expected_luns = $args{$_};
        }
    }

    # Look to see if WinClient's Disk Cache is defined and use if defined
    # If refresh is set, always call, which will refresh the cache.
    if (not defined($self->{WinClient}->{disks}) or $refresh == 1) {
        if ($r_timeout > -1) {
            $self->{WinClient}->getDevices($num_expected_luns, undef, $r_timeout);
        } else {
            $self->{WinClient}->getDevices($num_expected_luns);
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _configure_disk_san_policy()
# Objective:       configure the disk san policy
##############################################################################
sub _configure_disk_san_policy {
    my ($self) = @_;

    # burt490057, for win2k8 hosts, set SAN Policy to OnlineAll
    if ($self->{WinClient}->{hostver} =~ /2008/ or
        $self->{WinClient}->{hostver} =~ /windows server 7/) {
        my $disk_obj = WinClient::Disk->new(-host => $self->{WinClient});
        my ($status, @output) = $disk_obj->_diskpartCmd(commands => [("SAN policy=OnlineAll","exit")],
                                                        comment => "Setting SAN Policy to 'OnlineAll'");
        if ($status != 0) {
            $self->_set_err("Failed to set SAN Policy to OnlineAll");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

1;


=pod

=head1 Package Name

WinHost.pm

=head1 Author

David Pressley

=head1 Description

For description of methods see B<San::SanClient::Host.pod>.

=head1 Public Methods

=head2 B<get_iscsi_info()>

Collects all known iscsi information from the client.  Packages the information
into a complex data structure.

 Builds a hash of target nodenames.
 Under each target nodename, there is a hash for each iscsi cmd:
  alias      - contains scalar value (Can't get for windows)
  connection - contains an array of hashes
  initiator  - contains an array of hashes
  interface  - contains an array of hashes
  nodename   - contains scalar
  security   - contains a hash
  session    - contains an array of hashes
  show       - contains a hash
  status     - contains a hash
  tpgroup    - contains an array of hashes
 Each hash contains keys which correspond to the c-mode cmd options

=over 4

=item Inputs

=item * None

=item Outputs

=item * $iscsi_hash - hash of client iscsi info

=item * empty hash (do not set_err)  - no sessions or target portals

=item * empty hash (set_err) - error condition

=item * Sample output:

 $VAR1 = {
           'iqn.1992-08.com.netapp:sn.6dc5e7aff25311dfad2d123478563412:vs.6' => {
                                                                                  'connection' => [
                                                                                                    {
                                                                                                      'has-session' => 'true',
                                                                                                      'header-digest-enabled' => 'false',
                                                                                                      'remote-ip-port' => 61450,
                                                                                                      'authentication-method' => 'none',
                                                                                                      'local-ip-port' => '3260',
                                                                                                      'tsih' => 280,
                                                                                                      'connection-id' => 1,
                                                                                                      'connection-state' => 'Full_Feature_Phase',
                                                                                                      'remote-address' => '10.225.115.127',
                                                                                                      'local-address' => '10.228.128.125',
                                                                                                      'data-digest-enabled' => 'false'
                                                                                                    }
                                                                                                  ],
                                                                                  'show' => {
                                                                                              'status-admin' => 'up',
                                                                                              'target-name' => 'iqn.1992-08.com.netapp:sn.6dc5e7aff25311dfad2d123478563412:vs.6'
                                                                                            },
                                                                                  'nodename' => 'iqn.1992-08.com.netapp:sn.6dc5e7aff25311dfad2d123478563412:vs.6',
                                                                                  'status' => {
                                                                                                'status-admin' => 'up',
                                                                                                'is-available' => 'true'
                                                                                              },
                                                                                  'discovery_type' => 'sendtargets',
                                                                                  'initiator' => [
                                                                                                   {
                                                                                                     'tsih' => 280,
                                                                                                     'initiator-name' => 'iqn.1991-05.com.microsoft:ssan-w2k8-7.smokejumper.rtp.netapp.com',
                                                                                                     'initiator-alias' => '-',
                                                                                                     'isid' => '40:00:01:37:00:00'
                                                                                                   }
                                                                                                 ],
                                                                                  'tpgroup' => [
                                                                                                 {
                                                                                                   'lif' => [
                                                                                                              '10.228.128.125'
                                                                                                            ]
                                                                                                 }
                                                                                               ],
                                                                                  'session' => [
                                                                                                 {
                                                                                                   'tsih' => 280,
                                                                                                   'initiator-name' => 'iqn.1991-05.com.microsoft:ssan-w2k8-7.smokejumper.rtp.netapp.com',
                                                                                                   'connection-ids' => '1',
                                                                                                   'isid' => '40:00:01:37:00:00',
                                                                                                   'initiator-alias' => '-',
                                                                                                   'session-type' => 'Normal'
                                                                                                 }
                                                                                               ],
                                                                                  'security' => {
                                                                                                  'initiator-name' => 'iqn.1991-05.com.microsoft:ssan-w2k8-7.smokejumper.rtp.netapp.com',
                                                                                                  'auth-type' => 'none'
                                                                                                  'outbound-user-name' => '-',
                                                                                                  'user-name' => '-'
                                                                                                },
                                                                                  'interface' => [
                                                                                                   {
                                                                                                     'is-home' => 'true',
                                                                                                     'address' => '10.228.128.125',
                                                                                                     'enabled' => 'true',
                                                                                                     'ip-port' => '3260'
                                                                                                   }
                                                                                                 ]
                                                                                }
         };

=item Examples

    my $iscsi_info = $Client_Obj->get_iscsi_info();

=back

=cut
