# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to client initiators
## @author knopp@netapp.com, thomasm@netapp.com
## @status Private
## @pod here

# Package Name
##############
package Initiator;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use Error;

use vars qw(@ISA);

@ISA = qw(Error);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: display()
# Objective:       Display Initiator information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;
    my ($cmd, $name, $path, @output);

    $self->_clear_err();

    @output = (
           "Name: " . $self->get_name(),
           "========================================".
           "========================================",
           "Model: " . $self->get_model(),
           "Description: " . $self->get_description(),
           "Version: " . $self->get_version(),
           "Firmware: ". $self->get_firmware()
           );

    if ($self->get_protocol() =~ /fcp/i) {
        push(@output, "Nodename: ". $self->get_nodename(),
                      "Portname: ". $self->get_portname());
    } else{
        push(@output, "Initiator: ".$self->get_initiator());
    }

    push (@output, "\n");

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Initiator object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined ($self->{name})) {
        return $self->{name};
    } else {
        $self->_set_err("Error: Initiator::get_name() Initiator object is missing name");
        return "";
    }

}

################################################################################
# Subroutine Name: get_initiator
# Objective:       Return Initiator name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_initiator {
    my ($self) = @_;

    my ($initiator);

    $self->_clear_err();
    
    if (defined($self->{initiator})) {
        $initiator = $self->{initiator};
    } else {
        $self->_set_err("Error: Initiator::get_initiator() Initiator object does not have initiator name");
        $initiator = "";
    }

    return $initiator;
}

################################################################################
# Subroutine Name: get_nodename
# Objective:       Return Initiator nodename
# Details:         See POD documentation at the end of this file
################################################################################
sub get_nodename {
    my ($self, %args) = @_;
    my ($format, $nodename);

    $self->_clear_err();
    $format = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?format/) {
            if (not length($format = $args{$_}) and $args{$_} !~ /1|0|y|n|o/i) {
               $self->_set_err("Error: Initiator::get_nodename() Invalid value for argument '-format'");
               return "";
            }        
            next;
        }
    }    

    if (defined($self->{nodename})) {
        $nodename = $self->{nodename};
        if ($format =~  /1|y/i) {
            $nodename = $self->{nodename};
            $nodename =~ s/(..)(..)(..)(..)(..)(..)(..)(..)/$1:$2:$3:$4:$5:$6:$7:$8/;
        }
        return $nodename;
    } else {
        $self->_set_err("Error: Initiator::get_nodename() Initiator object is missing nodename");
        return "";
    }
}

################################################################################
# Subroutine Name: get_portname
# Objective:       Return Initiator portname
# Details:         See POD documentation at the end of this file
################################################################################
sub get_portname {
    my ($self, %args) = @_;
    my ($format, $portname); 

    $self->_clear_err();
    $format = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?format/) {
            if (not length($format = $args{$_}) and $args{$_} !~ /1|0|y|n|o/i) {
                $self->_set_err("Error: Initiator::get_portname() Invalid value for argument '-format'");
                return "";
            }        
            next;
        }
    }    
    
    if (defined($self->{portname})) {
        $portname = $self->{portname};
        if ($format =~ /1|y/i) {
            $portname = $self->{portname};
            $portname =~ s/(..)(..)(..)(..)(..)(..)(..)(..)/$1:$2:$3:$4:$5:$6:$7:$8/;
        }

        return $portname;
    } else {
        $self->_set_err("Error: Initiator::get_portname() Initiator object is missing portname");
        return "";
    }

}

################################################################################
# Subroutine Name: get_version
# Objective:       Return Initiator driver version
# Details:         See POD documentation at the end of this file
################################################################################
sub get_version {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{version})) {
        return $self->{version};
    } else {
        $self->_set_err("Error: Initiator::get_version() Initiator object is missing driver version");
        return "";
    }
}

################################################################################
# Subroutine Name: get_description
# Objective:       Return Initiator description
# Details:         See POD documentation at the end of this file
################################################################################
sub get_description {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{description})) {
        return $self->{description};
    } else {
        $self->_set_err("Error: Initiator::get_description() Initiator object is missing description");
        return "";
    }
}

################################################################################
# Subroutine Name: get_firmware
# Objective:       Return firmware status
# Details:         See POD documentation at the end of this file
################################################################################
sub get_firmware {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{firmware})) {
        return $self->{firmware};
    } else {
        $self->_set_err("Error: Initiator::get_status() Initiator object is missing firmware version");
        return "";
    }
}

################################################################################
# Subroutine Name: get_model
# Objective:       Return model
# Details:         See POD documentation at the end of this file
################################################################################
sub get_model {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{model})) {
        return $self->{model};
    } else {
        $self->_set_err("Error: Initiator::get_model() Initiator object is missing model");
        return "";
    }
}

################################################################################
# Subroutine Name: get_protocol
# Objective:       Return protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_protocol {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{protocol})) {
        return $self->{protocol};
    } else {
        $self->_set_err("Error: Initiator::get_protocol() Initiator object is missing protocol");
        return "";
    }
}

################################################################################
# Subroutine Name: is_fcp
# Objective:       Determine if initiator used fcp protocol
# Details:         See POD documentation at the end of this file
################################################################################
sub is_fcp {
    my ($self) = @_;

    $self->_clear_err();
    
    if ($self->get_protocol() =~ /fcp/i) {
        return 1;
    } else {
        return 0;
    }
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new Initiator object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, $args) = @_;

    my ($description, $firmware, $log, $model, $name, $nodename, $parent);
    my ($portname, $protocol, $self, $version);

    $self = {};

    bless($self, $class);    
    
    foreach (keys %$args) {
        if (/name/i) {
            if (not length($name = $args->{$_})) {
                $self->_set_err("Error: Initiator::_new() Invalid value for argument '-name'");
                return $self;
            }        
            next;
        } elsif (/^model$/i) {
            if (not length($model = $args->{$_})) {
                $self->_set_err("Error: Initiator::_new() Invalid value for argument '-model'");
                return $self;
            }        
            next;
        } elsif (/driver version/i) {
            if (not length($version = $args->{$_})) {
                $self->_set_err("Error: Initiator::_new() Invalid value for argument '-driver version'");
                return $self;
            }        
            next;
        } elsif (/firmware version/i) {
            if (not length($firmware = $args->{$_})) {
                $self->_set_err("Error: Initiator::_new() Invalid value for argument '-firmware version'");
                return $self; 
            }        
            next;
        } elsif (/WWNN/i) {
            if (not length($nodename = $args->{$_})) {
                $self->_set_err("Error: Initiator::_new() Invalid value for argument 'WWNN'");
                return $self; 
            }        
            next;
        } elsif (/WWPN/i) {
            if (not length($portname = $args->{$_})) {
                $self->_set_err("Error: Initiator::_new() Invalid value for argument '-cnt'");
                return $self;
            }        
            next;
        } elsif (/model description/i) {
            if (not length($description = $args->{$_})) {
                $self->_set_err("Error: Initiator::_new() Invalid value for argument '-model description'");
                return $self; 
            }        
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args->{$_})) {
                $self->_set_err("Error: Initiator::_new() Invalid value for argument '-parent'");
                return $self; 
            }        
            next;
        } elsif (/-?protocol/i) {
            if (not length($protocol = $args->{$_})) {
                $self->_set_err("Error: Initiator::_new() Invalid value for argument '-protocol'");
                return $self;
            }        
            next;
        } 
    }     
    
    $self->{name} = $name;
    $self->{protocol} = $protocol;
    $self->{parent} = $parent;
    $self->{model} = $model;
    $self->{description} = $description;
    $self->{portname} = $portname;
    $self->{nodename} = $nodename;
    $self->{initiator} = $nodename;
    $self->{firmware} = $firmware;
    $self->{version} = $version;

    return $self;

}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Initator.pm

=head1 Author

=begin html 

Gary Knopp


=end html

=head1 Description

Initiator provides an API interface to clients initiator(hardware or software).  The package should  
be instantiated through either of the two SanClient2 package methods; get_initiator and get_initiators(). 

=head1 See Also

Package inherits from the following modules. See these modules for further 
interface detail.

B<Error.pm>

=head1 Attributes

=head2 B<name>

string = Name of initiator

=head2 B<log>

filehandle = Filehandle to log output to.

=head2 B<protocol>

string = initiator protocol

=head2 B<parent>

reference = Reference to parent SanClient object

=head2 B<model>

string = model of initiator

=head2 B<description>

string = description of initiator

=head2 B<portname>

string = WWPN of initiator

=head2 B<nodename>

string = WWNN of initiator

=head2 B<initiator>

string = WWNN of initiator

=head2 B<firmware>

string = initiator firmware version

=head2 B<version>

string = initiator driver version

=head1 Public Methods     

=head2 B<display()>

Display detailed initiator information

=over 4

=item Inputs

None

=item Outputs

Returns an array or array reference if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and an empty array or array reference
is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving initiator information");

    $Test->log($initObj->display());
    if ($initObj->error){
        $Test->log($initObj->errmsg());
        return $TCase::FAIL;
    }


=back

=head2 B<get_description()>

Description of initiator.

=over 4

=item Inputs

none

=item Outputs

Returns the description of the client initiator.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The initiator description is '".$initObj->get_description()."'");


=back

=head2 B<get_firmware()>

Initiator firmware

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the initiator firmware.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The initiator firmware is '".$initObj->get_firmware()."'");

=back

=head2 B<get_model()>

Initiator model

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the initiator model. If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The initiator model '".$initObj->get_model()."'");

=back


=head2 B<get_name()>

Name of initiator

=over 4

=item Inputs

none

=item Outputs

Returns the name of the initiator.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The initiator name is '".$initObj->get_name()."'");

=back


=head2 B<get_initiator()>

Retrieve iscsi initiator node name

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the iscsi initiator name.  If an error occurs
the error flag is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The initiator name is ".$initObj->get_initiator());

=back

=head2 B<get_nodename(I<%args>)>

Retrieve nodename of initiator

=over 4

=item Inputs

=item * I<-format> (default: 0);

A boolean flag to indicate whether the nodename should be formatted 
(ex 10:a9:80:01:02:00:8e:1c), otherwise it is returned without the 
':' seperators.


=item Outputs

Returns a scalar containing the initiator nodename.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The initiator nodename is ".$initObj->get_nodename(-format=>1));

=back

=head2 B<get_portname(I<%args>)>

Retrieve portname of initiator

=over 4

=item Inputs

=item * I<-format> (default: 0)

A boolean flag to indicate whether the portname should be formatted 
(ex 10:a9:80:01:02:00:8e:1c), otherwise it is returned without the 
':' seperators.

=item Outputs

Returns a scalar containing the initiator portname.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The initiator portname is ".$initObj->get_portname(-format=>1));

=back

=head2 B<get_protocol()>

Retrieve the initiator protocol

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the initiator protocol.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        

=item Examples

    $Test->nlog("The initiator protocol is ".$initObj->get_protocol());


=back


=head2 B<get_version()>

Retrieve the version of the initiator

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the initiator version.  If 
an error occurs the error flag is set, errmsg is set to the failure 
reason, and "" is returned.        


=item Examples

    $Test->nlog("The initiator firmware version is ".$initObj->get_version());

=back

=cut

