# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to LVM
## @author knopp@netapp.com, pressley@netapp.com, latif@netapp.com
## @status Private
## @pod here

package Lvm;

use strict;

use Error;

use vars qw(@ISA);

@ISA = qw(Error);

sub _new {
    my ($class, %args) = @_;
    my ($self, $hosttype);

    $self = {};

    # bless the object upfront in case an error occurs.
    bless $self, $class;

    # check for existance of required -lvm argument
    if (grep (/^-lvm$/, (keys %args))) {
        unless ($args{-lvm} =~ /^vxvm$|^nativelvm$|^svm$|^dmp$|^zfs$|^rdm$|^vmfs$/i) {
            $self->_set_err(
                 "Error: Lvm::new() Unsupported or invalid volume manager for argument '-lvm' \(" . $args{-lvm} . "\)");
            return $self;
        }
    } else {
        $self->_set_err("Error: Lvm::new() Missing volume manager name '-lvm'", "ex: vxvm, nativelvm, or svm");
        return $self;
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Error: Lvm::new() Invalid value for argument '-parent'");
            return $self;
        }
        $hosttype = $args{-parent}->get_connection()->{hostp}{hosttype};
    } else {
        $self->_set_err("Error: Lvm::new() Missing required argument '-parent'");
        return $self;
    }

    # load the appropriate lvm for the hosttype
    if ($hosttype =~ /unix-solaris/i and $args{-lvm} =~ /vxvm|dmp/i) {
        require San::SanClient::Lvm::Solaris::Vxvm;
        $self = San::SanClient::Lvm::Solaris::Vxvm->_new(%args);
    } elsif ($hosttype =~ /unix-solaris/i and $args{-lvm} =~ /svm/i) {
        require San::SanClient::Lvm::Solaris::SVM;
        $self = San::SanClient::Lvm::Solaris::SVM->_new(%args);
    } elsif ($hosttype =~ /unix-hpux/i and $args{-lvm} =~ /nativelvm/i) {
        require San::SanClient::Lvm::HPUX::NativeLvm;
        $self = San::SanClient::Lvm::HPUX::NativeLvm->_new(%args);
    } elsif ($hosttype =~ /unix-hpux/i and $args{-lvm} =~ /vxvm|dmp/i) {
        require San::SanClient::Lvm::HPUX::Vxvm;
        $self = San::SanClient::Lvm::HPUX::Vxvm->_new(%args);
    } elsif ($hosttype =~ /unix-aix/i and $args{-lvm} =~ /nativelvm/i) {
        require San::SanClient::Lvm::Aix::NativeLvm;
        $self = San::SanClient::Lvm::Aix::NativeLvm->_new(%args);
    } elsif ($hosttype =~ /unix-aix/i and $args{-lvm} =~ /vxvm/i) {
        require San::SanClient::Lvm::Aix::Vxvm;
        $self = San::SanClient::Lvm::Aix::Vxvm->_new(%args);
    } elsif ($hosttype =~ /windows/i and $args{-lvm} =~ /vxvm|dmp/i) {
        require San::SanClient::Lvm::Windows::Vxvm;
        $self = San::SanClient::Lvm::Windows::Vxvm->_new(%args);
    } elsif ($hosttype =~ /unix-linux/i and $args{-lvm} =~ /vxvm|dmp/i) {
        require San::SanClient::Lvm::Linux::Vxvm;
        $self = San::SanClient::Lvm::Linux::Vxvm->_new(%args);
    } elsif ($hosttype =~ /unix-linux/i and $args{-lvm} =~ /nativelvm/i) {
        require San::SanClient::Lvm::Linux::NativeLvm;
        $self = San::SanClient::Lvm::Linux::NativeLvm->_new(%args);
    } elsif ($hosttype =~ /unix-solaris/i and $args{-lvm} =~ /zfs/i) {
        require San::SanClient::Lvm::Solaris::ZFS;
        $self = San::SanClient::Lvm::Solaris::ZFS->_new(%args);
    } elsif ($hosttype =~ /linux-vmware/i and $args{-lvm} =~ /vmfs/i) {
        if ($args{-parent}->has_vma_host()) {
            require San::SanClient::Lvm::ESXi::VMFS;
            $self = San::SanClient::Lvm::ESXi::VMFS->_new(%args);
        } else {
            require San::SanClient::Lvm::ESX::VMFS;
            $self = San::SanClient::Lvm::ESX::VMFS->_new(%args);
        }
    } elsif ($hosttype =~ /linux-vmware/i and $args{-lvm} =~ /rdm/i) {
        if ($args{-parent}->has_vma_host()) {
            require San::SanClient::Lvm::ESXi::RDM;
            $self = San::SanClient::Lvm::ESXi::RDM->_new(%args);
        } else {
            require San::SanClient::Lvm::ESX::RDM;
            $self = San::SanClient::Lvm::ESX::RDM->_new(%args);
        }
    } else {
        $self->_set_err("Error: Lvm::new() Lvm '" . $args{-lvm} . "'" . "not currently supported with this library");
        return $self;
    }

    return $self if ($self->error);

    $self->{parent} = $args{-parent};
    $self->set_debug_level($self->{parent}->get_debug_level());

    #
    # By returning $self here we are returning a parent class such
    # as Vxvm, not a Lvm object.  This is OK because
    # the OS specific Lvm::XXXLvm classes add Lvm in @ISA
    # so Lvm is the base class for the parent
    #
    return ($self);

}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Lvm.pm

=head1 Author

Gary Knopp

=head1 Description

Lvm provides an abstract API interface to a host logical volume manager.
This package should be instantiated through the SanClient package
create_lvm() method.

    $lvm_default=$Client->get_lvm_default();
    $lvm_obj=$Client->create_lvm(-lvm=>$lvm_default);
    if ($Client->error()){
    $Test->nlog($Client->errmsg());
    return 0;
    }

=head1 See Also

Package inherits from the following modules. The abstract methods for LVM::* are included below.

B<San::SanClient::Lvm::[Solaris::Vxvm|HPUX::NativeLvm.pm|Aix::SanPath.pm|Linux::NativeLvm.pm]>

B<Error.pm>

=head1 Attributes

=head2 B<parent>

SanClient object = Reference to a parnet SanClient object

=head1 Public Methods

=head2 B<create_vg(%args)>

Creates a new volume group object.

=over 4

=item Inputs

=item * -name

The name of the volume group

=item * -devices

The array reference of initialized devices to use in the volume group

=item Outputs

Returns a volume group object. If an error occurs the error flag is set, errmsg is
set to the failure reason.

=item Examples

    # Create a volume Group object.
    $Test->log("-- Creating volume group 'mydg'");
    my $dgObj = $Lvm->create_vg(-name=>"mydg",-devices=>\@devs);
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<destroy_vg($name)>

Destroy a volume group object by name

=over 4

=item Inputs

=item * $name

The name of the volume group object to destroy

=item Outputs

Returns a scalar value of 1 if successful.  If the volume group does not exist or
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.


=item Examples

    $Test->log("-- Destroy disk group object 'mydg'");
    if (not $Lvm->destroy_vg("mydg")){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<display()>

Display detailed information for all volume groups

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM version information");

    $Test->log($Lvm->display());
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }


=back

=head2 B<get_format_default()>

Retrieve the default disk format for initializing device under lvm control

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value containing the default format if successful. If no
default exists than an empty scalar is returned.


=item Examples

    $Test->log("-- Retrieving default fisk format");
    $disk_format=$Lvm->get_format_default("c3t1d5");


=back

=head2 B<get_disk_format($name)>

Retrieve disk format used by device name

=over 4

=item Inputs

=item * $name

The name of the disk device

=item Outputs

Returns a scalar value containing the format if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving disk group object 'dg1'");
    $disk_format=$Lvm->get_disk_format("c3t1d5");
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<get_disk_formats()>

Returns a list of disk formats supported by the volume manager for initializing the device.

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of format values if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and an empty array is returned.


=item Examples

    $Test->log("-- Retrieving lvm disk formats");
    @disk_formats=$Lvm->get_disk_formats();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }


=back

=head2 B<get_free_devices(I<%args>)>

Returns a list of devices that are not being used by other
LVM objects.  The list of devices can be selected by any combination
of filer, volume, and lun.

=over 4

=item Inputs

=item * I<-cnt>

The number of free devices to retrieve.  If the number of free devices found is less
than the count than the error flag is set.

=item * I<-filer>

Retrieve host devices belonging to a specific filer.

=item * I<-vol>    (ex: vol2)

Retrieve host devices belonging to a specific filer volume.  Only supply the filer
volume name and not the full volume path.

=item * I<-lun> (ex: mylun)

Retrieve the host device mapped to a specific lun.  Only supply the lun name and not
the full path

=item * I<-refresh>    (default: 0)

Forces the host device to filers maps to be refeshed.  This is automatically done the
first time the method is called.  It needs to be called whenever a test program
dynamically creates and discovers new luns in a test method. This will cause the map
to be updated and correctly reflect device metadata on the host.


=item Outputs

Returns an array or array reference of free devices.  If an error occured the error flag
is set and an errmsg is set to the failure reason.


=item Examples

    $Test->log("Get 3 free devices from filer 'myfiler'");
    @devs=$Lvm->get_free_devices(-cnt=>3,-filer=>'myfiler');
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }


=back

=head2 B<get_name()>

Retrieve name of the installed logical volume manager.

=over 4

=item Inputs

none

=item Outputs

Returns an scalar value if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and "" is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM name");
    $name=$Lvm->get_name();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

    $Test->log("LVM is ".$name);

=back

=head2 B<get_pkginfo()>

Retrieve package information on the installed logical volume manager.

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM package information");
    $buf=$Lvm->get_pkginfo();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

    foreach $line (@$buf) {
        $Test->log($line);
    }

=back

=head2 B<get_revision()>

Retrieve revision of the installed logical volume manager.

=over 4

=item Inputs

none

=item Outputs

Returns an scalar value if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and "" is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM revision");
    $rev=$Lvm->get_revision();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

    $Test->log("LVM revision is ".$rev);

=back

=head2 B<get_version()>

Retrieve version of the installed logical volume manager.

=over 4

=item Inputs

none

=item Outputs

Returns an scalar value if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and "" is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM version information");
    $version=$Lvm->get_version();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

    $Test->log("LVM version is ".$version);

=back

=head2 B<get_vg($name)>

Retrieve a volume group object by name

=over 4

=item Inputs

=item * $name

The name of the volume group object to retrieve

=item Outputs

Returns a volume group object if successful.  If the volume group does not exist or
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is
passed back as the return value.


=item Examples

    $Test->log("-- Retrieving disk group object 'dg1'");
    $vgObj=$Lvm->get_vg("dg1");
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<get_vgs()>

Retrieve a list of all volume group objects

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of volume groups if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and an empty array is returned.


=item Examples

    $Test->log("-- Retrieving disk group objects");
    @vgObjs=$Lvm->get_vgs();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }


=back

=head2 B<get_vgnames()>

Retrieve a list of all volume group names

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of volume group names if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and an empty array is returned.


=item Examples

    $Test->log("-- Retrieving disk group names");
    @neames=$Lvm->get_vgnames();
    if ($Lvm->error){
        $Test->nlog($Lvm->errmsg());
        return $TCase::FAIL;
    }


=back

=head2 B<initialize(%args)>

Initialize a list of devices under LVM control

=over 4

=item Inputs

=item * -devices

The list of devices to initialize under LVM control

=item Outputs

Returns a scalar value of 1 if successful.  If an error occurs the error flag
is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples


    $Test->log("-- Initialize devices under lvm control");
    $Lvm->initialize(-devices=>\@devs);
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }


=back

=head2 B<load_vg($name)>

Load a volume group object that is not currently in LVM object cache

=over 4

=item Inputs

=item * $name

The name of the volume group object to load

=item Outputs

Returns a volume group object if successful.  If the volume group does not exist or
an error occured the error flag is set, errmsg is set to the failure reason, and an empty
hash ref is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving disk group object 'dg1'");
    $vgObj=$Lvm->load_vg("dg1");
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<rescan()>

Rescan the disks on the host.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Rescan devices");

    $Lvm->rescan();
    if ($Lvm->error){
    $Test->log($Lvm->errmsg());
    return $TCase::FAIL;
    }


=back

=head2 B<unload_vg($name)>

Unload a volume group object that is currently in LVM object cache

=over 4

=item Inputs

=item * $name

The name of the volume group object to load

=item Outputs

Returns a scalar value of 1 if successful.  If the volume group is not loaded or
an error occured the error flag is set, errmsg is set to the failure reason, and a 0
is passed back as the return value.


=item Examples

    $Test->log("-- Unloading disk group object 'dg1'");
    $Lvm->unload_vg("dg1");
    if ($Lvm->error){
    $Test->log($Lvm->errmsg());
    return $TCase::FAIL;
    }

=back

=cut
