# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the NativeLvm Logical Volume Manager on Aix
## @author thomasm@netapp.com
## @status private
## @pod ../../Lvm.pm

# Package Name
##############
package San::SanClient::Lvm::Aix::NativeLvm;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanClient::Lvm;
use San::SanLog;

require San::SanClient::Lvm::Aix::NativeLvm::Vg;

use Error;

use vars qw(@ISA);

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: initialize
# Objective:       initialize disk under veritas control
#                  This function is a NO-OP for NativeLvm on AIX. 
# Details:         See POD documentation at the end of this file
################################################################################
sub initialize () {
    my ($self,%args) = @_;

    my ($cmd, @already_inited, @output, $status, $devices);

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: NativeLvm::initialize() Invalid value for argument '-devices'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        } 
    }     

    # check for required parameters
    unless(defined $devices and length $devices){
        $self->_set_err("Error: NativeLvm::initialize() -devices is a required argument");
        return ($ClientUtil::FALSE);
    }

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: clear_cache
# Objective:       clear cache objects.
# Details:         See POD documentation at the end of this file
################################################################################
sub clear_cache {
    my ($self) = @_;
    
    # Clear cache of vg objects
    $self->{vgs} = ();         

    return ($ClientUtil::TRUE);
    
} 

################################################################################
# Subroutine Name: create_vg
# Objective:       create and return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vg {
    my ($self, %args) = @_;

    my ($vgObj);
    
    $vgObj = San::SanClient::Lvm::Aix::NativeLvm::Vg->_new(-parent => $self,
                                                           -conn => $self->{conn},
                                                           %args);

    if ($vgObj->error){
        $self->_set_err($vgObj->errmsg());
        return ($self);
    }        

    # Check and remove any cached copy of the Vg that
    # may still exist in the LVM due to a deletion
    # occuring outside the test's control.  Ignore error
    # if id doesn't currently exist.

    $self->unload_vg($vgObj->get_name());
    if ($self->error) {
        $self->_clear_err();
    }

    push (@{$self->{vgs}}, $vgObj);
    
    return ($vgObj);

}

################################################################################
# Subroutine Name: destroy_vg
# Objective:       destroy a named diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vg {
    my ($self, $name) = @_;

    my ($vgObj, @vgs);
    
    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: NativeLvm::destroy_vg() vg name is a required argument");
        return ($ClientUtil::FALSE);
    }
    
    # Delete the of vg object and update the parent vg list.
    foreach $vgObj ( @{$self->{vgs}}) { 
        if ($vgObj->get_name() =~ /^$name$/) {
            $vgObj->destroy();
            if ($vgObj->error) { 
                $self->_set_err($vgObj->errmsg());
                return ($ClientUtil::FALSE);
            }
        } else {
            push (@vgs,$vgObj);
        }
    }
    
    $self->{vgs} = \@vgs;

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: display
# Objective:       display volume group information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;

    my ($cmd, $status, $verbose, @buffer, @output, @vgObjs);
    
    $self->_clear_err();

    $self->sanlog(-msg => "Display diskgroup information");

    @vgObjs = $self->get_vgs();
    if ($self->error) {
        return ();
    }

    foreach my $vgObj (@vgObjs) {                
        @output = $vgObj->display(%args);
        return (wantarray()) ? () : [] if ($self->error);

        push @buffer,
        @output[0..scalar(@output) -2],   
        "=====================================".
            "=====================================";
    }

    return (wantarray()) ? @buffer : \@buffer;    
    
}

###############################################################################
# Method:    get_disk_format
# Objective: Get list of supported disk formats
#            This is a no-op for Aix.  Function supports veritas lvm
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_format {
    my ($self, $disk) = @_;

    my ($format);

    $self->_clear_err();

    $format = "n/a";

    return ($format); 

}

###############################################################################
# Method:    get_disk_formats
# Objective: Get list of supported disk formats
#            This is a no-op for Aix.  Function supports veritas lvm
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_formats {
    my ($self) = @_;

    my (@format);

    $self->_clear_err();

    @format = ();

    return (wantarray()) ? @format : \@format;  

}

###############################################################################
# Method:    get_format_default
# Objective: Get default type to use for disk format
#            This is a no-op for Aix.  Function supports veritas lvm
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_format_default {
    my ($self) = @_;

    my ($format);

    $self->_clear_err();

    $format = "n/a";
    
    return ($format);   

}


################################################################################
# Subroutine Name: get_vg
# Objective:       return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vg {
    my ($self, $name) = @_;

    my ($vgObj);
    
    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: NativeLvm::get_vg() vg name is a required argument");
        return ($ClientUtil::FALSE);
    }

    foreach my $obj ( @{$self->{vgs}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return ($obj);
        }
    }
    
    $self->_set_err("Error: NativeLvm::get_vg() Unable to find diskgroup '$name'");

    return ($ClientUtil::FALSE);

}

################################################################################
# Subroutine Name: get_vgs
# Objective:       return array of volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgs {
    my ($self) = @_;

    my (@vgs);
    
    $self->_clear_err();

    if (defined $self->{vgs}) {
        @vgs = @{$self->{vgs}};
    } else {
        $self->_set_err("Error: NativeLvm::get_vgs() Vg object has no volumes groups");
        return ();                
    }

    return (wantarray()) ? @vgs : \@vgs;
    
} 

################################################################################
# Subroutine Name: get_vgnames
# Objective:       return array of volume group names
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgnames {
    my ($self) = @_;

    my ($cmd, $status, @buffer, @output);
    
    $self->_clear_err();

    # ignore rootvg
    $cmd = "lsvg | grep -v root";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: NativeLvm::get_vgnames Command ($cmd) returned a bad status ($status)");     
        return (wantarray()) ? () : [];
    }

    # Remove dev prefix to vg name
    foreach my $line (@output) {
        push (@buffer, $line);
    }

    return (wantarray()) ? @buffer : \@buffer;
    
} 

################################################################################
# Subroutine Name: get_pkginfo()
# Objective:       get detailed lvm pkg information
# Details:         See POD documentation at the end of this file
################################################################################
sub get_pkginfo {
    my ($self) = @_;

    my ($cmd, @output, $status);
    
    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{pkginfo});
    
    return (wantarray()) ? @{$self->{pkginfo}} : $self->{pkginfo};   

}

################################################################################
# Subroutine Name: get_revision()
# Objective:       Get revision of AIX LVM
# Details:         See POD documentation at the end of this file
################################################################################
sub get_revision {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{rev});

    if ($self->error()) {
        return ("");
    }

    return ($self->{rev});  
    
}

################################################################################
# Subroutine Name: get_name()
# Objective:       Get lvm name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{name});

    if ($self->error()) {
       return ("");
    }

    return ($self->{name});  
    
}

################################################################################
# Subroutine Name: get_version()
# Objective:       Check version of AIX LVM
# Details:         See POD documentation at the end of this file
################################################################################
sub get_version {
    my ($self) = @_;

    my ($cmd, @output, $status, $version);

    $self->_clear_err();        
    
    $self->_get_pkginfo() unless (defined $self->{vers});

    if ($self->error()) {
        return("");
    }

    return $self->{vers};
    
}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices { 
    my ($self, %args) = @_;

    my ($bCnt, $cmd, $cnt, $device, $filer, $filername, $lun, $refresh);
    my ($status, $total, $vol, %hDevs, @all_devices, @attr);
    my (@devices, @lvm_devices, @output, @used);
    my (@inited_states, $inited_state, $bootdevice, $filers, $protocol);

    $total = 0;
    @{$filers} = ();

    $self->_clear_err();

    # Asssume user supplied -cnt argument
    $bCnt = 1;
    $refresh = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Error: AixNativeLvm::get_free_devices() Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [];                          
            }        
            next;
        } elsif (/-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Error: AixNativeLvm::get_free_devices() Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];                          
            }
            next;
        } elsif (/^-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Error: AixNativeLvm::get_free_devices() Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Error: AixNativeLvm::get_free_devices() Invalid value for argument '-vol'");
                return (wantarray()) ? () : [];                          
            }        
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Error: AixNativeLvm::get_free_devices() Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];                          
            }        
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: AixNativeLvm::get_free_devices() Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [];                          
            }        
            next;
        } elsif (/^-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: AixNativeLvm::get_free_devices() Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }     
    
    # Check if user wants free devices from a specified filer, volume, or lun
    if (defined $filer or defined $vol or defined $lun or defined $protocol or scalar(@{$filers}) != 0 or $refresh == 1) {
        # Check if a device to filer map 'dfmap' has already been created or 
        # refresh it if requested.
        if (not defined $self->{dfmap} or $refresh eq 1) {
            $self->sanlog(-msg => "Creating translation maps; device to filer, vol, and lun");
            @output = $self->{sanlun}->lun_show();
            if ($self->{sanlun}->error) {
                $self->_set_err($self->{sanlun}->errmsg());
                return (wantarray()) ? () : [];                          
            }

            foreach my $line (@output) {
                # skip first line (header)
                if ($line =~ /lun-pathname/i) {
                    next;
                }

                # depending on filername length sanlun may not have display blank space before name
                my ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp);
                if ($line  =~ /\s{0}(\S+):*\s+\/vol\/(\w+)\/(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+/) {
                    # The line is in the following format
                    # p720_aix03c1_data:  /vol/p720_aix03c1_data32_0/p720_aix03c1_data32_lun_0  hdisk1           fcs0     FCP            2g (2147483648)     GOOD
                    # p720_aix03c1_data:  /vol/p720_aix03c1_data64_1/p720_aix03c1_data32_lun_0  hdisk2           fcs0     FCP            2g (2147483648)     GOOD
                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp) = ($1, $2, $3, $4, lc($5));
                } elsif ($line =~ /\s{0}(\S+):*\s+\/vol\/(\w+)\/(\S+)\s+(\S+)\s+/) {
                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp) = ($1, $2, $3, $4);
                }

                if ($device_tmp eq "" or $device_tmp =~ /^\s*$/) {
                    $self->sanlog(-msg => "Output from sanlun is incorrect:\n$line\n");
                    next;
                }

                my $dev_tmp = basename($device_tmp);

                $hDevs{$dev_tmp} = {filer    => $filer_tmp,
                                    vol      => $vol_tmp,
                                    lun      => $lun_tmp,
                                    protocol => $protocol_tmp};

                $self->sanlog(-msg => "Device ($dev_tmp) information:\n\tFiler: $filer_tmp\n\tVolume: $vol_tmp\n" .
                                      "\tLUN: $lun_tmp\n\tProtocol: $protocol_tmp\n",
                              -msg_level => 3);

            }
            
            # save map for next call of function
            $self->{dfmap} = \%hDevs;

        }
    }

    # If not -cnt then assume all free disks 
    # under lvm control are wanted
    unless (defined $cnt and length($cnt)) {
        $cnt = 10000;
        $bCnt = 0;
    }

    # Get list of devices. Make sure you take in account deported vgs.
    if ($self->{parent}->get_protocol() =~ /fcp/i) {

        if ($self->{parent}->{mp_type} =~ /sanpath/i) {
            $cmd = "setsp | egrep  'spd[0-999]' | grep 'NETAPP' | awk '{print \$3;}'";
        } elsif ($self->{parent}->{mp_type} =~ /mpio/i) {
            $cmd = "lsdev -C -c disk | egrep \"\(MPIO)\|\(Virtual\)\" | egrep -i \"\(NetApp\)\|\(ONTAP\)\|\(Virtual\) FCP\" | awk '{print \$1}'"; 
        }

        ($status, @lvm_devices) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status gt 0) {
            $self->_set_err("Error: AixNativeLvm::get_free_devices() Command ($cmd) returned a bad status ($status)");
            return (wantarray()) ? () : [];                          
        }

        # determine if there were any sanpath devices available
        if ($self->{parent}->{mp_type} =~ /sanpath/i) {
            if ( grep(/no spd devices found/i, @lvm_devices) > 0 ) {
                $self->_set_err("Error: AixNativeLvm::get__free_devices() Command ($cmd) failed to find Sanpath devices.  Check license sppath -v.");
                return (wantarray()) ? () : [];                          
            }
        }
    } elsif ($self->{parent}->get_protocol() =~ /iscsi/i) {
        if ($self->{parent}->{mp_type} =~ /mpio/i) {
            $cmd = "lsdev -C -c disk | egrep \"\(MPIO)\|\(Virtual\)\" | egrep -i \"\(NetApp\)\|\(ONTAP\)\|\(Virtual\) iSCSI\" | awk '{print \$1;}'"; 
        } else {
            $cmd = "lsdev -C -c disk | egrep -i \"\(NetApp\)\|\(ONTAP\)\|\(Virtual\)\" | awk '{print \$1}'";
        }
        ($status, @lvm_devices) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status gt 0) {
            $self->_set_err("Error: AixNativeLvm::get_free_devices() Command ($cmd) returned a bad status ($status)");
            return (wantarray()) ? () : [];                          
        }
    }

    # Get list of all free devices 
    $cmd = "lspv | grep -i none | awk '{print \$1}'";  
    ($status, @all_devices) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: AixHost::get_free_devices() Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];                          
    } 

    # get the boot device - do not attempt to rmdev this - could
    # be sanbooted netapp device.
    $cmd = "bootinfo -b";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: AixDevMgr::get_free_devices() Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];                          
    }
    $bootdevice = $output[0];

    # iterate through the devices and make sure that none of the disks 
    # are being used other LVM disk object.  
    foreach my $device (@lvm_devices) {                

        # skip the boot device
        if ($device eq $bootdevice) {
            $self->sanlog(-msg => "\nFilter bootdevice: $bootdevice\nDevice: $device",
                          -msg_level => 3);
            next;
        }

        # If  the device is being used by other VGs, skip it
        if (grep(/^$device$/, @all_devices) == 0) {
            $self->sanlog(-msg => "\nFilter alldevices: " . join(" " , @all_devices) . "\nDevice: $device",
                          -msg_level => 3);
            next;
        }

        # Apply filters to any of the free devices found
        if (defined $filer and length $filer) {
            $self->sanlog(-msg => "\nFilter filer: $filer\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i);
        }

        if (defined $vol and length $vol) {
            $self->sanlog(-msg => "\nFilter volume: $vol\nDevice: $device\nDevice volume: " . $self->{dfmap}{$device}{vol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i);
        }

        if (defined $lun and length $lun) {
            $self->sanlog(-msg => "\nFilter lun: $lun\nDevice: $device\nDevice lun: " . $self->{dfmap}{$device}{lun},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{lun} !~ /^$lun$/i);
        }

        if (defined $protocol and length $protocol) {
            $self->sanlog(-msg => "\nFilter protocol: $protocol\nDevice: $device\nDevice protocol: " . $self->{dfmap}{$device}{protocol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{protocol} !~ /$protocol/i);
        }

        # Filter on a filer list if provided by caller
        if (defined $filers and scalar(@{$filers}) != 0) {
            my $found = 0;
            # Look through each filer in the filers list
            foreach my $filer_f (@{$filers}) {
                $self->sanlog(-msg => "\nFilter filer(s): $filer_f\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                              -msg_level => 3);
                # If filer matches devices filer, set found to 1
                if ($self->{dfmap}{$device}{filer} =~ /^$filer_f$/i) {
                    $found = 1;
                }
            }

            # If the device is not part of the filers list caller requested skip it.
            if ($found == 0) {
                next;
            }
        }

        push @devices, $device;
    } 

    # If user supplied cnt and free devices fell short
    # generate error
    if ($bCnt eq 1 and scalar(@devices) < $cnt){
        $self->_set_err("Error: AixNativeLvm::get_free_devices() ",
                        "Only ". scalar(@devices) . 
                        " of $cnt requested devices were found");
    }

    return (wantarray()) ? @devices : \@devices; 

}

################################################################################
# Subroutine Name: delete_devices
# Objective:       delete devices from host
# Details:         See POD documentation at the end of this file
################################################################################
sub delete_devices {
    my ($self, @devices) = @_;

    my ($cmd, $status, @output); 

    $self->_clear_err();

    # Remove devices from host.  This is because orphaned
    # devices maintain metadata of prior lun
    if ($self->{parent}->get_protocol() =~ /fcp/) {
        
        foreach my $device (@devices){
            if ($self->{parent}->{mp_type} =~ /sanpath/i) {
                # Determine spd id for master device name
                my $spd_id = $self->_get_spd_id($device);
                $self->sanlog(-msg => $self->errmsg()) if ($self->error);
        
                # Retrieve all device paths for spd_id
                my @alldevpaths = $self->_get_spd_devices($spd_id);
                $self->sanlog(-msg => $self->errmsg()) if ($self->error);
        
                # Stop master spd device
                $self->_stop_spd_id($spd_id);
                $self->sanlog(-msg => $self->errmsg()) if ($self->error);
    
                # Remove master spd device
                $self->_remove_spd_id($spd_id);
                $self->sanlog(-msg => $self->errmsg()) if ($self->error);

                # Remove device from host
                foreach my $device (@alldevpaths) {
                    $cmd = "rmdev -d -l " . $device;
                    ($status, @output) = &ClientUtil::execute($self->{conn}, 
                                                              $cmd);
                    if ($status gt 0) {
                        $self->_set_err("Error: AixNativeLvm::delete_devices() Unable to remove device '" .
                                        $device . "'", @output);
                        return ($ClientUtil::FALSE);             
                    } 
                }
            } elsif ($self->{parent}->{mp_type} =~ /mpio/i) {
                $cmd = "rmdev -d -l " . $device;
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
                if ($status gt 0) {
                    $self->_set_err("Error: AixNativeLvm::delete_devices() Unable to remove device '" . 
                                    $device . "'", @output);
                    return($ClientUtil::FALSE);                       
                }
            }
        }
    } else {
        foreach my $device (@devices) {
            $cmd = "rmdev -d -l " . $device;
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status gt 0) {
                $self->_set_err("Error: AixNativeLvm::delete_devices() Unable to remove device '" . 
                                $device . "'", @output);
                
                return($ClientUtil::FALSE);                       
            }
        }
    }

    return($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: load_vg
# Objective:       load an existing diskgroup that is not in memory
# Details:         See POD documentation at the end of this file
################################################################################
sub load_vg {
    my ($self, $name) = @_;

    my ($vgObj); 

    $self->_clear_err();

    unless (defined $name and length $name){
        $self->_set_err("Error: AixNativeLvm::load_vg() Invalid or missing required paramter \$name");
        return {};
    }

    # Check if vg has already been loaded
    $vgObj = $self->get_vg($name);

    if ($self->error) {
        # VG has not been load yet
        $self->_clear_err();

        # load VG
        $self->_create_vgs($name);
        if ($self->error) {
            return {};
        }

        $vgObj = $self->get_vg($name);                
    } else {
        $self->sanlog(-msg => "AixNativeLvm::load_vg() Volume group '$name' was already loaded");
    }

    return $vgObj;
    
}

################################################################################
# Subroutine Name: unload_vg
# Objective:       Remove the vg object under LVM object control
# Details:         See POD documentation at the end of this file
################################################################################
sub unload_vg {
    my ($self, $name) = @_;

    my ($vgObj, $bFound, @vgs); 

    $self->_clear_err();

    unless (defined $name and length $name){
        $self->_set_err("ERROR: AixNativeLvm::unload_vg() Invalid or missing required paramter \$name");
        return {};
    }

    $bFound = 0;
    
    foreach $vgObj (@{$self->{vgs}}) {
        ($vgObj->get_name() =~ /^$name$/)
            ? $bFound=1
                : push (@vgs, $vgObj);
    }
    
    $self->{vgs} = (@vgs) ? \@vgs : (); 

    unless($bFound) {
        $self->_set_err("Error: AixNativeLvm::unload_vg() Nothing to unload. Vg name '" .
                        $name . "' was not loaded");
        return ($ClientUtil::FALSE);
    }        

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: enable_controllers
# Objective:       enable controllers (NO-OP for AIX)
# Details:         See POD documentation at the end of this file
################################################################################
sub enable_controllers {
    my ($self) = @_;
    
    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: rescan
# Objective:       To run the sanpath command to rescan  list of
#                  disks that it thinks are under it's control.
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();        

    if ($self->{parent}->{mp_type} =~ /sanpath/i) {
        if ($self->{parent}->get_protocol() =~ /fcp/) {
            # refresh sanpath devices
            $cmd = "setsp -T; setsp -S";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            # ignore errors since any busy devices will complain
        }
    }

    return ($ClientUtil::TRUE);  

}

################################################################################
# Subroutine Name: restart_recovery_daemon
# Objective:       Method that will stop the recovery deamon and reset 
#                  (NO-OP for AIX)
# Details:         See POD documentation at the end of this file
################################################################################
sub restart_recovery_daemon {
    my ($self, $interval) = @_;
    
    return ($ClientUtil::TRUE);  

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($conn, $loadcfg, $log, $lvm, $parent, $self);

    $self = {};

    bless ($self, $class);        

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_})) {
                $self->_set_err("Error: NativeLvm::_new() Invalid value for argument '-lvm'");
                return ($ClientUtil::FALSE); 
            }        
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: NativeLvm::_new() Invalid value for argument '-parent'");
                return ($ClientUtil::FALSE); 
            }
            next;
        } elsif (/-?loadcfg/) {
            if (not length($loadcfg = $args{$_}) or $loadcfg !~ /no|yes|1|0|off|on/i) {
                $loadcfg = 1;
            }
            next;
        } 
    }     
    
    # check for existance of required -lvm argument
    if (grep ( /^-lvm$/, (keys %args) )) {
        unless ($args{-lvm} =~ /nativelvm/i) {
            $self->_set_err("Error: NativeLvm::_new()",
                            "Invalid value for argument '-lvm' \(" .
                            $args{-lvm} . "\)");
            return ($self);         
        }
    } else {
        $self->_set_err("Error: NativeLvm::_new()",
                        "Missing required argument'-lvm'");
        return ($self); 
    }

    # check for existance of required -parent argument
    if (grep ( /^-?parent$/, (keys %args) )) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Error: NativeLvm::_new() ",
                            "Invalid value for argument '-conn'");
            return ($self); 
        }
    } else {
        $self->_set_err("Error: NativeLvm::_new()",
                        "Missing required argument '-conn'");
        return ($self); 
    }

    # initialize the object members
    $self->{conn} = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    
    $self->{vgs} = [];

    # create vg objects for existing volume groups
    if ($loadcfg) {
        $self->_create_vgs();
    }
   
    if ($self->error) {
        return ($self);
    }

    return ($self);

}

################################################################################
# Subroutine Name: _create_vgs
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vgs {
    my ($self, $vgname) = @_;

    my ($cmd, $name, @online, @output, $status, $vgObj, %hDeport);

    $cmd = (defined ($vgname) and length ($vgname))
        ? "lsvg | grep $vgname | awk '{print \$1}'"        
            : "lsvg | grep -v root | awk '{print \$1}'";
    
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);                                                                    
    if ($status gt 0) {
        $self->_set_err("Error: NativeLvm::_create_vgs()",
                        "Unable to load existing diskgroups");
        
        return ($ClientUtil::FALSE);                      
    } 

    $cmd = "lsvg -o | grep -v root | awk '{print \$1}'";
    ($status, @online) = &ClientUtil::execute($self->{conn}, $cmd);                                                                    
    if ($status gt 0) {
        $self->_set_err("Error: NativeLvm::_create_vgs()",
                        "Unable to determine online diskgroups");
        
        return ($ClientUtil::FALSE);                      
    } 
    
    # Skip the header and rootvg lines
    foreach my $name (@output){
        $self->create_vg(-name => $name,
                         -deported => grep(/$name/,@online) ? 0 : 1);

        if ($self->error) {
            return ($ClientUtil::FALSE);
        }
    }

    return ($ClientUtil::TRUE);
    
}

################################################################################
# Subroutine Name: _get_disk_id
# Objective:       Return disk id
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_disk_id {
    my ($self, $device) = @_;

    my ($cmd, $name, @output, $status, $num); 

    $self->_clear_err();

    # get via native commands and not sanlun
    $cmd = "lscfg -vpl $device | grep -i 'Serial Number'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: NativeLvm::get_diskid() Unable to get diskid for device \'$device\'",
                         @output);
        return ($ClientUtil::FALSE);                       
    } 

    $num = $output[0];
    $num =~ s/\s+Serial Number\.+//;
    
    return ($num);
    
}

################################################################################
# Subroutine Name: _get_spd_devices
# Objective:       get all device id paths. Only valid for SANPath MP
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_spd_devices {
    my ($self, $spd_id) = @_;
    my ($cmd, $id, $status, @devices, @output);
    
    $cmd = "setsp -l" . $spd_id;      
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
        $self->_set_err("Error: AixNativeLvm::_get_spd_devices() Unable to get devices for spd_id \'$spd_id\'",
                        @output);
    
        return (wantarray()) ? () : [];                          
    }

    foreach my $line (@output) {
        next if ($line !~ /hdisk\d+\(/i);
        push @devices, $line =~ /(hdisk\d+)\(/;;
    }
    
    return (wantarray()) ? @devices : \@devices; 

}

################################################################################
# Subroutine Name: _get_spd_id
# Objective:       get sanpath device id. Only valid for SANPath MP
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_spd_id {
    my ($self, $device) = @_;

    my ($cmd, $id, $status, @output);

    $cmd = "cat /etc/sppath.conf | egrep \':$device\$\' | awk '{print substr(\$1,5)}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
        $self->_set_err("Error: AixNativeLvm::_get_spd_id() Unable to retrieve sanpath id for device \'$device\'", 
                        @output);
    
        return ""                          
    }

    $id = $output[0];

    return $id;
}

################################################################################
# Subroutine Name: _stop_spd_id
# Objective:       stop sanpath device. Only valid for SANPath MP
# Details:         See POD documentation at the end of this file
################################################################################
sub _stop_spd_id {
    my ($self, $spd_id) = @_;

    my ($cmd, $id, $status, @output);
    
    $cmd = "setsp -T -l" . $spd_id;      
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
        $self->_set_err("Error: AixNativeLvm::_stop_spd_id() Unable to stop sanpath device id\'$spd_id\'", 
                        @output);
        return ($ClientUtil::FALSE);
    
    }

    return ($ClientUtil::TRUE);
}

################################################################################
# Subroutine Name: _remove_spd_id
# Objective:       remove sanpath device. Only valid for SANPath MP
# Details:         See POD documentation at the end of this file
################################################################################
sub _remove_spd_id {
    my ($self, $spd_id) = @_;

    my ($cmd, $id, $status, @output);
    
    $cmd = "setsp -x -l" . $spd_id;      
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
        $self->_set_err("Error: AixNativeLvm::_remove_spd_id() Unable to remove sanpath device id\'$spd_id\'",
                        @output);
        return ($ClientUtil::FALSE);
    }

    return ($ClientUtil::TRUE);
    
}

################################################################################
# Subroutine Name: _get_pkginfo
# Objective:       Get lvm package info
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self) = @_;

    my ($cmd, $name, $vers, $rev, @buffer, @output, $status);

    $cmd = "lslpp -l | grep lvm | awk '{print \$2}'";
    ($status, @buffer) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: NativeLvm::get_version() Command ($cmd) returned a bad status ($status)");
        return ($ClientUtil::FALSE);                             
    } 

    push (@output, "NativeLvm version: " . $buffer[0]);

    $self->{name} = "NativeLvm";
    $self->{rev} = $buffer[0];
    $self->{vers} = "1.0";
    $self->{pkginfo} = \@output;

    return ($ClientUtil::TRUE);
    
}

1;
