# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the Native Voume Manager (LVM) on AIX
## @author knopp@netapp.com, thomasm@netapp.com
## @status private
## @pod ../../Vg.pod

package San::SanClient::Lvm::Aix::NativeLvm::Vg;

use strict;

use Tharn;
use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

require San::SanClient::Lvm::Aix::NativeLvm::Vg::Vol;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# Subroutine Name: add_device()
# Objective:       add a device to the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub add_device {
    my ($self, $device) = @_;

    my ($cmd, $disk_cnt, $name, @output, $status);
    
    $self->_clear_err();
    
    # Get name of volume group.         
    $name = $self->get_name();

    $cmd = "extendvg -f $name $device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: Vg::add_device Unable to extend volumegroup \'$name\'",
                        @output);
        return ($ClientUtil::FALSE);                            
    } 

    $self->{devices}->{$device} = $device;
    $self->{disk_cnt}++;

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: create_vol
# Objective:       create and return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vol {
    my ($self, %args) = @_;

    my ($volObj);

    # Change layout to concat if VG contains only 1 device
    if (exists $args{-layout} and length $args{-layout}) {
        if ($args{-layout} =~ /stripe/i and $self->get_device_cnt() eq "1") {
            $args{-layout} = "concat";
        }
    }

    $volObj = San::SanClient::Lvm::Aix::NativeLvm::Vg::Vol->_new(-parent=>$self,
                                                                 -conn=>$self->{conn},
                                                                 %args);

    if ($volObj->error) {
        $self->_set_err($volObj->errmsg());
        return ($volObj);
    }

    push (@{$self->{vols}}, $volObj);

    return ($volObj);

}

################################################################################
# Subroutine Name: deport()
# Objective:       deport Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub deport {
    my ($self) = @_;

    my ($cmd, @cmds, $name, @output, $status);
    
    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vg::deport() VolumeGroup name is a required parameter");
        return($ClientUtil::FALSE);      
    }

    @cmds = ("varyoffvg $name",
             "exportvg $name");

    foreach $cmd (@cmds) {
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: Vg::deport() Unable to execute command \'$cmd\'",
                            @output);
            return ($ClientUtil::FALSE);                           
        } 
    }

    # Remove volume objects from vg cache
    $self->{vols} = ();

    $self->{deported} = 1;
    
    return ($ClientUtil::TRUE);        

}

################################################################################
# Subroutine Name: destroy_vol
# Objective:       destroy a named vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vol {
    my ($self, $name) = @_;

    my ($volObj, @vols);
    
    $self->_clear_err();

    unless(defined $name and length $name) {
        $self->_set_err("Error: Vg::destroy_vol() volname is a required argument");
        return $ClientUtil::FALSE
    }
    
    # Delete the of vg object and update the parent vg list.
    foreach $volObj ( @{$self->{vols}}) { 
        if ($volObj->get_name() =~ /^$name$/) {
            $volObj->destroy(-force=>1);
            if ($volObj->error) { 
                $self->_set_err($volObj->errmsg());
                return ($ClientUtil::FALSE);
            }
        } else {
            push (@vols, $volObj);
        }
    }
    
    $self->{vols} = \@vols;

    return ($ClientUtil::TRUE);        

}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;
    
    my (@cmds, $cmd, $name, @devices, @output, $status, @vgs);
    
    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vg::destroy() VolumeGroup name is a required parameter");
        return ($ClientUtil::FALSE);
    }

    # agreed in code review to simply varyoff and export the vg instead
    # of going through the entire reduce process.
    if ($self->{deported}) {
        @cmds = ("exportvg $name");
    } else {
        @cmds = ("varyoffvg $name",
                 "exportvg $name");
    }

    foreach $cmd (@cmds) {
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: Vg::destroy() Unable to execute command \'$cmd\'",
                            @output);
            return ($ClientUtil::FALSE);
        }
    }

    # Remove the vgobj from the parent lvm object
    foreach my $vgObj (@{$self->{parent}->{vgs}}) { 
        if ($vgObj->get_name() =~ /^$name$/) {
            next
        } else {
            push (@vgs, $vgObj);
        }
    }
    
    $self->{parent}->{vgs} = \@vgs;
    @devices = (keys %{$self->{devices}});

    $self->{parent}->delete_devices(@devices);
    if ($self->{parent}->error) {
        $self->_set_err($self->{parent}->errmsg());
        return ($ClientUtil::FALSE);
    }

    return ($ClientUtil::TRUE);        

}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;

    my ($cmd, $header, $name, $status, $summary, $verbose);
    my (@output, @real_volnames, @summary, @verbose);
    
    $self->_clear_err();

    $verbose = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Error: Vg::display() ",
                                "Invalid value for argument '-verbose'");
                return (wantarray()) ? () : [];                            
            }        
            next;
        } else {
            $self->_set_err("Error: Vg::display()" ,
                            "Invalid parameter specified \'$_\'");
            return (wantarray()) ? () : [];                            
        }
    }  

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vg::display() VolumeGroup name is a required parameter");
        return (wantarray()) ? () : []; 
    }

    push (@summary, "VG Name:  " . $self->get_name());
    push (@summary, "Devices:  " . $self->get_device_cnt());

    $header = 
        sprintf("Vol Summary:\n".
                "%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-32.32s\n".
                "%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-32.32s",
                "NAME","LAYOUT","TYPE","STATE","SIZE(mb)","MOUNT_PT",
                "="x12,"="x7,"="x5,"="x12,"="x12,"="x32);
    
    push (@summary, $header);

    foreach my $volObj (@{$self->{vols}}) {
        $summary = 
            sprintf("%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-s",
                    $volObj->get_name(),
                    $volObj->get_layout(),
                    $volObj->get_fs(),
                    $volObj->get_state(),
                    $volObj->get_size(),
                    $volObj->get_mount());
        
        push (@summary, $summary);

        # Rather than make a seperate telnet call to display the 
        # detail for each volume, send all the commands at once.
        push (@real_volnames,
              "\\\n && echo '' && lslv -L " . $name . "_" . $volObj->get_name());
    }

    if (defined $verbose and $verbose =~ /1|on|y|yes|t|true/i) {
        $cmd = "lsvg -L $name @real_volnames";

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: Vg::display() Unable to display info for volumegroup \'$name\'",
                            @output);
            return (wantarray()) ? () : [];                            
        }

        # Commands that use the unix continuation character '\ ' will 
        # add a '> ' to the output.  This should be removed.
        my $output_wo_cont_chars = join ("\n", @output, "\n");
        $output_wo_cont_chars =~ s/\> //g;

        @verbose = (@summary, "\nVgroup Detail:", $output_wo_cont_chars);

        push (@verbose, "\n\n");

        return (wantarray()) ? @verbose : \@verbose;
    } else {
        push (@summary, "\n\n");

        return (wantarray()) ? @summary : \@summary;
    }

}

################################################################################
# Subroutine Name: get_device_cnt
# Objective:       Return number of devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device_cnt {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{disk_cnt}) {
        return $self->{disk_cnt};
    } else {
        $self->_set_err("Error: Vg::get_device_cnt() Unable to display disk count for volumegroup");
        return (0);                                 
    }                                         
    
}

################################################################################
# Subroutine Name: get_devices
# Objective:       Return devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;

    my (@output);

    $self->_clear_err();
    
    @output = (keys %{$self->{devices}});
    
    return (wantarray()) ? @output : \@output;        

}

################################################################################
# Subroutine Name: get_freespace
# Objective:       Return available free space
# Details:         See POD documentation at the end of this file
################################################################################
sub get_freespace {
    my ($self) = @_;

    my($cmd, $name, @output, $status, $size);
    
    $self->_clear_err();

    $size = 0;

    $name = $self->get_name();

    $cmd = "lsvg -L $name | grep 'FREE PPs'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: Vg::get_freespace() Unable to get free space (mb) for volumegroup \'$name\'",
                        @output);
        return(0);
    } 

    ($size) = $output[0] =~ /\((\d+)\s+/;
    
    return ($size);        
    
}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vg object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{name}) {
        return ($self->{name});
    } else {
        $self->_set_err("Error: Vg::get_name() Vg object is missing name");
        return ("");
    }

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size);
    
    $self->_clear_err();

    $size = 0;

    $name = $self->get_name();

    $cmd = "lsvg -L  $name |  grep 'TOTAL PPs'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: Vg::get_size() Unable to get size (mb) for volumegroup \'$name\'",
                         @output);
        return($ClientUtil::FALSE);                       
    } 

    ($size) = $output[0] =~ /\((\d+)\s+/;
    
    return ($size);        
    
}

################################################################################
# Subroutine Name: get_vol
# Objective:       return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol {
    my ($self,$name) = @_;

    my ($vgObj, @vols, $str);
    
    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Vg::get_vol() volume name is a required argument");
        return $ClientUtil::FALSE
    }

    @vols = $self->get_vols();
    foreach my $obj (@vols) {
        $str = $name;
        if ($obj->get_name() =~ /$str$/) {
            return $obj;
        }
    }
    
    $self->_set_err("Error: Vg::get_vol() Unable to find volume '$name'");
    return (wantarray()) ? () : [];                            

}

################################################################################
# Subroutine Name: get_vol_cnt
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol_cnt {
    my ($self, %args) = @_;

    my (@vols);
    
    $self->_clear_err();

    if (defined $self->{vols}) {
        # exclude the log logical volume used for JFS
        if (exists $args{-ignore_log}) {            
            @vols = $self->get_vols(-ignore_log => 1);
            return (scalar(@vols));
        } else {
            @vols = $self->get_vols();
            return (scalar(@vols));
        }
    } else {
        $self->_set_err("Error: Vg::get_vol_cnt() Vg object has no vol count");
        return (0);        
    }

    return (0);     

} 

################################################################################
# Subroutine Name: get_vols
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vols {
    my ($self, %args) = @_;

    my (@volObjs, $logObj);
    
    $self->_clear_err();

    if (defined $self->{vols}) {
        foreach my $volObj (@{$self->{vols}}) {
            if ($volObj->{type} =~ /jfslog/i or
                $volObj->{type} =~ /jfs2log/i) {
                $logObj =  $volObj;
            } else {
                push @volObjs, $volObj;
            }
        }
    } else {
        $self->_set_err("Error: Vg::get_vols() Vg object has no volumes");
        return ();        
    }

    # Check to see if jfslog file should not be included.
    if (not (exists $args{-ignore_log})) {            
        if (defined $logObj) {
            push (@volObjs, $logObj) 
        }
    }

    return (wantarray()) ? @volObjs : \@volObjs;
    
} 

################################################################################
# Subroutine Name: import()
# Objective:       Import Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub import {
    my ($self) = @_;

    my ($cmd, @devs, $name, @output, $status);
    
    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vg::import() VolumeGroup name is a required parameter");
        return ($ClientUtil::FALSE);  
    }

    @devs = (keys %{$self->{devices}});

    $cmd = "importvg -f -y $name ".$devs[0];
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: Vg::import() Unable to import volumegroup \'$name\'",
                        @output);
        return ($ClientUtil::FALSE);                           
    } 

    $self->{deported} = 0;

    # requery device list
    $self->{devices} = $self->_get_devices();
    if ($self->error) {
         return ($ClientUtil::FALSE);
    }

    $self->_create_vols(-_import => 1);
    if ($self->error) {
        return($ClientUtil::FALSE);
    }

    return ($ClientUtil::TRUE);        

}

################################################################################
# Subroutine Name: is_deported
# Objective:       Determine if volgroup has been deported
# Details:         See POD documentation at the end of this file
################################################################################
sub is_deported {
    my ($self) = @_;

    $self->_clear_err();
    
    return ($self->{deported});

}

################################################################################
# Subroutine Name: is_inline
# Objective:       Determine if volgroup uses inline logging
# Details:         See POD documentation at the end of this file
################################################################################
sub is_inline {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{inline}) {
        return ($self->{inline});
    } else {
        return (0);
    }

}

################################################################################
# Subroutine Name: refresh
# Objective:       refresh content of volume group.
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh {
    my ($self) = @_;

    my (@devs);

    # Remove existing vol cache
    $self->{vols} = undef;
    
    $self->sanlog(-msg => "Refreshing volume group '" . 
                          $self->get_name() . "'");
    @devs = $self->_get_devices();
    if ($self->error) { 
        # If we cannot get devices, assume it's deported
        $self->{deported} = 1;
        return ($ClientUtil::TRUE); 
    } else { 
        $self->{devices} = \@devs;
        $self->_create_vols();
        if ($self->error) {
            return ($ClientUtil::FALSE);
        }
    }

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: remove_device()
# Objective:       remove a device from the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub remove_device {
    my ($self, %args) = @_;

    my ($cmd, $device, $force, $disk_cnt, $disk_name, $name);
    my (@output, @vols_mount, $status);
    
    $self->_clear_err();

    $force = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Error: Vg::remove_device()",
                                "Invalid value for argument '-device'");
                return ($ClientUtil::FALSE);
            }        
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Vg::remove_device()",
                                "Invalid value for argument '-force'");
                return ($ClientUtil::FALSE);
            } 
            $force = "-d -f";
            next;
        } 
    }

    # get name of volume group        
    $name = $self->get_name();

    # verify the device is being used in volume group
    unless (exists $self->{devices}->{$device}) {
        $self->_set_err("Error: Vg::remove_device()",
                        "Device '$device' doesn't exist in volumegroup");
        return ($ClientUtil::FALSE);
    }

    $disk_name = $self->{devices}->{$device};

    # if no force option is specified umount all volumes so deport
    # can occur without error.  Make note of which vols were
    # mounted so they can be remounted after removing device
    if (not defined $force) {
        my @volObjs = $self->get_vols();

        return if ($self->error);

        foreach my $volObj (@volObjs) {                                        
            
            if ($volObj->is_fs()) {
                if ($volObj->is_mounted()) {
                    $volObj->unmount();
                    if ($volObj->error) {
                        $self->_set_err($volObj->errmsg());
                        return ($ClientUtil::FALSE);
                    }

                    push (@vols_mount, $volObj);
                }                
            }                
        }
    }


    # Check for the force flag being specified and add it to the 
    # rmdisk line
    if (defined ($force)) {
        $cmd = "reducevg $force $name $device";
    } else {
        $cmd = "reducevg $name $device";
    }
    
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: Vg::remove_device Unable to remove $device from volumegroup \'$name\'",
                        @output);
        return ($ClientUtil::FALSE);                            
    } 

    # remove device from hash
    delete ($self->{devices}->{$device});

    # decrement device counter;
    $disk_cnt = $self->{disk_cnt};
    $disk_cnt--;
    $self->{disk_cnt} = $disk_cnt;

    # Restart and mount volumes that were previously mounted.
    if (not defined $force) {
        foreach my $vol (@vols_mount) {
            $vol->_vol_start();
            if ($vol->error) {
                $self->_set_err($vol->errmsg());
                return ($ClientUtil::FALSE);
            }

            $vol->mount();
            if ($vol->error) {
                $self->_set_err($vol->errmsg());
                return ($ClientUtil::FALSE);
            }
        }
    }

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: rename()
# Objective:       rename volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, %args) = @_;

    my ($cmd, $force, $name, $newname, $status, @disks);
    my (@output, @vols_mount, @volObjs);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($newname = $args{$_})) {
                $self->_set_err("Error: Vg::rename()",
                                "Invalid value for argument '-name'");
                return ($ClientUtil::FALSE);
            }        
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Vg::rename()",
                                "Invalid value for argument '-force'");
                return ($ClientUtil::FALSE);
            }        
            next;
        } else {
            $self->_set_err("Error: Vg::rename()",
                            "Invalid parameter specified \'$_\'");
            return ($ClientUtil::FALSE);                               
        }
    }

    # get name of disk group        
    $name = $self->get_name();

    # verify volume group has a name
    unless (defined $name) {
        $self->_set_err("Error: Vg::rename()",
                        "Volume group doesn't have a name");
        return ($ClientUtil::FALSE);
    }

    # verify that a new volume group name was specified
    unless (defined $newname) {
        $self->_set_err("Error: Vg::rename()",
                        "New name for Volume group not specified");
        return ($ClientUtil::FALSE);
    }

    # if force option specified umount all volumes so deport
    # can occur without error.  Make note of which vols were
    # mounted so they can be remounted after rename occurs
    if (defined $force and $force =~ /1|on|y/i) {
        @volObjs = $self->get_vols();
        return if ($self->error);

        foreach my $volObj (@volObjs) {                                        
            
            if ($volObj->is_fs()) {
                if ($volObj->is_mounted()) {
                    
                    # retrieve current mount path
                    # before unmounting the filesystem.
                    my $path = $volObj->get_mount();

                    # Remove the vg and vol path info
                    my $vgvol = "/" . $name . "/" . $volObj->get_name();
                    $path =~ s/$vgvol//g;
                    
                    # Unmount the filesystem
                    $volObj->unmount();
                    if ($volObj->error) {
                        $self->_set_err($volObj->errmsg());
                        return ($ClientUtil::FALSE);
                    }
                    
                    # Store info on volumes 
                    my $hVols = {};
                    $hVols->{volObj} = $volObj;
                    $hVols->{path} = $path;
                    push (@vols_mount, $hVols);
                }                
            }                
        }
    }

    # Deport the volumegroup
    $self->deport();
    return $ClientUtil::FALSE if ($self->error);

    # Get devices used by volume group
    @disks = $self->get_devices();
    return $ClientUtil::FALSE if ($self->error);

    # Rename the volumegroup
    $cmd = "recreatevg -y $newname @disks";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: Vg::rename Unable to import '$name' under a new volumegroup \'$newname\'",
                        @output);
        return ($ClientUtil::FALSE);                            
    } 

    $self->{name} = $newname;

    $self->{deported} = 0;

    # Update the volume name prefix with the new
    # volume group name.

    foreach my $volObj (@volObjs) {        
        my $old_name = "fs" . $name . "_" . $volObj->get_name();

        $cmd = "chlv -n $newname" . "_" . $volObj->get_name() . " $old_name";        
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status gt 0) {
            $self->_set_err("Error: Vg::rename Unable to import '$name' under a new volumegroup \'$newname\'",
                             @output);
            return ($ClientUtil::FALSE);                            
        } 
    }

    # Mount filesystems on volumes that were previously mounted 
    # before renaming volume group.
    foreach my $hVol (@vols_mount) {
        my $volObj = $hVol->{volObj};
        $volObj->mount(-path => $hVol->{path});                
        if ($volObj->error) {
            $self->_set_err($volObj->errmsg());
            return ($ClientUtil::FALSE);
        }
    }

    return ($ClientUtil::TRUE);

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _create_vols
# Objective:       create vol object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vols {
    my ($self, %args) = @_;

    my ($cmd, $import, $vgname, $name, @names, @output);
    my ($status, $vgObj, $volObj, $log_lv, $pp_size); 

    # Identify volumes that are being imported so
    # that their automatically mounted with their
    # last filesystem mount point.
    $import = (exists $args{-_import})
               ? $args{-_import} 
               : 0;

    $name = $self->get_name();
    $log_lv = $name . "_log";

    # get PP size for the VG/LV
    $cmd = "lsvg $name | grep -i \"PP SIZE:\"";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: Vg::_create_vols()",
                        "Unable to get PP size from VG ($name)",
                        @output);
        return ($ClientUtil::FALSE);               
    } 

    $pp_size = (split(/:/, $output[0]))[2];
    $pp_size =~ s/^\s+(\d+)\s+(\S+)//;
    $pp_size = $1;
    
    # Get list of volumes (ignore log LV)
    $cmd = "lsvg -l $name | egrep -v '$name:|LV NAME|$log_lv' | awk '{print \$1,\$2,\$4}'";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: Vg::_create_vols()",
                        "Unable to list available disk group",
                        "status=$status",
                        @output);
        return ($ClientUtil::FALSE);               
    } 

    foreach my $line (@output) {
        my $prefix = "$name" . "_";

        my ($volname, $type, $size) = split(/\s+/, $line);

        # Multiply the Num PPs ($4) by the PP size (in MB)
        my $sz = $size * $pp_size;

        $volObj = $self->create_vol(-name => $volname,
                                    -size => "$sz" . "m",
                                    -type => $type,
                                    -exist => 1,
                                    -_import => $import);

        if ($self->error) {
            return ($ClientUtil::FALSE);
        }
    }

    return ($ClientUtil::TRUE); 

}

################################################################################
# Subroutine Name: _create()
# Objective:       create a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub _create{
    my ($self, $name, $devices) = @_;

    my ($cmd, @cmds, @output, $status, %hDevice);

    # if device count is greater (or equal to) 32, then
    # use flag to allow for more devices in VG (-B)
    if (scalar(@$devices) >= 32) {
        @cmds = ("mkvg -f -B -y $name @$devices",
                 "chvg -s y $name",
                 "chvg -Q n $name",
                 "varyoffvg $name",
                 "varyonvg $name");
    } else {
        @cmds = ("mkvg -f -y $name @$devices",
                 "chvg -s y $name",
                 "chvg -Q n $name",
                 "varyoffvg $name",
                 "varyonvg $name");
    }

    foreach $cmd (@cmds) {         
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status gt 0) {
            $self->_set_err("Error: Vg::_create Unable to execute command: ",
                            $cmd, @output);
            return ($ClientUtil::FALSE);                            
        } 
    }
    
    foreach my $device (@$devices) {
        $hDevice{$device} = $device; 
    }
    
    $self->{devices} = \%hDevice;
    $self->{disk_cnt} = scalar(@$devices);

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: _get_devices
# Objective:       get list of devices used by disk group
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices {
    my ($self) = @_;

    my ($cmd, $name, $status, @output, %hDevice);

    $name = $self->get_name();
    
    $cmd = "lsvg -p $name | grep hdisk | awk '{print \$1}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: Vg::_get_devices() Unable to get devices for volgroup \'$name\'",
                        @output);
        return {};                          
    }

    $self->{disk_cnt} = scalar(@output);
    
    foreach my $device (@output) {                        
        $hDevice{$device} = $device;           
    }

    return (\%hDevice);
    
}

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd,$conn, $devices, $disk_cnt, $jlog, $parent, $log, $name);
    my ($self, $status, $deported, @output, $type, $value, $fs);

    $self={};

    bless ($self, $class);        

    $self->_clear_err();

    $deported = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: Vg::_new()",
                                "Invalid value for argument '-devices'");
                return ($self);
            }        
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: Vg::_new()",
                                "Missing value for argument '-name'");
                return ($self);
            }
            next;
        } elsif (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: Vg::_new()",
                                "Missing value for argument '-conn'");
                return ($self);
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Vg::_new()",
                                "Missing value for argument '-lvmobj'");
                return ($self);
            }
            next;
        } elsif (/-?deported/) {
            if (not length($deported = $args{$_})) {
                $self->_set_err("Error: Vg::_new()",
                                "Missing value for argument '-deported'");
                return ($self);
            }
            next;
        }
    }  
    
    # check for existance of required -lvm (parent) argument
    unless (defined $parent) {
        $self->_set_err("Error: Vg::_new()",
                        "Missing required argument '-parent'");
        return ($self); 
    }

    # check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: Vg::_new()",
                        "Missing required argument '-name'");
        return ($self); 
    }

    # check for existance of required -conn argument
    if (grep ( /^-?conn$/, (keys %args) )) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: Vg::_new()",
                            "Invalid value for argument '-conn'");
            return ($self); 
        }        
    } else {
        $self->_set_err("Error: Vg::_new()",
                        "Missing required argument '-conn'");
        return ($self); 
    }

    # initilize the object members
    $self->{conn} = $conn;
    $self->{name} = $name;
    $self->{deported} = $deported;
    $self->{disk_cnt} = 0;
    $self->{parent} = $parent;
    
    # create the volumegroup if devices were supplied
    if (defined $devices and ref($devices) eq "ARRAY") {
        $self->_create($name, $devices);                
        return $self if ($self->error());

        # create a journal log LV for possible use (if jfs)
        # will not use for JFS2
        $type = "jfslog";

        $self->create_vol(-name => "log",
                          -size => "50m", 
                          -layout => "concat",
                          -type => $type);
        return ($self) if ($self->error());

        # save response to logform prompt
        $cmd = "echo 'y' > /tmp/yes";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: Vg::new Unable to create input file ($cmd)");
            return($self);
        } 

        # run logform to format the logging volume
        $cmd = "/usr/sbin/logform /dev/$name" . "_log < /tmp/yes";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: Vg::new()",
                            "Error running $cmd to format the logging volume");
            return($self);
        }
    } else {
        # vg must already exist. set device list
        # for all online VGs.
        if (!$self->{deported}) {
            $self->{devices} = $self->_get_devices();
            return ($self) if ($self->error);        
            $self->_create_vols();
            return $self if ($self->error);
        }

    }

    return ($self);

}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Vg.pm

=head1 Authors

=begin html 

Gary Knopp<br>
Tom Masterson

=end html

=head1 Description

Vg provides an API interface to LVM volume groups through the use of the Telnet library.  

=head1 Public Methods

=head2 B<add_device($name)>

Add a new device to the volume group

=over 4

=item Inputs

=item * $name

The name of the device.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Add a new device to volume group");
    $vgObj=$Lvm->get_vg("vg1");
    if ($Lvm->error){
                $Test->log($Lvm->errmsg());
                return $TCase::FAIL;
    }

        $vgObj->add_device("c3t0d8");
         if ($vgObj->error){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }

=back

=head2 B<create_vol(%args)>

Creates a new volume group object.

=over 4

=item Inputs

=item * -name

The name of the volume

=item * -size (note: specify units 500k, 1m, 2g)

The size of the volume (note: specify units 500k, 1m, 2g)

=item * I<-layout> (default: concat)

The layout of the volume on the devices

=item Outputs

Returns a volume object. If an error occurs the error flag is set, errmsg is
set to the failure reason.  

=item Examples

    # Create a volume Group object.
        $Test->log("-- Creating volume group 'myvg'");
        $volObj = $vgObj->create_vol(-name=>"vol_A",-size=>'100m',-layout=>'stripe');
        if ($vgObj->error){
            $Test->log($vgObj->errmsg());        
            return $TCase::FAIL;
        }

=back

=head2 B<deport()>

Import a volume group 

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Deport a volume group");

        $vgObj->deport();
    if ($vgObj->error()){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }

=back

=head2 B<destroy()>

Destroy the volume group 

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Destroy volume group");

        $vgObj->destroy();
    if ($vgObj->error()){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }

=back

=head2 B<destroy_vol($name)>

Destroy a volume object by name

=over 4

=item Inputs

=item * $name

The name of the volume object to destroy

=item Outputs

Returns a scalar value of 1 if successful.  If the volume group does not exist or 
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is 
passed back as the return value.


=item Examples

    $Test->log("-- Destroy volume object 'vol_A'");
    if (not $vgObj->destroy_vol("vol_A")){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }

=back

=head2 B<display()>

Display detailed volume group information

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM version information");

    $Test->log($vgObj->display());
    if ($vgObj->error){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }

=back

=head2 B<error()>

This method allows you to determine if any errors occured in the
previous method call.  Note that the error flag is reset at the
beginning of each method, so it should be checked after B<every>
major method call.  If you do not check for errors, your calling
program may fail at some point and you will not know why.

=over 4 

=item Inputs

None

=item Outputs

Returns 1 if an error occured in the previous call, 0 otherwise.

=item Examples

 # Determine if error occurred...
 if ($obj->error) {
    print "Error encountered in previous method call\n";
 }

=back

=head2 B<errmsg()>

This method allows you to get the text of the error message if an error
occurred on the previous method call.

=over 4

=item Inputs

None

=item Outputs

Returns the text of the last error message set.  If no error message is
set, an empty string will be returned.

=item Examples

 # Check for errors.
 if ($obj->error) {
    $errmsg = $obj->errmsg;
    print "An error occurred: $errmsg\n";
 }

=back

=head2 B<get_devices()>

List of devices used  by volume group

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of devices if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and an empty array is returned.        


=item Examples

    $Test->log("-- Retrieving disk group devices");
    @devs=$vgObj->get_devices();
    if ($vgObj->error){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }  
    foreach $dev (@devs){
                print $dev."\n";
        }

=back

=head2 B<get_device_cnt()>

Number of devices used by volume group

=over 4

=item Inputs

none
                                                                           
=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is returned.        


=item Examples

    $size=$vgObj->get_size();
    if ($vgObj->error){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }  
   
    print "VG size is $size mb\n";

=back


=head2 B<get_freespace()>

Estimate of free disk space (megabytes) in volume group

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is returned.        


=item Examples

    $size=$vgObj->get_freespace();
    if ($vgObj->error){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }  
   
    print "VG available space is $size mb\n";

=back

=head2 B<get_name()>

Name of volume group

=over 4

=item Inputs

none

=item Outputs

Returns the name of the volume group.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Volume group name is '".$vgObj->get_name()."'";

=back


=head2 B<get_size()>

Estimate of volume group size in Megabytes

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is returned.        


=item Examples

    $size=$vgObj->get_size();
    if ($vgObj->error){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }  
   
    print "VG size is $size mb\n";

=back

=head2 B<get_vol($name)>

Retrieve a volume object by name

=over 4

=item Inputs

=item * $name

The name of the volume to retrieve

=item Outputs

Returns a volume object if successful.  If the volume does not exist or 
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is 
passed back as the return value.


=item Examples

    $Test->log("-- Retrieving vol object 'vol_A'");
    $volObj=$vgObj->get_vol("vol_A");
    if ($Lvm->error){
                $Test->log($Lvm->errmsg());
                return $TCase::FAIL;
    }

=back

=head2 B<get_vols()>

Retrieve a list of all volumes in  volume group

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of volumes if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and an empty array is returned.        


=item Examples

    $Test->log("-- Retrieving volume objecs objects");
    @volObjs=$vgObj->get_vols();
    if ($vgObj->error){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }  
    

=back

=head2 B<get_vol_cnt()>

Number of volumes in the volume group

=over 4

=item Inputs

none
                                                                           
=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is returned.        


=item Examples

    $cnt=$vgObj->get_vol_cnt();
    if ($vgObj->error){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }  
   
    print $vgObj->get_name()." has is $cnt volumes\n";

=back



=head2 B<import()>

Import a volume group 

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Import a volume group");

        $vgObj->import();
    if ($vgObj->error()){
                $Test->log($vgObj->errmsg());
                return $TCase::FAIL;
    }

=back

=head2 B<is_deported()>

Determine if a volume group is deported 

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if deported, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Examples

        if($vgObj->is_deported()){
                $vgObj->import();
                ...
                <error handler>
                ...
        }
   

=back

=head2 B<refresh()>

Refresh volume group volume objects.  This method will rebuild its 
volume object cache.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("Refresh volume group");
    if (not $vgObj->refresh()) {
    if ($vgObj->error){
        $Test->log($vgObj->errmsg());
    }

=back

=head2 B<remove_device(I<%args>)>

Removes a device from the volume group.  

=over 4

=item Inputs

=item * -device

The name of the device.        

=item * I<-force> (values: 1 | on | yes )         

Remove the device without unmounting any filesystems that exist.  Without the force
option, all filesystems that are mounted will be unmounted during the device removal.
Upon success the filesystems that were previously mounted will be remounted if they 
were not effected by the device removal.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

   $vgObj->remove_device(-device=>$device,-force=>1);
   if ($vgObj->error){
       $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
   }


=back

=head2 B<rename(I<%args>)>

Rename the volume group

=over 4

=item Inputs

=item * -name

The new name of the volume group.        

=item * I<-force> (values: 1 | on | yes )         

The force option will automatically locate and unmount all filesystems that are mounted 
so that a volume group rename can occur without error.  Upon success the filesystems 
that were previously mounted will be remounted.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

   $vgObj->rename(-name=>"myvg_new",-force=>1);
   if ($vgObj->error){
       $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
   }


=back
