# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the NativeLvm logical volumes
## @author knopp@netapp.com, thomasm@netapp.com
## @status private
## @pod ../../../Vol.pod
 
package San::SanClient::Lvm::Aix::NativeLvm::Vg::Vol;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use File::Basename;
use Data::Dumper;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
# Method:    create_fs
# Objective: create a filesystem on a volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_fs {
    my ($self, %args) = @_;

    my ($cmd, $type, $vgname, $name, $status, @output);

    $vgname = $self->get_vgname();
    $name = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?type$/) {
            $type = $args{$_};
            next;
        }
    } 
    
    # Default to jfs2 filesystems if not specified
    if (not defined($type) or $type eq "") {
        $type = "jfs2";
    }

    # set inline logging if jfs2 is the filesystem
    if ($type =~ /^jfs2$/) {
        $self->{inline} = 1;
        $self->{jlog} = "inline";
    } else {
        $self->{inline} = 0;
        $self->{jlog} = "external";
    }

    if ($type =~ /^jfs$|^jfs2$/i) {
        # Create a filesytem using requested journal logging mechanism
        if ($self->{inline} == 1) {
            $cmd = "yes | mkfs -V $type -o log=INLINE,bf=true /dev/$name";
        } else {
            $cmd = "yes | mkfs -V $type -o bf=true,log=/dev/$vgname" . 
                   "_log /dev/$name";
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status gt 0) {
            $self->_set_err("Error: Vol::create_fs Unable to create $type filesystem on volume '$name'",
                            @output);                        
            return ($ClientUtil::FALSE);  
        } 
        
        $cmd = "chlv -t $type $name ";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status gt 0) {
            $self->_set_err("Error: Vol::create_fs Unable to change filesystem type to $type volume '$name'",
                             @output);
            return ($ClientUtil::FALSE);  
        } 
    } else {
        $self->_set_err("Error: Vol::create_fs()",
                        "Invalid Filesystem type '$type'");
        return ($ClientUtil::FALSE);  
    }

    $self->{filesystem} = $type;

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self, %args) = @_;
    
    my ($cmd, $force, $vgname, $name, @output);
    my ($status, @vols, $real_vol_name);
    
    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Vol::destroy()",
                                "Invalid value for argument '-force'");
                return ($ClientUtil::FALSE);
            }        
            next;
        } 
    }

    $vgname = $self->get_vgname();
    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vol::destroy() VolumeGroup name is a required parameter");
        return ($ClientUtil::FALSE);
    }

    # The jfslog can't be deleted if its being used for
    # logging by another mounted filesystem
    return ($ClientUtil::TRUE) if ($self->{type} =~ /jfslog/i);

    # if force option specified umount all volumes so deport
    # can occur without error.  Make note of which vols were
    # mounted so they can be remounted after rename occurs
    if (defined $force and $force =~ /1|on|y/i) {
        if ($self->is_fs()) {
            if ($self->is_mounted()) {
                $self->unmount();
                return ($ClientUtil::FALSE) if ($self->error());
            }        
        }        
    }

    $cmd = "rmlv -f $name";
    
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);                                  
    if ($status gt 0) {
        $self->_set_err("Error: Vol::destroy() Unable to destroy volume \'$name\' in volumegroup '$vgname'",
                        $cmd, @output);
        return ($ClientUtil::FALSE);  
    } 

    # Remove the volume from the parent vol group object
    foreach my $volObj (@{$self->{parent}->{vols}}) { 
        if ($volObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push (@vols, $volObj);
        }
    }
    
    $self->{parent}->{vols} = \@vols;
    
    return ($ClientUtil::TRUE);    
    
}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;

    my ($cmd, $header, $name, $status, $summary);
    my ($verbose, $vgname, @buf, @output);
    
    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Error: Vol::display()",
                                "Invalid value for argument '-verbose'");
                return (wantarray()) ? () : []; 
            }        
            next;
        } else {
            $self->_set_err("Error: Vol::display()",
                            "Invalid parameter specified \'$_\'");
            return (wantarray()) ? () : []; 
        }
    }  

    $vgname = $self->get_vgname();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vol::display() volume name is a required parameter");
        return (wantarray()) ? () : []; 
    }

    $header = 
        sprintf("\nVol Summary:\n".
                "%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-s\n".
                "%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-32.32s",
                "NAME","LAYOUT","TYPE","STATE","SIZE(mb)","MOUNT_PT",
                "="x12,"="x7,"="x5,"="x12,"="x12,"="x32);

    $summary = 
        sprintf("%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-s",
                $self->get_name(),
                $self->get_layout(),
                $self->get_fs(),
                $self->get_state(),
                $self->get_size(),
                $self->get_mount());
    
    push (@buf, $header, $summary);

    if (defined $verbose and $verbose =~ /1|on|y|yes|t|true/i) {
        $cmd = "lslv -L $name";

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);                                  
        if ($status gt 0) {
            $self->_set_err("Error: Vol::display() Unable to display info for volume \'$name\' in volume group \'$vgname\'", 
                            @output);
            return (wantarray()) ? () : [];                            
        } 

        push (@buf, "\nVol Detail:", @output);
     }

     return (wantarray()) ? @buf : \@buf;         

}

################################################################################
# Subroutine Name: get_fs
# Objective:       Return Vol object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{filesystem}) and length($self->{filesystem})) {
        return ($self->{filesystem});
    } else {
        $self->_set_err("Error: Vol::get_fs() Vol object is missing filesystem");
        return "";
    }

}

################################################################################
# Subroutine Name: get_layout
# Objective:       Return volume layout
# Details:         See POD documentation at the end of this file
################################################################################
sub get_layout {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{layout})) {
        return ($self->{layout});
    } else {
        $self->_set_err("Error: Vol::get_layout() Vol object is missing layout");
        return "";
    }

}

################################################################################
# Subroutine Name: get_mount
# Objective:       Return mount point of volume
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{mount}) and length($self->{mount})) {
        return ($self->{mount});
    } else {
        $self->_set_err("Error: Vol::get_mount() Vol filesystem is not mounted");
        return "";
    }

}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vol object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;
    
    $self->_clear_err();
    
    if (defined($self->{name})) {
        return ($self->{name});
    } else {
        $self->_set_err("Error: Vol::get_name() Vol object is missing name");
        return "";
    }

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return ooject size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{size})) {
        return ($self->{size});
    } else {
        $self->_set_err("Error: Vol::get_size() Vol object is missing size");
        return "";
    }

}

################################################################################
# Subroutine Name: get_state
# Objective:       Return object state
# Details:         See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{state})) {
        return ($self->{state});
    } else {
        $self->_set_err("Error: Vol::get_state() Vol object is missing state");
        return "";
    }

}

################################################################################
# Subroutine Name: get_vgname
# Objective:       Return Vol object volumegroup name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgname {
    my ($self) = @_;

    my ($name);

    $self->_clear_err();
    
    $name = $self->{parent}->get_name();

    if (defined($name) and length($name)) {
        return ($name);
    } else {
        $self->_set_err("Error: Vol::get_vgname() Vol object is missing parent name");
        return "";
    }

}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       Return path to raw vol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    $self->_clear_err();

    # get name of volume    
    my $name = $self->get_name();

    return "/dev/r$name";

}

################################################################################
# Subroutine Name: is_fs
# Objective:       Determine if volume has a filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub is_fs {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{filesystem}) and $self->{filesystem} !~ /raw/i) {
        return ($ClientUtil::TRUE);
    } else {
        return ($ClientUtil::FALSE);
    }

}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if volume is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined($self->{mount}) and length($self->{mount})) {
        return ($ClientUtil::TRUE);
    } else {
        return ($ClientUtil::FALSE);
    }

}

################################################################################
# Subroutine Name: mount
# Objective:       mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub mount {
    my ($self, %args) = @_;

    my ($cmd, $type, $directio, $pass_thru, $vgname, 
        $mount, $name, $status, @output);

    # Get attributes 
    $vgname = $self->get_vgname();
    $name = $self->get_name();
    $type = $self->get_fs();
    $mount = undef;
    $directio = 0;
    $pass_thru = "";

    # check if volume is unformatted
    if (defined $type and $type =~ /raw/) {
        $self->_set_err("Error: Vol::mount()",
                        "Cannot mount a raw filesystem");
        return ($ClientUtil::FALSE); 
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?path$/) {
            if (not length($mount = $args{$_})) {
                $self->_set_err("Error: Vol::mount()",
                                "Invalid value for argument '-path'");
                return ($ClientUtil::FALSE);
            }        
            next;
        } elsif (/-?directio$/) {
            if (not length($directio = $args{$_}) or $args{$_} !~ /0|1/) {
                $self->_set_err("Error: Dev::mount_fs()",
                                "Invalid value for argument '-directio'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?pass_thru$/) {
            $pass_thru = $args{$_};
            $pass_thru =~ s/^\s+//;
            $pass_thru =~ s/\s+$//;
            if ($pass_thru ne "") {
                $pass_thru = "," . $pass_thru;
            }
            next;
        }
    }

    # check if the volume is already mounted
    if ($self->is_mounted()) {
        $self->_set_err("Error: Vol::mount() volume '$name' is already mounted");                    
        return ($ClientUtil::FALSE);                            
    }
    
    # Use default path if not defined
    unless (defined $mount) {
        $mount = "/mnt/qa/$vgname/" . $self->{basename};
    } else {
        $mount .= "/$vgname/" . $self->{basename};
    }
    
    $self->{mount} = $mount;

    # Save the mount point with the vol metadata in case
    # a vol is renamed while mounted
    $self->_set_metadata("LABEL", $mount);
    return ($ClientUtil::FALSE) if ($self->error); 

    # Create the mount point and open up dir privs 
    $cmd = "mkdir -p $mount; chmod 777 $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: Vol::mount() Unable to create mount point directory for volume '$name'",
                         @output);
        return ($ClientUtil::FALSE);                            
    } 

    # Mount the volume's filesystem
    if ($self->{inline} == 1 or $self->{jlog} =~ /inline/i) {
        if ($directio == 1) {
           $cmd = "mount -V $type -o log=INLINE${pass_thru} /dev/$name $mount";
        } else {
           $cmd = "mount -V $type -o log=INLINE${pass_thru} /dev/$name $mount";
        }
    } else {
        if ($directio == 1) {
            $cmd = "mount -V $type $pass_thru -o dio,log=/dev/$vgname${pass_thru}" . 
                   "_log /dev/$name $mount";
        } else {
            $cmd = "mount -V $type $pass_thru -o log=/dev/$vgname${pass_thru}" . 
                   "_log /dev/$name $mount";
        }
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: Vol::mount() Unable to mount volume filesystem",
                         @output);
        return ($ClientUtil::FALSE);                            
    } 
    
    $self->{state} = $self->_init_state();
    return ($ClientUtil::FALSE) if ($self->error);

    # Get handle to client object 
    $self->{client} = $self->{parent}{parent}{parent};        

    # Commenting out this code and want it left in.
    # Some systems can hang because of stale entries in /etc/filesystems
    # when abnormal failures in prior tests occur. 
    #$self->{client}->mount_entry_add(-device_mount => "/dev/$name",
    #                                 -mount_pt => $mount,
    #                                 -fs_type => $type,
    #                                 -mount_options => 
    #                                  ($self->{jlog} =~ /inline/i) ? 
    #                                   "INLINE" : 
    #                                   "/dev/$vgname" . "_log");

    #if ($self->{client}->error) {
    #    $self->_set_err($self->{client}->errmsg());
    #    return ($ClientUtil::FALSE);
    #}   

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: rename()
# Objective:       rename volume
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, $newname) = @_;

    my ($bMounted, $cmd, $name, $vgname, @output, $status); 

    $self->_clear_err();

    $bMounted = 0;

    # get name of volume group
    $vgname = $self->get_vgname();

    # get name of volume    
    $name = $self->get_name();

    # verify the device is being used in volume group
    unless (defined $name) {
        $self->_set_err("Error: Vol::rename()",
                        "Volume doesn't have a name");
        return ($ClientUtil::FALSE);
    }

    # Don't allow volumes of type 'jfslog' to be modified.
    # Although this is supported 
    if ($self->{type} =~ /jfslog/i) {
        $self->_set_err("Error: Vol::rename()",
                        "Names of volumes used for log cannot be changed");
        return ($ClientUtil::FALSE);
    }

    # verify the device is being used in volume group
    unless (defined $newname) {
        $self->_set_err("Error: Vol::rename()",
                        "New name for Volume not specified");
        return ($ClientUtil::FALSE);
    }

    # Volume must be unmounted in order to rename it 
    if ($self->is_fs()) {
        if ($self->is_mounted()) {
            # flag to inform method of state before name change
            $bMounted = 1;        
            $self->unmount(); 
            return ($ClientUtil::FALSE) if ($self->error);
        }
    }


    # Rename the volume
    $cmd = "chlv -n $vgname" . "_" . $newname . " " . $name;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: Vol::rename Unable to rename volume '$name' to \'$newname\'",
                        @output);
        return ($ClientUtil::FALSE);                            
    } 

    $self->{basename} = $newname;
    $self->{name} = $vgname . "_" . $newname;

    # Remount the volume if it was mounted previously
    if ($bMounted eq 1) {
        $self->mount();
        return ($ClientUtil::FALSE) if ($self->error);
    }

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: rescan()
# Objective:       rescan volume
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;

    my ($mount);

    # Currently just handles remounting.   
    # Needed after snapdrive restore
    $self->_clear_err();
    
    $mount = $self->_init_mount();
    if (defined $mount and length $mount) {
        $self->mount(-path => $mount);
    } else {
        $self->{mount} = undef;
    }

    return ($ClientUtil::FALSE) if ($self->error);

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: unmount()
# Objective:       unmount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount {
    my ($self, %args) = @_;

    my ($cmd, $type, $vgname, $fs, $mount, $name, $status);
    my ($force, $option, @output);

    $self->_clear_err();

    $option = "";
    
    # parse hash
    foreach (keys %args) {
        if (/-?force$/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Vol::unmount()",
                                "Invalid value for argument '-force'");
                return ($ClientUtil::FALSE);
            }        
            next;
        } 
    }  

    $option = "-f" if (defined $force);

    # Get attributes 
    $vgname = $self->get_vgname();
    $name = $self->get_name();
    $mount = $self->get_mount();
    $fs = $self->get_fs();

    # check if volume contains a filesystem
    unless ($fs !~ /raw/i) {
        $self->_set_err("Error: Vol::unmount()",
                        "Cannot unmount a raw device");
        return ($ClientUtil::FALSE); 
    }

    # check if volume is mounted
    unless (defined $mount and length($mount)) {
        $self->_set_err("Error: Vol::unmount()",
                        "Cannot unmount a filesystem that is not mounted");
        return ($ClientUtil::FALSE); 
    } 

    # Unmount the volume's filesystem
    $cmd = "umount $option $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: Vol::unmount() Unable to unmount volume's filesystem",
                        @output);
        return ($ClientUtil::FALSE);                            
    } 

    # Remove mount pt metadata from object
    $self->{mount} = "";
    $self->{state} = $self->_init_state();
    return ($ClientUtil::FALSE) if ($self->error);

    # Clear vfstab mount entry
    # client = lvm <-- vg <-- vol
    $self->{client} = $self->{parent}{parent}{parent};

    # Commenting out this code and want it left in.
    # Some systems can hang because of stale entries in /etc/filesystems
    # when abnormal failures in prior tests occur.
    #$self->{client}->mount_entry_delete($mount);
    #if ($self->{client}->error) {
    #        $self->_set_error($self->{client}->errmsg());
    #        return ($ClientUtil::FALSE);
    #}   

    # Remove directory head
    my ($base, $parent) = fileparse($mount);
    if ($mount ne "/") {
       $cmd = "rm -rf $mount";
       ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);                                      
       if ($status gt 0) {
            $self->_set_err("Error: Vol::unmount() Unable to remove mount point '$mount'", 
                            @output);
            return ($ClientUtil::FALSE); 
       } 
    } else {
       $self->_set_err("Error: Vol::unmount() mount point '$mount' was not appropriate formed to remove", 
                       @output);
       return ($ClientUtil::FALSE); 
    }

    return ($ClientUtil::TRUE);

}

################################################################################
#  P R I V A T E   S U B R O U T I N E S
################################################################################

################################################################################
# Subroutine Name: _create()
# Objective:       create a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self) = @_;

    my ($cmd, $device_cnt, $dg, $layout, $name);
    my ($size, $status, $type, @devs, @output);

    $dg = $self->{parent}->get_name();
    $device_cnt = $self->{parent}->get_device_cnt();
    $name = $self->get_name();
    $size = $self->get_size();
    $layout = $self->{layout};
    $type = $self->{type};

    @devs = (keys %{$self->{parent}->{devices}});

    $cmd = "mklv -y $name ";
    if ($device_cnt > 1 and $layout =~ /stripe/i) {
        $cmd .= "-S " . $self->{stripe_size} . " ";
    }
    $cmd .= "-t $type -v y $dg $size @devs";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status gt 0) {
        $self->_set_err("Error: Vol::_create Unable to create volume \'$name\'",
                        @output);
        return ($ClientUtil::FALSE);                            
    } 

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: _get_metadata
# Objective:      Retrieve metadata for attribute 
# Details:        See POD documentation at the end of this file
################################################################################
sub _get_metadata{
    my ($self, $attribute) = @_;
    
    my ($cmd, $line, $vgname, $name, @output, $status, $value, $str);
    
    $value = "";

    $vgname = $self->get_vgname();
    $name = $self->get_name();

    $str = $name;
    $cmd = "lslv -L $str | grep '$attribute': ";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($self->error) {
        $self->_set_err("Error: Vol::_get_metadata() ",
                        "Unable to retrieve volume \'$name\' metadata for",
                        "attribute '$attribute' in volumegroup '$vgname'",
                        @output);
        return ""; 
    }

    # return the attribute value if exists.
    if (defined $output[0]) {
        ($value) = $output[0] =~ /$attribute\:\s+(\S+)/;
        return "" if ($value =~ /none/i);
    }
    
    return ($value);    
    
}

################################################################################
# Subroutine Name: _init_fstype
# Objective:      Initialize fstype for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_fstype {
    my ($self) = @_;
    
    my ($cmd, $fstype, $vgname, $name, @output, $status);
    
    $vgname = $self->get_vgname();
    $name = $self->get_name();

    # redirect to /dev/null since this output could be large
    # and we are not looking for the output to validate, just the
    # return code.
    $cmd = "dumpfs /dev/$name > /dev/null";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    # check for corruption 
    if (grep(/Cannot read superblock/i, @output) > 0) {
        $self->_set_err("Error: Vol::_init_fstype(): Error executing comand ($cmd)");
        # return "" since that will trigger _new() to clear_err 
        # and move on to getting the mount point (if this is a filesystem)
        return ""; 
    }
    
    if ($status gt 0) {
        $fstype = "raw";
    } else {
        $fstype = $self->{type};    
    }

    return ($fstype);    
    
}

################################################################################
# Subroutine Name: _init_jlog
# Objective:      Initialize journal log type for exist host volume.
#                 Note: jfs will always be external log, jfs2 always INLINE
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_jlog {
    my ($self) = @_;
    
    my ($cmd, $jlog, $name, @output, $status);

    # Jfs filesystems can only have external logs.
    # There is no 'inline' attribute in the filesystem
    # header.  So set to external and return.
    if ($self->get_fs() =~ /^jfs$/i) {
        return "external";
    }
    
    $name = $self->get_name();

    $cmd = "dumpfs /dev/$name | grep 'Inline Log:' | awk '{print \$3}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($output[0] eq "0") {
        $jlog = "external";
    } else {
        $jlog = "inline";    
    }

    return ($jlog);    
    
}

################################################################################
# Subroutine Name: _init_layout
# Objective:      Initialize layout of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_layout {
    my ($self) = @_;
    
    my ($basename, $cmd, $vgname, $layout, $name, $volname, @output, $status);
    
    $layout = "";

    $vgname = $self->get_vgname();
    $volname = $self->{name};
    $basename = $self->{basename};   
    
    # Assume volname is format '[vgname]_[volname]' for uniqueness 
    $cmd = "lslv -L $volname | grep 'SCHED POLICY:' | awk '{print \$5}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: Vol::_init_layout() ".
                        "Unable to retreive layout for volume \'$basename\' in volumegroup '$vgname'",
                        @output);
        return "";        
    } 

    if ($output[0] =~ /paral/i) {
        $layout = "concat";
    } else {
        $layout = $output[0];
    }

    return ($layout);
    
}

################################################################################
# Subroutine Name: _init_mount
# Objective:      Initialize mount point for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_mount {
    my ($self, $import) = @_;
    
    my ($cmd, $mount, $name, $path, $status, $vgname, @output);
    
    $mount = "";

    $vgname = $self->get_vgname();
    $name = $self->get_name();

    # Check to see if mount path has been stored
    # in volume metadata under 'comment' attribute.
    # This is to done in order to associate a mount
    # pt with a vg/vol in case the vol was renames while
    # still mounted.
    $cmd = "mount | grep $name | awk '{if (\$2) print \$2; else exit 1; }'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 

    # AIX doesn't return error when if can't find
    # mount point. So look for empty output
    if ($status gt 0 or scalar(@output) == 0) {                 
        # If volumes are bing imported assume then 
        # automatically mount fs with last stored 
        # mount point (or default path) if a mount
        # point does not exist
        if ($import) {
            $mount = $self->_get_metadata("LABEL");    
            if (defined $mount and length($mount)) {
               ($path, $vgname, $name) = $mount =~ /(\S+)\/(\w+)\/(\w+)$/;
            }
          
            $self->mount(-path => $path);
            return "" if $self->error();
        }  
    } else {
       # return the volume mount point.
       $mount = $output[0];
    }

    return ($mount);    
    
}

################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self) = @_;
    
    my ($basename, $cmd, $vgname, $size, $name);
    my ($pp_size,  @output, $sectors, $status);
    
    $size = "";

    $vgname = $self->get_vgname();
    $name = $self->get_name();
    $basename = $self->{basename};  

    # get PP size for the VG/LV
    $cmd = "lsvg $vgname | grep -i \"PP SIZE:\"";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: Vol::_init_size()",
                        "Unable to get PP size from VG ($name)",
                        @output);
        return ($ClientUtil::FALSE);
    }

    $pp_size = (split(/:/, $output[0]))[2];
    $pp_size =~ s/^\s+(\d+)\s+(\S+)//;
    $pp_size = $1;

    $cmd = "lsvg -l $vgname | grep \"$name\" | awk '{print \$4}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: Vol::_init_size() ".
                        "Unable to retrieve  size (MB) for volume \'$basename\' in volumegroup '$vgname'",
                        @output);
        return "";        
    } 

    # size in MB is multiplied by the PP size
    $size = $output[0];
    $size = $size * $pp_size;
    
    return ($size);    
    
}

################################################################################
# Subroutine Name: _init_state
# Objective:      Initialize state of existing host volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_state {
    my ($self) = @_;
    
    my ($cmd, $basename, $state, $vgname, $status, @output);
    
    $state = "";

    $vgname = $self->get_vgname();
    $basename = $self->{basename};

    $cmd = "lsvg -l $vgname | grep '$basename' | awk '{print \$6;}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: Vol::_init_state() ".
                        "Unable to retrieve state for volume \'$basename\' in volumegroup '$vgname'",
                        @output);
        
        return "";        
    } 

    $state = $output[0];

    return ($state);
    
}

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $exist, $import, $name, $size, $conn, $layout);
    my ($log, $parent, $self, $status, $type, @output, $stripe_size);

    $self = {};

    bless ($self, $class);    

    #default layout is concatenation
    $layout = "concat";
    $type = "raw";
    $import = 0;
    $exist = 0;
    $stripe_size = "64k";

    # parse arguments
    foreach (keys %args) {
        if (/-?layout$/) {
            if (not length($layout = $args{$_})) {
                $self->_set_err("Error: Vol::_new()",
                                "Invalid value for argument '-layout'");
                return ($self);
            }        
            next;
        } elsif (/-?_import$/) {  # private member
            if (not length($import = $args{$_})) {
                $self->_set_err("Error: Vol::_new()",
                                "Invalid value for argument '-_import'");
                return ($self);
            }
            next;
        } elsif (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: Vol::_new()",
                                "Invalid value for argument '-name'");
                return ($self);
            }
            next;
        } elsif (/-?size$/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Error: Vol::_new()",
                                "Invalid value for argument '-size'");
                return ($self);
            }
            next;
        } elsif (/-?type$/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: Vol::_new()",
                                "Invalid value for argument '-type'");
                return ($self);
            }
            next;
        } elsif (/-?exist$/) {
            if (not length($exist = $args{$_})) {
                $self->_set_err("Error: Vol::_new()",
                                "Invalid value for argument '-exist'");
                return ($self);
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: Vol::_new()",
                                "Invalid value for argument '-conn'");
                return ($self);
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Vol::_new()",
                                "Missing value for argument '-parent'");
                return ($self);
            }
            next;
        } elsif (/-?stripe_size$/) {
            if (not length($stripe_size = $args{$_})) {
                $self->_set_err("Error: Vol::_new()",
                                "Missing value for argument '-stripe_size'");
                return ($self);
            }
            next;
        }
    }  

    # check for existance of required -lvm argument
    unless (defined $parent) {
        $self->_set_err("Error: Vol::_new()",
                        "Missing required argument '-parent'");
        return ($self); 
    }

    
    # check for existance of required -type argument
    unless (defined $type) {
        $self->_set_err("Error: Vol::_new()",
                        "Missing required argument '-type'");
        return ($self); 
    }

    # check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: Vol::_new()",
                        "Missing required argument '-name'");
        return ($self); 
    }

    # check for existance of required -size argument
    unless (defined $size) {
        $self->_set_err("Error: Vol::_new()",
                        "Missing required argument '-size'");
        return ($self); 
    }

    # check for existance of required -conn argument
    if (grep( /^-?conn$/, (keys %args) )) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: Vg::_new()",
                            "Invalid value for argument '-conn'");
            return ($self); 
        }    
    } else {
        $self->_set_err("Error: Vg::_new()",
                        "Missing required argument '-conn'");
        return ($self); 
    }

    # Default stripe_size to 64k if not defined
    if (not defined($stripe_size)) {
        $self->{stripe_size} = "64k";
    } else {
        $self->{stripe_size} = $stripe_size;
    }

    # initialize the object members
    $self->{conn} = $conn;
    $self->{layout} = $layout;
    $self->{size} = $size;
    $self->{type} = $type;
    $self->{parent} = $parent;
    $self->{state} = 'closed/syncd';
    $self->{filesystem} = 'raw';

    if (not $exist) {
        $self->{basename} = $name;
        $self->{name} = $parent->get_name() . "_" . $name;
        $self->_create();
        return ($self) if ($self->error());
    } else {
        $self->{name} = $name; 

        # Determine the volume name input.  Aix required
        # that all logical volumes be unique which is not 
        # the case of other platform lvms.  To get around
        # will allow the same input as other lvms but 
        # perform the uniqueness underneath by appending
        # the volume group name as a prefix.  Keep track
        # of the lv input name in 'basename' of object.
        my $tmp = $self->{parent}->get_name() . "_";

        $self->{basename} = $name;       
        $self->{basename} =~ s/$tmp//;      

        $self->{state} = $self->_init_state();
        return ($self) if ($self->error);

        $self->{filesystem} = $self->_init_fstype();
        if ($self->error()) {
            # allow vol to be loaded although it 
            # can't be started to determine fs
            if ($self->{filesystem} eq "") {
                $self->_clear_err();
            }  else {
                return ($self);
            }
        }
        
        if ($self->{filesystem} !~ /raw/i) {
            $self->{jlog} = $self->_init_jlog();
            $self->{mount} = $self->_init_mount($import);

            return ($self) if ($self->error);
        }

        $self->{size} = $self->_init_size();
        return ($self) if ($self->error);

        $self->{layout} = $self->_init_layout();
        return ($self) if ($self->error);
    }
    
    return ($self);

}


################################################################################
# Subroutine Name: _set_metadata
# Objective:      Set metadata for attribute 
# Details:        See POD documentation at the end of this file
################################################################################
sub _set_metadata {
    my ($self, $attribute, $value) = @_;
    
    my ($cmd, $vgname, $name, @output, $status);
    
    $vgname = $self->get_vgname();
    $name = $self->get_name();
    
    # Notes: This needs to be modifed to allow various modification
    # depending on attribute.  Currently hardcoded for label only.
    if ($attribute eq "LABEL") {
        $cmd = "chlv -L $value $name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($self->error) {
            $self->_set_err("Error: Vol::_set_metadata() ",
                            "Unable to set volume \'$name\' metadata for",
                            "attribute '$attribute' in volumegroup '$vgname'",
                            @output);
            return ($ClientUtil::FALSE); 
        }
    } else {
        $self->_set_err("Error: Vol::_set_metadata() LABEL attribute, ",
                        "($attribute) was passed");
        return ($ClientUtil::FALSE); 
    }
    
    return ($ClientUtil::TRUE);  
    
}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Vol.pm

=head1 Authors

=begin html 

Gary Knopp

=end html

=head1 Description

Vol provides an API interface to LVM volumes through the use of the Telnet library.  

=head1 Public Methods

=head2 B<clean()>

Remove file content from a volume

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("Remove volume content");
    if (not $dtObj->clean()) {
        $Test->log($dtObj->errmsg);
    }

    

=back

=head2 B<create_dt(I<%args>)>

Create a datatest object that generates I/O on filesystem or raw device under test.

=over 4

=item Inputs

=item * I<-proc> (default: 1)

The number of datatest processes to execute in parallel

=item * I<-limit> (default: '2m')

The amount of data each process thread will write per datatest cycle.  Include unit 
of storage with value ((b)ytes, (k)ilobytes, (m)egabytes, (g)igabytes)

=item * I<-compare> (default: 1)

Determine whether datatest will verify the pattern it wrote to the filesystem.  

=item * I<-dispose> (default: 1)

Determine whether datatest deletes the data it created when its done.  

=item * I<-runtime> 

The length of time the datatest program will run. Include
unit of time with value ((s)econds, (m)inutes, (h)ours, (d)ays)


=item Outputs

Returns a Datatest object if the background process was created sucessfully. If an error occurs the
error flag is set, errmsg is set to the failure reason, and undef hash reference is returned.

=item Examples

        $Test->log("-- Starting I/O");                                                
        $dtObj = $volObj->create_dt(-procs=>4,-limit=>"2m",-runtime=>'10s');
        if ($volObj->error) {
            $Test->log($volObj->errmsg());
            return $TCase::FAIL;
        }

=back

=head2 B<create_fs(%args)>

Creates a filesystem on the volume.  The filesystem is currently created with the 'largefiles' option.

=over 4

=item Inputs

=item * -type

The type of filesystem to create (vxfs or ufs)

=item Outputs

Returns a scalar value of 1 if the filesystem was successfully created, otherwise 0. If an error occurs the
error flag is set, errmsg is set to the failure reason.  

=item Examples

    # Create a vxfs filesystem
    $Test->log("-- Creating volume group 'myvg'");
    if ($volObj->create_fs(-type=>"vxfs")) { 
        $Test->log($volObj->errmsg());
        return $TCase::FAIL; 
    }

=back

=head2 B<destroy()>

Destroy the volume 

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Destroy volume");

    $volObj->destroy();
    if ($volObj->error()) {
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<display()>

Display detailed volume information

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of volume information if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving volumeinformation");

    $Test->log($volObj->display());
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<error()>

This method allows you to determine if any errors occured in the
previous method call.  Note that the error flag is reset at the
beginning of each method, so it should be checked after B<every>
major method call.  If you do not check for errors, your calling
program may fail at some point and you will not know why.

=over 4 

=item Inputs

None

=item Outputs

Returns 1 if an error occured in the previous call, 0 otherwise.

=item Examples

 # Determine if error occurred...
 if ($obj->error) {
    print "Error encountered in previous method call\n";
 }

=back

=head2 B<errmsg()>

This method allows you to get the text of the error message if an error
occurred on the previous method call.

=over 4

=item Inputs

None

=item Outputs

Returns the text of the last error message set.  If no error message is
set, an empty string will be returned.

=item Examples

 # Check for errors.
 if ($obj->error) {
    $errmsg = $obj->errmsg;
    print "An error occurred: $errmsg\n";
 }

=back

=head2 B<get_files()>

Retrieve a list of files

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of file names created by datatest if successful. 
If an error occurs the error flag is set, errmsg is set to the failure reason, and an 
empty array is returned.        

=item Examples

    $Test->log("-- Retrieving files");
    @files = $volObj->get_files();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }  
    

=back

=head2 B<get_fs()>

Type of filesystem 

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the filesystem structure.  If no filesystem exists than the value 
'raw' is returned. If an error occurs the error flag is set, errmsg is set to the failure reason,
and "" is returned.        


=item Examples

    print "Volume filesystem is '".$volObj->get_fs()."'";

=back

=head2 B<get_layout()>

Structure of volume 

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the volume structure.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Volume layout is '".$volObj->get_layout()."'";

=back

=head2 B<get_mount()>

Mount point of volume

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the the volume mount point.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Volume is mounted at '".$volObj->get_mount()."'";

=back

=head2 B<get_name()>

Name of volume

=over 4

=item Inputs

none

=item Outputs

Returns the name of the volume.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Volume name is '".$volObj->get_name()."'";

=back

=head2 B<get_pattern()>

Retrieve pattern that will be used to generate file/raw content 

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the pattern if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Pattern used to generate data content is   '".$volObj->get_pattern()."'";

=back

=head2 B<get_size()>

Volume size in megabytes

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is returned.        


=item Examples

    $size = $volObj->get_size();
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }  
   
    print "Volume size is $size mb\n";

=back

=head2 B<get_vgname()>

Name of volume group 

=over 4

=item Inputs

none

=item Outputs

Returns the name of the volume group.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Volume group is '".$volObj->get_vgname()."'";

=back

=head2 B<is_fs()>

Determine if a volume has a filesystem 

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if a filesystem exists, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Examples

    if(not $volObj->is_fs()) {
        ...
        <error handler>
        ...
    }
   

=back

=head2 B<is_mounted()>

Determine if volume filesystem is mounted.

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if filesystem is mounted, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Examples

    if(not $volObj->is_mounted()) {
        $volObj->mount();
        ...
        <error handler>
        ...
    }
   

=back

=head2 B<mount(I<%args>)>

Mount the filesystem

=over 4

=item Inputs

=item * I<-path> (Default: /usr/tmp/mnts/[vgName]/[volName])

The mount point location where the filesystem will be accessible. If the path does not exist
it is automatically created with full user access.

=item Outputs

Returns a value of 1 the filesystem is successfully mounted, otherwise 0.  If an error occured 
the error flag is set, errmsg is set to the failure reason.

=item Examples

   $volObj->mount();
   if ($volObj->error) {
       $Test->log($volObj->errmsg());
       return $TCase::FAIL;
   }  
       

=back

=head2 B<rename($name)>

Rename a volume to a new name

=over 4

=item Inputs

=item * $name

The new name of the volume

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("Rename volumes");
    foreach my $volObj (@volObjs) {
        $volObj->rename($volObj->get_name()."_new");
        if ($volObj->error) {
            $Test->log($volObj->errmsg);
            return $TCase::FAIL;
        }
    }

=back

=head2 B<rescan()>

Rescan volume to refresh state.  Currently, this method is only refreshing 
the mount point if it advertantly goes away during a restore.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("Refresh volume state");
    $volObj->rescan() if ($volObj->is_fs());
    if ($volObj->error) {
        $Test->log($volObj->errmsg());
    }

=back

=head2 B<set_pattern($pattern)>

Sets pattern that will be used in generating file/raw content 

=over 4

=item Inputs

=item * $pattern (default: 0x39c39c39c);

The pattern to use when generating file content.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    my $pattern = "Welcome to the Host SAN Project";
    $volObj->set_pattern($pattern);
    if ($volObj->error()) {
        $Test->log($volObj->errmsg());
    }

=back

=head2 B<unmount(I<%args>)>

Mount the filesystem

=over 4

=item Inputs

=item * I<-force> (1/0)

Setting this value to true will force the volume to be unmounted even if its busy.

=item Outputs

Returns a value of 1 the filesystem is successfully unmounted, otherwise 0.  If an error occured 
the error flag is set, errmsg is set to the failure reason.

=item Examples

   $volObj->unmount(-forece=>1);
   if ($volObj->error) {
       $Test->log($volObj->errmsg());
       return $TCase::FAIL;
   }  
       

=back

1
