# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the NativeLvm Logical Volume Manager on Aix
## @author knopp@netapp.com
## @status private
## @pod ../../Lvm.pm


# Package Name
##############
package San::SanClient::Lvm::Aix::SanPath;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanClient::Lvm;
use San::SanClient::Lvm::Aix::SanPath::Vg;

use Error;
use San::SanLog;

use vars qw( @ISA );

@ISA=qw(Error Sanlog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################


################################################################################
# Subroutine Name: initialize
# Objective:       initialize disk under veritas control
# Details:         See POD documentation at the end of this file
################################################################################
sub initialize ($)
{
    my ($self,%args) = @_;
    my (
	$cmd, 
	@already_inited,
	@output, 
	$status,
	$devices,
	);


    #  This function is a NOOP for SanPath on AIX. 

    # parse arguments
    foreach (keys %args) {
	if (/-?devices/) {
	    if (not length($devices = $args{$_})) {
		$self->_set_err("Error: SanPath::initialize() Invalid value for argument '-devices'");
		return $ClientUtil::FALSE; 
	    }        
	    next;
	}
	elsif (/-?format/) { #  Noop for Aix   
	    next;
	}
	else {
	    $self->_set_err("Error: SanPath::initialize() Invalid parameter specified \'$_\'");
	    return $ClientUtil::FALSE;                       	  
	}
    }     

    #  check for required parameters
    unless(defined $devices and length $devices){
	$self->_set_err("Error: SanPath::initialize() -devices is a required argument");
	return $ClientUtil::FALSE;
    }

    return($ClientUtil::TRUE);

}
################################################################################
# Subroutine Name: clear_cache
# Objective:       clear cache objects.
# Details:         See POD documentation at the end of this file
################################################################################
sub clear_cache {
    my ($self) = @_;
    
    #  Clear cache of vg objects
    $self->{vgs}=(); 	

    return $ClientUtil::TRUE;
    
} 
################################################################################
# Subroutine Name: create_vg
# Objective:       create and return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vg {
    my ($self,%args) = @_;
    my (
	$vgObj,
	);
    
    $vgObj=Vg->_new(	
			-parent=>$self,
			-conn=>$self->{conn},
			-loghdl=>$self->{log},
			%args);

    if ($vgObj->error){
	$self->_set_err($vgObj->errmsg());
	return $self;
    }	

    #  Check and remove any cached copy of the Vg that
    #  may still exist in the LVM due to a deletion
    #  occuring outside the test's control.  Ignore error
    #  if id doesn't currently exist.

    $self->unload_vg($vgObj->get_name());
    $self->_clear_err if ($self->error); 

    push @{$self->{vgs}}, $vgObj;
    

    return $vgObj;
}

################################################################################
# Subroutine Name: destroy_vg
# Objective:       destroy a named diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vg {
    my ($self,$name) = @_;
    my (
	$vgObj,
	@vgs,
	);
    
    $self->_clear_err();

    unless(defined $name and length $name){
	$self->_set_err("Error: SanPath::destroy_vg() vg name is a required argument");
	return $ClientUtil::FALSE
	}
    
    #  Delete the of vg object and update the parent vg list.
    foreach $vgObj ( @{$self->{vgs}}){ 
	if ($vgObj->get_name() =~ /^$name$/){
	    $vgObj->destroy();
	    if ($vgObj->error){ 
		$self->_set_err($vgObj->errmsg());
		return $ClientUtil::FALSE	
		}
	}
	else {
	    push @vgs,$vgObj;
	}
    }
    
    $self->{vgs}=\@vgs;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display
# Objective:       display disk group information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self,%args) = @_;
    my (
	$cmd,
	$status,
	$verbose,
	@buffer,
	@output,
	@vgObjs,
	);
    
    $self->_clear_err();

    $self->_log("Display diskgroup information");


    @vgObjs=$self->get_vgs();
    return () if ($self->error);

    foreach my $vgObj(@vgObjs){		
	@output = $vgObj->display(%args);
	return (wantarray()) ? () : [] if ($self->error);

	push @buffer,
	@output[0..scalar(@output)-2],   
	"=====================================".
	    "=====================================";
    }

    return (wantarray()) ? @buffer : \@buffer;    
    

}

###############################################################################
# Method:    get_disk_format
# Objective: Get list of supported disk formats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_format {
    my ($self,$disk) = @_;
    my (
	$format,
	);

    $self->_clear_err();

    #  This is a noop for Aix.  Function
    #  supports veritas lvm

    $format="";

    return $format; 
}

###############################################################################
# Method:    get_disk_formats
# Objective: Get list of supported disk formats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_formats {
    my ($self) = @_;
    my (
	@format
	);

    $self->_clear_err();
    #  This is a noop for Aix.  Function
    #  supports veritas lvm

    @format=();

    return (wantarray()) ? @format : \@format;  
}

###############################################################################
# Method:    get_format_default
# Objective: Get default type to use for disk format
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_format_default {
    my ($self) = @_;
    my (
	$format,
	);
    

    $self->_clear_err();
    #  This is a noop for Aix.  Function
    #  supports veritas lvm

    $format="";
    
    return $format;   
}


################################################################################
# Subroutine Name: get_vg
# Objective:       return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vg {
    my ($self,$name) = @_;
    my (
	$vgObj,
	);
    
    $self->_clear_err();

    unless(defined $name and length $name){
	$self->_set_err("Error: SanPath::get_vg() vg name is a required argument");
	return $ClientUtil::FALSE ;
    }

    #$vgObj={};

    foreach my $obj ( @{$self->{vgs}}){
	if ($obj->get_name() =~ /^$name$/){
	    return $obj;
	}
    }
    
    $self->_set_err("Error: SanPath::get_vg() Unable to find diskgroup '$name'");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_vgs
# Objective:       return array of volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgs {
    my ($self) = @_;
    my (
	@vgs,
	);
    
    $self->_clear_err();

    if (defined $self->{vgs}) {
	@vgs = @{$self->{vgs}};
    } 
    else {
	$self->_set_err("Error: SanPath::get_vgs() Vxvm object has no volumes groups");
	return ();		
    }

    return (wantarray()) ? @vgs : \@vgs;
    
} 
################################################################################
# Subroutine Name: get_vgnames
# Objective:       return array of volume group names
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgnames {
    my ($self) = @_;
    my (
	$cmd,
	$status,
	@buffer,
	@output,
	);
    
    $self->_clear_err();

    $cmd = "lsvg | grep -v root";
    

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
	$self->_set_err("Error: SanPath::get_vgnames Command ($cmd) returned a bad status ($status)");     
	return (wantarray()) ? () : [];
    }

    #  Remove dev prefix to vg name
    foreach my $line (@output) {
	$line=~s/\/dev\///;
	push @buffer,$line;
    }


    return (wantarray()) ? @buffer : \@buffer;
    
} 
################################################################################
# Subroutine Name: get_pkginfo()
# Objective:       get detailed lvm pkg information
# Details:         See POD documentation at the end of this file
################################################################################
sub get_pkginfo {
    my ($self) = @_;
    my (
	$cmd, 
	@output, 
	$status,
	);
    
    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{pkginfo});
    
    return (wantarray()) ? @{$self->{pkginfo}} : $self->{pkginfo};   

}

################################################################################
# Subroutine Name: _get_pkginfo
# Objective:       Get lvm package info
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self) = @_;
    my (
	$cmd,
	$name,
	$vers,
	$rev,	
	@buffer,
	@output, 
	$status,
	);

    $cmd = "lslpp -l | grep lvm | awk '{print \$2}'";
    ($status, @buffer) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if command worked
    if ($status gt 0) {
	$self->_set_err( "Error: SanPath::get_version() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 

    push @output,"NativeLvm version: ".$buffer[0];


    
    if ($self->{parent}->get_protocol()=~/fcp/i){
	$cmd = "setsp -V";

	($status, @buffer) = &ClientUtil::execute($self->{conn}, $cmd); 
	
	# determine if command worked
	if ($status gt 0) {
	    $self->_set_err( "Error: SanPath::get_version() Command ($cmd) returned a bad status ($status)");
	    return $ClientUtil::FALSE;                             
	} 

	($vers)=$buffer[0]=~/:\s(\S+)/;

	#  The aix lvm is NativeLvm and not "SanPath" which is
	#  a 3rd party multipathing product.  This needs to be
	#  corrected once the impact to existing testcases is
	#  understood and corrected.

	$self->{name}="SanPath";
	$self->{vers}=$vers;
	$self->{rev}=$vers;
	push @output,@buffer;
    }



    $self->{pkginfo}=\@output;
    
}
################################################################################
# Subroutine Name: get_revision()
# Objective:       Get revision of Veritas
# Details:         See POD documentation at the end of this file
################################################################################
sub get_revision {
    my ($self) = @_;
    my (
	$cmd, 
	@output, 
	$status,
	);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{rev});
    return "" if ($self->error());	

    return $self->{rev};  
    
}

################################################################################
# Subroutine Name: get_name()
# Objective:       Get lvm name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;
    my (
	$cmd, 
	@output, 
	$status,
	);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{name});
    return "" if ($self->error());	

    return $self->{name};  
    
}

################################################################################
# Subroutine Name: get_version()
# Objective:       Check version of Veritas
# Details:         See POD documentation at the end of this file
################################################################################
sub get_version {
    my ($self) = @_;
    my (
	$cmd, 
	@output, 
	$status,
	$version,
	);

    $self->_clear_err();	
    
    $self->_get_pkginfo() unless (defined $self->{vers});
    return "" if ($self->error());	

    return $self->{vers};
    
}


################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices{ 
    my ($self,%args)=@_;
    my (
        $bCnt, 
        $cmd,
        $cnt,
        $device,
        $filer,
        $filername,
	$lun,
	$refresh,
        $status,
        $total,
        $type,
	$vol,
        %hDevs,
	@all_devices,
	@attr,
        @devices,
	@lvm_devices,
        @output,
        @used,
	);

    $type=".*";
    $total=0;

    $self->_clear_err();

    #  Asssume user supplied -cnt argument
    $bCnt=1;
    $refresh=0;

    # parse arguments
    foreach (keys %args) {
	if (/-?cnt/) {
	    if (not length($cnt = $args{$_})) {
		$self->_set_err("Error: AixSanPath::get_free_devices() Invalid value for argument '-cnt'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	}
	elsif (/-?type/) {
	    if (not length($type = $args{$_})) {
		$self->_set_err("Error: AixSanPath::get_free_devices() Invalid value for argument '-type'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	elsif (/-?filer/) {
	    if (not length($filer = $args{$_})) {
		$self->_set_err("Error: AixSanPath::get_free_devices() Invalid value for argument '-filer'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	elsif (/-?vol/) {
	    if (not length($vol = $args{$_})) {
		$self->_set_err("Error: AixSanPath::get_free_devices() Invalid value for argument '-vol'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	}
	elsif (/-?lun/) {
	    if (not length($lun = $args{$_})) {
		$self->_set_err("Error: AixSanPath::get_free_devices() Invalid value for argument '-lun'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	}
	elsif (/-?refresh/) {
	    if (not length($refresh = $args{$_})) {
		$self->_set_err("Error: AixSanPath::get_free_devices() Invalid value for argument '-refresh'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	}
	else {
	    $self->_set_err("Error: AixSanPath::get_free_devices() Invalid parameter specified \'$_\'");
	    return ($ClientUtil::FALSE);                       	  
	}
    }     
    
    #  Check if user wants free devices from a specified filer,volume, or lun
    if (defined $filer or defined $vol or defined $lun) {

	#  Check if a device to filer map 'dfmap' has already been created or 
	#  refresh it if requested.
	if ( not defined $self->{dfmap} or $refresh eq 1){
	    $self->_log("creating translation maps; device to filer, vol, and lun");
	    @output=$self->{sanlun}->lun_show();
	    if ($self->{sanlun}->error){
		$self->_set_err($self->{sanlun}->errmsg());
		return $ClientUtil::FAIL;
	    }

	    foreach my $line (@output){

                # skip first line (header)
                if ($line =~ /lun-pathname/i) {
                    next;
                }
		
		my ($filer_tmp,$vol_tmp,$device_tmp)=$line=~/\s{0}(\S+):\s+\/vol\/(\w+)\/\S+\s+(\S+)\s+/;
		
		@attr=split(/\s+/,$line);
		my $lun_tmp=basename($attr[1]);
		
		$hDevs{$device_tmp}={
		    filer=>$filer_tmp,
		    vol=>$vol_tmp,
		    lun=>$lun_tmp
		    };
	    }
	    
	    #  save map for next call of function
	    $self->{dfmap}=\%hDevs;

	}
    }

    #  If not -cnt then assume all free disks 
    #  under lvm control are wanted
    unless (defined $cnt and length($cnt)){
	$cnt=10000;
	$bCnt=0;
    }

    # Get list of devices. Make sure you take in account deported vgs.
    # execute setsp (as root)

    #  Use sanpath if protocol is fcp to acquire master device name. 
    if ($self->{parent}->get_protocol()=~/fcp/i){
	$cmd="setsp | egrep  'spd[0-999]' | grep 'NETAPP' | awk '{print \$3;}'"; 
	
	($status, @lvm_devices) = &ClientUtil::execute($self->{conn}, $cmd); 
	
	# determine if lsdev command worked
	if ($status gt 0) {
	    $self->_set_err( "Error: AixHost::_get_devices() Command ($cmd) returned a bad status ($status)");
	    return $ClientUtil::FALSE;                             
	}

	# determine if there were any sanpath devices available
	if ( grep(/no spd devices found/i, @lvm_devices) > 0 ) {
	    $self->_set_err( "Error: AixHost::_get_devices() Command ($cmd) failed to find Sanpath devices.  Check license.");
	    return $ClientUtil::FALSE;
	}
    }
    else {
	@lvm_devices=$self->{parent}->{sanlun}->lun_show(-_primary_paths_only=>1);
    }

    #  Get list of all free devices and paths
    $cmd="lspv | grep 'None' | awk '{print \$1}'";  

    ($status, @all_devices) = &ClientUtil::execute($self->{conn}, $cmd); 
    
    # determine if command worked
    if ($status gt 0) {
	$self->_set_err( "Error: AixHost::get_free_devices() Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    } 
    
    # iterate through the devices and make sure that none of the disks 
    # are being used other LVM disk object.  
    foreach my $device (@lvm_devices) {		
	
	#  Skip devices being used other vgs.
	if (grep /^$device$/,@all_devices) { 

	    #  Apply filters to any of the free devices found
	    if ($device){
		
		if (defined $filer and length $filer) {
		    next if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i);
		} 

		if (defined $vol and length $vol) {
		    next if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i);
		} 
		
		if (defined $lun and length $lun) {
		    next if ($self->{dfmap}{$device}{lun} !~ /$lun/i);
		} 
		
		push @devices,$device if (++$total <= $cnt); 
	    }	
	}
    } 

    #  If user supplied cnt and free devices fell short
    #  generate error
    if ($bCnt eq 1 and scalar(@devices) < $cnt){
	$self->_set_err(
			"Error: AixSanPath::get_free_devices() ",
			"Only ".scalar(@devices)." of $cnt requested devices were found");
    }

    return (wantarray()) ? @devices : \@devices; 
}


################################################################################
# Subroutine Name: delete_devices
# Objective:       delete devices from host
# Details:         See POD documentation at the end of this file
################################################################################
sub delete_devices {
    my ($self,@devices)=@_;
    my (
	$cmd,
	$status,
	@output,
	);

    $self->_clear_err();

    #  Remove devices from host.  This is because orphaned
    #  devices maintain metadata of prior lun
    if ($self->{parent}->get_protocol=~/fcp/){
	
	#  Capture sanpath information priot to deletion 
	#  for debugging purposes
	#$self->_log("Debug:  Display Sanpath /etc/sppath.conf /etc/spd.conf",
	#           "before removing entries");

        #$cmd="cat \/etc\/sppath.conf";	    
        #($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        #$self->_log("FILE: /etc/sppath.conf",@output);
	
        #$cmd="cat \/etc\/spd.conf";	    
        #($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        #$self->_log("FILE: /etc/spd.conf",@output);

	foreach my $device (@devices){
	    
	    $self->_log("device: ".$device);

	    #  Determine spd id for master device name
	    my $spd_id=$self->_get_spd_id($device);   
	    $self->_log($self->errmsg()) if ($self->error);	    
	    $self->_log("spd_id: ".$spd_id);

	    #  Retrieve all device paths for spd_id
	    my @alldevpaths=$self->_get_spd_devices($spd_id);
            $self->_log($self->errmsg()) if ($self->error);
            #$self->_log("alldevs: ",@alldevpaths);

	    #  Stop master spd device
	    $self->_stop_spd_id($spd_id);
            $self->_log($self->errmsg()) if ($self->error);

            #  Remove master spd device
	    $self->_remove_spd_id($spd_id);
            $self->_log($self->errmsg()) if ($self->error);

	    #  Remove device from host
	    foreach my $device (@alldevpaths){

		$cmd = "rmdev -d -l".$device;

		$self->_log($cmd);

		($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 								 
		if ($status gt 0) {
		    $self->_set_err( "Error: Sanpath::destroy() Unable to remove device '".$device."'",				     
				     @output);
		    
		    return $ClientUtil::FALSE;                       
		} 
	    }

            #$cmd="cat \/etc\/sppath.conf";	    
            #($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
            #$self->_log("FILE AFTER: /etc/sppath.conf",@output,"\n");

	}
    }
    else {
	
	foreach my $device (@devices){
	    
	    $cmd = "rmdev -d -l".$device;

	    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 								 
	    if ($status gt 0) {
		$self->_set_err( "Error: Sanpth::destroy() Unable to remove device '".$device."'",
				 @output);
		
		return $ClientUtil::FALSE;                       
	    } 	    

	}

    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: load_vg
# Objective:       load an existing diskgroup that is not in memory
# Details:         See POD documentation at the end of this file
################################################################################
sub load_vg {
    my ($self,$name) = @_;
    my (
	$vgObj,
	);

    $self->_clear_err();

    unless (defined $name and length $name){
	$self->_set_err("ERROR: AixSanPath::load_vg() Invalid or missing required paramter \$name");
	return {};
    }

    #  Check if vg has already been loaded
    $vgObj=$self->get_vg($name);
    if ($self->error) {

	#  dg has not been load yet
	$self->_clear_err();

	#  load dg
	$self->_create_vgs($name);
	return {} if ($self->error);
	$vgObj=$self->get_vg($name);		
    } 
    else {
	$self->_log("INFO: AixSanPath::load_vg() Volume group '$name' was already loaded");
    }

    return $vgObj;
    
}
################################################################################
# Subroutine Name: unload_vg
# Objective:       Remove the vg object under LVM object control
# Details:         See POD documentation at the end of this file
################################################################################
sub unload_vg {
    my ($self,$name) = @_;
    my (
	$vgObj,
	$bFound,
	@vgs,
	);

    $self->_clear_err();

    unless (defined $name and length $name){
	$self->_set_err("ERROR: AixSanPath::unload_vg() Invalid or missing required paramter \$name");
	return {};
    }

    $bFound=0;
    
    foreach $vgObj ( @{$self->{vgs}}){
	($vgObj->get_name() =~ /^$name$/)
	    ? $bFound=1
		: push @vgs, $vgObj;
    }
    
    $self->{vgs}=(@vgs) ? \@vgs : (); 

    unless($bFound) {
	$self->_set_err("ERROR: AixSanPath::unload_vg() Nothing to unload. Vg name '".$name."' was not loaded");
	return $ClientUtil::FALSE;
    }	

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: enable_controllers
# Objective:       enable controllers
# Details:         See POD documentation at the end of this file
################################################################################

sub enable_controllers{
    my ($self) = @_;
    
    my ($cmd, @output, $status);
    my ($ctlr, @controllers);

    $self->_clear_err();

  Tharn::tstlog('-comment', "Enabling all controllers for Veritas DMP");

    if ($self->is_installed()) {
	
        $self->rescan();
	return if ($self->error());

        # Get a list of all the controllers
        $cmd  =  "/etc/vx/bin/vxcntrllist";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

	if ($status > 0) {
	    $self->_set_err("Error: AixSanPath::enable_controllers() Command ($cmd) returned a bad status ($status)");     
            return $ClientUtil::FALSE;
        }

        # skip controller 0 (root disk)
        @controllers = grep (! /c0/, @output);
	
        # traverse through the controllers and enable each.
        foreach $ctlr (@controllers) {
            $cmd  = "/usr/sbin/vxdmpadm enable ctlr=$ctlr";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
		$self->_set_err("Error: AixSanPath::enable_controllers() Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

}

################################################################################
# Subroutine Name: rescan
# Objective:       To run the sanpath command to rescan  list of
#                  disks that it thinks are under it's control.
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;
    my (
	$cmd, 
	@output, 
	$status,
	);

    $self->_clear_err();	
    
    if ($self->{parent}->get_protocol()=~/fcp/) {

	# refresh sanpath devices
	$cmd = "setsp -T;setsp -S";

	($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
	
	#  igrnore errors since any busy devices will complani
	
    }
    

    return $ClientUtil::TRUE;  

}


################################################################################
# Subroutine Name: restart_recovery_daemon
# Objective:       Method that will stop the recovery deamon and reset
# Details:         See POD documentation at the end of this file
################################################################################

sub restart_recovery_daemon{
    my ($self, $interval) = @_;
    
    my ($cmd, @output, $status);

    $self->_clear_err();

    if ($self->is_installed()) {
      Tharn::tstlog('-comment', "Setting Veritas Recover timing to $interval seconds");
	
        # stop the restored
        $cmd  = "/usr/sbin/vxdmpadm stop restore";
        ($status, @output) = &ClientUtil::execute($self->{conn}, 
                                                  $cmd);
        if ($status > 0) {
	    $self->_set_err("Error: AixSanPath::restart_recovery_daemon()",
			    "Command ($cmd) returned a bad status ($status)");

            return $ClientUtil::FALSE;
        }

        # start the restored
        $cmd  = "/usr/sbin/vxdmpadm start restore interval=$interval";
        ($status, @output) = &ClientUtil::execute($self->{conn},$cmd);
        if ($status > 0) {
	    $self->_set_err("Error: AixSanPath::restart_recovery_daemon()",
			    "Command ($cmd) returned a bad status ($status)");

            return $ClientUtil::FALSE;
        }

        # set stats on the setting, just to get it in the logs.
        $cmd  = "/usr/sbin/vxdmpadm stat restored";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
	if ($status > 0) {
	    $self->_set_err("Error: AixSanPath::restart_recovery_daemon()",
			    "Command ($cmd) returned a bad status ($status)");

            return $ClientUtil::FALSE;
        }

    }
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ( 
	 $conn,
	 $loadcfg,
	 $log,
	 $lvm,
	 $parent,
	 $hosttype,
	 $self,
	 );

    $self={};

    bless $self,$class;	


    # parse arguments
    foreach (keys %args) {
	if (/-?lvm/) {
	    if (not length($lvm = $args{$_})) {
		$self->_set_err("Error: SanPath::_new() Invalid value for argument '-lvm'");
		return ($ClientUtil::FALSE); 
	    }        
	    next;
	}
	elsif (/-?parent/) {
	    if (not length($parent = $args{$_})) {
		$self->_set_err("Error: SanPath::_new() Invalid value for argument '-parent'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	elsif (/-?loghdl/) {
	    if (not length($log = $args{$_})) {
		$self->_set_err("Error: SanPath::_new() Invalid value for argument '-loghdl'");
		return ($ClientUtil::FALSE); 
	    }
	    next;
	}
	elsif (/-?loadcfg/) {
	    if (not length($loadcfg = $args{$_}) or $loadcfg !~ /no|yes|1|0|off|on/i) {
		$loadcfg=1;
	    }
	    next;
	}
	else {
	    $self->_set_err("Error: SanPath::_new() Invalid parameter specified \'$_\'");
	    return ($ClientUtil::FALSE);                       	  
	}
    }     
    
    #  Allow CLI commands to be logged;
    $self->{log}=$log if (defined $log);
    

    #  check for existance of required -lvm argument
    if (grep ( /^-lvm$/, (keys %args) )) {
	unless ($args{-lvm} =~ /sanpath|vxvm/i) {
	    $self->_set_err("Error: SanPath::_new()",
			    "Invalid value for argument '-lvm' \(".$args{-lvm}."\)");
	    return $self; 	
	}
    }
    else {
	$self->_set_err("Error: SanPath::_new()",
			"Missing required argument'-lvm'");
	return $self; 
    }

    #  check for existance of required -parent argument
    if (grep ( /^-?parent$/, (keys %args) )) {
	unless (ref($args{-parent}) =~ "SanClient") {
	    $self->_set_err("Error: SanPath::_new()",
			    "Invalid value for argument '-conn'");
	    return $self; 
	}
	$hosttype=$args{-parent}->{hostp}{hosttype};
    }
    else {
	$self->_set_err("Error: SanPath::_new()",
			"Missing required argument '-conn'");
	return $self; 
    }

    # initilize the object members
    $self->{conn}=$parent->get_connection();
    $self->{sanlun}=$parent->get_sanlun();
    
    $self->{vgs}= [];

    #  check to make sure that LVM is installed
    if (not $self->_is_installed()){
     	return $self if ($self->error);     
    } 

    #  create sanpath vg objects for existing disk groups
    $self->_create_vgs()if ($loadcfg);
    return $self if ($self->error);

    return($self);

}

################################################################################
# Subroutine Name: _create_vgs
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vgs {
    my ($self,$vgname)=@_;
    my (
	$cmd,
	$name,
	@output,
	$status,
	$vgObj,
	%hDeport,
	);

    $cmd= (defined ($vgname) and length ($vgname))
	? "lsvg | grep $vgname | awk '{print \$1}'"	
	    : "lsvg | grep -v root | awk '{print \$1}'";
    
    
    
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);   								 
    if ($status gt 0) {
	$self->_set_err("Error: SanPath::_create_vgs()",
			"Unable to load existing diskgroups");
	
	return $ClientUtil::FALSE;                      
    } 

    #  Skip the header and rootdg lines
    foreach my $name (@output){
	$self->create_vg(-name=>$name,-deported=>0);
	return $ClientUtil::FALSE if ($self->error);
    }

    #  Aix NativeLvm doesn't have a solution (from what I see) for 
    #  determining if a deported volume group exists on any of the 
    #  host devices (as do other platforms/lvms)


    return $ClientUtil::TRUE;
    
}

################################################################################
# Subroutine Name: _get_disk_id
# Objective:       Return disk id
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_disk_id {
    my ($self,$device) = @_;

    my(
       $cmd,
       $name,
       @output,
       $status,		
       );
    
    $self->_clear_err();

    $name="/dev/rdsk/$device"."s2";

    $cmd = "sanlun lun show -v -d $name | grep 'Serial number'";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 								 
    if ($status gt 0) {
	$self->_set_err( "Error: SanPath::get_diskid() Unable to get diskid for device \'$device\'",
			 @output);
	
	return $ClientUtil::FALSE;                       
    } 
    
    return $output[0];	
    
}

################################################################################
# Subroutine Name: _get_spd_devices
# Objective:       get all device id paths
# Details:         See POD documentation at the end of this file
################################################################################

sub _get_spd_devices {
    my ($self,$spd_id)=@_;
    my (
	$cmd,
	$id,
	$status,
	@devices,
	@output,
	);
    
    $cmd = "setsp -l".$spd_id;  	

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
	$self->_set_err( "Error: Sanpath::_get_spd_devices() Unable to get devices for spd_id \'$spd_id\'",
			 @output);
	
	return (wantarray()) ? () : [];                          
    }

    foreach my $line (@output) {
	next if ($line!~/hdisk\d+\(/i);
	push @devices,$line=~/(hdisk\d+)\(/;;
    }
    
    return (wantarray()) ? @devices : \@devices; 
}
################################################################################
# Subroutine Name: _get_spd_id
# Objective:       get sanpath device id
# Details:         See POD documentation at the end of this file
################################################################################

sub _get_spd_id {
    my ($self,$device)=@_;
    my (
	$cmd,
	$id,
	$status,
	@output,
	);

    $cmd = "cat \/etc\/sppath.conf | egrep \':$device\$\' | awk '{print substr(\$1,5)}'";

    # $cmd = "setsp | grep '= $device '";  	

    #$self->_log("CMD: ".$cmd);

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
	$self->_set_err( "Error: Sanpath::_get_spd_id() Unable to retrieve sanpath id for device \'$device\'",
			 @output);
	
	return ""                          
	}


    #($id)=$output[0]=~/spd(\d+) \=/i;

    $id=$output[0];

    return $id;
}

################################################################################
# Subroutine Name: _stop_spd_id
# Objective:       stop sanpath device
# Details:         See POD documentation at the end of this file
################################################################################

sub _stop_spd_id {
    my ($self,$spd_id)=@_;
    my (
	$cmd,
	$id,
	$status,
	@output,
	);
    
    $cmd = "setsp -T -l".$spd_id;  	

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
	$self->_set_err( "Error: Sanpath::_stop_spd_id() Unable to stop sanpath device id\'$spd_id\'",
			 @output);
	
    }

    return;
}
################################################################################
# Subroutine Name: _remove_spd_id
# Objective:       remove sanpath device
# Details:         See POD documentation at the end of this file
################################################################################

sub _remove_spd_id {
    my ($self,$spd_id)=@_;
    my (
	$cmd,
	$id,
	$status,
	@output,
	);
    
    $cmd = "setsp -x -l".$spd_id;  	

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
    if ($status gt 0) {
	$self->_set_err( "Error: Sanpath::_remove_spd_id() Unable to remove sanpath device id\'$spd_id\'",
			 @output);
	
    }

    return;
}
################################################################################
# Subroutine Name: _get_disk_id
# Objective:       Return disk id
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_disk_id_OLD {
    my ($self,$device) = @_;

    my(
       $cmd,
       @output,
       $status,		
       );
    
    $self->_clear_err();


    $cmd = "vxdisk -s list  $device | grep diskid | awk '{print \$2}'";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 								 
    if ($status gt 0) {
	$self->_set_err( "Error: SanPath::get_diskid() Unable to get diskid for device \'$device\'",
			 @output);
	
	return $ClientUtil::FALSE;                       
    } 
    
    return $output[0];	
    
}


################################################################################
# Subroutine Name: _is_installed
# Objective:       Check to make sure that the LVM is installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_installed {
    my ($self) = @_;
    my (
	$cmd, 
	@output, 
	$status,
	);

    $self->_clear_err();


    $cmd = "lslpp -l | grep lvm";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($status == 0) {
	$self->_log("lvm is installed");
        return $ClientUtil::TRUE;
    } 
    else {
	$self->_log("lvm is not installed");
	$self->_set_err( "Error: Command ($cmd) returned a bad status ($status)");
	return $ClientUtil::FALSE;                             
    }  

}


###############################################################################
# Method:    _log
# Objective: Log test output to stdout and file
# Details:   See POD documentation at the end of this file.
###############################################################################
sub _log{
    my ($self,@line) = @_;
    my (
	$bFirst,
	$dash,
	);
    
    $bFirst=1;
    
    
    if (defined $self->{log}){

	foreach my $line (@line){
	    

	    $dash = ($bFirst == 1) ? "- " : "  ";
	    $bFirst=0;

	    print {${$self->{log}}} "   ".$dash.$line."\n";
    }

  Tharn::tstlog(-comment,@line);
}

}


1;

