# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to the RDM Disk files on ESX
## @author hrivera@netapp.com
## @status private
## @pod ../../Lvm.pm

# Package Name
##############
package San::SanClient::Lvm::ESX::RDM;

use strict;

# Package Requirements
######################
use File::Basename;
use POSIX;
use Data::Dumper;

use San::SanClient::Lvm;
use San::SanClient::ClientUtil;
use San::SanUtil;

use Error;

use vars qw( @ISA );

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: create_vmdks()
# Objective:       creates rdm vmdk files for each vmhba device in @rdm_list
# Returns:         a list of rdm vmdk file paths.
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vmdks {
    my ($self, %args) = @_;

    my ($cmd,
        @output,
        $status,
        @rdm_devices,
        $controller,
        $rdm_type,
        $vm_name,
        $flag,
        @rdm_vmdk_files
      );

    # parse arguments
    foreach (keys %args) {
        if (/-?rdm_devices/) {
            if (not scalar(@rdm_devices = @{$args{$_}})) {
                $self->_set_err("Error: Lvm::ESX::RDM::create_vmdks() ".
                                "Invalid value for argument '-rdm_devices'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?controller/) {
            if (not length($controller = $args{$_})
                    or $controller !~ /^(lsilogic|buslogic)$/ ) {
                $self->_set_err("Error: Lvm::ESX::RDM::create_vmdks() ".
                                "Invalid value for argument '-controller'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?rdm_type/) {
            if (not length($rdm_type = $args{$_})
                    or $rdm_type !~ /^(physical|virtual)$/ ) {
                $self->_set_err("Error: Lvm::ESX::RDM::create_vmdks() ".
                                "Invalid value for argument '-rdm_type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::RDM::create_vmdks() ".
                                "Invalid value for argument '-vm_name'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@rdm_devices)) {
        $self->_set_err("Error: Lvm::ESX::RDM::create_vmdks() " .
                        "missing required argument \'-rdm_devices\'.");
        return (wantarray()) ? () : [];
    } elsif (not defined($controller)) {
        $self->_set_err("Error: ESX:RDM::create_vmdks() ".
                        "missing required argument '-controller'. ".
                        "Use lsilogic or buslogic.");
        return (wantarray()) ? () : [];
    } elsif (not defined($rdm_type)) {
        $self->_set_err("Error: ESX:RDM::create_vmdks() ".
                        "missing required argument '-rdm_type'. ".
                        "Use 'physical' or 'virtual'.");
        return (wantarray()) ? () : [];
    } elsif (not defined($vm_name)) {
        $self->_set_err("Error: ESX:RDM::create_vmdks() ".
                        "missing required argument '-vm_name'.");
        return (wantarray()) ? () : [];
    }

    # Since we will store the RDM vmdk files with the VM's files,
    # we will need to know the path of the VM's config file
    my $vm_config_file_path =
        $self->{parent}->_get_vm_config_file(-vm_name => $vm_name);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return (wantarray()) ? () : [];
    }

    # pull out the directory name of the vmx file path
    my $vm_directory = &File::Basename::dirname($vm_config_file_path);
    my $rdm_vmdk_directory = $vm_directory;

    # set the rdm command flag
    if ($rdm_type =~ /physical/i) {
        $flag = '-z'; # pass-through (physical)
    } elsif ($rdm_type =~ /virtual/i) {
        $flag = '-r'; # non-pass-through (virtual)
    } else {
        $self->_set_err("Error: Lvm::ESX::RDM::create_vmdks()" .
                        " Invalid rdm type. Use 'physical' or 'virtual'");
        return (wantarray()) ? () : [];
    }

    my $retry_count = 1;

    # for each rdm device, construct a vmdk file path and create the vmdk
    foreach my $device (@rdm_devices) {
        # HACK: if this device is a mpdev device, then we need
        # to make it backwards compatible with DevMgr
        if (ref($device) =~ /MPDev/) {
            $device->{vmkdisk_name} = $device->mp_device_name();
            $device->{filer_name} = $device->ontap_controller();
            $device->{lun_path} = $device->ontap_lun_path();
        }

        my $vmkdisk_name = $device->{vmkdisk_name};

        my ($vmdk_file_path, $rdm_device);

        # if on ESX4 we need to not add the :0 for RDM disks
        if ($self->{parent}->get_os_version() =~ /^4\./) {
            $vmdk_file_path = $rdm_vmdk_directory . "/nate_rdm_" .
                              $vmkdisk_name . ".vmdk";
            $rdm_device = "/vmfs/devices/disks/" .  $vmkdisk_name;
        } else {
            $vmdk_file_path = $rdm_vmdk_directory . "/nate_rdm_" .
                              $vmkdisk_name . ":0.vmdk";
            $rdm_device = "/vmfs/devices/disks/" .  $vmkdisk_name . ":0";
        }

        my $lun_path = $device->{filer_name} . ": " . $device->{lun_path};
        $self->sanlog(-msg => "Creating RDM VMDK for $vmkdisk_name ($lun_path)");

        # create the vmdk
        $cmd = "/usr/sbin/vmkfstools $flag $rdm_device $vmdk_file_path -a $controller";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if ($status > 0) {
            # sometimes an orphaned vmdk file may be lingering around that has
            # the same name as the one we are creating. delete it and retry.
            if (grep(/Failed to create virtual disk: The file already exists/i, @output) and
                $retry_count == 1) {
                $cmd = "/usr/sbin/vmkfstools -U $vmdk_file_path";
                ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
                if ($status > 0) {
                    # somtimes it can't be removed, in this case, just remove the file.
                    $cmd = "rm -f $vmdk_file_path";
                    ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);

                    # also remove the rdmp mapping file
                    my $rdmp_file_mapping = $vmdk_file_path;
                    $rdmp_file_mapping =~ s/\.vmdk/\-rdmp\.vmdk/;
                    $cmd = "rm -f $rdmp_file_mapping";
                    ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
                }
                &SanUtil::nate_sleep(5);
                $retry_count = 0;
                redo;
            }

            $self->_set_err("Error: Lvm::ESX::RDM::create_vmdks() Command ($cmd) " .
                            "returned a bad status ($status): You may have an orhpaned ".
                            "vmdk lun mapping.");
            return (wantarray()) ? () : [];
        }

        $retry_count = 1;
        push(@rdm_vmdk_files, $vmdk_file_path);
    }

    $self->{rdm_vmdk_files} = @rdm_vmdk_files;
    $self->sanlog(-msg => "RDM VMDKs created");

    return (wantarray()) ? @rdm_vmdk_files : \@rdm_vmdk_files;
}

################################################################################
# Subroutine Name: destroy_vmdks()
# Objective:       destroys the rdm vmdk files on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vmdks {
    my ($self, %args) = @_;

    my ($cmd, @output, $status);
    my ($vm_name, @rdm_vmdk_files);

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::RDM::destroy_vmdks() ".
                                "Invalid value for argument '-vm_name'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required arguments
    if (not defined($vm_name)) {
        $self->_set_err("Error: Lvm::ESX::RDM::destroy_vmdks() ".
                        "Missing required argument -vm_name");
        return $ClientUtil::FALSE;
    }

    # get the rdm vmdk files in use by $vm_name
    @rdm_vmdk_files = $self->{parent}->get_vmdks_from_vm(-vm_name => $vm_name,
                                                         -vmdk_filepath_regex_filter => qr/nate_rdm_/i);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # search for nate rdm vmdk files that may be lingering around in the config file directory
    # for this vm and add them to the list of vmdks if they are not already in there
    @rdm_vmdk_files = $self->_find_spare_rdm_vmdks(-vm_name => $vm_name,
                                                   -vmdk_filepath_filter => "nate_rdm_",
                                                   -filter_out => \@rdm_vmdk_files);
    if ($self->error()) {
        $self->_set_err("Failed to find spare RDM VMDKs", $self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "";
    my $retry_count = 0;
    my $num_vmdks_destroyed = 0;

    # destroy each rdm vmdk file
    foreach my $rdm_vmdk_file_path (@rdm_vmdk_files) {
        $cmd = "/usr/sbin/vmkfstools -U $rdm_vmdk_file_path";

        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if ($status > 0) {
            # if its a "system cannot find the file specified" error,
            # log a comment concerning it, but dont set an error, just continue.
            if (grep {/cannot find the file specified/i} @output) {
                $self->sanlog(-msg => "Warning: No RDM VMDK was destroyed. ".
                                      "The command '$cmd' attempted to ".
                                      "remove a vmdk that does not exist.");
            } else {
                $self->_set_err("Error: Lvm::ESX::RDM::destroy_vmdks() ".
                                "Could not destroy $rdm_vmdk_file_path");
                return $ClientUtil::FALSE;
            }
        }

        # give esx some time to process the command
        &SanUtil::nate_sleep(1);

        # verify that the vmdk file was destroyed.
        $cmd = "ls $rdm_vmdk_file_path";
        ($status, @output) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
        if ($status == 0) {
            # give up and move on if we tried to delete this vmdk file 4 times.
            if ($retry_count > 4) {
                $retry_count = 0;
                $self->sanlog(-msg => "WARNING: tried to delete $rdm_vmdk_file_path ".
                                      "4 times with no success, moving on.");
                next;
            }

            # the file was not deleted properly, sleep and try again.
            &SanUtil::nate_sleep(1);
            $retry_count++;
            redo;
        }

        # reset the retry count
        $retry_count = 0;
        $num_vmdks_destroyed++;
    }

    # delete the nate rdms directory for this VM
    my $vm_config_file_path = $self->{parent}->_get_vm_config_file(-vm_name => $vm_name);
    if ($self->{parent}->error() or
        not length($vm_config_file_path)) {
        $self->_set_err("Failed to obtain config file path for $vm_name",
                        $self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # delete the nate rdms directory if it exists
    my $nate_rdms_dir = dirname($vm_config_file_path) . "/nate_rdms";
    $cmd = "rm -rf $nate_rdms_dir";
    $self->{parent}->_esx_execute_cmd(-cmd => $cmd);

    # if there are no nate rdm vmdk files in use by this guest,
    # then just return.
    if (scalar(@rdm_vmdk_files) == 0) {
        $self->sanlog(-msg => "No RDM VMDKs to destroy for $vm_name");
        return $ClientUtil::TRUE;
    }

    $self->sanlog(-msg => "($num_vmdks_destroyed) RDM VMDK(s) destroyed for $vm_name");

    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my (
        $lvm,
        $parent,
        $self,
       );

    $self = {};

    bless ($self, $class);

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_}) or $lvm !~ /^rdm$/) {
                $self->_set_err("Error: Lvm::ESX::RDM::_new() ".
                                "Invalid value for argument '-lvm'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::RDM::_new() ".
                                "Invalid value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -lvm argument
    if (not defined ($lvm)) {
        $self->_set_err("Error: Lvm::ESX::RDM::_new()",
                        "Missing required parameter -lvm'");
        return $self;
    }

    # check for existance of required -parent argument
    if ((not defined ($parent)) or
        (ref($parent) !~ /SanClient/)) {
        $self->_set_err("Error: Lvm::ESX::RDM::_new()",
                        "Missing or Invalid value for argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;

    return $self;
}

################################################################################
# Subroutine Name: _find_spare_rdm_vmdks()
# Objective:       Find spare RDM vmdks that match the filter
################################################################################
sub _find_spare_rdm_vmdks {
    my ($self, %args) = @_;

    my $vm_name = $args{-vm_name};
    my $filter_string = $args{-vmdk_filepath_filter};
    my $filter_list = $args{-filter_out};

    # convert vm_name to its vmx file path, then convert the vmx file path to datastore format '[...] path/to/vmx'
    my $vm_config_file_path = $self->{parent}->_get_vm_config_file(-vm_name => $vm_name);
    if ($self->{parent}->error()) {
        $self->_set_err("Failed to obtain config file path for $vm_name",
                        $self->{parent}->errmsg());
        return (wantarray()) ? () : [];
    }

    # parse the vmx directory
    my ($directory) = File::Basename::dirname($vm_config_file_path);

    # search the vmx directory for nate_rdm files
    # verify that the vmdk file was destroyed.
    my $cmd = "ls '$directory' | cat | grep $filter_string";
    my ($status, @output) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
    if ($status == 0) {
        # we found something, splice out all elements in $filter_list from @output
        # first convert all the files into the proper format
        @output = map {
            "${directory}/$_";
        } @output;

        # remove rdmp files
        @output = grep { $_ !~ /\-rdmp?\.vmdk/ } @output;

        # remove dups
        my %hash;
        foreach my $file (@output, @{$filter_list}) {
            $hash{$file}++;
        }
        my @return_files = keys %hash;

        # return the final list of rdm files
        return (wantarray()) ? @return_files : \@return_files;
    }

    return (wantarray()) ? () : [];
}

1;
