# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to the VMFS Logical Volume Manager on ESX
## @author hrivera@netapp.com
## @status private
## @pod ../../Lvm.pm

# Package Name
##############
package San::SanClient::Lvm::ESX::VMFS;

use strict;

# Package Requirements
######################
use File::Basename;
use POSIX;

use NATE::Log qw(log_global);
use San::SanClient::Lvm;
use San::SanClient::ClientUtil;
use San::SanUtil;

use Error;

use vars qw( @ISA $Log );

$Log = log_global();

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: create_datastores()
# Objective:       creates all vmfs datastores for every device passed into the
#                  -vmfs_devices parameter.
# Details:         See POD documentation at the end of this file
################################################################################
sub create_datastores {
    my ($self, %args) = @_;

    my @vmfs_devices = ();
    my $datastore_type = "one-to-one"; # (one-to-one | concat | multi-initiator)
    my ($cmd, @output, $status, @datastores);
    my ($vmfs_type, $block_size, $hostname, $num_luns_per_datastore);

    # parse args
    foreach (keys %args) {
        if (/-?vmfs_devices/) {
            if (not scalar(@vmfs_devices = @{$args{$_}})) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                                "Invalid value for argument '-vmfs_devices'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?datastore_type/) {
            if (not length($datastore_type = $args{$_}) or
                    $datastore_type !~ /^(one-to-one|concat|multi-initiator)$/i) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                                "Invalid value for argument '-datastore_type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?num_luns_per_dstore/) {
            if (not length($num_luns_per_datastore = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                                "Invalid value for argument '-num_luns_per_dstore'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vmfs_type/) {
            if (not length($vmfs_type = $args{$_}) or
                    $vmfs_type !~ /^(vmfs3|vmfs2)$/) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                                "Invalid value for argument '-vmfs_type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?block_size/) {
            if (not length($block_size = $args{$_}) or
                    $block_size !~ /^([1248][mMkK]|auto)$/i) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                                "Invalid value for argument '-block_size'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?hostname/) {
            if (not length($hostname = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                                "Invalid value for argument '-hostname'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@vmfs_devices)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                        "Missing required parameter '-vmfs_devices'");
        return (wantarray()) ? () : [];
    } elsif (not length($datastore_type)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                        "Missing required parameter '-datastore_type'");
        return (wantarray()) ? () : [];
    } elsif (not length($num_luns_per_datastore)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                        "Missing required parameter '-num_luns_per_dstore'");
        return (wantarray()) ? () : [];
    } elsif (not length($vmfs_type)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                        "Missing required parameter '-vmfs_type'");
        return (wantarray()) ? () : [];
    } elsif (not length($block_size)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                        "Missing required parameter '-block_size'");
        return (wantarray()) ? () : [];
    } elsif (not length($hostname)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                        "Missing required parameter '-hostname'");
        return (wantarray()) ? () : [];
    }

    # auto generate the blocksize if requested.
    if ($block_size =~ /^auto$/i) {
        $block_size = $self->_calculate_block_size($vmfs_devices[0],
                                                   $datastore_type,
                                                   $num_luns_per_datastore);
    }

    $self->sanlog(-msg => "VMFS Datastore type is set to '$datastore_type'");

    # for "one-to-one", we will make one datastore for each vmfs lun.
    if ($datastore_type =~ /one-to-one|multi-initiator/i) {

        # for each vmfs device, tack on :0 at the end, fdisk it,
        # and run vmkfstools on the first partition.
        foreach my $vmfs_device (@vmfs_devices) {
            # HACK: if this device is a mpdev device, then we need
            # to make it backwards compatible with DevMgr
            if (ref($vmfs_device) =~ /MPDev/) {
                $vmfs_device->{vmkdisk_name} = $vmfs_device->mp_device_name();
                $vmfs_device->{filer_name} = $vmfs_device->ontap_controller();
                $vmfs_device->{lun_path} = $vmfs_device->ontap_lun_path();
                $vmfs_device->{filer_id} = $vmfs_device->ontap_node_id();
            }

            my $vmfs_main_lun;

            if ($self->{parent}->get_os_version() =~ /^4\./) {
                # no change is needed for ESX4
                $vmfs_main_lun = $vmfs_device->{vmkdisk_name};
            } else {
                # concat the :0 to the end of the lun for ESX3
                $vmfs_main_lun = $vmfs_device->{vmkdisk_name} . ":0";
            }

            my $lun_path = $vmfs_device->{filer_name} . ": " . $vmfs_device->{lun_path};

            $self->sanlog(-msg => "Formating $vmfs_main_lun ($lun_path)");

            # fdisk the lun
            $cmd = "echo 'n\np\n1\n\n\nt\nfb\nx\nb\n1\n128\np\nw\n' ".
                   "| fdisk /vmfs/devices/disks/" . $vmfs_main_lun ;
            ($status, @output) = &ClientUtil::execute($self->{parent}->{conn},
                                                     $cmd);
            if ($status > 0) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                                "Could not fdisk /vmfs/devices/disks/$vmfs_main_lun");
                return (wantarray()) ? () : [];
            }

            # run vmkfstools to create the vmfs filesystem (datastore)
            # on the partition we made. This requires we concat :1 to the
            # end of $vmfs_device
            my $vmfs_partition = $vmfs_device->{vmkdisk_name} . ":1";
            my $datastore_name = "nate_datastore_" . $hostname . "_" .
                                 $vmfs_device->{vmkdisk_name};
            if ($datastore_type =~ /multi-initiator/i) {
                # tag the datastore name with "multi${filer_id}" if its a multi_initiator type datastore.
                $datastore_name = "nate_datastore_" . $hostname . "_" .
                                  $vmfs_device->{vmkdisk_name} . "_multi" . $vmfs_device->{filer_id};
            }

            $self->sanlog(-msg => "Creating VMFS Datastore '$datastore_name'");
            $cmd = "/usr/sbin/vmkfstools -C $vmfs_type -b $block_size ".
                   "-S $datastore_name ".
                   "/vmfs/devices/disks/" . $vmfs_partition;
            ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
            if ($status > 0) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                                "Could not create datastore on " .
                                "/vmfs/devices/disks/$vmfs_partition");
                return (wantarray()) ? () : [];
            }

            # store this datastore in our datastores array
            push(@datastores, $datastore_name);
        }
    }

    # for "concat," we will make one datastore that holds at most 16 luns
    elsif ($datastore_type =~ /concat/i) {

        my $on_first_vmfs_device = "yes";
        my $first_vmfs_partition;
        my $lun_extent_count = 0;

        # for each vmfs device, fdisk the ':0' partition, then run vmkfstools
        # on the ':1' partition.
        # We will concatenate each ':1' partition with the first vmfs
        # partition that we created.
        # Since each datastore can only have 16 extents (luns),
        # we need to create a new datastore for every $num_luns_per_datastore luns.
        foreach my $vmfs_device (@vmfs_devices) {
            # concat the :0 to the end of the lun.
            my $vmfs_main_lun = $vmfs_device->{vmkdisk_name} . ":0";
            my $lun_path = $vmfs_device->{filer_name} . ": " . $vmfs_device->{lun_path};

            $self->sanlog(-msg => "Formatting $vmfs_main_lun ($lun_path)");

            # fdisk the lun
            $cmd = "echo 'n\np\n1\n\n\nt\nfb\nx\nb\n1\n128\np\nw\n' " .
                   "| fdisk /vmfs/devices/disks/" . $vmfs_main_lun ;
            ($status, @output) = &ClientUtil::execute($self->{parent}->{conn},
                                                     $cmd);
            if ($status > 0) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() ".
                                "Could not fdisk /vmfs/devices/disks/$vmfs_main_lun");
                return (wantarray()) ? () : [];
            }

            # each datastore can only have a maximum of $num_luns_per_datastore extents.
            if ($lun_extent_count == $num_luns_per_datastore) {
                $lun_extent_count = 0;
                $on_first_vmfs_device = "yes";
            }

            # if we are on the first vmfs lun from the list, create
            # the datastore that we will concat to.
            if ($on_first_vmfs_device =~ /yes/) {
                $on_first_vmfs_device = "no";

                # run vmkfstools to create the vmfs filesystem (datastore)
                # on the partition we made. This requires we concat :1 to the
                # end of $vmfs_device->{vmkdisk_name}
                $first_vmfs_partition = $vmfs_device->{vmkdisk_name} . ":1";
                my $datastore_name = "nate_datastore_" . $hostname . "_" .
                                     $vmfs_device->{vmkdisk_name};
                $self->sanlog(-msg => "Creating VMFS Datastore '$datastore_name' " .
                                      "(cnt=$lun_extent_count)");
                $cmd = "/usr/sbin/vmkfstools -C $vmfs_type -b $block_size ".
                       "-S $datastore_name ".
                       "/vmfs/devices/disks/" . $first_vmfs_partition;

                # store this datastore in our datastores array
                push(@datastores, $datastore_name);
            } else {
                # otherwise, concat (extend) the :1 partition of this vmfs lun
                # to $first_vmfs_partition by using vmkfstools -Z
                my $vmfs_partition = $vmfs_device->{vmkdisk_name} . ":1";
                $self->sanlog(-msg => "Extending '$first_vmfs_partition' with " .
                                      "'$vmfs_partition'. (cnt=$lun_extent_count)");
                $cmd = "echo -e '0' | vmkfstools -Z " .
                       "$vmfs_partition $first_vmfs_partition";
            }

            # run the vmkfs command
            ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
            if ($status > 0) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_datastores() " .
                                "Vmkfstools failed!");
                return (wantarray()) ? () : [];
            }

            # increment the number of lun extents that are in this datastore
            $lun_extent_count += 1;
        }
    }

    # store the datastore array and datastore type in the lvm and return
    # the datastore array to the caller.
    $self->{datastores} = @datastores;
    $self->{datastore_type} = $datastore_type;

    my $num_datastores_created = scalar(@datastores);

    $self->sanlog(-msg => "($num_datastores_created) VMFS Datastore(s) created");

    return (wantarray()) ? @datastores : \@datastores;
}

################################################################################
# Subroutine Name: create_vmdks()
# Objective:       creates vmfs vmdk files for each datastore
# Returns:         a list of vmfs vmdk file paths.
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vmdks {
    my ($self, %args) = @_;

    my (@datastores, @vmfs_vmdk_files);
    my ($cmd, @output, $status, $controller, $num_vmfs_vmdks, $vmfs_vmdk_size_kbytes, $vmfs_vmdk_disk_type, $suffix, $reuse);
    my $datastore_type = "one-to-one"; # (one-to-one | concat | multi-initiator)
    my $percent = 100;
    $vmfs_vmdk_disk_type = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?datastores/) {
            if (not scalar(@datastores = @{$args{$_}})) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_vmdks() " .
                                "Invalid value for argument '-datastores'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?controller/) {
            if (not length($controller = $args{$_})
                    or $controller !~ /^(lsilogic|buslogic)$/) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_vmdks() " .
                                "Invalid value for argument '-controller'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?datastore_type/) {
            if (not length($datastore_type = $args{$_})
                    or $datastore_type !~ /^(one-to-one|concat|multi-initiator)$/i ) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_vmdks() " .
                                "Invalid value for argument '-datastore_type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?num_vmfs_vmdks/) {
            if (not length($num_vmfs_vmdks = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_vmdks() " .
                                "Invalid value for argument ".
                                "'-num_vmfs_vmdks'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vmfs_vmdk_size_kbytes/) {
            if (not length($vmfs_vmdk_size_kbytes = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::VMFS::create_vmdks() " .
                                "Invalid value for argument ".
                                "'-vmfs_vmdk_size_kbytes'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?datastore_usage_percentage/) {
            if (not length($percent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-datastore_usage_percentage'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vmfs_vmdk_disk_type/) {
            if (length $args{$_}) {
                $vmfs_vmdk_disk_type = $args{$_};
            }
            next;
        } elsif (/-?suffix/) {
            $suffix = $args{$_};
            next;
        } elsif (/-?reuse/) {
            $reuse = $args{$_};
            next;
        }
    }

    # check for required parameters
    if (not scalar(@datastores)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::create_vmdks() " .
                        "missing required argument '-datastores'.");
        return (wantarray()) ? () : [];
    } elsif (not defined($controller)) {
        $self->_set_err("Error: Lvm::ESX:VMFS::create_vmdks() " .
                        "missing required argument '-controller'." .
                        " Use lsilogic or buslogic.");
        return (wantarray()) ? () : [];
    } elsif ($datastore_type =~ /concat/i and not defined($num_vmfs_vmdks)) {
        $self->_set_err("Error: Lvm::ESX:VMFS::create_vmdks() " .
                        "concatenated -datastore_type requires the " .
                        "argument '-num_vmfs_vmdks'.");
        return (wantarray()) ? () : [];
    }

    # calculate the number of vmfs vmdks that will be made.
    $num_vmfs_vmdks = $num_vmfs_vmdks / scalar(@datastores);

    my $datastore_hash = {};
    my $datastore_count = 0;

    my $esx_hostname = $self->{parent}->hostname();

    # iterate over each datastore and create vmdks out of them
    foreach my $datastore (@datastores) {
        # calculate the size of each vmdk file
        my $vmdk_size_in_kbytes;
        ($vmdk_size_in_kbytes, $datastore) = $self->_calculate_vmdk_size($datastore,
                                                                         $num_vmfs_vmdks,
                                                                         $percent);
        if($self->error()) {
            return (wantarray()) ? () : [];
        }

        # if the calling function passed in the size to use then use that instead.
        if (defined($vmfs_vmdk_size_kbytes)) {
            # convert all K's to lowercase
            $vmfs_vmdk_size_kbytes =~ s/K/k/g;

            # append a lowercase 'k' if it does not exist.
            if ($vmfs_vmdk_size_kbytes !~ /k\s*$/) {
                $vmfs_vmdk_size_kbytes .= "k";
            }

            $vmdk_size_in_kbytes = $vmfs_vmdk_size_kbytes;
        }

        # convert the datastore to it's uuid value
        my $datastore_uuid;
        ($datastore_uuid, $datastore) = $self->_convert_datastore_to_uuid($datastore);
        if($self->error()) {
            return (wantarray()) ? () : [];
        }

        # remove spaces, parentheses, and back slashes from the datastore name
        $datastore =~ s/\s/_/g;
        $datastore =~ s/\(//g;
        $datastore =~ s/\)//g;
        $datastore =~ s/\\//g;

        my $retry_count = 1;

        # create $num_vmfs_vmdks files in this datastore that have size
        # $vmdk_size_in_kbytes
        for (my $i = 0; $i < $num_vmfs_vmdks; $i++) {
            my $vmdk_file_name = "/vmfs/volumes/" .$datastore_uuid . "/" .
                                 $datastore . "_${esx_hostname}";
            $vmdk_file_name = length($suffix) ? $vmdk_file_name . "_$suffix" . "_$i.vmdk" : $vmdk_file_name . "_$i.vmdk";

            # now create the VMFS vmdk file
            ($status, @output) = $self->{parent}->create_vmdk(-vmdk_file_path      => $vmdk_file_name,
                                                              -disk_format         => $vmfs_vmdk_disk_type,
                                                              -adapter_type        => $controller,
                                                              -vmdk_size_in_kbytes => $vmdk_size_in_kbytes);
            if ($self->{parent}->error()) {
                $self->_set_err("Failed to create vmdk", $self->{parent}->errmsg());
                return (wantarray()) ? () : [];
            }

            if ($status > 0) {
                if (grep(/Failed to create virtual disk: The file already exists/i, @output)) {
                    if ($reuse == 1) {
                        # do nothing (reuse existing VMDK)
                        $Log->comment("Reusing existing VMDK $vmdk_file_name");
                    } elsif ($retry_count == 1) {
                        # sometimes an orphaned vmdk file may be lingering around that has
                        # the same name as the one we are creating. delete it and retry.
                        $cmd = "/usr/sbin/vmkfstools -U $vmdk_file_name";
                        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
                        if ($status > 0) {
                            $self->_set_err("Could not destroy the orphaned vmdk file " .
                                            "'$vmdk_file_name'");
                            return (wantarray()) ? () : [];
                        }
                        &SanUtil::nate_sleep(5);
                        $retry_count = 0;
                        redo;
                    } else {
                        $self->_set_err("Could not create the VMDK '$vmdk_file_name' " .
                                        "it already exists.");
                        return (wantarray()) ? () : [];
                    }
                } elsif (grep(/Failed to create virtual disk: Device or resource busy/i, @output)) {
                    # sometimes the vmdk creation can fail because a resource or device is busy.
                    # Sleep for a bit and try again
                    &SanUtil::nate_sleep(30);
                    redo;
                } else {
                    $self->_set_err("Could not create the VMDK '$vmdk_file_name'");
                    return (wantarray()) ? () : [];
                }
            }

            $retry_count = 1;
            push(@vmfs_vmdk_files, $vmdk_file_name);
        }

        # We need to balance the vmdk files across all the controllers here.
        # We assume that @datastores contains an alternating list of datastores evenly balanced
        # across the controllers.
        # Build a hash where each element is a datastore and each value is @vmfs_vmdk_files.
        @{$datastore_hash->{$datastore_count}} = @vmfs_vmdk_files;
        $datastore_count++;
        @vmfs_vmdk_files = ();
    }

    my @balanced_vmfs_vmdk_file_list;

    # Shift one vmdk file out of each datastore and push it on the balanced list
    # until all the vmdk files have been added to the balanced list.
    while (keys %{$datastore_hash}) {
        my @keys = keys %{$datastore_hash};
        foreach my $datastore_id (sort {$a <=> $b} @keys) {
            my $vmdk_file = shift(@{$datastore_hash->{$datastore_id}});
            # add this vmdk file to the balanced list.
            push(@balanced_vmfs_vmdk_file_list, $vmdk_file);

            # if this datastore is empty, then remove it from the hash.
            if (scalar(@{$datastore_hash->{$datastore_id}}) <= 0) {
                delete($datastore_hash->{$datastore_id});
            }
        }
    }

    # store the list in the lvm object
    $self->{vmdk_files} = @balanced_vmfs_vmdk_file_list;
    my $num_vmfs_vmdks_created = scalar(@balanced_vmfs_vmdk_file_list);

    $self->sanlog(-msg => "($num_vmfs_vmdks_created) VMFS VMDK(s) created");

    return (wantarray()) ? @balanced_vmfs_vmdk_file_list : \@balanced_vmfs_vmdk_file_list;
}

################################################################################
# Subroutine Name: destroy_datastores()
# Objective:       destroys datastores that are no longer in use.
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_datastores {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, @datastores, $hostname);
    my $count = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?hostname/) {
            if (not length($hostname = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::VMFS::destroy_datastores() " .
                                    "Invalid value for argument '-hostname'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (not defined($hostname)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::destroy_datastores() " .
                        "Missing required argument '-hostname'");
        return $ClientUtil::FALSE;
    }

    # list all the datastore names on the host.
    $cmd = "ls /vmfs/volumes/ | cat";
    ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Lvm::ESX::VMFS::destroy_datastores() " .
                        "Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    # pull out only the datastores that were created by nate and belong to
    # this host, ie: search for "nate_datastore_$hostname_"
    @datastores = grep {/nate_datastore_($hostname)_/} @output;

    # run the vmdk remove command for each vmdk file
    foreach my $datastore (@datastores) {

        # skip this item if its empty.
        next if ($datastore =~ /^\s*$/);

        # sometimes ESX magically places a \s(\d+) at the end of a datastore
        # if it detects duplicate datastore names that have different UUIDs
        # across multiple ESX machines. Escape the parentheses and spaces.
        $datastore =~ s#\(#\\\(#g;
        $datastore =~ s#\)#\\\)#g;
        $datastore =~ s# #\\ #g;

        # if the datastore has vmdk files in it, then it is still in use
        # by some vm, and should not be destroyed.
        $cmd = "ls /vmfs/volumes/". $datastore;
        ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
        if ($status > 0) {
            # this error most likely means that the datastore link is bad and
            # was deleted improperly. Ignore this datastore, ESX will clean it
            # up next time it creates a datastore with that name, or reboots.
            $self->sanlog(-msg => "Warning: The datastore '$datastore' " .
                                  "is a dangling symbolic link. Its LUN was probably orphaned.");
            next;
        }
        if (scalar(@output) > 0) {
            # datastore is not empty, this is very bad
            $self->sanlog(-msg => "WARNING WARNING: The VMFS Datastore '$datastore' belongs to this host and still has VMDKs in it that are \n" .
                                  "not associated with any of the virtual machines that are a part of this test. This is probably going to \n" .
                                  "cause severe problems the next time you try to configure this host.");
            next;
        }

        # if the name has a '/' at the end, chop it off.
        chop($datastore) if ($datastore =~ /\/$/);

        # remove the symbolic link of the datastore so we dont get any
        # broken links.
        $cmd = "rm -f /vmfs/volumes/$datastore";
        ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Lvm::ESX::VMFS::destroy_datastores() " .
                            "Could not remove the symbolic link " .
                            "/vmfs/volumes/$datastore");
            return $ClientUtil::FALSE;
        }

        # remove the datastore from the vmware db
        $cmd = "vimsh -n -e \"hostsvc/datastore/remove $datastore\" 2>&1";
        ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Lvm::ESX::VMFS::destroy_datastores() " .
                            "Could not remove /vmfs/volumes/$datastore from the Datastore database.");
            return $ClientUtil::FALSE;
        }

        $self->sanlog(-msg => "VMFS Datastore '$datastore' has been removed.");
        $count++;
    }

    $self->sanlog(-msg => "VMFS Datastore Cleanup is complete. " .
                          "($count) Datastore(s) has/have been removed.");
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy_vmdks()
# Objective:       destroys the vmfs vmdk files in use by -vm_name
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vmdks {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, @datastore_paths, $datastore);
    my ($vm_name, @vmfs_vmdk_files);

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::VMFS::destroy_vmdks() " .
                                "Invalid value for argument '-vm_name'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required arguments
    if (not defined($vm_name)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::destroy_vmdks() " .
                        "Missing required argument -vm_name");
        return $ClientUtil::FALSE;
    }

    # get the vmfs vmdk files in use by $vm_name
    @vmfs_vmdk_files = $self->{parent}->get_vmdks_from_vm(-vm_name => $vm_name,
                                                          -vmdk_filepath_regex_filter => qr/nate_datastore_/i);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # if there are no nate vmfs vmdk files in use by this guest,
    # then just return.
    if (scalar(@vmfs_vmdk_files) == 0) {
        $self->sanlog(-msg => "No VMFS VMDKs to destroy for $vm_name");
        return $ClientUtil::TRUE;
    }

    my $retry_count = 1;
    my $number_of_vmdks_destroyed = 0;

    # destroy each vmfs vmdk file
    foreach my $vmfs_vmdk_file_path (@vmfs_vmdk_files) {
        $cmd = "/usr/sbin/vmkfstools -U $vmfs_vmdk_file_path";

        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if ($status > 0) {
            # if its a "system cannot find the file specified" error,
            # log a comment concerning it, but dont set an error, just continue.
            if(grep {/cannot find the file specified/i} @output) {
                $self->sanlog(-msg => "Warning: The command '$cmd' attempted " .
                                      "to remove a vmdk that does not exist.");
            } else {
                $self->_set_err("Error: Lvm::ESX::VMFS::destroy_vmdks() " .
                                "Command ($cmd) returned a bad " .
                                "status ($status)");
                return $ClientUtil::FALSE;
            }
        }

        # give esx some time to process the command
        &SanUtil::nate_sleep(1);

        # verify that the vmdk file was destroyed.
        $cmd = "ls $vmfs_vmdk_file_path";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if ($status == 0 and $retry_count == 1) {
            # the file was not deleted properly, sleep and try again.
            &SanUtil::nate_sleep(1);
            $retry_count = 0;
            redo;
        }

        $number_of_vmdks_destroyed++;
        $retry_count = 1;
    }

    $self->sanlog(-msg => "($number_of_vmdks_destroyed) VMFS VMDK(s) destroyed for $vm_name");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy_spare_vmdks()
# Objective:       destroys spare vmdks passed into the function
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_spare_vmdks {
    my ($self, %args) = @_;

    my ($cmd, @output, $status);
    my (@vmfs_vmdk_files);

    # parse arguments
    foreach (keys %args) {
        if (/-?vmfs_vmdk_files/) {
            @vmfs_vmdk_files = @{$args{$_}};
        }
    }

    # if there are no nate vmfs vmdk files in use by this guest,
    # then just return.
    if (scalar(@vmfs_vmdk_files) == 0) {
        $self->sanlog(-msg => "No spare VMFS VMDKs to destroy.");
        return $ClientUtil::TRUE;
    }

    my $retry_count = 1;
    my $delete_retry_count = 3;
    my $number_of_vmdks_destroyed = 0;

    # destroy each vmfs vmdk file
    foreach my $vmfs_vmdk_file_path (@vmfs_vmdk_files) {
        $cmd = "/usr/sbin/vmkfstools -U $vmfs_vmdk_file_path";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if ($status > 0) {
            # if its a "system cannot find the file specified" error,
            # log a comment concerning it, but dont set an error, just continue.
            if (grep {/cannot find the file specified/i} @output) {
                $self->sanlog(-msg => "Warning: The command '$cmd' attempted " .
                                      "to remove a vmdk that does not exist.");
            } elsif (grep {/Failed to delete virtual disk: Resource temporarily unavailable/i} @output and
                     $delete_retry_count > 0) {
                # if we failed to delete it, then sleep a bit and try again.
                &SanUtil::nate_sleep(10);
                $delete_retry_count--;
                redo;
            } else {
                $self->_set_err("Error: Lvm::ESX::VMFS::destroy_spare_vmdks() " .
                                "Command ($cmd) returned a bad " .
                                "status ($status)");
                return $ClientUtil::FALSE;
            }
        }

        # give esx some time to process the command
        &SanUtil::nate_sleep(1);

        # verify that the vmdk file was destroyed.
        $cmd = "ls $vmfs_vmdk_file_path";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if ($status == 0 and $retry_count == 1) {
            # the file was not deleted properly, sleep and try again.
            &SanUtil::nate_sleep(1);
            $retry_count = 0;
            redo;
        }

        $number_of_vmdks_destroyed++;
        $retry_count = 1;
    }

    $self->sanlog(-msg => "($number_of_vmdks_destroyed) Spare VMFS VMDK(s) destroyed.");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_multi_init_datastores()
# Objective:       obtain an alternating list (by filers) of multi initiator
#                  datastores
# Details:         See POD documentation at the end of this file
################################################################################
sub get_multi_init_datastores {
    my ($self, %args) = @_;

    my $balance;

    # parse arguments
    foreach (keys %args) {
        if (/-?balance/) {
            if (not length($balance = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::VMFS::get_multi_init_datastores() " .
                                "Invalid value for argument '-balance'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # list all the datastores on the system.
    my ($cmd, @output, $status);
    $cmd = "ls /vmfs/volumes/ | cat";
    ($status, @output) = &ClientUtil::execute($self->{parent}->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Lvm::ESX::VMFS::get_multi_init_datastores() " .
                        "The command '$cmd' failed.");
        return (wantarray()) ? () : [];
    }

    my $datastore_hash = {};

    # pull out the filer IDs from the datastore name and store in a hash
    foreach my $line (@output) {
        if ($line =~ /nate_datastore_.+_multi(\d+)/) {
            my $filer_id = $1;
            # sometimes ESX magically places a \s(\d+) at the end of a datastore
            # if it detects duplicate datastore names that have different UUIDs
            # across multiple ESX machines. Escape the parentheses and spaces.
            $line =~ s#\(#\\\(#g;
            $line =~ s#\)#\\\)#g;
            $line =~ s# #\\ #g;

            # store this datastore in the hash
            push(@{$datastore_hash->{$filer_id}}, $line);
        }
    }

    my @balanced_datastore_list = ();

    # build the balanced datastore list
    while (keys %{$datastore_hash}) {
        my @keys = keys %{$datastore_hash};
        foreach my $filer_id (sort {$a <=> $b} @keys) {
            my $datastore = shift(@{$datastore_hash->{$filer_id}});
            # add this datastore to the balanced list.
            push(@balanced_datastore_list, $datastore);

            # if there are no datastores left for this filer id, then remove it from the hash.
            if (scalar(@{$datastore_hash->{$filer_id}}) <= 0) {
                delete($datastore_hash->{$filer_id});
            }
        }
    }

    # if a balance parameter was passed in, then reverse the list if $balance is odd
    if (defined($balance)) {
        if (($balance % 2) == 1) {
            # reverse the device list
            @balanced_datastore_list = reverse @balanced_datastore_list;
        }
    }

    return (wantarray()) ? @balanced_datastore_list : \@balanced_datastore_list;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($lvm, $parent, $self);

    $self = {};

    bless ($self, $class);

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_}) or $lvm !~ /^vmfs$/) {
                $self->_set_err("Error: Lvm::ESX::VMFS::_new() " .
                                "Invalid value for argument '-lvm'");
                return $self;
            }
            next;
        }
        elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Lvm::ESX::VMFS::_new() " .
                                "Invalid value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -lvm argument
    if (not defined ($lvm)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::_new()",
                        "Missing required parameter -lvm'");
        return $self;
    }

    # check for existance of required -parent argument
    if ((not defined ($parent)) or
        (ref($parent) !~ /SanClient/)) {
        $self->_set_err("Error: Lvm::ESX::VMFS::_new()",
                        "Missing or Invalid value for argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{datastore_type} = "one-to-one";
    $self->{parent} = $parent;

    return $self;
}

################################################################################
# Subroutine Name: _convert_datastore_to_uuid()
# Objective:       Given the datastore name, convert to the UUID label
# Details:         See POD documentation at the end of this file
################################################################################
sub _convert_datastore_to_uuid {
    my ($self, $datastore) = @_;

    my ($cmd, $status, @output, $uuid);
    my $retry_count = 2;
    my $original_datastore = $datastore;

    while (1) {
        # grab this datastores uuid and return it
        $cmd = "/usr/sbin/vmkfstools -P -h /vmfs/volumes/$datastore | grep UUID | " .
               "awk {'print \$2'}";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if (($status > 0 or grep(/Could not open/, @output)) and $retry_count > 0) {
            # sometimes the datastore named changed because it is shared across multiple hosts.
            # attempt to discover the new name
            my $ls_cmd = "ls /vmfs/volumes/ | grep $original_datastore";
            my ($status, @datastore_output) = &ClientUtil::execute($self->{parent}->{conn}, $ls_cmd);
            if ($status > 0 or scalar(@datastore_output) != 1) {
                $self->_set_err("Error: Lvm::ESX::VMFS::_convert_datastore_to_uuid() ".
                                "Could not find the datastore $original_datastore.");
                return ("", $original_datastore);
            }

            # update the datastore with the new name
            $datastore = $datastore_output[0];

            # escape spaces and parenthesis characters
            $datastore =~ s#\(#\\\(#g;
            $datastore =~ s#\)#\\\)#g;
            $datastore =~ s# #\\ #g;

            # retry the loop
            $retry_count--;
            redo;
        } elsif ($status == 0) {
            # then we successfully ran the command, break out
            last;
        } else {
            # otherwise we could not find this datastore
            $self->_set_err("Error: Lvm::ESX:VMFS::_convert_datastore_to_uuid() ".
                            "Could not obtain the UUID value for the datastore ".
                            "'$original_datastore'");
            return ("", $original_datastore);

        }
    }

    return ($output[0], $datastore);
}

################################################################################
# Subroutine Name: _calculate_vmdk_size()
# Objective:       Given the size of a datastore and the number of vmdks to
#                  carve out of the datastore, calculate the size of each vmdk
#                  in kilobytes.
# Details:         See POD documentation at the end of this file
################################################################################
sub _calculate_vmdk_size {
    my ($self, $datastore, $num_vmdks, $percentage) = @_;

    my ($cmd, $status, @output);
    my ($numbers, $size_letter);
    my $retry_count = 2;
    my $original_datastore = $datastore;

    # default the percentage of the datastore to be 100% utilization
    if (not defined($percentage)) {
        $percentage = 100;
    }
    $percentage /= 100;

    # calculate the amount of space available on this datastore.
    # we will be creating a vmdk file that allocates the maximum size of
    # this datastore's vmfs partition.
    while(1) {
        $cmd = "/usr/sbin/vmkfstools -P -h /vmfs/volumes/$datastore";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if($status > 0 and $retry_count > 0) {
            # sometimes the datastore named changed because it is shared across multiple hosts.
            # attempt to discover the new name
            my $ls_cmd = "ls /vmfs/volumes/ | grep $original_datastore";
            my ($status, @datastore_output) = &ClientUtil::execute($self->{parent}->{conn}, $ls_cmd);
            if ($status > 0 or scalar(@datastore_output) != 1) {
                $self->_set_err("Error: Lvm::ESX::VMFS::_calculate_vmdk_size() " .
                                "Could not find the datastore $original_datastore.");
                return (0, $original_datastore);
            }

            # update the datastore with the new name
            $datastore = $datastore_output[0];

            # escape spaces and parenthesis characters
            $datastore =~ s#\(#\\\(#g;
            $datastore =~ s#\)#\\\)#g;
            $datastore =~ s# #\\ #g;

            # retry the loop
            $retry_count--;
            redo;
        } elsif ($status == 0) {
            # then we successfully ran the command, break out
            last;
        } else {
            # otherwise our retry count has been exhausted. error out
            $self->_set_err("Error: Lvm::ESX:VMFS::_calculate_vmdk_size() " .
                            "Could not obtain VMFS information on the datastore " .
                            "'$original_datastore'");
            return (0, $original_datastore);
        }
    }

    # pull out the capacity of this datastore from the @output
    foreach my $line (@output) {
        # $line can look like:
        # ESX 4:
        # Capacity 3.8 GB, 3.5 GB available, file block size 1 MB
        # ESX 3.5 or lower:
        # Capacity 129G, 129G available, file block size 1.0M
        if ($line =~ /Capacity\s+[0-9\.]+\s*[A-Za-z]+,\s+([0-9\.]+)\s*([A-Za-z]+)/) {
            $numbers = "$1";
            $size_letter = lc($2);

            # convert $numbers to kbytes
            if ($size_letter =~ /(t|g|m)/) {
                # if numbers looks like "45.4" or "45.0"
                if ($numbers =~ /(\d*)[.](\d)/) {
                    my $first_set = $1;
                    my $second_set = $2;

                    $numbers = $first_set . $second_set . "00"
                        if ($size_letter =~ /m/);
                    $numbers = $first_set . $second_set . "00000"
                        if ($size_letter =~ /g/);
                    $numbers = $first_set . $second_set . "00000000"
                        if ($size_letter =~ /t/);
                }
                # otherwise, $numbers looks like "45"
                else {
                    $numbers = $numbers . "000" if ($size_letter =~ /m/);
                    $numbers = $numbers . "000000" if ($size_letter =~ /g/);
                    $numbers = $numbers . "000000000" if ($size_letter =~ /t/);
                }

                $size_letter = "k";
            }

            # now divide the size by the number of vmdks to make
            my $size_of_each_vmdk = $numbers / $num_vmdks;

            # subtract a bit from the size to take into account for
            # the metadata reservation. This is a hack to make ESX happy.
            $size_of_each_vmdk -= 800;

            # floor it to get rid of repeating decimals.
            $size_of_each_vmdk = POSIX::floor($size_of_each_vmdk * $percentage);

            # add the kilobyte letter to the size
            $size_of_each_vmdk .= "k";

            return ($size_of_each_vmdk, $datastore);
        }
    }

    # if we get to this point then we could not find the capacity
    # of the datastore
    $self->_set_err("Error: Lvm::ESX:VMFS::_calculate_vmdk_size() Could not " .
                    "determine capacity of datastore '$datastore'");
    return (0, $datastore);
}

################################################################################
# Subroutine Name: _calculate_block_size()
# Objective:       given the size of device, datastore type, and the number
#                  of vmfs luns to be used per datastore, calculate blocksize
# Details:         See POD documentation at the end of this file
################################################################################
sub _calculate_block_size {
    my ($self,
        $vmfs_device,
        $datastore_type,
        $num_luns_per_datastore) = @_;

    my $range_block_size_1m = 256 * 1024 * 1024 * 1024;  # 256g
    my $range_block_size_2m = 512 * 1024 * 1024 * 1024;  # 512g
    my $range_block_size_4m = 1024 * 1024 * 1024 * 1024; # 1024g

    my $datastore_size = 0;

    # if vmfs device is a MPDev obj, we need to make it bw compatible with DevMgr
    if (ref($vmfs_device) =~ /MPDev/) {
        $vmfs_device->{lun_size} = $vmfs_device->device_size();
    }

    # determine the size of the datastore that will be created.
    if ($datastore_type =~ /concat/i) {
        $datastore_size = $vmfs_device->{lun_size} * $num_luns_per_datastore;
    } else {
        $datastore_size = $vmfs_device->{lun_size};
    }

    # determine the range and return the appropriate blocksize
    if ($datastore_size < $range_block_size_1m) {
        return "1m";
    } elsif ($datastore_size < $range_block_size_2m) {
        return "2m";
    } elsif ($datastore_size < $range_block_size_4m) {
        return "4m";
    } else {
        return "8m";
    }

}

1;
