# Copyright (c) 2010 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to the RDM Disk files on ESXi
## @author hrivera@netapp.com
## @status private
## @pod ../../Lvm.pm

# Package Name
##############
package San::SanClient::Lvm::ESXi::RDM;

use strict;
use warnings;

# Package Requirements
######################
use File::Basename;
use POSIX;
use Data::Dumper;

use San::SanClient::Lvm;
use San::SanClient::ClientUtil;
use San::SanUtil;

use Error;

use vars qw( @ISA );

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: create_vmdks()
# Objective:       creates rdm vmdk files for each vmhba device in @rdm_list
# Returns:         a list of rdm vmdk file paths.
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vmdks {
    my ($self, %args) = @_;

    my ($cmd,
        @output,
        $status,
        @rdm_devices,
        $controller,
        $rdm_type,
        $vm_name,
        $flag,
        @rdm_vmdk_files
      );

    # parse arguments
    foreach (keys %args) {
        if (/-?rdm_devices/) {
            if (not scalar(@rdm_devices = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-rdm_devices'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?controller/) {
            if (not length($controller = $args{$_})
                    or $controller !~ /^(lsilogic|buslogic)$/ ) {
                $self->_set_err("Invalid value for argument '-controller'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?rdm_type/) {
            if (not length($rdm_type = $args{$_})
                    or $rdm_type !~ /^(physical|virtual)$/ ) {
                $self->_set_err("Invalid value for argument '-rdm_type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vm_name'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # check for required parameters
    if (not scalar(@rdm_devices)) {
        $self->_set_err("Missing required argument \'-rdm_devices\'.");
        return (wantarray()) ? () : [];
    } elsif (not defined($controller)) {
        $self->_set_err("Missing required argument '-controller'. ".
                        "Use 'lsilogic' or 'buslogic'.");
        return (wantarray()) ? () : [];
    } elsif (not defined($rdm_type)) {
        $self->_set_err("Missing required argument '-rdm_type'. ".
                        "Use 'physical' or 'virtual'.");
        return (wantarray()) ? () : [];
    } elsif (not defined($vm_name)) {
        $self->_set_err("Missing required argument '-vm_name'.");
        return (wantarray()) ? () : [];
    }

    # Since we will store the RDM vmdk files with the VM's files,
    # we will need to know the path of the VM's config file
    my $vm_config_file_path =
        $self->{parent}->_get_vm_config_file(-vm_name => $vm_name);
    if ($self->{parent}->error()) {
        $self->_set_err($self->{parent}->errmsg());
        return (wantarray()) ? () : [];
    }

    # pull out the directory name of the vmx file path
    my $vm_directory = &File::Basename::dirname($vm_config_file_path);

    # the nate rdms will be located with the virtual machine
    my $rdm_vmdk_directory = $vm_directory;

    # convert the directory to a valid datastore path
    my $rdm_vmdk_datastore_path = $self->{parent}->_convert_to_datastore_path(-file_path => $rdm_vmdk_directory);
    if ($self->{parent}->error()) {
        $self->_set_err("Failed to convert datastore path", $self->{parent}->errmsg());
        return (wantarray()) ? () : [];
    }

    # rdm_vmdk_datastore_path contains the UUID of the datastore rather
    # than the datastore label, the next command requires that we pass
    # in the datastore label rather than the UUID,
    # we'll need to convert from UUID to label
    my $datastore_label;
    my $datastore_uuid;
    my $datastore_type;
    if ($rdm_vmdk_datastore_path =~ /\[(\S+)\]/) {
        $datastore_uuid = $1;
    } else {
        $self->_set_err("Unable to parse UUID from datastore path $rdm_vmdk_datastore_path");
        return (wantarray()) ? () : [];
    }

    ($datastore_label, $datastore_type) = $self->_convert_datastore_uuid_to_label(-datastore_uuid => $datastore_uuid);
    if ($self->error()) {
        return (wantarray()) ? () : [];
    }

    # datastore type has to be VMFS for rdms to work
    if ($datastore_type !~ /VMFS/i) {
        $self->_set_err("Cannot create RDM files on the datastore $datastore_label " .
                        " of type $datastore_type. RDM files can only be stored on VMFS datastores. You will need " .
                        "to perform a storage vmotion of $vm_name to a VMFS datastore for this to work properly.");
        return (wantarray()) ? () : [];
    }

    # now replace the datastore uuid with the datastore label in the datastore
    # path
    $rdm_vmdk_datastore_path =~ s/\[$datastore_uuid\]/\[$datastore_label\]/;

    # set the rdm command flag
    if ($rdm_type =~ /physical/i) {
        $flag = '--createrdmpassthru'; # pass-through (physical)
    } elsif ($rdm_type =~ /virtual/i) {
        $flag = '--createrdm'; # non-pass-through (virtual)
    } else {
        $self->_set_err("Invalid rdm type. Use 'physical' or 'virtual'");
        return (wantarray()) ? () : [];
    }

    my $retry_count = 1;

    # for each rdm device, construct a vmdk file path and create the vmdk
    foreach my $device (@rdm_devices) {
        my $naa_device_name = $device->mp_device_name();
        my $naa_device_path = $device->mp_device_path();

        my $vmdk_file_path = $rdm_vmdk_datastore_path . "/nate_rdm_" . $naa_device_name . ".vmdk";

        my $lun_path = $device->ontap_controller() . ": " . $device->ontap_lun_path();
        $self->sanlog(-msg => "Creating RDM VMDK for $naa_device_name ($lun_path)");

        # create the RDM vmdk
        $cmd = "vmkfstools $flag $naa_device_path '$vmdk_file_path' --adapterType $controller";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if ($status > 0) {
            # sometimes an orphaned vmdk file may be lingering around that has
            # the same name as the one we are creating. delete it and retry.
            if (grep(/Failed to create virtual disk: The file already exists/i, @output) and
                $retry_count == 1) {
                $cmd = "vmkfstools --deletevirtualdisk '$vmdk_file_path'";
                ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
                if ($status > 0) {
                    # somtimes it can't be removed, in this case, just remove the file.
                    $cmd = "vifs --rm '$vmdk_file_path' --force";
                    ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);

                    # also remove the rdmp mapping file
                    my $rdmp_file_mapping = $vmdk_file_path;
                    $rdmp_file_mapping =~ s/\.vmdk/\-rdmp\.vmdk/;
                    $cmd = "vifs --rm '$rdmp_file_mapping' --force";
                    ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);

                    # or the rdm file mapping...
                    my $rdm_file_mapping = $vmdk_file_path;
                    $rdmp_file_mapping =~ s/\.vmdk/\-rdm\.vmdk/;
                    $cmd = "vifs --rm '$rdmp_file_mapping' --force";
                    ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
                }
                &SanUtil::nate_sleep(5);
                $retry_count = 0;
                redo;
            }

            $self->_set_err("Command ($cmd) returned a bad status ($status): You may have an orhpaned ".
                            "vmdk lun mapping.");
            return (wantarray()) ? () : [];
        }

        $retry_count = 1;
        push(@rdm_vmdk_files, $vmdk_file_path);
    }

    $self->{rdm_vmdk_files} = @rdm_vmdk_files;
    $self->sanlog(-msg => "RDM VMDKs created");

    return (wantarray()) ? @rdm_vmdk_files : \@rdm_vmdk_files;
}

################################################################################
# Subroutine Name: destroy_vmdks()
# Objective:       destroys the rdm vmdk files on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vmdks {
    my ($self, %args) = @_;

    my ($cmd, @output, $status);
    my ($vm_name, @rdm_vmdk_files);

    # parse arguments
    foreach (keys %args) {
        if (/-?vm_name/) {
            if (not length($vm_name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vm_name'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required arguments
    if (not defined($vm_name)) {
        $self->_set_err("Error: Lvm::ESX::RDM::destroy_vmdks() ".
                        "Missing required argument -vm_name");
        return $ClientUtil::FALSE;
    }

    # get the vmdk disk filepaths
    @rdm_vmdk_files = $self->{parent}->get_vmdks_from_vm(-vm_name => $vm_name,
                                                         -vmdk_filepath_regex_filter => qr/nate_rdm_/i);
    if ($self->{parent}->error()) {
        $self->_set_err("Failed to obtain VMDKS from vm", $self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    # search for nate rdm vmdk files that may be lingering around in the config file directory
    # for this vm and add them to the list of vmdks if they are not already in there
    my ($relationship_hash, $orphaned_data_file_hash);
    ($relationship_hash, $orphaned_data_file_hash, @rdm_vmdk_files) =
        $self->_find_spare_rdm_vmdks(-vm_name => $vm_name,
                                     -vmdk_filepath_filter => "nate_rdm_",
                                     -filter_out => \@rdm_vmdk_files);
    if ($self->error()) {
        $self->_set_err("Failed to find spare RDM VMDKs", $self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    $cmd = "";
    my $retry_count = 0;
    my $num_vmdks_destroyed = 0;

    # destroy each rdm vmdk file
    foreach my $rdm_vmdk_file_path (@rdm_vmdk_files) {
        $cmd = "vmkfstools --deletevirtualdisk '$rdm_vmdk_file_path'";
        ($status, @output) = $self->{parent}->_vmkfstools_cmd($cmd);
        if ($status > 0) {
            # if its a "system cannot find the file specified" error,
            # log a comment concerning it, but dont set an error, just continue.
            if (grep {/was not found/i} @output) {
                $self->sanlog(-msg => "Warning: No RDM VMDK was destroyed. ".
                                      "The command '$cmd' attempted to ".
                                      "remove a vmdk that does not exist.");
                $retry_count = 0;
                next;
            } else {
                $self->{parent}->_clear_err();
                # hard remove it using vifs (ignore the error)
                $cmd = "vifs --rm '$rdm_vmdk_file_path' --force";
                $self->{parent}->_esx_execute_cmd(-cmd => $cmd);

                # delete the data file if there is a mapping file for it in the relationship hash
                if (exists($relationship_hash->{$rdm_vmdk_file_path})) {
                    my $data_file = $relationship_hash->{$rdm_vmdk_file_path};
                    $cmd = "vifs --rm '$data_file' --force";
                    $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
                }
            }
        }

        # give esx some time to process the command
        &SanUtil::nate_sleep(1);

        # parse the vmdk file name and directory
        my ($directory, $vmdk_filename) = $self->{parent}->_parse_vmdk_file_path(-vmdk_file_path => $rdm_vmdk_file_path);
        if ($self->{parent}->error()) {
            $self->_set_err("Failed to parse vmdk_file_path '$rdm_vmdk_file_path'", $self->{parent}->errmsg());
            return $ClientUtil::FALSE;
        }

        # verify that the vmdk file was destroyed.
        $cmd = "vifs --dir '$directory' | grep $vmdk_filename";
        ($status, @output) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
        if ($status == 0) {
            # give up and move on if we tried to delete this vmdk file 4 times.
            if ($retry_count > 4) {
                $retry_count = 0;
                $self->sanlog(-msg => "WARNING: tried to delete $rdm_vmdk_file_path ".
                                      "4 times with no success, moving on.");
                next;
            }

            # the file was not deleted properly, sleep and try again.
            &SanUtil::nate_sleep(1);
            $retry_count++;
            redo;
        }

        # reset the retry count
        $retry_count = 0;
        $num_vmdks_destroyed++;
    }

    # delete any orphaned rdm vmdk files left over (ignore results)
    foreach my $mapping_file (keys %{$orphaned_data_file_hash}) {
        my $data_file = $orphaned_data_file_hash->{$mapping_file};
        $cmd = "vifs --rm '$data_file' --force";
        $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
    }

    # delete the nate RDM directory for this vm
    $self->_delete_nate_rdm_directory(-vm_name => $vm_name);
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    # if there are no nate rdm vmdk files in use by this guest,
    # then just return.
    if (scalar(@rdm_vmdk_files) == 0) {
        $self->sanlog(-msg => "No RDM VMDKs to destroy for $vm_name");
        return $ClientUtil::TRUE;
    }

    $self->sanlog(-msg => "($num_vmdks_destroyed) RDM VMDK(s) destroyed for $vm_name");

    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my (
        $lvm,
        $parent,
        $self,
       );

    $self = {};

    bless ($self, $class);

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_}) or $lvm !~ /^rdm$/) {
                $self->_set_err("Error: Lvm::ESX::RDM::_new() ".
                                "Invalid value for argument '-lvm'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -lvm argument
    if (not defined ($lvm)) {
        $self->_set_err("Missing required parameter -lvm'");
        return $self;
    }

    # check for existance of required -parent argument
    if ((not defined ($parent)) or
        (ref($parent) !~ /SanClient/)) {
        $self->_set_err("Missing or Invalid value for argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{parent} = $parent;

    return $self;
}

################################################################################
# Subroutine Name: _find_spare_rdm_vmdks()
# Objective:       Find spare RDM vmdks that match the filter
################################################################################
sub _find_spare_rdm_vmdks {
    my ($self, %args) = @_;

    my $vm_name = $args{-vm_name};
    my $filter_string = $args{-vmdk_filepath_filter};
    my $filter_list = $args{-filter_out};

    # convert vm_name to its vmx file path, then convert the vmx file path to datastore format '[...] path/to/vmx'
    my $vm_config_file_path = $self->{parent}->_get_vm_config_file(-vm_name => $vm_name);
    if ($self->{parent}->error()) {
        $self->_set_err("Failed to obtain config file path for $vm_name",
                        $self->{parent}->errmsg());
        return ();
    }
    $vm_config_file_path = $self->{parent}->_convert_to_datastore_path(-file_path => $vm_config_file_path,
                                                                       -format => "datastore_path",
                                                                       -use_uuid => 0);
    if ($self->{parent}->error()) {
        $self->_set_err("Failed to obtain config file path for $vm_name",
                        $self->{parent}->errmsg());
        return ();
    }

    # parse the vmx directory
    my ($directory) = $self->{parent}->_parse_vmdk_file_path(-vmdk_file_path => $vm_config_file_path);
    if ($self->{parent}->error()) {
        $self->_set_err("Failed to parse vmdk_file_path '$vm_config_file_path'",
                        $self->{parent}->errmsg());
        return ();
    }

    my $relationship_hash = {};

    # search the vmx directory for nate_rdm files
    # verify that the vmdk file was destroyed.
    my $cmd = "vifs --dir '$directory' | grep $filter_string";
    my ($status, @output) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
    if ($status == 0) {
        # we found something, splice out all elements in $filter_list from @output
        # first convert all the files into the proper format
        @output = map {
            "${directory}/$_";
        } @output;

        foreach my $file (@output) {
            if ($file =~ /\-rdmp?\.vmdk/) {
                my $mapping_vmdk = $file;
                $mapping_vmdk =~ s/\-rdmp?\.vmdk/\.vmdk/;
                $relationship_hash->{$mapping_vmdk} = $file;
            }
        }

        # remove rdmp files
        @output = grep { $_ !~ /\-rdmp?\.vmdk/ } @output;

        # remove dups
        my %hash;
        foreach my $file (@output, @{$filter_list}) {
            $hash{$file}++;
        }
        my @return_files = keys %hash;

        # scan each file in the relationship_hash, if it is not in there, then we might have
        # orphaned -rdm.vmdk files left on the system
        my $orphaned_data_file_hash = {};
        foreach my $mapping_vmdk (keys %{$relationship_hash}) {
            my $found = 0;
            foreach my $filename (@return_files) {
                if ($mapping_vmdk eq $filename) {
                    $found = 1;
                    last;
                }
            }
            if (not $found) {
                # this is an orphaned mapping vmdk,
                $orphaned_data_file_hash->{$mapping_vmdk} = $relationship_hash->{$mapping_vmdk};
            }
        }

        # return the final list of rdm files
        return ($relationship_hash, $orphaned_data_file_hash, @return_files);
    }

    return ();
}

################################################################################
# Subroutine Name: _delete_nate_rdm_directory()
# Objective:       Remove the nate rdms directory for this VM
# Details:         See POD documentation at the end of this file
################################################################################
sub _delete_nate_rdm_directory {
    my ($self, %args) = @_;

    my $vm_name = $args{-vm_name};

    # delete the nate rdms directory for this VM
    my $vm_config_file_path = $self->{parent}->_get_vm_config_file(-vm_name => $vm_name);
    if ($self->{parent}->error() or
        not length($vm_config_file_path)) {
        $self->_set_err("Failed to obtain config file path for $vm_name",
                        $self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }
    my $vm_directory = &File::Basename::dirname($vm_config_file_path);
    $vm_directory = $self->{parent}->_convert_to_datastore_path(-file_path => $vm_directory,
                                                                -format => "datastore_path",
                                                                -use_uuid => 0);
    if ($self->{parent}->error()) {
        $self->_set_err("Failed to convert config file path to datastore format for $vm_name",
                        $self->{parent}->errmsg());
        return $ClientUtil::FALSE;
    }

    my $nate_rdms_dir = $vm_directory . "/nate_rdms";

    # only delete it if it exists
    my $cmd = "vifs --dir '$nate_rdms_dir'";
    my ($status, @output) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
    if (not grep(/Can not list directory/i, @output)) {
        # then the directory exists,
        # delete the contents of the directory before deleting the directory
        @output = grep { !/^-+$|^\s*$|^\s*Content\s*Listing\s*$/i } @output;
        foreach my $line (@output) {
            $cmd = "vifs --force --rm '$nate_rdms_dir\/$line'";
            $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
        }
        # now delete the nate_rdms/ dir
        $cmd = "vifs --force --rm '$nate_rdms_dir'";
        $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _convert_datastore_uuid_to_label()
# Objective:       Convert datastore UUID to the datastore label
# Details:         See POD documentation at the end of this file
################################################################################
sub _convert_datastore_uuid_to_label {
    my ($self, %args) = @_;

    my $datastore_uuid = $args{-datastore_uuid};

    my $datastore_label;
    my $datastore_type;

    # first try using the esx_datastore_info script
    my $cmd = "esx_datastore_info.pl";
    my ($status, @output) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd);
    if ($status > 0) {
        # if it fails or doesn't exist, then fall back to esxcfg-scsidevs (does not report NFS datastores)
        $cmd = "esxcfg-scsidevs --vmfs | grep $datastore_uuid";
        ($status, @output) = $self->{parent}->_esx_execute_cmd(-cmd => $cmd,
                                                               -retries => 3);
        if ($status > 0) {
            # There is no VMFS datastore for this UUID its probably NFS
            return ($datastore_uuid, "NFS");
        }

        if ($output[0] =~ /^\S+\s+\S+\s+${datastore_uuid}\s+\S+\s+(.+)$/) {
            $datastore_label = $1;
        }
        return ($datastore_label, "VMFS");
    } else {
        # parse out the datastore label
        foreach my $line (@output) {
            if ($line =~ /^\/vmfs\/volumes\/${datastore_uuid}\s+(.+?)\s+(VMFS|NFS)/) {
                $datastore_label = $1;
                $datastore_type = $2;
                last;
            }
        }
    }

    if (not defined($datastore_label)) {
        $self->_set_err("Unable to determine datastore label from datastore UUID '$datastore_uuid'");
        return ();
    }

    return ($datastore_label, $datastore_type);
}

1;
