# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the NativeLvm Logical Volume Manager on HPUX
## @author knopp@netapp.com latif@netapp.com
## @status private
## @pod ../../Lvm.pm

# Package Name
##############
package San::SanClient::Lvm::HPUX::NativeLvm;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanClient::Lvm;
use San::SanLog;
use Error;

require San::SanClient::Lvm::HPUX::NativeLvm::Vg;

use vars qw( @ISA );

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: initialize
# Objective:       initialize disk under native HP-UX LVM control
# Details:         See POD documentation at the end of this file
################################################################################
sub initialize {
    my ($self, %args) = @_;
    my ($cmd, @already_inited, @output, $status, $devices);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: HPUXNativeLvm::initialize() Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required parameters
    unless (defined $devices and length $devices) {
        $self->_set_err("Error: HPUXNativeLvm::initialize() -devices is a required argument");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: create_vg
# Objective:       create and return a Volume Group object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vg {
    my ($self, %args) = @_;
    my ($vgObj);

    $self->_clear_err();

    # add on /dev/<vgname>
    if ($args{-name} !~ /\/dev/) {
        $args{-name} = "/dev/$args{-name}";
    }

    $vgObj = San::SanClient::Lvm::HPUX::NativeLvm::Vg->_new(-parent => $self,
                                                            -conn   => $self->{conn},
                                                            %args);

    if ($vgObj->error) {
        $self->_set_err($vgObj->errmsg());
        return $vgObj;
    }

    push @{$self->{vgs}}, $vgObj;

    return $vgObj;

}

################################################################################
# Subroutine Name: destroy_vg
# Objective:       destroy a named Volume Group object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vg {
    my ($self, $name) = @_;
    my ($vgObj, @vgs);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: HPUXNativeLvm::destroy_vg() vg name is a required argument");
        return $ClientUtil::FALSE;
    }

    # add the HP pre-fix (if not specified)
    if ($name !~ /dev/) {
        $name = "/dev/$name";
    }

    # Delete the vg object and update the parent vg list.
    foreach $vgObj (@{$self->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            $vgObj->destroy();
            if ($vgObj->error) {
                $self->_set_err($vgObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vgs, $vgObj;
        }
    }

    $self->{vgs} = \@vgs;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display
# Objective:       display volume group information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;
    my ($cmd, $status, @output);

    $self->_clear_err();

    $self->sanlog(-msg => "Display Volume Group information");

    $cmd = "/usr/sbin/vgdisplay";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($status > 0) {
        $self->_set_err("Error: HPUXNativeLvm::display() - Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

}

###############################################################################
# Method:    get_disk_format
# Objective: Get list of supported disk formats
#            This is a noop for HPUX.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_format {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

###############################################################################
# Method:    get_disk_formats
# Objective: Get list of supported disk formats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_formats {
    my ($self) = @_;
    my (@format);

    $self->_clear_err();

    # This is a noop for HPUX.  Function
    # supports veritas lvm

    @format = ();

    return (wantarray()) ? @format : \@format;
}

###############################################################################
# Method:    get_format_default
# Objective: Get default type to use for disk format
#            This is a noop for HPUX.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_format_default {
    my ($self) = @_;

    $self->_clear_err();

    return "";
}

################################################################################
# Subroutine Name: get_vg
# Objective:       return a Volume Group object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vg {
    my ($self, $name) = @_;
    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: HPUXNativeLvm::get_vg() vg name is a required argument");
        return {};
    }

    foreach my $obj (@{$self->{vgs}}) {
        if ($obj->get_name() =~ /^$name$/ or
            $obj->get_name() =~ /^\/dev\/$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: HPUXNativeLvm::get_vg() Unable to find Volume Group '$name'");

    return {};

}

################################################################################
# Subroutine Name: get_vgs
# Objective:       return array of volume group objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgs {
    my ($self) = @_;
    my (@vgs);

    $self->_clear_err();

    if (defined $self->{vgs}) {
        @vgs = @{$self->{vgs}};
    } else {
        $self->_set_err("Error: HPUXNativeLvm::get_vgs() NativeLvm object has no volumes groups");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @vgs : \@vgs;

}

################################################################################
# Subroutine Name: get_vgnames
# Objective:       return array of volume group names
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgnames {
    my ($self) = @_;
    my ($cmd, $status, @buffer, @output);

    $self->_clear_err();

    $cmd = "vgdisplay | grep 'VG Name' | grep -v '/dev/vg00' | awk '{print \$3}'";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXNativeLvm::get_vgnames() Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    # Remove dev prefix to vg name
    foreach my $line (@output) {
        $line =~ s/\/dev\///;
        push @buffer, $line;
    }

    return (wantarray()) ? @buffer : \@buffer;

}

################################################################################
# Subroutine Name: get_pkginfo()
# Objective:       get detailed lvm pkg information
# Details:         See POD documentation at the end of this file
################################################################################
sub get_pkginfo {
    my ($self) = @_;
    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{pkginfo});

    return (wantarray()) ? @{$self->{pkginfo}} : $self->{pkginfo};

}

################################################################################
# Subroutine Name: _get_pkginfo
# Objective:       Get lvm package info
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self) = @_;
    my ($cmd, @output, $status, @type);

    # determine machine type (i.e PA-RISC vs. Itanium)
    if ($self->{parent}->{arch} =~ /ia64|9000/ and $self->{parent}->{os_version} =~ /B\.11\.2|B\.11\.1/i) {
        # run swlist and grab the LVM version information from
        # the HPUXBaseOS information (Itanium) 11iv1 && 11iv2
        $cmd = "/usr/sbin/swlist HPUXBaseOS | grep HPUXBaseOS.LVM | awk '{print \$2}'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        if ($status > 0) {
            # run swlist and grab the LVM version information from
            # the HPUXBase64 information (Itanium) 11iv1 && 11iv2
            # We do this check because the software package could
            # be HPUXBase64 not HPUXBaseOS
            $cmd = "/usr/sbin/swlist HPUXBase64 | grep HPUXBase64.LVM | awk '{print \$2}'";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: HPUXNativeLvm::_get_pkginfo() - Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }

    } elsif ($self->{parent}->{arch} =~ /ia64/ and $self->{parent}->{os_version} =~ /B\.11\.3/i) {
        # run swlist and grab the LVM version information from
        # the HPUXBaseOS information (Itanium) 11iv3
        $cmd = "/usr/sbin/swlist BaseLVM | grep BaseLVM.LVM | awk '{print \$2}'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        if ($status > 0) {
            $self->_set_err("Error: HPUXNativeLvm::_get_pkginfo() - Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

    } else {

        # run swlist and grab the LVM version information from
        # the HPUXBase64 information (PA-RISC)
        $cmd = "/usr/sbin/swlist HPUXBase64 | grep HPUXBase64.LVM | awk '{print \$2}'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        if ($status > 0) {
            $self->_set_err("Error: HPUXNativeLvm::_get_pkginfo() - Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }
    }

    # some fields kept for Solaris/VxVM compatability
    $self->{name}    = "LVM";
    $self->{vers}    = $output[0];
    $self->{rev}     = $output[0];
    $self->{pkginfo} = \@output;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: get_revision()
# Objective:       Get revision of Native LVM
# Details:         See POD documentation at the end of this file
################################################################################
sub get_revision {
    my ($self) = @_;
    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{rev});
    return "" if ($self->error());

    return $self->{rev};

}

################################################################################
# Subroutine Name: get_name()
# Objective:       Get lvm name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;
    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{name});
    return "" if ($self->error());

    return $self->{name};

}

################################################################################
# Subroutine Name: get_version()
# Objective:       Check version of Veritas
# Details:         See POD documentation at the end of this file
################################################################################
sub get_version {
    my ($self) = @_;
    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{vers});
    return "" if ($self->error());

    return $self->{vers};

}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;
    my ($bCnt,    $cmd,     $cnt,       $device,      $filer, $filername, $lun,
        $lun_tmp, $refresh, $status,    $total,       $vol,   %hDevs,
        @attr,    @devices, @output,    @used,        $done,  @split,     $dev,
        $line,    @array,   @tmp_array, @tmp_devices, @sanlun_output, $filers,
        $protocol);

    $total = 0;
    @{$filers} = ();

    $self->_clear_err();

    # Asssume user supplied -cnt argument
    $bCnt    = 1;
    $refresh = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Error: HPUXNativeLvm::get_free_devices() Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Error: HPUXNativeLvm::get_free_devices() Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Error: HPUXNativeLvm::get_free_devices() Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Error: HPUXNativeLvm::get_free_devices() Invalid value for argument '-vol'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Error: HPUXNativeLvm::get_free_devices() Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: HPUXNativeLvm::get_free_devices() Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: HPUXNativeLvm::get_free_devices() Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Check if user wants free devices from a specified filer, volume, or lun
    if (defined $filer or defined $vol or defined $lun or defined $protocol or scalar(@{$filers}) != 0 or $refresh == 1) {
        # Check if a device to filer map 'dfmap' has already been created or
        # refresh it if requested.
        if (not defined $self->{dfmap} or $refresh == 1) {
            $self->sanlog(-msg => "refreshing device to filer, vol and lun map");
            @output = $self->{sanlun}->lun_show();
            if ($self->{sanlun}->error) {
                $self->_set_err($self->{sanlun}->errmsg());
                return $ClientUtil::FALSE;
            }

            foreach my $line (@output) {
                # skip first line (header)
                if ($line =~ /lun-pathname/i) {
                    next;
                }

                # depending on filername length sanlun may not have display blank space before name
                my ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp);
                if ($line  =~ /^\s*(\S+?):?\s+\/vol\/(\w+)\/(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+/) {
                    # The line is in the following format
                    #sanmgr_vserv_0:  /vol/vol_1/i_5_0  /dev/rdsk/c5t0d0   lan2     iSCSI          3g (3221225472)     GOOD
                    #sanmgr_vserv_1:  /vol/vol_1/f_5_0  /dev/rdsk/c19t0d0  td0      FCP            3g (3221225472)     GOOD

                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp) = ($1, $2, $3, $4, lc($5));
                } elsif ($line =~ /\s{0}(\S+):\s+\/vol\/(\w+)\/(\S+)\s+\/dev\/rdsk\/(\w+)/) {
                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp) = ($1, $2, $3, $4);
                }

                if ($device_tmp eq "" or $device_tmp =~ /^\s*$/) {
                    $self->sanlog(-msg => "Output from sanlun is incorrect:\n$line\n");
                    next;
                }

                my $dev_tmp = basename($device_tmp);

                $hDevs{$dev_tmp} = {filer    => $filer_tmp,
                                    vol      => $vol_tmp,
                                    lun      => $lun_tmp,
                                    protocol => $protocol_tmp};

                $self->sanlog(-msg => "Device ($dev_tmp) information:\n\tFiler: $filer_tmp\n\tVolume: $vol_tmp\n" .
                                      "\tLUN: $lun_tmp\n\tProtocol: $protocol_tmp\n",
                              -msg_level => 3);

            }

            # save map for next call of function
            $self->{dfmap} = \%hDevs;
        }
    }

    # If not -cnt then assume all free disks
    # under lvm control are wanted
    unless (defined $cnt and length($cnt)) {
        $cnt  = 10000;
        $bCnt = 0;
    }

    # Get list of the root to the filer devices. That is,
    # one primary path that is "one of" the paths to a particular
    # LUN. This uses the sanlun lun show -s (hidden flag). It is
    # OK to use this since this is used for ntap_config_paths.
    @sanlun_output = $self->{sanlun}->lun_show(-primary_paths => 1);
    if ($self->{sanlun}->error) {
        $self->_set_err($self->{sanlun}->errmsg());
        return (wantarray()) ? () : [];
    }

    # The foreach below will take sanlun output and create CTD and diskXX relationships.
    # It will update the dfmap on $self to properly reflect the CTD and diskXX relationships.
    my $found_device = "";
    # Process sanlun output, match up CTD to diskXXX
    foreach my $line (@sanlun_output) {
        # Example output of what this foreach is processing:
        # sanmgr_vserv_1:/vol/vol_1/i_5_0 (LUN 0)
        # Agile Filename: /dev/rdisk/disk103
        #    3g (3221225472)     lun state: GOOD
        # Controller_CF_State: Cluster Enabled  Controller Partner: partner
        # Multipath_Provider: None
        # -------- ---------- -------------------- ------ ---------- ----------
        # host     controller                             primary    partner
        # path     path                   /dev/dsk host   controller controller
        # state    type                   filename HBA    port       port
        # -------- ---------- -------------------- ------ ---------- ----------
        # up       iscsi           /dev/dsk/c8t0d0 lan2   192.168.202.49
        # up       iscsi           /dev/dsk/c9t0d0 lan2   192.168.202.83

        # If we find an Agile device, set $found_device
        if ($line =~ /Agile Filename:\s+\/dev\/rdisk\/(\S+)/) {
            $found_device = $1;
        }

        # If we have not found a device, skip the line
        if ($found_device eq "") {
            next;
        }

        if ($line =~ /\s+\/dev\/dsk\/(\S+)\s+/) {
            # Found a CTD device filename, take the value of $found_device and
            # match it up with dfmap
            my $ctd_device = $1;
            if (defined($self->{dfmap}{$ctd_device})) {
                # We know it exists, now set the CTD_device field as $found_device
                $self->{dfmap}{$found_device} = $self->{dfmap}{$ctd_device};
                $self->{dfmap}{$found_device}{ctd_device} = $ctd_device;
                delete($self->{dfmap}{$ctd_device});
            } else {
            }
        }
    }

    # Grab a list of the devices visible from the host, this can be either MP or non-MP devices
    my @hDevices = $self->{parent}->_get_devices();
    if ($self->{parent}->error()) {
        my $error = join("\n", $self->{parent}->errmsg());
        $self->_set_err("Error: HPUXNativeLvm::get_free_devices() Unable to get list of devices from system " . $error );
        return (wantarray()) ? () : [];
    }

    # get boot device (to exclude)
    my $boot_device = $self->{parent}->_get_boot_device();
    # Get the basename of the boot device, no need for the /dev/rdsk path
    $boot_device = basename($boot_device);

    foreach my $dev (@hDevices) {
        if ($dev eq "") {
            $self->sanlog(-msg => "HPUXNativeLvm::get_free_devices() - Unable to locate " .
                                  "$dev within hDevices list from HPUXHost::_get_devices()");
            next;
        }

        # run pvdisplay to make sure this disk is free
        $cmd = "/usr/sbin/pvdisplay " . $dev->{dev_path} . "/" . $dev->{name};
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        if (grep(/Cannot display physical volume/i, @output) == 0) {
            next;
        }

        # skip the boot device
        if ($boot_device eq $dev->{name}) {
            $self->sanlog(-msg =>
                          "HPUXNativeLvm::_get_free_devices() - skipping the boot device ($boot_device)");
            next;
        }

        # Apply filters to any of the free devices found
        my $device = $dev->{name};

        if (defined $filer and length $filer) {
            $self->sanlog(-msg => "\nFilter filer: $filer\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i);
        }

        if (defined $vol and length $vol) {
            $self->sanlog(-msg => "\nFilter volume: $vol\nDevice: $device\nDevice volume: " . $self->{dfmap}{$device}{vol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i);
        }

        if (defined $lun and length $lun) {
            $self->sanlog(-msg => "\nFilter lun: $lun\nDevice: $device\nDevice lun: " . $self->{dfmap}{$device}{lun},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{lun} !~ /^$lun$/i);
        }

        if (defined $protocol and length $protocol) {
            $self->sanlog(-msg => "\nFilter protocol: $protocol\nDevice: $device\nDevice protocol: " . $self->{dfmap}{$device}{protocol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{protocol} !~ /$protocol/i);
        }

        # Filter on a filer list if provided by caller
        if (defined $filers and scalar(@{$filers}) != 0) {
            my $found = 0;
            # Look through each filer in the filers list
            foreach my $filer_f (@{$filers}) {
                $self->sanlog(-msg => "\nFilter filer(s): $filer_f\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                              -msg_level => 3);
                # If filer matches devices filer, set found to 1
                if ($self->{dfmap}{$device}{filer} =~ /^$filer_f$/i) {
                    $found = 1;
                }
            }

            # If the device is not part of the filers list caller requested skip it.
            if ($found == 0) {
                next;
            }
        }

        push @devices, $dev->{name} if (++$total <= $cnt);
    }

    # If user supplied cnt and free devices fell short
    # generate error
    if ($bCnt == 1 and scalar(@devices) < $cnt) {
        $self->_set_err("Error: HPUXNativeLvm::get_free_devices() ",
                        "Only " . scalar(@devices) . " of $cnt requested devices were found");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @devices : \@devices;

}

################################################################################
# Subroutine Name: load_vg
# Objective:       load an existing diskgroup that is not in memory
# Details:         See POD documentation at the end of this file
################################################################################
sub load_vg {
    my ($self, $name) = @_;
    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("ERROR: HPUXNativeLvm::load_vg() Invalid or missing required parameter \$name");
        return {};
    }

    # Check if vg has already been loaded
    $vgObj = $self->get_vg($name);
    if ($self->error) {
        # vg has not been load yet
        $self->_clear_err();

        # load vg
        $self->_create_vgs($name);
        return {} if ($self->error);
        $vgObj = $self->get_vg($name);
    } else {
        $self->sanlog(-msg => "HPUXNativeLvm::load_vg() Volume group '$name' was already loaded");
    }

    return $vgObj;

}
################################################################################
# Subroutine Name: unload_vg
# Objective:       Remove the vg object under LVM object control
# Details:         See POD documentation at the end of this file
################################################################################
sub unload_vg {
    my ($self, $name) = @_;
    my ($vgObj, $bFound, @vgs);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("ERROR: HPUXNativeLvm::unload_vg() Invalid or missing required parameter \$name");
        return {};
    }

    $bFound = 0;

    foreach $vgObj (@{$self->{vgs}}) {
        ($vgObj->get_name() =~ /^$name$/)
          ? $bFound = 1
          : push @vgs, $vgObj;
    }

    $self->{vgs} = \@vgs;

    unless ($bFound) {
        $self->_set_err("ERROR: HPUXNativeLvm::unload_vg() Nothing to unload. Vg name '$name' was not loaded");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: enable_controllers
# Objective:       No equivalent call exists for HPUX. 
# Details:         See POD documentation at the end of this file
################################################################################
sub enable_controllers {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: rescan
# Objective:       No equivalent call exists for HPUX.
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: restart_recovery_daemon
# Objective:       No equivalent call exists for HPUX. Leaving in for
#                  compatability with Solaris
# Details:         See POD documentation at the end of this file
################################################################################
sub restart_recovery_daemon {
    my ($self, $interval) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($cmd, $log, $loadcfg, $lvm, $parent, $self, $status, @output);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_})) {
                $self->_set_err("Error: HPUXNativeLvm::_new() Invalid value for argument '-lvm'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: HPUXNativeLvm::_new() Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?loadcfg/) {
            if (not length($loadcfg = $args{$_}) or $loadcfg !~ /no|yes|1|0|off|on/i) {
                $loadcfg = 1;
            }
            next;
        }
    }

    unless ($args{-lvm} =~ /^nativelvm$/i) {
            $self->_set_err("Error: HPUXNativeLvm::_new()",
                            "Invalid value for argument '-lvm' \(" . $args{-lvm} . "\)");
            return $self;
    }

    unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Error: HPUXNativeLvm::_new()", "Invalid value for argument '-parent'");
            return $self;
    }

    # initilize the object members
    $self->{conn}   = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{debug_level} = $parent->{debug_level};

    $self->{vgs} = [];

    # check to make sure that LVM is installed
    if (not $self->_get_pkginfo()) {
        return $self if ($self->error);
    }

    # create directory to hold vg metadata and mapfiles
    $cmd = "mkdir -p /usr/tmp/mnts";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXNativeLvm::_new()", "Command ($cmd) returned a bad status ($status)", @output);

        return $self if ($self->error);
    }

    # create veritas vg objects for existing volume groups
    $self->_create_vgs() if ($loadcfg);
    return $self if ($self->error);

    return ($self);

}

################################################################################
# Subroutine Name: _create_vgs
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vgs {
    my ($self, $vgname) = @_;
    my ($cmd, @output, $status, $vgObj, %hDeport, $vg, @all_vgs, @no_root);

    # execute the vgdisplay command to get a listing of all
    # the VGs.

    $cmd =
      (defined($vgname) and length($vgname))
      ? "/usr/sbin/vgdisplay | egrep '$vgname\$'"
      : "/usr/sbin/vgdisplay";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXNativeLvm::_create_vgs() Unable to load existing Volume Groups");
        return $ClientUtil::FALSE;
    }

    # Find each VG, plus exclude the rootvg
    @all_vgs = grep(/VG Name/, @output);
    @no_root = grep (!/\/dev\/vg00/, @all_vgs);

    # foreach volume group, create volume group object
    foreach my $line (@no_root) {
        $vg = $line;
        $vg =~ s/VG Name\s+//;
        $self->create_vg(-name     => $vg,
                         -deported => 0);

        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _get_disk_id
# Objective:       Return disk id
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_disk_id {
    my ($self, $device) = @_;

    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    $name = "/dev/rdsk/$device";

    my $path = $self->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show -v -d $name | grep 'Serial number'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXNativeLvm::_get_disk_id() Unable to get disk id for device '$device'", @output);
        return $ClientUtil::FALSE;
    }

    return $output[0];

}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

NativeLvm.pm

=head1 Authors

=begin html 

Gary Knopp, 
Tom Masterson
Michael Latif

=end html

=head1 Description

NativeLvm provides an API interface to a host logical volume manager 
on HP-UX.

=head1 Public Methods

=head2 B<create_vg(%args)>

Creates a new volume group object.

=over 4

=item Inputs

=item * -name

The name of the volume group

=item * -devices

The array reference of initialized devices to use in the volume group

=item Outputs

Returns a volume group object. If an error occurs the error flag is set, errmsg is
set to the failure reason.  

=item Examples

    # Create a volume Group object.
    $Test->log("-- Creating volume group 'myvg'");
    my $vgObj = $Lvm->create_vg(-name=>"myvg",-devices=>\@devs);
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());    
        return $TCase::FAIL;
    }

=back

=head2 B<destroy_vg($name)>

Destroy a volume group object by name

=over 4

=item Inputs

=item * $name

The name of the volume group object to destroy

=item Outputs

Returns a scalar value of 1 if successful.  If the volume group does not exist or 
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is 
passed back as the return value.


=item Examples

    $Test->log("-- Destroy disk group object 'myvg'");
    if (not $Lvm->destroy_vg("myvg")){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<display()>

Display detailed information for all volume groups

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM version information");

    $Test->log($Lvm->display());
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }


=back


=head2 B<error()>

This method allows you to determine if any errors occured in the
previous method call.  Note that the error flag is reset at the
beginning of each method, so it should be checked after B<every>
major method call.  If you do not check for errors, your calling
program may fail at some point and you will not know why.

=over 4 

=item Inputs

None

=item Outputs

Returns 1 if an error occured in the previous call, 0 otherwise.

=item Examples

 # Determine if error occurred...
 if ($obj->error) {
    print "Error encountered in previous method call\n";
 }

=back

=head2 B<errmsg()>

This method allows you to get the text of the error message if an error
occurred on the previous method call.

=over 4

=item Inputs

None

=item Outputs

Returns the text of the last error message set.  If no error message is
set, an empty string will be returned.

=item Examples

 # Check for errors.
 if ($obj->error) {
    $errmsg = $obj->errmsg;
    print "An error occurred: $errmsg\n";
 }

=back

=head2 B<get_free_devices(I<%args>)>

Returns a list of devices that are not being used by other 
LVM objects.  The list of devices can be selected by any combination 
of filer, volume, and lun.

=over 4

=item Inputs

=item * I<-cnt> 

The number of free devices to retrieve.  If the number of free devices found is less 
than the count than the error flag is set.

=item * I<-filer>

Retrieve host devices belonging to a specific filer. 

=item * I<-vol>    (ex: vol2)

Retrieve host devices belonging to a specific filer volume.  Only supply the filer 
volume name and not the full volume path.

=item * I<-lun> (ex: mylun)

Retrieve the host device mapped to a specific lun.  Only supply the lun name and not 
the full path

=item * I<-refresh>    (default: 0)

Forces the host device to filers maps to be refeshed.  This is automatically done the 
first time the method is called.  It needs to be called whenever a test program 
dynamically creates and discovers new luns in a test method. This will cause the map 
to be updated and correctly reflect device metadata on the host.


=item Outputs

Returns an array or array reference of free devices.  If an error occured the error flag 
is set and an errmsg is set to the failure reason.       


=item Examples

    $Test->log("Get 3 free devices from filer 'myfiler'");
    @devs=$Lvm->get_free_devices(-cnt=>3,-filer=>'myfiler');
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());    
        return $TCase::FAIL;
    }


=back

=head2 B<get_name()>

Retrieve name of the installed logical volume manager.

=over 4

=item Inputs

none

=item Outputs

Returns an scalar value if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and "" is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM name");
    $name=$Lvm->get_name();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

    $Test->log("LVM is ".$name);

=back

=head2 B<get_pkginfo()>

Retrieve package information on the installed logical volume manager.

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM package information");
    $buf=$Lvm->get_pkginfo();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

    foreach $line (@$buf) {
        $Test->log($line);
    }

=back

=head2 B<get_revision()>

Retrieve revision of the installed logical volume manager.

=over 4

=item Inputs

none

=item Outputs

Returns an scalar value if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and "" is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM revision");
    $rev=$Lvm->get_revision();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

    $Test->log("LVM revision is ".$rev);

=back

=head2 B<get_version()>

Retrieve version of the installed logical volume manager.

=over 4

=item Inputs

none

=item Outputs

Returns an scalar value if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and "" is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM version information");
    $version=$Lvm->get_version();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

    $Test->log("LVM version is ".$version);

=back

=head2 B<get_vg($name)>

Retrieve a volume group object by name

=over 4

=item Inputs

=item * $name

The name of the volume object to retrieve

=item Outputs

Returns a volume group object if successful.  If the volume group does not exist or 
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is 
passed back as the return value.


=item Examples

    $Test->log("-- Retrieving disk group object 'vg1'");
    $vgObj=$Lvm->get_vg("vg1");
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<get_vgs()>

Retrieve a list of all volume groups

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of volume groups if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and an empty array is returned.        


=item Examples

    $Test->log("-- Retrieving Volume Group objects");
    @vgObjs=$Lvm->get_vgs();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }  
    

=back

=head2 B<initialize(%args)>

Initialize a list of devices under LVM control

=over 4

=item Inputs

=item * -devices 

The list of devices to initialize under LVM control  

=item Outputs

Returns a scalar value of 1 if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is passed back as the return value.       


=item Examples

    
    $Test->log("-- Initialize devices under lvm control");
    $Lvm->initialize(-devices=>\@devs);
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());    
        return $TCase::FAIL;
    }


=back

=head2 B<rescan()>

Rescan the disks on the host.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Rescan devices");

    $Lvm->rescan();
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }


=back
