# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the Vxvm Volume Groups on Solaris
## @author knopp@netapp.com latif@netapp.com
## @status private
## @pod ../../Vg.pod

package San::SanClient::Lvm::HPUX::NativeLvm::Vg;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

require San::SanClient::Lvm::HPUX::NativeLvm::Vg::Vol;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($cmd, $conn, $devices, $disk_cnt, $parent, $log, $name, $self, $status);
    my ($deported, @output, %hDevice);

    $self = {};

    bless $self, $class;

    $self->_clear_err();

    $deported = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: Vg::new()", "Invalid value for argument '-devices'");
                return $self;
            }
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: Vg::new()", "Missing value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: Vg::new()", "Missing value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Vg::new()", "Missing value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?deported/) {
            if (not length($deported = $args{$_})) {
                $self->_set_err("Error: Vg::new()", "Missing value for argument '-deported'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -lvm argument
    unless (defined $parent) {
        $self->_set_err("Error: Vg::new()", "Missing required argument '-parent'");
        return $self;
    }

    # check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: Vg::new()", "Missing required argument '-name'");
        return $self;
    }

    # check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: Vg::new()", "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: Vg::new()", "Missing required argument '-conn'");
        return $self;
    }

    $self->{log} = $log if (defined $log);

    # initilize the object members
    $self->{conn}     = $conn;
    $self->{name}     = $name;
    $self->{deported} = $deported;
    $self->{disk_cnt} = 0;
    $self->{parent}   = $parent;

    # create the Volume Group if devices were supplied
    if (defined $devices and ref($devices) eq "ARRAY") {
        $self->_create($name, $devices);
        return $self if ($self->error());
    } else {
        # vg must already exist. set device list
        # for all existing and deported dgs.
        $self->sanlog(-msg => "loading existing volume group '$name'");
        $self->{devices} = $self->_get_devices();
        return $self if ($self->error);

        $self->_create_vols();
        return $self if ($self->error);
    }

    return $self;

}

################################################################################
# Subroutine Name: _create_vols
# Objective:       create vol object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vols {
    my ($self, %args) = @_;
    my ($cmd, $import, $name, @output, $status, $vgObj, $volObj, %hDeport);
    my (@lv_output, $size);

    # Identify volumes that are being imported so
    # that their automatically mounted with their
    # last filesystem mount point.
    $import =
      (exists $args{-_import})
      ? $args{-_import}
      : 0;

    $name = $self->{name};

    # get the list of volume groups
    $cmd = "/usr/sbin/vgdisplay -v $name | grep 'LV Name'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 1) {
        $self->_set_err("Error: Vg::_create_vols()",
                        "Unable to display existing volume groups",
                        "status=$status", @output);
        return $ClientUtil::FALSE;
    }

    # load logical volume information
    foreach my $lv (@output) {
        $lv =~ s/\s+LV Name\s+//;

        $cmd = "/usr/sbin/lvdisplay $lv";
        ($status, @lv_output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 1) {
            $self->_set_err("Error: Vg::_create_vols()",
                            " Unable to run command ($cmd)", @output);
            return $ClientUtil::FALSE;
        }
        $size = (grep(/LV Size \(Mbytes\)/, @lv_output))[0];
        $size =~ s/LV Size \(Mbytes\)\s+//;

        $volObj = $self->create_vol(-name    => $lv,
                                    -size    => $size,
                                    -exist   => 1,
                                    -_import => $import);
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: create_vol
# Objective:       create and return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vol {
    my ($self, %args) = @_;
    my ($volObj);

    $volObj = San::SanClient::Lvm::HPUX::NativeLvm::Vg::Vol->_new(-parent => $self,
                                                                  -conn   => $self->{conn},
                                                                  %args);

    if ($volObj->error) {
        $self->_set_err($volObj->errmsg());
        return $volObj;
    }

    push @{$self->{vols}}, $volObj;

    return $volObj;

}

################################################################################
# Subroutine Name: destroy_vol
# Objective:       destroy a named vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vol {
    my ($self, $name) = @_;
    my ($volObj, @vols);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Vg::destroy_vol() volname is a required argument");
        return $ClientUtil::FALSE;
    }

    # Delete the of vg object and update the parent vg list.
    foreach $volObj (@{$self->{vols}}) {
        if ($volObj->get_name() =~ /^$name$/) {
            $volObj->destroy();
            if ($volObj->error) {
                $self->_set_err($volObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vols, $volObj;
        }
    }

    $self->{vols} = \@vols;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _get_devices
# Objective:       get list of devices used by volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices {
    my ($self) = @_;
    my ($cmd, $name, $deported, $status, @output, %hDevice, $cnt);

    $name     = $self->get_name();
    $deported = $self->{deported};

    $cmd = "/usr/sbin/vgdisplay -v $name | grep \"PV Name\"";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($status > 0) {
        $self->_set_err("Error: Vg::_get_devices() Unable to get devices for volgroup \'$name\'", @output);
        return {};
    }

    $cnt = 0;
    foreach my $line (@output) {
        $line =~ s/\s+PV Name\s+//;

        # If this is not a alternate link, increment the disk count
        if ($line !~ /Alternate Link/) {
            $hDevice{$line} = "NA";
            $cnt++;
        }
    }

    $self->{disk_cnt} = $cnt;

    return \%hDevice;

}

################################################################################
# Subroutine Name: _create()
# Objective:       create a volume group. For good document on VGs, please see
#                  http://docs.hp.com/en/lvm-v2/L2_whitepaper_8.pdf
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self, $name, $devices) = @_;
    my ($cmd, $disk_cnt, $disk_name, @output, $status, %hDevice, $len);
    my ($max_vg_size, $extent_size, $highest, $highest_num, $entry, $major, $minor);

    # get the revision of volume group to be used and the extent size to use
    my $vm_hpux_revision = $self->{parent}->{parent}->{vm_hpux_revision}; 
    my $vm_hpux_extent_size = $self->{parent}->{parent}->{vm_hpux_extent_size};
    my $mp_type = $self->{parent}->{parent}->{mp_type};

    # major number is different for V1.0 and V2.x volumes groups
    if ($vm_hpux_revision =~ /2\./) {
        $major = 128;
    } else {
        $major = 64;
    }

    $disk_cnt = 0;

    my @hDevices = $self->{parent}->{parent}->_get_devices();
    if ($self->{parent}->{parent}->error()) {
        my $error = join("\n", $self->{parent}->{parent}->errmsg());
        $self->_set_err("Error: Vg::_create Unable to get list of devices from system " . $error );
        return $ClientUtil::FALSE;
    }

    # check to see if a VG exists with this name
    $cmd = "/usr/sbin/vgdisplay $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    # if there is no VG and a directory remains, delete it
    if ($status != 0) {
        if ($name =~ /^$/ || $name =~ /^\/$/) {
            $self->_set_err("Error: Vg::_create Invalid name ($name)");
            return $ClientUtil::FALSE;
        }
        # force remove the directory and it's contents.
        $cmd = "rm -rf $name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    }

    # creation of VG directory is not needed with 2.1 volume groups
    if ($vm_hpux_revision !~ /2\.1/) {
        # create the LVM VG group directory
        $cmd = "mkdir $name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # determine if mkdir of LVM directory worked.
        if ($status > 0) {
            $self->_set_err("Error: Vg::_create Unable to create vol group directory \'$name\'", @output);
            return $ClientUtil::FALSE;
        }
    
        # determine the next minor number to use for the mknod call
        # in the volume group directory
        $cmd = "ls -l /dev/*/group";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # determine if ls to determine minor number worked
        if (grep(/not found/i, @output)) {
            # No vg groups are created on this client, start with the first minor number
            $highest = "000000";
        } elsif ($status > 0) {
            $self->_set_err("Error: Vg::_create Unable to query minor number", @output);
            return $ClientUtil::FALSE;
        }
    
        foreach $entry (@output) {
            $minor = (split(/\s+/, $entry))[5];
            $minor =~ s/0x//;
            if (not defined($highest) or $minor > $highest) {
                $highest = $minor;
            }
        }
    
        # get the highest minor number and increment.
        # v2.x is high order 12 bits of minor, v1.0 is high order 8 bits of minor
        if ($vm_hpux_revision =~ /2\./) {
            $highest_num = substr($highest, 1, 2);
        } else {
            $highest_num = substr($highest, 0, 2);
        }
        $highest_num++;
    
        # get the size of the string representing the new value
        $len = length($highest_num);
    
        # next, run the mknod command with the newly aquired minor device number
        # (plus one).
        if ($len > 1) {
            if ($vm_hpux_revision =~ /2\./) {
                $cmd = "/usr/sbin/mknod $name/group c $major 0x0" . $highest_num . "000";
            } else {
                $cmd = "/usr/sbin/mknod $name/group c $major 0x" . $highest_num . "0000";
            }
        } else {
            if ($vm_hpux_revision =~ /2\./) {
                $cmd = "/usr/sbin/mknod $name/group c $major 0x00" . $highest_num . "000";
            } else {
                $cmd = "/usr/sbin/mknod $name/group c $major 0x0" . $highest_num . "0000";
            }
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # check to make sure the mknod worked
        if ($status > 0) {
            $self->_set_err("Error: Vg::_create Unable to run mknod", @output);
            return $ClientUtil::FALSE;
        }
    }

    # iterate through the devices
    foreach my $device (@$devices) {
        my $hdev = "";
        foreach my $cdev (@hDevices) {
            if ($cdev->{name} eq $device) {
                $hdev = $cdev;
                last;
            }
        }

        if ($hdev eq "") {
            $self->sanlog(-msg => "Vg::_createget_free_devices() - Unable to locate " .
                                  "$device within hDevices list from HPUXHost::_get_devices()");
            next;
        }

        $disk_cnt++;

        # create the physical volume
        # if mp_type is mass, use the new style DSF
        if ($mp_type eq "mass") {
            $cmd = "/usr/sbin/pvcreate -f " . $hdev->{path} . "/" . $hdev->{name_disk};
        } else {
            $cmd = "/usr/sbin/pvcreate -f " . $hdev->{path} . "/" . $hdev->{name};
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Vg::_create Unable to pvcreate device \'" . $hdev->{path} . "/" . $hdev->{name} . "\'", @output);
            return $ClientUtil::FALSE;
        }

        # determine the extext size to use for volume group creation (default will be 
        # be 32 (MB). Extext size needs to be a mutiple a power of 2.
        if ($vm_hpux_extent_size eq "") {
            $extent_size = 32;
        } else {
            $extent_size = $vm_hpux_extent_size;
        }

        # max number of extents =  32M (in G). 
        # Reason we divide by 4 is to limit the overhead of space
        # needed on each disk for meta data. If we use MAX size it
        # can require up to 500M and with smaller disks this will not
        # be efficient and could fail tests. With 1/4 size we get up
        # to 256TB and overhead is 100M
        $max_vg_size = int((32768 * $extent_size) / 4);

        if ($disk_cnt == 1) {
            # Next we need to create a volume group (local to this test)
            # and put all the devices into that Volume Group. First init the new
            # VG with the first device. Then for each subsequent device, add it
            # to the new VG. Create the volume group with the maximum number
            # of physical volume contributors (255). Also, turn off autobackup
            # so that we dont get a ton of stale data in the /etc/lvmconf
            # directory. 

            # Now, handle v1.0, v2.0 (11iv3), and v2.1 (11iv3) creation of VG
            if ($vm_hpux_revision =~ /2\.1/) {
                $cmd = "/usr/sbin/vgcreate -V 2.1 -A n -S ${max_vg_size}g -s $extent_size $name " . $hdev->{dev_path} . "/" . $hdev->{name};
            } elsif ($vm_hpux_revision =~ /2\.0/) {
                $cmd = "/usr/sbin/vgcreate -V 2.0 -A n -S ${max_vg_size}g -s $extent_size $name " . $hdev->{dev_path} . "/" . $hdev->{name};
            } else {
                $cmd = "/usr/sbin/vgcreate -A n -p 255 -s $extent_size $name " . $hdev->{dev_path} . "/" . $hdev->{name};
            }
        } else {
            $cmd = "/usr/sbin/vgextend $name " . $hdev->{dev_path} . "/" . $hdev->{name};
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # determine if the volume group command worked
        if ($status > 0) {
            $self->_set_err("Error: Vg::_create Unable to create/extend volume group \'$name\'", @output);
            return $ClientUtil::FALSE;
        }

        # the disk number if not needed for HPUX.
        $hDevice{$device} = "NA";
    }

    # next, we need to re-run ntap_config paths to add the new paths
    # for the device that was added.
    $self->{parent}->{parent}->config_paths();
    if ($self->{parent}->{parent}->error) {
        my @error = $self->{parent}->{parent}->errmsg();
        $self->_set_err("Error: Vg::_create Unable to run ntap_config_paths on volume group $name", @output, @error);
        return $ClientUtil::FALSE;
    }

    $self->{devices}  = \%hDevice;
    $self->{disk_cnt} = $disk_cnt;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: add_device()
# Objective:       add a device to the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub add_device {
    my ($self, $device) = @_;
    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    # Get name of volume group.
    $name = $self->{name};

    # Increment device counter
    $self->{disk_cnt}++;

    # create the physical volume
    $cmd = "/usr/sbin/pvcreate -f /dev/rdsk/$device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # determine if the pvcreate worked.
    if ($status > 0) {
        $self->_set_err("Error: Vg::add_device Unable to pvcreate device $device", @output);
        return $ClientUtil::FALSE;
    }

    # extend the volume group
    $cmd = "/usr/sbin/vgextend $name /dev/dsk/$device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check to make sure the extend worked.
    if ($status > 0) {
        $self->_set_err("Error: Vg::add_device Unable to extend the volume group $name", @output);
        return $ClientUtil::FALSE;
    }

    # next, we need to re-run ntap_config paths to add the new paths
    # for the device that was added.
    $self->{parent}->{parent}->config_paths();
    if ($self->{parent}->{parent}->error) {
        my @error = $self->{parent}->{parent}->errmsg();
        $self->_set_err("Error: Vg::add_device Unable to run ntap_config_paths on volume group $name", @output, @error);
        return $ClientUtil::FALSE;
    }

    $self->{devices}->{$device} = "NA";

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: remove_device()
# Objective:       remove a device from the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub remove_device {
    my ($self, %args) = @_;
    my ($cmd,        $device, $force, $disk_cnt, $disk_name, $name, @output,
        @vols_mount, $status, $line,  @devices,  $path,      $found);

    $self->_clear_err();

    $force = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Error: Vg::remove_device()", "Invalid value for argument '-device'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Vg::remove_device()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            $force = "-l";
            next;
        }
    }

    # get name of the volume group
    $name = $self->{name};

    # verify the device is being used in volume group
    unless (exists $self->{devices}->{$device}) {
        $self->_set_err("Error: Vg::remove_device()", "Device '$device' doesn't exist in Volume Group");
        return $ClientUtil::FALSE;
    }

    $disk_name = $self->{devices}->{$device};

    # if no force option is specified umount all volumes so deport
    # can occur without error.  Make note of which vols were
    # mounted so they can be remounted after removing device
    if (not defined $force) {
        my @volObjs = $self->get_vols();
        return $ClientUtil::FALSE if ($self->error);

        foreach my $volObj (@volObjs) {
            if ($volObj->is_fs()) {
                if ($volObj->is_mounted()) {
                    $self->sanlog(-msg => "unmounting filesystem on volume " . $volObj->get_name());
                    $volObj->unmount();
                    if ($volObj->error) {
                        $self->_set_err($volObj->errmsg());
                        return $ClientUtil::FALSE;
                    }
                    push @vols_mount, $volObj;
                }
            }
        }
    }

    # formulate the vgdisplay command
    $cmd = "/usr/sbin/vgdisplay -v $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check status from the command
    if ($status > 0) {
        $self->_set_err("Error: Vg::remove_device Unable to get infomation from  volume group $name", @output);
        return $ClientUtil::FALSE;
    }

    # we need to remove the Alternate paths for this device,
    # get the alternate paths before we remove the original device.
    $found = 0;
    foreach $line (@output) {
        if ($line =~ /PV Name.*$device/) {
            $found = 1;
            next;
        }

        # If we found the device, now look for Alternate paths.
        if ($found == 1) {
            if ($line =~ /PV Name/) {

                # If we hit a PV Name that is not an alternate, we are
                # done with that LUN
                if ($line !~ /Alternate/) {
                    $found = 0;
                } else {

                    # strip information to get down to just the device
                    $line =~ s/.*PV Name.*\/dev\/dsk\///;
                    $line =~ s/\s+Alternate Link.*//;
                    push(@devices, "/dev/dsk/$line");
                }
            }
        }
    }

    # add the original device into the list
    push(@devices, "/dev/dsk/$device");

    # now remove all the paths to the LUN
    foreach $path (@devices) {
        # formulate the vgreduce command
        $cmd = "/usr/sbin/vgreduce $name $path";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # check status from the command
        if ($status > 0) {
            $self->_set_err("Error: Vg::remove_device Unable to remove $path from volume group $name", @output);
            return $ClientUtil::FALSE;
        }
    }

    # formulate the pvremove command
    $cmd = "/usr/sbin/pvremove /dev/rdsk/$device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check status from the command
    if ($status > 0) {
        $self->_set_err("Error: Vg::remove_device Unable to pvremove $device $name", @output);
        return $ClientUtil::FALSE;
    }

    # remove device from hash
    delete($self->{devices}->{$device});

    # decrement device counter;
    $self->{disk_cnt}--;

    # Restart and mount volumes that were previously mounted.
    foreach my $vol (@vols_mount) {
        $vol->_vol_start();
        if ($vol->error) {
            $self->_set_err($vol->errmsg());
            return $ClientUtil::FALSE;
        }

        $vol->mount();
        if ($vol->error) {
            $self->_set_err($vol->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: rename()
# Objective:       NOOP for HPUX Native LVM
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, %args) = @_;
    return $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;
    my ($cmd, $name, @output, $status, @vgs, @pvs, @lvs, $lv, 
        $pv, $num_pvs, $rdsk, $cnt);

    $self->_clear_err();

    $name = "/dev/" . $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vg::destroy() Volume Group name is a required parameter");
        return $ClientUtil::FALSE;
    }

    # Get the logical volumes in the VG
    $cmd = "/usr/sbin/vgdisplay -v $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::destroy() Unable to get logical volumes $name", @output);
        return $ClientUtil::FALSE;
    }

    @lvs = grep(/LV Name/, @output);

    # remove the logical volumes in the volume group
    foreach $lv (@lvs) {
        $lv =~ s/\s+LV Name\s+//;

        # get the lv object
        my $lv_obj = $self->get_vol($lv);

        $cmd = "/usr/sbin/lvremove -f $lv";
        ($status, @lvs) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Vg::destroy() Unable to remove logical volumes $lv", @output);
            return $ClientUtil::FALSE;
        }

        # remove the mount point file
        $lv_obj->_set_metadata("");
    }

    # Get the physical volumes in the VG
    $cmd = "/usr/sbin/vgdisplay -v $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::destroy() Unable to get physical volumes $name", @output);
        return $ClientUtil::FALSE;
    }

    @pvs = grep(/PV Name/, @output);

    # get the number of pvs
    $num_pvs = scalar(@pvs);

    # remove the physical volumes in the volume group (all but one - this
    # last one will be remove when removing the VG.
    if ($num_pvs > 1) {
        $cnt = 0;
        while ($cnt < ($num_pvs - 1)) {
            $pv = $pvs[$cnt];
            $pv =~ s/\s+PV Name\s+//;
            $pv =~ s/Alternate Link//;

            # first reduce the Volume group by this physical volume
            $cmd = "/usr/sbin/vgreduce $name $pv";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 1) {
                $self->_set_err("Error: Vg::destroy() Unable to reduce the volume group $name - $pv", @output);
                return $ClientUtil::FALSE;
            }
            $cnt++;
        }
    }

    # now, remove the Volume group and the last physical volume
    $cmd = "/usr/sbin/vgremove $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        # Unable to remove Volume group, deport it
        $self->deport();
        if ($self->error()) {
            $self->_set_err("Error: Vg::destroy() Unable to remove volume group $name", @output);
            return $ClientUtil::FALSE;
        }
    }

    # cleanup the physical volumes and special files
    foreach $pv (@pvs) {
        $pv =~ s/\s+PV Name\s+//;

        # only need to remove the primary links, if this is
        # an alternate link, skip.
        # Also check to see if the VG has been deported, if so
        # skip through the pv's
        if ($pv =~ /Alternate Link/ or $self->is_deported()) {
            next;
        }

        # get the rdsk entry of the physical volume
        $rdsk = $pv;
        $rdsk =~ s/dsk/rdsk/;

        # If /dev/disk, replace pv with rdisk, need char device not phys device
        # This is needed to support MASS stack
        if ($rdsk =~ /\/dev\/disk\//) {
            $rdsk =~ s/disk/rdisk/;
        }

        # remove the physical volume
        $cmd = "/usr/sbin/pvremove $rdsk";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Vg::destroy() Unable to remove physical volume $pv", @output);
            return $ClientUtil::FALSE;
        }
    }

    # remove special file (on group file - part of VG)
    $cmd = "/usr/sbin/rmsf $name/group";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::destroy() Unable to remove special file $name/group", @output);
        return $ClientUtil::FALSE;
    }

    # now, remove the Volume group directory
    $cmd = "rmdir $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::destroy() Unable to remove directory $cmd", @output);
        return $ClientUtil::FALSE;
    }

    # now, remove the entries in the /etc/lvmconf directory that
    # correspond to this VG
    my $vg_name = $name;
    $vg_name =~ s/\/dev\///;
    $cmd = "rm -f /etc/lvmconf/$vg_name.conf /etc/lvmconf/$vg_name.conf.old";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::destroy() Unable to remove lvmconf files", @output);
        return $ClientUtil::FALSE;
    }

    # Remove the vgobj from the parent lvm object
    # Strip the "/dev/" prefix off of $name, to match
    # return value from get_name()
    $name =~ s/^\/dev\///;
    foreach my $vgObj (@{$self->{parent}->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vgs, $vgObj;
        }
    }

    $self->{parent}->{vgs} = \@vgs;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;
    my (@verbose, $cmd, $name, @output, $status, $header,
        $summary, @summary, $verbose);

    $self->_clear_err();

    $verbose = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Error: Vg::display()", "Invalid value for argument '-verbose'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    $name = $self->{name};

    unless (defined $name) {
        $self->_set_err("Error: Vg::display() Volume Group name is a required parameter");
        return (wantarray()) ? () : [];
    }

    push @summary, "VG Name:  " . $self->get_name();
    push @summary, "Devices:  " . $self->get_device_cnt();

    $header = sprintf("Vol Summary:\n"
                        . "%-12.12s %-7.7s %-5.5s %-12.12s %-32.32s\n"
                        . "%-12.12s %-7.7s %-5.5s %-12.12s %-32.32s",
                      "NAME",   "LAYOUT", "TYPE",  "SIZE(mb)", "MOUNT_PT",
                      "=" x 12, "=" x 7,  "=" x 5, "=" x 12,   "=" x 32);

    push @summary, $header;

    foreach my $volObj (@{$self->{vols}}) {
        $summary = sprintf("%-12.12s %-7.7s %-5.5s %-12.12s %-s",
                           $volObj->get_name(), $volObj->get_layout(), $volObj->get_fs(),
                           $volObj->get_size(), $volObj->get_mount());

        push @summary, $summary;
    }

    if (defined $verbose and $verbose =~ /1|on|y|yes|t|true/) {
        $cmd = "/usr/sbin/vgdisplay -v $name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Vg::display() Unable to display info for Volume Group \'$name\'", @output);
            return (wantarray()) ? () : [];
        }

        @verbose = (@summary, "\nVgroup Detail:", @output);

        push @verbose, "\n\n";
        return (wantarray()) ? @verbose : \@verbose;
    } else {
        push @summary, "\n\n";
        return (wantarray()) ? @summary : \@summary;
    }

}

################################################################################
# Subroutine Name: get_device_cnt
# Objective:       Return number of devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device_cnt {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{disk_cnt})) {
        return $self->{disk_cnt};
    } else {
        $self->_set_err("Error: Vg::get_device_cnt() Unable to display disk count for Volume Group");
        return 0;
    }

}

################################################################################
# Subroutine Name: get_devices
# Objective:       Return devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    @output = (keys %{$self->{devices}});

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vg object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    my ($tmp);

    $self->_clear_err();

    # strip the /dev/ prefix since snapdrive currently cannot handle this
    if (defined($self->{name})) {
        $tmp = $self->{name};
        $tmp =~ s/\/dev\///;
        return $tmp;
    } else {
        $self->_set_err("Error: Vg::get_name() Vg object is missing name");
        return "";
    }

}

################################################################################
# Subroutine Name: get_vol
# Objective:       return a Volume Group object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol {
    my ($self, $name) = @_;
    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Vg::get_vol() volume name is a required argument");
        return {};
    }

    foreach my $obj (@{$self->{vols}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: Vg::get_vol() Unable to find volume '$name'");

    return {};

}

################################################################################
# Subroutine Name: get_vols
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vols {
    my ($self) = @_;
    my (@volObjs);

    $self->_clear_err();
    if (defined $self->{vols}) {
        @volObjs = @{$self->{vols}};
    } else {
        $self->_set_err("Error: Vg::get_vols() Vg object has no volumes");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @volObjs : \@volObjs;

}

################################################################################
# Subroutine Name: get_vol_cnt
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol_cnt {
    my ($self) = @_;

    $self->_clear_err();
    if (defined $self->{vols}) {
        return scalar(@{$self->{vols}});
    } else {
        $self->_set_err("Error: Vg::get_vol_cnt() Vg object has no vol count");
        return 0;
    }

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size, $pe_size, $total_pe);

    $self->_clear_err();

    $size = 0;

    $name = $self->get_name();

    $cmd = "/usr/sbin/vgdisplay -v $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::get_size() unable to execute \'$cmd\'", @output);
        return $size;
    }

    # get the Physical Extent size
    $pe_size = (grep(/PE Size \(Mbytes\)/, @output))[0];
    $pe_size =~ s/PE Size \(Mbytes\)\s+//;

    # get the total number of Physical Extents
    $total_pe = (grep(/Total PE/, @output))[0];
    $total_pe =~ s/Total PE\s+//;

    # size is just the multiplication of these.
    $size = $pe_size * $total_pe;

    return $size;

}

################################################################################
# Subroutine Name: get_freespace
# Objective:       Return available free space
# Details:         See POD documentation at the end of this file
################################################################################
sub get_freespace {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size, $pe_size, $free_pe);

    $self->_clear_err();

    $size = 0;

    $name = $self->get_name();

    $cmd = "/usr/sbin/vgdisplay -v $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::get_freespace() Unable to get free space (mb) for Volume Group \'$name\'", @output);
        return $size;
    }

    # get the Physical Extent size
    $pe_size = (grep(/PE Size \(Mbytes\)/, @output))[0];
    $pe_size =~ s/PE Size \(Mbytes\)\s+//;

    # get the total number of Free Physical Extents
    $free_pe = (grep(/Free PE/, @output))[0];
    $free_pe =~ s/Free PE\s+//;

    # size is just the multiplication of these.
    $size = $pe_size * $free_pe;

    return $size;

}

################################################################################
# Subroutine Name: is_deported
# Objective:       Determine if volgroup has been deported
# Details:         See POD documentation at the end of this file
################################################################################
sub is_deported {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{deported})) {
        return $self->{deported};
    } else {
        return 0;
    }

}

################################################################################
# Subroutine Name: deport()
# Objective:       deport Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub deport {
    my ($self) = @_;
    my ($cmd, $name, @output, $status, @devices);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined($name)) {
        $self->_set_err("Error: Vg::deport() Volume Group name is a required parameter");
        return $ClientUtil::FALSE;
    }

    # first run vgchange to allow the volume group to be exported.
    $cmd = "/usr/sbin/vgchange -a n $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::deport() Unable to change state of the volume group \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    # Then, actually export the volume group
    my $tmp = $name;
    $tmp =~ s/\/dev\///;
    $cmd = "/usr/sbin/vgexport -m /usr/tmp/mnts/$tmp.mapfile $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::deport() Unable to export volume group \'$name\'", @output);
        return $ClientUtil::FALSE;

    }

    # Remove volume objects from vg cache
    $self->{vols} = ();

    $self->{deported} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: import()
# Objective:       Import Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub import {
    my ($self) = @_;
    
    # we are stupidly overriding the import method
    if (!ref($self)) {
        return;
    }

    my ($cmd,     $name,        @output,  $status,   $entry, $device, $minor,
        $highest, $highest_num, $dev_list, $len,   @devices);

    $self->_clear_err();

    $name = $self->get_name();

    @devices = $self->get_devices();

    unless (defined $name) {
        $self->_set_err("Error: Vg::import() Volume Group name is a required parameter");
        return $ClientUtil::FALSE;
    }

    # check to see if a VG exists with this name
    $cmd = "/usr/sbin/vgdisplay $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # if there is no VG and a directory remains, delete it
    if ($status != 0) {
        # force remove the directory and it's contents.
        $cmd = "rm -rf $name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    }

    # create the volume group directory again for import.
    $cmd = "mkdir $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    # check to make sure the mkdir worked
    if ($status > 0) {
        $self->_set_err("Error: Vg::import() Unable to run mkdir \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    # next, determine the next minor number to use for the mknod call
    # in the volume group directory
    $cmd = "ls -l /dev/*/group";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # determine if ls to determine minor number worked
    if (grep(/not found/i, @output)) {

        # No vg groups are created on this client, start with the first minor number
        $highest = "000000";
    } elsif ($status > 0) {
        $self->_set_err("Error: Vg::import() Unable to query minor number", @output);
        return $ClientUtil::FALSE;
    }

    foreach $entry (@output) {
        $minor = (split(/\s+/, $entry))[5];
        $minor =~ s/0x//;
        if (not defined($highest) or $minor > $highest) {
            $highest = $minor;
        }
    }

    # get the highest minor number and increment.
    $highest_num = substr($highest, 1, 1);
    $highest_num++;

    # get the size of the string representing the new value
    $len = length($highest_num);

    # next, run the mknod command with the newly aquired minor device number
    # (plus one).
    if ($len > 1) {
        $cmd = "/usr/sbin/mknod $name/group c 64 0x0" . $highest_num . "000";
    } else {
        $cmd = "/usr/sbin/mknod $name/group c 64 0x0" . $highest_num . "0000";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check to make sure the mknod worked
    if ($status > 0) {
        $self->_set_err("Error: Vg::import() Unable to run mknod", @output);
        return $ClientUtil::FALSE;
    }

    # get the list of devices ready for the import format
    foreach $device (@devices) {
        if ($device !~ /\/dev\//) {
            $dev_list .= " /dev/dsk/$device";
        } else {
            $dev_list .= " $device";
        }
    }

    # execute the import command with the list of saved devices
    my $tmp = $name;
    $tmp =~ s/\/dev\///;
    $self->sanlog(-msg => "importing volume group '" . $tmp . "'");
    $cmd = "/usr/sbin/vgimport -m /usr/tmp/mnts/$tmp.mapfile -v $name $dev_list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::import() Unable to import Volume Group \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    # Activate the Volume Group
    $self->sanlog(-msg => "activating volume group '" . $tmp . "'");
    $cmd = "/usr/sbin/vgchange -a y $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::import() Unable to activate Volume Group \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    $self->{deported} = 0;

    # requery device list
    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    $self->_create_vols(-_import => 1);
    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: refresh
# Objective:       refresh content of volume group.
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh {
    my ($self) = @_;
    my ($name, @devs);

    # Remove existing vol cache
    $self->{vols} = undef;
    $name = $self->get_name();

    $self->sanlog(-msg => "refreshing volume group $name");

    @devs = $self->_get_devices();
    if ($self->error) {
        $self->{deported} = 1;
        return $ClientUtil::TRUE;
    } else {
        $self->{devices} = \@devs;
        $self->_create_vols();
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;

}
################################################################################
# Subroutine Name: _create_mapfile
# Objective:       create mapfile for volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_mapfile {
    my ($self) = @_;
    my ($cmd, $name, $status, @output);

    $name = $self->{name};

    # Remove dev prefix
    my $tmp = $name;
    $tmp =~ s/\/dev\///;

    # Create map entry w/o actually exporting volume group.
    # Use the (-p)review option.
    $cmd = "/usr/sbin/vgexport -p -m /usr/tmp/mnts/$tmp.mapfile $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::_create_mapfile() Unable to create mapfile for volume group \'$tmp\'", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Vg.pm

=head1 Authors

=begin html 

Gary Knopp<br>
Tom Masterson
Michael Latif

=end html

=head1 Description

Vg provides an API interface to LVM volume groups through the use of the Telnet library.  

=head1 Public Methods

=head2 B<add_device($name)>

Add a new device to the volume group

=over 4

=item Inputs

=item * $name

The name of the device.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Add a new device to volume group");
    $vgObj=$Lvm->get_vg("dg1");
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

    $vgObj->add_device("c3t0d8");
     if ($vgObj->error){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<create_vol(%args)>

Creates a new volume group object.

=over 4

=item Inputs

=item * -name

The name of the volume

=item * -size (note: specify units 500k, 1m, 2g)

The size of the volume (note: specify units 500k, 1m, 2g)

=item * I<-layout> (default: concat)

The layout of the volume on the devices

=item Outputs

Returns a volume object. If an error occurs the error flag is set, errmsg is
set to the failure reason.  

=item Examples

    # Create a volume Group object.
    $Test->log("-- Creating volume group 'mydg'");
    $volObj = $vgObj->create_vol(-name=>"vol_A",-size=>'100m',-layout=>'stripe');
    if ($vgObj->error){
        $Test->log($vgObj->errmsg());    
        return $TCase::FAIL;
    }

=back

=head2 B<deport()>

Import a volume group 

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Deport a volume group");

    $vgObj->deport();
    if ($vgObj->error()){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<destroy()>

Destroy the volume group 

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Destroy volume group");

    $vgObj->destroy();
    if ($vgObj->error()){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<destroy_vol($name)>

Destroy a volume object by name

=over 4

=item Inputs

=item * $name

The name of the volume object to destroy

=item Outputs

Returns a scalar value of 1 if successful.  If the volume group does not exist or 
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is 
passed back as the return value.


=item Examples

    $Test->log("-- Destroy volume object 'vol_A'");
    if (not $vgObj->destroy_vol("vol_A")){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<display()>

Display detailed volume group information

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving LVM version information");

    $Test->log($vgObj->display());
    if ($vgObj->error){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<error()>

This method allows you to determine if any errors occured in the
previous method call.  Note that the error flag is reset at the
beginning of each method, so it should be checked after B<every>
major method call.  If you do not check for errors, your calling
program may fail at some point and you will not know why.

=over 4 

=item Inputs

None

=item Outputs

Returns 1 if an error occured in the previous call, 0 otherwise.

=item Examples

 # Determine if error occurred...
 if ($obj->error) {
    print "Error encountered in previous method call\n";
 }

=back

=head2 B<errmsg()>

This method allows you to get the text of the error message if an error
occurred on the previous method call.

=over 4

=item Inputs

None

=item Outputs

Returns the text of the last error message set.  If no error message is
set, an empty string will be returned.

=item Examples

 # Check for errors.
 if ($obj->error) {
    $errmsg = $obj->errmsg;
    print "An error occurred: $errmsg\n";
 }

=back

=head2 B<get_devices()>

List of devices used  by volume group

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of devices if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and an empty array is returned.        


=item Examples

    $Test->log("-- Retrieving volume group devices");
    @devs=$vgObj->get_devices();
    if ($vgObj->error){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }  
    foreach $dev (@devs){
        print $dev."\n";
    }

=back

=head2 B<get_device_cnt()>

Number of devices used by volume group

=over 4

=item Inputs

none
                                       
=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is returned.        


=item Examples

    $size=$vgObj->get_size();
    if ($vgObj->error){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }  
   
    print "VG size is $size mb\n";

=back


=head2 B<get_freespace()>

Estimate of free disk space (megabytes) in volume group

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is returned.        


=item Examples

    $size=$vgObj->get_freespace();
    if ($vgObj->error){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }  
   
    print "VG available space is $size mb\n";

=back

=head2 B<get_name()>

Name of volume group

=over 4

=item Inputs

none

=item Outputs

Returns the name of the volume group.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Volume group name is '".$vgObj->get_name()."'";

=back


=head2 B<get_size()>

Estimate of volume group size in Megabytes

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is returned.        


=item Examples

    $size=$vgObj->get_size();
    if ($vgObj->error){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }  
   
    print "VG size is $size mb\n";

=back

=head2 B<get_vol($name)>

Retrieve a volume object by name

=over 4

=item Inputs

=item * $name

The name of the volume to retrieve

=item Outputs

Returns a volume object if successful.  If the volume does not exist or 
an error occured the error flag is set, errmsg is set to the failure reason, and 0 is 
passed back as the return value.


=item Examples

    $Test->log("-- Retrieving vol object 'vol_A'");
    $volObj=$vgObj->get_vol("vol_A");
    if ($Lvm->error){
        $Test->log($Lvm->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<get_vols()>

Retrieve a list of all volumes in  volume group

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of volumes if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and an empty array is returned.        


=item Examples

    $Test->log("-- Retrieving volume objecs objects");
    @volObjs=$vgObj->get_vols();
    if ($vgObj->error){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }  
    

=back

=head2 B<get_vol_cnt()>

Number of volumes in the volume group

=over 4

=item Inputs

none
                                       
=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is returned.        


=item Examples

    $cnt=$vgObj->get_vol_cnt();
    if ($vgObj->error){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }  
   
    print $vgObj->get_name()." has is $cnt volumes\n";

=back



=head2 B<import()>

Import a volume group 

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Import a volume group");

    $vgObj->import();
    if ($vgObj->error()){
        $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<is_deported()>

Determine if a volume group is deported 

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if deported, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Examples

    if($vgObj->is_deported()){
        $vgObj->import();
        ...
        <error handler>
        ...
    }
   

=back


=head2 B<remove_device(I<%args>)>

Removes a device from the volume group.  

=over 4

=item Inputs

=item * -device

The name of the device.    

=item * I<-force> (values: 1 | on | yes )     

Remove the device without unmounting any filesystems that exist.  Without the force
option, all filesystems that are mounted will be unmounted during the device removal.
Upon success the filesystems that were previously mounted will be remounted if they 
were not effected by the device removal.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

   $vgObj->remove_device(-device=>$device,-force=>1);
   if ($vgObj->error){
       $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
   }


=back

=head2 B<rename(I<%args>)>

Rename the volume group

=over 4

=item Inputs

=item * -name

The new name of the volume group.    

=item * I<-force> (values: 1 | on | yes )     

The force option will automatically locate and unmount all filesystems that are mounted 
so that a volume group rename can occur without error.  Upon success the filesystems 
that were previously mounted will be remounted.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

   $vgObj->rename(-name=>"mydg_new",-force=>1);
   if ($vgObj->error){
       $Test->log($vgObj->errmsg());
        return $TCase::FAIL;
   }


=back
