# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the NativeLvm logical volumes
## @author knopp@netapp.com latif@netapp.com
## @status private
## @pod ../../../Vol.pod

package San::SanClient::Lvm::HPUX::NativeLvm::Vg::Vol;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use File::Basename;
use Data::Dumper;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
# Method:    clean
# Objective: clean volume of data
# Details:   See POD documentation at the end of this file.
###############################################################################
sub clean {
    my ($self) = @_;
    my ($cmd, $mount_dir, $name, $vgname, $status, @output);

    $self->_clear_err();

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();

    if (not $self->is_mounted()) {
        $self->_set_err("Error: Vol::clean() Volume is not currently mounted");
        return $ClientUtil::FALSE;
    }

    $mount_dir = $self->get_mount();
    unless (    defined $mount_dir
            and length $mount_dir
            and $mount_dir !~ /^\\$/) {
        $self->_set_err("Error: Vol::clean() Mount point is missing");
        return $ClientUtil::FALSE;
    }

    # Verify that the mount_dir is NOT root
    if ($mount_dir =~ /^\/$/) {
        $self->_set_err("Error: Vol::clean() " . "Supplied mount_dir is invalid ($mount_dir), can not specify ROOT!! ");
        return $ClientUtil::FALSE;
    }

    # Remove datatest files '_io*-' in datatest name.
    $cmd = "rm $mount_dir/\*_io\*\-\*";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 and $status != 2) {    # Not filename found
        $self->_set_err(
                       "Error: Vol::clean() " . "Unable to execute command on volume \'$name\' in diskgroup '$vgname' ",
                       @output);

        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: create_fs
# Objective:       create a filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub create_fs {
    my ($self, %args) = @_;
    my ($cmd, $type, $vgname, $name, $status, @output);

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?type$/) {
            $type = $args{$_};
            next;
        }
    }

    ### check for existance of required -type argument, if not
    ### set default to UFS
    if (not defined($type) or $type eq "") {
        $type = "vxfs";
    }

    # create the filesystem
    $cmd = "/usr/sbin/mkfs -F $type -o largefiles $vgname/r$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($status > 0) {
        $self->_set_err("Error: Vol::create_fs Unable to create $type filesystem on volume '$name'", @output);
        return $ClientUtil::FALSE;
    }

    $self->{filesystem} = $type;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self, %args) = @_;

    my ($cmd, $force, $vgname, $name, @output, $status, @vols);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Vol::destroy()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vol::destroy() Volume Group name is a required parameter");
        return $ClientUtil::FALSE;
    }

    # if force option specified umount all volumes so deport
    # can occur without error.  Make note of which vols were
    # mounted so they can be remounted after rename occurs
    if (defined $force and $force =~ /1|on|y/i) {
        $self->sanlog(-msg => "determine if volume has filesystem");
        if ($self->is_fs()) {
            $self->sanlog(-msg => "determine if volume is mounted");
            if ($self->is_mounted()) {
                $self->sanlog(-msg => "unmount filesystem " . $self->get_name());
                $self->unmount();
                return $ClientUtil::FALSE if ($self->error());
            }
        }
    }

    # remove the logical volume
    if ($name =~ /dev/) {
        $name =~ s/$vgname\///;
    }

    $cmd = "/usr/sbin/lvremove -f $vgname/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vol::destroy() Unable to destroy volume \'$name\' in Volume Group '$vgname'", @output);

        return $ClientUtil::FALSE;
    }

    # Remove the volume from the parent vol group object
    foreach my $volObj (@{$self->{parent}->get_vols()}) {
        if ($volObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vols, $volObj;
        }
    }

    $self->{parent}->{vols} = \@vols;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;
    my (@buf, $cmd, $vgname, $name, @output, $status, $header, $summary);

    $self->_clear_err();

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vol::display() volume name is a required parameter");
        return (wantarray()) ? () : [];
    }

    # display the logical volume
    $cmd = "/usr/sbin/lvdisplay $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vol::display() Unable to display info for volume \'$name\' in volume group \'$vgname\'",
                        @output);

        return (wantarray()) ? () : [];
    }

    $header = sprintf("\nVol Summary:\n"
                        . "%-12.12s %-7.7s %-5.5s %-12.12s %-s\n"
                        . "%-12.12s %-7.7s %-5.5s %-12.12s %-32.32s",
                      "NAME",   "LAYOUT", "TYPE",  "SIZE(mb)", "MOUNT_PT",
                      "=" x 12, "=" x 7,  "=" x 5, "=" x 12,   "=" x 32);

    $summary = sprintf("%-12.12s %-7.7s %-5.5s %-12.12s %-s",
                       $self->get_name(), $self->get_layout(), $self->get_fs(),
                       $self->get_size(), $self->get_mount());

    @buf = ($header, $summary, "\nVol Detail:", @output);

    return (wantarray()) ? @buf : \@buf;

}

################################################################################
# Subroutine Name: get_fs
# Objective:       Return Vol object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{filesystem}) and length($self->{filesystem})) {
        return $self->{filesystem};
    } else {
        $self->_set_err("Error: Vol::get_fs() Vol object is missing filesystem");
        return "";
    }
}

################################################################################
# Subroutine Name: get_layout
# Objective:       Return volume layout
# Details:         See POD documentation at the end of this file
################################################################################
sub get_layout {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{layout})) {
        return $self->{layout};
    } else {
        $self->_set_err("Error: Vol::get_layout() Vol object is missing layout");
        return "";
    }
}

################################################################################
# Subroutine Name: get_mount
# Objective:       Return mount point of volume
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{mount}) and length $self->{mount}) {
        return $self->{mount};
    } else {
        $self->_set_err("Error: Vol::get_mount() Vol filesystem has no mount point");
        return "";
    }

}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vol object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    my ($vgname, $tmp);

    $self->_clear_err();

    # make sure that the vgname is removed from the
    # LV name since snapdrive cannot handle this.
    $vgname = "/dev/" . $self->get_vgname() . "/";

    if (defined($self->{name})) {
        $tmp = $self->{name};
        $tmp =~ s/$vgname//;
        return $tmp;
    } else {
        $self->_set_err("Error: Vol::get_name() Vol object is missing name");
        return "";
    }

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return ooject size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{size})) {
        return $self->{size};
    } else {
        $self->_set_err("Error: Vol::get_size() Vol object is missing size");
        return "";
    }

}

################################################################################
# Subroutine Name: get_state
# Objective:       Return object state
# Details:         See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{state})) {
        return $self->{state};
    } else {
        $self->_set_err("Error: Vol::get_state() Vol object is missing state");
        return "";
    }

}

################################################################################
# Subroutine Name: get_vgname
# Objective:       Return Vol object Volume Group name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgname {
    my ($self) = @_;
    my ($name);

    $self->_clear_err();

    $name = $self->{parent}->get_name();

    if (defined($name) and length $name) {
        return $name;
    } else {
        $self->_set_err("Error: Vol::get_vgname() Vol object is missing parent name");
        return "";
    }

}

################################################################################
# Subroutine Name: is_fs
# Objective:       Determine if volume has a filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub is_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined($self->{filesystem}) and $self->{filesystem} !~ /raw/i) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if volume is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    $self->_clear_err();
    return $self->{mounted_p};

}

################################################################################
# Subroutine Name: mount
# Objective:       mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub mount {
    my ($self, %args) = @_;
    my ($cmd, $type, $vgname, $mount, $name, $pass_thru,
        $status, @output);

    # Get attributes
    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();
    $type   = $self->{filesystem};
    $mount  = undef;
    $pass_thru  = "";

    # check if volume is unformatted
    if (defined $type and $type =~ /raw/) {
        $self->_set_err("Error: Vol::mount()", "Cannot mount a raw filesystem");
        return $ClientUtil::FALSE;
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?path$/) {
            if (not length($mount = $args{$_})) {
                $self->_set_err("Error: Vol::mount()", "Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }

            # Added needed suffixes to mount point.
            my $tmp = $vgname;
            $name =~ s/$vgname\///;
            $tmp  =~ s/\/dev\///;
            $mount .= "/$tmp/$name";
            $self->sanlog(-msg => "using mount pt '$mount'");
            next;
        } elsif (/-?pass_thru$/) {
            $pass_thru = $args{$_};
            $pass_thru =~ s/^\s+//;
            $pass_thru =~ s/\s+$//;
            if ($pass_thru ne "") {
                $pass_thru = "," . $pass_thru;
            }
        }
    }

    # check if the volume is already mounted
    if ($self->is_mounted()) {
        $self->_set_err("Error: Vol::mount() volume '$name' is already mounted");
        return $ClientUtil::FALSE;
    }

    # Use default path if not defined
    unless (defined $mount and length($mount)) {
        my $tmp = $vgname;
        $name =~ s/$vgname\///;
        $tmp  =~ s/\/dev\///;
        $mount = "/mnt/qa/$tmp/$name";
        $self->sanlog(-msg => "using default mount pt '$mount'");
    }

    $self->{mount} = $mount;

    # Create the mount point and open up dir privs
    $cmd = "mkdir -p $mount; chmod 777 $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check status
    if ($status > 0) {
        $self->_set_err("Error: Vol::mount() Unable to create mount point directory for volume '$name'", @output);

        return $ClientUtil::FALSE;
    }

    # Save the mount point with the vol metadata in case
    # a vol is renamed while mounted
    $self->_set_metadata($mount);
    return $ClientUtil::FALSE if ($self->error);

    # Mount the volume's filesystem
    $cmd = "/usr/sbin/mount -F $type -o largefiles$pass_thru $vgname/$name $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($status > 0) {
        $self->_set_err("Error: Vol::mount() Unable to mount volume filesystem", @output);

        return $ClientUtil::FALSE;
    }

    $self->{mounted_p} = 1;

    # Get handle to client object
    $self->{client} = $self->{parent}{parent}{parent};

    # Comment out for now, can cause reboot to hang with stale entries in vfstab
    #$self->{client}->mount_entry_add(-device_mount  => "/dev/" . $self->get_vgname() . "/$name",
    #                                 -mount_pt      => $mount,
    #                                 -fs_type       => $self->get_fs(),
    #                                 -mount_options => "log,largefiles");
    #if ($self->{client}->error) {
    #    $self->_set_err($self->{client}->errmsg());
    #    return $ClientUtil::FALSE;
    #}

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: rename()
# Objective:       NOOP FOR HPUX
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, $newname) = @_;

    return $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: rescan()
# Objective:       rescan volume
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;
    my ($mount);

    # Currently just handles remounting.
    # Needed after snapdrive restore
    $self->_clear_err();

    $mount = $self->_init_mount();
    unless (defined $mount and length $mount) {
        $self->{mount} = undef;
    } else {
        $self->{mounted_p} = $self->_is_mounted_real();
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: unmount()
# Objective:       unmount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount {
    my ($self, %args) = @_;
    my ($cmd, $type, $vgname, $fs, $mount, $name, $status, $force, $option, @output);

    $self->_clear_err();

    $option = "";

    # parse hash
    foreach (keys %args) {
        if (/-?force$/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Vol::unmount()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # HP-UX does not support -f option, just inform user
    if (defined $force and $force =~ /1|y|yes|t|true/) {
        $self->sanlog(-msg => "HP-UX does not support the force option on umount");
    }

    # Get attributes
    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();

    $mount = $self->{mount};

    $fs = $self->{filesystem};

    # check if volume contains a filesystem

    unless ($fs !~ /raw/i) {
        $self->_set_err("Error: Vol::unmount()", "Cannot unmount a raw device");
        return $ClientUtil::FALSE;
    }

    # check if volume is mounted
    unless (defined $mount and length($mount)) {
        $self->_set_err("Error: Vol::unmount()", "Cannot unmount a filesystem that is not mounted");
        return $ClientUtil::FALSE;
    }

    # Unmount the volume's filesystem
    $cmd = "/usr/sbin/umount $option $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($status > 0) {
        $self->_set_err("Error: Vol::unmount() Unable to unmount volume's filesystem", @output);
        return $ClientUtil::FALSE;
    }

    $self->{mounted_p} = 0;

    # Do NOT cleanup the mount point file. Handle this in the
    # Vg:destroy method since a configuration change with a restore
    # can leave us in a state where the mount point file does not
    # exist and it will cause other tests downsteam to fail.

    # Clear vfstab mount entry - remove for now
    # client = lvm <-- vg <-- vol
    $self->{client} = $self->{parent}{parent}{parent};

    #$self->{client}->mount_entry_delete($mount);
    #if ($self->{client}->error) {
    #    $self->_set_error($self->{client}->errmsg());
    #    return $ClientUtil::FALSE;
    #}

    $self->sanlog(-msg => "removing mount point '$mount'");

    # Remove directory head
    my ($base, $parent) = fileparse($mount);
    $cmd = "rm -rf $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vol::unmount() Unable to remove mount point '$mount'", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
## Subroutine Name: get_raw_device
## Objective:       Return path to raw vol
## Details:         See POD documentation at the end of this file
#################################################################################
sub get_raw_device {
    my ($self) = @_;

    $self->_clear_err();
    ### get name of disk group
    my $vgname = $self->get_vgname();

    ### get name of volume
    my $name = $self->get_name();

    my $tmp  = $name;
    my $tmp2 = $vgname;

    if ($name =~ /dev/) {
        $tmp =~ s/$vgname\///;
    }

    # the vgname might not contain "/dev"
    if ($name =~ /dev/) {
        $tmp =~ s/\/dev\///;
    }

    if ($vgname =~ /dev/) {
        $tmp2 =~ s/\/dev\///;
    }

    return "/dev/$tmp2/r$tmp";

}

################################################################################
##
# P R I V A T E   S U B R O U T I N E S
##
################################################################################

################################################################################
# Subroutine Name: _create()
# Objective:       create a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self) = @_;
    my ($cmd, $device_cnt, $vg, $name, $layout, $size, @output, $status);

    $vg         = "/dev/" . $self->get_vgname();
    $device_cnt = $self->{parent}->get_device_cnt();
    $name       = $self->get_name();
    $size       = $self->{size};
    $layout     = $self->{layout};

    # check size, convert to MB
    my ($nbr, $unit) = $size =~ /(\d+)(\w+)/;

    if ($unit =~ /g/i) {
        $size = $nbr * 1024;
    }

    # formulate the lvcreate command
    $cmd = "/usr/sbin/lvcreate -A n -L $size -n $name ";

    # add striping options
    if ($device_cnt > 1 and $layout =~ /stripe/i) {
        $cmd .= "-i $device_cnt -I 32k ";
    }

    $cmd .= "$vg";

    # execute the command
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check status from the lvcreate
    if ($status > 1) {
        $self->_set_err("Error: Vol::_create Unable to create volume \'$name\'", "CMD: $cmd", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _get_metadata
# Objective:      Retrieve metadata for attribute
# Details:        See POD documentation at the end of this file
################################################################################
sub _get_metadata {
    my ($self) = @_;

    my ($cmd, $vgname, $name, @output, $status, $value, $mount_dir, $filename);

    $value = "";

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();

    # For HP, we just read the file that has been
    # created in _set_metadata. This should contain
    # the last location of where the filesystem was
    # mounted.
    $name   =~ s/$vgname\///;
    $vgname =~ s/\/dev\///;
    $mount_dir = "/mnt/qa/";
    $filename  = $mount_dir . $vgname . "_" . $name;

    $cmd = "cat $filename";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check the status from the call. If it failed, assume
    # the filesystem is not mounted.
    if ($status > 0) {
        return "";
    }

    return $output[0];

}

################################################################################
# Subroutine Name: _init_fstype
# Objective:      Initialize fstype for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_fstype {
    my ($self) = @_;

    my ($cmd, $vgname, $name, @output, $status);

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $vgname . "/" . $self->get_name();

    # Start volume if not enabled
    if ($self->get_state() =~ /unavailable/i) {
        $self->_vol_start();
    }

    # formulate the fstyp command
    $cmd = "/usr/sbin/fstyp $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check the results
    if ($status > 0) {
        if ($output[0] =~ /unknown/i) {
            return "raw";
        } else {
            $self->_set_err("Error: Vol::_init_fstype() ",
                            "Unable to retrieve  filesystem type for volume \'$name\' in Volume Group '$vgname'",
                            @output);
            return "";
        }
    }

    return $output[0];

}

################################################################################
# Subroutine Name: _init_layout
# Objective:      Initialize layout of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_layout {
    my ($self) = @_;

    my ($cmd, $vgname, $layout, $name, @output, $status, $stripes);

    $layout = "";

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $vgname . "/" . $self->get_name();

    # get lvm information
    $cmd = "/usr/sbin/lvdisplay $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vol::_init_layout() "
                          . "Unable to retreive layout for volume \'$name\' in Volume Group '$vgname'",
                        @output);

        return "";
    }

    # find field Stripes in the output
    $stripes = (grep(/Stripes/, @output))[0];

    # simply return if this is a concatentation or a striped logical volume
    if ($stripes !~ /0/) {
        $layout = "striped";
    } else {
        $layout = "concat";
    }

    return $layout;

}

################################################################################
# Subroutine Name: _init_mount
# Objective:      Initialize mount point for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_mount {
    my ($self, $import) = @_;

    my ($cmd, $vgname, $mount, $name, $path, @output, $status);

    $mount  = "";
    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();

    $cmd = "grep '$name' /etc/mnttab | grep -v vg00";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($status > 0) {
        # If volumes are bing imported assume then
        # automatically mount fs with last stored
        # mount point (or default path) if a mount
        # point does not exist
        if ($import) {
            $mount = $self->_get_metadata();
            if (defined $mount and length($mount)) {
                ($path, $vgname, $name) = $mount =~ /(\S+)\/(\w+)\/(\w+)$/;
            }

            if (defined $path and length $path) {
                $self->mount(-path => $path);
            } else {
                $self->mount();
            }

            return "" if $self->error();
        }
    } else {
        # return the volume mount point.
        my @buf = split(/\s+/, $output[0]);
        $mount = $buf[1];
    }

    return $mount;

}

################################################################################
# Subroutine Name: _is_mounted_real
# Objective:      Check system to see if actually mounted.
# Details:        See POD documentation at the end of this file
################################################################################
sub _is_mounted_real {
    my ($self) = @_;
    my ($cmd, $status, $mount);
    my (@output);

    $mount = $self->{mount};

    if (defined $mount and length($mount)) {
        $cmd = "bdf | grep $mount";

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            # volume is not mounted
            return $ClientUtil::TRUE;
        }
    }

    return $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self) = @_;

    my ($cmd, $vgname, $size, $name, @output, $sectors, $status);

    $size = "";

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $vgname . "/" . $self->get_name();

    # formulate lvdisplay command
    $cmd = "/usr/sbin/lvdisplay $name | grep \"LV Size (Mbytes)\"";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check status
    if ($status > 0) {
        $self->_set_err("Error: Vol::_init_size() "
                          . "Unable to retrieve  size (MB) for volume \'$name\' in Volume Group '$vgname'",
                        @output);
        return "";
    }

    # get the size in Mbytes returned by the lvdisplay
    $size = $output[0];
    $size =~ m/LV Size \(Mbytes\)\s+(\d+)/;
    $size = $1;

    return $size;

}

################################################################################
# Subroutine Name: _init_state
# Objective:      Initialize state of existing host volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_state {
    my ($self) = @_;

    my ($cmd, $name, $state, $vgname, $status, @output);

    $state = "";

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $vgname . "/" . $self->get_name();

    # formulate lvdisplay command
    $cmd = "/usr/sbin/lvdisplay $name | grep \"LV Status\"";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check status
    if ($status > 0) {
        $self->_set_err(
               "Error: Vol::_init_state() " . "Unable to retrieve state for volume \'$name\' in Volume Group '$vgname'",
               @output);

        return "";
    }

    # get the status from the lvdisplay output
    $state = $output[0];
    $state =~ s/LV Status\s+//;

    return $state;

}

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($cmd, $exist, $import, $name, $size, $conn, $layout,
        $log, $parent, $self, $status, @output);

    $self = {};

    bless $self, $class;

    #default layout is concatenation
    $layout = "concat";
    $import = 0;

    $exist = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?layout$/) {
            if (not length($layout = $args{$_})) {
                $self->_set_err("Error: Vol::new()", "Invalid value for argument '-layout'");
                return $self;
            }
            next;
        } elsif (/-?_import$/) {    # private member
            if (not length($import = $args{$_})) {
                $self->_set_err("Error: Vol::new()", "Invalid value for argument '-_import'");
                return $self;
            }
            next;
        } elsif (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: Vol::new()", "Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?size$/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Error: Vol::new()", "Invalid value for argument '-size'");
                return $self;
            }
            next;
        } elsif (/-?exist$/) {
            if (not length($exist = $args{$_})) {
                $self->_set_err("Error: Vol::new()", "Invalid value for argument '-exist'");
                return $self;
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: Vol::new()", "Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Vol::new()", "Missing value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -lvm argument
    unless (defined $parent) {
        $self->_set_err("Error: Vol::new()", "Missing required argument '-parent'");
        return $self;
    }

    # check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: Vol::new()", "Missing required argument '-name'");
        return $self;
    }

    # check for existance of required -name argument
    unless (defined $size) {
        $self->_set_err("Error: Vol::new()", "Missing required argument '-size'");
        return $self;
    }

    # check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: Vg::new()", "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: Vg::new()", "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn}       = $conn;
    $self->{name}       = $name;
    $self->{layout}     = $layout;
    $self->{size}       = $size;
    $self->{parent}     = $parent;
    $self->{state}      = 'ENABLED';
    $self->{filesystem} = 'raw';
    $self->{dt_pattern} = "0x39c39c39c";

    if (not $exist) {
        $self->_create();
        return $self if ($self->error());
    } else {
        $self->sanlog(-msg => "");
        $self->sanlog(-msg => "loading existing volume '$name'");

        $self->{state} = $self->_init_state();
        return $self if ($self->error);
        $self->sanlog(-msg => "  state: " . $self->{state});

        $self->{filesystem} = $self->_init_fstype();
        $self->sanlog(-msg => "  fs: " . $self->{filesystem});
        if ($self->error()) {

            # allow vol to be loaded although it
            # can't be started to determine fs
            if ($self->{filesystem} eq "") {
                $self->_clear_err();
            } else {
                return $self;
            }
        }

        if ($self->{filesystem} !~ /raw/i) {
            $self->{mount} = $self->_init_mount($import);
            return $self if ($self->error);
            $self->sanlog(-msg => "  mnt: " . $self->{mount});
            $self->{mounted_p} = $self->_is_mounted_real();
        }

        $self->{size} = $self->_init_size();
        return $self if ($self->error);
        $self->sanlog(-msg => "  size: " . $self->{size} . " mb");

        $self->{layout} = $self->_init_layout();
        return $self if ($self->error);
        $self->sanlog(-msg => "  layout: " . $self->{layout});
    }

    return $self;

}

################################################################################
# Subroutine Name: _set_metadata
# Objective:      Set metadata for attribute
# Details:        See POD documentation at the end of this file
################################################################################
sub _set_metadata {
    my ($self, $value) = @_;

    my ($cmd, $vgname, $name, @output, $status, $filename, $mount_dir, $tmp);

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();

    # For HP, we will write file into the root of the
    # standard mount directory to store the information
    # information about where the filesystem is mounted
    $mount_dir = "/mnt/qa/";
    $name =~ s/$vgname\///;
    $tmp = $vgname;
    $tmp =~ s/\/dev\///;
    $filename = $mount_dir . $tmp . "_" . $name;

    # simply remove the file to setup for this mount
    $cmd = "rm -f $filename";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check status
    if ($status > 0) {
        $self->_set_err("Error: Vol::_set_metadata() " . "Unable to remove file \'$filename\'", @output);
        return $ClientUtil::FALSE;
    }

    # If the user passed a mount point to this function,
    # add it to the file on the local system.
    if ($value ne "") {

        # place into the file the mount point
        $cmd = "echo $value > $filename";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # check status
        if ($status > 0) {
            $self->_set_err("Error: Vol::_set_metadata() " . "Unable to add data \'$value\'", @output);
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _vol_start
# Objective:      start up volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _vol_start {
    my ($self) = @_;

    my ($cmd, $vgname, $name, @output, $status);

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();

    # formulate the lvchange line
    $self->sanlog(-msg => "activating logical volume '" . $name . "'");
    $cmd = "/usr/sbin/lvchange -a y $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check status
    if ($status > 0) {
        $self->_set_err("Error: Vol::_vol_start() " . "Unable to start volume \'$name\' in Volume Group '$vgname'",
                        @output);

        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _vol_stop
# Objective:      stop volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _vol_stop {
    my ($self) = @_;

    my ($cmd, $vgname, $name, @output, $status);

    $vgname = "/dev/" . $self->get_vgname();
    $name   = $self->get_name();

    # formulate the lvchange line
    $cmd = "/usr/sbin/lvchange -a n $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # check status
    if ($status > 0) {
        $self->_set_err("Error: Vol::_vol_stop() " . "Unable to stop volume \'$name\' in Volume Group '$vgname'",
                        @output);

        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

1;

#############################################################################
#                             POD Documentation
#############################################################################

=pod

=head1 Package Name

Vol.pm

=head1 Authors

=begin html 

Gary Knopp,
Tom Masterson

=end html

=head1 Description

Vol provides an API interface to LVM volumes through the use of the Telnet library.  

=head1 Public Methods

=head2 B<create_fs(%args)>

Creates a filesystem on the volume.  The filesystem is currently created with the 'largefiles' option.

=over 4

=item Inputs

=item * -type

The type of filesystem to create (vxfs or ufs)

=item Outputs

Returns a scalar value of 1 if the filesystem was successfully created, otherwise 0. If an error occurs the
error flag is set, errmsg is set to the failure reason.  

=item Examples

    # Create a vxfs filesystem
    $Test->log("-- Creating volume group 'myvg'");
    if ($volObj->create_fs(-type=>"vxfs")){ 
        $Test->log($volObj->errmsg());
        return $TCase::FAIL; 
    }

=back

=head2 B<destroy()>

Destroy the volume 

=over 4

=item Inputs

None

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set, errmsg
is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("-- Destroy volume");

    $volObj->destroy();
    if ($volObj->error()){
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<display()>

Display detailed volume information

=over 4

=item Inputs

none

=item Outputs

Returns an array or array reference of volume information if successful.  If an error occured the error 
flag is set, errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    $Test->log("-- Retrieving volumeinformation");

    $Test->log($volObj->display());
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }

=back

=head2 B<error()>

This method allows you to determine if any errors occured in the
previous method call.  Note that the error flag is reset at the
beginning of each method, so it should be checked after B<every>
major method call.  If you do not check for errors, your calling
program may fail at some point and you will not know why.

=over 4 

=item Inputs

None

=item Outputs

Returns 1 if an error occured in the previous call, 0 otherwise.

=item Examples

 # Determine if error occurred...
 if ($obj->error) {
    print "Error encountered in previous method call\n";
 }

=back

=head2 B<errmsg()>

This method allows you to get the text of the error message if an error
occurred on the previous method call.

=over 4

=item Inputs

None

=item Outputs

Returns the text of the last error message set.  If no error message is
set, an empty string will be returned.

=item Examples

 # Check for errors.
 if ($obj->error) {
    $errmsg = $obj->errmsg;
    print "An error occurred: $errmsg\n";
 }

=back

=head2 B<get_fs()>

Type of filesystem 

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the filesystem structure.  If no filesystem exists than the value 
'raw' is returned. If an error occurs the error flag is set, errmsg is set to the failure reason,
and "" is returned.        


=item Examples

    print "Volume filesystem is '".$volObj->get_fs()."'";

=back

=head2 B<get_layout()>

Structure of volume 

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the volume structure.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Volume layout is '".$volObj->get_layout()."'";

=back

=head2 B<get_mount()>

Mount point of volume

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the the volume mount point.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Volume is mounted at '".$volObj->get_mount()."'";

=back

=head2 B<get_name()>

Name of volume

=over 4

=item Inputs

none

=item Outputs

Returns the name of the volume.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Volume name is '".$volObj->get_name()."'";

=back

=head2 B<get_size()>

Volume size in megabytes

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and 0 is returned.        


=item Examples

    $size=$volObj->get_size();
    if ($volObj->error){
        $Test->log($volObj->errmsg());
        return $TCase::FAIL;
    }  
   
    print "Volume size is $size mb\n";

=back

=head2 B<get_vgname()>

Name of volume group 

=over 4

=item Inputs

none

=item Outputs

Returns the name of the volume group.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Volume group is '".$volObj->get_vgname()."'";

=back

=head2 B<is_fs()>

Determine if a volume has a filesystem 

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if a filesystem exists, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Examples

    if(not $volObj->is_fs()){
        $volObj->create_fs(-type=>"vxfs");
        ...
        <error handler>
        ...
    }

=back

=head2 B<is_mounted()>

Determine if volume filesystem is mounted.

=over 4

=item Inputs

None

=item Outputs

Returns a value of 1 if filesystem is mounted, otherwise 0.  If an error occured the error flag is set, errmsg
is set to the failure reason.

=item Examples

    if(not $volObj->is_mounted()){
        $volObj->mount();
        ...
        <error handler>
        ...
    }

=back

=head2 B<mount(I<%args>)>

Mount the filesystem

=over 4

=item Inputs

=item * I<-path> (Default: /mnt/qa/[vgName]/[volName])

The mount point location where the filesystem will be accessible. If the path does not exist
it is automatically created with full user access.

=item Outputs

Returns a value of 1 the filesystem is successfully mounted, otherwise 0.  If an error occured 
the error flag is set, errmsg is set to the failure reason.

=item Examples

   $volObj->mount();
   if ($volObj->error){
       $Test->log($volObj->errmsg());
       return $TCase::FAIL;
   }  

=back

=head2 B<unmount(I<%args>)>

Un-Mount the filesystem

=over 4

=item Inputs

=item * I<-f> 

Force the unmount of the filesystem. Note, this is not supported on HP-UX
since the OS does not support this functionality. Since this is not supported,
this flag will be ignored on HP-UX.

=item Outputs

Returns a value of 1 the filesystem is successfully unmounted, otherwise 0.  
If an error occured the error flag is set, errmsg is set to the failure reason.

=item Examples

   $volObj->unmount();
   if ($volObj->error){
       $Test->log($volObj->errmsg());
       return $TCase::FAIL;
   }  

=back

=head2 B<rename($name)>

Rename a volume to a new name

=over 4

=item Inputs

=item * $name

The new name of the volume

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("Rename volumes");
    foreach my $volObj (@volObjs){
        $volObj->rename($volObj->get_name()."_new");
        if ($volObj->error){
            $Test->log($volObj->errmsg);
            return $TCase::FAIL;
        }
    }

=back

=head2 B<rescan()>

Rescan volume to refresh state.  Currently, this method is only refreshing 
the mount point if it advertantly goes away during a restore.

=over 4

=item Inputs

none

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.

=item Examples

    $Test->log("Refresh volume state");
    $volObj->rescan() if ($volObj->is_fs());
    if ($volObj->error){
    $Test->log($volObj->errmsg());
    }

=back

=head2 B<create_dt(I<%args>)>

Create a datatest object that generates I/O on filesystem or raw device under test.

=over 4

=item Inputs

=item * I<-proc> (default: 1)

The number of datatest processes to execute in parallel

=item * I<-limit> (default: '2m')

The amount of data each process thread will write per datatest cycle.  Include unit 
of storage with value ((b)ytes, (k)ilobytes, (m)egabytes, (g)igabytes)

=item * I<-compare> (default: 1)

Determine whether datatest will verify the pattern it wrote to the filesystem.  

=item * I<-dispose> (default: 1)

Determine whether datatest deletes the data it created when its done.  

=item * I<-runtime> 

The length of time the datatest program will run. Include
unit of time with value ((s)econds, (m)inutes, (h)ours, (d)ays)


=item Outputs

Returns a Datatest object if the background process was created sucessfully. If an error occurs the
error flag is set, errmsg is set to the failure reason, and undef hash reference is returned.

=item Examples

        $Test->log("-- Starting I/O");                        
        $dtObj=$volObj->create_dt(-procs=>4,-limit=>"2m",-runtime=>'10s');
        if ($volObj->error) {
        $Test->log($volObj->errmsg());
            return $TCase::FAIL;
        }

=back

=head2 B<set_pattern($pattern)>

Sets pattern that will be used in generating file/raw content 

=over 4

=item Inputs

=item * $pattern (default: 0x39c39c39c);

The pattern to use when generating file content.

=item Outputs

Returns a scalar value of 1 if successful.  If an error occured the error flag is set,
errmsg is set to the failure reason, and 0 is passed back as the return value.


=item Examples

    my $pattern="Welcome to the Sawzall Project";
    $volObj->set_pattern($pattern);
    if ($volObj->error()){
        $Test->log($volObj->errmsg());
    }

=back

=head2 B<get_pattern()>

Retrieve pattern that will be used to generate file/raw content 

=over 4

=item Inputs

none

=item Outputs

Returns a scalar containing the pattern if successful.  If an error occurs the error flag 
is set, errmsg is set to the failure reason, and "" is returned.        


=item Examples

    print "Pattern used to generate data content is   '".$volObj->get_pattern()."'";

=back


