# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the Vxvm Logical Volume Manager on HPUX
## @author latif@netapp.com
## @status private
## @pod ../../Lvm.pm

# Package Name
##############
package San::SanClient::Lvm::HPUX::Vxvm;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanClient::Lvm;
use San::SanLog;
use Error;

require San::SanClient::Lvm::HPUX::Vxvm::Vg;

use vars qw(@ISA %hFormat);

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: initialize
# Objective:       initialize disk under veritas control
# Details:         See POD documentation at the end of this file
################################################################################
sub initialize {
    my ($self, %args) = @_;

    my ($attribute, $cmd, $format, @already_inited, @output, $status, $devices);

    ### default disk type
    $format = "cdsdisk";

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::initialize() Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?format/) {
            if (not length($format = $args{$_}) and $format !~ /cdsdisk|sliced/i) {
                $self->_set_err("Error: HPUX::Vxvm::initialize() Invalid value for argument '-format'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### check for required parameters
    unless (defined $devices and length $devices) {
        $self->_set_err("Error: HPUX::Vxvm::initialize() -devices is a required argument");
        return $ClientUtil::FALSE;
    }

    # add rescan - needed on some system after format and before vxdisksetup
    $self->rescan();
    return $ClientUtil::FALSE if ($self->error());

    ### Get a list of all the devices that are free (not in a dg),
    ### online, "online thinrclm", and already formatted'.  "online thin" is a new online format
    @already_inited = $self->get_free_devices(-inited_states => ["online", "online thin", "online thinrclm"]);

    if ($self->get_version() >= 4) {
        if (not defined $format) {
            $format = "cdsdisk";
        }

        $attribute = "format=" . lc($format);
    } elsif ($self->get_version() =~ /^3\.5/) {
        $format    = "sliced";
        $attribute = "";
    }

    ### Initialize each device in the user specified devices array
    ### that is not already inited (not in @already_inited array).
    foreach my $device (@$devices) {

        # initialized devices that have not been inited or are inited with the wrong format
        if (not(grep /^$device$/, @already_inited) or $hFormat{$device} !~ /$format/i) {
            $self->sanlog(-msg => "$device init ($format)");

            $cmd = "/etc/vx/bin/vxdisksetup -f -i $device $attribute";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)",
                                "Veritas Device setup (initialize) failed.",
                                "OUTPUT = @output");

                return  $ClientUtil::FALSE;
            }
        } else {
            $self->sanlog(-msg => "$device already inited ($format)");
        }

        ### Update the Format hash
        $self->sanlog(-msg => "Device '$device' is $format");
        $hFormat{$device} = $format;
    }

    return  $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: create_vg
# Objective:       create and return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vg {
    my ($self, %args) = @_;

    my ($vgObj, $bAdd);

    $bAdd = 1;

    if (exists $args{-no_add}) {
        $self->sanlog(-msg => "LVM object will not manage VG object");
        delete $args{-no_add};
        $bAdd = 0;
    }

    $vgObj = San::SanClient::Lvm::HPUX::Vxvm::Vg->_new(-parent => $self,
                                                       -conn   => $self->{conn},
                                                       %args);

    if ($vgObj->error) {
        $self->_set_err($vgObj->errmsg());
        return {};
    }

    if ($bAdd eq 1) {
        push @{$self->{vgs}}, $vgObj;
    } else {
        $vgObj = {};
    }

    return $vgObj;

}

################################################################################
# Subroutine Name: destroy_vg
# Objective:       destroy a named diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vg {
    my ($self, $name) = @_;

    my ($vgObj, @vgs);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: HPUX::Vxvm::destroy_vg() vg name is a required argument");
        return $ClientUtil::FALSE;
    }

    ### Delete the vg object and update the parent vg list.
    foreach $vgObj (@{$self->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            $vgObj->destroy();
            if ($vgObj->error) {
                $self->_set_err($vgObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vgs, $vgObj;
        }
    }

    $self->{vgs} = \@vgs;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display
# Objective:       display disk group information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    $self->sanlog(-msg => "Display diskgroup information");

    $cmd = "/usr/sbin/vxprint";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::display() Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

}

###############################################################################
# Method:    get_disk_format
# Objective: Get list of supported disk formats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_format {
    my ($self, $disk) = @_;

    my ($format);

    $self->_clear_err();

    unless (defined $disk and length $disk) {
        $self->_set_err("Error: HPUX::Vxvm::get_disk_format() disk name is a required argument");
        return "";
    }

    $format = $hFormat{$disk};

    unless (defined $format and length $format) {
        $self->_set_err("Error: HPUX::Vxvm::get_disk_format() disk has not been initialized");
        return "";
    }

    ### If disk format is prefixed with 'auto:' remove it.
    $format =~ s/auto://g;

    return $format;
}

###############################################################################
# Method:    get_disk_formats
# Objective: Get list of supported disk formats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_formats {
    my ($self) = @_;

    my (@format);

    $self->_clear_err();

    if ($self->get_version() =~ /^4|5/) {
        push @format, qw(cdsdisk sliced);
    } else {
        push @format, qw(sliced);
    }

    return (wantarray()) ? @format : \@format;
}

###############################################################################
# Method:    get_format_default
# Objective: Get default type to use for disk format
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_format_default {
    my ($self) = @_;

    my ($format);

    $self->_clear_err();

    if ($self->get_version() =~ /^4|5/) {
        $format = "cdsdisk";
    } else {
        $format = "sliced";
    }

    return $format;
}

################################################################################
# Subroutine Name: get_vg
# Objective:       return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vg {
    my ($self, $name) = @_;

    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: HPUX::Vxvm::get_vg() vg name is a required argument");
        return {};
    }

    foreach my $obj (@{$self->{vgs}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: HPUX::Vxvm::get_vg() Unable to find diskgroup '$name'");

    return {};
}

################################################################################
# Subroutine Name: get_vgs
# Objective:       return array of volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgs {
    my ($self) = @_;

    my (@vgs);

    $self->_clear_err();

    if (defined $self->{vgs}) {
        @vgs = @{$self->{vgs}};
    } else {
        $self->_set_err("Error: Vxvg::get_vgs() Vxvm object has no volumes groups");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @vgs : \@vgs;

}

################################################################################
# Subroutine Name: get_vgnames
# Objective:       return array of volume group names
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgnames {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    $cmd = "/usr/sbin/vxdg list | egrep -iv 'NAME|root' | awk '{print \$1}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::get_vgnames() Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_pkginfo()
# Objective:       get detailed lvm pkg information
# Details:         See POD documentation at the end of this file
################################################################################
sub get_pkginfo {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{pkginfo});

    return (wantarray()) ? @{$self->{pkginfo}} : $self->{pkginfo};

}

################################################################################
# Subroutine Name: get_revision()
# Objective:       Get revision of Veritas
# Details:         See POD documentation at the end of this file
################################################################################
sub get_revision {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{rev});
    return "" if ($self->error());

    return $self->{rev};

}

################################################################################
# Subroutine Name: get_name()
# Objective:       Get lvm name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{name});
    return "" if ($self->error());

    return $self->{name};

}

################################################################################
# Subroutine Name: get_version()
# Objective:       Check version of Veritas
# Details:         See POD documentation at the end of this file
################################################################################
sub get_version {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{vers});
    return "" if ($self->error());

    return $self->{vers};

}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;
    my ($bCnt, $cmd, $cnt, $device, $filer, $filername, $lun,
        $refresh, $status, $total, $type, $vol, %hDevs, @all_devices,
        @attr, @devices, @lvm_devices, @output, @used,  $filers, $protocol,
        @inited_states, $inited_state);

    $type  = ".*";
    $total = 0;

    $self->_clear_err();

    # Asssume user supplied -cnt argument
    $bCnt    = 1;
    $refresh = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [];
            }
            delete($args{$_});
            next;
        } elsif (/-?inited_states/) {
            if (not scalar(@inited_states = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-inited_states'");
                return (wantarray()) ? () : [];
            }
            delete($args{$_});
            next;
        } elsif (/-?inited_state$/) {
            if (not length($inited_state = $args{$_})) {
                $self->_set_err("Invalid value for argument '-inited_state'");
                return (wantarray()) ? () : [];
            }
            delete($args{$_});
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vol'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    if (scalar(@inited_states) != 0) {
        my @temp_devices = ();
        foreach my $inited_state (@inited_states) {
            $self->sanlog(-msg => "Getting LUNs with state '$inited_state'");
            push @temp_devices, $self->get_free_devices(-inited_state => "$inited_state", %args);
            if ($self->error) {
                return (wantarray()) ? () : [];
            }
        }
        if (defined($cnt) and $cnt > 0) {
            # Check to make sure we got enough devices back.  It is possible we got back
            # more than we need so handle it if we did
            if (scalar(@temp_devices) < $cnt) {
                $self->_set_err("Only " . scalar(@temp_devices) . " of $cnt requested devices were found");
                return (wantarray()) ? () : [];
            } elsif (scalar(@temp_devices) > $cnt) {
                my @temp_devices2 = ();
                for (my $index = 0; $index < $cnt; $index++) {
                    push(@temp_devices2, $temp_devices[$index]);
                }
                @temp_devices = @temp_devices2;
            }
        }
        return (wantarray()) ? @temp_devices : \@temp_devices;
    }

    # add rescan - needed on some system after format
    $self->rescan();
    if ($self->error()) {
        return (wantarray()) ? () : [];
    }

    ## Build a hash of devices that are in use by LVM
    ## We will run lspv and walk the output.  It should
    ## look something like this:
    # bash-2.05b# vgdisplay -v | grep "PV Name"
    #    PV Name                     /dev/disk/disk2_p2
    #    PV Name                     /dev/dsk/c10t0d0
    my @pv_output;
    $cmd = "/usr/sbin/vgdisplay -v|grep \"PV Name\"";
    ($status, @pv_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    my $lvm_in_use = {};
    foreach my $pv_device (@pv_output) {
        chomp($pv_device);
        # Remove PV Name from the line
        $pv_device =~ s/PV Name//;

        # Remove any leading whitespace
        $pv_device =~ s/^\s+//g;

        # See if this is a legacy device map
        # Run ioscan and you will get this output:
        #
        # Persistent DSF           Legacy DSF(s)
        # ========================================
        # /dev/disk/disk2_p1       /dev/dsk/c2t0d0s1

        my $device;
        if ($pv_device =~ /disk/i) {
            $cmd = "/usr/sbin/ioscan -kN -m dsf $pv_device";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return (wantarray()) ? () : [];
            }
            foreach my $line (@output) {
                if ($line =~ /===|Legacy DSF/i) {
                    next;
                } elsif ($line =~ /\/dev\/dsk\/(\S+)/) {
                    $device = $1;
                }
            }
        } else {
            # Looks like this wasn't a legacy device so just set $device to $pv_device;
            $device = basename($pv_device);
        }

        $self->sanlog(-msg=>"Device '$device' is in use by a volume group", -msg_level => 3);
        $lvm_in_use->{$device} = 1;
    }##END foreach my $pv_device (@pv_output) {

    # get boot device (to exclude)
    my $boot_device = $self->{parent}->_get_boot_device();
    if ($self->error) {
        return (wantarray()) ? () : [];
    }

    # if applicable get mapping of agile to legacy device map
    # See if this is a legacy device map
    # Run ioscan and you will get this output:
    #
    # Persistent DSF           Legacy DSF(s)
    # ========================================
    # /dev/disk/disk2_p1       /dev/dsk/c2t0d0s1

    my $boot_device_hash = {};
    if ($boot_device =~ /rdisk/i) {
        $cmd = "/usr/sbin/ioscan -kN -m dsf $boot_device";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return (wantarray()) ? () : [];
        }
        foreach my $line (@output) {
            if ($line =~ /===|Legacy DSF/i) {
                next;
            } elsif ($line =~ m/\/dev\/r?dsk\/(\S+)/) {
                $boot_device_hash->{$1} = 1;
            }
        }
    } else {
        $boot_device_hash->{basename($boot_device)} = 1;
    }

    # Check if user wants free devices from a specified filer,volume, or lun
    my $vxdisk_hash = {};
    if (not defined($self->{dfmap}) or not defined($self->{raw_vxdisk_list_output})
        or not defined($self->{raw_sanlun_output}) or $refresh == 1) {
        $self->sanlog(-msg=>"creating translation maps:");
        $refresh = 1;

        # Grab the vxdisk -o alldgs list output
        # DEVICE       TYPE            DISK         GROUP        STATUS
        # das_fcp_0_0  auto            -            -            error
        # das_fcp_0_1  auto            -            -            error
        # das_fcp_0_2  auto            -            -            error
        # das_fcp_0_3  auto            -            -            error
        # das_iscs0_0  auto            -            -            error
        # das_iscs0_1  auto            -            -            error
        # das_iscs0_2  auto            -            -            error
        # das_iscs0_3  auto            -            -            error
        # sda          auto:none       -            -            online invalid
        my @vxdisk_list_output;
        $cmd = "/usr/sbin/vxdisk -o alldgs list";
        ($status, @vxdisk_list_output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return (wantarray()) ? () : [];
        }

        # Store the data
        $self->{raw_vxdisk_list_output} = \@vxdisk_list_output;

        # Build a relationship hash keyed by device that looks like this:
        # vxdisk_hash => {device} => {
        #                               dmpnode    = "node_name",
        #                               primary    = "0|1",
        #                               init_state = "online"
        #                            }
        foreach my $line (sort @vxdisk_list_output) {
            $line =~ s/^\s+//g;
            next if ($line =~ /^DEVICE/);

            # Skip any devices flagged as boot, root, or reserved
            next if ($line =~ /boot|root|reserved/i);

            my ($dmpnode, $type, $disk, $diskgroup, @init_state) = split(/\s+/, $line);
            # Update $diskgroup to remove any ()
            $diskgroup =~ s/\(|\)//;

            # Skip any disks who have a disk group defined that isn't "-";
            if ($diskgroup ne "-") {
                $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since the it is used in diskgroup '$diskgroup'",
                              -msg_level => 3);
                next;
            }

            # Make sure type is "auto"
            unless ($type =~ /auto/) {
                $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since the type isn't 'auto'",
                              -msg_level => 3);
                next;
            }

            # Now that we have the dmpnode, run vxdisk list on that node.
            # and find out primary/secondary status
            my @vxdisk_list_node_output;
            $cmd = "/usr/sbin/vxdisk list " . $dmpnode;
            ($status, @vxdisk_list_node_output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return (wantarray()) ? () : [];
            }

            # Output should be
            # [root@auto-rhel5u6-02 ~]# vxdisk list das_fcp_0_0
            # Device:    das_fcp_0_0
            # devicetag: das_fcp_0_0
            # type:      auto
            # flags:     error private autoconfig
            # pubpaths:  block=/dev/vx/dmp/das_fcp_0_0 char=/dev/vx/rdmp/das_fcp_0_0
            # guid:      -
            # udid:      NETAPP%5FLUN%20C-Mode%5F2FiIn%3F-lCptj%5F2FiIn%3F-lCptm
            # site:      -
            # Multipathing information:
            # numpaths:   4
            # sdj             state=enabled   type=primary
            # sdn             state=enabled   type=secondary
            # sdb             state=enabled   type=secondary
            # sdf             state=enabled   type=primary
            foreach my $line (@vxdisk_list_node_output) {
                chomp($line);
                if ($line =~ /^udid\:\s+(.+)/) {
                    # Make sure this is a netapp disk
                    my $udid = $1;
                    unless ($udid =~ /netapp/i) {
                        $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since it isn't a NETAPP lun",
                                      -msg_level => 3);
                        last;
                    }
                } elsif ($line =~ /^flags\:\s+(.+)/i) {
                    my $flags = $1;
                    if ($flags =~ /coordinator/i) {
                        # Disk has coordinator flag, which means it is a fencing device, skip it
                        $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since it is a fenced device");
                        last;
                    }
                } elsif ($line =~ /^(\S+)\s+state=(\S+)\s+type=(\S+)/) {
                    my $device_tmp = $1;
                    my $state = $2;
                    my $type = $3;

                    # Skip the device if it isn't enabled
                    if ($state !~ /enabled/i) {
                        $self->sanlog(-msg => "Skipping device '$device_tmp' since the path isn't enabled",
                                      -msg_level => 3);
                        next;
                    }

                    # Now let's add it to $vxdisk_hash
                    $vxdisk_hash->{$device_tmp}->{"dmpnode"}  = $dmpnode;
                    $vxdisk_hash->{$device_tmp}->{"init_state"} = join(" ", @init_state);
                    if ($type =~ /primary/i) {
                        $vxdisk_hash->{$device_tmp}->{"primary"} = 1;
                    } else {
                        $vxdisk_hash->{$device_tmp}->{"primary"} = 0;
                    }
                } elsif ($line =~ /^(\S+)\s+state=(\S+)$/) {
                    my $device_tmp = $1;
                    my $state = $2;

                    # Skip the device if it isn't enabled
                    if ($state !~ /enabled/i) {
                        $self->sanlog(-msg => "Skipping device '$device_tmp' since the path isn't enabled",
                                      -msg_level => 3);
                        next;
                    }

                    # Now let's add it to $vxdisk_hash.  Since we don't have a "type", we will set
                    # it as a primary path
                    $vxdisk_hash->{$device_tmp}->{"dmpnode"} = $dmpnode;
                    $vxdisk_hash->{$device_tmp}->{"primary"} = 1;
                    $vxdisk_hash->{$device_tmp}->{"init_state"} = join(" ", @init_state);
                } ##END if ($line =~ /^udid\:\s+(.+)/) {
            } ##END foreach my $line (@vxdisk_list_output) {
        } ##END foreach my $device (sort keys %$vxdisk_hash) {

        # make sure our $vxdisk_hash ended up with something in it.
        unless (scalar(keys %$vxdisk_hash)) {
            $self->_set_err("The vxdisk relationship hash is empty. Unable to continue.");
            return (wantarray()) ? () : [];
        }

        # Collect sanlun so we can later sort out internal devices
        my @sanlun_output = $self->{sanlun}->lun_show();

        if ($self->{sanlun}->error) {
            $self->_set_err($self->{sanlun}->errmsg());
            return (wantarray()) ? () : [];
        }

        # Store the raw sanlun output for later
        $self->{raw_sanlun_output} = \@sanlun_output;

        foreach my $line (@sanlun_output) {
            # remove a known string that sanlun is prepending
            # to the first lun it finds (hack).
            $line =~ s|Report bugs to <bug-findutils\@gnu.org>.||;
            $line =~ s/^\s+//g;

            # Example Output depending on architecture
            # controller:                lun-pathname                device filename  adapter  protocol          lun size         lun state
            # controller:                            lun-pathname                          device filename   adapter  protocol          lun size         lun state
            # auto-fas3020-03:  /vol/nate_1_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_0  /dev/rdsk/c10t0d0  td1      FCP            8g (8589934592)     GOOD
            # auto-fas3020-03:  /vol/nate_1_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_0  /dev/rdsk/c12t0d0  td1      FCP            8g (8589934592)     GOOD
            # auto-fas3020-03:  /vol/nate_1_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_0  /dev/rdsk/c6t0d0   td0      FCP            8g (8589934592)     GOOD
            # auto-fas3020-03:  /vol/nate_1_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_0  /dev/rdsk/c8t0d0   td0      FCP            8g (8589934592)     GOOD
            # auto-fas3020-03:  /vol/nate_2_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_1  /dev/rdsk/c10t0d1  td1      FCP            8g (8589934592)     GOOD
            # auto-fas3020-03:  /vol/nate_2_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_1  /dev/rdsk/c12t0d1  td1      FCP            8g (8589934592)     GOOD
            # auto-fas3020-03:  /vol/nate_2_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_1  /dev/rdsk/c6t0d1   td0      FCP            8g (8589934592)     GOOD
            # auto-fas3020-03:  /vol/nate_2_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_1  /dev/rdsk/c8t0d1   td0      FCP            8g (8589934592)     GOOD
            # auto-fas3020-04:  /vol/nate_1_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_0  /dev/rdsk/c12t0d2  td1      FCP            8g (8589934592)     GOOD
            # auto-fas3020-04:  /vol/nate_1_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_0  /dev/rdsk/c10t0d2  td1      FCP            8g (8589934592)     GOOD
            # auto-fas3020-04:  /vol/nate_1_1_auto_hprx2600_01/nate_auto-hprx2600-01_hpux_0  /dev/rdsk/c8t0d2   td0      FCP            8g (8589934592)     GOOD
            if ($line =~ /^(\S+):\s*\/vol\/(\S+)\/(\S+)\s+\/dev\/rdsk\/(\S+)\s+\S+\s+(\S+)\s+/) {
                my $filer_tmp    = $1;
                my $vol_tmp      = $2;
                my $lun_tmp      = $3;
                my $device_tmp   = $4;
                my $protocol_tmp = $5;

                # Now we want to look in the $vxdisk_hash for the device.  If it is reserved then we will
                # skip it.  If it isn't reserved, we will grab the dmpnode and add it to the hash with the
                # rest of the information.
                print "Checking for device '$device_tmp'\n";
                if (exists($vxdisk_hash->{$device_tmp})) {
                    print "Found device '$device_tmp'\n";
                    # Store all of the collected information
                    $hDevs{$device_tmp} = {
                                            filer       => $filer_tmp,
                                            vol         => $vol_tmp,
                                            lun         => $lun_tmp,
                                            protocol    => $protocol_tmp,
                                            dmpnode     => $vxdisk_hash->{$device_tmp}->{'dmpnode'},
                                            init_state  => $vxdisk_hash->{$device_tmp}->{'init_state'},
                                          };
                }
            } ##END if ($line =~ /^(\S+):\s*\/vol\/(\S+)\/(\S+)\s+\/dev\/rdsk\/(\S+)\s+\S+\s+(\S+)\s+/) {
        } ##END foreach my $line (@sanlun_output) {

        # We now need to iterate over the hDevs hash and look to see if it is lvm_in_use hash or
        # the boot device hash.
        my $dmp_nodes_to_skip = {};
        foreach my $device (keys %hDevs) {
            if (exists($lvm_in_use->{$device}) or exists($boot_device_hash->{$device})) {
                $dmp_nodes_to_skip->{$hDevs{$device}{'dmpnode'}} = 1;
            }
        }

        # Now we want to walk through the $hDevs hash and eliminate any duplicate paths
        # for dmpnodes.  We likely have multiple primary and secondary paths for each dmpnode
        # and we only want to represent one primary path for each dmpnode in the final hash.
        my $temp_dmpnode_hash = {};
        my $hDevs_primary = {};
        foreach my $device (sort keys %hDevs) {
            # Check the primary/secondary status for the device in the
            # previously created vxdisk_hash.
            if (exists($vxdisk_hash->{$device})) {
                if (defined($vxdisk_hash->{$device}->{'primary'})
                    and $vxdisk_hash->{$device}->{'primary'} == 1) {
                    # Looks like this is a primary path so see if
                    # the dmpnode has already been represented.
                    my $dmpnode = $hDevs{$device}{'dmpnode'};

                    # Make sure we don't want to skip this dmp node because it is in use by LVM or
                    # is a boot device
                    if (exists($dmp_nodes_to_skip->{$dmpnode})) {
                        next;
                    }

                    # Add the dmp node to the temp hash so long as it isn't already there.
                    unless (exists($temp_dmpnode_hash->{$dmpnode})) {
                        # This dmpnode has not been represented yet.
                        $temp_dmpnode_hash->{$dmpnode}->{$device} = $hDevs{$device};
                    }
                } ##END if (defined($vxdisk_hash->{$device}->{'primary'}
            } ##END if (exists($vxdisk_hash->{$device})) {
        } ##END foreach my $device (sort keys %hDevs) {

        # Now take all of the primary paths and store them in a new hDevs_primary hash
        foreach my $dmpnode (sort keys %$temp_dmpnode_hash) {
            foreach my $device (sort keys %{$temp_dmpnode_hash->{$dmpnode}}) {
                $hDevs_primary->{$device} = $temp_dmpnode_hash->{$dmpnode}->{$device};
            }
        }

        # Make sure we found veritas luns.  If we didn't, then stop now.
        unless (scalar(keys %$hDevs_primary)) {
            $self->_set_err("Failed to find any valid devices under Veritas control");
            return (wantarray()) ? () : [];
        }

        # save map for next call of function
        $self->{dfmap} = $hDevs_primary;
    }

    # Make sure we have the dfmap and raw_sanlun_output.  If
    # we don't, then something went wrong earlier and we can't continue.
    unless (scalar(keys %{$self->{dfmap}} > 0)) {
        $self->_set_err("Unable to continue because the Device Map (dfmap) is empty");
        return (wantarray()) ? () : [];
    }
    unless (scalar(@{$self->{raw_sanlun_output}} > 0)) {
        $self->_set_err("Unable to continue becasue 'sanlun lun show -p' raw device output is empty");
        return (wantarray()) ? () : [];
    }
    unless (scalar(@{$self->{raw_vxdisk_list_output}} > 0)) {
        $self->_set_err("Unable to continue because 'vxdisk -o alldgs list' raw output is empty");
        return (wantarray()) ? () : [];
    }

    # If not -cnt then assume all free disks
    # under lvm control are wanted
    unless (defined $cnt and length($cnt)) {
        $cnt  = 10000;
        $bCnt = 0;
    }

    # Walk through the  dfmap and filter
    foreach my $device (sort keys %{$self->{dfmap}}) {
        # Filter based on single filer
        if (defined $filer and length $filer) {
            if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i) {
                $self->sanlog(-msg => "Skipping device '$device' since it doesn't match filer '$filer'",
                              -msg_level => 3);
                next;
            }
        }

        # Filter based on volume
        if (defined $vol and length $vol) {
            if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i) {
                $self->sanlog(-msg => "Skipping device '$device' since it doesn't match volume '$vol'",
                              -msg_level => 3);
                next;
            }
        }

        # Filter based on lun
        if (defined $lun and length $lun) {
            if ($self->{dfmap}{$device}{lun} !~ /$lun/i) {
                $self->sanlog(-msg => "Skipping device '$device' since it doesn't match lun '$lun'",
                              -msg_level => 3);
                next;
            }
        }

        # Filter based on protocol
        if (defined $protocol and length $protocol) {
            if ($self->{dfmap}{$device}{protocol} !~ /$protocol/i) {
                $self->sanlog(-msg => "Skipping device '$device' since it doesn't match protocol '$protocol'",
                              -msg_level => 3);
                next;
            }
        }

        # Filter based on list of filers
        if (defined $filers and scalar(@{$filers}) > 0) {
            my $found = 0;
            # Look through each filer in the filers list
            my $filer_list = join("|", @$filers);
            foreach my $filer_f (@{$filers}) {
                if ($self->{dfmap}{$device}{filer} =~ /^$filer_f$/i) {
                    # If filer matches devices filer, set found to 1
                    $found = 1;
                }
            }

            # If the device is not part of the filers list caller requested skip it.
            if ($found == 0) {
                $self->sanlog(-msg => "Skipping device '$device' since it doesn't match filers '$filer_list'",
                              -msg_level => 3);
                next;
            }
        }

        # Filter based on init_state
        if (defined $inited_state and length $inited_state) {
            next unless ($self->{dfmap}{$device}{init_state} =~ /^$inited_state$/i);
            $self->sanlog(-msg => "Selecting device '$device' since it matches inited_state '$inited_state'",
                          -msg_level => 3);
        }

        # Since we are returning veritas devices, we need to return the dmpnodes
        push @devices, $self->{dfmap}{$device}{dmpnode} if (++$total <= $cnt);
    }

    # If user supplied cnt and free devices fell short
    # generate error
    if ($bCnt == 1 and scalar(@devices) < $cnt) {
        $self->_set_err("Only " . scalar(@devices) . " of $cnt requested devices were found");
        return (wantarray()) ? () : [] ;
    } elsif (scalar(@devices) == 0) {
        $self->sanlog(-msg => "WARNING requested device list is EMPTY");
        return (wantarray()) ? () : [] ;
    }

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: load_vg
# Objective:       load an existing diskgroup that is not in memory
# Details:         See POD documentation at the end of this file
################################################################################
sub load_vg {
    my ($self, $name) = @_;

    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("ERROR: Vxvm::load_vg() Invalid or missing required paramter \$name");
        return {};
    }

    ### Check if vg has already been loaded
    $vgObj = $self->get_vg($name);
    if ($self->error) {
        # dg has not been load yet
        $self->_clear_err();

        # load dg
        $self->_create_vgs($name);
        return {} if ($self->error);
        $vgObj = $self->get_vg($name);
    } else {
        $self->sanlog(-msg => "INFO: Lvm::load_vg() Volume group '$name' was already loaded");
    }

    return $vgObj;

}

################################################################################
# Subroutine Name: unload_vg
# Objective:       Remove the vg object under LVM object control
# Details:         See POD documentation at the end of this file
################################################################################
sub unload_vg {
    my ($self, $name) = @_;

    my ($vgObj, $bFound, @vgs);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("ERROR: Vxvm::unload_vg() Invalid or missing required paramter \$name");
        return $ClientUtil::FALSE;
    }

    $bFound = 0;

    foreach $vgObj (@{$self->{vgs}}) {
        ($vgObj->get_name() =~ /^$name$/)
          ? $bFound = 1
          : push @vgs, $vgObj;
    }

    $self->{vgs} = \@vgs;

    unless ($bFound) {
        $self->_set_err("ERROR: Vxvm::unload_vg() Nothing to unload. Vg name '" . $name . "' was not loaded");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: enable_controllers
# Objective:       enable controllers
# Details:         See POD documentation at the end of this file
################################################################################
sub enable_controllers {
    my ($self) = @_;

    my ($cmd, @output, $status);
    my ($ctlr, @controllers);

    $self->_clear_err();

    $self->sanlog(-msg => "Enabling all controllers for Veritas DMP");

    if ($self->is_installed()) {

        $self->rescan();
        return $ClientUtil::FALSE if ($self->error());

        # Get a list of all the controllers
        $cmd = "/etc/vx/bin/vxcntrllist";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: HPUX::Vxvm::enable_controllers() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # skip controller 0 (root disk)
        @controllers = grep (!/c0/, @output);

        # traverse through the controllers and enable each.
        foreach $ctlr (@controllers) {
            $cmd = "/usr/sbin/vxdmpadm enable ctlr=$ctlr";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: HPUX::Vxvm::enable_controllers() Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: rescan
# Objective:       To run the vxdctl command to rescan veritas's list of
#                  disks that it thinks are under it's control.
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    # Call the apporiate enable function on the mp object
    if ($self->{parent}->{mp_type} ne "" && defined($self->{parent}->{mp})) {
        $self->{parent}->{mp}->enable(-conn => $self->{conn});
        if ($self->{parent}->{mp}->error()) {
            $self->_set_err($self->{parent}->{mp}->errmsg());
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Error: HPUX::Vxvm::rescan() no mp_type specified, need an multipath type set");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: restart_recovery_daemon
# Objective:       Method that will stop the recovery deamon and reset
# Details:         See POD documentation at the end of this file
################################################################################
sub restart_recovery_daemon {
    my ($self, $interval) = @_;

    my ($cmd, @output, $status, $policy, $policy_string);

    $self->_clear_err();

    if (not defined ($interval)) {
        $self->sanlog(-msg => "Restore interval was not defined, thus it will not be reset (could be cleanup).");
        return $ClientUtil::TRUE;
    }

    my $version = $self->get_version();
    my $mp = $self->_get_version_mp();
    $mp =~ s/MP//;
    if ($mp eq "" or !defined($mp)) {
        $mp = 0;
    }

    if ($version =~ /^4/) {
        $policy = "check_all";
    } else {
        # default for 5.0+ is check_disabled
        $policy = $self->{parent}->{vx_restore_policy};
    }

    if ($self->_is_installed()) {
        $self->sanlog(-msg => "Setting Veritas Recovery Dameon Restore interval to $interval seconds");
        $self->sanlog(-msg => "Setting the Veritas Restore Policy to $policy");

        # with 5.0 MP3, there is new way to set restore daemon, use this if your at that revision.
        # This allows the setting to be persistent across reboots
        if (($version == 5.0 and $mp >= 3) or $version >= 5.1) {
            # set dmp_restore_interval
            $cmd = "/usr/sbin/vxdmpadm settune dmp_restore_interval=$interval";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                foreach my $line (@output) {
                    if (($line =~ /vxdmpadm ERROR/) && ($line !~ /same as its existing/)) {
                        $self->_set_err("Unable to set dmp_restore_interval");
                        return $ClientUtil::FALSE;
                    }
                }
            }

            # set dmp_restore_policy
            $cmd = "/usr/sbin/vxdmpadm settune dmp_restore_policy=$policy";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                foreach my $line (@output) {
                    if (($line =~ /vxdmpadm ERROR/) && ($line !~ /same as its existing/)) {
                        $self->_set_err("Unable to set dmp_restore_policy");
                        return $ClientUtil::FALSE;
                    }
                }
            }
        } else {
            # stop the restored
            $cmd = "/usr/sbin/vxdmpadm stop restore";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }

            if (defined($policy)) {
                $policy_string = "policy=$policy";
            }

            # Sleep for 2 seconds to make sure the restored completely stops
            sleep(2);

            # start the restored
            $cmd = "/usr/sbin/vxdmpadm start restore interval=$interval $policy_string";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }

            # set status on the setting, just to get it in the logs.
            $cmd  = "/usr/sbin/vxdmpadm stat restored";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($conn, $log, $lvm, $loadcfg, $parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::_new() Invalid value for argument '-lvm'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::_new() Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?loadcfg/) {
            if (not length($loadcfg = $args{$_}) or $loadcfg !~ /no|yes|1|0|off|on/i) {
                $loadcfg = 1;
            }
            next;
        }
    }

    ### check for existance of required -lvm argument
    if (grep (/^-lvm$/, (keys %args))) {
        unless ($args{-lvm} =~ /^vxvm|dmp$/i) {
            $self->_set_err("Error: HPUX::Vxvm::_new()", "Invalid value for argument '-lvm' \(" . $args{-lvm} . "\)");
            return $self;
        }
    } else {
        $self->_set_err("Error: HPUX::Vxvm::_new()", "Missing required argument'-lvm'");
        return $self;
    }

    ### check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanClient/i) {
            $self->_set_err("Error: HPUX::Vxvm::_new()", "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: HPUX::Vxvm::_new()", "Missing required argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn}   = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{vgs} = [];

    ### check to make sure that LVM is installed
    if (not $self->_is_installed()) {
        return $self if ($self->error);
    }

    ### create veritas vg objects for existing disk groups
    if ($loadcfg) {
        $self->_create_vgs();
        return $self if ($self->error);

        ### update map to determine disk format
        $self->_update_disk_format_map();
        return $self if ($self->error);
    }

    # After we create a new vg we need to set enclosure tunings
    $self->_set_vxvm_encl_tunings();
    return $self if ($self->error);

    # Set Veritas recovery deamon settings.
    $self->restart_recovery_daemon($self->{parent}->{vx_restore_interval});
    return $self if ($self->error);

    return  $self;

}

################################################################################
# Subroutine Name: _update_disk_format_map
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _update_disk_format_map {
    my ($self) = @_;

    my ($cmd, @output, $status);

    ### Retrieve the disk groups.
    $cmd = "/usr/sbin/vxdisk -o alldgs list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 1) {
        $self->_set_err("Error: HPUX::Vxvm::_update_disk_format_map()", "Unable to update map");

        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        $line =~ s/auto://;
        my ($device, $format) = $line =~ /^(\S+)s2\s+(\S+)/;
        $hFormat{$device} = $format;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _create_vgs
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vgs {
    my ($self, $vgname) = @_;

    my ($cmd, @output, $status, $vgObj, %hDeport);

    $cmd =
      (defined($vgname) and length($vgname))
      ? "/usr/sbin/vxdg list | egrep -v 'NAME' | egrep '$vgname '"
      : "/usr/sbin/vxdg list | egrep -v 'NAME' ";

    ### Retrieve the volume groups.
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 1) {
        $self->_set_err("Error: HPUX::Vxvm::_create_vgs()", "Unable to load existing diskgroup\(s\)");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        next if ($line =~ /root/i);    ### skip over rootdg
        $line =~ /(\w+).*/;
        $self->create_vg(-name => $1, -deported => 0);
        return $ClientUtil::FALSE if ($self->error);
    }

    if (not defined $vgname) {

        # Create objects for the deported disk groups as well
        $cmd = "/usr/sbin/vxdisk -o alldgs list";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: HPUX::Vxvm::_create_vgs()", "Unable to load deported diskgroups");
            return $ClientUtil::FALSE;
        }

    } else {

        # Determine if dg is deported
        $cmd = "/usr/sbin/vxdisk -o alldgs list | grep '(" . $vgname . ")'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # determine if vxdisk command worked
        if ($status gt 1) {
            $self->_set_err("Error: HPUX::Vxvm::_create_vgs()", "Unable to load deported diskgroups");
            return $ClientUtil::FALSE;
        }
    }

    ### Retrieve the deported disk groups
    foreach my $line (@output) {
        my ($dg) = $line =~ /^\w+s2\s+\S+\s+-\s+\((.*)\)\s+/i;
        if ($dg and $dg !~ /root/i) {
            $hDeport{$dg} += 1;
            ### Vg with multple devices only need to
            ### be created once.
            if ($hDeport{$dg} eq 1) {
                $vgObj = $self->create_vg(-name => $dg, -deported => 1);
                return $ClientUtil::FALSE if ($self->error);
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_disk_id
# Objective:       Return disk id
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_disk_id {
    my ($self, $device) = @_;

    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    $name = "/dev/rdsk/$device" . "s2";

    my $path = $self->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show -v -d $name | grep 'Serial number'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: HPUX::Vxvm::_get_diskid() Unable to get diskid for device \'$device\'", @output);
        return "";
    }

    return $output[0];

}

################################################################################
# Subroutine Name: _is_installed
# Objective:       Check to make sure that the LVM is installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_installed {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "/usr/sbin/swlist VRTSvxvm";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 0) {
        $self->sanlog(-msg => "lvm is installed");
        return $ClientUtil::TRUE;
    } else {
        $self->sanlog(-msg => "lvm is not installed");
        $self->_set_err("Error: HPUX::Vxvm::_is_installed() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Method:    _set_vxvm_encl_tunings
# Objective: Set Veritas enclosure ebased tunings.
# Details:   See POD documentation at the end of this file.
################################################################################
sub _set_vxvm_encl_tunings {
    my ($self) = @_;

    my (@cmds, $line, $cmd, $status, @output, @enclosure_names, $version);
    my ($rec_option, $queue_option);

    my ($vx_50_fast_recovery,
        $vx_50_recovery_option,
        $vx_50_queue_depth,
        $vx_io_policy,
        $vx_50_dmp_path_age,
        $vx_50_dmp_log_level,
        $vx_50_probe_idle_lun,
        $vx_50_monitor_fabric) =
          ($self->{parent}->{vx_50_fast_recovery},
           $self->{parent}->{vx_50_recovery_option},
           $self->{parent}->{vx_50_queue_depth},
           $self->{parent}->{vx_io_policy},
           $self->{parent}->{vx_50_dmp_path_age},
           $self->{parent}->{vx_50_dmp_log_level},
           $self->{parent}->{vx_50_probe_idle_lun},
           $self->{parent}->{vx_50_monitor_fabric});

    # As currently designed, this relies on the thpl sending default values
    # down when the tester has not specified any optional parameters.
    # vx_50_fast recovery, vx_50_recovery_options, vx_50_queue_depth
    # and vx_io_policy
    if ($vx_50_fast_recovery eq "" or $vx_50_recovery_option eq "" or
        $vx_50_queue_depth eq "" or $vx_io_policy eq "") {
        $self->sanlog(-msg => "_set_vxvm_encl_tunings: One or more options parameters have not been passed in.\n" .
                              "This is either because a cleanup is happening, or the parameters were not " .
                              "properly defined.");

        # Return true here because this may not have happened in error.
        # This can happen in cases where cleanup is being done, or
        # optional parameters simply were not provided.
        return $ClientUtil::TRUE;
    }

    @enclosure_names = $self->_get_vxvm_enclosure_name();
    if (scalar(@enclosure_names) == 0) {
        $self->_set_err("No Valid enclosure name was found");
        return $ClientUtil::FALSE;
    }

    # get version (including MP)
    $version = $self->get_version();

    # Get the MP
    my $mp = $self->_get_version_mp();
    $mp =~ s/MP//;
    if ($mp eq "" or not defined($mp)) {
        $mp = 0;
    }

    # Get the RP
    my $rp = $self->_get_version_rp();
    $rp =~ s/RP//;
    if ($rp eq "" or not defined($rp)) {
        $rp = 0;
    }

    # Get the SP
    my $sp = $self->_get_version_sp();
    $sp =~ s/SP//;
    if ($sp eq "" or not defined($sp)) {
        $sp = 0;
    }

    # Log a message indicating the version we are tuning.
    if ($sp > 0) {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version SP" . $sp);
    } elsif ($rp > 0) {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version RP" . $rp);
    } elsif ($mp > 0) {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version MP" . $mp);
    } else {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version flat");
    }

    # Veritas has backported the new 5.0 tunings to Veritas 4.2. Thus we allow
    # the 5.0 tunings to be set for 4.2.
    if  ($version =~ /^(5|6)/ or ($version =~ /4\.1/ and $mp >= 2)) {
        my $count = 0;

        # for 5.1 or greater always set recoveryoption=default
        if ($version >= 5.1) {
            $rec_option = "recoveryoption=default";
            $count++;
        } elsif (defined $vx_50_recovery_option and $vx_50_recovery_option ne "NOT_SET") {
            $rec_option = "recoveryoption=$vx_50_recovery_option";
            $count++;
        } elsif (defined $vx_50_recovery_option and $vx_50_recovery_option eq "NOT_SET") {
            $rec_option = "recoveryoption=throttle";
            $count++;
        }

        # for 5.1 or greater always set recoveryoption=default
        if ($version >= 5.1 or (defined($rec_option) and $rec_option eq "default")) {
            $queue_option = "";
            $count++;
        } elsif (defined $vx_50_queue_depth and $vx_50_queue_depth ne "NOT_SET") {
            $queue_option = "queuedepth=$vx_50_queue_depth";
            $count++;
        } elsif (defined $vx_50_queue_depth and $vx_50_queue_depth eq "NOT_SET") {
            $queue_option = "queuedepth=20";
            $count++;
        }

        # if vx_50_recovery_option or vx_50_queue_depth is set, tune.
        if (defined $queue_option and defined $rec_option) {
            foreach my $enclosure_name (@enclosure_names) {
                $cmd = "/usr/sbin/vxdmpadm setattr enclosure $enclosure_name $rec_option $queue_option";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Failed to set recovery option or queuedepth");
                    return $ClientUtil::FALSE;
                }
            }
        } elsif ($count == 1) {
            $self->_set_err("Veritas requires both queuedepth and recoveryoption to be tuned together " .
                            "(for VRTS versions < 5.1). Please specify values for both " .
                            "VX_50_RECOVERY_OPTION and VX_50_QUEUE_DEPTH");
            return $ClientUtil::FALSE;
        }
    }

    # set fast recovery and dmp log level options
    my $fast_r;
    if ($version =~ /^(5|6)/ or ($version =~ /4\.1/ and $mp >= 2)) {
        if ($version =~ /4\.1/ and $mp >= 2) {
            $fast_r = "off";
        } else {
            $fast_r = $vx_50_fast_recovery;
        }
        $cmd = "/usr/sbin/vxdmpadm settune dmp_fast_recovery=$fast_r";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            foreach $line (@output) {
                if (($line =~ /vxdmpadm ERROR/) and ($line !~ /same as its existing/)) {
                    $self->_set_err("Failed to set DMP fast recovery.");
                    return $ClientUtil::FALSE;
                }
            }
        }

        $cmd = "/usr/sbin/vxdmpadm settune dmp_log_level=$vx_50_dmp_log_level";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            foreach $line (@output) {
                if (($line =~ /vxdmpadm ERROR/) and ($line !~ /same as its existing/)) {
                    $self->_set_err("Failed to set DMP log level.");
                    return $ClientUtil::FALSE;
                 }
            }
        }
    }

    # set probe_idle_lun setting for 5.0MP3+
    if ($version =~ /^(5|6)/) {
        if ($vx_50_probe_idle_lun =~ /NOT_SET/i) {
            # If user did not set the option, then default to off
            $vx_50_probe_idle_lun = "off";
        }

        if ($vx_50_probe_idle_lun !~ /NOT_SET/i) {
            # if a value was set by user or from MP3+ default, then set tunable
            $cmd = "/usr/sbin/vxdmpadm settune dmp_probe_idle_lun=$vx_50_probe_idle_lun";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                foreach $line (@output) {
                    if (($line =~ /vxdmpadm ERROR/) and ($line !~ /same as its existing/)) {
                        $self->_set_err("Failed to set DMP dmp_probe_idle_lun tunable.");
                        return $ClientUtil::FALSE;
                     }
                }
            }
        }
    }

    # set dmp_monitor_fabric tunable for Veritas 5.X and 6.X
    if ($version =~ /^(5|6)/) {
        if (defined($vx_50_monitor_fabric)) {
            if ($vx_50_monitor_fabric eq "NOT_SET") {
                # If the user didn't set the value, set it to the veritas default of "on"
                $vx_50_monitor_fabric = "on";
            }

            # Now tune it
            # changes to tool to make change in MP3+ or 5.1 and greater
            if (($version == 5.0 and $mp >= 3) or $version >= 5.1) {
                $cmd = "/usr/sbin/vxdmpadm settune dmp_monitor_fabric=$vx_50_monitor_fabric";
            } else {
                $cmd = "/usr/sbin/vxddladm settune monitor_fabric=$vx_50_monitor_fabric";
            }
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                foreach $line (@output) {
                    if (($line =~ /vxdmpadm ERROR/) and ($line !~ /same as its existing/)) {
                        $self->_set_err("Failed to set DMP monitor_fabric tunable.");
                        return $ClientUtil::FALSE;
                    } elsif ($line=~/vxddladm ERROR/) {
                        $self->_set_err("Failed to set DMP monitor_fabric tunable.");
                        return $ClientUtil::FALSE;
                    }
                }
            }
        }
    }

    # set IO Policy (if passed in and not empty string)
    if (defined $vx_io_policy and $vx_io_policy eq "NOT_SET") {
        $vx_io_policy = "adaptive";
    }

    if (defined $vx_io_policy) {
        foreach my $enclosure_name (@enclosure_names) {
            $cmd = "/usr/sbin/vxdmpadm setattr enclosure $enclosure_name iopolicy=$vx_io_policy";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Failed to set the IO Policy for Enclosure: $enclosure_name .");
                return $ClientUtil::FALSE;
            }
        }
    }

    # Set the dmp_path_age if it is specified.. If version is 5.1SP1 or later
    # and value of vx_50_dmp_path_age is "NOT_SET" then hard code it to "120".
    # First check to see if we are running 5.1 SP1 or later and force it to 120
    # if it isn't set.
    if ($version =~ /^(5|6)/) {
        if (defined($vx_50_dmp_path_age)) {
            if ($vx_50_dmp_path_age eq "NOT_SET") {
                # We found the vx_50_dmp_path_age variable and it wasn't set by the user
                # or it wasn't defined so adjust it based on the version you are running
                if (($version == 5.1 and $sp > 0) or $version > 5.1) {
                    # Looks like we are running 5.1SP1 or higher so force it to 120
                    $vx_50_dmp_path_age = "120";
                } else {
                    # We are running a version prior to 5.1SP1 so set it to the
                    # veritas default of 300.
                    $vx_50_dmp_path_age = "300";
                }
            }

            # Now tune dmp_path_age if we find $vx_50_dmp_path_age defined
            $cmd = "/usr/sbin/vxdmpadm settune dmp_path_age=$vx_50_dmp_path_age";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                foreach $line (@output) {
                    if (($line =~ /vxdmpadm ERROR/) and ($line !~ /same as its existing/)) {
                        $self->_set_err("Failed to set DMP path_age tunable.");
                    return $ClientUtil::FALSE;
                    }
                }
            }
        }
    }

    # print out all the tunables for each enclosure to get in log
    foreach my $enclosure_name (@enclosure_names) {
        if ($version >= 5.1) {
            @cmds = ("/usr/sbin/vxdmpadm getattr enclosure $enclosure_name iopolicy",
                     "/usr/sbin/vxdmpadm getattr enclosure $enclosure_name recoveryoption",
                     "/usr/sbin/vxdmpadm gettune all");
        } elsif ($version =~ /^5/) {
            @cmds = ("/usr/sbin/vxdmpadm listapm dmpnetapp",
                     "/usr/sbin/vxdmpadm getattr enclosure $enclosure_name iopolicy",
                     "/usr/sbin/vxdmpadm getattr enclosure $enclosure_name recoveryoption",
                     "/usr/sbin/vxdmpadm gettune all",
                     "/usr/sbin/vxddladm gettune all");
        } elsif ($version =~ /4\.1/ and $mp >= 2) {
            @cmds = ("/usr/sbin/vxdmpadm getattr enclosure $enclosure_name iopolicy",
                     "/usr/sbin/vxdmpadm getattr enclosure $enclosure_name recoveryoption",
                     "/usr/sbin/vxdmpadm gettune all",
                     "/usr/sbin/vxddladm gettune all");
        } else {
            @cmds = ("/usr/sbin/vxdmpadm getattr enclosure $enclosure_name iopolicy");
        }

        # Execute, don't bother to check status - just need info in log
        foreach $cmd (@cmds) {
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method:    _get_vxvm_enclosure_name
# Objective: Get the enclosure name of the attached Veritas Enclosure
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_vxvm_enclosure_name {
    my ($self) = @_;

    my ($cmd, $line, $status, @output, @enclosure_names);

    @enclosure_names = ();

    $cmd = "/usr/sbin/vxdmpadm listenclosure all";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("HPUX::Vxvm::_get_vxvm_enclosure_name: Failed to execute command ($cmd)");
        return (wantarray()) ? () : [];
    }

    foreach $line (@output) {
        if (($line !~ /DISCONNECTED/i) && ($line =~ /CONNECTED/i) && ($line !~ /Disk/i)) {
            my ($enclosure_name, undef) = split('\s+', $line);
            $self->sanlog(-msg => "The Enclosure Name found is $enclosure_name.");
            push(@enclosure_names, $enclosure_name);
        }
    }

    if (scalar(@enclosure_names) == 0) {
        $self->_set_err("Error: HPUX::Vxvm::_get_vxvm_enclosure_name: Failed to collect enclosure names");
        return (wantarray()) ? () : [];
    } else {
        $self->sanlog(-msg => "The list of Enclosure Names is @enclosure_names");
        return (wantarray()) ? @enclosure_names : \@enclosure_names;
    }
}

################################################################################
# Subroutine Name: _get_pkginfo
# Objective:       Get lvm package info
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self) = @_;

    my ($cmd, @output, $status, $version_string, $name, $full_version, $rev);
    my ($major_version, $sp, $rp, $mp);

    $self->_clear_err();

    $major_version = "";
    $sp            = "";
    $rp            = "";
    $mp            = "";

    $cmd = "/usr/sbin/swlist VRTSvxvm | grep VRTSvxvm";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to execute command $cmd");
        return $ClientUtil::FALSE;
    }

    $version_string = $output[0];
    $version_string =~ s/\#//;
    $version_string =~ s/^\s+//;

    $name         = (split(/\s+/, $version_string))[0];
    $full_version = (split(/\s+/, $version_string))[1];
    $rev          = $full_version;

    $self->sanlog(-msg => "Veritas version is: $full_version");

    #================================================
    # format of swlist VRTSvxvm returns:
    #
    # 6.0:
    # VRTSvxvm                      6.0.000.000    Veritas Volume Manager by Symantec
    #
    # 5.1SP1RP1:
    # VRTSvxvm                      5.1.101.000    Veritas Volume Manager by Symantec
    #
    # 5.1SP1:
    # VRTSvxvm                      5.1.100.000    Veritas Volume Manager by Symantec
    #
    # 5.1RP2:
    # VRTSvxvm                      5.1.2.000    Veritas Volume Manager by Symantec
    #
    # 5.0.1:
    # VRTSvxvm                      5.0.31.5     Veritas Volume Manager by Symantec
    #================================================
    my ($d1, $d2, $d3, $d4) = split(/\./, $full_version);

    # The version is the first 2 put together
    $major_version = $d1 . "." . $d2;

    # Now is where things get kludgy.  We now have to use the output from
    # lspp above to determin the sp, mp, rp
    if ($major_version >= 5.1) {
        if ($d3 eq "100") {
            $sp = "SP1";
        } elsif ($d3 eq "101") {
            $sp = "SP1";
            $rp = "RP1";
        } elsif ($d3 eq "2") {
            $rp = "RP2";
        }
    } elsif ($major_version == 5.0) {
        if ($d3 == 3) {
            $mp = "MP3";
            if ($d4 eq "200") {
                $rp = "RP2";
            }
        } elsif ($d3 >= 31) {
            # This is 5.0.1.  It should be treated like 5.0 MP3 so
            # we will set the MP to MP3.
            $mp = "MP3";
        } else {
            # Default to MP1 for all 5.0 versions except MP3
            $mp = "MP1";
        }
    }

    $self->{name}    = "VRTSvxvm";
    $self->{vers}    = $major_version;
    $self->{rev}     = $major_version;
    $self->{mp}      = $mp;
    $self->{rp}      = $rp;
    $self->{sp}      = $sp;
    $self->{pkginfo} = \@output;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_version_mp()
# Objective:       Check version of Veritas MP revision
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_version_mp {
    my ($self) = @_;

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{mp});
    return "" if ($self->error());

    return $self->{mp};
}

################################################################################
# Subroutine Name: _get_version_rp()
# Objective:       Check version of Veritas RP revision
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_version_rp {
    my ($self) = @_;

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{rp});
    return "" if ($self->error());

    return $self->{rp};
}

################################################################################
# Subroutine Name: _get_version_sp()
# Objective:       Check version of Veritas SP revision
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_version_sp {
    my ($self) = @_;

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{sp});
    return "" if ($self->error());

    return $self->{sp};
}

1;

