# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the Vxvm Volume Groups on HPUX
## @author latif@netapp.com
## @status private
## @pod ../../Vg.pod

package San::SanClient::Lvm::HPUX::Vxvm::Vg;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

require San::SanClient::Lvm::HPUX::Vxvm::Vg::Vol;

use Tie::IxHash;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $conn, $devices, $disk_cnt, $parent, $log, $name,
        $self, $status, $deported, @output, %hDevice);

    $self = {};

    bless $self, $class;

    $self->_clear_err();

    $deported = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::_new()", "Invalid value for argument '-devices'");
                return $self;
            }
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::_new()", "Missing value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::_new()", "Missing value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::_new()", "Missing value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?deported/) {
            if (not length($deported = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::_new()", "Missing value for argument '-deported'");
                return $self;
            }
            next;
        }
    }

    ### check for existance of required -parent argument
    unless (defined $parent) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::_new()", "Missing required argument '-parent'");
        return $self;
    }

    ### check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::_new()", "Missing required argument '-name'");
        return $self;
    }

    ### check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: HPUX::Vxvm::Vg::_new()", "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: HPUX::Vxvm::Vg::_new()", "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn}     = $conn;
    $self->{name}     = $name;
    $self->{deported} = $deported;
    $self->{disk_cnt} = 0;
    $self->{parent}   = $parent;

    # create the diskgroup if devices were supplied
    if (defined $devices and ref($devices) eq "ARRAY") {
        $self->_create($name, $devices);
        return $self if ($self->error());
    } else {

        # vg must already exist. set device list
        # for all existing and deported dgs.
        $self->sanlog(-msg => "loading existing volume group '$name'");
        $self->{devices} = $self->_get_devices();
        return $self if ($self->error);

        unless ($self->{deported}) {
            $self->_create_vols();
            return $self if ($self->error);
        }

    }

    return $self;

}

################################################################################
# Subroutine Name: _create_vols
# Objective:       create vol object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vols {
    my ($self, %args) = @_;

    my ($cmd, $import, $name, $fsck, $status);
    my (@umount_output,  $vgObj, $volObj, %hDeport, @output);

    ### Identify volumes that are being imported so
    ### that their automatically mounted with their
    ### last filesystem mount point.
    $import =
      (exists $args{-_import})
      ? $args{-_import}
      : 0;

    $fsck =
      (exists $args{-fsck})
      ? $args{-fsck}
      : 0;

    $name = $self->get_name();

    $cmd = "/usr/sbin/vxprint -qhtg $name | egrep '^v  '";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    # Allow status > 1 becuase if no vols are found, egrep will return
    # status of 1, this is ok, we may not have any vols on this vg
    if ($status > 1) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::_create_vols()",
                        "Unable to load existing volumes in disk group",
                        "status=$status", @output);

        return $ClientUtil::FALSE;
    }

    # if there is no diskgroup (because it is disabled), skip the load
    if (grep(/No such disk group/i, @output) == 0) {
        foreach my $line (@output) {
            my (undef, $name, undef, undef, undef, $size) = split(/\s+/, $line);
            my $sz = $size * 512 / (1024 * 1024);
            $volObj = $self->create_vol(-name => $name,
                                        -size => $sz,
                                        -exist => 1,
                                        -_import => $import,
                                        -fsck => $fsck);

            return $ClientUtil::FALSE if ($self->error);
        }
    } else {
        # If no disk group (disabled), look for stale mounted filesystems
        # and umount (or later deport will fail)
        $cmd = "/usr/sbin/mount | grep -e \'$name\'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            foreach my $line (@output) {
                my ($mount) = split(/\s+/, $line);
                $cmd = "/usr/sbin/umount $mount";
                ($status, @umount_output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status > 0) {
                    $self->_set_err("Error: Vg::_create_vols()",
                                    "Unable to umount stale filesystem ($mount)",
                                    "status=$status",
                                    @output);
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_vol
# Objective:       create and return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vol {
    my ($self, %args) = @_;

    my ($volObj);

    ### Change layout to concat if dg contains only 1 device
    if (exists $args{-layout} and length $args{-layout}) {
        if ($args{-layout} =~ /stripe/i and $self->get_device_cnt() eq "1") {
            $args{-layout} = "CONCAT";
        }
    }

    $volObj = San::SanClient::Lvm::HPUX::Vxvm::Vg::Vol->_new(-parent => $self,
                                                             -conn   => $self->{conn},
                                                             %args);

    if ($volObj->error) {
        $self->_set_err($volObj->errmsg());
        return $volObj;
    }

    push @{$self->{vols}}, $volObj;

    return $volObj;
}

################################################################################
# Subroutine Name: destroy_vol
# Objective:       destroy a named vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vol {
    my ($self, $name) = @_;

    my ($volObj, @vols);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::destroy_vol() volname is a required argument");
        return $ClientUtil::FALSE;
    }

    ### Delete the of vg object and update the parent vg list.
    foreach $volObj (@{$self->{vols}}) {
        if ($volObj->get_name() =~ /^$name$/) {
            $volObj->destroy(-force => 1);
            if ($volObj->error) {
                $self->_set_err($volObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vols, $volObj;
        }
    }

    $self->{vols} = \@vols;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_devices
# Objective:       get list of devices used by disk group
# Details:         See POD documentation at the end of this file
################################################################################

sub _get_devices {
    my ($self) = @_;

    my ($cmd, $name, $deported, $status, @output, %hDevice, $disk_cnt);

    tie %hDevice, 'Tie::IxHash';

    $name = $self->get_name();
    $deported = $self->{deported};

    if ($deported == 1) {
        $cmd = "/usr/sbin/vxdisk -o alldgs list | grep '($name)'";
    } else {
        $cmd = "/usr/sbin/vxdisk -o alldgs list | grep ' $name '";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0 && $status != 1) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::_get_devices() Unable to get devices for volgroup \'$name\'", @output);

        return {};
    }

    $self->{disk_cnt} = scalar(@output);

    $disk_cnt = 0;
    foreach my $line (@output) {
        my @flds = split(/\s+/, $line);
        if ($flds[0] =~ /s2/) {
            $flds[0] =~ s/s2//;
            ### remove auto prefix if exists
            $flds[1] =~ s/auto://g;
            $self->{$flds[0]} = $flds[1];
            $hDevice{$flds[0]} = $flds[2];
            $disk_cnt++;
        }
    }

    $self->{disk_cnt} = $disk_cnt;

    return \%hDevice;

}

################################################################################
# Subroutine Name: _create()
# Objective:       create a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self, $name, $devices) = @_;

    my ($attribute, $cmd, $disk_cnt, $disk_name, $format);
    my ($power_fail, @output, $status, %hDevice);

    $disk_cnt  = 0;
    $attribute = "";
    if ($self->{parent}->get_version() >= 4) {
        ### Use the first device to determine the
        ### format of the disk group
        $format = $self->{parent}->get_disk_format((@$devices)[0]);
        $attribute = "cds=off" if ($format !~ /cds/);
    }

    foreach my $device (@$devices) {
        $disk_cnt++;

        $disk_name = "disk-$disk_cnt";

        if ($disk_cnt == 1) {
            $cmd = "/usr/sbin/vxdg init $name $disk_name=$device $attribute";
        } else {
            $cmd = "/usr/sbin/vxdg -g $name adddisk $disk_name=$device";
        }

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # determine if vxdg command worked
        if ($status > 0) {
            $self->_set_err("Error: HPUX::Vxvm::Vg::_create Unable to create diskgroup \'$name\'", @output);

            return $ClientUtil::FALSE;
        }
        $self->{$device} = $format;
        $hDevice{$device} = $disk_name;

    }

    # set VRTS power fail timeout value. Hardcode to 120 for 4.1
    if ($self->{parent}->get_version() =~ /^4/) {
        $power_fail = 120;
    } else {
        $power_fail = $self->{parent}->{parent}->{vx_power_fail_timeout};
    }


    $cmd = "/opt/VRTS/bin/vxpfto -g $name -t $power_fail";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::_create Unable to set VxVM Powerfail Timeout on \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    $self->{devices}  = \%hDevice;
    $self->{disk_cnt} = $disk_cnt;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: add_device()
# Objective:       add a device to the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub add_device {
    my ($self, $device) = @_;

    my ($cmd, $disk_cnt, $disk_name, $name, @output, $status);

    $self->_clear_err();

    ### Get name of volume group.
    $name = $self->get_name();

    ### Increment device counter
    $disk_cnt = $self->{disk_cnt};

    $disk_cnt++;

    $disk_name = "disk-$disk_cnt";

    $cmd = "/usr/sbin/vxdg -g $name adddisk $disk_name=$device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::add_device Unable to create diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    $self->{devices}->{$device} = $disk_name;
    $self->{disk_cnt}++;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remove_device()
# Objective:       remove a device from the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub remove_device {
    my ($self, %args) = @_;

    my ($cmd, $device, $force, $disk_cnt, $disk_name, $name,
        @output, @vols_mount, $status);

    $self->_clear_err();

    $force = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::remove_device()", "Invalid value for argument '-device'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::rename()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            $force = "-k";
            next;
        }
    }

    ### get name of volume group
    $name = $self->get_name();

    ### verify the device is being used in volume group
    unless (exists $self->{devices}->{$device}) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::remove_device()", "Device '$device' doesn't exist in diskgroup");
        return $ClientUtil::FALSE;
    }

    $disk_name = $self->{devices}->{$device};

    ### if no force option is specified umount all volumes so deport
    ### can occur without error.  Make note of which vols were
    ### mounted so they can be remounted after removing device
    if (not defined $force) {
        my @volObjs = $self->get_vols();
        return if ($self->error);

        foreach my $volObj (@volObjs) {

            if ($volObj->is_fs()) {
                if ($volObj->is_mounted()) {
                    $self->sanlog(-msg => "unmounting filesystem on volume " . $volObj->get_name());
                    $volObj->unmount();
                    if ($volObj->error) {
                        $self->_set_err($volObj->errmsg());
                        return $ClientUtil::FALSE;
                    }

                    push @vols_mount, $volObj;
                }
            }
        }
    }

    # Check for the force flag being specified and add it to the
    # rmdisk line
    if (defined($force)) {
        $cmd = "/usr/sbin/vxdg -g $name $force rmdisk $disk_name";
    } else {
        $cmd = "/usr/sbin/vxdg -g $name rmdisk $disk_name";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::remove_device Unable to remove $device from diskgroup \'$name\'", @output);

        return $ClientUtil::FALSE;
    }

    ### remove device from hash
    delete($self->{devices}->{$device});

    ### decrement device counter;
    $self->{disk_cnt}--;

    ### Restart and mount volumes that were previously mounted.
    foreach my $vol (@vols_mount) {

        $vol->_vol_start();
        if ($vol->error) {
            $self->_set_err($vol->errmsg());
            return $ClientUtil::FALSE;
        }

        $vol->mount();
        if ($vol->error) {
            $self->_set_err($vol->errmsg());
            return $ClientUtil::FALSE;
        }

    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rename()
# Objective:       rename volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, %args) = @_;

    my ($cmd, $force, $name, $newname, @output, $status, @vols_mount);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($newname = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::rename()", "Invalid value for argument '-name'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::rename()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }

    }

    ### get name of disk group
    $name = $self->get_name();

    ### verify the device is being used in volume group
    unless (defined $name) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::rename()", "Volume group doesn't have a name");
        return $ClientUtil::FALSE;
    }

    ### verify the device is being used in volume group
    unless (defined $newname) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::rename()", "New name for Volume group not specified");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Rename volume group", "Old Name: $name New Name: " . $newname);

    ### if force option specified umount all volumes so deport
    ### can occur without error.  Make note of which vols were
    ### mounted so they can be remounted after rename occurs
    if (defined $force) {
        my @volObjs = $self->get_vols();
        return if ($self->error);

        foreach my $volObj (@volObjs) {

            if ($volObj->is_fs()) {
                if ($volObj->is_mounted()) {
                    $self->sanlog(-msg => "unmounting filesystem on volume " . $volObj->get_name());
                    $volObj->unmount();
                    if ($volObj->error) {
                        $self->_set_err($volObj->errmsg());
                        return $ClientUtil::FALSE;
                    }

                    push @vols_mount, $volObj;
                }
            }
        }
    }

    ### Deport the diskgroup
    $self->deport();
    return $ClientUtil::FALSE if ($self->error);

    ### Rename the diskgroup
    $cmd = "/usr/sbin/vxdg -n $newname import $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::rename Unable to import '$name' under a new diskgroup \'$newname\'", @output);

        return $ClientUtil::FALSE;
    }

    $self->{name} = $newname;

    $self->{deported} = 0;

    ### Restart and mount volumes that were previously mounted.
    foreach my $vol (@vols_mount) {

        $vol->_vol_start();
        if ($vol->error) {
            $self->_set_err($vol->errmsg());
            return $ClientUtil::FALSE;
        }

        $vol->mount();
        if ($vol->error) {
            $self->_set_err($vol->errmsg());
            return $ClientUtil::FALSE;
        }

    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, @vgs);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::destroy() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/sbin/vxdg destroy $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        ### Ignore the following error message.  This message seems to
        ### be Veritas 4.0 related".
        ### VxVM vxdg ERROR V-5-1-586 Disk group dg2: deport failed: Format error in disk private region
        if (grep(/V\-5\-1\-586/, @output)) {
            $self->sanlog(-msg => "IGNORING Error: HPUX::Vxvm::Vg::destroy() Unable to destroy diskgroup \'$name\'",
                          @output);
        } else {
            # sometimes this is failure is a results of the DG being not active
            # Try a rescan and then do a deport.
            $self->{parent}->rescan();

            $cmd = "vxdg deport $name";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0 and
                grep(/No such disk group is imported/i, @output) == 0) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::destroy() Unable to destroy diskgroup \'$name\'");
                return $ClientUtil::FALSE;
            }
        }
    }

    ### Remove the vgobj from the parent lvm object
    foreach my $vgObj (@{$self->{parent}->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vgs, $vgObj;
        }
    }

    $self->{parent}->{vgs} = \@vgs;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;

    my (@verbose, $cmd, $name, @output, $status, $header, $summary, @summary, $verbose);

    $self->_clear_err();

    $verbose = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Error: HPUX::Vxvm::Vg::display()", "Invalid value for argument '-verbose'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::display() Diskgroup name is a required parameter");
        return (wantarray()) ? () : [];
    }

    push @summary, "VG Name:  " . $self->get_name();
    push @summary, "Device Cnt:  " . $self->get_device_cnt();

    foreach my $device ($self->get_devices()) {
        push @summary, "Device:  " . $device . " (" . $self->{parent}->get_disk_format($device) . ")";

    }

    $header = sprintf("Vol Summary:\n"
                        . "%-12.12s %-12.12s %-7.7s %-5.5s %-12.12s %-32.32s\n"
                        . "%-12.12s %-12.12s %-7.7s %-5.5s %-12.12s %-32.32s",
                      "VGROUP", "VOLUME", "LAYOUT", "TYPE",  "SIZE(mb)", "MOUNT_PT",
                      "=" x 12, "=" x 12, "=" x 7,  "=" x 5, "=" x 12,   "=" x 32);

    push @summary, $header;

    foreach my $volObj (@{$self->{vols}}) {
        $summary = sprintf("%-12.12s %-12.12s %-7.7s %-5.5s %-12.12s %-s",
                           $self->get_name(), $volObj->get_name(), $volObj->get_layout(),
                           $volObj->get_fs(), $volObj->get_size(), $volObj->get_mount());

        push @summary, $summary;
    }

    if (defined $verbose and $verbose =~ /1|on|y|yes|t|true/) {

        $cmd = "/usr/sbin/vxprint -lg $name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: HPUX::Vxvm::Vg::display() Unable to display info for diskgroup \'$name\'", @output);

            return (wantarray()) ? () : [];
        }

        @verbose = (@summary, "\nVgroup Detail:", @output);

        push @verbose, "\n\n";
        return (wantarray()) ? @verbose : \@verbose;
    } else {
        push @summary, "\n\n";
        return (wantarray()) ? @summary : \@summary;
    }
}

################################################################################
# Subroutine Name: get_device_cnt
# Objective:       Return number of devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device_cnt {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{disk_cnt}) {
        return $self->{disk_cnt};
    } else {
        $self->_set_err("Error: HPUX::Vxvm::Vg::get_device_cnt() Unable to display disk count for diskgroup");
        return 0;
    }

}

################################################################################
# Subroutine Name: get_devices
# Objective:       Return devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;

    my (@output);

    $self->_clear_err();

    @output = (keys %{$self->{devices}});

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vg object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: HPUX::Vxvm::Vg::get_name() Vg object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vol
# Objective:       return a vilume object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol {
    my ($self, $name) = @_;

    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::get_vol() volume name is a required argument");
        return {};
    }

    foreach my $obj (@{$self->{vols}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: HPUX::Vxvm::Vg::get_vol() Unable to find volume '$name'");

    return {};

}

################################################################################
# Subroutine Name: get_vols
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vols {
    my ($self) = @_;

    my (@volObjs);

    $self->_clear_err();
    if (defined $self->{vols}) {
        @volObjs = @{$self->{vols}};
    } else {
        $self->_set_err("Error: HPUX::Vxvm::Vg::get_vols() Vg object has no volumes");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @volObjs : \@volObjs;

}

################################################################################
# Subroutine Name: get_vol_cnt
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol_cnt {
    my ($self) = @_;

    $self->_clear_err();
    if (defined $self->{vols}) {
        return scalar(@{$self->{vols}});
    } else {
        $self->_set_err("Error: HPUX::Vxvm::Vg::get_vol_cnt() Vg object has no vol count");
        return 0;
    }

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size, $sectors);

    $self->_clear_err();

    $size    = 0;
    $sectors = 0;

    $name = $self->get_name();

    $cmd = "/usr/sbin/vxprint -Vdg $name | grep dm | awk '{print \$5}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::get_size() Unable to get size (mb) for diskgroup \'$name\'", @output);

        return $size;
    }

    foreach my $sector (@output) {
        $sectors += $sector;
    }

    $size = $sectors * 512;    #bytes
    $size = $size / 1024;      #kilobytes
    $size = $size / 1024;      #megabytes

    return $size;

}

################################################################################
# Subroutine Name: get_freespace
# Objective:       Return available free space
# Details:         See POD documentation at the end of this file
################################################################################
sub get_freespace {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size, $sectors);

    $self->_clear_err();

    $size    = 0;
    $sectors = 0;

    $name = $self->get_name();

    $cmd = "/usr/sbin/vxdg -g $name free | grep disk | awk '{print \$5}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::get_freespace() Unable to get free space (mb) for diskgroup \'$name\'", @output);

        return $size;
    }

    foreach my $sector (@output) {
        $sectors += $sector;
    }

    $size = $sectors * 512;    #bytes
    $size = $size / 1024;      #kilobytes
    $size = $size / 1024;      #megabytes

    return $size;

}

################################################################################
# Subroutine Name: is_deported
# Objective:       Determine if volgroup has been deported
# Details:         See POD documentation at the end of this file
################################################################################
sub is_deported {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{deported}) {
        return $self->{deported};
    } else {
        return 0;
    }
}

################################################################################
# Subroutine Name: deport()
# Objective:       deport Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub deport {
    my ($self) = @_;

    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::deport() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/sbin/vxdg deport $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::deport() Unable to deport diskgroup \'$name\'", @output);

        return $ClientUtil::FALSE;
    }

    ### Remove volume objects from vg cache
    $self->{vols} = ();

    $self->{deported} = 1;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: import()
# Objective:       Import Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub import {
    my ($self, %args) = @_;

    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::import() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/sbin/vxdg import $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::import() Unable to import diskgroup \'$name\'", @output);

        return $ClientUtil::FALSE;
    }

    $self->{deported} = 0;

    ### requery device list
    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    $cmd = "/usr/sbin/vxvol -g $name startall";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::import() Unable to start up volumes in diskgroup \'$name\'", @output);

        return $ClientUtil::FALSE;
    }

    $self->_create_vols(-_import => 1, %args);

    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: refresh
# Objective:       refresh content of volume group.
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    ### Remove existing vol cache
    $self->{vols} = undef;

    $self->sanlog(-msg => "refreshing volume group '" . $self->get_name() . "'");

    ### Determine if dg is deported
    $cmd = "/usr/sbin/vxdisk -o alldgs list | grep '(" . $self->get_name() . ")'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 1) {
        $self->_set_err("Error: HPUX::Vxvm::Vg::refresh()", "Unable to load deported diskgroups");

        return $ClientUtil::FALSE;
    }

    $self->{deported} = 1 if (@output);

    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    if (not $self->is_deported()) {
        $self->_create_vols();
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;
}

1;

