# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the NativeLvm Logical Volume Manager on Linux
## @author knopp@netapp.com,rawat@netapp.com
## @status private
## @pod ../../Lvm.pm

# Package Name
##############
package San::SanClient::Lvm::Linux::NativeLvm;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanClient::Lvm;
use San::SanLog;
use Error;

require San::SanClient::Lvm::Linux::NativeLvm::Vg;

use vars qw(@ISA);

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: initialize
# Objective:       initialize disk under lvm control
# Details:         See POD documentation at the end of this file
################################################################################
sub initialize {
    my ($self, %args) = @_;
    my ($cmd, $devices, $label, $status, @all_free_devices, @format_cmds, @used, @output, @hDevices);

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: NativeLvm::initialize() Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Check that required parameter '-device' was provided
    unless (defined $devices and ref($devices) eq "ARRAY") {
        $self->_set_err("Error: NativeLvm::initialize() ", "Missing required argument '-devices'");
        return $ClientUtil::FALSE;
    }

    # Locate list of all devices that have not been initialized
    $cmd = "lvmdiskscan | egrep '/dev' | egrep -v '/dev/sda\$' | grep 'free' | awk '{print \$3}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: NativeLvm::initialize() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }
    push @all_free_devices, basename($_) foreach (@output);

    foreach my $device (@$devices) {

        # Only initialize devices that need it
        if (grep /^$device$/, @all_free_devices) {

            # formulate the pvcreate command
            $cmd = "pvcreate /dev/$device";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0 ) {
                $self->_set_err("Error NativeLvm::initialize() error while running cmd ($cmd)\n");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_vg
# Objective:       create and return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vg {
    my ($self, %args) = @_;

    my ($vgObj);
    $vgObj = San::SanClient::Lvm::Linux::NativeLvm::Vg->_new(-parent => $self,
                                                             -conn   => $self->{conn},
                                                             %args);

    if ($vgObj->error) {
        $self->_set_err($vgObj->errmsg());
        return $vgObj;
    }

    push @{$self->{vgs}}, $vgObj;

    return $vgObj;
}

################################################################################
# Subroutine Name: destroy_vg
# Objective:       destroy a named diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vg {
    my ($self, $name) = @_;

    my ($vgObj, @vgs,);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: NativeLvm::destroy_vg() vg name is a required argument");
        return $ClientUtil::FALSE;
    }

    # Delete the of vg object and update the parent vg list.
    foreach $vgObj (@{$self->{vgs}}) {
        my $vgname = $vgObj->get_name();
        if ($vgObj->get_name() =~ /^$name$/) {
            $vgObj->destroy(-force => 1);
            if ($vgObj->error) {
                $self->_set_err($vgObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vgs, $vgObj;
        }
    }

    $self->{vgs} = \@vgs;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display
# Objective:       display disk group information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;

    my ($cmd, $status, $verbose, @buffer, @output, @vgObjs,);

    $self->_clear_err();

    $self->sanlog(-msg => "Display diskgroup information");

    @vgObjs = $self->get_vgs();
    return (wantarray()) ? () : [] if ($self->error);

    foreach my $vgObj (@vgObjs) {
        @output = $vgObj->display(%args);
        return (wantarray()) ? () : [] if ($self->error);

        push @buffer, @output[0 .. scalar(@output) - 2],
          "=====================================" . "=====================================";
    }

    return (wantarray()) ? @buffer : \@buffer;
}

###############################################################################
# Method:    get_disk_format
# Objective: This is a noop for linux
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_format {
    my ($self, $disk) = @_;
    my ($format,);

    $self->_clear_err();

    return "";
}

###############################################################################
# Method:    get_disk_formats
# Objective:  This is a noop for linux
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_formats {
    my ($self) = @_;
    my (@format);

    $self->_clear_err();

    @format = ();

    return (wantarray()) ? @format : \@format;
}

###############################################################################
# Method:    get_format_default
# Objective:  This is a noop for linux
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_format_default {
    my ($self) = @_;
    my ($format,);

    $self->_clear_err();

    return "";
}

################################################################################
# Subroutine Name: get_vg
# Objective:       return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vg {
    my ($self, $name) = @_;
    my ($vgObj,);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: NativeLvm::get_vg() vg name is a required argument");
        return {};
    }

    foreach my $obj (@{$self->{vgs}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: NativeLvm::get_vg() Unable to find diskgroup '$name'");
    return {};
}

################################################################################
# Subroutine Name: get_vgs
# Objective:       return array of volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgs {
    my ($self) = @_;
    my (@vgs,);

    $self->_clear_err();

    if (defined $self->{vgs}) {
        @vgs = @{$self->{vgs}};
    } else {
        $self->_set_err("Error: NativeLvm::get_vgs() Vxvm object has no volumes groups");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @vgs : \@vgs;

}

################################################################################
# Subroutine Name: get_vgnames
# Objective:       return array of volume group names
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgnames {
    my ($self) = @_;
    my ($cmd, $status, @output, @vgs);

    $self->_clear_err();

    $cmd = "vgdisplay";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error NativeLvm::get_vgnames() error while running cmd ($cmd)\n" .
                        "Manual cleanup maybe required, make sure all lvs/vgs/pvs have\n" .
                        "been removed. If you can't remove these items, a reboot of the\n" .
                        "system will be required");
        return (wantarray()) ? () : [];
    }

    (@output) = grep (/VG\s+Name/,@output);
    foreach my $line (@output) {
        my @vg_info = split(/\s+/,$line);
        push @vgs,$vg_info[2];
    }

    return (wantarray()) ? @vgs : \@vgs;

}

################################################################################
# Subroutine Name: load_vg
# Objective:       load an existing diskgroup that is not in memory
# Details:         See POD documentation at the end of this file
################################################################################
sub load_vg {
    my ($self, $name) = @_;
    my ($vgObj,);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: NativeLvm::load_vg() Invalid or missing required paramter \$name");
        return {};
    }

    # Check if vg has already been loaded
    $vgObj = $self->get_vg($name);
    if ($self->error) {

        # dg has not been load yet
        $self->_clear_err();

        # load dg
        $self->_create_vgs($name);
        return {} if ($self->error);
        $vgObj = $self->get_vg($name);
    } else {
        $self->sanlog(-msg => "INFO: NativeLvm::load_vg() Volume group '$name' was already loaded");
    }

    return $vgObj;

}
################################################################################
# Subroutine Name: unload_vg
# Objective:       Remove the vg object under LVM object control
# Details:         See POD documentation at the end of this file
################################################################################
sub unload_vg {
    my ($self, $name) = @_;
    my ($vgObj, $bFound, @vgs,);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: NativeLvm::unload_vg() Invalid or missing required paramter \$name");
        return $ClientUtil::FALSE;
    }

    $bFound = 0;

    foreach $vgObj (@{$self->{vgs}}) {
        ($vgObj->get_name() =~ /^$name$/)
          ? $bFound = 1
          : push @vgs, $vgObj;
    }

    $self->{vgs} = \@vgs;

    unless ($bFound) {
        $self->_set_err("Error: NativeLvm::unload_vg() Nothing to unload. Vg name '" . $name . "' was not loaded");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}
################################################################################
# Subroutine Name: get_pkginfo()
# Objective:       get detailed lvm pkg information
# Details:         See POD documentation at the end of this file
################################################################################
sub get_pkginfo {
    my ($self) = @_;
    my ($cmd, @output, $status,);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{pkginfo});

    return (wantarray()) ? @{$self->{pkginfo}} : $self->{pkginfo};

}

################################################################################
# Subroutine Name: _get_pkginfo
# Objective:       Get lvm package info
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self) = @_;
    my ($cmd, $name, $vers, $rev, @output, $status,);

    $cmd = "rpm -q lvm";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # determine if command worked
    if ($status > 0) {
        $self->_set_err("Error: NativeLvm::get_version() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    ($vers, $rev) = $output[0] =~ /\-(\S+)\-(\S+)/;

    $self->{name}    = "lvm";
    $self->{vers}    = $vers;
    $self->{rev}     = $rev;
    $self->{pkginfo} = \@output;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_revision()
# Objective:       Get revision of NativeLvm
# Details:         See POD documentation at the end of this file
################################################################################
sub get_revision {
    my ($self) = @_;
    my ($cmd, @output, $status,);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{rev});
    return "" if ($self->error());

    return $self->{rev};

}

################################################################################
# Subroutine Name: get_name()
# Objective:       Get lvm name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;
    my ($cmd, @output, $status,);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{name});
    return "" if ($self->error());

    return $self->{name};

}

################################################################################
# Subroutine Name: get_version()
# Objective:       Check version of Veritas
# Details:         See POD documentation at the end of this file
################################################################################
sub get_version {
    my ($self) = @_;
    my ($cmd, @output, $status, $version,);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{vers});
    return "" if ($self->error());

    return $self->{vers};

}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;
    my ($bCnt,    $cmd,     $cnt,         $device, $filer, $filername, $lun,
        $refresh, $status,  $total,       $type,   $vol,   %hDevs,     @all_devices,
        @attr,    @devices, @lvm_devices, @output, @used,  $protocol, $filers);

    $type  = ".*";
    $total = 0;
    @{$filers} = ();

    $self->_clear_err();

    # Asssume user supplied -cnt argument
    $bCnt    = 1;
    $refresh = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Error: NativeLvm::get_free_devices() Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: NativeLvm::get_free_devices() Invalid value for argument '-type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Error: NativeLvm::get_free_devices() Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/^-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Error: NativeLvm::get_free_devices() Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Error: NativeLvm::get_free_devices() Invalid value for argument '-vol'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Error: NativeLvm::get_free_devices() Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: NativeLvm::get_free_devices() Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: NativeLvm::get_free_devices() Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Check if user wants free devices from a specified filer,volume, or lun
    if (defined $filer or defined $vol or defined $lun or defined $protocol or defined $filers) {
        $self->sanlog(-msg => "\nFilter options defined:\n\tFiler: $filer\n\tVolume: $vol\n\tLun: $lun\n\tProtocol: $protocol",
                      -msg_level => 3);
        $refresh = 1;

        # Check if a device to filer map 'dfmap' has already been created or
        # refresh it if requested.
        if (not defined $self->{dfmap} or $refresh == 1) {
            @output = $self->{sanlun}->lun_show();
            if ($self->{sanlun}->error) {
                $self->_set_err($self->{sanlun}->errmsg());
                return (wantarray()) ? () : [];
            }

            foreach my $line (@output) {
                my $dev_tmp = "";

                # skip first line (header)
                if ($line =~ /lun-pathname/i) {
                    next;
                }

                # remove leading white spaces if any
                $line =~ s/^\s*//;

                my ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp);
                if ($line  =~ /(\S+)\s+\/vol\/(\w+)\/(\S+)\s+(\S+)\s+\S+\s+(\S+)\s+/) {
                    # old sanlun kit output looks like below,
                    #    controller:          lun-pathname        device filename  adapter  protocol          lun size         lun state
                    # mml_sanmgr_vserv:  /vol/mml_vol_2/mml_lun_0  /dev/sda         host37   iSCSI          3g (3221225472)     GOOD
                    #
                    # new sanlun kit output looks like below,
                    # controller(7mode)/                                                 device          host                  lun
                    # vserver(Cmode)       lun-pathname                                  filename        adapter    protocol   size    mode
                    # ---------------------------------------------------------------------------------------------------------------------
                    # fas3170-201-76       /vol/nate_1_1_1/nate_IBMblade-200-218_linux_0 /dev/sde        host22     iSCSI      4g      7

                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp) = ($1, $2, $3, $4, lc($5));
                } elsif ($line =~ /(\S+)\s+\/vol\/(\w+)\/(\S+)\s+(\S+)\s+/) {
                    ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp) = ($1, $2, $3, $4);
                }

                # Remove colon in the end of filer/vserver name if presents
                $filer_tmp =~ s/:$//;

                next unless (defined $vol_tmp and length($vol_tmp));    # skip orphaned luns

                if ($self->{parent}->{mp_type} =~ /dmmp/i) {
                    $cmd = "multipath -l -d $device_tmp |grep 360";

                    my ($status, $output) = &ClientUtil::execute($self->{conn}, $cmd);
                    if ($status > 0) {
                        $self->_set_err(
                                 "Error: NativeLvm::get_free_devices() Command ($cmd) returned a bad status ($status)");
                        return (wantarray()) ? () : [];
                    }

                    #make sure We are not doing same work for the same device again.
                    next if ($dev_tmp =~ /$output/);

                    # Trying to parse the Lun ID from the cmd output
                    my @lun_id = split(//, $output);
                    my $i = 0;

                    #Lun ID start with 3
                    while ($lun_id[$i] != 3) {
                        $i++;
                    }
                    my @dev;

                    # Lun ID is 33 chars long
                    for (my $j = 0; $j < 33; $j++) {
                        $dev[$j] = $lun_id[$i++];
                    }

                    $device_tmp = join("", @dev);

                }

                @attr = split(/\s+/, $line);
                $dev_tmp = basename($device_tmp);

                $hDevs{$dev_tmp} = {filer    => $filer_tmp,
                                    vol      => $vol_tmp,
                                    lun      => $lun_tmp,
                                    protocol => $protocol_tmp};

                $self->sanlog(-msg => "Device ($dev_tmp) information:\n\tFiler: $filer_tmp\n\tVolume: $vol_tmp\n" .
                                      "\tLUN: $lun_tmp\n\tProtocol: $protocol_tmp\n",
                              -msg_level => 5);
            }

            # save map for next call of function
            $self->{dfmap} = \%hDevs;
        }
    }

    # If not -cnt then assume all free disks
    # under lvm control are wanted
    unless (defined $cnt and length($cnt)) {
        $cnt  = 10000;
        $bCnt = 0;
    }

    @lvm_devices = $self->{parent}->_get_unlabeled_devices();
    if ($self->{parent}->error) {
        $self->_set_err($self->{parent}->errmsg());
    }

    # iterate through the devices and make sure that none of the disks
    # are being used other LVM disk object.
    foreach my $device (@lvm_devices) {
        my $device_not_free = 0;
        if (defined $filer and length $filer) {
            $self->sanlog(-msg => "\nFilter filer: $filer\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i);
        }

        if (defined $vol and length $vol) {
            $self->sanlog(-msg => "\nFilter volume: $vol\nDevice: $device\nDevice volume: " . $self->{dfmap}{$device}{vol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i);
        }

        if (defined $lun and length $lun) {
            $self->sanlog(-msg => "\nFilter lun: $lun\nDevice: $device\nDevice lun: " . $self->{dfmap}{$device}{lun},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{lun} !~ /$lun/i);
        }

        if (defined $protocol and length $protocol) {
            $self->sanlog(-msg => "\nFilter protocol: $protocol\nDevice: $device\nDevice protocol: " . $self->{dfmap}{$device}{protocol},
                          -msg_level => 3);
            next if ($self->{dfmap}{$device}{protocol} !~ /$protocol/i);
        }

        # Filter on a filer list if provided by caller
        if (defined $filers and scalar(@{$filers}) != 0) {
            my $found = 0;
            # Look through each filer in the filers list
            foreach my $filer_f (@{$filers}) {
                $self->sanlog(-msg => "\nFilter filer(s): $filer_f\nDevice: $device\nDevice filer: " . $self->{dfmap}{$device}{filer},
                              -msg_level => 3);
                # If filer matches devices filer, set found to 1
                if ($self->{dfmap}{$device}{filer} =~ /^$filer_f$/i) {
                    $found = 1;
                }
            }

            # If the device is not part of the filers list caller requested skip it.
            if ($found == 0) {
                next;
            }
        }

        if ($self->{parent}->{mp_type} =~ /dmmp/i) {
            #########################################################################
            # We need to remove devices which have partitions, like Root LUN etc
            # we get the follwoing outupt if we have partitions on device
            #
            # fdisk -l /dev/mapper/mpath0
            # Disk /dev/mapper/mpath0: 18.2 GB, 18254659584 bytes
            # 255 heads, 63 sectors/track, 2219 cylinders
            # Units = cylinders of 16065 * 512 = 8225280 bytes
            #
            # Device Boot      Start         End      Blocks   Id  System
            # /dev/mapper/mpath0p1   *           1          13      104391   83  Linux
            # /dev/mapper/mpath0p2              14        2219    17719695   8e  Linux LVM
            #
            # If device is raw then
            # fdisk -l /dev/mapper/mpath7
            # Disk /dev/mapper/mpath7: 5368 MB, 5368709120 bytes
            # 255 heads, 63 sectors/track, 652 cylinders
            # Units = cylinders of 16065 * 512 = 8225280 bytes
            #
            # Disk /dev/mapper/mpath7 doesn't contain a valid partition table
            #########################################################################
            $cmd = "fdisk -l /dev/mapper/$device";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error: NativeLvm::get_free_devices() $cmd failed");
                return  (wantarray()) ? () : [];
            }
            if (grep(/Device\s+Boot\s+Start/i, @output)) {
                $device_not_free = 1;
            }
            if ($device_not_free != 1) {
                push @devices, "/dev/mapper/" . $device if (++$total <= $cnt);
            }
        } else {
            $self->sanlog(-msg => "Adding device (/dev/$device) to devices list",
                          -msg_level => 0);
            push @devices, "/dev/" . $device if (++$total <= $cnt);
        }
    }

    # If user supplied cnt and free devices fell short
    # generate error
    if ($bCnt eq 1 and scalar(@devices) < $cnt) {
        $self->_set_err("Error: NativeLvm::get_free_devices() ",
                        "Only " . scalar(@devices) . " of $cnt requested devices were found");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: enable_controllers
# Objective:       This is a noop for Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub enable_controllers {
    my ($self) = @_;

    my ($cmd, @output, $status);
    my ($ctlr, @controllers);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rescan
# Objective:       This is a noop for Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;
    my ($cmd, @output, $status,);

    $self->_clear_err();

    # This is a noop of sanpath on aix

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: restart_recovery_daemon
# Objective:       This is a noop for Linux
# Details:         See POD documentation at the end of this file
################################################################################
sub restart_recovery_daemon {
    my ($self, $interval) = @_;

    my ($cmd, @output, $status);

    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($conn, $log, $loadcfg, $lvm, $parent, $hosttype, $self);
    my ($cmd, $status, @output);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_})) {
                $self->_set_err("Error: NativeLvm::_new() Invalid value for argument '-lvm'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: NativeLvm::_new() Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?loadcfg/) {
            if (not length($loadcfg = $args{$_})
                or $loadcfg !~ /no|yes|1|0|off|on/i) {
                $loadcfg = 1;
            }
            next;
        }
    }

    # Allow CLI commands to be logged;
    $self->{log} = $log if (defined $log);

    # check for existance of required -lvm argument
    if (grep (/^-lvm$/, (keys %args))) {
        unless ($args{-lvm} =~ /NativeLvm/i) {
            $self->_set_err("Error: NativeLvm::_new()", "Invalid value for argument '-lvm' \(" . $args{-lvm} . "\)");
            return $self;
        }
    } else {
        $self->_set_err("Error: NativeLvm::_new()", "Missing required argument'-lvm'");
        return $self;
    }

    # check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ "SanClient") {
            $self->_set_err("Error: NativeLvm::_new()", "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: NativeLvm::_new()", "Missing required argument '-parent'");
        return $self;
    }

    # Set debug level on self
    $self->set_debug_level($parent->get_debug_level());

    # initilize the object members
    $self->{conn}   = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();

    $self->{parent} = $parent;

    $self->{vgs} = [];

    # check to make sure that LVM is installed
    if (not $self->_is_installed()) {
        return $self if ($self->error);
    }

    # We are getting 'Found duplicate PV' while running any LVs related cmd
    # by changing filters in /etc/lvm/lvm.conf file we can get rid of these messages
    if ($parent->{major_revision} =~ /sles9/ or $parent->{major_revision} =~ /sles10/) {
        $self->_modify_lvm_filter();
        if ($self->error) {
            return $self;
        }

        # Need to run following commands if lvm filter is modified
        $cmd = "pvscan";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to run command $cmd");
            return $self;
        }
        $cmd = "vgscan";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to run command $cmd");
            return $self;
        }
    }

    # create sanpath vg objects for existing disk groups
    $self->_create_vgs() if ($loadcfg);
    return $self         if ($self->error);

    return $self;

}

################################################################################
# Subroutine Name: _modify_lvm_filter
# Objective:       configure /etc/lvm/lvm.conf file
# Details:         See POD documentation at the end of this file
################################################################################
sub _modify_lvm_filter {
    my ($self) = @_;
    my ($filter_line, $cmd, $no_of_matches, $status, @output, @output_tmp);

    $self->_clear_err();

    my $client_obj = $self->{parent};

    $cmd = "rm -f /etc/lvm/lvm.conf.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to remove /etc/lvm/lvm.conf.tmp");
        return $ClientUtil::FALSE;
    }

    $cmd = "touch /etc/lvm/lvm.conf.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to create /etc/lvm/lvm.conf.tmp");
        return $ClientUtil::FALSE;
    }

    $cmd = "cat /etc/lvm/lvm.conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to cat /etc/lvm/lvm.conf");
        return $ClientUtil::FALSE;
    }

    # A filter that tells LVM2 to only use a restricted set of devices
    # If we dont reject the /dev/sd* devices then in case of SLES we get
    # "Found duplicate PV" for all LVM related cmds that we execute. For more info read
    # /etc/lvm/lvm.conf

    my $filter_is_not_proper = 0;
    my @all_filters = grep (/^\s*filter\s*=\s*(.*)\s*]$/i, @output);
    my ($filters) = ($all_filters[-1] =~ /^\s*filter\s*=\s*\[\s*(.*)\s*\]$/i);
    # Verify that filter is proper
    # If there are multiple filter lines use the last one
    my @filters = split (/\,\s*/, $filters);

    # make sure that accept all devices setting is the last piece of the filter
    my $accept_all_dev = $filters[-1];
    if ($accept_all_dev ne "\"a/.*/\" " and $accept_all_dev ne "\"a/.*/\"") {
        $filter_is_not_proper = 1;
    }

    #
    # Don't do anything if proper filter is already in place
    #

    # Checking for by-id filter
    my $by_id_filter = "\"r|\/dev\/.*\/by-id\/.*|\"";
    if (not grep(/$by_id_filter/, @filters)){
        $filter_is_not_proper = 1;
    }
    # Checking for by-path filter
    my $by_path_filter = "\"r|\/dev\/.*\/by-path\/.*|\"";
    if (not grep(/$by_path_filter/, @filters)){
        $filter_is_not_proper = 1;
    }
    # Checking for /dev/disk filter
    my $disk_filter = "\"r|\/dev\/disk\/*|\"";
    if (not grep(/$disk_filter/, @filters)){
        $filter_is_not_proper = 1;
    }
    # Checking for /dev/block filter
    my $block_filter = "\"r|\/dev\/block\/*|\"";
    if (not grep(/$block_filter/, @filters)){
        $filter_is_not_proper = 1;
    }

    # check for preferred_names settings in /etc/lvm/lvm.conf
    my $preferred_names_is_not_proper = 0;

    my @preferred_names_list = grep (/^\s*preferred_names\s*=\s*.*]$/i, @output);
    my ($preferred_names) = ($preferred_names_list[-1] =~ /^\s*preferred_names\s*=\s*\[\s*(.*)\s*\]$/i);


    # Verify that filter is proper
    # If there are multiple preferred_names lines use the last one
    my @preferred_names = split(/\,\s*/, $preferred_names);

    # Checking for preferred_name /dev/mapper/mpath
    my $preferred_name_not_exists = 1;
    foreach my $preferred_name (@preferred_names) {
        if ($preferred_name eq "\"^/dev/mapper/mpath\"" or $preferred_name eq "\"^/dev/mapper/mpath\" ") {
            $preferred_name_not_exists = 0;
        }
    }

    # Checking for preferred_name /dev/hd
    foreach my $preferred_name (@preferred_names) {
        if ($preferred_name eq "\"^/dev/hd\" " or $preferred_name eq "\"^/dev/hd\"") {
            $preferred_name_not_exists = 0;
        }
    }
    if ($preferred_name_not_exists == 1) {
        $preferred_names_is_not_proper = 1;
    }

    # make sure it has only 2 elements
    if (scalar(@preferred_names) != 2) {
        $preferred_names_is_not_proper = 1;
    }

    # Calling this to set value of $client_obj->{sanbooted}
    my $mp_obj = $client_obj->{mp};
    if (! defined($mp_obj->{root_mp_devices})) {
        $client_obj->get_root_dev();
    }

    #  if sanbooted make sure /dev/sd* filter is there
    if ($client_obj->{sanbooted} == 1) {
        my $block_filter = "\"r|\/dev\/sd.*\/|\"";
        # Checking for /dev/sd filter
        if (not grep(/$block_filter/, @filters)){
            $filter_is_not_proper = 1;
        }
        if (scalar(@filters) != 6) {
            $filter_is_not_proper = 1;
        }
    } else {
        if (scalar(@filters) != 5) {
            $filter_is_not_proper = 1;
        }
    }

    # Modify lvm.conf file if one of them is not proper
    if ($filter_is_not_proper == 1 or $preferred_names_is_not_proper == 1) {
        # Setting the filter
        if (defined $client_obj->{sanbooted} == 0) {
            $filter_line = "\ filter =\ \[\ \"r\|\/dev\/\.\*\/by-path\/\.\*\|\"" . "\,\ \"r\|\/dev\/\.\*\/by-id\/\.\*\|\"" . "\,\ \"r\|\/dev\/block\/\.\*\|\"" .  "\,\ \"r\|\/dev\/disk\/\.\*\|\"" . "\,\ \"a\/\.\*\/\"\ \]";
        } else {
            $filter_line = "\ filter =\ \[ \ \"r\|\/dev\/sd\.\*\|\"" . "\,\ \"r\|\/dev\/\.\*\/by-path\/\.\*\|\"" . "\,\ \"r\|\/dev\/\.\*\/by-id\/\.\*\|\"" . "\,\ \"r\|\/dev\/block\/\.\*\|\"" .  "\,\ \"r\|\/dev\/disk\/\.\*\|\"" . "\,\ \"a\/\.\*\/\"\ \]";
        }

        my $preferred_names_line = "\ preferred_names\ =\ \[\ \"\^\/dev\/mapper\/mpath\"\,\ \"\^\/dev\/hd\"\ \]";
        my $filter_changes_made = 0;
        my $preferred_names_changes_made = 0;

        foreach my $line (@output) {
            if (($filter_changes_made == 0) and ($line =~ /^\s*filter\s*=\s*.*]$/i)) {
                $line = $filter_line;
                $filter_changes_made = 1;
            } elsif (($preferred_names_changes_made == 0) and ($line =~ /^\s*preferred_names\s*=\s*.*]$/i)) {
                $line = $preferred_names_line;
                $preferred_names_changes_made = 1;
            }

            $no_of_matches = ($line =~ /\'/g);
            if ($no_of_matches > 1){
                $line =~ s/\'/\"/g;
            } else {
                $line =~ s/\'//g;
            }

            $cmd = "echo \'$line\' >> /etc/lvm/lvm.conf.tmp";
            ($status, @output_tmp) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Unable to append $line to /etc/lvm/lvm.conf file");
                return $ClientUtil::FALSE;
            }
        }

        # If filter is not there append the filter at the end
        if ($filter_changes_made == 0) {
            $cmd = "echo \'$filter_line\' >> /etc/lvm/lvm.conf.tmp";
            ($status, @output_tmp) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Unable to append $filter_line to /etc/lvm/lvm.conf file");
                return $ClientUtil::FALSE;
            }
        }

        # If preferred_names settings is not there append preferred_names line at the end
        if ($preferred_names_changes_made == 0) {
            $cmd = "echo \'$preferred_names_line\' >> /etc/lvm/lvm.conf.tmp";
            ($status, @output_tmp) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Unable to append $preferred_names_line to /etc/lvm/lvm.conf file");
                return $ClientUtil::FALSE;
            }
        }

        # -b is to take backup of destination file
        $cmd = "mv -b /etc/lvm/lvm.conf.tmp /etc/lvm/lvm.conf";
        ($status, @output) = &ClientUtil::execute($self->{conn} , $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to overide /etc/lvm/lvm.conf file");
            return $ClientUtil::FALSE;
        }

        # Recreating initrd to reflect the changes in initramsf
        $cmd = "uname -r";
        ($status, @output) = &ClientUtil::execute($self->{conn} , $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to get kernel version");
            return $ClientUtil::FALSE;
        }
        my $kernel_ver = $output[0];
        $cmd = "mkinitrd";
        if ($client_obj->{sanbooted} == 1) {
            if ($client_obj->{major_revision} =~ /sles10/ and $client_obj->{major_revision} >= 2) {
                $cmd .= " -f mpath";
            } elsif($client_obj->{major_revision} =~ /sles11/) {
                $cmd .= " -f multipath";
            }
        }
        ($status, @output) = &ClientUtil::execute($self->{conn} , $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to recreate initrd");
            return $ClientUtil::FALSE;
        }

    } else {
        $self->sanlog(-msg => "/etc/lvm/lvm.conf is setup properly. No need to make any changes");
    }
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _create_vgs
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vgs {
    my ($self, $vgname) = @_;
    my ($cmd, $deported, $name, @output, @volObjs, $status, $vgObj, %hDeport,);
    my ($vgdisplay_cmd, @vg_output);
    my (@root_dev_arr, @root_lv, $root_vg, $root_dev, $line, @mount_output, @lvs_output, $index);

    $deported = 0;

    # Create objects foreach disk group (including deported/inactive)
    # Here is the output of the following cmd from a system on which
    # one vg is exported
    #----------------cmd output-------------#
    #  VolGroup00:r/w:772:-1:0:2:2:-1:0:1:1:71434240:32768:2180:2178:2:oBzRO1-4GJt-M6DP-hdnv-IqiX-o7pd-4jkiyf
    #  Volume group n_vg is exported
    #  n_vg:r/w:774:-1:0:5:0:-1:0:4:4:20955136:4096:5116:1280:3836:BdTc4u-af07-JTwt-Kdy8-de5A-KQrA-iRXzQJ

    if (($self->{parent}->{major_revision} =~ /rhel5/ and $self->{parent}->{minor_revision} > 2)
       or ($self->{parent}->{major_revision} =~ /ol5/ and $self->{parent}->{minor_revision} >= 6)
       or $self->{parent}->{major_revision} =~ /rhel6|rhel7|ol6|sles11|fedora/) {
        $vgdisplay_cmd = "vgdisplay";
    } else {
        $vgdisplay_cmd = "vgdisplay -D";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $vgdisplay_cmd);
    if ($status > 0) {
        # vgdisplay either hanged or timed out. Rebooting the host
        # so that vgdisplay cmd would pass
        $self->sanlog(-msg => "Vgdisplay failed. Rebooting the host");
        $self->{parent}->reboot();
        if ($self->{parent}->error) {
            $self->_set_err($self->{parent}->errmsg());
            return $ClientUtil::FALSE;
        }

        ($status, @output) = &ClientUtil::execute($self->{conn}, $vgdisplay_cmd);
        if ($status > 0) {
            # if still failing then fail out
            $self->_set_err("Error while running cmd ($vgdisplay_cmd).\n The host has already been " .
                            "rebooted once, but that didn't allow vgdisplay to run properly\n" .
                            "Manual cleanup maybe required, make sure all lvs/vgs/pvs are removed.");
            return $ClientUtil::FALSE;
        }
    }

    if (defined($vgname) and length($vgname)) {
        @output = grep(/$vgname/ , @output);
    }

    # Get the root device from mount command
    $cmd = "mount";
    ($status, @mount_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Command ($cmd) returned a bad status ($status) ");
        return $ClientUtil::FALSE;
    }

    my ($root_line) = grep (/\s+\/\s+/, @mount_output);
    $root_line =~ s/^\s+//;    # strip leading white space
    @root_dev_arr = split (/\s+/, $root_line);
    $root_dev = $root_dev_arr[0];

    # We need to skip the VG, if root is on LV.
    # lvs will return corresponding VG name if root device is on LV, else it will return error status.We can skip error status
    # since we need only VG name if root is on LV.
    $cmd = "lvs $root_dev";
    ($status, @lvs_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 0) {
        $index = 0;
        foreach my $line (@lvs_output) {
            $index = $index + 1;
            if ($line =~ /\s*LV\s+/ ) {
                last;
            }
        }
        $line = $lvs_output[$index];
        $line =~ s/^\s+//;
        @root_lv = split (/\s+/, $line);
        $root_vg = $root_lv[1];
    }

    @output = grep(/VG\s+Name/i,@output);

    # iterate though dg names found
    foreach my $name (@output) {

        # first field is vg name and 3rd field is its state.
        my @vg_info_arr = split(/\s+/ , $name);

        my $vgname = $vg_info_arr[3];

        next if ($vgname eq $root_vg);

        $cmd = "mount |grep $vgname";
        my ($status, @mount_output) = &ClientUtil::execute($self->{conn}, $cmd);
        if (@mount_output) {
            my @mntpnt = split(/\s+/ , $mount_output[0]);
            next if ($mntpnt[2] eq "/");
        }

        # If vg is exported then vg status will be exported
        ($status, @vg_output) = &ClientUtil::execute($self->{conn}, $vgdisplay_cmd);
        if ($status > 0) {
            $self->_set_err("Error NativeLvm::_create_vgs() error while running cmd ($vgdisplay_cmd)\n" .
                            "Manual cleanup maybe required, make sure all lvs/vgs/pvs have" .
                            "been removed\n If you can't remove these items, a reboot of the\n" .
                            "system is required");
        }
        @vg_output = grep(/VG\s+Status/i,@vg_output);
        if ($vg_output[0] =~ /exported/i) {
            $deported = 1;
        }

        $vgObj = $self->create_vg(-name => $vgname, -deported => $deported);
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _is_installed
# Objective:       Check to make sure that the LVM is installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_installed {
    my ($self) = @_;
    my ($cmd, @output, $status,);

    $self->_clear_err();

    $cmd = "ls -l /proc/lvm";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 0) {

        #
        # Checking for /proc/lvm will work for redhat but not SuSE
        #
        $self->sanlog(-msg => "lvm is installed");
        return $ClientUtil::TRUE;
    } else {

        #
        # Checking for /etc/lvm will work for SuSE
        #
        $cmd = "ls -l /etc/lvm";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        if ($status == 0) {

            #
            # Checking for /proc/lvm will work for redhat but not SuSE
            #
            $self->sanlog(-msg => "lvm is installed");
            return $ClientUtil::TRUE;
        } else {
            $self->sanlog(-msg => "lvm is not installed");
            $self->_set_err("Error: Command ($cmd) returned a bad status ($status) " . "lvm is not installed");
            return $ClientUtil::FALSE;
        }
    }
}

1;
