# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the NativeLvm Volume Groups on Linux
## @author knopp@netapp.com , rawat@netapp.com
## @status private
## @pod ../../Vg.pod

package San::SanClient::Lvm::Linux::NativeLvm::Vg;

use strict;

use File::Basename;

use Tharn;
use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

require San::SanClient::Lvm::Linux::NativeLvm::Vg::Vol;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# Subroutine Name: add_device()
# Objective:       add a device to the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub add_device {
    my ($self, $device) = @_;

    my ($cmd, $disk_cnt, $name, @output, $status, $value);

    $self->_clear_err();

    # Get name of volume group.
    $name = $self->get_name();

    # Increment device counter
    $disk_cnt = $self->{disk_cnt};
    $disk_cnt++;

    unless ($device =~ /\/dev/i) {
        if ($self->{parent}->{parent}->{mp_type} =~ /dmmp/i){
            $device = "/dev/mapper/$device";
        } else {
            $device = "/dev/$device";
        }
    }

    $cmd = "vgextend $name " . $device;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error NativeLvm::Vg::add_device() error while running cmd ($cmd)\n");
        return $ClientUtil::FALSE;
    }

    if ($self->{parent}->{parent}->{mp_type} =~ /dmmp/i){
        ($value) = $device =~ /\/dev\/mapper\/(\S+)/i;
    } else {
        ($value) = $device =~ /\/dev\/(\S+)/i;
    }
    $self->{devices}->{$device} = $value;
    $self->{disk_cnt} = $disk_cnt;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_vol
# Objective:       create and return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vol {
    my ($self, %args) = @_;
    my ($volObj);

    # Change layout to concat if dg contains only 1 device
    if (exists $args{-layout} and length $args{-layout}) {
        if ($args{-layout} =~ /stripe/i and $self->get_device_cnt() == 1) {
            $args{-layout} = "concat";
        }
    }

    $volObj = San::SanClient::Lvm::Linux::NativeLvm::Vg::Vol->_new(-parent => $self,
                                                                   -conn   => $self->{conn},
                                                                   %args);

    if ($volObj->error) {
        $self->_set_err($volObj->errmsg());
        return $volObj;
    }

    push @{$self->{vols}}, $volObj;

    return $volObj;
}

################################################################################
# Subroutine Name: deport()
# Objective:       deport Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub deport {
    my ($self) = @_;
    my ($cmd, @cmds, $name, @output, @volObjs, $status);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vg::deport() Diskgroup type cannot be determined");
        return $ClientUtil::FALSE;
    }

    @volObjs = $self->get_vols();

    foreach my $volObj (@volObjs) {

        # Unmount volumes containing filesystems
        if ($volObj->is_mounted()) {
            if ($volObj->error) {
                $self->_set_err($volObj->errmsg());
                return $ClientUtil::FALSE;
            }
            $volObj->unmount();
            if ($volObj->error) {
                $self->_set_err($volObj->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    }

    push @cmds, "vgchange -a n $name", "vgexport $name";

    foreach $cmd (@cmds) {
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error NativeLvm::Vg::deport() error while running cmd ($cmd)\n");
            return $ClientUtil::FALSE;
        }
    }

    # Remove volume objects from vg cache
    $self->{vols}     = ();
    $self->{deported} = 1;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy_vol
# Objective:       destroy a named vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vol {
    my ($self, $name) = @_;
    my ($volObj, @vols);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Vg::destroy_vol() volname is a required argument");
        return $ClientUtil::FALSE;
    }

    # Delete the of vg object and update the parent vg list.
    foreach $volObj (@{$self->{vols}}) {
        if ($volObj->get_name() =~ /^$name$/) {
            $volObj->destroy(-force => 1);
            if ($volObj->error) {
                $self->_set_err($volObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vols, $volObj;
        }
    }

    $self->{vols} = \@vols;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self, %args) = @_;

    my ($cmd, $force, $name, @cmds, @devices, @output, $status, @vgs, @vols, $clientObj);

    $self->_clear_err();

    $name = $self->get_name();
    $clientObj = $self->{parent}->{parent};

    # parse arguments
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Vg::destroy()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    unless (defined $name) {
        $self->_set_err("Error: Vg::destroy() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    # Check if VG is exported. If it is then import if otherwise continue
    if ($self->{deported} == 1) {
        $self->import();
    }

    # If force specified then remove existing vol objects.
    if (defined $force) {

        my @volObjs = $self->get_vols();
        return $ClientUtil::FALSE if ($self->error);

        foreach my $volObj (@volObjs) {
            my $volname = $volObj->get_name();
            
            # Unmount volumes containing filesystems
            if ($volObj->is_mounted()) {
                if ($volObj->error) {
                    $self->_set_err($volObj->errmsg());
                    return $ClientUtil::FALSE;
                }
                $volObj->unmount();
                if ($volObj->error) {
                    $self->_set_err($volObj->errmsg());
                    return $ClientUtil::FALSE;
                }
            }
            push @cmds, "lvremove -f /dev/" . $self->get_name() . "/" . $volObj->get_name();
        }
    }

    # Push VolGroup for the removal
    push @cmds, "vgchange -a n $name", "yes | vgremove $name";

    @devices = $self->get_devices();
    return $ClientUtil::FALSE if ($self->error);

    # Remove all the PV from the devics
    foreach my $device (@devices){
        $self->sanlog(-msg=>"destroy: devices $device going to be removed");
        push @cmds, "pvremove -f $device" ;
    }

    foreach $cmd (@cmds) {
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error NativeLvm::Vg::destroy() error while running cmd ($cmd)");
            return $ClientUtil::FALSE;
        }
    }

    # Remove the vgobj from the parent lvm object
    foreach my $vgObj (@{$self->{parent}->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vgs, $vgObj;
        }
    }

    # On sles9 sp4 host lvremove doesnt remove the entries from /dev. Next time when 
    # Config try to create VG (with the same name) it fails because stale entry is already there
    if ($clientObj->{major_revision} =~ /sles9/i) {
        $cmd = "ls /dev/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            $cmd = "rm -rf /dev/$name";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Error NativeLvm::Vg::destroy() failed to delete VG/LV's stale " .
                                "entries from /dev");
                return $ClientUtil::FALSE;
            }
        }

    }
    
    $self->{parent}->{vgs} = \@vgs;
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;
    my ($cmd, $header, $name, $status, $summary, $verbose, @output, @real_volnames, @summary, @verbose);

    $self->_clear_err();

    $verbose = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Error: Vg::display()", "Invalid value for argument '-verbose'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    $name = $self->{name};

    unless (defined $name) {
        $self->_set_err("Error: Vg::display() Diskgroup name is a required parameter");
        return (wantarray()) ? () : [];
    }

    push @summary, "VG Name:  " . $self->get_name();
    push @summary, "Devices:  " . $self->get_device_cnt();

    $header = sprintf("Vol Summary:\n"
                        . "%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-32.32s\n"
                        . "%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-32.32s",
                      "NAME",   "LAYOUT", "TYPE",  "STATE",  "SIZE(mb)", "MOUNT_PT",
                      "=" x 12, "=" x 7,  "=" x 5, "=" x 12, "=" x 12,   "=" x 32);

    push @summary, $header;

    foreach my $volObj (@{$self->{vols}}) {
        $summary = sprintf("%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-s",
                           $volObj->get_name(),  $volObj->get_layout(), $volObj->get_fs(),
                           $volObj->get_state(), $volObj->get_size(),   $volObj->get_mount());

        push @summary, $summary;
    }

    if (defined $verbose and $verbose =~ /1|on|y|yes|t|true/i) {

        $cmd = "vgdisplay -v $name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error NativeLvm::Vg::display() error while running cmd ($cmd)\n" .
                            "Manual cleanup maybe required, make sure all lvs/vgs/pvs have" .
                            "been removed\n If you can.t remove these items, a reboot of the\n" .
                            "system is required");
            return (wantarray()) ? () : [];
        }

        # Nate will barf on logging lines that have a '--' in the
        # output.  These characters need to be substituted.

        my $output_wo_dash_chars = join "\n", @output, "\n";
        $output_wo_dash_chars =~ s/\-/\=/g;

        @verbose = (@summary, "\nVol Group Detail:", $output_wo_dash_chars);

        push @verbose, "\n\n";
        return (wantarray()) ? @verbose : \@verbose;
    } else {
        push @summary, "\n\n";
        return (wantarray()) ? @summary : \@summary;
    }
}

################################################################################
# Subroutine Name: get_device_cnt
# Objective:       Return number of devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device_cnt {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{disk_cnt}) {
        return $self->{disk_cnt};
    } else {
        $self->_set_err("Error: Vg::get_device_cnt() Unable to display disk count for diskgroup");
        return 0;
    }

}

################################################################################
# Subroutine Name: get_devices
# Objective:       Return devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    # Bug exists where '/dev' prefix is sometimes
    # part of device entry.

    my @tmp = (keys %{$self->{devices}});

    if ("@tmp" =~ /dev/i) {
        @output = @tmp;
    } else {
        foreach my $dev (keys %{$self->{devices}}) {
            unless ($dev =~ /\/dev/i) {
                if ($self->{parent}->{parent}->{mp_type} =~ /dmmp/i){
                    push @output, "/dev/mapper/" . $dev;
                } else {
                    push @output, "/dev/" . $dev;
                }
            }
        }
    }

    return (wantarray()) ? @output : \@output;
}

################################################################################
# Subroutine Name: get_freespace
# Objective:       Return available free space
# Details:         See POD documentation at the end of this file
################################################################################
sub get_freespace {
    my ($self) = @_;
    my ($cmd, $name, @output, $status, $size);

    $self->_clear_err();

    $size = 0;

    $name = $self->{name};

    $cmd = "vgdisplay -c";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error NativeLvm::Vg::get_freespace() error while running cmd ($cmd)\n" .
                        "Manual cleanup maybe required, make sure all lvs/vgs/pvs have" .
                        "been removed\n If you can.t remove these items, a reboot of the\n" .
                        "system is required");
        return $ClientUtil::FALSE;
    }

    (@output) = grep (/$name/,@output);
    (@output) = grep (!/exported/,@output);

    # There will be only 1 item in @output. If this is not the case then we are not filtering properly
    my @vg_info_arr = split (/:/,$output[0]);
    
    # 13th field is PE size (in KBs) and 16th field is no of free PE in the VG
    $size = ($vg_info_arr[15] * $vg_info_arr[12]);
    $size = $self->_convert_to_mbytes($size,'k');

    return $size;

}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vg object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: Vg::get_name() Vg object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size in MB
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size);

    $self->_clear_err();

    $size = 0;

    $name = $self->{name};

    $cmd = "vgdisplay -c";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error NativeLvm::Vg::get_size() error while running cmd ($cmd)\n" .
                        "Manual cleanup maybe required, make sure all lvs/vgs/pvs have" .
                        "been removed\n If you can.t remove these items, a reboot of the\n" .
                        "system is required");
        return $ClientUtil::FALSE;
    }

    (@output) = grep (/$name/,@output);
    (@output) = grep (!/exported/,@output);

    # There will be only 1 item in @output. If this is not the case then we are not filtering properly
    my @vg_info_arr = split (/:/,$output[0]);
    
    # 13th field is PE size (in KBs) and 14th field is no of free PE in the VG
    $size = ($vg_info_arr[13] * $vg_info_arr[12]);
    $size = $self->_convert_to_mbytes($size . 'k');

    return $size;
}

################################################################################
# Subroutine Name: get_vol
# Objective:       return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol {
    my ($self, $name) = @_;
    my ($vgObj, @vols, $str);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Vg::get_vol() volume name is a required argument");
        return {};
    }

    @vols = $self->get_vols();
    foreach my $obj (@vols) {
        if ($obj->get_name() =~ /$name/) {
            return $obj;
        }
    }

    $self->_set_err("Error: Vg::get_vol() Unable to find volume '$name'");

    return {};
}

################################################################################
# Subroutine Name: get_vol_cnt
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol_cnt {
    my ($self) = @_;
    my ($cnt, @vols,);

    $self->_clear_err();

    if (defined $self->{vols}) {
        @vols = $self->get_vols();
        $cnt  = scalar(@vols);
    } else {
        $self->_set_err("Error: Vg::get_vol_cnt() Vg object has no vol count");
        $cnt = 0;
    }

    return $cnt;
}

################################################################################
# Subroutine Name: get_vols
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vols {
    my ($self, %args) = @_;
    my (@volObjs);

    $self->_clear_err();

    if (defined $self->{vols}) {
        @volObjs = @{$self->{vols}};
    } else {
        $self->_set_err("Error: Vg::get_vols() Vg object has no volumes");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @volObjs : \@volObjs;

}

################################################################################
# Subroutine Name: import()
# Objective:       Import Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub import {
    my ($self) = @_;
    my ($cmd, @devs, $name, @output, $status);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Vg::import() Diskgroup Type cannot be determined");
        return $ClientUtil::FALSE;
    }

    $cmd = "vgimport $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error NativeLvm::Vg::import() error while running cmd ($cmd)");
        return $ClientUtil::FALSE;
    }

    $cmd = "vgchange -ay $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error NativeLvm::Vg::import() error while running cmd ($cmd)");
        return $ClientUtil::FALSE;
    }

    $self->{deported} = 0;

    # requery device list
    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    $self->_create_vols(-_import => 1);
    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: is_deported
# Objective:       Determine if volgroup has been deported
# Details:         See POD documentation at the end of this file
################################################################################
sub is_deported {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{deported}) {
        return $self->{deported};
    } else {
        return 0;
    }
}

################################################################################
# Subroutine Name: refresh
# Objective:       refresh content of volume group.
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh {
    my ($self) = @_;
    my ($cmd, $status, @output, $name);

    # Remove existing vol cache
    $self->{vols} = undef;
    $name = $self->get_name();

    $self->sanlog(-msg => "refreshing volume group $name ");

    # Determine if dg is deported
    $cmd = "vgdisplay -c | grep $name" ;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 1) {
        $self->_set_err("Error NativeLvm::Vg::refresh() error while running cmd ($cmd)\n" .
                        "Manual cleanup maybe required, make sure all lvs/vgs/pvs have" .
                        "been removed\n If you can't remove these items, a reboot of the\n" .
                        "system is required");
        return $ClientUtil::FALSE;
    }
    
    foreach my $line (@output) {
        if ( grep (/exported/ , $line)) {
            $self->{deported} = 1;
            last;
        }
    }

    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    # If not deported then lets try to load
    # any existing logical volumes.
    if (not $self->is_deported()) {
        $self->_create_vols();
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remove_device()
# Objective:       remove a device from the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub remove_device {
    my ($self, %args) = @_;
    my ($cmd, $device, $force, $disk_cnt, $dev, $name, @output, @vols_mount, $status);

    $self->_clear_err();

    $force = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Error: Vg::remove_device()", "Invalid value for argument '-device'");
                return $ClientUtil::FALSE;
            }
            next;
        }

    }

    # Add path to device if not provided
    unless ($device =~ /\/dev/i) {
        if ($self->{parent}->{parent}->{mp_type} =~ /dmmp/i){
            $device = "/dev/mapper/$device";
        } else {
            $device = "/dev/$device";
        }
    }

    # get name of colume group
    $name = $self->{name};

    # verify the device is being used in volume group
    unless (exists $self->{devices}->{$device}) {
        $self->_set_err("Error: Vg::remove_device()", "Device '$device' doesn't exist in diskgroup");
        return $ClientUtil::FALSE;
    }

    $cmd = "vgreduce $name $device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error NativeLvm::Vg::remove_device() error while running cmd ($cmd)");
        return $ClientUtil::FALSE;
    }

    # remove device from hash
    delete($self->{devices}->{$device});

    # decrement device counter;
    $self->{disk_cnt}--;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rename()
# Objective:       rename volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, %args) = @_;
    my ($cmd, $force, $name, $newname, $status, @disks, @output, @vols_mount, @volObjs,);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($newname = $args{$_})) {
                $self->_set_err("Error: Vg::rename()", "Invalid value for argument '-name'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Vg::rename()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # get name of disk group
    $name = $self->get_name();

    # verify volume group has a name
    unless (defined $name) {
        $self->_set_err("Error: Vg::rename()", "Volume group doesn't have a name");
        return $ClientUtil::FALSE;
    }

    # verify that a new volume group name was specified
    unless (defined $newname) {
        $self->_set_err("Error: Vg::rename()", "New name for Volume group not specified");
        return $ClientUtil::FALSE;
    }

    # if force option specified umount all volumes so deport
    # can occur without error.  Make note of which vols were
    # mounted so they can be remounted after rename occurs

    if (defined $force) {
        @volObjs = $self->get_vols();
        return $ClientUtil::FALSE if ($self->error);

        foreach my $volObj (@volObjs) {
            if ($volObj->is_fs()) {
                if ($volObj->is_mounted()) {

                    # retrieve current mount path
                    # before unmounting the filesystem.
                    my $path = $volObj->get_mount();
                    if ($volObj->error) {
                        $self->_set_err($volObj->errmsg);
                        return $ClientUtil::FALSE;
                    }

                    # Remove the vg and vol path info
                    my $vgvol = "/" . $name . "/" . $volObj->get_name();
                    $path =~ s/$vgvol//g;

                    # Unmount the filesystem
                    $volObj->unmount();
                    if ($volObj->error) {
                        $self->_set_err($volObj->errmsg());
                        return $ClientUtil::FALSE;
                    }

                    # Store info on volumes
                    my $hVols = {};
                    $hVols->{volObj} = $volObj;
                    $hVols->{path}   = $path;
                    push @vols_mount, $hVols;
                }
            }
        }
    }

    # Rename the diskgroup
    $cmd = "vgrename $name $newname";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error NativeLvm::Vg::rename() error while running cmd ($cmd)\n");
        return $ClientUtil::FALSE;
    }

    $self->{name} = $newname;

    # Mount filesystems on volumes that were previously mounted
    # before renaming volume group.
    foreach my $hVol (@vols_mount) {
        my $volObj = $hVol->{volObj};
        $volObj->mount(-path => $hVol->{path});
        if ($volObj->error) {
            $self->_set_err($volObj->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _create_vols
# Objective:       create vol object for a volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vols {
    my ($self, %args) = @_;
    my ($cmd, $import, $name, $size, $status, $type, $unit);
    my ($vgname, $vgObj, $vol, $volObj, @names, @output,);

    # Identify volumes that are being imported so
    # that their automatically mounted with their
    # last filesystem mount point.
    $import = (exists $args{-_import}) ? $args{-_import} : 0;

    $name = $self->{name};

    ## Get list of volumes
    $cmd = "lvs $name";
    ($status, @names) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 and (!grep (/exported/, @names))) {
        $self->_set_err("Error: Vg::_create_vols()" . "Unable to list available disk group @names");
        return $ClientUtil::FALSE;
    }

    # Filter out the header
    if (@names) {
        @names = grep(/$name/i, @names);
    }    

    foreach my $line (@names) {
        my (undef, $volname, undef, undef, $size) = split (/\s+/,$line);

        # Assume each volume is concatenated unless otherwise specified
        $type = "concat";

        $size = $self->_convert_to_mbytes("$size");

        $volObj = $self->create_vol(-name => $volname, -size => "$size" . "M", 
                                    -type => $type, -exist => 1, 
                                    -_import => $import);
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _convert_to_mbytes()
# Objective:       convert size strings to bytes.
# Details:         See POD documentation at the end of this file
################################################################################
sub _convert_to_mbytes {
    my ($self, $size) = @_;
    my ($mbytes, $nbr, $unit);

    # If no units then mb
    $size .= "m" if ($size =~ /^\d+$/);

    ($nbr, $unit) = $size =~ /(\S+)(\w\w)$/;

    if ($unit =~ /k/i) {
        $mbytes = $nbr / 1024;
    } elsif ($unit =~ /m/i) {
        $mbytes = $nbr;
    } elsif ($unit =~ /g/i) {
        $mbytes = $nbr * 1024;
    }

    return $mbytes;
}

################################################################################
# Subroutine Name: _create()
# Objective:       create a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self, $name, $devices) = @_;
    my ($cmd, @cmds, @devices, $device_with_path, @output, $status, %hDevice);

    foreach my $device (@$devices) {
        $device_with_path = $device;
        unless ($device =~ /\/dev/i) {
            if ($self->{parent}->{parent}->{mp_type} =~ /dmmp/i){
                $device_with_path = "/dev/mapper/" . $device;
            } else {
                $device_with_path = "/dev/" . $device;
            }
        }
        $self->sanlog(-msg => "refreshing kernel cache for device '" . $device_with_path. "'");

        $cmd = "pvcreate " . $device_with_path;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error NativeLvm::Vg::_create() error while running cmd ($cmd)");
            return $ClientUtil::FALSE;
        }

        push @devices,  $device_with_path;
    }

    # Sometime VG's stale entry(n_vg) left in /dev dir because of some 
    # improper cleanup. It is not possible to create VG with same name 
    # Remove the stale VG entry from /dev if exists.
    if (defined ($name) and $name ne "") {
        $cmd = "ls /dev/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            $cmd = "rm -rf /dev/$name";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to remove stale VG ($name) entry from /dev");
                return $ClientUtil::FALSE;
            }
        }
    }

    @cmds = ("vgcreate -s 4m $name @devices", "vgchange -a y $name",);
    foreach $cmd (@cmds) {
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0 ) {
            $self->_set_err("Error NativeLvm::Vg::_create() error while running cmd ($cmd)");
            return $ClientUtil::FALSE;
        }
    }

    foreach my $device (@devices) {
        my $value;
        if ($self->{parent}->{parent}->{mp_type} =~ /dmmp/i){
            ($value) = $device =~ /\/dev\/mapper\/(\S+)/;
        } else {
            ($value) = $device =~ /\/dev\/(\S+)/;
        }
        $hDevice{$device} = $value;
    }

    $self->{devices}  = \%hDevice;
    $self->{disk_cnt} = scalar(@$devices);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_deviceelf->{devices}
# Objective:       get list of devices used by disk group
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices {
    my ($self) = @_;
    my ($cmd, $name, $status, @output, %hDevice,);

    $name = $self->get_name();

    $cmd = "pvs";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error NativeLvm::Vg::_create() error while running cmd ($cmd)");
        return {};
    }
    if (@output) {
        @output = grep(/$name/,@output);
    }    

    $self->{disk_cnt} = scalar(@output);

    foreach my $line (@output) {
        my (undef,$device) = split(/\s+/,$line);
        $hDevice{$device} = $device;
    }

    return \%hDevice;

}

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($cmd,    $conn,     $devices, $disk_cnt, $parent, $log,   $name, $self,
        $status, $deported, @output,  %hDevice,  $type,   $value, $fs);

    $self = {};

    bless $self, $class;

    $self->_clear_err();

    $deported = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: Vg::_new()", "Invalid value for argument '-devices'");
                return $self;
            }
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: Vg::_new()", "Missing value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: Vg::_new()", "Missing value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Vg::_new()", "Missing value for argument '-lvmobj'");
                return $self;
            }
            next;
        } elsif (/-?deported/) {
            if (not length($deported = $args{$_})) {
                $self->_set_err("Error: Vg::_new()", "Missing value for argument '-deported'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -lvm argument
    unless (defined $parent) {
        $self->_set_err("Error: Vg::_new()", "Missing required argument '-parent'");
        return $self;
    }

    # check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: Vg::_new()", "Missing required argument '-name'");
        return $self;
    }

    # check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: Vg::_new()", "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: Vg::new()", "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn}     = $conn;
    $self->{name}     = $name;
    $self->{deported} = $deported;
    $self->{disk_cnt} = 0;
    $self->{parent}   = $parent;

    # create the diskgroup if devices were supplied
    if (defined $devices and ref($devices) eq "ARRAY") {
        $self->_create($name, $devices);
        return $self if ($self->error());
    } else {

        if ($self->{deported} == 1) {
            $self->import();
            return $self if ($self->error);
        } else {

        # vg must already exist. set device list
        # for all existing and deported dgs.
        $self->{devices} = $self->_get_devices();
        return $self if ($self->error);

        $self->_create_vols();
        return $self if ($self->error);
        }
    }

    return $self;
}

1;
