# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to NativeLvm logical volumes on Linux
## @author knopp@netapp.com, rawat@netapp.com
## @status private
## @pod ../../../Vol.pod

package San::SanClient::Lvm::Linux::NativeLvm::Vg::Vol;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use San::SanUtil;
use Error;

use Data::Dumper;
use File::Basename;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
# Method:    create_fs
# Objective: create a filesystem on a volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_fs {
    my ($self, %args) = @_;
    my ($cmd, $type, $vgname, $name, $status, @output,);

    $vgname = $self->get_vgname();

    $name = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?type$/) {
            $type = $args{$_};
            next;
        }
    }

    my $clientObj = $self->_get_client();

    # check for existance of required -lvm argument
    if (not defined $type or $type eq "") {
        if ($clientObj->{major_revision} =~ /rhel6|rhel7|ol6|fedora/i) {
            # From RHEL6 and OL6 onwards, the preferred file system is ext4.
            # So, we are creating ext4 file system by default.
            $type = "ext4";
        } else {
            $type = "ext3";
        }
    }

    # Remove leading and trailing white spaces
    $type =~ s/^\s*//;
    $type =~ s/\s*$//;

    if ($type =~ /^ext2$|^ext3$|^ext4$/i) {
        # Create a filesystem
        # "-t fs-type" option is supported only from RHEL6, OL6 and SLES11 OSes
        $cmd = "mke2fs ";
        $cmd .= "-t $type " if ($clientObj->{major_revision} =~ /rhel6|rhel7|ol6|sles11|fedora/i);
        $cmd .= "-j " if ($type =~ /ext3/i);
        $cmd .= "/dev/$vgname/$name";

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to create $type filesystem on volume '$name'", @output);
            return $ClientUtil::FALSE;
        }
    } elsif ($type =~ /^xfs$/i) {
        $cmd = "yes | mkfs.xfs /dev/$vgname/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Unable to create $type filesystem on volume '$name'. " .
                            "Make sure that xfsprogs package is installed", @output);
            return $ClientUtil::FALSE;
        }
    } elsif ($type =~ /^btrfs$/i) {
        $cmd = "yes | mkfs.btrfs /dev/$vgname/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            $self->_set_err("Unable to create $type filesystem on volume '$name'. " .
                            "Make sure that btrfsprogs package is installed", @output);
            return $ClientUtil::FALSE;
        }
    } elsif ($type =~ /^gfs$/i) {
        my $lockproto = "lock_dlm";
        my $cluster = $self->{cluster};
        my ($vol_num) = ($name =~ /nvol-(\d+)/);
        my $gfs_name = "gfs" . $vol_num;
        my $num_of_journals = 8;
        $cmd = "yes | /sbin/gfs_mkfs -p $lockproto -t $cluster:$gfs_name -j $num_of_journals /dev/$vgname/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to create $type filesystem on volume '$name'", @output);
            return $ClientUtil::FALSE;
        }
    } elsif ($type =~ /^reiserfs/i) {
        $cmd = "yes | mkfs.reiserfs /dev/$vgname/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to create $type filesystem on volume '$name'", @output);
            return $ClientUtil::FALSE;
        }
    } else {
        $self->_set_err("Invalid Filesystem type '$type'");
        return $ClientUtil::FALSE;
    }

    $self->{filesystem} = $type;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self, %args) = @_;

    my ($cmd, $force, $dgname, $name, @output, $status, @vols,);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }

    }

    $dgname = $self->get_vgname();
    $name   = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    # if force option specified umount all volumes so deport
    # can occur without error.  Make note of which vols were
    # mounted so they can be remounted after rename occurs
    if (defined $force) {
        if ($self->is_fs()) {
            if ($self->is_mounted()) {
                $self->unmount();
                return $ClientUtil::FALSE if ($self->error());
            }
        }
    }

    # Added extra command because lvremove is failing on OL5.6. Added command is working for other OSs
    $cmd = "vgchange -a y " . $dgname . "; " . "lvchange -a n /dev/$dgname/$name" . "; " . "lvremove -f /dev/" . $dgname . "/" . $name;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error while running cmd ($cmd)");
        return $ClientUtil::FALSE;
    }

    # Remove the volume from the parent vol group object
    foreach my $volObj (@{$self->{parent}->{vols}}) {
        if ($volObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vols, $volObj;
        }
    }

    $self->{parent}->{vols} = \@vols;

    return $ClientUtil::TRUE;

}
################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;
    my ($cmd, $header, $name, $status, $summary, $verbose, $vgname, @buf, @output,);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Invalid value for argument '-verbose'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    $vgname = $self->{parent}->get_name();

    $name = $self->{name};

    unless (defined $name) {
        $self->_set_err("volume name is a required parameter");
        return (wantarray()) ? () : [];
    }

    $header = sprintf("\nVol Summary:\n"
                        . "%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-s\n"
                        . "%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-32.32s",
                      "NAME",   "LAYOUT", "TYPE",  "STATE",  "SIZE(mb)", "MOUNT_PT",
                      "=" x 12, "=" x 7,  "=" x 5, "=" x 12, "=" x 12,   "=" x 32);

    $summary = sprintf("%-12.12s %-7.7s %-5.5s %-12.12s %-12.12s %-s",
                       $self->get_name(),  $self->get_layout(), $self->get_fs(),
                       $self->get_state(), $self->get_size(),   $self->get_mount());

    push @buf, $header, $summary;

    if (defined $verbose and $verbose =~ /1|on|y|yes|t|true/i) {
        $cmd = "lvdisplay /dev/" . $vgname . "/" . $name;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error while running cmd ($cmd)");
            return (wantarray()) ? () : [];
        }

        # Nate will barf on logging lines that have a '--' in the
        # output.  These characters need to be substituted.

        my $output_wo_dash_chars = join "\n", @output, "\n";
        $output_wo_dash_chars =~ s/\-/\=/g;

        push @buf, "\nVol Detail:", $output_wo_dash_chars;
    }

    return (wantarray()) ? @buf : \@buf;
}

################################################################################
# Subroutine Name: get_fs
# Objective:       Return Vol object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{filesystem} and length $self->{filesystem}) {
        return $self->{filesystem};
    } else {
        $self->_set_err("Vol object is missing filesystem");
        return "";
    }
}

################################################################################
# Subroutine Name: get_layout
# Objective:       Return volume layout
# Details:         See POD documentation at the end of this file
################################################################################
sub get_layout {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{layout}) {
        return $self->{layout};
    } else {
        $self->_set_err("Vol object is missing layout");
        return "";
    }
}
################################################################################
# Subroutine Name: get_mount
# Objective:       Return mount point of volume
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{mount} and length $self->{mount}) {
        return $self->{mount};
    } else {
        $self->_set_err("Vol filesystem is not mounted");
        return "";
    }
}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vol object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    my ($vg_name, $lv_name);

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Vol object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return ooject size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{size}) {
        return $self->{size};
    } else {
        $self->_set_err("Vol object is missing size");
        return "";
    }
}

################################################################################
# Subroutine Name: get_state
# Objective:       Return object state
# Details:         See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{state}) {
        return $self->{state};
    } else {
        $self->_set_err("Vol object is missing state");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vgname
# Objective:       Return Vol object diskgroup name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgname {
    my ($self) = @_;
    my ($name,);

    $self->_clear_err();

    $name = $self->{parent}->get_name();

    if (defined $name and length $name) {
        return $name;
    } else {
        $self->_set_err("Vol object is missing parent name");
        return "";
    }
}

################################################################################
# Subroutine Name: is_fs
# Objective:       Determine if volume has a filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub is_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{filesystem} and $self->{filesystem} !~ /raw/i) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if volume is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    my $name   = $self->get_name();
    my $cmd    = "mount |grep $name";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    $self->_clear_err();
    my @arr = split(/\s+/, $output[0]);
    my $mount = $arr[2];
    if (length $mount) {
         $self->{mount} = $mount;
         return $ClientUtil::TRUE;
    }

    if (defined $self->{mount} and length $self->{mount}) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: mount
# Objective:       mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub mount {
    my ($self, %args) = @_;
    my ($cmd, $type, $vgname, $mount, $name, $pass_thru, $discard,
        $status, @output);

    # Get attributes
    $vgname = $self->{parent}->get_name();
    $name   = $self->{name};
    $type   = $self->{filesystem};
    $mount  = undef;
    $discard = "";
    $pass_thru  = "";

    # check if volume is unformatted
    if (defined $type and $type =~ /raw/) {
        $self->_set_err("Cannot mount a raw filesystem");
        return $ClientUtil::FALSE;
    }

    # get client obj
    my $clientObj = $self->_get_client();

    # parse arguments
    foreach (keys %args) {
        if (/-?path$/) {
            if (not length($mount = $args{$_})) {
                $self->_set_err("Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?pass_thru$/) {
            $pass_thru = $args{$_};
            $pass_thru =~ s/^\s+//;
            $pass_thru =~ s/\s+$//;
            if ($pass_thru ne "") {
                $pass_thru = "-o $pass_thru";
            }
        } elsif (/-?discard$/) {
            $discard = $args{$_};
        }
    }

    if ($discard =~ /^y$|^yes$|^1$/i) {
        if ($type =~ /ext4|btrfs|xfs/i) {
            # Discard option is supported with only ext4, btrfs and xfs file systems
            $discard = "-o discard";
        } else {
            &SanUtil::nate_log('-warn', "Mount option (discard) is not supported with the file system ($type)");
            $discard = "";
        }
    } else {
        $discard = "";
    }

    # check if the volume is already mounted
    if ($self->is_mounted()) {
        $self->_set_err("Volume '$name' is already mounted");
        return $ClientUtil::FALSE;
    }

    # Use default path if not defined
    if (not defined $mount) {
        $mount = "/mnt/qa/$vgname/$name";
        $self->sanlog(-msg => "using default mount pt '$mount'");
    } else {
        $mount .= "/$vgname/$name";
        $self->sanlog(-msg => "using mount pt '$mount'");
    }

    # Save the mount point with the vol metadata in case
    # a vol is renamed while mounted
    $self->_set_metadata("mounted", $mount);
    return $ClientUtil::FALSE if ($self->error);

    # Create the mount point and open up dir privs
    $cmd = "mkdir -p $mount; chmod 777 $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to create mount point directory for volume '$name'", @output);
        return $ClientUtil::FALSE;
    }

    # Mount the volume's filesystem
    $cmd = "mount -t $type ";
    $cmd .= "$discard " if $discard !~ /^\s*$/;
    $cmd .= "$pass_thru " if $pass_thru !~ /^\s*$/;
    $cmd .= "/dev/$vgname/$name $mount";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to mount volume filesystem", @output);
        return $ClientUtil::FALSE;
    }

    # open up permissions on mount point
    $cmd = "chmod 777 $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to change permissions on mount ($mount)",  @output);
        return $ClientUtil::FALSE;
    }

    $self->{state} = $self->_init_state();
    return $ClientUtil::FALSE if ($self->error);

    # Get handle to client object
    #$self->{client} = $self->{parent}{parent}{parent};

    #$self->{client}->mount_entry_add(-device_mount => "/dev/$vgname/$name",
    #                                 -mount_pt     => $mount,
    #                                 -fs_type      => $self->get_fs());
    #if ($self->{client}->error) {
    #    $self->_set_err($self->{client}->errmsg());
    #    return $ClientUtil::FALSE;
    #}

    $self->{mount} = $mount;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rename()
# Objective:       rename volume
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, $newname) = @_;
    my ($bMounted, $cmd, $oldPath, $name, $vgname, @output, $status);

    $self->_clear_err();

    $bMounted = 0;

    # get name of disk group
    $vgname = $self->{parent}->get_name();

    # get name of volume
    $name = $self->{name};

    # verify the device is being used in volume group
    unless (defined $name) {
        $self->_set_err("Volume doesn't have a name");
        return $ClientUtil::FALSE;
    }

    # Don't allow volumes of type 'jfslog' to be modified.
    # Although this is supported
    if ($self->{type} =~ /jfslog/i) {
        $self->_set_err("Names of volumes used for log cannot be changed");
        return $ClientUtil::FALSE;
    }

    # verify the device is being used in volume group
    unless (defined $newname) {
        $self->_set_err("New name for Volume not specified");
        return $ClientUtil::FALSE;
    }

    # Volume must be unmounted in order to rename it
    if ($self->is_fs()) {
        if ($self->is_mounted()) {

            # flag to inform method of state
            # before name change
            ($oldPath) = $self->get_mount() =~ /(\S+)\/$vgname\/$name$/;
            return $ClientUtil::FALSE if ($self->error);
            $bMounted = 1;
            $self->unmount();
            return $ClientUtil::FALSE if ($self->error);
        }
    }

    # Rename the volume
    $cmd = "lvrename /dev/$vgname/$name /dev/$vgname/$newname";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error while running cmd ($cmd)");
        return $ClientUtil::FALSE;
    }

    $self->{name} = $newname;

    # Remount the volume if it was mounted previously
    if ($bMounted == 1) {
        (defined $oldPath and length $oldPath)
          ? $self->mount(-path => $oldPath)
          : $self->mount();

        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rescan()
# Objective:       rescan volume
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;
    my ($mount);

    # Currently just handles remounting.
    # Needed after snapdrive restore
    $self->_clear_err();

    $mount = $self->_init_mount();
    if (defined $mount and length $mount) {
        $self->mount(-path => $mount);
    } else {
        $self->{mount} = undef;
    }

    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unmount()
# Objective:       unmount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount {
    my ($self, %args) = @_;
    my ($cmd, $type, $vgname, $fs, $mount, $name, $status, $force, $option, @output);

    $self->_clear_err();

    $option = "";

    # parse hash
    foreach (keys %args) {
        if (/-?force$/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $option = "-f" if (defined $force);

    # Get attributes
    $vgname = $self->get_vgname();
    $name   = $self->get_name();

    $mount = $self->{mount};

    $fs = $self->get_fs();

    # check if volume contains a filesystem
    unless ($fs !~ /raw/i) {
        $self->_set_err("Cannot unmount a raw device");
        return $ClientUtil::FALSE;
    }

    # check if volume is mounted
    unless (defined $mount and length($mount)) {
        $self->_set_err("Cannot unmount a filesystem that is not mounted");
        return $ClientUtil::FALSE;
    }

    # Unmount the volume's filesystem
    $cmd = "umount $option $mount";
    foreach my $loop (1 .. 3) {

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {

            # Unmount -force does not always seem to work
            # so wait a second try again.
            if (($option =~ /f/i) and ($loop < 4)) {
                $self->sanlog(-msg => "NativeLvm::Vg::Vol::unmount() Unable to unmount with '-force', trying again");
                &SanUtil::nate_sleep(1);
                next;
            }
            $self->_set_err("Unable to unmount volume's filesystem", @output);
            return $ClientUtil::FALSE;
        }

        last;
    }

    # Remove mount pt metadata from object
    $self->{mount} = "";

    # remove the mount entry from /etc/fstab
    # Clear fstab mount entry
    # client = lvm <-- vg <-- vol
    $self->{client} = $self->_get_client();

    #$self->{client}->mount_entry_delete($mount);
    #if ($self->{client}->error) {
    #    $self->_set_error($self->{client}->errmsg());
    #    return $ClientUtil::FALSE;
    #}

    # Remove directory head
    my ($base, $parent) = fileparse($mount);
    $cmd = "rm -rf $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to remove mount point '$mount'", @output);
    }

    $self->{state} = $self->_init_state();
    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       Return client object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;
    my ($obj);

    if (defined $self->{raw_device}) {
        return $self->{raw_device};
    } else {
        $self->_set_err("Vol object is missing raw device binding");
        return {};
    }
}

################################################################################
#
# P R I V A T E   S U B R O U T I N E S
##
################################################################################
################################################################################
# Subroutine Name: _create()
# Objective:       create a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self) = @_;
    my ($cmd, $device_cnt, $dg, $layout, $name, $size, $status, $type, @devs, @output,);

    $dg         = $self->{parent}->get_name();
    $device_cnt = $self->{parent}->get_device_cnt();
    $name       = $self->get_name();
    $size       = $self->{size};
    $layout     = $self->{layout};
    $type       = $self->{type};

    @devs = (keys %{$self->{parent}->{devices}});

    $cmd = "lvcreate -n $name ";
    $cmd .= "-i " . $device_cnt . " " if ($device_cnt > 1 and $layout =~ /stripe/i);
    $cmd .= "-L $size " . $dg ;

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error while running cmd ($cmd)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_client
# Objective:       Return client object
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_client {
    my ($self) = @_;
    my ($obj);

    $obj = $self->{parent}{parent}{parent};

    if (defined $obj) {
        return $obj;
    } else {
        $self->_set_err("Vol object is missing client obj");
        return {};
    }
}

################################################################################
# Subroutine Name: _get_metadata
# Objective:      Retrieve metadata for attribute
# Details:        See POD documentation at the end of this file
################################################################################
sub _get_metadata {
    my ($self, $attribute) = @_;

    my ($cmd, $line, $vgname, $name, @output, $status, $value, $str);

    $value = "";

    $vgname = $self->get_vgname();
    $name   = $self->get_name;
    
    # In case of reiser filesystem it is not possibe to store metadata (such as mount point)
    # using dumpfs inside filesystem. So Storing the metadata in /tmp for reiserfs.
    if ($self->{filesystem} =~ /reiserfs|gfs/) {
        $cmd = "cat /tmp/fs.metadata";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to cat /tmp/fs.metadata");
            return "";
        }
        foreach my $line (@output) {
            if ($line =~ /$vgname\/$name/){
                ($value) = ($line =~ /$vgname\/$name:(.*)/);
                chomp($value);
                last;
            }
        }
    } else {
        $cmd = "dumpe2fs -h /dev/$vgname/$name | grep '$attribute'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($self->error) { 
            $self->_set_err("Unable to retrieve volume \'$name\' metadata for",
                            "attribute '$attribute' in diskgroup '$vgname'", @output);
            return "";
        }

        # return the attribute value if exists.
        if (defined $output[scalar(@output) - 1]) {
            ($value) = $output[scalar(@output) - 1] =~ /:\s+(.*)/;
            return "" if ($value =~ /\</i);
        }
    }

    return $value;
}

################################################################################
# Subroutine Name: _init_block_device
# Objective:      Initialize block device for existing host volumes
#				  Will return major:minor of a block device which will be
#				  used by _create_raw_bindings to bind a raw character device
#				  to a block device
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_block_device {
    my ($self, @buf) = @_;

    my ($cmd, $vgname, $device, $name, @output, $sectors, $status);

    $device = "";

    my $clientObj = $self->_get_client();
    $vgname = $self->get_vgname();
    $name   = $self->get_name();

    # If info buffer passed in don't requery for data
    if (scalar(@buf) > 0) {
        foreach my $line (@buf) {
            if ($line =~ /Block device/i) {
                ($device) = $line =~ /Block device\s+(\S+)/;
                last;
            }
        }
        unless (defined $device and length $device) {
            $self->_set_err("Unable to retrieve block device for volume \'$name\' in diskgroup '$vgname'");
            return "";
        }
    } else {
        if (($clientObj->{major_revision} =~ /rhel5/ and $clientObj->{minor_revision} > 2)
           or ($clientObj->{major_revision} =~ /ol5/ and $clientObj->{minor_revision} >= 6)
           or $clientObj->{major_revision} =~ /rhel6|rhel7|ol6|sles11|fedora/) {
            $cmd = "lvdisplay /dev/$vgname/$name | grep 'Block device' | gawk '{if (\$3) print \$3; else exit 1; }'";
        } else {
            $cmd = "lvdisplay -D /dev/$vgname/$name | grep 'Block device' | gawk '{if (\$3) print \$3; else exit 1; }'";
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error while running cmd ($cmd)");
            return "";
        }

        $device = $output[0];
    }

    return $device;

}

################################################################################
# Subroutine Name: _init_fstype
# Objective:      Initialize fstype for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_fstype {
    my ($self) = @_;

    my ($cmd, $fstype, $vgname, $name, @output, $status, $clientObj);

    $vgname = $self->get_vgname();
    $name   = $self->get_name();
    $clientObj = $self->_get_client();

    if ($clientObj->{major_revision} =~ /rhel7|rhel6|fedora|ol6/i) {
        # mount option "--guess-fstype" is not supported in RHEL7
        $cmd = "blkid \/dev\/$vgname\/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            if ($output[0] =~ /TYPE="(\S+)"/i) {
                $fstype = $1;
            } else {
                $self->_set_err("Recieved unexpected output", @output);
                return "";
            }
        }
    } else {
        $cmd = "mount -f --guess-fstype \/dev\/$vgname\/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            $fstype = $output[0];
        }
    }

    if ($status != 0) {
        # command fails even if the device is not there in the system.
        # Make sure that you set the error properly
        $cmd = "ls \/dev\/$vgname\/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("$name is not available on the system");
            return "";
        }

        # Make sure it is block device
        $cmd = "test -b \/dev\/$vgname\/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status > 0) {
            $self->_set_err("$name is not available on the system");
            return "";
        }
        $fstype = "raw";
    }

    return $fstype;
}
################################################################################
# Subroutine Name: _init_layout
# Objective:      Initialize layout of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_layout {
    my ($self, @buf) = @_;

    my ($cmd, $vgname, $layout, $name, @output, $status);

    $layout = "";

    my $clientObj = $self->_get_client();
    $vgname = $self->{parent}->get_name();
    $name   = $self->{name};

    # If info buffer passed in don't requery for data
    if (scalar(@buf) > 0) {
        foreach my $line (@buf) {
            if ($line =~ /Stripe/i) {
                ($layout) = "striped";
                last;
            }
        }

        $layout = "concat";

    } else {
        if (($clientObj->{major_revision} =~ /rhel5/ and $clientObj->{minor_revision} > 2)
           or ($clientObj->{major_revision} =~ /ol5/ and $clientObj->{minor_revision} >= 6)
           or $clientObj->{major_revision} =~ /rhel6|rhel7|ol6|sles11|fedora/) {
            $cmd = "lvdisplay  /dev/$vgname/$name " . "| egrep 'LV Name|Stripes' " . "| wc -l";
        } else {
            $cmd = "lvdisplay -D  /dev/$vgname/$name " . "| egrep 'LV Name|Stripes' " . "| wc -l";
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error while running cmd ($cmd)");
            return "";
        } 
        
        if ($output[0] == 1) {
            $layout = "concat";
        } elsif ($output[0] == 2) {
            $layout = "striped";
        }
    }

    return $layout;
}
################################################################################
# Subroutine Name: _init_mount
# Objective:      Initialize mount point for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_mount {
    my ($self, $import) = @_;

    my ($cmd, $mount, $name, $path, $status, $vgname, @output);

    $mount = "";

    $vgname = $self->{parent}->get_name();
    $name   = $self->{name};

    # Check to see if mount path has been store
    # in volume metadata under 'mounted' attribute.
    # This is to done in order to associate a mount
    # pt with a vg/vol in case the vol is renamed while
    # still mounted. (Note: Can vol rename occur if mounted?).

    # Need the -P to get the df -k output all on one line
    $cmd = "df -kP | egrep -e \'/$vgname/$name\' | awk '{print \$6;} END {if (\$count==0) exit 1;}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {

        # If volumes are being imported then
        # automatically mount fs with last stored
        # mount point. If one does not exist then
        # mount point with default mount path.
        if ($import) {
            $mount = $self->_get_metadata("mounted");
            return "" if ($self->error);

            if (defined $mount and length($mount)) {
                ($path, $vgname, $name) = $mount =~ /(\S+)\/(\w+)\/(\w+)$/;
            }

            $self->mount(-path => $path);
            return "" if $self->error();
        }
    } else {
        $mount = $output[0];
    }

    # return the volume mount point.
    return $mount;

}

################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self, @buf) = @_;

    my ($cmd, $vgname, $size, $name, @output, $sectors, $status);

    $size = "";

    my $clientObj = $self->_get_client();
    $vgname = $self->get_vgname();
    $name   = $self->get_name();

    # If info buffer passed in don't requery for data
    if (scalar(@buf) > 0) {
        foreach my $line (@buf) {
            if ($line =~ /LV Size/i) {
                my ($sz, $unit) = $line =~ /LV Size\s+(\S+)\s+(\S+)/;
                $size = &ClientUtil::convert_to_bytes(int $sz . $unit);
                last;
            }
        }
        unless (defined $size and length $size) {
            $self->_set_err(  "Unable to retrieve  size (MB) for volume \'$name\' in diskgroup '$vgname'");
            return 0;
        }

    } else {
        if (($clientObj->{major_revision} =~ /rhel5/ and $clientObj->{minor_revision} > 2)
           or ($clientObj->{major_revision} =~ /ol5/ and $clientObj->{minor_revision} >= 6)
           or $clientObj->{major_revision} =~ /rhel6|rhel7|ol6|sles11|fedora/) {
            $cmd = "lvdisplay /dev/$vgname/$name";
        } else {
            $cmd = "lvdisplay -D /dev/$vgname/$name";
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error while running cmd ($cmd)");
            return 0;
        }

        @output = grep(/LV\s+Size/i , @output);
        my ($sz, $unit) = ($output[0] =~ /LV\s+Size\s+(\d+\.?\d*)\s+(\S+)/i);
        $size = &ClientUtil::convert_to_bytes($sz . $unit);

    }
    return $size;

}

################################################################################
# Subroutine Name: _init_state
# Objective:      Initialize state of existing host volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_state {
    my ($self, @buf) = @_;

    my ($cmd, $name, $state, $vgname, $status, @output);

    $state = "";

    my $clientObj = $self->_get_client();
    $vgname = $self->get_vgname();
    $name   = $self->get_name();

    # If info buffer passed in don't requery for data
    if (scalar(@buf) > 0) {
        foreach my $line (@buf) {
            if ($line =~ /LV Status/i) {
                ($state) = ($line =~ /LV Status\s+(.*)/i);
                last;
            }
        }

        unless (defined $state and length $state) {
            $self->_set_err("Unable to retrieve state for volume \'$name\' in diskgroup '$vgname'");
            return "";
        }

    } else {
        if (($clientObj->{major_revision} =~ /rhel5/ and $clientObj->{minor_revision} > 2)
           or ($clientObj->{major_revision} =~ /ol5/ and $clientObj->{minor_revision} >= 6)
           or $clientObj->{major_revision} =~ /rhel6|rhel7|ol6|sles11|fedora/) {
            $cmd = "lvdisplay /dev/$vgname/$name";
        } else {
            $cmd = "lvdisplay -D /dev/$vgname/$name";
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error while running cmd ($cmd)");
            return "";
        }

        @output = grep(/LV\s+Status/i , @output);
        ($state) = ($output[0] =~ /LV\s+Status\s+(\S+)/i);
    }

    return $state;

}

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($cmd, $exist, $import, $name, $size, $conn, $layout, $log);
    my ($cluster, $parent, $self, $status, $type, $vgname, @output,);

    $self = {};

    bless $self, $class;

    #default layout is concatenation
    $layout = "concat";
    $import = 0;
    $type   = "raw";

    $exist = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?layout$/) {
            if (not length($layout = $args{$_})) {
                $self->_set_err("Invalid value for argument '-layout'");
                return $self;
            }
            next;
        } elsif (/-?_import$/) {    # private member
            if (not length($import = $args{$_})) {
                $self->_set_err("Invalid value for argument '-_import'");
                return $self;
            }
            next;
        } elsif (/-?cluster$/) {
            if (not length($cluster = $args{$_})) {
                $self->_set_err("Missing value for argument '-cluster'");
                return $self;
            }
            next;
        } elsif (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?size$/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Invalid value for argument '-size'");
                return $self;
            }
            next;
        } elsif (/-?type$/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return $self;
            }
            next;
        } elsif (/-?exist$/) {
            if (not length($exist = $args{$_})) {
                $self->_set_err("Invalid value for argument '-exist'");
                return $self;
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Missing value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    # check for existance of required -parent argument
    unless (defined $parent) {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # check for existance of required -type argument
    unless (defined $type) {
        $self->_set_err("Missing required argument '-type'");
        return $self;
    }

    # check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Missing required argument '-name'");
        return $self;
    }

    # check for existance of required -size argument
    unless (defined $size) {
        $self->_set_err("Missing required argument '-size'");
        return $self;
    }

    # check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-conn'");
        return $self;
    }

    # initialize the object members
    $self->{client}     = $self->{parent}{parent}{parent};
    $self->{conn}       = $conn;
    $self->{name}       = $name;
    $self->{layout}     = $layout;
    $self->{size}       = $size;
    $self->{type}       = $type;
    $self->{parent}     = $parent;
    $self->{state}      = 'available';
    $self->{filesystem} = 'raw';
    $self->{cluster}    = $cluster;

    $vgname = $self->{parent}->get_name();

    if (not $exist) {
        $self->_create();
        return $self if ($self->error());

        # Assume all logical volumes will not have a fs and
        # will be accessed using raw device.
        # _init_block_devices will bind raw character device to a block device
        $self->{block_device} = $self->_init_block_device();
        return $self if ($self->error);

        my $clientObj = $self->_get_client();

        if ($clientObj->{major_revision} !~ /fedora|rhel7/) {
            $self->{raw_device} = $clientObj->_create_raw_binding($self->{block_device});
            if ($clientObj->error()) {
                $self->_set_err($clientObj->errmsg());
                return $self;
            }
            $self->sanlog(-msg => "if _new:raw device:" . $self->{raw_device});
        }
    } else {
        $self->sanlog(-msg       => "loading existing volume '$name'",
                      -msg_level => 7);

        my $clientObj = $self->_get_client();
        # Retrieve as much needed metadata as possible in a single request.
        # Doesn't make sense to query each element seperately.
        if (($clientObj->{major_revision} =~ /rhel5/ and $clientObj->{minor_revision} > 2)
           or ($clientObj->{major_revision} =~ /ol5/ and $clientObj->{minor_revision} >= 6)
           or $clientObj->{major_revision} =~ /rhel6|rhel7|ol6|sles11|fedora/) {
            $cmd = "lvdisplay ";
        } else {
            $cmd = "lvdisplay -D ";
        }
        $cmd .= "/dev/$vgname/$name | egrep 'LV Status|LV Size|Stripes|Block device'" .
               "| gawk '{print \$1,\$2,\$3,\$4,\$5} END {if (\$count==0) exit 1}'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error while running cmd ($cmd)");
            return $self;
        }

        $self->{state} = $self->_init_state(@output);
        return $self if ($self->error);
        $self->sanlog(-msg       => "  state: " . $self->{state},
                      -msg_level => 7);

        # Make the status of LV ready if its not available
        # We are doing this because we cant perform any action on it
        # if its status is unavailable (cant even be destroyed)
        if ($self->{state} =~ /not\s+available/i) {
            $cmd = "lvchange -ay /dev/$vgname/$name";
            my ($status, @lvchange_output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to run cmd ($cmd)");
                return $self;
            }
        }

        # Determine what filesystem (if any) is on
        # logical volume.
        $self->{filesystem} = $self->_init_fstype();
        return $self if ($self->error());
        $self->sanlog(-msg       => "  fs: " . $self->{filesystem},
                      -msg_level => 7);

        if ($self->{filesystem} !~ /raw/i) {
            $self->{mount} = $self->_init_mount($import);
            return $self if ($self->error);
            $self->sanlog(-msg => "  mnt: " . $self->{mount});

        } else {
            $self->{block_device} = $self->_init_block_device(@output);
            return $self if ($self->error);

            my $clientObj = $self->_get_client();

            $self->{raw_device} = $clientObj->_create_raw_binding($self->{block_device});
            if ($clientObj->error()) {
                $self->_set_err($clientObj->errmsg());
                return $self;
            }

            $self->sanlog(-msg => "else _new:raw device: " . $self->{raw_device});
        }

        $self->{size} = $self->_init_size(@output);
        return $self if ($self->error);
        $self->sanlog(-msg => "  size: " . $self->{size} . " mb");

        $self->{layout} = $self->_init_layout(@output);
        return $self if ($self->error);
        $self->sanlog(-msg => "  layout: " . $self->{layout});

    }

    return $self;
}

################################################################################
# Subroutine Name: _set_metadata
# Objective:      Set metadata for attribute
# Details:        See POD documentation at the end of this file
################################################################################
sub _set_metadata {
    my ($self, $attribute, $value) = @_;

    my ($cmd, $vgname, $name, @output, $status, $no_entry_exists);

    # This code will break if there are more than one instances writing to this file
    # In that case the last write will override all the previous entries
    $vgname = $self->get_vgname();
    $name   = $self->get_name();
    $no_entry_exists = 1;

    if ($self->{filesystem} =~ /reiserfs|gfs|xfs|btrfs/) {
        
        # for reiserfs there is no way of storing the mount point as metadata so 
        # storing it in /tmp/reiserfs.matadata. 
        $cmd = "touch /tmp/fs.metadata";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to touch file /tmp/fs.metadata");
            return $ClientUtil::FALSE;
        }

        $cmd = "rm -f /tmp/fs.metadata.tmp";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to remove file /tmp/fs.metadata.tmp");
            return $ClientUtil::FALSE;
        }

        $cmd = "touch /tmp/fs.metadata.tmp";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to create file /tmp/fs.metadata.tmp");
            return $ClientUtil::FALSE;
        }

        $cmd = "cat /tmp/fs.metadata";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to cat file /tmp/fs.metadata.tmp");
            return $ClientUtil::FALSE;
        }

        foreach my $line (@output) {
            if ($line =~ /$vgname\/$name/){
                $cmd = "echo $vgname\/$name:$value >> /tmp/fs.metadata.tmp";
                $no_entry_exists = 0; 
            } else {
                $cmd = "echo $line >> /tmp/fs.metadata.tmp";
            }
            my ($status, @output_tmp) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Unable to copy mount point $value into /tmp/fs.metadata");
                return $ClientUtil::FALSE;
            }
        }

        # Entry is not there
        if ($no_entry_exists == 1) {
            $cmd = "echo $vgname\/$name:$value >> /tmp/fs.metadata.tmp";
            my ($status, @output_tmp) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Unable to copy mount point $value into /tmp/fs.metadata.tmp");
                return $ClientUtil::FALSE;
            }
        }

        $cmd = "mv -f  /tmp/fs.metadata.tmp /tmp/fs.metadata";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to replace /tmp/fs.metadata by /tmp/fs.metadata.tmp");
            return $ClientUtil::FALSE;
        }
    } else {

        # Notes: This needs to be modifed to allow various modification
        # depending on attribute.  Currently hardcoded for label only.
        $cmd = "tune2fs ";
        if ($attribute =~ /mounted/i) {
            $cmd .= "-M $value ";
        } else {
            $self->_set_err("Attribute '$attribute' is currently", "not supported");
            return $ClientUtil::FALSE;
        }
        $cmd .= "/dev/" . $vgname . "/" . $name;
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to set volume \'$name\' metadata for",
                            "attribute '$attribute' in diskgroup '$vgname'", @output);
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

1;

