# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the SVM Logical Volume Manager on Solaris
## @author latif@netapp.com, rroberts@netapp.com
## @status private

# Package Name
##############
package San::SanClient::Lvm::Solaris::SVM;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;
use Data::Dumper;

require San::SanClient::Lvm::Solaris::SVM::Vg;

use vars qw(@ISA %hFormat);

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: initialize
# Objective:       initialize disk under svm control
# Details:         See POD documentation at the end of this file
################################################################################
sub initialize {
    my ($self, %args) = @_;

    my ($attribute, $cmd, $format, @already_inited, @output, $status, $devices);

    ### default disk type
    $format = "cdsdisk";

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: SVM::initialize() Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?format/) {
            if (not length($format = $args{$_}) and $format !~ /cdsdisk|sliced/i) {
                $self->_set_err("Error: SVM::initialize() Invalid value for argument '-format'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### check for required parameters
    unless (defined $devices and length $devices) {
        $self->_set_err("Error: SVM::initialize() -devices is a required argument");
        return $ClientUtil::FALSE;
    }

    ### Check to see if metadb exists
    $cmd = "metadb";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (scalar(@output) == 0 || grep(/no existing/, @output) != 0) {
        # Meta db does not exist, create it
        # First detect root volume
        $cmd = "df /";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        # Example output:
        # "/                (/dev/dsk/c1t0d0s0 ):51674752 blocks  3466272 files"
        my $root_dev;
        # Parse root dev
        if ($output[0] =~ /(\/dev\/dsk\/.+)\s*\)/){
            $root_dev = $1;
        } elsif ($output[0] =~ /rpool/){
            # First detect root volume
            $cmd = "zpool status rpool";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            foreach my $line (@output) {
            #zpool status rpool
            #  pool: rpool
            # state: ONLINE
            #  scan: none requested
            #config:
            #
            #        NAME        STATE     READ WRITE CKSUM
            #        rpool       ONLINE       0     0     0
            #          c3t0d0s0  ONLINE       0     0     0
                if($line =~ /^\s*(\S+)\s+\S+\s+\d\s+\d\s+\d/) {
                    $root_dev = $1;
                    if($root_dev !~ /rpool/i) {
                        $root_dev = '/dev/dsk/' . $root_dev;
                        last;
                    }
                }
            }
        } else {
            $self->_set_err("Error: SVM::initialize()- \"df \/\" output doesn't match known regexes");
            return $ClientUtil::FALSE;
        }
        $root_dev =~ s/\s*//g;
        $root_dev =~ s/dsk/rdsk/;
        $root_dev =~ s/s\d+$/s7/;

        $self->sanlog(-msg => "Root device is $root_dev.",
                      -msg_level => 3);
        # Make sure that the root dev has slice 7 and 7 is not Mounted to an incorrect directory
        $cmd = "prtvtoc -s $root_dev";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        #*                          First     Sector    Last
        #* Partition  Tag  Flags    Sector     Count    Sector  Mount Directory
        #       0      0    00    1060290 487299645 488359934
        #       2      5    01          0 488359935 488359934
        #       7      0    00      16065   1028160   1044224
        #       8      1    01          0     16065     16064

        # Grep for slice 7
        my $slice_7 = grep(/^\s*7\s+\d+/, @output);
        if ($slice_7 eq "" or $slice_7 =~ /^\s*$/) {
            $self->_set_err("Error: SVM::initialize()- slice 7 doesn't exist on $root_dev");
            return $ClientUtil::FALSE;
        }

        $slice_7 =~ /\s*7\s+\d+\s+\d+\s+\d+\s+\d+\s+\d+\s*(.*)/;
        if(defined ($1) and length($1)) {
            $self->_set_err("Error: SVM::initialize()- slice 7 on $root_dev " .
                            "has Mount Directory \'$1\' and should not");
            return $ClientUtil::FALSE;
        }

        # Create metadb using root dev
        $cmd = "metadb -a -c3 -f $root_dev";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        # Exit if the status of the 'metadb' output is not 0.
        if ($status > 0) {
            $self->_set_err("Error: SVM::initialize()- metadb failed, returned status- $status");
            return $ClientUtil::FALSE;
        }
    }

    ### Get a list of all the devices that are free (not in a metaset),
    ### online, and already formatted'.
    @already_inited = $self->get_free_devices();

    ### Initialize each device in the user specified devices array
    ### that is not aready inited (not in @already_inited array).
    foreach my $device (@$devices) {
        # initialized devices that have not been inited or are inited with
        # the wrong format
        if (not(grep /^$device$/, @already_inited) or
            $hFormat{$device} !~ /$format/i) {
            $self->sanlog(-msg => "$device init ($format)",
                          -msg_level => 3);

            ### Check if cdsdisk, if so re-add the volume label.  Current CDS
            ### bug that removes the label during initialization.
            if ($format =~ /cds/i) {
                ### Re-add the default volume name
                $self->{parent}->format(-device => [($device)],
                                        -label => "readd");
            }
        } else {
            $self->sanlog(-msg => "$device already inited ($format)",
                          -msg_level => 3);
        }

        ### Update the Format hash
        $self->sanlog(-msg => "Device '$device' is $format",
                      -msg_level => 3);
        $hFormat{$device} = $format;
    }

    return ($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: create_vg
# Objective:       create and return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vg {
    my ($self, %args) = @_;

    my ($vgObj, $bAdd,);

    $bAdd = 1;

    if (exists $args{-no_add}) {
        $self->sanlog(-msg => "LVM object will not manage VG object",
                      -msg_level => 3);
        delete $args{-no_add};
        $bAdd = 0;
    }

    $vgObj = San::SanClient::Lvm::Solaris::SVM::Vg->_new(-parent => $self,
                                                         -conn   => $self->{conn},
                                                         -loghdl => $self->{log},
                                                         %args);

    if ($vgObj->error) {
        $self->_set_err($vgObj->errmsg());
        return $vgObj;
    }

    if ($bAdd eq 1) {
        push @{$self->{vgs}}, $vgObj;
    } else {
        $vgObj = {};
    }

    return $vgObj;

}

################################################################################
# Subroutine Name: destroy_vg
# Objective:       destroy a named diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vg {
    my ($self, $name) = @_;

    my ($vgObj, @vgs,);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: SVM::destroy_vg() vg name is a required argument");
        return $ClientUtil::FALSE;
    }

    ### Delete the of vg object and update the parent vg list.
    foreach $vgObj (@{$self->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            $vgObj->destroy();
            if ($vgObj->error) {
                $self->_set_err($vgObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vgs, $vgObj;
        }
    }

    $self->{vgs} = \@vgs;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display
# Objective:       display disk group information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;

    my ($cmd, $status, @output,);

    $self->_clear_err();

    $self->sanlog(-msg => "Display diskgroup information",
                  -msg_level => 3);

    $cmd = "metastat -a";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::display() Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

}

###############################################################################
# Method:    get_disk_format
# Objective: Get list of supported disk formats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_format {
    my ($self, $disk) = @_;

    my ($format);

    $self->_clear_err();

    unless (defined $disk and length $disk) {
        $self->_set_err("Error: SVM::get_disk_format() disk name is a required argument");
        return "";
    }

    $format = $hFormat{$disk};

    unless (defined $format and length $format) {
        $self->_set_err("Error: SVM::get_disk_format() disk has not been initialized");
        return "";
    }

    ### If disk format is prefixed with 'auto:' remove it.
    $format =~ s/auto://g;

    return $format;

}

###############################################################################
# Method:    get_disk_formats
# Objective: Get list of supported disk formats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_formats {
    my ($self) = @_;

    my (@format);

    $self->_clear_err();

    push @format, qw(cdsdisk);

    return (wantarray()) ? @format : \@format;

}

###############################################################################
# Method:    get_format_default
# Objective: Get default type to use for disk format
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_format_default {
    my ($self) = @_;

    my ($format);

    $self->_clear_err();

    $format = "cdsdisk";

    return $format;

}

################################################################################
# Subroutine Name: get_vg
# Objective:       return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vg {
    my ($self, $name) = @_;

    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: SVM::get_vg() vg name is a required argument");
        return {};
    }

    foreach my $obj (@{$self->{vgs}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: SVM::get_vg() Unable to find diskgroup '$name'");

    return {};

}

################################################################################
# Subroutine Name: get_vgs
# Objective:       return array of volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgs {
    my ($self) = @_;

    my (@vgs);

    $self->_clear_err();

    if (defined $self->{vgs}) {
        @vgs = @{$self->{vgs}};
    } else {
        $self->_set_err("Error: SVM::get_vgs() SVM object has no volumes groups");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @vgs : \@vgs;

}

################################################################################
# Subroutine Name: get_vgnames
# Objective:       return array of volume group names
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgnames {
    my ($self) = @_;

    my ($cmd, $status, @output, @keep);

    $self->_clear_err();

    $cmd = "metastat -a";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::get_vgnames() Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    foreach my $line (@output) {
        #trial/d0: Concat/Stripe
        if ($line =~ /(\S+)\/\S+:\s+\S+\/\S+/) {
            push @keep, $1;
        }
    }

    @output = @keep;

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_pkginfo()
# Objective:       get detailed lvm pkg information
# Details:         See POD documentation at the end of this file
################################################################################
sub get_pkginfo {
    my ($self) = @_;

    my ($cmd, @output, $status,);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{pkginfo});

    return (wantarray()) ? @{$self->{pkginfo}} : $self->{pkginfo};

}

################################################################################
# Subroutine Name: _get_pkginfo
# Objective:       Get lvm package info
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self) = @_;

    my ($cmd, $name, $vers, $rev, @output, $status);

    $cmd = "/usr/bin/pkginfo -l SUNWmdr";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::_get_pkginfo() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        ($name) = $line =~ /PKGINST:\s+(\w+)/i if ($line =~ /PKGINST/i);
        ($vers, $rev) =
              $line =~ /VERSION:\s+(.*),REV=\b(.*)\b/i if ($line =~ /VERSION/i);
    }
    $self->{name}    = $name;
    $self->{vers}    = $vers;
    $self->{rev}     = $rev;
    $self->{pkginfo} = \@output;

}

################################################################################
# Subroutine Name: get_revision()
# Objective:       Get revision of SVM
# Details:         See POD documentation at the end of this file
################################################################################
sub get_revision {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{rev});

    return "" if ($self->error());

    return $self->{rev};

}

################################################################################
# Subroutine Name: get_name()
# Objective:       Get lvm name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    my ($cmd, @output, $status,);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{name});

    return "" if ($self->error());

    return $self->{name};

}

################################################################################
# Subroutine Name: get_version()
# Objective:       Check version of SVM
# Details:         See POD documentation at the end of this file
################################################################################
sub get_version {
    my ($self) = @_;

    my ($cmd, @output, $status,);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{vers});

    return "" if ($self->error());

    return $self->{vers};

}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;
    my ($alldgs_option, $bCnt, $cmd,     $cnt,     $device, $filer, $filers,
        $filername,     $lun,  $refresh, $status,  $total,  $vol, $protocol,
        %hDevs,         @attr, @buffer,  @devices, @output, @used, @sanlun_output);

    $total = 0;

    $self->_clear_err();

    ### Asssume user supplied -cnt argument
    $bCnt    = 1;
    $refresh = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Invalid value for argument '-cnt'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?inited_state/) {
            # NOOP for SVM
            next;
        } elsif (/-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Invalid value for argument '-filer'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
        } elsif (/-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vol'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Invalid value for argument '-lun'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Invalid value for argument '-refresh'");
                return ($ClientUtil::FALSE);
            }
            next;
        }
    }

    ### Check if user wants free devices from a specified filer,volume, or lun
    if (defined $filer or defined $vol or defined $lun or defined $protocol or defined $filers) {
        ### Check if a device to filer map 'dfmap' has already been created or
        ### refresh it if requested.
        if (scalar(keys %{$self->{dfmap}}) == 0 or $refresh == 1) {
            $self->sanlog(-msg => "Refreshing device to filer, vol and lun map",
                          -msg_level => 3);

            # Grab some sanlun output
            @sanlun_output = $self->{sanlun}->lun_show();
            if ($self->{sanlun}->error) {
                $self->_set_err($self->{sanlun}->errmsg());
                return (wantarray()) ? () : [] ;
            }

            # E.g. of output:
            # fas880-rtp02: /vol/nate_1_1/nate_sun280r-fak01_1        /dev/rdsk/c4t500A098192D0DFB6d5s2           lpfc0       FCP         1g (1073741824)     GOOD
            foreach my $line (@sanlun_output) {
                # Skip the header line
                next if ($line =~ /lun-pathname/i);

                $self->sanlog(-msg => "LINE=|$line|", -msg_level => 3);

                # Figure out what we are dealing with from sanlun
                my ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp);
                if ($line =~ /(\S+)\s+\/vol\/(\S+)\/(\S+)\s+\/dev\/rdsk\/(\S+)s2\s+\S+\s+(\S+)\s+/) {
                    # old sanlun kit output looks like below,
                    # controller:                 lun-pathname                                 device filename                     adapter   protocol          lun size         lun state
                    # iscsi_vs:  /vol/iscsi_vol_1/iscsi_auto-sunx2200-01_0  /dev/rdsk/c0t600A0980324668762D3F41734170386Fd0s2  scsi_vhci0  iSCSI          5g (5368709120)     GOOD
                    #
                    # new sanlun kit output looks like below,
                    # controller(7mode)/                                             device                                            host                  lun
                    # vserver(Cmode)       lun-pathname                              filename                                          adapter    protocol   size    mode
                    # ---------------------------------------------------------------------------------------------------------------------------------------------------
                    # iscsi_vs             /vol/iscsi_vol_1/iscsi_auto-sunx2200-01_0 /dev/rdsk/c0t600A0980324668762D3F41734170386Fd0s2 scsi_vhci0 iSCSI      5g      C
                    ($filer_tmp,$vol_tmp,$lun_tmp,$device_tmp,$protocol_tmp) = ($1,$2,$3,$4,lc($5));
                } elsif ($line =~ /\s{0}(\S+):\s+\/vol\/(\w+)\/(\S+)\s+\/dev\/rdsk\/(\w+)s2/) {
                    ($filer_tmp,$vol_tmp,$lun_tmp,$device_tmp,$protocol_tmp) = ($1,$2,$3,$4,"");
                } else {
                    next;
                }

                # Remove colon in the end of filer/vserver name if presents
                $filer_tmp =~ s/:$//;

                $self->sanlog(-msg => "Filer: $filer_tmp   Vol: $vol_tmp   " .
                                      "Device: $device_tmp   Lun: $lun_tmp",
                              -msg_level => 3);
                $hDevs{$device_tmp} = {
                                         filer    => $filer_tmp,
                                         vol      => $vol_tmp,
                                         lun      => $lun_tmp,
                                         protocol => $protocol_tmp,
                                      };
            }
            ### save map for next call of function
            $self->{dfmap} = \%hDevs;
        }
    }

    ### If not -cnt then assume all free disks
    ### under lvm control are wanted
    unless (defined $cnt and length($cnt)) {
        $cnt  = 10000;
        $bCnt = 0;
    }

    # Check to see if we already ran sanlun.  If not, then run it.
    unless (scalar(@sanlun_output) > 0) {
            # Grab some sanlun output
            @sanlun_output = $self->{sanlun}->lun_show();
            if ($self->{sanlun}->error) {
                $self->_set_err($self->{sanlun}->errmsg());
                return (wantarray()) ? () : [] ;
            }
    }

    # Run "metaset" so we can determine if the disk is already in use
    my $ms_cmd = "/usr/sbin/metaset";
    my ($ms_status, @ms_output) = &ClientUtil::execute($self->{conn}, $ms_cmd);
    if ($ms_status > 0) {
        $self->_set_err("Command ($ms_cmd) returned a bad status ($ms_status)");
        return (wantarray()) ? () : [];
    }

    # iterate through the devices and make sure that none of the disks
    # are being used by another LVM disk object.
    foreach my $line (@sanlun_output) {
        # Skip the header line
        next if ($line =~ /lun-pathname/i);

        my $format = "ufs";
        my $dev    = "";

        # fas880-rtp02: /vol/nate_1_1/nate_sun280r-fak01_1
        # /dev/rdsk/c4t500A098192D0DFB6d5s2           lpfc0       FCP
        # 1g (1073741824)     GOOD
        if ($line =~ /\/dev\/rdsk\/(\S+)s2\s+/) {
            $dev = $1;
        }

        # Apply filters to any of the free devices found
        if ($dev) {
            # See if the dev is used in a metaset
            # We are looking for something like this:
            # /dev/dsk/c3t60A9800043346163614A2F6B4E376D70d0   Yes
            if (grep(/\/dev\/dsk\/${dev}\s+/, @ms_output) > 0) {
                $self->sanlog(-msg => "SVM Skipping device '$dev' since it is used in a metaset",
                              -msg_level => 3);
                next;
            }

            if (defined $filer and length $filer) {
                $self->sanlog(-msg => "\nFilter filer: $filer\nDevice: $dev\nDevice filer: " .
                                      $self->{dfmap}{$dev}{filer},
                              -msg_level => 3);
                next if ($self->{dfmap}{$dev}{filer} !~ /^$filer$/i);
            }

            if (defined $vol and length $vol) {
                $self->sanlog(-msg => "\nFilter vol: $vol\nDevice: $dev\nDevice vol: " .
                                      $self->{dfmap}{$dev}{vol},
                              -msg_level => 3);
                next if ($self->{dfmap}{$dev}{vol} !~ /^$vol$/i);
            }

            if (defined $lun and length $lun) {
                $self->sanlog(-msg => "\nFilter lun: $lun\nDevice: $dev\nDevice lun: " .
                                      $self->{dfmap}{$dev}{lun},
                              -msg_level => 3);
                next if ($self->{dfmap}{$dev}{lun} !~ /$lun/i);
            }

            if (defined $protocol and length $protocol) {
                $self->sanlog(-msg => "\nFilter protocol: $protocol\nDevice: $dev\nDevice protocol: " .
                                      $self->{dfmap}{$dev}{protocol},
                              -msg_level => 3);
                next if ($self->{dfmap}{$dev}{protocol} !~ /$protocol/i);
            }

            # Filter on a filer list if provided by caller
            if (defined $filers and scalar(@{$filers}) != 0) {
                my $found = 0;
                # Look through each filer in the filers list
                foreach my $filer_f (@{$filers}) {
                    $self->sanlog(-msg => "\nFilter filer(s): $filer_f\nDevice: $dev\nDevice filer: " .
                                          $self->{dfmap}{$dev}{filer},
                                  -msg_level => 3);
                    # If filer matches devices filer, set found to 1
                    if ($self->{dfmap}{$dev}{filer} =~ /^$filer_f$/i) {
                        $found = 1;
                    }
                }

                # If the device is not part of the filers list caller requested skip it.
                if ($found == 0) {
                    next;
                }
            }

            ### Keep a hash of device formats
            $hFormat{$dev} = $format;

            # See if the lun has a volume label on it
            my $prt_cmd = "prtvtoc /dev/rdsk/" . $dev;
            if ($dev !~ /s2/) {
                $prt_cmd .= "s2";
            }
            my ($prt_status, @prt_output) = &ClientUtil::execute($self->{conn}, $prt_cmd);
            if ($prt_status > 0) {
                $self->_set_err("Command ($prt_cmd) returned a bad status ($prt_status)");
                return (wantarray()) ? () : [];
            }

            # Example output:
            # * /dev/rdsk/c4t60A98000486E544D575A4B58646A5774d0s2 (volume "NATE") partition map
            # Grep out just the line with the device in it (this has the words "partition map" in it)
            @prt_output = grep(/partition map/, @prt_output);

            # Check to see if the line contains volname of reserved types
            if (grep(/root|boot|reserved/, @prt_output)) {
                $self->sanlog(-msg => "Ignoring device because it has a special volume label ($dev)");
                next;
            }

            $self->sanlog(-msg => "SVM push $dev", -msg_level => 3);

            push @devices, $dev if (++$total <= $cnt);
        }
    }

    ### If user supplied cnt and free devices fell short
    ### generate error
    if ($bCnt eq 1 and scalar(@devices) < $cnt) {
        $self->_set_err("Only " . scalar(@devices) . "of $cnt requested devices were found");
    }

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: load_vg
# Objective:       load an existing diskgroup that is not in memory
# Details:         See POD documentation at the end of this file
################################################################################
sub load_vg {
    my ($self, $name) = @_;

    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: SVM::load_vg() Invalid or missing required paramter \$name");
        return {};
    }

    ### Check if vg has already been loaded
    $vgObj = $self->get_vg($name);
    if ($self->error) {
        # dg has not been loaded yet
        $self->_clear_err();

        # load dg
        $self->_create_vgs($name);
        return {} if ($self->error);
        $vgObj = $self->get_vg($name);
    } else {
        $self->sanlog(-msg => "INFO: SVM::load_vg() Volume group '$name' was already loaded", -msg_level => 1);
    }

    return $vgObj;

}
################################################################################
# Subroutine Name: unload_vg
# Objective:       Remove the vg object under LVM object control
# Details:         See POD documentation at the end of this file
################################################################################
sub unload_vg {
    my ($self, $name) = @_;

    my ($vgObj, $bFound, @vgs);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: SVM::unload_vg() Invalid or missing required paramter \$name");
        return $ClientUtil::FALSE;
    }

    $bFound = 0;

    foreach $vgObj (@{$self->{vgs}}) {
        ($vgObj->get_name() =~ /^$name$/)
          ? $bFound = 1
          : push @vgs, $vgObj;
    }

    $self->{vgs} = \@vgs;

    unless ($bFound) {
        $self->_set_err("Error: SVM::unload_vg() Nothing to unload. Vg name '" .
                        $name . "' was not loaded");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: enable_controllers
# Objective:       NOOP FOR SVM
# Details:         See POD documentation at the end of this file
################################################################################
sub enable_controllers {

    my ($self) = @_;

    # Unsupported function in SVM, NOOP
    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: rescan
# Objective:       NOOP FOR SVM
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {

    my ($self) = @_;

    # Unsupported function in SVM, NOOP
    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: restart_recovery_daemon
# Objective:       NOOP FOR SVM
# Details:         See POD documentation at the end of this file
################################################################################
sub restart_recovery_daemon {

    my ($self, $interval) = @_;

    # Unsupported function in SVM, NOOP
    return $ClientUtil::TRUE;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($conn, $log, $lvm, $loadcfg, $parent, $self,);

    $self = {};

    # creating lvm object for SVM
    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_})) {
                $self->_set_err("Error: SVM::_new() Invalid value for argument '-lvm'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: SVM::_new() Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?loghdl/) {
            if (not length($log = $args{$_})) {
                $self->_set_err("Error: SVM::_new() Invalid value for argument '-loghdl'");
                return $self;
            }
            next;
        } elsif (/-?loadcfg/) {
            if (not length($loadcfg = $args{$_}) or $loadcfg !~ /no|yes|1|0|off|on/i) {
                $loadcfg = 1;
            }
            next;
        }
    }

    ### check for existance of required -lvm argument
    if ($lvm !~ /svm/i) {
        $self->_set_err("Error: SVM::_new()",
                        "Invalid value for argument '-lvm' \($lvm\)");
        return $self;
    } elsif (not defined($lvm)) {
        $self->_set_err("Error: SVM::_new()",
                        "Missing required argument'-lvm'");
        return $self;
    }

    ### check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /Solaris/i) {
            $self->_set_err("Error: SVM::_new()",
                            "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: SVM::_new()",
                        "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn}   = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();

    $self->{vgs} = [];

    ### check to make sure that LVM is installed
    if (not $self->_is_installed()) {
        return $self if ($self->error);
    }

    ### create svm vg objects for existing disk groups
    if ($loadcfg) {
        $self->_create_vgs();
        return $self if ($self->error);

        ### update map to determine disk format
        $self->_update_disk_format_map();
        return $self if ($self->error);
    }

    # Returning LVM object created for SVM
    return ($self);

}

################################################################################
# Subroutine Name: _update_disk_format_map
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _update_disk_format_map {
    my ($self) = @_;

    my ($cmd, @output, $status,);

    ### Retrieve the volume groups.
    $cmd = "metastat -a";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 1) {
        $self->_set_err("Error: SVM::_update_disk_format_map()",
                        "Unable to update map");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        $line =~ s/:/ /;
        my ($device, $format) = $line =~ /^(\S+)s2\s+(\S+)/;
        $hFormat{$device} = $format;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _create_vgs
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vgs {
    my ($self, $vgname) = @_;

    my ($cmd, @output, @set_output, $status, $vgObj, $deported);

    ### Retrieve the volume groups.
    $cmd = "metaset|egrep \"Set name\"|awk '{print \$4}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 1) {
        $self->_set_err("Error: SVM::_create_vgs()",
                        "Unable to load existing diskgroup\(s\)");
        return $ClientUtil::FALSE;
    }

    ### Check to see if service is offline. Example:
    ### system/mdmonitor:default: service not online in SMF
    if (grep(/service not online/i, @output) > 0) {
        $self->_set_err("Error: SVM::_create_vgs()",
                        "Found service offline when running $cmd, please investigate");
        return $ClientUtil::FALSE;
    }

    # Remove extra stderr output from buffer
    @output = grep(!/there are no existing databases/i, @output);

    if ($vgname ne "") {
        @output = grep(/$vgname/, @output);
    }

    foreach my $line (@output) {
        $line =~ s/\,//g;

        $cmd = "metaset -s $line";
        ($status, @set_output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: SVM::_create_vgs() Unable to get owner information");
            return $ClientUtil::FALSE;;
        }

        # If we aren't the owner of the set, then it should be considered deported
        if (grep(/No \(auto\)/, @set_output)) {
            $deported = 1;
        } else {
            $deported = 0;
        }

        $self->create_vg(-name => $line, -deported => $deported);
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _get_disk_id
# Objective:       Return disk id
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_disk_id {
    my ($self, $device) = @_;

    my ($cmd, $name, @output, $status,);

    $self->_clear_err();

    $name = "/dev/rdsk/$device" . "s2";

    my $path = $self->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show -v -d $name | grep 'Serial number'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::_get_disk_id() Unable to get diskid for device \'$device\'", @output);
        return "";
    }

    return $output[0];

}

################################################################################
# Subroutine Name: _is_installed
# Objective:       Check to make sure that the LVM is installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_installed {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "/usr/bin/pkginfo SUNWmdr";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 0) {
        $self->sanlog(-msg => "lvm is installed", -msg_level => 3);
        return $ClientUtil::TRUE;
    } else {
        $self->sanlog(-msg => "lvm is not installed", -msg_level => 3);
        $self->_set_err("Error: SVM::_is_installed() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }
}

1;
