# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the SVM Volume Groups on Solaris
## @authors latif@netapp.com, rroberts@netapp.com
## @status private
## @pod ../../Vg.pod

package San::SanClient::Lvm::Solaris::SVM::Vg;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

require San::SanClient::Lvm::Solaris::SVM::Vg::Vol;

use Tie::IxHash;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $conn, $devices, $disk_cnt, $parent, $log, $name);
    my ($self, $status, $deported, @output, %hDevice);

    $self = {};

    bless $self, $class;

    $self->_clear_err();

    $deported = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::new()", "Invalid value for argument '-devices'");
                return $self;
            }
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::new()", "Missing value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::new()", "Missing value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::new()", "Missing value for argument '-lvmobj'");
                return $self;
            }
            next;
        } elsif (/-?deported/) {
            if (not length($deported = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::new()", "Missing value for argument '-deported'");
                return $self;
            }
            next;
        } 
    }

    ### check for existance of required -lvm argument
    unless (defined $parent) {
        $self->_set_err("Error: SVM::Vg::new()", 
                        "Missing required argument '-parent'");
        return $self;
    }

    ### check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: SVM::Vg::new()", 
                        "Missing required argument '-name'");
        return $self;
    }

    ### check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: SVM::Vg::new()", 
                            "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: SVM::Vg::new()", 
                        "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn}     = $conn;
    $self->{name}     = $name;
    $self->{deported} = $deported;
    $self->{disk_cnt} = 0;
    $self->{parent}   = $parent;

    # create the diskgroup if devices were supplied
    if (defined $devices and ref($devices) eq "ARRAY") {
        $self->sanlog(-msg => "Creating new VG named ($name) with devices ($devices)", -msg_level => 3);
        $self->_create($name, $devices);
        if ($self->error()) {
            return $self;
        }
    } else {
        # vg must already exist. set device list
        # for all existing and deported dgs.
        $self->sanlog(-msg => "loading existing volume group '$name'", 
                      -msg_level => 3);
        $self->{devices} = $self->_get_devices();
        return $self if ($self->error);
        unless ($self->{deported}) {
            $self->_create_vols();
            return $self if ($self->error);
        }
    }

    return $self;

}

################################################################################
# Subroutine Name: _create_vols
# Objective:       create vol object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vols {
    my ($self, %args) = @_;

    my ($cmd, $import, $name, $fsck, $status, $vgObj, $volObj);
    my (%hDeport, @output);

    ### Identify volumes that are being imported so
    ### that their automatically mounted with their
    ### last filesystem mount point.
    $import =
      (exists $args{-_import})
      ? $args{-_import}
      : 0;

    $fsck =
      (exists $args{-fsck})
      ? $args{-fsck}
      : 0;

    $name = $self->get_name();

    $cmd = "metastat -s $name";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::_create_vols()",
                        "Unable to load existing volumes in disk group",
                        "status=$status", @output);
        return $ClientUtil::FALSE;
    }

    my @tmp_vol_array;
    foreach my $line (@output) {
        if ($line =~ /^$name/) {
            push @tmp_vol_array, $line;
            next;
        }
    }

    foreach my $line (@tmp_vol_array) {
        if ($line =~ /^$name\/(\w+\d+)\:/) {
            # n_vg/d5: Concat/Stripe
            $cmd = "metastat -s $name $1|head -n2";
            ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
            if ($status gt 0) {
                $self->_set_err("Error: SVM::Vg::_create_vols()",
                                "Unable to load existing volumes in disk group",
                                "status=$status", @output);

                return $ClientUtil::FALSE;
            }

            # E.g of output
            # trial/d0: Concat/Stripe Size: 6248448 blocks (3.0 GB)
            foreach my $line (join(" ", @output)) {
                my ($vol_name, undef, undef, $size, undef) = 
                                                            split(/\s+/, $line);
                $size = $size * 512 / (1024 * 1024);
                $vol_name =~ s/^\///;
                $vol_name =~ s/\://;
                $volObj = $self->create_vol(-name    => $vol_name,
                                            -size    => $size,
                                            -exist   => 1,
                                            -_import => $import,
                                            -fsck    => $fsck);
                return $ClientUtil::FALSE if ($self->error);
            }
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: create_vol
# Objective:       create and return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vol {
    my ($self, %args) = @_;

    my ($volObj);

    ### Change layout to concat if dg contains only 1 device
    if (exists $args{-layout} and length $args{-layout}) {
        if ($args{-layout} =~ /stripe/i and $self->get_device_cnt() eq "1") {
            $args{-layout} = "CONCAT";
        }
    }

    $volObj = San::SanClient::Lvm::Solaris::SVM::Vg::Vol->_new(-parent => $self,
                                                               -conn   => $self->{conn},
                                                               -loghdl => $self->{log},
                                                               %args);

    if ($volObj->error) {
        $self->_set_err($volObj->errmsg());
        return $self;
    }

    push @{$self->{vols}}, $volObj;

    return $volObj;

}

################################################################################
# Subroutine Name: destroy_vol
# Objective:       destroy a named vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vol {
    my ($self, $name) = @_;

    my ($volObj, @vols,);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: SVM::Vg::destroy_vol() volname is a required argument");
        return $ClientUtil::FALSE;
    }

    ### Delete the of vg object and update the parent vg list.
    foreach $volObj (@{$self->{vols}}) {
        if ($volObj->get_name() =~ /^$name$/) {
            $volObj->destroy(-force => 1);
            if ($volObj->error) {
                $self->_set_err($volObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vols, $volObj;
        }
    }

    $self->{vols} = \@vols;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _get_devices
# Objective:       get list of devices used by disk group
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices {
    my ($self) = @_;

    my ($cmd, $name, $deported, $status, @output, %hDevice, $disk_cnt);

    $name     = $self->get_name();
    $deported = $self->{deported};

    $cmd = "metaset -s $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::_get_devices() Unable to get devices for volgroup \'$name\'", @output);
        return {};
    }

    # E.g of output:-
    #
    # Solaris 10:
    # Device                                           Reloc  Device ID
    # /dev/dsk/c3t60A980004F6444524A34456A78324C7Ad0   Yes    id1,sd@n60a980004f6444524a34456a78324c7a
    #
    # Solaris 11:
    # Drive                                   Dbase
    # c0t60A98000486E61366D346D3655587231d0   Yes
    #
    $disk_cnt = 0;
    foreach my $line (@output) {
        if ($line =~ /^(\/dev\/dsk\/\S+)\s+/) {
            $self->{$1} = $1;
            $hDevice{$1} = $1;
            $disk_cnt++;
        } elsif ($line =~ /^(c\d+t\S+)\s+/) {
            my $ctd = $1;
            my $disk = "/dev/dsk/" . $ctd;
            $self->{$disk} = $disk;
            $hDevice{$disk} = $disk;
            $disk_cnt++;
        }
    }

    $self->{disk_cnt} = $disk_cnt;

    return \%hDevice;
}

################################################################################
# Subroutine Name: _create()
# Objective:       create a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self, $name, $devices) = @_;

    my ($attribute, $cmd, $disk_cnt, $disk_name, $format, @output);
    my ($status, %hDevice);

    $disk_cnt  = 0;
    $attribute = "";

    $format = $self->{parent}->get_disk_format((@$devices)[0]);
    $attribute = "cds=off" if ($format !~ /cds/);

    # Check to see if VG already exists and if it does return
    #metastat: sunx2100-shu04: setname "nate_vga": no such set
    $cmd = "metastat -s $name 2>&1";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if (grep(/no such set/, @output) == 0) {
        foreach my $device (@$devices) {
            $device =~ s/s\d+$//;
            $self->{$device} = $format;
            $hDevice{$device} = $name;
        }

        $self->{devices}  = \%hDevice;
        $self->{disk_cnt} = scalar(@$devices);
        return $ClientUtil::TRUE;
    }

    # should really get hostname from the host instead
    # of from NATE.
    $cmd = "/bin/hostname";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::_create() Unable to execute $cmd");
        return $ClientUtil::FALSE;
    }
    my $hostname = $output[0];

    # Create the metaset
    $cmd = "metaset -s $name -a -A enable -h $hostname";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::_create() Unable to create (metaset) diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    my $dev_list = "";
    foreach my $device (@$devices) {
        $dev_list .= $device . " ";
        $device =~ s/s\d+$//;

        # determine if add worked
        if (!$self->add_device($device)) {
            $self->_set_err("Error: SVM::Vg::_create() Unable to create (diskgroup \'$name\'", @output);
            return $ClientUtil::FALSE;
        }

        $self->{$device} = $format;
        $hDevice{$device} = $name;
    }

    $self->{devices}  = \%hDevice;
    $self->{disk_cnt} = $disk_cnt;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: add_device()
# Objective:       add a device to the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub add_device {
    my ($self, $device) = @_;

    my ($cmd, $disk_cnt, $disk_name, $name, @output, $status);

    $self->_clear_err();

    ### Get name of volume group.
    $name = $self->get_name();

    $device =~ s/s\d+$//;

    $cmd = "metaset -s $name -a $device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::add_device() Unable to add $device to diskgroup \'$name\'", @output);

        return $ClientUtil::FALSE;
    }

    $self->{devices}->{$device} = $device;
    $self->{disk_cnt} = $disk_cnt;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: remove_device()
# Objective:       remove a device from the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub remove_device {
    my ($self, %args) = @_;

    my ($cmd, $device, $force, $disk_name, $name, @output);
    my (@vols_mount, $status);

    $self->_clear_err();

    $force = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::remove_device()", "Invalid value for argument '-device'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::remove_device()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            $force = "-k";
            next;
        }
    }

    ### get name of colume group
    $name = $self->get_name();

    ### verify the device is being used in volume group
    unless (exists $self->{devices}->{$device}) {
        $self->_set_err("Error: SVM::Vg::remove_device()", "Device '$device' doesn't exist in diskgroup");
        return $ClientUtil::FALSE;
    }

    $disk_name = $self->{devices}->{$device};

    ### if no force option is specified umount all volumes so deport
    ### can occur without error.  Make note of which vols were
    ### mounted so they can be remounted after removing device
    if (not defined $force) {
        my @volObjs = $self->get_vols();
        return $ClientUtil::FALSE if ($self->error);

        foreach my $volObj (@volObjs) {

            if ($volObj->is_fs()) {
                if ($volObj->is_mounted()) {
                    $self->sanlog(-msg => "unmounting filesystem on volume " . 
                                          $volObj->get_name(), 
                                  -msg_level => 3);
                    $volObj->unmount();
                    if ($volObj->error) {
                        $self->_set_err($volObj->errmsg());
                        return $ClientUtil::FALSE;
                    }

                    push @vols_mount, $volObj;
                }
            }
        }
    }

    $cmd = "metaset -s $name -d $device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::remove_device() Unable to remove $device from diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    ### remove device from hash
    delete($self->{devices}->{$device});

    ### decrement device counter;
    $self->{disk_cnt}--;

    ### Restart and mount volumes that were previously mounted.
    foreach my $vol (@vols_mount) {
        $vol->mount();
        if ($vol->error) {
            $self->_set_err($vol->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: rename()
# Objective:       rename volume group - THIS IS A NOOP OP FOR SVM
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, %args) = @_;

    my ($cmd, $force, $name, $newname, @output, $status, @vols_mount);

    $self->_clear_err();

    $self->_set_err("Error: SVM::Vg::rename()", 
                    "Unsupported function for SVM::Vg.pm");
    return $ClientUtil::FALSE;

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($newname = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::rename()", "Invalid value for argument '-name'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::rename()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### get name of disk group
    $name = $self->get_name();

    ### verify the device is being used in volume group
    unless (defined $name) {
        $self->_set_err("Error: SVM::Vg::rename()", 
                        "Volume group doesn't have a name");
        return $ClientUtil::FALSE;
    }

    ### verify the device is being used in volume group
    unless (defined $newname) {
        $self->_set_err("Error: SVM::Vg::rename()", 
                        "New name for Volume group not specified");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Rename volume group" . 
                          "Old Name: " . $self->get_name() . 
                          "New Name: " . $newname);

    ### if force option specified umount all volumes so deport
    ### can occur without error.  Make note of which vols were
    ### mounted so they can be remounted after rename occurs
    if (defined $force and $force =~ /1|on|y/i) {
        my @volObjs = $self->get_vols();
        return $ClientUtil::FALSE if ($self->error);

        foreach my $volObj (@volObjs) {
            if ($volObj->is_fs()) {
                if ($volObj->is_mounted()) {
                    $self->sanlog(-msg => "unmounting filesystem on volume " . 
                                          $volObj->get_name(), 
                                  -msg_level => 1);
                    $volObj->unmount();
                    if ($volObj->error) {
                        $self->_set_err($volObj->errmsg());
                        return $ClientUtil::FALSE;
                    }
                    push @vols_mount, $volObj;
                }
            }
        }
    }

    ### Deport the diskgroup
    $self->deport();
    return $ClientUtil::FALSE if ($self->error);

    ### Rename the diskgroup
    $cmd = "metarename -s $name $name $newname";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::rename() Unable to import '$name' under a new diskgroup \'$newname\'",
                        @output);

        return $ClientUtil::FALSE;
    }

    $self->{name} = $newname;

    $self->{deported} = 0;

    ### Restart and mount volumes that were previously mounted.
    foreach my $vol (@vols_mount) {
        $vol->mount();
        if ($vol->error) {
            $self->_set_err($vol->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, @vgs);

    $self->_clear_err();

    $name = $self->{name};

    unless (defined $name) {
        $self->_set_err("Error: SVM::Vg::destroy() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    if ($self->{deported}) {
        # We are not the owner of the metaset, to destroy we can only purge
        $cmd = "metaset -s $name -P";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: SVM::Vg::destroy() Unable to destroy diskgroup \'$name\'", @output);
            return $ClientUtil::FALSE;
        }
    } else {
        $cmd = "metaclear -s $name -a";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: SVM::Vg::destroy() Unable to destroy diskgroup \'$name\'", @output);
            return $ClientUtil::FALSE;
        }

        foreach my $device ($self->get_devices()) {
            $cmd = "metaset -s $name -f -d $device";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status gt 0) {
                $self->_set_err("Error: SVM::Vg::destroy() Unable to destroy diskgroup \'$name\'", @output);
                return $ClientUtil::FALSE;
            }
        }

        # should really get hostname from the host instead
        # of from NATE.
        $cmd = "/bin/hostname";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: SVM::Vg::_create() Unable to execute $cmd");
            return $ClientUtil::FALSE;
        }
        my $hostname = $output[0];

        $cmd = "metaset -s $name -d -h $hostname";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: SVM::Vg::destroy() Unable to destroy diskgroup \'$name\'", @output);
            return $ClientUtil::FALSE;
        }
    }

    ### Remove the vgobj from the parent lvm object
    foreach my $vgObj (@{$self->{parent}->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vgs, $vgObj;
        }
    }

    $self->{parent}->{vgs} = \@vgs;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;

    my (@verbose, $cmd, $name, @output, $status, $header, $summary);
    my (@summary, $verbose);

    $self->_clear_err();

    $verbose = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::display()", "Invalid value for argument '-verbose'");
                return $self;
            }
            next;
        }
    }

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: SVM::Vg::display() Diskgroup name is a required parameter");
        return (wantarray()) ? () : [];
    }

    push @summary, "VG Name:  " . $self->get_name();
    push @summary, "Device Cnt:  " . $self->get_device_cnt();

    foreach my $device ($self->get_devices()) {
        push @summary, "Device:  " . $device . " (" . $self->{parent}->get_disk_format($device) . ")";
    }

    $header = sprintf("Vol Summary:\n"
                        . "%-12.12s %-12.12s %-7.7s %-5.5s %-12.12s %-32.32s\n"
                        . "%-12.12s %-12.12s %-7.7s %-5.5s %-12.12s %-32.32s",
                      "VGROUP", "VOLUME", "LAYOUT", "TYPE",  "SIZE(mb)", "MOUNT_PT",
                      "=" x 12, "=" x 12, "=" x 7,  "=" x 5, "=" x 12,   "=" x 32);

    push @summary, $header;

    foreach my $volObj (@{$self->{vols}}) {
        $summary = sprintf("%-12.12s %-12.12s %-7.7s %-5.5s %-12.12s %-s",
                           $self->get_name(), $volObj->get_name(), $volObj->get_layout(),
                           $volObj->get_fs(), $volObj->get_size(), $volObj->get_mount());

        push @summary, $summary;
    }

    if (defined $verbose and $verbose =~ /1|on|y|yes|t|true/) {

        $cmd = "metastat -s $name";

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: SVM::Vg::display() Unable to display info for diskgroup \'$name\'", @output);
            return (wantarray()) ? () : [];
        }

        @verbose = (@summary, "\nVgroup Detail:", @output);

        push @verbose, "\n\n";
        return (wantarray()) ? @verbose : \@verbose;
    } else {
        push @summary, "\n\n";
        return (wantarray()) ? @summary : \@summary;
    }

}

################################################################################
# Subroutine Name: get_device_cnt
# Objective:       Return number of devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device_cnt {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{disk_cnt}) {
        return $self->{disk_cnt};
    } else {
        $self->_set_err("Error: SVM::Vg::get_device_cnt() Unable to display disk count for diskgroup");
        return 0;
    }
}

################################################################################
# Subroutine Name: get_devices
# Objective:       Return devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;

    my (@output);

    $self->_clear_err();

    @output = (keys %{$self->{devices}});

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vg object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: SVM::Vg::get_name() Vg object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vol
# Objective:       return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol {
    my ($self, $name) = @_;

    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: SVM::Vg::get_vol() volume name is a required argument");
        return $vgObj;
    }

    foreach my $obj (@{$self->{vols}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: SVM::Vg::get_vol() Unable to find volume '$name'");

    return $vgObj;

}

################################################################################
# Subroutine Name: get_vols
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vols {
    my ($self) = @_;

    my (@volObjs);

    $self->_clear_err();
    if (defined $self->{vols}) {
        @volObjs = @{$self->{vols}};
    } else {
        $self->_set_err("Error: SVM::Vg::get_vols() Vg object has no volumes");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @volObjs : \@volObjs;

}

################################################################################
# Subroutine Name: get_vol_cnt
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol_cnt {
    my ($self) = @_;

    $self->_clear_err();
    if (defined $self->{vols}) {
        return scalar(@{$self->{vols}});
    } else {
        $self->_set_err("Error: SVM::Vg::get_vol_cnt() Vg object has no vol count");
        return 0;
    }
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size, $sectors);

    $self->_clear_err();

    $size    = 0;
    $sectors = 0;

    $name = $self->get_name();

    $cmd = "metastat -s $name|grep Size:";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::get_size() Unable to get size (mb) for diskgroup \'$name\'", @output);
        return 0;
    }

    $output[0] =~ /Size:\s+(\d+)\s+blocks/;
    $sectors = $1;

    $size = $sectors * 512;    #bytes
    $size = $size / 1024;      #kilobytes
    $size = $size / 1024;      #megabytes

    return $size;

}

################################################################################
# Subroutine Name: get_freespace
# Objective:       Return available free space
# Details:         See POD documentation at the end of this file
################################################################################
sub get_freespace {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size, $sectors);

    $self->_clear_err();

    $size    = 0;
    $sectors = 0;

    $name = $self->get_name();

    $cmd = "metastat -s $name|grep Size:";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::get_size() Unable to get size (mb) for diskgroup \'$name\'", @output);
        return 0;
    }

    $output[0] =~ /Size:\s+(\d+)\s+blocks/;
    $sectors = $1;

    $size = $sectors * 512;    #bytes
    $size = $size / 1024;      #kilobytes
    $size = $size / 1024;      #megabytes

    return $size;

}

################################################################################
# Subroutine Name: is_deported
# Objective:       Determine if volgroup has been deported
# Details:         See POD documentation at the end of this file
################################################################################
sub is_deported {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{deported}) {
        return $self->{deported};
    } else {
        return 0;
    }

}

################################################################################
# Subroutine Name: deport()
# Objective:       deport Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub deport {
    my ($self) = @_;

    my ($cmd, $name, @output, $status,);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: SVM::Vg::deport() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    $cmd = "metadetach -s $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::deport() Unable to deport diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    ### Remove volume objects from vg cache
    $self->{vols} = ();

    $self->{deported} = 1;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: import()
# Objective:       Import Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub import {
    my ($self, %args) = @_;

    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: SVM::Vg::import() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    $cmd = "metaimport -s $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::import() Unable to import diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    $self->{deported} = 0;

    ### requery device list
    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    $cmd = "metaonline -s $name";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status gt 0) {
        $self->_set_err("Error: SVM::Vg::import() Unable to start up volumes in diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    $self->_create_vols(-_import => 1, %args);

    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: refresh
# Objective:       refresh content of volume group.
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh {
    my ($self) = @_;

    my ($cmd, $status, @output,);

    $self->_clear_err();

    ### Remove existing vol cache
    $self->{vols} = undef;

    $self->sanlog(-msg => "refreshing volume group '" . $self->get_name() . "'",
                  -msg_level => 3);

    ### Determine if dg is deported
    $cmd = "metastat -a|grep '(" . $self->get_name() . ")'";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    # determine if devfsadm command worked
    if ($status gt 1) {
        $self->_set_err("Error: SVM::Vg::refresh()", 
                        "Unable to load deported diskgroups");

        return $ClientUtil::FALSE;
    }

    $self->{deported} = 1 if (@output);

    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    if (not $self->is_deported()) {
        $self->_create_vols();
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;

}

1;
