# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the SVM logical volumes
## @authors latif@netapp.com, rroberts@netapp.com
## @status private
## @pod ../../../Vol.pod

package San::SanClient::Lvm::Solaris::SVM::Vg::Vol;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use Data::Dumper;
use File::Basename;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
# Method:    create_fs
# Objective: create a filesystem on a volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_fs {
    my ($self, %args) = @_;

    my ($cmd, $type, $vgname, $name, $status, @output);

    $vgname = $self->{parent}->get_name();

    $name = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?type$/) {
            $type = $args{$_};
            next;
        }
    }

    ### check for existance of required -type argument, if not
    ### set default to UFS
    if (not defined($type) or $type eq "") {
        $type = "ufs";
    }

    unless (defined $type) {
        $self->_set_err("Error: SVM::Vg::Vol::create_fs()", 
                        "Missing required argument '-type'");
        return $ClientUtil::FALSE;
    }

    if ($type =~ /ufs/) {
        $cmd = "echo 'y' > /tmp/yes";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: SVM::Vg::Vol::create_fs Unable to create $type filesystem on volume '$name'",
                            @output);

            return $ClientUtil::FALSE;
        }

        $cmd = "/usr/sbin/newfs " . $self->get_raw_device() . " </tmp/yes";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: SVM::Vg::Vol::create_fs Unable to create $type filesystem on volume '$name'",
                            @output);
            return $ClientUtil::FALSE;
        }
    }

    $self->{filesystem} = $type;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self, %args) = @_;

    my ($cmd, $force, $dgname, $name, @output, $status, @vols);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::destroy()", 
                                "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $dgname = $self->{parent}->get_name();
    $name   = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: SVM::Vg::Vol::destroy() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    ### if force option specified umount all volumes so deport
    ### can occur without error.  Make note of which vols were
    ### mounted so they can be remounted after rename occurs
    if (defined $force) {
        $self->sanlog(-msg => "determine if volume has filesystem", 
                      -msg_level => 3);
        if ($self->is_fs()) {
            $self->sanlog(-msg => "determine if volume is mounted", 
                          -msg_level => 3);
            if ($self->is_mounted()) {
                $self->sanlog(-msg => "unmount filesystem " . $self->get_name(),
                              -msg_level => 3);
                $self->unmount();
                return $ClientUtil::FALSE if ($self->error());
            }
        }
    }

    $cmd = "metaclear -s $dgname $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::destroy() Unable to destroy volume \'$name\' in diskgroup '$dgname'",
                        $cmd, @output);
        return $ClientUtil::FALSE;
    }

    ### Remove the volume from the parent vol group object
    foreach my $volObj (@{$self->{parent}->{vols}}) {
        if ($volObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vols, $volObj;
        }
    }

    $self->{parent}->{vols} = \@vols;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;

    my (@buf, $cmd, $vgname, $name, @output, $status, $header, $summary);

    $self->_clear_err();

    $vgname = $self->{parent}->get_name();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: SVM::Vg::Vol::display() volume name is a required parameter");
        return (wantarray()) ? () : [];
    }

    $cmd = "metastat -s $vgname $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err(
               "Error: SVM::Vg::Vol::display() Unable to display info for volume \'$name\' in volume group \'$vgname\'",
               @output);
        return (wantarray()) ? () : [];
    }

    $header = sprintf("\nVol Summary:\n"
                        . "%-12.12s %-7.7s %-5.5s %-12.12s %-s\n"
                        . "%-12.12s %-7.7s %-5.5s %-12.12s %-32.32s",
                      "NAME",   "LAYOUT", "TYPE",  "SIZE(mb)", "MOUNT_PT",
                      "=" x 12, "=" x 7,  "=" x 5, "=" x 12,   "=" x 32);

    $summary = sprintf("%-12.12s %-7.7s %-5.5s %-12.12s %-s",
                       $self->get_name(), $self->get_layout(), $self->get_fs(),
                       $self->get_size(), $self->get_mount());

    @buf = ($header, $summary, "\nVol Detail:", @output);

    return (wantarray()) ? @buf : \@buf;

}

################################################################################
# Subroutine Name: get_fs
# Objective:       Return Vol fs type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{filesystem} and length $self->{filesystem}) {
        return $self->{filesystem};
    } else {
        $self->_set_err("Error: SVM::Vg::Vol::get_fs() Vol object is missing filesystem");
        return "";
    }

}

################################################################################
# Subroutine Name: get_layout
# Objective:       Return volume layout
# Details:         See POD documentation at the end of this file
################################################################################
sub get_layout {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{layout}) {
        return $self->{layout};
    } else {
        $self->_set_err("Error: SVM::Vg::Vol::get_layout() Vol object is missing layout");
        return "";
    }

}

################################################################################
# Subroutine Name: get_mount
# Objective:       Return mount point of volume
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{mount} and length $self->{mount}) {
        return $self->{mount};
    } else {
        $self->_set_err("Error: SVM::Vg::Vol::get_mount() Vol filesystem is not mounted");
        return "";
    }
}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vol object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: SVM::Vg::Vol::get_name() Vol object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return ooject size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{size}) {
        return $self->{size};
    } else {
        $self->_set_err("Error: SVM::Vg::Vol::get_size() Vol object is missing size");
        return "";
    }
}

################################################################################
# Subroutine Name: get_state
# Objective:       Return object state
# Details:         See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{state}) {
        return $self->{state};
    } else {
        $self->_set_err("Error: SVM::Vg::Vol::get_state() Vol object is missing state");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vgname
# Objective:       Return Vol object diskgroup name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgname {
    my ($self) = @_;

    my ($name);

    $self->_clear_err();

    $name = $self->{parent}->get_name();

    if (defined $name and length $name) {
        return $name;
    } else {
        $self->_set_err("Error: SVM::Vg::Vol::get_vgname() Vol object is missing parent name");
        return "";
    }

}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       Return path to raw vol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    $self->_clear_err();

    ### get name of disk group
    my $vgname = $self->{parent}->get_name();

    ### get name of volume
    my $name = $self->get_name();

    return "/dev/md/$vgname/rdsk/$name";

}

################################################################################
# Subroutine Name: is_fs
# Objective:       Determine if volume has a filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub is_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{filesystem} and $self->{filesystem} !~ /raw/i) {
        return 1;
    } else {
        return 0;
    }

}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if volume is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{mount} and length $self->{mount}) {
        return 1;
    } else {
        return 0;
    }

}

################################################################################
# Subroutine Name: mount
# Objective:       mount filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub mount {
    my ($self, %args) = @_;

    my ($cmd, $directio, $fsck, $type, $vgname, $mount, $name);
    my ($options, $pass_thru, $status, @output);

    ### Get attributes
    $vgname = $self->{parent}->get_name();
    $name   = $self->get_name();
    $type   = $self->get_fs();
    $mount  = undef;
    $pass_thru  = "";

    ### check if volume is unformatted
    if (defined $type and $type =~ /raw/) {
        $self->_set_err("Error: SVM::Vg::Vol::mount()", 
                        "Cannot mount a raw filesystem");
        return $ClientUtil::FALSE;
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?path$/) {
            if (not length($mount = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::mount()", "Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?directio$/) {
            if (not length($directio = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::mount()", "Invalid value for argument '-directio'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?fsck$/) {
            if (not length($fsck = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::mount()", "Invalid value for argument '-fsck'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?pass_thru$/) {
            $pass_thru = $args{$_};
            $pass_thru =~ s/^\s+//;
            $pass_thru =~ s/\s+$//;
            if ($pass_thru ne "") {
                $pass_thru = "," . $pass_thru;
            }
        }
    }

    ### check if the volume is already mounted
    if ($self->is_mounted()) {
        $self->_set_err("Error: SVM::Vg::Vol::mount() volume '$name' is already mounted");
        return $ClientUtil::FALSE;
    }

    ### Use default path if not defined
    if (not defined $mount) {
        $mount = "/mnt/qa/$vgname/$name";
        $self->sanlog(-msg => "using default mount pt '$mount'", 
                      -msg_level => 3);
    } else {
        $mount .= "/$vgname/$name";
        $self->sanlog(-msg => "using mount pt '$mount'", 
                      -msg_level => 3);
    }

    $self->{mount} = $mount;

    ### Create the mount point and open up dir privs
    $cmd = "mkdir -p $mount; chmod 777 $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::mount() Unable to create mount point directory for volume '$name'",
                        @output);
        return $ClientUtil::FALSE;
    }

    if ($fsck) {
        $cmd = "fsck -F " . $self->get_fs() . " -y /dev/md/$vgname/rdsk/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: SVM::Vg::Vol::mount() Unable to fsck volume filesystem", @output);
            return $ClientUtil::FALSE;
        }
    }

    ### Mount the volume's filesystem
    $options = "largefiles";
    if ($directio) {
        $options .= ",forcedirectio";
    }

    $cmd = "/sbin/mount -F $type -o " . $options . $pass_thru . " /dev/md/$vgname/dsk/$name $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::mount() Unable to mount volume filesystem", @output);
        return $ClientUtil::FALSE;
    }

    ### Get handle to client object
    $self->{client} = $self->{parent}{parent}{parent};

    # Comment out for now, can cause system to fail reboot when stale entries 
    # are in vftsb file
    #$self->{client}->mount_entry_add(-device_mount => "/dev/md/$vgname/dsk/$name",
    #                                 -device_fsck  => "/dev/md/$vgname/rdsk/$name",
    #                                 -mount_pt     => $mount,
    #                                 -fs_type      => $self->get_fs());
    #if ($self->{client}->error) {
    #    $self->_set_err($self->{client}->errmsg());
    #    return $ClientUtil::FALSE;
    #}

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: rename()
# Objective:       rename volume
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, $newname) = @_;

    my ($cmd, $name, $vgname, @output, $status);

    $self->_clear_err();

    ### get name of disk group
    $vgname = $self->{parent}->get_name();

    ### get name of volume
    $name = $self->get_name();

    ### verify the device is being used in volume group
    unless (defined $name) {
        $self->_set_err("Error: SVM::Vg::Vol::rename()", 
                        "Volume doesn't have a name");
        return $ClientUtil::FALSE;
    }

    ### verify the device is being used in volume group
    unless (defined $newname) {
        $self->_set_err("Error: SVM::Vg::Vol::rename()", 
                        "New name for Volume not specified");
        return $ClientUtil::FALSE;
    }

    ### Rename the volume
    $cmd = "metarename -s $vgname $name $newname";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::rename Unable to rename volume '$name' to \'$newname\'", @output);
        return $ClientUtil::FALSE;
    }

    $self->{name} = $newname;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: rescan()
# Objective:       rescan volume
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;

    my ($mount);

    ### Currently just handles remounting.
    ### Needed after snapdrive restore
    $self->_clear_err();

    $mount = $self->_init_mount();
    if (defined $mount and length $mount) {
        $self->mount(-path => $mount);
    } else {
        $self->{mount} = undef;
    }

    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unmount()
# Objective:       unmount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount {
    my ($self, %args) = @_;

    my ($cmd, $type, $vgname, $fs, $mount, $name, $status, $force);
    my ($option, @output);

    $self->_clear_err();

    $option = "";

    # parse hash
    foreach (keys %args) {
        if (/-?force$/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::unmount()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $option = "-f" if (defined $force);

    ### Get attributes
    $vgname = $self->{parent}->get_name();
    $name   = $self->get_name();
    $mount  = $self->get_mount();
    $fs     = $self->get_fs();

    ### check if volume contains a filesystem

    unless ($fs !~ /raw/i) {
        $self->_set_err("Error: SVM::Vg::Vol::unmount()", 
                        "Cannot unmount a raw device");
        return $ClientUtil::FALSE;
    }

    ### check if volume is mounted
    unless (defined $mount and length($mount)) {
        $self->_set_err("Error: SVM::Vg::Vol::unmount()", 
                        "Cannot unmount a filesystem that is not mounted");
        return $ClientUtil::FALSE;
    }

    ### Unmount the volume's filesystem
    $cmd = "/usr/sbin/umount $option $mount";
    ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd, 180, 2);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::unmount() Unable to unmount volume's filesystem", @output);
        return $ClientUtil::FALSE;
    }

    ### Remove mount pt metadata from object
    $self->{mount} = "";

    ### Clear vfstab mount entry
    ### client = lvm <-- vg <-- vol
    $self->{client} = $self->{parent}{parent}{parent};

    # Comment out for now, can cause system to fail reboot when stale entries 
    # are in vftsb file
    #$self->{client}->mount_entry_delete($mount);
    #if ($self->{client}->error) {
    #    $self->_set_error($self->{client}->errmsg());
    #    return $ClientUtil::FALSE;
    #}

    ### Remove directory head
    my ($base, $parent) = fileparse($mount);
    $cmd = "rm -rf $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::unmount() Unable to remove mount point '$mount'", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
#
# P R I V A T E   S U B R O U T I N E S
#
###############################################################################
################################################################################
# Subroutine Name: _create()
# Objective:       create a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self) = @_;

    my ($cmd, $device_cnt, $dg, $name, $layout, $size, @output, $status);
    my ($no_of_vols, $stripe_size);
    my ($num_dev_per_vol, $device_list_per_vol);

    $dg                  = $self->{parent}->get_name();
    $device_cnt          = $self->{parent}->get_device_cnt();
    $name                = $self->{name};
    $size                = $self->{size};
    $layout              = $self->{layout};
    $no_of_vols          = $self->{no_of_vols};
    $num_dev_per_vol     = $self->{num_dev_per_vol};
    $device_list_per_vol = $self->{dev_per_vol_list};

    # Setting the default stripe size
    $stripe_size = "64k";

    if (not defined($no_of_vols)) {
        $no_of_vols = 1;
    }

    my $vol_name = $self->{name};

    # Creating volume with the parameters specific to SVM, passed from the 
    # thpl.

    # To get around the long command limitation, we use `cat /tmp/file` to 
    # input the device list to the metainit command.
    # Have to clear out the temp file IF it exists, this keeps us from 
    # having old data in file causing us issues later on down the road.

    $cmd = "printf \"\" >/tmp/nate_" . $vol_name;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::_create() Unable to create volume - $vol_name", "CMD: $cmd", @output);
        return $ClientUtil::FALSE;
    }

    # Pushing the list of devices to be included per volume in the temp file.
    $cmd = "printf \"$device_list_per_vol \" >>/tmp/nate_" . $vol_name;
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::_create() Unable to create volume ", "CMD: $cmd", @output);
        return $ClientUtil::FALSE;
    }

    # Execute the metainit command for the volume object passed by the thpl 
    # using `cat /tmp/nate_x`, this will get us past the long command limitation
    $cmd =
      "metainit -s $dg " . $vol_name . " 1 " . $num_dev_per_vol . " `cat /tmp/nate_" . $vol_name . "` -i $stripe_size";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::_create() Unable to create volume \'" . 
                        $vol_name . "\'", "CMD: $cmd", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _get_metadata
# Objective:      Retrieve metadata for attribute
# Details:        See POD documentation at the end of this file
################################################################################
sub _get_metadata {
    my ($self, $attribute) = @_;

    my ($cmd, $vgname, $name, @output, $status, $value);

    $value = "";

    $vgname = $self->{parent}->get_name();
    $name   = $self->get_name();

    $cmd = "metastat -s $vgname $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($self->error) {
        $self->_set_err("Error: SVM::Vg::Vol::_get_metadata() ",
                        "Unable to retrieve volume \'$name\' metadata for",
                        "attribute '$attribute' in diskgroup '$vgname'", 
                        @output);
        return "";
    }

    ### return the attribute value if exists.
    if (defined $output[0]) {
        ($value) = $output[0] =~ /$attribute:\s+(.*)/;
    }

    return $value;

}

################################################################################
# Subroutine Name: _init_fstype
# Objective:      Initialize fstype for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_fstype {
    my ($self) = @_;

    my ($cmd, $fstype, $vgname, $name, @output, $status);

    $vgname = $self->{parent}->get_name();
    $name   = $self->get_name();

    ### Start volume if not enabled
    if ($self->get_state() =~ /disable/i) {
        $self->_vol_start();
        return "";
    }

    $cmd = "fstyp /dev/md/$vgname/dsk/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status eq 1) {
        # Check if it is perhaps a broken filesystem
        $cmd = "mount | grep /dev/md/$vgname/dsk/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if (grep(m"/dev/md/$vgname/dsk/$name", @output)) {
            $fstype = "";
        } else {
            $fstype = "raw";
        }
    } elsif ($status > 1) {
        $self->_set_err("Error: SVM::Vg::Vol::_init_fstype() " .
                        "Unable to retrieve  filesystem type for volume \'$name\' in diskgroup '$vgname'",
                        @output);
        $fstype = "";
    } else {
        ($fstype) = $output[0] =~ /(\w+)/i;
    }

    return $fstype;

}

################################################################################
# Subroutine Name: _init_layout
# Objective:      Initialize layout of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_layout {
    my ($self) = @_;

    my ($cmd, $vgname, $layout, $name, @output, $status);

    $layout = "";

    $vgname = $self->{parent}->get_name();
    $name   = $self->get_name();

    # don't exist if this returns > 0, since if volume has single
    # device, interlace will not appear and this really is not a 
    # problem, just means that this not a stripe.
    $cmd = "metastat -s $vgname $name |grep interlace";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 0) {
        # get layout value
        ($layout, undef) = split(/\s+/, $output[0]);
    } else {
        $layout = "concat";
    }

    return lc($layout);

}

################################################################################
# Subroutine Name: _init_mount
# Objective:      Initialize mount point for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_mount {
    my ($self, $import, $fsck) = @_;

    my ($cmd, $vgname, $mount, $name, $path, @output, $status);

    $mount = "";

    $vgname = $self->{parent}->get_name();
    $name   = $self->get_name();

    ### Check to see if mount path has been store
    ### in volume metadata under 'comment' attribute.
    ### This is to done in order to associate a mount
    ### pt with a vg/vol in case the vol was renames while
    ### still mounted.

    $cmd = "df -k | /usr/xpg4/bin/grep -e \'$vgname/$name$\'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    
    if ($status > 0) {
        ### If volumes are bing imported assume then
        ### automatically mount fs with last stored
        ### mount point (or default path) if a mount
        ### point does not exist
        if ($import) {
            $mount = $self->_get_metadata("comment");
            if (defined $mount and length($mount)) {
                ($path, $vgname, $name) = $mount =~ /(\S+)\/(\w+)\/(\w+)$/;
            }
            $self->mount(-path => $path, -fsck => $fsck);
            return "" if $self->error();
        }
    } else {
        foreach my $line (@output) {
            if ($line =~ /\s+\d+\%\s+(\S+)/) {
                $mount = $1;
                last;
            }
        }
    }

    return $mount;
}

################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self) = @_;

    my ($cmd, $vgname, $size, $name, @output, $sectors, $status);

    $size = "";

    $vgname = $self->{parent}->get_name();
    $name   = $self->get_name();

    $cmd = "metastat -s $vgname $name |grep Size:";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::_init_size() " .
                        "Unable to retrieve  size (MB) for volume \'$name\' in diskgroup '$vgname'",
                        @output);
        return "";
    }

    ### convert sectors to megabytes

    (undef, $sectors) = split(/\s+/, $output[0]);

    $size = $sectors * 512;    #bytes
    $size = $size / 1024;      #kilobytes
    $size = $size / 1024;      #megabytes

    return $size;

}

################################################################################
# Subroutine Name: _init_state
# Objective:      Initialize state of existing host volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_state {
    my ($self) = @_;

    my ($cmd, $name, $state, $vgname, $status, @output);

    $state = "";

    $vgname = $self->{parent}->get_name();
    $name   = $self->get_name();

    $cmd = "metastat -s $vgname";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: SVM::Vg::Vol::_init_state() " .
                        "Unable to retrieve state for volume \'$name\' in diskgroup '$vgname'",
                        @output);
        return "";
    }

    if (scalar(@output) == 0) {
        $state = "offline";
    } else {
        $state = "online";
    }

    return $state;

}

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $exist, $fsck, $import, $name, $size, $conn, $layout);
    my ($log, $parent, $self, $status, @output, $no_of_vols);
    my ($num_dev_per_vol, $device_list_per_vol);

    $self = {};

    bless $self, $class;

    #default layout is concatenation
    $layout = "concat";
    $import = 0;
    $exist  = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?layout$/) {
            if (not length($layout = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Invalid value for argument '-layout'");
                return $self;
            }
            next;
        } elsif (/-?_import$/) {    ### private member
            if (not length($import = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Invalid value for argument '-_import'");
                return $self;
            }
            next;
        } elsif (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Invalid value for argument '-name' ($name)");
                return $self;
            }
            next;
        } elsif (/-?size$/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Invalid value for argument '-size'");
                return $self;
            }
            next;
        } elsif (/-?fsck$/) {
            if (not length($fsck = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Invalid value for argument '-fsck'");
                return $self;
            }
            next;
        } elsif (/-?exist$/) {
            if (not length($exist = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Invalid value for argument '-exist'");
                return $self;
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Missing value for argument '-parent'");
                return $self;
            }
            next;

            # Checking the arguement for no of volumes specified
        } elsif (/-?num_of_vols$/) {
            if (not length($no_of_vols = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Missing value for arguement '-num_of_vols'");
            }
        } elsif (/-?num_dev_per_vol$/) {
            if (not length($num_dev_per_vol = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Missing value for arguement '-num_dev_per_vol'");
            }
        } elsif (/-?dev_per_vol_list$/) {
            if (not length($device_list_per_vol = $args{$_})) {
                $self->_set_err("Error: SVM::Vg::Vol::_new()", "Missing value for arguement '-dev_per_vol_list'");
            }
        }
    }

    ### check for existance of required -lvm argument
    unless (defined $parent) {
        $self->_set_err("Error: SVM::Vg::Vol::_new()", 
                        "Missing required argument '-parent'");
        return $self;
    }

    ### check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: SVM::Vg::Vol::_new()", 
                        "Missing required argument '-name'");
        return $self;
    }

    ### check for existance of required -name argument
    unless (defined $size) {
        $self->_set_err("Error: SVM::Vg::Vol::_new()", 
                        "Missing required argument '-size'");
        return $self;
    }

    ### check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: SVM::Vg::Vol::_new()", 
                            "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: SVM::Vg::Vol::_new()", 
                        "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn}             = $conn;
    $self->{name}             = $name;
    $self->{layout}           = $layout;
    $self->{size}             = $size;
    $self->{parent}           = $parent;
    $self->{state}            = 'ENABLED';
    $self->{filesystem}       = 'raw';
    $self->{no_of_vols}       = $no_of_vols;
    $self->{num_dev_per_vol}  = $num_dev_per_vol;
    $self->{dev_per_vol_list} = $device_list_per_vol;

    # Map the name to an SVM usable volume name
    # All names must start with a "d" and end with
    # a digit (up to 3 digits long)
    if ($self->{name} =~ /^d\s+/) {
        $self->sanlog(-msg => "SVM Vol::_new() Correct vg name given, no conversion needed", 
                      -msg_level => 3);
    } else {
        # Search and replace all chars non digit and remove them
        $self->{name} =~ s/\D+//g;

        # Prepend "d" to the name
        if ($self->{name} !~ /\d+/ || $self->{name} eq "") {
            # Checking the list of existing volumes
            # and finding the unused volume name
            my $cmd = "metastat -a";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

            if ($status > 0) {
                $self->_set_err("Error: SVM::Vg::Vol::_create() Unable to check existing volumes", "CMD: $cmd", @output);
                return $self;
            }

            # Creating a loop to find the next available volume name
            my $vol_num  = 0;
            my $vol_name = "d" . $vol_num;
            my $done     = 0;
            my $dg = $self->{parent}->get_name();

            while (!$done) {
                my $vol_exists = 0;
                foreach my $line (@output) {
                    if ($line =~ /$dg\/$vol_name/) {
                        $vol_exists = 1;
                    }
                }
                if ($vol_exists eq 1) {
                    $vol_num++;
                    $vol_name = "d" . $vol_num;
                } else {
                    $done = 1;
                }
            }

            $self->{name} = $vol_name;
        } else {
            $self->{name} = "d" . $self->{name};
        }
    }

    if (not $exist) {
        $self->_create();
        return $self if ($self->error());
    } else {
        $self->sanlog(-msg => "loading existing volume '$name'", 
                      -msg_level => 3);

        $self->{state} = $self->_init_state();
        return $self if ($self->error);
        $self->sanlog(-msg => "  state: " . $self->{state}, 
                      -msg_level => 3);

        $self->{filesystem} = $self->_init_fstype();
        $self->sanlog(-msg => "  fs: " . $self->{filesystem}, 
                      -msg_level => 3);
        if ($self->error()) {
            ### allow vol to be loaded although it
            ### can't be started to determine fs
            if ($self->{filesystem} eq "") {
                $self->_clear_err();
            } else {
                return $self;
            }
        }

        if ($self->{filesystem} !~ /raw/i) {
            $self->{mount} = $self->_init_mount($import, $fsck);
            return $self if ($self->error);
            $self->sanlog(-msg => "  mnt: " . $self->{mount}, 
                          -msg_level => 3);
        }

        $self->{size} = $self->_init_size();
        return $self if ($self->error);
        $self->sanlog(-msg => "  size: " . $self->{size} . " mb", 
                      -msg_level => 3);

        $self->{layout} = $self->_init_layout();
        return $self if ($self->error);
        $self->sanlog(-msg => "  layout: " . $self->{layout}, 
                      -msg_level => 3);
    }

    return $self;

}

1;
