# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the Vxvm Logical Volume Manager on Solaris
## @author knopp@netapp.com, latif@netapp.com
## @status private
## @pod ../../Lvm.pm

# Package Name
##############
package San::SanClient::Lvm::Solaris::Vxvm;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanClient::Lvm;
use San::SanLog;
use Error;
use Data::Dumper;

require San::SanClient::Lvm::Solaris::Vxvm::Vg;

use vars qw(@ISA %hFormat);

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: initialize
# Objective:       initialize disk under veritas control
# Details:         See POD documentation at the end of this file
################################################################################
sub initialize {
    my ($self, %args) = @_;

    my ($attribute, $cmd, $format, @already_inited, @output);
    my ($status, $devices);

    ### default disk type
    $format = "cdsdisk";

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?format/) {
            if (not length($format = $args{$_}) and $format !~ /cdsdisk|sliced/i) {
                $self->_set_err("Invalid value for argument '-format'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### check for required parameters
    unless(defined $devices and length $devices) {
        $self->_set_err("Missing required argument '-devices'");
        return $ClientUtil::FALSE;
    }

    # add rescan - needed on some system after format and before vxdisksetup
    $self->rescan();
    return $ClientUtil::FALSE if ($self->error());

    ### Get a list of all the devices that are free (not in a dg),
    ### online, and already formatted'.  "online thinrclm" is a new online format
    ### for 5.1 and later.  thinrclm means the lun supports thin reclamation.
    @already_inited = $self->get_free_devices(-inited_states => ["online", "online thinrclm"]);

    if ($self->get_version() =~ /^(4|5|6)/) {
        if (not defined $format) {
            $format="cdsdisk";
        }
        $attribute = "format=" . lc($format);
    } elsif ($self->get_version() =~ /^3\.5/) {
        $format = "sliced";
        $attribute = "";
    }

    ### Initialize each device in the user specified devices array
    ### that is not already inited (not in @already_inited array).
    foreach my $device (@$devices) {
        # remove s2 from the end so this will match output from get_free_devices
        # also vxdisksetup device must be of the format cCtTdD, mcCtTdD or <enclosurename>_<diskno>
        $device =~ s/s\d+$//;

        # initialized devices that have not been inited or are inited with
        # the wrong format
        if (not (grep /^$device$/, @already_inited) or
            $hFormat{$device} !~ /$format/i) {

            $self->sanlog(-msg => "$device init ($format)");

            $cmd = "/etc/vx/bin/vxdisksetup -f -i $device $attribute";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)",
                                "Veritas Device setup (initialize) failed.",
                                "OUTPUT = @output");
                return($ClientUtil::FALSE);
            }
        } else {
            $self->sanlog(-msg => "$device already inited ($format)");
        }

        ### Update the Format hash
        $self->sanlog(-msg => "Device '$device' is $format");
        $hFormat{$device} = $format;
    }

    return($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: create_vg
# Objective:       create and return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vg {
    my ($self, %args) = @_;
    my ($vgObj, $bAdd);

    $bAdd = 1;

    if (exists $args{-no_add}) {
        $self->sanlog(-msg => "LVM object will not manage VG object");
        delete $args{-no_add};
        $bAdd = 0;
    }

    $vgObj = San::SanClient::Lvm::Solaris::Vxvm::Vg->_new(-parent=>$self,
                                                          -conn=>$self->{conn},
                                                          %args);

    if ($vgObj->error) {
        $self->_set_err($vgObj->errmsg());
        return $self;
    }

    if ($bAdd == 1) {
        push @{$self->{vgs}}, $vgObj;
    } else {
        $vgObj = {};
    }

    return $vgObj;

}

################################################################################
# Subroutine Name: destroy_vg
# Objective:       destroy a named diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vg {
    my ($self, $name) = @_;

    my ($vgObj, @vgs);

    $self->_clear_err();

    unless(defined $name and length $name) {
        $self->_set_err("vg name is a required argument");
        return $ClientUtil::FALSE
    }

    ### Delete the of vg object and update the parent vg list.
    foreach $vgObj (@{$self->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            $vgObj->destroy();
            if ($vgObj->error) {
                $self->_set_err($vgObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vgs,$vgObj;
        }
    }

    $self->{vgs} = \@vgs;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display
# Objective:       display disk group information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    $self->sanlog(-msg => "Display diskgroup information");

    $cmd = "vxprint";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

}

###############################################################################
# Method:    get_disk_format
# Objective: Get list of supported disk formats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_format {
    my ($self,$disk) = @_;

    my ($format);

    $self->_clear_err();

    unless(defined $disk and length $disk) {
        $self->_set_err("Disk name is a required argument");
        return $ClientUtil::FALSE ;
    }

    $format = $hFormat{$disk};

    unless(defined $format and length $format) {
        $self->_set_err("Disk has not been initialized");
        return $ClientUtil::FALSE ;
    }

    ### If disk format is prefixed with 'auto:' remove it.
    $format =~ s/auto://g;

    return $format;

}

###############################################################################
# Method:    get_disk_formats
# Objective: Get list of supported disk formats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_formats {
    my ($self) = @_;

    my (@format);

    $self->_clear_err();

    if ($self->get_version() =~ /^(4|5|6)/) {
        push @format, qw(cdsdisk sliced);
    } else {
        push @format, qw(sliced);
    }

    return (wantarray()) ? @format : \@format;

}

###############################################################################
# Method:    get_format_default
# Objective: Get default type to use for disk format
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_format_default {
    my ($self) = @_;

    my ($format);

    $self->_clear_err();

    if ($self->get_version() =~ /^(4|5|6)/) {
        $format = "cdsdisk";
    } else {
        $format = "sliced";
    }

    return $format;

}

################################################################################
# Subroutine Name: get_vg
# Objective:       return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vg {
    my ($self, $name) = @_;

    my ($vgObj);

    $self->_clear_err();

    unless(defined $name and length $name) {
        $self->_set_err("Vg name is a required argument");
        return $ClientUtil::FALSE ;
    }

    foreach my $obj (@{$self->{vgs}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Unable to find diskgroup '$name'");
    return $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: get_vgs
# Objective:       return array of volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgs {
    my ($self) = @_;
    my (@vgs);

    $self->_clear_err();

    if (defined $self->{vgs}) {
        @vgs = @{$self->{vgs}};
    } else {
        $self->_set_err("Vxvm object has no volumes groups");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @vgs : \@vgs;

}

################################################################################
# Subroutine Name: get_vgnames
# Objective:       return array of volume group names
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgnames {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    $cmd = "vxdg list | egrep -v 'NAME|root|boot|reserved' | awk '{print \$1}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_pkginfo()
# Objective:       get detailed lvm pkg information
# Details:         See POD documentation at the end of this file
################################################################################
sub get_pkginfo {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{pkginfo});

    return (wantarray()) ? @{$self->{pkginfo}} : $self->{pkginfo};

}

################################################################################
# Subroutine Name: _get_pkginfo
# Objective:       Get lvm package info
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self) = @_;

    my ($mp, $sp, $rp, $cmd, $name, $vers, $rev, @output, $status);

    $cmd = "/usr/bin/pkginfo -l VRTSvxvm";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        ($name) = $line =~ /PKGINST:\s+(\w+)/i
              if ($line =~ /PKGINST/i);
        ($vers,$rev) = $line =~ /VERSION:\s+(.*),REV=\b(.*)\b/i
              if ($line =~ /VERSION/i);
        if ($line =~ /PSTAMP/i) {
            # PSTAMP Output can look like this:
            #
            # 6.0
            # PSTAMP:  6.0.000.000-6.0-2011-08-01
            #
            # 5.1SP1
            # PSTAMP:  5.1.100.000-5.1SP1-2010-10-10
            #
            # 5.1 P1
            # PSTAMP:  Veritas-5.1-P1.1:2009-11-24
            #
            # 5.1 RP1
            # PSTAMP:  5.1.001.000-5.1RP1_x86-2010-02-23-VERITAS-VM
            # PSTAMP:  5.1.001.000-5.1RP1-2010-02-23-VERITAS-VM
            #
            # 5.0 MP3
            # PSTAMP:  VERITAS-5.0_x86_MP3RP2HF7:2009-09-10
            # PSTAMP:  Veritas-5.0_MP3_RP2HF7:2009-09-11
            #
            # 5.0 MP1
            # PSTAMP:  Veritas-5.0_MP1_RP3.2:2007-08-28

            # We need to normalize the PSTAMP line
            my $line_temp = $line;

            # Remove PSTAMP
            $line_temp =~ s/^\s*PSTAMP:\s+//g;

            # new "long" style versioning of D.D.DDD.DDD followed by dash or underscore
            $line_temp =~ s/^\s*(Veritas|\d+\.\d+\.\d+\.\d+)(\-|\_)//gi;

            # Remove any _x86 or -x86
            $line_temp =~ s/(\-|\_)x86//g;

            # Remove all date stamps
            $line_temp =~ s/\d\d\d\d\-\d\d\-\d\d//g;

            # Remove all remaining underscores and dashes
            $line_temp =~ s/\-|\_//g;

            # That should leave us with the following:
            # 5.0MP3RP2HF7:
            # 5.1RP1VERITASVM
            # 5.1P1.1:
            # 5.0MP1RP3.2:
            # 5.1SP1

            # Look up MP or P info
            (undef, $mp) = $line_temp =~ /^(\d+\.\d+)(MP\d+|P\d+).*/i;

            # Look for RP info
            (undef, $rp) = $line_temp =~ /^(\d+\.\d+)(RP\d+).*/i;

            # Look for SP info
            (undef, $sp) = $line_temp =~ /^(\d+\.\d+)(SP\d+).*/i;
        }
    }

    $self->{name}    = $name;
    $self->{vers}    = $vers;
    $self->{rev}     = $rev;
    $self->{mp}      = $mp;
    $self->{sp}      = $sp;
    $self->{rp}      = $rp;
    $self->{pkginfo} = \@output;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_revision()
# Objective:       Get revision of Veritas
# Details:         See POD documentation at the end of this file
################################################################################
sub get_revision {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{rev});
    return "" if ($self->error());

    return $self->{rev};

}

################################################################################
# Subroutine Name: get_name()
# Objective:       Get lvm name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{name});
    return "" if ($self->error());

    return $self->{name};

}

################################################################################
# Subroutine Name: get_version()
# Objective:       Check version of Veritas
# Details:         See POD documentation at the end of this file
################################################################################
sub get_version {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{vers});
    return "" if ($self->error());

    return $self->{vers};

}

################################################################################
# Subroutine Name: _get_version_mp()
# Objective:       Check version of Veritas MP revision
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_version_mp {
    my ($self) = @_;

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{mp});
    return "" if ($self->error());

    return $self->{mp};
}

################################################################################
# Subroutine Name: _get_version_rp()
# Objective:       Check version of Veritas RP revision
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_version_rp {
    my ($self) = @_;

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{rp});
    return "" if ($self->error());

    return $self->{rp};
}

################################################################################
# Subroutine Name: _get_version_sp()
# Objective:       Check version of Veritas SP revision
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_version_sp {
    my ($self) = @_;

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{sp});
    return "" if ($self->error());

    return $self->{sp};
}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;
    my ($bCnt, $cmd, $cnt, $device, $filer, $filername, $lun,
        $refresh, $status, $total, $type, $vol, %hDevs, @all_devices,
        @attr, @devices, @lvm_devices, @output, @used,  $filers, $protocol,
        @inited_states, $inited_state);

    $type  = ".*";
    $total = 0;

    $self->_clear_err();

    # Asssume user supplied -cnt argument
    $bCnt    = 1;
    $refresh = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [];
            }
            delete($args{$_});
            next;
        } elsif (/-?inited_states/) {
            if (not scalar(@inited_states = @{$args{$_}})) {
                $self->_set_err("Invalid value for argument '-inited_states'");
                return (wantarray()) ? () : [];
            }
            delete($args{$_});
            next;
        } elsif (/-?inited_state$/) {
            if (not length($inited_state = $args{$_})) {
                $self->_set_err("Invalid value for argument '-inited_state'");
                return (wantarray()) ? () : [];
            }
            delete($args{$_});
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Invalid value for argument '-type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-vol'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    if (scalar(@inited_states) != 0) {
        my @temp_devices = ();
        foreach my $inited_state (@inited_states) {
            $self->sanlog(-msg => "Getting LUNs with state '$inited_state'");
            push @temp_devices, $self->get_free_devices(-inited_state => "$inited_state", %args);
            if ($self->error) {
                return (wantarray()) ? () : [];
            }
        }
        if (defined($cnt) and $cnt > 0) {
            # Check to make sure we got enough devices back.  It is possible we got back
            # more than we need so handle it if we did
            if (scalar(@temp_devices) < $cnt) {
                $self->_set_err("Only " . scalar(@temp_devices) . " of $cnt requested devices were found");
                return (wantarray()) ? () : [];
            } elsif (scalar(@temp_devices) > $cnt) {
                my @temp_devices2 = ();
                for (my $index = 0; $index < $cnt; $index++) {
                    push(@temp_devices2, $temp_devices[$index]);
                }
                @temp_devices = @temp_devices2;
            }
        }
        return (wantarray()) ? @temp_devices : \@temp_devices;
    }

    # add rescan - needed on some system after format
    $self->rescan();
    if ($self->error()) {
        return (wantarray()) ? () : [];
    }

    # Check if user wants free devices from a specified filer,volume, or lun
    my $vxdisk_hash = {};
    if (not defined($self->{dfmap}) or not defined($self->{raw_vxdisk_list_output})
        or not defined($self->{raw_sanlun_output}) or $refresh == 1) {
        $self->sanlog(-msg=>"creating translation maps:");
        $refresh = 1;

        # Grab the vxdisk -o alldgs list output
        # DEVICE       TYPE            DISK         GROUP        STATUS
        # das_fcp_0_0  auto            -            -            error
        # das_fcp_0_1  auto            -            -            error
        # das_fcp_0_2  auto            -            -            error
        # das_fcp_0_3  auto            -            -            error
        # das_iscs0_0  auto            -            -            error
        # das_iscs0_1  auto            -            -            error
        # das_iscs0_2  auto            -            -            error
        # das_iscs0_3  auto            -            -            error
        # sda          auto:none       -            -            online invalid
        my @vxdisk_list_output;
        $cmd = "/usr/sbin/vxdisk -o alldgs list";
        ($status, @vxdisk_list_output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return (wantarray()) ? () : [];
        }

        # Store the data
        $self->{raw_vxdisk_list_output} = \@vxdisk_list_output;

        # Build a relationship hash keyed by device that looks like this:
        # vxdisk_hash => {device} => {
        #                               dmpnode    = "node_name",
        #                               primary    = "0|1",
        #                               init_state = "online"
        #                            }
        foreach my $line (sort @vxdisk_list_output) {
            $line =~ s/^\s+//g;
            next if ($line =~ /^DEVICE/);

            # Skip any devices flagged as boot, root, or reserved
            next if ($line =~ /boot|root|reserved/i);

            my ($dmpnode, $type, $disk, $diskgroup, @init_state) = split(/\s+/, $line);
            # Update $diskgroup to remove any ()
            $diskgroup =~ s/\(|\)//;

            # Skip any disks who have a disk group defined that isn't "-";
            if ($diskgroup ne "-") {
                $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since the it is used in diskgroup '$diskgroup'");
                next;
            }

            # Make sure type is "auto"
            unless ($type =~ /auto/) {
                $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since the type isn't 'auto'");
                next;
            }

            # Now that we have the dmpnode, run vxdisk list on that node.
            # and find out primary/secondary status
            my @vxdisk_list_node_output;
            $cmd = "/usr/sbin/vxdisk list " . $dmpnode;
            ($status, @vxdisk_list_node_output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return (wantarray()) ? () : [];
            }

            # Output should be
            # [root@auto-rhel5u6-02 ~]# vxdisk list das_fcp_0_0
            # Device:    das_fcp_0_0
            # devicetag: das_fcp_0_0
            # type:      auto
            # flags:     error private autoconfig
            # pubpaths:  block=/dev/vx/dmp/das_fcp_0_0 char=/dev/vx/rdmp/das_fcp_0_0
            # guid:      -
            # udid:      NETAPP%5FLUN%20C-Mode%5F2FiIn%3F-lCptj%5F2FiIn%3F-lCptm
            # site:      -
            # Multipathing information:
            # numpaths:   4
            # sdj             state=enabled   type=primary
            # sdn             state=enabled   type=secondary
            # sdb             state=enabled   type=secondary
            # sdf             state=enabled   type=primary
            foreach my $line (@vxdisk_list_node_output) {
                chomp($line);
                if ($line =~ /^udid\:\s+(.+)/) {
                    # Make sure this is a netapp disk
                    my $udid = $1;
                    unless ($udid =~ /netapp/i) {
                        $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since it isn't a NETAPP lun");
                        last;
                    }
                } elsif ($line =~ /^flags\:\s+(.+)/i) {
                    my $flags = $1;
                    if ($flags =~ /coordinator/i) {
                        # Disk has coordinator flag, which means it is a fencing device, skip it
                        $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since it is a fenced device");
                        last;
                    }
                } elsif ($line =~ /^(\S+)\s+state=(\S+)\s+type=(\S+)/) {
                    my $device_tmp = $1;
                    my $state = $2;
                    my $type = $3;

                    # Skip the device if it isn't enabled
                    if ($state !~ /enabled/i) {
                        $self->sanlog(-msg => "Skipping device '$device_tmp' since the path isn't enabled");
                        next;
                    }

                    # remove s2 from the end so this will match the later parsed output from Sanlun
                    $device_tmp =~ s/s\d+$//;

                    # Now let's add it to $vxdisk_hash
                    $vxdisk_hash->{$device_tmp}->{"dmpnode"}  = $dmpnode;
                    $vxdisk_hash->{$device_tmp}->{"init_state"} = join(" ", @init_state);
                    if ($type =~ /primary/i) {
                        $vxdisk_hash->{$device_tmp}->{"primary"} = 1;
                    } else {
                        $vxdisk_hash->{$device_tmp}->{"primary"} = 0;
                    }
                } elsif ($line =~ /^(\S+)\s+state=(\S+)$/) {
                    my $device_tmp = $1;
                    my $state = $2;

                    # Skip the device if it isn't enabled
                    if ($state !~ /enabled/i) {
                        $self->sanlog(-msg => "Skipping device '$device_tmp' since the path isn't enabled");
                        next;
                    }

                    # remove s2 from the end so this will match the later parsed output from Sanlun
                    $device_tmp =~ s/s\d+$//;

                    # Now let's add it to $vxdisk_hash.  Since we don't have a "type", we will set
                    # it as a primary path
                    $vxdisk_hash->{$device_tmp}->{"dmpnode"} = $dmpnode;
                    $vxdisk_hash->{$device_tmp}->{"primary"} = 1;
                    $vxdisk_hash->{$device_tmp}->{"init_state"} = join(" ", @init_state);
                } ##END if ($line =~ /^udid\:\s+(.+)/) {
            } ##END foreach my $line (@vxdisk_list_output) {
        } ##END foreach my $device (sort keys %$vxdisk_hash) {

        # make sure our $vxdisk_hash ended up with something in it.
        unless (scalar(keys %$vxdisk_hash)) {
            $self->_set_err("The vxdisk relationship hash is empty. Unable to continue.");
            return (wantarray()) ? () : [];
        }

        # Collect sanlun so we can later sort out internal devices
        my @sanlun_output = $self->{sanlun}->lun_show();

        if ($self->{sanlun}->error) {
            $self->_set_err($self->{sanlun}->errmsg());
            return (wantarray()) ? () : [];
        }

        # Store the raw sanlun output for later
        $self->{raw_sanlun_output} = \@sanlun_output;

        foreach my $line (@sanlun_output) {
            # remove a known string that sanlun is prepending
            # to the first lun it finds (hack).
            $line =~ s|Report bugs to <bug-findutils\@gnu.org>.||;
            $line =~ s/^\s+//g;

            # Example Output depending on architecture
            # controller:                lun-pathname                device filename  adapter  protocol          lun size         lun state
            # auto-fas3020-03:  /vol/nate_2_1_auto_sunv440_01/nate_auto-sunv440-01_solaris_1  /dev/rdsk/c2t500A09849667B1D1d3s2  emlxs0   FCP           15g (16106127360)    GOOD
            if ($line =~ /(\S+)\s+\/vol\/(\S+)\/(\S+)\s+\/dev\/rdsk\/(\S+)s2\s+\S+\s+(\S+)\s+/) {
                # old sanlun kit output looks like below,
                # controller:                 lun-pathname                                 device filename                     adapter   protocol          lun size         lun state
                # iscsi_vs:  /vol/iscsi_vol_1/iscsi_auto-sunx2200-01_0  /dev/rdsk/c0t600A0980324668762D3F41734170386Fd0s2  scsi_vhci0  iSCSI          5g (5368709120)     GOOD
                #
                # new sanlun kit output looks like below,
                # controller(7mode)/                                             device                                            host                  lun
                # vserver(Cmode)       lun-pathname                              filename                                          adapter    protocol   size    mode
                # ---------------------------------------------------------------------------------------------------------------------------------------------------
                # iscsi_vs             /vol/iscsi_vol_1/iscsi_auto-sunx2200-01_0 /dev/rdsk/c0t600A0980324668762D3F41734170386Fd0s2 scsi_vhci0 iSCSI      5g      C
                my $filer_tmp    = $1;
                my $vol_tmp      = $2;
                my $lun_tmp      = $3;
                my $device_tmp   = $4;
                my $protocol_tmp = $5;

                # Remove colon in the end of filer/vserver name if presents
                $filer_tmp =~ s/:$//;

                # Now we want to look in the $vxdisk_hash for the device.  If it is reserved then we will
                # skip it.  If it isn't reserved, we will grab the dmpnode and add it to the hash with the
                # rest of the information.
                if (exists($vxdisk_hash->{$device_tmp})) {
                    # Store all of the collected information
                    $hDevs{$device_tmp} = {
                                            filer      => $filer_tmp,
                                            vol        => $vol_tmp,
                                            lun        => $lun_tmp,
                                            protocol   => $protocol_tmp,
                                            dmpnode    => $vxdisk_hash->{$device_tmp}->{'dmpnode'},
                                            init_state => $vxdisk_hash->{$device_tmp}->{'init_state'},
                                          };
                }
            }
        }

        # Now we want to walk through the $hDevs hash and eliminate any duplicate paths
        # for dmpnodes.  We likely have multiple primary and secondary paths for each dmpnode
        # and we only want to represent one primary path for each dmpnode in the final hash.
        my $temp_dmpnode_hash = {};
        my $hDevs_primary = {};
        foreach my $device (sort keys %hDevs) {
            # Check the primary/secondary status for the device in the
            # previously created vxdisk_hash.
            if (exists($vxdisk_hash->{$device})) {
                if (defined($vxdisk_hash->{$device}->{'primary'})
                    and $vxdisk_hash->{$device}->{'primary'} == 1) {
                    # Looks like this is a primary path so see if
                    # the dmpnode has already been represented.
                    my $dmpnode = $hDevs{$device}{'dmpnode'};
                    unless (exists($temp_dmpnode_hash->{$dmpnode})) {
                        # This dmpnode has not been represented yet.
                        $temp_dmpnode_hash->{$dmpnode}->{$device} = $hDevs{$device};
                    }
                } ##END if (defined($vxdisk_hash->{$device}->{'primary'}
            } ##END if (exists($vxdisk_hash->{$device})) {
        } ##END foreach my $device (sort keys %hDevs) {

        # Now take all of the primary paths and store them in a new hDevs_primary hash
        foreach my $dmpnode (sort keys %$temp_dmpnode_hash) {
            foreach my $device (sort keys %{$temp_dmpnode_hash->{$dmpnode}}) {
                $hDevs_primary->{$device} = $temp_dmpnode_hash->{$dmpnode}->{$device};
            }
        }

        # Make sure we found veritas luns.  If we didn't, then stop now.
        unless (scalar(keys %$hDevs_primary)) {
            $self->_set_err("Failed to find any valid devices under Veritas control");
            return (wantarray()) ? () : [];
        }

        # save map for next call of function
        $self->{dfmap} = $hDevs_primary;
    }

    # Make sure we have the dfmap and raw_sanlun_output.  If
    # we don't, then something went wrong earlier and we can't continue.
    unless (scalar(keys %{$self->{dfmap}} > 0)) {
        $self->_set_err("Unable to continue because the Device Map (dfmap) is empty");
        return (wantarray()) ? () : [];
    }
    unless (scalar(@{$self->{raw_sanlun_output}} > 0)) {
        $self->_set_err("Unable to continue becasue 'sanlun lun show -p' raw device output is empty");
        return (wantarray()) ? () : [];
    }
    unless (scalar(@{$self->{raw_vxdisk_list_output}} > 0)) {
        $self->_set_err("Unable to continue because 'vxdisk -o alldgs list' raw output is empty");
        return (wantarray()) ? () : [];
    }

    # If not -cnt then assume all free disks
    # under lvm control are wanted
    unless (defined $cnt and length($cnt)) {
        $cnt  = 10000;
        $bCnt = 0;
    }

    # Walk through the  dfmap and filter
    foreach my $device (sort keys %{$self->{dfmap}}) {
        # Filter based on single filer
        if (defined $filer and length $filer) {
            if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i) {
                $self->sanlog(-msg => "Skipping device '$device' since it doesn't match filer '$filer'",
                              -msg_level => 3);
                next;
            }
        }

        # Filter based on volume
        if (defined $vol and length $vol) {
            if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i) {
                $self->sanlog(-msg => "Skipping device '$device' since it doesn't match volume '$vol'",
                              -msg_level => 3);
                next;
            }
        }

        # Filter based on lun
        if (defined $lun and length $lun) {
            if ($self->{dfmap}{$device}{lun} !~ /$lun/i) {
                $self->sanlog(-msg => "Skipping device '$device' since it doesn't match lun '$lun'",
                              -msg_level => 3);
                next;
            }
        }

        # Filter based on protocol
        if (defined $protocol and length $protocol) {
            if ($self->{dfmap}{$device}{protocol} !~ /$protocol/i) {
                $self->sanlog(-msg => "Skipping device '$device' since it doesn't match protocol '$protocol'",
                              -msg_level => 3);
                next;
            }
        }

        # Filter based on list of filers
        if (defined $filers and scalar(@{$filers}) > 0) {
            my $found = 0;
            # Look through each filer in the filers list
            my $filer_list = join("|", @$filers);
            foreach my $filer_f (@{$filers}) {
                if ($self->{dfmap}{$device}{filer} =~ /^$filer_f$/i) {
                    # If filer matches devices filer, set found to 1
                    $found = 1;
                }
            }

            # If the device is not part of the filers list caller requested skip it.
            if ($found == 0) {
                $self->sanlog(-msg => "Skipping device '$device' since it doesn't match filers '$filer_list'",
                              -msg_level => 3);
                next;
            }
        }

        # Filter based on init_state
        if (defined $inited_state and length $inited_state) {
            next unless ($self->{dfmap}{$device}{init_state} =~ /^$inited_state$/i);
            $self->sanlog(-msg => "Selecting device '$device' since it matches inited_state '$inited_state'",
                          -msg_level => 3);
        }

        # See if the lun has a volume label on it
        my $prt_cmd = "prtvtoc /dev/rdsk/" . $device;
        if ($device !~ /s2\s*$/) {
            $prt_cmd .= "s2";
        }

        my ($prt_status, @prt_output) = &ClientUtil::execute($self->{conn}, $prt_cmd);
        if ($prt_status > 0) {
            $self->_set_err("Command ($prt_cmd) returned a bad status ($prt_status)");
            return (wantarray()) ? () : [];
        }

        # Example output:
        # * /dev/rdsk/c4t60A98000486E544D575A4B58646A5774d0s2 (volume "NATE") partition map
        # Grep out just the line with the device in it (this has the words "partition map" in it)
        @prt_output = grep(/partition map/, @prt_output);

        # Check to see if the line contains volname of reserved types
        if (grep(/root|boot|reserved/, @prt_output)) {
            $self->sanlog(-msg => "Ignoring device because it has a special volume label ($device)");
            next;
        }

        # Since we are returning veritas devices, we need to return the dmpnodes
        push @devices, $self->{dfmap}{$device}{dmpnode} if (++$total <= $cnt);
    }

    # If user supplied cnt and free devices fell short
    # generate error
    if ($bCnt == 1 and scalar(@devices) < $cnt) {
        $self->_set_err("Only " . scalar(@devices) . " of $cnt requested devices were found");
        return (wantarray()) ? () : [] ;
    } elsif (scalar(@devices) == 0) {
        $self->sanlog(-msg => "WARNING requested device list is EMPTY");
        return (wantarray()) ? () : [] ;
    }

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: load_vg
# Objective:       load an existing diskgroup that is not in memory
# Details:         See POD documentation at the end of this file
################################################################################
sub load_vg {
    my ($self, $name) = @_;
    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Invalid or missing required parameter 'dg name'");
        return {};
    }

    ### Check if vg has already been loaded
    $vgObj = $self->get_vg($name);
    if ($self->error) {
        # dg has not been load yet
        $self->_clear_err();

        # load dg
        $self->_create_vgs($name);
        return {} if ($self->error);
        $vgObj=$self->get_vg($name);
    } else {
        $self->sanlog(-msg => "Volume group '$name' was already loaded");
    }

    return $vgObj;

}

################################################################################
# Subroutine Name: unload_vg
# Objective:       Remove the vg object under LVM object control
# Details:         See POD documentation at the end of this file
################################################################################
sub unload_vg {
    my ($self, $name) = @_;
    my ($vgObj, $bFound, @vgs);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Invalid or missing required parameter 'dg name'");
        return {};
    }

    $bFound = 0;

    foreach $vgObj (@{$self->{vgs}}) {
        ($vgObj->get_name() =~ /^$name$/)
            ? $bFound=1
        : push @vgs, $vgObj;
    }

    $self->{vgs} = \@vgs;

    unless($bFound) {
        $self->_set_err("Nothing to unload. Vg name '".$name."' was not loaded");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: enable_controllers
# Objective:       enable controllers
# Details:         See POD documentation at the end of this file
################################################################################
sub enable_controllers {
    my ($self) = @_;

    my ($cmd, @output, $status);
    my ($ctlr, @controllers);

    $self->_clear_err();

    $self->sanlog('-msg', "Enabling all controllers for Veritas DMP");

    if ($self->is_installed()) {
        $self->rescan();
        return $ClientUtil::FALSE if ($self->error());

        # Get a list of all the controllers
        $cmd  =  "/etc/vx/bin/vxcntrllist";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # skip controller 0 (root disk)
        @controllers = grep (! /c0/, @output);

        # traverse through the controllers and enable each.
        foreach $ctlr (@controllers) {
            $cmd = "/usr/sbin/vxdmpadm enable ctlr=$ctlr";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }

}

################################################################################
# Subroutine Name: rescan
# Objective:       To run the vxdctl command to rescan veritas's list of
#                  disks that it thinks are under it's control.
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;
    my ($cmd, @output, $status);

    $self->_clear_err();

    # need to run the vxdctl script after the new VTOC is down
    # to get Veritas to see the disk again (properly) .. not completely
    # sure why
    $cmd  = "vxdctl enable";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: restart_recovery_daemon
# Objective:       Method that will stop the recovery deamon and reset
# Details:         See POD documentation at the end of this file
################################################################################
sub restart_recovery_daemon {
    my ($self, $interval) = @_;

    my ($cmd, @output, $status, $policy, $policy_string);

    $self->_clear_err();

    if (not defined ($interval)) {
        $self->sanlog(-msg => "Restore interval was not defined, thus it will not be reset (could be cleanup).");
        return $ClientUtil::TRUE;
    }

    my $version = $self->get_version();
    my $mp = $self->_get_version_mp();
    $mp =~ s/MP//;
    if ($mp eq "" or !defined($mp)) {
        $mp = 0;
    }

    if ($version =~ /^4/) {
        $policy = "check_all";
    } else {
        # default for 5.0+ is check_disabled
        $policy = $self->{parent}->{vx_restore_policy};
    }

    if ($self->_is_installed()) {
        $self->sanlog(-msg => "Setting Veritas Recovery Dameon Restore interval to $interval seconds");
        $self->sanlog(-msg => "Setting the Veritas Restore Policy to $policy");

        # with 5.0 MP3, there is new way to set restore daemon, use this if your at that revision.
        # This allows the setting to be persistent across reboots
        if (($version =~ /^5/ and $mp >= 3) or $version >= 5.1) {
            # set dmp_restore_interval
            $cmd = "/usr/sbin/vxdmpadm settune dmp_restore_interval=$interval";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                foreach my $line (@output) {
                    if (($line =~ /vxdmpadm ERROR/) && ($line !~ /same as its existing/)) {
                        $self->_set_err("Unable to set dmp_restore_interval");
                        return $ClientUtil::FALSE;
                    }
                }
            }

            # set dmp_restore_policy
            $cmd = "/usr/sbin/vxdmpadm settune dmp_restore_policy=$policy";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                foreach my $line (@output) {
                    if (($line =~ /vxdmpadm ERROR/) && ($line !~ /same as its existing/)) {
                        $self->_set_err("Unable to set dmp_restore_policy");
                        return $ClientUtil::FALSE;
                    }
                }
            }
        } else {
            # stop the restored
            $cmd = "/usr/sbin/vxdmpadm stop restore";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }

            if (defined($policy)) {
                $policy_string = "policy=$policy";
            }

            # Sleep for 2 seconds to make sure the restored completely stops
            sleep(2);

            # start the restored
            $cmd = "/usr/sbin/vxdmpadm start restore interval=$interval $policy_string";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }

            # set status on the setting, just to get it in the logs.
            $cmd  = "/usr/sbin/vxdmpadm stat restored";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($conn, $log, $lvm, $loadcfg, $parent, $self);

    $self = {};

    bless $self,$class;

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_})) {
                $self->_set_err("Invalid value for argument '-lvm'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent'");
                return ($ClientUtil::FALSE);
            }
            next;
        } elsif (/-?loadcfg/) {
            if (not length($loadcfg = $args{$_}) or $loadcfg !~ /no|yes|1|0|off|on/i) {
                $loadcfg = 1;
            }
            next;
        }
    }

    ### check for existance of required -lvm argument
    if (grep ( /^-lvm$/, (keys %args) )) {
        unless ($args{-lvm} =~ /^vxvm|dmp$/i) {
            $self->_set_err("Invalid value for argument '-lvm' \(".$args{-lvm}."\)");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument'-lvm'");
        return $self;
    }

    ### check for existance of required -parent argument
    if (grep ( /^-?parent$/, (keys %args) )) {
        unless (ref($args{-parent}) =~ /SanClient/i) {
            $self->_set_err("Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Missing required argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{vgs} = [];

    ### check to make sure that LVM is installed
    if (not $self->_is_installed()) {
        return $self if ($self->error);
    }

    ### create veritas vg objects for existing disk groups
    if ($loadcfg) {
        $self->_create_vgs() ;
        return $self if ($self->error);

        ### update map to determine disk format
        $self->_update_disk_format_map();
        return $self if ($self->error);
    }

    # set VRTS enclosure tuning and restart recovery daemon with passed
    # settings.
    $self->_set_vxvm_encl_tunings();
    return $self if ($self->error);

    $self->restart_recovery_daemon($self->{parent}->{vx_restore_interval});
    return $self if ($self->error);

    # this is the case with VxVM volume manager, but MPXIO
    # multipathing. Make sure in this context that fast
    # recovery is disabled.
    if ($self->{parent}->{mp}->{mp_type} =~ /mpxio/i) {
        $self->_set_vxvm_fast_recovery_mpxio();
        return $self if ($self->error);
    }

    return($self);

}

################################################################################
# Subroutine Name: _update_disk_format_map
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _update_disk_format_map {
    my ($self) = @_;

    my ($cmd, @output, $status);

    ### Retrieve the volume groups.
    $cmd = "vxdisk -o alldgs list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 1) {
        $self->_set_err("Unable to update map");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        $line =~ s/auto://;
        my ($device,$format) = $line =~ /^(\S+)s2\s+(\S+)/;
        $hFormat{$device} = $format;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _create_vgs
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vgs {
    my ($self,$vgname) = @_;

    my ($cmd, @output, $status, $vgObj, %hDeport);

    $cmd = (defined ($vgname) and length ($vgname))
        ? "vxdg list | egrep -v 'NAME' | egrep '$vgname '"
        : "vxdg list | egrep -v 'NAME' ";

    ### Retrieve the volume groups.
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 1) {
        $self->_set_err("Unable to load existing diskgroup\(s\)");

        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        next if ($line =~ /root|boot|reserved/);  ### skip over rootdg (<4.0)
        $line =~ /(\w+).*/;
        $self->create_vg(-name => $1,
                         -deported => 0);
        return $ClientUtil::FALSE if ($self->error);
    }

    if (not defined $vgname) {
        # Create objects for the deported disk groups as well
        $cmd = "vxdisk -o alldgs list";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to load deported diskgroups");
            return $ClientUtil::FALSE;
        }
    } else {
        # Determine if dg is deported
        $cmd = "vxdisk -o alldgs list | grep '(".$vgname.")'";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 1) {
            $self->_set_err("Unable to load deported diskgroups");
            return $ClientUtil::FALSE;
        }
    }

    ### Retrieve the deported disk groups
    foreach my $line (@output) {
        my ($dg) = $line =~ /^\w+s2\s+\S+\s+-\s+\((.*)\)\s+/i;
        if ($dg and $dg !~ /root|boot|reserved/i) {
            $hDeport{$dg} += 1;
            ### Vg with multple devices only need to
            ### be created once.
            if ($hDeport{$dg} == 1) {
                $vgObj=$self->create_vg(-name => $dg,
                                        -deported => 1);
                return $ClientUtil::FALSE if ($self->error);
            }
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _get_disk_id
# Objective:       Return disk id
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_disk_id {
    my ($self, $device) = @_;

    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    $name = "/dev/rdsk/$device" . "s2";

    my $path = $self->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show -v -d $name | grep 'Serial number'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to get diskid for device \'$device\'",
                         @output);
        return $ClientUtil::FALSE;
    }

    return $output[0];

}

################################################################################
# Subroutine Name: _is_installed
# Objective:       Check to make sure that the LVM is installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_installed {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "/usr/bin/pkginfo VRTSvxvm";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 0) {
        $self->sanlog(-msg => "lvm is installed");
        return $ClientUtil::TRUE;
    } else {
        $self->sanlog(-msg => "lvm is not installed");
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Method:    _set_vxvm_fast_recovery_mpxio
# Objective: This is called in MPXIO MP case, and in that case we should
#            turn off fast recovery.
# Details:   See POD documentation at the end of this file.
################################################################################
sub _set_vxvm_fast_recovery_mpxio {
    my ($self) = @_;

    my (@cmds, $fast_r, $line, $cmd, $status, @output, $version, $mp, $rp, $sp);
    my ($vx_50_fast_recovery,
        $vx_50_recovery_option,
        $vx_50_queue_depth,
        $vx_io_policy,
        $vx_50_dmp_log_level) =
          ($self->{parent}->{vx_50_fast_recovery},
           $self->{parent}->{vx_50_recovery_option},
           $self->{parent}->{vx_50_queue_depth},
           $self->{parent}->{vx_io_policy},
           $self->{parent}->{vx_50_dmp_log_level});

    # As currently designed, this relies on the thpl sending default values
    # down when the tester has not specified any optional parameters.
    # vx_50_fast recovery, vx_50_recovery_options, vx_50_queue_depth
    # and vx_io_policy
    if ($vx_50_fast_recovery eq "" or $vx_50_recovery_option eq "" or
        $vx_50_queue_depth eq "" or $vx_io_policy eq "") {

        $self->sanlog(-msg => "_set_vxvm_fast_recovery_mpxio: One or more options parameters have not been passed in.\n" .
                              "This is either because a cleanup is happening, or the parameters\n" .
                              "were not properly defined.");

        # Return true here because this may not have happened in error.
        # This can happen in cases where cleanup is being done, or
        # optional parameters simply were not provided.
        return $ClientUtil::TRUE;
    }

    # get version (including MP)
    $version = $self->get_version();

    # Get the MP
    $mp = $self->_get_version_mp();
    $mp =~ s/MP//;
    if ($mp eq "" or not defined($mp)) {
        $mp = 0;
    }

    # Get the RP
    $rp = $self->_get_version_rp();
    $rp =~ s/RP//;
    if ($rp eq "" or not defined($rp)) {
        $rp = 0;
    }

    # Get the SP
    $sp = $self->_get_version_sp();
    $sp =~ s/RP//;    if ($sp eq "" or not defined($sp)) {
        $sp = 0;
    }

    # Log a message indicating the version we are tuning.
    if ($sp > 0) {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version SP" . $sp);
    } elsif ($rp > 0) {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version RP" . $rp);
    } elsif ($mp > 0) {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version MP" . $mp);
    } else {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version flat");
    }

    # in addition to turn off fast recovery, make sure
    # APM is not installed.
    if ($version =~ /^5/) {
        $cmd = "pkginfo -l VRTSNTAPapm";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status == 0) {
            $self->_set_err("Package (VRTSNTAPapm) is installed and should NOT be with MPXIO MultiPathing");
            return $ClientUtil::FALSE;
        }
    }

    # clear error from previous check (expected it to fail)
    $self->_clear_err();

    # set fast recovery off
    if ($version =~ /^5/ or ($version =~ /4\.1/ and $mp >= 2)) {
        $fast_r = $vx_50_fast_recovery;
        if (!defined($fast_r) or $fast_r eq "") {
            $fast_r = "off";
        }
        $cmd = "/usr/sbin/vxdmpadm settune dmp_fast_recovery=$fast_r";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            foreach $line (@output) {
                if (($line =~ /vxdmpadm ERROR/) && ($line !~ /same as its existing/)) {
                    $self->_set_err("Failed to set DMP fast recovery.");
                    return $ClientUtil::FALSE;
                }
            }
        }
    }
}

################################################################################
# Method: _set_vxvm_encl_tunings
# Objective: Set Veritas enclosure ebased tunings.
# Details:   See POD documentation at the end of this file.
################################################################################
sub _set_vxvm_encl_tunings {
    my ($self) = @_;

    my (@cmds, $line, $cmd, $status, @output, @enclosure_names, $version);
    my ($mp, $rp, $sp, $rec_option, $queue_option, $fast_r);

    my ($vx_50_fast_recovery,
        $vx_50_recovery_option,
        $vx_50_queue_depth,
        $vx_io_policy,
        $vx_50_dmp_path_age,
        $vx_50_dmp_log_level,
        $vx_50_probe_idle_lun,
        $vx_50_monitor_fabric) =
          ($self->{parent}->{vx_50_fast_recovery},
           $self->{parent}->{vx_50_recovery_option},
           $self->{parent}->{vx_50_queue_depth},
           $self->{parent}->{vx_io_policy},
           $self->{parent}->{vx_50_dmp_path_age},
           $self->{parent}->{vx_50_dmp_log_level},
           $self->{parent}->{vx_50_probe_idle_lun},
           $self->{parent}->{vx_50_monitor_fabric});

    # As currently designed, this relies on the thpl sending default values
    # down when the tester has not specified any optional parameters.
    # vx_50_fast recovery, vx_50_recovery_options, vx_50_queue_depth
    # and vx_io_policy
    if ($vx_50_fast_recovery eq "" or $vx_50_recovery_option eq "" or
        $vx_50_queue_depth eq "" or $vx_io_policy eq "") {
        $self->sanlog(-msg => "_set_vxvm_encl_tunings: One or more options parameters have not been passed in.\n" .
                              "This is either because a cleanup is happening, or the parameters were not properly defined.");

        # Return true here because this may not have happened in error.
        # This can happen in cases where cleanup is being done, or
        # optional parameters simply were not provided.
        return $ClientUtil::TRUE;
    }

    @enclosure_names = $self->_get_vxvm_enclosure_name();
    if (scalar(@enclosure_names) == 0) {
        $self->_clear_err();
        $self->sanlog(-msg => "_set_vxvm_encl_tunings: No valid enclosure names were found.\n" .
                              "This is either because a cleanup is happening, or something was not properly setup.");
        # Return true here because this may not have happened in error.
        # This can happen in cases where cleanup is being done.
        return $ClientUtil::TRUE;
    }

    # get version (including MP)
    $version = $self->get_version();

    # Get the MP
    $mp = $self->_get_version_mp();
    $mp =~ s/MP//;
    if ($mp eq "" or not defined($mp)) {
        $mp = 0;
    }

    # Get the RP
    $rp = $self->_get_version_rp();
    $rp =~ s/RP//;
    if ($rp eq "" or not defined($rp)) {
        $rp = 0;
    }

    # Get the SP
    $sp = $self->_get_version_sp();
    $sp =~ s/SP//;
    if ($sp eq "" or not defined($sp)) {
        $sp = 0;
    }

    # Log a message indicating the version we are tuning.
    if ($sp > 0) {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version SP" . $sp);
    } elsif ($rp > 0) {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version RP" . $rp);
    } elsif ($mp > 0) {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version MP" . $mp);
    } else {
        $self->sanlog(-msg => "The Version of Veritas being tuned is $version flat");
    }

    # Veritas has backported the new 5.0 tunings to Veritas 4.1 MP2+.
    # Thus we allow the 5.0 tunings to be set for 4.1 MP2+ (no fast recovery)
    if  ($version =~ /^(5|6)/ or ($version =~ /4\.1/ and $mp >= 2)) {
        my $count = 0;

        # for 5.1 or greater always set recoveryoption=default
        if ($version >= 5.1) {
            $rec_option = "recoveryoption=default";
            $count++;
        } elsif (defined $vx_50_recovery_option and $vx_50_recovery_option ne "NOT_SET") {
            $rec_option = "recoveryoption=$vx_50_recovery_option";
            $count++;
        } elsif (defined $vx_50_recovery_option and $vx_50_recovery_option eq "NOT_SET") {
            $rec_option = "recoveryoption=throttle";
            $count++;
        }

        # for 5.1 or greater, queuedepth is not a valid option
        if ($version >= 5.1) {
            $queue_option = "";
            $count++;
        } elsif (defined $vx_50_queue_depth and $vx_50_queue_depth ne "NOT_SET") {
            $queue_option = "queuedepth=$vx_50_queue_depth";
            $count++;
        } elsif (defined $vx_50_queue_depth and $vx_50_queue_depth eq "NOT_SET") {
            $queue_option = "queuedepth=20";
            $count++;
        }

        # if vx_50_recovery_option or vx_50_queue_depth is set, tune.
        if (defined $queue_option and defined $rec_option) {
            foreach my $enclosure_name (@enclosure_names) {
                $cmd = "/usr/sbin/vxdmpadm setattr enclosure $enclosure_name $rec_option $queue_option";
                ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
                if ($status != 0) {
                    $self->_set_err("Failed to set recovery option or queuedepth");
                    return $ClientUtil::FALSE;
                }
            }
        } elsif ($count == 1) {
            $self->_set_err("Veritas requires both queuedepth and recoveryoption to be tuned together (for VRTS versions < 5.1). Please specify values for both VX_50_RECOVERY_OPTION and VX_50_QUEUE_DEPTH");
            return $ClientUtil::FALSE;
        }
    }

    # set fast recovery and dmp log level options
    if ($version =~ /^(5|6)/ or ($version =~ /4\.1/ and $mp >= 2)) {
        if ($version =~ /4\.1/ and $mp >= 2) {
            $fast_r = "off";
        } else {
            $fast_r = $vx_50_fast_recovery;
        }
        $cmd = "/usr/sbin/vxdmpadm settune dmp_fast_recovery=$fast_r";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            foreach $line (@output) {
                if (($line =~ /vxdmpadm ERROR/) and ($line !~ /same as its existing/)) {
                    $self->_set_err("Failed to set DMP fast recovery.");
                    return $ClientUtil::FALSE;
                }
            }
        }

        $cmd = "/usr/sbin/vxdmpadm settune dmp_log_level=$vx_50_dmp_log_level";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status != 0) {
            foreach $line (@output) {
                if (($line =~ /vxdmpadm ERROR/) and ($line !~ /same as its existing/)) {
                    $self->_set_err("Failed to set DMP log level.");
                    return $ClientUtil::FALSE;
                 }
            }
        }
    }

    # set probe_idle_lun setting for 5.0MP3+
    if ($version =~ /^(5|6)/) {
        if ($vx_50_probe_idle_lun =~ /NOT_SET/i) {
            # If user did not set the option, then default to off for MP3+
            if (($version == 5.0 and $mp >= 3) or $version >= 5.1) {
                $vx_50_probe_idle_lun = "off";
            } else {
                $vx_50_probe_idle_lun = "on";
            }
        }

        if ($vx_50_probe_idle_lun !~ /NOT_SET/i) {
            # if a value was set by user or from MP3+ default, then set tunable
            $cmd = "/usr/sbin/vxdmpadm settune dmp_probe_idle_lun=$vx_50_probe_idle_lun";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                foreach $line (@output) {
                    if (($line =~ /vxdmpadm ERROR/) and ($line !~ /same as its existing/)) {
                        $self->_set_err("Failed to set DMP dmp_probe_idle_lun tunable.");
                        return $ClientUtil::FALSE;
                     }
                }
            }
        }
    }

    # set dmp_monitor_fabric tunable for Veritas 5.X and 6.X
    if ($version =~ /^(5|6)/) {
        if (defined($vx_50_monitor_fabric)) {
            if ($vx_50_monitor_fabric eq "NOT_SET") {
                # If the user didn't set the value, set it off for mpxio or on for everything
                # else.  "on" is the veritas defualt.
                if ($self->{parent}->{mp}->{mp_type} =~ /mpxio/i) {
                    $vx_50_monitor_fabric = "off";
                } else {
                    $vx_50_monitor_fabric = "on";
                }
            }

            # Now tune it
            # changes to tool to make change in MP3+ or 5.1 and greater
            if (($version == 5.0 and $mp >= 3) or $version >= 5.1) {
                $cmd = "/usr/sbin/vxdmpadm settune dmp_monitor_fabric=$vx_50_monitor_fabric";
            } else {
                $cmd = "/usr/sbin/vxddladm settune monitor_fabric=$vx_50_monitor_fabric";
            }
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                foreach $line (@output) {
                    if (($line =~ /vxdmpadm ERROR/) and ($line !~ /same as its existing/)) {
                        $self->_set_err("Failed to set DMP monitor_fabric tunable.");
                        return $ClientUtil::FALSE;
                    } elsif ($line=~/vxddladm ERROR/) {
                        $self->_set_err("Failed to set DMP monitor_fabric tunable.");
                        return $ClientUtil::FALSE;
                    }
                }
            }
        }
    }

    # set IO Policy (if passed in and not empty string)
    if (defined $vx_io_policy and $vx_io_policy eq "NOT_SET") {
        $vx_io_policy = "adaptive";
    }

    if (defined $vx_io_policy) {
        foreach my $enclosure_name (@enclosure_names) {
            $cmd = "/usr/sbin/vxdmpadm setattr enclosure $enclosure_name iopolicy=$vx_io_policy";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Failed to set the IO Policy for Enclosure: $enclosure_name .");
                return $ClientUtil::FALSE;
            }
        }
    }

    # Set the dmp_path_age if it is specified.. If version is 5.1SP1 or later
    # and value of vx_50_dmp_path_age is "NOT_SET" then hard code it to "120".
    # This is only needed if we are not running mpxio since veritas only sees
    # one path with mpxio configs.

    # First check to see if we are running 5.1 SP1 or later and force it to 120
    # if it isn't set.
    if ($version =~ /^(5|6)/) {
        if (defined($vx_50_dmp_path_age)) {
            if ($vx_50_dmp_path_age eq "NOT_SET") {
                # We found the vx_50_dmp_path_age variable and it wasn't set by the user
                # or it wasn't defined so adjust it based on the version you are running
                if (($version == 5.1 and $sp > 0) or $version > 5.1) {
                    # Looks like we are running 5.1SP1 or higher so force it to 120
                    $vx_50_dmp_path_age = "120";
                } else {
                    # We are running a version prior to 5.1SP1 so set it to the
                    # veritas default of 300.
                    $vx_50_dmp_path_age = "300";
                }
            }

            # Now tune dmp_path_age if we find $vx_50_dmp_path_age defined
            $cmd = "/usr/sbin/vxdmpadm settune dmp_path_age=$vx_50_dmp_path_age";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                foreach $line (@output) {
                    if (($line =~ /vxdmpadm ERROR/) and ($line !~ /same as its existing/)) {
                        $self->_set_err("Failed to set DMP path_age tunable.");
                    return $ClientUtil::FALSE;
                    }
                }
            }
        }
    }

    # print out all the tunables for each enclosure to get in log
    foreach my $enclosure_name (@enclosure_names) {
        if ($version >= 5.1) {
            @cmds = ("/usr/sbin/vxdmpadm getattr enclosure $enclosure_name iopolicy",
                     "/usr/sbin/vxdmpadm getattr enclosure $enclosure_name recoveryoption",
                     "/usr/sbin/vxdmpadm gettune all");
        } elsif ($version =~ /^5/) {
            @cmds = ("/usr/sbin/vxdmpadm listapm dmpnetapp",
                     "/usr/sbin/vxdmpadm getattr enclosure $enclosure_name iopolicy",
                     "/usr/sbin/vxdmpadm getattr enclosure $enclosure_name recoveryoption",
                     "/usr/sbin/vxdmpadm gettune all",
                     "/usr/sbin/vxddladm gettune all");
        } elsif ($version =~ /4\.1/ and $mp >= 2) {
            @cmds = ("/usr/sbin/vxdmpadm getattr enclosure $enclosure_name iopolicy",
                     "/usr/sbin/vxdmpadm getattr enclosure $enclosure_name recoveryoption",
                     "/usr/sbin/vxdmpadm gettune all",
                     "/usr/sbin/vxddladm gettune all");
        } else {
            @cmds = ("/usr/sbin/vxdmpadm getattr enclosure $enclosure_name iopolicy");
        }

        # Execute, don't bother to check status - just need info in log
        foreach $cmd (@cmds) {
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Method: _get_vxvm_enclosure_name
# Objective: Get the enclosure name of the attached Veritas Enclosure
# Details:   See POD documentation at the end of this file.
################################################################################
sub _get_vxvm_enclosure_name {
    my ($self) = @_;

    my ($cmd, $line, $status, @output, @enclosure_names);

    # need to add in case enclosure is not yet connected.
    $self->rescan();
    if ($self->error()) {
        return (wantarray()) ? () : [];
    }

    @enclosure_names = ();

    $cmd = "/usr/sbin/vxdmpadm listenclosure all";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to execute command ($cmd)");
        return (wantarray()) ? () : [];
    }

    foreach $line (@output) {
        chomp ($line);
        $line =~ s/\r//g;
        if (($line !~ /DISCONNECTED/) &&
            ($line =~ /CONNECTED/) && ($line !~ /Disk/)) {
            my ($enclosure_name, undef) = split ('\s+',$line);
            push (@enclosure_names, $enclosure_name);
        }
    }

    if (scalar(@enclosure_names) == 0) {
        $self->_set_err("Failed to collect enclosure names");
        return (wantarray()) ? () : [];
    } else {
        return (wantarray()) ? @enclosure_names : \@enclosure_names;
    }

}

1;
