# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface the Vxvm logical volumes
## @author knopp@netapp.com, latif@netapp.com
## @status private
## @pod ../../../Vol.pod

package San::SanClient::Lvm::Solaris::Vxvm::Vg::Vol;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use Data::Dumper;
use File::Basename;

use vars qw($_NAME_CNT @ISA);

@ISA = qw(Error SanLog);

###############################################################################
# Method:    create_fs
# Objective: create a filesystem on a volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_fs {
    my ($self, %args) = @_;

    my ($cmd, $type, $vgname, $name, $status, @output);

    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?type$/) {
            $type = $args{$_};
            next;
        } else {
            $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                            "Invalid parameter specified \'$_\'");
            return $self;                            
        }
    }

    if (not defined($type) or $type eq "") {
        # Default -type to value of vxfs, since we are in the vxfs libs
        # This can be overridden with an argument flag of -type
        $type = "vxfs";
    }
    
    ### check for existance of required -lvm argument
    unless (defined $type) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::create_fs()",
                        "Missing required argument '-type'");
        return $self; 
    }

    if ($type =~ /vxfs/i) {
        ### Rename the volume
        $cmd = "/usr/sbin/mkfs -F $type -o largefiles /dev/vx/rdsk/$vgname/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        
        if ($status > 0) {
            $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::create_fs Unable to create $type filesystem on volume '$name'",
                            @output);
            return $ClientUtil::FALSE;                            
        } 
    } elsif ($type =~ /ufs/) {
        $cmd = "yes | /usr/sbin/newfs /dev/vx/rdsk/$vgname/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
        if ($status > 0) {
            $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::create_fs Unable to create $type filesystem on volume '$name'",
                            @output);
            return $ClientUtil::FALSE;                            
        }    
    }

    ### Update the volume metadata with the fstype.
    $cmd = "vxedit -g $vgname set fstype=$type $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::create_fs Unable to update the volume attribute 'fstype'",
                        @output);
        return $ClientUtil::FALSE;                            
    } 

    $self->{filesystem} = $type;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self, %args) = @_;
    
    my ($cmd, $force, $dgname, $name, @output, $status, @vols);
    
    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::destroy()",
                                "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }        
            next;
        } 
    }

    $dgname = $self->{parent}->get_name();
    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::destroy() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    ### if force option specified umount all volumes so deport
    ### can occur without error.  Make note of which vols were
    ### mounted so they can be remounted after rename occurs
    if (defined $force) {
        $self->sanlog(-msg => "determine if volume has filesystem");
        if ($self->is_fs()) {
            $self->sanlog(-msg => "determine if volume is mounted");
            if ($self->is_mounted()) {
                $self->sanlog(-msg => "unmount filesystem " . $self->get_name());
                $self->unmount();
                return $ClientUtil::FALSE if ($self->error());
            }        
        }        
    }

    $cmd = "vxassist -g $dgname remove volume $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::destroy() Unable to destroy volume \'$name\' in diskgroup '$dgname'",$cmd, @output);
        return $ClientUtil::FALSE;  
    } 

    ### Remove the volume from the parent vol group object
    foreach my $volObj ( @{$self->{parent}->{vols}}) { 
        if ($volObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vols,$volObj;
        }
    }
    
    $self->{parent}->{vols} = \@vols;

    return $ClientUtil::TRUE;    
    
}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;

    my (@buf, $cmd, $vgname, $name, @output, $status, $header, $summary);
    
    $self->_clear_err();
    
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::display() volume name is a required parameter");
        return (wantarray()) ? () : []; 
    }
    
    $cmd = "vxprint -g $vgname -l $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::display() Unable to display info for volume \'$name\' in volume group \'$vgname\'",
                        @output);
        return (wantarray()) ? () : [];                            
    } 
    
    $header = 
        sprintf("\nVol Summary:\n".
                "%-12.12s %-7.7s %-5.5s %-12.12s %-s\n".
                "%-12.12s %-7.7s %-5.5s %-12.12s %-32.32s",
                "NAME","LAYOUT","TYPE","SIZE(mb)","MOUNT_PT",
                "="x12,"="x7,"="x5,"="x12,"="x32);

    $summary = 
        sprintf("%-12.12s %-7.7s %-5.5s %-12.12s %-s",
                $self->get_name(),
                $self->get_layout(),
                $self->get_fs(),
                $self->get_size(),
                $self->get_mount());

    @buf = ($header,$summary, "\nVol Detail:", @output);

    return (wantarray()) ? @buf : \@buf;

}

################################################################################
# Subroutine Name: get_fs
# Objective:       Return Vol fs type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{filesystem} and length $self->{filesystem}) {
        return $self->{filesystem};
    } else {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::get_fs() Vol object is missing filesystem");
        return "";
    }

}

################################################################################
# Subroutine Name: get_layout
# Objective:       Return volume layout
# Details:         See POD documentation at the end of this file
################################################################################
sub get_layout {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{layout}) {
        return $self->{layout};
    } else {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::get_layout() Vol object is missing layout");
        return "";
    }

}

################################################################################
# Subroutine Name: get_mount
# Objective:       Return mount point of volume
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{mount} and length $self->{mount}) {
        return $self->{mount};
    } else {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::get_mount() Vol filesystem is not mounted");
        return "";
    }

}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vol object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::get_name() Vol object is missing name");
        return "";
    }

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return ooject size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{size}) {
        return $self->{size};
    } else {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::get_size() Vol object is missing size");
        return "";
    }

}

################################################################################
# Subroutine Name: get_state
# Objective:       Return object state
# Details:         See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{state}) {
        return $self->{state};
    } else {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::get_state() Vol object is missing state");
        return "";
    }

}

################################################################################
# Subroutine Name: get_vgname
# Objective:       Return Vol object diskgroup name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgname {
    my ($self) = @_;

    my ($name);

    $self->_clear_err();
    
    $name = $self->{parent}->get_name();

    if (defined $name and length $name) {
        return $name;
    } else {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::get_vgname() Vol object is missing parent name");
        return "";
    }

}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       Return path to raw vol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    $self->_clear_err();

    ### get name of disk group
    my $vgname = $self->{parent}->get_name();

    ### get name of volume    
    my $name = $self->get_name();

    return "/dev/vx/rdsk/$vgname/$name";

}

################################################################################
# Subroutine Name: is_fs
# Objective:       Determine if volume has a filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub is_fs {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{filesystem} and $self->{filesystem} !~ /raw/i) {
        return 1;
    } else {
        return 0;
    }

}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if volume is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    $self->_clear_err();
    
    if (defined $self->{mount} and length $self->{mount}) {
        return 1;
    } else {
        return 0;
    }
}

################################################################################
# Subroutine Name: mount
# Objective:       mount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub mount {
    my ($self, %args) = @_;

    my ($cmd, $directio, $fsck, $type, $vgname, $mount, $name);
    my ($options, $pass_thru, $status, @output);

    ### Get attributes 
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();
    $type = $self->{filesystem};
    $mount = undef;
    $pass_thru = "";

    ### check if volume is unformatted
    if (defined $type and $type =~ /raw/) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::mount()",
                        "Cannot mount a raw filesystem");
        return $ClientUtil::FALSE; 
    }

    # parse arguments
    foreach (keys %args) {
        if (/-?path$/) {
            if (not length($mount = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::mount()",
                                "Invalid value for argument '-path'");
                return $ClientUtil::FALSE;
            }        
            next;
        } elsif (/-?directio$/) {
            if (not length($directio = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::mount()",
                                "Invalid value for argument '-directio'");
                return $ClientUtil::FALSE;
            }        
            next;
        } elsif (/-?fsck$/) {
            if (not length($fsck = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::mount()",
                                "Invalid value for argument '-fsck'");
                return $ClientUtil::FALSE;
            }        
            next;
        } elsif (/-?pass_thru$/) {
            $pass_thru = $args{$_};
            $pass_thru =~ s/^\s+//;
            $pass_thru =~ s/\s+$//;
            if ($pass_thru ne "") {
                $pass_thru = "," . $pass_thru;
            }
        }
    }

    ### check if the volume is already mounted
    if ($self->is_mounted()) {
        $self->_set_err( "Error: Solaris::Vxvm::Vg::Vol::mount() volume '$name' is already mounted");                    
        return $ClientUtil::FALSE;                            
    }
    

    ### Use default path if not defined
    if (not defined $mount) {
        $mount = "/mnt/qa/$vgname/$name";
        $self->sanlog(-msg => "using default mount pt '$mount'");      
    } else {
        $mount .= "/$vgname/$name";
        $self->sanlog(-msg => "using mount pt '$mount'"); 
    }
    
    $self->{mount} = $mount;

    ### Save the mount point with the vol metadata in case
    ### a vol is renamed while mounted
    $self->_set_metadata("comment",$mount);
    return if ($self->error); 

    ### Create the mount point and open up dir privs 
    $cmd = "mkdir -p $mount; chmod 777 $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::mount() Unable to create mount point directory for volume '$name'",
                        @output);
        
        return $ClientUtil::FALSE;                            
    } 

    if ($fsck) {
        $cmd = "fsck -F " . $self->get_fs() . " -y /dev/vx/dsk/$vgname/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);     
        if ($status > 0) {
            $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::mount() Unable to fsck volume filesystem",
                            @output);
            
            return $ClientUtil::FALSE;                            
        } 
    }

    ### Mount the volume's filesystem
    $options = "largefiles";
    if ($directio) {
        $options .= ",forcedirectio";
    }

    $cmd = "/sbin/mount -F $type -o " . 
           $options . $pass_thru . " /dev/vx/dsk/$vgname/$name $mount";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::mount() Unable to mount volume filesystem",
                        @output);
        return $ClientUtil::FALSE;                            
    } 

    ### Store the mount string in the volumes metadata attribute '-comment'
    $cmd = "vxedit -g $vgname set comment='$mount' $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::mount() Unable to save mount pt in volume metadata",
                        @output);
        return $ClientUtil::FALSE;                            
    }

    ### Get handle to client object 
    $self->{client} = $self->{parent}{parent}{parent};    

    # Comment out for now, can cause reboot to hang with stale entries in vfstab
    #$self->{client}->mount_entry_add(
    #                             -device_mount=>"/dev/vx/dsk/$vgname/$name",
    #                             -device_fsck=>"/dev/vx/rdsk/$vgname/$name",
    #                             -mount_pt=>$mount,
    #                             -fs_type=>$self->get_fs()
    #                             );
    #if ($self->{client}->error) {
    #$self->_set_err($self->{client}->errmsg());
    #return $ClientUtil::FALSE;
    #}   

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rename()
# Objective:       rename volume
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, $newname) = @_;

    my ($cmd, $name, $vgname, @output, $status); 

    $self->_clear_err();

    ### get name of disk group
    $vgname = $self->{parent}->get_name();

    ### get name of volume    
    $name = $self->get_name();

    ### verify the device is being used in volume group
    unless (defined $name) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::rename()",
                        "Volume doesn't have a name");
        return $ClientUtil::FALSE;
    }

    ### verify the device is being used in volume group
    unless (defined $newname) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::rename()",
                        "New name for Volume not specified");
        return $ClientUtil::FALSE;
    }

    ### Rename the volume
    $cmd = "vxedit -g $vgname -v rename $name $newname";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::rename Unable to rename volume '$name' to \'$newname\'", @output);
        return $ClientUtil::FALSE;                            
    } 

    $self->{name} = $newname;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: rescan()
# Objective:       rescan volume
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;

    my ($mount);

    ### Currently just handles remounting.   
    ### Needed after snapdrive restore
    $self->_clear_err();
    
    $mount = $self->_init_mount();
    if (defined $mount and length $mount) {
        $self->mount(-path => $mount);
    } else {
        $self->{mount} = undef;
    }

    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: fspec_sign()
# Objective:       Create fspec signature on storage
# Details:         See POD documentation at the end of this file
################################################################################
sub fspec_sign {
    my ($self) = @_;

    my ($cmd, $vgname, $name, $status, @output);

    $self->_clear_err();

    ### get name of disk group
    $vgname = $self->{parent}->get_name();

    ### get name of volume    
    $name = $self->get_name();

    if ($self->is_fs()) { 
        if ($self->is_mounted()) {
            $cmd = "echo $vgname/$name > " . $self->get_mount() . "/fspec.sig";
        } else {
            $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::fspec_sign() Unable to create fspec signature",
                            "Logical volume is currently unmounted", @output);
            return $ClientUtil::FALSE;   
        }
    } else {
        $cmd = "echo $vgname/$name | dd of=/dev/vx/rdsk/" . $vgname . "/" . $name . " bs=512 conv=sync";
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::fspec_sign() Unable to create fspec signature",
                        @output);
        return $ClientUtil::FALSE;                            
    } 
    
    return $ClientUtil::TRUE;
}


################################################################################
# Subroutine Name: get_fspec()
# Objective:       Retrieve fspec signature on storage
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fspec {
    my ($self) = @_;

    my ($cmd, $vgname, $name, $status, @buf, @output);

    $self->_clear_err();

    ### get name of disk group
    $vgname = $self->{parent}->get_name();

    ### get name of volume    
    $name = $self->get_name();

    if ($self->is_fs()) { 
        if ($self->is_mounted()) {
            $cmd = "cat " . $self->get_mount() . "/fspec.sig";
        } else {
            $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::get_fspec() Unable to retrieve fspec signature",
                            "Logical volume is currently unmounted", @output);
            
            return ""; 
        }
    } else {
        $cmd = "dd if=/dev/vx/rdsk/" . $vgname ."/" . $name . " count=1|strings";
    }


    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err( "Error: Solaris::Vxvm::Vg::Vol::get_fspec() Unable to get fspec signature",
                         @output);
        return "";                        
    } 

    ### Remove dd summary msg from output
    @buf = grep{$_ !~ /record/} @output;
    return $buf[0];

}

################################################################################
# Subroutine Name: unmount()
# Objective:       unmount volume
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount {
    my ($self, %args) = @_;

    my ($cmd, $type, $vgname, $fs, $mount, $name, $status, $force);
    my ($option, @output); 

    $self->_clear_err();

    $option = "";
    
    # parse hash
    foreach (keys %args) {
        if (/-?force$/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::unmount()",
                                "Invalid value for argument '-force'");
                return $self;
            }        
            next;
        } 
    }  

    $option = "-f" if (defined $force);

    ### Get attributes 
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();
    $mount = $self->{mount};
    $fs = $self->{filesystem};

    ### check if volume contains a filesystem
    unless ($fs !~ /raw/i) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::unmount()",
                        "Cannot unmount a raw device");
        return $ClientUtil::FALSE; 
    }

    ### check if volume is mounted
    unless (defined $mount and length($mount)) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::unmount()",
                        "Cannot unmount a filesystem that is not mounted");
        return $ClientUtil::FALSE; 
    } 

    ### Unmount the volume's filesystem
    $cmd = "/usr/sbin/umount $option $mount";
    ($status, @output) = &ClientUtil::execute_retry($self->{conn}, $cmd, 180, 2); 
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::unmount() Unable to unmount volume's filesystem",
                        @output);
        return $ClientUtil::FALSE;                            
    } 

    ### Remove mount pt metadata from object
    $self->{mount} = "";

    ### Clear the mount point with the vol metadata
    #$self->_set_metadata("comment","");
    #return $ClientUtil::FALSE if ($self->error); 

    ### Clear vfstab mount entry
    ### client = lvm <-- vg <-- vol
    $self->{client}=$self->{parent}{parent}{parent};

    # Comment out for now, can cause reboot to hang with stale entries in vfstab
    #$self->{client}->mount_entry_delete($mount);
    #if ($self->{client}->error) {
        #$self->_set_error($self->{client}->errmsg());
        #return $ClientUtil::FALSE;
    #}   

    ### Remove directory head
    my ($base,$parent) = fileparse($mount);
    $cmd = "rm -rf $mount";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::unmount() Unable to remove mount point '$mount'",
                        @output);
    } 

    return $ClientUtil::TRUE;

}

################################################################################
# P R I V A T E   S U B R O U T I N E S
###############################################################################
################################################################################
# Subroutine Name: _create()
# Objective:       create a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self) = @_;

    my ($cmd, $device_cnt, $dg, $name, $layout, $size, @output, $status);

    $dg = $self->{parent}->get_name();
    $device_cnt = $self->{parent}->get_device_cnt();
    $name = $self->get_name();
    $size = $self->get_size();
    $layout = $self->get_layout();

    $cmd = "vxassist -g $dg make $name $size ";

    if ($device_cnt > 1) {
        $cmd .= "layout=$layout ".  
                "stripe_stripeunit=16k ".
                "max_ncolumn=$device_cnt ".
                "nstripe=$device_cnt";
    }
    
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::_create Unable to create volume \'$name\'", @output);
        return $ClientUtil::FALSE;                            
    } 

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _get_metadata
# Objective:      Retrieve metadata for attribute 
# Details:        See POD documentation at the end of this file
################################################################################
sub _get_metadata {
    my ($self, $attribute) = @_;
    
    my ($cmd, $vgname, $name, @output, $status, $value);
    
    $value = "";
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    $cmd = "vxprint -g $vgname -lv $name | grep $attribute:";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::_get_metadata() ",
                        "Unable to retrieve volume \'$name\' metadata for",
                        "attribute '$attribute' in diskgroup '$vgname'",
                        @output);
        return ""; 
    }
    
    ### return the attribute value if exists.
    if (defined $output[0]) {
        ($value) = $output[0] =~ /$attribute:\s+(.*)/;
    }
    
    return $value;    
    
}

################################################################################
# Subroutine Name: _init_fstype
# Objective:      Initialize fstype for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_fstype {
    my ($self) = @_;
    
    my ($cmd, $fstype, $vgname, $name, @output, $status);
    
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    ### Start volume if not enabled
    if ($self->get_state() =~ /disable/i) {
        $self->_vol_start();
        if ($self->error()) {
            return "";
        }
    }

    $cmd = "fstyp /dev/vx/dsk/$vgname/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status eq 1) {
        $fstype = "raw";
    } elsif ($status > 1) {    
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::_init_fstype() ".
                        "Unable to retrieve  filesystem type for volume \'$name\' in diskgroup '$vgname'",
                        @output);
        $fstype = "";
    } else {
        ($fstype) = $output[0] =~ /(\w+)/i;    
    }

    return $fstype;    
    
}

################################################################################
# Subroutine Name: _init_layout
# Objective:      Initialize layout of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_layout {
    my ($self) = @_;
    
    my ($cmd, $vgname, $layout, $name, @output, $status);
    
    $layout = "";

    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    $cmd = "vxprint -g $vgname -shlv $name | grep layout | awk '{print \$2}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::_init_layout() ".
                        "Unable to retreive layout for volume \'$name\' in diskgroup '$vgname'",
                        @output);
        return "";        
    } 

    ### convert sectors to megabytes
    (undef, $layout) = split(/=/, $output[0]);
    return $layout;
    
}

################################################################################
# Subroutine Name: _init_mount
# Objective:      Initialize mount point for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_mount {
    my ($self, $import, $fsck) = @_;
    
    my ($cmd, $vgname, $mount, $name, $path, @output, $status);
    
    $mount = "";
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    ### Check to see if mount path has been store
    ### in volume metadata under 'comment' attribute.
    ### This is to done in order to associate a mount
    ### pt with a vg/vol in case the vol was renames while
    ### still mounted.
    $cmd = "mount | /usr/xpg4/bin/grep -e \'$vgname/$name$\'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($status > 0) {        
        ### If volumes are bing imported assume then 
        ### automatically mount fs with last stored 
        ### mount point (or default path) if a mount
        ### point does not exist
        if ($import) {
            $mount = $self->_get_metadata("comment");    
            if (defined $mount and length($mount)) {
                ($path,$vgname,$name)=$mount=~/(\S+)\/(\w+)\/(\w+)$/;
            }
            
            $self->mount(-path => $path,
                         -fsck => $fsck);
            return "" if $self->error();
        }
    } else {
        ### return the volume mount point.
        ($mount) = split(/\s+/, $output[0]);
    }
    
    return $mount;    
    
}

################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self) = @_;
    
    my ($cmd, $vgname, $size, $name, @output, $sectors, $status);
    
    $size = "";
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    $cmd = "vxprint -g $vgname -Qqv $name | awk '{ print \$5 }'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::_init_fstype() ".
                        "Unable to retrieve  size (MB) for volume \'$name\' in diskgroup '$vgname'",
                        @output);
        return "";        
    } 

    ### convert sectors to megabytes
    $sectors = $output[0];

    $size = $sectors * 512; # bytes
    $size = $size / 1024;   # kilobytes
    $size = $size / 1024;   # megabytes
    
    return $size;    
    
}

################################################################################
# Subroutine Name: _init_state
# Objective:      Initialize state of existing host volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_state {
    my ($self) = @_;
    
    my ($cmd, $name, $state, $vgname, $status, @output);
    
    $state = "";

    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    $cmd = "vxprint -g $vgname -v $name | grep 'v ' | awk '{print \$4}'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 || grep(/VxVM vxprint/i, @output)) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::_init_state() ".
                        "Unable to retrieve state for volume \'$name\' in diskgroup '$vgname'",
                        @output);
        return "";        
    } 

    ### convert sectors to megabytes
    $state = $output[0];

    return $state;
    
}

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $exist, $fsck, $import, $name, $size, $conn, $layout, $log);
    my ($parent, $self, $status, @output);

    $self = {};

    bless $self, $class;    

    # default layout is concatenation
    $layout = "concat";
    $import = 0;
    $exist = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?layout$/) {
            if (not length($layout = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                                "Invalid value for argument '-layout'");
                return $self;
            }        
            next;
        } elsif (/-?_import$/) {  ### private member
            if (not length($import = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                                "Invalid value for argument '-_import'");
                return $self;
            }
            next;
        } elsif (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                                "Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?size$/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                                "Invalid value for argument '-size'");
                return $self;
            }
            next;
        } elsif (/-?fsck$/) {
            if (not length($fsck = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                                "Invalid value for argument '-fsck'");
                return $self;
            }
            next;
        } elsif (/-?exist$/) {
            if (not length($exist = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                                "Invalid value for argument '-exist'");
                return $self;
            }
            next;
        }
        elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                                "Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                                "Missing value for argument '-parent'");
                return $self;
            }
            next;
        } 
    }  
    
    ### check for existance of required -lvm argument
    unless (defined $parent) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                        "Missing required argument '-parent'");
        return $self; 
    }

    ### check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                        "Missing required argument '-name'");
        return $self; 
    }

    ### check for existance of required -size argument
    unless (defined $size) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::new()",
                        "Missing required argument '-size'");
        return $self; 
    }

    ### check for existance of required -conn argument
    if (grep ( /^-?conn$/, (keys %args) )) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: Solaris::Vxvm::Vg::Vg::new()",
                            "Invalid value for argument '-conn'");
            return $self; 
        }    
    } else {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vg::new()",
                        "Missing required argument '-conn'");
        return $self; 
    }

    # initilize the object members
    $self->{conn} = $conn;
    $self->{name} = $name;
    $self->{layout} = $layout;
    $self->{size} = $size;
    $self->{parent} = $parent;
    $self->{state} = 'ENABLED';
    $self->{filesystem} = 'raw';

    if (not $exist) {
        $self->_create();
        return $self if ($self->error());
    } else {
        $self->sanlog(-msg => "loading existing volume '$name'"); 
        
        $self->{state} = $self->_init_state();
        return $self if ($self->error);
        $self->sanlog(-msg => "state: ".$self->{state}); 
        
        $self->{filesystem} = $self->_init_fstype();
        $self->sanlog(-msg => "fs: ".$self->{filesystem}); 
        if ($self->error()) {
            ### allow vol to be loaded although it 
            ### can't be started to determine fs
            if ($self->{filesystem} eq "") {
                $self->_clear_err();
            }  else {
                return $self;
            }
        }
        
        if ($self->{filesystem} !~ /raw/i) {      
            $self->{mount} = $self->_init_mount($import, $fsck);
            return $self if ($self->error);
            $self->sanlog(-msg => "mnt: " . $self->{mount}); 
        }

        $self->{size} = $self->_init_size();
        return $self if ($self->error);
        $self->sanlog(-msg => "size: " . $self->{size} . " mb"); 

        $self->{layout} = $self->_init_layout();
        return $self if ($self->error);
        $self->sanlog(-msg => "layout: " . $self->{layout});
    }
    
    return $self;

}

################################################################################
# Subroutine Name: _set_metadata
# Objective:      Set metadata for attribute 
# Details:        See POD documentation at the end of this file
################################################################################
sub _set_metadata {
    my ($self, $attribute, $value) = @_;
    
    my ($cmd, $vgname, $name, @output, $status);

    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    $cmd = "vxedit -g $vgname set $attribute='$value' $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd); 
    if ($self->error) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::_set_metadata() ",
                        "Unable to set volume \'$name\' metadata for",
                        "attribute '$attribute' in diskgroup '$vgname'",
                        @output);
        return $ClientUtil::FALSE; 
    }
    
    return $ClientUtil::TRUE;  
    
}

################################################################################
# Subroutine Name: _vol_start
# Objective:      start up volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _vol_start {
    my ($self) = @_;
    
    my ($cmd, $vgname, $name, @output, $status);
    
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    $cmd = "vxvol -g $vgname start $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::_vol_start() ".
                        "Unable to start volume \'$name\' in diskgroup '$vgname'",
                        @output);
        return $ClientUtil::FALSE;    
    } 

    return $ClientUtil::TRUE;
    
}

################################################################################
# Subroutine Name: _vol_stop
# Objective:      stop volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _vol_stop {
    my ($self) = @_;
    
    my ($cmd, $vgname, $name, @output, $status);
    
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    $cmd = "vxvol -g $vgname stop $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Solaris::Vxvm::Vg::Vol::_vol_stop() ".
                        "Unable to stop volume \'$name\' in diskgroup '$vgname'",
                        @output);
        return $ClientUtil::FALSE;    
    } 

    return $ClientUtil::TRUE;
    
}

1;
