# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for the ZFS Logical Volume Manager on Solaris
## @author rroberts@netapp.com
## @author sabnis@netapp.com
## @status private

# Package Name
##############
package San::SanClient::Lvm::Solaris::ZFS;

use strict;

# Package Requirements
######################
use File::Basename;
use File::Copy;
use POSIX;

use San::SanClient::ClientUtil;
use San::SanClient::Lvm;
use San::SanLog;
use Error;

require San::SanClient::Lvm::Solaris::ZFS::Vg;

use vars qw(@ISA);

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: initialize
# Objective:       initialize disk under ZFS control
# Details:         See POD documentation at the end of this file
################################################################################
sub initialize {
    my ($self, %args) = @_;

    my ($attribute, $cmd, $format, @already_inited, @output, $status, $devices);

    ### default disk type
    $format = "n/a";

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: ZFS::initialize() Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### check for required parameters
    unless(defined $devices and length $devices) {
        $self->_set_err("Error: ZFS::initialize() -devices is a required argument");
        return $ClientUtil::FALSE;
    }

    ### Get a list of all the devices that are free (not in a metaset),
    ### online, and already formatted'.
    @already_inited = $self->get_free_devices();

    ### Initialize each device in the user specified devices array
    ### that is not aready inited (not in @already_inited array).

    foreach my $device (@$devices) {
        # initialized devices that have not been inited or are inited with the wrong format
        if (not (grep /^$device$/,@already_inited)) {
            $self->sanlog(-msg => "ZFS::initialize() $device init ($format)", -msg_level => 3);
        } else {
            $self->sanlog(-msg => "ZFS::initialize() $device already inited ($format)", -msg_level => 3);
        }
        $self->sanlog(-msg => "ZFS::initialize() Device '$device' is $format", -msg_level => 3);
    }

    return($ClientUtil::TRUE);

}

################################################################################
# Subroutine Name: create_vg
# Objective:       create and return a Zpool object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vg {
    my ($self, %args) = @_;
    my ($vgObj, $bAdd);

    $bAdd = 1;

    if (exists $args{-no_add}){
        $self->sanlog(-msg => "ZFS::create_vg() LVM object will not manage zpool object", -msg_level => 2);
        delete $args{-no_add};
        $bAdd=0;
    }

    $vgObj = San::SanClient::Lvm::Solaris::ZFS::Vg->_new(-parent => $self,
                                                         -conn => $self->{conn},
                                                         %args);

    if ($vgObj->error) {
        $self->_set_err($vgObj->errmsg());
        return $vgObj;
    }

    if ($bAdd == 1) {
        push @{$self->{vgs}}, $vgObj;
    } else {
        $vgObj = {};
    }

    return $vgObj;
}

################################################################################
# Subroutine Name: destroy_vg
# Objective:       destroy a named Zpool object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vg {
    my ($self, $name) = @_;

    my ($vgObj, @vgs);

    $self->_clear_err();

    unless(defined $name and length $name) {
        $self->_set_err("Error: ZFS::destroy_vg() Zpoolname is a required argument");
        return $ClientUtil::FALSE;
    }

     ### Delete the of vg object and update the parent vg list.
     foreach $vgObj (@{$self->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            $vgObj->destroy();
            if ($vgObj->error) {
                $self->_set_err($vgObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vgs,$vgObj;
        }
    }

    $self->{vgs} = \@vgs;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display
# Objective:       display zpool information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    $self->sanlog(-msg => "ZFS::display() Display zpool information", -msg_level => 9);

    $cmd = "zpool list";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err( "Error: ZFS::display()->  Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [] ;
    }

    return (wantarray()) ? @output : \@output;
}

###############################################################################
# Method:    get_disk_format
# Objective: Get list of supported disk formats  *NO-OP for ZFS
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_format {
    my ($self, $disk) = @_;

    my ($format);

    $self->_clear_err();

    $format = "n/a";

    return $format;

}

###############################################################################
# Method:    get_disk_formats
# Objective: Get list of supported disk formats **NOOP FOR ZFS
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_formats {
    my ($self) = @_;

    my (@format);

    @format = ();

    return (wantarray()) ? @format : \@format;
}

###############################################################################
# Method:    get_format_default
# Objective: Get default type to use for disk format *NOOP FOR ZFS
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_format_default {
    my ($self) = @_;

    my $format;

    $self->_clear_err();

    $format = "n/a";

    return $format;

}

################################################################################
# Subroutine Name: get_vg
# Objective:       return a zpool object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vg {
    my ($self, $name) = @_;

    my ($vgObj);

    $self->_clear_err();

    unless(defined $name and length $name) {
        $self->_set_err("Error: ZFS::get_vg() Zpool name is a required argument");
        return {};
    }

    foreach my $obj (@{$self->{vgs}}) {
        if ($obj->get_name() =~ /^$name$/){
            return $obj;
        }
    }

    $self->_set_err("Error: ZFS::get_vg() Unable to find zpool '$name'");

    return {};

}

################################################################################
# Subroutine Name: get_vgs
# Objective:       return array of zpool objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgs {
    my ($self) = @_;

    my (@vgs);

    $self->_clear_err();

    if (defined $self->{vgs}) {
        @vgs = @{$self->{vgs}};
    } else {
        $self->_set_err("Error: ZFS::get_vgs() No Zpool objects found");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @vgs : \@vgs;

}

################################################################################
# Subroutine Name: get_vgnames
# Objective:       return array of zpool names
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgnames {
    my ($self) = @_;

    my ($cmd, $status, @output);

    $self->_clear_err();

    $cmd = "zpool status -v";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::get_vgnames() Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    my @tmp_array;
    foreach my $line (@output) {
        # e.g. output :-
        #  pool: test_vg
        if ($line =~ /^\s*pool:\s+(\S+)\s+$/) {
            push @tmp_array, $1;
        }
    }

    @output = @tmp_array;

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_pkginfo()
# Objective:       get detailed lvm pkg information
# Details:         See POD documentation at the end of this file
################################################################################
sub get_pkginfo {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{pkginfo});

    return (wantarray()) ? @{$self->{pkginfo}} : $self->{pkginfo};

}

################################################################################
# Subroutine Name: get_revision()
# Objective:       Get revision of ZFS
# Details:         See POD documentation at the end of this file
################################################################################
sub get_revision {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{rev});
    return "" if ($self->error());

    return $self->{rev};

}

################################################################################
# Subroutine Name: get_name()
# Objective:       Get lvm name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{name});
    return "" if ($self->error());

    return $self->{name};

}

################################################################################
# Subroutine Name: get_version()
# Objective:       Check version of ZFS
# Details:         See POD documentation at the end of this file
################################################################################
sub get_version {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $self->_get_pkginfo() unless (defined $self->{vers});
    return "" if ($self->error());

    return $self->{vers};

}


################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;

    my ($bCnt, $cmd, $cnt, $device, $filer, $filername, $lun, $refresh, $filers, $protocol);
    my ($status, $total, $vol, %hDevs, @attr, @buffer, @devices, @output, @used, @sanlun_output);

    $total = 0;

    $self->_clear_err();

    ### Asssume user supplied -cnt argument
    $bCnt = 1;
    $refresh = 0;

     # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Error: ZFS::get_free_devices() Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [] ;
            }
            next;
        } elsif (/-?inited_state/) {
            # NOOP for ZFS
            next;
        } elsif (/-?filer$/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Error: ZFS::get_free_devices() Invalid value for argument '-filer'");
                return (wantarray()) ? () : [] ;
            }
            next;
        } elsif (/-?filers$/) {
            if (not length($filers = $args{$_})) {
                $self->_set_err("Error: ZFS::get_free_devices() Invalid value for argument '-filers'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Error: ZFS::get_free_devices() Invalid value for argument '-vol'");
                return (wantarray()) ? () : [] ;
            }
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Error: ZFS::get_free_devices() Invalid value for argument '-lun'");
                return (wantarray()) ? () : [] ;
            }
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: ZFS::get_free_devices() Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [] ;
            }
            next;
        } elsif (/-?protocol$/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Error: ZFS::get_free_devices() Invalid value for argument '-protocol'");
                return (wantarray()) ? () : [];
            }
        }
    }

    ### Check if user wants free devices from a specified filer,volume, or lun
    if (defined $filer or defined $vol or defined $lun or defined $protocol or defined $filers) {
        ### Check if a device to filer map 'dfmap' has already been created or
        ### refresh it if requested.
        if (scalar(keys %{$self->{dfmap}}) == 0 or $refresh == 1) {
            $self->sanlog(-msg => "ZFS::get_free_devices() refreshing device to filer, vol and lun map",
                          -msg_level => 9);

            # Grab some sanlun output
            @sanlun_output = $self->{sanlun}->lun_show();
            if ($self->{sanlun}->error) {
                $self->_set_err($self->{sanlun}->errmsg());
                return (wantarray()) ? () : [] ;
            }

            # E.g. of output:
            # auto-fas3020-03:  /vol/nate_2_3_auto_sunx2200_01/nate_auto-sunx2200-01_solaris_efi_5  /dev/rdsk/c3t60A9800043346163614A2F6B4E375477d0s2  scsi_vhci0  FCP           15g (16106127360)    GOOD
            #
            foreach my $line (@sanlun_output) {
                chomp($line);

                # Skip the header line
                next if ($line =~ /lun-pathname/i);

                # Remove any leading or trailing whitespace
                $line =~ s/^\s+|\s+$//g;

                $self->sanlog(-msg => "ZFS::get_free_devices() LINE=|$line|", -msg_level => 3);

                # Figure out what we are dealing with from sanlun
                my ($filer_tmp, $vol_tmp, $lun_tmp, $device_tmp, $protocol_tmp);
                if ($line =~ /(\S+)\s+\/vol\/(\S+)\/(\S+)\s+\/dev\/rdsk\/(\S+)s2\s+\S+\s+(\S+)\s+/) {
                        # old sanlun kit output looks like below,
                        # controller:                 lun-pathname                                 device filename                     adapter   protocol          lun size         lun state
                        # iscsi_vs:  /vol/iscsi_vol_1/iscsi_auto-sunx2200-01_0  /dev/rdsk/c0t600A0980324668762D3F41734170386Fd0s2  scsi_vhci0  iSCSI          5g (5368709120)     GOOD
                        #
                        # new sanlun kit output looks like below,
                        # controller(7mode)/                                             device                                            host                  lun
                        # vserver(Cmode)       lun-pathname                              filename                                          adapter    protocol   size    mode
                        # ---------------------------------------------------------------------------------------------------------------------------------------------------
                        # iscsi_vs             /vol/iscsi_vol_1/iscsi_auto-sunx2200-01_0 /dev/rdsk/c0t600A0980324668762D3F41734170386Fd0s2 scsi_vhci0 iSCSI      5g      C
                        ($filer_tmp,$vol_tmp,$lun_tmp,$device_tmp,$protocol_tmp) = ($1,$2,$3,$4,lc($5));
                } elsif ($line =~ /(\S+):\s+\/vol\/(\S+)\/(\S+)\s+\/dev\/rdsk\/(\S+)s2/) {
                    # No Protocol
                    ($filer_tmp,$vol_tmp,$lun_tmp,$device_tmp,$protocol_tmp) = ($1,$2,$3,$4,"");
                } else {
                    next;
                }

                # Remove colon in the end of filer/vserver name if presents
                $filer_tmp =~ s/:$//;

                $self->sanlog(-msg => "ZFS::get_free_devices() Filer: $filer_tmp   Vol: $vol_tmp   " .
                                      "Device: $device_tmp   Lun: $lun_tmp",
                              -msg_level => 9);
                $hDevs{$device_tmp} = {
                                         filer    => $filer_tmp,
                                         vol      => $vol_tmp,
                                         lun      => $lun_tmp,
                                         protocol => $protocol_tmp,
                                      };
            }

            ### save map for next call of function
            $self->{dfmap} = \%hDevs;
        }
    }

    ### If not -cnt then assume all free disks
    ### under lvm control are wanted
    unless (defined $cnt and length($cnt)) {
        $cnt = 10000;
        $bCnt = 0;
    }

    # Check to see if we already ran sanlun.  If not, then run it.
    unless (scalar(@sanlun_output) > 0) {
            # Grab some sanlun output
            @sanlun_output = $self->{sanlun}->lun_show();
            if ($self->{sanlun}->error) {
                $self->_set_err($self->{sanlun}->errmsg());
                return (wantarray()) ? () : [] ;
            }
    }

    # Run zpool status and save the output so we can use it to determine
    # if the device is in a zpool already.
    my $zp_cmd = "zpool status";
    my ($zp_status, @zp_output) = &ClientUtil::execute($self->{conn}, $zp_cmd);
    if ($zp_status > 0) {
        $self->_set_err("Error: ZFS::get_free_devices() " .
                        "Command ($zp_cmd) returned a bad status ($zp_status)");
        return (wantarray()) ? () : [];
    }

    # iterate through the devices and make sure that none of the disks
    # are being used by another LVM disk object.
    foreach my $line (@sanlun_output) {
        # Skip the header line
        next if ($line =~ /lun-pathname/i);
        my $dev = "";
        # E.g. of output:
        # fas880-rtp02: /vol/nate_1_1/nate_sun280r-fak01_1        /dev/rdsk/c4t500A098192D0DFB6d5s2           lpfc0       FCP         1g (1073741824)     GOOD
        if ($line =~ /\/dev\/rdsk\/(\S+)s2\s+/) {
            $dev = $1;

            # See if the device is used in a zpool
            if (grep(/\s+${dev}\s+/, @zp_output) > 0) {
                $self->sanlog(-msg => "Skipping device '$dev' since it is used in a zpool",
                              -msg_level => 3);
                next;
            }

            # Apply filters to any of the free devices found
            if (defined $filer and length $filer) {
                $self->sanlog(-msg => "\nFilter Filer: $filer\nDevice: $dev\nDevice Filer: " .
                                      $self->{dfmap}{$dev}{filer},
                              -msg_level => 3);
                next if ($self->{dfmap}{$dev}{filer} !~ /$filer/i);
            }

            if (defined $vol and length $vol) {
                $self->sanlog(-msg => "\nFilter Vol: $vol\nDevice: $dev\nDevice Vol: " .
                                      $self->{dfmap}{$dev}{vol},
                              -msg_level => 3);
                next if ($self->{dfmap}{$dev}{vol} !~ /$vol/i);
            }

            if (defined $lun and length $lun) {
                $self->sanlog(-msg => "\nFilter LUN: $lun\nDevice: $dev\nDevice LUN: " .
                                      $self->{dfmap}{$dev}{lun},
                              -msg_level => 3);
                next if ($self->{dfmap}{$dev}{lun} !~ /$lun/i);
            }

            if (defined $protocol and length $protocol) {
                $self->sanlog(-msg => "\nFilter protocol: $protocol\nDevice: $dev\nDevice protocol: " .
                                      $self->{dfmap}{$dev}{protocol},
                              -msg_level => 3);
                next if ($self->{dfmap}{$dev}{protocol} !~ /$protocol/i);
            }

            # Filter on a filer list if provided by caller
            if (defined $filers and scalar(@{$filers}) != 0) {
                my $found = 0;
                # Look through each filer in the filers list
                foreach my $filer_f (@{$filers}) {
                    $self->sanlog(-msg => "\nFilter filer(s): $filer_f\nDevice: $dev\nDevice filer: " .
                                          $self->{dfmap}{$dev}{filer},
                                  -msg_level => 3);
                    # If filer matches devices filer, set found to 1
                    if ($self->{dfmap}{$dev}{filer} =~ /^$filer_f$/i) {
                        $found = 1;
                    }
                }

                # If the device is not part of the filers list caller requested skip it.
                if ($found == 0) {
                    next;
                }
            }

            # See if the lun has a volume label on it
            my $prt_cmd = "/usr/sbin/prtvtoc /dev/rdsk/" . $dev;
            if ($dev !~ /s2/) {
                $prt_cmd .= "s2";
            }
            my ($prt_status, @prt_output) = &ClientUtil::execute($self->{conn}, $prt_cmd);
            if ($prt_status > 0) {
                $self->_set_err("Error: ZFS::get_free_devices() " .
                                "Command ($prt_cmd) returned a bad status ($prt_status)");
                return (wantarray()) ? () : [];
            }

            # Example output:
            # * /dev/rdsk/c4t60A98000486E544D575A4B58646A5774d0s2 (volume "NATE") partition map
            # Grep out just the line with the device in it (this has the words "partition map" in it)
            @prt_output = grep(/partition map/, @prt_output);

            # Check to see if the line contains volname of reserved types
            if (grep(/root|boot|reserved/, @prt_output)) {
                $self->sanlog(-msg => "ZFS::get_free_devices() ignoring device because " .
                                      "it has a special volume label ($dev)");
                next;
            }

            $self->sanlog(-msg => "ZFS push $dev", -msg_level => 9);
            push (@devices,$dev) if (++$total <= $cnt);
        }
    }

    ### If user supplied cnt and free devices fell short
    ### generate error
    if ($bCnt == 1 and scalar(@devices) < $cnt) {
        $self->_set_err( "Only " . scalar(@devices) . " of $cnt requested devices were found");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: load_vg
# Objective:       load an existing diskgroup that is not in memory
# Details:         See POD documentation at the end of this file
################################################################################
sub load_vg {
    my ($self, $name) = @_;

    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name){
        $self->_set_err("ERROR: ZFS::load_vg() Invalid or missing required parameter - zpool name \$name");
        return {};
    }

    ### Check if vg has already been loaded
    $vgObj = $self->get_vg($name);
    if ($self->error) {
        # dg has not been load yet
        $self->_clear_err();

        # load dg
        $self->_create_vgs($name);
        return {} if ($self->error);
        $vgObj = $self->get_vg($name);
    } else {
        $self->sanlog(-msg => "INFO: ZFS::load_vg() Volume group '$name' was already loaded", -msg_level => 9);
    }

    return $vgObj;

}


################################################################################
# Subroutine Name: unload_vg
# Objective:       Remove the vg object under LVM object control
# Details:         See POD documentation at the end of this file
################################################################################
sub unload_vg {
    my ($self,$name) = @_;

    my ($vgObj, $bFound, @vgs);

    $self->_clear_err();

    unless (defined $name and length $name){
        $self->_set_err("ERROR: ZFS::unload_vg() Invalid or missing required parameter - zpool name \$name");
        return $ClientUtil::FALSE;
    }

    $bFound = 0;

    foreach $vgObj ( @{$self->{vgs}}){
        ($vgObj->get_name() =~ /^$name$/)
            ? $bFound=1
            : push @vgs, $vgObj;
    }

    $self->{vgs} = \@vgs;

    unless($bFound) {
        $self->_set_err("ERROR: ZFS::unload_vg() Nothing to unload. Vg name '".$name."' was not loaded");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: enable_controllers
# Objective:       NOOP FOR ZFS
# Details:         See POD documentation at the end of this file
################################################################################
sub enable_controllers {
    my ($self) = @_;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: rescan
# Objective:       NOOP FOR ZFS
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: restart_recovery_daemon
# Objective:       NOOP FOR ZFS
# Details:         See POD documentation at the end of this file
################################################################################
sub restart_recovery_daemon {
    my ($self, $interval) = @_;

    return $ClientUtil::TRUE;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($conn, $log, $lvm, $loadcfg, $parent, $self);

    $self = {};

    bless $self,$class;

    # creating lvm object for ZFS

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_})) {
                $self->_set_err("Error: ZFS::_new() Invalid value for argument '-lvm'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: ZFS::_new() Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?loadcfg/) {
            if (not length($loadcfg = $args{$_}) or $loadcfg !~ /no|yes|1|0|off|on/i) {
                $loadcfg=1;
            }
            next;
        }
    }

    ### check for existance of required -lvm argument
    if (grep(/^-lvm$/, (keys %args) )) {
        unless ($args{-lvm} =~ /zfs/i) {
            $self->_set_err("Error: ZFS::_new()",
                            "Invalid value for argument '-lvm' \(".$args{-lvm}."\)");
            return $self;
        }
    } else {
        $self->_set_err("Error: ZFS::_new()",
                        "Missing required argument'-lvm'");
        return $self;
    }

    ### check for existance of required -parent argument
    if (grep ( /^-?parent$/,(keys %args))) {
        unless (ref($args{-parent}) =~ /SanClient/i) {
            $self->_set_err("Error: ZFS::_new()",
                            "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: ZFS::_new()",
                        "Missing required argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn} = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{vgs} = [];

    ### check to make sure that LVM is installed
    if (not $self->_is_installed()) {
        return $self if ($self->error);
    }

    ### create ZFS vg objects for existing disk groups
    if ($loadcfg) {
        $self->_create_vgs() ;
        return $self if ($self->error);
    }

    # returning LVM object created for ZFS
    return($self);

}

################################################################################
# Subroutine Name: _get_pkginfo
# Objective:       Get lvm package info
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self) = @_;
    my ($cmd, $name, $vers, $rev, @output, $status);

    $cmd = "/usr/bin/pkginfo -l SUNWzfskr";
    ($status, @output) = ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err( "Error: ZFS::_get_pkginfo() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        ($name) = $line =~ /PKGINST:\s+(\w+)/i if ($line =~ /PKGINST/i);
        ($vers,$rev) = $line =~ /VERSION:\s+(.*),REV=\b(.*)\b/i if ($line =~ /VERSION/i);
    }

    $self->{name} = $name;
    $self->{vers} = $vers;
    $self->{rev} = $rev;
    $self->{pkginfo} = \@output;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _create_vgs
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vgs {
    my ($self, $vgname) = @_;
    my ($cmd, @output, @mount_output, $status, $vgObj);

    # get the root filesystem mount point to exclude
    $cmd = "df -kl / \| grep -v \"Filesystem\"";
    ($status, @mount_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::_create_vgs()",
                        "Unable to get root mount point");
        return $ClientUtil::FALSE;
    }

    ### Retrieve the volume groups.
    $cmd = "zpool status -v|grep pool";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::_create_vgs()",
                        "Unable to load existing zpool\(s\)");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        if ($line =~ /^\s*pool\:\s+(\S+)/) {
            my $zpool_name = $1;
            # look to see if zpool name is mounted as / and exclude
            if ((grep(/^$zpool_name/, @mount_output) == 0) and
                $zpool_name !~ /boot|reserved/i) {
                $self->create_vg(-name => $zpool_name,
                                 -deported => 0);
            } else {
                $self->sanlog(-msg =>
                              "ZFS::_create_vgs() excluding zpool ($zpool_name) " .
                              "since it looks like it is mount as / or it's name contains reserved words");
            }
        }
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_disk_id
# Objective:       Return disk id
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_disk_id {
    my ($self, $device) = @_;

    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    $name = "/dev/rdsk/$device" . "s2";

    my $path = $self->{parent}->get_hu_bin_path();
    $cmd = $path . "/sanlun lun show -v -d $name | grep 'Serial number'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err( "Error: ZFS::_get_disk_id() Unable to get diskid for device \'$device\'",
                         @output);
        return $ClientUtil::FALSE;
    }

    return $output[0];

}

################################################################################
# Subroutine Name: _is_installed
# Objective:       Check to make sure that the LVM is installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_installed {
    my ($self) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "/usr/bin/pkginfo SUNWzfskr";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status == 0) {
        $self->sanlog(-msg => "ZFS::_is_installed() ZFS is installed", -msg_level => 9);
        return $ClientUtil::TRUE;
    } else {
        $self->_set_err("Error: ZFS::_is_installed() ZFS is not installed",
                        "Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }
}

1;
