# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for the ZFS Zpools (Volume Groups) on Solaris
## @author rroberts@netapp.com
## @author sabnis@netapp.com
## @status private
## @pod ../../Vg.pod

package San::SanClient::Lvm::Solaris::ZFS::Vg;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

require San::SanClient::Lvm::Solaris::ZFS::Vg::Vol;

use Tie::IxHash;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new zpool (volume group) object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $conn, $devices, $disk_cnt, $parent, $log, $name, $self);
    my ($status, $deported, @output, %hDevice, $zfs_checksum_enabled);

    $self={};

    bless $self,$class;

    $self->_clear_err();

    $deported = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
            $self->_set_err("Error: ZFS::Vg::new()",
                            "Invalid value for argument '-devices'");
            return $self;
            }
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::new()",
                                "Missing value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::new()",
                                "Missing value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::new()",
                                "Missing value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?deported/) {
            if (not length($deported = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::new()",
                                "Missing value for argument '-deported'");
                return $self;
            }
            next;
        }
    }
    
    ### check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: ZFS::Vg::new()",
                        "Missing required argument '-name'");
        return $self;
    }

    ### check for existance of required -conn argument
    if (grep ( /^-?conn$/, (keys %args) )) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: ZFS::Vg::new()",
                            "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: ZFS::Vg::new()",
                        "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn} = $conn;
    $self->{name} = $name;
    $self->{deported} = $deported;
    $self->{disk_cnt} = 0;
    $self->{parent} = $parent;

    # create the zpool if devices were supplied
    if (defined $devices and ref($devices) eq "ARRAY") {
        $self->_create($name,$devices);
        if ($self->error()) {
            return $self;
        }
    } else {
        # vg must already exist. set device list
        # for all existing and deported dgs.
        $self->sanlog(-msg => "loading existing zpool '$name'", -msg_level => 3);
        $self->{devices} = $self->_get_devices();
        return $self if ($self->error);

        unless ($self->{deported}) {
            $self->_create_vols();
            return $self if ($self->error);
        }
    }

    # Set the ZFS checksum option for the zpools. Checksum should be off unless otherwise set by the caller.
    if (defined($self->{parent}->{parent}->{zfs_checksum_enabled})) {
        $zfs_checksum_enabled = $self->{parent}->{parent}->{zfs_checksum_enabled}; 
    } else {
        $zfs_checksum_enabled = 0;
    }

    if ($zfs_checksum_enabled == 0 or $zfs_checksum_enabled eq "") {
        $cmd = "zfs set checksum=off $name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ZFS::Vg::new()",
                            "Unable to disable zfs checksum on zpool $name");
            return $self;
        }
    } elsif ($zfs_checksum_enabled == 1) {
        $cmd = "zfs set checksum=on $name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ZFS::Vg::new()",
                            "Unable to enable zfs checksum on zpool $name");
            return $self;
        }
    } else {
       $self->_set_err("Error: ZFS::Vg::new()",
                       "Invalid value provided for ZFS_CHECKSUM_ENABLED");
             return $self;
    } 

    return $self;

}

################################################################################
# Subroutine Name: _create_vols
# Objective:       create vol object for each zpool on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vols {
    my ($self, %args) = @_;

    my ($cmd, $import, $name, $status, $vgObj, $volObj, %hDeport);
    my (@output, @output_new, $fs_name, $fs_size);

    ### Identify volumes that are being imported so
    ### that they are automatically mounted with their
    ### last filesystem mount point.
    foreach (keys %args) {
        if (/-?_import/) {
            if (not length($import = $args{$_})) {
                $import = 0;
            }
            next;
        }    
    }

    $name = $self->get_name();
    
    $cmd = "zfs list -t filesystem -o name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::_create_vols()",
                        "Unable to load existing volumes in zpool",
                        @output);
        return $ClientUtil::FALSE;
    }
    
    # Eg of output:
    # NAME
    # n_vg
    # n_vg/nvol-1
    # n_vg/nvol-2
    # n_vg/nvol-3
    # n_vg/nvol-4
    # n_vg/nvol-5
    
    # Creating volume object for each of the filesytems in the specified zpool. 
    foreach my $line(@output) {
        if ($line =~ /$name\/(\S+)$/) {
            $fs_name = $1;  

            # Setting $fs_size to 1M as an arbirtrary value.  Since we are setting -exist to 1, 
            # the Vol code ignores the passed in value (even though it is required) and uses
            # it's own routine to determine the size of the volume.
  
            $fs_size = "1M";

            $self->sanlog(-msg => "ZFS::Vg::_create_vols() - Creating object for filesystem $fs_name.");
            $volObj = $self->create_vol(-name => $fs_name,
                                        -size => $fs_size,
                                        -exist => 1);
            return $ClientUtil::FALSE if ($self->error);
        } 
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_vol
# Objective:       create and return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vol {
    my ($self, %args) = @_;

    my ($volObj);

    $volObj = San::SanClient::Lvm::Solaris::ZFS::Vg::Vol->_new(-parent => $self,
                                                               -conn => $self->{conn},
                                                               %args);

    if ($volObj->error) {
        $self->_set_err($volObj->errmsg());
        return $volObj;
    }

    push @{$self->{vols}}, $volObj;

    return $volObj;
}

################################################################################
# Subroutine Name: destroy_vol
# Objective:       destroy a named vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vol {
    my ($self, $name) = @_;

    my ($volObj, @vols); 

    $self->_clear_err();

    unless(defined $name and length $name){
        $self->_set_err("Error: ZFS::Vg::destroy_vol() volname is a required argument");
        return $ClientUtil::FALSE
    }

    ### Delete the of vg object and update the parent vg list.
    foreach $volObj (@{$self->{vols}}) {
        if ($volObj->get_name() =~ /^$name$/) {
            $volObj->destroy(-force => 1);
        } else {
            push @vols,$volObj;
        }
        if ($volObj->error) {
                $self->_set_err($volObj->errmsg());
                return $ClientUtil::FALSE;
        }   
    }

    $self->{vols} = \@vols;
    return $ClientUtil::TRUE;
}


################################################################################
# Subroutine Name: _get_devices
# Objective:       get list of devices used by zpool 
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices {
    my ($self) = @_;

    my ($cmd, $name, $deported, $status, @output, %hDevice, $disk_cnt);

    $name = $self->get_name();
    $deported = $self->{deported};

    $cmd = "zpool status $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::_get_devices() Unable to get devices for zpool \'$name\'",
                        @output);
        return {};
    }

    $disk_cnt = 0;
    foreach my $line (@output) {
        # e.g.-           c4t60A980004F6444524A3445486659454Dd0  ONLINE       0     0     0
        if ($line =~ /^\s*(c\d+t\S+d\d+)\s*/) {
            $self->{$1} = $1;
            $hDevice{$1} = $1;
            $disk_cnt++;
        }
    }

    $self->{disk_cnt} = $disk_cnt;
    return \%hDevice;

}


################################################################################
# Subroutine Name: _create()
# Objective:       create a zpool
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self, $name, $devices) = @_;

    my ($attribute, $cmd, $disk_cnt, $disk_name, $format);
    my (@output, $status, %hDevice);

    $disk_cnt = 0;
    $attribute = "";
    $format = "n/a"; 

    # Check to see if VG already exists and if it does return
    # "cannot open 'test': no such pool"
    
    $cmd = "zpool status $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    
    if (grep(/state: ONLINE/i, @output) == 1) {
        foreach my $device (@$devices){
            $hDevice{$device} = $name;
        }
        $self->{devices} = \%hDevice;
        $self->{disk_cnt} = scalar(@$devices);
        return $ClientUtil::TRUE;
    }
    
    # Formatting device names in device array
    my @dev_array;
    foreach my $device(@$devices) {
        # remove /dev/rdsk from the begining of device name
        $device =~ s/^\/dev\/rdsk\///;
        # remove s# from the end to get the device name
        $device =~ s/s\d+$//;
        push @dev_array,$device;
    }
    
    # Create the zpool 
    $cmd = "zpool create $name $dev_array[0]";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::_create() - Unable to create zpool $name",
                        @output);

        return $ClientUtil::FALSE;
    }
    
    my $device;
    $hDevice{$device} = $name;
    # Setting disk count to 1 because 1 disk(device) was used while creating zpool
    $self->{disk_cnt} = 1; 
    
    # Shifting the device array to get the next device to be added to zpool
    shift @dev_array;

    # Adding remaining devices to zpool
    foreach $device (@dev_array) {
        # determine if add_device worked
        if (not $self->add_device($device)) {
            $self->_set_err("Error: ZFS::Vg::_create() - Unable to add remaining devices to zpool $name",
                            @output);
            return $ClientUtil::FALSE;
        }
        $hDevice{$device} = $name;
    }

    $self->{devices} = \%hDevice;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: add_device()
# Objective:       add a device to the zpool 
# Details:         See POD documentation at the end of this file
################################################################################
sub add_device {
    my ($self, $device) = @_;

    my ($cmd, $disk_cnt, $disk_name, $name, @output, $status);

    $self->_clear_err();

    ### Get name of zpool
    $name = $self->get_name();
    
    # Removing slice number from the end of device name
    # e.g. - c5t60A980004F644C774234454865784D67d0s2 (Removing s2)
    $device =~ s/s\d+$//;
    
    $cmd = "zpool add -f $name $device";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::add_device() Unable to add $device to zpool $name",
                        @output);

        return $ClientUtil::FALSE;
    }

    $self->{devices}->{$device} = $device;
    $self->{disk_cnt} += 1; 
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remove_device()
# Objective:       remove a device from the zpool 
# Details:         See POD documentation at the end of this file
################################################################################
sub remove_device {
    my ($self, %args) = @_;

    my ($cmd, $device, $force, $disk_cnt, $disk_name, $name);
    my (@output, @vols_mount, $status);

    $self->_clear_err();

    $force = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::remove_device()",
                                "Invalid value for argument '-device'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::remove_device()",
                                "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            $force = "-f";
            next;
        }
    }
    
    ### get name of zpool 
    $name = $self->get_name();

    ### verify the device is being used in zpool 
    unless (exists $self->{devices}->{$device}) {
        $self->_set_err("Error: ZFS::Vg::remove_device()",
                        "Device '$device' doesn't exist in zpool");
        return $ClientUtil::FALSE;
    }

    $disk_name = $self->{devices}->{$device};

    ### if no force option is specified umount all volumes so deport
    ### can occur without error.  Make note of which vols were
    ### mounted so they can be remounted after removing device
    if (not defined $force) {
        my @volObjs = $self->get_vols();
        return $ClientUtil::FALSE if ($self->error);
        foreach my $volObj (@volObjs) {
            if ($volObj->is_fs()) {
                if ($volObj->is_mounted()) {
                    $self->sanlog(-msg => "unmounting filesystem on volume before removing the device" .
                                          $volObj->get_name(),
                                  -msg_level => 9);
                    $volObj->unmount();
                    if ($volObj->error) {
                        $self->_set_err($volObj->errmsg());
                        return $ClientUtil::FALSE;
                    }   
                    push @vols_mount,$volObj;
                }
            }
        }
    }

    $cmd = "zpool remove $force $name $device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::remove_device Unable to remove $device from zpool $name",
                        @output);

        return $ClientUtil::FALSE;
    }

    ### remove device from hash
    delete($self->{devices}->{$device});

    ### decrement device counter;
    $self->{disk_cnt}--;

    ### Restart and mount volumes that were previously mounted.
    foreach my $vol (@vols_mount) {
        $vol->mount();
        if ($vol->error){
            $self->_set_err($vol->errmsg());
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rename()
# Objective:       rename zpool - THIS IS A NOOP FOR ZFS 
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, %args) = @_;
    
    $self->_set_err( "Error: ZFS::Vg::rename() ZFS doesn't have a provision for renaming zpool");

    return $ClientUtil::FALSE;

}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a zpool 
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;

    my($cmd, $name, @output, $status, @vgs);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: ZFS::Vg::destroy() Zpool name is a required parameter");
        return $ClientUtil::FALSE;
    }

    $cmd = "zpool destroy $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        if (grep (/pool is faulted/i, @output) == 1) {
            foreach my $line(@output) {
                $self->sanlog(-msg=> "$line", -msg_level => 1);
            } 
        }
        $self->_set_err("Error: ZFS::Vg::destroy() Unable to destroy zpool $name", @output);

        return $ClientUtil::FALSE;
    }

    ### Remove the vgobj from the parent lvm object
    foreach my $vgObj ( @{$self->{parent}->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/){
            next;
        } else {
            push @vgs,$vgObj;
        }
    }

    $self->{parent}->{vgs}=\@vgs;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;

    my (@verbose, $cmd, $name, @output, $status, $header);
    my ($summary, @summary, $verbose);

    $self->_clear_err();

    $verbose = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::display()",
                                "Invalid value for argument '-verbose'");
                return (wantarray()) ? () : [];
            }   
            next;
        } 
    }

    $name = $self->get_name();

    unless (defined $name){
        $self->_set_err("Error: ZFS::Vg::display() zpool name is a required parameter");
        return (wantarray()) ? () : [];
    }
    
    push @summary, "VG Name:  ".$self->get_name();
    push @summary, "Device Cnt:  ".$self->get_device_cnt();

    foreach my $device ($self->get_devices()) {
        push @summary, "Device:  ".$device." (".$self->{parent}->get_disk_format($device).")";
    }
    
    if (defined $verbose and $verbose =~ /1|on|y|yes|t|true/i) {

        $cmd = "zpool status $name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ZFS::Vg::display() Unable to display info for zpool $name",
                            @output);
            return (wantarray()) ? () : [];
        }

        @verbose = (@summary, "\nZpool Detail:", @output);
        push @verbose,"\n\n";
        return (wantarray()) ? @verbose : \@verbose;
    } else {
        push @summary,"\n\n";
        return (wantarray()) ? @summary : \@summary;
    }

}

################################################################################
# Subroutine Name: get_device_cnt
# Objective:       Return number of devices in zpool 
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device_cnt {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{disk_cnt}) {
        return $self->{disk_cnt};
    } else {
        $self->_set_err( "Error: ZFS::Vg::get_device_cnt() Unable to display device count for zpool");
        return 0;
    }

}

################################################################################
# Subroutine Name: get_devices
# Objective:       Return devices in zpool
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    @output = (keys %{$self->{devices}});

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vg object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: ZFS::Vg::get_name() Vg object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vols
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vols {
    my ($self) = @_;
    my (@volObjs);

    $self->_clear_err();
    if (defined $self->{vols}) {
        @volObjs = @{$self->{vols}};
    } else {
        $self->_set_err("Error: ZFS::Vg::get_vols() Vg object has no volumes");
        return (wantarray()) ? () : [];;
    }

    return (wantarray()) ? @volObjs : \@volObjs;
}

################################################################################
# Subroutine Name: get_vol_cnt
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol_cnt {
    my ($self) = @_;

    $self->_clear_err();
    if (defined $self->{vols}) {
        return scalar(@{$self->{vols}});
    } else {
        $self->_set_err("Error: ZFS::Vg::get_vol_cnt() Vg object has no vol count");
        return 0;
    }
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;
    my ($cmd, $name, @output, $status, $size);

    $self->_clear_err();

    $size = 0;

    $name = $self->get_name();

    $cmd = "zpool list -o name,size";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::get_size() Unable to get size for zpool $name",
                        @output);
        return 0;
    }

    # Eg.format of output   
    # NAME                  SIZE
    # test                  1.97G

    if (grep(/$name/,@output) == 1) { 
        foreach my $line(@output) {
            chomp($line);
            if ($line =~ /${name}\s+(\S+)/) {
                $size = $1;
                last;
            }
        }
    } else {
        $self->_set_err("Error: ZFS::Vg::get_size() Unable to get size for zpool $name",
                @output);
        return 0;
    }

    # Converting unit of size to MB
    $size = &ClientUtil::convert_to_bytes($size, "m");
    $size = int($size);

    return $size;

}

################################################################################
# Subroutine Name: get_freespace
# Objective:       Return available free space
# Details:         See POD documentation at the end of this file
################################################################################
sub get_freespace {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size);

    $self->_clear_err();

    $size = 0;
    $name = $self->get_name();

    $cmd = "zpool list -o name,available";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::get_freespace() Unable to get size for zpool $name",
                         @output);
        return 0;
    }
    
    # E.g. of output:-
    # NAME                  AVAIL
    # test                  1.97G

    if (grep(/$name/,@output) == 1) { 
        foreach my $line(@output) {
            chomp($line);
            if ($line =~ /${name}\s+(\S+)/) {
                $size = $1;
                last;
            }
        }
    } else {
        $self->_set_err("Error: ZFS::Vg::get_freespace() Unable to get size for zpool $name",
                @output);   
        return 0;
    }

    # Converting size to MB
    $size = &ClientUtil::convert_to_bytes($size, "m");
    $size = int($size);

    return $size;
}

################################################################################
# Subroutine Name: is_deported
# Objective:       Determine if zpool has been deported
# Details:         See POD documentation at the end of this file
################################################################################
sub is_deported {
    my ($self) = @_;

    $self->_clear_err();

    return $self->{deported};
}


################################################################################
# Subroutine Name: deport()
# Objective:       deport Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub deport {
    my ($self) = @_;

    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: ZFS::Vg::deport() Zpool name is a required parameter");
        return $ClientUtil::FALSE;
    }

    $cmd = "zpool export $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::deport() Unable to deport zpool $name",
                        @output);

        return $ClientUtil::FALSE;
    }

    ### Remove volume objects from vg cache
    $self->{vols} = ();
    $self->{deported} = 1;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: import()
# Objective:       Import Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub import {
    my ($self, %args) = @_;
    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: ZFS::Vg::import() zpool name is a required parameter");
        return $ClientUtil::FALSE;
    }

    $cmd = "zpool import $name";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

    if ($status > 0) {
        $self->_set_err( "Error: ZFS::Vg::import() Unable to import zpool \'$name\'",
                @output);

        return $ClientUtil::FALSE;
    }

    $self->{deported} = 0;

    ### requery device list
    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    $self->_create_vols(-_import => 1,
                        %args);
    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;
}


################################################################################
# Subroutine Name: refresh
# Objective:       refresh content of zpool 
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh {
    my ($self) = @_;
    my ($cmd, $status, @output);

    $self->_clear_err();

    ### Remove existing vol cache
    $self->{vols} = undef;

    $self->sanlog(-msg => "refreshing zpool '" . $self->get_name() . "'", 
                  -msg_level => 3);

    ### Determine if dg is deported
    $cmd = "zpool status| grep '(".$self->get_name().")'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::_refresh()",
            "Unable to load deported zpools");

        return $ClientUtil::FALSE;
    }

    $self->{deported} = 1 if (@output);

    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    if (not $self->is_deported()) {
        $self->_create_vols();
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;
}

1;
