# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface for the ZFS filesystems  
## @author rroberts@netapp.com
## @author sabnis@netapp.com
## @status private
## @pod ../../../Vol.pod

package San::SanClient::Lvm::Solaris::ZFS::Vg::Vol;

use strict;

use San::SanClient::ClientUtil;
use San::SanLog;
use Error;

use Data::Dumper;
use File::Basename;

use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
# Method:    create_fs
# Objective: create a filesystem 
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_fs {
    my ($self, %args) = @_;

    my $fstype;

    # This function only sets fs type attribute of the object
    # as the zfs filesystems are directly created on the zpools
    # by the _create function

    # parse arguments
    foreach (keys %args) {
        if (/-?type$/) {
            $fstype = $args{$_};
        }
    }

    # Setting the default fs type to zfs 
    # since for now, zfs is tested only with its own filesystems.
    if (not defined $fstype  or $fstype eq "") {
        $fstype = "zfs";
    }

    if ($fstype !~ /zfs/i) {
        $self->_set_err("Error: ZFS::Vg::Vol::create_fs()",
                        "Invalid value for argument '-type'",
                        "FS type allowed is only zfs.");
        return $ClientUtil::FALSE;
    }

    $self->{filesystem} = $fstype;  

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a filesystem in a zpool
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self, %args) = @_;

    my ($cmd, $force, $vgname, $name, @output, $status, @vols, $option);

    $self->_clear_err();
    $option = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::Vol::destroy()",
                                "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
        next;
        } 
    }

    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: ZFS::Vg::Vol::destroy() Invalid Vol name - $name");
        return $ClientUtil::FALSE;
    }

    ### if force option specified umount all volumes so deport
    ### can occur without error.  Make note of which vols were
    ### mounted so they can be remounted after rename occurs
    if (defined $force) {
        $option = "-f";
        $self->sanlog(-msg=> "ZFS::Vg::Vol::destroy() determine if volume has filesystem");
        if ($self->is_fs()) {
            $self->sanlog(-msg=> "ZFS::Vg::Vol::destroy() determine if volume is mounted");
            if ($self->is_mounted()) {
                $self->sanlog(-msg=> "ZFS::Vg::Vol::destroy() unmount filesystem ".$self->get_name());
                $self->unmount();
                return $ClientUtil::FALSE if ($self->error());
            }
        }
    }

    $cmd = "zfs destroy $option $vgname/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::Vol::destroy() Unable to destroy volume \'$name\' in zpool '$vgname'",$cmd,
                        @output);
        return $ClientUtil::FALSE;

    }

    ### Remove the (filesystem) volume from the parent vol group object
    foreach my $volObj (@{$self->{parent}->{vols}}) {
        if ($volObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vols,$volObj;
        }
    }
    
    $self->{parent}->{vols} = \@vols;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;

    my (@buf, $cmd, $vgname, $name, @output, $status, $header, $summary);

    $self->_clear_err();

    $vgname = $self->{parent}->get_name();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: ZFS::Vg::Vol::display() volume name is a required parameter");
        return (wantarray()) ? () : [];
    }

    $cmd = "zfs list -t filesystem $vgname/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err( "Error: ZFS::Vg::Vol::display() Unable to display info for volume \'$name\' in volume group \'$vgname\'",
                         @output);

        return (wantarray()) ? () : [];
    }

    $header =
              sprintf("\nVol Summary:\n".
                      "%-12.12s %-7.7s %-5.5s %-12.12s %-s\n".
                      "%-12.12s %-7.7s %-5.5s %-12.12s %-32.32s",
                      "NAME","LAYOUT","TYPE","SIZE(mb)","MOUNT_PT",
                      "="x12,"="x7,"="x5,"="x12,"="x32);

    $summary =
               sprintf("%-12.12s %-7.7s %-5.5s %-12.12s %-s",
                       $self->get_name(),
                       $self->get_layout(),
                       $self->get_fs(),
                       $self->get_size(),
                       $self->get_mount());

    @buf = ($header,$summary,"\nVol Detail:",@output);

    return (wantarray()) ? @buf : \@buf;
}
                                             
################################################################################
# Subroutine Name: get_fs
# Objective:       Return Vol fs type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{filesystem} and length $self->{filesystem}) {
        return $self->{filesystem};
    } else {
        $self->_set_err("Error: ZFS::Vg::Vol::get_fs() Vol object is missing filesystem");
        return "";
    }
}

################################################################################
# Subroutine Name: get_layout
# Objective:       Return volume layout
# Details:         See POD documentation at the end of this file
################################################################################
sub get_layout {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{layout}) {
        return $self->{layout};
    } else {
        $self->_set_err("Error: ZFS::Vg::Vol::get_layout() Vol object is missing layout");
        return "";
    }
}

################################################################################
# Subroutine Name: get_mount
# Objective:       Return mount point of volume
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{mount} and length $self->{mount}) {
        return $self->{mount};
    } else {
        $self->_set_err("Error: ZFS::Vg::Vol::get_mount() Vol filesystem is not mounted");
        return "";
    }
}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vol object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: ZFS::Vg::Vol::get_name() Vol object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{size}) {
        return $self->{size};
    } else {
        $self->_set_err("Error: ZFS::Vg::Vol::get_size() Vol object is missing size");
        return "";
    }
}

################################################################################
# Subroutine Name: get_state
# Objective:       Return object state
# Details:         See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{state}) {
        return $self->{state};
    } else {
        $self->_set_err("Error: ZFS::Vg::Vol::get_state() Vol object is missing state");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vgname
# Objective:       Return Vol object diskgroup name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgname {
    my ($self) = @_;

    my ($name);

    $self->_clear_err();

    $name = $self->{parent}->get_name();

    if (defined $name and length $name) { 
        return $name;
    } else {
        $self->_set_err("Error: ZFS::Vg::Vol::get_vgname() Vol object is missing parent name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       Return path to raw vol *NOOP for ZFS
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    $self->_clear_err();

    ### get name of zpool (disk group) 
    my $vgname = $self->{parent}->get_name();

    ### get name of volume (zfs filesystem)
    my $name = $self->{name};

    $self->_set_err("Error: ZFS::Vg::Vol::get_raw_device() No provision for raw I/O with ZFS");
    return "";

}

################################################################################
# Subroutine Name: is_fs
# Objective:       Determine if volume has a filesystem *NOOP for ZFS
# Details:         See POD documentation at the end of this file
################################################################################
sub is_fs {
    my ($self) = @_;

    $self->_clear_err();

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: is_mounted
# Objective:       Determine if volume is mounted
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;

    my ($cmd, $status, @output, $name, $vgname);
    
    $self->_clear_err();
    
    $vgname = $self->{parent}->get_name();  
    $name = $self->get_name();

    $cmd = "zfs get mounted $vgname/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::Vol::mount() Unable to mount volume (filesystem)",
                        @output);
        return $ClientUtil::FALSE;
    }
    
    # Eg of output:
    # zfs get mounted pool/filesystem
    # NAME                 PROPERTY       VALUE        SOURCE
    # pool/filesystem      mounted        yes          -
    foreach my $line(@output) {
        if ($line =~ /^$vgname\/$name\s+mounted\s+yes/i) {
            return $ClientUtil::TRUE;
        } elsif ($line =~ /^$vgname\/$name\s+mounted\s+no/i) {
            return $ClientUtil::FALSE;
        }
    }
}

################################################################################
# Subroutine Name: mount
# Objective:       mount ZFS filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub mount {
    my ($self,%args) = @_;
    my ($cmd, $vgname, $name, $status, @output, $pass_thru,
        $mount, $fstype);

    # Parse the arguements to check if mount point is passed in
    foreach (keys %args) {
        if (/-?path$/) {
            if (not length($mount = $args{$_})) {
                $self->sanlog(-msg => "ZFS::Vg::Vol::new() value for argument '-path' not specified",
                              -msg_level => 3);
            }
        } elsif (/-?pass_thru$/) {
            $pass_thru = $args{$_};
            $pass_thru =~ s/^\s+//;
            $pass_thru =~ s/\s+$//;
            if ($pass_thru ne "") {
                $pass_thru = "-o $pass_thru";
            }
        }
    }

    # Get attributes
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();
    $fstype = $self->get_fs();
    $pass_thru = "";

    # Use default path if not defined
    if (not defined $mount) {
        $mount = "/$vgname/$name";
        $self->sanlog(-msg=> "ZFS::Vg::Vol::mount() using default mount pt '$mount'", -msg_level => 3);
    } else { 
        # else set the supplied mountpoint
        $cmd = "zfs set mountpoint = $mount";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ZFS::Vg::Vol::mount() Unable to set given mountpoint - $mount",
                            @output);
            return $ClientUtil::FALSE;
        }
    }

    # check if volume is unformatted
    if (defined $fstype and $fstype =~ /raw/) {
        $self->_set_err("Error: ZFS::Vg::Vol::mount()",
                        "Cannot mount a raw filesystem");
        return $ClientUtil::FALSE;
    }
    
    # check if the volume is already mounted
    if ($self->is_mounted()) {
        $self->_set_err("Error: ZFS::Vg::Vol::mount() Filesystem already mounted");
        return $ClientUtil::FALSE;
    } else {
        # else mount the filesystem 
        $cmd = "zfs mount $pass_thru $vgname/$name";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: ZFS::Vg::Vol::mount() Unable to mount volume (filesystem)",
                            @output);
            return $ClientUtil::FALSE;
        }
    }

    $self->{mount} = $mount;

    return $ClientUtil::TRUE;
    
} 

################################################################################
# Subroutine Name: rename()
# Objective:       rename filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self,$newname) = @_;
    my ($cmd, $name, $vgname, @output, $status);

    $self->_clear_err();

    ### get name of disk group
    $vgname = $self->{parent}->get_name();

    ### get name of volume
    $name = $self->get_name();

    ### verify the device is being used in volume group
    unless (defined $name) {
        $self->_set_err("Error: ZFS::Vg::Vol::rename()",
                        "Filesystem doesn't have a name");
        return $ClientUtil::FALSE;
    }
    
    ### verify whether new name is specified or not 
    unless (defined $newname) {
        $self->_set_err("Error: ZFS::Vg::Vol::rename()",
                        "New name for filesystem not specified");
        return $ClientUtil::FALSE;
    }

    ### Rename the filesystem 
    $cmd = "zfs rename /$vgname/$name /$vgname/$newname";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::Vol::rename Unable to rename filesystem '$name' to \'$newname\'",
                        @output);

        return $ClientUtil::FALSE;
    }

    $self->{name} = $newname;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rescan()
# Objective:       rescan volume - NOOP for ZFS
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;
    
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unmount()
# Objective:       unmount volume (filesystem)
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount {
    my ($self,%args) = @_;
    my ($cmd, $type, $vgname, $fs, $mount, $name, $status, 
        $force, $option, @output);

    $option = "";

    foreach (keys %args) {
        if (/-?force$/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::Vol::unmount()",
                                "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (defined $force) {
        $option = "-f";
    }   

    ### Get attributes
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();
    $mount = $self->get_mount();
    $fs = $self->get_fs();

    ### check if volume contains a filesystem

    if ($fs =~ /raw/i) {
        $self->_set_err("Error: ZFS::Vg::Vol::unmount()",
                        "Cannot unmount a raw device");
        return $ClientUtil::FALSE;
    }

    $cmd = "zfs unmount $option $vgname/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err( "Error: ZFS::Vg::Vol::unmount() Unable to unmount volume's filesystem",
                         @output);
        return $ClientUtil::FALSE;
    }

    ### Remove mount pt metadata from object
    $self->{mount} = "";

    return $ClientUtil::TRUE;

}

################################################################################
# P R I V A T E   S U B R O U T I N E S
###############################################################################

################################################################################
# Subroutine Name: _create()
# Objective:       create a filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self) = @_;

    my ($cmd, $device_cnt, $vgname, $name, $layout, $size, @output, $status);

    $vgname = $self->{parent}->get_name();
    $device_cnt = $self->{parent}->get_device_cnt();
    $name = $self->get_name();
    $size = $self->get_size();
    $layout = $self->get_layout();

    $cmd = "zfs create $vgname/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err(
                        "Error: ZFS::Vg::Vol::_create() Unable to create ZFS filesystem - '$name'",
                        @output);
        return $ClientUtil::FALSE;
    }

    # Setting the size limit for the new filesystem created
    $cmd = "zfs set quota=$size $vgname/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err(
                        "Error: ZFS::Vg::Vol::_create() Unable to set size for filesystem - '$name'",
                        @output);
        return $ClientUtil::FALSE;
    }

    # Unmounting the zfs filesystem 
    # The thpl will mount the filesystem later.
    $self->unmount();
    if ($self->error) {
        $self->sanlog(-msg =>"Error: ZFS::Vg::Vol::_create() Unable to unmount the filesystem '$name'");
    }

    return $ClientUtil::TRUE;

}


################################################################################
# Subroutine Name: _init_mount
# Objective:      Initialize mount point for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_mount {
    my ($self) = @_;

    my ($cmd, $vgname, $mount, $name, $path, @output, $status);

    $mount = "";
    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();

    $cmd = "zfs list -t filesystem -o name,mountpoint $vgname/$name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Vg::_init_mount() Unable to display mountpoint for $vgname/$name");
        return "";  
    }    

    # E.g of output: 
    # NAME                  MOUNTPOINT
    # test/file_1           /test/file_1
    foreach my $line(@output) {
        if (grep (/$vgname\/$name/,@output) == 1) {
            if ($line =~ /\S+\s+(\S+\/$name)/) {
                $mount = $1;
            }
        }
    }

    return $mount;
}


################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self) = @_;

    my ($cmd, $vgname, $size, $name, @output, $status);
    my ($size_used, $size_avail);

    $size = "";

    $vgname = $self->{parent}->get_name();
    $name = $self->{name};

    $cmd = "zfs list -t filesystem -o name,avail,used $vgname/$name";    
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::Vol::_init_size() ",
                        "Unable to retrieve size for filesystem \'$name\' in zpool '$vgname",
                        @output);
        return "";
    }
    
    # E.g of output   
    # NAME         AVAIL   USED
    # zpool1/test  10.0G    18K

    if (grep (/$vgname\/$name/,@output) == 1) {
        foreach my $line(@output) {
            chomp ($line);
            if ($line =~ /${vgname}\/${name}\s+(\S+)\s+(\S+)/) {
                $size_avail = $1;
                $size_used = $2;
                last;
            }
        }
    } else {
        $self->_set_err("Error: ZFS::Vg::Vol::_init_size() ",
                "Unable to retrieve size for filesystem \'$name\' in zpool '$vgname",
                @output);
        return "";
    }
    
    # Converting size_used to MB.
    $size_used = &ClientUtil::convert_to_bytes($size_used, "m");


    # Converting size_avail to MB.
    $size_avail = &ClientUtil::convert_to_bytes($size_avail, "m");

    # Add $size_avail and $size_used together

    $size = int($size_used + $size_avail);

    return $size;
}   

################################################################################
# Subroutine Name: _init_state
# Objective:      Initialize state of existing host volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_state {
    my ($self) = @_;

    my ($cmd, $name, $state, $vgname, $status, @output);

    $state = "";

    $vgname = $self->{parent}->get_name();
    $name = $self->get_name();
    
    $cmd = "zfs list $vgname/$name";    
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: ZFS::Vg::Vol::_init_state() ",
                        "Unable to retrieve state for filesystem '$name' in zpool '$vgname'",
                        @output);
        return "";
    }

    if (grep (/dataset does not exist/i, @output) == 1) {
        $state = "offline";
    } else {
        $state = "online";
    }

    return $state;

}

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($cmd, $exist, $import, $name, $size, $conn, $layout, $log);
    my ($parent, $self, $status, @output);

    $self = {};

    bless $self, $class;

    # ZFS does not have a default layout 
    $layout = "n/a";
    $exist = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?layout$/) {
            if (not length($layout = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::Vol::new()",
                                "Invalid value for argument '-layout'");
                return $self;
            }
            next;
        } elsif (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::Vol::new()",
                                "Invalid value for argument '-name'");
                return $self;
            }   
            next;
        } elsif (/-?size$/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::Vol::new()",
                                "Invalid value for argument '-size'");
                return $self;
            }
            next;
        } elsif (/-?exist$/) {
            if (not length($exist = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::Vol::new()",
                                "Invalid value for argument '-exist'");
                return $self;
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::Vol::new()",
                                "Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: ZFS::Vg::Vol::new()",
                                "Missing value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    ### check for existance of required -parent argument
    unless (defined $parent) {
        $self->_set_err("Error: ZFS::Vg::Vol::new()",
                        "Missing required argument '-parent'");
        return $self;
    }

    ### check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: ZFS::Vg::Vol::new()",
                        "Missing required argument '-name'");
        return $self;
    }

    ### check for existance of required -size argument
    unless (defined $size) {
        $self->_set_err("Error: ZFS::Vg::Vol::new()",
                        "Missing required argument '-size'");
        return $self;
    }

    ### check for existance of required -conn argument
    if (grep ( /^-?conn$/,(keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: ZFS::Vg::Vol::new()",
                            "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: ZFS::Vg::Vol::new()",
                        "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn} = $conn;
    $self->{name} = $name;
    $self->{layout} = $layout;
    $self->{size} = $size;
    $self->{parent} = $parent;
    $self->{state} = 'ENABLED';
    $self->{filesystem} = 'zfs';

    if (not $exist) {
        $self->_create();
        return $self if ($self->error());
    } else {
        $self->sanlog(-msg => "ZFS::Vg::Vol::new() loading existing volume '$name'", -msg_level => 3);
        $self->{state} = $self->_init_state();
        return $self if ($self->error);
        $self->sanlog(-msg => "ZFS::Vg::Vol::new()  state: " . $self->{state}, -msg_level => 3);

        $self->{filesystem} = "zfs";
        if ($self->error()) {
            ### allow vol to be loaded although it
            ### can't be started to determine fs
            if ($self->{filesystem} eq "") {
                $self->_clear_err();
            } else {
                return $self;
            }
        }
        $self->sanlog(-msg => "ZFS::Vg::Vol::new() fs: " . $self->{filesystem}, -msg_level => 3);

        if ($self->{filesystem} !~ /raw/i) {
            $self->{mount} = $self->_init_mount();
            return $self if ($self->error);
            $self->sanlog(-msg => "ZFS::Vg::Vol::new() mnt: " . $self->{mount}, -msg_level => 3);
        }

        $self->{size} = $self->_init_size();
        return $self if ($self->error);
        $self->sanlog(-msg => "ZFS::Vg::Vol::new() size: " . $self->{size}." mb", -msg_level => 3);

        $self->{layout} = "n/a";
        return $self if ($self->error);
        $self->sanlog(-msg =>"ZFS::Vg::Vol::new() layout: " . $self->{layout}, -msg_level => 3);
    }

    return $self;

}

1;
