## @summary Provide an interface the Vxvm Logical Volume Manager on Windows
## @author rawat@netapp.com
## @status private
## @pod ../../Lvm.pm

# Package Name
##############
package San::SanClient::Lvm::Windows::Vxvm;

use strict;

# Package Requirements
######################
use File::Basename;

use San::SanClient::ClientUtil;
use San::SanClient::Lvm;
require San::SanClient::Lvm::Windows::Vxvm::Vg;

use Error;
use San::SanLog;

use vars qw(@ISA  %hFormat);

@ISA = qw(Error SanLog Lvm);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: initialize
# Objective:       initialize disk under veritas control
# Details:         See POD documentation at the end of this file
################################################################################
sub initialize {
    my ($self, %args) = @_;
    my ($attribute, $cmd, $disk_style, $controller, @already_inited, @output, $status, $devices, @vx_devices);

    ### default disk type
    $disk_style = "MBR";
    $attribute = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::initialize() Invalid value for argument '-devices'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?format/) {
            if (not length($disk_style = $args{$_}) and $disk_style !~ /MBR|GPT/i) {
                $self->_set_err("Error: Windows::Vxvm::initialize() Invalid value for argument '-format'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### check for required parameters
    unless (defined $devices and length $devices) {
        $self->_set_err("Error: Windows::Vxvm::initialize() -devices is a required argument");
        return $ClientUtil::FALSE;
    }


    # We need to rescan the Host so that we dont left with the old stale entries from the previous run

    $cmd = "vxassist rescan";
    ($status, @vx_devices) = &ClientUtil::execute($self->{conn}, $cmd, 3600);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::initialize() no valid devices found");
    }

    # We need to get the controller name of the device to which it belongs. This cmd has the
    # following output. In this output 8th field (seperated by spaces) is the controller name

    ## Disk    LUN     DSM     Policy  Paths   Serial          CFMode  Filer           Path
    ## ----    ---     -----   ------  ------  -----           ------  -----           ----
    ## Disk2   0       unkown  FO      2       HnWLg4Hasol7    None    f3070-200-73    /vol/nate_1_1/nate_w2k3200-104_windows_0

    $cmd = "%SystemDrive%\\TOOLS\\mpc.exe show luns";
    my ($status, @luns) = &ClientUtil::execute($self->{conn}, $cmd, 3600);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::initialize() unable to display the luns");
    }  

    $cmd = "vxdisk list";
    ($status, @vx_devices) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::initialize() no valid devices found");
    }

    my $tmp_ctrl = "";
    my $ctrl_assinged = 0;

    foreach my $device (@$devices) {

        # initialized devices that have not been inited or are inited with the wrong format
        my ($drive_num) = ($device =~ /\\\\.\\PHYSICALDRIVE(\d+)/);
        my $tmp_dev = "Harddisk".$drive_num ;
        my ($line) = grep (/^$tmp_dev\s+/, @vx_devices);
        my ($disk_id) = ($line =~ /\w+.*\s+(P\d+C\d+T\d+L\d+)/);

        
        my ($line) = grep (/^Disk$drive_num/i, @luns);
        ($controller) = ($line =~ /\S+\s+\d+\s+\w+\s+\w+\s+\d+\s+\S+\s+\w+\s+(\S+)\s+.*/);

        $cmd = "vxdisk diskinfo $disk_id";
        my ($status, @disk_info) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Windows::Vxvm::initialize() unable to get the info of dev $tmp_dev");
        }  

        my ($line) = grep (/Signature/i,@disk_info);
        my ($signature) = ($line =~ /\w+\s+:\s+(\d+)/);
        my ($line) = grep (/Length/i,@disk_info);
        my ($size) = ($line =~ /\w+\s+:\s+(\d+)/);
        my ($line) = grep (/status/i,@disk_info);
        my ($status) = ($line =~ /\w+\s+:\s+(\w+)/);
        if ($status =~ /offline/i) {
            $cmd = "vxdisk online $disk_id";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)",
                                "Veritas Device setup (making online) failed.",
                                "OUTPUT = @output");
                return $ClientUtil::FALSE;
            }
        }

        if ($signature == 0) {
            $cmd = "vxdisk sig $disk_id diskstype = $disk_style";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err("Command ($cmd) returned a bad status ($status)",
                                "Veritas Device setup (initialize) failed.",
                                "OUTPUT = @output");
                return $ClientUtil::FALSE;
            }

        } else {
            $self->sanlog(-msg => "Signature has already been put on the device $tmp_dev");
        }

        if ( $tmp_ctrl ne "" and $tmp_ctrl ne $controller and $ctrl_assinged == 0) {
            $self->{controller_1} = $tmp_ctrl;
            $self->{controller_2} = $controller;
            $ctrl_assinged = 1;
        }

        $self->{$controller}{unused_disks}++;

        $tmp_ctrl = $controller;
        ## This hash we are going to use in Vol.pm while creating the Volumes
        $self->{$tmp_dev}{controller} = $controller;

        ## converting the size to MB
        $self->{$tmp_dev}{size} = int $size/(1024*1024);
        $self->{$tmp_dev}{in_use} = 0;
        $self->{$tmp_dev}{vol} = "";
        #$self->{free_disks}++;

        ## Update the Format hash
        $self->sanlog(-msg => "Device '$device' is $signature");
        $self->{$device}->{signature} = $signature;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: create_vg
# Objective:       create and return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vg {
    my ($self, %args) = @_;
    my ($vgObj, $bAdd);

    $bAdd = 1;

    if (exists $args{-no_add}) {
        $self->sanlog(-msg => "VxVM object will not manage VG object");
        delete $args{-no_add};
        $bAdd = 0;
    }

    $vgObj = San::SanClient::Lvm::Windows::Vxvm::Vg->_new(-parent => $self,
                                                        -conn   => $self->{conn},
                                                        %args);   

    if ($vgObj->error) {
        $self->_set_err($vgObj->errmsg());
        return {};
    }

    if ($bAdd == 1) {
        push @{$self->{vgs}}, $vgObj;
    } else {
        $vgObj = {};
    }

    return $vgObj;
}

################################################################################
# Subroutine Name: destroy_vg
# Objective:       destroy a named diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vg {
    my ($self, $name) = @_;
    my ($vgObj, @vgs);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Windows::Vxvm::destroy_vg() vg name is a required argument");
        return $ClientUtil::FALSE;
    }

    ### Delete the of vg object and update the parent vg list.
    foreach $vgObj (@{$self->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            $vgObj->destroy();
            if ($vgObj->error) {
                $self->_set_err($vgObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vgs, $vgObj;
        }
    }

    $self->{vgs} = \@vgs;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display
# Objective:       display disk group information
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self) = @_;
    my ($cmd, $status, @output);

    $self->_clear_err();

    $self->sanlog(-msg => "Display diskgroup information");

    $cmd = "/sbin/vxprint";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Command ($cmd) returned a bad status ($status)");
        return  (wantarray()) ? () : [];
    }

    return (wantarray()) ? @output : \@output;

}

###############################################################################
# Method:    get_disk_format
# Objective: No-op for windows
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_format {
    my ($self, $disk) = @_;
    my ($format);

    $self->_clear_err();

    unless (defined $disk and length $disk) {
        $self->_set_err("Error: Windows::Vxvm::get_disk_format() disk name is a required argument");
        return "";
    }

    $format = "MBR";

    unless (defined $format and length $format) {
        $self->_set_err("Error: Windows::Vxvm::get_disk_format() disk has not been initialized");
        return "";
    }

    return $format;
}

###############################################################################
# Method:    get_disk_formats
# Objective: Get list of supported disk formats
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_disk_formats {
    my ($self) = @_;
    my (@format);

    $self->_clear_err();

    push @format, qw(MBR GPT);

    return (wantarray()) ? @format : \@format;
}

###############################################################################
# Method:    get_format_default
# Objective: Get default type to use for disk format
# Details:   See POD documentation at the end of this file.
###############################################################################
sub get_format_default {
    my ($self) = @_;
    my ($format);

    $self->_clear_err();

    $format = "MBR";
    return $format;
}

################################################################################
# Subroutine Name: get_lv_supported_types 
# Objective:       return supported LV types
# Details:         See POD documentation at the end of this file
################################################################################
sub get_lv_supported_types {
    my ($self) = @_;
    my (@supported_lvs);

    push (@supported_lvs, ('striped-mirror','concat','mirror','stripe','raid5'));
    return (wantarray()) ? @supported_lvs : \@supported_lvs;

}

################################################################################
# Subroutine Name: get_vg
# Objective:       return a diskgroup object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vg {
    my ($self, $name) = @_;
    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Windows::Vxvm::get_vg() vg name is a required argument");
        return {};
    }

    foreach my $obj (@{$self->{vgs}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: Windows::Vxvm::get_vg() Unable to find diskgroup '$name'");

    return {};
}

################################################################################
# Subroutine Name: get_vgs
# Objective:       return array of volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgs {
    my ($self) = @_;
    my (@vgs);

    $self->_clear_err();

    if (defined $self->{vgs}) {
        @vgs = @{$self->{vgs}};
    } else {
        $self->_set_err("Error: Windows::Vxvm::get_vgs() object has no volumes groups");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @vgs : \@vgs;

}

################################################################################
# Subroutine Name: get_vgnames
# Objective:       return array of dynamic disk group names
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgnames {
    my ($self) = @_;
    my ($cmd, $status, @output, @vgnames);

    $self->_clear_err();

    $cmd = "vxdg list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::get_vgnames Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }
    
    my @dg_names = grep (/Name\s+:/i,@output);
    foreach my $line (@dg_names) {
        my ($dgname) = ($line =~ /Name\s+:\s+(\w+)/);
        next if ($dgname eq "BasicGroup");
        push (@vgnames, $dgname);
    }
        
    return (wantarray()) ? @vgnames : \@vgnames;

}

################################################################################
# Subroutine Name: get_free_devices
# Objective:       Retrieve list of free devices for use.
# Details:         See POD documentation at the end of this file
################################################################################
sub get_free_devices {
    my ($self, %args) = @_;
    my ($bCnt,    $cmd,     $cnt,         $device, $filer, $filername, $lun,
        $refresh, $status,  $total,  $client_obj,      $type,   $vol,   %hDevs,     @all_devices,
        @attr,    @devices, @lvm_devices, @output, @used);

    $type  = ".*";
    $total = 0;

    $self->_clear_err();

    # Asssume user supplied -cnt argument
    $bCnt    = 1;
    $refresh = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?cnt/) {
            if (not length($cnt = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::get_free_devices() Invalid value for argument '-cnt'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::get_free_devices() Invalid value for argument '-type'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?filer/) {
            if (not length($filer = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::get_free_devices() Invalid value for argument '-filer'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?vol/) {
            if (not length($vol = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::get_free_devices() Invalid value for argument '-vol'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?lun/) {
            if (not length($lun = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::get_free_devices() Invalid value for argument '-lun'");
                return (wantarray()) ? () : [];
            }
            next;
        } elsif (/-?refresh/) {
            if (not length($refresh = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::get_free_devices() Invalid value for argument '-refresh'");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Check if user wants free devices from a specified filer,volume, or lun
    if (defined $filer or defined $vol or defined $lun) {
        $self->sanlog(-msg => "creating translation maps:");
        $refresh = 1;

        # Check if a device to filer map 'dfmap' has already been created or
        # refresh it if requested.
    if ( not defined $self->{dfmap} or $refresh == 1) {
        $self->sanlog(-msg => "Creating translation maps; " .
                          "device to filer, vol, and lun");
        my @output=$self->{sanlun}->lun_show();
        if ($self->{sanlun}->error) {
            $self->_set_err($self->{sanlun}->errmsg());
            return (wantarray()) ? () : [];
        }

        #
        # We are expecting @output to look like the following for physical windows machines:
        #
        #   filer:         lun-pathname         device filename          adapter        protocol        lun size        lun state
        # fas270-rtp08: /vol/nate_1_1/nate_dell2650-rtp8_windows_0   \\\\.\\PHYSICALDRIVE1         TBD       iscsi         2104M (2206203904)     OK
        # fas270-rtp07: /vol/nate_1_1/nate_dell2650-rtp8_windows_0   \\\\.\\PHYSICALDRIVE2         TBD       iscsi         2104M (2206203904)     OK
        #
        # but @output can also look like this for virtual windows machines:
        #
        #   filer:         lun-pathname         device filename          adapter        protocol        lun size        lun state
        # fas270-rtp07: /vol/nate_3_1_ibm3550_rtp01_guest02/nate_ibm3550-rtp01-guest02_windows_0   \\\\.\\PHYSICALDRIVE4         TBD       none         4200M (4404019200)     OK
        # :    \\\\.\\PHYSICALDRIVE1         TBD       none         3092M (3242196992)     OK
        # fas270-rtp08: /vol/nate_3_1_ibm3550_rtp01_guest02/nate_ibm3550-rtp01-guest02_windows_0   \\\\.\\PHYSICALDRIVE3         TBD       none         4200M (4404019200)     OK
        # :    \\\\.\\PHYSICALDRIVE2         TBD       none         3092M (3242196992)     OK
        #
        
        #
        # Look at each line from sanlun and create a hash for each lun
        # Take this hash and add it to a dfmap.
        #

        $self->sanlog(-msg => "sanlun output is:");
        foreach my $line (@output) {
            
            # skip first line (header)
            $self->sanlog(-msg=>"$line");
            next if ($line =~ /lun-pathname/i);

            # extract the info from the lun show line
            my ($filer_name, $lunpath, $device_filename) = 
                ($line =~ /(.*?):(.*?)([\\\.]*PHYSICALDRIVE\d+)/);

            # eliminate whitespace in the lun path
            $lunpath =~ s/\s//g;
            
            # extract the vol name and lun name.
            my ($vol_name, $lun_name) = ($lunpath =~ /^\/vol\/(.*?)\/(.*)/);
        
            # pull out the drive number
            my ($drive_number) = ($device_filename =~ /PHYSICALDRIVE(\d+)/);
                        
            $hDevs{$device_filename} = { filer        => $filer_name,
                                         vol          => $vol_name,
                                         lun          => $lun_name,
                                         drive_number => $drive_number
                                     };
            
            $self->sanlog(-msg_level => 3,
                          -msg       => "\nDevice Info for ($device_filename) dfmap:\n" .
                                        "  filer = $filer_name\n"                       .
                                        "  vol   = $vol_name\n"                         .
                                        "  lun   = $lun_name\n");
        }
        
        # save map for next call of function
        $self->{dfmap}=\%hDevs;
    } 
    }

    # If not -cnt then assume all free disks
    # under lvm control are wanted
    unless (defined $cnt and length($cnt)) {
        $cnt  = 10000;
        $bCnt = 0;
    }

    $cmd = "vxdisk list";
    my ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::get_free_devices() Could not execute $cmd cmd");
        return (wantarray()) ? () : [] ;
    }

    $client_obj = $self->{parent};
    foreach my $line (@output) {
        if(grep (/netapp/i,$line)) {
            my ($drive_num) = ($line =~ /^Harddisk(\d+)\s+.*/);
            my $device = "\\\\.\\PHYSICALDRIVE" . $drive_num;
            if (($client_obj->{WinClient}->{disks}->{$drive_num}->{boot} == 0) and
                ($client_obj->{WinClient}->{disks}->{$drive_num}->{isQuorum} == 0)) {
                push @lvm_devices, $device;
            }
        }
    }

    # iterate through the devices and make sure that none of the disks
    # are being used other LVM disk object.
    foreach my $device (@lvm_devices) {
        if (defined $filer and length $filer) {

            $self->sanlog(-msg => "MAP: ".$self->{dfmap}{$device}{filer}." FILER: ".$filer, 
                          -msg_level => 7);
            next if ($self->{dfmap}{$device}{filer} !~ /^$filer$/i);
        }

        if (defined $vol and length $vol) {

            $self->sanlog(-msg=>"MAP: ".$self->{dfmap}{$device}{vol}." VOL: ".$vol,
                          -msg_level => 7);
            next if ($self->{dfmap}{$device}{vol} !~ /^$vol$/i);
        }

        if (defined $lun and length $lun) {

            $self->sanlog(-msg=>"MAP: ".$self->{dfmap}{$device}{lun}." LUN: ".$lun,
                          -msg_level => 7);
            next if ($self->{dfmap}{$device}{lun} !~ /$lun/i);
        }
        push @devices, $device if (++$total <= $cnt);
    }

    # If user supplied cnt and free devices fell short
    # generate error
    if ($bCnt == 1 and scalar(@devices) < $cnt) {
        $self->_set_err("Error: Windows::Vxvm::get_free_devices() ",
                        "Only " . scalar(@devices) . " of $cnt requested devices were found");
        return (wantarray()) ? () : [] ;
    }

    return (wantarray()) ? @devices : \@devices;
}

################################################################################
# Subroutine Name: load_vg
# Objective:       load an existing diskgroup that is not in memory
# Details:         See POD documentation at the end of this file
################################################################################
sub load_vg {
    my ($self, $name) = @_;
    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Windows::Vxvm::load_vg() Invalid or missing required paramter \$name");
        return {};
    }

    ### Check if vg has already been loaded
    $vgObj = $self->get_vg($name);
    if ($self->error) {

        # dg has not been load yet
        $self->_clear_err();

        # load dg
        $self->_create_vgs($name);
        return {} if ($self->error);
        $vgObj = $self->get_vg($name);
    } else {
        $self->sanlog(-msg => "INFO: Windows::Vxvm::load_vg() Volume group '$name' was already loaded"
                      -msg_level => 7);
    }

    return $vgObj;

}

################################################################################
# Subroutine Name: unload_vg
# Objective:       Remove the vg object under LVM object control
# Details:         See POD documentation at the end of this file
################################################################################
sub unload_vg {
    my ($self, $name) = @_;
    my ($vgObj, $bFound, @vgs);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Windows::Vxvm::unload_vg() Invalid or missing required paramter \$name");
        return $ClientUtil::FALSE;
    }

    $bFound = 0;

    foreach $vgObj (@{$self->{vgs}}) {
        ($vgObj->get_name() =~ /^$name$/)
          ? $bFound = 1
          : push @vgs, $vgObj;
    }

    $self->{vgs} = \@vgs;

    unless ($bFound) {
        $self->_set_err("Error: Windows::Vxvm::unload_vg() Nothing to unload. Vg name '" . $name . "' was not loaded");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: enable_controllers
# Objective:       enable controllers
# Details:         See POD documentation at the end of this file
################################################################################
sub enable_controllers {
    my ($self) = @_;

    my ($cmd, @output, $status, $ctlr, @controllers);

    $self->_clear_err();

    if ($self->_is_installed()) {

        $self->rescan();
        return $ClientUtil::FALSE if ($self->error());

        # Get a list of all the controllers
        $cmd = "/etc/vx/bin/vxcntrllist";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);

        if ($status > 0) {
            $self->_set_err("Error: Windows::Vxvm::enable_controllers() Command ($cmd) returned a bad status ($status)");
            return $ClientUtil::FALSE;
        }

        # skip controller 0 (root disk)
        @controllers = grep (!/c0/, @output);

        # traverse through the controllers and enable each.
        foreach $ctlr (@controllers) {
            $cmd = "/usr/sbin/vxdmpadm enable ctlr=$ctlr";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 0) {
                $self->_set_err(
                               "Error: Windows::Vxvm::enable_controllers() Command ($cmd) returned a bad status ($status)");
                return $ClientUtil::FALSE;
            }
        }
    }
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rescan
# Objective:       To run the vxdctl command to rescan veritas's list of
#                  disks that it thinks are under it's control.
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;
    my ($cmd, @output, $status);

    $self->_clear_err();

    # need to run the vxdctl script after the new VTOC is down
    # to get Veritas to see the disk again (properly) .. not completely
    # sure why
    $cmd = "vxassist rescan";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::rescan() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: restart_recovery_daemon
# Objective:       Method that will stop the recovery deamon and reset
# Details:         See POD documentation at the end of this file
################################################################################
sub restart_recovery_daemon {
    my ($self, $interval) = @_;

    my ($cmd, @output, $status, $policy, $policy_string);
    
    $self->_clear_err();

    if (not defined ($interval)) {        
        $self->sanlog(-msg => "Restore interval was not defined, thus it will not be reset (could be cleanup).");        
        return $ClientUtil::TRUE;
    }

    $policy = $self->{parent}->{vx_restore_policy};
    if ($self->_is_installed()) {

        # stop the restore
        $cmd = "/usr/sbin/vxdmpadm stop restore";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Windows::Vxvm::restart_recovery_daemon()",
                            "Command ($cmd) returned a bad status ($status)");

            return $ClientUtil::FALSE;
        }

        if (defined ($policy)) {
            $policy_string = "policy=$policy";
        }

        # start the restore
        $cmd = "/usr/sbin/vxdmpadm start restore interval=$interval $policy_string";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Windows::Vxvm::restart_recovery_daemon()",
                            "Command ($cmd) returned a bad status ($status)");

            return $ClientUtil::FALSE;
        }

        # set stats on the setting, just to get it in the logs.
        $cmd = "/usr/sbin/vxdmpadm stat restored";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Windows::Vxvm::restart_recovery_daemon()",
                            "Command ($cmd) returned a bad status ($status)");

            return $ClientUtil::FALSE;
        }

    }
    return $ClientUtil::TRUE;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new lvm object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($conn, $log, $lvm, $loadcfg, $parent, $self);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?lvm/) {
            if (not length($lvm = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::_new() Invalid value for argument '-lvm'");
                return  $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::_new() Invalid value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?loadcfg/) {
            if (not length($loadcfg = $args{$_}) or $loadcfg !~ /no|yes|1|0|off|on/i) {
                $loadcfg = 1;
            }
            next;
        }
    }

    ### check for existance of required -lvm argument
    if (grep (/^-lvm$/, (keys %args))) {
        unless ($args{-lvm} =~ /^vxvm|dmp$/i) {
            $self->_set_err("Error: Windows::Vxvm::_new()", "Invalid value for argument '-lvm' \(" . $args{-lvm} . "\)");
            return $self;
        }
    } else {
        $self->_set_err("Error: Windows::Vxvm::_new()", "Missing required argument'-lvm'");
        return $self;
    }

    ### check for existance of required -parent argument
    if (grep (/^-?parent$/, (keys %args))) {
        unless (ref($args{-parent}) =~ /SanClient/i) {
            $self->_set_err("Error: Windows::Vxvm::_new()", "Invalid value for argument '-parent'");
            return $self;
        }
    } else {
        $self->_set_err("Error: Windows::Vxvm::_new()", "Missing required argument '-parent'");
        return $self;
    }

    # initilize the object members
    $self->{conn}   = $parent->get_connection();
    $self->{sanlun} = $parent->get_sanlun();
    $self->{parent} = $parent;
    $self->{vgs} = [];

    ### check to make sure that LVM is installed
    ### no need to check anything for Windows
    if (not $self->_is_installed()) {
        return $self if ($self->error);
    }

    ### create veritas vg objects for existing disk groups
    if ($loadcfg) {
        $self->_create_vgs();
        return $self if ($self->error);

        ### update map to determine disk format
        $self->_update_disk_format_map();
        return $self if ($self->error);
    }
    
    # set VRTS enclosure tuning and restart recovery daemon with passed settings.
#    $self->_set_vxvm_encl_tunings();
#    return $self if ($self->error);
#    $self->restart_recovery_daemon($self->{parent}->{vx_restore_interval});
#    return $self if ($self->error);

    $self->_set_load_balancing_policy();
    return $self if ($self->error);

    return $self;

}

################################################################################
# Subroutine Name: _update_disk_format_map
# Objective:       Update disk format map
# Details:         See POD documentation at the end of this file
################################################################################
sub _update_disk_format_map {
    my ($self) = @_;
    my ($cmd, @output, $status, @netapp_device);

    $cmd = "vxdisk list";

    ### Retrieve the volume groups.
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 1) {
        $self->_set_err("Error: Windows::Vxvm::_update_disk_format_map()", "Unable to update map");
        return $ClientUtil::FALSE;
    }
    @netapp_device = grep (/netapp/i,@output);

    foreach my $line (@netapp_device) {
        my ($device, $format) = ($line =~ /^(\S+)\s+.*\s+(MBR|GPT)\s+/i);
        $hFormat{$device} = $format;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _create_vgs
# Objective:       create a vg object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vgs {
    my ($self, $vgname) = @_;
    my ($cmd, @output, $status, $vgObj, %hDeport);

    $cmd = "vxdg list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 1) {
        $self->_set_err("Error: Windows::Vxvm::_create_vgs()", "Unable to load existing diskgroup\(s\)");
        return $ClientUtil::FALSE;
    }
    
    if (defined($vgname) and length($vgname)) {
        my ($line) = grep (/$vgname/, @output);
        if ($line eq "") {
            self->_set_err("Error: Windows::Vxvm::_create_vgs() No Dynamic Disk group of name".
                            $vgname ."exists on the Client");
            return $ClientUtil::FALSE;
        } else {
            $self->create_vg(-name => $vgname, -deported => 0);
            return $ClientUtil::FALSE if ($self->error);
        }
    } else {
        
        # I am calling Public subroutine from inside a private subroutine. Is it fine?
        my @alldgs = $self->get_vgnames();
        foreach my $line (@alldgs) {
            $cmd = "vxdg -g $line dginfo"; 
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status > 1) {
                $self->_set_err("Error: Windows::Vxvm::_create_vgs()", "Unable to retrieve DG info");
                return $ClientUtil::FALSE;
            }
            my $deported = 0;
            if (grep (/Deported/i,@output)) {
                $deported = 1;
            }
            $self->create_vg(-name => $line, -deported => $deported);
            return $ClientUtil::FALSE if ($self->error);
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_disk_id
# Objective:       Return disk id
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_disk_id {
    my ($self, $device) = @_;

    my ($cmd, @output, $status);

    $self->_clear_err();

    $cmd = "vxdisk -s list  $device | grep diskid | awk '{print \$2}'";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::get_diskid() Unable to get diskid for device \'$device\'", @output);

        return $ClientUtil::FALSE;
    }

    return $output[0];

}

################################################################################
# Subroutine Name: _is_installed
# Objective:       No-op
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_installed {
    my ($self) = @_;
    my ($status, @output, $cmd);

    # Will use some vx cmd to 
    return $ClientUtil::TRUE;

}

################################################################################
# Method: _set_load_balancing_policy
# Objective: Set Veritas  enclosure ebased tunings.
# Details:   See POD documentation at the end of this file.
################################################################################
sub _set_load_balancing_policy {
    my ($self) = @_;
    my ($cmd, $status, $protocol, $policy, @output);
    
    $self->_clear_err();
}

################################################################################
# Method: _set_vxvm_encl_tunings
# Objective: Set Veritas  enclosure ebased tunings.
# Details:   See POD documentation at the end of this file.
################################################################################
sub _set_vxvm_encl_tunings {
    my ($self) = @_;

    my ($line, $cmd, $status, @output, @enclosure_names);
    my ($rec_option, $queue_option);

    my ($vx_50_fast_recovery, $vx_50_recovery_option, $vx_50_queue_depth, $vx_io_policy) = (
                                       $self->{parent}->{vx_50_fast_recovery}, $self->{parent}->{vx_50_recovery_option},
                                       $self->{parent}->{vx_50_queue_depth},   $self->{parent}->{vx_io_policy});

    if ($vx_50_fast_recovery eq "" || $vx_50_recovery_option eq "" ||
        $vx_50_queue_depth eq "" || $vx_io_policy eq "") {
        $self->sanlog(-msg => "One or more options parameters have not been passed in. This is either because a cleanup is happening, or the parameters were not properly defined.");
        
        # Return true here because this may not have happened in error. 
        # This can happen in cases where cleanup is being done, or 
        # optional parameters simply were not provided.
        return $ClientUtil::TRUE;
    }

    @enclosure_names = $self->_get_vxvm_enclosure_name();
    if (scalar(@enclosure_names) == 0) {
        $self->_set_err("No Valid enclosure name was found");
        return $ClientUtil::FALSE;
    }

    my $count = 0;
    if (defined $vx_50_recovery_option and $vx_50_recovery_option ne "NOT_SET") {
        $rec_option = "recoveryoption=$vx_50_recovery_option";
        $count++;
    }

    if (defined $vx_50_queue_depth and $vx_50_queue_depth ne "NOT_SET") {
        $queue_option = "queuedepth=$vx_50_queue_depth";
        $count++;
    }

    if (defined $queue_option and defined $rec_option) {
        foreach my $enclosure_name (@enclosure_names) {
            $cmd = "/usr/sbin/vxdmpadm setattr enclosure $enclosure_name $rec_option $queue_option";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Error: Windows::Vxvm::_set_vxvm_encl_tunings() Failed to set recovery option or queuedepth");
                return $ClientUtil::FALSE;
            }
        }
    } elsif ($count == 1) {
        $self->_set_err("Error: Windows::Vxvm::_set_vxvm_encl_tunings() Veritas requires both queuedepth and recoveryoption to be tuned together. Please specify values for both VX_50_RECOVERY_OPTION and VX_50_QUEUE_DEPTH");
        return $ClientUtil::FALSE;
    }

    $cmd = "/usr/sbin/vxdmpadm settune dmp_fast_recovery=$vx_50_fast_recovery";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        foreach $line (@output) {
            $self->sanlog(-msg => "DEBUG: the ouput trapped was: @output");
            if (($line =~ /vxdmpadm ERROR/) && ($line !~ /same as its existing/)) {
                $self->_set_err("Failed to set DMP fast recovery.");
                return $ClientUtil::FALSE;
            }
        }
    }

    if (defined $vx_io_policy and $vx_io_policy ne "NOT_SET") {
        foreach my $enclosure_name (@enclosure_names) {
            $cmd = "/usr/sbin/vxdmpadm setattr enclosure $enclosure_name iopolicy=$vx_io_policy";
            ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
            if ($status != 0) {
                $self->_set_err("Failed to set the IO Policy for Enclosure: $enclosure_name .");
                return $ClientUtil::FALSE;
            }
        }
    }
    return $ClientUtil::TRUE;

}

#################################################################################
# Method: _get_vxvm_enclosure_name
# Objective: Get the enclosure name of the attached Veritas Enclosure
# Details:   See POD documentation at the end of this file.
#################################################################################
sub _get_vxvm_enclosure_name () {
    my ($self) = @_;

    my ($cmd, $line, $status, @output, @enclosure_names);

    @enclosure_names = ();
    $cmd = "/usr/sbin/vxdmpadm listenclosure all";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Windows::VxVM::_get_vxvm_enclosure_name: Failed to execute command ($cmd)");
        return ();
    }
    foreach $line (@output) {
        chomp($line);
        $line =~ s/\r//g;
        if (($line !~ /DISCONNECTED/) && ($line =~ /CONNECTED/) && ($line !~ /Disk/)) {
            my ($enclosure_name, undef) = split('\s+', $line);
            $self->sanlog(-msg => "The Enclosure Name found is $enclosure_name.");
            push(@enclosure_names, $enclosure_name);

        }
    }
    if (scalar(@enclosure_names) == 0) {
        $self->_set_err("Windows::VxVM::_get_vxvm_enclosure_name: Failed to collect enclosure names");
        return ();
    } else {
        $self->sanlog(-msg => "The list of Enclosure Names is @enclosure_names");
        return (wantarray()) ? @enclosure_names : \@enclosure_names;
    }
}

1;

