## @summary Provide an interface the Vxvm Volume Groups on Windows
## @author rawat@netapp.com
## @status private
## @pod ../../Vg.pod

package San::SanClient::Lvm::Windows::Vxvm::Vg;

use strict;

use San::SanLog;
use San::SanClient::ClientUtil;
require San::SanClient::Lvm::Windows::Vxvm::Vg::Vol;

use Error;
use Tie::IxHash;
use vars qw(@ISA);

@ISA = qw(Error SanLog);

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume group object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($cmd, $conn, $devices, $disk_cnt, $parent, $log, $name, $self, $status, $deported, @output, %hDevice);

    $self = {};
    bless $self, $class;

    $self->_clear_err();

    $deported = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?devices/) {
            if (not length($devices = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Invalid value for argument '-devices'");
                return $self;
            }
            next;
        } elsif (/-?name/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Missing value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Missing value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Missing value for argument '-parent'");
                return $self;
            }
            next;
        } elsif (/-?deported/) {
            if (not length($deported = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Missing value for argument '-deported'");
                return $self;
            }
            next;
        }
    }

    ### check for existance of required -parent argument
    unless (defined $parent) {
        $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Missing required argument '-parent'");
        return $self;
    }

    ### check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Missing required argument '-name'");
        return $self;
    }

    ### check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn}     = $conn;
    $self->{name}     = $name;
    $self->{deported} = $deported;
    $self->{disk_cnt} = 0;
    $self->{parent}   = $parent;

    # create the diskgroup if devices were supplied
    if (defined $devices and ref($devices) eq "ARRAY") {
        $self->_create($name, $devices);
        return $self if ($self->error());
    } else {

        # vg must already exist. set device list
        # for all existing and deported dgs.
        $self->sanlog(-msg => "loading existing volume group '$name'");
        $self->{devices} = $self->_get_devices();
        return $self if ($self->error);

        unless ($self->{deported}) {
            $self->_create_vols();
            return $self if ($self->error);
        }

    }

    return $self;

}

################################################################################
# Subroutine Name: _create_vols
# Objective:       create vol object for each volume group on the host
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_vols {
    my ($self, %args) = @_;
    my ($cmd, $import,@vols, $name, $fsck, $status, $vgObj, $volObj, %hDeport, @output);

    ### Identify volumes that are being imported so
    ### that they are automatically mounted with their
    ### last filesystem mount point.
    $import = (exists $args{-_import}) ? $args{-_import} : 0;

    $fsck = (exists $args{-fsck}) ? $args{-fsck} : 0;

    $name = $self->get_name();

    $cmd = "vxdg -g $name dginfo";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::_create_vols()",
                        "Unable to load existing volumes in disk group",
                        "status=$status", @output);
        return $ClientUtil::FALSE;
    }
    
    # The above cmd will have the following type of output
    
    #### Names of the Volumes are....
    #### \Device\HarddiskDmVolumes\n_vg\nvol-1 (A:)
    #### \Device\HarddiskDmVolumes\n_vg\nvol-2 (B:)
    #### \Device\HarddiskDmVolumes\n_vg\nvol-3 (E:)
    #### \Device\HarddiskDmVolumes\n_vg\nvol-4 (F:)
    
    # From the above output I need to grep Vol names 

    @vols = grep (/\\Device\\HarddiskDmVolumes/i,@output);
    foreach my $line (@vols) {
        my (undef, undef, undef, undef, $volname) = split(/\\/, $line);
        $cmd = "vxvol -g $name volinfo $volname";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Windows::Vxvm::Vg::_create_vols()",
                            "Unable to get Vol info of Volume $volname");
            return $ClientUtil::FALSE;
        }
        
        # The above cmd's output will be like this

        #### Volume information
        #### Name        :   (A:)
        #### DeviceName  :   \Device\HarddiskDmVolumes\n_vg\nvol-1
        #### DriveLetter :   A:
        #### Disk Group  :   n_vg
        #### Size        :   1073741824
        #### Boot        :   No
        #### System      :   No
        #### PageFile    :   No
        #### Type        :   Mirrored Striped
        #### Status      :   Started
        #### Access      :   Read/Write
        #### Comment     :
        #### Name of the plex are...

        # from above output we need to extract Vol size
        my ($line) = grep (/Size\s+:/i,@output);
        my ($size) = ($line =~ /Size\s+:\s+(\d+)/i) ;
        my $sz = $size / (1024 * 1024);

        $volObj = $self->create_vol(-name => $volname, -size => $sz, 
                                    -exist => 1, -_import => $import, 
                                    -fsck => $fsck);
        return $ClientUtil::FALSE  if ($self->error);
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: create_vol
# Objective:       create and return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub create_vol {
    my ($self, %args) = @_;
    my ($volObj);

    ### Change layout to concat if dg contains only 1 device
    if (exists $args{-layout} and length $args{-layout}) {
        if ($args{-layout} =~ /stripe/i and $self->get_device_cnt() == 1) {
            $args{-layout} = "CONCAT";
        }
    }

    $volObj = San::SanClient::Lvm::Windows::Vxvm::Vg::Vol->_new(-parent => $self,
                                                              -conn   => $self->{conn},
                                                              %args);

    if ($volObj->error) {
        $self->_set_err($volObj->errmsg());
        return $volObj;
    }

    push @{$self->{vols}}, $volObj;

    return $volObj;
}

################################################################################
# Subroutine Name: destroy_vol
# Objective:       destroy a named vol object
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_vol {
    my ($self, $name) = @_;
    my ($volObj, @vols);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::destroy_vol() volname is a required argument");
        return $ClientUtil::FALSE;
    }

    # while creating or deleting the vols we should make sure Dynamic DG is imported
    $self->import();
    return $ClientUtil::FALSE if ($self->error);

    ### Delete the of vg object and update the parent vg list.
    foreach $volObj (@{$self->{vols}}) {
        if ($volObj->get_name() =~ /^$name$/) {
            $volObj->destroy(-force => 1);
            if ($volObj->error) {
                $self->_set_err($volObj->errmsg());
                return $ClientUtil::FALSE;
            }
        } else {
            push @vols, $volObj;
        }
    }

    $self->{vols} = \@vols;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_devices
# Objective:       get list of devices used by disk group
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_devices {
    my ($self) = @_;
    my ($cmd, $name, $deported, $status, @output, %hDevice);

    tie %hDevice, 'Tie::IxHash';

    $name     = $self->get_name();
    $deported = $self->{deported};

    $cmd = "vxdisk list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::_get_devices() Unable to get devices for volgroup \'$name\'", @output);
        return {};
    }

    my @devices = grep (/$name/,@output);

    $self->{disk_cnt} = scalar(@devices);

    foreach my $line (@devices) {
        my @flds = split(/\s+/, $line);
        $hDevice{$flds[0]} = $flds[1];
    }

    return \%hDevice;

}

################################################################################
# Subroutine Name: _create()
# Objective:       create a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self, $name, $devices) = @_;
    my ($attribute, $cmd, $disk_cnt, $disk_name, $format, @output, $status, %hDevice);

    $disk_cnt  = 0;

    ### Use the first device to determine the
    ### format of the disk group
    my ($drive_num) = (@$devices[0] =~ /\\\\.\\PHYSICALDRIVE(\d+)/);
    my $tmp_dev = "Harddisk".$drive_num ;
    $format = $self->{parent}->get_disk_format($tmp_dev);
    if ($self->{parent}->error) {
        $self->_set_err($self->{parent}->errmsg);
        return $ClientUtil::FALSE;
    }

    foreach my $device (@$devices) {
        my ($drive_num) = ($device =~ /\\\\.\\PHYSICALDRIVE(\d+)/);
        my $tmp_dev = "Harddisk".$drive_num ;

        $disk_cnt++;
        $disk_name = "disk-$disk_cnt";
        if ($disk_cnt == 1) {
            $cmd = "vxdg -g $name init $disk_name=$tmp_dev";
        } else {
            $cmd = "vxdg -g $name adddisk $disk_name=$tmp_dev";
        }
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Windows::Vxvm::Vg::_create Unable to create diskgroup \'$name\'", @output);
            return $ClientUtil::FALSE;
        }
        $self->{$device} = $format;
        $hDevice{$device} = $disk_name;

    }

    $self->{devices}  = \%hDevice;
    $self->{disk_cnt} = $disk_cnt;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: add_device()
# Objective:       add a device to the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub add_device {
    my ($self, $device) = @_;
    my ($cmd, $disk_cnt, $disk_name, $name, @output, $status);

    $self->_clear_err();

    ### Get name of volume group.
    $name = $self->get_name();

    ### Increment device counter
    $disk_cnt = $self->{disk_cnt};
    $disk_cnt++;
    $disk_name = "disk-$disk_cnt";

    $cmd = "vxdg -g $name adddisk $disk_name=$device";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::add_device Unable to create diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    $self->{devices}->{$device} = $disk_name;
    $self->{disk_cnt} = $disk_cnt;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: remove_device()
# Objective:       remove a device from the volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub remove_device {
    my ($self, %args) = @_;
    my ($cmd, $device, $force, $disk_cnt, $disk_name, $name, @output, @vols_mount, $status);

    $self->_clear_err();

    $force = "";

    # parse arguments
    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::remove_device()", "Invalid value for argument '-device'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::remove_device()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            $force = "-f";
            next;
        }

    }

    ### get name of volume group
    $name = $self->get_name();

    ### verify the device is being used in volume group
    unless (exists $self->{devices}->{$device}) {
        $self->_set_err("Error: Windows::Vxvm::Vg::remove_device()", "Device '$device' doesn't exist in diskgroup");
        return $ClientUtil::FALSE;
    }

    $disk_name = $self->{devices}->{$device};

    # Check for the force flag being specified and add it to the
    # rmdisk line
    $cmd = "vxdg -g $name $force rmdisk $disk_name";

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::remove_device Unable to remove $device from diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    ### remove device from hash
    delete($self->{devices}->{$device});

    ### decrement device counter;
    $self->{disk_cnt}--;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rename()
# Objective:       rename volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, %args) = @_;
    my ($cmd, $force, $name, $newname, @output, $status, @vols_mount);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?name/) {
            if (not length($newname = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::rename()", "Invalid value for argument '-name'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::rename()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }

    }

    ### get name of disk group
    $name = $self->get_name();

    ### verify the device is being used in volume group
    unless (defined $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::rename()", "Volume group doesn't have a name");
        return $ClientUtil::FALSE;
    }

    ### verify the device is being used in volume group
    unless (defined $newname) {
        $self->_set_err("Error: Windows::Vxvm::Vg::rename()", "New name for Volume group not specified");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "Rename volume group Old Name: " . $name, "New Name: " . $newname);

    ### Deport the diskgroup
    $self->deport();
    return $ClientUtil::FALSE if ($self->error);

    $self->import(-rename_to => $newname);
    return $ClientUtil::FALSE if ($self->error);

    $self->{name} = $newname;

    $self->{deported} = 0;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self) = @_;

    my (@disks, $cmd, $name, @output, $status, @vgs);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::destroy() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    # This might be weird but I could not find a cmd to destroy Dynamic Disk Group
    # But once all the volumes got deleted and all the disk are removed and 
    # DG will automatically got deleted. We just need to run "vxassist rescan"

    $cmd = "vxdisk list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::destroy() Unable to destroy diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    # But before we try to remove the Disks from Dynamic DG we should import it.
    $self->import();
    return $ClientUtil::FALSE if ($self->error);

    @disks = grep(/$name/,@output);
    foreach my $line (@disks) {
        (my $disk) = split(/\s+/,$line);
        $self->remove_device(-device=>$disk, -force =>1);
        return $ClientUtil::FALSE if ($self->error);
    }

    # Rescan the storage devices
    $cmd = "vxassist rescan";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::destroy() Unable to destroy diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    ### Remove the vgobj from the parent lvm object
    foreach my $vgObj (@{$self->{parent}->{vgs}}) {
        if ($vgObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vgs, $vgObj;
        }
    }

    $self->{parent}->{vgs} = \@vgs;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: display()
# Objective:       Display contents of Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    my ($self, %args) = @_;
    my (@verbose, $cmd, $name, @output, $status, $header, $summary, @summary, $verbose);

    $self->_clear_err();

    $verbose = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?verbose/) {
            if (not length($verbose = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::display()", "Invalid value for argument '-verbose'");
                return $self;
            }
            next;
        }
    }

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::display() Diskgroup name is a required parameter");
        return (wantarray()) ? () : [];
    }

    push @summary, "VG Name:  " . $self->get_name();
    push @summary, "Device Cnt:  " . $self->get_device_cnt();

    foreach my $device ($self->get_devices()) {
        push @summary, "Device:  " . $device . " (" . $self->{parent}->get_disk_format($device) . ")";

    }

    $header = sprintf("Vol Summary:\n"
                        . "%-12.12s %-12.12s %-7.7s %-5.5s %-12.12s %-32.32s\n"
                        . "%-12.12s %-12.12s %-7.7s %-5.5s %-12.12s %-32.32s",
                      "VGROUP", "VOLUME", "LAYOUT", "TYPE",  "SIZE(mb)", "MOUNT_PT",
                      "=" x 12, "=" x 12, "=" x 7,  "=" x 5, "=" x 12,   "=" x 32);

    push @summary, $header;

    foreach my $volObj (@{$self->{vols}}) {
        $summary = sprintf("%-12.12s %-12.12s %-7.7s %-5.5s %-12.12s %-s",
                           $self->get_name(), $volObj->get_name(), $volObj->get_layout(),
                           $volObj->get_fs(), $volObj->get_size(), $volObj->get_mount());

        push @summary, $summary;
    }

    if (defined $verbose and $verbose =~ /1|on|y|yes|t|true/) {

        $cmd = "/sbin/vxprint -lg $name";

        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status gt 0) {
            $self->_set_err("Error: Windows::Vxvm::Vg::display() Unable to display info for diskgroup \'$name\'", @output);

            return (wantarray()) ? () : [];
        }

        @verbose = (@summary, "\nVgroup Detail:", @output);

        push @verbose, "\n\n";
        return (wantarray()) ? @verbose : \@verbose;
    } else {
        push @summary, "\n\n";
        return (wantarray()) ? @summary : \@summary;
    }
}

################################################################################
# Subroutine Name: get_device_cnt
# Objective:       Return number of devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_device_cnt {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{disk_cnt}) {
        return $self->{disk_cnt};
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::get_device_cnt() Unable to display disk count for diskgroup");
        return 0;
    }

}

################################################################################
# Subroutine Name: get_devices
# Objective:       Return devices in volume group
# Details:         See POD documentation at the end of this file
################################################################################
sub get_devices {
    my ($self) = @_;
    my (@output);

    $self->_clear_err();

    @output = (keys %{$self->{devices}});

    return (wantarray()) ? @output : \@output;

}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vg object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::get_name() Vg object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vol
# Objective:       return a volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol {
    my ($self, $name) = @_;
    my ($vgObj);

    $self->_clear_err();

    unless (defined $name and length $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::get_vol() volume name is a required argument");
        return {};
    }

    foreach my $obj (@{$self->{vols}}) {
        if ($obj->get_name() =~ /^$name$/) {
            return $obj;
        }
    }

    $self->_set_err("Error: Windows::Vxvm::Vg::get_vol() Unable to find volume '$name'");

    return {};
}

################################################################################
# Subroutine Name: get_vols
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vols {
    my ($self) = @_;
    my (@volObjs);

    $self->_clear_err();
    if (defined $self->{vols}) {
        @volObjs = @{$self->{vols}};
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::get_vols() Vg object has no volumes");
        return (wantarray()) ? () : [];
    }

    return (wantarray()) ? @volObjs : \@volObjs;

}

################################################################################
# Subroutine Name: get_vol_cnt
# Objective:       return array of volume objects
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vol_cnt {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{vols}) {
        return scalar(@{$self->{vols}});
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::get_vol_cnt() Vg object has no vol count");
        return 0;
    }

}

################################################################################
# Subroutine Name: get_size
# Objective:       Return object size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size);

    $self->_clear_err();

    $size    = 0;

    $name = $self->get_name();

    $cmd = "vxdg -g $name dginfo";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::get_size() Unable to get size (mb) for diskgroup \'$name\'", @output);

        return 0;
    }
    
    my @devices = grep (/\s+Harddisk\d+/i, @output);
    foreach my $device (@devices) { 
        $cmd = "vxdisk -g $name diskinfo $device";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Windows::Vxvm::Vg::get_size() Unable to get size (mb)" . 
                            "for diskgroup \'$name\'", @output);
            return 0;
        }
        my ($line) = grep (/Length\s+:\s+\d+/i,@output);
        my ($length) = ($line =~ /Length\s+:\s+(\d+)/i);
        $size = $size + $length;
    }

    $size = $size / 1024;      #kilobytes
    $size = $size / 1024;      #megabytes

    return $size;

}

################################################################################
# Subroutine Name: get_freespace
# Objective:       Return available free space
# Details:         See POD documentation at the end of this file
################################################################################
sub get_freespace {
    my ($self) = @_;

    my ($cmd, $name, @output, $status, $size);

    $self->_clear_err();

    $name = $self->get_name();

    $cmd = "vxdg -g $name dginfo";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::get_freespace() Unable to get size (mb) for diskgroup \'$name\'", @output);

        return 0;
    }
    
    my @devices = grep (/\s+Harddisk\d+/i, @output);
    foreach my $device (@devices) { 
        $cmd = "vxdisk -g $name diskinfo $device";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Windows::Vxvm::Vg::get_size() Unable to get size (mb)" . 
                            "for diskgroup \'$name\'", @output);
            return 0;
        }
        my ($line) = grep (/FreeSpace\s+:\s+\d+/i,@output);
        my ($free_space) = ($line =~ /FreeSpace\s+:\s+(\d+)/i);
        $size = $size + $free_space;
    }

    $size = $size / 1024;      #kilobytes
    $size = $size / 1024;      #megabytes

    return $size;

}

################################################################################
# Subroutine Name: is_deported
# Objective:       Determine if volgroup has been deported
# Details:         See POD documentation at the end of this file
################################################################################
sub is_deported {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{deported}) {
        return $self->{deported};
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::is_deported() \$self->{deported} is not defined in _new");
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: deport()
# Objective:       deport Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub deport {
    my ($self) = @_;
    my ($cmd, $name, @output, $status);

    $self->_clear_err();

    $name = $self->get_name();

    unless (defined $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::deport() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    $cmd = "vxdg -g $name deport";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::deport() Unable to deport diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    }

    ### Remove volume objects from vg cache
    $self->{vols} = ();

    $self->{deported} = 1;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: import()
# Objective:       Import Vg object
# Details:         See POD documentation at the end of this file
################################################################################
sub import {
    my ($self, %args) = @_;
    my ($cmd, $name, @output, $status);

    my $new_name = "";
    $self->_clear_err();

    $name = $self->get_name();

    # parse arguments
    foreach (keys %args) {
        if (/-?rename_to/) {
            if (not length($new_name = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::import", "Invalid value for argument '-rename_to'");
                return $self;
            }
            next;
        }
    }

    unless (defined $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::import() Diskgroup name is a required parameter");
        return $ClientUtil::FALSE;
    }

    # The cmd will fail if the Dynamic DG is already imported. In this case the output will be like this
    # Failed to complete the operation...
    # V-76-58645-122: Dynamic disk group exists and is imported

    if ($new_name ne "") {
        $cmd = "vxdg -g $name -n $new_name import";
    } else {
        $cmd = "vxdg -g $name import";
    }
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0 and (!grep(/.*exists\s+and\s+is\s+imported/i, @output))) {
        $self->_set_err("Error: Windows::Vxvm::Vg::import() Unable to import diskgroup \'$name\'", @output);
        return $ClientUtil::FALSE;
    } elsif (grep(/.*exists\s+and\s+is\s+imported/i, @output)) {
        $self->sanlog(-msg=> "Dynamic DG $name was already imported");
        return $ClientUtil::TRUE;
    }

    $self->{deported} = 0;

    ### requery device list
    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    $self->_create_vols(-_import => 1, %args);

    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: refresh
# Objective:       refresh content of volume group.
# Details:         See POD documentation at the end of this file
################################################################################
sub refresh {
    my ($self) = @_;
    my ($cmd, $status, @output);

    $self->_clear_err();

    ### Remove existing vol cache
    $self->{vols} = undef;

    $self->sanlog(-msg => "refreshing volume group '" . $self->get_name() . "'");

    ### Determine if dg is deported
    $cmd = "/sbin/vxdisk -o alldgs list | grep '(" . $self->get_name() . ")'";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 1) {
        $self->_set_err("Error: Windows::Vxvm::Vg::refresh()", "Unable to load deported diskgroups");

        return $ClientUtil::FALSE;
    }

    $self->{deported} = 1 if (@output);

    $self->{devices} = $self->_get_devices();
    return $ClientUtil::FALSE if ($self->error);

    if (not $self->is_deported()) {
        $self->_create_vols();
        return $ClientUtil::FALSE if ($self->error);
    }

    return $ClientUtil::TRUE;
}

1;

