## @summary Provide an interface the Vxvm logical volumes
## @author rawat@netapp.com
## @status private
## @pod ../../../Vol.pod

package San::SanClient::Lvm::Windows::Vxvm::Vg::Vol;

use strict;

use San::SanClient::ClientUtil;
use Error;
use San::SanLog;
use San::SanUtil;

use File::Basename;

use vars qw( @ISA );
@ISA = qw(Error SanLog);

###############################################################################
# Method:    create_fs
# Objective: create a filesystem on a volume
# Details:   See POD documentation at the end of this file.
###############################################################################
sub create_fs {
    my ($self, %args) = @_;
    my ($cmd, $type, $status, @output, $mount);

    $self->_clear_err();

    $mount = $self->get_mount();

    # parse arguments
    foreach (keys %args) {
        if (/-?type$/) {
            $type = $args{$_};
            next;
        }
    }

    ### Default -type to value of ntfs, since we are in the vxfs libs
    # This can be overridden with an argument flag of -type
    if (not defined($type) or $type eq "") {
        $type = "ntfs";
    }

    $cmd = "echo y | format $mount: /FS:$type /Q";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::create_fs Unable to create $type filesystem on volume '$mount'", @output);
        return $ClientUtil::FALSE;
    }   

    # There is no way to find out the fstype of Volume. The following
    # cmd will set comment field of the volume to 'FS:$type"
    # Later when we need fstype we can use this field.
    $cmd = "vxedit set comment=\'FS:$type\' $mount:";
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::create_fs Unable to set comments for the volume '$mount'");
        return $ClientUtil::FALSE;
    }   


    $self->{filesystem} = $type;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: destroy
# Objective:       destroy a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy {
    my ($self, %args) = @_;

    my ($cmd, $force, $dgname, $name, $mount, $status, @vols, @output);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?force/) {
            if (not length($force = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::Vol::destroy()", "Invalid value for argument '-force'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $dgname = $self->get_vgname();
    $name   = $self->get_name();
    $mount  = $self->get_mount();

    unless (defined $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::destroy() Diskgroup is a required parameter");
        return $ClientUtil::FALSE;
    }
    my $force_option ;
    if ($force) {
        $force_option = "-f";
    }
    $cmd = "vxassist $force_option remove volume $mount:";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::destroy() Unable to destroy volume \'$name\'".
                            "in diskgroup '$dgname'");
        return $ClientUtil::FALSE;
    }

    ### Remove the volume from the parent vol group object
    foreach my $volObj (@{$self->{parent}->{vols}}) {
        if ($volObj->get_name() =~ /^$name$/) {
            next;
        } else {
            push @vols, $volObj;
        }
    }
    $self->{parent}->{vols} = \@vols;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: display()
# Objective:       No-op
# Details:         See POD documentation at the end of this file
################################################################################
sub display {
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_fs
# Objective:       Return Vol fs type
# Details:         See POD documentation at the end of this file
################################################################################
sub get_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{filesystem} and length $self->{filesystem}) {
        return $self->{filesystem};
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::get_fs() Vol object is missing filesystem");
        return "";
    }
}

################################################################################
# Subroutine Name: get_layout
# Objective:       Return volume layout
# Details:         See POD documentation at the end of this file
################################################################################
sub get_layout {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{layout}) {
        return $self->{layout};
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::get_layout() Vol object is missing layout");
        return "";
    }
}

################################################################################
# Subroutine Name: get_name
# Objective:       Return Vol object name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_name {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{name}) {
        return $self->{name};
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::get_name() Vol object is missing name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_size
# Objective:       Return ooject size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{size}) {
        return $self->{size};
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::get_size() Vol object is missing size");
        return 0;
    }
}

################################################################################
# Subroutine Name: get_state
# Objective:       Return object state
# Details:         See POD documentation at the end of this file
################################################################################
sub get_state {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{state}) {
        return $self->{state};
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::get_state() Vol object is missing state");
        return "";
    }
}

################################################################################
# Subroutine Name: get_vgname
# Objective:       Return Vol object diskgroup name
# Details:         See POD documentation at the end of this file
################################################################################
sub get_vgname {
    my ($self) = @_;
    my ($name);

    $self->_clear_err();

    $name = $self->{parent}->get_name();

    if (defined $name and length $name) {
        return $name;
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::get_vgname() Vol object is missing parent name");
        return "";
    }
}

################################################################################
# Subroutine Name: get_raw_device
# Objective:       Return path to raw vol
# Details:         See POD documentation at the end of this file
################################################################################
sub get_raw_device {
    my ($self) = @_;

    $self->_clear_err();

    ### get name of disk group
    my $vgname = $self->get_vgname();

    ### get name of volume
    my $name = $self->get_name();

    return "\\Device\\HarddiskDmVolumes\\$vgname\\$name";
}

################################################################################
# Subroutine Name: is_fs
# Objective:       Determine if volume has a filesystem
# Details:         See POD documentation at the end of this file
################################################################################
sub is_fs {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{filesystem} and $self->{filesystem} !~ /raw/i) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: is_mounted
# Objective:       No-op for Windows
# Details:         See POD documentation at the end of this file
################################################################################
sub is_mounted {
    my ($self) = @_;
    $self->_clear_err();

    if (defined $self->{mount} and length $self->{mount}) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: mount
# Objective:       No-op for Windows
# Details:         See POD documentation at the end of this file
################################################################################
sub mount {
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rename()
# Objective:       rename volume
# Details:         See POD documentation at the end of this file
################################################################################
sub rename {
    my ($self, $newname) = @_;
    my ($cmd, $name, $vgname, @output, $status);

    $self->_clear_err();

    ### get name of disk group
    $vgname = $self->get_vgname();

    ### get name of volume
    $name = $self->get_name();

    ### verify the device is being used in volume group
    unless (defined $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::rename()", "Volume doesn't have a name");
        return $ClientUtil::FALSE;
    }

    ### verify the device is being used in volume group
    unless (defined $newname) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::rename()", "New name for Volume is not specified");
        return $ClientUtil::FALSE;
    }

    ### Rename the volume
    $cmd = "vxedit -g $vgname $name $newname";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::rename Unable to rename volume '$name' to \'$newname\'", @output);
        return $ClientUtil::FALSE;
    }

    $self->{name} = $newname;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: rescan()
# Objective:       No-op for Windows
# Details:         See POD documentation at the end of this file
################################################################################
sub rescan {
    my ($self) = @_;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: unmount()
# Objective:       No-op for Windows
# Details:         See POD documentation at the end of this file
################################################################################
sub unmount {
    my ($self) = @_;

    return $ClientUtil::TRUE;
}

################################################################################
#
# P R I V A T E   S U B R O U T I N E S
#
###############################################################################
################################################################################
# Subroutine Name: _create()
# Objective:       create a volume
# Details:         See POD documentation at the end of this file
################################################################################
sub _create {
    my ($self) = @_;
    my ($cmd, $device_cnt, $dg, $name, $layout, $size, $required_size, $min_disks_required);
    my ($lvmObj, $status, @output, @disks);

    $dg         = $self->get_vgname();
    $device_cnt = $self->{parent}->get_device_cnt();
    $name       = $self->get_name();
    $size       = $self->get_size();
    $layout     = $self->get_layout();
    $lvmObj     = $self->{parent}->{parent};

    ## The hash below has the available drive letters that can be used to assign to LVs
    my $next_available_letter = shift(@{$self->{client_obj}->{availLetters}});
    my $no_of_columns = 2;
    my $no_of_mirrors = 2;
    my $attribute ="";
 
    ## We need to find the disks that are not used by any volume
    ## those disks will be used to create the new Vol 
    ## sub _get_unused_disks() will return enough unused disks to 
    ## just create this LV
    @disks = $self->_get_unused_disks();
    if($self->error) {
        $self->_set_err($self->errmsg());
        return $ClientUtil::FALSE;
    }

    ## No of columns used will depend on the layout and no of disks 
    ## used. e.g. for layout type is raid then no_of_columns = no_of_disks
    if (($layout =~ /stripe/i) and ($layout =~ /mirror/i)) {
        $layout = "stripe,mirror";
        $no_of_columns = scalar(@disks)/2;
        $attribute = "mirror=$no_of_mirrors";
    } elsif ($layout eq "stripe") {
        $no_of_columns = scalar(@disks);
    } elsif ($layout eq "mirror") {
        $attribute = "mirror=$no_of_mirrors";
    } elsif ($layout =~ /raid/i) {
        $no_of_columns = scalar(@disks);
    }

    if ($layout eq "concat") {
        $cmd = "vxassist -b -g $dg make $name $size $attribute";
    } else {
        $cmd = "vxassist -b -g $dg make $name $size type=$layout $attribute";
    }
    if ($layout =~ /stripe/i or $layout =~ /raid/i) {
        $cmd .= " column=$no_of_columns";
    }
    $cmd .= " DriveLetter=$next_available_letter";

    foreach my $disk (@disks) {
        
        ## We can specify the disks that we want to be used in volume creation 
        $cmd .= " $disk";
        $lvmObj->{$disk}{vol} = $name;
    }

    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::_create Unable to create volume \'$name\'", "CMD: $cmd", @output);
        return $ClientUtil::FALSE;
    }

    my $state = "Resynching";
    while ($state =~ /Resynching/i) {
        $cmd = "vxvol -g $dg volinfo $next_available_letter:";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: Windows::Vxvm::Vg::Vol::_create Unable to get Volinfo of vol $next_available_letter:");
            return $ClientUtil::FALSE;
        }

        my ($line) = grep (/Status\s+:/i,@output);
        ($state) = ($line =~ /Status\s+:\s+\w+,\s+(\w+)/i);
        &SanUtil::nate_sleep(60);
    }

    # Storing the drive letters for Dynamic Volume
    $self->{mount} = $next_available_letter;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_unused_disks 
# Objective:       Return unused disks
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_unused_disks {
    my ($self) = @_;
    my ($free_disks, $lvmObj, $controller_1, $controller_2, $count1, $count2);
    my ($layout, $dg, $dev, $no_of_disks, $min_size, $disks_required, $cmd);
    my ($status, @output, @devices, @disks, $controller_to_use);

    $layout = $self->get_layout();
    ## We assume that the luns supplied are of same size

    $self->_clear_err();

    $dg = $self->get_vgname();

    $lvmObj = $self->{parent}->{parent};

    ## Minimum number of disks that we are going to use to create volume is 4
    ## This is the recommended value
    $disks_required = 4;

    $controller_1 = $lvmObj->{controller_1};
    $controller_2 = $lvmObj->{controller_2};
    $free_disks = $lvmObj->{$controller_1}{unused_disks} + $lvmObj->{$controller_2}{unused_disks};

    ## failed out if required number of disks are more than available free disks
    if ($free_disks < $disks_required) {
        $self->_set_err ("Error: Windows::Vxvm::Vg::Vol::_create Not enough devices are available");
        return (wantarray()) ? () : [];
    }

    $cmd = "vxdisk list";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::_create unable to list the disks");
        return (wantarray()) ? () : [];
    }

    foreach my $line (@output) {
        if ($line =~ /$dg/) {
            ($dev) = ($line =~ /(\w+\d+)\s+.+/i);
            push (@devices, $dev);
        }
    }

    ## Here assumption is that all disks are of same size. But user can pass disks
    ## of different sizes. Now before creating the volume make sure that
    ## there are enough disks available to create the LV. The calculation (whether
    ## LV can be created on not) depends on the size (assumption:all disks are of same size)
    ## and number of disk. If user passes disks of different size then least size(of disks)
    ## should be passed for the calculation because if largest size is passed then 
    ## calculation will show that volume can be created but it might not be the case
    
    ## If all the disks are of same size then min_size will be equal to the disk size
    $min_size = 1000000000;

    $count1 = 0;
    $count2 = 0;
    foreach $dev (@devices) {

        ## Initially we select 2 disks from each controller
        if ($controller_1 eq $lvmObj->{$dev}{controller} and $lvmObj->{$dev}{in_use} != 1 
            and $count1 < 2) {
            push(@disks,$dev);
            $lvmObj->{$dev}{in_use} = 1;
            if ($min_size > $lvmObj->{$dev}{size}) {
                $min_size = $lvmObj->{$dev}{size};
            }
            $lvmObj->{$controller_1}{unused_disks}--;
            $free_disks--;
            $count1++;
        }
        if ($controller_2 eq $lvmObj->{$dev}{controller} and $lvmObj->{$dev}{in_use} != 1 
            and $count2 < 2) {
            push(@disks,$dev);
            $lvmObj->{$dev}{in_use} = 1;
            if ($min_size > $lvmObj->{$dev}{size}) {
                $min_size = $lvmObj->{$dev}{size};
            }
            $lvmObj->{$controller_2}{unused_disks}--;
            $free_disks--;
            $count2++;
        }
        last if ($count2 + $count1 == 4);
    }
    
    $no_of_disks = 4;

    if($lvmObj->{$controller_2}{unused_disks} > $lvmObj->{$controller_1}{unused_disks}) {
        $controller_to_use = $controller_2;
    } else {
        $controller_to_use = $controller_1;
    }

    ## Here each time we see if it is possible to create the Volume of given size
    ## with the extracted free disks from unused_disk_list. If not then each time we select the extra 
    ## (1 or 2 depeding upon the layout) disks from unused_disk_list and repeat the process until
    ## we have enough free disks to create the Vol.
    ## If unused_disk_list exhausted and we still dont have enough disk to create the volume of 
    ## required size then we error out
    while ( $self->_calculate_vol_size($min_size, $no_of_disks) == $ClientUtil::FALSE ) {
        if ($free_disks == 0) {
            $self->_set_err("Error: Windows::Vxvm::Vg::Vol::_create No devices are left");
            return (wantarray()) ? () : [];
        }
        my $count = 0;
        foreach my $dev (@devices) {
            if ($controller_to_use eq $lvmObj->{$dev}{controller} and $lvmObj->{$dev}{in_use} != 1) {
                push(@disks,$dev);
                $no_of_disks++;
                $lvmObj->{$dev}{in_use} = 1;
                if ($min_size > $lvmObj->{$dev}{size}) {
                    $min_size = $lvmObj->{$dev}{size};
                }
                $lvmObj->{$controller_to_use}{unused_disks}--;
                $free_disks--;
                $controller_to_use = ($controller_to_use eq $controller_1) ? $controller_2 : $controller_1;
                
                ## if layout type is mirror, stripe or stripped-mirror then we should 
                ## pass even no of disks to the subroutine otherwise our calculation 
                ## to calculate vol_size will not work be accurate
                if ( $layout =! /mirror|stripe/ or $count == 1) {
                    last ;
                }
                $count++;
            }
        }
    }
    
    return (wantarray()) ? @disks : \@disks;
}

################################################################################
# Subroutine Name: _calculate_vol_size
# Objective:       Calculate the size of the volume that can be created by the
#                  disks passed
# Details:         See POD documentation at the end of this file
################################################################################
sub _calculate_vol_size {
    my ($self, $disk_size, $no_of_disks) = @_;
    my ($size_tmp, $size_in_mb) ;

    my $layout = $self->get_layout();
    my $size = $self->get_size();

    my $size_in_bytes = &ClientUtil::convert_to_bytes($size);
    $size_in_mb = $size_in_bytes/(1024*1024);
    
    $size_tmp = 0;

    if (($layout =~ /stripe/i) and ($layout =~ /mirror/i)) {
        $size_tmp = int ($disk_size * $no_of_disks)/2;
        if ($size_in_mb < $size_tmp) {
            return $ClientUtil::TRUE;
        }
    } elsif ($layout eq "concat") {
        $size_tmp = $disk_size * $no_of_disks;
        if ($size_in_mb < $size_tmp) {
            return $ClientUtil::TRUE;
        }
    } elsif ($layout eq "stripe") {
        $size_tmp = $disk_size * $no_of_disks;
        if ($size_in_mb < $size_tmp) {
            return $ClientUtil::TRUE;
        }
    } elsif ($layout eq "mirror") {
        $size_tmp = int ($disk_size * $no_of_disks)/2;
        if ($size_in_mb < $size_tmp) {
            return $ClientUtil::TRUE;
        }
    } elsif ($layout =~ /raid/i) {
        $size_tmp = $disk_size * ($no_of_disks - 2);
        if ($size_in_mb < $size_tmp) {
            return $ClientUtil::TRUE;
        }
    }
    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: get_mount
# Objective:       Return ooject size
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mount {
    my ($self) = @_;

    $self->_clear_err();

    if (defined $self->{mount}) {
        return $self->{mount};
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::get_mount() Vol object is missing mount");
        return "";
    }
}

################################################################################
# Subroutine Name: _init_fstype
# Objective:      Initialize fstype for existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_fstype {
    my ($self) = @_;
    my ($cmd, $fstype, $vgname, $name, @output, $status, $mount);

    $name   = $self->get_name();
    $mount= $self->get_mount();

    $cmd = "vxvol volinfo $mount:";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) { 
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::init_fstype Unable to get volinfo of $name($mount:)");
        return $ClientUtil::FALSE;
    }
    my ($line) = grep (/Comment\s+:/i,@output);
    ($fstype) = ($line =~ /.*FS:(\w+)/i);

    return $fstype;

}

################################################################################
# Subroutine Name: _init_layout
# Objective:      Initialize layout of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_layout {
    my ($self) = @_;

    my ($cmd, $vgname, $layout, $name, @output, $status, $mount);

    $layout = "";

    $mount= $self->get_mount();
    $name   = $self->get_name();

    $cmd = "vxvol volinfo $mount:";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::init_layout",
                        "Unable to get layout info of Volume $name");
        return "";
    }
    my ($line) = grep (/Type\s+:/i,@output);
    ($layout) = ($line =~ /Type\s+:\s+(\w+)/i) ;

    return $layout;

}

################################################################################
# Subroutine Name: _init_size
# Objective:      Initialize size of existing host volumes
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_size {
    my ($self) = @_;

    my ($cmd, $vgname, $size, $name, @output, $sectors, $status, $mount);

    $mount= $self->get_mount();
    $name   = $self->get_name();
    $cmd = "vxvol  volinfo $mount:";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::_create_vols()",
                        "Unable to get Vol info of Volume $name");
        return 0;
    }
    my ($line) = grep (/Size\s+:/i,@output);
    ($size) = ($line =~ /Size\s+:\s+(\d+)/i) ;
    my $sz = $size / (1024 * 1024);

    ### The return will return a integer that has been rounded down
    ### It will also handle negative values
    return int($sz + .5 * ($sz <=> 0));

}

################################################################################
# Subroutine Name: _init_mount
# Objective:      Initialize state of existing host volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_mount {
    my ($self) = @_;

    my ($cmd, $name, $state, $vgname, $status, @output, $mount);

    $mount = "";

    $vgname = $self->get_vgname();
    $name   = $self->get_name();

    $cmd = "vxvol -g $vgname volinfo $name";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::_init_state()"
                        . "Unable to retrieve state for volume \'$name\' in diskgroup '$vgname'",
                   @output);
        return "";
    }

    my ($line) = grep (/DriveLetter\s+:/i,@output);
    ($mount) = ($line =~ /DriveLetter\s+:\s+(\w+):/i);

    return $mount;
}

################################################################################
# Subroutine Name: _init_state
# Objective:      Initialize state of existing host volume
# Details:        See POD documentation at the end of this file
################################################################################
sub _init_state {
    my ($self) = @_;

    my ($cmd, $name, $state, $vgname, $status, @output, $mount);

    $state = "";

    $mount= $self->get_mount();
    $vgname = $self->get_vgname();
    $name   = $self->get_name();

    $cmd = "vxvol volinfo $mount:";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::_init_state()"
                        . "Unable to retrieve state for volume \'$name\' in diskgroup '$vgname'",
                   @output);
        return "";
    }

    my ($line) = grep (/Status\s+:/i,@output);
    ($state) = ($line =~ /Status\s+:\s+(\w+)/i);

    return $state;
}

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new volume object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($cmd, $exist,  $fsck, $import, $name, $size, $conn, $layout, 
        $log, $parent, $self, $status, @output);

    $self = {};

    bless $self, $class;

    #default layout is concatenation
    $layout = "concat";
    $import = 0;
    $exist  = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?layout$/) {
            if (not length($layout = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Invalid value for argument '-layout'");
                return $self;
            }
            next;
        } elsif (/-?_import$/) {    ### private member
            if (not length($import = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Invalid value for argument '-_import'");
                return $self;
            }
            next;
        } elsif (/-?name$/) {
            if (not length($name = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Invalid value for argument '-name'");
                return $self;
            }
            next;
        } elsif (/-?size$/) {
            if (not length($size = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Invalid value for argument '-size'");
                return $self;
            }
            next;
        } elsif (/-?fsck$/) {
            if (not length($fsck = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Invalid value for argument '-fsck'");
                return $self;
            }
            next;
        } elsif (/-?exist$/) {
            if (not length($exist = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Invalid value for argument '-exist'");
                return $self;
            }
            next;
        } elsif (/-?conn$/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Invalid value for argument '-conn'");
                return $self;
            }
            next;
        } elsif (/-?parent$/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Missing value for argument '-parent'");
                return $self;
            }
            next;
        }
    }

    ### check for existance of required -parent argument
    unless (defined $parent) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Missing required argument '-parent'");
        return $self;
    }

    ### check for existance of required -name argument
    unless (defined $name) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Missing required argument '-name'");
        return $self;
    }

    ### check for existance of required -size argument
    unless (defined $size) {
        $self->_set_err("Error: Windows::Vxvm::Vg::Vol::new()", "Missing required argument '-size'");
        return $self;
    }

    ### check for existance of required -conn argument
    if (grep (/^-?conn$/, (keys %args))) {
        unless (ref($args{-conn}) eq "Connectrec") {
            $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Invalid value for argument '-conn'");
            return $self;
        }
    } else {
        $self->_set_err("Error: Windows::Vxvm::Vg::new()", "Missing required argument '-conn'");
        return $self;
    }

    # initilize the object members
    $self->{conn}       = $conn;
    $self->{name}       = $name;
    $self->{layout}     = $layout;
    $self->{size}       = $size;
    $self->{parent}     = $parent;
    $self->{state}      = 'Started';
    $self->{filesystem} = 'raw';
    $self->{client_obj} = $self->{parent}->{parent}->{parent}->{WinClient};

    if (not $exist) {
        $self->_create();
        return $self if ($self->error);
    } else {
        $self->sanlog(-msg => "loading existing volume '$name'");
        $self->{mount} = $self->_init_mount();
        $self->{state} = $self->_init_state();
        return $self if ($self->error);
        $self->sanlog(-msg => "  state: " . $self->{state}, -msg_level => 7);
        $self->{filesystem} = $self->_init_fstype();
        $self->sanlog(-msg => "  fs: " . $self->{filesystem}, -msg_level => 7);
        if ($self->error) {

            ### allow vol to be loaded although it
            ### can't be started to determine fs
            if ($self->{filesystem} eq "") {
                $self->_clear_err();
            } else {
                return $self;
            }
        }

        if ($self->{filesystem} !~ /raw/i) {
            $self->{mount} = $self->_init_mount($import, $fsck);
            return $self if ($self->error);
            $self->sanlog(-msg => "  mnt: " . $self->{mount}, -msg_level => 7);
        }

        $self->{size} = $self->_init_size();
        return $self if ($self->error);
        $self->sanlog(-msg => "  size: " . $self->{size} . " mb" , -msg_level => 7);

        $self->{layout} = $self->_init_layout();
        return $self if ($self->error);
        $self->sanlog(-msg => "  layout: " . $self->{layout}, -msg_level => 7);
    }

    return $self;

}

1;

