# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to MP (MultiPathing) software
## @author latif@netapp.com, thomasm@netapp.com
## @status private
## @pod here

package MP;

# Compiler directives.
use strict;

# Module import.
use San::SanLog;
use San::SanClient::ClientUtil;
use Error;
use Data::Dumper;
use vars qw(@ISA);

@ISA = qw(Error SanLog);

###############################################################################
# Public Methods
###############################################################################

###############################################################################
# Method:    new
# Objective: Create new Host MP object.
# Details:   See POD documentation at the end of this file.
###############################################################################
sub new {
    my ($class, %args) = @_;

    my ($conn, $self, $hosttype, $mp_type, $parent);

    $self = {};

    ## bless the object upfront in case an error occurs.
    bless $self, $class;

    # Process arguments.  If an error occurs during the arg processing,
    # The error flag and errmsg values are set and the object is returned.
    # It is up to the caller to check for errors after the new call.
    foreach (keys %args) {
        if (/-?hosttype$/) {
            length($hosttype = $args{$_})
              or do {
                $self->_set_err("Invalid value for \"-hosttype\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?mp_type$/) {
            length($mp_type = $args{$_})
              or do {
                $self->_set_err("Invalid value for \"-mp_type\":", "$args{$_}");
                return $self;
              };
        } elsif (/-?parent$/) {
            length($parent = delete $args{$_})
              or do {
                $self->_set_err("Invalid value for \"-parent\":", "$args{$_}");
                return $self;
              };
        }
    }

    # save parent in object
    $self->{parent} = $parent;

    # determine the right platform and require/import/call the appropriate
    # MP object's new call.
    if ($hosttype =~ /unix-solaris/i) {
        require San::SanClient::MP::SolarisMP;
        $self = San::SanClient::MP::SolarisMP->_new(-parent => $self,
                                                     %args);
    } elsif ($hosttype =~ /unix-linux/i) {
        require San::SanClient::MP::LinuxMP;
        $self = San::SanClient::MP::LinuxMP->_new(-parent => $self,
                                                  %args);
    } elsif ($hosttype =~ /unix-aix/i) {
        require San::SanClient::MP::AixMP;
        $self = San::SanClient::MP::AixMP->_new(-parent => $self,
                                                %args);
    } elsif ($hosttype =~ /unix-hpux/i) {
        require San::SanClient::MP::HPUXMP;
        $self = San::SanClient::MP::HPUXMP->_new(-parent => $self,
                                                 %args);
    } elsif ($hosttype =~ /linux-vmware/i) {
        if ($self->{parent}->has_vma_host()) {
            require San::SanClient::MP::ESXiMP;
            $self = San::SanClient::MP::ESXiMP->_new(-parent => $self,
                                                     %args);
        } else {
            require San::SanClient::MP::ESXMP;
            $self = San::SanClient::MP::ESXMP->_new(-parent => $self,
                                                    %args);
        }
    } elsif ($hosttype =~ /windows/i) {
        require San::SanClient::MP::WinMP;
        $self = San::SanClient::MP::WinMP->_new(-parent => $self,
                                                %args);
    } else {
        $self->_set_err("Host platform $hosttype not currently supported by MP object");
        return $self;
    }

    # save SanClient connection handle in object
    $conn = $self->{parent}->{parent}->{parent}->{conn};
    $self->{conn} = $conn;

    # MP set-up completed
    return $self;

}

###############################################################################
# Method:    gettype
# Objective: Return mp type
# Details:   See POD documentation at the end of this file.
###############################################################################
sub gettype {
    my ($self) = @_;

    return $self->{mp}->{mp_type};

}

################################################################################
# Subroutine Name: validate_paths()
# Objective:       validate the current grab of pathing information versus
#                  what was originally stored in the object
# Details:         See POD documentation at the end of this file
################################################################################
sub validate_paths {
    my ($self, $client_name, $takeover_state, $rule_type, $num_ports_disabled) = @_;

    my ($timeout, $level, $error, $errstr);
    my ($steady_paths, $curr_paths);
    my ($conn, $num_devices);

    $self->_clear_err();

    # set takeover_state to default if not set
    if (not defined($takeover_state)) {
        $takeover_state = 0;
    }

    # default the validate rule type to 7mode
    if (not defined($rule_type)) {
        $rule_type = "7mode";
    }

    # set num_ports_disabled to default if not set
    if (not defined($num_ports_disabled)) {
        $num_ports_disabled = 0;
    }

    if ($num_ports_disabled != 0 and $takeover_state != 0) {
        $self->_set_err("'$num_ports_disabled' and '$takeover_state' are mutually exclusive");
        return $ClientUtil::FALSE;
    }

    # validate the the rule is correct
    if ($rule_type !~ /^7mode$|^Cmode$/i) {
        $self->_set_err("Invalid validation rule: '$rule_type', use '7mode' or 'Cmode'");
        return $ClientUtil::FALSE;
    }

    # get steady state (before faults) path information.
    # This data was retrived before faults and stored in
    # Client object.
    $steady_paths = $self->get_steady_state();
    if ($self->error or
        not defined $steady_paths or $steady_paths eq "") {
        $self->_set_err("Cannot retrieve steady state information for client ($client_name)");
        return $ClientUtil::FALSE;
    }

    # gather pathing information from the client.
    $curr_paths = $self->get_path_info(-validate_mode => 1,
                                       -takeover_state => $takeover_state,
                                       -num_ports_disabled => $num_ports_disabled);
    if ($self->error or $curr_paths eq undef) {
        my $error_msg = join "\n", $self->errmsg();
        $self->_set_err("Failure calling get_path_info() for client ($client_name)\n" .
                        "Could not gather the pathing information, so no validation versus steady state will happen.\n" .
                        "Error Message: $error_msg");
        return $ClientUtil::FALSE;
    }

    # check if in takeover state since if we are the nexts will be different.
    # if not in takeover state and any port is not disabled, simply verify versus steady state.
    if ($takeover_state == 0 and $num_ports_disabled == 0) {
        $error = 0;
        # search through all paths to devices and look to see if
        # current pathing information to same as store information
        # about steady state.
        foreach my $key (keys (%{$steady_paths})) {
            # special case - check for no paths to current LUN
            # report the information differently
            if ($curr_paths->{$key}->{num_paths} == 0) {
                $error = 1;
                $errstr .= "No paths to LUN ($key) exist in current state.\n" .
                           "In steady state the LUN ($key) had ($steady_paths->{$key}->{num_paths}) paths\n\n";
            } else {
                # check num paths
                if ($steady_paths->{$key}->{num_paths} !=
                    $curr_paths->{$key}->{num_paths}) {
                    $error = 1;
                    $errstr .= "The number of paths to the LUN ($key) does not match\n" .
                               "Steady State  : $steady_paths->{$key}->{num_paths}\n" .
                               "Current State : $curr_paths->{$key}->{num_paths}\n\n";
                } else {
                    $self->sanlog(-msg =>
                                  "Path check ($key) passsed - num_paths: " .
                                  $steady_paths->{$key}->{num_paths} . " = " .
                                  $curr_paths->{$key}->{num_paths},
                                  -msg_level => 3);
                }

                # check alive paths
                if ($steady_paths->{$key}->{num_paths_alive} !=
                    $curr_paths->{$key}->{num_paths_alive}) {
                    $error = 1;
                    $errstr .= "The number of alive paths to the LUN ($key) does not match\n" .
                               "Steady State  : $steady_paths->{$key}->{num_paths_alive}\n" .
                               "Current State : $curr_paths->{$key}->{num_paths_alive}\n\n";
                } else {
                    $self->sanlog(-msg =>
                                  "Path check ($key) passsed - alive_paths: " .
                                  $steady_paths->{$key}->{num_paths_alive} . " = " .
                                  $curr_paths->{$key}->{num_paths_alive},
                                  -msg_level => 3);
                }

                # check dead paths
                if ($steady_paths->{$key}->{num_paths_dead} !=
                    $curr_paths->{$key}->{num_paths_dead}) {
                    $error = 1;
                    $errstr .= "The number of dead paths to the LUN ($key) does not match\n" .
                               "Steady State  : $steady_paths->{$key}->{num_paths_dead}\n" .
                               "Current State : $curr_paths->{$key}->{num_paths_dead}\n\n";
                } else {
                    $self->sanlog(-msg =>
                                  "Path check ($key) passsed  - dead_paths: " .
                                  $steady_paths->{$key}->{num_paths_dead} . " = " .
                                  $curr_paths->{$key}->{num_paths_dead},
                                  -msg_level => 3);
                }
            }
        }
    } else {
        #### Takeover rules ####
        # 7Mode Rules:
        # FCP - validate in takeover that you have at least 1/2 the paths available.
        # don't check exact half since with some stacks, if the path is not used
        # it won't be marked failed and that will screw up the check.
        # iSCSI - validate that we have the same number of path since this is
        # simply IP failover versus FCP and proxy/SSI behavior

        # CMode Rules:
        # FCP - validate in takeover that you have at least 1/2 the paths available.
        # don't check exact half since with some stacks, if the path is not used
        # it won't be marked failed and that will screw up the check.
        # iSCSI - validate in takeover that you have at least 1/2 the paths available.
        # don't check exact half since with some stacks, if the path is not used
        # it won't be marked failed and that will screw up the check.

        #### Port blocks rules ####
        # Validate in the current state that you have exactly the number of paths in
        # steady state minus number of ports disabled

        $error = 0;
        foreach my $key (keys (%{$steady_paths})) {
            # special case - check for no paths to current LUN
            # report the information differently
            if ($curr_paths->{$key}->{num_paths} == 0) {
                $error = 1;
                $errstr .= "No paths to LUN ($key) exist in current state.\n" .
                           "In steady state the LUN ($key) had ($steady_paths->{$key}->{num_paths}) paths\n\n";
            } else {
                my $num_curr_paths = $curr_paths->{$key}->{num_paths};
                my $num_steady_paths = $steady_paths->{$key}->{num_paths};
                my $protocol = $steady_paths->{$key}->{protocol};
                my ($expected_paths);

                if ($num_ports_disabled > 0) {
                    # In case of port blocks, expected paths should be atleast the difference of
                    # number of steady state paths and number of ports disabled
                    $expected_paths = $num_steady_paths - $num_ports_disabled;
                } else {
                    # calculate number of expected paths, don't expect 0 paths
                    # (iscsi case mostly), expect at least 1.
                    # Also, look to see what prtocol and depending on protocol,
                    # update the expected number of paths (see rules above).
                    if ($protocol =~ /iscsi/i and $rule_type =~ /7mode/i) {
                        # 7mode iSCSI configs have target port failover so expect no change in steady paths
                        $expected_paths = $num_steady_paths;
                    } else {
                        # Cmode configs (iscsi/fcp) and 7mode configs (fcp) do not have target
                        # port failover so expect at least half the number of paths during takeover.
                        $expected_paths = int($num_steady_paths / 2);
                        $expected_paths = 1 if ($expected_paths == 0);
                    }
                }

                if ($num_curr_paths < $expected_paths) {
                    $error = 1;
                    $errstr .= "The number of paths to the LUN ($key) is not correct\n" .
                               "Steady State  : $steady_paths->{$key}->{num_paths}\n" .
                               "Current State : $curr_paths->{$key}->{num_paths}\n";
                    if ($num_ports_disabled > 0) {
                        $errstr .= "Number of ports disabled: $num_ports_disabled\n" .
                                   "In case of port blocks, you should have atleast the number of paths which is " .
                                   "the difference of number of steady state paths and number of ports disabled\n\n";
                    } else {
                        $errstr .= "In takeover state you should have at least 1/2 the number\n" .
                                   "of paths as were recorded in steady state for (7mode FCP, Cmode FCP & iSCSI) or\n" .
                                   "the same number of paths as were recorded in steady state for (7mode iSCSI)\n\n";
                    }
                } else {
                    my $error_string = "Path check ($key) (takeover state) passsed - num_paths: $num_curr_paths\n";
                    if ($num_ports_disabled > 0) {
                        $error_string .= "paths to the LUN was atleast the difference of number of steady" .
                                         "state paths ($num_steady_paths) and number of ports disabled " .
                                         "($num_ports_disabled)";
                    } else {
                        $error_string .= "paths to the LUN was at least 1/2 the number as in steady state " .
                                         "($num_steady_paths) for (7mode FCP, Cmode FCP & iSCSI) or\n" .
                                         "the same number of paths as steady state for (7mode iSCSI)";
                    }
                    $self->sanlog(-msg => $error_string,
                                  -msg_level => 3);
                }

                my $num_curr_paths_alive = $curr_paths->{$key}->{num_paths_alive};
                my $num_steady_paths_alive = $steady_paths->{$key}->{num_paths_alive};

                if ($num_ports_disabled > 0) {
                    # In case of port blocks, expected paths should be exactly equal to
                    # number of steady state alive paths minus number of ports disabled
                    $expected_paths = $num_steady_paths_alive - $num_ports_disabled;
                } else {
                    # calculate number of expected alive paths, don't expect 0 paths
                    # (iscsi case mostly), expect at least 1.
                    if ($protocol =~ /iscsi/i and $rule_type =~ /7mode/i) {
                        # 7mode iSCSI configs have target port failover so expect no change in steady paths
                        $expected_paths = $num_steady_paths_alive;
                    } else {
                        # Cmode configs (iscsi/fcp) and 7mode configs (fcp) do not have target
                        # port failover so expect at least half the number of paths during takeover.
                        $expected_paths = int($num_steady_paths_alive / 2);
                        $expected_paths = 1 if ($expected_paths == 0);
                    }
                }

                if ($num_curr_paths_alive < $expected_paths) {
                    $error = 1;
                    $errstr .= "The number of alive paths to the LUN ($key) is not correct\n" .
                               "Steady State  : $steady_paths->{$key}->{num_paths_alive}\n" .
                               "Current State : $curr_paths->{$key}->{num_paths_alive}\n";
                    if ($num_ports_disabled > 0) {
                        $errstr .= "Number of ports disabled: $num_ports_disabled\n" .
                                   "In case of port blocks, you should have exactly the number of alive paths which is " .
                                   "the difference of number of steady state alive paths and number of ports disabled\n\n";
                    } else {
                        $errstr .= "In takeover state you should have at least 1/2 the number\n" .
                                   "of paths as were recorded in steady state for (7mode FCP, Cmode FCP & iSCSI) or\n" .
                                   "the same number of paths as were recorded in steady state for (7mode iSCSI)\n\n";
                    }
                } else {
                    my $error_string = "Path check ($key) (takeover state) passsed - alive_paths: $num_curr_paths_alive\n";
                    if ($num_ports_disabled > 0) {
                        $error_string .= "paths to the LUN was atleast the difference of number of steady" .
                                         "state paths ($num_steady_paths_alive) and number of ports disabled " .
                                         "($num_ports_disabled)";
                    } else {
                        $error_string .= "paths to the LUN was at least 1/2 the number as in steady state " .
                                         "($num_steady_paths_alive) for (7mode FCP, Cmode FCP & iSCSI) or\n" .
                                         "the same number of paths as steady state for (7mode iSCSI)";
                    }
                    $self->sanlog(-msg => $error_string,
                                  -msg_level => 3);
                }
            }
        }
    }

    # print out error information
    if ($error == 1) {
        $self->_set_err("Path validation failed for client ($client_name)\nDetails:\n\n$errstr");
         return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: validate_initial_paths
# Objective:       Verify the the number of paths available to lun against the
#                  expected paths available to lun(passed by user)
# Details:         See POD documentation at the end of this file
################################################################################
sub validate_initial_paths {
    my ($self, %args) = @_;

    my ($client_name, $num_expected_paths);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?client_name$/) {
            if (not length($client_name = $args{$_})) {
                $self->_set_err("Invalid value for \"-client_name\":", "$args{$_}");
                return $self;
            }
            next;
        } elsif (/-?num_expected_paths$/) {
            if (not length($num_expected_paths = $args{$_})) {
                $self->_set_err("Invalid value for \"-num_expected_paths\":", "$args{$_}");
                return $self;
            }
            next;
        }
    }

    # get the path informaton from the system
    my $path_info = $self->get_path_info(-validate_mode => 0);
    return $ClientUtil::FALSE if ($self->error);

    my $error = 0;
    my $errstr = "";

    # Before starting the faults(during client validation), verify whether the paths are
    # configured properly against the expected paths(passed by user)
    foreach my $dev (keys (%{$path_info})) {
        my $num_alive_paths = $path_info->{$dev}->{num_paths_alive};
        my $num_dead_paths = $path_info->{$dev}->{num_paths_dead};
        my $num_available_paths = $path_info->{$dev}->{num_paths};
        if ($num_dead_paths == 0) {
            if ($num_alive_paths != $num_expected_paths) {
                $error = 1;
                $errstr .= "\tActual number of alive paths available for device ($dev) doesn't match " .
                           "with number of expected paths.\n" .
                           "\t\tNumber of Expected paths: $num_expected_paths\n" .
                           "\t\tNumber of Available paths: $num_available_paths\n" .
                           "\t\tNumber of Alive paths: $num_alive_paths\n" .
                           "\t\tNumber of Dead paths: $num_dead_paths\n\n";
            }
        } else {
            $error = 1;
            $errstr .= "\tDevice ($dev) has dead paths.\n" .
                       "\t\tNumber of Available paths: $num_available_paths\n" .
                       "\t\tNumber of Alive paths: $num_alive_paths\n" .
                       "\t\tNumber of Dead paths: $num_dead_paths\n\n";
        }
    }

    # print out error information
    if ($error == 1) {
        $self->_set_err("Initial path validation failed for client ($client_name)\nDetails:\n$errstr");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: dump_path_info
# Objective:       Dump path information to log
# Details:         See POD documentation at the end of this file
################################################################################
sub dump_path_info {
    my ($self) = @_;

    my $path_info = $self->{path_info};

    foreach my $key (keys %{$path_info}) {
        my $dev = $path_info->{$key};
        $self->sanlog(-msg =>
              "Multi-Pathing Path Information for Device $key:\n" .
              "LUN protocol              = $dev->{protocol}\n" .
              "Number of Paths           = $dev->{num_paths}\n" .
              "Number of Alive Paths     = $dev->{num_paths_alive}\n" .
              "Number of Dead Paths      = $dev->{num_paths_dead}\n" .
              "Number of Primary Paths   = $dev->{num_paths_primary}\n" .
              "Number of Secondary Paths = $dev->{num_paths_secondary}\n");
    }

}

################################################################################
# Subroutine Name: load_path_info
# Objective:       calls get_path_info and loads that data into
#                  this object. Will be used to initially load config before
#                  fault injection starts
# Details:         See POD documentation at the end of this file
################################################################################
sub load_path_info {
    my ($self) = @_;

    # get the path informaton from the system
    my $path_info = $self->get_path_info(-validate_mode => 0);
    return $ClientUtil::FALSE if ($self->error);

    # Store the pathing information in the object
    $self->{path_info} = $path_info;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: get_steady_state
# Objective:       returns the steady state array representing path state info
# Details:         See POD documentation at the end of this file
################################################################################
sub get_steady_state {
    my ($self) = @_;

    return ($self->{path_info});
}

################################################################################
# Subroutine Name: set_steady_state
# Objective:       sets the steady state array representing path state info
# Details:         See POD documentation at the end of this file
################################################################################
sub set_steady_state {
    my ($self, $steady_state) = @_;

    # store the steady state value in path_info
    $self->{path_info} = $steady_state;

    return $ClientUtil::TRUE;
}

1;
