# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools
## @author thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::AixMP::DMP;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::AixMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       enable DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output,);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::DMP::enable() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "/usr/sbin/vxdctl enable";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixMP::DMP::enable() Error while enabling vxdctl (" . @output . ")");
        return $ClientUtil::FALSE;
    }

    #### Verify that vxdctl daemon is running
    $self->_is_vxdctl_running(-conn => $conn);
    if ($self->error) {
        $self->_set_err("Error: AixMP::DMP::enable() _is_vxdctl_running reported error");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       disable DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output,);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::DMP::disable() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "/usr/sbin/vxdctl disable";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixMP::DMP::disable() Error while disabling vxdctl");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify()
# Objective:       verify proper enviroment for DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    my (@pkgs, $conn, $status, $cmd, @output,);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::DMP::verify() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### Verufy that package for DMP is installed (
    foreach my $pkg ("VRTSvxvm", "VRTSvlic") {
        $cmd = "lslpp -l | grep $pkg | awk '{print \$2}'";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Error: AixMP::DMP::verify() Failed to verify packaged ($pkg) is installed");
            return $ClientUtil::FALSE;
        } else {
            $self->sanlog(-msg => "AixMP::DMP::verify() Verified that package ($pkg) is installed",
                          -msg_level => 9);
        }
    }

    ### Verify that the DMP module is loaded on the client
    $cmd = "ps -ef | grep -i [v]xconfig";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status != 0) {
        $self->_set_err("Error: AixMP::DMP::verify() Failed to verify that DMP daemon is running (" . @output . ")");
        return $ClientUtil::FALSE;
    }

    # Verify that the ASL/APM is installed (in case of versions < 5.1).
    # Case of 5.1+, support is in VRTS package itself.
    if ($self->{version} >= 5.1) {
        @pkgs = ("VRTSaslapm");
    } elsif ($self->{version} =~ /^5/) {
        @pkgs = ("VRTSNTAPasl", "VRTSNTAPapm");
    } else {
        @pkgs = ("NTAPasl");
    }

    foreach my $pkg (@pkgs) {
        $cmd = "lslpp -l | grep $pkg";
        ($status, @output) = &ClientUtil::execute($conn , $cmd);
        if ($status != 0) {
            $self->_set_err("Error: AixMP::DMP::verify() Failed to verify packaged ($pkg) is installed");
            return $ClientUtil::FALSE;
        }
    }

    # Now if this is 5.1, verify that the older requirements/packages
    # required with 5.0 are NOT installed.
    if ($self->{version} >= 5.1) {
        @pkgs = ("VRTSNTAPasl", "VRTSNTAPapm");
        foreach my $pkg (@pkgs) {
            $cmd = "lslpp -l | grep $pkg";
            ($status, @output) = &ClientUtil::execute($conn, $cmd);
            if ($status == 0) {
                $self->_set_err("Error: AixMP::DMP::verify() Failed to verify packaged ($pkg) is NOT installed\n" .
                                "VRTSNTAPasl and VRTSNTAPapm should NOT be installed with VRTS 5.1 and greater");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       setup proper enviroment for DMP daemon. No-op for now on AIX
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    return $self;
}

################################################################################
# Subroutine Name: config()
# Objective:       verify proper enviroment for DMP daemon. No-op for now on AIX
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    return $self;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # no-op for DMP on AIX

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each DMP node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($per_lun_hash, $cmd, @vx_output, @output, $status, $conn);
    my ($line, $enclosure, @encl_output);
    my ($dmp_node, $state, $num_paths, $num_alive, $num_dead);
    my (@lun_show_output, @vxdisk_output);
    my ($protocol, $device);
    my ($validate_mode);

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Error: AixMP::DMP::get_path_info() Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # get HU path for path's to binaries
    my $hu_path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();

    $per_lun_hash = {};
    @output = ();

    # get connection from object
    $conn = $self->{conn};

    # run "vxdmpadm listenclosure all" to get list of enclosures
    $cmd = "/usr/sbin/vxdmpadm listenclosure all | grep -v 'DISCONNECTED' | grep 'CONNECTED' | grep -v 'DISKS'";
    ($status, @vx_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                       180, 5, "-warn");
    if ($status != 0) {
        $self->_set_err("Error: AixMP::DMP::get_path_info() Failed to run vxdmpadm listenclosure all");
        return {};
    }

    foreach $line (@vx_output) {
        $enclosure = (split(/\s+/, $line))[0];
        if ($enclosure ne "") {
            $cmd = "/usr/sbin/vxdmpadm getdmpnode enclosure=$enclosure";
            ($status, @encl_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                 180, 5, "-comment");
            if ($status != 0) {
                $self->_set_err("Error: AixMP::DMP::get_path_info() Failed to run $cmd to get dmpnode and path information");
                return {};
            }
            push(@output, @encl_output);
        }
    }

    # only grab sanlun to get protocol if we are in validation mode
    # meaning we are checking paths. We will get the protocol only
    # in steady state, since sanlun is very slow in takover mode
    if ($validate_mode == 0) {
        # need normal sanlun output to get protocol
        $cmd = $hu_path . "/sanlun lun show";
        ($status, @lun_show_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                 360, 6, "-warn");
        if ($status != 0) {
            $self->_set_err("Error: AixMP::DMPMP::get_path_info() Error while calling $cmd");
            return {};
        }

        # need vxdisk -e list output to map the enclosure based name
        $cmd = "/usr/sbin/vxdisk -e list";
        ($status, @vxdisk_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                               180, 6, "-warn");
        if ($status != 0) {
            $self->_set_err("Error: AixMP::DMPMP::get_path_info() Error while calling $cmd");
            return {};
        }
    }

    # go over the enclosure output and pick apart each line (which
    # will be the dmp_node followed by it's characteristics
    # Example:
    #  NAME                 STATE     ENCLR-TYPE   PATHS  ENBL  DSBL  ENCLR-NAME
    # =========================================================================
    # c5t0d0s2             ENABLED      FAS960       4      4     0     FAS9600
    foreach $line (@output) {
        # skip header and newlines
        if ($line =~ /ENCLR-TYPE/i or
            $line =~ /=====/i or
            $line =~ /^\s*$/) {
            next;
        }

        # get per dmp node pathing information from output
        ($dmp_node, $state, undef,
         $num_paths, $num_alive, $num_dead, undef) = split(/\s+/, $line);

        # store number of paths
        $per_lun_hash->{$dmp_node}->{num_paths} = $num_paths;

        # store number of paths that are enabled (alive)
        $per_lun_hash->{$dmp_node}->{num_paths_alive} = $num_alive;

        # store number of paths that are disabled (dead)
        $per_lun_hash->{$dmp_node}->{num_paths_dead} = $num_dead;

        # store number of paths that are primary
        $per_lun_hash->{$dmp_node}->{num_paths_primary} = "NA";

        # store number of paths that are secondary
        $per_lun_hash->{$dmp_node}->{num_paths_secondary} = "NA";

        # only set protocol in validation mode (comparing paths)
        if ($validate_mode == 0) {
            # get protocol from the sanlun output
            my ($protocol);
            # lookup the device handle from getdmp node in
            # vxdisk -e list ouput to get the actual device
            my $vxdisk_line = (grep(/$dmp_node/, @vxdisk_output))[0];
            my $device = (split(/\s+/, $vxdisk_line))[5];

            # lookup that device in sanlun output to get protocol
            my $sanlun_dev = (grep(/$device/, @lun_show_output))[0];
            if ($sanlun_dev =~ /\s+iscsi\s+/i) {
                $protocol = "iSCSI";
            } else {
                $protocol = "FCP";
            }

            $per_lun_hash->{$dmp_node}->{protocol} = $protocol;
        } else {
            $per_lun_hash->{$dmp_node}->{protocol} = "takeover";
        }

    }

    return $per_lun_hash;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $parent, $obj_template);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument \'-parent\'");
                return $self;
            }
            next;
                } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        # fix the parent link and return
        $self->{parent} = $parent;
        return $self;
    }

    $self->{parent} = $parent;

    $self->_get_pkginfo(-conn => $self->{parent}->{parent}->{parent}->{conn});

    # Returning DMP object
    return $self;
}

################################################################################
# Subroutine Name: _get_pkginfo()
# Objective:       returns the version of Veritas running on the system
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, $conn);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return ("", "");
            }
            next;
        }
    }

    my $vers = "";
    my $sp   = "";
    my $rp   = "";
    my $mp   = "";

    $cmd = "lslpp -l | grep VRTSvxvm | awk '{print \$2}'|sort -u";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to execute command $cmd");
        return $ClientUtil::FALSE;
    }

    #================================================
    # format of lslpp -l VRTSvxvm returns:
    #
    # 6.0:
    #
    # 6.0.0.0
    #
    # 5.1SP1RP1:
    #
    # 5.1.101.0
    #
    # 5.1SP1:
    #
    # 5.1.100.0
    #
    # 5.1RP2:
    #
    # 5.1.2.0
    #
    # 5.0MP3RP2:
    #
    # 5.0.3.200
    #
    # 5.0MP3:
    #
    # 5.0.3.0
    #================================================
    foreach my $line (@output) {
        my ($d1, $d2, $d3, $d4) = split(/\./, $line);

        # The version is the first 2 put together
        $vers = $d1 . "." . $d2;

        # Now is where things get kludgy.  We now have to use the output from
        # lspp above to determin the sp, mp, rp
        if ($vers >= 5.1) {
            if ($d3 eq "100") {
                $sp = "SP1";
            } elsif ($d3 eq "101") {
                $sp = "SP1";
                $rp = "RP1";
            } elsif ($d3 eq "2") {
                $rp = "RP2";
            }
        } elsif ($vers eq 5.0) {
            if ($d3 eq "3") {
                $mp = "MP3";
                if ($d4 eq "200") {
                    $rp = "RP2";
                }
            } else {
                # Default to MP1 for all 5.0 versions except MP3
                $mp = "MP1";
            }
        }
    }

    $self->{name}     = "VRTSvxvm";
    $self->{version}  = $vers;
    $self->{revision} = $vers;
    $self->{mp}       = $mp;
    $self->{rp}       = $rp;
    $self->{sp}       = $sp;
    $self->{pkginfo}  = \@output;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _is_vxdctl_running()
# Objective:       returns ClientUtil::TRUE or ClientUtil::FALSE depending on
#                  if vxdctl is running or not.
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_vxdctl_running {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::DMP::_is_vxdctl_runing() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check status of vxdctl service
    my $cmd = "/usr/sbin/vxdctl mode";
    my ($status, @output) = &ClientUtil::execute($conn, $cmd);

    if (grep /enable/i, @output) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: _rescan
# Objective:       To run the vxdctl command to rescan vxvm's list of
#                  disks that it thinks are under it's control.
# Details:         See POD documentation at the end of this file
################################################################################
sub _rescan {
    my ($self, %args) = @_;

    my ($conn, $cmd, @output, $status, $host_obj,);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::DMP::_rescan() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "/usr/sbin/vxdctl enable";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: AixMP::DMP::rescan() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _dmp_version
# Objective:       To version and revision of VXVM package installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _dmp_version () {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, $version_string, $conn, $vers);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::DMP::_dmp_version() Invalid value for argument '-conn'");
                return ("", "");
            }
            next;
        }
    }

    $self->_clear_err();

    $cmd = "lslpp -l | grep VRTSvxvm | awk '{print \$2}'";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: Aix::Vxvm::_get_pkginfo() Command ($cmd) returned a bad status ($status)");
        return ("", "");
    }

    $vers = $output[0];

    return $vers, $vers;

}

