# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools for AIX MPIO
## @author thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::AixMP::MPIO;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::AixMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       enable MPIO daemon - nothing to do for AIX/MPIO
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::MPIO::enable() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: disable()
# Objective:       disable MPIO daemon. Nothing to do for AIX/MPIO.
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::MPIO::disable() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: verify()
# Objective:       MPIO is built into the device driver stack. There is no
#                  validation available that the AIX teams knows about (or is
#                  need to check). Will check for the NetApp ODM definitions for
#                  MPIO that are needed for this to function.
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output, @packages);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::MPIO::verify() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # get HU path for path's to binaries
    my $oem_rebrand = $self->{parent}->{parent}->{parent}->{general}->{OEM_REBRAND};
    if ($oem_rebrand =~ /ibm/i) {
        # List of IBM (OEM rebranded) packages for AIX
        @packages = ("ontap.mpio.*config",
                     "ontap.*fcp",
                     "ontap.*iscsi",
                     "ontap.*pcmodm");
    } else {
        # List of NetApp packages for AIX (note, FCP and iSCSI delivered in same HU).
        @packages = ("netapp.mpio.*config",
                     "netapp.*fcp",
                     "netapp.*iscsi",
                     "netapp.*pcmodm");
    }

    ### Verify that package for MPIO is installed
    foreach my $pkg (@packages) {
        $cmd = "lslpp -l | egrep -i \"$pkg\"";
        ($status, @output) = ClientUtil::execute($conn , $cmd);

        if ($status != 0) {
            $self->_set_err("Error: AixMP::MPIO::verify() Failed to verify packaged ($pkg) is installed");
            return $ClientUtil::FALSE;
        } else {
            $self->{parent}->{parent}->{parent}->sanlog(-msg => "AixMP::MPIO::verify() Verified that package ($pkg) is installed",
                                                        -msg_level => 9);
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: setup()
# Objective:       setup proper enviroment for MPIO daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::MPIO::setup() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return ($self);

}

################################################################################
# Subroutine Name: config()
# Objective:       verify proper enviroment for MPIO daemon. Nothing needed
#                  for AIX/MPIO
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::MPIO::config() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return ($self);

}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($cmd, @odm_output, @disk_output, @output, $status, $conn);
    my ($device, $details, @attr_output, @sanlun_output, @spt_output);
    my ($weight, $num_proxy, $num_primary);
    my (@lun_show_output, $protocol);
    my ($validate_mode);

    # parse arguments
    $details = 0;
    foreach (keys %args) {
       if (/-?details/) {
            if (not length($details = $args{$_})) {
                $self->_set_err("Error: AixMP::MPIOMP::get_path_info() Invalid value for argument \'-details\'");
                return $ClientUtil::FALSE;
            }
            next;
       } elsif (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Error: AixMP::MPIOMP::get_path_info() Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
       }
    }

    # get HU path for path's to binaries
    my $hu_path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();

    my $per_lun_hash = {};

    # get connection from object
    $conn = $self->{conn};

    # get boot device to exclude from check
    $cmd = "/usr/sbin/bootinfo -b";
    ($status, @output) = &ClientUtil::execute_retry($conn, $cmd,
                                                    180, 5, "-warn");
    if ($status != 0) {
        $self->_set_err("Error: AixMP::MPIO::get_path_info() Command ($cmd) returned a bad status ($status)");
    }
    my $bootdevice = $output[0];

    # get list of MPIO NetApp devices
    $cmd = "/usr/sbin/lsdev -C -c disk | egrep \"\(MPIO)\|\(Virtual\)\" | egrep -i \"\(NetApp\)\|\(ONTAP\)|\(Virtual\)\" | awk '{print \$1;}'";
    ($status, @disk_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                         180, 5, "-warn");
    if ($status != 0) {
         $self->_set_err("Error: AixMP::MPIO::get_path_info() Failed to run lsdev to discover devices");
         return {};
    }

    foreach $device (@disk_output) {
        # exclude boot device
        if ($device eq $bootdevice) {
            next;
        }

        # get the list of subpaths to the LUN
        $cmd = "/usr/sbin/lspath -l $device -F 'path_id name parent connection status' | grep hdisk";
        ($status, @output) = &ClientUtil::execute_retry($conn, $cmd,
                                                        300, 5, "-warn");
        if ($status != 0) {
            $self->_set_err("Error: AixMP::MPIO::get_path_info() Failed to run lspath info on $device");
            return {};
        }

        # store number of paths
        my @filtered = grep(/$device/, @output);
        my $num_paths = scalar(@filtered);
        $per_lun_hash->{$device}->{num_paths} = $num_paths;

        # store number of paths that are enabled (alive)
        my $num_alive = scalar(grep(/Enabled/i, @output));
        $per_lun_hash->{$device}->{num_paths_alive} = $num_alive;

        # store number of paths that are disabled (dead)
        my $num_dead = scalar(grep(/Failed/i, @output));
        $per_lun_hash->{$device}->{num_paths_dead} = $num_dead;

        # store number of paths that are primary
        $per_lun_hash->{$device}->{num_paths_primary} = "NA";

        # store number of paths that are secondary
        $per_lun_hash->{$device}->{num_paths_secondary} = "NA";

        # lookup that device path information and store the
        # type of mapped (iSCSI, FCP) based on the lspath information
        if (grep(/iqn/i, @output) > 0) {
            $protocol = "iSCSI";
        } else {
            $protocol = "FCP";
        }

        # set protocol
        $per_lun_hash->{$device}->{protocol} = $protocol;

        # generate additional device/path details if requested.
        if ($details == 1) {
            # get details from sanlun on this device
            my $path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();
            $cmd = $path . "/sanlun lun show -v -d $device";
            ($status, @sanlun_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                   300, 5, "-warn");
            if ($status != 0) {
                $self->_set_err("Error: AixMP::MPIO::get_path_info() Could not run sanlun to device $device");
                return {};
            }

            # pick out serial number
            my $serial_number = (grep(/Serial number:/i, @sanlun_output))[0];
            $serial_number =~ s/\s+Serial number://i;

            # get controller/lunpath information
            my $controller_info = $sanlun_output[1];
            $controller_info =~ s/^\s+//;
            my ($controller, $controller_path, undef, undef,
                $protocol, undef, $lun_size, undef) = split(/\s+/,  $controller_info);
            $lun_size =~ s/\(|\)//g;
            $controller =~ s/://;

            # get LUN (ID) - cannot get via sanlun standard
            $cmd = "/usr/sbin/lsattr -El $device -a \'lun_id\'";
            ($status, @attr_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                 300, 5, "-warn");
            if ($status != 0) {
                $self->_set_err("Error: AixMP::MPIO::get_path_info() Could not run lsattr to device $device");
                return {};
            }
            my $lun_id = $attr_output[0];
            $lun_id = (split(/\s+/, $lun_id))[1];

            $num_proxy = 0;
            $num_primary = 0;
            # pull information (additional) from lspath output
            foreach my $path_details (@output) {
                my ($path_id, undef, $hba, $target_info, $state, $lun_state) =
                   split(/\s+/, $path_details);

                # remove not needed info from target
                $target_info =~ s/\,.*//;

                # set device (non-path) specific data
                $per_lun_hash->{$device}->{controller} = $controller;
                $per_lun_hash->{$device}->{controller_path} = $controller_path;
                $per_lun_hash->{$device}->{lun_id} = $lun_id;
                $per_lun_hash->{$device}->{lun_state} = $lun_state;
                $per_lun_hash->{$device}->{lun_size} = $lun_size;
                $per_lun_hash->{$device}->{serial_num} = $serial_number;

                # run spt to get vdisk information on the individual path.
                # VTIC paths will show up with v. as there representation
                my $test_bin = $self->{parent}->{parent}->{parent}->{TEST_BIN};
                $cmd = "$test_bin/spt dsf=/dev/$device cdb=\"c0 00 02 0a 98 0a 21 00 ff 00\" dir=read length=255 path=$path_id";
                ($status, @spt_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                    300, 5, "-warn");
                if ($status != 0) {
                    $self->_set_err("Error: AixMP::MPIO::get_path_info() Could not run spt to device $device and path_id ($path_id)");
                    return {};
                }

                if (grep(/v\./, @spt_output) > 0) {
                    $per_lun_hash->{$device}->{per_path_hash}->{$path_id}->{is_proxy} = "True";
                    $num_proxy++;
                } else {
                    $per_lun_hash->{$device}->{per_path_hash}->{$path_id}->{is_proxy} = "False";
                    $num_primary++;
                }

                # get from ODM the weight on this particular path.
                $cmd = "/usr/bin/odmget -q \"name=$device and attribute=priority and path_id=$path_id\" CuPathAt";
                ($status, @odm_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                    300, 5, "-warn");
                if ($status != 0) {
                    $self->_set_err("Error: AixMP::MPIO::get_path_info() Could not odmget device $device and path_id ($path_id)");
                    return {};
                }

                # No return from ODM = path weight = 1, otherwise it's in the value line
                if (scalar(@odm_output) == 0) {
                    $weight = 1;
                } else {
                    $weight = (grep(/value \=/, @odm_output))[0];
                    $weight =~ s/\s+value\s+\=\s+//;
                    $weight =~ s/\"//g;
                }

                $per_lun_hash->{$device}->{per_path_hash}->{$path_id}->{init_port} = $hba;
                $per_lun_hash->{$device}->{per_path_hash}->{$path_id}->{target_port} = $target_info;
                $per_lun_hash->{$device}->{per_path_hash}->{$path_id}->{lb_policy} = "round_robin";
                $per_lun_hash->{$device}->{per_path_hash}->{$path_id}->{reserve} = "NA";
                $per_lun_hash->{$device}->{per_path_hash}->{$path_id}->{weight} = $weight;
                $per_lun_hash->{$device}->{per_path_hash}->{$path_id}->{dsm_id} = "NA";
                $per_lun_hash->{$device}->{per_path_hash}->{$path_id}->{state} = $state;
                $per_lun_hash->{$device}->{per_path_hash}->{$path_id}->{protocol} = $protocol;
            }
            # store number of paths that are primary
            $per_lun_hash->{$device}->{num_paths_primary} = $num_primary;

            # store number of paths that are secondary/proxy
            $per_lun_hash->{$device}->{num_paths_secondary} = $num_proxy;
        }
    }

    return $per_lun_hash;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $parent, $obj_template);

    $self = {};

    bless ($self, $class);

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument \'-parent\'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        # fix the parent link and return
        $self->{parent} = $parent;
        return $self;
    }

    $self->{parent} = $parent;

    ($self->{version}, $self->{revision}) =
              $self->_mpio_version(-conn =>
                                   $self->{parent}->{parent}->{parent}->{conn});

    return ($self);

}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # No-op for MPIO on AIX

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _mpio_version
# Objective:       MPIO is built-in to the device driver stack. It does not have
#                  a version number per-say. Recommendation is to go with
#                  keyword "native" (for both version and revision)
# Details:         See POD documentation at the end of this file
################################################################################
sub _mpio_version {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, $conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::MPIO::_mpio_version() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->_clear_err();

    return ("native", "native");

}

1;
