# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools for AIX SanPath (DotHill)
## @author thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::AixMP::SANPATH;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::AixMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       enable SANPATH daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::SANPATH::enable() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: disable()
# Objective:       disable SANPATH daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::SANPATH::disable() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: verify()
# Objective:       verify proper enviroment for SANPATH daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::SANPATH::verify() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### Verufy that package for SANPATH is installed (
    foreach my $pkg ("DotHill.SANpath") {
        $cmd = "lslpp -l | grep $pkg";
        ($status, @output) = &ClientUtil::execute($conn , $cmd);
        if ($status != 0) {
            $self->_set_err("Error: AixMP::SANPATH::verify() Failed to verify packaged ($pkg) is installed");
            return $ClientUtil::FALSE;
        } else {
            $self->sanlog(-msg => "AixMP::SANPATH::verify() Verified that package ($pkg) is installed",
                      -msg_level => 9);
        }
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # no-op for SANPATH on AIX

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       setup proper enviroment for SANPATH daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::SANPATH::setup() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return $self;

}

################################################################################
# Subroutine Name: config()
# Objective:       verify proper enviroment for SANPATH daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::SANPATH::config() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return $self;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($cmd, @disk_output, @output, $status, $conn);
    my ($device);
    my ($validate_mode);

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Error: AixMP::SANPATH::get_path_info() Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my $per_lun_hash = {};

    # get connection from object
    $conn = $self->{conn};

    # get list of SANPATH spd devices
    $cmd = "/usr/sbin/setsp | egrep 'spd[0-999]' | grep 'NETAPP' | awk '{print \$1;}'";
    ($status, @disk_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                         180, 5, "-warn");
    if ($status != 0) {
         $self->_set_err("Error: AixMP::SANPATH::get_path_info() Failed to run lsdev to discover devices");
         return {};
    }

    foreach $device (@disk_output) {
        # get the list of subpaths to the LUN
        $device =~ s/spd//;
        $cmd = "/usr/sbin/setsp -v -l $device | grep hdisk | grep -v \"ID =\"";
        ($status, @output) = &ClientUtil::execute_retry($conn, $cmd,
                                                        180, 5, "-warn");
        if ($status != 0) {
            $self->_set_err("Error: AixMP::SANPATH::get_path_info() Failed to run lspath info on $device");
            return {};
        }

        # store number of paths
        my $num_paths = scalar(@output);
        $per_lun_hash->{$device}->{num_paths} = $num_paths;

        # store number of paths that are enabled (alive)
        my $num_alive = scalar(grep(/Good/i, @output));
        $per_lun_hash->{$device}->{num_paths_alive} = $num_alive;

        # store number of paths that are disabled (dead)
        my $num_dead = scalar(grep(/Bad/i, @output));
        $per_lun_hash->{$device}->{num_paths_dead} = $num_dead;

        # store number of paths that are primary
        my $num_primary = scalar(grep(/\s+P\s+/, @output));
        $per_lun_hash->{$device}->{num_paths_primary} = $num_primary;

        # store number of paths that are secondary
        my $num_secondary = scalar(grep(/\s+S\s+/, @output));
        $per_lun_hash->{$device}->{num_paths_secondary} = $num_secondary;

        # SANPATH is only supported with FCP
        $per_lun_hash->{$device}->{protocol} = "FCP";
    }

    return $per_lun_hash;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $parent, $obj_template);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument \'-parent\'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        # fix the parent link and return
        $self->{parent} = $parent;
        return $self;
    }

    $self->{parent} = $parent;

    ($self->{version}, $self->{revision}) =
           $self->_sanpath_version(-conn =>
                                   $self->{parent}->{parent}->{parent}->{conn});

    return $self;

}

################################################################################
# Subroutine Name: _sanpath_version
# Objective:       To version and revision of VXVM package installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _sanpath_version {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, $conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: AixMP::SANPATH::_sanpath_version() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->_clear_err();

    $cmd = "lslpp -l DotHill.SANpath | grep DotHill.SANpath";
    ($status, @output) = &ClientUtil::execute($conn , $cmd);

    if ($status > 0) {
        $self->_set_err("Error: AixMP::SANPATH::_sanpath_version() Failed to execute command $cmd");
        return ("", "");
    }

    my $ver = $output[0];
    $ver =~ s/^\s+//;
    $ver = (split(/\s+/, $ver))[1];

    return $ver, $ver;

}

1;
