# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools
## @author hrivera@netapp.com, thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::ESXMP::NativeMP;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::ESXMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Error: ESXMP::NativeMP.pm::enable() Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    $self->_clear_err();
    $self->_set_err("Error: ESXMP::NativeMP.pm::disable() Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: verify()
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Error: ESXMP::NativeMP.pm::verify() Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # no-op for ESX

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Error: ESXMP::NativeMP.pm::setup() Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: config()
# Objective:       NOOP For ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Error: ESXMP::NativeMP.pm::config() Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: config_paths()
# Objective:       This will activate all primary paths and set up load
#                  balancing.
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self, %args) = @_;

    $self->_clear_err();

    # set up primary/loadbalanced/persistent paths for all multipathed vmhbas
    # need to distinguish between netapp|ontap
    my $path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();
    my $cmd = $path . "/config_mpath --primary --loadbalance --persistent";
    my ($status, @output) = &ClientUtil::execute($self->{parent}->{parent}->{parent}->{conn},
                                                 $cmd);
    if ($status > 0 and not grep(/No FC Luns were found/i, @output)) {
        # sometimes the error from config_mpath complains about dns errors.
        # dont set an error in this case.
        # if it was a DNS error, bomb out. Paths may not get set correctly in non alua configs
        if (grep(/Your DNS is improperly configured/i, @output)) {
            $self->_set_err("cmd ($cmd) returned status ($status). Either: " .
                            "1) DNS is not set up on this host or " .
                            "2) your NetApp controllers are not in DNS." .
                            "Please update your DNS or modify your /etc/hosts " .
                            "file to map the IP addresses of your NetApp " .
                            "controllers to their respective hostnames.");
            return $ClientUtil::FALSE;
        } else {
            # otherwise it was an unknown error.
            $self->_set_err("cmd ($cmd) returned status ($status). " .
                            "Check and make sure your EHU is up-to-date.");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: set_psp_policy
# Objective: NOOP
################################################################################
sub set_psp_policy {
    my ($self) = @_;

    # NOOP for ESX versions earlier than 4.0

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($num_alive, $num_dead, $num_paths, $device);
    my ($cmd, $status, $conn, $path);
    my ($validate_mode);
    my @all_devs = ();
    my @adapters = ();
    my @lun_output = ();
    my @dev_output = ();

    my $per_lun_hash = {};

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Error: ESXMP::NativeMP::get_path_info() Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # get connection from object
    $conn = $self->{conn};

    # based on protocol and type of iscsi initiator get the right array
    # of adapters to look for.
    my $client_obj = $self->{parent}->{parent}->{parent};
    @adapters = @{$client_obj->{vmhba_adapter}};

    # Execute esxcfg-vmhbadevs to get all devices and their
    # /dev/sd* path (single device here)
    foreach my $adapter (@adapters) {
        $cmd = "esxcfg-vmhbadevs | grep $adapter:";
        ($status, @dev_output) = &ClientUtil::execute($conn, $cmd);
        if (@dev_output) {
           push(@all_devs,  @dev_output);
        }
    }

    # check to make sure we have discover some devices to work with
    # there should be devices by the time this code is run.
    if (!@all_devs) {
        $self->_set_err("Error: ESXMP::NativeMP::get_path_info() No devices " .
                        "were found from the provided list of adapters (@adapters)");
        return {};
    }

    # traverse output and generate per LUN hash of pathing information
    foreach my $line (@all_devs) {
        $num_alive = $num_dead = $num_paths = 0;

        # We are going to run esxcfg-mpath on each device in the list
        # and then pull the information about pathing out from that
        # output.
        my ($lun, $sd_path) = split(/\s+/, $line);

        # skip if lun is not in the correct format
        next if ($lun !~ /^\s*vmhba\d+:\d+:\d+\s*$/);

        $cmd = "esxcfg-mpath -q --lun=$lun";
        ($status, @lun_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                            180, 5, "-warn");
        if ($status != 0) {
            $self->_set_err("Error: ESXMP::NativeMP::get_path_info() Command ($cmd) " .
                            " returned status ($status).");
             return {};
        }

        # Skip SAS controllers
        if (grep(/SAS/, @lun_output)) {
            next;
        }

        # traverse all the paths to the LUN
        foreach my $path (@lun_output) {
            # remove blank lines
            next if ($line =~ /^\s*$/);

            if ($path =~ /^Disk vmhba/i) {
                $num_paths = (split(/\s+/, $path))[5];
            }

            # get alive/dead paths
            # must anchor dead because it can appear in the wwpn...
            # and cause a false positive
            if ($path =~ /\sDead/i or $path =~ /N\/A/i) {
                $num_dead++;
            } elsif ($path =~ /\s+On\s+/i or
                     $path =~ /\s+On$/i) {
                $num_alive++;
            }
        }

        $per_lun_hash->{$lun}->{num_paths} = $num_paths;
        $per_lun_hash->{$lun}->{num_paths_alive} = $num_alive;
        $per_lun_hash->{$lun}->{num_paths_dead} = $num_dead;
        $per_lun_hash->{$lun}->{num_paths_primary} = "NA";
        $per_lun_hash->{$lun}->{num_paths_secondary} = "NA";
        $per_lun_hash->{$lun}->{protocol} = $client_obj->{protocol};
    }

    return $per_lun_hash;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $parent, $obj_template);

    $self = {};
    bless($self, $class);

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent' => $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
    }

    # plug in the parent
    $self->{parent} = $parent;

    return $self;
}

1;
