# Copyright (c) 2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools for ESX4
## @author hrivera@netapp.com, thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::ESXMP::NativeMP4;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::ESXMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    $self->_clear_err();
    $self->_set_err("Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: verify()
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # NOOP for ESX

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       NOOP for ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: config()
# Objective:       NOOP For ESX
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: config_paths()
# Objective:       This will activate all primary paths and set up load
#                  balancing.
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self, %args) = @_;

    $self->_clear_err();

    my $retry_count = 3;

    while (1) {
        # set up primary/loadbalanced/persistent paths for all multipathed vmhbas
        # need to distinguish between netapp|ontap
        my $path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();
        my $cmd = $path . "/config_mpath --primary --loadbalance";

        # from BURT #382524: if alua support is off, and protocol is fcp, set policy to be fixed
        if ($self->{parent}->{parent}->{parent}->get_alua_support() == 0 and
            $self->{parent}->{parent}->{parent}->get_protocol() =~ /fcp/i) {
            $cmd = $path . "/config_mpath --primary --policy fixed";
        }

        my ($status, @output) = &ClientUtil::execute($self->{parent}->{parent}->{parent}->{conn},
                                                     $cmd);
        if ($status > 0) {
            # if we see "The script was unable to execute the command necessary to set the policy.", then
            # config_mpath had trouble running an internal esx command, it is ok if this happens, we just need
            # to retry the command.
            if (grep(/The script was unable to execute the command necessary to set the policy/i, @output) and
                $retry_count > 0) {
                # decrease the retry count, clear error, and sleep
                $retry_count--;
                $self->sanlog(-msg => "config_mpath failed because of a known issue, will retry in 10 seconds" .
                                      " (count=$retry_count)");
                $self->_clear_err();
                &SanUtil::nate_sleep(10);
                redo;
            }

            # ignoring reboot statement. assumption is that if doing ALUA, then
            # the user should have already run config_mpath before and the default
            # psp policy should be set.
            unless (grep(/Please reboot your system for changes to take [ae]ffect/i, @output)) {
                # if it was a DNS error, bomb out. Paths may not get set correctly in non alua configs
                if (grep(/Your DNS is improperly configured/i, @output)) {
                    $self->_set_err("cmd ($cmd) returned status ($status). Either: " .
                                    "1) DNS is not set up on this host or " .
                                    "2) your NetApp controllers are not in DNS." .
                                    "Please update your DNS or modify your /etc/hosts " .
                                    "file to map the IP addresses of your NetApp " .
                                    "controllers to their respective hostnames.");
                    return $ClientUtil::FALSE;
                } elsif ($status =~ /^(2|3|4)$/) {
                    # these return codes are only warnings from config mpath and shouldnt be considered fatals
                    last;
                } else {
                    # otherwise it was an unknown error.
                    $self->_set_err("cmd ($cmd) returned status ($status). " .
                                        "Check and make sure your EHU is up-to-date.");
                    return $ClientUtil::FALSE;
                }
            }
        } else {
            # otherwise the command worked, break out of the loop
            last;
        }
    }

    return $ClientUtil::TRUE;
}


################################################################################
# Subroutine Name: set_psp_policy
# Objective:       Given a specific device, and requested path selection policy
# change the policy
################################################################################
sub set_psp_policy {
    my ($self, %args) = @_;

    my ($naa_device, $psp_policy, $native_cmd);

    $native_cmd = 0;

    # parse arguments
    foreach (keys %args) {
        if (/-?naa_device/) {
            if (not length($naa_device = $args{$_})) {
                $self->_set_err("Invalid value for argument '-naa_device'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?psp_policy/) {
            if (not length($psp_policy = $args{$_})) {
                $self->_set_err("Invalid value for argument '-psp_policy'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?native_cmd/) {
            if (not length($native_cmd = $args{$_})) {
                $self->_set_err("Invalid value for argument '-native_cmd'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not defined($naa_device)) {
        $self->_set_err("Missing required parameter -naa_device");
        return $ClientUtil::FALSE;
    }
    if (not defined($psp_policy)) {
        $self->_set_err("Missing required parameter -psp_policy");
        return $ClientUtil::FALSE;
    }

    # set the psp policy on this naa device using either config_mpath or the native command
    my $cmd;
    if ($native_cmd == 0) {
        my $path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();
        $cmd = $path . "/config_mpath -m -n ${naa_device} -p ${psp_policy}";
    } else {
        # convert the policy to the proper tag
        $psp_policy = "VMW_PSP_RR" if ($psp_policy eq "rr");
        $psp_policy = "VMW_PSP_MRU" if ($psp_policy eq "mru");
        $psp_policy = "VMW_PSP_FIXED" if ($psp_policy eq "fixed");
        $psp_policy = "Data_ONTAP_PSP" if ($psp_policy eq "ontap");
        my $nmp_namespace = $self->{parent}->{parent}->{parent}->{esxcli_nmp_namespace};

        if ($self->{parent}->{parent}->{parent}->get_os_version() =~ /^(5|6)\./) {
            $cmd = "esxcli ${nmp_namespace} device set --device ${naa_device} --psp \"${psp_policy}\"";
        } else {
            $cmd = "esxcli ${nmp_namespace} device setpolicy --device ${naa_device} --psp \"${psp_policy}\"";
        }
    }

    my ($status, @output) = $self->{parent}->{parent}->{parent}->_esx_execute_cmd(-cmd => $cmd,
                                                                                  -include_host_name => 1,
                                                                                  -reconnect_on_failure => 1);
    if ($status != 0) {
        $self->_set_err("Failed to set psp policy '$psp_policy' on device '$naa_device'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

sub get_path_info {
    my ($self, %args) = @_;

    my ($validate_mode);
    my $per_lun_hash = {};

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Invalid value for argument '-validate_mode'");
                return {};
            }
            next;
        }
    }

    # get path info for each 'working' adapter
    #@adapters = @{$client_obj->{vmhba_adapter}};
    my $sc_obj = $self->{parent}->{parent}->{parent};

    # if there is an MPDevMgr use it, otherwise, create a new one
    if (not defined($sc_obj->mpdevmgr())) {
        $sc_obj->create_mpdevmgr();
        if ($sc_obj->error()) {
            $self->_set_err("Failed to create MPDevMgr object",
                            $sc_obj->errmsg());
            return {};
        }
    }

    my $mpdevmgr = $sc_obj->mpdevmgr();
    my @mpdev_objs = $mpdevmgr->get_device_objs();

    # TODO: filter the mpdev objs by the @adapters array

    # check to make sure we have discovered some devices to work with
    # there should be devices by the time this code is run.
    if (not scalar(@mpdev_objs)) {
        $self->_set_err("No devices were found from the provided list of adapters");
        return {};
    }

    foreach my $mpdev_obj (@mpdev_objs) {
        my $naa_device = $mpdev_obj->mp_device_name();
        my $curr_path_state = $mpdev_obj->get_current_path_state();

        $per_lun_hash->{$naa_device}->{num_paths} = $curr_path_state->total_paths();
        $per_lun_hash->{$naa_device}->{num_paths_alive} = $curr_path_state->total_alive_paths();
        $per_lun_hash->{$naa_device}->{num_paths_dead} = $curr_path_state->total_dead_paths();
        $per_lun_hash->{$naa_device}->{num_paths_primary} = "NA";
        $per_lun_hash->{$naa_device}->{num_paths_secondary} = "NA";
        $per_lun_hash->{$naa_device}->{protocol} = $mpdev_obj->protocol();
    }

    return $per_lun_hash;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info_old {
    my ($self, %args) = @_;

    my ($num_alive, $num_dead, $num_paths, $device);
    my ($cmd, $status, $conn, $path);
    my ($validate_mode);
    my @all_devs = ();
    my @adapters = ();
    my @lun_output = ();
    my @dev_output = ();

    my $per_lun_hash = {};

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # get connection from object
    $conn = $self->{conn};

    # get path info for each 'working' adapter
    my $client_obj = $self->{parent}->{parent}->{parent};
    @adapters = @{$client_obj->{vmhba_adapter}};

    $cmd = "esxcfg-scsidevs -A";
    ($status, @dev_output) = $client_obj->_esx_execute_cmd(-cmd => $cmd,
                                                           -retries => 3,
                                                           -retry_failure_type => '-comment');
    foreach my $line (@dev_output) {
        foreach my $adapter (@adapters) {
            if ($line =~ /$adapter /) {
                push(@all_devs,  $line);
                last;
            }
        }
    }

    # output can look like:
    # esxcfg-scsidevs -A | grep "vmhba1 "
    #vmhba1      naa.60a98000686f6477536f4f346e613369
    #vmhba1      naa.60a98000686f652d2f5a4f346e594739
    #vmhba1      naa.60a98000686f6477536f4f346e614159
    #vmhba1      naa.60a980004334652f4f5a4e6834614f44
    #vmhba1      naa.60a98000686f652d2f5a4f346e614356
    #vmhba1      naa.60a98000686f6477536f4f346e613568
    #vmhba1      naa.60a98000686f652d2f5a4f346e594936
    #vmhba1      naa.60a98000686f652d2f5a4f346e61395a
    #vmhba1      naa.60a98000686f6477536f4f346e594230
    #vmhba1      naa.60a98000686f652d2f5a4f346e613763
    #vmhba1      naa.60a980004334652f4f5a4e6834595057
    #vmhba1      naa.60a98000686f6477536f4f346e592d32
    #vmhba1      naa.60a98000686f652d2f5a4f346e594542
    #vmhba1      naa.60a980004334652f4f5a4d79505a3355
    #vmhba1      naa.60a98000686f6477536f4f346e593832
    #vmhba1      naa.60a980004334652f4f5a4e68342f2d30

    # check to make sure we have discovered some devices to work with
    # there should be devices by the time this code is run.
    if (not scalar(@all_devs)) {
        $self->_set_err("No devices were found from the provided list of adapters (@adapters)");
        return {};
    }

    # list out all the device paths
    $cmd = "esxcfg-mpath -b";
    ($status, @lun_output) = $client_obj->_esx_execute_cmd(-cmd => $cmd,
                                                           -timeout => 180,
                                                           -retries => 5,
                                                           -retry_failure_type => "-warn");
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned status ($status).");
        return {};
    }

    # OUTPUT:
    #esxcfg-mpath -b
    #naa.60a980004334652f4f5a4e6834614f44 : NETAPP Fibre Channel Disk (naa.60a980004334652f4f5a4e6834614f44)
    #  vmhba1:C0:T5:L1 LUN:1 state:active fc Adapter: WWNN: 20:00:00:1b:32:00:d0:37 WWPN: 21:00:00:1b:32:00:d0:37  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:81:96:97:d7:94
    #  vmhba2:C0:T4:L1 LUN:1 state:active fc Adapter: WWNN: 20:01:00:1b:32:20:d0:37 WWPN: 21:01:00:1b:32:20:d0:37  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:81:86:97:d7:94
    #  vmhba2:C0:T5:L1 LUN:1 state:active fc Adapter: WWNN: 20:01:00:1b:32:20:d0:37 WWPN: 21:01:00:1b:32:20:d0:37  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:81:96:97:d7:94
    #  vmhba1:C0:T4:L1 LUN:1 state:active fc Adapter: WWNN: 20:00:00:1b:32:00:d0:37 WWPN: 21:00:00:1b:32:00:d0:37  Target: WWNN: 50:0a:09:80:86:97:d7:94 WWPN: 50:0a:09:81:86:97:d7:94
    #
    #naa.60a9800056716e39474a5168414c5243 : NETAPP iSCSI Disk (naa.60a9800056716e39474a5168414c5243)
    #  vmhba33:C0:T1:L1 LUN:1 state:active iscsi Adapter: iqn.1998-01.com.vmware:ibm3550-rtp10-4fb6605f  Target: IQN=iqn.1992-08.com.netapp:sn.50399615 Alias= Session=00023d000001 PortalTag=1001
    #  vmhba33:C2:T1:L1 LUN:1 state:active iscsi Adapter: iqn.1998-01.com.vmware:ibm3550-rtp10-4fb6605f  Target: IQN=iqn.1992-08.com.netapp:sn.50399615 Alias= Session=00023d000002 PortalTag=1002

    # join the output by newlines, then split on two newlines, this will organize each chunk into an element in the array
    my @lun_paths_output = split(/\n\n/, join("\n", @lun_output));

    # filter out non-NetApp devices
    @lun_paths_output = grep { /NETAPP/i } @lun_paths_output;

    # traverse output and generate per LUN hash of pathing information
    foreach my $line (@all_devs) {
        $num_alive = $num_dead = $num_paths = 0;

        # We are going to run esxcfg-mpath on each device in the list
        # and then pull the information about pathing out from that
        # output.
        my ($hba, $naa_device) = split(/\s+/, $line);

        my $protocol = "";

        foreach my $lun_paths (@lun_paths_output) {
            # if we match the naa device, then start counting paths
            if ($lun_paths =~ /^$naa_device\s/) {
                # handle fcp luns
                if ($lun_paths =~ /fc Adapter/i) {
                    $protocol = "fcp";
                }

                # handle iscsi luns
                if ($lun_paths =~ /iscsi Adapter/i) {
                    $protocol = "iscsi";
                }

                # handle case where a lun is multiprotocol
                if ($lun_paths =~ /iscsi Adapter/i and $lun_paths =~ /fc Adapter/i) {
                    $protocol = "fcp|iscsi";
                }

                # convert the output to an array and remove all lines that do not
                # represent paths.
                my @lun_paths = split(/\n/, $lun_paths);
                @lun_paths = grep { /state\:/i } @lun_paths;

                # traverse all the paths to the LUN
                foreach my $path (@lun_paths) {

                    # get alive/dead paths
                    if ($path =~ /state:dead/i) {
                        $num_dead++;
                    } elsif ($path =~ /state:active/i) {
                        $num_alive++;
                    } elsif ($path =~ /state:standby/i) {
                        $num_alive++;
                    }
                }

                $num_paths = scalar(@lun_paths);

                # break out of this loop since we are done counting paths for this lun
                last;
            }
        }

        $per_lun_hash->{$naa_device}->{num_paths} = $num_paths;
        $per_lun_hash->{$naa_device}->{num_paths_alive} = $num_alive;
        $per_lun_hash->{$naa_device}->{num_paths_dead} = $num_dead;
        $per_lun_hash->{$naa_device}->{num_paths_primary} = "NA";
        $per_lun_hash->{$naa_device}->{num_paths_secondary} = "NA";
        $per_lun_hash->{$naa_device}->{protocol} = $protocol;
    }

    return $per_lun_hash;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $parent, $obj_template);

    $self = {};
    bless($self, $class);

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent' => $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
    }

    # plug in the parent
    $self->{parent} = $parent;

    return $self;
}

1;
