# Copyright (c) 2010,2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to native multipath tools for ESXi4/5
## @author hrivera@netapp.com, arrance@netapp.com
## @status Private
## @pod noneo

# Package Name
##############
package San::SanClient::MP::ESXiMP::NativeMP4;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::ESXMP::NativeMP4;
use San::SanClient::MP::ESXiMP;
use Error;
use San::SanLog;
use Data::Dumper;
use vars qw(@ISA);

@ISA = qw(Error SanLog MP San::SanClient::MP::ESXMP::NativeMP4);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       NOOP for ESXi
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       NOOP for ESXi
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    $self->_clear_err();
    $self->_set_err("Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: verify()
# Objective:       NOOP for ESXi
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # NOOP for ESXi

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       NOOP for ESXi
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: config()
# Objective:       NOOP For ESXi
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    $self->_clear_err();

    $self->_set_err("Function not supported");

    return $ClientUtil::FALSE;
}

################################################################################
# Subroutine Name: config_paths()
# Objective:       NOOP for ESXi
# Details:         See POD documentation at the end of this file
################################################################################
sub config_paths {
    my ($self, %args) = @_;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: set_psp_policy
# Objective:       Given a specific device, and requested path selection policy
# change the policy
################################################################################
sub set_psp_policy {
    my ($self, %args) = @_;

    # invoke ESXMP::NativeMP4::set_psp_policy() using native esx cmds
    return $self->SUPER::set_psp_policy(%args,
                                        -native_cmd => 1);
}

################################################################################
# Subroutine Name: set_vxdmp_io_policy
# Objective:       Given an enclosure, set the io policy provided
################################################################################
sub set_vxdmp_io_policy {
    my ($self, %args) = @_;
    my $enclosure;
    my $io_policy;

    # parse arguments
    foreach (keys %args) {
        if (/^-?enclosure$/) {
            if (not length($enclosure = $args{$_})) {
                $self->_set_err("Invalid value for argument '-enclosure' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/^-?io_policy$/) {
            if (not length($io_policy = $args{$_})) {
                $self->_set_err("Invalid value for argument '-io_policy' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check for required params
    if (not $enclosure) {
        $self->_set_err("Missing required parameter -enclosure");
        return $ClientUtil::FALSE;
    }
    if (not $io_policy) {
        $self->_set_err("Missing required parameter -io_policy");
        return $ClientUtil::FALSE;
    }

    # get the sanclient object
    my $client_obj = $self->{parent}->{parent}->{parent};

    # execute the vxdmpadm command to set the iopolicy
    my $cmd = "setattr enclosure ${enclosure} iopolicy=${io_policy}";
    my ($status, @output) = $client_obj->_vxdmpadm_execute_cmd(-args => $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to set VxDMP IO Policy $io_policy on Enclosure $enclosure", $self->errmsg());
        return $ClientUtil::FALSE;
    }

    # verify it was set
    my $curr_io_policy = $self->get_vxdmp_io_policy(-enclosure => $enclosure);
    if ($curr_io_policy ne $io_policy) {
        $self->_set_err("VxDMP IOpolicy for enclosure $enclosure did not change after setting " .
                        "it to $io_policy (currently is $curr_io_policy)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_vxdmp_io_policy
# Objective:       Given an enclosure, get the io policy
################################################################################
sub get_vxdmp_io_policy {
    my ($self, %args) = @_;

    my $enclosure;
    my $io_policy;

    # parse arguments
    foreach (keys %args) {
        if (/^-?enclosure$/) {
            if (not length($enclosure = $args{$_})) {
                $self->_set_err("Invalid value for argument '-enclosure' => $args{$_}");
                return undef;
            }
            next;
        }
    }

    # check for required params
    if (not $enclosure) {
        $self->_set_err("Missing required parameter -enclosure");
        return undef;
    }

    # get the sanclient object
    my $client_obj = $self->{parent}->{parent}->{parent};

    # verify it was set (ignore any errors)
    my $cmd = "getattr enclosure ${enclosure} iopolicy";
    my ($status, @output) = $client_obj->_vxdmpadm_execute_cmd(-args => $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to get VxDMP IO Policy on Enclosure $enclosure", $self->errmsg());
        return undef;
    }
    foreach my $line (@output) {
        if ($line =~ /^\s*${enclosure}\s+iopolicy\s+\S+\s+(\S+)/) {
            $io_policy = $1;
            last;
        }
    }

    return lc($io_policy);
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $parent, $obj_template);

    $self = {};
    bless($self, $class);

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent' => $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
    }

    # plug in the parent
    $self->{parent} = $parent;

    return $self;
}

1;
