# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools
## @author latif@netapp.com, thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::HPUXMP::DMP;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::HPUXMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       enable DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output,);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::DMP::enable() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "/usr/sbin/vxdctl enable";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXMP::DMP::enable() Error while enabling vxdctl (" . @output . ")");
        return $ClientUtil::FALSE;
    }

    #### Verify that vxdctl daemon is running
    $self->_is_vxdctl_running(-conn => $conn);
    if ($self->error) {
        $self->_set_err("Error: HPUXMP::DMP::enable() _is_vxdctl_running reported error");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       disable DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::DMP::disable() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "/usr/sbin/vxdctl disable";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXMP::DMP::disable() Error while disabling vxdctl");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify()
# Objective:       verify proper enviroment for DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    my (@pkgs, $conn, $status, $cmd, @output,);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::DMP::verify() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ### Verufy that package for DMP is installed (
    foreach my $pkg ("VRTSvxvm", "VRTSvlic") {
        $cmd = "/usr/sbin/swlist $pkg";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Error: HPUXMP::DMP::verify() Failed to verify packaged ($pkg) is installed");
            return $ClientUtil::FALSE;
        } else {
            $self->sanlog(-msg => "HPUXMP::DMP::verify() Verified that package ($pkg) is installed",
                          -msg_level => 9);
        }
    }

    ### Verify that the DMP module is loaded on the client
    $cmd = "ps -ef|grep -i [v]xconfig";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status != 0) {
        $self->_set_err("Error: HPUXMP::DMP::verify() Failed to verify that DMP daemon is running (" . @output . ")");
        return $ClientUtil::FALSE;
    }

    # Verify that the ASL/APM is installed. For VRTS 5.0.1 on
    # HP-UX 11iv3 (only one supported on 11iv3), this is ALUA only
    # and does not require the APM.
    # Case of 5.1+, support is in VRTS package itself
    if ($self->{version} >= 5) {
        if ($self->{version} >= 5.1) {
            @pkgs = ("VRTSaslapm");
        } else {
            if ($self->{parent}->{parent}->{parent}->{os_version} !~ /11\.3/i) {
                @pkgs = ("VRTSNTAPasl", "VRTSNTAPapm");
            } else {
                @pkgs = ("VRTSNTAPasl");
            }
        }
    } else {
        @pkgs = ("NTAPasl");
    }

    foreach my $pkg (@pkgs) {
        $cmd = "/usr/sbin/swlist $pkg";
        ($status, @output) = &ClientUtil::execute($conn , $cmd);
        if ($status != 0) {
            $self->_set_err("Error: HPUXMP::DMP::verify() Failed to verify packaged ($pkg) is installed");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       setup proper enviroment for DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::DMP::setup() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return $self;
}

################################################################################
# Subroutine Name: config()
# Objective:       verify proper enviroment for DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::DMP::config() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    return $self;
}

################################################################################
# Subroutine Name: get_mp_devices()
# Objective:       get all mp devices
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mp_devices {
    my ($self, %args) = @_;

    my ($cmd, @output, @hDevices, $conn, $status, $cnt, $found);

    $self->_clear_err();
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::DMP::get_mp_devices() Invalid value for argument '-conn' => $args{$_}");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # ioscan will return the following device list.
    # disk    262  0/3/1/0.61.15.0.0.7.2  sdisk   CLAIMED     DEVICE       NETAPP  LUN
    #                        /dev/dsk/c19t7d2   /dev/rdsk/c19t7d2
    # disk    263  0/3/1/0.61.15.0.0.7.3  sdisk   CLAIMED     DEVICE       NETAPP  LUN
    #                        /dev/dsk/c19t7d3   /dev/rdsk/c19t7d3
    $cmd = "/usr/sbin/ioscan -fnk -C disk";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0 or scalar(@output) == 0) {
        $self->_set_err("Error: HPUXMP::DMP::get_mp_devices() No multipath devices found");
        return (wantarray()) ? () : [];
    }

    $cnt = 0;
    $found = $ClientUtil::FALSE;
    foreach my $line (@output) {
        if ($line =~ /netapp/i and $line =~ /claimed/i) {
            $found = $ClientUtil::TRUE;
            next;
        }

        if ($found) {
            my $device = (split(' ', $line))[1];
            $device =~ s#/dev/rdsk/##;

            # create hash of lun description and attributes
            my $hDevs = {};
            $hDevs->{dev_path}  = "/dev/dsk";
            $hDevs->{path}  = "/dev/rdsk";
            $hDevs->{name}  = $device;
            $hDevs->{name_ctd} = $device;
            $hDevs->{type}  = "NETAPP";
            $hDevs->{label} = "NA";
            $hDevs->{id}    = $cnt;
            $hDevs->{primary_path} = 0;

            push @hDevices, $hDevs;
            $cnt++;
        }
        $found = $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "HPUXMP::DMP::get_mp_devices() Found " . scalar(@hDevices) . " multipathing devices");

    return (wantarray()) ? @hDevices : \@hDevices;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # no-op for HPUX

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each DMP node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($per_lun_hash, $cmd, @vx_output, @output, $status, $conn);
    my ($line, $enclosure, @encl_output);
    my ($dmp_node, $state, $num_paths, $num_alive, $num_dead);
    my ($validate_mode);

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Error: HPUXMP::DMP::get_path_info() Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $per_lun_hash = {};
    @output = ();

    # get connection from object
    $conn = $self->{conn};

    # run "vxdmpadm listenclosure all" to get list of enclosures
    $cmd = "/usr/sbin/vxdmpadm listenclosure all | grep -v 'DISCONNECTED' | grep 'CONNECTED' | grep -v 'DISKS'";
    ($status, @vx_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                       180, 5, "-warn");
    if ($status != 0) {
        $self->_set_err("Error: HPUXMP::DMP::get_path_info() Failed to run vxdmpadm listenclosure all");
        return {};
    }

    foreach $line (@vx_output) {
        $enclosure = (split(/\s+/, $line))[0];
        if ($enclosure ne "") {
            $cmd = "/usr/sbin/vxdmpadm getdmpnode enclosure=$enclosure";
            ($status, @encl_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                 180, 5, "-warn");
            if ($status != 0) {
                $self->_set_err("Error: HPUXMP::DMP::get_path_info() Failed to run $cmd to get dmpnode and path information");
                return {};
            }
            push(@output, @encl_output);
        }
    }

    # go over the enclosure output and pick apart each line (which
    # will be the dmp_node followed by it's characteristics
    # Example:
    #  NAME                 STATE     ENCLR-TYPE   PATHS  ENBL  DSBL  ENCLR-NAME
    # =========================================================================
    # c5t0d0s2             ENABLED      FAS960       4      4     0     FAS9600

    foreach $line (@output) {

        # skip header and newlines
        if ($line =~ /ENCLR-TYPE/i or
            $line =~ /=====/i or
            $line =~ /^\s*$/) {
            next;
        }

        # get per dmp node pathing information from output
        ($dmp_node, $state, undef,
         $num_paths, $num_alive, $num_dead, undef) = split(/\s+/, $line);

        # store number of paths
        $per_lun_hash->{$dmp_node}->{num_paths} = $num_paths;

        # store number of paths that are enabled (alive)
        $per_lun_hash->{$dmp_node}->{num_paths_alive} = $num_alive;

        # store number of paths that are disabled (dead)
        $per_lun_hash->{$dmp_node}->{num_paths_dead} = $num_dead;

        # store number of paths that are primary
        $per_lun_hash->{$dmp_node}->{num_paths_primary} = "NA";

        # store number of paths that are secondary
        $per_lun_hash->{$dmp_node}->{num_paths_secondary} = "NA";

        # no iSCSI support with DMP - hardcode to FCP
        $per_lun_hash->{$dmp_node}->{protocol} = "FCP";
    }

    return $per_lun_hash;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self, $parent, $obj_template);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument \'-parent\'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        # fix the parent link and return
        $self->{parent} = $parent;
        return $self;
    }

    $self->{parent} = $parent;

    # This attribute denote if the devices provided by this package are
    # primary paths or all paths to all luns
    $self->{primary_paths} = 0;

    $self->_get_pkginfo(-conn => $self->{parent}->{parent}->{parent}->{conn});

    # Returning DMP object
    return $self;
}

################################################################################
# Subroutine Name: _get_pkginfo
# Objective:       Get lvm package info
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self, %args) = @_;

    $self->_clear_err();

    my ($cmd, @output, $status, $version_string, $name, $full_version);
    my ($major_version, $sp, $rp, $mp, $conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return ("", "");
            }
            next;
        }
    }

    $self->_clear_err();

    $major_version = "";
    $sp            = "";
    $rp            = "";
    $mp            = "";

    $cmd = "/usr/sbin/swlist VRTSvxvm | grep VRTSvxvm";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to execute command $cmd");
        return $ClientUtil::FALSE;
    }

    $version_string = $output[0];
    $version_string =~ s/\#//;
    $version_string =~ s/^\s+//;

    $name         = (split(/\s+/, $version_string))[0];
    $full_version = (split(/\s+/, $version_string))[1];

    $self->sanlog(-msg => "Veritas version is: $full_version");

    #================================================
    # format of swlist VRTSvxvm returns:
    #
    # 6.0:
    # VRTSvxvm                      6.0.000.000    Veritas Volume Manager by Symantec
    #
    # 5.1SP1RP1:
    # VRTSvxvm                      5.1.101.000    Veritas Volume Manager by Symantec
    #
    # 5.1SP1:
    # VRTSvxvm                      5.1.100.000    Veritas Volume Manager by Symantec
    #
    # 5.1RP2:
    # VRTSvxvm                      5.1.2.000    Veritas Volume Manager by Symantec
    #
    # 5.0.1:
    # VRTSvxvm                      5.0.31.5     Veritas Volume Manager by Symantec
    #================================================
    my ($d1, $d2, $d3, $d4) = split(/\./, $full_version);

    # The version is the first 2 put together
    $major_version = $d1 . "." . $d2;

    # Now is where things get kludgy.  We now have to use the output from
    # lspp above to determin the sp, mp, rp
    if ($major_version >= 5.1) {
        if ($d3 eq "100") {
            $sp = "SP1";
        } elsif ($d3 eq "101") {
            $sp = "SP1";
            $rp = "RP1";
        } elsif ($d3 eq "2") {
            $rp = "RP2";
        }
    } elsif ($major_version == 5.0) {
        if ($d3 == 3) {
            $mp = "MP3";
            if ($d4 eq "200") {
                $rp = "RP2";
            }
        } elsif ($d3 >= 31) {
            # This is 5.0.1.  It should be treated like 5.0 MP3 so
            # we will set the MP to MP3.
            $mp = "MP3";
        } else {
            # Default to MP1 for all 5.0 versions except MP3
            $mp = "MP1";
        }
    }

    $self->{name}     = "VRTSvxvm";
    $self->{version}  = $major_version;
    $self->{revision} = $major_version;
    $self->{mp}       = $mp;
    $self->{rp}       = $rp;
    $self->{sp}       = $sp;
    $self->{pkginfo}  = \@output;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _is_vxdctl_running()
# Objective:       returns ClientUtil::TRUE or ClientUtil::FALSE depending on
#                  if vxdctl is running or not.
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_vxdctl_running {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::DMP::_is_vxdctl_runing() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check status of vxdctl service
    my $cmd = "/usr/sbin/vxdctl mode";
    my ($status, @output) = &ClientUtil::execute($conn, $cmd);

    if (grep /enable/i, @output) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: _rescan
# Objective:       To run the vxdctl command to rescan vxvm's list of
#                  disks that it thinks are under it's control.
# Details:         See POD documentation at the end of this file
################################################################################
sub _rescan {
    my ($self, %args) = @_;

    my ($conn, $cmd, @output, $status, $host_obj,);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::DMP::_rescan() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->_clear_err();

    # need to run the vxdctl script after the new VTOC is down
    # to get DMP to see the disk again (properly) .. not completely
    # sure why
    $cmd = "/usr/sbin/vxdctl enable";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXMP::DMP::rescan() Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _dmp_version
# Objective:       To version and revision of VXVM package installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _dmp_version () {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, $version_string, $conn, $vers);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::DMP::_dmp_version() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->_clear_err();

    $cmd = "/usr/sbin/swlist VRTSvxvm| grep VRTSvxvm";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error: HPUXMP::DMP::_dmp_version() Failed to execute command $cmd");
        return ("", "");
    }

    $version_string = $output[0];
    $version_string =~ s/\#//;
    $version_string =~ s/^\s+//;
    $vers = (split(/\s+/, $version_string))[1];

    return $vers, $vers;
}

1;
