# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools
## @author latif@netapp.com, thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::HPUXMP::MASS;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::HPUXMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       enable MASS MP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       disable MASS MP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify()
# Objective:       verify proper enviroment for MASS MP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::MASS::verify() Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Output example for system level
    # PHKL_38054                            1.0            esctl cumulative patch
    $cmd = "/usr/sbin/swlist -l fileset |grep \"esctl cumulative patch\"";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0 or scalar(@output) == 0) {
        $self->_set_err("Error: HPUXMP::MASS::verify() Error while getting OS version");
        return $ClientUtil::FALSE;
    }

    # Check the output to see if it contains the "esctl cumm..." string
    # This tells us that the system is at the apporiate level to run MP type of MASS
    if (grep(/esctl\s+cumulative\s+patch/i, @output)) {
        $self->sanlog(-msg => "Current installed OS version does meet " .
                              "\nminimal OS requirements to use MASS multipathing stack.");
    } else {
        $self->_set_err("Error: HPUXMP::MASS::verify() Current installed OS version " .
                        "\nDOES NOT meet minimal OS requirements to use MASS multipathing stack." .
                        "\nMake sure that you have 11iv3 Feb 2008 or the required set of patches " .
                        "\ninstalled on your system");
        return $ClientUtil::FALSE;
    }

    # Check to make sure that the tool kit is at least the minimal 4.2
    if ($self->{parent}->{parent}->{parent}->get_hu_toolkit_version() < 4.2) {
        # Toolkit is not correct version
        $self->_set_err("Error: HPUXMP::MASS::verify() Current HU version (" .
                        $self->{parent}->{parent}->{parent}->get_hu_toolkit_version() . ")" .
                        "\nDOES NOT meet minimal HU requirements to use MASS multipathing stack." .
                        "\nMake sure that you have HU toolkit 4.2 or later installed on your system");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # no-op for HPUX

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       setup proper enviroment for MASS MP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    # NO-OP

    return $self;
}

################################################################################
# Subroutine Name: config()
# Objective:       verify proper enviroment for MASS MP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    # NO-OP

    return $self;
}

################################################################################
# Subroutine Name: get_mp_devices()
# Objective:       get all mp devices
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mp_devices {
    my ($self, %args) = @_;

    my ($cmd, @output, @hDevices, $conn, $status, $cnt, $disk_name);

    $self->_clear_err();
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::MASS::get_mp_devices() Invalid value for argument '-conn' => $args{$_}");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # get the boot device (to exclude). Call from Host object
    my $boot_device = $self->{parent}->{parent}->{parent}->_get_boot_device();
    if ($self->error) {
        return (wantarray()) ? () : [];
    }

    # ioscan will return the following device list.
    # disk    262  0/3/1/0.61.15.0.0.7.2  sdisk   CLAIMED     DEVICE       NETAPP  LUN
    #                        /dev/dsk/c19t7d2   /dev/rdsk/c19t7d2
    # disk    263  0/3/1/0.61.15.0.0.7.3  sdisk   CLAIMED     DEVICE       NETAPP  LUN
    #                        /dev/dsk/c19t7d3   /dev/rdsk/c19t7d3
    $cmd = "/usr/sbin/ioscan -fnk -C disk";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0 or scalar(@output) == 0) {
        $self->_set_err("Error: HPUXMP::MASS::get_mp_devices() No multipath devices found");
        return (wantarray()) ? () : [];
    }

    my $current_device = "";
    foreach my $line (@output) {
        if ($line =~ /netapp/i and $line =~ /claimed/i) {
            # Parse the disk number from the output
            $line =~ /^\s*disk\s+(\d+)\s+/;
            $current_device = "disk" . $1;
            next;
        }

        if ($current_device ne "" and $line =~ /\/dev\//) {
            my $device = (split(' ', $line))[1];
            $device =~ s#/dev/rdsk/##;

            # create hash of lun descriptions
            my $hDevs = {};
            $hDevs->{dev_path}  = "/dev/dsk";
            $hDevs->{path}  = "/dev/rdsk";
            $hDevs->{name}  = $device;
            $hDevs->{name_ctd} = $device;
            $hDevs->{type}  = "NETAPP";
            $hDevs->{label} = "NA";
            $hDevs->{id}    = $cnt;
            $hDevs->{primary_path} = 0;

            # Run the following command to get the proper /dev/rdisk/diskXXX
            # $> ioscan -m dsf /dev/dsk/c39t1d4
            # Persistent DSF           Legacy DSF(s)
            # ========================================
            # /dev/disk/disk579        /dev/dsk/c39t1d4
            $cmd = "/usr/sbin/ioscan -m dsf /dev/dsk/" . $device;
            my ($cmd_status, @cmd_output) = &ClientUtil::execute($conn, $cmd);
            if ($cmd_status > 0 or scalar(@cmd_output) == 0) {
                $self->_set_err("Error: HPUXMP::MASS::get_mp_devices() Unable to get rdisk for $device");
                return (wantarray()) ? () : [];
            }

            # Grep out the one line we need to get the disk number
            my ($rdisk) = grep(/\/dev\/disk/, @cmd_output);
            if ($rdisk =~ /\/dev\/disk\/(.+)\s+/) {
                # Save it on the hash
                $hDevs->{name_disk} = $1;
            } else {
                $self->_set_err("Error: HPUXMP::MASS::get_mp_devices() Unable to get rdisk for $device ($rdisk)");
                return (wantarray()) ? () : [];
            }

            push @hDevices, $hDevs;

            # Clear the current device
            $current_device = "";
        }
    }

    $self->sanlog(-msg => "HPUXMP::MASS::get_mp_devices() Found " . scalar(@hDevices) . " multipathing devices");

    return (wantarray()) ? @hDevices : \@hDevices;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($cmd, @format_output, @output, $status, $conn);
    my ($validate_mode);

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Error: HPUXMP::MASS::get_path_info() Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my $per_lun_hash = {};

    # get connection from object
    $conn = $self->{conn};

    # get the root node for all LUNs
    my @devices = $self->get_mp_devices(-conn => $conn);
    return {} if $self->error();

    foreach my $dev (@devices) {
        my ($state, $line, $protocol, $path);
        my ($num_paths, $lun, $start_path);

        $cmd = "/usr/sbin/scsimgr lun_map -D /dev/rdisk/" . $dev->{name_disk};
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status > 0) {
            $self->_set_err("Error: HPUXMP::MASS::get_path_info() failure executing ($cmd)");
            return {};
        }

        # process scsimgr output:
        # example output:
        #         LUN PATH INFORMATION FOR LUN : /dev/rdisk/disk1389
        # Total number of LUN paths     = 4
        # World Wide Identifier(WWID)    = 0x60a98000486e2f4c6c4a4d3731796a63

        # LUN path : lunpath508
        # Class                         = lunpath
        # Instance                      = 508
        # Hardware path                 = 0/2/1/0.0x500a098187197d3b.0x4003000000000000
        # SCSI transport protocol       = fibre_channel
        # State                         = ACTIVE
        # Last Open or Close state      = ACTIVE

        $lun = (grep(/LUN PATH INFORMATION FOR LUN/i, @output))[0];
        $lun =~ s/.+\://;
        $lun =~ s/\s+//;

        $num_paths = (grep(/Total number of LUN paths/i, @output))[0];
        $num_paths =~ s/.+\=//;
        $num_paths =~ s/\s+//;
        $per_lun_hash->{$lun}->{num_paths} = $num_paths;

        $start_path = 0;
        foreach $line (@output) {
            if ($line =~ /^LUN path\s+:/i) {
                $start_path = 1;
            }

            if ($line =~ /^\s?$/i) {
                $start_path = 0;
                next;
            }

            if ($start_path == 1) {
                # get the path
                if ($line =~ /^LUN path\s+:\s+(\S+)/i) {
                    $path = $1;
                }

                # get the protocol
                if ($line =~ /^SCSI transport protocol/i) {
                    if ($line =~ /fibre_channel/) {
                         $protocol = "FCP";
                    } else {
                         $protocol = "iSCSI";
                    }
                }

                # get the state
                if ($line =~ /^State\s+\=\s+(\S+)/i) {
                    $state = $1;
                }
            }

            if ($path ne "") {
                $per_lun_hash->{$lun}->{$path}->{path_type} = "NA";
                $per_lun_hash->{$lun}->{$path}->{path_state} = $state;
                $per_lun_hash->{$lun}->{protocol} = $protocol;
            }
        }
    }

    # and storing in the same hash
    foreach my $key (keys %{$per_lun_hash}) {
        my $alive_cnt = 0;
        my $dead_cnt = 0;
        my $proxy_cnt = 0;
        my $primary_cnt = 0;
        foreach my $path_key (keys %{$per_lun_hash->{$key}}) {
            if ($path_key =~ /num_paths|protocol/) {
                next;
            }

            # store number of available (alive) and unavailable
            # (dead) paths
            if ($per_lun_hash->{$key}->{$path_key}->{path_state} =~ /ACTIVE|STANDBY|UNOPEN/i) {
                $alive_cnt++;
            } else {
                $dead_cnt++;
            }

            # store proxy/primary information
            if ($per_lun_hash->{$key}->{$path_key}->{path_state} =~ /ACTIVE/i) {
                $primary_cnt++;
            } elsif ($per_lun_hash->{$key}->{$path_key}->{path_state} =~ /STANDBY/i) {
                $proxy_cnt++;
            }
        }

        # store number of paths that are enabled (alive)
        $per_lun_hash->{$key}->{num_paths_alive} = $alive_cnt;

        # store number of paths that are disabled (dead)
        $per_lun_hash->{$key}->{num_paths_dead} = $dead_cnt;

        # store number of paths that are primary
        $per_lun_hash->{$key}->{num_paths_primary} = $primary_cnt;

        # store number of paths that are secondary
        $per_lun_hash->{$key}->{num_paths_secondary} = $proxy_cnt;
    }

    return $per_lun_hash;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self, $parent, $obj_template);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument \'-parent\'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        # fix the parent link and return
        $self->{parent} = $parent;
        return $self;
    }

    $self->{parent} = $parent;

    # This attribute denotes if the devices provided by this package are
    # primary paths or all paths to all luns
    $self->{primary_paths} = 1;

    # Returning MASSMP object
    return $self;
}

1;
