# Copyright (c) 2007-2008 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools
## @author latif@netapp.com, thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::HPUXMP::PvLinks;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::HPUXMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       enable PvLinksMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       disable PvLinksMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # no-op for HPUX

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify()
# Objective:       verify proper enviroment for PvLinksMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    # NO-OP

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       setup proper enviroment for PvLinksMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    # NO-OP

    return $self;
}

################################################################################
# Subroutine Name: config()
# Objective:       verify proper enviroment for PvLinksMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    # NO-OP

    return $self;
}

################################################################################
# Subroutine Name: get_mp_devices()
# Objective:       get all mp devices
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mp_devices {
    my ($self, %args) = @_;

    my ($cmd, @output, @hDevices, $conn, $status, $cnt, $found);

    $self->_clear_err();
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Error: HPUXMP::PVLinks::get_mp_devices() Invalid value for argument '-conn' => $args{$_}");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # ioscan will return the following device list.
    # disk    262  0/3/1/0.61.15.0.0.7.2  sdisk   CLAIMED     DEVICE       NETAPP  LUN
    #                        /dev/dsk/c19t7d2   /dev/rdsk/c19t7d2
    # disk    263  0/3/1/0.61.15.0.0.7.3  sdisk   CLAIMED     DEVICE       NETAPP  LUN
    #                        /dev/dsk/c19t7d3   /dev/rdsk/c19t7d3
    $cmd = "/usr/sbin/ioscan -fnk -C disk";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0 or scalar(@output) == 0) {
        $self->_set_err("Error: HPUXMP::PVLinks::get_mp_devices() No multipath devices found");
        return (wantarray()) ? () : [];
    }

    my $current_device = "";
    foreach my $line (@output) {
        if ($line =~ /netapp/i and $line =~ /claimed/i) {
            # Parse the disk number from the output
            $line =~ /^\s*disk\s+(\d+)\s+/;
            $current_device = "disk" . $1;
            next;
        }

        if ($current_device ne "" and $line =~ /\/dev\//) {
            my $device = (split(' ', $line))[1];
            $device =~ s#/dev/rdsk/##;

            # create hash of lun descriptions
            my $hDevs = {};
            $hDevs->{dev_path}  = "/dev/dsk";
            $hDevs->{path}  = "/dev/rdsk";
            $hDevs->{name}  = $device;
            $hDevs->{name_disk}  = $current_device;
            $hDevs->{name_ctd} = $device;
            $hDevs->{type}  = "NETAPP";
            $hDevs->{label} = "NA";
            $hDevs->{id}    = $cnt;
            $hDevs->{primary_path} = 0;

            push @hDevices, $hDevs;

            # Clear the current device
            $current_device = "";
        }
    }

    $self->sanlog(-msg => "HPUXMP::PVLinks::get_mp_devices() Found " . scalar(@hDevices) . " multipathing devices");

    return (wantarray()) ? @hDevices : \@hDevices;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($cmd, @format_output, @output, $status, $conn);
    my ($line);
    my ($validate_mode);

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Error: HPUXMP::PVLinks::get_path_info() Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my $per_lun_hash = {};

    # get connection from object
    $conn = $self->{conn};

    # run sanlun lun show to get LUNs and
    # LUNs with pathing information.
    my $path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();
    $cmd = "export PATH=\$PATH:/usr/sbin; " . $path . "/sanlun lun show";
    ($status, @output) = &ClientUtil::execute_retry($conn, $cmd,
                                                    360, 5, "-warn");
    if ($status != 0) {
        $self->_set_err("Error: HPUXMP::PVLinks::get_path_info() Failed to run sanlun lun show");
        return {};
    }

    # parse through output generating hash of unique LUNs and their
    # state (good|bad)
    foreach $line (@output) {
        chomp($line);

        # skip sanlun header
        next if ($line =~ /lun-pathname/);

        # parse sanlun output and generate hash of paths
        # Example:
        # fas3070-rtp08:  /vol/nate_1_8_hprx2600_fak03_nate_aggr1/nate_hprx2600-fak03_hpux_21  /dev/rdsk/c25t0d0  td1 FCP  1g (1073741824) GOOD
        $line =~ s/^\s+//;
        my ($filer, $filer_path, $path, undef,
            $type, undef, undef, $state) = split (/\s+/, $line);
        my $lun = $filer . $filer_path;
        $type =~ s/\s+//g;
        $per_lun_hash->{$lun}->{$path}->{path_type} = "NA";
        $per_lun_hash->{$lun}->{protocol} = $type;
        $state =~ s/\s+//g;
        $per_lun_hash->{$lun}->{$path}->{path_state} = $state;
    }

    # traverse the per LUN hash extracting the pathing information
    # and storing in the same hash
    foreach my $key (keys %{$per_lun_hash}) {
        # store number of paths.
        # account for protocol being on the hash.
        $per_lun_hash->{$key}->{num_paths} =
                                   (scalar(keys %{$per_lun_hash->{$key}}) - 1);

        my $alive_cnt = 0;
        my $dead_cnt = 0;
        my $proxy_cnt = 0;
        my $primary_cnt = 0;
        foreach my $path_key (keys %{$per_lun_hash->{$key}}) {
            if ($path_key =~ /num_paths|protocol/) {
                next;
            }

            # store number of available (alive) and unavailable
            # (dead) paths
            if ($per_lun_hash->{$key}->{$path_key}->{path_state} =~ /good/i) {
                $alive_cnt++;
            } else {
                $dead_cnt++;
            }

            if ($per_lun_hash->{$key}->{$path_key}->{path_type} =~ /primary/i) {
                $primary_cnt++;
            } else {
                $proxy_cnt++;
            }
        }

        # store number of paths that are enabled (alive)
        $per_lun_hash->{$key}->{num_paths_alive} = $alive_cnt;

        # store number of paths that are disabled (dead)
        $per_lun_hash->{$key}->{num_paths_dead} = $dead_cnt;

        # store number of paths that are primary
        $per_lun_hash->{$key}->{num_paths_primary} = $primary_cnt;

        # store number of paths that are secondary
        $per_lun_hash->{$key}->{num_paths_secondary} = $proxy_cnt;

    }

    return $per_lun_hash;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self, $parent, $obj_template);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument \'-parent\'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        # fix the parent link and return
        $self->{parent} = $parent;
        return $self;
    }

    $self->{parent} = $parent;

    # This attribute denote if the devices provided by this package are
    # primary paths or all paths to all luns
    $self->{primary_paths} = 0;

    # Returning PvLinksMP object
    return $self;
}

1;
