# Copyright (c) 2007-2009 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools
## @author  rawat@netapp.com, thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::LinuxMP::DMMP;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::LinuxMP;
use San::SanUtil;
use Error;
use San::SanLog;
use Connectrec;
use Data::Dumper;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       Start the multipath service, load modules, and run multipath.
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, $conn, $client_obj);

    $client_obj = $self->{parent}->{parent}->{parent};
    my $hosttype = $client_obj->hosttype();

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn' => %args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (not defined($conn)) {
        $self->_set_err("'-conn' parameter is required.");
        return $ClientUtil::FALSE;
    }

    if ($client_obj->{major_revision} =~ /cxs/i) {
        my $host_uuid = $client_obj->{uuid};

        # goto maintenance mode
        $cmd = "xe host-disable uuid=$host_uuid";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to run the cmd: $cmd");
            return $ClientUtil::FALSE;
        }

        # enable multipath
        $cmd = "xe host-param-set other-config:multipathing=true uuid=$host_uuid";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to run the cmd: $cmd");
            return $ClientUtil::FALSE;
        }

        $cmd = "xe host-param-set other-config:multipathhandle=dmp uuid=$host_uuid";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to run the cmd: $cmd");
            return $ClientUtil::FALSE;
        }

        # release maintence mode
        $cmd = "xe host-enable uuid=$host_uuid";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to run the cmd: $cmd");
            return $ClientUtil::FALSE;
        }
    }

    # if on redhat, remove the automation_blacklisted_all_devices line in the multipath.conf file if it exists
    if ($hosttype =~ /rhel|redhat|fedora/i) {
        $cmd = "perl -p -i.bak -e 'END {exit 0}; s/^.*#\\s*automation_blacklisted_all_devices//g' /etc/multipath.conf";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to remove 'automation_blacklisted_all_devices' line in /etc/multipath.conf");
            return $ClientUtil::FALSE;
        }
    }

    # The first thing we should do is turn on the multipathd service.
    # if the service is not running, turn it on.
    my $multipathd_status = $self->_is_multipathd_running(-conn => $conn);
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    if ($multipathd_status == $ClientUtil::FALSE) {
        $self->start_multipath_daemon();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    # check if dm_multipath module is enabled
    my $dm_loaded = $client_obj->_is_module_loaded(-module => "dm_multipath");
    if ($client_obj->error) {
        $self->_set_err($client_obj->errmsg);
        return $ClientUtil::FALSE;
    }
    if ($dm_loaded == $ClientUtil::FALSE) {
        # its not enabled, modprobe it
        $cmd = "modprobe dm_multipath";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status > 0) {
            $self->_set_err("Error while running cmd:($cmd)");
            return $ClientUtil::FALSE;
        }

        # make sure it loaded correctly
        $dm_loaded = $client_obj->_is_module_loaded(-module => "dm_multipath");
        if ($client_obj->error) {
            $self->_set_err($client_obj->errmsg);
            return $ClientUtil::FALSE;
        }
        if ($dm_loaded == $ClientUtil::FALSE) {
            $self->_set_err("dm_multipath still  not loaded after modprobing (dm-multipath)");
            return $ClientUtil::FALSE;
        }
    }

    # Currently there is bug in RHEL6 and OL6 in creating bindings file automatically by
    # multipath in case of root on multipath. As a workaround, we need to manually
    # add the bindings to the file /etc/multipath/bindings during host configuration
    # The format of the binding for the root dm-multipath device is:
    # <Alias><space><wwid>
    # For Example:
    #------------------------------------------------------------------------------------
    # mpatha 3600d0230000000000e13955cc3757801
    #------------------------------------------------------------------------------------
    # This issue is only when user_friendly_name set to 1 in multipath.conf which is default
    # setting in RHEL6 or OL6
    if ($client_obj->{major_revision} =~ /rhel6|ol6/i and $client_obj->{config_dmmp_use_user_friendly_names} =~ /^1$|^yes$/i) {
        # Check if this client has root(/) on multipath-ed device
        if (not defined $self->{root_mp_devices}) {
            $self->{root_mp_devices} = $client_obj->get_root_dev();
            return $ClientUtil::FALSE if ($self->error);
        }

        my @mp_root_dev = @{$self->{root_mp_devices}};
        if (scalar(@mp_root_dev) > 0) {
            # Delete binding file to get rid of old entries
            $cmd = "rm -f /etc/multipath/bindings";
            ($status, @output) = &ClientUtil::execute($conn, $cmd);
            if ($status > 0 and (not grep(/no\s+such\s+file/, @output))) {
                $self->_set_err("Failed to delete binding file (/etc/multipath/bindings)");
                return $ClientUtil::FALSE;
            }

            # Create empty binding file
            $cmd = "touch /etc/multipath/bindings";
            ($status, @output) = &ClientUtil::execute($conn, $cmd);
            if ($status > 0) {
                $self->_set_err("Failed to create empty binding file(/etc/multipath/bindings)");
                return $ClientUtil::FALSE;
            }

            # Add the entry for the root device to the binding file
            foreach my $alias (@mp_root_dev) {
                my ($wwid);
                if ($alias =~ /mpath/) {
                    $cmd = "/sbin/multipath -ll $alias | grep $alias";
                    ($status, @output) = &ClientUtil::execute($conn, $cmd);
                    if ($status > 0) {
                        $self->_set_err("Failed to run multipath command");
                        return $ClientUtil::FALSE;
                    }
                    # Grep wwid of the multipathed device
                   ($wwid) = ($output[0] =~ /(360a9\w+|3600a09\w+)/i);
                } else {
                    # else alias is wwid
                    $wwid = $alias;
                }
                $self->_create_binding_entry_for_root_dev(-wwid => $wwid);
                if ($self->error) {
                    return $ClientUtil::FALSE;
                }
            }
        }
    }

    # FLUSH the old mutipath devices entries
    if ($client_obj->{major_revision} !~ /cxs/i) {
        $self->flush(-conn => $conn);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    my $alua_support = $client_obj->get_alua_support();
    if ($client_obj->error) {
        $self->_set_err($client_obj->errmsg);
        return $ClientUtil::FALSE;
    }

    # Need to load extra driver scsi_dh_alua if OS is SLES11 SP1 and ALUA is enabled
    if ($alua_support == 1 and $client_obj->{major_revision} =~ /sles11/i and $client_obj->{minor_revision} == 1) {
        $cmd = "modprobe scsi_dh_alua";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Error while running cmd: $cmd");
            return $ClientUtil::FALSE;
        }
    }

    my $action;
    # Reload/Restart multipath daemon for the new changes
    if ($client_obj->{major_revision} =~ /sles/i) {
        $action = "restart";
    } else {
        if ($client_obj->{sanbooted} == 1) {
            $action = "reload";
        } else {
            $action = "restart";
        }
    }

    if ($client_obj->{major_revision} =~ /rhel7|fedora/) {
        $cmd = "systemctl $action multipathd.service";
    } else {
        $cmd = "/etc/init.d/multipathd $action";
    }

    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status != 0) {
        $self->_set_err("Error while running cmd:($cmd) output:(@output)");
        return $ClientUtil::FALSE;
    }

    # Run multipath. Running multipath with increase verbosity will help in debugging
    $cmd = "/sbin/multipath -v3";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status != 0) {
        $self->_set_err("Error while running cmd:($cmd) output:(@output)");
        return $ClientUtil::FALSE;
    }

    # Need to make chkconfig on for nultipathd. It will start the daemon during bootup
    if ($client_obj->{major_revision} =~ /rhel7|fedora/) {
        $cmd = "systemctl enable multipathd.service";
    } else {
        $cmd = "/sbin/chkconfig multipathd on";
    }
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error while running \'$cmd\'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: recreate_maps()
# Objective:       Will flush old multipath devices and recreate the maps
# Details:         See POD documentation at the end of this file
################################################################################
sub recreate_maps {
    my ($self, %args) = @_;

    my ($hosttype, $cmd, $blacklist_block, $status, @output, $conn, $client_obj);

    $client_obj = $self->{parent}->{parent}->{parent};
    $hosttype = $client_obj->hosttype();

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn' => %args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (not defined($conn)) {
        $self->_set_err("'-conn' parameter is required.");
        return $ClientUtil::FALSE;
    }

    # Flush the maps before creating again
    $self->flush(-conn => $conn);
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    # Run multipath. Running multipath with increase verbosity will help in debugging
    $cmd = "multipath -v3";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error while running \'$cmd\'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       stops the multipath daemon, remove mutlipath kernel modules
#                  (dm_*) and blacklist all the devices.
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;
    my ($hosttype, $cmd, $blacklist_block, $status, @output, $conn, $client_obj);

    $client_obj = $self->{parent}->{parent}->{parent};
    $hosttype = $client_obj->hosttype();

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Flush all the entries from /dev/mapper
    $self->flush(-conn => $conn);
    if($self->error) {
        return $ClientUtil::FALSE;
    }

    # If it is SANbooted host, don't do the following steps and simply return
    if ($client_obj->{sanbooted} == 1) {
        return $ClientUtil::TRUE;
    }


    if ($client_obj->{major_revision} !~ /rhel4/i) {
        $cmd = "echo \'MULTIPATH=NO\' > /etc/sysconfig/mkinitrd/multipath";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Error while running \'$cmd\'");
            return $ClientUtil::FALSE;
        }
    }

    # if the multipath service is running, turn it off.
    my $multipathd_status = $self->_is_multipathd_running(-conn => $conn);
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    if ($multipathd_status == $ClientUtil::TRUE) {
        $self->stop_multipath_daemon();
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    $cmd = "rm -f /etc/multipath.conf.tmp";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to remove /etc/multipath.conf.tmp");
        return $ClientUtil::FALSE;
    }

    $cmd = "touch /etc/multipath.conf.tmp";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to create /etc/multipath.conf.tmp");
        return $ClientUtil::FALSE;
    }

    $cmd = "cat /etc/multipath.conf";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to cat /etc/multipath.conf");
        return $ClientUtil::FALSE;
    }

    # string to add to black list all devices
    my $blacklist = "       devnode \"*\" # automation_blacklisted_all_devices\n";

    # look to see if the file already contains the blacklist of all
    # if it does not, add, if it does, skip.
    if (grep(/automation_blacklisted_all_devices/i, @output) == 0) {
        foreach my $line (@output) {
            $cmd = "echo \'$line\' >> /etc/multipath.conf.tmp";
            my ($status, @output_edit) = &ClientUtil::execute($conn, $cmd);
            if ($status > 0) {
                $self->_set_err("Unable to append $line to /etc/multipath.conf.tmp");
                return $ClientUtil::FALSE;
            }
            if ($line =~ /^\s*devnode_blacklist\s*\{/ or
                $line =~ /^\s*blacklist\s*\{/) {
                $cmd = "echo \'$blacklist\' >> /etc/multipath.conf.tmp";
                my ($status, @output_cp) = &ClientUtil::execute($conn, $cmd);
                if ($status > 0) {
                    $self->_set_err("Unable to append $line to /etc/multipath.conf.tmp");
                    return $ClientUtil::FALSE;
                }
            }
        }

        $cmd = "mv -f /etc/multipath.conf.tmp /etc/multipath.conf";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to override /etc/multipath.conf");
            return $ClientUtil::FALSE;
        }
    } else {
        $self->sanlog(-msg => "Blacklist file already contains blacklist of all devices");
    }

    if ($client_obj->{major_revision} =~ /rhel7|fedora/i) {
        $cmd = "systemctl disable multipathd.service";
    } else {
        $cmd = "/sbin/chkconfig multipathd off";
    }
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error while running \'$cmd\'");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: flush()
# Objective:       Flush the multipath device entries from /dev/mapper and destroy
#                  the mutlipath maps
#                  NOTE: Before calling this method, you must unmount any
#                  multipathed devices.
# Details:         See POD documentation at the end of this file
################################################################################
sub flush {
    my ($self, %args) = @_;
    my ($cmd, $status, @output, $conn, $client_obj, $device);

    $client_obj = $self->{parent}->{parent}->{parent};

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Invalid value for argument '-device' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Check if device argument is passed
    if (defined($device)) {
        # flush the multipathed device which is passed
        $cmd = "multipath -f $device";
        ($status, @output) = &ClientUtil::execute($client_obj->{conn}, $cmd);
        if ($status != 0) {
            if (grep /map\s+in\s+use/i, @output) {
                $self->_set_err("Error while flushing the dm device $device. " .
                                "Make sure that the device is not in use");
            } else {
                $self->_set_err("Command ($cmd) returned bad status");
            }
            return $ClientUtil::FALSE;
        }
        return $ClientUtil::TRUE;
    } else {
        if (!defined $self->{root_mp_devices}) {
            $self->{root_mp_devices} = $client_obj->get_root_dev();
            return $ClientUtil::FALSE if ($self->error);
        }
        my @mp_root_dev = @{$self->{root_mp_devices}};

        # Grepping all the multipath devices which start with the name 360a9|3600a09 and mpath
        $cmd = "ls /dev/mapper/* | grep -P \"360a9|3600a09|mpath\"";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status > 0 and scalar(@output) != 0) {
            $self->_set_err("No multipathed devices were found");
            return $ClientUtil::FALSE;
        }

        # This will remove the all the dm devices used by root in case of SANBoot from the list
        # of discovered dm devices
        my @mp_devices;
        foreach my $line (@output) {
            my ($dev) = ($line =~ /([a-z0-9]{33}|mpath\w+).*/i);
            # remove partition designation - want to skip all partitions on root device
            # In RHEL and OL partition ends with pxx. In SLES it ends with _partxx
            $dev =~ s/p\d+//;
            $dev =~ s/_part\d+//;
            if (grep(/$dev/, @mp_root_dev)) {
                next;
            } else {
                push(@mp_devices, $dev);
            }
        }

        # try and flush the multipath
        # if the host is SANBooted ( and / is on multipath device) then we can not flush the scsi
        # device corresponding to the /. So we need to flush individual scsi devices.
        foreach my $dev (@mp_devices) {
            $cmd = "multipath -f $dev";
            ($status, @output) = &ClientUtil::execute($conn, $cmd);
            if ((grep /map\s+in\s+use/i, @output) or $status > 1) {
                $self->_set_err("Error while flushing the dm device $dev. " .
                                "Make sure all the file systems are unmounted");
                return $ClientUtil::FALSE;
            }
        }

        return $ClientUtil::TRUE;;
    }
}

################################################################################
# Subroutine Name: reconfigure()
# Objective:       It will reconfigure multipath maps
# Details:         See POD documentation at the end of this file
################################################################################
sub reconfigure {
    my ($self) = @_;
    my ($cmd, $status, @output, $client_obj);

    $self->_clear_err();

    $client_obj = $self->{parent}->{parent}->{parent};

    # The first thing we should do is turn on the multipathd service.
    $self->start_multipath_daemon();
    if ($self->error) {
        $self->_set_err("multipathd is not running and not able to start");
        return $ClientUtil::FAIL;
    }

    $cmd = "multipathd -k\"reconfigure\"";
    ($status, @output) = &ClientUtil::execute($client_obj->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("cmd ($cmd) returned bad status");
        return $ClientUtil::FAIL;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_all_partitioned_devices()
# Objective        Will return all multipathed devices which has One or
#                  more partitions
# Details:         See POD documentation at the end of this file
################################################################################
sub get_all_partitioned_devices {
    my ($self, %args) = @_;
    my ($cmd, $status, @output, $conn, @mp_devices);

    $conn = $self->{conn};
    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn' => $args{$_}");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    $cmd = "ls /dev/mapper/* | grep -P \"360a9|3600a09|mpath\"";
    ($status, @mp_devices) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0 and scalar(@mp_devices) != 0) {
        $self->_set_err("Unable to write to /tmp/part.txt", @output);
        return (wantarray()) ? () : [];
    }

    my @partitioned_devices = ();
    foreach my $dev (@mp_devices) {
        if ($dev =~ /([0-9a-z]{33}|mpath)p\w+/) {
            push(@partitioned_devices, $1);
        }
    }

    return (wantarray()) ? @partitioned_devices : \@partitioned_devices;
}

################################################################################
# Subroutine Name: destroy_partition()
# Objective        Will destroy device partition
# Details:         See POD documentation at the end of this file
################################################################################
sub destroy_partition {
    my ($self, %args) = @_;
    my ($cmd, $status, @output, $conn, $device);

    $conn = $self->{conn};
    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Invalid value for argument '-device' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # text file holds input to fdisk command
    $cmd = 'printf "d\\nw"' . " > /tmp/part.txt";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to write to /tmp/part.txt", @output);
        return $ClientUtil::FALSE;
    }

    # deleting partition using fdisk
    $cmd = "fdisk /dev/mapper/" . $device . " < /tmp/part.txt";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 1) {
        $self->_set_err("Unable to destroy partition on device '$device'", @output);
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_size()
# Objective        It will return the size of multipath device
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self, $mpath_device) = @_;
    my ($cmd, $status, @output, $conn);

    my $clientObj =  $self->{parent}->{parent}->{parent};
    # get connection from client object
    $conn = $clientObj->{conn};

    $cmd = "multipath -ll -d $mpath_device";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status != 0) {
        $self->_set_err("Error while calling running cmd ($cmd)");
        return "";
    }

    foreach my $line (@output) {
        if ($line =~ /size\=([0-9A-Za-z.]+)/i) {
            return $1;
        }
    }

    # if we are here, then we are not able to find size
    $self->_set_err("Not able to find size from ($cmd)");
    return "";
}

################################################################################
# Subroutine Name: stop_multipath_daemon()
# Objective        Will stop the multipath daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub stop_multipath_daemon {
    my ($self) = @_;
    my ($cmd, $status, @output, $conn);

    $conn = $self->{conn};

    my $client_obj =  $self->{parent}->{parent}->{parent};

    if ($client_obj->{major_revision} =~ /rhel7|fedora/) {
        $cmd = "systemctl stop multipathd.service";
    } else {
        $cmd = "/etc/init.d/multipathd stop";
    }
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0 and scalar(@output) != 0) {
        # take care of case where you are SANBooted and Multipathed
        # and you cannot stop the MP daemon
        if (grep(/Root\s+is\s+on\s+a\s+multipathed\s+device/i, @output) == 0) {
            $self->_set_err("Failed to stop the multipath daemon");
            return $ClientUtil::FALSE;
        } else {
            $self->sanlog(-msg => "Cannot stop multipath daemon because host is SANBooted");
            return $ClientUtil::TRUE;
        }
    }

    # for sles hosts we need to stop the boot.multipath also
    if ($client_obj->{major_revision} =~ /sles/i) {
        $cmd = "/etc/init.d/boot.multipath stop";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Error while running \'$cmd\'");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: start_multipath_daemon()
# Objective        Will start the multipath daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub start_multipath_daemon {
    my ($self) = @_;
    my ($cmd, $status, @output, $conn);

    $self->_clear_err();

    $conn = $self->{conn};

    my $client_obj =  $self->{parent}->{parent}->{parent};

    if ($client_obj->{major_revision} =~ /rhel7|fedora/) {
        $cmd = "systemctl start multipathd.service";
    } else {
        $cmd = "/etc/init.d/multipathd start";
    }
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to start the multipath daemon");
        return $ClientUtil::FALSE;
    }

    # for sles hosts we need to start the boot.multipath also
    if ($client_obj->{major_revision} =~ /sles/i) {
        $cmd = "/etc/init.d/boot.multipath start";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Error while running \'$cmd\'");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify()
# Objective:       verify proper enviroment for DMMP to work properly
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    my ($cmd, $status, @output, $conn, $clientObj);

    $clientObj =  $self->{parent}->{parent}->{parent};

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # verify that device-mapper package is installed
    $cmd = "rpm -q device-mapper";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("device-mapper package is not installed!");
        return $ClientUtil::FALSE;
    }

    # determine hosttype
    my $hosttype = $clientObj->hosttype();

    my $multipath_package;

    # if SUSE, check for multipath-tools
    if ($hosttype =~ /unix-linux-suse/i or $hosttype =~ /unix-linux-sles/i) {
        $multipath_package = "multipath-tools";
    }

    # if RHEL4, check for device-mapper-multipath
    elsif ($hosttype =~ /unix-linux/i) {
        $multipath_package = "device-mapper-multipath";
    }

    $cmd = "rpm -q $multipath_package";

    # verify that device-mapper-multipath package is installed
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("$multipath_package package is not installed!");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Suroutine Name: _get_default_multipath_params
# Objectiive:     Returns hash reference contains default parameters belong to
#                 devices and defaults sections of /etc/multipath.conf file
################################################################################
sub _get_default_multipath_params {
    my ($self) = @_;

    my ($cmd, @output, $status, $conn, $clientObj, $mpath_params_hash, $protocol);

    $self->_clear_err();

    # get the client object
    $clientObj =  $self->{parent}->{parent}->{parent};

    # get connection from the client object
    $conn = $clientObj->{conn};

    # get protocol from client object
    $protocol = $clientObj->{protocol};

    $mpath_params_hash = {};

    if ($clientObj->{major_revision} =~ /rhel7|fedora/i
       or ($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 4)) {
        # Just return as these OSes don't require devices and defaults sections
        return $mpath_params_hash;
    }

    # DEFAULTS BLOCK PARAMS
    unless (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)) {
        # No need to have this parameter from RHEL6.3, default value will be taken from
        # hardware table
        $mpath_params_hash->{defaults}->{user_friendly_names} = "no";
    }

    if ($clientObj->{major_revision} =~ /rhel4\.7|rhel5\.2/i) {
        $mpath_params_hash->{defaults}->{max_fds} = 4096;
    } elsif ($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} == 0) {
        $mpath_params_hash->{defaults}->{max_fds} = 8192;
    } elsif ($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3) {
        # No need to have this parameter from RHEL6.3, default value will be taken from
        # hardware table
    } elsif (($clientObj->{major_revision} =~ /rhel4/i and $clientObj->{minor_revision} >= 8)
       or ($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} >= 1)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /rhel6|ol6/i)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} <= 1)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 6)
       or ($clientObj->{major_revision} =~ /sles10/i and $clientObj->{minor_revision} >= 4)) {
        $mpath_params_hash->{defaults}->{max_fds} = "max";
    }

    if ($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3) {
        # No need to have this parameter from RHEL6.3, default value will be taken from
        # hardware table
    } elsif (($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /rhel6|ol6/i)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} <= 1)
       or ($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} >= 2)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 6)) {
        $mpath_params_hash->{defaults}->{queue_without_daemon} = "no";
    }

    if ($clientObj->{major_revision} =~ /rhel4/i and $clientObj->{minor_revision} <= 7) {
        $mpath_params_hash->{defaults}->{rr_min_io} = 128;
    }

    if ($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3) {
        # No need to have this parameter from RHEL6.3, default value will be taken from
        # hardware table
    } elsif (($clientObj->{major_revision} =~ /rhel4/i and $clientObj->{minor_revision} >= 7)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 6)
       or ($clientObj->{major_revision} =~ /rhel6|ol6/i)
       or ($clientObj->{major_revision} =~ /sles10/i and $clientObj->{minor_revision} >= 4)
       or ($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} >= 1)) {
        $mpath_params_hash->{defaults}->{flush_on_last_del} = "yes";
    } elsif ($clientObj->{major_revision} =~ /cxs6/i) {
        $mpath_params_hash->{defaults}->{flush_on_last_del} = "no";
    }

    if ($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} == 1) {
        $mpath_params_hash->{defaults}->{dev_loss_tmo} = 2147483647;
    } elsif ($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} == 2) {
        $mpath_params_hash->{defaults}->{dev_loss_tmo} = "infinity";
    } elsif ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2) {
        $mpath_params_hash->{defaults}->{dev_loss_tmo} = "30";
    }

    if ($clientObj->{major_revision} =~ /rhel6|ol6/i and
       ($clientObj->{minor_revision} == 1 or $clientObj->{minor_revision} == 2)) {
        $mpath_params_hash->{defaults}->{fast_io_fail_tmo} = 5;
    } elsif ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2) {
        $mpath_params_hash->{defaults}->{fast_io_fail_tmo} = "off";
    }

    # OL5.6 and RHEL5.6 onwards a new multipath.conf setting 'pg_prio_calc' has been
    # introduced. By default this is set to 'sum'. Setting it to 'avg' causes
    # dm-multipath to calculate priorities using the average priority of paths within
    # a group - which is what is preferred when using NetApp LUNs.
    # No need to have this param from OL5.8, because default value will be taken hardware table.
    if (($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 6)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} <= 1)
       or ($clientObj->{major_revision} =~ /ol5/i and
       ($clientObj->{minor_revision} == 6 or $clientObj->{minor_revision} == 7))) {
        $mpath_params_hash->{defaults}->{pg_prio_calc} = "avg";
    }

    # DEVICES BLOCK PARAMS
    # From RHEL 6.2 onwards, no need to have some of the parameters in multipath.conf file.
    # Because the default value for these parameters will be taken from hardware
    # tables automatically.

    my $path = $clientObj->get_hu_bin_path();
    if ($clientObj->error()) {
        $self->_set_err($clientObj->errmsg());
        return $mpath_params_hash;
    }

    my $alua_support = $clientObj->get_alua_support();
    if ($clientObj->error) {
        $self->_set_err($clientObj->errmsg());
        return $mpath_params_hash;
    }

    if (($alua_support == 1) and
       (($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} >= 1)
       or ($clientObj->{major_revision} =~ /sles10/i and $clientObj->{minor_revision} >= 4)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 5)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 6)
       or ($clientObj->{major_revision} =~ /rhel6|ol6|cxs6/))) {
        $mpath_params_hash->{devices}->{hardware_handler} = "\"1 alua\"";
    } elsif (($alua_support == 0) and
       (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9))) {
        # No need to have this parameter from RHEL6.3 and RHEL5.9 for non-alua configs, default value
        #will be taken from hardware table
    } else {
        $mpath_params_hash->{devices}->{hardware_handler} = "\"0\"";
    }

    if ($alua_support == 1) {
        # make necessary changes to callout program if ALUA_SUPPORT==1
        if (($clientObj->{major_revision} =~ /sles10/i and $clientObj->{minor_revision} >= 2)
           or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 6)
           or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
           or ($clientObj->{major_revision} =~ /sles11/i)
           or ($clientObj->{major_revision} =~ /rhel6|ol6/i)) {
            $mpath_params_hash->{devices}->{prio} = "\"alua\"";
        } else {
            $mpath_params_hash->{devices}->{prio_callout} = "\"\/sbin\/mpath_prio_alua \/dev\/\%n\"";
        }
    } else {
        if (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
           or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9)) {
            # No need to have this parameter from RHEL6.3, RHEL5.9 for non-alua configs, default value
            # will be taken from hardware table
        } elsif (($clientObj->{major_revision} =~ /sles10/i and $clientObj->{minor_revision} >= 2)
           or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 6)
           or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
           or ($clientObj->{major_revision} =~ /sles11/i)
           or ($clientObj->{major_revision} =~ /rhel6|ol6/i)) {
            $mpath_params_hash->{devices}->{prio} = "ontap";
        } elsif ($clientObj->{major_revision} =~ /rhel4|sles9/i) {
            $cmd = "ls /sbin/mpath_prio_netapp";
            ($status, @output) = &ClientUtil::execute($conn, $cmd);
            if ($status != 0) {
                $self->_set_err("file /sbin/mpath_prio_netapp doesnot exist. " .
                                "please make sure mpath_prio_netapp exists in /sbin dir");
                return $mpath_params_hash;
            }
            $mpath_params_hash->{devices}->{prio_callout} = "\"\/sbin\/mpath_prio_netapp \/dev\/\%n\"";
        } else {
            $cmd = "ls /sbin/mpath_prio_ontap";
            ($status, @output) = &ClientUtil::execute($conn, $cmd);
            if ($status > 0) {
                $cmd = "cp -f " . $path . "/mpath_prio_ontap /sbin/mpath_prio_ontap";
                ($status, @output) = &ClientUtil::execute($conn, $cmd);
                if ($status > 0) {
                    $self->_set_err("Unable to copy mpath_prio_ontap " .
                                    "from $path to /sbin. Copy it manually and rerun");
                    return $mpath_params_hash;
                }
            }
            $mpath_params_hash->{devices}->{prio_callout} = "\"\/sbin\/mpath_prio_ontap \/dev\/\%n\"";
        }
    }

    if (($alua_support == 0) and
       (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9))) {
        # No need to have this parameter from RHEL6.3, RHEL5.9 with non-alua configs, because devices
        # section is not required
    } elsif (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 8)) {
        # From RHEL6.3, OL6.3 and OL5.8 onwards product should be "LUN.*" instead of LUN
        $mpath_params_hash->{devices}->{product} = "\"LUN.*\"";
    } else {
        $mpath_params_hash->{devices}->{product} = "\"LUN\"";
    }

    if (($alua_support == 0) and
       (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9))) {
        # No need to have this parameter from RHEL6.3, RHEL5.9 with non-alua configs, because devices
        # section is not required
    } else {
        $mpath_params_hash->{devices}->{vendor} = "\"NETAPP\"";
    }

    if ($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} < 3) {
        # No need to have this parameter from RHEL6.3 and OL6.3, default value will be taken from
        # hardware table
        $mpath_params_hash->{devices}->{path_selector} = "\"round-robin 0\"";
    }

    if (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 8)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9 and $alua_support == 0)) {
        # No need to have this parameter from RHEL6.3, RHEL5.9(non-alua configs), OL6.3 and OL5.8, default value
        # will be taken from hardware table
    } elsif (($clientObj->{major_revision} =~ /sles9/i)
       or ($clientObj->{major_revision} =~ /rhel4/i and $clientObj->{minor_revision} <= 6)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} == 0)) {
        $mpath_params_hash->{devices}->{path_checker} = "readsector0";
    } elsif (($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} >= 1)
       or ($clientObj->{major_revision} =~ /sles10/i and $clientObj->{minor_revision} >= 4)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 7)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} <= 1)
       or ($clientObj->{major_revision} =~ /rhel6|ol6/i and
       ($clientObj->{minor_revision} == 1 or $clientObj->{minor_revision} == 2))) {
        $mpath_params_hash->{devices}->{path_checker} = "tur";
    } else {
        $mpath_params_hash->{devices}->{path_checker} = "directio";
    }

    if (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 8)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9 and $alua_support == 0)) {
        # No need to have this parameter from RHEL6.3, RHEL5.9(non-alua configs), OL6.3 and OL5.8, default value
        # will be taken from hardware table
    } elsif (($clientObj->{major_revision} =~ /rhel4/i and $clientObj->{minor_revision} >= 8)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 4)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 6)
       or ($clientObj->{major_revision} =~ /rhel6|ol6|cxs6/i)
       or ($clientObj->{major_revision} =~ /sles10/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /sles11/i)) {
        $mpath_params_hash->{devices}->{path_grouping_policy} = "group_by_prio";
    } elsif ($alua_support == 1 or $protocol =~ /fcp/i) {
        $mpath_params_hash->{devices}->{path_grouping_policy} = "group_by_prio";
    } elsif ($protocol =~ /iscsi/i) {
        $mpath_params_hash->{devices}->{path_grouping_policy} = "multibus";
    }

    unless (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 8)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9 and $alua_support == 0)) {
        # No need to have these parameters from RHEL6.3, RHEL5.9(non-alua configs), OL6.3 and OL5.8, default value
        # will be taken from hardware table
        $mpath_params_hash->{devices}->{failback} = "immediate";
        $mpath_params_hash->{devices}->{rr_weight} = "uniform";
    }

    if (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 8)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9 and $alua_support == 0)) {
       # No need to have this parameter from RHEL6.3, RHEL5.9(non-alua configs), OL6.3 and OL5.8, default value
       # will be taken from hardware table
    } elsif (($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 7)
       or ($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} >= 2)
       or ($clientObj->{major_revision} =~ /rhel6|ol6/i and
       ($clientObj->{minor_revision} == 1 or $clientObj->{minor_revision} == 2))) {
        $mpath_params_hash->{devices}->{features} = "\"3 queue_if_no_path pg_init_retries 50\"";
    } else {
        $mpath_params_hash->{devices}->{features} = "\"1 queue_if_no_path\"";
    }

    if (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 8)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9 and $alua_support == 0)) {
        # No need to have this parameter from RHEL6.3, RHEL5.9(non-alua configs), OL6.3 and OL5.8, default value
        # will be taken from hardware table
    } elsif (($clientObj->{major_revision} !~ /rhel4/i)
       or ($clientObj->{major_revision} =~ /rhel4/i and $clientObj->{minor_revision} >= 8)) {
        $mpath_params_hash->{devices}->{rr_min_io} = 128;
    }

    if (($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 8)
       or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
       or ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9 and $alua_support == 0)) {
        # No need to have this parameter from RHEL6.3, RHEL5.9(non-alua configs), OL6.3 and OL5.8, default value
        # will be taken from hardware table
    } elsif ($clientObj->{major_revision} =~ /sles11|rhel6|ol6/i) {
        $mpath_params_hash->{devices}->{getuid_callout} = "\"\/lib\/udev\/scsi_id -g -u -d \/dev\/\%n\"";
    } else {
        $mpath_params_hash->{devices}->{getuid_callout} = "\"\/sbin\/scsi_id -g -u -s \/block\/\%n\"";
    }

    return $mpath_params_hash;
}

################################################################################
# Subroutine Name: config()
# Objective:       Configure a proper enviroment for DMMP. In other words,
#                  configure the /etc/multipath.conf file.
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, $conn, $protocol, $clientObj);
    my (@multipath_conf, $syntax_errors, $param);

    $clientObj =  $self->{parent}->{parent}->{parent};

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
        if (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-protocol' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    ############
    # STEP ONE #
    ############
    # all the values left in the scsi hash should go into the blacklist
    # Now we use wwid to blacklist all non-netapp scsi devices. Modifying the code to do the same
    my @wwid_scsi_blacklist;
    my @devnode_scsi_blacklist;
    $cmd = "ls /sys/block | grep sd.*";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status == 1 and scalar(@output) == 0) {
        # No matching entry found
        $self->sanlog(-msg => "No SCSI device found");
    } elsif ($status != 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }
    foreach my $line (@output) {
        my $wwid = $clientObj->_get_wwid($line);
        if ($clientObj->error()) {
            # if scsi_id fails and there is no ID for the device, it might not be netapp
            # we'll have to check the vendor to be sure it isnt netapp.
            # if it isnt, then we'll push it on the devnode scsi blacklist
            $clientObj->_clear_err();
            $cmd = "cat /sys/block/$line/device/vendor";
            my ($vendor_status, @vendor_output) = &ClientUtil::execute($conn, $cmd);
            if ($vendor_status > 0) {
                $self->_set_err("Unable to determine vendor for device '$line'");
                return $ClientUtil::FALSE;
            }

            # if netapp is not the vendor, push it on the devnode blacklist and go to the next device
            if (not grep(/NETAPP/i, @vendor_output)) {
                push(@devnode_scsi_blacklist, $line);
                next;
            } else {
                # netapp is the vendor, dont blacklist this
                next;
            }
        }

        if ($wwid =~ /^(360a9|3600a09)/) {
            next;
        } else {
            push(@wwid_scsi_blacklist, $wwid);
        }
    }

    # lets build a string that delimits each blacklisted device in this
    # format: e.g wwid "(1HITACHI_HUS103073FL3800_V3W8GV6A | 1HITACHI_HUS103073FL3800_V3W72GWA)"
    my $scsi_blacklist_string ;
    if (scalar(@wwid_scsi_blacklist) > 0) {
        $scsi_blacklist_string = "    wwid (" . join("|", @wwid_scsi_blacklist) . ")";
    } else {
        $scsi_blacklist_string = "";
    }

    # add the non netapp scsi blacklist entries to the list
    # format: devnode "^sda$|^sdb$"
    if (scalar(@devnode_scsi_blacklist)) {
        if ($scsi_blacklist_string =~ /^\s*$/) {
            $scsi_blacklist_string .= "    devnode \"^" . join("\$|^", @devnode_scsi_blacklist) . "\$\"";
        } else {
            # add a newline if its non empty already
            $scsi_blacklist_string .= "\n    devnode \"^" . join("\$|^", @devnode_scsi_blacklist) . "\$\"";
        }
    }

    if (defined($scsi_blacklist_string)) {
        $self->sanlog(-msg => "Blacklist string is:\n$scsi_blacklist_string");
    }

    my $ide_blacklist_string = "    devnode \"^cciss.*\"\n";

    unless ($clientObj->{major_revision} =~ /rhel5/i and $clientObj->{minor_revision} >= 9) {
        $ide_blacklist_string .=  "    devnode \"^hd[a-z]\"\n" .
                                  "    devnode \"^(ram|raw|loop|fd|md|dm-|sr|scd|st)[0-9]*\"\n";
    }

    ############
    # STEP TWO #
    ############
    # We now have to make sure that the multipath.conf file on this
    # host is compatible! Bear with me on this!

    # Make sure multipath.conf exists on the host
    $cmd = "ls /etc/multipath.conf";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status != 0) {
        @multipath_conf = ();
    } else {
        # lets open up the multipath.conf file on the host
        $cmd = "cat /etc/multipath.conf";
        ($status, @multipath_conf) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("cmd ($cmd) returned a status of ($status)");
            return $ClientUtil::FALSE;
        }
    }

    foreach my $line (@multipath_conf) {

        # Remove all comments from the multipath_conf file because they will cause
        # problems otherwise.
        $line =~ s/^#[^_][^#][^_]//;

        # will remove if there is any device that we black listed using wwid
        $line =~ s/.*wwid.*//;

        # Remove devnode entries
        $line =~ s/.*devnode.*//;

        # also make sure that all '\t' tab characters are converted to 4 spaces
        $line =~ s/\t/    /g;

        # and remove all carriage returns
        $line =~ s/\r//g;
    }

    # Just in case, chomp the newlines off of every element in the output and
    # join the elements together into a string, delimited by newlines.
    chomp(@multipath_conf);
    my $multipath_conf_string = join "\n", @multipath_conf;

    # convert two or more blank lines into just one blank line to save space.
    $multipath_conf_string =~ s/^\s*\n\s*\n/\n/mg;

    # Now we have to parse through and modify the $multipath_conf_string
    # These will be our sentinels for our modifications.
    my $begin_sentinel = "#_#_ BEGIN multipath.conf AUTOMATION ##";
    my $end_sentinel   = "#_#_ END multipath.conf AUTOMATION ##";

    # get default parameters
    my $mpath_params_hash = $self->_get_default_multipath_params();
    if ($self->error()) {
        $self->_set_err($self->errmsg());
        return $ClientUtil::FALSE;
    }

    ### DEFAULTS BLOCK MODIFICATIONS

    # Get user passed value from client object
    my $user_friendly_names = $clientObj->{config_dmmp_use_user_friendly_names};
    my $max_fds = $clientObj->{config_dmmp_max_fds};
    my $queue_without_daemon = $clientObj->{config_dmmp_queue_without_daemon};
    my $flush_on_last_del = $clientObj->{config_dmmp_flush_on_last_del};
    my $pg_prio_calc = $clientObj->{config_dmmp_pg_prio_calc};
    my $dev_loss_tmo = $clientObj->{config_dmmp_dev_loss_tmo};
    my $fast_io_fail_tmo = $clientObj->{config_dmmp_fast_io_fail_tmo};

    # populate 'defaults' section of multipath.conf file.
    # Go through the default parameters, check if user want to
    # override any param, then update the param with user passed value.
    # otherwise update the param with default value.

    $param = "user_friendly_names";
    if ($user_friendly_names =~ /^1$|^yes$/i) {
        if ($mpath_params_hash->{defaults}->{$param} !~ /$user_friendly_names/) {
            &SanUtil::nate_log('-warn', "Default default_value of multipath parameter($param) " .
                                        "has been overridden with the value(yes)");
        }
        $mpath_params_hash->{defaults}->{$param} = "yes";
    }

    $param = "max_fds";
    if ($max_fds !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{defaults}->{$param} !~ /$max_fds/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($max_fds)");
        }
        $mpath_params_hash->{defaults}->{$param} = $max_fds;
    }

    $param = "queue_without_daemon";
    if ($queue_without_daemon !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{defaults}->{$param} !~ /$queue_without_daemon/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($queue_without_daemon)");
        }
        $mpath_params_hash->{defaults}->{$param} = $queue_without_daemon;
    }

    $param = "flush_on_last_del";
    if ($flush_on_last_del !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{defaults}->{$param} !~ /$flush_on_last_del/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($flush_on_last_del)");
        }
        $mpath_params_hash->{defaults}->{$param} = $flush_on_last_del;
    }

    $param = "pg_prio_calc";
    if ($pg_prio_calc !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{defaults}->{$param} !~ /$pg_prio_calc/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($pg_prio_calc)");
        }
        $mpath_params_hash->{defaults}->{$param} = $pg_prio_calc;
    }

    $param = "dev_loss_tmo";
    if ($dev_loss_tmo !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{defaults}->{$param} !~ /$dev_loss_tmo/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($dev_loss_tmo)");
        }
        $mpath_params_hash->{defaults}->{$param} = $dev_loss_tmo;
    }

    $param = "fast_io_fail_tmo";
    if ($fast_io_fail_tmo !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{defaults}->{$param} !~ /$fast_io_fail_tmo/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($fast_io_fail_tmo)");
        }
        $mpath_params_hash->{defaults}->{$param} = $fast_io_fail_tmo;
    }

    my $defaults_block = "";
    foreach $param (keys %{$mpath_params_hash->{defaults}}) {
        $defaults_block .= "    $param" . "     $mpath_params_hash->{defaults}->{$param}\n";
    }

    chomp($defaults_block);

    if ($defaults_block !~ /^\s*$/) {
        $defaults_block = "$begin_sentinel\n$defaults_block\n$end_sentinel\n";
    }

    ### DEVNODE BLACKLIST BLOCK MODFICATIONS
    my $devnode_blacklist_block = "";

    if (($clientObj->{major_revision} =~ /rhel7|fedora/i)
       or ($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} >= 3)
       or ($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 3)) {
        # We don't need to add any devnode entries in blacklist section for these OSes,
        # only $scsi_blacklist_string need to be added.
        # Hence, blacklist section is not required when $scsi_blacklist_string is not defined.
        if ($scsi_blacklist_string !~ /^\s*$/) {
            # insert the scsi blacklist string
            $devnode_blacklist_block .= "$begin_sentinel\n$scsi_blacklist_string\n$end_sentinel\n";
        }
    } else {
        if ($multipath_conf_string =~ /^\s*blacklist\s*\n*\s*{(.*?)}/sm) {
            $devnode_blacklist_block = $1;
        }

        my $syntax_errors = 0;

        # check for invalid braces in devnode blacklist  block
        $syntax_errors = ($devnode_blacklist_block =~ /{|}/g);

        # if there are syntax errors, then ignore all existing data in this block
        # and build one from scratch
        if ($syntax_errors > 0) {
            $devnode_blacklist_block = "";
            if ($scsi_blacklist_string !~ /^\s*$/) {
                # insert the scsi blacklist sting
                $devnode_blacklist_block .= "$begin_sentinel\n$scsi_blacklist_string\n$end_sentinel\n";
            }
            $devnode_blacklist_block .= "\n$ide_blacklist_string\n";
        }

        # Otherwise, Modify our blacklist string if needed, otherwise concat a new
        # field for it into the blacklist.
        else {
            # first remove the contents of the sentinel
            $devnode_blacklist_block =~ s/${begin_sentinel}.*?${end_sentinel}//gs;

            if ($scsi_blacklist_string !~ /^\s*$/) {
                # insert the scsi blacklist sting
                $devnode_blacklist_block .= "$begin_sentinel\n$scsi_blacklist_string\n$end_sentinel\n";
            }

            # check if there is an entry that blocks the ide disk devices,
            # otherwise, add one to the block
            unless ($devnode_blacklist_block =~ s/(devnode\s+\"\^hd\[)(.*?)(\]\")/$1a-z$3/) {
                $devnode_blacklist_block .= "$ide_blacklist_string\n";
            }
        }

        # by default, the /etc/multipath.conf file has a devnode "*" entry
        # that blocks every device. We must remove it if it exists.
        $devnode_blacklist_block =~ s/devnode\s*\"\*\"//;

        # remove all consecutive newlines
        $devnode_blacklist_block =~ s/\n+/\n/g;
    }

    chomp($devnode_blacklist_block);

    ### DEVICES BLOCK MODIFICATIONS
    my $devices_block = "";
    if ($multipath_conf_string =~ /^\s*devices\s*\n*\s*{\s*\n*\s*(.*)\s*\n*\s*}/ms) {
        $devices_block = $1;
    }

    # Get user passed value from client object
    my $hardware_handler = $clientObj->{config_dmmp_hardware_handler};
    my $prio_pgm = $clientObj->{config_dmmp_prio};
    my $prio_callout = $clientObj->{config_dmmp_prio_callout};
    my $product = $clientObj->{config_dmmp_product};
    my $vendor = $clientObj->{config_dmmp_vendor};
    my $path_selector = $clientObj->{config_dmmp_path_selector};
    my $path_checker = $clientObj->{config_dmmp_path_checker};
    my $grouping_policy = $clientObj->{config_dmmp_path_grouping_policy};
    my $failback_policy = $clientObj->{config_dmmp_failback};
    my $rr_weight = $clientObj->{config_dmmp_rr_weight};
    my $features = $clientObj->{config_dmmp_features};
    my $rr_min_io = $clientObj->{config_dmmp_rr_min_io};
    my $getuid_callout = $clientObj->{config_dmmp_getuid_callout};

    # populate 'devices' section of multipath.conf file.
    # Go through the devices parameters, check if user want to
    # override any param, then update the param with user passed value.
    # otherwise update the param with default value.

    $param = "product";
    if ($product !~ /NOT_SET|^\s*$/i) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$product/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($product)");
        }
        $mpath_params_hash->{devices}->{$param} = "\"$product\"";
    }

    $param = "vendor";
    if ($vendor !~ /NOT_SET|^\s*$/i) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$vendor/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($vendor)");
        }
        $mpath_params_hash->{devices}->{$param} = "\"$vendor\"";
    }

    $param = "path_grouping_policy";
    if ($grouping_policy !~ /NOT_SET|^\s*$/i) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$grouping_policy/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($grouping_policy)");
        }
        $mpath_params_hash->{devices}->{$param} = $grouping_policy;
    }

    $param = "path_selector";
    if ($path_selector !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$path_selector/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($path_selector)");
        }
        $mpath_params_hash->{devices}->{$param} = "\"$path_selector\"";
    }

    $param = "path_checker";
    if ($path_checker !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$path_checker/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($path_checker)");
        }
        $mpath_params_hash->{devices}->{$param} = $path_checker;
    }

    $param = "features";
    if ($features !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$features/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($features)");
        }
        $mpath_params_hash->{devices}->{$param} = "\"$features\"";
    }

    $param = "rr_weight";
    if ($rr_weight !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$rr_weight/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($rr_weight)");
        }
        $mpath_params_hash->{devices}->{$param} = $rr_weight;
    }

    $param = "rr_min_io";
    if ($rr_min_io !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$rr_min_io/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($rr_min_io)");
        }
        $mpath_params_hash->{devices}->{$param} = $rr_min_io;
    }

    $param = "hardware_handler";
    if ($hardware_handler !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$hardware_handler/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($hardware_handler)");
        }
        $mpath_params_hash->{devices}->{$param} = "\"$hardware_handler\"";
    }

    $param = "failback";
    if ($failback_policy !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$failback_policy/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($failback_policy)");
        }
        $mpath_params_hash->{devices}->{$param} = $failback_policy;
    }

    $param = "getuid_callout";
    if ($getuid_callout !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$getuid_callout/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($getuid_callout)");
        }
        $mpath_params_hash->{devices}->{$param} = "\"$getuid_callout\"";
    }

    $param = "prio_callout";
    if ($prio_callout !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$prio_callout/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($prio_callout)");
        }
        $mpath_params_hash->{devices}->{$param} = "\"$prio_callout\"";
    }

    $param = "prio";
    if ($prio_pgm !~ /NOT_SET|^\s*$/) {
        if ($mpath_params_hash->{devices}->{$param} !~ /$prio_pgm/) {
            &SanUtil::nate_log('-warn', "Default value of multipath parameter($param) " .
                                        "has been overridden with the value($prio_pgm)");
        }
        $mpath_params_hash->{devices}->{$param} = "\"$prio_pgm\"";
    }

    my $netapp_device = "";
    foreach $param (keys %{$mpath_params_hash->{devices}}) {
        $netapp_device .= "    $param" . "     $mpath_params_hash->{devices}->{$param}\n";
    }

    chomp($netapp_device);

    if ($netapp_device !~ /^\s*$/) {
        $netapp_device = "$begin_sentinel\n$netapp_device\n$end_sentinel\n";
    }

    # check for syntax errors in the devices blocks,
    $syntax_errors = 0;

    # make sure that the number of left braces and right braces are the same.
    $syntax_errors++ if (($devices_block =~ s/{/{/g) != ($devices_block =~ s/}/}/g));

    # make sure that for each device there are no dangling braces inside each block.
    $syntax_errors += (
        $devices_block =~ /device.*\{
                                          (
                                          [^\}]*\{
                                          |
                                          .*?\}[^\{]*\}
                                          )
                                          /gsx);

    # if there are syntax errors, then ignore all existing data in this block
    # and start from scratch
    if ($syntax_errors != 0) {
        $devices_block = "";

        if ($netapp_device !~ /^\s*$/) {
            $devices_block .= "\ndevice {\n$netapp_device\n}\n";
        }
    }

    # else make the changes to the device block if a netapp_device exists,
    else {
        if ($devices_block =~ /device\s*\n*\s*{\s*\n*\s*.*vendor\s*\"NETAPP\".*?}/msg) {
            if ($netapp_device =~ /^\s*$/) {
                # If netapp_device is empty, delete the existing entries for netapp devices
                $devices_block =~ s/(device\s*\n*\s*{)\s*\n*\s*(.*vendor\s*\"NETAPP\".*?)(})//msg;
            } else {
                # Replace existing netapp device section with netapp_device
                $devices_block =~ s/(device\s*\n*\s*{)\s*\n*\s*(.*vendor\s*\"NETAPP\".*?)(})/$1\n$netapp_device\n$3/msg;
            }
        } else {
            # if we didnt find any netapp devices, then we should concat our
            # netapp_device on the device block string.
            if ($netapp_device !~ /^\s*$/) {
                $devices_block .= "device {\n$netapp_device\n}\n";
            }
        }
    }

    ### REBUILD THE FILE
    # rebuild all the blocks by prepending and appending their respective block
    # headers

    # In some OSes, defaults block might not be needed. Hence, add defaults section
    # only if some entries are there in defaults_block
    if ($defaults_block !~ /^\s*$/) {
        $defaults_block          = "defaults {\n$defaults_block\n}\n";
    }

    if ($devnode_blacklist_block =~ /^\s*$/) {
        # if $devnode_blacklist_block is empty, then no need to have blacklist section.
        # This is only possible from RHEL 6.3 onwards. Because from RHEL 6.3 onwards
        # blacklist section will have only scsi blocklist strings and not devnode entries
        $devnode_blacklist_block = "";
    } elsif ($clientObj->{major_revision} =~ /rhel4|sles9/) {
        $devnode_blacklist_block = "devnode_blacklist {\n$devnode_blacklist_block\n}\n";
    } else {
        $devnode_blacklist_block = "blacklist {\n$devnode_blacklist_block\n}\n";
    }

    # In some OSes, devices block might not be needed. Hence, add devices section
    # only if some entries are there in devices_block
    if ($devices_block !~ /^\s*$/) {
        $devices_block           = "devices {\n$devices_block}\n";
    }

    # put all the pieces back together and split them into a list line by line
    my $new_multipath_conf_string = $defaults_block . $devnode_blacklist_block . $devices_block;

    # In some OSes (e.g., RHEL6.4) multipath settings (defaults and devices section) are not required.
    # But for successful dm-multipath configuration, still we need to provide atleast an empty multipath.conf file.
    if ($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 4) {
        # Just adding a comment saying multipath settings not required
        $new_multipath_conf_string = "### multipath.conf settings not required AUTOMATION ###\n" . $new_multipath_conf_string;
    }

    if ($new_multipath_conf_string =~ /^\s*$/) {
        $self->sanlog(-msg => "multipath.conf setting is not required");
        # if multipath.conf already exists, take backup and remove it
        $cmd = "ls /etc/multipath.conf";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status == 0) {
            $cmd = "mv /etc/multipath.conf /etc/multipath.conf.tmp";
            ($status, @output) = &ClientUtil::execute($conn, $cmd);
            if ($status != 0) {
                $self->_set_err("Cmd failed to backup the multipath.conf file");
                return $ClientUtil::FALSE;
            }
        }
        goto EMPTY_MULTIPATH;
    }

    my @new_multipath_conf = split(/\n/, $new_multipath_conf_string);
    chomp(@new_multipath_conf);

    ##############
    # STEP THREE #
    ##############
    # We will now update the /etc/multipath.conf on the host

    $cmd = "rm -f /etc/multipath.conf.tmp";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed to backup the multipath.conf file");
        return $ClientUtil::FALSE;
    }

    # first back up the previous file on the host by placing a
    # timestamp on the end of the filename
    $cmd = "touch /etc/multipath.conf.tmp";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed to backup the multipath.conf file");
        return $ClientUtil::FALSE;
    }

    # now for each line in our array, append to a new multipath.conf file
    foreach my $line (@new_multipath_conf) {
        $cmd = "echo \'$line\' >> /etc/multipath.conf.tmp";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Not able to appnd $line to multipath.conf");
            return $ClientUtil::FALSE;
        }
    }

    # overwrite the main version of the file, but create a backup of the original
    $cmd = "mv --backup --suffix=\".bak\" /etc/multipath.conf.tmp /etc/multipath.conf";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed to backup the multipath.conf file");
        return $ClientUtil::FALSE;
    }

    # Displaying modified multipath.conf in logs. Will be useful for debugging
    $cmd = "cat /etc/multipath.conf";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed to display multipath.conf file");
        return $ClientUtil::FALSE;
    }

    EMPTY_MULTIPATH:
    my $client_reboot_required = 0;
    my $verify_proc_cmdline = 0;

    if ($clientObj->{major_revision} =~ /rhel6|ol6/i and $clientObj->{minor_revision} >= 4) {
        if ($clientObj->verify_proc_cmdline(-option => "rdloaddriver=scsi_dh_alua") != $ClientUtil::TRUE) {
            # clear the error
            $clientObj->_clear_err();

            $self->sanlog(-msg => "Updating grub kernel parameter. Reboot is required.");
            $self->_update_grub_kernel_param();
            if ($self->error()){
                $self->_set_err("Unable to update grub kernel parameter");
                return $ClientUtil::FALSE;
            }

            # Set the flag to reboot the host. Host will be rebooted once all
            # the changes are done. After the reboot, make sure that scsi_dh_alua is loaded
            $client_reboot_required = 1;
            $verify_proc_cmdline = 1;
        }
    }

    # If the host is sanbooted and user_friendly_names is set to "no",
    # then we need to change the device map names in /etc/fstab and /boot/grub/device.map to the wwid.
    # This change is required only for RHEL and OL. Because in those OSs only, we install OS on
    # multipathed device. In SLES, we don't install OS on multipathed device.

    # Check if this client has root(/) on multipath-ed device
    if (not defined $self->{root_mp_devices}) {
        $self->{root_mp_devices} = $clientObj->get_root_dev();
        return $ClientUtil::FALSE if ($self->error);
    }
    my $sanboot_value = $clientObj->{sanbooted};
    if (($sanboot_value == 1) and ($clientObj->{config_dmmp_use_user_friendly_names} !~ /^1$|^yes$/i) and
       ($clientObj->{major_revision} =~ /rhel|ol|fedora/i)) {
        $self->sanlog(-msg => "Modifing multipath user_friendly_names setting to no. Reboot is required.");
        $self->_dev_map_changes(-conn => $conn);
        if ($self->error()){
           $self->_set_err("Unable to make required changes to /etc/fstab, /boot/grub/grub.conf and /boot/grub/device.map");
           return $ClientUtil::FALSE;
        }

        # Set the flag to reboot the host. Host will be rebooted once all
        # the changes are done.
        $client_reboot_required = 1;
    }

    if ($client_reboot_required == 1) {
        # Reboot the host after made all the changes
        $self->sanlog(-msg => "Rebooting the host to make all the new changes get affected");
        $clientObj->reboot();
        if ($clientObj->error()) {
            $self->_set_err($clientObj->errmsg);
            return $ClientUtil::FALSE;
        }
    }

    if ($verify_proc_cmdline == 1) {
        # Verify whether scsi_dh_alua driver is loaded after the reboot
        if ($clientObj->verify_proc_cmdline(-option => "rdloaddriver=scsi_dh_alua") != $ClientUtil::TRUE) {
            $self->_set_err("scsi_dh_alua driver is not loaded which is required");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

###################################################################################
# Suroutine Name: _update_grub_kernel_param
# Objectiive:     Modifies the grub kernel param to load scsi_dh_alua automatically
#                 during boot time
###################################################################################
sub _update_grub_kernel_param {
    my ($self) = @_;

    my ($client_obj, $conn, $status, @output, $cmd);
    my (@grub_conf, $grub_conf_path, $grub2);

    $self->_clear_err();

    $client_obj = $self->{parent}->{parent}->{parent};
    $conn = $client_obj->{conn};

    # If grub2 is installed, grub configuration file name changes.So find whether grub/grub2 is installed and
    # make changes in appropriate configuration file.
    # grub
    #   ls -al /etc | grep grub
    #   lrwxrwxrwx.   1 root root       31 Apr 15 12:41 grub2-efi.cfg -> ../boot/efi/EFI/fedora/grub.cfg
    # grub2
    #   ls -al /etc | grep grub
    #   lrwxrwxrwx.   1 root root        22 Apr 15 22:02 grub.conf -> ../boot/grub/grub.conf
    # /etc will be having soft link to the grub configuration file inside /boot directory.
    $grub2 = 0;
    $cmd = "rpm -qa | grep grub";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Command $cmd returned a bad status $status");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        if ($line =~ /grub2/){
            $grub2 = 1;
            last;
        }
    }

    $grub_conf_path = "";
    if ($grub2 == 1) {
        $cmd = "ls /etc | cat";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status > 0) {
            $self->_set_err("Command $cmd returned a bad status $status");
            return $ClientUtil::FALSE;
         }
         @grub_conf = grep (/^grub.*cfg$/, @output);
         $grub_conf_path = "/etc/$grub_conf[0]";
         $grub_conf_path =~ s/^\s+//;
    } else {
        $grub_conf_path = "/etc/grub.conf";
    }


    # Creating a backup of grub configuration file
    $cmd = "cp -f $grub_conf_path ${grub_conf_path}.orig";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed create a backup of $grub_conf_path file");
        return $ClientUtil::FALSE;
    }

    # Delete any previously existing temporary file
    $cmd = "rm -f ${grub_conf_path}.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in deleting temp file");
        return $ClientUtil::FALSE;
    }

    # Create a temporary file to make all the changes
    $cmd = "touch ${grub_conf_path}.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in creating temp file");
        return $ClientUtil::FALSE;
    }

    my @grub_conf_output;
    my $new_grub_conf_string;
    $cmd = "cat $grub_conf_path";
    ($status, @grub_conf_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in displaying $grub_conf_path file");
        return $ClientUtil::FALSE;
    } else {
        # Copy the contents of /boot/grub/grub.conf file to a temporary file /boot/grub/grub.conf.tmp
        # and during this process if we encounter kernel param, append "rdloaddriver=scsi_dh_alua"
        # if not present already
         foreach my $line (@grub_conf_output) {
            if (($line =~ /^\s*kernel.*/) or ($line =~ /^\s*linux.*/)){
                if ($line !~ /rdloaddriver\=scsi_dh_alua/) {
                    $line .= " rdloaddriver=scsi_dh_alua";
                }
            }
            $new_grub_conf_string .= $line . "\n" ;
        }
    }

    $new_grub_conf_string =~ s/\t/   /g;
    my @new_menu_lst = split(/\n/, $new_grub_conf_string);
    chomp(@new_menu_lst);

    # Append each line in temporary file
    foreach my $line (@new_menu_lst) {
        $cmd = "echo \'$line\' >> ${grub_conf_path}.tmp";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error in copying the data back into ${grub_conf_path}.tmp file");
            return $ClientUtil::FALSE;
        }
    }

    # Copy the contents of the temporary file to grub configuration file inside /etc
    # Could not use mv cmd , since mv will replace the existing soft link and create a new file on that name.
    $cmd = "yes | cp -f ${grub_conf_path}.tmp $grub_conf_path";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in copying the contents of ${grub_conf_path}.tmp to $grub_conf_path");
        return $ClientUtil::FALSE;
    }

    # Delete the temporary file
    $cmd = "rm -f ${grub_conf_path}.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in deleting temp file");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Suroutine Name: _dev_map_changes
# Objectiive:      Changes to /etc/fstab and /boot/grub/device.map
# Details:         See POD documentation at the end of this file
################################################################################
sub _dev_map_changes {
    my ($self, %args) = @_;

    my ($cmd, $conn, @output, $status, $clientObj, @fstab_output, @mpath_output, %wwid_hash);
    my ($alias, $init_filename, @device_output, $line, $file);

    $clientObj = $self->{parent}->{parent}->{parent};
    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Get the kernel release of the host which is used later to create the initrd image
    my $kernel_version = $clientObj->{revision};

    # RHEL6 and OL6 onwards, initrd image is represented as initramfs
    if ($clientObj->{major_revision} =~ /rhel6|rhel7|ol6|fedora/i) {
        $init_filename = "/boot/initramfs-" . $kernel_version . "\.img";
    } else {
        $init_filename = "/boot/initrd-" . $kernel_version . "\.img";
    }

    # Take a backup of the existing initrd image.
    $cmd = "cp -f $init_filename $init_filename.bak";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed to create a backup of initrd image");
        return $ClientUtil::FALSE;
    }

    # Recreate initrd
    # cmd to create initrd image differs from each OS
    if ($clientObj->{major_revision} =~ /rhel6|rhel7|ol6|fedora/i) {
        $cmd = "dracut -f $init_filename";
    } elsif ($clientObj->{major_revision} =~ /rhel4/i) {
        $cmd = "mkinitrd -f $init_filename $kernel_version --with=qioctlmod";
    } else {
        $cmd = "mkinitrd -f $init_filename $kernel_version";
    }
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to create an initrd image");
        return $ClientUtil::FALSE;
    }

    # Using "multipath -ll" command to store all the multipath devices and their corresponding
    # wwpns in the hash
    $cmd = "multipath -ll";
    ($status, @mpath_output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to execute multipath -ll");
         return $ClientUtil::FALSE;
    }

    foreach my $alias (@mpath_output) {
        if ($alias =~ /(mpath[0-9a-zA-Z]*)\s+\(([0-9a-zA-Z]*)\)/) {
             $wwid_hash{$1} = $2;
        }
    }

    # Creating a backup file of /etc/fstab file
    $cmd = "cp -f /etc/fstab /etc/fstab.bak";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed create a backup of /etc/fstab file");
        return $ClientUtil::FALSE;
    }

    # Delete any previously existing temporary files
    $cmd = "rm -f /etc/fstab.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in deleting temp file");
        return $ClientUtil::FALSE;
    }

    # Create a temporary file to make all the changes
    $cmd = "touch /etc/fstab.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in creating temp file");
        return $ClientUtil::FALSE;
    }

    my $new_fstab_string;
    $cmd = "cat /etc/fstab";
    ($status, @fstab_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in displaying /etc/fstab file");
        return $ClientUtil::FALSE;
    } else {
        # Copy the contents of /etc/fstab file to a temporary file /etc/fstab.tmp
        # and during this process if we encounter any user friendly names, replace them
        # with their corresponding wwpns from the hash
         foreach $line (@fstab_output) {
            if ($line =~ /(mpath[0-9a-zA-Z]*)/){
                my $mpathvar = $1;
                # remove partition designation - need to remove on root device
                # because hash entry wont contain partition designation
                $mpathvar =~ s/p\d+//;
                $line =~ s/$mpathvar/$wwid_hash{$mpathvar}/g;
            }
            $new_fstab_string .= $line . "\n" ;
        }
    }

    my @new_fstab = split(/\n/, $new_fstab_string);
    chomp(@new_fstab);

    # Append each line in new /etc/fstab file
    foreach $line (@new_fstab) {
        $cmd = "echo \'$line\' >> /etc/fstab.tmp";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error in copying the data back into /etc/fstab.tmp file");
            return $ClientUtil::FALSE;
        }
    }

    # Rename the temporary file as /etc/fstab
    $cmd = "mv -f /etc/fstab.tmp /etc/fstab";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in replacing the contents of /etc/fstab");
        return $ClientUtil::FALSE;
    }

    # Creating a backup of the /boot/grub/device.map file
    $cmd = "cp -f /boot/grub/device.map /boot/grub/device.map.bak";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed to create a backup of /boot/grub/device.map file");
        return $ClientUtil::FALSE;
    }

    # Delete existing temporary file
    $cmd = "rm -f /boot/grub/device.map.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed to delete /boot/grub/device.map.tmp file");
        return $ClientUtil::FALSE;
    }

    # Creating a temporary file of device.map
    $cmd = "touch /boot/grub/device.map.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed to delete /boot/grub/device.map.tmp file");
        return $ClientUtil::FALSE;
    }

    my $new_device_map_string;
    $cmd = "cat /boot/grub/device.map";
    ($status, @device_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in displaying /boot/grub/device.map file");
        return $ClientUtil::FALSE;
    } else {
        # Copy /boot/grub/device.map to a temporary file /boot/grub/device.map.tmp
        # and during this process if we encounter any user friendly names, replace them
        # with their corresponding wwpns from the hash
        foreach $line (@device_output) {
           if ($line =~ /(mpath[0-9a-zA-Z]*)/i) {
               my $mpathvar = $1;
               # remove partition designation - need to remove on root device
               # because hash entry wont contain partition designation
               $mpathvar =~ s/p\d+//;
               $line =~ s/$mpathvar/$wwid_hash{$mpathvar}/g;
           }
           $new_device_map_string .= $line . "\n";
        }
    }

    my @new_device_map = split(/\n/, $new_device_map_string);
    chomp(@new_device_map);

    # Append each line in new /boot/grub/device.map file
    foreach $line (@new_device_map) {
        $cmd = "echo \'$line\' >> /boot/grub/device.map.tmp";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error in copying the data into /boot/grub/device.map.tmp file");
            return $ClientUtil::FALSE;
        }
    }

    # Rename the temporary file as /boot/grub/device.map file
    $cmd = "mv /boot/grub/device.map.tmp /boot/grub/device.map";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in copying from device.map.tmp back to device.map file");
        return $ClientUtil::FALSE;
    }

    # Creating a backup file of /boot/grub/grub.conf file
    $cmd = "cp -f /boot/grub/grub.conf /boot/grub/grub.conf.bak";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Cmd failed create a backup of /boot/grub/grub.conf file");
        return $ClientUtil::FALSE;
    }

    # Delete any previously existing temporary file
    $cmd = "rm -f /boot/grub/grub.conf.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in deleting temp file");
        return $ClientUtil::FALSE;
    }

    # Create a temporary file to make all the changes
    $cmd = "touch /boot/grub/grub.conf.tmp";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in creating temp file");
        return $ClientUtil::FALSE;
    }

    my @grub_conf_output;
    my $new_grub_conf_string;
    $cmd = "cat /boot/grub/grub.conf";
    ($status, @grub_conf_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in displaying /boot/grub/grub.conf file");
        return $ClientUtil::FALSE;
    } else {
        # Copy the contents of /boot/grub/grub.conf file to a temporary file /boot/grub/grub.conf.tmp
        # and during this process if we encounter any user friendly names, replace them
        # with their corresponding wwpns from the hash
         foreach $line (@grub_conf_output) {
            if ($line =~ /(mpath[0-9a-zA-Z]*)/){
                my $mpathvar = $1;
                # remove partition designation - need to remove on root device
                # because hash entry wont contain partition designation
                $mpathvar =~ s/p\d+//;
                $line =~ s/$mpathvar/$wwid_hash{$mpathvar}/g;
            }
            $new_grub_conf_string .= $line . "\n" ;
        }
    }

    my @new_grub_conf = split(/\n/, $new_grub_conf_string);
    chomp(@new_grub_conf);

    # Append each line in /boot/grub/grub.conf file
    foreach $line (@new_grub_conf) {
        $cmd = "echo \'$line\' >> /boot/grub/grub.conf.tmp";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Error in copying the data back into /boot/grub/grub.conf.tmp file");
            return $ClientUtil::FALSE;
        }
    }

    # Rename the temporary file as /boot/grub/grub.conf
    $cmd = "mv -f /boot/grub/grub.conf.tmp /boot/grub/grub.conf";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in replacing the contents of /boot/grub/grub.conf");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_mp_devices()
# Objective:       get all mp devices
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mp_devices {
    my ($self, %args) = @_;
    my ($cmd, @output, @hDevices, $conn, $status, $clientObj, @tmp_output, @netapp_dev, @luns);

    $clientObj = $self->{parent}->{parent}->{parent};
    $conn = $clientObj->{conn};

    $self->_clear_err();
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn' => $args{$_}");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Netapp LunID starts with 360a9 or 3600a09, so I am going to grep all the mp devices
    # which starts with 360a9|3600a09 or with mpath (user friendly name) and will return those devices as
    # dm devices
    $cmd = "ls /dev/mapper/ | grep -P \"360a9|3600a09|mpath\"";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0 or scalar(@output) == 0) {
        return (wantarray()) ? () : [];
    }

    if (!defined $self->{root_mp_devices}) {
        $self->{root_mp_devices} = $clientObj->get_root_dev();
        if ($self->error) {
            return (wantarray()) ? () : [];
        }
    }
    my @root_mp_dev = @{$self->{root_mp_devices}};

    foreach my $name (@output) {
        my $hDevs = {};
        $name =~ s/\/dev\/mapper\///;
        next if (grep (/$name/, @root_mp_dev));

        # Need to store a single sdX device(a path) that corresponds to
        # mpath device. This mapping will be used in LinuxDev.pm _new() when
        # get_lun_path is called. It will be used to get the lun path on filer side
        # Here is the sample output of "multipath -ll <mpath_device>"
        #------------------------------
        #    [root@lnx236-107 ~]# multipath -ll 360a980004334647150344c3056785552
        #    360a980004334647150344c3056785552dm-15 NETAPP,LUN
        #    [size=5.0G][features=1 queue_if_no_path][hwhandler=0]
        #    \_ round-robin 0 [prio=2][active]
        #     \_ 57:0:0:2 sdk 8:160 [active][ready]
        #------------------------------
        $cmd = "multipath -ll $name";
        my ($status, @output_dev) = &ClientUtil::execute($conn, $cmd);

        # in case of sles10 and rhel4u3/4 multipath -ll returns 1 even if the cmd passes.
        # Putting the check condition so we dont fail for wrong reasons
        if ($clientObj->{major_revision} eq "sles10.0" or
            $clientObj->{major_revision} eq "rhel4.3" or
            $clientObj->{major_revision} eq "rhel4.4" or
            $clientObj->{major_revision} eq "rhel4.5") {
            if ($status > 1) {
                $self->_set_err("Cmd $cmd failed. Output: (@output_dev)");
                return (wantarray()) ? () : [];
            }
        } elsif ($status > 0) {
            $self->_set_err("Cmd $cmd failed. Output: (@output_dev)");
            return (wantarray()) ? () : [];
        }

        # Skip if device doesnt belong to NetApp
        next if (not grep(/NETAPP/i, @output_dev));

        @output_dev = grep (/\s+sd\w+\s+/, @output_dev);
        my @first_path_info = split(/\s+/, $output_dev[0]);
        my $sdX = $first_path_info[3];

        $self->{mapping}->{$name}->{sdX}= $sdX;

        $hDevs->{name}  = $name;
        $hDevs->{path}  = "/dev/mapper/";
        $hDevs->{label} = "free";
        push @hDevices, $hDevs;
    }

    return (wantarray()) ? @hDevices : \@hDevices;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self, %args) = @_;

    my ($takeover_state);

    $self->_clear_err();

    # Hardcoding this for time being, but might make this a param if
    # it is used by other platforms
    my $io_threshold = 100;

    # Will store the all paths in $self along with their types (primary/seconadary)
    $self->_store_all_paths_type();
    return $ClientUtil::FALSE if $self->error();

    # Will return iostats of all the paths of the type(pri/sec/all) passed to it
    my $paths_iostats = $self->_get_paths_iostats(%args);
    return $ClientUtil::FALSE if $self->error();

    # Verify whether io is running or not on proper paths
    foreach my $path (keys %{$paths_iostats}) {
        my $path_iostats = $paths_iostats->{$path}->{blk_read} + $paths_iostats->{$path}->{blk_wrtn};
        $self->sanlog(-msg => "IO stats(block read + block written) for path $path are: $path_iostats\n");
        # Failout if IO is less than the threshold value
        # Setting the threshold value to 100 because even if there
        # is no actual IO running on luns, multipathd reads 0th block
        # to find out whether path is dead or alive. Because of this
        # value of iostats will never be zero. But we have observed that
        # it is always considerably less than 100.
        if ($path_iostats < $io_threshold) {
            $self->_set_err("IO activity for the path /dev/$path is less than threshold ($io_threshold)");
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($cmd, $device, @output, $status, $conn, $dead_paths_cnt, $alive_paths_cnt);
    my ($file, $line, $start, $device_num, $per_lun_hash, $clientObj);
    my ($protocol, @lun_show_output, $output);
    my ($validate_mode, @wwids, @mp_devices, $multipath_error, $takeover_state);

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $per_lun_hash = {};

    # get the client object
    $clientObj =  $self->{parent}->{parent}->{parent};

    # get connection from the client object
    $conn = $self->{conn};

    # get HU path for path's to binaries
    my $hu_path = $clientObj->get_hu_bin_path();

    # only grab sanlun to get protocol if we are in validation mode
    # meaning we are checking paths. We will get the protocol only
    # in steady state, since sanlun is very slow in takover mode
    if ($validate_mode == 0) {
        # need normal sanlun output to get protocol
        $cmd = $hu_path . "/sanlun lun show";
        ($status, @lun_show_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                             180, 2, "-comment");
        if ($status != 0) {
            # if sanlun is not available, try to get info from netapp_scsi_show script
            @lun_show_output = $clientObj->netapp_scsi_show();
            if ($clientObj->error()) {
                $self->_set_err($clientObj->errmsg());
                return {};
            }
        }
    }

    # run multipath -ll -d to get list of the devices and paths
    # to those devices.
    if ($clientObj->{major_revision} =~ /cxs/i) {
        $cmd = "mpathutil status";
    } else {
        $cmd = "/sbin/multipath -ll -d";
    }

    if ($clientObj->{major_revision} eq "sles10.0" or
        $clientObj->{major_revision} eq "rhel4.3" or
        $clientObj->{major_revision} eq "rhel4.4" or
        $clientObj->{major_revision} eq "rhel4.5") {
        # In sles10 and rhel4u3,u4 and u5, there is a bug in multipath package
        # which causes the multipath -ll cmd to return 1 even if it passes.
        # execute() defined in ClientUtil.pm returns 1 if cmd times out.
        # In this case, if we use execute() from ClientUtil.pm, we can't differentiate
        # whether the cmd passed or timed out. To avoid this, using nate defined execute
        # directly
        my $retry_count = 6;
        while ($retry_count > 0) {
            eval {
                $output = $conn->execute(command => $cmd,
                                         get_status => \$status,
                                         timeout => 180);
                if ($status != 1) {
                    $multipath_error = 1;
                    $retry_count--;
                    $self->sanlog(-msg => "Attempted command ($cmd) and it failed with status ($status).\n" .
                                          "Attempts left ($retry_count)");
                    next;
                } else {
                    # command passed
                    @output = split(/\n/, $output);
                    $multipath_error = 0;
                    last;
                }
            };

            if ($@) {
                # command timed out
                $multipath_error = 1;
                $retry_count--;
                $self->sanlog(-msg => "Attempted command ($cmd) and it has timed out.\n" .
                                      "Attempts left ($retry_count)");
                next;
            }
        }
    } else {
        ($status, @output) = &ClientUtil::execute_retry($conn, $cmd,
                                                    180, 6, "-warn", 0);
        if ($status != 0) {
            $multipath_error = 1;
        }
    }

    if ($multipath_error == 1) {
        # Will store the all paths in $self with their types (primary/seconadary)
        $self->_store_all_paths_type();
        return $ClientUtil::FALSE if $self->error();


        # If we are here it means that cmd "/sbin/multipath -ll -d"
        # either hung or timed-out. We are going to  use alternate
        # method('sg_turs /dev/sdX') to validate the paths
        if ($self->_store_status_of_each_path() == $ClientUtil::FALSE) {
            return (wantarray()) ? () : [];
        }

        if ($clientObj->{config_dmmp_use_user_friendly_names} =~ /^1$|^yes$/i) {
            $cmd = "ls /dev/mapper/ | grep mpath";
        } else {
            $cmd = "ls /dev/mapper/ | grep -P \"360a9|3600a09\"";
        }
        ($status, @wwids) = &ClientUtil::execute($conn, $cmd);
        if ($status > 0 or scalar (@wwids) == 0) {
            $self->_set_err("Unable to find any DM devices");
            return (wantarray()) ? () : [];
        }

        foreach my $name (@wwids) {
            # skip partitioned device
            if ($name =~ /p\d+$|_part\d+$/i) {
                next;
            }

            # get rid of /dev/mapper from DM device's name
            $name =~ s/\/dev\/mapper\///;

            push (@mp_devices, $name);
        }

        # Here we are populating $per_lun_hash. $self is populated
        # with num_paths_alive/num_paths_dead by _store_status_of_each_path
        foreach my $wwid (@mp_devices) {
           my @dev_paths = keys %{$self->{$wwid}->{paths}};
           my $path = $dev_paths[0];
           # only set protocol in validation mode (comparing paths)
           if ($validate_mode == 0) {
               # lookup that device in sanlun output to get protocol
               my $sanlun_dev = (grep(/$path\s+/, @lun_show_output))[0];
               if ($sanlun_dev =~ /\s+iscsi\s+/i) {
                   $protocol = "iSCSI";
               } else {
                   $protocol = "FCP";
               }
               $per_lun_hash->{$wwid}->{protocol} = $protocol;
           } else {
               $per_lun_hash->{$wwid}->{protocol} = "takeover";
           }

           $per_lun_hash->{$wwid}->{num_paths_alive} = $self->{$wwid}->{num_paths_alive};
           $per_lun_hash->{$wwid}->{num_paths_dead} = $self->{$wwid}->{num_paths_dead};
           $per_lun_hash->{$wwid}->{num_paths_primary} = "NA";
           $per_lun_hash->{$wwid}->{num_paths_secondary} = "NA";
           $per_lun_hash->{$wwid}->{num_paths} = scalar (keys %{$self->{$wwid}->{paths}});
        }

        if ($clientObj->{major_revision} =~ /cxs/i) {
            goto SR_SANITY;
        }

        return $per_lun_hash;
    }

    $start = 0;
    foreach $line (@output) {
        chomp($line);

        # can appear when paths are down
        if ($line =~ /checker msg is/i) {
            next;
        }

        # when we get the next line with device we are done
        # with the previous representation.
        if ($line =~ /360a9[a-z|A-Z|0-9]{28}|3600a09[a-z|A-Z|0-9]{26}/) {
            $start = 0;
        }

        # if this is SLES11SP1+ then we need to skip
        # non-NETAPP device properly since it might not be
        # blacklisted.
        if ($clientObj->{major_revision} =~ /sles11\.1/i) {
            if ($line =~ /\s+dm\-\d+\s+/ and $line !~ /NETAPP/i) {
                $start = 0;
            }
        }

        # Start collecting LUN paths
        if ($start == 1) {
            my $path;
            if (($clientObj->{major_revision} =~ /rhel6|rhel7|ol6|fedora/i)
               or ($clientObj->{major_revision} =~ /ovm3/i)
               or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
               or ($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} >= 1)
               or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 6)) {
                # ignore size and feature information (may use later)
                # ignore lines in format:
                # dm table 360a98000433465716b5a454878675759  N
                #-------------------------------------------------------------
                # Here is an example of 'multipath -ll' on RHEL6, OL6 and SLES11SP1
                #-------------------------------------------------------------
                # 360a98000486e2f76535a5544354b5332 dm-9 NETAPP,LUN
                # size=150G features='1 queue_if_no_path' hwhandler='0' wp=rw
                #|-+- policy='round-robin 0' prio=8 status=enabled
                #| |- 5:0:0:7 sdi  8:128  active ready running
                #| `- 6:0:1:7 sdam 66:96  active ready running
                #`-+- policy='round-robin 0' prio=2 status=enabled
                #  |- 5:0:1:7 sdaa 65:160 active ready running
                #    `- 6:0:0:7 sdz  65:144 active ready running
                #-------------------------------------------------------------

                next if ($line =~ /^size=/ or /^dm\s+/);

                # Skip lines providing policy info
                next if (grep(/policy=/, $line));

                # Look for the lines which contains path status and store path status info in hash
                if ($line =~ /\s+.+\s+sd\w+\s+.+\s+\w+\s+\w+\s+\w+/i) {
                    # this regex will look for the lines like below:
                    #  5:0:0:7 sdi  8:128  active ready running
                    #  1:0:1:33 sdfa 129:192 failed faulty running
                    my @path_info = split(/\s+/, $line);
                    $path = $path_info[-5];
                    $per_lun_hash->{$device_num}->{$path}->{path_state_1} = $path_info[-3];
                    $per_lun_hash->{$device_num}->{$path}->{path_state_2} = $path_info[-2];
                    $per_lun_hash->{$device_num}->{$path}->{path_state_3} = $path_info[-1];
                }

            } else {
                # ignore size and feature information (may use later)
                # ignore lines in format:
                # dm table 360a98000433465716b5a454878675759  N
                if ($line =~ /^\\_/ or /\[size=/ or /^dm\s+/) {
                    next;
                }

                # start collecting paths (sd* names) as paths to this LUN
                if ($line =~ /^\s+\\_/) {
                    $line =~ m/\s+\\_\s+.*\s(sd\w+)\s+.*\s+\[(\w+)\]\[(\w+)\]/;
                    $path = $1;
                    my $state_1 = $2;
                    my $state_2 = $3;

                    $per_lun_hash->{$device_num}->{$path}->{path_state_1} = $state_1;
                    $per_lun_hash->{$device_num}->{$path}->{path_state_2} = $state_2;
                }
            }

            # only set protocol in validation mode (comparing paths)
            if ($validate_mode == 0) {
                # lookup that device in sanlun output to get protocol
                my $sanlun_dev = (grep(/$path\s+/, @lun_show_output))[0];
                if ($sanlun_dev =~ /\s+iscsi\s+/i) {
                    $protocol = "iSCSI";
                } else {
                    $protocol = "FCP";
                }
                $per_lun_hash->{$device_num}->{protocol} = $protocol;
            } else {
                $per_lun_hash->{$device_num}->{protocol} = "takeover";
            }
        }

        # if we see the following, we know we have a
        # new LUN (very dependent on OS revision and
        # multipath.conf differences):
        # mpath30 (360a9800043346149644a474f4f495776) or
        # 360a9800043346149644a474f4f495776 or
        # 3600a09800043346149644a474f4f4957 or
        # 360a9800043346141614a474f4f493639NETAPP,LUN or
        # 360a9800056724a774e5a47716272712ddm-10 NETAPP,LUN
        # using 360a9* as representation of device/LUN
        if ($line =~ /(360a9[a-z|A-Z|0-9]{28}|3600a09[a-z|A-Z|0-9]{26})/ and $start == 0) {
            $device_num = $1;
            $device = $line;
            $start = 1;
        }
    }

    # traverse the hash we generated and generate information
    # about the number of paths, alive paths, and dead paths.
    foreach my $key (keys %{$per_lun_hash}) {
        # store number of paths (take into consideration
        # protocol being set on the hash)
        $per_lun_hash->{$key}->{num_paths} =
                    (scalar(keys %{$per_lun_hash->{$key}}) - 1);

        my $alive_cnt = 0;
        my $dead_cnt = 0;
        my $proxy_cnt = 0;
        my $primary_cnt = 0;
        foreach my $path_key (keys %{$per_lun_hash->{$key}}) {
            if ($path_key =~ /protocol|num_paths/) {
                next;
            }

            # Store alive and dead paths
            if ($per_lun_hash->{$key}->{$path_key}->{path_state_1} =~ /active/i and
                $per_lun_hash->{$key}->{$path_key}->{path_state_2} =~ /ready/i) {
                if (($clientObj->{major_revision} =~ /rhel6|rhel7|ol6|fedora/i)
                 or ($clientObj->{major_revision} =~ /ovm3/i)
                 or ($clientObj->{major_revision} =~ /cxs6/i and $clientObj->{minor_revision} >= 2)
                 or ($clientObj->{major_revision} =~ /sles11/i and $clientObj->{minor_revision} >= 1)
                 or ($clientObj->{major_revision} =~ /ol5/i and $clientObj->{minor_revision} >= 6)) {
                    if ($per_lun_hash->{$key}->{$path_key}->{path_state_3} =~ /running/i) {
                        $alive_cnt++;
                    } else {
                        $dead_cnt++;
                    }
                } else {
                    $alive_cnt++;
                }
            } else {
                $dead_cnt++;
            }
        }

        # store number of paths that are enabled (alive)
        $per_lun_hash->{$key}->{num_paths_alive} = $alive_cnt;

        # store number of paths that are disabled (dead)
        $per_lun_hash->{$key}->{num_paths_dead} = $dead_cnt;

        # store number of paths that are primary
        $per_lun_hash->{$key}->{num_paths_primary} = "NA";

        # store number of paths that are secondary
        $per_lun_hash->{$key}->{num_paths_secondary} = "NA";
    }

    SR_SANITY:
    # Validate SR sanity in case of Citrix XenServer
    if ($clientObj->{major_revision} =~ /cxs/i) {
        $self->validate_sr_sanity(-path_info_hash => $per_lun_hash);
        if ($self->error()) {
            return {};
        }
    }

    return $per_lun_hash;
}

################################################################################
# Subroutine Name: validate_sr_sanity
# Objective:       Validates that SR path information just aligns with mpathutil
#                  information
################################################################################
sub validate_sr_sanity {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, $conn);
    my $path_info_hash;

    # parse arguments
    foreach (keys %args) {
        if (/-?path_info_hash/) {
            if (not length($path_info_hash = $args{$_})) {
                $self->_set_err("Invalid value for argument '-path_info_hash'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (not defined($path_info_hash)) {
        $self->_set_err("Required argument '-path_info_hash'");
        return $ClientUtil::FALSE;
    }

    # get the client object
    my $client_obj =  $self->{parent}->{parent}->{parent};

    # get connection from the client object
    $conn = $client_obj->{conn};

    $self->sanlog(-msg => "Performing SR sanity check (xe pbd-list versus mpathutil status)");

    # get SR path information
    my $sr_path_hash = $self->get_sr_path_info();
    if ($self->error()) {
        return $ClientUtil::FALSE;
    }

    if (not defined($sr_path_hash)) {
        $self->sanlog(-msg => "No SAN SR found. Skipping!");
        return $ClientUtil::TRUE;
    }

    my $errstr = "";
    my $error = 0;
    foreach my $scsi_id (keys %{$sr_path_hash}) {
        # more than one PBD can be available if SR is shared.
        foreach my $pbd (values %{$sr_path_hash->{$scsi_id}->{pbds}}) {
            # skip the check if the PBD is not currently attached
            next if $pbd->{currently_attached} !~ /true/i;

            my $num_paths_alive = $path_info_hash->{$scsi_id}->{num_paths_alive};
            if ($num_paths_alive != $pbd->{num_paths_alive}) {
                $error++;
                my $sr_name_label = $sr_path_hash->{$scsi_id}->{sr_name_label};
                my $pbd_uuid = $pbd->{uuid};
                $errstr .= "Number of alive paths to the PBD ($pbd_uuid) from SR ($sr_name_label) doesn't match\n" .
                           "Alive paths reported in pbd-list: " . $pbd->{num_paths_alive} . "\n" .
                           "Alive paths reported in mpathutil: $num_paths_alive\n\n";
            }
        }
    }

    if ($error > 0) {
        $self->_set_err("SR sanity check failed. Details:\n$errstr");
        return $ClientUtil::FALSE;
    }

    $self->sanlog(-msg => "SR sanity check passed");

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_sr_path_info
# Objective:       return a hash which has path information of all SR
################################################################################
sub get_sr_path_info {
    my ($self, %args) = @_;

    my ($cmd, @output, $status, $conn);

    my $sr_path_hash = {};

    # get the client object
    my $client_obj =  $self->{parent}->{parent}->{parent};

    # get connection from the client object
    $conn = $client_obj->{conn};

    $cmd = "xe sr-list params=uuid,name-label,PBDs,type,sm-config";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to get SR information");
        return  $sr_path_hash;
    }

    my $sr_uuid = "";
    my $sr_name_label = "";
    my @pbds = ();
    my $found_valid_sr = 0;

    foreach my $line (@output) {

        next if $line =~ /^\s*$/;

        if ($line =~ /uuid.*:(.+)/i) {
            $sr_uuid = $1;
            $sr_uuid =~ s/^\s*//;
            $found_valid_sr = 1;
            next;
        } elsif ($line =~ /name-label.*:(.+)/i and $found_valid_sr == 1) {
            $sr_name_label = $1;
            $sr_name_label =~ s/^\s*//;
            #if ($sr_name_label !~ /NATE_SR/i) {
            #    $sr_uuid = "";
            #    $sr_name_label = "";
            #    $found_valid_sr = 0;
            #}
            next;
        } elsif ($line =~ /PBDs.*:(.+)/i and $found_valid_sr == 1) {
            my $pbds_string = $1;
            $pbds_string =~ s/\s+//g;
            @pbds = split(/;/, $pbds_string);
            next;
        } elsif ($line =~ /type.*:(.+)/i) {
            if ($1 !~ /lvmohba|lvmoiscsi/i) {
                $sr_uuid = "";
                $sr_name_label = "";
                @pbds = ();
                $found_valid_sr = 0;
            }
            next;
        } elsif ($line =~ /sm-config.*:.*scsi-([\S]{33})/i and $found_valid_sr == 1) {
            my $scsi_id = $1;
            $sr_path_hash->{$scsi_id}->{sr_uuid} = $sr_uuid;
            $sr_path_hash->{$scsi_id}->{sr_name_label} = $sr_name_label;

            foreach my $pbd (@pbds) {
                my @pbd_output = ();
                $cmd = "xe pbd-list uuid=$pbd params=other-config,currently-attached";
                ($status, @pbd_output) = &ClientUtil::execute($conn, $cmd);
                if ($status != 0) {
                    $self->_set_err("Failed to get path information of PBD ($pbd)");
                    return $sr_path_hash;
                }
                my $num_paths = 0;
                my $num_paths_alive = 0;
                my $currently_attached = "false";

                foreach my $line (@pbd_output) {
                    if ($line =~ /currently-attached.*:\s*(\S+)/i) {
                        $currently_attached = $1;
                    } elsif ($line =~ /\s+mpath-\S+:\s*\[(\d+),\s*(\d+)\]/i) {
                        $num_paths = $2;
                        $num_paths_alive = $1;
                    }
                }

                $sr_path_hash->{$scsi_id}->{pbds}->{$pbd}->{uuid} = $pbd;
                $sr_path_hash->{$scsi_id}->{pbds}->{$pbd}->{currently_attached} = $currently_attached;
                $sr_path_hash->{$scsi_id}->{pbds}->{$pbd}->{num_paths} = $num_paths;
                $sr_path_hash->{$scsi_id}->{pbds}->{$pbd}->{num_paths_alive} = $num_paths_alive;
            }
            $sr_uuid = "";
            $sr_name_label = "";
            @pbds = ();
            $found_valid_sr = 0;
            next;
        }
    }

    return $sr_path_hash;
}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self, $parent, $obj_template);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '-parent' => $args{$_}");
                return $self;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
    }

    # plug in the parent
    $self->{parent} = $parent;

    return $self;
}

################################################################################
# Subroutine Name: _is_multipathd_running()
# Objective:       returns ClientUtil::TRUE or ClientUtil::FALSE depending on
#                  if /etc/init.d/multipathd is running or not.
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_multipathd_running {
    my ($self, %args) = @_;

    my ($status, $cmd, $conn, @output);

    # parse arguments

    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid argument value '-conn'=> $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    if (not defined($conn)) {
        $self->_set_err("'-conn' parameter is required.");
        return $ClientUtil::FALSE;
    }

    my $client_obj =  $self->{parent}->{parent}->{parent};

    # check status of multipathd service
    if ($client_obj->{major_revision} =~ /rhel7|fedora/) {
        $cmd = "systemctl status multipathd.service";
    } else {
        $cmd = "/etc/init.d/multipathd status";
    }
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ((grep /running/i, @output) and $status == 0) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: _create_binding_entry_for_root_dev()
# Objective:       Will create binding entry between mpath device and its alias
#                  in mpath binding file (/etc/multipath/bindings) for root device
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_binding_entry_for_root_dev {
    my ($self, %args) = @_;

    my ($wwid);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?wwid/) {
            if (not length($wwid = $args{$_})) {
                $self->_set_err("Invalid value for argument '-wwid' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Add binding entry for root device with alias default to mpatha
    # We can do this because there wont be any binding entry in
    # file /etc/multipath/bindings
    $self->_create_mpath_binding_entry(-wwid    => $wwid,
                                       -alias   => "mpatha");
    return $ClientUtil::FALSE if ($self->error);

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _create_mpath_binding_entry()
# Objective:       Will create binding entry between mpath device and its alias
#                  in mpath binding file (/etc/multipath/bindings)
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_mpath_binding_entry {
    my ($self, %args) = @_;
    my ($cmd , $status, @output, $wwid, $alias);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?wwid/) {
            if (not length($wwid = $args{$_})) {
                $self->_set_err("Invalid value for argument '-wwid' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?alias/) {
            if (not length($alias = $args{$_})) {
                $self->_set_err("Invalid value for argument '-alias' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Check for required params
    if (not defined($wwid)) {
        $self->_set_err("'-wwid' parameter is required.");
        return $ClientUtil::FALSE;
    }

    # Here is an example of /etc/multipath/bindings
    ##-----------------------------------------------------------------------
    ## Multipath bindings, Version : 1.0
    ## NOTE: this file is automatically maintained by the multipath program.
    ## You should not need to edit this file in normal circumstances.
    ##
    ## Format:
    ## alias wwid
    ## mpatha 360a98000486e2f66426f5831337a326e
    ## mpathb 360a98000486e2f65686f596f5576384b
    ## mpathc 3600a098000486e2f65686f596f557638
    ##-----------------------------------------------------------------------
    $cmd = "cat /etc/multipath/bindings";
    ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to display /etc/multipath/bindings file");
        return  $ClientUtil::FALSE;
    }
    # Dont do anything if binding entry already exists in the file
    if (grep(/$alias\s+$wwid/, @output)) {
        $self->sanlog(-msg => "Binding entry \'$alias $wwid\' " .
                              "already exists in /etc/multipath/bindings file");
        return  $ClientUtil::TRUE;
    } else { # Else append binding entry to the file
        $cmd = "echo \'$alias $wwid\' >> /etc/multipath/bindings";
        ($status, @output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Failed to add binding entry \'$alias $wwid\' " .
                            "to /etc/multipath/bindings file");
            return  $ClientUtil::FALSE;
        }
    }

    return  $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_paths_iostats()
# Objective:       Gather the iostats of all paths of type $type for all devices
#                  Type could be primary, secondary or all(get all the paths)
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_paths_iostats {
    my ($self, %args) = @_;

    my ($cmd, $device, $type, @devices, @paths, @output, $status, $conn, @lun_show_output);
    my ($clientObj, $validate_mode, @paths_to_check_for_io, $takenover_filer, $takeover_state);
    my ($paths_stats);

    $conn = $self->{conn};

    $self->_clear_err();
    foreach (keys %args) {
        if (/-?takeover_state/) {
            if (not length($takeover_state = $args{$_})) {
                $self->_set_err("'-takeover_state' is a required parameter");
                return {};
            }
            next;
        } elsif (/-?takenover_filer/) {
            $takenover_filer = $args{$_};
            next;
        }
    }

    # Make sure user passed in the takeover state
    if (not defined($takeover_state)) {
        $self->_set_err("'-takeover_state' is a required parameter");
        return {};
    }
    if ($takeover_state == 1) {
        if (not defined($takenover_filer)) {
            $self->_set_err("'-takenover_filer' need to be passed if takeover_state is set");
        } else {
            $self->sanlog(-msg => "Taken over filer is '$takenover_filer'");
        }
    }

    # get the client object
    $clientObj =  $self->{parent}->{parent}->{parent};

    # get HU path for path's to binaries
    my $hu_path = $clientObj->get_hu_bin_path();

    # need normal sanlun output to get all the paths to all the devices
    $cmd = $hu_path . "/sanlun lun show";
    ($status, @lun_show_output) = &ClientUtil::execute_retry($conn, $cmd, 180, 6, "-warn");
    if ($status > 0) {
        $self->_set_err("Error while calling $cmd");
        return {};
    }

    # Get all the paths to check for io if protocol is iSCSI or
    # Get all the primary paths if controller is not in takeover state or
    # in takeover state get all the secondary paths for taken over controller
    # and all primary paths for other controller
    my $protocol = $clientObj->{protocol};
    if ($protocol =~ /iscsi/i) {
        @paths_to_check_for_io = $self->_get_paths(-type => "all");
    } elsif ($takeover_state == 0) {
        @paths_to_check_for_io = $self->_get_paths(-type => "primary");
    } else {
        foreach my $lun_info (@lun_show_output) {
            # Skip the first line because this line is not the actual lun info
            next if ($lun_info =~ /^\s+controller/i);
            my ($filer, undef, $path, undef, undef, undef) = split (/\s+/, $lun_info);
            $path =~ s/\/dev\///;
            my $wwid = $clientObj->_get_wwid($path);
            if ($lun_info =~ /$takenover_filer/i) {
                if ($self->{$wwid}->{paths}->{$path}->{type} eq "secondary") {
                    push(@paths_to_check_for_io, $path);
                }
            } else {
                if ($self->{$wwid}->{paths}->{$path}->{type} eq "primary") {
                    push(@paths_to_check_for_io, $path);
                }
            }
        }
    }

    # Initialize all path's iostats to ZERO
    foreach my $path (@paths_to_check_for_io) {
        $paths_stats->{$path}->{blk_read} = 0;
        $paths_stats->{$path}->{blk_wrtn} = 0;
    }
    # Display 13 reports at five second intervals for $all_pri_paths
    # For a total of 1 minuts. First report it displays immediately
    # Here is an example of iostat output
    #--------------------------------------------------------------
    # Linux 2.6.16.60-0.21-smp (IBMx336-200-185)      05/18/09
    #
    # avg-cpu:  %user   %nice %system %iowait  %steal   %idle
    #           68.72    0.00   18.69    8.73    0.00    3.85
    # Device:            tps   Blk_read/s   Blk_wrtn/s   Blk_read   Blk_wrtn
    # sdg             190.00         1.72      3675.29      34960   74760568
    # sdh             190.05         1.71      3673.60      34872   74726264
    # sdi             189.93         1.72      3670.03      35040   74653552
    # sdj             189.82         1.73      3669.03      35232   74633280
    # sdk             190.19         1.73      3683.70      35288   74931672
    # sdl             189.94         1.72      3681.13      35080   74879376
    # sdm             189.68         1.69      3676.09      34472   74776952
    # sdn             189.62         1.71      3674.85      34736   74751672
    #--------------------------------------------------------------
    #
    # sysstat rpm is need to be installed on the host to run the following cmd
    my $all_paths = join (" ", @paths_to_check_for_io);
    $cmd = "/usr/bin/iostat 5 13 $all_paths > /tmp/path_stats";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        if (grep(/command\s+not\s+found|command-not-found/i, @output)) {
            $self->sanlog(-msg => "Install 'sysstat' rpm on the host $clientObj->hostname()");
        }
        $self->_set_err("Unable to run command $cmd");
        return {};
    }

    $cmd = "cat /tmp/path_stats";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to run command $cmd");
        return {};
    }

    # The  first  report  generated  by  the  iostat command provides statistics
    # concerning the time since the system was booted. Each subsequent
    # report covers the time since the previous report. Thats why we dont
    # add io stats until second cycle of iostat starts
    my $counter = 0;
    foreach my $line (@output) {
        if (grep (/^avg-cpu:/i, $line)) {
            $counter = $counter + 1;
        }
        next if ($counter < 2);
        if (grep (/^sd\w+/, $line)) {
            my ($path, undef, undef, undef, $blk_read, $blk_wrtn) = split(/\s+/, $line);
            $paths_stats->{$path}->{blk_read} = $paths_stats->{$path}->{blk_read} + $blk_read;
            $paths_stats->{$path}->{blk_wrtn} = $paths_stats->{$path}->{blk_wrtn} + $blk_wrtn;
        }
    }

    return $paths_stats;
}

################################################################################
# Subroutine Name: create_encrypted_device()
# Objective:       Will make the device as encrypted device
# Details:         See POD documentation at the end of this file
################################################################################
sub create_encrypted_device {
    my ($self, %args) = @_;

    my ($device, $encrypted_dev, $clientObj);

    $clientObj = $self->{parent}->{parent}->{parent};
    $self->_clear_err();

    foreach (keys %args) {
        if (/-?device/) {
            if (not length($device = $args{$_})) {
                $self->_set_err("Invalid value for argument '-device'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Make sure that required parameters are passed
    if (not defined($device)) {
        $self->_set_err("'-device' is a required parameter");
        return $ClientUtil::FALSE;
    }

    # Create encrypted device using LUKS method
    if ($clientObj->{encryption_method} =~ /LUKS/i) {
        $self->_create_encrypted_device_luks($device);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _create_encrypted_device_luks()
# Objective:       Will create encrypted device using LUKS
# Details:         See POD documentation at the end of this file
################################################################################
sub _create_encrypted_device_luks {
    my ($self, $device) = @_;

    my ($cmd, @output, $status, $conn);
    my ($clientObj, $encrypted_dev);

    $self->_clear_err();
    $conn = $self->{conn};

    # Make sure device has full path. If not prepend the path
    if (not grep(/\/dev\/mapper/, $device)) {
        $device = "/dev/mapper/" . $device;
    }

    # Format the device with LUKS encryption
    # /var/tmp/nate_crypt_luks.key key will be used to do the encryption
    $cmd = "/sbin/cryptsetup -q luksFormat $device /var/tmp/nate_encryption_luks.key";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to create LUKS encrypted device");
        return $ClientUtil::FALSE;
    }

    my ($tmp) = ($device =~ /\/dev\/mapper\/(\w+)/);
    # Will prepend "crypt" to multipath device name
    # to show that these are encrypted devices
    $encrypted_dev = "crypt-" . $tmp;
    # Open the encrypted device with new name
    $cmd = "/sbin/cryptsetup -q luksOpen $device $encrypted_dev --key-file /var/tmp/nate_encryption_luks.key";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Unable to open the LUKS encrypted device");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _get_paths()
# Objective:       Return all the primary or secondary paths of all devices
#                  depending on the value of the argument -type
#                  These paths will be used to see the io activity on individual
#                  paths in takeover mode when cmd "multipath -ll" timed-out
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_paths {
    my ($self, %args) = @_;

    my ($cmd, $device, @devices, @paths, $type, @output, $status, $conn);
    my ($clientObj);

    $self->_clear_err();
    $conn = $self->{conn};
    $clientObj = $self->{parent}->{parent}->{parent};

    foreach (keys %args) {
        if (/-?type/) {
            if (not length($type = $args{$_})) {
                $self->_set_err("'-type' is a required parameter");
                return (wantarray()) ? () : [];
            }
            next;
        }
    }

    # Make sure user passed in the path type
    if (not defined($type)) {
        $self->_set_err("'-type' is a required parameter");
        return (wantarray()) ? () : [];
    }

    if ($clientObj->{client_dmmp_use_user_friendly_name} =~ /^1$|^yes$/i) {
        $cmd = "ls /dev/mapper/* | grep 'mpath'";
    } else {
        $cmd = "ls /dev/mapper/* | grep -P \"360a9|3600a09\"";
    }
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0 or scalar (@output) == 0) {
        $self->_set_err("No dm devices found");
        return (wantarray()) ? () : [];
    }
    # Remove /dev/mapper part from device's name
    foreach my $name (@output) {
        $name =~ s/\/dev\/mapper\///;
        push (@devices, $name);
    }

    # Store all the paths along with their type(primary/secondary/all)
    foreach $device (@devices) {
        foreach my $path (keys %{$self->{$device}->{paths}}) {
            if ($type eq "all") {
                push (@paths, $path);
            } elsif ($self->{$device}->{paths}->{$path}->{type} eq $type) {
                push (@paths, $path);
            }
        }
    }

    return (wantarray()) ? @paths : \@paths;
}

################################################################################
# Subroutine Name: _store_status_of_each_path()
# Objective:       Store the state of the path whether path is dead or alive
# Details:         See POD documentation at the end of this file
################################################################################
sub _store_status_of_each_path {
    my ($self) = @_;

    my ($cmd, $device, @devices, @paths, $type, @output, $status, $conn, $clientObj);

    $conn = $self->{conn};
    $self->_clear_err();
    $clientObj = $self->{parent}->{parent}->{parent};

    # get all the paths of type=>$type for all the devices
    @paths = $self->_get_paths(-type => "all");
    return $ClientUtil::FALSE if $self->error();

    my $paths_dead = 0;
    my $paths_alive = 0;
    foreach my $path (@paths) {
        # Get wwid of the path
        my $wwid = $clientObj->_get_wwid($path);
        if ($self->error) {
            return $ClientUtil::FALSE;
        }

        # This cmd could fail because of any of the following reasons
        # Make sure to mark the path as dead for correct reasons
        $cmd = "/usr/bin/sg_turs -v /dev/$path";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status > 0) {
            if (grep(/command\s+not\s+found|command-not-found/i, @output)) {
                # if required rpms are not installed for this commnad
                $self->_set_err("sg3_utils rpms are not installed on this host. " .
                                "User has to manually install rpm(s) to run this cmd");
                return $ClientUtil::FALSE;
            } elsif (grep(/no\s+such\s+file\s+or\s+directory/i, @output)) {
                # if path is missing
                $self->_set_err("device /dev/$path is missing");
                return $ClientUtil::FALSE;
            } else {
                # If both above conditions are not true then path is dead
                $self->{$wwid}->{num_paths_dead}++;
                $paths_dead = $paths_dead + 1;
            }
        } else {
            # if the return status is 0 then path is alive
            $self->{$wwid}->{num_paths_alive}++;
            $paths_alive = $paths_alive + 1;
        }
    }

    # store total dead and alive paths
    $self->{paths}->{status}->{total_dead} = $paths_dead;
    $self->{paths}->{status}->{total_alive} = $paths_alive;

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _store_all_paths_type()
# Objective:       Store all the paths of all devices along with their types
# Details:         See POD documentation at the end of this file
################################################################################
sub _store_all_paths_type {
    my ($self) = @_;

    my ($cmd, $device, $type, @output, $status, $conn, $clientObj, $test_bin);

    $conn = $self->{conn};
    $clientObj = $self->{parent}->{parent}->{parent};
    $test_bin = $clientObj->{TEST_BIN};

    $cmd = "ls /sys/block | egrep sd.*";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("No scsi device found");
        return $ClientUtil::FALSE;
    }

    foreach my $path (@output) {
        my $wwid = $clientObj->_get_wwid($path);
        if ($self->error) {
            return $ClientUtil::FALSE;
        } elsif ($wwid !~ /^(360a9\w+|3600a09\w+|mpath\w+)/) {
            next;
        }

        # determine whether the path is primary or secondary (proxy)
        # here is the output of 'spt' cmd
        #------------------------------------------------------------
        # If path is primary:
        #------------------------------------------------------------
        # Data Read for SCSI opcode 0xc0, 16 bytes:
        #
        # Off  Data
        # 0000 00 00 00 10 0a 98 0a 02 21 00 00 08 30 62 00 00 "        !   0b  "
        #------------------------------------------------------------
        # If path is secondary:
        #------------------------------------------------------------
        # Data Read for SCSI opcode 0xc0, 20 bytes:
        #
        # Off  Data
        # 0000 00 00 00 14 0a 98 0a 02 21 00 00 0c 76 2e 30 62 "        !   v.0b"
        # 0010 00 00 00 00                                     "    "
        #
        $cmd = "$test_bin/spt dsf=/dev/$path cdb=\"c0 00 02 0a 98 0a 21 00 ff 00\" dir=read length=255";
        my ($status, @spt_output) = &ClientUtil::execute($conn, $cmd);
        if ($status > 0) {
            $self->_set_err("Unable to run spt for the device /dev/$path");
            return $ClientUtil::FALSE;
        }

        # Output will contain "v" in it if path is secondary
        if (grep(/v\./, @spt_output) > 0) {
            $type = "secondary";
        } else {
            $type = "primary";
        }
        $self->{$wwid}->{paths}->{$path}->{type} = $type;
    }

    return $ClientUtil::TRUE;
}
1;
