# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools
## @author latif@netapp.com, rawat@netapp.com, thomasm@netapp.com
## @status Private
## @pod ../none_yet
# Package Name
##############
package San::SanClient::MP::LinuxMP::DMP;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::LinuxMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       enable DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "/sbin/vxconfigd -k -m enable";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);

    if ($status > 0) {
        $self->_set_err("Error while enabling DMP (" . @output . ")");
        return $ClientUtil::FALSE;
    }

    #### Verify that DMP daemon is running
    $self->_is_dmp_running(-conn => $conn);
    if ($self->error) {
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: flush()
# Objective:       No-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub flush {
    my ($self) = @_;
    # Noop
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: start_multipath_daemon()
# Objective:       No-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub start_multipath_daemon {
    my ($self) = @_;

    # Noop

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_size()
# Objective:       No-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub get_size {
    my ($self) = @_;

    # Noop

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: reconfigure()
# Objective:       No-OP
# Details:         See POD documentation at the end of this file
################################################################################
sub reconfigure {
    my ($self) = @_;

    # Noop

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: disable()
# Objective:       disable DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "/sbin/vxconfigd -k -m disable";

    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error while disabling dmp");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: recreate_maps()
# Objective:       No-op
# Details:         See POD documentation at the end of this file
################################################################################
sub recreate_maps {
    my ($self, %args) = @_;
    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_mp_devices()
# Objective:       get all mp devices
# Details:         See POD documentation at the end of this file
################################################################################
sub get_mp_devices {
    my ($self, %args) = @_;

    my ($cmd, @output, @hDevices, $conn, $status);
    my (@grep_netepp_devices);

    $self->_clear_err();

    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value " . "for argument '-conn' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Enabling encloser based naming
    $cmd = "/usr/sbin/vxddladm set namingscheme=ebn";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Error in setting the Naming Scheme");
        return $ClientUtil::FALSE;
    }

    # Grab the vxdisk -o alldgs list output
    # DEVICE       TYPE            DISK         GROUP        STATUS
    # das_fcp_0_0  auto            -            -            error
    # das_fcp_0_1  auto            -            -            error
    # das_fcp_0_2  auto            -            -            error
    # das_fcp_0_3  auto            -            -            error
    # das_iscs0_0  auto            -            -            error
    # das_iscs0_1  auto            -            -            error
    # das_iscs0_2  auto            -            -            error
    # das_iscs0_3  auto            -            -            error
    # sda          auto:none       -            -            online invalid
    my @vxdisk_list_output;
    $cmd = "/sbin/vxdisk -o alldgs list";
    ($status, @vxdisk_list_output) = &ClientUtil::execute($self->{conn}, $cmd);
    if ($status != 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return (wantarray()) ? () : [];
    }

    foreach my $line (sort @vxdisk_list_output) {
        $line =~ s/^\s+//g;
        next if ($line =~ /^DEVICE/);

        # Skip any devices flagged as boot, root, or reserved
        next if ($line =~ /boot|root|reserved/i);

        my ($dmpnode, $type, $disk, $diskgroup, @init_state) = split(/\s+/, $line);
        # Update $diskgroup to remove any ()
        $diskgroup =~ s/\(|\)//;

        # Skip any disks who have a disk group defined that isn't "-";
        if ($diskgroup ne "-") {
            $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since the it is used in diskgroup '$diskgroup'");
            next;
        }

        # Make sure type is "auto"
        unless ($type =~ /auto/) {
            $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since the type isn't 'auto'");
            next;
        }

        # Now that we have the dmpnode, run vxdisk list on that node.
        # and find out primary/secondary status

        my @vxdisk_list_node_output;
        $cmd = "/sbin/vxdisk list " . $dmpnode;
        ($status, @vxdisk_list_node_output) = &ClientUtil::execute($self->{conn}, $cmd);
        if ($status > 0) {
            $self->_set_err("Command ($cmd) returned a bad status ($status)");
            return (wantarray()) ? () : [];
        }

        # Output should be
        # [root@auto-rhel5u6-02 ~]# vxdisk list das_fcp_0_0
        # Device:    das_fcp_0_0
        # devicetag: das_fcp_0_0
        # type:      auto
        # flags:     error private autoconfig
        # pubpaths:  block=/dev/vx/dmp/das_fcp_0_0 char=/dev/vx/rdmp/das_fcp_0_0
        # guid:      -
        # udid:   dm list dmpnode dmpnodename=   NETAPP%5FLUN%20C-Mode%5F2FiIn%3F-lCptj%5F2FiIn%3F-lCptm
        # site:      -
        # Multipathing information:
        # numpaths:   4
        # sdj             state=enabled   type=primary
        # sdn             state=enabled   type=secondary
        # sdb             state=enabled   type=secondary
        # sdf             state=enabled   type=primary
        my $skip_dev_flag = 0;
        foreach my $line (@vxdisk_list_node_output) {
            chomp($line);
            if ($line =~ /^udid\:\s+(.+)/) {
                # Make sure this is a netapp disk
                my $udid = $1;
                unless (($udid =~ /netapp/i) and ($udid =~ /lun/i)) {
                    $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since it isn't a NETAPP lun");
                    # Set $skip_dev_flag as 1 to skip the device in the outer loop
                    $skip_dev_flag = 1;
                    last;
                }
            }
            if ($line =~ /^flags\:\s+(.+)/i) {
                my $flags = $1;
                if ($flags =~ /coordinator/i) {
                    # Disk has coordinator flag, which means it is a fencing device, skip it
                    $self->sanlog(-msg => "Skipping dmpnode '$dmpnode' since it is a fenced device");
                    # Set $skip_dev_flag as 1 to skip the device in the outer loop
                    $skip_dev_flag = 1;
                    last;
                }
            }
        }

        if ($skip_dev_flag == 1) {
            # $skip_dev_flag is set, skipping dmp device
            next;
        }

        my $hDevs = {};
        $hDevs->{name}  = $dmpnode;
        $hDevs->{path}  = "";
        $hDevs->{label} = "free";
        push @hDevices, $hDevs;

        # Need to store a single sdX device(a path) that corresponds to
        # mpath device. This mapping will be used in LinuxDev.pm _new() when
        # get_lun_path is called. It will be used to get the lun path on filer side
        # Here is the sample output of commnad
        # "vxdmpadm list dmpnode dmpnodename=<dmpnodename> |grep path"
        #--------------------------------------------------------------------
        # ###path         = name state type transport ctlr hwpath aportID aportWWN attr
        # path            = sdbu enabled(a) - iSCSI c16 c16 010.072.201.071:03260 10.72.201.71:3260 -
        # path            = sdci enabled(a) - iSCSI c17 c17 fd20.8b1e.b255.40c8.02a0.98ff.fe08.6a60:03260 244.16.240.208:3104 -
        # path            = sdcw enabled(a) - iSCSI c18 c18 192.168.200.096:03260 192.168.200.96:3260 -
        # path            = sdbg enabled(a) - iSCSI c15 c15 192.168.200.071:03260 192.168.200.71:3260 -
        #--------------------------------------------------------------------

        $cmd = "vxdmpadm list dmpnode dmpnodename=$dmpnode | grep ^path";
        my ($status, @tmp_arr) = &ClientUtil::execute($conn, $cmd);
        if ($status > 0) {
            $self->_set_err("No valid Vxdisk found");
            return $ClientUtil::FALSE;
        }
        my (undef, undef, $path) = split(/\s+/, $tmp_arr[0]);
        $self->{mapping}->{$dmpnode}->{sdX} = $path;
    }

    return (wantarray()) ? @hDevices : \@hDevices;
}

################################################################################
# Subroutine Name: verify()
# Objective:       verify whether the required packaged are installed or not
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output, @asl_pkgs, @full_pkgs);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Verify that the ASL/APM is installed (in case of versions < 5.1)
    # Case of 5.1+, support is in VRTS package itself.
    if ($self->{version} >= 5.1) {
        @full_pkgs = ("VRTSvxvm", "VRTSvxfs", "VRTSaslapm");
    } else {
        @full_pkgs = ("VRTSvxvm-common","VRTSvxfs-common","VRTSvxfs-platform","VRTSvxvm-platform","VRTSNTAPasl","VRTSNTAPapm");
    }

    ### Verify that package for DMP is installed
    foreach my $pkg (@full_pkgs) {
        $cmd = "rpm -q $pkg";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to verify packaged ($pkg) is installed");
            return $ClientUtil::FALSE;
        } else {
            $self->sanlog(-msg => "LinuxMP::DMP::verify() Verified that package ($pkg) is installed");
        }
    }

    # Now if this is 5.1, verify that the older requirements/packages
    # required with 5.0 are NOT installed.
    if ($self->{version} >= 5.1) {
        @asl_pkgs = ("VRTSNTAPasl", "VRTSNTAPapm");
        foreach my $pkg (@asl_pkgs) {
            $cmd = "rpm -q $pkg";
            ($status, @output) = &ClientUtil::execute($conn, $cmd);
            if ($status == 0) {
                $self->_set_err("Failed to verify packaged ($pkg) is NOT installed\n" .
                                "VRTSNTAPasl and VRTSNTAPapm should NOT be installed with VRTS 5.1 and greater");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       #no-op for linux
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {

    #no-op for linux

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: stop_multipath_daemon()
# Objective        Will stop the multipath daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub stop_multipath_daemon {

    #no-op for linux
    return $ClientUtil::TRUE;
}
################################################################################
# Subroutine Name: config()
# Objective:       #no-op for linux
# Details:         See POD documentation at the end of this file
################################################################################
sub config {

    #no-op for linux

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # no-op for linux

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each DMP node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($per_lun_hash, $cmd, @vx_output, @output, $status, $conn);
    my ($line, $enclosure, @encl_output);
    my ($dmp_node, $state, $num_paths, $num_alive, $num_dead);
    my (@lun_show_output, @vxdisk_output);
    my ($validate_mode);

    # parse arguments
    foreach (keys %args) {
        if (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # get HU path for path's to binaries
    my $hu_path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();

    $per_lun_hash = {};
    @output = ();

    # get connection from object
    $conn = $self->{conn};

    # run "vxdmpadm listenclosure all" to get list of enclosures
    $cmd = "/sbin/vxdmpadm listenclosure all | grep -v 'DISCONNECTED' | grep 'CONNECTED' | grep -v 'DISKS'";
    ($status, @vx_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                       180, 5, "-warn");
    if ($status != 0) {
        $self->_set_err("Failed to run vxdmpadm listenclosure all");
        return {};
    }

    foreach $line (@vx_output) {
        $enclosure = (split(/\s+/, $line))[0];
        if ($enclosure ne "") {
            $cmd = "/sbin/vxdmpadm getdmpnode enclosure=$enclosure";
            ($status, @encl_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                 180, 5, "-warn");
            if ($status != 0) {
                $self->_set_err("Failed to run $cmd to get dmpnode and path information");
                return {};
            }
            push(@output, @encl_output);
        }
    }

    # only grab sanlun to get protocol if we are in validation mode
    # meaning we are checking paths. We will get the protocol only
    # in steady state, since sanlun is very slow in takover mode
    if ($validate_mode == 0) {
        # need normal sanlun output to get protocol
        $cmd = $hu_path . "/sanlun lun show";
        ($status, @lun_show_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                 360, 6, "-warn");
        if ($status != 0) {
            $self->_set_err("Error while calling $cmd");
            return {};
        }

        # need vxdisk -e list output to map the enclosure based name
        $cmd = "vxdisk -e list";
        ($status, @vxdisk_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                               180, 6, "-warn");
        if ($status != 0) {
            $self->_set_err("Error while calling $cmd");
            return {};
        }
    }

    # go over the enclosure output and pick apart each line (which
    # will be the dmp_node followed by it's characteristics
    # Example:
    #  NAME                 STATE     ENCLR-TYPE   PATHS  ENBL  DSBL  ENCLR-NAME
    # =========================================================================
    # c5t0d0s2             ENABLED      FAS960       4      4     0     FAS9600

    foreach $line (@output) {

        # skip header and newlines
        if ($line =~ /ENCLR-TYPE/i or
            $line =~ /=====/i or
            $line =~ /^\s*$/) {
            next;
        }

        # get per dmp node pathing information from output
        ($dmp_node, $state, undef,
         $num_paths, $num_alive, $num_dead, undef) = split(/\s+/, $line);

        # store number of paths
        $per_lun_hash->{$dmp_node}->{num_paths} = $num_paths;

        # store number of paths that are enabled (alive)
        $per_lun_hash->{$dmp_node}->{num_paths_alive} = $num_alive;

        # store number of paths that are disabled (dead)
        $per_lun_hash->{$dmp_node}->{num_paths_dead} = $num_dead;

        # store number of paths that are primary
        $per_lun_hash->{$dmp_node}->{num_paths_primary} = "NA";

        # store number of paths that are secondary
        $per_lun_hash->{$dmp_node}->{num_paths_secondary} = "NA";

        # only set protocol in validation mode (comparing paths)
        if ($validate_mode == 0) {
            # get protocol from the sanlun output
            my ($protocol);
            # lookup the device handle from getdmp node in
            # vxdisk -e list ouput to get the actual device
            my $vxdisk_line = (grep(/$dmp_node/, @vxdisk_output))[0];
            my $device = (split(/\s+/, $vxdisk_line))[5];

            # lookup that device in sanlun output to get protocol
            my $sanlun_dev = (grep(/$device/, @lun_show_output))[0];
            if ($sanlun_dev =~ /\s+iscsi\s+/i) {
                $protocol = "iSCSI";
            } else {
                $protocol = "FCP";
            }
            $per_lun_hash->{$dmp_node}->{protocol} = $protocol;
        } else {
            $per_lun_hash->{$dmp_node}->{protocol} = "takeover";
        }

    }

    return $per_lun_hash;

}

################################################################################
# P R I V A T E   M E T H O D S
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;
    my ($self, $parent, $obj_template);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument \'-parent\'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents first
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }

        # store the parent and return
        $self->{parent} = $parent;
        return $self;
    }

    $self->{parent} = $parent;

    $self->_version_info(-conn => $self->{parent}->{parent}->{parent}->{conn});
    if ($self->error) {
        return $self;
    }

    # Returning DMP object
    return $self;

}

################################################################################
# Subroutine Name: _version_info()
# Objective:       returns the version of Veritas running on the system
# Details:         See POD documentation at the end of this file
################################################################################
sub _version_info {
    my ($self, %args) = @_;

    my ($conn, $cmd, @output, $status);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my $vers = "";
    my $rev  = "";
    my $sp   = "";
    my $rp   = "";
    my $mp   = "";

    $cmd = "rpm -qa VRTSvxvm VRTSvxvm-common";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to execute command $cmd");
        return $ClientUtil::FALSE;
    }

    # format of rpm returns:
    # 6.0
    # RHEL:
    # VRTSvxvm-6.0.000.000-internal_RHEL5
    # SLES:
    # Unknown
    #
    # 5.1SP1
    # RHEL:
    # VRTSvxvm-5.1.100.000-SP1_RHEL5
    # SLES:
    # VRTSvxvm-5.1.100.000-SP1_SLES10
    #
    # 5.1RP1:
    # From RHEL5U5 host:
    # VRTSvxvm-5.1.001.000-RP1_RHEL5
    #
    # From SLES11 host:
    # VRTSvxvm-5.1.001.000-RP1_SLES11
    #
    # 5.0MP4:
    # RHEL:
    # VRTSvxvm-platform-5.0.40.00-MP4_RHEL5
    # VRTSvxvm-common-5.0.40.00-MP4_RHEL5
    #
    # SUSE:
    # VRTSvxvm-common-5.0.40.00-MP4_SLES10
    # VRTSvxvm-platform-5.0.40.00-MP4_SLES10
    #
    # VRTSvxvm-5.1.00.01-P1_RHEL5

    foreach my $line (@output) {
        if ($line =~ /VRTSvxvm/i) {
            # We need to normalize the line to remove the VRTSvxvm-(common|platform)? stuff
            $line =~ s/^\s*(VRTSvxvm\-)((common|platform)\-)?//gi;

            # Set $rev to $line for the full revision
            $rev  = $line;

            # Look up the version
            ($vers, undef) = $line =~ /^(\d+\.\d+\.\d+\.\d+)(\-.*)/i;

            # Look up MP or P info
            (undef, $mp) = $line =~ /^(\d+\.\d+\.\d+\.\d+\-)(MP\d+|P\d+).*/i;

            # Look for RP info
            (undef, $rp) = $line =~ /^(\d+\.\d+\.\d+\.\d+\-)(RP\d+).*/i;

            # Look for SP info
            (undef, $sp) = $line =~ /^(\d+\.\d+\.\d+\.\d+\-)(SP\d+).*/i;

            last;
        }
    }

    $self->{name}     = "VRTSvxvm";
    $self->{version}  = $vers;
    $self->{revision} = $rev;
    $self->{mp}       = $mp;
    $self->{rp}       = $rp;
    $self->{sp}       = $sp;
    $self->{pkginfo}  = \@output;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _is_dmp_running()
# Objective:       returns ClientUtil::TRUE or ClientUtil::FALSE depending on
#                  if dmp is running or not.
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_dmp_running {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check status of dmp service
    my $cmd = "/sbin/vxdctl mode |grep enabled";
    my ($status, @output) = ClientUtil::execute($conn, $cmd);

    if ($status == 0) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }
}

################################################################################
# Subroutine Name: _dmp_version()
# Objective:       returns the version of Veritas running on the system
# Details:         See POD documentation at the end of this file
################################################################################
sub _dmp_version {
    my ($self, %args) = @_;

    my ($rev, $vers, $cmd, @output, $status, $conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->_clear_err();

    $vers = $rev = "";

    $cmd = "rpm -qa VRTSvxvm";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to execute command $cmd");
        return ("", "");
    }

    # format of rpm return:
    # VRTSvxvm-6.0.000.000-internal_RHEL5
    # VRTSvxvm-5.1.00.01-P1_RHEL5
    foreach my $line (@output) {
        if ($line =~ /VRTSvxvm/i) {
            $line =~ s/VRTSvxvm\-//;
            $line =~ m/(\d+\.\d+\.\d+\.\d+)(\-.*)/i;
            $vers = $1;
            $rev = $1 . $2;
            last;
        }
    }

    return ($vers, $rev);
}

1;
