# Copyright (c) 2007-2011 NetApp Inc.
# All rights reserved.
#
## @summary Provide an interface to multipath tools
## @author latif@netapp.com, thomasm@netapp.com
## @status Private
## @pod ../none_yet

# Package Name
##############
package San::SanClient::MP::SolarisMP::DMP;

use strict;

# Package Requirements
######################
use San::SanClient::ClientUtil;
use San::SanClient::MP::SolarisMP;
use Error;
use San::SanLog;

use vars qw(@ISA);

@ISA = qw(Error SanLog MP);

################################################################################
# P U B L I C   M E T H O D S
################################################################################

################################################################################
# Subroutine Name: enable()
# Objective:       enable DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub enable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "vxdctl enable";
    ($status, @output) = &ClientUtil::execute($conn , $cmd);
    if ($status > 0) {
        $self->_set_err("Error while enabling vxdctl (". @output . ")");
        return $ClientUtil::FALSE;
    }

    #### Verify that vxdctl daemon is running
    $self->_is_vxdctl_running(-conn => $conn);
    if ($self->error) {
        $self->_set_err("_is_vxdctl_running reported error");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: disable()
# Objective:       disable DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub disable {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $cmd = "vxdctl disable";
    ($status, @output) = &ClientUtil::execute($conn , $cmd);
    if ($status > 0) {
        $self->_set_err("Error while disabling vxdctl");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: verify()
# Objective:       verify proper enviroment for DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub verify {
    my ($self, %args) = @_;

    my (@pkgs, $conn, $status, $cmd, @output);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # Verify that package for DMP is installed
    foreach my $pkg ("VRTSvxvm", "VRTSvlic") {
        $cmd = "pkginfo -l $pkg";
        ($status, @output) = &ClientUtil::execute($conn , $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to verify packaged ($pkg) is installed");
            return $ClientUtil::FALSE;
        } else {
            $self->sanlog(-msg => "SolarisMP::DMP::verify() Verified that package ($pkg) is installed",
                                                        -msg_level => 9);
        }
    }

    ### Verify that the DMP module is loaded on the client
    $cmd = "modinfo | grep -i vxdmp";
    ($status, @output) = &ClientUtil::execute($conn , $cmd);
    if ($status != 0) {
        $self->_set_err("Failed to verify that DMP daemon is running (".@output.")");
        return $ClientUtil::FALSE;
    }

    # Verify that the ASL/APM is installed (in case of versions < 5.1).
    # Case of 5.1+, support is in VRTS package itself.
    if ($self->{version} >= 5.1) {
        @pkgs = ("VRTSaslapm");
    } elsif ($self->{version} =~ /^5/) {
        @pkgs = ("VRTSNTAPasl", "VRTSNTAPapm");
    } else {
        @pkgs = ("NTAPasl");
    }

    foreach my $pkg (@pkgs) {
        $cmd = "pkginfo -l $pkg";
        ($status, @output) = &ClientUtil::execute($conn, $cmd);
        if ($status != 0) {
            $self->_set_err("Failed to verify packaged ($pkg) is installed");
            return $ClientUtil::FALSE;
        }
    }

    # Now if this is 5.1, verify that the older requirements/packages
    # required with 5.0 are NOT installed.
    if ($self->{version} >= 5.1) {
        @pkgs = ("VRTSNTAPasl", "VRTSNTAPapm");
        foreach my $pkg (@pkgs) {
            $cmd = "pkginfo -l $pkg";
            ($status, @output) = &ClientUtil::execute($conn, $cmd);
            if ($status == 0) {
                $self->_set_err("Failed to verify packaged ($pkg) is NOT installed\n" .
                                "VRTSNTAPasl and VRTSNTAPapm should NOT be installed with VRTS 5.1 and greater");
                return $ClientUtil::FALSE;
            }
        }
    }

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: setup()
# Objective:       setup proper enviroment for DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub setup {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
        else {
            $self->_set_err("Invalid parameter specified \'$_\'");
            return $ClientUtil::FALSE;
        }
    }

    return $self;
}

################################################################################
# Subroutine Name: config()
# Objective:       verify proper enviroment for DMP daemon
# Details:         See POD documentation at the end of this file
################################################################################
sub config {
    my ($self, %args) = @_;

    my ($conn, $status, $cmd, @output, $protocol);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?protocol/) {
            if (not length($protocol = $args{$_})) {
                $self->_set_err("Invalid value for argument '-protocol' => $args{$_}");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # make sure that the system is NOT configured for MPXIO (stmsboot). If it is,
    # disable and next reboot will take care of the rest. Do only on Solaris 10.
    if ($self->{parent}->{parent}->{parent}->{revision} =~ /5\.10/i) {
        my $filename = "/tmp/stmsboot_input.txt";
        my @stms_cmds = ( "y", "n" );
        &ClientUtil::create_remote_file($conn, $filename, @stms_cmds);

        # Solaris 10 U5 changed how stmsboot is called, account for that here
        if ($self->{parent}->{parent}->{parent}->{minor_revision} >= 5) {
            $cmd = "stmsboot -D fp -d < $filename";
        } else {
            $cmd = "stmsboot -d < $filename";
        }
        ($status, @output) = &ClientUtil::execute($conn , $cmd);

        if ($status > 0) {
            $self->_set_err("Error while disabling MPXIO");
            return $ClientUtil::FALSE;
        }

        &ClientUtil::remove_remote_file($conn, $filename);

        if ($protocol =~ /iscsi/) {

            # Get the HU toolkit version
            my $hu_version = $self->{parent}->{parent}->{parent}->get_hu_toolkit_version();
            unless (defined($hu_version) and $hu_version ne "") {
                $self->_set_err("Unable to determine the Host Utility version");
                return $ClientUtil::FALSE;
            }

            if ($hu_version < 6.0) {
                # run mpxio_set -df to remove MPXIO settings if done already
                # (if platforms was iSCSI and mpxio_setup -e was run previously)
                my $path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();
                $cmd = $path . "/mpxio_set -df";
                ($status, @output) = &ClientUtil::execute($conn , $cmd);
                if ($status != 0 and $status != 1) {
                    $self->_set_err("Failed to execute command ($cmd:" . @output . ")");
                    return $ClientUtil::FALSE;
                }
            }

            # Disable MPXIO
            my $sanclient_obj = $self->{parent}->{parent}->{parent};
            $sanclient_obj->_modify_iscsi_mpxio(-state => "disable");
            if ($sanclient_obj->error()) {
                $self->_set_err("Error disabling MPxIO", $sanclient_obj->errmsg());
                return $ClientUtil::FALSE;
            }
        }
    }

    return $self;
}

################################################################################
# Subroutine Name: verify_io_switching_between_path_groups()
# Objective:       Gather the iostats of all paths of type $type of all devices.
#                  type could be primary, secondary or both
# Details:         See POD documentation at the end of this file
################################################################################
sub verify_io_switching_between_path_groups {
    my ($self) = @_;

    # NOOP for Solaris

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: get_path_info
# Objective:       get path information for each DMP node discovered on the
#                  system. Purpose is to print out the pathing information
#                  and generation of an array (than can be stored in the
#                  object with load_path_info on initial load).
# Details:         See POD documentation at the end of this file
################################################################################
sub get_path_info {
    my ($self, %args) = @_;

    my ($per_lun_hash, $cmd, @vx_output, @output, $status, $conn);
    my ($line, $enclosure, @encl_output);
    my ($dmp_node, $state, $num_paths, $num_alive, $num_dead);
    my ($dev_path, $protocol, $start);
    my ($details, @serial, @lun_show_output, @vxdisk_output);
    my ($validate_mode);

    # parse arguments
    $details = 0;
    foreach (keys %args) {
       if (/-?details/) {
            if (not length($details = $args{$_})) {
                $self->_set_err("Invalid value for argument \'-details\'");
                return $ClientUtil::FALSE;
            }
            next;
       } elsif (/-?validate_mode/) {
            if (not length($validate_mode = $args{$_})) {
                $self->_set_err("Invalid value for argument '-validate_mode'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check to see if details was defined and set to TRUE
    if (defined($details) and $details == $ClientUtil::TRUE) {
        $details = 1;
    } else {
        $details = 0;
    }

    # get HU path for path's to binaries
    my $hu_path = $self->{parent}->{parent}->{parent}->get_hu_bin_path();

    $per_lun_hash = {};
    @output = ();

    # get connection from object
    $conn = $self->{conn};

    # run "vxdmpadm listenclosure all" to get list of enclosures
    $cmd = "vxdmpadm listenclosure all | grep -v 'DISCONNECTED' | grep 'CONNECTED' | grep -v 'DISKS'";
    ($status, @vx_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                       180, 5, "-warn");
    if ($status != 0) {
        $self->_set_err("Failed to run vxdmpadm listenclosure all");
        return {};
    }

    foreach $line (@vx_output) {
        $enclosure = (split(/\s+/, $line))[0];
        if ($enclosure ne "") {
            $cmd = "vxdmpadm getdmpnode enclosure=$enclosure";
            ($status, @encl_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                 180, 5, "-warn");
            if ($status != 0) {
                $self->_set_err("Failed to run $cmd to get dmpnode and path information");
                return {};
            }
            push(@output, @encl_output);
        }
    }

    # only grab sanlun to get protocol if we are in validation mode
    # meaning we are checking paths. We will get the protocol only
    # in steady state, since sanlun is very slow in takover mode
    if ($validate_mode == 0) {
        # need normal sanlun output to get protocol
        $cmd = $hu_path . "/sanlun lun show";
        ($status, @lun_show_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                 360, 6, "-warn");
        if ($status != 0) {
            $self->_set_err("Error while calling $cmd");
            return {};
        }

        # need vxdisk -e list output to map the enclosure based name
        $cmd = "vxdisk -e list";
        ($status, @vxdisk_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                               180, 6, "-warn");
        if ($status != 0) {
            $self->_set_err("Error while calling $cmd");
            return {};
        }
    }

    # go over the enclosure output and pick apart each line (which
    # will be the dmp_node followed by it's characteristics
    # Example:
    #  NAME                 STATE     ENCLR-TYPE   PATHS  ENBL  DSBL  ENCLR-NAME
    # =========================================================================
    # c5t0d0s2             ENABLED      FAS960       4      4     0     FAS9600

    foreach $line (@output) {
        # skip header and newlines
        if ($line =~ /ENCLR-TYPE/i or
            $line =~ /=====/i or
            $line =~ /^\s*$/) {
            next;
        }

        # get per dmp node pathing information from output
        ($dmp_node, $state, undef,
         $num_paths, $num_alive, $num_dead, undef) = split(/\s+/, $line);

        # store number of paths
        $per_lun_hash->{$dmp_node}->{num_paths} = $num_paths;

        # store number of paths that are enabled (alive)
        $per_lun_hash->{$dmp_node}->{num_paths_alive} = $num_alive;

        # store number of paths that are disabled (dead)
        $per_lun_hash->{$dmp_node}->{num_paths_dead} = $num_dead;

        # store number of paths that are primary
        $per_lun_hash->{$dmp_node}->{num_paths_primary} = "NA";

        # store number of paths that are secondary
        $per_lun_hash->{$dmp_node}->{num_paths_secondary} = "NA";

        # only set protocol in validation mode (comparing paths)
        if ($validate_mode == 0) {
            # get protocol from the sanlun output
            my ($device, $protocol);
            # lookup the device handle from getdmp node in
            # vxdisk -e list ouput to get the actual device
            my $vxdisk_line = (grep(/$dmp_node/, @vxdisk_output))[0];
            if ($vxdisk_line =~ /thinrclm/i) {
                $device = (split(/\s+/, $vxdisk_line))[6];
            } else {
                $device = (split(/\s+/, $vxdisk_line))[5];
            }

            # lookup that device in sanlun output to get protocol
            my $sanlun_dev = (grep(/$device/, @lun_show_output))[0];
            if ($sanlun_dev =~ /\s+iscsi\s+/i) {
                $protocol = "iSCSI";
            } else {
                $protocol = "FCP";
            }

            # set protocol
            $per_lun_hash->{$dmp_node}->{protocol} = $protocol;
        } else {
            $per_lun_hash->{$dmp_node}->{protocol} = "takeover";
        }
    }

    # generate additional device/path details if requested.
    if ($details == 1) {
         # reset per lun hash since sanlun does NOT always
         # choose the DMP node returned by Veritas. We don't
         # want to use sanlun always because of reliability/speed
         # issues.
         $per_lun_hash = {};

         # get lun show -p to get full path details
         $cmd = $hu_path . "/sanlun lun show -p";
         ($status, @output) = &ClientUtil::execute_retry($conn, $cmd,
                                                         180, 6, "-warn");
         if ($status != 0) {
             $self->_set_err("Error while calling $cmd");
             return {};
         }

         # need normal sanlun output to get protocol
         $cmd = $hu_path . "/sanlun lun show";
         ($status, @lun_show_output) = &ClientUtil::execute_retry($conn, $cmd,
                                                                  180, 6, "-warn");
         if ($status != 0) {
             $self->_set_err("Error while calling $cmd");
             return {};
         }

         my $cnt = 0;
         # traverse lun show -p output to gather details
         foreach my $line (@output) {
            # skip header/blank lines
            if ($line =~ /^----/ or
                $line =~ /^path\s+path/ or
                $line =~ /^state\s+type/i or
                $line =~ /LUN State/i) {
                $cnt++;
                next;
            }

            # next newline after looking at per path details
            # should be next device, stop grabbing per path info
            if ($line =~ /^\s*$/ and $start == 1) {
                $start = 0;
            }

            # check to see if you are starting a new LUN
            if ($line =~ /^\S+:\//) {
                $start = 1;

                $line =~ m/(\S+)\s+\(LUN (\d+)\)\s+(\S+)\s*.*/;
                my $controller_info = $1;
                my $lun_id = $2;

                # $dev_path will be the key to the hash
                $dev_path = $3;
                # strip off /dev/rdsk
                $dev_path =~ s/\/dev\/rdsk\///;

                my ($controller, $controller_path) = split(/:/, $controller_info);

                my $next_line = $output[($cnt + 1)];
                $next_line =~ m/\s+(\S+)\s+\(\d+\)\s+LUN State\:\s+(\S+)/;
                my $lun_size = $1;
                my $lun_state = $2;

                $per_lun_hash->{$dev_path}->{controller} = $controller;
                $per_lun_hash->{$dev_path}->{controller_path} = $controller_path;
                $per_lun_hash->{$dev_path}->{lun_id} = $lun_id;
                $per_lun_hash->{$dev_path}->{lun_state} = $lun_state;
                $per_lun_hash->{$dev_path}->{lun_size} = $lun_size;

                # get serial number by another call to sanlun on per device path
                # Don't check status on purpose.
                $cmd = $hu_path . "/sanlun lun show -v -d /dev/rdsk/$dev_path | grep -i \"serial number\"";
                ($status, @serial) = &ClientUtil::execute_retry($conn, $cmd,
                                                                180, 2, "-warn");
                my $serial_number = $serial[0];
                $serial_number =~ s/\s+Serial number:\s+//;
                $per_lun_hash->{$dev_path}->{serial_num} = $serial_number;

                $cnt++;
                next;
            }

            # start == 1 tell us we should start looking at per path details
            if ($start == 1) {
                my ($p_state, $p_type, $device,
                    $hba, $target_port, $policy) = split(/\s+/, $line);

                # strip off /dev/rdsk
                $device =~ s/\/dev\/rdsk\///;

                # determine proxy/primary
                if ($p_type =~ /secondary/i) {
                    $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{is_proxy} = "True";
                } else {
                    $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{is_proxy} = "False";
                }

                $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{state} = $p_state;
                $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{init_port} = $hba;
                $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{target_port} = $target_port;
                $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{lb_policy} = $policy;
                $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{reserve} = "NA";
                $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{weight} = "NA";
                $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{dsm_id} = "VERITAS DMP";
                $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{state} = $p_state;

                # determine protocol (cannot be determine from -p output, need
                # standard lun show output).
                my $dev_line = (grep(/$device/, @lun_show_output))[0];
                if ($dev_line =~ /FCP/i) {
                   $protocol = "FCP";
                } elsif ($dev_line =~ /iSCSI/i) {
                   $protocol = "ISCSI";
                } else {
                   $protocol = "UNKNOWN";
                }
                $per_lun_hash->{$dev_path}->{per_path_hash}->{$device}->{protocol} = $protocol;
            }

            $cnt++;
        }

        # fill in num_paths / num_alive / num_dead
        foreach my $dev (keys(%{$per_lun_hash})) {
            $per_lun_hash->{$dev}->{num_paths} =
                      scalar(keys(%{$per_lun_hash->{$dev}->{per_path_hash}}));
            my $num_enabled = 0;
            foreach my $path (keys(%{$per_lun_hash->{$dev}->{per_path_hash}})) {
                if ($per_lun_hash->{$dev}->{per_path_hash}->{$path}->{state} =~ /enabled/i) {
                    $num_enabled++;
                }
            }
            $per_lun_hash->{$dev}->{num_paths_alive} = $num_enabled;
            $per_lun_hash->{$dev}->{num_paths_dead} = ($per_lun_hash->{$dev}->{num_paths} -
                                                       $num_enabled);
        }
    }

    return $per_lun_hash;

}

################################################################################
################################################################################
################################################################################
# Subroutine Name: _new()
# Objective:       Create a new mp object
# Details:         See POD documentation at the end of this file
################################################################################
sub _new {
    my ($class, %args) = @_;

    my ($self, $parent, $obj_template);

    $self = {};

    bless $self, $class;

    # parse arguments
    foreach (keys %args) {
        if (/-?parent/) {
            if (not length($parent = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        } elsif (/-?obj_template/) {
            if (not length($obj_template = $args{$_})) {
                $self->_set_err("Invalid value for argument '$_'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # if a template was provided, copy over all the contents and return
    if (defined($obj_template)) {
        foreach (keys %{$obj_template}) {
            $self->{$_} = $obj_template->{$_};
        }
        # store parent
        $self->{parent} = $parent;
        return $self;
    }

    $self->{parent} = $parent;

    # Get the Version information for veritas
    $self->_get_pkginfo(-conn => $self->{parent}->{parent}->{parent}->{conn});
    if ($self->error()) {
        return $self;
    }

    # Returning DMP object
    return $self;
}

################################################################################
# Subroutine Name: _get_pkginfo
# Objective:       Get dmp package info
# Details:         See POD documentation at the end of this file
################################################################################
sub _get_pkginfo {
    my ($self, %args) = @_;

    my ($conn, @output, $status);

    $self->_clear_err();

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    my $vers = "";
    my $rev  = "";
    my $sp   = "";
    my $rp   = "";
    my $mp   = "";
    my $name = "";

    my $cmd = "/usr/bin/pkginfo -l VRTSvxvm";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    foreach my $line (@output) {
        ($name) = $line =~ /PKGINST:\s+(\w+)/i
              if ($line =~ /PKGINST/i);
        ($vers,$rev) = $line =~ /VERSION:\s+(.*),REV=\b(.*)\b/i
              if ($line =~ /VERSION/i);
        if ($line =~ /PSTAMP/i) {
            # PSTAMP Output can look like this:
            #
            # 6.0
            # PSTAMP:  6.0.000.000-6.0-2011-08-01
            #
            # 5.1SP1
            # PSTAMP:  5.1.100.000-5.1SP1-2010-10-10
            #
            # 5.1 P1
            # PSTAMP:  Veritas-5.1-P1.1:2009-11-24
            #
            # 5.1 RP1
            # PSTAMP:  5.1.001.000-5.1RP1_x86-2010-02-23-VERITAS-VM
            # PSTAMP:  5.1.001.000-5.1RP1-2010-02-23-VERITAS-VM
            #
            # 5.0 MP3
            # PSTAMP:  VERITAS-5.0_x86_MP3RP2HF7:2009-09-10
            # PSTAMP:  Veritas-5.0_MP3_RP2HF7:2009-09-11
            #
            # 5.0 MP1
            # PSTAMP:  Veritas-5.0_MP1_RP3.2:2007-08-28

            # We need to normalize the PSTAMP line
            my $line_temp = $line;

            # Remove PSTAMP
            $line_temp =~ s/^\s*PSTAMP:\s+//g;

            # new "long" style versioning of D.D.DDD.DDD followed by dash or underscore
            $line_temp =~ s/^\s*(Veritas|\d+\.\d+\.\d+\.\d+)(\-|\_)//gi;

            # Remove any _x86 or -x86
            $line_temp =~ s/(\-|\_)x86//g;

            # Remove all date stamps
            $line_temp =~ s/\d\d\d\d\-\d\d\-\d\d//g;

            # Remove all remaining underscores and dashes
            $line_temp =~ s/\-|\_//g;

            # That should leave us with the following:
            # 5.0MP3RP2HF7:
            # 5.1RP1VERITASVM
            # 5.1P1.1:
            # 5.0MP1RP3.2:
            # 5.1SP1

            # Look up MP or P info
            (undef, $mp) = $line_temp =~ /^(\d+\.\d+)(MP\d+|P\d+).*/i;

            # Look for RP info
            (undef, $rp) = $line_temp =~ /^(\d+\.\d+)(RP\d+).*/i;

            # Look for SP info
            (undef, $sp) = $line_temp =~ /^(\d+\.\d+)(SP\d+).*/i;
        }
    }

    $self->{name}     = $name;
    $self->{revision} = $rev;
    $self->{version}  = $vers;
    $self->{mp}       = $mp;
    $self->{sp}       = $sp;
    $self->{rp}       = $rp;
    $self->{pkginfo}  = \@output;

    return $ClientUtil::TRUE;
}

################################################################################
# Subroutine Name: _is_vxdctl_running()
# Objective:       returns ClientUtil::TRUE or ClientUtil::FALSE depending on
#                  if vxdctl is running or not.
# Details:         See POD documentation at the end of this file
################################################################################
sub _is_vxdctl_running {
    my ($self, %args) = @_;

    my ($conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    # check status of vxdctl service
    my $cmd = "vxdctl mode";
    my ($status, @output) = &ClientUtil::execute($conn , $cmd);

    if (grep /enable/i, @output) {
        return $ClientUtil::TRUE;
    } else {
        return $ClientUtil::FALSE;
    }

}

################################################################################
# Subroutine Name: _rescan
# Objective:       To run the vxdctl command to rescan vxvm's list of
#                  disks that it thinks are under it's control.
# Details:         See POD documentation at the end of this file
################################################################################
sub _rescan {
    my ($self, %args) = @_;

    my ($conn, $cmd, @output, $status, $host_obj);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return $ClientUtil::FALSE;
            }
            next;
        }
    }

    $self->_clear_err();

    # need to run the vxdctl script after the new VTOC is down
    # to get MPXIO to see the disk again (properly) .. not completely
    # sure why
    $cmd  = "vxdctl enable";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    $cmd  = "vxdctl enable";
    ($status, @output) = &ClientUtil::execute($conn, $cmd);
    if ($status > 0) {
        $self->_set_err("Command ($cmd) returned a bad status ($status)");
        return $ClientUtil::FALSE;
    }

    return $ClientUtil::TRUE;

}

################################################################################
# Subroutine Name: _dmp_version
# Objective:       To version and revision of VXVM package installed
# Details:         See POD documentation at the end of this file
################################################################################
sub _dmp_version {
    my ($self, %args) = @_;

    my ($rev, $vers, $cmd, @output, $status, $conn);

    # parse arguments
    foreach (keys %args) {
        if (/-?conn/) {
            if (not length($conn = $args{$_})) {
                $self->_set_err("Invalid value for argument '-conn'");
                return ("", "");
            }
            next;
        }
    }

    $self->_clear_err();

    $cmd = "pkginfo -l VRTSvxvm";
    ($status, @output) = &ClientUtil::execute($conn , $cmd);
    if ($status > 0) {
        $self->_set_err("Failed to execute command $cmd");
        return ("", "");
    }

    foreach my $line (@output) {
        if ($line =~ /VERSION:\s+(.*),REV=\b(.*)\b/i) {
            $vers = $1;
            $rev = $2;
            last;
        }
    }

    if ($rev eq "" and $vers eq "") {
        $self->_set_err("Failed to determine version and revision of Veritas");
        return ("", "");
    }

    return ($vers, $rev);
}

1;
